
#define init_0_0_0_8
void FASTCALL call_0_0_0_8(SH2_struct* sh)
{
  SH2clrt(sh);
}
#define init_0_0_0_9
void FASTCALL call_0_0_0_9(SH2_struct* sh)
{
  SH2nop(sh);
}
#define init_0_0_0_11
void FASTCALL call_0_0_0_11(SH2_struct* sh)
{
  SH2rts(sh);
}
#define init_0_0_0_3
void FASTCALL call_0_0_0_3(SH2_struct* sh)
{
  SH2bsrf(sh,0);
}
#define init_0_1_0_3
void FASTCALL call_0_1_0_3(SH2_struct* sh)
{
  SH2bsrf(sh,1);
}
#define init_0_2_0_3
void FASTCALL call_0_2_0_3(SH2_struct* sh)
{
  SH2bsrf(sh,2);
}
#define init_0_3_0_3
void FASTCALL call_0_3_0_3(SH2_struct* sh)
{
  SH2bsrf(sh,3);
}
#define init_0_4_0_3
void FASTCALL call_0_4_0_3(SH2_struct* sh)
{
  SH2bsrf(sh,4);
}
#define init_0_5_0_3
void FASTCALL call_0_5_0_3(SH2_struct* sh)
{
  SH2bsrf(sh,5);
}
#define init_0_6_0_3
void FASTCALL call_0_6_0_3(SH2_struct* sh)
{
  SH2bsrf(sh,6);
}
#define init_0_7_0_3
void FASTCALL call_0_7_0_3(SH2_struct* sh)
{
  SH2bsrf(sh,7);
}
#define init_0_8_0_3
void FASTCALL call_0_8_0_3(SH2_struct* sh)
{
  SH2bsrf(sh,8);
}
#define init_0_9_0_3
void FASTCALL call_0_9_0_3(SH2_struct* sh)
{
  SH2bsrf(sh,9);
}
#define init_0_10_0_3
void FASTCALL call_0_10_0_3(SH2_struct* sh)
{
  SH2bsrf(sh,10);
}
#define init_0_11_0_3
void FASTCALL call_0_11_0_3(SH2_struct* sh)
{
  SH2bsrf(sh,11);
}
#define init_0_12_0_3
void FASTCALL call_0_12_0_3(SH2_struct* sh)
{
  SH2bsrf(sh,12);
}
#define init_0_13_0_3
void FASTCALL call_0_13_0_3(SH2_struct* sh)
{
  SH2bsrf(sh,13);
}
#define init_0_14_0_3
void FASTCALL call_0_14_0_3(SH2_struct* sh)
{
  SH2bsrf(sh,14);
}
#define init_0_15_0_3
void FASTCALL call_0_15_0_3(SH2_struct* sh)
{
  SH2bsrf(sh,15);
}
#define init_0_0_2_3
void FASTCALL call_0_0_2_3(SH2_struct* sh)
{
  SH2braf(sh,0);
}
#define init_0_1_2_3
void FASTCALL call_0_1_2_3(SH2_struct* sh)
{
  SH2braf(sh,1);
}
#define init_0_2_2_3
void FASTCALL call_0_2_2_3(SH2_struct* sh)
{
  SH2braf(sh,2);
}
#define init_0_3_2_3
void FASTCALL call_0_3_2_3(SH2_struct* sh)
{
  SH2braf(sh,3);
}
#define init_0_4_2_3
void FASTCALL call_0_4_2_3(SH2_struct* sh)
{
  SH2braf(sh,4);
}
#define init_0_5_2_3
void FASTCALL call_0_5_2_3(SH2_struct* sh)
{
  SH2braf(sh,5);
}
#define init_0_6_2_3
void FASTCALL call_0_6_2_3(SH2_struct* sh)
{
  SH2braf(sh,6);
}
#define init_0_7_2_3
void FASTCALL call_0_7_2_3(SH2_struct* sh)
{
  SH2braf(sh,7);
}
#define init_0_8_2_3
void FASTCALL call_0_8_2_3(SH2_struct* sh)
{
  SH2braf(sh,8);
}
#define init_0_9_2_3
void FASTCALL call_0_9_2_3(SH2_struct* sh)
{
  SH2braf(sh,9);
}
#define init_0_10_2_3
void FASTCALL call_0_10_2_3(SH2_struct* sh)
{
  SH2braf(sh,10);
}
#define init_0_11_2_3
void FASTCALL call_0_11_2_3(SH2_struct* sh)
{
  SH2braf(sh,11);
}
#define init_0_12_2_3
void FASTCALL call_0_12_2_3(SH2_struct* sh)
{
  SH2braf(sh,12);
}
#define init_0_13_2_3
void FASTCALL call_0_13_2_3(SH2_struct* sh)
{
  SH2braf(sh,13);
}
#define init_0_14_2_3
void FASTCALL call_0_14_2_3(SH2_struct* sh)
{
  SH2braf(sh,14);
}
#define init_0_15_2_3
void FASTCALL call_0_15_2_3(SH2_struct* sh)
{
  SH2braf(sh,15);
}
#define init_0_0_2_8
void FASTCALL call_0_0_2_8(SH2_struct* sh)
{
  SH2clrmac(sh);
}
#define init_0_0_1_9
void FASTCALL call_0_0_1_9(SH2_struct* sh)
{
  SH2div0u(sh);
}
#define init_0_0_1_11
void FASTCALL call_0_0_1_11(SH2_struct* sh)
{
  SH2sleep(sh);
}
#define init_0_0_2_10
void FASTCALL call_0_0_2_10(SH2_struct* sh)
{
  SH2stspr(sh,0);
}
#define init_0_1_2_10
void FASTCALL call_0_1_2_10(SH2_struct* sh)
{
  SH2stspr(sh,1);
}
#define init_0_2_2_10
void FASTCALL call_0_2_2_10(SH2_struct* sh)
{
  SH2stspr(sh,2);
}
#define init_0_3_2_10
void FASTCALL call_0_3_2_10(SH2_struct* sh)
{
  SH2stspr(sh,3);
}
#define init_0_4_2_10
void FASTCALL call_0_4_2_10(SH2_struct* sh)
{
  SH2stspr(sh,4);
}
#define init_0_5_2_10
void FASTCALL call_0_5_2_10(SH2_struct* sh)
{
  SH2stspr(sh,5);
}
#define init_0_6_2_10
void FASTCALL call_0_6_2_10(SH2_struct* sh)
{
  SH2stspr(sh,6);
}
#define init_0_7_2_10
void FASTCALL call_0_7_2_10(SH2_struct* sh)
{
  SH2stspr(sh,7);
}
#define init_0_8_2_10
void FASTCALL call_0_8_2_10(SH2_struct* sh)
{
  SH2stspr(sh,8);
}
#define init_0_9_2_10
void FASTCALL call_0_9_2_10(SH2_struct* sh)
{
  SH2stspr(sh,9);
}
#define init_0_10_2_10
void FASTCALL call_0_10_2_10(SH2_struct* sh)
{
  SH2stspr(sh,10);
}
#define init_0_11_2_10
void FASTCALL call_0_11_2_10(SH2_struct* sh)
{
  SH2stspr(sh,11);
}
#define init_0_12_2_10
void FASTCALL call_0_12_2_10(SH2_struct* sh)
{
  SH2stspr(sh,12);
}
#define init_0_13_2_10
void FASTCALL call_0_13_2_10(SH2_struct* sh)
{
  SH2stspr(sh,13);
}
#define init_0_14_2_10
void FASTCALL call_0_14_2_10(SH2_struct* sh)
{
  SH2stspr(sh,14);
}
#define init_0_15_2_10
void FASTCALL call_0_15_2_10(SH2_struct* sh)
{
  SH2stspr(sh,15);
}
#define init_0_0_1_10
void FASTCALL call_0_0_1_10(SH2_struct* sh)
{
  SH2stsmacl(sh,0);
}
#define init_0_1_1_10
void FASTCALL call_0_1_1_10(SH2_struct* sh)
{
  SH2stsmacl(sh,1);
}
#define init_0_2_1_10
void FASTCALL call_0_2_1_10(SH2_struct* sh)
{
  SH2stsmacl(sh,2);
}
#define init_0_3_1_10
void FASTCALL call_0_3_1_10(SH2_struct* sh)
{
  SH2stsmacl(sh,3);
}
#define init_0_4_1_10
void FASTCALL call_0_4_1_10(SH2_struct* sh)
{
  SH2stsmacl(sh,4);
}
#define init_0_5_1_10
void FASTCALL call_0_5_1_10(SH2_struct* sh)
{
  SH2stsmacl(sh,5);
}
#define init_0_6_1_10
void FASTCALL call_0_6_1_10(SH2_struct* sh)
{
  SH2stsmacl(sh,6);
}
#define init_0_7_1_10
void FASTCALL call_0_7_1_10(SH2_struct* sh)
{
  SH2stsmacl(sh,7);
}
#define init_0_8_1_10
void FASTCALL call_0_8_1_10(SH2_struct* sh)
{
  SH2stsmacl(sh,8);
}
#define init_0_9_1_10
void FASTCALL call_0_9_1_10(SH2_struct* sh)
{
  SH2stsmacl(sh,9);
}
#define init_0_10_1_10
void FASTCALL call_0_10_1_10(SH2_struct* sh)
{
  SH2stsmacl(sh,10);
}
#define init_0_11_1_10
void FASTCALL call_0_11_1_10(SH2_struct* sh)
{
  SH2stsmacl(sh,11);
}
#define init_0_12_1_10
void FASTCALL call_0_12_1_10(SH2_struct* sh)
{
  SH2stsmacl(sh,12);
}
#define init_0_13_1_10
void FASTCALL call_0_13_1_10(SH2_struct* sh)
{
  SH2stsmacl(sh,13);
}
#define init_0_14_1_10
void FASTCALL call_0_14_1_10(SH2_struct* sh)
{
  SH2stsmacl(sh,14);
}
#define init_0_15_1_10
void FASTCALL call_0_15_1_10(SH2_struct* sh)
{
  SH2stsmacl(sh,15);
}
#define init_0_0_0_10
void FASTCALL call_0_0_0_10(SH2_struct* sh)
{
  SH2stsmach(sh,0);
}
#define init_0_1_0_10
void FASTCALL call_0_1_0_10(SH2_struct* sh)
{
  SH2stsmach(sh,1);
}
#define init_0_2_0_10
void FASTCALL call_0_2_0_10(SH2_struct* sh)
{
  SH2stsmach(sh,2);
}
#define init_0_3_0_10
void FASTCALL call_0_3_0_10(SH2_struct* sh)
{
  SH2stsmach(sh,3);
}
#define init_0_4_0_10
void FASTCALL call_0_4_0_10(SH2_struct* sh)
{
  SH2stsmach(sh,4);
}
#define init_0_5_0_10
void FASTCALL call_0_5_0_10(SH2_struct* sh)
{
  SH2stsmach(sh,5);
}
#define init_0_6_0_10
void FASTCALL call_0_6_0_10(SH2_struct* sh)
{
  SH2stsmach(sh,6);
}
#define init_0_7_0_10
void FASTCALL call_0_7_0_10(SH2_struct* sh)
{
  SH2stsmach(sh,7);
}
#define init_0_8_0_10
void FASTCALL call_0_8_0_10(SH2_struct* sh)
{
  SH2stsmach(sh,8);
}
#define init_0_9_0_10
void FASTCALL call_0_9_0_10(SH2_struct* sh)
{
  SH2stsmach(sh,9);
}
#define init_0_10_0_10
void FASTCALL call_0_10_0_10(SH2_struct* sh)
{
  SH2stsmach(sh,10);
}
#define init_0_11_0_10
void FASTCALL call_0_11_0_10(SH2_struct* sh)
{
  SH2stsmach(sh,11);
}
#define init_0_12_0_10
void FASTCALL call_0_12_0_10(SH2_struct* sh)
{
  SH2stsmach(sh,12);
}
#define init_0_13_0_10
void FASTCALL call_0_13_0_10(SH2_struct* sh)
{
  SH2stsmach(sh,13);
}
#define init_0_14_0_10
void FASTCALL call_0_14_0_10(SH2_struct* sh)
{
  SH2stsmach(sh,14);
}
#define init_0_15_0_10
void FASTCALL call_0_15_0_10(SH2_struct* sh)
{
  SH2stsmach(sh,15);
}
#define init_0_0_2_2
void FASTCALL call_0_0_2_2(SH2_struct* sh)
{
  SH2stcvbr(sh,0);
}
#define init_0_1_2_2
void FASTCALL call_0_1_2_2(SH2_struct* sh)
{
  SH2stcvbr(sh,1);
}
#define init_0_2_2_2
void FASTCALL call_0_2_2_2(SH2_struct* sh)
{
  SH2stcvbr(sh,2);
}
#define init_0_3_2_2
void FASTCALL call_0_3_2_2(SH2_struct* sh)
{
  SH2stcvbr(sh,3);
}
#define init_0_4_2_2
void FASTCALL call_0_4_2_2(SH2_struct* sh)
{
  SH2stcvbr(sh,4);
}
#define init_0_5_2_2
void FASTCALL call_0_5_2_2(SH2_struct* sh)
{
  SH2stcvbr(sh,5);
}
#define init_0_6_2_2
void FASTCALL call_0_6_2_2(SH2_struct* sh)
{
  SH2stcvbr(sh,6);
}
#define init_0_7_2_2
void FASTCALL call_0_7_2_2(SH2_struct* sh)
{
  SH2stcvbr(sh,7);
}
#define init_0_8_2_2
void FASTCALL call_0_8_2_2(SH2_struct* sh)
{
  SH2stcvbr(sh,8);
}
#define init_0_9_2_2
void FASTCALL call_0_9_2_2(SH2_struct* sh)
{
  SH2stcvbr(sh,9);
}
#define init_0_10_2_2
void FASTCALL call_0_10_2_2(SH2_struct* sh)
{
  SH2stcvbr(sh,10);
}
#define init_0_11_2_2
void FASTCALL call_0_11_2_2(SH2_struct* sh)
{
  SH2stcvbr(sh,11);
}
#define init_0_12_2_2
void FASTCALL call_0_12_2_2(SH2_struct* sh)
{
  SH2stcvbr(sh,12);
}
#define init_0_13_2_2
void FASTCALL call_0_13_2_2(SH2_struct* sh)
{
  SH2stcvbr(sh,13);
}
#define init_0_14_2_2
void FASTCALL call_0_14_2_2(SH2_struct* sh)
{
  SH2stcvbr(sh,14);
}
#define init_0_15_2_2
void FASTCALL call_0_15_2_2(SH2_struct* sh)
{
  SH2stcvbr(sh,15);
}
#define init_0_0_0_2
void FASTCALL call_0_0_0_2(SH2_struct* sh)
{
  SH2stcsr(sh,0);
}
#define init_0_1_0_2
void FASTCALL call_0_1_0_2(SH2_struct* sh)
{
  SH2stcsr(sh,1);
}
#define init_0_2_0_2
void FASTCALL call_0_2_0_2(SH2_struct* sh)
{
  SH2stcsr(sh,2);
}
#define init_0_3_0_2
void FASTCALL call_0_3_0_2(SH2_struct* sh)
{
  SH2stcsr(sh,3);
}
#define init_0_4_0_2
void FASTCALL call_0_4_0_2(SH2_struct* sh)
{
  SH2stcsr(sh,4);
}
#define init_0_5_0_2
void FASTCALL call_0_5_0_2(SH2_struct* sh)
{
  SH2stcsr(sh,5);
}
#define init_0_6_0_2
void FASTCALL call_0_6_0_2(SH2_struct* sh)
{
  SH2stcsr(sh,6);
}
#define init_0_7_0_2
void FASTCALL call_0_7_0_2(SH2_struct* sh)
{
  SH2stcsr(sh,7);
}
#define init_0_8_0_2
void FASTCALL call_0_8_0_2(SH2_struct* sh)
{
  SH2stcsr(sh,8);
}
#define init_0_9_0_2
void FASTCALL call_0_9_0_2(SH2_struct* sh)
{
  SH2stcsr(sh,9);
}
#define init_0_10_0_2
void FASTCALL call_0_10_0_2(SH2_struct* sh)
{
  SH2stcsr(sh,10);
}
#define init_0_11_0_2
void FASTCALL call_0_11_0_2(SH2_struct* sh)
{
  SH2stcsr(sh,11);
}
#define init_0_12_0_2
void FASTCALL call_0_12_0_2(SH2_struct* sh)
{
  SH2stcsr(sh,12);
}
#define init_0_13_0_2
void FASTCALL call_0_13_0_2(SH2_struct* sh)
{
  SH2stcsr(sh,13);
}
#define init_0_14_0_2
void FASTCALL call_0_14_0_2(SH2_struct* sh)
{
  SH2stcsr(sh,14);
}
#define init_0_15_0_2
void FASTCALL call_0_15_0_2(SH2_struct* sh)
{
  SH2stcsr(sh,15);
}
#define init_0_0_1_2
void FASTCALL call_0_0_1_2(SH2_struct* sh)
{
  SH2stcgbr(sh,0);
}
#define init_0_1_1_2
void FASTCALL call_0_1_1_2(SH2_struct* sh)
{
  SH2stcgbr(sh,1);
}
#define init_0_2_1_2
void FASTCALL call_0_2_1_2(SH2_struct* sh)
{
  SH2stcgbr(sh,2);
}
#define init_0_3_1_2
void FASTCALL call_0_3_1_2(SH2_struct* sh)
{
  SH2stcgbr(sh,3);
}
#define init_0_4_1_2
void FASTCALL call_0_4_1_2(SH2_struct* sh)
{
  SH2stcgbr(sh,4);
}
#define init_0_5_1_2
void FASTCALL call_0_5_1_2(SH2_struct* sh)
{
  SH2stcgbr(sh,5);
}
#define init_0_6_1_2
void FASTCALL call_0_6_1_2(SH2_struct* sh)
{
  SH2stcgbr(sh,6);
}
#define init_0_7_1_2
void FASTCALL call_0_7_1_2(SH2_struct* sh)
{
  SH2stcgbr(sh,7);
}
#define init_0_8_1_2
void FASTCALL call_0_8_1_2(SH2_struct* sh)
{
  SH2stcgbr(sh,8);
}
#define init_0_9_1_2
void FASTCALL call_0_9_1_2(SH2_struct* sh)
{
  SH2stcgbr(sh,9);
}
#define init_0_10_1_2
void FASTCALL call_0_10_1_2(SH2_struct* sh)
{
  SH2stcgbr(sh,10);
}
#define init_0_11_1_2
void FASTCALL call_0_11_1_2(SH2_struct* sh)
{
  SH2stcgbr(sh,11);
}
#define init_0_12_1_2
void FASTCALL call_0_12_1_2(SH2_struct* sh)
{
  SH2stcgbr(sh,12);
}
#define init_0_13_1_2
void FASTCALL call_0_13_1_2(SH2_struct* sh)
{
  SH2stcgbr(sh,13);
}
#define init_0_14_1_2
void FASTCALL call_0_14_1_2(SH2_struct* sh)
{
  SH2stcgbr(sh,14);
}
#define init_0_15_1_2
void FASTCALL call_0_15_1_2(SH2_struct* sh)
{
  SH2stcgbr(sh,15);
}
#define init_0_0_1_8
void FASTCALL call_0_0_1_8(SH2_struct* sh)
{
  SH2sett(sh);
}
#define init_0_0_2_11
void FASTCALL call_0_0_2_11(SH2_struct* sh)
{
  SH2rte(sh);
}
#define init_0_0_0_7
void FASTCALL call_0_0_0_7(SH2_struct* sh)
{
  SH2mull(sh,0,0);
}
#define init_0_0_1_7
void FASTCALL call_0_0_1_7(SH2_struct* sh)
{
  SH2mull(sh,0,1);
}
#define init_0_0_2_7
void FASTCALL call_0_0_2_7(SH2_struct* sh)
{
  SH2mull(sh,0,2);
}
#define init_0_0_3_7
void FASTCALL call_0_0_3_7(SH2_struct* sh)
{
  SH2mull(sh,0,3);
}
#define init_0_0_4_7
void FASTCALL call_0_0_4_7(SH2_struct* sh)
{
  SH2mull(sh,0,4);
}
#define init_0_0_5_7
void FASTCALL call_0_0_5_7(SH2_struct* sh)
{
  SH2mull(sh,0,5);
}
#define init_0_0_6_7
void FASTCALL call_0_0_6_7(SH2_struct* sh)
{
  SH2mull(sh,0,6);
}
#define init_0_0_7_7
void FASTCALL call_0_0_7_7(SH2_struct* sh)
{
  SH2mull(sh,0,7);
}
#define init_0_0_8_7
void FASTCALL call_0_0_8_7(SH2_struct* sh)
{
  SH2mull(sh,0,8);
}
#define init_0_0_9_7
void FASTCALL call_0_0_9_7(SH2_struct* sh)
{
  SH2mull(sh,0,9);
}
#define init_0_0_10_7
void FASTCALL call_0_0_10_7(SH2_struct* sh)
{
  SH2mull(sh,0,10);
}
#define init_0_0_11_7
void FASTCALL call_0_0_11_7(SH2_struct* sh)
{
  SH2mull(sh,0,11);
}
#define init_0_0_12_7
void FASTCALL call_0_0_12_7(SH2_struct* sh)
{
  SH2mull(sh,0,12);
}
#define init_0_0_13_7
void FASTCALL call_0_0_13_7(SH2_struct* sh)
{
  SH2mull(sh,0,13);
}
#define init_0_0_14_7
void FASTCALL call_0_0_14_7(SH2_struct* sh)
{
  SH2mull(sh,0,14);
}
#define init_0_0_15_7
void FASTCALL call_0_0_15_7(SH2_struct* sh)
{
  SH2mull(sh,0,15);
}
#define init_0_1_0_7
void FASTCALL call_0_1_0_7(SH2_struct* sh)
{
  SH2mull(sh,1,0);
}
#define init_0_1_1_7
void FASTCALL call_0_1_1_7(SH2_struct* sh)
{
  SH2mull(sh,1,1);
}
#define init_0_1_2_7
void FASTCALL call_0_1_2_7(SH2_struct* sh)
{
  SH2mull(sh,1,2);
}
#define init_0_1_3_7
void FASTCALL call_0_1_3_7(SH2_struct* sh)
{
  SH2mull(sh,1,3);
}
#define init_0_1_4_7
void FASTCALL call_0_1_4_7(SH2_struct* sh)
{
  SH2mull(sh,1,4);
}
#define init_0_1_5_7
void FASTCALL call_0_1_5_7(SH2_struct* sh)
{
  SH2mull(sh,1,5);
}
#define init_0_1_6_7
void FASTCALL call_0_1_6_7(SH2_struct* sh)
{
  SH2mull(sh,1,6);
}
#define init_0_1_7_7
void FASTCALL call_0_1_7_7(SH2_struct* sh)
{
  SH2mull(sh,1,7);
}
#define init_0_1_8_7
void FASTCALL call_0_1_8_7(SH2_struct* sh)
{
  SH2mull(sh,1,8);
}
#define init_0_1_9_7
void FASTCALL call_0_1_9_7(SH2_struct* sh)
{
  SH2mull(sh,1,9);
}
#define init_0_1_10_7
void FASTCALL call_0_1_10_7(SH2_struct* sh)
{
  SH2mull(sh,1,10);
}
#define init_0_1_11_7
void FASTCALL call_0_1_11_7(SH2_struct* sh)
{
  SH2mull(sh,1,11);
}
#define init_0_1_12_7
void FASTCALL call_0_1_12_7(SH2_struct* sh)
{
  SH2mull(sh,1,12);
}
#define init_0_1_13_7
void FASTCALL call_0_1_13_7(SH2_struct* sh)
{
  SH2mull(sh,1,13);
}
#define init_0_1_14_7
void FASTCALL call_0_1_14_7(SH2_struct* sh)
{
  SH2mull(sh,1,14);
}
#define init_0_1_15_7
void FASTCALL call_0_1_15_7(SH2_struct* sh)
{
  SH2mull(sh,1,15);
}
#define init_0_2_0_7
void FASTCALL call_0_2_0_7(SH2_struct* sh)
{
  SH2mull(sh,2,0);
}
#define init_0_2_1_7
void FASTCALL call_0_2_1_7(SH2_struct* sh)
{
  SH2mull(sh,2,1);
}
#define init_0_2_2_7
void FASTCALL call_0_2_2_7(SH2_struct* sh)
{
  SH2mull(sh,2,2);
}
#define init_0_2_3_7
void FASTCALL call_0_2_3_7(SH2_struct* sh)
{
  SH2mull(sh,2,3);
}
#define init_0_2_4_7
void FASTCALL call_0_2_4_7(SH2_struct* sh)
{
  SH2mull(sh,2,4);
}
#define init_0_2_5_7
void FASTCALL call_0_2_5_7(SH2_struct* sh)
{
  SH2mull(sh,2,5);
}
#define init_0_2_6_7
void FASTCALL call_0_2_6_7(SH2_struct* sh)
{
  SH2mull(sh,2,6);
}
#define init_0_2_7_7
void FASTCALL call_0_2_7_7(SH2_struct* sh)
{
  SH2mull(sh,2,7);
}
#define init_0_2_8_7
void FASTCALL call_0_2_8_7(SH2_struct* sh)
{
  SH2mull(sh,2,8);
}
#define init_0_2_9_7
void FASTCALL call_0_2_9_7(SH2_struct* sh)
{
  SH2mull(sh,2,9);
}
#define init_0_2_10_7
void FASTCALL call_0_2_10_7(SH2_struct* sh)
{
  SH2mull(sh,2,10);
}
#define init_0_2_11_7
void FASTCALL call_0_2_11_7(SH2_struct* sh)
{
  SH2mull(sh,2,11);
}
#define init_0_2_12_7
void FASTCALL call_0_2_12_7(SH2_struct* sh)
{
  SH2mull(sh,2,12);
}
#define init_0_2_13_7
void FASTCALL call_0_2_13_7(SH2_struct* sh)
{
  SH2mull(sh,2,13);
}
#define init_0_2_14_7
void FASTCALL call_0_2_14_7(SH2_struct* sh)
{
  SH2mull(sh,2,14);
}
#define init_0_2_15_7
void FASTCALL call_0_2_15_7(SH2_struct* sh)
{
  SH2mull(sh,2,15);
}
#define init_0_3_0_7
void FASTCALL call_0_3_0_7(SH2_struct* sh)
{
  SH2mull(sh,3,0);
}
#define init_0_3_1_7
void FASTCALL call_0_3_1_7(SH2_struct* sh)
{
  SH2mull(sh,3,1);
}
#define init_0_3_2_7
void FASTCALL call_0_3_2_7(SH2_struct* sh)
{
  SH2mull(sh,3,2);
}
#define init_0_3_3_7
void FASTCALL call_0_3_3_7(SH2_struct* sh)
{
  SH2mull(sh,3,3);
}
#define init_0_3_4_7
void FASTCALL call_0_3_4_7(SH2_struct* sh)
{
  SH2mull(sh,3,4);
}
#define init_0_3_5_7
void FASTCALL call_0_3_5_7(SH2_struct* sh)
{
  SH2mull(sh,3,5);
}
#define init_0_3_6_7
void FASTCALL call_0_3_6_7(SH2_struct* sh)
{
  SH2mull(sh,3,6);
}
#define init_0_3_7_7
void FASTCALL call_0_3_7_7(SH2_struct* sh)
{
  SH2mull(sh,3,7);
}
#define init_0_3_8_7
void FASTCALL call_0_3_8_7(SH2_struct* sh)
{
  SH2mull(sh,3,8);
}
#define init_0_3_9_7
void FASTCALL call_0_3_9_7(SH2_struct* sh)
{
  SH2mull(sh,3,9);
}
#define init_0_3_10_7
void FASTCALL call_0_3_10_7(SH2_struct* sh)
{
  SH2mull(sh,3,10);
}
#define init_0_3_11_7
void FASTCALL call_0_3_11_7(SH2_struct* sh)
{
  SH2mull(sh,3,11);
}
#define init_0_3_12_7
void FASTCALL call_0_3_12_7(SH2_struct* sh)
{
  SH2mull(sh,3,12);
}
#define init_0_3_13_7
void FASTCALL call_0_3_13_7(SH2_struct* sh)
{
  SH2mull(sh,3,13);
}
#define init_0_3_14_7
void FASTCALL call_0_3_14_7(SH2_struct* sh)
{
  SH2mull(sh,3,14);
}
#define init_0_3_15_7
void FASTCALL call_0_3_15_7(SH2_struct* sh)
{
  SH2mull(sh,3,15);
}
#define init_0_4_0_7
void FASTCALL call_0_4_0_7(SH2_struct* sh)
{
  SH2mull(sh,4,0);
}
#define init_0_4_1_7
void FASTCALL call_0_4_1_7(SH2_struct* sh)
{
  SH2mull(sh,4,1);
}
#define init_0_4_2_7
void FASTCALL call_0_4_2_7(SH2_struct* sh)
{
  SH2mull(sh,4,2);
}
#define init_0_4_3_7
void FASTCALL call_0_4_3_7(SH2_struct* sh)
{
  SH2mull(sh,4,3);
}
#define init_0_4_4_7
void FASTCALL call_0_4_4_7(SH2_struct* sh)
{
  SH2mull(sh,4,4);
}
#define init_0_4_5_7
void FASTCALL call_0_4_5_7(SH2_struct* sh)
{
  SH2mull(sh,4,5);
}
#define init_0_4_6_7
void FASTCALL call_0_4_6_7(SH2_struct* sh)
{
  SH2mull(sh,4,6);
}
#define init_0_4_7_7
void FASTCALL call_0_4_7_7(SH2_struct* sh)
{
  SH2mull(sh,4,7);
}
#define init_0_4_8_7
void FASTCALL call_0_4_8_7(SH2_struct* sh)
{
  SH2mull(sh,4,8);
}
#define init_0_4_9_7
void FASTCALL call_0_4_9_7(SH2_struct* sh)
{
  SH2mull(sh,4,9);
}
#define init_0_4_10_7
void FASTCALL call_0_4_10_7(SH2_struct* sh)
{
  SH2mull(sh,4,10);
}
#define init_0_4_11_7
void FASTCALL call_0_4_11_7(SH2_struct* sh)
{
  SH2mull(sh,4,11);
}
#define init_0_4_12_7
void FASTCALL call_0_4_12_7(SH2_struct* sh)
{
  SH2mull(sh,4,12);
}
#define init_0_4_13_7
void FASTCALL call_0_4_13_7(SH2_struct* sh)
{
  SH2mull(sh,4,13);
}
#define init_0_4_14_7
void FASTCALL call_0_4_14_7(SH2_struct* sh)
{
  SH2mull(sh,4,14);
}
#define init_0_4_15_7
void FASTCALL call_0_4_15_7(SH2_struct* sh)
{
  SH2mull(sh,4,15);
}
#define init_0_5_0_7
void FASTCALL call_0_5_0_7(SH2_struct* sh)
{
  SH2mull(sh,5,0);
}
#define init_0_5_1_7
void FASTCALL call_0_5_1_7(SH2_struct* sh)
{
  SH2mull(sh,5,1);
}
#define init_0_5_2_7
void FASTCALL call_0_5_2_7(SH2_struct* sh)
{
  SH2mull(sh,5,2);
}
#define init_0_5_3_7
void FASTCALL call_0_5_3_7(SH2_struct* sh)
{
  SH2mull(sh,5,3);
}
#define init_0_5_4_7
void FASTCALL call_0_5_4_7(SH2_struct* sh)
{
  SH2mull(sh,5,4);
}
#define init_0_5_5_7
void FASTCALL call_0_5_5_7(SH2_struct* sh)
{
  SH2mull(sh,5,5);
}
#define init_0_5_6_7
void FASTCALL call_0_5_6_7(SH2_struct* sh)
{
  SH2mull(sh,5,6);
}
#define init_0_5_7_7
void FASTCALL call_0_5_7_7(SH2_struct* sh)
{
  SH2mull(sh,5,7);
}
#define init_0_5_8_7
void FASTCALL call_0_5_8_7(SH2_struct* sh)
{
  SH2mull(sh,5,8);
}
#define init_0_5_9_7
void FASTCALL call_0_5_9_7(SH2_struct* sh)
{
  SH2mull(sh,5,9);
}
#define init_0_5_10_7
void FASTCALL call_0_5_10_7(SH2_struct* sh)
{
  SH2mull(sh,5,10);
}
#define init_0_5_11_7
void FASTCALL call_0_5_11_7(SH2_struct* sh)
{
  SH2mull(sh,5,11);
}
#define init_0_5_12_7
void FASTCALL call_0_5_12_7(SH2_struct* sh)
{
  SH2mull(sh,5,12);
}
#define init_0_5_13_7
void FASTCALL call_0_5_13_7(SH2_struct* sh)
{
  SH2mull(sh,5,13);
}
#define init_0_5_14_7
void FASTCALL call_0_5_14_7(SH2_struct* sh)
{
  SH2mull(sh,5,14);
}
#define init_0_5_15_7
void FASTCALL call_0_5_15_7(SH2_struct* sh)
{
  SH2mull(sh,5,15);
}
#define init_0_6_0_7
void FASTCALL call_0_6_0_7(SH2_struct* sh)
{
  SH2mull(sh,6,0);
}
#define init_0_6_1_7
void FASTCALL call_0_6_1_7(SH2_struct* sh)
{
  SH2mull(sh,6,1);
}
#define init_0_6_2_7
void FASTCALL call_0_6_2_7(SH2_struct* sh)
{
  SH2mull(sh,6,2);
}
#define init_0_6_3_7
void FASTCALL call_0_6_3_7(SH2_struct* sh)
{
  SH2mull(sh,6,3);
}
#define init_0_6_4_7
void FASTCALL call_0_6_4_7(SH2_struct* sh)
{
  SH2mull(sh,6,4);
}
#define init_0_6_5_7
void FASTCALL call_0_6_5_7(SH2_struct* sh)
{
  SH2mull(sh,6,5);
}
#define init_0_6_6_7
void FASTCALL call_0_6_6_7(SH2_struct* sh)
{
  SH2mull(sh,6,6);
}
#define init_0_6_7_7
void FASTCALL call_0_6_7_7(SH2_struct* sh)
{
  SH2mull(sh,6,7);
}
#define init_0_6_8_7
void FASTCALL call_0_6_8_7(SH2_struct* sh)
{
  SH2mull(sh,6,8);
}
#define init_0_6_9_7
void FASTCALL call_0_6_9_7(SH2_struct* sh)
{
  SH2mull(sh,6,9);
}
#define init_0_6_10_7
void FASTCALL call_0_6_10_7(SH2_struct* sh)
{
  SH2mull(sh,6,10);
}
#define init_0_6_11_7
void FASTCALL call_0_6_11_7(SH2_struct* sh)
{
  SH2mull(sh,6,11);
}
#define init_0_6_12_7
void FASTCALL call_0_6_12_7(SH2_struct* sh)
{
  SH2mull(sh,6,12);
}
#define init_0_6_13_7
void FASTCALL call_0_6_13_7(SH2_struct* sh)
{
  SH2mull(sh,6,13);
}
#define init_0_6_14_7
void FASTCALL call_0_6_14_7(SH2_struct* sh)
{
  SH2mull(sh,6,14);
}
#define init_0_6_15_7
void FASTCALL call_0_6_15_7(SH2_struct* sh)
{
  SH2mull(sh,6,15);
}
#define init_0_7_0_7
void FASTCALL call_0_7_0_7(SH2_struct* sh)
{
  SH2mull(sh,7,0);
}
#define init_0_7_1_7
void FASTCALL call_0_7_1_7(SH2_struct* sh)
{
  SH2mull(sh,7,1);
}
#define init_0_7_2_7
void FASTCALL call_0_7_2_7(SH2_struct* sh)
{
  SH2mull(sh,7,2);
}
#define init_0_7_3_7
void FASTCALL call_0_7_3_7(SH2_struct* sh)
{
  SH2mull(sh,7,3);
}
#define init_0_7_4_7
void FASTCALL call_0_7_4_7(SH2_struct* sh)
{
  SH2mull(sh,7,4);
}
#define init_0_7_5_7
void FASTCALL call_0_7_5_7(SH2_struct* sh)
{
  SH2mull(sh,7,5);
}
#define init_0_7_6_7
void FASTCALL call_0_7_6_7(SH2_struct* sh)
{
  SH2mull(sh,7,6);
}
#define init_0_7_7_7
void FASTCALL call_0_7_7_7(SH2_struct* sh)
{
  SH2mull(sh,7,7);
}
#define init_0_7_8_7
void FASTCALL call_0_7_8_7(SH2_struct* sh)
{
  SH2mull(sh,7,8);
}
#define init_0_7_9_7
void FASTCALL call_0_7_9_7(SH2_struct* sh)
{
  SH2mull(sh,7,9);
}
#define init_0_7_10_7
void FASTCALL call_0_7_10_7(SH2_struct* sh)
{
  SH2mull(sh,7,10);
}
#define init_0_7_11_7
void FASTCALL call_0_7_11_7(SH2_struct* sh)
{
  SH2mull(sh,7,11);
}
#define init_0_7_12_7
void FASTCALL call_0_7_12_7(SH2_struct* sh)
{
  SH2mull(sh,7,12);
}
#define init_0_7_13_7
void FASTCALL call_0_7_13_7(SH2_struct* sh)
{
  SH2mull(sh,7,13);
}
#define init_0_7_14_7
void FASTCALL call_0_7_14_7(SH2_struct* sh)
{
  SH2mull(sh,7,14);
}
#define init_0_7_15_7
void FASTCALL call_0_7_15_7(SH2_struct* sh)
{
  SH2mull(sh,7,15);
}
#define init_0_8_0_7
void FASTCALL call_0_8_0_7(SH2_struct* sh)
{
  SH2mull(sh,8,0);
}
#define init_0_8_1_7
void FASTCALL call_0_8_1_7(SH2_struct* sh)
{
  SH2mull(sh,8,1);
}
#define init_0_8_2_7
void FASTCALL call_0_8_2_7(SH2_struct* sh)
{
  SH2mull(sh,8,2);
}
#define init_0_8_3_7
void FASTCALL call_0_8_3_7(SH2_struct* sh)
{
  SH2mull(sh,8,3);
}
#define init_0_8_4_7
void FASTCALL call_0_8_4_7(SH2_struct* sh)
{
  SH2mull(sh,8,4);
}
#define init_0_8_5_7
void FASTCALL call_0_8_5_7(SH2_struct* sh)
{
  SH2mull(sh,8,5);
}
#define init_0_8_6_7
void FASTCALL call_0_8_6_7(SH2_struct* sh)
{
  SH2mull(sh,8,6);
}
#define init_0_8_7_7
void FASTCALL call_0_8_7_7(SH2_struct* sh)
{
  SH2mull(sh,8,7);
}
#define init_0_8_8_7
void FASTCALL call_0_8_8_7(SH2_struct* sh)
{
  SH2mull(sh,8,8);
}
#define init_0_8_9_7
void FASTCALL call_0_8_9_7(SH2_struct* sh)
{
  SH2mull(sh,8,9);
}
#define init_0_8_10_7
void FASTCALL call_0_8_10_7(SH2_struct* sh)
{
  SH2mull(sh,8,10);
}
#define init_0_8_11_7
void FASTCALL call_0_8_11_7(SH2_struct* sh)
{
  SH2mull(sh,8,11);
}
#define init_0_8_12_7
void FASTCALL call_0_8_12_7(SH2_struct* sh)
{
  SH2mull(sh,8,12);
}
#define init_0_8_13_7
void FASTCALL call_0_8_13_7(SH2_struct* sh)
{
  SH2mull(sh,8,13);
}
#define init_0_8_14_7
void FASTCALL call_0_8_14_7(SH2_struct* sh)
{
  SH2mull(sh,8,14);
}
#define init_0_8_15_7
void FASTCALL call_0_8_15_7(SH2_struct* sh)
{
  SH2mull(sh,8,15);
}
#define init_0_9_0_7
void FASTCALL call_0_9_0_7(SH2_struct* sh)
{
  SH2mull(sh,9,0);
}
#define init_0_9_1_7
void FASTCALL call_0_9_1_7(SH2_struct* sh)
{
  SH2mull(sh,9,1);
}
#define init_0_9_2_7
void FASTCALL call_0_9_2_7(SH2_struct* sh)
{
  SH2mull(sh,9,2);
}
#define init_0_9_3_7
void FASTCALL call_0_9_3_7(SH2_struct* sh)
{
  SH2mull(sh,9,3);
}
#define init_0_9_4_7
void FASTCALL call_0_9_4_7(SH2_struct* sh)
{
  SH2mull(sh,9,4);
}
#define init_0_9_5_7
void FASTCALL call_0_9_5_7(SH2_struct* sh)
{
  SH2mull(sh,9,5);
}
#define init_0_9_6_7
void FASTCALL call_0_9_6_7(SH2_struct* sh)
{
  SH2mull(sh,9,6);
}
#define init_0_9_7_7
void FASTCALL call_0_9_7_7(SH2_struct* sh)
{
  SH2mull(sh,9,7);
}
#define init_0_9_8_7
void FASTCALL call_0_9_8_7(SH2_struct* sh)
{
  SH2mull(sh,9,8);
}
#define init_0_9_9_7
void FASTCALL call_0_9_9_7(SH2_struct* sh)
{
  SH2mull(sh,9,9);
}
#define init_0_9_10_7
void FASTCALL call_0_9_10_7(SH2_struct* sh)
{
  SH2mull(sh,9,10);
}
#define init_0_9_11_7
void FASTCALL call_0_9_11_7(SH2_struct* sh)
{
  SH2mull(sh,9,11);
}
#define init_0_9_12_7
void FASTCALL call_0_9_12_7(SH2_struct* sh)
{
  SH2mull(sh,9,12);
}
#define init_0_9_13_7
void FASTCALL call_0_9_13_7(SH2_struct* sh)
{
  SH2mull(sh,9,13);
}
#define init_0_9_14_7
void FASTCALL call_0_9_14_7(SH2_struct* sh)
{
  SH2mull(sh,9,14);
}
#define init_0_9_15_7
void FASTCALL call_0_9_15_7(SH2_struct* sh)
{
  SH2mull(sh,9,15);
}
#define init_0_10_0_7
void FASTCALL call_0_10_0_7(SH2_struct* sh)
{
  SH2mull(sh,10,0);
}
#define init_0_10_1_7
void FASTCALL call_0_10_1_7(SH2_struct* sh)
{
  SH2mull(sh,10,1);
}
#define init_0_10_2_7
void FASTCALL call_0_10_2_7(SH2_struct* sh)
{
  SH2mull(sh,10,2);
}
#define init_0_10_3_7
void FASTCALL call_0_10_3_7(SH2_struct* sh)
{
  SH2mull(sh,10,3);
}
#define init_0_10_4_7
void FASTCALL call_0_10_4_7(SH2_struct* sh)
{
  SH2mull(sh,10,4);
}
#define init_0_10_5_7
void FASTCALL call_0_10_5_7(SH2_struct* sh)
{
  SH2mull(sh,10,5);
}
#define init_0_10_6_7
void FASTCALL call_0_10_6_7(SH2_struct* sh)
{
  SH2mull(sh,10,6);
}
#define init_0_10_7_7
void FASTCALL call_0_10_7_7(SH2_struct* sh)
{
  SH2mull(sh,10,7);
}
#define init_0_10_8_7
void FASTCALL call_0_10_8_7(SH2_struct* sh)
{
  SH2mull(sh,10,8);
}
#define init_0_10_9_7
void FASTCALL call_0_10_9_7(SH2_struct* sh)
{
  SH2mull(sh,10,9);
}
#define init_0_10_10_7
void FASTCALL call_0_10_10_7(SH2_struct* sh)
{
  SH2mull(sh,10,10);
}
#define init_0_10_11_7
void FASTCALL call_0_10_11_7(SH2_struct* sh)
{
  SH2mull(sh,10,11);
}
#define init_0_10_12_7
void FASTCALL call_0_10_12_7(SH2_struct* sh)
{
  SH2mull(sh,10,12);
}
#define init_0_10_13_7
void FASTCALL call_0_10_13_7(SH2_struct* sh)
{
  SH2mull(sh,10,13);
}
#define init_0_10_14_7
void FASTCALL call_0_10_14_7(SH2_struct* sh)
{
  SH2mull(sh,10,14);
}
#define init_0_10_15_7
void FASTCALL call_0_10_15_7(SH2_struct* sh)
{
  SH2mull(sh,10,15);
}
#define init_0_11_0_7
void FASTCALL call_0_11_0_7(SH2_struct* sh)
{
  SH2mull(sh,11,0);
}
#define init_0_11_1_7
void FASTCALL call_0_11_1_7(SH2_struct* sh)
{
  SH2mull(sh,11,1);
}
#define init_0_11_2_7
void FASTCALL call_0_11_2_7(SH2_struct* sh)
{
  SH2mull(sh,11,2);
}
#define init_0_11_3_7
void FASTCALL call_0_11_3_7(SH2_struct* sh)
{
  SH2mull(sh,11,3);
}
#define init_0_11_4_7
void FASTCALL call_0_11_4_7(SH2_struct* sh)
{
  SH2mull(sh,11,4);
}
#define init_0_11_5_7
void FASTCALL call_0_11_5_7(SH2_struct* sh)
{
  SH2mull(sh,11,5);
}
#define init_0_11_6_7
void FASTCALL call_0_11_6_7(SH2_struct* sh)
{
  SH2mull(sh,11,6);
}
#define init_0_11_7_7
void FASTCALL call_0_11_7_7(SH2_struct* sh)
{
  SH2mull(sh,11,7);
}
#define init_0_11_8_7
void FASTCALL call_0_11_8_7(SH2_struct* sh)
{
  SH2mull(sh,11,8);
}
#define init_0_11_9_7
void FASTCALL call_0_11_9_7(SH2_struct* sh)
{
  SH2mull(sh,11,9);
}
#define init_0_11_10_7
void FASTCALL call_0_11_10_7(SH2_struct* sh)
{
  SH2mull(sh,11,10);
}
#define init_0_11_11_7
void FASTCALL call_0_11_11_7(SH2_struct* sh)
{
  SH2mull(sh,11,11);
}
#define init_0_11_12_7
void FASTCALL call_0_11_12_7(SH2_struct* sh)
{
  SH2mull(sh,11,12);
}
#define init_0_11_13_7
void FASTCALL call_0_11_13_7(SH2_struct* sh)
{
  SH2mull(sh,11,13);
}
#define init_0_11_14_7
void FASTCALL call_0_11_14_7(SH2_struct* sh)
{
  SH2mull(sh,11,14);
}
#define init_0_11_15_7
void FASTCALL call_0_11_15_7(SH2_struct* sh)
{
  SH2mull(sh,11,15);
}
#define init_0_12_0_7
void FASTCALL call_0_12_0_7(SH2_struct* sh)
{
  SH2mull(sh,12,0);
}
#define init_0_12_1_7
void FASTCALL call_0_12_1_7(SH2_struct* sh)
{
  SH2mull(sh,12,1);
}
#define init_0_12_2_7
void FASTCALL call_0_12_2_7(SH2_struct* sh)
{
  SH2mull(sh,12,2);
}
#define init_0_12_3_7
void FASTCALL call_0_12_3_7(SH2_struct* sh)
{
  SH2mull(sh,12,3);
}
#define init_0_12_4_7
void FASTCALL call_0_12_4_7(SH2_struct* sh)
{
  SH2mull(sh,12,4);
}
#define init_0_12_5_7
void FASTCALL call_0_12_5_7(SH2_struct* sh)
{
  SH2mull(sh,12,5);
}
#define init_0_12_6_7
void FASTCALL call_0_12_6_7(SH2_struct* sh)
{
  SH2mull(sh,12,6);
}
#define init_0_12_7_7
void FASTCALL call_0_12_7_7(SH2_struct* sh)
{
  SH2mull(sh,12,7);
}
#define init_0_12_8_7
void FASTCALL call_0_12_8_7(SH2_struct* sh)
{
  SH2mull(sh,12,8);
}
#define init_0_12_9_7
void FASTCALL call_0_12_9_7(SH2_struct* sh)
{
  SH2mull(sh,12,9);
}
#define init_0_12_10_7
void FASTCALL call_0_12_10_7(SH2_struct* sh)
{
  SH2mull(sh,12,10);
}
#define init_0_12_11_7
void FASTCALL call_0_12_11_7(SH2_struct* sh)
{
  SH2mull(sh,12,11);
}
#define init_0_12_12_7
void FASTCALL call_0_12_12_7(SH2_struct* sh)
{
  SH2mull(sh,12,12);
}
#define init_0_12_13_7
void FASTCALL call_0_12_13_7(SH2_struct* sh)
{
  SH2mull(sh,12,13);
}
#define init_0_12_14_7
void FASTCALL call_0_12_14_7(SH2_struct* sh)
{
  SH2mull(sh,12,14);
}
#define init_0_12_15_7
void FASTCALL call_0_12_15_7(SH2_struct* sh)
{
  SH2mull(sh,12,15);
}
#define init_0_13_0_7
void FASTCALL call_0_13_0_7(SH2_struct* sh)
{
  SH2mull(sh,13,0);
}
#define init_0_13_1_7
void FASTCALL call_0_13_1_7(SH2_struct* sh)
{
  SH2mull(sh,13,1);
}
#define init_0_13_2_7
void FASTCALL call_0_13_2_7(SH2_struct* sh)
{
  SH2mull(sh,13,2);
}
#define init_0_13_3_7
void FASTCALL call_0_13_3_7(SH2_struct* sh)
{
  SH2mull(sh,13,3);
}
#define init_0_13_4_7
void FASTCALL call_0_13_4_7(SH2_struct* sh)
{
  SH2mull(sh,13,4);
}
#define init_0_13_5_7
void FASTCALL call_0_13_5_7(SH2_struct* sh)
{
  SH2mull(sh,13,5);
}
#define init_0_13_6_7
void FASTCALL call_0_13_6_7(SH2_struct* sh)
{
  SH2mull(sh,13,6);
}
#define init_0_13_7_7
void FASTCALL call_0_13_7_7(SH2_struct* sh)
{
  SH2mull(sh,13,7);
}
#define init_0_13_8_7
void FASTCALL call_0_13_8_7(SH2_struct* sh)
{
  SH2mull(sh,13,8);
}
#define init_0_13_9_7
void FASTCALL call_0_13_9_7(SH2_struct* sh)
{
  SH2mull(sh,13,9);
}
#define init_0_13_10_7
void FASTCALL call_0_13_10_7(SH2_struct* sh)
{
  SH2mull(sh,13,10);
}
#define init_0_13_11_7
void FASTCALL call_0_13_11_7(SH2_struct* sh)
{
  SH2mull(sh,13,11);
}
#define init_0_13_12_7
void FASTCALL call_0_13_12_7(SH2_struct* sh)
{
  SH2mull(sh,13,12);
}
#define init_0_13_13_7
void FASTCALL call_0_13_13_7(SH2_struct* sh)
{
  SH2mull(sh,13,13);
}
#define init_0_13_14_7
void FASTCALL call_0_13_14_7(SH2_struct* sh)
{
  SH2mull(sh,13,14);
}
#define init_0_13_15_7
void FASTCALL call_0_13_15_7(SH2_struct* sh)
{
  SH2mull(sh,13,15);
}
#define init_0_14_0_7
void FASTCALL call_0_14_0_7(SH2_struct* sh)
{
  SH2mull(sh,14,0);
}
#define init_0_14_1_7
void FASTCALL call_0_14_1_7(SH2_struct* sh)
{
  SH2mull(sh,14,1);
}
#define init_0_14_2_7
void FASTCALL call_0_14_2_7(SH2_struct* sh)
{
  SH2mull(sh,14,2);
}
#define init_0_14_3_7
void FASTCALL call_0_14_3_7(SH2_struct* sh)
{
  SH2mull(sh,14,3);
}
#define init_0_14_4_7
void FASTCALL call_0_14_4_7(SH2_struct* sh)
{
  SH2mull(sh,14,4);
}
#define init_0_14_5_7
void FASTCALL call_0_14_5_7(SH2_struct* sh)
{
  SH2mull(sh,14,5);
}
#define init_0_14_6_7
void FASTCALL call_0_14_6_7(SH2_struct* sh)
{
  SH2mull(sh,14,6);
}
#define init_0_14_7_7
void FASTCALL call_0_14_7_7(SH2_struct* sh)
{
  SH2mull(sh,14,7);
}
#define init_0_14_8_7
void FASTCALL call_0_14_8_7(SH2_struct* sh)
{
  SH2mull(sh,14,8);
}
#define init_0_14_9_7
void FASTCALL call_0_14_9_7(SH2_struct* sh)
{
  SH2mull(sh,14,9);
}
#define init_0_14_10_7
void FASTCALL call_0_14_10_7(SH2_struct* sh)
{
  SH2mull(sh,14,10);
}
#define init_0_14_11_7
void FASTCALL call_0_14_11_7(SH2_struct* sh)
{
  SH2mull(sh,14,11);
}
#define init_0_14_12_7
void FASTCALL call_0_14_12_7(SH2_struct* sh)
{
  SH2mull(sh,14,12);
}
#define init_0_14_13_7
void FASTCALL call_0_14_13_7(SH2_struct* sh)
{
  SH2mull(sh,14,13);
}
#define init_0_14_14_7
void FASTCALL call_0_14_14_7(SH2_struct* sh)
{
  SH2mull(sh,14,14);
}
#define init_0_14_15_7
void FASTCALL call_0_14_15_7(SH2_struct* sh)
{
  SH2mull(sh,14,15);
}
#define init_0_15_0_7
void FASTCALL call_0_15_0_7(SH2_struct* sh)
{
  SH2mull(sh,15,0);
}
#define init_0_15_1_7
void FASTCALL call_0_15_1_7(SH2_struct* sh)
{
  SH2mull(sh,15,1);
}
#define init_0_15_2_7
void FASTCALL call_0_15_2_7(SH2_struct* sh)
{
  SH2mull(sh,15,2);
}
#define init_0_15_3_7
void FASTCALL call_0_15_3_7(SH2_struct* sh)
{
  SH2mull(sh,15,3);
}
#define init_0_15_4_7
void FASTCALL call_0_15_4_7(SH2_struct* sh)
{
  SH2mull(sh,15,4);
}
#define init_0_15_5_7
void FASTCALL call_0_15_5_7(SH2_struct* sh)
{
  SH2mull(sh,15,5);
}
#define init_0_15_6_7
void FASTCALL call_0_15_6_7(SH2_struct* sh)
{
  SH2mull(sh,15,6);
}
#define init_0_15_7_7
void FASTCALL call_0_15_7_7(SH2_struct* sh)
{
  SH2mull(sh,15,7);
}
#define init_0_15_8_7
void FASTCALL call_0_15_8_7(SH2_struct* sh)
{
  SH2mull(sh,15,8);
}
#define init_0_15_9_7
void FASTCALL call_0_15_9_7(SH2_struct* sh)
{
  SH2mull(sh,15,9);
}
#define init_0_15_10_7
void FASTCALL call_0_15_10_7(SH2_struct* sh)
{
  SH2mull(sh,15,10);
}
#define init_0_15_11_7
void FASTCALL call_0_15_11_7(SH2_struct* sh)
{
  SH2mull(sh,15,11);
}
#define init_0_15_12_7
void FASTCALL call_0_15_12_7(SH2_struct* sh)
{
  SH2mull(sh,15,12);
}
#define init_0_15_13_7
void FASTCALL call_0_15_13_7(SH2_struct* sh)
{
  SH2mull(sh,15,13);
}
#define init_0_15_14_7
void FASTCALL call_0_15_14_7(SH2_struct* sh)
{
  SH2mull(sh,15,14);
}
#define init_0_15_15_7
void FASTCALL call_0_15_15_7(SH2_struct* sh)
{
  SH2mull(sh,15,15);
}
#define init_0_0_0_13
void FASTCALL call_0_0_0_13(SH2_struct* sh)
{
  SH2movwl0(sh,0,0);
}
#define init_0_0_1_13
void FASTCALL call_0_0_1_13(SH2_struct* sh)
{
  SH2movwl0(sh,0,1);
}
#define init_0_0_2_13
void FASTCALL call_0_0_2_13(SH2_struct* sh)
{
  SH2movwl0(sh,0,2);
}
#define init_0_0_3_13
void FASTCALL call_0_0_3_13(SH2_struct* sh)
{
  SH2movwl0(sh,0,3);
}
#define init_0_0_4_13
void FASTCALL call_0_0_4_13(SH2_struct* sh)
{
  SH2movwl0(sh,0,4);
}
#define init_0_0_5_13
void FASTCALL call_0_0_5_13(SH2_struct* sh)
{
  SH2movwl0(sh,0,5);
}
#define init_0_0_6_13
void FASTCALL call_0_0_6_13(SH2_struct* sh)
{
  SH2movwl0(sh,0,6);
}
#define init_0_0_7_13
void FASTCALL call_0_0_7_13(SH2_struct* sh)
{
  SH2movwl0(sh,0,7);
}
#define init_0_0_8_13
void FASTCALL call_0_0_8_13(SH2_struct* sh)
{
  SH2movwl0(sh,0,8);
}
#define init_0_0_9_13
void FASTCALL call_0_0_9_13(SH2_struct* sh)
{
  SH2movwl0(sh,0,9);
}
#define init_0_0_10_13
void FASTCALL call_0_0_10_13(SH2_struct* sh)
{
  SH2movwl0(sh,0,10);
}
#define init_0_0_11_13
void FASTCALL call_0_0_11_13(SH2_struct* sh)
{
  SH2movwl0(sh,0,11);
}
#define init_0_0_12_13
void FASTCALL call_0_0_12_13(SH2_struct* sh)
{
  SH2movwl0(sh,0,12);
}
#define init_0_0_13_13
void FASTCALL call_0_0_13_13(SH2_struct* sh)
{
  SH2movwl0(sh,0,13);
}
#define init_0_0_14_13
void FASTCALL call_0_0_14_13(SH2_struct* sh)
{
  SH2movwl0(sh,0,14);
}
#define init_0_0_15_13
void FASTCALL call_0_0_15_13(SH2_struct* sh)
{
  SH2movwl0(sh,0,15);
}
#define init_0_1_0_13
void FASTCALL call_0_1_0_13(SH2_struct* sh)
{
  SH2movwl0(sh,1,0);
}
#define init_0_1_1_13
void FASTCALL call_0_1_1_13(SH2_struct* sh)
{
  SH2movwl0(sh,1,1);
}
#define init_0_1_2_13
void FASTCALL call_0_1_2_13(SH2_struct* sh)
{
  SH2movwl0(sh,1,2);
}
#define init_0_1_3_13
void FASTCALL call_0_1_3_13(SH2_struct* sh)
{
  SH2movwl0(sh,1,3);
}
#define init_0_1_4_13
void FASTCALL call_0_1_4_13(SH2_struct* sh)
{
  SH2movwl0(sh,1,4);
}
#define init_0_1_5_13
void FASTCALL call_0_1_5_13(SH2_struct* sh)
{
  SH2movwl0(sh,1,5);
}
#define init_0_1_6_13
void FASTCALL call_0_1_6_13(SH2_struct* sh)
{
  SH2movwl0(sh,1,6);
}
#define init_0_1_7_13
void FASTCALL call_0_1_7_13(SH2_struct* sh)
{
  SH2movwl0(sh,1,7);
}
#define init_0_1_8_13
void FASTCALL call_0_1_8_13(SH2_struct* sh)
{
  SH2movwl0(sh,1,8);
}
#define init_0_1_9_13
void FASTCALL call_0_1_9_13(SH2_struct* sh)
{
  SH2movwl0(sh,1,9);
}
#define init_0_1_10_13
void FASTCALL call_0_1_10_13(SH2_struct* sh)
{
  SH2movwl0(sh,1,10);
}
#define init_0_1_11_13
void FASTCALL call_0_1_11_13(SH2_struct* sh)
{
  SH2movwl0(sh,1,11);
}
#define init_0_1_12_13
void FASTCALL call_0_1_12_13(SH2_struct* sh)
{
  SH2movwl0(sh,1,12);
}
#define init_0_1_13_13
void FASTCALL call_0_1_13_13(SH2_struct* sh)
{
  SH2movwl0(sh,1,13);
}
#define init_0_1_14_13
void FASTCALL call_0_1_14_13(SH2_struct* sh)
{
  SH2movwl0(sh,1,14);
}
#define init_0_1_15_13
void FASTCALL call_0_1_15_13(SH2_struct* sh)
{
  SH2movwl0(sh,1,15);
}
#define init_0_2_0_13
void FASTCALL call_0_2_0_13(SH2_struct* sh)
{
  SH2movwl0(sh,2,0);
}
#define init_0_2_1_13
void FASTCALL call_0_2_1_13(SH2_struct* sh)
{
  SH2movwl0(sh,2,1);
}
#define init_0_2_2_13
void FASTCALL call_0_2_2_13(SH2_struct* sh)
{
  SH2movwl0(sh,2,2);
}
#define init_0_2_3_13
void FASTCALL call_0_2_3_13(SH2_struct* sh)
{
  SH2movwl0(sh,2,3);
}
#define init_0_2_4_13
void FASTCALL call_0_2_4_13(SH2_struct* sh)
{
  SH2movwl0(sh,2,4);
}
#define init_0_2_5_13
void FASTCALL call_0_2_5_13(SH2_struct* sh)
{
  SH2movwl0(sh,2,5);
}
#define init_0_2_6_13
void FASTCALL call_0_2_6_13(SH2_struct* sh)
{
  SH2movwl0(sh,2,6);
}
#define init_0_2_7_13
void FASTCALL call_0_2_7_13(SH2_struct* sh)
{
  SH2movwl0(sh,2,7);
}
#define init_0_2_8_13
void FASTCALL call_0_2_8_13(SH2_struct* sh)
{
  SH2movwl0(sh,2,8);
}
#define init_0_2_9_13
void FASTCALL call_0_2_9_13(SH2_struct* sh)
{
  SH2movwl0(sh,2,9);
}
#define init_0_2_10_13
void FASTCALL call_0_2_10_13(SH2_struct* sh)
{
  SH2movwl0(sh,2,10);
}
#define init_0_2_11_13
void FASTCALL call_0_2_11_13(SH2_struct* sh)
{
  SH2movwl0(sh,2,11);
}
#define init_0_2_12_13
void FASTCALL call_0_2_12_13(SH2_struct* sh)
{
  SH2movwl0(sh,2,12);
}
#define init_0_2_13_13
void FASTCALL call_0_2_13_13(SH2_struct* sh)
{
  SH2movwl0(sh,2,13);
}
#define init_0_2_14_13
void FASTCALL call_0_2_14_13(SH2_struct* sh)
{
  SH2movwl0(sh,2,14);
}
#define init_0_2_15_13
void FASTCALL call_0_2_15_13(SH2_struct* sh)
{
  SH2movwl0(sh,2,15);
}
#define init_0_3_0_13
void FASTCALL call_0_3_0_13(SH2_struct* sh)
{
  SH2movwl0(sh,3,0);
}
#define init_0_3_1_13
void FASTCALL call_0_3_1_13(SH2_struct* sh)
{
  SH2movwl0(sh,3,1);
}
#define init_0_3_2_13
void FASTCALL call_0_3_2_13(SH2_struct* sh)
{
  SH2movwl0(sh,3,2);
}
#define init_0_3_3_13
void FASTCALL call_0_3_3_13(SH2_struct* sh)
{
  SH2movwl0(sh,3,3);
}
#define init_0_3_4_13
void FASTCALL call_0_3_4_13(SH2_struct* sh)
{
  SH2movwl0(sh,3,4);
}
#define init_0_3_5_13
void FASTCALL call_0_3_5_13(SH2_struct* sh)
{
  SH2movwl0(sh,3,5);
}
#define init_0_3_6_13
void FASTCALL call_0_3_6_13(SH2_struct* sh)
{
  SH2movwl0(sh,3,6);
}
#define init_0_3_7_13
void FASTCALL call_0_3_7_13(SH2_struct* sh)
{
  SH2movwl0(sh,3,7);
}
#define init_0_3_8_13
void FASTCALL call_0_3_8_13(SH2_struct* sh)
{
  SH2movwl0(sh,3,8);
}
#define init_0_3_9_13
void FASTCALL call_0_3_9_13(SH2_struct* sh)
{
  SH2movwl0(sh,3,9);
}
#define init_0_3_10_13
void FASTCALL call_0_3_10_13(SH2_struct* sh)
{
  SH2movwl0(sh,3,10);
}
#define init_0_3_11_13
void FASTCALL call_0_3_11_13(SH2_struct* sh)
{
  SH2movwl0(sh,3,11);
}
#define init_0_3_12_13
void FASTCALL call_0_3_12_13(SH2_struct* sh)
{
  SH2movwl0(sh,3,12);
}
#define init_0_3_13_13
void FASTCALL call_0_3_13_13(SH2_struct* sh)
{
  SH2movwl0(sh,3,13);
}
#define init_0_3_14_13
void FASTCALL call_0_3_14_13(SH2_struct* sh)
{
  SH2movwl0(sh,3,14);
}
#define init_0_3_15_13
void FASTCALL call_0_3_15_13(SH2_struct* sh)
{
  SH2movwl0(sh,3,15);
}
#define init_0_4_0_13
void FASTCALL call_0_4_0_13(SH2_struct* sh)
{
  SH2movwl0(sh,4,0);
}
#define init_0_4_1_13
void FASTCALL call_0_4_1_13(SH2_struct* sh)
{
  SH2movwl0(sh,4,1);
}
#define init_0_4_2_13
void FASTCALL call_0_4_2_13(SH2_struct* sh)
{
  SH2movwl0(sh,4,2);
}
#define init_0_4_3_13
void FASTCALL call_0_4_3_13(SH2_struct* sh)
{
  SH2movwl0(sh,4,3);
}
#define init_0_4_4_13
void FASTCALL call_0_4_4_13(SH2_struct* sh)
{
  SH2movwl0(sh,4,4);
}
#define init_0_4_5_13
void FASTCALL call_0_4_5_13(SH2_struct* sh)
{
  SH2movwl0(sh,4,5);
}
#define init_0_4_6_13
void FASTCALL call_0_4_6_13(SH2_struct* sh)
{
  SH2movwl0(sh,4,6);
}
#define init_0_4_7_13
void FASTCALL call_0_4_7_13(SH2_struct* sh)
{
  SH2movwl0(sh,4,7);
}
#define init_0_4_8_13
void FASTCALL call_0_4_8_13(SH2_struct* sh)
{
  SH2movwl0(sh,4,8);
}
#define init_0_4_9_13
void FASTCALL call_0_4_9_13(SH2_struct* sh)
{
  SH2movwl0(sh,4,9);
}
#define init_0_4_10_13
void FASTCALL call_0_4_10_13(SH2_struct* sh)
{
  SH2movwl0(sh,4,10);
}
#define init_0_4_11_13
void FASTCALL call_0_4_11_13(SH2_struct* sh)
{
  SH2movwl0(sh,4,11);
}
#define init_0_4_12_13
void FASTCALL call_0_4_12_13(SH2_struct* sh)
{
  SH2movwl0(sh,4,12);
}
#define init_0_4_13_13
void FASTCALL call_0_4_13_13(SH2_struct* sh)
{
  SH2movwl0(sh,4,13);
}
#define init_0_4_14_13
void FASTCALL call_0_4_14_13(SH2_struct* sh)
{
  SH2movwl0(sh,4,14);
}
#define init_0_4_15_13
void FASTCALL call_0_4_15_13(SH2_struct* sh)
{
  SH2movwl0(sh,4,15);
}
#define init_0_5_0_13
void FASTCALL call_0_5_0_13(SH2_struct* sh)
{
  SH2movwl0(sh,5,0);
}
#define init_0_5_1_13
void FASTCALL call_0_5_1_13(SH2_struct* sh)
{
  SH2movwl0(sh,5,1);
}
#define init_0_5_2_13
void FASTCALL call_0_5_2_13(SH2_struct* sh)
{
  SH2movwl0(sh,5,2);
}
#define init_0_5_3_13
void FASTCALL call_0_5_3_13(SH2_struct* sh)
{
  SH2movwl0(sh,5,3);
}
#define init_0_5_4_13
void FASTCALL call_0_5_4_13(SH2_struct* sh)
{
  SH2movwl0(sh,5,4);
}
#define init_0_5_5_13
void FASTCALL call_0_5_5_13(SH2_struct* sh)
{
  SH2movwl0(sh,5,5);
}
#define init_0_5_6_13
void FASTCALL call_0_5_6_13(SH2_struct* sh)
{
  SH2movwl0(sh,5,6);
}
#define init_0_5_7_13
void FASTCALL call_0_5_7_13(SH2_struct* sh)
{
  SH2movwl0(sh,5,7);
}
#define init_0_5_8_13
void FASTCALL call_0_5_8_13(SH2_struct* sh)
{
  SH2movwl0(sh,5,8);
}
#define init_0_5_9_13
void FASTCALL call_0_5_9_13(SH2_struct* sh)
{
  SH2movwl0(sh,5,9);
}
#define init_0_5_10_13
void FASTCALL call_0_5_10_13(SH2_struct* sh)
{
  SH2movwl0(sh,5,10);
}
#define init_0_5_11_13
void FASTCALL call_0_5_11_13(SH2_struct* sh)
{
  SH2movwl0(sh,5,11);
}
#define init_0_5_12_13
void FASTCALL call_0_5_12_13(SH2_struct* sh)
{
  SH2movwl0(sh,5,12);
}
#define init_0_5_13_13
void FASTCALL call_0_5_13_13(SH2_struct* sh)
{
  SH2movwl0(sh,5,13);
}
#define init_0_5_14_13
void FASTCALL call_0_5_14_13(SH2_struct* sh)
{
  SH2movwl0(sh,5,14);
}
#define init_0_5_15_13
void FASTCALL call_0_5_15_13(SH2_struct* sh)
{
  SH2movwl0(sh,5,15);
}
#define init_0_6_0_13
void FASTCALL call_0_6_0_13(SH2_struct* sh)
{
  SH2movwl0(sh,6,0);
}
#define init_0_6_1_13
void FASTCALL call_0_6_1_13(SH2_struct* sh)
{
  SH2movwl0(sh,6,1);
}
#define init_0_6_2_13
void FASTCALL call_0_6_2_13(SH2_struct* sh)
{
  SH2movwl0(sh,6,2);
}
#define init_0_6_3_13
void FASTCALL call_0_6_3_13(SH2_struct* sh)
{
  SH2movwl0(sh,6,3);
}
#define init_0_6_4_13
void FASTCALL call_0_6_4_13(SH2_struct* sh)
{
  SH2movwl0(sh,6,4);
}
#define init_0_6_5_13
void FASTCALL call_0_6_5_13(SH2_struct* sh)
{
  SH2movwl0(sh,6,5);
}
#define init_0_6_6_13
void FASTCALL call_0_6_6_13(SH2_struct* sh)
{
  SH2movwl0(sh,6,6);
}
#define init_0_6_7_13
void FASTCALL call_0_6_7_13(SH2_struct* sh)
{
  SH2movwl0(sh,6,7);
}
#define init_0_6_8_13
void FASTCALL call_0_6_8_13(SH2_struct* sh)
{
  SH2movwl0(sh,6,8);
}
#define init_0_6_9_13
void FASTCALL call_0_6_9_13(SH2_struct* sh)
{
  SH2movwl0(sh,6,9);
}
#define init_0_6_10_13
void FASTCALL call_0_6_10_13(SH2_struct* sh)
{
  SH2movwl0(sh,6,10);
}
#define init_0_6_11_13
void FASTCALL call_0_6_11_13(SH2_struct* sh)
{
  SH2movwl0(sh,6,11);
}
#define init_0_6_12_13
void FASTCALL call_0_6_12_13(SH2_struct* sh)
{
  SH2movwl0(sh,6,12);
}
#define init_0_6_13_13
void FASTCALL call_0_6_13_13(SH2_struct* sh)
{
  SH2movwl0(sh,6,13);
}
#define init_0_6_14_13
void FASTCALL call_0_6_14_13(SH2_struct* sh)
{
  SH2movwl0(sh,6,14);
}
#define init_0_6_15_13
void FASTCALL call_0_6_15_13(SH2_struct* sh)
{
  SH2movwl0(sh,6,15);
}
#define init_0_7_0_13
void FASTCALL call_0_7_0_13(SH2_struct* sh)
{
  SH2movwl0(sh,7,0);
}
#define init_0_7_1_13
void FASTCALL call_0_7_1_13(SH2_struct* sh)
{
  SH2movwl0(sh,7,1);
}
#define init_0_7_2_13
void FASTCALL call_0_7_2_13(SH2_struct* sh)
{
  SH2movwl0(sh,7,2);
}
#define init_0_7_3_13
void FASTCALL call_0_7_3_13(SH2_struct* sh)
{
  SH2movwl0(sh,7,3);
}
#define init_0_7_4_13
void FASTCALL call_0_7_4_13(SH2_struct* sh)
{
  SH2movwl0(sh,7,4);
}
#define init_0_7_5_13
void FASTCALL call_0_7_5_13(SH2_struct* sh)
{
  SH2movwl0(sh,7,5);
}
#define init_0_7_6_13
void FASTCALL call_0_7_6_13(SH2_struct* sh)
{
  SH2movwl0(sh,7,6);
}
#define init_0_7_7_13
void FASTCALL call_0_7_7_13(SH2_struct* sh)
{
  SH2movwl0(sh,7,7);
}
#define init_0_7_8_13
void FASTCALL call_0_7_8_13(SH2_struct* sh)
{
  SH2movwl0(sh,7,8);
}
#define init_0_7_9_13
void FASTCALL call_0_7_9_13(SH2_struct* sh)
{
  SH2movwl0(sh,7,9);
}
#define init_0_7_10_13
void FASTCALL call_0_7_10_13(SH2_struct* sh)
{
  SH2movwl0(sh,7,10);
}
#define init_0_7_11_13
void FASTCALL call_0_7_11_13(SH2_struct* sh)
{
  SH2movwl0(sh,7,11);
}
#define init_0_7_12_13
void FASTCALL call_0_7_12_13(SH2_struct* sh)
{
  SH2movwl0(sh,7,12);
}
#define init_0_7_13_13
void FASTCALL call_0_7_13_13(SH2_struct* sh)
{
  SH2movwl0(sh,7,13);
}
#define init_0_7_14_13
void FASTCALL call_0_7_14_13(SH2_struct* sh)
{
  SH2movwl0(sh,7,14);
}
#define init_0_7_15_13
void FASTCALL call_0_7_15_13(SH2_struct* sh)
{
  SH2movwl0(sh,7,15);
}
#define init_0_8_0_13
void FASTCALL call_0_8_0_13(SH2_struct* sh)
{
  SH2movwl0(sh,8,0);
}
#define init_0_8_1_13
void FASTCALL call_0_8_1_13(SH2_struct* sh)
{
  SH2movwl0(sh,8,1);
}
#define init_0_8_2_13
void FASTCALL call_0_8_2_13(SH2_struct* sh)
{
  SH2movwl0(sh,8,2);
}
#define init_0_8_3_13
void FASTCALL call_0_8_3_13(SH2_struct* sh)
{
  SH2movwl0(sh,8,3);
}
#define init_0_8_4_13
void FASTCALL call_0_8_4_13(SH2_struct* sh)
{
  SH2movwl0(sh,8,4);
}
#define init_0_8_5_13
void FASTCALL call_0_8_5_13(SH2_struct* sh)
{
  SH2movwl0(sh,8,5);
}
#define init_0_8_6_13
void FASTCALL call_0_8_6_13(SH2_struct* sh)
{
  SH2movwl0(sh,8,6);
}
#define init_0_8_7_13
void FASTCALL call_0_8_7_13(SH2_struct* sh)
{
  SH2movwl0(sh,8,7);
}
#define init_0_8_8_13
void FASTCALL call_0_8_8_13(SH2_struct* sh)
{
  SH2movwl0(sh,8,8);
}
#define init_0_8_9_13
void FASTCALL call_0_8_9_13(SH2_struct* sh)
{
  SH2movwl0(sh,8,9);
}
#define init_0_8_10_13
void FASTCALL call_0_8_10_13(SH2_struct* sh)
{
  SH2movwl0(sh,8,10);
}
#define init_0_8_11_13
void FASTCALL call_0_8_11_13(SH2_struct* sh)
{
  SH2movwl0(sh,8,11);
}
#define init_0_8_12_13
void FASTCALL call_0_8_12_13(SH2_struct* sh)
{
  SH2movwl0(sh,8,12);
}
#define init_0_8_13_13
void FASTCALL call_0_8_13_13(SH2_struct* sh)
{
  SH2movwl0(sh,8,13);
}
#define init_0_8_14_13
void FASTCALL call_0_8_14_13(SH2_struct* sh)
{
  SH2movwl0(sh,8,14);
}
#define init_0_8_15_13
void FASTCALL call_0_8_15_13(SH2_struct* sh)
{
  SH2movwl0(sh,8,15);
}
#define init_0_9_0_13
void FASTCALL call_0_9_0_13(SH2_struct* sh)
{
  SH2movwl0(sh,9,0);
}
#define init_0_9_1_13
void FASTCALL call_0_9_1_13(SH2_struct* sh)
{
  SH2movwl0(sh,9,1);
}
#define init_0_9_2_13
void FASTCALL call_0_9_2_13(SH2_struct* sh)
{
  SH2movwl0(sh,9,2);
}
#define init_0_9_3_13
void FASTCALL call_0_9_3_13(SH2_struct* sh)
{
  SH2movwl0(sh,9,3);
}
#define init_0_9_4_13
void FASTCALL call_0_9_4_13(SH2_struct* sh)
{
  SH2movwl0(sh,9,4);
}
#define init_0_9_5_13
void FASTCALL call_0_9_5_13(SH2_struct* sh)
{
  SH2movwl0(sh,9,5);
}
#define init_0_9_6_13
void FASTCALL call_0_9_6_13(SH2_struct* sh)
{
  SH2movwl0(sh,9,6);
}
#define init_0_9_7_13
void FASTCALL call_0_9_7_13(SH2_struct* sh)
{
  SH2movwl0(sh,9,7);
}
#define init_0_9_8_13
void FASTCALL call_0_9_8_13(SH2_struct* sh)
{
  SH2movwl0(sh,9,8);
}
#define init_0_9_9_13
void FASTCALL call_0_9_9_13(SH2_struct* sh)
{
  SH2movwl0(sh,9,9);
}
#define init_0_9_10_13
void FASTCALL call_0_9_10_13(SH2_struct* sh)
{
  SH2movwl0(sh,9,10);
}
#define init_0_9_11_13
void FASTCALL call_0_9_11_13(SH2_struct* sh)
{
  SH2movwl0(sh,9,11);
}
#define init_0_9_12_13
void FASTCALL call_0_9_12_13(SH2_struct* sh)
{
  SH2movwl0(sh,9,12);
}
#define init_0_9_13_13
void FASTCALL call_0_9_13_13(SH2_struct* sh)
{
  SH2movwl0(sh,9,13);
}
#define init_0_9_14_13
void FASTCALL call_0_9_14_13(SH2_struct* sh)
{
  SH2movwl0(sh,9,14);
}
#define init_0_9_15_13
void FASTCALL call_0_9_15_13(SH2_struct* sh)
{
  SH2movwl0(sh,9,15);
}
#define init_0_10_0_13
void FASTCALL call_0_10_0_13(SH2_struct* sh)
{
  SH2movwl0(sh,10,0);
}
#define init_0_10_1_13
void FASTCALL call_0_10_1_13(SH2_struct* sh)
{
  SH2movwl0(sh,10,1);
}
#define init_0_10_2_13
void FASTCALL call_0_10_2_13(SH2_struct* sh)
{
  SH2movwl0(sh,10,2);
}
#define init_0_10_3_13
void FASTCALL call_0_10_3_13(SH2_struct* sh)
{
  SH2movwl0(sh,10,3);
}
#define init_0_10_4_13
void FASTCALL call_0_10_4_13(SH2_struct* sh)
{
  SH2movwl0(sh,10,4);
}
#define init_0_10_5_13
void FASTCALL call_0_10_5_13(SH2_struct* sh)
{
  SH2movwl0(sh,10,5);
}
#define init_0_10_6_13
void FASTCALL call_0_10_6_13(SH2_struct* sh)
{
  SH2movwl0(sh,10,6);
}
#define init_0_10_7_13
void FASTCALL call_0_10_7_13(SH2_struct* sh)
{
  SH2movwl0(sh,10,7);
}
#define init_0_10_8_13
void FASTCALL call_0_10_8_13(SH2_struct* sh)
{
  SH2movwl0(sh,10,8);
}
#define init_0_10_9_13
void FASTCALL call_0_10_9_13(SH2_struct* sh)
{
  SH2movwl0(sh,10,9);
}
#define init_0_10_10_13
void FASTCALL call_0_10_10_13(SH2_struct* sh)
{
  SH2movwl0(sh,10,10);
}
#define init_0_10_11_13
void FASTCALL call_0_10_11_13(SH2_struct* sh)
{
  SH2movwl0(sh,10,11);
}
#define init_0_10_12_13
void FASTCALL call_0_10_12_13(SH2_struct* sh)
{
  SH2movwl0(sh,10,12);
}
#define init_0_10_13_13
void FASTCALL call_0_10_13_13(SH2_struct* sh)
{
  SH2movwl0(sh,10,13);
}
#define init_0_10_14_13
void FASTCALL call_0_10_14_13(SH2_struct* sh)
{
  SH2movwl0(sh,10,14);
}
#define init_0_10_15_13
void FASTCALL call_0_10_15_13(SH2_struct* sh)
{
  SH2movwl0(sh,10,15);
}
#define init_0_11_0_13
void FASTCALL call_0_11_0_13(SH2_struct* sh)
{
  SH2movwl0(sh,11,0);
}
#define init_0_11_1_13
void FASTCALL call_0_11_1_13(SH2_struct* sh)
{
  SH2movwl0(sh,11,1);
}
#define init_0_11_2_13
void FASTCALL call_0_11_2_13(SH2_struct* sh)
{
  SH2movwl0(sh,11,2);
}
#define init_0_11_3_13
void FASTCALL call_0_11_3_13(SH2_struct* sh)
{
  SH2movwl0(sh,11,3);
}
#define init_0_11_4_13
void FASTCALL call_0_11_4_13(SH2_struct* sh)
{
  SH2movwl0(sh,11,4);
}
#define init_0_11_5_13
void FASTCALL call_0_11_5_13(SH2_struct* sh)
{
  SH2movwl0(sh,11,5);
}
#define init_0_11_6_13
void FASTCALL call_0_11_6_13(SH2_struct* sh)
{
  SH2movwl0(sh,11,6);
}
#define init_0_11_7_13
void FASTCALL call_0_11_7_13(SH2_struct* sh)
{
  SH2movwl0(sh,11,7);
}
#define init_0_11_8_13
void FASTCALL call_0_11_8_13(SH2_struct* sh)
{
  SH2movwl0(sh,11,8);
}
#define init_0_11_9_13
void FASTCALL call_0_11_9_13(SH2_struct* sh)
{
  SH2movwl0(sh,11,9);
}
#define init_0_11_10_13
void FASTCALL call_0_11_10_13(SH2_struct* sh)
{
  SH2movwl0(sh,11,10);
}
#define init_0_11_11_13
void FASTCALL call_0_11_11_13(SH2_struct* sh)
{
  SH2movwl0(sh,11,11);
}
#define init_0_11_12_13
void FASTCALL call_0_11_12_13(SH2_struct* sh)
{
  SH2movwl0(sh,11,12);
}
#define init_0_11_13_13
void FASTCALL call_0_11_13_13(SH2_struct* sh)
{
  SH2movwl0(sh,11,13);
}
#define init_0_11_14_13
void FASTCALL call_0_11_14_13(SH2_struct* sh)
{
  SH2movwl0(sh,11,14);
}
#define init_0_11_15_13
void FASTCALL call_0_11_15_13(SH2_struct* sh)
{
  SH2movwl0(sh,11,15);
}
#define init_0_12_0_13
void FASTCALL call_0_12_0_13(SH2_struct* sh)
{
  SH2movwl0(sh,12,0);
}
#define init_0_12_1_13
void FASTCALL call_0_12_1_13(SH2_struct* sh)
{
  SH2movwl0(sh,12,1);
}
#define init_0_12_2_13
void FASTCALL call_0_12_2_13(SH2_struct* sh)
{
  SH2movwl0(sh,12,2);
}
#define init_0_12_3_13
void FASTCALL call_0_12_3_13(SH2_struct* sh)
{
  SH2movwl0(sh,12,3);
}
#define init_0_12_4_13
void FASTCALL call_0_12_4_13(SH2_struct* sh)
{
  SH2movwl0(sh,12,4);
}
#define init_0_12_5_13
void FASTCALL call_0_12_5_13(SH2_struct* sh)
{
  SH2movwl0(sh,12,5);
}
#define init_0_12_6_13
void FASTCALL call_0_12_6_13(SH2_struct* sh)
{
  SH2movwl0(sh,12,6);
}
#define init_0_12_7_13
void FASTCALL call_0_12_7_13(SH2_struct* sh)
{
  SH2movwl0(sh,12,7);
}
#define init_0_12_8_13
void FASTCALL call_0_12_8_13(SH2_struct* sh)
{
  SH2movwl0(sh,12,8);
}
#define init_0_12_9_13
void FASTCALL call_0_12_9_13(SH2_struct* sh)
{
  SH2movwl0(sh,12,9);
}
#define init_0_12_10_13
void FASTCALL call_0_12_10_13(SH2_struct* sh)
{
  SH2movwl0(sh,12,10);
}
#define init_0_12_11_13
void FASTCALL call_0_12_11_13(SH2_struct* sh)
{
  SH2movwl0(sh,12,11);
}
#define init_0_12_12_13
void FASTCALL call_0_12_12_13(SH2_struct* sh)
{
  SH2movwl0(sh,12,12);
}
#define init_0_12_13_13
void FASTCALL call_0_12_13_13(SH2_struct* sh)
{
  SH2movwl0(sh,12,13);
}
#define init_0_12_14_13
void FASTCALL call_0_12_14_13(SH2_struct* sh)
{
  SH2movwl0(sh,12,14);
}
#define init_0_12_15_13
void FASTCALL call_0_12_15_13(SH2_struct* sh)
{
  SH2movwl0(sh,12,15);
}
#define init_0_13_0_13
void FASTCALL call_0_13_0_13(SH2_struct* sh)
{
  SH2movwl0(sh,13,0);
}
#define init_0_13_1_13
void FASTCALL call_0_13_1_13(SH2_struct* sh)
{
  SH2movwl0(sh,13,1);
}
#define init_0_13_2_13
void FASTCALL call_0_13_2_13(SH2_struct* sh)
{
  SH2movwl0(sh,13,2);
}
#define init_0_13_3_13
void FASTCALL call_0_13_3_13(SH2_struct* sh)
{
  SH2movwl0(sh,13,3);
}
#define init_0_13_4_13
void FASTCALL call_0_13_4_13(SH2_struct* sh)
{
  SH2movwl0(sh,13,4);
}
#define init_0_13_5_13
void FASTCALL call_0_13_5_13(SH2_struct* sh)
{
  SH2movwl0(sh,13,5);
}
#define init_0_13_6_13
void FASTCALL call_0_13_6_13(SH2_struct* sh)
{
  SH2movwl0(sh,13,6);
}
#define init_0_13_7_13
void FASTCALL call_0_13_7_13(SH2_struct* sh)
{
  SH2movwl0(sh,13,7);
}
#define init_0_13_8_13
void FASTCALL call_0_13_8_13(SH2_struct* sh)
{
  SH2movwl0(sh,13,8);
}
#define init_0_13_9_13
void FASTCALL call_0_13_9_13(SH2_struct* sh)
{
  SH2movwl0(sh,13,9);
}
#define init_0_13_10_13
void FASTCALL call_0_13_10_13(SH2_struct* sh)
{
  SH2movwl0(sh,13,10);
}
#define init_0_13_11_13
void FASTCALL call_0_13_11_13(SH2_struct* sh)
{
  SH2movwl0(sh,13,11);
}
#define init_0_13_12_13
void FASTCALL call_0_13_12_13(SH2_struct* sh)
{
  SH2movwl0(sh,13,12);
}
#define init_0_13_13_13
void FASTCALL call_0_13_13_13(SH2_struct* sh)
{
  SH2movwl0(sh,13,13);
}
#define init_0_13_14_13
void FASTCALL call_0_13_14_13(SH2_struct* sh)
{
  SH2movwl0(sh,13,14);
}
#define init_0_13_15_13
void FASTCALL call_0_13_15_13(SH2_struct* sh)
{
  SH2movwl0(sh,13,15);
}
#define init_0_14_0_13
void FASTCALL call_0_14_0_13(SH2_struct* sh)
{
  SH2movwl0(sh,14,0);
}
#define init_0_14_1_13
void FASTCALL call_0_14_1_13(SH2_struct* sh)
{
  SH2movwl0(sh,14,1);
}
#define init_0_14_2_13
void FASTCALL call_0_14_2_13(SH2_struct* sh)
{
  SH2movwl0(sh,14,2);
}
#define init_0_14_3_13
void FASTCALL call_0_14_3_13(SH2_struct* sh)
{
  SH2movwl0(sh,14,3);
}
#define init_0_14_4_13
void FASTCALL call_0_14_4_13(SH2_struct* sh)
{
  SH2movwl0(sh,14,4);
}
#define init_0_14_5_13
void FASTCALL call_0_14_5_13(SH2_struct* sh)
{
  SH2movwl0(sh,14,5);
}
#define init_0_14_6_13
void FASTCALL call_0_14_6_13(SH2_struct* sh)
{
  SH2movwl0(sh,14,6);
}
#define init_0_14_7_13
void FASTCALL call_0_14_7_13(SH2_struct* sh)
{
  SH2movwl0(sh,14,7);
}
#define init_0_14_8_13
void FASTCALL call_0_14_8_13(SH2_struct* sh)
{
  SH2movwl0(sh,14,8);
}
#define init_0_14_9_13
void FASTCALL call_0_14_9_13(SH2_struct* sh)
{
  SH2movwl0(sh,14,9);
}
#define init_0_14_10_13
void FASTCALL call_0_14_10_13(SH2_struct* sh)
{
  SH2movwl0(sh,14,10);
}
#define init_0_14_11_13
void FASTCALL call_0_14_11_13(SH2_struct* sh)
{
  SH2movwl0(sh,14,11);
}
#define init_0_14_12_13
void FASTCALL call_0_14_12_13(SH2_struct* sh)
{
  SH2movwl0(sh,14,12);
}
#define init_0_14_13_13
void FASTCALL call_0_14_13_13(SH2_struct* sh)
{
  SH2movwl0(sh,14,13);
}
#define init_0_14_14_13
void FASTCALL call_0_14_14_13(SH2_struct* sh)
{
  SH2movwl0(sh,14,14);
}
#define init_0_14_15_13
void FASTCALL call_0_14_15_13(SH2_struct* sh)
{
  SH2movwl0(sh,14,15);
}
#define init_0_15_0_13
void FASTCALL call_0_15_0_13(SH2_struct* sh)
{
  SH2movwl0(sh,15,0);
}
#define init_0_15_1_13
void FASTCALL call_0_15_1_13(SH2_struct* sh)
{
  SH2movwl0(sh,15,1);
}
#define init_0_15_2_13
void FASTCALL call_0_15_2_13(SH2_struct* sh)
{
  SH2movwl0(sh,15,2);
}
#define init_0_15_3_13
void FASTCALL call_0_15_3_13(SH2_struct* sh)
{
  SH2movwl0(sh,15,3);
}
#define init_0_15_4_13
void FASTCALL call_0_15_4_13(SH2_struct* sh)
{
  SH2movwl0(sh,15,4);
}
#define init_0_15_5_13
void FASTCALL call_0_15_5_13(SH2_struct* sh)
{
  SH2movwl0(sh,15,5);
}
#define init_0_15_6_13
void FASTCALL call_0_15_6_13(SH2_struct* sh)
{
  SH2movwl0(sh,15,6);
}
#define init_0_15_7_13
void FASTCALL call_0_15_7_13(SH2_struct* sh)
{
  SH2movwl0(sh,15,7);
}
#define init_0_15_8_13
void FASTCALL call_0_15_8_13(SH2_struct* sh)
{
  SH2movwl0(sh,15,8);
}
#define init_0_15_9_13
void FASTCALL call_0_15_9_13(SH2_struct* sh)
{
  SH2movwl0(sh,15,9);
}
#define init_0_15_10_13
void FASTCALL call_0_15_10_13(SH2_struct* sh)
{
  SH2movwl0(sh,15,10);
}
#define init_0_15_11_13
void FASTCALL call_0_15_11_13(SH2_struct* sh)
{
  SH2movwl0(sh,15,11);
}
#define init_0_15_12_13
void FASTCALL call_0_15_12_13(SH2_struct* sh)
{
  SH2movwl0(sh,15,12);
}
#define init_0_15_13_13
void FASTCALL call_0_15_13_13(SH2_struct* sh)
{
  SH2movwl0(sh,15,13);
}
#define init_0_15_14_13
void FASTCALL call_0_15_14_13(SH2_struct* sh)
{
  SH2movwl0(sh,15,14);
}
#define init_0_15_15_13
void FASTCALL call_0_15_15_13(SH2_struct* sh)
{
  SH2movwl0(sh,15,15);
}
#define init_0_0_0_6
void FASTCALL call_0_0_0_6(SH2_struct* sh)
{
  SH2movls0(sh,0,0);
}
#define init_0_0_1_6
void FASTCALL call_0_0_1_6(SH2_struct* sh)
{
  SH2movls0(sh,0,1);
}
#define init_0_0_2_6
void FASTCALL call_0_0_2_6(SH2_struct* sh)
{
  SH2movls0(sh,0,2);
}
#define init_0_0_3_6
void FASTCALL call_0_0_3_6(SH2_struct* sh)
{
  SH2movls0(sh,0,3);
}
#define init_0_0_4_6
void FASTCALL call_0_0_4_6(SH2_struct* sh)
{
  SH2movls0(sh,0,4);
}
#define init_0_0_5_6
void FASTCALL call_0_0_5_6(SH2_struct* sh)
{
  SH2movls0(sh,0,5);
}
#define init_0_0_6_6
void FASTCALL call_0_0_6_6(SH2_struct* sh)
{
  SH2movls0(sh,0,6);
}
#define init_0_0_7_6
void FASTCALL call_0_0_7_6(SH2_struct* sh)
{
  SH2movls0(sh,0,7);
}
#define init_0_0_8_6
void FASTCALL call_0_0_8_6(SH2_struct* sh)
{
  SH2movls0(sh,0,8);
}
#define init_0_0_9_6
void FASTCALL call_0_0_9_6(SH2_struct* sh)
{
  SH2movls0(sh,0,9);
}
#define init_0_0_10_6
void FASTCALL call_0_0_10_6(SH2_struct* sh)
{
  SH2movls0(sh,0,10);
}
#define init_0_0_11_6
void FASTCALL call_0_0_11_6(SH2_struct* sh)
{
  SH2movls0(sh,0,11);
}
#define init_0_0_12_6
void FASTCALL call_0_0_12_6(SH2_struct* sh)
{
  SH2movls0(sh,0,12);
}
#define init_0_0_13_6
void FASTCALL call_0_0_13_6(SH2_struct* sh)
{
  SH2movls0(sh,0,13);
}
#define init_0_0_14_6
void FASTCALL call_0_0_14_6(SH2_struct* sh)
{
  SH2movls0(sh,0,14);
}
#define init_0_0_15_6
void FASTCALL call_0_0_15_6(SH2_struct* sh)
{
  SH2movls0(sh,0,15);
}
#define init_0_1_0_6
void FASTCALL call_0_1_0_6(SH2_struct* sh)
{
  SH2movls0(sh,1,0);
}
#define init_0_1_1_6
void FASTCALL call_0_1_1_6(SH2_struct* sh)
{
  SH2movls0(sh,1,1);
}
#define init_0_1_2_6
void FASTCALL call_0_1_2_6(SH2_struct* sh)
{
  SH2movls0(sh,1,2);
}
#define init_0_1_3_6
void FASTCALL call_0_1_3_6(SH2_struct* sh)
{
  SH2movls0(sh,1,3);
}
#define init_0_1_4_6
void FASTCALL call_0_1_4_6(SH2_struct* sh)
{
  SH2movls0(sh,1,4);
}
#define init_0_1_5_6
void FASTCALL call_0_1_5_6(SH2_struct* sh)
{
  SH2movls0(sh,1,5);
}
#define init_0_1_6_6
void FASTCALL call_0_1_6_6(SH2_struct* sh)
{
  SH2movls0(sh,1,6);
}
#define init_0_1_7_6
void FASTCALL call_0_1_7_6(SH2_struct* sh)
{
  SH2movls0(sh,1,7);
}
#define init_0_1_8_6
void FASTCALL call_0_1_8_6(SH2_struct* sh)
{
  SH2movls0(sh,1,8);
}
#define init_0_1_9_6
void FASTCALL call_0_1_9_6(SH2_struct* sh)
{
  SH2movls0(sh,1,9);
}
#define init_0_1_10_6
void FASTCALL call_0_1_10_6(SH2_struct* sh)
{
  SH2movls0(sh,1,10);
}
#define init_0_1_11_6
void FASTCALL call_0_1_11_6(SH2_struct* sh)
{
  SH2movls0(sh,1,11);
}
#define init_0_1_12_6
void FASTCALL call_0_1_12_6(SH2_struct* sh)
{
  SH2movls0(sh,1,12);
}
#define init_0_1_13_6
void FASTCALL call_0_1_13_6(SH2_struct* sh)
{
  SH2movls0(sh,1,13);
}
#define init_0_1_14_6
void FASTCALL call_0_1_14_6(SH2_struct* sh)
{
  SH2movls0(sh,1,14);
}
#define init_0_1_15_6
void FASTCALL call_0_1_15_6(SH2_struct* sh)
{
  SH2movls0(sh,1,15);
}
#define init_0_2_0_6
void FASTCALL call_0_2_0_6(SH2_struct* sh)
{
  SH2movls0(sh,2,0);
}
#define init_0_2_1_6
void FASTCALL call_0_2_1_6(SH2_struct* sh)
{
  SH2movls0(sh,2,1);
}
#define init_0_2_2_6
void FASTCALL call_0_2_2_6(SH2_struct* sh)
{
  SH2movls0(sh,2,2);
}
#define init_0_2_3_6
void FASTCALL call_0_2_3_6(SH2_struct* sh)
{
  SH2movls0(sh,2,3);
}
#define init_0_2_4_6
void FASTCALL call_0_2_4_6(SH2_struct* sh)
{
  SH2movls0(sh,2,4);
}
#define init_0_2_5_6
void FASTCALL call_0_2_5_6(SH2_struct* sh)
{
  SH2movls0(sh,2,5);
}
#define init_0_2_6_6
void FASTCALL call_0_2_6_6(SH2_struct* sh)
{
  SH2movls0(sh,2,6);
}
#define init_0_2_7_6
void FASTCALL call_0_2_7_6(SH2_struct* sh)
{
  SH2movls0(sh,2,7);
}
#define init_0_2_8_6
void FASTCALL call_0_2_8_6(SH2_struct* sh)
{
  SH2movls0(sh,2,8);
}
#define init_0_2_9_6
void FASTCALL call_0_2_9_6(SH2_struct* sh)
{
  SH2movls0(sh,2,9);
}
#define init_0_2_10_6
void FASTCALL call_0_2_10_6(SH2_struct* sh)
{
  SH2movls0(sh,2,10);
}
#define init_0_2_11_6
void FASTCALL call_0_2_11_6(SH2_struct* sh)
{
  SH2movls0(sh,2,11);
}
#define init_0_2_12_6
void FASTCALL call_0_2_12_6(SH2_struct* sh)
{
  SH2movls0(sh,2,12);
}
#define init_0_2_13_6
void FASTCALL call_0_2_13_6(SH2_struct* sh)
{
  SH2movls0(sh,2,13);
}
#define init_0_2_14_6
void FASTCALL call_0_2_14_6(SH2_struct* sh)
{
  SH2movls0(sh,2,14);
}
#define init_0_2_15_6
void FASTCALL call_0_2_15_6(SH2_struct* sh)
{
  SH2movls0(sh,2,15);
}
#define init_0_3_0_6
void FASTCALL call_0_3_0_6(SH2_struct* sh)
{
  SH2movls0(sh,3,0);
}
#define init_0_3_1_6
void FASTCALL call_0_3_1_6(SH2_struct* sh)
{
  SH2movls0(sh,3,1);
}
#define init_0_3_2_6
void FASTCALL call_0_3_2_6(SH2_struct* sh)
{
  SH2movls0(sh,3,2);
}
#define init_0_3_3_6
void FASTCALL call_0_3_3_6(SH2_struct* sh)
{
  SH2movls0(sh,3,3);
}
#define init_0_3_4_6
void FASTCALL call_0_3_4_6(SH2_struct* sh)
{
  SH2movls0(sh,3,4);
}
#define init_0_3_5_6
void FASTCALL call_0_3_5_6(SH2_struct* sh)
{
  SH2movls0(sh,3,5);
}
#define init_0_3_6_6
void FASTCALL call_0_3_6_6(SH2_struct* sh)
{
  SH2movls0(sh,3,6);
}
#define init_0_3_7_6
void FASTCALL call_0_3_7_6(SH2_struct* sh)
{
  SH2movls0(sh,3,7);
}
#define init_0_3_8_6
void FASTCALL call_0_3_8_6(SH2_struct* sh)
{
  SH2movls0(sh,3,8);
}
#define init_0_3_9_6
void FASTCALL call_0_3_9_6(SH2_struct* sh)
{
  SH2movls0(sh,3,9);
}
#define init_0_3_10_6
void FASTCALL call_0_3_10_6(SH2_struct* sh)
{
  SH2movls0(sh,3,10);
}
#define init_0_3_11_6
void FASTCALL call_0_3_11_6(SH2_struct* sh)
{
  SH2movls0(sh,3,11);
}
#define init_0_3_12_6
void FASTCALL call_0_3_12_6(SH2_struct* sh)
{
  SH2movls0(sh,3,12);
}
#define init_0_3_13_6
void FASTCALL call_0_3_13_6(SH2_struct* sh)
{
  SH2movls0(sh,3,13);
}
#define init_0_3_14_6
void FASTCALL call_0_3_14_6(SH2_struct* sh)
{
  SH2movls0(sh,3,14);
}
#define init_0_3_15_6
void FASTCALL call_0_3_15_6(SH2_struct* sh)
{
  SH2movls0(sh,3,15);
}
#define init_0_4_0_6
void FASTCALL call_0_4_0_6(SH2_struct* sh)
{
  SH2movls0(sh,4,0);
}
#define init_0_4_1_6
void FASTCALL call_0_4_1_6(SH2_struct* sh)
{
  SH2movls0(sh,4,1);
}
#define init_0_4_2_6
void FASTCALL call_0_4_2_6(SH2_struct* sh)
{
  SH2movls0(sh,4,2);
}
#define init_0_4_3_6
void FASTCALL call_0_4_3_6(SH2_struct* sh)
{
  SH2movls0(sh,4,3);
}
#define init_0_4_4_6
void FASTCALL call_0_4_4_6(SH2_struct* sh)
{
  SH2movls0(sh,4,4);
}
#define init_0_4_5_6
void FASTCALL call_0_4_5_6(SH2_struct* sh)
{
  SH2movls0(sh,4,5);
}
#define init_0_4_6_6
void FASTCALL call_0_4_6_6(SH2_struct* sh)
{
  SH2movls0(sh,4,6);
}
#define init_0_4_7_6
void FASTCALL call_0_4_7_6(SH2_struct* sh)
{
  SH2movls0(sh,4,7);
}
#define init_0_4_8_6
void FASTCALL call_0_4_8_6(SH2_struct* sh)
{
  SH2movls0(sh,4,8);
}
#define init_0_4_9_6
void FASTCALL call_0_4_9_6(SH2_struct* sh)
{
  SH2movls0(sh,4,9);
}
#define init_0_4_10_6
void FASTCALL call_0_4_10_6(SH2_struct* sh)
{
  SH2movls0(sh,4,10);
}
#define init_0_4_11_6
void FASTCALL call_0_4_11_6(SH2_struct* sh)
{
  SH2movls0(sh,4,11);
}
#define init_0_4_12_6
void FASTCALL call_0_4_12_6(SH2_struct* sh)
{
  SH2movls0(sh,4,12);
}
#define init_0_4_13_6
void FASTCALL call_0_4_13_6(SH2_struct* sh)
{
  SH2movls0(sh,4,13);
}
#define init_0_4_14_6
void FASTCALL call_0_4_14_6(SH2_struct* sh)
{
  SH2movls0(sh,4,14);
}
#define init_0_4_15_6
void FASTCALL call_0_4_15_6(SH2_struct* sh)
{
  SH2movls0(sh,4,15);
}
#define init_0_5_0_6
void FASTCALL call_0_5_0_6(SH2_struct* sh)
{
  SH2movls0(sh,5,0);
}
#define init_0_5_1_6
void FASTCALL call_0_5_1_6(SH2_struct* sh)
{
  SH2movls0(sh,5,1);
}
#define init_0_5_2_6
void FASTCALL call_0_5_2_6(SH2_struct* sh)
{
  SH2movls0(sh,5,2);
}
#define init_0_5_3_6
void FASTCALL call_0_5_3_6(SH2_struct* sh)
{
  SH2movls0(sh,5,3);
}
#define init_0_5_4_6
void FASTCALL call_0_5_4_6(SH2_struct* sh)
{
  SH2movls0(sh,5,4);
}
#define init_0_5_5_6
void FASTCALL call_0_5_5_6(SH2_struct* sh)
{
  SH2movls0(sh,5,5);
}
#define init_0_5_6_6
void FASTCALL call_0_5_6_6(SH2_struct* sh)
{
  SH2movls0(sh,5,6);
}
#define init_0_5_7_6
void FASTCALL call_0_5_7_6(SH2_struct* sh)
{
  SH2movls0(sh,5,7);
}
#define init_0_5_8_6
void FASTCALL call_0_5_8_6(SH2_struct* sh)
{
  SH2movls0(sh,5,8);
}
#define init_0_5_9_6
void FASTCALL call_0_5_9_6(SH2_struct* sh)
{
  SH2movls0(sh,5,9);
}
#define init_0_5_10_6
void FASTCALL call_0_5_10_6(SH2_struct* sh)
{
  SH2movls0(sh,5,10);
}
#define init_0_5_11_6
void FASTCALL call_0_5_11_6(SH2_struct* sh)
{
  SH2movls0(sh,5,11);
}
#define init_0_5_12_6
void FASTCALL call_0_5_12_6(SH2_struct* sh)
{
  SH2movls0(sh,5,12);
}
#define init_0_5_13_6
void FASTCALL call_0_5_13_6(SH2_struct* sh)
{
  SH2movls0(sh,5,13);
}
#define init_0_5_14_6
void FASTCALL call_0_5_14_6(SH2_struct* sh)
{
  SH2movls0(sh,5,14);
}
#define init_0_5_15_6
void FASTCALL call_0_5_15_6(SH2_struct* sh)
{
  SH2movls0(sh,5,15);
}
#define init_0_6_0_6
void FASTCALL call_0_6_0_6(SH2_struct* sh)
{
  SH2movls0(sh,6,0);
}
#define init_0_6_1_6
void FASTCALL call_0_6_1_6(SH2_struct* sh)
{
  SH2movls0(sh,6,1);
}
#define init_0_6_2_6
void FASTCALL call_0_6_2_6(SH2_struct* sh)
{
  SH2movls0(sh,6,2);
}
#define init_0_6_3_6
void FASTCALL call_0_6_3_6(SH2_struct* sh)
{
  SH2movls0(sh,6,3);
}
#define init_0_6_4_6
void FASTCALL call_0_6_4_6(SH2_struct* sh)
{
  SH2movls0(sh,6,4);
}
#define init_0_6_5_6
void FASTCALL call_0_6_5_6(SH2_struct* sh)
{
  SH2movls0(sh,6,5);
}
#define init_0_6_6_6
void FASTCALL call_0_6_6_6(SH2_struct* sh)
{
  SH2movls0(sh,6,6);
}
#define init_0_6_7_6
void FASTCALL call_0_6_7_6(SH2_struct* sh)
{
  SH2movls0(sh,6,7);
}
#define init_0_6_8_6
void FASTCALL call_0_6_8_6(SH2_struct* sh)
{
  SH2movls0(sh,6,8);
}
#define init_0_6_9_6
void FASTCALL call_0_6_9_6(SH2_struct* sh)
{
  SH2movls0(sh,6,9);
}
#define init_0_6_10_6
void FASTCALL call_0_6_10_6(SH2_struct* sh)
{
  SH2movls0(sh,6,10);
}
#define init_0_6_11_6
void FASTCALL call_0_6_11_6(SH2_struct* sh)
{
  SH2movls0(sh,6,11);
}
#define init_0_6_12_6
void FASTCALL call_0_6_12_6(SH2_struct* sh)
{
  SH2movls0(sh,6,12);
}
#define init_0_6_13_6
void FASTCALL call_0_6_13_6(SH2_struct* sh)
{
  SH2movls0(sh,6,13);
}
#define init_0_6_14_6
void FASTCALL call_0_6_14_6(SH2_struct* sh)
{
  SH2movls0(sh,6,14);
}
#define init_0_6_15_6
void FASTCALL call_0_6_15_6(SH2_struct* sh)
{
  SH2movls0(sh,6,15);
}
#define init_0_7_0_6
void FASTCALL call_0_7_0_6(SH2_struct* sh)
{
  SH2movls0(sh,7,0);
}
#define init_0_7_1_6
void FASTCALL call_0_7_1_6(SH2_struct* sh)
{
  SH2movls0(sh,7,1);
}
#define init_0_7_2_6
void FASTCALL call_0_7_2_6(SH2_struct* sh)
{
  SH2movls0(sh,7,2);
}
#define init_0_7_3_6
void FASTCALL call_0_7_3_6(SH2_struct* sh)
{
  SH2movls0(sh,7,3);
}
#define init_0_7_4_6
void FASTCALL call_0_7_4_6(SH2_struct* sh)
{
  SH2movls0(sh,7,4);
}
#define init_0_7_5_6
void FASTCALL call_0_7_5_6(SH2_struct* sh)
{
  SH2movls0(sh,7,5);
}
#define init_0_7_6_6
void FASTCALL call_0_7_6_6(SH2_struct* sh)
{
  SH2movls0(sh,7,6);
}
#define init_0_7_7_6
void FASTCALL call_0_7_7_6(SH2_struct* sh)
{
  SH2movls0(sh,7,7);
}
#define init_0_7_8_6
void FASTCALL call_0_7_8_6(SH2_struct* sh)
{
  SH2movls0(sh,7,8);
}
#define init_0_7_9_6
void FASTCALL call_0_7_9_6(SH2_struct* sh)
{
  SH2movls0(sh,7,9);
}
#define init_0_7_10_6
void FASTCALL call_0_7_10_6(SH2_struct* sh)
{
  SH2movls0(sh,7,10);
}
#define init_0_7_11_6
void FASTCALL call_0_7_11_6(SH2_struct* sh)
{
  SH2movls0(sh,7,11);
}
#define init_0_7_12_6
void FASTCALL call_0_7_12_6(SH2_struct* sh)
{
  SH2movls0(sh,7,12);
}
#define init_0_7_13_6
void FASTCALL call_0_7_13_6(SH2_struct* sh)
{
  SH2movls0(sh,7,13);
}
#define init_0_7_14_6
void FASTCALL call_0_7_14_6(SH2_struct* sh)
{
  SH2movls0(sh,7,14);
}
#define init_0_7_15_6
void FASTCALL call_0_7_15_6(SH2_struct* sh)
{
  SH2movls0(sh,7,15);
}
#define init_0_8_0_6
void FASTCALL call_0_8_0_6(SH2_struct* sh)
{
  SH2movls0(sh,8,0);
}
#define init_0_8_1_6
void FASTCALL call_0_8_1_6(SH2_struct* sh)
{
  SH2movls0(sh,8,1);
}
#define init_0_8_2_6
void FASTCALL call_0_8_2_6(SH2_struct* sh)
{
  SH2movls0(sh,8,2);
}
#define init_0_8_3_6
void FASTCALL call_0_8_3_6(SH2_struct* sh)
{
  SH2movls0(sh,8,3);
}
#define init_0_8_4_6
void FASTCALL call_0_8_4_6(SH2_struct* sh)
{
  SH2movls0(sh,8,4);
}
#define init_0_8_5_6
void FASTCALL call_0_8_5_6(SH2_struct* sh)
{
  SH2movls0(sh,8,5);
}
#define init_0_8_6_6
void FASTCALL call_0_8_6_6(SH2_struct* sh)
{
  SH2movls0(sh,8,6);
}
#define init_0_8_7_6
void FASTCALL call_0_8_7_6(SH2_struct* sh)
{
  SH2movls0(sh,8,7);
}
#define init_0_8_8_6
void FASTCALL call_0_8_8_6(SH2_struct* sh)
{
  SH2movls0(sh,8,8);
}
#define init_0_8_9_6
void FASTCALL call_0_8_9_6(SH2_struct* sh)
{
  SH2movls0(sh,8,9);
}
#define init_0_8_10_6
void FASTCALL call_0_8_10_6(SH2_struct* sh)
{
  SH2movls0(sh,8,10);
}
#define init_0_8_11_6
void FASTCALL call_0_8_11_6(SH2_struct* sh)
{
  SH2movls0(sh,8,11);
}
#define init_0_8_12_6
void FASTCALL call_0_8_12_6(SH2_struct* sh)
{
  SH2movls0(sh,8,12);
}
#define init_0_8_13_6
void FASTCALL call_0_8_13_6(SH2_struct* sh)
{
  SH2movls0(sh,8,13);
}
#define init_0_8_14_6
void FASTCALL call_0_8_14_6(SH2_struct* sh)
{
  SH2movls0(sh,8,14);
}
#define init_0_8_15_6
void FASTCALL call_0_8_15_6(SH2_struct* sh)
{
  SH2movls0(sh,8,15);
}
#define init_0_9_0_6
void FASTCALL call_0_9_0_6(SH2_struct* sh)
{
  SH2movls0(sh,9,0);
}
#define init_0_9_1_6
void FASTCALL call_0_9_1_6(SH2_struct* sh)
{
  SH2movls0(sh,9,1);
}
#define init_0_9_2_6
void FASTCALL call_0_9_2_6(SH2_struct* sh)
{
  SH2movls0(sh,9,2);
}
#define init_0_9_3_6
void FASTCALL call_0_9_3_6(SH2_struct* sh)
{
  SH2movls0(sh,9,3);
}
#define init_0_9_4_6
void FASTCALL call_0_9_4_6(SH2_struct* sh)
{
  SH2movls0(sh,9,4);
}
#define init_0_9_5_6
void FASTCALL call_0_9_5_6(SH2_struct* sh)
{
  SH2movls0(sh,9,5);
}
#define init_0_9_6_6
void FASTCALL call_0_9_6_6(SH2_struct* sh)
{
  SH2movls0(sh,9,6);
}
#define init_0_9_7_6
void FASTCALL call_0_9_7_6(SH2_struct* sh)
{
  SH2movls0(sh,9,7);
}
#define init_0_9_8_6
void FASTCALL call_0_9_8_6(SH2_struct* sh)
{
  SH2movls0(sh,9,8);
}
#define init_0_9_9_6
void FASTCALL call_0_9_9_6(SH2_struct* sh)
{
  SH2movls0(sh,9,9);
}
#define init_0_9_10_6
void FASTCALL call_0_9_10_6(SH2_struct* sh)
{
  SH2movls0(sh,9,10);
}
#define init_0_9_11_6
void FASTCALL call_0_9_11_6(SH2_struct* sh)
{
  SH2movls0(sh,9,11);
}
#define init_0_9_12_6
void FASTCALL call_0_9_12_6(SH2_struct* sh)
{
  SH2movls0(sh,9,12);
}
#define init_0_9_13_6
void FASTCALL call_0_9_13_6(SH2_struct* sh)
{
  SH2movls0(sh,9,13);
}
#define init_0_9_14_6
void FASTCALL call_0_9_14_6(SH2_struct* sh)
{
  SH2movls0(sh,9,14);
}
#define init_0_9_15_6
void FASTCALL call_0_9_15_6(SH2_struct* sh)
{
  SH2movls0(sh,9,15);
}
#define init_0_10_0_6
void FASTCALL call_0_10_0_6(SH2_struct* sh)
{
  SH2movls0(sh,10,0);
}
#define init_0_10_1_6
void FASTCALL call_0_10_1_6(SH2_struct* sh)
{
  SH2movls0(sh,10,1);
}
#define init_0_10_2_6
void FASTCALL call_0_10_2_6(SH2_struct* sh)
{
  SH2movls0(sh,10,2);
}
#define init_0_10_3_6
void FASTCALL call_0_10_3_6(SH2_struct* sh)
{
  SH2movls0(sh,10,3);
}
#define init_0_10_4_6
void FASTCALL call_0_10_4_6(SH2_struct* sh)
{
  SH2movls0(sh,10,4);
}
#define init_0_10_5_6
void FASTCALL call_0_10_5_6(SH2_struct* sh)
{
  SH2movls0(sh,10,5);
}
#define init_0_10_6_6
void FASTCALL call_0_10_6_6(SH2_struct* sh)
{
  SH2movls0(sh,10,6);
}
#define init_0_10_7_6
void FASTCALL call_0_10_7_6(SH2_struct* sh)
{
  SH2movls0(sh,10,7);
}
#define init_0_10_8_6
void FASTCALL call_0_10_8_6(SH2_struct* sh)
{
  SH2movls0(sh,10,8);
}
#define init_0_10_9_6
void FASTCALL call_0_10_9_6(SH2_struct* sh)
{
  SH2movls0(sh,10,9);
}
#define init_0_10_10_6
void FASTCALL call_0_10_10_6(SH2_struct* sh)
{
  SH2movls0(sh,10,10);
}
#define init_0_10_11_6
void FASTCALL call_0_10_11_6(SH2_struct* sh)
{
  SH2movls0(sh,10,11);
}
#define init_0_10_12_6
void FASTCALL call_0_10_12_6(SH2_struct* sh)
{
  SH2movls0(sh,10,12);
}
#define init_0_10_13_6
void FASTCALL call_0_10_13_6(SH2_struct* sh)
{
  SH2movls0(sh,10,13);
}
#define init_0_10_14_6
void FASTCALL call_0_10_14_6(SH2_struct* sh)
{
  SH2movls0(sh,10,14);
}
#define init_0_10_15_6
void FASTCALL call_0_10_15_6(SH2_struct* sh)
{
  SH2movls0(sh,10,15);
}
#define init_0_11_0_6
void FASTCALL call_0_11_0_6(SH2_struct* sh)
{
  SH2movls0(sh,11,0);
}
#define init_0_11_1_6
void FASTCALL call_0_11_1_6(SH2_struct* sh)
{
  SH2movls0(sh,11,1);
}
#define init_0_11_2_6
void FASTCALL call_0_11_2_6(SH2_struct* sh)
{
  SH2movls0(sh,11,2);
}
#define init_0_11_3_6
void FASTCALL call_0_11_3_6(SH2_struct* sh)
{
  SH2movls0(sh,11,3);
}
#define init_0_11_4_6
void FASTCALL call_0_11_4_6(SH2_struct* sh)
{
  SH2movls0(sh,11,4);
}
#define init_0_11_5_6
void FASTCALL call_0_11_5_6(SH2_struct* sh)
{
  SH2movls0(sh,11,5);
}
#define init_0_11_6_6
void FASTCALL call_0_11_6_6(SH2_struct* sh)
{
  SH2movls0(sh,11,6);
}
#define init_0_11_7_6
void FASTCALL call_0_11_7_6(SH2_struct* sh)
{
  SH2movls0(sh,11,7);
}
#define init_0_11_8_6
void FASTCALL call_0_11_8_6(SH2_struct* sh)
{
  SH2movls0(sh,11,8);
}
#define init_0_11_9_6
void FASTCALL call_0_11_9_6(SH2_struct* sh)
{
  SH2movls0(sh,11,9);
}
#define init_0_11_10_6
void FASTCALL call_0_11_10_6(SH2_struct* sh)
{
  SH2movls0(sh,11,10);
}
#define init_0_11_11_6
void FASTCALL call_0_11_11_6(SH2_struct* sh)
{
  SH2movls0(sh,11,11);
}
#define init_0_11_12_6
void FASTCALL call_0_11_12_6(SH2_struct* sh)
{
  SH2movls0(sh,11,12);
}
#define init_0_11_13_6
void FASTCALL call_0_11_13_6(SH2_struct* sh)
{
  SH2movls0(sh,11,13);
}
#define init_0_11_14_6
void FASTCALL call_0_11_14_6(SH2_struct* sh)
{
  SH2movls0(sh,11,14);
}
#define init_0_11_15_6
void FASTCALL call_0_11_15_6(SH2_struct* sh)
{
  SH2movls0(sh,11,15);
}
#define init_0_12_0_6
void FASTCALL call_0_12_0_6(SH2_struct* sh)
{
  SH2movls0(sh,12,0);
}
#define init_0_12_1_6
void FASTCALL call_0_12_1_6(SH2_struct* sh)
{
  SH2movls0(sh,12,1);
}
#define init_0_12_2_6
void FASTCALL call_0_12_2_6(SH2_struct* sh)
{
  SH2movls0(sh,12,2);
}
#define init_0_12_3_6
void FASTCALL call_0_12_3_6(SH2_struct* sh)
{
  SH2movls0(sh,12,3);
}
#define init_0_12_4_6
void FASTCALL call_0_12_4_6(SH2_struct* sh)
{
  SH2movls0(sh,12,4);
}
#define init_0_12_5_6
void FASTCALL call_0_12_5_6(SH2_struct* sh)
{
  SH2movls0(sh,12,5);
}
#define init_0_12_6_6
void FASTCALL call_0_12_6_6(SH2_struct* sh)
{
  SH2movls0(sh,12,6);
}
#define init_0_12_7_6
void FASTCALL call_0_12_7_6(SH2_struct* sh)
{
  SH2movls0(sh,12,7);
}
#define init_0_12_8_6
void FASTCALL call_0_12_8_6(SH2_struct* sh)
{
  SH2movls0(sh,12,8);
}
#define init_0_12_9_6
void FASTCALL call_0_12_9_6(SH2_struct* sh)
{
  SH2movls0(sh,12,9);
}
#define init_0_12_10_6
void FASTCALL call_0_12_10_6(SH2_struct* sh)
{
  SH2movls0(sh,12,10);
}
#define init_0_12_11_6
void FASTCALL call_0_12_11_6(SH2_struct* sh)
{
  SH2movls0(sh,12,11);
}
#define init_0_12_12_6
void FASTCALL call_0_12_12_6(SH2_struct* sh)
{
  SH2movls0(sh,12,12);
}
#define init_0_12_13_6
void FASTCALL call_0_12_13_6(SH2_struct* sh)
{
  SH2movls0(sh,12,13);
}
#define init_0_12_14_6
void FASTCALL call_0_12_14_6(SH2_struct* sh)
{
  SH2movls0(sh,12,14);
}
#define init_0_12_15_6
void FASTCALL call_0_12_15_6(SH2_struct* sh)
{
  SH2movls0(sh,12,15);
}
#define init_0_13_0_6
void FASTCALL call_0_13_0_6(SH2_struct* sh)
{
  SH2movls0(sh,13,0);
}
#define init_0_13_1_6
void FASTCALL call_0_13_1_6(SH2_struct* sh)
{
  SH2movls0(sh,13,1);
}
#define init_0_13_2_6
void FASTCALL call_0_13_2_6(SH2_struct* sh)
{
  SH2movls0(sh,13,2);
}
#define init_0_13_3_6
void FASTCALL call_0_13_3_6(SH2_struct* sh)
{
  SH2movls0(sh,13,3);
}
#define init_0_13_4_6
void FASTCALL call_0_13_4_6(SH2_struct* sh)
{
  SH2movls0(sh,13,4);
}
#define init_0_13_5_6
void FASTCALL call_0_13_5_6(SH2_struct* sh)
{
  SH2movls0(sh,13,5);
}
#define init_0_13_6_6
void FASTCALL call_0_13_6_6(SH2_struct* sh)
{
  SH2movls0(sh,13,6);
}
#define init_0_13_7_6
void FASTCALL call_0_13_7_6(SH2_struct* sh)
{
  SH2movls0(sh,13,7);
}
#define init_0_13_8_6
void FASTCALL call_0_13_8_6(SH2_struct* sh)
{
  SH2movls0(sh,13,8);
}
#define init_0_13_9_6
void FASTCALL call_0_13_9_6(SH2_struct* sh)
{
  SH2movls0(sh,13,9);
}
#define init_0_13_10_6
void FASTCALL call_0_13_10_6(SH2_struct* sh)
{
  SH2movls0(sh,13,10);
}
#define init_0_13_11_6
void FASTCALL call_0_13_11_6(SH2_struct* sh)
{
  SH2movls0(sh,13,11);
}
#define init_0_13_12_6
void FASTCALL call_0_13_12_6(SH2_struct* sh)
{
  SH2movls0(sh,13,12);
}
#define init_0_13_13_6
void FASTCALL call_0_13_13_6(SH2_struct* sh)
{
  SH2movls0(sh,13,13);
}
#define init_0_13_14_6
void FASTCALL call_0_13_14_6(SH2_struct* sh)
{
  SH2movls0(sh,13,14);
}
#define init_0_13_15_6
void FASTCALL call_0_13_15_6(SH2_struct* sh)
{
  SH2movls0(sh,13,15);
}
#define init_0_14_0_6
void FASTCALL call_0_14_0_6(SH2_struct* sh)
{
  SH2movls0(sh,14,0);
}
#define init_0_14_1_6
void FASTCALL call_0_14_1_6(SH2_struct* sh)
{
  SH2movls0(sh,14,1);
}
#define init_0_14_2_6
void FASTCALL call_0_14_2_6(SH2_struct* sh)
{
  SH2movls0(sh,14,2);
}
#define init_0_14_3_6
void FASTCALL call_0_14_3_6(SH2_struct* sh)
{
  SH2movls0(sh,14,3);
}
#define init_0_14_4_6
void FASTCALL call_0_14_4_6(SH2_struct* sh)
{
  SH2movls0(sh,14,4);
}
#define init_0_14_5_6
void FASTCALL call_0_14_5_6(SH2_struct* sh)
{
  SH2movls0(sh,14,5);
}
#define init_0_14_6_6
void FASTCALL call_0_14_6_6(SH2_struct* sh)
{
  SH2movls0(sh,14,6);
}
#define init_0_14_7_6
void FASTCALL call_0_14_7_6(SH2_struct* sh)
{
  SH2movls0(sh,14,7);
}
#define init_0_14_8_6
void FASTCALL call_0_14_8_6(SH2_struct* sh)
{
  SH2movls0(sh,14,8);
}
#define init_0_14_9_6
void FASTCALL call_0_14_9_6(SH2_struct* sh)
{
  SH2movls0(sh,14,9);
}
#define init_0_14_10_6
void FASTCALL call_0_14_10_6(SH2_struct* sh)
{
  SH2movls0(sh,14,10);
}
#define init_0_14_11_6
void FASTCALL call_0_14_11_6(SH2_struct* sh)
{
  SH2movls0(sh,14,11);
}
#define init_0_14_12_6
void FASTCALL call_0_14_12_6(SH2_struct* sh)
{
  SH2movls0(sh,14,12);
}
#define init_0_14_13_6
void FASTCALL call_0_14_13_6(SH2_struct* sh)
{
  SH2movls0(sh,14,13);
}
#define init_0_14_14_6
void FASTCALL call_0_14_14_6(SH2_struct* sh)
{
  SH2movls0(sh,14,14);
}
#define init_0_14_15_6
void FASTCALL call_0_14_15_6(SH2_struct* sh)
{
  SH2movls0(sh,14,15);
}
#define init_0_15_0_6
void FASTCALL call_0_15_0_6(SH2_struct* sh)
{
  SH2movls0(sh,15,0);
}
#define init_0_15_1_6
void FASTCALL call_0_15_1_6(SH2_struct* sh)
{
  SH2movls0(sh,15,1);
}
#define init_0_15_2_6
void FASTCALL call_0_15_2_6(SH2_struct* sh)
{
  SH2movls0(sh,15,2);
}
#define init_0_15_3_6
void FASTCALL call_0_15_3_6(SH2_struct* sh)
{
  SH2movls0(sh,15,3);
}
#define init_0_15_4_6
void FASTCALL call_0_15_4_6(SH2_struct* sh)
{
  SH2movls0(sh,15,4);
}
#define init_0_15_5_6
void FASTCALL call_0_15_5_6(SH2_struct* sh)
{
  SH2movls0(sh,15,5);
}
#define init_0_15_6_6
void FASTCALL call_0_15_6_6(SH2_struct* sh)
{
  SH2movls0(sh,15,6);
}
#define init_0_15_7_6
void FASTCALL call_0_15_7_6(SH2_struct* sh)
{
  SH2movls0(sh,15,7);
}
#define init_0_15_8_6
void FASTCALL call_0_15_8_6(SH2_struct* sh)
{
  SH2movls0(sh,15,8);
}
#define init_0_15_9_6
void FASTCALL call_0_15_9_6(SH2_struct* sh)
{
  SH2movls0(sh,15,9);
}
#define init_0_15_10_6
void FASTCALL call_0_15_10_6(SH2_struct* sh)
{
  SH2movls0(sh,15,10);
}
#define init_0_15_11_6
void FASTCALL call_0_15_11_6(SH2_struct* sh)
{
  SH2movls0(sh,15,11);
}
#define init_0_15_12_6
void FASTCALL call_0_15_12_6(SH2_struct* sh)
{
  SH2movls0(sh,15,12);
}
#define init_0_15_13_6
void FASTCALL call_0_15_13_6(SH2_struct* sh)
{
  SH2movls0(sh,15,13);
}
#define init_0_15_14_6
void FASTCALL call_0_15_14_6(SH2_struct* sh)
{
  SH2movls0(sh,15,14);
}
#define init_0_15_15_6
void FASTCALL call_0_15_15_6(SH2_struct* sh)
{
  SH2movls0(sh,15,15);
}
#define init_0_0_2_9
void FASTCALL call_0_0_2_9(SH2_struct* sh)
{
  SH2movt(sh,0);
}
#define init_0_1_2_9
void FASTCALL call_0_1_2_9(SH2_struct* sh)
{
  SH2movt(sh,1);
}
#define init_0_2_2_9
void FASTCALL call_0_2_2_9(SH2_struct* sh)
{
  SH2movt(sh,2);
}
#define init_0_3_2_9
void FASTCALL call_0_3_2_9(SH2_struct* sh)
{
  SH2movt(sh,3);
}
#define init_0_4_2_9
void FASTCALL call_0_4_2_9(SH2_struct* sh)
{
  SH2movt(sh,4);
}
#define init_0_5_2_9
void FASTCALL call_0_5_2_9(SH2_struct* sh)
{
  SH2movt(sh,5);
}
#define init_0_6_2_9
void FASTCALL call_0_6_2_9(SH2_struct* sh)
{
  SH2movt(sh,6);
}
#define init_0_7_2_9
void FASTCALL call_0_7_2_9(SH2_struct* sh)
{
  SH2movt(sh,7);
}
#define init_0_8_2_9
void FASTCALL call_0_8_2_9(SH2_struct* sh)
{
  SH2movt(sh,8);
}
#define init_0_9_2_9
void FASTCALL call_0_9_2_9(SH2_struct* sh)
{
  SH2movt(sh,9);
}
#define init_0_10_2_9
void FASTCALL call_0_10_2_9(SH2_struct* sh)
{
  SH2movt(sh,10);
}
#define init_0_11_2_9
void FASTCALL call_0_11_2_9(SH2_struct* sh)
{
  SH2movt(sh,11);
}
#define init_0_12_2_9
void FASTCALL call_0_12_2_9(SH2_struct* sh)
{
  SH2movt(sh,12);
}
#define init_0_13_2_9
void FASTCALL call_0_13_2_9(SH2_struct* sh)
{
  SH2movt(sh,13);
}
#define init_0_14_2_9
void FASTCALL call_0_14_2_9(SH2_struct* sh)
{
  SH2movt(sh,14);
}
#define init_0_15_2_9
void FASTCALL call_0_15_2_9(SH2_struct* sh)
{
  SH2movt(sh,15);
}
#define init_0_0_0_5
void FASTCALL call_0_0_0_5(SH2_struct* sh)
{
  SH2movws0(sh,0,0);
}
#define init_0_0_1_5
void FASTCALL call_0_0_1_5(SH2_struct* sh)
{
  SH2movws0(sh,0,1);
}
#define init_0_0_2_5
void FASTCALL call_0_0_2_5(SH2_struct* sh)
{
  SH2movws0(sh,0,2);
}
#define init_0_0_3_5
void FASTCALL call_0_0_3_5(SH2_struct* sh)
{
  SH2movws0(sh,0,3);
}
#define init_0_0_4_5
void FASTCALL call_0_0_4_5(SH2_struct* sh)
{
  SH2movws0(sh,0,4);
}
#define init_0_0_5_5
void FASTCALL call_0_0_5_5(SH2_struct* sh)
{
  SH2movws0(sh,0,5);
}
#define init_0_0_6_5
void FASTCALL call_0_0_6_5(SH2_struct* sh)
{
  SH2movws0(sh,0,6);
}
#define init_0_0_7_5
void FASTCALL call_0_0_7_5(SH2_struct* sh)
{
  SH2movws0(sh,0,7);
}
#define init_0_0_8_5
void FASTCALL call_0_0_8_5(SH2_struct* sh)
{
  SH2movws0(sh,0,8);
}
#define init_0_0_9_5
void FASTCALL call_0_0_9_5(SH2_struct* sh)
{
  SH2movws0(sh,0,9);
}
#define init_0_0_10_5
void FASTCALL call_0_0_10_5(SH2_struct* sh)
{
  SH2movws0(sh,0,10);
}
#define init_0_0_11_5
void FASTCALL call_0_0_11_5(SH2_struct* sh)
{
  SH2movws0(sh,0,11);
}
#define init_0_0_12_5
void FASTCALL call_0_0_12_5(SH2_struct* sh)
{
  SH2movws0(sh,0,12);
}
#define init_0_0_13_5
void FASTCALL call_0_0_13_5(SH2_struct* sh)
{
  SH2movws0(sh,0,13);
}
#define init_0_0_14_5
void FASTCALL call_0_0_14_5(SH2_struct* sh)
{
  SH2movws0(sh,0,14);
}
#define init_0_0_15_5
void FASTCALL call_0_0_15_5(SH2_struct* sh)
{
  SH2movws0(sh,0,15);
}
#define init_0_1_0_5
void FASTCALL call_0_1_0_5(SH2_struct* sh)
{
  SH2movws0(sh,1,0);
}
#define init_0_1_1_5
void FASTCALL call_0_1_1_5(SH2_struct* sh)
{
  SH2movws0(sh,1,1);
}
#define init_0_1_2_5
void FASTCALL call_0_1_2_5(SH2_struct* sh)
{
  SH2movws0(sh,1,2);
}
#define init_0_1_3_5
void FASTCALL call_0_1_3_5(SH2_struct* sh)
{
  SH2movws0(sh,1,3);
}
#define init_0_1_4_5
void FASTCALL call_0_1_4_5(SH2_struct* sh)
{
  SH2movws0(sh,1,4);
}
#define init_0_1_5_5
void FASTCALL call_0_1_5_5(SH2_struct* sh)
{
  SH2movws0(sh,1,5);
}
#define init_0_1_6_5
void FASTCALL call_0_1_6_5(SH2_struct* sh)
{
  SH2movws0(sh,1,6);
}
#define init_0_1_7_5
void FASTCALL call_0_1_7_5(SH2_struct* sh)
{
  SH2movws0(sh,1,7);
}
#define init_0_1_8_5
void FASTCALL call_0_1_8_5(SH2_struct* sh)
{
  SH2movws0(sh,1,8);
}
#define init_0_1_9_5
void FASTCALL call_0_1_9_5(SH2_struct* sh)
{
  SH2movws0(sh,1,9);
}
#define init_0_1_10_5
void FASTCALL call_0_1_10_5(SH2_struct* sh)
{
  SH2movws0(sh,1,10);
}
#define init_0_1_11_5
void FASTCALL call_0_1_11_5(SH2_struct* sh)
{
  SH2movws0(sh,1,11);
}
#define init_0_1_12_5
void FASTCALL call_0_1_12_5(SH2_struct* sh)
{
  SH2movws0(sh,1,12);
}
#define init_0_1_13_5
void FASTCALL call_0_1_13_5(SH2_struct* sh)
{
  SH2movws0(sh,1,13);
}
#define init_0_1_14_5
void FASTCALL call_0_1_14_5(SH2_struct* sh)
{
  SH2movws0(sh,1,14);
}
#define init_0_1_15_5
void FASTCALL call_0_1_15_5(SH2_struct* sh)
{
  SH2movws0(sh,1,15);
}
#define init_0_2_0_5
void FASTCALL call_0_2_0_5(SH2_struct* sh)
{
  SH2movws0(sh,2,0);
}
#define init_0_2_1_5
void FASTCALL call_0_2_1_5(SH2_struct* sh)
{
  SH2movws0(sh,2,1);
}
#define init_0_2_2_5
void FASTCALL call_0_2_2_5(SH2_struct* sh)
{
  SH2movws0(sh,2,2);
}
#define init_0_2_3_5
void FASTCALL call_0_2_3_5(SH2_struct* sh)
{
  SH2movws0(sh,2,3);
}
#define init_0_2_4_5
void FASTCALL call_0_2_4_5(SH2_struct* sh)
{
  SH2movws0(sh,2,4);
}
#define init_0_2_5_5
void FASTCALL call_0_2_5_5(SH2_struct* sh)
{
  SH2movws0(sh,2,5);
}
#define init_0_2_6_5
void FASTCALL call_0_2_6_5(SH2_struct* sh)
{
  SH2movws0(sh,2,6);
}
#define init_0_2_7_5
void FASTCALL call_0_2_7_5(SH2_struct* sh)
{
  SH2movws0(sh,2,7);
}
#define init_0_2_8_5
void FASTCALL call_0_2_8_5(SH2_struct* sh)
{
  SH2movws0(sh,2,8);
}
#define init_0_2_9_5
void FASTCALL call_0_2_9_5(SH2_struct* sh)
{
  SH2movws0(sh,2,9);
}
#define init_0_2_10_5
void FASTCALL call_0_2_10_5(SH2_struct* sh)
{
  SH2movws0(sh,2,10);
}
#define init_0_2_11_5
void FASTCALL call_0_2_11_5(SH2_struct* sh)
{
  SH2movws0(sh,2,11);
}
#define init_0_2_12_5
void FASTCALL call_0_2_12_5(SH2_struct* sh)
{
  SH2movws0(sh,2,12);
}
#define init_0_2_13_5
void FASTCALL call_0_2_13_5(SH2_struct* sh)
{
  SH2movws0(sh,2,13);
}
#define init_0_2_14_5
void FASTCALL call_0_2_14_5(SH2_struct* sh)
{
  SH2movws0(sh,2,14);
}
#define init_0_2_15_5
void FASTCALL call_0_2_15_5(SH2_struct* sh)
{
  SH2movws0(sh,2,15);
}
#define init_0_3_0_5
void FASTCALL call_0_3_0_5(SH2_struct* sh)
{
  SH2movws0(sh,3,0);
}
#define init_0_3_1_5
void FASTCALL call_0_3_1_5(SH2_struct* sh)
{
  SH2movws0(sh,3,1);
}
#define init_0_3_2_5
void FASTCALL call_0_3_2_5(SH2_struct* sh)
{
  SH2movws0(sh,3,2);
}
#define init_0_3_3_5
void FASTCALL call_0_3_3_5(SH2_struct* sh)
{
  SH2movws0(sh,3,3);
}
#define init_0_3_4_5
void FASTCALL call_0_3_4_5(SH2_struct* sh)
{
  SH2movws0(sh,3,4);
}
#define init_0_3_5_5
void FASTCALL call_0_3_5_5(SH2_struct* sh)
{
  SH2movws0(sh,3,5);
}
#define init_0_3_6_5
void FASTCALL call_0_3_6_5(SH2_struct* sh)
{
  SH2movws0(sh,3,6);
}
#define init_0_3_7_5
void FASTCALL call_0_3_7_5(SH2_struct* sh)
{
  SH2movws0(sh,3,7);
}
#define init_0_3_8_5
void FASTCALL call_0_3_8_5(SH2_struct* sh)
{
  SH2movws0(sh,3,8);
}
#define init_0_3_9_5
void FASTCALL call_0_3_9_5(SH2_struct* sh)
{
  SH2movws0(sh,3,9);
}
#define init_0_3_10_5
void FASTCALL call_0_3_10_5(SH2_struct* sh)
{
  SH2movws0(sh,3,10);
}
#define init_0_3_11_5
void FASTCALL call_0_3_11_5(SH2_struct* sh)
{
  SH2movws0(sh,3,11);
}
#define init_0_3_12_5
void FASTCALL call_0_3_12_5(SH2_struct* sh)
{
  SH2movws0(sh,3,12);
}
#define init_0_3_13_5
void FASTCALL call_0_3_13_5(SH2_struct* sh)
{
  SH2movws0(sh,3,13);
}
#define init_0_3_14_5
void FASTCALL call_0_3_14_5(SH2_struct* sh)
{
  SH2movws0(sh,3,14);
}
#define init_0_3_15_5
void FASTCALL call_0_3_15_5(SH2_struct* sh)
{
  SH2movws0(sh,3,15);
}
#define init_0_4_0_5
void FASTCALL call_0_4_0_5(SH2_struct* sh)
{
  SH2movws0(sh,4,0);
}
#define init_0_4_1_5
void FASTCALL call_0_4_1_5(SH2_struct* sh)
{
  SH2movws0(sh,4,1);
}
#define init_0_4_2_5
void FASTCALL call_0_4_2_5(SH2_struct* sh)
{
  SH2movws0(sh,4,2);
}
#define init_0_4_3_5
void FASTCALL call_0_4_3_5(SH2_struct* sh)
{
  SH2movws0(sh,4,3);
}
#define init_0_4_4_5
void FASTCALL call_0_4_4_5(SH2_struct* sh)
{
  SH2movws0(sh,4,4);
}
#define init_0_4_5_5
void FASTCALL call_0_4_5_5(SH2_struct* sh)
{
  SH2movws0(sh,4,5);
}
#define init_0_4_6_5
void FASTCALL call_0_4_6_5(SH2_struct* sh)
{
  SH2movws0(sh,4,6);
}
#define init_0_4_7_5
void FASTCALL call_0_4_7_5(SH2_struct* sh)
{
  SH2movws0(sh,4,7);
}
#define init_0_4_8_5
void FASTCALL call_0_4_8_5(SH2_struct* sh)
{
  SH2movws0(sh,4,8);
}
#define init_0_4_9_5
void FASTCALL call_0_4_9_5(SH2_struct* sh)
{
  SH2movws0(sh,4,9);
}
#define init_0_4_10_5
void FASTCALL call_0_4_10_5(SH2_struct* sh)
{
  SH2movws0(sh,4,10);
}
#define init_0_4_11_5
void FASTCALL call_0_4_11_5(SH2_struct* sh)
{
  SH2movws0(sh,4,11);
}
#define init_0_4_12_5
void FASTCALL call_0_4_12_5(SH2_struct* sh)
{
  SH2movws0(sh,4,12);
}
#define init_0_4_13_5
void FASTCALL call_0_4_13_5(SH2_struct* sh)
{
  SH2movws0(sh,4,13);
}
#define init_0_4_14_5
void FASTCALL call_0_4_14_5(SH2_struct* sh)
{
  SH2movws0(sh,4,14);
}
#define init_0_4_15_5
void FASTCALL call_0_4_15_5(SH2_struct* sh)
{
  SH2movws0(sh,4,15);
}
#define init_0_5_0_5
void FASTCALL call_0_5_0_5(SH2_struct* sh)
{
  SH2movws0(sh,5,0);
}
#define init_0_5_1_5
void FASTCALL call_0_5_1_5(SH2_struct* sh)
{
  SH2movws0(sh,5,1);
}
#define init_0_5_2_5
void FASTCALL call_0_5_2_5(SH2_struct* sh)
{
  SH2movws0(sh,5,2);
}
#define init_0_5_3_5
void FASTCALL call_0_5_3_5(SH2_struct* sh)
{
  SH2movws0(sh,5,3);
}
#define init_0_5_4_5
void FASTCALL call_0_5_4_5(SH2_struct* sh)
{
  SH2movws0(sh,5,4);
}
#define init_0_5_5_5
void FASTCALL call_0_5_5_5(SH2_struct* sh)
{
  SH2movws0(sh,5,5);
}
#define init_0_5_6_5
void FASTCALL call_0_5_6_5(SH2_struct* sh)
{
  SH2movws0(sh,5,6);
}
#define init_0_5_7_5
void FASTCALL call_0_5_7_5(SH2_struct* sh)
{
  SH2movws0(sh,5,7);
}
#define init_0_5_8_5
void FASTCALL call_0_5_8_5(SH2_struct* sh)
{
  SH2movws0(sh,5,8);
}
#define init_0_5_9_5
void FASTCALL call_0_5_9_5(SH2_struct* sh)
{
  SH2movws0(sh,5,9);
}
#define init_0_5_10_5
void FASTCALL call_0_5_10_5(SH2_struct* sh)
{
  SH2movws0(sh,5,10);
}
#define init_0_5_11_5
void FASTCALL call_0_5_11_5(SH2_struct* sh)
{
  SH2movws0(sh,5,11);
}
#define init_0_5_12_5
void FASTCALL call_0_5_12_5(SH2_struct* sh)
{
  SH2movws0(sh,5,12);
}
#define init_0_5_13_5
void FASTCALL call_0_5_13_5(SH2_struct* sh)
{
  SH2movws0(sh,5,13);
}
#define init_0_5_14_5
void FASTCALL call_0_5_14_5(SH2_struct* sh)
{
  SH2movws0(sh,5,14);
}
#define init_0_5_15_5
void FASTCALL call_0_5_15_5(SH2_struct* sh)
{
  SH2movws0(sh,5,15);
}
#define init_0_6_0_5
void FASTCALL call_0_6_0_5(SH2_struct* sh)
{
  SH2movws0(sh,6,0);
}
#define init_0_6_1_5
void FASTCALL call_0_6_1_5(SH2_struct* sh)
{
  SH2movws0(sh,6,1);
}
#define init_0_6_2_5
void FASTCALL call_0_6_2_5(SH2_struct* sh)
{
  SH2movws0(sh,6,2);
}
#define init_0_6_3_5
void FASTCALL call_0_6_3_5(SH2_struct* sh)
{
  SH2movws0(sh,6,3);
}
#define init_0_6_4_5
void FASTCALL call_0_6_4_5(SH2_struct* sh)
{
  SH2movws0(sh,6,4);
}
#define init_0_6_5_5
void FASTCALL call_0_6_5_5(SH2_struct* sh)
{
  SH2movws0(sh,6,5);
}
#define init_0_6_6_5
void FASTCALL call_0_6_6_5(SH2_struct* sh)
{
  SH2movws0(sh,6,6);
}
#define init_0_6_7_5
void FASTCALL call_0_6_7_5(SH2_struct* sh)
{
  SH2movws0(sh,6,7);
}
#define init_0_6_8_5
void FASTCALL call_0_6_8_5(SH2_struct* sh)
{
  SH2movws0(sh,6,8);
}
#define init_0_6_9_5
void FASTCALL call_0_6_9_5(SH2_struct* sh)
{
  SH2movws0(sh,6,9);
}
#define init_0_6_10_5
void FASTCALL call_0_6_10_5(SH2_struct* sh)
{
  SH2movws0(sh,6,10);
}
#define init_0_6_11_5
void FASTCALL call_0_6_11_5(SH2_struct* sh)
{
  SH2movws0(sh,6,11);
}
#define init_0_6_12_5
void FASTCALL call_0_6_12_5(SH2_struct* sh)
{
  SH2movws0(sh,6,12);
}
#define init_0_6_13_5
void FASTCALL call_0_6_13_5(SH2_struct* sh)
{
  SH2movws0(sh,6,13);
}
#define init_0_6_14_5
void FASTCALL call_0_6_14_5(SH2_struct* sh)
{
  SH2movws0(sh,6,14);
}
#define init_0_6_15_5
void FASTCALL call_0_6_15_5(SH2_struct* sh)
{
  SH2movws0(sh,6,15);
}
#define init_0_7_0_5
void FASTCALL call_0_7_0_5(SH2_struct* sh)
{
  SH2movws0(sh,7,0);
}
#define init_0_7_1_5
void FASTCALL call_0_7_1_5(SH2_struct* sh)
{
  SH2movws0(sh,7,1);
}
#define init_0_7_2_5
void FASTCALL call_0_7_2_5(SH2_struct* sh)
{
  SH2movws0(sh,7,2);
}
#define init_0_7_3_5
void FASTCALL call_0_7_3_5(SH2_struct* sh)
{
  SH2movws0(sh,7,3);
}
#define init_0_7_4_5
void FASTCALL call_0_7_4_5(SH2_struct* sh)
{
  SH2movws0(sh,7,4);
}
#define init_0_7_5_5
void FASTCALL call_0_7_5_5(SH2_struct* sh)
{
  SH2movws0(sh,7,5);
}
#define init_0_7_6_5
void FASTCALL call_0_7_6_5(SH2_struct* sh)
{
  SH2movws0(sh,7,6);
}
#define init_0_7_7_5
void FASTCALL call_0_7_7_5(SH2_struct* sh)
{
  SH2movws0(sh,7,7);
}
#define init_0_7_8_5
void FASTCALL call_0_7_8_5(SH2_struct* sh)
{
  SH2movws0(sh,7,8);
}
#define init_0_7_9_5
void FASTCALL call_0_7_9_5(SH2_struct* sh)
{
  SH2movws0(sh,7,9);
}
#define init_0_7_10_5
void FASTCALL call_0_7_10_5(SH2_struct* sh)
{
  SH2movws0(sh,7,10);
}
#define init_0_7_11_5
void FASTCALL call_0_7_11_5(SH2_struct* sh)
{
  SH2movws0(sh,7,11);
}
#define init_0_7_12_5
void FASTCALL call_0_7_12_5(SH2_struct* sh)
{
  SH2movws0(sh,7,12);
}
#define init_0_7_13_5
void FASTCALL call_0_7_13_5(SH2_struct* sh)
{
  SH2movws0(sh,7,13);
}
#define init_0_7_14_5
void FASTCALL call_0_7_14_5(SH2_struct* sh)
{
  SH2movws0(sh,7,14);
}
#define init_0_7_15_5
void FASTCALL call_0_7_15_5(SH2_struct* sh)
{
  SH2movws0(sh,7,15);
}
#define init_0_8_0_5
void FASTCALL call_0_8_0_5(SH2_struct* sh)
{
  SH2movws0(sh,8,0);
}
#define init_0_8_1_5
void FASTCALL call_0_8_1_5(SH2_struct* sh)
{
  SH2movws0(sh,8,1);
}
#define init_0_8_2_5
void FASTCALL call_0_8_2_5(SH2_struct* sh)
{
  SH2movws0(sh,8,2);
}
#define init_0_8_3_5
void FASTCALL call_0_8_3_5(SH2_struct* sh)
{
  SH2movws0(sh,8,3);
}
#define init_0_8_4_5
void FASTCALL call_0_8_4_5(SH2_struct* sh)
{
  SH2movws0(sh,8,4);
}
#define init_0_8_5_5
void FASTCALL call_0_8_5_5(SH2_struct* sh)
{
  SH2movws0(sh,8,5);
}
#define init_0_8_6_5
void FASTCALL call_0_8_6_5(SH2_struct* sh)
{
  SH2movws0(sh,8,6);
}
#define init_0_8_7_5
void FASTCALL call_0_8_7_5(SH2_struct* sh)
{
  SH2movws0(sh,8,7);
}
#define init_0_8_8_5
void FASTCALL call_0_8_8_5(SH2_struct* sh)
{
  SH2movws0(sh,8,8);
}
#define init_0_8_9_5
void FASTCALL call_0_8_9_5(SH2_struct* sh)
{
  SH2movws0(sh,8,9);
}
#define init_0_8_10_5
void FASTCALL call_0_8_10_5(SH2_struct* sh)
{
  SH2movws0(sh,8,10);
}
#define init_0_8_11_5
void FASTCALL call_0_8_11_5(SH2_struct* sh)
{
  SH2movws0(sh,8,11);
}
#define init_0_8_12_5
void FASTCALL call_0_8_12_5(SH2_struct* sh)
{
  SH2movws0(sh,8,12);
}
#define init_0_8_13_5
void FASTCALL call_0_8_13_5(SH2_struct* sh)
{
  SH2movws0(sh,8,13);
}
#define init_0_8_14_5
void FASTCALL call_0_8_14_5(SH2_struct* sh)
{
  SH2movws0(sh,8,14);
}
#define init_0_8_15_5
void FASTCALL call_0_8_15_5(SH2_struct* sh)
{
  SH2movws0(sh,8,15);
}
#define init_0_9_0_5
void FASTCALL call_0_9_0_5(SH2_struct* sh)
{
  SH2movws0(sh,9,0);
}
#define init_0_9_1_5
void FASTCALL call_0_9_1_5(SH2_struct* sh)
{
  SH2movws0(sh,9,1);
}
#define init_0_9_2_5
void FASTCALL call_0_9_2_5(SH2_struct* sh)
{
  SH2movws0(sh,9,2);
}
#define init_0_9_3_5
void FASTCALL call_0_9_3_5(SH2_struct* sh)
{
  SH2movws0(sh,9,3);
}
#define init_0_9_4_5
void FASTCALL call_0_9_4_5(SH2_struct* sh)
{
  SH2movws0(sh,9,4);
}
#define init_0_9_5_5
void FASTCALL call_0_9_5_5(SH2_struct* sh)
{
  SH2movws0(sh,9,5);
}
#define init_0_9_6_5
void FASTCALL call_0_9_6_5(SH2_struct* sh)
{
  SH2movws0(sh,9,6);
}
#define init_0_9_7_5
void FASTCALL call_0_9_7_5(SH2_struct* sh)
{
  SH2movws0(sh,9,7);
}
#define init_0_9_8_5
void FASTCALL call_0_9_8_5(SH2_struct* sh)
{
  SH2movws0(sh,9,8);
}
#define init_0_9_9_5
void FASTCALL call_0_9_9_5(SH2_struct* sh)
{
  SH2movws0(sh,9,9);
}
#define init_0_9_10_5
void FASTCALL call_0_9_10_5(SH2_struct* sh)
{
  SH2movws0(sh,9,10);
}
#define init_0_9_11_5
void FASTCALL call_0_9_11_5(SH2_struct* sh)
{
  SH2movws0(sh,9,11);
}
#define init_0_9_12_5
void FASTCALL call_0_9_12_5(SH2_struct* sh)
{
  SH2movws0(sh,9,12);
}
#define init_0_9_13_5
void FASTCALL call_0_9_13_5(SH2_struct* sh)
{
  SH2movws0(sh,9,13);
}
#define init_0_9_14_5
void FASTCALL call_0_9_14_5(SH2_struct* sh)
{
  SH2movws0(sh,9,14);
}
#define init_0_9_15_5
void FASTCALL call_0_9_15_5(SH2_struct* sh)
{
  SH2movws0(sh,9,15);
}
#define init_0_10_0_5
void FASTCALL call_0_10_0_5(SH2_struct* sh)
{
  SH2movws0(sh,10,0);
}
#define init_0_10_1_5
void FASTCALL call_0_10_1_5(SH2_struct* sh)
{
  SH2movws0(sh,10,1);
}
#define init_0_10_2_5
void FASTCALL call_0_10_2_5(SH2_struct* sh)
{
  SH2movws0(sh,10,2);
}
#define init_0_10_3_5
void FASTCALL call_0_10_3_5(SH2_struct* sh)
{
  SH2movws0(sh,10,3);
}
#define init_0_10_4_5
void FASTCALL call_0_10_4_5(SH2_struct* sh)
{
  SH2movws0(sh,10,4);
}
#define init_0_10_5_5
void FASTCALL call_0_10_5_5(SH2_struct* sh)
{
  SH2movws0(sh,10,5);
}
#define init_0_10_6_5
void FASTCALL call_0_10_6_5(SH2_struct* sh)
{
  SH2movws0(sh,10,6);
}
#define init_0_10_7_5
void FASTCALL call_0_10_7_5(SH2_struct* sh)
{
  SH2movws0(sh,10,7);
}
#define init_0_10_8_5
void FASTCALL call_0_10_8_5(SH2_struct* sh)
{
  SH2movws0(sh,10,8);
}
#define init_0_10_9_5
void FASTCALL call_0_10_9_5(SH2_struct* sh)
{
  SH2movws0(sh,10,9);
}
#define init_0_10_10_5
void FASTCALL call_0_10_10_5(SH2_struct* sh)
{
  SH2movws0(sh,10,10);
}
#define init_0_10_11_5
void FASTCALL call_0_10_11_5(SH2_struct* sh)
{
  SH2movws0(sh,10,11);
}
#define init_0_10_12_5
void FASTCALL call_0_10_12_5(SH2_struct* sh)
{
  SH2movws0(sh,10,12);
}
#define init_0_10_13_5
void FASTCALL call_0_10_13_5(SH2_struct* sh)
{
  SH2movws0(sh,10,13);
}
#define init_0_10_14_5
void FASTCALL call_0_10_14_5(SH2_struct* sh)
{
  SH2movws0(sh,10,14);
}
#define init_0_10_15_5
void FASTCALL call_0_10_15_5(SH2_struct* sh)
{
  SH2movws0(sh,10,15);
}
#define init_0_11_0_5
void FASTCALL call_0_11_0_5(SH2_struct* sh)
{
  SH2movws0(sh,11,0);
}
#define init_0_11_1_5
void FASTCALL call_0_11_1_5(SH2_struct* sh)
{
  SH2movws0(sh,11,1);
}
#define init_0_11_2_5
void FASTCALL call_0_11_2_5(SH2_struct* sh)
{
  SH2movws0(sh,11,2);
}
#define init_0_11_3_5
void FASTCALL call_0_11_3_5(SH2_struct* sh)
{
  SH2movws0(sh,11,3);
}
#define init_0_11_4_5
void FASTCALL call_0_11_4_5(SH2_struct* sh)
{
  SH2movws0(sh,11,4);
}
#define init_0_11_5_5
void FASTCALL call_0_11_5_5(SH2_struct* sh)
{
  SH2movws0(sh,11,5);
}
#define init_0_11_6_5
void FASTCALL call_0_11_6_5(SH2_struct* sh)
{
  SH2movws0(sh,11,6);
}
#define init_0_11_7_5
void FASTCALL call_0_11_7_5(SH2_struct* sh)
{
  SH2movws0(sh,11,7);
}
#define init_0_11_8_5
void FASTCALL call_0_11_8_5(SH2_struct* sh)
{
  SH2movws0(sh,11,8);
}
#define init_0_11_9_5
void FASTCALL call_0_11_9_5(SH2_struct* sh)
{
  SH2movws0(sh,11,9);
}
#define init_0_11_10_5
void FASTCALL call_0_11_10_5(SH2_struct* sh)
{
  SH2movws0(sh,11,10);
}
#define init_0_11_11_5
void FASTCALL call_0_11_11_5(SH2_struct* sh)
{
  SH2movws0(sh,11,11);
}
#define init_0_11_12_5
void FASTCALL call_0_11_12_5(SH2_struct* sh)
{
  SH2movws0(sh,11,12);
}
#define init_0_11_13_5
void FASTCALL call_0_11_13_5(SH2_struct* sh)
{
  SH2movws0(sh,11,13);
}
#define init_0_11_14_5
void FASTCALL call_0_11_14_5(SH2_struct* sh)
{
  SH2movws0(sh,11,14);
}
#define init_0_11_15_5
void FASTCALL call_0_11_15_5(SH2_struct* sh)
{
  SH2movws0(sh,11,15);
}
#define init_0_12_0_5
void FASTCALL call_0_12_0_5(SH2_struct* sh)
{
  SH2movws0(sh,12,0);
}
#define init_0_12_1_5
void FASTCALL call_0_12_1_5(SH2_struct* sh)
{
  SH2movws0(sh,12,1);
}
#define init_0_12_2_5
void FASTCALL call_0_12_2_5(SH2_struct* sh)
{
  SH2movws0(sh,12,2);
}
#define init_0_12_3_5
void FASTCALL call_0_12_3_5(SH2_struct* sh)
{
  SH2movws0(sh,12,3);
}
#define init_0_12_4_5
void FASTCALL call_0_12_4_5(SH2_struct* sh)
{
  SH2movws0(sh,12,4);
}
#define init_0_12_5_5
void FASTCALL call_0_12_5_5(SH2_struct* sh)
{
  SH2movws0(sh,12,5);
}
#define init_0_12_6_5
void FASTCALL call_0_12_6_5(SH2_struct* sh)
{
  SH2movws0(sh,12,6);
}
#define init_0_12_7_5
void FASTCALL call_0_12_7_5(SH2_struct* sh)
{
  SH2movws0(sh,12,7);
}
#define init_0_12_8_5
void FASTCALL call_0_12_8_5(SH2_struct* sh)
{
  SH2movws0(sh,12,8);
}
#define init_0_12_9_5
void FASTCALL call_0_12_9_5(SH2_struct* sh)
{
  SH2movws0(sh,12,9);
}
#define init_0_12_10_5
void FASTCALL call_0_12_10_5(SH2_struct* sh)
{
  SH2movws0(sh,12,10);
}
#define init_0_12_11_5
void FASTCALL call_0_12_11_5(SH2_struct* sh)
{
  SH2movws0(sh,12,11);
}
#define init_0_12_12_5
void FASTCALL call_0_12_12_5(SH2_struct* sh)
{
  SH2movws0(sh,12,12);
}
#define init_0_12_13_5
void FASTCALL call_0_12_13_5(SH2_struct* sh)
{
  SH2movws0(sh,12,13);
}
#define init_0_12_14_5
void FASTCALL call_0_12_14_5(SH2_struct* sh)
{
  SH2movws0(sh,12,14);
}
#define init_0_12_15_5
void FASTCALL call_0_12_15_5(SH2_struct* sh)
{
  SH2movws0(sh,12,15);
}
#define init_0_13_0_5
void FASTCALL call_0_13_0_5(SH2_struct* sh)
{
  SH2movws0(sh,13,0);
}
#define init_0_13_1_5
void FASTCALL call_0_13_1_5(SH2_struct* sh)
{
  SH2movws0(sh,13,1);
}
#define init_0_13_2_5
void FASTCALL call_0_13_2_5(SH2_struct* sh)
{
  SH2movws0(sh,13,2);
}
#define init_0_13_3_5
void FASTCALL call_0_13_3_5(SH2_struct* sh)
{
  SH2movws0(sh,13,3);
}
#define init_0_13_4_5
void FASTCALL call_0_13_4_5(SH2_struct* sh)
{
  SH2movws0(sh,13,4);
}
#define init_0_13_5_5
void FASTCALL call_0_13_5_5(SH2_struct* sh)
{
  SH2movws0(sh,13,5);
}
#define init_0_13_6_5
void FASTCALL call_0_13_6_5(SH2_struct* sh)
{
  SH2movws0(sh,13,6);
}
#define init_0_13_7_5
void FASTCALL call_0_13_7_5(SH2_struct* sh)
{
  SH2movws0(sh,13,7);
}
#define init_0_13_8_5
void FASTCALL call_0_13_8_5(SH2_struct* sh)
{
  SH2movws0(sh,13,8);
}
#define init_0_13_9_5
void FASTCALL call_0_13_9_5(SH2_struct* sh)
{
  SH2movws0(sh,13,9);
}
#define init_0_13_10_5
void FASTCALL call_0_13_10_5(SH2_struct* sh)
{
  SH2movws0(sh,13,10);
}
#define init_0_13_11_5
void FASTCALL call_0_13_11_5(SH2_struct* sh)
{
  SH2movws0(sh,13,11);
}
#define init_0_13_12_5
void FASTCALL call_0_13_12_5(SH2_struct* sh)
{
  SH2movws0(sh,13,12);
}
#define init_0_13_13_5
void FASTCALL call_0_13_13_5(SH2_struct* sh)
{
  SH2movws0(sh,13,13);
}
#define init_0_13_14_5
void FASTCALL call_0_13_14_5(SH2_struct* sh)
{
  SH2movws0(sh,13,14);
}
#define init_0_13_15_5
void FASTCALL call_0_13_15_5(SH2_struct* sh)
{
  SH2movws0(sh,13,15);
}
#define init_0_14_0_5
void FASTCALL call_0_14_0_5(SH2_struct* sh)
{
  SH2movws0(sh,14,0);
}
#define init_0_14_1_5
void FASTCALL call_0_14_1_5(SH2_struct* sh)
{
  SH2movws0(sh,14,1);
}
#define init_0_14_2_5
void FASTCALL call_0_14_2_5(SH2_struct* sh)
{
  SH2movws0(sh,14,2);
}
#define init_0_14_3_5
void FASTCALL call_0_14_3_5(SH2_struct* sh)
{
  SH2movws0(sh,14,3);
}
#define init_0_14_4_5
void FASTCALL call_0_14_4_5(SH2_struct* sh)
{
  SH2movws0(sh,14,4);
}
#define init_0_14_5_5
void FASTCALL call_0_14_5_5(SH2_struct* sh)
{
  SH2movws0(sh,14,5);
}
#define init_0_14_6_5
void FASTCALL call_0_14_6_5(SH2_struct* sh)
{
  SH2movws0(sh,14,6);
}
#define init_0_14_7_5
void FASTCALL call_0_14_7_5(SH2_struct* sh)
{
  SH2movws0(sh,14,7);
}
#define init_0_14_8_5
void FASTCALL call_0_14_8_5(SH2_struct* sh)
{
  SH2movws0(sh,14,8);
}
#define init_0_14_9_5
void FASTCALL call_0_14_9_5(SH2_struct* sh)
{
  SH2movws0(sh,14,9);
}
#define init_0_14_10_5
void FASTCALL call_0_14_10_5(SH2_struct* sh)
{
  SH2movws0(sh,14,10);
}
#define init_0_14_11_5
void FASTCALL call_0_14_11_5(SH2_struct* sh)
{
  SH2movws0(sh,14,11);
}
#define init_0_14_12_5
void FASTCALL call_0_14_12_5(SH2_struct* sh)
{
  SH2movws0(sh,14,12);
}
#define init_0_14_13_5
void FASTCALL call_0_14_13_5(SH2_struct* sh)
{
  SH2movws0(sh,14,13);
}
#define init_0_14_14_5
void FASTCALL call_0_14_14_5(SH2_struct* sh)
{
  SH2movws0(sh,14,14);
}
#define init_0_14_15_5
void FASTCALL call_0_14_15_5(SH2_struct* sh)
{
  SH2movws0(sh,14,15);
}
#define init_0_15_0_5
void FASTCALL call_0_15_0_5(SH2_struct* sh)
{
  SH2movws0(sh,15,0);
}
#define init_0_15_1_5
void FASTCALL call_0_15_1_5(SH2_struct* sh)
{
  SH2movws0(sh,15,1);
}
#define init_0_15_2_5
void FASTCALL call_0_15_2_5(SH2_struct* sh)
{
  SH2movws0(sh,15,2);
}
#define init_0_15_3_5
void FASTCALL call_0_15_3_5(SH2_struct* sh)
{
  SH2movws0(sh,15,3);
}
#define init_0_15_4_5
void FASTCALL call_0_15_4_5(SH2_struct* sh)
{
  SH2movws0(sh,15,4);
}
#define init_0_15_5_5
void FASTCALL call_0_15_5_5(SH2_struct* sh)
{
  SH2movws0(sh,15,5);
}
#define init_0_15_6_5
void FASTCALL call_0_15_6_5(SH2_struct* sh)
{
  SH2movws0(sh,15,6);
}
#define init_0_15_7_5
void FASTCALL call_0_15_7_5(SH2_struct* sh)
{
  SH2movws0(sh,15,7);
}
#define init_0_15_8_5
void FASTCALL call_0_15_8_5(SH2_struct* sh)
{
  SH2movws0(sh,15,8);
}
#define init_0_15_9_5
void FASTCALL call_0_15_9_5(SH2_struct* sh)
{
  SH2movws0(sh,15,9);
}
#define init_0_15_10_5
void FASTCALL call_0_15_10_5(SH2_struct* sh)
{
  SH2movws0(sh,15,10);
}
#define init_0_15_11_5
void FASTCALL call_0_15_11_5(SH2_struct* sh)
{
  SH2movws0(sh,15,11);
}
#define init_0_15_12_5
void FASTCALL call_0_15_12_5(SH2_struct* sh)
{
  SH2movws0(sh,15,12);
}
#define init_0_15_13_5
void FASTCALL call_0_15_13_5(SH2_struct* sh)
{
  SH2movws0(sh,15,13);
}
#define init_0_15_14_5
void FASTCALL call_0_15_14_5(SH2_struct* sh)
{
  SH2movws0(sh,15,14);
}
#define init_0_15_15_5
void FASTCALL call_0_15_15_5(SH2_struct* sh)
{
  SH2movws0(sh,15,15);
}
#define init_0_0_0_14
void FASTCALL call_0_0_0_14(SH2_struct* sh)
{
  SH2movll0(sh,0,0);
}
#define init_0_0_1_14
void FASTCALL call_0_0_1_14(SH2_struct* sh)
{
  SH2movll0(sh,0,1);
}
#define init_0_0_2_14
void FASTCALL call_0_0_2_14(SH2_struct* sh)
{
  SH2movll0(sh,0,2);
}
#define init_0_0_3_14
void FASTCALL call_0_0_3_14(SH2_struct* sh)
{
  SH2movll0(sh,0,3);
}
#define init_0_0_4_14
void FASTCALL call_0_0_4_14(SH2_struct* sh)
{
  SH2movll0(sh,0,4);
}
#define init_0_0_5_14
void FASTCALL call_0_0_5_14(SH2_struct* sh)
{
  SH2movll0(sh,0,5);
}
#define init_0_0_6_14
void FASTCALL call_0_0_6_14(SH2_struct* sh)
{
  SH2movll0(sh,0,6);
}
#define init_0_0_7_14
void FASTCALL call_0_0_7_14(SH2_struct* sh)
{
  SH2movll0(sh,0,7);
}
#define init_0_0_8_14
void FASTCALL call_0_0_8_14(SH2_struct* sh)
{
  SH2movll0(sh,0,8);
}
#define init_0_0_9_14
void FASTCALL call_0_0_9_14(SH2_struct* sh)
{
  SH2movll0(sh,0,9);
}
#define init_0_0_10_14
void FASTCALL call_0_0_10_14(SH2_struct* sh)
{
  SH2movll0(sh,0,10);
}
#define init_0_0_11_14
void FASTCALL call_0_0_11_14(SH2_struct* sh)
{
  SH2movll0(sh,0,11);
}
#define init_0_0_12_14
void FASTCALL call_0_0_12_14(SH2_struct* sh)
{
  SH2movll0(sh,0,12);
}
#define init_0_0_13_14
void FASTCALL call_0_0_13_14(SH2_struct* sh)
{
  SH2movll0(sh,0,13);
}
#define init_0_0_14_14
void FASTCALL call_0_0_14_14(SH2_struct* sh)
{
  SH2movll0(sh,0,14);
}
#define init_0_0_15_14
void FASTCALL call_0_0_15_14(SH2_struct* sh)
{
  SH2movll0(sh,0,15);
}
#define init_0_1_0_14
void FASTCALL call_0_1_0_14(SH2_struct* sh)
{
  SH2movll0(sh,1,0);
}
#define init_0_1_1_14
void FASTCALL call_0_1_1_14(SH2_struct* sh)
{
  SH2movll0(sh,1,1);
}
#define init_0_1_2_14
void FASTCALL call_0_1_2_14(SH2_struct* sh)
{
  SH2movll0(sh,1,2);
}
#define init_0_1_3_14
void FASTCALL call_0_1_3_14(SH2_struct* sh)
{
  SH2movll0(sh,1,3);
}
#define init_0_1_4_14
void FASTCALL call_0_1_4_14(SH2_struct* sh)
{
  SH2movll0(sh,1,4);
}
#define init_0_1_5_14
void FASTCALL call_0_1_5_14(SH2_struct* sh)
{
  SH2movll0(sh,1,5);
}
#define init_0_1_6_14
void FASTCALL call_0_1_6_14(SH2_struct* sh)
{
  SH2movll0(sh,1,6);
}
#define init_0_1_7_14
void FASTCALL call_0_1_7_14(SH2_struct* sh)
{
  SH2movll0(sh,1,7);
}
#define init_0_1_8_14
void FASTCALL call_0_1_8_14(SH2_struct* sh)
{
  SH2movll0(sh,1,8);
}
#define init_0_1_9_14
void FASTCALL call_0_1_9_14(SH2_struct* sh)
{
  SH2movll0(sh,1,9);
}
#define init_0_1_10_14
void FASTCALL call_0_1_10_14(SH2_struct* sh)
{
  SH2movll0(sh,1,10);
}
#define init_0_1_11_14
void FASTCALL call_0_1_11_14(SH2_struct* sh)
{
  SH2movll0(sh,1,11);
}
#define init_0_1_12_14
void FASTCALL call_0_1_12_14(SH2_struct* sh)
{
  SH2movll0(sh,1,12);
}
#define init_0_1_13_14
void FASTCALL call_0_1_13_14(SH2_struct* sh)
{
  SH2movll0(sh,1,13);
}
#define init_0_1_14_14
void FASTCALL call_0_1_14_14(SH2_struct* sh)
{
  SH2movll0(sh,1,14);
}
#define init_0_1_15_14
void FASTCALL call_0_1_15_14(SH2_struct* sh)
{
  SH2movll0(sh,1,15);
}
#define init_0_2_0_14
void FASTCALL call_0_2_0_14(SH2_struct* sh)
{
  SH2movll0(sh,2,0);
}
#define init_0_2_1_14
void FASTCALL call_0_2_1_14(SH2_struct* sh)
{
  SH2movll0(sh,2,1);
}
#define init_0_2_2_14
void FASTCALL call_0_2_2_14(SH2_struct* sh)
{
  SH2movll0(sh,2,2);
}
#define init_0_2_3_14
void FASTCALL call_0_2_3_14(SH2_struct* sh)
{
  SH2movll0(sh,2,3);
}
#define init_0_2_4_14
void FASTCALL call_0_2_4_14(SH2_struct* sh)
{
  SH2movll0(sh,2,4);
}
#define init_0_2_5_14
void FASTCALL call_0_2_5_14(SH2_struct* sh)
{
  SH2movll0(sh,2,5);
}
#define init_0_2_6_14
void FASTCALL call_0_2_6_14(SH2_struct* sh)
{
  SH2movll0(sh,2,6);
}
#define init_0_2_7_14
void FASTCALL call_0_2_7_14(SH2_struct* sh)
{
  SH2movll0(sh,2,7);
}
#define init_0_2_8_14
void FASTCALL call_0_2_8_14(SH2_struct* sh)
{
  SH2movll0(sh,2,8);
}
#define init_0_2_9_14
void FASTCALL call_0_2_9_14(SH2_struct* sh)
{
  SH2movll0(sh,2,9);
}
#define init_0_2_10_14
void FASTCALL call_0_2_10_14(SH2_struct* sh)
{
  SH2movll0(sh,2,10);
}
#define init_0_2_11_14
void FASTCALL call_0_2_11_14(SH2_struct* sh)
{
  SH2movll0(sh,2,11);
}
#define init_0_2_12_14
void FASTCALL call_0_2_12_14(SH2_struct* sh)
{
  SH2movll0(sh,2,12);
}
#define init_0_2_13_14
void FASTCALL call_0_2_13_14(SH2_struct* sh)
{
  SH2movll0(sh,2,13);
}
#define init_0_2_14_14
void FASTCALL call_0_2_14_14(SH2_struct* sh)
{
  SH2movll0(sh,2,14);
}
#define init_0_2_15_14
void FASTCALL call_0_2_15_14(SH2_struct* sh)
{
  SH2movll0(sh,2,15);
}
#define init_0_3_0_14
void FASTCALL call_0_3_0_14(SH2_struct* sh)
{
  SH2movll0(sh,3,0);
}
#define init_0_3_1_14
void FASTCALL call_0_3_1_14(SH2_struct* sh)
{
  SH2movll0(sh,3,1);
}
#define init_0_3_2_14
void FASTCALL call_0_3_2_14(SH2_struct* sh)
{
  SH2movll0(sh,3,2);
}
#define init_0_3_3_14
void FASTCALL call_0_3_3_14(SH2_struct* sh)
{
  SH2movll0(sh,3,3);
}
#define init_0_3_4_14
void FASTCALL call_0_3_4_14(SH2_struct* sh)
{
  SH2movll0(sh,3,4);
}
#define init_0_3_5_14
void FASTCALL call_0_3_5_14(SH2_struct* sh)
{
  SH2movll0(sh,3,5);
}
#define init_0_3_6_14
void FASTCALL call_0_3_6_14(SH2_struct* sh)
{
  SH2movll0(sh,3,6);
}
#define init_0_3_7_14
void FASTCALL call_0_3_7_14(SH2_struct* sh)
{
  SH2movll0(sh,3,7);
}
#define init_0_3_8_14
void FASTCALL call_0_3_8_14(SH2_struct* sh)
{
  SH2movll0(sh,3,8);
}
#define init_0_3_9_14
void FASTCALL call_0_3_9_14(SH2_struct* sh)
{
  SH2movll0(sh,3,9);
}
#define init_0_3_10_14
void FASTCALL call_0_3_10_14(SH2_struct* sh)
{
  SH2movll0(sh,3,10);
}
#define init_0_3_11_14
void FASTCALL call_0_3_11_14(SH2_struct* sh)
{
  SH2movll0(sh,3,11);
}
#define init_0_3_12_14
void FASTCALL call_0_3_12_14(SH2_struct* sh)
{
  SH2movll0(sh,3,12);
}
#define init_0_3_13_14
void FASTCALL call_0_3_13_14(SH2_struct* sh)
{
  SH2movll0(sh,3,13);
}
#define init_0_3_14_14
void FASTCALL call_0_3_14_14(SH2_struct* sh)
{
  SH2movll0(sh,3,14);
}
#define init_0_3_15_14
void FASTCALL call_0_3_15_14(SH2_struct* sh)
{
  SH2movll0(sh,3,15);
}
#define init_0_4_0_14
void FASTCALL call_0_4_0_14(SH2_struct* sh)
{
  SH2movll0(sh,4,0);
}
#define init_0_4_1_14
void FASTCALL call_0_4_1_14(SH2_struct* sh)
{
  SH2movll0(sh,4,1);
}
#define init_0_4_2_14
void FASTCALL call_0_4_2_14(SH2_struct* sh)
{
  SH2movll0(sh,4,2);
}
#define init_0_4_3_14
void FASTCALL call_0_4_3_14(SH2_struct* sh)
{
  SH2movll0(sh,4,3);
}
#define init_0_4_4_14
void FASTCALL call_0_4_4_14(SH2_struct* sh)
{
  SH2movll0(sh,4,4);
}
#define init_0_4_5_14
void FASTCALL call_0_4_5_14(SH2_struct* sh)
{
  SH2movll0(sh,4,5);
}
#define init_0_4_6_14
void FASTCALL call_0_4_6_14(SH2_struct* sh)
{
  SH2movll0(sh,4,6);
}
#define init_0_4_7_14
void FASTCALL call_0_4_7_14(SH2_struct* sh)
{
  SH2movll0(sh,4,7);
}
#define init_0_4_8_14
void FASTCALL call_0_4_8_14(SH2_struct* sh)
{
  SH2movll0(sh,4,8);
}
#define init_0_4_9_14
void FASTCALL call_0_4_9_14(SH2_struct* sh)
{
  SH2movll0(sh,4,9);
}
#define init_0_4_10_14
void FASTCALL call_0_4_10_14(SH2_struct* sh)
{
  SH2movll0(sh,4,10);
}
#define init_0_4_11_14
void FASTCALL call_0_4_11_14(SH2_struct* sh)
{
  SH2movll0(sh,4,11);
}
#define init_0_4_12_14
void FASTCALL call_0_4_12_14(SH2_struct* sh)
{
  SH2movll0(sh,4,12);
}
#define init_0_4_13_14
void FASTCALL call_0_4_13_14(SH2_struct* sh)
{
  SH2movll0(sh,4,13);
}
#define init_0_4_14_14
void FASTCALL call_0_4_14_14(SH2_struct* sh)
{
  SH2movll0(sh,4,14);
}
#define init_0_4_15_14
void FASTCALL call_0_4_15_14(SH2_struct* sh)
{
  SH2movll0(sh,4,15);
}
#define init_0_5_0_14
void FASTCALL call_0_5_0_14(SH2_struct* sh)
{
  SH2movll0(sh,5,0);
}
#define init_0_5_1_14
void FASTCALL call_0_5_1_14(SH2_struct* sh)
{
  SH2movll0(sh,5,1);
}
#define init_0_5_2_14
void FASTCALL call_0_5_2_14(SH2_struct* sh)
{
  SH2movll0(sh,5,2);
}
#define init_0_5_3_14
void FASTCALL call_0_5_3_14(SH2_struct* sh)
{
  SH2movll0(sh,5,3);
}
#define init_0_5_4_14
void FASTCALL call_0_5_4_14(SH2_struct* sh)
{
  SH2movll0(sh,5,4);
}
#define init_0_5_5_14
void FASTCALL call_0_5_5_14(SH2_struct* sh)
{
  SH2movll0(sh,5,5);
}
#define init_0_5_6_14
void FASTCALL call_0_5_6_14(SH2_struct* sh)
{
  SH2movll0(sh,5,6);
}
#define init_0_5_7_14
void FASTCALL call_0_5_7_14(SH2_struct* sh)
{
  SH2movll0(sh,5,7);
}
#define init_0_5_8_14
void FASTCALL call_0_5_8_14(SH2_struct* sh)
{
  SH2movll0(sh,5,8);
}
#define init_0_5_9_14
void FASTCALL call_0_5_9_14(SH2_struct* sh)
{
  SH2movll0(sh,5,9);
}
#define init_0_5_10_14
void FASTCALL call_0_5_10_14(SH2_struct* sh)
{
  SH2movll0(sh,5,10);
}
#define init_0_5_11_14
void FASTCALL call_0_5_11_14(SH2_struct* sh)
{
  SH2movll0(sh,5,11);
}
#define init_0_5_12_14
void FASTCALL call_0_5_12_14(SH2_struct* sh)
{
  SH2movll0(sh,5,12);
}
#define init_0_5_13_14
void FASTCALL call_0_5_13_14(SH2_struct* sh)
{
  SH2movll0(sh,5,13);
}
#define init_0_5_14_14
void FASTCALL call_0_5_14_14(SH2_struct* sh)
{
  SH2movll0(sh,5,14);
}
#define init_0_5_15_14
void FASTCALL call_0_5_15_14(SH2_struct* sh)
{
  SH2movll0(sh,5,15);
}
#define init_0_6_0_14
void FASTCALL call_0_6_0_14(SH2_struct* sh)
{
  SH2movll0(sh,6,0);
}
#define init_0_6_1_14
void FASTCALL call_0_6_1_14(SH2_struct* sh)
{
  SH2movll0(sh,6,1);
}
#define init_0_6_2_14
void FASTCALL call_0_6_2_14(SH2_struct* sh)
{
  SH2movll0(sh,6,2);
}
#define init_0_6_3_14
void FASTCALL call_0_6_3_14(SH2_struct* sh)
{
  SH2movll0(sh,6,3);
}
#define init_0_6_4_14
void FASTCALL call_0_6_4_14(SH2_struct* sh)
{
  SH2movll0(sh,6,4);
}
#define init_0_6_5_14
void FASTCALL call_0_6_5_14(SH2_struct* sh)
{
  SH2movll0(sh,6,5);
}
#define init_0_6_6_14
void FASTCALL call_0_6_6_14(SH2_struct* sh)
{
  SH2movll0(sh,6,6);
}
#define init_0_6_7_14
void FASTCALL call_0_6_7_14(SH2_struct* sh)
{
  SH2movll0(sh,6,7);
}
#define init_0_6_8_14
void FASTCALL call_0_6_8_14(SH2_struct* sh)
{
  SH2movll0(sh,6,8);
}
#define init_0_6_9_14
void FASTCALL call_0_6_9_14(SH2_struct* sh)
{
  SH2movll0(sh,6,9);
}
#define init_0_6_10_14
void FASTCALL call_0_6_10_14(SH2_struct* sh)
{
  SH2movll0(sh,6,10);
}
#define init_0_6_11_14
void FASTCALL call_0_6_11_14(SH2_struct* sh)
{
  SH2movll0(sh,6,11);
}
#define init_0_6_12_14
void FASTCALL call_0_6_12_14(SH2_struct* sh)
{
  SH2movll0(sh,6,12);
}
#define init_0_6_13_14
void FASTCALL call_0_6_13_14(SH2_struct* sh)
{
  SH2movll0(sh,6,13);
}
#define init_0_6_14_14
void FASTCALL call_0_6_14_14(SH2_struct* sh)
{
  SH2movll0(sh,6,14);
}
#define init_0_6_15_14
void FASTCALL call_0_6_15_14(SH2_struct* sh)
{
  SH2movll0(sh,6,15);
}
#define init_0_7_0_14
void FASTCALL call_0_7_0_14(SH2_struct* sh)
{
  SH2movll0(sh,7,0);
}
#define init_0_7_1_14
void FASTCALL call_0_7_1_14(SH2_struct* sh)
{
  SH2movll0(sh,7,1);
}
#define init_0_7_2_14
void FASTCALL call_0_7_2_14(SH2_struct* sh)
{
  SH2movll0(sh,7,2);
}
#define init_0_7_3_14
void FASTCALL call_0_7_3_14(SH2_struct* sh)
{
  SH2movll0(sh,7,3);
}
#define init_0_7_4_14
void FASTCALL call_0_7_4_14(SH2_struct* sh)
{
  SH2movll0(sh,7,4);
}
#define init_0_7_5_14
void FASTCALL call_0_7_5_14(SH2_struct* sh)
{
  SH2movll0(sh,7,5);
}
#define init_0_7_6_14
void FASTCALL call_0_7_6_14(SH2_struct* sh)
{
  SH2movll0(sh,7,6);
}
#define init_0_7_7_14
void FASTCALL call_0_7_7_14(SH2_struct* sh)
{
  SH2movll0(sh,7,7);
}
#define init_0_7_8_14
void FASTCALL call_0_7_8_14(SH2_struct* sh)
{
  SH2movll0(sh,7,8);
}
#define init_0_7_9_14
void FASTCALL call_0_7_9_14(SH2_struct* sh)
{
  SH2movll0(sh,7,9);
}
#define init_0_7_10_14
void FASTCALL call_0_7_10_14(SH2_struct* sh)
{
  SH2movll0(sh,7,10);
}
#define init_0_7_11_14
void FASTCALL call_0_7_11_14(SH2_struct* sh)
{
  SH2movll0(sh,7,11);
}
#define init_0_7_12_14
void FASTCALL call_0_7_12_14(SH2_struct* sh)
{
  SH2movll0(sh,7,12);
}
#define init_0_7_13_14
void FASTCALL call_0_7_13_14(SH2_struct* sh)
{
  SH2movll0(sh,7,13);
}
#define init_0_7_14_14
void FASTCALL call_0_7_14_14(SH2_struct* sh)
{
  SH2movll0(sh,7,14);
}
#define init_0_7_15_14
void FASTCALL call_0_7_15_14(SH2_struct* sh)
{
  SH2movll0(sh,7,15);
}
#define init_0_8_0_14
void FASTCALL call_0_8_0_14(SH2_struct* sh)
{
  SH2movll0(sh,8,0);
}
#define init_0_8_1_14
void FASTCALL call_0_8_1_14(SH2_struct* sh)
{
  SH2movll0(sh,8,1);
}
#define init_0_8_2_14
void FASTCALL call_0_8_2_14(SH2_struct* sh)
{
  SH2movll0(sh,8,2);
}
#define init_0_8_3_14
void FASTCALL call_0_8_3_14(SH2_struct* sh)
{
  SH2movll0(sh,8,3);
}
#define init_0_8_4_14
void FASTCALL call_0_8_4_14(SH2_struct* sh)
{
  SH2movll0(sh,8,4);
}
#define init_0_8_5_14
void FASTCALL call_0_8_5_14(SH2_struct* sh)
{
  SH2movll0(sh,8,5);
}
#define init_0_8_6_14
void FASTCALL call_0_8_6_14(SH2_struct* sh)
{
  SH2movll0(sh,8,6);
}
#define init_0_8_7_14
void FASTCALL call_0_8_7_14(SH2_struct* sh)
{
  SH2movll0(sh,8,7);
}
#define init_0_8_8_14
void FASTCALL call_0_8_8_14(SH2_struct* sh)
{
  SH2movll0(sh,8,8);
}
#define init_0_8_9_14
void FASTCALL call_0_8_9_14(SH2_struct* sh)
{
  SH2movll0(sh,8,9);
}
#define init_0_8_10_14
void FASTCALL call_0_8_10_14(SH2_struct* sh)
{
  SH2movll0(sh,8,10);
}
#define init_0_8_11_14
void FASTCALL call_0_8_11_14(SH2_struct* sh)
{
  SH2movll0(sh,8,11);
}
#define init_0_8_12_14
void FASTCALL call_0_8_12_14(SH2_struct* sh)
{
  SH2movll0(sh,8,12);
}
#define init_0_8_13_14
void FASTCALL call_0_8_13_14(SH2_struct* sh)
{
  SH2movll0(sh,8,13);
}
#define init_0_8_14_14
void FASTCALL call_0_8_14_14(SH2_struct* sh)
{
  SH2movll0(sh,8,14);
}
#define init_0_8_15_14
void FASTCALL call_0_8_15_14(SH2_struct* sh)
{
  SH2movll0(sh,8,15);
}
#define init_0_9_0_14
void FASTCALL call_0_9_0_14(SH2_struct* sh)
{
  SH2movll0(sh,9,0);
}
#define init_0_9_1_14
void FASTCALL call_0_9_1_14(SH2_struct* sh)
{
  SH2movll0(sh,9,1);
}
#define init_0_9_2_14
void FASTCALL call_0_9_2_14(SH2_struct* sh)
{
  SH2movll0(sh,9,2);
}
#define init_0_9_3_14
void FASTCALL call_0_9_3_14(SH2_struct* sh)
{
  SH2movll0(sh,9,3);
}
#define init_0_9_4_14
void FASTCALL call_0_9_4_14(SH2_struct* sh)
{
  SH2movll0(sh,9,4);
}
#define init_0_9_5_14
void FASTCALL call_0_9_5_14(SH2_struct* sh)
{
  SH2movll0(sh,9,5);
}
#define init_0_9_6_14
void FASTCALL call_0_9_6_14(SH2_struct* sh)
{
  SH2movll0(sh,9,6);
}
#define init_0_9_7_14
void FASTCALL call_0_9_7_14(SH2_struct* sh)
{
  SH2movll0(sh,9,7);
}
#define init_0_9_8_14
void FASTCALL call_0_9_8_14(SH2_struct* sh)
{
  SH2movll0(sh,9,8);
}
#define init_0_9_9_14
void FASTCALL call_0_9_9_14(SH2_struct* sh)
{
  SH2movll0(sh,9,9);
}
#define init_0_9_10_14
void FASTCALL call_0_9_10_14(SH2_struct* sh)
{
  SH2movll0(sh,9,10);
}
#define init_0_9_11_14
void FASTCALL call_0_9_11_14(SH2_struct* sh)
{
  SH2movll0(sh,9,11);
}
#define init_0_9_12_14
void FASTCALL call_0_9_12_14(SH2_struct* sh)
{
  SH2movll0(sh,9,12);
}
#define init_0_9_13_14
void FASTCALL call_0_9_13_14(SH2_struct* sh)
{
  SH2movll0(sh,9,13);
}
#define init_0_9_14_14
void FASTCALL call_0_9_14_14(SH2_struct* sh)
{
  SH2movll0(sh,9,14);
}
#define init_0_9_15_14
void FASTCALL call_0_9_15_14(SH2_struct* sh)
{
  SH2movll0(sh,9,15);
}
#define init_0_10_0_14
void FASTCALL call_0_10_0_14(SH2_struct* sh)
{
  SH2movll0(sh,10,0);
}
#define init_0_10_1_14
void FASTCALL call_0_10_1_14(SH2_struct* sh)
{
  SH2movll0(sh,10,1);
}
#define init_0_10_2_14
void FASTCALL call_0_10_2_14(SH2_struct* sh)
{
  SH2movll0(sh,10,2);
}
#define init_0_10_3_14
void FASTCALL call_0_10_3_14(SH2_struct* sh)
{
  SH2movll0(sh,10,3);
}
#define init_0_10_4_14
void FASTCALL call_0_10_4_14(SH2_struct* sh)
{
  SH2movll0(sh,10,4);
}
#define init_0_10_5_14
void FASTCALL call_0_10_5_14(SH2_struct* sh)
{
  SH2movll0(sh,10,5);
}
#define init_0_10_6_14
void FASTCALL call_0_10_6_14(SH2_struct* sh)
{
  SH2movll0(sh,10,6);
}
#define init_0_10_7_14
void FASTCALL call_0_10_7_14(SH2_struct* sh)
{
  SH2movll0(sh,10,7);
}
#define init_0_10_8_14
void FASTCALL call_0_10_8_14(SH2_struct* sh)
{
  SH2movll0(sh,10,8);
}
#define init_0_10_9_14
void FASTCALL call_0_10_9_14(SH2_struct* sh)
{
  SH2movll0(sh,10,9);
}
#define init_0_10_10_14
void FASTCALL call_0_10_10_14(SH2_struct* sh)
{
  SH2movll0(sh,10,10);
}
#define init_0_10_11_14
void FASTCALL call_0_10_11_14(SH2_struct* sh)
{
  SH2movll0(sh,10,11);
}
#define init_0_10_12_14
void FASTCALL call_0_10_12_14(SH2_struct* sh)
{
  SH2movll0(sh,10,12);
}
#define init_0_10_13_14
void FASTCALL call_0_10_13_14(SH2_struct* sh)
{
  SH2movll0(sh,10,13);
}
#define init_0_10_14_14
void FASTCALL call_0_10_14_14(SH2_struct* sh)
{
  SH2movll0(sh,10,14);
}
#define init_0_10_15_14
void FASTCALL call_0_10_15_14(SH2_struct* sh)
{
  SH2movll0(sh,10,15);
}
#define init_0_11_0_14
void FASTCALL call_0_11_0_14(SH2_struct* sh)
{
  SH2movll0(sh,11,0);
}
#define init_0_11_1_14
void FASTCALL call_0_11_1_14(SH2_struct* sh)
{
  SH2movll0(sh,11,1);
}
#define init_0_11_2_14
void FASTCALL call_0_11_2_14(SH2_struct* sh)
{
  SH2movll0(sh,11,2);
}
#define init_0_11_3_14
void FASTCALL call_0_11_3_14(SH2_struct* sh)
{
  SH2movll0(sh,11,3);
}
#define init_0_11_4_14
void FASTCALL call_0_11_4_14(SH2_struct* sh)
{
  SH2movll0(sh,11,4);
}
#define init_0_11_5_14
void FASTCALL call_0_11_5_14(SH2_struct* sh)
{
  SH2movll0(sh,11,5);
}
#define init_0_11_6_14
void FASTCALL call_0_11_6_14(SH2_struct* sh)
{
  SH2movll0(sh,11,6);
}
#define init_0_11_7_14
void FASTCALL call_0_11_7_14(SH2_struct* sh)
{
  SH2movll0(sh,11,7);
}
#define init_0_11_8_14
void FASTCALL call_0_11_8_14(SH2_struct* sh)
{
  SH2movll0(sh,11,8);
}
#define init_0_11_9_14
void FASTCALL call_0_11_9_14(SH2_struct* sh)
{
  SH2movll0(sh,11,9);
}
#define init_0_11_10_14
void FASTCALL call_0_11_10_14(SH2_struct* sh)
{
  SH2movll0(sh,11,10);
}
#define init_0_11_11_14
void FASTCALL call_0_11_11_14(SH2_struct* sh)
{
  SH2movll0(sh,11,11);
}
#define init_0_11_12_14
void FASTCALL call_0_11_12_14(SH2_struct* sh)
{
  SH2movll0(sh,11,12);
}
#define init_0_11_13_14
void FASTCALL call_0_11_13_14(SH2_struct* sh)
{
  SH2movll0(sh,11,13);
}
#define init_0_11_14_14
void FASTCALL call_0_11_14_14(SH2_struct* sh)
{
  SH2movll0(sh,11,14);
}
#define init_0_11_15_14
void FASTCALL call_0_11_15_14(SH2_struct* sh)
{
  SH2movll0(sh,11,15);
}
#define init_0_12_0_14
void FASTCALL call_0_12_0_14(SH2_struct* sh)
{
  SH2movll0(sh,12,0);
}
#define init_0_12_1_14
void FASTCALL call_0_12_1_14(SH2_struct* sh)
{
  SH2movll0(sh,12,1);
}
#define init_0_12_2_14
void FASTCALL call_0_12_2_14(SH2_struct* sh)
{
  SH2movll0(sh,12,2);
}
#define init_0_12_3_14
void FASTCALL call_0_12_3_14(SH2_struct* sh)
{
  SH2movll0(sh,12,3);
}
#define init_0_12_4_14
void FASTCALL call_0_12_4_14(SH2_struct* sh)
{
  SH2movll0(sh,12,4);
}
#define init_0_12_5_14
void FASTCALL call_0_12_5_14(SH2_struct* sh)
{
  SH2movll0(sh,12,5);
}
#define init_0_12_6_14
void FASTCALL call_0_12_6_14(SH2_struct* sh)
{
  SH2movll0(sh,12,6);
}
#define init_0_12_7_14
void FASTCALL call_0_12_7_14(SH2_struct* sh)
{
  SH2movll0(sh,12,7);
}
#define init_0_12_8_14
void FASTCALL call_0_12_8_14(SH2_struct* sh)
{
  SH2movll0(sh,12,8);
}
#define init_0_12_9_14
void FASTCALL call_0_12_9_14(SH2_struct* sh)
{
  SH2movll0(sh,12,9);
}
#define init_0_12_10_14
void FASTCALL call_0_12_10_14(SH2_struct* sh)
{
  SH2movll0(sh,12,10);
}
#define init_0_12_11_14
void FASTCALL call_0_12_11_14(SH2_struct* sh)
{
  SH2movll0(sh,12,11);
}
#define init_0_12_12_14
void FASTCALL call_0_12_12_14(SH2_struct* sh)
{
  SH2movll0(sh,12,12);
}
#define init_0_12_13_14
void FASTCALL call_0_12_13_14(SH2_struct* sh)
{
  SH2movll0(sh,12,13);
}
#define init_0_12_14_14
void FASTCALL call_0_12_14_14(SH2_struct* sh)
{
  SH2movll0(sh,12,14);
}
#define init_0_12_15_14
void FASTCALL call_0_12_15_14(SH2_struct* sh)
{
  SH2movll0(sh,12,15);
}
#define init_0_13_0_14
void FASTCALL call_0_13_0_14(SH2_struct* sh)
{
  SH2movll0(sh,13,0);
}
#define init_0_13_1_14
void FASTCALL call_0_13_1_14(SH2_struct* sh)
{
  SH2movll0(sh,13,1);
}
#define init_0_13_2_14
void FASTCALL call_0_13_2_14(SH2_struct* sh)
{
  SH2movll0(sh,13,2);
}
#define init_0_13_3_14
void FASTCALL call_0_13_3_14(SH2_struct* sh)
{
  SH2movll0(sh,13,3);
}
#define init_0_13_4_14
void FASTCALL call_0_13_4_14(SH2_struct* sh)
{
  SH2movll0(sh,13,4);
}
#define init_0_13_5_14
void FASTCALL call_0_13_5_14(SH2_struct* sh)
{
  SH2movll0(sh,13,5);
}
#define init_0_13_6_14
void FASTCALL call_0_13_6_14(SH2_struct* sh)
{
  SH2movll0(sh,13,6);
}
#define init_0_13_7_14
void FASTCALL call_0_13_7_14(SH2_struct* sh)
{
  SH2movll0(sh,13,7);
}
#define init_0_13_8_14
void FASTCALL call_0_13_8_14(SH2_struct* sh)
{
  SH2movll0(sh,13,8);
}
#define init_0_13_9_14
void FASTCALL call_0_13_9_14(SH2_struct* sh)
{
  SH2movll0(sh,13,9);
}
#define init_0_13_10_14
void FASTCALL call_0_13_10_14(SH2_struct* sh)
{
  SH2movll0(sh,13,10);
}
#define init_0_13_11_14
void FASTCALL call_0_13_11_14(SH2_struct* sh)
{
  SH2movll0(sh,13,11);
}
#define init_0_13_12_14
void FASTCALL call_0_13_12_14(SH2_struct* sh)
{
  SH2movll0(sh,13,12);
}
#define init_0_13_13_14
void FASTCALL call_0_13_13_14(SH2_struct* sh)
{
  SH2movll0(sh,13,13);
}
#define init_0_13_14_14
void FASTCALL call_0_13_14_14(SH2_struct* sh)
{
  SH2movll0(sh,13,14);
}
#define init_0_13_15_14
void FASTCALL call_0_13_15_14(SH2_struct* sh)
{
  SH2movll0(sh,13,15);
}
#define init_0_14_0_14
void FASTCALL call_0_14_0_14(SH2_struct* sh)
{
  SH2movll0(sh,14,0);
}
#define init_0_14_1_14
void FASTCALL call_0_14_1_14(SH2_struct* sh)
{
  SH2movll0(sh,14,1);
}
#define init_0_14_2_14
void FASTCALL call_0_14_2_14(SH2_struct* sh)
{
  SH2movll0(sh,14,2);
}
#define init_0_14_3_14
void FASTCALL call_0_14_3_14(SH2_struct* sh)
{
  SH2movll0(sh,14,3);
}
#define init_0_14_4_14
void FASTCALL call_0_14_4_14(SH2_struct* sh)
{
  SH2movll0(sh,14,4);
}
#define init_0_14_5_14
void FASTCALL call_0_14_5_14(SH2_struct* sh)
{
  SH2movll0(sh,14,5);
}
#define init_0_14_6_14
void FASTCALL call_0_14_6_14(SH2_struct* sh)
{
  SH2movll0(sh,14,6);
}
#define init_0_14_7_14
void FASTCALL call_0_14_7_14(SH2_struct* sh)
{
  SH2movll0(sh,14,7);
}
#define init_0_14_8_14
void FASTCALL call_0_14_8_14(SH2_struct* sh)
{
  SH2movll0(sh,14,8);
}
#define init_0_14_9_14
void FASTCALL call_0_14_9_14(SH2_struct* sh)
{
  SH2movll0(sh,14,9);
}
#define init_0_14_10_14
void FASTCALL call_0_14_10_14(SH2_struct* sh)
{
  SH2movll0(sh,14,10);
}
#define init_0_14_11_14
void FASTCALL call_0_14_11_14(SH2_struct* sh)
{
  SH2movll0(sh,14,11);
}
#define init_0_14_12_14
void FASTCALL call_0_14_12_14(SH2_struct* sh)
{
  SH2movll0(sh,14,12);
}
#define init_0_14_13_14
void FASTCALL call_0_14_13_14(SH2_struct* sh)
{
  SH2movll0(sh,14,13);
}
#define init_0_14_14_14
void FASTCALL call_0_14_14_14(SH2_struct* sh)
{
  SH2movll0(sh,14,14);
}
#define init_0_14_15_14
void FASTCALL call_0_14_15_14(SH2_struct* sh)
{
  SH2movll0(sh,14,15);
}
#define init_0_15_0_14
void FASTCALL call_0_15_0_14(SH2_struct* sh)
{
  SH2movll0(sh,15,0);
}
#define init_0_15_1_14
void FASTCALL call_0_15_1_14(SH2_struct* sh)
{
  SH2movll0(sh,15,1);
}
#define init_0_15_2_14
void FASTCALL call_0_15_2_14(SH2_struct* sh)
{
  SH2movll0(sh,15,2);
}
#define init_0_15_3_14
void FASTCALL call_0_15_3_14(SH2_struct* sh)
{
  SH2movll0(sh,15,3);
}
#define init_0_15_4_14
void FASTCALL call_0_15_4_14(SH2_struct* sh)
{
  SH2movll0(sh,15,4);
}
#define init_0_15_5_14
void FASTCALL call_0_15_5_14(SH2_struct* sh)
{
  SH2movll0(sh,15,5);
}
#define init_0_15_6_14
void FASTCALL call_0_15_6_14(SH2_struct* sh)
{
  SH2movll0(sh,15,6);
}
#define init_0_15_7_14
void FASTCALL call_0_15_7_14(SH2_struct* sh)
{
  SH2movll0(sh,15,7);
}
#define init_0_15_8_14
void FASTCALL call_0_15_8_14(SH2_struct* sh)
{
  SH2movll0(sh,15,8);
}
#define init_0_15_9_14
void FASTCALL call_0_15_9_14(SH2_struct* sh)
{
  SH2movll0(sh,15,9);
}
#define init_0_15_10_14
void FASTCALL call_0_15_10_14(SH2_struct* sh)
{
  SH2movll0(sh,15,10);
}
#define init_0_15_11_14
void FASTCALL call_0_15_11_14(SH2_struct* sh)
{
  SH2movll0(sh,15,11);
}
#define init_0_15_12_14
void FASTCALL call_0_15_12_14(SH2_struct* sh)
{
  SH2movll0(sh,15,12);
}
#define init_0_15_13_14
void FASTCALL call_0_15_13_14(SH2_struct* sh)
{
  SH2movll0(sh,15,13);
}
#define init_0_15_14_14
void FASTCALL call_0_15_14_14(SH2_struct* sh)
{
  SH2movll0(sh,15,14);
}
#define init_0_15_15_14
void FASTCALL call_0_15_15_14(SH2_struct* sh)
{
  SH2movll0(sh,15,15);
}
#define init_0_0_0_4
void FASTCALL call_0_0_0_4(SH2_struct* sh)
{
  SH2movbs0(sh,0,0);
}
#define init_0_0_1_4
void FASTCALL call_0_0_1_4(SH2_struct* sh)
{
  SH2movbs0(sh,0,1);
}
#define init_0_0_2_4
void FASTCALL call_0_0_2_4(SH2_struct* sh)
{
  SH2movbs0(sh,0,2);
}
#define init_0_0_3_4
void FASTCALL call_0_0_3_4(SH2_struct* sh)
{
  SH2movbs0(sh,0,3);
}
#define init_0_0_4_4
void FASTCALL call_0_0_4_4(SH2_struct* sh)
{
  SH2movbs0(sh,0,4);
}
#define init_0_0_5_4
void FASTCALL call_0_0_5_4(SH2_struct* sh)
{
  SH2movbs0(sh,0,5);
}
#define init_0_0_6_4
void FASTCALL call_0_0_6_4(SH2_struct* sh)
{
  SH2movbs0(sh,0,6);
}
#define init_0_0_7_4
void FASTCALL call_0_0_7_4(SH2_struct* sh)
{
  SH2movbs0(sh,0,7);
}
#define init_0_0_8_4
void FASTCALL call_0_0_8_4(SH2_struct* sh)
{
  SH2movbs0(sh,0,8);
}
#define init_0_0_9_4
void FASTCALL call_0_0_9_4(SH2_struct* sh)
{
  SH2movbs0(sh,0,9);
}
#define init_0_0_10_4
void FASTCALL call_0_0_10_4(SH2_struct* sh)
{
  SH2movbs0(sh,0,10);
}
#define init_0_0_11_4
void FASTCALL call_0_0_11_4(SH2_struct* sh)
{
  SH2movbs0(sh,0,11);
}
#define init_0_0_12_4
void FASTCALL call_0_0_12_4(SH2_struct* sh)
{
  SH2movbs0(sh,0,12);
}
#define init_0_0_13_4
void FASTCALL call_0_0_13_4(SH2_struct* sh)
{
  SH2movbs0(sh,0,13);
}
#define init_0_0_14_4
void FASTCALL call_0_0_14_4(SH2_struct* sh)
{
  SH2movbs0(sh,0,14);
}
#define init_0_0_15_4
void FASTCALL call_0_0_15_4(SH2_struct* sh)
{
  SH2movbs0(sh,0,15);
}
#define init_0_1_0_4
void FASTCALL call_0_1_0_4(SH2_struct* sh)
{
  SH2movbs0(sh,1,0);
}
#define init_0_1_1_4
void FASTCALL call_0_1_1_4(SH2_struct* sh)
{
  SH2movbs0(sh,1,1);
}
#define init_0_1_2_4
void FASTCALL call_0_1_2_4(SH2_struct* sh)
{
  SH2movbs0(sh,1,2);
}
#define init_0_1_3_4
void FASTCALL call_0_1_3_4(SH2_struct* sh)
{
  SH2movbs0(sh,1,3);
}
#define init_0_1_4_4
void FASTCALL call_0_1_4_4(SH2_struct* sh)
{
  SH2movbs0(sh,1,4);
}
#define init_0_1_5_4
void FASTCALL call_0_1_5_4(SH2_struct* sh)
{
  SH2movbs0(sh,1,5);
}
#define init_0_1_6_4
void FASTCALL call_0_1_6_4(SH2_struct* sh)
{
  SH2movbs0(sh,1,6);
}
#define init_0_1_7_4
void FASTCALL call_0_1_7_4(SH2_struct* sh)
{
  SH2movbs0(sh,1,7);
}
#define init_0_1_8_4
void FASTCALL call_0_1_8_4(SH2_struct* sh)
{
  SH2movbs0(sh,1,8);
}
#define init_0_1_9_4
void FASTCALL call_0_1_9_4(SH2_struct* sh)
{
  SH2movbs0(sh,1,9);
}
#define init_0_1_10_4
void FASTCALL call_0_1_10_4(SH2_struct* sh)
{
  SH2movbs0(sh,1,10);
}
#define init_0_1_11_4
void FASTCALL call_0_1_11_4(SH2_struct* sh)
{
  SH2movbs0(sh,1,11);
}
#define init_0_1_12_4
void FASTCALL call_0_1_12_4(SH2_struct* sh)
{
  SH2movbs0(sh,1,12);
}
#define init_0_1_13_4
void FASTCALL call_0_1_13_4(SH2_struct* sh)
{
  SH2movbs0(sh,1,13);
}
#define init_0_1_14_4
void FASTCALL call_0_1_14_4(SH2_struct* sh)
{
  SH2movbs0(sh,1,14);
}
#define init_0_1_15_4
void FASTCALL call_0_1_15_4(SH2_struct* sh)
{
  SH2movbs0(sh,1,15);
}
#define init_0_2_0_4
void FASTCALL call_0_2_0_4(SH2_struct* sh)
{
  SH2movbs0(sh,2,0);
}
#define init_0_2_1_4
void FASTCALL call_0_2_1_4(SH2_struct* sh)
{
  SH2movbs0(sh,2,1);
}
#define init_0_2_2_4
void FASTCALL call_0_2_2_4(SH2_struct* sh)
{
  SH2movbs0(sh,2,2);
}
#define init_0_2_3_4
void FASTCALL call_0_2_3_4(SH2_struct* sh)
{
  SH2movbs0(sh,2,3);
}
#define init_0_2_4_4
void FASTCALL call_0_2_4_4(SH2_struct* sh)
{
  SH2movbs0(sh,2,4);
}
#define init_0_2_5_4
void FASTCALL call_0_2_5_4(SH2_struct* sh)
{
  SH2movbs0(sh,2,5);
}
#define init_0_2_6_4
void FASTCALL call_0_2_6_4(SH2_struct* sh)
{
  SH2movbs0(sh,2,6);
}
#define init_0_2_7_4
void FASTCALL call_0_2_7_4(SH2_struct* sh)
{
  SH2movbs0(sh,2,7);
}
#define init_0_2_8_4
void FASTCALL call_0_2_8_4(SH2_struct* sh)
{
  SH2movbs0(sh,2,8);
}
#define init_0_2_9_4
void FASTCALL call_0_2_9_4(SH2_struct* sh)
{
  SH2movbs0(sh,2,9);
}
#define init_0_2_10_4
void FASTCALL call_0_2_10_4(SH2_struct* sh)
{
  SH2movbs0(sh,2,10);
}
#define init_0_2_11_4
void FASTCALL call_0_2_11_4(SH2_struct* sh)
{
  SH2movbs0(sh,2,11);
}
#define init_0_2_12_4
void FASTCALL call_0_2_12_4(SH2_struct* sh)
{
  SH2movbs0(sh,2,12);
}
#define init_0_2_13_4
void FASTCALL call_0_2_13_4(SH2_struct* sh)
{
  SH2movbs0(sh,2,13);
}
#define init_0_2_14_4
void FASTCALL call_0_2_14_4(SH2_struct* sh)
{
  SH2movbs0(sh,2,14);
}
#define init_0_2_15_4
void FASTCALL call_0_2_15_4(SH2_struct* sh)
{
  SH2movbs0(sh,2,15);
}
#define init_0_3_0_4
void FASTCALL call_0_3_0_4(SH2_struct* sh)
{
  SH2movbs0(sh,3,0);
}
#define init_0_3_1_4
void FASTCALL call_0_3_1_4(SH2_struct* sh)
{
  SH2movbs0(sh,3,1);
}
#define init_0_3_2_4
void FASTCALL call_0_3_2_4(SH2_struct* sh)
{
  SH2movbs0(sh,3,2);
}
#define init_0_3_3_4
void FASTCALL call_0_3_3_4(SH2_struct* sh)
{
  SH2movbs0(sh,3,3);
}
#define init_0_3_4_4
void FASTCALL call_0_3_4_4(SH2_struct* sh)
{
  SH2movbs0(sh,3,4);
}
#define init_0_3_5_4
void FASTCALL call_0_3_5_4(SH2_struct* sh)
{
  SH2movbs0(sh,3,5);
}
#define init_0_3_6_4
void FASTCALL call_0_3_6_4(SH2_struct* sh)
{
  SH2movbs0(sh,3,6);
}
#define init_0_3_7_4
void FASTCALL call_0_3_7_4(SH2_struct* sh)
{
  SH2movbs0(sh,3,7);
}
#define init_0_3_8_4
void FASTCALL call_0_3_8_4(SH2_struct* sh)
{
  SH2movbs0(sh,3,8);
}
#define init_0_3_9_4
void FASTCALL call_0_3_9_4(SH2_struct* sh)
{
  SH2movbs0(sh,3,9);
}
#define init_0_3_10_4
void FASTCALL call_0_3_10_4(SH2_struct* sh)
{
  SH2movbs0(sh,3,10);
}
#define init_0_3_11_4
void FASTCALL call_0_3_11_4(SH2_struct* sh)
{
  SH2movbs0(sh,3,11);
}
#define init_0_3_12_4
void FASTCALL call_0_3_12_4(SH2_struct* sh)
{
  SH2movbs0(sh,3,12);
}
#define init_0_3_13_4
void FASTCALL call_0_3_13_4(SH2_struct* sh)
{
  SH2movbs0(sh,3,13);
}
#define init_0_3_14_4
void FASTCALL call_0_3_14_4(SH2_struct* sh)
{
  SH2movbs0(sh,3,14);
}
#define init_0_3_15_4
void FASTCALL call_0_3_15_4(SH2_struct* sh)
{
  SH2movbs0(sh,3,15);
}
#define init_0_4_0_4
void FASTCALL call_0_4_0_4(SH2_struct* sh)
{
  SH2movbs0(sh,4,0);
}
#define init_0_4_1_4
void FASTCALL call_0_4_1_4(SH2_struct* sh)
{
  SH2movbs0(sh,4,1);
}
#define init_0_4_2_4
void FASTCALL call_0_4_2_4(SH2_struct* sh)
{
  SH2movbs0(sh,4,2);
}
#define init_0_4_3_4
void FASTCALL call_0_4_3_4(SH2_struct* sh)
{
  SH2movbs0(sh,4,3);
}
#define init_0_4_4_4
void FASTCALL call_0_4_4_4(SH2_struct* sh)
{
  SH2movbs0(sh,4,4);
}
#define init_0_4_5_4
void FASTCALL call_0_4_5_4(SH2_struct* sh)
{
  SH2movbs0(sh,4,5);
}
#define init_0_4_6_4
void FASTCALL call_0_4_6_4(SH2_struct* sh)
{
  SH2movbs0(sh,4,6);
}
#define init_0_4_7_4
void FASTCALL call_0_4_7_4(SH2_struct* sh)
{
  SH2movbs0(sh,4,7);
}
#define init_0_4_8_4
void FASTCALL call_0_4_8_4(SH2_struct* sh)
{
  SH2movbs0(sh,4,8);
}
#define init_0_4_9_4
void FASTCALL call_0_4_9_4(SH2_struct* sh)
{
  SH2movbs0(sh,4,9);
}
#define init_0_4_10_4
void FASTCALL call_0_4_10_4(SH2_struct* sh)
{
  SH2movbs0(sh,4,10);
}
#define init_0_4_11_4
void FASTCALL call_0_4_11_4(SH2_struct* sh)
{
  SH2movbs0(sh,4,11);
}
#define init_0_4_12_4
void FASTCALL call_0_4_12_4(SH2_struct* sh)
{
  SH2movbs0(sh,4,12);
}
#define init_0_4_13_4
void FASTCALL call_0_4_13_4(SH2_struct* sh)
{
  SH2movbs0(sh,4,13);
}
#define init_0_4_14_4
void FASTCALL call_0_4_14_4(SH2_struct* sh)
{
  SH2movbs0(sh,4,14);
}
#define init_0_4_15_4
void FASTCALL call_0_4_15_4(SH2_struct* sh)
{
  SH2movbs0(sh,4,15);
}
#define init_0_5_0_4
void FASTCALL call_0_5_0_4(SH2_struct* sh)
{
  SH2movbs0(sh,5,0);
}
#define init_0_5_1_4
void FASTCALL call_0_5_1_4(SH2_struct* sh)
{
  SH2movbs0(sh,5,1);
}
#define init_0_5_2_4
void FASTCALL call_0_5_2_4(SH2_struct* sh)
{
  SH2movbs0(sh,5,2);
}
#define init_0_5_3_4
void FASTCALL call_0_5_3_4(SH2_struct* sh)
{
  SH2movbs0(sh,5,3);
}
#define init_0_5_4_4
void FASTCALL call_0_5_4_4(SH2_struct* sh)
{
  SH2movbs0(sh,5,4);
}
#define init_0_5_5_4
void FASTCALL call_0_5_5_4(SH2_struct* sh)
{
  SH2movbs0(sh,5,5);
}
#define init_0_5_6_4
void FASTCALL call_0_5_6_4(SH2_struct* sh)
{
  SH2movbs0(sh,5,6);
}
#define init_0_5_7_4
void FASTCALL call_0_5_7_4(SH2_struct* sh)
{
  SH2movbs0(sh,5,7);
}
#define init_0_5_8_4
void FASTCALL call_0_5_8_4(SH2_struct* sh)
{
  SH2movbs0(sh,5,8);
}
#define init_0_5_9_4
void FASTCALL call_0_5_9_4(SH2_struct* sh)
{
  SH2movbs0(sh,5,9);
}
#define init_0_5_10_4
void FASTCALL call_0_5_10_4(SH2_struct* sh)
{
  SH2movbs0(sh,5,10);
}
#define init_0_5_11_4
void FASTCALL call_0_5_11_4(SH2_struct* sh)
{
  SH2movbs0(sh,5,11);
}
#define init_0_5_12_4
void FASTCALL call_0_5_12_4(SH2_struct* sh)
{
  SH2movbs0(sh,5,12);
}
#define init_0_5_13_4
void FASTCALL call_0_5_13_4(SH2_struct* sh)
{
  SH2movbs0(sh,5,13);
}
#define init_0_5_14_4
void FASTCALL call_0_5_14_4(SH2_struct* sh)
{
  SH2movbs0(sh,5,14);
}
#define init_0_5_15_4
void FASTCALL call_0_5_15_4(SH2_struct* sh)
{
  SH2movbs0(sh,5,15);
}
#define init_0_6_0_4
void FASTCALL call_0_6_0_4(SH2_struct* sh)
{
  SH2movbs0(sh,6,0);
}
#define init_0_6_1_4
void FASTCALL call_0_6_1_4(SH2_struct* sh)
{
  SH2movbs0(sh,6,1);
}
#define init_0_6_2_4
void FASTCALL call_0_6_2_4(SH2_struct* sh)
{
  SH2movbs0(sh,6,2);
}
#define init_0_6_3_4
void FASTCALL call_0_6_3_4(SH2_struct* sh)
{
  SH2movbs0(sh,6,3);
}
#define init_0_6_4_4
void FASTCALL call_0_6_4_4(SH2_struct* sh)
{
  SH2movbs0(sh,6,4);
}
#define init_0_6_5_4
void FASTCALL call_0_6_5_4(SH2_struct* sh)
{
  SH2movbs0(sh,6,5);
}
#define init_0_6_6_4
void FASTCALL call_0_6_6_4(SH2_struct* sh)
{
  SH2movbs0(sh,6,6);
}
#define init_0_6_7_4
void FASTCALL call_0_6_7_4(SH2_struct* sh)
{
  SH2movbs0(sh,6,7);
}
#define init_0_6_8_4
void FASTCALL call_0_6_8_4(SH2_struct* sh)
{
  SH2movbs0(sh,6,8);
}
#define init_0_6_9_4
void FASTCALL call_0_6_9_4(SH2_struct* sh)
{
  SH2movbs0(sh,6,9);
}
#define init_0_6_10_4
void FASTCALL call_0_6_10_4(SH2_struct* sh)
{
  SH2movbs0(sh,6,10);
}
#define init_0_6_11_4
void FASTCALL call_0_6_11_4(SH2_struct* sh)
{
  SH2movbs0(sh,6,11);
}
#define init_0_6_12_4
void FASTCALL call_0_6_12_4(SH2_struct* sh)
{
  SH2movbs0(sh,6,12);
}
#define init_0_6_13_4
void FASTCALL call_0_6_13_4(SH2_struct* sh)
{
  SH2movbs0(sh,6,13);
}
#define init_0_6_14_4
void FASTCALL call_0_6_14_4(SH2_struct* sh)
{
  SH2movbs0(sh,6,14);
}
#define init_0_6_15_4
void FASTCALL call_0_6_15_4(SH2_struct* sh)
{
  SH2movbs0(sh,6,15);
}
#define init_0_7_0_4
void FASTCALL call_0_7_0_4(SH2_struct* sh)
{
  SH2movbs0(sh,7,0);
}
#define init_0_7_1_4
void FASTCALL call_0_7_1_4(SH2_struct* sh)
{
  SH2movbs0(sh,7,1);
}
#define init_0_7_2_4
void FASTCALL call_0_7_2_4(SH2_struct* sh)
{
  SH2movbs0(sh,7,2);
}
#define init_0_7_3_4
void FASTCALL call_0_7_3_4(SH2_struct* sh)
{
  SH2movbs0(sh,7,3);
}
#define init_0_7_4_4
void FASTCALL call_0_7_4_4(SH2_struct* sh)
{
  SH2movbs0(sh,7,4);
}
#define init_0_7_5_4
void FASTCALL call_0_7_5_4(SH2_struct* sh)
{
  SH2movbs0(sh,7,5);
}
#define init_0_7_6_4
void FASTCALL call_0_7_6_4(SH2_struct* sh)
{
  SH2movbs0(sh,7,6);
}
#define init_0_7_7_4
void FASTCALL call_0_7_7_4(SH2_struct* sh)
{
  SH2movbs0(sh,7,7);
}
#define init_0_7_8_4
void FASTCALL call_0_7_8_4(SH2_struct* sh)
{
  SH2movbs0(sh,7,8);
}
#define init_0_7_9_4
void FASTCALL call_0_7_9_4(SH2_struct* sh)
{
  SH2movbs0(sh,7,9);
}
#define init_0_7_10_4
void FASTCALL call_0_7_10_4(SH2_struct* sh)
{
  SH2movbs0(sh,7,10);
}
#define init_0_7_11_4
void FASTCALL call_0_7_11_4(SH2_struct* sh)
{
  SH2movbs0(sh,7,11);
}
#define init_0_7_12_4
void FASTCALL call_0_7_12_4(SH2_struct* sh)
{
  SH2movbs0(sh,7,12);
}
#define init_0_7_13_4
void FASTCALL call_0_7_13_4(SH2_struct* sh)
{
  SH2movbs0(sh,7,13);
}
#define init_0_7_14_4
void FASTCALL call_0_7_14_4(SH2_struct* sh)
{
  SH2movbs0(sh,7,14);
}
#define init_0_7_15_4
void FASTCALL call_0_7_15_4(SH2_struct* sh)
{
  SH2movbs0(sh,7,15);
}
#define init_0_8_0_4
void FASTCALL call_0_8_0_4(SH2_struct* sh)
{
  SH2movbs0(sh,8,0);
}
#define init_0_8_1_4
void FASTCALL call_0_8_1_4(SH2_struct* sh)
{
  SH2movbs0(sh,8,1);
}
#define init_0_8_2_4
void FASTCALL call_0_8_2_4(SH2_struct* sh)
{
  SH2movbs0(sh,8,2);
}
#define init_0_8_3_4
void FASTCALL call_0_8_3_4(SH2_struct* sh)
{
  SH2movbs0(sh,8,3);
}
#define init_0_8_4_4
void FASTCALL call_0_8_4_4(SH2_struct* sh)
{
  SH2movbs0(sh,8,4);
}
#define init_0_8_5_4
void FASTCALL call_0_8_5_4(SH2_struct* sh)
{
  SH2movbs0(sh,8,5);
}
#define init_0_8_6_4
void FASTCALL call_0_8_6_4(SH2_struct* sh)
{
  SH2movbs0(sh,8,6);
}
#define init_0_8_7_4
void FASTCALL call_0_8_7_4(SH2_struct* sh)
{
  SH2movbs0(sh,8,7);
}
#define init_0_8_8_4
void FASTCALL call_0_8_8_4(SH2_struct* sh)
{
  SH2movbs0(sh,8,8);
}
#define init_0_8_9_4
void FASTCALL call_0_8_9_4(SH2_struct* sh)
{
  SH2movbs0(sh,8,9);
}
#define init_0_8_10_4
void FASTCALL call_0_8_10_4(SH2_struct* sh)
{
  SH2movbs0(sh,8,10);
}
#define init_0_8_11_4
void FASTCALL call_0_8_11_4(SH2_struct* sh)
{
  SH2movbs0(sh,8,11);
}
#define init_0_8_12_4
void FASTCALL call_0_8_12_4(SH2_struct* sh)
{
  SH2movbs0(sh,8,12);
}
#define init_0_8_13_4
void FASTCALL call_0_8_13_4(SH2_struct* sh)
{
  SH2movbs0(sh,8,13);
}
#define init_0_8_14_4
void FASTCALL call_0_8_14_4(SH2_struct* sh)
{
  SH2movbs0(sh,8,14);
}
#define init_0_8_15_4
void FASTCALL call_0_8_15_4(SH2_struct* sh)
{
  SH2movbs0(sh,8,15);
}
#define init_0_9_0_4
void FASTCALL call_0_9_0_4(SH2_struct* sh)
{
  SH2movbs0(sh,9,0);
}
#define init_0_9_1_4
void FASTCALL call_0_9_1_4(SH2_struct* sh)
{
  SH2movbs0(sh,9,1);
}
#define init_0_9_2_4
void FASTCALL call_0_9_2_4(SH2_struct* sh)
{
  SH2movbs0(sh,9,2);
}
#define init_0_9_3_4
void FASTCALL call_0_9_3_4(SH2_struct* sh)
{
  SH2movbs0(sh,9,3);
}
#define init_0_9_4_4
void FASTCALL call_0_9_4_4(SH2_struct* sh)
{
  SH2movbs0(sh,9,4);
}
#define init_0_9_5_4
void FASTCALL call_0_9_5_4(SH2_struct* sh)
{
  SH2movbs0(sh,9,5);
}
#define init_0_9_6_4
void FASTCALL call_0_9_6_4(SH2_struct* sh)
{
  SH2movbs0(sh,9,6);
}
#define init_0_9_7_4
void FASTCALL call_0_9_7_4(SH2_struct* sh)
{
  SH2movbs0(sh,9,7);
}
#define init_0_9_8_4
void FASTCALL call_0_9_8_4(SH2_struct* sh)
{
  SH2movbs0(sh,9,8);
}
#define init_0_9_9_4
void FASTCALL call_0_9_9_4(SH2_struct* sh)
{
  SH2movbs0(sh,9,9);
}
#define init_0_9_10_4
void FASTCALL call_0_9_10_4(SH2_struct* sh)
{
  SH2movbs0(sh,9,10);
}
#define init_0_9_11_4
void FASTCALL call_0_9_11_4(SH2_struct* sh)
{
  SH2movbs0(sh,9,11);
}
#define init_0_9_12_4
void FASTCALL call_0_9_12_4(SH2_struct* sh)
{
  SH2movbs0(sh,9,12);
}
#define init_0_9_13_4
void FASTCALL call_0_9_13_4(SH2_struct* sh)
{
  SH2movbs0(sh,9,13);
}
#define init_0_9_14_4
void FASTCALL call_0_9_14_4(SH2_struct* sh)
{
  SH2movbs0(sh,9,14);
}
#define init_0_9_15_4
void FASTCALL call_0_9_15_4(SH2_struct* sh)
{
  SH2movbs0(sh,9,15);
}
#define init_0_10_0_4
void FASTCALL call_0_10_0_4(SH2_struct* sh)
{
  SH2movbs0(sh,10,0);
}
#define init_0_10_1_4
void FASTCALL call_0_10_1_4(SH2_struct* sh)
{
  SH2movbs0(sh,10,1);
}
#define init_0_10_2_4
void FASTCALL call_0_10_2_4(SH2_struct* sh)
{
  SH2movbs0(sh,10,2);
}
#define init_0_10_3_4
void FASTCALL call_0_10_3_4(SH2_struct* sh)
{
  SH2movbs0(sh,10,3);
}
#define init_0_10_4_4
void FASTCALL call_0_10_4_4(SH2_struct* sh)
{
  SH2movbs0(sh,10,4);
}
#define init_0_10_5_4
void FASTCALL call_0_10_5_4(SH2_struct* sh)
{
  SH2movbs0(sh,10,5);
}
#define init_0_10_6_4
void FASTCALL call_0_10_6_4(SH2_struct* sh)
{
  SH2movbs0(sh,10,6);
}
#define init_0_10_7_4
void FASTCALL call_0_10_7_4(SH2_struct* sh)
{
  SH2movbs0(sh,10,7);
}
#define init_0_10_8_4
void FASTCALL call_0_10_8_4(SH2_struct* sh)
{
  SH2movbs0(sh,10,8);
}
#define init_0_10_9_4
void FASTCALL call_0_10_9_4(SH2_struct* sh)
{
  SH2movbs0(sh,10,9);
}
#define init_0_10_10_4
void FASTCALL call_0_10_10_4(SH2_struct* sh)
{
  SH2movbs0(sh,10,10);
}
#define init_0_10_11_4
void FASTCALL call_0_10_11_4(SH2_struct* sh)
{
  SH2movbs0(sh,10,11);
}
#define init_0_10_12_4
void FASTCALL call_0_10_12_4(SH2_struct* sh)
{
  SH2movbs0(sh,10,12);
}
#define init_0_10_13_4
void FASTCALL call_0_10_13_4(SH2_struct* sh)
{
  SH2movbs0(sh,10,13);
}
#define init_0_10_14_4
void FASTCALL call_0_10_14_4(SH2_struct* sh)
{
  SH2movbs0(sh,10,14);
}
#define init_0_10_15_4
void FASTCALL call_0_10_15_4(SH2_struct* sh)
{
  SH2movbs0(sh,10,15);
}
#define init_0_11_0_4
void FASTCALL call_0_11_0_4(SH2_struct* sh)
{
  SH2movbs0(sh,11,0);
}
#define init_0_11_1_4
void FASTCALL call_0_11_1_4(SH2_struct* sh)
{
  SH2movbs0(sh,11,1);
}
#define init_0_11_2_4
void FASTCALL call_0_11_2_4(SH2_struct* sh)
{
  SH2movbs0(sh,11,2);
}
#define init_0_11_3_4
void FASTCALL call_0_11_3_4(SH2_struct* sh)
{
  SH2movbs0(sh,11,3);
}
#define init_0_11_4_4
void FASTCALL call_0_11_4_4(SH2_struct* sh)
{
  SH2movbs0(sh,11,4);
}
#define init_0_11_5_4
void FASTCALL call_0_11_5_4(SH2_struct* sh)
{
  SH2movbs0(sh,11,5);
}
#define init_0_11_6_4
void FASTCALL call_0_11_6_4(SH2_struct* sh)
{
  SH2movbs0(sh,11,6);
}
#define init_0_11_7_4
void FASTCALL call_0_11_7_4(SH2_struct* sh)
{
  SH2movbs0(sh,11,7);
}
#define init_0_11_8_4
void FASTCALL call_0_11_8_4(SH2_struct* sh)
{
  SH2movbs0(sh,11,8);
}
#define init_0_11_9_4
void FASTCALL call_0_11_9_4(SH2_struct* sh)
{
  SH2movbs0(sh,11,9);
}
#define init_0_11_10_4
void FASTCALL call_0_11_10_4(SH2_struct* sh)
{
  SH2movbs0(sh,11,10);
}
#define init_0_11_11_4
void FASTCALL call_0_11_11_4(SH2_struct* sh)
{
  SH2movbs0(sh,11,11);
}
#define init_0_11_12_4
void FASTCALL call_0_11_12_4(SH2_struct* sh)
{
  SH2movbs0(sh,11,12);
}
#define init_0_11_13_4
void FASTCALL call_0_11_13_4(SH2_struct* sh)
{
  SH2movbs0(sh,11,13);
}
#define init_0_11_14_4
void FASTCALL call_0_11_14_4(SH2_struct* sh)
{
  SH2movbs0(sh,11,14);
}
#define init_0_11_15_4
void FASTCALL call_0_11_15_4(SH2_struct* sh)
{
  SH2movbs0(sh,11,15);
}
#define init_0_12_0_4
void FASTCALL call_0_12_0_4(SH2_struct* sh)
{
  SH2movbs0(sh,12,0);
}
#define init_0_12_1_4
void FASTCALL call_0_12_1_4(SH2_struct* sh)
{
  SH2movbs0(sh,12,1);
}
#define init_0_12_2_4
void FASTCALL call_0_12_2_4(SH2_struct* sh)
{
  SH2movbs0(sh,12,2);
}
#define init_0_12_3_4
void FASTCALL call_0_12_3_4(SH2_struct* sh)
{
  SH2movbs0(sh,12,3);
}
#define init_0_12_4_4
void FASTCALL call_0_12_4_4(SH2_struct* sh)
{
  SH2movbs0(sh,12,4);
}
#define init_0_12_5_4
void FASTCALL call_0_12_5_4(SH2_struct* sh)
{
  SH2movbs0(sh,12,5);
}
#define init_0_12_6_4
void FASTCALL call_0_12_6_4(SH2_struct* sh)
{
  SH2movbs0(sh,12,6);
}
#define init_0_12_7_4
void FASTCALL call_0_12_7_4(SH2_struct* sh)
{
  SH2movbs0(sh,12,7);
}
#define init_0_12_8_4
void FASTCALL call_0_12_8_4(SH2_struct* sh)
{
  SH2movbs0(sh,12,8);
}
#define init_0_12_9_4
void FASTCALL call_0_12_9_4(SH2_struct* sh)
{
  SH2movbs0(sh,12,9);
}
#define init_0_12_10_4
void FASTCALL call_0_12_10_4(SH2_struct* sh)
{
  SH2movbs0(sh,12,10);
}
#define init_0_12_11_4
void FASTCALL call_0_12_11_4(SH2_struct* sh)
{
  SH2movbs0(sh,12,11);
}
#define init_0_12_12_4
void FASTCALL call_0_12_12_4(SH2_struct* sh)
{
  SH2movbs0(sh,12,12);
}
#define init_0_12_13_4
void FASTCALL call_0_12_13_4(SH2_struct* sh)
{
  SH2movbs0(sh,12,13);
}
#define init_0_12_14_4
void FASTCALL call_0_12_14_4(SH2_struct* sh)
{
  SH2movbs0(sh,12,14);
}
#define init_0_12_15_4
void FASTCALL call_0_12_15_4(SH2_struct* sh)
{
  SH2movbs0(sh,12,15);
}
#define init_0_13_0_4
void FASTCALL call_0_13_0_4(SH2_struct* sh)
{
  SH2movbs0(sh,13,0);
}
#define init_0_13_1_4
void FASTCALL call_0_13_1_4(SH2_struct* sh)
{
  SH2movbs0(sh,13,1);
}
#define init_0_13_2_4
void FASTCALL call_0_13_2_4(SH2_struct* sh)
{
  SH2movbs0(sh,13,2);
}
#define init_0_13_3_4
void FASTCALL call_0_13_3_4(SH2_struct* sh)
{
  SH2movbs0(sh,13,3);
}
#define init_0_13_4_4
void FASTCALL call_0_13_4_4(SH2_struct* sh)
{
  SH2movbs0(sh,13,4);
}
#define init_0_13_5_4
void FASTCALL call_0_13_5_4(SH2_struct* sh)
{
  SH2movbs0(sh,13,5);
}
#define init_0_13_6_4
void FASTCALL call_0_13_6_4(SH2_struct* sh)
{
  SH2movbs0(sh,13,6);
}
#define init_0_13_7_4
void FASTCALL call_0_13_7_4(SH2_struct* sh)
{
  SH2movbs0(sh,13,7);
}
#define init_0_13_8_4
void FASTCALL call_0_13_8_4(SH2_struct* sh)
{
  SH2movbs0(sh,13,8);
}
#define init_0_13_9_4
void FASTCALL call_0_13_9_4(SH2_struct* sh)
{
  SH2movbs0(sh,13,9);
}
#define init_0_13_10_4
void FASTCALL call_0_13_10_4(SH2_struct* sh)
{
  SH2movbs0(sh,13,10);
}
#define init_0_13_11_4
void FASTCALL call_0_13_11_4(SH2_struct* sh)
{
  SH2movbs0(sh,13,11);
}
#define init_0_13_12_4
void FASTCALL call_0_13_12_4(SH2_struct* sh)
{
  SH2movbs0(sh,13,12);
}
#define init_0_13_13_4
void FASTCALL call_0_13_13_4(SH2_struct* sh)
{
  SH2movbs0(sh,13,13);
}
#define init_0_13_14_4
void FASTCALL call_0_13_14_4(SH2_struct* sh)
{
  SH2movbs0(sh,13,14);
}
#define init_0_13_15_4
void FASTCALL call_0_13_15_4(SH2_struct* sh)
{
  SH2movbs0(sh,13,15);
}
#define init_0_14_0_4
void FASTCALL call_0_14_0_4(SH2_struct* sh)
{
  SH2movbs0(sh,14,0);
}
#define init_0_14_1_4
void FASTCALL call_0_14_1_4(SH2_struct* sh)
{
  SH2movbs0(sh,14,1);
}
#define init_0_14_2_4
void FASTCALL call_0_14_2_4(SH2_struct* sh)
{
  SH2movbs0(sh,14,2);
}
#define init_0_14_3_4
void FASTCALL call_0_14_3_4(SH2_struct* sh)
{
  SH2movbs0(sh,14,3);
}
#define init_0_14_4_4
void FASTCALL call_0_14_4_4(SH2_struct* sh)
{
  SH2movbs0(sh,14,4);
}
#define init_0_14_5_4
void FASTCALL call_0_14_5_4(SH2_struct* sh)
{
  SH2movbs0(sh,14,5);
}
#define init_0_14_6_4
void FASTCALL call_0_14_6_4(SH2_struct* sh)
{
  SH2movbs0(sh,14,6);
}
#define init_0_14_7_4
void FASTCALL call_0_14_7_4(SH2_struct* sh)
{
  SH2movbs0(sh,14,7);
}
#define init_0_14_8_4
void FASTCALL call_0_14_8_4(SH2_struct* sh)
{
  SH2movbs0(sh,14,8);
}
#define init_0_14_9_4
void FASTCALL call_0_14_9_4(SH2_struct* sh)
{
  SH2movbs0(sh,14,9);
}
#define init_0_14_10_4
void FASTCALL call_0_14_10_4(SH2_struct* sh)
{
  SH2movbs0(sh,14,10);
}
#define init_0_14_11_4
void FASTCALL call_0_14_11_4(SH2_struct* sh)
{
  SH2movbs0(sh,14,11);
}
#define init_0_14_12_4
void FASTCALL call_0_14_12_4(SH2_struct* sh)
{
  SH2movbs0(sh,14,12);
}
#define init_0_14_13_4
void FASTCALL call_0_14_13_4(SH2_struct* sh)
{
  SH2movbs0(sh,14,13);
}
#define init_0_14_14_4
void FASTCALL call_0_14_14_4(SH2_struct* sh)
{
  SH2movbs0(sh,14,14);
}
#define init_0_14_15_4
void FASTCALL call_0_14_15_4(SH2_struct* sh)
{
  SH2movbs0(sh,14,15);
}
#define init_0_15_0_4
void FASTCALL call_0_15_0_4(SH2_struct* sh)
{
  SH2movbs0(sh,15,0);
}
#define init_0_15_1_4
void FASTCALL call_0_15_1_4(SH2_struct* sh)
{
  SH2movbs0(sh,15,1);
}
#define init_0_15_2_4
void FASTCALL call_0_15_2_4(SH2_struct* sh)
{
  SH2movbs0(sh,15,2);
}
#define init_0_15_3_4
void FASTCALL call_0_15_3_4(SH2_struct* sh)
{
  SH2movbs0(sh,15,3);
}
#define init_0_15_4_4
void FASTCALL call_0_15_4_4(SH2_struct* sh)
{
  SH2movbs0(sh,15,4);
}
#define init_0_15_5_4
void FASTCALL call_0_15_5_4(SH2_struct* sh)
{
  SH2movbs0(sh,15,5);
}
#define init_0_15_6_4
void FASTCALL call_0_15_6_4(SH2_struct* sh)
{
  SH2movbs0(sh,15,6);
}
#define init_0_15_7_4
void FASTCALL call_0_15_7_4(SH2_struct* sh)
{
  SH2movbs0(sh,15,7);
}
#define init_0_15_8_4
void FASTCALL call_0_15_8_4(SH2_struct* sh)
{
  SH2movbs0(sh,15,8);
}
#define init_0_15_9_4
void FASTCALL call_0_15_9_4(SH2_struct* sh)
{
  SH2movbs0(sh,15,9);
}
#define init_0_15_10_4
void FASTCALL call_0_15_10_4(SH2_struct* sh)
{
  SH2movbs0(sh,15,10);
}
#define init_0_15_11_4
void FASTCALL call_0_15_11_4(SH2_struct* sh)
{
  SH2movbs0(sh,15,11);
}
#define init_0_15_12_4
void FASTCALL call_0_15_12_4(SH2_struct* sh)
{
  SH2movbs0(sh,15,12);
}
#define init_0_15_13_4
void FASTCALL call_0_15_13_4(SH2_struct* sh)
{
  SH2movbs0(sh,15,13);
}
#define init_0_15_14_4
void FASTCALL call_0_15_14_4(SH2_struct* sh)
{
  SH2movbs0(sh,15,14);
}
#define init_0_15_15_4
void FASTCALL call_0_15_15_4(SH2_struct* sh)
{
  SH2movbs0(sh,15,15);
}
#define init_0_0_0_12
void FASTCALL call_0_0_0_12(SH2_struct* sh)
{
  SH2movbl0(sh,0,0);
}
#define init_0_0_1_12
void FASTCALL call_0_0_1_12(SH2_struct* sh)
{
  SH2movbl0(sh,0,1);
}
#define init_0_0_2_12
void FASTCALL call_0_0_2_12(SH2_struct* sh)
{
  SH2movbl0(sh,0,2);
}
#define init_0_0_3_12
void FASTCALL call_0_0_3_12(SH2_struct* sh)
{
  SH2movbl0(sh,0,3);
}
#define init_0_0_4_12
void FASTCALL call_0_0_4_12(SH2_struct* sh)
{
  SH2movbl0(sh,0,4);
}
#define init_0_0_5_12
void FASTCALL call_0_0_5_12(SH2_struct* sh)
{
  SH2movbl0(sh,0,5);
}
#define init_0_0_6_12
void FASTCALL call_0_0_6_12(SH2_struct* sh)
{
  SH2movbl0(sh,0,6);
}
#define init_0_0_7_12
void FASTCALL call_0_0_7_12(SH2_struct* sh)
{
  SH2movbl0(sh,0,7);
}
#define init_0_0_8_12
void FASTCALL call_0_0_8_12(SH2_struct* sh)
{
  SH2movbl0(sh,0,8);
}
#define init_0_0_9_12
void FASTCALL call_0_0_9_12(SH2_struct* sh)
{
  SH2movbl0(sh,0,9);
}
#define init_0_0_10_12
void FASTCALL call_0_0_10_12(SH2_struct* sh)
{
  SH2movbl0(sh,0,10);
}
#define init_0_0_11_12
void FASTCALL call_0_0_11_12(SH2_struct* sh)
{
  SH2movbl0(sh,0,11);
}
#define init_0_0_12_12
void FASTCALL call_0_0_12_12(SH2_struct* sh)
{
  SH2movbl0(sh,0,12);
}
#define init_0_0_13_12
void FASTCALL call_0_0_13_12(SH2_struct* sh)
{
  SH2movbl0(sh,0,13);
}
#define init_0_0_14_12
void FASTCALL call_0_0_14_12(SH2_struct* sh)
{
  SH2movbl0(sh,0,14);
}
#define init_0_0_15_12
void FASTCALL call_0_0_15_12(SH2_struct* sh)
{
  SH2movbl0(sh,0,15);
}
#define init_0_1_0_12
void FASTCALL call_0_1_0_12(SH2_struct* sh)
{
  SH2movbl0(sh,1,0);
}
#define init_0_1_1_12
void FASTCALL call_0_1_1_12(SH2_struct* sh)
{
  SH2movbl0(sh,1,1);
}
#define init_0_1_2_12
void FASTCALL call_0_1_2_12(SH2_struct* sh)
{
  SH2movbl0(sh,1,2);
}
#define init_0_1_3_12
void FASTCALL call_0_1_3_12(SH2_struct* sh)
{
  SH2movbl0(sh,1,3);
}
#define init_0_1_4_12
void FASTCALL call_0_1_4_12(SH2_struct* sh)
{
  SH2movbl0(sh,1,4);
}
#define init_0_1_5_12
void FASTCALL call_0_1_5_12(SH2_struct* sh)
{
  SH2movbl0(sh,1,5);
}
#define init_0_1_6_12
void FASTCALL call_0_1_6_12(SH2_struct* sh)
{
  SH2movbl0(sh,1,6);
}
#define init_0_1_7_12
void FASTCALL call_0_1_7_12(SH2_struct* sh)
{
  SH2movbl0(sh,1,7);
}
#define init_0_1_8_12
void FASTCALL call_0_1_8_12(SH2_struct* sh)
{
  SH2movbl0(sh,1,8);
}
#define init_0_1_9_12
void FASTCALL call_0_1_9_12(SH2_struct* sh)
{
  SH2movbl0(sh,1,9);
}
#define init_0_1_10_12
void FASTCALL call_0_1_10_12(SH2_struct* sh)
{
  SH2movbl0(sh,1,10);
}
#define init_0_1_11_12
void FASTCALL call_0_1_11_12(SH2_struct* sh)
{
  SH2movbl0(sh,1,11);
}
#define init_0_1_12_12
void FASTCALL call_0_1_12_12(SH2_struct* sh)
{
  SH2movbl0(sh,1,12);
}
#define init_0_1_13_12
void FASTCALL call_0_1_13_12(SH2_struct* sh)
{
  SH2movbl0(sh,1,13);
}
#define init_0_1_14_12
void FASTCALL call_0_1_14_12(SH2_struct* sh)
{
  SH2movbl0(sh,1,14);
}
#define init_0_1_15_12
void FASTCALL call_0_1_15_12(SH2_struct* sh)
{
  SH2movbl0(sh,1,15);
}
#define init_0_2_0_12
void FASTCALL call_0_2_0_12(SH2_struct* sh)
{
  SH2movbl0(sh,2,0);
}
#define init_0_2_1_12
void FASTCALL call_0_2_1_12(SH2_struct* sh)
{
  SH2movbl0(sh,2,1);
}
#define init_0_2_2_12
void FASTCALL call_0_2_2_12(SH2_struct* sh)
{
  SH2movbl0(sh,2,2);
}
#define init_0_2_3_12
void FASTCALL call_0_2_3_12(SH2_struct* sh)
{
  SH2movbl0(sh,2,3);
}
#define init_0_2_4_12
void FASTCALL call_0_2_4_12(SH2_struct* sh)
{
  SH2movbl0(sh,2,4);
}
#define init_0_2_5_12
void FASTCALL call_0_2_5_12(SH2_struct* sh)
{
  SH2movbl0(sh,2,5);
}
#define init_0_2_6_12
void FASTCALL call_0_2_6_12(SH2_struct* sh)
{
  SH2movbl0(sh,2,6);
}
#define init_0_2_7_12
void FASTCALL call_0_2_7_12(SH2_struct* sh)
{
  SH2movbl0(sh,2,7);
}
#define init_0_2_8_12
void FASTCALL call_0_2_8_12(SH2_struct* sh)
{
  SH2movbl0(sh,2,8);
}
#define init_0_2_9_12
void FASTCALL call_0_2_9_12(SH2_struct* sh)
{
  SH2movbl0(sh,2,9);
}
#define init_0_2_10_12
void FASTCALL call_0_2_10_12(SH2_struct* sh)
{
  SH2movbl0(sh,2,10);
}
#define init_0_2_11_12
void FASTCALL call_0_2_11_12(SH2_struct* sh)
{
  SH2movbl0(sh,2,11);
}
#define init_0_2_12_12
void FASTCALL call_0_2_12_12(SH2_struct* sh)
{
  SH2movbl0(sh,2,12);
}
#define init_0_2_13_12
void FASTCALL call_0_2_13_12(SH2_struct* sh)
{
  SH2movbl0(sh,2,13);
}
#define init_0_2_14_12
void FASTCALL call_0_2_14_12(SH2_struct* sh)
{
  SH2movbl0(sh,2,14);
}
#define init_0_2_15_12
void FASTCALL call_0_2_15_12(SH2_struct* sh)
{
  SH2movbl0(sh,2,15);
}
#define init_0_3_0_12
void FASTCALL call_0_3_0_12(SH2_struct* sh)
{
  SH2movbl0(sh,3,0);
}
#define init_0_3_1_12
void FASTCALL call_0_3_1_12(SH2_struct* sh)
{
  SH2movbl0(sh,3,1);
}
#define init_0_3_2_12
void FASTCALL call_0_3_2_12(SH2_struct* sh)
{
  SH2movbl0(sh,3,2);
}
#define init_0_3_3_12
void FASTCALL call_0_3_3_12(SH2_struct* sh)
{
  SH2movbl0(sh,3,3);
}
#define init_0_3_4_12
void FASTCALL call_0_3_4_12(SH2_struct* sh)
{
  SH2movbl0(sh,3,4);
}
#define init_0_3_5_12
void FASTCALL call_0_3_5_12(SH2_struct* sh)
{
  SH2movbl0(sh,3,5);
}
#define init_0_3_6_12
void FASTCALL call_0_3_6_12(SH2_struct* sh)
{
  SH2movbl0(sh,3,6);
}
#define init_0_3_7_12
void FASTCALL call_0_3_7_12(SH2_struct* sh)
{
  SH2movbl0(sh,3,7);
}
#define init_0_3_8_12
void FASTCALL call_0_3_8_12(SH2_struct* sh)
{
  SH2movbl0(sh,3,8);
}
#define init_0_3_9_12
void FASTCALL call_0_3_9_12(SH2_struct* sh)
{
  SH2movbl0(sh,3,9);
}
#define init_0_3_10_12
void FASTCALL call_0_3_10_12(SH2_struct* sh)
{
  SH2movbl0(sh,3,10);
}
#define init_0_3_11_12
void FASTCALL call_0_3_11_12(SH2_struct* sh)
{
  SH2movbl0(sh,3,11);
}
#define init_0_3_12_12
void FASTCALL call_0_3_12_12(SH2_struct* sh)
{
  SH2movbl0(sh,3,12);
}
#define init_0_3_13_12
void FASTCALL call_0_3_13_12(SH2_struct* sh)
{
  SH2movbl0(sh,3,13);
}
#define init_0_3_14_12
void FASTCALL call_0_3_14_12(SH2_struct* sh)
{
  SH2movbl0(sh,3,14);
}
#define init_0_3_15_12
void FASTCALL call_0_3_15_12(SH2_struct* sh)
{
  SH2movbl0(sh,3,15);
}
#define init_0_4_0_12
void FASTCALL call_0_4_0_12(SH2_struct* sh)
{
  SH2movbl0(sh,4,0);
}
#define init_0_4_1_12
void FASTCALL call_0_4_1_12(SH2_struct* sh)
{
  SH2movbl0(sh,4,1);
}
#define init_0_4_2_12
void FASTCALL call_0_4_2_12(SH2_struct* sh)
{
  SH2movbl0(sh,4,2);
}
#define init_0_4_3_12
void FASTCALL call_0_4_3_12(SH2_struct* sh)
{
  SH2movbl0(sh,4,3);
}
#define init_0_4_4_12
void FASTCALL call_0_4_4_12(SH2_struct* sh)
{
  SH2movbl0(sh,4,4);
}
#define init_0_4_5_12
void FASTCALL call_0_4_5_12(SH2_struct* sh)
{
  SH2movbl0(sh,4,5);
}
#define init_0_4_6_12
void FASTCALL call_0_4_6_12(SH2_struct* sh)
{
  SH2movbl0(sh,4,6);
}
#define init_0_4_7_12
void FASTCALL call_0_4_7_12(SH2_struct* sh)
{
  SH2movbl0(sh,4,7);
}
#define init_0_4_8_12
void FASTCALL call_0_4_8_12(SH2_struct* sh)
{
  SH2movbl0(sh,4,8);
}
#define init_0_4_9_12
void FASTCALL call_0_4_9_12(SH2_struct* sh)
{
  SH2movbl0(sh,4,9);
}
#define init_0_4_10_12
void FASTCALL call_0_4_10_12(SH2_struct* sh)
{
  SH2movbl0(sh,4,10);
}
#define init_0_4_11_12
void FASTCALL call_0_4_11_12(SH2_struct* sh)
{
  SH2movbl0(sh,4,11);
}
#define init_0_4_12_12
void FASTCALL call_0_4_12_12(SH2_struct* sh)
{
  SH2movbl0(sh,4,12);
}
#define init_0_4_13_12
void FASTCALL call_0_4_13_12(SH2_struct* sh)
{
  SH2movbl0(sh,4,13);
}
#define init_0_4_14_12
void FASTCALL call_0_4_14_12(SH2_struct* sh)
{
  SH2movbl0(sh,4,14);
}
#define init_0_4_15_12
void FASTCALL call_0_4_15_12(SH2_struct* sh)
{
  SH2movbl0(sh,4,15);
}
#define init_0_5_0_12
void FASTCALL call_0_5_0_12(SH2_struct* sh)
{
  SH2movbl0(sh,5,0);
}
#define init_0_5_1_12
void FASTCALL call_0_5_1_12(SH2_struct* sh)
{
  SH2movbl0(sh,5,1);
}
#define init_0_5_2_12
void FASTCALL call_0_5_2_12(SH2_struct* sh)
{
  SH2movbl0(sh,5,2);
}
#define init_0_5_3_12
void FASTCALL call_0_5_3_12(SH2_struct* sh)
{
  SH2movbl0(sh,5,3);
}
#define init_0_5_4_12
void FASTCALL call_0_5_4_12(SH2_struct* sh)
{
  SH2movbl0(sh,5,4);
}
#define init_0_5_5_12
void FASTCALL call_0_5_5_12(SH2_struct* sh)
{
  SH2movbl0(sh,5,5);
}
#define init_0_5_6_12
void FASTCALL call_0_5_6_12(SH2_struct* sh)
{
  SH2movbl0(sh,5,6);
}
#define init_0_5_7_12
void FASTCALL call_0_5_7_12(SH2_struct* sh)
{
  SH2movbl0(sh,5,7);
}
#define init_0_5_8_12
void FASTCALL call_0_5_8_12(SH2_struct* sh)
{
  SH2movbl0(sh,5,8);
}
#define init_0_5_9_12
void FASTCALL call_0_5_9_12(SH2_struct* sh)
{
  SH2movbl0(sh,5,9);
}
#define init_0_5_10_12
void FASTCALL call_0_5_10_12(SH2_struct* sh)
{
  SH2movbl0(sh,5,10);
}
#define init_0_5_11_12
void FASTCALL call_0_5_11_12(SH2_struct* sh)
{
  SH2movbl0(sh,5,11);
}
#define init_0_5_12_12
void FASTCALL call_0_5_12_12(SH2_struct* sh)
{
  SH2movbl0(sh,5,12);
}
#define init_0_5_13_12
void FASTCALL call_0_5_13_12(SH2_struct* sh)
{
  SH2movbl0(sh,5,13);
}
#define init_0_5_14_12
void FASTCALL call_0_5_14_12(SH2_struct* sh)
{
  SH2movbl0(sh,5,14);
}
#define init_0_5_15_12
void FASTCALL call_0_5_15_12(SH2_struct* sh)
{
  SH2movbl0(sh,5,15);
}
#define init_0_6_0_12
void FASTCALL call_0_6_0_12(SH2_struct* sh)
{
  SH2movbl0(sh,6,0);
}
#define init_0_6_1_12
void FASTCALL call_0_6_1_12(SH2_struct* sh)
{
  SH2movbl0(sh,6,1);
}
#define init_0_6_2_12
void FASTCALL call_0_6_2_12(SH2_struct* sh)
{
  SH2movbl0(sh,6,2);
}
#define init_0_6_3_12
void FASTCALL call_0_6_3_12(SH2_struct* sh)
{
  SH2movbl0(sh,6,3);
}
#define init_0_6_4_12
void FASTCALL call_0_6_4_12(SH2_struct* sh)
{
  SH2movbl0(sh,6,4);
}
#define init_0_6_5_12
void FASTCALL call_0_6_5_12(SH2_struct* sh)
{
  SH2movbl0(sh,6,5);
}
#define init_0_6_6_12
void FASTCALL call_0_6_6_12(SH2_struct* sh)
{
  SH2movbl0(sh,6,6);
}
#define init_0_6_7_12
void FASTCALL call_0_6_7_12(SH2_struct* sh)
{
  SH2movbl0(sh,6,7);
}
#define init_0_6_8_12
void FASTCALL call_0_6_8_12(SH2_struct* sh)
{
  SH2movbl0(sh,6,8);
}
#define init_0_6_9_12
void FASTCALL call_0_6_9_12(SH2_struct* sh)
{
  SH2movbl0(sh,6,9);
}
#define init_0_6_10_12
void FASTCALL call_0_6_10_12(SH2_struct* sh)
{
  SH2movbl0(sh,6,10);
}
#define init_0_6_11_12
void FASTCALL call_0_6_11_12(SH2_struct* sh)
{
  SH2movbl0(sh,6,11);
}
#define init_0_6_12_12
void FASTCALL call_0_6_12_12(SH2_struct* sh)
{
  SH2movbl0(sh,6,12);
}
#define init_0_6_13_12
void FASTCALL call_0_6_13_12(SH2_struct* sh)
{
  SH2movbl0(sh,6,13);
}
#define init_0_6_14_12
void FASTCALL call_0_6_14_12(SH2_struct* sh)
{
  SH2movbl0(sh,6,14);
}
#define init_0_6_15_12
void FASTCALL call_0_6_15_12(SH2_struct* sh)
{
  SH2movbl0(sh,6,15);
}
#define init_0_7_0_12
void FASTCALL call_0_7_0_12(SH2_struct* sh)
{
  SH2movbl0(sh,7,0);
}
#define init_0_7_1_12
void FASTCALL call_0_7_1_12(SH2_struct* sh)
{
  SH2movbl0(sh,7,1);
}
#define init_0_7_2_12
void FASTCALL call_0_7_2_12(SH2_struct* sh)
{
  SH2movbl0(sh,7,2);
}
#define init_0_7_3_12
void FASTCALL call_0_7_3_12(SH2_struct* sh)
{
  SH2movbl0(sh,7,3);
}
#define init_0_7_4_12
void FASTCALL call_0_7_4_12(SH2_struct* sh)
{
  SH2movbl0(sh,7,4);
}
#define init_0_7_5_12
void FASTCALL call_0_7_5_12(SH2_struct* sh)
{
  SH2movbl0(sh,7,5);
}
#define init_0_7_6_12
void FASTCALL call_0_7_6_12(SH2_struct* sh)
{
  SH2movbl0(sh,7,6);
}
#define init_0_7_7_12
void FASTCALL call_0_7_7_12(SH2_struct* sh)
{
  SH2movbl0(sh,7,7);
}
#define init_0_7_8_12
void FASTCALL call_0_7_8_12(SH2_struct* sh)
{
  SH2movbl0(sh,7,8);
}
#define init_0_7_9_12
void FASTCALL call_0_7_9_12(SH2_struct* sh)
{
  SH2movbl0(sh,7,9);
}
#define init_0_7_10_12
void FASTCALL call_0_7_10_12(SH2_struct* sh)
{
  SH2movbl0(sh,7,10);
}
#define init_0_7_11_12
void FASTCALL call_0_7_11_12(SH2_struct* sh)
{
  SH2movbl0(sh,7,11);
}
#define init_0_7_12_12
void FASTCALL call_0_7_12_12(SH2_struct* sh)
{
  SH2movbl0(sh,7,12);
}
#define init_0_7_13_12
void FASTCALL call_0_7_13_12(SH2_struct* sh)
{
  SH2movbl0(sh,7,13);
}
#define init_0_7_14_12
void FASTCALL call_0_7_14_12(SH2_struct* sh)
{
  SH2movbl0(sh,7,14);
}
#define init_0_7_15_12
void FASTCALL call_0_7_15_12(SH2_struct* sh)
{
  SH2movbl0(sh,7,15);
}
#define init_0_8_0_12
void FASTCALL call_0_8_0_12(SH2_struct* sh)
{
  SH2movbl0(sh,8,0);
}
#define init_0_8_1_12
void FASTCALL call_0_8_1_12(SH2_struct* sh)
{
  SH2movbl0(sh,8,1);
}
#define init_0_8_2_12
void FASTCALL call_0_8_2_12(SH2_struct* sh)
{
  SH2movbl0(sh,8,2);
}
#define init_0_8_3_12
void FASTCALL call_0_8_3_12(SH2_struct* sh)
{
  SH2movbl0(sh,8,3);
}
#define init_0_8_4_12
void FASTCALL call_0_8_4_12(SH2_struct* sh)
{
  SH2movbl0(sh,8,4);
}
#define init_0_8_5_12
void FASTCALL call_0_8_5_12(SH2_struct* sh)
{
  SH2movbl0(sh,8,5);
}
#define init_0_8_6_12
void FASTCALL call_0_8_6_12(SH2_struct* sh)
{
  SH2movbl0(sh,8,6);
}
#define init_0_8_7_12
void FASTCALL call_0_8_7_12(SH2_struct* sh)
{
  SH2movbl0(sh,8,7);
}
#define init_0_8_8_12
void FASTCALL call_0_8_8_12(SH2_struct* sh)
{
  SH2movbl0(sh,8,8);
}
#define init_0_8_9_12
void FASTCALL call_0_8_9_12(SH2_struct* sh)
{
  SH2movbl0(sh,8,9);
}
#define init_0_8_10_12
void FASTCALL call_0_8_10_12(SH2_struct* sh)
{
  SH2movbl0(sh,8,10);
}
#define init_0_8_11_12
void FASTCALL call_0_8_11_12(SH2_struct* sh)
{
  SH2movbl0(sh,8,11);
}
#define init_0_8_12_12
void FASTCALL call_0_8_12_12(SH2_struct* sh)
{
  SH2movbl0(sh,8,12);
}
#define init_0_8_13_12
void FASTCALL call_0_8_13_12(SH2_struct* sh)
{
  SH2movbl0(sh,8,13);
}
#define init_0_8_14_12
void FASTCALL call_0_8_14_12(SH2_struct* sh)
{
  SH2movbl0(sh,8,14);
}
#define init_0_8_15_12
void FASTCALL call_0_8_15_12(SH2_struct* sh)
{
  SH2movbl0(sh,8,15);
}
#define init_0_9_0_12
void FASTCALL call_0_9_0_12(SH2_struct* sh)
{
  SH2movbl0(sh,9,0);
}
#define init_0_9_1_12
void FASTCALL call_0_9_1_12(SH2_struct* sh)
{
  SH2movbl0(sh,9,1);
}
#define init_0_9_2_12
void FASTCALL call_0_9_2_12(SH2_struct* sh)
{
  SH2movbl0(sh,9,2);
}
#define init_0_9_3_12
void FASTCALL call_0_9_3_12(SH2_struct* sh)
{
  SH2movbl0(sh,9,3);
}
#define init_0_9_4_12
void FASTCALL call_0_9_4_12(SH2_struct* sh)
{
  SH2movbl0(sh,9,4);
}
#define init_0_9_5_12
void FASTCALL call_0_9_5_12(SH2_struct* sh)
{
  SH2movbl0(sh,9,5);
}
#define init_0_9_6_12
void FASTCALL call_0_9_6_12(SH2_struct* sh)
{
  SH2movbl0(sh,9,6);
}
#define init_0_9_7_12
void FASTCALL call_0_9_7_12(SH2_struct* sh)
{
  SH2movbl0(sh,9,7);
}
#define init_0_9_8_12
void FASTCALL call_0_9_8_12(SH2_struct* sh)
{
  SH2movbl0(sh,9,8);
}
#define init_0_9_9_12
void FASTCALL call_0_9_9_12(SH2_struct* sh)
{
  SH2movbl0(sh,9,9);
}
#define init_0_9_10_12
void FASTCALL call_0_9_10_12(SH2_struct* sh)
{
  SH2movbl0(sh,9,10);
}
#define init_0_9_11_12
void FASTCALL call_0_9_11_12(SH2_struct* sh)
{
  SH2movbl0(sh,9,11);
}
#define init_0_9_12_12
void FASTCALL call_0_9_12_12(SH2_struct* sh)
{
  SH2movbl0(sh,9,12);
}
#define init_0_9_13_12
void FASTCALL call_0_9_13_12(SH2_struct* sh)
{
  SH2movbl0(sh,9,13);
}
#define init_0_9_14_12
void FASTCALL call_0_9_14_12(SH2_struct* sh)
{
  SH2movbl0(sh,9,14);
}
#define init_0_9_15_12
void FASTCALL call_0_9_15_12(SH2_struct* sh)
{
  SH2movbl0(sh,9,15);
}
#define init_0_10_0_12
void FASTCALL call_0_10_0_12(SH2_struct* sh)
{
  SH2movbl0(sh,10,0);
}
#define init_0_10_1_12
void FASTCALL call_0_10_1_12(SH2_struct* sh)
{
  SH2movbl0(sh,10,1);
}
#define init_0_10_2_12
void FASTCALL call_0_10_2_12(SH2_struct* sh)
{
  SH2movbl0(sh,10,2);
}
#define init_0_10_3_12
void FASTCALL call_0_10_3_12(SH2_struct* sh)
{
  SH2movbl0(sh,10,3);
}
#define init_0_10_4_12
void FASTCALL call_0_10_4_12(SH2_struct* sh)
{
  SH2movbl0(sh,10,4);
}
#define init_0_10_5_12
void FASTCALL call_0_10_5_12(SH2_struct* sh)
{
  SH2movbl0(sh,10,5);
}
#define init_0_10_6_12
void FASTCALL call_0_10_6_12(SH2_struct* sh)
{
  SH2movbl0(sh,10,6);
}
#define init_0_10_7_12
void FASTCALL call_0_10_7_12(SH2_struct* sh)
{
  SH2movbl0(sh,10,7);
}
#define init_0_10_8_12
void FASTCALL call_0_10_8_12(SH2_struct* sh)
{
  SH2movbl0(sh,10,8);
}
#define init_0_10_9_12
void FASTCALL call_0_10_9_12(SH2_struct* sh)
{
  SH2movbl0(sh,10,9);
}
#define init_0_10_10_12
void FASTCALL call_0_10_10_12(SH2_struct* sh)
{
  SH2movbl0(sh,10,10);
}
#define init_0_10_11_12
void FASTCALL call_0_10_11_12(SH2_struct* sh)
{
  SH2movbl0(sh,10,11);
}
#define init_0_10_12_12
void FASTCALL call_0_10_12_12(SH2_struct* sh)
{
  SH2movbl0(sh,10,12);
}
#define init_0_10_13_12
void FASTCALL call_0_10_13_12(SH2_struct* sh)
{
  SH2movbl0(sh,10,13);
}
#define init_0_10_14_12
void FASTCALL call_0_10_14_12(SH2_struct* sh)
{
  SH2movbl0(sh,10,14);
}
#define init_0_10_15_12
void FASTCALL call_0_10_15_12(SH2_struct* sh)
{
  SH2movbl0(sh,10,15);
}
#define init_0_11_0_12
void FASTCALL call_0_11_0_12(SH2_struct* sh)
{
  SH2movbl0(sh,11,0);
}
#define init_0_11_1_12
void FASTCALL call_0_11_1_12(SH2_struct* sh)
{
  SH2movbl0(sh,11,1);
}
#define init_0_11_2_12
void FASTCALL call_0_11_2_12(SH2_struct* sh)
{
  SH2movbl0(sh,11,2);
}
#define init_0_11_3_12
void FASTCALL call_0_11_3_12(SH2_struct* sh)
{
  SH2movbl0(sh,11,3);
}
#define init_0_11_4_12
void FASTCALL call_0_11_4_12(SH2_struct* sh)
{
  SH2movbl0(sh,11,4);
}
#define init_0_11_5_12
void FASTCALL call_0_11_5_12(SH2_struct* sh)
{
  SH2movbl0(sh,11,5);
}
#define init_0_11_6_12
void FASTCALL call_0_11_6_12(SH2_struct* sh)
{
  SH2movbl0(sh,11,6);
}
#define init_0_11_7_12
void FASTCALL call_0_11_7_12(SH2_struct* sh)
{
  SH2movbl0(sh,11,7);
}
#define init_0_11_8_12
void FASTCALL call_0_11_8_12(SH2_struct* sh)
{
  SH2movbl0(sh,11,8);
}
#define init_0_11_9_12
void FASTCALL call_0_11_9_12(SH2_struct* sh)
{
  SH2movbl0(sh,11,9);
}
#define init_0_11_10_12
void FASTCALL call_0_11_10_12(SH2_struct* sh)
{
  SH2movbl0(sh,11,10);
}
#define init_0_11_11_12
void FASTCALL call_0_11_11_12(SH2_struct* sh)
{
  SH2movbl0(sh,11,11);
}
#define init_0_11_12_12
void FASTCALL call_0_11_12_12(SH2_struct* sh)
{
  SH2movbl0(sh,11,12);
}
#define init_0_11_13_12
void FASTCALL call_0_11_13_12(SH2_struct* sh)
{
  SH2movbl0(sh,11,13);
}
#define init_0_11_14_12
void FASTCALL call_0_11_14_12(SH2_struct* sh)
{
  SH2movbl0(sh,11,14);
}
#define init_0_11_15_12
void FASTCALL call_0_11_15_12(SH2_struct* sh)
{
  SH2movbl0(sh,11,15);
}
#define init_0_12_0_12
void FASTCALL call_0_12_0_12(SH2_struct* sh)
{
  SH2movbl0(sh,12,0);
}
#define init_0_12_1_12
void FASTCALL call_0_12_1_12(SH2_struct* sh)
{
  SH2movbl0(sh,12,1);
}
#define init_0_12_2_12
void FASTCALL call_0_12_2_12(SH2_struct* sh)
{
  SH2movbl0(sh,12,2);
}
#define init_0_12_3_12
void FASTCALL call_0_12_3_12(SH2_struct* sh)
{
  SH2movbl0(sh,12,3);
}
#define init_0_12_4_12
void FASTCALL call_0_12_4_12(SH2_struct* sh)
{
  SH2movbl0(sh,12,4);
}
#define init_0_12_5_12
void FASTCALL call_0_12_5_12(SH2_struct* sh)
{
  SH2movbl0(sh,12,5);
}
#define init_0_12_6_12
void FASTCALL call_0_12_6_12(SH2_struct* sh)
{
  SH2movbl0(sh,12,6);
}
#define init_0_12_7_12
void FASTCALL call_0_12_7_12(SH2_struct* sh)
{
  SH2movbl0(sh,12,7);
}
#define init_0_12_8_12
void FASTCALL call_0_12_8_12(SH2_struct* sh)
{
  SH2movbl0(sh,12,8);
}
#define init_0_12_9_12
void FASTCALL call_0_12_9_12(SH2_struct* sh)
{
  SH2movbl0(sh,12,9);
}
#define init_0_12_10_12
void FASTCALL call_0_12_10_12(SH2_struct* sh)
{
  SH2movbl0(sh,12,10);
}
#define init_0_12_11_12
void FASTCALL call_0_12_11_12(SH2_struct* sh)
{
  SH2movbl0(sh,12,11);
}
#define init_0_12_12_12
void FASTCALL call_0_12_12_12(SH2_struct* sh)
{
  SH2movbl0(sh,12,12);
}
#define init_0_12_13_12
void FASTCALL call_0_12_13_12(SH2_struct* sh)
{
  SH2movbl0(sh,12,13);
}
#define init_0_12_14_12
void FASTCALL call_0_12_14_12(SH2_struct* sh)
{
  SH2movbl0(sh,12,14);
}
#define init_0_12_15_12
void FASTCALL call_0_12_15_12(SH2_struct* sh)
{
  SH2movbl0(sh,12,15);
}
#define init_0_13_0_12
void FASTCALL call_0_13_0_12(SH2_struct* sh)
{
  SH2movbl0(sh,13,0);
}
#define init_0_13_1_12
void FASTCALL call_0_13_1_12(SH2_struct* sh)
{
  SH2movbl0(sh,13,1);
}
#define init_0_13_2_12
void FASTCALL call_0_13_2_12(SH2_struct* sh)
{
  SH2movbl0(sh,13,2);
}
#define init_0_13_3_12
void FASTCALL call_0_13_3_12(SH2_struct* sh)
{
  SH2movbl0(sh,13,3);
}
#define init_0_13_4_12
void FASTCALL call_0_13_4_12(SH2_struct* sh)
{
  SH2movbl0(sh,13,4);
}
#define init_0_13_5_12
void FASTCALL call_0_13_5_12(SH2_struct* sh)
{
  SH2movbl0(sh,13,5);
}
#define init_0_13_6_12
void FASTCALL call_0_13_6_12(SH2_struct* sh)
{
  SH2movbl0(sh,13,6);
}
#define init_0_13_7_12
void FASTCALL call_0_13_7_12(SH2_struct* sh)
{
  SH2movbl0(sh,13,7);
}
#define init_0_13_8_12
void FASTCALL call_0_13_8_12(SH2_struct* sh)
{
  SH2movbl0(sh,13,8);
}
#define init_0_13_9_12
void FASTCALL call_0_13_9_12(SH2_struct* sh)
{
  SH2movbl0(sh,13,9);
}
#define init_0_13_10_12
void FASTCALL call_0_13_10_12(SH2_struct* sh)
{
  SH2movbl0(sh,13,10);
}
#define init_0_13_11_12
void FASTCALL call_0_13_11_12(SH2_struct* sh)
{
  SH2movbl0(sh,13,11);
}
#define init_0_13_12_12
void FASTCALL call_0_13_12_12(SH2_struct* sh)
{
  SH2movbl0(sh,13,12);
}
#define init_0_13_13_12
void FASTCALL call_0_13_13_12(SH2_struct* sh)
{
  SH2movbl0(sh,13,13);
}
#define init_0_13_14_12
void FASTCALL call_0_13_14_12(SH2_struct* sh)
{
  SH2movbl0(sh,13,14);
}
#define init_0_13_15_12
void FASTCALL call_0_13_15_12(SH2_struct* sh)
{
  SH2movbl0(sh,13,15);
}
#define init_0_14_0_12
void FASTCALL call_0_14_0_12(SH2_struct* sh)
{
  SH2movbl0(sh,14,0);
}
#define init_0_14_1_12
void FASTCALL call_0_14_1_12(SH2_struct* sh)
{
  SH2movbl0(sh,14,1);
}
#define init_0_14_2_12
void FASTCALL call_0_14_2_12(SH2_struct* sh)
{
  SH2movbl0(sh,14,2);
}
#define init_0_14_3_12
void FASTCALL call_0_14_3_12(SH2_struct* sh)
{
  SH2movbl0(sh,14,3);
}
#define init_0_14_4_12
void FASTCALL call_0_14_4_12(SH2_struct* sh)
{
  SH2movbl0(sh,14,4);
}
#define init_0_14_5_12
void FASTCALL call_0_14_5_12(SH2_struct* sh)
{
  SH2movbl0(sh,14,5);
}
#define init_0_14_6_12
void FASTCALL call_0_14_6_12(SH2_struct* sh)
{
  SH2movbl0(sh,14,6);
}
#define init_0_14_7_12
void FASTCALL call_0_14_7_12(SH2_struct* sh)
{
  SH2movbl0(sh,14,7);
}
#define init_0_14_8_12
void FASTCALL call_0_14_8_12(SH2_struct* sh)
{
  SH2movbl0(sh,14,8);
}
#define init_0_14_9_12
void FASTCALL call_0_14_9_12(SH2_struct* sh)
{
  SH2movbl0(sh,14,9);
}
#define init_0_14_10_12
void FASTCALL call_0_14_10_12(SH2_struct* sh)
{
  SH2movbl0(sh,14,10);
}
#define init_0_14_11_12
void FASTCALL call_0_14_11_12(SH2_struct* sh)
{
  SH2movbl0(sh,14,11);
}
#define init_0_14_12_12
void FASTCALL call_0_14_12_12(SH2_struct* sh)
{
  SH2movbl0(sh,14,12);
}
#define init_0_14_13_12
void FASTCALL call_0_14_13_12(SH2_struct* sh)
{
  SH2movbl0(sh,14,13);
}
#define init_0_14_14_12
void FASTCALL call_0_14_14_12(SH2_struct* sh)
{
  SH2movbl0(sh,14,14);
}
#define init_0_14_15_12
void FASTCALL call_0_14_15_12(SH2_struct* sh)
{
  SH2movbl0(sh,14,15);
}
#define init_0_15_0_12
void FASTCALL call_0_15_0_12(SH2_struct* sh)
{
  SH2movbl0(sh,15,0);
}
#define init_0_15_1_12
void FASTCALL call_0_15_1_12(SH2_struct* sh)
{
  SH2movbl0(sh,15,1);
}
#define init_0_15_2_12
void FASTCALL call_0_15_2_12(SH2_struct* sh)
{
  SH2movbl0(sh,15,2);
}
#define init_0_15_3_12
void FASTCALL call_0_15_3_12(SH2_struct* sh)
{
  SH2movbl0(sh,15,3);
}
#define init_0_15_4_12
void FASTCALL call_0_15_4_12(SH2_struct* sh)
{
  SH2movbl0(sh,15,4);
}
#define init_0_15_5_12
void FASTCALL call_0_15_5_12(SH2_struct* sh)
{
  SH2movbl0(sh,15,5);
}
#define init_0_15_6_12
void FASTCALL call_0_15_6_12(SH2_struct* sh)
{
  SH2movbl0(sh,15,6);
}
#define init_0_15_7_12
void FASTCALL call_0_15_7_12(SH2_struct* sh)
{
  SH2movbl0(sh,15,7);
}
#define init_0_15_8_12
void FASTCALL call_0_15_8_12(SH2_struct* sh)
{
  SH2movbl0(sh,15,8);
}
#define init_0_15_9_12
void FASTCALL call_0_15_9_12(SH2_struct* sh)
{
  SH2movbl0(sh,15,9);
}
#define init_0_15_10_12
void FASTCALL call_0_15_10_12(SH2_struct* sh)
{
  SH2movbl0(sh,15,10);
}
#define init_0_15_11_12
void FASTCALL call_0_15_11_12(SH2_struct* sh)
{
  SH2movbl0(sh,15,11);
}
#define init_0_15_12_12
void FASTCALL call_0_15_12_12(SH2_struct* sh)
{
  SH2movbl0(sh,15,12);
}
#define init_0_15_13_12
void FASTCALL call_0_15_13_12(SH2_struct* sh)
{
  SH2movbl0(sh,15,13);
}
#define init_0_15_14_12
void FASTCALL call_0_15_14_12(SH2_struct* sh)
{
  SH2movbl0(sh,15,14);
}
#define init_0_15_15_12
void FASTCALL call_0_15_15_12(SH2_struct* sh)
{
  SH2movbl0(sh,15,15);
}
#define init_0_0_0_15
void FASTCALL call_0_0_0_15(SH2_struct* sh)
{
  SH2macl(sh,0,0);
}
#define init_0_0_1_15
void FASTCALL call_0_0_1_15(SH2_struct* sh)
{
  SH2macl(sh,0,1);
}
#define init_0_0_2_15
void FASTCALL call_0_0_2_15(SH2_struct* sh)
{
  SH2macl(sh,0,2);
}
#define init_0_0_3_15
void FASTCALL call_0_0_3_15(SH2_struct* sh)
{
  SH2macl(sh,0,3);
}
#define init_0_0_4_15
void FASTCALL call_0_0_4_15(SH2_struct* sh)
{
  SH2macl(sh,0,4);
}
#define init_0_0_5_15
void FASTCALL call_0_0_5_15(SH2_struct* sh)
{
  SH2macl(sh,0,5);
}
#define init_0_0_6_15
void FASTCALL call_0_0_6_15(SH2_struct* sh)
{
  SH2macl(sh,0,6);
}
#define init_0_0_7_15
void FASTCALL call_0_0_7_15(SH2_struct* sh)
{
  SH2macl(sh,0,7);
}
#define init_0_0_8_15
void FASTCALL call_0_0_8_15(SH2_struct* sh)
{
  SH2macl(sh,0,8);
}
#define init_0_0_9_15
void FASTCALL call_0_0_9_15(SH2_struct* sh)
{
  SH2macl(sh,0,9);
}
#define init_0_0_10_15
void FASTCALL call_0_0_10_15(SH2_struct* sh)
{
  SH2macl(sh,0,10);
}
#define init_0_0_11_15
void FASTCALL call_0_0_11_15(SH2_struct* sh)
{
  SH2macl(sh,0,11);
}
#define init_0_0_12_15
void FASTCALL call_0_0_12_15(SH2_struct* sh)
{
  SH2macl(sh,0,12);
}
#define init_0_0_13_15
void FASTCALL call_0_0_13_15(SH2_struct* sh)
{
  SH2macl(sh,0,13);
}
#define init_0_0_14_15
void FASTCALL call_0_0_14_15(SH2_struct* sh)
{
  SH2macl(sh,0,14);
}
#define init_0_0_15_15
void FASTCALL call_0_0_15_15(SH2_struct* sh)
{
  SH2macl(sh,0,15);
}
#define init_0_1_0_15
void FASTCALL call_0_1_0_15(SH2_struct* sh)
{
  SH2macl(sh,1,0);
}
#define init_0_1_1_15
void FASTCALL call_0_1_1_15(SH2_struct* sh)
{
  SH2macl(sh,1,1);
}
#define init_0_1_2_15
void FASTCALL call_0_1_2_15(SH2_struct* sh)
{
  SH2macl(sh,1,2);
}
#define init_0_1_3_15
void FASTCALL call_0_1_3_15(SH2_struct* sh)
{
  SH2macl(sh,1,3);
}
#define init_0_1_4_15
void FASTCALL call_0_1_4_15(SH2_struct* sh)
{
  SH2macl(sh,1,4);
}
#define init_0_1_5_15
void FASTCALL call_0_1_5_15(SH2_struct* sh)
{
  SH2macl(sh,1,5);
}
#define init_0_1_6_15
void FASTCALL call_0_1_6_15(SH2_struct* sh)
{
  SH2macl(sh,1,6);
}
#define init_0_1_7_15
void FASTCALL call_0_1_7_15(SH2_struct* sh)
{
  SH2macl(sh,1,7);
}
#define init_0_1_8_15
void FASTCALL call_0_1_8_15(SH2_struct* sh)
{
  SH2macl(sh,1,8);
}
#define init_0_1_9_15
void FASTCALL call_0_1_9_15(SH2_struct* sh)
{
  SH2macl(sh,1,9);
}
#define init_0_1_10_15
void FASTCALL call_0_1_10_15(SH2_struct* sh)
{
  SH2macl(sh,1,10);
}
#define init_0_1_11_15
void FASTCALL call_0_1_11_15(SH2_struct* sh)
{
  SH2macl(sh,1,11);
}
#define init_0_1_12_15
void FASTCALL call_0_1_12_15(SH2_struct* sh)
{
  SH2macl(sh,1,12);
}
#define init_0_1_13_15
void FASTCALL call_0_1_13_15(SH2_struct* sh)
{
  SH2macl(sh,1,13);
}
#define init_0_1_14_15
void FASTCALL call_0_1_14_15(SH2_struct* sh)
{
  SH2macl(sh,1,14);
}
#define init_0_1_15_15
void FASTCALL call_0_1_15_15(SH2_struct* sh)
{
  SH2macl(sh,1,15);
}
#define init_0_2_0_15
void FASTCALL call_0_2_0_15(SH2_struct* sh)
{
  SH2macl(sh,2,0);
}
#define init_0_2_1_15
void FASTCALL call_0_2_1_15(SH2_struct* sh)
{
  SH2macl(sh,2,1);
}
#define init_0_2_2_15
void FASTCALL call_0_2_2_15(SH2_struct* sh)
{
  SH2macl(sh,2,2);
}
#define init_0_2_3_15
void FASTCALL call_0_2_3_15(SH2_struct* sh)
{
  SH2macl(sh,2,3);
}
#define init_0_2_4_15
void FASTCALL call_0_2_4_15(SH2_struct* sh)
{
  SH2macl(sh,2,4);
}
#define init_0_2_5_15
void FASTCALL call_0_2_5_15(SH2_struct* sh)
{
  SH2macl(sh,2,5);
}
#define init_0_2_6_15
void FASTCALL call_0_2_6_15(SH2_struct* sh)
{
  SH2macl(sh,2,6);
}
#define init_0_2_7_15
void FASTCALL call_0_2_7_15(SH2_struct* sh)
{
  SH2macl(sh,2,7);
}
#define init_0_2_8_15
void FASTCALL call_0_2_8_15(SH2_struct* sh)
{
  SH2macl(sh,2,8);
}
#define init_0_2_9_15
void FASTCALL call_0_2_9_15(SH2_struct* sh)
{
  SH2macl(sh,2,9);
}
#define init_0_2_10_15
void FASTCALL call_0_2_10_15(SH2_struct* sh)
{
  SH2macl(sh,2,10);
}
#define init_0_2_11_15
void FASTCALL call_0_2_11_15(SH2_struct* sh)
{
  SH2macl(sh,2,11);
}
#define init_0_2_12_15
void FASTCALL call_0_2_12_15(SH2_struct* sh)
{
  SH2macl(sh,2,12);
}
#define init_0_2_13_15
void FASTCALL call_0_2_13_15(SH2_struct* sh)
{
  SH2macl(sh,2,13);
}
#define init_0_2_14_15
void FASTCALL call_0_2_14_15(SH2_struct* sh)
{
  SH2macl(sh,2,14);
}
#define init_0_2_15_15
void FASTCALL call_0_2_15_15(SH2_struct* sh)
{
  SH2macl(sh,2,15);
}
#define init_0_3_0_15
void FASTCALL call_0_3_0_15(SH2_struct* sh)
{
  SH2macl(sh,3,0);
}
#define init_0_3_1_15
void FASTCALL call_0_3_1_15(SH2_struct* sh)
{
  SH2macl(sh,3,1);
}
#define init_0_3_2_15
void FASTCALL call_0_3_2_15(SH2_struct* sh)
{
  SH2macl(sh,3,2);
}
#define init_0_3_3_15
void FASTCALL call_0_3_3_15(SH2_struct* sh)
{
  SH2macl(sh,3,3);
}
#define init_0_3_4_15
void FASTCALL call_0_3_4_15(SH2_struct* sh)
{
  SH2macl(sh,3,4);
}
#define init_0_3_5_15
void FASTCALL call_0_3_5_15(SH2_struct* sh)
{
  SH2macl(sh,3,5);
}
#define init_0_3_6_15
void FASTCALL call_0_3_6_15(SH2_struct* sh)
{
  SH2macl(sh,3,6);
}
#define init_0_3_7_15
void FASTCALL call_0_3_7_15(SH2_struct* sh)
{
  SH2macl(sh,3,7);
}
#define init_0_3_8_15
void FASTCALL call_0_3_8_15(SH2_struct* sh)
{
  SH2macl(sh,3,8);
}
#define init_0_3_9_15
void FASTCALL call_0_3_9_15(SH2_struct* sh)
{
  SH2macl(sh,3,9);
}
#define init_0_3_10_15
void FASTCALL call_0_3_10_15(SH2_struct* sh)
{
  SH2macl(sh,3,10);
}
#define init_0_3_11_15
void FASTCALL call_0_3_11_15(SH2_struct* sh)
{
  SH2macl(sh,3,11);
}
#define init_0_3_12_15
void FASTCALL call_0_3_12_15(SH2_struct* sh)
{
  SH2macl(sh,3,12);
}
#define init_0_3_13_15
void FASTCALL call_0_3_13_15(SH2_struct* sh)
{
  SH2macl(sh,3,13);
}
#define init_0_3_14_15
void FASTCALL call_0_3_14_15(SH2_struct* sh)
{
  SH2macl(sh,3,14);
}
#define init_0_3_15_15
void FASTCALL call_0_3_15_15(SH2_struct* sh)
{
  SH2macl(sh,3,15);
}
#define init_0_4_0_15
void FASTCALL call_0_4_0_15(SH2_struct* sh)
{
  SH2macl(sh,4,0);
}
#define init_0_4_1_15
void FASTCALL call_0_4_1_15(SH2_struct* sh)
{
  SH2macl(sh,4,1);
}
#define init_0_4_2_15
void FASTCALL call_0_4_2_15(SH2_struct* sh)
{
  SH2macl(sh,4,2);
}
#define init_0_4_3_15
void FASTCALL call_0_4_3_15(SH2_struct* sh)
{
  SH2macl(sh,4,3);
}
#define init_0_4_4_15
void FASTCALL call_0_4_4_15(SH2_struct* sh)
{
  SH2macl(sh,4,4);
}
#define init_0_4_5_15
void FASTCALL call_0_4_5_15(SH2_struct* sh)
{
  SH2macl(sh,4,5);
}
#define init_0_4_6_15
void FASTCALL call_0_4_6_15(SH2_struct* sh)
{
  SH2macl(sh,4,6);
}
#define init_0_4_7_15
void FASTCALL call_0_4_7_15(SH2_struct* sh)
{
  SH2macl(sh,4,7);
}
#define init_0_4_8_15
void FASTCALL call_0_4_8_15(SH2_struct* sh)
{
  SH2macl(sh,4,8);
}
#define init_0_4_9_15
void FASTCALL call_0_4_9_15(SH2_struct* sh)
{
  SH2macl(sh,4,9);
}
#define init_0_4_10_15
void FASTCALL call_0_4_10_15(SH2_struct* sh)
{
  SH2macl(sh,4,10);
}
#define init_0_4_11_15
void FASTCALL call_0_4_11_15(SH2_struct* sh)
{
  SH2macl(sh,4,11);
}
#define init_0_4_12_15
void FASTCALL call_0_4_12_15(SH2_struct* sh)
{
  SH2macl(sh,4,12);
}
#define init_0_4_13_15
void FASTCALL call_0_4_13_15(SH2_struct* sh)
{
  SH2macl(sh,4,13);
}
#define init_0_4_14_15
void FASTCALL call_0_4_14_15(SH2_struct* sh)
{
  SH2macl(sh,4,14);
}
#define init_0_4_15_15
void FASTCALL call_0_4_15_15(SH2_struct* sh)
{
  SH2macl(sh,4,15);
}
#define init_0_5_0_15
void FASTCALL call_0_5_0_15(SH2_struct* sh)
{
  SH2macl(sh,5,0);
}
#define init_0_5_1_15
void FASTCALL call_0_5_1_15(SH2_struct* sh)
{
  SH2macl(sh,5,1);
}
#define init_0_5_2_15
void FASTCALL call_0_5_2_15(SH2_struct* sh)
{
  SH2macl(sh,5,2);
}
#define init_0_5_3_15
void FASTCALL call_0_5_3_15(SH2_struct* sh)
{
  SH2macl(sh,5,3);
}
#define init_0_5_4_15
void FASTCALL call_0_5_4_15(SH2_struct* sh)
{
  SH2macl(sh,5,4);
}
#define init_0_5_5_15
void FASTCALL call_0_5_5_15(SH2_struct* sh)
{
  SH2macl(sh,5,5);
}
#define init_0_5_6_15
void FASTCALL call_0_5_6_15(SH2_struct* sh)
{
  SH2macl(sh,5,6);
}
#define init_0_5_7_15
void FASTCALL call_0_5_7_15(SH2_struct* sh)
{
  SH2macl(sh,5,7);
}
#define init_0_5_8_15
void FASTCALL call_0_5_8_15(SH2_struct* sh)
{
  SH2macl(sh,5,8);
}
#define init_0_5_9_15
void FASTCALL call_0_5_9_15(SH2_struct* sh)
{
  SH2macl(sh,5,9);
}
#define init_0_5_10_15
void FASTCALL call_0_5_10_15(SH2_struct* sh)
{
  SH2macl(sh,5,10);
}
#define init_0_5_11_15
void FASTCALL call_0_5_11_15(SH2_struct* sh)
{
  SH2macl(sh,5,11);
}
#define init_0_5_12_15
void FASTCALL call_0_5_12_15(SH2_struct* sh)
{
  SH2macl(sh,5,12);
}
#define init_0_5_13_15
void FASTCALL call_0_5_13_15(SH2_struct* sh)
{
  SH2macl(sh,5,13);
}
#define init_0_5_14_15
void FASTCALL call_0_5_14_15(SH2_struct* sh)
{
  SH2macl(sh,5,14);
}
#define init_0_5_15_15
void FASTCALL call_0_5_15_15(SH2_struct* sh)
{
  SH2macl(sh,5,15);
}
#define init_0_6_0_15
void FASTCALL call_0_6_0_15(SH2_struct* sh)
{
  SH2macl(sh,6,0);
}
#define init_0_6_1_15
void FASTCALL call_0_6_1_15(SH2_struct* sh)
{
  SH2macl(sh,6,1);
}
#define init_0_6_2_15
void FASTCALL call_0_6_2_15(SH2_struct* sh)
{
  SH2macl(sh,6,2);
}
#define init_0_6_3_15
void FASTCALL call_0_6_3_15(SH2_struct* sh)
{
  SH2macl(sh,6,3);
}
#define init_0_6_4_15
void FASTCALL call_0_6_4_15(SH2_struct* sh)
{
  SH2macl(sh,6,4);
}
#define init_0_6_5_15
void FASTCALL call_0_6_5_15(SH2_struct* sh)
{
  SH2macl(sh,6,5);
}
#define init_0_6_6_15
void FASTCALL call_0_6_6_15(SH2_struct* sh)
{
  SH2macl(sh,6,6);
}
#define init_0_6_7_15
void FASTCALL call_0_6_7_15(SH2_struct* sh)
{
  SH2macl(sh,6,7);
}
#define init_0_6_8_15
void FASTCALL call_0_6_8_15(SH2_struct* sh)
{
  SH2macl(sh,6,8);
}
#define init_0_6_9_15
void FASTCALL call_0_6_9_15(SH2_struct* sh)
{
  SH2macl(sh,6,9);
}
#define init_0_6_10_15
void FASTCALL call_0_6_10_15(SH2_struct* sh)
{
  SH2macl(sh,6,10);
}
#define init_0_6_11_15
void FASTCALL call_0_6_11_15(SH2_struct* sh)
{
  SH2macl(sh,6,11);
}
#define init_0_6_12_15
void FASTCALL call_0_6_12_15(SH2_struct* sh)
{
  SH2macl(sh,6,12);
}
#define init_0_6_13_15
void FASTCALL call_0_6_13_15(SH2_struct* sh)
{
  SH2macl(sh,6,13);
}
#define init_0_6_14_15
void FASTCALL call_0_6_14_15(SH2_struct* sh)
{
  SH2macl(sh,6,14);
}
#define init_0_6_15_15
void FASTCALL call_0_6_15_15(SH2_struct* sh)
{
  SH2macl(sh,6,15);
}
#define init_0_7_0_15
void FASTCALL call_0_7_0_15(SH2_struct* sh)
{
  SH2macl(sh,7,0);
}
#define init_0_7_1_15
void FASTCALL call_0_7_1_15(SH2_struct* sh)
{
  SH2macl(sh,7,1);
}
#define init_0_7_2_15
void FASTCALL call_0_7_2_15(SH2_struct* sh)
{
  SH2macl(sh,7,2);
}
#define init_0_7_3_15
void FASTCALL call_0_7_3_15(SH2_struct* sh)
{
  SH2macl(sh,7,3);
}
#define init_0_7_4_15
void FASTCALL call_0_7_4_15(SH2_struct* sh)
{
  SH2macl(sh,7,4);
}
#define init_0_7_5_15
void FASTCALL call_0_7_5_15(SH2_struct* sh)
{
  SH2macl(sh,7,5);
}
#define init_0_7_6_15
void FASTCALL call_0_7_6_15(SH2_struct* sh)
{
  SH2macl(sh,7,6);
}
#define init_0_7_7_15
void FASTCALL call_0_7_7_15(SH2_struct* sh)
{
  SH2macl(sh,7,7);
}
#define init_0_7_8_15
void FASTCALL call_0_7_8_15(SH2_struct* sh)
{
  SH2macl(sh,7,8);
}
#define init_0_7_9_15
void FASTCALL call_0_7_9_15(SH2_struct* sh)
{
  SH2macl(sh,7,9);
}
#define init_0_7_10_15
void FASTCALL call_0_7_10_15(SH2_struct* sh)
{
  SH2macl(sh,7,10);
}
#define init_0_7_11_15
void FASTCALL call_0_7_11_15(SH2_struct* sh)
{
  SH2macl(sh,7,11);
}
#define init_0_7_12_15
void FASTCALL call_0_7_12_15(SH2_struct* sh)
{
  SH2macl(sh,7,12);
}
#define init_0_7_13_15
void FASTCALL call_0_7_13_15(SH2_struct* sh)
{
  SH2macl(sh,7,13);
}
#define init_0_7_14_15
void FASTCALL call_0_7_14_15(SH2_struct* sh)
{
  SH2macl(sh,7,14);
}
#define init_0_7_15_15
void FASTCALL call_0_7_15_15(SH2_struct* sh)
{
  SH2macl(sh,7,15);
}
#define init_0_8_0_15
void FASTCALL call_0_8_0_15(SH2_struct* sh)
{
  SH2macl(sh,8,0);
}
#define init_0_8_1_15
void FASTCALL call_0_8_1_15(SH2_struct* sh)
{
  SH2macl(sh,8,1);
}
#define init_0_8_2_15
void FASTCALL call_0_8_2_15(SH2_struct* sh)
{
  SH2macl(sh,8,2);
}
#define init_0_8_3_15
void FASTCALL call_0_8_3_15(SH2_struct* sh)
{
  SH2macl(sh,8,3);
}
#define init_0_8_4_15
void FASTCALL call_0_8_4_15(SH2_struct* sh)
{
  SH2macl(sh,8,4);
}
#define init_0_8_5_15
void FASTCALL call_0_8_5_15(SH2_struct* sh)
{
  SH2macl(sh,8,5);
}
#define init_0_8_6_15
void FASTCALL call_0_8_6_15(SH2_struct* sh)
{
  SH2macl(sh,8,6);
}
#define init_0_8_7_15
void FASTCALL call_0_8_7_15(SH2_struct* sh)
{
  SH2macl(sh,8,7);
}
#define init_0_8_8_15
void FASTCALL call_0_8_8_15(SH2_struct* sh)
{
  SH2macl(sh,8,8);
}
#define init_0_8_9_15
void FASTCALL call_0_8_9_15(SH2_struct* sh)
{
  SH2macl(sh,8,9);
}
#define init_0_8_10_15
void FASTCALL call_0_8_10_15(SH2_struct* sh)
{
  SH2macl(sh,8,10);
}
#define init_0_8_11_15
void FASTCALL call_0_8_11_15(SH2_struct* sh)
{
  SH2macl(sh,8,11);
}
#define init_0_8_12_15
void FASTCALL call_0_8_12_15(SH2_struct* sh)
{
  SH2macl(sh,8,12);
}
#define init_0_8_13_15
void FASTCALL call_0_8_13_15(SH2_struct* sh)
{
  SH2macl(sh,8,13);
}
#define init_0_8_14_15
void FASTCALL call_0_8_14_15(SH2_struct* sh)
{
  SH2macl(sh,8,14);
}
#define init_0_8_15_15
void FASTCALL call_0_8_15_15(SH2_struct* sh)
{
  SH2macl(sh,8,15);
}
#define init_0_9_0_15
void FASTCALL call_0_9_0_15(SH2_struct* sh)
{
  SH2macl(sh,9,0);
}
#define init_0_9_1_15
void FASTCALL call_0_9_1_15(SH2_struct* sh)
{
  SH2macl(sh,9,1);
}
#define init_0_9_2_15
void FASTCALL call_0_9_2_15(SH2_struct* sh)
{
  SH2macl(sh,9,2);
}
#define init_0_9_3_15
void FASTCALL call_0_9_3_15(SH2_struct* sh)
{
  SH2macl(sh,9,3);
}
#define init_0_9_4_15
void FASTCALL call_0_9_4_15(SH2_struct* sh)
{
  SH2macl(sh,9,4);
}
#define init_0_9_5_15
void FASTCALL call_0_9_5_15(SH2_struct* sh)
{
  SH2macl(sh,9,5);
}
#define init_0_9_6_15
void FASTCALL call_0_9_6_15(SH2_struct* sh)
{
  SH2macl(sh,9,6);
}
#define init_0_9_7_15
void FASTCALL call_0_9_7_15(SH2_struct* sh)
{
  SH2macl(sh,9,7);
}
#define init_0_9_8_15
void FASTCALL call_0_9_8_15(SH2_struct* sh)
{
  SH2macl(sh,9,8);
}
#define init_0_9_9_15
void FASTCALL call_0_9_9_15(SH2_struct* sh)
{
  SH2macl(sh,9,9);
}
#define init_0_9_10_15
void FASTCALL call_0_9_10_15(SH2_struct* sh)
{
  SH2macl(sh,9,10);
}
#define init_0_9_11_15
void FASTCALL call_0_9_11_15(SH2_struct* sh)
{
  SH2macl(sh,9,11);
}
#define init_0_9_12_15
void FASTCALL call_0_9_12_15(SH2_struct* sh)
{
  SH2macl(sh,9,12);
}
#define init_0_9_13_15
void FASTCALL call_0_9_13_15(SH2_struct* sh)
{
  SH2macl(sh,9,13);
}
#define init_0_9_14_15
void FASTCALL call_0_9_14_15(SH2_struct* sh)
{
  SH2macl(sh,9,14);
}
#define init_0_9_15_15
void FASTCALL call_0_9_15_15(SH2_struct* sh)
{
  SH2macl(sh,9,15);
}
#define init_0_10_0_15
void FASTCALL call_0_10_0_15(SH2_struct* sh)
{
  SH2macl(sh,10,0);
}
#define init_0_10_1_15
void FASTCALL call_0_10_1_15(SH2_struct* sh)
{
  SH2macl(sh,10,1);
}
#define init_0_10_2_15
void FASTCALL call_0_10_2_15(SH2_struct* sh)
{
  SH2macl(sh,10,2);
}
#define init_0_10_3_15
void FASTCALL call_0_10_3_15(SH2_struct* sh)
{
  SH2macl(sh,10,3);
}
#define init_0_10_4_15
void FASTCALL call_0_10_4_15(SH2_struct* sh)
{
  SH2macl(sh,10,4);
}
#define init_0_10_5_15
void FASTCALL call_0_10_5_15(SH2_struct* sh)
{
  SH2macl(sh,10,5);
}
#define init_0_10_6_15
void FASTCALL call_0_10_6_15(SH2_struct* sh)
{
  SH2macl(sh,10,6);
}
#define init_0_10_7_15
void FASTCALL call_0_10_7_15(SH2_struct* sh)
{
  SH2macl(sh,10,7);
}
#define init_0_10_8_15
void FASTCALL call_0_10_8_15(SH2_struct* sh)
{
  SH2macl(sh,10,8);
}
#define init_0_10_9_15
void FASTCALL call_0_10_9_15(SH2_struct* sh)
{
  SH2macl(sh,10,9);
}
#define init_0_10_10_15
void FASTCALL call_0_10_10_15(SH2_struct* sh)
{
  SH2macl(sh,10,10);
}
#define init_0_10_11_15
void FASTCALL call_0_10_11_15(SH2_struct* sh)
{
  SH2macl(sh,10,11);
}
#define init_0_10_12_15
void FASTCALL call_0_10_12_15(SH2_struct* sh)
{
  SH2macl(sh,10,12);
}
#define init_0_10_13_15
void FASTCALL call_0_10_13_15(SH2_struct* sh)
{
  SH2macl(sh,10,13);
}
#define init_0_10_14_15
void FASTCALL call_0_10_14_15(SH2_struct* sh)
{
  SH2macl(sh,10,14);
}
#define init_0_10_15_15
void FASTCALL call_0_10_15_15(SH2_struct* sh)
{
  SH2macl(sh,10,15);
}
#define init_0_11_0_15
void FASTCALL call_0_11_0_15(SH2_struct* sh)
{
  SH2macl(sh,11,0);
}
#define init_0_11_1_15
void FASTCALL call_0_11_1_15(SH2_struct* sh)
{
  SH2macl(sh,11,1);
}
#define init_0_11_2_15
void FASTCALL call_0_11_2_15(SH2_struct* sh)
{
  SH2macl(sh,11,2);
}
#define init_0_11_3_15
void FASTCALL call_0_11_3_15(SH2_struct* sh)
{
  SH2macl(sh,11,3);
}
#define init_0_11_4_15
void FASTCALL call_0_11_4_15(SH2_struct* sh)
{
  SH2macl(sh,11,4);
}
#define init_0_11_5_15
void FASTCALL call_0_11_5_15(SH2_struct* sh)
{
  SH2macl(sh,11,5);
}
#define init_0_11_6_15
void FASTCALL call_0_11_6_15(SH2_struct* sh)
{
  SH2macl(sh,11,6);
}
#define init_0_11_7_15
void FASTCALL call_0_11_7_15(SH2_struct* sh)
{
  SH2macl(sh,11,7);
}
#define init_0_11_8_15
void FASTCALL call_0_11_8_15(SH2_struct* sh)
{
  SH2macl(sh,11,8);
}
#define init_0_11_9_15
void FASTCALL call_0_11_9_15(SH2_struct* sh)
{
  SH2macl(sh,11,9);
}
#define init_0_11_10_15
void FASTCALL call_0_11_10_15(SH2_struct* sh)
{
  SH2macl(sh,11,10);
}
#define init_0_11_11_15
void FASTCALL call_0_11_11_15(SH2_struct* sh)
{
  SH2macl(sh,11,11);
}
#define init_0_11_12_15
void FASTCALL call_0_11_12_15(SH2_struct* sh)
{
  SH2macl(sh,11,12);
}
#define init_0_11_13_15
void FASTCALL call_0_11_13_15(SH2_struct* sh)
{
  SH2macl(sh,11,13);
}
#define init_0_11_14_15
void FASTCALL call_0_11_14_15(SH2_struct* sh)
{
  SH2macl(sh,11,14);
}
#define init_0_11_15_15
void FASTCALL call_0_11_15_15(SH2_struct* sh)
{
  SH2macl(sh,11,15);
}
#define init_0_12_0_15
void FASTCALL call_0_12_0_15(SH2_struct* sh)
{
  SH2macl(sh,12,0);
}
#define init_0_12_1_15
void FASTCALL call_0_12_1_15(SH2_struct* sh)
{
  SH2macl(sh,12,1);
}
#define init_0_12_2_15
void FASTCALL call_0_12_2_15(SH2_struct* sh)
{
  SH2macl(sh,12,2);
}
#define init_0_12_3_15
void FASTCALL call_0_12_3_15(SH2_struct* sh)
{
  SH2macl(sh,12,3);
}
#define init_0_12_4_15
void FASTCALL call_0_12_4_15(SH2_struct* sh)
{
  SH2macl(sh,12,4);
}
#define init_0_12_5_15
void FASTCALL call_0_12_5_15(SH2_struct* sh)
{
  SH2macl(sh,12,5);
}
#define init_0_12_6_15
void FASTCALL call_0_12_6_15(SH2_struct* sh)
{
  SH2macl(sh,12,6);
}
#define init_0_12_7_15
void FASTCALL call_0_12_7_15(SH2_struct* sh)
{
  SH2macl(sh,12,7);
}
#define init_0_12_8_15
void FASTCALL call_0_12_8_15(SH2_struct* sh)
{
  SH2macl(sh,12,8);
}
#define init_0_12_9_15
void FASTCALL call_0_12_9_15(SH2_struct* sh)
{
  SH2macl(sh,12,9);
}
#define init_0_12_10_15
void FASTCALL call_0_12_10_15(SH2_struct* sh)
{
  SH2macl(sh,12,10);
}
#define init_0_12_11_15
void FASTCALL call_0_12_11_15(SH2_struct* sh)
{
  SH2macl(sh,12,11);
}
#define init_0_12_12_15
void FASTCALL call_0_12_12_15(SH2_struct* sh)
{
  SH2macl(sh,12,12);
}
#define init_0_12_13_15
void FASTCALL call_0_12_13_15(SH2_struct* sh)
{
  SH2macl(sh,12,13);
}
#define init_0_12_14_15
void FASTCALL call_0_12_14_15(SH2_struct* sh)
{
  SH2macl(sh,12,14);
}
#define init_0_12_15_15
void FASTCALL call_0_12_15_15(SH2_struct* sh)
{
  SH2macl(sh,12,15);
}
#define init_0_13_0_15
void FASTCALL call_0_13_0_15(SH2_struct* sh)
{
  SH2macl(sh,13,0);
}
#define init_0_13_1_15
void FASTCALL call_0_13_1_15(SH2_struct* sh)
{
  SH2macl(sh,13,1);
}
#define init_0_13_2_15
void FASTCALL call_0_13_2_15(SH2_struct* sh)
{
  SH2macl(sh,13,2);
}
#define init_0_13_3_15
void FASTCALL call_0_13_3_15(SH2_struct* sh)
{
  SH2macl(sh,13,3);
}
#define init_0_13_4_15
void FASTCALL call_0_13_4_15(SH2_struct* sh)
{
  SH2macl(sh,13,4);
}
#define init_0_13_5_15
void FASTCALL call_0_13_5_15(SH2_struct* sh)
{
  SH2macl(sh,13,5);
}
#define init_0_13_6_15
void FASTCALL call_0_13_6_15(SH2_struct* sh)
{
  SH2macl(sh,13,6);
}
#define init_0_13_7_15
void FASTCALL call_0_13_7_15(SH2_struct* sh)
{
  SH2macl(sh,13,7);
}
#define init_0_13_8_15
void FASTCALL call_0_13_8_15(SH2_struct* sh)
{
  SH2macl(sh,13,8);
}
#define init_0_13_9_15
void FASTCALL call_0_13_9_15(SH2_struct* sh)
{
  SH2macl(sh,13,9);
}
#define init_0_13_10_15
void FASTCALL call_0_13_10_15(SH2_struct* sh)
{
  SH2macl(sh,13,10);
}
#define init_0_13_11_15
void FASTCALL call_0_13_11_15(SH2_struct* sh)
{
  SH2macl(sh,13,11);
}
#define init_0_13_12_15
void FASTCALL call_0_13_12_15(SH2_struct* sh)
{
  SH2macl(sh,13,12);
}
#define init_0_13_13_15
void FASTCALL call_0_13_13_15(SH2_struct* sh)
{
  SH2macl(sh,13,13);
}
#define init_0_13_14_15
void FASTCALL call_0_13_14_15(SH2_struct* sh)
{
  SH2macl(sh,13,14);
}
#define init_0_13_15_15
void FASTCALL call_0_13_15_15(SH2_struct* sh)
{
  SH2macl(sh,13,15);
}
#define init_0_14_0_15
void FASTCALL call_0_14_0_15(SH2_struct* sh)
{
  SH2macl(sh,14,0);
}
#define init_0_14_1_15
void FASTCALL call_0_14_1_15(SH2_struct* sh)
{
  SH2macl(sh,14,1);
}
#define init_0_14_2_15
void FASTCALL call_0_14_2_15(SH2_struct* sh)
{
  SH2macl(sh,14,2);
}
#define init_0_14_3_15
void FASTCALL call_0_14_3_15(SH2_struct* sh)
{
  SH2macl(sh,14,3);
}
#define init_0_14_4_15
void FASTCALL call_0_14_4_15(SH2_struct* sh)
{
  SH2macl(sh,14,4);
}
#define init_0_14_5_15
void FASTCALL call_0_14_5_15(SH2_struct* sh)
{
  SH2macl(sh,14,5);
}
#define init_0_14_6_15
void FASTCALL call_0_14_6_15(SH2_struct* sh)
{
  SH2macl(sh,14,6);
}
#define init_0_14_7_15
void FASTCALL call_0_14_7_15(SH2_struct* sh)
{
  SH2macl(sh,14,7);
}
#define init_0_14_8_15
void FASTCALL call_0_14_8_15(SH2_struct* sh)
{
  SH2macl(sh,14,8);
}
#define init_0_14_9_15
void FASTCALL call_0_14_9_15(SH2_struct* sh)
{
  SH2macl(sh,14,9);
}
#define init_0_14_10_15
void FASTCALL call_0_14_10_15(SH2_struct* sh)
{
  SH2macl(sh,14,10);
}
#define init_0_14_11_15
void FASTCALL call_0_14_11_15(SH2_struct* sh)
{
  SH2macl(sh,14,11);
}
#define init_0_14_12_15
void FASTCALL call_0_14_12_15(SH2_struct* sh)
{
  SH2macl(sh,14,12);
}
#define init_0_14_13_15
void FASTCALL call_0_14_13_15(SH2_struct* sh)
{
  SH2macl(sh,14,13);
}
#define init_0_14_14_15
void FASTCALL call_0_14_14_15(SH2_struct* sh)
{
  SH2macl(sh,14,14);
}
#define init_0_14_15_15
void FASTCALL call_0_14_15_15(SH2_struct* sh)
{
  SH2macl(sh,14,15);
}
#define init_0_15_0_15
void FASTCALL call_0_15_0_15(SH2_struct* sh)
{
  SH2macl(sh,15,0);
}
#define init_0_15_1_15
void FASTCALL call_0_15_1_15(SH2_struct* sh)
{
  SH2macl(sh,15,1);
}
#define init_0_15_2_15
void FASTCALL call_0_15_2_15(SH2_struct* sh)
{
  SH2macl(sh,15,2);
}
#define init_0_15_3_15
void FASTCALL call_0_15_3_15(SH2_struct* sh)
{
  SH2macl(sh,15,3);
}
#define init_0_15_4_15
void FASTCALL call_0_15_4_15(SH2_struct* sh)
{
  SH2macl(sh,15,4);
}
#define init_0_15_5_15
void FASTCALL call_0_15_5_15(SH2_struct* sh)
{
  SH2macl(sh,15,5);
}
#define init_0_15_6_15
void FASTCALL call_0_15_6_15(SH2_struct* sh)
{
  SH2macl(sh,15,6);
}
#define init_0_15_7_15
void FASTCALL call_0_15_7_15(SH2_struct* sh)
{
  SH2macl(sh,15,7);
}
#define init_0_15_8_15
void FASTCALL call_0_15_8_15(SH2_struct* sh)
{
  SH2macl(sh,15,8);
}
#define init_0_15_9_15
void FASTCALL call_0_15_9_15(SH2_struct* sh)
{
  SH2macl(sh,15,9);
}
#define init_0_15_10_15
void FASTCALL call_0_15_10_15(SH2_struct* sh)
{
  SH2macl(sh,15,10);
}
#define init_0_15_11_15
void FASTCALL call_0_15_11_15(SH2_struct* sh)
{
  SH2macl(sh,15,11);
}
#define init_0_15_12_15
void FASTCALL call_0_15_12_15(SH2_struct* sh)
{
  SH2macl(sh,15,12);
}
#define init_0_15_13_15
void FASTCALL call_0_15_13_15(SH2_struct* sh)
{
  SH2macl(sh,15,13);
}
#define init_0_15_14_15
void FASTCALL call_0_15_14_15(SH2_struct* sh)
{
  SH2macl(sh,15,14);
}
#define init_0_15_15_15
void FASTCALL call_0_15_15_15(SH2_struct* sh)
{
  SH2macl(sh,15,15);
}
#define init_2_0_0_8
void FASTCALL call_2_0_0_8(SH2_struct* sh)
{
  SH2tst(sh,0,0);
}
#define init_2_0_1_8
void FASTCALL call_2_0_1_8(SH2_struct* sh)
{
  SH2tst(sh,0,1);
}
#define init_2_0_2_8
void FASTCALL call_2_0_2_8(SH2_struct* sh)
{
  SH2tst(sh,0,2);
}
#define init_2_0_3_8
void FASTCALL call_2_0_3_8(SH2_struct* sh)
{
  SH2tst(sh,0,3);
}
#define init_2_0_4_8
void FASTCALL call_2_0_4_8(SH2_struct* sh)
{
  SH2tst(sh,0,4);
}
#define init_2_0_5_8
void FASTCALL call_2_0_5_8(SH2_struct* sh)
{
  SH2tst(sh,0,5);
}
#define init_2_0_6_8
void FASTCALL call_2_0_6_8(SH2_struct* sh)
{
  SH2tst(sh,0,6);
}
#define init_2_0_7_8
void FASTCALL call_2_0_7_8(SH2_struct* sh)
{
  SH2tst(sh,0,7);
}
#define init_2_0_8_8
void FASTCALL call_2_0_8_8(SH2_struct* sh)
{
  SH2tst(sh,0,8);
}
#define init_2_0_9_8
void FASTCALL call_2_0_9_8(SH2_struct* sh)
{
  SH2tst(sh,0,9);
}
#define init_2_0_10_8
void FASTCALL call_2_0_10_8(SH2_struct* sh)
{
  SH2tst(sh,0,10);
}
#define init_2_0_11_8
void FASTCALL call_2_0_11_8(SH2_struct* sh)
{
  SH2tst(sh,0,11);
}
#define init_2_0_12_8
void FASTCALL call_2_0_12_8(SH2_struct* sh)
{
  SH2tst(sh,0,12);
}
#define init_2_0_13_8
void FASTCALL call_2_0_13_8(SH2_struct* sh)
{
  SH2tst(sh,0,13);
}
#define init_2_0_14_8
void FASTCALL call_2_0_14_8(SH2_struct* sh)
{
  SH2tst(sh,0,14);
}
#define init_2_0_15_8
void FASTCALL call_2_0_15_8(SH2_struct* sh)
{
  SH2tst(sh,0,15);
}
#define init_2_1_0_8
void FASTCALL call_2_1_0_8(SH2_struct* sh)
{
  SH2tst(sh,1,0);
}
#define init_2_1_1_8
void FASTCALL call_2_1_1_8(SH2_struct* sh)
{
  SH2tst(sh,1,1);
}
#define init_2_1_2_8
void FASTCALL call_2_1_2_8(SH2_struct* sh)
{
  SH2tst(sh,1,2);
}
#define init_2_1_3_8
void FASTCALL call_2_1_3_8(SH2_struct* sh)
{
  SH2tst(sh,1,3);
}
#define init_2_1_4_8
void FASTCALL call_2_1_4_8(SH2_struct* sh)
{
  SH2tst(sh,1,4);
}
#define init_2_1_5_8
void FASTCALL call_2_1_5_8(SH2_struct* sh)
{
  SH2tst(sh,1,5);
}
#define init_2_1_6_8
void FASTCALL call_2_1_6_8(SH2_struct* sh)
{
  SH2tst(sh,1,6);
}
#define init_2_1_7_8
void FASTCALL call_2_1_7_8(SH2_struct* sh)
{
  SH2tst(sh,1,7);
}
#define init_2_1_8_8
void FASTCALL call_2_1_8_8(SH2_struct* sh)
{
  SH2tst(sh,1,8);
}
#define init_2_1_9_8
void FASTCALL call_2_1_9_8(SH2_struct* sh)
{
  SH2tst(sh,1,9);
}
#define init_2_1_10_8
void FASTCALL call_2_1_10_8(SH2_struct* sh)
{
  SH2tst(sh,1,10);
}
#define init_2_1_11_8
void FASTCALL call_2_1_11_8(SH2_struct* sh)
{
  SH2tst(sh,1,11);
}
#define init_2_1_12_8
void FASTCALL call_2_1_12_8(SH2_struct* sh)
{
  SH2tst(sh,1,12);
}
#define init_2_1_13_8
void FASTCALL call_2_1_13_8(SH2_struct* sh)
{
  SH2tst(sh,1,13);
}
#define init_2_1_14_8
void FASTCALL call_2_1_14_8(SH2_struct* sh)
{
  SH2tst(sh,1,14);
}
#define init_2_1_15_8
void FASTCALL call_2_1_15_8(SH2_struct* sh)
{
  SH2tst(sh,1,15);
}
#define init_2_2_0_8
void FASTCALL call_2_2_0_8(SH2_struct* sh)
{
  SH2tst(sh,2,0);
}
#define init_2_2_1_8
void FASTCALL call_2_2_1_8(SH2_struct* sh)
{
  SH2tst(sh,2,1);
}
#define init_2_2_2_8
void FASTCALL call_2_2_2_8(SH2_struct* sh)
{
  SH2tst(sh,2,2);
}
#define init_2_2_3_8
void FASTCALL call_2_2_3_8(SH2_struct* sh)
{
  SH2tst(sh,2,3);
}
#define init_2_2_4_8
void FASTCALL call_2_2_4_8(SH2_struct* sh)
{
  SH2tst(sh,2,4);
}
#define init_2_2_5_8
void FASTCALL call_2_2_5_8(SH2_struct* sh)
{
  SH2tst(sh,2,5);
}
#define init_2_2_6_8
void FASTCALL call_2_2_6_8(SH2_struct* sh)
{
  SH2tst(sh,2,6);
}
#define init_2_2_7_8
void FASTCALL call_2_2_7_8(SH2_struct* sh)
{
  SH2tst(sh,2,7);
}
#define init_2_2_8_8
void FASTCALL call_2_2_8_8(SH2_struct* sh)
{
  SH2tst(sh,2,8);
}
#define init_2_2_9_8
void FASTCALL call_2_2_9_8(SH2_struct* sh)
{
  SH2tst(sh,2,9);
}
#define init_2_2_10_8
void FASTCALL call_2_2_10_8(SH2_struct* sh)
{
  SH2tst(sh,2,10);
}
#define init_2_2_11_8
void FASTCALL call_2_2_11_8(SH2_struct* sh)
{
  SH2tst(sh,2,11);
}
#define init_2_2_12_8
void FASTCALL call_2_2_12_8(SH2_struct* sh)
{
  SH2tst(sh,2,12);
}
#define init_2_2_13_8
void FASTCALL call_2_2_13_8(SH2_struct* sh)
{
  SH2tst(sh,2,13);
}
#define init_2_2_14_8
void FASTCALL call_2_2_14_8(SH2_struct* sh)
{
  SH2tst(sh,2,14);
}
#define init_2_2_15_8
void FASTCALL call_2_2_15_8(SH2_struct* sh)
{
  SH2tst(sh,2,15);
}
#define init_2_3_0_8
void FASTCALL call_2_3_0_8(SH2_struct* sh)
{
  SH2tst(sh,3,0);
}
#define init_2_3_1_8
void FASTCALL call_2_3_1_8(SH2_struct* sh)
{
  SH2tst(sh,3,1);
}
#define init_2_3_2_8
void FASTCALL call_2_3_2_8(SH2_struct* sh)
{
  SH2tst(sh,3,2);
}
#define init_2_3_3_8
void FASTCALL call_2_3_3_8(SH2_struct* sh)
{
  SH2tst(sh,3,3);
}
#define init_2_3_4_8
void FASTCALL call_2_3_4_8(SH2_struct* sh)
{
  SH2tst(sh,3,4);
}
#define init_2_3_5_8
void FASTCALL call_2_3_5_8(SH2_struct* sh)
{
  SH2tst(sh,3,5);
}
#define init_2_3_6_8
void FASTCALL call_2_3_6_8(SH2_struct* sh)
{
  SH2tst(sh,3,6);
}
#define init_2_3_7_8
void FASTCALL call_2_3_7_8(SH2_struct* sh)
{
  SH2tst(sh,3,7);
}
#define init_2_3_8_8
void FASTCALL call_2_3_8_8(SH2_struct* sh)
{
  SH2tst(sh,3,8);
}
#define init_2_3_9_8
void FASTCALL call_2_3_9_8(SH2_struct* sh)
{
  SH2tst(sh,3,9);
}
#define init_2_3_10_8
void FASTCALL call_2_3_10_8(SH2_struct* sh)
{
  SH2tst(sh,3,10);
}
#define init_2_3_11_8
void FASTCALL call_2_3_11_8(SH2_struct* sh)
{
  SH2tst(sh,3,11);
}
#define init_2_3_12_8
void FASTCALL call_2_3_12_8(SH2_struct* sh)
{
  SH2tst(sh,3,12);
}
#define init_2_3_13_8
void FASTCALL call_2_3_13_8(SH2_struct* sh)
{
  SH2tst(sh,3,13);
}
#define init_2_3_14_8
void FASTCALL call_2_3_14_8(SH2_struct* sh)
{
  SH2tst(sh,3,14);
}
#define init_2_3_15_8
void FASTCALL call_2_3_15_8(SH2_struct* sh)
{
  SH2tst(sh,3,15);
}
#define init_2_4_0_8
void FASTCALL call_2_4_0_8(SH2_struct* sh)
{
  SH2tst(sh,4,0);
}
#define init_2_4_1_8
void FASTCALL call_2_4_1_8(SH2_struct* sh)
{
  SH2tst(sh,4,1);
}
#define init_2_4_2_8
void FASTCALL call_2_4_2_8(SH2_struct* sh)
{
  SH2tst(sh,4,2);
}
#define init_2_4_3_8
void FASTCALL call_2_4_3_8(SH2_struct* sh)
{
  SH2tst(sh,4,3);
}
#define init_2_4_4_8
void FASTCALL call_2_4_4_8(SH2_struct* sh)
{
  SH2tst(sh,4,4);
}
#define init_2_4_5_8
void FASTCALL call_2_4_5_8(SH2_struct* sh)
{
  SH2tst(sh,4,5);
}
#define init_2_4_6_8
void FASTCALL call_2_4_6_8(SH2_struct* sh)
{
  SH2tst(sh,4,6);
}
#define init_2_4_7_8
void FASTCALL call_2_4_7_8(SH2_struct* sh)
{
  SH2tst(sh,4,7);
}
#define init_2_4_8_8
void FASTCALL call_2_4_8_8(SH2_struct* sh)
{
  SH2tst(sh,4,8);
}
#define init_2_4_9_8
void FASTCALL call_2_4_9_8(SH2_struct* sh)
{
  SH2tst(sh,4,9);
}
#define init_2_4_10_8
void FASTCALL call_2_4_10_8(SH2_struct* sh)
{
  SH2tst(sh,4,10);
}
#define init_2_4_11_8
void FASTCALL call_2_4_11_8(SH2_struct* sh)
{
  SH2tst(sh,4,11);
}
#define init_2_4_12_8
void FASTCALL call_2_4_12_8(SH2_struct* sh)
{
  SH2tst(sh,4,12);
}
#define init_2_4_13_8
void FASTCALL call_2_4_13_8(SH2_struct* sh)
{
  SH2tst(sh,4,13);
}
#define init_2_4_14_8
void FASTCALL call_2_4_14_8(SH2_struct* sh)
{
  SH2tst(sh,4,14);
}
#define init_2_4_15_8
void FASTCALL call_2_4_15_8(SH2_struct* sh)
{
  SH2tst(sh,4,15);
}
#define init_2_5_0_8
void FASTCALL call_2_5_0_8(SH2_struct* sh)
{
  SH2tst(sh,5,0);
}
#define init_2_5_1_8
void FASTCALL call_2_5_1_8(SH2_struct* sh)
{
  SH2tst(sh,5,1);
}
#define init_2_5_2_8
void FASTCALL call_2_5_2_8(SH2_struct* sh)
{
  SH2tst(sh,5,2);
}
#define init_2_5_3_8
void FASTCALL call_2_5_3_8(SH2_struct* sh)
{
  SH2tst(sh,5,3);
}
#define init_2_5_4_8
void FASTCALL call_2_5_4_8(SH2_struct* sh)
{
  SH2tst(sh,5,4);
}
#define init_2_5_5_8
void FASTCALL call_2_5_5_8(SH2_struct* sh)
{
  SH2tst(sh,5,5);
}
#define init_2_5_6_8
void FASTCALL call_2_5_6_8(SH2_struct* sh)
{
  SH2tst(sh,5,6);
}
#define init_2_5_7_8
void FASTCALL call_2_5_7_8(SH2_struct* sh)
{
  SH2tst(sh,5,7);
}
#define init_2_5_8_8
void FASTCALL call_2_5_8_8(SH2_struct* sh)
{
  SH2tst(sh,5,8);
}
#define init_2_5_9_8
void FASTCALL call_2_5_9_8(SH2_struct* sh)
{
  SH2tst(sh,5,9);
}
#define init_2_5_10_8
void FASTCALL call_2_5_10_8(SH2_struct* sh)
{
  SH2tst(sh,5,10);
}
#define init_2_5_11_8
void FASTCALL call_2_5_11_8(SH2_struct* sh)
{
  SH2tst(sh,5,11);
}
#define init_2_5_12_8
void FASTCALL call_2_5_12_8(SH2_struct* sh)
{
  SH2tst(sh,5,12);
}
#define init_2_5_13_8
void FASTCALL call_2_5_13_8(SH2_struct* sh)
{
  SH2tst(sh,5,13);
}
#define init_2_5_14_8
void FASTCALL call_2_5_14_8(SH2_struct* sh)
{
  SH2tst(sh,5,14);
}
#define init_2_5_15_8
void FASTCALL call_2_5_15_8(SH2_struct* sh)
{
  SH2tst(sh,5,15);
}
#define init_2_6_0_8
void FASTCALL call_2_6_0_8(SH2_struct* sh)
{
  SH2tst(sh,6,0);
}
#define init_2_6_1_8
void FASTCALL call_2_6_1_8(SH2_struct* sh)
{
  SH2tst(sh,6,1);
}
#define init_2_6_2_8
void FASTCALL call_2_6_2_8(SH2_struct* sh)
{
  SH2tst(sh,6,2);
}
#define init_2_6_3_8
void FASTCALL call_2_6_3_8(SH2_struct* sh)
{
  SH2tst(sh,6,3);
}
#define init_2_6_4_8
void FASTCALL call_2_6_4_8(SH2_struct* sh)
{
  SH2tst(sh,6,4);
}
#define init_2_6_5_8
void FASTCALL call_2_6_5_8(SH2_struct* sh)
{
  SH2tst(sh,6,5);
}
#define init_2_6_6_8
void FASTCALL call_2_6_6_8(SH2_struct* sh)
{
  SH2tst(sh,6,6);
}
#define init_2_6_7_8
void FASTCALL call_2_6_7_8(SH2_struct* sh)
{
  SH2tst(sh,6,7);
}
#define init_2_6_8_8
void FASTCALL call_2_6_8_8(SH2_struct* sh)
{
  SH2tst(sh,6,8);
}
#define init_2_6_9_8
void FASTCALL call_2_6_9_8(SH2_struct* sh)
{
  SH2tst(sh,6,9);
}
#define init_2_6_10_8
void FASTCALL call_2_6_10_8(SH2_struct* sh)
{
  SH2tst(sh,6,10);
}
#define init_2_6_11_8
void FASTCALL call_2_6_11_8(SH2_struct* sh)
{
  SH2tst(sh,6,11);
}
#define init_2_6_12_8
void FASTCALL call_2_6_12_8(SH2_struct* sh)
{
  SH2tst(sh,6,12);
}
#define init_2_6_13_8
void FASTCALL call_2_6_13_8(SH2_struct* sh)
{
  SH2tst(sh,6,13);
}
#define init_2_6_14_8
void FASTCALL call_2_6_14_8(SH2_struct* sh)
{
  SH2tst(sh,6,14);
}
#define init_2_6_15_8
void FASTCALL call_2_6_15_8(SH2_struct* sh)
{
  SH2tst(sh,6,15);
}
#define init_2_7_0_8
void FASTCALL call_2_7_0_8(SH2_struct* sh)
{
  SH2tst(sh,7,0);
}
#define init_2_7_1_8
void FASTCALL call_2_7_1_8(SH2_struct* sh)
{
  SH2tst(sh,7,1);
}
#define init_2_7_2_8
void FASTCALL call_2_7_2_8(SH2_struct* sh)
{
  SH2tst(sh,7,2);
}
#define init_2_7_3_8
void FASTCALL call_2_7_3_8(SH2_struct* sh)
{
  SH2tst(sh,7,3);
}
#define init_2_7_4_8
void FASTCALL call_2_7_4_8(SH2_struct* sh)
{
  SH2tst(sh,7,4);
}
#define init_2_7_5_8
void FASTCALL call_2_7_5_8(SH2_struct* sh)
{
  SH2tst(sh,7,5);
}
#define init_2_7_6_8
void FASTCALL call_2_7_6_8(SH2_struct* sh)
{
  SH2tst(sh,7,6);
}
#define init_2_7_7_8
void FASTCALL call_2_7_7_8(SH2_struct* sh)
{
  SH2tst(sh,7,7);
}
#define init_2_7_8_8
void FASTCALL call_2_7_8_8(SH2_struct* sh)
{
  SH2tst(sh,7,8);
}
#define init_2_7_9_8
void FASTCALL call_2_7_9_8(SH2_struct* sh)
{
  SH2tst(sh,7,9);
}
#define init_2_7_10_8
void FASTCALL call_2_7_10_8(SH2_struct* sh)
{
  SH2tst(sh,7,10);
}
#define init_2_7_11_8
void FASTCALL call_2_7_11_8(SH2_struct* sh)
{
  SH2tst(sh,7,11);
}
#define init_2_7_12_8
void FASTCALL call_2_7_12_8(SH2_struct* sh)
{
  SH2tst(sh,7,12);
}
#define init_2_7_13_8
void FASTCALL call_2_7_13_8(SH2_struct* sh)
{
  SH2tst(sh,7,13);
}
#define init_2_7_14_8
void FASTCALL call_2_7_14_8(SH2_struct* sh)
{
  SH2tst(sh,7,14);
}
#define init_2_7_15_8
void FASTCALL call_2_7_15_8(SH2_struct* sh)
{
  SH2tst(sh,7,15);
}
#define init_2_8_0_8
void FASTCALL call_2_8_0_8(SH2_struct* sh)
{
  SH2tst(sh,8,0);
}
#define init_2_8_1_8
void FASTCALL call_2_8_1_8(SH2_struct* sh)
{
  SH2tst(sh,8,1);
}
#define init_2_8_2_8
void FASTCALL call_2_8_2_8(SH2_struct* sh)
{
  SH2tst(sh,8,2);
}
#define init_2_8_3_8
void FASTCALL call_2_8_3_8(SH2_struct* sh)
{
  SH2tst(sh,8,3);
}
#define init_2_8_4_8
void FASTCALL call_2_8_4_8(SH2_struct* sh)
{
  SH2tst(sh,8,4);
}
#define init_2_8_5_8
void FASTCALL call_2_8_5_8(SH2_struct* sh)
{
  SH2tst(sh,8,5);
}
#define init_2_8_6_8
void FASTCALL call_2_8_6_8(SH2_struct* sh)
{
  SH2tst(sh,8,6);
}
#define init_2_8_7_8
void FASTCALL call_2_8_7_8(SH2_struct* sh)
{
  SH2tst(sh,8,7);
}
#define init_2_8_8_8
void FASTCALL call_2_8_8_8(SH2_struct* sh)
{
  SH2tst(sh,8,8);
}
#define init_2_8_9_8
void FASTCALL call_2_8_9_8(SH2_struct* sh)
{
  SH2tst(sh,8,9);
}
#define init_2_8_10_8
void FASTCALL call_2_8_10_8(SH2_struct* sh)
{
  SH2tst(sh,8,10);
}
#define init_2_8_11_8
void FASTCALL call_2_8_11_8(SH2_struct* sh)
{
  SH2tst(sh,8,11);
}
#define init_2_8_12_8
void FASTCALL call_2_8_12_8(SH2_struct* sh)
{
  SH2tst(sh,8,12);
}
#define init_2_8_13_8
void FASTCALL call_2_8_13_8(SH2_struct* sh)
{
  SH2tst(sh,8,13);
}
#define init_2_8_14_8
void FASTCALL call_2_8_14_8(SH2_struct* sh)
{
  SH2tst(sh,8,14);
}
#define init_2_8_15_8
void FASTCALL call_2_8_15_8(SH2_struct* sh)
{
  SH2tst(sh,8,15);
}
#define init_2_9_0_8
void FASTCALL call_2_9_0_8(SH2_struct* sh)
{
  SH2tst(sh,9,0);
}
#define init_2_9_1_8
void FASTCALL call_2_9_1_8(SH2_struct* sh)
{
  SH2tst(sh,9,1);
}
#define init_2_9_2_8
void FASTCALL call_2_9_2_8(SH2_struct* sh)
{
  SH2tst(sh,9,2);
}
#define init_2_9_3_8
void FASTCALL call_2_9_3_8(SH2_struct* sh)
{
  SH2tst(sh,9,3);
}
#define init_2_9_4_8
void FASTCALL call_2_9_4_8(SH2_struct* sh)
{
  SH2tst(sh,9,4);
}
#define init_2_9_5_8
void FASTCALL call_2_9_5_8(SH2_struct* sh)
{
  SH2tst(sh,9,5);
}
#define init_2_9_6_8
void FASTCALL call_2_9_6_8(SH2_struct* sh)
{
  SH2tst(sh,9,6);
}
#define init_2_9_7_8
void FASTCALL call_2_9_7_8(SH2_struct* sh)
{
  SH2tst(sh,9,7);
}
#define init_2_9_8_8
void FASTCALL call_2_9_8_8(SH2_struct* sh)
{
  SH2tst(sh,9,8);
}
#define init_2_9_9_8
void FASTCALL call_2_9_9_8(SH2_struct* sh)
{
  SH2tst(sh,9,9);
}
#define init_2_9_10_8
void FASTCALL call_2_9_10_8(SH2_struct* sh)
{
  SH2tst(sh,9,10);
}
#define init_2_9_11_8
void FASTCALL call_2_9_11_8(SH2_struct* sh)
{
  SH2tst(sh,9,11);
}
#define init_2_9_12_8
void FASTCALL call_2_9_12_8(SH2_struct* sh)
{
  SH2tst(sh,9,12);
}
#define init_2_9_13_8
void FASTCALL call_2_9_13_8(SH2_struct* sh)
{
  SH2tst(sh,9,13);
}
#define init_2_9_14_8
void FASTCALL call_2_9_14_8(SH2_struct* sh)
{
  SH2tst(sh,9,14);
}
#define init_2_9_15_8
void FASTCALL call_2_9_15_8(SH2_struct* sh)
{
  SH2tst(sh,9,15);
}
#define init_2_10_0_8
void FASTCALL call_2_10_0_8(SH2_struct* sh)
{
  SH2tst(sh,10,0);
}
#define init_2_10_1_8
void FASTCALL call_2_10_1_8(SH2_struct* sh)
{
  SH2tst(sh,10,1);
}
#define init_2_10_2_8
void FASTCALL call_2_10_2_8(SH2_struct* sh)
{
  SH2tst(sh,10,2);
}
#define init_2_10_3_8
void FASTCALL call_2_10_3_8(SH2_struct* sh)
{
  SH2tst(sh,10,3);
}
#define init_2_10_4_8
void FASTCALL call_2_10_4_8(SH2_struct* sh)
{
  SH2tst(sh,10,4);
}
#define init_2_10_5_8
void FASTCALL call_2_10_5_8(SH2_struct* sh)
{
  SH2tst(sh,10,5);
}
#define init_2_10_6_8
void FASTCALL call_2_10_6_8(SH2_struct* sh)
{
  SH2tst(sh,10,6);
}
#define init_2_10_7_8
void FASTCALL call_2_10_7_8(SH2_struct* sh)
{
  SH2tst(sh,10,7);
}
#define init_2_10_8_8
void FASTCALL call_2_10_8_8(SH2_struct* sh)
{
  SH2tst(sh,10,8);
}
#define init_2_10_9_8
void FASTCALL call_2_10_9_8(SH2_struct* sh)
{
  SH2tst(sh,10,9);
}
#define init_2_10_10_8
void FASTCALL call_2_10_10_8(SH2_struct* sh)
{
  SH2tst(sh,10,10);
}
#define init_2_10_11_8
void FASTCALL call_2_10_11_8(SH2_struct* sh)
{
  SH2tst(sh,10,11);
}
#define init_2_10_12_8
void FASTCALL call_2_10_12_8(SH2_struct* sh)
{
  SH2tst(sh,10,12);
}
#define init_2_10_13_8
void FASTCALL call_2_10_13_8(SH2_struct* sh)
{
  SH2tst(sh,10,13);
}
#define init_2_10_14_8
void FASTCALL call_2_10_14_8(SH2_struct* sh)
{
  SH2tst(sh,10,14);
}
#define init_2_10_15_8
void FASTCALL call_2_10_15_8(SH2_struct* sh)
{
  SH2tst(sh,10,15);
}
#define init_2_11_0_8
void FASTCALL call_2_11_0_8(SH2_struct* sh)
{
  SH2tst(sh,11,0);
}
#define init_2_11_1_8
void FASTCALL call_2_11_1_8(SH2_struct* sh)
{
  SH2tst(sh,11,1);
}
#define init_2_11_2_8
void FASTCALL call_2_11_2_8(SH2_struct* sh)
{
  SH2tst(sh,11,2);
}
#define init_2_11_3_8
void FASTCALL call_2_11_3_8(SH2_struct* sh)
{
  SH2tst(sh,11,3);
}
#define init_2_11_4_8
void FASTCALL call_2_11_4_8(SH2_struct* sh)
{
  SH2tst(sh,11,4);
}
#define init_2_11_5_8
void FASTCALL call_2_11_5_8(SH2_struct* sh)
{
  SH2tst(sh,11,5);
}
#define init_2_11_6_8
void FASTCALL call_2_11_6_8(SH2_struct* sh)
{
  SH2tst(sh,11,6);
}
#define init_2_11_7_8
void FASTCALL call_2_11_7_8(SH2_struct* sh)
{
  SH2tst(sh,11,7);
}
#define init_2_11_8_8
void FASTCALL call_2_11_8_8(SH2_struct* sh)
{
  SH2tst(sh,11,8);
}
#define init_2_11_9_8
void FASTCALL call_2_11_9_8(SH2_struct* sh)
{
  SH2tst(sh,11,9);
}
#define init_2_11_10_8
void FASTCALL call_2_11_10_8(SH2_struct* sh)
{
  SH2tst(sh,11,10);
}
#define init_2_11_11_8
void FASTCALL call_2_11_11_8(SH2_struct* sh)
{
  SH2tst(sh,11,11);
}
#define init_2_11_12_8
void FASTCALL call_2_11_12_8(SH2_struct* sh)
{
  SH2tst(sh,11,12);
}
#define init_2_11_13_8
void FASTCALL call_2_11_13_8(SH2_struct* sh)
{
  SH2tst(sh,11,13);
}
#define init_2_11_14_8
void FASTCALL call_2_11_14_8(SH2_struct* sh)
{
  SH2tst(sh,11,14);
}
#define init_2_11_15_8
void FASTCALL call_2_11_15_8(SH2_struct* sh)
{
  SH2tst(sh,11,15);
}
#define init_2_12_0_8
void FASTCALL call_2_12_0_8(SH2_struct* sh)
{
  SH2tst(sh,12,0);
}
#define init_2_12_1_8
void FASTCALL call_2_12_1_8(SH2_struct* sh)
{
  SH2tst(sh,12,1);
}
#define init_2_12_2_8
void FASTCALL call_2_12_2_8(SH2_struct* sh)
{
  SH2tst(sh,12,2);
}
#define init_2_12_3_8
void FASTCALL call_2_12_3_8(SH2_struct* sh)
{
  SH2tst(sh,12,3);
}
#define init_2_12_4_8
void FASTCALL call_2_12_4_8(SH2_struct* sh)
{
  SH2tst(sh,12,4);
}
#define init_2_12_5_8
void FASTCALL call_2_12_5_8(SH2_struct* sh)
{
  SH2tst(sh,12,5);
}
#define init_2_12_6_8
void FASTCALL call_2_12_6_8(SH2_struct* sh)
{
  SH2tst(sh,12,6);
}
#define init_2_12_7_8
void FASTCALL call_2_12_7_8(SH2_struct* sh)
{
  SH2tst(sh,12,7);
}
#define init_2_12_8_8
void FASTCALL call_2_12_8_8(SH2_struct* sh)
{
  SH2tst(sh,12,8);
}
#define init_2_12_9_8
void FASTCALL call_2_12_9_8(SH2_struct* sh)
{
  SH2tst(sh,12,9);
}
#define init_2_12_10_8
void FASTCALL call_2_12_10_8(SH2_struct* sh)
{
  SH2tst(sh,12,10);
}
#define init_2_12_11_8
void FASTCALL call_2_12_11_8(SH2_struct* sh)
{
  SH2tst(sh,12,11);
}
#define init_2_12_12_8
void FASTCALL call_2_12_12_8(SH2_struct* sh)
{
  SH2tst(sh,12,12);
}
#define init_2_12_13_8
void FASTCALL call_2_12_13_8(SH2_struct* sh)
{
  SH2tst(sh,12,13);
}
#define init_2_12_14_8
void FASTCALL call_2_12_14_8(SH2_struct* sh)
{
  SH2tst(sh,12,14);
}
#define init_2_12_15_8
void FASTCALL call_2_12_15_8(SH2_struct* sh)
{
  SH2tst(sh,12,15);
}
#define init_2_13_0_8
void FASTCALL call_2_13_0_8(SH2_struct* sh)
{
  SH2tst(sh,13,0);
}
#define init_2_13_1_8
void FASTCALL call_2_13_1_8(SH2_struct* sh)
{
  SH2tst(sh,13,1);
}
#define init_2_13_2_8
void FASTCALL call_2_13_2_8(SH2_struct* sh)
{
  SH2tst(sh,13,2);
}
#define init_2_13_3_8
void FASTCALL call_2_13_3_8(SH2_struct* sh)
{
  SH2tst(sh,13,3);
}
#define init_2_13_4_8
void FASTCALL call_2_13_4_8(SH2_struct* sh)
{
  SH2tst(sh,13,4);
}
#define init_2_13_5_8
void FASTCALL call_2_13_5_8(SH2_struct* sh)
{
  SH2tst(sh,13,5);
}
#define init_2_13_6_8
void FASTCALL call_2_13_6_8(SH2_struct* sh)
{
  SH2tst(sh,13,6);
}
#define init_2_13_7_8
void FASTCALL call_2_13_7_8(SH2_struct* sh)
{
  SH2tst(sh,13,7);
}
#define init_2_13_8_8
void FASTCALL call_2_13_8_8(SH2_struct* sh)
{
  SH2tst(sh,13,8);
}
#define init_2_13_9_8
void FASTCALL call_2_13_9_8(SH2_struct* sh)
{
  SH2tst(sh,13,9);
}
#define init_2_13_10_8
void FASTCALL call_2_13_10_8(SH2_struct* sh)
{
  SH2tst(sh,13,10);
}
#define init_2_13_11_8
void FASTCALL call_2_13_11_8(SH2_struct* sh)
{
  SH2tst(sh,13,11);
}
#define init_2_13_12_8
void FASTCALL call_2_13_12_8(SH2_struct* sh)
{
  SH2tst(sh,13,12);
}
#define init_2_13_13_8
void FASTCALL call_2_13_13_8(SH2_struct* sh)
{
  SH2tst(sh,13,13);
}
#define init_2_13_14_8
void FASTCALL call_2_13_14_8(SH2_struct* sh)
{
  SH2tst(sh,13,14);
}
#define init_2_13_15_8
void FASTCALL call_2_13_15_8(SH2_struct* sh)
{
  SH2tst(sh,13,15);
}
#define init_2_14_0_8
void FASTCALL call_2_14_0_8(SH2_struct* sh)
{
  SH2tst(sh,14,0);
}
#define init_2_14_1_8
void FASTCALL call_2_14_1_8(SH2_struct* sh)
{
  SH2tst(sh,14,1);
}
#define init_2_14_2_8
void FASTCALL call_2_14_2_8(SH2_struct* sh)
{
  SH2tst(sh,14,2);
}
#define init_2_14_3_8
void FASTCALL call_2_14_3_8(SH2_struct* sh)
{
  SH2tst(sh,14,3);
}
#define init_2_14_4_8
void FASTCALL call_2_14_4_8(SH2_struct* sh)
{
  SH2tst(sh,14,4);
}
#define init_2_14_5_8
void FASTCALL call_2_14_5_8(SH2_struct* sh)
{
  SH2tst(sh,14,5);
}
#define init_2_14_6_8
void FASTCALL call_2_14_6_8(SH2_struct* sh)
{
  SH2tst(sh,14,6);
}
#define init_2_14_7_8
void FASTCALL call_2_14_7_8(SH2_struct* sh)
{
  SH2tst(sh,14,7);
}
#define init_2_14_8_8
void FASTCALL call_2_14_8_8(SH2_struct* sh)
{
  SH2tst(sh,14,8);
}
#define init_2_14_9_8
void FASTCALL call_2_14_9_8(SH2_struct* sh)
{
  SH2tst(sh,14,9);
}
#define init_2_14_10_8
void FASTCALL call_2_14_10_8(SH2_struct* sh)
{
  SH2tst(sh,14,10);
}
#define init_2_14_11_8
void FASTCALL call_2_14_11_8(SH2_struct* sh)
{
  SH2tst(sh,14,11);
}
#define init_2_14_12_8
void FASTCALL call_2_14_12_8(SH2_struct* sh)
{
  SH2tst(sh,14,12);
}
#define init_2_14_13_8
void FASTCALL call_2_14_13_8(SH2_struct* sh)
{
  SH2tst(sh,14,13);
}
#define init_2_14_14_8
void FASTCALL call_2_14_14_8(SH2_struct* sh)
{
  SH2tst(sh,14,14);
}
#define init_2_14_15_8
void FASTCALL call_2_14_15_8(SH2_struct* sh)
{
  SH2tst(sh,14,15);
}
#define init_2_15_0_8
void FASTCALL call_2_15_0_8(SH2_struct* sh)
{
  SH2tst(sh,15,0);
}
#define init_2_15_1_8
void FASTCALL call_2_15_1_8(SH2_struct* sh)
{
  SH2tst(sh,15,1);
}
#define init_2_15_2_8
void FASTCALL call_2_15_2_8(SH2_struct* sh)
{
  SH2tst(sh,15,2);
}
#define init_2_15_3_8
void FASTCALL call_2_15_3_8(SH2_struct* sh)
{
  SH2tst(sh,15,3);
}
#define init_2_15_4_8
void FASTCALL call_2_15_4_8(SH2_struct* sh)
{
  SH2tst(sh,15,4);
}
#define init_2_15_5_8
void FASTCALL call_2_15_5_8(SH2_struct* sh)
{
  SH2tst(sh,15,5);
}
#define init_2_15_6_8
void FASTCALL call_2_15_6_8(SH2_struct* sh)
{
  SH2tst(sh,15,6);
}
#define init_2_15_7_8
void FASTCALL call_2_15_7_8(SH2_struct* sh)
{
  SH2tst(sh,15,7);
}
#define init_2_15_8_8
void FASTCALL call_2_15_8_8(SH2_struct* sh)
{
  SH2tst(sh,15,8);
}
#define init_2_15_9_8
void FASTCALL call_2_15_9_8(SH2_struct* sh)
{
  SH2tst(sh,15,9);
}
#define init_2_15_10_8
void FASTCALL call_2_15_10_8(SH2_struct* sh)
{
  SH2tst(sh,15,10);
}
#define init_2_15_11_8
void FASTCALL call_2_15_11_8(SH2_struct* sh)
{
  SH2tst(sh,15,11);
}
#define init_2_15_12_8
void FASTCALL call_2_15_12_8(SH2_struct* sh)
{
  SH2tst(sh,15,12);
}
#define init_2_15_13_8
void FASTCALL call_2_15_13_8(SH2_struct* sh)
{
  SH2tst(sh,15,13);
}
#define init_2_15_14_8
void FASTCALL call_2_15_14_8(SH2_struct* sh)
{
  SH2tst(sh,15,14);
}
#define init_2_15_15_8
void FASTCALL call_2_15_15_8(SH2_struct* sh)
{
  SH2tst(sh,15,15);
}
#define init_2_0_0_10
void FASTCALL call_2_0_0_10(SH2_struct* sh)
{
  SH2xor(sh,0,0);
}
#define init_2_0_1_10
void FASTCALL call_2_0_1_10(SH2_struct* sh)
{
  SH2xor(sh,0,1);
}
#define init_2_0_2_10
void FASTCALL call_2_0_2_10(SH2_struct* sh)
{
  SH2xor(sh,0,2);
}
#define init_2_0_3_10
void FASTCALL call_2_0_3_10(SH2_struct* sh)
{
  SH2xor(sh,0,3);
}
#define init_2_0_4_10
void FASTCALL call_2_0_4_10(SH2_struct* sh)
{
  SH2xor(sh,0,4);
}
#define init_2_0_5_10
void FASTCALL call_2_0_5_10(SH2_struct* sh)
{
  SH2xor(sh,0,5);
}
#define init_2_0_6_10
void FASTCALL call_2_0_6_10(SH2_struct* sh)
{
  SH2xor(sh,0,6);
}
#define init_2_0_7_10
void FASTCALL call_2_0_7_10(SH2_struct* sh)
{
  SH2xor(sh,0,7);
}
#define init_2_0_8_10
void FASTCALL call_2_0_8_10(SH2_struct* sh)
{
  SH2xor(sh,0,8);
}
#define init_2_0_9_10
void FASTCALL call_2_0_9_10(SH2_struct* sh)
{
  SH2xor(sh,0,9);
}
#define init_2_0_10_10
void FASTCALL call_2_0_10_10(SH2_struct* sh)
{
  SH2xor(sh,0,10);
}
#define init_2_0_11_10
void FASTCALL call_2_0_11_10(SH2_struct* sh)
{
  SH2xor(sh,0,11);
}
#define init_2_0_12_10
void FASTCALL call_2_0_12_10(SH2_struct* sh)
{
  SH2xor(sh,0,12);
}
#define init_2_0_13_10
void FASTCALL call_2_0_13_10(SH2_struct* sh)
{
  SH2xor(sh,0,13);
}
#define init_2_0_14_10
void FASTCALL call_2_0_14_10(SH2_struct* sh)
{
  SH2xor(sh,0,14);
}
#define init_2_0_15_10
void FASTCALL call_2_0_15_10(SH2_struct* sh)
{
  SH2xor(sh,0,15);
}
#define init_2_1_0_10
void FASTCALL call_2_1_0_10(SH2_struct* sh)
{
  SH2xor(sh,1,0);
}
#define init_2_1_1_10
void FASTCALL call_2_1_1_10(SH2_struct* sh)
{
  SH2xor(sh,1,1);
}
#define init_2_1_2_10
void FASTCALL call_2_1_2_10(SH2_struct* sh)
{
  SH2xor(sh,1,2);
}
#define init_2_1_3_10
void FASTCALL call_2_1_3_10(SH2_struct* sh)
{
  SH2xor(sh,1,3);
}
#define init_2_1_4_10
void FASTCALL call_2_1_4_10(SH2_struct* sh)
{
  SH2xor(sh,1,4);
}
#define init_2_1_5_10
void FASTCALL call_2_1_5_10(SH2_struct* sh)
{
  SH2xor(sh,1,5);
}
#define init_2_1_6_10
void FASTCALL call_2_1_6_10(SH2_struct* sh)
{
  SH2xor(sh,1,6);
}
#define init_2_1_7_10
void FASTCALL call_2_1_7_10(SH2_struct* sh)
{
  SH2xor(sh,1,7);
}
#define init_2_1_8_10
void FASTCALL call_2_1_8_10(SH2_struct* sh)
{
  SH2xor(sh,1,8);
}
#define init_2_1_9_10
void FASTCALL call_2_1_9_10(SH2_struct* sh)
{
  SH2xor(sh,1,9);
}
#define init_2_1_10_10
void FASTCALL call_2_1_10_10(SH2_struct* sh)
{
  SH2xor(sh,1,10);
}
#define init_2_1_11_10
void FASTCALL call_2_1_11_10(SH2_struct* sh)
{
  SH2xor(sh,1,11);
}
#define init_2_1_12_10
void FASTCALL call_2_1_12_10(SH2_struct* sh)
{
  SH2xor(sh,1,12);
}
#define init_2_1_13_10
void FASTCALL call_2_1_13_10(SH2_struct* sh)
{
  SH2xor(sh,1,13);
}
#define init_2_1_14_10
void FASTCALL call_2_1_14_10(SH2_struct* sh)
{
  SH2xor(sh,1,14);
}
#define init_2_1_15_10
void FASTCALL call_2_1_15_10(SH2_struct* sh)
{
  SH2xor(sh,1,15);
}
#define init_2_2_0_10
void FASTCALL call_2_2_0_10(SH2_struct* sh)
{
  SH2xor(sh,2,0);
}
#define init_2_2_1_10
void FASTCALL call_2_2_1_10(SH2_struct* sh)
{
  SH2xor(sh,2,1);
}
#define init_2_2_2_10
void FASTCALL call_2_2_2_10(SH2_struct* sh)
{
  SH2xor(sh,2,2);
}
#define init_2_2_3_10
void FASTCALL call_2_2_3_10(SH2_struct* sh)
{
  SH2xor(sh,2,3);
}
#define init_2_2_4_10
void FASTCALL call_2_2_4_10(SH2_struct* sh)
{
  SH2xor(sh,2,4);
}
#define init_2_2_5_10
void FASTCALL call_2_2_5_10(SH2_struct* sh)
{
  SH2xor(sh,2,5);
}
#define init_2_2_6_10
void FASTCALL call_2_2_6_10(SH2_struct* sh)
{
  SH2xor(sh,2,6);
}
#define init_2_2_7_10
void FASTCALL call_2_2_7_10(SH2_struct* sh)
{
  SH2xor(sh,2,7);
}
#define init_2_2_8_10
void FASTCALL call_2_2_8_10(SH2_struct* sh)
{
  SH2xor(sh,2,8);
}
#define init_2_2_9_10
void FASTCALL call_2_2_9_10(SH2_struct* sh)
{
  SH2xor(sh,2,9);
}
#define init_2_2_10_10
void FASTCALL call_2_2_10_10(SH2_struct* sh)
{
  SH2xor(sh,2,10);
}
#define init_2_2_11_10
void FASTCALL call_2_2_11_10(SH2_struct* sh)
{
  SH2xor(sh,2,11);
}
#define init_2_2_12_10
void FASTCALL call_2_2_12_10(SH2_struct* sh)
{
  SH2xor(sh,2,12);
}
#define init_2_2_13_10
void FASTCALL call_2_2_13_10(SH2_struct* sh)
{
  SH2xor(sh,2,13);
}
#define init_2_2_14_10
void FASTCALL call_2_2_14_10(SH2_struct* sh)
{
  SH2xor(sh,2,14);
}
#define init_2_2_15_10
void FASTCALL call_2_2_15_10(SH2_struct* sh)
{
  SH2xor(sh,2,15);
}
#define init_2_3_0_10
void FASTCALL call_2_3_0_10(SH2_struct* sh)
{
  SH2xor(sh,3,0);
}
#define init_2_3_1_10
void FASTCALL call_2_3_1_10(SH2_struct* sh)
{
  SH2xor(sh,3,1);
}
#define init_2_3_2_10
void FASTCALL call_2_3_2_10(SH2_struct* sh)
{
  SH2xor(sh,3,2);
}
#define init_2_3_3_10
void FASTCALL call_2_3_3_10(SH2_struct* sh)
{
  SH2xor(sh,3,3);
}
#define init_2_3_4_10
void FASTCALL call_2_3_4_10(SH2_struct* sh)
{
  SH2xor(sh,3,4);
}
#define init_2_3_5_10
void FASTCALL call_2_3_5_10(SH2_struct* sh)
{
  SH2xor(sh,3,5);
}
#define init_2_3_6_10
void FASTCALL call_2_3_6_10(SH2_struct* sh)
{
  SH2xor(sh,3,6);
}
#define init_2_3_7_10
void FASTCALL call_2_3_7_10(SH2_struct* sh)
{
  SH2xor(sh,3,7);
}
#define init_2_3_8_10
void FASTCALL call_2_3_8_10(SH2_struct* sh)
{
  SH2xor(sh,3,8);
}
#define init_2_3_9_10
void FASTCALL call_2_3_9_10(SH2_struct* sh)
{
  SH2xor(sh,3,9);
}
#define init_2_3_10_10
void FASTCALL call_2_3_10_10(SH2_struct* sh)
{
  SH2xor(sh,3,10);
}
#define init_2_3_11_10
void FASTCALL call_2_3_11_10(SH2_struct* sh)
{
  SH2xor(sh,3,11);
}
#define init_2_3_12_10
void FASTCALL call_2_3_12_10(SH2_struct* sh)
{
  SH2xor(sh,3,12);
}
#define init_2_3_13_10
void FASTCALL call_2_3_13_10(SH2_struct* sh)
{
  SH2xor(sh,3,13);
}
#define init_2_3_14_10
void FASTCALL call_2_3_14_10(SH2_struct* sh)
{
  SH2xor(sh,3,14);
}
#define init_2_3_15_10
void FASTCALL call_2_3_15_10(SH2_struct* sh)
{
  SH2xor(sh,3,15);
}
#define init_2_4_0_10
void FASTCALL call_2_4_0_10(SH2_struct* sh)
{
  SH2xor(sh,4,0);
}
#define init_2_4_1_10
void FASTCALL call_2_4_1_10(SH2_struct* sh)
{
  SH2xor(sh,4,1);
}
#define init_2_4_2_10
void FASTCALL call_2_4_2_10(SH2_struct* sh)
{
  SH2xor(sh,4,2);
}
#define init_2_4_3_10
void FASTCALL call_2_4_3_10(SH2_struct* sh)
{
  SH2xor(sh,4,3);
}
#define init_2_4_4_10
void FASTCALL call_2_4_4_10(SH2_struct* sh)
{
  SH2xor(sh,4,4);
}
#define init_2_4_5_10
void FASTCALL call_2_4_5_10(SH2_struct* sh)
{
  SH2xor(sh,4,5);
}
#define init_2_4_6_10
void FASTCALL call_2_4_6_10(SH2_struct* sh)
{
  SH2xor(sh,4,6);
}
#define init_2_4_7_10
void FASTCALL call_2_4_7_10(SH2_struct* sh)
{
  SH2xor(sh,4,7);
}
#define init_2_4_8_10
void FASTCALL call_2_4_8_10(SH2_struct* sh)
{
  SH2xor(sh,4,8);
}
#define init_2_4_9_10
void FASTCALL call_2_4_9_10(SH2_struct* sh)
{
  SH2xor(sh,4,9);
}
#define init_2_4_10_10
void FASTCALL call_2_4_10_10(SH2_struct* sh)
{
  SH2xor(sh,4,10);
}
#define init_2_4_11_10
void FASTCALL call_2_4_11_10(SH2_struct* sh)
{
  SH2xor(sh,4,11);
}
#define init_2_4_12_10
void FASTCALL call_2_4_12_10(SH2_struct* sh)
{
  SH2xor(sh,4,12);
}
#define init_2_4_13_10
void FASTCALL call_2_4_13_10(SH2_struct* sh)
{
  SH2xor(sh,4,13);
}
#define init_2_4_14_10
void FASTCALL call_2_4_14_10(SH2_struct* sh)
{
  SH2xor(sh,4,14);
}
#define init_2_4_15_10
void FASTCALL call_2_4_15_10(SH2_struct* sh)
{
  SH2xor(sh,4,15);
}
#define init_2_5_0_10
void FASTCALL call_2_5_0_10(SH2_struct* sh)
{
  SH2xor(sh,5,0);
}
#define init_2_5_1_10
void FASTCALL call_2_5_1_10(SH2_struct* sh)
{
  SH2xor(sh,5,1);
}
#define init_2_5_2_10
void FASTCALL call_2_5_2_10(SH2_struct* sh)
{
  SH2xor(sh,5,2);
}
#define init_2_5_3_10
void FASTCALL call_2_5_3_10(SH2_struct* sh)
{
  SH2xor(sh,5,3);
}
#define init_2_5_4_10
void FASTCALL call_2_5_4_10(SH2_struct* sh)
{
  SH2xor(sh,5,4);
}
#define init_2_5_5_10
void FASTCALL call_2_5_5_10(SH2_struct* sh)
{
  SH2xor(sh,5,5);
}
#define init_2_5_6_10
void FASTCALL call_2_5_6_10(SH2_struct* sh)
{
  SH2xor(sh,5,6);
}
#define init_2_5_7_10
void FASTCALL call_2_5_7_10(SH2_struct* sh)
{
  SH2xor(sh,5,7);
}
#define init_2_5_8_10
void FASTCALL call_2_5_8_10(SH2_struct* sh)
{
  SH2xor(sh,5,8);
}
#define init_2_5_9_10
void FASTCALL call_2_5_9_10(SH2_struct* sh)
{
  SH2xor(sh,5,9);
}
#define init_2_5_10_10
void FASTCALL call_2_5_10_10(SH2_struct* sh)
{
  SH2xor(sh,5,10);
}
#define init_2_5_11_10
void FASTCALL call_2_5_11_10(SH2_struct* sh)
{
  SH2xor(sh,5,11);
}
#define init_2_5_12_10
void FASTCALL call_2_5_12_10(SH2_struct* sh)
{
  SH2xor(sh,5,12);
}
#define init_2_5_13_10
void FASTCALL call_2_5_13_10(SH2_struct* sh)
{
  SH2xor(sh,5,13);
}
#define init_2_5_14_10
void FASTCALL call_2_5_14_10(SH2_struct* sh)
{
  SH2xor(sh,5,14);
}
#define init_2_5_15_10
void FASTCALL call_2_5_15_10(SH2_struct* sh)
{
  SH2xor(sh,5,15);
}
#define init_2_6_0_10
void FASTCALL call_2_6_0_10(SH2_struct* sh)
{
  SH2xor(sh,6,0);
}
#define init_2_6_1_10
void FASTCALL call_2_6_1_10(SH2_struct* sh)
{
  SH2xor(sh,6,1);
}
#define init_2_6_2_10
void FASTCALL call_2_6_2_10(SH2_struct* sh)
{
  SH2xor(sh,6,2);
}
#define init_2_6_3_10
void FASTCALL call_2_6_3_10(SH2_struct* sh)
{
  SH2xor(sh,6,3);
}
#define init_2_6_4_10
void FASTCALL call_2_6_4_10(SH2_struct* sh)
{
  SH2xor(sh,6,4);
}
#define init_2_6_5_10
void FASTCALL call_2_6_5_10(SH2_struct* sh)
{
  SH2xor(sh,6,5);
}
#define init_2_6_6_10
void FASTCALL call_2_6_6_10(SH2_struct* sh)
{
  SH2xor(sh,6,6);
}
#define init_2_6_7_10
void FASTCALL call_2_6_7_10(SH2_struct* sh)
{
  SH2xor(sh,6,7);
}
#define init_2_6_8_10
void FASTCALL call_2_6_8_10(SH2_struct* sh)
{
  SH2xor(sh,6,8);
}
#define init_2_6_9_10
void FASTCALL call_2_6_9_10(SH2_struct* sh)
{
  SH2xor(sh,6,9);
}
#define init_2_6_10_10
void FASTCALL call_2_6_10_10(SH2_struct* sh)
{
  SH2xor(sh,6,10);
}
#define init_2_6_11_10
void FASTCALL call_2_6_11_10(SH2_struct* sh)
{
  SH2xor(sh,6,11);
}
#define init_2_6_12_10
void FASTCALL call_2_6_12_10(SH2_struct* sh)
{
  SH2xor(sh,6,12);
}
#define init_2_6_13_10
void FASTCALL call_2_6_13_10(SH2_struct* sh)
{
  SH2xor(sh,6,13);
}
#define init_2_6_14_10
void FASTCALL call_2_6_14_10(SH2_struct* sh)
{
  SH2xor(sh,6,14);
}
#define init_2_6_15_10
void FASTCALL call_2_6_15_10(SH2_struct* sh)
{
  SH2xor(sh,6,15);
}
#define init_2_7_0_10
void FASTCALL call_2_7_0_10(SH2_struct* sh)
{
  SH2xor(sh,7,0);
}
#define init_2_7_1_10
void FASTCALL call_2_7_1_10(SH2_struct* sh)
{
  SH2xor(sh,7,1);
}
#define init_2_7_2_10
void FASTCALL call_2_7_2_10(SH2_struct* sh)
{
  SH2xor(sh,7,2);
}
#define init_2_7_3_10
void FASTCALL call_2_7_3_10(SH2_struct* sh)
{
  SH2xor(sh,7,3);
}
#define init_2_7_4_10
void FASTCALL call_2_7_4_10(SH2_struct* sh)
{
  SH2xor(sh,7,4);
}
#define init_2_7_5_10
void FASTCALL call_2_7_5_10(SH2_struct* sh)
{
  SH2xor(sh,7,5);
}
#define init_2_7_6_10
void FASTCALL call_2_7_6_10(SH2_struct* sh)
{
  SH2xor(sh,7,6);
}
#define init_2_7_7_10
void FASTCALL call_2_7_7_10(SH2_struct* sh)
{
  SH2xor(sh,7,7);
}
#define init_2_7_8_10
void FASTCALL call_2_7_8_10(SH2_struct* sh)
{
  SH2xor(sh,7,8);
}
#define init_2_7_9_10
void FASTCALL call_2_7_9_10(SH2_struct* sh)
{
  SH2xor(sh,7,9);
}
#define init_2_7_10_10
void FASTCALL call_2_7_10_10(SH2_struct* sh)
{
  SH2xor(sh,7,10);
}
#define init_2_7_11_10
void FASTCALL call_2_7_11_10(SH2_struct* sh)
{
  SH2xor(sh,7,11);
}
#define init_2_7_12_10
void FASTCALL call_2_7_12_10(SH2_struct* sh)
{
  SH2xor(sh,7,12);
}
#define init_2_7_13_10
void FASTCALL call_2_7_13_10(SH2_struct* sh)
{
  SH2xor(sh,7,13);
}
#define init_2_7_14_10
void FASTCALL call_2_7_14_10(SH2_struct* sh)
{
  SH2xor(sh,7,14);
}
#define init_2_7_15_10
void FASTCALL call_2_7_15_10(SH2_struct* sh)
{
  SH2xor(sh,7,15);
}
#define init_2_8_0_10
void FASTCALL call_2_8_0_10(SH2_struct* sh)
{
  SH2xor(sh,8,0);
}
#define init_2_8_1_10
void FASTCALL call_2_8_1_10(SH2_struct* sh)
{
  SH2xor(sh,8,1);
}
#define init_2_8_2_10
void FASTCALL call_2_8_2_10(SH2_struct* sh)
{
  SH2xor(sh,8,2);
}
#define init_2_8_3_10
void FASTCALL call_2_8_3_10(SH2_struct* sh)
{
  SH2xor(sh,8,3);
}
#define init_2_8_4_10
void FASTCALL call_2_8_4_10(SH2_struct* sh)
{
  SH2xor(sh,8,4);
}
#define init_2_8_5_10
void FASTCALL call_2_8_5_10(SH2_struct* sh)
{
  SH2xor(sh,8,5);
}
#define init_2_8_6_10
void FASTCALL call_2_8_6_10(SH2_struct* sh)
{
  SH2xor(sh,8,6);
}
#define init_2_8_7_10
void FASTCALL call_2_8_7_10(SH2_struct* sh)
{
  SH2xor(sh,8,7);
}
#define init_2_8_8_10
void FASTCALL call_2_8_8_10(SH2_struct* sh)
{
  SH2xor(sh,8,8);
}
#define init_2_8_9_10
void FASTCALL call_2_8_9_10(SH2_struct* sh)
{
  SH2xor(sh,8,9);
}
#define init_2_8_10_10
void FASTCALL call_2_8_10_10(SH2_struct* sh)
{
  SH2xor(sh,8,10);
}
#define init_2_8_11_10
void FASTCALL call_2_8_11_10(SH2_struct* sh)
{
  SH2xor(sh,8,11);
}
#define init_2_8_12_10
void FASTCALL call_2_8_12_10(SH2_struct* sh)
{
  SH2xor(sh,8,12);
}
#define init_2_8_13_10
void FASTCALL call_2_8_13_10(SH2_struct* sh)
{
  SH2xor(sh,8,13);
}
#define init_2_8_14_10
void FASTCALL call_2_8_14_10(SH2_struct* sh)
{
  SH2xor(sh,8,14);
}
#define init_2_8_15_10
void FASTCALL call_2_8_15_10(SH2_struct* sh)
{
  SH2xor(sh,8,15);
}
#define init_2_9_0_10
void FASTCALL call_2_9_0_10(SH2_struct* sh)
{
  SH2xor(sh,9,0);
}
#define init_2_9_1_10
void FASTCALL call_2_9_1_10(SH2_struct* sh)
{
  SH2xor(sh,9,1);
}
#define init_2_9_2_10
void FASTCALL call_2_9_2_10(SH2_struct* sh)
{
  SH2xor(sh,9,2);
}
#define init_2_9_3_10
void FASTCALL call_2_9_3_10(SH2_struct* sh)
{
  SH2xor(sh,9,3);
}
#define init_2_9_4_10
void FASTCALL call_2_9_4_10(SH2_struct* sh)
{
  SH2xor(sh,9,4);
}
#define init_2_9_5_10
void FASTCALL call_2_9_5_10(SH2_struct* sh)
{
  SH2xor(sh,9,5);
}
#define init_2_9_6_10
void FASTCALL call_2_9_6_10(SH2_struct* sh)
{
  SH2xor(sh,9,6);
}
#define init_2_9_7_10
void FASTCALL call_2_9_7_10(SH2_struct* sh)
{
  SH2xor(sh,9,7);
}
#define init_2_9_8_10
void FASTCALL call_2_9_8_10(SH2_struct* sh)
{
  SH2xor(sh,9,8);
}
#define init_2_9_9_10
void FASTCALL call_2_9_9_10(SH2_struct* sh)
{
  SH2xor(sh,9,9);
}
#define init_2_9_10_10
void FASTCALL call_2_9_10_10(SH2_struct* sh)
{
  SH2xor(sh,9,10);
}
#define init_2_9_11_10
void FASTCALL call_2_9_11_10(SH2_struct* sh)
{
  SH2xor(sh,9,11);
}
#define init_2_9_12_10
void FASTCALL call_2_9_12_10(SH2_struct* sh)
{
  SH2xor(sh,9,12);
}
#define init_2_9_13_10
void FASTCALL call_2_9_13_10(SH2_struct* sh)
{
  SH2xor(sh,9,13);
}
#define init_2_9_14_10
void FASTCALL call_2_9_14_10(SH2_struct* sh)
{
  SH2xor(sh,9,14);
}
#define init_2_9_15_10
void FASTCALL call_2_9_15_10(SH2_struct* sh)
{
  SH2xor(sh,9,15);
}
#define init_2_10_0_10
void FASTCALL call_2_10_0_10(SH2_struct* sh)
{
  SH2xor(sh,10,0);
}
#define init_2_10_1_10
void FASTCALL call_2_10_1_10(SH2_struct* sh)
{
  SH2xor(sh,10,1);
}
#define init_2_10_2_10
void FASTCALL call_2_10_2_10(SH2_struct* sh)
{
  SH2xor(sh,10,2);
}
#define init_2_10_3_10
void FASTCALL call_2_10_3_10(SH2_struct* sh)
{
  SH2xor(sh,10,3);
}
#define init_2_10_4_10
void FASTCALL call_2_10_4_10(SH2_struct* sh)
{
  SH2xor(sh,10,4);
}
#define init_2_10_5_10
void FASTCALL call_2_10_5_10(SH2_struct* sh)
{
  SH2xor(sh,10,5);
}
#define init_2_10_6_10
void FASTCALL call_2_10_6_10(SH2_struct* sh)
{
  SH2xor(sh,10,6);
}
#define init_2_10_7_10
void FASTCALL call_2_10_7_10(SH2_struct* sh)
{
  SH2xor(sh,10,7);
}
#define init_2_10_8_10
void FASTCALL call_2_10_8_10(SH2_struct* sh)
{
  SH2xor(sh,10,8);
}
#define init_2_10_9_10
void FASTCALL call_2_10_9_10(SH2_struct* sh)
{
  SH2xor(sh,10,9);
}
#define init_2_10_10_10
void FASTCALL call_2_10_10_10(SH2_struct* sh)
{
  SH2xor(sh,10,10);
}
#define init_2_10_11_10
void FASTCALL call_2_10_11_10(SH2_struct* sh)
{
  SH2xor(sh,10,11);
}
#define init_2_10_12_10
void FASTCALL call_2_10_12_10(SH2_struct* sh)
{
  SH2xor(sh,10,12);
}
#define init_2_10_13_10
void FASTCALL call_2_10_13_10(SH2_struct* sh)
{
  SH2xor(sh,10,13);
}
#define init_2_10_14_10
void FASTCALL call_2_10_14_10(SH2_struct* sh)
{
  SH2xor(sh,10,14);
}
#define init_2_10_15_10
void FASTCALL call_2_10_15_10(SH2_struct* sh)
{
  SH2xor(sh,10,15);
}
#define init_2_11_0_10
void FASTCALL call_2_11_0_10(SH2_struct* sh)
{
  SH2xor(sh,11,0);
}
#define init_2_11_1_10
void FASTCALL call_2_11_1_10(SH2_struct* sh)
{
  SH2xor(sh,11,1);
}
#define init_2_11_2_10
void FASTCALL call_2_11_2_10(SH2_struct* sh)
{
  SH2xor(sh,11,2);
}
#define init_2_11_3_10
void FASTCALL call_2_11_3_10(SH2_struct* sh)
{
  SH2xor(sh,11,3);
}
#define init_2_11_4_10
void FASTCALL call_2_11_4_10(SH2_struct* sh)
{
  SH2xor(sh,11,4);
}
#define init_2_11_5_10
void FASTCALL call_2_11_5_10(SH2_struct* sh)
{
  SH2xor(sh,11,5);
}
#define init_2_11_6_10
void FASTCALL call_2_11_6_10(SH2_struct* sh)
{
  SH2xor(sh,11,6);
}
#define init_2_11_7_10
void FASTCALL call_2_11_7_10(SH2_struct* sh)
{
  SH2xor(sh,11,7);
}
#define init_2_11_8_10
void FASTCALL call_2_11_8_10(SH2_struct* sh)
{
  SH2xor(sh,11,8);
}
#define init_2_11_9_10
void FASTCALL call_2_11_9_10(SH2_struct* sh)
{
  SH2xor(sh,11,9);
}
#define init_2_11_10_10
void FASTCALL call_2_11_10_10(SH2_struct* sh)
{
  SH2xor(sh,11,10);
}
#define init_2_11_11_10
void FASTCALL call_2_11_11_10(SH2_struct* sh)
{
  SH2xor(sh,11,11);
}
#define init_2_11_12_10
void FASTCALL call_2_11_12_10(SH2_struct* sh)
{
  SH2xor(sh,11,12);
}
#define init_2_11_13_10
void FASTCALL call_2_11_13_10(SH2_struct* sh)
{
  SH2xor(sh,11,13);
}
#define init_2_11_14_10
void FASTCALL call_2_11_14_10(SH2_struct* sh)
{
  SH2xor(sh,11,14);
}
#define init_2_11_15_10
void FASTCALL call_2_11_15_10(SH2_struct* sh)
{
  SH2xor(sh,11,15);
}
#define init_2_12_0_10
void FASTCALL call_2_12_0_10(SH2_struct* sh)
{
  SH2xor(sh,12,0);
}
#define init_2_12_1_10
void FASTCALL call_2_12_1_10(SH2_struct* sh)
{
  SH2xor(sh,12,1);
}
#define init_2_12_2_10
void FASTCALL call_2_12_2_10(SH2_struct* sh)
{
  SH2xor(sh,12,2);
}
#define init_2_12_3_10
void FASTCALL call_2_12_3_10(SH2_struct* sh)
{
  SH2xor(sh,12,3);
}
#define init_2_12_4_10
void FASTCALL call_2_12_4_10(SH2_struct* sh)
{
  SH2xor(sh,12,4);
}
#define init_2_12_5_10
void FASTCALL call_2_12_5_10(SH2_struct* sh)
{
  SH2xor(sh,12,5);
}
#define init_2_12_6_10
void FASTCALL call_2_12_6_10(SH2_struct* sh)
{
  SH2xor(sh,12,6);
}
#define init_2_12_7_10
void FASTCALL call_2_12_7_10(SH2_struct* sh)
{
  SH2xor(sh,12,7);
}
#define init_2_12_8_10
void FASTCALL call_2_12_8_10(SH2_struct* sh)
{
  SH2xor(sh,12,8);
}
#define init_2_12_9_10
void FASTCALL call_2_12_9_10(SH2_struct* sh)
{
  SH2xor(sh,12,9);
}
#define init_2_12_10_10
void FASTCALL call_2_12_10_10(SH2_struct* sh)
{
  SH2xor(sh,12,10);
}
#define init_2_12_11_10
void FASTCALL call_2_12_11_10(SH2_struct* sh)
{
  SH2xor(sh,12,11);
}
#define init_2_12_12_10
void FASTCALL call_2_12_12_10(SH2_struct* sh)
{
  SH2xor(sh,12,12);
}
#define init_2_12_13_10
void FASTCALL call_2_12_13_10(SH2_struct* sh)
{
  SH2xor(sh,12,13);
}
#define init_2_12_14_10
void FASTCALL call_2_12_14_10(SH2_struct* sh)
{
  SH2xor(sh,12,14);
}
#define init_2_12_15_10
void FASTCALL call_2_12_15_10(SH2_struct* sh)
{
  SH2xor(sh,12,15);
}
#define init_2_13_0_10
void FASTCALL call_2_13_0_10(SH2_struct* sh)
{
  SH2xor(sh,13,0);
}
#define init_2_13_1_10
void FASTCALL call_2_13_1_10(SH2_struct* sh)
{
  SH2xor(sh,13,1);
}
#define init_2_13_2_10
void FASTCALL call_2_13_2_10(SH2_struct* sh)
{
  SH2xor(sh,13,2);
}
#define init_2_13_3_10
void FASTCALL call_2_13_3_10(SH2_struct* sh)
{
  SH2xor(sh,13,3);
}
#define init_2_13_4_10
void FASTCALL call_2_13_4_10(SH2_struct* sh)
{
  SH2xor(sh,13,4);
}
#define init_2_13_5_10
void FASTCALL call_2_13_5_10(SH2_struct* sh)
{
  SH2xor(sh,13,5);
}
#define init_2_13_6_10
void FASTCALL call_2_13_6_10(SH2_struct* sh)
{
  SH2xor(sh,13,6);
}
#define init_2_13_7_10
void FASTCALL call_2_13_7_10(SH2_struct* sh)
{
  SH2xor(sh,13,7);
}
#define init_2_13_8_10
void FASTCALL call_2_13_8_10(SH2_struct* sh)
{
  SH2xor(sh,13,8);
}
#define init_2_13_9_10
void FASTCALL call_2_13_9_10(SH2_struct* sh)
{
  SH2xor(sh,13,9);
}
#define init_2_13_10_10
void FASTCALL call_2_13_10_10(SH2_struct* sh)
{
  SH2xor(sh,13,10);
}
#define init_2_13_11_10
void FASTCALL call_2_13_11_10(SH2_struct* sh)
{
  SH2xor(sh,13,11);
}
#define init_2_13_12_10
void FASTCALL call_2_13_12_10(SH2_struct* sh)
{
  SH2xor(sh,13,12);
}
#define init_2_13_13_10
void FASTCALL call_2_13_13_10(SH2_struct* sh)
{
  SH2xor(sh,13,13);
}
#define init_2_13_14_10
void FASTCALL call_2_13_14_10(SH2_struct* sh)
{
  SH2xor(sh,13,14);
}
#define init_2_13_15_10
void FASTCALL call_2_13_15_10(SH2_struct* sh)
{
  SH2xor(sh,13,15);
}
#define init_2_14_0_10
void FASTCALL call_2_14_0_10(SH2_struct* sh)
{
  SH2xor(sh,14,0);
}
#define init_2_14_1_10
void FASTCALL call_2_14_1_10(SH2_struct* sh)
{
  SH2xor(sh,14,1);
}
#define init_2_14_2_10
void FASTCALL call_2_14_2_10(SH2_struct* sh)
{
  SH2xor(sh,14,2);
}
#define init_2_14_3_10
void FASTCALL call_2_14_3_10(SH2_struct* sh)
{
  SH2xor(sh,14,3);
}
#define init_2_14_4_10
void FASTCALL call_2_14_4_10(SH2_struct* sh)
{
  SH2xor(sh,14,4);
}
#define init_2_14_5_10
void FASTCALL call_2_14_5_10(SH2_struct* sh)
{
  SH2xor(sh,14,5);
}
#define init_2_14_6_10
void FASTCALL call_2_14_6_10(SH2_struct* sh)
{
  SH2xor(sh,14,6);
}
#define init_2_14_7_10
void FASTCALL call_2_14_7_10(SH2_struct* sh)
{
  SH2xor(sh,14,7);
}
#define init_2_14_8_10
void FASTCALL call_2_14_8_10(SH2_struct* sh)
{
  SH2xor(sh,14,8);
}
#define init_2_14_9_10
void FASTCALL call_2_14_9_10(SH2_struct* sh)
{
  SH2xor(sh,14,9);
}
#define init_2_14_10_10
void FASTCALL call_2_14_10_10(SH2_struct* sh)
{
  SH2xor(sh,14,10);
}
#define init_2_14_11_10
void FASTCALL call_2_14_11_10(SH2_struct* sh)
{
  SH2xor(sh,14,11);
}
#define init_2_14_12_10
void FASTCALL call_2_14_12_10(SH2_struct* sh)
{
  SH2xor(sh,14,12);
}
#define init_2_14_13_10
void FASTCALL call_2_14_13_10(SH2_struct* sh)
{
  SH2xor(sh,14,13);
}
#define init_2_14_14_10
void FASTCALL call_2_14_14_10(SH2_struct* sh)
{
  SH2xor(sh,14,14);
}
#define init_2_14_15_10
void FASTCALL call_2_14_15_10(SH2_struct* sh)
{
  SH2xor(sh,14,15);
}
#define init_2_15_0_10
void FASTCALL call_2_15_0_10(SH2_struct* sh)
{
  SH2xor(sh,15,0);
}
#define init_2_15_1_10
void FASTCALL call_2_15_1_10(SH2_struct* sh)
{
  SH2xor(sh,15,1);
}
#define init_2_15_2_10
void FASTCALL call_2_15_2_10(SH2_struct* sh)
{
  SH2xor(sh,15,2);
}
#define init_2_15_3_10
void FASTCALL call_2_15_3_10(SH2_struct* sh)
{
  SH2xor(sh,15,3);
}
#define init_2_15_4_10
void FASTCALL call_2_15_4_10(SH2_struct* sh)
{
  SH2xor(sh,15,4);
}
#define init_2_15_5_10
void FASTCALL call_2_15_5_10(SH2_struct* sh)
{
  SH2xor(sh,15,5);
}
#define init_2_15_6_10
void FASTCALL call_2_15_6_10(SH2_struct* sh)
{
  SH2xor(sh,15,6);
}
#define init_2_15_7_10
void FASTCALL call_2_15_7_10(SH2_struct* sh)
{
  SH2xor(sh,15,7);
}
#define init_2_15_8_10
void FASTCALL call_2_15_8_10(SH2_struct* sh)
{
  SH2xor(sh,15,8);
}
#define init_2_15_9_10
void FASTCALL call_2_15_9_10(SH2_struct* sh)
{
  SH2xor(sh,15,9);
}
#define init_2_15_10_10
void FASTCALL call_2_15_10_10(SH2_struct* sh)
{
  SH2xor(sh,15,10);
}
#define init_2_15_11_10
void FASTCALL call_2_15_11_10(SH2_struct* sh)
{
  SH2xor(sh,15,11);
}
#define init_2_15_12_10
void FASTCALL call_2_15_12_10(SH2_struct* sh)
{
  SH2xor(sh,15,12);
}
#define init_2_15_13_10
void FASTCALL call_2_15_13_10(SH2_struct* sh)
{
  SH2xor(sh,15,13);
}
#define init_2_15_14_10
void FASTCALL call_2_15_14_10(SH2_struct* sh)
{
  SH2xor(sh,15,14);
}
#define init_2_15_15_10
void FASTCALL call_2_15_15_10(SH2_struct* sh)
{
  SH2xor(sh,15,15);
}
#define init_2_0_0_13
void FASTCALL call_2_0_0_13(SH2_struct* sh)
{
  SH2xtrct(sh,0,0);
}
#define init_2_0_1_13
void FASTCALL call_2_0_1_13(SH2_struct* sh)
{
  SH2xtrct(sh,0,1);
}
#define init_2_0_2_13
void FASTCALL call_2_0_2_13(SH2_struct* sh)
{
  SH2xtrct(sh,0,2);
}
#define init_2_0_3_13
void FASTCALL call_2_0_3_13(SH2_struct* sh)
{
  SH2xtrct(sh,0,3);
}
#define init_2_0_4_13
void FASTCALL call_2_0_4_13(SH2_struct* sh)
{
  SH2xtrct(sh,0,4);
}
#define init_2_0_5_13
void FASTCALL call_2_0_5_13(SH2_struct* sh)
{
  SH2xtrct(sh,0,5);
}
#define init_2_0_6_13
void FASTCALL call_2_0_6_13(SH2_struct* sh)
{
  SH2xtrct(sh,0,6);
}
#define init_2_0_7_13
void FASTCALL call_2_0_7_13(SH2_struct* sh)
{
  SH2xtrct(sh,0,7);
}
#define init_2_0_8_13
void FASTCALL call_2_0_8_13(SH2_struct* sh)
{
  SH2xtrct(sh,0,8);
}
#define init_2_0_9_13
void FASTCALL call_2_0_9_13(SH2_struct* sh)
{
  SH2xtrct(sh,0,9);
}
#define init_2_0_10_13
void FASTCALL call_2_0_10_13(SH2_struct* sh)
{
  SH2xtrct(sh,0,10);
}
#define init_2_0_11_13
void FASTCALL call_2_0_11_13(SH2_struct* sh)
{
  SH2xtrct(sh,0,11);
}
#define init_2_0_12_13
void FASTCALL call_2_0_12_13(SH2_struct* sh)
{
  SH2xtrct(sh,0,12);
}
#define init_2_0_13_13
void FASTCALL call_2_0_13_13(SH2_struct* sh)
{
  SH2xtrct(sh,0,13);
}
#define init_2_0_14_13
void FASTCALL call_2_0_14_13(SH2_struct* sh)
{
  SH2xtrct(sh,0,14);
}
#define init_2_0_15_13
void FASTCALL call_2_0_15_13(SH2_struct* sh)
{
  SH2xtrct(sh,0,15);
}
#define init_2_1_0_13
void FASTCALL call_2_1_0_13(SH2_struct* sh)
{
  SH2xtrct(sh,1,0);
}
#define init_2_1_1_13
void FASTCALL call_2_1_1_13(SH2_struct* sh)
{
  SH2xtrct(sh,1,1);
}
#define init_2_1_2_13
void FASTCALL call_2_1_2_13(SH2_struct* sh)
{
  SH2xtrct(sh,1,2);
}
#define init_2_1_3_13
void FASTCALL call_2_1_3_13(SH2_struct* sh)
{
  SH2xtrct(sh,1,3);
}
#define init_2_1_4_13
void FASTCALL call_2_1_4_13(SH2_struct* sh)
{
  SH2xtrct(sh,1,4);
}
#define init_2_1_5_13
void FASTCALL call_2_1_5_13(SH2_struct* sh)
{
  SH2xtrct(sh,1,5);
}
#define init_2_1_6_13
void FASTCALL call_2_1_6_13(SH2_struct* sh)
{
  SH2xtrct(sh,1,6);
}
#define init_2_1_7_13
void FASTCALL call_2_1_7_13(SH2_struct* sh)
{
  SH2xtrct(sh,1,7);
}
#define init_2_1_8_13
void FASTCALL call_2_1_8_13(SH2_struct* sh)
{
  SH2xtrct(sh,1,8);
}
#define init_2_1_9_13
void FASTCALL call_2_1_9_13(SH2_struct* sh)
{
  SH2xtrct(sh,1,9);
}
#define init_2_1_10_13
void FASTCALL call_2_1_10_13(SH2_struct* sh)
{
  SH2xtrct(sh,1,10);
}
#define init_2_1_11_13
void FASTCALL call_2_1_11_13(SH2_struct* sh)
{
  SH2xtrct(sh,1,11);
}
#define init_2_1_12_13
void FASTCALL call_2_1_12_13(SH2_struct* sh)
{
  SH2xtrct(sh,1,12);
}
#define init_2_1_13_13
void FASTCALL call_2_1_13_13(SH2_struct* sh)
{
  SH2xtrct(sh,1,13);
}
#define init_2_1_14_13
void FASTCALL call_2_1_14_13(SH2_struct* sh)
{
  SH2xtrct(sh,1,14);
}
#define init_2_1_15_13
void FASTCALL call_2_1_15_13(SH2_struct* sh)
{
  SH2xtrct(sh,1,15);
}
#define init_2_2_0_13
void FASTCALL call_2_2_0_13(SH2_struct* sh)
{
  SH2xtrct(sh,2,0);
}
#define init_2_2_1_13
void FASTCALL call_2_2_1_13(SH2_struct* sh)
{
  SH2xtrct(sh,2,1);
}
#define init_2_2_2_13
void FASTCALL call_2_2_2_13(SH2_struct* sh)
{
  SH2xtrct(sh,2,2);
}
#define init_2_2_3_13
void FASTCALL call_2_2_3_13(SH2_struct* sh)
{
  SH2xtrct(sh,2,3);
}
#define init_2_2_4_13
void FASTCALL call_2_2_4_13(SH2_struct* sh)
{
  SH2xtrct(sh,2,4);
}
#define init_2_2_5_13
void FASTCALL call_2_2_5_13(SH2_struct* sh)
{
  SH2xtrct(sh,2,5);
}
#define init_2_2_6_13
void FASTCALL call_2_2_6_13(SH2_struct* sh)
{
  SH2xtrct(sh,2,6);
}
#define init_2_2_7_13
void FASTCALL call_2_2_7_13(SH2_struct* sh)
{
  SH2xtrct(sh,2,7);
}
#define init_2_2_8_13
void FASTCALL call_2_2_8_13(SH2_struct* sh)
{
  SH2xtrct(sh,2,8);
}
#define init_2_2_9_13
void FASTCALL call_2_2_9_13(SH2_struct* sh)
{
  SH2xtrct(sh,2,9);
}
#define init_2_2_10_13
void FASTCALL call_2_2_10_13(SH2_struct* sh)
{
  SH2xtrct(sh,2,10);
}
#define init_2_2_11_13
void FASTCALL call_2_2_11_13(SH2_struct* sh)
{
  SH2xtrct(sh,2,11);
}
#define init_2_2_12_13
void FASTCALL call_2_2_12_13(SH2_struct* sh)
{
  SH2xtrct(sh,2,12);
}
#define init_2_2_13_13
void FASTCALL call_2_2_13_13(SH2_struct* sh)
{
  SH2xtrct(sh,2,13);
}
#define init_2_2_14_13
void FASTCALL call_2_2_14_13(SH2_struct* sh)
{
  SH2xtrct(sh,2,14);
}
#define init_2_2_15_13
void FASTCALL call_2_2_15_13(SH2_struct* sh)
{
  SH2xtrct(sh,2,15);
}
#define init_2_3_0_13
void FASTCALL call_2_3_0_13(SH2_struct* sh)
{
  SH2xtrct(sh,3,0);
}
#define init_2_3_1_13
void FASTCALL call_2_3_1_13(SH2_struct* sh)
{
  SH2xtrct(sh,3,1);
}
#define init_2_3_2_13
void FASTCALL call_2_3_2_13(SH2_struct* sh)
{
  SH2xtrct(sh,3,2);
}
#define init_2_3_3_13
void FASTCALL call_2_3_3_13(SH2_struct* sh)
{
  SH2xtrct(sh,3,3);
}
#define init_2_3_4_13
void FASTCALL call_2_3_4_13(SH2_struct* sh)
{
  SH2xtrct(sh,3,4);
}
#define init_2_3_5_13
void FASTCALL call_2_3_5_13(SH2_struct* sh)
{
  SH2xtrct(sh,3,5);
}
#define init_2_3_6_13
void FASTCALL call_2_3_6_13(SH2_struct* sh)
{
  SH2xtrct(sh,3,6);
}
#define init_2_3_7_13
void FASTCALL call_2_3_7_13(SH2_struct* sh)
{
  SH2xtrct(sh,3,7);
}
#define init_2_3_8_13
void FASTCALL call_2_3_8_13(SH2_struct* sh)
{
  SH2xtrct(sh,3,8);
}
#define init_2_3_9_13
void FASTCALL call_2_3_9_13(SH2_struct* sh)
{
  SH2xtrct(sh,3,9);
}
#define init_2_3_10_13
void FASTCALL call_2_3_10_13(SH2_struct* sh)
{
  SH2xtrct(sh,3,10);
}
#define init_2_3_11_13
void FASTCALL call_2_3_11_13(SH2_struct* sh)
{
  SH2xtrct(sh,3,11);
}
#define init_2_3_12_13
void FASTCALL call_2_3_12_13(SH2_struct* sh)
{
  SH2xtrct(sh,3,12);
}
#define init_2_3_13_13
void FASTCALL call_2_3_13_13(SH2_struct* sh)
{
  SH2xtrct(sh,3,13);
}
#define init_2_3_14_13
void FASTCALL call_2_3_14_13(SH2_struct* sh)
{
  SH2xtrct(sh,3,14);
}
#define init_2_3_15_13
void FASTCALL call_2_3_15_13(SH2_struct* sh)
{
  SH2xtrct(sh,3,15);
}
#define init_2_4_0_13
void FASTCALL call_2_4_0_13(SH2_struct* sh)
{
  SH2xtrct(sh,4,0);
}
#define init_2_4_1_13
void FASTCALL call_2_4_1_13(SH2_struct* sh)
{
  SH2xtrct(sh,4,1);
}
#define init_2_4_2_13
void FASTCALL call_2_4_2_13(SH2_struct* sh)
{
  SH2xtrct(sh,4,2);
}
#define init_2_4_3_13
void FASTCALL call_2_4_3_13(SH2_struct* sh)
{
  SH2xtrct(sh,4,3);
}
#define init_2_4_4_13
void FASTCALL call_2_4_4_13(SH2_struct* sh)
{
  SH2xtrct(sh,4,4);
}
#define init_2_4_5_13
void FASTCALL call_2_4_5_13(SH2_struct* sh)
{
  SH2xtrct(sh,4,5);
}
#define init_2_4_6_13
void FASTCALL call_2_4_6_13(SH2_struct* sh)
{
  SH2xtrct(sh,4,6);
}
#define init_2_4_7_13
void FASTCALL call_2_4_7_13(SH2_struct* sh)
{
  SH2xtrct(sh,4,7);
}
#define init_2_4_8_13
void FASTCALL call_2_4_8_13(SH2_struct* sh)
{
  SH2xtrct(sh,4,8);
}
#define init_2_4_9_13
void FASTCALL call_2_4_9_13(SH2_struct* sh)
{
  SH2xtrct(sh,4,9);
}
#define init_2_4_10_13
void FASTCALL call_2_4_10_13(SH2_struct* sh)
{
  SH2xtrct(sh,4,10);
}
#define init_2_4_11_13
void FASTCALL call_2_4_11_13(SH2_struct* sh)
{
  SH2xtrct(sh,4,11);
}
#define init_2_4_12_13
void FASTCALL call_2_4_12_13(SH2_struct* sh)
{
  SH2xtrct(sh,4,12);
}
#define init_2_4_13_13
void FASTCALL call_2_4_13_13(SH2_struct* sh)
{
  SH2xtrct(sh,4,13);
}
#define init_2_4_14_13
void FASTCALL call_2_4_14_13(SH2_struct* sh)
{
  SH2xtrct(sh,4,14);
}
#define init_2_4_15_13
void FASTCALL call_2_4_15_13(SH2_struct* sh)
{
  SH2xtrct(sh,4,15);
}
#define init_2_5_0_13
void FASTCALL call_2_5_0_13(SH2_struct* sh)
{
  SH2xtrct(sh,5,0);
}
#define init_2_5_1_13
void FASTCALL call_2_5_1_13(SH2_struct* sh)
{
  SH2xtrct(sh,5,1);
}
#define init_2_5_2_13
void FASTCALL call_2_5_2_13(SH2_struct* sh)
{
  SH2xtrct(sh,5,2);
}
#define init_2_5_3_13
void FASTCALL call_2_5_3_13(SH2_struct* sh)
{
  SH2xtrct(sh,5,3);
}
#define init_2_5_4_13
void FASTCALL call_2_5_4_13(SH2_struct* sh)
{
  SH2xtrct(sh,5,4);
}
#define init_2_5_5_13
void FASTCALL call_2_5_5_13(SH2_struct* sh)
{
  SH2xtrct(sh,5,5);
}
#define init_2_5_6_13
void FASTCALL call_2_5_6_13(SH2_struct* sh)
{
  SH2xtrct(sh,5,6);
}
#define init_2_5_7_13
void FASTCALL call_2_5_7_13(SH2_struct* sh)
{
  SH2xtrct(sh,5,7);
}
#define init_2_5_8_13
void FASTCALL call_2_5_8_13(SH2_struct* sh)
{
  SH2xtrct(sh,5,8);
}
#define init_2_5_9_13
void FASTCALL call_2_5_9_13(SH2_struct* sh)
{
  SH2xtrct(sh,5,9);
}
#define init_2_5_10_13
void FASTCALL call_2_5_10_13(SH2_struct* sh)
{
  SH2xtrct(sh,5,10);
}
#define init_2_5_11_13
void FASTCALL call_2_5_11_13(SH2_struct* sh)
{
  SH2xtrct(sh,5,11);
}
#define init_2_5_12_13
void FASTCALL call_2_5_12_13(SH2_struct* sh)
{
  SH2xtrct(sh,5,12);
}
#define init_2_5_13_13
void FASTCALL call_2_5_13_13(SH2_struct* sh)
{
  SH2xtrct(sh,5,13);
}
#define init_2_5_14_13
void FASTCALL call_2_5_14_13(SH2_struct* sh)
{
  SH2xtrct(sh,5,14);
}
#define init_2_5_15_13
void FASTCALL call_2_5_15_13(SH2_struct* sh)
{
  SH2xtrct(sh,5,15);
}
#define init_2_6_0_13
void FASTCALL call_2_6_0_13(SH2_struct* sh)
{
  SH2xtrct(sh,6,0);
}
#define init_2_6_1_13
void FASTCALL call_2_6_1_13(SH2_struct* sh)
{
  SH2xtrct(sh,6,1);
}
#define init_2_6_2_13
void FASTCALL call_2_6_2_13(SH2_struct* sh)
{
  SH2xtrct(sh,6,2);
}
#define init_2_6_3_13
void FASTCALL call_2_6_3_13(SH2_struct* sh)
{
  SH2xtrct(sh,6,3);
}
#define init_2_6_4_13
void FASTCALL call_2_6_4_13(SH2_struct* sh)
{
  SH2xtrct(sh,6,4);
}
#define init_2_6_5_13
void FASTCALL call_2_6_5_13(SH2_struct* sh)
{
  SH2xtrct(sh,6,5);
}
#define init_2_6_6_13
void FASTCALL call_2_6_6_13(SH2_struct* sh)
{
  SH2xtrct(sh,6,6);
}
#define init_2_6_7_13
void FASTCALL call_2_6_7_13(SH2_struct* sh)
{
  SH2xtrct(sh,6,7);
}
#define init_2_6_8_13
void FASTCALL call_2_6_8_13(SH2_struct* sh)
{
  SH2xtrct(sh,6,8);
}
#define init_2_6_9_13
void FASTCALL call_2_6_9_13(SH2_struct* sh)
{
  SH2xtrct(sh,6,9);
}
#define init_2_6_10_13
void FASTCALL call_2_6_10_13(SH2_struct* sh)
{
  SH2xtrct(sh,6,10);
}
#define init_2_6_11_13
void FASTCALL call_2_6_11_13(SH2_struct* sh)
{
  SH2xtrct(sh,6,11);
}
#define init_2_6_12_13
void FASTCALL call_2_6_12_13(SH2_struct* sh)
{
  SH2xtrct(sh,6,12);
}
#define init_2_6_13_13
void FASTCALL call_2_6_13_13(SH2_struct* sh)
{
  SH2xtrct(sh,6,13);
}
#define init_2_6_14_13
void FASTCALL call_2_6_14_13(SH2_struct* sh)
{
  SH2xtrct(sh,6,14);
}
#define init_2_6_15_13
void FASTCALL call_2_6_15_13(SH2_struct* sh)
{
  SH2xtrct(sh,6,15);
}
#define init_2_7_0_13
void FASTCALL call_2_7_0_13(SH2_struct* sh)
{
  SH2xtrct(sh,7,0);
}
#define init_2_7_1_13
void FASTCALL call_2_7_1_13(SH2_struct* sh)
{
  SH2xtrct(sh,7,1);
}
#define init_2_7_2_13
void FASTCALL call_2_7_2_13(SH2_struct* sh)
{
  SH2xtrct(sh,7,2);
}
#define init_2_7_3_13
void FASTCALL call_2_7_3_13(SH2_struct* sh)
{
  SH2xtrct(sh,7,3);
}
#define init_2_7_4_13
void FASTCALL call_2_7_4_13(SH2_struct* sh)
{
  SH2xtrct(sh,7,4);
}
#define init_2_7_5_13
void FASTCALL call_2_7_5_13(SH2_struct* sh)
{
  SH2xtrct(sh,7,5);
}
#define init_2_7_6_13
void FASTCALL call_2_7_6_13(SH2_struct* sh)
{
  SH2xtrct(sh,7,6);
}
#define init_2_7_7_13
void FASTCALL call_2_7_7_13(SH2_struct* sh)
{
  SH2xtrct(sh,7,7);
}
#define init_2_7_8_13
void FASTCALL call_2_7_8_13(SH2_struct* sh)
{
  SH2xtrct(sh,7,8);
}
#define init_2_7_9_13
void FASTCALL call_2_7_9_13(SH2_struct* sh)
{
  SH2xtrct(sh,7,9);
}
#define init_2_7_10_13
void FASTCALL call_2_7_10_13(SH2_struct* sh)
{
  SH2xtrct(sh,7,10);
}
#define init_2_7_11_13
void FASTCALL call_2_7_11_13(SH2_struct* sh)
{
  SH2xtrct(sh,7,11);
}
#define init_2_7_12_13
void FASTCALL call_2_7_12_13(SH2_struct* sh)
{
  SH2xtrct(sh,7,12);
}
#define init_2_7_13_13
void FASTCALL call_2_7_13_13(SH2_struct* sh)
{
  SH2xtrct(sh,7,13);
}
#define init_2_7_14_13
void FASTCALL call_2_7_14_13(SH2_struct* sh)
{
  SH2xtrct(sh,7,14);
}
#define init_2_7_15_13
void FASTCALL call_2_7_15_13(SH2_struct* sh)
{
  SH2xtrct(sh,7,15);
}
#define init_2_8_0_13
void FASTCALL call_2_8_0_13(SH2_struct* sh)
{
  SH2xtrct(sh,8,0);
}
#define init_2_8_1_13
void FASTCALL call_2_8_1_13(SH2_struct* sh)
{
  SH2xtrct(sh,8,1);
}
#define init_2_8_2_13
void FASTCALL call_2_8_2_13(SH2_struct* sh)
{
  SH2xtrct(sh,8,2);
}
#define init_2_8_3_13
void FASTCALL call_2_8_3_13(SH2_struct* sh)
{
  SH2xtrct(sh,8,3);
}
#define init_2_8_4_13
void FASTCALL call_2_8_4_13(SH2_struct* sh)
{
  SH2xtrct(sh,8,4);
}
#define init_2_8_5_13
void FASTCALL call_2_8_5_13(SH2_struct* sh)
{
  SH2xtrct(sh,8,5);
}
#define init_2_8_6_13
void FASTCALL call_2_8_6_13(SH2_struct* sh)
{
  SH2xtrct(sh,8,6);
}
#define init_2_8_7_13
void FASTCALL call_2_8_7_13(SH2_struct* sh)
{
  SH2xtrct(sh,8,7);
}
#define init_2_8_8_13
void FASTCALL call_2_8_8_13(SH2_struct* sh)
{
  SH2xtrct(sh,8,8);
}
#define init_2_8_9_13
void FASTCALL call_2_8_9_13(SH2_struct* sh)
{
  SH2xtrct(sh,8,9);
}
#define init_2_8_10_13
void FASTCALL call_2_8_10_13(SH2_struct* sh)
{
  SH2xtrct(sh,8,10);
}
#define init_2_8_11_13
void FASTCALL call_2_8_11_13(SH2_struct* sh)
{
  SH2xtrct(sh,8,11);
}
#define init_2_8_12_13
void FASTCALL call_2_8_12_13(SH2_struct* sh)
{
  SH2xtrct(sh,8,12);
}
#define init_2_8_13_13
void FASTCALL call_2_8_13_13(SH2_struct* sh)
{
  SH2xtrct(sh,8,13);
}
#define init_2_8_14_13
void FASTCALL call_2_8_14_13(SH2_struct* sh)
{
  SH2xtrct(sh,8,14);
}
#define init_2_8_15_13
void FASTCALL call_2_8_15_13(SH2_struct* sh)
{
  SH2xtrct(sh,8,15);
}
#define init_2_9_0_13
void FASTCALL call_2_9_0_13(SH2_struct* sh)
{
  SH2xtrct(sh,9,0);
}
#define init_2_9_1_13
void FASTCALL call_2_9_1_13(SH2_struct* sh)
{
  SH2xtrct(sh,9,1);
}
#define init_2_9_2_13
void FASTCALL call_2_9_2_13(SH2_struct* sh)
{
  SH2xtrct(sh,9,2);
}
#define init_2_9_3_13
void FASTCALL call_2_9_3_13(SH2_struct* sh)
{
  SH2xtrct(sh,9,3);
}
#define init_2_9_4_13
void FASTCALL call_2_9_4_13(SH2_struct* sh)
{
  SH2xtrct(sh,9,4);
}
#define init_2_9_5_13
void FASTCALL call_2_9_5_13(SH2_struct* sh)
{
  SH2xtrct(sh,9,5);
}
#define init_2_9_6_13
void FASTCALL call_2_9_6_13(SH2_struct* sh)
{
  SH2xtrct(sh,9,6);
}
#define init_2_9_7_13
void FASTCALL call_2_9_7_13(SH2_struct* sh)
{
  SH2xtrct(sh,9,7);
}
#define init_2_9_8_13
void FASTCALL call_2_9_8_13(SH2_struct* sh)
{
  SH2xtrct(sh,9,8);
}
#define init_2_9_9_13
void FASTCALL call_2_9_9_13(SH2_struct* sh)
{
  SH2xtrct(sh,9,9);
}
#define init_2_9_10_13
void FASTCALL call_2_9_10_13(SH2_struct* sh)
{
  SH2xtrct(sh,9,10);
}
#define init_2_9_11_13
void FASTCALL call_2_9_11_13(SH2_struct* sh)
{
  SH2xtrct(sh,9,11);
}
#define init_2_9_12_13
void FASTCALL call_2_9_12_13(SH2_struct* sh)
{
  SH2xtrct(sh,9,12);
}
#define init_2_9_13_13
void FASTCALL call_2_9_13_13(SH2_struct* sh)
{
  SH2xtrct(sh,9,13);
}
#define init_2_9_14_13
void FASTCALL call_2_9_14_13(SH2_struct* sh)
{
  SH2xtrct(sh,9,14);
}
#define init_2_9_15_13
void FASTCALL call_2_9_15_13(SH2_struct* sh)
{
  SH2xtrct(sh,9,15);
}
#define init_2_10_0_13
void FASTCALL call_2_10_0_13(SH2_struct* sh)
{
  SH2xtrct(sh,10,0);
}
#define init_2_10_1_13
void FASTCALL call_2_10_1_13(SH2_struct* sh)
{
  SH2xtrct(sh,10,1);
}
#define init_2_10_2_13
void FASTCALL call_2_10_2_13(SH2_struct* sh)
{
  SH2xtrct(sh,10,2);
}
#define init_2_10_3_13
void FASTCALL call_2_10_3_13(SH2_struct* sh)
{
  SH2xtrct(sh,10,3);
}
#define init_2_10_4_13
void FASTCALL call_2_10_4_13(SH2_struct* sh)
{
  SH2xtrct(sh,10,4);
}
#define init_2_10_5_13
void FASTCALL call_2_10_5_13(SH2_struct* sh)
{
  SH2xtrct(sh,10,5);
}
#define init_2_10_6_13
void FASTCALL call_2_10_6_13(SH2_struct* sh)
{
  SH2xtrct(sh,10,6);
}
#define init_2_10_7_13
void FASTCALL call_2_10_7_13(SH2_struct* sh)
{
  SH2xtrct(sh,10,7);
}
#define init_2_10_8_13
void FASTCALL call_2_10_8_13(SH2_struct* sh)
{
  SH2xtrct(sh,10,8);
}
#define init_2_10_9_13
void FASTCALL call_2_10_9_13(SH2_struct* sh)
{
  SH2xtrct(sh,10,9);
}
#define init_2_10_10_13
void FASTCALL call_2_10_10_13(SH2_struct* sh)
{
  SH2xtrct(sh,10,10);
}
#define init_2_10_11_13
void FASTCALL call_2_10_11_13(SH2_struct* sh)
{
  SH2xtrct(sh,10,11);
}
#define init_2_10_12_13
void FASTCALL call_2_10_12_13(SH2_struct* sh)
{
  SH2xtrct(sh,10,12);
}
#define init_2_10_13_13
void FASTCALL call_2_10_13_13(SH2_struct* sh)
{
  SH2xtrct(sh,10,13);
}
#define init_2_10_14_13
void FASTCALL call_2_10_14_13(SH2_struct* sh)
{
  SH2xtrct(sh,10,14);
}
#define init_2_10_15_13
void FASTCALL call_2_10_15_13(SH2_struct* sh)
{
  SH2xtrct(sh,10,15);
}
#define init_2_11_0_13
void FASTCALL call_2_11_0_13(SH2_struct* sh)
{
  SH2xtrct(sh,11,0);
}
#define init_2_11_1_13
void FASTCALL call_2_11_1_13(SH2_struct* sh)
{
  SH2xtrct(sh,11,1);
}
#define init_2_11_2_13
void FASTCALL call_2_11_2_13(SH2_struct* sh)
{
  SH2xtrct(sh,11,2);
}
#define init_2_11_3_13
void FASTCALL call_2_11_3_13(SH2_struct* sh)
{
  SH2xtrct(sh,11,3);
}
#define init_2_11_4_13
void FASTCALL call_2_11_4_13(SH2_struct* sh)
{
  SH2xtrct(sh,11,4);
}
#define init_2_11_5_13
void FASTCALL call_2_11_5_13(SH2_struct* sh)
{
  SH2xtrct(sh,11,5);
}
#define init_2_11_6_13
void FASTCALL call_2_11_6_13(SH2_struct* sh)
{
  SH2xtrct(sh,11,6);
}
#define init_2_11_7_13
void FASTCALL call_2_11_7_13(SH2_struct* sh)
{
  SH2xtrct(sh,11,7);
}
#define init_2_11_8_13
void FASTCALL call_2_11_8_13(SH2_struct* sh)
{
  SH2xtrct(sh,11,8);
}
#define init_2_11_9_13
void FASTCALL call_2_11_9_13(SH2_struct* sh)
{
  SH2xtrct(sh,11,9);
}
#define init_2_11_10_13
void FASTCALL call_2_11_10_13(SH2_struct* sh)
{
  SH2xtrct(sh,11,10);
}
#define init_2_11_11_13
void FASTCALL call_2_11_11_13(SH2_struct* sh)
{
  SH2xtrct(sh,11,11);
}
#define init_2_11_12_13
void FASTCALL call_2_11_12_13(SH2_struct* sh)
{
  SH2xtrct(sh,11,12);
}
#define init_2_11_13_13
void FASTCALL call_2_11_13_13(SH2_struct* sh)
{
  SH2xtrct(sh,11,13);
}
#define init_2_11_14_13
void FASTCALL call_2_11_14_13(SH2_struct* sh)
{
  SH2xtrct(sh,11,14);
}
#define init_2_11_15_13
void FASTCALL call_2_11_15_13(SH2_struct* sh)
{
  SH2xtrct(sh,11,15);
}
#define init_2_12_0_13
void FASTCALL call_2_12_0_13(SH2_struct* sh)
{
  SH2xtrct(sh,12,0);
}
#define init_2_12_1_13
void FASTCALL call_2_12_1_13(SH2_struct* sh)
{
  SH2xtrct(sh,12,1);
}
#define init_2_12_2_13
void FASTCALL call_2_12_2_13(SH2_struct* sh)
{
  SH2xtrct(sh,12,2);
}
#define init_2_12_3_13
void FASTCALL call_2_12_3_13(SH2_struct* sh)
{
  SH2xtrct(sh,12,3);
}
#define init_2_12_4_13
void FASTCALL call_2_12_4_13(SH2_struct* sh)
{
  SH2xtrct(sh,12,4);
}
#define init_2_12_5_13
void FASTCALL call_2_12_5_13(SH2_struct* sh)
{
  SH2xtrct(sh,12,5);
}
#define init_2_12_6_13
void FASTCALL call_2_12_6_13(SH2_struct* sh)
{
  SH2xtrct(sh,12,6);
}
#define init_2_12_7_13
void FASTCALL call_2_12_7_13(SH2_struct* sh)
{
  SH2xtrct(sh,12,7);
}
#define init_2_12_8_13
void FASTCALL call_2_12_8_13(SH2_struct* sh)
{
  SH2xtrct(sh,12,8);
}
#define init_2_12_9_13
void FASTCALL call_2_12_9_13(SH2_struct* sh)
{
  SH2xtrct(sh,12,9);
}
#define init_2_12_10_13
void FASTCALL call_2_12_10_13(SH2_struct* sh)
{
  SH2xtrct(sh,12,10);
}
#define init_2_12_11_13
void FASTCALL call_2_12_11_13(SH2_struct* sh)
{
  SH2xtrct(sh,12,11);
}
#define init_2_12_12_13
void FASTCALL call_2_12_12_13(SH2_struct* sh)
{
  SH2xtrct(sh,12,12);
}
#define init_2_12_13_13
void FASTCALL call_2_12_13_13(SH2_struct* sh)
{
  SH2xtrct(sh,12,13);
}
#define init_2_12_14_13
void FASTCALL call_2_12_14_13(SH2_struct* sh)
{
  SH2xtrct(sh,12,14);
}
#define init_2_12_15_13
void FASTCALL call_2_12_15_13(SH2_struct* sh)
{
  SH2xtrct(sh,12,15);
}
#define init_2_13_0_13
void FASTCALL call_2_13_0_13(SH2_struct* sh)
{
  SH2xtrct(sh,13,0);
}
#define init_2_13_1_13
void FASTCALL call_2_13_1_13(SH2_struct* sh)
{
  SH2xtrct(sh,13,1);
}
#define init_2_13_2_13
void FASTCALL call_2_13_2_13(SH2_struct* sh)
{
  SH2xtrct(sh,13,2);
}
#define init_2_13_3_13
void FASTCALL call_2_13_3_13(SH2_struct* sh)
{
  SH2xtrct(sh,13,3);
}
#define init_2_13_4_13
void FASTCALL call_2_13_4_13(SH2_struct* sh)
{
  SH2xtrct(sh,13,4);
}
#define init_2_13_5_13
void FASTCALL call_2_13_5_13(SH2_struct* sh)
{
  SH2xtrct(sh,13,5);
}
#define init_2_13_6_13
void FASTCALL call_2_13_6_13(SH2_struct* sh)
{
  SH2xtrct(sh,13,6);
}
#define init_2_13_7_13
void FASTCALL call_2_13_7_13(SH2_struct* sh)
{
  SH2xtrct(sh,13,7);
}
#define init_2_13_8_13
void FASTCALL call_2_13_8_13(SH2_struct* sh)
{
  SH2xtrct(sh,13,8);
}
#define init_2_13_9_13
void FASTCALL call_2_13_9_13(SH2_struct* sh)
{
  SH2xtrct(sh,13,9);
}
#define init_2_13_10_13
void FASTCALL call_2_13_10_13(SH2_struct* sh)
{
  SH2xtrct(sh,13,10);
}
#define init_2_13_11_13
void FASTCALL call_2_13_11_13(SH2_struct* sh)
{
  SH2xtrct(sh,13,11);
}
#define init_2_13_12_13
void FASTCALL call_2_13_12_13(SH2_struct* sh)
{
  SH2xtrct(sh,13,12);
}
#define init_2_13_13_13
void FASTCALL call_2_13_13_13(SH2_struct* sh)
{
  SH2xtrct(sh,13,13);
}
#define init_2_13_14_13
void FASTCALL call_2_13_14_13(SH2_struct* sh)
{
  SH2xtrct(sh,13,14);
}
#define init_2_13_15_13
void FASTCALL call_2_13_15_13(SH2_struct* sh)
{
  SH2xtrct(sh,13,15);
}
#define init_2_14_0_13
void FASTCALL call_2_14_0_13(SH2_struct* sh)
{
  SH2xtrct(sh,14,0);
}
#define init_2_14_1_13
void FASTCALL call_2_14_1_13(SH2_struct* sh)
{
  SH2xtrct(sh,14,1);
}
#define init_2_14_2_13
void FASTCALL call_2_14_2_13(SH2_struct* sh)
{
  SH2xtrct(sh,14,2);
}
#define init_2_14_3_13
void FASTCALL call_2_14_3_13(SH2_struct* sh)
{
  SH2xtrct(sh,14,3);
}
#define init_2_14_4_13
void FASTCALL call_2_14_4_13(SH2_struct* sh)
{
  SH2xtrct(sh,14,4);
}
#define init_2_14_5_13
void FASTCALL call_2_14_5_13(SH2_struct* sh)
{
  SH2xtrct(sh,14,5);
}
#define init_2_14_6_13
void FASTCALL call_2_14_6_13(SH2_struct* sh)
{
  SH2xtrct(sh,14,6);
}
#define init_2_14_7_13
void FASTCALL call_2_14_7_13(SH2_struct* sh)
{
  SH2xtrct(sh,14,7);
}
#define init_2_14_8_13
void FASTCALL call_2_14_8_13(SH2_struct* sh)
{
  SH2xtrct(sh,14,8);
}
#define init_2_14_9_13
void FASTCALL call_2_14_9_13(SH2_struct* sh)
{
  SH2xtrct(sh,14,9);
}
#define init_2_14_10_13
void FASTCALL call_2_14_10_13(SH2_struct* sh)
{
  SH2xtrct(sh,14,10);
}
#define init_2_14_11_13
void FASTCALL call_2_14_11_13(SH2_struct* sh)
{
  SH2xtrct(sh,14,11);
}
#define init_2_14_12_13
void FASTCALL call_2_14_12_13(SH2_struct* sh)
{
  SH2xtrct(sh,14,12);
}
#define init_2_14_13_13
void FASTCALL call_2_14_13_13(SH2_struct* sh)
{
  SH2xtrct(sh,14,13);
}
#define init_2_14_14_13
void FASTCALL call_2_14_14_13(SH2_struct* sh)
{
  SH2xtrct(sh,14,14);
}
#define init_2_14_15_13
void FASTCALL call_2_14_15_13(SH2_struct* sh)
{
  SH2xtrct(sh,14,15);
}
#define init_2_15_0_13
void FASTCALL call_2_15_0_13(SH2_struct* sh)
{
  SH2xtrct(sh,15,0);
}
#define init_2_15_1_13
void FASTCALL call_2_15_1_13(SH2_struct* sh)
{
  SH2xtrct(sh,15,1);
}
#define init_2_15_2_13
void FASTCALL call_2_15_2_13(SH2_struct* sh)
{
  SH2xtrct(sh,15,2);
}
#define init_2_15_3_13
void FASTCALL call_2_15_3_13(SH2_struct* sh)
{
  SH2xtrct(sh,15,3);
}
#define init_2_15_4_13
void FASTCALL call_2_15_4_13(SH2_struct* sh)
{
  SH2xtrct(sh,15,4);
}
#define init_2_15_5_13
void FASTCALL call_2_15_5_13(SH2_struct* sh)
{
  SH2xtrct(sh,15,5);
}
#define init_2_15_6_13
void FASTCALL call_2_15_6_13(SH2_struct* sh)
{
  SH2xtrct(sh,15,6);
}
#define init_2_15_7_13
void FASTCALL call_2_15_7_13(SH2_struct* sh)
{
  SH2xtrct(sh,15,7);
}
#define init_2_15_8_13
void FASTCALL call_2_15_8_13(SH2_struct* sh)
{
  SH2xtrct(sh,15,8);
}
#define init_2_15_9_13
void FASTCALL call_2_15_9_13(SH2_struct* sh)
{
  SH2xtrct(sh,15,9);
}
#define init_2_15_10_13
void FASTCALL call_2_15_10_13(SH2_struct* sh)
{
  SH2xtrct(sh,15,10);
}
#define init_2_15_11_13
void FASTCALL call_2_15_11_13(SH2_struct* sh)
{
  SH2xtrct(sh,15,11);
}
#define init_2_15_12_13
void FASTCALL call_2_15_12_13(SH2_struct* sh)
{
  SH2xtrct(sh,15,12);
}
#define init_2_15_13_13
void FASTCALL call_2_15_13_13(SH2_struct* sh)
{
  SH2xtrct(sh,15,13);
}
#define init_2_15_14_13
void FASTCALL call_2_15_14_13(SH2_struct* sh)
{
  SH2xtrct(sh,15,14);
}
#define init_2_15_15_13
void FASTCALL call_2_15_15_13(SH2_struct* sh)
{
  SH2xtrct(sh,15,15);
}
#define init_12_14_0_0
void FASTCALL call_12_14_0_0(SH2_struct* sh)
{
  SH2xorm(sh,(0<<4)+0);
}
#define init_12_14_0_1
void FASTCALL call_12_14_0_1(SH2_struct* sh)
{
  SH2xorm(sh,(0<<4)+1);
}
#define init_12_14_0_2
void FASTCALL call_12_14_0_2(SH2_struct* sh)
{
  SH2xorm(sh,(0<<4)+2);
}
#define init_12_14_0_3
void FASTCALL call_12_14_0_3(SH2_struct* sh)
{
  SH2xorm(sh,(0<<4)+3);
}
#define init_12_14_0_4
void FASTCALL call_12_14_0_4(SH2_struct* sh)
{
  SH2xorm(sh,(0<<4)+4);
}
#define init_12_14_0_5
void FASTCALL call_12_14_0_5(SH2_struct* sh)
{
  SH2xorm(sh,(0<<4)+5);
}
#define init_12_14_0_6
void FASTCALL call_12_14_0_6(SH2_struct* sh)
{
  SH2xorm(sh,(0<<4)+6);
}
#define init_12_14_0_7
void FASTCALL call_12_14_0_7(SH2_struct* sh)
{
  SH2xorm(sh,(0<<4)+7);
}
#define init_12_14_0_8
void FASTCALL call_12_14_0_8(SH2_struct* sh)
{
  SH2xorm(sh,(0<<4)+8);
}
#define init_12_14_0_9
void FASTCALL call_12_14_0_9(SH2_struct* sh)
{
  SH2xorm(sh,(0<<4)+9);
}
#define init_12_14_0_10
void FASTCALL call_12_14_0_10(SH2_struct* sh)
{
  SH2xorm(sh,(0<<4)+10);
}
#define init_12_14_0_11
void FASTCALL call_12_14_0_11(SH2_struct* sh)
{
  SH2xorm(sh,(0<<4)+11);
}
#define init_12_14_0_12
void FASTCALL call_12_14_0_12(SH2_struct* sh)
{
  SH2xorm(sh,(0<<4)+12);
}
#define init_12_14_0_13
void FASTCALL call_12_14_0_13(SH2_struct* sh)
{
  SH2xorm(sh,(0<<4)+13);
}
#define init_12_14_0_14
void FASTCALL call_12_14_0_14(SH2_struct* sh)
{
  SH2xorm(sh,(0<<4)+14);
}
#define init_12_14_0_15
void FASTCALL call_12_14_0_15(SH2_struct* sh)
{
  SH2xorm(sh,(0<<4)+15);
}
#define init_12_14_1_0
void FASTCALL call_12_14_1_0(SH2_struct* sh)
{
  SH2xorm(sh,(1<<4)+0);
}
#define init_12_14_1_1
void FASTCALL call_12_14_1_1(SH2_struct* sh)
{
  SH2xorm(sh,(1<<4)+1);
}
#define init_12_14_1_2
void FASTCALL call_12_14_1_2(SH2_struct* sh)
{
  SH2xorm(sh,(1<<4)+2);
}
#define init_12_14_1_3
void FASTCALL call_12_14_1_3(SH2_struct* sh)
{
  SH2xorm(sh,(1<<4)+3);
}
#define init_12_14_1_4
void FASTCALL call_12_14_1_4(SH2_struct* sh)
{
  SH2xorm(sh,(1<<4)+4);
}
#define init_12_14_1_5
void FASTCALL call_12_14_1_5(SH2_struct* sh)
{
  SH2xorm(sh,(1<<4)+5);
}
#define init_12_14_1_6
void FASTCALL call_12_14_1_6(SH2_struct* sh)
{
  SH2xorm(sh,(1<<4)+6);
}
#define init_12_14_1_7
void FASTCALL call_12_14_1_7(SH2_struct* sh)
{
  SH2xorm(sh,(1<<4)+7);
}
#define init_12_14_1_8
void FASTCALL call_12_14_1_8(SH2_struct* sh)
{
  SH2xorm(sh,(1<<4)+8);
}
#define init_12_14_1_9
void FASTCALL call_12_14_1_9(SH2_struct* sh)
{
  SH2xorm(sh,(1<<4)+9);
}
#define init_12_14_1_10
void FASTCALL call_12_14_1_10(SH2_struct* sh)
{
  SH2xorm(sh,(1<<4)+10);
}
#define init_12_14_1_11
void FASTCALL call_12_14_1_11(SH2_struct* sh)
{
  SH2xorm(sh,(1<<4)+11);
}
#define init_12_14_1_12
void FASTCALL call_12_14_1_12(SH2_struct* sh)
{
  SH2xorm(sh,(1<<4)+12);
}
#define init_12_14_1_13
void FASTCALL call_12_14_1_13(SH2_struct* sh)
{
  SH2xorm(sh,(1<<4)+13);
}
#define init_12_14_1_14
void FASTCALL call_12_14_1_14(SH2_struct* sh)
{
  SH2xorm(sh,(1<<4)+14);
}
#define init_12_14_1_15
void FASTCALL call_12_14_1_15(SH2_struct* sh)
{
  SH2xorm(sh,(1<<4)+15);
}
#define init_12_14_2_0
void FASTCALL call_12_14_2_0(SH2_struct* sh)
{
  SH2xorm(sh,(2<<4)+0);
}
#define init_12_14_2_1
void FASTCALL call_12_14_2_1(SH2_struct* sh)
{
  SH2xorm(sh,(2<<4)+1);
}
#define init_12_14_2_2
void FASTCALL call_12_14_2_2(SH2_struct* sh)
{
  SH2xorm(sh,(2<<4)+2);
}
#define init_12_14_2_3
void FASTCALL call_12_14_2_3(SH2_struct* sh)
{
  SH2xorm(sh,(2<<4)+3);
}
#define init_12_14_2_4
void FASTCALL call_12_14_2_4(SH2_struct* sh)
{
  SH2xorm(sh,(2<<4)+4);
}
#define init_12_14_2_5
void FASTCALL call_12_14_2_5(SH2_struct* sh)
{
  SH2xorm(sh,(2<<4)+5);
}
#define init_12_14_2_6
void FASTCALL call_12_14_2_6(SH2_struct* sh)
{
  SH2xorm(sh,(2<<4)+6);
}
#define init_12_14_2_7
void FASTCALL call_12_14_2_7(SH2_struct* sh)
{
  SH2xorm(sh,(2<<4)+7);
}
#define init_12_14_2_8
void FASTCALL call_12_14_2_8(SH2_struct* sh)
{
  SH2xorm(sh,(2<<4)+8);
}
#define init_12_14_2_9
void FASTCALL call_12_14_2_9(SH2_struct* sh)
{
  SH2xorm(sh,(2<<4)+9);
}
#define init_12_14_2_10
void FASTCALL call_12_14_2_10(SH2_struct* sh)
{
  SH2xorm(sh,(2<<4)+10);
}
#define init_12_14_2_11
void FASTCALL call_12_14_2_11(SH2_struct* sh)
{
  SH2xorm(sh,(2<<4)+11);
}
#define init_12_14_2_12
void FASTCALL call_12_14_2_12(SH2_struct* sh)
{
  SH2xorm(sh,(2<<4)+12);
}
#define init_12_14_2_13
void FASTCALL call_12_14_2_13(SH2_struct* sh)
{
  SH2xorm(sh,(2<<4)+13);
}
#define init_12_14_2_14
void FASTCALL call_12_14_2_14(SH2_struct* sh)
{
  SH2xorm(sh,(2<<4)+14);
}
#define init_12_14_2_15
void FASTCALL call_12_14_2_15(SH2_struct* sh)
{
  SH2xorm(sh,(2<<4)+15);
}
#define init_12_14_3_0
void FASTCALL call_12_14_3_0(SH2_struct* sh)
{
  SH2xorm(sh,(3<<4)+0);
}
#define init_12_14_3_1
void FASTCALL call_12_14_3_1(SH2_struct* sh)
{
  SH2xorm(sh,(3<<4)+1);
}
#define init_12_14_3_2
void FASTCALL call_12_14_3_2(SH2_struct* sh)
{
  SH2xorm(sh,(3<<4)+2);
}
#define init_12_14_3_3
void FASTCALL call_12_14_3_3(SH2_struct* sh)
{
  SH2xorm(sh,(3<<4)+3);
}
#define init_12_14_3_4
void FASTCALL call_12_14_3_4(SH2_struct* sh)
{
  SH2xorm(sh,(3<<4)+4);
}
#define init_12_14_3_5
void FASTCALL call_12_14_3_5(SH2_struct* sh)
{
  SH2xorm(sh,(3<<4)+5);
}
#define init_12_14_3_6
void FASTCALL call_12_14_3_6(SH2_struct* sh)
{
  SH2xorm(sh,(3<<4)+6);
}
#define init_12_14_3_7
void FASTCALL call_12_14_3_7(SH2_struct* sh)
{
  SH2xorm(sh,(3<<4)+7);
}
#define init_12_14_3_8
void FASTCALL call_12_14_3_8(SH2_struct* sh)
{
  SH2xorm(sh,(3<<4)+8);
}
#define init_12_14_3_9
void FASTCALL call_12_14_3_9(SH2_struct* sh)
{
  SH2xorm(sh,(3<<4)+9);
}
#define init_12_14_3_10
void FASTCALL call_12_14_3_10(SH2_struct* sh)
{
  SH2xorm(sh,(3<<4)+10);
}
#define init_12_14_3_11
void FASTCALL call_12_14_3_11(SH2_struct* sh)
{
  SH2xorm(sh,(3<<4)+11);
}
#define init_12_14_3_12
void FASTCALL call_12_14_3_12(SH2_struct* sh)
{
  SH2xorm(sh,(3<<4)+12);
}
#define init_12_14_3_13
void FASTCALL call_12_14_3_13(SH2_struct* sh)
{
  SH2xorm(sh,(3<<4)+13);
}
#define init_12_14_3_14
void FASTCALL call_12_14_3_14(SH2_struct* sh)
{
  SH2xorm(sh,(3<<4)+14);
}
#define init_12_14_3_15
void FASTCALL call_12_14_3_15(SH2_struct* sh)
{
  SH2xorm(sh,(3<<4)+15);
}
#define init_12_14_4_0
void FASTCALL call_12_14_4_0(SH2_struct* sh)
{
  SH2xorm(sh,(4<<4)+0);
}
#define init_12_14_4_1
void FASTCALL call_12_14_4_1(SH2_struct* sh)
{
  SH2xorm(sh,(4<<4)+1);
}
#define init_12_14_4_2
void FASTCALL call_12_14_4_2(SH2_struct* sh)
{
  SH2xorm(sh,(4<<4)+2);
}
#define init_12_14_4_3
void FASTCALL call_12_14_4_3(SH2_struct* sh)
{
  SH2xorm(sh,(4<<4)+3);
}
#define init_12_14_4_4
void FASTCALL call_12_14_4_4(SH2_struct* sh)
{
  SH2xorm(sh,(4<<4)+4);
}
#define init_12_14_4_5
void FASTCALL call_12_14_4_5(SH2_struct* sh)
{
  SH2xorm(sh,(4<<4)+5);
}
#define init_12_14_4_6
void FASTCALL call_12_14_4_6(SH2_struct* sh)
{
  SH2xorm(sh,(4<<4)+6);
}
#define init_12_14_4_7
void FASTCALL call_12_14_4_7(SH2_struct* sh)
{
  SH2xorm(sh,(4<<4)+7);
}
#define init_12_14_4_8
void FASTCALL call_12_14_4_8(SH2_struct* sh)
{
  SH2xorm(sh,(4<<4)+8);
}
#define init_12_14_4_9
void FASTCALL call_12_14_4_9(SH2_struct* sh)
{
  SH2xorm(sh,(4<<4)+9);
}
#define init_12_14_4_10
void FASTCALL call_12_14_4_10(SH2_struct* sh)
{
  SH2xorm(sh,(4<<4)+10);
}
#define init_12_14_4_11
void FASTCALL call_12_14_4_11(SH2_struct* sh)
{
  SH2xorm(sh,(4<<4)+11);
}
#define init_12_14_4_12
void FASTCALL call_12_14_4_12(SH2_struct* sh)
{
  SH2xorm(sh,(4<<4)+12);
}
#define init_12_14_4_13
void FASTCALL call_12_14_4_13(SH2_struct* sh)
{
  SH2xorm(sh,(4<<4)+13);
}
#define init_12_14_4_14
void FASTCALL call_12_14_4_14(SH2_struct* sh)
{
  SH2xorm(sh,(4<<4)+14);
}
#define init_12_14_4_15
void FASTCALL call_12_14_4_15(SH2_struct* sh)
{
  SH2xorm(sh,(4<<4)+15);
}
#define init_12_14_5_0
void FASTCALL call_12_14_5_0(SH2_struct* sh)
{
  SH2xorm(sh,(5<<4)+0);
}
#define init_12_14_5_1
void FASTCALL call_12_14_5_1(SH2_struct* sh)
{
  SH2xorm(sh,(5<<4)+1);
}
#define init_12_14_5_2
void FASTCALL call_12_14_5_2(SH2_struct* sh)
{
  SH2xorm(sh,(5<<4)+2);
}
#define init_12_14_5_3
void FASTCALL call_12_14_5_3(SH2_struct* sh)
{
  SH2xorm(sh,(5<<4)+3);
}
#define init_12_14_5_4
void FASTCALL call_12_14_5_4(SH2_struct* sh)
{
  SH2xorm(sh,(5<<4)+4);
}
#define init_12_14_5_5
void FASTCALL call_12_14_5_5(SH2_struct* sh)
{
  SH2xorm(sh,(5<<4)+5);
}
#define init_12_14_5_6
void FASTCALL call_12_14_5_6(SH2_struct* sh)
{
  SH2xorm(sh,(5<<4)+6);
}
#define init_12_14_5_7
void FASTCALL call_12_14_5_7(SH2_struct* sh)
{
  SH2xorm(sh,(5<<4)+7);
}
#define init_12_14_5_8
void FASTCALL call_12_14_5_8(SH2_struct* sh)
{
  SH2xorm(sh,(5<<4)+8);
}
#define init_12_14_5_9
void FASTCALL call_12_14_5_9(SH2_struct* sh)
{
  SH2xorm(sh,(5<<4)+9);
}
#define init_12_14_5_10
void FASTCALL call_12_14_5_10(SH2_struct* sh)
{
  SH2xorm(sh,(5<<4)+10);
}
#define init_12_14_5_11
void FASTCALL call_12_14_5_11(SH2_struct* sh)
{
  SH2xorm(sh,(5<<4)+11);
}
#define init_12_14_5_12
void FASTCALL call_12_14_5_12(SH2_struct* sh)
{
  SH2xorm(sh,(5<<4)+12);
}
#define init_12_14_5_13
void FASTCALL call_12_14_5_13(SH2_struct* sh)
{
  SH2xorm(sh,(5<<4)+13);
}
#define init_12_14_5_14
void FASTCALL call_12_14_5_14(SH2_struct* sh)
{
  SH2xorm(sh,(5<<4)+14);
}
#define init_12_14_5_15
void FASTCALL call_12_14_5_15(SH2_struct* sh)
{
  SH2xorm(sh,(5<<4)+15);
}
#define init_12_14_6_0
void FASTCALL call_12_14_6_0(SH2_struct* sh)
{
  SH2xorm(sh,(6<<4)+0);
}
#define init_12_14_6_1
void FASTCALL call_12_14_6_1(SH2_struct* sh)
{
  SH2xorm(sh,(6<<4)+1);
}
#define init_12_14_6_2
void FASTCALL call_12_14_6_2(SH2_struct* sh)
{
  SH2xorm(sh,(6<<4)+2);
}
#define init_12_14_6_3
void FASTCALL call_12_14_6_3(SH2_struct* sh)
{
  SH2xorm(sh,(6<<4)+3);
}
#define init_12_14_6_4
void FASTCALL call_12_14_6_4(SH2_struct* sh)
{
  SH2xorm(sh,(6<<4)+4);
}
#define init_12_14_6_5
void FASTCALL call_12_14_6_5(SH2_struct* sh)
{
  SH2xorm(sh,(6<<4)+5);
}
#define init_12_14_6_6
void FASTCALL call_12_14_6_6(SH2_struct* sh)
{
  SH2xorm(sh,(6<<4)+6);
}
#define init_12_14_6_7
void FASTCALL call_12_14_6_7(SH2_struct* sh)
{
  SH2xorm(sh,(6<<4)+7);
}
#define init_12_14_6_8
void FASTCALL call_12_14_6_8(SH2_struct* sh)
{
  SH2xorm(sh,(6<<4)+8);
}
#define init_12_14_6_9
void FASTCALL call_12_14_6_9(SH2_struct* sh)
{
  SH2xorm(sh,(6<<4)+9);
}
#define init_12_14_6_10
void FASTCALL call_12_14_6_10(SH2_struct* sh)
{
  SH2xorm(sh,(6<<4)+10);
}
#define init_12_14_6_11
void FASTCALL call_12_14_6_11(SH2_struct* sh)
{
  SH2xorm(sh,(6<<4)+11);
}
#define init_12_14_6_12
void FASTCALL call_12_14_6_12(SH2_struct* sh)
{
  SH2xorm(sh,(6<<4)+12);
}
#define init_12_14_6_13
void FASTCALL call_12_14_6_13(SH2_struct* sh)
{
  SH2xorm(sh,(6<<4)+13);
}
#define init_12_14_6_14
void FASTCALL call_12_14_6_14(SH2_struct* sh)
{
  SH2xorm(sh,(6<<4)+14);
}
#define init_12_14_6_15
void FASTCALL call_12_14_6_15(SH2_struct* sh)
{
  SH2xorm(sh,(6<<4)+15);
}
#define init_12_14_7_0
void FASTCALL call_12_14_7_0(SH2_struct* sh)
{
  SH2xorm(sh,(7<<4)+0);
}
#define init_12_14_7_1
void FASTCALL call_12_14_7_1(SH2_struct* sh)
{
  SH2xorm(sh,(7<<4)+1);
}
#define init_12_14_7_2
void FASTCALL call_12_14_7_2(SH2_struct* sh)
{
  SH2xorm(sh,(7<<4)+2);
}
#define init_12_14_7_3
void FASTCALL call_12_14_7_3(SH2_struct* sh)
{
  SH2xorm(sh,(7<<4)+3);
}
#define init_12_14_7_4
void FASTCALL call_12_14_7_4(SH2_struct* sh)
{
  SH2xorm(sh,(7<<4)+4);
}
#define init_12_14_7_5
void FASTCALL call_12_14_7_5(SH2_struct* sh)
{
  SH2xorm(sh,(7<<4)+5);
}
#define init_12_14_7_6
void FASTCALL call_12_14_7_6(SH2_struct* sh)
{
  SH2xorm(sh,(7<<4)+6);
}
#define init_12_14_7_7
void FASTCALL call_12_14_7_7(SH2_struct* sh)
{
  SH2xorm(sh,(7<<4)+7);
}
#define init_12_14_7_8
void FASTCALL call_12_14_7_8(SH2_struct* sh)
{
  SH2xorm(sh,(7<<4)+8);
}
#define init_12_14_7_9
void FASTCALL call_12_14_7_9(SH2_struct* sh)
{
  SH2xorm(sh,(7<<4)+9);
}
#define init_12_14_7_10
void FASTCALL call_12_14_7_10(SH2_struct* sh)
{
  SH2xorm(sh,(7<<4)+10);
}
#define init_12_14_7_11
void FASTCALL call_12_14_7_11(SH2_struct* sh)
{
  SH2xorm(sh,(7<<4)+11);
}
#define init_12_14_7_12
void FASTCALL call_12_14_7_12(SH2_struct* sh)
{
  SH2xorm(sh,(7<<4)+12);
}
#define init_12_14_7_13
void FASTCALL call_12_14_7_13(SH2_struct* sh)
{
  SH2xorm(sh,(7<<4)+13);
}
#define init_12_14_7_14
void FASTCALL call_12_14_7_14(SH2_struct* sh)
{
  SH2xorm(sh,(7<<4)+14);
}
#define init_12_14_7_15
void FASTCALL call_12_14_7_15(SH2_struct* sh)
{
  SH2xorm(sh,(7<<4)+15);
}
#define init_12_14_8_0
void FASTCALL call_12_14_8_0(SH2_struct* sh)
{
  SH2xorm(sh,(8<<4)+0);
}
#define init_12_14_8_1
void FASTCALL call_12_14_8_1(SH2_struct* sh)
{
  SH2xorm(sh,(8<<4)+1);
}
#define init_12_14_8_2
void FASTCALL call_12_14_8_2(SH2_struct* sh)
{
  SH2xorm(sh,(8<<4)+2);
}
#define init_12_14_8_3
void FASTCALL call_12_14_8_3(SH2_struct* sh)
{
  SH2xorm(sh,(8<<4)+3);
}
#define init_12_14_8_4
void FASTCALL call_12_14_8_4(SH2_struct* sh)
{
  SH2xorm(sh,(8<<4)+4);
}
#define init_12_14_8_5
void FASTCALL call_12_14_8_5(SH2_struct* sh)
{
  SH2xorm(sh,(8<<4)+5);
}
#define init_12_14_8_6
void FASTCALL call_12_14_8_6(SH2_struct* sh)
{
  SH2xorm(sh,(8<<4)+6);
}
#define init_12_14_8_7
void FASTCALL call_12_14_8_7(SH2_struct* sh)
{
  SH2xorm(sh,(8<<4)+7);
}
#define init_12_14_8_8
void FASTCALL call_12_14_8_8(SH2_struct* sh)
{
  SH2xorm(sh,(8<<4)+8);
}
#define init_12_14_8_9
void FASTCALL call_12_14_8_9(SH2_struct* sh)
{
  SH2xorm(sh,(8<<4)+9);
}
#define init_12_14_8_10
void FASTCALL call_12_14_8_10(SH2_struct* sh)
{
  SH2xorm(sh,(8<<4)+10);
}
#define init_12_14_8_11
void FASTCALL call_12_14_8_11(SH2_struct* sh)
{
  SH2xorm(sh,(8<<4)+11);
}
#define init_12_14_8_12
void FASTCALL call_12_14_8_12(SH2_struct* sh)
{
  SH2xorm(sh,(8<<4)+12);
}
#define init_12_14_8_13
void FASTCALL call_12_14_8_13(SH2_struct* sh)
{
  SH2xorm(sh,(8<<4)+13);
}
#define init_12_14_8_14
void FASTCALL call_12_14_8_14(SH2_struct* sh)
{
  SH2xorm(sh,(8<<4)+14);
}
#define init_12_14_8_15
void FASTCALL call_12_14_8_15(SH2_struct* sh)
{
  SH2xorm(sh,(8<<4)+15);
}
#define init_12_14_9_0
void FASTCALL call_12_14_9_0(SH2_struct* sh)
{
  SH2xorm(sh,(9<<4)+0);
}
#define init_12_14_9_1
void FASTCALL call_12_14_9_1(SH2_struct* sh)
{
  SH2xorm(sh,(9<<4)+1);
}
#define init_12_14_9_2
void FASTCALL call_12_14_9_2(SH2_struct* sh)
{
  SH2xorm(sh,(9<<4)+2);
}
#define init_12_14_9_3
void FASTCALL call_12_14_9_3(SH2_struct* sh)
{
  SH2xorm(sh,(9<<4)+3);
}
#define init_12_14_9_4
void FASTCALL call_12_14_9_4(SH2_struct* sh)
{
  SH2xorm(sh,(9<<4)+4);
}
#define init_12_14_9_5
void FASTCALL call_12_14_9_5(SH2_struct* sh)
{
  SH2xorm(sh,(9<<4)+5);
}
#define init_12_14_9_6
void FASTCALL call_12_14_9_6(SH2_struct* sh)
{
  SH2xorm(sh,(9<<4)+6);
}
#define init_12_14_9_7
void FASTCALL call_12_14_9_7(SH2_struct* sh)
{
  SH2xorm(sh,(9<<4)+7);
}
#define init_12_14_9_8
void FASTCALL call_12_14_9_8(SH2_struct* sh)
{
  SH2xorm(sh,(9<<4)+8);
}
#define init_12_14_9_9
void FASTCALL call_12_14_9_9(SH2_struct* sh)
{
  SH2xorm(sh,(9<<4)+9);
}
#define init_12_14_9_10
void FASTCALL call_12_14_9_10(SH2_struct* sh)
{
  SH2xorm(sh,(9<<4)+10);
}
#define init_12_14_9_11
void FASTCALL call_12_14_9_11(SH2_struct* sh)
{
  SH2xorm(sh,(9<<4)+11);
}
#define init_12_14_9_12
void FASTCALL call_12_14_9_12(SH2_struct* sh)
{
  SH2xorm(sh,(9<<4)+12);
}
#define init_12_14_9_13
void FASTCALL call_12_14_9_13(SH2_struct* sh)
{
  SH2xorm(sh,(9<<4)+13);
}
#define init_12_14_9_14
void FASTCALL call_12_14_9_14(SH2_struct* sh)
{
  SH2xorm(sh,(9<<4)+14);
}
#define init_12_14_9_15
void FASTCALL call_12_14_9_15(SH2_struct* sh)
{
  SH2xorm(sh,(9<<4)+15);
}
#define init_12_14_10_0
void FASTCALL call_12_14_10_0(SH2_struct* sh)
{
  SH2xorm(sh,(10<<4)+0);
}
#define init_12_14_10_1
void FASTCALL call_12_14_10_1(SH2_struct* sh)
{
  SH2xorm(sh,(10<<4)+1);
}
#define init_12_14_10_2
void FASTCALL call_12_14_10_2(SH2_struct* sh)
{
  SH2xorm(sh,(10<<4)+2);
}
#define init_12_14_10_3
void FASTCALL call_12_14_10_3(SH2_struct* sh)
{
  SH2xorm(sh,(10<<4)+3);
}
#define init_12_14_10_4
void FASTCALL call_12_14_10_4(SH2_struct* sh)
{
  SH2xorm(sh,(10<<4)+4);
}
#define init_12_14_10_5
void FASTCALL call_12_14_10_5(SH2_struct* sh)
{
  SH2xorm(sh,(10<<4)+5);
}
#define init_12_14_10_6
void FASTCALL call_12_14_10_6(SH2_struct* sh)
{
  SH2xorm(sh,(10<<4)+6);
}
#define init_12_14_10_7
void FASTCALL call_12_14_10_7(SH2_struct* sh)
{
  SH2xorm(sh,(10<<4)+7);
}
#define init_12_14_10_8
void FASTCALL call_12_14_10_8(SH2_struct* sh)
{
  SH2xorm(sh,(10<<4)+8);
}
#define init_12_14_10_9
void FASTCALL call_12_14_10_9(SH2_struct* sh)
{
  SH2xorm(sh,(10<<4)+9);
}
#define init_12_14_10_10
void FASTCALL call_12_14_10_10(SH2_struct* sh)
{
  SH2xorm(sh,(10<<4)+10);
}
#define init_12_14_10_11
void FASTCALL call_12_14_10_11(SH2_struct* sh)
{
  SH2xorm(sh,(10<<4)+11);
}
#define init_12_14_10_12
void FASTCALL call_12_14_10_12(SH2_struct* sh)
{
  SH2xorm(sh,(10<<4)+12);
}
#define init_12_14_10_13
void FASTCALL call_12_14_10_13(SH2_struct* sh)
{
  SH2xorm(sh,(10<<4)+13);
}
#define init_12_14_10_14
void FASTCALL call_12_14_10_14(SH2_struct* sh)
{
  SH2xorm(sh,(10<<4)+14);
}
#define init_12_14_10_15
void FASTCALL call_12_14_10_15(SH2_struct* sh)
{
  SH2xorm(sh,(10<<4)+15);
}
#define init_12_14_11_0
void FASTCALL call_12_14_11_0(SH2_struct* sh)
{
  SH2xorm(sh,(11<<4)+0);
}
#define init_12_14_11_1
void FASTCALL call_12_14_11_1(SH2_struct* sh)
{
  SH2xorm(sh,(11<<4)+1);
}
#define init_12_14_11_2
void FASTCALL call_12_14_11_2(SH2_struct* sh)
{
  SH2xorm(sh,(11<<4)+2);
}
#define init_12_14_11_3
void FASTCALL call_12_14_11_3(SH2_struct* sh)
{
  SH2xorm(sh,(11<<4)+3);
}
#define init_12_14_11_4
void FASTCALL call_12_14_11_4(SH2_struct* sh)
{
  SH2xorm(sh,(11<<4)+4);
}
#define init_12_14_11_5
void FASTCALL call_12_14_11_5(SH2_struct* sh)
{
  SH2xorm(sh,(11<<4)+5);
}
#define init_12_14_11_6
void FASTCALL call_12_14_11_6(SH2_struct* sh)
{
  SH2xorm(sh,(11<<4)+6);
}
#define init_12_14_11_7
void FASTCALL call_12_14_11_7(SH2_struct* sh)
{
  SH2xorm(sh,(11<<4)+7);
}
#define init_12_14_11_8
void FASTCALL call_12_14_11_8(SH2_struct* sh)
{
  SH2xorm(sh,(11<<4)+8);
}
#define init_12_14_11_9
void FASTCALL call_12_14_11_9(SH2_struct* sh)
{
  SH2xorm(sh,(11<<4)+9);
}
#define init_12_14_11_10
void FASTCALL call_12_14_11_10(SH2_struct* sh)
{
  SH2xorm(sh,(11<<4)+10);
}
#define init_12_14_11_11
void FASTCALL call_12_14_11_11(SH2_struct* sh)
{
  SH2xorm(sh,(11<<4)+11);
}
#define init_12_14_11_12
void FASTCALL call_12_14_11_12(SH2_struct* sh)
{
  SH2xorm(sh,(11<<4)+12);
}
#define init_12_14_11_13
void FASTCALL call_12_14_11_13(SH2_struct* sh)
{
  SH2xorm(sh,(11<<4)+13);
}
#define init_12_14_11_14
void FASTCALL call_12_14_11_14(SH2_struct* sh)
{
  SH2xorm(sh,(11<<4)+14);
}
#define init_12_14_11_15
void FASTCALL call_12_14_11_15(SH2_struct* sh)
{
  SH2xorm(sh,(11<<4)+15);
}
#define init_12_14_12_0
void FASTCALL call_12_14_12_0(SH2_struct* sh)
{
  SH2xorm(sh,(12<<4)+0);
}
#define init_12_14_12_1
void FASTCALL call_12_14_12_1(SH2_struct* sh)
{
  SH2xorm(sh,(12<<4)+1);
}
#define init_12_14_12_2
void FASTCALL call_12_14_12_2(SH2_struct* sh)
{
  SH2xorm(sh,(12<<4)+2);
}
#define init_12_14_12_3
void FASTCALL call_12_14_12_3(SH2_struct* sh)
{
  SH2xorm(sh,(12<<4)+3);
}
#define init_12_14_12_4
void FASTCALL call_12_14_12_4(SH2_struct* sh)
{
  SH2xorm(sh,(12<<4)+4);
}
#define init_12_14_12_5
void FASTCALL call_12_14_12_5(SH2_struct* sh)
{
  SH2xorm(sh,(12<<4)+5);
}
#define init_12_14_12_6
void FASTCALL call_12_14_12_6(SH2_struct* sh)
{
  SH2xorm(sh,(12<<4)+6);
}
#define init_12_14_12_7
void FASTCALL call_12_14_12_7(SH2_struct* sh)
{
  SH2xorm(sh,(12<<4)+7);
}
#define init_12_14_12_8
void FASTCALL call_12_14_12_8(SH2_struct* sh)
{
  SH2xorm(sh,(12<<4)+8);
}
#define init_12_14_12_9
void FASTCALL call_12_14_12_9(SH2_struct* sh)
{
  SH2xorm(sh,(12<<4)+9);
}
#define init_12_14_12_10
void FASTCALL call_12_14_12_10(SH2_struct* sh)
{
  SH2xorm(sh,(12<<4)+10);
}
#define init_12_14_12_11
void FASTCALL call_12_14_12_11(SH2_struct* sh)
{
  SH2xorm(sh,(12<<4)+11);
}
#define init_12_14_12_12
void FASTCALL call_12_14_12_12(SH2_struct* sh)
{
  SH2xorm(sh,(12<<4)+12);
}
#define init_12_14_12_13
void FASTCALL call_12_14_12_13(SH2_struct* sh)
{
  SH2xorm(sh,(12<<4)+13);
}
#define init_12_14_12_14
void FASTCALL call_12_14_12_14(SH2_struct* sh)
{
  SH2xorm(sh,(12<<4)+14);
}
#define init_12_14_12_15
void FASTCALL call_12_14_12_15(SH2_struct* sh)
{
  SH2xorm(sh,(12<<4)+15);
}
#define init_12_14_13_0
void FASTCALL call_12_14_13_0(SH2_struct* sh)
{
  SH2xorm(sh,(13<<4)+0);
}
#define init_12_14_13_1
void FASTCALL call_12_14_13_1(SH2_struct* sh)
{
  SH2xorm(sh,(13<<4)+1);
}
#define init_12_14_13_2
void FASTCALL call_12_14_13_2(SH2_struct* sh)
{
  SH2xorm(sh,(13<<4)+2);
}
#define init_12_14_13_3
void FASTCALL call_12_14_13_3(SH2_struct* sh)
{
  SH2xorm(sh,(13<<4)+3);
}
#define init_12_14_13_4
void FASTCALL call_12_14_13_4(SH2_struct* sh)
{
  SH2xorm(sh,(13<<4)+4);
}
#define init_12_14_13_5
void FASTCALL call_12_14_13_5(SH2_struct* sh)
{
  SH2xorm(sh,(13<<4)+5);
}
#define init_12_14_13_6
void FASTCALL call_12_14_13_6(SH2_struct* sh)
{
  SH2xorm(sh,(13<<4)+6);
}
#define init_12_14_13_7
void FASTCALL call_12_14_13_7(SH2_struct* sh)
{
  SH2xorm(sh,(13<<4)+7);
}
#define init_12_14_13_8
void FASTCALL call_12_14_13_8(SH2_struct* sh)
{
  SH2xorm(sh,(13<<4)+8);
}
#define init_12_14_13_9
void FASTCALL call_12_14_13_9(SH2_struct* sh)
{
  SH2xorm(sh,(13<<4)+9);
}
#define init_12_14_13_10
void FASTCALL call_12_14_13_10(SH2_struct* sh)
{
  SH2xorm(sh,(13<<4)+10);
}
#define init_12_14_13_11
void FASTCALL call_12_14_13_11(SH2_struct* sh)
{
  SH2xorm(sh,(13<<4)+11);
}
#define init_12_14_13_12
void FASTCALL call_12_14_13_12(SH2_struct* sh)
{
  SH2xorm(sh,(13<<4)+12);
}
#define init_12_14_13_13
void FASTCALL call_12_14_13_13(SH2_struct* sh)
{
  SH2xorm(sh,(13<<4)+13);
}
#define init_12_14_13_14
void FASTCALL call_12_14_13_14(SH2_struct* sh)
{
  SH2xorm(sh,(13<<4)+14);
}
#define init_12_14_13_15
void FASTCALL call_12_14_13_15(SH2_struct* sh)
{
  SH2xorm(sh,(13<<4)+15);
}
#define init_12_14_14_0
void FASTCALL call_12_14_14_0(SH2_struct* sh)
{
  SH2xorm(sh,(14<<4)+0);
}
#define init_12_14_14_1
void FASTCALL call_12_14_14_1(SH2_struct* sh)
{
  SH2xorm(sh,(14<<4)+1);
}
#define init_12_14_14_2
void FASTCALL call_12_14_14_2(SH2_struct* sh)
{
  SH2xorm(sh,(14<<4)+2);
}
#define init_12_14_14_3
void FASTCALL call_12_14_14_3(SH2_struct* sh)
{
  SH2xorm(sh,(14<<4)+3);
}
#define init_12_14_14_4
void FASTCALL call_12_14_14_4(SH2_struct* sh)
{
  SH2xorm(sh,(14<<4)+4);
}
#define init_12_14_14_5
void FASTCALL call_12_14_14_5(SH2_struct* sh)
{
  SH2xorm(sh,(14<<4)+5);
}
#define init_12_14_14_6
void FASTCALL call_12_14_14_6(SH2_struct* sh)
{
  SH2xorm(sh,(14<<4)+6);
}
#define init_12_14_14_7
void FASTCALL call_12_14_14_7(SH2_struct* sh)
{
  SH2xorm(sh,(14<<4)+7);
}
#define init_12_14_14_8
void FASTCALL call_12_14_14_8(SH2_struct* sh)
{
  SH2xorm(sh,(14<<4)+8);
}
#define init_12_14_14_9
void FASTCALL call_12_14_14_9(SH2_struct* sh)
{
  SH2xorm(sh,(14<<4)+9);
}
#define init_12_14_14_10
void FASTCALL call_12_14_14_10(SH2_struct* sh)
{
  SH2xorm(sh,(14<<4)+10);
}
#define init_12_14_14_11
void FASTCALL call_12_14_14_11(SH2_struct* sh)
{
  SH2xorm(sh,(14<<4)+11);
}
#define init_12_14_14_12
void FASTCALL call_12_14_14_12(SH2_struct* sh)
{
  SH2xorm(sh,(14<<4)+12);
}
#define init_12_14_14_13
void FASTCALL call_12_14_14_13(SH2_struct* sh)
{
  SH2xorm(sh,(14<<4)+13);
}
#define init_12_14_14_14
void FASTCALL call_12_14_14_14(SH2_struct* sh)
{
  SH2xorm(sh,(14<<4)+14);
}
#define init_12_14_14_15
void FASTCALL call_12_14_14_15(SH2_struct* sh)
{
  SH2xorm(sh,(14<<4)+15);
}
#define init_12_14_15_0
void FASTCALL call_12_14_15_0(SH2_struct* sh)
{
  SH2xorm(sh,(15<<4)+0);
}
#define init_12_14_15_1
void FASTCALL call_12_14_15_1(SH2_struct* sh)
{
  SH2xorm(sh,(15<<4)+1);
}
#define init_12_14_15_2
void FASTCALL call_12_14_15_2(SH2_struct* sh)
{
  SH2xorm(sh,(15<<4)+2);
}
#define init_12_14_15_3
void FASTCALL call_12_14_15_3(SH2_struct* sh)
{
  SH2xorm(sh,(15<<4)+3);
}
#define init_12_14_15_4
void FASTCALL call_12_14_15_4(SH2_struct* sh)
{
  SH2xorm(sh,(15<<4)+4);
}
#define init_12_14_15_5
void FASTCALL call_12_14_15_5(SH2_struct* sh)
{
  SH2xorm(sh,(15<<4)+5);
}
#define init_12_14_15_6
void FASTCALL call_12_14_15_6(SH2_struct* sh)
{
  SH2xorm(sh,(15<<4)+6);
}
#define init_12_14_15_7
void FASTCALL call_12_14_15_7(SH2_struct* sh)
{
  SH2xorm(sh,(15<<4)+7);
}
#define init_12_14_15_8
void FASTCALL call_12_14_15_8(SH2_struct* sh)
{
  SH2xorm(sh,(15<<4)+8);
}
#define init_12_14_15_9
void FASTCALL call_12_14_15_9(SH2_struct* sh)
{
  SH2xorm(sh,(15<<4)+9);
}
#define init_12_14_15_10
void FASTCALL call_12_14_15_10(SH2_struct* sh)
{
  SH2xorm(sh,(15<<4)+10);
}
#define init_12_14_15_11
void FASTCALL call_12_14_15_11(SH2_struct* sh)
{
  SH2xorm(sh,(15<<4)+11);
}
#define init_12_14_15_12
void FASTCALL call_12_14_15_12(SH2_struct* sh)
{
  SH2xorm(sh,(15<<4)+12);
}
#define init_12_14_15_13
void FASTCALL call_12_14_15_13(SH2_struct* sh)
{
  SH2xorm(sh,(15<<4)+13);
}
#define init_12_14_15_14
void FASTCALL call_12_14_15_14(SH2_struct* sh)
{
  SH2xorm(sh,(15<<4)+14);
}
#define init_12_14_15_15
void FASTCALL call_12_14_15_15(SH2_struct* sh)
{
  SH2xorm(sh,(15<<4)+15);
}
#define init_12_10_0_0
void FASTCALL call_12_10_0_0(SH2_struct* sh)
{
  SH2xori(sh,(0<<4)+0);
}
#define init_12_10_0_1
void FASTCALL call_12_10_0_1(SH2_struct* sh)
{
  SH2xori(sh,(0<<4)+1);
}
#define init_12_10_0_2
void FASTCALL call_12_10_0_2(SH2_struct* sh)
{
  SH2xori(sh,(0<<4)+2);
}
#define init_12_10_0_3
void FASTCALL call_12_10_0_3(SH2_struct* sh)
{
  SH2xori(sh,(0<<4)+3);
}
#define init_12_10_0_4
void FASTCALL call_12_10_0_4(SH2_struct* sh)
{
  SH2xori(sh,(0<<4)+4);
}
#define init_12_10_0_5
void FASTCALL call_12_10_0_5(SH2_struct* sh)
{
  SH2xori(sh,(0<<4)+5);
}
#define init_12_10_0_6
void FASTCALL call_12_10_0_6(SH2_struct* sh)
{
  SH2xori(sh,(0<<4)+6);
}
#define init_12_10_0_7
void FASTCALL call_12_10_0_7(SH2_struct* sh)
{
  SH2xori(sh,(0<<4)+7);
}
#define init_12_10_0_8
void FASTCALL call_12_10_0_8(SH2_struct* sh)
{
  SH2xori(sh,(0<<4)+8);
}
#define init_12_10_0_9
void FASTCALL call_12_10_0_9(SH2_struct* sh)
{
  SH2xori(sh,(0<<4)+9);
}
#define init_12_10_0_10
void FASTCALL call_12_10_0_10(SH2_struct* sh)
{
  SH2xori(sh,(0<<4)+10);
}
#define init_12_10_0_11
void FASTCALL call_12_10_0_11(SH2_struct* sh)
{
  SH2xori(sh,(0<<4)+11);
}
#define init_12_10_0_12
void FASTCALL call_12_10_0_12(SH2_struct* sh)
{
  SH2xori(sh,(0<<4)+12);
}
#define init_12_10_0_13
void FASTCALL call_12_10_0_13(SH2_struct* sh)
{
  SH2xori(sh,(0<<4)+13);
}
#define init_12_10_0_14
void FASTCALL call_12_10_0_14(SH2_struct* sh)
{
  SH2xori(sh,(0<<4)+14);
}
#define init_12_10_0_15
void FASTCALL call_12_10_0_15(SH2_struct* sh)
{
  SH2xori(sh,(0<<4)+15);
}
#define init_12_10_1_0
void FASTCALL call_12_10_1_0(SH2_struct* sh)
{
  SH2xori(sh,(1<<4)+0);
}
#define init_12_10_1_1
void FASTCALL call_12_10_1_1(SH2_struct* sh)
{
  SH2xori(sh,(1<<4)+1);
}
#define init_12_10_1_2
void FASTCALL call_12_10_1_2(SH2_struct* sh)
{
  SH2xori(sh,(1<<4)+2);
}
#define init_12_10_1_3
void FASTCALL call_12_10_1_3(SH2_struct* sh)
{
  SH2xori(sh,(1<<4)+3);
}
#define init_12_10_1_4
void FASTCALL call_12_10_1_4(SH2_struct* sh)
{
  SH2xori(sh,(1<<4)+4);
}
#define init_12_10_1_5
void FASTCALL call_12_10_1_5(SH2_struct* sh)
{
  SH2xori(sh,(1<<4)+5);
}
#define init_12_10_1_6
void FASTCALL call_12_10_1_6(SH2_struct* sh)
{
  SH2xori(sh,(1<<4)+6);
}
#define init_12_10_1_7
void FASTCALL call_12_10_1_7(SH2_struct* sh)
{
  SH2xori(sh,(1<<4)+7);
}
#define init_12_10_1_8
void FASTCALL call_12_10_1_8(SH2_struct* sh)
{
  SH2xori(sh,(1<<4)+8);
}
#define init_12_10_1_9
void FASTCALL call_12_10_1_9(SH2_struct* sh)
{
  SH2xori(sh,(1<<4)+9);
}
#define init_12_10_1_10
void FASTCALL call_12_10_1_10(SH2_struct* sh)
{
  SH2xori(sh,(1<<4)+10);
}
#define init_12_10_1_11
void FASTCALL call_12_10_1_11(SH2_struct* sh)
{
  SH2xori(sh,(1<<4)+11);
}
#define init_12_10_1_12
void FASTCALL call_12_10_1_12(SH2_struct* sh)
{
  SH2xori(sh,(1<<4)+12);
}
#define init_12_10_1_13
void FASTCALL call_12_10_1_13(SH2_struct* sh)
{
  SH2xori(sh,(1<<4)+13);
}
#define init_12_10_1_14
void FASTCALL call_12_10_1_14(SH2_struct* sh)
{
  SH2xori(sh,(1<<4)+14);
}
#define init_12_10_1_15
void FASTCALL call_12_10_1_15(SH2_struct* sh)
{
  SH2xori(sh,(1<<4)+15);
}
#define init_12_10_2_0
void FASTCALL call_12_10_2_0(SH2_struct* sh)
{
  SH2xori(sh,(2<<4)+0);
}
#define init_12_10_2_1
void FASTCALL call_12_10_2_1(SH2_struct* sh)
{
  SH2xori(sh,(2<<4)+1);
}
#define init_12_10_2_2
void FASTCALL call_12_10_2_2(SH2_struct* sh)
{
  SH2xori(sh,(2<<4)+2);
}
#define init_12_10_2_3
void FASTCALL call_12_10_2_3(SH2_struct* sh)
{
  SH2xori(sh,(2<<4)+3);
}
#define init_12_10_2_4
void FASTCALL call_12_10_2_4(SH2_struct* sh)
{
  SH2xori(sh,(2<<4)+4);
}
#define init_12_10_2_5
void FASTCALL call_12_10_2_5(SH2_struct* sh)
{
  SH2xori(sh,(2<<4)+5);
}
#define init_12_10_2_6
void FASTCALL call_12_10_2_6(SH2_struct* sh)
{
  SH2xori(sh,(2<<4)+6);
}
#define init_12_10_2_7
void FASTCALL call_12_10_2_7(SH2_struct* sh)
{
  SH2xori(sh,(2<<4)+7);
}
#define init_12_10_2_8
void FASTCALL call_12_10_2_8(SH2_struct* sh)
{
  SH2xori(sh,(2<<4)+8);
}
#define init_12_10_2_9
void FASTCALL call_12_10_2_9(SH2_struct* sh)
{
  SH2xori(sh,(2<<4)+9);
}
#define init_12_10_2_10
void FASTCALL call_12_10_2_10(SH2_struct* sh)
{
  SH2xori(sh,(2<<4)+10);
}
#define init_12_10_2_11
void FASTCALL call_12_10_2_11(SH2_struct* sh)
{
  SH2xori(sh,(2<<4)+11);
}
#define init_12_10_2_12
void FASTCALL call_12_10_2_12(SH2_struct* sh)
{
  SH2xori(sh,(2<<4)+12);
}
#define init_12_10_2_13
void FASTCALL call_12_10_2_13(SH2_struct* sh)
{
  SH2xori(sh,(2<<4)+13);
}
#define init_12_10_2_14
void FASTCALL call_12_10_2_14(SH2_struct* sh)
{
  SH2xori(sh,(2<<4)+14);
}
#define init_12_10_2_15
void FASTCALL call_12_10_2_15(SH2_struct* sh)
{
  SH2xori(sh,(2<<4)+15);
}
#define init_12_10_3_0
void FASTCALL call_12_10_3_0(SH2_struct* sh)
{
  SH2xori(sh,(3<<4)+0);
}
#define init_12_10_3_1
void FASTCALL call_12_10_3_1(SH2_struct* sh)
{
  SH2xori(sh,(3<<4)+1);
}
#define init_12_10_3_2
void FASTCALL call_12_10_3_2(SH2_struct* sh)
{
  SH2xori(sh,(3<<4)+2);
}
#define init_12_10_3_3
void FASTCALL call_12_10_3_3(SH2_struct* sh)
{
  SH2xori(sh,(3<<4)+3);
}
#define init_12_10_3_4
void FASTCALL call_12_10_3_4(SH2_struct* sh)
{
  SH2xori(sh,(3<<4)+4);
}
#define init_12_10_3_5
void FASTCALL call_12_10_3_5(SH2_struct* sh)
{
  SH2xori(sh,(3<<4)+5);
}
#define init_12_10_3_6
void FASTCALL call_12_10_3_6(SH2_struct* sh)
{
  SH2xori(sh,(3<<4)+6);
}
#define init_12_10_3_7
void FASTCALL call_12_10_3_7(SH2_struct* sh)
{
  SH2xori(sh,(3<<4)+7);
}
#define init_12_10_3_8
void FASTCALL call_12_10_3_8(SH2_struct* sh)
{
  SH2xori(sh,(3<<4)+8);
}
#define init_12_10_3_9
void FASTCALL call_12_10_3_9(SH2_struct* sh)
{
  SH2xori(sh,(3<<4)+9);
}
#define init_12_10_3_10
void FASTCALL call_12_10_3_10(SH2_struct* sh)
{
  SH2xori(sh,(3<<4)+10);
}
#define init_12_10_3_11
void FASTCALL call_12_10_3_11(SH2_struct* sh)
{
  SH2xori(sh,(3<<4)+11);
}
#define init_12_10_3_12
void FASTCALL call_12_10_3_12(SH2_struct* sh)
{
  SH2xori(sh,(3<<4)+12);
}
#define init_12_10_3_13
void FASTCALL call_12_10_3_13(SH2_struct* sh)
{
  SH2xori(sh,(3<<4)+13);
}
#define init_12_10_3_14
void FASTCALL call_12_10_3_14(SH2_struct* sh)
{
  SH2xori(sh,(3<<4)+14);
}
#define init_12_10_3_15
void FASTCALL call_12_10_3_15(SH2_struct* sh)
{
  SH2xori(sh,(3<<4)+15);
}
#define init_12_10_4_0
void FASTCALL call_12_10_4_0(SH2_struct* sh)
{
  SH2xori(sh,(4<<4)+0);
}
#define init_12_10_4_1
void FASTCALL call_12_10_4_1(SH2_struct* sh)
{
  SH2xori(sh,(4<<4)+1);
}
#define init_12_10_4_2
void FASTCALL call_12_10_4_2(SH2_struct* sh)
{
  SH2xori(sh,(4<<4)+2);
}
#define init_12_10_4_3
void FASTCALL call_12_10_4_3(SH2_struct* sh)
{
  SH2xori(sh,(4<<4)+3);
}
#define init_12_10_4_4
void FASTCALL call_12_10_4_4(SH2_struct* sh)
{
  SH2xori(sh,(4<<4)+4);
}
#define init_12_10_4_5
void FASTCALL call_12_10_4_5(SH2_struct* sh)
{
  SH2xori(sh,(4<<4)+5);
}
#define init_12_10_4_6
void FASTCALL call_12_10_4_6(SH2_struct* sh)
{
  SH2xori(sh,(4<<4)+6);
}
#define init_12_10_4_7
void FASTCALL call_12_10_4_7(SH2_struct* sh)
{
  SH2xori(sh,(4<<4)+7);
}
#define init_12_10_4_8
void FASTCALL call_12_10_4_8(SH2_struct* sh)
{
  SH2xori(sh,(4<<4)+8);
}
#define init_12_10_4_9
void FASTCALL call_12_10_4_9(SH2_struct* sh)
{
  SH2xori(sh,(4<<4)+9);
}
#define init_12_10_4_10
void FASTCALL call_12_10_4_10(SH2_struct* sh)
{
  SH2xori(sh,(4<<4)+10);
}
#define init_12_10_4_11
void FASTCALL call_12_10_4_11(SH2_struct* sh)
{
  SH2xori(sh,(4<<4)+11);
}
#define init_12_10_4_12
void FASTCALL call_12_10_4_12(SH2_struct* sh)
{
  SH2xori(sh,(4<<4)+12);
}
#define init_12_10_4_13
void FASTCALL call_12_10_4_13(SH2_struct* sh)
{
  SH2xori(sh,(4<<4)+13);
}
#define init_12_10_4_14
void FASTCALL call_12_10_4_14(SH2_struct* sh)
{
  SH2xori(sh,(4<<4)+14);
}
#define init_12_10_4_15
void FASTCALL call_12_10_4_15(SH2_struct* sh)
{
  SH2xori(sh,(4<<4)+15);
}
#define init_12_10_5_0
void FASTCALL call_12_10_5_0(SH2_struct* sh)
{
  SH2xori(sh,(5<<4)+0);
}
#define init_12_10_5_1
void FASTCALL call_12_10_5_1(SH2_struct* sh)
{
  SH2xori(sh,(5<<4)+1);
}
#define init_12_10_5_2
void FASTCALL call_12_10_5_2(SH2_struct* sh)
{
  SH2xori(sh,(5<<4)+2);
}
#define init_12_10_5_3
void FASTCALL call_12_10_5_3(SH2_struct* sh)
{
  SH2xori(sh,(5<<4)+3);
}
#define init_12_10_5_4
void FASTCALL call_12_10_5_4(SH2_struct* sh)
{
  SH2xori(sh,(5<<4)+4);
}
#define init_12_10_5_5
void FASTCALL call_12_10_5_5(SH2_struct* sh)
{
  SH2xori(sh,(5<<4)+5);
}
#define init_12_10_5_6
void FASTCALL call_12_10_5_6(SH2_struct* sh)
{
  SH2xori(sh,(5<<4)+6);
}
#define init_12_10_5_7
void FASTCALL call_12_10_5_7(SH2_struct* sh)
{
  SH2xori(sh,(5<<4)+7);
}
#define init_12_10_5_8
void FASTCALL call_12_10_5_8(SH2_struct* sh)
{
  SH2xori(sh,(5<<4)+8);
}
#define init_12_10_5_9
void FASTCALL call_12_10_5_9(SH2_struct* sh)
{
  SH2xori(sh,(5<<4)+9);
}
#define init_12_10_5_10
void FASTCALL call_12_10_5_10(SH2_struct* sh)
{
  SH2xori(sh,(5<<4)+10);
}
#define init_12_10_5_11
void FASTCALL call_12_10_5_11(SH2_struct* sh)
{
  SH2xori(sh,(5<<4)+11);
}
#define init_12_10_5_12
void FASTCALL call_12_10_5_12(SH2_struct* sh)
{
  SH2xori(sh,(5<<4)+12);
}
#define init_12_10_5_13
void FASTCALL call_12_10_5_13(SH2_struct* sh)
{
  SH2xori(sh,(5<<4)+13);
}
#define init_12_10_5_14
void FASTCALL call_12_10_5_14(SH2_struct* sh)
{
  SH2xori(sh,(5<<4)+14);
}
#define init_12_10_5_15
void FASTCALL call_12_10_5_15(SH2_struct* sh)
{
  SH2xori(sh,(5<<4)+15);
}
#define init_12_10_6_0
void FASTCALL call_12_10_6_0(SH2_struct* sh)
{
  SH2xori(sh,(6<<4)+0);
}
#define init_12_10_6_1
void FASTCALL call_12_10_6_1(SH2_struct* sh)
{
  SH2xori(sh,(6<<4)+1);
}
#define init_12_10_6_2
void FASTCALL call_12_10_6_2(SH2_struct* sh)
{
  SH2xori(sh,(6<<4)+2);
}
#define init_12_10_6_3
void FASTCALL call_12_10_6_3(SH2_struct* sh)
{
  SH2xori(sh,(6<<4)+3);
}
#define init_12_10_6_4
void FASTCALL call_12_10_6_4(SH2_struct* sh)
{
  SH2xori(sh,(6<<4)+4);
}
#define init_12_10_6_5
void FASTCALL call_12_10_6_5(SH2_struct* sh)
{
  SH2xori(sh,(6<<4)+5);
}
#define init_12_10_6_6
void FASTCALL call_12_10_6_6(SH2_struct* sh)
{
  SH2xori(sh,(6<<4)+6);
}
#define init_12_10_6_7
void FASTCALL call_12_10_6_7(SH2_struct* sh)
{
  SH2xori(sh,(6<<4)+7);
}
#define init_12_10_6_8
void FASTCALL call_12_10_6_8(SH2_struct* sh)
{
  SH2xori(sh,(6<<4)+8);
}
#define init_12_10_6_9
void FASTCALL call_12_10_6_9(SH2_struct* sh)
{
  SH2xori(sh,(6<<4)+9);
}
#define init_12_10_6_10
void FASTCALL call_12_10_6_10(SH2_struct* sh)
{
  SH2xori(sh,(6<<4)+10);
}
#define init_12_10_6_11
void FASTCALL call_12_10_6_11(SH2_struct* sh)
{
  SH2xori(sh,(6<<4)+11);
}
#define init_12_10_6_12
void FASTCALL call_12_10_6_12(SH2_struct* sh)
{
  SH2xori(sh,(6<<4)+12);
}
#define init_12_10_6_13
void FASTCALL call_12_10_6_13(SH2_struct* sh)
{
  SH2xori(sh,(6<<4)+13);
}
#define init_12_10_6_14
void FASTCALL call_12_10_6_14(SH2_struct* sh)
{
  SH2xori(sh,(6<<4)+14);
}
#define init_12_10_6_15
void FASTCALL call_12_10_6_15(SH2_struct* sh)
{
  SH2xori(sh,(6<<4)+15);
}
#define init_12_10_7_0
void FASTCALL call_12_10_7_0(SH2_struct* sh)
{
  SH2xori(sh,(7<<4)+0);
}
#define init_12_10_7_1
void FASTCALL call_12_10_7_1(SH2_struct* sh)
{
  SH2xori(sh,(7<<4)+1);
}
#define init_12_10_7_2
void FASTCALL call_12_10_7_2(SH2_struct* sh)
{
  SH2xori(sh,(7<<4)+2);
}
#define init_12_10_7_3
void FASTCALL call_12_10_7_3(SH2_struct* sh)
{
  SH2xori(sh,(7<<4)+3);
}
#define init_12_10_7_4
void FASTCALL call_12_10_7_4(SH2_struct* sh)
{
  SH2xori(sh,(7<<4)+4);
}
#define init_12_10_7_5
void FASTCALL call_12_10_7_5(SH2_struct* sh)
{
  SH2xori(sh,(7<<4)+5);
}
#define init_12_10_7_6
void FASTCALL call_12_10_7_6(SH2_struct* sh)
{
  SH2xori(sh,(7<<4)+6);
}
#define init_12_10_7_7
void FASTCALL call_12_10_7_7(SH2_struct* sh)
{
  SH2xori(sh,(7<<4)+7);
}
#define init_12_10_7_8
void FASTCALL call_12_10_7_8(SH2_struct* sh)
{
  SH2xori(sh,(7<<4)+8);
}
#define init_12_10_7_9
void FASTCALL call_12_10_7_9(SH2_struct* sh)
{
  SH2xori(sh,(7<<4)+9);
}
#define init_12_10_7_10
void FASTCALL call_12_10_7_10(SH2_struct* sh)
{
  SH2xori(sh,(7<<4)+10);
}
#define init_12_10_7_11
void FASTCALL call_12_10_7_11(SH2_struct* sh)
{
  SH2xori(sh,(7<<4)+11);
}
#define init_12_10_7_12
void FASTCALL call_12_10_7_12(SH2_struct* sh)
{
  SH2xori(sh,(7<<4)+12);
}
#define init_12_10_7_13
void FASTCALL call_12_10_7_13(SH2_struct* sh)
{
  SH2xori(sh,(7<<4)+13);
}
#define init_12_10_7_14
void FASTCALL call_12_10_7_14(SH2_struct* sh)
{
  SH2xori(sh,(7<<4)+14);
}
#define init_12_10_7_15
void FASTCALL call_12_10_7_15(SH2_struct* sh)
{
  SH2xori(sh,(7<<4)+15);
}
#define init_12_10_8_0
void FASTCALL call_12_10_8_0(SH2_struct* sh)
{
  SH2xori(sh,(8<<4)+0);
}
#define init_12_10_8_1
void FASTCALL call_12_10_8_1(SH2_struct* sh)
{
  SH2xori(sh,(8<<4)+1);
}
#define init_12_10_8_2
void FASTCALL call_12_10_8_2(SH2_struct* sh)
{
  SH2xori(sh,(8<<4)+2);
}
#define init_12_10_8_3
void FASTCALL call_12_10_8_3(SH2_struct* sh)
{
  SH2xori(sh,(8<<4)+3);
}
#define init_12_10_8_4
void FASTCALL call_12_10_8_4(SH2_struct* sh)
{
  SH2xori(sh,(8<<4)+4);
}
#define init_12_10_8_5
void FASTCALL call_12_10_8_5(SH2_struct* sh)
{
  SH2xori(sh,(8<<4)+5);
}
#define init_12_10_8_6
void FASTCALL call_12_10_8_6(SH2_struct* sh)
{
  SH2xori(sh,(8<<4)+6);
}
#define init_12_10_8_7
void FASTCALL call_12_10_8_7(SH2_struct* sh)
{
  SH2xori(sh,(8<<4)+7);
}
#define init_12_10_8_8
void FASTCALL call_12_10_8_8(SH2_struct* sh)
{
  SH2xori(sh,(8<<4)+8);
}
#define init_12_10_8_9
void FASTCALL call_12_10_8_9(SH2_struct* sh)
{
  SH2xori(sh,(8<<4)+9);
}
#define init_12_10_8_10
void FASTCALL call_12_10_8_10(SH2_struct* sh)
{
  SH2xori(sh,(8<<4)+10);
}
#define init_12_10_8_11
void FASTCALL call_12_10_8_11(SH2_struct* sh)
{
  SH2xori(sh,(8<<4)+11);
}
#define init_12_10_8_12
void FASTCALL call_12_10_8_12(SH2_struct* sh)
{
  SH2xori(sh,(8<<4)+12);
}
#define init_12_10_8_13
void FASTCALL call_12_10_8_13(SH2_struct* sh)
{
  SH2xori(sh,(8<<4)+13);
}
#define init_12_10_8_14
void FASTCALL call_12_10_8_14(SH2_struct* sh)
{
  SH2xori(sh,(8<<4)+14);
}
#define init_12_10_8_15
void FASTCALL call_12_10_8_15(SH2_struct* sh)
{
  SH2xori(sh,(8<<4)+15);
}
#define init_12_10_9_0
void FASTCALL call_12_10_9_0(SH2_struct* sh)
{
  SH2xori(sh,(9<<4)+0);
}
#define init_12_10_9_1
void FASTCALL call_12_10_9_1(SH2_struct* sh)
{
  SH2xori(sh,(9<<4)+1);
}
#define init_12_10_9_2
void FASTCALL call_12_10_9_2(SH2_struct* sh)
{
  SH2xori(sh,(9<<4)+2);
}
#define init_12_10_9_3
void FASTCALL call_12_10_9_3(SH2_struct* sh)
{
  SH2xori(sh,(9<<4)+3);
}
#define init_12_10_9_4
void FASTCALL call_12_10_9_4(SH2_struct* sh)
{
  SH2xori(sh,(9<<4)+4);
}
#define init_12_10_9_5
void FASTCALL call_12_10_9_5(SH2_struct* sh)
{
  SH2xori(sh,(9<<4)+5);
}
#define init_12_10_9_6
void FASTCALL call_12_10_9_6(SH2_struct* sh)
{
  SH2xori(sh,(9<<4)+6);
}
#define init_12_10_9_7
void FASTCALL call_12_10_9_7(SH2_struct* sh)
{
  SH2xori(sh,(9<<4)+7);
}
#define init_12_10_9_8
void FASTCALL call_12_10_9_8(SH2_struct* sh)
{
  SH2xori(sh,(9<<4)+8);
}
#define init_12_10_9_9
void FASTCALL call_12_10_9_9(SH2_struct* sh)
{
  SH2xori(sh,(9<<4)+9);
}
#define init_12_10_9_10
void FASTCALL call_12_10_9_10(SH2_struct* sh)
{
  SH2xori(sh,(9<<4)+10);
}
#define init_12_10_9_11
void FASTCALL call_12_10_9_11(SH2_struct* sh)
{
  SH2xori(sh,(9<<4)+11);
}
#define init_12_10_9_12
void FASTCALL call_12_10_9_12(SH2_struct* sh)
{
  SH2xori(sh,(9<<4)+12);
}
#define init_12_10_9_13
void FASTCALL call_12_10_9_13(SH2_struct* sh)
{
  SH2xori(sh,(9<<4)+13);
}
#define init_12_10_9_14
void FASTCALL call_12_10_9_14(SH2_struct* sh)
{
  SH2xori(sh,(9<<4)+14);
}
#define init_12_10_9_15
void FASTCALL call_12_10_9_15(SH2_struct* sh)
{
  SH2xori(sh,(9<<4)+15);
}
#define init_12_10_10_0
void FASTCALL call_12_10_10_0(SH2_struct* sh)
{
  SH2xori(sh,(10<<4)+0);
}
#define init_12_10_10_1
void FASTCALL call_12_10_10_1(SH2_struct* sh)
{
  SH2xori(sh,(10<<4)+1);
}
#define init_12_10_10_2
void FASTCALL call_12_10_10_2(SH2_struct* sh)
{
  SH2xori(sh,(10<<4)+2);
}
#define init_12_10_10_3
void FASTCALL call_12_10_10_3(SH2_struct* sh)
{
  SH2xori(sh,(10<<4)+3);
}
#define init_12_10_10_4
void FASTCALL call_12_10_10_4(SH2_struct* sh)
{
  SH2xori(sh,(10<<4)+4);
}
#define init_12_10_10_5
void FASTCALL call_12_10_10_5(SH2_struct* sh)
{
  SH2xori(sh,(10<<4)+5);
}
#define init_12_10_10_6
void FASTCALL call_12_10_10_6(SH2_struct* sh)
{
  SH2xori(sh,(10<<4)+6);
}
#define init_12_10_10_7
void FASTCALL call_12_10_10_7(SH2_struct* sh)
{
  SH2xori(sh,(10<<4)+7);
}
#define init_12_10_10_8
void FASTCALL call_12_10_10_8(SH2_struct* sh)
{
  SH2xori(sh,(10<<4)+8);
}
#define init_12_10_10_9
void FASTCALL call_12_10_10_9(SH2_struct* sh)
{
  SH2xori(sh,(10<<4)+9);
}
#define init_12_10_10_10
void FASTCALL call_12_10_10_10(SH2_struct* sh)
{
  SH2xori(sh,(10<<4)+10);
}
#define init_12_10_10_11
void FASTCALL call_12_10_10_11(SH2_struct* sh)
{
  SH2xori(sh,(10<<4)+11);
}
#define init_12_10_10_12
void FASTCALL call_12_10_10_12(SH2_struct* sh)
{
  SH2xori(sh,(10<<4)+12);
}
#define init_12_10_10_13
void FASTCALL call_12_10_10_13(SH2_struct* sh)
{
  SH2xori(sh,(10<<4)+13);
}
#define init_12_10_10_14
void FASTCALL call_12_10_10_14(SH2_struct* sh)
{
  SH2xori(sh,(10<<4)+14);
}
#define init_12_10_10_15
void FASTCALL call_12_10_10_15(SH2_struct* sh)
{
  SH2xori(sh,(10<<4)+15);
}
#define init_12_10_11_0
void FASTCALL call_12_10_11_0(SH2_struct* sh)
{
  SH2xori(sh,(11<<4)+0);
}
#define init_12_10_11_1
void FASTCALL call_12_10_11_1(SH2_struct* sh)
{
  SH2xori(sh,(11<<4)+1);
}
#define init_12_10_11_2
void FASTCALL call_12_10_11_2(SH2_struct* sh)
{
  SH2xori(sh,(11<<4)+2);
}
#define init_12_10_11_3
void FASTCALL call_12_10_11_3(SH2_struct* sh)
{
  SH2xori(sh,(11<<4)+3);
}
#define init_12_10_11_4
void FASTCALL call_12_10_11_4(SH2_struct* sh)
{
  SH2xori(sh,(11<<4)+4);
}
#define init_12_10_11_5
void FASTCALL call_12_10_11_5(SH2_struct* sh)
{
  SH2xori(sh,(11<<4)+5);
}
#define init_12_10_11_6
void FASTCALL call_12_10_11_6(SH2_struct* sh)
{
  SH2xori(sh,(11<<4)+6);
}
#define init_12_10_11_7
void FASTCALL call_12_10_11_7(SH2_struct* sh)
{
  SH2xori(sh,(11<<4)+7);
}
#define init_12_10_11_8
void FASTCALL call_12_10_11_8(SH2_struct* sh)
{
  SH2xori(sh,(11<<4)+8);
}
#define init_12_10_11_9
void FASTCALL call_12_10_11_9(SH2_struct* sh)
{
  SH2xori(sh,(11<<4)+9);
}
#define init_12_10_11_10
void FASTCALL call_12_10_11_10(SH2_struct* sh)
{
  SH2xori(sh,(11<<4)+10);
}
#define init_12_10_11_11
void FASTCALL call_12_10_11_11(SH2_struct* sh)
{
  SH2xori(sh,(11<<4)+11);
}
#define init_12_10_11_12
void FASTCALL call_12_10_11_12(SH2_struct* sh)
{
  SH2xori(sh,(11<<4)+12);
}
#define init_12_10_11_13
void FASTCALL call_12_10_11_13(SH2_struct* sh)
{
  SH2xori(sh,(11<<4)+13);
}
#define init_12_10_11_14
void FASTCALL call_12_10_11_14(SH2_struct* sh)
{
  SH2xori(sh,(11<<4)+14);
}
#define init_12_10_11_15
void FASTCALL call_12_10_11_15(SH2_struct* sh)
{
  SH2xori(sh,(11<<4)+15);
}
#define init_12_10_12_0
void FASTCALL call_12_10_12_0(SH2_struct* sh)
{
  SH2xori(sh,(12<<4)+0);
}
#define init_12_10_12_1
void FASTCALL call_12_10_12_1(SH2_struct* sh)
{
  SH2xori(sh,(12<<4)+1);
}
#define init_12_10_12_2
void FASTCALL call_12_10_12_2(SH2_struct* sh)
{
  SH2xori(sh,(12<<4)+2);
}
#define init_12_10_12_3
void FASTCALL call_12_10_12_3(SH2_struct* sh)
{
  SH2xori(sh,(12<<4)+3);
}
#define init_12_10_12_4
void FASTCALL call_12_10_12_4(SH2_struct* sh)
{
  SH2xori(sh,(12<<4)+4);
}
#define init_12_10_12_5
void FASTCALL call_12_10_12_5(SH2_struct* sh)
{
  SH2xori(sh,(12<<4)+5);
}
#define init_12_10_12_6
void FASTCALL call_12_10_12_6(SH2_struct* sh)
{
  SH2xori(sh,(12<<4)+6);
}
#define init_12_10_12_7
void FASTCALL call_12_10_12_7(SH2_struct* sh)
{
  SH2xori(sh,(12<<4)+7);
}
#define init_12_10_12_8
void FASTCALL call_12_10_12_8(SH2_struct* sh)
{
  SH2xori(sh,(12<<4)+8);
}
#define init_12_10_12_9
void FASTCALL call_12_10_12_9(SH2_struct* sh)
{
  SH2xori(sh,(12<<4)+9);
}
#define init_12_10_12_10
void FASTCALL call_12_10_12_10(SH2_struct* sh)
{
  SH2xori(sh,(12<<4)+10);
}
#define init_12_10_12_11
void FASTCALL call_12_10_12_11(SH2_struct* sh)
{
  SH2xori(sh,(12<<4)+11);
}
#define init_12_10_12_12
void FASTCALL call_12_10_12_12(SH2_struct* sh)
{
  SH2xori(sh,(12<<4)+12);
}
#define init_12_10_12_13
void FASTCALL call_12_10_12_13(SH2_struct* sh)
{
  SH2xori(sh,(12<<4)+13);
}
#define init_12_10_12_14
void FASTCALL call_12_10_12_14(SH2_struct* sh)
{
  SH2xori(sh,(12<<4)+14);
}
#define init_12_10_12_15
void FASTCALL call_12_10_12_15(SH2_struct* sh)
{
  SH2xori(sh,(12<<4)+15);
}
#define init_12_10_13_0
void FASTCALL call_12_10_13_0(SH2_struct* sh)
{
  SH2xori(sh,(13<<4)+0);
}
#define init_12_10_13_1
void FASTCALL call_12_10_13_1(SH2_struct* sh)
{
  SH2xori(sh,(13<<4)+1);
}
#define init_12_10_13_2
void FASTCALL call_12_10_13_2(SH2_struct* sh)
{
  SH2xori(sh,(13<<4)+2);
}
#define init_12_10_13_3
void FASTCALL call_12_10_13_3(SH2_struct* sh)
{
  SH2xori(sh,(13<<4)+3);
}
#define init_12_10_13_4
void FASTCALL call_12_10_13_4(SH2_struct* sh)
{
  SH2xori(sh,(13<<4)+4);
}
#define init_12_10_13_5
void FASTCALL call_12_10_13_5(SH2_struct* sh)
{
  SH2xori(sh,(13<<4)+5);
}
#define init_12_10_13_6
void FASTCALL call_12_10_13_6(SH2_struct* sh)
{
  SH2xori(sh,(13<<4)+6);
}
#define init_12_10_13_7
void FASTCALL call_12_10_13_7(SH2_struct* sh)
{
  SH2xori(sh,(13<<4)+7);
}
#define init_12_10_13_8
void FASTCALL call_12_10_13_8(SH2_struct* sh)
{
  SH2xori(sh,(13<<4)+8);
}
#define init_12_10_13_9
void FASTCALL call_12_10_13_9(SH2_struct* sh)
{
  SH2xori(sh,(13<<4)+9);
}
#define init_12_10_13_10
void FASTCALL call_12_10_13_10(SH2_struct* sh)
{
  SH2xori(sh,(13<<4)+10);
}
#define init_12_10_13_11
void FASTCALL call_12_10_13_11(SH2_struct* sh)
{
  SH2xori(sh,(13<<4)+11);
}
#define init_12_10_13_12
void FASTCALL call_12_10_13_12(SH2_struct* sh)
{
  SH2xori(sh,(13<<4)+12);
}
#define init_12_10_13_13
void FASTCALL call_12_10_13_13(SH2_struct* sh)
{
  SH2xori(sh,(13<<4)+13);
}
#define init_12_10_13_14
void FASTCALL call_12_10_13_14(SH2_struct* sh)
{
  SH2xori(sh,(13<<4)+14);
}
#define init_12_10_13_15
void FASTCALL call_12_10_13_15(SH2_struct* sh)
{
  SH2xori(sh,(13<<4)+15);
}
#define init_12_10_14_0
void FASTCALL call_12_10_14_0(SH2_struct* sh)
{
  SH2xori(sh,(14<<4)+0);
}
#define init_12_10_14_1
void FASTCALL call_12_10_14_1(SH2_struct* sh)
{
  SH2xori(sh,(14<<4)+1);
}
#define init_12_10_14_2
void FASTCALL call_12_10_14_2(SH2_struct* sh)
{
  SH2xori(sh,(14<<4)+2);
}
#define init_12_10_14_3
void FASTCALL call_12_10_14_3(SH2_struct* sh)
{
  SH2xori(sh,(14<<4)+3);
}
#define init_12_10_14_4
void FASTCALL call_12_10_14_4(SH2_struct* sh)
{
  SH2xori(sh,(14<<4)+4);
}
#define init_12_10_14_5
void FASTCALL call_12_10_14_5(SH2_struct* sh)
{
  SH2xori(sh,(14<<4)+5);
}
#define init_12_10_14_6
void FASTCALL call_12_10_14_6(SH2_struct* sh)
{
  SH2xori(sh,(14<<4)+6);
}
#define init_12_10_14_7
void FASTCALL call_12_10_14_7(SH2_struct* sh)
{
  SH2xori(sh,(14<<4)+7);
}
#define init_12_10_14_8
void FASTCALL call_12_10_14_8(SH2_struct* sh)
{
  SH2xori(sh,(14<<4)+8);
}
#define init_12_10_14_9
void FASTCALL call_12_10_14_9(SH2_struct* sh)
{
  SH2xori(sh,(14<<4)+9);
}
#define init_12_10_14_10
void FASTCALL call_12_10_14_10(SH2_struct* sh)
{
  SH2xori(sh,(14<<4)+10);
}
#define init_12_10_14_11
void FASTCALL call_12_10_14_11(SH2_struct* sh)
{
  SH2xori(sh,(14<<4)+11);
}
#define init_12_10_14_12
void FASTCALL call_12_10_14_12(SH2_struct* sh)
{
  SH2xori(sh,(14<<4)+12);
}
#define init_12_10_14_13
void FASTCALL call_12_10_14_13(SH2_struct* sh)
{
  SH2xori(sh,(14<<4)+13);
}
#define init_12_10_14_14
void FASTCALL call_12_10_14_14(SH2_struct* sh)
{
  SH2xori(sh,(14<<4)+14);
}
#define init_12_10_14_15
void FASTCALL call_12_10_14_15(SH2_struct* sh)
{
  SH2xori(sh,(14<<4)+15);
}
#define init_12_10_15_0
void FASTCALL call_12_10_15_0(SH2_struct* sh)
{
  SH2xori(sh,(15<<4)+0);
}
#define init_12_10_15_1
void FASTCALL call_12_10_15_1(SH2_struct* sh)
{
  SH2xori(sh,(15<<4)+1);
}
#define init_12_10_15_2
void FASTCALL call_12_10_15_2(SH2_struct* sh)
{
  SH2xori(sh,(15<<4)+2);
}
#define init_12_10_15_3
void FASTCALL call_12_10_15_3(SH2_struct* sh)
{
  SH2xori(sh,(15<<4)+3);
}
#define init_12_10_15_4
void FASTCALL call_12_10_15_4(SH2_struct* sh)
{
  SH2xori(sh,(15<<4)+4);
}
#define init_12_10_15_5
void FASTCALL call_12_10_15_5(SH2_struct* sh)
{
  SH2xori(sh,(15<<4)+5);
}
#define init_12_10_15_6
void FASTCALL call_12_10_15_6(SH2_struct* sh)
{
  SH2xori(sh,(15<<4)+6);
}
#define init_12_10_15_7
void FASTCALL call_12_10_15_7(SH2_struct* sh)
{
  SH2xori(sh,(15<<4)+7);
}
#define init_12_10_15_8
void FASTCALL call_12_10_15_8(SH2_struct* sh)
{
  SH2xori(sh,(15<<4)+8);
}
#define init_12_10_15_9
void FASTCALL call_12_10_15_9(SH2_struct* sh)
{
  SH2xori(sh,(15<<4)+9);
}
#define init_12_10_15_10
void FASTCALL call_12_10_15_10(SH2_struct* sh)
{
  SH2xori(sh,(15<<4)+10);
}
#define init_12_10_15_11
void FASTCALL call_12_10_15_11(SH2_struct* sh)
{
  SH2xori(sh,(15<<4)+11);
}
#define init_12_10_15_12
void FASTCALL call_12_10_15_12(SH2_struct* sh)
{
  SH2xori(sh,(15<<4)+12);
}
#define init_12_10_15_13
void FASTCALL call_12_10_15_13(SH2_struct* sh)
{
  SH2xori(sh,(15<<4)+13);
}
#define init_12_10_15_14
void FASTCALL call_12_10_15_14(SH2_struct* sh)
{
  SH2xori(sh,(15<<4)+14);
}
#define init_12_10_15_15
void FASTCALL call_12_10_15_15(SH2_struct* sh)
{
  SH2xori(sh,(15<<4)+15);
}
#define init_12_12_0_0
void FASTCALL call_12_12_0_0(SH2_struct* sh)
{
  SH2tstm(sh,(0<<4)+0);
}
#define init_12_12_0_1
void FASTCALL call_12_12_0_1(SH2_struct* sh)
{
  SH2tstm(sh,(0<<4)+1);
}
#define init_12_12_0_2
void FASTCALL call_12_12_0_2(SH2_struct* sh)
{
  SH2tstm(sh,(0<<4)+2);
}
#define init_12_12_0_3
void FASTCALL call_12_12_0_3(SH2_struct* sh)
{
  SH2tstm(sh,(0<<4)+3);
}
#define init_12_12_0_4
void FASTCALL call_12_12_0_4(SH2_struct* sh)
{
  SH2tstm(sh,(0<<4)+4);
}
#define init_12_12_0_5
void FASTCALL call_12_12_0_5(SH2_struct* sh)
{
  SH2tstm(sh,(0<<4)+5);
}
#define init_12_12_0_6
void FASTCALL call_12_12_0_6(SH2_struct* sh)
{
  SH2tstm(sh,(0<<4)+6);
}
#define init_12_12_0_7
void FASTCALL call_12_12_0_7(SH2_struct* sh)
{
  SH2tstm(sh,(0<<4)+7);
}
#define init_12_12_0_8
void FASTCALL call_12_12_0_8(SH2_struct* sh)
{
  SH2tstm(sh,(0<<4)+8);
}
#define init_12_12_0_9
void FASTCALL call_12_12_0_9(SH2_struct* sh)
{
  SH2tstm(sh,(0<<4)+9);
}
#define init_12_12_0_10
void FASTCALL call_12_12_0_10(SH2_struct* sh)
{
  SH2tstm(sh,(0<<4)+10);
}
#define init_12_12_0_11
void FASTCALL call_12_12_0_11(SH2_struct* sh)
{
  SH2tstm(sh,(0<<4)+11);
}
#define init_12_12_0_12
void FASTCALL call_12_12_0_12(SH2_struct* sh)
{
  SH2tstm(sh,(0<<4)+12);
}
#define init_12_12_0_13
void FASTCALL call_12_12_0_13(SH2_struct* sh)
{
  SH2tstm(sh,(0<<4)+13);
}
#define init_12_12_0_14
void FASTCALL call_12_12_0_14(SH2_struct* sh)
{
  SH2tstm(sh,(0<<4)+14);
}
#define init_12_12_0_15
void FASTCALL call_12_12_0_15(SH2_struct* sh)
{
  SH2tstm(sh,(0<<4)+15);
}
#define init_12_12_1_0
void FASTCALL call_12_12_1_0(SH2_struct* sh)
{
  SH2tstm(sh,(1<<4)+0);
}
#define init_12_12_1_1
void FASTCALL call_12_12_1_1(SH2_struct* sh)
{
  SH2tstm(sh,(1<<4)+1);
}
#define init_12_12_1_2
void FASTCALL call_12_12_1_2(SH2_struct* sh)
{
  SH2tstm(sh,(1<<4)+2);
}
#define init_12_12_1_3
void FASTCALL call_12_12_1_3(SH2_struct* sh)
{
  SH2tstm(sh,(1<<4)+3);
}
#define init_12_12_1_4
void FASTCALL call_12_12_1_4(SH2_struct* sh)
{
  SH2tstm(sh,(1<<4)+4);
}
#define init_12_12_1_5
void FASTCALL call_12_12_1_5(SH2_struct* sh)
{
  SH2tstm(sh,(1<<4)+5);
}
#define init_12_12_1_6
void FASTCALL call_12_12_1_6(SH2_struct* sh)
{
  SH2tstm(sh,(1<<4)+6);
}
#define init_12_12_1_7
void FASTCALL call_12_12_1_7(SH2_struct* sh)
{
  SH2tstm(sh,(1<<4)+7);
}
#define init_12_12_1_8
void FASTCALL call_12_12_1_8(SH2_struct* sh)
{
  SH2tstm(sh,(1<<4)+8);
}
#define init_12_12_1_9
void FASTCALL call_12_12_1_9(SH2_struct* sh)
{
  SH2tstm(sh,(1<<4)+9);
}
#define init_12_12_1_10
void FASTCALL call_12_12_1_10(SH2_struct* sh)
{
  SH2tstm(sh,(1<<4)+10);
}
#define init_12_12_1_11
void FASTCALL call_12_12_1_11(SH2_struct* sh)
{
  SH2tstm(sh,(1<<4)+11);
}
#define init_12_12_1_12
void FASTCALL call_12_12_1_12(SH2_struct* sh)
{
  SH2tstm(sh,(1<<4)+12);
}
#define init_12_12_1_13
void FASTCALL call_12_12_1_13(SH2_struct* sh)
{
  SH2tstm(sh,(1<<4)+13);
}
#define init_12_12_1_14
void FASTCALL call_12_12_1_14(SH2_struct* sh)
{
  SH2tstm(sh,(1<<4)+14);
}
#define init_12_12_1_15
void FASTCALL call_12_12_1_15(SH2_struct* sh)
{
  SH2tstm(sh,(1<<4)+15);
}
#define init_12_12_2_0
void FASTCALL call_12_12_2_0(SH2_struct* sh)
{
  SH2tstm(sh,(2<<4)+0);
}
#define init_12_12_2_1
void FASTCALL call_12_12_2_1(SH2_struct* sh)
{
  SH2tstm(sh,(2<<4)+1);
}
#define init_12_12_2_2
void FASTCALL call_12_12_2_2(SH2_struct* sh)
{
  SH2tstm(sh,(2<<4)+2);
}
#define init_12_12_2_3
void FASTCALL call_12_12_2_3(SH2_struct* sh)
{
  SH2tstm(sh,(2<<4)+3);
}
#define init_12_12_2_4
void FASTCALL call_12_12_2_4(SH2_struct* sh)
{
  SH2tstm(sh,(2<<4)+4);
}
#define init_12_12_2_5
void FASTCALL call_12_12_2_5(SH2_struct* sh)
{
  SH2tstm(sh,(2<<4)+5);
}
#define init_12_12_2_6
void FASTCALL call_12_12_2_6(SH2_struct* sh)
{
  SH2tstm(sh,(2<<4)+6);
}
#define init_12_12_2_7
void FASTCALL call_12_12_2_7(SH2_struct* sh)
{
  SH2tstm(sh,(2<<4)+7);
}
#define init_12_12_2_8
void FASTCALL call_12_12_2_8(SH2_struct* sh)
{
  SH2tstm(sh,(2<<4)+8);
}
#define init_12_12_2_9
void FASTCALL call_12_12_2_9(SH2_struct* sh)
{
  SH2tstm(sh,(2<<4)+9);
}
#define init_12_12_2_10
void FASTCALL call_12_12_2_10(SH2_struct* sh)
{
  SH2tstm(sh,(2<<4)+10);
}
#define init_12_12_2_11
void FASTCALL call_12_12_2_11(SH2_struct* sh)
{
  SH2tstm(sh,(2<<4)+11);
}
#define init_12_12_2_12
void FASTCALL call_12_12_2_12(SH2_struct* sh)
{
  SH2tstm(sh,(2<<4)+12);
}
#define init_12_12_2_13
void FASTCALL call_12_12_2_13(SH2_struct* sh)
{
  SH2tstm(sh,(2<<4)+13);
}
#define init_12_12_2_14
void FASTCALL call_12_12_2_14(SH2_struct* sh)
{
  SH2tstm(sh,(2<<4)+14);
}
#define init_12_12_2_15
void FASTCALL call_12_12_2_15(SH2_struct* sh)
{
  SH2tstm(sh,(2<<4)+15);
}
#define init_12_12_3_0
void FASTCALL call_12_12_3_0(SH2_struct* sh)
{
  SH2tstm(sh,(3<<4)+0);
}
#define init_12_12_3_1
void FASTCALL call_12_12_3_1(SH2_struct* sh)
{
  SH2tstm(sh,(3<<4)+1);
}
#define init_12_12_3_2
void FASTCALL call_12_12_3_2(SH2_struct* sh)
{
  SH2tstm(sh,(3<<4)+2);
}
#define init_12_12_3_3
void FASTCALL call_12_12_3_3(SH2_struct* sh)
{
  SH2tstm(sh,(3<<4)+3);
}
#define init_12_12_3_4
void FASTCALL call_12_12_3_4(SH2_struct* sh)
{
  SH2tstm(sh,(3<<4)+4);
}
#define init_12_12_3_5
void FASTCALL call_12_12_3_5(SH2_struct* sh)
{
  SH2tstm(sh,(3<<4)+5);
}
#define init_12_12_3_6
void FASTCALL call_12_12_3_6(SH2_struct* sh)
{
  SH2tstm(sh,(3<<4)+6);
}
#define init_12_12_3_7
void FASTCALL call_12_12_3_7(SH2_struct* sh)
{
  SH2tstm(sh,(3<<4)+7);
}
#define init_12_12_3_8
void FASTCALL call_12_12_3_8(SH2_struct* sh)
{
  SH2tstm(sh,(3<<4)+8);
}
#define init_12_12_3_9
void FASTCALL call_12_12_3_9(SH2_struct* sh)
{
  SH2tstm(sh,(3<<4)+9);
}
#define init_12_12_3_10
void FASTCALL call_12_12_3_10(SH2_struct* sh)
{
  SH2tstm(sh,(3<<4)+10);
}
#define init_12_12_3_11
void FASTCALL call_12_12_3_11(SH2_struct* sh)
{
  SH2tstm(sh,(3<<4)+11);
}
#define init_12_12_3_12
void FASTCALL call_12_12_3_12(SH2_struct* sh)
{
  SH2tstm(sh,(3<<4)+12);
}
#define init_12_12_3_13
void FASTCALL call_12_12_3_13(SH2_struct* sh)
{
  SH2tstm(sh,(3<<4)+13);
}
#define init_12_12_3_14
void FASTCALL call_12_12_3_14(SH2_struct* sh)
{
  SH2tstm(sh,(3<<4)+14);
}
#define init_12_12_3_15
void FASTCALL call_12_12_3_15(SH2_struct* sh)
{
  SH2tstm(sh,(3<<4)+15);
}
#define init_12_12_4_0
void FASTCALL call_12_12_4_0(SH2_struct* sh)
{
  SH2tstm(sh,(4<<4)+0);
}
#define init_12_12_4_1
void FASTCALL call_12_12_4_1(SH2_struct* sh)
{
  SH2tstm(sh,(4<<4)+1);
}
#define init_12_12_4_2
void FASTCALL call_12_12_4_2(SH2_struct* sh)
{
  SH2tstm(sh,(4<<4)+2);
}
#define init_12_12_4_3
void FASTCALL call_12_12_4_3(SH2_struct* sh)
{
  SH2tstm(sh,(4<<4)+3);
}
#define init_12_12_4_4
void FASTCALL call_12_12_4_4(SH2_struct* sh)
{
  SH2tstm(sh,(4<<4)+4);
}
#define init_12_12_4_5
void FASTCALL call_12_12_4_5(SH2_struct* sh)
{
  SH2tstm(sh,(4<<4)+5);
}
#define init_12_12_4_6
void FASTCALL call_12_12_4_6(SH2_struct* sh)
{
  SH2tstm(sh,(4<<4)+6);
}
#define init_12_12_4_7
void FASTCALL call_12_12_4_7(SH2_struct* sh)
{
  SH2tstm(sh,(4<<4)+7);
}
#define init_12_12_4_8
void FASTCALL call_12_12_4_8(SH2_struct* sh)
{
  SH2tstm(sh,(4<<4)+8);
}
#define init_12_12_4_9
void FASTCALL call_12_12_4_9(SH2_struct* sh)
{
  SH2tstm(sh,(4<<4)+9);
}
#define init_12_12_4_10
void FASTCALL call_12_12_4_10(SH2_struct* sh)
{
  SH2tstm(sh,(4<<4)+10);
}
#define init_12_12_4_11
void FASTCALL call_12_12_4_11(SH2_struct* sh)
{
  SH2tstm(sh,(4<<4)+11);
}
#define init_12_12_4_12
void FASTCALL call_12_12_4_12(SH2_struct* sh)
{
  SH2tstm(sh,(4<<4)+12);
}
#define init_12_12_4_13
void FASTCALL call_12_12_4_13(SH2_struct* sh)
{
  SH2tstm(sh,(4<<4)+13);
}
#define init_12_12_4_14
void FASTCALL call_12_12_4_14(SH2_struct* sh)
{
  SH2tstm(sh,(4<<4)+14);
}
#define init_12_12_4_15
void FASTCALL call_12_12_4_15(SH2_struct* sh)
{
  SH2tstm(sh,(4<<4)+15);
}
#define init_12_12_5_0
void FASTCALL call_12_12_5_0(SH2_struct* sh)
{
  SH2tstm(sh,(5<<4)+0);
}
#define init_12_12_5_1
void FASTCALL call_12_12_5_1(SH2_struct* sh)
{
  SH2tstm(sh,(5<<4)+1);
}
#define init_12_12_5_2
void FASTCALL call_12_12_5_2(SH2_struct* sh)
{
  SH2tstm(sh,(5<<4)+2);
}
#define init_12_12_5_3
void FASTCALL call_12_12_5_3(SH2_struct* sh)
{
  SH2tstm(sh,(5<<4)+3);
}
#define init_12_12_5_4
void FASTCALL call_12_12_5_4(SH2_struct* sh)
{
  SH2tstm(sh,(5<<4)+4);
}
#define init_12_12_5_5
void FASTCALL call_12_12_5_5(SH2_struct* sh)
{
  SH2tstm(sh,(5<<4)+5);
}
#define init_12_12_5_6
void FASTCALL call_12_12_5_6(SH2_struct* sh)
{
  SH2tstm(sh,(5<<4)+6);
}
#define init_12_12_5_7
void FASTCALL call_12_12_5_7(SH2_struct* sh)
{
  SH2tstm(sh,(5<<4)+7);
}
#define init_12_12_5_8
void FASTCALL call_12_12_5_8(SH2_struct* sh)
{
  SH2tstm(sh,(5<<4)+8);
}
#define init_12_12_5_9
void FASTCALL call_12_12_5_9(SH2_struct* sh)
{
  SH2tstm(sh,(5<<4)+9);
}
#define init_12_12_5_10
void FASTCALL call_12_12_5_10(SH2_struct* sh)
{
  SH2tstm(sh,(5<<4)+10);
}
#define init_12_12_5_11
void FASTCALL call_12_12_5_11(SH2_struct* sh)
{
  SH2tstm(sh,(5<<4)+11);
}
#define init_12_12_5_12
void FASTCALL call_12_12_5_12(SH2_struct* sh)
{
  SH2tstm(sh,(5<<4)+12);
}
#define init_12_12_5_13
void FASTCALL call_12_12_5_13(SH2_struct* sh)
{
  SH2tstm(sh,(5<<4)+13);
}
#define init_12_12_5_14
void FASTCALL call_12_12_5_14(SH2_struct* sh)
{
  SH2tstm(sh,(5<<4)+14);
}
#define init_12_12_5_15
void FASTCALL call_12_12_5_15(SH2_struct* sh)
{
  SH2tstm(sh,(5<<4)+15);
}
#define init_12_12_6_0
void FASTCALL call_12_12_6_0(SH2_struct* sh)
{
  SH2tstm(sh,(6<<4)+0);
}
#define init_12_12_6_1
void FASTCALL call_12_12_6_1(SH2_struct* sh)
{
  SH2tstm(sh,(6<<4)+1);
}
#define init_12_12_6_2
void FASTCALL call_12_12_6_2(SH2_struct* sh)
{
  SH2tstm(sh,(6<<4)+2);
}
#define init_12_12_6_3
void FASTCALL call_12_12_6_3(SH2_struct* sh)
{
  SH2tstm(sh,(6<<4)+3);
}
#define init_12_12_6_4
void FASTCALL call_12_12_6_4(SH2_struct* sh)
{
  SH2tstm(sh,(6<<4)+4);
}
#define init_12_12_6_5
void FASTCALL call_12_12_6_5(SH2_struct* sh)
{
  SH2tstm(sh,(6<<4)+5);
}
#define init_12_12_6_6
void FASTCALL call_12_12_6_6(SH2_struct* sh)
{
  SH2tstm(sh,(6<<4)+6);
}
#define init_12_12_6_7
void FASTCALL call_12_12_6_7(SH2_struct* sh)
{
  SH2tstm(sh,(6<<4)+7);
}
#define init_12_12_6_8
void FASTCALL call_12_12_6_8(SH2_struct* sh)
{
  SH2tstm(sh,(6<<4)+8);
}
#define init_12_12_6_9
void FASTCALL call_12_12_6_9(SH2_struct* sh)
{
  SH2tstm(sh,(6<<4)+9);
}
#define init_12_12_6_10
void FASTCALL call_12_12_6_10(SH2_struct* sh)
{
  SH2tstm(sh,(6<<4)+10);
}
#define init_12_12_6_11
void FASTCALL call_12_12_6_11(SH2_struct* sh)
{
  SH2tstm(sh,(6<<4)+11);
}
#define init_12_12_6_12
void FASTCALL call_12_12_6_12(SH2_struct* sh)
{
  SH2tstm(sh,(6<<4)+12);
}
#define init_12_12_6_13
void FASTCALL call_12_12_6_13(SH2_struct* sh)
{
  SH2tstm(sh,(6<<4)+13);
}
#define init_12_12_6_14
void FASTCALL call_12_12_6_14(SH2_struct* sh)
{
  SH2tstm(sh,(6<<4)+14);
}
#define init_12_12_6_15
void FASTCALL call_12_12_6_15(SH2_struct* sh)
{
  SH2tstm(sh,(6<<4)+15);
}
#define init_12_12_7_0
void FASTCALL call_12_12_7_0(SH2_struct* sh)
{
  SH2tstm(sh,(7<<4)+0);
}
#define init_12_12_7_1
void FASTCALL call_12_12_7_1(SH2_struct* sh)
{
  SH2tstm(sh,(7<<4)+1);
}
#define init_12_12_7_2
void FASTCALL call_12_12_7_2(SH2_struct* sh)
{
  SH2tstm(sh,(7<<4)+2);
}
#define init_12_12_7_3
void FASTCALL call_12_12_7_3(SH2_struct* sh)
{
  SH2tstm(sh,(7<<4)+3);
}
#define init_12_12_7_4
void FASTCALL call_12_12_7_4(SH2_struct* sh)
{
  SH2tstm(sh,(7<<4)+4);
}
#define init_12_12_7_5
void FASTCALL call_12_12_7_5(SH2_struct* sh)
{
  SH2tstm(sh,(7<<4)+5);
}
#define init_12_12_7_6
void FASTCALL call_12_12_7_6(SH2_struct* sh)
{
  SH2tstm(sh,(7<<4)+6);
}
#define init_12_12_7_7
void FASTCALL call_12_12_7_7(SH2_struct* sh)
{
  SH2tstm(sh,(7<<4)+7);
}
#define init_12_12_7_8
void FASTCALL call_12_12_7_8(SH2_struct* sh)
{
  SH2tstm(sh,(7<<4)+8);
}
#define init_12_12_7_9
void FASTCALL call_12_12_7_9(SH2_struct* sh)
{
  SH2tstm(sh,(7<<4)+9);
}
#define init_12_12_7_10
void FASTCALL call_12_12_7_10(SH2_struct* sh)
{
  SH2tstm(sh,(7<<4)+10);
}
#define init_12_12_7_11
void FASTCALL call_12_12_7_11(SH2_struct* sh)
{
  SH2tstm(sh,(7<<4)+11);
}
#define init_12_12_7_12
void FASTCALL call_12_12_7_12(SH2_struct* sh)
{
  SH2tstm(sh,(7<<4)+12);
}
#define init_12_12_7_13
void FASTCALL call_12_12_7_13(SH2_struct* sh)
{
  SH2tstm(sh,(7<<4)+13);
}
#define init_12_12_7_14
void FASTCALL call_12_12_7_14(SH2_struct* sh)
{
  SH2tstm(sh,(7<<4)+14);
}
#define init_12_12_7_15
void FASTCALL call_12_12_7_15(SH2_struct* sh)
{
  SH2tstm(sh,(7<<4)+15);
}
#define init_12_12_8_0
void FASTCALL call_12_12_8_0(SH2_struct* sh)
{
  SH2tstm(sh,(8<<4)+0);
}
#define init_12_12_8_1
void FASTCALL call_12_12_8_1(SH2_struct* sh)
{
  SH2tstm(sh,(8<<4)+1);
}
#define init_12_12_8_2
void FASTCALL call_12_12_8_2(SH2_struct* sh)
{
  SH2tstm(sh,(8<<4)+2);
}
#define init_12_12_8_3
void FASTCALL call_12_12_8_3(SH2_struct* sh)
{
  SH2tstm(sh,(8<<4)+3);
}
#define init_12_12_8_4
void FASTCALL call_12_12_8_4(SH2_struct* sh)
{
  SH2tstm(sh,(8<<4)+4);
}
#define init_12_12_8_5
void FASTCALL call_12_12_8_5(SH2_struct* sh)
{
  SH2tstm(sh,(8<<4)+5);
}
#define init_12_12_8_6
void FASTCALL call_12_12_8_6(SH2_struct* sh)
{
  SH2tstm(sh,(8<<4)+6);
}
#define init_12_12_8_7
void FASTCALL call_12_12_8_7(SH2_struct* sh)
{
  SH2tstm(sh,(8<<4)+7);
}
#define init_12_12_8_8
void FASTCALL call_12_12_8_8(SH2_struct* sh)
{
  SH2tstm(sh,(8<<4)+8);
}
#define init_12_12_8_9
void FASTCALL call_12_12_8_9(SH2_struct* sh)
{
  SH2tstm(sh,(8<<4)+9);
}
#define init_12_12_8_10
void FASTCALL call_12_12_8_10(SH2_struct* sh)
{
  SH2tstm(sh,(8<<4)+10);
}
#define init_12_12_8_11
void FASTCALL call_12_12_8_11(SH2_struct* sh)
{
  SH2tstm(sh,(8<<4)+11);
}
#define init_12_12_8_12
void FASTCALL call_12_12_8_12(SH2_struct* sh)
{
  SH2tstm(sh,(8<<4)+12);
}
#define init_12_12_8_13
void FASTCALL call_12_12_8_13(SH2_struct* sh)
{
  SH2tstm(sh,(8<<4)+13);
}
#define init_12_12_8_14
void FASTCALL call_12_12_8_14(SH2_struct* sh)
{
  SH2tstm(sh,(8<<4)+14);
}
#define init_12_12_8_15
void FASTCALL call_12_12_8_15(SH2_struct* sh)
{
  SH2tstm(sh,(8<<4)+15);
}
#define init_12_12_9_0
void FASTCALL call_12_12_9_0(SH2_struct* sh)
{
  SH2tstm(sh,(9<<4)+0);
}
#define init_12_12_9_1
void FASTCALL call_12_12_9_1(SH2_struct* sh)
{
  SH2tstm(sh,(9<<4)+1);
}
#define init_12_12_9_2
void FASTCALL call_12_12_9_2(SH2_struct* sh)
{
  SH2tstm(sh,(9<<4)+2);
}
#define init_12_12_9_3
void FASTCALL call_12_12_9_3(SH2_struct* sh)
{
  SH2tstm(sh,(9<<4)+3);
}
#define init_12_12_9_4
void FASTCALL call_12_12_9_4(SH2_struct* sh)
{
  SH2tstm(sh,(9<<4)+4);
}
#define init_12_12_9_5
void FASTCALL call_12_12_9_5(SH2_struct* sh)
{
  SH2tstm(sh,(9<<4)+5);
}
#define init_12_12_9_6
void FASTCALL call_12_12_9_6(SH2_struct* sh)
{
  SH2tstm(sh,(9<<4)+6);
}
#define init_12_12_9_7
void FASTCALL call_12_12_9_7(SH2_struct* sh)
{
  SH2tstm(sh,(9<<4)+7);
}
#define init_12_12_9_8
void FASTCALL call_12_12_9_8(SH2_struct* sh)
{
  SH2tstm(sh,(9<<4)+8);
}
#define init_12_12_9_9
void FASTCALL call_12_12_9_9(SH2_struct* sh)
{
  SH2tstm(sh,(9<<4)+9);
}
#define init_12_12_9_10
void FASTCALL call_12_12_9_10(SH2_struct* sh)
{
  SH2tstm(sh,(9<<4)+10);
}
#define init_12_12_9_11
void FASTCALL call_12_12_9_11(SH2_struct* sh)
{
  SH2tstm(sh,(9<<4)+11);
}
#define init_12_12_9_12
void FASTCALL call_12_12_9_12(SH2_struct* sh)
{
  SH2tstm(sh,(9<<4)+12);
}
#define init_12_12_9_13
void FASTCALL call_12_12_9_13(SH2_struct* sh)
{
  SH2tstm(sh,(9<<4)+13);
}
#define init_12_12_9_14
void FASTCALL call_12_12_9_14(SH2_struct* sh)
{
  SH2tstm(sh,(9<<4)+14);
}
#define init_12_12_9_15
void FASTCALL call_12_12_9_15(SH2_struct* sh)
{
  SH2tstm(sh,(9<<4)+15);
}
#define init_12_12_10_0
void FASTCALL call_12_12_10_0(SH2_struct* sh)
{
  SH2tstm(sh,(10<<4)+0);
}
#define init_12_12_10_1
void FASTCALL call_12_12_10_1(SH2_struct* sh)
{
  SH2tstm(sh,(10<<4)+1);
}
#define init_12_12_10_2
void FASTCALL call_12_12_10_2(SH2_struct* sh)
{
  SH2tstm(sh,(10<<4)+2);
}
#define init_12_12_10_3
void FASTCALL call_12_12_10_3(SH2_struct* sh)
{
  SH2tstm(sh,(10<<4)+3);
}
#define init_12_12_10_4
void FASTCALL call_12_12_10_4(SH2_struct* sh)
{
  SH2tstm(sh,(10<<4)+4);
}
#define init_12_12_10_5
void FASTCALL call_12_12_10_5(SH2_struct* sh)
{
  SH2tstm(sh,(10<<4)+5);
}
#define init_12_12_10_6
void FASTCALL call_12_12_10_6(SH2_struct* sh)
{
  SH2tstm(sh,(10<<4)+6);
}
#define init_12_12_10_7
void FASTCALL call_12_12_10_7(SH2_struct* sh)
{
  SH2tstm(sh,(10<<4)+7);
}
#define init_12_12_10_8
void FASTCALL call_12_12_10_8(SH2_struct* sh)
{
  SH2tstm(sh,(10<<4)+8);
}
#define init_12_12_10_9
void FASTCALL call_12_12_10_9(SH2_struct* sh)
{
  SH2tstm(sh,(10<<4)+9);
}
#define init_12_12_10_10
void FASTCALL call_12_12_10_10(SH2_struct* sh)
{
  SH2tstm(sh,(10<<4)+10);
}
#define init_12_12_10_11
void FASTCALL call_12_12_10_11(SH2_struct* sh)
{
  SH2tstm(sh,(10<<4)+11);
}
#define init_12_12_10_12
void FASTCALL call_12_12_10_12(SH2_struct* sh)
{
  SH2tstm(sh,(10<<4)+12);
}
#define init_12_12_10_13
void FASTCALL call_12_12_10_13(SH2_struct* sh)
{
  SH2tstm(sh,(10<<4)+13);
}
#define init_12_12_10_14
void FASTCALL call_12_12_10_14(SH2_struct* sh)
{
  SH2tstm(sh,(10<<4)+14);
}
#define init_12_12_10_15
void FASTCALL call_12_12_10_15(SH2_struct* sh)
{
  SH2tstm(sh,(10<<4)+15);
}
#define init_12_12_11_0
void FASTCALL call_12_12_11_0(SH2_struct* sh)
{
  SH2tstm(sh,(11<<4)+0);
}
#define init_12_12_11_1
void FASTCALL call_12_12_11_1(SH2_struct* sh)
{
  SH2tstm(sh,(11<<4)+1);
}
#define init_12_12_11_2
void FASTCALL call_12_12_11_2(SH2_struct* sh)
{
  SH2tstm(sh,(11<<4)+2);
}
#define init_12_12_11_3
void FASTCALL call_12_12_11_3(SH2_struct* sh)
{
  SH2tstm(sh,(11<<4)+3);
}
#define init_12_12_11_4
void FASTCALL call_12_12_11_4(SH2_struct* sh)
{
  SH2tstm(sh,(11<<4)+4);
}
#define init_12_12_11_5
void FASTCALL call_12_12_11_5(SH2_struct* sh)
{
  SH2tstm(sh,(11<<4)+5);
}
#define init_12_12_11_6
void FASTCALL call_12_12_11_6(SH2_struct* sh)
{
  SH2tstm(sh,(11<<4)+6);
}
#define init_12_12_11_7
void FASTCALL call_12_12_11_7(SH2_struct* sh)
{
  SH2tstm(sh,(11<<4)+7);
}
#define init_12_12_11_8
void FASTCALL call_12_12_11_8(SH2_struct* sh)
{
  SH2tstm(sh,(11<<4)+8);
}
#define init_12_12_11_9
void FASTCALL call_12_12_11_9(SH2_struct* sh)
{
  SH2tstm(sh,(11<<4)+9);
}
#define init_12_12_11_10
void FASTCALL call_12_12_11_10(SH2_struct* sh)
{
  SH2tstm(sh,(11<<4)+10);
}
#define init_12_12_11_11
void FASTCALL call_12_12_11_11(SH2_struct* sh)
{
  SH2tstm(sh,(11<<4)+11);
}
#define init_12_12_11_12
void FASTCALL call_12_12_11_12(SH2_struct* sh)
{
  SH2tstm(sh,(11<<4)+12);
}
#define init_12_12_11_13
void FASTCALL call_12_12_11_13(SH2_struct* sh)
{
  SH2tstm(sh,(11<<4)+13);
}
#define init_12_12_11_14
void FASTCALL call_12_12_11_14(SH2_struct* sh)
{
  SH2tstm(sh,(11<<4)+14);
}
#define init_12_12_11_15
void FASTCALL call_12_12_11_15(SH2_struct* sh)
{
  SH2tstm(sh,(11<<4)+15);
}
#define init_12_12_12_0
void FASTCALL call_12_12_12_0(SH2_struct* sh)
{
  SH2tstm(sh,(12<<4)+0);
}
#define init_12_12_12_1
void FASTCALL call_12_12_12_1(SH2_struct* sh)
{
  SH2tstm(sh,(12<<4)+1);
}
#define init_12_12_12_2
void FASTCALL call_12_12_12_2(SH2_struct* sh)
{
  SH2tstm(sh,(12<<4)+2);
}
#define init_12_12_12_3
void FASTCALL call_12_12_12_3(SH2_struct* sh)
{
  SH2tstm(sh,(12<<4)+3);
}
#define init_12_12_12_4
void FASTCALL call_12_12_12_4(SH2_struct* sh)
{
  SH2tstm(sh,(12<<4)+4);
}
#define init_12_12_12_5
void FASTCALL call_12_12_12_5(SH2_struct* sh)
{
  SH2tstm(sh,(12<<4)+5);
}
#define init_12_12_12_6
void FASTCALL call_12_12_12_6(SH2_struct* sh)
{
  SH2tstm(sh,(12<<4)+6);
}
#define init_12_12_12_7
void FASTCALL call_12_12_12_7(SH2_struct* sh)
{
  SH2tstm(sh,(12<<4)+7);
}
#define init_12_12_12_8
void FASTCALL call_12_12_12_8(SH2_struct* sh)
{
  SH2tstm(sh,(12<<4)+8);
}
#define init_12_12_12_9
void FASTCALL call_12_12_12_9(SH2_struct* sh)
{
  SH2tstm(sh,(12<<4)+9);
}
#define init_12_12_12_10
void FASTCALL call_12_12_12_10(SH2_struct* sh)
{
  SH2tstm(sh,(12<<4)+10);
}
#define init_12_12_12_11
void FASTCALL call_12_12_12_11(SH2_struct* sh)
{
  SH2tstm(sh,(12<<4)+11);
}
#define init_12_12_12_12
void FASTCALL call_12_12_12_12(SH2_struct* sh)
{
  SH2tstm(sh,(12<<4)+12);
}
#define init_12_12_12_13
void FASTCALL call_12_12_12_13(SH2_struct* sh)
{
  SH2tstm(sh,(12<<4)+13);
}
#define init_12_12_12_14
void FASTCALL call_12_12_12_14(SH2_struct* sh)
{
  SH2tstm(sh,(12<<4)+14);
}
#define init_12_12_12_15
void FASTCALL call_12_12_12_15(SH2_struct* sh)
{
  SH2tstm(sh,(12<<4)+15);
}
#define init_12_12_13_0
void FASTCALL call_12_12_13_0(SH2_struct* sh)
{
  SH2tstm(sh,(13<<4)+0);
}
#define init_12_12_13_1
void FASTCALL call_12_12_13_1(SH2_struct* sh)
{
  SH2tstm(sh,(13<<4)+1);
}
#define init_12_12_13_2
void FASTCALL call_12_12_13_2(SH2_struct* sh)
{
  SH2tstm(sh,(13<<4)+2);
}
#define init_12_12_13_3
void FASTCALL call_12_12_13_3(SH2_struct* sh)
{
  SH2tstm(sh,(13<<4)+3);
}
#define init_12_12_13_4
void FASTCALL call_12_12_13_4(SH2_struct* sh)
{
  SH2tstm(sh,(13<<4)+4);
}
#define init_12_12_13_5
void FASTCALL call_12_12_13_5(SH2_struct* sh)
{
  SH2tstm(sh,(13<<4)+5);
}
#define init_12_12_13_6
void FASTCALL call_12_12_13_6(SH2_struct* sh)
{
  SH2tstm(sh,(13<<4)+6);
}
#define init_12_12_13_7
void FASTCALL call_12_12_13_7(SH2_struct* sh)
{
  SH2tstm(sh,(13<<4)+7);
}
#define init_12_12_13_8
void FASTCALL call_12_12_13_8(SH2_struct* sh)
{
  SH2tstm(sh,(13<<4)+8);
}
#define init_12_12_13_9
void FASTCALL call_12_12_13_9(SH2_struct* sh)
{
  SH2tstm(sh,(13<<4)+9);
}
#define init_12_12_13_10
void FASTCALL call_12_12_13_10(SH2_struct* sh)
{
  SH2tstm(sh,(13<<4)+10);
}
#define init_12_12_13_11
void FASTCALL call_12_12_13_11(SH2_struct* sh)
{
  SH2tstm(sh,(13<<4)+11);
}
#define init_12_12_13_12
void FASTCALL call_12_12_13_12(SH2_struct* sh)
{
  SH2tstm(sh,(13<<4)+12);
}
#define init_12_12_13_13
void FASTCALL call_12_12_13_13(SH2_struct* sh)
{
  SH2tstm(sh,(13<<4)+13);
}
#define init_12_12_13_14
void FASTCALL call_12_12_13_14(SH2_struct* sh)
{
  SH2tstm(sh,(13<<4)+14);
}
#define init_12_12_13_15
void FASTCALL call_12_12_13_15(SH2_struct* sh)
{
  SH2tstm(sh,(13<<4)+15);
}
#define init_12_12_14_0
void FASTCALL call_12_12_14_0(SH2_struct* sh)
{
  SH2tstm(sh,(14<<4)+0);
}
#define init_12_12_14_1
void FASTCALL call_12_12_14_1(SH2_struct* sh)
{
  SH2tstm(sh,(14<<4)+1);
}
#define init_12_12_14_2
void FASTCALL call_12_12_14_2(SH2_struct* sh)
{
  SH2tstm(sh,(14<<4)+2);
}
#define init_12_12_14_3
void FASTCALL call_12_12_14_3(SH2_struct* sh)
{
  SH2tstm(sh,(14<<4)+3);
}
#define init_12_12_14_4
void FASTCALL call_12_12_14_4(SH2_struct* sh)
{
  SH2tstm(sh,(14<<4)+4);
}
#define init_12_12_14_5
void FASTCALL call_12_12_14_5(SH2_struct* sh)
{
  SH2tstm(sh,(14<<4)+5);
}
#define init_12_12_14_6
void FASTCALL call_12_12_14_6(SH2_struct* sh)
{
  SH2tstm(sh,(14<<4)+6);
}
#define init_12_12_14_7
void FASTCALL call_12_12_14_7(SH2_struct* sh)
{
  SH2tstm(sh,(14<<4)+7);
}
#define init_12_12_14_8
void FASTCALL call_12_12_14_8(SH2_struct* sh)
{
  SH2tstm(sh,(14<<4)+8);
}
#define init_12_12_14_9
void FASTCALL call_12_12_14_9(SH2_struct* sh)
{
  SH2tstm(sh,(14<<4)+9);
}
#define init_12_12_14_10
void FASTCALL call_12_12_14_10(SH2_struct* sh)
{
  SH2tstm(sh,(14<<4)+10);
}
#define init_12_12_14_11
void FASTCALL call_12_12_14_11(SH2_struct* sh)
{
  SH2tstm(sh,(14<<4)+11);
}
#define init_12_12_14_12
void FASTCALL call_12_12_14_12(SH2_struct* sh)
{
  SH2tstm(sh,(14<<4)+12);
}
#define init_12_12_14_13
void FASTCALL call_12_12_14_13(SH2_struct* sh)
{
  SH2tstm(sh,(14<<4)+13);
}
#define init_12_12_14_14
void FASTCALL call_12_12_14_14(SH2_struct* sh)
{
  SH2tstm(sh,(14<<4)+14);
}
#define init_12_12_14_15
void FASTCALL call_12_12_14_15(SH2_struct* sh)
{
  SH2tstm(sh,(14<<4)+15);
}
#define init_12_12_15_0
void FASTCALL call_12_12_15_0(SH2_struct* sh)
{
  SH2tstm(sh,(15<<4)+0);
}
#define init_12_12_15_1
void FASTCALL call_12_12_15_1(SH2_struct* sh)
{
  SH2tstm(sh,(15<<4)+1);
}
#define init_12_12_15_2
void FASTCALL call_12_12_15_2(SH2_struct* sh)
{
  SH2tstm(sh,(15<<4)+2);
}
#define init_12_12_15_3
void FASTCALL call_12_12_15_3(SH2_struct* sh)
{
  SH2tstm(sh,(15<<4)+3);
}
#define init_12_12_15_4
void FASTCALL call_12_12_15_4(SH2_struct* sh)
{
  SH2tstm(sh,(15<<4)+4);
}
#define init_12_12_15_5
void FASTCALL call_12_12_15_5(SH2_struct* sh)
{
  SH2tstm(sh,(15<<4)+5);
}
#define init_12_12_15_6
void FASTCALL call_12_12_15_6(SH2_struct* sh)
{
  SH2tstm(sh,(15<<4)+6);
}
#define init_12_12_15_7
void FASTCALL call_12_12_15_7(SH2_struct* sh)
{
  SH2tstm(sh,(15<<4)+7);
}
#define init_12_12_15_8
void FASTCALL call_12_12_15_8(SH2_struct* sh)
{
  SH2tstm(sh,(15<<4)+8);
}
#define init_12_12_15_9
void FASTCALL call_12_12_15_9(SH2_struct* sh)
{
  SH2tstm(sh,(15<<4)+9);
}
#define init_12_12_15_10
void FASTCALL call_12_12_15_10(SH2_struct* sh)
{
  SH2tstm(sh,(15<<4)+10);
}
#define init_12_12_15_11
void FASTCALL call_12_12_15_11(SH2_struct* sh)
{
  SH2tstm(sh,(15<<4)+11);
}
#define init_12_12_15_12
void FASTCALL call_12_12_15_12(SH2_struct* sh)
{
  SH2tstm(sh,(15<<4)+12);
}
#define init_12_12_15_13
void FASTCALL call_12_12_15_13(SH2_struct* sh)
{
  SH2tstm(sh,(15<<4)+13);
}
#define init_12_12_15_14
void FASTCALL call_12_12_15_14(SH2_struct* sh)
{
  SH2tstm(sh,(15<<4)+14);
}
#define init_12_12_15_15
void FASTCALL call_12_12_15_15(SH2_struct* sh)
{
  SH2tstm(sh,(15<<4)+15);
}
#define init_12_8_0_0
void FASTCALL call_12_8_0_0(SH2_struct* sh)
{
  SH2tsti(sh,(0<<4)+0);
}
#define init_12_8_0_1
void FASTCALL call_12_8_0_1(SH2_struct* sh)
{
  SH2tsti(sh,(0<<4)+1);
}
#define init_12_8_0_2
void FASTCALL call_12_8_0_2(SH2_struct* sh)
{
  SH2tsti(sh,(0<<4)+2);
}
#define init_12_8_0_3
void FASTCALL call_12_8_0_3(SH2_struct* sh)
{
  SH2tsti(sh,(0<<4)+3);
}
#define init_12_8_0_4
void FASTCALL call_12_8_0_4(SH2_struct* sh)
{
  SH2tsti(sh,(0<<4)+4);
}
#define init_12_8_0_5
void FASTCALL call_12_8_0_5(SH2_struct* sh)
{
  SH2tsti(sh,(0<<4)+5);
}
#define init_12_8_0_6
void FASTCALL call_12_8_0_6(SH2_struct* sh)
{
  SH2tsti(sh,(0<<4)+6);
}
#define init_12_8_0_7
void FASTCALL call_12_8_0_7(SH2_struct* sh)
{
  SH2tsti(sh,(0<<4)+7);
}
#define init_12_8_0_8
void FASTCALL call_12_8_0_8(SH2_struct* sh)
{
  SH2tsti(sh,(0<<4)+8);
}
#define init_12_8_0_9
void FASTCALL call_12_8_0_9(SH2_struct* sh)
{
  SH2tsti(sh,(0<<4)+9);
}
#define init_12_8_0_10
void FASTCALL call_12_8_0_10(SH2_struct* sh)
{
  SH2tsti(sh,(0<<4)+10);
}
#define init_12_8_0_11
void FASTCALL call_12_8_0_11(SH2_struct* sh)
{
  SH2tsti(sh,(0<<4)+11);
}
#define init_12_8_0_12
void FASTCALL call_12_8_0_12(SH2_struct* sh)
{
  SH2tsti(sh,(0<<4)+12);
}
#define init_12_8_0_13
void FASTCALL call_12_8_0_13(SH2_struct* sh)
{
  SH2tsti(sh,(0<<4)+13);
}
#define init_12_8_0_14
void FASTCALL call_12_8_0_14(SH2_struct* sh)
{
  SH2tsti(sh,(0<<4)+14);
}
#define init_12_8_0_15
void FASTCALL call_12_8_0_15(SH2_struct* sh)
{
  SH2tsti(sh,(0<<4)+15);
}
#define init_12_8_1_0
void FASTCALL call_12_8_1_0(SH2_struct* sh)
{
  SH2tsti(sh,(1<<4)+0);
}
#define init_12_8_1_1
void FASTCALL call_12_8_1_1(SH2_struct* sh)
{
  SH2tsti(sh,(1<<4)+1);
}
#define init_12_8_1_2
void FASTCALL call_12_8_1_2(SH2_struct* sh)
{
  SH2tsti(sh,(1<<4)+2);
}
#define init_12_8_1_3
void FASTCALL call_12_8_1_3(SH2_struct* sh)
{
  SH2tsti(sh,(1<<4)+3);
}
#define init_12_8_1_4
void FASTCALL call_12_8_1_4(SH2_struct* sh)
{
  SH2tsti(sh,(1<<4)+4);
}
#define init_12_8_1_5
void FASTCALL call_12_8_1_5(SH2_struct* sh)
{
  SH2tsti(sh,(1<<4)+5);
}
#define init_12_8_1_6
void FASTCALL call_12_8_1_6(SH2_struct* sh)
{
  SH2tsti(sh,(1<<4)+6);
}
#define init_12_8_1_7
void FASTCALL call_12_8_1_7(SH2_struct* sh)
{
  SH2tsti(sh,(1<<4)+7);
}
#define init_12_8_1_8
void FASTCALL call_12_8_1_8(SH2_struct* sh)
{
  SH2tsti(sh,(1<<4)+8);
}
#define init_12_8_1_9
void FASTCALL call_12_8_1_9(SH2_struct* sh)
{
  SH2tsti(sh,(1<<4)+9);
}
#define init_12_8_1_10
void FASTCALL call_12_8_1_10(SH2_struct* sh)
{
  SH2tsti(sh,(1<<4)+10);
}
#define init_12_8_1_11
void FASTCALL call_12_8_1_11(SH2_struct* sh)
{
  SH2tsti(sh,(1<<4)+11);
}
#define init_12_8_1_12
void FASTCALL call_12_8_1_12(SH2_struct* sh)
{
  SH2tsti(sh,(1<<4)+12);
}
#define init_12_8_1_13
void FASTCALL call_12_8_1_13(SH2_struct* sh)
{
  SH2tsti(sh,(1<<4)+13);
}
#define init_12_8_1_14
void FASTCALL call_12_8_1_14(SH2_struct* sh)
{
  SH2tsti(sh,(1<<4)+14);
}
#define init_12_8_1_15
void FASTCALL call_12_8_1_15(SH2_struct* sh)
{
  SH2tsti(sh,(1<<4)+15);
}
#define init_12_8_2_0
void FASTCALL call_12_8_2_0(SH2_struct* sh)
{
  SH2tsti(sh,(2<<4)+0);
}
#define init_12_8_2_1
void FASTCALL call_12_8_2_1(SH2_struct* sh)
{
  SH2tsti(sh,(2<<4)+1);
}
#define init_12_8_2_2
void FASTCALL call_12_8_2_2(SH2_struct* sh)
{
  SH2tsti(sh,(2<<4)+2);
}
#define init_12_8_2_3
void FASTCALL call_12_8_2_3(SH2_struct* sh)
{
  SH2tsti(sh,(2<<4)+3);
}
#define init_12_8_2_4
void FASTCALL call_12_8_2_4(SH2_struct* sh)
{
  SH2tsti(sh,(2<<4)+4);
}
#define init_12_8_2_5
void FASTCALL call_12_8_2_5(SH2_struct* sh)
{
  SH2tsti(sh,(2<<4)+5);
}
#define init_12_8_2_6
void FASTCALL call_12_8_2_6(SH2_struct* sh)
{
  SH2tsti(sh,(2<<4)+6);
}
#define init_12_8_2_7
void FASTCALL call_12_8_2_7(SH2_struct* sh)
{
  SH2tsti(sh,(2<<4)+7);
}
#define init_12_8_2_8
void FASTCALL call_12_8_2_8(SH2_struct* sh)
{
  SH2tsti(sh,(2<<4)+8);
}
#define init_12_8_2_9
void FASTCALL call_12_8_2_9(SH2_struct* sh)
{
  SH2tsti(sh,(2<<4)+9);
}
#define init_12_8_2_10
void FASTCALL call_12_8_2_10(SH2_struct* sh)
{
  SH2tsti(sh,(2<<4)+10);
}
#define init_12_8_2_11
void FASTCALL call_12_8_2_11(SH2_struct* sh)
{
  SH2tsti(sh,(2<<4)+11);
}
#define init_12_8_2_12
void FASTCALL call_12_8_2_12(SH2_struct* sh)
{
  SH2tsti(sh,(2<<4)+12);
}
#define init_12_8_2_13
void FASTCALL call_12_8_2_13(SH2_struct* sh)
{
  SH2tsti(sh,(2<<4)+13);
}
#define init_12_8_2_14
void FASTCALL call_12_8_2_14(SH2_struct* sh)
{
  SH2tsti(sh,(2<<4)+14);
}
#define init_12_8_2_15
void FASTCALL call_12_8_2_15(SH2_struct* sh)
{
  SH2tsti(sh,(2<<4)+15);
}
#define init_12_8_3_0
void FASTCALL call_12_8_3_0(SH2_struct* sh)
{
  SH2tsti(sh,(3<<4)+0);
}
#define init_12_8_3_1
void FASTCALL call_12_8_3_1(SH2_struct* sh)
{
  SH2tsti(sh,(3<<4)+1);
}
#define init_12_8_3_2
void FASTCALL call_12_8_3_2(SH2_struct* sh)
{
  SH2tsti(sh,(3<<4)+2);
}
#define init_12_8_3_3
void FASTCALL call_12_8_3_3(SH2_struct* sh)
{
  SH2tsti(sh,(3<<4)+3);
}
#define init_12_8_3_4
void FASTCALL call_12_8_3_4(SH2_struct* sh)
{
  SH2tsti(sh,(3<<4)+4);
}
#define init_12_8_3_5
void FASTCALL call_12_8_3_5(SH2_struct* sh)
{
  SH2tsti(sh,(3<<4)+5);
}
#define init_12_8_3_6
void FASTCALL call_12_8_3_6(SH2_struct* sh)
{
  SH2tsti(sh,(3<<4)+6);
}
#define init_12_8_3_7
void FASTCALL call_12_8_3_7(SH2_struct* sh)
{
  SH2tsti(sh,(3<<4)+7);
}
#define init_12_8_3_8
void FASTCALL call_12_8_3_8(SH2_struct* sh)
{
  SH2tsti(sh,(3<<4)+8);
}
#define init_12_8_3_9
void FASTCALL call_12_8_3_9(SH2_struct* sh)
{
  SH2tsti(sh,(3<<4)+9);
}
#define init_12_8_3_10
void FASTCALL call_12_8_3_10(SH2_struct* sh)
{
  SH2tsti(sh,(3<<4)+10);
}
#define init_12_8_3_11
void FASTCALL call_12_8_3_11(SH2_struct* sh)
{
  SH2tsti(sh,(3<<4)+11);
}
#define init_12_8_3_12
void FASTCALL call_12_8_3_12(SH2_struct* sh)
{
  SH2tsti(sh,(3<<4)+12);
}
#define init_12_8_3_13
void FASTCALL call_12_8_3_13(SH2_struct* sh)
{
  SH2tsti(sh,(3<<4)+13);
}
#define init_12_8_3_14
void FASTCALL call_12_8_3_14(SH2_struct* sh)
{
  SH2tsti(sh,(3<<4)+14);
}
#define init_12_8_3_15
void FASTCALL call_12_8_3_15(SH2_struct* sh)
{
  SH2tsti(sh,(3<<4)+15);
}
#define init_12_8_4_0
void FASTCALL call_12_8_4_0(SH2_struct* sh)
{
  SH2tsti(sh,(4<<4)+0);
}
#define init_12_8_4_1
void FASTCALL call_12_8_4_1(SH2_struct* sh)
{
  SH2tsti(sh,(4<<4)+1);
}
#define init_12_8_4_2
void FASTCALL call_12_8_4_2(SH2_struct* sh)
{
  SH2tsti(sh,(4<<4)+2);
}
#define init_12_8_4_3
void FASTCALL call_12_8_4_3(SH2_struct* sh)
{
  SH2tsti(sh,(4<<4)+3);
}
#define init_12_8_4_4
void FASTCALL call_12_8_4_4(SH2_struct* sh)
{
  SH2tsti(sh,(4<<4)+4);
}
#define init_12_8_4_5
void FASTCALL call_12_8_4_5(SH2_struct* sh)
{
  SH2tsti(sh,(4<<4)+5);
}
#define init_12_8_4_6
void FASTCALL call_12_8_4_6(SH2_struct* sh)
{
  SH2tsti(sh,(4<<4)+6);
}
#define init_12_8_4_7
void FASTCALL call_12_8_4_7(SH2_struct* sh)
{
  SH2tsti(sh,(4<<4)+7);
}
#define init_12_8_4_8
void FASTCALL call_12_8_4_8(SH2_struct* sh)
{
  SH2tsti(sh,(4<<4)+8);
}
#define init_12_8_4_9
void FASTCALL call_12_8_4_9(SH2_struct* sh)
{
  SH2tsti(sh,(4<<4)+9);
}
#define init_12_8_4_10
void FASTCALL call_12_8_4_10(SH2_struct* sh)
{
  SH2tsti(sh,(4<<4)+10);
}
#define init_12_8_4_11
void FASTCALL call_12_8_4_11(SH2_struct* sh)
{
  SH2tsti(sh,(4<<4)+11);
}
#define init_12_8_4_12
void FASTCALL call_12_8_4_12(SH2_struct* sh)
{
  SH2tsti(sh,(4<<4)+12);
}
#define init_12_8_4_13
void FASTCALL call_12_8_4_13(SH2_struct* sh)
{
  SH2tsti(sh,(4<<4)+13);
}
#define init_12_8_4_14
void FASTCALL call_12_8_4_14(SH2_struct* sh)
{
  SH2tsti(sh,(4<<4)+14);
}
#define init_12_8_4_15
void FASTCALL call_12_8_4_15(SH2_struct* sh)
{
  SH2tsti(sh,(4<<4)+15);
}
#define init_12_8_5_0
void FASTCALL call_12_8_5_0(SH2_struct* sh)
{
  SH2tsti(sh,(5<<4)+0);
}
#define init_12_8_5_1
void FASTCALL call_12_8_5_1(SH2_struct* sh)
{
  SH2tsti(sh,(5<<4)+1);
}
#define init_12_8_5_2
void FASTCALL call_12_8_5_2(SH2_struct* sh)
{
  SH2tsti(sh,(5<<4)+2);
}
#define init_12_8_5_3
void FASTCALL call_12_8_5_3(SH2_struct* sh)
{
  SH2tsti(sh,(5<<4)+3);
}
#define init_12_8_5_4
void FASTCALL call_12_8_5_4(SH2_struct* sh)
{
  SH2tsti(sh,(5<<4)+4);
}
#define init_12_8_5_5
void FASTCALL call_12_8_5_5(SH2_struct* sh)
{
  SH2tsti(sh,(5<<4)+5);
}
#define init_12_8_5_6
void FASTCALL call_12_8_5_6(SH2_struct* sh)
{
  SH2tsti(sh,(5<<4)+6);
}
#define init_12_8_5_7
void FASTCALL call_12_8_5_7(SH2_struct* sh)
{
  SH2tsti(sh,(5<<4)+7);
}
#define init_12_8_5_8
void FASTCALL call_12_8_5_8(SH2_struct* sh)
{
  SH2tsti(sh,(5<<4)+8);
}
#define init_12_8_5_9
void FASTCALL call_12_8_5_9(SH2_struct* sh)
{
  SH2tsti(sh,(5<<4)+9);
}
#define init_12_8_5_10
void FASTCALL call_12_8_5_10(SH2_struct* sh)
{
  SH2tsti(sh,(5<<4)+10);
}
#define init_12_8_5_11
void FASTCALL call_12_8_5_11(SH2_struct* sh)
{
  SH2tsti(sh,(5<<4)+11);
}
#define init_12_8_5_12
void FASTCALL call_12_8_5_12(SH2_struct* sh)
{
  SH2tsti(sh,(5<<4)+12);
}
#define init_12_8_5_13
void FASTCALL call_12_8_5_13(SH2_struct* sh)
{
  SH2tsti(sh,(5<<4)+13);
}
#define init_12_8_5_14
void FASTCALL call_12_8_5_14(SH2_struct* sh)
{
  SH2tsti(sh,(5<<4)+14);
}
#define init_12_8_5_15
void FASTCALL call_12_8_5_15(SH2_struct* sh)
{
  SH2tsti(sh,(5<<4)+15);
}
#define init_12_8_6_0
void FASTCALL call_12_8_6_0(SH2_struct* sh)
{
  SH2tsti(sh,(6<<4)+0);
}
#define init_12_8_6_1
void FASTCALL call_12_8_6_1(SH2_struct* sh)
{
  SH2tsti(sh,(6<<4)+1);
}
#define init_12_8_6_2
void FASTCALL call_12_8_6_2(SH2_struct* sh)
{
  SH2tsti(sh,(6<<4)+2);
}
#define init_12_8_6_3
void FASTCALL call_12_8_6_3(SH2_struct* sh)
{
  SH2tsti(sh,(6<<4)+3);
}
#define init_12_8_6_4
void FASTCALL call_12_8_6_4(SH2_struct* sh)
{
  SH2tsti(sh,(6<<4)+4);
}
#define init_12_8_6_5
void FASTCALL call_12_8_6_5(SH2_struct* sh)
{
  SH2tsti(sh,(6<<4)+5);
}
#define init_12_8_6_6
void FASTCALL call_12_8_6_6(SH2_struct* sh)
{
  SH2tsti(sh,(6<<4)+6);
}
#define init_12_8_6_7
void FASTCALL call_12_8_6_7(SH2_struct* sh)
{
  SH2tsti(sh,(6<<4)+7);
}
#define init_12_8_6_8
void FASTCALL call_12_8_6_8(SH2_struct* sh)
{
  SH2tsti(sh,(6<<4)+8);
}
#define init_12_8_6_9
void FASTCALL call_12_8_6_9(SH2_struct* sh)
{
  SH2tsti(sh,(6<<4)+9);
}
#define init_12_8_6_10
void FASTCALL call_12_8_6_10(SH2_struct* sh)
{
  SH2tsti(sh,(6<<4)+10);
}
#define init_12_8_6_11
void FASTCALL call_12_8_6_11(SH2_struct* sh)
{
  SH2tsti(sh,(6<<4)+11);
}
#define init_12_8_6_12
void FASTCALL call_12_8_6_12(SH2_struct* sh)
{
  SH2tsti(sh,(6<<4)+12);
}
#define init_12_8_6_13
void FASTCALL call_12_8_6_13(SH2_struct* sh)
{
  SH2tsti(sh,(6<<4)+13);
}
#define init_12_8_6_14
void FASTCALL call_12_8_6_14(SH2_struct* sh)
{
  SH2tsti(sh,(6<<4)+14);
}
#define init_12_8_6_15
void FASTCALL call_12_8_6_15(SH2_struct* sh)
{
  SH2tsti(sh,(6<<4)+15);
}
#define init_12_8_7_0
void FASTCALL call_12_8_7_0(SH2_struct* sh)
{
  SH2tsti(sh,(7<<4)+0);
}
#define init_12_8_7_1
void FASTCALL call_12_8_7_1(SH2_struct* sh)
{
  SH2tsti(sh,(7<<4)+1);
}
#define init_12_8_7_2
void FASTCALL call_12_8_7_2(SH2_struct* sh)
{
  SH2tsti(sh,(7<<4)+2);
}
#define init_12_8_7_3
void FASTCALL call_12_8_7_3(SH2_struct* sh)
{
  SH2tsti(sh,(7<<4)+3);
}
#define init_12_8_7_4
void FASTCALL call_12_8_7_4(SH2_struct* sh)
{
  SH2tsti(sh,(7<<4)+4);
}
#define init_12_8_7_5
void FASTCALL call_12_8_7_5(SH2_struct* sh)
{
  SH2tsti(sh,(7<<4)+5);
}
#define init_12_8_7_6
void FASTCALL call_12_8_7_6(SH2_struct* sh)
{
  SH2tsti(sh,(7<<4)+6);
}
#define init_12_8_7_7
void FASTCALL call_12_8_7_7(SH2_struct* sh)
{
  SH2tsti(sh,(7<<4)+7);
}
#define init_12_8_7_8
void FASTCALL call_12_8_7_8(SH2_struct* sh)
{
  SH2tsti(sh,(7<<4)+8);
}
#define init_12_8_7_9
void FASTCALL call_12_8_7_9(SH2_struct* sh)
{
  SH2tsti(sh,(7<<4)+9);
}
#define init_12_8_7_10
void FASTCALL call_12_8_7_10(SH2_struct* sh)
{
  SH2tsti(sh,(7<<4)+10);
}
#define init_12_8_7_11
void FASTCALL call_12_8_7_11(SH2_struct* sh)
{
  SH2tsti(sh,(7<<4)+11);
}
#define init_12_8_7_12
void FASTCALL call_12_8_7_12(SH2_struct* sh)
{
  SH2tsti(sh,(7<<4)+12);
}
#define init_12_8_7_13
void FASTCALL call_12_8_7_13(SH2_struct* sh)
{
  SH2tsti(sh,(7<<4)+13);
}
#define init_12_8_7_14
void FASTCALL call_12_8_7_14(SH2_struct* sh)
{
  SH2tsti(sh,(7<<4)+14);
}
#define init_12_8_7_15
void FASTCALL call_12_8_7_15(SH2_struct* sh)
{
  SH2tsti(sh,(7<<4)+15);
}
#define init_12_8_8_0
void FASTCALL call_12_8_8_0(SH2_struct* sh)
{
  SH2tsti(sh,(8<<4)+0);
}
#define init_12_8_8_1
void FASTCALL call_12_8_8_1(SH2_struct* sh)
{
  SH2tsti(sh,(8<<4)+1);
}
#define init_12_8_8_2
void FASTCALL call_12_8_8_2(SH2_struct* sh)
{
  SH2tsti(sh,(8<<4)+2);
}
#define init_12_8_8_3
void FASTCALL call_12_8_8_3(SH2_struct* sh)
{
  SH2tsti(sh,(8<<4)+3);
}
#define init_12_8_8_4
void FASTCALL call_12_8_8_4(SH2_struct* sh)
{
  SH2tsti(sh,(8<<4)+4);
}
#define init_12_8_8_5
void FASTCALL call_12_8_8_5(SH2_struct* sh)
{
  SH2tsti(sh,(8<<4)+5);
}
#define init_12_8_8_6
void FASTCALL call_12_8_8_6(SH2_struct* sh)
{
  SH2tsti(sh,(8<<4)+6);
}
#define init_12_8_8_7
void FASTCALL call_12_8_8_7(SH2_struct* sh)
{
  SH2tsti(sh,(8<<4)+7);
}
#define init_12_8_8_8
void FASTCALL call_12_8_8_8(SH2_struct* sh)
{
  SH2tsti(sh,(8<<4)+8);
}
#define init_12_8_8_9
void FASTCALL call_12_8_8_9(SH2_struct* sh)
{
  SH2tsti(sh,(8<<4)+9);
}
#define init_12_8_8_10
void FASTCALL call_12_8_8_10(SH2_struct* sh)
{
  SH2tsti(sh,(8<<4)+10);
}
#define init_12_8_8_11
void FASTCALL call_12_8_8_11(SH2_struct* sh)
{
  SH2tsti(sh,(8<<4)+11);
}
#define init_12_8_8_12
void FASTCALL call_12_8_8_12(SH2_struct* sh)
{
  SH2tsti(sh,(8<<4)+12);
}
#define init_12_8_8_13
void FASTCALL call_12_8_8_13(SH2_struct* sh)
{
  SH2tsti(sh,(8<<4)+13);
}
#define init_12_8_8_14
void FASTCALL call_12_8_8_14(SH2_struct* sh)
{
  SH2tsti(sh,(8<<4)+14);
}
#define init_12_8_8_15
void FASTCALL call_12_8_8_15(SH2_struct* sh)
{
  SH2tsti(sh,(8<<4)+15);
}
#define init_12_8_9_0
void FASTCALL call_12_8_9_0(SH2_struct* sh)
{
  SH2tsti(sh,(9<<4)+0);
}
#define init_12_8_9_1
void FASTCALL call_12_8_9_1(SH2_struct* sh)
{
  SH2tsti(sh,(9<<4)+1);
}
#define init_12_8_9_2
void FASTCALL call_12_8_9_2(SH2_struct* sh)
{
  SH2tsti(sh,(9<<4)+2);
}
#define init_12_8_9_3
void FASTCALL call_12_8_9_3(SH2_struct* sh)
{
  SH2tsti(sh,(9<<4)+3);
}
#define init_12_8_9_4
void FASTCALL call_12_8_9_4(SH2_struct* sh)
{
  SH2tsti(sh,(9<<4)+4);
}
#define init_12_8_9_5
void FASTCALL call_12_8_9_5(SH2_struct* sh)
{
  SH2tsti(sh,(9<<4)+5);
}
#define init_12_8_9_6
void FASTCALL call_12_8_9_6(SH2_struct* sh)
{
  SH2tsti(sh,(9<<4)+6);
}
#define init_12_8_9_7
void FASTCALL call_12_8_9_7(SH2_struct* sh)
{
  SH2tsti(sh,(9<<4)+7);
}
#define init_12_8_9_8
void FASTCALL call_12_8_9_8(SH2_struct* sh)
{
  SH2tsti(sh,(9<<4)+8);
}
#define init_12_8_9_9
void FASTCALL call_12_8_9_9(SH2_struct* sh)
{
  SH2tsti(sh,(9<<4)+9);
}
#define init_12_8_9_10
void FASTCALL call_12_8_9_10(SH2_struct* sh)
{
  SH2tsti(sh,(9<<4)+10);
}
#define init_12_8_9_11
void FASTCALL call_12_8_9_11(SH2_struct* sh)
{
  SH2tsti(sh,(9<<4)+11);
}
#define init_12_8_9_12
void FASTCALL call_12_8_9_12(SH2_struct* sh)
{
  SH2tsti(sh,(9<<4)+12);
}
#define init_12_8_9_13
void FASTCALL call_12_8_9_13(SH2_struct* sh)
{
  SH2tsti(sh,(9<<4)+13);
}
#define init_12_8_9_14
void FASTCALL call_12_8_9_14(SH2_struct* sh)
{
  SH2tsti(sh,(9<<4)+14);
}
#define init_12_8_9_15
void FASTCALL call_12_8_9_15(SH2_struct* sh)
{
  SH2tsti(sh,(9<<4)+15);
}
#define init_12_8_10_0
void FASTCALL call_12_8_10_0(SH2_struct* sh)
{
  SH2tsti(sh,(10<<4)+0);
}
#define init_12_8_10_1
void FASTCALL call_12_8_10_1(SH2_struct* sh)
{
  SH2tsti(sh,(10<<4)+1);
}
#define init_12_8_10_2
void FASTCALL call_12_8_10_2(SH2_struct* sh)
{
  SH2tsti(sh,(10<<4)+2);
}
#define init_12_8_10_3
void FASTCALL call_12_8_10_3(SH2_struct* sh)
{
  SH2tsti(sh,(10<<4)+3);
}
#define init_12_8_10_4
void FASTCALL call_12_8_10_4(SH2_struct* sh)
{
  SH2tsti(sh,(10<<4)+4);
}
#define init_12_8_10_5
void FASTCALL call_12_8_10_5(SH2_struct* sh)
{
  SH2tsti(sh,(10<<4)+5);
}
#define init_12_8_10_6
void FASTCALL call_12_8_10_6(SH2_struct* sh)
{
  SH2tsti(sh,(10<<4)+6);
}
#define init_12_8_10_7
void FASTCALL call_12_8_10_7(SH2_struct* sh)
{
  SH2tsti(sh,(10<<4)+7);
}
#define init_12_8_10_8
void FASTCALL call_12_8_10_8(SH2_struct* sh)
{
  SH2tsti(sh,(10<<4)+8);
}
#define init_12_8_10_9
void FASTCALL call_12_8_10_9(SH2_struct* sh)
{
  SH2tsti(sh,(10<<4)+9);
}
#define init_12_8_10_10
void FASTCALL call_12_8_10_10(SH2_struct* sh)
{
  SH2tsti(sh,(10<<4)+10);
}
#define init_12_8_10_11
void FASTCALL call_12_8_10_11(SH2_struct* sh)
{
  SH2tsti(sh,(10<<4)+11);
}
#define init_12_8_10_12
void FASTCALL call_12_8_10_12(SH2_struct* sh)
{
  SH2tsti(sh,(10<<4)+12);
}
#define init_12_8_10_13
void FASTCALL call_12_8_10_13(SH2_struct* sh)
{
  SH2tsti(sh,(10<<4)+13);
}
#define init_12_8_10_14
void FASTCALL call_12_8_10_14(SH2_struct* sh)
{
  SH2tsti(sh,(10<<4)+14);
}
#define init_12_8_10_15
void FASTCALL call_12_8_10_15(SH2_struct* sh)
{
  SH2tsti(sh,(10<<4)+15);
}
#define init_12_8_11_0
void FASTCALL call_12_8_11_0(SH2_struct* sh)
{
  SH2tsti(sh,(11<<4)+0);
}
#define init_12_8_11_1
void FASTCALL call_12_8_11_1(SH2_struct* sh)
{
  SH2tsti(sh,(11<<4)+1);
}
#define init_12_8_11_2
void FASTCALL call_12_8_11_2(SH2_struct* sh)
{
  SH2tsti(sh,(11<<4)+2);
}
#define init_12_8_11_3
void FASTCALL call_12_8_11_3(SH2_struct* sh)
{
  SH2tsti(sh,(11<<4)+3);
}
#define init_12_8_11_4
void FASTCALL call_12_8_11_4(SH2_struct* sh)
{
  SH2tsti(sh,(11<<4)+4);
}
#define init_12_8_11_5
void FASTCALL call_12_8_11_5(SH2_struct* sh)
{
  SH2tsti(sh,(11<<4)+5);
}
#define init_12_8_11_6
void FASTCALL call_12_8_11_6(SH2_struct* sh)
{
  SH2tsti(sh,(11<<4)+6);
}
#define init_12_8_11_7
void FASTCALL call_12_8_11_7(SH2_struct* sh)
{
  SH2tsti(sh,(11<<4)+7);
}
#define init_12_8_11_8
void FASTCALL call_12_8_11_8(SH2_struct* sh)
{
  SH2tsti(sh,(11<<4)+8);
}
#define init_12_8_11_9
void FASTCALL call_12_8_11_9(SH2_struct* sh)
{
  SH2tsti(sh,(11<<4)+9);
}
#define init_12_8_11_10
void FASTCALL call_12_8_11_10(SH2_struct* sh)
{
  SH2tsti(sh,(11<<4)+10);
}
#define init_12_8_11_11
void FASTCALL call_12_8_11_11(SH2_struct* sh)
{
  SH2tsti(sh,(11<<4)+11);
}
#define init_12_8_11_12
void FASTCALL call_12_8_11_12(SH2_struct* sh)
{
  SH2tsti(sh,(11<<4)+12);
}
#define init_12_8_11_13
void FASTCALL call_12_8_11_13(SH2_struct* sh)
{
  SH2tsti(sh,(11<<4)+13);
}
#define init_12_8_11_14
void FASTCALL call_12_8_11_14(SH2_struct* sh)
{
  SH2tsti(sh,(11<<4)+14);
}
#define init_12_8_11_15
void FASTCALL call_12_8_11_15(SH2_struct* sh)
{
  SH2tsti(sh,(11<<4)+15);
}
#define init_12_8_12_0
void FASTCALL call_12_8_12_0(SH2_struct* sh)
{
  SH2tsti(sh,(12<<4)+0);
}
#define init_12_8_12_1
void FASTCALL call_12_8_12_1(SH2_struct* sh)
{
  SH2tsti(sh,(12<<4)+1);
}
#define init_12_8_12_2
void FASTCALL call_12_8_12_2(SH2_struct* sh)
{
  SH2tsti(sh,(12<<4)+2);
}
#define init_12_8_12_3
void FASTCALL call_12_8_12_3(SH2_struct* sh)
{
  SH2tsti(sh,(12<<4)+3);
}
#define init_12_8_12_4
void FASTCALL call_12_8_12_4(SH2_struct* sh)
{
  SH2tsti(sh,(12<<4)+4);
}
#define init_12_8_12_5
void FASTCALL call_12_8_12_5(SH2_struct* sh)
{
  SH2tsti(sh,(12<<4)+5);
}
#define init_12_8_12_6
void FASTCALL call_12_8_12_6(SH2_struct* sh)
{
  SH2tsti(sh,(12<<4)+6);
}
#define init_12_8_12_7
void FASTCALL call_12_8_12_7(SH2_struct* sh)
{
  SH2tsti(sh,(12<<4)+7);
}
#define init_12_8_12_8
void FASTCALL call_12_8_12_8(SH2_struct* sh)
{
  SH2tsti(sh,(12<<4)+8);
}
#define init_12_8_12_9
void FASTCALL call_12_8_12_9(SH2_struct* sh)
{
  SH2tsti(sh,(12<<4)+9);
}
#define init_12_8_12_10
void FASTCALL call_12_8_12_10(SH2_struct* sh)
{
  SH2tsti(sh,(12<<4)+10);
}
#define init_12_8_12_11
void FASTCALL call_12_8_12_11(SH2_struct* sh)
{
  SH2tsti(sh,(12<<4)+11);
}
#define init_12_8_12_12
void FASTCALL call_12_8_12_12(SH2_struct* sh)
{
  SH2tsti(sh,(12<<4)+12);
}
#define init_12_8_12_13
void FASTCALL call_12_8_12_13(SH2_struct* sh)
{
  SH2tsti(sh,(12<<4)+13);
}
#define init_12_8_12_14
void FASTCALL call_12_8_12_14(SH2_struct* sh)
{
  SH2tsti(sh,(12<<4)+14);
}
#define init_12_8_12_15
void FASTCALL call_12_8_12_15(SH2_struct* sh)
{
  SH2tsti(sh,(12<<4)+15);
}
#define init_12_8_13_0
void FASTCALL call_12_8_13_0(SH2_struct* sh)
{
  SH2tsti(sh,(13<<4)+0);
}
#define init_12_8_13_1
void FASTCALL call_12_8_13_1(SH2_struct* sh)
{
  SH2tsti(sh,(13<<4)+1);
}
#define init_12_8_13_2
void FASTCALL call_12_8_13_2(SH2_struct* sh)
{
  SH2tsti(sh,(13<<4)+2);
}
#define init_12_8_13_3
void FASTCALL call_12_8_13_3(SH2_struct* sh)
{
  SH2tsti(sh,(13<<4)+3);
}
#define init_12_8_13_4
void FASTCALL call_12_8_13_4(SH2_struct* sh)
{
  SH2tsti(sh,(13<<4)+4);
}
#define init_12_8_13_5
void FASTCALL call_12_8_13_5(SH2_struct* sh)
{
  SH2tsti(sh,(13<<4)+5);
}
#define init_12_8_13_6
void FASTCALL call_12_8_13_6(SH2_struct* sh)
{
  SH2tsti(sh,(13<<4)+6);
}
#define init_12_8_13_7
void FASTCALL call_12_8_13_7(SH2_struct* sh)
{
  SH2tsti(sh,(13<<4)+7);
}
#define init_12_8_13_8
void FASTCALL call_12_8_13_8(SH2_struct* sh)
{
  SH2tsti(sh,(13<<4)+8);
}
#define init_12_8_13_9
void FASTCALL call_12_8_13_9(SH2_struct* sh)
{
  SH2tsti(sh,(13<<4)+9);
}
#define init_12_8_13_10
void FASTCALL call_12_8_13_10(SH2_struct* sh)
{
  SH2tsti(sh,(13<<4)+10);
}
#define init_12_8_13_11
void FASTCALL call_12_8_13_11(SH2_struct* sh)
{
  SH2tsti(sh,(13<<4)+11);
}
#define init_12_8_13_12
void FASTCALL call_12_8_13_12(SH2_struct* sh)
{
  SH2tsti(sh,(13<<4)+12);
}
#define init_12_8_13_13
void FASTCALL call_12_8_13_13(SH2_struct* sh)
{
  SH2tsti(sh,(13<<4)+13);
}
#define init_12_8_13_14
void FASTCALL call_12_8_13_14(SH2_struct* sh)
{
  SH2tsti(sh,(13<<4)+14);
}
#define init_12_8_13_15
void FASTCALL call_12_8_13_15(SH2_struct* sh)
{
  SH2tsti(sh,(13<<4)+15);
}
#define init_12_8_14_0
void FASTCALL call_12_8_14_0(SH2_struct* sh)
{
  SH2tsti(sh,(14<<4)+0);
}
#define init_12_8_14_1
void FASTCALL call_12_8_14_1(SH2_struct* sh)
{
  SH2tsti(sh,(14<<4)+1);
}
#define init_12_8_14_2
void FASTCALL call_12_8_14_2(SH2_struct* sh)
{
  SH2tsti(sh,(14<<4)+2);
}
#define init_12_8_14_3
void FASTCALL call_12_8_14_3(SH2_struct* sh)
{
  SH2tsti(sh,(14<<4)+3);
}
#define init_12_8_14_4
void FASTCALL call_12_8_14_4(SH2_struct* sh)
{
  SH2tsti(sh,(14<<4)+4);
}
#define init_12_8_14_5
void FASTCALL call_12_8_14_5(SH2_struct* sh)
{
  SH2tsti(sh,(14<<4)+5);
}
#define init_12_8_14_6
void FASTCALL call_12_8_14_6(SH2_struct* sh)
{
  SH2tsti(sh,(14<<4)+6);
}
#define init_12_8_14_7
void FASTCALL call_12_8_14_7(SH2_struct* sh)
{
  SH2tsti(sh,(14<<4)+7);
}
#define init_12_8_14_8
void FASTCALL call_12_8_14_8(SH2_struct* sh)
{
  SH2tsti(sh,(14<<4)+8);
}
#define init_12_8_14_9
void FASTCALL call_12_8_14_9(SH2_struct* sh)
{
  SH2tsti(sh,(14<<4)+9);
}
#define init_12_8_14_10
void FASTCALL call_12_8_14_10(SH2_struct* sh)
{
  SH2tsti(sh,(14<<4)+10);
}
#define init_12_8_14_11
void FASTCALL call_12_8_14_11(SH2_struct* sh)
{
  SH2tsti(sh,(14<<4)+11);
}
#define init_12_8_14_12
void FASTCALL call_12_8_14_12(SH2_struct* sh)
{
  SH2tsti(sh,(14<<4)+12);
}
#define init_12_8_14_13
void FASTCALL call_12_8_14_13(SH2_struct* sh)
{
  SH2tsti(sh,(14<<4)+13);
}
#define init_12_8_14_14
void FASTCALL call_12_8_14_14(SH2_struct* sh)
{
  SH2tsti(sh,(14<<4)+14);
}
#define init_12_8_14_15
void FASTCALL call_12_8_14_15(SH2_struct* sh)
{
  SH2tsti(sh,(14<<4)+15);
}
#define init_12_8_15_0
void FASTCALL call_12_8_15_0(SH2_struct* sh)
{
  SH2tsti(sh,(15<<4)+0);
}
#define init_12_8_15_1
void FASTCALL call_12_8_15_1(SH2_struct* sh)
{
  SH2tsti(sh,(15<<4)+1);
}
#define init_12_8_15_2
void FASTCALL call_12_8_15_2(SH2_struct* sh)
{
  SH2tsti(sh,(15<<4)+2);
}
#define init_12_8_15_3
void FASTCALL call_12_8_15_3(SH2_struct* sh)
{
  SH2tsti(sh,(15<<4)+3);
}
#define init_12_8_15_4
void FASTCALL call_12_8_15_4(SH2_struct* sh)
{
  SH2tsti(sh,(15<<4)+4);
}
#define init_12_8_15_5
void FASTCALL call_12_8_15_5(SH2_struct* sh)
{
  SH2tsti(sh,(15<<4)+5);
}
#define init_12_8_15_6
void FASTCALL call_12_8_15_6(SH2_struct* sh)
{
  SH2tsti(sh,(15<<4)+6);
}
#define init_12_8_15_7
void FASTCALL call_12_8_15_7(SH2_struct* sh)
{
  SH2tsti(sh,(15<<4)+7);
}
#define init_12_8_15_8
void FASTCALL call_12_8_15_8(SH2_struct* sh)
{
  SH2tsti(sh,(15<<4)+8);
}
#define init_12_8_15_9
void FASTCALL call_12_8_15_9(SH2_struct* sh)
{
  SH2tsti(sh,(15<<4)+9);
}
#define init_12_8_15_10
void FASTCALL call_12_8_15_10(SH2_struct* sh)
{
  SH2tsti(sh,(15<<4)+10);
}
#define init_12_8_15_11
void FASTCALL call_12_8_15_11(SH2_struct* sh)
{
  SH2tsti(sh,(15<<4)+11);
}
#define init_12_8_15_12
void FASTCALL call_12_8_15_12(SH2_struct* sh)
{
  SH2tsti(sh,(15<<4)+12);
}
#define init_12_8_15_13
void FASTCALL call_12_8_15_13(SH2_struct* sh)
{
  SH2tsti(sh,(15<<4)+13);
}
#define init_12_8_15_14
void FASTCALL call_12_8_15_14(SH2_struct* sh)
{
  SH2tsti(sh,(15<<4)+14);
}
#define init_12_8_15_15
void FASTCALL call_12_8_15_15(SH2_struct* sh)
{
  SH2tsti(sh,(15<<4)+15);
}
#define init_12_3_0_0
void FASTCALL call_12_3_0_0(SH2_struct* sh)
{
  SH2trapa(sh,(0<<4)+0);
}
#define init_12_3_0_1
void FASTCALL call_12_3_0_1(SH2_struct* sh)
{
  SH2trapa(sh,(0<<4)+1);
}
#define init_12_3_0_2
void FASTCALL call_12_3_0_2(SH2_struct* sh)
{
  SH2trapa(sh,(0<<4)+2);
}
#define init_12_3_0_3
void FASTCALL call_12_3_0_3(SH2_struct* sh)
{
  SH2trapa(sh,(0<<4)+3);
}
#define init_12_3_0_4
void FASTCALL call_12_3_0_4(SH2_struct* sh)
{
  SH2trapa(sh,(0<<4)+4);
}
#define init_12_3_0_5
void FASTCALL call_12_3_0_5(SH2_struct* sh)
{
  SH2trapa(sh,(0<<4)+5);
}
#define init_12_3_0_6
void FASTCALL call_12_3_0_6(SH2_struct* sh)
{
  SH2trapa(sh,(0<<4)+6);
}
#define init_12_3_0_7
void FASTCALL call_12_3_0_7(SH2_struct* sh)
{
  SH2trapa(sh,(0<<4)+7);
}
#define init_12_3_0_8
void FASTCALL call_12_3_0_8(SH2_struct* sh)
{
  SH2trapa(sh,(0<<4)+8);
}
#define init_12_3_0_9
void FASTCALL call_12_3_0_9(SH2_struct* sh)
{
  SH2trapa(sh,(0<<4)+9);
}
#define init_12_3_0_10
void FASTCALL call_12_3_0_10(SH2_struct* sh)
{
  SH2trapa(sh,(0<<4)+10);
}
#define init_12_3_0_11
void FASTCALL call_12_3_0_11(SH2_struct* sh)
{
  SH2trapa(sh,(0<<4)+11);
}
#define init_12_3_0_12
void FASTCALL call_12_3_0_12(SH2_struct* sh)
{
  SH2trapa(sh,(0<<4)+12);
}
#define init_12_3_0_13
void FASTCALL call_12_3_0_13(SH2_struct* sh)
{
  SH2trapa(sh,(0<<4)+13);
}
#define init_12_3_0_14
void FASTCALL call_12_3_0_14(SH2_struct* sh)
{
  SH2trapa(sh,(0<<4)+14);
}
#define init_12_3_0_15
void FASTCALL call_12_3_0_15(SH2_struct* sh)
{
  SH2trapa(sh,(0<<4)+15);
}
#define init_12_3_1_0
void FASTCALL call_12_3_1_0(SH2_struct* sh)
{
  SH2trapa(sh,(1<<4)+0);
}
#define init_12_3_1_1
void FASTCALL call_12_3_1_1(SH2_struct* sh)
{
  SH2trapa(sh,(1<<4)+1);
}
#define init_12_3_1_2
void FASTCALL call_12_3_1_2(SH2_struct* sh)
{
  SH2trapa(sh,(1<<4)+2);
}
#define init_12_3_1_3
void FASTCALL call_12_3_1_3(SH2_struct* sh)
{
  SH2trapa(sh,(1<<4)+3);
}
#define init_12_3_1_4
void FASTCALL call_12_3_1_4(SH2_struct* sh)
{
  SH2trapa(sh,(1<<4)+4);
}
#define init_12_3_1_5
void FASTCALL call_12_3_1_5(SH2_struct* sh)
{
  SH2trapa(sh,(1<<4)+5);
}
#define init_12_3_1_6
void FASTCALL call_12_3_1_6(SH2_struct* sh)
{
  SH2trapa(sh,(1<<4)+6);
}
#define init_12_3_1_7
void FASTCALL call_12_3_1_7(SH2_struct* sh)
{
  SH2trapa(sh,(1<<4)+7);
}
#define init_12_3_1_8
void FASTCALL call_12_3_1_8(SH2_struct* sh)
{
  SH2trapa(sh,(1<<4)+8);
}
#define init_12_3_1_9
void FASTCALL call_12_3_1_9(SH2_struct* sh)
{
  SH2trapa(sh,(1<<4)+9);
}
#define init_12_3_1_10
void FASTCALL call_12_3_1_10(SH2_struct* sh)
{
  SH2trapa(sh,(1<<4)+10);
}
#define init_12_3_1_11
void FASTCALL call_12_3_1_11(SH2_struct* sh)
{
  SH2trapa(sh,(1<<4)+11);
}
#define init_12_3_1_12
void FASTCALL call_12_3_1_12(SH2_struct* sh)
{
  SH2trapa(sh,(1<<4)+12);
}
#define init_12_3_1_13
void FASTCALL call_12_3_1_13(SH2_struct* sh)
{
  SH2trapa(sh,(1<<4)+13);
}
#define init_12_3_1_14
void FASTCALL call_12_3_1_14(SH2_struct* sh)
{
  SH2trapa(sh,(1<<4)+14);
}
#define init_12_3_1_15
void FASTCALL call_12_3_1_15(SH2_struct* sh)
{
  SH2trapa(sh,(1<<4)+15);
}
#define init_12_3_2_0
void FASTCALL call_12_3_2_0(SH2_struct* sh)
{
  SH2trapa(sh,(2<<4)+0);
}
#define init_12_3_2_1
void FASTCALL call_12_3_2_1(SH2_struct* sh)
{
  SH2trapa(sh,(2<<4)+1);
}
#define init_12_3_2_2
void FASTCALL call_12_3_2_2(SH2_struct* sh)
{
  SH2trapa(sh,(2<<4)+2);
}
#define init_12_3_2_3
void FASTCALL call_12_3_2_3(SH2_struct* sh)
{
  SH2trapa(sh,(2<<4)+3);
}
#define init_12_3_2_4
void FASTCALL call_12_3_2_4(SH2_struct* sh)
{
  SH2trapa(sh,(2<<4)+4);
}
#define init_12_3_2_5
void FASTCALL call_12_3_2_5(SH2_struct* sh)
{
  SH2trapa(sh,(2<<4)+5);
}
#define init_12_3_2_6
void FASTCALL call_12_3_2_6(SH2_struct* sh)
{
  SH2trapa(sh,(2<<4)+6);
}
#define init_12_3_2_7
void FASTCALL call_12_3_2_7(SH2_struct* sh)
{
  SH2trapa(sh,(2<<4)+7);
}
#define init_12_3_2_8
void FASTCALL call_12_3_2_8(SH2_struct* sh)
{
  SH2trapa(sh,(2<<4)+8);
}
#define init_12_3_2_9
void FASTCALL call_12_3_2_9(SH2_struct* sh)
{
  SH2trapa(sh,(2<<4)+9);
}
#define init_12_3_2_10
void FASTCALL call_12_3_2_10(SH2_struct* sh)
{
  SH2trapa(sh,(2<<4)+10);
}
#define init_12_3_2_11
void FASTCALL call_12_3_2_11(SH2_struct* sh)
{
  SH2trapa(sh,(2<<4)+11);
}
#define init_12_3_2_12
void FASTCALL call_12_3_2_12(SH2_struct* sh)
{
  SH2trapa(sh,(2<<4)+12);
}
#define init_12_3_2_13
void FASTCALL call_12_3_2_13(SH2_struct* sh)
{
  SH2trapa(sh,(2<<4)+13);
}
#define init_12_3_2_14
void FASTCALL call_12_3_2_14(SH2_struct* sh)
{
  SH2trapa(sh,(2<<4)+14);
}
#define init_12_3_2_15
void FASTCALL call_12_3_2_15(SH2_struct* sh)
{
  SH2trapa(sh,(2<<4)+15);
}
#define init_12_3_3_0
void FASTCALL call_12_3_3_0(SH2_struct* sh)
{
  SH2trapa(sh,(3<<4)+0);
}
#define init_12_3_3_1
void FASTCALL call_12_3_3_1(SH2_struct* sh)
{
  SH2trapa(sh,(3<<4)+1);
}
#define init_12_3_3_2
void FASTCALL call_12_3_3_2(SH2_struct* sh)
{
  SH2trapa(sh,(3<<4)+2);
}
#define init_12_3_3_3
void FASTCALL call_12_3_3_3(SH2_struct* sh)
{
  SH2trapa(sh,(3<<4)+3);
}
#define init_12_3_3_4
void FASTCALL call_12_3_3_4(SH2_struct* sh)
{
  SH2trapa(sh,(3<<4)+4);
}
#define init_12_3_3_5
void FASTCALL call_12_3_3_5(SH2_struct* sh)
{
  SH2trapa(sh,(3<<4)+5);
}
#define init_12_3_3_6
void FASTCALL call_12_3_3_6(SH2_struct* sh)
{
  SH2trapa(sh,(3<<4)+6);
}
#define init_12_3_3_7
void FASTCALL call_12_3_3_7(SH2_struct* sh)
{
  SH2trapa(sh,(3<<4)+7);
}
#define init_12_3_3_8
void FASTCALL call_12_3_3_8(SH2_struct* sh)
{
  SH2trapa(sh,(3<<4)+8);
}
#define init_12_3_3_9
void FASTCALL call_12_3_3_9(SH2_struct* sh)
{
  SH2trapa(sh,(3<<4)+9);
}
#define init_12_3_3_10
void FASTCALL call_12_3_3_10(SH2_struct* sh)
{
  SH2trapa(sh,(3<<4)+10);
}
#define init_12_3_3_11
void FASTCALL call_12_3_3_11(SH2_struct* sh)
{
  SH2trapa(sh,(3<<4)+11);
}
#define init_12_3_3_12
void FASTCALL call_12_3_3_12(SH2_struct* sh)
{
  SH2trapa(sh,(3<<4)+12);
}
#define init_12_3_3_13
void FASTCALL call_12_3_3_13(SH2_struct* sh)
{
  SH2trapa(sh,(3<<4)+13);
}
#define init_12_3_3_14
void FASTCALL call_12_3_3_14(SH2_struct* sh)
{
  SH2trapa(sh,(3<<4)+14);
}
#define init_12_3_3_15
void FASTCALL call_12_3_3_15(SH2_struct* sh)
{
  SH2trapa(sh,(3<<4)+15);
}
#define init_12_3_4_0
void FASTCALL call_12_3_4_0(SH2_struct* sh)
{
  SH2trapa(sh,(4<<4)+0);
}
#define init_12_3_4_1
void FASTCALL call_12_3_4_1(SH2_struct* sh)
{
  SH2trapa(sh,(4<<4)+1);
}
#define init_12_3_4_2
void FASTCALL call_12_3_4_2(SH2_struct* sh)
{
  SH2trapa(sh,(4<<4)+2);
}
#define init_12_3_4_3
void FASTCALL call_12_3_4_3(SH2_struct* sh)
{
  SH2trapa(sh,(4<<4)+3);
}
#define init_12_3_4_4
void FASTCALL call_12_3_4_4(SH2_struct* sh)
{
  SH2trapa(sh,(4<<4)+4);
}
#define init_12_3_4_5
void FASTCALL call_12_3_4_5(SH2_struct* sh)
{
  SH2trapa(sh,(4<<4)+5);
}
#define init_12_3_4_6
void FASTCALL call_12_3_4_6(SH2_struct* sh)
{
  SH2trapa(sh,(4<<4)+6);
}
#define init_12_3_4_7
void FASTCALL call_12_3_4_7(SH2_struct* sh)
{
  SH2trapa(sh,(4<<4)+7);
}
#define init_12_3_4_8
void FASTCALL call_12_3_4_8(SH2_struct* sh)
{
  SH2trapa(sh,(4<<4)+8);
}
#define init_12_3_4_9
void FASTCALL call_12_3_4_9(SH2_struct* sh)
{
  SH2trapa(sh,(4<<4)+9);
}
#define init_12_3_4_10
void FASTCALL call_12_3_4_10(SH2_struct* sh)
{
  SH2trapa(sh,(4<<4)+10);
}
#define init_12_3_4_11
void FASTCALL call_12_3_4_11(SH2_struct* sh)
{
  SH2trapa(sh,(4<<4)+11);
}
#define init_12_3_4_12
void FASTCALL call_12_3_4_12(SH2_struct* sh)
{
  SH2trapa(sh,(4<<4)+12);
}
#define init_12_3_4_13
void FASTCALL call_12_3_4_13(SH2_struct* sh)
{
  SH2trapa(sh,(4<<4)+13);
}
#define init_12_3_4_14
void FASTCALL call_12_3_4_14(SH2_struct* sh)
{
  SH2trapa(sh,(4<<4)+14);
}
#define init_12_3_4_15
void FASTCALL call_12_3_4_15(SH2_struct* sh)
{
  SH2trapa(sh,(4<<4)+15);
}
#define init_12_3_5_0
void FASTCALL call_12_3_5_0(SH2_struct* sh)
{
  SH2trapa(sh,(5<<4)+0);
}
#define init_12_3_5_1
void FASTCALL call_12_3_5_1(SH2_struct* sh)
{
  SH2trapa(sh,(5<<4)+1);
}
#define init_12_3_5_2
void FASTCALL call_12_3_5_2(SH2_struct* sh)
{
  SH2trapa(sh,(5<<4)+2);
}
#define init_12_3_5_3
void FASTCALL call_12_3_5_3(SH2_struct* sh)
{
  SH2trapa(sh,(5<<4)+3);
}
#define init_12_3_5_4
void FASTCALL call_12_3_5_4(SH2_struct* sh)
{
  SH2trapa(sh,(5<<4)+4);
}
#define init_12_3_5_5
void FASTCALL call_12_3_5_5(SH2_struct* sh)
{
  SH2trapa(sh,(5<<4)+5);
}
#define init_12_3_5_6
void FASTCALL call_12_3_5_6(SH2_struct* sh)
{
  SH2trapa(sh,(5<<4)+6);
}
#define init_12_3_5_7
void FASTCALL call_12_3_5_7(SH2_struct* sh)
{
  SH2trapa(sh,(5<<4)+7);
}
#define init_12_3_5_8
void FASTCALL call_12_3_5_8(SH2_struct* sh)
{
  SH2trapa(sh,(5<<4)+8);
}
#define init_12_3_5_9
void FASTCALL call_12_3_5_9(SH2_struct* sh)
{
  SH2trapa(sh,(5<<4)+9);
}
#define init_12_3_5_10
void FASTCALL call_12_3_5_10(SH2_struct* sh)
{
  SH2trapa(sh,(5<<4)+10);
}
#define init_12_3_5_11
void FASTCALL call_12_3_5_11(SH2_struct* sh)
{
  SH2trapa(sh,(5<<4)+11);
}
#define init_12_3_5_12
void FASTCALL call_12_3_5_12(SH2_struct* sh)
{
  SH2trapa(sh,(5<<4)+12);
}
#define init_12_3_5_13
void FASTCALL call_12_3_5_13(SH2_struct* sh)
{
  SH2trapa(sh,(5<<4)+13);
}
#define init_12_3_5_14
void FASTCALL call_12_3_5_14(SH2_struct* sh)
{
  SH2trapa(sh,(5<<4)+14);
}
#define init_12_3_5_15
void FASTCALL call_12_3_5_15(SH2_struct* sh)
{
  SH2trapa(sh,(5<<4)+15);
}
#define init_12_3_6_0
void FASTCALL call_12_3_6_0(SH2_struct* sh)
{
  SH2trapa(sh,(6<<4)+0);
}
#define init_12_3_6_1
void FASTCALL call_12_3_6_1(SH2_struct* sh)
{
  SH2trapa(sh,(6<<4)+1);
}
#define init_12_3_6_2
void FASTCALL call_12_3_6_2(SH2_struct* sh)
{
  SH2trapa(sh,(6<<4)+2);
}
#define init_12_3_6_3
void FASTCALL call_12_3_6_3(SH2_struct* sh)
{
  SH2trapa(sh,(6<<4)+3);
}
#define init_12_3_6_4
void FASTCALL call_12_3_6_4(SH2_struct* sh)
{
  SH2trapa(sh,(6<<4)+4);
}
#define init_12_3_6_5
void FASTCALL call_12_3_6_5(SH2_struct* sh)
{
  SH2trapa(sh,(6<<4)+5);
}
#define init_12_3_6_6
void FASTCALL call_12_3_6_6(SH2_struct* sh)
{
  SH2trapa(sh,(6<<4)+6);
}
#define init_12_3_6_7
void FASTCALL call_12_3_6_7(SH2_struct* sh)
{
  SH2trapa(sh,(6<<4)+7);
}
#define init_12_3_6_8
void FASTCALL call_12_3_6_8(SH2_struct* sh)
{
  SH2trapa(sh,(6<<4)+8);
}
#define init_12_3_6_9
void FASTCALL call_12_3_6_9(SH2_struct* sh)
{
  SH2trapa(sh,(6<<4)+9);
}
#define init_12_3_6_10
void FASTCALL call_12_3_6_10(SH2_struct* sh)
{
  SH2trapa(sh,(6<<4)+10);
}
#define init_12_3_6_11
void FASTCALL call_12_3_6_11(SH2_struct* sh)
{
  SH2trapa(sh,(6<<4)+11);
}
#define init_12_3_6_12
void FASTCALL call_12_3_6_12(SH2_struct* sh)
{
  SH2trapa(sh,(6<<4)+12);
}
#define init_12_3_6_13
void FASTCALL call_12_3_6_13(SH2_struct* sh)
{
  SH2trapa(sh,(6<<4)+13);
}
#define init_12_3_6_14
void FASTCALL call_12_3_6_14(SH2_struct* sh)
{
  SH2trapa(sh,(6<<4)+14);
}
#define init_12_3_6_15
void FASTCALL call_12_3_6_15(SH2_struct* sh)
{
  SH2trapa(sh,(6<<4)+15);
}
#define init_12_3_7_0
void FASTCALL call_12_3_7_0(SH2_struct* sh)
{
  SH2trapa(sh,(7<<4)+0);
}
#define init_12_3_7_1
void FASTCALL call_12_3_7_1(SH2_struct* sh)
{
  SH2trapa(sh,(7<<4)+1);
}
#define init_12_3_7_2
void FASTCALL call_12_3_7_2(SH2_struct* sh)
{
  SH2trapa(sh,(7<<4)+2);
}
#define init_12_3_7_3
void FASTCALL call_12_3_7_3(SH2_struct* sh)
{
  SH2trapa(sh,(7<<4)+3);
}
#define init_12_3_7_4
void FASTCALL call_12_3_7_4(SH2_struct* sh)
{
  SH2trapa(sh,(7<<4)+4);
}
#define init_12_3_7_5
void FASTCALL call_12_3_7_5(SH2_struct* sh)
{
  SH2trapa(sh,(7<<4)+5);
}
#define init_12_3_7_6
void FASTCALL call_12_3_7_6(SH2_struct* sh)
{
  SH2trapa(sh,(7<<4)+6);
}
#define init_12_3_7_7
void FASTCALL call_12_3_7_7(SH2_struct* sh)
{
  SH2trapa(sh,(7<<4)+7);
}
#define init_12_3_7_8
void FASTCALL call_12_3_7_8(SH2_struct* sh)
{
  SH2trapa(sh,(7<<4)+8);
}
#define init_12_3_7_9
void FASTCALL call_12_3_7_9(SH2_struct* sh)
{
  SH2trapa(sh,(7<<4)+9);
}
#define init_12_3_7_10
void FASTCALL call_12_3_7_10(SH2_struct* sh)
{
  SH2trapa(sh,(7<<4)+10);
}
#define init_12_3_7_11
void FASTCALL call_12_3_7_11(SH2_struct* sh)
{
  SH2trapa(sh,(7<<4)+11);
}
#define init_12_3_7_12
void FASTCALL call_12_3_7_12(SH2_struct* sh)
{
  SH2trapa(sh,(7<<4)+12);
}
#define init_12_3_7_13
void FASTCALL call_12_3_7_13(SH2_struct* sh)
{
  SH2trapa(sh,(7<<4)+13);
}
#define init_12_3_7_14
void FASTCALL call_12_3_7_14(SH2_struct* sh)
{
  SH2trapa(sh,(7<<4)+14);
}
#define init_12_3_7_15
void FASTCALL call_12_3_7_15(SH2_struct* sh)
{
  SH2trapa(sh,(7<<4)+15);
}
#define init_12_3_8_0
void FASTCALL call_12_3_8_0(SH2_struct* sh)
{
  SH2trapa(sh,(8<<4)+0);
}
#define init_12_3_8_1
void FASTCALL call_12_3_8_1(SH2_struct* sh)
{
  SH2trapa(sh,(8<<4)+1);
}
#define init_12_3_8_2
void FASTCALL call_12_3_8_2(SH2_struct* sh)
{
  SH2trapa(sh,(8<<4)+2);
}
#define init_12_3_8_3
void FASTCALL call_12_3_8_3(SH2_struct* sh)
{
  SH2trapa(sh,(8<<4)+3);
}
#define init_12_3_8_4
void FASTCALL call_12_3_8_4(SH2_struct* sh)
{
  SH2trapa(sh,(8<<4)+4);
}
#define init_12_3_8_5
void FASTCALL call_12_3_8_5(SH2_struct* sh)
{
  SH2trapa(sh,(8<<4)+5);
}
#define init_12_3_8_6
void FASTCALL call_12_3_8_6(SH2_struct* sh)
{
  SH2trapa(sh,(8<<4)+6);
}
#define init_12_3_8_7
void FASTCALL call_12_3_8_7(SH2_struct* sh)
{
  SH2trapa(sh,(8<<4)+7);
}
#define init_12_3_8_8
void FASTCALL call_12_3_8_8(SH2_struct* sh)
{
  SH2trapa(sh,(8<<4)+8);
}
#define init_12_3_8_9
void FASTCALL call_12_3_8_9(SH2_struct* sh)
{
  SH2trapa(sh,(8<<4)+9);
}
#define init_12_3_8_10
void FASTCALL call_12_3_8_10(SH2_struct* sh)
{
  SH2trapa(sh,(8<<4)+10);
}
#define init_12_3_8_11
void FASTCALL call_12_3_8_11(SH2_struct* sh)
{
  SH2trapa(sh,(8<<4)+11);
}
#define init_12_3_8_12
void FASTCALL call_12_3_8_12(SH2_struct* sh)
{
  SH2trapa(sh,(8<<4)+12);
}
#define init_12_3_8_13
void FASTCALL call_12_3_8_13(SH2_struct* sh)
{
  SH2trapa(sh,(8<<4)+13);
}
#define init_12_3_8_14
void FASTCALL call_12_3_8_14(SH2_struct* sh)
{
  SH2trapa(sh,(8<<4)+14);
}
#define init_12_3_8_15
void FASTCALL call_12_3_8_15(SH2_struct* sh)
{
  SH2trapa(sh,(8<<4)+15);
}
#define init_12_3_9_0
void FASTCALL call_12_3_9_0(SH2_struct* sh)
{
  SH2trapa(sh,(9<<4)+0);
}
#define init_12_3_9_1
void FASTCALL call_12_3_9_1(SH2_struct* sh)
{
  SH2trapa(sh,(9<<4)+1);
}
#define init_12_3_9_2
void FASTCALL call_12_3_9_2(SH2_struct* sh)
{
  SH2trapa(sh,(9<<4)+2);
}
#define init_12_3_9_3
void FASTCALL call_12_3_9_3(SH2_struct* sh)
{
  SH2trapa(sh,(9<<4)+3);
}
#define init_12_3_9_4
void FASTCALL call_12_3_9_4(SH2_struct* sh)
{
  SH2trapa(sh,(9<<4)+4);
}
#define init_12_3_9_5
void FASTCALL call_12_3_9_5(SH2_struct* sh)
{
  SH2trapa(sh,(9<<4)+5);
}
#define init_12_3_9_6
void FASTCALL call_12_3_9_6(SH2_struct* sh)
{
  SH2trapa(sh,(9<<4)+6);
}
#define init_12_3_9_7
void FASTCALL call_12_3_9_7(SH2_struct* sh)
{
  SH2trapa(sh,(9<<4)+7);
}
#define init_12_3_9_8
void FASTCALL call_12_3_9_8(SH2_struct* sh)
{
  SH2trapa(sh,(9<<4)+8);
}
#define init_12_3_9_9
void FASTCALL call_12_3_9_9(SH2_struct* sh)
{
  SH2trapa(sh,(9<<4)+9);
}
#define init_12_3_9_10
void FASTCALL call_12_3_9_10(SH2_struct* sh)
{
  SH2trapa(sh,(9<<4)+10);
}
#define init_12_3_9_11
void FASTCALL call_12_3_9_11(SH2_struct* sh)
{
  SH2trapa(sh,(9<<4)+11);
}
#define init_12_3_9_12
void FASTCALL call_12_3_9_12(SH2_struct* sh)
{
  SH2trapa(sh,(9<<4)+12);
}
#define init_12_3_9_13
void FASTCALL call_12_3_9_13(SH2_struct* sh)
{
  SH2trapa(sh,(9<<4)+13);
}
#define init_12_3_9_14
void FASTCALL call_12_3_9_14(SH2_struct* sh)
{
  SH2trapa(sh,(9<<4)+14);
}
#define init_12_3_9_15
void FASTCALL call_12_3_9_15(SH2_struct* sh)
{
  SH2trapa(sh,(9<<4)+15);
}
#define init_12_3_10_0
void FASTCALL call_12_3_10_0(SH2_struct* sh)
{
  SH2trapa(sh,(10<<4)+0);
}
#define init_12_3_10_1
void FASTCALL call_12_3_10_1(SH2_struct* sh)
{
  SH2trapa(sh,(10<<4)+1);
}
#define init_12_3_10_2
void FASTCALL call_12_3_10_2(SH2_struct* sh)
{
  SH2trapa(sh,(10<<4)+2);
}
#define init_12_3_10_3
void FASTCALL call_12_3_10_3(SH2_struct* sh)
{
  SH2trapa(sh,(10<<4)+3);
}
#define init_12_3_10_4
void FASTCALL call_12_3_10_4(SH2_struct* sh)
{
  SH2trapa(sh,(10<<4)+4);
}
#define init_12_3_10_5
void FASTCALL call_12_3_10_5(SH2_struct* sh)
{
  SH2trapa(sh,(10<<4)+5);
}
#define init_12_3_10_6
void FASTCALL call_12_3_10_6(SH2_struct* sh)
{
  SH2trapa(sh,(10<<4)+6);
}
#define init_12_3_10_7
void FASTCALL call_12_3_10_7(SH2_struct* sh)
{
  SH2trapa(sh,(10<<4)+7);
}
#define init_12_3_10_8
void FASTCALL call_12_3_10_8(SH2_struct* sh)
{
  SH2trapa(sh,(10<<4)+8);
}
#define init_12_3_10_9
void FASTCALL call_12_3_10_9(SH2_struct* sh)
{
  SH2trapa(sh,(10<<4)+9);
}
#define init_12_3_10_10
void FASTCALL call_12_3_10_10(SH2_struct* sh)
{
  SH2trapa(sh,(10<<4)+10);
}
#define init_12_3_10_11
void FASTCALL call_12_3_10_11(SH2_struct* sh)
{
  SH2trapa(sh,(10<<4)+11);
}
#define init_12_3_10_12
void FASTCALL call_12_3_10_12(SH2_struct* sh)
{
  SH2trapa(sh,(10<<4)+12);
}
#define init_12_3_10_13
void FASTCALL call_12_3_10_13(SH2_struct* sh)
{
  SH2trapa(sh,(10<<4)+13);
}
#define init_12_3_10_14
void FASTCALL call_12_3_10_14(SH2_struct* sh)
{
  SH2trapa(sh,(10<<4)+14);
}
#define init_12_3_10_15
void FASTCALL call_12_3_10_15(SH2_struct* sh)
{
  SH2trapa(sh,(10<<4)+15);
}
#define init_12_3_11_0
void FASTCALL call_12_3_11_0(SH2_struct* sh)
{
  SH2trapa(sh,(11<<4)+0);
}
#define init_12_3_11_1
void FASTCALL call_12_3_11_1(SH2_struct* sh)
{
  SH2trapa(sh,(11<<4)+1);
}
#define init_12_3_11_2
void FASTCALL call_12_3_11_2(SH2_struct* sh)
{
  SH2trapa(sh,(11<<4)+2);
}
#define init_12_3_11_3
void FASTCALL call_12_3_11_3(SH2_struct* sh)
{
  SH2trapa(sh,(11<<4)+3);
}
#define init_12_3_11_4
void FASTCALL call_12_3_11_4(SH2_struct* sh)
{
  SH2trapa(sh,(11<<4)+4);
}
#define init_12_3_11_5
void FASTCALL call_12_3_11_5(SH2_struct* sh)
{
  SH2trapa(sh,(11<<4)+5);
}
#define init_12_3_11_6
void FASTCALL call_12_3_11_6(SH2_struct* sh)
{
  SH2trapa(sh,(11<<4)+6);
}
#define init_12_3_11_7
void FASTCALL call_12_3_11_7(SH2_struct* sh)
{
  SH2trapa(sh,(11<<4)+7);
}
#define init_12_3_11_8
void FASTCALL call_12_3_11_8(SH2_struct* sh)
{
  SH2trapa(sh,(11<<4)+8);
}
#define init_12_3_11_9
void FASTCALL call_12_3_11_9(SH2_struct* sh)
{
  SH2trapa(sh,(11<<4)+9);
}
#define init_12_3_11_10
void FASTCALL call_12_3_11_10(SH2_struct* sh)
{
  SH2trapa(sh,(11<<4)+10);
}
#define init_12_3_11_11
void FASTCALL call_12_3_11_11(SH2_struct* sh)
{
  SH2trapa(sh,(11<<4)+11);
}
#define init_12_3_11_12
void FASTCALL call_12_3_11_12(SH2_struct* sh)
{
  SH2trapa(sh,(11<<4)+12);
}
#define init_12_3_11_13
void FASTCALL call_12_3_11_13(SH2_struct* sh)
{
  SH2trapa(sh,(11<<4)+13);
}
#define init_12_3_11_14
void FASTCALL call_12_3_11_14(SH2_struct* sh)
{
  SH2trapa(sh,(11<<4)+14);
}
#define init_12_3_11_15
void FASTCALL call_12_3_11_15(SH2_struct* sh)
{
  SH2trapa(sh,(11<<4)+15);
}
#define init_12_3_12_0
void FASTCALL call_12_3_12_0(SH2_struct* sh)
{
  SH2trapa(sh,(12<<4)+0);
}
#define init_12_3_12_1
void FASTCALL call_12_3_12_1(SH2_struct* sh)
{
  SH2trapa(sh,(12<<4)+1);
}
#define init_12_3_12_2
void FASTCALL call_12_3_12_2(SH2_struct* sh)
{
  SH2trapa(sh,(12<<4)+2);
}
#define init_12_3_12_3
void FASTCALL call_12_3_12_3(SH2_struct* sh)
{
  SH2trapa(sh,(12<<4)+3);
}
#define init_12_3_12_4
void FASTCALL call_12_3_12_4(SH2_struct* sh)
{
  SH2trapa(sh,(12<<4)+4);
}
#define init_12_3_12_5
void FASTCALL call_12_3_12_5(SH2_struct* sh)
{
  SH2trapa(sh,(12<<4)+5);
}
#define init_12_3_12_6
void FASTCALL call_12_3_12_6(SH2_struct* sh)
{
  SH2trapa(sh,(12<<4)+6);
}
#define init_12_3_12_7
void FASTCALL call_12_3_12_7(SH2_struct* sh)
{
  SH2trapa(sh,(12<<4)+7);
}
#define init_12_3_12_8
void FASTCALL call_12_3_12_8(SH2_struct* sh)
{
  SH2trapa(sh,(12<<4)+8);
}
#define init_12_3_12_9
void FASTCALL call_12_3_12_9(SH2_struct* sh)
{
  SH2trapa(sh,(12<<4)+9);
}
#define init_12_3_12_10
void FASTCALL call_12_3_12_10(SH2_struct* sh)
{
  SH2trapa(sh,(12<<4)+10);
}
#define init_12_3_12_11
void FASTCALL call_12_3_12_11(SH2_struct* sh)
{
  SH2trapa(sh,(12<<4)+11);
}
#define init_12_3_12_12
void FASTCALL call_12_3_12_12(SH2_struct* sh)
{
  SH2trapa(sh,(12<<4)+12);
}
#define init_12_3_12_13
void FASTCALL call_12_3_12_13(SH2_struct* sh)
{
  SH2trapa(sh,(12<<4)+13);
}
#define init_12_3_12_14
void FASTCALL call_12_3_12_14(SH2_struct* sh)
{
  SH2trapa(sh,(12<<4)+14);
}
#define init_12_3_12_15
void FASTCALL call_12_3_12_15(SH2_struct* sh)
{
  SH2trapa(sh,(12<<4)+15);
}
#define init_12_3_13_0
void FASTCALL call_12_3_13_0(SH2_struct* sh)
{
  SH2trapa(sh,(13<<4)+0);
}
#define init_12_3_13_1
void FASTCALL call_12_3_13_1(SH2_struct* sh)
{
  SH2trapa(sh,(13<<4)+1);
}
#define init_12_3_13_2
void FASTCALL call_12_3_13_2(SH2_struct* sh)
{
  SH2trapa(sh,(13<<4)+2);
}
#define init_12_3_13_3
void FASTCALL call_12_3_13_3(SH2_struct* sh)
{
  SH2trapa(sh,(13<<4)+3);
}
#define init_12_3_13_4
void FASTCALL call_12_3_13_4(SH2_struct* sh)
{
  SH2trapa(sh,(13<<4)+4);
}
#define init_12_3_13_5
void FASTCALL call_12_3_13_5(SH2_struct* sh)
{
  SH2trapa(sh,(13<<4)+5);
}
#define init_12_3_13_6
void FASTCALL call_12_3_13_6(SH2_struct* sh)
{
  SH2trapa(sh,(13<<4)+6);
}
#define init_12_3_13_7
void FASTCALL call_12_3_13_7(SH2_struct* sh)
{
  SH2trapa(sh,(13<<4)+7);
}
#define init_12_3_13_8
void FASTCALL call_12_3_13_8(SH2_struct* sh)
{
  SH2trapa(sh,(13<<4)+8);
}
#define init_12_3_13_9
void FASTCALL call_12_3_13_9(SH2_struct* sh)
{
  SH2trapa(sh,(13<<4)+9);
}
#define init_12_3_13_10
void FASTCALL call_12_3_13_10(SH2_struct* sh)
{
  SH2trapa(sh,(13<<4)+10);
}
#define init_12_3_13_11
void FASTCALL call_12_3_13_11(SH2_struct* sh)
{
  SH2trapa(sh,(13<<4)+11);
}
#define init_12_3_13_12
void FASTCALL call_12_3_13_12(SH2_struct* sh)
{
  SH2trapa(sh,(13<<4)+12);
}
#define init_12_3_13_13
void FASTCALL call_12_3_13_13(SH2_struct* sh)
{
  SH2trapa(sh,(13<<4)+13);
}
#define init_12_3_13_14
void FASTCALL call_12_3_13_14(SH2_struct* sh)
{
  SH2trapa(sh,(13<<4)+14);
}
#define init_12_3_13_15
void FASTCALL call_12_3_13_15(SH2_struct* sh)
{
  SH2trapa(sh,(13<<4)+15);
}
#define init_12_3_14_0
void FASTCALL call_12_3_14_0(SH2_struct* sh)
{
  SH2trapa(sh,(14<<4)+0);
}
#define init_12_3_14_1
void FASTCALL call_12_3_14_1(SH2_struct* sh)
{
  SH2trapa(sh,(14<<4)+1);
}
#define init_12_3_14_2
void FASTCALL call_12_3_14_2(SH2_struct* sh)
{
  SH2trapa(sh,(14<<4)+2);
}
#define init_12_3_14_3
void FASTCALL call_12_3_14_3(SH2_struct* sh)
{
  SH2trapa(sh,(14<<4)+3);
}
#define init_12_3_14_4
void FASTCALL call_12_3_14_4(SH2_struct* sh)
{
  SH2trapa(sh,(14<<4)+4);
}
#define init_12_3_14_5
void FASTCALL call_12_3_14_5(SH2_struct* sh)
{
  SH2trapa(sh,(14<<4)+5);
}
#define init_12_3_14_6
void FASTCALL call_12_3_14_6(SH2_struct* sh)
{
  SH2trapa(sh,(14<<4)+6);
}
#define init_12_3_14_7
void FASTCALL call_12_3_14_7(SH2_struct* sh)
{
  SH2trapa(sh,(14<<4)+7);
}
#define init_12_3_14_8
void FASTCALL call_12_3_14_8(SH2_struct* sh)
{
  SH2trapa(sh,(14<<4)+8);
}
#define init_12_3_14_9
void FASTCALL call_12_3_14_9(SH2_struct* sh)
{
  SH2trapa(sh,(14<<4)+9);
}
#define init_12_3_14_10
void FASTCALL call_12_3_14_10(SH2_struct* sh)
{
  SH2trapa(sh,(14<<4)+10);
}
#define init_12_3_14_11
void FASTCALL call_12_3_14_11(SH2_struct* sh)
{
  SH2trapa(sh,(14<<4)+11);
}
#define init_12_3_14_12
void FASTCALL call_12_3_14_12(SH2_struct* sh)
{
  SH2trapa(sh,(14<<4)+12);
}
#define init_12_3_14_13
void FASTCALL call_12_3_14_13(SH2_struct* sh)
{
  SH2trapa(sh,(14<<4)+13);
}
#define init_12_3_14_14
void FASTCALL call_12_3_14_14(SH2_struct* sh)
{
  SH2trapa(sh,(14<<4)+14);
}
#define init_12_3_14_15
void FASTCALL call_12_3_14_15(SH2_struct* sh)
{
  SH2trapa(sh,(14<<4)+15);
}
#define init_12_3_15_0
void FASTCALL call_12_3_15_0(SH2_struct* sh)
{
  SH2trapa(sh,(15<<4)+0);
}
#define init_12_3_15_1
void FASTCALL call_12_3_15_1(SH2_struct* sh)
{
  SH2trapa(sh,(15<<4)+1);
}
#define init_12_3_15_2
void FASTCALL call_12_3_15_2(SH2_struct* sh)
{
  SH2trapa(sh,(15<<4)+2);
}
#define init_12_3_15_3
void FASTCALL call_12_3_15_3(SH2_struct* sh)
{
  SH2trapa(sh,(15<<4)+3);
}
#define init_12_3_15_4
void FASTCALL call_12_3_15_4(SH2_struct* sh)
{
  SH2trapa(sh,(15<<4)+4);
}
#define init_12_3_15_5
void FASTCALL call_12_3_15_5(SH2_struct* sh)
{
  SH2trapa(sh,(15<<4)+5);
}
#define init_12_3_15_6
void FASTCALL call_12_3_15_6(SH2_struct* sh)
{
  SH2trapa(sh,(15<<4)+6);
}
#define init_12_3_15_7
void FASTCALL call_12_3_15_7(SH2_struct* sh)
{
  SH2trapa(sh,(15<<4)+7);
}
#define init_12_3_15_8
void FASTCALL call_12_3_15_8(SH2_struct* sh)
{
  SH2trapa(sh,(15<<4)+8);
}
#define init_12_3_15_9
void FASTCALL call_12_3_15_9(SH2_struct* sh)
{
  SH2trapa(sh,(15<<4)+9);
}
#define init_12_3_15_10
void FASTCALL call_12_3_15_10(SH2_struct* sh)
{
  SH2trapa(sh,(15<<4)+10);
}
#define init_12_3_15_11
void FASTCALL call_12_3_15_11(SH2_struct* sh)
{
  SH2trapa(sh,(15<<4)+11);
}
#define init_12_3_15_12
void FASTCALL call_12_3_15_12(SH2_struct* sh)
{
  SH2trapa(sh,(15<<4)+12);
}
#define init_12_3_15_13
void FASTCALL call_12_3_15_13(SH2_struct* sh)
{
  SH2trapa(sh,(15<<4)+13);
}
#define init_12_3_15_14
void FASTCALL call_12_3_15_14(SH2_struct* sh)
{
  SH2trapa(sh,(15<<4)+14);
}
#define init_12_3_15_15
void FASTCALL call_12_3_15_15(SH2_struct* sh)
{
  SH2trapa(sh,(15<<4)+15);
}
#define init_4_0_1_11
void FASTCALL call_4_0_1_11(SH2_struct* sh)
{
  SH2tas(sh,0);
}
#define init_4_1_1_11
void FASTCALL call_4_1_1_11(SH2_struct* sh)
{
  SH2tas(sh,1);
}
#define init_4_2_1_11
void FASTCALL call_4_2_1_11(SH2_struct* sh)
{
  SH2tas(sh,2);
}
#define init_4_3_1_11
void FASTCALL call_4_3_1_11(SH2_struct* sh)
{
  SH2tas(sh,3);
}
#define init_4_4_1_11
void FASTCALL call_4_4_1_11(SH2_struct* sh)
{
  SH2tas(sh,4);
}
#define init_4_5_1_11
void FASTCALL call_4_5_1_11(SH2_struct* sh)
{
  SH2tas(sh,5);
}
#define init_4_6_1_11
void FASTCALL call_4_6_1_11(SH2_struct* sh)
{
  SH2tas(sh,6);
}
#define init_4_7_1_11
void FASTCALL call_4_7_1_11(SH2_struct* sh)
{
  SH2tas(sh,7);
}
#define init_4_8_1_11
void FASTCALL call_4_8_1_11(SH2_struct* sh)
{
  SH2tas(sh,8);
}
#define init_4_9_1_11
void FASTCALL call_4_9_1_11(SH2_struct* sh)
{
  SH2tas(sh,9);
}
#define init_4_10_1_11
void FASTCALL call_4_10_1_11(SH2_struct* sh)
{
  SH2tas(sh,10);
}
#define init_4_11_1_11
void FASTCALL call_4_11_1_11(SH2_struct* sh)
{
  SH2tas(sh,11);
}
#define init_4_12_1_11
void FASTCALL call_4_12_1_11(SH2_struct* sh)
{
  SH2tas(sh,12);
}
#define init_4_13_1_11
void FASTCALL call_4_13_1_11(SH2_struct* sh)
{
  SH2tas(sh,13);
}
#define init_4_14_1_11
void FASTCALL call_4_14_1_11(SH2_struct* sh)
{
  SH2tas(sh,14);
}
#define init_4_15_1_11
void FASTCALL call_4_15_1_11(SH2_struct* sh)
{
  SH2tas(sh,15);
}
#define init_6_0_0_9
void FASTCALL call_6_0_0_9(SH2_struct* sh)
{
  SH2swapw(sh,0,0);
}
#define init_6_0_1_9
void FASTCALL call_6_0_1_9(SH2_struct* sh)
{
  SH2swapw(sh,0,1);
}
#define init_6_0_2_9
void FASTCALL call_6_0_2_9(SH2_struct* sh)
{
  SH2swapw(sh,0,2);
}
#define init_6_0_3_9
void FASTCALL call_6_0_3_9(SH2_struct* sh)
{
  SH2swapw(sh,0,3);
}
#define init_6_0_4_9
void FASTCALL call_6_0_4_9(SH2_struct* sh)
{
  SH2swapw(sh,0,4);
}
#define init_6_0_5_9
void FASTCALL call_6_0_5_9(SH2_struct* sh)
{
  SH2swapw(sh,0,5);
}
#define init_6_0_6_9
void FASTCALL call_6_0_6_9(SH2_struct* sh)
{
  SH2swapw(sh,0,6);
}
#define init_6_0_7_9
void FASTCALL call_6_0_7_9(SH2_struct* sh)
{
  SH2swapw(sh,0,7);
}
#define init_6_0_8_9
void FASTCALL call_6_0_8_9(SH2_struct* sh)
{
  SH2swapw(sh,0,8);
}
#define init_6_0_9_9
void FASTCALL call_6_0_9_9(SH2_struct* sh)
{
  SH2swapw(sh,0,9);
}
#define init_6_0_10_9
void FASTCALL call_6_0_10_9(SH2_struct* sh)
{
  SH2swapw(sh,0,10);
}
#define init_6_0_11_9
void FASTCALL call_6_0_11_9(SH2_struct* sh)
{
  SH2swapw(sh,0,11);
}
#define init_6_0_12_9
void FASTCALL call_6_0_12_9(SH2_struct* sh)
{
  SH2swapw(sh,0,12);
}
#define init_6_0_13_9
void FASTCALL call_6_0_13_9(SH2_struct* sh)
{
  SH2swapw(sh,0,13);
}
#define init_6_0_14_9
void FASTCALL call_6_0_14_9(SH2_struct* sh)
{
  SH2swapw(sh,0,14);
}
#define init_6_0_15_9
void FASTCALL call_6_0_15_9(SH2_struct* sh)
{
  SH2swapw(sh,0,15);
}
#define init_6_1_0_9
void FASTCALL call_6_1_0_9(SH2_struct* sh)
{
  SH2swapw(sh,1,0);
}
#define init_6_1_1_9
void FASTCALL call_6_1_1_9(SH2_struct* sh)
{
  SH2swapw(sh,1,1);
}
#define init_6_1_2_9
void FASTCALL call_6_1_2_9(SH2_struct* sh)
{
  SH2swapw(sh,1,2);
}
#define init_6_1_3_9
void FASTCALL call_6_1_3_9(SH2_struct* sh)
{
  SH2swapw(sh,1,3);
}
#define init_6_1_4_9
void FASTCALL call_6_1_4_9(SH2_struct* sh)
{
  SH2swapw(sh,1,4);
}
#define init_6_1_5_9
void FASTCALL call_6_1_5_9(SH2_struct* sh)
{
  SH2swapw(sh,1,5);
}
#define init_6_1_6_9
void FASTCALL call_6_1_6_9(SH2_struct* sh)
{
  SH2swapw(sh,1,6);
}
#define init_6_1_7_9
void FASTCALL call_6_1_7_9(SH2_struct* sh)
{
  SH2swapw(sh,1,7);
}
#define init_6_1_8_9
void FASTCALL call_6_1_8_9(SH2_struct* sh)
{
  SH2swapw(sh,1,8);
}
#define init_6_1_9_9
void FASTCALL call_6_1_9_9(SH2_struct* sh)
{
  SH2swapw(sh,1,9);
}
#define init_6_1_10_9
void FASTCALL call_6_1_10_9(SH2_struct* sh)
{
  SH2swapw(sh,1,10);
}
#define init_6_1_11_9
void FASTCALL call_6_1_11_9(SH2_struct* sh)
{
  SH2swapw(sh,1,11);
}
#define init_6_1_12_9
void FASTCALL call_6_1_12_9(SH2_struct* sh)
{
  SH2swapw(sh,1,12);
}
#define init_6_1_13_9
void FASTCALL call_6_1_13_9(SH2_struct* sh)
{
  SH2swapw(sh,1,13);
}
#define init_6_1_14_9
void FASTCALL call_6_1_14_9(SH2_struct* sh)
{
  SH2swapw(sh,1,14);
}
#define init_6_1_15_9
void FASTCALL call_6_1_15_9(SH2_struct* sh)
{
  SH2swapw(sh,1,15);
}
#define init_6_2_0_9
void FASTCALL call_6_2_0_9(SH2_struct* sh)
{
  SH2swapw(sh,2,0);
}
#define init_6_2_1_9
void FASTCALL call_6_2_1_9(SH2_struct* sh)
{
  SH2swapw(sh,2,1);
}
#define init_6_2_2_9
void FASTCALL call_6_2_2_9(SH2_struct* sh)
{
  SH2swapw(sh,2,2);
}
#define init_6_2_3_9
void FASTCALL call_6_2_3_9(SH2_struct* sh)
{
  SH2swapw(sh,2,3);
}
#define init_6_2_4_9
void FASTCALL call_6_2_4_9(SH2_struct* sh)
{
  SH2swapw(sh,2,4);
}
#define init_6_2_5_9
void FASTCALL call_6_2_5_9(SH2_struct* sh)
{
  SH2swapw(sh,2,5);
}
#define init_6_2_6_9
void FASTCALL call_6_2_6_9(SH2_struct* sh)
{
  SH2swapw(sh,2,6);
}
#define init_6_2_7_9
void FASTCALL call_6_2_7_9(SH2_struct* sh)
{
  SH2swapw(sh,2,7);
}
#define init_6_2_8_9
void FASTCALL call_6_2_8_9(SH2_struct* sh)
{
  SH2swapw(sh,2,8);
}
#define init_6_2_9_9
void FASTCALL call_6_2_9_9(SH2_struct* sh)
{
  SH2swapw(sh,2,9);
}
#define init_6_2_10_9
void FASTCALL call_6_2_10_9(SH2_struct* sh)
{
  SH2swapw(sh,2,10);
}
#define init_6_2_11_9
void FASTCALL call_6_2_11_9(SH2_struct* sh)
{
  SH2swapw(sh,2,11);
}
#define init_6_2_12_9
void FASTCALL call_6_2_12_9(SH2_struct* sh)
{
  SH2swapw(sh,2,12);
}
#define init_6_2_13_9
void FASTCALL call_6_2_13_9(SH2_struct* sh)
{
  SH2swapw(sh,2,13);
}
#define init_6_2_14_9
void FASTCALL call_6_2_14_9(SH2_struct* sh)
{
  SH2swapw(sh,2,14);
}
#define init_6_2_15_9
void FASTCALL call_6_2_15_9(SH2_struct* sh)
{
  SH2swapw(sh,2,15);
}
#define init_6_3_0_9
void FASTCALL call_6_3_0_9(SH2_struct* sh)
{
  SH2swapw(sh,3,0);
}
#define init_6_3_1_9
void FASTCALL call_6_3_1_9(SH2_struct* sh)
{
  SH2swapw(sh,3,1);
}
#define init_6_3_2_9
void FASTCALL call_6_3_2_9(SH2_struct* sh)
{
  SH2swapw(sh,3,2);
}
#define init_6_3_3_9
void FASTCALL call_6_3_3_9(SH2_struct* sh)
{
  SH2swapw(sh,3,3);
}
#define init_6_3_4_9
void FASTCALL call_6_3_4_9(SH2_struct* sh)
{
  SH2swapw(sh,3,4);
}
#define init_6_3_5_9
void FASTCALL call_6_3_5_9(SH2_struct* sh)
{
  SH2swapw(sh,3,5);
}
#define init_6_3_6_9
void FASTCALL call_6_3_6_9(SH2_struct* sh)
{
  SH2swapw(sh,3,6);
}
#define init_6_3_7_9
void FASTCALL call_6_3_7_9(SH2_struct* sh)
{
  SH2swapw(sh,3,7);
}
#define init_6_3_8_9
void FASTCALL call_6_3_8_9(SH2_struct* sh)
{
  SH2swapw(sh,3,8);
}
#define init_6_3_9_9
void FASTCALL call_6_3_9_9(SH2_struct* sh)
{
  SH2swapw(sh,3,9);
}
#define init_6_3_10_9
void FASTCALL call_6_3_10_9(SH2_struct* sh)
{
  SH2swapw(sh,3,10);
}
#define init_6_3_11_9
void FASTCALL call_6_3_11_9(SH2_struct* sh)
{
  SH2swapw(sh,3,11);
}
#define init_6_3_12_9
void FASTCALL call_6_3_12_9(SH2_struct* sh)
{
  SH2swapw(sh,3,12);
}
#define init_6_3_13_9
void FASTCALL call_6_3_13_9(SH2_struct* sh)
{
  SH2swapw(sh,3,13);
}
#define init_6_3_14_9
void FASTCALL call_6_3_14_9(SH2_struct* sh)
{
  SH2swapw(sh,3,14);
}
#define init_6_3_15_9
void FASTCALL call_6_3_15_9(SH2_struct* sh)
{
  SH2swapw(sh,3,15);
}
#define init_6_4_0_9
void FASTCALL call_6_4_0_9(SH2_struct* sh)
{
  SH2swapw(sh,4,0);
}
#define init_6_4_1_9
void FASTCALL call_6_4_1_9(SH2_struct* sh)
{
  SH2swapw(sh,4,1);
}
#define init_6_4_2_9
void FASTCALL call_6_4_2_9(SH2_struct* sh)
{
  SH2swapw(sh,4,2);
}
#define init_6_4_3_9
void FASTCALL call_6_4_3_9(SH2_struct* sh)
{
  SH2swapw(sh,4,3);
}
#define init_6_4_4_9
void FASTCALL call_6_4_4_9(SH2_struct* sh)
{
  SH2swapw(sh,4,4);
}
#define init_6_4_5_9
void FASTCALL call_6_4_5_9(SH2_struct* sh)
{
  SH2swapw(sh,4,5);
}
#define init_6_4_6_9
void FASTCALL call_6_4_6_9(SH2_struct* sh)
{
  SH2swapw(sh,4,6);
}
#define init_6_4_7_9
void FASTCALL call_6_4_7_9(SH2_struct* sh)
{
  SH2swapw(sh,4,7);
}
#define init_6_4_8_9
void FASTCALL call_6_4_8_9(SH2_struct* sh)
{
  SH2swapw(sh,4,8);
}
#define init_6_4_9_9
void FASTCALL call_6_4_9_9(SH2_struct* sh)
{
  SH2swapw(sh,4,9);
}
#define init_6_4_10_9
void FASTCALL call_6_4_10_9(SH2_struct* sh)
{
  SH2swapw(sh,4,10);
}
#define init_6_4_11_9
void FASTCALL call_6_4_11_9(SH2_struct* sh)
{
  SH2swapw(sh,4,11);
}
#define init_6_4_12_9
void FASTCALL call_6_4_12_9(SH2_struct* sh)
{
  SH2swapw(sh,4,12);
}
#define init_6_4_13_9
void FASTCALL call_6_4_13_9(SH2_struct* sh)
{
  SH2swapw(sh,4,13);
}
#define init_6_4_14_9
void FASTCALL call_6_4_14_9(SH2_struct* sh)
{
  SH2swapw(sh,4,14);
}
#define init_6_4_15_9
void FASTCALL call_6_4_15_9(SH2_struct* sh)
{
  SH2swapw(sh,4,15);
}
#define init_6_5_0_9
void FASTCALL call_6_5_0_9(SH2_struct* sh)
{
  SH2swapw(sh,5,0);
}
#define init_6_5_1_9
void FASTCALL call_6_5_1_9(SH2_struct* sh)
{
  SH2swapw(sh,5,1);
}
#define init_6_5_2_9
void FASTCALL call_6_5_2_9(SH2_struct* sh)
{
  SH2swapw(sh,5,2);
}
#define init_6_5_3_9
void FASTCALL call_6_5_3_9(SH2_struct* sh)
{
  SH2swapw(sh,5,3);
}
#define init_6_5_4_9
void FASTCALL call_6_5_4_9(SH2_struct* sh)
{
  SH2swapw(sh,5,4);
}
#define init_6_5_5_9
void FASTCALL call_6_5_5_9(SH2_struct* sh)
{
  SH2swapw(sh,5,5);
}
#define init_6_5_6_9
void FASTCALL call_6_5_6_9(SH2_struct* sh)
{
  SH2swapw(sh,5,6);
}
#define init_6_5_7_9
void FASTCALL call_6_5_7_9(SH2_struct* sh)
{
  SH2swapw(sh,5,7);
}
#define init_6_5_8_9
void FASTCALL call_6_5_8_9(SH2_struct* sh)
{
  SH2swapw(sh,5,8);
}
#define init_6_5_9_9
void FASTCALL call_6_5_9_9(SH2_struct* sh)
{
  SH2swapw(sh,5,9);
}
#define init_6_5_10_9
void FASTCALL call_6_5_10_9(SH2_struct* sh)
{
  SH2swapw(sh,5,10);
}
#define init_6_5_11_9
void FASTCALL call_6_5_11_9(SH2_struct* sh)
{
  SH2swapw(sh,5,11);
}
#define init_6_5_12_9
void FASTCALL call_6_5_12_9(SH2_struct* sh)
{
  SH2swapw(sh,5,12);
}
#define init_6_5_13_9
void FASTCALL call_6_5_13_9(SH2_struct* sh)
{
  SH2swapw(sh,5,13);
}
#define init_6_5_14_9
void FASTCALL call_6_5_14_9(SH2_struct* sh)
{
  SH2swapw(sh,5,14);
}
#define init_6_5_15_9
void FASTCALL call_6_5_15_9(SH2_struct* sh)
{
  SH2swapw(sh,5,15);
}
#define init_6_6_0_9
void FASTCALL call_6_6_0_9(SH2_struct* sh)
{
  SH2swapw(sh,6,0);
}
#define init_6_6_1_9
void FASTCALL call_6_6_1_9(SH2_struct* sh)
{
  SH2swapw(sh,6,1);
}
#define init_6_6_2_9
void FASTCALL call_6_6_2_9(SH2_struct* sh)
{
  SH2swapw(sh,6,2);
}
#define init_6_6_3_9
void FASTCALL call_6_6_3_9(SH2_struct* sh)
{
  SH2swapw(sh,6,3);
}
#define init_6_6_4_9
void FASTCALL call_6_6_4_9(SH2_struct* sh)
{
  SH2swapw(sh,6,4);
}
#define init_6_6_5_9
void FASTCALL call_6_6_5_9(SH2_struct* sh)
{
  SH2swapw(sh,6,5);
}
#define init_6_6_6_9
void FASTCALL call_6_6_6_9(SH2_struct* sh)
{
  SH2swapw(sh,6,6);
}
#define init_6_6_7_9
void FASTCALL call_6_6_7_9(SH2_struct* sh)
{
  SH2swapw(sh,6,7);
}
#define init_6_6_8_9
void FASTCALL call_6_6_8_9(SH2_struct* sh)
{
  SH2swapw(sh,6,8);
}
#define init_6_6_9_9
void FASTCALL call_6_6_9_9(SH2_struct* sh)
{
  SH2swapw(sh,6,9);
}
#define init_6_6_10_9
void FASTCALL call_6_6_10_9(SH2_struct* sh)
{
  SH2swapw(sh,6,10);
}
#define init_6_6_11_9
void FASTCALL call_6_6_11_9(SH2_struct* sh)
{
  SH2swapw(sh,6,11);
}
#define init_6_6_12_9
void FASTCALL call_6_6_12_9(SH2_struct* sh)
{
  SH2swapw(sh,6,12);
}
#define init_6_6_13_9
void FASTCALL call_6_6_13_9(SH2_struct* sh)
{
  SH2swapw(sh,6,13);
}
#define init_6_6_14_9
void FASTCALL call_6_6_14_9(SH2_struct* sh)
{
  SH2swapw(sh,6,14);
}
#define init_6_6_15_9
void FASTCALL call_6_6_15_9(SH2_struct* sh)
{
  SH2swapw(sh,6,15);
}
#define init_6_7_0_9
void FASTCALL call_6_7_0_9(SH2_struct* sh)
{
  SH2swapw(sh,7,0);
}
#define init_6_7_1_9
void FASTCALL call_6_7_1_9(SH2_struct* sh)
{
  SH2swapw(sh,7,1);
}
#define init_6_7_2_9
void FASTCALL call_6_7_2_9(SH2_struct* sh)
{
  SH2swapw(sh,7,2);
}
#define init_6_7_3_9
void FASTCALL call_6_7_3_9(SH2_struct* sh)
{
  SH2swapw(sh,7,3);
}
#define init_6_7_4_9
void FASTCALL call_6_7_4_9(SH2_struct* sh)
{
  SH2swapw(sh,7,4);
}
#define init_6_7_5_9
void FASTCALL call_6_7_5_9(SH2_struct* sh)
{
  SH2swapw(sh,7,5);
}
#define init_6_7_6_9
void FASTCALL call_6_7_6_9(SH2_struct* sh)
{
  SH2swapw(sh,7,6);
}
#define init_6_7_7_9
void FASTCALL call_6_7_7_9(SH2_struct* sh)
{
  SH2swapw(sh,7,7);
}
#define init_6_7_8_9
void FASTCALL call_6_7_8_9(SH2_struct* sh)
{
  SH2swapw(sh,7,8);
}
#define init_6_7_9_9
void FASTCALL call_6_7_9_9(SH2_struct* sh)
{
  SH2swapw(sh,7,9);
}
#define init_6_7_10_9
void FASTCALL call_6_7_10_9(SH2_struct* sh)
{
  SH2swapw(sh,7,10);
}
#define init_6_7_11_9
void FASTCALL call_6_7_11_9(SH2_struct* sh)
{
  SH2swapw(sh,7,11);
}
#define init_6_7_12_9
void FASTCALL call_6_7_12_9(SH2_struct* sh)
{
  SH2swapw(sh,7,12);
}
#define init_6_7_13_9
void FASTCALL call_6_7_13_9(SH2_struct* sh)
{
  SH2swapw(sh,7,13);
}
#define init_6_7_14_9
void FASTCALL call_6_7_14_9(SH2_struct* sh)
{
  SH2swapw(sh,7,14);
}
#define init_6_7_15_9
void FASTCALL call_6_7_15_9(SH2_struct* sh)
{
  SH2swapw(sh,7,15);
}
#define init_6_8_0_9
void FASTCALL call_6_8_0_9(SH2_struct* sh)
{
  SH2swapw(sh,8,0);
}
#define init_6_8_1_9
void FASTCALL call_6_8_1_9(SH2_struct* sh)
{
  SH2swapw(sh,8,1);
}
#define init_6_8_2_9
void FASTCALL call_6_8_2_9(SH2_struct* sh)
{
  SH2swapw(sh,8,2);
}
#define init_6_8_3_9
void FASTCALL call_6_8_3_9(SH2_struct* sh)
{
  SH2swapw(sh,8,3);
}
#define init_6_8_4_9
void FASTCALL call_6_8_4_9(SH2_struct* sh)
{
  SH2swapw(sh,8,4);
}
#define init_6_8_5_9
void FASTCALL call_6_8_5_9(SH2_struct* sh)
{
  SH2swapw(sh,8,5);
}
#define init_6_8_6_9
void FASTCALL call_6_8_6_9(SH2_struct* sh)
{
  SH2swapw(sh,8,6);
}
#define init_6_8_7_9
void FASTCALL call_6_8_7_9(SH2_struct* sh)
{
  SH2swapw(sh,8,7);
}
#define init_6_8_8_9
void FASTCALL call_6_8_8_9(SH2_struct* sh)
{
  SH2swapw(sh,8,8);
}
#define init_6_8_9_9
void FASTCALL call_6_8_9_9(SH2_struct* sh)
{
  SH2swapw(sh,8,9);
}
#define init_6_8_10_9
void FASTCALL call_6_8_10_9(SH2_struct* sh)
{
  SH2swapw(sh,8,10);
}
#define init_6_8_11_9
void FASTCALL call_6_8_11_9(SH2_struct* sh)
{
  SH2swapw(sh,8,11);
}
#define init_6_8_12_9
void FASTCALL call_6_8_12_9(SH2_struct* sh)
{
  SH2swapw(sh,8,12);
}
#define init_6_8_13_9
void FASTCALL call_6_8_13_9(SH2_struct* sh)
{
  SH2swapw(sh,8,13);
}
#define init_6_8_14_9
void FASTCALL call_6_8_14_9(SH2_struct* sh)
{
  SH2swapw(sh,8,14);
}
#define init_6_8_15_9
void FASTCALL call_6_8_15_9(SH2_struct* sh)
{
  SH2swapw(sh,8,15);
}
#define init_6_9_0_9
void FASTCALL call_6_9_0_9(SH2_struct* sh)
{
  SH2swapw(sh,9,0);
}
#define init_6_9_1_9
void FASTCALL call_6_9_1_9(SH2_struct* sh)
{
  SH2swapw(sh,9,1);
}
#define init_6_9_2_9
void FASTCALL call_6_9_2_9(SH2_struct* sh)
{
  SH2swapw(sh,9,2);
}
#define init_6_9_3_9
void FASTCALL call_6_9_3_9(SH2_struct* sh)
{
  SH2swapw(sh,9,3);
}
#define init_6_9_4_9
void FASTCALL call_6_9_4_9(SH2_struct* sh)
{
  SH2swapw(sh,9,4);
}
#define init_6_9_5_9
void FASTCALL call_6_9_5_9(SH2_struct* sh)
{
  SH2swapw(sh,9,5);
}
#define init_6_9_6_9
void FASTCALL call_6_9_6_9(SH2_struct* sh)
{
  SH2swapw(sh,9,6);
}
#define init_6_9_7_9
void FASTCALL call_6_9_7_9(SH2_struct* sh)
{
  SH2swapw(sh,9,7);
}
#define init_6_9_8_9
void FASTCALL call_6_9_8_9(SH2_struct* sh)
{
  SH2swapw(sh,9,8);
}
#define init_6_9_9_9
void FASTCALL call_6_9_9_9(SH2_struct* sh)
{
  SH2swapw(sh,9,9);
}
#define init_6_9_10_9
void FASTCALL call_6_9_10_9(SH2_struct* sh)
{
  SH2swapw(sh,9,10);
}
#define init_6_9_11_9
void FASTCALL call_6_9_11_9(SH2_struct* sh)
{
  SH2swapw(sh,9,11);
}
#define init_6_9_12_9
void FASTCALL call_6_9_12_9(SH2_struct* sh)
{
  SH2swapw(sh,9,12);
}
#define init_6_9_13_9
void FASTCALL call_6_9_13_9(SH2_struct* sh)
{
  SH2swapw(sh,9,13);
}
#define init_6_9_14_9
void FASTCALL call_6_9_14_9(SH2_struct* sh)
{
  SH2swapw(sh,9,14);
}
#define init_6_9_15_9
void FASTCALL call_6_9_15_9(SH2_struct* sh)
{
  SH2swapw(sh,9,15);
}
#define init_6_10_0_9
void FASTCALL call_6_10_0_9(SH2_struct* sh)
{
  SH2swapw(sh,10,0);
}
#define init_6_10_1_9
void FASTCALL call_6_10_1_9(SH2_struct* sh)
{
  SH2swapw(sh,10,1);
}
#define init_6_10_2_9
void FASTCALL call_6_10_2_9(SH2_struct* sh)
{
  SH2swapw(sh,10,2);
}
#define init_6_10_3_9
void FASTCALL call_6_10_3_9(SH2_struct* sh)
{
  SH2swapw(sh,10,3);
}
#define init_6_10_4_9
void FASTCALL call_6_10_4_9(SH2_struct* sh)
{
  SH2swapw(sh,10,4);
}
#define init_6_10_5_9
void FASTCALL call_6_10_5_9(SH2_struct* sh)
{
  SH2swapw(sh,10,5);
}
#define init_6_10_6_9
void FASTCALL call_6_10_6_9(SH2_struct* sh)
{
  SH2swapw(sh,10,6);
}
#define init_6_10_7_9
void FASTCALL call_6_10_7_9(SH2_struct* sh)
{
  SH2swapw(sh,10,7);
}
#define init_6_10_8_9
void FASTCALL call_6_10_8_9(SH2_struct* sh)
{
  SH2swapw(sh,10,8);
}
#define init_6_10_9_9
void FASTCALL call_6_10_9_9(SH2_struct* sh)
{
  SH2swapw(sh,10,9);
}
#define init_6_10_10_9
void FASTCALL call_6_10_10_9(SH2_struct* sh)
{
  SH2swapw(sh,10,10);
}
#define init_6_10_11_9
void FASTCALL call_6_10_11_9(SH2_struct* sh)
{
  SH2swapw(sh,10,11);
}
#define init_6_10_12_9
void FASTCALL call_6_10_12_9(SH2_struct* sh)
{
  SH2swapw(sh,10,12);
}
#define init_6_10_13_9
void FASTCALL call_6_10_13_9(SH2_struct* sh)
{
  SH2swapw(sh,10,13);
}
#define init_6_10_14_9
void FASTCALL call_6_10_14_9(SH2_struct* sh)
{
  SH2swapw(sh,10,14);
}
#define init_6_10_15_9
void FASTCALL call_6_10_15_9(SH2_struct* sh)
{
  SH2swapw(sh,10,15);
}
#define init_6_11_0_9
void FASTCALL call_6_11_0_9(SH2_struct* sh)
{
  SH2swapw(sh,11,0);
}
#define init_6_11_1_9
void FASTCALL call_6_11_1_9(SH2_struct* sh)
{
  SH2swapw(sh,11,1);
}
#define init_6_11_2_9
void FASTCALL call_6_11_2_9(SH2_struct* sh)
{
  SH2swapw(sh,11,2);
}
#define init_6_11_3_9
void FASTCALL call_6_11_3_9(SH2_struct* sh)
{
  SH2swapw(sh,11,3);
}
#define init_6_11_4_9
void FASTCALL call_6_11_4_9(SH2_struct* sh)
{
  SH2swapw(sh,11,4);
}
#define init_6_11_5_9
void FASTCALL call_6_11_5_9(SH2_struct* sh)
{
  SH2swapw(sh,11,5);
}
#define init_6_11_6_9
void FASTCALL call_6_11_6_9(SH2_struct* sh)
{
  SH2swapw(sh,11,6);
}
#define init_6_11_7_9
void FASTCALL call_6_11_7_9(SH2_struct* sh)
{
  SH2swapw(sh,11,7);
}
#define init_6_11_8_9
void FASTCALL call_6_11_8_9(SH2_struct* sh)
{
  SH2swapw(sh,11,8);
}
#define init_6_11_9_9
void FASTCALL call_6_11_9_9(SH2_struct* sh)
{
  SH2swapw(sh,11,9);
}
#define init_6_11_10_9
void FASTCALL call_6_11_10_9(SH2_struct* sh)
{
  SH2swapw(sh,11,10);
}
#define init_6_11_11_9
void FASTCALL call_6_11_11_9(SH2_struct* sh)
{
  SH2swapw(sh,11,11);
}
#define init_6_11_12_9
void FASTCALL call_6_11_12_9(SH2_struct* sh)
{
  SH2swapw(sh,11,12);
}
#define init_6_11_13_9
void FASTCALL call_6_11_13_9(SH2_struct* sh)
{
  SH2swapw(sh,11,13);
}
#define init_6_11_14_9
void FASTCALL call_6_11_14_9(SH2_struct* sh)
{
  SH2swapw(sh,11,14);
}
#define init_6_11_15_9
void FASTCALL call_6_11_15_9(SH2_struct* sh)
{
  SH2swapw(sh,11,15);
}
#define init_6_12_0_9
void FASTCALL call_6_12_0_9(SH2_struct* sh)
{
  SH2swapw(sh,12,0);
}
#define init_6_12_1_9
void FASTCALL call_6_12_1_9(SH2_struct* sh)
{
  SH2swapw(sh,12,1);
}
#define init_6_12_2_9
void FASTCALL call_6_12_2_9(SH2_struct* sh)
{
  SH2swapw(sh,12,2);
}
#define init_6_12_3_9
void FASTCALL call_6_12_3_9(SH2_struct* sh)
{
  SH2swapw(sh,12,3);
}
#define init_6_12_4_9
void FASTCALL call_6_12_4_9(SH2_struct* sh)
{
  SH2swapw(sh,12,4);
}
#define init_6_12_5_9
void FASTCALL call_6_12_5_9(SH2_struct* sh)
{
  SH2swapw(sh,12,5);
}
#define init_6_12_6_9
void FASTCALL call_6_12_6_9(SH2_struct* sh)
{
  SH2swapw(sh,12,6);
}
#define init_6_12_7_9
void FASTCALL call_6_12_7_9(SH2_struct* sh)
{
  SH2swapw(sh,12,7);
}
#define init_6_12_8_9
void FASTCALL call_6_12_8_9(SH2_struct* sh)
{
  SH2swapw(sh,12,8);
}
#define init_6_12_9_9
void FASTCALL call_6_12_9_9(SH2_struct* sh)
{
  SH2swapw(sh,12,9);
}
#define init_6_12_10_9
void FASTCALL call_6_12_10_9(SH2_struct* sh)
{
  SH2swapw(sh,12,10);
}
#define init_6_12_11_9
void FASTCALL call_6_12_11_9(SH2_struct* sh)
{
  SH2swapw(sh,12,11);
}
#define init_6_12_12_9
void FASTCALL call_6_12_12_9(SH2_struct* sh)
{
  SH2swapw(sh,12,12);
}
#define init_6_12_13_9
void FASTCALL call_6_12_13_9(SH2_struct* sh)
{
  SH2swapw(sh,12,13);
}
#define init_6_12_14_9
void FASTCALL call_6_12_14_9(SH2_struct* sh)
{
  SH2swapw(sh,12,14);
}
#define init_6_12_15_9
void FASTCALL call_6_12_15_9(SH2_struct* sh)
{
  SH2swapw(sh,12,15);
}
#define init_6_13_0_9
void FASTCALL call_6_13_0_9(SH2_struct* sh)
{
  SH2swapw(sh,13,0);
}
#define init_6_13_1_9
void FASTCALL call_6_13_1_9(SH2_struct* sh)
{
  SH2swapw(sh,13,1);
}
#define init_6_13_2_9
void FASTCALL call_6_13_2_9(SH2_struct* sh)
{
  SH2swapw(sh,13,2);
}
#define init_6_13_3_9
void FASTCALL call_6_13_3_9(SH2_struct* sh)
{
  SH2swapw(sh,13,3);
}
#define init_6_13_4_9
void FASTCALL call_6_13_4_9(SH2_struct* sh)
{
  SH2swapw(sh,13,4);
}
#define init_6_13_5_9
void FASTCALL call_6_13_5_9(SH2_struct* sh)
{
  SH2swapw(sh,13,5);
}
#define init_6_13_6_9
void FASTCALL call_6_13_6_9(SH2_struct* sh)
{
  SH2swapw(sh,13,6);
}
#define init_6_13_7_9
void FASTCALL call_6_13_7_9(SH2_struct* sh)
{
  SH2swapw(sh,13,7);
}
#define init_6_13_8_9
void FASTCALL call_6_13_8_9(SH2_struct* sh)
{
  SH2swapw(sh,13,8);
}
#define init_6_13_9_9
void FASTCALL call_6_13_9_9(SH2_struct* sh)
{
  SH2swapw(sh,13,9);
}
#define init_6_13_10_9
void FASTCALL call_6_13_10_9(SH2_struct* sh)
{
  SH2swapw(sh,13,10);
}
#define init_6_13_11_9
void FASTCALL call_6_13_11_9(SH2_struct* sh)
{
  SH2swapw(sh,13,11);
}
#define init_6_13_12_9
void FASTCALL call_6_13_12_9(SH2_struct* sh)
{
  SH2swapw(sh,13,12);
}
#define init_6_13_13_9
void FASTCALL call_6_13_13_9(SH2_struct* sh)
{
  SH2swapw(sh,13,13);
}
#define init_6_13_14_9
void FASTCALL call_6_13_14_9(SH2_struct* sh)
{
  SH2swapw(sh,13,14);
}
#define init_6_13_15_9
void FASTCALL call_6_13_15_9(SH2_struct* sh)
{
  SH2swapw(sh,13,15);
}
#define init_6_14_0_9
void FASTCALL call_6_14_0_9(SH2_struct* sh)
{
  SH2swapw(sh,14,0);
}
#define init_6_14_1_9
void FASTCALL call_6_14_1_9(SH2_struct* sh)
{
  SH2swapw(sh,14,1);
}
#define init_6_14_2_9
void FASTCALL call_6_14_2_9(SH2_struct* sh)
{
  SH2swapw(sh,14,2);
}
#define init_6_14_3_9
void FASTCALL call_6_14_3_9(SH2_struct* sh)
{
  SH2swapw(sh,14,3);
}
#define init_6_14_4_9
void FASTCALL call_6_14_4_9(SH2_struct* sh)
{
  SH2swapw(sh,14,4);
}
#define init_6_14_5_9
void FASTCALL call_6_14_5_9(SH2_struct* sh)
{
  SH2swapw(sh,14,5);
}
#define init_6_14_6_9
void FASTCALL call_6_14_6_9(SH2_struct* sh)
{
  SH2swapw(sh,14,6);
}
#define init_6_14_7_9
void FASTCALL call_6_14_7_9(SH2_struct* sh)
{
  SH2swapw(sh,14,7);
}
#define init_6_14_8_9
void FASTCALL call_6_14_8_9(SH2_struct* sh)
{
  SH2swapw(sh,14,8);
}
#define init_6_14_9_9
void FASTCALL call_6_14_9_9(SH2_struct* sh)
{
  SH2swapw(sh,14,9);
}
#define init_6_14_10_9
void FASTCALL call_6_14_10_9(SH2_struct* sh)
{
  SH2swapw(sh,14,10);
}
#define init_6_14_11_9
void FASTCALL call_6_14_11_9(SH2_struct* sh)
{
  SH2swapw(sh,14,11);
}
#define init_6_14_12_9
void FASTCALL call_6_14_12_9(SH2_struct* sh)
{
  SH2swapw(sh,14,12);
}
#define init_6_14_13_9
void FASTCALL call_6_14_13_9(SH2_struct* sh)
{
  SH2swapw(sh,14,13);
}
#define init_6_14_14_9
void FASTCALL call_6_14_14_9(SH2_struct* sh)
{
  SH2swapw(sh,14,14);
}
#define init_6_14_15_9
void FASTCALL call_6_14_15_9(SH2_struct* sh)
{
  SH2swapw(sh,14,15);
}
#define init_6_15_0_9
void FASTCALL call_6_15_0_9(SH2_struct* sh)
{
  SH2swapw(sh,15,0);
}
#define init_6_15_1_9
void FASTCALL call_6_15_1_9(SH2_struct* sh)
{
  SH2swapw(sh,15,1);
}
#define init_6_15_2_9
void FASTCALL call_6_15_2_9(SH2_struct* sh)
{
  SH2swapw(sh,15,2);
}
#define init_6_15_3_9
void FASTCALL call_6_15_3_9(SH2_struct* sh)
{
  SH2swapw(sh,15,3);
}
#define init_6_15_4_9
void FASTCALL call_6_15_4_9(SH2_struct* sh)
{
  SH2swapw(sh,15,4);
}
#define init_6_15_5_9
void FASTCALL call_6_15_5_9(SH2_struct* sh)
{
  SH2swapw(sh,15,5);
}
#define init_6_15_6_9
void FASTCALL call_6_15_6_9(SH2_struct* sh)
{
  SH2swapw(sh,15,6);
}
#define init_6_15_7_9
void FASTCALL call_6_15_7_9(SH2_struct* sh)
{
  SH2swapw(sh,15,7);
}
#define init_6_15_8_9
void FASTCALL call_6_15_8_9(SH2_struct* sh)
{
  SH2swapw(sh,15,8);
}
#define init_6_15_9_9
void FASTCALL call_6_15_9_9(SH2_struct* sh)
{
  SH2swapw(sh,15,9);
}
#define init_6_15_10_9
void FASTCALL call_6_15_10_9(SH2_struct* sh)
{
  SH2swapw(sh,15,10);
}
#define init_6_15_11_9
void FASTCALL call_6_15_11_9(SH2_struct* sh)
{
  SH2swapw(sh,15,11);
}
#define init_6_15_12_9
void FASTCALL call_6_15_12_9(SH2_struct* sh)
{
  SH2swapw(sh,15,12);
}
#define init_6_15_13_9
void FASTCALL call_6_15_13_9(SH2_struct* sh)
{
  SH2swapw(sh,15,13);
}
#define init_6_15_14_9
void FASTCALL call_6_15_14_9(SH2_struct* sh)
{
  SH2swapw(sh,15,14);
}
#define init_6_15_15_9
void FASTCALL call_6_15_15_9(SH2_struct* sh)
{
  SH2swapw(sh,15,15);
}
#define init_6_0_0_8
void FASTCALL call_6_0_0_8(SH2_struct* sh)
{
  SH2swapb(sh,0,0);
}
#define init_6_0_1_8
void FASTCALL call_6_0_1_8(SH2_struct* sh)
{
  SH2swapb(sh,0,1);
}
#define init_6_0_2_8
void FASTCALL call_6_0_2_8(SH2_struct* sh)
{
  SH2swapb(sh,0,2);
}
#define init_6_0_3_8
void FASTCALL call_6_0_3_8(SH2_struct* sh)
{
  SH2swapb(sh,0,3);
}
#define init_6_0_4_8
void FASTCALL call_6_0_4_8(SH2_struct* sh)
{
  SH2swapb(sh,0,4);
}
#define init_6_0_5_8
void FASTCALL call_6_0_5_8(SH2_struct* sh)
{
  SH2swapb(sh,0,5);
}
#define init_6_0_6_8
void FASTCALL call_6_0_6_8(SH2_struct* sh)
{
  SH2swapb(sh,0,6);
}
#define init_6_0_7_8
void FASTCALL call_6_0_7_8(SH2_struct* sh)
{
  SH2swapb(sh,0,7);
}
#define init_6_0_8_8
void FASTCALL call_6_0_8_8(SH2_struct* sh)
{
  SH2swapb(sh,0,8);
}
#define init_6_0_9_8
void FASTCALL call_6_0_9_8(SH2_struct* sh)
{
  SH2swapb(sh,0,9);
}
#define init_6_0_10_8
void FASTCALL call_6_0_10_8(SH2_struct* sh)
{
  SH2swapb(sh,0,10);
}
#define init_6_0_11_8
void FASTCALL call_6_0_11_8(SH2_struct* sh)
{
  SH2swapb(sh,0,11);
}
#define init_6_0_12_8
void FASTCALL call_6_0_12_8(SH2_struct* sh)
{
  SH2swapb(sh,0,12);
}
#define init_6_0_13_8
void FASTCALL call_6_0_13_8(SH2_struct* sh)
{
  SH2swapb(sh,0,13);
}
#define init_6_0_14_8
void FASTCALL call_6_0_14_8(SH2_struct* sh)
{
  SH2swapb(sh,0,14);
}
#define init_6_0_15_8
void FASTCALL call_6_0_15_8(SH2_struct* sh)
{
  SH2swapb(sh,0,15);
}
#define init_6_1_0_8
void FASTCALL call_6_1_0_8(SH2_struct* sh)
{
  SH2swapb(sh,1,0);
}
#define init_6_1_1_8
void FASTCALL call_6_1_1_8(SH2_struct* sh)
{
  SH2swapb(sh,1,1);
}
#define init_6_1_2_8
void FASTCALL call_6_1_2_8(SH2_struct* sh)
{
  SH2swapb(sh,1,2);
}
#define init_6_1_3_8
void FASTCALL call_6_1_3_8(SH2_struct* sh)
{
  SH2swapb(sh,1,3);
}
#define init_6_1_4_8
void FASTCALL call_6_1_4_8(SH2_struct* sh)
{
  SH2swapb(sh,1,4);
}
#define init_6_1_5_8
void FASTCALL call_6_1_5_8(SH2_struct* sh)
{
  SH2swapb(sh,1,5);
}
#define init_6_1_6_8
void FASTCALL call_6_1_6_8(SH2_struct* sh)
{
  SH2swapb(sh,1,6);
}
#define init_6_1_7_8
void FASTCALL call_6_1_7_8(SH2_struct* sh)
{
  SH2swapb(sh,1,7);
}
#define init_6_1_8_8
void FASTCALL call_6_1_8_8(SH2_struct* sh)
{
  SH2swapb(sh,1,8);
}
#define init_6_1_9_8
void FASTCALL call_6_1_9_8(SH2_struct* sh)
{
  SH2swapb(sh,1,9);
}
#define init_6_1_10_8
void FASTCALL call_6_1_10_8(SH2_struct* sh)
{
  SH2swapb(sh,1,10);
}
#define init_6_1_11_8
void FASTCALL call_6_1_11_8(SH2_struct* sh)
{
  SH2swapb(sh,1,11);
}
#define init_6_1_12_8
void FASTCALL call_6_1_12_8(SH2_struct* sh)
{
  SH2swapb(sh,1,12);
}
#define init_6_1_13_8
void FASTCALL call_6_1_13_8(SH2_struct* sh)
{
  SH2swapb(sh,1,13);
}
#define init_6_1_14_8
void FASTCALL call_6_1_14_8(SH2_struct* sh)
{
  SH2swapb(sh,1,14);
}
#define init_6_1_15_8
void FASTCALL call_6_1_15_8(SH2_struct* sh)
{
  SH2swapb(sh,1,15);
}
#define init_6_2_0_8
void FASTCALL call_6_2_0_8(SH2_struct* sh)
{
  SH2swapb(sh,2,0);
}
#define init_6_2_1_8
void FASTCALL call_6_2_1_8(SH2_struct* sh)
{
  SH2swapb(sh,2,1);
}
#define init_6_2_2_8
void FASTCALL call_6_2_2_8(SH2_struct* sh)
{
  SH2swapb(sh,2,2);
}
#define init_6_2_3_8
void FASTCALL call_6_2_3_8(SH2_struct* sh)
{
  SH2swapb(sh,2,3);
}
#define init_6_2_4_8
void FASTCALL call_6_2_4_8(SH2_struct* sh)
{
  SH2swapb(sh,2,4);
}
#define init_6_2_5_8
void FASTCALL call_6_2_5_8(SH2_struct* sh)
{
  SH2swapb(sh,2,5);
}
#define init_6_2_6_8
void FASTCALL call_6_2_6_8(SH2_struct* sh)
{
  SH2swapb(sh,2,6);
}
#define init_6_2_7_8
void FASTCALL call_6_2_7_8(SH2_struct* sh)
{
  SH2swapb(sh,2,7);
}
#define init_6_2_8_8
void FASTCALL call_6_2_8_8(SH2_struct* sh)
{
  SH2swapb(sh,2,8);
}
#define init_6_2_9_8
void FASTCALL call_6_2_9_8(SH2_struct* sh)
{
  SH2swapb(sh,2,9);
}
#define init_6_2_10_8
void FASTCALL call_6_2_10_8(SH2_struct* sh)
{
  SH2swapb(sh,2,10);
}
#define init_6_2_11_8
void FASTCALL call_6_2_11_8(SH2_struct* sh)
{
  SH2swapb(sh,2,11);
}
#define init_6_2_12_8
void FASTCALL call_6_2_12_8(SH2_struct* sh)
{
  SH2swapb(sh,2,12);
}
#define init_6_2_13_8
void FASTCALL call_6_2_13_8(SH2_struct* sh)
{
  SH2swapb(sh,2,13);
}
#define init_6_2_14_8
void FASTCALL call_6_2_14_8(SH2_struct* sh)
{
  SH2swapb(sh,2,14);
}
#define init_6_2_15_8
void FASTCALL call_6_2_15_8(SH2_struct* sh)
{
  SH2swapb(sh,2,15);
}
#define init_6_3_0_8
void FASTCALL call_6_3_0_8(SH2_struct* sh)
{
  SH2swapb(sh,3,0);
}
#define init_6_3_1_8
void FASTCALL call_6_3_1_8(SH2_struct* sh)
{
  SH2swapb(sh,3,1);
}
#define init_6_3_2_8
void FASTCALL call_6_3_2_8(SH2_struct* sh)
{
  SH2swapb(sh,3,2);
}
#define init_6_3_3_8
void FASTCALL call_6_3_3_8(SH2_struct* sh)
{
  SH2swapb(sh,3,3);
}
#define init_6_3_4_8
void FASTCALL call_6_3_4_8(SH2_struct* sh)
{
  SH2swapb(sh,3,4);
}
#define init_6_3_5_8
void FASTCALL call_6_3_5_8(SH2_struct* sh)
{
  SH2swapb(sh,3,5);
}
#define init_6_3_6_8
void FASTCALL call_6_3_6_8(SH2_struct* sh)
{
  SH2swapb(sh,3,6);
}
#define init_6_3_7_8
void FASTCALL call_6_3_7_8(SH2_struct* sh)
{
  SH2swapb(sh,3,7);
}
#define init_6_3_8_8
void FASTCALL call_6_3_8_8(SH2_struct* sh)
{
  SH2swapb(sh,3,8);
}
#define init_6_3_9_8
void FASTCALL call_6_3_9_8(SH2_struct* sh)
{
  SH2swapb(sh,3,9);
}
#define init_6_3_10_8
void FASTCALL call_6_3_10_8(SH2_struct* sh)
{
  SH2swapb(sh,3,10);
}
#define init_6_3_11_8
void FASTCALL call_6_3_11_8(SH2_struct* sh)
{
  SH2swapb(sh,3,11);
}
#define init_6_3_12_8
void FASTCALL call_6_3_12_8(SH2_struct* sh)
{
  SH2swapb(sh,3,12);
}
#define init_6_3_13_8
void FASTCALL call_6_3_13_8(SH2_struct* sh)
{
  SH2swapb(sh,3,13);
}
#define init_6_3_14_8
void FASTCALL call_6_3_14_8(SH2_struct* sh)
{
  SH2swapb(sh,3,14);
}
#define init_6_3_15_8
void FASTCALL call_6_3_15_8(SH2_struct* sh)
{
  SH2swapb(sh,3,15);
}
#define init_6_4_0_8
void FASTCALL call_6_4_0_8(SH2_struct* sh)
{
  SH2swapb(sh,4,0);
}
#define init_6_4_1_8
void FASTCALL call_6_4_1_8(SH2_struct* sh)
{
  SH2swapb(sh,4,1);
}
#define init_6_4_2_8
void FASTCALL call_6_4_2_8(SH2_struct* sh)
{
  SH2swapb(sh,4,2);
}
#define init_6_4_3_8
void FASTCALL call_6_4_3_8(SH2_struct* sh)
{
  SH2swapb(sh,4,3);
}
#define init_6_4_4_8
void FASTCALL call_6_4_4_8(SH2_struct* sh)
{
  SH2swapb(sh,4,4);
}
#define init_6_4_5_8
void FASTCALL call_6_4_5_8(SH2_struct* sh)
{
  SH2swapb(sh,4,5);
}
#define init_6_4_6_8
void FASTCALL call_6_4_6_8(SH2_struct* sh)
{
  SH2swapb(sh,4,6);
}
#define init_6_4_7_8
void FASTCALL call_6_4_7_8(SH2_struct* sh)
{
  SH2swapb(sh,4,7);
}
#define init_6_4_8_8
void FASTCALL call_6_4_8_8(SH2_struct* sh)
{
  SH2swapb(sh,4,8);
}
#define init_6_4_9_8
void FASTCALL call_6_4_9_8(SH2_struct* sh)
{
  SH2swapb(sh,4,9);
}
#define init_6_4_10_8
void FASTCALL call_6_4_10_8(SH2_struct* sh)
{
  SH2swapb(sh,4,10);
}
#define init_6_4_11_8
void FASTCALL call_6_4_11_8(SH2_struct* sh)
{
  SH2swapb(sh,4,11);
}
#define init_6_4_12_8
void FASTCALL call_6_4_12_8(SH2_struct* sh)
{
  SH2swapb(sh,4,12);
}
#define init_6_4_13_8
void FASTCALL call_6_4_13_8(SH2_struct* sh)
{
  SH2swapb(sh,4,13);
}
#define init_6_4_14_8
void FASTCALL call_6_4_14_8(SH2_struct* sh)
{
  SH2swapb(sh,4,14);
}
#define init_6_4_15_8
void FASTCALL call_6_4_15_8(SH2_struct* sh)
{
  SH2swapb(sh,4,15);
}
#define init_6_5_0_8
void FASTCALL call_6_5_0_8(SH2_struct* sh)
{
  SH2swapb(sh,5,0);
}
#define init_6_5_1_8
void FASTCALL call_6_5_1_8(SH2_struct* sh)
{
  SH2swapb(sh,5,1);
}
#define init_6_5_2_8
void FASTCALL call_6_5_2_8(SH2_struct* sh)
{
  SH2swapb(sh,5,2);
}
#define init_6_5_3_8
void FASTCALL call_6_5_3_8(SH2_struct* sh)
{
  SH2swapb(sh,5,3);
}
#define init_6_5_4_8
void FASTCALL call_6_5_4_8(SH2_struct* sh)
{
  SH2swapb(sh,5,4);
}
#define init_6_5_5_8
void FASTCALL call_6_5_5_8(SH2_struct* sh)
{
  SH2swapb(sh,5,5);
}
#define init_6_5_6_8
void FASTCALL call_6_5_6_8(SH2_struct* sh)
{
  SH2swapb(sh,5,6);
}
#define init_6_5_7_8
void FASTCALL call_6_5_7_8(SH2_struct* sh)
{
  SH2swapb(sh,5,7);
}
#define init_6_5_8_8
void FASTCALL call_6_5_8_8(SH2_struct* sh)
{
  SH2swapb(sh,5,8);
}
#define init_6_5_9_8
void FASTCALL call_6_5_9_8(SH2_struct* sh)
{
  SH2swapb(sh,5,9);
}
#define init_6_5_10_8
void FASTCALL call_6_5_10_8(SH2_struct* sh)
{
  SH2swapb(sh,5,10);
}
#define init_6_5_11_8
void FASTCALL call_6_5_11_8(SH2_struct* sh)
{
  SH2swapb(sh,5,11);
}
#define init_6_5_12_8
void FASTCALL call_6_5_12_8(SH2_struct* sh)
{
  SH2swapb(sh,5,12);
}
#define init_6_5_13_8
void FASTCALL call_6_5_13_8(SH2_struct* sh)
{
  SH2swapb(sh,5,13);
}
#define init_6_5_14_8
void FASTCALL call_6_5_14_8(SH2_struct* sh)
{
  SH2swapb(sh,5,14);
}
#define init_6_5_15_8
void FASTCALL call_6_5_15_8(SH2_struct* sh)
{
  SH2swapb(sh,5,15);
}
#define init_6_6_0_8
void FASTCALL call_6_6_0_8(SH2_struct* sh)
{
  SH2swapb(sh,6,0);
}
#define init_6_6_1_8
void FASTCALL call_6_6_1_8(SH2_struct* sh)
{
  SH2swapb(sh,6,1);
}
#define init_6_6_2_8
void FASTCALL call_6_6_2_8(SH2_struct* sh)
{
  SH2swapb(sh,6,2);
}
#define init_6_6_3_8
void FASTCALL call_6_6_3_8(SH2_struct* sh)
{
  SH2swapb(sh,6,3);
}
#define init_6_6_4_8
void FASTCALL call_6_6_4_8(SH2_struct* sh)
{
  SH2swapb(sh,6,4);
}
#define init_6_6_5_8
void FASTCALL call_6_6_5_8(SH2_struct* sh)
{
  SH2swapb(sh,6,5);
}
#define init_6_6_6_8
void FASTCALL call_6_6_6_8(SH2_struct* sh)
{
  SH2swapb(sh,6,6);
}
#define init_6_6_7_8
void FASTCALL call_6_6_7_8(SH2_struct* sh)
{
  SH2swapb(sh,6,7);
}
#define init_6_6_8_8
void FASTCALL call_6_6_8_8(SH2_struct* sh)
{
  SH2swapb(sh,6,8);
}
#define init_6_6_9_8
void FASTCALL call_6_6_9_8(SH2_struct* sh)
{
  SH2swapb(sh,6,9);
}
#define init_6_6_10_8
void FASTCALL call_6_6_10_8(SH2_struct* sh)
{
  SH2swapb(sh,6,10);
}
#define init_6_6_11_8
void FASTCALL call_6_6_11_8(SH2_struct* sh)
{
  SH2swapb(sh,6,11);
}
#define init_6_6_12_8
void FASTCALL call_6_6_12_8(SH2_struct* sh)
{
  SH2swapb(sh,6,12);
}
#define init_6_6_13_8
void FASTCALL call_6_6_13_8(SH2_struct* sh)
{
  SH2swapb(sh,6,13);
}
#define init_6_6_14_8
void FASTCALL call_6_6_14_8(SH2_struct* sh)
{
  SH2swapb(sh,6,14);
}
#define init_6_6_15_8
void FASTCALL call_6_6_15_8(SH2_struct* sh)
{
  SH2swapb(sh,6,15);
}
#define init_6_7_0_8
void FASTCALL call_6_7_0_8(SH2_struct* sh)
{
  SH2swapb(sh,7,0);
}
#define init_6_7_1_8
void FASTCALL call_6_7_1_8(SH2_struct* sh)
{
  SH2swapb(sh,7,1);
}
#define init_6_7_2_8
void FASTCALL call_6_7_2_8(SH2_struct* sh)
{
  SH2swapb(sh,7,2);
}
#define init_6_7_3_8
void FASTCALL call_6_7_3_8(SH2_struct* sh)
{
  SH2swapb(sh,7,3);
}
#define init_6_7_4_8
void FASTCALL call_6_7_4_8(SH2_struct* sh)
{
  SH2swapb(sh,7,4);
}
#define init_6_7_5_8
void FASTCALL call_6_7_5_8(SH2_struct* sh)
{
  SH2swapb(sh,7,5);
}
#define init_6_7_6_8
void FASTCALL call_6_7_6_8(SH2_struct* sh)
{
  SH2swapb(sh,7,6);
}
#define init_6_7_7_8
void FASTCALL call_6_7_7_8(SH2_struct* sh)
{
  SH2swapb(sh,7,7);
}
#define init_6_7_8_8
void FASTCALL call_6_7_8_8(SH2_struct* sh)
{
  SH2swapb(sh,7,8);
}
#define init_6_7_9_8
void FASTCALL call_6_7_9_8(SH2_struct* sh)
{
  SH2swapb(sh,7,9);
}
#define init_6_7_10_8
void FASTCALL call_6_7_10_8(SH2_struct* sh)
{
  SH2swapb(sh,7,10);
}
#define init_6_7_11_8
void FASTCALL call_6_7_11_8(SH2_struct* sh)
{
  SH2swapb(sh,7,11);
}
#define init_6_7_12_8
void FASTCALL call_6_7_12_8(SH2_struct* sh)
{
  SH2swapb(sh,7,12);
}
#define init_6_7_13_8
void FASTCALL call_6_7_13_8(SH2_struct* sh)
{
  SH2swapb(sh,7,13);
}
#define init_6_7_14_8
void FASTCALL call_6_7_14_8(SH2_struct* sh)
{
  SH2swapb(sh,7,14);
}
#define init_6_7_15_8
void FASTCALL call_6_7_15_8(SH2_struct* sh)
{
  SH2swapb(sh,7,15);
}
#define init_6_8_0_8
void FASTCALL call_6_8_0_8(SH2_struct* sh)
{
  SH2swapb(sh,8,0);
}
#define init_6_8_1_8
void FASTCALL call_6_8_1_8(SH2_struct* sh)
{
  SH2swapb(sh,8,1);
}
#define init_6_8_2_8
void FASTCALL call_6_8_2_8(SH2_struct* sh)
{
  SH2swapb(sh,8,2);
}
#define init_6_8_3_8
void FASTCALL call_6_8_3_8(SH2_struct* sh)
{
  SH2swapb(sh,8,3);
}
#define init_6_8_4_8
void FASTCALL call_6_8_4_8(SH2_struct* sh)
{
  SH2swapb(sh,8,4);
}
#define init_6_8_5_8
void FASTCALL call_6_8_5_8(SH2_struct* sh)
{
  SH2swapb(sh,8,5);
}
#define init_6_8_6_8
void FASTCALL call_6_8_6_8(SH2_struct* sh)
{
  SH2swapb(sh,8,6);
}
#define init_6_8_7_8
void FASTCALL call_6_8_7_8(SH2_struct* sh)
{
  SH2swapb(sh,8,7);
}
#define init_6_8_8_8
void FASTCALL call_6_8_8_8(SH2_struct* sh)
{
  SH2swapb(sh,8,8);
}
#define init_6_8_9_8
void FASTCALL call_6_8_9_8(SH2_struct* sh)
{
  SH2swapb(sh,8,9);
}
#define init_6_8_10_8
void FASTCALL call_6_8_10_8(SH2_struct* sh)
{
  SH2swapb(sh,8,10);
}
#define init_6_8_11_8
void FASTCALL call_6_8_11_8(SH2_struct* sh)
{
  SH2swapb(sh,8,11);
}
#define init_6_8_12_8
void FASTCALL call_6_8_12_8(SH2_struct* sh)
{
  SH2swapb(sh,8,12);
}
#define init_6_8_13_8
void FASTCALL call_6_8_13_8(SH2_struct* sh)
{
  SH2swapb(sh,8,13);
}
#define init_6_8_14_8
void FASTCALL call_6_8_14_8(SH2_struct* sh)
{
  SH2swapb(sh,8,14);
}
#define init_6_8_15_8
void FASTCALL call_6_8_15_8(SH2_struct* sh)
{
  SH2swapb(sh,8,15);
}
#define init_6_9_0_8
void FASTCALL call_6_9_0_8(SH2_struct* sh)
{
  SH2swapb(sh,9,0);
}
#define init_6_9_1_8
void FASTCALL call_6_9_1_8(SH2_struct* sh)
{
  SH2swapb(sh,9,1);
}
#define init_6_9_2_8
void FASTCALL call_6_9_2_8(SH2_struct* sh)
{
  SH2swapb(sh,9,2);
}
#define init_6_9_3_8
void FASTCALL call_6_9_3_8(SH2_struct* sh)
{
  SH2swapb(sh,9,3);
}
#define init_6_9_4_8
void FASTCALL call_6_9_4_8(SH2_struct* sh)
{
  SH2swapb(sh,9,4);
}
#define init_6_9_5_8
void FASTCALL call_6_9_5_8(SH2_struct* sh)
{
  SH2swapb(sh,9,5);
}
#define init_6_9_6_8
void FASTCALL call_6_9_6_8(SH2_struct* sh)
{
  SH2swapb(sh,9,6);
}
#define init_6_9_7_8
void FASTCALL call_6_9_7_8(SH2_struct* sh)
{
  SH2swapb(sh,9,7);
}
#define init_6_9_8_8
void FASTCALL call_6_9_8_8(SH2_struct* sh)
{
  SH2swapb(sh,9,8);
}
#define init_6_9_9_8
void FASTCALL call_6_9_9_8(SH2_struct* sh)
{
  SH2swapb(sh,9,9);
}
#define init_6_9_10_8
void FASTCALL call_6_9_10_8(SH2_struct* sh)
{
  SH2swapb(sh,9,10);
}
#define init_6_9_11_8
void FASTCALL call_6_9_11_8(SH2_struct* sh)
{
  SH2swapb(sh,9,11);
}
#define init_6_9_12_8
void FASTCALL call_6_9_12_8(SH2_struct* sh)
{
  SH2swapb(sh,9,12);
}
#define init_6_9_13_8
void FASTCALL call_6_9_13_8(SH2_struct* sh)
{
  SH2swapb(sh,9,13);
}
#define init_6_9_14_8
void FASTCALL call_6_9_14_8(SH2_struct* sh)
{
  SH2swapb(sh,9,14);
}
#define init_6_9_15_8
void FASTCALL call_6_9_15_8(SH2_struct* sh)
{
  SH2swapb(sh,9,15);
}
#define init_6_10_0_8
void FASTCALL call_6_10_0_8(SH2_struct* sh)
{
  SH2swapb(sh,10,0);
}
#define init_6_10_1_8
void FASTCALL call_6_10_1_8(SH2_struct* sh)
{
  SH2swapb(sh,10,1);
}
#define init_6_10_2_8
void FASTCALL call_6_10_2_8(SH2_struct* sh)
{
  SH2swapb(sh,10,2);
}
#define init_6_10_3_8
void FASTCALL call_6_10_3_8(SH2_struct* sh)
{
  SH2swapb(sh,10,3);
}
#define init_6_10_4_8
void FASTCALL call_6_10_4_8(SH2_struct* sh)
{
  SH2swapb(sh,10,4);
}
#define init_6_10_5_8
void FASTCALL call_6_10_5_8(SH2_struct* sh)
{
  SH2swapb(sh,10,5);
}
#define init_6_10_6_8
void FASTCALL call_6_10_6_8(SH2_struct* sh)
{
  SH2swapb(sh,10,6);
}
#define init_6_10_7_8
void FASTCALL call_6_10_7_8(SH2_struct* sh)
{
  SH2swapb(sh,10,7);
}
#define init_6_10_8_8
void FASTCALL call_6_10_8_8(SH2_struct* sh)
{
  SH2swapb(sh,10,8);
}
#define init_6_10_9_8
void FASTCALL call_6_10_9_8(SH2_struct* sh)
{
  SH2swapb(sh,10,9);
}
#define init_6_10_10_8
void FASTCALL call_6_10_10_8(SH2_struct* sh)
{
  SH2swapb(sh,10,10);
}
#define init_6_10_11_8
void FASTCALL call_6_10_11_8(SH2_struct* sh)
{
  SH2swapb(sh,10,11);
}
#define init_6_10_12_8
void FASTCALL call_6_10_12_8(SH2_struct* sh)
{
  SH2swapb(sh,10,12);
}
#define init_6_10_13_8
void FASTCALL call_6_10_13_8(SH2_struct* sh)
{
  SH2swapb(sh,10,13);
}
#define init_6_10_14_8
void FASTCALL call_6_10_14_8(SH2_struct* sh)
{
  SH2swapb(sh,10,14);
}
#define init_6_10_15_8
void FASTCALL call_6_10_15_8(SH2_struct* sh)
{
  SH2swapb(sh,10,15);
}
#define init_6_11_0_8
void FASTCALL call_6_11_0_8(SH2_struct* sh)
{
  SH2swapb(sh,11,0);
}
#define init_6_11_1_8
void FASTCALL call_6_11_1_8(SH2_struct* sh)
{
  SH2swapb(sh,11,1);
}
#define init_6_11_2_8
void FASTCALL call_6_11_2_8(SH2_struct* sh)
{
  SH2swapb(sh,11,2);
}
#define init_6_11_3_8
void FASTCALL call_6_11_3_8(SH2_struct* sh)
{
  SH2swapb(sh,11,3);
}
#define init_6_11_4_8
void FASTCALL call_6_11_4_8(SH2_struct* sh)
{
  SH2swapb(sh,11,4);
}
#define init_6_11_5_8
void FASTCALL call_6_11_5_8(SH2_struct* sh)
{
  SH2swapb(sh,11,5);
}
#define init_6_11_6_8
void FASTCALL call_6_11_6_8(SH2_struct* sh)
{
  SH2swapb(sh,11,6);
}
#define init_6_11_7_8
void FASTCALL call_6_11_7_8(SH2_struct* sh)
{
  SH2swapb(sh,11,7);
}
#define init_6_11_8_8
void FASTCALL call_6_11_8_8(SH2_struct* sh)
{
  SH2swapb(sh,11,8);
}
#define init_6_11_9_8
void FASTCALL call_6_11_9_8(SH2_struct* sh)
{
  SH2swapb(sh,11,9);
}
#define init_6_11_10_8
void FASTCALL call_6_11_10_8(SH2_struct* sh)
{
  SH2swapb(sh,11,10);
}
#define init_6_11_11_8
void FASTCALL call_6_11_11_8(SH2_struct* sh)
{
  SH2swapb(sh,11,11);
}
#define init_6_11_12_8
void FASTCALL call_6_11_12_8(SH2_struct* sh)
{
  SH2swapb(sh,11,12);
}
#define init_6_11_13_8
void FASTCALL call_6_11_13_8(SH2_struct* sh)
{
  SH2swapb(sh,11,13);
}
#define init_6_11_14_8
void FASTCALL call_6_11_14_8(SH2_struct* sh)
{
  SH2swapb(sh,11,14);
}
#define init_6_11_15_8
void FASTCALL call_6_11_15_8(SH2_struct* sh)
{
  SH2swapb(sh,11,15);
}
#define init_6_12_0_8
void FASTCALL call_6_12_0_8(SH2_struct* sh)
{
  SH2swapb(sh,12,0);
}
#define init_6_12_1_8
void FASTCALL call_6_12_1_8(SH2_struct* sh)
{
  SH2swapb(sh,12,1);
}
#define init_6_12_2_8
void FASTCALL call_6_12_2_8(SH2_struct* sh)
{
  SH2swapb(sh,12,2);
}
#define init_6_12_3_8
void FASTCALL call_6_12_3_8(SH2_struct* sh)
{
  SH2swapb(sh,12,3);
}
#define init_6_12_4_8
void FASTCALL call_6_12_4_8(SH2_struct* sh)
{
  SH2swapb(sh,12,4);
}
#define init_6_12_5_8
void FASTCALL call_6_12_5_8(SH2_struct* sh)
{
  SH2swapb(sh,12,5);
}
#define init_6_12_6_8
void FASTCALL call_6_12_6_8(SH2_struct* sh)
{
  SH2swapb(sh,12,6);
}
#define init_6_12_7_8
void FASTCALL call_6_12_7_8(SH2_struct* sh)
{
  SH2swapb(sh,12,7);
}
#define init_6_12_8_8
void FASTCALL call_6_12_8_8(SH2_struct* sh)
{
  SH2swapb(sh,12,8);
}
#define init_6_12_9_8
void FASTCALL call_6_12_9_8(SH2_struct* sh)
{
  SH2swapb(sh,12,9);
}
#define init_6_12_10_8
void FASTCALL call_6_12_10_8(SH2_struct* sh)
{
  SH2swapb(sh,12,10);
}
#define init_6_12_11_8
void FASTCALL call_6_12_11_8(SH2_struct* sh)
{
  SH2swapb(sh,12,11);
}
#define init_6_12_12_8
void FASTCALL call_6_12_12_8(SH2_struct* sh)
{
  SH2swapb(sh,12,12);
}
#define init_6_12_13_8
void FASTCALL call_6_12_13_8(SH2_struct* sh)
{
  SH2swapb(sh,12,13);
}
#define init_6_12_14_8
void FASTCALL call_6_12_14_8(SH2_struct* sh)
{
  SH2swapb(sh,12,14);
}
#define init_6_12_15_8
void FASTCALL call_6_12_15_8(SH2_struct* sh)
{
  SH2swapb(sh,12,15);
}
#define init_6_13_0_8
void FASTCALL call_6_13_0_8(SH2_struct* sh)
{
  SH2swapb(sh,13,0);
}
#define init_6_13_1_8
void FASTCALL call_6_13_1_8(SH2_struct* sh)
{
  SH2swapb(sh,13,1);
}
#define init_6_13_2_8
void FASTCALL call_6_13_2_8(SH2_struct* sh)
{
  SH2swapb(sh,13,2);
}
#define init_6_13_3_8
void FASTCALL call_6_13_3_8(SH2_struct* sh)
{
  SH2swapb(sh,13,3);
}
#define init_6_13_4_8
void FASTCALL call_6_13_4_8(SH2_struct* sh)
{
  SH2swapb(sh,13,4);
}
#define init_6_13_5_8
void FASTCALL call_6_13_5_8(SH2_struct* sh)
{
  SH2swapb(sh,13,5);
}
#define init_6_13_6_8
void FASTCALL call_6_13_6_8(SH2_struct* sh)
{
  SH2swapb(sh,13,6);
}
#define init_6_13_7_8
void FASTCALL call_6_13_7_8(SH2_struct* sh)
{
  SH2swapb(sh,13,7);
}
#define init_6_13_8_8
void FASTCALL call_6_13_8_8(SH2_struct* sh)
{
  SH2swapb(sh,13,8);
}
#define init_6_13_9_8
void FASTCALL call_6_13_9_8(SH2_struct* sh)
{
  SH2swapb(sh,13,9);
}
#define init_6_13_10_8
void FASTCALL call_6_13_10_8(SH2_struct* sh)
{
  SH2swapb(sh,13,10);
}
#define init_6_13_11_8
void FASTCALL call_6_13_11_8(SH2_struct* sh)
{
  SH2swapb(sh,13,11);
}
#define init_6_13_12_8
void FASTCALL call_6_13_12_8(SH2_struct* sh)
{
  SH2swapb(sh,13,12);
}
#define init_6_13_13_8
void FASTCALL call_6_13_13_8(SH2_struct* sh)
{
  SH2swapb(sh,13,13);
}
#define init_6_13_14_8
void FASTCALL call_6_13_14_8(SH2_struct* sh)
{
  SH2swapb(sh,13,14);
}
#define init_6_13_15_8
void FASTCALL call_6_13_15_8(SH2_struct* sh)
{
  SH2swapb(sh,13,15);
}
#define init_6_14_0_8
void FASTCALL call_6_14_0_8(SH2_struct* sh)
{
  SH2swapb(sh,14,0);
}
#define init_6_14_1_8
void FASTCALL call_6_14_1_8(SH2_struct* sh)
{
  SH2swapb(sh,14,1);
}
#define init_6_14_2_8
void FASTCALL call_6_14_2_8(SH2_struct* sh)
{
  SH2swapb(sh,14,2);
}
#define init_6_14_3_8
void FASTCALL call_6_14_3_8(SH2_struct* sh)
{
  SH2swapb(sh,14,3);
}
#define init_6_14_4_8
void FASTCALL call_6_14_4_8(SH2_struct* sh)
{
  SH2swapb(sh,14,4);
}
#define init_6_14_5_8
void FASTCALL call_6_14_5_8(SH2_struct* sh)
{
  SH2swapb(sh,14,5);
}
#define init_6_14_6_8
void FASTCALL call_6_14_6_8(SH2_struct* sh)
{
  SH2swapb(sh,14,6);
}
#define init_6_14_7_8
void FASTCALL call_6_14_7_8(SH2_struct* sh)
{
  SH2swapb(sh,14,7);
}
#define init_6_14_8_8
void FASTCALL call_6_14_8_8(SH2_struct* sh)
{
  SH2swapb(sh,14,8);
}
#define init_6_14_9_8
void FASTCALL call_6_14_9_8(SH2_struct* sh)
{
  SH2swapb(sh,14,9);
}
#define init_6_14_10_8
void FASTCALL call_6_14_10_8(SH2_struct* sh)
{
  SH2swapb(sh,14,10);
}
#define init_6_14_11_8
void FASTCALL call_6_14_11_8(SH2_struct* sh)
{
  SH2swapb(sh,14,11);
}
#define init_6_14_12_8
void FASTCALL call_6_14_12_8(SH2_struct* sh)
{
  SH2swapb(sh,14,12);
}
#define init_6_14_13_8
void FASTCALL call_6_14_13_8(SH2_struct* sh)
{
  SH2swapb(sh,14,13);
}
#define init_6_14_14_8
void FASTCALL call_6_14_14_8(SH2_struct* sh)
{
  SH2swapb(sh,14,14);
}
#define init_6_14_15_8
void FASTCALL call_6_14_15_8(SH2_struct* sh)
{
  SH2swapb(sh,14,15);
}
#define init_6_15_0_8
void FASTCALL call_6_15_0_8(SH2_struct* sh)
{
  SH2swapb(sh,15,0);
}
#define init_6_15_1_8
void FASTCALL call_6_15_1_8(SH2_struct* sh)
{
  SH2swapb(sh,15,1);
}
#define init_6_15_2_8
void FASTCALL call_6_15_2_8(SH2_struct* sh)
{
  SH2swapb(sh,15,2);
}
#define init_6_15_3_8
void FASTCALL call_6_15_3_8(SH2_struct* sh)
{
  SH2swapb(sh,15,3);
}
#define init_6_15_4_8
void FASTCALL call_6_15_4_8(SH2_struct* sh)
{
  SH2swapb(sh,15,4);
}
#define init_6_15_5_8
void FASTCALL call_6_15_5_8(SH2_struct* sh)
{
  SH2swapb(sh,15,5);
}
#define init_6_15_6_8
void FASTCALL call_6_15_6_8(SH2_struct* sh)
{
  SH2swapb(sh,15,6);
}
#define init_6_15_7_8
void FASTCALL call_6_15_7_8(SH2_struct* sh)
{
  SH2swapb(sh,15,7);
}
#define init_6_15_8_8
void FASTCALL call_6_15_8_8(SH2_struct* sh)
{
  SH2swapb(sh,15,8);
}
#define init_6_15_9_8
void FASTCALL call_6_15_9_8(SH2_struct* sh)
{
  SH2swapb(sh,15,9);
}
#define init_6_15_10_8
void FASTCALL call_6_15_10_8(SH2_struct* sh)
{
  SH2swapb(sh,15,10);
}
#define init_6_15_11_8
void FASTCALL call_6_15_11_8(SH2_struct* sh)
{
  SH2swapb(sh,15,11);
}
#define init_6_15_12_8
void FASTCALL call_6_15_12_8(SH2_struct* sh)
{
  SH2swapb(sh,15,12);
}
#define init_6_15_13_8
void FASTCALL call_6_15_13_8(SH2_struct* sh)
{
  SH2swapb(sh,15,13);
}
#define init_6_15_14_8
void FASTCALL call_6_15_14_8(SH2_struct* sh)
{
  SH2swapb(sh,15,14);
}
#define init_6_15_15_8
void FASTCALL call_6_15_15_8(SH2_struct* sh)
{
  SH2swapb(sh,15,15);
}
#define init_3_0_0_11
void FASTCALL call_3_0_0_11(SH2_struct* sh)
{
  SH2subv(sh,0,0);
}
#define init_3_0_1_11
void FASTCALL call_3_0_1_11(SH2_struct* sh)
{
  SH2subv(sh,0,1);
}
#define init_3_0_2_11
void FASTCALL call_3_0_2_11(SH2_struct* sh)
{
  SH2subv(sh,0,2);
}
#define init_3_0_3_11
void FASTCALL call_3_0_3_11(SH2_struct* sh)
{
  SH2subv(sh,0,3);
}
#define init_3_0_4_11
void FASTCALL call_3_0_4_11(SH2_struct* sh)
{
  SH2subv(sh,0,4);
}
#define init_3_0_5_11
void FASTCALL call_3_0_5_11(SH2_struct* sh)
{
  SH2subv(sh,0,5);
}
#define init_3_0_6_11
void FASTCALL call_3_0_6_11(SH2_struct* sh)
{
  SH2subv(sh,0,6);
}
#define init_3_0_7_11
void FASTCALL call_3_0_7_11(SH2_struct* sh)
{
  SH2subv(sh,0,7);
}
#define init_3_0_8_11
void FASTCALL call_3_0_8_11(SH2_struct* sh)
{
  SH2subv(sh,0,8);
}
#define init_3_0_9_11
void FASTCALL call_3_0_9_11(SH2_struct* sh)
{
  SH2subv(sh,0,9);
}
#define init_3_0_10_11
void FASTCALL call_3_0_10_11(SH2_struct* sh)
{
  SH2subv(sh,0,10);
}
#define init_3_0_11_11
void FASTCALL call_3_0_11_11(SH2_struct* sh)
{
  SH2subv(sh,0,11);
}
#define init_3_0_12_11
void FASTCALL call_3_0_12_11(SH2_struct* sh)
{
  SH2subv(sh,0,12);
}
#define init_3_0_13_11
void FASTCALL call_3_0_13_11(SH2_struct* sh)
{
  SH2subv(sh,0,13);
}
#define init_3_0_14_11
void FASTCALL call_3_0_14_11(SH2_struct* sh)
{
  SH2subv(sh,0,14);
}
#define init_3_0_15_11
void FASTCALL call_3_0_15_11(SH2_struct* sh)
{
  SH2subv(sh,0,15);
}
#define init_3_1_0_11
void FASTCALL call_3_1_0_11(SH2_struct* sh)
{
  SH2subv(sh,1,0);
}
#define init_3_1_1_11
void FASTCALL call_3_1_1_11(SH2_struct* sh)
{
  SH2subv(sh,1,1);
}
#define init_3_1_2_11
void FASTCALL call_3_1_2_11(SH2_struct* sh)
{
  SH2subv(sh,1,2);
}
#define init_3_1_3_11
void FASTCALL call_3_1_3_11(SH2_struct* sh)
{
  SH2subv(sh,1,3);
}
#define init_3_1_4_11
void FASTCALL call_3_1_4_11(SH2_struct* sh)
{
  SH2subv(sh,1,4);
}
#define init_3_1_5_11
void FASTCALL call_3_1_5_11(SH2_struct* sh)
{
  SH2subv(sh,1,5);
}
#define init_3_1_6_11
void FASTCALL call_3_1_6_11(SH2_struct* sh)
{
  SH2subv(sh,1,6);
}
#define init_3_1_7_11
void FASTCALL call_3_1_7_11(SH2_struct* sh)
{
  SH2subv(sh,1,7);
}
#define init_3_1_8_11
void FASTCALL call_3_1_8_11(SH2_struct* sh)
{
  SH2subv(sh,1,8);
}
#define init_3_1_9_11
void FASTCALL call_3_1_9_11(SH2_struct* sh)
{
  SH2subv(sh,1,9);
}
#define init_3_1_10_11
void FASTCALL call_3_1_10_11(SH2_struct* sh)
{
  SH2subv(sh,1,10);
}
#define init_3_1_11_11
void FASTCALL call_3_1_11_11(SH2_struct* sh)
{
  SH2subv(sh,1,11);
}
#define init_3_1_12_11
void FASTCALL call_3_1_12_11(SH2_struct* sh)
{
  SH2subv(sh,1,12);
}
#define init_3_1_13_11
void FASTCALL call_3_1_13_11(SH2_struct* sh)
{
  SH2subv(sh,1,13);
}
#define init_3_1_14_11
void FASTCALL call_3_1_14_11(SH2_struct* sh)
{
  SH2subv(sh,1,14);
}
#define init_3_1_15_11
void FASTCALL call_3_1_15_11(SH2_struct* sh)
{
  SH2subv(sh,1,15);
}
#define init_3_2_0_11
void FASTCALL call_3_2_0_11(SH2_struct* sh)
{
  SH2subv(sh,2,0);
}
#define init_3_2_1_11
void FASTCALL call_3_2_1_11(SH2_struct* sh)
{
  SH2subv(sh,2,1);
}
#define init_3_2_2_11
void FASTCALL call_3_2_2_11(SH2_struct* sh)
{
  SH2subv(sh,2,2);
}
#define init_3_2_3_11
void FASTCALL call_3_2_3_11(SH2_struct* sh)
{
  SH2subv(sh,2,3);
}
#define init_3_2_4_11
void FASTCALL call_3_2_4_11(SH2_struct* sh)
{
  SH2subv(sh,2,4);
}
#define init_3_2_5_11
void FASTCALL call_3_2_5_11(SH2_struct* sh)
{
  SH2subv(sh,2,5);
}
#define init_3_2_6_11
void FASTCALL call_3_2_6_11(SH2_struct* sh)
{
  SH2subv(sh,2,6);
}
#define init_3_2_7_11
void FASTCALL call_3_2_7_11(SH2_struct* sh)
{
  SH2subv(sh,2,7);
}
#define init_3_2_8_11
void FASTCALL call_3_2_8_11(SH2_struct* sh)
{
  SH2subv(sh,2,8);
}
#define init_3_2_9_11
void FASTCALL call_3_2_9_11(SH2_struct* sh)
{
  SH2subv(sh,2,9);
}
#define init_3_2_10_11
void FASTCALL call_3_2_10_11(SH2_struct* sh)
{
  SH2subv(sh,2,10);
}
#define init_3_2_11_11
void FASTCALL call_3_2_11_11(SH2_struct* sh)
{
  SH2subv(sh,2,11);
}
#define init_3_2_12_11
void FASTCALL call_3_2_12_11(SH2_struct* sh)
{
  SH2subv(sh,2,12);
}
#define init_3_2_13_11
void FASTCALL call_3_2_13_11(SH2_struct* sh)
{
  SH2subv(sh,2,13);
}
#define init_3_2_14_11
void FASTCALL call_3_2_14_11(SH2_struct* sh)
{
  SH2subv(sh,2,14);
}
#define init_3_2_15_11
void FASTCALL call_3_2_15_11(SH2_struct* sh)
{
  SH2subv(sh,2,15);
}
#define init_3_3_0_11
void FASTCALL call_3_3_0_11(SH2_struct* sh)
{
  SH2subv(sh,3,0);
}
#define init_3_3_1_11
void FASTCALL call_3_3_1_11(SH2_struct* sh)
{
  SH2subv(sh,3,1);
}
#define init_3_3_2_11
void FASTCALL call_3_3_2_11(SH2_struct* sh)
{
  SH2subv(sh,3,2);
}
#define init_3_3_3_11
void FASTCALL call_3_3_3_11(SH2_struct* sh)
{
  SH2subv(sh,3,3);
}
#define init_3_3_4_11
void FASTCALL call_3_3_4_11(SH2_struct* sh)
{
  SH2subv(sh,3,4);
}
#define init_3_3_5_11
void FASTCALL call_3_3_5_11(SH2_struct* sh)
{
  SH2subv(sh,3,5);
}
#define init_3_3_6_11
void FASTCALL call_3_3_6_11(SH2_struct* sh)
{
  SH2subv(sh,3,6);
}
#define init_3_3_7_11
void FASTCALL call_3_3_7_11(SH2_struct* sh)
{
  SH2subv(sh,3,7);
}
#define init_3_3_8_11
void FASTCALL call_3_3_8_11(SH2_struct* sh)
{
  SH2subv(sh,3,8);
}
#define init_3_3_9_11
void FASTCALL call_3_3_9_11(SH2_struct* sh)
{
  SH2subv(sh,3,9);
}
#define init_3_3_10_11
void FASTCALL call_3_3_10_11(SH2_struct* sh)
{
  SH2subv(sh,3,10);
}
#define init_3_3_11_11
void FASTCALL call_3_3_11_11(SH2_struct* sh)
{
  SH2subv(sh,3,11);
}
#define init_3_3_12_11
void FASTCALL call_3_3_12_11(SH2_struct* sh)
{
  SH2subv(sh,3,12);
}
#define init_3_3_13_11
void FASTCALL call_3_3_13_11(SH2_struct* sh)
{
  SH2subv(sh,3,13);
}
#define init_3_3_14_11
void FASTCALL call_3_3_14_11(SH2_struct* sh)
{
  SH2subv(sh,3,14);
}
#define init_3_3_15_11
void FASTCALL call_3_3_15_11(SH2_struct* sh)
{
  SH2subv(sh,3,15);
}
#define init_3_4_0_11
void FASTCALL call_3_4_0_11(SH2_struct* sh)
{
  SH2subv(sh,4,0);
}
#define init_3_4_1_11
void FASTCALL call_3_4_1_11(SH2_struct* sh)
{
  SH2subv(sh,4,1);
}
#define init_3_4_2_11
void FASTCALL call_3_4_2_11(SH2_struct* sh)
{
  SH2subv(sh,4,2);
}
#define init_3_4_3_11
void FASTCALL call_3_4_3_11(SH2_struct* sh)
{
  SH2subv(sh,4,3);
}
#define init_3_4_4_11
void FASTCALL call_3_4_4_11(SH2_struct* sh)
{
  SH2subv(sh,4,4);
}
#define init_3_4_5_11
void FASTCALL call_3_4_5_11(SH2_struct* sh)
{
  SH2subv(sh,4,5);
}
#define init_3_4_6_11
void FASTCALL call_3_4_6_11(SH2_struct* sh)
{
  SH2subv(sh,4,6);
}
#define init_3_4_7_11
void FASTCALL call_3_4_7_11(SH2_struct* sh)
{
  SH2subv(sh,4,7);
}
#define init_3_4_8_11
void FASTCALL call_3_4_8_11(SH2_struct* sh)
{
  SH2subv(sh,4,8);
}
#define init_3_4_9_11
void FASTCALL call_3_4_9_11(SH2_struct* sh)
{
  SH2subv(sh,4,9);
}
#define init_3_4_10_11
void FASTCALL call_3_4_10_11(SH2_struct* sh)
{
  SH2subv(sh,4,10);
}
#define init_3_4_11_11
void FASTCALL call_3_4_11_11(SH2_struct* sh)
{
  SH2subv(sh,4,11);
}
#define init_3_4_12_11
void FASTCALL call_3_4_12_11(SH2_struct* sh)
{
  SH2subv(sh,4,12);
}
#define init_3_4_13_11
void FASTCALL call_3_4_13_11(SH2_struct* sh)
{
  SH2subv(sh,4,13);
}
#define init_3_4_14_11
void FASTCALL call_3_4_14_11(SH2_struct* sh)
{
  SH2subv(sh,4,14);
}
#define init_3_4_15_11
void FASTCALL call_3_4_15_11(SH2_struct* sh)
{
  SH2subv(sh,4,15);
}
#define init_3_5_0_11
void FASTCALL call_3_5_0_11(SH2_struct* sh)
{
  SH2subv(sh,5,0);
}
#define init_3_5_1_11
void FASTCALL call_3_5_1_11(SH2_struct* sh)
{
  SH2subv(sh,5,1);
}
#define init_3_5_2_11
void FASTCALL call_3_5_2_11(SH2_struct* sh)
{
  SH2subv(sh,5,2);
}
#define init_3_5_3_11
void FASTCALL call_3_5_3_11(SH2_struct* sh)
{
  SH2subv(sh,5,3);
}
#define init_3_5_4_11
void FASTCALL call_3_5_4_11(SH2_struct* sh)
{
  SH2subv(sh,5,4);
}
#define init_3_5_5_11
void FASTCALL call_3_5_5_11(SH2_struct* sh)
{
  SH2subv(sh,5,5);
}
#define init_3_5_6_11
void FASTCALL call_3_5_6_11(SH2_struct* sh)
{
  SH2subv(sh,5,6);
}
#define init_3_5_7_11
void FASTCALL call_3_5_7_11(SH2_struct* sh)
{
  SH2subv(sh,5,7);
}
#define init_3_5_8_11
void FASTCALL call_3_5_8_11(SH2_struct* sh)
{
  SH2subv(sh,5,8);
}
#define init_3_5_9_11
void FASTCALL call_3_5_9_11(SH2_struct* sh)
{
  SH2subv(sh,5,9);
}
#define init_3_5_10_11
void FASTCALL call_3_5_10_11(SH2_struct* sh)
{
  SH2subv(sh,5,10);
}
#define init_3_5_11_11
void FASTCALL call_3_5_11_11(SH2_struct* sh)
{
  SH2subv(sh,5,11);
}
#define init_3_5_12_11
void FASTCALL call_3_5_12_11(SH2_struct* sh)
{
  SH2subv(sh,5,12);
}
#define init_3_5_13_11
void FASTCALL call_3_5_13_11(SH2_struct* sh)
{
  SH2subv(sh,5,13);
}
#define init_3_5_14_11
void FASTCALL call_3_5_14_11(SH2_struct* sh)
{
  SH2subv(sh,5,14);
}
#define init_3_5_15_11
void FASTCALL call_3_5_15_11(SH2_struct* sh)
{
  SH2subv(sh,5,15);
}
#define init_3_6_0_11
void FASTCALL call_3_6_0_11(SH2_struct* sh)
{
  SH2subv(sh,6,0);
}
#define init_3_6_1_11
void FASTCALL call_3_6_1_11(SH2_struct* sh)
{
  SH2subv(sh,6,1);
}
#define init_3_6_2_11
void FASTCALL call_3_6_2_11(SH2_struct* sh)
{
  SH2subv(sh,6,2);
}
#define init_3_6_3_11
void FASTCALL call_3_6_3_11(SH2_struct* sh)
{
  SH2subv(sh,6,3);
}
#define init_3_6_4_11
void FASTCALL call_3_6_4_11(SH2_struct* sh)
{
  SH2subv(sh,6,4);
}
#define init_3_6_5_11
void FASTCALL call_3_6_5_11(SH2_struct* sh)
{
  SH2subv(sh,6,5);
}
#define init_3_6_6_11
void FASTCALL call_3_6_6_11(SH2_struct* sh)
{
  SH2subv(sh,6,6);
}
#define init_3_6_7_11
void FASTCALL call_3_6_7_11(SH2_struct* sh)
{
  SH2subv(sh,6,7);
}
#define init_3_6_8_11
void FASTCALL call_3_6_8_11(SH2_struct* sh)
{
  SH2subv(sh,6,8);
}
#define init_3_6_9_11
void FASTCALL call_3_6_9_11(SH2_struct* sh)
{
  SH2subv(sh,6,9);
}
#define init_3_6_10_11
void FASTCALL call_3_6_10_11(SH2_struct* sh)
{
  SH2subv(sh,6,10);
}
#define init_3_6_11_11
void FASTCALL call_3_6_11_11(SH2_struct* sh)
{
  SH2subv(sh,6,11);
}
#define init_3_6_12_11
void FASTCALL call_3_6_12_11(SH2_struct* sh)
{
  SH2subv(sh,6,12);
}
#define init_3_6_13_11
void FASTCALL call_3_6_13_11(SH2_struct* sh)
{
  SH2subv(sh,6,13);
}
#define init_3_6_14_11
void FASTCALL call_3_6_14_11(SH2_struct* sh)
{
  SH2subv(sh,6,14);
}
#define init_3_6_15_11
void FASTCALL call_3_6_15_11(SH2_struct* sh)
{
  SH2subv(sh,6,15);
}
#define init_3_7_0_11
void FASTCALL call_3_7_0_11(SH2_struct* sh)
{
  SH2subv(sh,7,0);
}
#define init_3_7_1_11
void FASTCALL call_3_7_1_11(SH2_struct* sh)
{
  SH2subv(sh,7,1);
}
#define init_3_7_2_11
void FASTCALL call_3_7_2_11(SH2_struct* sh)
{
  SH2subv(sh,7,2);
}
#define init_3_7_3_11
void FASTCALL call_3_7_3_11(SH2_struct* sh)
{
  SH2subv(sh,7,3);
}
#define init_3_7_4_11
void FASTCALL call_3_7_4_11(SH2_struct* sh)
{
  SH2subv(sh,7,4);
}
#define init_3_7_5_11
void FASTCALL call_3_7_5_11(SH2_struct* sh)
{
  SH2subv(sh,7,5);
}
#define init_3_7_6_11
void FASTCALL call_3_7_6_11(SH2_struct* sh)
{
  SH2subv(sh,7,6);
}
#define init_3_7_7_11
void FASTCALL call_3_7_7_11(SH2_struct* sh)
{
  SH2subv(sh,7,7);
}
#define init_3_7_8_11
void FASTCALL call_3_7_8_11(SH2_struct* sh)
{
  SH2subv(sh,7,8);
}
#define init_3_7_9_11
void FASTCALL call_3_7_9_11(SH2_struct* sh)
{
  SH2subv(sh,7,9);
}
#define init_3_7_10_11
void FASTCALL call_3_7_10_11(SH2_struct* sh)
{
  SH2subv(sh,7,10);
}
#define init_3_7_11_11
void FASTCALL call_3_7_11_11(SH2_struct* sh)
{
  SH2subv(sh,7,11);
}
#define init_3_7_12_11
void FASTCALL call_3_7_12_11(SH2_struct* sh)
{
  SH2subv(sh,7,12);
}
#define init_3_7_13_11
void FASTCALL call_3_7_13_11(SH2_struct* sh)
{
  SH2subv(sh,7,13);
}
#define init_3_7_14_11
void FASTCALL call_3_7_14_11(SH2_struct* sh)
{
  SH2subv(sh,7,14);
}
#define init_3_7_15_11
void FASTCALL call_3_7_15_11(SH2_struct* sh)
{
  SH2subv(sh,7,15);
}
#define init_3_8_0_11
void FASTCALL call_3_8_0_11(SH2_struct* sh)
{
  SH2subv(sh,8,0);
}
#define init_3_8_1_11
void FASTCALL call_3_8_1_11(SH2_struct* sh)
{
  SH2subv(sh,8,1);
}
#define init_3_8_2_11
void FASTCALL call_3_8_2_11(SH2_struct* sh)
{
  SH2subv(sh,8,2);
}
#define init_3_8_3_11
void FASTCALL call_3_8_3_11(SH2_struct* sh)
{
  SH2subv(sh,8,3);
}
#define init_3_8_4_11
void FASTCALL call_3_8_4_11(SH2_struct* sh)
{
  SH2subv(sh,8,4);
}
#define init_3_8_5_11
void FASTCALL call_3_8_5_11(SH2_struct* sh)
{
  SH2subv(sh,8,5);
}
#define init_3_8_6_11
void FASTCALL call_3_8_6_11(SH2_struct* sh)
{
  SH2subv(sh,8,6);
}
#define init_3_8_7_11
void FASTCALL call_3_8_7_11(SH2_struct* sh)
{
  SH2subv(sh,8,7);
}
#define init_3_8_8_11
void FASTCALL call_3_8_8_11(SH2_struct* sh)
{
  SH2subv(sh,8,8);
}
#define init_3_8_9_11
void FASTCALL call_3_8_9_11(SH2_struct* sh)
{
  SH2subv(sh,8,9);
}
#define init_3_8_10_11
void FASTCALL call_3_8_10_11(SH2_struct* sh)
{
  SH2subv(sh,8,10);
}
#define init_3_8_11_11
void FASTCALL call_3_8_11_11(SH2_struct* sh)
{
  SH2subv(sh,8,11);
}
#define init_3_8_12_11
void FASTCALL call_3_8_12_11(SH2_struct* sh)
{
  SH2subv(sh,8,12);
}
#define init_3_8_13_11
void FASTCALL call_3_8_13_11(SH2_struct* sh)
{
  SH2subv(sh,8,13);
}
#define init_3_8_14_11
void FASTCALL call_3_8_14_11(SH2_struct* sh)
{
  SH2subv(sh,8,14);
}
#define init_3_8_15_11
void FASTCALL call_3_8_15_11(SH2_struct* sh)
{
  SH2subv(sh,8,15);
}
#define init_3_9_0_11
void FASTCALL call_3_9_0_11(SH2_struct* sh)
{
  SH2subv(sh,9,0);
}
#define init_3_9_1_11
void FASTCALL call_3_9_1_11(SH2_struct* sh)
{
  SH2subv(sh,9,1);
}
#define init_3_9_2_11
void FASTCALL call_3_9_2_11(SH2_struct* sh)
{
  SH2subv(sh,9,2);
}
#define init_3_9_3_11
void FASTCALL call_3_9_3_11(SH2_struct* sh)
{
  SH2subv(sh,9,3);
}
#define init_3_9_4_11
void FASTCALL call_3_9_4_11(SH2_struct* sh)
{
  SH2subv(sh,9,4);
}
#define init_3_9_5_11
void FASTCALL call_3_9_5_11(SH2_struct* sh)
{
  SH2subv(sh,9,5);
}
#define init_3_9_6_11
void FASTCALL call_3_9_6_11(SH2_struct* sh)
{
  SH2subv(sh,9,6);
}
#define init_3_9_7_11
void FASTCALL call_3_9_7_11(SH2_struct* sh)
{
  SH2subv(sh,9,7);
}
#define init_3_9_8_11
void FASTCALL call_3_9_8_11(SH2_struct* sh)
{
  SH2subv(sh,9,8);
}
#define init_3_9_9_11
void FASTCALL call_3_9_9_11(SH2_struct* sh)
{
  SH2subv(sh,9,9);
}
#define init_3_9_10_11
void FASTCALL call_3_9_10_11(SH2_struct* sh)
{
  SH2subv(sh,9,10);
}
#define init_3_9_11_11
void FASTCALL call_3_9_11_11(SH2_struct* sh)
{
  SH2subv(sh,9,11);
}
#define init_3_9_12_11
void FASTCALL call_3_9_12_11(SH2_struct* sh)
{
  SH2subv(sh,9,12);
}
#define init_3_9_13_11
void FASTCALL call_3_9_13_11(SH2_struct* sh)
{
  SH2subv(sh,9,13);
}
#define init_3_9_14_11
void FASTCALL call_3_9_14_11(SH2_struct* sh)
{
  SH2subv(sh,9,14);
}
#define init_3_9_15_11
void FASTCALL call_3_9_15_11(SH2_struct* sh)
{
  SH2subv(sh,9,15);
}
#define init_3_10_0_11
void FASTCALL call_3_10_0_11(SH2_struct* sh)
{
  SH2subv(sh,10,0);
}
#define init_3_10_1_11
void FASTCALL call_3_10_1_11(SH2_struct* sh)
{
  SH2subv(sh,10,1);
}
#define init_3_10_2_11
void FASTCALL call_3_10_2_11(SH2_struct* sh)
{
  SH2subv(sh,10,2);
}
#define init_3_10_3_11
void FASTCALL call_3_10_3_11(SH2_struct* sh)
{
  SH2subv(sh,10,3);
}
#define init_3_10_4_11
void FASTCALL call_3_10_4_11(SH2_struct* sh)
{
  SH2subv(sh,10,4);
}
#define init_3_10_5_11
void FASTCALL call_3_10_5_11(SH2_struct* sh)
{
  SH2subv(sh,10,5);
}
#define init_3_10_6_11
void FASTCALL call_3_10_6_11(SH2_struct* sh)
{
  SH2subv(sh,10,6);
}
#define init_3_10_7_11
void FASTCALL call_3_10_7_11(SH2_struct* sh)
{
  SH2subv(sh,10,7);
}
#define init_3_10_8_11
void FASTCALL call_3_10_8_11(SH2_struct* sh)
{
  SH2subv(sh,10,8);
}
#define init_3_10_9_11
void FASTCALL call_3_10_9_11(SH2_struct* sh)
{
  SH2subv(sh,10,9);
}
#define init_3_10_10_11
void FASTCALL call_3_10_10_11(SH2_struct* sh)
{
  SH2subv(sh,10,10);
}
#define init_3_10_11_11
void FASTCALL call_3_10_11_11(SH2_struct* sh)
{
  SH2subv(sh,10,11);
}
#define init_3_10_12_11
void FASTCALL call_3_10_12_11(SH2_struct* sh)
{
  SH2subv(sh,10,12);
}
#define init_3_10_13_11
void FASTCALL call_3_10_13_11(SH2_struct* sh)
{
  SH2subv(sh,10,13);
}
#define init_3_10_14_11
void FASTCALL call_3_10_14_11(SH2_struct* sh)
{
  SH2subv(sh,10,14);
}
#define init_3_10_15_11
void FASTCALL call_3_10_15_11(SH2_struct* sh)
{
  SH2subv(sh,10,15);
}
#define init_3_11_0_11
void FASTCALL call_3_11_0_11(SH2_struct* sh)
{
  SH2subv(sh,11,0);
}
#define init_3_11_1_11
void FASTCALL call_3_11_1_11(SH2_struct* sh)
{
  SH2subv(sh,11,1);
}
#define init_3_11_2_11
void FASTCALL call_3_11_2_11(SH2_struct* sh)
{
  SH2subv(sh,11,2);
}
#define init_3_11_3_11
void FASTCALL call_3_11_3_11(SH2_struct* sh)
{
  SH2subv(sh,11,3);
}
#define init_3_11_4_11
void FASTCALL call_3_11_4_11(SH2_struct* sh)
{
  SH2subv(sh,11,4);
}
#define init_3_11_5_11
void FASTCALL call_3_11_5_11(SH2_struct* sh)
{
  SH2subv(sh,11,5);
}
#define init_3_11_6_11
void FASTCALL call_3_11_6_11(SH2_struct* sh)
{
  SH2subv(sh,11,6);
}
#define init_3_11_7_11
void FASTCALL call_3_11_7_11(SH2_struct* sh)
{
  SH2subv(sh,11,7);
}
#define init_3_11_8_11
void FASTCALL call_3_11_8_11(SH2_struct* sh)
{
  SH2subv(sh,11,8);
}
#define init_3_11_9_11
void FASTCALL call_3_11_9_11(SH2_struct* sh)
{
  SH2subv(sh,11,9);
}
#define init_3_11_10_11
void FASTCALL call_3_11_10_11(SH2_struct* sh)
{
  SH2subv(sh,11,10);
}
#define init_3_11_11_11
void FASTCALL call_3_11_11_11(SH2_struct* sh)
{
  SH2subv(sh,11,11);
}
#define init_3_11_12_11
void FASTCALL call_3_11_12_11(SH2_struct* sh)
{
  SH2subv(sh,11,12);
}
#define init_3_11_13_11
void FASTCALL call_3_11_13_11(SH2_struct* sh)
{
  SH2subv(sh,11,13);
}
#define init_3_11_14_11
void FASTCALL call_3_11_14_11(SH2_struct* sh)
{
  SH2subv(sh,11,14);
}
#define init_3_11_15_11
void FASTCALL call_3_11_15_11(SH2_struct* sh)
{
  SH2subv(sh,11,15);
}
#define init_3_12_0_11
void FASTCALL call_3_12_0_11(SH2_struct* sh)
{
  SH2subv(sh,12,0);
}
#define init_3_12_1_11
void FASTCALL call_3_12_1_11(SH2_struct* sh)
{
  SH2subv(sh,12,1);
}
#define init_3_12_2_11
void FASTCALL call_3_12_2_11(SH2_struct* sh)
{
  SH2subv(sh,12,2);
}
#define init_3_12_3_11
void FASTCALL call_3_12_3_11(SH2_struct* sh)
{
  SH2subv(sh,12,3);
}
#define init_3_12_4_11
void FASTCALL call_3_12_4_11(SH2_struct* sh)
{
  SH2subv(sh,12,4);
}
#define init_3_12_5_11
void FASTCALL call_3_12_5_11(SH2_struct* sh)
{
  SH2subv(sh,12,5);
}
#define init_3_12_6_11
void FASTCALL call_3_12_6_11(SH2_struct* sh)
{
  SH2subv(sh,12,6);
}
#define init_3_12_7_11
void FASTCALL call_3_12_7_11(SH2_struct* sh)
{
  SH2subv(sh,12,7);
}
#define init_3_12_8_11
void FASTCALL call_3_12_8_11(SH2_struct* sh)
{
  SH2subv(sh,12,8);
}
#define init_3_12_9_11
void FASTCALL call_3_12_9_11(SH2_struct* sh)
{
  SH2subv(sh,12,9);
}
#define init_3_12_10_11
void FASTCALL call_3_12_10_11(SH2_struct* sh)
{
  SH2subv(sh,12,10);
}
#define init_3_12_11_11
void FASTCALL call_3_12_11_11(SH2_struct* sh)
{
  SH2subv(sh,12,11);
}
#define init_3_12_12_11
void FASTCALL call_3_12_12_11(SH2_struct* sh)
{
  SH2subv(sh,12,12);
}
#define init_3_12_13_11
void FASTCALL call_3_12_13_11(SH2_struct* sh)
{
  SH2subv(sh,12,13);
}
#define init_3_12_14_11
void FASTCALL call_3_12_14_11(SH2_struct* sh)
{
  SH2subv(sh,12,14);
}
#define init_3_12_15_11
void FASTCALL call_3_12_15_11(SH2_struct* sh)
{
  SH2subv(sh,12,15);
}
#define init_3_13_0_11
void FASTCALL call_3_13_0_11(SH2_struct* sh)
{
  SH2subv(sh,13,0);
}
#define init_3_13_1_11
void FASTCALL call_3_13_1_11(SH2_struct* sh)
{
  SH2subv(sh,13,1);
}
#define init_3_13_2_11
void FASTCALL call_3_13_2_11(SH2_struct* sh)
{
  SH2subv(sh,13,2);
}
#define init_3_13_3_11
void FASTCALL call_3_13_3_11(SH2_struct* sh)
{
  SH2subv(sh,13,3);
}
#define init_3_13_4_11
void FASTCALL call_3_13_4_11(SH2_struct* sh)
{
  SH2subv(sh,13,4);
}
#define init_3_13_5_11
void FASTCALL call_3_13_5_11(SH2_struct* sh)
{
  SH2subv(sh,13,5);
}
#define init_3_13_6_11
void FASTCALL call_3_13_6_11(SH2_struct* sh)
{
  SH2subv(sh,13,6);
}
#define init_3_13_7_11
void FASTCALL call_3_13_7_11(SH2_struct* sh)
{
  SH2subv(sh,13,7);
}
#define init_3_13_8_11
void FASTCALL call_3_13_8_11(SH2_struct* sh)
{
  SH2subv(sh,13,8);
}
#define init_3_13_9_11
void FASTCALL call_3_13_9_11(SH2_struct* sh)
{
  SH2subv(sh,13,9);
}
#define init_3_13_10_11
void FASTCALL call_3_13_10_11(SH2_struct* sh)
{
  SH2subv(sh,13,10);
}
#define init_3_13_11_11
void FASTCALL call_3_13_11_11(SH2_struct* sh)
{
  SH2subv(sh,13,11);
}
#define init_3_13_12_11
void FASTCALL call_3_13_12_11(SH2_struct* sh)
{
  SH2subv(sh,13,12);
}
#define init_3_13_13_11
void FASTCALL call_3_13_13_11(SH2_struct* sh)
{
  SH2subv(sh,13,13);
}
#define init_3_13_14_11
void FASTCALL call_3_13_14_11(SH2_struct* sh)
{
  SH2subv(sh,13,14);
}
#define init_3_13_15_11
void FASTCALL call_3_13_15_11(SH2_struct* sh)
{
  SH2subv(sh,13,15);
}
#define init_3_14_0_11
void FASTCALL call_3_14_0_11(SH2_struct* sh)
{
  SH2subv(sh,14,0);
}
#define init_3_14_1_11
void FASTCALL call_3_14_1_11(SH2_struct* sh)
{
  SH2subv(sh,14,1);
}
#define init_3_14_2_11
void FASTCALL call_3_14_2_11(SH2_struct* sh)
{
  SH2subv(sh,14,2);
}
#define init_3_14_3_11
void FASTCALL call_3_14_3_11(SH2_struct* sh)
{
  SH2subv(sh,14,3);
}
#define init_3_14_4_11
void FASTCALL call_3_14_4_11(SH2_struct* sh)
{
  SH2subv(sh,14,4);
}
#define init_3_14_5_11
void FASTCALL call_3_14_5_11(SH2_struct* sh)
{
  SH2subv(sh,14,5);
}
#define init_3_14_6_11
void FASTCALL call_3_14_6_11(SH2_struct* sh)
{
  SH2subv(sh,14,6);
}
#define init_3_14_7_11
void FASTCALL call_3_14_7_11(SH2_struct* sh)
{
  SH2subv(sh,14,7);
}
#define init_3_14_8_11
void FASTCALL call_3_14_8_11(SH2_struct* sh)
{
  SH2subv(sh,14,8);
}
#define init_3_14_9_11
void FASTCALL call_3_14_9_11(SH2_struct* sh)
{
  SH2subv(sh,14,9);
}
#define init_3_14_10_11
void FASTCALL call_3_14_10_11(SH2_struct* sh)
{
  SH2subv(sh,14,10);
}
#define init_3_14_11_11
void FASTCALL call_3_14_11_11(SH2_struct* sh)
{
  SH2subv(sh,14,11);
}
#define init_3_14_12_11
void FASTCALL call_3_14_12_11(SH2_struct* sh)
{
  SH2subv(sh,14,12);
}
#define init_3_14_13_11
void FASTCALL call_3_14_13_11(SH2_struct* sh)
{
  SH2subv(sh,14,13);
}
#define init_3_14_14_11
void FASTCALL call_3_14_14_11(SH2_struct* sh)
{
  SH2subv(sh,14,14);
}
#define init_3_14_15_11
void FASTCALL call_3_14_15_11(SH2_struct* sh)
{
  SH2subv(sh,14,15);
}
#define init_3_15_0_11
void FASTCALL call_3_15_0_11(SH2_struct* sh)
{
  SH2subv(sh,15,0);
}
#define init_3_15_1_11
void FASTCALL call_3_15_1_11(SH2_struct* sh)
{
  SH2subv(sh,15,1);
}
#define init_3_15_2_11
void FASTCALL call_3_15_2_11(SH2_struct* sh)
{
  SH2subv(sh,15,2);
}
#define init_3_15_3_11
void FASTCALL call_3_15_3_11(SH2_struct* sh)
{
  SH2subv(sh,15,3);
}
#define init_3_15_4_11
void FASTCALL call_3_15_4_11(SH2_struct* sh)
{
  SH2subv(sh,15,4);
}
#define init_3_15_5_11
void FASTCALL call_3_15_5_11(SH2_struct* sh)
{
  SH2subv(sh,15,5);
}
#define init_3_15_6_11
void FASTCALL call_3_15_6_11(SH2_struct* sh)
{
  SH2subv(sh,15,6);
}
#define init_3_15_7_11
void FASTCALL call_3_15_7_11(SH2_struct* sh)
{
  SH2subv(sh,15,7);
}
#define init_3_15_8_11
void FASTCALL call_3_15_8_11(SH2_struct* sh)
{
  SH2subv(sh,15,8);
}
#define init_3_15_9_11
void FASTCALL call_3_15_9_11(SH2_struct* sh)
{
  SH2subv(sh,15,9);
}
#define init_3_15_10_11
void FASTCALL call_3_15_10_11(SH2_struct* sh)
{
  SH2subv(sh,15,10);
}
#define init_3_15_11_11
void FASTCALL call_3_15_11_11(SH2_struct* sh)
{
  SH2subv(sh,15,11);
}
#define init_3_15_12_11
void FASTCALL call_3_15_12_11(SH2_struct* sh)
{
  SH2subv(sh,15,12);
}
#define init_3_15_13_11
void FASTCALL call_3_15_13_11(SH2_struct* sh)
{
  SH2subv(sh,15,13);
}
#define init_3_15_14_11
void FASTCALL call_3_15_14_11(SH2_struct* sh)
{
  SH2subv(sh,15,14);
}
#define init_3_15_15_11
void FASTCALL call_3_15_15_11(SH2_struct* sh)
{
  SH2subv(sh,15,15);
}
#define init_3_0_0_10
void FASTCALL call_3_0_0_10(SH2_struct* sh)
{
  SH2subc(sh,0,0);
}
#define init_3_0_1_10
void FASTCALL call_3_0_1_10(SH2_struct* sh)
{
  SH2subc(sh,0,1);
}
#define init_3_0_2_10
void FASTCALL call_3_0_2_10(SH2_struct* sh)
{
  SH2subc(sh,0,2);
}
#define init_3_0_3_10
void FASTCALL call_3_0_3_10(SH2_struct* sh)
{
  SH2subc(sh,0,3);
}
#define init_3_0_4_10
void FASTCALL call_3_0_4_10(SH2_struct* sh)
{
  SH2subc(sh,0,4);
}
#define init_3_0_5_10
void FASTCALL call_3_0_5_10(SH2_struct* sh)
{
  SH2subc(sh,0,5);
}
#define init_3_0_6_10
void FASTCALL call_3_0_6_10(SH2_struct* sh)
{
  SH2subc(sh,0,6);
}
#define init_3_0_7_10
void FASTCALL call_3_0_7_10(SH2_struct* sh)
{
  SH2subc(sh,0,7);
}
#define init_3_0_8_10
void FASTCALL call_3_0_8_10(SH2_struct* sh)
{
  SH2subc(sh,0,8);
}
#define init_3_0_9_10
void FASTCALL call_3_0_9_10(SH2_struct* sh)
{
  SH2subc(sh,0,9);
}
#define init_3_0_10_10
void FASTCALL call_3_0_10_10(SH2_struct* sh)
{
  SH2subc(sh,0,10);
}
#define init_3_0_11_10
void FASTCALL call_3_0_11_10(SH2_struct* sh)
{
  SH2subc(sh,0,11);
}
#define init_3_0_12_10
void FASTCALL call_3_0_12_10(SH2_struct* sh)
{
  SH2subc(sh,0,12);
}
#define init_3_0_13_10
void FASTCALL call_3_0_13_10(SH2_struct* sh)
{
  SH2subc(sh,0,13);
}
#define init_3_0_14_10
void FASTCALL call_3_0_14_10(SH2_struct* sh)
{
  SH2subc(sh,0,14);
}
#define init_3_0_15_10
void FASTCALL call_3_0_15_10(SH2_struct* sh)
{
  SH2subc(sh,0,15);
}
#define init_3_1_0_10
void FASTCALL call_3_1_0_10(SH2_struct* sh)
{
  SH2subc(sh,1,0);
}
#define init_3_1_1_10
void FASTCALL call_3_1_1_10(SH2_struct* sh)
{
  SH2subc(sh,1,1);
}
#define init_3_1_2_10
void FASTCALL call_3_1_2_10(SH2_struct* sh)
{
  SH2subc(sh,1,2);
}
#define init_3_1_3_10
void FASTCALL call_3_1_3_10(SH2_struct* sh)
{
  SH2subc(sh,1,3);
}
#define init_3_1_4_10
void FASTCALL call_3_1_4_10(SH2_struct* sh)
{
  SH2subc(sh,1,4);
}
#define init_3_1_5_10
void FASTCALL call_3_1_5_10(SH2_struct* sh)
{
  SH2subc(sh,1,5);
}
#define init_3_1_6_10
void FASTCALL call_3_1_6_10(SH2_struct* sh)
{
  SH2subc(sh,1,6);
}
#define init_3_1_7_10
void FASTCALL call_3_1_7_10(SH2_struct* sh)
{
  SH2subc(sh,1,7);
}
#define init_3_1_8_10
void FASTCALL call_3_1_8_10(SH2_struct* sh)
{
  SH2subc(sh,1,8);
}
#define init_3_1_9_10
void FASTCALL call_3_1_9_10(SH2_struct* sh)
{
  SH2subc(sh,1,9);
}
#define init_3_1_10_10
void FASTCALL call_3_1_10_10(SH2_struct* sh)
{
  SH2subc(sh,1,10);
}
#define init_3_1_11_10
void FASTCALL call_3_1_11_10(SH2_struct* sh)
{
  SH2subc(sh,1,11);
}
#define init_3_1_12_10
void FASTCALL call_3_1_12_10(SH2_struct* sh)
{
  SH2subc(sh,1,12);
}
#define init_3_1_13_10
void FASTCALL call_3_1_13_10(SH2_struct* sh)
{
  SH2subc(sh,1,13);
}
#define init_3_1_14_10
void FASTCALL call_3_1_14_10(SH2_struct* sh)
{
  SH2subc(sh,1,14);
}
#define init_3_1_15_10
void FASTCALL call_3_1_15_10(SH2_struct* sh)
{
  SH2subc(sh,1,15);
}
#define init_3_2_0_10
void FASTCALL call_3_2_0_10(SH2_struct* sh)
{
  SH2subc(sh,2,0);
}
#define init_3_2_1_10
void FASTCALL call_3_2_1_10(SH2_struct* sh)
{
  SH2subc(sh,2,1);
}
#define init_3_2_2_10
void FASTCALL call_3_2_2_10(SH2_struct* sh)
{
  SH2subc(sh,2,2);
}
#define init_3_2_3_10
void FASTCALL call_3_2_3_10(SH2_struct* sh)
{
  SH2subc(sh,2,3);
}
#define init_3_2_4_10
void FASTCALL call_3_2_4_10(SH2_struct* sh)
{
  SH2subc(sh,2,4);
}
#define init_3_2_5_10
void FASTCALL call_3_2_5_10(SH2_struct* sh)
{
  SH2subc(sh,2,5);
}
#define init_3_2_6_10
void FASTCALL call_3_2_6_10(SH2_struct* sh)
{
  SH2subc(sh,2,6);
}
#define init_3_2_7_10
void FASTCALL call_3_2_7_10(SH2_struct* sh)
{
  SH2subc(sh,2,7);
}
#define init_3_2_8_10
void FASTCALL call_3_2_8_10(SH2_struct* sh)
{
  SH2subc(sh,2,8);
}
#define init_3_2_9_10
void FASTCALL call_3_2_9_10(SH2_struct* sh)
{
  SH2subc(sh,2,9);
}
#define init_3_2_10_10
void FASTCALL call_3_2_10_10(SH2_struct* sh)
{
  SH2subc(sh,2,10);
}
#define init_3_2_11_10
void FASTCALL call_3_2_11_10(SH2_struct* sh)
{
  SH2subc(sh,2,11);
}
#define init_3_2_12_10
void FASTCALL call_3_2_12_10(SH2_struct* sh)
{
  SH2subc(sh,2,12);
}
#define init_3_2_13_10
void FASTCALL call_3_2_13_10(SH2_struct* sh)
{
  SH2subc(sh,2,13);
}
#define init_3_2_14_10
void FASTCALL call_3_2_14_10(SH2_struct* sh)
{
  SH2subc(sh,2,14);
}
#define init_3_2_15_10
void FASTCALL call_3_2_15_10(SH2_struct* sh)
{
  SH2subc(sh,2,15);
}
#define init_3_3_0_10
void FASTCALL call_3_3_0_10(SH2_struct* sh)
{
  SH2subc(sh,3,0);
}
#define init_3_3_1_10
void FASTCALL call_3_3_1_10(SH2_struct* sh)
{
  SH2subc(sh,3,1);
}
#define init_3_3_2_10
void FASTCALL call_3_3_2_10(SH2_struct* sh)
{
  SH2subc(sh,3,2);
}
#define init_3_3_3_10
void FASTCALL call_3_3_3_10(SH2_struct* sh)
{
  SH2subc(sh,3,3);
}
#define init_3_3_4_10
void FASTCALL call_3_3_4_10(SH2_struct* sh)
{
  SH2subc(sh,3,4);
}
#define init_3_3_5_10
void FASTCALL call_3_3_5_10(SH2_struct* sh)
{
  SH2subc(sh,3,5);
}
#define init_3_3_6_10
void FASTCALL call_3_3_6_10(SH2_struct* sh)
{
  SH2subc(sh,3,6);
}
#define init_3_3_7_10
void FASTCALL call_3_3_7_10(SH2_struct* sh)
{
  SH2subc(sh,3,7);
}
#define init_3_3_8_10
void FASTCALL call_3_3_8_10(SH2_struct* sh)
{
  SH2subc(sh,3,8);
}
#define init_3_3_9_10
void FASTCALL call_3_3_9_10(SH2_struct* sh)
{
  SH2subc(sh,3,9);
}
#define init_3_3_10_10
void FASTCALL call_3_3_10_10(SH2_struct* sh)
{
  SH2subc(sh,3,10);
}
#define init_3_3_11_10
void FASTCALL call_3_3_11_10(SH2_struct* sh)
{
  SH2subc(sh,3,11);
}
#define init_3_3_12_10
void FASTCALL call_3_3_12_10(SH2_struct* sh)
{
  SH2subc(sh,3,12);
}
#define init_3_3_13_10
void FASTCALL call_3_3_13_10(SH2_struct* sh)
{
  SH2subc(sh,3,13);
}
#define init_3_3_14_10
void FASTCALL call_3_3_14_10(SH2_struct* sh)
{
  SH2subc(sh,3,14);
}
#define init_3_3_15_10
void FASTCALL call_3_3_15_10(SH2_struct* sh)
{
  SH2subc(sh,3,15);
}
#define init_3_4_0_10
void FASTCALL call_3_4_0_10(SH2_struct* sh)
{
  SH2subc(sh,4,0);
}
#define init_3_4_1_10
void FASTCALL call_3_4_1_10(SH2_struct* sh)
{
  SH2subc(sh,4,1);
}
#define init_3_4_2_10
void FASTCALL call_3_4_2_10(SH2_struct* sh)
{
  SH2subc(sh,4,2);
}
#define init_3_4_3_10
void FASTCALL call_3_4_3_10(SH2_struct* sh)
{
  SH2subc(sh,4,3);
}
#define init_3_4_4_10
void FASTCALL call_3_4_4_10(SH2_struct* sh)
{
  SH2subc(sh,4,4);
}
#define init_3_4_5_10
void FASTCALL call_3_4_5_10(SH2_struct* sh)
{
  SH2subc(sh,4,5);
}
#define init_3_4_6_10
void FASTCALL call_3_4_6_10(SH2_struct* sh)
{
  SH2subc(sh,4,6);
}
#define init_3_4_7_10
void FASTCALL call_3_4_7_10(SH2_struct* sh)
{
  SH2subc(sh,4,7);
}
#define init_3_4_8_10
void FASTCALL call_3_4_8_10(SH2_struct* sh)
{
  SH2subc(sh,4,8);
}
#define init_3_4_9_10
void FASTCALL call_3_4_9_10(SH2_struct* sh)
{
  SH2subc(sh,4,9);
}
#define init_3_4_10_10
void FASTCALL call_3_4_10_10(SH2_struct* sh)
{
  SH2subc(sh,4,10);
}
#define init_3_4_11_10
void FASTCALL call_3_4_11_10(SH2_struct* sh)
{
  SH2subc(sh,4,11);
}
#define init_3_4_12_10
void FASTCALL call_3_4_12_10(SH2_struct* sh)
{
  SH2subc(sh,4,12);
}
#define init_3_4_13_10
void FASTCALL call_3_4_13_10(SH2_struct* sh)
{
  SH2subc(sh,4,13);
}
#define init_3_4_14_10
void FASTCALL call_3_4_14_10(SH2_struct* sh)
{
  SH2subc(sh,4,14);
}
#define init_3_4_15_10
void FASTCALL call_3_4_15_10(SH2_struct* sh)
{
  SH2subc(sh,4,15);
}
#define init_3_5_0_10
void FASTCALL call_3_5_0_10(SH2_struct* sh)
{
  SH2subc(sh,5,0);
}
#define init_3_5_1_10
void FASTCALL call_3_5_1_10(SH2_struct* sh)
{
  SH2subc(sh,5,1);
}
#define init_3_5_2_10
void FASTCALL call_3_5_2_10(SH2_struct* sh)
{
  SH2subc(sh,5,2);
}
#define init_3_5_3_10
void FASTCALL call_3_5_3_10(SH2_struct* sh)
{
  SH2subc(sh,5,3);
}
#define init_3_5_4_10
void FASTCALL call_3_5_4_10(SH2_struct* sh)
{
  SH2subc(sh,5,4);
}
#define init_3_5_5_10
void FASTCALL call_3_5_5_10(SH2_struct* sh)
{
  SH2subc(sh,5,5);
}
#define init_3_5_6_10
void FASTCALL call_3_5_6_10(SH2_struct* sh)
{
  SH2subc(sh,5,6);
}
#define init_3_5_7_10
void FASTCALL call_3_5_7_10(SH2_struct* sh)
{
  SH2subc(sh,5,7);
}
#define init_3_5_8_10
void FASTCALL call_3_5_8_10(SH2_struct* sh)
{
  SH2subc(sh,5,8);
}
#define init_3_5_9_10
void FASTCALL call_3_5_9_10(SH2_struct* sh)
{
  SH2subc(sh,5,9);
}
#define init_3_5_10_10
void FASTCALL call_3_5_10_10(SH2_struct* sh)
{
  SH2subc(sh,5,10);
}
#define init_3_5_11_10
void FASTCALL call_3_5_11_10(SH2_struct* sh)
{
  SH2subc(sh,5,11);
}
#define init_3_5_12_10
void FASTCALL call_3_5_12_10(SH2_struct* sh)
{
  SH2subc(sh,5,12);
}
#define init_3_5_13_10
void FASTCALL call_3_5_13_10(SH2_struct* sh)
{
  SH2subc(sh,5,13);
}
#define init_3_5_14_10
void FASTCALL call_3_5_14_10(SH2_struct* sh)
{
  SH2subc(sh,5,14);
}
#define init_3_5_15_10
void FASTCALL call_3_5_15_10(SH2_struct* sh)
{
  SH2subc(sh,5,15);
}
#define init_3_6_0_10
void FASTCALL call_3_6_0_10(SH2_struct* sh)
{
  SH2subc(sh,6,0);
}
#define init_3_6_1_10
void FASTCALL call_3_6_1_10(SH2_struct* sh)
{
  SH2subc(sh,6,1);
}
#define init_3_6_2_10
void FASTCALL call_3_6_2_10(SH2_struct* sh)
{
  SH2subc(sh,6,2);
}
#define init_3_6_3_10
void FASTCALL call_3_6_3_10(SH2_struct* sh)
{
  SH2subc(sh,6,3);
}
#define init_3_6_4_10
void FASTCALL call_3_6_4_10(SH2_struct* sh)
{
  SH2subc(sh,6,4);
}
#define init_3_6_5_10
void FASTCALL call_3_6_5_10(SH2_struct* sh)
{
  SH2subc(sh,6,5);
}
#define init_3_6_6_10
void FASTCALL call_3_6_6_10(SH2_struct* sh)
{
  SH2subc(sh,6,6);
}
#define init_3_6_7_10
void FASTCALL call_3_6_7_10(SH2_struct* sh)
{
  SH2subc(sh,6,7);
}
#define init_3_6_8_10
void FASTCALL call_3_6_8_10(SH2_struct* sh)
{
  SH2subc(sh,6,8);
}
#define init_3_6_9_10
void FASTCALL call_3_6_9_10(SH2_struct* sh)
{
  SH2subc(sh,6,9);
}
#define init_3_6_10_10
void FASTCALL call_3_6_10_10(SH2_struct* sh)
{
  SH2subc(sh,6,10);
}
#define init_3_6_11_10
void FASTCALL call_3_6_11_10(SH2_struct* sh)
{
  SH2subc(sh,6,11);
}
#define init_3_6_12_10
void FASTCALL call_3_6_12_10(SH2_struct* sh)
{
  SH2subc(sh,6,12);
}
#define init_3_6_13_10
void FASTCALL call_3_6_13_10(SH2_struct* sh)
{
  SH2subc(sh,6,13);
}
#define init_3_6_14_10
void FASTCALL call_3_6_14_10(SH2_struct* sh)
{
  SH2subc(sh,6,14);
}
#define init_3_6_15_10
void FASTCALL call_3_6_15_10(SH2_struct* sh)
{
  SH2subc(sh,6,15);
}
#define init_3_7_0_10
void FASTCALL call_3_7_0_10(SH2_struct* sh)
{
  SH2subc(sh,7,0);
}
#define init_3_7_1_10
void FASTCALL call_3_7_1_10(SH2_struct* sh)
{
  SH2subc(sh,7,1);
}
#define init_3_7_2_10
void FASTCALL call_3_7_2_10(SH2_struct* sh)
{
  SH2subc(sh,7,2);
}
#define init_3_7_3_10
void FASTCALL call_3_7_3_10(SH2_struct* sh)
{
  SH2subc(sh,7,3);
}
#define init_3_7_4_10
void FASTCALL call_3_7_4_10(SH2_struct* sh)
{
  SH2subc(sh,7,4);
}
#define init_3_7_5_10
void FASTCALL call_3_7_5_10(SH2_struct* sh)
{
  SH2subc(sh,7,5);
}
#define init_3_7_6_10
void FASTCALL call_3_7_6_10(SH2_struct* sh)
{
  SH2subc(sh,7,6);
}
#define init_3_7_7_10
void FASTCALL call_3_7_7_10(SH2_struct* sh)
{
  SH2subc(sh,7,7);
}
#define init_3_7_8_10
void FASTCALL call_3_7_8_10(SH2_struct* sh)
{
  SH2subc(sh,7,8);
}
#define init_3_7_9_10
void FASTCALL call_3_7_9_10(SH2_struct* sh)
{
  SH2subc(sh,7,9);
}
#define init_3_7_10_10
void FASTCALL call_3_7_10_10(SH2_struct* sh)
{
  SH2subc(sh,7,10);
}
#define init_3_7_11_10
void FASTCALL call_3_7_11_10(SH2_struct* sh)
{
  SH2subc(sh,7,11);
}
#define init_3_7_12_10
void FASTCALL call_3_7_12_10(SH2_struct* sh)
{
  SH2subc(sh,7,12);
}
#define init_3_7_13_10
void FASTCALL call_3_7_13_10(SH2_struct* sh)
{
  SH2subc(sh,7,13);
}
#define init_3_7_14_10
void FASTCALL call_3_7_14_10(SH2_struct* sh)
{
  SH2subc(sh,7,14);
}
#define init_3_7_15_10
void FASTCALL call_3_7_15_10(SH2_struct* sh)
{
  SH2subc(sh,7,15);
}
#define init_3_8_0_10
void FASTCALL call_3_8_0_10(SH2_struct* sh)
{
  SH2subc(sh,8,0);
}
#define init_3_8_1_10
void FASTCALL call_3_8_1_10(SH2_struct* sh)
{
  SH2subc(sh,8,1);
}
#define init_3_8_2_10
void FASTCALL call_3_8_2_10(SH2_struct* sh)
{
  SH2subc(sh,8,2);
}
#define init_3_8_3_10
void FASTCALL call_3_8_3_10(SH2_struct* sh)
{
  SH2subc(sh,8,3);
}
#define init_3_8_4_10
void FASTCALL call_3_8_4_10(SH2_struct* sh)
{
  SH2subc(sh,8,4);
}
#define init_3_8_5_10
void FASTCALL call_3_8_5_10(SH2_struct* sh)
{
  SH2subc(sh,8,5);
}
#define init_3_8_6_10
void FASTCALL call_3_8_6_10(SH2_struct* sh)
{
  SH2subc(sh,8,6);
}
#define init_3_8_7_10
void FASTCALL call_3_8_7_10(SH2_struct* sh)
{
  SH2subc(sh,8,7);
}
#define init_3_8_8_10
void FASTCALL call_3_8_8_10(SH2_struct* sh)
{
  SH2subc(sh,8,8);
}
#define init_3_8_9_10
void FASTCALL call_3_8_9_10(SH2_struct* sh)
{
  SH2subc(sh,8,9);
}
#define init_3_8_10_10
void FASTCALL call_3_8_10_10(SH2_struct* sh)
{
  SH2subc(sh,8,10);
}
#define init_3_8_11_10
void FASTCALL call_3_8_11_10(SH2_struct* sh)
{
  SH2subc(sh,8,11);
}
#define init_3_8_12_10
void FASTCALL call_3_8_12_10(SH2_struct* sh)
{
  SH2subc(sh,8,12);
}
#define init_3_8_13_10
void FASTCALL call_3_8_13_10(SH2_struct* sh)
{
  SH2subc(sh,8,13);
}
#define init_3_8_14_10
void FASTCALL call_3_8_14_10(SH2_struct* sh)
{
  SH2subc(sh,8,14);
}
#define init_3_8_15_10
void FASTCALL call_3_8_15_10(SH2_struct* sh)
{
  SH2subc(sh,8,15);
}
#define init_3_9_0_10
void FASTCALL call_3_9_0_10(SH2_struct* sh)
{
  SH2subc(sh,9,0);
}
#define init_3_9_1_10
void FASTCALL call_3_9_1_10(SH2_struct* sh)
{
  SH2subc(sh,9,1);
}
#define init_3_9_2_10
void FASTCALL call_3_9_2_10(SH2_struct* sh)
{
  SH2subc(sh,9,2);
}
#define init_3_9_3_10
void FASTCALL call_3_9_3_10(SH2_struct* sh)
{
  SH2subc(sh,9,3);
}
#define init_3_9_4_10
void FASTCALL call_3_9_4_10(SH2_struct* sh)
{
  SH2subc(sh,9,4);
}
#define init_3_9_5_10
void FASTCALL call_3_9_5_10(SH2_struct* sh)
{
  SH2subc(sh,9,5);
}
#define init_3_9_6_10
void FASTCALL call_3_9_6_10(SH2_struct* sh)
{
  SH2subc(sh,9,6);
}
#define init_3_9_7_10
void FASTCALL call_3_9_7_10(SH2_struct* sh)
{
  SH2subc(sh,9,7);
}
#define init_3_9_8_10
void FASTCALL call_3_9_8_10(SH2_struct* sh)
{
  SH2subc(sh,9,8);
}
#define init_3_9_9_10
void FASTCALL call_3_9_9_10(SH2_struct* sh)
{
  SH2subc(sh,9,9);
}
#define init_3_9_10_10
void FASTCALL call_3_9_10_10(SH2_struct* sh)
{
  SH2subc(sh,9,10);
}
#define init_3_9_11_10
void FASTCALL call_3_9_11_10(SH2_struct* sh)
{
  SH2subc(sh,9,11);
}
#define init_3_9_12_10
void FASTCALL call_3_9_12_10(SH2_struct* sh)
{
  SH2subc(sh,9,12);
}
#define init_3_9_13_10
void FASTCALL call_3_9_13_10(SH2_struct* sh)
{
  SH2subc(sh,9,13);
}
#define init_3_9_14_10
void FASTCALL call_3_9_14_10(SH2_struct* sh)
{
  SH2subc(sh,9,14);
}
#define init_3_9_15_10
void FASTCALL call_3_9_15_10(SH2_struct* sh)
{
  SH2subc(sh,9,15);
}
#define init_3_10_0_10
void FASTCALL call_3_10_0_10(SH2_struct* sh)
{
  SH2subc(sh,10,0);
}
#define init_3_10_1_10
void FASTCALL call_3_10_1_10(SH2_struct* sh)
{
  SH2subc(sh,10,1);
}
#define init_3_10_2_10
void FASTCALL call_3_10_2_10(SH2_struct* sh)
{
  SH2subc(sh,10,2);
}
#define init_3_10_3_10
void FASTCALL call_3_10_3_10(SH2_struct* sh)
{
  SH2subc(sh,10,3);
}
#define init_3_10_4_10
void FASTCALL call_3_10_4_10(SH2_struct* sh)
{
  SH2subc(sh,10,4);
}
#define init_3_10_5_10
void FASTCALL call_3_10_5_10(SH2_struct* sh)
{
  SH2subc(sh,10,5);
}
#define init_3_10_6_10
void FASTCALL call_3_10_6_10(SH2_struct* sh)
{
  SH2subc(sh,10,6);
}
#define init_3_10_7_10
void FASTCALL call_3_10_7_10(SH2_struct* sh)
{
  SH2subc(sh,10,7);
}
#define init_3_10_8_10
void FASTCALL call_3_10_8_10(SH2_struct* sh)
{
  SH2subc(sh,10,8);
}
#define init_3_10_9_10
void FASTCALL call_3_10_9_10(SH2_struct* sh)
{
  SH2subc(sh,10,9);
}
#define init_3_10_10_10
void FASTCALL call_3_10_10_10(SH2_struct* sh)
{
  SH2subc(sh,10,10);
}
#define init_3_10_11_10
void FASTCALL call_3_10_11_10(SH2_struct* sh)
{
  SH2subc(sh,10,11);
}
#define init_3_10_12_10
void FASTCALL call_3_10_12_10(SH2_struct* sh)
{
  SH2subc(sh,10,12);
}
#define init_3_10_13_10
void FASTCALL call_3_10_13_10(SH2_struct* sh)
{
  SH2subc(sh,10,13);
}
#define init_3_10_14_10
void FASTCALL call_3_10_14_10(SH2_struct* sh)
{
  SH2subc(sh,10,14);
}
#define init_3_10_15_10
void FASTCALL call_3_10_15_10(SH2_struct* sh)
{
  SH2subc(sh,10,15);
}
#define init_3_11_0_10
void FASTCALL call_3_11_0_10(SH2_struct* sh)
{
  SH2subc(sh,11,0);
}
#define init_3_11_1_10
void FASTCALL call_3_11_1_10(SH2_struct* sh)
{
  SH2subc(sh,11,1);
}
#define init_3_11_2_10
void FASTCALL call_3_11_2_10(SH2_struct* sh)
{
  SH2subc(sh,11,2);
}
#define init_3_11_3_10
void FASTCALL call_3_11_3_10(SH2_struct* sh)
{
  SH2subc(sh,11,3);
}
#define init_3_11_4_10
void FASTCALL call_3_11_4_10(SH2_struct* sh)
{
  SH2subc(sh,11,4);
}
#define init_3_11_5_10
void FASTCALL call_3_11_5_10(SH2_struct* sh)
{
  SH2subc(sh,11,5);
}
#define init_3_11_6_10
void FASTCALL call_3_11_6_10(SH2_struct* sh)
{
  SH2subc(sh,11,6);
}
#define init_3_11_7_10
void FASTCALL call_3_11_7_10(SH2_struct* sh)
{
  SH2subc(sh,11,7);
}
#define init_3_11_8_10
void FASTCALL call_3_11_8_10(SH2_struct* sh)
{
  SH2subc(sh,11,8);
}
#define init_3_11_9_10
void FASTCALL call_3_11_9_10(SH2_struct* sh)
{
  SH2subc(sh,11,9);
}
#define init_3_11_10_10
void FASTCALL call_3_11_10_10(SH2_struct* sh)
{
  SH2subc(sh,11,10);
}
#define init_3_11_11_10
void FASTCALL call_3_11_11_10(SH2_struct* sh)
{
  SH2subc(sh,11,11);
}
#define init_3_11_12_10
void FASTCALL call_3_11_12_10(SH2_struct* sh)
{
  SH2subc(sh,11,12);
}
#define init_3_11_13_10
void FASTCALL call_3_11_13_10(SH2_struct* sh)
{
  SH2subc(sh,11,13);
}
#define init_3_11_14_10
void FASTCALL call_3_11_14_10(SH2_struct* sh)
{
  SH2subc(sh,11,14);
}
#define init_3_11_15_10
void FASTCALL call_3_11_15_10(SH2_struct* sh)
{
  SH2subc(sh,11,15);
}
#define init_3_12_0_10
void FASTCALL call_3_12_0_10(SH2_struct* sh)
{
  SH2subc(sh,12,0);
}
#define init_3_12_1_10
void FASTCALL call_3_12_1_10(SH2_struct* sh)
{
  SH2subc(sh,12,1);
}
#define init_3_12_2_10
void FASTCALL call_3_12_2_10(SH2_struct* sh)
{
  SH2subc(sh,12,2);
}
#define init_3_12_3_10
void FASTCALL call_3_12_3_10(SH2_struct* sh)
{
  SH2subc(sh,12,3);
}
#define init_3_12_4_10
void FASTCALL call_3_12_4_10(SH2_struct* sh)
{
  SH2subc(sh,12,4);
}
#define init_3_12_5_10
void FASTCALL call_3_12_5_10(SH2_struct* sh)
{
  SH2subc(sh,12,5);
}
#define init_3_12_6_10
void FASTCALL call_3_12_6_10(SH2_struct* sh)
{
  SH2subc(sh,12,6);
}
#define init_3_12_7_10
void FASTCALL call_3_12_7_10(SH2_struct* sh)
{
  SH2subc(sh,12,7);
}
#define init_3_12_8_10
void FASTCALL call_3_12_8_10(SH2_struct* sh)
{
  SH2subc(sh,12,8);
}
#define init_3_12_9_10
void FASTCALL call_3_12_9_10(SH2_struct* sh)
{
  SH2subc(sh,12,9);
}
#define init_3_12_10_10
void FASTCALL call_3_12_10_10(SH2_struct* sh)
{
  SH2subc(sh,12,10);
}
#define init_3_12_11_10
void FASTCALL call_3_12_11_10(SH2_struct* sh)
{
  SH2subc(sh,12,11);
}
#define init_3_12_12_10
void FASTCALL call_3_12_12_10(SH2_struct* sh)
{
  SH2subc(sh,12,12);
}
#define init_3_12_13_10
void FASTCALL call_3_12_13_10(SH2_struct* sh)
{
  SH2subc(sh,12,13);
}
#define init_3_12_14_10
void FASTCALL call_3_12_14_10(SH2_struct* sh)
{
  SH2subc(sh,12,14);
}
#define init_3_12_15_10
void FASTCALL call_3_12_15_10(SH2_struct* sh)
{
  SH2subc(sh,12,15);
}
#define init_3_13_0_10
void FASTCALL call_3_13_0_10(SH2_struct* sh)
{
  SH2subc(sh,13,0);
}
#define init_3_13_1_10
void FASTCALL call_3_13_1_10(SH2_struct* sh)
{
  SH2subc(sh,13,1);
}
#define init_3_13_2_10
void FASTCALL call_3_13_2_10(SH2_struct* sh)
{
  SH2subc(sh,13,2);
}
#define init_3_13_3_10
void FASTCALL call_3_13_3_10(SH2_struct* sh)
{
  SH2subc(sh,13,3);
}
#define init_3_13_4_10
void FASTCALL call_3_13_4_10(SH2_struct* sh)
{
  SH2subc(sh,13,4);
}
#define init_3_13_5_10
void FASTCALL call_3_13_5_10(SH2_struct* sh)
{
  SH2subc(sh,13,5);
}
#define init_3_13_6_10
void FASTCALL call_3_13_6_10(SH2_struct* sh)
{
  SH2subc(sh,13,6);
}
#define init_3_13_7_10
void FASTCALL call_3_13_7_10(SH2_struct* sh)
{
  SH2subc(sh,13,7);
}
#define init_3_13_8_10
void FASTCALL call_3_13_8_10(SH2_struct* sh)
{
  SH2subc(sh,13,8);
}
#define init_3_13_9_10
void FASTCALL call_3_13_9_10(SH2_struct* sh)
{
  SH2subc(sh,13,9);
}
#define init_3_13_10_10
void FASTCALL call_3_13_10_10(SH2_struct* sh)
{
  SH2subc(sh,13,10);
}
#define init_3_13_11_10
void FASTCALL call_3_13_11_10(SH2_struct* sh)
{
  SH2subc(sh,13,11);
}
#define init_3_13_12_10
void FASTCALL call_3_13_12_10(SH2_struct* sh)
{
  SH2subc(sh,13,12);
}
#define init_3_13_13_10
void FASTCALL call_3_13_13_10(SH2_struct* sh)
{
  SH2subc(sh,13,13);
}
#define init_3_13_14_10
void FASTCALL call_3_13_14_10(SH2_struct* sh)
{
  SH2subc(sh,13,14);
}
#define init_3_13_15_10
void FASTCALL call_3_13_15_10(SH2_struct* sh)
{
  SH2subc(sh,13,15);
}
#define init_3_14_0_10
void FASTCALL call_3_14_0_10(SH2_struct* sh)
{
  SH2subc(sh,14,0);
}
#define init_3_14_1_10
void FASTCALL call_3_14_1_10(SH2_struct* sh)
{
  SH2subc(sh,14,1);
}
#define init_3_14_2_10
void FASTCALL call_3_14_2_10(SH2_struct* sh)
{
  SH2subc(sh,14,2);
}
#define init_3_14_3_10
void FASTCALL call_3_14_3_10(SH2_struct* sh)
{
  SH2subc(sh,14,3);
}
#define init_3_14_4_10
void FASTCALL call_3_14_4_10(SH2_struct* sh)
{
  SH2subc(sh,14,4);
}
#define init_3_14_5_10
void FASTCALL call_3_14_5_10(SH2_struct* sh)
{
  SH2subc(sh,14,5);
}
#define init_3_14_6_10
void FASTCALL call_3_14_6_10(SH2_struct* sh)
{
  SH2subc(sh,14,6);
}
#define init_3_14_7_10
void FASTCALL call_3_14_7_10(SH2_struct* sh)
{
  SH2subc(sh,14,7);
}
#define init_3_14_8_10
void FASTCALL call_3_14_8_10(SH2_struct* sh)
{
  SH2subc(sh,14,8);
}
#define init_3_14_9_10
void FASTCALL call_3_14_9_10(SH2_struct* sh)
{
  SH2subc(sh,14,9);
}
#define init_3_14_10_10
void FASTCALL call_3_14_10_10(SH2_struct* sh)
{
  SH2subc(sh,14,10);
}
#define init_3_14_11_10
void FASTCALL call_3_14_11_10(SH2_struct* sh)
{
  SH2subc(sh,14,11);
}
#define init_3_14_12_10
void FASTCALL call_3_14_12_10(SH2_struct* sh)
{
  SH2subc(sh,14,12);
}
#define init_3_14_13_10
void FASTCALL call_3_14_13_10(SH2_struct* sh)
{
  SH2subc(sh,14,13);
}
#define init_3_14_14_10
void FASTCALL call_3_14_14_10(SH2_struct* sh)
{
  SH2subc(sh,14,14);
}
#define init_3_14_15_10
void FASTCALL call_3_14_15_10(SH2_struct* sh)
{
  SH2subc(sh,14,15);
}
#define init_3_15_0_10
void FASTCALL call_3_15_0_10(SH2_struct* sh)
{
  SH2subc(sh,15,0);
}
#define init_3_15_1_10
void FASTCALL call_3_15_1_10(SH2_struct* sh)
{
  SH2subc(sh,15,1);
}
#define init_3_15_2_10
void FASTCALL call_3_15_2_10(SH2_struct* sh)
{
  SH2subc(sh,15,2);
}
#define init_3_15_3_10
void FASTCALL call_3_15_3_10(SH2_struct* sh)
{
  SH2subc(sh,15,3);
}
#define init_3_15_4_10
void FASTCALL call_3_15_4_10(SH2_struct* sh)
{
  SH2subc(sh,15,4);
}
#define init_3_15_5_10
void FASTCALL call_3_15_5_10(SH2_struct* sh)
{
  SH2subc(sh,15,5);
}
#define init_3_15_6_10
void FASTCALL call_3_15_6_10(SH2_struct* sh)
{
  SH2subc(sh,15,6);
}
#define init_3_15_7_10
void FASTCALL call_3_15_7_10(SH2_struct* sh)
{
  SH2subc(sh,15,7);
}
#define init_3_15_8_10
void FASTCALL call_3_15_8_10(SH2_struct* sh)
{
  SH2subc(sh,15,8);
}
#define init_3_15_9_10
void FASTCALL call_3_15_9_10(SH2_struct* sh)
{
  SH2subc(sh,15,9);
}
#define init_3_15_10_10
void FASTCALL call_3_15_10_10(SH2_struct* sh)
{
  SH2subc(sh,15,10);
}
#define init_3_15_11_10
void FASTCALL call_3_15_11_10(SH2_struct* sh)
{
  SH2subc(sh,15,11);
}
#define init_3_15_12_10
void FASTCALL call_3_15_12_10(SH2_struct* sh)
{
  SH2subc(sh,15,12);
}
#define init_3_15_13_10
void FASTCALL call_3_15_13_10(SH2_struct* sh)
{
  SH2subc(sh,15,13);
}
#define init_3_15_14_10
void FASTCALL call_3_15_14_10(SH2_struct* sh)
{
  SH2subc(sh,15,14);
}
#define init_3_15_15_10
void FASTCALL call_3_15_15_10(SH2_struct* sh)
{
  SH2subc(sh,15,15);
}
#define init_3_0_0_8
void FASTCALL call_3_0_0_8(SH2_struct* sh)
{
  SH2sub(sh,0,0);
}
#define init_3_0_1_8
void FASTCALL call_3_0_1_8(SH2_struct* sh)
{
  SH2sub(sh,0,1);
}
#define init_3_0_2_8
void FASTCALL call_3_0_2_8(SH2_struct* sh)
{
  SH2sub(sh,0,2);
}
#define init_3_0_3_8
void FASTCALL call_3_0_3_8(SH2_struct* sh)
{
  SH2sub(sh,0,3);
}
#define init_3_0_4_8
void FASTCALL call_3_0_4_8(SH2_struct* sh)
{
  SH2sub(sh,0,4);
}
#define init_3_0_5_8
void FASTCALL call_3_0_5_8(SH2_struct* sh)
{
  SH2sub(sh,0,5);
}
#define init_3_0_6_8
void FASTCALL call_3_0_6_8(SH2_struct* sh)
{
  SH2sub(sh,0,6);
}
#define init_3_0_7_8
void FASTCALL call_3_0_7_8(SH2_struct* sh)
{
  SH2sub(sh,0,7);
}
#define init_3_0_8_8
void FASTCALL call_3_0_8_8(SH2_struct* sh)
{
  SH2sub(sh,0,8);
}
#define init_3_0_9_8
void FASTCALL call_3_0_9_8(SH2_struct* sh)
{
  SH2sub(sh,0,9);
}
#define init_3_0_10_8
void FASTCALL call_3_0_10_8(SH2_struct* sh)
{
  SH2sub(sh,0,10);
}
#define init_3_0_11_8
void FASTCALL call_3_0_11_8(SH2_struct* sh)
{
  SH2sub(sh,0,11);
}
#define init_3_0_12_8
void FASTCALL call_3_0_12_8(SH2_struct* sh)
{
  SH2sub(sh,0,12);
}
#define init_3_0_13_8
void FASTCALL call_3_0_13_8(SH2_struct* sh)
{
  SH2sub(sh,0,13);
}
#define init_3_0_14_8
void FASTCALL call_3_0_14_8(SH2_struct* sh)
{
  SH2sub(sh,0,14);
}
#define init_3_0_15_8
void FASTCALL call_3_0_15_8(SH2_struct* sh)
{
  SH2sub(sh,0,15);
}
#define init_3_1_0_8
void FASTCALL call_3_1_0_8(SH2_struct* sh)
{
  SH2sub(sh,1,0);
}
#define init_3_1_1_8
void FASTCALL call_3_1_1_8(SH2_struct* sh)
{
  SH2sub(sh,1,1);
}
#define init_3_1_2_8
void FASTCALL call_3_1_2_8(SH2_struct* sh)
{
  SH2sub(sh,1,2);
}
#define init_3_1_3_8
void FASTCALL call_3_1_3_8(SH2_struct* sh)
{
  SH2sub(sh,1,3);
}
#define init_3_1_4_8
void FASTCALL call_3_1_4_8(SH2_struct* sh)
{
  SH2sub(sh,1,4);
}
#define init_3_1_5_8
void FASTCALL call_3_1_5_8(SH2_struct* sh)
{
  SH2sub(sh,1,5);
}
#define init_3_1_6_8
void FASTCALL call_3_1_6_8(SH2_struct* sh)
{
  SH2sub(sh,1,6);
}
#define init_3_1_7_8
void FASTCALL call_3_1_7_8(SH2_struct* sh)
{
  SH2sub(sh,1,7);
}
#define init_3_1_8_8
void FASTCALL call_3_1_8_8(SH2_struct* sh)
{
  SH2sub(sh,1,8);
}
#define init_3_1_9_8
void FASTCALL call_3_1_9_8(SH2_struct* sh)
{
  SH2sub(sh,1,9);
}
#define init_3_1_10_8
void FASTCALL call_3_1_10_8(SH2_struct* sh)
{
  SH2sub(sh,1,10);
}
#define init_3_1_11_8
void FASTCALL call_3_1_11_8(SH2_struct* sh)
{
  SH2sub(sh,1,11);
}
#define init_3_1_12_8
void FASTCALL call_3_1_12_8(SH2_struct* sh)
{
  SH2sub(sh,1,12);
}
#define init_3_1_13_8
void FASTCALL call_3_1_13_8(SH2_struct* sh)
{
  SH2sub(sh,1,13);
}
#define init_3_1_14_8
void FASTCALL call_3_1_14_8(SH2_struct* sh)
{
  SH2sub(sh,1,14);
}
#define init_3_1_15_8
void FASTCALL call_3_1_15_8(SH2_struct* sh)
{
  SH2sub(sh,1,15);
}
#define init_3_2_0_8
void FASTCALL call_3_2_0_8(SH2_struct* sh)
{
  SH2sub(sh,2,0);
}
#define init_3_2_1_8
void FASTCALL call_3_2_1_8(SH2_struct* sh)
{
  SH2sub(sh,2,1);
}
#define init_3_2_2_8
void FASTCALL call_3_2_2_8(SH2_struct* sh)
{
  SH2sub(sh,2,2);
}
#define init_3_2_3_8
void FASTCALL call_3_2_3_8(SH2_struct* sh)
{
  SH2sub(sh,2,3);
}
#define init_3_2_4_8
void FASTCALL call_3_2_4_8(SH2_struct* sh)
{
  SH2sub(sh,2,4);
}
#define init_3_2_5_8
void FASTCALL call_3_2_5_8(SH2_struct* sh)
{
  SH2sub(sh,2,5);
}
#define init_3_2_6_8
void FASTCALL call_3_2_6_8(SH2_struct* sh)
{
  SH2sub(sh,2,6);
}
#define init_3_2_7_8
void FASTCALL call_3_2_7_8(SH2_struct* sh)
{
  SH2sub(sh,2,7);
}
#define init_3_2_8_8
void FASTCALL call_3_2_8_8(SH2_struct* sh)
{
  SH2sub(sh,2,8);
}
#define init_3_2_9_8
void FASTCALL call_3_2_9_8(SH2_struct* sh)
{
  SH2sub(sh,2,9);
}
#define init_3_2_10_8
void FASTCALL call_3_2_10_8(SH2_struct* sh)
{
  SH2sub(sh,2,10);
}
#define init_3_2_11_8
void FASTCALL call_3_2_11_8(SH2_struct* sh)
{
  SH2sub(sh,2,11);
}
#define init_3_2_12_8
void FASTCALL call_3_2_12_8(SH2_struct* sh)
{
  SH2sub(sh,2,12);
}
#define init_3_2_13_8
void FASTCALL call_3_2_13_8(SH2_struct* sh)
{
  SH2sub(sh,2,13);
}
#define init_3_2_14_8
void FASTCALL call_3_2_14_8(SH2_struct* sh)
{
  SH2sub(sh,2,14);
}
#define init_3_2_15_8
void FASTCALL call_3_2_15_8(SH2_struct* sh)
{
  SH2sub(sh,2,15);
}
#define init_3_3_0_8
void FASTCALL call_3_3_0_8(SH2_struct* sh)
{
  SH2sub(sh,3,0);
}
#define init_3_3_1_8
void FASTCALL call_3_3_1_8(SH2_struct* sh)
{
  SH2sub(sh,3,1);
}
#define init_3_3_2_8
void FASTCALL call_3_3_2_8(SH2_struct* sh)
{
  SH2sub(sh,3,2);
}
#define init_3_3_3_8
void FASTCALL call_3_3_3_8(SH2_struct* sh)
{
  SH2sub(sh,3,3);
}
#define init_3_3_4_8
void FASTCALL call_3_3_4_8(SH2_struct* sh)
{
  SH2sub(sh,3,4);
}
#define init_3_3_5_8
void FASTCALL call_3_3_5_8(SH2_struct* sh)
{
  SH2sub(sh,3,5);
}
#define init_3_3_6_8
void FASTCALL call_3_3_6_8(SH2_struct* sh)
{
  SH2sub(sh,3,6);
}
#define init_3_3_7_8
void FASTCALL call_3_3_7_8(SH2_struct* sh)
{
  SH2sub(sh,3,7);
}
#define init_3_3_8_8
void FASTCALL call_3_3_8_8(SH2_struct* sh)
{
  SH2sub(sh,3,8);
}
#define init_3_3_9_8
void FASTCALL call_3_3_9_8(SH2_struct* sh)
{
  SH2sub(sh,3,9);
}
#define init_3_3_10_8
void FASTCALL call_3_3_10_8(SH2_struct* sh)
{
  SH2sub(sh,3,10);
}
#define init_3_3_11_8
void FASTCALL call_3_3_11_8(SH2_struct* sh)
{
  SH2sub(sh,3,11);
}
#define init_3_3_12_8
void FASTCALL call_3_3_12_8(SH2_struct* sh)
{
  SH2sub(sh,3,12);
}
#define init_3_3_13_8
void FASTCALL call_3_3_13_8(SH2_struct* sh)
{
  SH2sub(sh,3,13);
}
#define init_3_3_14_8
void FASTCALL call_3_3_14_8(SH2_struct* sh)
{
  SH2sub(sh,3,14);
}
#define init_3_3_15_8
void FASTCALL call_3_3_15_8(SH2_struct* sh)
{
  SH2sub(sh,3,15);
}
#define init_3_4_0_8
void FASTCALL call_3_4_0_8(SH2_struct* sh)
{
  SH2sub(sh,4,0);
}
#define init_3_4_1_8
void FASTCALL call_3_4_1_8(SH2_struct* sh)
{
  SH2sub(sh,4,1);
}
#define init_3_4_2_8
void FASTCALL call_3_4_2_8(SH2_struct* sh)
{
  SH2sub(sh,4,2);
}
#define init_3_4_3_8
void FASTCALL call_3_4_3_8(SH2_struct* sh)
{
  SH2sub(sh,4,3);
}
#define init_3_4_4_8
void FASTCALL call_3_4_4_8(SH2_struct* sh)
{
  SH2sub(sh,4,4);
}
#define init_3_4_5_8
void FASTCALL call_3_4_5_8(SH2_struct* sh)
{
  SH2sub(sh,4,5);
}
#define init_3_4_6_8
void FASTCALL call_3_4_6_8(SH2_struct* sh)
{
  SH2sub(sh,4,6);
}
#define init_3_4_7_8
void FASTCALL call_3_4_7_8(SH2_struct* sh)
{
  SH2sub(sh,4,7);
}
#define init_3_4_8_8
void FASTCALL call_3_4_8_8(SH2_struct* sh)
{
  SH2sub(sh,4,8);
}
#define init_3_4_9_8
void FASTCALL call_3_4_9_8(SH2_struct* sh)
{
  SH2sub(sh,4,9);
}
#define init_3_4_10_8
void FASTCALL call_3_4_10_8(SH2_struct* sh)
{
  SH2sub(sh,4,10);
}
#define init_3_4_11_8
void FASTCALL call_3_4_11_8(SH2_struct* sh)
{
  SH2sub(sh,4,11);
}
#define init_3_4_12_8
void FASTCALL call_3_4_12_8(SH2_struct* sh)
{
  SH2sub(sh,4,12);
}
#define init_3_4_13_8
void FASTCALL call_3_4_13_8(SH2_struct* sh)
{
  SH2sub(sh,4,13);
}
#define init_3_4_14_8
void FASTCALL call_3_4_14_8(SH2_struct* sh)
{
  SH2sub(sh,4,14);
}
#define init_3_4_15_8
void FASTCALL call_3_4_15_8(SH2_struct* sh)
{
  SH2sub(sh,4,15);
}
#define init_3_5_0_8
void FASTCALL call_3_5_0_8(SH2_struct* sh)
{
  SH2sub(sh,5,0);
}
#define init_3_5_1_8
void FASTCALL call_3_5_1_8(SH2_struct* sh)
{
  SH2sub(sh,5,1);
}
#define init_3_5_2_8
void FASTCALL call_3_5_2_8(SH2_struct* sh)
{
  SH2sub(sh,5,2);
}
#define init_3_5_3_8
void FASTCALL call_3_5_3_8(SH2_struct* sh)
{
  SH2sub(sh,5,3);
}
#define init_3_5_4_8
void FASTCALL call_3_5_4_8(SH2_struct* sh)
{
  SH2sub(sh,5,4);
}
#define init_3_5_5_8
void FASTCALL call_3_5_5_8(SH2_struct* sh)
{
  SH2sub(sh,5,5);
}
#define init_3_5_6_8
void FASTCALL call_3_5_6_8(SH2_struct* sh)
{
  SH2sub(sh,5,6);
}
#define init_3_5_7_8
void FASTCALL call_3_5_7_8(SH2_struct* sh)
{
  SH2sub(sh,5,7);
}
#define init_3_5_8_8
void FASTCALL call_3_5_8_8(SH2_struct* sh)
{
  SH2sub(sh,5,8);
}
#define init_3_5_9_8
void FASTCALL call_3_5_9_8(SH2_struct* sh)
{
  SH2sub(sh,5,9);
}
#define init_3_5_10_8
void FASTCALL call_3_5_10_8(SH2_struct* sh)
{
  SH2sub(sh,5,10);
}
#define init_3_5_11_8
void FASTCALL call_3_5_11_8(SH2_struct* sh)
{
  SH2sub(sh,5,11);
}
#define init_3_5_12_8
void FASTCALL call_3_5_12_8(SH2_struct* sh)
{
  SH2sub(sh,5,12);
}
#define init_3_5_13_8
void FASTCALL call_3_5_13_8(SH2_struct* sh)
{
  SH2sub(sh,5,13);
}
#define init_3_5_14_8
void FASTCALL call_3_5_14_8(SH2_struct* sh)
{
  SH2sub(sh,5,14);
}
#define init_3_5_15_8
void FASTCALL call_3_5_15_8(SH2_struct* sh)
{
  SH2sub(sh,5,15);
}
#define init_3_6_0_8
void FASTCALL call_3_6_0_8(SH2_struct* sh)
{
  SH2sub(sh,6,0);
}
#define init_3_6_1_8
void FASTCALL call_3_6_1_8(SH2_struct* sh)
{
  SH2sub(sh,6,1);
}
#define init_3_6_2_8
void FASTCALL call_3_6_2_8(SH2_struct* sh)
{
  SH2sub(sh,6,2);
}
#define init_3_6_3_8
void FASTCALL call_3_6_3_8(SH2_struct* sh)
{
  SH2sub(sh,6,3);
}
#define init_3_6_4_8
void FASTCALL call_3_6_4_8(SH2_struct* sh)
{
  SH2sub(sh,6,4);
}
#define init_3_6_5_8
void FASTCALL call_3_6_5_8(SH2_struct* sh)
{
  SH2sub(sh,6,5);
}
#define init_3_6_6_8
void FASTCALL call_3_6_6_8(SH2_struct* sh)
{
  SH2sub(sh,6,6);
}
#define init_3_6_7_8
void FASTCALL call_3_6_7_8(SH2_struct* sh)
{
  SH2sub(sh,6,7);
}
#define init_3_6_8_8
void FASTCALL call_3_6_8_8(SH2_struct* sh)
{
  SH2sub(sh,6,8);
}
#define init_3_6_9_8
void FASTCALL call_3_6_9_8(SH2_struct* sh)
{
  SH2sub(sh,6,9);
}
#define init_3_6_10_8
void FASTCALL call_3_6_10_8(SH2_struct* sh)
{
  SH2sub(sh,6,10);
}
#define init_3_6_11_8
void FASTCALL call_3_6_11_8(SH2_struct* sh)
{
  SH2sub(sh,6,11);
}
#define init_3_6_12_8
void FASTCALL call_3_6_12_8(SH2_struct* sh)
{
  SH2sub(sh,6,12);
}
#define init_3_6_13_8
void FASTCALL call_3_6_13_8(SH2_struct* sh)
{
  SH2sub(sh,6,13);
}
#define init_3_6_14_8
void FASTCALL call_3_6_14_8(SH2_struct* sh)
{
  SH2sub(sh,6,14);
}
#define init_3_6_15_8
void FASTCALL call_3_6_15_8(SH2_struct* sh)
{
  SH2sub(sh,6,15);
}
#define init_3_7_0_8
void FASTCALL call_3_7_0_8(SH2_struct* sh)
{
  SH2sub(sh,7,0);
}
#define init_3_7_1_8
void FASTCALL call_3_7_1_8(SH2_struct* sh)
{
  SH2sub(sh,7,1);
}
#define init_3_7_2_8
void FASTCALL call_3_7_2_8(SH2_struct* sh)
{
  SH2sub(sh,7,2);
}
#define init_3_7_3_8
void FASTCALL call_3_7_3_8(SH2_struct* sh)
{
  SH2sub(sh,7,3);
}
#define init_3_7_4_8
void FASTCALL call_3_7_4_8(SH2_struct* sh)
{
  SH2sub(sh,7,4);
}
#define init_3_7_5_8
void FASTCALL call_3_7_5_8(SH2_struct* sh)
{
  SH2sub(sh,7,5);
}
#define init_3_7_6_8
void FASTCALL call_3_7_6_8(SH2_struct* sh)
{
  SH2sub(sh,7,6);
}
#define init_3_7_7_8
void FASTCALL call_3_7_7_8(SH2_struct* sh)
{
  SH2sub(sh,7,7);
}
#define init_3_7_8_8
void FASTCALL call_3_7_8_8(SH2_struct* sh)
{
  SH2sub(sh,7,8);
}
#define init_3_7_9_8
void FASTCALL call_3_7_9_8(SH2_struct* sh)
{
  SH2sub(sh,7,9);
}
#define init_3_7_10_8
void FASTCALL call_3_7_10_8(SH2_struct* sh)
{
  SH2sub(sh,7,10);
}
#define init_3_7_11_8
void FASTCALL call_3_7_11_8(SH2_struct* sh)
{
  SH2sub(sh,7,11);
}
#define init_3_7_12_8
void FASTCALL call_3_7_12_8(SH2_struct* sh)
{
  SH2sub(sh,7,12);
}
#define init_3_7_13_8
void FASTCALL call_3_7_13_8(SH2_struct* sh)
{
  SH2sub(sh,7,13);
}
#define init_3_7_14_8
void FASTCALL call_3_7_14_8(SH2_struct* sh)
{
  SH2sub(sh,7,14);
}
#define init_3_7_15_8
void FASTCALL call_3_7_15_8(SH2_struct* sh)
{
  SH2sub(sh,7,15);
}
#define init_3_8_0_8
void FASTCALL call_3_8_0_8(SH2_struct* sh)
{
  SH2sub(sh,8,0);
}
#define init_3_8_1_8
void FASTCALL call_3_8_1_8(SH2_struct* sh)
{
  SH2sub(sh,8,1);
}
#define init_3_8_2_8
void FASTCALL call_3_8_2_8(SH2_struct* sh)
{
  SH2sub(sh,8,2);
}
#define init_3_8_3_8
void FASTCALL call_3_8_3_8(SH2_struct* sh)
{
  SH2sub(sh,8,3);
}
#define init_3_8_4_8
void FASTCALL call_3_8_4_8(SH2_struct* sh)
{
  SH2sub(sh,8,4);
}
#define init_3_8_5_8
void FASTCALL call_3_8_5_8(SH2_struct* sh)
{
  SH2sub(sh,8,5);
}
#define init_3_8_6_8
void FASTCALL call_3_8_6_8(SH2_struct* sh)
{
  SH2sub(sh,8,6);
}
#define init_3_8_7_8
void FASTCALL call_3_8_7_8(SH2_struct* sh)
{
  SH2sub(sh,8,7);
}
#define init_3_8_8_8
void FASTCALL call_3_8_8_8(SH2_struct* sh)
{
  SH2sub(sh,8,8);
}
#define init_3_8_9_8
void FASTCALL call_3_8_9_8(SH2_struct* sh)
{
  SH2sub(sh,8,9);
}
#define init_3_8_10_8
void FASTCALL call_3_8_10_8(SH2_struct* sh)
{
  SH2sub(sh,8,10);
}
#define init_3_8_11_8
void FASTCALL call_3_8_11_8(SH2_struct* sh)
{
  SH2sub(sh,8,11);
}
#define init_3_8_12_8
void FASTCALL call_3_8_12_8(SH2_struct* sh)
{
  SH2sub(sh,8,12);
}
#define init_3_8_13_8
void FASTCALL call_3_8_13_8(SH2_struct* sh)
{
  SH2sub(sh,8,13);
}
#define init_3_8_14_8
void FASTCALL call_3_8_14_8(SH2_struct* sh)
{
  SH2sub(sh,8,14);
}
#define init_3_8_15_8
void FASTCALL call_3_8_15_8(SH2_struct* sh)
{
  SH2sub(sh,8,15);
}
#define init_3_9_0_8
void FASTCALL call_3_9_0_8(SH2_struct* sh)
{
  SH2sub(sh,9,0);
}
#define init_3_9_1_8
void FASTCALL call_3_9_1_8(SH2_struct* sh)
{
  SH2sub(sh,9,1);
}
#define init_3_9_2_8
void FASTCALL call_3_9_2_8(SH2_struct* sh)
{
  SH2sub(sh,9,2);
}
#define init_3_9_3_8
void FASTCALL call_3_9_3_8(SH2_struct* sh)
{
  SH2sub(sh,9,3);
}
#define init_3_9_4_8
void FASTCALL call_3_9_4_8(SH2_struct* sh)
{
  SH2sub(sh,9,4);
}
#define init_3_9_5_8
void FASTCALL call_3_9_5_8(SH2_struct* sh)
{
  SH2sub(sh,9,5);
}
#define init_3_9_6_8
void FASTCALL call_3_9_6_8(SH2_struct* sh)
{
  SH2sub(sh,9,6);
}
#define init_3_9_7_8
void FASTCALL call_3_9_7_8(SH2_struct* sh)
{
  SH2sub(sh,9,7);
}
#define init_3_9_8_8
void FASTCALL call_3_9_8_8(SH2_struct* sh)
{
  SH2sub(sh,9,8);
}
#define init_3_9_9_8
void FASTCALL call_3_9_9_8(SH2_struct* sh)
{
  SH2sub(sh,9,9);
}
#define init_3_9_10_8
void FASTCALL call_3_9_10_8(SH2_struct* sh)
{
  SH2sub(sh,9,10);
}
#define init_3_9_11_8
void FASTCALL call_3_9_11_8(SH2_struct* sh)
{
  SH2sub(sh,9,11);
}
#define init_3_9_12_8
void FASTCALL call_3_9_12_8(SH2_struct* sh)
{
  SH2sub(sh,9,12);
}
#define init_3_9_13_8
void FASTCALL call_3_9_13_8(SH2_struct* sh)
{
  SH2sub(sh,9,13);
}
#define init_3_9_14_8
void FASTCALL call_3_9_14_8(SH2_struct* sh)
{
  SH2sub(sh,9,14);
}
#define init_3_9_15_8
void FASTCALL call_3_9_15_8(SH2_struct* sh)
{
  SH2sub(sh,9,15);
}
#define init_3_10_0_8
void FASTCALL call_3_10_0_8(SH2_struct* sh)
{
  SH2sub(sh,10,0);
}
#define init_3_10_1_8
void FASTCALL call_3_10_1_8(SH2_struct* sh)
{
  SH2sub(sh,10,1);
}
#define init_3_10_2_8
void FASTCALL call_3_10_2_8(SH2_struct* sh)
{
  SH2sub(sh,10,2);
}
#define init_3_10_3_8
void FASTCALL call_3_10_3_8(SH2_struct* sh)
{
  SH2sub(sh,10,3);
}
#define init_3_10_4_8
void FASTCALL call_3_10_4_8(SH2_struct* sh)
{
  SH2sub(sh,10,4);
}
#define init_3_10_5_8
void FASTCALL call_3_10_5_8(SH2_struct* sh)
{
  SH2sub(sh,10,5);
}
#define init_3_10_6_8
void FASTCALL call_3_10_6_8(SH2_struct* sh)
{
  SH2sub(sh,10,6);
}
#define init_3_10_7_8
void FASTCALL call_3_10_7_8(SH2_struct* sh)
{
  SH2sub(sh,10,7);
}
#define init_3_10_8_8
void FASTCALL call_3_10_8_8(SH2_struct* sh)
{
  SH2sub(sh,10,8);
}
#define init_3_10_9_8
void FASTCALL call_3_10_9_8(SH2_struct* sh)
{
  SH2sub(sh,10,9);
}
#define init_3_10_10_8
void FASTCALL call_3_10_10_8(SH2_struct* sh)
{
  SH2sub(sh,10,10);
}
#define init_3_10_11_8
void FASTCALL call_3_10_11_8(SH2_struct* sh)
{
  SH2sub(sh,10,11);
}
#define init_3_10_12_8
void FASTCALL call_3_10_12_8(SH2_struct* sh)
{
  SH2sub(sh,10,12);
}
#define init_3_10_13_8
void FASTCALL call_3_10_13_8(SH2_struct* sh)
{
  SH2sub(sh,10,13);
}
#define init_3_10_14_8
void FASTCALL call_3_10_14_8(SH2_struct* sh)
{
  SH2sub(sh,10,14);
}
#define init_3_10_15_8
void FASTCALL call_3_10_15_8(SH2_struct* sh)
{
  SH2sub(sh,10,15);
}
#define init_3_11_0_8
void FASTCALL call_3_11_0_8(SH2_struct* sh)
{
  SH2sub(sh,11,0);
}
#define init_3_11_1_8
void FASTCALL call_3_11_1_8(SH2_struct* sh)
{
  SH2sub(sh,11,1);
}
#define init_3_11_2_8
void FASTCALL call_3_11_2_8(SH2_struct* sh)
{
  SH2sub(sh,11,2);
}
#define init_3_11_3_8
void FASTCALL call_3_11_3_8(SH2_struct* sh)
{
  SH2sub(sh,11,3);
}
#define init_3_11_4_8
void FASTCALL call_3_11_4_8(SH2_struct* sh)
{
  SH2sub(sh,11,4);
}
#define init_3_11_5_8
void FASTCALL call_3_11_5_8(SH2_struct* sh)
{
  SH2sub(sh,11,5);
}
#define init_3_11_6_8
void FASTCALL call_3_11_6_8(SH2_struct* sh)
{
  SH2sub(sh,11,6);
}
#define init_3_11_7_8
void FASTCALL call_3_11_7_8(SH2_struct* sh)
{
  SH2sub(sh,11,7);
}
#define init_3_11_8_8
void FASTCALL call_3_11_8_8(SH2_struct* sh)
{
  SH2sub(sh,11,8);
}
#define init_3_11_9_8
void FASTCALL call_3_11_9_8(SH2_struct* sh)
{
  SH2sub(sh,11,9);
}
#define init_3_11_10_8
void FASTCALL call_3_11_10_8(SH2_struct* sh)
{
  SH2sub(sh,11,10);
}
#define init_3_11_11_8
void FASTCALL call_3_11_11_8(SH2_struct* sh)
{
  SH2sub(sh,11,11);
}
#define init_3_11_12_8
void FASTCALL call_3_11_12_8(SH2_struct* sh)
{
  SH2sub(sh,11,12);
}
#define init_3_11_13_8
void FASTCALL call_3_11_13_8(SH2_struct* sh)
{
  SH2sub(sh,11,13);
}
#define init_3_11_14_8
void FASTCALL call_3_11_14_8(SH2_struct* sh)
{
  SH2sub(sh,11,14);
}
#define init_3_11_15_8
void FASTCALL call_3_11_15_8(SH2_struct* sh)
{
  SH2sub(sh,11,15);
}
#define init_3_12_0_8
void FASTCALL call_3_12_0_8(SH2_struct* sh)
{
  SH2sub(sh,12,0);
}
#define init_3_12_1_8
void FASTCALL call_3_12_1_8(SH2_struct* sh)
{
  SH2sub(sh,12,1);
}
#define init_3_12_2_8
void FASTCALL call_3_12_2_8(SH2_struct* sh)
{
  SH2sub(sh,12,2);
}
#define init_3_12_3_8
void FASTCALL call_3_12_3_8(SH2_struct* sh)
{
  SH2sub(sh,12,3);
}
#define init_3_12_4_8
void FASTCALL call_3_12_4_8(SH2_struct* sh)
{
  SH2sub(sh,12,4);
}
#define init_3_12_5_8
void FASTCALL call_3_12_5_8(SH2_struct* sh)
{
  SH2sub(sh,12,5);
}
#define init_3_12_6_8
void FASTCALL call_3_12_6_8(SH2_struct* sh)
{
  SH2sub(sh,12,6);
}
#define init_3_12_7_8
void FASTCALL call_3_12_7_8(SH2_struct* sh)
{
  SH2sub(sh,12,7);
}
#define init_3_12_8_8
void FASTCALL call_3_12_8_8(SH2_struct* sh)
{
  SH2sub(sh,12,8);
}
#define init_3_12_9_8
void FASTCALL call_3_12_9_8(SH2_struct* sh)
{
  SH2sub(sh,12,9);
}
#define init_3_12_10_8
void FASTCALL call_3_12_10_8(SH2_struct* sh)
{
  SH2sub(sh,12,10);
}
#define init_3_12_11_8
void FASTCALL call_3_12_11_8(SH2_struct* sh)
{
  SH2sub(sh,12,11);
}
#define init_3_12_12_8
void FASTCALL call_3_12_12_8(SH2_struct* sh)
{
  SH2sub(sh,12,12);
}
#define init_3_12_13_8
void FASTCALL call_3_12_13_8(SH2_struct* sh)
{
  SH2sub(sh,12,13);
}
#define init_3_12_14_8
void FASTCALL call_3_12_14_8(SH2_struct* sh)
{
  SH2sub(sh,12,14);
}
#define init_3_12_15_8
void FASTCALL call_3_12_15_8(SH2_struct* sh)
{
  SH2sub(sh,12,15);
}
#define init_3_13_0_8
void FASTCALL call_3_13_0_8(SH2_struct* sh)
{
  SH2sub(sh,13,0);
}
#define init_3_13_1_8
void FASTCALL call_3_13_1_8(SH2_struct* sh)
{
  SH2sub(sh,13,1);
}
#define init_3_13_2_8
void FASTCALL call_3_13_2_8(SH2_struct* sh)
{
  SH2sub(sh,13,2);
}
#define init_3_13_3_8
void FASTCALL call_3_13_3_8(SH2_struct* sh)
{
  SH2sub(sh,13,3);
}
#define init_3_13_4_8
void FASTCALL call_3_13_4_8(SH2_struct* sh)
{
  SH2sub(sh,13,4);
}
#define init_3_13_5_8
void FASTCALL call_3_13_5_8(SH2_struct* sh)
{
  SH2sub(sh,13,5);
}
#define init_3_13_6_8
void FASTCALL call_3_13_6_8(SH2_struct* sh)
{
  SH2sub(sh,13,6);
}
#define init_3_13_7_8
void FASTCALL call_3_13_7_8(SH2_struct* sh)
{
  SH2sub(sh,13,7);
}
#define init_3_13_8_8
void FASTCALL call_3_13_8_8(SH2_struct* sh)
{
  SH2sub(sh,13,8);
}
#define init_3_13_9_8
void FASTCALL call_3_13_9_8(SH2_struct* sh)
{
  SH2sub(sh,13,9);
}
#define init_3_13_10_8
void FASTCALL call_3_13_10_8(SH2_struct* sh)
{
  SH2sub(sh,13,10);
}
#define init_3_13_11_8
void FASTCALL call_3_13_11_8(SH2_struct* sh)
{
  SH2sub(sh,13,11);
}
#define init_3_13_12_8
void FASTCALL call_3_13_12_8(SH2_struct* sh)
{
  SH2sub(sh,13,12);
}
#define init_3_13_13_8
void FASTCALL call_3_13_13_8(SH2_struct* sh)
{
  SH2sub(sh,13,13);
}
#define init_3_13_14_8
void FASTCALL call_3_13_14_8(SH2_struct* sh)
{
  SH2sub(sh,13,14);
}
#define init_3_13_15_8
void FASTCALL call_3_13_15_8(SH2_struct* sh)
{
  SH2sub(sh,13,15);
}
#define init_3_14_0_8
void FASTCALL call_3_14_0_8(SH2_struct* sh)
{
  SH2sub(sh,14,0);
}
#define init_3_14_1_8
void FASTCALL call_3_14_1_8(SH2_struct* sh)
{
  SH2sub(sh,14,1);
}
#define init_3_14_2_8
void FASTCALL call_3_14_2_8(SH2_struct* sh)
{
  SH2sub(sh,14,2);
}
#define init_3_14_3_8
void FASTCALL call_3_14_3_8(SH2_struct* sh)
{
  SH2sub(sh,14,3);
}
#define init_3_14_4_8
void FASTCALL call_3_14_4_8(SH2_struct* sh)
{
  SH2sub(sh,14,4);
}
#define init_3_14_5_8
void FASTCALL call_3_14_5_8(SH2_struct* sh)
{
  SH2sub(sh,14,5);
}
#define init_3_14_6_8
void FASTCALL call_3_14_6_8(SH2_struct* sh)
{
  SH2sub(sh,14,6);
}
#define init_3_14_7_8
void FASTCALL call_3_14_7_8(SH2_struct* sh)
{
  SH2sub(sh,14,7);
}
#define init_3_14_8_8
void FASTCALL call_3_14_8_8(SH2_struct* sh)
{
  SH2sub(sh,14,8);
}
#define init_3_14_9_8
void FASTCALL call_3_14_9_8(SH2_struct* sh)
{
  SH2sub(sh,14,9);
}
#define init_3_14_10_8
void FASTCALL call_3_14_10_8(SH2_struct* sh)
{
  SH2sub(sh,14,10);
}
#define init_3_14_11_8
void FASTCALL call_3_14_11_8(SH2_struct* sh)
{
  SH2sub(sh,14,11);
}
#define init_3_14_12_8
void FASTCALL call_3_14_12_8(SH2_struct* sh)
{
  SH2sub(sh,14,12);
}
#define init_3_14_13_8
void FASTCALL call_3_14_13_8(SH2_struct* sh)
{
  SH2sub(sh,14,13);
}
#define init_3_14_14_8
void FASTCALL call_3_14_14_8(SH2_struct* sh)
{
  SH2sub(sh,14,14);
}
#define init_3_14_15_8
void FASTCALL call_3_14_15_8(SH2_struct* sh)
{
  SH2sub(sh,14,15);
}
#define init_3_15_0_8
void FASTCALL call_3_15_0_8(SH2_struct* sh)
{
  SH2sub(sh,15,0);
}
#define init_3_15_1_8
void FASTCALL call_3_15_1_8(SH2_struct* sh)
{
  SH2sub(sh,15,1);
}
#define init_3_15_2_8
void FASTCALL call_3_15_2_8(SH2_struct* sh)
{
  SH2sub(sh,15,2);
}
#define init_3_15_3_8
void FASTCALL call_3_15_3_8(SH2_struct* sh)
{
  SH2sub(sh,15,3);
}
#define init_3_15_4_8
void FASTCALL call_3_15_4_8(SH2_struct* sh)
{
  SH2sub(sh,15,4);
}
#define init_3_15_5_8
void FASTCALL call_3_15_5_8(SH2_struct* sh)
{
  SH2sub(sh,15,5);
}
#define init_3_15_6_8
void FASTCALL call_3_15_6_8(SH2_struct* sh)
{
  SH2sub(sh,15,6);
}
#define init_3_15_7_8
void FASTCALL call_3_15_7_8(SH2_struct* sh)
{
  SH2sub(sh,15,7);
}
#define init_3_15_8_8
void FASTCALL call_3_15_8_8(SH2_struct* sh)
{
  SH2sub(sh,15,8);
}
#define init_3_15_9_8
void FASTCALL call_3_15_9_8(SH2_struct* sh)
{
  SH2sub(sh,15,9);
}
#define init_3_15_10_8
void FASTCALL call_3_15_10_8(SH2_struct* sh)
{
  SH2sub(sh,15,10);
}
#define init_3_15_11_8
void FASTCALL call_3_15_11_8(SH2_struct* sh)
{
  SH2sub(sh,15,11);
}
#define init_3_15_12_8
void FASTCALL call_3_15_12_8(SH2_struct* sh)
{
  SH2sub(sh,15,12);
}
#define init_3_15_13_8
void FASTCALL call_3_15_13_8(SH2_struct* sh)
{
  SH2sub(sh,15,13);
}
#define init_3_15_14_8
void FASTCALL call_3_15_14_8(SH2_struct* sh)
{
  SH2sub(sh,15,14);
}
#define init_3_15_15_8
void FASTCALL call_3_15_15_8(SH2_struct* sh)
{
  SH2sub(sh,15,15);
}
#define init_4_0_2_2
void FASTCALL call_4_0_2_2(SH2_struct* sh)
{
  SH2stsmpr(sh,0);
}
#define init_4_1_2_2
void FASTCALL call_4_1_2_2(SH2_struct* sh)
{
  SH2stsmpr(sh,1);
}
#define init_4_2_2_2
void FASTCALL call_4_2_2_2(SH2_struct* sh)
{
  SH2stsmpr(sh,2);
}
#define init_4_3_2_2
void FASTCALL call_4_3_2_2(SH2_struct* sh)
{
  SH2stsmpr(sh,3);
}
#define init_4_4_2_2
void FASTCALL call_4_4_2_2(SH2_struct* sh)
{
  SH2stsmpr(sh,4);
}
#define init_4_5_2_2
void FASTCALL call_4_5_2_2(SH2_struct* sh)
{
  SH2stsmpr(sh,5);
}
#define init_4_6_2_2
void FASTCALL call_4_6_2_2(SH2_struct* sh)
{
  SH2stsmpr(sh,6);
}
#define init_4_7_2_2
void FASTCALL call_4_7_2_2(SH2_struct* sh)
{
  SH2stsmpr(sh,7);
}
#define init_4_8_2_2
void FASTCALL call_4_8_2_2(SH2_struct* sh)
{
  SH2stsmpr(sh,8);
}
#define init_4_9_2_2
void FASTCALL call_4_9_2_2(SH2_struct* sh)
{
  SH2stsmpr(sh,9);
}
#define init_4_10_2_2
void FASTCALL call_4_10_2_2(SH2_struct* sh)
{
  SH2stsmpr(sh,10);
}
#define init_4_11_2_2
void FASTCALL call_4_11_2_2(SH2_struct* sh)
{
  SH2stsmpr(sh,11);
}
#define init_4_12_2_2
void FASTCALL call_4_12_2_2(SH2_struct* sh)
{
  SH2stsmpr(sh,12);
}
#define init_4_13_2_2
void FASTCALL call_4_13_2_2(SH2_struct* sh)
{
  SH2stsmpr(sh,13);
}
#define init_4_14_2_2
void FASTCALL call_4_14_2_2(SH2_struct* sh)
{
  SH2stsmpr(sh,14);
}
#define init_4_15_2_2
void FASTCALL call_4_15_2_2(SH2_struct* sh)
{
  SH2stsmpr(sh,15);
}
#define init_4_0_1_2
void FASTCALL call_4_0_1_2(SH2_struct* sh)
{
  SH2stsmmacl(sh,0);
}
#define init_4_1_1_2
void FASTCALL call_4_1_1_2(SH2_struct* sh)
{
  SH2stsmmacl(sh,1);
}
#define init_4_2_1_2
void FASTCALL call_4_2_1_2(SH2_struct* sh)
{
  SH2stsmmacl(sh,2);
}
#define init_4_3_1_2
void FASTCALL call_4_3_1_2(SH2_struct* sh)
{
  SH2stsmmacl(sh,3);
}
#define init_4_4_1_2
void FASTCALL call_4_4_1_2(SH2_struct* sh)
{
  SH2stsmmacl(sh,4);
}
#define init_4_5_1_2
void FASTCALL call_4_5_1_2(SH2_struct* sh)
{
  SH2stsmmacl(sh,5);
}
#define init_4_6_1_2
void FASTCALL call_4_6_1_2(SH2_struct* sh)
{
  SH2stsmmacl(sh,6);
}
#define init_4_7_1_2
void FASTCALL call_4_7_1_2(SH2_struct* sh)
{
  SH2stsmmacl(sh,7);
}
#define init_4_8_1_2
void FASTCALL call_4_8_1_2(SH2_struct* sh)
{
  SH2stsmmacl(sh,8);
}
#define init_4_9_1_2
void FASTCALL call_4_9_1_2(SH2_struct* sh)
{
  SH2stsmmacl(sh,9);
}
#define init_4_10_1_2
void FASTCALL call_4_10_1_2(SH2_struct* sh)
{
  SH2stsmmacl(sh,10);
}
#define init_4_11_1_2
void FASTCALL call_4_11_1_2(SH2_struct* sh)
{
  SH2stsmmacl(sh,11);
}
#define init_4_12_1_2
void FASTCALL call_4_12_1_2(SH2_struct* sh)
{
  SH2stsmmacl(sh,12);
}
#define init_4_13_1_2
void FASTCALL call_4_13_1_2(SH2_struct* sh)
{
  SH2stsmmacl(sh,13);
}
#define init_4_14_1_2
void FASTCALL call_4_14_1_2(SH2_struct* sh)
{
  SH2stsmmacl(sh,14);
}
#define init_4_15_1_2
void FASTCALL call_4_15_1_2(SH2_struct* sh)
{
  SH2stsmmacl(sh,15);
}
#define init_4_0_0_2
void FASTCALL call_4_0_0_2(SH2_struct* sh)
{
  SH2stsmmach(sh,0);
}
#define init_4_1_0_2
void FASTCALL call_4_1_0_2(SH2_struct* sh)
{
  SH2stsmmach(sh,1);
}
#define init_4_2_0_2
void FASTCALL call_4_2_0_2(SH2_struct* sh)
{
  SH2stsmmach(sh,2);
}
#define init_4_3_0_2
void FASTCALL call_4_3_0_2(SH2_struct* sh)
{
  SH2stsmmach(sh,3);
}
#define init_4_4_0_2
void FASTCALL call_4_4_0_2(SH2_struct* sh)
{
  SH2stsmmach(sh,4);
}
#define init_4_5_0_2
void FASTCALL call_4_5_0_2(SH2_struct* sh)
{
  SH2stsmmach(sh,5);
}
#define init_4_6_0_2
void FASTCALL call_4_6_0_2(SH2_struct* sh)
{
  SH2stsmmach(sh,6);
}
#define init_4_7_0_2
void FASTCALL call_4_7_0_2(SH2_struct* sh)
{
  SH2stsmmach(sh,7);
}
#define init_4_8_0_2
void FASTCALL call_4_8_0_2(SH2_struct* sh)
{
  SH2stsmmach(sh,8);
}
#define init_4_9_0_2
void FASTCALL call_4_9_0_2(SH2_struct* sh)
{
  SH2stsmmach(sh,9);
}
#define init_4_10_0_2
void FASTCALL call_4_10_0_2(SH2_struct* sh)
{
  SH2stsmmach(sh,10);
}
#define init_4_11_0_2
void FASTCALL call_4_11_0_2(SH2_struct* sh)
{
  SH2stsmmach(sh,11);
}
#define init_4_12_0_2
void FASTCALL call_4_12_0_2(SH2_struct* sh)
{
  SH2stsmmach(sh,12);
}
#define init_4_13_0_2
void FASTCALL call_4_13_0_2(SH2_struct* sh)
{
  SH2stsmmach(sh,13);
}
#define init_4_14_0_2
void FASTCALL call_4_14_0_2(SH2_struct* sh)
{
  SH2stsmmach(sh,14);
}
#define init_4_15_0_2
void FASTCALL call_4_15_0_2(SH2_struct* sh)
{
  SH2stsmmach(sh,15);
}
#define init_4_0_2_3
void FASTCALL call_4_0_2_3(SH2_struct* sh)
{
  SH2stcmvbr(sh,0);
}
#define init_4_1_2_3
void FASTCALL call_4_1_2_3(SH2_struct* sh)
{
  SH2stcmvbr(sh,1);
}
#define init_4_2_2_3
void FASTCALL call_4_2_2_3(SH2_struct* sh)
{
  SH2stcmvbr(sh,2);
}
#define init_4_3_2_3
void FASTCALL call_4_3_2_3(SH2_struct* sh)
{
  SH2stcmvbr(sh,3);
}
#define init_4_4_2_3
void FASTCALL call_4_4_2_3(SH2_struct* sh)
{
  SH2stcmvbr(sh,4);
}
#define init_4_5_2_3
void FASTCALL call_4_5_2_3(SH2_struct* sh)
{
  SH2stcmvbr(sh,5);
}
#define init_4_6_2_3
void FASTCALL call_4_6_2_3(SH2_struct* sh)
{
  SH2stcmvbr(sh,6);
}
#define init_4_7_2_3
void FASTCALL call_4_7_2_3(SH2_struct* sh)
{
  SH2stcmvbr(sh,7);
}
#define init_4_8_2_3
void FASTCALL call_4_8_2_3(SH2_struct* sh)
{
  SH2stcmvbr(sh,8);
}
#define init_4_9_2_3
void FASTCALL call_4_9_2_3(SH2_struct* sh)
{
  SH2stcmvbr(sh,9);
}
#define init_4_10_2_3
void FASTCALL call_4_10_2_3(SH2_struct* sh)
{
  SH2stcmvbr(sh,10);
}
#define init_4_11_2_3
void FASTCALL call_4_11_2_3(SH2_struct* sh)
{
  SH2stcmvbr(sh,11);
}
#define init_4_12_2_3
void FASTCALL call_4_12_2_3(SH2_struct* sh)
{
  SH2stcmvbr(sh,12);
}
#define init_4_13_2_3
void FASTCALL call_4_13_2_3(SH2_struct* sh)
{
  SH2stcmvbr(sh,13);
}
#define init_4_14_2_3
void FASTCALL call_4_14_2_3(SH2_struct* sh)
{
  SH2stcmvbr(sh,14);
}
#define init_4_15_2_3
void FASTCALL call_4_15_2_3(SH2_struct* sh)
{
  SH2stcmvbr(sh,15);
}
#define init_4_0_0_3
void FASTCALL call_4_0_0_3(SH2_struct* sh)
{
  SH2stcmsr(sh,0);
}
#define init_4_1_0_3
void FASTCALL call_4_1_0_3(SH2_struct* sh)
{
  SH2stcmsr(sh,1);
}
#define init_4_2_0_3
void FASTCALL call_4_2_0_3(SH2_struct* sh)
{
  SH2stcmsr(sh,2);
}
#define init_4_3_0_3
void FASTCALL call_4_3_0_3(SH2_struct* sh)
{
  SH2stcmsr(sh,3);
}
#define init_4_4_0_3
void FASTCALL call_4_4_0_3(SH2_struct* sh)
{
  SH2stcmsr(sh,4);
}
#define init_4_5_0_3
void FASTCALL call_4_5_0_3(SH2_struct* sh)
{
  SH2stcmsr(sh,5);
}
#define init_4_6_0_3
void FASTCALL call_4_6_0_3(SH2_struct* sh)
{
  SH2stcmsr(sh,6);
}
#define init_4_7_0_3
void FASTCALL call_4_7_0_3(SH2_struct* sh)
{
  SH2stcmsr(sh,7);
}
#define init_4_8_0_3
void FASTCALL call_4_8_0_3(SH2_struct* sh)
{
  SH2stcmsr(sh,8);
}
#define init_4_9_0_3
void FASTCALL call_4_9_0_3(SH2_struct* sh)
{
  SH2stcmsr(sh,9);
}
#define init_4_10_0_3
void FASTCALL call_4_10_0_3(SH2_struct* sh)
{
  SH2stcmsr(sh,10);
}
#define init_4_11_0_3
void FASTCALL call_4_11_0_3(SH2_struct* sh)
{
  SH2stcmsr(sh,11);
}
#define init_4_12_0_3
void FASTCALL call_4_12_0_3(SH2_struct* sh)
{
  SH2stcmsr(sh,12);
}
#define init_4_13_0_3
void FASTCALL call_4_13_0_3(SH2_struct* sh)
{
  SH2stcmsr(sh,13);
}
#define init_4_14_0_3
void FASTCALL call_4_14_0_3(SH2_struct* sh)
{
  SH2stcmsr(sh,14);
}
#define init_4_15_0_3
void FASTCALL call_4_15_0_3(SH2_struct* sh)
{
  SH2stcmsr(sh,15);
}
#define init_4_0_1_3
void FASTCALL call_4_0_1_3(SH2_struct* sh)
{
  SH2stcmgbr(sh,0);
}
#define init_4_1_1_3
void FASTCALL call_4_1_1_3(SH2_struct* sh)
{
  SH2stcmgbr(sh,1);
}
#define init_4_2_1_3
void FASTCALL call_4_2_1_3(SH2_struct* sh)
{
  SH2stcmgbr(sh,2);
}
#define init_4_3_1_3
void FASTCALL call_4_3_1_3(SH2_struct* sh)
{
  SH2stcmgbr(sh,3);
}
#define init_4_4_1_3
void FASTCALL call_4_4_1_3(SH2_struct* sh)
{
  SH2stcmgbr(sh,4);
}
#define init_4_5_1_3
void FASTCALL call_4_5_1_3(SH2_struct* sh)
{
  SH2stcmgbr(sh,5);
}
#define init_4_6_1_3
void FASTCALL call_4_6_1_3(SH2_struct* sh)
{
  SH2stcmgbr(sh,6);
}
#define init_4_7_1_3
void FASTCALL call_4_7_1_3(SH2_struct* sh)
{
  SH2stcmgbr(sh,7);
}
#define init_4_8_1_3
void FASTCALL call_4_8_1_3(SH2_struct* sh)
{
  SH2stcmgbr(sh,8);
}
#define init_4_9_1_3
void FASTCALL call_4_9_1_3(SH2_struct* sh)
{
  SH2stcmgbr(sh,9);
}
#define init_4_10_1_3
void FASTCALL call_4_10_1_3(SH2_struct* sh)
{
  SH2stcmgbr(sh,10);
}
#define init_4_11_1_3
void FASTCALL call_4_11_1_3(SH2_struct* sh)
{
  SH2stcmgbr(sh,11);
}
#define init_4_12_1_3
void FASTCALL call_4_12_1_3(SH2_struct* sh)
{
  SH2stcmgbr(sh,12);
}
#define init_4_13_1_3
void FASTCALL call_4_13_1_3(SH2_struct* sh)
{
  SH2stcmgbr(sh,13);
}
#define init_4_14_1_3
void FASTCALL call_4_14_1_3(SH2_struct* sh)
{
  SH2stcmgbr(sh,14);
}
#define init_4_15_1_3
void FASTCALL call_4_15_1_3(SH2_struct* sh)
{
  SH2stcmgbr(sh,15);
}
#define init_4_0_2_9
void FASTCALL call_4_0_2_9(SH2_struct* sh)
{
  SH2shlr16(sh,0);
}
#define init_4_1_2_9
void FASTCALL call_4_1_2_9(SH2_struct* sh)
{
  SH2shlr16(sh,1);
}
#define init_4_2_2_9
void FASTCALL call_4_2_2_9(SH2_struct* sh)
{
  SH2shlr16(sh,2);
}
#define init_4_3_2_9
void FASTCALL call_4_3_2_9(SH2_struct* sh)
{
  SH2shlr16(sh,3);
}
#define init_4_4_2_9
void FASTCALL call_4_4_2_9(SH2_struct* sh)
{
  SH2shlr16(sh,4);
}
#define init_4_5_2_9
void FASTCALL call_4_5_2_9(SH2_struct* sh)
{
  SH2shlr16(sh,5);
}
#define init_4_6_2_9
void FASTCALL call_4_6_2_9(SH2_struct* sh)
{
  SH2shlr16(sh,6);
}
#define init_4_7_2_9
void FASTCALL call_4_7_2_9(SH2_struct* sh)
{
  SH2shlr16(sh,7);
}
#define init_4_8_2_9
void FASTCALL call_4_8_2_9(SH2_struct* sh)
{
  SH2shlr16(sh,8);
}
#define init_4_9_2_9
void FASTCALL call_4_9_2_9(SH2_struct* sh)
{
  SH2shlr16(sh,9);
}
#define init_4_10_2_9
void FASTCALL call_4_10_2_9(SH2_struct* sh)
{
  SH2shlr16(sh,10);
}
#define init_4_11_2_9
void FASTCALL call_4_11_2_9(SH2_struct* sh)
{
  SH2shlr16(sh,11);
}
#define init_4_12_2_9
void FASTCALL call_4_12_2_9(SH2_struct* sh)
{
  SH2shlr16(sh,12);
}
#define init_4_13_2_9
void FASTCALL call_4_13_2_9(SH2_struct* sh)
{
  SH2shlr16(sh,13);
}
#define init_4_14_2_9
void FASTCALL call_4_14_2_9(SH2_struct* sh)
{
  SH2shlr16(sh,14);
}
#define init_4_15_2_9
void FASTCALL call_4_15_2_9(SH2_struct* sh)
{
  SH2shlr16(sh,15);
}
#define init_4_0_1_9
void FASTCALL call_4_0_1_9(SH2_struct* sh)
{
  SH2shlr8(sh,0);
}
#define init_4_1_1_9
void FASTCALL call_4_1_1_9(SH2_struct* sh)
{
  SH2shlr8(sh,1);
}
#define init_4_2_1_9
void FASTCALL call_4_2_1_9(SH2_struct* sh)
{
  SH2shlr8(sh,2);
}
#define init_4_3_1_9
void FASTCALL call_4_3_1_9(SH2_struct* sh)
{
  SH2shlr8(sh,3);
}
#define init_4_4_1_9
void FASTCALL call_4_4_1_9(SH2_struct* sh)
{
  SH2shlr8(sh,4);
}
#define init_4_5_1_9
void FASTCALL call_4_5_1_9(SH2_struct* sh)
{
  SH2shlr8(sh,5);
}
#define init_4_6_1_9
void FASTCALL call_4_6_1_9(SH2_struct* sh)
{
  SH2shlr8(sh,6);
}
#define init_4_7_1_9
void FASTCALL call_4_7_1_9(SH2_struct* sh)
{
  SH2shlr8(sh,7);
}
#define init_4_8_1_9
void FASTCALL call_4_8_1_9(SH2_struct* sh)
{
  SH2shlr8(sh,8);
}
#define init_4_9_1_9
void FASTCALL call_4_9_1_9(SH2_struct* sh)
{
  SH2shlr8(sh,9);
}
#define init_4_10_1_9
void FASTCALL call_4_10_1_9(SH2_struct* sh)
{
  SH2shlr8(sh,10);
}
#define init_4_11_1_9
void FASTCALL call_4_11_1_9(SH2_struct* sh)
{
  SH2shlr8(sh,11);
}
#define init_4_12_1_9
void FASTCALL call_4_12_1_9(SH2_struct* sh)
{
  SH2shlr8(sh,12);
}
#define init_4_13_1_9
void FASTCALL call_4_13_1_9(SH2_struct* sh)
{
  SH2shlr8(sh,13);
}
#define init_4_14_1_9
void FASTCALL call_4_14_1_9(SH2_struct* sh)
{
  SH2shlr8(sh,14);
}
#define init_4_15_1_9
void FASTCALL call_4_15_1_9(SH2_struct* sh)
{
  SH2shlr8(sh,15);
}
#define init_4_0_0_9
void FASTCALL call_4_0_0_9(SH2_struct* sh)
{
  SH2shlr2(sh,0);
}
#define init_4_1_0_9
void FASTCALL call_4_1_0_9(SH2_struct* sh)
{
  SH2shlr2(sh,1);
}
#define init_4_2_0_9
void FASTCALL call_4_2_0_9(SH2_struct* sh)
{
  SH2shlr2(sh,2);
}
#define init_4_3_0_9
void FASTCALL call_4_3_0_9(SH2_struct* sh)
{
  SH2shlr2(sh,3);
}
#define init_4_4_0_9
void FASTCALL call_4_4_0_9(SH2_struct* sh)
{
  SH2shlr2(sh,4);
}
#define init_4_5_0_9
void FASTCALL call_4_5_0_9(SH2_struct* sh)
{
  SH2shlr2(sh,5);
}
#define init_4_6_0_9
void FASTCALL call_4_6_0_9(SH2_struct* sh)
{
  SH2shlr2(sh,6);
}
#define init_4_7_0_9
void FASTCALL call_4_7_0_9(SH2_struct* sh)
{
  SH2shlr2(sh,7);
}
#define init_4_8_0_9
void FASTCALL call_4_8_0_9(SH2_struct* sh)
{
  SH2shlr2(sh,8);
}
#define init_4_9_0_9
void FASTCALL call_4_9_0_9(SH2_struct* sh)
{
  SH2shlr2(sh,9);
}
#define init_4_10_0_9
void FASTCALL call_4_10_0_9(SH2_struct* sh)
{
  SH2shlr2(sh,10);
}
#define init_4_11_0_9
void FASTCALL call_4_11_0_9(SH2_struct* sh)
{
  SH2shlr2(sh,11);
}
#define init_4_12_0_9
void FASTCALL call_4_12_0_9(SH2_struct* sh)
{
  SH2shlr2(sh,12);
}
#define init_4_13_0_9
void FASTCALL call_4_13_0_9(SH2_struct* sh)
{
  SH2shlr2(sh,13);
}
#define init_4_14_0_9
void FASTCALL call_4_14_0_9(SH2_struct* sh)
{
  SH2shlr2(sh,14);
}
#define init_4_15_0_9
void FASTCALL call_4_15_0_9(SH2_struct* sh)
{
  SH2shlr2(sh,15);
}
#define init_4_0_0_1
void FASTCALL call_4_0_0_1(SH2_struct* sh)
{
  SH2shlr(sh,0);
}
#define init_4_1_0_1
void FASTCALL call_4_1_0_1(SH2_struct* sh)
{
  SH2shlr(sh,1);
}
#define init_4_2_0_1
void FASTCALL call_4_2_0_1(SH2_struct* sh)
{
  SH2shlr(sh,2);
}
#define init_4_3_0_1
void FASTCALL call_4_3_0_1(SH2_struct* sh)
{
  SH2shlr(sh,3);
}
#define init_4_4_0_1
void FASTCALL call_4_4_0_1(SH2_struct* sh)
{
  SH2shlr(sh,4);
}
#define init_4_5_0_1
void FASTCALL call_4_5_0_1(SH2_struct* sh)
{
  SH2shlr(sh,5);
}
#define init_4_6_0_1
void FASTCALL call_4_6_0_1(SH2_struct* sh)
{
  SH2shlr(sh,6);
}
#define init_4_7_0_1
void FASTCALL call_4_7_0_1(SH2_struct* sh)
{
  SH2shlr(sh,7);
}
#define init_4_8_0_1
void FASTCALL call_4_8_0_1(SH2_struct* sh)
{
  SH2shlr(sh,8);
}
#define init_4_9_0_1
void FASTCALL call_4_9_0_1(SH2_struct* sh)
{
  SH2shlr(sh,9);
}
#define init_4_10_0_1
void FASTCALL call_4_10_0_1(SH2_struct* sh)
{
  SH2shlr(sh,10);
}
#define init_4_11_0_1
void FASTCALL call_4_11_0_1(SH2_struct* sh)
{
  SH2shlr(sh,11);
}
#define init_4_12_0_1
void FASTCALL call_4_12_0_1(SH2_struct* sh)
{
  SH2shlr(sh,12);
}
#define init_4_13_0_1
void FASTCALL call_4_13_0_1(SH2_struct* sh)
{
  SH2shlr(sh,13);
}
#define init_4_14_0_1
void FASTCALL call_4_14_0_1(SH2_struct* sh)
{
  SH2shlr(sh,14);
}
#define init_4_15_0_1
void FASTCALL call_4_15_0_1(SH2_struct* sh)
{
  SH2shlr(sh,15);
}
#define init_4_0_2_8
void FASTCALL call_4_0_2_8(SH2_struct* sh)
{
  SH2shll16(sh,0);
}
#define init_4_1_2_8
void FASTCALL call_4_1_2_8(SH2_struct* sh)
{
  SH2shll16(sh,1);
}
#define init_4_2_2_8
void FASTCALL call_4_2_2_8(SH2_struct* sh)
{
  SH2shll16(sh,2);
}
#define init_4_3_2_8
void FASTCALL call_4_3_2_8(SH2_struct* sh)
{
  SH2shll16(sh,3);
}
#define init_4_4_2_8
void FASTCALL call_4_4_2_8(SH2_struct* sh)
{
  SH2shll16(sh,4);
}
#define init_4_5_2_8
void FASTCALL call_4_5_2_8(SH2_struct* sh)
{
  SH2shll16(sh,5);
}
#define init_4_6_2_8
void FASTCALL call_4_6_2_8(SH2_struct* sh)
{
  SH2shll16(sh,6);
}
#define init_4_7_2_8
void FASTCALL call_4_7_2_8(SH2_struct* sh)
{
  SH2shll16(sh,7);
}
#define init_4_8_2_8
void FASTCALL call_4_8_2_8(SH2_struct* sh)
{
  SH2shll16(sh,8);
}
#define init_4_9_2_8
void FASTCALL call_4_9_2_8(SH2_struct* sh)
{
  SH2shll16(sh,9);
}
#define init_4_10_2_8
void FASTCALL call_4_10_2_8(SH2_struct* sh)
{
  SH2shll16(sh,10);
}
#define init_4_11_2_8
void FASTCALL call_4_11_2_8(SH2_struct* sh)
{
  SH2shll16(sh,11);
}
#define init_4_12_2_8
void FASTCALL call_4_12_2_8(SH2_struct* sh)
{
  SH2shll16(sh,12);
}
#define init_4_13_2_8
void FASTCALL call_4_13_2_8(SH2_struct* sh)
{
  SH2shll16(sh,13);
}
#define init_4_14_2_8
void FASTCALL call_4_14_2_8(SH2_struct* sh)
{
  SH2shll16(sh,14);
}
#define init_4_15_2_8
void FASTCALL call_4_15_2_8(SH2_struct* sh)
{
  SH2shll16(sh,15);
}
#define init_4_0_1_8
void FASTCALL call_4_0_1_8(SH2_struct* sh)
{
  SH2shll8(sh,0);
}
#define init_4_1_1_8
void FASTCALL call_4_1_1_8(SH2_struct* sh)
{
  SH2shll8(sh,1);
}
#define init_4_2_1_8
void FASTCALL call_4_2_1_8(SH2_struct* sh)
{
  SH2shll8(sh,2);
}
#define init_4_3_1_8
void FASTCALL call_4_3_1_8(SH2_struct* sh)
{
  SH2shll8(sh,3);
}
#define init_4_4_1_8
void FASTCALL call_4_4_1_8(SH2_struct* sh)
{
  SH2shll8(sh,4);
}
#define init_4_5_1_8
void FASTCALL call_4_5_1_8(SH2_struct* sh)
{
  SH2shll8(sh,5);
}
#define init_4_6_1_8
void FASTCALL call_4_6_1_8(SH2_struct* sh)
{
  SH2shll8(sh,6);
}
#define init_4_7_1_8
void FASTCALL call_4_7_1_8(SH2_struct* sh)
{
  SH2shll8(sh,7);
}
#define init_4_8_1_8
void FASTCALL call_4_8_1_8(SH2_struct* sh)
{
  SH2shll8(sh,8);
}
#define init_4_9_1_8
void FASTCALL call_4_9_1_8(SH2_struct* sh)
{
  SH2shll8(sh,9);
}
#define init_4_10_1_8
void FASTCALL call_4_10_1_8(SH2_struct* sh)
{
  SH2shll8(sh,10);
}
#define init_4_11_1_8
void FASTCALL call_4_11_1_8(SH2_struct* sh)
{
  SH2shll8(sh,11);
}
#define init_4_12_1_8
void FASTCALL call_4_12_1_8(SH2_struct* sh)
{
  SH2shll8(sh,12);
}
#define init_4_13_1_8
void FASTCALL call_4_13_1_8(SH2_struct* sh)
{
  SH2shll8(sh,13);
}
#define init_4_14_1_8
void FASTCALL call_4_14_1_8(SH2_struct* sh)
{
  SH2shll8(sh,14);
}
#define init_4_15_1_8
void FASTCALL call_4_15_1_8(SH2_struct* sh)
{
  SH2shll8(sh,15);
}
#define init_4_0_0_8
void FASTCALL call_4_0_0_8(SH2_struct* sh)
{
  SH2shll2(sh,0);
}
#define init_4_1_0_8
void FASTCALL call_4_1_0_8(SH2_struct* sh)
{
  SH2shll2(sh,1);
}
#define init_4_2_0_8
void FASTCALL call_4_2_0_8(SH2_struct* sh)
{
  SH2shll2(sh,2);
}
#define init_4_3_0_8
void FASTCALL call_4_3_0_8(SH2_struct* sh)
{
  SH2shll2(sh,3);
}
#define init_4_4_0_8
void FASTCALL call_4_4_0_8(SH2_struct* sh)
{
  SH2shll2(sh,4);
}
#define init_4_5_0_8
void FASTCALL call_4_5_0_8(SH2_struct* sh)
{
  SH2shll2(sh,5);
}
#define init_4_6_0_8
void FASTCALL call_4_6_0_8(SH2_struct* sh)
{
  SH2shll2(sh,6);
}
#define init_4_7_0_8
void FASTCALL call_4_7_0_8(SH2_struct* sh)
{
  SH2shll2(sh,7);
}
#define init_4_8_0_8
void FASTCALL call_4_8_0_8(SH2_struct* sh)
{
  SH2shll2(sh,8);
}
#define init_4_9_0_8
void FASTCALL call_4_9_0_8(SH2_struct* sh)
{
  SH2shll2(sh,9);
}
#define init_4_10_0_8
void FASTCALL call_4_10_0_8(SH2_struct* sh)
{
  SH2shll2(sh,10);
}
#define init_4_11_0_8
void FASTCALL call_4_11_0_8(SH2_struct* sh)
{
  SH2shll2(sh,11);
}
#define init_4_12_0_8
void FASTCALL call_4_12_0_8(SH2_struct* sh)
{
  SH2shll2(sh,12);
}
#define init_4_13_0_8
void FASTCALL call_4_13_0_8(SH2_struct* sh)
{
  SH2shll2(sh,13);
}
#define init_4_14_0_8
void FASTCALL call_4_14_0_8(SH2_struct* sh)
{
  SH2shll2(sh,14);
}
#define init_4_15_0_8
void FASTCALL call_4_15_0_8(SH2_struct* sh)
{
  SH2shll2(sh,15);
}
#define init_4_0_0_0
void FASTCALL call_4_0_0_0(SH2_struct* sh)
{
  SH2shll(sh,0);
}
#define init_4_1_0_0
void FASTCALL call_4_1_0_0(SH2_struct* sh)
{
  SH2shll(sh,1);
}
#define init_4_2_0_0
void FASTCALL call_4_2_0_0(SH2_struct* sh)
{
  SH2shll(sh,2);
}
#define init_4_3_0_0
void FASTCALL call_4_3_0_0(SH2_struct* sh)
{
  SH2shll(sh,3);
}
#define init_4_4_0_0
void FASTCALL call_4_4_0_0(SH2_struct* sh)
{
  SH2shll(sh,4);
}
#define init_4_5_0_0
void FASTCALL call_4_5_0_0(SH2_struct* sh)
{
  SH2shll(sh,5);
}
#define init_4_6_0_0
void FASTCALL call_4_6_0_0(SH2_struct* sh)
{
  SH2shll(sh,6);
}
#define init_4_7_0_0
void FASTCALL call_4_7_0_0(SH2_struct* sh)
{
  SH2shll(sh,7);
}
#define init_4_8_0_0
void FASTCALL call_4_8_0_0(SH2_struct* sh)
{
  SH2shll(sh,8);
}
#define init_4_9_0_0
void FASTCALL call_4_9_0_0(SH2_struct* sh)
{
  SH2shll(sh,9);
}
#define init_4_10_0_0
void FASTCALL call_4_10_0_0(SH2_struct* sh)
{
  SH2shll(sh,10);
}
#define init_4_11_0_0
void FASTCALL call_4_11_0_0(SH2_struct* sh)
{
  SH2shll(sh,11);
}
#define init_4_12_0_0
void FASTCALL call_4_12_0_0(SH2_struct* sh)
{
  SH2shll(sh,12);
}
#define init_4_13_0_0
void FASTCALL call_4_13_0_0(SH2_struct* sh)
{
  SH2shll(sh,13);
}
#define init_4_14_0_0
void FASTCALL call_4_14_0_0(SH2_struct* sh)
{
  SH2shll(sh,14);
}
#define init_4_15_0_0
void FASTCALL call_4_15_0_0(SH2_struct* sh)
{
  SH2shll(sh,15);
}
#define init_4_0_2_1
void FASTCALL call_4_0_2_1(SH2_struct* sh)
{
  SH2shar(sh,0);
}
#define init_4_1_2_1
void FASTCALL call_4_1_2_1(SH2_struct* sh)
{
  SH2shar(sh,1);
}
#define init_4_2_2_1
void FASTCALL call_4_2_2_1(SH2_struct* sh)
{
  SH2shar(sh,2);
}
#define init_4_3_2_1
void FASTCALL call_4_3_2_1(SH2_struct* sh)
{
  SH2shar(sh,3);
}
#define init_4_4_2_1
void FASTCALL call_4_4_2_1(SH2_struct* sh)
{
  SH2shar(sh,4);
}
#define init_4_5_2_1
void FASTCALL call_4_5_2_1(SH2_struct* sh)
{
  SH2shar(sh,5);
}
#define init_4_6_2_1
void FASTCALL call_4_6_2_1(SH2_struct* sh)
{
  SH2shar(sh,6);
}
#define init_4_7_2_1
void FASTCALL call_4_7_2_1(SH2_struct* sh)
{
  SH2shar(sh,7);
}
#define init_4_8_2_1
void FASTCALL call_4_8_2_1(SH2_struct* sh)
{
  SH2shar(sh,8);
}
#define init_4_9_2_1
void FASTCALL call_4_9_2_1(SH2_struct* sh)
{
  SH2shar(sh,9);
}
#define init_4_10_2_1
void FASTCALL call_4_10_2_1(SH2_struct* sh)
{
  SH2shar(sh,10);
}
#define init_4_11_2_1
void FASTCALL call_4_11_2_1(SH2_struct* sh)
{
  SH2shar(sh,11);
}
#define init_4_12_2_1
void FASTCALL call_4_12_2_1(SH2_struct* sh)
{
  SH2shar(sh,12);
}
#define init_4_13_2_1
void FASTCALL call_4_13_2_1(SH2_struct* sh)
{
  SH2shar(sh,13);
}
#define init_4_14_2_1
void FASTCALL call_4_14_2_1(SH2_struct* sh)
{
  SH2shar(sh,14);
}
#define init_4_15_2_1
void FASTCALL call_4_15_2_1(SH2_struct* sh)
{
  SH2shar(sh,15);
}
#define init_4_0_2_0
void FASTCALL call_4_0_2_0(SH2_struct* sh)
{
  SH2shal(sh,0);
}
#define init_4_1_2_0
void FASTCALL call_4_1_2_0(SH2_struct* sh)
{
  SH2shal(sh,1);
}
#define init_4_2_2_0
void FASTCALL call_4_2_2_0(SH2_struct* sh)
{
  SH2shal(sh,2);
}
#define init_4_3_2_0
void FASTCALL call_4_3_2_0(SH2_struct* sh)
{
  SH2shal(sh,3);
}
#define init_4_4_2_0
void FASTCALL call_4_4_2_0(SH2_struct* sh)
{
  SH2shal(sh,4);
}
#define init_4_5_2_0
void FASTCALL call_4_5_2_0(SH2_struct* sh)
{
  SH2shal(sh,5);
}
#define init_4_6_2_0
void FASTCALL call_4_6_2_0(SH2_struct* sh)
{
  SH2shal(sh,6);
}
#define init_4_7_2_0
void FASTCALL call_4_7_2_0(SH2_struct* sh)
{
  SH2shal(sh,7);
}
#define init_4_8_2_0
void FASTCALL call_4_8_2_0(SH2_struct* sh)
{
  SH2shal(sh,8);
}
#define init_4_9_2_0
void FASTCALL call_4_9_2_0(SH2_struct* sh)
{
  SH2shal(sh,9);
}
#define init_4_10_2_0
void FASTCALL call_4_10_2_0(SH2_struct* sh)
{
  SH2shal(sh,10);
}
#define init_4_11_2_0
void FASTCALL call_4_11_2_0(SH2_struct* sh)
{
  SH2shal(sh,11);
}
#define init_4_12_2_0
void FASTCALL call_4_12_2_0(SH2_struct* sh)
{
  SH2shal(sh,12);
}
#define init_4_13_2_0
void FASTCALL call_4_13_2_0(SH2_struct* sh)
{
  SH2shal(sh,13);
}
#define init_4_14_2_0
void FASTCALL call_4_14_2_0(SH2_struct* sh)
{
  SH2shal(sh,14);
}
#define init_4_15_2_0
void FASTCALL call_4_15_2_0(SH2_struct* sh)
{
  SH2shal(sh,15);
}
#define init_4_0_0_5
void FASTCALL call_4_0_0_5(SH2_struct* sh)
{
  SH2rotr(sh,0);
}
#define init_4_1_0_5
void FASTCALL call_4_1_0_5(SH2_struct* sh)
{
  SH2rotr(sh,1);
}
#define init_4_2_0_5
void FASTCALL call_4_2_0_5(SH2_struct* sh)
{
  SH2rotr(sh,2);
}
#define init_4_3_0_5
void FASTCALL call_4_3_0_5(SH2_struct* sh)
{
  SH2rotr(sh,3);
}
#define init_4_4_0_5
void FASTCALL call_4_4_0_5(SH2_struct* sh)
{
  SH2rotr(sh,4);
}
#define init_4_5_0_5
void FASTCALL call_4_5_0_5(SH2_struct* sh)
{
  SH2rotr(sh,5);
}
#define init_4_6_0_5
void FASTCALL call_4_6_0_5(SH2_struct* sh)
{
  SH2rotr(sh,6);
}
#define init_4_7_0_5
void FASTCALL call_4_7_0_5(SH2_struct* sh)
{
  SH2rotr(sh,7);
}
#define init_4_8_0_5
void FASTCALL call_4_8_0_5(SH2_struct* sh)
{
  SH2rotr(sh,8);
}
#define init_4_9_0_5
void FASTCALL call_4_9_0_5(SH2_struct* sh)
{
  SH2rotr(sh,9);
}
#define init_4_10_0_5
void FASTCALL call_4_10_0_5(SH2_struct* sh)
{
  SH2rotr(sh,10);
}
#define init_4_11_0_5
void FASTCALL call_4_11_0_5(SH2_struct* sh)
{
  SH2rotr(sh,11);
}
#define init_4_12_0_5
void FASTCALL call_4_12_0_5(SH2_struct* sh)
{
  SH2rotr(sh,12);
}
#define init_4_13_0_5
void FASTCALL call_4_13_0_5(SH2_struct* sh)
{
  SH2rotr(sh,13);
}
#define init_4_14_0_5
void FASTCALL call_4_14_0_5(SH2_struct* sh)
{
  SH2rotr(sh,14);
}
#define init_4_15_0_5
void FASTCALL call_4_15_0_5(SH2_struct* sh)
{
  SH2rotr(sh,15);
}
#define init_4_0_0_4
void FASTCALL call_4_0_0_4(SH2_struct* sh)
{
  SH2rotl(sh,0);
}
#define init_4_1_0_4
void FASTCALL call_4_1_0_4(SH2_struct* sh)
{
  SH2rotl(sh,1);
}
#define init_4_2_0_4
void FASTCALL call_4_2_0_4(SH2_struct* sh)
{
  SH2rotl(sh,2);
}
#define init_4_3_0_4
void FASTCALL call_4_3_0_4(SH2_struct* sh)
{
  SH2rotl(sh,3);
}
#define init_4_4_0_4
void FASTCALL call_4_4_0_4(SH2_struct* sh)
{
  SH2rotl(sh,4);
}
#define init_4_5_0_4
void FASTCALL call_4_5_0_4(SH2_struct* sh)
{
  SH2rotl(sh,5);
}
#define init_4_6_0_4
void FASTCALL call_4_6_0_4(SH2_struct* sh)
{
  SH2rotl(sh,6);
}
#define init_4_7_0_4
void FASTCALL call_4_7_0_4(SH2_struct* sh)
{
  SH2rotl(sh,7);
}
#define init_4_8_0_4
void FASTCALL call_4_8_0_4(SH2_struct* sh)
{
  SH2rotl(sh,8);
}
#define init_4_9_0_4
void FASTCALL call_4_9_0_4(SH2_struct* sh)
{
  SH2rotl(sh,9);
}
#define init_4_10_0_4
void FASTCALL call_4_10_0_4(SH2_struct* sh)
{
  SH2rotl(sh,10);
}
#define init_4_11_0_4
void FASTCALL call_4_11_0_4(SH2_struct* sh)
{
  SH2rotl(sh,11);
}
#define init_4_12_0_4
void FASTCALL call_4_12_0_4(SH2_struct* sh)
{
  SH2rotl(sh,12);
}
#define init_4_13_0_4
void FASTCALL call_4_13_0_4(SH2_struct* sh)
{
  SH2rotl(sh,13);
}
#define init_4_14_0_4
void FASTCALL call_4_14_0_4(SH2_struct* sh)
{
  SH2rotl(sh,14);
}
#define init_4_15_0_4
void FASTCALL call_4_15_0_4(SH2_struct* sh)
{
  SH2rotl(sh,15);
}
#define init_4_0_2_5
void FASTCALL call_4_0_2_5(SH2_struct* sh)
{
  SH2rotcr(sh,0);
}
#define init_4_1_2_5
void FASTCALL call_4_1_2_5(SH2_struct* sh)
{
  SH2rotcr(sh,1);
}
#define init_4_2_2_5
void FASTCALL call_4_2_2_5(SH2_struct* sh)
{
  SH2rotcr(sh,2);
}
#define init_4_3_2_5
void FASTCALL call_4_3_2_5(SH2_struct* sh)
{
  SH2rotcr(sh,3);
}
#define init_4_4_2_5
void FASTCALL call_4_4_2_5(SH2_struct* sh)
{
  SH2rotcr(sh,4);
}
#define init_4_5_2_5
void FASTCALL call_4_5_2_5(SH2_struct* sh)
{
  SH2rotcr(sh,5);
}
#define init_4_6_2_5
void FASTCALL call_4_6_2_5(SH2_struct* sh)
{
  SH2rotcr(sh,6);
}
#define init_4_7_2_5
void FASTCALL call_4_7_2_5(SH2_struct* sh)
{
  SH2rotcr(sh,7);
}
#define init_4_8_2_5
void FASTCALL call_4_8_2_5(SH2_struct* sh)
{
  SH2rotcr(sh,8);
}
#define init_4_9_2_5
void FASTCALL call_4_9_2_5(SH2_struct* sh)
{
  SH2rotcr(sh,9);
}
#define init_4_10_2_5
void FASTCALL call_4_10_2_5(SH2_struct* sh)
{
  SH2rotcr(sh,10);
}
#define init_4_11_2_5
void FASTCALL call_4_11_2_5(SH2_struct* sh)
{
  SH2rotcr(sh,11);
}
#define init_4_12_2_5
void FASTCALL call_4_12_2_5(SH2_struct* sh)
{
  SH2rotcr(sh,12);
}
#define init_4_13_2_5
void FASTCALL call_4_13_2_5(SH2_struct* sh)
{
  SH2rotcr(sh,13);
}
#define init_4_14_2_5
void FASTCALL call_4_14_2_5(SH2_struct* sh)
{
  SH2rotcr(sh,14);
}
#define init_4_15_2_5
void FASTCALL call_4_15_2_5(SH2_struct* sh)
{
  SH2rotcr(sh,15);
}
#define init_4_0_2_4
void FASTCALL call_4_0_2_4(SH2_struct* sh)
{
  SH2rotcl(sh,0);
}
#define init_4_1_2_4
void FASTCALL call_4_1_2_4(SH2_struct* sh)
{
  SH2rotcl(sh,1);
}
#define init_4_2_2_4
void FASTCALL call_4_2_2_4(SH2_struct* sh)
{
  SH2rotcl(sh,2);
}
#define init_4_3_2_4
void FASTCALL call_4_3_2_4(SH2_struct* sh)
{
  SH2rotcl(sh,3);
}
#define init_4_4_2_4
void FASTCALL call_4_4_2_4(SH2_struct* sh)
{
  SH2rotcl(sh,4);
}
#define init_4_5_2_4
void FASTCALL call_4_5_2_4(SH2_struct* sh)
{
  SH2rotcl(sh,5);
}
#define init_4_6_2_4
void FASTCALL call_4_6_2_4(SH2_struct* sh)
{
  SH2rotcl(sh,6);
}
#define init_4_7_2_4
void FASTCALL call_4_7_2_4(SH2_struct* sh)
{
  SH2rotcl(sh,7);
}
#define init_4_8_2_4
void FASTCALL call_4_8_2_4(SH2_struct* sh)
{
  SH2rotcl(sh,8);
}
#define init_4_9_2_4
void FASTCALL call_4_9_2_4(SH2_struct* sh)
{
  SH2rotcl(sh,9);
}
#define init_4_10_2_4
void FASTCALL call_4_10_2_4(SH2_struct* sh)
{
  SH2rotcl(sh,10);
}
#define init_4_11_2_4
void FASTCALL call_4_11_2_4(SH2_struct* sh)
{
  SH2rotcl(sh,11);
}
#define init_4_12_2_4
void FASTCALL call_4_12_2_4(SH2_struct* sh)
{
  SH2rotcl(sh,12);
}
#define init_4_13_2_4
void FASTCALL call_4_13_2_4(SH2_struct* sh)
{
  SH2rotcl(sh,13);
}
#define init_4_14_2_4
void FASTCALL call_4_14_2_4(SH2_struct* sh)
{
  SH2rotcl(sh,14);
}
#define init_4_15_2_4
void FASTCALL call_4_15_2_4(SH2_struct* sh)
{
  SH2rotcl(sh,15);
}
#define init_12_15_0_0
void FASTCALL call_12_15_0_0(SH2_struct* sh)
{
  SH2orm(sh,(0<<4)+0);
}
#define init_12_15_0_1
void FASTCALL call_12_15_0_1(SH2_struct* sh)
{
  SH2orm(sh,(0<<4)+1);
}
#define init_12_15_0_2
void FASTCALL call_12_15_0_2(SH2_struct* sh)
{
  SH2orm(sh,(0<<4)+2);
}
#define init_12_15_0_3
void FASTCALL call_12_15_0_3(SH2_struct* sh)
{
  SH2orm(sh,(0<<4)+3);
}
#define init_12_15_0_4
void FASTCALL call_12_15_0_4(SH2_struct* sh)
{
  SH2orm(sh,(0<<4)+4);
}
#define init_12_15_0_5
void FASTCALL call_12_15_0_5(SH2_struct* sh)
{
  SH2orm(sh,(0<<4)+5);
}
#define init_12_15_0_6
void FASTCALL call_12_15_0_6(SH2_struct* sh)
{
  SH2orm(sh,(0<<4)+6);
}
#define init_12_15_0_7
void FASTCALL call_12_15_0_7(SH2_struct* sh)
{
  SH2orm(sh,(0<<4)+7);
}
#define init_12_15_0_8
void FASTCALL call_12_15_0_8(SH2_struct* sh)
{
  SH2orm(sh,(0<<4)+8);
}
#define init_12_15_0_9
void FASTCALL call_12_15_0_9(SH2_struct* sh)
{
  SH2orm(sh,(0<<4)+9);
}
#define init_12_15_0_10
void FASTCALL call_12_15_0_10(SH2_struct* sh)
{
  SH2orm(sh,(0<<4)+10);
}
#define init_12_15_0_11
void FASTCALL call_12_15_0_11(SH2_struct* sh)
{
  SH2orm(sh,(0<<4)+11);
}
#define init_12_15_0_12
void FASTCALL call_12_15_0_12(SH2_struct* sh)
{
  SH2orm(sh,(0<<4)+12);
}
#define init_12_15_0_13
void FASTCALL call_12_15_0_13(SH2_struct* sh)
{
  SH2orm(sh,(0<<4)+13);
}
#define init_12_15_0_14
void FASTCALL call_12_15_0_14(SH2_struct* sh)
{
  SH2orm(sh,(0<<4)+14);
}
#define init_12_15_0_15
void FASTCALL call_12_15_0_15(SH2_struct* sh)
{
  SH2orm(sh,(0<<4)+15);
}
#define init_12_15_1_0
void FASTCALL call_12_15_1_0(SH2_struct* sh)
{
  SH2orm(sh,(1<<4)+0);
}
#define init_12_15_1_1
void FASTCALL call_12_15_1_1(SH2_struct* sh)
{
  SH2orm(sh,(1<<4)+1);
}
#define init_12_15_1_2
void FASTCALL call_12_15_1_2(SH2_struct* sh)
{
  SH2orm(sh,(1<<4)+2);
}
#define init_12_15_1_3
void FASTCALL call_12_15_1_3(SH2_struct* sh)
{
  SH2orm(sh,(1<<4)+3);
}
#define init_12_15_1_4
void FASTCALL call_12_15_1_4(SH2_struct* sh)
{
  SH2orm(sh,(1<<4)+4);
}
#define init_12_15_1_5
void FASTCALL call_12_15_1_5(SH2_struct* sh)
{
  SH2orm(sh,(1<<4)+5);
}
#define init_12_15_1_6
void FASTCALL call_12_15_1_6(SH2_struct* sh)
{
  SH2orm(sh,(1<<4)+6);
}
#define init_12_15_1_7
void FASTCALL call_12_15_1_7(SH2_struct* sh)
{
  SH2orm(sh,(1<<4)+7);
}
#define init_12_15_1_8
void FASTCALL call_12_15_1_8(SH2_struct* sh)
{
  SH2orm(sh,(1<<4)+8);
}
#define init_12_15_1_9
void FASTCALL call_12_15_1_9(SH2_struct* sh)
{
  SH2orm(sh,(1<<4)+9);
}
#define init_12_15_1_10
void FASTCALL call_12_15_1_10(SH2_struct* sh)
{
  SH2orm(sh,(1<<4)+10);
}
#define init_12_15_1_11
void FASTCALL call_12_15_1_11(SH2_struct* sh)
{
  SH2orm(sh,(1<<4)+11);
}
#define init_12_15_1_12
void FASTCALL call_12_15_1_12(SH2_struct* sh)
{
  SH2orm(sh,(1<<4)+12);
}
#define init_12_15_1_13
void FASTCALL call_12_15_1_13(SH2_struct* sh)
{
  SH2orm(sh,(1<<4)+13);
}
#define init_12_15_1_14
void FASTCALL call_12_15_1_14(SH2_struct* sh)
{
  SH2orm(sh,(1<<4)+14);
}
#define init_12_15_1_15
void FASTCALL call_12_15_1_15(SH2_struct* sh)
{
  SH2orm(sh,(1<<4)+15);
}
#define init_12_15_2_0
void FASTCALL call_12_15_2_0(SH2_struct* sh)
{
  SH2orm(sh,(2<<4)+0);
}
#define init_12_15_2_1
void FASTCALL call_12_15_2_1(SH2_struct* sh)
{
  SH2orm(sh,(2<<4)+1);
}
#define init_12_15_2_2
void FASTCALL call_12_15_2_2(SH2_struct* sh)
{
  SH2orm(sh,(2<<4)+2);
}
#define init_12_15_2_3
void FASTCALL call_12_15_2_3(SH2_struct* sh)
{
  SH2orm(sh,(2<<4)+3);
}
#define init_12_15_2_4
void FASTCALL call_12_15_2_4(SH2_struct* sh)
{
  SH2orm(sh,(2<<4)+4);
}
#define init_12_15_2_5
void FASTCALL call_12_15_2_5(SH2_struct* sh)
{
  SH2orm(sh,(2<<4)+5);
}
#define init_12_15_2_6
void FASTCALL call_12_15_2_6(SH2_struct* sh)
{
  SH2orm(sh,(2<<4)+6);
}
#define init_12_15_2_7
void FASTCALL call_12_15_2_7(SH2_struct* sh)
{
  SH2orm(sh,(2<<4)+7);
}
#define init_12_15_2_8
void FASTCALL call_12_15_2_8(SH2_struct* sh)
{
  SH2orm(sh,(2<<4)+8);
}
#define init_12_15_2_9
void FASTCALL call_12_15_2_9(SH2_struct* sh)
{
  SH2orm(sh,(2<<4)+9);
}
#define init_12_15_2_10
void FASTCALL call_12_15_2_10(SH2_struct* sh)
{
  SH2orm(sh,(2<<4)+10);
}
#define init_12_15_2_11
void FASTCALL call_12_15_2_11(SH2_struct* sh)
{
  SH2orm(sh,(2<<4)+11);
}
#define init_12_15_2_12
void FASTCALL call_12_15_2_12(SH2_struct* sh)
{
  SH2orm(sh,(2<<4)+12);
}
#define init_12_15_2_13
void FASTCALL call_12_15_2_13(SH2_struct* sh)
{
  SH2orm(sh,(2<<4)+13);
}
#define init_12_15_2_14
void FASTCALL call_12_15_2_14(SH2_struct* sh)
{
  SH2orm(sh,(2<<4)+14);
}
#define init_12_15_2_15
void FASTCALL call_12_15_2_15(SH2_struct* sh)
{
  SH2orm(sh,(2<<4)+15);
}
#define init_12_15_3_0
void FASTCALL call_12_15_3_0(SH2_struct* sh)
{
  SH2orm(sh,(3<<4)+0);
}
#define init_12_15_3_1
void FASTCALL call_12_15_3_1(SH2_struct* sh)
{
  SH2orm(sh,(3<<4)+1);
}
#define init_12_15_3_2
void FASTCALL call_12_15_3_2(SH2_struct* sh)
{
  SH2orm(sh,(3<<4)+2);
}
#define init_12_15_3_3
void FASTCALL call_12_15_3_3(SH2_struct* sh)
{
  SH2orm(sh,(3<<4)+3);
}
#define init_12_15_3_4
void FASTCALL call_12_15_3_4(SH2_struct* sh)
{
  SH2orm(sh,(3<<4)+4);
}
#define init_12_15_3_5
void FASTCALL call_12_15_3_5(SH2_struct* sh)
{
  SH2orm(sh,(3<<4)+5);
}
#define init_12_15_3_6
void FASTCALL call_12_15_3_6(SH2_struct* sh)
{
  SH2orm(sh,(3<<4)+6);
}
#define init_12_15_3_7
void FASTCALL call_12_15_3_7(SH2_struct* sh)
{
  SH2orm(sh,(3<<4)+7);
}
#define init_12_15_3_8
void FASTCALL call_12_15_3_8(SH2_struct* sh)
{
  SH2orm(sh,(3<<4)+8);
}
#define init_12_15_3_9
void FASTCALL call_12_15_3_9(SH2_struct* sh)
{
  SH2orm(sh,(3<<4)+9);
}
#define init_12_15_3_10
void FASTCALL call_12_15_3_10(SH2_struct* sh)
{
  SH2orm(sh,(3<<4)+10);
}
#define init_12_15_3_11
void FASTCALL call_12_15_3_11(SH2_struct* sh)
{
  SH2orm(sh,(3<<4)+11);
}
#define init_12_15_3_12
void FASTCALL call_12_15_3_12(SH2_struct* sh)
{
  SH2orm(sh,(3<<4)+12);
}
#define init_12_15_3_13
void FASTCALL call_12_15_3_13(SH2_struct* sh)
{
  SH2orm(sh,(3<<4)+13);
}
#define init_12_15_3_14
void FASTCALL call_12_15_3_14(SH2_struct* sh)
{
  SH2orm(sh,(3<<4)+14);
}
#define init_12_15_3_15
void FASTCALL call_12_15_3_15(SH2_struct* sh)
{
  SH2orm(sh,(3<<4)+15);
}
#define init_12_15_4_0
void FASTCALL call_12_15_4_0(SH2_struct* sh)
{
  SH2orm(sh,(4<<4)+0);
}
#define init_12_15_4_1
void FASTCALL call_12_15_4_1(SH2_struct* sh)
{
  SH2orm(sh,(4<<4)+1);
}
#define init_12_15_4_2
void FASTCALL call_12_15_4_2(SH2_struct* sh)
{
  SH2orm(sh,(4<<4)+2);
}
#define init_12_15_4_3
void FASTCALL call_12_15_4_3(SH2_struct* sh)
{
  SH2orm(sh,(4<<4)+3);
}
#define init_12_15_4_4
void FASTCALL call_12_15_4_4(SH2_struct* sh)
{
  SH2orm(sh,(4<<4)+4);
}
#define init_12_15_4_5
void FASTCALL call_12_15_4_5(SH2_struct* sh)
{
  SH2orm(sh,(4<<4)+5);
}
#define init_12_15_4_6
void FASTCALL call_12_15_4_6(SH2_struct* sh)
{
  SH2orm(sh,(4<<4)+6);
}
#define init_12_15_4_7
void FASTCALL call_12_15_4_7(SH2_struct* sh)
{
  SH2orm(sh,(4<<4)+7);
}
#define init_12_15_4_8
void FASTCALL call_12_15_4_8(SH2_struct* sh)
{
  SH2orm(sh,(4<<4)+8);
}
#define init_12_15_4_9
void FASTCALL call_12_15_4_9(SH2_struct* sh)
{
  SH2orm(sh,(4<<4)+9);
}
#define init_12_15_4_10
void FASTCALL call_12_15_4_10(SH2_struct* sh)
{
  SH2orm(sh,(4<<4)+10);
}
#define init_12_15_4_11
void FASTCALL call_12_15_4_11(SH2_struct* sh)
{
  SH2orm(sh,(4<<4)+11);
}
#define init_12_15_4_12
void FASTCALL call_12_15_4_12(SH2_struct* sh)
{
  SH2orm(sh,(4<<4)+12);
}
#define init_12_15_4_13
void FASTCALL call_12_15_4_13(SH2_struct* sh)
{
  SH2orm(sh,(4<<4)+13);
}
#define init_12_15_4_14
void FASTCALL call_12_15_4_14(SH2_struct* sh)
{
  SH2orm(sh,(4<<4)+14);
}
#define init_12_15_4_15
void FASTCALL call_12_15_4_15(SH2_struct* sh)
{
  SH2orm(sh,(4<<4)+15);
}
#define init_12_15_5_0
void FASTCALL call_12_15_5_0(SH2_struct* sh)
{
  SH2orm(sh,(5<<4)+0);
}
#define init_12_15_5_1
void FASTCALL call_12_15_5_1(SH2_struct* sh)
{
  SH2orm(sh,(5<<4)+1);
}
#define init_12_15_5_2
void FASTCALL call_12_15_5_2(SH2_struct* sh)
{
  SH2orm(sh,(5<<4)+2);
}
#define init_12_15_5_3
void FASTCALL call_12_15_5_3(SH2_struct* sh)
{
  SH2orm(sh,(5<<4)+3);
}
#define init_12_15_5_4
void FASTCALL call_12_15_5_4(SH2_struct* sh)
{
  SH2orm(sh,(5<<4)+4);
}
#define init_12_15_5_5
void FASTCALL call_12_15_5_5(SH2_struct* sh)
{
  SH2orm(sh,(5<<4)+5);
}
#define init_12_15_5_6
void FASTCALL call_12_15_5_6(SH2_struct* sh)
{
  SH2orm(sh,(5<<4)+6);
}
#define init_12_15_5_7
void FASTCALL call_12_15_5_7(SH2_struct* sh)
{
  SH2orm(sh,(5<<4)+7);
}
#define init_12_15_5_8
void FASTCALL call_12_15_5_8(SH2_struct* sh)
{
  SH2orm(sh,(5<<4)+8);
}
#define init_12_15_5_9
void FASTCALL call_12_15_5_9(SH2_struct* sh)
{
  SH2orm(sh,(5<<4)+9);
}
#define init_12_15_5_10
void FASTCALL call_12_15_5_10(SH2_struct* sh)
{
  SH2orm(sh,(5<<4)+10);
}
#define init_12_15_5_11
void FASTCALL call_12_15_5_11(SH2_struct* sh)
{
  SH2orm(sh,(5<<4)+11);
}
#define init_12_15_5_12
void FASTCALL call_12_15_5_12(SH2_struct* sh)
{
  SH2orm(sh,(5<<4)+12);
}
#define init_12_15_5_13
void FASTCALL call_12_15_5_13(SH2_struct* sh)
{
  SH2orm(sh,(5<<4)+13);
}
#define init_12_15_5_14
void FASTCALL call_12_15_5_14(SH2_struct* sh)
{
  SH2orm(sh,(5<<4)+14);
}
#define init_12_15_5_15
void FASTCALL call_12_15_5_15(SH2_struct* sh)
{
  SH2orm(sh,(5<<4)+15);
}
#define init_12_15_6_0
void FASTCALL call_12_15_6_0(SH2_struct* sh)
{
  SH2orm(sh,(6<<4)+0);
}
#define init_12_15_6_1
void FASTCALL call_12_15_6_1(SH2_struct* sh)
{
  SH2orm(sh,(6<<4)+1);
}
#define init_12_15_6_2
void FASTCALL call_12_15_6_2(SH2_struct* sh)
{
  SH2orm(sh,(6<<4)+2);
}
#define init_12_15_6_3
void FASTCALL call_12_15_6_3(SH2_struct* sh)
{
  SH2orm(sh,(6<<4)+3);
}
#define init_12_15_6_4
void FASTCALL call_12_15_6_4(SH2_struct* sh)
{
  SH2orm(sh,(6<<4)+4);
}
#define init_12_15_6_5
void FASTCALL call_12_15_6_5(SH2_struct* sh)
{
  SH2orm(sh,(6<<4)+5);
}
#define init_12_15_6_6
void FASTCALL call_12_15_6_6(SH2_struct* sh)
{
  SH2orm(sh,(6<<4)+6);
}
#define init_12_15_6_7
void FASTCALL call_12_15_6_7(SH2_struct* sh)
{
  SH2orm(sh,(6<<4)+7);
}
#define init_12_15_6_8
void FASTCALL call_12_15_6_8(SH2_struct* sh)
{
  SH2orm(sh,(6<<4)+8);
}
#define init_12_15_6_9
void FASTCALL call_12_15_6_9(SH2_struct* sh)
{
  SH2orm(sh,(6<<4)+9);
}
#define init_12_15_6_10
void FASTCALL call_12_15_6_10(SH2_struct* sh)
{
  SH2orm(sh,(6<<4)+10);
}
#define init_12_15_6_11
void FASTCALL call_12_15_6_11(SH2_struct* sh)
{
  SH2orm(sh,(6<<4)+11);
}
#define init_12_15_6_12
void FASTCALL call_12_15_6_12(SH2_struct* sh)
{
  SH2orm(sh,(6<<4)+12);
}
#define init_12_15_6_13
void FASTCALL call_12_15_6_13(SH2_struct* sh)
{
  SH2orm(sh,(6<<4)+13);
}
#define init_12_15_6_14
void FASTCALL call_12_15_6_14(SH2_struct* sh)
{
  SH2orm(sh,(6<<4)+14);
}
#define init_12_15_6_15
void FASTCALL call_12_15_6_15(SH2_struct* sh)
{
  SH2orm(sh,(6<<4)+15);
}
#define init_12_15_7_0
void FASTCALL call_12_15_7_0(SH2_struct* sh)
{
  SH2orm(sh,(7<<4)+0);
}
#define init_12_15_7_1
void FASTCALL call_12_15_7_1(SH2_struct* sh)
{
  SH2orm(sh,(7<<4)+1);
}
#define init_12_15_7_2
void FASTCALL call_12_15_7_2(SH2_struct* sh)
{
  SH2orm(sh,(7<<4)+2);
}
#define init_12_15_7_3
void FASTCALL call_12_15_7_3(SH2_struct* sh)
{
  SH2orm(sh,(7<<4)+3);
}
#define init_12_15_7_4
void FASTCALL call_12_15_7_4(SH2_struct* sh)
{
  SH2orm(sh,(7<<4)+4);
}
#define init_12_15_7_5
void FASTCALL call_12_15_7_5(SH2_struct* sh)
{
  SH2orm(sh,(7<<4)+5);
}
#define init_12_15_7_6
void FASTCALL call_12_15_7_6(SH2_struct* sh)
{
  SH2orm(sh,(7<<4)+6);
}
#define init_12_15_7_7
void FASTCALL call_12_15_7_7(SH2_struct* sh)
{
  SH2orm(sh,(7<<4)+7);
}
#define init_12_15_7_8
void FASTCALL call_12_15_7_8(SH2_struct* sh)
{
  SH2orm(sh,(7<<4)+8);
}
#define init_12_15_7_9
void FASTCALL call_12_15_7_9(SH2_struct* sh)
{
  SH2orm(sh,(7<<4)+9);
}
#define init_12_15_7_10
void FASTCALL call_12_15_7_10(SH2_struct* sh)
{
  SH2orm(sh,(7<<4)+10);
}
#define init_12_15_7_11
void FASTCALL call_12_15_7_11(SH2_struct* sh)
{
  SH2orm(sh,(7<<4)+11);
}
#define init_12_15_7_12
void FASTCALL call_12_15_7_12(SH2_struct* sh)
{
  SH2orm(sh,(7<<4)+12);
}
#define init_12_15_7_13
void FASTCALL call_12_15_7_13(SH2_struct* sh)
{
  SH2orm(sh,(7<<4)+13);
}
#define init_12_15_7_14
void FASTCALL call_12_15_7_14(SH2_struct* sh)
{
  SH2orm(sh,(7<<4)+14);
}
#define init_12_15_7_15
void FASTCALL call_12_15_7_15(SH2_struct* sh)
{
  SH2orm(sh,(7<<4)+15);
}
#define init_12_15_8_0
void FASTCALL call_12_15_8_0(SH2_struct* sh)
{
  SH2orm(sh,(8<<4)+0);
}
#define init_12_15_8_1
void FASTCALL call_12_15_8_1(SH2_struct* sh)
{
  SH2orm(sh,(8<<4)+1);
}
#define init_12_15_8_2
void FASTCALL call_12_15_8_2(SH2_struct* sh)
{
  SH2orm(sh,(8<<4)+2);
}
#define init_12_15_8_3
void FASTCALL call_12_15_8_3(SH2_struct* sh)
{
  SH2orm(sh,(8<<4)+3);
}
#define init_12_15_8_4
void FASTCALL call_12_15_8_4(SH2_struct* sh)
{
  SH2orm(sh,(8<<4)+4);
}
#define init_12_15_8_5
void FASTCALL call_12_15_8_5(SH2_struct* sh)
{
  SH2orm(sh,(8<<4)+5);
}
#define init_12_15_8_6
void FASTCALL call_12_15_8_6(SH2_struct* sh)
{
  SH2orm(sh,(8<<4)+6);
}
#define init_12_15_8_7
void FASTCALL call_12_15_8_7(SH2_struct* sh)
{
  SH2orm(sh,(8<<4)+7);
}
#define init_12_15_8_8
void FASTCALL call_12_15_8_8(SH2_struct* sh)
{
  SH2orm(sh,(8<<4)+8);
}
#define init_12_15_8_9
void FASTCALL call_12_15_8_9(SH2_struct* sh)
{
  SH2orm(sh,(8<<4)+9);
}
#define init_12_15_8_10
void FASTCALL call_12_15_8_10(SH2_struct* sh)
{
  SH2orm(sh,(8<<4)+10);
}
#define init_12_15_8_11
void FASTCALL call_12_15_8_11(SH2_struct* sh)
{
  SH2orm(sh,(8<<4)+11);
}
#define init_12_15_8_12
void FASTCALL call_12_15_8_12(SH2_struct* sh)
{
  SH2orm(sh,(8<<4)+12);
}
#define init_12_15_8_13
void FASTCALL call_12_15_8_13(SH2_struct* sh)
{
  SH2orm(sh,(8<<4)+13);
}
#define init_12_15_8_14
void FASTCALL call_12_15_8_14(SH2_struct* sh)
{
  SH2orm(sh,(8<<4)+14);
}
#define init_12_15_8_15
void FASTCALL call_12_15_8_15(SH2_struct* sh)
{
  SH2orm(sh,(8<<4)+15);
}
#define init_12_15_9_0
void FASTCALL call_12_15_9_0(SH2_struct* sh)
{
  SH2orm(sh,(9<<4)+0);
}
#define init_12_15_9_1
void FASTCALL call_12_15_9_1(SH2_struct* sh)
{
  SH2orm(sh,(9<<4)+1);
}
#define init_12_15_9_2
void FASTCALL call_12_15_9_2(SH2_struct* sh)
{
  SH2orm(sh,(9<<4)+2);
}
#define init_12_15_9_3
void FASTCALL call_12_15_9_3(SH2_struct* sh)
{
  SH2orm(sh,(9<<4)+3);
}
#define init_12_15_9_4
void FASTCALL call_12_15_9_4(SH2_struct* sh)
{
  SH2orm(sh,(9<<4)+4);
}
#define init_12_15_9_5
void FASTCALL call_12_15_9_5(SH2_struct* sh)
{
  SH2orm(sh,(9<<4)+5);
}
#define init_12_15_9_6
void FASTCALL call_12_15_9_6(SH2_struct* sh)
{
  SH2orm(sh,(9<<4)+6);
}
#define init_12_15_9_7
void FASTCALL call_12_15_9_7(SH2_struct* sh)
{
  SH2orm(sh,(9<<4)+7);
}
#define init_12_15_9_8
void FASTCALL call_12_15_9_8(SH2_struct* sh)
{
  SH2orm(sh,(9<<4)+8);
}
#define init_12_15_9_9
void FASTCALL call_12_15_9_9(SH2_struct* sh)
{
  SH2orm(sh,(9<<4)+9);
}
#define init_12_15_9_10
void FASTCALL call_12_15_9_10(SH2_struct* sh)
{
  SH2orm(sh,(9<<4)+10);
}
#define init_12_15_9_11
void FASTCALL call_12_15_9_11(SH2_struct* sh)
{
  SH2orm(sh,(9<<4)+11);
}
#define init_12_15_9_12
void FASTCALL call_12_15_9_12(SH2_struct* sh)
{
  SH2orm(sh,(9<<4)+12);
}
#define init_12_15_9_13
void FASTCALL call_12_15_9_13(SH2_struct* sh)
{
  SH2orm(sh,(9<<4)+13);
}
#define init_12_15_9_14
void FASTCALL call_12_15_9_14(SH2_struct* sh)
{
  SH2orm(sh,(9<<4)+14);
}
#define init_12_15_9_15
void FASTCALL call_12_15_9_15(SH2_struct* sh)
{
  SH2orm(sh,(9<<4)+15);
}
#define init_12_15_10_0
void FASTCALL call_12_15_10_0(SH2_struct* sh)
{
  SH2orm(sh,(10<<4)+0);
}
#define init_12_15_10_1
void FASTCALL call_12_15_10_1(SH2_struct* sh)
{
  SH2orm(sh,(10<<4)+1);
}
#define init_12_15_10_2
void FASTCALL call_12_15_10_2(SH2_struct* sh)
{
  SH2orm(sh,(10<<4)+2);
}
#define init_12_15_10_3
void FASTCALL call_12_15_10_3(SH2_struct* sh)
{
  SH2orm(sh,(10<<4)+3);
}
#define init_12_15_10_4
void FASTCALL call_12_15_10_4(SH2_struct* sh)
{
  SH2orm(sh,(10<<4)+4);
}
#define init_12_15_10_5
void FASTCALL call_12_15_10_5(SH2_struct* sh)
{
  SH2orm(sh,(10<<4)+5);
}
#define init_12_15_10_6
void FASTCALL call_12_15_10_6(SH2_struct* sh)
{
  SH2orm(sh,(10<<4)+6);
}
#define init_12_15_10_7
void FASTCALL call_12_15_10_7(SH2_struct* sh)
{
  SH2orm(sh,(10<<4)+7);
}
#define init_12_15_10_8
void FASTCALL call_12_15_10_8(SH2_struct* sh)
{
  SH2orm(sh,(10<<4)+8);
}
#define init_12_15_10_9
void FASTCALL call_12_15_10_9(SH2_struct* sh)
{
  SH2orm(sh,(10<<4)+9);
}
#define init_12_15_10_10
void FASTCALL call_12_15_10_10(SH2_struct* sh)
{
  SH2orm(sh,(10<<4)+10);
}
#define init_12_15_10_11
void FASTCALL call_12_15_10_11(SH2_struct* sh)
{
  SH2orm(sh,(10<<4)+11);
}
#define init_12_15_10_12
void FASTCALL call_12_15_10_12(SH2_struct* sh)
{
  SH2orm(sh,(10<<4)+12);
}
#define init_12_15_10_13
void FASTCALL call_12_15_10_13(SH2_struct* sh)
{
  SH2orm(sh,(10<<4)+13);
}
#define init_12_15_10_14
void FASTCALL call_12_15_10_14(SH2_struct* sh)
{
  SH2orm(sh,(10<<4)+14);
}
#define init_12_15_10_15
void FASTCALL call_12_15_10_15(SH2_struct* sh)
{
  SH2orm(sh,(10<<4)+15);
}
#define init_12_15_11_0
void FASTCALL call_12_15_11_0(SH2_struct* sh)
{
  SH2orm(sh,(11<<4)+0);
}
#define init_12_15_11_1
void FASTCALL call_12_15_11_1(SH2_struct* sh)
{
  SH2orm(sh,(11<<4)+1);
}
#define init_12_15_11_2
void FASTCALL call_12_15_11_2(SH2_struct* sh)
{
  SH2orm(sh,(11<<4)+2);
}
#define init_12_15_11_3
void FASTCALL call_12_15_11_3(SH2_struct* sh)
{
  SH2orm(sh,(11<<4)+3);
}
#define init_12_15_11_4
void FASTCALL call_12_15_11_4(SH2_struct* sh)
{
  SH2orm(sh,(11<<4)+4);
}
#define init_12_15_11_5
void FASTCALL call_12_15_11_5(SH2_struct* sh)
{
  SH2orm(sh,(11<<4)+5);
}
#define init_12_15_11_6
void FASTCALL call_12_15_11_6(SH2_struct* sh)
{
  SH2orm(sh,(11<<4)+6);
}
#define init_12_15_11_7
void FASTCALL call_12_15_11_7(SH2_struct* sh)
{
  SH2orm(sh,(11<<4)+7);
}
#define init_12_15_11_8
void FASTCALL call_12_15_11_8(SH2_struct* sh)
{
  SH2orm(sh,(11<<4)+8);
}
#define init_12_15_11_9
void FASTCALL call_12_15_11_9(SH2_struct* sh)
{
  SH2orm(sh,(11<<4)+9);
}
#define init_12_15_11_10
void FASTCALL call_12_15_11_10(SH2_struct* sh)
{
  SH2orm(sh,(11<<4)+10);
}
#define init_12_15_11_11
void FASTCALL call_12_15_11_11(SH2_struct* sh)
{
  SH2orm(sh,(11<<4)+11);
}
#define init_12_15_11_12
void FASTCALL call_12_15_11_12(SH2_struct* sh)
{
  SH2orm(sh,(11<<4)+12);
}
#define init_12_15_11_13
void FASTCALL call_12_15_11_13(SH2_struct* sh)
{
  SH2orm(sh,(11<<4)+13);
}
#define init_12_15_11_14
void FASTCALL call_12_15_11_14(SH2_struct* sh)
{
  SH2orm(sh,(11<<4)+14);
}
#define init_12_15_11_15
void FASTCALL call_12_15_11_15(SH2_struct* sh)
{
  SH2orm(sh,(11<<4)+15);
}
#define init_12_15_12_0
void FASTCALL call_12_15_12_0(SH2_struct* sh)
{
  SH2orm(sh,(12<<4)+0);
}
#define init_12_15_12_1
void FASTCALL call_12_15_12_1(SH2_struct* sh)
{
  SH2orm(sh,(12<<4)+1);
}
#define init_12_15_12_2
void FASTCALL call_12_15_12_2(SH2_struct* sh)
{
  SH2orm(sh,(12<<4)+2);
}
#define init_12_15_12_3
void FASTCALL call_12_15_12_3(SH2_struct* sh)
{
  SH2orm(sh,(12<<4)+3);
}
#define init_12_15_12_4
void FASTCALL call_12_15_12_4(SH2_struct* sh)
{
  SH2orm(sh,(12<<4)+4);
}
#define init_12_15_12_5
void FASTCALL call_12_15_12_5(SH2_struct* sh)
{
  SH2orm(sh,(12<<4)+5);
}
#define init_12_15_12_6
void FASTCALL call_12_15_12_6(SH2_struct* sh)
{
  SH2orm(sh,(12<<4)+6);
}
#define init_12_15_12_7
void FASTCALL call_12_15_12_7(SH2_struct* sh)
{
  SH2orm(sh,(12<<4)+7);
}
#define init_12_15_12_8
void FASTCALL call_12_15_12_8(SH2_struct* sh)
{
  SH2orm(sh,(12<<4)+8);
}
#define init_12_15_12_9
void FASTCALL call_12_15_12_9(SH2_struct* sh)
{
  SH2orm(sh,(12<<4)+9);
}
#define init_12_15_12_10
void FASTCALL call_12_15_12_10(SH2_struct* sh)
{
  SH2orm(sh,(12<<4)+10);
}
#define init_12_15_12_11
void FASTCALL call_12_15_12_11(SH2_struct* sh)
{
  SH2orm(sh,(12<<4)+11);
}
#define init_12_15_12_12
void FASTCALL call_12_15_12_12(SH2_struct* sh)
{
  SH2orm(sh,(12<<4)+12);
}
#define init_12_15_12_13
void FASTCALL call_12_15_12_13(SH2_struct* sh)
{
  SH2orm(sh,(12<<4)+13);
}
#define init_12_15_12_14
void FASTCALL call_12_15_12_14(SH2_struct* sh)
{
  SH2orm(sh,(12<<4)+14);
}
#define init_12_15_12_15
void FASTCALL call_12_15_12_15(SH2_struct* sh)
{
  SH2orm(sh,(12<<4)+15);
}
#define init_12_15_13_0
void FASTCALL call_12_15_13_0(SH2_struct* sh)
{
  SH2orm(sh,(13<<4)+0);
}
#define init_12_15_13_1
void FASTCALL call_12_15_13_1(SH2_struct* sh)
{
  SH2orm(sh,(13<<4)+1);
}
#define init_12_15_13_2
void FASTCALL call_12_15_13_2(SH2_struct* sh)
{
  SH2orm(sh,(13<<4)+2);
}
#define init_12_15_13_3
void FASTCALL call_12_15_13_3(SH2_struct* sh)
{
  SH2orm(sh,(13<<4)+3);
}
#define init_12_15_13_4
void FASTCALL call_12_15_13_4(SH2_struct* sh)
{
  SH2orm(sh,(13<<4)+4);
}
#define init_12_15_13_5
void FASTCALL call_12_15_13_5(SH2_struct* sh)
{
  SH2orm(sh,(13<<4)+5);
}
#define init_12_15_13_6
void FASTCALL call_12_15_13_6(SH2_struct* sh)
{
  SH2orm(sh,(13<<4)+6);
}
#define init_12_15_13_7
void FASTCALL call_12_15_13_7(SH2_struct* sh)
{
  SH2orm(sh,(13<<4)+7);
}
#define init_12_15_13_8
void FASTCALL call_12_15_13_8(SH2_struct* sh)
{
  SH2orm(sh,(13<<4)+8);
}
#define init_12_15_13_9
void FASTCALL call_12_15_13_9(SH2_struct* sh)
{
  SH2orm(sh,(13<<4)+9);
}
#define init_12_15_13_10
void FASTCALL call_12_15_13_10(SH2_struct* sh)
{
  SH2orm(sh,(13<<4)+10);
}
#define init_12_15_13_11
void FASTCALL call_12_15_13_11(SH2_struct* sh)
{
  SH2orm(sh,(13<<4)+11);
}
#define init_12_15_13_12
void FASTCALL call_12_15_13_12(SH2_struct* sh)
{
  SH2orm(sh,(13<<4)+12);
}
#define init_12_15_13_13
void FASTCALL call_12_15_13_13(SH2_struct* sh)
{
  SH2orm(sh,(13<<4)+13);
}
#define init_12_15_13_14
void FASTCALL call_12_15_13_14(SH2_struct* sh)
{
  SH2orm(sh,(13<<4)+14);
}
#define init_12_15_13_15
void FASTCALL call_12_15_13_15(SH2_struct* sh)
{
  SH2orm(sh,(13<<4)+15);
}
#define init_12_15_14_0
void FASTCALL call_12_15_14_0(SH2_struct* sh)
{
  SH2orm(sh,(14<<4)+0);
}
#define init_12_15_14_1
void FASTCALL call_12_15_14_1(SH2_struct* sh)
{
  SH2orm(sh,(14<<4)+1);
}
#define init_12_15_14_2
void FASTCALL call_12_15_14_2(SH2_struct* sh)
{
  SH2orm(sh,(14<<4)+2);
}
#define init_12_15_14_3
void FASTCALL call_12_15_14_3(SH2_struct* sh)
{
  SH2orm(sh,(14<<4)+3);
}
#define init_12_15_14_4
void FASTCALL call_12_15_14_4(SH2_struct* sh)
{
  SH2orm(sh,(14<<4)+4);
}
#define init_12_15_14_5
void FASTCALL call_12_15_14_5(SH2_struct* sh)
{
  SH2orm(sh,(14<<4)+5);
}
#define init_12_15_14_6
void FASTCALL call_12_15_14_6(SH2_struct* sh)
{
  SH2orm(sh,(14<<4)+6);
}
#define init_12_15_14_7
void FASTCALL call_12_15_14_7(SH2_struct* sh)
{
  SH2orm(sh,(14<<4)+7);
}
#define init_12_15_14_8
void FASTCALL call_12_15_14_8(SH2_struct* sh)
{
  SH2orm(sh,(14<<4)+8);
}
#define init_12_15_14_9
void FASTCALL call_12_15_14_9(SH2_struct* sh)
{
  SH2orm(sh,(14<<4)+9);
}
#define init_12_15_14_10
void FASTCALL call_12_15_14_10(SH2_struct* sh)
{
  SH2orm(sh,(14<<4)+10);
}
#define init_12_15_14_11
void FASTCALL call_12_15_14_11(SH2_struct* sh)
{
  SH2orm(sh,(14<<4)+11);
}
#define init_12_15_14_12
void FASTCALL call_12_15_14_12(SH2_struct* sh)
{
  SH2orm(sh,(14<<4)+12);
}
#define init_12_15_14_13
void FASTCALL call_12_15_14_13(SH2_struct* sh)
{
  SH2orm(sh,(14<<4)+13);
}
#define init_12_15_14_14
void FASTCALL call_12_15_14_14(SH2_struct* sh)
{
  SH2orm(sh,(14<<4)+14);
}
#define init_12_15_14_15
void FASTCALL call_12_15_14_15(SH2_struct* sh)
{
  SH2orm(sh,(14<<4)+15);
}
#define init_12_15_15_0
void FASTCALL call_12_15_15_0(SH2_struct* sh)
{
  SH2orm(sh,(15<<4)+0);
}
#define init_12_15_15_1
void FASTCALL call_12_15_15_1(SH2_struct* sh)
{
  SH2orm(sh,(15<<4)+1);
}
#define init_12_15_15_2
void FASTCALL call_12_15_15_2(SH2_struct* sh)
{
  SH2orm(sh,(15<<4)+2);
}
#define init_12_15_15_3
void FASTCALL call_12_15_15_3(SH2_struct* sh)
{
  SH2orm(sh,(15<<4)+3);
}
#define init_12_15_15_4
void FASTCALL call_12_15_15_4(SH2_struct* sh)
{
  SH2orm(sh,(15<<4)+4);
}
#define init_12_15_15_5
void FASTCALL call_12_15_15_5(SH2_struct* sh)
{
  SH2orm(sh,(15<<4)+5);
}
#define init_12_15_15_6
void FASTCALL call_12_15_15_6(SH2_struct* sh)
{
  SH2orm(sh,(15<<4)+6);
}
#define init_12_15_15_7
void FASTCALL call_12_15_15_7(SH2_struct* sh)
{
  SH2orm(sh,(15<<4)+7);
}
#define init_12_15_15_8
void FASTCALL call_12_15_15_8(SH2_struct* sh)
{
  SH2orm(sh,(15<<4)+8);
}
#define init_12_15_15_9
void FASTCALL call_12_15_15_9(SH2_struct* sh)
{
  SH2orm(sh,(15<<4)+9);
}
#define init_12_15_15_10
void FASTCALL call_12_15_15_10(SH2_struct* sh)
{
  SH2orm(sh,(15<<4)+10);
}
#define init_12_15_15_11
void FASTCALL call_12_15_15_11(SH2_struct* sh)
{
  SH2orm(sh,(15<<4)+11);
}
#define init_12_15_15_12
void FASTCALL call_12_15_15_12(SH2_struct* sh)
{
  SH2orm(sh,(15<<4)+12);
}
#define init_12_15_15_13
void FASTCALL call_12_15_15_13(SH2_struct* sh)
{
  SH2orm(sh,(15<<4)+13);
}
#define init_12_15_15_14
void FASTCALL call_12_15_15_14(SH2_struct* sh)
{
  SH2orm(sh,(15<<4)+14);
}
#define init_12_15_15_15
void FASTCALL call_12_15_15_15(SH2_struct* sh)
{
  SH2orm(sh,(15<<4)+15);
}
#define init_12_11_0_0
void FASTCALL call_12_11_0_0(SH2_struct* sh)
{
  SH2ori(sh,(0<<4)+0);
}
#define init_12_11_0_1
void FASTCALL call_12_11_0_1(SH2_struct* sh)
{
  SH2ori(sh,(0<<4)+1);
}
#define init_12_11_0_2
void FASTCALL call_12_11_0_2(SH2_struct* sh)
{
  SH2ori(sh,(0<<4)+2);
}
#define init_12_11_0_3
void FASTCALL call_12_11_0_3(SH2_struct* sh)
{
  SH2ori(sh,(0<<4)+3);
}
#define init_12_11_0_4
void FASTCALL call_12_11_0_4(SH2_struct* sh)
{
  SH2ori(sh,(0<<4)+4);
}
#define init_12_11_0_5
void FASTCALL call_12_11_0_5(SH2_struct* sh)
{
  SH2ori(sh,(0<<4)+5);
}
#define init_12_11_0_6
void FASTCALL call_12_11_0_6(SH2_struct* sh)
{
  SH2ori(sh,(0<<4)+6);
}
#define init_12_11_0_7
void FASTCALL call_12_11_0_7(SH2_struct* sh)
{
  SH2ori(sh,(0<<4)+7);
}
#define init_12_11_0_8
void FASTCALL call_12_11_0_8(SH2_struct* sh)
{
  SH2ori(sh,(0<<4)+8);
}
#define init_12_11_0_9
void FASTCALL call_12_11_0_9(SH2_struct* sh)
{
  SH2ori(sh,(0<<4)+9);
}
#define init_12_11_0_10
void FASTCALL call_12_11_0_10(SH2_struct* sh)
{
  SH2ori(sh,(0<<4)+10);
}
#define init_12_11_0_11
void FASTCALL call_12_11_0_11(SH2_struct* sh)
{
  SH2ori(sh,(0<<4)+11);
}
#define init_12_11_0_12
void FASTCALL call_12_11_0_12(SH2_struct* sh)
{
  SH2ori(sh,(0<<4)+12);
}
#define init_12_11_0_13
void FASTCALL call_12_11_0_13(SH2_struct* sh)
{
  SH2ori(sh,(0<<4)+13);
}
#define init_12_11_0_14
void FASTCALL call_12_11_0_14(SH2_struct* sh)
{
  SH2ori(sh,(0<<4)+14);
}
#define init_12_11_0_15
void FASTCALL call_12_11_0_15(SH2_struct* sh)
{
  SH2ori(sh,(0<<4)+15);
}
#define init_12_11_1_0
void FASTCALL call_12_11_1_0(SH2_struct* sh)
{
  SH2ori(sh,(1<<4)+0);
}
#define init_12_11_1_1
void FASTCALL call_12_11_1_1(SH2_struct* sh)
{
  SH2ori(sh,(1<<4)+1);
}
#define init_12_11_1_2
void FASTCALL call_12_11_1_2(SH2_struct* sh)
{
  SH2ori(sh,(1<<4)+2);
}
#define init_12_11_1_3
void FASTCALL call_12_11_1_3(SH2_struct* sh)
{
  SH2ori(sh,(1<<4)+3);
}
#define init_12_11_1_4
void FASTCALL call_12_11_1_4(SH2_struct* sh)
{
  SH2ori(sh,(1<<4)+4);
}
#define init_12_11_1_5
void FASTCALL call_12_11_1_5(SH2_struct* sh)
{
  SH2ori(sh,(1<<4)+5);
}
#define init_12_11_1_6
void FASTCALL call_12_11_1_6(SH2_struct* sh)
{
  SH2ori(sh,(1<<4)+6);
}
#define init_12_11_1_7
void FASTCALL call_12_11_1_7(SH2_struct* sh)
{
  SH2ori(sh,(1<<4)+7);
}
#define init_12_11_1_8
void FASTCALL call_12_11_1_8(SH2_struct* sh)
{
  SH2ori(sh,(1<<4)+8);
}
#define init_12_11_1_9
void FASTCALL call_12_11_1_9(SH2_struct* sh)
{
  SH2ori(sh,(1<<4)+9);
}
#define init_12_11_1_10
void FASTCALL call_12_11_1_10(SH2_struct* sh)
{
  SH2ori(sh,(1<<4)+10);
}
#define init_12_11_1_11
void FASTCALL call_12_11_1_11(SH2_struct* sh)
{
  SH2ori(sh,(1<<4)+11);
}
#define init_12_11_1_12
void FASTCALL call_12_11_1_12(SH2_struct* sh)
{
  SH2ori(sh,(1<<4)+12);
}
#define init_12_11_1_13
void FASTCALL call_12_11_1_13(SH2_struct* sh)
{
  SH2ori(sh,(1<<4)+13);
}
#define init_12_11_1_14
void FASTCALL call_12_11_1_14(SH2_struct* sh)
{
  SH2ori(sh,(1<<4)+14);
}
#define init_12_11_1_15
void FASTCALL call_12_11_1_15(SH2_struct* sh)
{
  SH2ori(sh,(1<<4)+15);
}
#define init_12_11_2_0
void FASTCALL call_12_11_2_0(SH2_struct* sh)
{
  SH2ori(sh,(2<<4)+0);
}
#define init_12_11_2_1
void FASTCALL call_12_11_2_1(SH2_struct* sh)
{
  SH2ori(sh,(2<<4)+1);
}
#define init_12_11_2_2
void FASTCALL call_12_11_2_2(SH2_struct* sh)
{
  SH2ori(sh,(2<<4)+2);
}
#define init_12_11_2_3
void FASTCALL call_12_11_2_3(SH2_struct* sh)
{
  SH2ori(sh,(2<<4)+3);
}
#define init_12_11_2_4
void FASTCALL call_12_11_2_4(SH2_struct* sh)
{
  SH2ori(sh,(2<<4)+4);
}
#define init_12_11_2_5
void FASTCALL call_12_11_2_5(SH2_struct* sh)
{
  SH2ori(sh,(2<<4)+5);
}
#define init_12_11_2_6
void FASTCALL call_12_11_2_6(SH2_struct* sh)
{
  SH2ori(sh,(2<<4)+6);
}
#define init_12_11_2_7
void FASTCALL call_12_11_2_7(SH2_struct* sh)
{
  SH2ori(sh,(2<<4)+7);
}
#define init_12_11_2_8
void FASTCALL call_12_11_2_8(SH2_struct* sh)
{
  SH2ori(sh,(2<<4)+8);
}
#define init_12_11_2_9
void FASTCALL call_12_11_2_9(SH2_struct* sh)
{
  SH2ori(sh,(2<<4)+9);
}
#define init_12_11_2_10
void FASTCALL call_12_11_2_10(SH2_struct* sh)
{
  SH2ori(sh,(2<<4)+10);
}
#define init_12_11_2_11
void FASTCALL call_12_11_2_11(SH2_struct* sh)
{
  SH2ori(sh,(2<<4)+11);
}
#define init_12_11_2_12
void FASTCALL call_12_11_2_12(SH2_struct* sh)
{
  SH2ori(sh,(2<<4)+12);
}
#define init_12_11_2_13
void FASTCALL call_12_11_2_13(SH2_struct* sh)
{
  SH2ori(sh,(2<<4)+13);
}
#define init_12_11_2_14
void FASTCALL call_12_11_2_14(SH2_struct* sh)
{
  SH2ori(sh,(2<<4)+14);
}
#define init_12_11_2_15
void FASTCALL call_12_11_2_15(SH2_struct* sh)
{
  SH2ori(sh,(2<<4)+15);
}
#define init_12_11_3_0
void FASTCALL call_12_11_3_0(SH2_struct* sh)
{
  SH2ori(sh,(3<<4)+0);
}
#define init_12_11_3_1
void FASTCALL call_12_11_3_1(SH2_struct* sh)
{
  SH2ori(sh,(3<<4)+1);
}
#define init_12_11_3_2
void FASTCALL call_12_11_3_2(SH2_struct* sh)
{
  SH2ori(sh,(3<<4)+2);
}
#define init_12_11_3_3
void FASTCALL call_12_11_3_3(SH2_struct* sh)
{
  SH2ori(sh,(3<<4)+3);
}
#define init_12_11_3_4
void FASTCALL call_12_11_3_4(SH2_struct* sh)
{
  SH2ori(sh,(3<<4)+4);
}
#define init_12_11_3_5
void FASTCALL call_12_11_3_5(SH2_struct* sh)
{
  SH2ori(sh,(3<<4)+5);
}
#define init_12_11_3_6
void FASTCALL call_12_11_3_6(SH2_struct* sh)
{
  SH2ori(sh,(3<<4)+6);
}
#define init_12_11_3_7
void FASTCALL call_12_11_3_7(SH2_struct* sh)
{
  SH2ori(sh,(3<<4)+7);
}
#define init_12_11_3_8
void FASTCALL call_12_11_3_8(SH2_struct* sh)
{
  SH2ori(sh,(3<<4)+8);
}
#define init_12_11_3_9
void FASTCALL call_12_11_3_9(SH2_struct* sh)
{
  SH2ori(sh,(3<<4)+9);
}
#define init_12_11_3_10
void FASTCALL call_12_11_3_10(SH2_struct* sh)
{
  SH2ori(sh,(3<<4)+10);
}
#define init_12_11_3_11
void FASTCALL call_12_11_3_11(SH2_struct* sh)
{
  SH2ori(sh,(3<<4)+11);
}
#define init_12_11_3_12
void FASTCALL call_12_11_3_12(SH2_struct* sh)
{
  SH2ori(sh,(3<<4)+12);
}
#define init_12_11_3_13
void FASTCALL call_12_11_3_13(SH2_struct* sh)
{
  SH2ori(sh,(3<<4)+13);
}
#define init_12_11_3_14
void FASTCALL call_12_11_3_14(SH2_struct* sh)
{
  SH2ori(sh,(3<<4)+14);
}
#define init_12_11_3_15
void FASTCALL call_12_11_3_15(SH2_struct* sh)
{
  SH2ori(sh,(3<<4)+15);
}
#define init_12_11_4_0
void FASTCALL call_12_11_4_0(SH2_struct* sh)
{
  SH2ori(sh,(4<<4)+0);
}
#define init_12_11_4_1
void FASTCALL call_12_11_4_1(SH2_struct* sh)
{
  SH2ori(sh,(4<<4)+1);
}
#define init_12_11_4_2
void FASTCALL call_12_11_4_2(SH2_struct* sh)
{
  SH2ori(sh,(4<<4)+2);
}
#define init_12_11_4_3
void FASTCALL call_12_11_4_3(SH2_struct* sh)
{
  SH2ori(sh,(4<<4)+3);
}
#define init_12_11_4_4
void FASTCALL call_12_11_4_4(SH2_struct* sh)
{
  SH2ori(sh,(4<<4)+4);
}
#define init_12_11_4_5
void FASTCALL call_12_11_4_5(SH2_struct* sh)
{
  SH2ori(sh,(4<<4)+5);
}
#define init_12_11_4_6
void FASTCALL call_12_11_4_6(SH2_struct* sh)
{
  SH2ori(sh,(4<<4)+6);
}
#define init_12_11_4_7
void FASTCALL call_12_11_4_7(SH2_struct* sh)
{
  SH2ori(sh,(4<<4)+7);
}
#define init_12_11_4_8
void FASTCALL call_12_11_4_8(SH2_struct* sh)
{
  SH2ori(sh,(4<<4)+8);
}
#define init_12_11_4_9
void FASTCALL call_12_11_4_9(SH2_struct* sh)
{
  SH2ori(sh,(4<<4)+9);
}
#define init_12_11_4_10
void FASTCALL call_12_11_4_10(SH2_struct* sh)
{
  SH2ori(sh,(4<<4)+10);
}
#define init_12_11_4_11
void FASTCALL call_12_11_4_11(SH2_struct* sh)
{
  SH2ori(sh,(4<<4)+11);
}
#define init_12_11_4_12
void FASTCALL call_12_11_4_12(SH2_struct* sh)
{
  SH2ori(sh,(4<<4)+12);
}
#define init_12_11_4_13
void FASTCALL call_12_11_4_13(SH2_struct* sh)
{
  SH2ori(sh,(4<<4)+13);
}
#define init_12_11_4_14
void FASTCALL call_12_11_4_14(SH2_struct* sh)
{
  SH2ori(sh,(4<<4)+14);
}
#define init_12_11_4_15
void FASTCALL call_12_11_4_15(SH2_struct* sh)
{
  SH2ori(sh,(4<<4)+15);
}
#define init_12_11_5_0
void FASTCALL call_12_11_5_0(SH2_struct* sh)
{
  SH2ori(sh,(5<<4)+0);
}
#define init_12_11_5_1
void FASTCALL call_12_11_5_1(SH2_struct* sh)
{
  SH2ori(sh,(5<<4)+1);
}
#define init_12_11_5_2
void FASTCALL call_12_11_5_2(SH2_struct* sh)
{
  SH2ori(sh,(5<<4)+2);
}
#define init_12_11_5_3
void FASTCALL call_12_11_5_3(SH2_struct* sh)
{
  SH2ori(sh,(5<<4)+3);
}
#define init_12_11_5_4
void FASTCALL call_12_11_5_4(SH2_struct* sh)
{
  SH2ori(sh,(5<<4)+4);
}
#define init_12_11_5_5
void FASTCALL call_12_11_5_5(SH2_struct* sh)
{
  SH2ori(sh,(5<<4)+5);
}
#define init_12_11_5_6
void FASTCALL call_12_11_5_6(SH2_struct* sh)
{
  SH2ori(sh,(5<<4)+6);
}
#define init_12_11_5_7
void FASTCALL call_12_11_5_7(SH2_struct* sh)
{
  SH2ori(sh,(5<<4)+7);
}
#define init_12_11_5_8
void FASTCALL call_12_11_5_8(SH2_struct* sh)
{
  SH2ori(sh,(5<<4)+8);
}
#define init_12_11_5_9
void FASTCALL call_12_11_5_9(SH2_struct* sh)
{
  SH2ori(sh,(5<<4)+9);
}
#define init_12_11_5_10
void FASTCALL call_12_11_5_10(SH2_struct* sh)
{
  SH2ori(sh,(5<<4)+10);
}
#define init_12_11_5_11
void FASTCALL call_12_11_5_11(SH2_struct* sh)
{
  SH2ori(sh,(5<<4)+11);
}
#define init_12_11_5_12
void FASTCALL call_12_11_5_12(SH2_struct* sh)
{
  SH2ori(sh,(5<<4)+12);
}
#define init_12_11_5_13
void FASTCALL call_12_11_5_13(SH2_struct* sh)
{
  SH2ori(sh,(5<<4)+13);
}
#define init_12_11_5_14
void FASTCALL call_12_11_5_14(SH2_struct* sh)
{
  SH2ori(sh,(5<<4)+14);
}
#define init_12_11_5_15
void FASTCALL call_12_11_5_15(SH2_struct* sh)
{
  SH2ori(sh,(5<<4)+15);
}
#define init_12_11_6_0
void FASTCALL call_12_11_6_0(SH2_struct* sh)
{
  SH2ori(sh,(6<<4)+0);
}
#define init_12_11_6_1
void FASTCALL call_12_11_6_1(SH2_struct* sh)
{
  SH2ori(sh,(6<<4)+1);
}
#define init_12_11_6_2
void FASTCALL call_12_11_6_2(SH2_struct* sh)
{
  SH2ori(sh,(6<<4)+2);
}
#define init_12_11_6_3
void FASTCALL call_12_11_6_3(SH2_struct* sh)
{
  SH2ori(sh,(6<<4)+3);
}
#define init_12_11_6_4
void FASTCALL call_12_11_6_4(SH2_struct* sh)
{
  SH2ori(sh,(6<<4)+4);
}
#define init_12_11_6_5
void FASTCALL call_12_11_6_5(SH2_struct* sh)
{
  SH2ori(sh,(6<<4)+5);
}
#define init_12_11_6_6
void FASTCALL call_12_11_6_6(SH2_struct* sh)
{
  SH2ori(sh,(6<<4)+6);
}
#define init_12_11_6_7
void FASTCALL call_12_11_6_7(SH2_struct* sh)
{
  SH2ori(sh,(6<<4)+7);
}
#define init_12_11_6_8
void FASTCALL call_12_11_6_8(SH2_struct* sh)
{
  SH2ori(sh,(6<<4)+8);
}
#define init_12_11_6_9
void FASTCALL call_12_11_6_9(SH2_struct* sh)
{
  SH2ori(sh,(6<<4)+9);
}
#define init_12_11_6_10
void FASTCALL call_12_11_6_10(SH2_struct* sh)
{
  SH2ori(sh,(6<<4)+10);
}
#define init_12_11_6_11
void FASTCALL call_12_11_6_11(SH2_struct* sh)
{
  SH2ori(sh,(6<<4)+11);
}
#define init_12_11_6_12
void FASTCALL call_12_11_6_12(SH2_struct* sh)
{
  SH2ori(sh,(6<<4)+12);
}
#define init_12_11_6_13
void FASTCALL call_12_11_6_13(SH2_struct* sh)
{
  SH2ori(sh,(6<<4)+13);
}
#define init_12_11_6_14
void FASTCALL call_12_11_6_14(SH2_struct* sh)
{
  SH2ori(sh,(6<<4)+14);
}
#define init_12_11_6_15
void FASTCALL call_12_11_6_15(SH2_struct* sh)
{
  SH2ori(sh,(6<<4)+15);
}
#define init_12_11_7_0
void FASTCALL call_12_11_7_0(SH2_struct* sh)
{
  SH2ori(sh,(7<<4)+0);
}
#define init_12_11_7_1
void FASTCALL call_12_11_7_1(SH2_struct* sh)
{
  SH2ori(sh,(7<<4)+1);
}
#define init_12_11_7_2
void FASTCALL call_12_11_7_2(SH2_struct* sh)
{
  SH2ori(sh,(7<<4)+2);
}
#define init_12_11_7_3
void FASTCALL call_12_11_7_3(SH2_struct* sh)
{
  SH2ori(sh,(7<<4)+3);
}
#define init_12_11_7_4
void FASTCALL call_12_11_7_4(SH2_struct* sh)
{
  SH2ori(sh,(7<<4)+4);
}
#define init_12_11_7_5
void FASTCALL call_12_11_7_5(SH2_struct* sh)
{
  SH2ori(sh,(7<<4)+5);
}
#define init_12_11_7_6
void FASTCALL call_12_11_7_6(SH2_struct* sh)
{
  SH2ori(sh,(7<<4)+6);
}
#define init_12_11_7_7
void FASTCALL call_12_11_7_7(SH2_struct* sh)
{
  SH2ori(sh,(7<<4)+7);
}
#define init_12_11_7_8
void FASTCALL call_12_11_7_8(SH2_struct* sh)
{
  SH2ori(sh,(7<<4)+8);
}
#define init_12_11_7_9
void FASTCALL call_12_11_7_9(SH2_struct* sh)
{
  SH2ori(sh,(7<<4)+9);
}
#define init_12_11_7_10
void FASTCALL call_12_11_7_10(SH2_struct* sh)
{
  SH2ori(sh,(7<<4)+10);
}
#define init_12_11_7_11
void FASTCALL call_12_11_7_11(SH2_struct* sh)
{
  SH2ori(sh,(7<<4)+11);
}
#define init_12_11_7_12
void FASTCALL call_12_11_7_12(SH2_struct* sh)
{
  SH2ori(sh,(7<<4)+12);
}
#define init_12_11_7_13
void FASTCALL call_12_11_7_13(SH2_struct* sh)
{
  SH2ori(sh,(7<<4)+13);
}
#define init_12_11_7_14
void FASTCALL call_12_11_7_14(SH2_struct* sh)
{
  SH2ori(sh,(7<<4)+14);
}
#define init_12_11_7_15
void FASTCALL call_12_11_7_15(SH2_struct* sh)
{
  SH2ori(sh,(7<<4)+15);
}
#define init_12_11_8_0
void FASTCALL call_12_11_8_0(SH2_struct* sh)
{
  SH2ori(sh,(8<<4)+0);
}
#define init_12_11_8_1
void FASTCALL call_12_11_8_1(SH2_struct* sh)
{
  SH2ori(sh,(8<<4)+1);
}
#define init_12_11_8_2
void FASTCALL call_12_11_8_2(SH2_struct* sh)
{
  SH2ori(sh,(8<<4)+2);
}
#define init_12_11_8_3
void FASTCALL call_12_11_8_3(SH2_struct* sh)
{
  SH2ori(sh,(8<<4)+3);
}
#define init_12_11_8_4
void FASTCALL call_12_11_8_4(SH2_struct* sh)
{
  SH2ori(sh,(8<<4)+4);
}
#define init_12_11_8_5
void FASTCALL call_12_11_8_5(SH2_struct* sh)
{
  SH2ori(sh,(8<<4)+5);
}
#define init_12_11_8_6
void FASTCALL call_12_11_8_6(SH2_struct* sh)
{
  SH2ori(sh,(8<<4)+6);
}
#define init_12_11_8_7
void FASTCALL call_12_11_8_7(SH2_struct* sh)
{
  SH2ori(sh,(8<<4)+7);
}
#define init_12_11_8_8
void FASTCALL call_12_11_8_8(SH2_struct* sh)
{
  SH2ori(sh,(8<<4)+8);
}
#define init_12_11_8_9
void FASTCALL call_12_11_8_9(SH2_struct* sh)
{
  SH2ori(sh,(8<<4)+9);
}
#define init_12_11_8_10
void FASTCALL call_12_11_8_10(SH2_struct* sh)
{
  SH2ori(sh,(8<<4)+10);
}
#define init_12_11_8_11
void FASTCALL call_12_11_8_11(SH2_struct* sh)
{
  SH2ori(sh,(8<<4)+11);
}
#define init_12_11_8_12
void FASTCALL call_12_11_8_12(SH2_struct* sh)
{
  SH2ori(sh,(8<<4)+12);
}
#define init_12_11_8_13
void FASTCALL call_12_11_8_13(SH2_struct* sh)
{
  SH2ori(sh,(8<<4)+13);
}
#define init_12_11_8_14
void FASTCALL call_12_11_8_14(SH2_struct* sh)
{
  SH2ori(sh,(8<<4)+14);
}
#define init_12_11_8_15
void FASTCALL call_12_11_8_15(SH2_struct* sh)
{
  SH2ori(sh,(8<<4)+15);
}
#define init_12_11_9_0
void FASTCALL call_12_11_9_0(SH2_struct* sh)
{
  SH2ori(sh,(9<<4)+0);
}
#define init_12_11_9_1
void FASTCALL call_12_11_9_1(SH2_struct* sh)
{
  SH2ori(sh,(9<<4)+1);
}
#define init_12_11_9_2
void FASTCALL call_12_11_9_2(SH2_struct* sh)
{
  SH2ori(sh,(9<<4)+2);
}
#define init_12_11_9_3
void FASTCALL call_12_11_9_3(SH2_struct* sh)
{
  SH2ori(sh,(9<<4)+3);
}
#define init_12_11_9_4
void FASTCALL call_12_11_9_4(SH2_struct* sh)
{
  SH2ori(sh,(9<<4)+4);
}
#define init_12_11_9_5
void FASTCALL call_12_11_9_5(SH2_struct* sh)
{
  SH2ori(sh,(9<<4)+5);
}
#define init_12_11_9_6
void FASTCALL call_12_11_9_6(SH2_struct* sh)
{
  SH2ori(sh,(9<<4)+6);
}
#define init_12_11_9_7
void FASTCALL call_12_11_9_7(SH2_struct* sh)
{
  SH2ori(sh,(9<<4)+7);
}
#define init_12_11_9_8
void FASTCALL call_12_11_9_8(SH2_struct* sh)
{
  SH2ori(sh,(9<<4)+8);
}
#define init_12_11_9_9
void FASTCALL call_12_11_9_9(SH2_struct* sh)
{
  SH2ori(sh,(9<<4)+9);
}
#define init_12_11_9_10
void FASTCALL call_12_11_9_10(SH2_struct* sh)
{
  SH2ori(sh,(9<<4)+10);
}
#define init_12_11_9_11
void FASTCALL call_12_11_9_11(SH2_struct* sh)
{
  SH2ori(sh,(9<<4)+11);
}
#define init_12_11_9_12
void FASTCALL call_12_11_9_12(SH2_struct* sh)
{
  SH2ori(sh,(9<<4)+12);
}
#define init_12_11_9_13
void FASTCALL call_12_11_9_13(SH2_struct* sh)
{
  SH2ori(sh,(9<<4)+13);
}
#define init_12_11_9_14
void FASTCALL call_12_11_9_14(SH2_struct* sh)
{
  SH2ori(sh,(9<<4)+14);
}
#define init_12_11_9_15
void FASTCALL call_12_11_9_15(SH2_struct* sh)
{
  SH2ori(sh,(9<<4)+15);
}
#define init_12_11_10_0
void FASTCALL call_12_11_10_0(SH2_struct* sh)
{
  SH2ori(sh,(10<<4)+0);
}
#define init_12_11_10_1
void FASTCALL call_12_11_10_1(SH2_struct* sh)
{
  SH2ori(sh,(10<<4)+1);
}
#define init_12_11_10_2
void FASTCALL call_12_11_10_2(SH2_struct* sh)
{
  SH2ori(sh,(10<<4)+2);
}
#define init_12_11_10_3
void FASTCALL call_12_11_10_3(SH2_struct* sh)
{
  SH2ori(sh,(10<<4)+3);
}
#define init_12_11_10_4
void FASTCALL call_12_11_10_4(SH2_struct* sh)
{
  SH2ori(sh,(10<<4)+4);
}
#define init_12_11_10_5
void FASTCALL call_12_11_10_5(SH2_struct* sh)
{
  SH2ori(sh,(10<<4)+5);
}
#define init_12_11_10_6
void FASTCALL call_12_11_10_6(SH2_struct* sh)
{
  SH2ori(sh,(10<<4)+6);
}
#define init_12_11_10_7
void FASTCALL call_12_11_10_7(SH2_struct* sh)
{
  SH2ori(sh,(10<<4)+7);
}
#define init_12_11_10_8
void FASTCALL call_12_11_10_8(SH2_struct* sh)
{
  SH2ori(sh,(10<<4)+8);
}
#define init_12_11_10_9
void FASTCALL call_12_11_10_9(SH2_struct* sh)
{
  SH2ori(sh,(10<<4)+9);
}
#define init_12_11_10_10
void FASTCALL call_12_11_10_10(SH2_struct* sh)
{
  SH2ori(sh,(10<<4)+10);
}
#define init_12_11_10_11
void FASTCALL call_12_11_10_11(SH2_struct* sh)
{
  SH2ori(sh,(10<<4)+11);
}
#define init_12_11_10_12
void FASTCALL call_12_11_10_12(SH2_struct* sh)
{
  SH2ori(sh,(10<<4)+12);
}
#define init_12_11_10_13
void FASTCALL call_12_11_10_13(SH2_struct* sh)
{
  SH2ori(sh,(10<<4)+13);
}
#define init_12_11_10_14
void FASTCALL call_12_11_10_14(SH2_struct* sh)
{
  SH2ori(sh,(10<<4)+14);
}
#define init_12_11_10_15
void FASTCALL call_12_11_10_15(SH2_struct* sh)
{
  SH2ori(sh,(10<<4)+15);
}
#define init_12_11_11_0
void FASTCALL call_12_11_11_0(SH2_struct* sh)
{
  SH2ori(sh,(11<<4)+0);
}
#define init_12_11_11_1
void FASTCALL call_12_11_11_1(SH2_struct* sh)
{
  SH2ori(sh,(11<<4)+1);
}
#define init_12_11_11_2
void FASTCALL call_12_11_11_2(SH2_struct* sh)
{
  SH2ori(sh,(11<<4)+2);
}
#define init_12_11_11_3
void FASTCALL call_12_11_11_3(SH2_struct* sh)
{
  SH2ori(sh,(11<<4)+3);
}
#define init_12_11_11_4
void FASTCALL call_12_11_11_4(SH2_struct* sh)
{
  SH2ori(sh,(11<<4)+4);
}
#define init_12_11_11_5
void FASTCALL call_12_11_11_5(SH2_struct* sh)
{
  SH2ori(sh,(11<<4)+5);
}
#define init_12_11_11_6
void FASTCALL call_12_11_11_6(SH2_struct* sh)
{
  SH2ori(sh,(11<<4)+6);
}
#define init_12_11_11_7
void FASTCALL call_12_11_11_7(SH2_struct* sh)
{
  SH2ori(sh,(11<<4)+7);
}
#define init_12_11_11_8
void FASTCALL call_12_11_11_8(SH2_struct* sh)
{
  SH2ori(sh,(11<<4)+8);
}
#define init_12_11_11_9
void FASTCALL call_12_11_11_9(SH2_struct* sh)
{
  SH2ori(sh,(11<<4)+9);
}
#define init_12_11_11_10
void FASTCALL call_12_11_11_10(SH2_struct* sh)
{
  SH2ori(sh,(11<<4)+10);
}
#define init_12_11_11_11
void FASTCALL call_12_11_11_11(SH2_struct* sh)
{
  SH2ori(sh,(11<<4)+11);
}
#define init_12_11_11_12
void FASTCALL call_12_11_11_12(SH2_struct* sh)
{
  SH2ori(sh,(11<<4)+12);
}
#define init_12_11_11_13
void FASTCALL call_12_11_11_13(SH2_struct* sh)
{
  SH2ori(sh,(11<<4)+13);
}
#define init_12_11_11_14
void FASTCALL call_12_11_11_14(SH2_struct* sh)
{
  SH2ori(sh,(11<<4)+14);
}
#define init_12_11_11_15
void FASTCALL call_12_11_11_15(SH2_struct* sh)
{
  SH2ori(sh,(11<<4)+15);
}
#define init_12_11_12_0
void FASTCALL call_12_11_12_0(SH2_struct* sh)
{
  SH2ori(sh,(12<<4)+0);
}
#define init_12_11_12_1
void FASTCALL call_12_11_12_1(SH2_struct* sh)
{
  SH2ori(sh,(12<<4)+1);
}
#define init_12_11_12_2
void FASTCALL call_12_11_12_2(SH2_struct* sh)
{
  SH2ori(sh,(12<<4)+2);
}
#define init_12_11_12_3
void FASTCALL call_12_11_12_3(SH2_struct* sh)
{
  SH2ori(sh,(12<<4)+3);
}
#define init_12_11_12_4
void FASTCALL call_12_11_12_4(SH2_struct* sh)
{
  SH2ori(sh,(12<<4)+4);
}
#define init_12_11_12_5
void FASTCALL call_12_11_12_5(SH2_struct* sh)
{
  SH2ori(sh,(12<<4)+5);
}
#define init_12_11_12_6
void FASTCALL call_12_11_12_6(SH2_struct* sh)
{
  SH2ori(sh,(12<<4)+6);
}
#define init_12_11_12_7
void FASTCALL call_12_11_12_7(SH2_struct* sh)
{
  SH2ori(sh,(12<<4)+7);
}
#define init_12_11_12_8
void FASTCALL call_12_11_12_8(SH2_struct* sh)
{
  SH2ori(sh,(12<<4)+8);
}
#define init_12_11_12_9
void FASTCALL call_12_11_12_9(SH2_struct* sh)
{
  SH2ori(sh,(12<<4)+9);
}
#define init_12_11_12_10
void FASTCALL call_12_11_12_10(SH2_struct* sh)
{
  SH2ori(sh,(12<<4)+10);
}
#define init_12_11_12_11
void FASTCALL call_12_11_12_11(SH2_struct* sh)
{
  SH2ori(sh,(12<<4)+11);
}
#define init_12_11_12_12
void FASTCALL call_12_11_12_12(SH2_struct* sh)
{
  SH2ori(sh,(12<<4)+12);
}
#define init_12_11_12_13
void FASTCALL call_12_11_12_13(SH2_struct* sh)
{
  SH2ori(sh,(12<<4)+13);
}
#define init_12_11_12_14
void FASTCALL call_12_11_12_14(SH2_struct* sh)
{
  SH2ori(sh,(12<<4)+14);
}
#define init_12_11_12_15
void FASTCALL call_12_11_12_15(SH2_struct* sh)
{
  SH2ori(sh,(12<<4)+15);
}
#define init_12_11_13_0
void FASTCALL call_12_11_13_0(SH2_struct* sh)
{
  SH2ori(sh,(13<<4)+0);
}
#define init_12_11_13_1
void FASTCALL call_12_11_13_1(SH2_struct* sh)
{
  SH2ori(sh,(13<<4)+1);
}
#define init_12_11_13_2
void FASTCALL call_12_11_13_2(SH2_struct* sh)
{
  SH2ori(sh,(13<<4)+2);
}
#define init_12_11_13_3
void FASTCALL call_12_11_13_3(SH2_struct* sh)
{
  SH2ori(sh,(13<<4)+3);
}
#define init_12_11_13_4
void FASTCALL call_12_11_13_4(SH2_struct* sh)
{
  SH2ori(sh,(13<<4)+4);
}
#define init_12_11_13_5
void FASTCALL call_12_11_13_5(SH2_struct* sh)
{
  SH2ori(sh,(13<<4)+5);
}
#define init_12_11_13_6
void FASTCALL call_12_11_13_6(SH2_struct* sh)
{
  SH2ori(sh,(13<<4)+6);
}
#define init_12_11_13_7
void FASTCALL call_12_11_13_7(SH2_struct* sh)
{
  SH2ori(sh,(13<<4)+7);
}
#define init_12_11_13_8
void FASTCALL call_12_11_13_8(SH2_struct* sh)
{
  SH2ori(sh,(13<<4)+8);
}
#define init_12_11_13_9
void FASTCALL call_12_11_13_9(SH2_struct* sh)
{
  SH2ori(sh,(13<<4)+9);
}
#define init_12_11_13_10
void FASTCALL call_12_11_13_10(SH2_struct* sh)
{
  SH2ori(sh,(13<<4)+10);
}
#define init_12_11_13_11
void FASTCALL call_12_11_13_11(SH2_struct* sh)
{
  SH2ori(sh,(13<<4)+11);
}
#define init_12_11_13_12
void FASTCALL call_12_11_13_12(SH2_struct* sh)
{
  SH2ori(sh,(13<<4)+12);
}
#define init_12_11_13_13
void FASTCALL call_12_11_13_13(SH2_struct* sh)
{
  SH2ori(sh,(13<<4)+13);
}
#define init_12_11_13_14
void FASTCALL call_12_11_13_14(SH2_struct* sh)
{
  SH2ori(sh,(13<<4)+14);
}
#define init_12_11_13_15
void FASTCALL call_12_11_13_15(SH2_struct* sh)
{
  SH2ori(sh,(13<<4)+15);
}
#define init_12_11_14_0
void FASTCALL call_12_11_14_0(SH2_struct* sh)
{
  SH2ori(sh,(14<<4)+0);
}
#define init_12_11_14_1
void FASTCALL call_12_11_14_1(SH2_struct* sh)
{
  SH2ori(sh,(14<<4)+1);
}
#define init_12_11_14_2
void FASTCALL call_12_11_14_2(SH2_struct* sh)
{
  SH2ori(sh,(14<<4)+2);
}
#define init_12_11_14_3
void FASTCALL call_12_11_14_3(SH2_struct* sh)
{
  SH2ori(sh,(14<<4)+3);
}
#define init_12_11_14_4
void FASTCALL call_12_11_14_4(SH2_struct* sh)
{
  SH2ori(sh,(14<<4)+4);
}
#define init_12_11_14_5
void FASTCALL call_12_11_14_5(SH2_struct* sh)
{
  SH2ori(sh,(14<<4)+5);
}
#define init_12_11_14_6
void FASTCALL call_12_11_14_6(SH2_struct* sh)
{
  SH2ori(sh,(14<<4)+6);
}
#define init_12_11_14_7
void FASTCALL call_12_11_14_7(SH2_struct* sh)
{
  SH2ori(sh,(14<<4)+7);
}
#define init_12_11_14_8
void FASTCALL call_12_11_14_8(SH2_struct* sh)
{
  SH2ori(sh,(14<<4)+8);
}
#define init_12_11_14_9
void FASTCALL call_12_11_14_9(SH2_struct* sh)
{
  SH2ori(sh,(14<<4)+9);
}
#define init_12_11_14_10
void FASTCALL call_12_11_14_10(SH2_struct* sh)
{
  SH2ori(sh,(14<<4)+10);
}
#define init_12_11_14_11
void FASTCALL call_12_11_14_11(SH2_struct* sh)
{
  SH2ori(sh,(14<<4)+11);
}
#define init_12_11_14_12
void FASTCALL call_12_11_14_12(SH2_struct* sh)
{
  SH2ori(sh,(14<<4)+12);
}
#define init_12_11_14_13
void FASTCALL call_12_11_14_13(SH2_struct* sh)
{
  SH2ori(sh,(14<<4)+13);
}
#define init_12_11_14_14
void FASTCALL call_12_11_14_14(SH2_struct* sh)
{
  SH2ori(sh,(14<<4)+14);
}
#define init_12_11_14_15
void FASTCALL call_12_11_14_15(SH2_struct* sh)
{
  SH2ori(sh,(14<<4)+15);
}
#define init_12_11_15_0
void FASTCALL call_12_11_15_0(SH2_struct* sh)
{
  SH2ori(sh,(15<<4)+0);
}
#define init_12_11_15_1
void FASTCALL call_12_11_15_1(SH2_struct* sh)
{
  SH2ori(sh,(15<<4)+1);
}
#define init_12_11_15_2
void FASTCALL call_12_11_15_2(SH2_struct* sh)
{
  SH2ori(sh,(15<<4)+2);
}
#define init_12_11_15_3
void FASTCALL call_12_11_15_3(SH2_struct* sh)
{
  SH2ori(sh,(15<<4)+3);
}
#define init_12_11_15_4
void FASTCALL call_12_11_15_4(SH2_struct* sh)
{
  SH2ori(sh,(15<<4)+4);
}
#define init_12_11_15_5
void FASTCALL call_12_11_15_5(SH2_struct* sh)
{
  SH2ori(sh,(15<<4)+5);
}
#define init_12_11_15_6
void FASTCALL call_12_11_15_6(SH2_struct* sh)
{
  SH2ori(sh,(15<<4)+6);
}
#define init_12_11_15_7
void FASTCALL call_12_11_15_7(SH2_struct* sh)
{
  SH2ori(sh,(15<<4)+7);
}
#define init_12_11_15_8
void FASTCALL call_12_11_15_8(SH2_struct* sh)
{
  SH2ori(sh,(15<<4)+8);
}
#define init_12_11_15_9
void FASTCALL call_12_11_15_9(SH2_struct* sh)
{
  SH2ori(sh,(15<<4)+9);
}
#define init_12_11_15_10
void FASTCALL call_12_11_15_10(SH2_struct* sh)
{
  SH2ori(sh,(15<<4)+10);
}
#define init_12_11_15_11
void FASTCALL call_12_11_15_11(SH2_struct* sh)
{
  SH2ori(sh,(15<<4)+11);
}
#define init_12_11_15_12
void FASTCALL call_12_11_15_12(SH2_struct* sh)
{
  SH2ori(sh,(15<<4)+12);
}
#define init_12_11_15_13
void FASTCALL call_12_11_15_13(SH2_struct* sh)
{
  SH2ori(sh,(15<<4)+13);
}
#define init_12_11_15_14
void FASTCALL call_12_11_15_14(SH2_struct* sh)
{
  SH2ori(sh,(15<<4)+14);
}
#define init_12_11_15_15
void FASTCALL call_12_11_15_15(SH2_struct* sh)
{
  SH2ori(sh,(15<<4)+15);
}
#define init_2_0_0_11
void FASTCALL call_2_0_0_11(SH2_struct* sh)
{
  SH2or(sh,0,0);
}
#define init_2_0_1_11
void FASTCALL call_2_0_1_11(SH2_struct* sh)
{
  SH2or(sh,0,1);
}
#define init_2_0_2_11
void FASTCALL call_2_0_2_11(SH2_struct* sh)
{
  SH2or(sh,0,2);
}
#define init_2_0_3_11
void FASTCALL call_2_0_3_11(SH2_struct* sh)
{
  SH2or(sh,0,3);
}
#define init_2_0_4_11
void FASTCALL call_2_0_4_11(SH2_struct* sh)
{
  SH2or(sh,0,4);
}
#define init_2_0_5_11
void FASTCALL call_2_0_5_11(SH2_struct* sh)
{
  SH2or(sh,0,5);
}
#define init_2_0_6_11
void FASTCALL call_2_0_6_11(SH2_struct* sh)
{
  SH2or(sh,0,6);
}
#define init_2_0_7_11
void FASTCALL call_2_0_7_11(SH2_struct* sh)
{
  SH2or(sh,0,7);
}
#define init_2_0_8_11
void FASTCALL call_2_0_8_11(SH2_struct* sh)
{
  SH2or(sh,0,8);
}
#define init_2_0_9_11
void FASTCALL call_2_0_9_11(SH2_struct* sh)
{
  SH2or(sh,0,9);
}
#define init_2_0_10_11
void FASTCALL call_2_0_10_11(SH2_struct* sh)
{
  SH2or(sh,0,10);
}
#define init_2_0_11_11
void FASTCALL call_2_0_11_11(SH2_struct* sh)
{
  SH2or(sh,0,11);
}
#define init_2_0_12_11
void FASTCALL call_2_0_12_11(SH2_struct* sh)
{
  SH2or(sh,0,12);
}
#define init_2_0_13_11
void FASTCALL call_2_0_13_11(SH2_struct* sh)
{
  SH2or(sh,0,13);
}
#define init_2_0_14_11
void FASTCALL call_2_0_14_11(SH2_struct* sh)
{
  SH2or(sh,0,14);
}
#define init_2_0_15_11
void FASTCALL call_2_0_15_11(SH2_struct* sh)
{
  SH2or(sh,0,15);
}
#define init_2_1_0_11
void FASTCALL call_2_1_0_11(SH2_struct* sh)
{
  SH2or(sh,1,0);
}
#define init_2_1_1_11
void FASTCALL call_2_1_1_11(SH2_struct* sh)
{
  SH2or(sh,1,1);
}
#define init_2_1_2_11
void FASTCALL call_2_1_2_11(SH2_struct* sh)
{
  SH2or(sh,1,2);
}
#define init_2_1_3_11
void FASTCALL call_2_1_3_11(SH2_struct* sh)
{
  SH2or(sh,1,3);
}
#define init_2_1_4_11
void FASTCALL call_2_1_4_11(SH2_struct* sh)
{
  SH2or(sh,1,4);
}
#define init_2_1_5_11
void FASTCALL call_2_1_5_11(SH2_struct* sh)
{
  SH2or(sh,1,5);
}
#define init_2_1_6_11
void FASTCALL call_2_1_6_11(SH2_struct* sh)
{
  SH2or(sh,1,6);
}
#define init_2_1_7_11
void FASTCALL call_2_1_7_11(SH2_struct* sh)
{
  SH2or(sh,1,7);
}
#define init_2_1_8_11
void FASTCALL call_2_1_8_11(SH2_struct* sh)
{
  SH2or(sh,1,8);
}
#define init_2_1_9_11
void FASTCALL call_2_1_9_11(SH2_struct* sh)
{
  SH2or(sh,1,9);
}
#define init_2_1_10_11
void FASTCALL call_2_1_10_11(SH2_struct* sh)
{
  SH2or(sh,1,10);
}
#define init_2_1_11_11
void FASTCALL call_2_1_11_11(SH2_struct* sh)
{
  SH2or(sh,1,11);
}
#define init_2_1_12_11
void FASTCALL call_2_1_12_11(SH2_struct* sh)
{
  SH2or(sh,1,12);
}
#define init_2_1_13_11
void FASTCALL call_2_1_13_11(SH2_struct* sh)
{
  SH2or(sh,1,13);
}
#define init_2_1_14_11
void FASTCALL call_2_1_14_11(SH2_struct* sh)
{
  SH2or(sh,1,14);
}
#define init_2_1_15_11
void FASTCALL call_2_1_15_11(SH2_struct* sh)
{
  SH2or(sh,1,15);
}
#define init_2_2_0_11
void FASTCALL call_2_2_0_11(SH2_struct* sh)
{
  SH2or(sh,2,0);
}
#define init_2_2_1_11
void FASTCALL call_2_2_1_11(SH2_struct* sh)
{
  SH2or(sh,2,1);
}
#define init_2_2_2_11
void FASTCALL call_2_2_2_11(SH2_struct* sh)
{
  SH2or(sh,2,2);
}
#define init_2_2_3_11
void FASTCALL call_2_2_3_11(SH2_struct* sh)
{
  SH2or(sh,2,3);
}
#define init_2_2_4_11
void FASTCALL call_2_2_4_11(SH2_struct* sh)
{
  SH2or(sh,2,4);
}
#define init_2_2_5_11
void FASTCALL call_2_2_5_11(SH2_struct* sh)
{
  SH2or(sh,2,5);
}
#define init_2_2_6_11
void FASTCALL call_2_2_6_11(SH2_struct* sh)
{
  SH2or(sh,2,6);
}
#define init_2_2_7_11
void FASTCALL call_2_2_7_11(SH2_struct* sh)
{
  SH2or(sh,2,7);
}
#define init_2_2_8_11
void FASTCALL call_2_2_8_11(SH2_struct* sh)
{
  SH2or(sh,2,8);
}
#define init_2_2_9_11
void FASTCALL call_2_2_9_11(SH2_struct* sh)
{
  SH2or(sh,2,9);
}
#define init_2_2_10_11
void FASTCALL call_2_2_10_11(SH2_struct* sh)
{
  SH2or(sh,2,10);
}
#define init_2_2_11_11
void FASTCALL call_2_2_11_11(SH2_struct* sh)
{
  SH2or(sh,2,11);
}
#define init_2_2_12_11
void FASTCALL call_2_2_12_11(SH2_struct* sh)
{
  SH2or(sh,2,12);
}
#define init_2_2_13_11
void FASTCALL call_2_2_13_11(SH2_struct* sh)
{
  SH2or(sh,2,13);
}
#define init_2_2_14_11
void FASTCALL call_2_2_14_11(SH2_struct* sh)
{
  SH2or(sh,2,14);
}
#define init_2_2_15_11
void FASTCALL call_2_2_15_11(SH2_struct* sh)
{
  SH2or(sh,2,15);
}
#define init_2_3_0_11
void FASTCALL call_2_3_0_11(SH2_struct* sh)
{
  SH2or(sh,3,0);
}
#define init_2_3_1_11
void FASTCALL call_2_3_1_11(SH2_struct* sh)
{
  SH2or(sh,3,1);
}
#define init_2_3_2_11
void FASTCALL call_2_3_2_11(SH2_struct* sh)
{
  SH2or(sh,3,2);
}
#define init_2_3_3_11
void FASTCALL call_2_3_3_11(SH2_struct* sh)
{
  SH2or(sh,3,3);
}
#define init_2_3_4_11
void FASTCALL call_2_3_4_11(SH2_struct* sh)
{
  SH2or(sh,3,4);
}
#define init_2_3_5_11
void FASTCALL call_2_3_5_11(SH2_struct* sh)
{
  SH2or(sh,3,5);
}
#define init_2_3_6_11
void FASTCALL call_2_3_6_11(SH2_struct* sh)
{
  SH2or(sh,3,6);
}
#define init_2_3_7_11
void FASTCALL call_2_3_7_11(SH2_struct* sh)
{
  SH2or(sh,3,7);
}
#define init_2_3_8_11
void FASTCALL call_2_3_8_11(SH2_struct* sh)
{
  SH2or(sh,3,8);
}
#define init_2_3_9_11
void FASTCALL call_2_3_9_11(SH2_struct* sh)
{
  SH2or(sh,3,9);
}
#define init_2_3_10_11
void FASTCALL call_2_3_10_11(SH2_struct* sh)
{
  SH2or(sh,3,10);
}
#define init_2_3_11_11
void FASTCALL call_2_3_11_11(SH2_struct* sh)
{
  SH2or(sh,3,11);
}
#define init_2_3_12_11
void FASTCALL call_2_3_12_11(SH2_struct* sh)
{
  SH2or(sh,3,12);
}
#define init_2_3_13_11
void FASTCALL call_2_3_13_11(SH2_struct* sh)
{
  SH2or(sh,3,13);
}
#define init_2_3_14_11
void FASTCALL call_2_3_14_11(SH2_struct* sh)
{
  SH2or(sh,3,14);
}
#define init_2_3_15_11
void FASTCALL call_2_3_15_11(SH2_struct* sh)
{
  SH2or(sh,3,15);
}
#define init_2_4_0_11
void FASTCALL call_2_4_0_11(SH2_struct* sh)
{
  SH2or(sh,4,0);
}
#define init_2_4_1_11
void FASTCALL call_2_4_1_11(SH2_struct* sh)
{
  SH2or(sh,4,1);
}
#define init_2_4_2_11
void FASTCALL call_2_4_2_11(SH2_struct* sh)
{
  SH2or(sh,4,2);
}
#define init_2_4_3_11
void FASTCALL call_2_4_3_11(SH2_struct* sh)
{
  SH2or(sh,4,3);
}
#define init_2_4_4_11
void FASTCALL call_2_4_4_11(SH2_struct* sh)
{
  SH2or(sh,4,4);
}
#define init_2_4_5_11
void FASTCALL call_2_4_5_11(SH2_struct* sh)
{
  SH2or(sh,4,5);
}
#define init_2_4_6_11
void FASTCALL call_2_4_6_11(SH2_struct* sh)
{
  SH2or(sh,4,6);
}
#define init_2_4_7_11
void FASTCALL call_2_4_7_11(SH2_struct* sh)
{
  SH2or(sh,4,7);
}
#define init_2_4_8_11
void FASTCALL call_2_4_8_11(SH2_struct* sh)
{
  SH2or(sh,4,8);
}
#define init_2_4_9_11
void FASTCALL call_2_4_9_11(SH2_struct* sh)
{
  SH2or(sh,4,9);
}
#define init_2_4_10_11
void FASTCALL call_2_4_10_11(SH2_struct* sh)
{
  SH2or(sh,4,10);
}
#define init_2_4_11_11
void FASTCALL call_2_4_11_11(SH2_struct* sh)
{
  SH2or(sh,4,11);
}
#define init_2_4_12_11
void FASTCALL call_2_4_12_11(SH2_struct* sh)
{
  SH2or(sh,4,12);
}
#define init_2_4_13_11
void FASTCALL call_2_4_13_11(SH2_struct* sh)
{
  SH2or(sh,4,13);
}
#define init_2_4_14_11
void FASTCALL call_2_4_14_11(SH2_struct* sh)
{
  SH2or(sh,4,14);
}
#define init_2_4_15_11
void FASTCALL call_2_4_15_11(SH2_struct* sh)
{
  SH2or(sh,4,15);
}
#define init_2_5_0_11
void FASTCALL call_2_5_0_11(SH2_struct* sh)
{
  SH2or(sh,5,0);
}
#define init_2_5_1_11
void FASTCALL call_2_5_1_11(SH2_struct* sh)
{
  SH2or(sh,5,1);
}
#define init_2_5_2_11
void FASTCALL call_2_5_2_11(SH2_struct* sh)
{
  SH2or(sh,5,2);
}
#define init_2_5_3_11
void FASTCALL call_2_5_3_11(SH2_struct* sh)
{
  SH2or(sh,5,3);
}
#define init_2_5_4_11
void FASTCALL call_2_5_4_11(SH2_struct* sh)
{
  SH2or(sh,5,4);
}
#define init_2_5_5_11
void FASTCALL call_2_5_5_11(SH2_struct* sh)
{
  SH2or(sh,5,5);
}
#define init_2_5_6_11
void FASTCALL call_2_5_6_11(SH2_struct* sh)
{
  SH2or(sh,5,6);
}
#define init_2_5_7_11
void FASTCALL call_2_5_7_11(SH2_struct* sh)
{
  SH2or(sh,5,7);
}
#define init_2_5_8_11
void FASTCALL call_2_5_8_11(SH2_struct* sh)
{
  SH2or(sh,5,8);
}
#define init_2_5_9_11
void FASTCALL call_2_5_9_11(SH2_struct* sh)
{
  SH2or(sh,5,9);
}
#define init_2_5_10_11
void FASTCALL call_2_5_10_11(SH2_struct* sh)
{
  SH2or(sh,5,10);
}
#define init_2_5_11_11
void FASTCALL call_2_5_11_11(SH2_struct* sh)
{
  SH2or(sh,5,11);
}
#define init_2_5_12_11
void FASTCALL call_2_5_12_11(SH2_struct* sh)
{
  SH2or(sh,5,12);
}
#define init_2_5_13_11
void FASTCALL call_2_5_13_11(SH2_struct* sh)
{
  SH2or(sh,5,13);
}
#define init_2_5_14_11
void FASTCALL call_2_5_14_11(SH2_struct* sh)
{
  SH2or(sh,5,14);
}
#define init_2_5_15_11
void FASTCALL call_2_5_15_11(SH2_struct* sh)
{
  SH2or(sh,5,15);
}
#define init_2_6_0_11
void FASTCALL call_2_6_0_11(SH2_struct* sh)
{
  SH2or(sh,6,0);
}
#define init_2_6_1_11
void FASTCALL call_2_6_1_11(SH2_struct* sh)
{
  SH2or(sh,6,1);
}
#define init_2_6_2_11
void FASTCALL call_2_6_2_11(SH2_struct* sh)
{
  SH2or(sh,6,2);
}
#define init_2_6_3_11
void FASTCALL call_2_6_3_11(SH2_struct* sh)
{
  SH2or(sh,6,3);
}
#define init_2_6_4_11
void FASTCALL call_2_6_4_11(SH2_struct* sh)
{
  SH2or(sh,6,4);
}
#define init_2_6_5_11
void FASTCALL call_2_6_5_11(SH2_struct* sh)
{
  SH2or(sh,6,5);
}
#define init_2_6_6_11
void FASTCALL call_2_6_6_11(SH2_struct* sh)
{
  SH2or(sh,6,6);
}
#define init_2_6_7_11
void FASTCALL call_2_6_7_11(SH2_struct* sh)
{
  SH2or(sh,6,7);
}
#define init_2_6_8_11
void FASTCALL call_2_6_8_11(SH2_struct* sh)
{
  SH2or(sh,6,8);
}
#define init_2_6_9_11
void FASTCALL call_2_6_9_11(SH2_struct* sh)
{
  SH2or(sh,6,9);
}
#define init_2_6_10_11
void FASTCALL call_2_6_10_11(SH2_struct* sh)
{
  SH2or(sh,6,10);
}
#define init_2_6_11_11
void FASTCALL call_2_6_11_11(SH2_struct* sh)
{
  SH2or(sh,6,11);
}
#define init_2_6_12_11
void FASTCALL call_2_6_12_11(SH2_struct* sh)
{
  SH2or(sh,6,12);
}
#define init_2_6_13_11
void FASTCALL call_2_6_13_11(SH2_struct* sh)
{
  SH2or(sh,6,13);
}
#define init_2_6_14_11
void FASTCALL call_2_6_14_11(SH2_struct* sh)
{
  SH2or(sh,6,14);
}
#define init_2_6_15_11
void FASTCALL call_2_6_15_11(SH2_struct* sh)
{
  SH2or(sh,6,15);
}
#define init_2_7_0_11
void FASTCALL call_2_7_0_11(SH2_struct* sh)
{
  SH2or(sh,7,0);
}
#define init_2_7_1_11
void FASTCALL call_2_7_1_11(SH2_struct* sh)
{
  SH2or(sh,7,1);
}
#define init_2_7_2_11
void FASTCALL call_2_7_2_11(SH2_struct* sh)
{
  SH2or(sh,7,2);
}
#define init_2_7_3_11
void FASTCALL call_2_7_3_11(SH2_struct* sh)
{
  SH2or(sh,7,3);
}
#define init_2_7_4_11
void FASTCALL call_2_7_4_11(SH2_struct* sh)
{
  SH2or(sh,7,4);
}
#define init_2_7_5_11
void FASTCALL call_2_7_5_11(SH2_struct* sh)
{
  SH2or(sh,7,5);
}
#define init_2_7_6_11
void FASTCALL call_2_7_6_11(SH2_struct* sh)
{
  SH2or(sh,7,6);
}
#define init_2_7_7_11
void FASTCALL call_2_7_7_11(SH2_struct* sh)
{
  SH2or(sh,7,7);
}
#define init_2_7_8_11
void FASTCALL call_2_7_8_11(SH2_struct* sh)
{
  SH2or(sh,7,8);
}
#define init_2_7_9_11
void FASTCALL call_2_7_9_11(SH2_struct* sh)
{
  SH2or(sh,7,9);
}
#define init_2_7_10_11
void FASTCALL call_2_7_10_11(SH2_struct* sh)
{
  SH2or(sh,7,10);
}
#define init_2_7_11_11
void FASTCALL call_2_7_11_11(SH2_struct* sh)
{
  SH2or(sh,7,11);
}
#define init_2_7_12_11
void FASTCALL call_2_7_12_11(SH2_struct* sh)
{
  SH2or(sh,7,12);
}
#define init_2_7_13_11
void FASTCALL call_2_7_13_11(SH2_struct* sh)
{
  SH2or(sh,7,13);
}
#define init_2_7_14_11
void FASTCALL call_2_7_14_11(SH2_struct* sh)
{
  SH2or(sh,7,14);
}
#define init_2_7_15_11
void FASTCALL call_2_7_15_11(SH2_struct* sh)
{
  SH2or(sh,7,15);
}
#define init_2_8_0_11
void FASTCALL call_2_8_0_11(SH2_struct* sh)
{
  SH2or(sh,8,0);
}
#define init_2_8_1_11
void FASTCALL call_2_8_1_11(SH2_struct* sh)
{
  SH2or(sh,8,1);
}
#define init_2_8_2_11
void FASTCALL call_2_8_2_11(SH2_struct* sh)
{
  SH2or(sh,8,2);
}
#define init_2_8_3_11
void FASTCALL call_2_8_3_11(SH2_struct* sh)
{
  SH2or(sh,8,3);
}
#define init_2_8_4_11
void FASTCALL call_2_8_4_11(SH2_struct* sh)
{
  SH2or(sh,8,4);
}
#define init_2_8_5_11
void FASTCALL call_2_8_5_11(SH2_struct* sh)
{
  SH2or(sh,8,5);
}
#define init_2_8_6_11
void FASTCALL call_2_8_6_11(SH2_struct* sh)
{
  SH2or(sh,8,6);
}
#define init_2_8_7_11
void FASTCALL call_2_8_7_11(SH2_struct* sh)
{
  SH2or(sh,8,7);
}
#define init_2_8_8_11
void FASTCALL call_2_8_8_11(SH2_struct* sh)
{
  SH2or(sh,8,8);
}
#define init_2_8_9_11
void FASTCALL call_2_8_9_11(SH2_struct* sh)
{
  SH2or(sh,8,9);
}
#define init_2_8_10_11
void FASTCALL call_2_8_10_11(SH2_struct* sh)
{
  SH2or(sh,8,10);
}
#define init_2_8_11_11
void FASTCALL call_2_8_11_11(SH2_struct* sh)
{
  SH2or(sh,8,11);
}
#define init_2_8_12_11
void FASTCALL call_2_8_12_11(SH2_struct* sh)
{
  SH2or(sh,8,12);
}
#define init_2_8_13_11
void FASTCALL call_2_8_13_11(SH2_struct* sh)
{
  SH2or(sh,8,13);
}
#define init_2_8_14_11
void FASTCALL call_2_8_14_11(SH2_struct* sh)
{
  SH2or(sh,8,14);
}
#define init_2_8_15_11
void FASTCALL call_2_8_15_11(SH2_struct* sh)
{
  SH2or(sh,8,15);
}
#define init_2_9_0_11
void FASTCALL call_2_9_0_11(SH2_struct* sh)
{
  SH2or(sh,9,0);
}
#define init_2_9_1_11
void FASTCALL call_2_9_1_11(SH2_struct* sh)
{
  SH2or(sh,9,1);
}
#define init_2_9_2_11
void FASTCALL call_2_9_2_11(SH2_struct* sh)
{
  SH2or(sh,9,2);
}
#define init_2_9_3_11
void FASTCALL call_2_9_3_11(SH2_struct* sh)
{
  SH2or(sh,9,3);
}
#define init_2_9_4_11
void FASTCALL call_2_9_4_11(SH2_struct* sh)
{
  SH2or(sh,9,4);
}
#define init_2_9_5_11
void FASTCALL call_2_9_5_11(SH2_struct* sh)
{
  SH2or(sh,9,5);
}
#define init_2_9_6_11
void FASTCALL call_2_9_6_11(SH2_struct* sh)
{
  SH2or(sh,9,6);
}
#define init_2_9_7_11
void FASTCALL call_2_9_7_11(SH2_struct* sh)
{
  SH2or(sh,9,7);
}
#define init_2_9_8_11
void FASTCALL call_2_9_8_11(SH2_struct* sh)
{
  SH2or(sh,9,8);
}
#define init_2_9_9_11
void FASTCALL call_2_9_9_11(SH2_struct* sh)
{
  SH2or(sh,9,9);
}
#define init_2_9_10_11
void FASTCALL call_2_9_10_11(SH2_struct* sh)
{
  SH2or(sh,9,10);
}
#define init_2_9_11_11
void FASTCALL call_2_9_11_11(SH2_struct* sh)
{
  SH2or(sh,9,11);
}
#define init_2_9_12_11
void FASTCALL call_2_9_12_11(SH2_struct* sh)
{
  SH2or(sh,9,12);
}
#define init_2_9_13_11
void FASTCALL call_2_9_13_11(SH2_struct* sh)
{
  SH2or(sh,9,13);
}
#define init_2_9_14_11
void FASTCALL call_2_9_14_11(SH2_struct* sh)
{
  SH2or(sh,9,14);
}
#define init_2_9_15_11
void FASTCALL call_2_9_15_11(SH2_struct* sh)
{
  SH2or(sh,9,15);
}
#define init_2_10_0_11
void FASTCALL call_2_10_0_11(SH2_struct* sh)
{
  SH2or(sh,10,0);
}
#define init_2_10_1_11
void FASTCALL call_2_10_1_11(SH2_struct* sh)
{
  SH2or(sh,10,1);
}
#define init_2_10_2_11
void FASTCALL call_2_10_2_11(SH2_struct* sh)
{
  SH2or(sh,10,2);
}
#define init_2_10_3_11
void FASTCALL call_2_10_3_11(SH2_struct* sh)
{
  SH2or(sh,10,3);
}
#define init_2_10_4_11
void FASTCALL call_2_10_4_11(SH2_struct* sh)
{
  SH2or(sh,10,4);
}
#define init_2_10_5_11
void FASTCALL call_2_10_5_11(SH2_struct* sh)
{
  SH2or(sh,10,5);
}
#define init_2_10_6_11
void FASTCALL call_2_10_6_11(SH2_struct* sh)
{
  SH2or(sh,10,6);
}
#define init_2_10_7_11
void FASTCALL call_2_10_7_11(SH2_struct* sh)
{
  SH2or(sh,10,7);
}
#define init_2_10_8_11
void FASTCALL call_2_10_8_11(SH2_struct* sh)
{
  SH2or(sh,10,8);
}
#define init_2_10_9_11
void FASTCALL call_2_10_9_11(SH2_struct* sh)
{
  SH2or(sh,10,9);
}
#define init_2_10_10_11
void FASTCALL call_2_10_10_11(SH2_struct* sh)
{
  SH2or(sh,10,10);
}
#define init_2_10_11_11
void FASTCALL call_2_10_11_11(SH2_struct* sh)
{
  SH2or(sh,10,11);
}
#define init_2_10_12_11
void FASTCALL call_2_10_12_11(SH2_struct* sh)
{
  SH2or(sh,10,12);
}
#define init_2_10_13_11
void FASTCALL call_2_10_13_11(SH2_struct* sh)
{
  SH2or(sh,10,13);
}
#define init_2_10_14_11
void FASTCALL call_2_10_14_11(SH2_struct* sh)
{
  SH2or(sh,10,14);
}
#define init_2_10_15_11
void FASTCALL call_2_10_15_11(SH2_struct* sh)
{
  SH2or(sh,10,15);
}
#define init_2_11_0_11
void FASTCALL call_2_11_0_11(SH2_struct* sh)
{
  SH2or(sh,11,0);
}
#define init_2_11_1_11
void FASTCALL call_2_11_1_11(SH2_struct* sh)
{
  SH2or(sh,11,1);
}
#define init_2_11_2_11
void FASTCALL call_2_11_2_11(SH2_struct* sh)
{
  SH2or(sh,11,2);
}
#define init_2_11_3_11
void FASTCALL call_2_11_3_11(SH2_struct* sh)
{
  SH2or(sh,11,3);
}
#define init_2_11_4_11
void FASTCALL call_2_11_4_11(SH2_struct* sh)
{
  SH2or(sh,11,4);
}
#define init_2_11_5_11
void FASTCALL call_2_11_5_11(SH2_struct* sh)
{
  SH2or(sh,11,5);
}
#define init_2_11_6_11
void FASTCALL call_2_11_6_11(SH2_struct* sh)
{
  SH2or(sh,11,6);
}
#define init_2_11_7_11
void FASTCALL call_2_11_7_11(SH2_struct* sh)
{
  SH2or(sh,11,7);
}
#define init_2_11_8_11
void FASTCALL call_2_11_8_11(SH2_struct* sh)
{
  SH2or(sh,11,8);
}
#define init_2_11_9_11
void FASTCALL call_2_11_9_11(SH2_struct* sh)
{
  SH2or(sh,11,9);
}
#define init_2_11_10_11
void FASTCALL call_2_11_10_11(SH2_struct* sh)
{
  SH2or(sh,11,10);
}
#define init_2_11_11_11
void FASTCALL call_2_11_11_11(SH2_struct* sh)
{
  SH2or(sh,11,11);
}
#define init_2_11_12_11
void FASTCALL call_2_11_12_11(SH2_struct* sh)
{
  SH2or(sh,11,12);
}
#define init_2_11_13_11
void FASTCALL call_2_11_13_11(SH2_struct* sh)
{
  SH2or(sh,11,13);
}
#define init_2_11_14_11
void FASTCALL call_2_11_14_11(SH2_struct* sh)
{
  SH2or(sh,11,14);
}
#define init_2_11_15_11
void FASTCALL call_2_11_15_11(SH2_struct* sh)
{
  SH2or(sh,11,15);
}
#define init_2_12_0_11
void FASTCALL call_2_12_0_11(SH2_struct* sh)
{
  SH2or(sh,12,0);
}
#define init_2_12_1_11
void FASTCALL call_2_12_1_11(SH2_struct* sh)
{
  SH2or(sh,12,1);
}
#define init_2_12_2_11
void FASTCALL call_2_12_2_11(SH2_struct* sh)
{
  SH2or(sh,12,2);
}
#define init_2_12_3_11
void FASTCALL call_2_12_3_11(SH2_struct* sh)
{
  SH2or(sh,12,3);
}
#define init_2_12_4_11
void FASTCALL call_2_12_4_11(SH2_struct* sh)
{
  SH2or(sh,12,4);
}
#define init_2_12_5_11
void FASTCALL call_2_12_5_11(SH2_struct* sh)
{
  SH2or(sh,12,5);
}
#define init_2_12_6_11
void FASTCALL call_2_12_6_11(SH2_struct* sh)
{
  SH2or(sh,12,6);
}
#define init_2_12_7_11
void FASTCALL call_2_12_7_11(SH2_struct* sh)
{
  SH2or(sh,12,7);
}
#define init_2_12_8_11
void FASTCALL call_2_12_8_11(SH2_struct* sh)
{
  SH2or(sh,12,8);
}
#define init_2_12_9_11
void FASTCALL call_2_12_9_11(SH2_struct* sh)
{
  SH2or(sh,12,9);
}
#define init_2_12_10_11
void FASTCALL call_2_12_10_11(SH2_struct* sh)
{
  SH2or(sh,12,10);
}
#define init_2_12_11_11
void FASTCALL call_2_12_11_11(SH2_struct* sh)
{
  SH2or(sh,12,11);
}
#define init_2_12_12_11
void FASTCALL call_2_12_12_11(SH2_struct* sh)
{
  SH2or(sh,12,12);
}
#define init_2_12_13_11
void FASTCALL call_2_12_13_11(SH2_struct* sh)
{
  SH2or(sh,12,13);
}
#define init_2_12_14_11
void FASTCALL call_2_12_14_11(SH2_struct* sh)
{
  SH2or(sh,12,14);
}
#define init_2_12_15_11
void FASTCALL call_2_12_15_11(SH2_struct* sh)
{
  SH2or(sh,12,15);
}
#define init_2_13_0_11
void FASTCALL call_2_13_0_11(SH2_struct* sh)
{
  SH2or(sh,13,0);
}
#define init_2_13_1_11
void FASTCALL call_2_13_1_11(SH2_struct* sh)
{
  SH2or(sh,13,1);
}
#define init_2_13_2_11
void FASTCALL call_2_13_2_11(SH2_struct* sh)
{
  SH2or(sh,13,2);
}
#define init_2_13_3_11
void FASTCALL call_2_13_3_11(SH2_struct* sh)
{
  SH2or(sh,13,3);
}
#define init_2_13_4_11
void FASTCALL call_2_13_4_11(SH2_struct* sh)
{
  SH2or(sh,13,4);
}
#define init_2_13_5_11
void FASTCALL call_2_13_5_11(SH2_struct* sh)
{
  SH2or(sh,13,5);
}
#define init_2_13_6_11
void FASTCALL call_2_13_6_11(SH2_struct* sh)
{
  SH2or(sh,13,6);
}
#define init_2_13_7_11
void FASTCALL call_2_13_7_11(SH2_struct* sh)
{
  SH2or(sh,13,7);
}
#define init_2_13_8_11
void FASTCALL call_2_13_8_11(SH2_struct* sh)
{
  SH2or(sh,13,8);
}
#define init_2_13_9_11
void FASTCALL call_2_13_9_11(SH2_struct* sh)
{
  SH2or(sh,13,9);
}
#define init_2_13_10_11
void FASTCALL call_2_13_10_11(SH2_struct* sh)
{
  SH2or(sh,13,10);
}
#define init_2_13_11_11
void FASTCALL call_2_13_11_11(SH2_struct* sh)
{
  SH2or(sh,13,11);
}
#define init_2_13_12_11
void FASTCALL call_2_13_12_11(SH2_struct* sh)
{
  SH2or(sh,13,12);
}
#define init_2_13_13_11
void FASTCALL call_2_13_13_11(SH2_struct* sh)
{
  SH2or(sh,13,13);
}
#define init_2_13_14_11
void FASTCALL call_2_13_14_11(SH2_struct* sh)
{
  SH2or(sh,13,14);
}
#define init_2_13_15_11
void FASTCALL call_2_13_15_11(SH2_struct* sh)
{
  SH2or(sh,13,15);
}
#define init_2_14_0_11
void FASTCALL call_2_14_0_11(SH2_struct* sh)
{
  SH2or(sh,14,0);
}
#define init_2_14_1_11
void FASTCALL call_2_14_1_11(SH2_struct* sh)
{
  SH2or(sh,14,1);
}
#define init_2_14_2_11
void FASTCALL call_2_14_2_11(SH2_struct* sh)
{
  SH2or(sh,14,2);
}
#define init_2_14_3_11
void FASTCALL call_2_14_3_11(SH2_struct* sh)
{
  SH2or(sh,14,3);
}
#define init_2_14_4_11
void FASTCALL call_2_14_4_11(SH2_struct* sh)
{
  SH2or(sh,14,4);
}
#define init_2_14_5_11
void FASTCALL call_2_14_5_11(SH2_struct* sh)
{
  SH2or(sh,14,5);
}
#define init_2_14_6_11
void FASTCALL call_2_14_6_11(SH2_struct* sh)
{
  SH2or(sh,14,6);
}
#define init_2_14_7_11
void FASTCALL call_2_14_7_11(SH2_struct* sh)
{
  SH2or(sh,14,7);
}
#define init_2_14_8_11
void FASTCALL call_2_14_8_11(SH2_struct* sh)
{
  SH2or(sh,14,8);
}
#define init_2_14_9_11
void FASTCALL call_2_14_9_11(SH2_struct* sh)
{
  SH2or(sh,14,9);
}
#define init_2_14_10_11
void FASTCALL call_2_14_10_11(SH2_struct* sh)
{
  SH2or(sh,14,10);
}
#define init_2_14_11_11
void FASTCALL call_2_14_11_11(SH2_struct* sh)
{
  SH2or(sh,14,11);
}
#define init_2_14_12_11
void FASTCALL call_2_14_12_11(SH2_struct* sh)
{
  SH2or(sh,14,12);
}
#define init_2_14_13_11
void FASTCALL call_2_14_13_11(SH2_struct* sh)
{
  SH2or(sh,14,13);
}
#define init_2_14_14_11
void FASTCALL call_2_14_14_11(SH2_struct* sh)
{
  SH2or(sh,14,14);
}
#define init_2_14_15_11
void FASTCALL call_2_14_15_11(SH2_struct* sh)
{
  SH2or(sh,14,15);
}
#define init_2_15_0_11
void FASTCALL call_2_15_0_11(SH2_struct* sh)
{
  SH2or(sh,15,0);
}
#define init_2_15_1_11
void FASTCALL call_2_15_1_11(SH2_struct* sh)
{
  SH2or(sh,15,1);
}
#define init_2_15_2_11
void FASTCALL call_2_15_2_11(SH2_struct* sh)
{
  SH2or(sh,15,2);
}
#define init_2_15_3_11
void FASTCALL call_2_15_3_11(SH2_struct* sh)
{
  SH2or(sh,15,3);
}
#define init_2_15_4_11
void FASTCALL call_2_15_4_11(SH2_struct* sh)
{
  SH2or(sh,15,4);
}
#define init_2_15_5_11
void FASTCALL call_2_15_5_11(SH2_struct* sh)
{
  SH2or(sh,15,5);
}
#define init_2_15_6_11
void FASTCALL call_2_15_6_11(SH2_struct* sh)
{
  SH2or(sh,15,6);
}
#define init_2_15_7_11
void FASTCALL call_2_15_7_11(SH2_struct* sh)
{
  SH2or(sh,15,7);
}
#define init_2_15_8_11
void FASTCALL call_2_15_8_11(SH2_struct* sh)
{
  SH2or(sh,15,8);
}
#define init_2_15_9_11
void FASTCALL call_2_15_9_11(SH2_struct* sh)
{
  SH2or(sh,15,9);
}
#define init_2_15_10_11
void FASTCALL call_2_15_10_11(SH2_struct* sh)
{
  SH2or(sh,15,10);
}
#define init_2_15_11_11
void FASTCALL call_2_15_11_11(SH2_struct* sh)
{
  SH2or(sh,15,11);
}
#define init_2_15_12_11
void FASTCALL call_2_15_12_11(SH2_struct* sh)
{
  SH2or(sh,15,12);
}
#define init_2_15_13_11
void FASTCALL call_2_15_13_11(SH2_struct* sh)
{
  SH2or(sh,15,13);
}
#define init_2_15_14_11
void FASTCALL call_2_15_14_11(SH2_struct* sh)
{
  SH2or(sh,15,14);
}
#define init_2_15_15_11
void FASTCALL call_2_15_15_11(SH2_struct* sh)
{
  SH2or(sh,15,15);
}
#define init_6_0_0_7
void FASTCALL call_6_0_0_7(SH2_struct* sh)
{
  SH2not(sh,0,0);
}
#define init_6_0_1_7
void FASTCALL call_6_0_1_7(SH2_struct* sh)
{
  SH2not(sh,0,1);
}
#define init_6_0_2_7
void FASTCALL call_6_0_2_7(SH2_struct* sh)
{
  SH2not(sh,0,2);
}
#define init_6_0_3_7
void FASTCALL call_6_0_3_7(SH2_struct* sh)
{
  SH2not(sh,0,3);
}
#define init_6_0_4_7
void FASTCALL call_6_0_4_7(SH2_struct* sh)
{
  SH2not(sh,0,4);
}
#define init_6_0_5_7
void FASTCALL call_6_0_5_7(SH2_struct* sh)
{
  SH2not(sh,0,5);
}
#define init_6_0_6_7
void FASTCALL call_6_0_6_7(SH2_struct* sh)
{
  SH2not(sh,0,6);
}
#define init_6_0_7_7
void FASTCALL call_6_0_7_7(SH2_struct* sh)
{
  SH2not(sh,0,7);
}
#define init_6_0_8_7
void FASTCALL call_6_0_8_7(SH2_struct* sh)
{
  SH2not(sh,0,8);
}
#define init_6_0_9_7
void FASTCALL call_6_0_9_7(SH2_struct* sh)
{
  SH2not(sh,0,9);
}
#define init_6_0_10_7
void FASTCALL call_6_0_10_7(SH2_struct* sh)
{
  SH2not(sh,0,10);
}
#define init_6_0_11_7
void FASTCALL call_6_0_11_7(SH2_struct* sh)
{
  SH2not(sh,0,11);
}
#define init_6_0_12_7
void FASTCALL call_6_0_12_7(SH2_struct* sh)
{
  SH2not(sh,0,12);
}
#define init_6_0_13_7
void FASTCALL call_6_0_13_7(SH2_struct* sh)
{
  SH2not(sh,0,13);
}
#define init_6_0_14_7
void FASTCALL call_6_0_14_7(SH2_struct* sh)
{
  SH2not(sh,0,14);
}
#define init_6_0_15_7
void FASTCALL call_6_0_15_7(SH2_struct* sh)
{
  SH2not(sh,0,15);
}
#define init_6_1_0_7
void FASTCALL call_6_1_0_7(SH2_struct* sh)
{
  SH2not(sh,1,0);
}
#define init_6_1_1_7
void FASTCALL call_6_1_1_7(SH2_struct* sh)
{
  SH2not(sh,1,1);
}
#define init_6_1_2_7
void FASTCALL call_6_1_2_7(SH2_struct* sh)
{
  SH2not(sh,1,2);
}
#define init_6_1_3_7
void FASTCALL call_6_1_3_7(SH2_struct* sh)
{
  SH2not(sh,1,3);
}
#define init_6_1_4_7
void FASTCALL call_6_1_4_7(SH2_struct* sh)
{
  SH2not(sh,1,4);
}
#define init_6_1_5_7
void FASTCALL call_6_1_5_7(SH2_struct* sh)
{
  SH2not(sh,1,5);
}
#define init_6_1_6_7
void FASTCALL call_6_1_6_7(SH2_struct* sh)
{
  SH2not(sh,1,6);
}
#define init_6_1_7_7
void FASTCALL call_6_1_7_7(SH2_struct* sh)
{
  SH2not(sh,1,7);
}
#define init_6_1_8_7
void FASTCALL call_6_1_8_7(SH2_struct* sh)
{
  SH2not(sh,1,8);
}
#define init_6_1_9_7
void FASTCALL call_6_1_9_7(SH2_struct* sh)
{
  SH2not(sh,1,9);
}
#define init_6_1_10_7
void FASTCALL call_6_1_10_7(SH2_struct* sh)
{
  SH2not(sh,1,10);
}
#define init_6_1_11_7
void FASTCALL call_6_1_11_7(SH2_struct* sh)
{
  SH2not(sh,1,11);
}
#define init_6_1_12_7
void FASTCALL call_6_1_12_7(SH2_struct* sh)
{
  SH2not(sh,1,12);
}
#define init_6_1_13_7
void FASTCALL call_6_1_13_7(SH2_struct* sh)
{
  SH2not(sh,1,13);
}
#define init_6_1_14_7
void FASTCALL call_6_1_14_7(SH2_struct* sh)
{
  SH2not(sh,1,14);
}
#define init_6_1_15_7
void FASTCALL call_6_1_15_7(SH2_struct* sh)
{
  SH2not(sh,1,15);
}
#define init_6_2_0_7
void FASTCALL call_6_2_0_7(SH2_struct* sh)
{
  SH2not(sh,2,0);
}
#define init_6_2_1_7
void FASTCALL call_6_2_1_7(SH2_struct* sh)
{
  SH2not(sh,2,1);
}
#define init_6_2_2_7
void FASTCALL call_6_2_2_7(SH2_struct* sh)
{
  SH2not(sh,2,2);
}
#define init_6_2_3_7
void FASTCALL call_6_2_3_7(SH2_struct* sh)
{
  SH2not(sh,2,3);
}
#define init_6_2_4_7
void FASTCALL call_6_2_4_7(SH2_struct* sh)
{
  SH2not(sh,2,4);
}
#define init_6_2_5_7
void FASTCALL call_6_2_5_7(SH2_struct* sh)
{
  SH2not(sh,2,5);
}
#define init_6_2_6_7
void FASTCALL call_6_2_6_7(SH2_struct* sh)
{
  SH2not(sh,2,6);
}
#define init_6_2_7_7
void FASTCALL call_6_2_7_7(SH2_struct* sh)
{
  SH2not(sh,2,7);
}
#define init_6_2_8_7
void FASTCALL call_6_2_8_7(SH2_struct* sh)
{
  SH2not(sh,2,8);
}
#define init_6_2_9_7
void FASTCALL call_6_2_9_7(SH2_struct* sh)
{
  SH2not(sh,2,9);
}
#define init_6_2_10_7
void FASTCALL call_6_2_10_7(SH2_struct* sh)
{
  SH2not(sh,2,10);
}
#define init_6_2_11_7
void FASTCALL call_6_2_11_7(SH2_struct* sh)
{
  SH2not(sh,2,11);
}
#define init_6_2_12_7
void FASTCALL call_6_2_12_7(SH2_struct* sh)
{
  SH2not(sh,2,12);
}
#define init_6_2_13_7
void FASTCALL call_6_2_13_7(SH2_struct* sh)
{
  SH2not(sh,2,13);
}
#define init_6_2_14_7
void FASTCALL call_6_2_14_7(SH2_struct* sh)
{
  SH2not(sh,2,14);
}
#define init_6_2_15_7
void FASTCALL call_6_2_15_7(SH2_struct* sh)
{
  SH2not(sh,2,15);
}
#define init_6_3_0_7
void FASTCALL call_6_3_0_7(SH2_struct* sh)
{
  SH2not(sh,3,0);
}
#define init_6_3_1_7
void FASTCALL call_6_3_1_7(SH2_struct* sh)
{
  SH2not(sh,3,1);
}
#define init_6_3_2_7
void FASTCALL call_6_3_2_7(SH2_struct* sh)
{
  SH2not(sh,3,2);
}
#define init_6_3_3_7
void FASTCALL call_6_3_3_7(SH2_struct* sh)
{
  SH2not(sh,3,3);
}
#define init_6_3_4_7
void FASTCALL call_6_3_4_7(SH2_struct* sh)
{
  SH2not(sh,3,4);
}
#define init_6_3_5_7
void FASTCALL call_6_3_5_7(SH2_struct* sh)
{
  SH2not(sh,3,5);
}
#define init_6_3_6_7
void FASTCALL call_6_3_6_7(SH2_struct* sh)
{
  SH2not(sh,3,6);
}
#define init_6_3_7_7
void FASTCALL call_6_3_7_7(SH2_struct* sh)
{
  SH2not(sh,3,7);
}
#define init_6_3_8_7
void FASTCALL call_6_3_8_7(SH2_struct* sh)
{
  SH2not(sh,3,8);
}
#define init_6_3_9_7
void FASTCALL call_6_3_9_7(SH2_struct* sh)
{
  SH2not(sh,3,9);
}
#define init_6_3_10_7
void FASTCALL call_6_3_10_7(SH2_struct* sh)
{
  SH2not(sh,3,10);
}
#define init_6_3_11_7
void FASTCALL call_6_3_11_7(SH2_struct* sh)
{
  SH2not(sh,3,11);
}
#define init_6_3_12_7
void FASTCALL call_6_3_12_7(SH2_struct* sh)
{
  SH2not(sh,3,12);
}
#define init_6_3_13_7
void FASTCALL call_6_3_13_7(SH2_struct* sh)
{
  SH2not(sh,3,13);
}
#define init_6_3_14_7
void FASTCALL call_6_3_14_7(SH2_struct* sh)
{
  SH2not(sh,3,14);
}
#define init_6_3_15_7
void FASTCALL call_6_3_15_7(SH2_struct* sh)
{
  SH2not(sh,3,15);
}
#define init_6_4_0_7
void FASTCALL call_6_4_0_7(SH2_struct* sh)
{
  SH2not(sh,4,0);
}
#define init_6_4_1_7
void FASTCALL call_6_4_1_7(SH2_struct* sh)
{
  SH2not(sh,4,1);
}
#define init_6_4_2_7
void FASTCALL call_6_4_2_7(SH2_struct* sh)
{
  SH2not(sh,4,2);
}
#define init_6_4_3_7
void FASTCALL call_6_4_3_7(SH2_struct* sh)
{
  SH2not(sh,4,3);
}
#define init_6_4_4_7
void FASTCALL call_6_4_4_7(SH2_struct* sh)
{
  SH2not(sh,4,4);
}
#define init_6_4_5_7
void FASTCALL call_6_4_5_7(SH2_struct* sh)
{
  SH2not(sh,4,5);
}
#define init_6_4_6_7
void FASTCALL call_6_4_6_7(SH2_struct* sh)
{
  SH2not(sh,4,6);
}
#define init_6_4_7_7
void FASTCALL call_6_4_7_7(SH2_struct* sh)
{
  SH2not(sh,4,7);
}
#define init_6_4_8_7
void FASTCALL call_6_4_8_7(SH2_struct* sh)
{
  SH2not(sh,4,8);
}
#define init_6_4_9_7
void FASTCALL call_6_4_9_7(SH2_struct* sh)
{
  SH2not(sh,4,9);
}
#define init_6_4_10_7
void FASTCALL call_6_4_10_7(SH2_struct* sh)
{
  SH2not(sh,4,10);
}
#define init_6_4_11_7
void FASTCALL call_6_4_11_7(SH2_struct* sh)
{
  SH2not(sh,4,11);
}
#define init_6_4_12_7
void FASTCALL call_6_4_12_7(SH2_struct* sh)
{
  SH2not(sh,4,12);
}
#define init_6_4_13_7
void FASTCALL call_6_4_13_7(SH2_struct* sh)
{
  SH2not(sh,4,13);
}
#define init_6_4_14_7
void FASTCALL call_6_4_14_7(SH2_struct* sh)
{
  SH2not(sh,4,14);
}
#define init_6_4_15_7
void FASTCALL call_6_4_15_7(SH2_struct* sh)
{
  SH2not(sh,4,15);
}
#define init_6_5_0_7
void FASTCALL call_6_5_0_7(SH2_struct* sh)
{
  SH2not(sh,5,0);
}
#define init_6_5_1_7
void FASTCALL call_6_5_1_7(SH2_struct* sh)
{
  SH2not(sh,5,1);
}
#define init_6_5_2_7
void FASTCALL call_6_5_2_7(SH2_struct* sh)
{
  SH2not(sh,5,2);
}
#define init_6_5_3_7
void FASTCALL call_6_5_3_7(SH2_struct* sh)
{
  SH2not(sh,5,3);
}
#define init_6_5_4_7
void FASTCALL call_6_5_4_7(SH2_struct* sh)
{
  SH2not(sh,5,4);
}
#define init_6_5_5_7
void FASTCALL call_6_5_5_7(SH2_struct* sh)
{
  SH2not(sh,5,5);
}
#define init_6_5_6_7
void FASTCALL call_6_5_6_7(SH2_struct* sh)
{
  SH2not(sh,5,6);
}
#define init_6_5_7_7
void FASTCALL call_6_5_7_7(SH2_struct* sh)
{
  SH2not(sh,5,7);
}
#define init_6_5_8_7
void FASTCALL call_6_5_8_7(SH2_struct* sh)
{
  SH2not(sh,5,8);
}
#define init_6_5_9_7
void FASTCALL call_6_5_9_7(SH2_struct* sh)
{
  SH2not(sh,5,9);
}
#define init_6_5_10_7
void FASTCALL call_6_5_10_7(SH2_struct* sh)
{
  SH2not(sh,5,10);
}
#define init_6_5_11_7
void FASTCALL call_6_5_11_7(SH2_struct* sh)
{
  SH2not(sh,5,11);
}
#define init_6_5_12_7
void FASTCALL call_6_5_12_7(SH2_struct* sh)
{
  SH2not(sh,5,12);
}
#define init_6_5_13_7
void FASTCALL call_6_5_13_7(SH2_struct* sh)
{
  SH2not(sh,5,13);
}
#define init_6_5_14_7
void FASTCALL call_6_5_14_7(SH2_struct* sh)
{
  SH2not(sh,5,14);
}
#define init_6_5_15_7
void FASTCALL call_6_5_15_7(SH2_struct* sh)
{
  SH2not(sh,5,15);
}
#define init_6_6_0_7
void FASTCALL call_6_6_0_7(SH2_struct* sh)
{
  SH2not(sh,6,0);
}
#define init_6_6_1_7
void FASTCALL call_6_6_1_7(SH2_struct* sh)
{
  SH2not(sh,6,1);
}
#define init_6_6_2_7
void FASTCALL call_6_6_2_7(SH2_struct* sh)
{
  SH2not(sh,6,2);
}
#define init_6_6_3_7
void FASTCALL call_6_6_3_7(SH2_struct* sh)
{
  SH2not(sh,6,3);
}
#define init_6_6_4_7
void FASTCALL call_6_6_4_7(SH2_struct* sh)
{
  SH2not(sh,6,4);
}
#define init_6_6_5_7
void FASTCALL call_6_6_5_7(SH2_struct* sh)
{
  SH2not(sh,6,5);
}
#define init_6_6_6_7
void FASTCALL call_6_6_6_7(SH2_struct* sh)
{
  SH2not(sh,6,6);
}
#define init_6_6_7_7
void FASTCALL call_6_6_7_7(SH2_struct* sh)
{
  SH2not(sh,6,7);
}
#define init_6_6_8_7
void FASTCALL call_6_6_8_7(SH2_struct* sh)
{
  SH2not(sh,6,8);
}
#define init_6_6_9_7
void FASTCALL call_6_6_9_7(SH2_struct* sh)
{
  SH2not(sh,6,9);
}
#define init_6_6_10_7
void FASTCALL call_6_6_10_7(SH2_struct* sh)
{
  SH2not(sh,6,10);
}
#define init_6_6_11_7
void FASTCALL call_6_6_11_7(SH2_struct* sh)
{
  SH2not(sh,6,11);
}
#define init_6_6_12_7
void FASTCALL call_6_6_12_7(SH2_struct* sh)
{
  SH2not(sh,6,12);
}
#define init_6_6_13_7
void FASTCALL call_6_6_13_7(SH2_struct* sh)
{
  SH2not(sh,6,13);
}
#define init_6_6_14_7
void FASTCALL call_6_6_14_7(SH2_struct* sh)
{
  SH2not(sh,6,14);
}
#define init_6_6_15_7
void FASTCALL call_6_6_15_7(SH2_struct* sh)
{
  SH2not(sh,6,15);
}
#define init_6_7_0_7
void FASTCALL call_6_7_0_7(SH2_struct* sh)
{
  SH2not(sh,7,0);
}
#define init_6_7_1_7
void FASTCALL call_6_7_1_7(SH2_struct* sh)
{
  SH2not(sh,7,1);
}
#define init_6_7_2_7
void FASTCALL call_6_7_2_7(SH2_struct* sh)
{
  SH2not(sh,7,2);
}
#define init_6_7_3_7
void FASTCALL call_6_7_3_7(SH2_struct* sh)
{
  SH2not(sh,7,3);
}
#define init_6_7_4_7
void FASTCALL call_6_7_4_7(SH2_struct* sh)
{
  SH2not(sh,7,4);
}
#define init_6_7_5_7
void FASTCALL call_6_7_5_7(SH2_struct* sh)
{
  SH2not(sh,7,5);
}
#define init_6_7_6_7
void FASTCALL call_6_7_6_7(SH2_struct* sh)
{
  SH2not(sh,7,6);
}
#define init_6_7_7_7
void FASTCALL call_6_7_7_7(SH2_struct* sh)
{
  SH2not(sh,7,7);
}
#define init_6_7_8_7
void FASTCALL call_6_7_8_7(SH2_struct* sh)
{
  SH2not(sh,7,8);
}
#define init_6_7_9_7
void FASTCALL call_6_7_9_7(SH2_struct* sh)
{
  SH2not(sh,7,9);
}
#define init_6_7_10_7
void FASTCALL call_6_7_10_7(SH2_struct* sh)
{
  SH2not(sh,7,10);
}
#define init_6_7_11_7
void FASTCALL call_6_7_11_7(SH2_struct* sh)
{
  SH2not(sh,7,11);
}
#define init_6_7_12_7
void FASTCALL call_6_7_12_7(SH2_struct* sh)
{
  SH2not(sh,7,12);
}
#define init_6_7_13_7
void FASTCALL call_6_7_13_7(SH2_struct* sh)
{
  SH2not(sh,7,13);
}
#define init_6_7_14_7
void FASTCALL call_6_7_14_7(SH2_struct* sh)
{
  SH2not(sh,7,14);
}
#define init_6_7_15_7
void FASTCALL call_6_7_15_7(SH2_struct* sh)
{
  SH2not(sh,7,15);
}
#define init_6_8_0_7
void FASTCALL call_6_8_0_7(SH2_struct* sh)
{
  SH2not(sh,8,0);
}
#define init_6_8_1_7
void FASTCALL call_6_8_1_7(SH2_struct* sh)
{
  SH2not(sh,8,1);
}
#define init_6_8_2_7
void FASTCALL call_6_8_2_7(SH2_struct* sh)
{
  SH2not(sh,8,2);
}
#define init_6_8_3_7
void FASTCALL call_6_8_3_7(SH2_struct* sh)
{
  SH2not(sh,8,3);
}
#define init_6_8_4_7
void FASTCALL call_6_8_4_7(SH2_struct* sh)
{
  SH2not(sh,8,4);
}
#define init_6_8_5_7
void FASTCALL call_6_8_5_7(SH2_struct* sh)
{
  SH2not(sh,8,5);
}
#define init_6_8_6_7
void FASTCALL call_6_8_6_7(SH2_struct* sh)
{
  SH2not(sh,8,6);
}
#define init_6_8_7_7
void FASTCALL call_6_8_7_7(SH2_struct* sh)
{
  SH2not(sh,8,7);
}
#define init_6_8_8_7
void FASTCALL call_6_8_8_7(SH2_struct* sh)
{
  SH2not(sh,8,8);
}
#define init_6_8_9_7
void FASTCALL call_6_8_9_7(SH2_struct* sh)
{
  SH2not(sh,8,9);
}
#define init_6_8_10_7
void FASTCALL call_6_8_10_7(SH2_struct* sh)
{
  SH2not(sh,8,10);
}
#define init_6_8_11_7
void FASTCALL call_6_8_11_7(SH2_struct* sh)
{
  SH2not(sh,8,11);
}
#define init_6_8_12_7
void FASTCALL call_6_8_12_7(SH2_struct* sh)
{
  SH2not(sh,8,12);
}
#define init_6_8_13_7
void FASTCALL call_6_8_13_7(SH2_struct* sh)
{
  SH2not(sh,8,13);
}
#define init_6_8_14_7
void FASTCALL call_6_8_14_7(SH2_struct* sh)
{
  SH2not(sh,8,14);
}
#define init_6_8_15_7
void FASTCALL call_6_8_15_7(SH2_struct* sh)
{
  SH2not(sh,8,15);
}
#define init_6_9_0_7
void FASTCALL call_6_9_0_7(SH2_struct* sh)
{
  SH2not(sh,9,0);
}
#define init_6_9_1_7
void FASTCALL call_6_9_1_7(SH2_struct* sh)
{
  SH2not(sh,9,1);
}
#define init_6_9_2_7
void FASTCALL call_6_9_2_7(SH2_struct* sh)
{
  SH2not(sh,9,2);
}
#define init_6_9_3_7
void FASTCALL call_6_9_3_7(SH2_struct* sh)
{
  SH2not(sh,9,3);
}
#define init_6_9_4_7
void FASTCALL call_6_9_4_7(SH2_struct* sh)
{
  SH2not(sh,9,4);
}
#define init_6_9_5_7
void FASTCALL call_6_9_5_7(SH2_struct* sh)
{
  SH2not(sh,9,5);
}
#define init_6_9_6_7
void FASTCALL call_6_9_6_7(SH2_struct* sh)
{
  SH2not(sh,9,6);
}
#define init_6_9_7_7
void FASTCALL call_6_9_7_7(SH2_struct* sh)
{
  SH2not(sh,9,7);
}
#define init_6_9_8_7
void FASTCALL call_6_9_8_7(SH2_struct* sh)
{
  SH2not(sh,9,8);
}
#define init_6_9_9_7
void FASTCALL call_6_9_9_7(SH2_struct* sh)
{
  SH2not(sh,9,9);
}
#define init_6_9_10_7
void FASTCALL call_6_9_10_7(SH2_struct* sh)
{
  SH2not(sh,9,10);
}
#define init_6_9_11_7
void FASTCALL call_6_9_11_7(SH2_struct* sh)
{
  SH2not(sh,9,11);
}
#define init_6_9_12_7
void FASTCALL call_6_9_12_7(SH2_struct* sh)
{
  SH2not(sh,9,12);
}
#define init_6_9_13_7
void FASTCALL call_6_9_13_7(SH2_struct* sh)
{
  SH2not(sh,9,13);
}
#define init_6_9_14_7
void FASTCALL call_6_9_14_7(SH2_struct* sh)
{
  SH2not(sh,9,14);
}
#define init_6_9_15_7
void FASTCALL call_6_9_15_7(SH2_struct* sh)
{
  SH2not(sh,9,15);
}
#define init_6_10_0_7
void FASTCALL call_6_10_0_7(SH2_struct* sh)
{
  SH2not(sh,10,0);
}
#define init_6_10_1_7
void FASTCALL call_6_10_1_7(SH2_struct* sh)
{
  SH2not(sh,10,1);
}
#define init_6_10_2_7
void FASTCALL call_6_10_2_7(SH2_struct* sh)
{
  SH2not(sh,10,2);
}
#define init_6_10_3_7
void FASTCALL call_6_10_3_7(SH2_struct* sh)
{
  SH2not(sh,10,3);
}
#define init_6_10_4_7
void FASTCALL call_6_10_4_7(SH2_struct* sh)
{
  SH2not(sh,10,4);
}
#define init_6_10_5_7
void FASTCALL call_6_10_5_7(SH2_struct* sh)
{
  SH2not(sh,10,5);
}
#define init_6_10_6_7
void FASTCALL call_6_10_6_7(SH2_struct* sh)
{
  SH2not(sh,10,6);
}
#define init_6_10_7_7
void FASTCALL call_6_10_7_7(SH2_struct* sh)
{
  SH2not(sh,10,7);
}
#define init_6_10_8_7
void FASTCALL call_6_10_8_7(SH2_struct* sh)
{
  SH2not(sh,10,8);
}
#define init_6_10_9_7
void FASTCALL call_6_10_9_7(SH2_struct* sh)
{
  SH2not(sh,10,9);
}
#define init_6_10_10_7
void FASTCALL call_6_10_10_7(SH2_struct* sh)
{
  SH2not(sh,10,10);
}
#define init_6_10_11_7
void FASTCALL call_6_10_11_7(SH2_struct* sh)
{
  SH2not(sh,10,11);
}
#define init_6_10_12_7
void FASTCALL call_6_10_12_7(SH2_struct* sh)
{
  SH2not(sh,10,12);
}
#define init_6_10_13_7
void FASTCALL call_6_10_13_7(SH2_struct* sh)
{
  SH2not(sh,10,13);
}
#define init_6_10_14_7
void FASTCALL call_6_10_14_7(SH2_struct* sh)
{
  SH2not(sh,10,14);
}
#define init_6_10_15_7
void FASTCALL call_6_10_15_7(SH2_struct* sh)
{
  SH2not(sh,10,15);
}
#define init_6_11_0_7
void FASTCALL call_6_11_0_7(SH2_struct* sh)
{
  SH2not(sh,11,0);
}
#define init_6_11_1_7
void FASTCALL call_6_11_1_7(SH2_struct* sh)
{
  SH2not(sh,11,1);
}
#define init_6_11_2_7
void FASTCALL call_6_11_2_7(SH2_struct* sh)
{
  SH2not(sh,11,2);
}
#define init_6_11_3_7
void FASTCALL call_6_11_3_7(SH2_struct* sh)
{
  SH2not(sh,11,3);
}
#define init_6_11_4_7
void FASTCALL call_6_11_4_7(SH2_struct* sh)
{
  SH2not(sh,11,4);
}
#define init_6_11_5_7
void FASTCALL call_6_11_5_7(SH2_struct* sh)
{
  SH2not(sh,11,5);
}
#define init_6_11_6_7
void FASTCALL call_6_11_6_7(SH2_struct* sh)
{
  SH2not(sh,11,6);
}
#define init_6_11_7_7
void FASTCALL call_6_11_7_7(SH2_struct* sh)
{
  SH2not(sh,11,7);
}
#define init_6_11_8_7
void FASTCALL call_6_11_8_7(SH2_struct* sh)
{
  SH2not(sh,11,8);
}
#define init_6_11_9_7
void FASTCALL call_6_11_9_7(SH2_struct* sh)
{
  SH2not(sh,11,9);
}
#define init_6_11_10_7
void FASTCALL call_6_11_10_7(SH2_struct* sh)
{
  SH2not(sh,11,10);
}
#define init_6_11_11_7
void FASTCALL call_6_11_11_7(SH2_struct* sh)
{
  SH2not(sh,11,11);
}
#define init_6_11_12_7
void FASTCALL call_6_11_12_7(SH2_struct* sh)
{
  SH2not(sh,11,12);
}
#define init_6_11_13_7
void FASTCALL call_6_11_13_7(SH2_struct* sh)
{
  SH2not(sh,11,13);
}
#define init_6_11_14_7
void FASTCALL call_6_11_14_7(SH2_struct* sh)
{
  SH2not(sh,11,14);
}
#define init_6_11_15_7
void FASTCALL call_6_11_15_7(SH2_struct* sh)
{
  SH2not(sh,11,15);
}
#define init_6_12_0_7
void FASTCALL call_6_12_0_7(SH2_struct* sh)
{
  SH2not(sh,12,0);
}
#define init_6_12_1_7
void FASTCALL call_6_12_1_7(SH2_struct* sh)
{
  SH2not(sh,12,1);
}
#define init_6_12_2_7
void FASTCALL call_6_12_2_7(SH2_struct* sh)
{
  SH2not(sh,12,2);
}
#define init_6_12_3_7
void FASTCALL call_6_12_3_7(SH2_struct* sh)
{
  SH2not(sh,12,3);
}
#define init_6_12_4_7
void FASTCALL call_6_12_4_7(SH2_struct* sh)
{
  SH2not(sh,12,4);
}
#define init_6_12_5_7
void FASTCALL call_6_12_5_7(SH2_struct* sh)
{
  SH2not(sh,12,5);
}
#define init_6_12_6_7
void FASTCALL call_6_12_6_7(SH2_struct* sh)
{
  SH2not(sh,12,6);
}
#define init_6_12_7_7
void FASTCALL call_6_12_7_7(SH2_struct* sh)
{
  SH2not(sh,12,7);
}
#define init_6_12_8_7
void FASTCALL call_6_12_8_7(SH2_struct* sh)
{
  SH2not(sh,12,8);
}
#define init_6_12_9_7
void FASTCALL call_6_12_9_7(SH2_struct* sh)
{
  SH2not(sh,12,9);
}
#define init_6_12_10_7
void FASTCALL call_6_12_10_7(SH2_struct* sh)
{
  SH2not(sh,12,10);
}
#define init_6_12_11_7
void FASTCALL call_6_12_11_7(SH2_struct* sh)
{
  SH2not(sh,12,11);
}
#define init_6_12_12_7
void FASTCALL call_6_12_12_7(SH2_struct* sh)
{
  SH2not(sh,12,12);
}
#define init_6_12_13_7
void FASTCALL call_6_12_13_7(SH2_struct* sh)
{
  SH2not(sh,12,13);
}
#define init_6_12_14_7
void FASTCALL call_6_12_14_7(SH2_struct* sh)
{
  SH2not(sh,12,14);
}
#define init_6_12_15_7
void FASTCALL call_6_12_15_7(SH2_struct* sh)
{
  SH2not(sh,12,15);
}
#define init_6_13_0_7
void FASTCALL call_6_13_0_7(SH2_struct* sh)
{
  SH2not(sh,13,0);
}
#define init_6_13_1_7
void FASTCALL call_6_13_1_7(SH2_struct* sh)
{
  SH2not(sh,13,1);
}
#define init_6_13_2_7
void FASTCALL call_6_13_2_7(SH2_struct* sh)
{
  SH2not(sh,13,2);
}
#define init_6_13_3_7
void FASTCALL call_6_13_3_7(SH2_struct* sh)
{
  SH2not(sh,13,3);
}
#define init_6_13_4_7
void FASTCALL call_6_13_4_7(SH2_struct* sh)
{
  SH2not(sh,13,4);
}
#define init_6_13_5_7
void FASTCALL call_6_13_5_7(SH2_struct* sh)
{
  SH2not(sh,13,5);
}
#define init_6_13_6_7
void FASTCALL call_6_13_6_7(SH2_struct* sh)
{
  SH2not(sh,13,6);
}
#define init_6_13_7_7
void FASTCALL call_6_13_7_7(SH2_struct* sh)
{
  SH2not(sh,13,7);
}
#define init_6_13_8_7
void FASTCALL call_6_13_8_7(SH2_struct* sh)
{
  SH2not(sh,13,8);
}
#define init_6_13_9_7
void FASTCALL call_6_13_9_7(SH2_struct* sh)
{
  SH2not(sh,13,9);
}
#define init_6_13_10_7
void FASTCALL call_6_13_10_7(SH2_struct* sh)
{
  SH2not(sh,13,10);
}
#define init_6_13_11_7
void FASTCALL call_6_13_11_7(SH2_struct* sh)
{
  SH2not(sh,13,11);
}
#define init_6_13_12_7
void FASTCALL call_6_13_12_7(SH2_struct* sh)
{
  SH2not(sh,13,12);
}
#define init_6_13_13_7
void FASTCALL call_6_13_13_7(SH2_struct* sh)
{
  SH2not(sh,13,13);
}
#define init_6_13_14_7
void FASTCALL call_6_13_14_7(SH2_struct* sh)
{
  SH2not(sh,13,14);
}
#define init_6_13_15_7
void FASTCALL call_6_13_15_7(SH2_struct* sh)
{
  SH2not(sh,13,15);
}
#define init_6_14_0_7
void FASTCALL call_6_14_0_7(SH2_struct* sh)
{
  SH2not(sh,14,0);
}
#define init_6_14_1_7
void FASTCALL call_6_14_1_7(SH2_struct* sh)
{
  SH2not(sh,14,1);
}
#define init_6_14_2_7
void FASTCALL call_6_14_2_7(SH2_struct* sh)
{
  SH2not(sh,14,2);
}
#define init_6_14_3_7
void FASTCALL call_6_14_3_7(SH2_struct* sh)
{
  SH2not(sh,14,3);
}
#define init_6_14_4_7
void FASTCALL call_6_14_4_7(SH2_struct* sh)
{
  SH2not(sh,14,4);
}
#define init_6_14_5_7
void FASTCALL call_6_14_5_7(SH2_struct* sh)
{
  SH2not(sh,14,5);
}
#define init_6_14_6_7
void FASTCALL call_6_14_6_7(SH2_struct* sh)
{
  SH2not(sh,14,6);
}
#define init_6_14_7_7
void FASTCALL call_6_14_7_7(SH2_struct* sh)
{
  SH2not(sh,14,7);
}
#define init_6_14_8_7
void FASTCALL call_6_14_8_7(SH2_struct* sh)
{
  SH2not(sh,14,8);
}
#define init_6_14_9_7
void FASTCALL call_6_14_9_7(SH2_struct* sh)
{
  SH2not(sh,14,9);
}
#define init_6_14_10_7
void FASTCALL call_6_14_10_7(SH2_struct* sh)
{
  SH2not(sh,14,10);
}
#define init_6_14_11_7
void FASTCALL call_6_14_11_7(SH2_struct* sh)
{
  SH2not(sh,14,11);
}
#define init_6_14_12_7
void FASTCALL call_6_14_12_7(SH2_struct* sh)
{
  SH2not(sh,14,12);
}
#define init_6_14_13_7
void FASTCALL call_6_14_13_7(SH2_struct* sh)
{
  SH2not(sh,14,13);
}
#define init_6_14_14_7
void FASTCALL call_6_14_14_7(SH2_struct* sh)
{
  SH2not(sh,14,14);
}
#define init_6_14_15_7
void FASTCALL call_6_14_15_7(SH2_struct* sh)
{
  SH2not(sh,14,15);
}
#define init_6_15_0_7
void FASTCALL call_6_15_0_7(SH2_struct* sh)
{
  SH2not(sh,15,0);
}
#define init_6_15_1_7
void FASTCALL call_6_15_1_7(SH2_struct* sh)
{
  SH2not(sh,15,1);
}
#define init_6_15_2_7
void FASTCALL call_6_15_2_7(SH2_struct* sh)
{
  SH2not(sh,15,2);
}
#define init_6_15_3_7
void FASTCALL call_6_15_3_7(SH2_struct* sh)
{
  SH2not(sh,15,3);
}
#define init_6_15_4_7
void FASTCALL call_6_15_4_7(SH2_struct* sh)
{
  SH2not(sh,15,4);
}
#define init_6_15_5_7
void FASTCALL call_6_15_5_7(SH2_struct* sh)
{
  SH2not(sh,15,5);
}
#define init_6_15_6_7
void FASTCALL call_6_15_6_7(SH2_struct* sh)
{
  SH2not(sh,15,6);
}
#define init_6_15_7_7
void FASTCALL call_6_15_7_7(SH2_struct* sh)
{
  SH2not(sh,15,7);
}
#define init_6_15_8_7
void FASTCALL call_6_15_8_7(SH2_struct* sh)
{
  SH2not(sh,15,8);
}
#define init_6_15_9_7
void FASTCALL call_6_15_9_7(SH2_struct* sh)
{
  SH2not(sh,15,9);
}
#define init_6_15_10_7
void FASTCALL call_6_15_10_7(SH2_struct* sh)
{
  SH2not(sh,15,10);
}
#define init_6_15_11_7
void FASTCALL call_6_15_11_7(SH2_struct* sh)
{
  SH2not(sh,15,11);
}
#define init_6_15_12_7
void FASTCALL call_6_15_12_7(SH2_struct* sh)
{
  SH2not(sh,15,12);
}
#define init_6_15_13_7
void FASTCALL call_6_15_13_7(SH2_struct* sh)
{
  SH2not(sh,15,13);
}
#define init_6_15_14_7
void FASTCALL call_6_15_14_7(SH2_struct* sh)
{
  SH2not(sh,15,14);
}
#define init_6_15_15_7
void FASTCALL call_6_15_15_7(SH2_struct* sh)
{
  SH2not(sh,15,15);
}
#define init_6_0_0_10
void FASTCALL call_6_0_0_10(SH2_struct* sh)
{
  SH2negc(sh,0,0);
}
#define init_6_0_1_10
void FASTCALL call_6_0_1_10(SH2_struct* sh)
{
  SH2negc(sh,0,1);
}
#define init_6_0_2_10
void FASTCALL call_6_0_2_10(SH2_struct* sh)
{
  SH2negc(sh,0,2);
}
#define init_6_0_3_10
void FASTCALL call_6_0_3_10(SH2_struct* sh)
{
  SH2negc(sh,0,3);
}
#define init_6_0_4_10
void FASTCALL call_6_0_4_10(SH2_struct* sh)
{
  SH2negc(sh,0,4);
}
#define init_6_0_5_10
void FASTCALL call_6_0_5_10(SH2_struct* sh)
{
  SH2negc(sh,0,5);
}
#define init_6_0_6_10
void FASTCALL call_6_0_6_10(SH2_struct* sh)
{
  SH2negc(sh,0,6);
}
#define init_6_0_7_10
void FASTCALL call_6_0_7_10(SH2_struct* sh)
{
  SH2negc(sh,0,7);
}
#define init_6_0_8_10
void FASTCALL call_6_0_8_10(SH2_struct* sh)
{
  SH2negc(sh,0,8);
}
#define init_6_0_9_10
void FASTCALL call_6_0_9_10(SH2_struct* sh)
{
  SH2negc(sh,0,9);
}
#define init_6_0_10_10
void FASTCALL call_6_0_10_10(SH2_struct* sh)
{
  SH2negc(sh,0,10);
}
#define init_6_0_11_10
void FASTCALL call_6_0_11_10(SH2_struct* sh)
{
  SH2negc(sh,0,11);
}
#define init_6_0_12_10
void FASTCALL call_6_0_12_10(SH2_struct* sh)
{
  SH2negc(sh,0,12);
}
#define init_6_0_13_10
void FASTCALL call_6_0_13_10(SH2_struct* sh)
{
  SH2negc(sh,0,13);
}
#define init_6_0_14_10
void FASTCALL call_6_0_14_10(SH2_struct* sh)
{
  SH2negc(sh,0,14);
}
#define init_6_0_15_10
void FASTCALL call_6_0_15_10(SH2_struct* sh)
{
  SH2negc(sh,0,15);
}
#define init_6_1_0_10
void FASTCALL call_6_1_0_10(SH2_struct* sh)
{
  SH2negc(sh,1,0);
}
#define init_6_1_1_10
void FASTCALL call_6_1_1_10(SH2_struct* sh)
{
  SH2negc(sh,1,1);
}
#define init_6_1_2_10
void FASTCALL call_6_1_2_10(SH2_struct* sh)
{
  SH2negc(sh,1,2);
}
#define init_6_1_3_10
void FASTCALL call_6_1_3_10(SH2_struct* sh)
{
  SH2negc(sh,1,3);
}
#define init_6_1_4_10
void FASTCALL call_6_1_4_10(SH2_struct* sh)
{
  SH2negc(sh,1,4);
}
#define init_6_1_5_10
void FASTCALL call_6_1_5_10(SH2_struct* sh)
{
  SH2negc(sh,1,5);
}
#define init_6_1_6_10
void FASTCALL call_6_1_6_10(SH2_struct* sh)
{
  SH2negc(sh,1,6);
}
#define init_6_1_7_10
void FASTCALL call_6_1_7_10(SH2_struct* sh)
{
  SH2negc(sh,1,7);
}
#define init_6_1_8_10
void FASTCALL call_6_1_8_10(SH2_struct* sh)
{
  SH2negc(sh,1,8);
}
#define init_6_1_9_10
void FASTCALL call_6_1_9_10(SH2_struct* sh)
{
  SH2negc(sh,1,9);
}
#define init_6_1_10_10
void FASTCALL call_6_1_10_10(SH2_struct* sh)
{
  SH2negc(sh,1,10);
}
#define init_6_1_11_10
void FASTCALL call_6_1_11_10(SH2_struct* sh)
{
  SH2negc(sh,1,11);
}
#define init_6_1_12_10
void FASTCALL call_6_1_12_10(SH2_struct* sh)
{
  SH2negc(sh,1,12);
}
#define init_6_1_13_10
void FASTCALL call_6_1_13_10(SH2_struct* sh)
{
  SH2negc(sh,1,13);
}
#define init_6_1_14_10
void FASTCALL call_6_1_14_10(SH2_struct* sh)
{
  SH2negc(sh,1,14);
}
#define init_6_1_15_10
void FASTCALL call_6_1_15_10(SH2_struct* sh)
{
  SH2negc(sh,1,15);
}
#define init_6_2_0_10
void FASTCALL call_6_2_0_10(SH2_struct* sh)
{
  SH2negc(sh,2,0);
}
#define init_6_2_1_10
void FASTCALL call_6_2_1_10(SH2_struct* sh)
{
  SH2negc(sh,2,1);
}
#define init_6_2_2_10
void FASTCALL call_6_2_2_10(SH2_struct* sh)
{
  SH2negc(sh,2,2);
}
#define init_6_2_3_10
void FASTCALL call_6_2_3_10(SH2_struct* sh)
{
  SH2negc(sh,2,3);
}
#define init_6_2_4_10
void FASTCALL call_6_2_4_10(SH2_struct* sh)
{
  SH2negc(sh,2,4);
}
#define init_6_2_5_10
void FASTCALL call_6_2_5_10(SH2_struct* sh)
{
  SH2negc(sh,2,5);
}
#define init_6_2_6_10
void FASTCALL call_6_2_6_10(SH2_struct* sh)
{
  SH2negc(sh,2,6);
}
#define init_6_2_7_10
void FASTCALL call_6_2_7_10(SH2_struct* sh)
{
  SH2negc(sh,2,7);
}
#define init_6_2_8_10
void FASTCALL call_6_2_8_10(SH2_struct* sh)
{
  SH2negc(sh,2,8);
}
#define init_6_2_9_10
void FASTCALL call_6_2_9_10(SH2_struct* sh)
{
  SH2negc(sh,2,9);
}
#define init_6_2_10_10
void FASTCALL call_6_2_10_10(SH2_struct* sh)
{
  SH2negc(sh,2,10);
}
#define init_6_2_11_10
void FASTCALL call_6_2_11_10(SH2_struct* sh)
{
  SH2negc(sh,2,11);
}
#define init_6_2_12_10
void FASTCALL call_6_2_12_10(SH2_struct* sh)
{
  SH2negc(sh,2,12);
}
#define init_6_2_13_10
void FASTCALL call_6_2_13_10(SH2_struct* sh)
{
  SH2negc(sh,2,13);
}
#define init_6_2_14_10
void FASTCALL call_6_2_14_10(SH2_struct* sh)
{
  SH2negc(sh,2,14);
}
#define init_6_2_15_10
void FASTCALL call_6_2_15_10(SH2_struct* sh)
{
  SH2negc(sh,2,15);
}
#define init_6_3_0_10
void FASTCALL call_6_3_0_10(SH2_struct* sh)
{
  SH2negc(sh,3,0);
}
#define init_6_3_1_10
void FASTCALL call_6_3_1_10(SH2_struct* sh)
{
  SH2negc(sh,3,1);
}
#define init_6_3_2_10
void FASTCALL call_6_3_2_10(SH2_struct* sh)
{
  SH2negc(sh,3,2);
}
#define init_6_3_3_10
void FASTCALL call_6_3_3_10(SH2_struct* sh)
{
  SH2negc(sh,3,3);
}
#define init_6_3_4_10
void FASTCALL call_6_3_4_10(SH2_struct* sh)
{
  SH2negc(sh,3,4);
}
#define init_6_3_5_10
void FASTCALL call_6_3_5_10(SH2_struct* sh)
{
  SH2negc(sh,3,5);
}
#define init_6_3_6_10
void FASTCALL call_6_3_6_10(SH2_struct* sh)
{
  SH2negc(sh,3,6);
}
#define init_6_3_7_10
void FASTCALL call_6_3_7_10(SH2_struct* sh)
{
  SH2negc(sh,3,7);
}
#define init_6_3_8_10
void FASTCALL call_6_3_8_10(SH2_struct* sh)
{
  SH2negc(sh,3,8);
}
#define init_6_3_9_10
void FASTCALL call_6_3_9_10(SH2_struct* sh)
{
  SH2negc(sh,3,9);
}
#define init_6_3_10_10
void FASTCALL call_6_3_10_10(SH2_struct* sh)
{
  SH2negc(sh,3,10);
}
#define init_6_3_11_10
void FASTCALL call_6_3_11_10(SH2_struct* sh)
{
  SH2negc(sh,3,11);
}
#define init_6_3_12_10
void FASTCALL call_6_3_12_10(SH2_struct* sh)
{
  SH2negc(sh,3,12);
}
#define init_6_3_13_10
void FASTCALL call_6_3_13_10(SH2_struct* sh)
{
  SH2negc(sh,3,13);
}
#define init_6_3_14_10
void FASTCALL call_6_3_14_10(SH2_struct* sh)
{
  SH2negc(sh,3,14);
}
#define init_6_3_15_10
void FASTCALL call_6_3_15_10(SH2_struct* sh)
{
  SH2negc(sh,3,15);
}
#define init_6_4_0_10
void FASTCALL call_6_4_0_10(SH2_struct* sh)
{
  SH2negc(sh,4,0);
}
#define init_6_4_1_10
void FASTCALL call_6_4_1_10(SH2_struct* sh)
{
  SH2negc(sh,4,1);
}
#define init_6_4_2_10
void FASTCALL call_6_4_2_10(SH2_struct* sh)
{
  SH2negc(sh,4,2);
}
#define init_6_4_3_10
void FASTCALL call_6_4_3_10(SH2_struct* sh)
{
  SH2negc(sh,4,3);
}
#define init_6_4_4_10
void FASTCALL call_6_4_4_10(SH2_struct* sh)
{
  SH2negc(sh,4,4);
}
#define init_6_4_5_10
void FASTCALL call_6_4_5_10(SH2_struct* sh)
{
  SH2negc(sh,4,5);
}
#define init_6_4_6_10
void FASTCALL call_6_4_6_10(SH2_struct* sh)
{
  SH2negc(sh,4,6);
}
#define init_6_4_7_10
void FASTCALL call_6_4_7_10(SH2_struct* sh)
{
  SH2negc(sh,4,7);
}
#define init_6_4_8_10
void FASTCALL call_6_4_8_10(SH2_struct* sh)
{
  SH2negc(sh,4,8);
}
#define init_6_4_9_10
void FASTCALL call_6_4_9_10(SH2_struct* sh)
{
  SH2negc(sh,4,9);
}
#define init_6_4_10_10
void FASTCALL call_6_4_10_10(SH2_struct* sh)
{
  SH2negc(sh,4,10);
}
#define init_6_4_11_10
void FASTCALL call_6_4_11_10(SH2_struct* sh)
{
  SH2negc(sh,4,11);
}
#define init_6_4_12_10
void FASTCALL call_6_4_12_10(SH2_struct* sh)
{
  SH2negc(sh,4,12);
}
#define init_6_4_13_10
void FASTCALL call_6_4_13_10(SH2_struct* sh)
{
  SH2negc(sh,4,13);
}
#define init_6_4_14_10
void FASTCALL call_6_4_14_10(SH2_struct* sh)
{
  SH2negc(sh,4,14);
}
#define init_6_4_15_10
void FASTCALL call_6_4_15_10(SH2_struct* sh)
{
  SH2negc(sh,4,15);
}
#define init_6_5_0_10
void FASTCALL call_6_5_0_10(SH2_struct* sh)
{
  SH2negc(sh,5,0);
}
#define init_6_5_1_10
void FASTCALL call_6_5_1_10(SH2_struct* sh)
{
  SH2negc(sh,5,1);
}
#define init_6_5_2_10
void FASTCALL call_6_5_2_10(SH2_struct* sh)
{
  SH2negc(sh,5,2);
}
#define init_6_5_3_10
void FASTCALL call_6_5_3_10(SH2_struct* sh)
{
  SH2negc(sh,5,3);
}
#define init_6_5_4_10
void FASTCALL call_6_5_4_10(SH2_struct* sh)
{
  SH2negc(sh,5,4);
}
#define init_6_5_5_10
void FASTCALL call_6_5_5_10(SH2_struct* sh)
{
  SH2negc(sh,5,5);
}
#define init_6_5_6_10
void FASTCALL call_6_5_6_10(SH2_struct* sh)
{
  SH2negc(sh,5,6);
}
#define init_6_5_7_10
void FASTCALL call_6_5_7_10(SH2_struct* sh)
{
  SH2negc(sh,5,7);
}
#define init_6_5_8_10
void FASTCALL call_6_5_8_10(SH2_struct* sh)
{
  SH2negc(sh,5,8);
}
#define init_6_5_9_10
void FASTCALL call_6_5_9_10(SH2_struct* sh)
{
  SH2negc(sh,5,9);
}
#define init_6_5_10_10
void FASTCALL call_6_5_10_10(SH2_struct* sh)
{
  SH2negc(sh,5,10);
}
#define init_6_5_11_10
void FASTCALL call_6_5_11_10(SH2_struct* sh)
{
  SH2negc(sh,5,11);
}
#define init_6_5_12_10
void FASTCALL call_6_5_12_10(SH2_struct* sh)
{
  SH2negc(sh,5,12);
}
#define init_6_5_13_10
void FASTCALL call_6_5_13_10(SH2_struct* sh)
{
  SH2negc(sh,5,13);
}
#define init_6_5_14_10
void FASTCALL call_6_5_14_10(SH2_struct* sh)
{
  SH2negc(sh,5,14);
}
#define init_6_5_15_10
void FASTCALL call_6_5_15_10(SH2_struct* sh)
{
  SH2negc(sh,5,15);
}
#define init_6_6_0_10
void FASTCALL call_6_6_0_10(SH2_struct* sh)
{
  SH2negc(sh,6,0);
}
#define init_6_6_1_10
void FASTCALL call_6_6_1_10(SH2_struct* sh)
{
  SH2negc(sh,6,1);
}
#define init_6_6_2_10
void FASTCALL call_6_6_2_10(SH2_struct* sh)
{
  SH2negc(sh,6,2);
}
#define init_6_6_3_10
void FASTCALL call_6_6_3_10(SH2_struct* sh)
{
  SH2negc(sh,6,3);
}
#define init_6_6_4_10
void FASTCALL call_6_6_4_10(SH2_struct* sh)
{
  SH2negc(sh,6,4);
}
#define init_6_6_5_10
void FASTCALL call_6_6_5_10(SH2_struct* sh)
{
  SH2negc(sh,6,5);
}
#define init_6_6_6_10
void FASTCALL call_6_6_6_10(SH2_struct* sh)
{
  SH2negc(sh,6,6);
}
#define init_6_6_7_10
void FASTCALL call_6_6_7_10(SH2_struct* sh)
{
  SH2negc(sh,6,7);
}
#define init_6_6_8_10
void FASTCALL call_6_6_8_10(SH2_struct* sh)
{
  SH2negc(sh,6,8);
}
#define init_6_6_9_10
void FASTCALL call_6_6_9_10(SH2_struct* sh)
{
  SH2negc(sh,6,9);
}
#define init_6_6_10_10
void FASTCALL call_6_6_10_10(SH2_struct* sh)
{
  SH2negc(sh,6,10);
}
#define init_6_6_11_10
void FASTCALL call_6_6_11_10(SH2_struct* sh)
{
  SH2negc(sh,6,11);
}
#define init_6_6_12_10
void FASTCALL call_6_6_12_10(SH2_struct* sh)
{
  SH2negc(sh,6,12);
}
#define init_6_6_13_10
void FASTCALL call_6_6_13_10(SH2_struct* sh)
{
  SH2negc(sh,6,13);
}
#define init_6_6_14_10
void FASTCALL call_6_6_14_10(SH2_struct* sh)
{
  SH2negc(sh,6,14);
}
#define init_6_6_15_10
void FASTCALL call_6_6_15_10(SH2_struct* sh)
{
  SH2negc(sh,6,15);
}
#define init_6_7_0_10
void FASTCALL call_6_7_0_10(SH2_struct* sh)
{
  SH2negc(sh,7,0);
}
#define init_6_7_1_10
void FASTCALL call_6_7_1_10(SH2_struct* sh)
{
  SH2negc(sh,7,1);
}
#define init_6_7_2_10
void FASTCALL call_6_7_2_10(SH2_struct* sh)
{
  SH2negc(sh,7,2);
}
#define init_6_7_3_10
void FASTCALL call_6_7_3_10(SH2_struct* sh)
{
  SH2negc(sh,7,3);
}
#define init_6_7_4_10
void FASTCALL call_6_7_4_10(SH2_struct* sh)
{
  SH2negc(sh,7,4);
}
#define init_6_7_5_10
void FASTCALL call_6_7_5_10(SH2_struct* sh)
{
  SH2negc(sh,7,5);
}
#define init_6_7_6_10
void FASTCALL call_6_7_6_10(SH2_struct* sh)
{
  SH2negc(sh,7,6);
}
#define init_6_7_7_10
void FASTCALL call_6_7_7_10(SH2_struct* sh)
{
  SH2negc(sh,7,7);
}
#define init_6_7_8_10
void FASTCALL call_6_7_8_10(SH2_struct* sh)
{
  SH2negc(sh,7,8);
}
#define init_6_7_9_10
void FASTCALL call_6_7_9_10(SH2_struct* sh)
{
  SH2negc(sh,7,9);
}
#define init_6_7_10_10
void FASTCALL call_6_7_10_10(SH2_struct* sh)
{
  SH2negc(sh,7,10);
}
#define init_6_7_11_10
void FASTCALL call_6_7_11_10(SH2_struct* sh)
{
  SH2negc(sh,7,11);
}
#define init_6_7_12_10
void FASTCALL call_6_7_12_10(SH2_struct* sh)
{
  SH2negc(sh,7,12);
}
#define init_6_7_13_10
void FASTCALL call_6_7_13_10(SH2_struct* sh)
{
  SH2negc(sh,7,13);
}
#define init_6_7_14_10
void FASTCALL call_6_7_14_10(SH2_struct* sh)
{
  SH2negc(sh,7,14);
}
#define init_6_7_15_10
void FASTCALL call_6_7_15_10(SH2_struct* sh)
{
  SH2negc(sh,7,15);
}
#define init_6_8_0_10
void FASTCALL call_6_8_0_10(SH2_struct* sh)
{
  SH2negc(sh,8,0);
}
#define init_6_8_1_10
void FASTCALL call_6_8_1_10(SH2_struct* sh)
{
  SH2negc(sh,8,1);
}
#define init_6_8_2_10
void FASTCALL call_6_8_2_10(SH2_struct* sh)
{
  SH2negc(sh,8,2);
}
#define init_6_8_3_10
void FASTCALL call_6_8_3_10(SH2_struct* sh)
{
  SH2negc(sh,8,3);
}
#define init_6_8_4_10
void FASTCALL call_6_8_4_10(SH2_struct* sh)
{
  SH2negc(sh,8,4);
}
#define init_6_8_5_10
void FASTCALL call_6_8_5_10(SH2_struct* sh)
{
  SH2negc(sh,8,5);
}
#define init_6_8_6_10
void FASTCALL call_6_8_6_10(SH2_struct* sh)
{
  SH2negc(sh,8,6);
}
#define init_6_8_7_10
void FASTCALL call_6_8_7_10(SH2_struct* sh)
{
  SH2negc(sh,8,7);
}
#define init_6_8_8_10
void FASTCALL call_6_8_8_10(SH2_struct* sh)
{
  SH2negc(sh,8,8);
}
#define init_6_8_9_10
void FASTCALL call_6_8_9_10(SH2_struct* sh)
{
  SH2negc(sh,8,9);
}
#define init_6_8_10_10
void FASTCALL call_6_8_10_10(SH2_struct* sh)
{
  SH2negc(sh,8,10);
}
#define init_6_8_11_10
void FASTCALL call_6_8_11_10(SH2_struct* sh)
{
  SH2negc(sh,8,11);
}
#define init_6_8_12_10
void FASTCALL call_6_8_12_10(SH2_struct* sh)
{
  SH2negc(sh,8,12);
}
#define init_6_8_13_10
void FASTCALL call_6_8_13_10(SH2_struct* sh)
{
  SH2negc(sh,8,13);
}
#define init_6_8_14_10
void FASTCALL call_6_8_14_10(SH2_struct* sh)
{
  SH2negc(sh,8,14);
}
#define init_6_8_15_10
void FASTCALL call_6_8_15_10(SH2_struct* sh)
{
  SH2negc(sh,8,15);
}
#define init_6_9_0_10
void FASTCALL call_6_9_0_10(SH2_struct* sh)
{
  SH2negc(sh,9,0);
}
#define init_6_9_1_10
void FASTCALL call_6_9_1_10(SH2_struct* sh)
{
  SH2negc(sh,9,1);
}
#define init_6_9_2_10
void FASTCALL call_6_9_2_10(SH2_struct* sh)
{
  SH2negc(sh,9,2);
}
#define init_6_9_3_10
void FASTCALL call_6_9_3_10(SH2_struct* sh)
{
  SH2negc(sh,9,3);
}
#define init_6_9_4_10
void FASTCALL call_6_9_4_10(SH2_struct* sh)
{
  SH2negc(sh,9,4);
}
#define init_6_9_5_10
void FASTCALL call_6_9_5_10(SH2_struct* sh)
{
  SH2negc(sh,9,5);
}
#define init_6_9_6_10
void FASTCALL call_6_9_6_10(SH2_struct* sh)
{
  SH2negc(sh,9,6);
}
#define init_6_9_7_10
void FASTCALL call_6_9_7_10(SH2_struct* sh)
{
  SH2negc(sh,9,7);
}
#define init_6_9_8_10
void FASTCALL call_6_9_8_10(SH2_struct* sh)
{
  SH2negc(sh,9,8);
}
#define init_6_9_9_10
void FASTCALL call_6_9_9_10(SH2_struct* sh)
{
  SH2negc(sh,9,9);
}
#define init_6_9_10_10
void FASTCALL call_6_9_10_10(SH2_struct* sh)
{
  SH2negc(sh,9,10);
}
#define init_6_9_11_10
void FASTCALL call_6_9_11_10(SH2_struct* sh)
{
  SH2negc(sh,9,11);
}
#define init_6_9_12_10
void FASTCALL call_6_9_12_10(SH2_struct* sh)
{
  SH2negc(sh,9,12);
}
#define init_6_9_13_10
void FASTCALL call_6_9_13_10(SH2_struct* sh)
{
  SH2negc(sh,9,13);
}
#define init_6_9_14_10
void FASTCALL call_6_9_14_10(SH2_struct* sh)
{
  SH2negc(sh,9,14);
}
#define init_6_9_15_10
void FASTCALL call_6_9_15_10(SH2_struct* sh)
{
  SH2negc(sh,9,15);
}
#define init_6_10_0_10
void FASTCALL call_6_10_0_10(SH2_struct* sh)
{
  SH2negc(sh,10,0);
}
#define init_6_10_1_10
void FASTCALL call_6_10_1_10(SH2_struct* sh)
{
  SH2negc(sh,10,1);
}
#define init_6_10_2_10
void FASTCALL call_6_10_2_10(SH2_struct* sh)
{
  SH2negc(sh,10,2);
}
#define init_6_10_3_10
void FASTCALL call_6_10_3_10(SH2_struct* sh)
{
  SH2negc(sh,10,3);
}
#define init_6_10_4_10
void FASTCALL call_6_10_4_10(SH2_struct* sh)
{
  SH2negc(sh,10,4);
}
#define init_6_10_5_10
void FASTCALL call_6_10_5_10(SH2_struct* sh)
{
  SH2negc(sh,10,5);
}
#define init_6_10_6_10
void FASTCALL call_6_10_6_10(SH2_struct* sh)
{
  SH2negc(sh,10,6);
}
#define init_6_10_7_10
void FASTCALL call_6_10_7_10(SH2_struct* sh)
{
  SH2negc(sh,10,7);
}
#define init_6_10_8_10
void FASTCALL call_6_10_8_10(SH2_struct* sh)
{
  SH2negc(sh,10,8);
}
#define init_6_10_9_10
void FASTCALL call_6_10_9_10(SH2_struct* sh)
{
  SH2negc(sh,10,9);
}
#define init_6_10_10_10
void FASTCALL call_6_10_10_10(SH2_struct* sh)
{
  SH2negc(sh,10,10);
}
#define init_6_10_11_10
void FASTCALL call_6_10_11_10(SH2_struct* sh)
{
  SH2negc(sh,10,11);
}
#define init_6_10_12_10
void FASTCALL call_6_10_12_10(SH2_struct* sh)
{
  SH2negc(sh,10,12);
}
#define init_6_10_13_10
void FASTCALL call_6_10_13_10(SH2_struct* sh)
{
  SH2negc(sh,10,13);
}
#define init_6_10_14_10
void FASTCALL call_6_10_14_10(SH2_struct* sh)
{
  SH2negc(sh,10,14);
}
#define init_6_10_15_10
void FASTCALL call_6_10_15_10(SH2_struct* sh)
{
  SH2negc(sh,10,15);
}
#define init_6_11_0_10
void FASTCALL call_6_11_0_10(SH2_struct* sh)
{
  SH2negc(sh,11,0);
}
#define init_6_11_1_10
void FASTCALL call_6_11_1_10(SH2_struct* sh)
{
  SH2negc(sh,11,1);
}
#define init_6_11_2_10
void FASTCALL call_6_11_2_10(SH2_struct* sh)
{
  SH2negc(sh,11,2);
}
#define init_6_11_3_10
void FASTCALL call_6_11_3_10(SH2_struct* sh)
{
  SH2negc(sh,11,3);
}
#define init_6_11_4_10
void FASTCALL call_6_11_4_10(SH2_struct* sh)
{
  SH2negc(sh,11,4);
}
#define init_6_11_5_10
void FASTCALL call_6_11_5_10(SH2_struct* sh)
{
  SH2negc(sh,11,5);
}
#define init_6_11_6_10
void FASTCALL call_6_11_6_10(SH2_struct* sh)
{
  SH2negc(sh,11,6);
}
#define init_6_11_7_10
void FASTCALL call_6_11_7_10(SH2_struct* sh)
{
  SH2negc(sh,11,7);
}
#define init_6_11_8_10
void FASTCALL call_6_11_8_10(SH2_struct* sh)
{
  SH2negc(sh,11,8);
}
#define init_6_11_9_10
void FASTCALL call_6_11_9_10(SH2_struct* sh)
{
  SH2negc(sh,11,9);
}
#define init_6_11_10_10
void FASTCALL call_6_11_10_10(SH2_struct* sh)
{
  SH2negc(sh,11,10);
}
#define init_6_11_11_10
void FASTCALL call_6_11_11_10(SH2_struct* sh)
{
  SH2negc(sh,11,11);
}
#define init_6_11_12_10
void FASTCALL call_6_11_12_10(SH2_struct* sh)
{
  SH2negc(sh,11,12);
}
#define init_6_11_13_10
void FASTCALL call_6_11_13_10(SH2_struct* sh)
{
  SH2negc(sh,11,13);
}
#define init_6_11_14_10
void FASTCALL call_6_11_14_10(SH2_struct* sh)
{
  SH2negc(sh,11,14);
}
#define init_6_11_15_10
void FASTCALL call_6_11_15_10(SH2_struct* sh)
{
  SH2negc(sh,11,15);
}
#define init_6_12_0_10
void FASTCALL call_6_12_0_10(SH2_struct* sh)
{
  SH2negc(sh,12,0);
}
#define init_6_12_1_10
void FASTCALL call_6_12_1_10(SH2_struct* sh)
{
  SH2negc(sh,12,1);
}
#define init_6_12_2_10
void FASTCALL call_6_12_2_10(SH2_struct* sh)
{
  SH2negc(sh,12,2);
}
#define init_6_12_3_10
void FASTCALL call_6_12_3_10(SH2_struct* sh)
{
  SH2negc(sh,12,3);
}
#define init_6_12_4_10
void FASTCALL call_6_12_4_10(SH2_struct* sh)
{
  SH2negc(sh,12,4);
}
#define init_6_12_5_10
void FASTCALL call_6_12_5_10(SH2_struct* sh)
{
  SH2negc(sh,12,5);
}
#define init_6_12_6_10
void FASTCALL call_6_12_6_10(SH2_struct* sh)
{
  SH2negc(sh,12,6);
}
#define init_6_12_7_10
void FASTCALL call_6_12_7_10(SH2_struct* sh)
{
  SH2negc(sh,12,7);
}
#define init_6_12_8_10
void FASTCALL call_6_12_8_10(SH2_struct* sh)
{
  SH2negc(sh,12,8);
}
#define init_6_12_9_10
void FASTCALL call_6_12_9_10(SH2_struct* sh)
{
  SH2negc(sh,12,9);
}
#define init_6_12_10_10
void FASTCALL call_6_12_10_10(SH2_struct* sh)
{
  SH2negc(sh,12,10);
}
#define init_6_12_11_10
void FASTCALL call_6_12_11_10(SH2_struct* sh)
{
  SH2negc(sh,12,11);
}
#define init_6_12_12_10
void FASTCALL call_6_12_12_10(SH2_struct* sh)
{
  SH2negc(sh,12,12);
}
#define init_6_12_13_10
void FASTCALL call_6_12_13_10(SH2_struct* sh)
{
  SH2negc(sh,12,13);
}
#define init_6_12_14_10
void FASTCALL call_6_12_14_10(SH2_struct* sh)
{
  SH2negc(sh,12,14);
}
#define init_6_12_15_10
void FASTCALL call_6_12_15_10(SH2_struct* sh)
{
  SH2negc(sh,12,15);
}
#define init_6_13_0_10
void FASTCALL call_6_13_0_10(SH2_struct* sh)
{
  SH2negc(sh,13,0);
}
#define init_6_13_1_10
void FASTCALL call_6_13_1_10(SH2_struct* sh)
{
  SH2negc(sh,13,1);
}
#define init_6_13_2_10
void FASTCALL call_6_13_2_10(SH2_struct* sh)
{
  SH2negc(sh,13,2);
}
#define init_6_13_3_10
void FASTCALL call_6_13_3_10(SH2_struct* sh)
{
  SH2negc(sh,13,3);
}
#define init_6_13_4_10
void FASTCALL call_6_13_4_10(SH2_struct* sh)
{
  SH2negc(sh,13,4);
}
#define init_6_13_5_10
void FASTCALL call_6_13_5_10(SH2_struct* sh)
{
  SH2negc(sh,13,5);
}
#define init_6_13_6_10
void FASTCALL call_6_13_6_10(SH2_struct* sh)
{
  SH2negc(sh,13,6);
}
#define init_6_13_7_10
void FASTCALL call_6_13_7_10(SH2_struct* sh)
{
  SH2negc(sh,13,7);
}
#define init_6_13_8_10
void FASTCALL call_6_13_8_10(SH2_struct* sh)
{
  SH2negc(sh,13,8);
}
#define init_6_13_9_10
void FASTCALL call_6_13_9_10(SH2_struct* sh)
{
  SH2negc(sh,13,9);
}
#define init_6_13_10_10
void FASTCALL call_6_13_10_10(SH2_struct* sh)
{
  SH2negc(sh,13,10);
}
#define init_6_13_11_10
void FASTCALL call_6_13_11_10(SH2_struct* sh)
{
  SH2negc(sh,13,11);
}
#define init_6_13_12_10
void FASTCALL call_6_13_12_10(SH2_struct* sh)
{
  SH2negc(sh,13,12);
}
#define init_6_13_13_10
void FASTCALL call_6_13_13_10(SH2_struct* sh)
{
  SH2negc(sh,13,13);
}
#define init_6_13_14_10
void FASTCALL call_6_13_14_10(SH2_struct* sh)
{
  SH2negc(sh,13,14);
}
#define init_6_13_15_10
void FASTCALL call_6_13_15_10(SH2_struct* sh)
{
  SH2negc(sh,13,15);
}
#define init_6_14_0_10
void FASTCALL call_6_14_0_10(SH2_struct* sh)
{
  SH2negc(sh,14,0);
}
#define init_6_14_1_10
void FASTCALL call_6_14_1_10(SH2_struct* sh)
{
  SH2negc(sh,14,1);
}
#define init_6_14_2_10
void FASTCALL call_6_14_2_10(SH2_struct* sh)
{
  SH2negc(sh,14,2);
}
#define init_6_14_3_10
void FASTCALL call_6_14_3_10(SH2_struct* sh)
{
  SH2negc(sh,14,3);
}
#define init_6_14_4_10
void FASTCALL call_6_14_4_10(SH2_struct* sh)
{
  SH2negc(sh,14,4);
}
#define init_6_14_5_10
void FASTCALL call_6_14_5_10(SH2_struct* sh)
{
  SH2negc(sh,14,5);
}
#define init_6_14_6_10
void FASTCALL call_6_14_6_10(SH2_struct* sh)
{
  SH2negc(sh,14,6);
}
#define init_6_14_7_10
void FASTCALL call_6_14_7_10(SH2_struct* sh)
{
  SH2negc(sh,14,7);
}
#define init_6_14_8_10
void FASTCALL call_6_14_8_10(SH2_struct* sh)
{
  SH2negc(sh,14,8);
}
#define init_6_14_9_10
void FASTCALL call_6_14_9_10(SH2_struct* sh)
{
  SH2negc(sh,14,9);
}
#define init_6_14_10_10
void FASTCALL call_6_14_10_10(SH2_struct* sh)
{
  SH2negc(sh,14,10);
}
#define init_6_14_11_10
void FASTCALL call_6_14_11_10(SH2_struct* sh)
{
  SH2negc(sh,14,11);
}
#define init_6_14_12_10
void FASTCALL call_6_14_12_10(SH2_struct* sh)
{
  SH2negc(sh,14,12);
}
#define init_6_14_13_10
void FASTCALL call_6_14_13_10(SH2_struct* sh)
{
  SH2negc(sh,14,13);
}
#define init_6_14_14_10
void FASTCALL call_6_14_14_10(SH2_struct* sh)
{
  SH2negc(sh,14,14);
}
#define init_6_14_15_10
void FASTCALL call_6_14_15_10(SH2_struct* sh)
{
  SH2negc(sh,14,15);
}
#define init_6_15_0_10
void FASTCALL call_6_15_0_10(SH2_struct* sh)
{
  SH2negc(sh,15,0);
}
#define init_6_15_1_10
void FASTCALL call_6_15_1_10(SH2_struct* sh)
{
  SH2negc(sh,15,1);
}
#define init_6_15_2_10
void FASTCALL call_6_15_2_10(SH2_struct* sh)
{
  SH2negc(sh,15,2);
}
#define init_6_15_3_10
void FASTCALL call_6_15_3_10(SH2_struct* sh)
{
  SH2negc(sh,15,3);
}
#define init_6_15_4_10
void FASTCALL call_6_15_4_10(SH2_struct* sh)
{
  SH2negc(sh,15,4);
}
#define init_6_15_5_10
void FASTCALL call_6_15_5_10(SH2_struct* sh)
{
  SH2negc(sh,15,5);
}
#define init_6_15_6_10
void FASTCALL call_6_15_6_10(SH2_struct* sh)
{
  SH2negc(sh,15,6);
}
#define init_6_15_7_10
void FASTCALL call_6_15_7_10(SH2_struct* sh)
{
  SH2negc(sh,15,7);
}
#define init_6_15_8_10
void FASTCALL call_6_15_8_10(SH2_struct* sh)
{
  SH2negc(sh,15,8);
}
#define init_6_15_9_10
void FASTCALL call_6_15_9_10(SH2_struct* sh)
{
  SH2negc(sh,15,9);
}
#define init_6_15_10_10
void FASTCALL call_6_15_10_10(SH2_struct* sh)
{
  SH2negc(sh,15,10);
}
#define init_6_15_11_10
void FASTCALL call_6_15_11_10(SH2_struct* sh)
{
  SH2negc(sh,15,11);
}
#define init_6_15_12_10
void FASTCALL call_6_15_12_10(SH2_struct* sh)
{
  SH2negc(sh,15,12);
}
#define init_6_15_13_10
void FASTCALL call_6_15_13_10(SH2_struct* sh)
{
  SH2negc(sh,15,13);
}
#define init_6_15_14_10
void FASTCALL call_6_15_14_10(SH2_struct* sh)
{
  SH2negc(sh,15,14);
}
#define init_6_15_15_10
void FASTCALL call_6_15_15_10(SH2_struct* sh)
{
  SH2negc(sh,15,15);
}
#define init_6_0_0_11
void FASTCALL call_6_0_0_11(SH2_struct* sh)
{
  SH2neg(sh,0,0);
}
#define init_6_0_1_11
void FASTCALL call_6_0_1_11(SH2_struct* sh)
{
  SH2neg(sh,0,1);
}
#define init_6_0_2_11
void FASTCALL call_6_0_2_11(SH2_struct* sh)
{
  SH2neg(sh,0,2);
}
#define init_6_0_3_11
void FASTCALL call_6_0_3_11(SH2_struct* sh)
{
  SH2neg(sh,0,3);
}
#define init_6_0_4_11
void FASTCALL call_6_0_4_11(SH2_struct* sh)
{
  SH2neg(sh,0,4);
}
#define init_6_0_5_11
void FASTCALL call_6_0_5_11(SH2_struct* sh)
{
  SH2neg(sh,0,5);
}
#define init_6_0_6_11
void FASTCALL call_6_0_6_11(SH2_struct* sh)
{
  SH2neg(sh,0,6);
}
#define init_6_0_7_11
void FASTCALL call_6_0_7_11(SH2_struct* sh)
{
  SH2neg(sh,0,7);
}
#define init_6_0_8_11
void FASTCALL call_6_0_8_11(SH2_struct* sh)
{
  SH2neg(sh,0,8);
}
#define init_6_0_9_11
void FASTCALL call_6_0_9_11(SH2_struct* sh)
{
  SH2neg(sh,0,9);
}
#define init_6_0_10_11
void FASTCALL call_6_0_10_11(SH2_struct* sh)
{
  SH2neg(sh,0,10);
}
#define init_6_0_11_11
void FASTCALL call_6_0_11_11(SH2_struct* sh)
{
  SH2neg(sh,0,11);
}
#define init_6_0_12_11
void FASTCALL call_6_0_12_11(SH2_struct* sh)
{
  SH2neg(sh,0,12);
}
#define init_6_0_13_11
void FASTCALL call_6_0_13_11(SH2_struct* sh)
{
  SH2neg(sh,0,13);
}
#define init_6_0_14_11
void FASTCALL call_6_0_14_11(SH2_struct* sh)
{
  SH2neg(sh,0,14);
}
#define init_6_0_15_11
void FASTCALL call_6_0_15_11(SH2_struct* sh)
{
  SH2neg(sh,0,15);
}
#define init_6_1_0_11
void FASTCALL call_6_1_0_11(SH2_struct* sh)
{
  SH2neg(sh,1,0);
}
#define init_6_1_1_11
void FASTCALL call_6_1_1_11(SH2_struct* sh)
{
  SH2neg(sh,1,1);
}
#define init_6_1_2_11
void FASTCALL call_6_1_2_11(SH2_struct* sh)
{
  SH2neg(sh,1,2);
}
#define init_6_1_3_11
void FASTCALL call_6_1_3_11(SH2_struct* sh)
{
  SH2neg(sh,1,3);
}
#define init_6_1_4_11
void FASTCALL call_6_1_4_11(SH2_struct* sh)
{
  SH2neg(sh,1,4);
}
#define init_6_1_5_11
void FASTCALL call_6_1_5_11(SH2_struct* sh)
{
  SH2neg(sh,1,5);
}
#define init_6_1_6_11
void FASTCALL call_6_1_6_11(SH2_struct* sh)
{
  SH2neg(sh,1,6);
}
#define init_6_1_7_11
void FASTCALL call_6_1_7_11(SH2_struct* sh)
{
  SH2neg(sh,1,7);
}
#define init_6_1_8_11
void FASTCALL call_6_1_8_11(SH2_struct* sh)
{
  SH2neg(sh,1,8);
}
#define init_6_1_9_11
void FASTCALL call_6_1_9_11(SH2_struct* sh)
{
  SH2neg(sh,1,9);
}
#define init_6_1_10_11
void FASTCALL call_6_1_10_11(SH2_struct* sh)
{
  SH2neg(sh,1,10);
}
#define init_6_1_11_11
void FASTCALL call_6_1_11_11(SH2_struct* sh)
{
  SH2neg(sh,1,11);
}
#define init_6_1_12_11
void FASTCALL call_6_1_12_11(SH2_struct* sh)
{
  SH2neg(sh,1,12);
}
#define init_6_1_13_11
void FASTCALL call_6_1_13_11(SH2_struct* sh)
{
  SH2neg(sh,1,13);
}
#define init_6_1_14_11
void FASTCALL call_6_1_14_11(SH2_struct* sh)
{
  SH2neg(sh,1,14);
}
#define init_6_1_15_11
void FASTCALL call_6_1_15_11(SH2_struct* sh)
{
  SH2neg(sh,1,15);
}
#define init_6_2_0_11
void FASTCALL call_6_2_0_11(SH2_struct* sh)
{
  SH2neg(sh,2,0);
}
#define init_6_2_1_11
void FASTCALL call_6_2_1_11(SH2_struct* sh)
{
  SH2neg(sh,2,1);
}
#define init_6_2_2_11
void FASTCALL call_6_2_2_11(SH2_struct* sh)
{
  SH2neg(sh,2,2);
}
#define init_6_2_3_11
void FASTCALL call_6_2_3_11(SH2_struct* sh)
{
  SH2neg(sh,2,3);
}
#define init_6_2_4_11
void FASTCALL call_6_2_4_11(SH2_struct* sh)
{
  SH2neg(sh,2,4);
}
#define init_6_2_5_11
void FASTCALL call_6_2_5_11(SH2_struct* sh)
{
  SH2neg(sh,2,5);
}
#define init_6_2_6_11
void FASTCALL call_6_2_6_11(SH2_struct* sh)
{
  SH2neg(sh,2,6);
}
#define init_6_2_7_11
void FASTCALL call_6_2_7_11(SH2_struct* sh)
{
  SH2neg(sh,2,7);
}
#define init_6_2_8_11
void FASTCALL call_6_2_8_11(SH2_struct* sh)
{
  SH2neg(sh,2,8);
}
#define init_6_2_9_11
void FASTCALL call_6_2_9_11(SH2_struct* sh)
{
  SH2neg(sh,2,9);
}
#define init_6_2_10_11
void FASTCALL call_6_2_10_11(SH2_struct* sh)
{
  SH2neg(sh,2,10);
}
#define init_6_2_11_11
void FASTCALL call_6_2_11_11(SH2_struct* sh)
{
  SH2neg(sh,2,11);
}
#define init_6_2_12_11
void FASTCALL call_6_2_12_11(SH2_struct* sh)
{
  SH2neg(sh,2,12);
}
#define init_6_2_13_11
void FASTCALL call_6_2_13_11(SH2_struct* sh)
{
  SH2neg(sh,2,13);
}
#define init_6_2_14_11
void FASTCALL call_6_2_14_11(SH2_struct* sh)
{
  SH2neg(sh,2,14);
}
#define init_6_2_15_11
void FASTCALL call_6_2_15_11(SH2_struct* sh)
{
  SH2neg(sh,2,15);
}
#define init_6_3_0_11
void FASTCALL call_6_3_0_11(SH2_struct* sh)
{
  SH2neg(sh,3,0);
}
#define init_6_3_1_11
void FASTCALL call_6_3_1_11(SH2_struct* sh)
{
  SH2neg(sh,3,1);
}
#define init_6_3_2_11
void FASTCALL call_6_3_2_11(SH2_struct* sh)
{
  SH2neg(sh,3,2);
}
#define init_6_3_3_11
void FASTCALL call_6_3_3_11(SH2_struct* sh)
{
  SH2neg(sh,3,3);
}
#define init_6_3_4_11
void FASTCALL call_6_3_4_11(SH2_struct* sh)
{
  SH2neg(sh,3,4);
}
#define init_6_3_5_11
void FASTCALL call_6_3_5_11(SH2_struct* sh)
{
  SH2neg(sh,3,5);
}
#define init_6_3_6_11
void FASTCALL call_6_3_6_11(SH2_struct* sh)
{
  SH2neg(sh,3,6);
}
#define init_6_3_7_11
void FASTCALL call_6_3_7_11(SH2_struct* sh)
{
  SH2neg(sh,3,7);
}
#define init_6_3_8_11
void FASTCALL call_6_3_8_11(SH2_struct* sh)
{
  SH2neg(sh,3,8);
}
#define init_6_3_9_11
void FASTCALL call_6_3_9_11(SH2_struct* sh)
{
  SH2neg(sh,3,9);
}
#define init_6_3_10_11
void FASTCALL call_6_3_10_11(SH2_struct* sh)
{
  SH2neg(sh,3,10);
}
#define init_6_3_11_11
void FASTCALL call_6_3_11_11(SH2_struct* sh)
{
  SH2neg(sh,3,11);
}
#define init_6_3_12_11
void FASTCALL call_6_3_12_11(SH2_struct* sh)
{
  SH2neg(sh,3,12);
}
#define init_6_3_13_11
void FASTCALL call_6_3_13_11(SH2_struct* sh)
{
  SH2neg(sh,3,13);
}
#define init_6_3_14_11
void FASTCALL call_6_3_14_11(SH2_struct* sh)
{
  SH2neg(sh,3,14);
}
#define init_6_3_15_11
void FASTCALL call_6_3_15_11(SH2_struct* sh)
{
  SH2neg(sh,3,15);
}
#define init_6_4_0_11
void FASTCALL call_6_4_0_11(SH2_struct* sh)
{
  SH2neg(sh,4,0);
}
#define init_6_4_1_11
void FASTCALL call_6_4_1_11(SH2_struct* sh)
{
  SH2neg(sh,4,1);
}
#define init_6_4_2_11
void FASTCALL call_6_4_2_11(SH2_struct* sh)
{
  SH2neg(sh,4,2);
}
#define init_6_4_3_11
void FASTCALL call_6_4_3_11(SH2_struct* sh)
{
  SH2neg(sh,4,3);
}
#define init_6_4_4_11
void FASTCALL call_6_4_4_11(SH2_struct* sh)
{
  SH2neg(sh,4,4);
}
#define init_6_4_5_11
void FASTCALL call_6_4_5_11(SH2_struct* sh)
{
  SH2neg(sh,4,5);
}
#define init_6_4_6_11
void FASTCALL call_6_4_6_11(SH2_struct* sh)
{
  SH2neg(sh,4,6);
}
#define init_6_4_7_11
void FASTCALL call_6_4_7_11(SH2_struct* sh)
{
  SH2neg(sh,4,7);
}
#define init_6_4_8_11
void FASTCALL call_6_4_8_11(SH2_struct* sh)
{
  SH2neg(sh,4,8);
}
#define init_6_4_9_11
void FASTCALL call_6_4_9_11(SH2_struct* sh)
{
  SH2neg(sh,4,9);
}
#define init_6_4_10_11
void FASTCALL call_6_4_10_11(SH2_struct* sh)
{
  SH2neg(sh,4,10);
}
#define init_6_4_11_11
void FASTCALL call_6_4_11_11(SH2_struct* sh)
{
  SH2neg(sh,4,11);
}
#define init_6_4_12_11
void FASTCALL call_6_4_12_11(SH2_struct* sh)
{
  SH2neg(sh,4,12);
}
#define init_6_4_13_11
void FASTCALL call_6_4_13_11(SH2_struct* sh)
{
  SH2neg(sh,4,13);
}
#define init_6_4_14_11
void FASTCALL call_6_4_14_11(SH2_struct* sh)
{
  SH2neg(sh,4,14);
}
#define init_6_4_15_11
void FASTCALL call_6_4_15_11(SH2_struct* sh)
{
  SH2neg(sh,4,15);
}
#define init_6_5_0_11
void FASTCALL call_6_5_0_11(SH2_struct* sh)
{
  SH2neg(sh,5,0);
}
#define init_6_5_1_11
void FASTCALL call_6_5_1_11(SH2_struct* sh)
{
  SH2neg(sh,5,1);
}
#define init_6_5_2_11
void FASTCALL call_6_5_2_11(SH2_struct* sh)
{
  SH2neg(sh,5,2);
}
#define init_6_5_3_11
void FASTCALL call_6_5_3_11(SH2_struct* sh)
{
  SH2neg(sh,5,3);
}
#define init_6_5_4_11
void FASTCALL call_6_5_4_11(SH2_struct* sh)
{
  SH2neg(sh,5,4);
}
#define init_6_5_5_11
void FASTCALL call_6_5_5_11(SH2_struct* sh)
{
  SH2neg(sh,5,5);
}
#define init_6_5_6_11
void FASTCALL call_6_5_6_11(SH2_struct* sh)
{
  SH2neg(sh,5,6);
}
#define init_6_5_7_11
void FASTCALL call_6_5_7_11(SH2_struct* sh)
{
  SH2neg(sh,5,7);
}
#define init_6_5_8_11
void FASTCALL call_6_5_8_11(SH2_struct* sh)
{
  SH2neg(sh,5,8);
}
#define init_6_5_9_11
void FASTCALL call_6_5_9_11(SH2_struct* sh)
{
  SH2neg(sh,5,9);
}
#define init_6_5_10_11
void FASTCALL call_6_5_10_11(SH2_struct* sh)
{
  SH2neg(sh,5,10);
}
#define init_6_5_11_11
void FASTCALL call_6_5_11_11(SH2_struct* sh)
{
  SH2neg(sh,5,11);
}
#define init_6_5_12_11
void FASTCALL call_6_5_12_11(SH2_struct* sh)
{
  SH2neg(sh,5,12);
}
#define init_6_5_13_11
void FASTCALL call_6_5_13_11(SH2_struct* sh)
{
  SH2neg(sh,5,13);
}
#define init_6_5_14_11
void FASTCALL call_6_5_14_11(SH2_struct* sh)
{
  SH2neg(sh,5,14);
}
#define init_6_5_15_11
void FASTCALL call_6_5_15_11(SH2_struct* sh)
{
  SH2neg(sh,5,15);
}
#define init_6_6_0_11
void FASTCALL call_6_6_0_11(SH2_struct* sh)
{
  SH2neg(sh,6,0);
}
#define init_6_6_1_11
void FASTCALL call_6_6_1_11(SH2_struct* sh)
{
  SH2neg(sh,6,1);
}
#define init_6_6_2_11
void FASTCALL call_6_6_2_11(SH2_struct* sh)
{
  SH2neg(sh,6,2);
}
#define init_6_6_3_11
void FASTCALL call_6_6_3_11(SH2_struct* sh)
{
  SH2neg(sh,6,3);
}
#define init_6_6_4_11
void FASTCALL call_6_6_4_11(SH2_struct* sh)
{
  SH2neg(sh,6,4);
}
#define init_6_6_5_11
void FASTCALL call_6_6_5_11(SH2_struct* sh)
{
  SH2neg(sh,6,5);
}
#define init_6_6_6_11
void FASTCALL call_6_6_6_11(SH2_struct* sh)
{
  SH2neg(sh,6,6);
}
#define init_6_6_7_11
void FASTCALL call_6_6_7_11(SH2_struct* sh)
{
  SH2neg(sh,6,7);
}
#define init_6_6_8_11
void FASTCALL call_6_6_8_11(SH2_struct* sh)
{
  SH2neg(sh,6,8);
}
#define init_6_6_9_11
void FASTCALL call_6_6_9_11(SH2_struct* sh)
{
  SH2neg(sh,6,9);
}
#define init_6_6_10_11
void FASTCALL call_6_6_10_11(SH2_struct* sh)
{
  SH2neg(sh,6,10);
}
#define init_6_6_11_11
void FASTCALL call_6_6_11_11(SH2_struct* sh)
{
  SH2neg(sh,6,11);
}
#define init_6_6_12_11
void FASTCALL call_6_6_12_11(SH2_struct* sh)
{
  SH2neg(sh,6,12);
}
#define init_6_6_13_11
void FASTCALL call_6_6_13_11(SH2_struct* sh)
{
  SH2neg(sh,6,13);
}
#define init_6_6_14_11
void FASTCALL call_6_6_14_11(SH2_struct* sh)
{
  SH2neg(sh,6,14);
}
#define init_6_6_15_11
void FASTCALL call_6_6_15_11(SH2_struct* sh)
{
  SH2neg(sh,6,15);
}
#define init_6_7_0_11
void FASTCALL call_6_7_0_11(SH2_struct* sh)
{
  SH2neg(sh,7,0);
}
#define init_6_7_1_11
void FASTCALL call_6_7_1_11(SH2_struct* sh)
{
  SH2neg(sh,7,1);
}
#define init_6_7_2_11
void FASTCALL call_6_7_2_11(SH2_struct* sh)
{
  SH2neg(sh,7,2);
}
#define init_6_7_3_11
void FASTCALL call_6_7_3_11(SH2_struct* sh)
{
  SH2neg(sh,7,3);
}
#define init_6_7_4_11
void FASTCALL call_6_7_4_11(SH2_struct* sh)
{
  SH2neg(sh,7,4);
}
#define init_6_7_5_11
void FASTCALL call_6_7_5_11(SH2_struct* sh)
{
  SH2neg(sh,7,5);
}
#define init_6_7_6_11
void FASTCALL call_6_7_6_11(SH2_struct* sh)
{
  SH2neg(sh,7,6);
}
#define init_6_7_7_11
void FASTCALL call_6_7_7_11(SH2_struct* sh)
{
  SH2neg(sh,7,7);
}
#define init_6_7_8_11
void FASTCALL call_6_7_8_11(SH2_struct* sh)
{
  SH2neg(sh,7,8);
}
#define init_6_7_9_11
void FASTCALL call_6_7_9_11(SH2_struct* sh)
{
  SH2neg(sh,7,9);
}
#define init_6_7_10_11
void FASTCALL call_6_7_10_11(SH2_struct* sh)
{
  SH2neg(sh,7,10);
}
#define init_6_7_11_11
void FASTCALL call_6_7_11_11(SH2_struct* sh)
{
  SH2neg(sh,7,11);
}
#define init_6_7_12_11
void FASTCALL call_6_7_12_11(SH2_struct* sh)
{
  SH2neg(sh,7,12);
}
#define init_6_7_13_11
void FASTCALL call_6_7_13_11(SH2_struct* sh)
{
  SH2neg(sh,7,13);
}
#define init_6_7_14_11
void FASTCALL call_6_7_14_11(SH2_struct* sh)
{
  SH2neg(sh,7,14);
}
#define init_6_7_15_11
void FASTCALL call_6_7_15_11(SH2_struct* sh)
{
  SH2neg(sh,7,15);
}
#define init_6_8_0_11
void FASTCALL call_6_8_0_11(SH2_struct* sh)
{
  SH2neg(sh,8,0);
}
#define init_6_8_1_11
void FASTCALL call_6_8_1_11(SH2_struct* sh)
{
  SH2neg(sh,8,1);
}
#define init_6_8_2_11
void FASTCALL call_6_8_2_11(SH2_struct* sh)
{
  SH2neg(sh,8,2);
}
#define init_6_8_3_11
void FASTCALL call_6_8_3_11(SH2_struct* sh)
{
  SH2neg(sh,8,3);
}
#define init_6_8_4_11
void FASTCALL call_6_8_4_11(SH2_struct* sh)
{
  SH2neg(sh,8,4);
}
#define init_6_8_5_11
void FASTCALL call_6_8_5_11(SH2_struct* sh)
{
  SH2neg(sh,8,5);
}
#define init_6_8_6_11
void FASTCALL call_6_8_6_11(SH2_struct* sh)
{
  SH2neg(sh,8,6);
}
#define init_6_8_7_11
void FASTCALL call_6_8_7_11(SH2_struct* sh)
{
  SH2neg(sh,8,7);
}
#define init_6_8_8_11
void FASTCALL call_6_8_8_11(SH2_struct* sh)
{
  SH2neg(sh,8,8);
}
#define init_6_8_9_11
void FASTCALL call_6_8_9_11(SH2_struct* sh)
{
  SH2neg(sh,8,9);
}
#define init_6_8_10_11
void FASTCALL call_6_8_10_11(SH2_struct* sh)
{
  SH2neg(sh,8,10);
}
#define init_6_8_11_11
void FASTCALL call_6_8_11_11(SH2_struct* sh)
{
  SH2neg(sh,8,11);
}
#define init_6_8_12_11
void FASTCALL call_6_8_12_11(SH2_struct* sh)
{
  SH2neg(sh,8,12);
}
#define init_6_8_13_11
void FASTCALL call_6_8_13_11(SH2_struct* sh)
{
  SH2neg(sh,8,13);
}
#define init_6_8_14_11
void FASTCALL call_6_8_14_11(SH2_struct* sh)
{
  SH2neg(sh,8,14);
}
#define init_6_8_15_11
void FASTCALL call_6_8_15_11(SH2_struct* sh)
{
  SH2neg(sh,8,15);
}
#define init_6_9_0_11
void FASTCALL call_6_9_0_11(SH2_struct* sh)
{
  SH2neg(sh,9,0);
}
#define init_6_9_1_11
void FASTCALL call_6_9_1_11(SH2_struct* sh)
{
  SH2neg(sh,9,1);
}
#define init_6_9_2_11
void FASTCALL call_6_9_2_11(SH2_struct* sh)
{
  SH2neg(sh,9,2);
}
#define init_6_9_3_11
void FASTCALL call_6_9_3_11(SH2_struct* sh)
{
  SH2neg(sh,9,3);
}
#define init_6_9_4_11
void FASTCALL call_6_9_4_11(SH2_struct* sh)
{
  SH2neg(sh,9,4);
}
#define init_6_9_5_11
void FASTCALL call_6_9_5_11(SH2_struct* sh)
{
  SH2neg(sh,9,5);
}
#define init_6_9_6_11
void FASTCALL call_6_9_6_11(SH2_struct* sh)
{
  SH2neg(sh,9,6);
}
#define init_6_9_7_11
void FASTCALL call_6_9_7_11(SH2_struct* sh)
{
  SH2neg(sh,9,7);
}
#define init_6_9_8_11
void FASTCALL call_6_9_8_11(SH2_struct* sh)
{
  SH2neg(sh,9,8);
}
#define init_6_9_9_11
void FASTCALL call_6_9_9_11(SH2_struct* sh)
{
  SH2neg(sh,9,9);
}
#define init_6_9_10_11
void FASTCALL call_6_9_10_11(SH2_struct* sh)
{
  SH2neg(sh,9,10);
}
#define init_6_9_11_11
void FASTCALL call_6_9_11_11(SH2_struct* sh)
{
  SH2neg(sh,9,11);
}
#define init_6_9_12_11
void FASTCALL call_6_9_12_11(SH2_struct* sh)
{
  SH2neg(sh,9,12);
}
#define init_6_9_13_11
void FASTCALL call_6_9_13_11(SH2_struct* sh)
{
  SH2neg(sh,9,13);
}
#define init_6_9_14_11
void FASTCALL call_6_9_14_11(SH2_struct* sh)
{
  SH2neg(sh,9,14);
}
#define init_6_9_15_11
void FASTCALL call_6_9_15_11(SH2_struct* sh)
{
  SH2neg(sh,9,15);
}
#define init_6_10_0_11
void FASTCALL call_6_10_0_11(SH2_struct* sh)
{
  SH2neg(sh,10,0);
}
#define init_6_10_1_11
void FASTCALL call_6_10_1_11(SH2_struct* sh)
{
  SH2neg(sh,10,1);
}
#define init_6_10_2_11
void FASTCALL call_6_10_2_11(SH2_struct* sh)
{
  SH2neg(sh,10,2);
}
#define init_6_10_3_11
void FASTCALL call_6_10_3_11(SH2_struct* sh)
{
  SH2neg(sh,10,3);
}
#define init_6_10_4_11
void FASTCALL call_6_10_4_11(SH2_struct* sh)
{
  SH2neg(sh,10,4);
}
#define init_6_10_5_11
void FASTCALL call_6_10_5_11(SH2_struct* sh)
{
  SH2neg(sh,10,5);
}
#define init_6_10_6_11
void FASTCALL call_6_10_6_11(SH2_struct* sh)
{
  SH2neg(sh,10,6);
}
#define init_6_10_7_11
void FASTCALL call_6_10_7_11(SH2_struct* sh)
{
  SH2neg(sh,10,7);
}
#define init_6_10_8_11
void FASTCALL call_6_10_8_11(SH2_struct* sh)
{
  SH2neg(sh,10,8);
}
#define init_6_10_9_11
void FASTCALL call_6_10_9_11(SH2_struct* sh)
{
  SH2neg(sh,10,9);
}
#define init_6_10_10_11
void FASTCALL call_6_10_10_11(SH2_struct* sh)
{
  SH2neg(sh,10,10);
}
#define init_6_10_11_11
void FASTCALL call_6_10_11_11(SH2_struct* sh)
{
  SH2neg(sh,10,11);
}
#define init_6_10_12_11
void FASTCALL call_6_10_12_11(SH2_struct* sh)
{
  SH2neg(sh,10,12);
}
#define init_6_10_13_11
void FASTCALL call_6_10_13_11(SH2_struct* sh)
{
  SH2neg(sh,10,13);
}
#define init_6_10_14_11
void FASTCALL call_6_10_14_11(SH2_struct* sh)
{
  SH2neg(sh,10,14);
}
#define init_6_10_15_11
void FASTCALL call_6_10_15_11(SH2_struct* sh)
{
  SH2neg(sh,10,15);
}
#define init_6_11_0_11
void FASTCALL call_6_11_0_11(SH2_struct* sh)
{
  SH2neg(sh,11,0);
}
#define init_6_11_1_11
void FASTCALL call_6_11_1_11(SH2_struct* sh)
{
  SH2neg(sh,11,1);
}
#define init_6_11_2_11
void FASTCALL call_6_11_2_11(SH2_struct* sh)
{
  SH2neg(sh,11,2);
}
#define init_6_11_3_11
void FASTCALL call_6_11_3_11(SH2_struct* sh)
{
  SH2neg(sh,11,3);
}
#define init_6_11_4_11
void FASTCALL call_6_11_4_11(SH2_struct* sh)
{
  SH2neg(sh,11,4);
}
#define init_6_11_5_11
void FASTCALL call_6_11_5_11(SH2_struct* sh)
{
  SH2neg(sh,11,5);
}
#define init_6_11_6_11
void FASTCALL call_6_11_6_11(SH2_struct* sh)
{
  SH2neg(sh,11,6);
}
#define init_6_11_7_11
void FASTCALL call_6_11_7_11(SH2_struct* sh)
{
  SH2neg(sh,11,7);
}
#define init_6_11_8_11
void FASTCALL call_6_11_8_11(SH2_struct* sh)
{
  SH2neg(sh,11,8);
}
#define init_6_11_9_11
void FASTCALL call_6_11_9_11(SH2_struct* sh)
{
  SH2neg(sh,11,9);
}
#define init_6_11_10_11
void FASTCALL call_6_11_10_11(SH2_struct* sh)
{
  SH2neg(sh,11,10);
}
#define init_6_11_11_11
void FASTCALL call_6_11_11_11(SH2_struct* sh)
{
  SH2neg(sh,11,11);
}
#define init_6_11_12_11
void FASTCALL call_6_11_12_11(SH2_struct* sh)
{
  SH2neg(sh,11,12);
}
#define init_6_11_13_11
void FASTCALL call_6_11_13_11(SH2_struct* sh)
{
  SH2neg(sh,11,13);
}
#define init_6_11_14_11
void FASTCALL call_6_11_14_11(SH2_struct* sh)
{
  SH2neg(sh,11,14);
}
#define init_6_11_15_11
void FASTCALL call_6_11_15_11(SH2_struct* sh)
{
  SH2neg(sh,11,15);
}
#define init_6_12_0_11
void FASTCALL call_6_12_0_11(SH2_struct* sh)
{
  SH2neg(sh,12,0);
}
#define init_6_12_1_11
void FASTCALL call_6_12_1_11(SH2_struct* sh)
{
  SH2neg(sh,12,1);
}
#define init_6_12_2_11
void FASTCALL call_6_12_2_11(SH2_struct* sh)
{
  SH2neg(sh,12,2);
}
#define init_6_12_3_11
void FASTCALL call_6_12_3_11(SH2_struct* sh)
{
  SH2neg(sh,12,3);
}
#define init_6_12_4_11
void FASTCALL call_6_12_4_11(SH2_struct* sh)
{
  SH2neg(sh,12,4);
}
#define init_6_12_5_11
void FASTCALL call_6_12_5_11(SH2_struct* sh)
{
  SH2neg(sh,12,5);
}
#define init_6_12_6_11
void FASTCALL call_6_12_6_11(SH2_struct* sh)
{
  SH2neg(sh,12,6);
}
#define init_6_12_7_11
void FASTCALL call_6_12_7_11(SH2_struct* sh)
{
  SH2neg(sh,12,7);
}
#define init_6_12_8_11
void FASTCALL call_6_12_8_11(SH2_struct* sh)
{
  SH2neg(sh,12,8);
}
#define init_6_12_9_11
void FASTCALL call_6_12_9_11(SH2_struct* sh)
{
  SH2neg(sh,12,9);
}
#define init_6_12_10_11
void FASTCALL call_6_12_10_11(SH2_struct* sh)
{
  SH2neg(sh,12,10);
}
#define init_6_12_11_11
void FASTCALL call_6_12_11_11(SH2_struct* sh)
{
  SH2neg(sh,12,11);
}
#define init_6_12_12_11
void FASTCALL call_6_12_12_11(SH2_struct* sh)
{
  SH2neg(sh,12,12);
}
#define init_6_12_13_11
void FASTCALL call_6_12_13_11(SH2_struct* sh)
{
  SH2neg(sh,12,13);
}
#define init_6_12_14_11
void FASTCALL call_6_12_14_11(SH2_struct* sh)
{
  SH2neg(sh,12,14);
}
#define init_6_12_15_11
void FASTCALL call_6_12_15_11(SH2_struct* sh)
{
  SH2neg(sh,12,15);
}
#define init_6_13_0_11
void FASTCALL call_6_13_0_11(SH2_struct* sh)
{
  SH2neg(sh,13,0);
}
#define init_6_13_1_11
void FASTCALL call_6_13_1_11(SH2_struct* sh)
{
  SH2neg(sh,13,1);
}
#define init_6_13_2_11
void FASTCALL call_6_13_2_11(SH2_struct* sh)
{
  SH2neg(sh,13,2);
}
#define init_6_13_3_11
void FASTCALL call_6_13_3_11(SH2_struct* sh)
{
  SH2neg(sh,13,3);
}
#define init_6_13_4_11
void FASTCALL call_6_13_4_11(SH2_struct* sh)
{
  SH2neg(sh,13,4);
}
#define init_6_13_5_11
void FASTCALL call_6_13_5_11(SH2_struct* sh)
{
  SH2neg(sh,13,5);
}
#define init_6_13_6_11
void FASTCALL call_6_13_6_11(SH2_struct* sh)
{
  SH2neg(sh,13,6);
}
#define init_6_13_7_11
void FASTCALL call_6_13_7_11(SH2_struct* sh)
{
  SH2neg(sh,13,7);
}
#define init_6_13_8_11
void FASTCALL call_6_13_8_11(SH2_struct* sh)
{
  SH2neg(sh,13,8);
}
#define init_6_13_9_11
void FASTCALL call_6_13_9_11(SH2_struct* sh)
{
  SH2neg(sh,13,9);
}
#define init_6_13_10_11
void FASTCALL call_6_13_10_11(SH2_struct* sh)
{
  SH2neg(sh,13,10);
}
#define init_6_13_11_11
void FASTCALL call_6_13_11_11(SH2_struct* sh)
{
  SH2neg(sh,13,11);
}
#define init_6_13_12_11
void FASTCALL call_6_13_12_11(SH2_struct* sh)
{
  SH2neg(sh,13,12);
}
#define init_6_13_13_11
void FASTCALL call_6_13_13_11(SH2_struct* sh)
{
  SH2neg(sh,13,13);
}
#define init_6_13_14_11
void FASTCALL call_6_13_14_11(SH2_struct* sh)
{
  SH2neg(sh,13,14);
}
#define init_6_13_15_11
void FASTCALL call_6_13_15_11(SH2_struct* sh)
{
  SH2neg(sh,13,15);
}
#define init_6_14_0_11
void FASTCALL call_6_14_0_11(SH2_struct* sh)
{
  SH2neg(sh,14,0);
}
#define init_6_14_1_11
void FASTCALL call_6_14_1_11(SH2_struct* sh)
{
  SH2neg(sh,14,1);
}
#define init_6_14_2_11
void FASTCALL call_6_14_2_11(SH2_struct* sh)
{
  SH2neg(sh,14,2);
}
#define init_6_14_3_11
void FASTCALL call_6_14_3_11(SH2_struct* sh)
{
  SH2neg(sh,14,3);
}
#define init_6_14_4_11
void FASTCALL call_6_14_4_11(SH2_struct* sh)
{
  SH2neg(sh,14,4);
}
#define init_6_14_5_11
void FASTCALL call_6_14_5_11(SH2_struct* sh)
{
  SH2neg(sh,14,5);
}
#define init_6_14_6_11
void FASTCALL call_6_14_6_11(SH2_struct* sh)
{
  SH2neg(sh,14,6);
}
#define init_6_14_7_11
void FASTCALL call_6_14_7_11(SH2_struct* sh)
{
  SH2neg(sh,14,7);
}
#define init_6_14_8_11
void FASTCALL call_6_14_8_11(SH2_struct* sh)
{
  SH2neg(sh,14,8);
}
#define init_6_14_9_11
void FASTCALL call_6_14_9_11(SH2_struct* sh)
{
  SH2neg(sh,14,9);
}
#define init_6_14_10_11
void FASTCALL call_6_14_10_11(SH2_struct* sh)
{
  SH2neg(sh,14,10);
}
#define init_6_14_11_11
void FASTCALL call_6_14_11_11(SH2_struct* sh)
{
  SH2neg(sh,14,11);
}
#define init_6_14_12_11
void FASTCALL call_6_14_12_11(SH2_struct* sh)
{
  SH2neg(sh,14,12);
}
#define init_6_14_13_11
void FASTCALL call_6_14_13_11(SH2_struct* sh)
{
  SH2neg(sh,14,13);
}
#define init_6_14_14_11
void FASTCALL call_6_14_14_11(SH2_struct* sh)
{
  SH2neg(sh,14,14);
}
#define init_6_14_15_11
void FASTCALL call_6_14_15_11(SH2_struct* sh)
{
  SH2neg(sh,14,15);
}
#define init_6_15_0_11
void FASTCALL call_6_15_0_11(SH2_struct* sh)
{
  SH2neg(sh,15,0);
}
#define init_6_15_1_11
void FASTCALL call_6_15_1_11(SH2_struct* sh)
{
  SH2neg(sh,15,1);
}
#define init_6_15_2_11
void FASTCALL call_6_15_2_11(SH2_struct* sh)
{
  SH2neg(sh,15,2);
}
#define init_6_15_3_11
void FASTCALL call_6_15_3_11(SH2_struct* sh)
{
  SH2neg(sh,15,3);
}
#define init_6_15_4_11
void FASTCALL call_6_15_4_11(SH2_struct* sh)
{
  SH2neg(sh,15,4);
}
#define init_6_15_5_11
void FASTCALL call_6_15_5_11(SH2_struct* sh)
{
  SH2neg(sh,15,5);
}
#define init_6_15_6_11
void FASTCALL call_6_15_6_11(SH2_struct* sh)
{
  SH2neg(sh,15,6);
}
#define init_6_15_7_11
void FASTCALL call_6_15_7_11(SH2_struct* sh)
{
  SH2neg(sh,15,7);
}
#define init_6_15_8_11
void FASTCALL call_6_15_8_11(SH2_struct* sh)
{
  SH2neg(sh,15,8);
}
#define init_6_15_9_11
void FASTCALL call_6_15_9_11(SH2_struct* sh)
{
  SH2neg(sh,15,9);
}
#define init_6_15_10_11
void FASTCALL call_6_15_10_11(SH2_struct* sh)
{
  SH2neg(sh,15,10);
}
#define init_6_15_11_11
void FASTCALL call_6_15_11_11(SH2_struct* sh)
{
  SH2neg(sh,15,11);
}
#define init_6_15_12_11
void FASTCALL call_6_15_12_11(SH2_struct* sh)
{
  SH2neg(sh,15,12);
}
#define init_6_15_13_11
void FASTCALL call_6_15_13_11(SH2_struct* sh)
{
  SH2neg(sh,15,13);
}
#define init_6_15_14_11
void FASTCALL call_6_15_14_11(SH2_struct* sh)
{
  SH2neg(sh,15,14);
}
#define init_6_15_15_11
void FASTCALL call_6_15_15_11(SH2_struct* sh)
{
  SH2neg(sh,15,15);
}
#define init_2_0_0_14
void FASTCALL call_2_0_0_14(SH2_struct* sh)
{
  SH2mulu(sh,0,0);
}
#define init_2_0_1_14
void FASTCALL call_2_0_1_14(SH2_struct* sh)
{
  SH2mulu(sh,0,1);
}
#define init_2_0_2_14
void FASTCALL call_2_0_2_14(SH2_struct* sh)
{
  SH2mulu(sh,0,2);
}
#define init_2_0_3_14
void FASTCALL call_2_0_3_14(SH2_struct* sh)
{
  SH2mulu(sh,0,3);
}
#define init_2_0_4_14
void FASTCALL call_2_0_4_14(SH2_struct* sh)
{
  SH2mulu(sh,0,4);
}
#define init_2_0_5_14
void FASTCALL call_2_0_5_14(SH2_struct* sh)
{
  SH2mulu(sh,0,5);
}
#define init_2_0_6_14
void FASTCALL call_2_0_6_14(SH2_struct* sh)
{
  SH2mulu(sh,0,6);
}
#define init_2_0_7_14
void FASTCALL call_2_0_7_14(SH2_struct* sh)
{
  SH2mulu(sh,0,7);
}
#define init_2_0_8_14
void FASTCALL call_2_0_8_14(SH2_struct* sh)
{
  SH2mulu(sh,0,8);
}
#define init_2_0_9_14
void FASTCALL call_2_0_9_14(SH2_struct* sh)
{
  SH2mulu(sh,0,9);
}
#define init_2_0_10_14
void FASTCALL call_2_0_10_14(SH2_struct* sh)
{
  SH2mulu(sh,0,10);
}
#define init_2_0_11_14
void FASTCALL call_2_0_11_14(SH2_struct* sh)
{
  SH2mulu(sh,0,11);
}
#define init_2_0_12_14
void FASTCALL call_2_0_12_14(SH2_struct* sh)
{
  SH2mulu(sh,0,12);
}
#define init_2_0_13_14
void FASTCALL call_2_0_13_14(SH2_struct* sh)
{
  SH2mulu(sh,0,13);
}
#define init_2_0_14_14
void FASTCALL call_2_0_14_14(SH2_struct* sh)
{
  SH2mulu(sh,0,14);
}
#define init_2_0_15_14
void FASTCALL call_2_0_15_14(SH2_struct* sh)
{
  SH2mulu(sh,0,15);
}
#define init_2_1_0_14
void FASTCALL call_2_1_0_14(SH2_struct* sh)
{
  SH2mulu(sh,1,0);
}
#define init_2_1_1_14
void FASTCALL call_2_1_1_14(SH2_struct* sh)
{
  SH2mulu(sh,1,1);
}
#define init_2_1_2_14
void FASTCALL call_2_1_2_14(SH2_struct* sh)
{
  SH2mulu(sh,1,2);
}
#define init_2_1_3_14
void FASTCALL call_2_1_3_14(SH2_struct* sh)
{
  SH2mulu(sh,1,3);
}
#define init_2_1_4_14
void FASTCALL call_2_1_4_14(SH2_struct* sh)
{
  SH2mulu(sh,1,4);
}
#define init_2_1_5_14
void FASTCALL call_2_1_5_14(SH2_struct* sh)
{
  SH2mulu(sh,1,5);
}
#define init_2_1_6_14
void FASTCALL call_2_1_6_14(SH2_struct* sh)
{
  SH2mulu(sh,1,6);
}
#define init_2_1_7_14
void FASTCALL call_2_1_7_14(SH2_struct* sh)
{
  SH2mulu(sh,1,7);
}
#define init_2_1_8_14
void FASTCALL call_2_1_8_14(SH2_struct* sh)
{
  SH2mulu(sh,1,8);
}
#define init_2_1_9_14
void FASTCALL call_2_1_9_14(SH2_struct* sh)
{
  SH2mulu(sh,1,9);
}
#define init_2_1_10_14
void FASTCALL call_2_1_10_14(SH2_struct* sh)
{
  SH2mulu(sh,1,10);
}
#define init_2_1_11_14
void FASTCALL call_2_1_11_14(SH2_struct* sh)
{
  SH2mulu(sh,1,11);
}
#define init_2_1_12_14
void FASTCALL call_2_1_12_14(SH2_struct* sh)
{
  SH2mulu(sh,1,12);
}
#define init_2_1_13_14
void FASTCALL call_2_1_13_14(SH2_struct* sh)
{
  SH2mulu(sh,1,13);
}
#define init_2_1_14_14
void FASTCALL call_2_1_14_14(SH2_struct* sh)
{
  SH2mulu(sh,1,14);
}
#define init_2_1_15_14
void FASTCALL call_2_1_15_14(SH2_struct* sh)
{
  SH2mulu(sh,1,15);
}
#define init_2_2_0_14
void FASTCALL call_2_2_0_14(SH2_struct* sh)
{
  SH2mulu(sh,2,0);
}
#define init_2_2_1_14
void FASTCALL call_2_2_1_14(SH2_struct* sh)
{
  SH2mulu(sh,2,1);
}
#define init_2_2_2_14
void FASTCALL call_2_2_2_14(SH2_struct* sh)
{
  SH2mulu(sh,2,2);
}
#define init_2_2_3_14
void FASTCALL call_2_2_3_14(SH2_struct* sh)
{
  SH2mulu(sh,2,3);
}
#define init_2_2_4_14
void FASTCALL call_2_2_4_14(SH2_struct* sh)
{
  SH2mulu(sh,2,4);
}
#define init_2_2_5_14
void FASTCALL call_2_2_5_14(SH2_struct* sh)
{
  SH2mulu(sh,2,5);
}
#define init_2_2_6_14
void FASTCALL call_2_2_6_14(SH2_struct* sh)
{
  SH2mulu(sh,2,6);
}
#define init_2_2_7_14
void FASTCALL call_2_2_7_14(SH2_struct* sh)
{
  SH2mulu(sh,2,7);
}
#define init_2_2_8_14
void FASTCALL call_2_2_8_14(SH2_struct* sh)
{
  SH2mulu(sh,2,8);
}
#define init_2_2_9_14
void FASTCALL call_2_2_9_14(SH2_struct* sh)
{
  SH2mulu(sh,2,9);
}
#define init_2_2_10_14
void FASTCALL call_2_2_10_14(SH2_struct* sh)
{
  SH2mulu(sh,2,10);
}
#define init_2_2_11_14
void FASTCALL call_2_2_11_14(SH2_struct* sh)
{
  SH2mulu(sh,2,11);
}
#define init_2_2_12_14
void FASTCALL call_2_2_12_14(SH2_struct* sh)
{
  SH2mulu(sh,2,12);
}
#define init_2_2_13_14
void FASTCALL call_2_2_13_14(SH2_struct* sh)
{
  SH2mulu(sh,2,13);
}
#define init_2_2_14_14
void FASTCALL call_2_2_14_14(SH2_struct* sh)
{
  SH2mulu(sh,2,14);
}
#define init_2_2_15_14
void FASTCALL call_2_2_15_14(SH2_struct* sh)
{
  SH2mulu(sh,2,15);
}
#define init_2_3_0_14
void FASTCALL call_2_3_0_14(SH2_struct* sh)
{
  SH2mulu(sh,3,0);
}
#define init_2_3_1_14
void FASTCALL call_2_3_1_14(SH2_struct* sh)
{
  SH2mulu(sh,3,1);
}
#define init_2_3_2_14
void FASTCALL call_2_3_2_14(SH2_struct* sh)
{
  SH2mulu(sh,3,2);
}
#define init_2_3_3_14
void FASTCALL call_2_3_3_14(SH2_struct* sh)
{
  SH2mulu(sh,3,3);
}
#define init_2_3_4_14
void FASTCALL call_2_3_4_14(SH2_struct* sh)
{
  SH2mulu(sh,3,4);
}
#define init_2_3_5_14
void FASTCALL call_2_3_5_14(SH2_struct* sh)
{
  SH2mulu(sh,3,5);
}
#define init_2_3_6_14
void FASTCALL call_2_3_6_14(SH2_struct* sh)
{
  SH2mulu(sh,3,6);
}
#define init_2_3_7_14
void FASTCALL call_2_3_7_14(SH2_struct* sh)
{
  SH2mulu(sh,3,7);
}
#define init_2_3_8_14
void FASTCALL call_2_3_8_14(SH2_struct* sh)
{
  SH2mulu(sh,3,8);
}
#define init_2_3_9_14
void FASTCALL call_2_3_9_14(SH2_struct* sh)
{
  SH2mulu(sh,3,9);
}
#define init_2_3_10_14
void FASTCALL call_2_3_10_14(SH2_struct* sh)
{
  SH2mulu(sh,3,10);
}
#define init_2_3_11_14
void FASTCALL call_2_3_11_14(SH2_struct* sh)
{
  SH2mulu(sh,3,11);
}
#define init_2_3_12_14
void FASTCALL call_2_3_12_14(SH2_struct* sh)
{
  SH2mulu(sh,3,12);
}
#define init_2_3_13_14
void FASTCALL call_2_3_13_14(SH2_struct* sh)
{
  SH2mulu(sh,3,13);
}
#define init_2_3_14_14
void FASTCALL call_2_3_14_14(SH2_struct* sh)
{
  SH2mulu(sh,3,14);
}
#define init_2_3_15_14
void FASTCALL call_2_3_15_14(SH2_struct* sh)
{
  SH2mulu(sh,3,15);
}
#define init_2_4_0_14
void FASTCALL call_2_4_0_14(SH2_struct* sh)
{
  SH2mulu(sh,4,0);
}
#define init_2_4_1_14
void FASTCALL call_2_4_1_14(SH2_struct* sh)
{
  SH2mulu(sh,4,1);
}
#define init_2_4_2_14
void FASTCALL call_2_4_2_14(SH2_struct* sh)
{
  SH2mulu(sh,4,2);
}
#define init_2_4_3_14
void FASTCALL call_2_4_3_14(SH2_struct* sh)
{
  SH2mulu(sh,4,3);
}
#define init_2_4_4_14
void FASTCALL call_2_4_4_14(SH2_struct* sh)
{
  SH2mulu(sh,4,4);
}
#define init_2_4_5_14
void FASTCALL call_2_4_5_14(SH2_struct* sh)
{
  SH2mulu(sh,4,5);
}
#define init_2_4_6_14
void FASTCALL call_2_4_6_14(SH2_struct* sh)
{
  SH2mulu(sh,4,6);
}
#define init_2_4_7_14
void FASTCALL call_2_4_7_14(SH2_struct* sh)
{
  SH2mulu(sh,4,7);
}
#define init_2_4_8_14
void FASTCALL call_2_4_8_14(SH2_struct* sh)
{
  SH2mulu(sh,4,8);
}
#define init_2_4_9_14
void FASTCALL call_2_4_9_14(SH2_struct* sh)
{
  SH2mulu(sh,4,9);
}
#define init_2_4_10_14
void FASTCALL call_2_4_10_14(SH2_struct* sh)
{
  SH2mulu(sh,4,10);
}
#define init_2_4_11_14
void FASTCALL call_2_4_11_14(SH2_struct* sh)
{
  SH2mulu(sh,4,11);
}
#define init_2_4_12_14
void FASTCALL call_2_4_12_14(SH2_struct* sh)
{
  SH2mulu(sh,4,12);
}
#define init_2_4_13_14
void FASTCALL call_2_4_13_14(SH2_struct* sh)
{
  SH2mulu(sh,4,13);
}
#define init_2_4_14_14
void FASTCALL call_2_4_14_14(SH2_struct* sh)
{
  SH2mulu(sh,4,14);
}
#define init_2_4_15_14
void FASTCALL call_2_4_15_14(SH2_struct* sh)
{
  SH2mulu(sh,4,15);
}
#define init_2_5_0_14
void FASTCALL call_2_5_0_14(SH2_struct* sh)
{
  SH2mulu(sh,5,0);
}
#define init_2_5_1_14
void FASTCALL call_2_5_1_14(SH2_struct* sh)
{
  SH2mulu(sh,5,1);
}
#define init_2_5_2_14
void FASTCALL call_2_5_2_14(SH2_struct* sh)
{
  SH2mulu(sh,5,2);
}
#define init_2_5_3_14
void FASTCALL call_2_5_3_14(SH2_struct* sh)
{
  SH2mulu(sh,5,3);
}
#define init_2_5_4_14
void FASTCALL call_2_5_4_14(SH2_struct* sh)
{
  SH2mulu(sh,5,4);
}
#define init_2_5_5_14
void FASTCALL call_2_5_5_14(SH2_struct* sh)
{
  SH2mulu(sh,5,5);
}
#define init_2_5_6_14
void FASTCALL call_2_5_6_14(SH2_struct* sh)
{
  SH2mulu(sh,5,6);
}
#define init_2_5_7_14
void FASTCALL call_2_5_7_14(SH2_struct* sh)
{
  SH2mulu(sh,5,7);
}
#define init_2_5_8_14
void FASTCALL call_2_5_8_14(SH2_struct* sh)
{
  SH2mulu(sh,5,8);
}
#define init_2_5_9_14
void FASTCALL call_2_5_9_14(SH2_struct* sh)
{
  SH2mulu(sh,5,9);
}
#define init_2_5_10_14
void FASTCALL call_2_5_10_14(SH2_struct* sh)
{
  SH2mulu(sh,5,10);
}
#define init_2_5_11_14
void FASTCALL call_2_5_11_14(SH2_struct* sh)
{
  SH2mulu(sh,5,11);
}
#define init_2_5_12_14
void FASTCALL call_2_5_12_14(SH2_struct* sh)
{
  SH2mulu(sh,5,12);
}
#define init_2_5_13_14
void FASTCALL call_2_5_13_14(SH2_struct* sh)
{
  SH2mulu(sh,5,13);
}
#define init_2_5_14_14
void FASTCALL call_2_5_14_14(SH2_struct* sh)
{
  SH2mulu(sh,5,14);
}
#define init_2_5_15_14
void FASTCALL call_2_5_15_14(SH2_struct* sh)
{
  SH2mulu(sh,5,15);
}
#define init_2_6_0_14
void FASTCALL call_2_6_0_14(SH2_struct* sh)
{
  SH2mulu(sh,6,0);
}
#define init_2_6_1_14
void FASTCALL call_2_6_1_14(SH2_struct* sh)
{
  SH2mulu(sh,6,1);
}
#define init_2_6_2_14
void FASTCALL call_2_6_2_14(SH2_struct* sh)
{
  SH2mulu(sh,6,2);
}
#define init_2_6_3_14
void FASTCALL call_2_6_3_14(SH2_struct* sh)
{
  SH2mulu(sh,6,3);
}
#define init_2_6_4_14
void FASTCALL call_2_6_4_14(SH2_struct* sh)
{
  SH2mulu(sh,6,4);
}
#define init_2_6_5_14
void FASTCALL call_2_6_5_14(SH2_struct* sh)
{
  SH2mulu(sh,6,5);
}
#define init_2_6_6_14
void FASTCALL call_2_6_6_14(SH2_struct* sh)
{
  SH2mulu(sh,6,6);
}
#define init_2_6_7_14
void FASTCALL call_2_6_7_14(SH2_struct* sh)
{
  SH2mulu(sh,6,7);
}
#define init_2_6_8_14
void FASTCALL call_2_6_8_14(SH2_struct* sh)
{
  SH2mulu(sh,6,8);
}
#define init_2_6_9_14
void FASTCALL call_2_6_9_14(SH2_struct* sh)
{
  SH2mulu(sh,6,9);
}
#define init_2_6_10_14
void FASTCALL call_2_6_10_14(SH2_struct* sh)
{
  SH2mulu(sh,6,10);
}
#define init_2_6_11_14
void FASTCALL call_2_6_11_14(SH2_struct* sh)
{
  SH2mulu(sh,6,11);
}
#define init_2_6_12_14
void FASTCALL call_2_6_12_14(SH2_struct* sh)
{
  SH2mulu(sh,6,12);
}
#define init_2_6_13_14
void FASTCALL call_2_6_13_14(SH2_struct* sh)
{
  SH2mulu(sh,6,13);
}
#define init_2_6_14_14
void FASTCALL call_2_6_14_14(SH2_struct* sh)
{
  SH2mulu(sh,6,14);
}
#define init_2_6_15_14
void FASTCALL call_2_6_15_14(SH2_struct* sh)
{
  SH2mulu(sh,6,15);
}
#define init_2_7_0_14
void FASTCALL call_2_7_0_14(SH2_struct* sh)
{
  SH2mulu(sh,7,0);
}
#define init_2_7_1_14
void FASTCALL call_2_7_1_14(SH2_struct* sh)
{
  SH2mulu(sh,7,1);
}
#define init_2_7_2_14
void FASTCALL call_2_7_2_14(SH2_struct* sh)
{
  SH2mulu(sh,7,2);
}
#define init_2_7_3_14
void FASTCALL call_2_7_3_14(SH2_struct* sh)
{
  SH2mulu(sh,7,3);
}
#define init_2_7_4_14
void FASTCALL call_2_7_4_14(SH2_struct* sh)
{
  SH2mulu(sh,7,4);
}
#define init_2_7_5_14
void FASTCALL call_2_7_5_14(SH2_struct* sh)
{
  SH2mulu(sh,7,5);
}
#define init_2_7_6_14
void FASTCALL call_2_7_6_14(SH2_struct* sh)
{
  SH2mulu(sh,7,6);
}
#define init_2_7_7_14
void FASTCALL call_2_7_7_14(SH2_struct* sh)
{
  SH2mulu(sh,7,7);
}
#define init_2_7_8_14
void FASTCALL call_2_7_8_14(SH2_struct* sh)
{
  SH2mulu(sh,7,8);
}
#define init_2_7_9_14
void FASTCALL call_2_7_9_14(SH2_struct* sh)
{
  SH2mulu(sh,7,9);
}
#define init_2_7_10_14
void FASTCALL call_2_7_10_14(SH2_struct* sh)
{
  SH2mulu(sh,7,10);
}
#define init_2_7_11_14
void FASTCALL call_2_7_11_14(SH2_struct* sh)
{
  SH2mulu(sh,7,11);
}
#define init_2_7_12_14
void FASTCALL call_2_7_12_14(SH2_struct* sh)
{
  SH2mulu(sh,7,12);
}
#define init_2_7_13_14
void FASTCALL call_2_7_13_14(SH2_struct* sh)
{
  SH2mulu(sh,7,13);
}
#define init_2_7_14_14
void FASTCALL call_2_7_14_14(SH2_struct* sh)
{
  SH2mulu(sh,7,14);
}
#define init_2_7_15_14
void FASTCALL call_2_7_15_14(SH2_struct* sh)
{
  SH2mulu(sh,7,15);
}
#define init_2_8_0_14
void FASTCALL call_2_8_0_14(SH2_struct* sh)
{
  SH2mulu(sh,8,0);
}
#define init_2_8_1_14
void FASTCALL call_2_8_1_14(SH2_struct* sh)
{
  SH2mulu(sh,8,1);
}
#define init_2_8_2_14
void FASTCALL call_2_8_2_14(SH2_struct* sh)
{
  SH2mulu(sh,8,2);
}
#define init_2_8_3_14
void FASTCALL call_2_8_3_14(SH2_struct* sh)
{
  SH2mulu(sh,8,3);
}
#define init_2_8_4_14
void FASTCALL call_2_8_4_14(SH2_struct* sh)
{
  SH2mulu(sh,8,4);
}
#define init_2_8_5_14
void FASTCALL call_2_8_5_14(SH2_struct* sh)
{
  SH2mulu(sh,8,5);
}
#define init_2_8_6_14
void FASTCALL call_2_8_6_14(SH2_struct* sh)
{
  SH2mulu(sh,8,6);
}
#define init_2_8_7_14
void FASTCALL call_2_8_7_14(SH2_struct* sh)
{
  SH2mulu(sh,8,7);
}
#define init_2_8_8_14
void FASTCALL call_2_8_8_14(SH2_struct* sh)
{
  SH2mulu(sh,8,8);
}
#define init_2_8_9_14
void FASTCALL call_2_8_9_14(SH2_struct* sh)
{
  SH2mulu(sh,8,9);
}
#define init_2_8_10_14
void FASTCALL call_2_8_10_14(SH2_struct* sh)
{
  SH2mulu(sh,8,10);
}
#define init_2_8_11_14
void FASTCALL call_2_8_11_14(SH2_struct* sh)
{
  SH2mulu(sh,8,11);
}
#define init_2_8_12_14
void FASTCALL call_2_8_12_14(SH2_struct* sh)
{
  SH2mulu(sh,8,12);
}
#define init_2_8_13_14
void FASTCALL call_2_8_13_14(SH2_struct* sh)
{
  SH2mulu(sh,8,13);
}
#define init_2_8_14_14
void FASTCALL call_2_8_14_14(SH2_struct* sh)
{
  SH2mulu(sh,8,14);
}
#define init_2_8_15_14
void FASTCALL call_2_8_15_14(SH2_struct* sh)
{
  SH2mulu(sh,8,15);
}
#define init_2_9_0_14
void FASTCALL call_2_9_0_14(SH2_struct* sh)
{
  SH2mulu(sh,9,0);
}
#define init_2_9_1_14
void FASTCALL call_2_9_1_14(SH2_struct* sh)
{
  SH2mulu(sh,9,1);
}
#define init_2_9_2_14
void FASTCALL call_2_9_2_14(SH2_struct* sh)
{
  SH2mulu(sh,9,2);
}
#define init_2_9_3_14
void FASTCALL call_2_9_3_14(SH2_struct* sh)
{
  SH2mulu(sh,9,3);
}
#define init_2_9_4_14
void FASTCALL call_2_9_4_14(SH2_struct* sh)
{
  SH2mulu(sh,9,4);
}
#define init_2_9_5_14
void FASTCALL call_2_9_5_14(SH2_struct* sh)
{
  SH2mulu(sh,9,5);
}
#define init_2_9_6_14
void FASTCALL call_2_9_6_14(SH2_struct* sh)
{
  SH2mulu(sh,9,6);
}
#define init_2_9_7_14
void FASTCALL call_2_9_7_14(SH2_struct* sh)
{
  SH2mulu(sh,9,7);
}
#define init_2_9_8_14
void FASTCALL call_2_9_8_14(SH2_struct* sh)
{
  SH2mulu(sh,9,8);
}
#define init_2_9_9_14
void FASTCALL call_2_9_9_14(SH2_struct* sh)
{
  SH2mulu(sh,9,9);
}
#define init_2_9_10_14
void FASTCALL call_2_9_10_14(SH2_struct* sh)
{
  SH2mulu(sh,9,10);
}
#define init_2_9_11_14
void FASTCALL call_2_9_11_14(SH2_struct* sh)
{
  SH2mulu(sh,9,11);
}
#define init_2_9_12_14
void FASTCALL call_2_9_12_14(SH2_struct* sh)
{
  SH2mulu(sh,9,12);
}
#define init_2_9_13_14
void FASTCALL call_2_9_13_14(SH2_struct* sh)
{
  SH2mulu(sh,9,13);
}
#define init_2_9_14_14
void FASTCALL call_2_9_14_14(SH2_struct* sh)
{
  SH2mulu(sh,9,14);
}
#define init_2_9_15_14
void FASTCALL call_2_9_15_14(SH2_struct* sh)
{
  SH2mulu(sh,9,15);
}
#define init_2_10_0_14
void FASTCALL call_2_10_0_14(SH2_struct* sh)
{
  SH2mulu(sh,10,0);
}
#define init_2_10_1_14
void FASTCALL call_2_10_1_14(SH2_struct* sh)
{
  SH2mulu(sh,10,1);
}
#define init_2_10_2_14
void FASTCALL call_2_10_2_14(SH2_struct* sh)
{
  SH2mulu(sh,10,2);
}
#define init_2_10_3_14
void FASTCALL call_2_10_3_14(SH2_struct* sh)
{
  SH2mulu(sh,10,3);
}
#define init_2_10_4_14
void FASTCALL call_2_10_4_14(SH2_struct* sh)
{
  SH2mulu(sh,10,4);
}
#define init_2_10_5_14
void FASTCALL call_2_10_5_14(SH2_struct* sh)
{
  SH2mulu(sh,10,5);
}
#define init_2_10_6_14
void FASTCALL call_2_10_6_14(SH2_struct* sh)
{
  SH2mulu(sh,10,6);
}
#define init_2_10_7_14
void FASTCALL call_2_10_7_14(SH2_struct* sh)
{
  SH2mulu(sh,10,7);
}
#define init_2_10_8_14
void FASTCALL call_2_10_8_14(SH2_struct* sh)
{
  SH2mulu(sh,10,8);
}
#define init_2_10_9_14
void FASTCALL call_2_10_9_14(SH2_struct* sh)
{
  SH2mulu(sh,10,9);
}
#define init_2_10_10_14
void FASTCALL call_2_10_10_14(SH2_struct* sh)
{
  SH2mulu(sh,10,10);
}
#define init_2_10_11_14
void FASTCALL call_2_10_11_14(SH2_struct* sh)
{
  SH2mulu(sh,10,11);
}
#define init_2_10_12_14
void FASTCALL call_2_10_12_14(SH2_struct* sh)
{
  SH2mulu(sh,10,12);
}
#define init_2_10_13_14
void FASTCALL call_2_10_13_14(SH2_struct* sh)
{
  SH2mulu(sh,10,13);
}
#define init_2_10_14_14
void FASTCALL call_2_10_14_14(SH2_struct* sh)
{
  SH2mulu(sh,10,14);
}
#define init_2_10_15_14
void FASTCALL call_2_10_15_14(SH2_struct* sh)
{
  SH2mulu(sh,10,15);
}
#define init_2_11_0_14
void FASTCALL call_2_11_0_14(SH2_struct* sh)
{
  SH2mulu(sh,11,0);
}
#define init_2_11_1_14
void FASTCALL call_2_11_1_14(SH2_struct* sh)
{
  SH2mulu(sh,11,1);
}
#define init_2_11_2_14
void FASTCALL call_2_11_2_14(SH2_struct* sh)
{
  SH2mulu(sh,11,2);
}
#define init_2_11_3_14
void FASTCALL call_2_11_3_14(SH2_struct* sh)
{
  SH2mulu(sh,11,3);
}
#define init_2_11_4_14
void FASTCALL call_2_11_4_14(SH2_struct* sh)
{
  SH2mulu(sh,11,4);
}
#define init_2_11_5_14
void FASTCALL call_2_11_5_14(SH2_struct* sh)
{
  SH2mulu(sh,11,5);
}
#define init_2_11_6_14
void FASTCALL call_2_11_6_14(SH2_struct* sh)
{
  SH2mulu(sh,11,6);
}
#define init_2_11_7_14
void FASTCALL call_2_11_7_14(SH2_struct* sh)
{
  SH2mulu(sh,11,7);
}
#define init_2_11_8_14
void FASTCALL call_2_11_8_14(SH2_struct* sh)
{
  SH2mulu(sh,11,8);
}
#define init_2_11_9_14
void FASTCALL call_2_11_9_14(SH2_struct* sh)
{
  SH2mulu(sh,11,9);
}
#define init_2_11_10_14
void FASTCALL call_2_11_10_14(SH2_struct* sh)
{
  SH2mulu(sh,11,10);
}
#define init_2_11_11_14
void FASTCALL call_2_11_11_14(SH2_struct* sh)
{
  SH2mulu(sh,11,11);
}
#define init_2_11_12_14
void FASTCALL call_2_11_12_14(SH2_struct* sh)
{
  SH2mulu(sh,11,12);
}
#define init_2_11_13_14
void FASTCALL call_2_11_13_14(SH2_struct* sh)
{
  SH2mulu(sh,11,13);
}
#define init_2_11_14_14
void FASTCALL call_2_11_14_14(SH2_struct* sh)
{
  SH2mulu(sh,11,14);
}
#define init_2_11_15_14
void FASTCALL call_2_11_15_14(SH2_struct* sh)
{
  SH2mulu(sh,11,15);
}
#define init_2_12_0_14
void FASTCALL call_2_12_0_14(SH2_struct* sh)
{
  SH2mulu(sh,12,0);
}
#define init_2_12_1_14
void FASTCALL call_2_12_1_14(SH2_struct* sh)
{
  SH2mulu(sh,12,1);
}
#define init_2_12_2_14
void FASTCALL call_2_12_2_14(SH2_struct* sh)
{
  SH2mulu(sh,12,2);
}
#define init_2_12_3_14
void FASTCALL call_2_12_3_14(SH2_struct* sh)
{
  SH2mulu(sh,12,3);
}
#define init_2_12_4_14
void FASTCALL call_2_12_4_14(SH2_struct* sh)
{
  SH2mulu(sh,12,4);
}
#define init_2_12_5_14
void FASTCALL call_2_12_5_14(SH2_struct* sh)
{
  SH2mulu(sh,12,5);
}
#define init_2_12_6_14
void FASTCALL call_2_12_6_14(SH2_struct* sh)
{
  SH2mulu(sh,12,6);
}
#define init_2_12_7_14
void FASTCALL call_2_12_7_14(SH2_struct* sh)
{
  SH2mulu(sh,12,7);
}
#define init_2_12_8_14
void FASTCALL call_2_12_8_14(SH2_struct* sh)
{
  SH2mulu(sh,12,8);
}
#define init_2_12_9_14
void FASTCALL call_2_12_9_14(SH2_struct* sh)
{
  SH2mulu(sh,12,9);
}
#define init_2_12_10_14
void FASTCALL call_2_12_10_14(SH2_struct* sh)
{
  SH2mulu(sh,12,10);
}
#define init_2_12_11_14
void FASTCALL call_2_12_11_14(SH2_struct* sh)
{
  SH2mulu(sh,12,11);
}
#define init_2_12_12_14
void FASTCALL call_2_12_12_14(SH2_struct* sh)
{
  SH2mulu(sh,12,12);
}
#define init_2_12_13_14
void FASTCALL call_2_12_13_14(SH2_struct* sh)
{
  SH2mulu(sh,12,13);
}
#define init_2_12_14_14
void FASTCALL call_2_12_14_14(SH2_struct* sh)
{
  SH2mulu(sh,12,14);
}
#define init_2_12_15_14
void FASTCALL call_2_12_15_14(SH2_struct* sh)
{
  SH2mulu(sh,12,15);
}
#define init_2_13_0_14
void FASTCALL call_2_13_0_14(SH2_struct* sh)
{
  SH2mulu(sh,13,0);
}
#define init_2_13_1_14
void FASTCALL call_2_13_1_14(SH2_struct* sh)
{
  SH2mulu(sh,13,1);
}
#define init_2_13_2_14
void FASTCALL call_2_13_2_14(SH2_struct* sh)
{
  SH2mulu(sh,13,2);
}
#define init_2_13_3_14
void FASTCALL call_2_13_3_14(SH2_struct* sh)
{
  SH2mulu(sh,13,3);
}
#define init_2_13_4_14
void FASTCALL call_2_13_4_14(SH2_struct* sh)
{
  SH2mulu(sh,13,4);
}
#define init_2_13_5_14
void FASTCALL call_2_13_5_14(SH2_struct* sh)
{
  SH2mulu(sh,13,5);
}
#define init_2_13_6_14
void FASTCALL call_2_13_6_14(SH2_struct* sh)
{
  SH2mulu(sh,13,6);
}
#define init_2_13_7_14
void FASTCALL call_2_13_7_14(SH2_struct* sh)
{
  SH2mulu(sh,13,7);
}
#define init_2_13_8_14
void FASTCALL call_2_13_8_14(SH2_struct* sh)
{
  SH2mulu(sh,13,8);
}
#define init_2_13_9_14
void FASTCALL call_2_13_9_14(SH2_struct* sh)
{
  SH2mulu(sh,13,9);
}
#define init_2_13_10_14
void FASTCALL call_2_13_10_14(SH2_struct* sh)
{
  SH2mulu(sh,13,10);
}
#define init_2_13_11_14
void FASTCALL call_2_13_11_14(SH2_struct* sh)
{
  SH2mulu(sh,13,11);
}
#define init_2_13_12_14
void FASTCALL call_2_13_12_14(SH2_struct* sh)
{
  SH2mulu(sh,13,12);
}
#define init_2_13_13_14
void FASTCALL call_2_13_13_14(SH2_struct* sh)
{
  SH2mulu(sh,13,13);
}
#define init_2_13_14_14
void FASTCALL call_2_13_14_14(SH2_struct* sh)
{
  SH2mulu(sh,13,14);
}
#define init_2_13_15_14
void FASTCALL call_2_13_15_14(SH2_struct* sh)
{
  SH2mulu(sh,13,15);
}
#define init_2_14_0_14
void FASTCALL call_2_14_0_14(SH2_struct* sh)
{
  SH2mulu(sh,14,0);
}
#define init_2_14_1_14
void FASTCALL call_2_14_1_14(SH2_struct* sh)
{
  SH2mulu(sh,14,1);
}
#define init_2_14_2_14
void FASTCALL call_2_14_2_14(SH2_struct* sh)
{
  SH2mulu(sh,14,2);
}
#define init_2_14_3_14
void FASTCALL call_2_14_3_14(SH2_struct* sh)
{
  SH2mulu(sh,14,3);
}
#define init_2_14_4_14
void FASTCALL call_2_14_4_14(SH2_struct* sh)
{
  SH2mulu(sh,14,4);
}
#define init_2_14_5_14
void FASTCALL call_2_14_5_14(SH2_struct* sh)
{
  SH2mulu(sh,14,5);
}
#define init_2_14_6_14
void FASTCALL call_2_14_6_14(SH2_struct* sh)
{
  SH2mulu(sh,14,6);
}
#define init_2_14_7_14
void FASTCALL call_2_14_7_14(SH2_struct* sh)
{
  SH2mulu(sh,14,7);
}
#define init_2_14_8_14
void FASTCALL call_2_14_8_14(SH2_struct* sh)
{
  SH2mulu(sh,14,8);
}
#define init_2_14_9_14
void FASTCALL call_2_14_9_14(SH2_struct* sh)
{
  SH2mulu(sh,14,9);
}
#define init_2_14_10_14
void FASTCALL call_2_14_10_14(SH2_struct* sh)
{
  SH2mulu(sh,14,10);
}
#define init_2_14_11_14
void FASTCALL call_2_14_11_14(SH2_struct* sh)
{
  SH2mulu(sh,14,11);
}
#define init_2_14_12_14
void FASTCALL call_2_14_12_14(SH2_struct* sh)
{
  SH2mulu(sh,14,12);
}
#define init_2_14_13_14
void FASTCALL call_2_14_13_14(SH2_struct* sh)
{
  SH2mulu(sh,14,13);
}
#define init_2_14_14_14
void FASTCALL call_2_14_14_14(SH2_struct* sh)
{
  SH2mulu(sh,14,14);
}
#define init_2_14_15_14
void FASTCALL call_2_14_15_14(SH2_struct* sh)
{
  SH2mulu(sh,14,15);
}
#define init_2_15_0_14
void FASTCALL call_2_15_0_14(SH2_struct* sh)
{
  SH2mulu(sh,15,0);
}
#define init_2_15_1_14
void FASTCALL call_2_15_1_14(SH2_struct* sh)
{
  SH2mulu(sh,15,1);
}
#define init_2_15_2_14
void FASTCALL call_2_15_2_14(SH2_struct* sh)
{
  SH2mulu(sh,15,2);
}
#define init_2_15_3_14
void FASTCALL call_2_15_3_14(SH2_struct* sh)
{
  SH2mulu(sh,15,3);
}
#define init_2_15_4_14
void FASTCALL call_2_15_4_14(SH2_struct* sh)
{
  SH2mulu(sh,15,4);
}
#define init_2_15_5_14
void FASTCALL call_2_15_5_14(SH2_struct* sh)
{
  SH2mulu(sh,15,5);
}
#define init_2_15_6_14
void FASTCALL call_2_15_6_14(SH2_struct* sh)
{
  SH2mulu(sh,15,6);
}
#define init_2_15_7_14
void FASTCALL call_2_15_7_14(SH2_struct* sh)
{
  SH2mulu(sh,15,7);
}
#define init_2_15_8_14
void FASTCALL call_2_15_8_14(SH2_struct* sh)
{
  SH2mulu(sh,15,8);
}
#define init_2_15_9_14
void FASTCALL call_2_15_9_14(SH2_struct* sh)
{
  SH2mulu(sh,15,9);
}
#define init_2_15_10_14
void FASTCALL call_2_15_10_14(SH2_struct* sh)
{
  SH2mulu(sh,15,10);
}
#define init_2_15_11_14
void FASTCALL call_2_15_11_14(SH2_struct* sh)
{
  SH2mulu(sh,15,11);
}
#define init_2_15_12_14
void FASTCALL call_2_15_12_14(SH2_struct* sh)
{
  SH2mulu(sh,15,12);
}
#define init_2_15_13_14
void FASTCALL call_2_15_13_14(SH2_struct* sh)
{
  SH2mulu(sh,15,13);
}
#define init_2_15_14_14
void FASTCALL call_2_15_14_14(SH2_struct* sh)
{
  SH2mulu(sh,15,14);
}
#define init_2_15_15_14
void FASTCALL call_2_15_15_14(SH2_struct* sh)
{
  SH2mulu(sh,15,15);
}
#define init_2_0_0_15
void FASTCALL call_2_0_0_15(SH2_struct* sh)
{
  SH2muls(sh,0,0);
}
#define init_2_0_1_15
void FASTCALL call_2_0_1_15(SH2_struct* sh)
{
  SH2muls(sh,0,1);
}
#define init_2_0_2_15
void FASTCALL call_2_0_2_15(SH2_struct* sh)
{
  SH2muls(sh,0,2);
}
#define init_2_0_3_15
void FASTCALL call_2_0_3_15(SH2_struct* sh)
{
  SH2muls(sh,0,3);
}
#define init_2_0_4_15
void FASTCALL call_2_0_4_15(SH2_struct* sh)
{
  SH2muls(sh,0,4);
}
#define init_2_0_5_15
void FASTCALL call_2_0_5_15(SH2_struct* sh)
{
  SH2muls(sh,0,5);
}
#define init_2_0_6_15
void FASTCALL call_2_0_6_15(SH2_struct* sh)
{
  SH2muls(sh,0,6);
}
#define init_2_0_7_15
void FASTCALL call_2_0_7_15(SH2_struct* sh)
{
  SH2muls(sh,0,7);
}
#define init_2_0_8_15
void FASTCALL call_2_0_8_15(SH2_struct* sh)
{
  SH2muls(sh,0,8);
}
#define init_2_0_9_15
void FASTCALL call_2_0_9_15(SH2_struct* sh)
{
  SH2muls(sh,0,9);
}
#define init_2_0_10_15
void FASTCALL call_2_0_10_15(SH2_struct* sh)
{
  SH2muls(sh,0,10);
}
#define init_2_0_11_15
void FASTCALL call_2_0_11_15(SH2_struct* sh)
{
  SH2muls(sh,0,11);
}
#define init_2_0_12_15
void FASTCALL call_2_0_12_15(SH2_struct* sh)
{
  SH2muls(sh,0,12);
}
#define init_2_0_13_15
void FASTCALL call_2_0_13_15(SH2_struct* sh)
{
  SH2muls(sh,0,13);
}
#define init_2_0_14_15
void FASTCALL call_2_0_14_15(SH2_struct* sh)
{
  SH2muls(sh,0,14);
}
#define init_2_0_15_15
void FASTCALL call_2_0_15_15(SH2_struct* sh)
{
  SH2muls(sh,0,15);
}
#define init_2_1_0_15
void FASTCALL call_2_1_0_15(SH2_struct* sh)
{
  SH2muls(sh,1,0);
}
#define init_2_1_1_15
void FASTCALL call_2_1_1_15(SH2_struct* sh)
{
  SH2muls(sh,1,1);
}
#define init_2_1_2_15
void FASTCALL call_2_1_2_15(SH2_struct* sh)
{
  SH2muls(sh,1,2);
}
#define init_2_1_3_15
void FASTCALL call_2_1_3_15(SH2_struct* sh)
{
  SH2muls(sh,1,3);
}
#define init_2_1_4_15
void FASTCALL call_2_1_4_15(SH2_struct* sh)
{
  SH2muls(sh,1,4);
}
#define init_2_1_5_15
void FASTCALL call_2_1_5_15(SH2_struct* sh)
{
  SH2muls(sh,1,5);
}
#define init_2_1_6_15
void FASTCALL call_2_1_6_15(SH2_struct* sh)
{
  SH2muls(sh,1,6);
}
#define init_2_1_7_15
void FASTCALL call_2_1_7_15(SH2_struct* sh)
{
  SH2muls(sh,1,7);
}
#define init_2_1_8_15
void FASTCALL call_2_1_8_15(SH2_struct* sh)
{
  SH2muls(sh,1,8);
}
#define init_2_1_9_15
void FASTCALL call_2_1_9_15(SH2_struct* sh)
{
  SH2muls(sh,1,9);
}
#define init_2_1_10_15
void FASTCALL call_2_1_10_15(SH2_struct* sh)
{
  SH2muls(sh,1,10);
}
#define init_2_1_11_15
void FASTCALL call_2_1_11_15(SH2_struct* sh)
{
  SH2muls(sh,1,11);
}
#define init_2_1_12_15
void FASTCALL call_2_1_12_15(SH2_struct* sh)
{
  SH2muls(sh,1,12);
}
#define init_2_1_13_15
void FASTCALL call_2_1_13_15(SH2_struct* sh)
{
  SH2muls(sh,1,13);
}
#define init_2_1_14_15
void FASTCALL call_2_1_14_15(SH2_struct* sh)
{
  SH2muls(sh,1,14);
}
#define init_2_1_15_15
void FASTCALL call_2_1_15_15(SH2_struct* sh)
{
  SH2muls(sh,1,15);
}
#define init_2_2_0_15
void FASTCALL call_2_2_0_15(SH2_struct* sh)
{
  SH2muls(sh,2,0);
}
#define init_2_2_1_15
void FASTCALL call_2_2_1_15(SH2_struct* sh)
{
  SH2muls(sh,2,1);
}
#define init_2_2_2_15
void FASTCALL call_2_2_2_15(SH2_struct* sh)
{
  SH2muls(sh,2,2);
}
#define init_2_2_3_15
void FASTCALL call_2_2_3_15(SH2_struct* sh)
{
  SH2muls(sh,2,3);
}
#define init_2_2_4_15
void FASTCALL call_2_2_4_15(SH2_struct* sh)
{
  SH2muls(sh,2,4);
}
#define init_2_2_5_15
void FASTCALL call_2_2_5_15(SH2_struct* sh)
{
  SH2muls(sh,2,5);
}
#define init_2_2_6_15
void FASTCALL call_2_2_6_15(SH2_struct* sh)
{
  SH2muls(sh,2,6);
}
#define init_2_2_7_15
void FASTCALL call_2_2_7_15(SH2_struct* sh)
{
  SH2muls(sh,2,7);
}
#define init_2_2_8_15
void FASTCALL call_2_2_8_15(SH2_struct* sh)
{
  SH2muls(sh,2,8);
}
#define init_2_2_9_15
void FASTCALL call_2_2_9_15(SH2_struct* sh)
{
  SH2muls(sh,2,9);
}
#define init_2_2_10_15
void FASTCALL call_2_2_10_15(SH2_struct* sh)
{
  SH2muls(sh,2,10);
}
#define init_2_2_11_15
void FASTCALL call_2_2_11_15(SH2_struct* sh)
{
  SH2muls(sh,2,11);
}
#define init_2_2_12_15
void FASTCALL call_2_2_12_15(SH2_struct* sh)
{
  SH2muls(sh,2,12);
}
#define init_2_2_13_15
void FASTCALL call_2_2_13_15(SH2_struct* sh)
{
  SH2muls(sh,2,13);
}
#define init_2_2_14_15
void FASTCALL call_2_2_14_15(SH2_struct* sh)
{
  SH2muls(sh,2,14);
}
#define init_2_2_15_15
void FASTCALL call_2_2_15_15(SH2_struct* sh)
{
  SH2muls(sh,2,15);
}
#define init_2_3_0_15
void FASTCALL call_2_3_0_15(SH2_struct* sh)
{
  SH2muls(sh,3,0);
}
#define init_2_3_1_15
void FASTCALL call_2_3_1_15(SH2_struct* sh)
{
  SH2muls(sh,3,1);
}
#define init_2_3_2_15
void FASTCALL call_2_3_2_15(SH2_struct* sh)
{
  SH2muls(sh,3,2);
}
#define init_2_3_3_15
void FASTCALL call_2_3_3_15(SH2_struct* sh)
{
  SH2muls(sh,3,3);
}
#define init_2_3_4_15
void FASTCALL call_2_3_4_15(SH2_struct* sh)
{
  SH2muls(sh,3,4);
}
#define init_2_3_5_15
void FASTCALL call_2_3_5_15(SH2_struct* sh)
{
  SH2muls(sh,3,5);
}
#define init_2_3_6_15
void FASTCALL call_2_3_6_15(SH2_struct* sh)
{
  SH2muls(sh,3,6);
}
#define init_2_3_7_15
void FASTCALL call_2_3_7_15(SH2_struct* sh)
{
  SH2muls(sh,3,7);
}
#define init_2_3_8_15
void FASTCALL call_2_3_8_15(SH2_struct* sh)
{
  SH2muls(sh,3,8);
}
#define init_2_3_9_15
void FASTCALL call_2_3_9_15(SH2_struct* sh)
{
  SH2muls(sh,3,9);
}
#define init_2_3_10_15
void FASTCALL call_2_3_10_15(SH2_struct* sh)
{
  SH2muls(sh,3,10);
}
#define init_2_3_11_15
void FASTCALL call_2_3_11_15(SH2_struct* sh)
{
  SH2muls(sh,3,11);
}
#define init_2_3_12_15
void FASTCALL call_2_3_12_15(SH2_struct* sh)
{
  SH2muls(sh,3,12);
}
#define init_2_3_13_15
void FASTCALL call_2_3_13_15(SH2_struct* sh)
{
  SH2muls(sh,3,13);
}
#define init_2_3_14_15
void FASTCALL call_2_3_14_15(SH2_struct* sh)
{
  SH2muls(sh,3,14);
}
#define init_2_3_15_15
void FASTCALL call_2_3_15_15(SH2_struct* sh)
{
  SH2muls(sh,3,15);
}
#define init_2_4_0_15
void FASTCALL call_2_4_0_15(SH2_struct* sh)
{
  SH2muls(sh,4,0);
}
#define init_2_4_1_15
void FASTCALL call_2_4_1_15(SH2_struct* sh)
{
  SH2muls(sh,4,1);
}
#define init_2_4_2_15
void FASTCALL call_2_4_2_15(SH2_struct* sh)
{
  SH2muls(sh,4,2);
}
#define init_2_4_3_15
void FASTCALL call_2_4_3_15(SH2_struct* sh)
{
  SH2muls(sh,4,3);
}
#define init_2_4_4_15
void FASTCALL call_2_4_4_15(SH2_struct* sh)
{
  SH2muls(sh,4,4);
}
#define init_2_4_5_15
void FASTCALL call_2_4_5_15(SH2_struct* sh)
{
  SH2muls(sh,4,5);
}
#define init_2_4_6_15
void FASTCALL call_2_4_6_15(SH2_struct* sh)
{
  SH2muls(sh,4,6);
}
#define init_2_4_7_15
void FASTCALL call_2_4_7_15(SH2_struct* sh)
{
  SH2muls(sh,4,7);
}
#define init_2_4_8_15
void FASTCALL call_2_4_8_15(SH2_struct* sh)
{
  SH2muls(sh,4,8);
}
#define init_2_4_9_15
void FASTCALL call_2_4_9_15(SH2_struct* sh)
{
  SH2muls(sh,4,9);
}
#define init_2_4_10_15
void FASTCALL call_2_4_10_15(SH2_struct* sh)
{
  SH2muls(sh,4,10);
}
#define init_2_4_11_15
void FASTCALL call_2_4_11_15(SH2_struct* sh)
{
  SH2muls(sh,4,11);
}
#define init_2_4_12_15
void FASTCALL call_2_4_12_15(SH2_struct* sh)
{
  SH2muls(sh,4,12);
}
#define init_2_4_13_15
void FASTCALL call_2_4_13_15(SH2_struct* sh)
{
  SH2muls(sh,4,13);
}
#define init_2_4_14_15
void FASTCALL call_2_4_14_15(SH2_struct* sh)
{
  SH2muls(sh,4,14);
}
#define init_2_4_15_15
void FASTCALL call_2_4_15_15(SH2_struct* sh)
{
  SH2muls(sh,4,15);
}
#define init_2_5_0_15
void FASTCALL call_2_5_0_15(SH2_struct* sh)
{
  SH2muls(sh,5,0);
}
#define init_2_5_1_15
void FASTCALL call_2_5_1_15(SH2_struct* sh)
{
  SH2muls(sh,5,1);
}
#define init_2_5_2_15
void FASTCALL call_2_5_2_15(SH2_struct* sh)
{
  SH2muls(sh,5,2);
}
#define init_2_5_3_15
void FASTCALL call_2_5_3_15(SH2_struct* sh)
{
  SH2muls(sh,5,3);
}
#define init_2_5_4_15
void FASTCALL call_2_5_4_15(SH2_struct* sh)
{
  SH2muls(sh,5,4);
}
#define init_2_5_5_15
void FASTCALL call_2_5_5_15(SH2_struct* sh)
{
  SH2muls(sh,5,5);
}
#define init_2_5_6_15
void FASTCALL call_2_5_6_15(SH2_struct* sh)
{
  SH2muls(sh,5,6);
}
#define init_2_5_7_15
void FASTCALL call_2_5_7_15(SH2_struct* sh)
{
  SH2muls(sh,5,7);
}
#define init_2_5_8_15
void FASTCALL call_2_5_8_15(SH2_struct* sh)
{
  SH2muls(sh,5,8);
}
#define init_2_5_9_15
void FASTCALL call_2_5_9_15(SH2_struct* sh)
{
  SH2muls(sh,5,9);
}
#define init_2_5_10_15
void FASTCALL call_2_5_10_15(SH2_struct* sh)
{
  SH2muls(sh,5,10);
}
#define init_2_5_11_15
void FASTCALL call_2_5_11_15(SH2_struct* sh)
{
  SH2muls(sh,5,11);
}
#define init_2_5_12_15
void FASTCALL call_2_5_12_15(SH2_struct* sh)
{
  SH2muls(sh,5,12);
}
#define init_2_5_13_15
void FASTCALL call_2_5_13_15(SH2_struct* sh)
{
  SH2muls(sh,5,13);
}
#define init_2_5_14_15
void FASTCALL call_2_5_14_15(SH2_struct* sh)
{
  SH2muls(sh,5,14);
}
#define init_2_5_15_15
void FASTCALL call_2_5_15_15(SH2_struct* sh)
{
  SH2muls(sh,5,15);
}
#define init_2_6_0_15
void FASTCALL call_2_6_0_15(SH2_struct* sh)
{
  SH2muls(sh,6,0);
}
#define init_2_6_1_15
void FASTCALL call_2_6_1_15(SH2_struct* sh)
{
  SH2muls(sh,6,1);
}
#define init_2_6_2_15
void FASTCALL call_2_6_2_15(SH2_struct* sh)
{
  SH2muls(sh,6,2);
}
#define init_2_6_3_15
void FASTCALL call_2_6_3_15(SH2_struct* sh)
{
  SH2muls(sh,6,3);
}
#define init_2_6_4_15
void FASTCALL call_2_6_4_15(SH2_struct* sh)
{
  SH2muls(sh,6,4);
}
#define init_2_6_5_15
void FASTCALL call_2_6_5_15(SH2_struct* sh)
{
  SH2muls(sh,6,5);
}
#define init_2_6_6_15
void FASTCALL call_2_6_6_15(SH2_struct* sh)
{
  SH2muls(sh,6,6);
}
#define init_2_6_7_15
void FASTCALL call_2_6_7_15(SH2_struct* sh)
{
  SH2muls(sh,6,7);
}
#define init_2_6_8_15
void FASTCALL call_2_6_8_15(SH2_struct* sh)
{
  SH2muls(sh,6,8);
}
#define init_2_6_9_15
void FASTCALL call_2_6_9_15(SH2_struct* sh)
{
  SH2muls(sh,6,9);
}
#define init_2_6_10_15
void FASTCALL call_2_6_10_15(SH2_struct* sh)
{
  SH2muls(sh,6,10);
}
#define init_2_6_11_15
void FASTCALL call_2_6_11_15(SH2_struct* sh)
{
  SH2muls(sh,6,11);
}
#define init_2_6_12_15
void FASTCALL call_2_6_12_15(SH2_struct* sh)
{
  SH2muls(sh,6,12);
}
#define init_2_6_13_15
void FASTCALL call_2_6_13_15(SH2_struct* sh)
{
  SH2muls(sh,6,13);
}
#define init_2_6_14_15
void FASTCALL call_2_6_14_15(SH2_struct* sh)
{
  SH2muls(sh,6,14);
}
#define init_2_6_15_15
void FASTCALL call_2_6_15_15(SH2_struct* sh)
{
  SH2muls(sh,6,15);
}
#define init_2_7_0_15
void FASTCALL call_2_7_0_15(SH2_struct* sh)
{
  SH2muls(sh,7,0);
}
#define init_2_7_1_15
void FASTCALL call_2_7_1_15(SH2_struct* sh)
{
  SH2muls(sh,7,1);
}
#define init_2_7_2_15
void FASTCALL call_2_7_2_15(SH2_struct* sh)
{
  SH2muls(sh,7,2);
}
#define init_2_7_3_15
void FASTCALL call_2_7_3_15(SH2_struct* sh)
{
  SH2muls(sh,7,3);
}
#define init_2_7_4_15
void FASTCALL call_2_7_4_15(SH2_struct* sh)
{
  SH2muls(sh,7,4);
}
#define init_2_7_5_15
void FASTCALL call_2_7_5_15(SH2_struct* sh)
{
  SH2muls(sh,7,5);
}
#define init_2_7_6_15
void FASTCALL call_2_7_6_15(SH2_struct* sh)
{
  SH2muls(sh,7,6);
}
#define init_2_7_7_15
void FASTCALL call_2_7_7_15(SH2_struct* sh)
{
  SH2muls(sh,7,7);
}
#define init_2_7_8_15
void FASTCALL call_2_7_8_15(SH2_struct* sh)
{
  SH2muls(sh,7,8);
}
#define init_2_7_9_15
void FASTCALL call_2_7_9_15(SH2_struct* sh)
{
  SH2muls(sh,7,9);
}
#define init_2_7_10_15
void FASTCALL call_2_7_10_15(SH2_struct* sh)
{
  SH2muls(sh,7,10);
}
#define init_2_7_11_15
void FASTCALL call_2_7_11_15(SH2_struct* sh)
{
  SH2muls(sh,7,11);
}
#define init_2_7_12_15
void FASTCALL call_2_7_12_15(SH2_struct* sh)
{
  SH2muls(sh,7,12);
}
#define init_2_7_13_15
void FASTCALL call_2_7_13_15(SH2_struct* sh)
{
  SH2muls(sh,7,13);
}
#define init_2_7_14_15
void FASTCALL call_2_7_14_15(SH2_struct* sh)
{
  SH2muls(sh,7,14);
}
#define init_2_7_15_15
void FASTCALL call_2_7_15_15(SH2_struct* sh)
{
  SH2muls(sh,7,15);
}
#define init_2_8_0_15
void FASTCALL call_2_8_0_15(SH2_struct* sh)
{
  SH2muls(sh,8,0);
}
#define init_2_8_1_15
void FASTCALL call_2_8_1_15(SH2_struct* sh)
{
  SH2muls(sh,8,1);
}
#define init_2_8_2_15
void FASTCALL call_2_8_2_15(SH2_struct* sh)
{
  SH2muls(sh,8,2);
}
#define init_2_8_3_15
void FASTCALL call_2_8_3_15(SH2_struct* sh)
{
  SH2muls(sh,8,3);
}
#define init_2_8_4_15
void FASTCALL call_2_8_4_15(SH2_struct* sh)
{
  SH2muls(sh,8,4);
}
#define init_2_8_5_15
void FASTCALL call_2_8_5_15(SH2_struct* sh)
{
  SH2muls(sh,8,5);
}
#define init_2_8_6_15
void FASTCALL call_2_8_6_15(SH2_struct* sh)
{
  SH2muls(sh,8,6);
}
#define init_2_8_7_15
void FASTCALL call_2_8_7_15(SH2_struct* sh)
{
  SH2muls(sh,8,7);
}
#define init_2_8_8_15
void FASTCALL call_2_8_8_15(SH2_struct* sh)
{
  SH2muls(sh,8,8);
}
#define init_2_8_9_15
void FASTCALL call_2_8_9_15(SH2_struct* sh)
{
  SH2muls(sh,8,9);
}
#define init_2_8_10_15
void FASTCALL call_2_8_10_15(SH2_struct* sh)
{
  SH2muls(sh,8,10);
}
#define init_2_8_11_15
void FASTCALL call_2_8_11_15(SH2_struct* sh)
{
  SH2muls(sh,8,11);
}
#define init_2_8_12_15
void FASTCALL call_2_8_12_15(SH2_struct* sh)
{
  SH2muls(sh,8,12);
}
#define init_2_8_13_15
void FASTCALL call_2_8_13_15(SH2_struct* sh)
{
  SH2muls(sh,8,13);
}
#define init_2_8_14_15
void FASTCALL call_2_8_14_15(SH2_struct* sh)
{
  SH2muls(sh,8,14);
}
#define init_2_8_15_15
void FASTCALL call_2_8_15_15(SH2_struct* sh)
{
  SH2muls(sh,8,15);
}
#define init_2_9_0_15
void FASTCALL call_2_9_0_15(SH2_struct* sh)
{
  SH2muls(sh,9,0);
}
#define init_2_9_1_15
void FASTCALL call_2_9_1_15(SH2_struct* sh)
{
  SH2muls(sh,9,1);
}
#define init_2_9_2_15
void FASTCALL call_2_9_2_15(SH2_struct* sh)
{
  SH2muls(sh,9,2);
}
#define init_2_9_3_15
void FASTCALL call_2_9_3_15(SH2_struct* sh)
{
  SH2muls(sh,9,3);
}
#define init_2_9_4_15
void FASTCALL call_2_9_4_15(SH2_struct* sh)
{
  SH2muls(sh,9,4);
}
#define init_2_9_5_15
void FASTCALL call_2_9_5_15(SH2_struct* sh)
{
  SH2muls(sh,9,5);
}
#define init_2_9_6_15
void FASTCALL call_2_9_6_15(SH2_struct* sh)
{
  SH2muls(sh,9,6);
}
#define init_2_9_7_15
void FASTCALL call_2_9_7_15(SH2_struct* sh)
{
  SH2muls(sh,9,7);
}
#define init_2_9_8_15
void FASTCALL call_2_9_8_15(SH2_struct* sh)
{
  SH2muls(sh,9,8);
}
#define init_2_9_9_15
void FASTCALL call_2_9_9_15(SH2_struct* sh)
{
  SH2muls(sh,9,9);
}
#define init_2_9_10_15
void FASTCALL call_2_9_10_15(SH2_struct* sh)
{
  SH2muls(sh,9,10);
}
#define init_2_9_11_15
void FASTCALL call_2_9_11_15(SH2_struct* sh)
{
  SH2muls(sh,9,11);
}
#define init_2_9_12_15
void FASTCALL call_2_9_12_15(SH2_struct* sh)
{
  SH2muls(sh,9,12);
}
#define init_2_9_13_15
void FASTCALL call_2_9_13_15(SH2_struct* sh)
{
  SH2muls(sh,9,13);
}
#define init_2_9_14_15
void FASTCALL call_2_9_14_15(SH2_struct* sh)
{
  SH2muls(sh,9,14);
}
#define init_2_9_15_15
void FASTCALL call_2_9_15_15(SH2_struct* sh)
{
  SH2muls(sh,9,15);
}
#define init_2_10_0_15
void FASTCALL call_2_10_0_15(SH2_struct* sh)
{
  SH2muls(sh,10,0);
}
#define init_2_10_1_15
void FASTCALL call_2_10_1_15(SH2_struct* sh)
{
  SH2muls(sh,10,1);
}
#define init_2_10_2_15
void FASTCALL call_2_10_2_15(SH2_struct* sh)
{
  SH2muls(sh,10,2);
}
#define init_2_10_3_15
void FASTCALL call_2_10_3_15(SH2_struct* sh)
{
  SH2muls(sh,10,3);
}
#define init_2_10_4_15
void FASTCALL call_2_10_4_15(SH2_struct* sh)
{
  SH2muls(sh,10,4);
}
#define init_2_10_5_15
void FASTCALL call_2_10_5_15(SH2_struct* sh)
{
  SH2muls(sh,10,5);
}
#define init_2_10_6_15
void FASTCALL call_2_10_6_15(SH2_struct* sh)
{
  SH2muls(sh,10,6);
}
#define init_2_10_7_15
void FASTCALL call_2_10_7_15(SH2_struct* sh)
{
  SH2muls(sh,10,7);
}
#define init_2_10_8_15
void FASTCALL call_2_10_8_15(SH2_struct* sh)
{
  SH2muls(sh,10,8);
}
#define init_2_10_9_15
void FASTCALL call_2_10_9_15(SH2_struct* sh)
{
  SH2muls(sh,10,9);
}
#define init_2_10_10_15
void FASTCALL call_2_10_10_15(SH2_struct* sh)
{
  SH2muls(sh,10,10);
}
#define init_2_10_11_15
void FASTCALL call_2_10_11_15(SH2_struct* sh)
{
  SH2muls(sh,10,11);
}
#define init_2_10_12_15
void FASTCALL call_2_10_12_15(SH2_struct* sh)
{
  SH2muls(sh,10,12);
}
#define init_2_10_13_15
void FASTCALL call_2_10_13_15(SH2_struct* sh)
{
  SH2muls(sh,10,13);
}
#define init_2_10_14_15
void FASTCALL call_2_10_14_15(SH2_struct* sh)
{
  SH2muls(sh,10,14);
}
#define init_2_10_15_15
void FASTCALL call_2_10_15_15(SH2_struct* sh)
{
  SH2muls(sh,10,15);
}
#define init_2_11_0_15
void FASTCALL call_2_11_0_15(SH2_struct* sh)
{
  SH2muls(sh,11,0);
}
#define init_2_11_1_15
void FASTCALL call_2_11_1_15(SH2_struct* sh)
{
  SH2muls(sh,11,1);
}
#define init_2_11_2_15
void FASTCALL call_2_11_2_15(SH2_struct* sh)
{
  SH2muls(sh,11,2);
}
#define init_2_11_3_15
void FASTCALL call_2_11_3_15(SH2_struct* sh)
{
  SH2muls(sh,11,3);
}
#define init_2_11_4_15
void FASTCALL call_2_11_4_15(SH2_struct* sh)
{
  SH2muls(sh,11,4);
}
#define init_2_11_5_15
void FASTCALL call_2_11_5_15(SH2_struct* sh)
{
  SH2muls(sh,11,5);
}
#define init_2_11_6_15
void FASTCALL call_2_11_6_15(SH2_struct* sh)
{
  SH2muls(sh,11,6);
}
#define init_2_11_7_15
void FASTCALL call_2_11_7_15(SH2_struct* sh)
{
  SH2muls(sh,11,7);
}
#define init_2_11_8_15
void FASTCALL call_2_11_8_15(SH2_struct* sh)
{
  SH2muls(sh,11,8);
}
#define init_2_11_9_15
void FASTCALL call_2_11_9_15(SH2_struct* sh)
{
  SH2muls(sh,11,9);
}
#define init_2_11_10_15
void FASTCALL call_2_11_10_15(SH2_struct* sh)
{
  SH2muls(sh,11,10);
}
#define init_2_11_11_15
void FASTCALL call_2_11_11_15(SH2_struct* sh)
{
  SH2muls(sh,11,11);
}
#define init_2_11_12_15
void FASTCALL call_2_11_12_15(SH2_struct* sh)
{
  SH2muls(sh,11,12);
}
#define init_2_11_13_15
void FASTCALL call_2_11_13_15(SH2_struct* sh)
{
  SH2muls(sh,11,13);
}
#define init_2_11_14_15
void FASTCALL call_2_11_14_15(SH2_struct* sh)
{
  SH2muls(sh,11,14);
}
#define init_2_11_15_15
void FASTCALL call_2_11_15_15(SH2_struct* sh)
{
  SH2muls(sh,11,15);
}
#define init_2_12_0_15
void FASTCALL call_2_12_0_15(SH2_struct* sh)
{
  SH2muls(sh,12,0);
}
#define init_2_12_1_15
void FASTCALL call_2_12_1_15(SH2_struct* sh)
{
  SH2muls(sh,12,1);
}
#define init_2_12_2_15
void FASTCALL call_2_12_2_15(SH2_struct* sh)
{
  SH2muls(sh,12,2);
}
#define init_2_12_3_15
void FASTCALL call_2_12_3_15(SH2_struct* sh)
{
  SH2muls(sh,12,3);
}
#define init_2_12_4_15
void FASTCALL call_2_12_4_15(SH2_struct* sh)
{
  SH2muls(sh,12,4);
}
#define init_2_12_5_15
void FASTCALL call_2_12_5_15(SH2_struct* sh)
{
  SH2muls(sh,12,5);
}
#define init_2_12_6_15
void FASTCALL call_2_12_6_15(SH2_struct* sh)
{
  SH2muls(sh,12,6);
}
#define init_2_12_7_15
void FASTCALL call_2_12_7_15(SH2_struct* sh)
{
  SH2muls(sh,12,7);
}
#define init_2_12_8_15
void FASTCALL call_2_12_8_15(SH2_struct* sh)
{
  SH2muls(sh,12,8);
}
#define init_2_12_9_15
void FASTCALL call_2_12_9_15(SH2_struct* sh)
{
  SH2muls(sh,12,9);
}
#define init_2_12_10_15
void FASTCALL call_2_12_10_15(SH2_struct* sh)
{
  SH2muls(sh,12,10);
}
#define init_2_12_11_15
void FASTCALL call_2_12_11_15(SH2_struct* sh)
{
  SH2muls(sh,12,11);
}
#define init_2_12_12_15
void FASTCALL call_2_12_12_15(SH2_struct* sh)
{
  SH2muls(sh,12,12);
}
#define init_2_12_13_15
void FASTCALL call_2_12_13_15(SH2_struct* sh)
{
  SH2muls(sh,12,13);
}
#define init_2_12_14_15
void FASTCALL call_2_12_14_15(SH2_struct* sh)
{
  SH2muls(sh,12,14);
}
#define init_2_12_15_15
void FASTCALL call_2_12_15_15(SH2_struct* sh)
{
  SH2muls(sh,12,15);
}
#define init_2_13_0_15
void FASTCALL call_2_13_0_15(SH2_struct* sh)
{
  SH2muls(sh,13,0);
}
#define init_2_13_1_15
void FASTCALL call_2_13_1_15(SH2_struct* sh)
{
  SH2muls(sh,13,1);
}
#define init_2_13_2_15
void FASTCALL call_2_13_2_15(SH2_struct* sh)
{
  SH2muls(sh,13,2);
}
#define init_2_13_3_15
void FASTCALL call_2_13_3_15(SH2_struct* sh)
{
  SH2muls(sh,13,3);
}
#define init_2_13_4_15
void FASTCALL call_2_13_4_15(SH2_struct* sh)
{
  SH2muls(sh,13,4);
}
#define init_2_13_5_15
void FASTCALL call_2_13_5_15(SH2_struct* sh)
{
  SH2muls(sh,13,5);
}
#define init_2_13_6_15
void FASTCALL call_2_13_6_15(SH2_struct* sh)
{
  SH2muls(sh,13,6);
}
#define init_2_13_7_15
void FASTCALL call_2_13_7_15(SH2_struct* sh)
{
  SH2muls(sh,13,7);
}
#define init_2_13_8_15
void FASTCALL call_2_13_8_15(SH2_struct* sh)
{
  SH2muls(sh,13,8);
}
#define init_2_13_9_15
void FASTCALL call_2_13_9_15(SH2_struct* sh)
{
  SH2muls(sh,13,9);
}
#define init_2_13_10_15
void FASTCALL call_2_13_10_15(SH2_struct* sh)
{
  SH2muls(sh,13,10);
}
#define init_2_13_11_15
void FASTCALL call_2_13_11_15(SH2_struct* sh)
{
  SH2muls(sh,13,11);
}
#define init_2_13_12_15
void FASTCALL call_2_13_12_15(SH2_struct* sh)
{
  SH2muls(sh,13,12);
}
#define init_2_13_13_15
void FASTCALL call_2_13_13_15(SH2_struct* sh)
{
  SH2muls(sh,13,13);
}
#define init_2_13_14_15
void FASTCALL call_2_13_14_15(SH2_struct* sh)
{
  SH2muls(sh,13,14);
}
#define init_2_13_15_15
void FASTCALL call_2_13_15_15(SH2_struct* sh)
{
  SH2muls(sh,13,15);
}
#define init_2_14_0_15
void FASTCALL call_2_14_0_15(SH2_struct* sh)
{
  SH2muls(sh,14,0);
}
#define init_2_14_1_15
void FASTCALL call_2_14_1_15(SH2_struct* sh)
{
  SH2muls(sh,14,1);
}
#define init_2_14_2_15
void FASTCALL call_2_14_2_15(SH2_struct* sh)
{
  SH2muls(sh,14,2);
}
#define init_2_14_3_15
void FASTCALL call_2_14_3_15(SH2_struct* sh)
{
  SH2muls(sh,14,3);
}
#define init_2_14_4_15
void FASTCALL call_2_14_4_15(SH2_struct* sh)
{
  SH2muls(sh,14,4);
}
#define init_2_14_5_15
void FASTCALL call_2_14_5_15(SH2_struct* sh)
{
  SH2muls(sh,14,5);
}
#define init_2_14_6_15
void FASTCALL call_2_14_6_15(SH2_struct* sh)
{
  SH2muls(sh,14,6);
}
#define init_2_14_7_15
void FASTCALL call_2_14_7_15(SH2_struct* sh)
{
  SH2muls(sh,14,7);
}
#define init_2_14_8_15
void FASTCALL call_2_14_8_15(SH2_struct* sh)
{
  SH2muls(sh,14,8);
}
#define init_2_14_9_15
void FASTCALL call_2_14_9_15(SH2_struct* sh)
{
  SH2muls(sh,14,9);
}
#define init_2_14_10_15
void FASTCALL call_2_14_10_15(SH2_struct* sh)
{
  SH2muls(sh,14,10);
}
#define init_2_14_11_15
void FASTCALL call_2_14_11_15(SH2_struct* sh)
{
  SH2muls(sh,14,11);
}
#define init_2_14_12_15
void FASTCALL call_2_14_12_15(SH2_struct* sh)
{
  SH2muls(sh,14,12);
}
#define init_2_14_13_15
void FASTCALL call_2_14_13_15(SH2_struct* sh)
{
  SH2muls(sh,14,13);
}
#define init_2_14_14_15
void FASTCALL call_2_14_14_15(SH2_struct* sh)
{
  SH2muls(sh,14,14);
}
#define init_2_14_15_15
void FASTCALL call_2_14_15_15(SH2_struct* sh)
{
  SH2muls(sh,14,15);
}
#define init_2_15_0_15
void FASTCALL call_2_15_0_15(SH2_struct* sh)
{
  SH2muls(sh,15,0);
}
#define init_2_15_1_15
void FASTCALL call_2_15_1_15(SH2_struct* sh)
{
  SH2muls(sh,15,1);
}
#define init_2_15_2_15
void FASTCALL call_2_15_2_15(SH2_struct* sh)
{
  SH2muls(sh,15,2);
}
#define init_2_15_3_15
void FASTCALL call_2_15_3_15(SH2_struct* sh)
{
  SH2muls(sh,15,3);
}
#define init_2_15_4_15
void FASTCALL call_2_15_4_15(SH2_struct* sh)
{
  SH2muls(sh,15,4);
}
#define init_2_15_5_15
void FASTCALL call_2_15_5_15(SH2_struct* sh)
{
  SH2muls(sh,15,5);
}
#define init_2_15_6_15
void FASTCALL call_2_15_6_15(SH2_struct* sh)
{
  SH2muls(sh,15,6);
}
#define init_2_15_7_15
void FASTCALL call_2_15_7_15(SH2_struct* sh)
{
  SH2muls(sh,15,7);
}
#define init_2_15_8_15
void FASTCALL call_2_15_8_15(SH2_struct* sh)
{
  SH2muls(sh,15,8);
}
#define init_2_15_9_15
void FASTCALL call_2_15_9_15(SH2_struct* sh)
{
  SH2muls(sh,15,9);
}
#define init_2_15_10_15
void FASTCALL call_2_15_10_15(SH2_struct* sh)
{
  SH2muls(sh,15,10);
}
#define init_2_15_11_15
void FASTCALL call_2_15_11_15(SH2_struct* sh)
{
  SH2muls(sh,15,11);
}
#define init_2_15_12_15
void FASTCALL call_2_15_12_15(SH2_struct* sh)
{
  SH2muls(sh,15,12);
}
#define init_2_15_13_15
void FASTCALL call_2_15_13_15(SH2_struct* sh)
{
  SH2muls(sh,15,13);
}
#define init_2_15_14_15
void FASTCALL call_2_15_14_15(SH2_struct* sh)
{
  SH2muls(sh,15,14);
}
#define init_2_15_15_15
void FASTCALL call_2_15_15_15(SH2_struct* sh)
{
  SH2muls(sh,15,15);
}
#define init_12_1_0_0
void FASTCALL call_12_1_0_0(SH2_struct* sh)
{
  SH2movwsg(sh,(0<<4)+0);
}
#define init_12_1_0_1
void FASTCALL call_12_1_0_1(SH2_struct* sh)
{
  SH2movwsg(sh,(0<<4)+1);
}
#define init_12_1_0_2
void FASTCALL call_12_1_0_2(SH2_struct* sh)
{
  SH2movwsg(sh,(0<<4)+2);
}
#define init_12_1_0_3
void FASTCALL call_12_1_0_3(SH2_struct* sh)
{
  SH2movwsg(sh,(0<<4)+3);
}
#define init_12_1_0_4
void FASTCALL call_12_1_0_4(SH2_struct* sh)
{
  SH2movwsg(sh,(0<<4)+4);
}
#define init_12_1_0_5
void FASTCALL call_12_1_0_5(SH2_struct* sh)
{
  SH2movwsg(sh,(0<<4)+5);
}
#define init_12_1_0_6
void FASTCALL call_12_1_0_6(SH2_struct* sh)
{
  SH2movwsg(sh,(0<<4)+6);
}
#define init_12_1_0_7
void FASTCALL call_12_1_0_7(SH2_struct* sh)
{
  SH2movwsg(sh,(0<<4)+7);
}
#define init_12_1_0_8
void FASTCALL call_12_1_0_8(SH2_struct* sh)
{
  SH2movwsg(sh,(0<<4)+8);
}
#define init_12_1_0_9
void FASTCALL call_12_1_0_9(SH2_struct* sh)
{
  SH2movwsg(sh,(0<<4)+9);
}
#define init_12_1_0_10
void FASTCALL call_12_1_0_10(SH2_struct* sh)
{
  SH2movwsg(sh,(0<<4)+10);
}
#define init_12_1_0_11
void FASTCALL call_12_1_0_11(SH2_struct* sh)
{
  SH2movwsg(sh,(0<<4)+11);
}
#define init_12_1_0_12
void FASTCALL call_12_1_0_12(SH2_struct* sh)
{
  SH2movwsg(sh,(0<<4)+12);
}
#define init_12_1_0_13
void FASTCALL call_12_1_0_13(SH2_struct* sh)
{
  SH2movwsg(sh,(0<<4)+13);
}
#define init_12_1_0_14
void FASTCALL call_12_1_0_14(SH2_struct* sh)
{
  SH2movwsg(sh,(0<<4)+14);
}
#define init_12_1_0_15
void FASTCALL call_12_1_0_15(SH2_struct* sh)
{
  SH2movwsg(sh,(0<<4)+15);
}
#define init_12_1_1_0
void FASTCALL call_12_1_1_0(SH2_struct* sh)
{
  SH2movwsg(sh,(1<<4)+0);
}
#define init_12_1_1_1
void FASTCALL call_12_1_1_1(SH2_struct* sh)
{
  SH2movwsg(sh,(1<<4)+1);
}
#define init_12_1_1_2
void FASTCALL call_12_1_1_2(SH2_struct* sh)
{
  SH2movwsg(sh,(1<<4)+2);
}
#define init_12_1_1_3
void FASTCALL call_12_1_1_3(SH2_struct* sh)
{
  SH2movwsg(sh,(1<<4)+3);
}
#define init_12_1_1_4
void FASTCALL call_12_1_1_4(SH2_struct* sh)
{
  SH2movwsg(sh,(1<<4)+4);
}
#define init_12_1_1_5
void FASTCALL call_12_1_1_5(SH2_struct* sh)
{
  SH2movwsg(sh,(1<<4)+5);
}
#define init_12_1_1_6
void FASTCALL call_12_1_1_6(SH2_struct* sh)
{
  SH2movwsg(sh,(1<<4)+6);
}
#define init_12_1_1_7
void FASTCALL call_12_1_1_7(SH2_struct* sh)
{
  SH2movwsg(sh,(1<<4)+7);
}
#define init_12_1_1_8
void FASTCALL call_12_1_1_8(SH2_struct* sh)
{
  SH2movwsg(sh,(1<<4)+8);
}
#define init_12_1_1_9
void FASTCALL call_12_1_1_9(SH2_struct* sh)
{
  SH2movwsg(sh,(1<<4)+9);
}
#define init_12_1_1_10
void FASTCALL call_12_1_1_10(SH2_struct* sh)
{
  SH2movwsg(sh,(1<<4)+10);
}
#define init_12_1_1_11
void FASTCALL call_12_1_1_11(SH2_struct* sh)
{
  SH2movwsg(sh,(1<<4)+11);
}
#define init_12_1_1_12
void FASTCALL call_12_1_1_12(SH2_struct* sh)
{
  SH2movwsg(sh,(1<<4)+12);
}
#define init_12_1_1_13
void FASTCALL call_12_1_1_13(SH2_struct* sh)
{
  SH2movwsg(sh,(1<<4)+13);
}
#define init_12_1_1_14
void FASTCALL call_12_1_1_14(SH2_struct* sh)
{
  SH2movwsg(sh,(1<<4)+14);
}
#define init_12_1_1_15
void FASTCALL call_12_1_1_15(SH2_struct* sh)
{
  SH2movwsg(sh,(1<<4)+15);
}
#define init_12_1_2_0
void FASTCALL call_12_1_2_0(SH2_struct* sh)
{
  SH2movwsg(sh,(2<<4)+0);
}
#define init_12_1_2_1
void FASTCALL call_12_1_2_1(SH2_struct* sh)
{
  SH2movwsg(sh,(2<<4)+1);
}
#define init_12_1_2_2
void FASTCALL call_12_1_2_2(SH2_struct* sh)
{
  SH2movwsg(sh,(2<<4)+2);
}
#define init_12_1_2_3
void FASTCALL call_12_1_2_3(SH2_struct* sh)
{
  SH2movwsg(sh,(2<<4)+3);
}
#define init_12_1_2_4
void FASTCALL call_12_1_2_4(SH2_struct* sh)
{
  SH2movwsg(sh,(2<<4)+4);
}
#define init_12_1_2_5
void FASTCALL call_12_1_2_5(SH2_struct* sh)
{
  SH2movwsg(sh,(2<<4)+5);
}
#define init_12_1_2_6
void FASTCALL call_12_1_2_6(SH2_struct* sh)
{
  SH2movwsg(sh,(2<<4)+6);
}
#define init_12_1_2_7
void FASTCALL call_12_1_2_7(SH2_struct* sh)
{
  SH2movwsg(sh,(2<<4)+7);
}
#define init_12_1_2_8
void FASTCALL call_12_1_2_8(SH2_struct* sh)
{
  SH2movwsg(sh,(2<<4)+8);
}
#define init_12_1_2_9
void FASTCALL call_12_1_2_9(SH2_struct* sh)
{
  SH2movwsg(sh,(2<<4)+9);
}
#define init_12_1_2_10
void FASTCALL call_12_1_2_10(SH2_struct* sh)
{
  SH2movwsg(sh,(2<<4)+10);
}
#define init_12_1_2_11
void FASTCALL call_12_1_2_11(SH2_struct* sh)
{
  SH2movwsg(sh,(2<<4)+11);
}
#define init_12_1_2_12
void FASTCALL call_12_1_2_12(SH2_struct* sh)
{
  SH2movwsg(sh,(2<<4)+12);
}
#define init_12_1_2_13
void FASTCALL call_12_1_2_13(SH2_struct* sh)
{
  SH2movwsg(sh,(2<<4)+13);
}
#define init_12_1_2_14
void FASTCALL call_12_1_2_14(SH2_struct* sh)
{
  SH2movwsg(sh,(2<<4)+14);
}
#define init_12_1_2_15
void FASTCALL call_12_1_2_15(SH2_struct* sh)
{
  SH2movwsg(sh,(2<<4)+15);
}
#define init_12_1_3_0
void FASTCALL call_12_1_3_0(SH2_struct* sh)
{
  SH2movwsg(sh,(3<<4)+0);
}
#define init_12_1_3_1
void FASTCALL call_12_1_3_1(SH2_struct* sh)
{
  SH2movwsg(sh,(3<<4)+1);
}
#define init_12_1_3_2
void FASTCALL call_12_1_3_2(SH2_struct* sh)
{
  SH2movwsg(sh,(3<<4)+2);
}
#define init_12_1_3_3
void FASTCALL call_12_1_3_3(SH2_struct* sh)
{
  SH2movwsg(sh,(3<<4)+3);
}
#define init_12_1_3_4
void FASTCALL call_12_1_3_4(SH2_struct* sh)
{
  SH2movwsg(sh,(3<<4)+4);
}
#define init_12_1_3_5
void FASTCALL call_12_1_3_5(SH2_struct* sh)
{
  SH2movwsg(sh,(3<<4)+5);
}
#define init_12_1_3_6
void FASTCALL call_12_1_3_6(SH2_struct* sh)
{
  SH2movwsg(sh,(3<<4)+6);
}
#define init_12_1_3_7
void FASTCALL call_12_1_3_7(SH2_struct* sh)
{
  SH2movwsg(sh,(3<<4)+7);
}
#define init_12_1_3_8
void FASTCALL call_12_1_3_8(SH2_struct* sh)
{
  SH2movwsg(sh,(3<<4)+8);
}
#define init_12_1_3_9
void FASTCALL call_12_1_3_9(SH2_struct* sh)
{
  SH2movwsg(sh,(3<<4)+9);
}
#define init_12_1_3_10
void FASTCALL call_12_1_3_10(SH2_struct* sh)
{
  SH2movwsg(sh,(3<<4)+10);
}
#define init_12_1_3_11
void FASTCALL call_12_1_3_11(SH2_struct* sh)
{
  SH2movwsg(sh,(3<<4)+11);
}
#define init_12_1_3_12
void FASTCALL call_12_1_3_12(SH2_struct* sh)
{
  SH2movwsg(sh,(3<<4)+12);
}
#define init_12_1_3_13
void FASTCALL call_12_1_3_13(SH2_struct* sh)
{
  SH2movwsg(sh,(3<<4)+13);
}
#define init_12_1_3_14
void FASTCALL call_12_1_3_14(SH2_struct* sh)
{
  SH2movwsg(sh,(3<<4)+14);
}
#define init_12_1_3_15
void FASTCALL call_12_1_3_15(SH2_struct* sh)
{
  SH2movwsg(sh,(3<<4)+15);
}
#define init_12_1_4_0
void FASTCALL call_12_1_4_0(SH2_struct* sh)
{
  SH2movwsg(sh,(4<<4)+0);
}
#define init_12_1_4_1
void FASTCALL call_12_1_4_1(SH2_struct* sh)
{
  SH2movwsg(sh,(4<<4)+1);
}
#define init_12_1_4_2
void FASTCALL call_12_1_4_2(SH2_struct* sh)
{
  SH2movwsg(sh,(4<<4)+2);
}
#define init_12_1_4_3
void FASTCALL call_12_1_4_3(SH2_struct* sh)
{
  SH2movwsg(sh,(4<<4)+3);
}
#define init_12_1_4_4
void FASTCALL call_12_1_4_4(SH2_struct* sh)
{
  SH2movwsg(sh,(4<<4)+4);
}
#define init_12_1_4_5
void FASTCALL call_12_1_4_5(SH2_struct* sh)
{
  SH2movwsg(sh,(4<<4)+5);
}
#define init_12_1_4_6
void FASTCALL call_12_1_4_6(SH2_struct* sh)
{
  SH2movwsg(sh,(4<<4)+6);
}
#define init_12_1_4_7
void FASTCALL call_12_1_4_7(SH2_struct* sh)
{
  SH2movwsg(sh,(4<<4)+7);
}
#define init_12_1_4_8
void FASTCALL call_12_1_4_8(SH2_struct* sh)
{
  SH2movwsg(sh,(4<<4)+8);
}
#define init_12_1_4_9
void FASTCALL call_12_1_4_9(SH2_struct* sh)
{
  SH2movwsg(sh,(4<<4)+9);
}
#define init_12_1_4_10
void FASTCALL call_12_1_4_10(SH2_struct* sh)
{
  SH2movwsg(sh,(4<<4)+10);
}
#define init_12_1_4_11
void FASTCALL call_12_1_4_11(SH2_struct* sh)
{
  SH2movwsg(sh,(4<<4)+11);
}
#define init_12_1_4_12
void FASTCALL call_12_1_4_12(SH2_struct* sh)
{
  SH2movwsg(sh,(4<<4)+12);
}
#define init_12_1_4_13
void FASTCALL call_12_1_4_13(SH2_struct* sh)
{
  SH2movwsg(sh,(4<<4)+13);
}
#define init_12_1_4_14
void FASTCALL call_12_1_4_14(SH2_struct* sh)
{
  SH2movwsg(sh,(4<<4)+14);
}
#define init_12_1_4_15
void FASTCALL call_12_1_4_15(SH2_struct* sh)
{
  SH2movwsg(sh,(4<<4)+15);
}
#define init_12_1_5_0
void FASTCALL call_12_1_5_0(SH2_struct* sh)
{
  SH2movwsg(sh,(5<<4)+0);
}
#define init_12_1_5_1
void FASTCALL call_12_1_5_1(SH2_struct* sh)
{
  SH2movwsg(sh,(5<<4)+1);
}
#define init_12_1_5_2
void FASTCALL call_12_1_5_2(SH2_struct* sh)
{
  SH2movwsg(sh,(5<<4)+2);
}
#define init_12_1_5_3
void FASTCALL call_12_1_5_3(SH2_struct* sh)
{
  SH2movwsg(sh,(5<<4)+3);
}
#define init_12_1_5_4
void FASTCALL call_12_1_5_4(SH2_struct* sh)
{
  SH2movwsg(sh,(5<<4)+4);
}
#define init_12_1_5_5
void FASTCALL call_12_1_5_5(SH2_struct* sh)
{
  SH2movwsg(sh,(5<<4)+5);
}
#define init_12_1_5_6
void FASTCALL call_12_1_5_6(SH2_struct* sh)
{
  SH2movwsg(sh,(5<<4)+6);
}
#define init_12_1_5_7
void FASTCALL call_12_1_5_7(SH2_struct* sh)
{
  SH2movwsg(sh,(5<<4)+7);
}
#define init_12_1_5_8
void FASTCALL call_12_1_5_8(SH2_struct* sh)
{
  SH2movwsg(sh,(5<<4)+8);
}
#define init_12_1_5_9
void FASTCALL call_12_1_5_9(SH2_struct* sh)
{
  SH2movwsg(sh,(5<<4)+9);
}
#define init_12_1_5_10
void FASTCALL call_12_1_5_10(SH2_struct* sh)
{
  SH2movwsg(sh,(5<<4)+10);
}
#define init_12_1_5_11
void FASTCALL call_12_1_5_11(SH2_struct* sh)
{
  SH2movwsg(sh,(5<<4)+11);
}
#define init_12_1_5_12
void FASTCALL call_12_1_5_12(SH2_struct* sh)
{
  SH2movwsg(sh,(5<<4)+12);
}
#define init_12_1_5_13
void FASTCALL call_12_1_5_13(SH2_struct* sh)
{
  SH2movwsg(sh,(5<<4)+13);
}
#define init_12_1_5_14
void FASTCALL call_12_1_5_14(SH2_struct* sh)
{
  SH2movwsg(sh,(5<<4)+14);
}
#define init_12_1_5_15
void FASTCALL call_12_1_5_15(SH2_struct* sh)
{
  SH2movwsg(sh,(5<<4)+15);
}
#define init_12_1_6_0
void FASTCALL call_12_1_6_0(SH2_struct* sh)
{
  SH2movwsg(sh,(6<<4)+0);
}
#define init_12_1_6_1
void FASTCALL call_12_1_6_1(SH2_struct* sh)
{
  SH2movwsg(sh,(6<<4)+1);
}
#define init_12_1_6_2
void FASTCALL call_12_1_6_2(SH2_struct* sh)
{
  SH2movwsg(sh,(6<<4)+2);
}
#define init_12_1_6_3
void FASTCALL call_12_1_6_3(SH2_struct* sh)
{
  SH2movwsg(sh,(6<<4)+3);
}
#define init_12_1_6_4
void FASTCALL call_12_1_6_4(SH2_struct* sh)
{
  SH2movwsg(sh,(6<<4)+4);
}
#define init_12_1_6_5
void FASTCALL call_12_1_6_5(SH2_struct* sh)
{
  SH2movwsg(sh,(6<<4)+5);
}
#define init_12_1_6_6
void FASTCALL call_12_1_6_6(SH2_struct* sh)
{
  SH2movwsg(sh,(6<<4)+6);
}
#define init_12_1_6_7
void FASTCALL call_12_1_6_7(SH2_struct* sh)
{
  SH2movwsg(sh,(6<<4)+7);
}
#define init_12_1_6_8
void FASTCALL call_12_1_6_8(SH2_struct* sh)
{
  SH2movwsg(sh,(6<<4)+8);
}
#define init_12_1_6_9
void FASTCALL call_12_1_6_9(SH2_struct* sh)
{
  SH2movwsg(sh,(6<<4)+9);
}
#define init_12_1_6_10
void FASTCALL call_12_1_6_10(SH2_struct* sh)
{
  SH2movwsg(sh,(6<<4)+10);
}
#define init_12_1_6_11
void FASTCALL call_12_1_6_11(SH2_struct* sh)
{
  SH2movwsg(sh,(6<<4)+11);
}
#define init_12_1_6_12
void FASTCALL call_12_1_6_12(SH2_struct* sh)
{
  SH2movwsg(sh,(6<<4)+12);
}
#define init_12_1_6_13
void FASTCALL call_12_1_6_13(SH2_struct* sh)
{
  SH2movwsg(sh,(6<<4)+13);
}
#define init_12_1_6_14
void FASTCALL call_12_1_6_14(SH2_struct* sh)
{
  SH2movwsg(sh,(6<<4)+14);
}
#define init_12_1_6_15
void FASTCALL call_12_1_6_15(SH2_struct* sh)
{
  SH2movwsg(sh,(6<<4)+15);
}
#define init_12_1_7_0
void FASTCALL call_12_1_7_0(SH2_struct* sh)
{
  SH2movwsg(sh,(7<<4)+0);
}
#define init_12_1_7_1
void FASTCALL call_12_1_7_1(SH2_struct* sh)
{
  SH2movwsg(sh,(7<<4)+1);
}
#define init_12_1_7_2
void FASTCALL call_12_1_7_2(SH2_struct* sh)
{
  SH2movwsg(sh,(7<<4)+2);
}
#define init_12_1_7_3
void FASTCALL call_12_1_7_3(SH2_struct* sh)
{
  SH2movwsg(sh,(7<<4)+3);
}
#define init_12_1_7_4
void FASTCALL call_12_1_7_4(SH2_struct* sh)
{
  SH2movwsg(sh,(7<<4)+4);
}
#define init_12_1_7_5
void FASTCALL call_12_1_7_5(SH2_struct* sh)
{
  SH2movwsg(sh,(7<<4)+5);
}
#define init_12_1_7_6
void FASTCALL call_12_1_7_6(SH2_struct* sh)
{
  SH2movwsg(sh,(7<<4)+6);
}
#define init_12_1_7_7
void FASTCALL call_12_1_7_7(SH2_struct* sh)
{
  SH2movwsg(sh,(7<<4)+7);
}
#define init_12_1_7_8
void FASTCALL call_12_1_7_8(SH2_struct* sh)
{
  SH2movwsg(sh,(7<<4)+8);
}
#define init_12_1_7_9
void FASTCALL call_12_1_7_9(SH2_struct* sh)
{
  SH2movwsg(sh,(7<<4)+9);
}
#define init_12_1_7_10
void FASTCALL call_12_1_7_10(SH2_struct* sh)
{
  SH2movwsg(sh,(7<<4)+10);
}
#define init_12_1_7_11
void FASTCALL call_12_1_7_11(SH2_struct* sh)
{
  SH2movwsg(sh,(7<<4)+11);
}
#define init_12_1_7_12
void FASTCALL call_12_1_7_12(SH2_struct* sh)
{
  SH2movwsg(sh,(7<<4)+12);
}
#define init_12_1_7_13
void FASTCALL call_12_1_7_13(SH2_struct* sh)
{
  SH2movwsg(sh,(7<<4)+13);
}
#define init_12_1_7_14
void FASTCALL call_12_1_7_14(SH2_struct* sh)
{
  SH2movwsg(sh,(7<<4)+14);
}
#define init_12_1_7_15
void FASTCALL call_12_1_7_15(SH2_struct* sh)
{
  SH2movwsg(sh,(7<<4)+15);
}
#define init_12_1_8_0
void FASTCALL call_12_1_8_0(SH2_struct* sh)
{
  SH2movwsg(sh,(8<<4)+0);
}
#define init_12_1_8_1
void FASTCALL call_12_1_8_1(SH2_struct* sh)
{
  SH2movwsg(sh,(8<<4)+1);
}
#define init_12_1_8_2
void FASTCALL call_12_1_8_2(SH2_struct* sh)
{
  SH2movwsg(sh,(8<<4)+2);
}
#define init_12_1_8_3
void FASTCALL call_12_1_8_3(SH2_struct* sh)
{
  SH2movwsg(sh,(8<<4)+3);
}
#define init_12_1_8_4
void FASTCALL call_12_1_8_4(SH2_struct* sh)
{
  SH2movwsg(sh,(8<<4)+4);
}
#define init_12_1_8_5
void FASTCALL call_12_1_8_5(SH2_struct* sh)
{
  SH2movwsg(sh,(8<<4)+5);
}
#define init_12_1_8_6
void FASTCALL call_12_1_8_6(SH2_struct* sh)
{
  SH2movwsg(sh,(8<<4)+6);
}
#define init_12_1_8_7
void FASTCALL call_12_1_8_7(SH2_struct* sh)
{
  SH2movwsg(sh,(8<<4)+7);
}
#define init_12_1_8_8
void FASTCALL call_12_1_8_8(SH2_struct* sh)
{
  SH2movwsg(sh,(8<<4)+8);
}
#define init_12_1_8_9
void FASTCALL call_12_1_8_9(SH2_struct* sh)
{
  SH2movwsg(sh,(8<<4)+9);
}
#define init_12_1_8_10
void FASTCALL call_12_1_8_10(SH2_struct* sh)
{
  SH2movwsg(sh,(8<<4)+10);
}
#define init_12_1_8_11
void FASTCALL call_12_1_8_11(SH2_struct* sh)
{
  SH2movwsg(sh,(8<<4)+11);
}
#define init_12_1_8_12
void FASTCALL call_12_1_8_12(SH2_struct* sh)
{
  SH2movwsg(sh,(8<<4)+12);
}
#define init_12_1_8_13
void FASTCALL call_12_1_8_13(SH2_struct* sh)
{
  SH2movwsg(sh,(8<<4)+13);
}
#define init_12_1_8_14
void FASTCALL call_12_1_8_14(SH2_struct* sh)
{
  SH2movwsg(sh,(8<<4)+14);
}
#define init_12_1_8_15
void FASTCALL call_12_1_8_15(SH2_struct* sh)
{
  SH2movwsg(sh,(8<<4)+15);
}
#define init_12_1_9_0
void FASTCALL call_12_1_9_0(SH2_struct* sh)
{
  SH2movwsg(sh,(9<<4)+0);
}
#define init_12_1_9_1
void FASTCALL call_12_1_9_1(SH2_struct* sh)
{
  SH2movwsg(sh,(9<<4)+1);
}
#define init_12_1_9_2
void FASTCALL call_12_1_9_2(SH2_struct* sh)
{
  SH2movwsg(sh,(9<<4)+2);
}
#define init_12_1_9_3
void FASTCALL call_12_1_9_3(SH2_struct* sh)
{
  SH2movwsg(sh,(9<<4)+3);
}
#define init_12_1_9_4
void FASTCALL call_12_1_9_4(SH2_struct* sh)
{
  SH2movwsg(sh,(9<<4)+4);
}
#define init_12_1_9_5
void FASTCALL call_12_1_9_5(SH2_struct* sh)
{
  SH2movwsg(sh,(9<<4)+5);
}
#define init_12_1_9_6
void FASTCALL call_12_1_9_6(SH2_struct* sh)
{
  SH2movwsg(sh,(9<<4)+6);
}
#define init_12_1_9_7
void FASTCALL call_12_1_9_7(SH2_struct* sh)
{
  SH2movwsg(sh,(9<<4)+7);
}
#define init_12_1_9_8
void FASTCALL call_12_1_9_8(SH2_struct* sh)
{
  SH2movwsg(sh,(9<<4)+8);
}
#define init_12_1_9_9
void FASTCALL call_12_1_9_9(SH2_struct* sh)
{
  SH2movwsg(sh,(9<<4)+9);
}
#define init_12_1_9_10
void FASTCALL call_12_1_9_10(SH2_struct* sh)
{
  SH2movwsg(sh,(9<<4)+10);
}
#define init_12_1_9_11
void FASTCALL call_12_1_9_11(SH2_struct* sh)
{
  SH2movwsg(sh,(9<<4)+11);
}
#define init_12_1_9_12
void FASTCALL call_12_1_9_12(SH2_struct* sh)
{
  SH2movwsg(sh,(9<<4)+12);
}
#define init_12_1_9_13
void FASTCALL call_12_1_9_13(SH2_struct* sh)
{
  SH2movwsg(sh,(9<<4)+13);
}
#define init_12_1_9_14
void FASTCALL call_12_1_9_14(SH2_struct* sh)
{
  SH2movwsg(sh,(9<<4)+14);
}
#define init_12_1_9_15
void FASTCALL call_12_1_9_15(SH2_struct* sh)
{
  SH2movwsg(sh,(9<<4)+15);
}
#define init_12_1_10_0
void FASTCALL call_12_1_10_0(SH2_struct* sh)
{
  SH2movwsg(sh,(10<<4)+0);
}
#define init_12_1_10_1
void FASTCALL call_12_1_10_1(SH2_struct* sh)
{
  SH2movwsg(sh,(10<<4)+1);
}
#define init_12_1_10_2
void FASTCALL call_12_1_10_2(SH2_struct* sh)
{
  SH2movwsg(sh,(10<<4)+2);
}
#define init_12_1_10_3
void FASTCALL call_12_1_10_3(SH2_struct* sh)
{
  SH2movwsg(sh,(10<<4)+3);
}
#define init_12_1_10_4
void FASTCALL call_12_1_10_4(SH2_struct* sh)
{
  SH2movwsg(sh,(10<<4)+4);
}
#define init_12_1_10_5
void FASTCALL call_12_1_10_5(SH2_struct* sh)
{
  SH2movwsg(sh,(10<<4)+5);
}
#define init_12_1_10_6
void FASTCALL call_12_1_10_6(SH2_struct* sh)
{
  SH2movwsg(sh,(10<<4)+6);
}
#define init_12_1_10_7
void FASTCALL call_12_1_10_7(SH2_struct* sh)
{
  SH2movwsg(sh,(10<<4)+7);
}
#define init_12_1_10_8
void FASTCALL call_12_1_10_8(SH2_struct* sh)
{
  SH2movwsg(sh,(10<<4)+8);
}
#define init_12_1_10_9
void FASTCALL call_12_1_10_9(SH2_struct* sh)
{
  SH2movwsg(sh,(10<<4)+9);
}
#define init_12_1_10_10
void FASTCALL call_12_1_10_10(SH2_struct* sh)
{
  SH2movwsg(sh,(10<<4)+10);
}
#define init_12_1_10_11
void FASTCALL call_12_1_10_11(SH2_struct* sh)
{
  SH2movwsg(sh,(10<<4)+11);
}
#define init_12_1_10_12
void FASTCALL call_12_1_10_12(SH2_struct* sh)
{
  SH2movwsg(sh,(10<<4)+12);
}
#define init_12_1_10_13
void FASTCALL call_12_1_10_13(SH2_struct* sh)
{
  SH2movwsg(sh,(10<<4)+13);
}
#define init_12_1_10_14
void FASTCALL call_12_1_10_14(SH2_struct* sh)
{
  SH2movwsg(sh,(10<<4)+14);
}
#define init_12_1_10_15
void FASTCALL call_12_1_10_15(SH2_struct* sh)
{
  SH2movwsg(sh,(10<<4)+15);
}
#define init_12_1_11_0
void FASTCALL call_12_1_11_0(SH2_struct* sh)
{
  SH2movwsg(sh,(11<<4)+0);
}
#define init_12_1_11_1
void FASTCALL call_12_1_11_1(SH2_struct* sh)
{
  SH2movwsg(sh,(11<<4)+1);
}
#define init_12_1_11_2
void FASTCALL call_12_1_11_2(SH2_struct* sh)
{
  SH2movwsg(sh,(11<<4)+2);
}
#define init_12_1_11_3
void FASTCALL call_12_1_11_3(SH2_struct* sh)
{
  SH2movwsg(sh,(11<<4)+3);
}
#define init_12_1_11_4
void FASTCALL call_12_1_11_4(SH2_struct* sh)
{
  SH2movwsg(sh,(11<<4)+4);
}
#define init_12_1_11_5
void FASTCALL call_12_1_11_5(SH2_struct* sh)
{
  SH2movwsg(sh,(11<<4)+5);
}
#define init_12_1_11_6
void FASTCALL call_12_1_11_6(SH2_struct* sh)
{
  SH2movwsg(sh,(11<<4)+6);
}
#define init_12_1_11_7
void FASTCALL call_12_1_11_7(SH2_struct* sh)
{
  SH2movwsg(sh,(11<<4)+7);
}
#define init_12_1_11_8
void FASTCALL call_12_1_11_8(SH2_struct* sh)
{
  SH2movwsg(sh,(11<<4)+8);
}
#define init_12_1_11_9
void FASTCALL call_12_1_11_9(SH2_struct* sh)
{
  SH2movwsg(sh,(11<<4)+9);
}
#define init_12_1_11_10
void FASTCALL call_12_1_11_10(SH2_struct* sh)
{
  SH2movwsg(sh,(11<<4)+10);
}
#define init_12_1_11_11
void FASTCALL call_12_1_11_11(SH2_struct* sh)
{
  SH2movwsg(sh,(11<<4)+11);
}
#define init_12_1_11_12
void FASTCALL call_12_1_11_12(SH2_struct* sh)
{
  SH2movwsg(sh,(11<<4)+12);
}
#define init_12_1_11_13
void FASTCALL call_12_1_11_13(SH2_struct* sh)
{
  SH2movwsg(sh,(11<<4)+13);
}
#define init_12_1_11_14
void FASTCALL call_12_1_11_14(SH2_struct* sh)
{
  SH2movwsg(sh,(11<<4)+14);
}
#define init_12_1_11_15
void FASTCALL call_12_1_11_15(SH2_struct* sh)
{
  SH2movwsg(sh,(11<<4)+15);
}
#define init_12_1_12_0
void FASTCALL call_12_1_12_0(SH2_struct* sh)
{
  SH2movwsg(sh,(12<<4)+0);
}
#define init_12_1_12_1
void FASTCALL call_12_1_12_1(SH2_struct* sh)
{
  SH2movwsg(sh,(12<<4)+1);
}
#define init_12_1_12_2
void FASTCALL call_12_1_12_2(SH2_struct* sh)
{
  SH2movwsg(sh,(12<<4)+2);
}
#define init_12_1_12_3
void FASTCALL call_12_1_12_3(SH2_struct* sh)
{
  SH2movwsg(sh,(12<<4)+3);
}
#define init_12_1_12_4
void FASTCALL call_12_1_12_4(SH2_struct* sh)
{
  SH2movwsg(sh,(12<<4)+4);
}
#define init_12_1_12_5
void FASTCALL call_12_1_12_5(SH2_struct* sh)
{
  SH2movwsg(sh,(12<<4)+5);
}
#define init_12_1_12_6
void FASTCALL call_12_1_12_6(SH2_struct* sh)
{
  SH2movwsg(sh,(12<<4)+6);
}
#define init_12_1_12_7
void FASTCALL call_12_1_12_7(SH2_struct* sh)
{
  SH2movwsg(sh,(12<<4)+7);
}
#define init_12_1_12_8
void FASTCALL call_12_1_12_8(SH2_struct* sh)
{
  SH2movwsg(sh,(12<<4)+8);
}
#define init_12_1_12_9
void FASTCALL call_12_1_12_9(SH2_struct* sh)
{
  SH2movwsg(sh,(12<<4)+9);
}
#define init_12_1_12_10
void FASTCALL call_12_1_12_10(SH2_struct* sh)
{
  SH2movwsg(sh,(12<<4)+10);
}
#define init_12_1_12_11
void FASTCALL call_12_1_12_11(SH2_struct* sh)
{
  SH2movwsg(sh,(12<<4)+11);
}
#define init_12_1_12_12
void FASTCALL call_12_1_12_12(SH2_struct* sh)
{
  SH2movwsg(sh,(12<<4)+12);
}
#define init_12_1_12_13
void FASTCALL call_12_1_12_13(SH2_struct* sh)
{
  SH2movwsg(sh,(12<<4)+13);
}
#define init_12_1_12_14
void FASTCALL call_12_1_12_14(SH2_struct* sh)
{
  SH2movwsg(sh,(12<<4)+14);
}
#define init_12_1_12_15
void FASTCALL call_12_1_12_15(SH2_struct* sh)
{
  SH2movwsg(sh,(12<<4)+15);
}
#define init_12_1_13_0
void FASTCALL call_12_1_13_0(SH2_struct* sh)
{
  SH2movwsg(sh,(13<<4)+0);
}
#define init_12_1_13_1
void FASTCALL call_12_1_13_1(SH2_struct* sh)
{
  SH2movwsg(sh,(13<<4)+1);
}
#define init_12_1_13_2
void FASTCALL call_12_1_13_2(SH2_struct* sh)
{
  SH2movwsg(sh,(13<<4)+2);
}
#define init_12_1_13_3
void FASTCALL call_12_1_13_3(SH2_struct* sh)
{
  SH2movwsg(sh,(13<<4)+3);
}
#define init_12_1_13_4
void FASTCALL call_12_1_13_4(SH2_struct* sh)
{
  SH2movwsg(sh,(13<<4)+4);
}
#define init_12_1_13_5
void FASTCALL call_12_1_13_5(SH2_struct* sh)
{
  SH2movwsg(sh,(13<<4)+5);
}
#define init_12_1_13_6
void FASTCALL call_12_1_13_6(SH2_struct* sh)
{
  SH2movwsg(sh,(13<<4)+6);
}
#define init_12_1_13_7
void FASTCALL call_12_1_13_7(SH2_struct* sh)
{
  SH2movwsg(sh,(13<<4)+7);
}
#define init_12_1_13_8
void FASTCALL call_12_1_13_8(SH2_struct* sh)
{
  SH2movwsg(sh,(13<<4)+8);
}
#define init_12_1_13_9
void FASTCALL call_12_1_13_9(SH2_struct* sh)
{
  SH2movwsg(sh,(13<<4)+9);
}
#define init_12_1_13_10
void FASTCALL call_12_1_13_10(SH2_struct* sh)
{
  SH2movwsg(sh,(13<<4)+10);
}
#define init_12_1_13_11
void FASTCALL call_12_1_13_11(SH2_struct* sh)
{
  SH2movwsg(sh,(13<<4)+11);
}
#define init_12_1_13_12
void FASTCALL call_12_1_13_12(SH2_struct* sh)
{
  SH2movwsg(sh,(13<<4)+12);
}
#define init_12_1_13_13
void FASTCALL call_12_1_13_13(SH2_struct* sh)
{
  SH2movwsg(sh,(13<<4)+13);
}
#define init_12_1_13_14
void FASTCALL call_12_1_13_14(SH2_struct* sh)
{
  SH2movwsg(sh,(13<<4)+14);
}
#define init_12_1_13_15
void FASTCALL call_12_1_13_15(SH2_struct* sh)
{
  SH2movwsg(sh,(13<<4)+15);
}
#define init_12_1_14_0
void FASTCALL call_12_1_14_0(SH2_struct* sh)
{
  SH2movwsg(sh,(14<<4)+0);
}
#define init_12_1_14_1
void FASTCALL call_12_1_14_1(SH2_struct* sh)
{
  SH2movwsg(sh,(14<<4)+1);
}
#define init_12_1_14_2
void FASTCALL call_12_1_14_2(SH2_struct* sh)
{
  SH2movwsg(sh,(14<<4)+2);
}
#define init_12_1_14_3
void FASTCALL call_12_1_14_3(SH2_struct* sh)
{
  SH2movwsg(sh,(14<<4)+3);
}
#define init_12_1_14_4
void FASTCALL call_12_1_14_4(SH2_struct* sh)
{
  SH2movwsg(sh,(14<<4)+4);
}
#define init_12_1_14_5
void FASTCALL call_12_1_14_5(SH2_struct* sh)
{
  SH2movwsg(sh,(14<<4)+5);
}
#define init_12_1_14_6
void FASTCALL call_12_1_14_6(SH2_struct* sh)
{
  SH2movwsg(sh,(14<<4)+6);
}
#define init_12_1_14_7
void FASTCALL call_12_1_14_7(SH2_struct* sh)
{
  SH2movwsg(sh,(14<<4)+7);
}
#define init_12_1_14_8
void FASTCALL call_12_1_14_8(SH2_struct* sh)
{
  SH2movwsg(sh,(14<<4)+8);
}
#define init_12_1_14_9
void FASTCALL call_12_1_14_9(SH2_struct* sh)
{
  SH2movwsg(sh,(14<<4)+9);
}
#define init_12_1_14_10
void FASTCALL call_12_1_14_10(SH2_struct* sh)
{
  SH2movwsg(sh,(14<<4)+10);
}
#define init_12_1_14_11
void FASTCALL call_12_1_14_11(SH2_struct* sh)
{
  SH2movwsg(sh,(14<<4)+11);
}
#define init_12_1_14_12
void FASTCALL call_12_1_14_12(SH2_struct* sh)
{
  SH2movwsg(sh,(14<<4)+12);
}
#define init_12_1_14_13
void FASTCALL call_12_1_14_13(SH2_struct* sh)
{
  SH2movwsg(sh,(14<<4)+13);
}
#define init_12_1_14_14
void FASTCALL call_12_1_14_14(SH2_struct* sh)
{
  SH2movwsg(sh,(14<<4)+14);
}
#define init_12_1_14_15
void FASTCALL call_12_1_14_15(SH2_struct* sh)
{
  SH2movwsg(sh,(14<<4)+15);
}
#define init_12_1_15_0
void FASTCALL call_12_1_15_0(SH2_struct* sh)
{
  SH2movwsg(sh,(15<<4)+0);
}
#define init_12_1_15_1
void FASTCALL call_12_1_15_1(SH2_struct* sh)
{
  SH2movwsg(sh,(15<<4)+1);
}
#define init_12_1_15_2
void FASTCALL call_12_1_15_2(SH2_struct* sh)
{
  SH2movwsg(sh,(15<<4)+2);
}
#define init_12_1_15_3
void FASTCALL call_12_1_15_3(SH2_struct* sh)
{
  SH2movwsg(sh,(15<<4)+3);
}
#define init_12_1_15_4
void FASTCALL call_12_1_15_4(SH2_struct* sh)
{
  SH2movwsg(sh,(15<<4)+4);
}
#define init_12_1_15_5
void FASTCALL call_12_1_15_5(SH2_struct* sh)
{
  SH2movwsg(sh,(15<<4)+5);
}
#define init_12_1_15_6
void FASTCALL call_12_1_15_6(SH2_struct* sh)
{
  SH2movwsg(sh,(15<<4)+6);
}
#define init_12_1_15_7
void FASTCALL call_12_1_15_7(SH2_struct* sh)
{
  SH2movwsg(sh,(15<<4)+7);
}
#define init_12_1_15_8
void FASTCALL call_12_1_15_8(SH2_struct* sh)
{
  SH2movwsg(sh,(15<<4)+8);
}
#define init_12_1_15_9
void FASTCALL call_12_1_15_9(SH2_struct* sh)
{
  SH2movwsg(sh,(15<<4)+9);
}
#define init_12_1_15_10
void FASTCALL call_12_1_15_10(SH2_struct* sh)
{
  SH2movwsg(sh,(15<<4)+10);
}
#define init_12_1_15_11
void FASTCALL call_12_1_15_11(SH2_struct* sh)
{
  SH2movwsg(sh,(15<<4)+11);
}
#define init_12_1_15_12
void FASTCALL call_12_1_15_12(SH2_struct* sh)
{
  SH2movwsg(sh,(15<<4)+12);
}
#define init_12_1_15_13
void FASTCALL call_12_1_15_13(SH2_struct* sh)
{
  SH2movwsg(sh,(15<<4)+13);
}
#define init_12_1_15_14
void FASTCALL call_12_1_15_14(SH2_struct* sh)
{
  SH2movwsg(sh,(15<<4)+14);
}
#define init_12_1_15_15
void FASTCALL call_12_1_15_15(SH2_struct* sh)
{
  SH2movwsg(sh,(15<<4)+15);
}
#define init_8_1_0_0
void FASTCALL call_8_1_0_0(SH2_struct* sh)
{
  SH2movws4(sh,0,0);
}
#define init_8_1_0_1
void FASTCALL call_8_1_0_1(SH2_struct* sh)
{
  SH2movws4(sh,0,1);
}
#define init_8_1_0_2
void FASTCALL call_8_1_0_2(SH2_struct* sh)
{
  SH2movws4(sh,0,2);
}
#define init_8_1_0_3
void FASTCALL call_8_1_0_3(SH2_struct* sh)
{
  SH2movws4(sh,0,3);
}
#define init_8_1_0_4
void FASTCALL call_8_1_0_4(SH2_struct* sh)
{
  SH2movws4(sh,0,4);
}
#define init_8_1_0_5
void FASTCALL call_8_1_0_5(SH2_struct* sh)
{
  SH2movws4(sh,0,5);
}
#define init_8_1_0_6
void FASTCALL call_8_1_0_6(SH2_struct* sh)
{
  SH2movws4(sh,0,6);
}
#define init_8_1_0_7
void FASTCALL call_8_1_0_7(SH2_struct* sh)
{
  SH2movws4(sh,0,7);
}
#define init_8_1_0_8
void FASTCALL call_8_1_0_8(SH2_struct* sh)
{
  SH2movws4(sh,0,8);
}
#define init_8_1_0_9
void FASTCALL call_8_1_0_9(SH2_struct* sh)
{
  SH2movws4(sh,0,9);
}
#define init_8_1_0_10
void FASTCALL call_8_1_0_10(SH2_struct* sh)
{
  SH2movws4(sh,0,10);
}
#define init_8_1_0_11
void FASTCALL call_8_1_0_11(SH2_struct* sh)
{
  SH2movws4(sh,0,11);
}
#define init_8_1_0_12
void FASTCALL call_8_1_0_12(SH2_struct* sh)
{
  SH2movws4(sh,0,12);
}
#define init_8_1_0_13
void FASTCALL call_8_1_0_13(SH2_struct* sh)
{
  SH2movws4(sh,0,13);
}
#define init_8_1_0_14
void FASTCALL call_8_1_0_14(SH2_struct* sh)
{
  SH2movws4(sh,0,14);
}
#define init_8_1_0_15
void FASTCALL call_8_1_0_15(SH2_struct* sh)
{
  SH2movws4(sh,0,15);
}
#define init_8_1_1_0
void FASTCALL call_8_1_1_0(SH2_struct* sh)
{
  SH2movws4(sh,1,0);
}
#define init_8_1_1_1
void FASTCALL call_8_1_1_1(SH2_struct* sh)
{
  SH2movws4(sh,1,1);
}
#define init_8_1_1_2
void FASTCALL call_8_1_1_2(SH2_struct* sh)
{
  SH2movws4(sh,1,2);
}
#define init_8_1_1_3
void FASTCALL call_8_1_1_3(SH2_struct* sh)
{
  SH2movws4(sh,1,3);
}
#define init_8_1_1_4
void FASTCALL call_8_1_1_4(SH2_struct* sh)
{
  SH2movws4(sh,1,4);
}
#define init_8_1_1_5
void FASTCALL call_8_1_1_5(SH2_struct* sh)
{
  SH2movws4(sh,1,5);
}
#define init_8_1_1_6
void FASTCALL call_8_1_1_6(SH2_struct* sh)
{
  SH2movws4(sh,1,6);
}
#define init_8_1_1_7
void FASTCALL call_8_1_1_7(SH2_struct* sh)
{
  SH2movws4(sh,1,7);
}
#define init_8_1_1_8
void FASTCALL call_8_1_1_8(SH2_struct* sh)
{
  SH2movws4(sh,1,8);
}
#define init_8_1_1_9
void FASTCALL call_8_1_1_9(SH2_struct* sh)
{
  SH2movws4(sh,1,9);
}
#define init_8_1_1_10
void FASTCALL call_8_1_1_10(SH2_struct* sh)
{
  SH2movws4(sh,1,10);
}
#define init_8_1_1_11
void FASTCALL call_8_1_1_11(SH2_struct* sh)
{
  SH2movws4(sh,1,11);
}
#define init_8_1_1_12
void FASTCALL call_8_1_1_12(SH2_struct* sh)
{
  SH2movws4(sh,1,12);
}
#define init_8_1_1_13
void FASTCALL call_8_1_1_13(SH2_struct* sh)
{
  SH2movws4(sh,1,13);
}
#define init_8_1_1_14
void FASTCALL call_8_1_1_14(SH2_struct* sh)
{
  SH2movws4(sh,1,14);
}
#define init_8_1_1_15
void FASTCALL call_8_1_1_15(SH2_struct* sh)
{
  SH2movws4(sh,1,15);
}
#define init_8_1_2_0
void FASTCALL call_8_1_2_0(SH2_struct* sh)
{
  SH2movws4(sh,2,0);
}
#define init_8_1_2_1
void FASTCALL call_8_1_2_1(SH2_struct* sh)
{
  SH2movws4(sh,2,1);
}
#define init_8_1_2_2
void FASTCALL call_8_1_2_2(SH2_struct* sh)
{
  SH2movws4(sh,2,2);
}
#define init_8_1_2_3
void FASTCALL call_8_1_2_3(SH2_struct* sh)
{
  SH2movws4(sh,2,3);
}
#define init_8_1_2_4
void FASTCALL call_8_1_2_4(SH2_struct* sh)
{
  SH2movws4(sh,2,4);
}
#define init_8_1_2_5
void FASTCALL call_8_1_2_5(SH2_struct* sh)
{
  SH2movws4(sh,2,5);
}
#define init_8_1_2_6
void FASTCALL call_8_1_2_6(SH2_struct* sh)
{
  SH2movws4(sh,2,6);
}
#define init_8_1_2_7
void FASTCALL call_8_1_2_7(SH2_struct* sh)
{
  SH2movws4(sh,2,7);
}
#define init_8_1_2_8
void FASTCALL call_8_1_2_8(SH2_struct* sh)
{
  SH2movws4(sh,2,8);
}
#define init_8_1_2_9
void FASTCALL call_8_1_2_9(SH2_struct* sh)
{
  SH2movws4(sh,2,9);
}
#define init_8_1_2_10
void FASTCALL call_8_1_2_10(SH2_struct* sh)
{
  SH2movws4(sh,2,10);
}
#define init_8_1_2_11
void FASTCALL call_8_1_2_11(SH2_struct* sh)
{
  SH2movws4(sh,2,11);
}
#define init_8_1_2_12
void FASTCALL call_8_1_2_12(SH2_struct* sh)
{
  SH2movws4(sh,2,12);
}
#define init_8_1_2_13
void FASTCALL call_8_1_2_13(SH2_struct* sh)
{
  SH2movws4(sh,2,13);
}
#define init_8_1_2_14
void FASTCALL call_8_1_2_14(SH2_struct* sh)
{
  SH2movws4(sh,2,14);
}
#define init_8_1_2_15
void FASTCALL call_8_1_2_15(SH2_struct* sh)
{
  SH2movws4(sh,2,15);
}
#define init_8_1_3_0
void FASTCALL call_8_1_3_0(SH2_struct* sh)
{
  SH2movws4(sh,3,0);
}
#define init_8_1_3_1
void FASTCALL call_8_1_3_1(SH2_struct* sh)
{
  SH2movws4(sh,3,1);
}
#define init_8_1_3_2
void FASTCALL call_8_1_3_2(SH2_struct* sh)
{
  SH2movws4(sh,3,2);
}
#define init_8_1_3_3
void FASTCALL call_8_1_3_3(SH2_struct* sh)
{
  SH2movws4(sh,3,3);
}
#define init_8_1_3_4
void FASTCALL call_8_1_3_4(SH2_struct* sh)
{
  SH2movws4(sh,3,4);
}
#define init_8_1_3_5
void FASTCALL call_8_1_3_5(SH2_struct* sh)
{
  SH2movws4(sh,3,5);
}
#define init_8_1_3_6
void FASTCALL call_8_1_3_6(SH2_struct* sh)
{
  SH2movws4(sh,3,6);
}
#define init_8_1_3_7
void FASTCALL call_8_1_3_7(SH2_struct* sh)
{
  SH2movws4(sh,3,7);
}
#define init_8_1_3_8
void FASTCALL call_8_1_3_8(SH2_struct* sh)
{
  SH2movws4(sh,3,8);
}
#define init_8_1_3_9
void FASTCALL call_8_1_3_9(SH2_struct* sh)
{
  SH2movws4(sh,3,9);
}
#define init_8_1_3_10
void FASTCALL call_8_1_3_10(SH2_struct* sh)
{
  SH2movws4(sh,3,10);
}
#define init_8_1_3_11
void FASTCALL call_8_1_3_11(SH2_struct* sh)
{
  SH2movws4(sh,3,11);
}
#define init_8_1_3_12
void FASTCALL call_8_1_3_12(SH2_struct* sh)
{
  SH2movws4(sh,3,12);
}
#define init_8_1_3_13
void FASTCALL call_8_1_3_13(SH2_struct* sh)
{
  SH2movws4(sh,3,13);
}
#define init_8_1_3_14
void FASTCALL call_8_1_3_14(SH2_struct* sh)
{
  SH2movws4(sh,3,14);
}
#define init_8_1_3_15
void FASTCALL call_8_1_3_15(SH2_struct* sh)
{
  SH2movws4(sh,3,15);
}
#define init_8_1_4_0
void FASTCALL call_8_1_4_0(SH2_struct* sh)
{
  SH2movws4(sh,4,0);
}
#define init_8_1_4_1
void FASTCALL call_8_1_4_1(SH2_struct* sh)
{
  SH2movws4(sh,4,1);
}
#define init_8_1_4_2
void FASTCALL call_8_1_4_2(SH2_struct* sh)
{
  SH2movws4(sh,4,2);
}
#define init_8_1_4_3
void FASTCALL call_8_1_4_3(SH2_struct* sh)
{
  SH2movws4(sh,4,3);
}
#define init_8_1_4_4
void FASTCALL call_8_1_4_4(SH2_struct* sh)
{
  SH2movws4(sh,4,4);
}
#define init_8_1_4_5
void FASTCALL call_8_1_4_5(SH2_struct* sh)
{
  SH2movws4(sh,4,5);
}
#define init_8_1_4_6
void FASTCALL call_8_1_4_6(SH2_struct* sh)
{
  SH2movws4(sh,4,6);
}
#define init_8_1_4_7
void FASTCALL call_8_1_4_7(SH2_struct* sh)
{
  SH2movws4(sh,4,7);
}
#define init_8_1_4_8
void FASTCALL call_8_1_4_8(SH2_struct* sh)
{
  SH2movws4(sh,4,8);
}
#define init_8_1_4_9
void FASTCALL call_8_1_4_9(SH2_struct* sh)
{
  SH2movws4(sh,4,9);
}
#define init_8_1_4_10
void FASTCALL call_8_1_4_10(SH2_struct* sh)
{
  SH2movws4(sh,4,10);
}
#define init_8_1_4_11
void FASTCALL call_8_1_4_11(SH2_struct* sh)
{
  SH2movws4(sh,4,11);
}
#define init_8_1_4_12
void FASTCALL call_8_1_4_12(SH2_struct* sh)
{
  SH2movws4(sh,4,12);
}
#define init_8_1_4_13
void FASTCALL call_8_1_4_13(SH2_struct* sh)
{
  SH2movws4(sh,4,13);
}
#define init_8_1_4_14
void FASTCALL call_8_1_4_14(SH2_struct* sh)
{
  SH2movws4(sh,4,14);
}
#define init_8_1_4_15
void FASTCALL call_8_1_4_15(SH2_struct* sh)
{
  SH2movws4(sh,4,15);
}
#define init_8_1_5_0
void FASTCALL call_8_1_5_0(SH2_struct* sh)
{
  SH2movws4(sh,5,0);
}
#define init_8_1_5_1
void FASTCALL call_8_1_5_1(SH2_struct* sh)
{
  SH2movws4(sh,5,1);
}
#define init_8_1_5_2
void FASTCALL call_8_1_5_2(SH2_struct* sh)
{
  SH2movws4(sh,5,2);
}
#define init_8_1_5_3
void FASTCALL call_8_1_5_3(SH2_struct* sh)
{
  SH2movws4(sh,5,3);
}
#define init_8_1_5_4
void FASTCALL call_8_1_5_4(SH2_struct* sh)
{
  SH2movws4(sh,5,4);
}
#define init_8_1_5_5
void FASTCALL call_8_1_5_5(SH2_struct* sh)
{
  SH2movws4(sh,5,5);
}
#define init_8_1_5_6
void FASTCALL call_8_1_5_6(SH2_struct* sh)
{
  SH2movws4(sh,5,6);
}
#define init_8_1_5_7
void FASTCALL call_8_1_5_7(SH2_struct* sh)
{
  SH2movws4(sh,5,7);
}
#define init_8_1_5_8
void FASTCALL call_8_1_5_8(SH2_struct* sh)
{
  SH2movws4(sh,5,8);
}
#define init_8_1_5_9
void FASTCALL call_8_1_5_9(SH2_struct* sh)
{
  SH2movws4(sh,5,9);
}
#define init_8_1_5_10
void FASTCALL call_8_1_5_10(SH2_struct* sh)
{
  SH2movws4(sh,5,10);
}
#define init_8_1_5_11
void FASTCALL call_8_1_5_11(SH2_struct* sh)
{
  SH2movws4(sh,5,11);
}
#define init_8_1_5_12
void FASTCALL call_8_1_5_12(SH2_struct* sh)
{
  SH2movws4(sh,5,12);
}
#define init_8_1_5_13
void FASTCALL call_8_1_5_13(SH2_struct* sh)
{
  SH2movws4(sh,5,13);
}
#define init_8_1_5_14
void FASTCALL call_8_1_5_14(SH2_struct* sh)
{
  SH2movws4(sh,5,14);
}
#define init_8_1_5_15
void FASTCALL call_8_1_5_15(SH2_struct* sh)
{
  SH2movws4(sh,5,15);
}
#define init_8_1_6_0
void FASTCALL call_8_1_6_0(SH2_struct* sh)
{
  SH2movws4(sh,6,0);
}
#define init_8_1_6_1
void FASTCALL call_8_1_6_1(SH2_struct* sh)
{
  SH2movws4(sh,6,1);
}
#define init_8_1_6_2
void FASTCALL call_8_1_6_2(SH2_struct* sh)
{
  SH2movws4(sh,6,2);
}
#define init_8_1_6_3
void FASTCALL call_8_1_6_3(SH2_struct* sh)
{
  SH2movws4(sh,6,3);
}
#define init_8_1_6_4
void FASTCALL call_8_1_6_4(SH2_struct* sh)
{
  SH2movws4(sh,6,4);
}
#define init_8_1_6_5
void FASTCALL call_8_1_6_5(SH2_struct* sh)
{
  SH2movws4(sh,6,5);
}
#define init_8_1_6_6
void FASTCALL call_8_1_6_6(SH2_struct* sh)
{
  SH2movws4(sh,6,6);
}
#define init_8_1_6_7
void FASTCALL call_8_1_6_7(SH2_struct* sh)
{
  SH2movws4(sh,6,7);
}
#define init_8_1_6_8
void FASTCALL call_8_1_6_8(SH2_struct* sh)
{
  SH2movws4(sh,6,8);
}
#define init_8_1_6_9
void FASTCALL call_8_1_6_9(SH2_struct* sh)
{
  SH2movws4(sh,6,9);
}
#define init_8_1_6_10
void FASTCALL call_8_1_6_10(SH2_struct* sh)
{
  SH2movws4(sh,6,10);
}
#define init_8_1_6_11
void FASTCALL call_8_1_6_11(SH2_struct* sh)
{
  SH2movws4(sh,6,11);
}
#define init_8_1_6_12
void FASTCALL call_8_1_6_12(SH2_struct* sh)
{
  SH2movws4(sh,6,12);
}
#define init_8_1_6_13
void FASTCALL call_8_1_6_13(SH2_struct* sh)
{
  SH2movws4(sh,6,13);
}
#define init_8_1_6_14
void FASTCALL call_8_1_6_14(SH2_struct* sh)
{
  SH2movws4(sh,6,14);
}
#define init_8_1_6_15
void FASTCALL call_8_1_6_15(SH2_struct* sh)
{
  SH2movws4(sh,6,15);
}
#define init_8_1_7_0
void FASTCALL call_8_1_7_0(SH2_struct* sh)
{
  SH2movws4(sh,7,0);
}
#define init_8_1_7_1
void FASTCALL call_8_1_7_1(SH2_struct* sh)
{
  SH2movws4(sh,7,1);
}
#define init_8_1_7_2
void FASTCALL call_8_1_7_2(SH2_struct* sh)
{
  SH2movws4(sh,7,2);
}
#define init_8_1_7_3
void FASTCALL call_8_1_7_3(SH2_struct* sh)
{
  SH2movws4(sh,7,3);
}
#define init_8_1_7_4
void FASTCALL call_8_1_7_4(SH2_struct* sh)
{
  SH2movws4(sh,7,4);
}
#define init_8_1_7_5
void FASTCALL call_8_1_7_5(SH2_struct* sh)
{
  SH2movws4(sh,7,5);
}
#define init_8_1_7_6
void FASTCALL call_8_1_7_6(SH2_struct* sh)
{
  SH2movws4(sh,7,6);
}
#define init_8_1_7_7
void FASTCALL call_8_1_7_7(SH2_struct* sh)
{
  SH2movws4(sh,7,7);
}
#define init_8_1_7_8
void FASTCALL call_8_1_7_8(SH2_struct* sh)
{
  SH2movws4(sh,7,8);
}
#define init_8_1_7_9
void FASTCALL call_8_1_7_9(SH2_struct* sh)
{
  SH2movws4(sh,7,9);
}
#define init_8_1_7_10
void FASTCALL call_8_1_7_10(SH2_struct* sh)
{
  SH2movws4(sh,7,10);
}
#define init_8_1_7_11
void FASTCALL call_8_1_7_11(SH2_struct* sh)
{
  SH2movws4(sh,7,11);
}
#define init_8_1_7_12
void FASTCALL call_8_1_7_12(SH2_struct* sh)
{
  SH2movws4(sh,7,12);
}
#define init_8_1_7_13
void FASTCALL call_8_1_7_13(SH2_struct* sh)
{
  SH2movws4(sh,7,13);
}
#define init_8_1_7_14
void FASTCALL call_8_1_7_14(SH2_struct* sh)
{
  SH2movws4(sh,7,14);
}
#define init_8_1_7_15
void FASTCALL call_8_1_7_15(SH2_struct* sh)
{
  SH2movws4(sh,7,15);
}
#define init_8_1_8_0
void FASTCALL call_8_1_8_0(SH2_struct* sh)
{
  SH2movws4(sh,8,0);
}
#define init_8_1_8_1
void FASTCALL call_8_1_8_1(SH2_struct* sh)
{
  SH2movws4(sh,8,1);
}
#define init_8_1_8_2
void FASTCALL call_8_1_8_2(SH2_struct* sh)
{
  SH2movws4(sh,8,2);
}
#define init_8_1_8_3
void FASTCALL call_8_1_8_3(SH2_struct* sh)
{
  SH2movws4(sh,8,3);
}
#define init_8_1_8_4
void FASTCALL call_8_1_8_4(SH2_struct* sh)
{
  SH2movws4(sh,8,4);
}
#define init_8_1_8_5
void FASTCALL call_8_1_8_5(SH2_struct* sh)
{
  SH2movws4(sh,8,5);
}
#define init_8_1_8_6
void FASTCALL call_8_1_8_6(SH2_struct* sh)
{
  SH2movws4(sh,8,6);
}
#define init_8_1_8_7
void FASTCALL call_8_1_8_7(SH2_struct* sh)
{
  SH2movws4(sh,8,7);
}
#define init_8_1_8_8
void FASTCALL call_8_1_8_8(SH2_struct* sh)
{
  SH2movws4(sh,8,8);
}
#define init_8_1_8_9
void FASTCALL call_8_1_8_9(SH2_struct* sh)
{
  SH2movws4(sh,8,9);
}
#define init_8_1_8_10
void FASTCALL call_8_1_8_10(SH2_struct* sh)
{
  SH2movws4(sh,8,10);
}
#define init_8_1_8_11
void FASTCALL call_8_1_8_11(SH2_struct* sh)
{
  SH2movws4(sh,8,11);
}
#define init_8_1_8_12
void FASTCALL call_8_1_8_12(SH2_struct* sh)
{
  SH2movws4(sh,8,12);
}
#define init_8_1_8_13
void FASTCALL call_8_1_8_13(SH2_struct* sh)
{
  SH2movws4(sh,8,13);
}
#define init_8_1_8_14
void FASTCALL call_8_1_8_14(SH2_struct* sh)
{
  SH2movws4(sh,8,14);
}
#define init_8_1_8_15
void FASTCALL call_8_1_8_15(SH2_struct* sh)
{
  SH2movws4(sh,8,15);
}
#define init_8_1_9_0
void FASTCALL call_8_1_9_0(SH2_struct* sh)
{
  SH2movws4(sh,9,0);
}
#define init_8_1_9_1
void FASTCALL call_8_1_9_1(SH2_struct* sh)
{
  SH2movws4(sh,9,1);
}
#define init_8_1_9_2
void FASTCALL call_8_1_9_2(SH2_struct* sh)
{
  SH2movws4(sh,9,2);
}
#define init_8_1_9_3
void FASTCALL call_8_1_9_3(SH2_struct* sh)
{
  SH2movws4(sh,9,3);
}
#define init_8_1_9_4
void FASTCALL call_8_1_9_4(SH2_struct* sh)
{
  SH2movws4(sh,9,4);
}
#define init_8_1_9_5
void FASTCALL call_8_1_9_5(SH2_struct* sh)
{
  SH2movws4(sh,9,5);
}
#define init_8_1_9_6
void FASTCALL call_8_1_9_6(SH2_struct* sh)
{
  SH2movws4(sh,9,6);
}
#define init_8_1_9_7
void FASTCALL call_8_1_9_7(SH2_struct* sh)
{
  SH2movws4(sh,9,7);
}
#define init_8_1_9_8
void FASTCALL call_8_1_9_8(SH2_struct* sh)
{
  SH2movws4(sh,9,8);
}
#define init_8_1_9_9
void FASTCALL call_8_1_9_9(SH2_struct* sh)
{
  SH2movws4(sh,9,9);
}
#define init_8_1_9_10
void FASTCALL call_8_1_9_10(SH2_struct* sh)
{
  SH2movws4(sh,9,10);
}
#define init_8_1_9_11
void FASTCALL call_8_1_9_11(SH2_struct* sh)
{
  SH2movws4(sh,9,11);
}
#define init_8_1_9_12
void FASTCALL call_8_1_9_12(SH2_struct* sh)
{
  SH2movws4(sh,9,12);
}
#define init_8_1_9_13
void FASTCALL call_8_1_9_13(SH2_struct* sh)
{
  SH2movws4(sh,9,13);
}
#define init_8_1_9_14
void FASTCALL call_8_1_9_14(SH2_struct* sh)
{
  SH2movws4(sh,9,14);
}
#define init_8_1_9_15
void FASTCALL call_8_1_9_15(SH2_struct* sh)
{
  SH2movws4(sh,9,15);
}
#define init_8_1_10_0
void FASTCALL call_8_1_10_0(SH2_struct* sh)
{
  SH2movws4(sh,10,0);
}
#define init_8_1_10_1
void FASTCALL call_8_1_10_1(SH2_struct* sh)
{
  SH2movws4(sh,10,1);
}
#define init_8_1_10_2
void FASTCALL call_8_1_10_2(SH2_struct* sh)
{
  SH2movws4(sh,10,2);
}
#define init_8_1_10_3
void FASTCALL call_8_1_10_3(SH2_struct* sh)
{
  SH2movws4(sh,10,3);
}
#define init_8_1_10_4
void FASTCALL call_8_1_10_4(SH2_struct* sh)
{
  SH2movws4(sh,10,4);
}
#define init_8_1_10_5
void FASTCALL call_8_1_10_5(SH2_struct* sh)
{
  SH2movws4(sh,10,5);
}
#define init_8_1_10_6
void FASTCALL call_8_1_10_6(SH2_struct* sh)
{
  SH2movws4(sh,10,6);
}
#define init_8_1_10_7
void FASTCALL call_8_1_10_7(SH2_struct* sh)
{
  SH2movws4(sh,10,7);
}
#define init_8_1_10_8
void FASTCALL call_8_1_10_8(SH2_struct* sh)
{
  SH2movws4(sh,10,8);
}
#define init_8_1_10_9
void FASTCALL call_8_1_10_9(SH2_struct* sh)
{
  SH2movws4(sh,10,9);
}
#define init_8_1_10_10
void FASTCALL call_8_1_10_10(SH2_struct* sh)
{
  SH2movws4(sh,10,10);
}
#define init_8_1_10_11
void FASTCALL call_8_1_10_11(SH2_struct* sh)
{
  SH2movws4(sh,10,11);
}
#define init_8_1_10_12
void FASTCALL call_8_1_10_12(SH2_struct* sh)
{
  SH2movws4(sh,10,12);
}
#define init_8_1_10_13
void FASTCALL call_8_1_10_13(SH2_struct* sh)
{
  SH2movws4(sh,10,13);
}
#define init_8_1_10_14
void FASTCALL call_8_1_10_14(SH2_struct* sh)
{
  SH2movws4(sh,10,14);
}
#define init_8_1_10_15
void FASTCALL call_8_1_10_15(SH2_struct* sh)
{
  SH2movws4(sh,10,15);
}
#define init_8_1_11_0
void FASTCALL call_8_1_11_0(SH2_struct* sh)
{
  SH2movws4(sh,11,0);
}
#define init_8_1_11_1
void FASTCALL call_8_1_11_1(SH2_struct* sh)
{
  SH2movws4(sh,11,1);
}
#define init_8_1_11_2
void FASTCALL call_8_1_11_2(SH2_struct* sh)
{
  SH2movws4(sh,11,2);
}
#define init_8_1_11_3
void FASTCALL call_8_1_11_3(SH2_struct* sh)
{
  SH2movws4(sh,11,3);
}
#define init_8_1_11_4
void FASTCALL call_8_1_11_4(SH2_struct* sh)
{
  SH2movws4(sh,11,4);
}
#define init_8_1_11_5
void FASTCALL call_8_1_11_5(SH2_struct* sh)
{
  SH2movws4(sh,11,5);
}
#define init_8_1_11_6
void FASTCALL call_8_1_11_6(SH2_struct* sh)
{
  SH2movws4(sh,11,6);
}
#define init_8_1_11_7
void FASTCALL call_8_1_11_7(SH2_struct* sh)
{
  SH2movws4(sh,11,7);
}
#define init_8_1_11_8
void FASTCALL call_8_1_11_8(SH2_struct* sh)
{
  SH2movws4(sh,11,8);
}
#define init_8_1_11_9
void FASTCALL call_8_1_11_9(SH2_struct* sh)
{
  SH2movws4(sh,11,9);
}
#define init_8_1_11_10
void FASTCALL call_8_1_11_10(SH2_struct* sh)
{
  SH2movws4(sh,11,10);
}
#define init_8_1_11_11
void FASTCALL call_8_1_11_11(SH2_struct* sh)
{
  SH2movws4(sh,11,11);
}
#define init_8_1_11_12
void FASTCALL call_8_1_11_12(SH2_struct* sh)
{
  SH2movws4(sh,11,12);
}
#define init_8_1_11_13
void FASTCALL call_8_1_11_13(SH2_struct* sh)
{
  SH2movws4(sh,11,13);
}
#define init_8_1_11_14
void FASTCALL call_8_1_11_14(SH2_struct* sh)
{
  SH2movws4(sh,11,14);
}
#define init_8_1_11_15
void FASTCALL call_8_1_11_15(SH2_struct* sh)
{
  SH2movws4(sh,11,15);
}
#define init_8_1_12_0
void FASTCALL call_8_1_12_0(SH2_struct* sh)
{
  SH2movws4(sh,12,0);
}
#define init_8_1_12_1
void FASTCALL call_8_1_12_1(SH2_struct* sh)
{
  SH2movws4(sh,12,1);
}
#define init_8_1_12_2
void FASTCALL call_8_1_12_2(SH2_struct* sh)
{
  SH2movws4(sh,12,2);
}
#define init_8_1_12_3
void FASTCALL call_8_1_12_3(SH2_struct* sh)
{
  SH2movws4(sh,12,3);
}
#define init_8_1_12_4
void FASTCALL call_8_1_12_4(SH2_struct* sh)
{
  SH2movws4(sh,12,4);
}
#define init_8_1_12_5
void FASTCALL call_8_1_12_5(SH2_struct* sh)
{
  SH2movws4(sh,12,5);
}
#define init_8_1_12_6
void FASTCALL call_8_1_12_6(SH2_struct* sh)
{
  SH2movws4(sh,12,6);
}
#define init_8_1_12_7
void FASTCALL call_8_1_12_7(SH2_struct* sh)
{
  SH2movws4(sh,12,7);
}
#define init_8_1_12_8
void FASTCALL call_8_1_12_8(SH2_struct* sh)
{
  SH2movws4(sh,12,8);
}
#define init_8_1_12_9
void FASTCALL call_8_1_12_9(SH2_struct* sh)
{
  SH2movws4(sh,12,9);
}
#define init_8_1_12_10
void FASTCALL call_8_1_12_10(SH2_struct* sh)
{
  SH2movws4(sh,12,10);
}
#define init_8_1_12_11
void FASTCALL call_8_1_12_11(SH2_struct* sh)
{
  SH2movws4(sh,12,11);
}
#define init_8_1_12_12
void FASTCALL call_8_1_12_12(SH2_struct* sh)
{
  SH2movws4(sh,12,12);
}
#define init_8_1_12_13
void FASTCALL call_8_1_12_13(SH2_struct* sh)
{
  SH2movws4(sh,12,13);
}
#define init_8_1_12_14
void FASTCALL call_8_1_12_14(SH2_struct* sh)
{
  SH2movws4(sh,12,14);
}
#define init_8_1_12_15
void FASTCALL call_8_1_12_15(SH2_struct* sh)
{
  SH2movws4(sh,12,15);
}
#define init_8_1_13_0
void FASTCALL call_8_1_13_0(SH2_struct* sh)
{
  SH2movws4(sh,13,0);
}
#define init_8_1_13_1
void FASTCALL call_8_1_13_1(SH2_struct* sh)
{
  SH2movws4(sh,13,1);
}
#define init_8_1_13_2
void FASTCALL call_8_1_13_2(SH2_struct* sh)
{
  SH2movws4(sh,13,2);
}
#define init_8_1_13_3
void FASTCALL call_8_1_13_3(SH2_struct* sh)
{
  SH2movws4(sh,13,3);
}
#define init_8_1_13_4
void FASTCALL call_8_1_13_4(SH2_struct* sh)
{
  SH2movws4(sh,13,4);
}
#define init_8_1_13_5
void FASTCALL call_8_1_13_5(SH2_struct* sh)
{
  SH2movws4(sh,13,5);
}
#define init_8_1_13_6
void FASTCALL call_8_1_13_6(SH2_struct* sh)
{
  SH2movws4(sh,13,6);
}
#define init_8_1_13_7
void FASTCALL call_8_1_13_7(SH2_struct* sh)
{
  SH2movws4(sh,13,7);
}
#define init_8_1_13_8
void FASTCALL call_8_1_13_8(SH2_struct* sh)
{
  SH2movws4(sh,13,8);
}
#define init_8_1_13_9
void FASTCALL call_8_1_13_9(SH2_struct* sh)
{
  SH2movws4(sh,13,9);
}
#define init_8_1_13_10
void FASTCALL call_8_1_13_10(SH2_struct* sh)
{
  SH2movws4(sh,13,10);
}
#define init_8_1_13_11
void FASTCALL call_8_1_13_11(SH2_struct* sh)
{
  SH2movws4(sh,13,11);
}
#define init_8_1_13_12
void FASTCALL call_8_1_13_12(SH2_struct* sh)
{
  SH2movws4(sh,13,12);
}
#define init_8_1_13_13
void FASTCALL call_8_1_13_13(SH2_struct* sh)
{
  SH2movws4(sh,13,13);
}
#define init_8_1_13_14
void FASTCALL call_8_1_13_14(SH2_struct* sh)
{
  SH2movws4(sh,13,14);
}
#define init_8_1_13_15
void FASTCALL call_8_1_13_15(SH2_struct* sh)
{
  SH2movws4(sh,13,15);
}
#define init_8_1_14_0
void FASTCALL call_8_1_14_0(SH2_struct* sh)
{
  SH2movws4(sh,14,0);
}
#define init_8_1_14_1
void FASTCALL call_8_1_14_1(SH2_struct* sh)
{
  SH2movws4(sh,14,1);
}
#define init_8_1_14_2
void FASTCALL call_8_1_14_2(SH2_struct* sh)
{
  SH2movws4(sh,14,2);
}
#define init_8_1_14_3
void FASTCALL call_8_1_14_3(SH2_struct* sh)
{
  SH2movws4(sh,14,3);
}
#define init_8_1_14_4
void FASTCALL call_8_1_14_4(SH2_struct* sh)
{
  SH2movws4(sh,14,4);
}
#define init_8_1_14_5
void FASTCALL call_8_1_14_5(SH2_struct* sh)
{
  SH2movws4(sh,14,5);
}
#define init_8_1_14_6
void FASTCALL call_8_1_14_6(SH2_struct* sh)
{
  SH2movws4(sh,14,6);
}
#define init_8_1_14_7
void FASTCALL call_8_1_14_7(SH2_struct* sh)
{
  SH2movws4(sh,14,7);
}
#define init_8_1_14_8
void FASTCALL call_8_1_14_8(SH2_struct* sh)
{
  SH2movws4(sh,14,8);
}
#define init_8_1_14_9
void FASTCALL call_8_1_14_9(SH2_struct* sh)
{
  SH2movws4(sh,14,9);
}
#define init_8_1_14_10
void FASTCALL call_8_1_14_10(SH2_struct* sh)
{
  SH2movws4(sh,14,10);
}
#define init_8_1_14_11
void FASTCALL call_8_1_14_11(SH2_struct* sh)
{
  SH2movws4(sh,14,11);
}
#define init_8_1_14_12
void FASTCALL call_8_1_14_12(SH2_struct* sh)
{
  SH2movws4(sh,14,12);
}
#define init_8_1_14_13
void FASTCALL call_8_1_14_13(SH2_struct* sh)
{
  SH2movws4(sh,14,13);
}
#define init_8_1_14_14
void FASTCALL call_8_1_14_14(SH2_struct* sh)
{
  SH2movws4(sh,14,14);
}
#define init_8_1_14_15
void FASTCALL call_8_1_14_15(SH2_struct* sh)
{
  SH2movws4(sh,14,15);
}
#define init_8_1_15_0
void FASTCALL call_8_1_15_0(SH2_struct* sh)
{
  SH2movws4(sh,15,0);
}
#define init_8_1_15_1
void FASTCALL call_8_1_15_1(SH2_struct* sh)
{
  SH2movws4(sh,15,1);
}
#define init_8_1_15_2
void FASTCALL call_8_1_15_2(SH2_struct* sh)
{
  SH2movws4(sh,15,2);
}
#define init_8_1_15_3
void FASTCALL call_8_1_15_3(SH2_struct* sh)
{
  SH2movws4(sh,15,3);
}
#define init_8_1_15_4
void FASTCALL call_8_1_15_4(SH2_struct* sh)
{
  SH2movws4(sh,15,4);
}
#define init_8_1_15_5
void FASTCALL call_8_1_15_5(SH2_struct* sh)
{
  SH2movws4(sh,15,5);
}
#define init_8_1_15_6
void FASTCALL call_8_1_15_6(SH2_struct* sh)
{
  SH2movws4(sh,15,6);
}
#define init_8_1_15_7
void FASTCALL call_8_1_15_7(SH2_struct* sh)
{
  SH2movws4(sh,15,7);
}
#define init_8_1_15_8
void FASTCALL call_8_1_15_8(SH2_struct* sh)
{
  SH2movws4(sh,15,8);
}
#define init_8_1_15_9
void FASTCALL call_8_1_15_9(SH2_struct* sh)
{
  SH2movws4(sh,15,9);
}
#define init_8_1_15_10
void FASTCALL call_8_1_15_10(SH2_struct* sh)
{
  SH2movws4(sh,15,10);
}
#define init_8_1_15_11
void FASTCALL call_8_1_15_11(SH2_struct* sh)
{
  SH2movws4(sh,15,11);
}
#define init_8_1_15_12
void FASTCALL call_8_1_15_12(SH2_struct* sh)
{
  SH2movws4(sh,15,12);
}
#define init_8_1_15_13
void FASTCALL call_8_1_15_13(SH2_struct* sh)
{
  SH2movws4(sh,15,13);
}
#define init_8_1_15_14
void FASTCALL call_8_1_15_14(SH2_struct* sh)
{
  SH2movws4(sh,15,14);
}
#define init_8_1_15_15
void FASTCALL call_8_1_15_15(SH2_struct* sh)
{
  SH2movws4(sh,15,15);
}
#define init_2_0_0_1
void FASTCALL call_2_0_0_1(SH2_struct* sh)
{
  SH2movws(sh,0,0);
}
#define init_2_0_1_1
void FASTCALL call_2_0_1_1(SH2_struct* sh)
{
  SH2movws(sh,0,1);
}
#define init_2_0_2_1
void FASTCALL call_2_0_2_1(SH2_struct* sh)
{
  SH2movws(sh,0,2);
}
#define init_2_0_3_1
void FASTCALL call_2_0_3_1(SH2_struct* sh)
{
  SH2movws(sh,0,3);
}
#define init_2_0_4_1
void FASTCALL call_2_0_4_1(SH2_struct* sh)
{
  SH2movws(sh,0,4);
}
#define init_2_0_5_1
void FASTCALL call_2_0_5_1(SH2_struct* sh)
{
  SH2movws(sh,0,5);
}
#define init_2_0_6_1
void FASTCALL call_2_0_6_1(SH2_struct* sh)
{
  SH2movws(sh,0,6);
}
#define init_2_0_7_1
void FASTCALL call_2_0_7_1(SH2_struct* sh)
{
  SH2movws(sh,0,7);
}
#define init_2_0_8_1
void FASTCALL call_2_0_8_1(SH2_struct* sh)
{
  SH2movws(sh,0,8);
}
#define init_2_0_9_1
void FASTCALL call_2_0_9_1(SH2_struct* sh)
{
  SH2movws(sh,0,9);
}
#define init_2_0_10_1
void FASTCALL call_2_0_10_1(SH2_struct* sh)
{
  SH2movws(sh,0,10);
}
#define init_2_0_11_1
void FASTCALL call_2_0_11_1(SH2_struct* sh)
{
  SH2movws(sh,0,11);
}
#define init_2_0_12_1
void FASTCALL call_2_0_12_1(SH2_struct* sh)
{
  SH2movws(sh,0,12);
}
#define init_2_0_13_1
void FASTCALL call_2_0_13_1(SH2_struct* sh)
{
  SH2movws(sh,0,13);
}
#define init_2_0_14_1
void FASTCALL call_2_0_14_1(SH2_struct* sh)
{
  SH2movws(sh,0,14);
}
#define init_2_0_15_1
void FASTCALL call_2_0_15_1(SH2_struct* sh)
{
  SH2movws(sh,0,15);
}
#define init_2_1_0_1
void FASTCALL call_2_1_0_1(SH2_struct* sh)
{
  SH2movws(sh,1,0);
}
#define init_2_1_1_1
void FASTCALL call_2_1_1_1(SH2_struct* sh)
{
  SH2movws(sh,1,1);
}
#define init_2_1_2_1
void FASTCALL call_2_1_2_1(SH2_struct* sh)
{
  SH2movws(sh,1,2);
}
#define init_2_1_3_1
void FASTCALL call_2_1_3_1(SH2_struct* sh)
{
  SH2movws(sh,1,3);
}
#define init_2_1_4_1
void FASTCALL call_2_1_4_1(SH2_struct* sh)
{
  SH2movws(sh,1,4);
}
#define init_2_1_5_1
void FASTCALL call_2_1_5_1(SH2_struct* sh)
{
  SH2movws(sh,1,5);
}
#define init_2_1_6_1
void FASTCALL call_2_1_6_1(SH2_struct* sh)
{
  SH2movws(sh,1,6);
}
#define init_2_1_7_1
void FASTCALL call_2_1_7_1(SH2_struct* sh)
{
  SH2movws(sh,1,7);
}
#define init_2_1_8_1
void FASTCALL call_2_1_8_1(SH2_struct* sh)
{
  SH2movws(sh,1,8);
}
#define init_2_1_9_1
void FASTCALL call_2_1_9_1(SH2_struct* sh)
{
  SH2movws(sh,1,9);
}
#define init_2_1_10_1
void FASTCALL call_2_1_10_1(SH2_struct* sh)
{
  SH2movws(sh,1,10);
}
#define init_2_1_11_1
void FASTCALL call_2_1_11_1(SH2_struct* sh)
{
  SH2movws(sh,1,11);
}
#define init_2_1_12_1
void FASTCALL call_2_1_12_1(SH2_struct* sh)
{
  SH2movws(sh,1,12);
}
#define init_2_1_13_1
void FASTCALL call_2_1_13_1(SH2_struct* sh)
{
  SH2movws(sh,1,13);
}
#define init_2_1_14_1
void FASTCALL call_2_1_14_1(SH2_struct* sh)
{
  SH2movws(sh,1,14);
}
#define init_2_1_15_1
void FASTCALL call_2_1_15_1(SH2_struct* sh)
{
  SH2movws(sh,1,15);
}
#define init_2_2_0_1
void FASTCALL call_2_2_0_1(SH2_struct* sh)
{
  SH2movws(sh,2,0);
}
#define init_2_2_1_1
void FASTCALL call_2_2_1_1(SH2_struct* sh)
{
  SH2movws(sh,2,1);
}
#define init_2_2_2_1
void FASTCALL call_2_2_2_1(SH2_struct* sh)
{
  SH2movws(sh,2,2);
}
#define init_2_2_3_1
void FASTCALL call_2_2_3_1(SH2_struct* sh)
{
  SH2movws(sh,2,3);
}
#define init_2_2_4_1
void FASTCALL call_2_2_4_1(SH2_struct* sh)
{
  SH2movws(sh,2,4);
}
#define init_2_2_5_1
void FASTCALL call_2_2_5_1(SH2_struct* sh)
{
  SH2movws(sh,2,5);
}
#define init_2_2_6_1
void FASTCALL call_2_2_6_1(SH2_struct* sh)
{
  SH2movws(sh,2,6);
}
#define init_2_2_7_1
void FASTCALL call_2_2_7_1(SH2_struct* sh)
{
  SH2movws(sh,2,7);
}
#define init_2_2_8_1
void FASTCALL call_2_2_8_1(SH2_struct* sh)
{
  SH2movws(sh,2,8);
}
#define init_2_2_9_1
void FASTCALL call_2_2_9_1(SH2_struct* sh)
{
  SH2movws(sh,2,9);
}
#define init_2_2_10_1
void FASTCALL call_2_2_10_1(SH2_struct* sh)
{
  SH2movws(sh,2,10);
}
#define init_2_2_11_1
void FASTCALL call_2_2_11_1(SH2_struct* sh)
{
  SH2movws(sh,2,11);
}
#define init_2_2_12_1
void FASTCALL call_2_2_12_1(SH2_struct* sh)
{
  SH2movws(sh,2,12);
}
#define init_2_2_13_1
void FASTCALL call_2_2_13_1(SH2_struct* sh)
{
  SH2movws(sh,2,13);
}
#define init_2_2_14_1
void FASTCALL call_2_2_14_1(SH2_struct* sh)
{
  SH2movws(sh,2,14);
}
#define init_2_2_15_1
void FASTCALL call_2_2_15_1(SH2_struct* sh)
{
  SH2movws(sh,2,15);
}
#define init_2_3_0_1
void FASTCALL call_2_3_0_1(SH2_struct* sh)
{
  SH2movws(sh,3,0);
}
#define init_2_3_1_1
void FASTCALL call_2_3_1_1(SH2_struct* sh)
{
  SH2movws(sh,3,1);
}
#define init_2_3_2_1
void FASTCALL call_2_3_2_1(SH2_struct* sh)
{
  SH2movws(sh,3,2);
}
#define init_2_3_3_1
void FASTCALL call_2_3_3_1(SH2_struct* sh)
{
  SH2movws(sh,3,3);
}
#define init_2_3_4_1
void FASTCALL call_2_3_4_1(SH2_struct* sh)
{
  SH2movws(sh,3,4);
}
#define init_2_3_5_1
void FASTCALL call_2_3_5_1(SH2_struct* sh)
{
  SH2movws(sh,3,5);
}
#define init_2_3_6_1
void FASTCALL call_2_3_6_1(SH2_struct* sh)
{
  SH2movws(sh,3,6);
}
#define init_2_3_7_1
void FASTCALL call_2_3_7_1(SH2_struct* sh)
{
  SH2movws(sh,3,7);
}
#define init_2_3_8_1
void FASTCALL call_2_3_8_1(SH2_struct* sh)
{
  SH2movws(sh,3,8);
}
#define init_2_3_9_1
void FASTCALL call_2_3_9_1(SH2_struct* sh)
{
  SH2movws(sh,3,9);
}
#define init_2_3_10_1
void FASTCALL call_2_3_10_1(SH2_struct* sh)
{
  SH2movws(sh,3,10);
}
#define init_2_3_11_1
void FASTCALL call_2_3_11_1(SH2_struct* sh)
{
  SH2movws(sh,3,11);
}
#define init_2_3_12_1
void FASTCALL call_2_3_12_1(SH2_struct* sh)
{
  SH2movws(sh,3,12);
}
#define init_2_3_13_1
void FASTCALL call_2_3_13_1(SH2_struct* sh)
{
  SH2movws(sh,3,13);
}
#define init_2_3_14_1
void FASTCALL call_2_3_14_1(SH2_struct* sh)
{
  SH2movws(sh,3,14);
}
#define init_2_3_15_1
void FASTCALL call_2_3_15_1(SH2_struct* sh)
{
  SH2movws(sh,3,15);
}
#define init_2_4_0_1
void FASTCALL call_2_4_0_1(SH2_struct* sh)
{
  SH2movws(sh,4,0);
}
#define init_2_4_1_1
void FASTCALL call_2_4_1_1(SH2_struct* sh)
{
  SH2movws(sh,4,1);
}
#define init_2_4_2_1
void FASTCALL call_2_4_2_1(SH2_struct* sh)
{
  SH2movws(sh,4,2);
}
#define init_2_4_3_1
void FASTCALL call_2_4_3_1(SH2_struct* sh)
{
  SH2movws(sh,4,3);
}
#define init_2_4_4_1
void FASTCALL call_2_4_4_1(SH2_struct* sh)
{
  SH2movws(sh,4,4);
}
#define init_2_4_5_1
void FASTCALL call_2_4_5_1(SH2_struct* sh)
{
  SH2movws(sh,4,5);
}
#define init_2_4_6_1
void FASTCALL call_2_4_6_1(SH2_struct* sh)
{
  SH2movws(sh,4,6);
}
#define init_2_4_7_1
void FASTCALL call_2_4_7_1(SH2_struct* sh)
{
  SH2movws(sh,4,7);
}
#define init_2_4_8_1
void FASTCALL call_2_4_8_1(SH2_struct* sh)
{
  SH2movws(sh,4,8);
}
#define init_2_4_9_1
void FASTCALL call_2_4_9_1(SH2_struct* sh)
{
  SH2movws(sh,4,9);
}
#define init_2_4_10_1
void FASTCALL call_2_4_10_1(SH2_struct* sh)
{
  SH2movws(sh,4,10);
}
#define init_2_4_11_1
void FASTCALL call_2_4_11_1(SH2_struct* sh)
{
  SH2movws(sh,4,11);
}
#define init_2_4_12_1
void FASTCALL call_2_4_12_1(SH2_struct* sh)
{
  SH2movws(sh,4,12);
}
#define init_2_4_13_1
void FASTCALL call_2_4_13_1(SH2_struct* sh)
{
  SH2movws(sh,4,13);
}
#define init_2_4_14_1
void FASTCALL call_2_4_14_1(SH2_struct* sh)
{
  SH2movws(sh,4,14);
}
#define init_2_4_15_1
void FASTCALL call_2_4_15_1(SH2_struct* sh)
{
  SH2movws(sh,4,15);
}
#define init_2_5_0_1
void FASTCALL call_2_5_0_1(SH2_struct* sh)
{
  SH2movws(sh,5,0);
}
#define init_2_5_1_1
void FASTCALL call_2_5_1_1(SH2_struct* sh)
{
  SH2movws(sh,5,1);
}
#define init_2_5_2_1
void FASTCALL call_2_5_2_1(SH2_struct* sh)
{
  SH2movws(sh,5,2);
}
#define init_2_5_3_1
void FASTCALL call_2_5_3_1(SH2_struct* sh)
{
  SH2movws(sh,5,3);
}
#define init_2_5_4_1
void FASTCALL call_2_5_4_1(SH2_struct* sh)
{
  SH2movws(sh,5,4);
}
#define init_2_5_5_1
void FASTCALL call_2_5_5_1(SH2_struct* sh)
{
  SH2movws(sh,5,5);
}
#define init_2_5_6_1
void FASTCALL call_2_5_6_1(SH2_struct* sh)
{
  SH2movws(sh,5,6);
}
#define init_2_5_7_1
void FASTCALL call_2_5_7_1(SH2_struct* sh)
{
  SH2movws(sh,5,7);
}
#define init_2_5_8_1
void FASTCALL call_2_5_8_1(SH2_struct* sh)
{
  SH2movws(sh,5,8);
}
#define init_2_5_9_1
void FASTCALL call_2_5_9_1(SH2_struct* sh)
{
  SH2movws(sh,5,9);
}
#define init_2_5_10_1
void FASTCALL call_2_5_10_1(SH2_struct* sh)
{
  SH2movws(sh,5,10);
}
#define init_2_5_11_1
void FASTCALL call_2_5_11_1(SH2_struct* sh)
{
  SH2movws(sh,5,11);
}
#define init_2_5_12_1
void FASTCALL call_2_5_12_1(SH2_struct* sh)
{
  SH2movws(sh,5,12);
}
#define init_2_5_13_1
void FASTCALL call_2_5_13_1(SH2_struct* sh)
{
  SH2movws(sh,5,13);
}
#define init_2_5_14_1
void FASTCALL call_2_5_14_1(SH2_struct* sh)
{
  SH2movws(sh,5,14);
}
#define init_2_5_15_1
void FASTCALL call_2_5_15_1(SH2_struct* sh)
{
  SH2movws(sh,5,15);
}
#define init_2_6_0_1
void FASTCALL call_2_6_0_1(SH2_struct* sh)
{
  SH2movws(sh,6,0);
}
#define init_2_6_1_1
void FASTCALL call_2_6_1_1(SH2_struct* sh)
{
  SH2movws(sh,6,1);
}
#define init_2_6_2_1
void FASTCALL call_2_6_2_1(SH2_struct* sh)
{
  SH2movws(sh,6,2);
}
#define init_2_6_3_1
void FASTCALL call_2_6_3_1(SH2_struct* sh)
{
  SH2movws(sh,6,3);
}
#define init_2_6_4_1
void FASTCALL call_2_6_4_1(SH2_struct* sh)
{
  SH2movws(sh,6,4);
}
#define init_2_6_5_1
void FASTCALL call_2_6_5_1(SH2_struct* sh)
{
  SH2movws(sh,6,5);
}
#define init_2_6_6_1
void FASTCALL call_2_6_6_1(SH2_struct* sh)
{
  SH2movws(sh,6,6);
}
#define init_2_6_7_1
void FASTCALL call_2_6_7_1(SH2_struct* sh)
{
  SH2movws(sh,6,7);
}
#define init_2_6_8_1
void FASTCALL call_2_6_8_1(SH2_struct* sh)
{
  SH2movws(sh,6,8);
}
#define init_2_6_9_1
void FASTCALL call_2_6_9_1(SH2_struct* sh)
{
  SH2movws(sh,6,9);
}
#define init_2_6_10_1
void FASTCALL call_2_6_10_1(SH2_struct* sh)
{
  SH2movws(sh,6,10);
}
#define init_2_6_11_1
void FASTCALL call_2_6_11_1(SH2_struct* sh)
{
  SH2movws(sh,6,11);
}
#define init_2_6_12_1
void FASTCALL call_2_6_12_1(SH2_struct* sh)
{
  SH2movws(sh,6,12);
}
#define init_2_6_13_1
void FASTCALL call_2_6_13_1(SH2_struct* sh)
{
  SH2movws(sh,6,13);
}
#define init_2_6_14_1
void FASTCALL call_2_6_14_1(SH2_struct* sh)
{
  SH2movws(sh,6,14);
}
#define init_2_6_15_1
void FASTCALL call_2_6_15_1(SH2_struct* sh)
{
  SH2movws(sh,6,15);
}
#define init_2_7_0_1
void FASTCALL call_2_7_0_1(SH2_struct* sh)
{
  SH2movws(sh,7,0);
}
#define init_2_7_1_1
void FASTCALL call_2_7_1_1(SH2_struct* sh)
{
  SH2movws(sh,7,1);
}
#define init_2_7_2_1
void FASTCALL call_2_7_2_1(SH2_struct* sh)
{
  SH2movws(sh,7,2);
}
#define init_2_7_3_1
void FASTCALL call_2_7_3_1(SH2_struct* sh)
{
  SH2movws(sh,7,3);
}
#define init_2_7_4_1
void FASTCALL call_2_7_4_1(SH2_struct* sh)
{
  SH2movws(sh,7,4);
}
#define init_2_7_5_1
void FASTCALL call_2_7_5_1(SH2_struct* sh)
{
  SH2movws(sh,7,5);
}
#define init_2_7_6_1
void FASTCALL call_2_7_6_1(SH2_struct* sh)
{
  SH2movws(sh,7,6);
}
#define init_2_7_7_1
void FASTCALL call_2_7_7_1(SH2_struct* sh)
{
  SH2movws(sh,7,7);
}
#define init_2_7_8_1
void FASTCALL call_2_7_8_1(SH2_struct* sh)
{
  SH2movws(sh,7,8);
}
#define init_2_7_9_1
void FASTCALL call_2_7_9_1(SH2_struct* sh)
{
  SH2movws(sh,7,9);
}
#define init_2_7_10_1
void FASTCALL call_2_7_10_1(SH2_struct* sh)
{
  SH2movws(sh,7,10);
}
#define init_2_7_11_1
void FASTCALL call_2_7_11_1(SH2_struct* sh)
{
  SH2movws(sh,7,11);
}
#define init_2_7_12_1
void FASTCALL call_2_7_12_1(SH2_struct* sh)
{
  SH2movws(sh,7,12);
}
#define init_2_7_13_1
void FASTCALL call_2_7_13_1(SH2_struct* sh)
{
  SH2movws(sh,7,13);
}
#define init_2_7_14_1
void FASTCALL call_2_7_14_1(SH2_struct* sh)
{
  SH2movws(sh,7,14);
}
#define init_2_7_15_1
void FASTCALL call_2_7_15_1(SH2_struct* sh)
{
  SH2movws(sh,7,15);
}
#define init_2_8_0_1
void FASTCALL call_2_8_0_1(SH2_struct* sh)
{
  SH2movws(sh,8,0);
}
#define init_2_8_1_1
void FASTCALL call_2_8_1_1(SH2_struct* sh)
{
  SH2movws(sh,8,1);
}
#define init_2_8_2_1
void FASTCALL call_2_8_2_1(SH2_struct* sh)
{
  SH2movws(sh,8,2);
}
#define init_2_8_3_1
void FASTCALL call_2_8_3_1(SH2_struct* sh)
{
  SH2movws(sh,8,3);
}
#define init_2_8_4_1
void FASTCALL call_2_8_4_1(SH2_struct* sh)
{
  SH2movws(sh,8,4);
}
#define init_2_8_5_1
void FASTCALL call_2_8_5_1(SH2_struct* sh)
{
  SH2movws(sh,8,5);
}
#define init_2_8_6_1
void FASTCALL call_2_8_6_1(SH2_struct* sh)
{
  SH2movws(sh,8,6);
}
#define init_2_8_7_1
void FASTCALL call_2_8_7_1(SH2_struct* sh)
{
  SH2movws(sh,8,7);
}
#define init_2_8_8_1
void FASTCALL call_2_8_8_1(SH2_struct* sh)
{
  SH2movws(sh,8,8);
}
#define init_2_8_9_1
void FASTCALL call_2_8_9_1(SH2_struct* sh)
{
  SH2movws(sh,8,9);
}
#define init_2_8_10_1
void FASTCALL call_2_8_10_1(SH2_struct* sh)
{
  SH2movws(sh,8,10);
}
#define init_2_8_11_1
void FASTCALL call_2_8_11_1(SH2_struct* sh)
{
  SH2movws(sh,8,11);
}
#define init_2_8_12_1
void FASTCALL call_2_8_12_1(SH2_struct* sh)
{
  SH2movws(sh,8,12);
}
#define init_2_8_13_1
void FASTCALL call_2_8_13_1(SH2_struct* sh)
{
  SH2movws(sh,8,13);
}
#define init_2_8_14_1
void FASTCALL call_2_8_14_1(SH2_struct* sh)
{
  SH2movws(sh,8,14);
}
#define init_2_8_15_1
void FASTCALL call_2_8_15_1(SH2_struct* sh)
{
  SH2movws(sh,8,15);
}
#define init_2_9_0_1
void FASTCALL call_2_9_0_1(SH2_struct* sh)
{
  SH2movws(sh,9,0);
}
#define init_2_9_1_1
void FASTCALL call_2_9_1_1(SH2_struct* sh)
{
  SH2movws(sh,9,1);
}
#define init_2_9_2_1
void FASTCALL call_2_9_2_1(SH2_struct* sh)
{
  SH2movws(sh,9,2);
}
#define init_2_9_3_1
void FASTCALL call_2_9_3_1(SH2_struct* sh)
{
  SH2movws(sh,9,3);
}
#define init_2_9_4_1
void FASTCALL call_2_9_4_1(SH2_struct* sh)
{
  SH2movws(sh,9,4);
}
#define init_2_9_5_1
void FASTCALL call_2_9_5_1(SH2_struct* sh)
{
  SH2movws(sh,9,5);
}
#define init_2_9_6_1
void FASTCALL call_2_9_6_1(SH2_struct* sh)
{
  SH2movws(sh,9,6);
}
#define init_2_9_7_1
void FASTCALL call_2_9_7_1(SH2_struct* sh)
{
  SH2movws(sh,9,7);
}
#define init_2_9_8_1
void FASTCALL call_2_9_8_1(SH2_struct* sh)
{
  SH2movws(sh,9,8);
}
#define init_2_9_9_1
void FASTCALL call_2_9_9_1(SH2_struct* sh)
{
  SH2movws(sh,9,9);
}
#define init_2_9_10_1
void FASTCALL call_2_9_10_1(SH2_struct* sh)
{
  SH2movws(sh,9,10);
}
#define init_2_9_11_1
void FASTCALL call_2_9_11_1(SH2_struct* sh)
{
  SH2movws(sh,9,11);
}
#define init_2_9_12_1
void FASTCALL call_2_9_12_1(SH2_struct* sh)
{
  SH2movws(sh,9,12);
}
#define init_2_9_13_1
void FASTCALL call_2_9_13_1(SH2_struct* sh)
{
  SH2movws(sh,9,13);
}
#define init_2_9_14_1
void FASTCALL call_2_9_14_1(SH2_struct* sh)
{
  SH2movws(sh,9,14);
}
#define init_2_9_15_1
void FASTCALL call_2_9_15_1(SH2_struct* sh)
{
  SH2movws(sh,9,15);
}
#define init_2_10_0_1
void FASTCALL call_2_10_0_1(SH2_struct* sh)
{
  SH2movws(sh,10,0);
}
#define init_2_10_1_1
void FASTCALL call_2_10_1_1(SH2_struct* sh)
{
  SH2movws(sh,10,1);
}
#define init_2_10_2_1
void FASTCALL call_2_10_2_1(SH2_struct* sh)
{
  SH2movws(sh,10,2);
}
#define init_2_10_3_1
void FASTCALL call_2_10_3_1(SH2_struct* sh)
{
  SH2movws(sh,10,3);
}
#define init_2_10_4_1
void FASTCALL call_2_10_4_1(SH2_struct* sh)
{
  SH2movws(sh,10,4);
}
#define init_2_10_5_1
void FASTCALL call_2_10_5_1(SH2_struct* sh)
{
  SH2movws(sh,10,5);
}
#define init_2_10_6_1
void FASTCALL call_2_10_6_1(SH2_struct* sh)
{
  SH2movws(sh,10,6);
}
#define init_2_10_7_1
void FASTCALL call_2_10_7_1(SH2_struct* sh)
{
  SH2movws(sh,10,7);
}
#define init_2_10_8_1
void FASTCALL call_2_10_8_1(SH2_struct* sh)
{
  SH2movws(sh,10,8);
}
#define init_2_10_9_1
void FASTCALL call_2_10_9_1(SH2_struct* sh)
{
  SH2movws(sh,10,9);
}
#define init_2_10_10_1
void FASTCALL call_2_10_10_1(SH2_struct* sh)
{
  SH2movws(sh,10,10);
}
#define init_2_10_11_1
void FASTCALL call_2_10_11_1(SH2_struct* sh)
{
  SH2movws(sh,10,11);
}
#define init_2_10_12_1
void FASTCALL call_2_10_12_1(SH2_struct* sh)
{
  SH2movws(sh,10,12);
}
#define init_2_10_13_1
void FASTCALL call_2_10_13_1(SH2_struct* sh)
{
  SH2movws(sh,10,13);
}
#define init_2_10_14_1
void FASTCALL call_2_10_14_1(SH2_struct* sh)
{
  SH2movws(sh,10,14);
}
#define init_2_10_15_1
void FASTCALL call_2_10_15_1(SH2_struct* sh)
{
  SH2movws(sh,10,15);
}
#define init_2_11_0_1
void FASTCALL call_2_11_0_1(SH2_struct* sh)
{
  SH2movws(sh,11,0);
}
#define init_2_11_1_1
void FASTCALL call_2_11_1_1(SH2_struct* sh)
{
  SH2movws(sh,11,1);
}
#define init_2_11_2_1
void FASTCALL call_2_11_2_1(SH2_struct* sh)
{
  SH2movws(sh,11,2);
}
#define init_2_11_3_1
void FASTCALL call_2_11_3_1(SH2_struct* sh)
{
  SH2movws(sh,11,3);
}
#define init_2_11_4_1
void FASTCALL call_2_11_4_1(SH2_struct* sh)
{
  SH2movws(sh,11,4);
}
#define init_2_11_5_1
void FASTCALL call_2_11_5_1(SH2_struct* sh)
{
  SH2movws(sh,11,5);
}
#define init_2_11_6_1
void FASTCALL call_2_11_6_1(SH2_struct* sh)
{
  SH2movws(sh,11,6);
}
#define init_2_11_7_1
void FASTCALL call_2_11_7_1(SH2_struct* sh)
{
  SH2movws(sh,11,7);
}
#define init_2_11_8_1
void FASTCALL call_2_11_8_1(SH2_struct* sh)
{
  SH2movws(sh,11,8);
}
#define init_2_11_9_1
void FASTCALL call_2_11_9_1(SH2_struct* sh)
{
  SH2movws(sh,11,9);
}
#define init_2_11_10_1
void FASTCALL call_2_11_10_1(SH2_struct* sh)
{
  SH2movws(sh,11,10);
}
#define init_2_11_11_1
void FASTCALL call_2_11_11_1(SH2_struct* sh)
{
  SH2movws(sh,11,11);
}
#define init_2_11_12_1
void FASTCALL call_2_11_12_1(SH2_struct* sh)
{
  SH2movws(sh,11,12);
}
#define init_2_11_13_1
void FASTCALL call_2_11_13_1(SH2_struct* sh)
{
  SH2movws(sh,11,13);
}
#define init_2_11_14_1
void FASTCALL call_2_11_14_1(SH2_struct* sh)
{
  SH2movws(sh,11,14);
}
#define init_2_11_15_1
void FASTCALL call_2_11_15_1(SH2_struct* sh)
{
  SH2movws(sh,11,15);
}
#define init_2_12_0_1
void FASTCALL call_2_12_0_1(SH2_struct* sh)
{
  SH2movws(sh,12,0);
}
#define init_2_12_1_1
void FASTCALL call_2_12_1_1(SH2_struct* sh)
{
  SH2movws(sh,12,1);
}
#define init_2_12_2_1
void FASTCALL call_2_12_2_1(SH2_struct* sh)
{
  SH2movws(sh,12,2);
}
#define init_2_12_3_1
void FASTCALL call_2_12_3_1(SH2_struct* sh)
{
  SH2movws(sh,12,3);
}
#define init_2_12_4_1
void FASTCALL call_2_12_4_1(SH2_struct* sh)
{
  SH2movws(sh,12,4);
}
#define init_2_12_5_1
void FASTCALL call_2_12_5_1(SH2_struct* sh)
{
  SH2movws(sh,12,5);
}
#define init_2_12_6_1
void FASTCALL call_2_12_6_1(SH2_struct* sh)
{
  SH2movws(sh,12,6);
}
#define init_2_12_7_1
void FASTCALL call_2_12_7_1(SH2_struct* sh)
{
  SH2movws(sh,12,7);
}
#define init_2_12_8_1
void FASTCALL call_2_12_8_1(SH2_struct* sh)
{
  SH2movws(sh,12,8);
}
#define init_2_12_9_1
void FASTCALL call_2_12_9_1(SH2_struct* sh)
{
  SH2movws(sh,12,9);
}
#define init_2_12_10_1
void FASTCALL call_2_12_10_1(SH2_struct* sh)
{
  SH2movws(sh,12,10);
}
#define init_2_12_11_1
void FASTCALL call_2_12_11_1(SH2_struct* sh)
{
  SH2movws(sh,12,11);
}
#define init_2_12_12_1
void FASTCALL call_2_12_12_1(SH2_struct* sh)
{
  SH2movws(sh,12,12);
}
#define init_2_12_13_1
void FASTCALL call_2_12_13_1(SH2_struct* sh)
{
  SH2movws(sh,12,13);
}
#define init_2_12_14_1
void FASTCALL call_2_12_14_1(SH2_struct* sh)
{
  SH2movws(sh,12,14);
}
#define init_2_12_15_1
void FASTCALL call_2_12_15_1(SH2_struct* sh)
{
  SH2movws(sh,12,15);
}
#define init_2_13_0_1
void FASTCALL call_2_13_0_1(SH2_struct* sh)
{
  SH2movws(sh,13,0);
}
#define init_2_13_1_1
void FASTCALL call_2_13_1_1(SH2_struct* sh)
{
  SH2movws(sh,13,1);
}
#define init_2_13_2_1
void FASTCALL call_2_13_2_1(SH2_struct* sh)
{
  SH2movws(sh,13,2);
}
#define init_2_13_3_1
void FASTCALL call_2_13_3_1(SH2_struct* sh)
{
  SH2movws(sh,13,3);
}
#define init_2_13_4_1
void FASTCALL call_2_13_4_1(SH2_struct* sh)
{
  SH2movws(sh,13,4);
}
#define init_2_13_5_1
void FASTCALL call_2_13_5_1(SH2_struct* sh)
{
  SH2movws(sh,13,5);
}
#define init_2_13_6_1
void FASTCALL call_2_13_6_1(SH2_struct* sh)
{
  SH2movws(sh,13,6);
}
#define init_2_13_7_1
void FASTCALL call_2_13_7_1(SH2_struct* sh)
{
  SH2movws(sh,13,7);
}
#define init_2_13_8_1
void FASTCALL call_2_13_8_1(SH2_struct* sh)
{
  SH2movws(sh,13,8);
}
#define init_2_13_9_1
void FASTCALL call_2_13_9_1(SH2_struct* sh)
{
  SH2movws(sh,13,9);
}
#define init_2_13_10_1
void FASTCALL call_2_13_10_1(SH2_struct* sh)
{
  SH2movws(sh,13,10);
}
#define init_2_13_11_1
void FASTCALL call_2_13_11_1(SH2_struct* sh)
{
  SH2movws(sh,13,11);
}
#define init_2_13_12_1
void FASTCALL call_2_13_12_1(SH2_struct* sh)
{
  SH2movws(sh,13,12);
}
#define init_2_13_13_1
void FASTCALL call_2_13_13_1(SH2_struct* sh)
{
  SH2movws(sh,13,13);
}
#define init_2_13_14_1
void FASTCALL call_2_13_14_1(SH2_struct* sh)
{
  SH2movws(sh,13,14);
}
#define init_2_13_15_1
void FASTCALL call_2_13_15_1(SH2_struct* sh)
{
  SH2movws(sh,13,15);
}
#define init_2_14_0_1
void FASTCALL call_2_14_0_1(SH2_struct* sh)
{
  SH2movws(sh,14,0);
}
#define init_2_14_1_1
void FASTCALL call_2_14_1_1(SH2_struct* sh)
{
  SH2movws(sh,14,1);
}
#define init_2_14_2_1
void FASTCALL call_2_14_2_1(SH2_struct* sh)
{
  SH2movws(sh,14,2);
}
#define init_2_14_3_1
void FASTCALL call_2_14_3_1(SH2_struct* sh)
{
  SH2movws(sh,14,3);
}
#define init_2_14_4_1
void FASTCALL call_2_14_4_1(SH2_struct* sh)
{
  SH2movws(sh,14,4);
}
#define init_2_14_5_1
void FASTCALL call_2_14_5_1(SH2_struct* sh)
{
  SH2movws(sh,14,5);
}
#define init_2_14_6_1
void FASTCALL call_2_14_6_1(SH2_struct* sh)
{
  SH2movws(sh,14,6);
}
#define init_2_14_7_1
void FASTCALL call_2_14_7_1(SH2_struct* sh)
{
  SH2movws(sh,14,7);
}
#define init_2_14_8_1
void FASTCALL call_2_14_8_1(SH2_struct* sh)
{
  SH2movws(sh,14,8);
}
#define init_2_14_9_1
void FASTCALL call_2_14_9_1(SH2_struct* sh)
{
  SH2movws(sh,14,9);
}
#define init_2_14_10_1
void FASTCALL call_2_14_10_1(SH2_struct* sh)
{
  SH2movws(sh,14,10);
}
#define init_2_14_11_1
void FASTCALL call_2_14_11_1(SH2_struct* sh)
{
  SH2movws(sh,14,11);
}
#define init_2_14_12_1
void FASTCALL call_2_14_12_1(SH2_struct* sh)
{
  SH2movws(sh,14,12);
}
#define init_2_14_13_1
void FASTCALL call_2_14_13_1(SH2_struct* sh)
{
  SH2movws(sh,14,13);
}
#define init_2_14_14_1
void FASTCALL call_2_14_14_1(SH2_struct* sh)
{
  SH2movws(sh,14,14);
}
#define init_2_14_15_1
void FASTCALL call_2_14_15_1(SH2_struct* sh)
{
  SH2movws(sh,14,15);
}
#define init_2_15_0_1
void FASTCALL call_2_15_0_1(SH2_struct* sh)
{
  SH2movws(sh,15,0);
}
#define init_2_15_1_1
void FASTCALL call_2_15_1_1(SH2_struct* sh)
{
  SH2movws(sh,15,1);
}
#define init_2_15_2_1
void FASTCALL call_2_15_2_1(SH2_struct* sh)
{
  SH2movws(sh,15,2);
}
#define init_2_15_3_1
void FASTCALL call_2_15_3_1(SH2_struct* sh)
{
  SH2movws(sh,15,3);
}
#define init_2_15_4_1
void FASTCALL call_2_15_4_1(SH2_struct* sh)
{
  SH2movws(sh,15,4);
}
#define init_2_15_5_1
void FASTCALL call_2_15_5_1(SH2_struct* sh)
{
  SH2movws(sh,15,5);
}
#define init_2_15_6_1
void FASTCALL call_2_15_6_1(SH2_struct* sh)
{
  SH2movws(sh,15,6);
}
#define init_2_15_7_1
void FASTCALL call_2_15_7_1(SH2_struct* sh)
{
  SH2movws(sh,15,7);
}
#define init_2_15_8_1
void FASTCALL call_2_15_8_1(SH2_struct* sh)
{
  SH2movws(sh,15,8);
}
#define init_2_15_9_1
void FASTCALL call_2_15_9_1(SH2_struct* sh)
{
  SH2movws(sh,15,9);
}
#define init_2_15_10_1
void FASTCALL call_2_15_10_1(SH2_struct* sh)
{
  SH2movws(sh,15,10);
}
#define init_2_15_11_1
void FASTCALL call_2_15_11_1(SH2_struct* sh)
{
  SH2movws(sh,15,11);
}
#define init_2_15_12_1
void FASTCALL call_2_15_12_1(SH2_struct* sh)
{
  SH2movws(sh,15,12);
}
#define init_2_15_13_1
void FASTCALL call_2_15_13_1(SH2_struct* sh)
{
  SH2movws(sh,15,13);
}
#define init_2_15_14_1
void FASTCALL call_2_15_14_1(SH2_struct* sh)
{
  SH2movws(sh,15,14);
}
#define init_2_15_15_1
void FASTCALL call_2_15_15_1(SH2_struct* sh)
{
  SH2movws(sh,15,15);
}
#define init_6_0_0_5
void FASTCALL call_6_0_0_5(SH2_struct* sh)
{
  SH2movwp(sh,0,0);
}
#define init_6_0_1_5
void FASTCALL call_6_0_1_5(SH2_struct* sh)
{
  SH2movwp(sh,0,1);
}
#define init_6_0_2_5
void FASTCALL call_6_0_2_5(SH2_struct* sh)
{
  SH2movwp(sh,0,2);
}
#define init_6_0_3_5
void FASTCALL call_6_0_3_5(SH2_struct* sh)
{
  SH2movwp(sh,0,3);
}
#define init_6_0_4_5
void FASTCALL call_6_0_4_5(SH2_struct* sh)
{
  SH2movwp(sh,0,4);
}
#define init_6_0_5_5
void FASTCALL call_6_0_5_5(SH2_struct* sh)
{
  SH2movwp(sh,0,5);
}
#define init_6_0_6_5
void FASTCALL call_6_0_6_5(SH2_struct* sh)
{
  SH2movwp(sh,0,6);
}
#define init_6_0_7_5
void FASTCALL call_6_0_7_5(SH2_struct* sh)
{
  SH2movwp(sh,0,7);
}
#define init_6_0_8_5
void FASTCALL call_6_0_8_5(SH2_struct* sh)
{
  SH2movwp(sh,0,8);
}
#define init_6_0_9_5
void FASTCALL call_6_0_9_5(SH2_struct* sh)
{
  SH2movwp(sh,0,9);
}
#define init_6_0_10_5
void FASTCALL call_6_0_10_5(SH2_struct* sh)
{
  SH2movwp(sh,0,10);
}
#define init_6_0_11_5
void FASTCALL call_6_0_11_5(SH2_struct* sh)
{
  SH2movwp(sh,0,11);
}
#define init_6_0_12_5
void FASTCALL call_6_0_12_5(SH2_struct* sh)
{
  SH2movwp(sh,0,12);
}
#define init_6_0_13_5
void FASTCALL call_6_0_13_5(SH2_struct* sh)
{
  SH2movwp(sh,0,13);
}
#define init_6_0_14_5
void FASTCALL call_6_0_14_5(SH2_struct* sh)
{
  SH2movwp(sh,0,14);
}
#define init_6_0_15_5
void FASTCALL call_6_0_15_5(SH2_struct* sh)
{
  SH2movwp(sh,0,15);
}
#define init_6_1_0_5
void FASTCALL call_6_1_0_5(SH2_struct* sh)
{
  SH2movwp(sh,1,0);
}
#define init_6_1_1_5
void FASTCALL call_6_1_1_5(SH2_struct* sh)
{
  SH2movwp(sh,1,1);
}
#define init_6_1_2_5
void FASTCALL call_6_1_2_5(SH2_struct* sh)
{
  SH2movwp(sh,1,2);
}
#define init_6_1_3_5
void FASTCALL call_6_1_3_5(SH2_struct* sh)
{
  SH2movwp(sh,1,3);
}
#define init_6_1_4_5
void FASTCALL call_6_1_4_5(SH2_struct* sh)
{
  SH2movwp(sh,1,4);
}
#define init_6_1_5_5
void FASTCALL call_6_1_5_5(SH2_struct* sh)
{
  SH2movwp(sh,1,5);
}
#define init_6_1_6_5
void FASTCALL call_6_1_6_5(SH2_struct* sh)
{
  SH2movwp(sh,1,6);
}
#define init_6_1_7_5
void FASTCALL call_6_1_7_5(SH2_struct* sh)
{
  SH2movwp(sh,1,7);
}
#define init_6_1_8_5
void FASTCALL call_6_1_8_5(SH2_struct* sh)
{
  SH2movwp(sh,1,8);
}
#define init_6_1_9_5
void FASTCALL call_6_1_9_5(SH2_struct* sh)
{
  SH2movwp(sh,1,9);
}
#define init_6_1_10_5
void FASTCALL call_6_1_10_5(SH2_struct* sh)
{
  SH2movwp(sh,1,10);
}
#define init_6_1_11_5
void FASTCALL call_6_1_11_5(SH2_struct* sh)
{
  SH2movwp(sh,1,11);
}
#define init_6_1_12_5
void FASTCALL call_6_1_12_5(SH2_struct* sh)
{
  SH2movwp(sh,1,12);
}
#define init_6_1_13_5
void FASTCALL call_6_1_13_5(SH2_struct* sh)
{
  SH2movwp(sh,1,13);
}
#define init_6_1_14_5
void FASTCALL call_6_1_14_5(SH2_struct* sh)
{
  SH2movwp(sh,1,14);
}
#define init_6_1_15_5
void FASTCALL call_6_1_15_5(SH2_struct* sh)
{
  SH2movwp(sh,1,15);
}
#define init_6_2_0_5
void FASTCALL call_6_2_0_5(SH2_struct* sh)
{
  SH2movwp(sh,2,0);
}
#define init_6_2_1_5
void FASTCALL call_6_2_1_5(SH2_struct* sh)
{
  SH2movwp(sh,2,1);
}
#define init_6_2_2_5
void FASTCALL call_6_2_2_5(SH2_struct* sh)
{
  SH2movwp(sh,2,2);
}
#define init_6_2_3_5
void FASTCALL call_6_2_3_5(SH2_struct* sh)
{
  SH2movwp(sh,2,3);
}
#define init_6_2_4_5
void FASTCALL call_6_2_4_5(SH2_struct* sh)
{
  SH2movwp(sh,2,4);
}
#define init_6_2_5_5
void FASTCALL call_6_2_5_5(SH2_struct* sh)
{
  SH2movwp(sh,2,5);
}
#define init_6_2_6_5
void FASTCALL call_6_2_6_5(SH2_struct* sh)
{
  SH2movwp(sh,2,6);
}
#define init_6_2_7_5
void FASTCALL call_6_2_7_5(SH2_struct* sh)
{
  SH2movwp(sh,2,7);
}
#define init_6_2_8_5
void FASTCALL call_6_2_8_5(SH2_struct* sh)
{
  SH2movwp(sh,2,8);
}
#define init_6_2_9_5
void FASTCALL call_6_2_9_5(SH2_struct* sh)
{
  SH2movwp(sh,2,9);
}
#define init_6_2_10_5
void FASTCALL call_6_2_10_5(SH2_struct* sh)
{
  SH2movwp(sh,2,10);
}
#define init_6_2_11_5
void FASTCALL call_6_2_11_5(SH2_struct* sh)
{
  SH2movwp(sh,2,11);
}
#define init_6_2_12_5
void FASTCALL call_6_2_12_5(SH2_struct* sh)
{
  SH2movwp(sh,2,12);
}
#define init_6_2_13_5
void FASTCALL call_6_2_13_5(SH2_struct* sh)
{
  SH2movwp(sh,2,13);
}
#define init_6_2_14_5
void FASTCALL call_6_2_14_5(SH2_struct* sh)
{
  SH2movwp(sh,2,14);
}
#define init_6_2_15_5
void FASTCALL call_6_2_15_5(SH2_struct* sh)
{
  SH2movwp(sh,2,15);
}
#define init_6_3_0_5
void FASTCALL call_6_3_0_5(SH2_struct* sh)
{
  SH2movwp(sh,3,0);
}
#define init_6_3_1_5
void FASTCALL call_6_3_1_5(SH2_struct* sh)
{
  SH2movwp(sh,3,1);
}
#define init_6_3_2_5
void FASTCALL call_6_3_2_5(SH2_struct* sh)
{
  SH2movwp(sh,3,2);
}
#define init_6_3_3_5
void FASTCALL call_6_3_3_5(SH2_struct* sh)
{
  SH2movwp(sh,3,3);
}
#define init_6_3_4_5
void FASTCALL call_6_3_4_5(SH2_struct* sh)
{
  SH2movwp(sh,3,4);
}
#define init_6_3_5_5
void FASTCALL call_6_3_5_5(SH2_struct* sh)
{
  SH2movwp(sh,3,5);
}
#define init_6_3_6_5
void FASTCALL call_6_3_6_5(SH2_struct* sh)
{
  SH2movwp(sh,3,6);
}
#define init_6_3_7_5
void FASTCALL call_6_3_7_5(SH2_struct* sh)
{
  SH2movwp(sh,3,7);
}
#define init_6_3_8_5
void FASTCALL call_6_3_8_5(SH2_struct* sh)
{
  SH2movwp(sh,3,8);
}
#define init_6_3_9_5
void FASTCALL call_6_3_9_5(SH2_struct* sh)
{
  SH2movwp(sh,3,9);
}
#define init_6_3_10_5
void FASTCALL call_6_3_10_5(SH2_struct* sh)
{
  SH2movwp(sh,3,10);
}
#define init_6_3_11_5
void FASTCALL call_6_3_11_5(SH2_struct* sh)
{
  SH2movwp(sh,3,11);
}
#define init_6_3_12_5
void FASTCALL call_6_3_12_5(SH2_struct* sh)
{
  SH2movwp(sh,3,12);
}
#define init_6_3_13_5
void FASTCALL call_6_3_13_5(SH2_struct* sh)
{
  SH2movwp(sh,3,13);
}
#define init_6_3_14_5
void FASTCALL call_6_3_14_5(SH2_struct* sh)
{
  SH2movwp(sh,3,14);
}
#define init_6_3_15_5
void FASTCALL call_6_3_15_5(SH2_struct* sh)
{
  SH2movwp(sh,3,15);
}
#define init_6_4_0_5
void FASTCALL call_6_4_0_5(SH2_struct* sh)
{
  SH2movwp(sh,4,0);
}
#define init_6_4_1_5
void FASTCALL call_6_4_1_5(SH2_struct* sh)
{
  SH2movwp(sh,4,1);
}
#define init_6_4_2_5
void FASTCALL call_6_4_2_5(SH2_struct* sh)
{
  SH2movwp(sh,4,2);
}
#define init_6_4_3_5
void FASTCALL call_6_4_3_5(SH2_struct* sh)
{
  SH2movwp(sh,4,3);
}
#define init_6_4_4_5
void FASTCALL call_6_4_4_5(SH2_struct* sh)
{
  SH2movwp(sh,4,4);
}
#define init_6_4_5_5
void FASTCALL call_6_4_5_5(SH2_struct* sh)
{
  SH2movwp(sh,4,5);
}
#define init_6_4_6_5
void FASTCALL call_6_4_6_5(SH2_struct* sh)
{
  SH2movwp(sh,4,6);
}
#define init_6_4_7_5
void FASTCALL call_6_4_7_5(SH2_struct* sh)
{
  SH2movwp(sh,4,7);
}
#define init_6_4_8_5
void FASTCALL call_6_4_8_5(SH2_struct* sh)
{
  SH2movwp(sh,4,8);
}
#define init_6_4_9_5
void FASTCALL call_6_4_9_5(SH2_struct* sh)
{
  SH2movwp(sh,4,9);
}
#define init_6_4_10_5
void FASTCALL call_6_4_10_5(SH2_struct* sh)
{
  SH2movwp(sh,4,10);
}
#define init_6_4_11_5
void FASTCALL call_6_4_11_5(SH2_struct* sh)
{
  SH2movwp(sh,4,11);
}
#define init_6_4_12_5
void FASTCALL call_6_4_12_5(SH2_struct* sh)
{
  SH2movwp(sh,4,12);
}
#define init_6_4_13_5
void FASTCALL call_6_4_13_5(SH2_struct* sh)
{
  SH2movwp(sh,4,13);
}
#define init_6_4_14_5
void FASTCALL call_6_4_14_5(SH2_struct* sh)
{
  SH2movwp(sh,4,14);
}
#define init_6_4_15_5
void FASTCALL call_6_4_15_5(SH2_struct* sh)
{
  SH2movwp(sh,4,15);
}
#define init_6_5_0_5
void FASTCALL call_6_5_0_5(SH2_struct* sh)
{
  SH2movwp(sh,5,0);
}
#define init_6_5_1_5
void FASTCALL call_6_5_1_5(SH2_struct* sh)
{
  SH2movwp(sh,5,1);
}
#define init_6_5_2_5
void FASTCALL call_6_5_2_5(SH2_struct* sh)
{
  SH2movwp(sh,5,2);
}
#define init_6_5_3_5
void FASTCALL call_6_5_3_5(SH2_struct* sh)
{
  SH2movwp(sh,5,3);
}
#define init_6_5_4_5
void FASTCALL call_6_5_4_5(SH2_struct* sh)
{
  SH2movwp(sh,5,4);
}
#define init_6_5_5_5
void FASTCALL call_6_5_5_5(SH2_struct* sh)
{
  SH2movwp(sh,5,5);
}
#define init_6_5_6_5
void FASTCALL call_6_5_6_5(SH2_struct* sh)
{
  SH2movwp(sh,5,6);
}
#define init_6_5_7_5
void FASTCALL call_6_5_7_5(SH2_struct* sh)
{
  SH2movwp(sh,5,7);
}
#define init_6_5_8_5
void FASTCALL call_6_5_8_5(SH2_struct* sh)
{
  SH2movwp(sh,5,8);
}
#define init_6_5_9_5
void FASTCALL call_6_5_9_5(SH2_struct* sh)
{
  SH2movwp(sh,5,9);
}
#define init_6_5_10_5
void FASTCALL call_6_5_10_5(SH2_struct* sh)
{
  SH2movwp(sh,5,10);
}
#define init_6_5_11_5
void FASTCALL call_6_5_11_5(SH2_struct* sh)
{
  SH2movwp(sh,5,11);
}
#define init_6_5_12_5
void FASTCALL call_6_5_12_5(SH2_struct* sh)
{
  SH2movwp(sh,5,12);
}
#define init_6_5_13_5
void FASTCALL call_6_5_13_5(SH2_struct* sh)
{
  SH2movwp(sh,5,13);
}
#define init_6_5_14_5
void FASTCALL call_6_5_14_5(SH2_struct* sh)
{
  SH2movwp(sh,5,14);
}
#define init_6_5_15_5
void FASTCALL call_6_5_15_5(SH2_struct* sh)
{
  SH2movwp(sh,5,15);
}
#define init_6_6_0_5
void FASTCALL call_6_6_0_5(SH2_struct* sh)
{
  SH2movwp(sh,6,0);
}
#define init_6_6_1_5
void FASTCALL call_6_6_1_5(SH2_struct* sh)
{
  SH2movwp(sh,6,1);
}
#define init_6_6_2_5
void FASTCALL call_6_6_2_5(SH2_struct* sh)
{
  SH2movwp(sh,6,2);
}
#define init_6_6_3_5
void FASTCALL call_6_6_3_5(SH2_struct* sh)
{
  SH2movwp(sh,6,3);
}
#define init_6_6_4_5
void FASTCALL call_6_6_4_5(SH2_struct* sh)
{
  SH2movwp(sh,6,4);
}
#define init_6_6_5_5
void FASTCALL call_6_6_5_5(SH2_struct* sh)
{
  SH2movwp(sh,6,5);
}
#define init_6_6_6_5
void FASTCALL call_6_6_6_5(SH2_struct* sh)
{
  SH2movwp(sh,6,6);
}
#define init_6_6_7_5
void FASTCALL call_6_6_7_5(SH2_struct* sh)
{
  SH2movwp(sh,6,7);
}
#define init_6_6_8_5
void FASTCALL call_6_6_8_5(SH2_struct* sh)
{
  SH2movwp(sh,6,8);
}
#define init_6_6_9_5
void FASTCALL call_6_6_9_5(SH2_struct* sh)
{
  SH2movwp(sh,6,9);
}
#define init_6_6_10_5
void FASTCALL call_6_6_10_5(SH2_struct* sh)
{
  SH2movwp(sh,6,10);
}
#define init_6_6_11_5
void FASTCALL call_6_6_11_5(SH2_struct* sh)
{
  SH2movwp(sh,6,11);
}
#define init_6_6_12_5
void FASTCALL call_6_6_12_5(SH2_struct* sh)
{
  SH2movwp(sh,6,12);
}
#define init_6_6_13_5
void FASTCALL call_6_6_13_5(SH2_struct* sh)
{
  SH2movwp(sh,6,13);
}
#define init_6_6_14_5
void FASTCALL call_6_6_14_5(SH2_struct* sh)
{
  SH2movwp(sh,6,14);
}
#define init_6_6_15_5
void FASTCALL call_6_6_15_5(SH2_struct* sh)
{
  SH2movwp(sh,6,15);
}
#define init_6_7_0_5
void FASTCALL call_6_7_0_5(SH2_struct* sh)
{
  SH2movwp(sh,7,0);
}
#define init_6_7_1_5
void FASTCALL call_6_7_1_5(SH2_struct* sh)
{
  SH2movwp(sh,7,1);
}
#define init_6_7_2_5
void FASTCALL call_6_7_2_5(SH2_struct* sh)
{
  SH2movwp(sh,7,2);
}
#define init_6_7_3_5
void FASTCALL call_6_7_3_5(SH2_struct* sh)
{
  SH2movwp(sh,7,3);
}
#define init_6_7_4_5
void FASTCALL call_6_7_4_5(SH2_struct* sh)
{
  SH2movwp(sh,7,4);
}
#define init_6_7_5_5
void FASTCALL call_6_7_5_5(SH2_struct* sh)
{
  SH2movwp(sh,7,5);
}
#define init_6_7_6_5
void FASTCALL call_6_7_6_5(SH2_struct* sh)
{
  SH2movwp(sh,7,6);
}
#define init_6_7_7_5
void FASTCALL call_6_7_7_5(SH2_struct* sh)
{
  SH2movwp(sh,7,7);
}
#define init_6_7_8_5
void FASTCALL call_6_7_8_5(SH2_struct* sh)
{
  SH2movwp(sh,7,8);
}
#define init_6_7_9_5
void FASTCALL call_6_7_9_5(SH2_struct* sh)
{
  SH2movwp(sh,7,9);
}
#define init_6_7_10_5
void FASTCALL call_6_7_10_5(SH2_struct* sh)
{
  SH2movwp(sh,7,10);
}
#define init_6_7_11_5
void FASTCALL call_6_7_11_5(SH2_struct* sh)
{
  SH2movwp(sh,7,11);
}
#define init_6_7_12_5
void FASTCALL call_6_7_12_5(SH2_struct* sh)
{
  SH2movwp(sh,7,12);
}
#define init_6_7_13_5
void FASTCALL call_6_7_13_5(SH2_struct* sh)
{
  SH2movwp(sh,7,13);
}
#define init_6_7_14_5
void FASTCALL call_6_7_14_5(SH2_struct* sh)
{
  SH2movwp(sh,7,14);
}
#define init_6_7_15_5
void FASTCALL call_6_7_15_5(SH2_struct* sh)
{
  SH2movwp(sh,7,15);
}
#define init_6_8_0_5
void FASTCALL call_6_8_0_5(SH2_struct* sh)
{
  SH2movwp(sh,8,0);
}
#define init_6_8_1_5
void FASTCALL call_6_8_1_5(SH2_struct* sh)
{
  SH2movwp(sh,8,1);
}
#define init_6_8_2_5
void FASTCALL call_6_8_2_5(SH2_struct* sh)
{
  SH2movwp(sh,8,2);
}
#define init_6_8_3_5
void FASTCALL call_6_8_3_5(SH2_struct* sh)
{
  SH2movwp(sh,8,3);
}
#define init_6_8_4_5
void FASTCALL call_6_8_4_5(SH2_struct* sh)
{
  SH2movwp(sh,8,4);
}
#define init_6_8_5_5
void FASTCALL call_6_8_5_5(SH2_struct* sh)
{
  SH2movwp(sh,8,5);
}
#define init_6_8_6_5
void FASTCALL call_6_8_6_5(SH2_struct* sh)
{
  SH2movwp(sh,8,6);
}
#define init_6_8_7_5
void FASTCALL call_6_8_7_5(SH2_struct* sh)
{
  SH2movwp(sh,8,7);
}
#define init_6_8_8_5
void FASTCALL call_6_8_8_5(SH2_struct* sh)
{
  SH2movwp(sh,8,8);
}
#define init_6_8_9_5
void FASTCALL call_6_8_9_5(SH2_struct* sh)
{
  SH2movwp(sh,8,9);
}
#define init_6_8_10_5
void FASTCALL call_6_8_10_5(SH2_struct* sh)
{
  SH2movwp(sh,8,10);
}
#define init_6_8_11_5
void FASTCALL call_6_8_11_5(SH2_struct* sh)
{
  SH2movwp(sh,8,11);
}
#define init_6_8_12_5
void FASTCALL call_6_8_12_5(SH2_struct* sh)
{
  SH2movwp(sh,8,12);
}
#define init_6_8_13_5
void FASTCALL call_6_8_13_5(SH2_struct* sh)
{
  SH2movwp(sh,8,13);
}
#define init_6_8_14_5
void FASTCALL call_6_8_14_5(SH2_struct* sh)
{
  SH2movwp(sh,8,14);
}
#define init_6_8_15_5
void FASTCALL call_6_8_15_5(SH2_struct* sh)
{
  SH2movwp(sh,8,15);
}
#define init_6_9_0_5
void FASTCALL call_6_9_0_5(SH2_struct* sh)
{
  SH2movwp(sh,9,0);
}
#define init_6_9_1_5
void FASTCALL call_6_9_1_5(SH2_struct* sh)
{
  SH2movwp(sh,9,1);
}
#define init_6_9_2_5
void FASTCALL call_6_9_2_5(SH2_struct* sh)
{
  SH2movwp(sh,9,2);
}
#define init_6_9_3_5
void FASTCALL call_6_9_3_5(SH2_struct* sh)
{
  SH2movwp(sh,9,3);
}
#define init_6_9_4_5
void FASTCALL call_6_9_4_5(SH2_struct* sh)
{
  SH2movwp(sh,9,4);
}
#define init_6_9_5_5
void FASTCALL call_6_9_5_5(SH2_struct* sh)
{
  SH2movwp(sh,9,5);
}
#define init_6_9_6_5
void FASTCALL call_6_9_6_5(SH2_struct* sh)
{
  SH2movwp(sh,9,6);
}
#define init_6_9_7_5
void FASTCALL call_6_9_7_5(SH2_struct* sh)
{
  SH2movwp(sh,9,7);
}
#define init_6_9_8_5
void FASTCALL call_6_9_8_5(SH2_struct* sh)
{
  SH2movwp(sh,9,8);
}
#define init_6_9_9_5
void FASTCALL call_6_9_9_5(SH2_struct* sh)
{
  SH2movwp(sh,9,9);
}
#define init_6_9_10_5
void FASTCALL call_6_9_10_5(SH2_struct* sh)
{
  SH2movwp(sh,9,10);
}
#define init_6_9_11_5
void FASTCALL call_6_9_11_5(SH2_struct* sh)
{
  SH2movwp(sh,9,11);
}
#define init_6_9_12_5
void FASTCALL call_6_9_12_5(SH2_struct* sh)
{
  SH2movwp(sh,9,12);
}
#define init_6_9_13_5
void FASTCALL call_6_9_13_5(SH2_struct* sh)
{
  SH2movwp(sh,9,13);
}
#define init_6_9_14_5
void FASTCALL call_6_9_14_5(SH2_struct* sh)
{
  SH2movwp(sh,9,14);
}
#define init_6_9_15_5
void FASTCALL call_6_9_15_5(SH2_struct* sh)
{
  SH2movwp(sh,9,15);
}
#define init_6_10_0_5
void FASTCALL call_6_10_0_5(SH2_struct* sh)
{
  SH2movwp(sh,10,0);
}
#define init_6_10_1_5
void FASTCALL call_6_10_1_5(SH2_struct* sh)
{
  SH2movwp(sh,10,1);
}
#define init_6_10_2_5
void FASTCALL call_6_10_2_5(SH2_struct* sh)
{
  SH2movwp(sh,10,2);
}
#define init_6_10_3_5
void FASTCALL call_6_10_3_5(SH2_struct* sh)
{
  SH2movwp(sh,10,3);
}
#define init_6_10_4_5
void FASTCALL call_6_10_4_5(SH2_struct* sh)
{
  SH2movwp(sh,10,4);
}
#define init_6_10_5_5
void FASTCALL call_6_10_5_5(SH2_struct* sh)
{
  SH2movwp(sh,10,5);
}
#define init_6_10_6_5
void FASTCALL call_6_10_6_5(SH2_struct* sh)
{
  SH2movwp(sh,10,6);
}
#define init_6_10_7_5
void FASTCALL call_6_10_7_5(SH2_struct* sh)
{
  SH2movwp(sh,10,7);
}
#define init_6_10_8_5
void FASTCALL call_6_10_8_5(SH2_struct* sh)
{
  SH2movwp(sh,10,8);
}
#define init_6_10_9_5
void FASTCALL call_6_10_9_5(SH2_struct* sh)
{
  SH2movwp(sh,10,9);
}
#define init_6_10_10_5
void FASTCALL call_6_10_10_5(SH2_struct* sh)
{
  SH2movwp(sh,10,10);
}
#define init_6_10_11_5
void FASTCALL call_6_10_11_5(SH2_struct* sh)
{
  SH2movwp(sh,10,11);
}
#define init_6_10_12_5
void FASTCALL call_6_10_12_5(SH2_struct* sh)
{
  SH2movwp(sh,10,12);
}
#define init_6_10_13_5
void FASTCALL call_6_10_13_5(SH2_struct* sh)
{
  SH2movwp(sh,10,13);
}
#define init_6_10_14_5
void FASTCALL call_6_10_14_5(SH2_struct* sh)
{
  SH2movwp(sh,10,14);
}
#define init_6_10_15_5
void FASTCALL call_6_10_15_5(SH2_struct* sh)
{
  SH2movwp(sh,10,15);
}
#define init_6_11_0_5
void FASTCALL call_6_11_0_5(SH2_struct* sh)
{
  SH2movwp(sh,11,0);
}
#define init_6_11_1_5
void FASTCALL call_6_11_1_5(SH2_struct* sh)
{
  SH2movwp(sh,11,1);
}
#define init_6_11_2_5
void FASTCALL call_6_11_2_5(SH2_struct* sh)
{
  SH2movwp(sh,11,2);
}
#define init_6_11_3_5
void FASTCALL call_6_11_3_5(SH2_struct* sh)
{
  SH2movwp(sh,11,3);
}
#define init_6_11_4_5
void FASTCALL call_6_11_4_5(SH2_struct* sh)
{
  SH2movwp(sh,11,4);
}
#define init_6_11_5_5
void FASTCALL call_6_11_5_5(SH2_struct* sh)
{
  SH2movwp(sh,11,5);
}
#define init_6_11_6_5
void FASTCALL call_6_11_6_5(SH2_struct* sh)
{
  SH2movwp(sh,11,6);
}
#define init_6_11_7_5
void FASTCALL call_6_11_7_5(SH2_struct* sh)
{
  SH2movwp(sh,11,7);
}
#define init_6_11_8_5
void FASTCALL call_6_11_8_5(SH2_struct* sh)
{
  SH2movwp(sh,11,8);
}
#define init_6_11_9_5
void FASTCALL call_6_11_9_5(SH2_struct* sh)
{
  SH2movwp(sh,11,9);
}
#define init_6_11_10_5
void FASTCALL call_6_11_10_5(SH2_struct* sh)
{
  SH2movwp(sh,11,10);
}
#define init_6_11_11_5
void FASTCALL call_6_11_11_5(SH2_struct* sh)
{
  SH2movwp(sh,11,11);
}
#define init_6_11_12_5
void FASTCALL call_6_11_12_5(SH2_struct* sh)
{
  SH2movwp(sh,11,12);
}
#define init_6_11_13_5
void FASTCALL call_6_11_13_5(SH2_struct* sh)
{
  SH2movwp(sh,11,13);
}
#define init_6_11_14_5
void FASTCALL call_6_11_14_5(SH2_struct* sh)
{
  SH2movwp(sh,11,14);
}
#define init_6_11_15_5
void FASTCALL call_6_11_15_5(SH2_struct* sh)
{
  SH2movwp(sh,11,15);
}
#define init_6_12_0_5
void FASTCALL call_6_12_0_5(SH2_struct* sh)
{
  SH2movwp(sh,12,0);
}
#define init_6_12_1_5
void FASTCALL call_6_12_1_5(SH2_struct* sh)
{
  SH2movwp(sh,12,1);
}
#define init_6_12_2_5
void FASTCALL call_6_12_2_5(SH2_struct* sh)
{
  SH2movwp(sh,12,2);
}
#define init_6_12_3_5
void FASTCALL call_6_12_3_5(SH2_struct* sh)
{
  SH2movwp(sh,12,3);
}
#define init_6_12_4_5
void FASTCALL call_6_12_4_5(SH2_struct* sh)
{
  SH2movwp(sh,12,4);
}
#define init_6_12_5_5
void FASTCALL call_6_12_5_5(SH2_struct* sh)
{
  SH2movwp(sh,12,5);
}
#define init_6_12_6_5
void FASTCALL call_6_12_6_5(SH2_struct* sh)
{
  SH2movwp(sh,12,6);
}
#define init_6_12_7_5
void FASTCALL call_6_12_7_5(SH2_struct* sh)
{
  SH2movwp(sh,12,7);
}
#define init_6_12_8_5
void FASTCALL call_6_12_8_5(SH2_struct* sh)
{
  SH2movwp(sh,12,8);
}
#define init_6_12_9_5
void FASTCALL call_6_12_9_5(SH2_struct* sh)
{
  SH2movwp(sh,12,9);
}
#define init_6_12_10_5
void FASTCALL call_6_12_10_5(SH2_struct* sh)
{
  SH2movwp(sh,12,10);
}
#define init_6_12_11_5
void FASTCALL call_6_12_11_5(SH2_struct* sh)
{
  SH2movwp(sh,12,11);
}
#define init_6_12_12_5
void FASTCALL call_6_12_12_5(SH2_struct* sh)
{
  SH2movwp(sh,12,12);
}
#define init_6_12_13_5
void FASTCALL call_6_12_13_5(SH2_struct* sh)
{
  SH2movwp(sh,12,13);
}
#define init_6_12_14_5
void FASTCALL call_6_12_14_5(SH2_struct* sh)
{
  SH2movwp(sh,12,14);
}
#define init_6_12_15_5
void FASTCALL call_6_12_15_5(SH2_struct* sh)
{
  SH2movwp(sh,12,15);
}
#define init_6_13_0_5
void FASTCALL call_6_13_0_5(SH2_struct* sh)
{
  SH2movwp(sh,13,0);
}
#define init_6_13_1_5
void FASTCALL call_6_13_1_5(SH2_struct* sh)
{
  SH2movwp(sh,13,1);
}
#define init_6_13_2_5
void FASTCALL call_6_13_2_5(SH2_struct* sh)
{
  SH2movwp(sh,13,2);
}
#define init_6_13_3_5
void FASTCALL call_6_13_3_5(SH2_struct* sh)
{
  SH2movwp(sh,13,3);
}
#define init_6_13_4_5
void FASTCALL call_6_13_4_5(SH2_struct* sh)
{
  SH2movwp(sh,13,4);
}
#define init_6_13_5_5
void FASTCALL call_6_13_5_5(SH2_struct* sh)
{
  SH2movwp(sh,13,5);
}
#define init_6_13_6_5
void FASTCALL call_6_13_6_5(SH2_struct* sh)
{
  SH2movwp(sh,13,6);
}
#define init_6_13_7_5
void FASTCALL call_6_13_7_5(SH2_struct* sh)
{
  SH2movwp(sh,13,7);
}
#define init_6_13_8_5
void FASTCALL call_6_13_8_5(SH2_struct* sh)
{
  SH2movwp(sh,13,8);
}
#define init_6_13_9_5
void FASTCALL call_6_13_9_5(SH2_struct* sh)
{
  SH2movwp(sh,13,9);
}
#define init_6_13_10_5
void FASTCALL call_6_13_10_5(SH2_struct* sh)
{
  SH2movwp(sh,13,10);
}
#define init_6_13_11_5
void FASTCALL call_6_13_11_5(SH2_struct* sh)
{
  SH2movwp(sh,13,11);
}
#define init_6_13_12_5
void FASTCALL call_6_13_12_5(SH2_struct* sh)
{
  SH2movwp(sh,13,12);
}
#define init_6_13_13_5
void FASTCALL call_6_13_13_5(SH2_struct* sh)
{
  SH2movwp(sh,13,13);
}
#define init_6_13_14_5
void FASTCALL call_6_13_14_5(SH2_struct* sh)
{
  SH2movwp(sh,13,14);
}
#define init_6_13_15_5
void FASTCALL call_6_13_15_5(SH2_struct* sh)
{
  SH2movwp(sh,13,15);
}
#define init_6_14_0_5
void FASTCALL call_6_14_0_5(SH2_struct* sh)
{
  SH2movwp(sh,14,0);
}
#define init_6_14_1_5
void FASTCALL call_6_14_1_5(SH2_struct* sh)
{
  SH2movwp(sh,14,1);
}
#define init_6_14_2_5
void FASTCALL call_6_14_2_5(SH2_struct* sh)
{
  SH2movwp(sh,14,2);
}
#define init_6_14_3_5
void FASTCALL call_6_14_3_5(SH2_struct* sh)
{
  SH2movwp(sh,14,3);
}
#define init_6_14_4_5
void FASTCALL call_6_14_4_5(SH2_struct* sh)
{
  SH2movwp(sh,14,4);
}
#define init_6_14_5_5
void FASTCALL call_6_14_5_5(SH2_struct* sh)
{
  SH2movwp(sh,14,5);
}
#define init_6_14_6_5
void FASTCALL call_6_14_6_5(SH2_struct* sh)
{
  SH2movwp(sh,14,6);
}
#define init_6_14_7_5
void FASTCALL call_6_14_7_5(SH2_struct* sh)
{
  SH2movwp(sh,14,7);
}
#define init_6_14_8_5
void FASTCALL call_6_14_8_5(SH2_struct* sh)
{
  SH2movwp(sh,14,8);
}
#define init_6_14_9_5
void FASTCALL call_6_14_9_5(SH2_struct* sh)
{
  SH2movwp(sh,14,9);
}
#define init_6_14_10_5
void FASTCALL call_6_14_10_5(SH2_struct* sh)
{
  SH2movwp(sh,14,10);
}
#define init_6_14_11_5
void FASTCALL call_6_14_11_5(SH2_struct* sh)
{
  SH2movwp(sh,14,11);
}
#define init_6_14_12_5
void FASTCALL call_6_14_12_5(SH2_struct* sh)
{
  SH2movwp(sh,14,12);
}
#define init_6_14_13_5
void FASTCALL call_6_14_13_5(SH2_struct* sh)
{
  SH2movwp(sh,14,13);
}
#define init_6_14_14_5
void FASTCALL call_6_14_14_5(SH2_struct* sh)
{
  SH2movwp(sh,14,14);
}
#define init_6_14_15_5
void FASTCALL call_6_14_15_5(SH2_struct* sh)
{
  SH2movwp(sh,14,15);
}
#define init_6_15_0_5
void FASTCALL call_6_15_0_5(SH2_struct* sh)
{
  SH2movwp(sh,15,0);
}
#define init_6_15_1_5
void FASTCALL call_6_15_1_5(SH2_struct* sh)
{
  SH2movwp(sh,15,1);
}
#define init_6_15_2_5
void FASTCALL call_6_15_2_5(SH2_struct* sh)
{
  SH2movwp(sh,15,2);
}
#define init_6_15_3_5
void FASTCALL call_6_15_3_5(SH2_struct* sh)
{
  SH2movwp(sh,15,3);
}
#define init_6_15_4_5
void FASTCALL call_6_15_4_5(SH2_struct* sh)
{
  SH2movwp(sh,15,4);
}
#define init_6_15_5_5
void FASTCALL call_6_15_5_5(SH2_struct* sh)
{
  SH2movwp(sh,15,5);
}
#define init_6_15_6_5
void FASTCALL call_6_15_6_5(SH2_struct* sh)
{
  SH2movwp(sh,15,6);
}
#define init_6_15_7_5
void FASTCALL call_6_15_7_5(SH2_struct* sh)
{
  SH2movwp(sh,15,7);
}
#define init_6_15_8_5
void FASTCALL call_6_15_8_5(SH2_struct* sh)
{
  SH2movwp(sh,15,8);
}
#define init_6_15_9_5
void FASTCALL call_6_15_9_5(SH2_struct* sh)
{
  SH2movwp(sh,15,9);
}
#define init_6_15_10_5
void FASTCALL call_6_15_10_5(SH2_struct* sh)
{
  SH2movwp(sh,15,10);
}
#define init_6_15_11_5
void FASTCALL call_6_15_11_5(SH2_struct* sh)
{
  SH2movwp(sh,15,11);
}
#define init_6_15_12_5
void FASTCALL call_6_15_12_5(SH2_struct* sh)
{
  SH2movwp(sh,15,12);
}
#define init_6_15_13_5
void FASTCALL call_6_15_13_5(SH2_struct* sh)
{
  SH2movwp(sh,15,13);
}
#define init_6_15_14_5
void FASTCALL call_6_15_14_5(SH2_struct* sh)
{
  SH2movwp(sh,15,14);
}
#define init_6_15_15_5
void FASTCALL call_6_15_15_5(SH2_struct* sh)
{
  SH2movwp(sh,15,15);
}
#define init_2_0_0_5
void FASTCALL call_2_0_0_5(SH2_struct* sh)
{
  SH2movwm(sh,0,0);
}
#define init_2_0_1_5
void FASTCALL call_2_0_1_5(SH2_struct* sh)
{
  SH2movwm(sh,0,1);
}
#define init_2_0_2_5
void FASTCALL call_2_0_2_5(SH2_struct* sh)
{
  SH2movwm(sh,0,2);
}
#define init_2_0_3_5
void FASTCALL call_2_0_3_5(SH2_struct* sh)
{
  SH2movwm(sh,0,3);
}
#define init_2_0_4_5
void FASTCALL call_2_0_4_5(SH2_struct* sh)
{
  SH2movwm(sh,0,4);
}
#define init_2_0_5_5
void FASTCALL call_2_0_5_5(SH2_struct* sh)
{
  SH2movwm(sh,0,5);
}
#define init_2_0_6_5
void FASTCALL call_2_0_6_5(SH2_struct* sh)
{
  SH2movwm(sh,0,6);
}
#define init_2_0_7_5
void FASTCALL call_2_0_7_5(SH2_struct* sh)
{
  SH2movwm(sh,0,7);
}
#define init_2_0_8_5
void FASTCALL call_2_0_8_5(SH2_struct* sh)
{
  SH2movwm(sh,0,8);
}
#define init_2_0_9_5
void FASTCALL call_2_0_9_5(SH2_struct* sh)
{
  SH2movwm(sh,0,9);
}
#define init_2_0_10_5
void FASTCALL call_2_0_10_5(SH2_struct* sh)
{
  SH2movwm(sh,0,10);
}
#define init_2_0_11_5
void FASTCALL call_2_0_11_5(SH2_struct* sh)
{
  SH2movwm(sh,0,11);
}
#define init_2_0_12_5
void FASTCALL call_2_0_12_5(SH2_struct* sh)
{
  SH2movwm(sh,0,12);
}
#define init_2_0_13_5
void FASTCALL call_2_0_13_5(SH2_struct* sh)
{
  SH2movwm(sh,0,13);
}
#define init_2_0_14_5
void FASTCALL call_2_0_14_5(SH2_struct* sh)
{
  SH2movwm(sh,0,14);
}
#define init_2_0_15_5
void FASTCALL call_2_0_15_5(SH2_struct* sh)
{
  SH2movwm(sh,0,15);
}
#define init_2_1_0_5
void FASTCALL call_2_1_0_5(SH2_struct* sh)
{
  SH2movwm(sh,1,0);
}
#define init_2_1_1_5
void FASTCALL call_2_1_1_5(SH2_struct* sh)
{
  SH2movwm(sh,1,1);
}
#define init_2_1_2_5
void FASTCALL call_2_1_2_5(SH2_struct* sh)
{
  SH2movwm(sh,1,2);
}
#define init_2_1_3_5
void FASTCALL call_2_1_3_5(SH2_struct* sh)
{
  SH2movwm(sh,1,3);
}
#define init_2_1_4_5
void FASTCALL call_2_1_4_5(SH2_struct* sh)
{
  SH2movwm(sh,1,4);
}
#define init_2_1_5_5
void FASTCALL call_2_1_5_5(SH2_struct* sh)
{
  SH2movwm(sh,1,5);
}
#define init_2_1_6_5
void FASTCALL call_2_1_6_5(SH2_struct* sh)
{
  SH2movwm(sh,1,6);
}
#define init_2_1_7_5
void FASTCALL call_2_1_7_5(SH2_struct* sh)
{
  SH2movwm(sh,1,7);
}
#define init_2_1_8_5
void FASTCALL call_2_1_8_5(SH2_struct* sh)
{
  SH2movwm(sh,1,8);
}
#define init_2_1_9_5
void FASTCALL call_2_1_9_5(SH2_struct* sh)
{
  SH2movwm(sh,1,9);
}
#define init_2_1_10_5
void FASTCALL call_2_1_10_5(SH2_struct* sh)
{
  SH2movwm(sh,1,10);
}
#define init_2_1_11_5
void FASTCALL call_2_1_11_5(SH2_struct* sh)
{
  SH2movwm(sh,1,11);
}
#define init_2_1_12_5
void FASTCALL call_2_1_12_5(SH2_struct* sh)
{
  SH2movwm(sh,1,12);
}
#define init_2_1_13_5
void FASTCALL call_2_1_13_5(SH2_struct* sh)
{
  SH2movwm(sh,1,13);
}
#define init_2_1_14_5
void FASTCALL call_2_1_14_5(SH2_struct* sh)
{
  SH2movwm(sh,1,14);
}
#define init_2_1_15_5
void FASTCALL call_2_1_15_5(SH2_struct* sh)
{
  SH2movwm(sh,1,15);
}
#define init_2_2_0_5
void FASTCALL call_2_2_0_5(SH2_struct* sh)
{
  SH2movwm(sh,2,0);
}
#define init_2_2_1_5
void FASTCALL call_2_2_1_5(SH2_struct* sh)
{
  SH2movwm(sh,2,1);
}
#define init_2_2_2_5
void FASTCALL call_2_2_2_5(SH2_struct* sh)
{
  SH2movwm(sh,2,2);
}
#define init_2_2_3_5
void FASTCALL call_2_2_3_5(SH2_struct* sh)
{
  SH2movwm(sh,2,3);
}
#define init_2_2_4_5
void FASTCALL call_2_2_4_5(SH2_struct* sh)
{
  SH2movwm(sh,2,4);
}
#define init_2_2_5_5
void FASTCALL call_2_2_5_5(SH2_struct* sh)
{
  SH2movwm(sh,2,5);
}
#define init_2_2_6_5
void FASTCALL call_2_2_6_5(SH2_struct* sh)
{
  SH2movwm(sh,2,6);
}
#define init_2_2_7_5
void FASTCALL call_2_2_7_5(SH2_struct* sh)
{
  SH2movwm(sh,2,7);
}
#define init_2_2_8_5
void FASTCALL call_2_2_8_5(SH2_struct* sh)
{
  SH2movwm(sh,2,8);
}
#define init_2_2_9_5
void FASTCALL call_2_2_9_5(SH2_struct* sh)
{
  SH2movwm(sh,2,9);
}
#define init_2_2_10_5
void FASTCALL call_2_2_10_5(SH2_struct* sh)
{
  SH2movwm(sh,2,10);
}
#define init_2_2_11_5
void FASTCALL call_2_2_11_5(SH2_struct* sh)
{
  SH2movwm(sh,2,11);
}
#define init_2_2_12_5
void FASTCALL call_2_2_12_5(SH2_struct* sh)
{
  SH2movwm(sh,2,12);
}
#define init_2_2_13_5
void FASTCALL call_2_2_13_5(SH2_struct* sh)
{
  SH2movwm(sh,2,13);
}
#define init_2_2_14_5
void FASTCALL call_2_2_14_5(SH2_struct* sh)
{
  SH2movwm(sh,2,14);
}
#define init_2_2_15_5
void FASTCALL call_2_2_15_5(SH2_struct* sh)
{
  SH2movwm(sh,2,15);
}
#define init_2_3_0_5
void FASTCALL call_2_3_0_5(SH2_struct* sh)
{
  SH2movwm(sh,3,0);
}
#define init_2_3_1_5
void FASTCALL call_2_3_1_5(SH2_struct* sh)
{
  SH2movwm(sh,3,1);
}
#define init_2_3_2_5
void FASTCALL call_2_3_2_5(SH2_struct* sh)
{
  SH2movwm(sh,3,2);
}
#define init_2_3_3_5
void FASTCALL call_2_3_3_5(SH2_struct* sh)
{
  SH2movwm(sh,3,3);
}
#define init_2_3_4_5
void FASTCALL call_2_3_4_5(SH2_struct* sh)
{
  SH2movwm(sh,3,4);
}
#define init_2_3_5_5
void FASTCALL call_2_3_5_5(SH2_struct* sh)
{
  SH2movwm(sh,3,5);
}
#define init_2_3_6_5
void FASTCALL call_2_3_6_5(SH2_struct* sh)
{
  SH2movwm(sh,3,6);
}
#define init_2_3_7_5
void FASTCALL call_2_3_7_5(SH2_struct* sh)
{
  SH2movwm(sh,3,7);
}
#define init_2_3_8_5
void FASTCALL call_2_3_8_5(SH2_struct* sh)
{
  SH2movwm(sh,3,8);
}
#define init_2_3_9_5
void FASTCALL call_2_3_9_5(SH2_struct* sh)
{
  SH2movwm(sh,3,9);
}
#define init_2_3_10_5
void FASTCALL call_2_3_10_5(SH2_struct* sh)
{
  SH2movwm(sh,3,10);
}
#define init_2_3_11_5
void FASTCALL call_2_3_11_5(SH2_struct* sh)
{
  SH2movwm(sh,3,11);
}
#define init_2_3_12_5
void FASTCALL call_2_3_12_5(SH2_struct* sh)
{
  SH2movwm(sh,3,12);
}
#define init_2_3_13_5
void FASTCALL call_2_3_13_5(SH2_struct* sh)
{
  SH2movwm(sh,3,13);
}
#define init_2_3_14_5
void FASTCALL call_2_3_14_5(SH2_struct* sh)
{
  SH2movwm(sh,3,14);
}
#define init_2_3_15_5
void FASTCALL call_2_3_15_5(SH2_struct* sh)
{
  SH2movwm(sh,3,15);
}
#define init_2_4_0_5
void FASTCALL call_2_4_0_5(SH2_struct* sh)
{
  SH2movwm(sh,4,0);
}
#define init_2_4_1_5
void FASTCALL call_2_4_1_5(SH2_struct* sh)
{
  SH2movwm(sh,4,1);
}
#define init_2_4_2_5
void FASTCALL call_2_4_2_5(SH2_struct* sh)
{
  SH2movwm(sh,4,2);
}
#define init_2_4_3_5
void FASTCALL call_2_4_3_5(SH2_struct* sh)
{
  SH2movwm(sh,4,3);
}
#define init_2_4_4_5
void FASTCALL call_2_4_4_5(SH2_struct* sh)
{
  SH2movwm(sh,4,4);
}
#define init_2_4_5_5
void FASTCALL call_2_4_5_5(SH2_struct* sh)
{
  SH2movwm(sh,4,5);
}
#define init_2_4_6_5
void FASTCALL call_2_4_6_5(SH2_struct* sh)
{
  SH2movwm(sh,4,6);
}
#define init_2_4_7_5
void FASTCALL call_2_4_7_5(SH2_struct* sh)
{
  SH2movwm(sh,4,7);
}
#define init_2_4_8_5
void FASTCALL call_2_4_8_5(SH2_struct* sh)
{
  SH2movwm(sh,4,8);
}
#define init_2_4_9_5
void FASTCALL call_2_4_9_5(SH2_struct* sh)
{
  SH2movwm(sh,4,9);
}
#define init_2_4_10_5
void FASTCALL call_2_4_10_5(SH2_struct* sh)
{
  SH2movwm(sh,4,10);
}
#define init_2_4_11_5
void FASTCALL call_2_4_11_5(SH2_struct* sh)
{
  SH2movwm(sh,4,11);
}
#define init_2_4_12_5
void FASTCALL call_2_4_12_5(SH2_struct* sh)
{
  SH2movwm(sh,4,12);
}
#define init_2_4_13_5
void FASTCALL call_2_4_13_5(SH2_struct* sh)
{
  SH2movwm(sh,4,13);
}
#define init_2_4_14_5
void FASTCALL call_2_4_14_5(SH2_struct* sh)
{
  SH2movwm(sh,4,14);
}
#define init_2_4_15_5
void FASTCALL call_2_4_15_5(SH2_struct* sh)
{
  SH2movwm(sh,4,15);
}
#define init_2_5_0_5
void FASTCALL call_2_5_0_5(SH2_struct* sh)
{
  SH2movwm(sh,5,0);
}
#define init_2_5_1_5
void FASTCALL call_2_5_1_5(SH2_struct* sh)
{
  SH2movwm(sh,5,1);
}
#define init_2_5_2_5
void FASTCALL call_2_5_2_5(SH2_struct* sh)
{
  SH2movwm(sh,5,2);
}
#define init_2_5_3_5
void FASTCALL call_2_5_3_5(SH2_struct* sh)
{
  SH2movwm(sh,5,3);
}
#define init_2_5_4_5
void FASTCALL call_2_5_4_5(SH2_struct* sh)
{
  SH2movwm(sh,5,4);
}
#define init_2_5_5_5
void FASTCALL call_2_5_5_5(SH2_struct* sh)
{
  SH2movwm(sh,5,5);
}
#define init_2_5_6_5
void FASTCALL call_2_5_6_5(SH2_struct* sh)
{
  SH2movwm(sh,5,6);
}
#define init_2_5_7_5
void FASTCALL call_2_5_7_5(SH2_struct* sh)
{
  SH2movwm(sh,5,7);
}
#define init_2_5_8_5
void FASTCALL call_2_5_8_5(SH2_struct* sh)
{
  SH2movwm(sh,5,8);
}
#define init_2_5_9_5
void FASTCALL call_2_5_9_5(SH2_struct* sh)
{
  SH2movwm(sh,5,9);
}
#define init_2_5_10_5
void FASTCALL call_2_5_10_5(SH2_struct* sh)
{
  SH2movwm(sh,5,10);
}
#define init_2_5_11_5
void FASTCALL call_2_5_11_5(SH2_struct* sh)
{
  SH2movwm(sh,5,11);
}
#define init_2_5_12_5
void FASTCALL call_2_5_12_5(SH2_struct* sh)
{
  SH2movwm(sh,5,12);
}
#define init_2_5_13_5
void FASTCALL call_2_5_13_5(SH2_struct* sh)
{
  SH2movwm(sh,5,13);
}
#define init_2_5_14_5
void FASTCALL call_2_5_14_5(SH2_struct* sh)
{
  SH2movwm(sh,5,14);
}
#define init_2_5_15_5
void FASTCALL call_2_5_15_5(SH2_struct* sh)
{
  SH2movwm(sh,5,15);
}
#define init_2_6_0_5
void FASTCALL call_2_6_0_5(SH2_struct* sh)
{
  SH2movwm(sh,6,0);
}
#define init_2_6_1_5
void FASTCALL call_2_6_1_5(SH2_struct* sh)
{
  SH2movwm(sh,6,1);
}
#define init_2_6_2_5
void FASTCALL call_2_6_2_5(SH2_struct* sh)
{
  SH2movwm(sh,6,2);
}
#define init_2_6_3_5
void FASTCALL call_2_6_3_5(SH2_struct* sh)
{
  SH2movwm(sh,6,3);
}
#define init_2_6_4_5
void FASTCALL call_2_6_4_5(SH2_struct* sh)
{
  SH2movwm(sh,6,4);
}
#define init_2_6_5_5
void FASTCALL call_2_6_5_5(SH2_struct* sh)
{
  SH2movwm(sh,6,5);
}
#define init_2_6_6_5
void FASTCALL call_2_6_6_5(SH2_struct* sh)
{
  SH2movwm(sh,6,6);
}
#define init_2_6_7_5
void FASTCALL call_2_6_7_5(SH2_struct* sh)
{
  SH2movwm(sh,6,7);
}
#define init_2_6_8_5
void FASTCALL call_2_6_8_5(SH2_struct* sh)
{
  SH2movwm(sh,6,8);
}
#define init_2_6_9_5
void FASTCALL call_2_6_9_5(SH2_struct* sh)
{
  SH2movwm(sh,6,9);
}
#define init_2_6_10_5
void FASTCALL call_2_6_10_5(SH2_struct* sh)
{
  SH2movwm(sh,6,10);
}
#define init_2_6_11_5
void FASTCALL call_2_6_11_5(SH2_struct* sh)
{
  SH2movwm(sh,6,11);
}
#define init_2_6_12_5
void FASTCALL call_2_6_12_5(SH2_struct* sh)
{
  SH2movwm(sh,6,12);
}
#define init_2_6_13_5
void FASTCALL call_2_6_13_5(SH2_struct* sh)
{
  SH2movwm(sh,6,13);
}
#define init_2_6_14_5
void FASTCALL call_2_6_14_5(SH2_struct* sh)
{
  SH2movwm(sh,6,14);
}
#define init_2_6_15_5
void FASTCALL call_2_6_15_5(SH2_struct* sh)
{
  SH2movwm(sh,6,15);
}
#define init_2_7_0_5
void FASTCALL call_2_7_0_5(SH2_struct* sh)
{
  SH2movwm(sh,7,0);
}
#define init_2_7_1_5
void FASTCALL call_2_7_1_5(SH2_struct* sh)
{
  SH2movwm(sh,7,1);
}
#define init_2_7_2_5
void FASTCALL call_2_7_2_5(SH2_struct* sh)
{
  SH2movwm(sh,7,2);
}
#define init_2_7_3_5
void FASTCALL call_2_7_3_5(SH2_struct* sh)
{
  SH2movwm(sh,7,3);
}
#define init_2_7_4_5
void FASTCALL call_2_7_4_5(SH2_struct* sh)
{
  SH2movwm(sh,7,4);
}
#define init_2_7_5_5
void FASTCALL call_2_7_5_5(SH2_struct* sh)
{
  SH2movwm(sh,7,5);
}
#define init_2_7_6_5
void FASTCALL call_2_7_6_5(SH2_struct* sh)
{
  SH2movwm(sh,7,6);
}
#define init_2_7_7_5
void FASTCALL call_2_7_7_5(SH2_struct* sh)
{
  SH2movwm(sh,7,7);
}
#define init_2_7_8_5
void FASTCALL call_2_7_8_5(SH2_struct* sh)
{
  SH2movwm(sh,7,8);
}
#define init_2_7_9_5
void FASTCALL call_2_7_9_5(SH2_struct* sh)
{
  SH2movwm(sh,7,9);
}
#define init_2_7_10_5
void FASTCALL call_2_7_10_5(SH2_struct* sh)
{
  SH2movwm(sh,7,10);
}
#define init_2_7_11_5
void FASTCALL call_2_7_11_5(SH2_struct* sh)
{
  SH2movwm(sh,7,11);
}
#define init_2_7_12_5
void FASTCALL call_2_7_12_5(SH2_struct* sh)
{
  SH2movwm(sh,7,12);
}
#define init_2_7_13_5
void FASTCALL call_2_7_13_5(SH2_struct* sh)
{
  SH2movwm(sh,7,13);
}
#define init_2_7_14_5
void FASTCALL call_2_7_14_5(SH2_struct* sh)
{
  SH2movwm(sh,7,14);
}
#define init_2_7_15_5
void FASTCALL call_2_7_15_5(SH2_struct* sh)
{
  SH2movwm(sh,7,15);
}
#define init_2_8_0_5
void FASTCALL call_2_8_0_5(SH2_struct* sh)
{
  SH2movwm(sh,8,0);
}
#define init_2_8_1_5
void FASTCALL call_2_8_1_5(SH2_struct* sh)
{
  SH2movwm(sh,8,1);
}
#define init_2_8_2_5
void FASTCALL call_2_8_2_5(SH2_struct* sh)
{
  SH2movwm(sh,8,2);
}
#define init_2_8_3_5
void FASTCALL call_2_8_3_5(SH2_struct* sh)
{
  SH2movwm(sh,8,3);
}
#define init_2_8_4_5
void FASTCALL call_2_8_4_5(SH2_struct* sh)
{
  SH2movwm(sh,8,4);
}
#define init_2_8_5_5
void FASTCALL call_2_8_5_5(SH2_struct* sh)
{
  SH2movwm(sh,8,5);
}
#define init_2_8_6_5
void FASTCALL call_2_8_6_5(SH2_struct* sh)
{
  SH2movwm(sh,8,6);
}
#define init_2_8_7_5
void FASTCALL call_2_8_7_5(SH2_struct* sh)
{
  SH2movwm(sh,8,7);
}
#define init_2_8_8_5
void FASTCALL call_2_8_8_5(SH2_struct* sh)
{
  SH2movwm(sh,8,8);
}
#define init_2_8_9_5
void FASTCALL call_2_8_9_5(SH2_struct* sh)
{
  SH2movwm(sh,8,9);
}
#define init_2_8_10_5
void FASTCALL call_2_8_10_5(SH2_struct* sh)
{
  SH2movwm(sh,8,10);
}
#define init_2_8_11_5
void FASTCALL call_2_8_11_5(SH2_struct* sh)
{
  SH2movwm(sh,8,11);
}
#define init_2_8_12_5
void FASTCALL call_2_8_12_5(SH2_struct* sh)
{
  SH2movwm(sh,8,12);
}
#define init_2_8_13_5
void FASTCALL call_2_8_13_5(SH2_struct* sh)
{
  SH2movwm(sh,8,13);
}
#define init_2_8_14_5
void FASTCALL call_2_8_14_5(SH2_struct* sh)
{
  SH2movwm(sh,8,14);
}
#define init_2_8_15_5
void FASTCALL call_2_8_15_5(SH2_struct* sh)
{
  SH2movwm(sh,8,15);
}
#define init_2_9_0_5
void FASTCALL call_2_9_0_5(SH2_struct* sh)
{
  SH2movwm(sh,9,0);
}
#define init_2_9_1_5
void FASTCALL call_2_9_1_5(SH2_struct* sh)
{
  SH2movwm(sh,9,1);
}
#define init_2_9_2_5
void FASTCALL call_2_9_2_5(SH2_struct* sh)
{
  SH2movwm(sh,9,2);
}
#define init_2_9_3_5
void FASTCALL call_2_9_3_5(SH2_struct* sh)
{
  SH2movwm(sh,9,3);
}
#define init_2_9_4_5
void FASTCALL call_2_9_4_5(SH2_struct* sh)
{
  SH2movwm(sh,9,4);
}
#define init_2_9_5_5
void FASTCALL call_2_9_5_5(SH2_struct* sh)
{
  SH2movwm(sh,9,5);
}
#define init_2_9_6_5
void FASTCALL call_2_9_6_5(SH2_struct* sh)
{
  SH2movwm(sh,9,6);
}
#define init_2_9_7_5
void FASTCALL call_2_9_7_5(SH2_struct* sh)
{
  SH2movwm(sh,9,7);
}
#define init_2_9_8_5
void FASTCALL call_2_9_8_5(SH2_struct* sh)
{
  SH2movwm(sh,9,8);
}
#define init_2_9_9_5
void FASTCALL call_2_9_9_5(SH2_struct* sh)
{
  SH2movwm(sh,9,9);
}
#define init_2_9_10_5
void FASTCALL call_2_9_10_5(SH2_struct* sh)
{
  SH2movwm(sh,9,10);
}
#define init_2_9_11_5
void FASTCALL call_2_9_11_5(SH2_struct* sh)
{
  SH2movwm(sh,9,11);
}
#define init_2_9_12_5
void FASTCALL call_2_9_12_5(SH2_struct* sh)
{
  SH2movwm(sh,9,12);
}
#define init_2_9_13_5
void FASTCALL call_2_9_13_5(SH2_struct* sh)
{
  SH2movwm(sh,9,13);
}
#define init_2_9_14_5
void FASTCALL call_2_9_14_5(SH2_struct* sh)
{
  SH2movwm(sh,9,14);
}
#define init_2_9_15_5
void FASTCALL call_2_9_15_5(SH2_struct* sh)
{
  SH2movwm(sh,9,15);
}
#define init_2_10_0_5
void FASTCALL call_2_10_0_5(SH2_struct* sh)
{
  SH2movwm(sh,10,0);
}
#define init_2_10_1_5
void FASTCALL call_2_10_1_5(SH2_struct* sh)
{
  SH2movwm(sh,10,1);
}
#define init_2_10_2_5
void FASTCALL call_2_10_2_5(SH2_struct* sh)
{
  SH2movwm(sh,10,2);
}
#define init_2_10_3_5
void FASTCALL call_2_10_3_5(SH2_struct* sh)
{
  SH2movwm(sh,10,3);
}
#define init_2_10_4_5
void FASTCALL call_2_10_4_5(SH2_struct* sh)
{
  SH2movwm(sh,10,4);
}
#define init_2_10_5_5
void FASTCALL call_2_10_5_5(SH2_struct* sh)
{
  SH2movwm(sh,10,5);
}
#define init_2_10_6_5
void FASTCALL call_2_10_6_5(SH2_struct* sh)
{
  SH2movwm(sh,10,6);
}
#define init_2_10_7_5
void FASTCALL call_2_10_7_5(SH2_struct* sh)
{
  SH2movwm(sh,10,7);
}
#define init_2_10_8_5
void FASTCALL call_2_10_8_5(SH2_struct* sh)
{
  SH2movwm(sh,10,8);
}
#define init_2_10_9_5
void FASTCALL call_2_10_9_5(SH2_struct* sh)
{
  SH2movwm(sh,10,9);
}
#define init_2_10_10_5
void FASTCALL call_2_10_10_5(SH2_struct* sh)
{
  SH2movwm(sh,10,10);
}
#define init_2_10_11_5
void FASTCALL call_2_10_11_5(SH2_struct* sh)
{
  SH2movwm(sh,10,11);
}
#define init_2_10_12_5
void FASTCALL call_2_10_12_5(SH2_struct* sh)
{
  SH2movwm(sh,10,12);
}
#define init_2_10_13_5
void FASTCALL call_2_10_13_5(SH2_struct* sh)
{
  SH2movwm(sh,10,13);
}
#define init_2_10_14_5
void FASTCALL call_2_10_14_5(SH2_struct* sh)
{
  SH2movwm(sh,10,14);
}
#define init_2_10_15_5
void FASTCALL call_2_10_15_5(SH2_struct* sh)
{
  SH2movwm(sh,10,15);
}
#define init_2_11_0_5
void FASTCALL call_2_11_0_5(SH2_struct* sh)
{
  SH2movwm(sh,11,0);
}
#define init_2_11_1_5
void FASTCALL call_2_11_1_5(SH2_struct* sh)
{
  SH2movwm(sh,11,1);
}
#define init_2_11_2_5
void FASTCALL call_2_11_2_5(SH2_struct* sh)
{
  SH2movwm(sh,11,2);
}
#define init_2_11_3_5
void FASTCALL call_2_11_3_5(SH2_struct* sh)
{
  SH2movwm(sh,11,3);
}
#define init_2_11_4_5
void FASTCALL call_2_11_4_5(SH2_struct* sh)
{
  SH2movwm(sh,11,4);
}
#define init_2_11_5_5
void FASTCALL call_2_11_5_5(SH2_struct* sh)
{
  SH2movwm(sh,11,5);
}
#define init_2_11_6_5
void FASTCALL call_2_11_6_5(SH2_struct* sh)
{
  SH2movwm(sh,11,6);
}
#define init_2_11_7_5
void FASTCALL call_2_11_7_5(SH2_struct* sh)
{
  SH2movwm(sh,11,7);
}
#define init_2_11_8_5
void FASTCALL call_2_11_8_5(SH2_struct* sh)
{
  SH2movwm(sh,11,8);
}
#define init_2_11_9_5
void FASTCALL call_2_11_9_5(SH2_struct* sh)
{
  SH2movwm(sh,11,9);
}
#define init_2_11_10_5
void FASTCALL call_2_11_10_5(SH2_struct* sh)
{
  SH2movwm(sh,11,10);
}
#define init_2_11_11_5
void FASTCALL call_2_11_11_5(SH2_struct* sh)
{
  SH2movwm(sh,11,11);
}
#define init_2_11_12_5
void FASTCALL call_2_11_12_5(SH2_struct* sh)
{
  SH2movwm(sh,11,12);
}
#define init_2_11_13_5
void FASTCALL call_2_11_13_5(SH2_struct* sh)
{
  SH2movwm(sh,11,13);
}
#define init_2_11_14_5
void FASTCALL call_2_11_14_5(SH2_struct* sh)
{
  SH2movwm(sh,11,14);
}
#define init_2_11_15_5
void FASTCALL call_2_11_15_5(SH2_struct* sh)
{
  SH2movwm(sh,11,15);
}
#define init_2_12_0_5
void FASTCALL call_2_12_0_5(SH2_struct* sh)
{
  SH2movwm(sh,12,0);
}
#define init_2_12_1_5
void FASTCALL call_2_12_1_5(SH2_struct* sh)
{
  SH2movwm(sh,12,1);
}
#define init_2_12_2_5
void FASTCALL call_2_12_2_5(SH2_struct* sh)
{
  SH2movwm(sh,12,2);
}
#define init_2_12_3_5
void FASTCALL call_2_12_3_5(SH2_struct* sh)
{
  SH2movwm(sh,12,3);
}
#define init_2_12_4_5
void FASTCALL call_2_12_4_5(SH2_struct* sh)
{
  SH2movwm(sh,12,4);
}
#define init_2_12_5_5
void FASTCALL call_2_12_5_5(SH2_struct* sh)
{
  SH2movwm(sh,12,5);
}
#define init_2_12_6_5
void FASTCALL call_2_12_6_5(SH2_struct* sh)
{
  SH2movwm(sh,12,6);
}
#define init_2_12_7_5
void FASTCALL call_2_12_7_5(SH2_struct* sh)
{
  SH2movwm(sh,12,7);
}
#define init_2_12_8_5
void FASTCALL call_2_12_8_5(SH2_struct* sh)
{
  SH2movwm(sh,12,8);
}
#define init_2_12_9_5
void FASTCALL call_2_12_9_5(SH2_struct* sh)
{
  SH2movwm(sh,12,9);
}
#define init_2_12_10_5
void FASTCALL call_2_12_10_5(SH2_struct* sh)
{
  SH2movwm(sh,12,10);
}
#define init_2_12_11_5
void FASTCALL call_2_12_11_5(SH2_struct* sh)
{
  SH2movwm(sh,12,11);
}
#define init_2_12_12_5
void FASTCALL call_2_12_12_5(SH2_struct* sh)
{
  SH2movwm(sh,12,12);
}
#define init_2_12_13_5
void FASTCALL call_2_12_13_5(SH2_struct* sh)
{
  SH2movwm(sh,12,13);
}
#define init_2_12_14_5
void FASTCALL call_2_12_14_5(SH2_struct* sh)
{
  SH2movwm(sh,12,14);
}
#define init_2_12_15_5
void FASTCALL call_2_12_15_5(SH2_struct* sh)
{
  SH2movwm(sh,12,15);
}
#define init_2_13_0_5
void FASTCALL call_2_13_0_5(SH2_struct* sh)
{
  SH2movwm(sh,13,0);
}
#define init_2_13_1_5
void FASTCALL call_2_13_1_5(SH2_struct* sh)
{
  SH2movwm(sh,13,1);
}
#define init_2_13_2_5
void FASTCALL call_2_13_2_5(SH2_struct* sh)
{
  SH2movwm(sh,13,2);
}
#define init_2_13_3_5
void FASTCALL call_2_13_3_5(SH2_struct* sh)
{
  SH2movwm(sh,13,3);
}
#define init_2_13_4_5
void FASTCALL call_2_13_4_5(SH2_struct* sh)
{
  SH2movwm(sh,13,4);
}
#define init_2_13_5_5
void FASTCALL call_2_13_5_5(SH2_struct* sh)
{
  SH2movwm(sh,13,5);
}
#define init_2_13_6_5
void FASTCALL call_2_13_6_5(SH2_struct* sh)
{
  SH2movwm(sh,13,6);
}
#define init_2_13_7_5
void FASTCALL call_2_13_7_5(SH2_struct* sh)
{
  SH2movwm(sh,13,7);
}
#define init_2_13_8_5
void FASTCALL call_2_13_8_5(SH2_struct* sh)
{
  SH2movwm(sh,13,8);
}
#define init_2_13_9_5
void FASTCALL call_2_13_9_5(SH2_struct* sh)
{
  SH2movwm(sh,13,9);
}
#define init_2_13_10_5
void FASTCALL call_2_13_10_5(SH2_struct* sh)
{
  SH2movwm(sh,13,10);
}
#define init_2_13_11_5
void FASTCALL call_2_13_11_5(SH2_struct* sh)
{
  SH2movwm(sh,13,11);
}
#define init_2_13_12_5
void FASTCALL call_2_13_12_5(SH2_struct* sh)
{
  SH2movwm(sh,13,12);
}
#define init_2_13_13_5
void FASTCALL call_2_13_13_5(SH2_struct* sh)
{
  SH2movwm(sh,13,13);
}
#define init_2_13_14_5
void FASTCALL call_2_13_14_5(SH2_struct* sh)
{
  SH2movwm(sh,13,14);
}
#define init_2_13_15_5
void FASTCALL call_2_13_15_5(SH2_struct* sh)
{
  SH2movwm(sh,13,15);
}
#define init_2_14_0_5
void FASTCALL call_2_14_0_5(SH2_struct* sh)
{
  SH2movwm(sh,14,0);
}
#define init_2_14_1_5
void FASTCALL call_2_14_1_5(SH2_struct* sh)
{
  SH2movwm(sh,14,1);
}
#define init_2_14_2_5
void FASTCALL call_2_14_2_5(SH2_struct* sh)
{
  SH2movwm(sh,14,2);
}
#define init_2_14_3_5
void FASTCALL call_2_14_3_5(SH2_struct* sh)
{
  SH2movwm(sh,14,3);
}
#define init_2_14_4_5
void FASTCALL call_2_14_4_5(SH2_struct* sh)
{
  SH2movwm(sh,14,4);
}
#define init_2_14_5_5
void FASTCALL call_2_14_5_5(SH2_struct* sh)
{
  SH2movwm(sh,14,5);
}
#define init_2_14_6_5
void FASTCALL call_2_14_6_5(SH2_struct* sh)
{
  SH2movwm(sh,14,6);
}
#define init_2_14_7_5
void FASTCALL call_2_14_7_5(SH2_struct* sh)
{
  SH2movwm(sh,14,7);
}
#define init_2_14_8_5
void FASTCALL call_2_14_8_5(SH2_struct* sh)
{
  SH2movwm(sh,14,8);
}
#define init_2_14_9_5
void FASTCALL call_2_14_9_5(SH2_struct* sh)
{
  SH2movwm(sh,14,9);
}
#define init_2_14_10_5
void FASTCALL call_2_14_10_5(SH2_struct* sh)
{
  SH2movwm(sh,14,10);
}
#define init_2_14_11_5
void FASTCALL call_2_14_11_5(SH2_struct* sh)
{
  SH2movwm(sh,14,11);
}
#define init_2_14_12_5
void FASTCALL call_2_14_12_5(SH2_struct* sh)
{
  SH2movwm(sh,14,12);
}
#define init_2_14_13_5
void FASTCALL call_2_14_13_5(SH2_struct* sh)
{
  SH2movwm(sh,14,13);
}
#define init_2_14_14_5
void FASTCALL call_2_14_14_5(SH2_struct* sh)
{
  SH2movwm(sh,14,14);
}
#define init_2_14_15_5
void FASTCALL call_2_14_15_5(SH2_struct* sh)
{
  SH2movwm(sh,14,15);
}
#define init_2_15_0_5
void FASTCALL call_2_15_0_5(SH2_struct* sh)
{
  SH2movwm(sh,15,0);
}
#define init_2_15_1_5
void FASTCALL call_2_15_1_5(SH2_struct* sh)
{
  SH2movwm(sh,15,1);
}
#define init_2_15_2_5
void FASTCALL call_2_15_2_5(SH2_struct* sh)
{
  SH2movwm(sh,15,2);
}
#define init_2_15_3_5
void FASTCALL call_2_15_3_5(SH2_struct* sh)
{
  SH2movwm(sh,15,3);
}
#define init_2_15_4_5
void FASTCALL call_2_15_4_5(SH2_struct* sh)
{
  SH2movwm(sh,15,4);
}
#define init_2_15_5_5
void FASTCALL call_2_15_5_5(SH2_struct* sh)
{
  SH2movwm(sh,15,5);
}
#define init_2_15_6_5
void FASTCALL call_2_15_6_5(SH2_struct* sh)
{
  SH2movwm(sh,15,6);
}
#define init_2_15_7_5
void FASTCALL call_2_15_7_5(SH2_struct* sh)
{
  SH2movwm(sh,15,7);
}
#define init_2_15_8_5
void FASTCALL call_2_15_8_5(SH2_struct* sh)
{
  SH2movwm(sh,15,8);
}
#define init_2_15_9_5
void FASTCALL call_2_15_9_5(SH2_struct* sh)
{
  SH2movwm(sh,15,9);
}
#define init_2_15_10_5
void FASTCALL call_2_15_10_5(SH2_struct* sh)
{
  SH2movwm(sh,15,10);
}
#define init_2_15_11_5
void FASTCALL call_2_15_11_5(SH2_struct* sh)
{
  SH2movwm(sh,15,11);
}
#define init_2_15_12_5
void FASTCALL call_2_15_12_5(SH2_struct* sh)
{
  SH2movwm(sh,15,12);
}
#define init_2_15_13_5
void FASTCALL call_2_15_13_5(SH2_struct* sh)
{
  SH2movwm(sh,15,13);
}
#define init_2_15_14_5
void FASTCALL call_2_15_14_5(SH2_struct* sh)
{
  SH2movwm(sh,15,14);
}
#define init_2_15_15_5
void FASTCALL call_2_15_15_5(SH2_struct* sh)
{
  SH2movwm(sh,15,15);
}
#define init_12_5_0_0
void FASTCALL call_12_5_0_0(SH2_struct* sh)
{
  SH2movwlg(sh,(0<<4)+0);
}
#define init_12_5_0_1
void FASTCALL call_12_5_0_1(SH2_struct* sh)
{
  SH2movwlg(sh,(0<<4)+1);
}
#define init_12_5_0_2
void FASTCALL call_12_5_0_2(SH2_struct* sh)
{
  SH2movwlg(sh,(0<<4)+2);
}
#define init_12_5_0_3
void FASTCALL call_12_5_0_3(SH2_struct* sh)
{
  SH2movwlg(sh,(0<<4)+3);
}
#define init_12_5_0_4
void FASTCALL call_12_5_0_4(SH2_struct* sh)
{
  SH2movwlg(sh,(0<<4)+4);
}
#define init_12_5_0_5
void FASTCALL call_12_5_0_5(SH2_struct* sh)
{
  SH2movwlg(sh,(0<<4)+5);
}
#define init_12_5_0_6
void FASTCALL call_12_5_0_6(SH2_struct* sh)
{
  SH2movwlg(sh,(0<<4)+6);
}
#define init_12_5_0_7
void FASTCALL call_12_5_0_7(SH2_struct* sh)
{
  SH2movwlg(sh,(0<<4)+7);
}
#define init_12_5_0_8
void FASTCALL call_12_5_0_8(SH2_struct* sh)
{
  SH2movwlg(sh,(0<<4)+8);
}
#define init_12_5_0_9
void FASTCALL call_12_5_0_9(SH2_struct* sh)
{
  SH2movwlg(sh,(0<<4)+9);
}
#define init_12_5_0_10
void FASTCALL call_12_5_0_10(SH2_struct* sh)
{
  SH2movwlg(sh,(0<<4)+10);
}
#define init_12_5_0_11
void FASTCALL call_12_5_0_11(SH2_struct* sh)
{
  SH2movwlg(sh,(0<<4)+11);
}
#define init_12_5_0_12
void FASTCALL call_12_5_0_12(SH2_struct* sh)
{
  SH2movwlg(sh,(0<<4)+12);
}
#define init_12_5_0_13
void FASTCALL call_12_5_0_13(SH2_struct* sh)
{
  SH2movwlg(sh,(0<<4)+13);
}
#define init_12_5_0_14
void FASTCALL call_12_5_0_14(SH2_struct* sh)
{
  SH2movwlg(sh,(0<<4)+14);
}
#define init_12_5_0_15
void FASTCALL call_12_5_0_15(SH2_struct* sh)
{
  SH2movwlg(sh,(0<<4)+15);
}
#define init_12_5_1_0
void FASTCALL call_12_5_1_0(SH2_struct* sh)
{
  SH2movwlg(sh,(1<<4)+0);
}
#define init_12_5_1_1
void FASTCALL call_12_5_1_1(SH2_struct* sh)
{
  SH2movwlg(sh,(1<<4)+1);
}
#define init_12_5_1_2
void FASTCALL call_12_5_1_2(SH2_struct* sh)
{
  SH2movwlg(sh,(1<<4)+2);
}
#define init_12_5_1_3
void FASTCALL call_12_5_1_3(SH2_struct* sh)
{
  SH2movwlg(sh,(1<<4)+3);
}
#define init_12_5_1_4
void FASTCALL call_12_5_1_4(SH2_struct* sh)
{
  SH2movwlg(sh,(1<<4)+4);
}
#define init_12_5_1_5
void FASTCALL call_12_5_1_5(SH2_struct* sh)
{
  SH2movwlg(sh,(1<<4)+5);
}
#define init_12_5_1_6
void FASTCALL call_12_5_1_6(SH2_struct* sh)
{
  SH2movwlg(sh,(1<<4)+6);
}
#define init_12_5_1_7
void FASTCALL call_12_5_1_7(SH2_struct* sh)
{
  SH2movwlg(sh,(1<<4)+7);
}
#define init_12_5_1_8
void FASTCALL call_12_5_1_8(SH2_struct* sh)
{
  SH2movwlg(sh,(1<<4)+8);
}
#define init_12_5_1_9
void FASTCALL call_12_5_1_9(SH2_struct* sh)
{
  SH2movwlg(sh,(1<<4)+9);
}
#define init_12_5_1_10
void FASTCALL call_12_5_1_10(SH2_struct* sh)
{
  SH2movwlg(sh,(1<<4)+10);
}
#define init_12_5_1_11
void FASTCALL call_12_5_1_11(SH2_struct* sh)
{
  SH2movwlg(sh,(1<<4)+11);
}
#define init_12_5_1_12
void FASTCALL call_12_5_1_12(SH2_struct* sh)
{
  SH2movwlg(sh,(1<<4)+12);
}
#define init_12_5_1_13
void FASTCALL call_12_5_1_13(SH2_struct* sh)
{
  SH2movwlg(sh,(1<<4)+13);
}
#define init_12_5_1_14
void FASTCALL call_12_5_1_14(SH2_struct* sh)
{
  SH2movwlg(sh,(1<<4)+14);
}
#define init_12_5_1_15
void FASTCALL call_12_5_1_15(SH2_struct* sh)
{
  SH2movwlg(sh,(1<<4)+15);
}
#define init_12_5_2_0
void FASTCALL call_12_5_2_0(SH2_struct* sh)
{
  SH2movwlg(sh,(2<<4)+0);
}
#define init_12_5_2_1
void FASTCALL call_12_5_2_1(SH2_struct* sh)
{
  SH2movwlg(sh,(2<<4)+1);
}
#define init_12_5_2_2
void FASTCALL call_12_5_2_2(SH2_struct* sh)
{
  SH2movwlg(sh,(2<<4)+2);
}
#define init_12_5_2_3
void FASTCALL call_12_5_2_3(SH2_struct* sh)
{
  SH2movwlg(sh,(2<<4)+3);
}
#define init_12_5_2_4
void FASTCALL call_12_5_2_4(SH2_struct* sh)
{
  SH2movwlg(sh,(2<<4)+4);
}
#define init_12_5_2_5
void FASTCALL call_12_5_2_5(SH2_struct* sh)
{
  SH2movwlg(sh,(2<<4)+5);
}
#define init_12_5_2_6
void FASTCALL call_12_5_2_6(SH2_struct* sh)
{
  SH2movwlg(sh,(2<<4)+6);
}
#define init_12_5_2_7
void FASTCALL call_12_5_2_7(SH2_struct* sh)
{
  SH2movwlg(sh,(2<<4)+7);
}
#define init_12_5_2_8
void FASTCALL call_12_5_2_8(SH2_struct* sh)
{
  SH2movwlg(sh,(2<<4)+8);
}
#define init_12_5_2_9
void FASTCALL call_12_5_2_9(SH2_struct* sh)
{
  SH2movwlg(sh,(2<<4)+9);
}
#define init_12_5_2_10
void FASTCALL call_12_5_2_10(SH2_struct* sh)
{
  SH2movwlg(sh,(2<<4)+10);
}
#define init_12_5_2_11
void FASTCALL call_12_5_2_11(SH2_struct* sh)
{
  SH2movwlg(sh,(2<<4)+11);
}
#define init_12_5_2_12
void FASTCALL call_12_5_2_12(SH2_struct* sh)
{
  SH2movwlg(sh,(2<<4)+12);
}
#define init_12_5_2_13
void FASTCALL call_12_5_2_13(SH2_struct* sh)
{
  SH2movwlg(sh,(2<<4)+13);
}
#define init_12_5_2_14
void FASTCALL call_12_5_2_14(SH2_struct* sh)
{
  SH2movwlg(sh,(2<<4)+14);
}
#define init_12_5_2_15
void FASTCALL call_12_5_2_15(SH2_struct* sh)
{
  SH2movwlg(sh,(2<<4)+15);
}
#define init_12_5_3_0
void FASTCALL call_12_5_3_0(SH2_struct* sh)
{
  SH2movwlg(sh,(3<<4)+0);
}
#define init_12_5_3_1
void FASTCALL call_12_5_3_1(SH2_struct* sh)
{
  SH2movwlg(sh,(3<<4)+1);
}
#define init_12_5_3_2
void FASTCALL call_12_5_3_2(SH2_struct* sh)
{
  SH2movwlg(sh,(3<<4)+2);
}
#define init_12_5_3_3
void FASTCALL call_12_5_3_3(SH2_struct* sh)
{
  SH2movwlg(sh,(3<<4)+3);
}
#define init_12_5_3_4
void FASTCALL call_12_5_3_4(SH2_struct* sh)
{
  SH2movwlg(sh,(3<<4)+4);
}
#define init_12_5_3_5
void FASTCALL call_12_5_3_5(SH2_struct* sh)
{
  SH2movwlg(sh,(3<<4)+5);
}
#define init_12_5_3_6
void FASTCALL call_12_5_3_6(SH2_struct* sh)
{
  SH2movwlg(sh,(3<<4)+6);
}
#define init_12_5_3_7
void FASTCALL call_12_5_3_7(SH2_struct* sh)
{
  SH2movwlg(sh,(3<<4)+7);
}
#define init_12_5_3_8
void FASTCALL call_12_5_3_8(SH2_struct* sh)
{
  SH2movwlg(sh,(3<<4)+8);
}
#define init_12_5_3_9
void FASTCALL call_12_5_3_9(SH2_struct* sh)
{
  SH2movwlg(sh,(3<<4)+9);
}
#define init_12_5_3_10
void FASTCALL call_12_5_3_10(SH2_struct* sh)
{
  SH2movwlg(sh,(3<<4)+10);
}
#define init_12_5_3_11
void FASTCALL call_12_5_3_11(SH2_struct* sh)
{
  SH2movwlg(sh,(3<<4)+11);
}
#define init_12_5_3_12
void FASTCALL call_12_5_3_12(SH2_struct* sh)
{
  SH2movwlg(sh,(3<<4)+12);
}
#define init_12_5_3_13
void FASTCALL call_12_5_3_13(SH2_struct* sh)
{
  SH2movwlg(sh,(3<<4)+13);
}
#define init_12_5_3_14
void FASTCALL call_12_5_3_14(SH2_struct* sh)
{
  SH2movwlg(sh,(3<<4)+14);
}
#define init_12_5_3_15
void FASTCALL call_12_5_3_15(SH2_struct* sh)
{
  SH2movwlg(sh,(3<<4)+15);
}
#define init_12_5_4_0
void FASTCALL call_12_5_4_0(SH2_struct* sh)
{
  SH2movwlg(sh,(4<<4)+0);
}
#define init_12_5_4_1
void FASTCALL call_12_5_4_1(SH2_struct* sh)
{
  SH2movwlg(sh,(4<<4)+1);
}
#define init_12_5_4_2
void FASTCALL call_12_5_4_2(SH2_struct* sh)
{
  SH2movwlg(sh,(4<<4)+2);
}
#define init_12_5_4_3
void FASTCALL call_12_5_4_3(SH2_struct* sh)
{
  SH2movwlg(sh,(4<<4)+3);
}
#define init_12_5_4_4
void FASTCALL call_12_5_4_4(SH2_struct* sh)
{
  SH2movwlg(sh,(4<<4)+4);
}
#define init_12_5_4_5
void FASTCALL call_12_5_4_5(SH2_struct* sh)
{
  SH2movwlg(sh,(4<<4)+5);
}
#define init_12_5_4_6
void FASTCALL call_12_5_4_6(SH2_struct* sh)
{
  SH2movwlg(sh,(4<<4)+6);
}
#define init_12_5_4_7
void FASTCALL call_12_5_4_7(SH2_struct* sh)
{
  SH2movwlg(sh,(4<<4)+7);
}
#define init_12_5_4_8
void FASTCALL call_12_5_4_8(SH2_struct* sh)
{
  SH2movwlg(sh,(4<<4)+8);
}
#define init_12_5_4_9
void FASTCALL call_12_5_4_9(SH2_struct* sh)
{
  SH2movwlg(sh,(4<<4)+9);
}
#define init_12_5_4_10
void FASTCALL call_12_5_4_10(SH2_struct* sh)
{
  SH2movwlg(sh,(4<<4)+10);
}
#define init_12_5_4_11
void FASTCALL call_12_5_4_11(SH2_struct* sh)
{
  SH2movwlg(sh,(4<<4)+11);
}
#define init_12_5_4_12
void FASTCALL call_12_5_4_12(SH2_struct* sh)
{
  SH2movwlg(sh,(4<<4)+12);
}
#define init_12_5_4_13
void FASTCALL call_12_5_4_13(SH2_struct* sh)
{
  SH2movwlg(sh,(4<<4)+13);
}
#define init_12_5_4_14
void FASTCALL call_12_5_4_14(SH2_struct* sh)
{
  SH2movwlg(sh,(4<<4)+14);
}
#define init_12_5_4_15
void FASTCALL call_12_5_4_15(SH2_struct* sh)
{
  SH2movwlg(sh,(4<<4)+15);
}
#define init_12_5_5_0
void FASTCALL call_12_5_5_0(SH2_struct* sh)
{
  SH2movwlg(sh,(5<<4)+0);
}
#define init_12_5_5_1
void FASTCALL call_12_5_5_1(SH2_struct* sh)
{
  SH2movwlg(sh,(5<<4)+1);
}
#define init_12_5_5_2
void FASTCALL call_12_5_5_2(SH2_struct* sh)
{
  SH2movwlg(sh,(5<<4)+2);
}
#define init_12_5_5_3
void FASTCALL call_12_5_5_3(SH2_struct* sh)
{
  SH2movwlg(sh,(5<<4)+3);
}
#define init_12_5_5_4
void FASTCALL call_12_5_5_4(SH2_struct* sh)
{
  SH2movwlg(sh,(5<<4)+4);
}
#define init_12_5_5_5
void FASTCALL call_12_5_5_5(SH2_struct* sh)
{
  SH2movwlg(sh,(5<<4)+5);
}
#define init_12_5_5_6
void FASTCALL call_12_5_5_6(SH2_struct* sh)
{
  SH2movwlg(sh,(5<<4)+6);
}
#define init_12_5_5_7
void FASTCALL call_12_5_5_7(SH2_struct* sh)
{
  SH2movwlg(sh,(5<<4)+7);
}
#define init_12_5_5_8
void FASTCALL call_12_5_5_8(SH2_struct* sh)
{
  SH2movwlg(sh,(5<<4)+8);
}
#define init_12_5_5_9
void FASTCALL call_12_5_5_9(SH2_struct* sh)
{
  SH2movwlg(sh,(5<<4)+9);
}
#define init_12_5_5_10
void FASTCALL call_12_5_5_10(SH2_struct* sh)
{
  SH2movwlg(sh,(5<<4)+10);
}
#define init_12_5_5_11
void FASTCALL call_12_5_5_11(SH2_struct* sh)
{
  SH2movwlg(sh,(5<<4)+11);
}
#define init_12_5_5_12
void FASTCALL call_12_5_5_12(SH2_struct* sh)
{
  SH2movwlg(sh,(5<<4)+12);
}
#define init_12_5_5_13
void FASTCALL call_12_5_5_13(SH2_struct* sh)
{
  SH2movwlg(sh,(5<<4)+13);
}
#define init_12_5_5_14
void FASTCALL call_12_5_5_14(SH2_struct* sh)
{
  SH2movwlg(sh,(5<<4)+14);
}
#define init_12_5_5_15
void FASTCALL call_12_5_5_15(SH2_struct* sh)
{
  SH2movwlg(sh,(5<<4)+15);
}
#define init_12_5_6_0
void FASTCALL call_12_5_6_0(SH2_struct* sh)
{
  SH2movwlg(sh,(6<<4)+0);
}
#define init_12_5_6_1
void FASTCALL call_12_5_6_1(SH2_struct* sh)
{
  SH2movwlg(sh,(6<<4)+1);
}
#define init_12_5_6_2
void FASTCALL call_12_5_6_2(SH2_struct* sh)
{
  SH2movwlg(sh,(6<<4)+2);
}
#define init_12_5_6_3
void FASTCALL call_12_5_6_3(SH2_struct* sh)
{
  SH2movwlg(sh,(6<<4)+3);
}
#define init_12_5_6_4
void FASTCALL call_12_5_6_4(SH2_struct* sh)
{
  SH2movwlg(sh,(6<<4)+4);
}
#define init_12_5_6_5
void FASTCALL call_12_5_6_5(SH2_struct* sh)
{
  SH2movwlg(sh,(6<<4)+5);
}
#define init_12_5_6_6
void FASTCALL call_12_5_6_6(SH2_struct* sh)
{
  SH2movwlg(sh,(6<<4)+6);
}
#define init_12_5_6_7
void FASTCALL call_12_5_6_7(SH2_struct* sh)
{
  SH2movwlg(sh,(6<<4)+7);
}
#define init_12_5_6_8
void FASTCALL call_12_5_6_8(SH2_struct* sh)
{
  SH2movwlg(sh,(6<<4)+8);
}
#define init_12_5_6_9
void FASTCALL call_12_5_6_9(SH2_struct* sh)
{
  SH2movwlg(sh,(6<<4)+9);
}
#define init_12_5_6_10
void FASTCALL call_12_5_6_10(SH2_struct* sh)
{
  SH2movwlg(sh,(6<<4)+10);
}
#define init_12_5_6_11
void FASTCALL call_12_5_6_11(SH2_struct* sh)
{
  SH2movwlg(sh,(6<<4)+11);
}
#define init_12_5_6_12
void FASTCALL call_12_5_6_12(SH2_struct* sh)
{
  SH2movwlg(sh,(6<<4)+12);
}
#define init_12_5_6_13
void FASTCALL call_12_5_6_13(SH2_struct* sh)
{
  SH2movwlg(sh,(6<<4)+13);
}
#define init_12_5_6_14
void FASTCALL call_12_5_6_14(SH2_struct* sh)
{
  SH2movwlg(sh,(6<<4)+14);
}
#define init_12_5_6_15
void FASTCALL call_12_5_6_15(SH2_struct* sh)
{
  SH2movwlg(sh,(6<<4)+15);
}
#define init_12_5_7_0
void FASTCALL call_12_5_7_0(SH2_struct* sh)
{
  SH2movwlg(sh,(7<<4)+0);
}
#define init_12_5_7_1
void FASTCALL call_12_5_7_1(SH2_struct* sh)
{
  SH2movwlg(sh,(7<<4)+1);
}
#define init_12_5_7_2
void FASTCALL call_12_5_7_2(SH2_struct* sh)
{
  SH2movwlg(sh,(7<<4)+2);
}
#define init_12_5_7_3
void FASTCALL call_12_5_7_3(SH2_struct* sh)
{
  SH2movwlg(sh,(7<<4)+3);
}
#define init_12_5_7_4
void FASTCALL call_12_5_7_4(SH2_struct* sh)
{
  SH2movwlg(sh,(7<<4)+4);
}
#define init_12_5_7_5
void FASTCALL call_12_5_7_5(SH2_struct* sh)
{
  SH2movwlg(sh,(7<<4)+5);
}
#define init_12_5_7_6
void FASTCALL call_12_5_7_6(SH2_struct* sh)
{
  SH2movwlg(sh,(7<<4)+6);
}
#define init_12_5_7_7
void FASTCALL call_12_5_7_7(SH2_struct* sh)
{
  SH2movwlg(sh,(7<<4)+7);
}
#define init_12_5_7_8
void FASTCALL call_12_5_7_8(SH2_struct* sh)
{
  SH2movwlg(sh,(7<<4)+8);
}
#define init_12_5_7_9
void FASTCALL call_12_5_7_9(SH2_struct* sh)
{
  SH2movwlg(sh,(7<<4)+9);
}
#define init_12_5_7_10
void FASTCALL call_12_5_7_10(SH2_struct* sh)
{
  SH2movwlg(sh,(7<<4)+10);
}
#define init_12_5_7_11
void FASTCALL call_12_5_7_11(SH2_struct* sh)
{
  SH2movwlg(sh,(7<<4)+11);
}
#define init_12_5_7_12
void FASTCALL call_12_5_7_12(SH2_struct* sh)
{
  SH2movwlg(sh,(7<<4)+12);
}
#define init_12_5_7_13
void FASTCALL call_12_5_7_13(SH2_struct* sh)
{
  SH2movwlg(sh,(7<<4)+13);
}
#define init_12_5_7_14
void FASTCALL call_12_5_7_14(SH2_struct* sh)
{
  SH2movwlg(sh,(7<<4)+14);
}
#define init_12_5_7_15
void FASTCALL call_12_5_7_15(SH2_struct* sh)
{
  SH2movwlg(sh,(7<<4)+15);
}
#define init_12_5_8_0
void FASTCALL call_12_5_8_0(SH2_struct* sh)
{
  SH2movwlg(sh,(8<<4)+0);
}
#define init_12_5_8_1
void FASTCALL call_12_5_8_1(SH2_struct* sh)
{
  SH2movwlg(sh,(8<<4)+1);
}
#define init_12_5_8_2
void FASTCALL call_12_5_8_2(SH2_struct* sh)
{
  SH2movwlg(sh,(8<<4)+2);
}
#define init_12_5_8_3
void FASTCALL call_12_5_8_3(SH2_struct* sh)
{
  SH2movwlg(sh,(8<<4)+3);
}
#define init_12_5_8_4
void FASTCALL call_12_5_8_4(SH2_struct* sh)
{
  SH2movwlg(sh,(8<<4)+4);
}
#define init_12_5_8_5
void FASTCALL call_12_5_8_5(SH2_struct* sh)
{
  SH2movwlg(sh,(8<<4)+5);
}
#define init_12_5_8_6
void FASTCALL call_12_5_8_6(SH2_struct* sh)
{
  SH2movwlg(sh,(8<<4)+6);
}
#define init_12_5_8_7
void FASTCALL call_12_5_8_7(SH2_struct* sh)
{
  SH2movwlg(sh,(8<<4)+7);
}
#define init_12_5_8_8
void FASTCALL call_12_5_8_8(SH2_struct* sh)
{
  SH2movwlg(sh,(8<<4)+8);
}
#define init_12_5_8_9
void FASTCALL call_12_5_8_9(SH2_struct* sh)
{
  SH2movwlg(sh,(8<<4)+9);
}
#define init_12_5_8_10
void FASTCALL call_12_5_8_10(SH2_struct* sh)
{
  SH2movwlg(sh,(8<<4)+10);
}
#define init_12_5_8_11
void FASTCALL call_12_5_8_11(SH2_struct* sh)
{
  SH2movwlg(sh,(8<<4)+11);
}
#define init_12_5_8_12
void FASTCALL call_12_5_8_12(SH2_struct* sh)
{
  SH2movwlg(sh,(8<<4)+12);
}
#define init_12_5_8_13
void FASTCALL call_12_5_8_13(SH2_struct* sh)
{
  SH2movwlg(sh,(8<<4)+13);
}
#define init_12_5_8_14
void FASTCALL call_12_5_8_14(SH2_struct* sh)
{
  SH2movwlg(sh,(8<<4)+14);
}
#define init_12_5_8_15
void FASTCALL call_12_5_8_15(SH2_struct* sh)
{
  SH2movwlg(sh,(8<<4)+15);
}
#define init_12_5_9_0
void FASTCALL call_12_5_9_0(SH2_struct* sh)
{
  SH2movwlg(sh,(9<<4)+0);
}
#define init_12_5_9_1
void FASTCALL call_12_5_9_1(SH2_struct* sh)
{
  SH2movwlg(sh,(9<<4)+1);
}
#define init_12_5_9_2
void FASTCALL call_12_5_9_2(SH2_struct* sh)
{
  SH2movwlg(sh,(9<<4)+2);
}
#define init_12_5_9_3
void FASTCALL call_12_5_9_3(SH2_struct* sh)
{
  SH2movwlg(sh,(9<<4)+3);
}
#define init_12_5_9_4
void FASTCALL call_12_5_9_4(SH2_struct* sh)
{
  SH2movwlg(sh,(9<<4)+4);
}
#define init_12_5_9_5
void FASTCALL call_12_5_9_5(SH2_struct* sh)
{
  SH2movwlg(sh,(9<<4)+5);
}
#define init_12_5_9_6
void FASTCALL call_12_5_9_6(SH2_struct* sh)
{
  SH2movwlg(sh,(9<<4)+6);
}
#define init_12_5_9_7
void FASTCALL call_12_5_9_7(SH2_struct* sh)
{
  SH2movwlg(sh,(9<<4)+7);
}
#define init_12_5_9_8
void FASTCALL call_12_5_9_8(SH2_struct* sh)
{
  SH2movwlg(sh,(9<<4)+8);
}
#define init_12_5_9_9
void FASTCALL call_12_5_9_9(SH2_struct* sh)
{
  SH2movwlg(sh,(9<<4)+9);
}
#define init_12_5_9_10
void FASTCALL call_12_5_9_10(SH2_struct* sh)
{
  SH2movwlg(sh,(9<<4)+10);
}
#define init_12_5_9_11
void FASTCALL call_12_5_9_11(SH2_struct* sh)
{
  SH2movwlg(sh,(9<<4)+11);
}
#define init_12_5_9_12
void FASTCALL call_12_5_9_12(SH2_struct* sh)
{
  SH2movwlg(sh,(9<<4)+12);
}
#define init_12_5_9_13
void FASTCALL call_12_5_9_13(SH2_struct* sh)
{
  SH2movwlg(sh,(9<<4)+13);
}
#define init_12_5_9_14
void FASTCALL call_12_5_9_14(SH2_struct* sh)
{
  SH2movwlg(sh,(9<<4)+14);
}
#define init_12_5_9_15
void FASTCALL call_12_5_9_15(SH2_struct* sh)
{
  SH2movwlg(sh,(9<<4)+15);
}
#define init_12_5_10_0
void FASTCALL call_12_5_10_0(SH2_struct* sh)
{
  SH2movwlg(sh,(10<<4)+0);
}
#define init_12_5_10_1
void FASTCALL call_12_5_10_1(SH2_struct* sh)
{
  SH2movwlg(sh,(10<<4)+1);
}
#define init_12_5_10_2
void FASTCALL call_12_5_10_2(SH2_struct* sh)
{
  SH2movwlg(sh,(10<<4)+2);
}
#define init_12_5_10_3
void FASTCALL call_12_5_10_3(SH2_struct* sh)
{
  SH2movwlg(sh,(10<<4)+3);
}
#define init_12_5_10_4
void FASTCALL call_12_5_10_4(SH2_struct* sh)
{
  SH2movwlg(sh,(10<<4)+4);
}
#define init_12_5_10_5
void FASTCALL call_12_5_10_5(SH2_struct* sh)
{
  SH2movwlg(sh,(10<<4)+5);
}
#define init_12_5_10_6
void FASTCALL call_12_5_10_6(SH2_struct* sh)
{
  SH2movwlg(sh,(10<<4)+6);
}
#define init_12_5_10_7
void FASTCALL call_12_5_10_7(SH2_struct* sh)
{
  SH2movwlg(sh,(10<<4)+7);
}
#define init_12_5_10_8
void FASTCALL call_12_5_10_8(SH2_struct* sh)
{
  SH2movwlg(sh,(10<<4)+8);
}
#define init_12_5_10_9
void FASTCALL call_12_5_10_9(SH2_struct* sh)
{
  SH2movwlg(sh,(10<<4)+9);
}
#define init_12_5_10_10
void FASTCALL call_12_5_10_10(SH2_struct* sh)
{
  SH2movwlg(sh,(10<<4)+10);
}
#define init_12_5_10_11
void FASTCALL call_12_5_10_11(SH2_struct* sh)
{
  SH2movwlg(sh,(10<<4)+11);
}
#define init_12_5_10_12
void FASTCALL call_12_5_10_12(SH2_struct* sh)
{
  SH2movwlg(sh,(10<<4)+12);
}
#define init_12_5_10_13
void FASTCALL call_12_5_10_13(SH2_struct* sh)
{
  SH2movwlg(sh,(10<<4)+13);
}
#define init_12_5_10_14
void FASTCALL call_12_5_10_14(SH2_struct* sh)
{
  SH2movwlg(sh,(10<<4)+14);
}
#define init_12_5_10_15
void FASTCALL call_12_5_10_15(SH2_struct* sh)
{
  SH2movwlg(sh,(10<<4)+15);
}
#define init_12_5_11_0
void FASTCALL call_12_5_11_0(SH2_struct* sh)
{
  SH2movwlg(sh,(11<<4)+0);
}
#define init_12_5_11_1
void FASTCALL call_12_5_11_1(SH2_struct* sh)
{
  SH2movwlg(sh,(11<<4)+1);
}
#define init_12_5_11_2
void FASTCALL call_12_5_11_2(SH2_struct* sh)
{
  SH2movwlg(sh,(11<<4)+2);
}
#define init_12_5_11_3
void FASTCALL call_12_5_11_3(SH2_struct* sh)
{
  SH2movwlg(sh,(11<<4)+3);
}
#define init_12_5_11_4
void FASTCALL call_12_5_11_4(SH2_struct* sh)
{
  SH2movwlg(sh,(11<<4)+4);
}
#define init_12_5_11_5
void FASTCALL call_12_5_11_5(SH2_struct* sh)
{
  SH2movwlg(sh,(11<<4)+5);
}
#define init_12_5_11_6
void FASTCALL call_12_5_11_6(SH2_struct* sh)
{
  SH2movwlg(sh,(11<<4)+6);
}
#define init_12_5_11_7
void FASTCALL call_12_5_11_7(SH2_struct* sh)
{
  SH2movwlg(sh,(11<<4)+7);
}
#define init_12_5_11_8
void FASTCALL call_12_5_11_8(SH2_struct* sh)
{
  SH2movwlg(sh,(11<<4)+8);
}
#define init_12_5_11_9
void FASTCALL call_12_5_11_9(SH2_struct* sh)
{
  SH2movwlg(sh,(11<<4)+9);
}
#define init_12_5_11_10
void FASTCALL call_12_5_11_10(SH2_struct* sh)
{
  SH2movwlg(sh,(11<<4)+10);
}
#define init_12_5_11_11
void FASTCALL call_12_5_11_11(SH2_struct* sh)
{
  SH2movwlg(sh,(11<<4)+11);
}
#define init_12_5_11_12
void FASTCALL call_12_5_11_12(SH2_struct* sh)
{
  SH2movwlg(sh,(11<<4)+12);
}
#define init_12_5_11_13
void FASTCALL call_12_5_11_13(SH2_struct* sh)
{
  SH2movwlg(sh,(11<<4)+13);
}
#define init_12_5_11_14
void FASTCALL call_12_5_11_14(SH2_struct* sh)
{
  SH2movwlg(sh,(11<<4)+14);
}
#define init_12_5_11_15
void FASTCALL call_12_5_11_15(SH2_struct* sh)
{
  SH2movwlg(sh,(11<<4)+15);
}
#define init_12_5_12_0
void FASTCALL call_12_5_12_0(SH2_struct* sh)
{
  SH2movwlg(sh,(12<<4)+0);
}
#define init_12_5_12_1
void FASTCALL call_12_5_12_1(SH2_struct* sh)
{
  SH2movwlg(sh,(12<<4)+1);
}
#define init_12_5_12_2
void FASTCALL call_12_5_12_2(SH2_struct* sh)
{
  SH2movwlg(sh,(12<<4)+2);
}
#define init_12_5_12_3
void FASTCALL call_12_5_12_3(SH2_struct* sh)
{
  SH2movwlg(sh,(12<<4)+3);
}
#define init_12_5_12_4
void FASTCALL call_12_5_12_4(SH2_struct* sh)
{
  SH2movwlg(sh,(12<<4)+4);
}
#define init_12_5_12_5
void FASTCALL call_12_5_12_5(SH2_struct* sh)
{
  SH2movwlg(sh,(12<<4)+5);
}
#define init_12_5_12_6
void FASTCALL call_12_5_12_6(SH2_struct* sh)
{
  SH2movwlg(sh,(12<<4)+6);
}
#define init_12_5_12_7
void FASTCALL call_12_5_12_7(SH2_struct* sh)
{
  SH2movwlg(sh,(12<<4)+7);
}
#define init_12_5_12_8
void FASTCALL call_12_5_12_8(SH2_struct* sh)
{
  SH2movwlg(sh,(12<<4)+8);
}
#define init_12_5_12_9
void FASTCALL call_12_5_12_9(SH2_struct* sh)
{
  SH2movwlg(sh,(12<<4)+9);
}
#define init_12_5_12_10
void FASTCALL call_12_5_12_10(SH2_struct* sh)
{
  SH2movwlg(sh,(12<<4)+10);
}
#define init_12_5_12_11
void FASTCALL call_12_5_12_11(SH2_struct* sh)
{
  SH2movwlg(sh,(12<<4)+11);
}
#define init_12_5_12_12
void FASTCALL call_12_5_12_12(SH2_struct* sh)
{
  SH2movwlg(sh,(12<<4)+12);
}
#define init_12_5_12_13
void FASTCALL call_12_5_12_13(SH2_struct* sh)
{
  SH2movwlg(sh,(12<<4)+13);
}
#define init_12_5_12_14
void FASTCALL call_12_5_12_14(SH2_struct* sh)
{
  SH2movwlg(sh,(12<<4)+14);
}
#define init_12_5_12_15
void FASTCALL call_12_5_12_15(SH2_struct* sh)
{
  SH2movwlg(sh,(12<<4)+15);
}
#define init_12_5_13_0
void FASTCALL call_12_5_13_0(SH2_struct* sh)
{
  SH2movwlg(sh,(13<<4)+0);
}
#define init_12_5_13_1
void FASTCALL call_12_5_13_1(SH2_struct* sh)
{
  SH2movwlg(sh,(13<<4)+1);
}
#define init_12_5_13_2
void FASTCALL call_12_5_13_2(SH2_struct* sh)
{
  SH2movwlg(sh,(13<<4)+2);
}
#define init_12_5_13_3
void FASTCALL call_12_5_13_3(SH2_struct* sh)
{
  SH2movwlg(sh,(13<<4)+3);
}
#define init_12_5_13_4
void FASTCALL call_12_5_13_4(SH2_struct* sh)
{
  SH2movwlg(sh,(13<<4)+4);
}
#define init_12_5_13_5
void FASTCALL call_12_5_13_5(SH2_struct* sh)
{
  SH2movwlg(sh,(13<<4)+5);
}
#define init_12_5_13_6
void FASTCALL call_12_5_13_6(SH2_struct* sh)
{
  SH2movwlg(sh,(13<<4)+6);
}
#define init_12_5_13_7
void FASTCALL call_12_5_13_7(SH2_struct* sh)
{
  SH2movwlg(sh,(13<<4)+7);
}
#define init_12_5_13_8
void FASTCALL call_12_5_13_8(SH2_struct* sh)
{
  SH2movwlg(sh,(13<<4)+8);
}
#define init_12_5_13_9
void FASTCALL call_12_5_13_9(SH2_struct* sh)
{
  SH2movwlg(sh,(13<<4)+9);
}
#define init_12_5_13_10
void FASTCALL call_12_5_13_10(SH2_struct* sh)
{
  SH2movwlg(sh,(13<<4)+10);
}
#define init_12_5_13_11
void FASTCALL call_12_5_13_11(SH2_struct* sh)
{
  SH2movwlg(sh,(13<<4)+11);
}
#define init_12_5_13_12
void FASTCALL call_12_5_13_12(SH2_struct* sh)
{
  SH2movwlg(sh,(13<<4)+12);
}
#define init_12_5_13_13
void FASTCALL call_12_5_13_13(SH2_struct* sh)
{
  SH2movwlg(sh,(13<<4)+13);
}
#define init_12_5_13_14
void FASTCALL call_12_5_13_14(SH2_struct* sh)
{
  SH2movwlg(sh,(13<<4)+14);
}
#define init_12_5_13_15
void FASTCALL call_12_5_13_15(SH2_struct* sh)
{
  SH2movwlg(sh,(13<<4)+15);
}
#define init_12_5_14_0
void FASTCALL call_12_5_14_0(SH2_struct* sh)
{
  SH2movwlg(sh,(14<<4)+0);
}
#define init_12_5_14_1
void FASTCALL call_12_5_14_1(SH2_struct* sh)
{
  SH2movwlg(sh,(14<<4)+1);
}
#define init_12_5_14_2
void FASTCALL call_12_5_14_2(SH2_struct* sh)
{
  SH2movwlg(sh,(14<<4)+2);
}
#define init_12_5_14_3
void FASTCALL call_12_5_14_3(SH2_struct* sh)
{
  SH2movwlg(sh,(14<<4)+3);
}
#define init_12_5_14_4
void FASTCALL call_12_5_14_4(SH2_struct* sh)
{
  SH2movwlg(sh,(14<<4)+4);
}
#define init_12_5_14_5
void FASTCALL call_12_5_14_5(SH2_struct* sh)
{
  SH2movwlg(sh,(14<<4)+5);
}
#define init_12_5_14_6
void FASTCALL call_12_5_14_6(SH2_struct* sh)
{
  SH2movwlg(sh,(14<<4)+6);
}
#define init_12_5_14_7
void FASTCALL call_12_5_14_7(SH2_struct* sh)
{
  SH2movwlg(sh,(14<<4)+7);
}
#define init_12_5_14_8
void FASTCALL call_12_5_14_8(SH2_struct* sh)
{
  SH2movwlg(sh,(14<<4)+8);
}
#define init_12_5_14_9
void FASTCALL call_12_5_14_9(SH2_struct* sh)
{
  SH2movwlg(sh,(14<<4)+9);
}
#define init_12_5_14_10
void FASTCALL call_12_5_14_10(SH2_struct* sh)
{
  SH2movwlg(sh,(14<<4)+10);
}
#define init_12_5_14_11
void FASTCALL call_12_5_14_11(SH2_struct* sh)
{
  SH2movwlg(sh,(14<<4)+11);
}
#define init_12_5_14_12
void FASTCALL call_12_5_14_12(SH2_struct* sh)
{
  SH2movwlg(sh,(14<<4)+12);
}
#define init_12_5_14_13
void FASTCALL call_12_5_14_13(SH2_struct* sh)
{
  SH2movwlg(sh,(14<<4)+13);
}
#define init_12_5_14_14
void FASTCALL call_12_5_14_14(SH2_struct* sh)
{
  SH2movwlg(sh,(14<<4)+14);
}
#define init_12_5_14_15
void FASTCALL call_12_5_14_15(SH2_struct* sh)
{
  SH2movwlg(sh,(14<<4)+15);
}
#define init_12_5_15_0
void FASTCALL call_12_5_15_0(SH2_struct* sh)
{
  SH2movwlg(sh,(15<<4)+0);
}
#define init_12_5_15_1
void FASTCALL call_12_5_15_1(SH2_struct* sh)
{
  SH2movwlg(sh,(15<<4)+1);
}
#define init_12_5_15_2
void FASTCALL call_12_5_15_2(SH2_struct* sh)
{
  SH2movwlg(sh,(15<<4)+2);
}
#define init_12_5_15_3
void FASTCALL call_12_5_15_3(SH2_struct* sh)
{
  SH2movwlg(sh,(15<<4)+3);
}
#define init_12_5_15_4
void FASTCALL call_12_5_15_4(SH2_struct* sh)
{
  SH2movwlg(sh,(15<<4)+4);
}
#define init_12_5_15_5
void FASTCALL call_12_5_15_5(SH2_struct* sh)
{
  SH2movwlg(sh,(15<<4)+5);
}
#define init_12_5_15_6
void FASTCALL call_12_5_15_6(SH2_struct* sh)
{
  SH2movwlg(sh,(15<<4)+6);
}
#define init_12_5_15_7
void FASTCALL call_12_5_15_7(SH2_struct* sh)
{
  SH2movwlg(sh,(15<<4)+7);
}
#define init_12_5_15_8
void FASTCALL call_12_5_15_8(SH2_struct* sh)
{
  SH2movwlg(sh,(15<<4)+8);
}
#define init_12_5_15_9
void FASTCALL call_12_5_15_9(SH2_struct* sh)
{
  SH2movwlg(sh,(15<<4)+9);
}
#define init_12_5_15_10
void FASTCALL call_12_5_15_10(SH2_struct* sh)
{
  SH2movwlg(sh,(15<<4)+10);
}
#define init_12_5_15_11
void FASTCALL call_12_5_15_11(SH2_struct* sh)
{
  SH2movwlg(sh,(15<<4)+11);
}
#define init_12_5_15_12
void FASTCALL call_12_5_15_12(SH2_struct* sh)
{
  SH2movwlg(sh,(15<<4)+12);
}
#define init_12_5_15_13
void FASTCALL call_12_5_15_13(SH2_struct* sh)
{
  SH2movwlg(sh,(15<<4)+13);
}
#define init_12_5_15_14
void FASTCALL call_12_5_15_14(SH2_struct* sh)
{
  SH2movwlg(sh,(15<<4)+14);
}
#define init_12_5_15_15
void FASTCALL call_12_5_15_15(SH2_struct* sh)
{
  SH2movwlg(sh,(15<<4)+15);
}
#define init_8_5_0_0
void FASTCALL call_8_5_0_0(SH2_struct* sh)
{
  SH2movwl4(sh,0,0);
}
#define init_8_5_0_1
void FASTCALL call_8_5_0_1(SH2_struct* sh)
{
  SH2movwl4(sh,0,1);
}
#define init_8_5_0_2
void FASTCALL call_8_5_0_2(SH2_struct* sh)
{
  SH2movwl4(sh,0,2);
}
#define init_8_5_0_3
void FASTCALL call_8_5_0_3(SH2_struct* sh)
{
  SH2movwl4(sh,0,3);
}
#define init_8_5_0_4
void FASTCALL call_8_5_0_4(SH2_struct* sh)
{
  SH2movwl4(sh,0,4);
}
#define init_8_5_0_5
void FASTCALL call_8_5_0_5(SH2_struct* sh)
{
  SH2movwl4(sh,0,5);
}
#define init_8_5_0_6
void FASTCALL call_8_5_0_6(SH2_struct* sh)
{
  SH2movwl4(sh,0,6);
}
#define init_8_5_0_7
void FASTCALL call_8_5_0_7(SH2_struct* sh)
{
  SH2movwl4(sh,0,7);
}
#define init_8_5_0_8
void FASTCALL call_8_5_0_8(SH2_struct* sh)
{
  SH2movwl4(sh,0,8);
}
#define init_8_5_0_9
void FASTCALL call_8_5_0_9(SH2_struct* sh)
{
  SH2movwl4(sh,0,9);
}
#define init_8_5_0_10
void FASTCALL call_8_5_0_10(SH2_struct* sh)
{
  SH2movwl4(sh,0,10);
}
#define init_8_5_0_11
void FASTCALL call_8_5_0_11(SH2_struct* sh)
{
  SH2movwl4(sh,0,11);
}
#define init_8_5_0_12
void FASTCALL call_8_5_0_12(SH2_struct* sh)
{
  SH2movwl4(sh,0,12);
}
#define init_8_5_0_13
void FASTCALL call_8_5_0_13(SH2_struct* sh)
{
  SH2movwl4(sh,0,13);
}
#define init_8_5_0_14
void FASTCALL call_8_5_0_14(SH2_struct* sh)
{
  SH2movwl4(sh,0,14);
}
#define init_8_5_0_15
void FASTCALL call_8_5_0_15(SH2_struct* sh)
{
  SH2movwl4(sh,0,15);
}
#define init_8_5_1_0
void FASTCALL call_8_5_1_0(SH2_struct* sh)
{
  SH2movwl4(sh,1,0);
}
#define init_8_5_1_1
void FASTCALL call_8_5_1_1(SH2_struct* sh)
{
  SH2movwl4(sh,1,1);
}
#define init_8_5_1_2
void FASTCALL call_8_5_1_2(SH2_struct* sh)
{
  SH2movwl4(sh,1,2);
}
#define init_8_5_1_3
void FASTCALL call_8_5_1_3(SH2_struct* sh)
{
  SH2movwl4(sh,1,3);
}
#define init_8_5_1_4
void FASTCALL call_8_5_1_4(SH2_struct* sh)
{
  SH2movwl4(sh,1,4);
}
#define init_8_5_1_5
void FASTCALL call_8_5_1_5(SH2_struct* sh)
{
  SH2movwl4(sh,1,5);
}
#define init_8_5_1_6
void FASTCALL call_8_5_1_6(SH2_struct* sh)
{
  SH2movwl4(sh,1,6);
}
#define init_8_5_1_7
void FASTCALL call_8_5_1_7(SH2_struct* sh)
{
  SH2movwl4(sh,1,7);
}
#define init_8_5_1_8
void FASTCALL call_8_5_1_8(SH2_struct* sh)
{
  SH2movwl4(sh,1,8);
}
#define init_8_5_1_9
void FASTCALL call_8_5_1_9(SH2_struct* sh)
{
  SH2movwl4(sh,1,9);
}
#define init_8_5_1_10
void FASTCALL call_8_5_1_10(SH2_struct* sh)
{
  SH2movwl4(sh,1,10);
}
#define init_8_5_1_11
void FASTCALL call_8_5_1_11(SH2_struct* sh)
{
  SH2movwl4(sh,1,11);
}
#define init_8_5_1_12
void FASTCALL call_8_5_1_12(SH2_struct* sh)
{
  SH2movwl4(sh,1,12);
}
#define init_8_5_1_13
void FASTCALL call_8_5_1_13(SH2_struct* sh)
{
  SH2movwl4(sh,1,13);
}
#define init_8_5_1_14
void FASTCALL call_8_5_1_14(SH2_struct* sh)
{
  SH2movwl4(sh,1,14);
}
#define init_8_5_1_15
void FASTCALL call_8_5_1_15(SH2_struct* sh)
{
  SH2movwl4(sh,1,15);
}
#define init_8_5_2_0
void FASTCALL call_8_5_2_0(SH2_struct* sh)
{
  SH2movwl4(sh,2,0);
}
#define init_8_5_2_1
void FASTCALL call_8_5_2_1(SH2_struct* sh)
{
  SH2movwl4(sh,2,1);
}
#define init_8_5_2_2
void FASTCALL call_8_5_2_2(SH2_struct* sh)
{
  SH2movwl4(sh,2,2);
}
#define init_8_5_2_3
void FASTCALL call_8_5_2_3(SH2_struct* sh)
{
  SH2movwl4(sh,2,3);
}
#define init_8_5_2_4
void FASTCALL call_8_5_2_4(SH2_struct* sh)
{
  SH2movwl4(sh,2,4);
}
#define init_8_5_2_5
void FASTCALL call_8_5_2_5(SH2_struct* sh)
{
  SH2movwl4(sh,2,5);
}
#define init_8_5_2_6
void FASTCALL call_8_5_2_6(SH2_struct* sh)
{
  SH2movwl4(sh,2,6);
}
#define init_8_5_2_7
void FASTCALL call_8_5_2_7(SH2_struct* sh)
{
  SH2movwl4(sh,2,7);
}
#define init_8_5_2_8
void FASTCALL call_8_5_2_8(SH2_struct* sh)
{
  SH2movwl4(sh,2,8);
}
#define init_8_5_2_9
void FASTCALL call_8_5_2_9(SH2_struct* sh)
{
  SH2movwl4(sh,2,9);
}
#define init_8_5_2_10
void FASTCALL call_8_5_2_10(SH2_struct* sh)
{
  SH2movwl4(sh,2,10);
}
#define init_8_5_2_11
void FASTCALL call_8_5_2_11(SH2_struct* sh)
{
  SH2movwl4(sh,2,11);
}
#define init_8_5_2_12
void FASTCALL call_8_5_2_12(SH2_struct* sh)
{
  SH2movwl4(sh,2,12);
}
#define init_8_5_2_13
void FASTCALL call_8_5_2_13(SH2_struct* sh)
{
  SH2movwl4(sh,2,13);
}
#define init_8_5_2_14
void FASTCALL call_8_5_2_14(SH2_struct* sh)
{
  SH2movwl4(sh,2,14);
}
#define init_8_5_2_15
void FASTCALL call_8_5_2_15(SH2_struct* sh)
{
  SH2movwl4(sh,2,15);
}
#define init_8_5_3_0
void FASTCALL call_8_5_3_0(SH2_struct* sh)
{
  SH2movwl4(sh,3,0);
}
#define init_8_5_3_1
void FASTCALL call_8_5_3_1(SH2_struct* sh)
{
  SH2movwl4(sh,3,1);
}
#define init_8_5_3_2
void FASTCALL call_8_5_3_2(SH2_struct* sh)
{
  SH2movwl4(sh,3,2);
}
#define init_8_5_3_3
void FASTCALL call_8_5_3_3(SH2_struct* sh)
{
  SH2movwl4(sh,3,3);
}
#define init_8_5_3_4
void FASTCALL call_8_5_3_4(SH2_struct* sh)
{
  SH2movwl4(sh,3,4);
}
#define init_8_5_3_5
void FASTCALL call_8_5_3_5(SH2_struct* sh)
{
  SH2movwl4(sh,3,5);
}
#define init_8_5_3_6
void FASTCALL call_8_5_3_6(SH2_struct* sh)
{
  SH2movwl4(sh,3,6);
}
#define init_8_5_3_7
void FASTCALL call_8_5_3_7(SH2_struct* sh)
{
  SH2movwl4(sh,3,7);
}
#define init_8_5_3_8
void FASTCALL call_8_5_3_8(SH2_struct* sh)
{
  SH2movwl4(sh,3,8);
}
#define init_8_5_3_9
void FASTCALL call_8_5_3_9(SH2_struct* sh)
{
  SH2movwl4(sh,3,9);
}
#define init_8_5_3_10
void FASTCALL call_8_5_3_10(SH2_struct* sh)
{
  SH2movwl4(sh,3,10);
}
#define init_8_5_3_11
void FASTCALL call_8_5_3_11(SH2_struct* sh)
{
  SH2movwl4(sh,3,11);
}
#define init_8_5_3_12
void FASTCALL call_8_5_3_12(SH2_struct* sh)
{
  SH2movwl4(sh,3,12);
}
#define init_8_5_3_13
void FASTCALL call_8_5_3_13(SH2_struct* sh)
{
  SH2movwl4(sh,3,13);
}
#define init_8_5_3_14
void FASTCALL call_8_5_3_14(SH2_struct* sh)
{
  SH2movwl4(sh,3,14);
}
#define init_8_5_3_15
void FASTCALL call_8_5_3_15(SH2_struct* sh)
{
  SH2movwl4(sh,3,15);
}
#define init_8_5_4_0
void FASTCALL call_8_5_4_0(SH2_struct* sh)
{
  SH2movwl4(sh,4,0);
}
#define init_8_5_4_1
void FASTCALL call_8_5_4_1(SH2_struct* sh)
{
  SH2movwl4(sh,4,1);
}
#define init_8_5_4_2
void FASTCALL call_8_5_4_2(SH2_struct* sh)
{
  SH2movwl4(sh,4,2);
}
#define init_8_5_4_3
void FASTCALL call_8_5_4_3(SH2_struct* sh)
{
  SH2movwl4(sh,4,3);
}
#define init_8_5_4_4
void FASTCALL call_8_5_4_4(SH2_struct* sh)
{
  SH2movwl4(sh,4,4);
}
#define init_8_5_4_5
void FASTCALL call_8_5_4_5(SH2_struct* sh)
{
  SH2movwl4(sh,4,5);
}
#define init_8_5_4_6
void FASTCALL call_8_5_4_6(SH2_struct* sh)
{
  SH2movwl4(sh,4,6);
}
#define init_8_5_4_7
void FASTCALL call_8_5_4_7(SH2_struct* sh)
{
  SH2movwl4(sh,4,7);
}
#define init_8_5_4_8
void FASTCALL call_8_5_4_8(SH2_struct* sh)
{
  SH2movwl4(sh,4,8);
}
#define init_8_5_4_9
void FASTCALL call_8_5_4_9(SH2_struct* sh)
{
  SH2movwl4(sh,4,9);
}
#define init_8_5_4_10
void FASTCALL call_8_5_4_10(SH2_struct* sh)
{
  SH2movwl4(sh,4,10);
}
#define init_8_5_4_11
void FASTCALL call_8_5_4_11(SH2_struct* sh)
{
  SH2movwl4(sh,4,11);
}
#define init_8_5_4_12
void FASTCALL call_8_5_4_12(SH2_struct* sh)
{
  SH2movwl4(sh,4,12);
}
#define init_8_5_4_13
void FASTCALL call_8_5_4_13(SH2_struct* sh)
{
  SH2movwl4(sh,4,13);
}
#define init_8_5_4_14
void FASTCALL call_8_5_4_14(SH2_struct* sh)
{
  SH2movwl4(sh,4,14);
}
#define init_8_5_4_15
void FASTCALL call_8_5_4_15(SH2_struct* sh)
{
  SH2movwl4(sh,4,15);
}
#define init_8_5_5_0
void FASTCALL call_8_5_5_0(SH2_struct* sh)
{
  SH2movwl4(sh,5,0);
}
#define init_8_5_5_1
void FASTCALL call_8_5_5_1(SH2_struct* sh)
{
  SH2movwl4(sh,5,1);
}
#define init_8_5_5_2
void FASTCALL call_8_5_5_2(SH2_struct* sh)
{
  SH2movwl4(sh,5,2);
}
#define init_8_5_5_3
void FASTCALL call_8_5_5_3(SH2_struct* sh)
{
  SH2movwl4(sh,5,3);
}
#define init_8_5_5_4
void FASTCALL call_8_5_5_4(SH2_struct* sh)
{
  SH2movwl4(sh,5,4);
}
#define init_8_5_5_5
void FASTCALL call_8_5_5_5(SH2_struct* sh)
{
  SH2movwl4(sh,5,5);
}
#define init_8_5_5_6
void FASTCALL call_8_5_5_6(SH2_struct* sh)
{
  SH2movwl4(sh,5,6);
}
#define init_8_5_5_7
void FASTCALL call_8_5_5_7(SH2_struct* sh)
{
  SH2movwl4(sh,5,7);
}
#define init_8_5_5_8
void FASTCALL call_8_5_5_8(SH2_struct* sh)
{
  SH2movwl4(sh,5,8);
}
#define init_8_5_5_9
void FASTCALL call_8_5_5_9(SH2_struct* sh)
{
  SH2movwl4(sh,5,9);
}
#define init_8_5_5_10
void FASTCALL call_8_5_5_10(SH2_struct* sh)
{
  SH2movwl4(sh,5,10);
}
#define init_8_5_5_11
void FASTCALL call_8_5_5_11(SH2_struct* sh)
{
  SH2movwl4(sh,5,11);
}
#define init_8_5_5_12
void FASTCALL call_8_5_5_12(SH2_struct* sh)
{
  SH2movwl4(sh,5,12);
}
#define init_8_5_5_13
void FASTCALL call_8_5_5_13(SH2_struct* sh)
{
  SH2movwl4(sh,5,13);
}
#define init_8_5_5_14
void FASTCALL call_8_5_5_14(SH2_struct* sh)
{
  SH2movwl4(sh,5,14);
}
#define init_8_5_5_15
void FASTCALL call_8_5_5_15(SH2_struct* sh)
{
  SH2movwl4(sh,5,15);
}
#define init_8_5_6_0
void FASTCALL call_8_5_6_0(SH2_struct* sh)
{
  SH2movwl4(sh,6,0);
}
#define init_8_5_6_1
void FASTCALL call_8_5_6_1(SH2_struct* sh)
{
  SH2movwl4(sh,6,1);
}
#define init_8_5_6_2
void FASTCALL call_8_5_6_2(SH2_struct* sh)
{
  SH2movwl4(sh,6,2);
}
#define init_8_5_6_3
void FASTCALL call_8_5_6_3(SH2_struct* sh)
{
  SH2movwl4(sh,6,3);
}
#define init_8_5_6_4
void FASTCALL call_8_5_6_4(SH2_struct* sh)
{
  SH2movwl4(sh,6,4);
}
#define init_8_5_6_5
void FASTCALL call_8_5_6_5(SH2_struct* sh)
{
  SH2movwl4(sh,6,5);
}
#define init_8_5_6_6
void FASTCALL call_8_5_6_6(SH2_struct* sh)
{
  SH2movwl4(sh,6,6);
}
#define init_8_5_6_7
void FASTCALL call_8_5_6_7(SH2_struct* sh)
{
  SH2movwl4(sh,6,7);
}
#define init_8_5_6_8
void FASTCALL call_8_5_6_8(SH2_struct* sh)
{
  SH2movwl4(sh,6,8);
}
#define init_8_5_6_9
void FASTCALL call_8_5_6_9(SH2_struct* sh)
{
  SH2movwl4(sh,6,9);
}
#define init_8_5_6_10
void FASTCALL call_8_5_6_10(SH2_struct* sh)
{
  SH2movwl4(sh,6,10);
}
#define init_8_5_6_11
void FASTCALL call_8_5_6_11(SH2_struct* sh)
{
  SH2movwl4(sh,6,11);
}
#define init_8_5_6_12
void FASTCALL call_8_5_6_12(SH2_struct* sh)
{
  SH2movwl4(sh,6,12);
}
#define init_8_5_6_13
void FASTCALL call_8_5_6_13(SH2_struct* sh)
{
  SH2movwl4(sh,6,13);
}
#define init_8_5_6_14
void FASTCALL call_8_5_6_14(SH2_struct* sh)
{
  SH2movwl4(sh,6,14);
}
#define init_8_5_6_15
void FASTCALL call_8_5_6_15(SH2_struct* sh)
{
  SH2movwl4(sh,6,15);
}
#define init_8_5_7_0
void FASTCALL call_8_5_7_0(SH2_struct* sh)
{
  SH2movwl4(sh,7,0);
}
#define init_8_5_7_1
void FASTCALL call_8_5_7_1(SH2_struct* sh)
{
  SH2movwl4(sh,7,1);
}
#define init_8_5_7_2
void FASTCALL call_8_5_7_2(SH2_struct* sh)
{
  SH2movwl4(sh,7,2);
}
#define init_8_5_7_3
void FASTCALL call_8_5_7_3(SH2_struct* sh)
{
  SH2movwl4(sh,7,3);
}
#define init_8_5_7_4
void FASTCALL call_8_5_7_4(SH2_struct* sh)
{
  SH2movwl4(sh,7,4);
}
#define init_8_5_7_5
void FASTCALL call_8_5_7_5(SH2_struct* sh)
{
  SH2movwl4(sh,7,5);
}
#define init_8_5_7_6
void FASTCALL call_8_5_7_6(SH2_struct* sh)
{
  SH2movwl4(sh,7,6);
}
#define init_8_5_7_7
void FASTCALL call_8_5_7_7(SH2_struct* sh)
{
  SH2movwl4(sh,7,7);
}
#define init_8_5_7_8
void FASTCALL call_8_5_7_8(SH2_struct* sh)
{
  SH2movwl4(sh,7,8);
}
#define init_8_5_7_9
void FASTCALL call_8_5_7_9(SH2_struct* sh)
{
  SH2movwl4(sh,7,9);
}
#define init_8_5_7_10
void FASTCALL call_8_5_7_10(SH2_struct* sh)
{
  SH2movwl4(sh,7,10);
}
#define init_8_5_7_11
void FASTCALL call_8_5_7_11(SH2_struct* sh)
{
  SH2movwl4(sh,7,11);
}
#define init_8_5_7_12
void FASTCALL call_8_5_7_12(SH2_struct* sh)
{
  SH2movwl4(sh,7,12);
}
#define init_8_5_7_13
void FASTCALL call_8_5_7_13(SH2_struct* sh)
{
  SH2movwl4(sh,7,13);
}
#define init_8_5_7_14
void FASTCALL call_8_5_7_14(SH2_struct* sh)
{
  SH2movwl4(sh,7,14);
}
#define init_8_5_7_15
void FASTCALL call_8_5_7_15(SH2_struct* sh)
{
  SH2movwl4(sh,7,15);
}
#define init_8_5_8_0
void FASTCALL call_8_5_8_0(SH2_struct* sh)
{
  SH2movwl4(sh,8,0);
}
#define init_8_5_8_1
void FASTCALL call_8_5_8_1(SH2_struct* sh)
{
  SH2movwl4(sh,8,1);
}
#define init_8_5_8_2
void FASTCALL call_8_5_8_2(SH2_struct* sh)
{
  SH2movwl4(sh,8,2);
}
#define init_8_5_8_3
void FASTCALL call_8_5_8_3(SH2_struct* sh)
{
  SH2movwl4(sh,8,3);
}
#define init_8_5_8_4
void FASTCALL call_8_5_8_4(SH2_struct* sh)
{
  SH2movwl4(sh,8,4);
}
#define init_8_5_8_5
void FASTCALL call_8_5_8_5(SH2_struct* sh)
{
  SH2movwl4(sh,8,5);
}
#define init_8_5_8_6
void FASTCALL call_8_5_8_6(SH2_struct* sh)
{
  SH2movwl4(sh,8,6);
}
#define init_8_5_8_7
void FASTCALL call_8_5_8_7(SH2_struct* sh)
{
  SH2movwl4(sh,8,7);
}
#define init_8_5_8_8
void FASTCALL call_8_5_8_8(SH2_struct* sh)
{
  SH2movwl4(sh,8,8);
}
#define init_8_5_8_9
void FASTCALL call_8_5_8_9(SH2_struct* sh)
{
  SH2movwl4(sh,8,9);
}
#define init_8_5_8_10
void FASTCALL call_8_5_8_10(SH2_struct* sh)
{
  SH2movwl4(sh,8,10);
}
#define init_8_5_8_11
void FASTCALL call_8_5_8_11(SH2_struct* sh)
{
  SH2movwl4(sh,8,11);
}
#define init_8_5_8_12
void FASTCALL call_8_5_8_12(SH2_struct* sh)
{
  SH2movwl4(sh,8,12);
}
#define init_8_5_8_13
void FASTCALL call_8_5_8_13(SH2_struct* sh)
{
  SH2movwl4(sh,8,13);
}
#define init_8_5_8_14
void FASTCALL call_8_5_8_14(SH2_struct* sh)
{
  SH2movwl4(sh,8,14);
}
#define init_8_5_8_15
void FASTCALL call_8_5_8_15(SH2_struct* sh)
{
  SH2movwl4(sh,8,15);
}
#define init_8_5_9_0
void FASTCALL call_8_5_9_0(SH2_struct* sh)
{
  SH2movwl4(sh,9,0);
}
#define init_8_5_9_1
void FASTCALL call_8_5_9_1(SH2_struct* sh)
{
  SH2movwl4(sh,9,1);
}
#define init_8_5_9_2
void FASTCALL call_8_5_9_2(SH2_struct* sh)
{
  SH2movwl4(sh,9,2);
}
#define init_8_5_9_3
void FASTCALL call_8_5_9_3(SH2_struct* sh)
{
  SH2movwl4(sh,9,3);
}
#define init_8_5_9_4
void FASTCALL call_8_5_9_4(SH2_struct* sh)
{
  SH2movwl4(sh,9,4);
}
#define init_8_5_9_5
void FASTCALL call_8_5_9_5(SH2_struct* sh)
{
  SH2movwl4(sh,9,5);
}
#define init_8_5_9_6
void FASTCALL call_8_5_9_6(SH2_struct* sh)
{
  SH2movwl4(sh,9,6);
}
#define init_8_5_9_7
void FASTCALL call_8_5_9_7(SH2_struct* sh)
{
  SH2movwl4(sh,9,7);
}
#define init_8_5_9_8
void FASTCALL call_8_5_9_8(SH2_struct* sh)
{
  SH2movwl4(sh,9,8);
}
#define init_8_5_9_9
void FASTCALL call_8_5_9_9(SH2_struct* sh)
{
  SH2movwl4(sh,9,9);
}
#define init_8_5_9_10
void FASTCALL call_8_5_9_10(SH2_struct* sh)
{
  SH2movwl4(sh,9,10);
}
#define init_8_5_9_11
void FASTCALL call_8_5_9_11(SH2_struct* sh)
{
  SH2movwl4(sh,9,11);
}
#define init_8_5_9_12
void FASTCALL call_8_5_9_12(SH2_struct* sh)
{
  SH2movwl4(sh,9,12);
}
#define init_8_5_9_13
void FASTCALL call_8_5_9_13(SH2_struct* sh)
{
  SH2movwl4(sh,9,13);
}
#define init_8_5_9_14
void FASTCALL call_8_5_9_14(SH2_struct* sh)
{
  SH2movwl4(sh,9,14);
}
#define init_8_5_9_15
void FASTCALL call_8_5_9_15(SH2_struct* sh)
{
  SH2movwl4(sh,9,15);
}
#define init_8_5_10_0
void FASTCALL call_8_5_10_0(SH2_struct* sh)
{
  SH2movwl4(sh,10,0);
}
#define init_8_5_10_1
void FASTCALL call_8_5_10_1(SH2_struct* sh)
{
  SH2movwl4(sh,10,1);
}
#define init_8_5_10_2
void FASTCALL call_8_5_10_2(SH2_struct* sh)
{
  SH2movwl4(sh,10,2);
}
#define init_8_5_10_3
void FASTCALL call_8_5_10_3(SH2_struct* sh)
{
  SH2movwl4(sh,10,3);
}
#define init_8_5_10_4
void FASTCALL call_8_5_10_4(SH2_struct* sh)
{
  SH2movwl4(sh,10,4);
}
#define init_8_5_10_5
void FASTCALL call_8_5_10_5(SH2_struct* sh)
{
  SH2movwl4(sh,10,5);
}
#define init_8_5_10_6
void FASTCALL call_8_5_10_6(SH2_struct* sh)
{
  SH2movwl4(sh,10,6);
}
#define init_8_5_10_7
void FASTCALL call_8_5_10_7(SH2_struct* sh)
{
  SH2movwl4(sh,10,7);
}
#define init_8_5_10_8
void FASTCALL call_8_5_10_8(SH2_struct* sh)
{
  SH2movwl4(sh,10,8);
}
#define init_8_5_10_9
void FASTCALL call_8_5_10_9(SH2_struct* sh)
{
  SH2movwl4(sh,10,9);
}
#define init_8_5_10_10
void FASTCALL call_8_5_10_10(SH2_struct* sh)
{
  SH2movwl4(sh,10,10);
}
#define init_8_5_10_11
void FASTCALL call_8_5_10_11(SH2_struct* sh)
{
  SH2movwl4(sh,10,11);
}
#define init_8_5_10_12
void FASTCALL call_8_5_10_12(SH2_struct* sh)
{
  SH2movwl4(sh,10,12);
}
#define init_8_5_10_13
void FASTCALL call_8_5_10_13(SH2_struct* sh)
{
  SH2movwl4(sh,10,13);
}
#define init_8_5_10_14
void FASTCALL call_8_5_10_14(SH2_struct* sh)
{
  SH2movwl4(sh,10,14);
}
#define init_8_5_10_15
void FASTCALL call_8_5_10_15(SH2_struct* sh)
{
  SH2movwl4(sh,10,15);
}
#define init_8_5_11_0
void FASTCALL call_8_5_11_0(SH2_struct* sh)
{
  SH2movwl4(sh,11,0);
}
#define init_8_5_11_1
void FASTCALL call_8_5_11_1(SH2_struct* sh)
{
  SH2movwl4(sh,11,1);
}
#define init_8_5_11_2
void FASTCALL call_8_5_11_2(SH2_struct* sh)
{
  SH2movwl4(sh,11,2);
}
#define init_8_5_11_3
void FASTCALL call_8_5_11_3(SH2_struct* sh)
{
  SH2movwl4(sh,11,3);
}
#define init_8_5_11_4
void FASTCALL call_8_5_11_4(SH2_struct* sh)
{
  SH2movwl4(sh,11,4);
}
#define init_8_5_11_5
void FASTCALL call_8_5_11_5(SH2_struct* sh)
{
  SH2movwl4(sh,11,5);
}
#define init_8_5_11_6
void FASTCALL call_8_5_11_6(SH2_struct* sh)
{
  SH2movwl4(sh,11,6);
}
#define init_8_5_11_7
void FASTCALL call_8_5_11_7(SH2_struct* sh)
{
  SH2movwl4(sh,11,7);
}
#define init_8_5_11_8
void FASTCALL call_8_5_11_8(SH2_struct* sh)
{
  SH2movwl4(sh,11,8);
}
#define init_8_5_11_9
void FASTCALL call_8_5_11_9(SH2_struct* sh)
{
  SH2movwl4(sh,11,9);
}
#define init_8_5_11_10
void FASTCALL call_8_5_11_10(SH2_struct* sh)
{
  SH2movwl4(sh,11,10);
}
#define init_8_5_11_11
void FASTCALL call_8_5_11_11(SH2_struct* sh)
{
  SH2movwl4(sh,11,11);
}
#define init_8_5_11_12
void FASTCALL call_8_5_11_12(SH2_struct* sh)
{
  SH2movwl4(sh,11,12);
}
#define init_8_5_11_13
void FASTCALL call_8_5_11_13(SH2_struct* sh)
{
  SH2movwl4(sh,11,13);
}
#define init_8_5_11_14
void FASTCALL call_8_5_11_14(SH2_struct* sh)
{
  SH2movwl4(sh,11,14);
}
#define init_8_5_11_15
void FASTCALL call_8_5_11_15(SH2_struct* sh)
{
  SH2movwl4(sh,11,15);
}
#define init_8_5_12_0
void FASTCALL call_8_5_12_0(SH2_struct* sh)
{
  SH2movwl4(sh,12,0);
}
#define init_8_5_12_1
void FASTCALL call_8_5_12_1(SH2_struct* sh)
{
  SH2movwl4(sh,12,1);
}
#define init_8_5_12_2
void FASTCALL call_8_5_12_2(SH2_struct* sh)
{
  SH2movwl4(sh,12,2);
}
#define init_8_5_12_3
void FASTCALL call_8_5_12_3(SH2_struct* sh)
{
  SH2movwl4(sh,12,3);
}
#define init_8_5_12_4
void FASTCALL call_8_5_12_4(SH2_struct* sh)
{
  SH2movwl4(sh,12,4);
}
#define init_8_5_12_5
void FASTCALL call_8_5_12_5(SH2_struct* sh)
{
  SH2movwl4(sh,12,5);
}
#define init_8_5_12_6
void FASTCALL call_8_5_12_6(SH2_struct* sh)
{
  SH2movwl4(sh,12,6);
}
#define init_8_5_12_7
void FASTCALL call_8_5_12_7(SH2_struct* sh)
{
  SH2movwl4(sh,12,7);
}
#define init_8_5_12_8
void FASTCALL call_8_5_12_8(SH2_struct* sh)
{
  SH2movwl4(sh,12,8);
}
#define init_8_5_12_9
void FASTCALL call_8_5_12_9(SH2_struct* sh)
{
  SH2movwl4(sh,12,9);
}
#define init_8_5_12_10
void FASTCALL call_8_5_12_10(SH2_struct* sh)
{
  SH2movwl4(sh,12,10);
}
#define init_8_5_12_11
void FASTCALL call_8_5_12_11(SH2_struct* sh)
{
  SH2movwl4(sh,12,11);
}
#define init_8_5_12_12
void FASTCALL call_8_5_12_12(SH2_struct* sh)
{
  SH2movwl4(sh,12,12);
}
#define init_8_5_12_13
void FASTCALL call_8_5_12_13(SH2_struct* sh)
{
  SH2movwl4(sh,12,13);
}
#define init_8_5_12_14
void FASTCALL call_8_5_12_14(SH2_struct* sh)
{
  SH2movwl4(sh,12,14);
}
#define init_8_5_12_15
void FASTCALL call_8_5_12_15(SH2_struct* sh)
{
  SH2movwl4(sh,12,15);
}
#define init_8_5_13_0
void FASTCALL call_8_5_13_0(SH2_struct* sh)
{
  SH2movwl4(sh,13,0);
}
#define init_8_5_13_1
void FASTCALL call_8_5_13_1(SH2_struct* sh)
{
  SH2movwl4(sh,13,1);
}
#define init_8_5_13_2
void FASTCALL call_8_5_13_2(SH2_struct* sh)
{
  SH2movwl4(sh,13,2);
}
#define init_8_5_13_3
void FASTCALL call_8_5_13_3(SH2_struct* sh)
{
  SH2movwl4(sh,13,3);
}
#define init_8_5_13_4
void FASTCALL call_8_5_13_4(SH2_struct* sh)
{
  SH2movwl4(sh,13,4);
}
#define init_8_5_13_5
void FASTCALL call_8_5_13_5(SH2_struct* sh)
{
  SH2movwl4(sh,13,5);
}
#define init_8_5_13_6
void FASTCALL call_8_5_13_6(SH2_struct* sh)
{
  SH2movwl4(sh,13,6);
}
#define init_8_5_13_7
void FASTCALL call_8_5_13_7(SH2_struct* sh)
{
  SH2movwl4(sh,13,7);
}
#define init_8_5_13_8
void FASTCALL call_8_5_13_8(SH2_struct* sh)
{
  SH2movwl4(sh,13,8);
}
#define init_8_5_13_9
void FASTCALL call_8_5_13_9(SH2_struct* sh)
{
  SH2movwl4(sh,13,9);
}
#define init_8_5_13_10
void FASTCALL call_8_5_13_10(SH2_struct* sh)
{
  SH2movwl4(sh,13,10);
}
#define init_8_5_13_11
void FASTCALL call_8_5_13_11(SH2_struct* sh)
{
  SH2movwl4(sh,13,11);
}
#define init_8_5_13_12
void FASTCALL call_8_5_13_12(SH2_struct* sh)
{
  SH2movwl4(sh,13,12);
}
#define init_8_5_13_13
void FASTCALL call_8_5_13_13(SH2_struct* sh)
{
  SH2movwl4(sh,13,13);
}
#define init_8_5_13_14
void FASTCALL call_8_5_13_14(SH2_struct* sh)
{
  SH2movwl4(sh,13,14);
}
#define init_8_5_13_15
void FASTCALL call_8_5_13_15(SH2_struct* sh)
{
  SH2movwl4(sh,13,15);
}
#define init_8_5_14_0
void FASTCALL call_8_5_14_0(SH2_struct* sh)
{
  SH2movwl4(sh,14,0);
}
#define init_8_5_14_1
void FASTCALL call_8_5_14_1(SH2_struct* sh)
{
  SH2movwl4(sh,14,1);
}
#define init_8_5_14_2
void FASTCALL call_8_5_14_2(SH2_struct* sh)
{
  SH2movwl4(sh,14,2);
}
#define init_8_5_14_3
void FASTCALL call_8_5_14_3(SH2_struct* sh)
{
  SH2movwl4(sh,14,3);
}
#define init_8_5_14_4
void FASTCALL call_8_5_14_4(SH2_struct* sh)
{
  SH2movwl4(sh,14,4);
}
#define init_8_5_14_5
void FASTCALL call_8_5_14_5(SH2_struct* sh)
{
  SH2movwl4(sh,14,5);
}
#define init_8_5_14_6
void FASTCALL call_8_5_14_6(SH2_struct* sh)
{
  SH2movwl4(sh,14,6);
}
#define init_8_5_14_7
void FASTCALL call_8_5_14_7(SH2_struct* sh)
{
  SH2movwl4(sh,14,7);
}
#define init_8_5_14_8
void FASTCALL call_8_5_14_8(SH2_struct* sh)
{
  SH2movwl4(sh,14,8);
}
#define init_8_5_14_9
void FASTCALL call_8_5_14_9(SH2_struct* sh)
{
  SH2movwl4(sh,14,9);
}
#define init_8_5_14_10
void FASTCALL call_8_5_14_10(SH2_struct* sh)
{
  SH2movwl4(sh,14,10);
}
#define init_8_5_14_11
void FASTCALL call_8_5_14_11(SH2_struct* sh)
{
  SH2movwl4(sh,14,11);
}
#define init_8_5_14_12
void FASTCALL call_8_5_14_12(SH2_struct* sh)
{
  SH2movwl4(sh,14,12);
}
#define init_8_5_14_13
void FASTCALL call_8_5_14_13(SH2_struct* sh)
{
  SH2movwl4(sh,14,13);
}
#define init_8_5_14_14
void FASTCALL call_8_5_14_14(SH2_struct* sh)
{
  SH2movwl4(sh,14,14);
}
#define init_8_5_14_15
void FASTCALL call_8_5_14_15(SH2_struct* sh)
{
  SH2movwl4(sh,14,15);
}
#define init_8_5_15_0
void FASTCALL call_8_5_15_0(SH2_struct* sh)
{
  SH2movwl4(sh,15,0);
}
#define init_8_5_15_1
void FASTCALL call_8_5_15_1(SH2_struct* sh)
{
  SH2movwl4(sh,15,1);
}
#define init_8_5_15_2
void FASTCALL call_8_5_15_2(SH2_struct* sh)
{
  SH2movwl4(sh,15,2);
}
#define init_8_5_15_3
void FASTCALL call_8_5_15_3(SH2_struct* sh)
{
  SH2movwl4(sh,15,3);
}
#define init_8_5_15_4
void FASTCALL call_8_5_15_4(SH2_struct* sh)
{
  SH2movwl4(sh,15,4);
}
#define init_8_5_15_5
void FASTCALL call_8_5_15_5(SH2_struct* sh)
{
  SH2movwl4(sh,15,5);
}
#define init_8_5_15_6
void FASTCALL call_8_5_15_6(SH2_struct* sh)
{
  SH2movwl4(sh,15,6);
}
#define init_8_5_15_7
void FASTCALL call_8_5_15_7(SH2_struct* sh)
{
  SH2movwl4(sh,15,7);
}
#define init_8_5_15_8
void FASTCALL call_8_5_15_8(SH2_struct* sh)
{
  SH2movwl4(sh,15,8);
}
#define init_8_5_15_9
void FASTCALL call_8_5_15_9(SH2_struct* sh)
{
  SH2movwl4(sh,15,9);
}
#define init_8_5_15_10
void FASTCALL call_8_5_15_10(SH2_struct* sh)
{
  SH2movwl4(sh,15,10);
}
#define init_8_5_15_11
void FASTCALL call_8_5_15_11(SH2_struct* sh)
{
  SH2movwl4(sh,15,11);
}
#define init_8_5_15_12
void FASTCALL call_8_5_15_12(SH2_struct* sh)
{
  SH2movwl4(sh,15,12);
}
#define init_8_5_15_13
void FASTCALL call_8_5_15_13(SH2_struct* sh)
{
  SH2movwl4(sh,15,13);
}
#define init_8_5_15_14
void FASTCALL call_8_5_15_14(SH2_struct* sh)
{
  SH2movwl4(sh,15,14);
}
#define init_8_5_15_15
void FASTCALL call_8_5_15_15(SH2_struct* sh)
{
  SH2movwl4(sh,15,15);
}
#define init_6_0_0_1
void FASTCALL call_6_0_0_1(SH2_struct* sh)
{
  SH2movwl(sh,0,0);
}
#define init_6_0_1_1
void FASTCALL call_6_0_1_1(SH2_struct* sh)
{
  SH2movwl(sh,0,1);
}
#define init_6_0_2_1
void FASTCALL call_6_0_2_1(SH2_struct* sh)
{
  SH2movwl(sh,0,2);
}
#define init_6_0_3_1
void FASTCALL call_6_0_3_1(SH2_struct* sh)
{
  SH2movwl(sh,0,3);
}
#define init_6_0_4_1
void FASTCALL call_6_0_4_1(SH2_struct* sh)
{
  SH2movwl(sh,0,4);
}
#define init_6_0_5_1
void FASTCALL call_6_0_5_1(SH2_struct* sh)
{
  SH2movwl(sh,0,5);
}
#define init_6_0_6_1
void FASTCALL call_6_0_6_1(SH2_struct* sh)
{
  SH2movwl(sh,0,6);
}
#define init_6_0_7_1
void FASTCALL call_6_0_7_1(SH2_struct* sh)
{
  SH2movwl(sh,0,7);
}
#define init_6_0_8_1
void FASTCALL call_6_0_8_1(SH2_struct* sh)
{
  SH2movwl(sh,0,8);
}
#define init_6_0_9_1
void FASTCALL call_6_0_9_1(SH2_struct* sh)
{
  SH2movwl(sh,0,9);
}
#define init_6_0_10_1
void FASTCALL call_6_0_10_1(SH2_struct* sh)
{
  SH2movwl(sh,0,10);
}
#define init_6_0_11_1
void FASTCALL call_6_0_11_1(SH2_struct* sh)
{
  SH2movwl(sh,0,11);
}
#define init_6_0_12_1
void FASTCALL call_6_0_12_1(SH2_struct* sh)
{
  SH2movwl(sh,0,12);
}
#define init_6_0_13_1
void FASTCALL call_6_0_13_1(SH2_struct* sh)
{
  SH2movwl(sh,0,13);
}
#define init_6_0_14_1
void FASTCALL call_6_0_14_1(SH2_struct* sh)
{
  SH2movwl(sh,0,14);
}
#define init_6_0_15_1
void FASTCALL call_6_0_15_1(SH2_struct* sh)
{
  SH2movwl(sh,0,15);
}
#define init_6_1_0_1
void FASTCALL call_6_1_0_1(SH2_struct* sh)
{
  SH2movwl(sh,1,0);
}
#define init_6_1_1_1
void FASTCALL call_6_1_1_1(SH2_struct* sh)
{
  SH2movwl(sh,1,1);
}
#define init_6_1_2_1
void FASTCALL call_6_1_2_1(SH2_struct* sh)
{
  SH2movwl(sh,1,2);
}
#define init_6_1_3_1
void FASTCALL call_6_1_3_1(SH2_struct* sh)
{
  SH2movwl(sh,1,3);
}
#define init_6_1_4_1
void FASTCALL call_6_1_4_1(SH2_struct* sh)
{
  SH2movwl(sh,1,4);
}
#define init_6_1_5_1
void FASTCALL call_6_1_5_1(SH2_struct* sh)
{
  SH2movwl(sh,1,5);
}
#define init_6_1_6_1
void FASTCALL call_6_1_6_1(SH2_struct* sh)
{
  SH2movwl(sh,1,6);
}
#define init_6_1_7_1
void FASTCALL call_6_1_7_1(SH2_struct* sh)
{
  SH2movwl(sh,1,7);
}
#define init_6_1_8_1
void FASTCALL call_6_1_8_1(SH2_struct* sh)
{
  SH2movwl(sh,1,8);
}
#define init_6_1_9_1
void FASTCALL call_6_1_9_1(SH2_struct* sh)
{
  SH2movwl(sh,1,9);
}
#define init_6_1_10_1
void FASTCALL call_6_1_10_1(SH2_struct* sh)
{
  SH2movwl(sh,1,10);
}
#define init_6_1_11_1
void FASTCALL call_6_1_11_1(SH2_struct* sh)
{
  SH2movwl(sh,1,11);
}
#define init_6_1_12_1
void FASTCALL call_6_1_12_1(SH2_struct* sh)
{
  SH2movwl(sh,1,12);
}
#define init_6_1_13_1
void FASTCALL call_6_1_13_1(SH2_struct* sh)
{
  SH2movwl(sh,1,13);
}
#define init_6_1_14_1
void FASTCALL call_6_1_14_1(SH2_struct* sh)
{
  SH2movwl(sh,1,14);
}
#define init_6_1_15_1
void FASTCALL call_6_1_15_1(SH2_struct* sh)
{
  SH2movwl(sh,1,15);
}
#define init_6_2_0_1
void FASTCALL call_6_2_0_1(SH2_struct* sh)
{
  SH2movwl(sh,2,0);
}
#define init_6_2_1_1
void FASTCALL call_6_2_1_1(SH2_struct* sh)
{
  SH2movwl(sh,2,1);
}
#define init_6_2_2_1
void FASTCALL call_6_2_2_1(SH2_struct* sh)
{
  SH2movwl(sh,2,2);
}
#define init_6_2_3_1
void FASTCALL call_6_2_3_1(SH2_struct* sh)
{
  SH2movwl(sh,2,3);
}
#define init_6_2_4_1
void FASTCALL call_6_2_4_1(SH2_struct* sh)
{
  SH2movwl(sh,2,4);
}
#define init_6_2_5_1
void FASTCALL call_6_2_5_1(SH2_struct* sh)
{
  SH2movwl(sh,2,5);
}
#define init_6_2_6_1
void FASTCALL call_6_2_6_1(SH2_struct* sh)
{
  SH2movwl(sh,2,6);
}
#define init_6_2_7_1
void FASTCALL call_6_2_7_1(SH2_struct* sh)
{
  SH2movwl(sh,2,7);
}
#define init_6_2_8_1
void FASTCALL call_6_2_8_1(SH2_struct* sh)
{
  SH2movwl(sh,2,8);
}
#define init_6_2_9_1
void FASTCALL call_6_2_9_1(SH2_struct* sh)
{
  SH2movwl(sh,2,9);
}
#define init_6_2_10_1
void FASTCALL call_6_2_10_1(SH2_struct* sh)
{
  SH2movwl(sh,2,10);
}
#define init_6_2_11_1
void FASTCALL call_6_2_11_1(SH2_struct* sh)
{
  SH2movwl(sh,2,11);
}
#define init_6_2_12_1
void FASTCALL call_6_2_12_1(SH2_struct* sh)
{
  SH2movwl(sh,2,12);
}
#define init_6_2_13_1
void FASTCALL call_6_2_13_1(SH2_struct* sh)
{
  SH2movwl(sh,2,13);
}
#define init_6_2_14_1
void FASTCALL call_6_2_14_1(SH2_struct* sh)
{
  SH2movwl(sh,2,14);
}
#define init_6_2_15_1
void FASTCALL call_6_2_15_1(SH2_struct* sh)
{
  SH2movwl(sh,2,15);
}
#define init_6_3_0_1
void FASTCALL call_6_3_0_1(SH2_struct* sh)
{
  SH2movwl(sh,3,0);
}
#define init_6_3_1_1
void FASTCALL call_6_3_1_1(SH2_struct* sh)
{
  SH2movwl(sh,3,1);
}
#define init_6_3_2_1
void FASTCALL call_6_3_2_1(SH2_struct* sh)
{
  SH2movwl(sh,3,2);
}
#define init_6_3_3_1
void FASTCALL call_6_3_3_1(SH2_struct* sh)
{
  SH2movwl(sh,3,3);
}
#define init_6_3_4_1
void FASTCALL call_6_3_4_1(SH2_struct* sh)
{
  SH2movwl(sh,3,4);
}
#define init_6_3_5_1
void FASTCALL call_6_3_5_1(SH2_struct* sh)
{
  SH2movwl(sh,3,5);
}
#define init_6_3_6_1
void FASTCALL call_6_3_6_1(SH2_struct* sh)
{
  SH2movwl(sh,3,6);
}
#define init_6_3_7_1
void FASTCALL call_6_3_7_1(SH2_struct* sh)
{
  SH2movwl(sh,3,7);
}
#define init_6_3_8_1
void FASTCALL call_6_3_8_1(SH2_struct* sh)
{
  SH2movwl(sh,3,8);
}
#define init_6_3_9_1
void FASTCALL call_6_3_9_1(SH2_struct* sh)
{
  SH2movwl(sh,3,9);
}
#define init_6_3_10_1
void FASTCALL call_6_3_10_1(SH2_struct* sh)
{
  SH2movwl(sh,3,10);
}
#define init_6_3_11_1
void FASTCALL call_6_3_11_1(SH2_struct* sh)
{
  SH2movwl(sh,3,11);
}
#define init_6_3_12_1
void FASTCALL call_6_3_12_1(SH2_struct* sh)
{
  SH2movwl(sh,3,12);
}
#define init_6_3_13_1
void FASTCALL call_6_3_13_1(SH2_struct* sh)
{
  SH2movwl(sh,3,13);
}
#define init_6_3_14_1
void FASTCALL call_6_3_14_1(SH2_struct* sh)
{
  SH2movwl(sh,3,14);
}
#define init_6_3_15_1
void FASTCALL call_6_3_15_1(SH2_struct* sh)
{
  SH2movwl(sh,3,15);
}
#define init_6_4_0_1
void FASTCALL call_6_4_0_1(SH2_struct* sh)
{
  SH2movwl(sh,4,0);
}
#define init_6_4_1_1
void FASTCALL call_6_4_1_1(SH2_struct* sh)
{
  SH2movwl(sh,4,1);
}
#define init_6_4_2_1
void FASTCALL call_6_4_2_1(SH2_struct* sh)
{
  SH2movwl(sh,4,2);
}
#define init_6_4_3_1
void FASTCALL call_6_4_3_1(SH2_struct* sh)
{
  SH2movwl(sh,4,3);
}
#define init_6_4_4_1
void FASTCALL call_6_4_4_1(SH2_struct* sh)
{
  SH2movwl(sh,4,4);
}
#define init_6_4_5_1
void FASTCALL call_6_4_5_1(SH2_struct* sh)
{
  SH2movwl(sh,4,5);
}
#define init_6_4_6_1
void FASTCALL call_6_4_6_1(SH2_struct* sh)
{
  SH2movwl(sh,4,6);
}
#define init_6_4_7_1
void FASTCALL call_6_4_7_1(SH2_struct* sh)
{
  SH2movwl(sh,4,7);
}
#define init_6_4_8_1
void FASTCALL call_6_4_8_1(SH2_struct* sh)
{
  SH2movwl(sh,4,8);
}
#define init_6_4_9_1
void FASTCALL call_6_4_9_1(SH2_struct* sh)
{
  SH2movwl(sh,4,9);
}
#define init_6_4_10_1
void FASTCALL call_6_4_10_1(SH2_struct* sh)
{
  SH2movwl(sh,4,10);
}
#define init_6_4_11_1
void FASTCALL call_6_4_11_1(SH2_struct* sh)
{
  SH2movwl(sh,4,11);
}
#define init_6_4_12_1
void FASTCALL call_6_4_12_1(SH2_struct* sh)
{
  SH2movwl(sh,4,12);
}
#define init_6_4_13_1
void FASTCALL call_6_4_13_1(SH2_struct* sh)
{
  SH2movwl(sh,4,13);
}
#define init_6_4_14_1
void FASTCALL call_6_4_14_1(SH2_struct* sh)
{
  SH2movwl(sh,4,14);
}
#define init_6_4_15_1
void FASTCALL call_6_4_15_1(SH2_struct* sh)
{
  SH2movwl(sh,4,15);
}
#define init_6_5_0_1
void FASTCALL call_6_5_0_1(SH2_struct* sh)
{
  SH2movwl(sh,5,0);
}
#define init_6_5_1_1
void FASTCALL call_6_5_1_1(SH2_struct* sh)
{
  SH2movwl(sh,5,1);
}
#define init_6_5_2_1
void FASTCALL call_6_5_2_1(SH2_struct* sh)
{
  SH2movwl(sh,5,2);
}
#define init_6_5_3_1
void FASTCALL call_6_5_3_1(SH2_struct* sh)
{
  SH2movwl(sh,5,3);
}
#define init_6_5_4_1
void FASTCALL call_6_5_4_1(SH2_struct* sh)
{
  SH2movwl(sh,5,4);
}
#define init_6_5_5_1
void FASTCALL call_6_5_5_1(SH2_struct* sh)
{
  SH2movwl(sh,5,5);
}
#define init_6_5_6_1
void FASTCALL call_6_5_6_1(SH2_struct* sh)
{
  SH2movwl(sh,5,6);
}
#define init_6_5_7_1
void FASTCALL call_6_5_7_1(SH2_struct* sh)
{
  SH2movwl(sh,5,7);
}
#define init_6_5_8_1
void FASTCALL call_6_5_8_1(SH2_struct* sh)
{
  SH2movwl(sh,5,8);
}
#define init_6_5_9_1
void FASTCALL call_6_5_9_1(SH2_struct* sh)
{
  SH2movwl(sh,5,9);
}
#define init_6_5_10_1
void FASTCALL call_6_5_10_1(SH2_struct* sh)
{
  SH2movwl(sh,5,10);
}
#define init_6_5_11_1
void FASTCALL call_6_5_11_1(SH2_struct* sh)
{
  SH2movwl(sh,5,11);
}
#define init_6_5_12_1
void FASTCALL call_6_5_12_1(SH2_struct* sh)
{
  SH2movwl(sh,5,12);
}
#define init_6_5_13_1
void FASTCALL call_6_5_13_1(SH2_struct* sh)
{
  SH2movwl(sh,5,13);
}
#define init_6_5_14_1
void FASTCALL call_6_5_14_1(SH2_struct* sh)
{
  SH2movwl(sh,5,14);
}
#define init_6_5_15_1
void FASTCALL call_6_5_15_1(SH2_struct* sh)
{
  SH2movwl(sh,5,15);
}
#define init_6_6_0_1
void FASTCALL call_6_6_0_1(SH2_struct* sh)
{
  SH2movwl(sh,6,0);
}
#define init_6_6_1_1
void FASTCALL call_6_6_1_1(SH2_struct* sh)
{
  SH2movwl(sh,6,1);
}
#define init_6_6_2_1
void FASTCALL call_6_6_2_1(SH2_struct* sh)
{
  SH2movwl(sh,6,2);
}
#define init_6_6_3_1
void FASTCALL call_6_6_3_1(SH2_struct* sh)
{
  SH2movwl(sh,6,3);
}
#define init_6_6_4_1
void FASTCALL call_6_6_4_1(SH2_struct* sh)
{
  SH2movwl(sh,6,4);
}
#define init_6_6_5_1
void FASTCALL call_6_6_5_1(SH2_struct* sh)
{
  SH2movwl(sh,6,5);
}
#define init_6_6_6_1
void FASTCALL call_6_6_6_1(SH2_struct* sh)
{
  SH2movwl(sh,6,6);
}
#define init_6_6_7_1
void FASTCALL call_6_6_7_1(SH2_struct* sh)
{
  SH2movwl(sh,6,7);
}
#define init_6_6_8_1
void FASTCALL call_6_6_8_1(SH2_struct* sh)
{
  SH2movwl(sh,6,8);
}
#define init_6_6_9_1
void FASTCALL call_6_6_9_1(SH2_struct* sh)
{
  SH2movwl(sh,6,9);
}
#define init_6_6_10_1
void FASTCALL call_6_6_10_1(SH2_struct* sh)
{
  SH2movwl(sh,6,10);
}
#define init_6_6_11_1
void FASTCALL call_6_6_11_1(SH2_struct* sh)
{
  SH2movwl(sh,6,11);
}
#define init_6_6_12_1
void FASTCALL call_6_6_12_1(SH2_struct* sh)
{
  SH2movwl(sh,6,12);
}
#define init_6_6_13_1
void FASTCALL call_6_6_13_1(SH2_struct* sh)
{
  SH2movwl(sh,6,13);
}
#define init_6_6_14_1
void FASTCALL call_6_6_14_1(SH2_struct* sh)
{
  SH2movwl(sh,6,14);
}
#define init_6_6_15_1
void FASTCALL call_6_6_15_1(SH2_struct* sh)
{
  SH2movwl(sh,6,15);
}
#define init_6_7_0_1
void FASTCALL call_6_7_0_1(SH2_struct* sh)
{
  SH2movwl(sh,7,0);
}
#define init_6_7_1_1
void FASTCALL call_6_7_1_1(SH2_struct* sh)
{
  SH2movwl(sh,7,1);
}
#define init_6_7_2_1
void FASTCALL call_6_7_2_1(SH2_struct* sh)
{
  SH2movwl(sh,7,2);
}
#define init_6_7_3_1
void FASTCALL call_6_7_3_1(SH2_struct* sh)
{
  SH2movwl(sh,7,3);
}
#define init_6_7_4_1
void FASTCALL call_6_7_4_1(SH2_struct* sh)
{
  SH2movwl(sh,7,4);
}
#define init_6_7_5_1
void FASTCALL call_6_7_5_1(SH2_struct* sh)
{
  SH2movwl(sh,7,5);
}
#define init_6_7_6_1
void FASTCALL call_6_7_6_1(SH2_struct* sh)
{
  SH2movwl(sh,7,6);
}
#define init_6_7_7_1
void FASTCALL call_6_7_7_1(SH2_struct* sh)
{
  SH2movwl(sh,7,7);
}
#define init_6_7_8_1
void FASTCALL call_6_7_8_1(SH2_struct* sh)
{
  SH2movwl(sh,7,8);
}
#define init_6_7_9_1
void FASTCALL call_6_7_9_1(SH2_struct* sh)
{
  SH2movwl(sh,7,9);
}
#define init_6_7_10_1
void FASTCALL call_6_7_10_1(SH2_struct* sh)
{
  SH2movwl(sh,7,10);
}
#define init_6_7_11_1
void FASTCALL call_6_7_11_1(SH2_struct* sh)
{
  SH2movwl(sh,7,11);
}
#define init_6_7_12_1
void FASTCALL call_6_7_12_1(SH2_struct* sh)
{
  SH2movwl(sh,7,12);
}
#define init_6_7_13_1
void FASTCALL call_6_7_13_1(SH2_struct* sh)
{
  SH2movwl(sh,7,13);
}
#define init_6_7_14_1
void FASTCALL call_6_7_14_1(SH2_struct* sh)
{
  SH2movwl(sh,7,14);
}
#define init_6_7_15_1
void FASTCALL call_6_7_15_1(SH2_struct* sh)
{
  SH2movwl(sh,7,15);
}
#define init_6_8_0_1
void FASTCALL call_6_8_0_1(SH2_struct* sh)
{
  SH2movwl(sh,8,0);
}
#define init_6_8_1_1
void FASTCALL call_6_8_1_1(SH2_struct* sh)
{
  SH2movwl(sh,8,1);
}
#define init_6_8_2_1
void FASTCALL call_6_8_2_1(SH2_struct* sh)
{
  SH2movwl(sh,8,2);
}
#define init_6_8_3_1
void FASTCALL call_6_8_3_1(SH2_struct* sh)
{
  SH2movwl(sh,8,3);
}
#define init_6_8_4_1
void FASTCALL call_6_8_4_1(SH2_struct* sh)
{
  SH2movwl(sh,8,4);
}
#define init_6_8_5_1
void FASTCALL call_6_8_5_1(SH2_struct* sh)
{
  SH2movwl(sh,8,5);
}
#define init_6_8_6_1
void FASTCALL call_6_8_6_1(SH2_struct* sh)
{
  SH2movwl(sh,8,6);
}
#define init_6_8_7_1
void FASTCALL call_6_8_7_1(SH2_struct* sh)
{
  SH2movwl(sh,8,7);
}
#define init_6_8_8_1
void FASTCALL call_6_8_8_1(SH2_struct* sh)
{
  SH2movwl(sh,8,8);
}
#define init_6_8_9_1
void FASTCALL call_6_8_9_1(SH2_struct* sh)
{
  SH2movwl(sh,8,9);
}
#define init_6_8_10_1
void FASTCALL call_6_8_10_1(SH2_struct* sh)
{
  SH2movwl(sh,8,10);
}
#define init_6_8_11_1
void FASTCALL call_6_8_11_1(SH2_struct* sh)
{
  SH2movwl(sh,8,11);
}
#define init_6_8_12_1
void FASTCALL call_6_8_12_1(SH2_struct* sh)
{
  SH2movwl(sh,8,12);
}
#define init_6_8_13_1
void FASTCALL call_6_8_13_1(SH2_struct* sh)
{
  SH2movwl(sh,8,13);
}
#define init_6_8_14_1
void FASTCALL call_6_8_14_1(SH2_struct* sh)
{
  SH2movwl(sh,8,14);
}
#define init_6_8_15_1
void FASTCALL call_6_8_15_1(SH2_struct* sh)
{
  SH2movwl(sh,8,15);
}
#define init_6_9_0_1
void FASTCALL call_6_9_0_1(SH2_struct* sh)
{
  SH2movwl(sh,9,0);
}
#define init_6_9_1_1
void FASTCALL call_6_9_1_1(SH2_struct* sh)
{
  SH2movwl(sh,9,1);
}
#define init_6_9_2_1
void FASTCALL call_6_9_2_1(SH2_struct* sh)
{
  SH2movwl(sh,9,2);
}
#define init_6_9_3_1
void FASTCALL call_6_9_3_1(SH2_struct* sh)
{
  SH2movwl(sh,9,3);
}
#define init_6_9_4_1
void FASTCALL call_6_9_4_1(SH2_struct* sh)
{
  SH2movwl(sh,9,4);
}
#define init_6_9_5_1
void FASTCALL call_6_9_5_1(SH2_struct* sh)
{
  SH2movwl(sh,9,5);
}
#define init_6_9_6_1
void FASTCALL call_6_9_6_1(SH2_struct* sh)
{
  SH2movwl(sh,9,6);
}
#define init_6_9_7_1
void FASTCALL call_6_9_7_1(SH2_struct* sh)
{
  SH2movwl(sh,9,7);
}
#define init_6_9_8_1
void FASTCALL call_6_9_8_1(SH2_struct* sh)
{
  SH2movwl(sh,9,8);
}
#define init_6_9_9_1
void FASTCALL call_6_9_9_1(SH2_struct* sh)
{
  SH2movwl(sh,9,9);
}
#define init_6_9_10_1
void FASTCALL call_6_9_10_1(SH2_struct* sh)
{
  SH2movwl(sh,9,10);
}
#define init_6_9_11_1
void FASTCALL call_6_9_11_1(SH2_struct* sh)
{
  SH2movwl(sh,9,11);
}
#define init_6_9_12_1
void FASTCALL call_6_9_12_1(SH2_struct* sh)
{
  SH2movwl(sh,9,12);
}
#define init_6_9_13_1
void FASTCALL call_6_9_13_1(SH2_struct* sh)
{
  SH2movwl(sh,9,13);
}
#define init_6_9_14_1
void FASTCALL call_6_9_14_1(SH2_struct* sh)
{
  SH2movwl(sh,9,14);
}
#define init_6_9_15_1
void FASTCALL call_6_9_15_1(SH2_struct* sh)
{
  SH2movwl(sh,9,15);
}
#define init_6_10_0_1
void FASTCALL call_6_10_0_1(SH2_struct* sh)
{
  SH2movwl(sh,10,0);
}
#define init_6_10_1_1
void FASTCALL call_6_10_1_1(SH2_struct* sh)
{
  SH2movwl(sh,10,1);
}
#define init_6_10_2_1
void FASTCALL call_6_10_2_1(SH2_struct* sh)
{
  SH2movwl(sh,10,2);
}
#define init_6_10_3_1
void FASTCALL call_6_10_3_1(SH2_struct* sh)
{
  SH2movwl(sh,10,3);
}
#define init_6_10_4_1
void FASTCALL call_6_10_4_1(SH2_struct* sh)
{
  SH2movwl(sh,10,4);
}
#define init_6_10_5_1
void FASTCALL call_6_10_5_1(SH2_struct* sh)
{
  SH2movwl(sh,10,5);
}
#define init_6_10_6_1
void FASTCALL call_6_10_6_1(SH2_struct* sh)
{
  SH2movwl(sh,10,6);
}
#define init_6_10_7_1
void FASTCALL call_6_10_7_1(SH2_struct* sh)
{
  SH2movwl(sh,10,7);
}
#define init_6_10_8_1
void FASTCALL call_6_10_8_1(SH2_struct* sh)
{
  SH2movwl(sh,10,8);
}
#define init_6_10_9_1
void FASTCALL call_6_10_9_1(SH2_struct* sh)
{
  SH2movwl(sh,10,9);
}
#define init_6_10_10_1
void FASTCALL call_6_10_10_1(SH2_struct* sh)
{
  SH2movwl(sh,10,10);
}
#define init_6_10_11_1
void FASTCALL call_6_10_11_1(SH2_struct* sh)
{
  SH2movwl(sh,10,11);
}
#define init_6_10_12_1
void FASTCALL call_6_10_12_1(SH2_struct* sh)
{
  SH2movwl(sh,10,12);
}
#define init_6_10_13_1
void FASTCALL call_6_10_13_1(SH2_struct* sh)
{
  SH2movwl(sh,10,13);
}
#define init_6_10_14_1
void FASTCALL call_6_10_14_1(SH2_struct* sh)
{
  SH2movwl(sh,10,14);
}
#define init_6_10_15_1
void FASTCALL call_6_10_15_1(SH2_struct* sh)
{
  SH2movwl(sh,10,15);
}
#define init_6_11_0_1
void FASTCALL call_6_11_0_1(SH2_struct* sh)
{
  SH2movwl(sh,11,0);
}
#define init_6_11_1_1
void FASTCALL call_6_11_1_1(SH2_struct* sh)
{
  SH2movwl(sh,11,1);
}
#define init_6_11_2_1
void FASTCALL call_6_11_2_1(SH2_struct* sh)
{
  SH2movwl(sh,11,2);
}
#define init_6_11_3_1
void FASTCALL call_6_11_3_1(SH2_struct* sh)
{
  SH2movwl(sh,11,3);
}
#define init_6_11_4_1
void FASTCALL call_6_11_4_1(SH2_struct* sh)
{
  SH2movwl(sh,11,4);
}
#define init_6_11_5_1
void FASTCALL call_6_11_5_1(SH2_struct* sh)
{
  SH2movwl(sh,11,5);
}
#define init_6_11_6_1
void FASTCALL call_6_11_6_1(SH2_struct* sh)
{
  SH2movwl(sh,11,6);
}
#define init_6_11_7_1
void FASTCALL call_6_11_7_1(SH2_struct* sh)
{
  SH2movwl(sh,11,7);
}
#define init_6_11_8_1
void FASTCALL call_6_11_8_1(SH2_struct* sh)
{
  SH2movwl(sh,11,8);
}
#define init_6_11_9_1
void FASTCALL call_6_11_9_1(SH2_struct* sh)
{
  SH2movwl(sh,11,9);
}
#define init_6_11_10_1
void FASTCALL call_6_11_10_1(SH2_struct* sh)
{
  SH2movwl(sh,11,10);
}
#define init_6_11_11_1
void FASTCALL call_6_11_11_1(SH2_struct* sh)
{
  SH2movwl(sh,11,11);
}
#define init_6_11_12_1
void FASTCALL call_6_11_12_1(SH2_struct* sh)
{
  SH2movwl(sh,11,12);
}
#define init_6_11_13_1
void FASTCALL call_6_11_13_1(SH2_struct* sh)
{
  SH2movwl(sh,11,13);
}
#define init_6_11_14_1
void FASTCALL call_6_11_14_1(SH2_struct* sh)
{
  SH2movwl(sh,11,14);
}
#define init_6_11_15_1
void FASTCALL call_6_11_15_1(SH2_struct* sh)
{
  SH2movwl(sh,11,15);
}
#define init_6_12_0_1
void FASTCALL call_6_12_0_1(SH2_struct* sh)
{
  SH2movwl(sh,12,0);
}
#define init_6_12_1_1
void FASTCALL call_6_12_1_1(SH2_struct* sh)
{
  SH2movwl(sh,12,1);
}
#define init_6_12_2_1
void FASTCALL call_6_12_2_1(SH2_struct* sh)
{
  SH2movwl(sh,12,2);
}
#define init_6_12_3_1
void FASTCALL call_6_12_3_1(SH2_struct* sh)
{
  SH2movwl(sh,12,3);
}
#define init_6_12_4_1
void FASTCALL call_6_12_4_1(SH2_struct* sh)
{
  SH2movwl(sh,12,4);
}
#define init_6_12_5_1
void FASTCALL call_6_12_5_1(SH2_struct* sh)
{
  SH2movwl(sh,12,5);
}
#define init_6_12_6_1
void FASTCALL call_6_12_6_1(SH2_struct* sh)
{
  SH2movwl(sh,12,6);
}
#define init_6_12_7_1
void FASTCALL call_6_12_7_1(SH2_struct* sh)
{
  SH2movwl(sh,12,7);
}
#define init_6_12_8_1
void FASTCALL call_6_12_8_1(SH2_struct* sh)
{
  SH2movwl(sh,12,8);
}
#define init_6_12_9_1
void FASTCALL call_6_12_9_1(SH2_struct* sh)
{
  SH2movwl(sh,12,9);
}
#define init_6_12_10_1
void FASTCALL call_6_12_10_1(SH2_struct* sh)
{
  SH2movwl(sh,12,10);
}
#define init_6_12_11_1
void FASTCALL call_6_12_11_1(SH2_struct* sh)
{
  SH2movwl(sh,12,11);
}
#define init_6_12_12_1
void FASTCALL call_6_12_12_1(SH2_struct* sh)
{
  SH2movwl(sh,12,12);
}
#define init_6_12_13_1
void FASTCALL call_6_12_13_1(SH2_struct* sh)
{
  SH2movwl(sh,12,13);
}
#define init_6_12_14_1
void FASTCALL call_6_12_14_1(SH2_struct* sh)
{
  SH2movwl(sh,12,14);
}
#define init_6_12_15_1
void FASTCALL call_6_12_15_1(SH2_struct* sh)
{
  SH2movwl(sh,12,15);
}
#define init_6_13_0_1
void FASTCALL call_6_13_0_1(SH2_struct* sh)
{
  SH2movwl(sh,13,0);
}
#define init_6_13_1_1
void FASTCALL call_6_13_1_1(SH2_struct* sh)
{
  SH2movwl(sh,13,1);
}
#define init_6_13_2_1
void FASTCALL call_6_13_2_1(SH2_struct* sh)
{
  SH2movwl(sh,13,2);
}
#define init_6_13_3_1
void FASTCALL call_6_13_3_1(SH2_struct* sh)
{
  SH2movwl(sh,13,3);
}
#define init_6_13_4_1
void FASTCALL call_6_13_4_1(SH2_struct* sh)
{
  SH2movwl(sh,13,4);
}
#define init_6_13_5_1
void FASTCALL call_6_13_5_1(SH2_struct* sh)
{
  SH2movwl(sh,13,5);
}
#define init_6_13_6_1
void FASTCALL call_6_13_6_1(SH2_struct* sh)
{
  SH2movwl(sh,13,6);
}
#define init_6_13_7_1
void FASTCALL call_6_13_7_1(SH2_struct* sh)
{
  SH2movwl(sh,13,7);
}
#define init_6_13_8_1
void FASTCALL call_6_13_8_1(SH2_struct* sh)
{
  SH2movwl(sh,13,8);
}
#define init_6_13_9_1
void FASTCALL call_6_13_9_1(SH2_struct* sh)
{
  SH2movwl(sh,13,9);
}
#define init_6_13_10_1
void FASTCALL call_6_13_10_1(SH2_struct* sh)
{
  SH2movwl(sh,13,10);
}
#define init_6_13_11_1
void FASTCALL call_6_13_11_1(SH2_struct* sh)
{
  SH2movwl(sh,13,11);
}
#define init_6_13_12_1
void FASTCALL call_6_13_12_1(SH2_struct* sh)
{
  SH2movwl(sh,13,12);
}
#define init_6_13_13_1
void FASTCALL call_6_13_13_1(SH2_struct* sh)
{
  SH2movwl(sh,13,13);
}
#define init_6_13_14_1
void FASTCALL call_6_13_14_1(SH2_struct* sh)
{
  SH2movwl(sh,13,14);
}
#define init_6_13_15_1
void FASTCALL call_6_13_15_1(SH2_struct* sh)
{
  SH2movwl(sh,13,15);
}
#define init_6_14_0_1
void FASTCALL call_6_14_0_1(SH2_struct* sh)
{
  SH2movwl(sh,14,0);
}
#define init_6_14_1_1
void FASTCALL call_6_14_1_1(SH2_struct* sh)
{
  SH2movwl(sh,14,1);
}
#define init_6_14_2_1
void FASTCALL call_6_14_2_1(SH2_struct* sh)
{
  SH2movwl(sh,14,2);
}
#define init_6_14_3_1
void FASTCALL call_6_14_3_1(SH2_struct* sh)
{
  SH2movwl(sh,14,3);
}
#define init_6_14_4_1
void FASTCALL call_6_14_4_1(SH2_struct* sh)
{
  SH2movwl(sh,14,4);
}
#define init_6_14_5_1
void FASTCALL call_6_14_5_1(SH2_struct* sh)
{
  SH2movwl(sh,14,5);
}
#define init_6_14_6_1
void FASTCALL call_6_14_6_1(SH2_struct* sh)
{
  SH2movwl(sh,14,6);
}
#define init_6_14_7_1
void FASTCALL call_6_14_7_1(SH2_struct* sh)
{
  SH2movwl(sh,14,7);
}
#define init_6_14_8_1
void FASTCALL call_6_14_8_1(SH2_struct* sh)
{
  SH2movwl(sh,14,8);
}
#define init_6_14_9_1
void FASTCALL call_6_14_9_1(SH2_struct* sh)
{
  SH2movwl(sh,14,9);
}
#define init_6_14_10_1
void FASTCALL call_6_14_10_1(SH2_struct* sh)
{
  SH2movwl(sh,14,10);
}
#define init_6_14_11_1
void FASTCALL call_6_14_11_1(SH2_struct* sh)
{
  SH2movwl(sh,14,11);
}
#define init_6_14_12_1
void FASTCALL call_6_14_12_1(SH2_struct* sh)
{
  SH2movwl(sh,14,12);
}
#define init_6_14_13_1
void FASTCALL call_6_14_13_1(SH2_struct* sh)
{
  SH2movwl(sh,14,13);
}
#define init_6_14_14_1
void FASTCALL call_6_14_14_1(SH2_struct* sh)
{
  SH2movwl(sh,14,14);
}
#define init_6_14_15_1
void FASTCALL call_6_14_15_1(SH2_struct* sh)
{
  SH2movwl(sh,14,15);
}
#define init_6_15_0_1
void FASTCALL call_6_15_0_1(SH2_struct* sh)
{
  SH2movwl(sh,15,0);
}
#define init_6_15_1_1
void FASTCALL call_6_15_1_1(SH2_struct* sh)
{
  SH2movwl(sh,15,1);
}
#define init_6_15_2_1
void FASTCALL call_6_15_2_1(SH2_struct* sh)
{
  SH2movwl(sh,15,2);
}
#define init_6_15_3_1
void FASTCALL call_6_15_3_1(SH2_struct* sh)
{
  SH2movwl(sh,15,3);
}
#define init_6_15_4_1
void FASTCALL call_6_15_4_1(SH2_struct* sh)
{
  SH2movwl(sh,15,4);
}
#define init_6_15_5_1
void FASTCALL call_6_15_5_1(SH2_struct* sh)
{
  SH2movwl(sh,15,5);
}
#define init_6_15_6_1
void FASTCALL call_6_15_6_1(SH2_struct* sh)
{
  SH2movwl(sh,15,6);
}
#define init_6_15_7_1
void FASTCALL call_6_15_7_1(SH2_struct* sh)
{
  SH2movwl(sh,15,7);
}
#define init_6_15_8_1
void FASTCALL call_6_15_8_1(SH2_struct* sh)
{
  SH2movwl(sh,15,8);
}
#define init_6_15_9_1
void FASTCALL call_6_15_9_1(SH2_struct* sh)
{
  SH2movwl(sh,15,9);
}
#define init_6_15_10_1
void FASTCALL call_6_15_10_1(SH2_struct* sh)
{
  SH2movwl(sh,15,10);
}
#define init_6_15_11_1
void FASTCALL call_6_15_11_1(SH2_struct* sh)
{
  SH2movwl(sh,15,11);
}
#define init_6_15_12_1
void FASTCALL call_6_15_12_1(SH2_struct* sh)
{
  SH2movwl(sh,15,12);
}
#define init_6_15_13_1
void FASTCALL call_6_15_13_1(SH2_struct* sh)
{
  SH2movwl(sh,15,13);
}
#define init_6_15_14_1
void FASTCALL call_6_15_14_1(SH2_struct* sh)
{
  SH2movwl(sh,15,14);
}
#define init_6_15_15_1
void FASTCALL call_6_15_15_1(SH2_struct* sh)
{
  SH2movwl(sh,15,15);
}
#define init_9_0_0_0
void FASTCALL call_9_0_0_0(SH2_struct* sh)
{
  SH2movwi(sh, 0, (0<<4)+0);
}
#define init_9_0_0_1
void FASTCALL call_9_0_0_1(SH2_struct* sh)
{
  SH2movwi(sh, 0, (0<<4)+1);
}
#define init_9_0_0_2
void FASTCALL call_9_0_0_2(SH2_struct* sh)
{
  SH2movwi(sh, 0, (0<<4)+2);
}
#define init_9_0_0_3
void FASTCALL call_9_0_0_3(SH2_struct* sh)
{
  SH2movwi(sh, 0, (0<<4)+3);
}
#define init_9_0_0_4
void FASTCALL call_9_0_0_4(SH2_struct* sh)
{
  SH2movwi(sh, 0, (0<<4)+4);
}
#define init_9_0_0_5
void FASTCALL call_9_0_0_5(SH2_struct* sh)
{
  SH2movwi(sh, 0, (0<<4)+5);
}
#define init_9_0_0_6
void FASTCALL call_9_0_0_6(SH2_struct* sh)
{
  SH2movwi(sh, 0, (0<<4)+6);
}
#define init_9_0_0_7
void FASTCALL call_9_0_0_7(SH2_struct* sh)
{
  SH2movwi(sh, 0, (0<<4)+7);
}
#define init_9_0_0_8
void FASTCALL call_9_0_0_8(SH2_struct* sh)
{
  SH2movwi(sh, 0, (0<<4)+8);
}
#define init_9_0_0_9
void FASTCALL call_9_0_0_9(SH2_struct* sh)
{
  SH2movwi(sh, 0, (0<<4)+9);
}
#define init_9_0_0_10
void FASTCALL call_9_0_0_10(SH2_struct* sh)
{
  SH2movwi(sh, 0, (0<<4)+10);
}
#define init_9_0_0_11
void FASTCALL call_9_0_0_11(SH2_struct* sh)
{
  SH2movwi(sh, 0, (0<<4)+11);
}
#define init_9_0_0_12
void FASTCALL call_9_0_0_12(SH2_struct* sh)
{
  SH2movwi(sh, 0, (0<<4)+12);
}
#define init_9_0_0_13
void FASTCALL call_9_0_0_13(SH2_struct* sh)
{
  SH2movwi(sh, 0, (0<<4)+13);
}
#define init_9_0_0_14
void FASTCALL call_9_0_0_14(SH2_struct* sh)
{
  SH2movwi(sh, 0, (0<<4)+14);
}
#define init_9_0_0_15
void FASTCALL call_9_0_0_15(SH2_struct* sh)
{
  SH2movwi(sh, 0, (0<<4)+15);
}
#define init_9_0_1_0
void FASTCALL call_9_0_1_0(SH2_struct* sh)
{
  SH2movwi(sh, 0, (1<<4)+0);
}
#define init_9_0_1_1
void FASTCALL call_9_0_1_1(SH2_struct* sh)
{
  SH2movwi(sh, 0, (1<<4)+1);
}
#define init_9_0_1_2
void FASTCALL call_9_0_1_2(SH2_struct* sh)
{
  SH2movwi(sh, 0, (1<<4)+2);
}
#define init_9_0_1_3
void FASTCALL call_9_0_1_3(SH2_struct* sh)
{
  SH2movwi(sh, 0, (1<<4)+3);
}
#define init_9_0_1_4
void FASTCALL call_9_0_1_4(SH2_struct* sh)
{
  SH2movwi(sh, 0, (1<<4)+4);
}
#define init_9_0_1_5
void FASTCALL call_9_0_1_5(SH2_struct* sh)
{
  SH2movwi(sh, 0, (1<<4)+5);
}
#define init_9_0_1_6
void FASTCALL call_9_0_1_6(SH2_struct* sh)
{
  SH2movwi(sh, 0, (1<<4)+6);
}
#define init_9_0_1_7
void FASTCALL call_9_0_1_7(SH2_struct* sh)
{
  SH2movwi(sh, 0, (1<<4)+7);
}
#define init_9_0_1_8
void FASTCALL call_9_0_1_8(SH2_struct* sh)
{
  SH2movwi(sh, 0, (1<<4)+8);
}
#define init_9_0_1_9
void FASTCALL call_9_0_1_9(SH2_struct* sh)
{
  SH2movwi(sh, 0, (1<<4)+9);
}
#define init_9_0_1_10
void FASTCALL call_9_0_1_10(SH2_struct* sh)
{
  SH2movwi(sh, 0, (1<<4)+10);
}
#define init_9_0_1_11
void FASTCALL call_9_0_1_11(SH2_struct* sh)
{
  SH2movwi(sh, 0, (1<<4)+11);
}
#define init_9_0_1_12
void FASTCALL call_9_0_1_12(SH2_struct* sh)
{
  SH2movwi(sh, 0, (1<<4)+12);
}
#define init_9_0_1_13
void FASTCALL call_9_0_1_13(SH2_struct* sh)
{
  SH2movwi(sh, 0, (1<<4)+13);
}
#define init_9_0_1_14
void FASTCALL call_9_0_1_14(SH2_struct* sh)
{
  SH2movwi(sh, 0, (1<<4)+14);
}
#define init_9_0_1_15
void FASTCALL call_9_0_1_15(SH2_struct* sh)
{
  SH2movwi(sh, 0, (1<<4)+15);
}
#define init_9_0_2_0
void FASTCALL call_9_0_2_0(SH2_struct* sh)
{
  SH2movwi(sh, 0, (2<<4)+0);
}
#define init_9_0_2_1
void FASTCALL call_9_0_2_1(SH2_struct* sh)
{
  SH2movwi(sh, 0, (2<<4)+1);
}
#define init_9_0_2_2
void FASTCALL call_9_0_2_2(SH2_struct* sh)
{
  SH2movwi(sh, 0, (2<<4)+2);
}
#define init_9_0_2_3
void FASTCALL call_9_0_2_3(SH2_struct* sh)
{
  SH2movwi(sh, 0, (2<<4)+3);
}
#define init_9_0_2_4
void FASTCALL call_9_0_2_4(SH2_struct* sh)
{
  SH2movwi(sh, 0, (2<<4)+4);
}
#define init_9_0_2_5
void FASTCALL call_9_0_2_5(SH2_struct* sh)
{
  SH2movwi(sh, 0, (2<<4)+5);
}
#define init_9_0_2_6
void FASTCALL call_9_0_2_6(SH2_struct* sh)
{
  SH2movwi(sh, 0, (2<<4)+6);
}
#define init_9_0_2_7
void FASTCALL call_9_0_2_7(SH2_struct* sh)
{
  SH2movwi(sh, 0, (2<<4)+7);
}
#define init_9_0_2_8
void FASTCALL call_9_0_2_8(SH2_struct* sh)
{
  SH2movwi(sh, 0, (2<<4)+8);
}
#define init_9_0_2_9
void FASTCALL call_9_0_2_9(SH2_struct* sh)
{
  SH2movwi(sh, 0, (2<<4)+9);
}
#define init_9_0_2_10
void FASTCALL call_9_0_2_10(SH2_struct* sh)
{
  SH2movwi(sh, 0, (2<<4)+10);
}
#define init_9_0_2_11
void FASTCALL call_9_0_2_11(SH2_struct* sh)
{
  SH2movwi(sh, 0, (2<<4)+11);
}
#define init_9_0_2_12
void FASTCALL call_9_0_2_12(SH2_struct* sh)
{
  SH2movwi(sh, 0, (2<<4)+12);
}
#define init_9_0_2_13
void FASTCALL call_9_0_2_13(SH2_struct* sh)
{
  SH2movwi(sh, 0, (2<<4)+13);
}
#define init_9_0_2_14
void FASTCALL call_9_0_2_14(SH2_struct* sh)
{
  SH2movwi(sh, 0, (2<<4)+14);
}
#define init_9_0_2_15
void FASTCALL call_9_0_2_15(SH2_struct* sh)
{
  SH2movwi(sh, 0, (2<<4)+15);
}
#define init_9_0_3_0
void FASTCALL call_9_0_3_0(SH2_struct* sh)
{
  SH2movwi(sh, 0, (3<<4)+0);
}
#define init_9_0_3_1
void FASTCALL call_9_0_3_1(SH2_struct* sh)
{
  SH2movwi(sh, 0, (3<<4)+1);
}
#define init_9_0_3_2
void FASTCALL call_9_0_3_2(SH2_struct* sh)
{
  SH2movwi(sh, 0, (3<<4)+2);
}
#define init_9_0_3_3
void FASTCALL call_9_0_3_3(SH2_struct* sh)
{
  SH2movwi(sh, 0, (3<<4)+3);
}
#define init_9_0_3_4
void FASTCALL call_9_0_3_4(SH2_struct* sh)
{
  SH2movwi(sh, 0, (3<<4)+4);
}
#define init_9_0_3_5
void FASTCALL call_9_0_3_5(SH2_struct* sh)
{
  SH2movwi(sh, 0, (3<<4)+5);
}
#define init_9_0_3_6
void FASTCALL call_9_0_3_6(SH2_struct* sh)
{
  SH2movwi(sh, 0, (3<<4)+6);
}
#define init_9_0_3_7
void FASTCALL call_9_0_3_7(SH2_struct* sh)
{
  SH2movwi(sh, 0, (3<<4)+7);
}
#define init_9_0_3_8
void FASTCALL call_9_0_3_8(SH2_struct* sh)
{
  SH2movwi(sh, 0, (3<<4)+8);
}
#define init_9_0_3_9
void FASTCALL call_9_0_3_9(SH2_struct* sh)
{
  SH2movwi(sh, 0, (3<<4)+9);
}
#define init_9_0_3_10
void FASTCALL call_9_0_3_10(SH2_struct* sh)
{
  SH2movwi(sh, 0, (3<<4)+10);
}
#define init_9_0_3_11
void FASTCALL call_9_0_3_11(SH2_struct* sh)
{
  SH2movwi(sh, 0, (3<<4)+11);
}
#define init_9_0_3_12
void FASTCALL call_9_0_3_12(SH2_struct* sh)
{
  SH2movwi(sh, 0, (3<<4)+12);
}
#define init_9_0_3_13
void FASTCALL call_9_0_3_13(SH2_struct* sh)
{
  SH2movwi(sh, 0, (3<<4)+13);
}
#define init_9_0_3_14
void FASTCALL call_9_0_3_14(SH2_struct* sh)
{
  SH2movwi(sh, 0, (3<<4)+14);
}
#define init_9_0_3_15
void FASTCALL call_9_0_3_15(SH2_struct* sh)
{
  SH2movwi(sh, 0, (3<<4)+15);
}
#define init_9_0_4_0
void FASTCALL call_9_0_4_0(SH2_struct* sh)
{
  SH2movwi(sh, 0, (4<<4)+0);
}
#define init_9_0_4_1
void FASTCALL call_9_0_4_1(SH2_struct* sh)
{
  SH2movwi(sh, 0, (4<<4)+1);
}
#define init_9_0_4_2
void FASTCALL call_9_0_4_2(SH2_struct* sh)
{
  SH2movwi(sh, 0, (4<<4)+2);
}
#define init_9_0_4_3
void FASTCALL call_9_0_4_3(SH2_struct* sh)
{
  SH2movwi(sh, 0, (4<<4)+3);
}
#define init_9_0_4_4
void FASTCALL call_9_0_4_4(SH2_struct* sh)
{
  SH2movwi(sh, 0, (4<<4)+4);
}
#define init_9_0_4_5
void FASTCALL call_9_0_4_5(SH2_struct* sh)
{
  SH2movwi(sh, 0, (4<<4)+5);
}
#define init_9_0_4_6
void FASTCALL call_9_0_4_6(SH2_struct* sh)
{
  SH2movwi(sh, 0, (4<<4)+6);
}
#define init_9_0_4_7
void FASTCALL call_9_0_4_7(SH2_struct* sh)
{
  SH2movwi(sh, 0, (4<<4)+7);
}
#define init_9_0_4_8
void FASTCALL call_9_0_4_8(SH2_struct* sh)
{
  SH2movwi(sh, 0, (4<<4)+8);
}
#define init_9_0_4_9
void FASTCALL call_9_0_4_9(SH2_struct* sh)
{
  SH2movwi(sh, 0, (4<<4)+9);
}
#define init_9_0_4_10
void FASTCALL call_9_0_4_10(SH2_struct* sh)
{
  SH2movwi(sh, 0, (4<<4)+10);
}
#define init_9_0_4_11
void FASTCALL call_9_0_4_11(SH2_struct* sh)
{
  SH2movwi(sh, 0, (4<<4)+11);
}
#define init_9_0_4_12
void FASTCALL call_9_0_4_12(SH2_struct* sh)
{
  SH2movwi(sh, 0, (4<<4)+12);
}
#define init_9_0_4_13
void FASTCALL call_9_0_4_13(SH2_struct* sh)
{
  SH2movwi(sh, 0, (4<<4)+13);
}
#define init_9_0_4_14
void FASTCALL call_9_0_4_14(SH2_struct* sh)
{
  SH2movwi(sh, 0, (4<<4)+14);
}
#define init_9_0_4_15
void FASTCALL call_9_0_4_15(SH2_struct* sh)
{
  SH2movwi(sh, 0, (4<<4)+15);
}
#define init_9_0_5_0
void FASTCALL call_9_0_5_0(SH2_struct* sh)
{
  SH2movwi(sh, 0, (5<<4)+0);
}
#define init_9_0_5_1
void FASTCALL call_9_0_5_1(SH2_struct* sh)
{
  SH2movwi(sh, 0, (5<<4)+1);
}
#define init_9_0_5_2
void FASTCALL call_9_0_5_2(SH2_struct* sh)
{
  SH2movwi(sh, 0, (5<<4)+2);
}
#define init_9_0_5_3
void FASTCALL call_9_0_5_3(SH2_struct* sh)
{
  SH2movwi(sh, 0, (5<<4)+3);
}
#define init_9_0_5_4
void FASTCALL call_9_0_5_4(SH2_struct* sh)
{
  SH2movwi(sh, 0, (5<<4)+4);
}
#define init_9_0_5_5
void FASTCALL call_9_0_5_5(SH2_struct* sh)
{
  SH2movwi(sh, 0, (5<<4)+5);
}
#define init_9_0_5_6
void FASTCALL call_9_0_5_6(SH2_struct* sh)
{
  SH2movwi(sh, 0, (5<<4)+6);
}
#define init_9_0_5_7
void FASTCALL call_9_0_5_7(SH2_struct* sh)
{
  SH2movwi(sh, 0, (5<<4)+7);
}
#define init_9_0_5_8
void FASTCALL call_9_0_5_8(SH2_struct* sh)
{
  SH2movwi(sh, 0, (5<<4)+8);
}
#define init_9_0_5_9
void FASTCALL call_9_0_5_9(SH2_struct* sh)
{
  SH2movwi(sh, 0, (5<<4)+9);
}
#define init_9_0_5_10
void FASTCALL call_9_0_5_10(SH2_struct* sh)
{
  SH2movwi(sh, 0, (5<<4)+10);
}
#define init_9_0_5_11
void FASTCALL call_9_0_5_11(SH2_struct* sh)
{
  SH2movwi(sh, 0, (5<<4)+11);
}
#define init_9_0_5_12
void FASTCALL call_9_0_5_12(SH2_struct* sh)
{
  SH2movwi(sh, 0, (5<<4)+12);
}
#define init_9_0_5_13
void FASTCALL call_9_0_5_13(SH2_struct* sh)
{
  SH2movwi(sh, 0, (5<<4)+13);
}
#define init_9_0_5_14
void FASTCALL call_9_0_5_14(SH2_struct* sh)
{
  SH2movwi(sh, 0, (5<<4)+14);
}
#define init_9_0_5_15
void FASTCALL call_9_0_5_15(SH2_struct* sh)
{
  SH2movwi(sh, 0, (5<<4)+15);
}
#define init_9_0_6_0
void FASTCALL call_9_0_6_0(SH2_struct* sh)
{
  SH2movwi(sh, 0, (6<<4)+0);
}
#define init_9_0_6_1
void FASTCALL call_9_0_6_1(SH2_struct* sh)
{
  SH2movwi(sh, 0, (6<<4)+1);
}
#define init_9_0_6_2
void FASTCALL call_9_0_6_2(SH2_struct* sh)
{
  SH2movwi(sh, 0, (6<<4)+2);
}
#define init_9_0_6_3
void FASTCALL call_9_0_6_3(SH2_struct* sh)
{
  SH2movwi(sh, 0, (6<<4)+3);
}
#define init_9_0_6_4
void FASTCALL call_9_0_6_4(SH2_struct* sh)
{
  SH2movwi(sh, 0, (6<<4)+4);
}
#define init_9_0_6_5
void FASTCALL call_9_0_6_5(SH2_struct* sh)
{
  SH2movwi(sh, 0, (6<<4)+5);
}
#define init_9_0_6_6
void FASTCALL call_9_0_6_6(SH2_struct* sh)
{
  SH2movwi(sh, 0, (6<<4)+6);
}
#define init_9_0_6_7
void FASTCALL call_9_0_6_7(SH2_struct* sh)
{
  SH2movwi(sh, 0, (6<<4)+7);
}
#define init_9_0_6_8
void FASTCALL call_9_0_6_8(SH2_struct* sh)
{
  SH2movwi(sh, 0, (6<<4)+8);
}
#define init_9_0_6_9
void FASTCALL call_9_0_6_9(SH2_struct* sh)
{
  SH2movwi(sh, 0, (6<<4)+9);
}
#define init_9_0_6_10
void FASTCALL call_9_0_6_10(SH2_struct* sh)
{
  SH2movwi(sh, 0, (6<<4)+10);
}
#define init_9_0_6_11
void FASTCALL call_9_0_6_11(SH2_struct* sh)
{
  SH2movwi(sh, 0, (6<<4)+11);
}
#define init_9_0_6_12
void FASTCALL call_9_0_6_12(SH2_struct* sh)
{
  SH2movwi(sh, 0, (6<<4)+12);
}
#define init_9_0_6_13
void FASTCALL call_9_0_6_13(SH2_struct* sh)
{
  SH2movwi(sh, 0, (6<<4)+13);
}
#define init_9_0_6_14
void FASTCALL call_9_0_6_14(SH2_struct* sh)
{
  SH2movwi(sh, 0, (6<<4)+14);
}
#define init_9_0_6_15
void FASTCALL call_9_0_6_15(SH2_struct* sh)
{
  SH2movwi(sh, 0, (6<<4)+15);
}
#define init_9_0_7_0
void FASTCALL call_9_0_7_0(SH2_struct* sh)
{
  SH2movwi(sh, 0, (7<<4)+0);
}
#define init_9_0_7_1
void FASTCALL call_9_0_7_1(SH2_struct* sh)
{
  SH2movwi(sh, 0, (7<<4)+1);
}
#define init_9_0_7_2
void FASTCALL call_9_0_7_2(SH2_struct* sh)
{
  SH2movwi(sh, 0, (7<<4)+2);
}
#define init_9_0_7_3
void FASTCALL call_9_0_7_3(SH2_struct* sh)
{
  SH2movwi(sh, 0, (7<<4)+3);
}
#define init_9_0_7_4
void FASTCALL call_9_0_7_4(SH2_struct* sh)
{
  SH2movwi(sh, 0, (7<<4)+4);
}
#define init_9_0_7_5
void FASTCALL call_9_0_7_5(SH2_struct* sh)
{
  SH2movwi(sh, 0, (7<<4)+5);
}
#define init_9_0_7_6
void FASTCALL call_9_0_7_6(SH2_struct* sh)
{
  SH2movwi(sh, 0, (7<<4)+6);
}
#define init_9_0_7_7
void FASTCALL call_9_0_7_7(SH2_struct* sh)
{
  SH2movwi(sh, 0, (7<<4)+7);
}
#define init_9_0_7_8
void FASTCALL call_9_0_7_8(SH2_struct* sh)
{
  SH2movwi(sh, 0, (7<<4)+8);
}
#define init_9_0_7_9
void FASTCALL call_9_0_7_9(SH2_struct* sh)
{
  SH2movwi(sh, 0, (7<<4)+9);
}
#define init_9_0_7_10
void FASTCALL call_9_0_7_10(SH2_struct* sh)
{
  SH2movwi(sh, 0, (7<<4)+10);
}
#define init_9_0_7_11
void FASTCALL call_9_0_7_11(SH2_struct* sh)
{
  SH2movwi(sh, 0, (7<<4)+11);
}
#define init_9_0_7_12
void FASTCALL call_9_0_7_12(SH2_struct* sh)
{
  SH2movwi(sh, 0, (7<<4)+12);
}
#define init_9_0_7_13
void FASTCALL call_9_0_7_13(SH2_struct* sh)
{
  SH2movwi(sh, 0, (7<<4)+13);
}
#define init_9_0_7_14
void FASTCALL call_9_0_7_14(SH2_struct* sh)
{
  SH2movwi(sh, 0, (7<<4)+14);
}
#define init_9_0_7_15
void FASTCALL call_9_0_7_15(SH2_struct* sh)
{
  SH2movwi(sh, 0, (7<<4)+15);
}
#define init_9_0_8_0
void FASTCALL call_9_0_8_0(SH2_struct* sh)
{
  SH2movwi(sh, 0, (8<<4)+0);
}
#define init_9_0_8_1
void FASTCALL call_9_0_8_1(SH2_struct* sh)
{
  SH2movwi(sh, 0, (8<<4)+1);
}
#define init_9_0_8_2
void FASTCALL call_9_0_8_2(SH2_struct* sh)
{
  SH2movwi(sh, 0, (8<<4)+2);
}
#define init_9_0_8_3
void FASTCALL call_9_0_8_3(SH2_struct* sh)
{
  SH2movwi(sh, 0, (8<<4)+3);
}
#define init_9_0_8_4
void FASTCALL call_9_0_8_4(SH2_struct* sh)
{
  SH2movwi(sh, 0, (8<<4)+4);
}
#define init_9_0_8_5
void FASTCALL call_9_0_8_5(SH2_struct* sh)
{
  SH2movwi(sh, 0, (8<<4)+5);
}
#define init_9_0_8_6
void FASTCALL call_9_0_8_6(SH2_struct* sh)
{
  SH2movwi(sh, 0, (8<<4)+6);
}
#define init_9_0_8_7
void FASTCALL call_9_0_8_7(SH2_struct* sh)
{
  SH2movwi(sh, 0, (8<<4)+7);
}
#define init_9_0_8_8
void FASTCALL call_9_0_8_8(SH2_struct* sh)
{
  SH2movwi(sh, 0, (8<<4)+8);
}
#define init_9_0_8_9
void FASTCALL call_9_0_8_9(SH2_struct* sh)
{
  SH2movwi(sh, 0, (8<<4)+9);
}
#define init_9_0_8_10
void FASTCALL call_9_0_8_10(SH2_struct* sh)
{
  SH2movwi(sh, 0, (8<<4)+10);
}
#define init_9_0_8_11
void FASTCALL call_9_0_8_11(SH2_struct* sh)
{
  SH2movwi(sh, 0, (8<<4)+11);
}
#define init_9_0_8_12
void FASTCALL call_9_0_8_12(SH2_struct* sh)
{
  SH2movwi(sh, 0, (8<<4)+12);
}
#define init_9_0_8_13
void FASTCALL call_9_0_8_13(SH2_struct* sh)
{
  SH2movwi(sh, 0, (8<<4)+13);
}
#define init_9_0_8_14
void FASTCALL call_9_0_8_14(SH2_struct* sh)
{
  SH2movwi(sh, 0, (8<<4)+14);
}
#define init_9_0_8_15
void FASTCALL call_9_0_8_15(SH2_struct* sh)
{
  SH2movwi(sh, 0, (8<<4)+15);
}
#define init_9_0_9_0
void FASTCALL call_9_0_9_0(SH2_struct* sh)
{
  SH2movwi(sh, 0, (9<<4)+0);
}
#define init_9_0_9_1
void FASTCALL call_9_0_9_1(SH2_struct* sh)
{
  SH2movwi(sh, 0, (9<<4)+1);
}
#define init_9_0_9_2
void FASTCALL call_9_0_9_2(SH2_struct* sh)
{
  SH2movwi(sh, 0, (9<<4)+2);
}
#define init_9_0_9_3
void FASTCALL call_9_0_9_3(SH2_struct* sh)
{
  SH2movwi(sh, 0, (9<<4)+3);
}
#define init_9_0_9_4
void FASTCALL call_9_0_9_4(SH2_struct* sh)
{
  SH2movwi(sh, 0, (9<<4)+4);
}
#define init_9_0_9_5
void FASTCALL call_9_0_9_5(SH2_struct* sh)
{
  SH2movwi(sh, 0, (9<<4)+5);
}
#define init_9_0_9_6
void FASTCALL call_9_0_9_6(SH2_struct* sh)
{
  SH2movwi(sh, 0, (9<<4)+6);
}
#define init_9_0_9_7
void FASTCALL call_9_0_9_7(SH2_struct* sh)
{
  SH2movwi(sh, 0, (9<<4)+7);
}
#define init_9_0_9_8
void FASTCALL call_9_0_9_8(SH2_struct* sh)
{
  SH2movwi(sh, 0, (9<<4)+8);
}
#define init_9_0_9_9
void FASTCALL call_9_0_9_9(SH2_struct* sh)
{
  SH2movwi(sh, 0, (9<<4)+9);
}
#define init_9_0_9_10
void FASTCALL call_9_0_9_10(SH2_struct* sh)
{
  SH2movwi(sh, 0, (9<<4)+10);
}
#define init_9_0_9_11
void FASTCALL call_9_0_9_11(SH2_struct* sh)
{
  SH2movwi(sh, 0, (9<<4)+11);
}
#define init_9_0_9_12
void FASTCALL call_9_0_9_12(SH2_struct* sh)
{
  SH2movwi(sh, 0, (9<<4)+12);
}
#define init_9_0_9_13
void FASTCALL call_9_0_9_13(SH2_struct* sh)
{
  SH2movwi(sh, 0, (9<<4)+13);
}
#define init_9_0_9_14
void FASTCALL call_9_0_9_14(SH2_struct* sh)
{
  SH2movwi(sh, 0, (9<<4)+14);
}
#define init_9_0_9_15
void FASTCALL call_9_0_9_15(SH2_struct* sh)
{
  SH2movwi(sh, 0, (9<<4)+15);
}
#define init_9_0_10_0
void FASTCALL call_9_0_10_0(SH2_struct* sh)
{
  SH2movwi(sh, 0, (10<<4)+0);
}
#define init_9_0_10_1
void FASTCALL call_9_0_10_1(SH2_struct* sh)
{
  SH2movwi(sh, 0, (10<<4)+1);
}
#define init_9_0_10_2
void FASTCALL call_9_0_10_2(SH2_struct* sh)
{
  SH2movwi(sh, 0, (10<<4)+2);
}
#define init_9_0_10_3
void FASTCALL call_9_0_10_3(SH2_struct* sh)
{
  SH2movwi(sh, 0, (10<<4)+3);
}
#define init_9_0_10_4
void FASTCALL call_9_0_10_4(SH2_struct* sh)
{
  SH2movwi(sh, 0, (10<<4)+4);
}
#define init_9_0_10_5
void FASTCALL call_9_0_10_5(SH2_struct* sh)
{
  SH2movwi(sh, 0, (10<<4)+5);
}
#define init_9_0_10_6
void FASTCALL call_9_0_10_6(SH2_struct* sh)
{
  SH2movwi(sh, 0, (10<<4)+6);
}
#define init_9_0_10_7
void FASTCALL call_9_0_10_7(SH2_struct* sh)
{
  SH2movwi(sh, 0, (10<<4)+7);
}
#define init_9_0_10_8
void FASTCALL call_9_0_10_8(SH2_struct* sh)
{
  SH2movwi(sh, 0, (10<<4)+8);
}
#define init_9_0_10_9
void FASTCALL call_9_0_10_9(SH2_struct* sh)
{
  SH2movwi(sh, 0, (10<<4)+9);
}
#define init_9_0_10_10
void FASTCALL call_9_0_10_10(SH2_struct* sh)
{
  SH2movwi(sh, 0, (10<<4)+10);
}
#define init_9_0_10_11
void FASTCALL call_9_0_10_11(SH2_struct* sh)
{
  SH2movwi(sh, 0, (10<<4)+11);
}
#define init_9_0_10_12
void FASTCALL call_9_0_10_12(SH2_struct* sh)
{
  SH2movwi(sh, 0, (10<<4)+12);
}
#define init_9_0_10_13
void FASTCALL call_9_0_10_13(SH2_struct* sh)
{
  SH2movwi(sh, 0, (10<<4)+13);
}
#define init_9_0_10_14
void FASTCALL call_9_0_10_14(SH2_struct* sh)
{
  SH2movwi(sh, 0, (10<<4)+14);
}
#define init_9_0_10_15
void FASTCALL call_9_0_10_15(SH2_struct* sh)
{
  SH2movwi(sh, 0, (10<<4)+15);
}
#define init_9_0_11_0
void FASTCALL call_9_0_11_0(SH2_struct* sh)
{
  SH2movwi(sh, 0, (11<<4)+0);
}
#define init_9_0_11_1
void FASTCALL call_9_0_11_1(SH2_struct* sh)
{
  SH2movwi(sh, 0, (11<<4)+1);
}
#define init_9_0_11_2
void FASTCALL call_9_0_11_2(SH2_struct* sh)
{
  SH2movwi(sh, 0, (11<<4)+2);
}
#define init_9_0_11_3
void FASTCALL call_9_0_11_3(SH2_struct* sh)
{
  SH2movwi(sh, 0, (11<<4)+3);
}
#define init_9_0_11_4
void FASTCALL call_9_0_11_4(SH2_struct* sh)
{
  SH2movwi(sh, 0, (11<<4)+4);
}
#define init_9_0_11_5
void FASTCALL call_9_0_11_5(SH2_struct* sh)
{
  SH2movwi(sh, 0, (11<<4)+5);
}
#define init_9_0_11_6
void FASTCALL call_9_0_11_6(SH2_struct* sh)
{
  SH2movwi(sh, 0, (11<<4)+6);
}
#define init_9_0_11_7
void FASTCALL call_9_0_11_7(SH2_struct* sh)
{
  SH2movwi(sh, 0, (11<<4)+7);
}
#define init_9_0_11_8
void FASTCALL call_9_0_11_8(SH2_struct* sh)
{
  SH2movwi(sh, 0, (11<<4)+8);
}
#define init_9_0_11_9
void FASTCALL call_9_0_11_9(SH2_struct* sh)
{
  SH2movwi(sh, 0, (11<<4)+9);
}
#define init_9_0_11_10
void FASTCALL call_9_0_11_10(SH2_struct* sh)
{
  SH2movwi(sh, 0, (11<<4)+10);
}
#define init_9_0_11_11
void FASTCALL call_9_0_11_11(SH2_struct* sh)
{
  SH2movwi(sh, 0, (11<<4)+11);
}
#define init_9_0_11_12
void FASTCALL call_9_0_11_12(SH2_struct* sh)
{
  SH2movwi(sh, 0, (11<<4)+12);
}
#define init_9_0_11_13
void FASTCALL call_9_0_11_13(SH2_struct* sh)
{
  SH2movwi(sh, 0, (11<<4)+13);
}
#define init_9_0_11_14
void FASTCALL call_9_0_11_14(SH2_struct* sh)
{
  SH2movwi(sh, 0, (11<<4)+14);
}
#define init_9_0_11_15
void FASTCALL call_9_0_11_15(SH2_struct* sh)
{
  SH2movwi(sh, 0, (11<<4)+15);
}
#define init_9_0_12_0
void FASTCALL call_9_0_12_0(SH2_struct* sh)
{
  SH2movwi(sh, 0, (12<<4)+0);
}
#define init_9_0_12_1
void FASTCALL call_9_0_12_1(SH2_struct* sh)
{
  SH2movwi(sh, 0, (12<<4)+1);
}
#define init_9_0_12_2
void FASTCALL call_9_0_12_2(SH2_struct* sh)
{
  SH2movwi(sh, 0, (12<<4)+2);
}
#define init_9_0_12_3
void FASTCALL call_9_0_12_3(SH2_struct* sh)
{
  SH2movwi(sh, 0, (12<<4)+3);
}
#define init_9_0_12_4
void FASTCALL call_9_0_12_4(SH2_struct* sh)
{
  SH2movwi(sh, 0, (12<<4)+4);
}
#define init_9_0_12_5
void FASTCALL call_9_0_12_5(SH2_struct* sh)
{
  SH2movwi(sh, 0, (12<<4)+5);
}
#define init_9_0_12_6
void FASTCALL call_9_0_12_6(SH2_struct* sh)
{
  SH2movwi(sh, 0, (12<<4)+6);
}
#define init_9_0_12_7
void FASTCALL call_9_0_12_7(SH2_struct* sh)
{
  SH2movwi(sh, 0, (12<<4)+7);
}
#define init_9_0_12_8
void FASTCALL call_9_0_12_8(SH2_struct* sh)
{
  SH2movwi(sh, 0, (12<<4)+8);
}
#define init_9_0_12_9
void FASTCALL call_9_0_12_9(SH2_struct* sh)
{
  SH2movwi(sh, 0, (12<<4)+9);
}
#define init_9_0_12_10
void FASTCALL call_9_0_12_10(SH2_struct* sh)
{
  SH2movwi(sh, 0, (12<<4)+10);
}
#define init_9_0_12_11
void FASTCALL call_9_0_12_11(SH2_struct* sh)
{
  SH2movwi(sh, 0, (12<<4)+11);
}
#define init_9_0_12_12
void FASTCALL call_9_0_12_12(SH2_struct* sh)
{
  SH2movwi(sh, 0, (12<<4)+12);
}
#define init_9_0_12_13
void FASTCALL call_9_0_12_13(SH2_struct* sh)
{
  SH2movwi(sh, 0, (12<<4)+13);
}
#define init_9_0_12_14
void FASTCALL call_9_0_12_14(SH2_struct* sh)
{
  SH2movwi(sh, 0, (12<<4)+14);
}
#define init_9_0_12_15
void FASTCALL call_9_0_12_15(SH2_struct* sh)
{
  SH2movwi(sh, 0, (12<<4)+15);
}
#define init_9_0_13_0
void FASTCALL call_9_0_13_0(SH2_struct* sh)
{
  SH2movwi(sh, 0, (13<<4)+0);
}
#define init_9_0_13_1
void FASTCALL call_9_0_13_1(SH2_struct* sh)
{
  SH2movwi(sh, 0, (13<<4)+1);
}
#define init_9_0_13_2
void FASTCALL call_9_0_13_2(SH2_struct* sh)
{
  SH2movwi(sh, 0, (13<<4)+2);
}
#define init_9_0_13_3
void FASTCALL call_9_0_13_3(SH2_struct* sh)
{
  SH2movwi(sh, 0, (13<<4)+3);
}
#define init_9_0_13_4
void FASTCALL call_9_0_13_4(SH2_struct* sh)
{
  SH2movwi(sh, 0, (13<<4)+4);
}
#define init_9_0_13_5
void FASTCALL call_9_0_13_5(SH2_struct* sh)
{
  SH2movwi(sh, 0, (13<<4)+5);
}
#define init_9_0_13_6
void FASTCALL call_9_0_13_6(SH2_struct* sh)
{
  SH2movwi(sh, 0, (13<<4)+6);
}
#define init_9_0_13_7
void FASTCALL call_9_0_13_7(SH2_struct* sh)
{
  SH2movwi(sh, 0, (13<<4)+7);
}
#define init_9_0_13_8
void FASTCALL call_9_0_13_8(SH2_struct* sh)
{
  SH2movwi(sh, 0, (13<<4)+8);
}
#define init_9_0_13_9
void FASTCALL call_9_0_13_9(SH2_struct* sh)
{
  SH2movwi(sh, 0, (13<<4)+9);
}
#define init_9_0_13_10
void FASTCALL call_9_0_13_10(SH2_struct* sh)
{
  SH2movwi(sh, 0, (13<<4)+10);
}
#define init_9_0_13_11
void FASTCALL call_9_0_13_11(SH2_struct* sh)
{
  SH2movwi(sh, 0, (13<<4)+11);
}
#define init_9_0_13_12
void FASTCALL call_9_0_13_12(SH2_struct* sh)
{
  SH2movwi(sh, 0, (13<<4)+12);
}
#define init_9_0_13_13
void FASTCALL call_9_0_13_13(SH2_struct* sh)
{
  SH2movwi(sh, 0, (13<<4)+13);
}
#define init_9_0_13_14
void FASTCALL call_9_0_13_14(SH2_struct* sh)
{
  SH2movwi(sh, 0, (13<<4)+14);
}
#define init_9_0_13_15
void FASTCALL call_9_0_13_15(SH2_struct* sh)
{
  SH2movwi(sh, 0, (13<<4)+15);
}
#define init_9_0_14_0
void FASTCALL call_9_0_14_0(SH2_struct* sh)
{
  SH2movwi(sh, 0, (14<<4)+0);
}
#define init_9_0_14_1
void FASTCALL call_9_0_14_1(SH2_struct* sh)
{
  SH2movwi(sh, 0, (14<<4)+1);
}
#define init_9_0_14_2
void FASTCALL call_9_0_14_2(SH2_struct* sh)
{
  SH2movwi(sh, 0, (14<<4)+2);
}
#define init_9_0_14_3
void FASTCALL call_9_0_14_3(SH2_struct* sh)
{
  SH2movwi(sh, 0, (14<<4)+3);
}
#define init_9_0_14_4
void FASTCALL call_9_0_14_4(SH2_struct* sh)
{
  SH2movwi(sh, 0, (14<<4)+4);
}
#define init_9_0_14_5
void FASTCALL call_9_0_14_5(SH2_struct* sh)
{
  SH2movwi(sh, 0, (14<<4)+5);
}
#define init_9_0_14_6
void FASTCALL call_9_0_14_6(SH2_struct* sh)
{
  SH2movwi(sh, 0, (14<<4)+6);
}
#define init_9_0_14_7
void FASTCALL call_9_0_14_7(SH2_struct* sh)
{
  SH2movwi(sh, 0, (14<<4)+7);
}
#define init_9_0_14_8
void FASTCALL call_9_0_14_8(SH2_struct* sh)
{
  SH2movwi(sh, 0, (14<<4)+8);
}
#define init_9_0_14_9
void FASTCALL call_9_0_14_9(SH2_struct* sh)
{
  SH2movwi(sh, 0, (14<<4)+9);
}
#define init_9_0_14_10
void FASTCALL call_9_0_14_10(SH2_struct* sh)
{
  SH2movwi(sh, 0, (14<<4)+10);
}
#define init_9_0_14_11
void FASTCALL call_9_0_14_11(SH2_struct* sh)
{
  SH2movwi(sh, 0, (14<<4)+11);
}
#define init_9_0_14_12
void FASTCALL call_9_0_14_12(SH2_struct* sh)
{
  SH2movwi(sh, 0, (14<<4)+12);
}
#define init_9_0_14_13
void FASTCALL call_9_0_14_13(SH2_struct* sh)
{
  SH2movwi(sh, 0, (14<<4)+13);
}
#define init_9_0_14_14
void FASTCALL call_9_0_14_14(SH2_struct* sh)
{
  SH2movwi(sh, 0, (14<<4)+14);
}
#define init_9_0_14_15
void FASTCALL call_9_0_14_15(SH2_struct* sh)
{
  SH2movwi(sh, 0, (14<<4)+15);
}
#define init_9_0_15_0
void FASTCALL call_9_0_15_0(SH2_struct* sh)
{
  SH2movwi(sh, 0, (15<<4)+0);
}
#define init_9_0_15_1
void FASTCALL call_9_0_15_1(SH2_struct* sh)
{
  SH2movwi(sh, 0, (15<<4)+1);
}
#define init_9_0_15_2
void FASTCALL call_9_0_15_2(SH2_struct* sh)
{
  SH2movwi(sh, 0, (15<<4)+2);
}
#define init_9_0_15_3
void FASTCALL call_9_0_15_3(SH2_struct* sh)
{
  SH2movwi(sh, 0, (15<<4)+3);
}
#define init_9_0_15_4
void FASTCALL call_9_0_15_4(SH2_struct* sh)
{
  SH2movwi(sh, 0, (15<<4)+4);
}
#define init_9_0_15_5
void FASTCALL call_9_0_15_5(SH2_struct* sh)
{
  SH2movwi(sh, 0, (15<<4)+5);
}
#define init_9_0_15_6
void FASTCALL call_9_0_15_6(SH2_struct* sh)
{
  SH2movwi(sh, 0, (15<<4)+6);
}
#define init_9_0_15_7
void FASTCALL call_9_0_15_7(SH2_struct* sh)
{
  SH2movwi(sh, 0, (15<<4)+7);
}
#define init_9_0_15_8
void FASTCALL call_9_0_15_8(SH2_struct* sh)
{
  SH2movwi(sh, 0, (15<<4)+8);
}
#define init_9_0_15_9
void FASTCALL call_9_0_15_9(SH2_struct* sh)
{
  SH2movwi(sh, 0, (15<<4)+9);
}
#define init_9_0_15_10
void FASTCALL call_9_0_15_10(SH2_struct* sh)
{
  SH2movwi(sh, 0, (15<<4)+10);
}
#define init_9_0_15_11
void FASTCALL call_9_0_15_11(SH2_struct* sh)
{
  SH2movwi(sh, 0, (15<<4)+11);
}
#define init_9_0_15_12
void FASTCALL call_9_0_15_12(SH2_struct* sh)
{
  SH2movwi(sh, 0, (15<<4)+12);
}
#define init_9_0_15_13
void FASTCALL call_9_0_15_13(SH2_struct* sh)
{
  SH2movwi(sh, 0, (15<<4)+13);
}
#define init_9_0_15_14
void FASTCALL call_9_0_15_14(SH2_struct* sh)
{
  SH2movwi(sh, 0, (15<<4)+14);
}
#define init_9_0_15_15
void FASTCALL call_9_0_15_15(SH2_struct* sh)
{
  SH2movwi(sh, 0, (15<<4)+15);
}
#define init_9_1_0_0
void FASTCALL call_9_1_0_0(SH2_struct* sh)
{
  SH2movwi(sh, 1, (0<<4)+0);
}
#define init_9_1_0_1
void FASTCALL call_9_1_0_1(SH2_struct* sh)
{
  SH2movwi(sh, 1, (0<<4)+1);
}
#define init_9_1_0_2
void FASTCALL call_9_1_0_2(SH2_struct* sh)
{
  SH2movwi(sh, 1, (0<<4)+2);
}
#define init_9_1_0_3
void FASTCALL call_9_1_0_3(SH2_struct* sh)
{
  SH2movwi(sh, 1, (0<<4)+3);
}
#define init_9_1_0_4
void FASTCALL call_9_1_0_4(SH2_struct* sh)
{
  SH2movwi(sh, 1, (0<<4)+4);
}
#define init_9_1_0_5
void FASTCALL call_9_1_0_5(SH2_struct* sh)
{
  SH2movwi(sh, 1, (0<<4)+5);
}
#define init_9_1_0_6
void FASTCALL call_9_1_0_6(SH2_struct* sh)
{
  SH2movwi(sh, 1, (0<<4)+6);
}
#define init_9_1_0_7
void FASTCALL call_9_1_0_7(SH2_struct* sh)
{
  SH2movwi(sh, 1, (0<<4)+7);
}
#define init_9_1_0_8
void FASTCALL call_9_1_0_8(SH2_struct* sh)
{
  SH2movwi(sh, 1, (0<<4)+8);
}
#define init_9_1_0_9
void FASTCALL call_9_1_0_9(SH2_struct* sh)
{
  SH2movwi(sh, 1, (0<<4)+9);
}
#define init_9_1_0_10
void FASTCALL call_9_1_0_10(SH2_struct* sh)
{
  SH2movwi(sh, 1, (0<<4)+10);
}
#define init_9_1_0_11
void FASTCALL call_9_1_0_11(SH2_struct* sh)
{
  SH2movwi(sh, 1, (0<<4)+11);
}
#define init_9_1_0_12
void FASTCALL call_9_1_0_12(SH2_struct* sh)
{
  SH2movwi(sh, 1, (0<<4)+12);
}
#define init_9_1_0_13
void FASTCALL call_9_1_0_13(SH2_struct* sh)
{
  SH2movwi(sh, 1, (0<<4)+13);
}
#define init_9_1_0_14
void FASTCALL call_9_1_0_14(SH2_struct* sh)
{
  SH2movwi(sh, 1, (0<<4)+14);
}
#define init_9_1_0_15
void FASTCALL call_9_1_0_15(SH2_struct* sh)
{
  SH2movwi(sh, 1, (0<<4)+15);
}
#define init_9_1_1_0
void FASTCALL call_9_1_1_0(SH2_struct* sh)
{
  SH2movwi(sh, 1, (1<<4)+0);
}
#define init_9_1_1_1
void FASTCALL call_9_1_1_1(SH2_struct* sh)
{
  SH2movwi(sh, 1, (1<<4)+1);
}
#define init_9_1_1_2
void FASTCALL call_9_1_1_2(SH2_struct* sh)
{
  SH2movwi(sh, 1, (1<<4)+2);
}
#define init_9_1_1_3
void FASTCALL call_9_1_1_3(SH2_struct* sh)
{
  SH2movwi(sh, 1, (1<<4)+3);
}
#define init_9_1_1_4
void FASTCALL call_9_1_1_4(SH2_struct* sh)
{
  SH2movwi(sh, 1, (1<<4)+4);
}
#define init_9_1_1_5
void FASTCALL call_9_1_1_5(SH2_struct* sh)
{
  SH2movwi(sh, 1, (1<<4)+5);
}
#define init_9_1_1_6
void FASTCALL call_9_1_1_6(SH2_struct* sh)
{
  SH2movwi(sh, 1, (1<<4)+6);
}
#define init_9_1_1_7
void FASTCALL call_9_1_1_7(SH2_struct* sh)
{
  SH2movwi(sh, 1, (1<<4)+7);
}
#define init_9_1_1_8
void FASTCALL call_9_1_1_8(SH2_struct* sh)
{
  SH2movwi(sh, 1, (1<<4)+8);
}
#define init_9_1_1_9
void FASTCALL call_9_1_1_9(SH2_struct* sh)
{
  SH2movwi(sh, 1, (1<<4)+9);
}
#define init_9_1_1_10
void FASTCALL call_9_1_1_10(SH2_struct* sh)
{
  SH2movwi(sh, 1, (1<<4)+10);
}
#define init_9_1_1_11
void FASTCALL call_9_1_1_11(SH2_struct* sh)
{
  SH2movwi(sh, 1, (1<<4)+11);
}
#define init_9_1_1_12
void FASTCALL call_9_1_1_12(SH2_struct* sh)
{
  SH2movwi(sh, 1, (1<<4)+12);
}
#define init_9_1_1_13
void FASTCALL call_9_1_1_13(SH2_struct* sh)
{
  SH2movwi(sh, 1, (1<<4)+13);
}
#define init_9_1_1_14
void FASTCALL call_9_1_1_14(SH2_struct* sh)
{
  SH2movwi(sh, 1, (1<<4)+14);
}
#define init_9_1_1_15
void FASTCALL call_9_1_1_15(SH2_struct* sh)
{
  SH2movwi(sh, 1, (1<<4)+15);
}
#define init_9_1_2_0
void FASTCALL call_9_1_2_0(SH2_struct* sh)
{
  SH2movwi(sh, 1, (2<<4)+0);
}
#define init_9_1_2_1
void FASTCALL call_9_1_2_1(SH2_struct* sh)
{
  SH2movwi(sh, 1, (2<<4)+1);
}
#define init_9_1_2_2
void FASTCALL call_9_1_2_2(SH2_struct* sh)
{
  SH2movwi(sh, 1, (2<<4)+2);
}
#define init_9_1_2_3
void FASTCALL call_9_1_2_3(SH2_struct* sh)
{
  SH2movwi(sh, 1, (2<<4)+3);
}
#define init_9_1_2_4
void FASTCALL call_9_1_2_4(SH2_struct* sh)
{
  SH2movwi(sh, 1, (2<<4)+4);
}
#define init_9_1_2_5
void FASTCALL call_9_1_2_5(SH2_struct* sh)
{
  SH2movwi(sh, 1, (2<<4)+5);
}
#define init_9_1_2_6
void FASTCALL call_9_1_2_6(SH2_struct* sh)
{
  SH2movwi(sh, 1, (2<<4)+6);
}
#define init_9_1_2_7
void FASTCALL call_9_1_2_7(SH2_struct* sh)
{
  SH2movwi(sh, 1, (2<<4)+7);
}
#define init_9_1_2_8
void FASTCALL call_9_1_2_8(SH2_struct* sh)
{
  SH2movwi(sh, 1, (2<<4)+8);
}
#define init_9_1_2_9
void FASTCALL call_9_1_2_9(SH2_struct* sh)
{
  SH2movwi(sh, 1, (2<<4)+9);
}
#define init_9_1_2_10
void FASTCALL call_9_1_2_10(SH2_struct* sh)
{
  SH2movwi(sh, 1, (2<<4)+10);
}
#define init_9_1_2_11
void FASTCALL call_9_1_2_11(SH2_struct* sh)
{
  SH2movwi(sh, 1, (2<<4)+11);
}
#define init_9_1_2_12
void FASTCALL call_9_1_2_12(SH2_struct* sh)
{
  SH2movwi(sh, 1, (2<<4)+12);
}
#define init_9_1_2_13
void FASTCALL call_9_1_2_13(SH2_struct* sh)
{
  SH2movwi(sh, 1, (2<<4)+13);
}
#define init_9_1_2_14
void FASTCALL call_9_1_2_14(SH2_struct* sh)
{
  SH2movwi(sh, 1, (2<<4)+14);
}
#define init_9_1_2_15
void FASTCALL call_9_1_2_15(SH2_struct* sh)
{
  SH2movwi(sh, 1, (2<<4)+15);
}
#define init_9_1_3_0
void FASTCALL call_9_1_3_0(SH2_struct* sh)
{
  SH2movwi(sh, 1, (3<<4)+0);
}
#define init_9_1_3_1
void FASTCALL call_9_1_3_1(SH2_struct* sh)
{
  SH2movwi(sh, 1, (3<<4)+1);
}
#define init_9_1_3_2
void FASTCALL call_9_1_3_2(SH2_struct* sh)
{
  SH2movwi(sh, 1, (3<<4)+2);
}
#define init_9_1_3_3
void FASTCALL call_9_1_3_3(SH2_struct* sh)
{
  SH2movwi(sh, 1, (3<<4)+3);
}
#define init_9_1_3_4
void FASTCALL call_9_1_3_4(SH2_struct* sh)
{
  SH2movwi(sh, 1, (3<<4)+4);
}
#define init_9_1_3_5
void FASTCALL call_9_1_3_5(SH2_struct* sh)
{
  SH2movwi(sh, 1, (3<<4)+5);
}
#define init_9_1_3_6
void FASTCALL call_9_1_3_6(SH2_struct* sh)
{
  SH2movwi(sh, 1, (3<<4)+6);
}
#define init_9_1_3_7
void FASTCALL call_9_1_3_7(SH2_struct* sh)
{
  SH2movwi(sh, 1, (3<<4)+7);
}
#define init_9_1_3_8
void FASTCALL call_9_1_3_8(SH2_struct* sh)
{
  SH2movwi(sh, 1, (3<<4)+8);
}
#define init_9_1_3_9
void FASTCALL call_9_1_3_9(SH2_struct* sh)
{
  SH2movwi(sh, 1, (3<<4)+9);
}
#define init_9_1_3_10
void FASTCALL call_9_1_3_10(SH2_struct* sh)
{
  SH2movwi(sh, 1, (3<<4)+10);
}
#define init_9_1_3_11
void FASTCALL call_9_1_3_11(SH2_struct* sh)
{
  SH2movwi(sh, 1, (3<<4)+11);
}
#define init_9_1_3_12
void FASTCALL call_9_1_3_12(SH2_struct* sh)
{
  SH2movwi(sh, 1, (3<<4)+12);
}
#define init_9_1_3_13
void FASTCALL call_9_1_3_13(SH2_struct* sh)
{
  SH2movwi(sh, 1, (3<<4)+13);
}
#define init_9_1_3_14
void FASTCALL call_9_1_3_14(SH2_struct* sh)
{
  SH2movwi(sh, 1, (3<<4)+14);
}
#define init_9_1_3_15
void FASTCALL call_9_1_3_15(SH2_struct* sh)
{
  SH2movwi(sh, 1, (3<<4)+15);
}
#define init_9_1_4_0
void FASTCALL call_9_1_4_0(SH2_struct* sh)
{
  SH2movwi(sh, 1, (4<<4)+0);
}
#define init_9_1_4_1
void FASTCALL call_9_1_4_1(SH2_struct* sh)
{
  SH2movwi(sh, 1, (4<<4)+1);
}
#define init_9_1_4_2
void FASTCALL call_9_1_4_2(SH2_struct* sh)
{
  SH2movwi(sh, 1, (4<<4)+2);
}
#define init_9_1_4_3
void FASTCALL call_9_1_4_3(SH2_struct* sh)
{
  SH2movwi(sh, 1, (4<<4)+3);
}
#define init_9_1_4_4
void FASTCALL call_9_1_4_4(SH2_struct* sh)
{
  SH2movwi(sh, 1, (4<<4)+4);
}
#define init_9_1_4_5
void FASTCALL call_9_1_4_5(SH2_struct* sh)
{
  SH2movwi(sh, 1, (4<<4)+5);
}
#define init_9_1_4_6
void FASTCALL call_9_1_4_6(SH2_struct* sh)
{
  SH2movwi(sh, 1, (4<<4)+6);
}
#define init_9_1_4_7
void FASTCALL call_9_1_4_7(SH2_struct* sh)
{
  SH2movwi(sh, 1, (4<<4)+7);
}
#define init_9_1_4_8
void FASTCALL call_9_1_4_8(SH2_struct* sh)
{
  SH2movwi(sh, 1, (4<<4)+8);
}
#define init_9_1_4_9
void FASTCALL call_9_1_4_9(SH2_struct* sh)
{
  SH2movwi(sh, 1, (4<<4)+9);
}
#define init_9_1_4_10
void FASTCALL call_9_1_4_10(SH2_struct* sh)
{
  SH2movwi(sh, 1, (4<<4)+10);
}
#define init_9_1_4_11
void FASTCALL call_9_1_4_11(SH2_struct* sh)
{
  SH2movwi(sh, 1, (4<<4)+11);
}
#define init_9_1_4_12
void FASTCALL call_9_1_4_12(SH2_struct* sh)
{
  SH2movwi(sh, 1, (4<<4)+12);
}
#define init_9_1_4_13
void FASTCALL call_9_1_4_13(SH2_struct* sh)
{
  SH2movwi(sh, 1, (4<<4)+13);
}
#define init_9_1_4_14
void FASTCALL call_9_1_4_14(SH2_struct* sh)
{
  SH2movwi(sh, 1, (4<<4)+14);
}
#define init_9_1_4_15
void FASTCALL call_9_1_4_15(SH2_struct* sh)
{
  SH2movwi(sh, 1, (4<<4)+15);
}
#define init_9_1_5_0
void FASTCALL call_9_1_5_0(SH2_struct* sh)
{
  SH2movwi(sh, 1, (5<<4)+0);
}
#define init_9_1_5_1
void FASTCALL call_9_1_5_1(SH2_struct* sh)
{
  SH2movwi(sh, 1, (5<<4)+1);
}
#define init_9_1_5_2
void FASTCALL call_9_1_5_2(SH2_struct* sh)
{
  SH2movwi(sh, 1, (5<<4)+2);
}
#define init_9_1_5_3
void FASTCALL call_9_1_5_3(SH2_struct* sh)
{
  SH2movwi(sh, 1, (5<<4)+3);
}
#define init_9_1_5_4
void FASTCALL call_9_1_5_4(SH2_struct* sh)
{
  SH2movwi(sh, 1, (5<<4)+4);
}
#define init_9_1_5_5
void FASTCALL call_9_1_5_5(SH2_struct* sh)
{
  SH2movwi(sh, 1, (5<<4)+5);
}
#define init_9_1_5_6
void FASTCALL call_9_1_5_6(SH2_struct* sh)
{
  SH2movwi(sh, 1, (5<<4)+6);
}
#define init_9_1_5_7
void FASTCALL call_9_1_5_7(SH2_struct* sh)
{
  SH2movwi(sh, 1, (5<<4)+7);
}
#define init_9_1_5_8
void FASTCALL call_9_1_5_8(SH2_struct* sh)
{
  SH2movwi(sh, 1, (5<<4)+8);
}
#define init_9_1_5_9
void FASTCALL call_9_1_5_9(SH2_struct* sh)
{
  SH2movwi(sh, 1, (5<<4)+9);
}
#define init_9_1_5_10
void FASTCALL call_9_1_5_10(SH2_struct* sh)
{
  SH2movwi(sh, 1, (5<<4)+10);
}
#define init_9_1_5_11
void FASTCALL call_9_1_5_11(SH2_struct* sh)
{
  SH2movwi(sh, 1, (5<<4)+11);
}
#define init_9_1_5_12
void FASTCALL call_9_1_5_12(SH2_struct* sh)
{
  SH2movwi(sh, 1, (5<<4)+12);
}
#define init_9_1_5_13
void FASTCALL call_9_1_5_13(SH2_struct* sh)
{
  SH2movwi(sh, 1, (5<<4)+13);
}
#define init_9_1_5_14
void FASTCALL call_9_1_5_14(SH2_struct* sh)
{
  SH2movwi(sh, 1, (5<<4)+14);
}
#define init_9_1_5_15
void FASTCALL call_9_1_5_15(SH2_struct* sh)
{
  SH2movwi(sh, 1, (5<<4)+15);
}
#define init_9_1_6_0
void FASTCALL call_9_1_6_0(SH2_struct* sh)
{
  SH2movwi(sh, 1, (6<<4)+0);
}
#define init_9_1_6_1
void FASTCALL call_9_1_6_1(SH2_struct* sh)
{
  SH2movwi(sh, 1, (6<<4)+1);
}
#define init_9_1_6_2
void FASTCALL call_9_1_6_2(SH2_struct* sh)
{
  SH2movwi(sh, 1, (6<<4)+2);
}
#define init_9_1_6_3
void FASTCALL call_9_1_6_3(SH2_struct* sh)
{
  SH2movwi(sh, 1, (6<<4)+3);
}
#define init_9_1_6_4
void FASTCALL call_9_1_6_4(SH2_struct* sh)
{
  SH2movwi(sh, 1, (6<<4)+4);
}
#define init_9_1_6_5
void FASTCALL call_9_1_6_5(SH2_struct* sh)
{
  SH2movwi(sh, 1, (6<<4)+5);
}
#define init_9_1_6_6
void FASTCALL call_9_1_6_6(SH2_struct* sh)
{
  SH2movwi(sh, 1, (6<<4)+6);
}
#define init_9_1_6_7
void FASTCALL call_9_1_6_7(SH2_struct* sh)
{
  SH2movwi(sh, 1, (6<<4)+7);
}
#define init_9_1_6_8
void FASTCALL call_9_1_6_8(SH2_struct* sh)
{
  SH2movwi(sh, 1, (6<<4)+8);
}
#define init_9_1_6_9
void FASTCALL call_9_1_6_9(SH2_struct* sh)
{
  SH2movwi(sh, 1, (6<<4)+9);
}
#define init_9_1_6_10
void FASTCALL call_9_1_6_10(SH2_struct* sh)
{
  SH2movwi(sh, 1, (6<<4)+10);
}
#define init_9_1_6_11
void FASTCALL call_9_1_6_11(SH2_struct* sh)
{
  SH2movwi(sh, 1, (6<<4)+11);
}
#define init_9_1_6_12
void FASTCALL call_9_1_6_12(SH2_struct* sh)
{
  SH2movwi(sh, 1, (6<<4)+12);
}
#define init_9_1_6_13
void FASTCALL call_9_1_6_13(SH2_struct* sh)
{
  SH2movwi(sh, 1, (6<<4)+13);
}
#define init_9_1_6_14
void FASTCALL call_9_1_6_14(SH2_struct* sh)
{
  SH2movwi(sh, 1, (6<<4)+14);
}
#define init_9_1_6_15
void FASTCALL call_9_1_6_15(SH2_struct* sh)
{
  SH2movwi(sh, 1, (6<<4)+15);
}
#define init_9_1_7_0
void FASTCALL call_9_1_7_0(SH2_struct* sh)
{
  SH2movwi(sh, 1, (7<<4)+0);
}
#define init_9_1_7_1
void FASTCALL call_9_1_7_1(SH2_struct* sh)
{
  SH2movwi(sh, 1, (7<<4)+1);
}
#define init_9_1_7_2
void FASTCALL call_9_1_7_2(SH2_struct* sh)
{
  SH2movwi(sh, 1, (7<<4)+2);
}
#define init_9_1_7_3
void FASTCALL call_9_1_7_3(SH2_struct* sh)
{
  SH2movwi(sh, 1, (7<<4)+3);
}
#define init_9_1_7_4
void FASTCALL call_9_1_7_4(SH2_struct* sh)
{
  SH2movwi(sh, 1, (7<<4)+4);
}
#define init_9_1_7_5
void FASTCALL call_9_1_7_5(SH2_struct* sh)
{
  SH2movwi(sh, 1, (7<<4)+5);
}
#define init_9_1_7_6
void FASTCALL call_9_1_7_6(SH2_struct* sh)
{
  SH2movwi(sh, 1, (7<<4)+6);
}
#define init_9_1_7_7
void FASTCALL call_9_1_7_7(SH2_struct* sh)
{
  SH2movwi(sh, 1, (7<<4)+7);
}
#define init_9_1_7_8
void FASTCALL call_9_1_7_8(SH2_struct* sh)
{
  SH2movwi(sh, 1, (7<<4)+8);
}
#define init_9_1_7_9
void FASTCALL call_9_1_7_9(SH2_struct* sh)
{
  SH2movwi(sh, 1, (7<<4)+9);
}
#define init_9_1_7_10
void FASTCALL call_9_1_7_10(SH2_struct* sh)
{
  SH2movwi(sh, 1, (7<<4)+10);
}
#define init_9_1_7_11
void FASTCALL call_9_1_7_11(SH2_struct* sh)
{
  SH2movwi(sh, 1, (7<<4)+11);
}
#define init_9_1_7_12
void FASTCALL call_9_1_7_12(SH2_struct* sh)
{
  SH2movwi(sh, 1, (7<<4)+12);
}
#define init_9_1_7_13
void FASTCALL call_9_1_7_13(SH2_struct* sh)
{
  SH2movwi(sh, 1, (7<<4)+13);
}
#define init_9_1_7_14
void FASTCALL call_9_1_7_14(SH2_struct* sh)
{
  SH2movwi(sh, 1, (7<<4)+14);
}
#define init_9_1_7_15
void FASTCALL call_9_1_7_15(SH2_struct* sh)
{
  SH2movwi(sh, 1, (7<<4)+15);
}
#define init_9_1_8_0
void FASTCALL call_9_1_8_0(SH2_struct* sh)
{
  SH2movwi(sh, 1, (8<<4)+0);
}
#define init_9_1_8_1
void FASTCALL call_9_1_8_1(SH2_struct* sh)
{
  SH2movwi(sh, 1, (8<<4)+1);
}
#define init_9_1_8_2
void FASTCALL call_9_1_8_2(SH2_struct* sh)
{
  SH2movwi(sh, 1, (8<<4)+2);
}
#define init_9_1_8_3
void FASTCALL call_9_1_8_3(SH2_struct* sh)
{
  SH2movwi(sh, 1, (8<<4)+3);
}
#define init_9_1_8_4
void FASTCALL call_9_1_8_4(SH2_struct* sh)
{
  SH2movwi(sh, 1, (8<<4)+4);
}
#define init_9_1_8_5
void FASTCALL call_9_1_8_5(SH2_struct* sh)
{
  SH2movwi(sh, 1, (8<<4)+5);
}
#define init_9_1_8_6
void FASTCALL call_9_1_8_6(SH2_struct* sh)
{
  SH2movwi(sh, 1, (8<<4)+6);
}
#define init_9_1_8_7
void FASTCALL call_9_1_8_7(SH2_struct* sh)
{
  SH2movwi(sh, 1, (8<<4)+7);
}
#define init_9_1_8_8
void FASTCALL call_9_1_8_8(SH2_struct* sh)
{
  SH2movwi(sh, 1, (8<<4)+8);
}
#define init_9_1_8_9
void FASTCALL call_9_1_8_9(SH2_struct* sh)
{
  SH2movwi(sh, 1, (8<<4)+9);
}
#define init_9_1_8_10
void FASTCALL call_9_1_8_10(SH2_struct* sh)
{
  SH2movwi(sh, 1, (8<<4)+10);
}
#define init_9_1_8_11
void FASTCALL call_9_1_8_11(SH2_struct* sh)
{
  SH2movwi(sh, 1, (8<<4)+11);
}
#define init_9_1_8_12
void FASTCALL call_9_1_8_12(SH2_struct* sh)
{
  SH2movwi(sh, 1, (8<<4)+12);
}
#define init_9_1_8_13
void FASTCALL call_9_1_8_13(SH2_struct* sh)
{
  SH2movwi(sh, 1, (8<<4)+13);
}
#define init_9_1_8_14
void FASTCALL call_9_1_8_14(SH2_struct* sh)
{
  SH2movwi(sh, 1, (8<<4)+14);
}
#define init_9_1_8_15
void FASTCALL call_9_1_8_15(SH2_struct* sh)
{
  SH2movwi(sh, 1, (8<<4)+15);
}
#define init_9_1_9_0
void FASTCALL call_9_1_9_0(SH2_struct* sh)
{
  SH2movwi(sh, 1, (9<<4)+0);
}
#define init_9_1_9_1
void FASTCALL call_9_1_9_1(SH2_struct* sh)
{
  SH2movwi(sh, 1, (9<<4)+1);
}
#define init_9_1_9_2
void FASTCALL call_9_1_9_2(SH2_struct* sh)
{
  SH2movwi(sh, 1, (9<<4)+2);
}
#define init_9_1_9_3
void FASTCALL call_9_1_9_3(SH2_struct* sh)
{
  SH2movwi(sh, 1, (9<<4)+3);
}
#define init_9_1_9_4
void FASTCALL call_9_1_9_4(SH2_struct* sh)
{
  SH2movwi(sh, 1, (9<<4)+4);
}
#define init_9_1_9_5
void FASTCALL call_9_1_9_5(SH2_struct* sh)
{
  SH2movwi(sh, 1, (9<<4)+5);
}
#define init_9_1_9_6
void FASTCALL call_9_1_9_6(SH2_struct* sh)
{
  SH2movwi(sh, 1, (9<<4)+6);
}
#define init_9_1_9_7
void FASTCALL call_9_1_9_7(SH2_struct* sh)
{
  SH2movwi(sh, 1, (9<<4)+7);
}
#define init_9_1_9_8
void FASTCALL call_9_1_9_8(SH2_struct* sh)
{
  SH2movwi(sh, 1, (9<<4)+8);
}
#define init_9_1_9_9
void FASTCALL call_9_1_9_9(SH2_struct* sh)
{
  SH2movwi(sh, 1, (9<<4)+9);
}
#define init_9_1_9_10
void FASTCALL call_9_1_9_10(SH2_struct* sh)
{
  SH2movwi(sh, 1, (9<<4)+10);
}
#define init_9_1_9_11
void FASTCALL call_9_1_9_11(SH2_struct* sh)
{
  SH2movwi(sh, 1, (9<<4)+11);
}
#define init_9_1_9_12
void FASTCALL call_9_1_9_12(SH2_struct* sh)
{
  SH2movwi(sh, 1, (9<<4)+12);
}
#define init_9_1_9_13
void FASTCALL call_9_1_9_13(SH2_struct* sh)
{
  SH2movwi(sh, 1, (9<<4)+13);
}
#define init_9_1_9_14
void FASTCALL call_9_1_9_14(SH2_struct* sh)
{
  SH2movwi(sh, 1, (9<<4)+14);
}
#define init_9_1_9_15
void FASTCALL call_9_1_9_15(SH2_struct* sh)
{
  SH2movwi(sh, 1, (9<<4)+15);
}
#define init_9_1_10_0
void FASTCALL call_9_1_10_0(SH2_struct* sh)
{
  SH2movwi(sh, 1, (10<<4)+0);
}
#define init_9_1_10_1
void FASTCALL call_9_1_10_1(SH2_struct* sh)
{
  SH2movwi(sh, 1, (10<<4)+1);
}
#define init_9_1_10_2
void FASTCALL call_9_1_10_2(SH2_struct* sh)
{
  SH2movwi(sh, 1, (10<<4)+2);
}
#define init_9_1_10_3
void FASTCALL call_9_1_10_3(SH2_struct* sh)
{
  SH2movwi(sh, 1, (10<<4)+3);
}
#define init_9_1_10_4
void FASTCALL call_9_1_10_4(SH2_struct* sh)
{
  SH2movwi(sh, 1, (10<<4)+4);
}
#define init_9_1_10_5
void FASTCALL call_9_1_10_5(SH2_struct* sh)
{
  SH2movwi(sh, 1, (10<<4)+5);
}
#define init_9_1_10_6
void FASTCALL call_9_1_10_6(SH2_struct* sh)
{
  SH2movwi(sh, 1, (10<<4)+6);
}
#define init_9_1_10_7
void FASTCALL call_9_1_10_7(SH2_struct* sh)
{
  SH2movwi(sh, 1, (10<<4)+7);
}
#define init_9_1_10_8
void FASTCALL call_9_1_10_8(SH2_struct* sh)
{
  SH2movwi(sh, 1, (10<<4)+8);
}
#define init_9_1_10_9
void FASTCALL call_9_1_10_9(SH2_struct* sh)
{
  SH2movwi(sh, 1, (10<<4)+9);
}
#define init_9_1_10_10
void FASTCALL call_9_1_10_10(SH2_struct* sh)
{
  SH2movwi(sh, 1, (10<<4)+10);
}
#define init_9_1_10_11
void FASTCALL call_9_1_10_11(SH2_struct* sh)
{
  SH2movwi(sh, 1, (10<<4)+11);
}
#define init_9_1_10_12
void FASTCALL call_9_1_10_12(SH2_struct* sh)
{
  SH2movwi(sh, 1, (10<<4)+12);
}
#define init_9_1_10_13
void FASTCALL call_9_1_10_13(SH2_struct* sh)
{
  SH2movwi(sh, 1, (10<<4)+13);
}
#define init_9_1_10_14
void FASTCALL call_9_1_10_14(SH2_struct* sh)
{
  SH2movwi(sh, 1, (10<<4)+14);
}
#define init_9_1_10_15
void FASTCALL call_9_1_10_15(SH2_struct* sh)
{
  SH2movwi(sh, 1, (10<<4)+15);
}
#define init_9_1_11_0
void FASTCALL call_9_1_11_0(SH2_struct* sh)
{
  SH2movwi(sh, 1, (11<<4)+0);
}
#define init_9_1_11_1
void FASTCALL call_9_1_11_1(SH2_struct* sh)
{
  SH2movwi(sh, 1, (11<<4)+1);
}
#define init_9_1_11_2
void FASTCALL call_9_1_11_2(SH2_struct* sh)
{
  SH2movwi(sh, 1, (11<<4)+2);
}
#define init_9_1_11_3
void FASTCALL call_9_1_11_3(SH2_struct* sh)
{
  SH2movwi(sh, 1, (11<<4)+3);
}
#define init_9_1_11_4
void FASTCALL call_9_1_11_4(SH2_struct* sh)
{
  SH2movwi(sh, 1, (11<<4)+4);
}
#define init_9_1_11_5
void FASTCALL call_9_1_11_5(SH2_struct* sh)
{
  SH2movwi(sh, 1, (11<<4)+5);
}
#define init_9_1_11_6
void FASTCALL call_9_1_11_6(SH2_struct* sh)
{
  SH2movwi(sh, 1, (11<<4)+6);
}
#define init_9_1_11_7
void FASTCALL call_9_1_11_7(SH2_struct* sh)
{
  SH2movwi(sh, 1, (11<<4)+7);
}
#define init_9_1_11_8
void FASTCALL call_9_1_11_8(SH2_struct* sh)
{
  SH2movwi(sh, 1, (11<<4)+8);
}
#define init_9_1_11_9
void FASTCALL call_9_1_11_9(SH2_struct* sh)
{
  SH2movwi(sh, 1, (11<<4)+9);
}
#define init_9_1_11_10
void FASTCALL call_9_1_11_10(SH2_struct* sh)
{
  SH2movwi(sh, 1, (11<<4)+10);
}
#define init_9_1_11_11
void FASTCALL call_9_1_11_11(SH2_struct* sh)
{
  SH2movwi(sh, 1, (11<<4)+11);
}
#define init_9_1_11_12
void FASTCALL call_9_1_11_12(SH2_struct* sh)
{
  SH2movwi(sh, 1, (11<<4)+12);
}
#define init_9_1_11_13
void FASTCALL call_9_1_11_13(SH2_struct* sh)
{
  SH2movwi(sh, 1, (11<<4)+13);
}
#define init_9_1_11_14
void FASTCALL call_9_1_11_14(SH2_struct* sh)
{
  SH2movwi(sh, 1, (11<<4)+14);
}
#define init_9_1_11_15
void FASTCALL call_9_1_11_15(SH2_struct* sh)
{
  SH2movwi(sh, 1, (11<<4)+15);
}
#define init_9_1_12_0
void FASTCALL call_9_1_12_0(SH2_struct* sh)
{
  SH2movwi(sh, 1, (12<<4)+0);
}
#define init_9_1_12_1
void FASTCALL call_9_1_12_1(SH2_struct* sh)
{
  SH2movwi(sh, 1, (12<<4)+1);
}
#define init_9_1_12_2
void FASTCALL call_9_1_12_2(SH2_struct* sh)
{
  SH2movwi(sh, 1, (12<<4)+2);
}
#define init_9_1_12_3
void FASTCALL call_9_1_12_3(SH2_struct* sh)
{
  SH2movwi(sh, 1, (12<<4)+3);
}
#define init_9_1_12_4
void FASTCALL call_9_1_12_4(SH2_struct* sh)
{
  SH2movwi(sh, 1, (12<<4)+4);
}
#define init_9_1_12_5
void FASTCALL call_9_1_12_5(SH2_struct* sh)
{
  SH2movwi(sh, 1, (12<<4)+5);
}
#define init_9_1_12_6
void FASTCALL call_9_1_12_6(SH2_struct* sh)
{
  SH2movwi(sh, 1, (12<<4)+6);
}
#define init_9_1_12_7
void FASTCALL call_9_1_12_7(SH2_struct* sh)
{
  SH2movwi(sh, 1, (12<<4)+7);
}
#define init_9_1_12_8
void FASTCALL call_9_1_12_8(SH2_struct* sh)
{
  SH2movwi(sh, 1, (12<<4)+8);
}
#define init_9_1_12_9
void FASTCALL call_9_1_12_9(SH2_struct* sh)
{
  SH2movwi(sh, 1, (12<<4)+9);
}
#define init_9_1_12_10
void FASTCALL call_9_1_12_10(SH2_struct* sh)
{
  SH2movwi(sh, 1, (12<<4)+10);
}
#define init_9_1_12_11
void FASTCALL call_9_1_12_11(SH2_struct* sh)
{
  SH2movwi(sh, 1, (12<<4)+11);
}
#define init_9_1_12_12
void FASTCALL call_9_1_12_12(SH2_struct* sh)
{
  SH2movwi(sh, 1, (12<<4)+12);
}
#define init_9_1_12_13
void FASTCALL call_9_1_12_13(SH2_struct* sh)
{
  SH2movwi(sh, 1, (12<<4)+13);
}
#define init_9_1_12_14
void FASTCALL call_9_1_12_14(SH2_struct* sh)
{
  SH2movwi(sh, 1, (12<<4)+14);
}
#define init_9_1_12_15
void FASTCALL call_9_1_12_15(SH2_struct* sh)
{
  SH2movwi(sh, 1, (12<<4)+15);
}
#define init_9_1_13_0
void FASTCALL call_9_1_13_0(SH2_struct* sh)
{
  SH2movwi(sh, 1, (13<<4)+0);
}
#define init_9_1_13_1
void FASTCALL call_9_1_13_1(SH2_struct* sh)
{
  SH2movwi(sh, 1, (13<<4)+1);
}
#define init_9_1_13_2
void FASTCALL call_9_1_13_2(SH2_struct* sh)
{
  SH2movwi(sh, 1, (13<<4)+2);
}
#define init_9_1_13_3
void FASTCALL call_9_1_13_3(SH2_struct* sh)
{
  SH2movwi(sh, 1, (13<<4)+3);
}
#define init_9_1_13_4
void FASTCALL call_9_1_13_4(SH2_struct* sh)
{
  SH2movwi(sh, 1, (13<<4)+4);
}
#define init_9_1_13_5
void FASTCALL call_9_1_13_5(SH2_struct* sh)
{
  SH2movwi(sh, 1, (13<<4)+5);
}
#define init_9_1_13_6
void FASTCALL call_9_1_13_6(SH2_struct* sh)
{
  SH2movwi(sh, 1, (13<<4)+6);
}
#define init_9_1_13_7
void FASTCALL call_9_1_13_7(SH2_struct* sh)
{
  SH2movwi(sh, 1, (13<<4)+7);
}
#define init_9_1_13_8
void FASTCALL call_9_1_13_8(SH2_struct* sh)
{
  SH2movwi(sh, 1, (13<<4)+8);
}
#define init_9_1_13_9
void FASTCALL call_9_1_13_9(SH2_struct* sh)
{
  SH2movwi(sh, 1, (13<<4)+9);
}
#define init_9_1_13_10
void FASTCALL call_9_1_13_10(SH2_struct* sh)
{
  SH2movwi(sh, 1, (13<<4)+10);
}
#define init_9_1_13_11
void FASTCALL call_9_1_13_11(SH2_struct* sh)
{
  SH2movwi(sh, 1, (13<<4)+11);
}
#define init_9_1_13_12
void FASTCALL call_9_1_13_12(SH2_struct* sh)
{
  SH2movwi(sh, 1, (13<<4)+12);
}
#define init_9_1_13_13
void FASTCALL call_9_1_13_13(SH2_struct* sh)
{
  SH2movwi(sh, 1, (13<<4)+13);
}
#define init_9_1_13_14
void FASTCALL call_9_1_13_14(SH2_struct* sh)
{
  SH2movwi(sh, 1, (13<<4)+14);
}
#define init_9_1_13_15
void FASTCALL call_9_1_13_15(SH2_struct* sh)
{
  SH2movwi(sh, 1, (13<<4)+15);
}
#define init_9_1_14_0
void FASTCALL call_9_1_14_0(SH2_struct* sh)
{
  SH2movwi(sh, 1, (14<<4)+0);
}
#define init_9_1_14_1
void FASTCALL call_9_1_14_1(SH2_struct* sh)
{
  SH2movwi(sh, 1, (14<<4)+1);
}
#define init_9_1_14_2
void FASTCALL call_9_1_14_2(SH2_struct* sh)
{
  SH2movwi(sh, 1, (14<<4)+2);
}
#define init_9_1_14_3
void FASTCALL call_9_1_14_3(SH2_struct* sh)
{
  SH2movwi(sh, 1, (14<<4)+3);
}
#define init_9_1_14_4
void FASTCALL call_9_1_14_4(SH2_struct* sh)
{
  SH2movwi(sh, 1, (14<<4)+4);
}
#define init_9_1_14_5
void FASTCALL call_9_1_14_5(SH2_struct* sh)
{
  SH2movwi(sh, 1, (14<<4)+5);
}
#define init_9_1_14_6
void FASTCALL call_9_1_14_6(SH2_struct* sh)
{
  SH2movwi(sh, 1, (14<<4)+6);
}
#define init_9_1_14_7
void FASTCALL call_9_1_14_7(SH2_struct* sh)
{
  SH2movwi(sh, 1, (14<<4)+7);
}
#define init_9_1_14_8
void FASTCALL call_9_1_14_8(SH2_struct* sh)
{
  SH2movwi(sh, 1, (14<<4)+8);
}
#define init_9_1_14_9
void FASTCALL call_9_1_14_9(SH2_struct* sh)
{
  SH2movwi(sh, 1, (14<<4)+9);
}
#define init_9_1_14_10
void FASTCALL call_9_1_14_10(SH2_struct* sh)
{
  SH2movwi(sh, 1, (14<<4)+10);
}
#define init_9_1_14_11
void FASTCALL call_9_1_14_11(SH2_struct* sh)
{
  SH2movwi(sh, 1, (14<<4)+11);
}
#define init_9_1_14_12
void FASTCALL call_9_1_14_12(SH2_struct* sh)
{
  SH2movwi(sh, 1, (14<<4)+12);
}
#define init_9_1_14_13
void FASTCALL call_9_1_14_13(SH2_struct* sh)
{
  SH2movwi(sh, 1, (14<<4)+13);
}
#define init_9_1_14_14
void FASTCALL call_9_1_14_14(SH2_struct* sh)
{
  SH2movwi(sh, 1, (14<<4)+14);
}
#define init_9_1_14_15
void FASTCALL call_9_1_14_15(SH2_struct* sh)
{
  SH2movwi(sh, 1, (14<<4)+15);
}
#define init_9_1_15_0
void FASTCALL call_9_1_15_0(SH2_struct* sh)
{
  SH2movwi(sh, 1, (15<<4)+0);
}
#define init_9_1_15_1
void FASTCALL call_9_1_15_1(SH2_struct* sh)
{
  SH2movwi(sh, 1, (15<<4)+1);
}
#define init_9_1_15_2
void FASTCALL call_9_1_15_2(SH2_struct* sh)
{
  SH2movwi(sh, 1, (15<<4)+2);
}
#define init_9_1_15_3
void FASTCALL call_9_1_15_3(SH2_struct* sh)
{
  SH2movwi(sh, 1, (15<<4)+3);
}
#define init_9_1_15_4
void FASTCALL call_9_1_15_4(SH2_struct* sh)
{
  SH2movwi(sh, 1, (15<<4)+4);
}
#define init_9_1_15_5
void FASTCALL call_9_1_15_5(SH2_struct* sh)
{
  SH2movwi(sh, 1, (15<<4)+5);
}
#define init_9_1_15_6
void FASTCALL call_9_1_15_6(SH2_struct* sh)
{
  SH2movwi(sh, 1, (15<<4)+6);
}
#define init_9_1_15_7
void FASTCALL call_9_1_15_7(SH2_struct* sh)
{
  SH2movwi(sh, 1, (15<<4)+7);
}
#define init_9_1_15_8
void FASTCALL call_9_1_15_8(SH2_struct* sh)
{
  SH2movwi(sh, 1, (15<<4)+8);
}
#define init_9_1_15_9
void FASTCALL call_9_1_15_9(SH2_struct* sh)
{
  SH2movwi(sh, 1, (15<<4)+9);
}
#define init_9_1_15_10
void FASTCALL call_9_1_15_10(SH2_struct* sh)
{
  SH2movwi(sh, 1, (15<<4)+10);
}
#define init_9_1_15_11
void FASTCALL call_9_1_15_11(SH2_struct* sh)
{
  SH2movwi(sh, 1, (15<<4)+11);
}
#define init_9_1_15_12
void FASTCALL call_9_1_15_12(SH2_struct* sh)
{
  SH2movwi(sh, 1, (15<<4)+12);
}
#define init_9_1_15_13
void FASTCALL call_9_1_15_13(SH2_struct* sh)
{
  SH2movwi(sh, 1, (15<<4)+13);
}
#define init_9_1_15_14
void FASTCALL call_9_1_15_14(SH2_struct* sh)
{
  SH2movwi(sh, 1, (15<<4)+14);
}
#define init_9_1_15_15
void FASTCALL call_9_1_15_15(SH2_struct* sh)
{
  SH2movwi(sh, 1, (15<<4)+15);
}
#define init_9_2_0_0
void FASTCALL call_9_2_0_0(SH2_struct* sh)
{
  SH2movwi(sh, 2, (0<<4)+0);
}
#define init_9_2_0_1
void FASTCALL call_9_2_0_1(SH2_struct* sh)
{
  SH2movwi(sh, 2, (0<<4)+1);
}
#define init_9_2_0_2
void FASTCALL call_9_2_0_2(SH2_struct* sh)
{
  SH2movwi(sh, 2, (0<<4)+2);
}
#define init_9_2_0_3
void FASTCALL call_9_2_0_3(SH2_struct* sh)
{
  SH2movwi(sh, 2, (0<<4)+3);
}
#define init_9_2_0_4
void FASTCALL call_9_2_0_4(SH2_struct* sh)
{
  SH2movwi(sh, 2, (0<<4)+4);
}
#define init_9_2_0_5
void FASTCALL call_9_2_0_5(SH2_struct* sh)
{
  SH2movwi(sh, 2, (0<<4)+5);
}
#define init_9_2_0_6
void FASTCALL call_9_2_0_6(SH2_struct* sh)
{
  SH2movwi(sh, 2, (0<<4)+6);
}
#define init_9_2_0_7
void FASTCALL call_9_2_0_7(SH2_struct* sh)
{
  SH2movwi(sh, 2, (0<<4)+7);
}
#define init_9_2_0_8
void FASTCALL call_9_2_0_8(SH2_struct* sh)
{
  SH2movwi(sh, 2, (0<<4)+8);
}
#define init_9_2_0_9
void FASTCALL call_9_2_0_9(SH2_struct* sh)
{
  SH2movwi(sh, 2, (0<<4)+9);
}
#define init_9_2_0_10
void FASTCALL call_9_2_0_10(SH2_struct* sh)
{
  SH2movwi(sh, 2, (0<<4)+10);
}
#define init_9_2_0_11
void FASTCALL call_9_2_0_11(SH2_struct* sh)
{
  SH2movwi(sh, 2, (0<<4)+11);
}
#define init_9_2_0_12
void FASTCALL call_9_2_0_12(SH2_struct* sh)
{
  SH2movwi(sh, 2, (0<<4)+12);
}
#define init_9_2_0_13
void FASTCALL call_9_2_0_13(SH2_struct* sh)
{
  SH2movwi(sh, 2, (0<<4)+13);
}
#define init_9_2_0_14
void FASTCALL call_9_2_0_14(SH2_struct* sh)
{
  SH2movwi(sh, 2, (0<<4)+14);
}
#define init_9_2_0_15
void FASTCALL call_9_2_0_15(SH2_struct* sh)
{
  SH2movwi(sh, 2, (0<<4)+15);
}
#define init_9_2_1_0
void FASTCALL call_9_2_1_0(SH2_struct* sh)
{
  SH2movwi(sh, 2, (1<<4)+0);
}
#define init_9_2_1_1
void FASTCALL call_9_2_1_1(SH2_struct* sh)
{
  SH2movwi(sh, 2, (1<<4)+1);
}
#define init_9_2_1_2
void FASTCALL call_9_2_1_2(SH2_struct* sh)
{
  SH2movwi(sh, 2, (1<<4)+2);
}
#define init_9_2_1_3
void FASTCALL call_9_2_1_3(SH2_struct* sh)
{
  SH2movwi(sh, 2, (1<<4)+3);
}
#define init_9_2_1_4
void FASTCALL call_9_2_1_4(SH2_struct* sh)
{
  SH2movwi(sh, 2, (1<<4)+4);
}
#define init_9_2_1_5
void FASTCALL call_9_2_1_5(SH2_struct* sh)
{
  SH2movwi(sh, 2, (1<<4)+5);
}
#define init_9_2_1_6
void FASTCALL call_9_2_1_6(SH2_struct* sh)
{
  SH2movwi(sh, 2, (1<<4)+6);
}
#define init_9_2_1_7
void FASTCALL call_9_2_1_7(SH2_struct* sh)
{
  SH2movwi(sh, 2, (1<<4)+7);
}
#define init_9_2_1_8
void FASTCALL call_9_2_1_8(SH2_struct* sh)
{
  SH2movwi(sh, 2, (1<<4)+8);
}
#define init_9_2_1_9
void FASTCALL call_9_2_1_9(SH2_struct* sh)
{
  SH2movwi(sh, 2, (1<<4)+9);
}
#define init_9_2_1_10
void FASTCALL call_9_2_1_10(SH2_struct* sh)
{
  SH2movwi(sh, 2, (1<<4)+10);
}
#define init_9_2_1_11
void FASTCALL call_9_2_1_11(SH2_struct* sh)
{
  SH2movwi(sh, 2, (1<<4)+11);
}
#define init_9_2_1_12
void FASTCALL call_9_2_1_12(SH2_struct* sh)
{
  SH2movwi(sh, 2, (1<<4)+12);
}
#define init_9_2_1_13
void FASTCALL call_9_2_1_13(SH2_struct* sh)
{
  SH2movwi(sh, 2, (1<<4)+13);
}
#define init_9_2_1_14
void FASTCALL call_9_2_1_14(SH2_struct* sh)
{
  SH2movwi(sh, 2, (1<<4)+14);
}
#define init_9_2_1_15
void FASTCALL call_9_2_1_15(SH2_struct* sh)
{
  SH2movwi(sh, 2, (1<<4)+15);
}
#define init_9_2_2_0
void FASTCALL call_9_2_2_0(SH2_struct* sh)
{
  SH2movwi(sh, 2, (2<<4)+0);
}
#define init_9_2_2_1
void FASTCALL call_9_2_2_1(SH2_struct* sh)
{
  SH2movwi(sh, 2, (2<<4)+1);
}
#define init_9_2_2_2
void FASTCALL call_9_2_2_2(SH2_struct* sh)
{
  SH2movwi(sh, 2, (2<<4)+2);
}
#define init_9_2_2_3
void FASTCALL call_9_2_2_3(SH2_struct* sh)
{
  SH2movwi(sh, 2, (2<<4)+3);
}
#define init_9_2_2_4
void FASTCALL call_9_2_2_4(SH2_struct* sh)
{
  SH2movwi(sh, 2, (2<<4)+4);
}
#define init_9_2_2_5
void FASTCALL call_9_2_2_5(SH2_struct* sh)
{
  SH2movwi(sh, 2, (2<<4)+5);
}
#define init_9_2_2_6
void FASTCALL call_9_2_2_6(SH2_struct* sh)
{
  SH2movwi(sh, 2, (2<<4)+6);
}
#define init_9_2_2_7
void FASTCALL call_9_2_2_7(SH2_struct* sh)
{
  SH2movwi(sh, 2, (2<<4)+7);
}
#define init_9_2_2_8
void FASTCALL call_9_2_2_8(SH2_struct* sh)
{
  SH2movwi(sh, 2, (2<<4)+8);
}
#define init_9_2_2_9
void FASTCALL call_9_2_2_9(SH2_struct* sh)
{
  SH2movwi(sh, 2, (2<<4)+9);
}
#define init_9_2_2_10
void FASTCALL call_9_2_2_10(SH2_struct* sh)
{
  SH2movwi(sh, 2, (2<<4)+10);
}
#define init_9_2_2_11
void FASTCALL call_9_2_2_11(SH2_struct* sh)
{
  SH2movwi(sh, 2, (2<<4)+11);
}
#define init_9_2_2_12
void FASTCALL call_9_2_2_12(SH2_struct* sh)
{
  SH2movwi(sh, 2, (2<<4)+12);
}
#define init_9_2_2_13
void FASTCALL call_9_2_2_13(SH2_struct* sh)
{
  SH2movwi(sh, 2, (2<<4)+13);
}
#define init_9_2_2_14
void FASTCALL call_9_2_2_14(SH2_struct* sh)
{
  SH2movwi(sh, 2, (2<<4)+14);
}
#define init_9_2_2_15
void FASTCALL call_9_2_2_15(SH2_struct* sh)
{
  SH2movwi(sh, 2, (2<<4)+15);
}
#define init_9_2_3_0
void FASTCALL call_9_2_3_0(SH2_struct* sh)
{
  SH2movwi(sh, 2, (3<<4)+0);
}
#define init_9_2_3_1
void FASTCALL call_9_2_3_1(SH2_struct* sh)
{
  SH2movwi(sh, 2, (3<<4)+1);
}
#define init_9_2_3_2
void FASTCALL call_9_2_3_2(SH2_struct* sh)
{
  SH2movwi(sh, 2, (3<<4)+2);
}
#define init_9_2_3_3
void FASTCALL call_9_2_3_3(SH2_struct* sh)
{
  SH2movwi(sh, 2, (3<<4)+3);
}
#define init_9_2_3_4
void FASTCALL call_9_2_3_4(SH2_struct* sh)
{
  SH2movwi(sh, 2, (3<<4)+4);
}
#define init_9_2_3_5
void FASTCALL call_9_2_3_5(SH2_struct* sh)
{
  SH2movwi(sh, 2, (3<<4)+5);
}
#define init_9_2_3_6
void FASTCALL call_9_2_3_6(SH2_struct* sh)
{
  SH2movwi(sh, 2, (3<<4)+6);
}
#define init_9_2_3_7
void FASTCALL call_9_2_3_7(SH2_struct* sh)
{
  SH2movwi(sh, 2, (3<<4)+7);
}
#define init_9_2_3_8
void FASTCALL call_9_2_3_8(SH2_struct* sh)
{
  SH2movwi(sh, 2, (3<<4)+8);
}
#define init_9_2_3_9
void FASTCALL call_9_2_3_9(SH2_struct* sh)
{
  SH2movwi(sh, 2, (3<<4)+9);
}
#define init_9_2_3_10
void FASTCALL call_9_2_3_10(SH2_struct* sh)
{
  SH2movwi(sh, 2, (3<<4)+10);
}
#define init_9_2_3_11
void FASTCALL call_9_2_3_11(SH2_struct* sh)
{
  SH2movwi(sh, 2, (3<<4)+11);
}
#define init_9_2_3_12
void FASTCALL call_9_2_3_12(SH2_struct* sh)
{
  SH2movwi(sh, 2, (3<<4)+12);
}
#define init_9_2_3_13
void FASTCALL call_9_2_3_13(SH2_struct* sh)
{
  SH2movwi(sh, 2, (3<<4)+13);
}
#define init_9_2_3_14
void FASTCALL call_9_2_3_14(SH2_struct* sh)
{
  SH2movwi(sh, 2, (3<<4)+14);
}
#define init_9_2_3_15
void FASTCALL call_9_2_3_15(SH2_struct* sh)
{
  SH2movwi(sh, 2, (3<<4)+15);
}
#define init_9_2_4_0
void FASTCALL call_9_2_4_0(SH2_struct* sh)
{
  SH2movwi(sh, 2, (4<<4)+0);
}
#define init_9_2_4_1
void FASTCALL call_9_2_4_1(SH2_struct* sh)
{
  SH2movwi(sh, 2, (4<<4)+1);
}
#define init_9_2_4_2
void FASTCALL call_9_2_4_2(SH2_struct* sh)
{
  SH2movwi(sh, 2, (4<<4)+2);
}
#define init_9_2_4_3
void FASTCALL call_9_2_4_3(SH2_struct* sh)
{
  SH2movwi(sh, 2, (4<<4)+3);
}
#define init_9_2_4_4
void FASTCALL call_9_2_4_4(SH2_struct* sh)
{
  SH2movwi(sh, 2, (4<<4)+4);
}
#define init_9_2_4_5
void FASTCALL call_9_2_4_5(SH2_struct* sh)
{
  SH2movwi(sh, 2, (4<<4)+5);
}
#define init_9_2_4_6
void FASTCALL call_9_2_4_6(SH2_struct* sh)
{
  SH2movwi(sh, 2, (4<<4)+6);
}
#define init_9_2_4_7
void FASTCALL call_9_2_4_7(SH2_struct* sh)
{
  SH2movwi(sh, 2, (4<<4)+7);
}
#define init_9_2_4_8
void FASTCALL call_9_2_4_8(SH2_struct* sh)
{
  SH2movwi(sh, 2, (4<<4)+8);
}
#define init_9_2_4_9
void FASTCALL call_9_2_4_9(SH2_struct* sh)
{
  SH2movwi(sh, 2, (4<<4)+9);
}
#define init_9_2_4_10
void FASTCALL call_9_2_4_10(SH2_struct* sh)
{
  SH2movwi(sh, 2, (4<<4)+10);
}
#define init_9_2_4_11
void FASTCALL call_9_2_4_11(SH2_struct* sh)
{
  SH2movwi(sh, 2, (4<<4)+11);
}
#define init_9_2_4_12
void FASTCALL call_9_2_4_12(SH2_struct* sh)
{
  SH2movwi(sh, 2, (4<<4)+12);
}
#define init_9_2_4_13
void FASTCALL call_9_2_4_13(SH2_struct* sh)
{
  SH2movwi(sh, 2, (4<<4)+13);
}
#define init_9_2_4_14
void FASTCALL call_9_2_4_14(SH2_struct* sh)
{
  SH2movwi(sh, 2, (4<<4)+14);
}
#define init_9_2_4_15
void FASTCALL call_9_2_4_15(SH2_struct* sh)
{
  SH2movwi(sh, 2, (4<<4)+15);
}
#define init_9_2_5_0
void FASTCALL call_9_2_5_0(SH2_struct* sh)
{
  SH2movwi(sh, 2, (5<<4)+0);
}
#define init_9_2_5_1
void FASTCALL call_9_2_5_1(SH2_struct* sh)
{
  SH2movwi(sh, 2, (5<<4)+1);
}
#define init_9_2_5_2
void FASTCALL call_9_2_5_2(SH2_struct* sh)
{
  SH2movwi(sh, 2, (5<<4)+2);
}
#define init_9_2_5_3
void FASTCALL call_9_2_5_3(SH2_struct* sh)
{
  SH2movwi(sh, 2, (5<<4)+3);
}
#define init_9_2_5_4
void FASTCALL call_9_2_5_4(SH2_struct* sh)
{
  SH2movwi(sh, 2, (5<<4)+4);
}
#define init_9_2_5_5
void FASTCALL call_9_2_5_5(SH2_struct* sh)
{
  SH2movwi(sh, 2, (5<<4)+5);
}
#define init_9_2_5_6
void FASTCALL call_9_2_5_6(SH2_struct* sh)
{
  SH2movwi(sh, 2, (5<<4)+6);
}
#define init_9_2_5_7
void FASTCALL call_9_2_5_7(SH2_struct* sh)
{
  SH2movwi(sh, 2, (5<<4)+7);
}
#define init_9_2_5_8
void FASTCALL call_9_2_5_8(SH2_struct* sh)
{
  SH2movwi(sh, 2, (5<<4)+8);
}
#define init_9_2_5_9
void FASTCALL call_9_2_5_9(SH2_struct* sh)
{
  SH2movwi(sh, 2, (5<<4)+9);
}
#define init_9_2_5_10
void FASTCALL call_9_2_5_10(SH2_struct* sh)
{
  SH2movwi(sh, 2, (5<<4)+10);
}
#define init_9_2_5_11
void FASTCALL call_9_2_5_11(SH2_struct* sh)
{
  SH2movwi(sh, 2, (5<<4)+11);
}
#define init_9_2_5_12
void FASTCALL call_9_2_5_12(SH2_struct* sh)
{
  SH2movwi(sh, 2, (5<<4)+12);
}
#define init_9_2_5_13
void FASTCALL call_9_2_5_13(SH2_struct* sh)
{
  SH2movwi(sh, 2, (5<<4)+13);
}
#define init_9_2_5_14
void FASTCALL call_9_2_5_14(SH2_struct* sh)
{
  SH2movwi(sh, 2, (5<<4)+14);
}
#define init_9_2_5_15
void FASTCALL call_9_2_5_15(SH2_struct* sh)
{
  SH2movwi(sh, 2, (5<<4)+15);
}
#define init_9_2_6_0
void FASTCALL call_9_2_6_0(SH2_struct* sh)
{
  SH2movwi(sh, 2, (6<<4)+0);
}
#define init_9_2_6_1
void FASTCALL call_9_2_6_1(SH2_struct* sh)
{
  SH2movwi(sh, 2, (6<<4)+1);
}
#define init_9_2_6_2
void FASTCALL call_9_2_6_2(SH2_struct* sh)
{
  SH2movwi(sh, 2, (6<<4)+2);
}
#define init_9_2_6_3
void FASTCALL call_9_2_6_3(SH2_struct* sh)
{
  SH2movwi(sh, 2, (6<<4)+3);
}
#define init_9_2_6_4
void FASTCALL call_9_2_6_4(SH2_struct* sh)
{
  SH2movwi(sh, 2, (6<<4)+4);
}
#define init_9_2_6_5
void FASTCALL call_9_2_6_5(SH2_struct* sh)
{
  SH2movwi(sh, 2, (6<<4)+5);
}
#define init_9_2_6_6
void FASTCALL call_9_2_6_6(SH2_struct* sh)
{
  SH2movwi(sh, 2, (6<<4)+6);
}
#define init_9_2_6_7
void FASTCALL call_9_2_6_7(SH2_struct* sh)
{
  SH2movwi(sh, 2, (6<<4)+7);
}
#define init_9_2_6_8
void FASTCALL call_9_2_6_8(SH2_struct* sh)
{
  SH2movwi(sh, 2, (6<<4)+8);
}
#define init_9_2_6_9
void FASTCALL call_9_2_6_9(SH2_struct* sh)
{
  SH2movwi(sh, 2, (6<<4)+9);
}
#define init_9_2_6_10
void FASTCALL call_9_2_6_10(SH2_struct* sh)
{
  SH2movwi(sh, 2, (6<<4)+10);
}
#define init_9_2_6_11
void FASTCALL call_9_2_6_11(SH2_struct* sh)
{
  SH2movwi(sh, 2, (6<<4)+11);
}
#define init_9_2_6_12
void FASTCALL call_9_2_6_12(SH2_struct* sh)
{
  SH2movwi(sh, 2, (6<<4)+12);
}
#define init_9_2_6_13
void FASTCALL call_9_2_6_13(SH2_struct* sh)
{
  SH2movwi(sh, 2, (6<<4)+13);
}
#define init_9_2_6_14
void FASTCALL call_9_2_6_14(SH2_struct* sh)
{
  SH2movwi(sh, 2, (6<<4)+14);
}
#define init_9_2_6_15
void FASTCALL call_9_2_6_15(SH2_struct* sh)
{
  SH2movwi(sh, 2, (6<<4)+15);
}
#define init_9_2_7_0
void FASTCALL call_9_2_7_0(SH2_struct* sh)
{
  SH2movwi(sh, 2, (7<<4)+0);
}
#define init_9_2_7_1
void FASTCALL call_9_2_7_1(SH2_struct* sh)
{
  SH2movwi(sh, 2, (7<<4)+1);
}
#define init_9_2_7_2
void FASTCALL call_9_2_7_2(SH2_struct* sh)
{
  SH2movwi(sh, 2, (7<<4)+2);
}
#define init_9_2_7_3
void FASTCALL call_9_2_7_3(SH2_struct* sh)
{
  SH2movwi(sh, 2, (7<<4)+3);
}
#define init_9_2_7_4
void FASTCALL call_9_2_7_4(SH2_struct* sh)
{
  SH2movwi(sh, 2, (7<<4)+4);
}
#define init_9_2_7_5
void FASTCALL call_9_2_7_5(SH2_struct* sh)
{
  SH2movwi(sh, 2, (7<<4)+5);
}
#define init_9_2_7_6
void FASTCALL call_9_2_7_6(SH2_struct* sh)
{
  SH2movwi(sh, 2, (7<<4)+6);
}
#define init_9_2_7_7
void FASTCALL call_9_2_7_7(SH2_struct* sh)
{
  SH2movwi(sh, 2, (7<<4)+7);
}
#define init_9_2_7_8
void FASTCALL call_9_2_7_8(SH2_struct* sh)
{
  SH2movwi(sh, 2, (7<<4)+8);
}
#define init_9_2_7_9
void FASTCALL call_9_2_7_9(SH2_struct* sh)
{
  SH2movwi(sh, 2, (7<<4)+9);
}
#define init_9_2_7_10
void FASTCALL call_9_2_7_10(SH2_struct* sh)
{
  SH2movwi(sh, 2, (7<<4)+10);
}
#define init_9_2_7_11
void FASTCALL call_9_2_7_11(SH2_struct* sh)
{
  SH2movwi(sh, 2, (7<<4)+11);
}
#define init_9_2_7_12
void FASTCALL call_9_2_7_12(SH2_struct* sh)
{
  SH2movwi(sh, 2, (7<<4)+12);
}
#define init_9_2_7_13
void FASTCALL call_9_2_7_13(SH2_struct* sh)
{
  SH2movwi(sh, 2, (7<<4)+13);
}
#define init_9_2_7_14
void FASTCALL call_9_2_7_14(SH2_struct* sh)
{
  SH2movwi(sh, 2, (7<<4)+14);
}
#define init_9_2_7_15
void FASTCALL call_9_2_7_15(SH2_struct* sh)
{
  SH2movwi(sh, 2, (7<<4)+15);
}
#define init_9_2_8_0
void FASTCALL call_9_2_8_0(SH2_struct* sh)
{
  SH2movwi(sh, 2, (8<<4)+0);
}
#define init_9_2_8_1
void FASTCALL call_9_2_8_1(SH2_struct* sh)
{
  SH2movwi(sh, 2, (8<<4)+1);
}
#define init_9_2_8_2
void FASTCALL call_9_2_8_2(SH2_struct* sh)
{
  SH2movwi(sh, 2, (8<<4)+2);
}
#define init_9_2_8_3
void FASTCALL call_9_2_8_3(SH2_struct* sh)
{
  SH2movwi(sh, 2, (8<<4)+3);
}
#define init_9_2_8_4
void FASTCALL call_9_2_8_4(SH2_struct* sh)
{
  SH2movwi(sh, 2, (8<<4)+4);
}
#define init_9_2_8_5
void FASTCALL call_9_2_8_5(SH2_struct* sh)
{
  SH2movwi(sh, 2, (8<<4)+5);
}
#define init_9_2_8_6
void FASTCALL call_9_2_8_6(SH2_struct* sh)
{
  SH2movwi(sh, 2, (8<<4)+6);
}
#define init_9_2_8_7
void FASTCALL call_9_2_8_7(SH2_struct* sh)
{
  SH2movwi(sh, 2, (8<<4)+7);
}
#define init_9_2_8_8
void FASTCALL call_9_2_8_8(SH2_struct* sh)
{
  SH2movwi(sh, 2, (8<<4)+8);
}
#define init_9_2_8_9
void FASTCALL call_9_2_8_9(SH2_struct* sh)
{
  SH2movwi(sh, 2, (8<<4)+9);
}
#define init_9_2_8_10
void FASTCALL call_9_2_8_10(SH2_struct* sh)
{
  SH2movwi(sh, 2, (8<<4)+10);
}
#define init_9_2_8_11
void FASTCALL call_9_2_8_11(SH2_struct* sh)
{
  SH2movwi(sh, 2, (8<<4)+11);
}
#define init_9_2_8_12
void FASTCALL call_9_2_8_12(SH2_struct* sh)
{
  SH2movwi(sh, 2, (8<<4)+12);
}
#define init_9_2_8_13
void FASTCALL call_9_2_8_13(SH2_struct* sh)
{
  SH2movwi(sh, 2, (8<<4)+13);
}
#define init_9_2_8_14
void FASTCALL call_9_2_8_14(SH2_struct* sh)
{
  SH2movwi(sh, 2, (8<<4)+14);
}
#define init_9_2_8_15
void FASTCALL call_9_2_8_15(SH2_struct* sh)
{
  SH2movwi(sh, 2, (8<<4)+15);
}
#define init_9_2_9_0
void FASTCALL call_9_2_9_0(SH2_struct* sh)
{
  SH2movwi(sh, 2, (9<<4)+0);
}
#define init_9_2_9_1
void FASTCALL call_9_2_9_1(SH2_struct* sh)
{
  SH2movwi(sh, 2, (9<<4)+1);
}
#define init_9_2_9_2
void FASTCALL call_9_2_9_2(SH2_struct* sh)
{
  SH2movwi(sh, 2, (9<<4)+2);
}
#define init_9_2_9_3
void FASTCALL call_9_2_9_3(SH2_struct* sh)
{
  SH2movwi(sh, 2, (9<<4)+3);
}
#define init_9_2_9_4
void FASTCALL call_9_2_9_4(SH2_struct* sh)
{
  SH2movwi(sh, 2, (9<<4)+4);
}
#define init_9_2_9_5
void FASTCALL call_9_2_9_5(SH2_struct* sh)
{
  SH2movwi(sh, 2, (9<<4)+5);
}
#define init_9_2_9_6
void FASTCALL call_9_2_9_6(SH2_struct* sh)
{
  SH2movwi(sh, 2, (9<<4)+6);
}
#define init_9_2_9_7
void FASTCALL call_9_2_9_7(SH2_struct* sh)
{
  SH2movwi(sh, 2, (9<<4)+7);
}
#define init_9_2_9_8
void FASTCALL call_9_2_9_8(SH2_struct* sh)
{
  SH2movwi(sh, 2, (9<<4)+8);
}
#define init_9_2_9_9
void FASTCALL call_9_2_9_9(SH2_struct* sh)
{
  SH2movwi(sh, 2, (9<<4)+9);
}
#define init_9_2_9_10
void FASTCALL call_9_2_9_10(SH2_struct* sh)
{
  SH2movwi(sh, 2, (9<<4)+10);
}
#define init_9_2_9_11
void FASTCALL call_9_2_9_11(SH2_struct* sh)
{
  SH2movwi(sh, 2, (9<<4)+11);
}
#define init_9_2_9_12
void FASTCALL call_9_2_9_12(SH2_struct* sh)
{
  SH2movwi(sh, 2, (9<<4)+12);
}
#define init_9_2_9_13
void FASTCALL call_9_2_9_13(SH2_struct* sh)
{
  SH2movwi(sh, 2, (9<<4)+13);
}
#define init_9_2_9_14
void FASTCALL call_9_2_9_14(SH2_struct* sh)
{
  SH2movwi(sh, 2, (9<<4)+14);
}
#define init_9_2_9_15
void FASTCALL call_9_2_9_15(SH2_struct* sh)
{
  SH2movwi(sh, 2, (9<<4)+15);
}
#define init_9_2_10_0
void FASTCALL call_9_2_10_0(SH2_struct* sh)
{
  SH2movwi(sh, 2, (10<<4)+0);
}
#define init_9_2_10_1
void FASTCALL call_9_2_10_1(SH2_struct* sh)
{
  SH2movwi(sh, 2, (10<<4)+1);
}
#define init_9_2_10_2
void FASTCALL call_9_2_10_2(SH2_struct* sh)
{
  SH2movwi(sh, 2, (10<<4)+2);
}
#define init_9_2_10_3
void FASTCALL call_9_2_10_3(SH2_struct* sh)
{
  SH2movwi(sh, 2, (10<<4)+3);
}
#define init_9_2_10_4
void FASTCALL call_9_2_10_4(SH2_struct* sh)
{
  SH2movwi(sh, 2, (10<<4)+4);
}
#define init_9_2_10_5
void FASTCALL call_9_2_10_5(SH2_struct* sh)
{
  SH2movwi(sh, 2, (10<<4)+5);
}
#define init_9_2_10_6
void FASTCALL call_9_2_10_6(SH2_struct* sh)
{
  SH2movwi(sh, 2, (10<<4)+6);
}
#define init_9_2_10_7
void FASTCALL call_9_2_10_7(SH2_struct* sh)
{
  SH2movwi(sh, 2, (10<<4)+7);
}
#define init_9_2_10_8
void FASTCALL call_9_2_10_8(SH2_struct* sh)
{
  SH2movwi(sh, 2, (10<<4)+8);
}
#define init_9_2_10_9
void FASTCALL call_9_2_10_9(SH2_struct* sh)
{
  SH2movwi(sh, 2, (10<<4)+9);
}
#define init_9_2_10_10
void FASTCALL call_9_2_10_10(SH2_struct* sh)
{
  SH2movwi(sh, 2, (10<<4)+10);
}
#define init_9_2_10_11
void FASTCALL call_9_2_10_11(SH2_struct* sh)
{
  SH2movwi(sh, 2, (10<<4)+11);
}
#define init_9_2_10_12
void FASTCALL call_9_2_10_12(SH2_struct* sh)
{
  SH2movwi(sh, 2, (10<<4)+12);
}
#define init_9_2_10_13
void FASTCALL call_9_2_10_13(SH2_struct* sh)
{
  SH2movwi(sh, 2, (10<<4)+13);
}
#define init_9_2_10_14
void FASTCALL call_9_2_10_14(SH2_struct* sh)
{
  SH2movwi(sh, 2, (10<<4)+14);
}
#define init_9_2_10_15
void FASTCALL call_9_2_10_15(SH2_struct* sh)
{
  SH2movwi(sh, 2, (10<<4)+15);
}
#define init_9_2_11_0
void FASTCALL call_9_2_11_0(SH2_struct* sh)
{
  SH2movwi(sh, 2, (11<<4)+0);
}
#define init_9_2_11_1
void FASTCALL call_9_2_11_1(SH2_struct* sh)
{
  SH2movwi(sh, 2, (11<<4)+1);
}
#define init_9_2_11_2
void FASTCALL call_9_2_11_2(SH2_struct* sh)
{
  SH2movwi(sh, 2, (11<<4)+2);
}
#define init_9_2_11_3
void FASTCALL call_9_2_11_3(SH2_struct* sh)
{
  SH2movwi(sh, 2, (11<<4)+3);
}
#define init_9_2_11_4
void FASTCALL call_9_2_11_4(SH2_struct* sh)
{
  SH2movwi(sh, 2, (11<<4)+4);
}
#define init_9_2_11_5
void FASTCALL call_9_2_11_5(SH2_struct* sh)
{
  SH2movwi(sh, 2, (11<<4)+5);
}
#define init_9_2_11_6
void FASTCALL call_9_2_11_6(SH2_struct* sh)
{
  SH2movwi(sh, 2, (11<<4)+6);
}
#define init_9_2_11_7
void FASTCALL call_9_2_11_7(SH2_struct* sh)
{
  SH2movwi(sh, 2, (11<<4)+7);
}
#define init_9_2_11_8
void FASTCALL call_9_2_11_8(SH2_struct* sh)
{
  SH2movwi(sh, 2, (11<<4)+8);
}
#define init_9_2_11_9
void FASTCALL call_9_2_11_9(SH2_struct* sh)
{
  SH2movwi(sh, 2, (11<<4)+9);
}
#define init_9_2_11_10
void FASTCALL call_9_2_11_10(SH2_struct* sh)
{
  SH2movwi(sh, 2, (11<<4)+10);
}
#define init_9_2_11_11
void FASTCALL call_9_2_11_11(SH2_struct* sh)
{
  SH2movwi(sh, 2, (11<<4)+11);
}
#define init_9_2_11_12
void FASTCALL call_9_2_11_12(SH2_struct* sh)
{
  SH2movwi(sh, 2, (11<<4)+12);
}
#define init_9_2_11_13
void FASTCALL call_9_2_11_13(SH2_struct* sh)
{
  SH2movwi(sh, 2, (11<<4)+13);
}
#define init_9_2_11_14
void FASTCALL call_9_2_11_14(SH2_struct* sh)
{
  SH2movwi(sh, 2, (11<<4)+14);
}
#define init_9_2_11_15
void FASTCALL call_9_2_11_15(SH2_struct* sh)
{
  SH2movwi(sh, 2, (11<<4)+15);
}
#define init_9_2_12_0
void FASTCALL call_9_2_12_0(SH2_struct* sh)
{
  SH2movwi(sh, 2, (12<<4)+0);
}
#define init_9_2_12_1
void FASTCALL call_9_2_12_1(SH2_struct* sh)
{
  SH2movwi(sh, 2, (12<<4)+1);
}
#define init_9_2_12_2
void FASTCALL call_9_2_12_2(SH2_struct* sh)
{
  SH2movwi(sh, 2, (12<<4)+2);
}
#define init_9_2_12_3
void FASTCALL call_9_2_12_3(SH2_struct* sh)
{
  SH2movwi(sh, 2, (12<<4)+3);
}
#define init_9_2_12_4
void FASTCALL call_9_2_12_4(SH2_struct* sh)
{
  SH2movwi(sh, 2, (12<<4)+4);
}
#define init_9_2_12_5
void FASTCALL call_9_2_12_5(SH2_struct* sh)
{
  SH2movwi(sh, 2, (12<<4)+5);
}
#define init_9_2_12_6
void FASTCALL call_9_2_12_6(SH2_struct* sh)
{
  SH2movwi(sh, 2, (12<<4)+6);
}
#define init_9_2_12_7
void FASTCALL call_9_2_12_7(SH2_struct* sh)
{
  SH2movwi(sh, 2, (12<<4)+7);
}
#define init_9_2_12_8
void FASTCALL call_9_2_12_8(SH2_struct* sh)
{
  SH2movwi(sh, 2, (12<<4)+8);
}
#define init_9_2_12_9
void FASTCALL call_9_2_12_9(SH2_struct* sh)
{
  SH2movwi(sh, 2, (12<<4)+9);
}
#define init_9_2_12_10
void FASTCALL call_9_2_12_10(SH2_struct* sh)
{
  SH2movwi(sh, 2, (12<<4)+10);
}
#define init_9_2_12_11
void FASTCALL call_9_2_12_11(SH2_struct* sh)
{
  SH2movwi(sh, 2, (12<<4)+11);
}
#define init_9_2_12_12
void FASTCALL call_9_2_12_12(SH2_struct* sh)
{
  SH2movwi(sh, 2, (12<<4)+12);
}
#define init_9_2_12_13
void FASTCALL call_9_2_12_13(SH2_struct* sh)
{
  SH2movwi(sh, 2, (12<<4)+13);
}
#define init_9_2_12_14
void FASTCALL call_9_2_12_14(SH2_struct* sh)
{
  SH2movwi(sh, 2, (12<<4)+14);
}
#define init_9_2_12_15
void FASTCALL call_9_2_12_15(SH2_struct* sh)
{
  SH2movwi(sh, 2, (12<<4)+15);
}
#define init_9_2_13_0
void FASTCALL call_9_2_13_0(SH2_struct* sh)
{
  SH2movwi(sh, 2, (13<<4)+0);
}
#define init_9_2_13_1
void FASTCALL call_9_2_13_1(SH2_struct* sh)
{
  SH2movwi(sh, 2, (13<<4)+1);
}
#define init_9_2_13_2
void FASTCALL call_9_2_13_2(SH2_struct* sh)
{
  SH2movwi(sh, 2, (13<<4)+2);
}
#define init_9_2_13_3
void FASTCALL call_9_2_13_3(SH2_struct* sh)
{
  SH2movwi(sh, 2, (13<<4)+3);
}
#define init_9_2_13_4
void FASTCALL call_9_2_13_4(SH2_struct* sh)
{
  SH2movwi(sh, 2, (13<<4)+4);
}
#define init_9_2_13_5
void FASTCALL call_9_2_13_5(SH2_struct* sh)
{
  SH2movwi(sh, 2, (13<<4)+5);
}
#define init_9_2_13_6
void FASTCALL call_9_2_13_6(SH2_struct* sh)
{
  SH2movwi(sh, 2, (13<<4)+6);
}
#define init_9_2_13_7
void FASTCALL call_9_2_13_7(SH2_struct* sh)
{
  SH2movwi(sh, 2, (13<<4)+7);
}
#define init_9_2_13_8
void FASTCALL call_9_2_13_8(SH2_struct* sh)
{
  SH2movwi(sh, 2, (13<<4)+8);
}
#define init_9_2_13_9
void FASTCALL call_9_2_13_9(SH2_struct* sh)
{
  SH2movwi(sh, 2, (13<<4)+9);
}
#define init_9_2_13_10
void FASTCALL call_9_2_13_10(SH2_struct* sh)
{
  SH2movwi(sh, 2, (13<<4)+10);
}
#define init_9_2_13_11
void FASTCALL call_9_2_13_11(SH2_struct* sh)
{
  SH2movwi(sh, 2, (13<<4)+11);
}
#define init_9_2_13_12
void FASTCALL call_9_2_13_12(SH2_struct* sh)
{
  SH2movwi(sh, 2, (13<<4)+12);
}
#define init_9_2_13_13
void FASTCALL call_9_2_13_13(SH2_struct* sh)
{
  SH2movwi(sh, 2, (13<<4)+13);
}
#define init_9_2_13_14
void FASTCALL call_9_2_13_14(SH2_struct* sh)
{
  SH2movwi(sh, 2, (13<<4)+14);
}
#define init_9_2_13_15
void FASTCALL call_9_2_13_15(SH2_struct* sh)
{
  SH2movwi(sh, 2, (13<<4)+15);
}
#define init_9_2_14_0
void FASTCALL call_9_2_14_0(SH2_struct* sh)
{
  SH2movwi(sh, 2, (14<<4)+0);
}
#define init_9_2_14_1
void FASTCALL call_9_2_14_1(SH2_struct* sh)
{
  SH2movwi(sh, 2, (14<<4)+1);
}
#define init_9_2_14_2
void FASTCALL call_9_2_14_2(SH2_struct* sh)
{
  SH2movwi(sh, 2, (14<<4)+2);
}
#define init_9_2_14_3
void FASTCALL call_9_2_14_3(SH2_struct* sh)
{
  SH2movwi(sh, 2, (14<<4)+3);
}
#define init_9_2_14_4
void FASTCALL call_9_2_14_4(SH2_struct* sh)
{
  SH2movwi(sh, 2, (14<<4)+4);
}
#define init_9_2_14_5
void FASTCALL call_9_2_14_5(SH2_struct* sh)
{
  SH2movwi(sh, 2, (14<<4)+5);
}
#define init_9_2_14_6
void FASTCALL call_9_2_14_6(SH2_struct* sh)
{
  SH2movwi(sh, 2, (14<<4)+6);
}
#define init_9_2_14_7
void FASTCALL call_9_2_14_7(SH2_struct* sh)
{
  SH2movwi(sh, 2, (14<<4)+7);
}
#define init_9_2_14_8
void FASTCALL call_9_2_14_8(SH2_struct* sh)
{
  SH2movwi(sh, 2, (14<<4)+8);
}
#define init_9_2_14_9
void FASTCALL call_9_2_14_9(SH2_struct* sh)
{
  SH2movwi(sh, 2, (14<<4)+9);
}
#define init_9_2_14_10
void FASTCALL call_9_2_14_10(SH2_struct* sh)
{
  SH2movwi(sh, 2, (14<<4)+10);
}
#define init_9_2_14_11
void FASTCALL call_9_2_14_11(SH2_struct* sh)
{
  SH2movwi(sh, 2, (14<<4)+11);
}
#define init_9_2_14_12
void FASTCALL call_9_2_14_12(SH2_struct* sh)
{
  SH2movwi(sh, 2, (14<<4)+12);
}
#define init_9_2_14_13
void FASTCALL call_9_2_14_13(SH2_struct* sh)
{
  SH2movwi(sh, 2, (14<<4)+13);
}
#define init_9_2_14_14
void FASTCALL call_9_2_14_14(SH2_struct* sh)
{
  SH2movwi(sh, 2, (14<<4)+14);
}
#define init_9_2_14_15
void FASTCALL call_9_2_14_15(SH2_struct* sh)
{
  SH2movwi(sh, 2, (14<<4)+15);
}
#define init_9_2_15_0
void FASTCALL call_9_2_15_0(SH2_struct* sh)
{
  SH2movwi(sh, 2, (15<<4)+0);
}
#define init_9_2_15_1
void FASTCALL call_9_2_15_1(SH2_struct* sh)
{
  SH2movwi(sh, 2, (15<<4)+1);
}
#define init_9_2_15_2
void FASTCALL call_9_2_15_2(SH2_struct* sh)
{
  SH2movwi(sh, 2, (15<<4)+2);
}
#define init_9_2_15_3
void FASTCALL call_9_2_15_3(SH2_struct* sh)
{
  SH2movwi(sh, 2, (15<<4)+3);
}
#define init_9_2_15_4
void FASTCALL call_9_2_15_4(SH2_struct* sh)
{
  SH2movwi(sh, 2, (15<<4)+4);
}
#define init_9_2_15_5
void FASTCALL call_9_2_15_5(SH2_struct* sh)
{
  SH2movwi(sh, 2, (15<<4)+5);
}
#define init_9_2_15_6
void FASTCALL call_9_2_15_6(SH2_struct* sh)
{
  SH2movwi(sh, 2, (15<<4)+6);
}
#define init_9_2_15_7
void FASTCALL call_9_2_15_7(SH2_struct* sh)
{
  SH2movwi(sh, 2, (15<<4)+7);
}
#define init_9_2_15_8
void FASTCALL call_9_2_15_8(SH2_struct* sh)
{
  SH2movwi(sh, 2, (15<<4)+8);
}
#define init_9_2_15_9
void FASTCALL call_9_2_15_9(SH2_struct* sh)
{
  SH2movwi(sh, 2, (15<<4)+9);
}
#define init_9_2_15_10
void FASTCALL call_9_2_15_10(SH2_struct* sh)
{
  SH2movwi(sh, 2, (15<<4)+10);
}
#define init_9_2_15_11
void FASTCALL call_9_2_15_11(SH2_struct* sh)
{
  SH2movwi(sh, 2, (15<<4)+11);
}
#define init_9_2_15_12
void FASTCALL call_9_2_15_12(SH2_struct* sh)
{
  SH2movwi(sh, 2, (15<<4)+12);
}
#define init_9_2_15_13
void FASTCALL call_9_2_15_13(SH2_struct* sh)
{
  SH2movwi(sh, 2, (15<<4)+13);
}
#define init_9_2_15_14
void FASTCALL call_9_2_15_14(SH2_struct* sh)
{
  SH2movwi(sh, 2, (15<<4)+14);
}
#define init_9_2_15_15
void FASTCALL call_9_2_15_15(SH2_struct* sh)
{
  SH2movwi(sh, 2, (15<<4)+15);
}
#define init_9_3_0_0
void FASTCALL call_9_3_0_0(SH2_struct* sh)
{
  SH2movwi(sh, 3, (0<<4)+0);
}
#define init_9_3_0_1
void FASTCALL call_9_3_0_1(SH2_struct* sh)
{
  SH2movwi(sh, 3, (0<<4)+1);
}
#define init_9_3_0_2
void FASTCALL call_9_3_0_2(SH2_struct* sh)
{
  SH2movwi(sh, 3, (0<<4)+2);
}
#define init_9_3_0_3
void FASTCALL call_9_3_0_3(SH2_struct* sh)
{
  SH2movwi(sh, 3, (0<<4)+3);
}
#define init_9_3_0_4
void FASTCALL call_9_3_0_4(SH2_struct* sh)
{
  SH2movwi(sh, 3, (0<<4)+4);
}
#define init_9_3_0_5
void FASTCALL call_9_3_0_5(SH2_struct* sh)
{
  SH2movwi(sh, 3, (0<<4)+5);
}
#define init_9_3_0_6
void FASTCALL call_9_3_0_6(SH2_struct* sh)
{
  SH2movwi(sh, 3, (0<<4)+6);
}
#define init_9_3_0_7
void FASTCALL call_9_3_0_7(SH2_struct* sh)
{
  SH2movwi(sh, 3, (0<<4)+7);
}
#define init_9_3_0_8
void FASTCALL call_9_3_0_8(SH2_struct* sh)
{
  SH2movwi(sh, 3, (0<<4)+8);
}
#define init_9_3_0_9
void FASTCALL call_9_3_0_9(SH2_struct* sh)
{
  SH2movwi(sh, 3, (0<<4)+9);
}
#define init_9_3_0_10
void FASTCALL call_9_3_0_10(SH2_struct* sh)
{
  SH2movwi(sh, 3, (0<<4)+10);
}
#define init_9_3_0_11
void FASTCALL call_9_3_0_11(SH2_struct* sh)
{
  SH2movwi(sh, 3, (0<<4)+11);
}
#define init_9_3_0_12
void FASTCALL call_9_3_0_12(SH2_struct* sh)
{
  SH2movwi(sh, 3, (0<<4)+12);
}
#define init_9_3_0_13
void FASTCALL call_9_3_0_13(SH2_struct* sh)
{
  SH2movwi(sh, 3, (0<<4)+13);
}
#define init_9_3_0_14
void FASTCALL call_9_3_0_14(SH2_struct* sh)
{
  SH2movwi(sh, 3, (0<<4)+14);
}
#define init_9_3_0_15
void FASTCALL call_9_3_0_15(SH2_struct* sh)
{
  SH2movwi(sh, 3, (0<<4)+15);
}
#define init_9_3_1_0
void FASTCALL call_9_3_1_0(SH2_struct* sh)
{
  SH2movwi(sh, 3, (1<<4)+0);
}
#define init_9_3_1_1
void FASTCALL call_9_3_1_1(SH2_struct* sh)
{
  SH2movwi(sh, 3, (1<<4)+1);
}
#define init_9_3_1_2
void FASTCALL call_9_3_1_2(SH2_struct* sh)
{
  SH2movwi(sh, 3, (1<<4)+2);
}
#define init_9_3_1_3
void FASTCALL call_9_3_1_3(SH2_struct* sh)
{
  SH2movwi(sh, 3, (1<<4)+3);
}
#define init_9_3_1_4
void FASTCALL call_9_3_1_4(SH2_struct* sh)
{
  SH2movwi(sh, 3, (1<<4)+4);
}
#define init_9_3_1_5
void FASTCALL call_9_3_1_5(SH2_struct* sh)
{
  SH2movwi(sh, 3, (1<<4)+5);
}
#define init_9_3_1_6
void FASTCALL call_9_3_1_6(SH2_struct* sh)
{
  SH2movwi(sh, 3, (1<<4)+6);
}
#define init_9_3_1_7
void FASTCALL call_9_3_1_7(SH2_struct* sh)
{
  SH2movwi(sh, 3, (1<<4)+7);
}
#define init_9_3_1_8
void FASTCALL call_9_3_1_8(SH2_struct* sh)
{
  SH2movwi(sh, 3, (1<<4)+8);
}
#define init_9_3_1_9
void FASTCALL call_9_3_1_9(SH2_struct* sh)
{
  SH2movwi(sh, 3, (1<<4)+9);
}
#define init_9_3_1_10
void FASTCALL call_9_3_1_10(SH2_struct* sh)
{
  SH2movwi(sh, 3, (1<<4)+10);
}
#define init_9_3_1_11
void FASTCALL call_9_3_1_11(SH2_struct* sh)
{
  SH2movwi(sh, 3, (1<<4)+11);
}
#define init_9_3_1_12
void FASTCALL call_9_3_1_12(SH2_struct* sh)
{
  SH2movwi(sh, 3, (1<<4)+12);
}
#define init_9_3_1_13
void FASTCALL call_9_3_1_13(SH2_struct* sh)
{
  SH2movwi(sh, 3, (1<<4)+13);
}
#define init_9_3_1_14
void FASTCALL call_9_3_1_14(SH2_struct* sh)
{
  SH2movwi(sh, 3, (1<<4)+14);
}
#define init_9_3_1_15
void FASTCALL call_9_3_1_15(SH2_struct* sh)
{
  SH2movwi(sh, 3, (1<<4)+15);
}
#define init_9_3_2_0
void FASTCALL call_9_3_2_0(SH2_struct* sh)
{
  SH2movwi(sh, 3, (2<<4)+0);
}
#define init_9_3_2_1
void FASTCALL call_9_3_2_1(SH2_struct* sh)
{
  SH2movwi(sh, 3, (2<<4)+1);
}
#define init_9_3_2_2
void FASTCALL call_9_3_2_2(SH2_struct* sh)
{
  SH2movwi(sh, 3, (2<<4)+2);
}
#define init_9_3_2_3
void FASTCALL call_9_3_2_3(SH2_struct* sh)
{
  SH2movwi(sh, 3, (2<<4)+3);
}
#define init_9_3_2_4
void FASTCALL call_9_3_2_4(SH2_struct* sh)
{
  SH2movwi(sh, 3, (2<<4)+4);
}
#define init_9_3_2_5
void FASTCALL call_9_3_2_5(SH2_struct* sh)
{
  SH2movwi(sh, 3, (2<<4)+5);
}
#define init_9_3_2_6
void FASTCALL call_9_3_2_6(SH2_struct* sh)
{
  SH2movwi(sh, 3, (2<<4)+6);
}
#define init_9_3_2_7
void FASTCALL call_9_3_2_7(SH2_struct* sh)
{
  SH2movwi(sh, 3, (2<<4)+7);
}
#define init_9_3_2_8
void FASTCALL call_9_3_2_8(SH2_struct* sh)
{
  SH2movwi(sh, 3, (2<<4)+8);
}
#define init_9_3_2_9
void FASTCALL call_9_3_2_9(SH2_struct* sh)
{
  SH2movwi(sh, 3, (2<<4)+9);
}
#define init_9_3_2_10
void FASTCALL call_9_3_2_10(SH2_struct* sh)
{
  SH2movwi(sh, 3, (2<<4)+10);
}
#define init_9_3_2_11
void FASTCALL call_9_3_2_11(SH2_struct* sh)
{
  SH2movwi(sh, 3, (2<<4)+11);
}
#define init_9_3_2_12
void FASTCALL call_9_3_2_12(SH2_struct* sh)
{
  SH2movwi(sh, 3, (2<<4)+12);
}
#define init_9_3_2_13
void FASTCALL call_9_3_2_13(SH2_struct* sh)
{
  SH2movwi(sh, 3, (2<<4)+13);
}
#define init_9_3_2_14
void FASTCALL call_9_3_2_14(SH2_struct* sh)
{
  SH2movwi(sh, 3, (2<<4)+14);
}
#define init_9_3_2_15
void FASTCALL call_9_3_2_15(SH2_struct* sh)
{
  SH2movwi(sh, 3, (2<<4)+15);
}
#define init_9_3_3_0
void FASTCALL call_9_3_3_0(SH2_struct* sh)
{
  SH2movwi(sh, 3, (3<<4)+0);
}
#define init_9_3_3_1
void FASTCALL call_9_3_3_1(SH2_struct* sh)
{
  SH2movwi(sh, 3, (3<<4)+1);
}
#define init_9_3_3_2
void FASTCALL call_9_3_3_2(SH2_struct* sh)
{
  SH2movwi(sh, 3, (3<<4)+2);
}
#define init_9_3_3_3
void FASTCALL call_9_3_3_3(SH2_struct* sh)
{
  SH2movwi(sh, 3, (3<<4)+3);
}
#define init_9_3_3_4
void FASTCALL call_9_3_3_4(SH2_struct* sh)
{
  SH2movwi(sh, 3, (3<<4)+4);
}
#define init_9_3_3_5
void FASTCALL call_9_3_3_5(SH2_struct* sh)
{
  SH2movwi(sh, 3, (3<<4)+5);
}
#define init_9_3_3_6
void FASTCALL call_9_3_3_6(SH2_struct* sh)
{
  SH2movwi(sh, 3, (3<<4)+6);
}
#define init_9_3_3_7
void FASTCALL call_9_3_3_7(SH2_struct* sh)
{
  SH2movwi(sh, 3, (3<<4)+7);
}
#define init_9_3_3_8
void FASTCALL call_9_3_3_8(SH2_struct* sh)
{
  SH2movwi(sh, 3, (3<<4)+8);
}
#define init_9_3_3_9
void FASTCALL call_9_3_3_9(SH2_struct* sh)
{
  SH2movwi(sh, 3, (3<<4)+9);
}
#define init_9_3_3_10
void FASTCALL call_9_3_3_10(SH2_struct* sh)
{
  SH2movwi(sh, 3, (3<<4)+10);
}
#define init_9_3_3_11
void FASTCALL call_9_3_3_11(SH2_struct* sh)
{
  SH2movwi(sh, 3, (3<<4)+11);
}
#define init_9_3_3_12
void FASTCALL call_9_3_3_12(SH2_struct* sh)
{
  SH2movwi(sh, 3, (3<<4)+12);
}
#define init_9_3_3_13
void FASTCALL call_9_3_3_13(SH2_struct* sh)
{
  SH2movwi(sh, 3, (3<<4)+13);
}
#define init_9_3_3_14
void FASTCALL call_9_3_3_14(SH2_struct* sh)
{
  SH2movwi(sh, 3, (3<<4)+14);
}
#define init_9_3_3_15
void FASTCALL call_9_3_3_15(SH2_struct* sh)
{
  SH2movwi(sh, 3, (3<<4)+15);
}
#define init_9_3_4_0
void FASTCALL call_9_3_4_0(SH2_struct* sh)
{
  SH2movwi(sh, 3, (4<<4)+0);
}
#define init_9_3_4_1
void FASTCALL call_9_3_4_1(SH2_struct* sh)
{
  SH2movwi(sh, 3, (4<<4)+1);
}
#define init_9_3_4_2
void FASTCALL call_9_3_4_2(SH2_struct* sh)
{
  SH2movwi(sh, 3, (4<<4)+2);
}
#define init_9_3_4_3
void FASTCALL call_9_3_4_3(SH2_struct* sh)
{
  SH2movwi(sh, 3, (4<<4)+3);
}
#define init_9_3_4_4
void FASTCALL call_9_3_4_4(SH2_struct* sh)
{
  SH2movwi(sh, 3, (4<<4)+4);
}
#define init_9_3_4_5
void FASTCALL call_9_3_4_5(SH2_struct* sh)
{
  SH2movwi(sh, 3, (4<<4)+5);
}
#define init_9_3_4_6
void FASTCALL call_9_3_4_6(SH2_struct* sh)
{
  SH2movwi(sh, 3, (4<<4)+6);
}
#define init_9_3_4_7
void FASTCALL call_9_3_4_7(SH2_struct* sh)
{
  SH2movwi(sh, 3, (4<<4)+7);
}
#define init_9_3_4_8
void FASTCALL call_9_3_4_8(SH2_struct* sh)
{
  SH2movwi(sh, 3, (4<<4)+8);
}
#define init_9_3_4_9
void FASTCALL call_9_3_4_9(SH2_struct* sh)
{
  SH2movwi(sh, 3, (4<<4)+9);
}
#define init_9_3_4_10
void FASTCALL call_9_3_4_10(SH2_struct* sh)
{
  SH2movwi(sh, 3, (4<<4)+10);
}
#define init_9_3_4_11
void FASTCALL call_9_3_4_11(SH2_struct* sh)
{
  SH2movwi(sh, 3, (4<<4)+11);
}
#define init_9_3_4_12
void FASTCALL call_9_3_4_12(SH2_struct* sh)
{
  SH2movwi(sh, 3, (4<<4)+12);
}
#define init_9_3_4_13
void FASTCALL call_9_3_4_13(SH2_struct* sh)
{
  SH2movwi(sh, 3, (4<<4)+13);
}
#define init_9_3_4_14
void FASTCALL call_9_3_4_14(SH2_struct* sh)
{
  SH2movwi(sh, 3, (4<<4)+14);
}
#define init_9_3_4_15
void FASTCALL call_9_3_4_15(SH2_struct* sh)
{
  SH2movwi(sh, 3, (4<<4)+15);
}
#define init_9_3_5_0
void FASTCALL call_9_3_5_0(SH2_struct* sh)
{
  SH2movwi(sh, 3, (5<<4)+0);
}
#define init_9_3_5_1
void FASTCALL call_9_3_5_1(SH2_struct* sh)
{
  SH2movwi(sh, 3, (5<<4)+1);
}
#define init_9_3_5_2
void FASTCALL call_9_3_5_2(SH2_struct* sh)
{
  SH2movwi(sh, 3, (5<<4)+2);
}
#define init_9_3_5_3
void FASTCALL call_9_3_5_3(SH2_struct* sh)
{
  SH2movwi(sh, 3, (5<<4)+3);
}
#define init_9_3_5_4
void FASTCALL call_9_3_5_4(SH2_struct* sh)
{
  SH2movwi(sh, 3, (5<<4)+4);
}
#define init_9_3_5_5
void FASTCALL call_9_3_5_5(SH2_struct* sh)
{
  SH2movwi(sh, 3, (5<<4)+5);
}
#define init_9_3_5_6
void FASTCALL call_9_3_5_6(SH2_struct* sh)
{
  SH2movwi(sh, 3, (5<<4)+6);
}
#define init_9_3_5_7
void FASTCALL call_9_3_5_7(SH2_struct* sh)
{
  SH2movwi(sh, 3, (5<<4)+7);
}
#define init_9_3_5_8
void FASTCALL call_9_3_5_8(SH2_struct* sh)
{
  SH2movwi(sh, 3, (5<<4)+8);
}
#define init_9_3_5_9
void FASTCALL call_9_3_5_9(SH2_struct* sh)
{
  SH2movwi(sh, 3, (5<<4)+9);
}
#define init_9_3_5_10
void FASTCALL call_9_3_5_10(SH2_struct* sh)
{
  SH2movwi(sh, 3, (5<<4)+10);
}
#define init_9_3_5_11
void FASTCALL call_9_3_5_11(SH2_struct* sh)
{
  SH2movwi(sh, 3, (5<<4)+11);
}
#define init_9_3_5_12
void FASTCALL call_9_3_5_12(SH2_struct* sh)
{
  SH2movwi(sh, 3, (5<<4)+12);
}
#define init_9_3_5_13
void FASTCALL call_9_3_5_13(SH2_struct* sh)
{
  SH2movwi(sh, 3, (5<<4)+13);
}
#define init_9_3_5_14
void FASTCALL call_9_3_5_14(SH2_struct* sh)
{
  SH2movwi(sh, 3, (5<<4)+14);
}
#define init_9_3_5_15
void FASTCALL call_9_3_5_15(SH2_struct* sh)
{
  SH2movwi(sh, 3, (5<<4)+15);
}
#define init_9_3_6_0
void FASTCALL call_9_3_6_0(SH2_struct* sh)
{
  SH2movwi(sh, 3, (6<<4)+0);
}
#define init_9_3_6_1
void FASTCALL call_9_3_6_1(SH2_struct* sh)
{
  SH2movwi(sh, 3, (6<<4)+1);
}
#define init_9_3_6_2
void FASTCALL call_9_3_6_2(SH2_struct* sh)
{
  SH2movwi(sh, 3, (6<<4)+2);
}
#define init_9_3_6_3
void FASTCALL call_9_3_6_3(SH2_struct* sh)
{
  SH2movwi(sh, 3, (6<<4)+3);
}
#define init_9_3_6_4
void FASTCALL call_9_3_6_4(SH2_struct* sh)
{
  SH2movwi(sh, 3, (6<<4)+4);
}
#define init_9_3_6_5
void FASTCALL call_9_3_6_5(SH2_struct* sh)
{
  SH2movwi(sh, 3, (6<<4)+5);
}
#define init_9_3_6_6
void FASTCALL call_9_3_6_6(SH2_struct* sh)
{
  SH2movwi(sh, 3, (6<<4)+6);
}
#define init_9_3_6_7
void FASTCALL call_9_3_6_7(SH2_struct* sh)
{
  SH2movwi(sh, 3, (6<<4)+7);
}
#define init_9_3_6_8
void FASTCALL call_9_3_6_8(SH2_struct* sh)
{
  SH2movwi(sh, 3, (6<<4)+8);
}
#define init_9_3_6_9
void FASTCALL call_9_3_6_9(SH2_struct* sh)
{
  SH2movwi(sh, 3, (6<<4)+9);
}
#define init_9_3_6_10
void FASTCALL call_9_3_6_10(SH2_struct* sh)
{
  SH2movwi(sh, 3, (6<<4)+10);
}
#define init_9_3_6_11
void FASTCALL call_9_3_6_11(SH2_struct* sh)
{
  SH2movwi(sh, 3, (6<<4)+11);
}
#define init_9_3_6_12
void FASTCALL call_9_3_6_12(SH2_struct* sh)
{
  SH2movwi(sh, 3, (6<<4)+12);
}
#define init_9_3_6_13
void FASTCALL call_9_3_6_13(SH2_struct* sh)
{
  SH2movwi(sh, 3, (6<<4)+13);
}
#define init_9_3_6_14
void FASTCALL call_9_3_6_14(SH2_struct* sh)
{
  SH2movwi(sh, 3, (6<<4)+14);
}
#define init_9_3_6_15
void FASTCALL call_9_3_6_15(SH2_struct* sh)
{
  SH2movwi(sh, 3, (6<<4)+15);
}
#define init_9_3_7_0
void FASTCALL call_9_3_7_0(SH2_struct* sh)
{
  SH2movwi(sh, 3, (7<<4)+0);
}
#define init_9_3_7_1
void FASTCALL call_9_3_7_1(SH2_struct* sh)
{
  SH2movwi(sh, 3, (7<<4)+1);
}
#define init_9_3_7_2
void FASTCALL call_9_3_7_2(SH2_struct* sh)
{
  SH2movwi(sh, 3, (7<<4)+2);
}
#define init_9_3_7_3
void FASTCALL call_9_3_7_3(SH2_struct* sh)
{
  SH2movwi(sh, 3, (7<<4)+3);
}
#define init_9_3_7_4
void FASTCALL call_9_3_7_4(SH2_struct* sh)
{
  SH2movwi(sh, 3, (7<<4)+4);
}
#define init_9_3_7_5
void FASTCALL call_9_3_7_5(SH2_struct* sh)
{
  SH2movwi(sh, 3, (7<<4)+5);
}
#define init_9_3_7_6
void FASTCALL call_9_3_7_6(SH2_struct* sh)
{
  SH2movwi(sh, 3, (7<<4)+6);
}
#define init_9_3_7_7
void FASTCALL call_9_3_7_7(SH2_struct* sh)
{
  SH2movwi(sh, 3, (7<<4)+7);
}
#define init_9_3_7_8
void FASTCALL call_9_3_7_8(SH2_struct* sh)
{
  SH2movwi(sh, 3, (7<<4)+8);
}
#define init_9_3_7_9
void FASTCALL call_9_3_7_9(SH2_struct* sh)
{
  SH2movwi(sh, 3, (7<<4)+9);
}
#define init_9_3_7_10
void FASTCALL call_9_3_7_10(SH2_struct* sh)
{
  SH2movwi(sh, 3, (7<<4)+10);
}
#define init_9_3_7_11
void FASTCALL call_9_3_7_11(SH2_struct* sh)
{
  SH2movwi(sh, 3, (7<<4)+11);
}
#define init_9_3_7_12
void FASTCALL call_9_3_7_12(SH2_struct* sh)
{
  SH2movwi(sh, 3, (7<<4)+12);
}
#define init_9_3_7_13
void FASTCALL call_9_3_7_13(SH2_struct* sh)
{
  SH2movwi(sh, 3, (7<<4)+13);
}
#define init_9_3_7_14
void FASTCALL call_9_3_7_14(SH2_struct* sh)
{
  SH2movwi(sh, 3, (7<<4)+14);
}
#define init_9_3_7_15
void FASTCALL call_9_3_7_15(SH2_struct* sh)
{
  SH2movwi(sh, 3, (7<<4)+15);
}
#define init_9_3_8_0
void FASTCALL call_9_3_8_0(SH2_struct* sh)
{
  SH2movwi(sh, 3, (8<<4)+0);
}
#define init_9_3_8_1
void FASTCALL call_9_3_8_1(SH2_struct* sh)
{
  SH2movwi(sh, 3, (8<<4)+1);
}
#define init_9_3_8_2
void FASTCALL call_9_3_8_2(SH2_struct* sh)
{
  SH2movwi(sh, 3, (8<<4)+2);
}
#define init_9_3_8_3
void FASTCALL call_9_3_8_3(SH2_struct* sh)
{
  SH2movwi(sh, 3, (8<<4)+3);
}
#define init_9_3_8_4
void FASTCALL call_9_3_8_4(SH2_struct* sh)
{
  SH2movwi(sh, 3, (8<<4)+4);
}
#define init_9_3_8_5
void FASTCALL call_9_3_8_5(SH2_struct* sh)
{
  SH2movwi(sh, 3, (8<<4)+5);
}
#define init_9_3_8_6
void FASTCALL call_9_3_8_6(SH2_struct* sh)
{
  SH2movwi(sh, 3, (8<<4)+6);
}
#define init_9_3_8_7
void FASTCALL call_9_3_8_7(SH2_struct* sh)
{
  SH2movwi(sh, 3, (8<<4)+7);
}
#define init_9_3_8_8
void FASTCALL call_9_3_8_8(SH2_struct* sh)
{
  SH2movwi(sh, 3, (8<<4)+8);
}
#define init_9_3_8_9
void FASTCALL call_9_3_8_9(SH2_struct* sh)
{
  SH2movwi(sh, 3, (8<<4)+9);
}
#define init_9_3_8_10
void FASTCALL call_9_3_8_10(SH2_struct* sh)
{
  SH2movwi(sh, 3, (8<<4)+10);
}
#define init_9_3_8_11
void FASTCALL call_9_3_8_11(SH2_struct* sh)
{
  SH2movwi(sh, 3, (8<<4)+11);
}
#define init_9_3_8_12
void FASTCALL call_9_3_8_12(SH2_struct* sh)
{
  SH2movwi(sh, 3, (8<<4)+12);
}
#define init_9_3_8_13
void FASTCALL call_9_3_8_13(SH2_struct* sh)
{
  SH2movwi(sh, 3, (8<<4)+13);
}
#define init_9_3_8_14
void FASTCALL call_9_3_8_14(SH2_struct* sh)
{
  SH2movwi(sh, 3, (8<<4)+14);
}
#define init_9_3_8_15
void FASTCALL call_9_3_8_15(SH2_struct* sh)
{
  SH2movwi(sh, 3, (8<<4)+15);
}
#define init_9_3_9_0
void FASTCALL call_9_3_9_0(SH2_struct* sh)
{
  SH2movwi(sh, 3, (9<<4)+0);
}
#define init_9_3_9_1
void FASTCALL call_9_3_9_1(SH2_struct* sh)
{
  SH2movwi(sh, 3, (9<<4)+1);
}
#define init_9_3_9_2
void FASTCALL call_9_3_9_2(SH2_struct* sh)
{
  SH2movwi(sh, 3, (9<<4)+2);
}
#define init_9_3_9_3
void FASTCALL call_9_3_9_3(SH2_struct* sh)
{
  SH2movwi(sh, 3, (9<<4)+3);
}
#define init_9_3_9_4
void FASTCALL call_9_3_9_4(SH2_struct* sh)
{
  SH2movwi(sh, 3, (9<<4)+4);
}
#define init_9_3_9_5
void FASTCALL call_9_3_9_5(SH2_struct* sh)
{
  SH2movwi(sh, 3, (9<<4)+5);
}
#define init_9_3_9_6
void FASTCALL call_9_3_9_6(SH2_struct* sh)
{
  SH2movwi(sh, 3, (9<<4)+6);
}
#define init_9_3_9_7
void FASTCALL call_9_3_9_7(SH2_struct* sh)
{
  SH2movwi(sh, 3, (9<<4)+7);
}
#define init_9_3_9_8
void FASTCALL call_9_3_9_8(SH2_struct* sh)
{
  SH2movwi(sh, 3, (9<<4)+8);
}
#define init_9_3_9_9
void FASTCALL call_9_3_9_9(SH2_struct* sh)
{
  SH2movwi(sh, 3, (9<<4)+9);
}
#define init_9_3_9_10
void FASTCALL call_9_3_9_10(SH2_struct* sh)
{
  SH2movwi(sh, 3, (9<<4)+10);
}
#define init_9_3_9_11
void FASTCALL call_9_3_9_11(SH2_struct* sh)
{
  SH2movwi(sh, 3, (9<<4)+11);
}
#define init_9_3_9_12
void FASTCALL call_9_3_9_12(SH2_struct* sh)
{
  SH2movwi(sh, 3, (9<<4)+12);
}
#define init_9_3_9_13
void FASTCALL call_9_3_9_13(SH2_struct* sh)
{
  SH2movwi(sh, 3, (9<<4)+13);
}
#define init_9_3_9_14
void FASTCALL call_9_3_9_14(SH2_struct* sh)
{
  SH2movwi(sh, 3, (9<<4)+14);
}
#define init_9_3_9_15
void FASTCALL call_9_3_9_15(SH2_struct* sh)
{
  SH2movwi(sh, 3, (9<<4)+15);
}
#define init_9_3_10_0
void FASTCALL call_9_3_10_0(SH2_struct* sh)
{
  SH2movwi(sh, 3, (10<<4)+0);
}
#define init_9_3_10_1
void FASTCALL call_9_3_10_1(SH2_struct* sh)
{
  SH2movwi(sh, 3, (10<<4)+1);
}
#define init_9_3_10_2
void FASTCALL call_9_3_10_2(SH2_struct* sh)
{
  SH2movwi(sh, 3, (10<<4)+2);
}
#define init_9_3_10_3
void FASTCALL call_9_3_10_3(SH2_struct* sh)
{
  SH2movwi(sh, 3, (10<<4)+3);
}
#define init_9_3_10_4
void FASTCALL call_9_3_10_4(SH2_struct* sh)
{
  SH2movwi(sh, 3, (10<<4)+4);
}
#define init_9_3_10_5
void FASTCALL call_9_3_10_5(SH2_struct* sh)
{
  SH2movwi(sh, 3, (10<<4)+5);
}
#define init_9_3_10_6
void FASTCALL call_9_3_10_6(SH2_struct* sh)
{
  SH2movwi(sh, 3, (10<<4)+6);
}
#define init_9_3_10_7
void FASTCALL call_9_3_10_7(SH2_struct* sh)
{
  SH2movwi(sh, 3, (10<<4)+7);
}
#define init_9_3_10_8
void FASTCALL call_9_3_10_8(SH2_struct* sh)
{
  SH2movwi(sh, 3, (10<<4)+8);
}
#define init_9_3_10_9
void FASTCALL call_9_3_10_9(SH2_struct* sh)
{
  SH2movwi(sh, 3, (10<<4)+9);
}
#define init_9_3_10_10
void FASTCALL call_9_3_10_10(SH2_struct* sh)
{
  SH2movwi(sh, 3, (10<<4)+10);
}
#define init_9_3_10_11
void FASTCALL call_9_3_10_11(SH2_struct* sh)
{
  SH2movwi(sh, 3, (10<<4)+11);
}
#define init_9_3_10_12
void FASTCALL call_9_3_10_12(SH2_struct* sh)
{
  SH2movwi(sh, 3, (10<<4)+12);
}
#define init_9_3_10_13
void FASTCALL call_9_3_10_13(SH2_struct* sh)
{
  SH2movwi(sh, 3, (10<<4)+13);
}
#define init_9_3_10_14
void FASTCALL call_9_3_10_14(SH2_struct* sh)
{
  SH2movwi(sh, 3, (10<<4)+14);
}
#define init_9_3_10_15
void FASTCALL call_9_3_10_15(SH2_struct* sh)
{
  SH2movwi(sh, 3, (10<<4)+15);
}
#define init_9_3_11_0
void FASTCALL call_9_3_11_0(SH2_struct* sh)
{
  SH2movwi(sh, 3, (11<<4)+0);
}
#define init_9_3_11_1
void FASTCALL call_9_3_11_1(SH2_struct* sh)
{
  SH2movwi(sh, 3, (11<<4)+1);
}
#define init_9_3_11_2
void FASTCALL call_9_3_11_2(SH2_struct* sh)
{
  SH2movwi(sh, 3, (11<<4)+2);
}
#define init_9_3_11_3
void FASTCALL call_9_3_11_3(SH2_struct* sh)
{
  SH2movwi(sh, 3, (11<<4)+3);
}
#define init_9_3_11_4
void FASTCALL call_9_3_11_4(SH2_struct* sh)
{
  SH2movwi(sh, 3, (11<<4)+4);
}
#define init_9_3_11_5
void FASTCALL call_9_3_11_5(SH2_struct* sh)
{
  SH2movwi(sh, 3, (11<<4)+5);
}
#define init_9_3_11_6
void FASTCALL call_9_3_11_6(SH2_struct* sh)
{
  SH2movwi(sh, 3, (11<<4)+6);
}
#define init_9_3_11_7
void FASTCALL call_9_3_11_7(SH2_struct* sh)
{
  SH2movwi(sh, 3, (11<<4)+7);
}
#define init_9_3_11_8
void FASTCALL call_9_3_11_8(SH2_struct* sh)
{
  SH2movwi(sh, 3, (11<<4)+8);
}
#define init_9_3_11_9
void FASTCALL call_9_3_11_9(SH2_struct* sh)
{
  SH2movwi(sh, 3, (11<<4)+9);
}
#define init_9_3_11_10
void FASTCALL call_9_3_11_10(SH2_struct* sh)
{
  SH2movwi(sh, 3, (11<<4)+10);
}
#define init_9_3_11_11
void FASTCALL call_9_3_11_11(SH2_struct* sh)
{
  SH2movwi(sh, 3, (11<<4)+11);
}
#define init_9_3_11_12
void FASTCALL call_9_3_11_12(SH2_struct* sh)
{
  SH2movwi(sh, 3, (11<<4)+12);
}
#define init_9_3_11_13
void FASTCALL call_9_3_11_13(SH2_struct* sh)
{
  SH2movwi(sh, 3, (11<<4)+13);
}
#define init_9_3_11_14
void FASTCALL call_9_3_11_14(SH2_struct* sh)
{
  SH2movwi(sh, 3, (11<<4)+14);
}
#define init_9_3_11_15
void FASTCALL call_9_3_11_15(SH2_struct* sh)
{
  SH2movwi(sh, 3, (11<<4)+15);
}
#define init_9_3_12_0
void FASTCALL call_9_3_12_0(SH2_struct* sh)
{
  SH2movwi(sh, 3, (12<<4)+0);
}
#define init_9_3_12_1
void FASTCALL call_9_3_12_1(SH2_struct* sh)
{
  SH2movwi(sh, 3, (12<<4)+1);
}
#define init_9_3_12_2
void FASTCALL call_9_3_12_2(SH2_struct* sh)
{
  SH2movwi(sh, 3, (12<<4)+2);
}
#define init_9_3_12_3
void FASTCALL call_9_3_12_3(SH2_struct* sh)
{
  SH2movwi(sh, 3, (12<<4)+3);
}
#define init_9_3_12_4
void FASTCALL call_9_3_12_4(SH2_struct* sh)
{
  SH2movwi(sh, 3, (12<<4)+4);
}
#define init_9_3_12_5
void FASTCALL call_9_3_12_5(SH2_struct* sh)
{
  SH2movwi(sh, 3, (12<<4)+5);
}
#define init_9_3_12_6
void FASTCALL call_9_3_12_6(SH2_struct* sh)
{
  SH2movwi(sh, 3, (12<<4)+6);
}
#define init_9_3_12_7
void FASTCALL call_9_3_12_7(SH2_struct* sh)
{
  SH2movwi(sh, 3, (12<<4)+7);
}
#define init_9_3_12_8
void FASTCALL call_9_3_12_8(SH2_struct* sh)
{
  SH2movwi(sh, 3, (12<<4)+8);
}
#define init_9_3_12_9
void FASTCALL call_9_3_12_9(SH2_struct* sh)
{
  SH2movwi(sh, 3, (12<<4)+9);
}
#define init_9_3_12_10
void FASTCALL call_9_3_12_10(SH2_struct* sh)
{
  SH2movwi(sh, 3, (12<<4)+10);
}
#define init_9_3_12_11
void FASTCALL call_9_3_12_11(SH2_struct* sh)
{
  SH2movwi(sh, 3, (12<<4)+11);
}
#define init_9_3_12_12
void FASTCALL call_9_3_12_12(SH2_struct* sh)
{
  SH2movwi(sh, 3, (12<<4)+12);
}
#define init_9_3_12_13
void FASTCALL call_9_3_12_13(SH2_struct* sh)
{
  SH2movwi(sh, 3, (12<<4)+13);
}
#define init_9_3_12_14
void FASTCALL call_9_3_12_14(SH2_struct* sh)
{
  SH2movwi(sh, 3, (12<<4)+14);
}
#define init_9_3_12_15
void FASTCALL call_9_3_12_15(SH2_struct* sh)
{
  SH2movwi(sh, 3, (12<<4)+15);
}
#define init_9_3_13_0
void FASTCALL call_9_3_13_0(SH2_struct* sh)
{
  SH2movwi(sh, 3, (13<<4)+0);
}
#define init_9_3_13_1
void FASTCALL call_9_3_13_1(SH2_struct* sh)
{
  SH2movwi(sh, 3, (13<<4)+1);
}
#define init_9_3_13_2
void FASTCALL call_9_3_13_2(SH2_struct* sh)
{
  SH2movwi(sh, 3, (13<<4)+2);
}
#define init_9_3_13_3
void FASTCALL call_9_3_13_3(SH2_struct* sh)
{
  SH2movwi(sh, 3, (13<<4)+3);
}
#define init_9_3_13_4
void FASTCALL call_9_3_13_4(SH2_struct* sh)
{
  SH2movwi(sh, 3, (13<<4)+4);
}
#define init_9_3_13_5
void FASTCALL call_9_3_13_5(SH2_struct* sh)
{
  SH2movwi(sh, 3, (13<<4)+5);
}
#define init_9_3_13_6
void FASTCALL call_9_3_13_6(SH2_struct* sh)
{
  SH2movwi(sh, 3, (13<<4)+6);
}
#define init_9_3_13_7
void FASTCALL call_9_3_13_7(SH2_struct* sh)
{
  SH2movwi(sh, 3, (13<<4)+7);
}
#define init_9_3_13_8
void FASTCALL call_9_3_13_8(SH2_struct* sh)
{
  SH2movwi(sh, 3, (13<<4)+8);
}
#define init_9_3_13_9
void FASTCALL call_9_3_13_9(SH2_struct* sh)
{
  SH2movwi(sh, 3, (13<<4)+9);
}
#define init_9_3_13_10
void FASTCALL call_9_3_13_10(SH2_struct* sh)
{
  SH2movwi(sh, 3, (13<<4)+10);
}
#define init_9_3_13_11
void FASTCALL call_9_3_13_11(SH2_struct* sh)
{
  SH2movwi(sh, 3, (13<<4)+11);
}
#define init_9_3_13_12
void FASTCALL call_9_3_13_12(SH2_struct* sh)
{
  SH2movwi(sh, 3, (13<<4)+12);
}
#define init_9_3_13_13
void FASTCALL call_9_3_13_13(SH2_struct* sh)
{
  SH2movwi(sh, 3, (13<<4)+13);
}
#define init_9_3_13_14
void FASTCALL call_9_3_13_14(SH2_struct* sh)
{
  SH2movwi(sh, 3, (13<<4)+14);
}
#define init_9_3_13_15
void FASTCALL call_9_3_13_15(SH2_struct* sh)
{
  SH2movwi(sh, 3, (13<<4)+15);
}
#define init_9_3_14_0
void FASTCALL call_9_3_14_0(SH2_struct* sh)
{
  SH2movwi(sh, 3, (14<<4)+0);
}
#define init_9_3_14_1
void FASTCALL call_9_3_14_1(SH2_struct* sh)
{
  SH2movwi(sh, 3, (14<<4)+1);
}
#define init_9_3_14_2
void FASTCALL call_9_3_14_2(SH2_struct* sh)
{
  SH2movwi(sh, 3, (14<<4)+2);
}
#define init_9_3_14_3
void FASTCALL call_9_3_14_3(SH2_struct* sh)
{
  SH2movwi(sh, 3, (14<<4)+3);
}
#define init_9_3_14_4
void FASTCALL call_9_3_14_4(SH2_struct* sh)
{
  SH2movwi(sh, 3, (14<<4)+4);
}
#define init_9_3_14_5
void FASTCALL call_9_3_14_5(SH2_struct* sh)
{
  SH2movwi(sh, 3, (14<<4)+5);
}
#define init_9_3_14_6
void FASTCALL call_9_3_14_6(SH2_struct* sh)
{
  SH2movwi(sh, 3, (14<<4)+6);
}
#define init_9_3_14_7
void FASTCALL call_9_3_14_7(SH2_struct* sh)
{
  SH2movwi(sh, 3, (14<<4)+7);
}
#define init_9_3_14_8
void FASTCALL call_9_3_14_8(SH2_struct* sh)
{
  SH2movwi(sh, 3, (14<<4)+8);
}
#define init_9_3_14_9
void FASTCALL call_9_3_14_9(SH2_struct* sh)
{
  SH2movwi(sh, 3, (14<<4)+9);
}
#define init_9_3_14_10
void FASTCALL call_9_3_14_10(SH2_struct* sh)
{
  SH2movwi(sh, 3, (14<<4)+10);
}
#define init_9_3_14_11
void FASTCALL call_9_3_14_11(SH2_struct* sh)
{
  SH2movwi(sh, 3, (14<<4)+11);
}
#define init_9_3_14_12
void FASTCALL call_9_3_14_12(SH2_struct* sh)
{
  SH2movwi(sh, 3, (14<<4)+12);
}
#define init_9_3_14_13
void FASTCALL call_9_3_14_13(SH2_struct* sh)
{
  SH2movwi(sh, 3, (14<<4)+13);
}
#define init_9_3_14_14
void FASTCALL call_9_3_14_14(SH2_struct* sh)
{
  SH2movwi(sh, 3, (14<<4)+14);
}
#define init_9_3_14_15
void FASTCALL call_9_3_14_15(SH2_struct* sh)
{
  SH2movwi(sh, 3, (14<<4)+15);
}
#define init_9_3_15_0
void FASTCALL call_9_3_15_0(SH2_struct* sh)
{
  SH2movwi(sh, 3, (15<<4)+0);
}
#define init_9_3_15_1
void FASTCALL call_9_3_15_1(SH2_struct* sh)
{
  SH2movwi(sh, 3, (15<<4)+1);
}
#define init_9_3_15_2
void FASTCALL call_9_3_15_2(SH2_struct* sh)
{
  SH2movwi(sh, 3, (15<<4)+2);
}
#define init_9_3_15_3
void FASTCALL call_9_3_15_3(SH2_struct* sh)
{
  SH2movwi(sh, 3, (15<<4)+3);
}
#define init_9_3_15_4
void FASTCALL call_9_3_15_4(SH2_struct* sh)
{
  SH2movwi(sh, 3, (15<<4)+4);
}
#define init_9_3_15_5
void FASTCALL call_9_3_15_5(SH2_struct* sh)
{
  SH2movwi(sh, 3, (15<<4)+5);
}
#define init_9_3_15_6
void FASTCALL call_9_3_15_6(SH2_struct* sh)
{
  SH2movwi(sh, 3, (15<<4)+6);
}
#define init_9_3_15_7
void FASTCALL call_9_3_15_7(SH2_struct* sh)
{
  SH2movwi(sh, 3, (15<<4)+7);
}
#define init_9_3_15_8
void FASTCALL call_9_3_15_8(SH2_struct* sh)
{
  SH2movwi(sh, 3, (15<<4)+8);
}
#define init_9_3_15_9
void FASTCALL call_9_3_15_9(SH2_struct* sh)
{
  SH2movwi(sh, 3, (15<<4)+9);
}
#define init_9_3_15_10
void FASTCALL call_9_3_15_10(SH2_struct* sh)
{
  SH2movwi(sh, 3, (15<<4)+10);
}
#define init_9_3_15_11
void FASTCALL call_9_3_15_11(SH2_struct* sh)
{
  SH2movwi(sh, 3, (15<<4)+11);
}
#define init_9_3_15_12
void FASTCALL call_9_3_15_12(SH2_struct* sh)
{
  SH2movwi(sh, 3, (15<<4)+12);
}
#define init_9_3_15_13
void FASTCALL call_9_3_15_13(SH2_struct* sh)
{
  SH2movwi(sh, 3, (15<<4)+13);
}
#define init_9_3_15_14
void FASTCALL call_9_3_15_14(SH2_struct* sh)
{
  SH2movwi(sh, 3, (15<<4)+14);
}
#define init_9_3_15_15
void FASTCALL call_9_3_15_15(SH2_struct* sh)
{
  SH2movwi(sh, 3, (15<<4)+15);
}
#define init_9_4_0_0
void FASTCALL call_9_4_0_0(SH2_struct* sh)
{
  SH2movwi(sh, 4, (0<<4)+0);
}
#define init_9_4_0_1
void FASTCALL call_9_4_0_1(SH2_struct* sh)
{
  SH2movwi(sh, 4, (0<<4)+1);
}
#define init_9_4_0_2
void FASTCALL call_9_4_0_2(SH2_struct* sh)
{
  SH2movwi(sh, 4, (0<<4)+2);
}
#define init_9_4_0_3
void FASTCALL call_9_4_0_3(SH2_struct* sh)
{
  SH2movwi(sh, 4, (0<<4)+3);
}
#define init_9_4_0_4
void FASTCALL call_9_4_0_4(SH2_struct* sh)
{
  SH2movwi(sh, 4, (0<<4)+4);
}
#define init_9_4_0_5
void FASTCALL call_9_4_0_5(SH2_struct* sh)
{
  SH2movwi(sh, 4, (0<<4)+5);
}
#define init_9_4_0_6
void FASTCALL call_9_4_0_6(SH2_struct* sh)
{
  SH2movwi(sh, 4, (0<<4)+6);
}
#define init_9_4_0_7
void FASTCALL call_9_4_0_7(SH2_struct* sh)
{
  SH2movwi(sh, 4, (0<<4)+7);
}
#define init_9_4_0_8
void FASTCALL call_9_4_0_8(SH2_struct* sh)
{
  SH2movwi(sh, 4, (0<<4)+8);
}
#define init_9_4_0_9
void FASTCALL call_9_4_0_9(SH2_struct* sh)
{
  SH2movwi(sh, 4, (0<<4)+9);
}
#define init_9_4_0_10
void FASTCALL call_9_4_0_10(SH2_struct* sh)
{
  SH2movwi(sh, 4, (0<<4)+10);
}
#define init_9_4_0_11
void FASTCALL call_9_4_0_11(SH2_struct* sh)
{
  SH2movwi(sh, 4, (0<<4)+11);
}
#define init_9_4_0_12
void FASTCALL call_9_4_0_12(SH2_struct* sh)
{
  SH2movwi(sh, 4, (0<<4)+12);
}
#define init_9_4_0_13
void FASTCALL call_9_4_0_13(SH2_struct* sh)
{
  SH2movwi(sh, 4, (0<<4)+13);
}
#define init_9_4_0_14
void FASTCALL call_9_4_0_14(SH2_struct* sh)
{
  SH2movwi(sh, 4, (0<<4)+14);
}
#define init_9_4_0_15
void FASTCALL call_9_4_0_15(SH2_struct* sh)
{
  SH2movwi(sh, 4, (0<<4)+15);
}
#define init_9_4_1_0
void FASTCALL call_9_4_1_0(SH2_struct* sh)
{
  SH2movwi(sh, 4, (1<<4)+0);
}
#define init_9_4_1_1
void FASTCALL call_9_4_1_1(SH2_struct* sh)
{
  SH2movwi(sh, 4, (1<<4)+1);
}
#define init_9_4_1_2
void FASTCALL call_9_4_1_2(SH2_struct* sh)
{
  SH2movwi(sh, 4, (1<<4)+2);
}
#define init_9_4_1_3
void FASTCALL call_9_4_1_3(SH2_struct* sh)
{
  SH2movwi(sh, 4, (1<<4)+3);
}
#define init_9_4_1_4
void FASTCALL call_9_4_1_4(SH2_struct* sh)
{
  SH2movwi(sh, 4, (1<<4)+4);
}
#define init_9_4_1_5
void FASTCALL call_9_4_1_5(SH2_struct* sh)
{
  SH2movwi(sh, 4, (1<<4)+5);
}
#define init_9_4_1_6
void FASTCALL call_9_4_1_6(SH2_struct* sh)
{
  SH2movwi(sh, 4, (1<<4)+6);
}
#define init_9_4_1_7
void FASTCALL call_9_4_1_7(SH2_struct* sh)
{
  SH2movwi(sh, 4, (1<<4)+7);
}
#define init_9_4_1_8
void FASTCALL call_9_4_1_8(SH2_struct* sh)
{
  SH2movwi(sh, 4, (1<<4)+8);
}
#define init_9_4_1_9
void FASTCALL call_9_4_1_9(SH2_struct* sh)
{
  SH2movwi(sh, 4, (1<<4)+9);
}
#define init_9_4_1_10
void FASTCALL call_9_4_1_10(SH2_struct* sh)
{
  SH2movwi(sh, 4, (1<<4)+10);
}
#define init_9_4_1_11
void FASTCALL call_9_4_1_11(SH2_struct* sh)
{
  SH2movwi(sh, 4, (1<<4)+11);
}
#define init_9_4_1_12
void FASTCALL call_9_4_1_12(SH2_struct* sh)
{
  SH2movwi(sh, 4, (1<<4)+12);
}
#define init_9_4_1_13
void FASTCALL call_9_4_1_13(SH2_struct* sh)
{
  SH2movwi(sh, 4, (1<<4)+13);
}
#define init_9_4_1_14
void FASTCALL call_9_4_1_14(SH2_struct* sh)
{
  SH2movwi(sh, 4, (1<<4)+14);
}
#define init_9_4_1_15
void FASTCALL call_9_4_1_15(SH2_struct* sh)
{
  SH2movwi(sh, 4, (1<<4)+15);
}
#define init_9_4_2_0
void FASTCALL call_9_4_2_0(SH2_struct* sh)
{
  SH2movwi(sh, 4, (2<<4)+0);
}
#define init_9_4_2_1
void FASTCALL call_9_4_2_1(SH2_struct* sh)
{
  SH2movwi(sh, 4, (2<<4)+1);
}
#define init_9_4_2_2
void FASTCALL call_9_4_2_2(SH2_struct* sh)
{
  SH2movwi(sh, 4, (2<<4)+2);
}
#define init_9_4_2_3
void FASTCALL call_9_4_2_3(SH2_struct* sh)
{
  SH2movwi(sh, 4, (2<<4)+3);
}
#define init_9_4_2_4
void FASTCALL call_9_4_2_4(SH2_struct* sh)
{
  SH2movwi(sh, 4, (2<<4)+4);
}
#define init_9_4_2_5
void FASTCALL call_9_4_2_5(SH2_struct* sh)
{
  SH2movwi(sh, 4, (2<<4)+5);
}
#define init_9_4_2_6
void FASTCALL call_9_4_2_6(SH2_struct* sh)
{
  SH2movwi(sh, 4, (2<<4)+6);
}
#define init_9_4_2_7
void FASTCALL call_9_4_2_7(SH2_struct* sh)
{
  SH2movwi(sh, 4, (2<<4)+7);
}
#define init_9_4_2_8
void FASTCALL call_9_4_2_8(SH2_struct* sh)
{
  SH2movwi(sh, 4, (2<<4)+8);
}
#define init_9_4_2_9
void FASTCALL call_9_4_2_9(SH2_struct* sh)
{
  SH2movwi(sh, 4, (2<<4)+9);
}
#define init_9_4_2_10
void FASTCALL call_9_4_2_10(SH2_struct* sh)
{
  SH2movwi(sh, 4, (2<<4)+10);
}
#define init_9_4_2_11
void FASTCALL call_9_4_2_11(SH2_struct* sh)
{
  SH2movwi(sh, 4, (2<<4)+11);
}
#define init_9_4_2_12
void FASTCALL call_9_4_2_12(SH2_struct* sh)
{
  SH2movwi(sh, 4, (2<<4)+12);
}
#define init_9_4_2_13
void FASTCALL call_9_4_2_13(SH2_struct* sh)
{
  SH2movwi(sh, 4, (2<<4)+13);
}
#define init_9_4_2_14
void FASTCALL call_9_4_2_14(SH2_struct* sh)
{
  SH2movwi(sh, 4, (2<<4)+14);
}
#define init_9_4_2_15
void FASTCALL call_9_4_2_15(SH2_struct* sh)
{
  SH2movwi(sh, 4, (2<<4)+15);
}
#define init_9_4_3_0
void FASTCALL call_9_4_3_0(SH2_struct* sh)
{
  SH2movwi(sh, 4, (3<<4)+0);
}
#define init_9_4_3_1
void FASTCALL call_9_4_3_1(SH2_struct* sh)
{
  SH2movwi(sh, 4, (3<<4)+1);
}
#define init_9_4_3_2
void FASTCALL call_9_4_3_2(SH2_struct* sh)
{
  SH2movwi(sh, 4, (3<<4)+2);
}
#define init_9_4_3_3
void FASTCALL call_9_4_3_3(SH2_struct* sh)
{
  SH2movwi(sh, 4, (3<<4)+3);
}
#define init_9_4_3_4
void FASTCALL call_9_4_3_4(SH2_struct* sh)
{
  SH2movwi(sh, 4, (3<<4)+4);
}
#define init_9_4_3_5
void FASTCALL call_9_4_3_5(SH2_struct* sh)
{
  SH2movwi(sh, 4, (3<<4)+5);
}
#define init_9_4_3_6
void FASTCALL call_9_4_3_6(SH2_struct* sh)
{
  SH2movwi(sh, 4, (3<<4)+6);
}
#define init_9_4_3_7
void FASTCALL call_9_4_3_7(SH2_struct* sh)
{
  SH2movwi(sh, 4, (3<<4)+7);
}
#define init_9_4_3_8
void FASTCALL call_9_4_3_8(SH2_struct* sh)
{
  SH2movwi(sh, 4, (3<<4)+8);
}
#define init_9_4_3_9
void FASTCALL call_9_4_3_9(SH2_struct* sh)
{
  SH2movwi(sh, 4, (3<<4)+9);
}
#define init_9_4_3_10
void FASTCALL call_9_4_3_10(SH2_struct* sh)
{
  SH2movwi(sh, 4, (3<<4)+10);
}
#define init_9_4_3_11
void FASTCALL call_9_4_3_11(SH2_struct* sh)
{
  SH2movwi(sh, 4, (3<<4)+11);
}
#define init_9_4_3_12
void FASTCALL call_9_4_3_12(SH2_struct* sh)
{
  SH2movwi(sh, 4, (3<<4)+12);
}
#define init_9_4_3_13
void FASTCALL call_9_4_3_13(SH2_struct* sh)
{
  SH2movwi(sh, 4, (3<<4)+13);
}
#define init_9_4_3_14
void FASTCALL call_9_4_3_14(SH2_struct* sh)
{
  SH2movwi(sh, 4, (3<<4)+14);
}
#define init_9_4_3_15
void FASTCALL call_9_4_3_15(SH2_struct* sh)
{
  SH2movwi(sh, 4, (3<<4)+15);
}
#define init_9_4_4_0
void FASTCALL call_9_4_4_0(SH2_struct* sh)
{
  SH2movwi(sh, 4, (4<<4)+0);
}
#define init_9_4_4_1
void FASTCALL call_9_4_4_1(SH2_struct* sh)
{
  SH2movwi(sh, 4, (4<<4)+1);
}
#define init_9_4_4_2
void FASTCALL call_9_4_4_2(SH2_struct* sh)
{
  SH2movwi(sh, 4, (4<<4)+2);
}
#define init_9_4_4_3
void FASTCALL call_9_4_4_3(SH2_struct* sh)
{
  SH2movwi(sh, 4, (4<<4)+3);
}
#define init_9_4_4_4
void FASTCALL call_9_4_4_4(SH2_struct* sh)
{
  SH2movwi(sh, 4, (4<<4)+4);
}
#define init_9_4_4_5
void FASTCALL call_9_4_4_5(SH2_struct* sh)
{
  SH2movwi(sh, 4, (4<<4)+5);
}
#define init_9_4_4_6
void FASTCALL call_9_4_4_6(SH2_struct* sh)
{
  SH2movwi(sh, 4, (4<<4)+6);
}
#define init_9_4_4_7
void FASTCALL call_9_4_4_7(SH2_struct* sh)
{
  SH2movwi(sh, 4, (4<<4)+7);
}
#define init_9_4_4_8
void FASTCALL call_9_4_4_8(SH2_struct* sh)
{
  SH2movwi(sh, 4, (4<<4)+8);
}
#define init_9_4_4_9
void FASTCALL call_9_4_4_9(SH2_struct* sh)
{
  SH2movwi(sh, 4, (4<<4)+9);
}
#define init_9_4_4_10
void FASTCALL call_9_4_4_10(SH2_struct* sh)
{
  SH2movwi(sh, 4, (4<<4)+10);
}
#define init_9_4_4_11
void FASTCALL call_9_4_4_11(SH2_struct* sh)
{
  SH2movwi(sh, 4, (4<<4)+11);
}
#define init_9_4_4_12
void FASTCALL call_9_4_4_12(SH2_struct* sh)
{
  SH2movwi(sh, 4, (4<<4)+12);
}
#define init_9_4_4_13
void FASTCALL call_9_4_4_13(SH2_struct* sh)
{
  SH2movwi(sh, 4, (4<<4)+13);
}
#define init_9_4_4_14
void FASTCALL call_9_4_4_14(SH2_struct* sh)
{
  SH2movwi(sh, 4, (4<<4)+14);
}
#define init_9_4_4_15
void FASTCALL call_9_4_4_15(SH2_struct* sh)
{
  SH2movwi(sh, 4, (4<<4)+15);
}
#define init_9_4_5_0
void FASTCALL call_9_4_5_0(SH2_struct* sh)
{
  SH2movwi(sh, 4, (5<<4)+0);
}
#define init_9_4_5_1
void FASTCALL call_9_4_5_1(SH2_struct* sh)
{
  SH2movwi(sh, 4, (5<<4)+1);
}
#define init_9_4_5_2
void FASTCALL call_9_4_5_2(SH2_struct* sh)
{
  SH2movwi(sh, 4, (5<<4)+2);
}
#define init_9_4_5_3
void FASTCALL call_9_4_5_3(SH2_struct* sh)
{
  SH2movwi(sh, 4, (5<<4)+3);
}
#define init_9_4_5_4
void FASTCALL call_9_4_5_4(SH2_struct* sh)
{
  SH2movwi(sh, 4, (5<<4)+4);
}
#define init_9_4_5_5
void FASTCALL call_9_4_5_5(SH2_struct* sh)
{
  SH2movwi(sh, 4, (5<<4)+5);
}
#define init_9_4_5_6
void FASTCALL call_9_4_5_6(SH2_struct* sh)
{
  SH2movwi(sh, 4, (5<<4)+6);
}
#define init_9_4_5_7
void FASTCALL call_9_4_5_7(SH2_struct* sh)
{
  SH2movwi(sh, 4, (5<<4)+7);
}
#define init_9_4_5_8
void FASTCALL call_9_4_5_8(SH2_struct* sh)
{
  SH2movwi(sh, 4, (5<<4)+8);
}
#define init_9_4_5_9
void FASTCALL call_9_4_5_9(SH2_struct* sh)
{
  SH2movwi(sh, 4, (5<<4)+9);
}
#define init_9_4_5_10
void FASTCALL call_9_4_5_10(SH2_struct* sh)
{
  SH2movwi(sh, 4, (5<<4)+10);
}
#define init_9_4_5_11
void FASTCALL call_9_4_5_11(SH2_struct* sh)
{
  SH2movwi(sh, 4, (5<<4)+11);
}
#define init_9_4_5_12
void FASTCALL call_9_4_5_12(SH2_struct* sh)
{
  SH2movwi(sh, 4, (5<<4)+12);
}
#define init_9_4_5_13
void FASTCALL call_9_4_5_13(SH2_struct* sh)
{
  SH2movwi(sh, 4, (5<<4)+13);
}
#define init_9_4_5_14
void FASTCALL call_9_4_5_14(SH2_struct* sh)
{
  SH2movwi(sh, 4, (5<<4)+14);
}
#define init_9_4_5_15
void FASTCALL call_9_4_5_15(SH2_struct* sh)
{
  SH2movwi(sh, 4, (5<<4)+15);
}
#define init_9_4_6_0
void FASTCALL call_9_4_6_0(SH2_struct* sh)
{
  SH2movwi(sh, 4, (6<<4)+0);
}
#define init_9_4_6_1
void FASTCALL call_9_4_6_1(SH2_struct* sh)
{
  SH2movwi(sh, 4, (6<<4)+1);
}
#define init_9_4_6_2
void FASTCALL call_9_4_6_2(SH2_struct* sh)
{
  SH2movwi(sh, 4, (6<<4)+2);
}
#define init_9_4_6_3
void FASTCALL call_9_4_6_3(SH2_struct* sh)
{
  SH2movwi(sh, 4, (6<<4)+3);
}
#define init_9_4_6_4
void FASTCALL call_9_4_6_4(SH2_struct* sh)
{
  SH2movwi(sh, 4, (6<<4)+4);
}
#define init_9_4_6_5
void FASTCALL call_9_4_6_5(SH2_struct* sh)
{
  SH2movwi(sh, 4, (6<<4)+5);
}
#define init_9_4_6_6
void FASTCALL call_9_4_6_6(SH2_struct* sh)
{
  SH2movwi(sh, 4, (6<<4)+6);
}
#define init_9_4_6_7
void FASTCALL call_9_4_6_7(SH2_struct* sh)
{
  SH2movwi(sh, 4, (6<<4)+7);
}
#define init_9_4_6_8
void FASTCALL call_9_4_6_8(SH2_struct* sh)
{
  SH2movwi(sh, 4, (6<<4)+8);
}
#define init_9_4_6_9
void FASTCALL call_9_4_6_9(SH2_struct* sh)
{
  SH2movwi(sh, 4, (6<<4)+9);
}
#define init_9_4_6_10
void FASTCALL call_9_4_6_10(SH2_struct* sh)
{
  SH2movwi(sh, 4, (6<<4)+10);
}
#define init_9_4_6_11
void FASTCALL call_9_4_6_11(SH2_struct* sh)
{
  SH2movwi(sh, 4, (6<<4)+11);
}
#define init_9_4_6_12
void FASTCALL call_9_4_6_12(SH2_struct* sh)
{
  SH2movwi(sh, 4, (6<<4)+12);
}
#define init_9_4_6_13
void FASTCALL call_9_4_6_13(SH2_struct* sh)
{
  SH2movwi(sh, 4, (6<<4)+13);
}
#define init_9_4_6_14
void FASTCALL call_9_4_6_14(SH2_struct* sh)
{
  SH2movwi(sh, 4, (6<<4)+14);
}
#define init_9_4_6_15
void FASTCALL call_9_4_6_15(SH2_struct* sh)
{
  SH2movwi(sh, 4, (6<<4)+15);
}
#define init_9_4_7_0
void FASTCALL call_9_4_7_0(SH2_struct* sh)
{
  SH2movwi(sh, 4, (7<<4)+0);
}
#define init_9_4_7_1
void FASTCALL call_9_4_7_1(SH2_struct* sh)
{
  SH2movwi(sh, 4, (7<<4)+1);
}
#define init_9_4_7_2
void FASTCALL call_9_4_7_2(SH2_struct* sh)
{
  SH2movwi(sh, 4, (7<<4)+2);
}
#define init_9_4_7_3
void FASTCALL call_9_4_7_3(SH2_struct* sh)
{
  SH2movwi(sh, 4, (7<<4)+3);
}
#define init_9_4_7_4
void FASTCALL call_9_4_7_4(SH2_struct* sh)
{
  SH2movwi(sh, 4, (7<<4)+4);
}
#define init_9_4_7_5
void FASTCALL call_9_4_7_5(SH2_struct* sh)
{
  SH2movwi(sh, 4, (7<<4)+5);
}
#define init_9_4_7_6
void FASTCALL call_9_4_7_6(SH2_struct* sh)
{
  SH2movwi(sh, 4, (7<<4)+6);
}
#define init_9_4_7_7
void FASTCALL call_9_4_7_7(SH2_struct* sh)
{
  SH2movwi(sh, 4, (7<<4)+7);
}
#define init_9_4_7_8
void FASTCALL call_9_4_7_8(SH2_struct* sh)
{
  SH2movwi(sh, 4, (7<<4)+8);
}
#define init_9_4_7_9
void FASTCALL call_9_4_7_9(SH2_struct* sh)
{
  SH2movwi(sh, 4, (7<<4)+9);
}
#define init_9_4_7_10
void FASTCALL call_9_4_7_10(SH2_struct* sh)
{
  SH2movwi(sh, 4, (7<<4)+10);
}
#define init_9_4_7_11
void FASTCALL call_9_4_7_11(SH2_struct* sh)
{
  SH2movwi(sh, 4, (7<<4)+11);
}
#define init_9_4_7_12
void FASTCALL call_9_4_7_12(SH2_struct* sh)
{
  SH2movwi(sh, 4, (7<<4)+12);
}
#define init_9_4_7_13
void FASTCALL call_9_4_7_13(SH2_struct* sh)
{
  SH2movwi(sh, 4, (7<<4)+13);
}
#define init_9_4_7_14
void FASTCALL call_9_4_7_14(SH2_struct* sh)
{
  SH2movwi(sh, 4, (7<<4)+14);
}
#define init_9_4_7_15
void FASTCALL call_9_4_7_15(SH2_struct* sh)
{
  SH2movwi(sh, 4, (7<<4)+15);
}
#define init_9_4_8_0
void FASTCALL call_9_4_8_0(SH2_struct* sh)
{
  SH2movwi(sh, 4, (8<<4)+0);
}
#define init_9_4_8_1
void FASTCALL call_9_4_8_1(SH2_struct* sh)
{
  SH2movwi(sh, 4, (8<<4)+1);
}
#define init_9_4_8_2
void FASTCALL call_9_4_8_2(SH2_struct* sh)
{
  SH2movwi(sh, 4, (8<<4)+2);
}
#define init_9_4_8_3
void FASTCALL call_9_4_8_3(SH2_struct* sh)
{
  SH2movwi(sh, 4, (8<<4)+3);
}
#define init_9_4_8_4
void FASTCALL call_9_4_8_4(SH2_struct* sh)
{
  SH2movwi(sh, 4, (8<<4)+4);
}
#define init_9_4_8_5
void FASTCALL call_9_4_8_5(SH2_struct* sh)
{
  SH2movwi(sh, 4, (8<<4)+5);
}
#define init_9_4_8_6
void FASTCALL call_9_4_8_6(SH2_struct* sh)
{
  SH2movwi(sh, 4, (8<<4)+6);
}
#define init_9_4_8_7
void FASTCALL call_9_4_8_7(SH2_struct* sh)
{
  SH2movwi(sh, 4, (8<<4)+7);
}
#define init_9_4_8_8
void FASTCALL call_9_4_8_8(SH2_struct* sh)
{
  SH2movwi(sh, 4, (8<<4)+8);
}
#define init_9_4_8_9
void FASTCALL call_9_4_8_9(SH2_struct* sh)
{
  SH2movwi(sh, 4, (8<<4)+9);
}
#define init_9_4_8_10
void FASTCALL call_9_4_8_10(SH2_struct* sh)
{
  SH2movwi(sh, 4, (8<<4)+10);
}
#define init_9_4_8_11
void FASTCALL call_9_4_8_11(SH2_struct* sh)
{
  SH2movwi(sh, 4, (8<<4)+11);
}
#define init_9_4_8_12
void FASTCALL call_9_4_8_12(SH2_struct* sh)
{
  SH2movwi(sh, 4, (8<<4)+12);
}
#define init_9_4_8_13
void FASTCALL call_9_4_8_13(SH2_struct* sh)
{
  SH2movwi(sh, 4, (8<<4)+13);
}
#define init_9_4_8_14
void FASTCALL call_9_4_8_14(SH2_struct* sh)
{
  SH2movwi(sh, 4, (8<<4)+14);
}
#define init_9_4_8_15
void FASTCALL call_9_4_8_15(SH2_struct* sh)
{
  SH2movwi(sh, 4, (8<<4)+15);
}
#define init_9_4_9_0
void FASTCALL call_9_4_9_0(SH2_struct* sh)
{
  SH2movwi(sh, 4, (9<<4)+0);
}
#define init_9_4_9_1
void FASTCALL call_9_4_9_1(SH2_struct* sh)
{
  SH2movwi(sh, 4, (9<<4)+1);
}
#define init_9_4_9_2
void FASTCALL call_9_4_9_2(SH2_struct* sh)
{
  SH2movwi(sh, 4, (9<<4)+2);
}
#define init_9_4_9_3
void FASTCALL call_9_4_9_3(SH2_struct* sh)
{
  SH2movwi(sh, 4, (9<<4)+3);
}
#define init_9_4_9_4
void FASTCALL call_9_4_9_4(SH2_struct* sh)
{
  SH2movwi(sh, 4, (9<<4)+4);
}
#define init_9_4_9_5
void FASTCALL call_9_4_9_5(SH2_struct* sh)
{
  SH2movwi(sh, 4, (9<<4)+5);
}
#define init_9_4_9_6
void FASTCALL call_9_4_9_6(SH2_struct* sh)
{
  SH2movwi(sh, 4, (9<<4)+6);
}
#define init_9_4_9_7
void FASTCALL call_9_4_9_7(SH2_struct* sh)
{
  SH2movwi(sh, 4, (9<<4)+7);
}
#define init_9_4_9_8
void FASTCALL call_9_4_9_8(SH2_struct* sh)
{
  SH2movwi(sh, 4, (9<<4)+8);
}
#define init_9_4_9_9
void FASTCALL call_9_4_9_9(SH2_struct* sh)
{
  SH2movwi(sh, 4, (9<<4)+9);
}
#define init_9_4_9_10
void FASTCALL call_9_4_9_10(SH2_struct* sh)
{
  SH2movwi(sh, 4, (9<<4)+10);
}
#define init_9_4_9_11
void FASTCALL call_9_4_9_11(SH2_struct* sh)
{
  SH2movwi(sh, 4, (9<<4)+11);
}
#define init_9_4_9_12
void FASTCALL call_9_4_9_12(SH2_struct* sh)
{
  SH2movwi(sh, 4, (9<<4)+12);
}
#define init_9_4_9_13
void FASTCALL call_9_4_9_13(SH2_struct* sh)
{
  SH2movwi(sh, 4, (9<<4)+13);
}
#define init_9_4_9_14
void FASTCALL call_9_4_9_14(SH2_struct* sh)
{
  SH2movwi(sh, 4, (9<<4)+14);
}
#define init_9_4_9_15
void FASTCALL call_9_4_9_15(SH2_struct* sh)
{
  SH2movwi(sh, 4, (9<<4)+15);
}
#define init_9_4_10_0
void FASTCALL call_9_4_10_0(SH2_struct* sh)
{
  SH2movwi(sh, 4, (10<<4)+0);
}
#define init_9_4_10_1
void FASTCALL call_9_4_10_1(SH2_struct* sh)
{
  SH2movwi(sh, 4, (10<<4)+1);
}
#define init_9_4_10_2
void FASTCALL call_9_4_10_2(SH2_struct* sh)
{
  SH2movwi(sh, 4, (10<<4)+2);
}
#define init_9_4_10_3
void FASTCALL call_9_4_10_3(SH2_struct* sh)
{
  SH2movwi(sh, 4, (10<<4)+3);
}
#define init_9_4_10_4
void FASTCALL call_9_4_10_4(SH2_struct* sh)
{
  SH2movwi(sh, 4, (10<<4)+4);
}
#define init_9_4_10_5
void FASTCALL call_9_4_10_5(SH2_struct* sh)
{
  SH2movwi(sh, 4, (10<<4)+5);
}
#define init_9_4_10_6
void FASTCALL call_9_4_10_6(SH2_struct* sh)
{
  SH2movwi(sh, 4, (10<<4)+6);
}
#define init_9_4_10_7
void FASTCALL call_9_4_10_7(SH2_struct* sh)
{
  SH2movwi(sh, 4, (10<<4)+7);
}
#define init_9_4_10_8
void FASTCALL call_9_4_10_8(SH2_struct* sh)
{
  SH2movwi(sh, 4, (10<<4)+8);
}
#define init_9_4_10_9
void FASTCALL call_9_4_10_9(SH2_struct* sh)
{
  SH2movwi(sh, 4, (10<<4)+9);
}
#define init_9_4_10_10
void FASTCALL call_9_4_10_10(SH2_struct* sh)
{
  SH2movwi(sh, 4, (10<<4)+10);
}
#define init_9_4_10_11
void FASTCALL call_9_4_10_11(SH2_struct* sh)
{
  SH2movwi(sh, 4, (10<<4)+11);
}
#define init_9_4_10_12
void FASTCALL call_9_4_10_12(SH2_struct* sh)
{
  SH2movwi(sh, 4, (10<<4)+12);
}
#define init_9_4_10_13
void FASTCALL call_9_4_10_13(SH2_struct* sh)
{
  SH2movwi(sh, 4, (10<<4)+13);
}
#define init_9_4_10_14
void FASTCALL call_9_4_10_14(SH2_struct* sh)
{
  SH2movwi(sh, 4, (10<<4)+14);
}
#define init_9_4_10_15
void FASTCALL call_9_4_10_15(SH2_struct* sh)
{
  SH2movwi(sh, 4, (10<<4)+15);
}
#define init_9_4_11_0
void FASTCALL call_9_4_11_0(SH2_struct* sh)
{
  SH2movwi(sh, 4, (11<<4)+0);
}
#define init_9_4_11_1
void FASTCALL call_9_4_11_1(SH2_struct* sh)
{
  SH2movwi(sh, 4, (11<<4)+1);
}
#define init_9_4_11_2
void FASTCALL call_9_4_11_2(SH2_struct* sh)
{
  SH2movwi(sh, 4, (11<<4)+2);
}
#define init_9_4_11_3
void FASTCALL call_9_4_11_3(SH2_struct* sh)
{
  SH2movwi(sh, 4, (11<<4)+3);
}
#define init_9_4_11_4
void FASTCALL call_9_4_11_4(SH2_struct* sh)
{
  SH2movwi(sh, 4, (11<<4)+4);
}
#define init_9_4_11_5
void FASTCALL call_9_4_11_5(SH2_struct* sh)
{
  SH2movwi(sh, 4, (11<<4)+5);
}
#define init_9_4_11_6
void FASTCALL call_9_4_11_6(SH2_struct* sh)
{
  SH2movwi(sh, 4, (11<<4)+6);
}
#define init_9_4_11_7
void FASTCALL call_9_4_11_7(SH2_struct* sh)
{
  SH2movwi(sh, 4, (11<<4)+7);
}
#define init_9_4_11_8
void FASTCALL call_9_4_11_8(SH2_struct* sh)
{
  SH2movwi(sh, 4, (11<<4)+8);
}
#define init_9_4_11_9
void FASTCALL call_9_4_11_9(SH2_struct* sh)
{
  SH2movwi(sh, 4, (11<<4)+9);
}
#define init_9_4_11_10
void FASTCALL call_9_4_11_10(SH2_struct* sh)
{
  SH2movwi(sh, 4, (11<<4)+10);
}
#define init_9_4_11_11
void FASTCALL call_9_4_11_11(SH2_struct* sh)
{
  SH2movwi(sh, 4, (11<<4)+11);
}
#define init_9_4_11_12
void FASTCALL call_9_4_11_12(SH2_struct* sh)
{
  SH2movwi(sh, 4, (11<<4)+12);
}
#define init_9_4_11_13
void FASTCALL call_9_4_11_13(SH2_struct* sh)
{
  SH2movwi(sh, 4, (11<<4)+13);
}
#define init_9_4_11_14
void FASTCALL call_9_4_11_14(SH2_struct* sh)
{
  SH2movwi(sh, 4, (11<<4)+14);
}
#define init_9_4_11_15
void FASTCALL call_9_4_11_15(SH2_struct* sh)
{
  SH2movwi(sh, 4, (11<<4)+15);
}
#define init_9_4_12_0
void FASTCALL call_9_4_12_0(SH2_struct* sh)
{
  SH2movwi(sh, 4, (12<<4)+0);
}
#define init_9_4_12_1
void FASTCALL call_9_4_12_1(SH2_struct* sh)
{
  SH2movwi(sh, 4, (12<<4)+1);
}
#define init_9_4_12_2
void FASTCALL call_9_4_12_2(SH2_struct* sh)
{
  SH2movwi(sh, 4, (12<<4)+2);
}
#define init_9_4_12_3
void FASTCALL call_9_4_12_3(SH2_struct* sh)
{
  SH2movwi(sh, 4, (12<<4)+3);
}
#define init_9_4_12_4
void FASTCALL call_9_4_12_4(SH2_struct* sh)
{
  SH2movwi(sh, 4, (12<<4)+4);
}
#define init_9_4_12_5
void FASTCALL call_9_4_12_5(SH2_struct* sh)
{
  SH2movwi(sh, 4, (12<<4)+5);
}
#define init_9_4_12_6
void FASTCALL call_9_4_12_6(SH2_struct* sh)
{
  SH2movwi(sh, 4, (12<<4)+6);
}
#define init_9_4_12_7
void FASTCALL call_9_4_12_7(SH2_struct* sh)
{
  SH2movwi(sh, 4, (12<<4)+7);
}
#define init_9_4_12_8
void FASTCALL call_9_4_12_8(SH2_struct* sh)
{
  SH2movwi(sh, 4, (12<<4)+8);
}
#define init_9_4_12_9
void FASTCALL call_9_4_12_9(SH2_struct* sh)
{
  SH2movwi(sh, 4, (12<<4)+9);
}
#define init_9_4_12_10
void FASTCALL call_9_4_12_10(SH2_struct* sh)
{
  SH2movwi(sh, 4, (12<<4)+10);
}
#define init_9_4_12_11
void FASTCALL call_9_4_12_11(SH2_struct* sh)
{
  SH2movwi(sh, 4, (12<<4)+11);
}
#define init_9_4_12_12
void FASTCALL call_9_4_12_12(SH2_struct* sh)
{
  SH2movwi(sh, 4, (12<<4)+12);
}
#define init_9_4_12_13
void FASTCALL call_9_4_12_13(SH2_struct* sh)
{
  SH2movwi(sh, 4, (12<<4)+13);
}
#define init_9_4_12_14
void FASTCALL call_9_4_12_14(SH2_struct* sh)
{
  SH2movwi(sh, 4, (12<<4)+14);
}
#define init_9_4_12_15
void FASTCALL call_9_4_12_15(SH2_struct* sh)
{
  SH2movwi(sh, 4, (12<<4)+15);
}
#define init_9_4_13_0
void FASTCALL call_9_4_13_0(SH2_struct* sh)
{
  SH2movwi(sh, 4, (13<<4)+0);
}
#define init_9_4_13_1
void FASTCALL call_9_4_13_1(SH2_struct* sh)
{
  SH2movwi(sh, 4, (13<<4)+1);
}
#define init_9_4_13_2
void FASTCALL call_9_4_13_2(SH2_struct* sh)
{
  SH2movwi(sh, 4, (13<<4)+2);
}
#define init_9_4_13_3
void FASTCALL call_9_4_13_3(SH2_struct* sh)
{
  SH2movwi(sh, 4, (13<<4)+3);
}
#define init_9_4_13_4
void FASTCALL call_9_4_13_4(SH2_struct* sh)
{
  SH2movwi(sh, 4, (13<<4)+4);
}
#define init_9_4_13_5
void FASTCALL call_9_4_13_5(SH2_struct* sh)
{
  SH2movwi(sh, 4, (13<<4)+5);
}
#define init_9_4_13_6
void FASTCALL call_9_4_13_6(SH2_struct* sh)
{
  SH2movwi(sh, 4, (13<<4)+6);
}
#define init_9_4_13_7
void FASTCALL call_9_4_13_7(SH2_struct* sh)
{
  SH2movwi(sh, 4, (13<<4)+7);
}
#define init_9_4_13_8
void FASTCALL call_9_4_13_8(SH2_struct* sh)
{
  SH2movwi(sh, 4, (13<<4)+8);
}
#define init_9_4_13_9
void FASTCALL call_9_4_13_9(SH2_struct* sh)
{
  SH2movwi(sh, 4, (13<<4)+9);
}
#define init_9_4_13_10
void FASTCALL call_9_4_13_10(SH2_struct* sh)
{
  SH2movwi(sh, 4, (13<<4)+10);
}
#define init_9_4_13_11
void FASTCALL call_9_4_13_11(SH2_struct* sh)
{
  SH2movwi(sh, 4, (13<<4)+11);
}
#define init_9_4_13_12
void FASTCALL call_9_4_13_12(SH2_struct* sh)
{
  SH2movwi(sh, 4, (13<<4)+12);
}
#define init_9_4_13_13
void FASTCALL call_9_4_13_13(SH2_struct* sh)
{
  SH2movwi(sh, 4, (13<<4)+13);
}
#define init_9_4_13_14
void FASTCALL call_9_4_13_14(SH2_struct* sh)
{
  SH2movwi(sh, 4, (13<<4)+14);
}
#define init_9_4_13_15
void FASTCALL call_9_4_13_15(SH2_struct* sh)
{
  SH2movwi(sh, 4, (13<<4)+15);
}
#define init_9_4_14_0
void FASTCALL call_9_4_14_0(SH2_struct* sh)
{
  SH2movwi(sh, 4, (14<<4)+0);
}
#define init_9_4_14_1
void FASTCALL call_9_4_14_1(SH2_struct* sh)
{
  SH2movwi(sh, 4, (14<<4)+1);
}
#define init_9_4_14_2
void FASTCALL call_9_4_14_2(SH2_struct* sh)
{
  SH2movwi(sh, 4, (14<<4)+2);
}
#define init_9_4_14_3
void FASTCALL call_9_4_14_3(SH2_struct* sh)
{
  SH2movwi(sh, 4, (14<<4)+3);
}
#define init_9_4_14_4
void FASTCALL call_9_4_14_4(SH2_struct* sh)
{
  SH2movwi(sh, 4, (14<<4)+4);
}
#define init_9_4_14_5
void FASTCALL call_9_4_14_5(SH2_struct* sh)
{
  SH2movwi(sh, 4, (14<<4)+5);
}
#define init_9_4_14_6
void FASTCALL call_9_4_14_6(SH2_struct* sh)
{
  SH2movwi(sh, 4, (14<<4)+6);
}
#define init_9_4_14_7
void FASTCALL call_9_4_14_7(SH2_struct* sh)
{
  SH2movwi(sh, 4, (14<<4)+7);
}
#define init_9_4_14_8
void FASTCALL call_9_4_14_8(SH2_struct* sh)
{
  SH2movwi(sh, 4, (14<<4)+8);
}
#define init_9_4_14_9
void FASTCALL call_9_4_14_9(SH2_struct* sh)
{
  SH2movwi(sh, 4, (14<<4)+9);
}
#define init_9_4_14_10
void FASTCALL call_9_4_14_10(SH2_struct* sh)
{
  SH2movwi(sh, 4, (14<<4)+10);
}
#define init_9_4_14_11
void FASTCALL call_9_4_14_11(SH2_struct* sh)
{
  SH2movwi(sh, 4, (14<<4)+11);
}
#define init_9_4_14_12
void FASTCALL call_9_4_14_12(SH2_struct* sh)
{
  SH2movwi(sh, 4, (14<<4)+12);
}
#define init_9_4_14_13
void FASTCALL call_9_4_14_13(SH2_struct* sh)
{
  SH2movwi(sh, 4, (14<<4)+13);
}
#define init_9_4_14_14
void FASTCALL call_9_4_14_14(SH2_struct* sh)
{
  SH2movwi(sh, 4, (14<<4)+14);
}
#define init_9_4_14_15
void FASTCALL call_9_4_14_15(SH2_struct* sh)
{
  SH2movwi(sh, 4, (14<<4)+15);
}
#define init_9_4_15_0
void FASTCALL call_9_4_15_0(SH2_struct* sh)
{
  SH2movwi(sh, 4, (15<<4)+0);
}
#define init_9_4_15_1
void FASTCALL call_9_4_15_1(SH2_struct* sh)
{
  SH2movwi(sh, 4, (15<<4)+1);
}
#define init_9_4_15_2
void FASTCALL call_9_4_15_2(SH2_struct* sh)
{
  SH2movwi(sh, 4, (15<<4)+2);
}
#define init_9_4_15_3
void FASTCALL call_9_4_15_3(SH2_struct* sh)
{
  SH2movwi(sh, 4, (15<<4)+3);
}
#define init_9_4_15_4
void FASTCALL call_9_4_15_4(SH2_struct* sh)
{
  SH2movwi(sh, 4, (15<<4)+4);
}
#define init_9_4_15_5
void FASTCALL call_9_4_15_5(SH2_struct* sh)
{
  SH2movwi(sh, 4, (15<<4)+5);
}
#define init_9_4_15_6
void FASTCALL call_9_4_15_6(SH2_struct* sh)
{
  SH2movwi(sh, 4, (15<<4)+6);
}
#define init_9_4_15_7
void FASTCALL call_9_4_15_7(SH2_struct* sh)
{
  SH2movwi(sh, 4, (15<<4)+7);
}
#define init_9_4_15_8
void FASTCALL call_9_4_15_8(SH2_struct* sh)
{
  SH2movwi(sh, 4, (15<<4)+8);
}
#define init_9_4_15_9
void FASTCALL call_9_4_15_9(SH2_struct* sh)
{
  SH2movwi(sh, 4, (15<<4)+9);
}
#define init_9_4_15_10
void FASTCALL call_9_4_15_10(SH2_struct* sh)
{
  SH2movwi(sh, 4, (15<<4)+10);
}
#define init_9_4_15_11
void FASTCALL call_9_4_15_11(SH2_struct* sh)
{
  SH2movwi(sh, 4, (15<<4)+11);
}
#define init_9_4_15_12
void FASTCALL call_9_4_15_12(SH2_struct* sh)
{
  SH2movwi(sh, 4, (15<<4)+12);
}
#define init_9_4_15_13
void FASTCALL call_9_4_15_13(SH2_struct* sh)
{
  SH2movwi(sh, 4, (15<<4)+13);
}
#define init_9_4_15_14
void FASTCALL call_9_4_15_14(SH2_struct* sh)
{
  SH2movwi(sh, 4, (15<<4)+14);
}
#define init_9_4_15_15
void FASTCALL call_9_4_15_15(SH2_struct* sh)
{
  SH2movwi(sh, 4, (15<<4)+15);
}
#define init_9_5_0_0
void FASTCALL call_9_5_0_0(SH2_struct* sh)
{
  SH2movwi(sh, 5, (0<<4)+0);
}
#define init_9_5_0_1
void FASTCALL call_9_5_0_1(SH2_struct* sh)
{
  SH2movwi(sh, 5, (0<<4)+1);
}
#define init_9_5_0_2
void FASTCALL call_9_5_0_2(SH2_struct* sh)
{
  SH2movwi(sh, 5, (0<<4)+2);
}
#define init_9_5_0_3
void FASTCALL call_9_5_0_3(SH2_struct* sh)
{
  SH2movwi(sh, 5, (0<<4)+3);
}
#define init_9_5_0_4
void FASTCALL call_9_5_0_4(SH2_struct* sh)
{
  SH2movwi(sh, 5, (0<<4)+4);
}
#define init_9_5_0_5
void FASTCALL call_9_5_0_5(SH2_struct* sh)
{
  SH2movwi(sh, 5, (0<<4)+5);
}
#define init_9_5_0_6
void FASTCALL call_9_5_0_6(SH2_struct* sh)
{
  SH2movwi(sh, 5, (0<<4)+6);
}
#define init_9_5_0_7
void FASTCALL call_9_5_0_7(SH2_struct* sh)
{
  SH2movwi(sh, 5, (0<<4)+7);
}
#define init_9_5_0_8
void FASTCALL call_9_5_0_8(SH2_struct* sh)
{
  SH2movwi(sh, 5, (0<<4)+8);
}
#define init_9_5_0_9
void FASTCALL call_9_5_0_9(SH2_struct* sh)
{
  SH2movwi(sh, 5, (0<<4)+9);
}
#define init_9_5_0_10
void FASTCALL call_9_5_0_10(SH2_struct* sh)
{
  SH2movwi(sh, 5, (0<<4)+10);
}
#define init_9_5_0_11
void FASTCALL call_9_5_0_11(SH2_struct* sh)
{
  SH2movwi(sh, 5, (0<<4)+11);
}
#define init_9_5_0_12
void FASTCALL call_9_5_0_12(SH2_struct* sh)
{
  SH2movwi(sh, 5, (0<<4)+12);
}
#define init_9_5_0_13
void FASTCALL call_9_5_0_13(SH2_struct* sh)
{
  SH2movwi(sh, 5, (0<<4)+13);
}
#define init_9_5_0_14
void FASTCALL call_9_5_0_14(SH2_struct* sh)
{
  SH2movwi(sh, 5, (0<<4)+14);
}
#define init_9_5_0_15
void FASTCALL call_9_5_0_15(SH2_struct* sh)
{
  SH2movwi(sh, 5, (0<<4)+15);
}
#define init_9_5_1_0
void FASTCALL call_9_5_1_0(SH2_struct* sh)
{
  SH2movwi(sh, 5, (1<<4)+0);
}
#define init_9_5_1_1
void FASTCALL call_9_5_1_1(SH2_struct* sh)
{
  SH2movwi(sh, 5, (1<<4)+1);
}
#define init_9_5_1_2
void FASTCALL call_9_5_1_2(SH2_struct* sh)
{
  SH2movwi(sh, 5, (1<<4)+2);
}
#define init_9_5_1_3
void FASTCALL call_9_5_1_3(SH2_struct* sh)
{
  SH2movwi(sh, 5, (1<<4)+3);
}
#define init_9_5_1_4
void FASTCALL call_9_5_1_4(SH2_struct* sh)
{
  SH2movwi(sh, 5, (1<<4)+4);
}
#define init_9_5_1_5
void FASTCALL call_9_5_1_5(SH2_struct* sh)
{
  SH2movwi(sh, 5, (1<<4)+5);
}
#define init_9_5_1_6
void FASTCALL call_9_5_1_6(SH2_struct* sh)
{
  SH2movwi(sh, 5, (1<<4)+6);
}
#define init_9_5_1_7
void FASTCALL call_9_5_1_7(SH2_struct* sh)
{
  SH2movwi(sh, 5, (1<<4)+7);
}
#define init_9_5_1_8
void FASTCALL call_9_5_1_8(SH2_struct* sh)
{
  SH2movwi(sh, 5, (1<<4)+8);
}
#define init_9_5_1_9
void FASTCALL call_9_5_1_9(SH2_struct* sh)
{
  SH2movwi(sh, 5, (1<<4)+9);
}
#define init_9_5_1_10
void FASTCALL call_9_5_1_10(SH2_struct* sh)
{
  SH2movwi(sh, 5, (1<<4)+10);
}
#define init_9_5_1_11
void FASTCALL call_9_5_1_11(SH2_struct* sh)
{
  SH2movwi(sh, 5, (1<<4)+11);
}
#define init_9_5_1_12
void FASTCALL call_9_5_1_12(SH2_struct* sh)
{
  SH2movwi(sh, 5, (1<<4)+12);
}
#define init_9_5_1_13
void FASTCALL call_9_5_1_13(SH2_struct* sh)
{
  SH2movwi(sh, 5, (1<<4)+13);
}
#define init_9_5_1_14
void FASTCALL call_9_5_1_14(SH2_struct* sh)
{
  SH2movwi(sh, 5, (1<<4)+14);
}
#define init_9_5_1_15
void FASTCALL call_9_5_1_15(SH2_struct* sh)
{
  SH2movwi(sh, 5, (1<<4)+15);
}
#define init_9_5_2_0
void FASTCALL call_9_5_2_0(SH2_struct* sh)
{
  SH2movwi(sh, 5, (2<<4)+0);
}
#define init_9_5_2_1
void FASTCALL call_9_5_2_1(SH2_struct* sh)
{
  SH2movwi(sh, 5, (2<<4)+1);
}
#define init_9_5_2_2
void FASTCALL call_9_5_2_2(SH2_struct* sh)
{
  SH2movwi(sh, 5, (2<<4)+2);
}
#define init_9_5_2_3
void FASTCALL call_9_5_2_3(SH2_struct* sh)
{
  SH2movwi(sh, 5, (2<<4)+3);
}
#define init_9_5_2_4
void FASTCALL call_9_5_2_4(SH2_struct* sh)
{
  SH2movwi(sh, 5, (2<<4)+4);
}
#define init_9_5_2_5
void FASTCALL call_9_5_2_5(SH2_struct* sh)
{
  SH2movwi(sh, 5, (2<<4)+5);
}
#define init_9_5_2_6
void FASTCALL call_9_5_2_6(SH2_struct* sh)
{
  SH2movwi(sh, 5, (2<<4)+6);
}
#define init_9_5_2_7
void FASTCALL call_9_5_2_7(SH2_struct* sh)
{
  SH2movwi(sh, 5, (2<<4)+7);
}
#define init_9_5_2_8
void FASTCALL call_9_5_2_8(SH2_struct* sh)
{
  SH2movwi(sh, 5, (2<<4)+8);
}
#define init_9_5_2_9
void FASTCALL call_9_5_2_9(SH2_struct* sh)
{
  SH2movwi(sh, 5, (2<<4)+9);
}
#define init_9_5_2_10
void FASTCALL call_9_5_2_10(SH2_struct* sh)
{
  SH2movwi(sh, 5, (2<<4)+10);
}
#define init_9_5_2_11
void FASTCALL call_9_5_2_11(SH2_struct* sh)
{
  SH2movwi(sh, 5, (2<<4)+11);
}
#define init_9_5_2_12
void FASTCALL call_9_5_2_12(SH2_struct* sh)
{
  SH2movwi(sh, 5, (2<<4)+12);
}
#define init_9_5_2_13
void FASTCALL call_9_5_2_13(SH2_struct* sh)
{
  SH2movwi(sh, 5, (2<<4)+13);
}
#define init_9_5_2_14
void FASTCALL call_9_5_2_14(SH2_struct* sh)
{
  SH2movwi(sh, 5, (2<<4)+14);
}
#define init_9_5_2_15
void FASTCALL call_9_5_2_15(SH2_struct* sh)
{
  SH2movwi(sh, 5, (2<<4)+15);
}
#define init_9_5_3_0
void FASTCALL call_9_5_3_0(SH2_struct* sh)
{
  SH2movwi(sh, 5, (3<<4)+0);
}
#define init_9_5_3_1
void FASTCALL call_9_5_3_1(SH2_struct* sh)
{
  SH2movwi(sh, 5, (3<<4)+1);
}
#define init_9_5_3_2
void FASTCALL call_9_5_3_2(SH2_struct* sh)
{
  SH2movwi(sh, 5, (3<<4)+2);
}
#define init_9_5_3_3
void FASTCALL call_9_5_3_3(SH2_struct* sh)
{
  SH2movwi(sh, 5, (3<<4)+3);
}
#define init_9_5_3_4
void FASTCALL call_9_5_3_4(SH2_struct* sh)
{
  SH2movwi(sh, 5, (3<<4)+4);
}
#define init_9_5_3_5
void FASTCALL call_9_5_3_5(SH2_struct* sh)
{
  SH2movwi(sh, 5, (3<<4)+5);
}
#define init_9_5_3_6
void FASTCALL call_9_5_3_6(SH2_struct* sh)
{
  SH2movwi(sh, 5, (3<<4)+6);
}
#define init_9_5_3_7
void FASTCALL call_9_5_3_7(SH2_struct* sh)
{
  SH2movwi(sh, 5, (3<<4)+7);
}
#define init_9_5_3_8
void FASTCALL call_9_5_3_8(SH2_struct* sh)
{
  SH2movwi(sh, 5, (3<<4)+8);
}
#define init_9_5_3_9
void FASTCALL call_9_5_3_9(SH2_struct* sh)
{
  SH2movwi(sh, 5, (3<<4)+9);
}
#define init_9_5_3_10
void FASTCALL call_9_5_3_10(SH2_struct* sh)
{
  SH2movwi(sh, 5, (3<<4)+10);
}
#define init_9_5_3_11
void FASTCALL call_9_5_3_11(SH2_struct* sh)
{
  SH2movwi(sh, 5, (3<<4)+11);
}
#define init_9_5_3_12
void FASTCALL call_9_5_3_12(SH2_struct* sh)
{
  SH2movwi(sh, 5, (3<<4)+12);
}
#define init_9_5_3_13
void FASTCALL call_9_5_3_13(SH2_struct* sh)
{
  SH2movwi(sh, 5, (3<<4)+13);
}
#define init_9_5_3_14
void FASTCALL call_9_5_3_14(SH2_struct* sh)
{
  SH2movwi(sh, 5, (3<<4)+14);
}
#define init_9_5_3_15
void FASTCALL call_9_5_3_15(SH2_struct* sh)
{
  SH2movwi(sh, 5, (3<<4)+15);
}
#define init_9_5_4_0
void FASTCALL call_9_5_4_0(SH2_struct* sh)
{
  SH2movwi(sh, 5, (4<<4)+0);
}
#define init_9_5_4_1
void FASTCALL call_9_5_4_1(SH2_struct* sh)
{
  SH2movwi(sh, 5, (4<<4)+1);
}
#define init_9_5_4_2
void FASTCALL call_9_5_4_2(SH2_struct* sh)
{
  SH2movwi(sh, 5, (4<<4)+2);
}
#define init_9_5_4_3
void FASTCALL call_9_5_4_3(SH2_struct* sh)
{
  SH2movwi(sh, 5, (4<<4)+3);
}
#define init_9_5_4_4
void FASTCALL call_9_5_4_4(SH2_struct* sh)
{
  SH2movwi(sh, 5, (4<<4)+4);
}
#define init_9_5_4_5
void FASTCALL call_9_5_4_5(SH2_struct* sh)
{
  SH2movwi(sh, 5, (4<<4)+5);
}
#define init_9_5_4_6
void FASTCALL call_9_5_4_6(SH2_struct* sh)
{
  SH2movwi(sh, 5, (4<<4)+6);
}
#define init_9_5_4_7
void FASTCALL call_9_5_4_7(SH2_struct* sh)
{
  SH2movwi(sh, 5, (4<<4)+7);
}
#define init_9_5_4_8
void FASTCALL call_9_5_4_8(SH2_struct* sh)
{
  SH2movwi(sh, 5, (4<<4)+8);
}
#define init_9_5_4_9
void FASTCALL call_9_5_4_9(SH2_struct* sh)
{
  SH2movwi(sh, 5, (4<<4)+9);
}
#define init_9_5_4_10
void FASTCALL call_9_5_4_10(SH2_struct* sh)
{
  SH2movwi(sh, 5, (4<<4)+10);
}
#define init_9_5_4_11
void FASTCALL call_9_5_4_11(SH2_struct* sh)
{
  SH2movwi(sh, 5, (4<<4)+11);
}
#define init_9_5_4_12
void FASTCALL call_9_5_4_12(SH2_struct* sh)
{
  SH2movwi(sh, 5, (4<<4)+12);
}
#define init_9_5_4_13
void FASTCALL call_9_5_4_13(SH2_struct* sh)
{
  SH2movwi(sh, 5, (4<<4)+13);
}
#define init_9_5_4_14
void FASTCALL call_9_5_4_14(SH2_struct* sh)
{
  SH2movwi(sh, 5, (4<<4)+14);
}
#define init_9_5_4_15
void FASTCALL call_9_5_4_15(SH2_struct* sh)
{
  SH2movwi(sh, 5, (4<<4)+15);
}
#define init_9_5_5_0
void FASTCALL call_9_5_5_0(SH2_struct* sh)
{
  SH2movwi(sh, 5, (5<<4)+0);
}
#define init_9_5_5_1
void FASTCALL call_9_5_5_1(SH2_struct* sh)
{
  SH2movwi(sh, 5, (5<<4)+1);
}
#define init_9_5_5_2
void FASTCALL call_9_5_5_2(SH2_struct* sh)
{
  SH2movwi(sh, 5, (5<<4)+2);
}
#define init_9_5_5_3
void FASTCALL call_9_5_5_3(SH2_struct* sh)
{
  SH2movwi(sh, 5, (5<<4)+3);
}
#define init_9_5_5_4
void FASTCALL call_9_5_5_4(SH2_struct* sh)
{
  SH2movwi(sh, 5, (5<<4)+4);
}
#define init_9_5_5_5
void FASTCALL call_9_5_5_5(SH2_struct* sh)
{
  SH2movwi(sh, 5, (5<<4)+5);
}
#define init_9_5_5_6
void FASTCALL call_9_5_5_6(SH2_struct* sh)
{
  SH2movwi(sh, 5, (5<<4)+6);
}
#define init_9_5_5_7
void FASTCALL call_9_5_5_7(SH2_struct* sh)
{
  SH2movwi(sh, 5, (5<<4)+7);
}
#define init_9_5_5_8
void FASTCALL call_9_5_5_8(SH2_struct* sh)
{
  SH2movwi(sh, 5, (5<<4)+8);
}
#define init_9_5_5_9
void FASTCALL call_9_5_5_9(SH2_struct* sh)
{
  SH2movwi(sh, 5, (5<<4)+9);
}
#define init_9_5_5_10
void FASTCALL call_9_5_5_10(SH2_struct* sh)
{
  SH2movwi(sh, 5, (5<<4)+10);
}
#define init_9_5_5_11
void FASTCALL call_9_5_5_11(SH2_struct* sh)
{
  SH2movwi(sh, 5, (5<<4)+11);
}
#define init_9_5_5_12
void FASTCALL call_9_5_5_12(SH2_struct* sh)
{
  SH2movwi(sh, 5, (5<<4)+12);
}
#define init_9_5_5_13
void FASTCALL call_9_5_5_13(SH2_struct* sh)
{
  SH2movwi(sh, 5, (5<<4)+13);
}
#define init_9_5_5_14
void FASTCALL call_9_5_5_14(SH2_struct* sh)
{
  SH2movwi(sh, 5, (5<<4)+14);
}
#define init_9_5_5_15
void FASTCALL call_9_5_5_15(SH2_struct* sh)
{
  SH2movwi(sh, 5, (5<<4)+15);
}
#define init_9_5_6_0
void FASTCALL call_9_5_6_0(SH2_struct* sh)
{
  SH2movwi(sh, 5, (6<<4)+0);
}
#define init_9_5_6_1
void FASTCALL call_9_5_6_1(SH2_struct* sh)
{
  SH2movwi(sh, 5, (6<<4)+1);
}
#define init_9_5_6_2
void FASTCALL call_9_5_6_2(SH2_struct* sh)
{
  SH2movwi(sh, 5, (6<<4)+2);
}
#define init_9_5_6_3
void FASTCALL call_9_5_6_3(SH2_struct* sh)
{
  SH2movwi(sh, 5, (6<<4)+3);
}
#define init_9_5_6_4
void FASTCALL call_9_5_6_4(SH2_struct* sh)
{
  SH2movwi(sh, 5, (6<<4)+4);
}
#define init_9_5_6_5
void FASTCALL call_9_5_6_5(SH2_struct* sh)
{
  SH2movwi(sh, 5, (6<<4)+5);
}
#define init_9_5_6_6
void FASTCALL call_9_5_6_6(SH2_struct* sh)
{
  SH2movwi(sh, 5, (6<<4)+6);
}
#define init_9_5_6_7
void FASTCALL call_9_5_6_7(SH2_struct* sh)
{
  SH2movwi(sh, 5, (6<<4)+7);
}
#define init_9_5_6_8
void FASTCALL call_9_5_6_8(SH2_struct* sh)
{
  SH2movwi(sh, 5, (6<<4)+8);
}
#define init_9_5_6_9
void FASTCALL call_9_5_6_9(SH2_struct* sh)
{
  SH2movwi(sh, 5, (6<<4)+9);
}
#define init_9_5_6_10
void FASTCALL call_9_5_6_10(SH2_struct* sh)
{
  SH2movwi(sh, 5, (6<<4)+10);
}
#define init_9_5_6_11
void FASTCALL call_9_5_6_11(SH2_struct* sh)
{
  SH2movwi(sh, 5, (6<<4)+11);
}
#define init_9_5_6_12
void FASTCALL call_9_5_6_12(SH2_struct* sh)
{
  SH2movwi(sh, 5, (6<<4)+12);
}
#define init_9_5_6_13
void FASTCALL call_9_5_6_13(SH2_struct* sh)
{
  SH2movwi(sh, 5, (6<<4)+13);
}
#define init_9_5_6_14
void FASTCALL call_9_5_6_14(SH2_struct* sh)
{
  SH2movwi(sh, 5, (6<<4)+14);
}
#define init_9_5_6_15
void FASTCALL call_9_5_6_15(SH2_struct* sh)
{
  SH2movwi(sh, 5, (6<<4)+15);
}
#define init_9_5_7_0
void FASTCALL call_9_5_7_0(SH2_struct* sh)
{
  SH2movwi(sh, 5, (7<<4)+0);
}
#define init_9_5_7_1
void FASTCALL call_9_5_7_1(SH2_struct* sh)
{
  SH2movwi(sh, 5, (7<<4)+1);
}
#define init_9_5_7_2
void FASTCALL call_9_5_7_2(SH2_struct* sh)
{
  SH2movwi(sh, 5, (7<<4)+2);
}
#define init_9_5_7_3
void FASTCALL call_9_5_7_3(SH2_struct* sh)
{
  SH2movwi(sh, 5, (7<<4)+3);
}
#define init_9_5_7_4
void FASTCALL call_9_5_7_4(SH2_struct* sh)
{
  SH2movwi(sh, 5, (7<<4)+4);
}
#define init_9_5_7_5
void FASTCALL call_9_5_7_5(SH2_struct* sh)
{
  SH2movwi(sh, 5, (7<<4)+5);
}
#define init_9_5_7_6
void FASTCALL call_9_5_7_6(SH2_struct* sh)
{
  SH2movwi(sh, 5, (7<<4)+6);
}
#define init_9_5_7_7
void FASTCALL call_9_5_7_7(SH2_struct* sh)
{
  SH2movwi(sh, 5, (7<<4)+7);
}
#define init_9_5_7_8
void FASTCALL call_9_5_7_8(SH2_struct* sh)
{
  SH2movwi(sh, 5, (7<<4)+8);
}
#define init_9_5_7_9
void FASTCALL call_9_5_7_9(SH2_struct* sh)
{
  SH2movwi(sh, 5, (7<<4)+9);
}
#define init_9_5_7_10
void FASTCALL call_9_5_7_10(SH2_struct* sh)
{
  SH2movwi(sh, 5, (7<<4)+10);
}
#define init_9_5_7_11
void FASTCALL call_9_5_7_11(SH2_struct* sh)
{
  SH2movwi(sh, 5, (7<<4)+11);
}
#define init_9_5_7_12
void FASTCALL call_9_5_7_12(SH2_struct* sh)
{
  SH2movwi(sh, 5, (7<<4)+12);
}
#define init_9_5_7_13
void FASTCALL call_9_5_7_13(SH2_struct* sh)
{
  SH2movwi(sh, 5, (7<<4)+13);
}
#define init_9_5_7_14
void FASTCALL call_9_5_7_14(SH2_struct* sh)
{
  SH2movwi(sh, 5, (7<<4)+14);
}
#define init_9_5_7_15
void FASTCALL call_9_5_7_15(SH2_struct* sh)
{
  SH2movwi(sh, 5, (7<<4)+15);
}
#define init_9_5_8_0
void FASTCALL call_9_5_8_0(SH2_struct* sh)
{
  SH2movwi(sh, 5, (8<<4)+0);
}
#define init_9_5_8_1
void FASTCALL call_9_5_8_1(SH2_struct* sh)
{
  SH2movwi(sh, 5, (8<<4)+1);
}
#define init_9_5_8_2
void FASTCALL call_9_5_8_2(SH2_struct* sh)
{
  SH2movwi(sh, 5, (8<<4)+2);
}
#define init_9_5_8_3
void FASTCALL call_9_5_8_3(SH2_struct* sh)
{
  SH2movwi(sh, 5, (8<<4)+3);
}
#define init_9_5_8_4
void FASTCALL call_9_5_8_4(SH2_struct* sh)
{
  SH2movwi(sh, 5, (8<<4)+4);
}
#define init_9_5_8_5
void FASTCALL call_9_5_8_5(SH2_struct* sh)
{
  SH2movwi(sh, 5, (8<<4)+5);
}
#define init_9_5_8_6
void FASTCALL call_9_5_8_6(SH2_struct* sh)
{
  SH2movwi(sh, 5, (8<<4)+6);
}
#define init_9_5_8_7
void FASTCALL call_9_5_8_7(SH2_struct* sh)
{
  SH2movwi(sh, 5, (8<<4)+7);
}
#define init_9_5_8_8
void FASTCALL call_9_5_8_8(SH2_struct* sh)
{
  SH2movwi(sh, 5, (8<<4)+8);
}
#define init_9_5_8_9
void FASTCALL call_9_5_8_9(SH2_struct* sh)
{
  SH2movwi(sh, 5, (8<<4)+9);
}
#define init_9_5_8_10
void FASTCALL call_9_5_8_10(SH2_struct* sh)
{
  SH2movwi(sh, 5, (8<<4)+10);
}
#define init_9_5_8_11
void FASTCALL call_9_5_8_11(SH2_struct* sh)
{
  SH2movwi(sh, 5, (8<<4)+11);
}
#define init_9_5_8_12
void FASTCALL call_9_5_8_12(SH2_struct* sh)
{
  SH2movwi(sh, 5, (8<<4)+12);
}
#define init_9_5_8_13
void FASTCALL call_9_5_8_13(SH2_struct* sh)
{
  SH2movwi(sh, 5, (8<<4)+13);
}
#define init_9_5_8_14
void FASTCALL call_9_5_8_14(SH2_struct* sh)
{
  SH2movwi(sh, 5, (8<<4)+14);
}
#define init_9_5_8_15
void FASTCALL call_9_5_8_15(SH2_struct* sh)
{
  SH2movwi(sh, 5, (8<<4)+15);
}
#define init_9_5_9_0
void FASTCALL call_9_5_9_0(SH2_struct* sh)
{
  SH2movwi(sh, 5, (9<<4)+0);
}
#define init_9_5_9_1
void FASTCALL call_9_5_9_1(SH2_struct* sh)
{
  SH2movwi(sh, 5, (9<<4)+1);
}
#define init_9_5_9_2
void FASTCALL call_9_5_9_2(SH2_struct* sh)
{
  SH2movwi(sh, 5, (9<<4)+2);
}
#define init_9_5_9_3
void FASTCALL call_9_5_9_3(SH2_struct* sh)
{
  SH2movwi(sh, 5, (9<<4)+3);
}
#define init_9_5_9_4
void FASTCALL call_9_5_9_4(SH2_struct* sh)
{
  SH2movwi(sh, 5, (9<<4)+4);
}
#define init_9_5_9_5
void FASTCALL call_9_5_9_5(SH2_struct* sh)
{
  SH2movwi(sh, 5, (9<<4)+5);
}
#define init_9_5_9_6
void FASTCALL call_9_5_9_6(SH2_struct* sh)
{
  SH2movwi(sh, 5, (9<<4)+6);
}
#define init_9_5_9_7
void FASTCALL call_9_5_9_7(SH2_struct* sh)
{
  SH2movwi(sh, 5, (9<<4)+7);
}
#define init_9_5_9_8
void FASTCALL call_9_5_9_8(SH2_struct* sh)
{
  SH2movwi(sh, 5, (9<<4)+8);
}
#define init_9_5_9_9
void FASTCALL call_9_5_9_9(SH2_struct* sh)
{
  SH2movwi(sh, 5, (9<<4)+9);
}
#define init_9_5_9_10
void FASTCALL call_9_5_9_10(SH2_struct* sh)
{
  SH2movwi(sh, 5, (9<<4)+10);
}
#define init_9_5_9_11
void FASTCALL call_9_5_9_11(SH2_struct* sh)
{
  SH2movwi(sh, 5, (9<<4)+11);
}
#define init_9_5_9_12
void FASTCALL call_9_5_9_12(SH2_struct* sh)
{
  SH2movwi(sh, 5, (9<<4)+12);
}
#define init_9_5_9_13
void FASTCALL call_9_5_9_13(SH2_struct* sh)
{
  SH2movwi(sh, 5, (9<<4)+13);
}
#define init_9_5_9_14
void FASTCALL call_9_5_9_14(SH2_struct* sh)
{
  SH2movwi(sh, 5, (9<<4)+14);
}
#define init_9_5_9_15
void FASTCALL call_9_5_9_15(SH2_struct* sh)
{
  SH2movwi(sh, 5, (9<<4)+15);
}
#define init_9_5_10_0
void FASTCALL call_9_5_10_0(SH2_struct* sh)
{
  SH2movwi(sh, 5, (10<<4)+0);
}
#define init_9_5_10_1
void FASTCALL call_9_5_10_1(SH2_struct* sh)
{
  SH2movwi(sh, 5, (10<<4)+1);
}
#define init_9_5_10_2
void FASTCALL call_9_5_10_2(SH2_struct* sh)
{
  SH2movwi(sh, 5, (10<<4)+2);
}
#define init_9_5_10_3
void FASTCALL call_9_5_10_3(SH2_struct* sh)
{
  SH2movwi(sh, 5, (10<<4)+3);
}
#define init_9_5_10_4
void FASTCALL call_9_5_10_4(SH2_struct* sh)
{
  SH2movwi(sh, 5, (10<<4)+4);
}
#define init_9_5_10_5
void FASTCALL call_9_5_10_5(SH2_struct* sh)
{
  SH2movwi(sh, 5, (10<<4)+5);
}
#define init_9_5_10_6
void FASTCALL call_9_5_10_6(SH2_struct* sh)
{
  SH2movwi(sh, 5, (10<<4)+6);
}
#define init_9_5_10_7
void FASTCALL call_9_5_10_7(SH2_struct* sh)
{
  SH2movwi(sh, 5, (10<<4)+7);
}
#define init_9_5_10_8
void FASTCALL call_9_5_10_8(SH2_struct* sh)
{
  SH2movwi(sh, 5, (10<<4)+8);
}
#define init_9_5_10_9
void FASTCALL call_9_5_10_9(SH2_struct* sh)
{
  SH2movwi(sh, 5, (10<<4)+9);
}
#define init_9_5_10_10
void FASTCALL call_9_5_10_10(SH2_struct* sh)
{
  SH2movwi(sh, 5, (10<<4)+10);
}
#define init_9_5_10_11
void FASTCALL call_9_5_10_11(SH2_struct* sh)
{
  SH2movwi(sh, 5, (10<<4)+11);
}
#define init_9_5_10_12
void FASTCALL call_9_5_10_12(SH2_struct* sh)
{
  SH2movwi(sh, 5, (10<<4)+12);
}
#define init_9_5_10_13
void FASTCALL call_9_5_10_13(SH2_struct* sh)
{
  SH2movwi(sh, 5, (10<<4)+13);
}
#define init_9_5_10_14
void FASTCALL call_9_5_10_14(SH2_struct* sh)
{
  SH2movwi(sh, 5, (10<<4)+14);
}
#define init_9_5_10_15
void FASTCALL call_9_5_10_15(SH2_struct* sh)
{
  SH2movwi(sh, 5, (10<<4)+15);
}
#define init_9_5_11_0
void FASTCALL call_9_5_11_0(SH2_struct* sh)
{
  SH2movwi(sh, 5, (11<<4)+0);
}
#define init_9_5_11_1
void FASTCALL call_9_5_11_1(SH2_struct* sh)
{
  SH2movwi(sh, 5, (11<<4)+1);
}
#define init_9_5_11_2
void FASTCALL call_9_5_11_2(SH2_struct* sh)
{
  SH2movwi(sh, 5, (11<<4)+2);
}
#define init_9_5_11_3
void FASTCALL call_9_5_11_3(SH2_struct* sh)
{
  SH2movwi(sh, 5, (11<<4)+3);
}
#define init_9_5_11_4
void FASTCALL call_9_5_11_4(SH2_struct* sh)
{
  SH2movwi(sh, 5, (11<<4)+4);
}
#define init_9_5_11_5
void FASTCALL call_9_5_11_5(SH2_struct* sh)
{
  SH2movwi(sh, 5, (11<<4)+5);
}
#define init_9_5_11_6
void FASTCALL call_9_5_11_6(SH2_struct* sh)
{
  SH2movwi(sh, 5, (11<<4)+6);
}
#define init_9_5_11_7
void FASTCALL call_9_5_11_7(SH2_struct* sh)
{
  SH2movwi(sh, 5, (11<<4)+7);
}
#define init_9_5_11_8
void FASTCALL call_9_5_11_8(SH2_struct* sh)
{
  SH2movwi(sh, 5, (11<<4)+8);
}
#define init_9_5_11_9
void FASTCALL call_9_5_11_9(SH2_struct* sh)
{
  SH2movwi(sh, 5, (11<<4)+9);
}
#define init_9_5_11_10
void FASTCALL call_9_5_11_10(SH2_struct* sh)
{
  SH2movwi(sh, 5, (11<<4)+10);
}
#define init_9_5_11_11
void FASTCALL call_9_5_11_11(SH2_struct* sh)
{
  SH2movwi(sh, 5, (11<<4)+11);
}
#define init_9_5_11_12
void FASTCALL call_9_5_11_12(SH2_struct* sh)
{
  SH2movwi(sh, 5, (11<<4)+12);
}
#define init_9_5_11_13
void FASTCALL call_9_5_11_13(SH2_struct* sh)
{
  SH2movwi(sh, 5, (11<<4)+13);
}
#define init_9_5_11_14
void FASTCALL call_9_5_11_14(SH2_struct* sh)
{
  SH2movwi(sh, 5, (11<<4)+14);
}
#define init_9_5_11_15
void FASTCALL call_9_5_11_15(SH2_struct* sh)
{
  SH2movwi(sh, 5, (11<<4)+15);
}
#define init_9_5_12_0
void FASTCALL call_9_5_12_0(SH2_struct* sh)
{
  SH2movwi(sh, 5, (12<<4)+0);
}
#define init_9_5_12_1
void FASTCALL call_9_5_12_1(SH2_struct* sh)
{
  SH2movwi(sh, 5, (12<<4)+1);
}
#define init_9_5_12_2
void FASTCALL call_9_5_12_2(SH2_struct* sh)
{
  SH2movwi(sh, 5, (12<<4)+2);
}
#define init_9_5_12_3
void FASTCALL call_9_5_12_3(SH2_struct* sh)
{
  SH2movwi(sh, 5, (12<<4)+3);
}
#define init_9_5_12_4
void FASTCALL call_9_5_12_4(SH2_struct* sh)
{
  SH2movwi(sh, 5, (12<<4)+4);
}
#define init_9_5_12_5
void FASTCALL call_9_5_12_5(SH2_struct* sh)
{
  SH2movwi(sh, 5, (12<<4)+5);
}
#define init_9_5_12_6
void FASTCALL call_9_5_12_6(SH2_struct* sh)
{
  SH2movwi(sh, 5, (12<<4)+6);
}
#define init_9_5_12_7
void FASTCALL call_9_5_12_7(SH2_struct* sh)
{
  SH2movwi(sh, 5, (12<<4)+7);
}
#define init_9_5_12_8
void FASTCALL call_9_5_12_8(SH2_struct* sh)
{
  SH2movwi(sh, 5, (12<<4)+8);
}
#define init_9_5_12_9
void FASTCALL call_9_5_12_9(SH2_struct* sh)
{
  SH2movwi(sh, 5, (12<<4)+9);
}
#define init_9_5_12_10
void FASTCALL call_9_5_12_10(SH2_struct* sh)
{
  SH2movwi(sh, 5, (12<<4)+10);
}
#define init_9_5_12_11
void FASTCALL call_9_5_12_11(SH2_struct* sh)
{
  SH2movwi(sh, 5, (12<<4)+11);
}
#define init_9_5_12_12
void FASTCALL call_9_5_12_12(SH2_struct* sh)
{
  SH2movwi(sh, 5, (12<<4)+12);
}
#define init_9_5_12_13
void FASTCALL call_9_5_12_13(SH2_struct* sh)
{
  SH2movwi(sh, 5, (12<<4)+13);
}
#define init_9_5_12_14
void FASTCALL call_9_5_12_14(SH2_struct* sh)
{
  SH2movwi(sh, 5, (12<<4)+14);
}
#define init_9_5_12_15
void FASTCALL call_9_5_12_15(SH2_struct* sh)
{
  SH2movwi(sh, 5, (12<<4)+15);
}
#define init_9_5_13_0
void FASTCALL call_9_5_13_0(SH2_struct* sh)
{
  SH2movwi(sh, 5, (13<<4)+0);
}
#define init_9_5_13_1
void FASTCALL call_9_5_13_1(SH2_struct* sh)
{
  SH2movwi(sh, 5, (13<<4)+1);
}
#define init_9_5_13_2
void FASTCALL call_9_5_13_2(SH2_struct* sh)
{
  SH2movwi(sh, 5, (13<<4)+2);
}
#define init_9_5_13_3
void FASTCALL call_9_5_13_3(SH2_struct* sh)
{
  SH2movwi(sh, 5, (13<<4)+3);
}
#define init_9_5_13_4
void FASTCALL call_9_5_13_4(SH2_struct* sh)
{
  SH2movwi(sh, 5, (13<<4)+4);
}
#define init_9_5_13_5
void FASTCALL call_9_5_13_5(SH2_struct* sh)
{
  SH2movwi(sh, 5, (13<<4)+5);
}
#define init_9_5_13_6
void FASTCALL call_9_5_13_6(SH2_struct* sh)
{
  SH2movwi(sh, 5, (13<<4)+6);
}
#define init_9_5_13_7
void FASTCALL call_9_5_13_7(SH2_struct* sh)
{
  SH2movwi(sh, 5, (13<<4)+7);
}
#define init_9_5_13_8
void FASTCALL call_9_5_13_8(SH2_struct* sh)
{
  SH2movwi(sh, 5, (13<<4)+8);
}
#define init_9_5_13_9
void FASTCALL call_9_5_13_9(SH2_struct* sh)
{
  SH2movwi(sh, 5, (13<<4)+9);
}
#define init_9_5_13_10
void FASTCALL call_9_5_13_10(SH2_struct* sh)
{
  SH2movwi(sh, 5, (13<<4)+10);
}
#define init_9_5_13_11
void FASTCALL call_9_5_13_11(SH2_struct* sh)
{
  SH2movwi(sh, 5, (13<<4)+11);
}
#define init_9_5_13_12
void FASTCALL call_9_5_13_12(SH2_struct* sh)
{
  SH2movwi(sh, 5, (13<<4)+12);
}
#define init_9_5_13_13
void FASTCALL call_9_5_13_13(SH2_struct* sh)
{
  SH2movwi(sh, 5, (13<<4)+13);
}
#define init_9_5_13_14
void FASTCALL call_9_5_13_14(SH2_struct* sh)
{
  SH2movwi(sh, 5, (13<<4)+14);
}
#define init_9_5_13_15
void FASTCALL call_9_5_13_15(SH2_struct* sh)
{
  SH2movwi(sh, 5, (13<<4)+15);
}
#define init_9_5_14_0
void FASTCALL call_9_5_14_0(SH2_struct* sh)
{
  SH2movwi(sh, 5, (14<<4)+0);
}
#define init_9_5_14_1
void FASTCALL call_9_5_14_1(SH2_struct* sh)
{
  SH2movwi(sh, 5, (14<<4)+1);
}
#define init_9_5_14_2
void FASTCALL call_9_5_14_2(SH2_struct* sh)
{
  SH2movwi(sh, 5, (14<<4)+2);
}
#define init_9_5_14_3
void FASTCALL call_9_5_14_3(SH2_struct* sh)
{
  SH2movwi(sh, 5, (14<<4)+3);
}
#define init_9_5_14_4
void FASTCALL call_9_5_14_4(SH2_struct* sh)
{
  SH2movwi(sh, 5, (14<<4)+4);
}
#define init_9_5_14_5
void FASTCALL call_9_5_14_5(SH2_struct* sh)
{
  SH2movwi(sh, 5, (14<<4)+5);
}
#define init_9_5_14_6
void FASTCALL call_9_5_14_6(SH2_struct* sh)
{
  SH2movwi(sh, 5, (14<<4)+6);
}
#define init_9_5_14_7
void FASTCALL call_9_5_14_7(SH2_struct* sh)
{
  SH2movwi(sh, 5, (14<<4)+7);
}
#define init_9_5_14_8
void FASTCALL call_9_5_14_8(SH2_struct* sh)
{
  SH2movwi(sh, 5, (14<<4)+8);
}
#define init_9_5_14_9
void FASTCALL call_9_5_14_9(SH2_struct* sh)
{
  SH2movwi(sh, 5, (14<<4)+9);
}
#define init_9_5_14_10
void FASTCALL call_9_5_14_10(SH2_struct* sh)
{
  SH2movwi(sh, 5, (14<<4)+10);
}
#define init_9_5_14_11
void FASTCALL call_9_5_14_11(SH2_struct* sh)
{
  SH2movwi(sh, 5, (14<<4)+11);
}
#define init_9_5_14_12
void FASTCALL call_9_5_14_12(SH2_struct* sh)
{
  SH2movwi(sh, 5, (14<<4)+12);
}
#define init_9_5_14_13
void FASTCALL call_9_5_14_13(SH2_struct* sh)
{
  SH2movwi(sh, 5, (14<<4)+13);
}
#define init_9_5_14_14
void FASTCALL call_9_5_14_14(SH2_struct* sh)
{
  SH2movwi(sh, 5, (14<<4)+14);
}
#define init_9_5_14_15
void FASTCALL call_9_5_14_15(SH2_struct* sh)
{
  SH2movwi(sh, 5, (14<<4)+15);
}
#define init_9_5_15_0
void FASTCALL call_9_5_15_0(SH2_struct* sh)
{
  SH2movwi(sh, 5, (15<<4)+0);
}
#define init_9_5_15_1
void FASTCALL call_9_5_15_1(SH2_struct* sh)
{
  SH2movwi(sh, 5, (15<<4)+1);
}
#define init_9_5_15_2
void FASTCALL call_9_5_15_2(SH2_struct* sh)
{
  SH2movwi(sh, 5, (15<<4)+2);
}
#define init_9_5_15_3
void FASTCALL call_9_5_15_3(SH2_struct* sh)
{
  SH2movwi(sh, 5, (15<<4)+3);
}
#define init_9_5_15_4
void FASTCALL call_9_5_15_4(SH2_struct* sh)
{
  SH2movwi(sh, 5, (15<<4)+4);
}
#define init_9_5_15_5
void FASTCALL call_9_5_15_5(SH2_struct* sh)
{
  SH2movwi(sh, 5, (15<<4)+5);
}
#define init_9_5_15_6
void FASTCALL call_9_5_15_6(SH2_struct* sh)
{
  SH2movwi(sh, 5, (15<<4)+6);
}
#define init_9_5_15_7
void FASTCALL call_9_5_15_7(SH2_struct* sh)
{
  SH2movwi(sh, 5, (15<<4)+7);
}
#define init_9_5_15_8
void FASTCALL call_9_5_15_8(SH2_struct* sh)
{
  SH2movwi(sh, 5, (15<<4)+8);
}
#define init_9_5_15_9
void FASTCALL call_9_5_15_9(SH2_struct* sh)
{
  SH2movwi(sh, 5, (15<<4)+9);
}
#define init_9_5_15_10
void FASTCALL call_9_5_15_10(SH2_struct* sh)
{
  SH2movwi(sh, 5, (15<<4)+10);
}
#define init_9_5_15_11
void FASTCALL call_9_5_15_11(SH2_struct* sh)
{
  SH2movwi(sh, 5, (15<<4)+11);
}
#define init_9_5_15_12
void FASTCALL call_9_5_15_12(SH2_struct* sh)
{
  SH2movwi(sh, 5, (15<<4)+12);
}
#define init_9_5_15_13
void FASTCALL call_9_5_15_13(SH2_struct* sh)
{
  SH2movwi(sh, 5, (15<<4)+13);
}
#define init_9_5_15_14
void FASTCALL call_9_5_15_14(SH2_struct* sh)
{
  SH2movwi(sh, 5, (15<<4)+14);
}
#define init_9_5_15_15
void FASTCALL call_9_5_15_15(SH2_struct* sh)
{
  SH2movwi(sh, 5, (15<<4)+15);
}
#define init_9_6_0_0
void FASTCALL call_9_6_0_0(SH2_struct* sh)
{
  SH2movwi(sh, 6, (0<<4)+0);
}
#define init_9_6_0_1
void FASTCALL call_9_6_0_1(SH2_struct* sh)
{
  SH2movwi(sh, 6, (0<<4)+1);
}
#define init_9_6_0_2
void FASTCALL call_9_6_0_2(SH2_struct* sh)
{
  SH2movwi(sh, 6, (0<<4)+2);
}
#define init_9_6_0_3
void FASTCALL call_9_6_0_3(SH2_struct* sh)
{
  SH2movwi(sh, 6, (0<<4)+3);
}
#define init_9_6_0_4
void FASTCALL call_9_6_0_4(SH2_struct* sh)
{
  SH2movwi(sh, 6, (0<<4)+4);
}
#define init_9_6_0_5
void FASTCALL call_9_6_0_5(SH2_struct* sh)
{
  SH2movwi(sh, 6, (0<<4)+5);
}
#define init_9_6_0_6
void FASTCALL call_9_6_0_6(SH2_struct* sh)
{
  SH2movwi(sh, 6, (0<<4)+6);
}
#define init_9_6_0_7
void FASTCALL call_9_6_0_7(SH2_struct* sh)
{
  SH2movwi(sh, 6, (0<<4)+7);
}
#define init_9_6_0_8
void FASTCALL call_9_6_0_8(SH2_struct* sh)
{
  SH2movwi(sh, 6, (0<<4)+8);
}
#define init_9_6_0_9
void FASTCALL call_9_6_0_9(SH2_struct* sh)
{
  SH2movwi(sh, 6, (0<<4)+9);
}
#define init_9_6_0_10
void FASTCALL call_9_6_0_10(SH2_struct* sh)
{
  SH2movwi(sh, 6, (0<<4)+10);
}
#define init_9_6_0_11
void FASTCALL call_9_6_0_11(SH2_struct* sh)
{
  SH2movwi(sh, 6, (0<<4)+11);
}
#define init_9_6_0_12
void FASTCALL call_9_6_0_12(SH2_struct* sh)
{
  SH2movwi(sh, 6, (0<<4)+12);
}
#define init_9_6_0_13
void FASTCALL call_9_6_0_13(SH2_struct* sh)
{
  SH2movwi(sh, 6, (0<<4)+13);
}
#define init_9_6_0_14
void FASTCALL call_9_6_0_14(SH2_struct* sh)
{
  SH2movwi(sh, 6, (0<<4)+14);
}
#define init_9_6_0_15
void FASTCALL call_9_6_0_15(SH2_struct* sh)
{
  SH2movwi(sh, 6, (0<<4)+15);
}
#define init_9_6_1_0
void FASTCALL call_9_6_1_0(SH2_struct* sh)
{
  SH2movwi(sh, 6, (1<<4)+0);
}
#define init_9_6_1_1
void FASTCALL call_9_6_1_1(SH2_struct* sh)
{
  SH2movwi(sh, 6, (1<<4)+1);
}
#define init_9_6_1_2
void FASTCALL call_9_6_1_2(SH2_struct* sh)
{
  SH2movwi(sh, 6, (1<<4)+2);
}
#define init_9_6_1_3
void FASTCALL call_9_6_1_3(SH2_struct* sh)
{
  SH2movwi(sh, 6, (1<<4)+3);
}
#define init_9_6_1_4
void FASTCALL call_9_6_1_4(SH2_struct* sh)
{
  SH2movwi(sh, 6, (1<<4)+4);
}
#define init_9_6_1_5
void FASTCALL call_9_6_1_5(SH2_struct* sh)
{
  SH2movwi(sh, 6, (1<<4)+5);
}
#define init_9_6_1_6
void FASTCALL call_9_6_1_6(SH2_struct* sh)
{
  SH2movwi(sh, 6, (1<<4)+6);
}
#define init_9_6_1_7
void FASTCALL call_9_6_1_7(SH2_struct* sh)
{
  SH2movwi(sh, 6, (1<<4)+7);
}
#define init_9_6_1_8
void FASTCALL call_9_6_1_8(SH2_struct* sh)
{
  SH2movwi(sh, 6, (1<<4)+8);
}
#define init_9_6_1_9
void FASTCALL call_9_6_1_9(SH2_struct* sh)
{
  SH2movwi(sh, 6, (1<<4)+9);
}
#define init_9_6_1_10
void FASTCALL call_9_6_1_10(SH2_struct* sh)
{
  SH2movwi(sh, 6, (1<<4)+10);
}
#define init_9_6_1_11
void FASTCALL call_9_6_1_11(SH2_struct* sh)
{
  SH2movwi(sh, 6, (1<<4)+11);
}
#define init_9_6_1_12
void FASTCALL call_9_6_1_12(SH2_struct* sh)
{
  SH2movwi(sh, 6, (1<<4)+12);
}
#define init_9_6_1_13
void FASTCALL call_9_6_1_13(SH2_struct* sh)
{
  SH2movwi(sh, 6, (1<<4)+13);
}
#define init_9_6_1_14
void FASTCALL call_9_6_1_14(SH2_struct* sh)
{
  SH2movwi(sh, 6, (1<<4)+14);
}
#define init_9_6_1_15
void FASTCALL call_9_6_1_15(SH2_struct* sh)
{
  SH2movwi(sh, 6, (1<<4)+15);
}
#define init_9_6_2_0
void FASTCALL call_9_6_2_0(SH2_struct* sh)
{
  SH2movwi(sh, 6, (2<<4)+0);
}
#define init_9_6_2_1
void FASTCALL call_9_6_2_1(SH2_struct* sh)
{
  SH2movwi(sh, 6, (2<<4)+1);
}
#define init_9_6_2_2
void FASTCALL call_9_6_2_2(SH2_struct* sh)
{
  SH2movwi(sh, 6, (2<<4)+2);
}
#define init_9_6_2_3
void FASTCALL call_9_6_2_3(SH2_struct* sh)
{
  SH2movwi(sh, 6, (2<<4)+3);
}
#define init_9_6_2_4
void FASTCALL call_9_6_2_4(SH2_struct* sh)
{
  SH2movwi(sh, 6, (2<<4)+4);
}
#define init_9_6_2_5
void FASTCALL call_9_6_2_5(SH2_struct* sh)
{
  SH2movwi(sh, 6, (2<<4)+5);
}
#define init_9_6_2_6
void FASTCALL call_9_6_2_6(SH2_struct* sh)
{
  SH2movwi(sh, 6, (2<<4)+6);
}
#define init_9_6_2_7
void FASTCALL call_9_6_2_7(SH2_struct* sh)
{
  SH2movwi(sh, 6, (2<<4)+7);
}
#define init_9_6_2_8
void FASTCALL call_9_6_2_8(SH2_struct* sh)
{
  SH2movwi(sh, 6, (2<<4)+8);
}
#define init_9_6_2_9
void FASTCALL call_9_6_2_9(SH2_struct* sh)
{
  SH2movwi(sh, 6, (2<<4)+9);
}
#define init_9_6_2_10
void FASTCALL call_9_6_2_10(SH2_struct* sh)
{
  SH2movwi(sh, 6, (2<<4)+10);
}
#define init_9_6_2_11
void FASTCALL call_9_6_2_11(SH2_struct* sh)
{
  SH2movwi(sh, 6, (2<<4)+11);
}
#define init_9_6_2_12
void FASTCALL call_9_6_2_12(SH2_struct* sh)
{
  SH2movwi(sh, 6, (2<<4)+12);
}
#define init_9_6_2_13
void FASTCALL call_9_6_2_13(SH2_struct* sh)
{
  SH2movwi(sh, 6, (2<<4)+13);
}
#define init_9_6_2_14
void FASTCALL call_9_6_2_14(SH2_struct* sh)
{
  SH2movwi(sh, 6, (2<<4)+14);
}
#define init_9_6_2_15
void FASTCALL call_9_6_2_15(SH2_struct* sh)
{
  SH2movwi(sh, 6, (2<<4)+15);
}
#define init_9_6_3_0
void FASTCALL call_9_6_3_0(SH2_struct* sh)
{
  SH2movwi(sh, 6, (3<<4)+0);
}
#define init_9_6_3_1
void FASTCALL call_9_6_3_1(SH2_struct* sh)
{
  SH2movwi(sh, 6, (3<<4)+1);
}
#define init_9_6_3_2
void FASTCALL call_9_6_3_2(SH2_struct* sh)
{
  SH2movwi(sh, 6, (3<<4)+2);
}
#define init_9_6_3_3
void FASTCALL call_9_6_3_3(SH2_struct* sh)
{
  SH2movwi(sh, 6, (3<<4)+3);
}
#define init_9_6_3_4
void FASTCALL call_9_6_3_4(SH2_struct* sh)
{
  SH2movwi(sh, 6, (3<<4)+4);
}
#define init_9_6_3_5
void FASTCALL call_9_6_3_5(SH2_struct* sh)
{
  SH2movwi(sh, 6, (3<<4)+5);
}
#define init_9_6_3_6
void FASTCALL call_9_6_3_6(SH2_struct* sh)
{
  SH2movwi(sh, 6, (3<<4)+6);
}
#define init_9_6_3_7
void FASTCALL call_9_6_3_7(SH2_struct* sh)
{
  SH2movwi(sh, 6, (3<<4)+7);
}
#define init_9_6_3_8
void FASTCALL call_9_6_3_8(SH2_struct* sh)
{
  SH2movwi(sh, 6, (3<<4)+8);
}
#define init_9_6_3_9
void FASTCALL call_9_6_3_9(SH2_struct* sh)
{
  SH2movwi(sh, 6, (3<<4)+9);
}
#define init_9_6_3_10
void FASTCALL call_9_6_3_10(SH2_struct* sh)
{
  SH2movwi(sh, 6, (3<<4)+10);
}
#define init_9_6_3_11
void FASTCALL call_9_6_3_11(SH2_struct* sh)
{
  SH2movwi(sh, 6, (3<<4)+11);
}
#define init_9_6_3_12
void FASTCALL call_9_6_3_12(SH2_struct* sh)
{
  SH2movwi(sh, 6, (3<<4)+12);
}
#define init_9_6_3_13
void FASTCALL call_9_6_3_13(SH2_struct* sh)
{
  SH2movwi(sh, 6, (3<<4)+13);
}
#define init_9_6_3_14
void FASTCALL call_9_6_3_14(SH2_struct* sh)
{
  SH2movwi(sh, 6, (3<<4)+14);
}
#define init_9_6_3_15
void FASTCALL call_9_6_3_15(SH2_struct* sh)
{
  SH2movwi(sh, 6, (3<<4)+15);
}
#define init_9_6_4_0
void FASTCALL call_9_6_4_0(SH2_struct* sh)
{
  SH2movwi(sh, 6, (4<<4)+0);
}
#define init_9_6_4_1
void FASTCALL call_9_6_4_1(SH2_struct* sh)
{
  SH2movwi(sh, 6, (4<<4)+1);
}
#define init_9_6_4_2
void FASTCALL call_9_6_4_2(SH2_struct* sh)
{
  SH2movwi(sh, 6, (4<<4)+2);
}
#define init_9_6_4_3
void FASTCALL call_9_6_4_3(SH2_struct* sh)
{
  SH2movwi(sh, 6, (4<<4)+3);
}
#define init_9_6_4_4
void FASTCALL call_9_6_4_4(SH2_struct* sh)
{
  SH2movwi(sh, 6, (4<<4)+4);
}
#define init_9_6_4_5
void FASTCALL call_9_6_4_5(SH2_struct* sh)
{
  SH2movwi(sh, 6, (4<<4)+5);
}
#define init_9_6_4_6
void FASTCALL call_9_6_4_6(SH2_struct* sh)
{
  SH2movwi(sh, 6, (4<<4)+6);
}
#define init_9_6_4_7
void FASTCALL call_9_6_4_7(SH2_struct* sh)
{
  SH2movwi(sh, 6, (4<<4)+7);
}
#define init_9_6_4_8
void FASTCALL call_9_6_4_8(SH2_struct* sh)
{
  SH2movwi(sh, 6, (4<<4)+8);
}
#define init_9_6_4_9
void FASTCALL call_9_6_4_9(SH2_struct* sh)
{
  SH2movwi(sh, 6, (4<<4)+9);
}
#define init_9_6_4_10
void FASTCALL call_9_6_4_10(SH2_struct* sh)
{
  SH2movwi(sh, 6, (4<<4)+10);
}
#define init_9_6_4_11
void FASTCALL call_9_6_4_11(SH2_struct* sh)
{
  SH2movwi(sh, 6, (4<<4)+11);
}
#define init_9_6_4_12
void FASTCALL call_9_6_4_12(SH2_struct* sh)
{
  SH2movwi(sh, 6, (4<<4)+12);
}
#define init_9_6_4_13
void FASTCALL call_9_6_4_13(SH2_struct* sh)
{
  SH2movwi(sh, 6, (4<<4)+13);
}
#define init_9_6_4_14
void FASTCALL call_9_6_4_14(SH2_struct* sh)
{
  SH2movwi(sh, 6, (4<<4)+14);
}
#define init_9_6_4_15
void FASTCALL call_9_6_4_15(SH2_struct* sh)
{
  SH2movwi(sh, 6, (4<<4)+15);
}
#define init_9_6_5_0
void FASTCALL call_9_6_5_0(SH2_struct* sh)
{
  SH2movwi(sh, 6, (5<<4)+0);
}
#define init_9_6_5_1
void FASTCALL call_9_6_5_1(SH2_struct* sh)
{
  SH2movwi(sh, 6, (5<<4)+1);
}
#define init_9_6_5_2
void FASTCALL call_9_6_5_2(SH2_struct* sh)
{
  SH2movwi(sh, 6, (5<<4)+2);
}
#define init_9_6_5_3
void FASTCALL call_9_6_5_3(SH2_struct* sh)
{
  SH2movwi(sh, 6, (5<<4)+3);
}
#define init_9_6_5_4
void FASTCALL call_9_6_5_4(SH2_struct* sh)
{
  SH2movwi(sh, 6, (5<<4)+4);
}
#define init_9_6_5_5
void FASTCALL call_9_6_5_5(SH2_struct* sh)
{
  SH2movwi(sh, 6, (5<<4)+5);
}
#define init_9_6_5_6
void FASTCALL call_9_6_5_6(SH2_struct* sh)
{
  SH2movwi(sh, 6, (5<<4)+6);
}
#define init_9_6_5_7
void FASTCALL call_9_6_5_7(SH2_struct* sh)
{
  SH2movwi(sh, 6, (5<<4)+7);
}
#define init_9_6_5_8
void FASTCALL call_9_6_5_8(SH2_struct* sh)
{
  SH2movwi(sh, 6, (5<<4)+8);
}
#define init_9_6_5_9
void FASTCALL call_9_6_5_9(SH2_struct* sh)
{
  SH2movwi(sh, 6, (5<<4)+9);
}
#define init_9_6_5_10
void FASTCALL call_9_6_5_10(SH2_struct* sh)
{
  SH2movwi(sh, 6, (5<<4)+10);
}
#define init_9_6_5_11
void FASTCALL call_9_6_5_11(SH2_struct* sh)
{
  SH2movwi(sh, 6, (5<<4)+11);
}
#define init_9_6_5_12
void FASTCALL call_9_6_5_12(SH2_struct* sh)
{
  SH2movwi(sh, 6, (5<<4)+12);
}
#define init_9_6_5_13
void FASTCALL call_9_6_5_13(SH2_struct* sh)
{
  SH2movwi(sh, 6, (5<<4)+13);
}
#define init_9_6_5_14
void FASTCALL call_9_6_5_14(SH2_struct* sh)
{
  SH2movwi(sh, 6, (5<<4)+14);
}
#define init_9_6_5_15
void FASTCALL call_9_6_5_15(SH2_struct* sh)
{
  SH2movwi(sh, 6, (5<<4)+15);
}
#define init_9_6_6_0
void FASTCALL call_9_6_6_0(SH2_struct* sh)
{
  SH2movwi(sh, 6, (6<<4)+0);
}
#define init_9_6_6_1
void FASTCALL call_9_6_6_1(SH2_struct* sh)
{
  SH2movwi(sh, 6, (6<<4)+1);
}
#define init_9_6_6_2
void FASTCALL call_9_6_6_2(SH2_struct* sh)
{
  SH2movwi(sh, 6, (6<<4)+2);
}
#define init_9_6_6_3
void FASTCALL call_9_6_6_3(SH2_struct* sh)
{
  SH2movwi(sh, 6, (6<<4)+3);
}
#define init_9_6_6_4
void FASTCALL call_9_6_6_4(SH2_struct* sh)
{
  SH2movwi(sh, 6, (6<<4)+4);
}
#define init_9_6_6_5
void FASTCALL call_9_6_6_5(SH2_struct* sh)
{
  SH2movwi(sh, 6, (6<<4)+5);
}
#define init_9_6_6_6
void FASTCALL call_9_6_6_6(SH2_struct* sh)
{
  SH2movwi(sh, 6, (6<<4)+6);
}
#define init_9_6_6_7
void FASTCALL call_9_6_6_7(SH2_struct* sh)
{
  SH2movwi(sh, 6, (6<<4)+7);
}
#define init_9_6_6_8
void FASTCALL call_9_6_6_8(SH2_struct* sh)
{
  SH2movwi(sh, 6, (6<<4)+8);
}
#define init_9_6_6_9
void FASTCALL call_9_6_6_9(SH2_struct* sh)
{
  SH2movwi(sh, 6, (6<<4)+9);
}
#define init_9_6_6_10
void FASTCALL call_9_6_6_10(SH2_struct* sh)
{
  SH2movwi(sh, 6, (6<<4)+10);
}
#define init_9_6_6_11
void FASTCALL call_9_6_6_11(SH2_struct* sh)
{
  SH2movwi(sh, 6, (6<<4)+11);
}
#define init_9_6_6_12
void FASTCALL call_9_6_6_12(SH2_struct* sh)
{
  SH2movwi(sh, 6, (6<<4)+12);
}
#define init_9_6_6_13
void FASTCALL call_9_6_6_13(SH2_struct* sh)
{
  SH2movwi(sh, 6, (6<<4)+13);
}
#define init_9_6_6_14
void FASTCALL call_9_6_6_14(SH2_struct* sh)
{
  SH2movwi(sh, 6, (6<<4)+14);
}
#define init_9_6_6_15
void FASTCALL call_9_6_6_15(SH2_struct* sh)
{
  SH2movwi(sh, 6, (6<<4)+15);
}
#define init_9_6_7_0
void FASTCALL call_9_6_7_0(SH2_struct* sh)
{
  SH2movwi(sh, 6, (7<<4)+0);
}
#define init_9_6_7_1
void FASTCALL call_9_6_7_1(SH2_struct* sh)
{
  SH2movwi(sh, 6, (7<<4)+1);
}
#define init_9_6_7_2
void FASTCALL call_9_6_7_2(SH2_struct* sh)
{
  SH2movwi(sh, 6, (7<<4)+2);
}
#define init_9_6_7_3
void FASTCALL call_9_6_7_3(SH2_struct* sh)
{
  SH2movwi(sh, 6, (7<<4)+3);
}
#define init_9_6_7_4
void FASTCALL call_9_6_7_4(SH2_struct* sh)
{
  SH2movwi(sh, 6, (7<<4)+4);
}
#define init_9_6_7_5
void FASTCALL call_9_6_7_5(SH2_struct* sh)
{
  SH2movwi(sh, 6, (7<<4)+5);
}
#define init_9_6_7_6
void FASTCALL call_9_6_7_6(SH2_struct* sh)
{
  SH2movwi(sh, 6, (7<<4)+6);
}
#define init_9_6_7_7
void FASTCALL call_9_6_7_7(SH2_struct* sh)
{
  SH2movwi(sh, 6, (7<<4)+7);
}
#define init_9_6_7_8
void FASTCALL call_9_6_7_8(SH2_struct* sh)
{
  SH2movwi(sh, 6, (7<<4)+8);
}
#define init_9_6_7_9
void FASTCALL call_9_6_7_9(SH2_struct* sh)
{
  SH2movwi(sh, 6, (7<<4)+9);
}
#define init_9_6_7_10
void FASTCALL call_9_6_7_10(SH2_struct* sh)
{
  SH2movwi(sh, 6, (7<<4)+10);
}
#define init_9_6_7_11
void FASTCALL call_9_6_7_11(SH2_struct* sh)
{
  SH2movwi(sh, 6, (7<<4)+11);
}
#define init_9_6_7_12
void FASTCALL call_9_6_7_12(SH2_struct* sh)
{
  SH2movwi(sh, 6, (7<<4)+12);
}
#define init_9_6_7_13
void FASTCALL call_9_6_7_13(SH2_struct* sh)
{
  SH2movwi(sh, 6, (7<<4)+13);
}
#define init_9_6_7_14
void FASTCALL call_9_6_7_14(SH2_struct* sh)
{
  SH2movwi(sh, 6, (7<<4)+14);
}
#define init_9_6_7_15
void FASTCALL call_9_6_7_15(SH2_struct* sh)
{
  SH2movwi(sh, 6, (7<<4)+15);
}
#define init_9_6_8_0
void FASTCALL call_9_6_8_0(SH2_struct* sh)
{
  SH2movwi(sh, 6, (8<<4)+0);
}
#define init_9_6_8_1
void FASTCALL call_9_6_8_1(SH2_struct* sh)
{
  SH2movwi(sh, 6, (8<<4)+1);
}
#define init_9_6_8_2
void FASTCALL call_9_6_8_2(SH2_struct* sh)
{
  SH2movwi(sh, 6, (8<<4)+2);
}
#define init_9_6_8_3
void FASTCALL call_9_6_8_3(SH2_struct* sh)
{
  SH2movwi(sh, 6, (8<<4)+3);
}
#define init_9_6_8_4
void FASTCALL call_9_6_8_4(SH2_struct* sh)
{
  SH2movwi(sh, 6, (8<<4)+4);
}
#define init_9_6_8_5
void FASTCALL call_9_6_8_5(SH2_struct* sh)
{
  SH2movwi(sh, 6, (8<<4)+5);
}
#define init_9_6_8_6
void FASTCALL call_9_6_8_6(SH2_struct* sh)
{
  SH2movwi(sh, 6, (8<<4)+6);
}
#define init_9_6_8_7
void FASTCALL call_9_6_8_7(SH2_struct* sh)
{
  SH2movwi(sh, 6, (8<<4)+7);
}
#define init_9_6_8_8
void FASTCALL call_9_6_8_8(SH2_struct* sh)
{
  SH2movwi(sh, 6, (8<<4)+8);
}
#define init_9_6_8_9
void FASTCALL call_9_6_8_9(SH2_struct* sh)
{
  SH2movwi(sh, 6, (8<<4)+9);
}
#define init_9_6_8_10
void FASTCALL call_9_6_8_10(SH2_struct* sh)
{
  SH2movwi(sh, 6, (8<<4)+10);
}
#define init_9_6_8_11
void FASTCALL call_9_6_8_11(SH2_struct* sh)
{
  SH2movwi(sh, 6, (8<<4)+11);
}
#define init_9_6_8_12
void FASTCALL call_9_6_8_12(SH2_struct* sh)
{
  SH2movwi(sh, 6, (8<<4)+12);
}
#define init_9_6_8_13
void FASTCALL call_9_6_8_13(SH2_struct* sh)
{
  SH2movwi(sh, 6, (8<<4)+13);
}
#define init_9_6_8_14
void FASTCALL call_9_6_8_14(SH2_struct* sh)
{
  SH2movwi(sh, 6, (8<<4)+14);
}
#define init_9_6_8_15
void FASTCALL call_9_6_8_15(SH2_struct* sh)
{
  SH2movwi(sh, 6, (8<<4)+15);
}
#define init_9_6_9_0
void FASTCALL call_9_6_9_0(SH2_struct* sh)
{
  SH2movwi(sh, 6, (9<<4)+0);
}
#define init_9_6_9_1
void FASTCALL call_9_6_9_1(SH2_struct* sh)
{
  SH2movwi(sh, 6, (9<<4)+1);
}
#define init_9_6_9_2
void FASTCALL call_9_6_9_2(SH2_struct* sh)
{
  SH2movwi(sh, 6, (9<<4)+2);
}
#define init_9_6_9_3
void FASTCALL call_9_6_9_3(SH2_struct* sh)
{
  SH2movwi(sh, 6, (9<<4)+3);
}
#define init_9_6_9_4
void FASTCALL call_9_6_9_4(SH2_struct* sh)
{
  SH2movwi(sh, 6, (9<<4)+4);
}
#define init_9_6_9_5
void FASTCALL call_9_6_9_5(SH2_struct* sh)
{
  SH2movwi(sh, 6, (9<<4)+5);
}
#define init_9_6_9_6
void FASTCALL call_9_6_9_6(SH2_struct* sh)
{
  SH2movwi(sh, 6, (9<<4)+6);
}
#define init_9_6_9_7
void FASTCALL call_9_6_9_7(SH2_struct* sh)
{
  SH2movwi(sh, 6, (9<<4)+7);
}
#define init_9_6_9_8
void FASTCALL call_9_6_9_8(SH2_struct* sh)
{
  SH2movwi(sh, 6, (9<<4)+8);
}
#define init_9_6_9_9
void FASTCALL call_9_6_9_9(SH2_struct* sh)
{
  SH2movwi(sh, 6, (9<<4)+9);
}
#define init_9_6_9_10
void FASTCALL call_9_6_9_10(SH2_struct* sh)
{
  SH2movwi(sh, 6, (9<<4)+10);
}
#define init_9_6_9_11
void FASTCALL call_9_6_9_11(SH2_struct* sh)
{
  SH2movwi(sh, 6, (9<<4)+11);
}
#define init_9_6_9_12
void FASTCALL call_9_6_9_12(SH2_struct* sh)
{
  SH2movwi(sh, 6, (9<<4)+12);
}
#define init_9_6_9_13
void FASTCALL call_9_6_9_13(SH2_struct* sh)
{
  SH2movwi(sh, 6, (9<<4)+13);
}
#define init_9_6_9_14
void FASTCALL call_9_6_9_14(SH2_struct* sh)
{
  SH2movwi(sh, 6, (9<<4)+14);
}
#define init_9_6_9_15
void FASTCALL call_9_6_9_15(SH2_struct* sh)
{
  SH2movwi(sh, 6, (9<<4)+15);
}
#define init_9_6_10_0
void FASTCALL call_9_6_10_0(SH2_struct* sh)
{
  SH2movwi(sh, 6, (10<<4)+0);
}
#define init_9_6_10_1
void FASTCALL call_9_6_10_1(SH2_struct* sh)
{
  SH2movwi(sh, 6, (10<<4)+1);
}
#define init_9_6_10_2
void FASTCALL call_9_6_10_2(SH2_struct* sh)
{
  SH2movwi(sh, 6, (10<<4)+2);
}
#define init_9_6_10_3
void FASTCALL call_9_6_10_3(SH2_struct* sh)
{
  SH2movwi(sh, 6, (10<<4)+3);
}
#define init_9_6_10_4
void FASTCALL call_9_6_10_4(SH2_struct* sh)
{
  SH2movwi(sh, 6, (10<<4)+4);
}
#define init_9_6_10_5
void FASTCALL call_9_6_10_5(SH2_struct* sh)
{
  SH2movwi(sh, 6, (10<<4)+5);
}
#define init_9_6_10_6
void FASTCALL call_9_6_10_6(SH2_struct* sh)
{
  SH2movwi(sh, 6, (10<<4)+6);
}
#define init_9_6_10_7
void FASTCALL call_9_6_10_7(SH2_struct* sh)
{
  SH2movwi(sh, 6, (10<<4)+7);
}
#define init_9_6_10_8
void FASTCALL call_9_6_10_8(SH2_struct* sh)
{
  SH2movwi(sh, 6, (10<<4)+8);
}
#define init_9_6_10_9
void FASTCALL call_9_6_10_9(SH2_struct* sh)
{
  SH2movwi(sh, 6, (10<<4)+9);
}
#define init_9_6_10_10
void FASTCALL call_9_6_10_10(SH2_struct* sh)
{
  SH2movwi(sh, 6, (10<<4)+10);
}
#define init_9_6_10_11
void FASTCALL call_9_6_10_11(SH2_struct* sh)
{
  SH2movwi(sh, 6, (10<<4)+11);
}
#define init_9_6_10_12
void FASTCALL call_9_6_10_12(SH2_struct* sh)
{
  SH2movwi(sh, 6, (10<<4)+12);
}
#define init_9_6_10_13
void FASTCALL call_9_6_10_13(SH2_struct* sh)
{
  SH2movwi(sh, 6, (10<<4)+13);
}
#define init_9_6_10_14
void FASTCALL call_9_6_10_14(SH2_struct* sh)
{
  SH2movwi(sh, 6, (10<<4)+14);
}
#define init_9_6_10_15
void FASTCALL call_9_6_10_15(SH2_struct* sh)
{
  SH2movwi(sh, 6, (10<<4)+15);
}
#define init_9_6_11_0
void FASTCALL call_9_6_11_0(SH2_struct* sh)
{
  SH2movwi(sh, 6, (11<<4)+0);
}
#define init_9_6_11_1
void FASTCALL call_9_6_11_1(SH2_struct* sh)
{
  SH2movwi(sh, 6, (11<<4)+1);
}
#define init_9_6_11_2
void FASTCALL call_9_6_11_2(SH2_struct* sh)
{
  SH2movwi(sh, 6, (11<<4)+2);
}
#define init_9_6_11_3
void FASTCALL call_9_6_11_3(SH2_struct* sh)
{
  SH2movwi(sh, 6, (11<<4)+3);
}
#define init_9_6_11_4
void FASTCALL call_9_6_11_4(SH2_struct* sh)
{
  SH2movwi(sh, 6, (11<<4)+4);
}
#define init_9_6_11_5
void FASTCALL call_9_6_11_5(SH2_struct* sh)
{
  SH2movwi(sh, 6, (11<<4)+5);
}
#define init_9_6_11_6
void FASTCALL call_9_6_11_6(SH2_struct* sh)
{
  SH2movwi(sh, 6, (11<<4)+6);
}
#define init_9_6_11_7
void FASTCALL call_9_6_11_7(SH2_struct* sh)
{
  SH2movwi(sh, 6, (11<<4)+7);
}
#define init_9_6_11_8
void FASTCALL call_9_6_11_8(SH2_struct* sh)
{
  SH2movwi(sh, 6, (11<<4)+8);
}
#define init_9_6_11_9
void FASTCALL call_9_6_11_9(SH2_struct* sh)
{
  SH2movwi(sh, 6, (11<<4)+9);
}
#define init_9_6_11_10
void FASTCALL call_9_6_11_10(SH2_struct* sh)
{
  SH2movwi(sh, 6, (11<<4)+10);
}
#define init_9_6_11_11
void FASTCALL call_9_6_11_11(SH2_struct* sh)
{
  SH2movwi(sh, 6, (11<<4)+11);
}
#define init_9_6_11_12
void FASTCALL call_9_6_11_12(SH2_struct* sh)
{
  SH2movwi(sh, 6, (11<<4)+12);
}
#define init_9_6_11_13
void FASTCALL call_9_6_11_13(SH2_struct* sh)
{
  SH2movwi(sh, 6, (11<<4)+13);
}
#define init_9_6_11_14
void FASTCALL call_9_6_11_14(SH2_struct* sh)
{
  SH2movwi(sh, 6, (11<<4)+14);
}
#define init_9_6_11_15
void FASTCALL call_9_6_11_15(SH2_struct* sh)
{
  SH2movwi(sh, 6, (11<<4)+15);
}
#define init_9_6_12_0
void FASTCALL call_9_6_12_0(SH2_struct* sh)
{
  SH2movwi(sh, 6, (12<<4)+0);
}
#define init_9_6_12_1
void FASTCALL call_9_6_12_1(SH2_struct* sh)
{
  SH2movwi(sh, 6, (12<<4)+1);
}
#define init_9_6_12_2
void FASTCALL call_9_6_12_2(SH2_struct* sh)
{
  SH2movwi(sh, 6, (12<<4)+2);
}
#define init_9_6_12_3
void FASTCALL call_9_6_12_3(SH2_struct* sh)
{
  SH2movwi(sh, 6, (12<<4)+3);
}
#define init_9_6_12_4
void FASTCALL call_9_6_12_4(SH2_struct* sh)
{
  SH2movwi(sh, 6, (12<<4)+4);
}
#define init_9_6_12_5
void FASTCALL call_9_6_12_5(SH2_struct* sh)
{
  SH2movwi(sh, 6, (12<<4)+5);
}
#define init_9_6_12_6
void FASTCALL call_9_6_12_6(SH2_struct* sh)
{
  SH2movwi(sh, 6, (12<<4)+6);
}
#define init_9_6_12_7
void FASTCALL call_9_6_12_7(SH2_struct* sh)
{
  SH2movwi(sh, 6, (12<<4)+7);
}
#define init_9_6_12_8
void FASTCALL call_9_6_12_8(SH2_struct* sh)
{
  SH2movwi(sh, 6, (12<<4)+8);
}
#define init_9_6_12_9
void FASTCALL call_9_6_12_9(SH2_struct* sh)
{
  SH2movwi(sh, 6, (12<<4)+9);
}
#define init_9_6_12_10
void FASTCALL call_9_6_12_10(SH2_struct* sh)
{
  SH2movwi(sh, 6, (12<<4)+10);
}
#define init_9_6_12_11
void FASTCALL call_9_6_12_11(SH2_struct* sh)
{
  SH2movwi(sh, 6, (12<<4)+11);
}
#define init_9_6_12_12
void FASTCALL call_9_6_12_12(SH2_struct* sh)
{
  SH2movwi(sh, 6, (12<<4)+12);
}
#define init_9_6_12_13
void FASTCALL call_9_6_12_13(SH2_struct* sh)
{
  SH2movwi(sh, 6, (12<<4)+13);
}
#define init_9_6_12_14
void FASTCALL call_9_6_12_14(SH2_struct* sh)
{
  SH2movwi(sh, 6, (12<<4)+14);
}
#define init_9_6_12_15
void FASTCALL call_9_6_12_15(SH2_struct* sh)
{
  SH2movwi(sh, 6, (12<<4)+15);
}
#define init_9_6_13_0
void FASTCALL call_9_6_13_0(SH2_struct* sh)
{
  SH2movwi(sh, 6, (13<<4)+0);
}
#define init_9_6_13_1
void FASTCALL call_9_6_13_1(SH2_struct* sh)
{
  SH2movwi(sh, 6, (13<<4)+1);
}
#define init_9_6_13_2
void FASTCALL call_9_6_13_2(SH2_struct* sh)
{
  SH2movwi(sh, 6, (13<<4)+2);
}
#define init_9_6_13_3
void FASTCALL call_9_6_13_3(SH2_struct* sh)
{
  SH2movwi(sh, 6, (13<<4)+3);
}
#define init_9_6_13_4
void FASTCALL call_9_6_13_4(SH2_struct* sh)
{
  SH2movwi(sh, 6, (13<<4)+4);
}
#define init_9_6_13_5
void FASTCALL call_9_6_13_5(SH2_struct* sh)
{
  SH2movwi(sh, 6, (13<<4)+5);
}
#define init_9_6_13_6
void FASTCALL call_9_6_13_6(SH2_struct* sh)
{
  SH2movwi(sh, 6, (13<<4)+6);
}
#define init_9_6_13_7
void FASTCALL call_9_6_13_7(SH2_struct* sh)
{
  SH2movwi(sh, 6, (13<<4)+7);
}
#define init_9_6_13_8
void FASTCALL call_9_6_13_8(SH2_struct* sh)
{
  SH2movwi(sh, 6, (13<<4)+8);
}
#define init_9_6_13_9
void FASTCALL call_9_6_13_9(SH2_struct* sh)
{
  SH2movwi(sh, 6, (13<<4)+9);
}
#define init_9_6_13_10
void FASTCALL call_9_6_13_10(SH2_struct* sh)
{
  SH2movwi(sh, 6, (13<<4)+10);
}
#define init_9_6_13_11
void FASTCALL call_9_6_13_11(SH2_struct* sh)
{
  SH2movwi(sh, 6, (13<<4)+11);
}
#define init_9_6_13_12
void FASTCALL call_9_6_13_12(SH2_struct* sh)
{
  SH2movwi(sh, 6, (13<<4)+12);
}
#define init_9_6_13_13
void FASTCALL call_9_6_13_13(SH2_struct* sh)
{
  SH2movwi(sh, 6, (13<<4)+13);
}
#define init_9_6_13_14
void FASTCALL call_9_6_13_14(SH2_struct* sh)
{
  SH2movwi(sh, 6, (13<<4)+14);
}
#define init_9_6_13_15
void FASTCALL call_9_6_13_15(SH2_struct* sh)
{
  SH2movwi(sh, 6, (13<<4)+15);
}
#define init_9_6_14_0
void FASTCALL call_9_6_14_0(SH2_struct* sh)
{
  SH2movwi(sh, 6, (14<<4)+0);
}
#define init_9_6_14_1
void FASTCALL call_9_6_14_1(SH2_struct* sh)
{
  SH2movwi(sh, 6, (14<<4)+1);
}
#define init_9_6_14_2
void FASTCALL call_9_6_14_2(SH2_struct* sh)
{
  SH2movwi(sh, 6, (14<<4)+2);
}
#define init_9_6_14_3
void FASTCALL call_9_6_14_3(SH2_struct* sh)
{
  SH2movwi(sh, 6, (14<<4)+3);
}
#define init_9_6_14_4
void FASTCALL call_9_6_14_4(SH2_struct* sh)
{
  SH2movwi(sh, 6, (14<<4)+4);
}
#define init_9_6_14_5
void FASTCALL call_9_6_14_5(SH2_struct* sh)
{
  SH2movwi(sh, 6, (14<<4)+5);
}
#define init_9_6_14_6
void FASTCALL call_9_6_14_6(SH2_struct* sh)
{
  SH2movwi(sh, 6, (14<<4)+6);
}
#define init_9_6_14_7
void FASTCALL call_9_6_14_7(SH2_struct* sh)
{
  SH2movwi(sh, 6, (14<<4)+7);
}
#define init_9_6_14_8
void FASTCALL call_9_6_14_8(SH2_struct* sh)
{
  SH2movwi(sh, 6, (14<<4)+8);
}
#define init_9_6_14_9
void FASTCALL call_9_6_14_9(SH2_struct* sh)
{
  SH2movwi(sh, 6, (14<<4)+9);
}
#define init_9_6_14_10
void FASTCALL call_9_6_14_10(SH2_struct* sh)
{
  SH2movwi(sh, 6, (14<<4)+10);
}
#define init_9_6_14_11
void FASTCALL call_9_6_14_11(SH2_struct* sh)
{
  SH2movwi(sh, 6, (14<<4)+11);
}
#define init_9_6_14_12
void FASTCALL call_9_6_14_12(SH2_struct* sh)
{
  SH2movwi(sh, 6, (14<<4)+12);
}
#define init_9_6_14_13
void FASTCALL call_9_6_14_13(SH2_struct* sh)
{
  SH2movwi(sh, 6, (14<<4)+13);
}
#define init_9_6_14_14
void FASTCALL call_9_6_14_14(SH2_struct* sh)
{
  SH2movwi(sh, 6, (14<<4)+14);
}
#define init_9_6_14_15
void FASTCALL call_9_6_14_15(SH2_struct* sh)
{
  SH2movwi(sh, 6, (14<<4)+15);
}
#define init_9_6_15_0
void FASTCALL call_9_6_15_0(SH2_struct* sh)
{
  SH2movwi(sh, 6, (15<<4)+0);
}
#define init_9_6_15_1
void FASTCALL call_9_6_15_1(SH2_struct* sh)
{
  SH2movwi(sh, 6, (15<<4)+1);
}
#define init_9_6_15_2
void FASTCALL call_9_6_15_2(SH2_struct* sh)
{
  SH2movwi(sh, 6, (15<<4)+2);
}
#define init_9_6_15_3
void FASTCALL call_9_6_15_3(SH2_struct* sh)
{
  SH2movwi(sh, 6, (15<<4)+3);
}
#define init_9_6_15_4
void FASTCALL call_9_6_15_4(SH2_struct* sh)
{
  SH2movwi(sh, 6, (15<<4)+4);
}
#define init_9_6_15_5
void FASTCALL call_9_6_15_5(SH2_struct* sh)
{
  SH2movwi(sh, 6, (15<<4)+5);
}
#define init_9_6_15_6
void FASTCALL call_9_6_15_6(SH2_struct* sh)
{
  SH2movwi(sh, 6, (15<<4)+6);
}
#define init_9_6_15_7
void FASTCALL call_9_6_15_7(SH2_struct* sh)
{
  SH2movwi(sh, 6, (15<<4)+7);
}
#define init_9_6_15_8
void FASTCALL call_9_6_15_8(SH2_struct* sh)
{
  SH2movwi(sh, 6, (15<<4)+8);
}
#define init_9_6_15_9
void FASTCALL call_9_6_15_9(SH2_struct* sh)
{
  SH2movwi(sh, 6, (15<<4)+9);
}
#define init_9_6_15_10
void FASTCALL call_9_6_15_10(SH2_struct* sh)
{
  SH2movwi(sh, 6, (15<<4)+10);
}
#define init_9_6_15_11
void FASTCALL call_9_6_15_11(SH2_struct* sh)
{
  SH2movwi(sh, 6, (15<<4)+11);
}
#define init_9_6_15_12
void FASTCALL call_9_6_15_12(SH2_struct* sh)
{
  SH2movwi(sh, 6, (15<<4)+12);
}
#define init_9_6_15_13
void FASTCALL call_9_6_15_13(SH2_struct* sh)
{
  SH2movwi(sh, 6, (15<<4)+13);
}
#define init_9_6_15_14
void FASTCALL call_9_6_15_14(SH2_struct* sh)
{
  SH2movwi(sh, 6, (15<<4)+14);
}
#define init_9_6_15_15
void FASTCALL call_9_6_15_15(SH2_struct* sh)
{
  SH2movwi(sh, 6, (15<<4)+15);
}
#define init_9_7_0_0
void FASTCALL call_9_7_0_0(SH2_struct* sh)
{
  SH2movwi(sh, 7, (0<<4)+0);
}
#define init_9_7_0_1
void FASTCALL call_9_7_0_1(SH2_struct* sh)
{
  SH2movwi(sh, 7, (0<<4)+1);
}
#define init_9_7_0_2
void FASTCALL call_9_7_0_2(SH2_struct* sh)
{
  SH2movwi(sh, 7, (0<<4)+2);
}
#define init_9_7_0_3
void FASTCALL call_9_7_0_3(SH2_struct* sh)
{
  SH2movwi(sh, 7, (0<<4)+3);
}
#define init_9_7_0_4
void FASTCALL call_9_7_0_4(SH2_struct* sh)
{
  SH2movwi(sh, 7, (0<<4)+4);
}
#define init_9_7_0_5
void FASTCALL call_9_7_0_5(SH2_struct* sh)
{
  SH2movwi(sh, 7, (0<<4)+5);
}
#define init_9_7_0_6
void FASTCALL call_9_7_0_6(SH2_struct* sh)
{
  SH2movwi(sh, 7, (0<<4)+6);
}
#define init_9_7_0_7
void FASTCALL call_9_7_0_7(SH2_struct* sh)
{
  SH2movwi(sh, 7, (0<<4)+7);
}
#define init_9_7_0_8
void FASTCALL call_9_7_0_8(SH2_struct* sh)
{
  SH2movwi(sh, 7, (0<<4)+8);
}
#define init_9_7_0_9
void FASTCALL call_9_7_0_9(SH2_struct* sh)
{
  SH2movwi(sh, 7, (0<<4)+9);
}
#define init_9_7_0_10
void FASTCALL call_9_7_0_10(SH2_struct* sh)
{
  SH2movwi(sh, 7, (0<<4)+10);
}
#define init_9_7_0_11
void FASTCALL call_9_7_0_11(SH2_struct* sh)
{
  SH2movwi(sh, 7, (0<<4)+11);
}
#define init_9_7_0_12
void FASTCALL call_9_7_0_12(SH2_struct* sh)
{
  SH2movwi(sh, 7, (0<<4)+12);
}
#define init_9_7_0_13
void FASTCALL call_9_7_0_13(SH2_struct* sh)
{
  SH2movwi(sh, 7, (0<<4)+13);
}
#define init_9_7_0_14
void FASTCALL call_9_7_0_14(SH2_struct* sh)
{
  SH2movwi(sh, 7, (0<<4)+14);
}
#define init_9_7_0_15
void FASTCALL call_9_7_0_15(SH2_struct* sh)
{
  SH2movwi(sh, 7, (0<<4)+15);
}
#define init_9_7_1_0
void FASTCALL call_9_7_1_0(SH2_struct* sh)
{
  SH2movwi(sh, 7, (1<<4)+0);
}
#define init_9_7_1_1
void FASTCALL call_9_7_1_1(SH2_struct* sh)
{
  SH2movwi(sh, 7, (1<<4)+1);
}
#define init_9_7_1_2
void FASTCALL call_9_7_1_2(SH2_struct* sh)
{
  SH2movwi(sh, 7, (1<<4)+2);
}
#define init_9_7_1_3
void FASTCALL call_9_7_1_3(SH2_struct* sh)
{
  SH2movwi(sh, 7, (1<<4)+3);
}
#define init_9_7_1_4
void FASTCALL call_9_7_1_4(SH2_struct* sh)
{
  SH2movwi(sh, 7, (1<<4)+4);
}
#define init_9_7_1_5
void FASTCALL call_9_7_1_5(SH2_struct* sh)
{
  SH2movwi(sh, 7, (1<<4)+5);
}
#define init_9_7_1_6
void FASTCALL call_9_7_1_6(SH2_struct* sh)
{
  SH2movwi(sh, 7, (1<<4)+6);
}
#define init_9_7_1_7
void FASTCALL call_9_7_1_7(SH2_struct* sh)
{
  SH2movwi(sh, 7, (1<<4)+7);
}
#define init_9_7_1_8
void FASTCALL call_9_7_1_8(SH2_struct* sh)
{
  SH2movwi(sh, 7, (1<<4)+8);
}
#define init_9_7_1_9
void FASTCALL call_9_7_1_9(SH2_struct* sh)
{
  SH2movwi(sh, 7, (1<<4)+9);
}
#define init_9_7_1_10
void FASTCALL call_9_7_1_10(SH2_struct* sh)
{
  SH2movwi(sh, 7, (1<<4)+10);
}
#define init_9_7_1_11
void FASTCALL call_9_7_1_11(SH2_struct* sh)
{
  SH2movwi(sh, 7, (1<<4)+11);
}
#define init_9_7_1_12
void FASTCALL call_9_7_1_12(SH2_struct* sh)
{
  SH2movwi(sh, 7, (1<<4)+12);
}
#define init_9_7_1_13
void FASTCALL call_9_7_1_13(SH2_struct* sh)
{
  SH2movwi(sh, 7, (1<<4)+13);
}
#define init_9_7_1_14
void FASTCALL call_9_7_1_14(SH2_struct* sh)
{
  SH2movwi(sh, 7, (1<<4)+14);
}
#define init_9_7_1_15
void FASTCALL call_9_7_1_15(SH2_struct* sh)
{
  SH2movwi(sh, 7, (1<<4)+15);
}
#define init_9_7_2_0
void FASTCALL call_9_7_2_0(SH2_struct* sh)
{
  SH2movwi(sh, 7, (2<<4)+0);
}
#define init_9_7_2_1
void FASTCALL call_9_7_2_1(SH2_struct* sh)
{
  SH2movwi(sh, 7, (2<<4)+1);
}
#define init_9_7_2_2
void FASTCALL call_9_7_2_2(SH2_struct* sh)
{
  SH2movwi(sh, 7, (2<<4)+2);
}
#define init_9_7_2_3
void FASTCALL call_9_7_2_3(SH2_struct* sh)
{
  SH2movwi(sh, 7, (2<<4)+3);
}
#define init_9_7_2_4
void FASTCALL call_9_7_2_4(SH2_struct* sh)
{
  SH2movwi(sh, 7, (2<<4)+4);
}
#define init_9_7_2_5
void FASTCALL call_9_7_2_5(SH2_struct* sh)
{
  SH2movwi(sh, 7, (2<<4)+5);
}
#define init_9_7_2_6
void FASTCALL call_9_7_2_6(SH2_struct* sh)
{
  SH2movwi(sh, 7, (2<<4)+6);
}
#define init_9_7_2_7
void FASTCALL call_9_7_2_7(SH2_struct* sh)
{
  SH2movwi(sh, 7, (2<<4)+7);
}
#define init_9_7_2_8
void FASTCALL call_9_7_2_8(SH2_struct* sh)
{
  SH2movwi(sh, 7, (2<<4)+8);
}
#define init_9_7_2_9
void FASTCALL call_9_7_2_9(SH2_struct* sh)
{
  SH2movwi(sh, 7, (2<<4)+9);
}
#define init_9_7_2_10
void FASTCALL call_9_7_2_10(SH2_struct* sh)
{
  SH2movwi(sh, 7, (2<<4)+10);
}
#define init_9_7_2_11
void FASTCALL call_9_7_2_11(SH2_struct* sh)
{
  SH2movwi(sh, 7, (2<<4)+11);
}
#define init_9_7_2_12
void FASTCALL call_9_7_2_12(SH2_struct* sh)
{
  SH2movwi(sh, 7, (2<<4)+12);
}
#define init_9_7_2_13
void FASTCALL call_9_7_2_13(SH2_struct* sh)
{
  SH2movwi(sh, 7, (2<<4)+13);
}
#define init_9_7_2_14
void FASTCALL call_9_7_2_14(SH2_struct* sh)
{
  SH2movwi(sh, 7, (2<<4)+14);
}
#define init_9_7_2_15
void FASTCALL call_9_7_2_15(SH2_struct* sh)
{
  SH2movwi(sh, 7, (2<<4)+15);
}
#define init_9_7_3_0
void FASTCALL call_9_7_3_0(SH2_struct* sh)
{
  SH2movwi(sh, 7, (3<<4)+0);
}
#define init_9_7_3_1
void FASTCALL call_9_7_3_1(SH2_struct* sh)
{
  SH2movwi(sh, 7, (3<<4)+1);
}
#define init_9_7_3_2
void FASTCALL call_9_7_3_2(SH2_struct* sh)
{
  SH2movwi(sh, 7, (3<<4)+2);
}
#define init_9_7_3_3
void FASTCALL call_9_7_3_3(SH2_struct* sh)
{
  SH2movwi(sh, 7, (3<<4)+3);
}
#define init_9_7_3_4
void FASTCALL call_9_7_3_4(SH2_struct* sh)
{
  SH2movwi(sh, 7, (3<<4)+4);
}
#define init_9_7_3_5
void FASTCALL call_9_7_3_5(SH2_struct* sh)
{
  SH2movwi(sh, 7, (3<<4)+5);
}
#define init_9_7_3_6
void FASTCALL call_9_7_3_6(SH2_struct* sh)
{
  SH2movwi(sh, 7, (3<<4)+6);
}
#define init_9_7_3_7
void FASTCALL call_9_7_3_7(SH2_struct* sh)
{
  SH2movwi(sh, 7, (3<<4)+7);
}
#define init_9_7_3_8
void FASTCALL call_9_7_3_8(SH2_struct* sh)
{
  SH2movwi(sh, 7, (3<<4)+8);
}
#define init_9_7_3_9
void FASTCALL call_9_7_3_9(SH2_struct* sh)
{
  SH2movwi(sh, 7, (3<<4)+9);
}
#define init_9_7_3_10
void FASTCALL call_9_7_3_10(SH2_struct* sh)
{
  SH2movwi(sh, 7, (3<<4)+10);
}
#define init_9_7_3_11
void FASTCALL call_9_7_3_11(SH2_struct* sh)
{
  SH2movwi(sh, 7, (3<<4)+11);
}
#define init_9_7_3_12
void FASTCALL call_9_7_3_12(SH2_struct* sh)
{
  SH2movwi(sh, 7, (3<<4)+12);
}
#define init_9_7_3_13
void FASTCALL call_9_7_3_13(SH2_struct* sh)
{
  SH2movwi(sh, 7, (3<<4)+13);
}
#define init_9_7_3_14
void FASTCALL call_9_7_3_14(SH2_struct* sh)
{
  SH2movwi(sh, 7, (3<<4)+14);
}
#define init_9_7_3_15
void FASTCALL call_9_7_3_15(SH2_struct* sh)
{
  SH2movwi(sh, 7, (3<<4)+15);
}
#define init_9_7_4_0
void FASTCALL call_9_7_4_0(SH2_struct* sh)
{
  SH2movwi(sh, 7, (4<<4)+0);
}
#define init_9_7_4_1
void FASTCALL call_9_7_4_1(SH2_struct* sh)
{
  SH2movwi(sh, 7, (4<<4)+1);
}
#define init_9_7_4_2
void FASTCALL call_9_7_4_2(SH2_struct* sh)
{
  SH2movwi(sh, 7, (4<<4)+2);
}
#define init_9_7_4_3
void FASTCALL call_9_7_4_3(SH2_struct* sh)
{
  SH2movwi(sh, 7, (4<<4)+3);
}
#define init_9_7_4_4
void FASTCALL call_9_7_4_4(SH2_struct* sh)
{
  SH2movwi(sh, 7, (4<<4)+4);
}
#define init_9_7_4_5
void FASTCALL call_9_7_4_5(SH2_struct* sh)
{
  SH2movwi(sh, 7, (4<<4)+5);
}
#define init_9_7_4_6
void FASTCALL call_9_7_4_6(SH2_struct* sh)
{
  SH2movwi(sh, 7, (4<<4)+6);
}
#define init_9_7_4_7
void FASTCALL call_9_7_4_7(SH2_struct* sh)
{
  SH2movwi(sh, 7, (4<<4)+7);
}
#define init_9_7_4_8
void FASTCALL call_9_7_4_8(SH2_struct* sh)
{
  SH2movwi(sh, 7, (4<<4)+8);
}
#define init_9_7_4_9
void FASTCALL call_9_7_4_9(SH2_struct* sh)
{
  SH2movwi(sh, 7, (4<<4)+9);
}
#define init_9_7_4_10
void FASTCALL call_9_7_4_10(SH2_struct* sh)
{
  SH2movwi(sh, 7, (4<<4)+10);
}
#define init_9_7_4_11
void FASTCALL call_9_7_4_11(SH2_struct* sh)
{
  SH2movwi(sh, 7, (4<<4)+11);
}
#define init_9_7_4_12
void FASTCALL call_9_7_4_12(SH2_struct* sh)
{
  SH2movwi(sh, 7, (4<<4)+12);
}
#define init_9_7_4_13
void FASTCALL call_9_7_4_13(SH2_struct* sh)
{
  SH2movwi(sh, 7, (4<<4)+13);
}
#define init_9_7_4_14
void FASTCALL call_9_7_4_14(SH2_struct* sh)
{
  SH2movwi(sh, 7, (4<<4)+14);
}
#define init_9_7_4_15
void FASTCALL call_9_7_4_15(SH2_struct* sh)
{
  SH2movwi(sh, 7, (4<<4)+15);
}
#define init_9_7_5_0
void FASTCALL call_9_7_5_0(SH2_struct* sh)
{
  SH2movwi(sh, 7, (5<<4)+0);
}
#define init_9_7_5_1
void FASTCALL call_9_7_5_1(SH2_struct* sh)
{
  SH2movwi(sh, 7, (5<<4)+1);
}
#define init_9_7_5_2
void FASTCALL call_9_7_5_2(SH2_struct* sh)
{
  SH2movwi(sh, 7, (5<<4)+2);
}
#define init_9_7_5_3
void FASTCALL call_9_7_5_3(SH2_struct* sh)
{
  SH2movwi(sh, 7, (5<<4)+3);
}
#define init_9_7_5_4
void FASTCALL call_9_7_5_4(SH2_struct* sh)
{
  SH2movwi(sh, 7, (5<<4)+4);
}
#define init_9_7_5_5
void FASTCALL call_9_7_5_5(SH2_struct* sh)
{
  SH2movwi(sh, 7, (5<<4)+5);
}
#define init_9_7_5_6
void FASTCALL call_9_7_5_6(SH2_struct* sh)
{
  SH2movwi(sh, 7, (5<<4)+6);
}
#define init_9_7_5_7
void FASTCALL call_9_7_5_7(SH2_struct* sh)
{
  SH2movwi(sh, 7, (5<<4)+7);
}
#define init_9_7_5_8
void FASTCALL call_9_7_5_8(SH2_struct* sh)
{
  SH2movwi(sh, 7, (5<<4)+8);
}
#define init_9_7_5_9
void FASTCALL call_9_7_5_9(SH2_struct* sh)
{
  SH2movwi(sh, 7, (5<<4)+9);
}
#define init_9_7_5_10
void FASTCALL call_9_7_5_10(SH2_struct* sh)
{
  SH2movwi(sh, 7, (5<<4)+10);
}
#define init_9_7_5_11
void FASTCALL call_9_7_5_11(SH2_struct* sh)
{
  SH2movwi(sh, 7, (5<<4)+11);
}
#define init_9_7_5_12
void FASTCALL call_9_7_5_12(SH2_struct* sh)
{
  SH2movwi(sh, 7, (5<<4)+12);
}
#define init_9_7_5_13
void FASTCALL call_9_7_5_13(SH2_struct* sh)
{
  SH2movwi(sh, 7, (5<<4)+13);
}
#define init_9_7_5_14
void FASTCALL call_9_7_5_14(SH2_struct* sh)
{
  SH2movwi(sh, 7, (5<<4)+14);
}
#define init_9_7_5_15
void FASTCALL call_9_7_5_15(SH2_struct* sh)
{
  SH2movwi(sh, 7, (5<<4)+15);
}
#define init_9_7_6_0
void FASTCALL call_9_7_6_0(SH2_struct* sh)
{
  SH2movwi(sh, 7, (6<<4)+0);
}
#define init_9_7_6_1
void FASTCALL call_9_7_6_1(SH2_struct* sh)
{
  SH2movwi(sh, 7, (6<<4)+1);
}
#define init_9_7_6_2
void FASTCALL call_9_7_6_2(SH2_struct* sh)
{
  SH2movwi(sh, 7, (6<<4)+2);
}
#define init_9_7_6_3
void FASTCALL call_9_7_6_3(SH2_struct* sh)
{
  SH2movwi(sh, 7, (6<<4)+3);
}
#define init_9_7_6_4
void FASTCALL call_9_7_6_4(SH2_struct* sh)
{
  SH2movwi(sh, 7, (6<<4)+4);
}
#define init_9_7_6_5
void FASTCALL call_9_7_6_5(SH2_struct* sh)
{
  SH2movwi(sh, 7, (6<<4)+5);
}
#define init_9_7_6_6
void FASTCALL call_9_7_6_6(SH2_struct* sh)
{
  SH2movwi(sh, 7, (6<<4)+6);
}
#define init_9_7_6_7
void FASTCALL call_9_7_6_7(SH2_struct* sh)
{
  SH2movwi(sh, 7, (6<<4)+7);
}
#define init_9_7_6_8
void FASTCALL call_9_7_6_8(SH2_struct* sh)
{
  SH2movwi(sh, 7, (6<<4)+8);
}
#define init_9_7_6_9
void FASTCALL call_9_7_6_9(SH2_struct* sh)
{
  SH2movwi(sh, 7, (6<<4)+9);
}
#define init_9_7_6_10
void FASTCALL call_9_7_6_10(SH2_struct* sh)
{
  SH2movwi(sh, 7, (6<<4)+10);
}
#define init_9_7_6_11
void FASTCALL call_9_7_6_11(SH2_struct* sh)
{
  SH2movwi(sh, 7, (6<<4)+11);
}
#define init_9_7_6_12
void FASTCALL call_9_7_6_12(SH2_struct* sh)
{
  SH2movwi(sh, 7, (6<<4)+12);
}
#define init_9_7_6_13
void FASTCALL call_9_7_6_13(SH2_struct* sh)
{
  SH2movwi(sh, 7, (6<<4)+13);
}
#define init_9_7_6_14
void FASTCALL call_9_7_6_14(SH2_struct* sh)
{
  SH2movwi(sh, 7, (6<<4)+14);
}
#define init_9_7_6_15
void FASTCALL call_9_7_6_15(SH2_struct* sh)
{
  SH2movwi(sh, 7, (6<<4)+15);
}
#define init_9_7_7_0
void FASTCALL call_9_7_7_0(SH2_struct* sh)
{
  SH2movwi(sh, 7, (7<<4)+0);
}
#define init_9_7_7_1
void FASTCALL call_9_7_7_1(SH2_struct* sh)
{
  SH2movwi(sh, 7, (7<<4)+1);
}
#define init_9_7_7_2
void FASTCALL call_9_7_7_2(SH2_struct* sh)
{
  SH2movwi(sh, 7, (7<<4)+2);
}
#define init_9_7_7_3
void FASTCALL call_9_7_7_3(SH2_struct* sh)
{
  SH2movwi(sh, 7, (7<<4)+3);
}
#define init_9_7_7_4
void FASTCALL call_9_7_7_4(SH2_struct* sh)
{
  SH2movwi(sh, 7, (7<<4)+4);
}
#define init_9_7_7_5
void FASTCALL call_9_7_7_5(SH2_struct* sh)
{
  SH2movwi(sh, 7, (7<<4)+5);
}
#define init_9_7_7_6
void FASTCALL call_9_7_7_6(SH2_struct* sh)
{
  SH2movwi(sh, 7, (7<<4)+6);
}
#define init_9_7_7_7
void FASTCALL call_9_7_7_7(SH2_struct* sh)
{
  SH2movwi(sh, 7, (7<<4)+7);
}
#define init_9_7_7_8
void FASTCALL call_9_7_7_8(SH2_struct* sh)
{
  SH2movwi(sh, 7, (7<<4)+8);
}
#define init_9_7_7_9
void FASTCALL call_9_7_7_9(SH2_struct* sh)
{
  SH2movwi(sh, 7, (7<<4)+9);
}
#define init_9_7_7_10
void FASTCALL call_9_7_7_10(SH2_struct* sh)
{
  SH2movwi(sh, 7, (7<<4)+10);
}
#define init_9_7_7_11
void FASTCALL call_9_7_7_11(SH2_struct* sh)
{
  SH2movwi(sh, 7, (7<<4)+11);
}
#define init_9_7_7_12
void FASTCALL call_9_7_7_12(SH2_struct* sh)
{
  SH2movwi(sh, 7, (7<<4)+12);
}
#define init_9_7_7_13
void FASTCALL call_9_7_7_13(SH2_struct* sh)
{
  SH2movwi(sh, 7, (7<<4)+13);
}
#define init_9_7_7_14
void FASTCALL call_9_7_7_14(SH2_struct* sh)
{
  SH2movwi(sh, 7, (7<<4)+14);
}
#define init_9_7_7_15
void FASTCALL call_9_7_7_15(SH2_struct* sh)
{
  SH2movwi(sh, 7, (7<<4)+15);
}
#define init_9_7_8_0
void FASTCALL call_9_7_8_0(SH2_struct* sh)
{
  SH2movwi(sh, 7, (8<<4)+0);
}
#define init_9_7_8_1
void FASTCALL call_9_7_8_1(SH2_struct* sh)
{
  SH2movwi(sh, 7, (8<<4)+1);
}
#define init_9_7_8_2
void FASTCALL call_9_7_8_2(SH2_struct* sh)
{
  SH2movwi(sh, 7, (8<<4)+2);
}
#define init_9_7_8_3
void FASTCALL call_9_7_8_3(SH2_struct* sh)
{
  SH2movwi(sh, 7, (8<<4)+3);
}
#define init_9_7_8_4
void FASTCALL call_9_7_8_4(SH2_struct* sh)
{
  SH2movwi(sh, 7, (8<<4)+4);
}
#define init_9_7_8_5
void FASTCALL call_9_7_8_5(SH2_struct* sh)
{
  SH2movwi(sh, 7, (8<<4)+5);
}
#define init_9_7_8_6
void FASTCALL call_9_7_8_6(SH2_struct* sh)
{
  SH2movwi(sh, 7, (8<<4)+6);
}
#define init_9_7_8_7
void FASTCALL call_9_7_8_7(SH2_struct* sh)
{
  SH2movwi(sh, 7, (8<<4)+7);
}
#define init_9_7_8_8
void FASTCALL call_9_7_8_8(SH2_struct* sh)
{
  SH2movwi(sh, 7, (8<<4)+8);
}
#define init_9_7_8_9
void FASTCALL call_9_7_8_9(SH2_struct* sh)
{
  SH2movwi(sh, 7, (8<<4)+9);
}
#define init_9_7_8_10
void FASTCALL call_9_7_8_10(SH2_struct* sh)
{
  SH2movwi(sh, 7, (8<<4)+10);
}
#define init_9_7_8_11
void FASTCALL call_9_7_8_11(SH2_struct* sh)
{
  SH2movwi(sh, 7, (8<<4)+11);
}
#define init_9_7_8_12
void FASTCALL call_9_7_8_12(SH2_struct* sh)
{
  SH2movwi(sh, 7, (8<<4)+12);
}
#define init_9_7_8_13
void FASTCALL call_9_7_8_13(SH2_struct* sh)
{
  SH2movwi(sh, 7, (8<<4)+13);
}
#define init_9_7_8_14
void FASTCALL call_9_7_8_14(SH2_struct* sh)
{
  SH2movwi(sh, 7, (8<<4)+14);
}
#define init_9_7_8_15
void FASTCALL call_9_7_8_15(SH2_struct* sh)
{
  SH2movwi(sh, 7, (8<<4)+15);
}
#define init_9_7_9_0
void FASTCALL call_9_7_9_0(SH2_struct* sh)
{
  SH2movwi(sh, 7, (9<<4)+0);
}
#define init_9_7_9_1
void FASTCALL call_9_7_9_1(SH2_struct* sh)
{
  SH2movwi(sh, 7, (9<<4)+1);
}
#define init_9_7_9_2
void FASTCALL call_9_7_9_2(SH2_struct* sh)
{
  SH2movwi(sh, 7, (9<<4)+2);
}
#define init_9_7_9_3
void FASTCALL call_9_7_9_3(SH2_struct* sh)
{
  SH2movwi(sh, 7, (9<<4)+3);
}
#define init_9_7_9_4
void FASTCALL call_9_7_9_4(SH2_struct* sh)
{
  SH2movwi(sh, 7, (9<<4)+4);
}
#define init_9_7_9_5
void FASTCALL call_9_7_9_5(SH2_struct* sh)
{
  SH2movwi(sh, 7, (9<<4)+5);
}
#define init_9_7_9_6
void FASTCALL call_9_7_9_6(SH2_struct* sh)
{
  SH2movwi(sh, 7, (9<<4)+6);
}
#define init_9_7_9_7
void FASTCALL call_9_7_9_7(SH2_struct* sh)
{
  SH2movwi(sh, 7, (9<<4)+7);
}
#define init_9_7_9_8
void FASTCALL call_9_7_9_8(SH2_struct* sh)
{
  SH2movwi(sh, 7, (9<<4)+8);
}
#define init_9_7_9_9
void FASTCALL call_9_7_9_9(SH2_struct* sh)
{
  SH2movwi(sh, 7, (9<<4)+9);
}
#define init_9_7_9_10
void FASTCALL call_9_7_9_10(SH2_struct* sh)
{
  SH2movwi(sh, 7, (9<<4)+10);
}
#define init_9_7_9_11
void FASTCALL call_9_7_9_11(SH2_struct* sh)
{
  SH2movwi(sh, 7, (9<<4)+11);
}
#define init_9_7_9_12
void FASTCALL call_9_7_9_12(SH2_struct* sh)
{
  SH2movwi(sh, 7, (9<<4)+12);
}
#define init_9_7_9_13
void FASTCALL call_9_7_9_13(SH2_struct* sh)
{
  SH2movwi(sh, 7, (9<<4)+13);
}
#define init_9_7_9_14
void FASTCALL call_9_7_9_14(SH2_struct* sh)
{
  SH2movwi(sh, 7, (9<<4)+14);
}
#define init_9_7_9_15
void FASTCALL call_9_7_9_15(SH2_struct* sh)
{
  SH2movwi(sh, 7, (9<<4)+15);
}
#define init_9_7_10_0
void FASTCALL call_9_7_10_0(SH2_struct* sh)
{
  SH2movwi(sh, 7, (10<<4)+0);
}
#define init_9_7_10_1
void FASTCALL call_9_7_10_1(SH2_struct* sh)
{
  SH2movwi(sh, 7, (10<<4)+1);
}
#define init_9_7_10_2
void FASTCALL call_9_7_10_2(SH2_struct* sh)
{
  SH2movwi(sh, 7, (10<<4)+2);
}
#define init_9_7_10_3
void FASTCALL call_9_7_10_3(SH2_struct* sh)
{
  SH2movwi(sh, 7, (10<<4)+3);
}
#define init_9_7_10_4
void FASTCALL call_9_7_10_4(SH2_struct* sh)
{
  SH2movwi(sh, 7, (10<<4)+4);
}
#define init_9_7_10_5
void FASTCALL call_9_7_10_5(SH2_struct* sh)
{
  SH2movwi(sh, 7, (10<<4)+5);
}
#define init_9_7_10_6
void FASTCALL call_9_7_10_6(SH2_struct* sh)
{
  SH2movwi(sh, 7, (10<<4)+6);
}
#define init_9_7_10_7
void FASTCALL call_9_7_10_7(SH2_struct* sh)
{
  SH2movwi(sh, 7, (10<<4)+7);
}
#define init_9_7_10_8
void FASTCALL call_9_7_10_8(SH2_struct* sh)
{
  SH2movwi(sh, 7, (10<<4)+8);
}
#define init_9_7_10_9
void FASTCALL call_9_7_10_9(SH2_struct* sh)
{
  SH2movwi(sh, 7, (10<<4)+9);
}
#define init_9_7_10_10
void FASTCALL call_9_7_10_10(SH2_struct* sh)
{
  SH2movwi(sh, 7, (10<<4)+10);
}
#define init_9_7_10_11
void FASTCALL call_9_7_10_11(SH2_struct* sh)
{
  SH2movwi(sh, 7, (10<<4)+11);
}
#define init_9_7_10_12
void FASTCALL call_9_7_10_12(SH2_struct* sh)
{
  SH2movwi(sh, 7, (10<<4)+12);
}
#define init_9_7_10_13
void FASTCALL call_9_7_10_13(SH2_struct* sh)
{
  SH2movwi(sh, 7, (10<<4)+13);
}
#define init_9_7_10_14
void FASTCALL call_9_7_10_14(SH2_struct* sh)
{
  SH2movwi(sh, 7, (10<<4)+14);
}
#define init_9_7_10_15
void FASTCALL call_9_7_10_15(SH2_struct* sh)
{
  SH2movwi(sh, 7, (10<<4)+15);
}
#define init_9_7_11_0
void FASTCALL call_9_7_11_0(SH2_struct* sh)
{
  SH2movwi(sh, 7, (11<<4)+0);
}
#define init_9_7_11_1
void FASTCALL call_9_7_11_1(SH2_struct* sh)
{
  SH2movwi(sh, 7, (11<<4)+1);
}
#define init_9_7_11_2
void FASTCALL call_9_7_11_2(SH2_struct* sh)
{
  SH2movwi(sh, 7, (11<<4)+2);
}
#define init_9_7_11_3
void FASTCALL call_9_7_11_3(SH2_struct* sh)
{
  SH2movwi(sh, 7, (11<<4)+3);
}
#define init_9_7_11_4
void FASTCALL call_9_7_11_4(SH2_struct* sh)
{
  SH2movwi(sh, 7, (11<<4)+4);
}
#define init_9_7_11_5
void FASTCALL call_9_7_11_5(SH2_struct* sh)
{
  SH2movwi(sh, 7, (11<<4)+5);
}
#define init_9_7_11_6
void FASTCALL call_9_7_11_6(SH2_struct* sh)
{
  SH2movwi(sh, 7, (11<<4)+6);
}
#define init_9_7_11_7
void FASTCALL call_9_7_11_7(SH2_struct* sh)
{
  SH2movwi(sh, 7, (11<<4)+7);
}
#define init_9_7_11_8
void FASTCALL call_9_7_11_8(SH2_struct* sh)
{
  SH2movwi(sh, 7, (11<<4)+8);
}
#define init_9_7_11_9
void FASTCALL call_9_7_11_9(SH2_struct* sh)
{
  SH2movwi(sh, 7, (11<<4)+9);
}
#define init_9_7_11_10
void FASTCALL call_9_7_11_10(SH2_struct* sh)
{
  SH2movwi(sh, 7, (11<<4)+10);
}
#define init_9_7_11_11
void FASTCALL call_9_7_11_11(SH2_struct* sh)
{
  SH2movwi(sh, 7, (11<<4)+11);
}
#define init_9_7_11_12
void FASTCALL call_9_7_11_12(SH2_struct* sh)
{
  SH2movwi(sh, 7, (11<<4)+12);
}
#define init_9_7_11_13
void FASTCALL call_9_7_11_13(SH2_struct* sh)
{
  SH2movwi(sh, 7, (11<<4)+13);
}
#define init_9_7_11_14
void FASTCALL call_9_7_11_14(SH2_struct* sh)
{
  SH2movwi(sh, 7, (11<<4)+14);
}
#define init_9_7_11_15
void FASTCALL call_9_7_11_15(SH2_struct* sh)
{
  SH2movwi(sh, 7, (11<<4)+15);
}
#define init_9_7_12_0
void FASTCALL call_9_7_12_0(SH2_struct* sh)
{
  SH2movwi(sh, 7, (12<<4)+0);
}
#define init_9_7_12_1
void FASTCALL call_9_7_12_1(SH2_struct* sh)
{
  SH2movwi(sh, 7, (12<<4)+1);
}
#define init_9_7_12_2
void FASTCALL call_9_7_12_2(SH2_struct* sh)
{
  SH2movwi(sh, 7, (12<<4)+2);
}
#define init_9_7_12_3
void FASTCALL call_9_7_12_3(SH2_struct* sh)
{
  SH2movwi(sh, 7, (12<<4)+3);
}
#define init_9_7_12_4
void FASTCALL call_9_7_12_4(SH2_struct* sh)
{
  SH2movwi(sh, 7, (12<<4)+4);
}
#define init_9_7_12_5
void FASTCALL call_9_7_12_5(SH2_struct* sh)
{
  SH2movwi(sh, 7, (12<<4)+5);
}
#define init_9_7_12_6
void FASTCALL call_9_7_12_6(SH2_struct* sh)
{
  SH2movwi(sh, 7, (12<<4)+6);
}
#define init_9_7_12_7
void FASTCALL call_9_7_12_7(SH2_struct* sh)
{
  SH2movwi(sh, 7, (12<<4)+7);
}
#define init_9_7_12_8
void FASTCALL call_9_7_12_8(SH2_struct* sh)
{
  SH2movwi(sh, 7, (12<<4)+8);
}
#define init_9_7_12_9
void FASTCALL call_9_7_12_9(SH2_struct* sh)
{
  SH2movwi(sh, 7, (12<<4)+9);
}
#define init_9_7_12_10
void FASTCALL call_9_7_12_10(SH2_struct* sh)
{
  SH2movwi(sh, 7, (12<<4)+10);
}
#define init_9_7_12_11
void FASTCALL call_9_7_12_11(SH2_struct* sh)
{
  SH2movwi(sh, 7, (12<<4)+11);
}
#define init_9_7_12_12
void FASTCALL call_9_7_12_12(SH2_struct* sh)
{
  SH2movwi(sh, 7, (12<<4)+12);
}
#define init_9_7_12_13
void FASTCALL call_9_7_12_13(SH2_struct* sh)
{
  SH2movwi(sh, 7, (12<<4)+13);
}
#define init_9_7_12_14
void FASTCALL call_9_7_12_14(SH2_struct* sh)
{
  SH2movwi(sh, 7, (12<<4)+14);
}
#define init_9_7_12_15
void FASTCALL call_9_7_12_15(SH2_struct* sh)
{
  SH2movwi(sh, 7, (12<<4)+15);
}
#define init_9_7_13_0
void FASTCALL call_9_7_13_0(SH2_struct* sh)
{
  SH2movwi(sh, 7, (13<<4)+0);
}
#define init_9_7_13_1
void FASTCALL call_9_7_13_1(SH2_struct* sh)
{
  SH2movwi(sh, 7, (13<<4)+1);
}
#define init_9_7_13_2
void FASTCALL call_9_7_13_2(SH2_struct* sh)
{
  SH2movwi(sh, 7, (13<<4)+2);
}
#define init_9_7_13_3
void FASTCALL call_9_7_13_3(SH2_struct* sh)
{
  SH2movwi(sh, 7, (13<<4)+3);
}
#define init_9_7_13_4
void FASTCALL call_9_7_13_4(SH2_struct* sh)
{
  SH2movwi(sh, 7, (13<<4)+4);
}
#define init_9_7_13_5
void FASTCALL call_9_7_13_5(SH2_struct* sh)
{
  SH2movwi(sh, 7, (13<<4)+5);
}
#define init_9_7_13_6
void FASTCALL call_9_7_13_6(SH2_struct* sh)
{
  SH2movwi(sh, 7, (13<<4)+6);
}
#define init_9_7_13_7
void FASTCALL call_9_7_13_7(SH2_struct* sh)
{
  SH2movwi(sh, 7, (13<<4)+7);
}
#define init_9_7_13_8
void FASTCALL call_9_7_13_8(SH2_struct* sh)
{
  SH2movwi(sh, 7, (13<<4)+8);
}
#define init_9_7_13_9
void FASTCALL call_9_7_13_9(SH2_struct* sh)
{
  SH2movwi(sh, 7, (13<<4)+9);
}
#define init_9_7_13_10
void FASTCALL call_9_7_13_10(SH2_struct* sh)
{
  SH2movwi(sh, 7, (13<<4)+10);
}
#define init_9_7_13_11
void FASTCALL call_9_7_13_11(SH2_struct* sh)
{
  SH2movwi(sh, 7, (13<<4)+11);
}
#define init_9_7_13_12
void FASTCALL call_9_7_13_12(SH2_struct* sh)
{
  SH2movwi(sh, 7, (13<<4)+12);
}
#define init_9_7_13_13
void FASTCALL call_9_7_13_13(SH2_struct* sh)
{
  SH2movwi(sh, 7, (13<<4)+13);
}
#define init_9_7_13_14
void FASTCALL call_9_7_13_14(SH2_struct* sh)
{
  SH2movwi(sh, 7, (13<<4)+14);
}
#define init_9_7_13_15
void FASTCALL call_9_7_13_15(SH2_struct* sh)
{
  SH2movwi(sh, 7, (13<<4)+15);
}
#define init_9_7_14_0
void FASTCALL call_9_7_14_0(SH2_struct* sh)
{
  SH2movwi(sh, 7, (14<<4)+0);
}
#define init_9_7_14_1
void FASTCALL call_9_7_14_1(SH2_struct* sh)
{
  SH2movwi(sh, 7, (14<<4)+1);
}
#define init_9_7_14_2
void FASTCALL call_9_7_14_2(SH2_struct* sh)
{
  SH2movwi(sh, 7, (14<<4)+2);
}
#define init_9_7_14_3
void FASTCALL call_9_7_14_3(SH2_struct* sh)
{
  SH2movwi(sh, 7, (14<<4)+3);
}
#define init_9_7_14_4
void FASTCALL call_9_7_14_4(SH2_struct* sh)
{
  SH2movwi(sh, 7, (14<<4)+4);
}
#define init_9_7_14_5
void FASTCALL call_9_7_14_5(SH2_struct* sh)
{
  SH2movwi(sh, 7, (14<<4)+5);
}
#define init_9_7_14_6
void FASTCALL call_9_7_14_6(SH2_struct* sh)
{
  SH2movwi(sh, 7, (14<<4)+6);
}
#define init_9_7_14_7
void FASTCALL call_9_7_14_7(SH2_struct* sh)
{
  SH2movwi(sh, 7, (14<<4)+7);
}
#define init_9_7_14_8
void FASTCALL call_9_7_14_8(SH2_struct* sh)
{
  SH2movwi(sh, 7, (14<<4)+8);
}
#define init_9_7_14_9
void FASTCALL call_9_7_14_9(SH2_struct* sh)
{
  SH2movwi(sh, 7, (14<<4)+9);
}
#define init_9_7_14_10
void FASTCALL call_9_7_14_10(SH2_struct* sh)
{
  SH2movwi(sh, 7, (14<<4)+10);
}
#define init_9_7_14_11
void FASTCALL call_9_7_14_11(SH2_struct* sh)
{
  SH2movwi(sh, 7, (14<<4)+11);
}
#define init_9_7_14_12
void FASTCALL call_9_7_14_12(SH2_struct* sh)
{
  SH2movwi(sh, 7, (14<<4)+12);
}
#define init_9_7_14_13
void FASTCALL call_9_7_14_13(SH2_struct* sh)
{
  SH2movwi(sh, 7, (14<<4)+13);
}
#define init_9_7_14_14
void FASTCALL call_9_7_14_14(SH2_struct* sh)
{
  SH2movwi(sh, 7, (14<<4)+14);
}
#define init_9_7_14_15
void FASTCALL call_9_7_14_15(SH2_struct* sh)
{
  SH2movwi(sh, 7, (14<<4)+15);
}
#define init_9_7_15_0
void FASTCALL call_9_7_15_0(SH2_struct* sh)
{
  SH2movwi(sh, 7, (15<<4)+0);
}
#define init_9_7_15_1
void FASTCALL call_9_7_15_1(SH2_struct* sh)
{
  SH2movwi(sh, 7, (15<<4)+1);
}
#define init_9_7_15_2
void FASTCALL call_9_7_15_2(SH2_struct* sh)
{
  SH2movwi(sh, 7, (15<<4)+2);
}
#define init_9_7_15_3
void FASTCALL call_9_7_15_3(SH2_struct* sh)
{
  SH2movwi(sh, 7, (15<<4)+3);
}
#define init_9_7_15_4
void FASTCALL call_9_7_15_4(SH2_struct* sh)
{
  SH2movwi(sh, 7, (15<<4)+4);
}
#define init_9_7_15_5
void FASTCALL call_9_7_15_5(SH2_struct* sh)
{
  SH2movwi(sh, 7, (15<<4)+5);
}
#define init_9_7_15_6
void FASTCALL call_9_7_15_6(SH2_struct* sh)
{
  SH2movwi(sh, 7, (15<<4)+6);
}
#define init_9_7_15_7
void FASTCALL call_9_7_15_7(SH2_struct* sh)
{
  SH2movwi(sh, 7, (15<<4)+7);
}
#define init_9_7_15_8
void FASTCALL call_9_7_15_8(SH2_struct* sh)
{
  SH2movwi(sh, 7, (15<<4)+8);
}
#define init_9_7_15_9
void FASTCALL call_9_7_15_9(SH2_struct* sh)
{
  SH2movwi(sh, 7, (15<<4)+9);
}
#define init_9_7_15_10
void FASTCALL call_9_7_15_10(SH2_struct* sh)
{
  SH2movwi(sh, 7, (15<<4)+10);
}
#define init_9_7_15_11
void FASTCALL call_9_7_15_11(SH2_struct* sh)
{
  SH2movwi(sh, 7, (15<<4)+11);
}
#define init_9_7_15_12
void FASTCALL call_9_7_15_12(SH2_struct* sh)
{
  SH2movwi(sh, 7, (15<<4)+12);
}
#define init_9_7_15_13
void FASTCALL call_9_7_15_13(SH2_struct* sh)
{
  SH2movwi(sh, 7, (15<<4)+13);
}
#define init_9_7_15_14
void FASTCALL call_9_7_15_14(SH2_struct* sh)
{
  SH2movwi(sh, 7, (15<<4)+14);
}
#define init_9_7_15_15
void FASTCALL call_9_7_15_15(SH2_struct* sh)
{
  SH2movwi(sh, 7, (15<<4)+15);
}
#define init_9_8_0_0
void FASTCALL call_9_8_0_0(SH2_struct* sh)
{
  SH2movwi(sh, 8, (0<<4)+0);
}
#define init_9_8_0_1
void FASTCALL call_9_8_0_1(SH2_struct* sh)
{
  SH2movwi(sh, 8, (0<<4)+1);
}
#define init_9_8_0_2
void FASTCALL call_9_8_0_2(SH2_struct* sh)
{
  SH2movwi(sh, 8, (0<<4)+2);
}
#define init_9_8_0_3
void FASTCALL call_9_8_0_3(SH2_struct* sh)
{
  SH2movwi(sh, 8, (0<<4)+3);
}
#define init_9_8_0_4
void FASTCALL call_9_8_0_4(SH2_struct* sh)
{
  SH2movwi(sh, 8, (0<<4)+4);
}
#define init_9_8_0_5
void FASTCALL call_9_8_0_5(SH2_struct* sh)
{
  SH2movwi(sh, 8, (0<<4)+5);
}
#define init_9_8_0_6
void FASTCALL call_9_8_0_6(SH2_struct* sh)
{
  SH2movwi(sh, 8, (0<<4)+6);
}
#define init_9_8_0_7
void FASTCALL call_9_8_0_7(SH2_struct* sh)
{
  SH2movwi(sh, 8, (0<<4)+7);
}
#define init_9_8_0_8
void FASTCALL call_9_8_0_8(SH2_struct* sh)
{
  SH2movwi(sh, 8, (0<<4)+8);
}
#define init_9_8_0_9
void FASTCALL call_9_8_0_9(SH2_struct* sh)
{
  SH2movwi(sh, 8, (0<<4)+9);
}
#define init_9_8_0_10
void FASTCALL call_9_8_0_10(SH2_struct* sh)
{
  SH2movwi(sh, 8, (0<<4)+10);
}
#define init_9_8_0_11
void FASTCALL call_9_8_0_11(SH2_struct* sh)
{
  SH2movwi(sh, 8, (0<<4)+11);
}
#define init_9_8_0_12
void FASTCALL call_9_8_0_12(SH2_struct* sh)
{
  SH2movwi(sh, 8, (0<<4)+12);
}
#define init_9_8_0_13
void FASTCALL call_9_8_0_13(SH2_struct* sh)
{
  SH2movwi(sh, 8, (0<<4)+13);
}
#define init_9_8_0_14
void FASTCALL call_9_8_0_14(SH2_struct* sh)
{
  SH2movwi(sh, 8, (0<<4)+14);
}
#define init_9_8_0_15
void FASTCALL call_9_8_0_15(SH2_struct* sh)
{
  SH2movwi(sh, 8, (0<<4)+15);
}
#define init_9_8_1_0
void FASTCALL call_9_8_1_0(SH2_struct* sh)
{
  SH2movwi(sh, 8, (1<<4)+0);
}
#define init_9_8_1_1
void FASTCALL call_9_8_1_1(SH2_struct* sh)
{
  SH2movwi(sh, 8, (1<<4)+1);
}
#define init_9_8_1_2
void FASTCALL call_9_8_1_2(SH2_struct* sh)
{
  SH2movwi(sh, 8, (1<<4)+2);
}
#define init_9_8_1_3
void FASTCALL call_9_8_1_3(SH2_struct* sh)
{
  SH2movwi(sh, 8, (1<<4)+3);
}
#define init_9_8_1_4
void FASTCALL call_9_8_1_4(SH2_struct* sh)
{
  SH2movwi(sh, 8, (1<<4)+4);
}
#define init_9_8_1_5
void FASTCALL call_9_8_1_5(SH2_struct* sh)
{
  SH2movwi(sh, 8, (1<<4)+5);
}
#define init_9_8_1_6
void FASTCALL call_9_8_1_6(SH2_struct* sh)
{
  SH2movwi(sh, 8, (1<<4)+6);
}
#define init_9_8_1_7
void FASTCALL call_9_8_1_7(SH2_struct* sh)
{
  SH2movwi(sh, 8, (1<<4)+7);
}
#define init_9_8_1_8
void FASTCALL call_9_8_1_8(SH2_struct* sh)
{
  SH2movwi(sh, 8, (1<<4)+8);
}
#define init_9_8_1_9
void FASTCALL call_9_8_1_9(SH2_struct* sh)
{
  SH2movwi(sh, 8, (1<<4)+9);
}
#define init_9_8_1_10
void FASTCALL call_9_8_1_10(SH2_struct* sh)
{
  SH2movwi(sh, 8, (1<<4)+10);
}
#define init_9_8_1_11
void FASTCALL call_9_8_1_11(SH2_struct* sh)
{
  SH2movwi(sh, 8, (1<<4)+11);
}
#define init_9_8_1_12
void FASTCALL call_9_8_1_12(SH2_struct* sh)
{
  SH2movwi(sh, 8, (1<<4)+12);
}
#define init_9_8_1_13
void FASTCALL call_9_8_1_13(SH2_struct* sh)
{
  SH2movwi(sh, 8, (1<<4)+13);
}
#define init_9_8_1_14
void FASTCALL call_9_8_1_14(SH2_struct* sh)
{
  SH2movwi(sh, 8, (1<<4)+14);
}
#define init_9_8_1_15
void FASTCALL call_9_8_1_15(SH2_struct* sh)
{
  SH2movwi(sh, 8, (1<<4)+15);
}
#define init_9_8_2_0
void FASTCALL call_9_8_2_0(SH2_struct* sh)
{
  SH2movwi(sh, 8, (2<<4)+0);
}
#define init_9_8_2_1
void FASTCALL call_9_8_2_1(SH2_struct* sh)
{
  SH2movwi(sh, 8, (2<<4)+1);
}
#define init_9_8_2_2
void FASTCALL call_9_8_2_2(SH2_struct* sh)
{
  SH2movwi(sh, 8, (2<<4)+2);
}
#define init_9_8_2_3
void FASTCALL call_9_8_2_3(SH2_struct* sh)
{
  SH2movwi(sh, 8, (2<<4)+3);
}
#define init_9_8_2_4
void FASTCALL call_9_8_2_4(SH2_struct* sh)
{
  SH2movwi(sh, 8, (2<<4)+4);
}
#define init_9_8_2_5
void FASTCALL call_9_8_2_5(SH2_struct* sh)
{
  SH2movwi(sh, 8, (2<<4)+5);
}
#define init_9_8_2_6
void FASTCALL call_9_8_2_6(SH2_struct* sh)
{
  SH2movwi(sh, 8, (2<<4)+6);
}
#define init_9_8_2_7
void FASTCALL call_9_8_2_7(SH2_struct* sh)
{
  SH2movwi(sh, 8, (2<<4)+7);
}
#define init_9_8_2_8
void FASTCALL call_9_8_2_8(SH2_struct* sh)
{
  SH2movwi(sh, 8, (2<<4)+8);
}
#define init_9_8_2_9
void FASTCALL call_9_8_2_9(SH2_struct* sh)
{
  SH2movwi(sh, 8, (2<<4)+9);
}
#define init_9_8_2_10
void FASTCALL call_9_8_2_10(SH2_struct* sh)
{
  SH2movwi(sh, 8, (2<<4)+10);
}
#define init_9_8_2_11
void FASTCALL call_9_8_2_11(SH2_struct* sh)
{
  SH2movwi(sh, 8, (2<<4)+11);
}
#define init_9_8_2_12
void FASTCALL call_9_8_2_12(SH2_struct* sh)
{
  SH2movwi(sh, 8, (2<<4)+12);
}
#define init_9_8_2_13
void FASTCALL call_9_8_2_13(SH2_struct* sh)
{
  SH2movwi(sh, 8, (2<<4)+13);
}
#define init_9_8_2_14
void FASTCALL call_9_8_2_14(SH2_struct* sh)
{
  SH2movwi(sh, 8, (2<<4)+14);
}
#define init_9_8_2_15
void FASTCALL call_9_8_2_15(SH2_struct* sh)
{
  SH2movwi(sh, 8, (2<<4)+15);
}
#define init_9_8_3_0
void FASTCALL call_9_8_3_0(SH2_struct* sh)
{
  SH2movwi(sh, 8, (3<<4)+0);
}
#define init_9_8_3_1
void FASTCALL call_9_8_3_1(SH2_struct* sh)
{
  SH2movwi(sh, 8, (3<<4)+1);
}
#define init_9_8_3_2
void FASTCALL call_9_8_3_2(SH2_struct* sh)
{
  SH2movwi(sh, 8, (3<<4)+2);
}
#define init_9_8_3_3
void FASTCALL call_9_8_3_3(SH2_struct* sh)
{
  SH2movwi(sh, 8, (3<<4)+3);
}
#define init_9_8_3_4
void FASTCALL call_9_8_3_4(SH2_struct* sh)
{
  SH2movwi(sh, 8, (3<<4)+4);
}
#define init_9_8_3_5
void FASTCALL call_9_8_3_5(SH2_struct* sh)
{
  SH2movwi(sh, 8, (3<<4)+5);
}
#define init_9_8_3_6
void FASTCALL call_9_8_3_6(SH2_struct* sh)
{
  SH2movwi(sh, 8, (3<<4)+6);
}
#define init_9_8_3_7
void FASTCALL call_9_8_3_7(SH2_struct* sh)
{
  SH2movwi(sh, 8, (3<<4)+7);
}
#define init_9_8_3_8
void FASTCALL call_9_8_3_8(SH2_struct* sh)
{
  SH2movwi(sh, 8, (3<<4)+8);
}
#define init_9_8_3_9
void FASTCALL call_9_8_3_9(SH2_struct* sh)
{
  SH2movwi(sh, 8, (3<<4)+9);
}
#define init_9_8_3_10
void FASTCALL call_9_8_3_10(SH2_struct* sh)
{
  SH2movwi(sh, 8, (3<<4)+10);
}
#define init_9_8_3_11
void FASTCALL call_9_8_3_11(SH2_struct* sh)
{
  SH2movwi(sh, 8, (3<<4)+11);
}
#define init_9_8_3_12
void FASTCALL call_9_8_3_12(SH2_struct* sh)
{
  SH2movwi(sh, 8, (3<<4)+12);
}
#define init_9_8_3_13
void FASTCALL call_9_8_3_13(SH2_struct* sh)
{
  SH2movwi(sh, 8, (3<<4)+13);
}
#define init_9_8_3_14
void FASTCALL call_9_8_3_14(SH2_struct* sh)
{
  SH2movwi(sh, 8, (3<<4)+14);
}
#define init_9_8_3_15
void FASTCALL call_9_8_3_15(SH2_struct* sh)
{
  SH2movwi(sh, 8, (3<<4)+15);
}
#define init_9_8_4_0
void FASTCALL call_9_8_4_0(SH2_struct* sh)
{
  SH2movwi(sh, 8, (4<<4)+0);
}
#define init_9_8_4_1
void FASTCALL call_9_8_4_1(SH2_struct* sh)
{
  SH2movwi(sh, 8, (4<<4)+1);
}
#define init_9_8_4_2
void FASTCALL call_9_8_4_2(SH2_struct* sh)
{
  SH2movwi(sh, 8, (4<<4)+2);
}
#define init_9_8_4_3
void FASTCALL call_9_8_4_3(SH2_struct* sh)
{
  SH2movwi(sh, 8, (4<<4)+3);
}
#define init_9_8_4_4
void FASTCALL call_9_8_4_4(SH2_struct* sh)
{
  SH2movwi(sh, 8, (4<<4)+4);
}
#define init_9_8_4_5
void FASTCALL call_9_8_4_5(SH2_struct* sh)
{
  SH2movwi(sh, 8, (4<<4)+5);
}
#define init_9_8_4_6
void FASTCALL call_9_8_4_6(SH2_struct* sh)
{
  SH2movwi(sh, 8, (4<<4)+6);
}
#define init_9_8_4_7
void FASTCALL call_9_8_4_7(SH2_struct* sh)
{
  SH2movwi(sh, 8, (4<<4)+7);
}
#define init_9_8_4_8
void FASTCALL call_9_8_4_8(SH2_struct* sh)
{
  SH2movwi(sh, 8, (4<<4)+8);
}
#define init_9_8_4_9
void FASTCALL call_9_8_4_9(SH2_struct* sh)
{
  SH2movwi(sh, 8, (4<<4)+9);
}
#define init_9_8_4_10
void FASTCALL call_9_8_4_10(SH2_struct* sh)
{
  SH2movwi(sh, 8, (4<<4)+10);
}
#define init_9_8_4_11
void FASTCALL call_9_8_4_11(SH2_struct* sh)
{
  SH2movwi(sh, 8, (4<<4)+11);
}
#define init_9_8_4_12
void FASTCALL call_9_8_4_12(SH2_struct* sh)
{
  SH2movwi(sh, 8, (4<<4)+12);
}
#define init_9_8_4_13
void FASTCALL call_9_8_4_13(SH2_struct* sh)
{
  SH2movwi(sh, 8, (4<<4)+13);
}
#define init_9_8_4_14
void FASTCALL call_9_8_4_14(SH2_struct* sh)
{
  SH2movwi(sh, 8, (4<<4)+14);
}
#define init_9_8_4_15
void FASTCALL call_9_8_4_15(SH2_struct* sh)
{
  SH2movwi(sh, 8, (4<<4)+15);
}
#define init_9_8_5_0
void FASTCALL call_9_8_5_0(SH2_struct* sh)
{
  SH2movwi(sh, 8, (5<<4)+0);
}
#define init_9_8_5_1
void FASTCALL call_9_8_5_1(SH2_struct* sh)
{
  SH2movwi(sh, 8, (5<<4)+1);
}
#define init_9_8_5_2
void FASTCALL call_9_8_5_2(SH2_struct* sh)
{
  SH2movwi(sh, 8, (5<<4)+2);
}
#define init_9_8_5_3
void FASTCALL call_9_8_5_3(SH2_struct* sh)
{
  SH2movwi(sh, 8, (5<<4)+3);
}
#define init_9_8_5_4
void FASTCALL call_9_8_5_4(SH2_struct* sh)
{
  SH2movwi(sh, 8, (5<<4)+4);
}
#define init_9_8_5_5
void FASTCALL call_9_8_5_5(SH2_struct* sh)
{
  SH2movwi(sh, 8, (5<<4)+5);
}
#define init_9_8_5_6
void FASTCALL call_9_8_5_6(SH2_struct* sh)
{
  SH2movwi(sh, 8, (5<<4)+6);
}
#define init_9_8_5_7
void FASTCALL call_9_8_5_7(SH2_struct* sh)
{
  SH2movwi(sh, 8, (5<<4)+7);
}
#define init_9_8_5_8
void FASTCALL call_9_8_5_8(SH2_struct* sh)
{
  SH2movwi(sh, 8, (5<<4)+8);
}
#define init_9_8_5_9
void FASTCALL call_9_8_5_9(SH2_struct* sh)
{
  SH2movwi(sh, 8, (5<<4)+9);
}
#define init_9_8_5_10
void FASTCALL call_9_8_5_10(SH2_struct* sh)
{
  SH2movwi(sh, 8, (5<<4)+10);
}
#define init_9_8_5_11
void FASTCALL call_9_8_5_11(SH2_struct* sh)
{
  SH2movwi(sh, 8, (5<<4)+11);
}
#define init_9_8_5_12
void FASTCALL call_9_8_5_12(SH2_struct* sh)
{
  SH2movwi(sh, 8, (5<<4)+12);
}
#define init_9_8_5_13
void FASTCALL call_9_8_5_13(SH2_struct* sh)
{
  SH2movwi(sh, 8, (5<<4)+13);
}
#define init_9_8_5_14
void FASTCALL call_9_8_5_14(SH2_struct* sh)
{
  SH2movwi(sh, 8, (5<<4)+14);
}
#define init_9_8_5_15
void FASTCALL call_9_8_5_15(SH2_struct* sh)
{
  SH2movwi(sh, 8, (5<<4)+15);
}
#define init_9_8_6_0
void FASTCALL call_9_8_6_0(SH2_struct* sh)
{
  SH2movwi(sh, 8, (6<<4)+0);
}
#define init_9_8_6_1
void FASTCALL call_9_8_6_1(SH2_struct* sh)
{
  SH2movwi(sh, 8, (6<<4)+1);
}
#define init_9_8_6_2
void FASTCALL call_9_8_6_2(SH2_struct* sh)
{
  SH2movwi(sh, 8, (6<<4)+2);
}
#define init_9_8_6_3
void FASTCALL call_9_8_6_3(SH2_struct* sh)
{
  SH2movwi(sh, 8, (6<<4)+3);
}
#define init_9_8_6_4
void FASTCALL call_9_8_6_4(SH2_struct* sh)
{
  SH2movwi(sh, 8, (6<<4)+4);
}
#define init_9_8_6_5
void FASTCALL call_9_8_6_5(SH2_struct* sh)
{
  SH2movwi(sh, 8, (6<<4)+5);
}
#define init_9_8_6_6
void FASTCALL call_9_8_6_6(SH2_struct* sh)
{
  SH2movwi(sh, 8, (6<<4)+6);
}
#define init_9_8_6_7
void FASTCALL call_9_8_6_7(SH2_struct* sh)
{
  SH2movwi(sh, 8, (6<<4)+7);
}
#define init_9_8_6_8
void FASTCALL call_9_8_6_8(SH2_struct* sh)
{
  SH2movwi(sh, 8, (6<<4)+8);
}
#define init_9_8_6_9
void FASTCALL call_9_8_6_9(SH2_struct* sh)
{
  SH2movwi(sh, 8, (6<<4)+9);
}
#define init_9_8_6_10
void FASTCALL call_9_8_6_10(SH2_struct* sh)
{
  SH2movwi(sh, 8, (6<<4)+10);
}
#define init_9_8_6_11
void FASTCALL call_9_8_6_11(SH2_struct* sh)
{
  SH2movwi(sh, 8, (6<<4)+11);
}
#define init_9_8_6_12
void FASTCALL call_9_8_6_12(SH2_struct* sh)
{
  SH2movwi(sh, 8, (6<<4)+12);
}
#define init_9_8_6_13
void FASTCALL call_9_8_6_13(SH2_struct* sh)
{
  SH2movwi(sh, 8, (6<<4)+13);
}
#define init_9_8_6_14
void FASTCALL call_9_8_6_14(SH2_struct* sh)
{
  SH2movwi(sh, 8, (6<<4)+14);
}
#define init_9_8_6_15
void FASTCALL call_9_8_6_15(SH2_struct* sh)
{
  SH2movwi(sh, 8, (6<<4)+15);
}
#define init_9_8_7_0
void FASTCALL call_9_8_7_0(SH2_struct* sh)
{
  SH2movwi(sh, 8, (7<<4)+0);
}
#define init_9_8_7_1
void FASTCALL call_9_8_7_1(SH2_struct* sh)
{
  SH2movwi(sh, 8, (7<<4)+1);
}
#define init_9_8_7_2
void FASTCALL call_9_8_7_2(SH2_struct* sh)
{
  SH2movwi(sh, 8, (7<<4)+2);
}
#define init_9_8_7_3
void FASTCALL call_9_8_7_3(SH2_struct* sh)
{
  SH2movwi(sh, 8, (7<<4)+3);
}
#define init_9_8_7_4
void FASTCALL call_9_8_7_4(SH2_struct* sh)
{
  SH2movwi(sh, 8, (7<<4)+4);
}
#define init_9_8_7_5
void FASTCALL call_9_8_7_5(SH2_struct* sh)
{
  SH2movwi(sh, 8, (7<<4)+5);
}
#define init_9_8_7_6
void FASTCALL call_9_8_7_6(SH2_struct* sh)
{
  SH2movwi(sh, 8, (7<<4)+6);
}
#define init_9_8_7_7
void FASTCALL call_9_8_7_7(SH2_struct* sh)
{
  SH2movwi(sh, 8, (7<<4)+7);
}
#define init_9_8_7_8
void FASTCALL call_9_8_7_8(SH2_struct* sh)
{
  SH2movwi(sh, 8, (7<<4)+8);
}
#define init_9_8_7_9
void FASTCALL call_9_8_7_9(SH2_struct* sh)
{
  SH2movwi(sh, 8, (7<<4)+9);
}
#define init_9_8_7_10
void FASTCALL call_9_8_7_10(SH2_struct* sh)
{
  SH2movwi(sh, 8, (7<<4)+10);
}
#define init_9_8_7_11
void FASTCALL call_9_8_7_11(SH2_struct* sh)
{
  SH2movwi(sh, 8, (7<<4)+11);
}
#define init_9_8_7_12
void FASTCALL call_9_8_7_12(SH2_struct* sh)
{
  SH2movwi(sh, 8, (7<<4)+12);
}
#define init_9_8_7_13
void FASTCALL call_9_8_7_13(SH2_struct* sh)
{
  SH2movwi(sh, 8, (7<<4)+13);
}
#define init_9_8_7_14
void FASTCALL call_9_8_7_14(SH2_struct* sh)
{
  SH2movwi(sh, 8, (7<<4)+14);
}
#define init_9_8_7_15
void FASTCALL call_9_8_7_15(SH2_struct* sh)
{
  SH2movwi(sh, 8, (7<<4)+15);
}
#define init_9_8_8_0
void FASTCALL call_9_8_8_0(SH2_struct* sh)
{
  SH2movwi(sh, 8, (8<<4)+0);
}
#define init_9_8_8_1
void FASTCALL call_9_8_8_1(SH2_struct* sh)
{
  SH2movwi(sh, 8, (8<<4)+1);
}
#define init_9_8_8_2
void FASTCALL call_9_8_8_2(SH2_struct* sh)
{
  SH2movwi(sh, 8, (8<<4)+2);
}
#define init_9_8_8_3
void FASTCALL call_9_8_8_3(SH2_struct* sh)
{
  SH2movwi(sh, 8, (8<<4)+3);
}
#define init_9_8_8_4
void FASTCALL call_9_8_8_4(SH2_struct* sh)
{
  SH2movwi(sh, 8, (8<<4)+4);
}
#define init_9_8_8_5
void FASTCALL call_9_8_8_5(SH2_struct* sh)
{
  SH2movwi(sh, 8, (8<<4)+5);
}
#define init_9_8_8_6
void FASTCALL call_9_8_8_6(SH2_struct* sh)
{
  SH2movwi(sh, 8, (8<<4)+6);
}
#define init_9_8_8_7
void FASTCALL call_9_8_8_7(SH2_struct* sh)
{
  SH2movwi(sh, 8, (8<<4)+7);
}
#define init_9_8_8_8
void FASTCALL call_9_8_8_8(SH2_struct* sh)
{
  SH2movwi(sh, 8, (8<<4)+8);
}
#define init_9_8_8_9
void FASTCALL call_9_8_8_9(SH2_struct* sh)
{
  SH2movwi(sh, 8, (8<<4)+9);
}
#define init_9_8_8_10
void FASTCALL call_9_8_8_10(SH2_struct* sh)
{
  SH2movwi(sh, 8, (8<<4)+10);
}
#define init_9_8_8_11
void FASTCALL call_9_8_8_11(SH2_struct* sh)
{
  SH2movwi(sh, 8, (8<<4)+11);
}
#define init_9_8_8_12
void FASTCALL call_9_8_8_12(SH2_struct* sh)
{
  SH2movwi(sh, 8, (8<<4)+12);
}
#define init_9_8_8_13
void FASTCALL call_9_8_8_13(SH2_struct* sh)
{
  SH2movwi(sh, 8, (8<<4)+13);
}
#define init_9_8_8_14
void FASTCALL call_9_8_8_14(SH2_struct* sh)
{
  SH2movwi(sh, 8, (8<<4)+14);
}
#define init_9_8_8_15
void FASTCALL call_9_8_8_15(SH2_struct* sh)
{
  SH2movwi(sh, 8, (8<<4)+15);
}
#define init_9_8_9_0
void FASTCALL call_9_8_9_0(SH2_struct* sh)
{
  SH2movwi(sh, 8, (9<<4)+0);
}
#define init_9_8_9_1
void FASTCALL call_9_8_9_1(SH2_struct* sh)
{
  SH2movwi(sh, 8, (9<<4)+1);
}
#define init_9_8_9_2
void FASTCALL call_9_8_9_2(SH2_struct* sh)
{
  SH2movwi(sh, 8, (9<<4)+2);
}
#define init_9_8_9_3
void FASTCALL call_9_8_9_3(SH2_struct* sh)
{
  SH2movwi(sh, 8, (9<<4)+3);
}
#define init_9_8_9_4
void FASTCALL call_9_8_9_4(SH2_struct* sh)
{
  SH2movwi(sh, 8, (9<<4)+4);
}
#define init_9_8_9_5
void FASTCALL call_9_8_9_5(SH2_struct* sh)
{
  SH2movwi(sh, 8, (9<<4)+5);
}
#define init_9_8_9_6
void FASTCALL call_9_8_9_6(SH2_struct* sh)
{
  SH2movwi(sh, 8, (9<<4)+6);
}
#define init_9_8_9_7
void FASTCALL call_9_8_9_7(SH2_struct* sh)
{
  SH2movwi(sh, 8, (9<<4)+7);
}
#define init_9_8_9_8
void FASTCALL call_9_8_9_8(SH2_struct* sh)
{
  SH2movwi(sh, 8, (9<<4)+8);
}
#define init_9_8_9_9
void FASTCALL call_9_8_9_9(SH2_struct* sh)
{
  SH2movwi(sh, 8, (9<<4)+9);
}
#define init_9_8_9_10
void FASTCALL call_9_8_9_10(SH2_struct* sh)
{
  SH2movwi(sh, 8, (9<<4)+10);
}
#define init_9_8_9_11
void FASTCALL call_9_8_9_11(SH2_struct* sh)
{
  SH2movwi(sh, 8, (9<<4)+11);
}
#define init_9_8_9_12
void FASTCALL call_9_8_9_12(SH2_struct* sh)
{
  SH2movwi(sh, 8, (9<<4)+12);
}
#define init_9_8_9_13
void FASTCALL call_9_8_9_13(SH2_struct* sh)
{
  SH2movwi(sh, 8, (9<<4)+13);
}
#define init_9_8_9_14
void FASTCALL call_9_8_9_14(SH2_struct* sh)
{
  SH2movwi(sh, 8, (9<<4)+14);
}
#define init_9_8_9_15
void FASTCALL call_9_8_9_15(SH2_struct* sh)
{
  SH2movwi(sh, 8, (9<<4)+15);
}
#define init_9_8_10_0
void FASTCALL call_9_8_10_0(SH2_struct* sh)
{
  SH2movwi(sh, 8, (10<<4)+0);
}
#define init_9_8_10_1
void FASTCALL call_9_8_10_1(SH2_struct* sh)
{
  SH2movwi(sh, 8, (10<<4)+1);
}
#define init_9_8_10_2
void FASTCALL call_9_8_10_2(SH2_struct* sh)
{
  SH2movwi(sh, 8, (10<<4)+2);
}
#define init_9_8_10_3
void FASTCALL call_9_8_10_3(SH2_struct* sh)
{
  SH2movwi(sh, 8, (10<<4)+3);
}
#define init_9_8_10_4
void FASTCALL call_9_8_10_4(SH2_struct* sh)
{
  SH2movwi(sh, 8, (10<<4)+4);
}
#define init_9_8_10_5
void FASTCALL call_9_8_10_5(SH2_struct* sh)
{
  SH2movwi(sh, 8, (10<<4)+5);
}
#define init_9_8_10_6
void FASTCALL call_9_8_10_6(SH2_struct* sh)
{
  SH2movwi(sh, 8, (10<<4)+6);
}
#define init_9_8_10_7
void FASTCALL call_9_8_10_7(SH2_struct* sh)
{
  SH2movwi(sh, 8, (10<<4)+7);
}
#define init_9_8_10_8
void FASTCALL call_9_8_10_8(SH2_struct* sh)
{
  SH2movwi(sh, 8, (10<<4)+8);
}
#define init_9_8_10_9
void FASTCALL call_9_8_10_9(SH2_struct* sh)
{
  SH2movwi(sh, 8, (10<<4)+9);
}
#define init_9_8_10_10
void FASTCALL call_9_8_10_10(SH2_struct* sh)
{
  SH2movwi(sh, 8, (10<<4)+10);
}
#define init_9_8_10_11
void FASTCALL call_9_8_10_11(SH2_struct* sh)
{
  SH2movwi(sh, 8, (10<<4)+11);
}
#define init_9_8_10_12
void FASTCALL call_9_8_10_12(SH2_struct* sh)
{
  SH2movwi(sh, 8, (10<<4)+12);
}
#define init_9_8_10_13
void FASTCALL call_9_8_10_13(SH2_struct* sh)
{
  SH2movwi(sh, 8, (10<<4)+13);
}
#define init_9_8_10_14
void FASTCALL call_9_8_10_14(SH2_struct* sh)
{
  SH2movwi(sh, 8, (10<<4)+14);
}
#define init_9_8_10_15
void FASTCALL call_9_8_10_15(SH2_struct* sh)
{
  SH2movwi(sh, 8, (10<<4)+15);
}
#define init_9_8_11_0
void FASTCALL call_9_8_11_0(SH2_struct* sh)
{
  SH2movwi(sh, 8, (11<<4)+0);
}
#define init_9_8_11_1
void FASTCALL call_9_8_11_1(SH2_struct* sh)
{
  SH2movwi(sh, 8, (11<<4)+1);
}
#define init_9_8_11_2
void FASTCALL call_9_8_11_2(SH2_struct* sh)
{
  SH2movwi(sh, 8, (11<<4)+2);
}
#define init_9_8_11_3
void FASTCALL call_9_8_11_3(SH2_struct* sh)
{
  SH2movwi(sh, 8, (11<<4)+3);
}
#define init_9_8_11_4
void FASTCALL call_9_8_11_4(SH2_struct* sh)
{
  SH2movwi(sh, 8, (11<<4)+4);
}
#define init_9_8_11_5
void FASTCALL call_9_8_11_5(SH2_struct* sh)
{
  SH2movwi(sh, 8, (11<<4)+5);
}
#define init_9_8_11_6
void FASTCALL call_9_8_11_6(SH2_struct* sh)
{
  SH2movwi(sh, 8, (11<<4)+6);
}
#define init_9_8_11_7
void FASTCALL call_9_8_11_7(SH2_struct* sh)
{
  SH2movwi(sh, 8, (11<<4)+7);
}
#define init_9_8_11_8
void FASTCALL call_9_8_11_8(SH2_struct* sh)
{
  SH2movwi(sh, 8, (11<<4)+8);
}
#define init_9_8_11_9
void FASTCALL call_9_8_11_9(SH2_struct* sh)
{
  SH2movwi(sh, 8, (11<<4)+9);
}
#define init_9_8_11_10
void FASTCALL call_9_8_11_10(SH2_struct* sh)
{
  SH2movwi(sh, 8, (11<<4)+10);
}
#define init_9_8_11_11
void FASTCALL call_9_8_11_11(SH2_struct* sh)
{
  SH2movwi(sh, 8, (11<<4)+11);
}
#define init_9_8_11_12
void FASTCALL call_9_8_11_12(SH2_struct* sh)
{
  SH2movwi(sh, 8, (11<<4)+12);
}
#define init_9_8_11_13
void FASTCALL call_9_8_11_13(SH2_struct* sh)
{
  SH2movwi(sh, 8, (11<<4)+13);
}
#define init_9_8_11_14
void FASTCALL call_9_8_11_14(SH2_struct* sh)
{
  SH2movwi(sh, 8, (11<<4)+14);
}
#define init_9_8_11_15
void FASTCALL call_9_8_11_15(SH2_struct* sh)
{
  SH2movwi(sh, 8, (11<<4)+15);
}
#define init_9_8_12_0
void FASTCALL call_9_8_12_0(SH2_struct* sh)
{
  SH2movwi(sh, 8, (12<<4)+0);
}
#define init_9_8_12_1
void FASTCALL call_9_8_12_1(SH2_struct* sh)
{
  SH2movwi(sh, 8, (12<<4)+1);
}
#define init_9_8_12_2
void FASTCALL call_9_8_12_2(SH2_struct* sh)
{
  SH2movwi(sh, 8, (12<<4)+2);
}
#define init_9_8_12_3
void FASTCALL call_9_8_12_3(SH2_struct* sh)
{
  SH2movwi(sh, 8, (12<<4)+3);
}
#define init_9_8_12_4
void FASTCALL call_9_8_12_4(SH2_struct* sh)
{
  SH2movwi(sh, 8, (12<<4)+4);
}
#define init_9_8_12_5
void FASTCALL call_9_8_12_5(SH2_struct* sh)
{
  SH2movwi(sh, 8, (12<<4)+5);
}
#define init_9_8_12_6
void FASTCALL call_9_8_12_6(SH2_struct* sh)
{
  SH2movwi(sh, 8, (12<<4)+6);
}
#define init_9_8_12_7
void FASTCALL call_9_8_12_7(SH2_struct* sh)
{
  SH2movwi(sh, 8, (12<<4)+7);
}
#define init_9_8_12_8
void FASTCALL call_9_8_12_8(SH2_struct* sh)
{
  SH2movwi(sh, 8, (12<<4)+8);
}
#define init_9_8_12_9
void FASTCALL call_9_8_12_9(SH2_struct* sh)
{
  SH2movwi(sh, 8, (12<<4)+9);
}
#define init_9_8_12_10
void FASTCALL call_9_8_12_10(SH2_struct* sh)
{
  SH2movwi(sh, 8, (12<<4)+10);
}
#define init_9_8_12_11
void FASTCALL call_9_8_12_11(SH2_struct* sh)
{
  SH2movwi(sh, 8, (12<<4)+11);
}
#define init_9_8_12_12
void FASTCALL call_9_8_12_12(SH2_struct* sh)
{
  SH2movwi(sh, 8, (12<<4)+12);
}
#define init_9_8_12_13
void FASTCALL call_9_8_12_13(SH2_struct* sh)
{
  SH2movwi(sh, 8, (12<<4)+13);
}
#define init_9_8_12_14
void FASTCALL call_9_8_12_14(SH2_struct* sh)
{
  SH2movwi(sh, 8, (12<<4)+14);
}
#define init_9_8_12_15
void FASTCALL call_9_8_12_15(SH2_struct* sh)
{
  SH2movwi(sh, 8, (12<<4)+15);
}
#define init_9_8_13_0
void FASTCALL call_9_8_13_0(SH2_struct* sh)
{
  SH2movwi(sh, 8, (13<<4)+0);
}
#define init_9_8_13_1
void FASTCALL call_9_8_13_1(SH2_struct* sh)
{
  SH2movwi(sh, 8, (13<<4)+1);
}
#define init_9_8_13_2
void FASTCALL call_9_8_13_2(SH2_struct* sh)
{
  SH2movwi(sh, 8, (13<<4)+2);
}
#define init_9_8_13_3
void FASTCALL call_9_8_13_3(SH2_struct* sh)
{
  SH2movwi(sh, 8, (13<<4)+3);
}
#define init_9_8_13_4
void FASTCALL call_9_8_13_4(SH2_struct* sh)
{
  SH2movwi(sh, 8, (13<<4)+4);
}
#define init_9_8_13_5
void FASTCALL call_9_8_13_5(SH2_struct* sh)
{
  SH2movwi(sh, 8, (13<<4)+5);
}
#define init_9_8_13_6
void FASTCALL call_9_8_13_6(SH2_struct* sh)
{
  SH2movwi(sh, 8, (13<<4)+6);
}
#define init_9_8_13_7
void FASTCALL call_9_8_13_7(SH2_struct* sh)
{
  SH2movwi(sh, 8, (13<<4)+7);
}
#define init_9_8_13_8
void FASTCALL call_9_8_13_8(SH2_struct* sh)
{
  SH2movwi(sh, 8, (13<<4)+8);
}
#define init_9_8_13_9
void FASTCALL call_9_8_13_9(SH2_struct* sh)
{
  SH2movwi(sh, 8, (13<<4)+9);
}
#define init_9_8_13_10
void FASTCALL call_9_8_13_10(SH2_struct* sh)
{
  SH2movwi(sh, 8, (13<<4)+10);
}
#define init_9_8_13_11
void FASTCALL call_9_8_13_11(SH2_struct* sh)
{
  SH2movwi(sh, 8, (13<<4)+11);
}
#define init_9_8_13_12
void FASTCALL call_9_8_13_12(SH2_struct* sh)
{
  SH2movwi(sh, 8, (13<<4)+12);
}
#define init_9_8_13_13
void FASTCALL call_9_8_13_13(SH2_struct* sh)
{
  SH2movwi(sh, 8, (13<<4)+13);
}
#define init_9_8_13_14
void FASTCALL call_9_8_13_14(SH2_struct* sh)
{
  SH2movwi(sh, 8, (13<<4)+14);
}
#define init_9_8_13_15
void FASTCALL call_9_8_13_15(SH2_struct* sh)
{
  SH2movwi(sh, 8, (13<<4)+15);
}
#define init_9_8_14_0
void FASTCALL call_9_8_14_0(SH2_struct* sh)
{
  SH2movwi(sh, 8, (14<<4)+0);
}
#define init_9_8_14_1
void FASTCALL call_9_8_14_1(SH2_struct* sh)
{
  SH2movwi(sh, 8, (14<<4)+1);
}
#define init_9_8_14_2
void FASTCALL call_9_8_14_2(SH2_struct* sh)
{
  SH2movwi(sh, 8, (14<<4)+2);
}
#define init_9_8_14_3
void FASTCALL call_9_8_14_3(SH2_struct* sh)
{
  SH2movwi(sh, 8, (14<<4)+3);
}
#define init_9_8_14_4
void FASTCALL call_9_8_14_4(SH2_struct* sh)
{
  SH2movwi(sh, 8, (14<<4)+4);
}
#define init_9_8_14_5
void FASTCALL call_9_8_14_5(SH2_struct* sh)
{
  SH2movwi(sh, 8, (14<<4)+5);
}
#define init_9_8_14_6
void FASTCALL call_9_8_14_6(SH2_struct* sh)
{
  SH2movwi(sh, 8, (14<<4)+6);
}
#define init_9_8_14_7
void FASTCALL call_9_8_14_7(SH2_struct* sh)
{
  SH2movwi(sh, 8, (14<<4)+7);
}
#define init_9_8_14_8
void FASTCALL call_9_8_14_8(SH2_struct* sh)
{
  SH2movwi(sh, 8, (14<<4)+8);
}
#define init_9_8_14_9
void FASTCALL call_9_8_14_9(SH2_struct* sh)
{
  SH2movwi(sh, 8, (14<<4)+9);
}
#define init_9_8_14_10
void FASTCALL call_9_8_14_10(SH2_struct* sh)
{
  SH2movwi(sh, 8, (14<<4)+10);
}
#define init_9_8_14_11
void FASTCALL call_9_8_14_11(SH2_struct* sh)
{
  SH2movwi(sh, 8, (14<<4)+11);
}
#define init_9_8_14_12
void FASTCALL call_9_8_14_12(SH2_struct* sh)
{
  SH2movwi(sh, 8, (14<<4)+12);
}
#define init_9_8_14_13
void FASTCALL call_9_8_14_13(SH2_struct* sh)
{
  SH2movwi(sh, 8, (14<<4)+13);
}
#define init_9_8_14_14
void FASTCALL call_9_8_14_14(SH2_struct* sh)
{
  SH2movwi(sh, 8, (14<<4)+14);
}
#define init_9_8_14_15
void FASTCALL call_9_8_14_15(SH2_struct* sh)
{
  SH2movwi(sh, 8, (14<<4)+15);
}
#define init_9_8_15_0
void FASTCALL call_9_8_15_0(SH2_struct* sh)
{
  SH2movwi(sh, 8, (15<<4)+0);
}
#define init_9_8_15_1
void FASTCALL call_9_8_15_1(SH2_struct* sh)
{
  SH2movwi(sh, 8, (15<<4)+1);
}
#define init_9_8_15_2
void FASTCALL call_9_8_15_2(SH2_struct* sh)
{
  SH2movwi(sh, 8, (15<<4)+2);
}
#define init_9_8_15_3
void FASTCALL call_9_8_15_3(SH2_struct* sh)
{
  SH2movwi(sh, 8, (15<<4)+3);
}
#define init_9_8_15_4
void FASTCALL call_9_8_15_4(SH2_struct* sh)
{
  SH2movwi(sh, 8, (15<<4)+4);
}
#define init_9_8_15_5
void FASTCALL call_9_8_15_5(SH2_struct* sh)
{
  SH2movwi(sh, 8, (15<<4)+5);
}
#define init_9_8_15_6
void FASTCALL call_9_8_15_6(SH2_struct* sh)
{
  SH2movwi(sh, 8, (15<<4)+6);
}
#define init_9_8_15_7
void FASTCALL call_9_8_15_7(SH2_struct* sh)
{
  SH2movwi(sh, 8, (15<<4)+7);
}
#define init_9_8_15_8
void FASTCALL call_9_8_15_8(SH2_struct* sh)
{
  SH2movwi(sh, 8, (15<<4)+8);
}
#define init_9_8_15_9
void FASTCALL call_9_8_15_9(SH2_struct* sh)
{
  SH2movwi(sh, 8, (15<<4)+9);
}
#define init_9_8_15_10
void FASTCALL call_9_8_15_10(SH2_struct* sh)
{
  SH2movwi(sh, 8, (15<<4)+10);
}
#define init_9_8_15_11
void FASTCALL call_9_8_15_11(SH2_struct* sh)
{
  SH2movwi(sh, 8, (15<<4)+11);
}
#define init_9_8_15_12
void FASTCALL call_9_8_15_12(SH2_struct* sh)
{
  SH2movwi(sh, 8, (15<<4)+12);
}
#define init_9_8_15_13
void FASTCALL call_9_8_15_13(SH2_struct* sh)
{
  SH2movwi(sh, 8, (15<<4)+13);
}
#define init_9_8_15_14
void FASTCALL call_9_8_15_14(SH2_struct* sh)
{
  SH2movwi(sh, 8, (15<<4)+14);
}
#define init_9_8_15_15
void FASTCALL call_9_8_15_15(SH2_struct* sh)
{
  SH2movwi(sh, 8, (15<<4)+15);
}
#define init_9_9_0_0
void FASTCALL call_9_9_0_0(SH2_struct* sh)
{
  SH2movwi(sh, 9, (0<<4)+0);
}
#define init_9_9_0_1
void FASTCALL call_9_9_0_1(SH2_struct* sh)
{
  SH2movwi(sh, 9, (0<<4)+1);
}
#define init_9_9_0_2
void FASTCALL call_9_9_0_2(SH2_struct* sh)
{
  SH2movwi(sh, 9, (0<<4)+2);
}
#define init_9_9_0_3
void FASTCALL call_9_9_0_3(SH2_struct* sh)
{
  SH2movwi(sh, 9, (0<<4)+3);
}
#define init_9_9_0_4
void FASTCALL call_9_9_0_4(SH2_struct* sh)
{
  SH2movwi(sh, 9, (0<<4)+4);
}
#define init_9_9_0_5
void FASTCALL call_9_9_0_5(SH2_struct* sh)
{
  SH2movwi(sh, 9, (0<<4)+5);
}
#define init_9_9_0_6
void FASTCALL call_9_9_0_6(SH2_struct* sh)
{
  SH2movwi(sh, 9, (0<<4)+6);
}
#define init_9_9_0_7
void FASTCALL call_9_9_0_7(SH2_struct* sh)
{
  SH2movwi(sh, 9, (0<<4)+7);
}
#define init_9_9_0_8
void FASTCALL call_9_9_0_8(SH2_struct* sh)
{
  SH2movwi(sh, 9, (0<<4)+8);
}
#define init_9_9_0_9
void FASTCALL call_9_9_0_9(SH2_struct* sh)
{
  SH2movwi(sh, 9, (0<<4)+9);
}
#define init_9_9_0_10
void FASTCALL call_9_9_0_10(SH2_struct* sh)
{
  SH2movwi(sh, 9, (0<<4)+10);
}
#define init_9_9_0_11
void FASTCALL call_9_9_0_11(SH2_struct* sh)
{
  SH2movwi(sh, 9, (0<<4)+11);
}
#define init_9_9_0_12
void FASTCALL call_9_9_0_12(SH2_struct* sh)
{
  SH2movwi(sh, 9, (0<<4)+12);
}
#define init_9_9_0_13
void FASTCALL call_9_9_0_13(SH2_struct* sh)
{
  SH2movwi(sh, 9, (0<<4)+13);
}
#define init_9_9_0_14
void FASTCALL call_9_9_0_14(SH2_struct* sh)
{
  SH2movwi(sh, 9, (0<<4)+14);
}
#define init_9_9_0_15
void FASTCALL call_9_9_0_15(SH2_struct* sh)
{
  SH2movwi(sh, 9, (0<<4)+15);
}
#define init_9_9_1_0
void FASTCALL call_9_9_1_0(SH2_struct* sh)
{
  SH2movwi(sh, 9, (1<<4)+0);
}
#define init_9_9_1_1
void FASTCALL call_9_9_1_1(SH2_struct* sh)
{
  SH2movwi(sh, 9, (1<<4)+1);
}
#define init_9_9_1_2
void FASTCALL call_9_9_1_2(SH2_struct* sh)
{
  SH2movwi(sh, 9, (1<<4)+2);
}
#define init_9_9_1_3
void FASTCALL call_9_9_1_3(SH2_struct* sh)
{
  SH2movwi(sh, 9, (1<<4)+3);
}
#define init_9_9_1_4
void FASTCALL call_9_9_1_4(SH2_struct* sh)
{
  SH2movwi(sh, 9, (1<<4)+4);
}
#define init_9_9_1_5
void FASTCALL call_9_9_1_5(SH2_struct* sh)
{
  SH2movwi(sh, 9, (1<<4)+5);
}
#define init_9_9_1_6
void FASTCALL call_9_9_1_6(SH2_struct* sh)
{
  SH2movwi(sh, 9, (1<<4)+6);
}
#define init_9_9_1_7
void FASTCALL call_9_9_1_7(SH2_struct* sh)
{
  SH2movwi(sh, 9, (1<<4)+7);
}
#define init_9_9_1_8
void FASTCALL call_9_9_1_8(SH2_struct* sh)
{
  SH2movwi(sh, 9, (1<<4)+8);
}
#define init_9_9_1_9
void FASTCALL call_9_9_1_9(SH2_struct* sh)
{
  SH2movwi(sh, 9, (1<<4)+9);
}
#define init_9_9_1_10
void FASTCALL call_9_9_1_10(SH2_struct* sh)
{
  SH2movwi(sh, 9, (1<<4)+10);
}
#define init_9_9_1_11
void FASTCALL call_9_9_1_11(SH2_struct* sh)
{
  SH2movwi(sh, 9, (1<<4)+11);
}
#define init_9_9_1_12
void FASTCALL call_9_9_1_12(SH2_struct* sh)
{
  SH2movwi(sh, 9, (1<<4)+12);
}
#define init_9_9_1_13
void FASTCALL call_9_9_1_13(SH2_struct* sh)
{
  SH2movwi(sh, 9, (1<<4)+13);
}
#define init_9_9_1_14
void FASTCALL call_9_9_1_14(SH2_struct* sh)
{
  SH2movwi(sh, 9, (1<<4)+14);
}
#define init_9_9_1_15
void FASTCALL call_9_9_1_15(SH2_struct* sh)
{
  SH2movwi(sh, 9, (1<<4)+15);
}
#define init_9_9_2_0
void FASTCALL call_9_9_2_0(SH2_struct* sh)
{
  SH2movwi(sh, 9, (2<<4)+0);
}
#define init_9_9_2_1
void FASTCALL call_9_9_2_1(SH2_struct* sh)
{
  SH2movwi(sh, 9, (2<<4)+1);
}
#define init_9_9_2_2
void FASTCALL call_9_9_2_2(SH2_struct* sh)
{
  SH2movwi(sh, 9, (2<<4)+2);
}
#define init_9_9_2_3
void FASTCALL call_9_9_2_3(SH2_struct* sh)
{
  SH2movwi(sh, 9, (2<<4)+3);
}
#define init_9_9_2_4
void FASTCALL call_9_9_2_4(SH2_struct* sh)
{
  SH2movwi(sh, 9, (2<<4)+4);
}
#define init_9_9_2_5
void FASTCALL call_9_9_2_5(SH2_struct* sh)
{
  SH2movwi(sh, 9, (2<<4)+5);
}
#define init_9_9_2_6
void FASTCALL call_9_9_2_6(SH2_struct* sh)
{
  SH2movwi(sh, 9, (2<<4)+6);
}
#define init_9_9_2_7
void FASTCALL call_9_9_2_7(SH2_struct* sh)
{
  SH2movwi(sh, 9, (2<<4)+7);
}
#define init_9_9_2_8
void FASTCALL call_9_9_2_8(SH2_struct* sh)
{
  SH2movwi(sh, 9, (2<<4)+8);
}
#define init_9_9_2_9
void FASTCALL call_9_9_2_9(SH2_struct* sh)
{
  SH2movwi(sh, 9, (2<<4)+9);
}
#define init_9_9_2_10
void FASTCALL call_9_9_2_10(SH2_struct* sh)
{
  SH2movwi(sh, 9, (2<<4)+10);
}
#define init_9_9_2_11
void FASTCALL call_9_9_2_11(SH2_struct* sh)
{
  SH2movwi(sh, 9, (2<<4)+11);
}
#define init_9_9_2_12
void FASTCALL call_9_9_2_12(SH2_struct* sh)
{
  SH2movwi(sh, 9, (2<<4)+12);
}
#define init_9_9_2_13
void FASTCALL call_9_9_2_13(SH2_struct* sh)
{
  SH2movwi(sh, 9, (2<<4)+13);
}
#define init_9_9_2_14
void FASTCALL call_9_9_2_14(SH2_struct* sh)
{
  SH2movwi(sh, 9, (2<<4)+14);
}
#define init_9_9_2_15
void FASTCALL call_9_9_2_15(SH2_struct* sh)
{
  SH2movwi(sh, 9, (2<<4)+15);
}
#define init_9_9_3_0
void FASTCALL call_9_9_3_0(SH2_struct* sh)
{
  SH2movwi(sh, 9, (3<<4)+0);
}
#define init_9_9_3_1
void FASTCALL call_9_9_3_1(SH2_struct* sh)
{
  SH2movwi(sh, 9, (3<<4)+1);
}
#define init_9_9_3_2
void FASTCALL call_9_9_3_2(SH2_struct* sh)
{
  SH2movwi(sh, 9, (3<<4)+2);
}
#define init_9_9_3_3
void FASTCALL call_9_9_3_3(SH2_struct* sh)
{
  SH2movwi(sh, 9, (3<<4)+3);
}
#define init_9_9_3_4
void FASTCALL call_9_9_3_4(SH2_struct* sh)
{
  SH2movwi(sh, 9, (3<<4)+4);
}
#define init_9_9_3_5
void FASTCALL call_9_9_3_5(SH2_struct* sh)
{
  SH2movwi(sh, 9, (3<<4)+5);
}
#define init_9_9_3_6
void FASTCALL call_9_9_3_6(SH2_struct* sh)
{
  SH2movwi(sh, 9, (3<<4)+6);
}
#define init_9_9_3_7
void FASTCALL call_9_9_3_7(SH2_struct* sh)
{
  SH2movwi(sh, 9, (3<<4)+7);
}
#define init_9_9_3_8
void FASTCALL call_9_9_3_8(SH2_struct* sh)
{
  SH2movwi(sh, 9, (3<<4)+8);
}
#define init_9_9_3_9
void FASTCALL call_9_9_3_9(SH2_struct* sh)
{
  SH2movwi(sh, 9, (3<<4)+9);
}
#define init_9_9_3_10
void FASTCALL call_9_9_3_10(SH2_struct* sh)
{
  SH2movwi(sh, 9, (3<<4)+10);
}
#define init_9_9_3_11
void FASTCALL call_9_9_3_11(SH2_struct* sh)
{
  SH2movwi(sh, 9, (3<<4)+11);
}
#define init_9_9_3_12
void FASTCALL call_9_9_3_12(SH2_struct* sh)
{
  SH2movwi(sh, 9, (3<<4)+12);
}
#define init_9_9_3_13
void FASTCALL call_9_9_3_13(SH2_struct* sh)
{
  SH2movwi(sh, 9, (3<<4)+13);
}
#define init_9_9_3_14
void FASTCALL call_9_9_3_14(SH2_struct* sh)
{
  SH2movwi(sh, 9, (3<<4)+14);
}
#define init_9_9_3_15
void FASTCALL call_9_9_3_15(SH2_struct* sh)
{
  SH2movwi(sh, 9, (3<<4)+15);
}
#define init_9_9_4_0
void FASTCALL call_9_9_4_0(SH2_struct* sh)
{
  SH2movwi(sh, 9, (4<<4)+0);
}
#define init_9_9_4_1
void FASTCALL call_9_9_4_1(SH2_struct* sh)
{
  SH2movwi(sh, 9, (4<<4)+1);
}
#define init_9_9_4_2
void FASTCALL call_9_9_4_2(SH2_struct* sh)
{
  SH2movwi(sh, 9, (4<<4)+2);
}
#define init_9_9_4_3
void FASTCALL call_9_9_4_3(SH2_struct* sh)
{
  SH2movwi(sh, 9, (4<<4)+3);
}
#define init_9_9_4_4
void FASTCALL call_9_9_4_4(SH2_struct* sh)
{
  SH2movwi(sh, 9, (4<<4)+4);
}
#define init_9_9_4_5
void FASTCALL call_9_9_4_5(SH2_struct* sh)
{
  SH2movwi(sh, 9, (4<<4)+5);
}
#define init_9_9_4_6
void FASTCALL call_9_9_4_6(SH2_struct* sh)
{
  SH2movwi(sh, 9, (4<<4)+6);
}
#define init_9_9_4_7
void FASTCALL call_9_9_4_7(SH2_struct* sh)
{
  SH2movwi(sh, 9, (4<<4)+7);
}
#define init_9_9_4_8
void FASTCALL call_9_9_4_8(SH2_struct* sh)
{
  SH2movwi(sh, 9, (4<<4)+8);
}
#define init_9_9_4_9
void FASTCALL call_9_9_4_9(SH2_struct* sh)
{
  SH2movwi(sh, 9, (4<<4)+9);
}
#define init_9_9_4_10
void FASTCALL call_9_9_4_10(SH2_struct* sh)
{
  SH2movwi(sh, 9, (4<<4)+10);
}
#define init_9_9_4_11
void FASTCALL call_9_9_4_11(SH2_struct* sh)
{
  SH2movwi(sh, 9, (4<<4)+11);
}
#define init_9_9_4_12
void FASTCALL call_9_9_4_12(SH2_struct* sh)
{
  SH2movwi(sh, 9, (4<<4)+12);
}
#define init_9_9_4_13
void FASTCALL call_9_9_4_13(SH2_struct* sh)
{
  SH2movwi(sh, 9, (4<<4)+13);
}
#define init_9_9_4_14
void FASTCALL call_9_9_4_14(SH2_struct* sh)
{
  SH2movwi(sh, 9, (4<<4)+14);
}
#define init_9_9_4_15
void FASTCALL call_9_9_4_15(SH2_struct* sh)
{
  SH2movwi(sh, 9, (4<<4)+15);
}
#define init_9_9_5_0
void FASTCALL call_9_9_5_0(SH2_struct* sh)
{
  SH2movwi(sh, 9, (5<<4)+0);
}
#define init_9_9_5_1
void FASTCALL call_9_9_5_1(SH2_struct* sh)
{
  SH2movwi(sh, 9, (5<<4)+1);
}
#define init_9_9_5_2
void FASTCALL call_9_9_5_2(SH2_struct* sh)
{
  SH2movwi(sh, 9, (5<<4)+2);
}
#define init_9_9_5_3
void FASTCALL call_9_9_5_3(SH2_struct* sh)
{
  SH2movwi(sh, 9, (5<<4)+3);
}
#define init_9_9_5_4
void FASTCALL call_9_9_5_4(SH2_struct* sh)
{
  SH2movwi(sh, 9, (5<<4)+4);
}
#define init_9_9_5_5
void FASTCALL call_9_9_5_5(SH2_struct* sh)
{
  SH2movwi(sh, 9, (5<<4)+5);
}
#define init_9_9_5_6
void FASTCALL call_9_9_5_6(SH2_struct* sh)
{
  SH2movwi(sh, 9, (5<<4)+6);
}
#define init_9_9_5_7
void FASTCALL call_9_9_5_7(SH2_struct* sh)
{
  SH2movwi(sh, 9, (5<<4)+7);
}
#define init_9_9_5_8
void FASTCALL call_9_9_5_8(SH2_struct* sh)
{
  SH2movwi(sh, 9, (5<<4)+8);
}
#define init_9_9_5_9
void FASTCALL call_9_9_5_9(SH2_struct* sh)
{
  SH2movwi(sh, 9, (5<<4)+9);
}
#define init_9_9_5_10
void FASTCALL call_9_9_5_10(SH2_struct* sh)
{
  SH2movwi(sh, 9, (5<<4)+10);
}
#define init_9_9_5_11
void FASTCALL call_9_9_5_11(SH2_struct* sh)
{
  SH2movwi(sh, 9, (5<<4)+11);
}
#define init_9_9_5_12
void FASTCALL call_9_9_5_12(SH2_struct* sh)
{
  SH2movwi(sh, 9, (5<<4)+12);
}
#define init_9_9_5_13
void FASTCALL call_9_9_5_13(SH2_struct* sh)
{
  SH2movwi(sh, 9, (5<<4)+13);
}
#define init_9_9_5_14
void FASTCALL call_9_9_5_14(SH2_struct* sh)
{
  SH2movwi(sh, 9, (5<<4)+14);
}
#define init_9_9_5_15
void FASTCALL call_9_9_5_15(SH2_struct* sh)
{
  SH2movwi(sh, 9, (5<<4)+15);
}
#define init_9_9_6_0
void FASTCALL call_9_9_6_0(SH2_struct* sh)
{
  SH2movwi(sh, 9, (6<<4)+0);
}
#define init_9_9_6_1
void FASTCALL call_9_9_6_1(SH2_struct* sh)
{
  SH2movwi(sh, 9, (6<<4)+1);
}
#define init_9_9_6_2
void FASTCALL call_9_9_6_2(SH2_struct* sh)
{
  SH2movwi(sh, 9, (6<<4)+2);
}
#define init_9_9_6_3
void FASTCALL call_9_9_6_3(SH2_struct* sh)
{
  SH2movwi(sh, 9, (6<<4)+3);
}
#define init_9_9_6_4
void FASTCALL call_9_9_6_4(SH2_struct* sh)
{
  SH2movwi(sh, 9, (6<<4)+4);
}
#define init_9_9_6_5
void FASTCALL call_9_9_6_5(SH2_struct* sh)
{
  SH2movwi(sh, 9, (6<<4)+5);
}
#define init_9_9_6_6
void FASTCALL call_9_9_6_6(SH2_struct* sh)
{
  SH2movwi(sh, 9, (6<<4)+6);
}
#define init_9_9_6_7
void FASTCALL call_9_9_6_7(SH2_struct* sh)
{
  SH2movwi(sh, 9, (6<<4)+7);
}
#define init_9_9_6_8
void FASTCALL call_9_9_6_8(SH2_struct* sh)
{
  SH2movwi(sh, 9, (6<<4)+8);
}
#define init_9_9_6_9
void FASTCALL call_9_9_6_9(SH2_struct* sh)
{
  SH2movwi(sh, 9, (6<<4)+9);
}
#define init_9_9_6_10
void FASTCALL call_9_9_6_10(SH2_struct* sh)
{
  SH2movwi(sh, 9, (6<<4)+10);
}
#define init_9_9_6_11
void FASTCALL call_9_9_6_11(SH2_struct* sh)
{
  SH2movwi(sh, 9, (6<<4)+11);
}
#define init_9_9_6_12
void FASTCALL call_9_9_6_12(SH2_struct* sh)
{
  SH2movwi(sh, 9, (6<<4)+12);
}
#define init_9_9_6_13
void FASTCALL call_9_9_6_13(SH2_struct* sh)
{
  SH2movwi(sh, 9, (6<<4)+13);
}
#define init_9_9_6_14
void FASTCALL call_9_9_6_14(SH2_struct* sh)
{
  SH2movwi(sh, 9, (6<<4)+14);
}
#define init_9_9_6_15
void FASTCALL call_9_9_6_15(SH2_struct* sh)
{
  SH2movwi(sh, 9, (6<<4)+15);
}
#define init_9_9_7_0
void FASTCALL call_9_9_7_0(SH2_struct* sh)
{
  SH2movwi(sh, 9, (7<<4)+0);
}
#define init_9_9_7_1
void FASTCALL call_9_9_7_1(SH2_struct* sh)
{
  SH2movwi(sh, 9, (7<<4)+1);
}
#define init_9_9_7_2
void FASTCALL call_9_9_7_2(SH2_struct* sh)
{
  SH2movwi(sh, 9, (7<<4)+2);
}
#define init_9_9_7_3
void FASTCALL call_9_9_7_3(SH2_struct* sh)
{
  SH2movwi(sh, 9, (7<<4)+3);
}
#define init_9_9_7_4
void FASTCALL call_9_9_7_4(SH2_struct* sh)
{
  SH2movwi(sh, 9, (7<<4)+4);
}
#define init_9_9_7_5
void FASTCALL call_9_9_7_5(SH2_struct* sh)
{
  SH2movwi(sh, 9, (7<<4)+5);
}
#define init_9_9_7_6
void FASTCALL call_9_9_7_6(SH2_struct* sh)
{
  SH2movwi(sh, 9, (7<<4)+6);
}
#define init_9_9_7_7
void FASTCALL call_9_9_7_7(SH2_struct* sh)
{
  SH2movwi(sh, 9, (7<<4)+7);
}
#define init_9_9_7_8
void FASTCALL call_9_9_7_8(SH2_struct* sh)
{
  SH2movwi(sh, 9, (7<<4)+8);
}
#define init_9_9_7_9
void FASTCALL call_9_9_7_9(SH2_struct* sh)
{
  SH2movwi(sh, 9, (7<<4)+9);
}
#define init_9_9_7_10
void FASTCALL call_9_9_7_10(SH2_struct* sh)
{
  SH2movwi(sh, 9, (7<<4)+10);
}
#define init_9_9_7_11
void FASTCALL call_9_9_7_11(SH2_struct* sh)
{
  SH2movwi(sh, 9, (7<<4)+11);
}
#define init_9_9_7_12
void FASTCALL call_9_9_7_12(SH2_struct* sh)
{
  SH2movwi(sh, 9, (7<<4)+12);
}
#define init_9_9_7_13
void FASTCALL call_9_9_7_13(SH2_struct* sh)
{
  SH2movwi(sh, 9, (7<<4)+13);
}
#define init_9_9_7_14
void FASTCALL call_9_9_7_14(SH2_struct* sh)
{
  SH2movwi(sh, 9, (7<<4)+14);
}
#define init_9_9_7_15
void FASTCALL call_9_9_7_15(SH2_struct* sh)
{
  SH2movwi(sh, 9, (7<<4)+15);
}
#define init_9_9_8_0
void FASTCALL call_9_9_8_0(SH2_struct* sh)
{
  SH2movwi(sh, 9, (8<<4)+0);
}
#define init_9_9_8_1
void FASTCALL call_9_9_8_1(SH2_struct* sh)
{
  SH2movwi(sh, 9, (8<<4)+1);
}
#define init_9_9_8_2
void FASTCALL call_9_9_8_2(SH2_struct* sh)
{
  SH2movwi(sh, 9, (8<<4)+2);
}
#define init_9_9_8_3
void FASTCALL call_9_9_8_3(SH2_struct* sh)
{
  SH2movwi(sh, 9, (8<<4)+3);
}
#define init_9_9_8_4
void FASTCALL call_9_9_8_4(SH2_struct* sh)
{
  SH2movwi(sh, 9, (8<<4)+4);
}
#define init_9_9_8_5
void FASTCALL call_9_9_8_5(SH2_struct* sh)
{
  SH2movwi(sh, 9, (8<<4)+5);
}
#define init_9_9_8_6
void FASTCALL call_9_9_8_6(SH2_struct* sh)
{
  SH2movwi(sh, 9, (8<<4)+6);
}
#define init_9_9_8_7
void FASTCALL call_9_9_8_7(SH2_struct* sh)
{
  SH2movwi(sh, 9, (8<<4)+7);
}
#define init_9_9_8_8
void FASTCALL call_9_9_8_8(SH2_struct* sh)
{
  SH2movwi(sh, 9, (8<<4)+8);
}
#define init_9_9_8_9
void FASTCALL call_9_9_8_9(SH2_struct* sh)
{
  SH2movwi(sh, 9, (8<<4)+9);
}
#define init_9_9_8_10
void FASTCALL call_9_9_8_10(SH2_struct* sh)
{
  SH2movwi(sh, 9, (8<<4)+10);
}
#define init_9_9_8_11
void FASTCALL call_9_9_8_11(SH2_struct* sh)
{
  SH2movwi(sh, 9, (8<<4)+11);
}
#define init_9_9_8_12
void FASTCALL call_9_9_8_12(SH2_struct* sh)
{
  SH2movwi(sh, 9, (8<<4)+12);
}
#define init_9_9_8_13
void FASTCALL call_9_9_8_13(SH2_struct* sh)
{
  SH2movwi(sh, 9, (8<<4)+13);
}
#define init_9_9_8_14
void FASTCALL call_9_9_8_14(SH2_struct* sh)
{
  SH2movwi(sh, 9, (8<<4)+14);
}
#define init_9_9_8_15
void FASTCALL call_9_9_8_15(SH2_struct* sh)
{
  SH2movwi(sh, 9, (8<<4)+15);
}
#define init_9_9_9_0
void FASTCALL call_9_9_9_0(SH2_struct* sh)
{
  SH2movwi(sh, 9, (9<<4)+0);
}
#define init_9_9_9_1
void FASTCALL call_9_9_9_1(SH2_struct* sh)
{
  SH2movwi(sh, 9, (9<<4)+1);
}
#define init_9_9_9_2
void FASTCALL call_9_9_9_2(SH2_struct* sh)
{
  SH2movwi(sh, 9, (9<<4)+2);
}
#define init_9_9_9_3
void FASTCALL call_9_9_9_3(SH2_struct* sh)
{
  SH2movwi(sh, 9, (9<<4)+3);
}
#define init_9_9_9_4
void FASTCALL call_9_9_9_4(SH2_struct* sh)
{
  SH2movwi(sh, 9, (9<<4)+4);
}
#define init_9_9_9_5
void FASTCALL call_9_9_9_5(SH2_struct* sh)
{
  SH2movwi(sh, 9, (9<<4)+5);
}
#define init_9_9_9_6
void FASTCALL call_9_9_9_6(SH2_struct* sh)
{
  SH2movwi(sh, 9, (9<<4)+6);
}
#define init_9_9_9_7
void FASTCALL call_9_9_9_7(SH2_struct* sh)
{
  SH2movwi(sh, 9, (9<<4)+7);
}
#define init_9_9_9_8
void FASTCALL call_9_9_9_8(SH2_struct* sh)
{
  SH2movwi(sh, 9, (9<<4)+8);
}
#define init_9_9_9_9
void FASTCALL call_9_9_9_9(SH2_struct* sh)
{
  SH2movwi(sh, 9, (9<<4)+9);
}
#define init_9_9_9_10
void FASTCALL call_9_9_9_10(SH2_struct* sh)
{
  SH2movwi(sh, 9, (9<<4)+10);
}
#define init_9_9_9_11
void FASTCALL call_9_9_9_11(SH2_struct* sh)
{
  SH2movwi(sh, 9, (9<<4)+11);
}
#define init_9_9_9_12
void FASTCALL call_9_9_9_12(SH2_struct* sh)
{
  SH2movwi(sh, 9, (9<<4)+12);
}
#define init_9_9_9_13
void FASTCALL call_9_9_9_13(SH2_struct* sh)
{
  SH2movwi(sh, 9, (9<<4)+13);
}
#define init_9_9_9_14
void FASTCALL call_9_9_9_14(SH2_struct* sh)
{
  SH2movwi(sh, 9, (9<<4)+14);
}
#define init_9_9_9_15
void FASTCALL call_9_9_9_15(SH2_struct* sh)
{
  SH2movwi(sh, 9, (9<<4)+15);
}
#define init_9_9_10_0
void FASTCALL call_9_9_10_0(SH2_struct* sh)
{
  SH2movwi(sh, 9, (10<<4)+0);
}
#define init_9_9_10_1
void FASTCALL call_9_9_10_1(SH2_struct* sh)
{
  SH2movwi(sh, 9, (10<<4)+1);
}
#define init_9_9_10_2
void FASTCALL call_9_9_10_2(SH2_struct* sh)
{
  SH2movwi(sh, 9, (10<<4)+2);
}
#define init_9_9_10_3
void FASTCALL call_9_9_10_3(SH2_struct* sh)
{
  SH2movwi(sh, 9, (10<<4)+3);
}
#define init_9_9_10_4
void FASTCALL call_9_9_10_4(SH2_struct* sh)
{
  SH2movwi(sh, 9, (10<<4)+4);
}
#define init_9_9_10_5
void FASTCALL call_9_9_10_5(SH2_struct* sh)
{
  SH2movwi(sh, 9, (10<<4)+5);
}
#define init_9_9_10_6
void FASTCALL call_9_9_10_6(SH2_struct* sh)
{
  SH2movwi(sh, 9, (10<<4)+6);
}
#define init_9_9_10_7
void FASTCALL call_9_9_10_7(SH2_struct* sh)
{
  SH2movwi(sh, 9, (10<<4)+7);
}
#define init_9_9_10_8
void FASTCALL call_9_9_10_8(SH2_struct* sh)
{
  SH2movwi(sh, 9, (10<<4)+8);
}
#define init_9_9_10_9
void FASTCALL call_9_9_10_9(SH2_struct* sh)
{
  SH2movwi(sh, 9, (10<<4)+9);
}
#define init_9_9_10_10
void FASTCALL call_9_9_10_10(SH2_struct* sh)
{
  SH2movwi(sh, 9, (10<<4)+10);
}
#define init_9_9_10_11
void FASTCALL call_9_9_10_11(SH2_struct* sh)
{
  SH2movwi(sh, 9, (10<<4)+11);
}
#define init_9_9_10_12
void FASTCALL call_9_9_10_12(SH2_struct* sh)
{
  SH2movwi(sh, 9, (10<<4)+12);
}
#define init_9_9_10_13
void FASTCALL call_9_9_10_13(SH2_struct* sh)
{
  SH2movwi(sh, 9, (10<<4)+13);
}
#define init_9_9_10_14
void FASTCALL call_9_9_10_14(SH2_struct* sh)
{
  SH2movwi(sh, 9, (10<<4)+14);
}
#define init_9_9_10_15
void FASTCALL call_9_9_10_15(SH2_struct* sh)
{
  SH2movwi(sh, 9, (10<<4)+15);
}
#define init_9_9_11_0
void FASTCALL call_9_9_11_0(SH2_struct* sh)
{
  SH2movwi(sh, 9, (11<<4)+0);
}
#define init_9_9_11_1
void FASTCALL call_9_9_11_1(SH2_struct* sh)
{
  SH2movwi(sh, 9, (11<<4)+1);
}
#define init_9_9_11_2
void FASTCALL call_9_9_11_2(SH2_struct* sh)
{
  SH2movwi(sh, 9, (11<<4)+2);
}
#define init_9_9_11_3
void FASTCALL call_9_9_11_3(SH2_struct* sh)
{
  SH2movwi(sh, 9, (11<<4)+3);
}
#define init_9_9_11_4
void FASTCALL call_9_9_11_4(SH2_struct* sh)
{
  SH2movwi(sh, 9, (11<<4)+4);
}
#define init_9_9_11_5
void FASTCALL call_9_9_11_5(SH2_struct* sh)
{
  SH2movwi(sh, 9, (11<<4)+5);
}
#define init_9_9_11_6
void FASTCALL call_9_9_11_6(SH2_struct* sh)
{
  SH2movwi(sh, 9, (11<<4)+6);
}
#define init_9_9_11_7
void FASTCALL call_9_9_11_7(SH2_struct* sh)
{
  SH2movwi(sh, 9, (11<<4)+7);
}
#define init_9_9_11_8
void FASTCALL call_9_9_11_8(SH2_struct* sh)
{
  SH2movwi(sh, 9, (11<<4)+8);
}
#define init_9_9_11_9
void FASTCALL call_9_9_11_9(SH2_struct* sh)
{
  SH2movwi(sh, 9, (11<<4)+9);
}
#define init_9_9_11_10
void FASTCALL call_9_9_11_10(SH2_struct* sh)
{
  SH2movwi(sh, 9, (11<<4)+10);
}
#define init_9_9_11_11
void FASTCALL call_9_9_11_11(SH2_struct* sh)
{
  SH2movwi(sh, 9, (11<<4)+11);
}
#define init_9_9_11_12
void FASTCALL call_9_9_11_12(SH2_struct* sh)
{
  SH2movwi(sh, 9, (11<<4)+12);
}
#define init_9_9_11_13
void FASTCALL call_9_9_11_13(SH2_struct* sh)
{
  SH2movwi(sh, 9, (11<<4)+13);
}
#define init_9_9_11_14
void FASTCALL call_9_9_11_14(SH2_struct* sh)
{
  SH2movwi(sh, 9, (11<<4)+14);
}
#define init_9_9_11_15
void FASTCALL call_9_9_11_15(SH2_struct* sh)
{
  SH2movwi(sh, 9, (11<<4)+15);
}
#define init_9_9_12_0
void FASTCALL call_9_9_12_0(SH2_struct* sh)
{
  SH2movwi(sh, 9, (12<<4)+0);
}
#define init_9_9_12_1
void FASTCALL call_9_9_12_1(SH2_struct* sh)
{
  SH2movwi(sh, 9, (12<<4)+1);
}
#define init_9_9_12_2
void FASTCALL call_9_9_12_2(SH2_struct* sh)
{
  SH2movwi(sh, 9, (12<<4)+2);
}
#define init_9_9_12_3
void FASTCALL call_9_9_12_3(SH2_struct* sh)
{
  SH2movwi(sh, 9, (12<<4)+3);
}
#define init_9_9_12_4
void FASTCALL call_9_9_12_4(SH2_struct* sh)
{
  SH2movwi(sh, 9, (12<<4)+4);
}
#define init_9_9_12_5
void FASTCALL call_9_9_12_5(SH2_struct* sh)
{
  SH2movwi(sh, 9, (12<<4)+5);
}
#define init_9_9_12_6
void FASTCALL call_9_9_12_6(SH2_struct* sh)
{
  SH2movwi(sh, 9, (12<<4)+6);
}
#define init_9_9_12_7
void FASTCALL call_9_9_12_7(SH2_struct* sh)
{
  SH2movwi(sh, 9, (12<<4)+7);
}
#define init_9_9_12_8
void FASTCALL call_9_9_12_8(SH2_struct* sh)
{
  SH2movwi(sh, 9, (12<<4)+8);
}
#define init_9_9_12_9
void FASTCALL call_9_9_12_9(SH2_struct* sh)
{
  SH2movwi(sh, 9, (12<<4)+9);
}
#define init_9_9_12_10
void FASTCALL call_9_9_12_10(SH2_struct* sh)
{
  SH2movwi(sh, 9, (12<<4)+10);
}
#define init_9_9_12_11
void FASTCALL call_9_9_12_11(SH2_struct* sh)
{
  SH2movwi(sh, 9, (12<<4)+11);
}
#define init_9_9_12_12
void FASTCALL call_9_9_12_12(SH2_struct* sh)
{
  SH2movwi(sh, 9, (12<<4)+12);
}
#define init_9_9_12_13
void FASTCALL call_9_9_12_13(SH2_struct* sh)
{
  SH2movwi(sh, 9, (12<<4)+13);
}
#define init_9_9_12_14
void FASTCALL call_9_9_12_14(SH2_struct* sh)
{
  SH2movwi(sh, 9, (12<<4)+14);
}
#define init_9_9_12_15
void FASTCALL call_9_9_12_15(SH2_struct* sh)
{
  SH2movwi(sh, 9, (12<<4)+15);
}
#define init_9_9_13_0
void FASTCALL call_9_9_13_0(SH2_struct* sh)
{
  SH2movwi(sh, 9, (13<<4)+0);
}
#define init_9_9_13_1
void FASTCALL call_9_9_13_1(SH2_struct* sh)
{
  SH2movwi(sh, 9, (13<<4)+1);
}
#define init_9_9_13_2
void FASTCALL call_9_9_13_2(SH2_struct* sh)
{
  SH2movwi(sh, 9, (13<<4)+2);
}
#define init_9_9_13_3
void FASTCALL call_9_9_13_3(SH2_struct* sh)
{
  SH2movwi(sh, 9, (13<<4)+3);
}
#define init_9_9_13_4
void FASTCALL call_9_9_13_4(SH2_struct* sh)
{
  SH2movwi(sh, 9, (13<<4)+4);
}
#define init_9_9_13_5
void FASTCALL call_9_9_13_5(SH2_struct* sh)
{
  SH2movwi(sh, 9, (13<<4)+5);
}
#define init_9_9_13_6
void FASTCALL call_9_9_13_6(SH2_struct* sh)
{
  SH2movwi(sh, 9, (13<<4)+6);
}
#define init_9_9_13_7
void FASTCALL call_9_9_13_7(SH2_struct* sh)
{
  SH2movwi(sh, 9, (13<<4)+7);
}
#define init_9_9_13_8
void FASTCALL call_9_9_13_8(SH2_struct* sh)
{
  SH2movwi(sh, 9, (13<<4)+8);
}
#define init_9_9_13_9
void FASTCALL call_9_9_13_9(SH2_struct* sh)
{
  SH2movwi(sh, 9, (13<<4)+9);
}
#define init_9_9_13_10
void FASTCALL call_9_9_13_10(SH2_struct* sh)
{
  SH2movwi(sh, 9, (13<<4)+10);
}
#define init_9_9_13_11
void FASTCALL call_9_9_13_11(SH2_struct* sh)
{
  SH2movwi(sh, 9, (13<<4)+11);
}
#define init_9_9_13_12
void FASTCALL call_9_9_13_12(SH2_struct* sh)
{
  SH2movwi(sh, 9, (13<<4)+12);
}
#define init_9_9_13_13
void FASTCALL call_9_9_13_13(SH2_struct* sh)
{
  SH2movwi(sh, 9, (13<<4)+13);
}
#define init_9_9_13_14
void FASTCALL call_9_9_13_14(SH2_struct* sh)
{
  SH2movwi(sh, 9, (13<<4)+14);
}
#define init_9_9_13_15
void FASTCALL call_9_9_13_15(SH2_struct* sh)
{
  SH2movwi(sh, 9, (13<<4)+15);
}
#define init_9_9_14_0
void FASTCALL call_9_9_14_0(SH2_struct* sh)
{
  SH2movwi(sh, 9, (14<<4)+0);
}
#define init_9_9_14_1
void FASTCALL call_9_9_14_1(SH2_struct* sh)
{
  SH2movwi(sh, 9, (14<<4)+1);
}
#define init_9_9_14_2
void FASTCALL call_9_9_14_2(SH2_struct* sh)
{
  SH2movwi(sh, 9, (14<<4)+2);
}
#define init_9_9_14_3
void FASTCALL call_9_9_14_3(SH2_struct* sh)
{
  SH2movwi(sh, 9, (14<<4)+3);
}
#define init_9_9_14_4
void FASTCALL call_9_9_14_4(SH2_struct* sh)
{
  SH2movwi(sh, 9, (14<<4)+4);
}
#define init_9_9_14_5
void FASTCALL call_9_9_14_5(SH2_struct* sh)
{
  SH2movwi(sh, 9, (14<<4)+5);
}
#define init_9_9_14_6
void FASTCALL call_9_9_14_6(SH2_struct* sh)
{
  SH2movwi(sh, 9, (14<<4)+6);
}
#define init_9_9_14_7
void FASTCALL call_9_9_14_7(SH2_struct* sh)
{
  SH2movwi(sh, 9, (14<<4)+7);
}
#define init_9_9_14_8
void FASTCALL call_9_9_14_8(SH2_struct* sh)
{
  SH2movwi(sh, 9, (14<<4)+8);
}
#define init_9_9_14_9
void FASTCALL call_9_9_14_9(SH2_struct* sh)
{
  SH2movwi(sh, 9, (14<<4)+9);
}
#define init_9_9_14_10
void FASTCALL call_9_9_14_10(SH2_struct* sh)
{
  SH2movwi(sh, 9, (14<<4)+10);
}
#define init_9_9_14_11
void FASTCALL call_9_9_14_11(SH2_struct* sh)
{
  SH2movwi(sh, 9, (14<<4)+11);
}
#define init_9_9_14_12
void FASTCALL call_9_9_14_12(SH2_struct* sh)
{
  SH2movwi(sh, 9, (14<<4)+12);
}
#define init_9_9_14_13
void FASTCALL call_9_9_14_13(SH2_struct* sh)
{
  SH2movwi(sh, 9, (14<<4)+13);
}
#define init_9_9_14_14
void FASTCALL call_9_9_14_14(SH2_struct* sh)
{
  SH2movwi(sh, 9, (14<<4)+14);
}
#define init_9_9_14_15
void FASTCALL call_9_9_14_15(SH2_struct* sh)
{
  SH2movwi(sh, 9, (14<<4)+15);
}
#define init_9_9_15_0
void FASTCALL call_9_9_15_0(SH2_struct* sh)
{
  SH2movwi(sh, 9, (15<<4)+0);
}
#define init_9_9_15_1
void FASTCALL call_9_9_15_1(SH2_struct* sh)
{
  SH2movwi(sh, 9, (15<<4)+1);
}
#define init_9_9_15_2
void FASTCALL call_9_9_15_2(SH2_struct* sh)
{
  SH2movwi(sh, 9, (15<<4)+2);
}
#define init_9_9_15_3
void FASTCALL call_9_9_15_3(SH2_struct* sh)
{
  SH2movwi(sh, 9, (15<<4)+3);
}
#define init_9_9_15_4
void FASTCALL call_9_9_15_4(SH2_struct* sh)
{
  SH2movwi(sh, 9, (15<<4)+4);
}
#define init_9_9_15_5
void FASTCALL call_9_9_15_5(SH2_struct* sh)
{
  SH2movwi(sh, 9, (15<<4)+5);
}
#define init_9_9_15_6
void FASTCALL call_9_9_15_6(SH2_struct* sh)
{
  SH2movwi(sh, 9, (15<<4)+6);
}
#define init_9_9_15_7
void FASTCALL call_9_9_15_7(SH2_struct* sh)
{
  SH2movwi(sh, 9, (15<<4)+7);
}
#define init_9_9_15_8
void FASTCALL call_9_9_15_8(SH2_struct* sh)
{
  SH2movwi(sh, 9, (15<<4)+8);
}
#define init_9_9_15_9
void FASTCALL call_9_9_15_9(SH2_struct* sh)
{
  SH2movwi(sh, 9, (15<<4)+9);
}
#define init_9_9_15_10
void FASTCALL call_9_9_15_10(SH2_struct* sh)
{
  SH2movwi(sh, 9, (15<<4)+10);
}
#define init_9_9_15_11
void FASTCALL call_9_9_15_11(SH2_struct* sh)
{
  SH2movwi(sh, 9, (15<<4)+11);
}
#define init_9_9_15_12
void FASTCALL call_9_9_15_12(SH2_struct* sh)
{
  SH2movwi(sh, 9, (15<<4)+12);
}
#define init_9_9_15_13
void FASTCALL call_9_9_15_13(SH2_struct* sh)
{
  SH2movwi(sh, 9, (15<<4)+13);
}
#define init_9_9_15_14
void FASTCALL call_9_9_15_14(SH2_struct* sh)
{
  SH2movwi(sh, 9, (15<<4)+14);
}
#define init_9_9_15_15
void FASTCALL call_9_9_15_15(SH2_struct* sh)
{
  SH2movwi(sh, 9, (15<<4)+15);
}
#define init_9_10_0_0
void FASTCALL call_9_10_0_0(SH2_struct* sh)
{
  SH2movwi(sh, 10, (0<<4)+0);
}
#define init_9_10_0_1
void FASTCALL call_9_10_0_1(SH2_struct* sh)
{
  SH2movwi(sh, 10, (0<<4)+1);
}
#define init_9_10_0_2
void FASTCALL call_9_10_0_2(SH2_struct* sh)
{
  SH2movwi(sh, 10, (0<<4)+2);
}
#define init_9_10_0_3
void FASTCALL call_9_10_0_3(SH2_struct* sh)
{
  SH2movwi(sh, 10, (0<<4)+3);
}
#define init_9_10_0_4
void FASTCALL call_9_10_0_4(SH2_struct* sh)
{
  SH2movwi(sh, 10, (0<<4)+4);
}
#define init_9_10_0_5
void FASTCALL call_9_10_0_5(SH2_struct* sh)
{
  SH2movwi(sh, 10, (0<<4)+5);
}
#define init_9_10_0_6
void FASTCALL call_9_10_0_6(SH2_struct* sh)
{
  SH2movwi(sh, 10, (0<<4)+6);
}
#define init_9_10_0_7
void FASTCALL call_9_10_0_7(SH2_struct* sh)
{
  SH2movwi(sh, 10, (0<<4)+7);
}
#define init_9_10_0_8
void FASTCALL call_9_10_0_8(SH2_struct* sh)
{
  SH2movwi(sh, 10, (0<<4)+8);
}
#define init_9_10_0_9
void FASTCALL call_9_10_0_9(SH2_struct* sh)
{
  SH2movwi(sh, 10, (0<<4)+9);
}
#define init_9_10_0_10
void FASTCALL call_9_10_0_10(SH2_struct* sh)
{
  SH2movwi(sh, 10, (0<<4)+10);
}
#define init_9_10_0_11
void FASTCALL call_9_10_0_11(SH2_struct* sh)
{
  SH2movwi(sh, 10, (0<<4)+11);
}
#define init_9_10_0_12
void FASTCALL call_9_10_0_12(SH2_struct* sh)
{
  SH2movwi(sh, 10, (0<<4)+12);
}
#define init_9_10_0_13
void FASTCALL call_9_10_0_13(SH2_struct* sh)
{
  SH2movwi(sh, 10, (0<<4)+13);
}
#define init_9_10_0_14
void FASTCALL call_9_10_0_14(SH2_struct* sh)
{
  SH2movwi(sh, 10, (0<<4)+14);
}
#define init_9_10_0_15
void FASTCALL call_9_10_0_15(SH2_struct* sh)
{
  SH2movwi(sh, 10, (0<<4)+15);
}
#define init_9_10_1_0
void FASTCALL call_9_10_1_0(SH2_struct* sh)
{
  SH2movwi(sh, 10, (1<<4)+0);
}
#define init_9_10_1_1
void FASTCALL call_9_10_1_1(SH2_struct* sh)
{
  SH2movwi(sh, 10, (1<<4)+1);
}
#define init_9_10_1_2
void FASTCALL call_9_10_1_2(SH2_struct* sh)
{
  SH2movwi(sh, 10, (1<<4)+2);
}
#define init_9_10_1_3
void FASTCALL call_9_10_1_3(SH2_struct* sh)
{
  SH2movwi(sh, 10, (1<<4)+3);
}
#define init_9_10_1_4
void FASTCALL call_9_10_1_4(SH2_struct* sh)
{
  SH2movwi(sh, 10, (1<<4)+4);
}
#define init_9_10_1_5
void FASTCALL call_9_10_1_5(SH2_struct* sh)
{
  SH2movwi(sh, 10, (1<<4)+5);
}
#define init_9_10_1_6
void FASTCALL call_9_10_1_6(SH2_struct* sh)
{
  SH2movwi(sh, 10, (1<<4)+6);
}
#define init_9_10_1_7
void FASTCALL call_9_10_1_7(SH2_struct* sh)
{
  SH2movwi(sh, 10, (1<<4)+7);
}
#define init_9_10_1_8
void FASTCALL call_9_10_1_8(SH2_struct* sh)
{
  SH2movwi(sh, 10, (1<<4)+8);
}
#define init_9_10_1_9
void FASTCALL call_9_10_1_9(SH2_struct* sh)
{
  SH2movwi(sh, 10, (1<<4)+9);
}
#define init_9_10_1_10
void FASTCALL call_9_10_1_10(SH2_struct* sh)
{
  SH2movwi(sh, 10, (1<<4)+10);
}
#define init_9_10_1_11
void FASTCALL call_9_10_1_11(SH2_struct* sh)
{
  SH2movwi(sh, 10, (1<<4)+11);
}
#define init_9_10_1_12
void FASTCALL call_9_10_1_12(SH2_struct* sh)
{
  SH2movwi(sh, 10, (1<<4)+12);
}
#define init_9_10_1_13
void FASTCALL call_9_10_1_13(SH2_struct* sh)
{
  SH2movwi(sh, 10, (1<<4)+13);
}
#define init_9_10_1_14
void FASTCALL call_9_10_1_14(SH2_struct* sh)
{
  SH2movwi(sh, 10, (1<<4)+14);
}
#define init_9_10_1_15
void FASTCALL call_9_10_1_15(SH2_struct* sh)
{
  SH2movwi(sh, 10, (1<<4)+15);
}
#define init_9_10_2_0
void FASTCALL call_9_10_2_0(SH2_struct* sh)
{
  SH2movwi(sh, 10, (2<<4)+0);
}
#define init_9_10_2_1
void FASTCALL call_9_10_2_1(SH2_struct* sh)
{
  SH2movwi(sh, 10, (2<<4)+1);
}
#define init_9_10_2_2
void FASTCALL call_9_10_2_2(SH2_struct* sh)
{
  SH2movwi(sh, 10, (2<<4)+2);
}
#define init_9_10_2_3
void FASTCALL call_9_10_2_3(SH2_struct* sh)
{
  SH2movwi(sh, 10, (2<<4)+3);
}
#define init_9_10_2_4
void FASTCALL call_9_10_2_4(SH2_struct* sh)
{
  SH2movwi(sh, 10, (2<<4)+4);
}
#define init_9_10_2_5
void FASTCALL call_9_10_2_5(SH2_struct* sh)
{
  SH2movwi(sh, 10, (2<<4)+5);
}
#define init_9_10_2_6
void FASTCALL call_9_10_2_6(SH2_struct* sh)
{
  SH2movwi(sh, 10, (2<<4)+6);
}
#define init_9_10_2_7
void FASTCALL call_9_10_2_7(SH2_struct* sh)
{
  SH2movwi(sh, 10, (2<<4)+7);
}
#define init_9_10_2_8
void FASTCALL call_9_10_2_8(SH2_struct* sh)
{
  SH2movwi(sh, 10, (2<<4)+8);
}
#define init_9_10_2_9
void FASTCALL call_9_10_2_9(SH2_struct* sh)
{
  SH2movwi(sh, 10, (2<<4)+9);
}
#define init_9_10_2_10
void FASTCALL call_9_10_2_10(SH2_struct* sh)
{
  SH2movwi(sh, 10, (2<<4)+10);
}
#define init_9_10_2_11
void FASTCALL call_9_10_2_11(SH2_struct* sh)
{
  SH2movwi(sh, 10, (2<<4)+11);
}
#define init_9_10_2_12
void FASTCALL call_9_10_2_12(SH2_struct* sh)
{
  SH2movwi(sh, 10, (2<<4)+12);
}
#define init_9_10_2_13
void FASTCALL call_9_10_2_13(SH2_struct* sh)
{
  SH2movwi(sh, 10, (2<<4)+13);
}
#define init_9_10_2_14
void FASTCALL call_9_10_2_14(SH2_struct* sh)
{
  SH2movwi(sh, 10, (2<<4)+14);
}
#define init_9_10_2_15
void FASTCALL call_9_10_2_15(SH2_struct* sh)
{
  SH2movwi(sh, 10, (2<<4)+15);
}
#define init_9_10_3_0
void FASTCALL call_9_10_3_0(SH2_struct* sh)
{
  SH2movwi(sh, 10, (3<<4)+0);
}
#define init_9_10_3_1
void FASTCALL call_9_10_3_1(SH2_struct* sh)
{
  SH2movwi(sh, 10, (3<<4)+1);
}
#define init_9_10_3_2
void FASTCALL call_9_10_3_2(SH2_struct* sh)
{
  SH2movwi(sh, 10, (3<<4)+2);
}
#define init_9_10_3_3
void FASTCALL call_9_10_3_3(SH2_struct* sh)
{
  SH2movwi(sh, 10, (3<<4)+3);
}
#define init_9_10_3_4
void FASTCALL call_9_10_3_4(SH2_struct* sh)
{
  SH2movwi(sh, 10, (3<<4)+4);
}
#define init_9_10_3_5
void FASTCALL call_9_10_3_5(SH2_struct* sh)
{
  SH2movwi(sh, 10, (3<<4)+5);
}
#define init_9_10_3_6
void FASTCALL call_9_10_3_6(SH2_struct* sh)
{
  SH2movwi(sh, 10, (3<<4)+6);
}
#define init_9_10_3_7
void FASTCALL call_9_10_3_7(SH2_struct* sh)
{
  SH2movwi(sh, 10, (3<<4)+7);
}
#define init_9_10_3_8
void FASTCALL call_9_10_3_8(SH2_struct* sh)
{
  SH2movwi(sh, 10, (3<<4)+8);
}
#define init_9_10_3_9
void FASTCALL call_9_10_3_9(SH2_struct* sh)
{
  SH2movwi(sh, 10, (3<<4)+9);
}
#define init_9_10_3_10
void FASTCALL call_9_10_3_10(SH2_struct* sh)
{
  SH2movwi(sh, 10, (3<<4)+10);
}
#define init_9_10_3_11
void FASTCALL call_9_10_3_11(SH2_struct* sh)
{
  SH2movwi(sh, 10, (3<<4)+11);
}
#define init_9_10_3_12
void FASTCALL call_9_10_3_12(SH2_struct* sh)
{
  SH2movwi(sh, 10, (3<<4)+12);
}
#define init_9_10_3_13
void FASTCALL call_9_10_3_13(SH2_struct* sh)
{
  SH2movwi(sh, 10, (3<<4)+13);
}
#define init_9_10_3_14
void FASTCALL call_9_10_3_14(SH2_struct* sh)
{
  SH2movwi(sh, 10, (3<<4)+14);
}
#define init_9_10_3_15
void FASTCALL call_9_10_3_15(SH2_struct* sh)
{
  SH2movwi(sh, 10, (3<<4)+15);
}
#define init_9_10_4_0
void FASTCALL call_9_10_4_0(SH2_struct* sh)
{
  SH2movwi(sh, 10, (4<<4)+0);
}
#define init_9_10_4_1
void FASTCALL call_9_10_4_1(SH2_struct* sh)
{
  SH2movwi(sh, 10, (4<<4)+1);
}
#define init_9_10_4_2
void FASTCALL call_9_10_4_2(SH2_struct* sh)
{
  SH2movwi(sh, 10, (4<<4)+2);
}
#define init_9_10_4_3
void FASTCALL call_9_10_4_3(SH2_struct* sh)
{
  SH2movwi(sh, 10, (4<<4)+3);
}
#define init_9_10_4_4
void FASTCALL call_9_10_4_4(SH2_struct* sh)
{
  SH2movwi(sh, 10, (4<<4)+4);
}
#define init_9_10_4_5
void FASTCALL call_9_10_4_5(SH2_struct* sh)
{
  SH2movwi(sh, 10, (4<<4)+5);
}
#define init_9_10_4_6
void FASTCALL call_9_10_4_6(SH2_struct* sh)
{
  SH2movwi(sh, 10, (4<<4)+6);
}
#define init_9_10_4_7
void FASTCALL call_9_10_4_7(SH2_struct* sh)
{
  SH2movwi(sh, 10, (4<<4)+7);
}
#define init_9_10_4_8
void FASTCALL call_9_10_4_8(SH2_struct* sh)
{
  SH2movwi(sh, 10, (4<<4)+8);
}
#define init_9_10_4_9
void FASTCALL call_9_10_4_9(SH2_struct* sh)
{
  SH2movwi(sh, 10, (4<<4)+9);
}
#define init_9_10_4_10
void FASTCALL call_9_10_4_10(SH2_struct* sh)
{
  SH2movwi(sh, 10, (4<<4)+10);
}
#define init_9_10_4_11
void FASTCALL call_9_10_4_11(SH2_struct* sh)
{
  SH2movwi(sh, 10, (4<<4)+11);
}
#define init_9_10_4_12
void FASTCALL call_9_10_4_12(SH2_struct* sh)
{
  SH2movwi(sh, 10, (4<<4)+12);
}
#define init_9_10_4_13
void FASTCALL call_9_10_4_13(SH2_struct* sh)
{
  SH2movwi(sh, 10, (4<<4)+13);
}
#define init_9_10_4_14
void FASTCALL call_9_10_4_14(SH2_struct* sh)
{
  SH2movwi(sh, 10, (4<<4)+14);
}
#define init_9_10_4_15
void FASTCALL call_9_10_4_15(SH2_struct* sh)
{
  SH2movwi(sh, 10, (4<<4)+15);
}
#define init_9_10_5_0
void FASTCALL call_9_10_5_0(SH2_struct* sh)
{
  SH2movwi(sh, 10, (5<<4)+0);
}
#define init_9_10_5_1
void FASTCALL call_9_10_5_1(SH2_struct* sh)
{
  SH2movwi(sh, 10, (5<<4)+1);
}
#define init_9_10_5_2
void FASTCALL call_9_10_5_2(SH2_struct* sh)
{
  SH2movwi(sh, 10, (5<<4)+2);
}
#define init_9_10_5_3
void FASTCALL call_9_10_5_3(SH2_struct* sh)
{
  SH2movwi(sh, 10, (5<<4)+3);
}
#define init_9_10_5_4
void FASTCALL call_9_10_5_4(SH2_struct* sh)
{
  SH2movwi(sh, 10, (5<<4)+4);
}
#define init_9_10_5_5
void FASTCALL call_9_10_5_5(SH2_struct* sh)
{
  SH2movwi(sh, 10, (5<<4)+5);
}
#define init_9_10_5_6
void FASTCALL call_9_10_5_6(SH2_struct* sh)
{
  SH2movwi(sh, 10, (5<<4)+6);
}
#define init_9_10_5_7
void FASTCALL call_9_10_5_7(SH2_struct* sh)
{
  SH2movwi(sh, 10, (5<<4)+7);
}
#define init_9_10_5_8
void FASTCALL call_9_10_5_8(SH2_struct* sh)
{
  SH2movwi(sh, 10, (5<<4)+8);
}
#define init_9_10_5_9
void FASTCALL call_9_10_5_9(SH2_struct* sh)
{
  SH2movwi(sh, 10, (5<<4)+9);
}
#define init_9_10_5_10
void FASTCALL call_9_10_5_10(SH2_struct* sh)
{
  SH2movwi(sh, 10, (5<<4)+10);
}
#define init_9_10_5_11
void FASTCALL call_9_10_5_11(SH2_struct* sh)
{
  SH2movwi(sh, 10, (5<<4)+11);
}
#define init_9_10_5_12
void FASTCALL call_9_10_5_12(SH2_struct* sh)
{
  SH2movwi(sh, 10, (5<<4)+12);
}
#define init_9_10_5_13
void FASTCALL call_9_10_5_13(SH2_struct* sh)
{
  SH2movwi(sh, 10, (5<<4)+13);
}
#define init_9_10_5_14
void FASTCALL call_9_10_5_14(SH2_struct* sh)
{
  SH2movwi(sh, 10, (5<<4)+14);
}
#define init_9_10_5_15
void FASTCALL call_9_10_5_15(SH2_struct* sh)
{
  SH2movwi(sh, 10, (5<<4)+15);
}
#define init_9_10_6_0
void FASTCALL call_9_10_6_0(SH2_struct* sh)
{
  SH2movwi(sh, 10, (6<<4)+0);
}
#define init_9_10_6_1
void FASTCALL call_9_10_6_1(SH2_struct* sh)
{
  SH2movwi(sh, 10, (6<<4)+1);
}
#define init_9_10_6_2
void FASTCALL call_9_10_6_2(SH2_struct* sh)
{
  SH2movwi(sh, 10, (6<<4)+2);
}
#define init_9_10_6_3
void FASTCALL call_9_10_6_3(SH2_struct* sh)
{
  SH2movwi(sh, 10, (6<<4)+3);
}
#define init_9_10_6_4
void FASTCALL call_9_10_6_4(SH2_struct* sh)
{
  SH2movwi(sh, 10, (6<<4)+4);
}
#define init_9_10_6_5
void FASTCALL call_9_10_6_5(SH2_struct* sh)
{
  SH2movwi(sh, 10, (6<<4)+5);
}
#define init_9_10_6_6
void FASTCALL call_9_10_6_6(SH2_struct* sh)
{
  SH2movwi(sh, 10, (6<<4)+6);
}
#define init_9_10_6_7
void FASTCALL call_9_10_6_7(SH2_struct* sh)
{
  SH2movwi(sh, 10, (6<<4)+7);
}
#define init_9_10_6_8
void FASTCALL call_9_10_6_8(SH2_struct* sh)
{
  SH2movwi(sh, 10, (6<<4)+8);
}
#define init_9_10_6_9
void FASTCALL call_9_10_6_9(SH2_struct* sh)
{
  SH2movwi(sh, 10, (6<<4)+9);
}
#define init_9_10_6_10
void FASTCALL call_9_10_6_10(SH2_struct* sh)
{
  SH2movwi(sh, 10, (6<<4)+10);
}
#define init_9_10_6_11
void FASTCALL call_9_10_6_11(SH2_struct* sh)
{
  SH2movwi(sh, 10, (6<<4)+11);
}
#define init_9_10_6_12
void FASTCALL call_9_10_6_12(SH2_struct* sh)
{
  SH2movwi(sh, 10, (6<<4)+12);
}
#define init_9_10_6_13
void FASTCALL call_9_10_6_13(SH2_struct* sh)
{
  SH2movwi(sh, 10, (6<<4)+13);
}
#define init_9_10_6_14
void FASTCALL call_9_10_6_14(SH2_struct* sh)
{
  SH2movwi(sh, 10, (6<<4)+14);
}
#define init_9_10_6_15
void FASTCALL call_9_10_6_15(SH2_struct* sh)
{
  SH2movwi(sh, 10, (6<<4)+15);
}
#define init_9_10_7_0
void FASTCALL call_9_10_7_0(SH2_struct* sh)
{
  SH2movwi(sh, 10, (7<<4)+0);
}
#define init_9_10_7_1
void FASTCALL call_9_10_7_1(SH2_struct* sh)
{
  SH2movwi(sh, 10, (7<<4)+1);
}
#define init_9_10_7_2
void FASTCALL call_9_10_7_2(SH2_struct* sh)
{
  SH2movwi(sh, 10, (7<<4)+2);
}
#define init_9_10_7_3
void FASTCALL call_9_10_7_3(SH2_struct* sh)
{
  SH2movwi(sh, 10, (7<<4)+3);
}
#define init_9_10_7_4
void FASTCALL call_9_10_7_4(SH2_struct* sh)
{
  SH2movwi(sh, 10, (7<<4)+4);
}
#define init_9_10_7_5
void FASTCALL call_9_10_7_5(SH2_struct* sh)
{
  SH2movwi(sh, 10, (7<<4)+5);
}
#define init_9_10_7_6
void FASTCALL call_9_10_7_6(SH2_struct* sh)
{
  SH2movwi(sh, 10, (7<<4)+6);
}
#define init_9_10_7_7
void FASTCALL call_9_10_7_7(SH2_struct* sh)
{
  SH2movwi(sh, 10, (7<<4)+7);
}
#define init_9_10_7_8
void FASTCALL call_9_10_7_8(SH2_struct* sh)
{
  SH2movwi(sh, 10, (7<<4)+8);
}
#define init_9_10_7_9
void FASTCALL call_9_10_7_9(SH2_struct* sh)
{
  SH2movwi(sh, 10, (7<<4)+9);
}
#define init_9_10_7_10
void FASTCALL call_9_10_7_10(SH2_struct* sh)
{
  SH2movwi(sh, 10, (7<<4)+10);
}
#define init_9_10_7_11
void FASTCALL call_9_10_7_11(SH2_struct* sh)
{
  SH2movwi(sh, 10, (7<<4)+11);
}
#define init_9_10_7_12
void FASTCALL call_9_10_7_12(SH2_struct* sh)
{
  SH2movwi(sh, 10, (7<<4)+12);
}
#define init_9_10_7_13
void FASTCALL call_9_10_7_13(SH2_struct* sh)
{
  SH2movwi(sh, 10, (7<<4)+13);
}
#define init_9_10_7_14
void FASTCALL call_9_10_7_14(SH2_struct* sh)
{
  SH2movwi(sh, 10, (7<<4)+14);
}
#define init_9_10_7_15
void FASTCALL call_9_10_7_15(SH2_struct* sh)
{
  SH2movwi(sh, 10, (7<<4)+15);
}
#define init_9_10_8_0
void FASTCALL call_9_10_8_0(SH2_struct* sh)
{
  SH2movwi(sh, 10, (8<<4)+0);
}
#define init_9_10_8_1
void FASTCALL call_9_10_8_1(SH2_struct* sh)
{
  SH2movwi(sh, 10, (8<<4)+1);
}
#define init_9_10_8_2
void FASTCALL call_9_10_8_2(SH2_struct* sh)
{
  SH2movwi(sh, 10, (8<<4)+2);
}
#define init_9_10_8_3
void FASTCALL call_9_10_8_3(SH2_struct* sh)
{
  SH2movwi(sh, 10, (8<<4)+3);
}
#define init_9_10_8_4
void FASTCALL call_9_10_8_4(SH2_struct* sh)
{
  SH2movwi(sh, 10, (8<<4)+4);
}
#define init_9_10_8_5
void FASTCALL call_9_10_8_5(SH2_struct* sh)
{
  SH2movwi(sh, 10, (8<<4)+5);
}
#define init_9_10_8_6
void FASTCALL call_9_10_8_6(SH2_struct* sh)
{
  SH2movwi(sh, 10, (8<<4)+6);
}
#define init_9_10_8_7
void FASTCALL call_9_10_8_7(SH2_struct* sh)
{
  SH2movwi(sh, 10, (8<<4)+7);
}
#define init_9_10_8_8
void FASTCALL call_9_10_8_8(SH2_struct* sh)
{
  SH2movwi(sh, 10, (8<<4)+8);
}
#define init_9_10_8_9
void FASTCALL call_9_10_8_9(SH2_struct* sh)
{
  SH2movwi(sh, 10, (8<<4)+9);
}
#define init_9_10_8_10
void FASTCALL call_9_10_8_10(SH2_struct* sh)
{
  SH2movwi(sh, 10, (8<<4)+10);
}
#define init_9_10_8_11
void FASTCALL call_9_10_8_11(SH2_struct* sh)
{
  SH2movwi(sh, 10, (8<<4)+11);
}
#define init_9_10_8_12
void FASTCALL call_9_10_8_12(SH2_struct* sh)
{
  SH2movwi(sh, 10, (8<<4)+12);
}
#define init_9_10_8_13
void FASTCALL call_9_10_8_13(SH2_struct* sh)
{
  SH2movwi(sh, 10, (8<<4)+13);
}
#define init_9_10_8_14
void FASTCALL call_9_10_8_14(SH2_struct* sh)
{
  SH2movwi(sh, 10, (8<<4)+14);
}
#define init_9_10_8_15
void FASTCALL call_9_10_8_15(SH2_struct* sh)
{
  SH2movwi(sh, 10, (8<<4)+15);
}
#define init_9_10_9_0
void FASTCALL call_9_10_9_0(SH2_struct* sh)
{
  SH2movwi(sh, 10, (9<<4)+0);
}
#define init_9_10_9_1
void FASTCALL call_9_10_9_1(SH2_struct* sh)
{
  SH2movwi(sh, 10, (9<<4)+1);
}
#define init_9_10_9_2
void FASTCALL call_9_10_9_2(SH2_struct* sh)
{
  SH2movwi(sh, 10, (9<<4)+2);
}
#define init_9_10_9_3
void FASTCALL call_9_10_9_3(SH2_struct* sh)
{
  SH2movwi(sh, 10, (9<<4)+3);
}
#define init_9_10_9_4
void FASTCALL call_9_10_9_4(SH2_struct* sh)
{
  SH2movwi(sh, 10, (9<<4)+4);
}
#define init_9_10_9_5
void FASTCALL call_9_10_9_5(SH2_struct* sh)
{
  SH2movwi(sh, 10, (9<<4)+5);
}
#define init_9_10_9_6
void FASTCALL call_9_10_9_6(SH2_struct* sh)
{
  SH2movwi(sh, 10, (9<<4)+6);
}
#define init_9_10_9_7
void FASTCALL call_9_10_9_7(SH2_struct* sh)
{
  SH2movwi(sh, 10, (9<<4)+7);
}
#define init_9_10_9_8
void FASTCALL call_9_10_9_8(SH2_struct* sh)
{
  SH2movwi(sh, 10, (9<<4)+8);
}
#define init_9_10_9_9
void FASTCALL call_9_10_9_9(SH2_struct* sh)
{
  SH2movwi(sh, 10, (9<<4)+9);
}
#define init_9_10_9_10
void FASTCALL call_9_10_9_10(SH2_struct* sh)
{
  SH2movwi(sh, 10, (9<<4)+10);
}
#define init_9_10_9_11
void FASTCALL call_9_10_9_11(SH2_struct* sh)
{
  SH2movwi(sh, 10, (9<<4)+11);
}
#define init_9_10_9_12
void FASTCALL call_9_10_9_12(SH2_struct* sh)
{
  SH2movwi(sh, 10, (9<<4)+12);
}
#define init_9_10_9_13
void FASTCALL call_9_10_9_13(SH2_struct* sh)
{
  SH2movwi(sh, 10, (9<<4)+13);
}
#define init_9_10_9_14
void FASTCALL call_9_10_9_14(SH2_struct* sh)
{
  SH2movwi(sh, 10, (9<<4)+14);
}
#define init_9_10_9_15
void FASTCALL call_9_10_9_15(SH2_struct* sh)
{
  SH2movwi(sh, 10, (9<<4)+15);
}
#define init_9_10_10_0
void FASTCALL call_9_10_10_0(SH2_struct* sh)
{
  SH2movwi(sh, 10, (10<<4)+0);
}
#define init_9_10_10_1
void FASTCALL call_9_10_10_1(SH2_struct* sh)
{
  SH2movwi(sh, 10, (10<<4)+1);
}
#define init_9_10_10_2
void FASTCALL call_9_10_10_2(SH2_struct* sh)
{
  SH2movwi(sh, 10, (10<<4)+2);
}
#define init_9_10_10_3
void FASTCALL call_9_10_10_3(SH2_struct* sh)
{
  SH2movwi(sh, 10, (10<<4)+3);
}
#define init_9_10_10_4
void FASTCALL call_9_10_10_4(SH2_struct* sh)
{
  SH2movwi(sh, 10, (10<<4)+4);
}
#define init_9_10_10_5
void FASTCALL call_9_10_10_5(SH2_struct* sh)
{
  SH2movwi(sh, 10, (10<<4)+5);
}
#define init_9_10_10_6
void FASTCALL call_9_10_10_6(SH2_struct* sh)
{
  SH2movwi(sh, 10, (10<<4)+6);
}
#define init_9_10_10_7
void FASTCALL call_9_10_10_7(SH2_struct* sh)
{
  SH2movwi(sh, 10, (10<<4)+7);
}
#define init_9_10_10_8
void FASTCALL call_9_10_10_8(SH2_struct* sh)
{
  SH2movwi(sh, 10, (10<<4)+8);
}
#define init_9_10_10_9
void FASTCALL call_9_10_10_9(SH2_struct* sh)
{
  SH2movwi(sh, 10, (10<<4)+9);
}
#define init_9_10_10_10
void FASTCALL call_9_10_10_10(SH2_struct* sh)
{
  SH2movwi(sh, 10, (10<<4)+10);
}
#define init_9_10_10_11
void FASTCALL call_9_10_10_11(SH2_struct* sh)
{
  SH2movwi(sh, 10, (10<<4)+11);
}
#define init_9_10_10_12
void FASTCALL call_9_10_10_12(SH2_struct* sh)
{
  SH2movwi(sh, 10, (10<<4)+12);
}
#define init_9_10_10_13
void FASTCALL call_9_10_10_13(SH2_struct* sh)
{
  SH2movwi(sh, 10, (10<<4)+13);
}
#define init_9_10_10_14
void FASTCALL call_9_10_10_14(SH2_struct* sh)
{
  SH2movwi(sh, 10, (10<<4)+14);
}
#define init_9_10_10_15
void FASTCALL call_9_10_10_15(SH2_struct* sh)
{
  SH2movwi(sh, 10, (10<<4)+15);
}
#define init_9_10_11_0
void FASTCALL call_9_10_11_0(SH2_struct* sh)
{
  SH2movwi(sh, 10, (11<<4)+0);
}
#define init_9_10_11_1
void FASTCALL call_9_10_11_1(SH2_struct* sh)
{
  SH2movwi(sh, 10, (11<<4)+1);
}
#define init_9_10_11_2
void FASTCALL call_9_10_11_2(SH2_struct* sh)
{
  SH2movwi(sh, 10, (11<<4)+2);
}
#define init_9_10_11_3
void FASTCALL call_9_10_11_3(SH2_struct* sh)
{
  SH2movwi(sh, 10, (11<<4)+3);
}
#define init_9_10_11_4
void FASTCALL call_9_10_11_4(SH2_struct* sh)
{
  SH2movwi(sh, 10, (11<<4)+4);
}
#define init_9_10_11_5
void FASTCALL call_9_10_11_5(SH2_struct* sh)
{
  SH2movwi(sh, 10, (11<<4)+5);
}
#define init_9_10_11_6
void FASTCALL call_9_10_11_6(SH2_struct* sh)
{
  SH2movwi(sh, 10, (11<<4)+6);
}
#define init_9_10_11_7
void FASTCALL call_9_10_11_7(SH2_struct* sh)
{
  SH2movwi(sh, 10, (11<<4)+7);
}
#define init_9_10_11_8
void FASTCALL call_9_10_11_8(SH2_struct* sh)
{
  SH2movwi(sh, 10, (11<<4)+8);
}
#define init_9_10_11_9
void FASTCALL call_9_10_11_9(SH2_struct* sh)
{
  SH2movwi(sh, 10, (11<<4)+9);
}
#define init_9_10_11_10
void FASTCALL call_9_10_11_10(SH2_struct* sh)
{
  SH2movwi(sh, 10, (11<<4)+10);
}
#define init_9_10_11_11
void FASTCALL call_9_10_11_11(SH2_struct* sh)
{
  SH2movwi(sh, 10, (11<<4)+11);
}
#define init_9_10_11_12
void FASTCALL call_9_10_11_12(SH2_struct* sh)
{
  SH2movwi(sh, 10, (11<<4)+12);
}
#define init_9_10_11_13
void FASTCALL call_9_10_11_13(SH2_struct* sh)
{
  SH2movwi(sh, 10, (11<<4)+13);
}
#define init_9_10_11_14
void FASTCALL call_9_10_11_14(SH2_struct* sh)
{
  SH2movwi(sh, 10, (11<<4)+14);
}
#define init_9_10_11_15
void FASTCALL call_9_10_11_15(SH2_struct* sh)
{
  SH2movwi(sh, 10, (11<<4)+15);
}
#define init_9_10_12_0
void FASTCALL call_9_10_12_0(SH2_struct* sh)
{
  SH2movwi(sh, 10, (12<<4)+0);
}
#define init_9_10_12_1
void FASTCALL call_9_10_12_1(SH2_struct* sh)
{
  SH2movwi(sh, 10, (12<<4)+1);
}
#define init_9_10_12_2
void FASTCALL call_9_10_12_2(SH2_struct* sh)
{
  SH2movwi(sh, 10, (12<<4)+2);
}
#define init_9_10_12_3
void FASTCALL call_9_10_12_3(SH2_struct* sh)
{
  SH2movwi(sh, 10, (12<<4)+3);
}
#define init_9_10_12_4
void FASTCALL call_9_10_12_4(SH2_struct* sh)
{
  SH2movwi(sh, 10, (12<<4)+4);
}
#define init_9_10_12_5
void FASTCALL call_9_10_12_5(SH2_struct* sh)
{
  SH2movwi(sh, 10, (12<<4)+5);
}
#define init_9_10_12_6
void FASTCALL call_9_10_12_6(SH2_struct* sh)
{
  SH2movwi(sh, 10, (12<<4)+6);
}
#define init_9_10_12_7
void FASTCALL call_9_10_12_7(SH2_struct* sh)
{
  SH2movwi(sh, 10, (12<<4)+7);
}
#define init_9_10_12_8
void FASTCALL call_9_10_12_8(SH2_struct* sh)
{
  SH2movwi(sh, 10, (12<<4)+8);
}
#define init_9_10_12_9
void FASTCALL call_9_10_12_9(SH2_struct* sh)
{
  SH2movwi(sh, 10, (12<<4)+9);
}
#define init_9_10_12_10
void FASTCALL call_9_10_12_10(SH2_struct* sh)
{
  SH2movwi(sh, 10, (12<<4)+10);
}
#define init_9_10_12_11
void FASTCALL call_9_10_12_11(SH2_struct* sh)
{
  SH2movwi(sh, 10, (12<<4)+11);
}
#define init_9_10_12_12
void FASTCALL call_9_10_12_12(SH2_struct* sh)
{
  SH2movwi(sh, 10, (12<<4)+12);
}
#define init_9_10_12_13
void FASTCALL call_9_10_12_13(SH2_struct* sh)
{
  SH2movwi(sh, 10, (12<<4)+13);
}
#define init_9_10_12_14
void FASTCALL call_9_10_12_14(SH2_struct* sh)
{
  SH2movwi(sh, 10, (12<<4)+14);
}
#define init_9_10_12_15
void FASTCALL call_9_10_12_15(SH2_struct* sh)
{
  SH2movwi(sh, 10, (12<<4)+15);
}
#define init_9_10_13_0
void FASTCALL call_9_10_13_0(SH2_struct* sh)
{
  SH2movwi(sh, 10, (13<<4)+0);
}
#define init_9_10_13_1
void FASTCALL call_9_10_13_1(SH2_struct* sh)
{
  SH2movwi(sh, 10, (13<<4)+1);
}
#define init_9_10_13_2
void FASTCALL call_9_10_13_2(SH2_struct* sh)
{
  SH2movwi(sh, 10, (13<<4)+2);
}
#define init_9_10_13_3
void FASTCALL call_9_10_13_3(SH2_struct* sh)
{
  SH2movwi(sh, 10, (13<<4)+3);
}
#define init_9_10_13_4
void FASTCALL call_9_10_13_4(SH2_struct* sh)
{
  SH2movwi(sh, 10, (13<<4)+4);
}
#define init_9_10_13_5
void FASTCALL call_9_10_13_5(SH2_struct* sh)
{
  SH2movwi(sh, 10, (13<<4)+5);
}
#define init_9_10_13_6
void FASTCALL call_9_10_13_6(SH2_struct* sh)
{
  SH2movwi(sh, 10, (13<<4)+6);
}
#define init_9_10_13_7
void FASTCALL call_9_10_13_7(SH2_struct* sh)
{
  SH2movwi(sh, 10, (13<<4)+7);
}
#define init_9_10_13_8
void FASTCALL call_9_10_13_8(SH2_struct* sh)
{
  SH2movwi(sh, 10, (13<<4)+8);
}
#define init_9_10_13_9
void FASTCALL call_9_10_13_9(SH2_struct* sh)
{
  SH2movwi(sh, 10, (13<<4)+9);
}
#define init_9_10_13_10
void FASTCALL call_9_10_13_10(SH2_struct* sh)
{
  SH2movwi(sh, 10, (13<<4)+10);
}
#define init_9_10_13_11
void FASTCALL call_9_10_13_11(SH2_struct* sh)
{
  SH2movwi(sh, 10, (13<<4)+11);
}
#define init_9_10_13_12
void FASTCALL call_9_10_13_12(SH2_struct* sh)
{
  SH2movwi(sh, 10, (13<<4)+12);
}
#define init_9_10_13_13
void FASTCALL call_9_10_13_13(SH2_struct* sh)
{
  SH2movwi(sh, 10, (13<<4)+13);
}
#define init_9_10_13_14
void FASTCALL call_9_10_13_14(SH2_struct* sh)
{
  SH2movwi(sh, 10, (13<<4)+14);
}
#define init_9_10_13_15
void FASTCALL call_9_10_13_15(SH2_struct* sh)
{
  SH2movwi(sh, 10, (13<<4)+15);
}
#define init_9_10_14_0
void FASTCALL call_9_10_14_0(SH2_struct* sh)
{
  SH2movwi(sh, 10, (14<<4)+0);
}
#define init_9_10_14_1
void FASTCALL call_9_10_14_1(SH2_struct* sh)
{
  SH2movwi(sh, 10, (14<<4)+1);
}
#define init_9_10_14_2
void FASTCALL call_9_10_14_2(SH2_struct* sh)
{
  SH2movwi(sh, 10, (14<<4)+2);
}
#define init_9_10_14_3
void FASTCALL call_9_10_14_3(SH2_struct* sh)
{
  SH2movwi(sh, 10, (14<<4)+3);
}
#define init_9_10_14_4
void FASTCALL call_9_10_14_4(SH2_struct* sh)
{
  SH2movwi(sh, 10, (14<<4)+4);
}
#define init_9_10_14_5
void FASTCALL call_9_10_14_5(SH2_struct* sh)
{
  SH2movwi(sh, 10, (14<<4)+5);
}
#define init_9_10_14_6
void FASTCALL call_9_10_14_6(SH2_struct* sh)
{
  SH2movwi(sh, 10, (14<<4)+6);
}
#define init_9_10_14_7
void FASTCALL call_9_10_14_7(SH2_struct* sh)
{
  SH2movwi(sh, 10, (14<<4)+7);
}
#define init_9_10_14_8
void FASTCALL call_9_10_14_8(SH2_struct* sh)
{
  SH2movwi(sh, 10, (14<<4)+8);
}
#define init_9_10_14_9
void FASTCALL call_9_10_14_9(SH2_struct* sh)
{
  SH2movwi(sh, 10, (14<<4)+9);
}
#define init_9_10_14_10
void FASTCALL call_9_10_14_10(SH2_struct* sh)
{
  SH2movwi(sh, 10, (14<<4)+10);
}
#define init_9_10_14_11
void FASTCALL call_9_10_14_11(SH2_struct* sh)
{
  SH2movwi(sh, 10, (14<<4)+11);
}
#define init_9_10_14_12
void FASTCALL call_9_10_14_12(SH2_struct* sh)
{
  SH2movwi(sh, 10, (14<<4)+12);
}
#define init_9_10_14_13
void FASTCALL call_9_10_14_13(SH2_struct* sh)
{
  SH2movwi(sh, 10, (14<<4)+13);
}
#define init_9_10_14_14
void FASTCALL call_9_10_14_14(SH2_struct* sh)
{
  SH2movwi(sh, 10, (14<<4)+14);
}
#define init_9_10_14_15
void FASTCALL call_9_10_14_15(SH2_struct* sh)
{
  SH2movwi(sh, 10, (14<<4)+15);
}
#define init_9_10_15_0
void FASTCALL call_9_10_15_0(SH2_struct* sh)
{
  SH2movwi(sh, 10, (15<<4)+0);
}
#define init_9_10_15_1
void FASTCALL call_9_10_15_1(SH2_struct* sh)
{
  SH2movwi(sh, 10, (15<<4)+1);
}
#define init_9_10_15_2
void FASTCALL call_9_10_15_2(SH2_struct* sh)
{
  SH2movwi(sh, 10, (15<<4)+2);
}
#define init_9_10_15_3
void FASTCALL call_9_10_15_3(SH2_struct* sh)
{
  SH2movwi(sh, 10, (15<<4)+3);
}
#define init_9_10_15_4
void FASTCALL call_9_10_15_4(SH2_struct* sh)
{
  SH2movwi(sh, 10, (15<<4)+4);
}
#define init_9_10_15_5
void FASTCALL call_9_10_15_5(SH2_struct* sh)
{
  SH2movwi(sh, 10, (15<<4)+5);
}
#define init_9_10_15_6
void FASTCALL call_9_10_15_6(SH2_struct* sh)
{
  SH2movwi(sh, 10, (15<<4)+6);
}
#define init_9_10_15_7
void FASTCALL call_9_10_15_7(SH2_struct* sh)
{
  SH2movwi(sh, 10, (15<<4)+7);
}
#define init_9_10_15_8
void FASTCALL call_9_10_15_8(SH2_struct* sh)
{
  SH2movwi(sh, 10, (15<<4)+8);
}
#define init_9_10_15_9
void FASTCALL call_9_10_15_9(SH2_struct* sh)
{
  SH2movwi(sh, 10, (15<<4)+9);
}
#define init_9_10_15_10
void FASTCALL call_9_10_15_10(SH2_struct* sh)
{
  SH2movwi(sh, 10, (15<<4)+10);
}
#define init_9_10_15_11
void FASTCALL call_9_10_15_11(SH2_struct* sh)
{
  SH2movwi(sh, 10, (15<<4)+11);
}
#define init_9_10_15_12
void FASTCALL call_9_10_15_12(SH2_struct* sh)
{
  SH2movwi(sh, 10, (15<<4)+12);
}
#define init_9_10_15_13
void FASTCALL call_9_10_15_13(SH2_struct* sh)
{
  SH2movwi(sh, 10, (15<<4)+13);
}
#define init_9_10_15_14
void FASTCALL call_9_10_15_14(SH2_struct* sh)
{
  SH2movwi(sh, 10, (15<<4)+14);
}
#define init_9_10_15_15
void FASTCALL call_9_10_15_15(SH2_struct* sh)
{
  SH2movwi(sh, 10, (15<<4)+15);
}
#define init_9_11_0_0
void FASTCALL call_9_11_0_0(SH2_struct* sh)
{
  SH2movwi(sh, 11, (0<<4)+0);
}
#define init_9_11_0_1
void FASTCALL call_9_11_0_1(SH2_struct* sh)
{
  SH2movwi(sh, 11, (0<<4)+1);
}
#define init_9_11_0_2
void FASTCALL call_9_11_0_2(SH2_struct* sh)
{
  SH2movwi(sh, 11, (0<<4)+2);
}
#define init_9_11_0_3
void FASTCALL call_9_11_0_3(SH2_struct* sh)
{
  SH2movwi(sh, 11, (0<<4)+3);
}
#define init_9_11_0_4
void FASTCALL call_9_11_0_4(SH2_struct* sh)
{
  SH2movwi(sh, 11, (0<<4)+4);
}
#define init_9_11_0_5
void FASTCALL call_9_11_0_5(SH2_struct* sh)
{
  SH2movwi(sh, 11, (0<<4)+5);
}
#define init_9_11_0_6
void FASTCALL call_9_11_0_6(SH2_struct* sh)
{
  SH2movwi(sh, 11, (0<<4)+6);
}
#define init_9_11_0_7
void FASTCALL call_9_11_0_7(SH2_struct* sh)
{
  SH2movwi(sh, 11, (0<<4)+7);
}
#define init_9_11_0_8
void FASTCALL call_9_11_0_8(SH2_struct* sh)
{
  SH2movwi(sh, 11, (0<<4)+8);
}
#define init_9_11_0_9
void FASTCALL call_9_11_0_9(SH2_struct* sh)
{
  SH2movwi(sh, 11, (0<<4)+9);
}
#define init_9_11_0_10
void FASTCALL call_9_11_0_10(SH2_struct* sh)
{
  SH2movwi(sh, 11, (0<<4)+10);
}
#define init_9_11_0_11
void FASTCALL call_9_11_0_11(SH2_struct* sh)
{
  SH2movwi(sh, 11, (0<<4)+11);
}
#define init_9_11_0_12
void FASTCALL call_9_11_0_12(SH2_struct* sh)
{
  SH2movwi(sh, 11, (0<<4)+12);
}
#define init_9_11_0_13
void FASTCALL call_9_11_0_13(SH2_struct* sh)
{
  SH2movwi(sh, 11, (0<<4)+13);
}
#define init_9_11_0_14
void FASTCALL call_9_11_0_14(SH2_struct* sh)
{
  SH2movwi(sh, 11, (0<<4)+14);
}
#define init_9_11_0_15
void FASTCALL call_9_11_0_15(SH2_struct* sh)
{
  SH2movwi(sh, 11, (0<<4)+15);
}
#define init_9_11_1_0
void FASTCALL call_9_11_1_0(SH2_struct* sh)
{
  SH2movwi(sh, 11, (1<<4)+0);
}
#define init_9_11_1_1
void FASTCALL call_9_11_1_1(SH2_struct* sh)
{
  SH2movwi(sh, 11, (1<<4)+1);
}
#define init_9_11_1_2
void FASTCALL call_9_11_1_2(SH2_struct* sh)
{
  SH2movwi(sh, 11, (1<<4)+2);
}
#define init_9_11_1_3
void FASTCALL call_9_11_1_3(SH2_struct* sh)
{
  SH2movwi(sh, 11, (1<<4)+3);
}
#define init_9_11_1_4
void FASTCALL call_9_11_1_4(SH2_struct* sh)
{
  SH2movwi(sh, 11, (1<<4)+4);
}
#define init_9_11_1_5
void FASTCALL call_9_11_1_5(SH2_struct* sh)
{
  SH2movwi(sh, 11, (1<<4)+5);
}
#define init_9_11_1_6
void FASTCALL call_9_11_1_6(SH2_struct* sh)
{
  SH2movwi(sh, 11, (1<<4)+6);
}
#define init_9_11_1_7
void FASTCALL call_9_11_1_7(SH2_struct* sh)
{
  SH2movwi(sh, 11, (1<<4)+7);
}
#define init_9_11_1_8
void FASTCALL call_9_11_1_8(SH2_struct* sh)
{
  SH2movwi(sh, 11, (1<<4)+8);
}
#define init_9_11_1_9
void FASTCALL call_9_11_1_9(SH2_struct* sh)
{
  SH2movwi(sh, 11, (1<<4)+9);
}
#define init_9_11_1_10
void FASTCALL call_9_11_1_10(SH2_struct* sh)
{
  SH2movwi(sh, 11, (1<<4)+10);
}
#define init_9_11_1_11
void FASTCALL call_9_11_1_11(SH2_struct* sh)
{
  SH2movwi(sh, 11, (1<<4)+11);
}
#define init_9_11_1_12
void FASTCALL call_9_11_1_12(SH2_struct* sh)
{
  SH2movwi(sh, 11, (1<<4)+12);
}
#define init_9_11_1_13
void FASTCALL call_9_11_1_13(SH2_struct* sh)
{
  SH2movwi(sh, 11, (1<<4)+13);
}
#define init_9_11_1_14
void FASTCALL call_9_11_1_14(SH2_struct* sh)
{
  SH2movwi(sh, 11, (1<<4)+14);
}
#define init_9_11_1_15
void FASTCALL call_9_11_1_15(SH2_struct* sh)
{
  SH2movwi(sh, 11, (1<<4)+15);
}
#define init_9_11_2_0
void FASTCALL call_9_11_2_0(SH2_struct* sh)
{
  SH2movwi(sh, 11, (2<<4)+0);
}
#define init_9_11_2_1
void FASTCALL call_9_11_2_1(SH2_struct* sh)
{
  SH2movwi(sh, 11, (2<<4)+1);
}
#define init_9_11_2_2
void FASTCALL call_9_11_2_2(SH2_struct* sh)
{
  SH2movwi(sh, 11, (2<<4)+2);
}
#define init_9_11_2_3
void FASTCALL call_9_11_2_3(SH2_struct* sh)
{
  SH2movwi(sh, 11, (2<<4)+3);
}
#define init_9_11_2_4
void FASTCALL call_9_11_2_4(SH2_struct* sh)
{
  SH2movwi(sh, 11, (2<<4)+4);
}
#define init_9_11_2_5
void FASTCALL call_9_11_2_5(SH2_struct* sh)
{
  SH2movwi(sh, 11, (2<<4)+5);
}
#define init_9_11_2_6
void FASTCALL call_9_11_2_6(SH2_struct* sh)
{
  SH2movwi(sh, 11, (2<<4)+6);
}
#define init_9_11_2_7
void FASTCALL call_9_11_2_7(SH2_struct* sh)
{
  SH2movwi(sh, 11, (2<<4)+7);
}
#define init_9_11_2_8
void FASTCALL call_9_11_2_8(SH2_struct* sh)
{
  SH2movwi(sh, 11, (2<<4)+8);
}
#define init_9_11_2_9
void FASTCALL call_9_11_2_9(SH2_struct* sh)
{
  SH2movwi(sh, 11, (2<<4)+9);
}
#define init_9_11_2_10
void FASTCALL call_9_11_2_10(SH2_struct* sh)
{
  SH2movwi(sh, 11, (2<<4)+10);
}
#define init_9_11_2_11
void FASTCALL call_9_11_2_11(SH2_struct* sh)
{
  SH2movwi(sh, 11, (2<<4)+11);
}
#define init_9_11_2_12
void FASTCALL call_9_11_2_12(SH2_struct* sh)
{
  SH2movwi(sh, 11, (2<<4)+12);
}
#define init_9_11_2_13
void FASTCALL call_9_11_2_13(SH2_struct* sh)
{
  SH2movwi(sh, 11, (2<<4)+13);
}
#define init_9_11_2_14
void FASTCALL call_9_11_2_14(SH2_struct* sh)
{
  SH2movwi(sh, 11, (2<<4)+14);
}
#define init_9_11_2_15
void FASTCALL call_9_11_2_15(SH2_struct* sh)
{
  SH2movwi(sh, 11, (2<<4)+15);
}
#define init_9_11_3_0
void FASTCALL call_9_11_3_0(SH2_struct* sh)
{
  SH2movwi(sh, 11, (3<<4)+0);
}
#define init_9_11_3_1
void FASTCALL call_9_11_3_1(SH2_struct* sh)
{
  SH2movwi(sh, 11, (3<<4)+1);
}
#define init_9_11_3_2
void FASTCALL call_9_11_3_2(SH2_struct* sh)
{
  SH2movwi(sh, 11, (3<<4)+2);
}
#define init_9_11_3_3
void FASTCALL call_9_11_3_3(SH2_struct* sh)
{
  SH2movwi(sh, 11, (3<<4)+3);
}
#define init_9_11_3_4
void FASTCALL call_9_11_3_4(SH2_struct* sh)
{
  SH2movwi(sh, 11, (3<<4)+4);
}
#define init_9_11_3_5
void FASTCALL call_9_11_3_5(SH2_struct* sh)
{
  SH2movwi(sh, 11, (3<<4)+5);
}
#define init_9_11_3_6
void FASTCALL call_9_11_3_6(SH2_struct* sh)
{
  SH2movwi(sh, 11, (3<<4)+6);
}
#define init_9_11_3_7
void FASTCALL call_9_11_3_7(SH2_struct* sh)
{
  SH2movwi(sh, 11, (3<<4)+7);
}
#define init_9_11_3_8
void FASTCALL call_9_11_3_8(SH2_struct* sh)
{
  SH2movwi(sh, 11, (3<<4)+8);
}
#define init_9_11_3_9
void FASTCALL call_9_11_3_9(SH2_struct* sh)
{
  SH2movwi(sh, 11, (3<<4)+9);
}
#define init_9_11_3_10
void FASTCALL call_9_11_3_10(SH2_struct* sh)
{
  SH2movwi(sh, 11, (3<<4)+10);
}
#define init_9_11_3_11
void FASTCALL call_9_11_3_11(SH2_struct* sh)
{
  SH2movwi(sh, 11, (3<<4)+11);
}
#define init_9_11_3_12
void FASTCALL call_9_11_3_12(SH2_struct* sh)
{
  SH2movwi(sh, 11, (3<<4)+12);
}
#define init_9_11_3_13
void FASTCALL call_9_11_3_13(SH2_struct* sh)
{
  SH2movwi(sh, 11, (3<<4)+13);
}
#define init_9_11_3_14
void FASTCALL call_9_11_3_14(SH2_struct* sh)
{
  SH2movwi(sh, 11, (3<<4)+14);
}
#define init_9_11_3_15
void FASTCALL call_9_11_3_15(SH2_struct* sh)
{
  SH2movwi(sh, 11, (3<<4)+15);
}
#define init_9_11_4_0
void FASTCALL call_9_11_4_0(SH2_struct* sh)
{
  SH2movwi(sh, 11, (4<<4)+0);
}
#define init_9_11_4_1
void FASTCALL call_9_11_4_1(SH2_struct* sh)
{
  SH2movwi(sh, 11, (4<<4)+1);
}
#define init_9_11_4_2
void FASTCALL call_9_11_4_2(SH2_struct* sh)
{
  SH2movwi(sh, 11, (4<<4)+2);
}
#define init_9_11_4_3
void FASTCALL call_9_11_4_3(SH2_struct* sh)
{
  SH2movwi(sh, 11, (4<<4)+3);
}
#define init_9_11_4_4
void FASTCALL call_9_11_4_4(SH2_struct* sh)
{
  SH2movwi(sh, 11, (4<<4)+4);
}
#define init_9_11_4_5
void FASTCALL call_9_11_4_5(SH2_struct* sh)
{
  SH2movwi(sh, 11, (4<<4)+5);
}
#define init_9_11_4_6
void FASTCALL call_9_11_4_6(SH2_struct* sh)
{
  SH2movwi(sh, 11, (4<<4)+6);
}
#define init_9_11_4_7
void FASTCALL call_9_11_4_7(SH2_struct* sh)
{
  SH2movwi(sh, 11, (4<<4)+7);
}
#define init_9_11_4_8
void FASTCALL call_9_11_4_8(SH2_struct* sh)
{
  SH2movwi(sh, 11, (4<<4)+8);
}
#define init_9_11_4_9
void FASTCALL call_9_11_4_9(SH2_struct* sh)
{
  SH2movwi(sh, 11, (4<<4)+9);
}
#define init_9_11_4_10
void FASTCALL call_9_11_4_10(SH2_struct* sh)
{
  SH2movwi(sh, 11, (4<<4)+10);
}
#define init_9_11_4_11
void FASTCALL call_9_11_4_11(SH2_struct* sh)
{
  SH2movwi(sh, 11, (4<<4)+11);
}
#define init_9_11_4_12
void FASTCALL call_9_11_4_12(SH2_struct* sh)
{
  SH2movwi(sh, 11, (4<<4)+12);
}
#define init_9_11_4_13
void FASTCALL call_9_11_4_13(SH2_struct* sh)
{
  SH2movwi(sh, 11, (4<<4)+13);
}
#define init_9_11_4_14
void FASTCALL call_9_11_4_14(SH2_struct* sh)
{
  SH2movwi(sh, 11, (4<<4)+14);
}
#define init_9_11_4_15
void FASTCALL call_9_11_4_15(SH2_struct* sh)
{
  SH2movwi(sh, 11, (4<<4)+15);
}
#define init_9_11_5_0
void FASTCALL call_9_11_5_0(SH2_struct* sh)
{
  SH2movwi(sh, 11, (5<<4)+0);
}
#define init_9_11_5_1
void FASTCALL call_9_11_5_1(SH2_struct* sh)
{
  SH2movwi(sh, 11, (5<<4)+1);
}
#define init_9_11_5_2
void FASTCALL call_9_11_5_2(SH2_struct* sh)
{
  SH2movwi(sh, 11, (5<<4)+2);
}
#define init_9_11_5_3
void FASTCALL call_9_11_5_3(SH2_struct* sh)
{
  SH2movwi(sh, 11, (5<<4)+3);
}
#define init_9_11_5_4
void FASTCALL call_9_11_5_4(SH2_struct* sh)
{
  SH2movwi(sh, 11, (5<<4)+4);
}
#define init_9_11_5_5
void FASTCALL call_9_11_5_5(SH2_struct* sh)
{
  SH2movwi(sh, 11, (5<<4)+5);
}
#define init_9_11_5_6
void FASTCALL call_9_11_5_6(SH2_struct* sh)
{
  SH2movwi(sh, 11, (5<<4)+6);
}
#define init_9_11_5_7
void FASTCALL call_9_11_5_7(SH2_struct* sh)
{
  SH2movwi(sh, 11, (5<<4)+7);
}
#define init_9_11_5_8
void FASTCALL call_9_11_5_8(SH2_struct* sh)
{
  SH2movwi(sh, 11, (5<<4)+8);
}
#define init_9_11_5_9
void FASTCALL call_9_11_5_9(SH2_struct* sh)
{
  SH2movwi(sh, 11, (5<<4)+9);
}
#define init_9_11_5_10
void FASTCALL call_9_11_5_10(SH2_struct* sh)
{
  SH2movwi(sh, 11, (5<<4)+10);
}
#define init_9_11_5_11
void FASTCALL call_9_11_5_11(SH2_struct* sh)
{
  SH2movwi(sh, 11, (5<<4)+11);
}
#define init_9_11_5_12
void FASTCALL call_9_11_5_12(SH2_struct* sh)
{
  SH2movwi(sh, 11, (5<<4)+12);
}
#define init_9_11_5_13
void FASTCALL call_9_11_5_13(SH2_struct* sh)
{
  SH2movwi(sh, 11, (5<<4)+13);
}
#define init_9_11_5_14
void FASTCALL call_9_11_5_14(SH2_struct* sh)
{
  SH2movwi(sh, 11, (5<<4)+14);
}
#define init_9_11_5_15
void FASTCALL call_9_11_5_15(SH2_struct* sh)
{
  SH2movwi(sh, 11, (5<<4)+15);
}
#define init_9_11_6_0
void FASTCALL call_9_11_6_0(SH2_struct* sh)
{
  SH2movwi(sh, 11, (6<<4)+0);
}
#define init_9_11_6_1
void FASTCALL call_9_11_6_1(SH2_struct* sh)
{
  SH2movwi(sh, 11, (6<<4)+1);
}
#define init_9_11_6_2
void FASTCALL call_9_11_6_2(SH2_struct* sh)
{
  SH2movwi(sh, 11, (6<<4)+2);
}
#define init_9_11_6_3
void FASTCALL call_9_11_6_3(SH2_struct* sh)
{
  SH2movwi(sh, 11, (6<<4)+3);
}
#define init_9_11_6_4
void FASTCALL call_9_11_6_4(SH2_struct* sh)
{
  SH2movwi(sh, 11, (6<<4)+4);
}
#define init_9_11_6_5
void FASTCALL call_9_11_6_5(SH2_struct* sh)
{
  SH2movwi(sh, 11, (6<<4)+5);
}
#define init_9_11_6_6
void FASTCALL call_9_11_6_6(SH2_struct* sh)
{
  SH2movwi(sh, 11, (6<<4)+6);
}
#define init_9_11_6_7
void FASTCALL call_9_11_6_7(SH2_struct* sh)
{
  SH2movwi(sh, 11, (6<<4)+7);
}
#define init_9_11_6_8
void FASTCALL call_9_11_6_8(SH2_struct* sh)
{
  SH2movwi(sh, 11, (6<<4)+8);
}
#define init_9_11_6_9
void FASTCALL call_9_11_6_9(SH2_struct* sh)
{
  SH2movwi(sh, 11, (6<<4)+9);
}
#define init_9_11_6_10
void FASTCALL call_9_11_6_10(SH2_struct* sh)
{
  SH2movwi(sh, 11, (6<<4)+10);
}
#define init_9_11_6_11
void FASTCALL call_9_11_6_11(SH2_struct* sh)
{
  SH2movwi(sh, 11, (6<<4)+11);
}
#define init_9_11_6_12
void FASTCALL call_9_11_6_12(SH2_struct* sh)
{
  SH2movwi(sh, 11, (6<<4)+12);
}
#define init_9_11_6_13
void FASTCALL call_9_11_6_13(SH2_struct* sh)
{
  SH2movwi(sh, 11, (6<<4)+13);
}
#define init_9_11_6_14
void FASTCALL call_9_11_6_14(SH2_struct* sh)
{
  SH2movwi(sh, 11, (6<<4)+14);
}
#define init_9_11_6_15
void FASTCALL call_9_11_6_15(SH2_struct* sh)
{
  SH2movwi(sh, 11, (6<<4)+15);
}
#define init_9_11_7_0
void FASTCALL call_9_11_7_0(SH2_struct* sh)
{
  SH2movwi(sh, 11, (7<<4)+0);
}
#define init_9_11_7_1
void FASTCALL call_9_11_7_1(SH2_struct* sh)
{
  SH2movwi(sh, 11, (7<<4)+1);
}
#define init_9_11_7_2
void FASTCALL call_9_11_7_2(SH2_struct* sh)
{
  SH2movwi(sh, 11, (7<<4)+2);
}
#define init_9_11_7_3
void FASTCALL call_9_11_7_3(SH2_struct* sh)
{
  SH2movwi(sh, 11, (7<<4)+3);
}
#define init_9_11_7_4
void FASTCALL call_9_11_7_4(SH2_struct* sh)
{
  SH2movwi(sh, 11, (7<<4)+4);
}
#define init_9_11_7_5
void FASTCALL call_9_11_7_5(SH2_struct* sh)
{
  SH2movwi(sh, 11, (7<<4)+5);
}
#define init_9_11_7_6
void FASTCALL call_9_11_7_6(SH2_struct* sh)
{
  SH2movwi(sh, 11, (7<<4)+6);
}
#define init_9_11_7_7
void FASTCALL call_9_11_7_7(SH2_struct* sh)
{
  SH2movwi(sh, 11, (7<<4)+7);
}
#define init_9_11_7_8
void FASTCALL call_9_11_7_8(SH2_struct* sh)
{
  SH2movwi(sh, 11, (7<<4)+8);
}
#define init_9_11_7_9
void FASTCALL call_9_11_7_9(SH2_struct* sh)
{
  SH2movwi(sh, 11, (7<<4)+9);
}
#define init_9_11_7_10
void FASTCALL call_9_11_7_10(SH2_struct* sh)
{
  SH2movwi(sh, 11, (7<<4)+10);
}
#define init_9_11_7_11
void FASTCALL call_9_11_7_11(SH2_struct* sh)
{
  SH2movwi(sh, 11, (7<<4)+11);
}
#define init_9_11_7_12
void FASTCALL call_9_11_7_12(SH2_struct* sh)
{
  SH2movwi(sh, 11, (7<<4)+12);
}
#define init_9_11_7_13
void FASTCALL call_9_11_7_13(SH2_struct* sh)
{
  SH2movwi(sh, 11, (7<<4)+13);
}
#define init_9_11_7_14
void FASTCALL call_9_11_7_14(SH2_struct* sh)
{
  SH2movwi(sh, 11, (7<<4)+14);
}
#define init_9_11_7_15
void FASTCALL call_9_11_7_15(SH2_struct* sh)
{
  SH2movwi(sh, 11, (7<<4)+15);
}
#define init_9_11_8_0
void FASTCALL call_9_11_8_0(SH2_struct* sh)
{
  SH2movwi(sh, 11, (8<<4)+0);
}
#define init_9_11_8_1
void FASTCALL call_9_11_8_1(SH2_struct* sh)
{
  SH2movwi(sh, 11, (8<<4)+1);
}
#define init_9_11_8_2
void FASTCALL call_9_11_8_2(SH2_struct* sh)
{
  SH2movwi(sh, 11, (8<<4)+2);
}
#define init_9_11_8_3
void FASTCALL call_9_11_8_3(SH2_struct* sh)
{
  SH2movwi(sh, 11, (8<<4)+3);
}
#define init_9_11_8_4
void FASTCALL call_9_11_8_4(SH2_struct* sh)
{
  SH2movwi(sh, 11, (8<<4)+4);
}
#define init_9_11_8_5
void FASTCALL call_9_11_8_5(SH2_struct* sh)
{
  SH2movwi(sh, 11, (8<<4)+5);
}
#define init_9_11_8_6
void FASTCALL call_9_11_8_6(SH2_struct* sh)
{
  SH2movwi(sh, 11, (8<<4)+6);
}
#define init_9_11_8_7
void FASTCALL call_9_11_8_7(SH2_struct* sh)
{
  SH2movwi(sh, 11, (8<<4)+7);
}
#define init_9_11_8_8
void FASTCALL call_9_11_8_8(SH2_struct* sh)
{
  SH2movwi(sh, 11, (8<<4)+8);
}
#define init_9_11_8_9
void FASTCALL call_9_11_8_9(SH2_struct* sh)
{
  SH2movwi(sh, 11, (8<<4)+9);
}
#define init_9_11_8_10
void FASTCALL call_9_11_8_10(SH2_struct* sh)
{
  SH2movwi(sh, 11, (8<<4)+10);
}
#define init_9_11_8_11
void FASTCALL call_9_11_8_11(SH2_struct* sh)
{
  SH2movwi(sh, 11, (8<<4)+11);
}
#define init_9_11_8_12
void FASTCALL call_9_11_8_12(SH2_struct* sh)
{
  SH2movwi(sh, 11, (8<<4)+12);
}
#define init_9_11_8_13
void FASTCALL call_9_11_8_13(SH2_struct* sh)
{
  SH2movwi(sh, 11, (8<<4)+13);
}
#define init_9_11_8_14
void FASTCALL call_9_11_8_14(SH2_struct* sh)
{
  SH2movwi(sh, 11, (8<<4)+14);
}
#define init_9_11_8_15
void FASTCALL call_9_11_8_15(SH2_struct* sh)
{
  SH2movwi(sh, 11, (8<<4)+15);
}
#define init_9_11_9_0
void FASTCALL call_9_11_9_0(SH2_struct* sh)
{
  SH2movwi(sh, 11, (9<<4)+0);
}
#define init_9_11_9_1
void FASTCALL call_9_11_9_1(SH2_struct* sh)
{
  SH2movwi(sh, 11, (9<<4)+1);
}
#define init_9_11_9_2
void FASTCALL call_9_11_9_2(SH2_struct* sh)
{
  SH2movwi(sh, 11, (9<<4)+2);
}
#define init_9_11_9_3
void FASTCALL call_9_11_9_3(SH2_struct* sh)
{
  SH2movwi(sh, 11, (9<<4)+3);
}
#define init_9_11_9_4
void FASTCALL call_9_11_9_4(SH2_struct* sh)
{
  SH2movwi(sh, 11, (9<<4)+4);
}
#define init_9_11_9_5
void FASTCALL call_9_11_9_5(SH2_struct* sh)
{
  SH2movwi(sh, 11, (9<<4)+5);
}
#define init_9_11_9_6
void FASTCALL call_9_11_9_6(SH2_struct* sh)
{
  SH2movwi(sh, 11, (9<<4)+6);
}
#define init_9_11_9_7
void FASTCALL call_9_11_9_7(SH2_struct* sh)
{
  SH2movwi(sh, 11, (9<<4)+7);
}
#define init_9_11_9_8
void FASTCALL call_9_11_9_8(SH2_struct* sh)
{
  SH2movwi(sh, 11, (9<<4)+8);
}
#define init_9_11_9_9
void FASTCALL call_9_11_9_9(SH2_struct* sh)
{
  SH2movwi(sh, 11, (9<<4)+9);
}
#define init_9_11_9_10
void FASTCALL call_9_11_9_10(SH2_struct* sh)
{
  SH2movwi(sh, 11, (9<<4)+10);
}
#define init_9_11_9_11
void FASTCALL call_9_11_9_11(SH2_struct* sh)
{
  SH2movwi(sh, 11, (9<<4)+11);
}
#define init_9_11_9_12
void FASTCALL call_9_11_9_12(SH2_struct* sh)
{
  SH2movwi(sh, 11, (9<<4)+12);
}
#define init_9_11_9_13
void FASTCALL call_9_11_9_13(SH2_struct* sh)
{
  SH2movwi(sh, 11, (9<<4)+13);
}
#define init_9_11_9_14
void FASTCALL call_9_11_9_14(SH2_struct* sh)
{
  SH2movwi(sh, 11, (9<<4)+14);
}
#define init_9_11_9_15
void FASTCALL call_9_11_9_15(SH2_struct* sh)
{
  SH2movwi(sh, 11, (9<<4)+15);
}
#define init_9_11_10_0
void FASTCALL call_9_11_10_0(SH2_struct* sh)
{
  SH2movwi(sh, 11, (10<<4)+0);
}
#define init_9_11_10_1
void FASTCALL call_9_11_10_1(SH2_struct* sh)
{
  SH2movwi(sh, 11, (10<<4)+1);
}
#define init_9_11_10_2
void FASTCALL call_9_11_10_2(SH2_struct* sh)
{
  SH2movwi(sh, 11, (10<<4)+2);
}
#define init_9_11_10_3
void FASTCALL call_9_11_10_3(SH2_struct* sh)
{
  SH2movwi(sh, 11, (10<<4)+3);
}
#define init_9_11_10_4
void FASTCALL call_9_11_10_4(SH2_struct* sh)
{
  SH2movwi(sh, 11, (10<<4)+4);
}
#define init_9_11_10_5
void FASTCALL call_9_11_10_5(SH2_struct* sh)
{
  SH2movwi(sh, 11, (10<<4)+5);
}
#define init_9_11_10_6
void FASTCALL call_9_11_10_6(SH2_struct* sh)
{
  SH2movwi(sh, 11, (10<<4)+6);
}
#define init_9_11_10_7
void FASTCALL call_9_11_10_7(SH2_struct* sh)
{
  SH2movwi(sh, 11, (10<<4)+7);
}
#define init_9_11_10_8
void FASTCALL call_9_11_10_8(SH2_struct* sh)
{
  SH2movwi(sh, 11, (10<<4)+8);
}
#define init_9_11_10_9
void FASTCALL call_9_11_10_9(SH2_struct* sh)
{
  SH2movwi(sh, 11, (10<<4)+9);
}
#define init_9_11_10_10
void FASTCALL call_9_11_10_10(SH2_struct* sh)
{
  SH2movwi(sh, 11, (10<<4)+10);
}
#define init_9_11_10_11
void FASTCALL call_9_11_10_11(SH2_struct* sh)
{
  SH2movwi(sh, 11, (10<<4)+11);
}
#define init_9_11_10_12
void FASTCALL call_9_11_10_12(SH2_struct* sh)
{
  SH2movwi(sh, 11, (10<<4)+12);
}
#define init_9_11_10_13
void FASTCALL call_9_11_10_13(SH2_struct* sh)
{
  SH2movwi(sh, 11, (10<<4)+13);
}
#define init_9_11_10_14
void FASTCALL call_9_11_10_14(SH2_struct* sh)
{
  SH2movwi(sh, 11, (10<<4)+14);
}
#define init_9_11_10_15
void FASTCALL call_9_11_10_15(SH2_struct* sh)
{
  SH2movwi(sh, 11, (10<<4)+15);
}
#define init_9_11_11_0
void FASTCALL call_9_11_11_0(SH2_struct* sh)
{
  SH2movwi(sh, 11, (11<<4)+0);
}
#define init_9_11_11_1
void FASTCALL call_9_11_11_1(SH2_struct* sh)
{
  SH2movwi(sh, 11, (11<<4)+1);
}
#define init_9_11_11_2
void FASTCALL call_9_11_11_2(SH2_struct* sh)
{
  SH2movwi(sh, 11, (11<<4)+2);
}
#define init_9_11_11_3
void FASTCALL call_9_11_11_3(SH2_struct* sh)
{
  SH2movwi(sh, 11, (11<<4)+3);
}
#define init_9_11_11_4
void FASTCALL call_9_11_11_4(SH2_struct* sh)
{
  SH2movwi(sh, 11, (11<<4)+4);
}
#define init_9_11_11_5
void FASTCALL call_9_11_11_5(SH2_struct* sh)
{
  SH2movwi(sh, 11, (11<<4)+5);
}
#define init_9_11_11_6
void FASTCALL call_9_11_11_6(SH2_struct* sh)
{
  SH2movwi(sh, 11, (11<<4)+6);
}
#define init_9_11_11_7
void FASTCALL call_9_11_11_7(SH2_struct* sh)
{
  SH2movwi(sh, 11, (11<<4)+7);
}
#define init_9_11_11_8
void FASTCALL call_9_11_11_8(SH2_struct* sh)
{
  SH2movwi(sh, 11, (11<<4)+8);
}
#define init_9_11_11_9
void FASTCALL call_9_11_11_9(SH2_struct* sh)
{
  SH2movwi(sh, 11, (11<<4)+9);
}
#define init_9_11_11_10
void FASTCALL call_9_11_11_10(SH2_struct* sh)
{
  SH2movwi(sh, 11, (11<<4)+10);
}
#define init_9_11_11_11
void FASTCALL call_9_11_11_11(SH2_struct* sh)
{
  SH2movwi(sh, 11, (11<<4)+11);
}
#define init_9_11_11_12
void FASTCALL call_9_11_11_12(SH2_struct* sh)
{
  SH2movwi(sh, 11, (11<<4)+12);
}
#define init_9_11_11_13
void FASTCALL call_9_11_11_13(SH2_struct* sh)
{
  SH2movwi(sh, 11, (11<<4)+13);
}
#define init_9_11_11_14
void FASTCALL call_9_11_11_14(SH2_struct* sh)
{
  SH2movwi(sh, 11, (11<<4)+14);
}
#define init_9_11_11_15
void FASTCALL call_9_11_11_15(SH2_struct* sh)
{
  SH2movwi(sh, 11, (11<<4)+15);
}
#define init_9_11_12_0
void FASTCALL call_9_11_12_0(SH2_struct* sh)
{
  SH2movwi(sh, 11, (12<<4)+0);
}
#define init_9_11_12_1
void FASTCALL call_9_11_12_1(SH2_struct* sh)
{
  SH2movwi(sh, 11, (12<<4)+1);
}
#define init_9_11_12_2
void FASTCALL call_9_11_12_2(SH2_struct* sh)
{
  SH2movwi(sh, 11, (12<<4)+2);
}
#define init_9_11_12_3
void FASTCALL call_9_11_12_3(SH2_struct* sh)
{
  SH2movwi(sh, 11, (12<<4)+3);
}
#define init_9_11_12_4
void FASTCALL call_9_11_12_4(SH2_struct* sh)
{
  SH2movwi(sh, 11, (12<<4)+4);
}
#define init_9_11_12_5
void FASTCALL call_9_11_12_5(SH2_struct* sh)
{
  SH2movwi(sh, 11, (12<<4)+5);
}
#define init_9_11_12_6
void FASTCALL call_9_11_12_6(SH2_struct* sh)
{
  SH2movwi(sh, 11, (12<<4)+6);
}
#define init_9_11_12_7
void FASTCALL call_9_11_12_7(SH2_struct* sh)
{
  SH2movwi(sh, 11, (12<<4)+7);
}
#define init_9_11_12_8
void FASTCALL call_9_11_12_8(SH2_struct* sh)
{
  SH2movwi(sh, 11, (12<<4)+8);
}
#define init_9_11_12_9
void FASTCALL call_9_11_12_9(SH2_struct* sh)
{
  SH2movwi(sh, 11, (12<<4)+9);
}
#define init_9_11_12_10
void FASTCALL call_9_11_12_10(SH2_struct* sh)
{
  SH2movwi(sh, 11, (12<<4)+10);
}
#define init_9_11_12_11
void FASTCALL call_9_11_12_11(SH2_struct* sh)
{
  SH2movwi(sh, 11, (12<<4)+11);
}
#define init_9_11_12_12
void FASTCALL call_9_11_12_12(SH2_struct* sh)
{
  SH2movwi(sh, 11, (12<<4)+12);
}
#define init_9_11_12_13
void FASTCALL call_9_11_12_13(SH2_struct* sh)
{
  SH2movwi(sh, 11, (12<<4)+13);
}
#define init_9_11_12_14
void FASTCALL call_9_11_12_14(SH2_struct* sh)
{
  SH2movwi(sh, 11, (12<<4)+14);
}
#define init_9_11_12_15
void FASTCALL call_9_11_12_15(SH2_struct* sh)
{
  SH2movwi(sh, 11, (12<<4)+15);
}
#define init_9_11_13_0
void FASTCALL call_9_11_13_0(SH2_struct* sh)
{
  SH2movwi(sh, 11, (13<<4)+0);
}
#define init_9_11_13_1
void FASTCALL call_9_11_13_1(SH2_struct* sh)
{
  SH2movwi(sh, 11, (13<<4)+1);
}
#define init_9_11_13_2
void FASTCALL call_9_11_13_2(SH2_struct* sh)
{
  SH2movwi(sh, 11, (13<<4)+2);
}
#define init_9_11_13_3
void FASTCALL call_9_11_13_3(SH2_struct* sh)
{
  SH2movwi(sh, 11, (13<<4)+3);
}
#define init_9_11_13_4
void FASTCALL call_9_11_13_4(SH2_struct* sh)
{
  SH2movwi(sh, 11, (13<<4)+4);
}
#define init_9_11_13_5
void FASTCALL call_9_11_13_5(SH2_struct* sh)
{
  SH2movwi(sh, 11, (13<<4)+5);
}
#define init_9_11_13_6
void FASTCALL call_9_11_13_6(SH2_struct* sh)
{
  SH2movwi(sh, 11, (13<<4)+6);
}
#define init_9_11_13_7
void FASTCALL call_9_11_13_7(SH2_struct* sh)
{
  SH2movwi(sh, 11, (13<<4)+7);
}
#define init_9_11_13_8
void FASTCALL call_9_11_13_8(SH2_struct* sh)
{
  SH2movwi(sh, 11, (13<<4)+8);
}
#define init_9_11_13_9
void FASTCALL call_9_11_13_9(SH2_struct* sh)
{
  SH2movwi(sh, 11, (13<<4)+9);
}
#define init_9_11_13_10
void FASTCALL call_9_11_13_10(SH2_struct* sh)
{
  SH2movwi(sh, 11, (13<<4)+10);
}
#define init_9_11_13_11
void FASTCALL call_9_11_13_11(SH2_struct* sh)
{
  SH2movwi(sh, 11, (13<<4)+11);
}
#define init_9_11_13_12
void FASTCALL call_9_11_13_12(SH2_struct* sh)
{
  SH2movwi(sh, 11, (13<<4)+12);
}
#define init_9_11_13_13
void FASTCALL call_9_11_13_13(SH2_struct* sh)
{
  SH2movwi(sh, 11, (13<<4)+13);
}
#define init_9_11_13_14
void FASTCALL call_9_11_13_14(SH2_struct* sh)
{
  SH2movwi(sh, 11, (13<<4)+14);
}
#define init_9_11_13_15
void FASTCALL call_9_11_13_15(SH2_struct* sh)
{
  SH2movwi(sh, 11, (13<<4)+15);
}
#define init_9_11_14_0
void FASTCALL call_9_11_14_0(SH2_struct* sh)
{
  SH2movwi(sh, 11, (14<<4)+0);
}
#define init_9_11_14_1
void FASTCALL call_9_11_14_1(SH2_struct* sh)
{
  SH2movwi(sh, 11, (14<<4)+1);
}
#define init_9_11_14_2
void FASTCALL call_9_11_14_2(SH2_struct* sh)
{
  SH2movwi(sh, 11, (14<<4)+2);
}
#define init_9_11_14_3
void FASTCALL call_9_11_14_3(SH2_struct* sh)
{
  SH2movwi(sh, 11, (14<<4)+3);
}
#define init_9_11_14_4
void FASTCALL call_9_11_14_4(SH2_struct* sh)
{
  SH2movwi(sh, 11, (14<<4)+4);
}
#define init_9_11_14_5
void FASTCALL call_9_11_14_5(SH2_struct* sh)
{
  SH2movwi(sh, 11, (14<<4)+5);
}
#define init_9_11_14_6
void FASTCALL call_9_11_14_6(SH2_struct* sh)
{
  SH2movwi(sh, 11, (14<<4)+6);
}
#define init_9_11_14_7
void FASTCALL call_9_11_14_7(SH2_struct* sh)
{
  SH2movwi(sh, 11, (14<<4)+7);
}
#define init_9_11_14_8
void FASTCALL call_9_11_14_8(SH2_struct* sh)
{
  SH2movwi(sh, 11, (14<<4)+8);
}
#define init_9_11_14_9
void FASTCALL call_9_11_14_9(SH2_struct* sh)
{
  SH2movwi(sh, 11, (14<<4)+9);
}
#define init_9_11_14_10
void FASTCALL call_9_11_14_10(SH2_struct* sh)
{
  SH2movwi(sh, 11, (14<<4)+10);
}
#define init_9_11_14_11
void FASTCALL call_9_11_14_11(SH2_struct* sh)
{
  SH2movwi(sh, 11, (14<<4)+11);
}
#define init_9_11_14_12
void FASTCALL call_9_11_14_12(SH2_struct* sh)
{
  SH2movwi(sh, 11, (14<<4)+12);
}
#define init_9_11_14_13
void FASTCALL call_9_11_14_13(SH2_struct* sh)
{
  SH2movwi(sh, 11, (14<<4)+13);
}
#define init_9_11_14_14
void FASTCALL call_9_11_14_14(SH2_struct* sh)
{
  SH2movwi(sh, 11, (14<<4)+14);
}
#define init_9_11_14_15
void FASTCALL call_9_11_14_15(SH2_struct* sh)
{
  SH2movwi(sh, 11, (14<<4)+15);
}
#define init_9_11_15_0
void FASTCALL call_9_11_15_0(SH2_struct* sh)
{
  SH2movwi(sh, 11, (15<<4)+0);
}
#define init_9_11_15_1
void FASTCALL call_9_11_15_1(SH2_struct* sh)
{
  SH2movwi(sh, 11, (15<<4)+1);
}
#define init_9_11_15_2
void FASTCALL call_9_11_15_2(SH2_struct* sh)
{
  SH2movwi(sh, 11, (15<<4)+2);
}
#define init_9_11_15_3
void FASTCALL call_9_11_15_3(SH2_struct* sh)
{
  SH2movwi(sh, 11, (15<<4)+3);
}
#define init_9_11_15_4
void FASTCALL call_9_11_15_4(SH2_struct* sh)
{
  SH2movwi(sh, 11, (15<<4)+4);
}
#define init_9_11_15_5
void FASTCALL call_9_11_15_5(SH2_struct* sh)
{
  SH2movwi(sh, 11, (15<<4)+5);
}
#define init_9_11_15_6
void FASTCALL call_9_11_15_6(SH2_struct* sh)
{
  SH2movwi(sh, 11, (15<<4)+6);
}
#define init_9_11_15_7
void FASTCALL call_9_11_15_7(SH2_struct* sh)
{
  SH2movwi(sh, 11, (15<<4)+7);
}
#define init_9_11_15_8
void FASTCALL call_9_11_15_8(SH2_struct* sh)
{
  SH2movwi(sh, 11, (15<<4)+8);
}
#define init_9_11_15_9
void FASTCALL call_9_11_15_9(SH2_struct* sh)
{
  SH2movwi(sh, 11, (15<<4)+9);
}
#define init_9_11_15_10
void FASTCALL call_9_11_15_10(SH2_struct* sh)
{
  SH2movwi(sh, 11, (15<<4)+10);
}
#define init_9_11_15_11
void FASTCALL call_9_11_15_11(SH2_struct* sh)
{
  SH2movwi(sh, 11, (15<<4)+11);
}
#define init_9_11_15_12
void FASTCALL call_9_11_15_12(SH2_struct* sh)
{
  SH2movwi(sh, 11, (15<<4)+12);
}
#define init_9_11_15_13
void FASTCALL call_9_11_15_13(SH2_struct* sh)
{
  SH2movwi(sh, 11, (15<<4)+13);
}
#define init_9_11_15_14
void FASTCALL call_9_11_15_14(SH2_struct* sh)
{
  SH2movwi(sh, 11, (15<<4)+14);
}
#define init_9_11_15_15
void FASTCALL call_9_11_15_15(SH2_struct* sh)
{
  SH2movwi(sh, 11, (15<<4)+15);
}
#define init_9_12_0_0
void FASTCALL call_9_12_0_0(SH2_struct* sh)
{
  SH2movwi(sh, 12, (0<<4)+0);
}
#define init_9_12_0_1
void FASTCALL call_9_12_0_1(SH2_struct* sh)
{
  SH2movwi(sh, 12, (0<<4)+1);
}
#define init_9_12_0_2
void FASTCALL call_9_12_0_2(SH2_struct* sh)
{
  SH2movwi(sh, 12, (0<<4)+2);
}
#define init_9_12_0_3
void FASTCALL call_9_12_0_3(SH2_struct* sh)
{
  SH2movwi(sh, 12, (0<<4)+3);
}
#define init_9_12_0_4
void FASTCALL call_9_12_0_4(SH2_struct* sh)
{
  SH2movwi(sh, 12, (0<<4)+4);
}
#define init_9_12_0_5
void FASTCALL call_9_12_0_5(SH2_struct* sh)
{
  SH2movwi(sh, 12, (0<<4)+5);
}
#define init_9_12_0_6
void FASTCALL call_9_12_0_6(SH2_struct* sh)
{
  SH2movwi(sh, 12, (0<<4)+6);
}
#define init_9_12_0_7
void FASTCALL call_9_12_0_7(SH2_struct* sh)
{
  SH2movwi(sh, 12, (0<<4)+7);
}
#define init_9_12_0_8
void FASTCALL call_9_12_0_8(SH2_struct* sh)
{
  SH2movwi(sh, 12, (0<<4)+8);
}
#define init_9_12_0_9
void FASTCALL call_9_12_0_9(SH2_struct* sh)
{
  SH2movwi(sh, 12, (0<<4)+9);
}
#define init_9_12_0_10
void FASTCALL call_9_12_0_10(SH2_struct* sh)
{
  SH2movwi(sh, 12, (0<<4)+10);
}
#define init_9_12_0_11
void FASTCALL call_9_12_0_11(SH2_struct* sh)
{
  SH2movwi(sh, 12, (0<<4)+11);
}
#define init_9_12_0_12
void FASTCALL call_9_12_0_12(SH2_struct* sh)
{
  SH2movwi(sh, 12, (0<<4)+12);
}
#define init_9_12_0_13
void FASTCALL call_9_12_0_13(SH2_struct* sh)
{
  SH2movwi(sh, 12, (0<<4)+13);
}
#define init_9_12_0_14
void FASTCALL call_9_12_0_14(SH2_struct* sh)
{
  SH2movwi(sh, 12, (0<<4)+14);
}
#define init_9_12_0_15
void FASTCALL call_9_12_0_15(SH2_struct* sh)
{
  SH2movwi(sh, 12, (0<<4)+15);
}
#define init_9_12_1_0
void FASTCALL call_9_12_1_0(SH2_struct* sh)
{
  SH2movwi(sh, 12, (1<<4)+0);
}
#define init_9_12_1_1
void FASTCALL call_9_12_1_1(SH2_struct* sh)
{
  SH2movwi(sh, 12, (1<<4)+1);
}
#define init_9_12_1_2
void FASTCALL call_9_12_1_2(SH2_struct* sh)
{
  SH2movwi(sh, 12, (1<<4)+2);
}
#define init_9_12_1_3
void FASTCALL call_9_12_1_3(SH2_struct* sh)
{
  SH2movwi(sh, 12, (1<<4)+3);
}
#define init_9_12_1_4
void FASTCALL call_9_12_1_4(SH2_struct* sh)
{
  SH2movwi(sh, 12, (1<<4)+4);
}
#define init_9_12_1_5
void FASTCALL call_9_12_1_5(SH2_struct* sh)
{
  SH2movwi(sh, 12, (1<<4)+5);
}
#define init_9_12_1_6
void FASTCALL call_9_12_1_6(SH2_struct* sh)
{
  SH2movwi(sh, 12, (1<<4)+6);
}
#define init_9_12_1_7
void FASTCALL call_9_12_1_7(SH2_struct* sh)
{
  SH2movwi(sh, 12, (1<<4)+7);
}
#define init_9_12_1_8
void FASTCALL call_9_12_1_8(SH2_struct* sh)
{
  SH2movwi(sh, 12, (1<<4)+8);
}
#define init_9_12_1_9
void FASTCALL call_9_12_1_9(SH2_struct* sh)
{
  SH2movwi(sh, 12, (1<<4)+9);
}
#define init_9_12_1_10
void FASTCALL call_9_12_1_10(SH2_struct* sh)
{
  SH2movwi(sh, 12, (1<<4)+10);
}
#define init_9_12_1_11
void FASTCALL call_9_12_1_11(SH2_struct* sh)
{
  SH2movwi(sh, 12, (1<<4)+11);
}
#define init_9_12_1_12
void FASTCALL call_9_12_1_12(SH2_struct* sh)
{
  SH2movwi(sh, 12, (1<<4)+12);
}
#define init_9_12_1_13
void FASTCALL call_9_12_1_13(SH2_struct* sh)
{
  SH2movwi(sh, 12, (1<<4)+13);
}
#define init_9_12_1_14
void FASTCALL call_9_12_1_14(SH2_struct* sh)
{
  SH2movwi(sh, 12, (1<<4)+14);
}
#define init_9_12_1_15
void FASTCALL call_9_12_1_15(SH2_struct* sh)
{
  SH2movwi(sh, 12, (1<<4)+15);
}
#define init_9_12_2_0
void FASTCALL call_9_12_2_0(SH2_struct* sh)
{
  SH2movwi(sh, 12, (2<<4)+0);
}
#define init_9_12_2_1
void FASTCALL call_9_12_2_1(SH2_struct* sh)
{
  SH2movwi(sh, 12, (2<<4)+1);
}
#define init_9_12_2_2
void FASTCALL call_9_12_2_2(SH2_struct* sh)
{
  SH2movwi(sh, 12, (2<<4)+2);
}
#define init_9_12_2_3
void FASTCALL call_9_12_2_3(SH2_struct* sh)
{
  SH2movwi(sh, 12, (2<<4)+3);
}
#define init_9_12_2_4
void FASTCALL call_9_12_2_4(SH2_struct* sh)
{
  SH2movwi(sh, 12, (2<<4)+4);
}
#define init_9_12_2_5
void FASTCALL call_9_12_2_5(SH2_struct* sh)
{
  SH2movwi(sh, 12, (2<<4)+5);
}
#define init_9_12_2_6
void FASTCALL call_9_12_2_6(SH2_struct* sh)
{
  SH2movwi(sh, 12, (2<<4)+6);
}
#define init_9_12_2_7
void FASTCALL call_9_12_2_7(SH2_struct* sh)
{
  SH2movwi(sh, 12, (2<<4)+7);
}
#define init_9_12_2_8
void FASTCALL call_9_12_2_8(SH2_struct* sh)
{
  SH2movwi(sh, 12, (2<<4)+8);
}
#define init_9_12_2_9
void FASTCALL call_9_12_2_9(SH2_struct* sh)
{
  SH2movwi(sh, 12, (2<<4)+9);
}
#define init_9_12_2_10
void FASTCALL call_9_12_2_10(SH2_struct* sh)
{
  SH2movwi(sh, 12, (2<<4)+10);
}
#define init_9_12_2_11
void FASTCALL call_9_12_2_11(SH2_struct* sh)
{
  SH2movwi(sh, 12, (2<<4)+11);
}
#define init_9_12_2_12
void FASTCALL call_9_12_2_12(SH2_struct* sh)
{
  SH2movwi(sh, 12, (2<<4)+12);
}
#define init_9_12_2_13
void FASTCALL call_9_12_2_13(SH2_struct* sh)
{
  SH2movwi(sh, 12, (2<<4)+13);
}
#define init_9_12_2_14
void FASTCALL call_9_12_2_14(SH2_struct* sh)
{
  SH2movwi(sh, 12, (2<<4)+14);
}
#define init_9_12_2_15
void FASTCALL call_9_12_2_15(SH2_struct* sh)
{
  SH2movwi(sh, 12, (2<<4)+15);
}
#define init_9_12_3_0
void FASTCALL call_9_12_3_0(SH2_struct* sh)
{
  SH2movwi(sh, 12, (3<<4)+0);
}
#define init_9_12_3_1
void FASTCALL call_9_12_3_1(SH2_struct* sh)
{
  SH2movwi(sh, 12, (3<<4)+1);
}
#define init_9_12_3_2
void FASTCALL call_9_12_3_2(SH2_struct* sh)
{
  SH2movwi(sh, 12, (3<<4)+2);
}
#define init_9_12_3_3
void FASTCALL call_9_12_3_3(SH2_struct* sh)
{
  SH2movwi(sh, 12, (3<<4)+3);
}
#define init_9_12_3_4
void FASTCALL call_9_12_3_4(SH2_struct* sh)
{
  SH2movwi(sh, 12, (3<<4)+4);
}
#define init_9_12_3_5
void FASTCALL call_9_12_3_5(SH2_struct* sh)
{
  SH2movwi(sh, 12, (3<<4)+5);
}
#define init_9_12_3_6
void FASTCALL call_9_12_3_6(SH2_struct* sh)
{
  SH2movwi(sh, 12, (3<<4)+6);
}
#define init_9_12_3_7
void FASTCALL call_9_12_3_7(SH2_struct* sh)
{
  SH2movwi(sh, 12, (3<<4)+7);
}
#define init_9_12_3_8
void FASTCALL call_9_12_3_8(SH2_struct* sh)
{
  SH2movwi(sh, 12, (3<<4)+8);
}
#define init_9_12_3_9
void FASTCALL call_9_12_3_9(SH2_struct* sh)
{
  SH2movwi(sh, 12, (3<<4)+9);
}
#define init_9_12_3_10
void FASTCALL call_9_12_3_10(SH2_struct* sh)
{
  SH2movwi(sh, 12, (3<<4)+10);
}
#define init_9_12_3_11
void FASTCALL call_9_12_3_11(SH2_struct* sh)
{
  SH2movwi(sh, 12, (3<<4)+11);
}
#define init_9_12_3_12
void FASTCALL call_9_12_3_12(SH2_struct* sh)
{
  SH2movwi(sh, 12, (3<<4)+12);
}
#define init_9_12_3_13
void FASTCALL call_9_12_3_13(SH2_struct* sh)
{
  SH2movwi(sh, 12, (3<<4)+13);
}
#define init_9_12_3_14
void FASTCALL call_9_12_3_14(SH2_struct* sh)
{
  SH2movwi(sh, 12, (3<<4)+14);
}
#define init_9_12_3_15
void FASTCALL call_9_12_3_15(SH2_struct* sh)
{
  SH2movwi(sh, 12, (3<<4)+15);
}
#define init_9_12_4_0
void FASTCALL call_9_12_4_0(SH2_struct* sh)
{
  SH2movwi(sh, 12, (4<<4)+0);
}
#define init_9_12_4_1
void FASTCALL call_9_12_4_1(SH2_struct* sh)
{
  SH2movwi(sh, 12, (4<<4)+1);
}
#define init_9_12_4_2
void FASTCALL call_9_12_4_2(SH2_struct* sh)
{
  SH2movwi(sh, 12, (4<<4)+2);
}
#define init_9_12_4_3
void FASTCALL call_9_12_4_3(SH2_struct* sh)
{
  SH2movwi(sh, 12, (4<<4)+3);
}
#define init_9_12_4_4
void FASTCALL call_9_12_4_4(SH2_struct* sh)
{
  SH2movwi(sh, 12, (4<<4)+4);
}
#define init_9_12_4_5
void FASTCALL call_9_12_4_5(SH2_struct* sh)
{
  SH2movwi(sh, 12, (4<<4)+5);
}
#define init_9_12_4_6
void FASTCALL call_9_12_4_6(SH2_struct* sh)
{
  SH2movwi(sh, 12, (4<<4)+6);
}
#define init_9_12_4_7
void FASTCALL call_9_12_4_7(SH2_struct* sh)
{
  SH2movwi(sh, 12, (4<<4)+7);
}
#define init_9_12_4_8
void FASTCALL call_9_12_4_8(SH2_struct* sh)
{
  SH2movwi(sh, 12, (4<<4)+8);
}
#define init_9_12_4_9
void FASTCALL call_9_12_4_9(SH2_struct* sh)
{
  SH2movwi(sh, 12, (4<<4)+9);
}
#define init_9_12_4_10
void FASTCALL call_9_12_4_10(SH2_struct* sh)
{
  SH2movwi(sh, 12, (4<<4)+10);
}
#define init_9_12_4_11
void FASTCALL call_9_12_4_11(SH2_struct* sh)
{
  SH2movwi(sh, 12, (4<<4)+11);
}
#define init_9_12_4_12
void FASTCALL call_9_12_4_12(SH2_struct* sh)
{
  SH2movwi(sh, 12, (4<<4)+12);
}
#define init_9_12_4_13
void FASTCALL call_9_12_4_13(SH2_struct* sh)
{
  SH2movwi(sh, 12, (4<<4)+13);
}
#define init_9_12_4_14
void FASTCALL call_9_12_4_14(SH2_struct* sh)
{
  SH2movwi(sh, 12, (4<<4)+14);
}
#define init_9_12_4_15
void FASTCALL call_9_12_4_15(SH2_struct* sh)
{
  SH2movwi(sh, 12, (4<<4)+15);
}
#define init_9_12_5_0
void FASTCALL call_9_12_5_0(SH2_struct* sh)
{
  SH2movwi(sh, 12, (5<<4)+0);
}
#define init_9_12_5_1
void FASTCALL call_9_12_5_1(SH2_struct* sh)
{
  SH2movwi(sh, 12, (5<<4)+1);
}
#define init_9_12_5_2
void FASTCALL call_9_12_5_2(SH2_struct* sh)
{
  SH2movwi(sh, 12, (5<<4)+2);
}
#define init_9_12_5_3
void FASTCALL call_9_12_5_3(SH2_struct* sh)
{
  SH2movwi(sh, 12, (5<<4)+3);
}
#define init_9_12_5_4
void FASTCALL call_9_12_5_4(SH2_struct* sh)
{
  SH2movwi(sh, 12, (5<<4)+4);
}
#define init_9_12_5_5
void FASTCALL call_9_12_5_5(SH2_struct* sh)
{
  SH2movwi(sh, 12, (5<<4)+5);
}
#define init_9_12_5_6
void FASTCALL call_9_12_5_6(SH2_struct* sh)
{
  SH2movwi(sh, 12, (5<<4)+6);
}
#define init_9_12_5_7
void FASTCALL call_9_12_5_7(SH2_struct* sh)
{
  SH2movwi(sh, 12, (5<<4)+7);
}
#define init_9_12_5_8
void FASTCALL call_9_12_5_8(SH2_struct* sh)
{
  SH2movwi(sh, 12, (5<<4)+8);
}
#define init_9_12_5_9
void FASTCALL call_9_12_5_9(SH2_struct* sh)
{
  SH2movwi(sh, 12, (5<<4)+9);
}
#define init_9_12_5_10
void FASTCALL call_9_12_5_10(SH2_struct* sh)
{
  SH2movwi(sh, 12, (5<<4)+10);
}
#define init_9_12_5_11
void FASTCALL call_9_12_5_11(SH2_struct* sh)
{
  SH2movwi(sh, 12, (5<<4)+11);
}
#define init_9_12_5_12
void FASTCALL call_9_12_5_12(SH2_struct* sh)
{
  SH2movwi(sh, 12, (5<<4)+12);
}
#define init_9_12_5_13
void FASTCALL call_9_12_5_13(SH2_struct* sh)
{
  SH2movwi(sh, 12, (5<<4)+13);
}
#define init_9_12_5_14
void FASTCALL call_9_12_5_14(SH2_struct* sh)
{
  SH2movwi(sh, 12, (5<<4)+14);
}
#define init_9_12_5_15
void FASTCALL call_9_12_5_15(SH2_struct* sh)
{
  SH2movwi(sh, 12, (5<<4)+15);
}
#define init_9_12_6_0
void FASTCALL call_9_12_6_0(SH2_struct* sh)
{
  SH2movwi(sh, 12, (6<<4)+0);
}
#define init_9_12_6_1
void FASTCALL call_9_12_6_1(SH2_struct* sh)
{
  SH2movwi(sh, 12, (6<<4)+1);
}
#define init_9_12_6_2
void FASTCALL call_9_12_6_2(SH2_struct* sh)
{
  SH2movwi(sh, 12, (6<<4)+2);
}
#define init_9_12_6_3
void FASTCALL call_9_12_6_3(SH2_struct* sh)
{
  SH2movwi(sh, 12, (6<<4)+3);
}
#define init_9_12_6_4
void FASTCALL call_9_12_6_4(SH2_struct* sh)
{
  SH2movwi(sh, 12, (6<<4)+4);
}
#define init_9_12_6_5
void FASTCALL call_9_12_6_5(SH2_struct* sh)
{
  SH2movwi(sh, 12, (6<<4)+5);
}
#define init_9_12_6_6
void FASTCALL call_9_12_6_6(SH2_struct* sh)
{
  SH2movwi(sh, 12, (6<<4)+6);
}
#define init_9_12_6_7
void FASTCALL call_9_12_6_7(SH2_struct* sh)
{
  SH2movwi(sh, 12, (6<<4)+7);
}
#define init_9_12_6_8
void FASTCALL call_9_12_6_8(SH2_struct* sh)
{
  SH2movwi(sh, 12, (6<<4)+8);
}
#define init_9_12_6_9
void FASTCALL call_9_12_6_9(SH2_struct* sh)
{
  SH2movwi(sh, 12, (6<<4)+9);
}
#define init_9_12_6_10
void FASTCALL call_9_12_6_10(SH2_struct* sh)
{
  SH2movwi(sh, 12, (6<<4)+10);
}
#define init_9_12_6_11
void FASTCALL call_9_12_6_11(SH2_struct* sh)
{
  SH2movwi(sh, 12, (6<<4)+11);
}
#define init_9_12_6_12
void FASTCALL call_9_12_6_12(SH2_struct* sh)
{
  SH2movwi(sh, 12, (6<<4)+12);
}
#define init_9_12_6_13
void FASTCALL call_9_12_6_13(SH2_struct* sh)
{
  SH2movwi(sh, 12, (6<<4)+13);
}
#define init_9_12_6_14
void FASTCALL call_9_12_6_14(SH2_struct* sh)
{
  SH2movwi(sh, 12, (6<<4)+14);
}
#define init_9_12_6_15
void FASTCALL call_9_12_6_15(SH2_struct* sh)
{
  SH2movwi(sh, 12, (6<<4)+15);
}
#define init_9_12_7_0
void FASTCALL call_9_12_7_0(SH2_struct* sh)
{
  SH2movwi(sh, 12, (7<<4)+0);
}
#define init_9_12_7_1
void FASTCALL call_9_12_7_1(SH2_struct* sh)
{
  SH2movwi(sh, 12, (7<<4)+1);
}
#define init_9_12_7_2
void FASTCALL call_9_12_7_2(SH2_struct* sh)
{
  SH2movwi(sh, 12, (7<<4)+2);
}
#define init_9_12_7_3
void FASTCALL call_9_12_7_3(SH2_struct* sh)
{
  SH2movwi(sh, 12, (7<<4)+3);
}
#define init_9_12_7_4
void FASTCALL call_9_12_7_4(SH2_struct* sh)
{
  SH2movwi(sh, 12, (7<<4)+4);
}
#define init_9_12_7_5
void FASTCALL call_9_12_7_5(SH2_struct* sh)
{
  SH2movwi(sh, 12, (7<<4)+5);
}
#define init_9_12_7_6
void FASTCALL call_9_12_7_6(SH2_struct* sh)
{
  SH2movwi(sh, 12, (7<<4)+6);
}
#define init_9_12_7_7
void FASTCALL call_9_12_7_7(SH2_struct* sh)
{
  SH2movwi(sh, 12, (7<<4)+7);
}
#define init_9_12_7_8
void FASTCALL call_9_12_7_8(SH2_struct* sh)
{
  SH2movwi(sh, 12, (7<<4)+8);
}
#define init_9_12_7_9
void FASTCALL call_9_12_7_9(SH2_struct* sh)
{
  SH2movwi(sh, 12, (7<<4)+9);
}
#define init_9_12_7_10
void FASTCALL call_9_12_7_10(SH2_struct* sh)
{
  SH2movwi(sh, 12, (7<<4)+10);
}
#define init_9_12_7_11
void FASTCALL call_9_12_7_11(SH2_struct* sh)
{
  SH2movwi(sh, 12, (7<<4)+11);
}
#define init_9_12_7_12
void FASTCALL call_9_12_7_12(SH2_struct* sh)
{
  SH2movwi(sh, 12, (7<<4)+12);
}
#define init_9_12_7_13
void FASTCALL call_9_12_7_13(SH2_struct* sh)
{
  SH2movwi(sh, 12, (7<<4)+13);
}
#define init_9_12_7_14
void FASTCALL call_9_12_7_14(SH2_struct* sh)
{
  SH2movwi(sh, 12, (7<<4)+14);
}
#define init_9_12_7_15
void FASTCALL call_9_12_7_15(SH2_struct* sh)
{
  SH2movwi(sh, 12, (7<<4)+15);
}
#define init_9_12_8_0
void FASTCALL call_9_12_8_0(SH2_struct* sh)
{
  SH2movwi(sh, 12, (8<<4)+0);
}
#define init_9_12_8_1
void FASTCALL call_9_12_8_1(SH2_struct* sh)
{
  SH2movwi(sh, 12, (8<<4)+1);
}
#define init_9_12_8_2
void FASTCALL call_9_12_8_2(SH2_struct* sh)
{
  SH2movwi(sh, 12, (8<<4)+2);
}
#define init_9_12_8_3
void FASTCALL call_9_12_8_3(SH2_struct* sh)
{
  SH2movwi(sh, 12, (8<<4)+3);
}
#define init_9_12_8_4
void FASTCALL call_9_12_8_4(SH2_struct* sh)
{
  SH2movwi(sh, 12, (8<<4)+4);
}
#define init_9_12_8_5
void FASTCALL call_9_12_8_5(SH2_struct* sh)
{
  SH2movwi(sh, 12, (8<<4)+5);
}
#define init_9_12_8_6
void FASTCALL call_9_12_8_6(SH2_struct* sh)
{
  SH2movwi(sh, 12, (8<<4)+6);
}
#define init_9_12_8_7
void FASTCALL call_9_12_8_7(SH2_struct* sh)
{
  SH2movwi(sh, 12, (8<<4)+7);
}
#define init_9_12_8_8
void FASTCALL call_9_12_8_8(SH2_struct* sh)
{
  SH2movwi(sh, 12, (8<<4)+8);
}
#define init_9_12_8_9
void FASTCALL call_9_12_8_9(SH2_struct* sh)
{
  SH2movwi(sh, 12, (8<<4)+9);
}
#define init_9_12_8_10
void FASTCALL call_9_12_8_10(SH2_struct* sh)
{
  SH2movwi(sh, 12, (8<<4)+10);
}
#define init_9_12_8_11
void FASTCALL call_9_12_8_11(SH2_struct* sh)
{
  SH2movwi(sh, 12, (8<<4)+11);
}
#define init_9_12_8_12
void FASTCALL call_9_12_8_12(SH2_struct* sh)
{
  SH2movwi(sh, 12, (8<<4)+12);
}
#define init_9_12_8_13
void FASTCALL call_9_12_8_13(SH2_struct* sh)
{
  SH2movwi(sh, 12, (8<<4)+13);
}
#define init_9_12_8_14
void FASTCALL call_9_12_8_14(SH2_struct* sh)
{
  SH2movwi(sh, 12, (8<<4)+14);
}
#define init_9_12_8_15
void FASTCALL call_9_12_8_15(SH2_struct* sh)
{
  SH2movwi(sh, 12, (8<<4)+15);
}
#define init_9_12_9_0
void FASTCALL call_9_12_9_0(SH2_struct* sh)
{
  SH2movwi(sh, 12, (9<<4)+0);
}
#define init_9_12_9_1
void FASTCALL call_9_12_9_1(SH2_struct* sh)
{
  SH2movwi(sh, 12, (9<<4)+1);
}
#define init_9_12_9_2
void FASTCALL call_9_12_9_2(SH2_struct* sh)
{
  SH2movwi(sh, 12, (9<<4)+2);
}
#define init_9_12_9_3
void FASTCALL call_9_12_9_3(SH2_struct* sh)
{
  SH2movwi(sh, 12, (9<<4)+3);
}
#define init_9_12_9_4
void FASTCALL call_9_12_9_4(SH2_struct* sh)
{
  SH2movwi(sh, 12, (9<<4)+4);
}
#define init_9_12_9_5
void FASTCALL call_9_12_9_5(SH2_struct* sh)
{
  SH2movwi(sh, 12, (9<<4)+5);
}
#define init_9_12_9_6
void FASTCALL call_9_12_9_6(SH2_struct* sh)
{
  SH2movwi(sh, 12, (9<<4)+6);
}
#define init_9_12_9_7
void FASTCALL call_9_12_9_7(SH2_struct* sh)
{
  SH2movwi(sh, 12, (9<<4)+7);
}
#define init_9_12_9_8
void FASTCALL call_9_12_9_8(SH2_struct* sh)
{
  SH2movwi(sh, 12, (9<<4)+8);
}
#define init_9_12_9_9
void FASTCALL call_9_12_9_9(SH2_struct* sh)
{
  SH2movwi(sh, 12, (9<<4)+9);
}
#define init_9_12_9_10
void FASTCALL call_9_12_9_10(SH2_struct* sh)
{
  SH2movwi(sh, 12, (9<<4)+10);
}
#define init_9_12_9_11
void FASTCALL call_9_12_9_11(SH2_struct* sh)
{
  SH2movwi(sh, 12, (9<<4)+11);
}
#define init_9_12_9_12
void FASTCALL call_9_12_9_12(SH2_struct* sh)
{
  SH2movwi(sh, 12, (9<<4)+12);
}
#define init_9_12_9_13
void FASTCALL call_9_12_9_13(SH2_struct* sh)
{
  SH2movwi(sh, 12, (9<<4)+13);
}
#define init_9_12_9_14
void FASTCALL call_9_12_9_14(SH2_struct* sh)
{
  SH2movwi(sh, 12, (9<<4)+14);
}
#define init_9_12_9_15
void FASTCALL call_9_12_9_15(SH2_struct* sh)
{
  SH2movwi(sh, 12, (9<<4)+15);
}
#define init_9_12_10_0
void FASTCALL call_9_12_10_0(SH2_struct* sh)
{
  SH2movwi(sh, 12, (10<<4)+0);
}
#define init_9_12_10_1
void FASTCALL call_9_12_10_1(SH2_struct* sh)
{
  SH2movwi(sh, 12, (10<<4)+1);
}
#define init_9_12_10_2
void FASTCALL call_9_12_10_2(SH2_struct* sh)
{
  SH2movwi(sh, 12, (10<<4)+2);
}
#define init_9_12_10_3
void FASTCALL call_9_12_10_3(SH2_struct* sh)
{
  SH2movwi(sh, 12, (10<<4)+3);
}
#define init_9_12_10_4
void FASTCALL call_9_12_10_4(SH2_struct* sh)
{
  SH2movwi(sh, 12, (10<<4)+4);
}
#define init_9_12_10_5
void FASTCALL call_9_12_10_5(SH2_struct* sh)
{
  SH2movwi(sh, 12, (10<<4)+5);
}
#define init_9_12_10_6
void FASTCALL call_9_12_10_6(SH2_struct* sh)
{
  SH2movwi(sh, 12, (10<<4)+6);
}
#define init_9_12_10_7
void FASTCALL call_9_12_10_7(SH2_struct* sh)
{
  SH2movwi(sh, 12, (10<<4)+7);
}
#define init_9_12_10_8
void FASTCALL call_9_12_10_8(SH2_struct* sh)
{
  SH2movwi(sh, 12, (10<<4)+8);
}
#define init_9_12_10_9
void FASTCALL call_9_12_10_9(SH2_struct* sh)
{
  SH2movwi(sh, 12, (10<<4)+9);
}
#define init_9_12_10_10
void FASTCALL call_9_12_10_10(SH2_struct* sh)
{
  SH2movwi(sh, 12, (10<<4)+10);
}
#define init_9_12_10_11
void FASTCALL call_9_12_10_11(SH2_struct* sh)
{
  SH2movwi(sh, 12, (10<<4)+11);
}
#define init_9_12_10_12
void FASTCALL call_9_12_10_12(SH2_struct* sh)
{
  SH2movwi(sh, 12, (10<<4)+12);
}
#define init_9_12_10_13
void FASTCALL call_9_12_10_13(SH2_struct* sh)
{
  SH2movwi(sh, 12, (10<<4)+13);
}
#define init_9_12_10_14
void FASTCALL call_9_12_10_14(SH2_struct* sh)
{
  SH2movwi(sh, 12, (10<<4)+14);
}
#define init_9_12_10_15
void FASTCALL call_9_12_10_15(SH2_struct* sh)
{
  SH2movwi(sh, 12, (10<<4)+15);
}
#define init_9_12_11_0
void FASTCALL call_9_12_11_0(SH2_struct* sh)
{
  SH2movwi(sh, 12, (11<<4)+0);
}
#define init_9_12_11_1
void FASTCALL call_9_12_11_1(SH2_struct* sh)
{
  SH2movwi(sh, 12, (11<<4)+1);
}
#define init_9_12_11_2
void FASTCALL call_9_12_11_2(SH2_struct* sh)
{
  SH2movwi(sh, 12, (11<<4)+2);
}
#define init_9_12_11_3
void FASTCALL call_9_12_11_3(SH2_struct* sh)
{
  SH2movwi(sh, 12, (11<<4)+3);
}
#define init_9_12_11_4
void FASTCALL call_9_12_11_4(SH2_struct* sh)
{
  SH2movwi(sh, 12, (11<<4)+4);
}
#define init_9_12_11_5
void FASTCALL call_9_12_11_5(SH2_struct* sh)
{
  SH2movwi(sh, 12, (11<<4)+5);
}
#define init_9_12_11_6
void FASTCALL call_9_12_11_6(SH2_struct* sh)
{
  SH2movwi(sh, 12, (11<<4)+6);
}
#define init_9_12_11_7
void FASTCALL call_9_12_11_7(SH2_struct* sh)
{
  SH2movwi(sh, 12, (11<<4)+7);
}
#define init_9_12_11_8
void FASTCALL call_9_12_11_8(SH2_struct* sh)
{
  SH2movwi(sh, 12, (11<<4)+8);
}
#define init_9_12_11_9
void FASTCALL call_9_12_11_9(SH2_struct* sh)
{
  SH2movwi(sh, 12, (11<<4)+9);
}
#define init_9_12_11_10
void FASTCALL call_9_12_11_10(SH2_struct* sh)
{
  SH2movwi(sh, 12, (11<<4)+10);
}
#define init_9_12_11_11
void FASTCALL call_9_12_11_11(SH2_struct* sh)
{
  SH2movwi(sh, 12, (11<<4)+11);
}
#define init_9_12_11_12
void FASTCALL call_9_12_11_12(SH2_struct* sh)
{
  SH2movwi(sh, 12, (11<<4)+12);
}
#define init_9_12_11_13
void FASTCALL call_9_12_11_13(SH2_struct* sh)
{
  SH2movwi(sh, 12, (11<<4)+13);
}
#define init_9_12_11_14
void FASTCALL call_9_12_11_14(SH2_struct* sh)
{
  SH2movwi(sh, 12, (11<<4)+14);
}
#define init_9_12_11_15
void FASTCALL call_9_12_11_15(SH2_struct* sh)
{
  SH2movwi(sh, 12, (11<<4)+15);
}
#define init_9_12_12_0
void FASTCALL call_9_12_12_0(SH2_struct* sh)
{
  SH2movwi(sh, 12, (12<<4)+0);
}
#define init_9_12_12_1
void FASTCALL call_9_12_12_1(SH2_struct* sh)
{
  SH2movwi(sh, 12, (12<<4)+1);
}
#define init_9_12_12_2
void FASTCALL call_9_12_12_2(SH2_struct* sh)
{
  SH2movwi(sh, 12, (12<<4)+2);
}
#define init_9_12_12_3
void FASTCALL call_9_12_12_3(SH2_struct* sh)
{
  SH2movwi(sh, 12, (12<<4)+3);
}
#define init_9_12_12_4
void FASTCALL call_9_12_12_4(SH2_struct* sh)
{
  SH2movwi(sh, 12, (12<<4)+4);
}
#define init_9_12_12_5
void FASTCALL call_9_12_12_5(SH2_struct* sh)
{
  SH2movwi(sh, 12, (12<<4)+5);
}
#define init_9_12_12_6
void FASTCALL call_9_12_12_6(SH2_struct* sh)
{
  SH2movwi(sh, 12, (12<<4)+6);
}
#define init_9_12_12_7
void FASTCALL call_9_12_12_7(SH2_struct* sh)
{
  SH2movwi(sh, 12, (12<<4)+7);
}
#define init_9_12_12_8
void FASTCALL call_9_12_12_8(SH2_struct* sh)
{
  SH2movwi(sh, 12, (12<<4)+8);
}
#define init_9_12_12_9
void FASTCALL call_9_12_12_9(SH2_struct* sh)
{
  SH2movwi(sh, 12, (12<<4)+9);
}
#define init_9_12_12_10
void FASTCALL call_9_12_12_10(SH2_struct* sh)
{
  SH2movwi(sh, 12, (12<<4)+10);
}
#define init_9_12_12_11
void FASTCALL call_9_12_12_11(SH2_struct* sh)
{
  SH2movwi(sh, 12, (12<<4)+11);
}
#define init_9_12_12_12
void FASTCALL call_9_12_12_12(SH2_struct* sh)
{
  SH2movwi(sh, 12, (12<<4)+12);
}
#define init_9_12_12_13
void FASTCALL call_9_12_12_13(SH2_struct* sh)
{
  SH2movwi(sh, 12, (12<<4)+13);
}
#define init_9_12_12_14
void FASTCALL call_9_12_12_14(SH2_struct* sh)
{
  SH2movwi(sh, 12, (12<<4)+14);
}
#define init_9_12_12_15
void FASTCALL call_9_12_12_15(SH2_struct* sh)
{
  SH2movwi(sh, 12, (12<<4)+15);
}
#define init_9_12_13_0
void FASTCALL call_9_12_13_0(SH2_struct* sh)
{
  SH2movwi(sh, 12, (13<<4)+0);
}
#define init_9_12_13_1
void FASTCALL call_9_12_13_1(SH2_struct* sh)
{
  SH2movwi(sh, 12, (13<<4)+1);
}
#define init_9_12_13_2
void FASTCALL call_9_12_13_2(SH2_struct* sh)
{
  SH2movwi(sh, 12, (13<<4)+2);
}
#define init_9_12_13_3
void FASTCALL call_9_12_13_3(SH2_struct* sh)
{
  SH2movwi(sh, 12, (13<<4)+3);
}
#define init_9_12_13_4
void FASTCALL call_9_12_13_4(SH2_struct* sh)
{
  SH2movwi(sh, 12, (13<<4)+4);
}
#define init_9_12_13_5
void FASTCALL call_9_12_13_5(SH2_struct* sh)
{
  SH2movwi(sh, 12, (13<<4)+5);
}
#define init_9_12_13_6
void FASTCALL call_9_12_13_6(SH2_struct* sh)
{
  SH2movwi(sh, 12, (13<<4)+6);
}
#define init_9_12_13_7
void FASTCALL call_9_12_13_7(SH2_struct* sh)
{
  SH2movwi(sh, 12, (13<<4)+7);
}
#define init_9_12_13_8
void FASTCALL call_9_12_13_8(SH2_struct* sh)
{
  SH2movwi(sh, 12, (13<<4)+8);
}
#define init_9_12_13_9
void FASTCALL call_9_12_13_9(SH2_struct* sh)
{
  SH2movwi(sh, 12, (13<<4)+9);
}
#define init_9_12_13_10
void FASTCALL call_9_12_13_10(SH2_struct* sh)
{
  SH2movwi(sh, 12, (13<<4)+10);
}
#define init_9_12_13_11
void FASTCALL call_9_12_13_11(SH2_struct* sh)
{
  SH2movwi(sh, 12, (13<<4)+11);
}
#define init_9_12_13_12
void FASTCALL call_9_12_13_12(SH2_struct* sh)
{
  SH2movwi(sh, 12, (13<<4)+12);
}
#define init_9_12_13_13
void FASTCALL call_9_12_13_13(SH2_struct* sh)
{
  SH2movwi(sh, 12, (13<<4)+13);
}
#define init_9_12_13_14
void FASTCALL call_9_12_13_14(SH2_struct* sh)
{
  SH2movwi(sh, 12, (13<<4)+14);
}
#define init_9_12_13_15
void FASTCALL call_9_12_13_15(SH2_struct* sh)
{
  SH2movwi(sh, 12, (13<<4)+15);
}
#define init_9_12_14_0
void FASTCALL call_9_12_14_0(SH2_struct* sh)
{
  SH2movwi(sh, 12, (14<<4)+0);
}
#define init_9_12_14_1
void FASTCALL call_9_12_14_1(SH2_struct* sh)
{
  SH2movwi(sh, 12, (14<<4)+1);
}
#define init_9_12_14_2
void FASTCALL call_9_12_14_2(SH2_struct* sh)
{
  SH2movwi(sh, 12, (14<<4)+2);
}
#define init_9_12_14_3
void FASTCALL call_9_12_14_3(SH2_struct* sh)
{
  SH2movwi(sh, 12, (14<<4)+3);
}
#define init_9_12_14_4
void FASTCALL call_9_12_14_4(SH2_struct* sh)
{
  SH2movwi(sh, 12, (14<<4)+4);
}
#define init_9_12_14_5
void FASTCALL call_9_12_14_5(SH2_struct* sh)
{
  SH2movwi(sh, 12, (14<<4)+5);
}
#define init_9_12_14_6
void FASTCALL call_9_12_14_6(SH2_struct* sh)
{
  SH2movwi(sh, 12, (14<<4)+6);
}
#define init_9_12_14_7
void FASTCALL call_9_12_14_7(SH2_struct* sh)
{
  SH2movwi(sh, 12, (14<<4)+7);
}
#define init_9_12_14_8
void FASTCALL call_9_12_14_8(SH2_struct* sh)
{
  SH2movwi(sh, 12, (14<<4)+8);
}
#define init_9_12_14_9
void FASTCALL call_9_12_14_9(SH2_struct* sh)
{
  SH2movwi(sh, 12, (14<<4)+9);
}
#define init_9_12_14_10
void FASTCALL call_9_12_14_10(SH2_struct* sh)
{
  SH2movwi(sh, 12, (14<<4)+10);
}
#define init_9_12_14_11
void FASTCALL call_9_12_14_11(SH2_struct* sh)
{
  SH2movwi(sh, 12, (14<<4)+11);
}
#define init_9_12_14_12
void FASTCALL call_9_12_14_12(SH2_struct* sh)
{
  SH2movwi(sh, 12, (14<<4)+12);
}
#define init_9_12_14_13
void FASTCALL call_9_12_14_13(SH2_struct* sh)
{
  SH2movwi(sh, 12, (14<<4)+13);
}
#define init_9_12_14_14
void FASTCALL call_9_12_14_14(SH2_struct* sh)
{
  SH2movwi(sh, 12, (14<<4)+14);
}
#define init_9_12_14_15
void FASTCALL call_9_12_14_15(SH2_struct* sh)
{
  SH2movwi(sh, 12, (14<<4)+15);
}
#define init_9_12_15_0
void FASTCALL call_9_12_15_0(SH2_struct* sh)
{
  SH2movwi(sh, 12, (15<<4)+0);
}
#define init_9_12_15_1
void FASTCALL call_9_12_15_1(SH2_struct* sh)
{
  SH2movwi(sh, 12, (15<<4)+1);
}
#define init_9_12_15_2
void FASTCALL call_9_12_15_2(SH2_struct* sh)
{
  SH2movwi(sh, 12, (15<<4)+2);
}
#define init_9_12_15_3
void FASTCALL call_9_12_15_3(SH2_struct* sh)
{
  SH2movwi(sh, 12, (15<<4)+3);
}
#define init_9_12_15_4
void FASTCALL call_9_12_15_4(SH2_struct* sh)
{
  SH2movwi(sh, 12, (15<<4)+4);
}
#define init_9_12_15_5
void FASTCALL call_9_12_15_5(SH2_struct* sh)
{
  SH2movwi(sh, 12, (15<<4)+5);
}
#define init_9_12_15_6
void FASTCALL call_9_12_15_6(SH2_struct* sh)
{
  SH2movwi(sh, 12, (15<<4)+6);
}
#define init_9_12_15_7
void FASTCALL call_9_12_15_7(SH2_struct* sh)
{
  SH2movwi(sh, 12, (15<<4)+7);
}
#define init_9_12_15_8
void FASTCALL call_9_12_15_8(SH2_struct* sh)
{
  SH2movwi(sh, 12, (15<<4)+8);
}
#define init_9_12_15_9
void FASTCALL call_9_12_15_9(SH2_struct* sh)
{
  SH2movwi(sh, 12, (15<<4)+9);
}
#define init_9_12_15_10
void FASTCALL call_9_12_15_10(SH2_struct* sh)
{
  SH2movwi(sh, 12, (15<<4)+10);
}
#define init_9_12_15_11
void FASTCALL call_9_12_15_11(SH2_struct* sh)
{
  SH2movwi(sh, 12, (15<<4)+11);
}
#define init_9_12_15_12
void FASTCALL call_9_12_15_12(SH2_struct* sh)
{
  SH2movwi(sh, 12, (15<<4)+12);
}
#define init_9_12_15_13
void FASTCALL call_9_12_15_13(SH2_struct* sh)
{
  SH2movwi(sh, 12, (15<<4)+13);
}
#define init_9_12_15_14
void FASTCALL call_9_12_15_14(SH2_struct* sh)
{
  SH2movwi(sh, 12, (15<<4)+14);
}
#define init_9_12_15_15
void FASTCALL call_9_12_15_15(SH2_struct* sh)
{
  SH2movwi(sh, 12, (15<<4)+15);
}
#define init_9_13_0_0
void FASTCALL call_9_13_0_0(SH2_struct* sh)
{
  SH2movwi(sh, 13, (0<<4)+0);
}
#define init_9_13_0_1
void FASTCALL call_9_13_0_1(SH2_struct* sh)
{
  SH2movwi(sh, 13, (0<<4)+1);
}
#define init_9_13_0_2
void FASTCALL call_9_13_0_2(SH2_struct* sh)
{
  SH2movwi(sh, 13, (0<<4)+2);
}
#define init_9_13_0_3
void FASTCALL call_9_13_0_3(SH2_struct* sh)
{
  SH2movwi(sh, 13, (0<<4)+3);
}
#define init_9_13_0_4
void FASTCALL call_9_13_0_4(SH2_struct* sh)
{
  SH2movwi(sh, 13, (0<<4)+4);
}
#define init_9_13_0_5
void FASTCALL call_9_13_0_5(SH2_struct* sh)
{
  SH2movwi(sh, 13, (0<<4)+5);
}
#define init_9_13_0_6
void FASTCALL call_9_13_0_6(SH2_struct* sh)
{
  SH2movwi(sh, 13, (0<<4)+6);
}
#define init_9_13_0_7
void FASTCALL call_9_13_0_7(SH2_struct* sh)
{
  SH2movwi(sh, 13, (0<<4)+7);
}
#define init_9_13_0_8
void FASTCALL call_9_13_0_8(SH2_struct* sh)
{
  SH2movwi(sh, 13, (0<<4)+8);
}
#define init_9_13_0_9
void FASTCALL call_9_13_0_9(SH2_struct* sh)
{
  SH2movwi(sh, 13, (0<<4)+9);
}
#define init_9_13_0_10
void FASTCALL call_9_13_0_10(SH2_struct* sh)
{
  SH2movwi(sh, 13, (0<<4)+10);
}
#define init_9_13_0_11
void FASTCALL call_9_13_0_11(SH2_struct* sh)
{
  SH2movwi(sh, 13, (0<<4)+11);
}
#define init_9_13_0_12
void FASTCALL call_9_13_0_12(SH2_struct* sh)
{
  SH2movwi(sh, 13, (0<<4)+12);
}
#define init_9_13_0_13
void FASTCALL call_9_13_0_13(SH2_struct* sh)
{
  SH2movwi(sh, 13, (0<<4)+13);
}
#define init_9_13_0_14
void FASTCALL call_9_13_0_14(SH2_struct* sh)
{
  SH2movwi(sh, 13, (0<<4)+14);
}
#define init_9_13_0_15
void FASTCALL call_9_13_0_15(SH2_struct* sh)
{
  SH2movwi(sh, 13, (0<<4)+15);
}
#define init_9_13_1_0
void FASTCALL call_9_13_1_0(SH2_struct* sh)
{
  SH2movwi(sh, 13, (1<<4)+0);
}
#define init_9_13_1_1
void FASTCALL call_9_13_1_1(SH2_struct* sh)
{
  SH2movwi(sh, 13, (1<<4)+1);
}
#define init_9_13_1_2
void FASTCALL call_9_13_1_2(SH2_struct* sh)
{
  SH2movwi(sh, 13, (1<<4)+2);
}
#define init_9_13_1_3
void FASTCALL call_9_13_1_3(SH2_struct* sh)
{
  SH2movwi(sh, 13, (1<<4)+3);
}
#define init_9_13_1_4
void FASTCALL call_9_13_1_4(SH2_struct* sh)
{
  SH2movwi(sh, 13, (1<<4)+4);
}
#define init_9_13_1_5
void FASTCALL call_9_13_1_5(SH2_struct* sh)
{
  SH2movwi(sh, 13, (1<<4)+5);
}
#define init_9_13_1_6
void FASTCALL call_9_13_1_6(SH2_struct* sh)
{
  SH2movwi(sh, 13, (1<<4)+6);
}
#define init_9_13_1_7
void FASTCALL call_9_13_1_7(SH2_struct* sh)
{
  SH2movwi(sh, 13, (1<<4)+7);
}
#define init_9_13_1_8
void FASTCALL call_9_13_1_8(SH2_struct* sh)
{
  SH2movwi(sh, 13, (1<<4)+8);
}
#define init_9_13_1_9
void FASTCALL call_9_13_1_9(SH2_struct* sh)
{
  SH2movwi(sh, 13, (1<<4)+9);
}
#define init_9_13_1_10
void FASTCALL call_9_13_1_10(SH2_struct* sh)
{
  SH2movwi(sh, 13, (1<<4)+10);
}
#define init_9_13_1_11
void FASTCALL call_9_13_1_11(SH2_struct* sh)
{
  SH2movwi(sh, 13, (1<<4)+11);
}
#define init_9_13_1_12
void FASTCALL call_9_13_1_12(SH2_struct* sh)
{
  SH2movwi(sh, 13, (1<<4)+12);
}
#define init_9_13_1_13
void FASTCALL call_9_13_1_13(SH2_struct* sh)
{
  SH2movwi(sh, 13, (1<<4)+13);
}
#define init_9_13_1_14
void FASTCALL call_9_13_1_14(SH2_struct* sh)
{
  SH2movwi(sh, 13, (1<<4)+14);
}
#define init_9_13_1_15
void FASTCALL call_9_13_1_15(SH2_struct* sh)
{
  SH2movwi(sh, 13, (1<<4)+15);
}
#define init_9_13_2_0
void FASTCALL call_9_13_2_0(SH2_struct* sh)
{
  SH2movwi(sh, 13, (2<<4)+0);
}
#define init_9_13_2_1
void FASTCALL call_9_13_2_1(SH2_struct* sh)
{
  SH2movwi(sh, 13, (2<<4)+1);
}
#define init_9_13_2_2
void FASTCALL call_9_13_2_2(SH2_struct* sh)
{
  SH2movwi(sh, 13, (2<<4)+2);
}
#define init_9_13_2_3
void FASTCALL call_9_13_2_3(SH2_struct* sh)
{
  SH2movwi(sh, 13, (2<<4)+3);
}
#define init_9_13_2_4
void FASTCALL call_9_13_2_4(SH2_struct* sh)
{
  SH2movwi(sh, 13, (2<<4)+4);
}
#define init_9_13_2_5
void FASTCALL call_9_13_2_5(SH2_struct* sh)
{
  SH2movwi(sh, 13, (2<<4)+5);
}
#define init_9_13_2_6
void FASTCALL call_9_13_2_6(SH2_struct* sh)
{
  SH2movwi(sh, 13, (2<<4)+6);
}
#define init_9_13_2_7
void FASTCALL call_9_13_2_7(SH2_struct* sh)
{
  SH2movwi(sh, 13, (2<<4)+7);
}
#define init_9_13_2_8
void FASTCALL call_9_13_2_8(SH2_struct* sh)
{
  SH2movwi(sh, 13, (2<<4)+8);
}
#define init_9_13_2_9
void FASTCALL call_9_13_2_9(SH2_struct* sh)
{
  SH2movwi(sh, 13, (2<<4)+9);
}
#define init_9_13_2_10
void FASTCALL call_9_13_2_10(SH2_struct* sh)
{
  SH2movwi(sh, 13, (2<<4)+10);
}
#define init_9_13_2_11
void FASTCALL call_9_13_2_11(SH2_struct* sh)
{
  SH2movwi(sh, 13, (2<<4)+11);
}
#define init_9_13_2_12
void FASTCALL call_9_13_2_12(SH2_struct* sh)
{
  SH2movwi(sh, 13, (2<<4)+12);
}
#define init_9_13_2_13
void FASTCALL call_9_13_2_13(SH2_struct* sh)
{
  SH2movwi(sh, 13, (2<<4)+13);
}
#define init_9_13_2_14
void FASTCALL call_9_13_2_14(SH2_struct* sh)
{
  SH2movwi(sh, 13, (2<<4)+14);
}
#define init_9_13_2_15
void FASTCALL call_9_13_2_15(SH2_struct* sh)
{
  SH2movwi(sh, 13, (2<<4)+15);
}
#define init_9_13_3_0
void FASTCALL call_9_13_3_0(SH2_struct* sh)
{
  SH2movwi(sh, 13, (3<<4)+0);
}
#define init_9_13_3_1
void FASTCALL call_9_13_3_1(SH2_struct* sh)
{
  SH2movwi(sh, 13, (3<<4)+1);
}
#define init_9_13_3_2
void FASTCALL call_9_13_3_2(SH2_struct* sh)
{
  SH2movwi(sh, 13, (3<<4)+2);
}
#define init_9_13_3_3
void FASTCALL call_9_13_3_3(SH2_struct* sh)
{
  SH2movwi(sh, 13, (3<<4)+3);
}
#define init_9_13_3_4
void FASTCALL call_9_13_3_4(SH2_struct* sh)
{
  SH2movwi(sh, 13, (3<<4)+4);
}
#define init_9_13_3_5
void FASTCALL call_9_13_3_5(SH2_struct* sh)
{
  SH2movwi(sh, 13, (3<<4)+5);
}
#define init_9_13_3_6
void FASTCALL call_9_13_3_6(SH2_struct* sh)
{
  SH2movwi(sh, 13, (3<<4)+6);
}
#define init_9_13_3_7
void FASTCALL call_9_13_3_7(SH2_struct* sh)
{
  SH2movwi(sh, 13, (3<<4)+7);
}
#define init_9_13_3_8
void FASTCALL call_9_13_3_8(SH2_struct* sh)
{
  SH2movwi(sh, 13, (3<<4)+8);
}
#define init_9_13_3_9
void FASTCALL call_9_13_3_9(SH2_struct* sh)
{
  SH2movwi(sh, 13, (3<<4)+9);
}
#define init_9_13_3_10
void FASTCALL call_9_13_3_10(SH2_struct* sh)
{
  SH2movwi(sh, 13, (3<<4)+10);
}
#define init_9_13_3_11
void FASTCALL call_9_13_3_11(SH2_struct* sh)
{
  SH2movwi(sh, 13, (3<<4)+11);
}
#define init_9_13_3_12
void FASTCALL call_9_13_3_12(SH2_struct* sh)
{
  SH2movwi(sh, 13, (3<<4)+12);
}
#define init_9_13_3_13
void FASTCALL call_9_13_3_13(SH2_struct* sh)
{
  SH2movwi(sh, 13, (3<<4)+13);
}
#define init_9_13_3_14
void FASTCALL call_9_13_3_14(SH2_struct* sh)
{
  SH2movwi(sh, 13, (3<<4)+14);
}
#define init_9_13_3_15
void FASTCALL call_9_13_3_15(SH2_struct* sh)
{
  SH2movwi(sh, 13, (3<<4)+15);
}
#define init_9_13_4_0
void FASTCALL call_9_13_4_0(SH2_struct* sh)
{
  SH2movwi(sh, 13, (4<<4)+0);
}
#define init_9_13_4_1
void FASTCALL call_9_13_4_1(SH2_struct* sh)
{
  SH2movwi(sh, 13, (4<<4)+1);
}
#define init_9_13_4_2
void FASTCALL call_9_13_4_2(SH2_struct* sh)
{
  SH2movwi(sh, 13, (4<<4)+2);
}
#define init_9_13_4_3
void FASTCALL call_9_13_4_3(SH2_struct* sh)
{
  SH2movwi(sh, 13, (4<<4)+3);
}
#define init_9_13_4_4
void FASTCALL call_9_13_4_4(SH2_struct* sh)
{
  SH2movwi(sh, 13, (4<<4)+4);
}
#define init_9_13_4_5
void FASTCALL call_9_13_4_5(SH2_struct* sh)
{
  SH2movwi(sh, 13, (4<<4)+5);
}
#define init_9_13_4_6
void FASTCALL call_9_13_4_6(SH2_struct* sh)
{
  SH2movwi(sh, 13, (4<<4)+6);
}
#define init_9_13_4_7
void FASTCALL call_9_13_4_7(SH2_struct* sh)
{
  SH2movwi(sh, 13, (4<<4)+7);
}
#define init_9_13_4_8
void FASTCALL call_9_13_4_8(SH2_struct* sh)
{
  SH2movwi(sh, 13, (4<<4)+8);
}
#define init_9_13_4_9
void FASTCALL call_9_13_4_9(SH2_struct* sh)
{
  SH2movwi(sh, 13, (4<<4)+9);
}
#define init_9_13_4_10
void FASTCALL call_9_13_4_10(SH2_struct* sh)
{
  SH2movwi(sh, 13, (4<<4)+10);
}
#define init_9_13_4_11
void FASTCALL call_9_13_4_11(SH2_struct* sh)
{
  SH2movwi(sh, 13, (4<<4)+11);
}
#define init_9_13_4_12
void FASTCALL call_9_13_4_12(SH2_struct* sh)
{
  SH2movwi(sh, 13, (4<<4)+12);
}
#define init_9_13_4_13
void FASTCALL call_9_13_4_13(SH2_struct* sh)
{
  SH2movwi(sh, 13, (4<<4)+13);
}
#define init_9_13_4_14
void FASTCALL call_9_13_4_14(SH2_struct* sh)
{
  SH2movwi(sh, 13, (4<<4)+14);
}
#define init_9_13_4_15
void FASTCALL call_9_13_4_15(SH2_struct* sh)
{
  SH2movwi(sh, 13, (4<<4)+15);
}
#define init_9_13_5_0
void FASTCALL call_9_13_5_0(SH2_struct* sh)
{
  SH2movwi(sh, 13, (5<<4)+0);
}
#define init_9_13_5_1
void FASTCALL call_9_13_5_1(SH2_struct* sh)
{
  SH2movwi(sh, 13, (5<<4)+1);
}
#define init_9_13_5_2
void FASTCALL call_9_13_5_2(SH2_struct* sh)
{
  SH2movwi(sh, 13, (5<<4)+2);
}
#define init_9_13_5_3
void FASTCALL call_9_13_5_3(SH2_struct* sh)
{
  SH2movwi(sh, 13, (5<<4)+3);
}
#define init_9_13_5_4
void FASTCALL call_9_13_5_4(SH2_struct* sh)
{
  SH2movwi(sh, 13, (5<<4)+4);
}
#define init_9_13_5_5
void FASTCALL call_9_13_5_5(SH2_struct* sh)
{
  SH2movwi(sh, 13, (5<<4)+5);
}
#define init_9_13_5_6
void FASTCALL call_9_13_5_6(SH2_struct* sh)
{
  SH2movwi(sh, 13, (5<<4)+6);
}
#define init_9_13_5_7
void FASTCALL call_9_13_5_7(SH2_struct* sh)
{
  SH2movwi(sh, 13, (5<<4)+7);
}
#define init_9_13_5_8
void FASTCALL call_9_13_5_8(SH2_struct* sh)
{
  SH2movwi(sh, 13, (5<<4)+8);
}
#define init_9_13_5_9
void FASTCALL call_9_13_5_9(SH2_struct* sh)
{
  SH2movwi(sh, 13, (5<<4)+9);
}
#define init_9_13_5_10
void FASTCALL call_9_13_5_10(SH2_struct* sh)
{
  SH2movwi(sh, 13, (5<<4)+10);
}
#define init_9_13_5_11
void FASTCALL call_9_13_5_11(SH2_struct* sh)
{
  SH2movwi(sh, 13, (5<<4)+11);
}
#define init_9_13_5_12
void FASTCALL call_9_13_5_12(SH2_struct* sh)
{
  SH2movwi(sh, 13, (5<<4)+12);
}
#define init_9_13_5_13
void FASTCALL call_9_13_5_13(SH2_struct* sh)
{
  SH2movwi(sh, 13, (5<<4)+13);
}
#define init_9_13_5_14
void FASTCALL call_9_13_5_14(SH2_struct* sh)
{
  SH2movwi(sh, 13, (5<<4)+14);
}
#define init_9_13_5_15
void FASTCALL call_9_13_5_15(SH2_struct* sh)
{
  SH2movwi(sh, 13, (5<<4)+15);
}
#define init_9_13_6_0
void FASTCALL call_9_13_6_0(SH2_struct* sh)
{
  SH2movwi(sh, 13, (6<<4)+0);
}
#define init_9_13_6_1
void FASTCALL call_9_13_6_1(SH2_struct* sh)
{
  SH2movwi(sh, 13, (6<<4)+1);
}
#define init_9_13_6_2
void FASTCALL call_9_13_6_2(SH2_struct* sh)
{
  SH2movwi(sh, 13, (6<<4)+2);
}
#define init_9_13_6_3
void FASTCALL call_9_13_6_3(SH2_struct* sh)
{
  SH2movwi(sh, 13, (6<<4)+3);
}
#define init_9_13_6_4
void FASTCALL call_9_13_6_4(SH2_struct* sh)
{
  SH2movwi(sh, 13, (6<<4)+4);
}
#define init_9_13_6_5
void FASTCALL call_9_13_6_5(SH2_struct* sh)
{
  SH2movwi(sh, 13, (6<<4)+5);
}
#define init_9_13_6_6
void FASTCALL call_9_13_6_6(SH2_struct* sh)
{
  SH2movwi(sh, 13, (6<<4)+6);
}
#define init_9_13_6_7
void FASTCALL call_9_13_6_7(SH2_struct* sh)
{
  SH2movwi(sh, 13, (6<<4)+7);
}
#define init_9_13_6_8
void FASTCALL call_9_13_6_8(SH2_struct* sh)
{
  SH2movwi(sh, 13, (6<<4)+8);
}
#define init_9_13_6_9
void FASTCALL call_9_13_6_9(SH2_struct* sh)
{
  SH2movwi(sh, 13, (6<<4)+9);
}
#define init_9_13_6_10
void FASTCALL call_9_13_6_10(SH2_struct* sh)
{
  SH2movwi(sh, 13, (6<<4)+10);
}
#define init_9_13_6_11
void FASTCALL call_9_13_6_11(SH2_struct* sh)
{
  SH2movwi(sh, 13, (6<<4)+11);
}
#define init_9_13_6_12
void FASTCALL call_9_13_6_12(SH2_struct* sh)
{
  SH2movwi(sh, 13, (6<<4)+12);
}
#define init_9_13_6_13
void FASTCALL call_9_13_6_13(SH2_struct* sh)
{
  SH2movwi(sh, 13, (6<<4)+13);
}
#define init_9_13_6_14
void FASTCALL call_9_13_6_14(SH2_struct* sh)
{
  SH2movwi(sh, 13, (6<<4)+14);
}
#define init_9_13_6_15
void FASTCALL call_9_13_6_15(SH2_struct* sh)
{
  SH2movwi(sh, 13, (6<<4)+15);
}
#define init_9_13_7_0
void FASTCALL call_9_13_7_0(SH2_struct* sh)
{
  SH2movwi(sh, 13, (7<<4)+0);
}
#define init_9_13_7_1
void FASTCALL call_9_13_7_1(SH2_struct* sh)
{
  SH2movwi(sh, 13, (7<<4)+1);
}
#define init_9_13_7_2
void FASTCALL call_9_13_7_2(SH2_struct* sh)
{
  SH2movwi(sh, 13, (7<<4)+2);
}
#define init_9_13_7_3
void FASTCALL call_9_13_7_3(SH2_struct* sh)
{
  SH2movwi(sh, 13, (7<<4)+3);
}
#define init_9_13_7_4
void FASTCALL call_9_13_7_4(SH2_struct* sh)
{
  SH2movwi(sh, 13, (7<<4)+4);
}
#define init_9_13_7_5
void FASTCALL call_9_13_7_5(SH2_struct* sh)
{
  SH2movwi(sh, 13, (7<<4)+5);
}
#define init_9_13_7_6
void FASTCALL call_9_13_7_6(SH2_struct* sh)
{
  SH2movwi(sh, 13, (7<<4)+6);
}
#define init_9_13_7_7
void FASTCALL call_9_13_7_7(SH2_struct* sh)
{
  SH2movwi(sh, 13, (7<<4)+7);
}
#define init_9_13_7_8
void FASTCALL call_9_13_7_8(SH2_struct* sh)
{
  SH2movwi(sh, 13, (7<<4)+8);
}
#define init_9_13_7_9
void FASTCALL call_9_13_7_9(SH2_struct* sh)
{
  SH2movwi(sh, 13, (7<<4)+9);
}
#define init_9_13_7_10
void FASTCALL call_9_13_7_10(SH2_struct* sh)
{
  SH2movwi(sh, 13, (7<<4)+10);
}
#define init_9_13_7_11
void FASTCALL call_9_13_7_11(SH2_struct* sh)
{
  SH2movwi(sh, 13, (7<<4)+11);
}
#define init_9_13_7_12
void FASTCALL call_9_13_7_12(SH2_struct* sh)
{
  SH2movwi(sh, 13, (7<<4)+12);
}
#define init_9_13_7_13
void FASTCALL call_9_13_7_13(SH2_struct* sh)
{
  SH2movwi(sh, 13, (7<<4)+13);
}
#define init_9_13_7_14
void FASTCALL call_9_13_7_14(SH2_struct* sh)
{
  SH2movwi(sh, 13, (7<<4)+14);
}
#define init_9_13_7_15
void FASTCALL call_9_13_7_15(SH2_struct* sh)
{
  SH2movwi(sh, 13, (7<<4)+15);
}
#define init_9_13_8_0
void FASTCALL call_9_13_8_0(SH2_struct* sh)
{
  SH2movwi(sh, 13, (8<<4)+0);
}
#define init_9_13_8_1
void FASTCALL call_9_13_8_1(SH2_struct* sh)
{
  SH2movwi(sh, 13, (8<<4)+1);
}
#define init_9_13_8_2
void FASTCALL call_9_13_8_2(SH2_struct* sh)
{
  SH2movwi(sh, 13, (8<<4)+2);
}
#define init_9_13_8_3
void FASTCALL call_9_13_8_3(SH2_struct* sh)
{
  SH2movwi(sh, 13, (8<<4)+3);
}
#define init_9_13_8_4
void FASTCALL call_9_13_8_4(SH2_struct* sh)
{
  SH2movwi(sh, 13, (8<<4)+4);
}
#define init_9_13_8_5
void FASTCALL call_9_13_8_5(SH2_struct* sh)
{
  SH2movwi(sh, 13, (8<<4)+5);
}
#define init_9_13_8_6
void FASTCALL call_9_13_8_6(SH2_struct* sh)
{
  SH2movwi(sh, 13, (8<<4)+6);
}
#define init_9_13_8_7
void FASTCALL call_9_13_8_7(SH2_struct* sh)
{
  SH2movwi(sh, 13, (8<<4)+7);
}
#define init_9_13_8_8
void FASTCALL call_9_13_8_8(SH2_struct* sh)
{
  SH2movwi(sh, 13, (8<<4)+8);
}
#define init_9_13_8_9
void FASTCALL call_9_13_8_9(SH2_struct* sh)
{
  SH2movwi(sh, 13, (8<<4)+9);
}
#define init_9_13_8_10
void FASTCALL call_9_13_8_10(SH2_struct* sh)
{
  SH2movwi(sh, 13, (8<<4)+10);
}
#define init_9_13_8_11
void FASTCALL call_9_13_8_11(SH2_struct* sh)
{
  SH2movwi(sh, 13, (8<<4)+11);
}
#define init_9_13_8_12
void FASTCALL call_9_13_8_12(SH2_struct* sh)
{
  SH2movwi(sh, 13, (8<<4)+12);
}
#define init_9_13_8_13
void FASTCALL call_9_13_8_13(SH2_struct* sh)
{
  SH2movwi(sh, 13, (8<<4)+13);
}
#define init_9_13_8_14
void FASTCALL call_9_13_8_14(SH2_struct* sh)
{
  SH2movwi(sh, 13, (8<<4)+14);
}
#define init_9_13_8_15
void FASTCALL call_9_13_8_15(SH2_struct* sh)
{
  SH2movwi(sh, 13, (8<<4)+15);
}
#define init_9_13_9_0
void FASTCALL call_9_13_9_0(SH2_struct* sh)
{
  SH2movwi(sh, 13, (9<<4)+0);
}
#define init_9_13_9_1
void FASTCALL call_9_13_9_1(SH2_struct* sh)
{
  SH2movwi(sh, 13, (9<<4)+1);
}
#define init_9_13_9_2
void FASTCALL call_9_13_9_2(SH2_struct* sh)
{
  SH2movwi(sh, 13, (9<<4)+2);
}
#define init_9_13_9_3
void FASTCALL call_9_13_9_3(SH2_struct* sh)
{
  SH2movwi(sh, 13, (9<<4)+3);
}
#define init_9_13_9_4
void FASTCALL call_9_13_9_4(SH2_struct* sh)
{
  SH2movwi(sh, 13, (9<<4)+4);
}
#define init_9_13_9_5
void FASTCALL call_9_13_9_5(SH2_struct* sh)
{
  SH2movwi(sh, 13, (9<<4)+5);
}
#define init_9_13_9_6
void FASTCALL call_9_13_9_6(SH2_struct* sh)
{
  SH2movwi(sh, 13, (9<<4)+6);
}
#define init_9_13_9_7
void FASTCALL call_9_13_9_7(SH2_struct* sh)
{
  SH2movwi(sh, 13, (9<<4)+7);
}
#define init_9_13_9_8
void FASTCALL call_9_13_9_8(SH2_struct* sh)
{
  SH2movwi(sh, 13, (9<<4)+8);
}
#define init_9_13_9_9
void FASTCALL call_9_13_9_9(SH2_struct* sh)
{
  SH2movwi(sh, 13, (9<<4)+9);
}
#define init_9_13_9_10
void FASTCALL call_9_13_9_10(SH2_struct* sh)
{
  SH2movwi(sh, 13, (9<<4)+10);
}
#define init_9_13_9_11
void FASTCALL call_9_13_9_11(SH2_struct* sh)
{
  SH2movwi(sh, 13, (9<<4)+11);
}
#define init_9_13_9_12
void FASTCALL call_9_13_9_12(SH2_struct* sh)
{
  SH2movwi(sh, 13, (9<<4)+12);
}
#define init_9_13_9_13
void FASTCALL call_9_13_9_13(SH2_struct* sh)
{
  SH2movwi(sh, 13, (9<<4)+13);
}
#define init_9_13_9_14
void FASTCALL call_9_13_9_14(SH2_struct* sh)
{
  SH2movwi(sh, 13, (9<<4)+14);
}
#define init_9_13_9_15
void FASTCALL call_9_13_9_15(SH2_struct* sh)
{
  SH2movwi(sh, 13, (9<<4)+15);
}
#define init_9_13_10_0
void FASTCALL call_9_13_10_0(SH2_struct* sh)
{
  SH2movwi(sh, 13, (10<<4)+0);
}
#define init_9_13_10_1
void FASTCALL call_9_13_10_1(SH2_struct* sh)
{
  SH2movwi(sh, 13, (10<<4)+1);
}
#define init_9_13_10_2
void FASTCALL call_9_13_10_2(SH2_struct* sh)
{
  SH2movwi(sh, 13, (10<<4)+2);
}
#define init_9_13_10_3
void FASTCALL call_9_13_10_3(SH2_struct* sh)
{
  SH2movwi(sh, 13, (10<<4)+3);
}
#define init_9_13_10_4
void FASTCALL call_9_13_10_4(SH2_struct* sh)
{
  SH2movwi(sh, 13, (10<<4)+4);
}
#define init_9_13_10_5
void FASTCALL call_9_13_10_5(SH2_struct* sh)
{
  SH2movwi(sh, 13, (10<<4)+5);
}
#define init_9_13_10_6
void FASTCALL call_9_13_10_6(SH2_struct* sh)
{
  SH2movwi(sh, 13, (10<<4)+6);
}
#define init_9_13_10_7
void FASTCALL call_9_13_10_7(SH2_struct* sh)
{
  SH2movwi(sh, 13, (10<<4)+7);
}
#define init_9_13_10_8
void FASTCALL call_9_13_10_8(SH2_struct* sh)
{
  SH2movwi(sh, 13, (10<<4)+8);
}
#define init_9_13_10_9
void FASTCALL call_9_13_10_9(SH2_struct* sh)
{
  SH2movwi(sh, 13, (10<<4)+9);
}
#define init_9_13_10_10
void FASTCALL call_9_13_10_10(SH2_struct* sh)
{
  SH2movwi(sh, 13, (10<<4)+10);
}
#define init_9_13_10_11
void FASTCALL call_9_13_10_11(SH2_struct* sh)
{
  SH2movwi(sh, 13, (10<<4)+11);
}
#define init_9_13_10_12
void FASTCALL call_9_13_10_12(SH2_struct* sh)
{
  SH2movwi(sh, 13, (10<<4)+12);
}
#define init_9_13_10_13
void FASTCALL call_9_13_10_13(SH2_struct* sh)
{
  SH2movwi(sh, 13, (10<<4)+13);
}
#define init_9_13_10_14
void FASTCALL call_9_13_10_14(SH2_struct* sh)
{
  SH2movwi(sh, 13, (10<<4)+14);
}
#define init_9_13_10_15
void FASTCALL call_9_13_10_15(SH2_struct* sh)
{
  SH2movwi(sh, 13, (10<<4)+15);
}
#define init_9_13_11_0
void FASTCALL call_9_13_11_0(SH2_struct* sh)
{
  SH2movwi(sh, 13, (11<<4)+0);
}
#define init_9_13_11_1
void FASTCALL call_9_13_11_1(SH2_struct* sh)
{
  SH2movwi(sh, 13, (11<<4)+1);
}
#define init_9_13_11_2
void FASTCALL call_9_13_11_2(SH2_struct* sh)
{
  SH2movwi(sh, 13, (11<<4)+2);
}
#define init_9_13_11_3
void FASTCALL call_9_13_11_3(SH2_struct* sh)
{
  SH2movwi(sh, 13, (11<<4)+3);
}
#define init_9_13_11_4
void FASTCALL call_9_13_11_4(SH2_struct* sh)
{
  SH2movwi(sh, 13, (11<<4)+4);
}
#define init_9_13_11_5
void FASTCALL call_9_13_11_5(SH2_struct* sh)
{
  SH2movwi(sh, 13, (11<<4)+5);
}
#define init_9_13_11_6
void FASTCALL call_9_13_11_6(SH2_struct* sh)
{
  SH2movwi(sh, 13, (11<<4)+6);
}
#define init_9_13_11_7
void FASTCALL call_9_13_11_7(SH2_struct* sh)
{
  SH2movwi(sh, 13, (11<<4)+7);
}
#define init_9_13_11_8
void FASTCALL call_9_13_11_8(SH2_struct* sh)
{
  SH2movwi(sh, 13, (11<<4)+8);
}
#define init_9_13_11_9
void FASTCALL call_9_13_11_9(SH2_struct* sh)
{
  SH2movwi(sh, 13, (11<<4)+9);
}
#define init_9_13_11_10
void FASTCALL call_9_13_11_10(SH2_struct* sh)
{
  SH2movwi(sh, 13, (11<<4)+10);
}
#define init_9_13_11_11
void FASTCALL call_9_13_11_11(SH2_struct* sh)
{
  SH2movwi(sh, 13, (11<<4)+11);
}
#define init_9_13_11_12
void FASTCALL call_9_13_11_12(SH2_struct* sh)
{
  SH2movwi(sh, 13, (11<<4)+12);
}
#define init_9_13_11_13
void FASTCALL call_9_13_11_13(SH2_struct* sh)
{
  SH2movwi(sh, 13, (11<<4)+13);
}
#define init_9_13_11_14
void FASTCALL call_9_13_11_14(SH2_struct* sh)
{
  SH2movwi(sh, 13, (11<<4)+14);
}
#define init_9_13_11_15
void FASTCALL call_9_13_11_15(SH2_struct* sh)
{
  SH2movwi(sh, 13, (11<<4)+15);
}
#define init_9_13_12_0
void FASTCALL call_9_13_12_0(SH2_struct* sh)
{
  SH2movwi(sh, 13, (12<<4)+0);
}
#define init_9_13_12_1
void FASTCALL call_9_13_12_1(SH2_struct* sh)
{
  SH2movwi(sh, 13, (12<<4)+1);
}
#define init_9_13_12_2
void FASTCALL call_9_13_12_2(SH2_struct* sh)
{
  SH2movwi(sh, 13, (12<<4)+2);
}
#define init_9_13_12_3
void FASTCALL call_9_13_12_3(SH2_struct* sh)
{
  SH2movwi(sh, 13, (12<<4)+3);
}
#define init_9_13_12_4
void FASTCALL call_9_13_12_4(SH2_struct* sh)
{
  SH2movwi(sh, 13, (12<<4)+4);
}
#define init_9_13_12_5
void FASTCALL call_9_13_12_5(SH2_struct* sh)
{
  SH2movwi(sh, 13, (12<<4)+5);
}
#define init_9_13_12_6
void FASTCALL call_9_13_12_6(SH2_struct* sh)
{
  SH2movwi(sh, 13, (12<<4)+6);
}
#define init_9_13_12_7
void FASTCALL call_9_13_12_7(SH2_struct* sh)
{
  SH2movwi(sh, 13, (12<<4)+7);
}
#define init_9_13_12_8
void FASTCALL call_9_13_12_8(SH2_struct* sh)
{
  SH2movwi(sh, 13, (12<<4)+8);
}
#define init_9_13_12_9
void FASTCALL call_9_13_12_9(SH2_struct* sh)
{
  SH2movwi(sh, 13, (12<<4)+9);
}
#define init_9_13_12_10
void FASTCALL call_9_13_12_10(SH2_struct* sh)
{
  SH2movwi(sh, 13, (12<<4)+10);
}
#define init_9_13_12_11
void FASTCALL call_9_13_12_11(SH2_struct* sh)
{
  SH2movwi(sh, 13, (12<<4)+11);
}
#define init_9_13_12_12
void FASTCALL call_9_13_12_12(SH2_struct* sh)
{
  SH2movwi(sh, 13, (12<<4)+12);
}
#define init_9_13_12_13
void FASTCALL call_9_13_12_13(SH2_struct* sh)
{
  SH2movwi(sh, 13, (12<<4)+13);
}
#define init_9_13_12_14
void FASTCALL call_9_13_12_14(SH2_struct* sh)
{
  SH2movwi(sh, 13, (12<<4)+14);
}
#define init_9_13_12_15
void FASTCALL call_9_13_12_15(SH2_struct* sh)
{
  SH2movwi(sh, 13, (12<<4)+15);
}
#define init_9_13_13_0
void FASTCALL call_9_13_13_0(SH2_struct* sh)
{
  SH2movwi(sh, 13, (13<<4)+0);
}
#define init_9_13_13_1
void FASTCALL call_9_13_13_1(SH2_struct* sh)
{
  SH2movwi(sh, 13, (13<<4)+1);
}
#define init_9_13_13_2
void FASTCALL call_9_13_13_2(SH2_struct* sh)
{
  SH2movwi(sh, 13, (13<<4)+2);
}
#define init_9_13_13_3
void FASTCALL call_9_13_13_3(SH2_struct* sh)
{
  SH2movwi(sh, 13, (13<<4)+3);
}
#define init_9_13_13_4
void FASTCALL call_9_13_13_4(SH2_struct* sh)
{
  SH2movwi(sh, 13, (13<<4)+4);
}
#define init_9_13_13_5
void FASTCALL call_9_13_13_5(SH2_struct* sh)
{
  SH2movwi(sh, 13, (13<<4)+5);
}
#define init_9_13_13_6
void FASTCALL call_9_13_13_6(SH2_struct* sh)
{
  SH2movwi(sh, 13, (13<<4)+6);
}
#define init_9_13_13_7
void FASTCALL call_9_13_13_7(SH2_struct* sh)
{
  SH2movwi(sh, 13, (13<<4)+7);
}
#define init_9_13_13_8
void FASTCALL call_9_13_13_8(SH2_struct* sh)
{
  SH2movwi(sh, 13, (13<<4)+8);
}
#define init_9_13_13_9
void FASTCALL call_9_13_13_9(SH2_struct* sh)
{
  SH2movwi(sh, 13, (13<<4)+9);
}
#define init_9_13_13_10
void FASTCALL call_9_13_13_10(SH2_struct* sh)
{
  SH2movwi(sh, 13, (13<<4)+10);
}
#define init_9_13_13_11
void FASTCALL call_9_13_13_11(SH2_struct* sh)
{
  SH2movwi(sh, 13, (13<<4)+11);
}
#define init_9_13_13_12
void FASTCALL call_9_13_13_12(SH2_struct* sh)
{
  SH2movwi(sh, 13, (13<<4)+12);
}
#define init_9_13_13_13
void FASTCALL call_9_13_13_13(SH2_struct* sh)
{
  SH2movwi(sh, 13, (13<<4)+13);
}
#define init_9_13_13_14
void FASTCALL call_9_13_13_14(SH2_struct* sh)
{
  SH2movwi(sh, 13, (13<<4)+14);
}
#define init_9_13_13_15
void FASTCALL call_9_13_13_15(SH2_struct* sh)
{
  SH2movwi(sh, 13, (13<<4)+15);
}
#define init_9_13_14_0
void FASTCALL call_9_13_14_0(SH2_struct* sh)
{
  SH2movwi(sh, 13, (14<<4)+0);
}
#define init_9_13_14_1
void FASTCALL call_9_13_14_1(SH2_struct* sh)
{
  SH2movwi(sh, 13, (14<<4)+1);
}
#define init_9_13_14_2
void FASTCALL call_9_13_14_2(SH2_struct* sh)
{
  SH2movwi(sh, 13, (14<<4)+2);
}
#define init_9_13_14_3
void FASTCALL call_9_13_14_3(SH2_struct* sh)
{
  SH2movwi(sh, 13, (14<<4)+3);
}
#define init_9_13_14_4
void FASTCALL call_9_13_14_4(SH2_struct* sh)
{
  SH2movwi(sh, 13, (14<<4)+4);
}
#define init_9_13_14_5
void FASTCALL call_9_13_14_5(SH2_struct* sh)
{
  SH2movwi(sh, 13, (14<<4)+5);
}
#define init_9_13_14_6
void FASTCALL call_9_13_14_6(SH2_struct* sh)
{
  SH2movwi(sh, 13, (14<<4)+6);
}
#define init_9_13_14_7
void FASTCALL call_9_13_14_7(SH2_struct* sh)
{
  SH2movwi(sh, 13, (14<<4)+7);
}
#define init_9_13_14_8
void FASTCALL call_9_13_14_8(SH2_struct* sh)
{
  SH2movwi(sh, 13, (14<<4)+8);
}
#define init_9_13_14_9
void FASTCALL call_9_13_14_9(SH2_struct* sh)
{
  SH2movwi(sh, 13, (14<<4)+9);
}
#define init_9_13_14_10
void FASTCALL call_9_13_14_10(SH2_struct* sh)
{
  SH2movwi(sh, 13, (14<<4)+10);
}
#define init_9_13_14_11
void FASTCALL call_9_13_14_11(SH2_struct* sh)
{
  SH2movwi(sh, 13, (14<<4)+11);
}
#define init_9_13_14_12
void FASTCALL call_9_13_14_12(SH2_struct* sh)
{
  SH2movwi(sh, 13, (14<<4)+12);
}
#define init_9_13_14_13
void FASTCALL call_9_13_14_13(SH2_struct* sh)
{
  SH2movwi(sh, 13, (14<<4)+13);
}
#define init_9_13_14_14
void FASTCALL call_9_13_14_14(SH2_struct* sh)
{
  SH2movwi(sh, 13, (14<<4)+14);
}
#define init_9_13_14_15
void FASTCALL call_9_13_14_15(SH2_struct* sh)
{
  SH2movwi(sh, 13, (14<<4)+15);
}
#define init_9_13_15_0
void FASTCALL call_9_13_15_0(SH2_struct* sh)
{
  SH2movwi(sh, 13, (15<<4)+0);
}
#define init_9_13_15_1
void FASTCALL call_9_13_15_1(SH2_struct* sh)
{
  SH2movwi(sh, 13, (15<<4)+1);
}
#define init_9_13_15_2
void FASTCALL call_9_13_15_2(SH2_struct* sh)
{
  SH2movwi(sh, 13, (15<<4)+2);
}
#define init_9_13_15_3
void FASTCALL call_9_13_15_3(SH2_struct* sh)
{
  SH2movwi(sh, 13, (15<<4)+3);
}
#define init_9_13_15_4
void FASTCALL call_9_13_15_4(SH2_struct* sh)
{
  SH2movwi(sh, 13, (15<<4)+4);
}
#define init_9_13_15_5
void FASTCALL call_9_13_15_5(SH2_struct* sh)
{
  SH2movwi(sh, 13, (15<<4)+5);
}
#define init_9_13_15_6
void FASTCALL call_9_13_15_6(SH2_struct* sh)
{
  SH2movwi(sh, 13, (15<<4)+6);
}
#define init_9_13_15_7
void FASTCALL call_9_13_15_7(SH2_struct* sh)
{
  SH2movwi(sh, 13, (15<<4)+7);
}
#define init_9_13_15_8
void FASTCALL call_9_13_15_8(SH2_struct* sh)
{
  SH2movwi(sh, 13, (15<<4)+8);
}
#define init_9_13_15_9
void FASTCALL call_9_13_15_9(SH2_struct* sh)
{
  SH2movwi(sh, 13, (15<<4)+9);
}
#define init_9_13_15_10
void FASTCALL call_9_13_15_10(SH2_struct* sh)
{
  SH2movwi(sh, 13, (15<<4)+10);
}
#define init_9_13_15_11
void FASTCALL call_9_13_15_11(SH2_struct* sh)
{
  SH2movwi(sh, 13, (15<<4)+11);
}
#define init_9_13_15_12
void FASTCALL call_9_13_15_12(SH2_struct* sh)
{
  SH2movwi(sh, 13, (15<<4)+12);
}
#define init_9_13_15_13
void FASTCALL call_9_13_15_13(SH2_struct* sh)
{
  SH2movwi(sh, 13, (15<<4)+13);
}
#define init_9_13_15_14
void FASTCALL call_9_13_15_14(SH2_struct* sh)
{
  SH2movwi(sh, 13, (15<<4)+14);
}
#define init_9_13_15_15
void FASTCALL call_9_13_15_15(SH2_struct* sh)
{
  SH2movwi(sh, 13, (15<<4)+15);
}
#define init_9_14_0_0
void FASTCALL call_9_14_0_0(SH2_struct* sh)
{
  SH2movwi(sh, 14, (0<<4)+0);
}
#define init_9_14_0_1
void FASTCALL call_9_14_0_1(SH2_struct* sh)
{
  SH2movwi(sh, 14, (0<<4)+1);
}
#define init_9_14_0_2
void FASTCALL call_9_14_0_2(SH2_struct* sh)
{
  SH2movwi(sh, 14, (0<<4)+2);
}
#define init_9_14_0_3
void FASTCALL call_9_14_0_3(SH2_struct* sh)
{
  SH2movwi(sh, 14, (0<<4)+3);
}
#define init_9_14_0_4
void FASTCALL call_9_14_0_4(SH2_struct* sh)
{
  SH2movwi(sh, 14, (0<<4)+4);
}
#define init_9_14_0_5
void FASTCALL call_9_14_0_5(SH2_struct* sh)
{
  SH2movwi(sh, 14, (0<<4)+5);
}
#define init_9_14_0_6
void FASTCALL call_9_14_0_6(SH2_struct* sh)
{
  SH2movwi(sh, 14, (0<<4)+6);
}
#define init_9_14_0_7
void FASTCALL call_9_14_0_7(SH2_struct* sh)
{
  SH2movwi(sh, 14, (0<<4)+7);
}
#define init_9_14_0_8
void FASTCALL call_9_14_0_8(SH2_struct* sh)
{
  SH2movwi(sh, 14, (0<<4)+8);
}
#define init_9_14_0_9
void FASTCALL call_9_14_0_9(SH2_struct* sh)
{
  SH2movwi(sh, 14, (0<<4)+9);
}
#define init_9_14_0_10
void FASTCALL call_9_14_0_10(SH2_struct* sh)
{
  SH2movwi(sh, 14, (0<<4)+10);
}
#define init_9_14_0_11
void FASTCALL call_9_14_0_11(SH2_struct* sh)
{
  SH2movwi(sh, 14, (0<<4)+11);
}
#define init_9_14_0_12
void FASTCALL call_9_14_0_12(SH2_struct* sh)
{
  SH2movwi(sh, 14, (0<<4)+12);
}
#define init_9_14_0_13
void FASTCALL call_9_14_0_13(SH2_struct* sh)
{
  SH2movwi(sh, 14, (0<<4)+13);
}
#define init_9_14_0_14
void FASTCALL call_9_14_0_14(SH2_struct* sh)
{
  SH2movwi(sh, 14, (0<<4)+14);
}
#define init_9_14_0_15
void FASTCALL call_9_14_0_15(SH2_struct* sh)
{
  SH2movwi(sh, 14, (0<<4)+15);
}
#define init_9_14_1_0
void FASTCALL call_9_14_1_0(SH2_struct* sh)
{
  SH2movwi(sh, 14, (1<<4)+0);
}
#define init_9_14_1_1
void FASTCALL call_9_14_1_1(SH2_struct* sh)
{
  SH2movwi(sh, 14, (1<<4)+1);
}
#define init_9_14_1_2
void FASTCALL call_9_14_1_2(SH2_struct* sh)
{
  SH2movwi(sh, 14, (1<<4)+2);
}
#define init_9_14_1_3
void FASTCALL call_9_14_1_3(SH2_struct* sh)
{
  SH2movwi(sh, 14, (1<<4)+3);
}
#define init_9_14_1_4
void FASTCALL call_9_14_1_4(SH2_struct* sh)
{
  SH2movwi(sh, 14, (1<<4)+4);
}
#define init_9_14_1_5
void FASTCALL call_9_14_1_5(SH2_struct* sh)
{
  SH2movwi(sh, 14, (1<<4)+5);
}
#define init_9_14_1_6
void FASTCALL call_9_14_1_6(SH2_struct* sh)
{
  SH2movwi(sh, 14, (1<<4)+6);
}
#define init_9_14_1_7
void FASTCALL call_9_14_1_7(SH2_struct* sh)
{
  SH2movwi(sh, 14, (1<<4)+7);
}
#define init_9_14_1_8
void FASTCALL call_9_14_1_8(SH2_struct* sh)
{
  SH2movwi(sh, 14, (1<<4)+8);
}
#define init_9_14_1_9
void FASTCALL call_9_14_1_9(SH2_struct* sh)
{
  SH2movwi(sh, 14, (1<<4)+9);
}
#define init_9_14_1_10
void FASTCALL call_9_14_1_10(SH2_struct* sh)
{
  SH2movwi(sh, 14, (1<<4)+10);
}
#define init_9_14_1_11
void FASTCALL call_9_14_1_11(SH2_struct* sh)
{
  SH2movwi(sh, 14, (1<<4)+11);
}
#define init_9_14_1_12
void FASTCALL call_9_14_1_12(SH2_struct* sh)
{
  SH2movwi(sh, 14, (1<<4)+12);
}
#define init_9_14_1_13
void FASTCALL call_9_14_1_13(SH2_struct* sh)
{
  SH2movwi(sh, 14, (1<<4)+13);
}
#define init_9_14_1_14
void FASTCALL call_9_14_1_14(SH2_struct* sh)
{
  SH2movwi(sh, 14, (1<<4)+14);
}
#define init_9_14_1_15
void FASTCALL call_9_14_1_15(SH2_struct* sh)
{
  SH2movwi(sh, 14, (1<<4)+15);
}
#define init_9_14_2_0
void FASTCALL call_9_14_2_0(SH2_struct* sh)
{
  SH2movwi(sh, 14, (2<<4)+0);
}
#define init_9_14_2_1
void FASTCALL call_9_14_2_1(SH2_struct* sh)
{
  SH2movwi(sh, 14, (2<<4)+1);
}
#define init_9_14_2_2
void FASTCALL call_9_14_2_2(SH2_struct* sh)
{
  SH2movwi(sh, 14, (2<<4)+2);
}
#define init_9_14_2_3
void FASTCALL call_9_14_2_3(SH2_struct* sh)
{
  SH2movwi(sh, 14, (2<<4)+3);
}
#define init_9_14_2_4
void FASTCALL call_9_14_2_4(SH2_struct* sh)
{
  SH2movwi(sh, 14, (2<<4)+4);
}
#define init_9_14_2_5
void FASTCALL call_9_14_2_5(SH2_struct* sh)
{
  SH2movwi(sh, 14, (2<<4)+5);
}
#define init_9_14_2_6
void FASTCALL call_9_14_2_6(SH2_struct* sh)
{
  SH2movwi(sh, 14, (2<<4)+6);
}
#define init_9_14_2_7
void FASTCALL call_9_14_2_7(SH2_struct* sh)
{
  SH2movwi(sh, 14, (2<<4)+7);
}
#define init_9_14_2_8
void FASTCALL call_9_14_2_8(SH2_struct* sh)
{
  SH2movwi(sh, 14, (2<<4)+8);
}
#define init_9_14_2_9
void FASTCALL call_9_14_2_9(SH2_struct* sh)
{
  SH2movwi(sh, 14, (2<<4)+9);
}
#define init_9_14_2_10
void FASTCALL call_9_14_2_10(SH2_struct* sh)
{
  SH2movwi(sh, 14, (2<<4)+10);
}
#define init_9_14_2_11
void FASTCALL call_9_14_2_11(SH2_struct* sh)
{
  SH2movwi(sh, 14, (2<<4)+11);
}
#define init_9_14_2_12
void FASTCALL call_9_14_2_12(SH2_struct* sh)
{
  SH2movwi(sh, 14, (2<<4)+12);
}
#define init_9_14_2_13
void FASTCALL call_9_14_2_13(SH2_struct* sh)
{
  SH2movwi(sh, 14, (2<<4)+13);
}
#define init_9_14_2_14
void FASTCALL call_9_14_2_14(SH2_struct* sh)
{
  SH2movwi(sh, 14, (2<<4)+14);
}
#define init_9_14_2_15
void FASTCALL call_9_14_2_15(SH2_struct* sh)
{
  SH2movwi(sh, 14, (2<<4)+15);
}
#define init_9_14_3_0
void FASTCALL call_9_14_3_0(SH2_struct* sh)
{
  SH2movwi(sh, 14, (3<<4)+0);
}
#define init_9_14_3_1
void FASTCALL call_9_14_3_1(SH2_struct* sh)
{
  SH2movwi(sh, 14, (3<<4)+1);
}
#define init_9_14_3_2
void FASTCALL call_9_14_3_2(SH2_struct* sh)
{
  SH2movwi(sh, 14, (3<<4)+2);
}
#define init_9_14_3_3
void FASTCALL call_9_14_3_3(SH2_struct* sh)
{
  SH2movwi(sh, 14, (3<<4)+3);
}
#define init_9_14_3_4
void FASTCALL call_9_14_3_4(SH2_struct* sh)
{
  SH2movwi(sh, 14, (3<<4)+4);
}
#define init_9_14_3_5
void FASTCALL call_9_14_3_5(SH2_struct* sh)
{
  SH2movwi(sh, 14, (3<<4)+5);
}
#define init_9_14_3_6
void FASTCALL call_9_14_3_6(SH2_struct* sh)
{
  SH2movwi(sh, 14, (3<<4)+6);
}
#define init_9_14_3_7
void FASTCALL call_9_14_3_7(SH2_struct* sh)
{
  SH2movwi(sh, 14, (3<<4)+7);
}
#define init_9_14_3_8
void FASTCALL call_9_14_3_8(SH2_struct* sh)
{
  SH2movwi(sh, 14, (3<<4)+8);
}
#define init_9_14_3_9
void FASTCALL call_9_14_3_9(SH2_struct* sh)
{
  SH2movwi(sh, 14, (3<<4)+9);
}
#define init_9_14_3_10
void FASTCALL call_9_14_3_10(SH2_struct* sh)
{
  SH2movwi(sh, 14, (3<<4)+10);
}
#define init_9_14_3_11
void FASTCALL call_9_14_3_11(SH2_struct* sh)
{
  SH2movwi(sh, 14, (3<<4)+11);
}
#define init_9_14_3_12
void FASTCALL call_9_14_3_12(SH2_struct* sh)
{
  SH2movwi(sh, 14, (3<<4)+12);
}
#define init_9_14_3_13
void FASTCALL call_9_14_3_13(SH2_struct* sh)
{
  SH2movwi(sh, 14, (3<<4)+13);
}
#define init_9_14_3_14
void FASTCALL call_9_14_3_14(SH2_struct* sh)
{
  SH2movwi(sh, 14, (3<<4)+14);
}
#define init_9_14_3_15
void FASTCALL call_9_14_3_15(SH2_struct* sh)
{
  SH2movwi(sh, 14, (3<<4)+15);
}
#define init_9_14_4_0
void FASTCALL call_9_14_4_0(SH2_struct* sh)
{
  SH2movwi(sh, 14, (4<<4)+0);
}
#define init_9_14_4_1
void FASTCALL call_9_14_4_1(SH2_struct* sh)
{
  SH2movwi(sh, 14, (4<<4)+1);
}
#define init_9_14_4_2
void FASTCALL call_9_14_4_2(SH2_struct* sh)
{
  SH2movwi(sh, 14, (4<<4)+2);
}
#define init_9_14_4_3
void FASTCALL call_9_14_4_3(SH2_struct* sh)
{
  SH2movwi(sh, 14, (4<<4)+3);
}
#define init_9_14_4_4
void FASTCALL call_9_14_4_4(SH2_struct* sh)
{
  SH2movwi(sh, 14, (4<<4)+4);
}
#define init_9_14_4_5
void FASTCALL call_9_14_4_5(SH2_struct* sh)
{
  SH2movwi(sh, 14, (4<<4)+5);
}
#define init_9_14_4_6
void FASTCALL call_9_14_4_6(SH2_struct* sh)
{
  SH2movwi(sh, 14, (4<<4)+6);
}
#define init_9_14_4_7
void FASTCALL call_9_14_4_7(SH2_struct* sh)
{
  SH2movwi(sh, 14, (4<<4)+7);
}
#define init_9_14_4_8
void FASTCALL call_9_14_4_8(SH2_struct* sh)
{
  SH2movwi(sh, 14, (4<<4)+8);
}
#define init_9_14_4_9
void FASTCALL call_9_14_4_9(SH2_struct* sh)
{
  SH2movwi(sh, 14, (4<<4)+9);
}
#define init_9_14_4_10
void FASTCALL call_9_14_4_10(SH2_struct* sh)
{
  SH2movwi(sh, 14, (4<<4)+10);
}
#define init_9_14_4_11
void FASTCALL call_9_14_4_11(SH2_struct* sh)
{
  SH2movwi(sh, 14, (4<<4)+11);
}
#define init_9_14_4_12
void FASTCALL call_9_14_4_12(SH2_struct* sh)
{
  SH2movwi(sh, 14, (4<<4)+12);
}
#define init_9_14_4_13
void FASTCALL call_9_14_4_13(SH2_struct* sh)
{
  SH2movwi(sh, 14, (4<<4)+13);
}
#define init_9_14_4_14
void FASTCALL call_9_14_4_14(SH2_struct* sh)
{
  SH2movwi(sh, 14, (4<<4)+14);
}
#define init_9_14_4_15
void FASTCALL call_9_14_4_15(SH2_struct* sh)
{
  SH2movwi(sh, 14, (4<<4)+15);
}
#define init_9_14_5_0
void FASTCALL call_9_14_5_0(SH2_struct* sh)
{
  SH2movwi(sh, 14, (5<<4)+0);
}
#define init_9_14_5_1
void FASTCALL call_9_14_5_1(SH2_struct* sh)
{
  SH2movwi(sh, 14, (5<<4)+1);
}
#define init_9_14_5_2
void FASTCALL call_9_14_5_2(SH2_struct* sh)
{
  SH2movwi(sh, 14, (5<<4)+2);
}
#define init_9_14_5_3
void FASTCALL call_9_14_5_3(SH2_struct* sh)
{
  SH2movwi(sh, 14, (5<<4)+3);
}
#define init_9_14_5_4
void FASTCALL call_9_14_5_4(SH2_struct* sh)
{
  SH2movwi(sh, 14, (5<<4)+4);
}
#define init_9_14_5_5
void FASTCALL call_9_14_5_5(SH2_struct* sh)
{
  SH2movwi(sh, 14, (5<<4)+5);
}
#define init_9_14_5_6
void FASTCALL call_9_14_5_6(SH2_struct* sh)
{
  SH2movwi(sh, 14, (5<<4)+6);
}
#define init_9_14_5_7
void FASTCALL call_9_14_5_7(SH2_struct* sh)
{
  SH2movwi(sh, 14, (5<<4)+7);
}
#define init_9_14_5_8
void FASTCALL call_9_14_5_8(SH2_struct* sh)
{
  SH2movwi(sh, 14, (5<<4)+8);
}
#define init_9_14_5_9
void FASTCALL call_9_14_5_9(SH2_struct* sh)
{
  SH2movwi(sh, 14, (5<<4)+9);
}
#define init_9_14_5_10
void FASTCALL call_9_14_5_10(SH2_struct* sh)
{
  SH2movwi(sh, 14, (5<<4)+10);
}
#define init_9_14_5_11
void FASTCALL call_9_14_5_11(SH2_struct* sh)
{
  SH2movwi(sh, 14, (5<<4)+11);
}
#define init_9_14_5_12
void FASTCALL call_9_14_5_12(SH2_struct* sh)
{
  SH2movwi(sh, 14, (5<<4)+12);
}
#define init_9_14_5_13
void FASTCALL call_9_14_5_13(SH2_struct* sh)
{
  SH2movwi(sh, 14, (5<<4)+13);
}
#define init_9_14_5_14
void FASTCALL call_9_14_5_14(SH2_struct* sh)
{
  SH2movwi(sh, 14, (5<<4)+14);
}
#define init_9_14_5_15
void FASTCALL call_9_14_5_15(SH2_struct* sh)
{
  SH2movwi(sh, 14, (5<<4)+15);
}
#define init_9_14_6_0
void FASTCALL call_9_14_6_0(SH2_struct* sh)
{
  SH2movwi(sh, 14, (6<<4)+0);
}
#define init_9_14_6_1
void FASTCALL call_9_14_6_1(SH2_struct* sh)
{
  SH2movwi(sh, 14, (6<<4)+1);
}
#define init_9_14_6_2
void FASTCALL call_9_14_6_2(SH2_struct* sh)
{
  SH2movwi(sh, 14, (6<<4)+2);
}
#define init_9_14_6_3
void FASTCALL call_9_14_6_3(SH2_struct* sh)
{
  SH2movwi(sh, 14, (6<<4)+3);
}
#define init_9_14_6_4
void FASTCALL call_9_14_6_4(SH2_struct* sh)
{
  SH2movwi(sh, 14, (6<<4)+4);
}
#define init_9_14_6_5
void FASTCALL call_9_14_6_5(SH2_struct* sh)
{
  SH2movwi(sh, 14, (6<<4)+5);
}
#define init_9_14_6_6
void FASTCALL call_9_14_6_6(SH2_struct* sh)
{
  SH2movwi(sh, 14, (6<<4)+6);
}
#define init_9_14_6_7
void FASTCALL call_9_14_6_7(SH2_struct* sh)
{
  SH2movwi(sh, 14, (6<<4)+7);
}
#define init_9_14_6_8
void FASTCALL call_9_14_6_8(SH2_struct* sh)
{
  SH2movwi(sh, 14, (6<<4)+8);
}
#define init_9_14_6_9
void FASTCALL call_9_14_6_9(SH2_struct* sh)
{
  SH2movwi(sh, 14, (6<<4)+9);
}
#define init_9_14_6_10
void FASTCALL call_9_14_6_10(SH2_struct* sh)
{
  SH2movwi(sh, 14, (6<<4)+10);
}
#define init_9_14_6_11
void FASTCALL call_9_14_6_11(SH2_struct* sh)
{
  SH2movwi(sh, 14, (6<<4)+11);
}
#define init_9_14_6_12
void FASTCALL call_9_14_6_12(SH2_struct* sh)
{
  SH2movwi(sh, 14, (6<<4)+12);
}
#define init_9_14_6_13
void FASTCALL call_9_14_6_13(SH2_struct* sh)
{
  SH2movwi(sh, 14, (6<<4)+13);
}
#define init_9_14_6_14
void FASTCALL call_9_14_6_14(SH2_struct* sh)
{
  SH2movwi(sh, 14, (6<<4)+14);
}
#define init_9_14_6_15
void FASTCALL call_9_14_6_15(SH2_struct* sh)
{
  SH2movwi(sh, 14, (6<<4)+15);
}
#define init_9_14_7_0
void FASTCALL call_9_14_7_0(SH2_struct* sh)
{
  SH2movwi(sh, 14, (7<<4)+0);
}
#define init_9_14_7_1
void FASTCALL call_9_14_7_1(SH2_struct* sh)
{
  SH2movwi(sh, 14, (7<<4)+1);
}
#define init_9_14_7_2
void FASTCALL call_9_14_7_2(SH2_struct* sh)
{
  SH2movwi(sh, 14, (7<<4)+2);
}
#define init_9_14_7_3
void FASTCALL call_9_14_7_3(SH2_struct* sh)
{
  SH2movwi(sh, 14, (7<<4)+3);
}
#define init_9_14_7_4
void FASTCALL call_9_14_7_4(SH2_struct* sh)
{
  SH2movwi(sh, 14, (7<<4)+4);
}
#define init_9_14_7_5
void FASTCALL call_9_14_7_5(SH2_struct* sh)
{
  SH2movwi(sh, 14, (7<<4)+5);
}
#define init_9_14_7_6
void FASTCALL call_9_14_7_6(SH2_struct* sh)
{
  SH2movwi(sh, 14, (7<<4)+6);
}
#define init_9_14_7_7
void FASTCALL call_9_14_7_7(SH2_struct* sh)
{
  SH2movwi(sh, 14, (7<<4)+7);
}
#define init_9_14_7_8
void FASTCALL call_9_14_7_8(SH2_struct* sh)
{
  SH2movwi(sh, 14, (7<<4)+8);
}
#define init_9_14_7_9
void FASTCALL call_9_14_7_9(SH2_struct* sh)
{
  SH2movwi(sh, 14, (7<<4)+9);
}
#define init_9_14_7_10
void FASTCALL call_9_14_7_10(SH2_struct* sh)
{
  SH2movwi(sh, 14, (7<<4)+10);
}
#define init_9_14_7_11
void FASTCALL call_9_14_7_11(SH2_struct* sh)
{
  SH2movwi(sh, 14, (7<<4)+11);
}
#define init_9_14_7_12
void FASTCALL call_9_14_7_12(SH2_struct* sh)
{
  SH2movwi(sh, 14, (7<<4)+12);
}
#define init_9_14_7_13
void FASTCALL call_9_14_7_13(SH2_struct* sh)
{
  SH2movwi(sh, 14, (7<<4)+13);
}
#define init_9_14_7_14
void FASTCALL call_9_14_7_14(SH2_struct* sh)
{
  SH2movwi(sh, 14, (7<<4)+14);
}
#define init_9_14_7_15
void FASTCALL call_9_14_7_15(SH2_struct* sh)
{
  SH2movwi(sh, 14, (7<<4)+15);
}
#define init_9_14_8_0
void FASTCALL call_9_14_8_0(SH2_struct* sh)
{
  SH2movwi(sh, 14, (8<<4)+0);
}
#define init_9_14_8_1
void FASTCALL call_9_14_8_1(SH2_struct* sh)
{
  SH2movwi(sh, 14, (8<<4)+1);
}
#define init_9_14_8_2
void FASTCALL call_9_14_8_2(SH2_struct* sh)
{
  SH2movwi(sh, 14, (8<<4)+2);
}
#define init_9_14_8_3
void FASTCALL call_9_14_8_3(SH2_struct* sh)
{
  SH2movwi(sh, 14, (8<<4)+3);
}
#define init_9_14_8_4
void FASTCALL call_9_14_8_4(SH2_struct* sh)
{
  SH2movwi(sh, 14, (8<<4)+4);
}
#define init_9_14_8_5
void FASTCALL call_9_14_8_5(SH2_struct* sh)
{
  SH2movwi(sh, 14, (8<<4)+5);
}
#define init_9_14_8_6
void FASTCALL call_9_14_8_6(SH2_struct* sh)
{
  SH2movwi(sh, 14, (8<<4)+6);
}
#define init_9_14_8_7
void FASTCALL call_9_14_8_7(SH2_struct* sh)
{
  SH2movwi(sh, 14, (8<<4)+7);
}
#define init_9_14_8_8
void FASTCALL call_9_14_8_8(SH2_struct* sh)
{
  SH2movwi(sh, 14, (8<<4)+8);
}
#define init_9_14_8_9
void FASTCALL call_9_14_8_9(SH2_struct* sh)
{
  SH2movwi(sh, 14, (8<<4)+9);
}
#define init_9_14_8_10
void FASTCALL call_9_14_8_10(SH2_struct* sh)
{
  SH2movwi(sh, 14, (8<<4)+10);
}
#define init_9_14_8_11
void FASTCALL call_9_14_8_11(SH2_struct* sh)
{
  SH2movwi(sh, 14, (8<<4)+11);
}
#define init_9_14_8_12
void FASTCALL call_9_14_8_12(SH2_struct* sh)
{
  SH2movwi(sh, 14, (8<<4)+12);
}
#define init_9_14_8_13
void FASTCALL call_9_14_8_13(SH2_struct* sh)
{
  SH2movwi(sh, 14, (8<<4)+13);
}
#define init_9_14_8_14
void FASTCALL call_9_14_8_14(SH2_struct* sh)
{
  SH2movwi(sh, 14, (8<<4)+14);
}
#define init_9_14_8_15
void FASTCALL call_9_14_8_15(SH2_struct* sh)
{
  SH2movwi(sh, 14, (8<<4)+15);
}
#define init_9_14_9_0
void FASTCALL call_9_14_9_0(SH2_struct* sh)
{
  SH2movwi(sh, 14, (9<<4)+0);
}
#define init_9_14_9_1
void FASTCALL call_9_14_9_1(SH2_struct* sh)
{
  SH2movwi(sh, 14, (9<<4)+1);
}
#define init_9_14_9_2
void FASTCALL call_9_14_9_2(SH2_struct* sh)
{
  SH2movwi(sh, 14, (9<<4)+2);
}
#define init_9_14_9_3
void FASTCALL call_9_14_9_3(SH2_struct* sh)
{
  SH2movwi(sh, 14, (9<<4)+3);
}
#define init_9_14_9_4
void FASTCALL call_9_14_9_4(SH2_struct* sh)
{
  SH2movwi(sh, 14, (9<<4)+4);
}
#define init_9_14_9_5
void FASTCALL call_9_14_9_5(SH2_struct* sh)
{
  SH2movwi(sh, 14, (9<<4)+5);
}
#define init_9_14_9_6
void FASTCALL call_9_14_9_6(SH2_struct* sh)
{
  SH2movwi(sh, 14, (9<<4)+6);
}
#define init_9_14_9_7
void FASTCALL call_9_14_9_7(SH2_struct* sh)
{
  SH2movwi(sh, 14, (9<<4)+7);
}
#define init_9_14_9_8
void FASTCALL call_9_14_9_8(SH2_struct* sh)
{
  SH2movwi(sh, 14, (9<<4)+8);
}
#define init_9_14_9_9
void FASTCALL call_9_14_9_9(SH2_struct* sh)
{
  SH2movwi(sh, 14, (9<<4)+9);
}
#define init_9_14_9_10
void FASTCALL call_9_14_9_10(SH2_struct* sh)
{
  SH2movwi(sh, 14, (9<<4)+10);
}
#define init_9_14_9_11
void FASTCALL call_9_14_9_11(SH2_struct* sh)
{
  SH2movwi(sh, 14, (9<<4)+11);
}
#define init_9_14_9_12
void FASTCALL call_9_14_9_12(SH2_struct* sh)
{
  SH2movwi(sh, 14, (9<<4)+12);
}
#define init_9_14_9_13
void FASTCALL call_9_14_9_13(SH2_struct* sh)
{
  SH2movwi(sh, 14, (9<<4)+13);
}
#define init_9_14_9_14
void FASTCALL call_9_14_9_14(SH2_struct* sh)
{
  SH2movwi(sh, 14, (9<<4)+14);
}
#define init_9_14_9_15
void FASTCALL call_9_14_9_15(SH2_struct* sh)
{
  SH2movwi(sh, 14, (9<<4)+15);
}
#define init_9_14_10_0
void FASTCALL call_9_14_10_0(SH2_struct* sh)
{
  SH2movwi(sh, 14, (10<<4)+0);
}
#define init_9_14_10_1
void FASTCALL call_9_14_10_1(SH2_struct* sh)
{
  SH2movwi(sh, 14, (10<<4)+1);
}
#define init_9_14_10_2
void FASTCALL call_9_14_10_2(SH2_struct* sh)
{
  SH2movwi(sh, 14, (10<<4)+2);
}
#define init_9_14_10_3
void FASTCALL call_9_14_10_3(SH2_struct* sh)
{
  SH2movwi(sh, 14, (10<<4)+3);
}
#define init_9_14_10_4
void FASTCALL call_9_14_10_4(SH2_struct* sh)
{
  SH2movwi(sh, 14, (10<<4)+4);
}
#define init_9_14_10_5
void FASTCALL call_9_14_10_5(SH2_struct* sh)
{
  SH2movwi(sh, 14, (10<<4)+5);
}
#define init_9_14_10_6
void FASTCALL call_9_14_10_6(SH2_struct* sh)
{
  SH2movwi(sh, 14, (10<<4)+6);
}
#define init_9_14_10_7
void FASTCALL call_9_14_10_7(SH2_struct* sh)
{
  SH2movwi(sh, 14, (10<<4)+7);
}
#define init_9_14_10_8
void FASTCALL call_9_14_10_8(SH2_struct* sh)
{
  SH2movwi(sh, 14, (10<<4)+8);
}
#define init_9_14_10_9
void FASTCALL call_9_14_10_9(SH2_struct* sh)
{
  SH2movwi(sh, 14, (10<<4)+9);
}
#define init_9_14_10_10
void FASTCALL call_9_14_10_10(SH2_struct* sh)
{
  SH2movwi(sh, 14, (10<<4)+10);
}
#define init_9_14_10_11
void FASTCALL call_9_14_10_11(SH2_struct* sh)
{
  SH2movwi(sh, 14, (10<<4)+11);
}
#define init_9_14_10_12
void FASTCALL call_9_14_10_12(SH2_struct* sh)
{
  SH2movwi(sh, 14, (10<<4)+12);
}
#define init_9_14_10_13
void FASTCALL call_9_14_10_13(SH2_struct* sh)
{
  SH2movwi(sh, 14, (10<<4)+13);
}
#define init_9_14_10_14
void FASTCALL call_9_14_10_14(SH2_struct* sh)
{
  SH2movwi(sh, 14, (10<<4)+14);
}
#define init_9_14_10_15
void FASTCALL call_9_14_10_15(SH2_struct* sh)
{
  SH2movwi(sh, 14, (10<<4)+15);
}
#define init_9_14_11_0
void FASTCALL call_9_14_11_0(SH2_struct* sh)
{
  SH2movwi(sh, 14, (11<<4)+0);
}
#define init_9_14_11_1
void FASTCALL call_9_14_11_1(SH2_struct* sh)
{
  SH2movwi(sh, 14, (11<<4)+1);
}
#define init_9_14_11_2
void FASTCALL call_9_14_11_2(SH2_struct* sh)
{
  SH2movwi(sh, 14, (11<<4)+2);
}
#define init_9_14_11_3
void FASTCALL call_9_14_11_3(SH2_struct* sh)
{
  SH2movwi(sh, 14, (11<<4)+3);
}
#define init_9_14_11_4
void FASTCALL call_9_14_11_4(SH2_struct* sh)
{
  SH2movwi(sh, 14, (11<<4)+4);
}
#define init_9_14_11_5
void FASTCALL call_9_14_11_5(SH2_struct* sh)
{
  SH2movwi(sh, 14, (11<<4)+5);
}
#define init_9_14_11_6
void FASTCALL call_9_14_11_6(SH2_struct* sh)
{
  SH2movwi(sh, 14, (11<<4)+6);
}
#define init_9_14_11_7
void FASTCALL call_9_14_11_7(SH2_struct* sh)
{
  SH2movwi(sh, 14, (11<<4)+7);
}
#define init_9_14_11_8
void FASTCALL call_9_14_11_8(SH2_struct* sh)
{
  SH2movwi(sh, 14, (11<<4)+8);
}
#define init_9_14_11_9
void FASTCALL call_9_14_11_9(SH2_struct* sh)
{
  SH2movwi(sh, 14, (11<<4)+9);
}
#define init_9_14_11_10
void FASTCALL call_9_14_11_10(SH2_struct* sh)
{
  SH2movwi(sh, 14, (11<<4)+10);
}
#define init_9_14_11_11
void FASTCALL call_9_14_11_11(SH2_struct* sh)
{
  SH2movwi(sh, 14, (11<<4)+11);
}
#define init_9_14_11_12
void FASTCALL call_9_14_11_12(SH2_struct* sh)
{
  SH2movwi(sh, 14, (11<<4)+12);
}
#define init_9_14_11_13
void FASTCALL call_9_14_11_13(SH2_struct* sh)
{
  SH2movwi(sh, 14, (11<<4)+13);
}
#define init_9_14_11_14
void FASTCALL call_9_14_11_14(SH2_struct* sh)
{
  SH2movwi(sh, 14, (11<<4)+14);
}
#define init_9_14_11_15
void FASTCALL call_9_14_11_15(SH2_struct* sh)
{
  SH2movwi(sh, 14, (11<<4)+15);
}
#define init_9_14_12_0
void FASTCALL call_9_14_12_0(SH2_struct* sh)
{
  SH2movwi(sh, 14, (12<<4)+0);
}
#define init_9_14_12_1
void FASTCALL call_9_14_12_1(SH2_struct* sh)
{
  SH2movwi(sh, 14, (12<<4)+1);
}
#define init_9_14_12_2
void FASTCALL call_9_14_12_2(SH2_struct* sh)
{
  SH2movwi(sh, 14, (12<<4)+2);
}
#define init_9_14_12_3
void FASTCALL call_9_14_12_3(SH2_struct* sh)
{
  SH2movwi(sh, 14, (12<<4)+3);
}
#define init_9_14_12_4
void FASTCALL call_9_14_12_4(SH2_struct* sh)
{
  SH2movwi(sh, 14, (12<<4)+4);
}
#define init_9_14_12_5
void FASTCALL call_9_14_12_5(SH2_struct* sh)
{
  SH2movwi(sh, 14, (12<<4)+5);
}
#define init_9_14_12_6
void FASTCALL call_9_14_12_6(SH2_struct* sh)
{
  SH2movwi(sh, 14, (12<<4)+6);
}
#define init_9_14_12_7
void FASTCALL call_9_14_12_7(SH2_struct* sh)
{
  SH2movwi(sh, 14, (12<<4)+7);
}
#define init_9_14_12_8
void FASTCALL call_9_14_12_8(SH2_struct* sh)
{
  SH2movwi(sh, 14, (12<<4)+8);
}
#define init_9_14_12_9
void FASTCALL call_9_14_12_9(SH2_struct* sh)
{
  SH2movwi(sh, 14, (12<<4)+9);
}
#define init_9_14_12_10
void FASTCALL call_9_14_12_10(SH2_struct* sh)
{
  SH2movwi(sh, 14, (12<<4)+10);
}
#define init_9_14_12_11
void FASTCALL call_9_14_12_11(SH2_struct* sh)
{
  SH2movwi(sh, 14, (12<<4)+11);
}
#define init_9_14_12_12
void FASTCALL call_9_14_12_12(SH2_struct* sh)
{
  SH2movwi(sh, 14, (12<<4)+12);
}
#define init_9_14_12_13
void FASTCALL call_9_14_12_13(SH2_struct* sh)
{
  SH2movwi(sh, 14, (12<<4)+13);
}
#define init_9_14_12_14
void FASTCALL call_9_14_12_14(SH2_struct* sh)
{
  SH2movwi(sh, 14, (12<<4)+14);
}
#define init_9_14_12_15
void FASTCALL call_9_14_12_15(SH2_struct* sh)
{
  SH2movwi(sh, 14, (12<<4)+15);
}
#define init_9_14_13_0
void FASTCALL call_9_14_13_0(SH2_struct* sh)
{
  SH2movwi(sh, 14, (13<<4)+0);
}
#define init_9_14_13_1
void FASTCALL call_9_14_13_1(SH2_struct* sh)
{
  SH2movwi(sh, 14, (13<<4)+1);
}
#define init_9_14_13_2
void FASTCALL call_9_14_13_2(SH2_struct* sh)
{
  SH2movwi(sh, 14, (13<<4)+2);
}
#define init_9_14_13_3
void FASTCALL call_9_14_13_3(SH2_struct* sh)
{
  SH2movwi(sh, 14, (13<<4)+3);
}
#define init_9_14_13_4
void FASTCALL call_9_14_13_4(SH2_struct* sh)
{
  SH2movwi(sh, 14, (13<<4)+4);
}
#define init_9_14_13_5
void FASTCALL call_9_14_13_5(SH2_struct* sh)
{
  SH2movwi(sh, 14, (13<<4)+5);
}
#define init_9_14_13_6
void FASTCALL call_9_14_13_6(SH2_struct* sh)
{
  SH2movwi(sh, 14, (13<<4)+6);
}
#define init_9_14_13_7
void FASTCALL call_9_14_13_7(SH2_struct* sh)
{
  SH2movwi(sh, 14, (13<<4)+7);
}
#define init_9_14_13_8
void FASTCALL call_9_14_13_8(SH2_struct* sh)
{
  SH2movwi(sh, 14, (13<<4)+8);
}
#define init_9_14_13_9
void FASTCALL call_9_14_13_9(SH2_struct* sh)
{
  SH2movwi(sh, 14, (13<<4)+9);
}
#define init_9_14_13_10
void FASTCALL call_9_14_13_10(SH2_struct* sh)
{
  SH2movwi(sh, 14, (13<<4)+10);
}
#define init_9_14_13_11
void FASTCALL call_9_14_13_11(SH2_struct* sh)
{
  SH2movwi(sh, 14, (13<<4)+11);
}
#define init_9_14_13_12
void FASTCALL call_9_14_13_12(SH2_struct* sh)
{
  SH2movwi(sh, 14, (13<<4)+12);
}
#define init_9_14_13_13
void FASTCALL call_9_14_13_13(SH2_struct* sh)
{
  SH2movwi(sh, 14, (13<<4)+13);
}
#define init_9_14_13_14
void FASTCALL call_9_14_13_14(SH2_struct* sh)
{
  SH2movwi(sh, 14, (13<<4)+14);
}
#define init_9_14_13_15
void FASTCALL call_9_14_13_15(SH2_struct* sh)
{
  SH2movwi(sh, 14, (13<<4)+15);
}
#define init_9_14_14_0
void FASTCALL call_9_14_14_0(SH2_struct* sh)
{
  SH2movwi(sh, 14, (14<<4)+0);
}
#define init_9_14_14_1
void FASTCALL call_9_14_14_1(SH2_struct* sh)
{
  SH2movwi(sh, 14, (14<<4)+1);
}
#define init_9_14_14_2
void FASTCALL call_9_14_14_2(SH2_struct* sh)
{
  SH2movwi(sh, 14, (14<<4)+2);
}
#define init_9_14_14_3
void FASTCALL call_9_14_14_3(SH2_struct* sh)
{
  SH2movwi(sh, 14, (14<<4)+3);
}
#define init_9_14_14_4
void FASTCALL call_9_14_14_4(SH2_struct* sh)
{
  SH2movwi(sh, 14, (14<<4)+4);
}
#define init_9_14_14_5
void FASTCALL call_9_14_14_5(SH2_struct* sh)
{
  SH2movwi(sh, 14, (14<<4)+5);
}
#define init_9_14_14_6
void FASTCALL call_9_14_14_6(SH2_struct* sh)
{
  SH2movwi(sh, 14, (14<<4)+6);
}
#define init_9_14_14_7
void FASTCALL call_9_14_14_7(SH2_struct* sh)
{
  SH2movwi(sh, 14, (14<<4)+7);
}
#define init_9_14_14_8
void FASTCALL call_9_14_14_8(SH2_struct* sh)
{
  SH2movwi(sh, 14, (14<<4)+8);
}
#define init_9_14_14_9
void FASTCALL call_9_14_14_9(SH2_struct* sh)
{
  SH2movwi(sh, 14, (14<<4)+9);
}
#define init_9_14_14_10
void FASTCALL call_9_14_14_10(SH2_struct* sh)
{
  SH2movwi(sh, 14, (14<<4)+10);
}
#define init_9_14_14_11
void FASTCALL call_9_14_14_11(SH2_struct* sh)
{
  SH2movwi(sh, 14, (14<<4)+11);
}
#define init_9_14_14_12
void FASTCALL call_9_14_14_12(SH2_struct* sh)
{
  SH2movwi(sh, 14, (14<<4)+12);
}
#define init_9_14_14_13
void FASTCALL call_9_14_14_13(SH2_struct* sh)
{
  SH2movwi(sh, 14, (14<<4)+13);
}
#define init_9_14_14_14
void FASTCALL call_9_14_14_14(SH2_struct* sh)
{
  SH2movwi(sh, 14, (14<<4)+14);
}
#define init_9_14_14_15
void FASTCALL call_9_14_14_15(SH2_struct* sh)
{
  SH2movwi(sh, 14, (14<<4)+15);
}
#define init_9_14_15_0
void FASTCALL call_9_14_15_0(SH2_struct* sh)
{
  SH2movwi(sh, 14, (15<<4)+0);
}
#define init_9_14_15_1
void FASTCALL call_9_14_15_1(SH2_struct* sh)
{
  SH2movwi(sh, 14, (15<<4)+1);
}
#define init_9_14_15_2
void FASTCALL call_9_14_15_2(SH2_struct* sh)
{
  SH2movwi(sh, 14, (15<<4)+2);
}
#define init_9_14_15_3
void FASTCALL call_9_14_15_3(SH2_struct* sh)
{
  SH2movwi(sh, 14, (15<<4)+3);
}
#define init_9_14_15_4
void FASTCALL call_9_14_15_4(SH2_struct* sh)
{
  SH2movwi(sh, 14, (15<<4)+4);
}
#define init_9_14_15_5
void FASTCALL call_9_14_15_5(SH2_struct* sh)
{
  SH2movwi(sh, 14, (15<<4)+5);
}
#define init_9_14_15_6
void FASTCALL call_9_14_15_6(SH2_struct* sh)
{
  SH2movwi(sh, 14, (15<<4)+6);
}
#define init_9_14_15_7
void FASTCALL call_9_14_15_7(SH2_struct* sh)
{
  SH2movwi(sh, 14, (15<<4)+7);
}
#define init_9_14_15_8
void FASTCALL call_9_14_15_8(SH2_struct* sh)
{
  SH2movwi(sh, 14, (15<<4)+8);
}
#define init_9_14_15_9
void FASTCALL call_9_14_15_9(SH2_struct* sh)
{
  SH2movwi(sh, 14, (15<<4)+9);
}
#define init_9_14_15_10
void FASTCALL call_9_14_15_10(SH2_struct* sh)
{
  SH2movwi(sh, 14, (15<<4)+10);
}
#define init_9_14_15_11
void FASTCALL call_9_14_15_11(SH2_struct* sh)
{
  SH2movwi(sh, 14, (15<<4)+11);
}
#define init_9_14_15_12
void FASTCALL call_9_14_15_12(SH2_struct* sh)
{
  SH2movwi(sh, 14, (15<<4)+12);
}
#define init_9_14_15_13
void FASTCALL call_9_14_15_13(SH2_struct* sh)
{
  SH2movwi(sh, 14, (15<<4)+13);
}
#define init_9_14_15_14
void FASTCALL call_9_14_15_14(SH2_struct* sh)
{
  SH2movwi(sh, 14, (15<<4)+14);
}
#define init_9_14_15_15
void FASTCALL call_9_14_15_15(SH2_struct* sh)
{
  SH2movwi(sh, 14, (15<<4)+15);
}
#define init_9_15_0_0
void FASTCALL call_9_15_0_0(SH2_struct* sh)
{
  SH2movwi(sh, 15, (0<<4)+0);
}
#define init_9_15_0_1
void FASTCALL call_9_15_0_1(SH2_struct* sh)
{
  SH2movwi(sh, 15, (0<<4)+1);
}
#define init_9_15_0_2
void FASTCALL call_9_15_0_2(SH2_struct* sh)
{
  SH2movwi(sh, 15, (0<<4)+2);
}
#define init_9_15_0_3
void FASTCALL call_9_15_0_3(SH2_struct* sh)
{
  SH2movwi(sh, 15, (0<<4)+3);
}
#define init_9_15_0_4
void FASTCALL call_9_15_0_4(SH2_struct* sh)
{
  SH2movwi(sh, 15, (0<<4)+4);
}
#define init_9_15_0_5
void FASTCALL call_9_15_0_5(SH2_struct* sh)
{
  SH2movwi(sh, 15, (0<<4)+5);
}
#define init_9_15_0_6
void FASTCALL call_9_15_0_6(SH2_struct* sh)
{
  SH2movwi(sh, 15, (0<<4)+6);
}
#define init_9_15_0_7
void FASTCALL call_9_15_0_7(SH2_struct* sh)
{
  SH2movwi(sh, 15, (0<<4)+7);
}
#define init_9_15_0_8
void FASTCALL call_9_15_0_8(SH2_struct* sh)
{
  SH2movwi(sh, 15, (0<<4)+8);
}
#define init_9_15_0_9
void FASTCALL call_9_15_0_9(SH2_struct* sh)
{
  SH2movwi(sh, 15, (0<<4)+9);
}
#define init_9_15_0_10
void FASTCALL call_9_15_0_10(SH2_struct* sh)
{
  SH2movwi(sh, 15, (0<<4)+10);
}
#define init_9_15_0_11
void FASTCALL call_9_15_0_11(SH2_struct* sh)
{
  SH2movwi(sh, 15, (0<<4)+11);
}
#define init_9_15_0_12
void FASTCALL call_9_15_0_12(SH2_struct* sh)
{
  SH2movwi(sh, 15, (0<<4)+12);
}
#define init_9_15_0_13
void FASTCALL call_9_15_0_13(SH2_struct* sh)
{
  SH2movwi(sh, 15, (0<<4)+13);
}
#define init_9_15_0_14
void FASTCALL call_9_15_0_14(SH2_struct* sh)
{
  SH2movwi(sh, 15, (0<<4)+14);
}
#define init_9_15_0_15
void FASTCALL call_9_15_0_15(SH2_struct* sh)
{
  SH2movwi(sh, 15, (0<<4)+15);
}
#define init_9_15_1_0
void FASTCALL call_9_15_1_0(SH2_struct* sh)
{
  SH2movwi(sh, 15, (1<<4)+0);
}
#define init_9_15_1_1
void FASTCALL call_9_15_1_1(SH2_struct* sh)
{
  SH2movwi(sh, 15, (1<<4)+1);
}
#define init_9_15_1_2
void FASTCALL call_9_15_1_2(SH2_struct* sh)
{
  SH2movwi(sh, 15, (1<<4)+2);
}
#define init_9_15_1_3
void FASTCALL call_9_15_1_3(SH2_struct* sh)
{
  SH2movwi(sh, 15, (1<<4)+3);
}
#define init_9_15_1_4
void FASTCALL call_9_15_1_4(SH2_struct* sh)
{
  SH2movwi(sh, 15, (1<<4)+4);
}
#define init_9_15_1_5
void FASTCALL call_9_15_1_5(SH2_struct* sh)
{
  SH2movwi(sh, 15, (1<<4)+5);
}
#define init_9_15_1_6
void FASTCALL call_9_15_1_6(SH2_struct* sh)
{
  SH2movwi(sh, 15, (1<<4)+6);
}
#define init_9_15_1_7
void FASTCALL call_9_15_1_7(SH2_struct* sh)
{
  SH2movwi(sh, 15, (1<<4)+7);
}
#define init_9_15_1_8
void FASTCALL call_9_15_1_8(SH2_struct* sh)
{
  SH2movwi(sh, 15, (1<<4)+8);
}
#define init_9_15_1_9
void FASTCALL call_9_15_1_9(SH2_struct* sh)
{
  SH2movwi(sh, 15, (1<<4)+9);
}
#define init_9_15_1_10
void FASTCALL call_9_15_1_10(SH2_struct* sh)
{
  SH2movwi(sh, 15, (1<<4)+10);
}
#define init_9_15_1_11
void FASTCALL call_9_15_1_11(SH2_struct* sh)
{
  SH2movwi(sh, 15, (1<<4)+11);
}
#define init_9_15_1_12
void FASTCALL call_9_15_1_12(SH2_struct* sh)
{
  SH2movwi(sh, 15, (1<<4)+12);
}
#define init_9_15_1_13
void FASTCALL call_9_15_1_13(SH2_struct* sh)
{
  SH2movwi(sh, 15, (1<<4)+13);
}
#define init_9_15_1_14
void FASTCALL call_9_15_1_14(SH2_struct* sh)
{
  SH2movwi(sh, 15, (1<<4)+14);
}
#define init_9_15_1_15
void FASTCALL call_9_15_1_15(SH2_struct* sh)
{
  SH2movwi(sh, 15, (1<<4)+15);
}
#define init_9_15_2_0
void FASTCALL call_9_15_2_0(SH2_struct* sh)
{
  SH2movwi(sh, 15, (2<<4)+0);
}
#define init_9_15_2_1
void FASTCALL call_9_15_2_1(SH2_struct* sh)
{
  SH2movwi(sh, 15, (2<<4)+1);
}
#define init_9_15_2_2
void FASTCALL call_9_15_2_2(SH2_struct* sh)
{
  SH2movwi(sh, 15, (2<<4)+2);
}
#define init_9_15_2_3
void FASTCALL call_9_15_2_3(SH2_struct* sh)
{
  SH2movwi(sh, 15, (2<<4)+3);
}
#define init_9_15_2_4
void FASTCALL call_9_15_2_4(SH2_struct* sh)
{
  SH2movwi(sh, 15, (2<<4)+4);
}
#define init_9_15_2_5
void FASTCALL call_9_15_2_5(SH2_struct* sh)
{
  SH2movwi(sh, 15, (2<<4)+5);
}
#define init_9_15_2_6
void FASTCALL call_9_15_2_6(SH2_struct* sh)
{
  SH2movwi(sh, 15, (2<<4)+6);
}
#define init_9_15_2_7
void FASTCALL call_9_15_2_7(SH2_struct* sh)
{
  SH2movwi(sh, 15, (2<<4)+7);
}
#define init_9_15_2_8
void FASTCALL call_9_15_2_8(SH2_struct* sh)
{
  SH2movwi(sh, 15, (2<<4)+8);
}
#define init_9_15_2_9
void FASTCALL call_9_15_2_9(SH2_struct* sh)
{
  SH2movwi(sh, 15, (2<<4)+9);
}
#define init_9_15_2_10
void FASTCALL call_9_15_2_10(SH2_struct* sh)
{
  SH2movwi(sh, 15, (2<<4)+10);
}
#define init_9_15_2_11
void FASTCALL call_9_15_2_11(SH2_struct* sh)
{
  SH2movwi(sh, 15, (2<<4)+11);
}
#define init_9_15_2_12
void FASTCALL call_9_15_2_12(SH2_struct* sh)
{
  SH2movwi(sh, 15, (2<<4)+12);
}
#define init_9_15_2_13
void FASTCALL call_9_15_2_13(SH2_struct* sh)
{
  SH2movwi(sh, 15, (2<<4)+13);
}
#define init_9_15_2_14
void FASTCALL call_9_15_2_14(SH2_struct* sh)
{
  SH2movwi(sh, 15, (2<<4)+14);
}
#define init_9_15_2_15
void FASTCALL call_9_15_2_15(SH2_struct* sh)
{
  SH2movwi(sh, 15, (2<<4)+15);
}
#define init_9_15_3_0
void FASTCALL call_9_15_3_0(SH2_struct* sh)
{
  SH2movwi(sh, 15, (3<<4)+0);
}
#define init_9_15_3_1
void FASTCALL call_9_15_3_1(SH2_struct* sh)
{
  SH2movwi(sh, 15, (3<<4)+1);
}
#define init_9_15_3_2
void FASTCALL call_9_15_3_2(SH2_struct* sh)
{
  SH2movwi(sh, 15, (3<<4)+2);
}
#define init_9_15_3_3
void FASTCALL call_9_15_3_3(SH2_struct* sh)
{
  SH2movwi(sh, 15, (3<<4)+3);
}
#define init_9_15_3_4
void FASTCALL call_9_15_3_4(SH2_struct* sh)
{
  SH2movwi(sh, 15, (3<<4)+4);
}
#define init_9_15_3_5
void FASTCALL call_9_15_3_5(SH2_struct* sh)
{
  SH2movwi(sh, 15, (3<<4)+5);
}
#define init_9_15_3_6
void FASTCALL call_9_15_3_6(SH2_struct* sh)
{
  SH2movwi(sh, 15, (3<<4)+6);
}
#define init_9_15_3_7
void FASTCALL call_9_15_3_7(SH2_struct* sh)
{
  SH2movwi(sh, 15, (3<<4)+7);
}
#define init_9_15_3_8
void FASTCALL call_9_15_3_8(SH2_struct* sh)
{
  SH2movwi(sh, 15, (3<<4)+8);
}
#define init_9_15_3_9
void FASTCALL call_9_15_3_9(SH2_struct* sh)
{
  SH2movwi(sh, 15, (3<<4)+9);
}
#define init_9_15_3_10
void FASTCALL call_9_15_3_10(SH2_struct* sh)
{
  SH2movwi(sh, 15, (3<<4)+10);
}
#define init_9_15_3_11
void FASTCALL call_9_15_3_11(SH2_struct* sh)
{
  SH2movwi(sh, 15, (3<<4)+11);
}
#define init_9_15_3_12
void FASTCALL call_9_15_3_12(SH2_struct* sh)
{
  SH2movwi(sh, 15, (3<<4)+12);
}
#define init_9_15_3_13
void FASTCALL call_9_15_3_13(SH2_struct* sh)
{
  SH2movwi(sh, 15, (3<<4)+13);
}
#define init_9_15_3_14
void FASTCALL call_9_15_3_14(SH2_struct* sh)
{
  SH2movwi(sh, 15, (3<<4)+14);
}
#define init_9_15_3_15
void FASTCALL call_9_15_3_15(SH2_struct* sh)
{
  SH2movwi(sh, 15, (3<<4)+15);
}
#define init_9_15_4_0
void FASTCALL call_9_15_4_0(SH2_struct* sh)
{
  SH2movwi(sh, 15, (4<<4)+0);
}
#define init_9_15_4_1
void FASTCALL call_9_15_4_1(SH2_struct* sh)
{
  SH2movwi(sh, 15, (4<<4)+1);
}
#define init_9_15_4_2
void FASTCALL call_9_15_4_2(SH2_struct* sh)
{
  SH2movwi(sh, 15, (4<<4)+2);
}
#define init_9_15_4_3
void FASTCALL call_9_15_4_3(SH2_struct* sh)
{
  SH2movwi(sh, 15, (4<<4)+3);
}
#define init_9_15_4_4
void FASTCALL call_9_15_4_4(SH2_struct* sh)
{
  SH2movwi(sh, 15, (4<<4)+4);
}
#define init_9_15_4_5
void FASTCALL call_9_15_4_5(SH2_struct* sh)
{
  SH2movwi(sh, 15, (4<<4)+5);
}
#define init_9_15_4_6
void FASTCALL call_9_15_4_6(SH2_struct* sh)
{
  SH2movwi(sh, 15, (4<<4)+6);
}
#define init_9_15_4_7
void FASTCALL call_9_15_4_7(SH2_struct* sh)
{
  SH2movwi(sh, 15, (4<<4)+7);
}
#define init_9_15_4_8
void FASTCALL call_9_15_4_8(SH2_struct* sh)
{
  SH2movwi(sh, 15, (4<<4)+8);
}
#define init_9_15_4_9
void FASTCALL call_9_15_4_9(SH2_struct* sh)
{
  SH2movwi(sh, 15, (4<<4)+9);
}
#define init_9_15_4_10
void FASTCALL call_9_15_4_10(SH2_struct* sh)
{
  SH2movwi(sh, 15, (4<<4)+10);
}
#define init_9_15_4_11
void FASTCALL call_9_15_4_11(SH2_struct* sh)
{
  SH2movwi(sh, 15, (4<<4)+11);
}
#define init_9_15_4_12
void FASTCALL call_9_15_4_12(SH2_struct* sh)
{
  SH2movwi(sh, 15, (4<<4)+12);
}
#define init_9_15_4_13
void FASTCALL call_9_15_4_13(SH2_struct* sh)
{
  SH2movwi(sh, 15, (4<<4)+13);
}
#define init_9_15_4_14
void FASTCALL call_9_15_4_14(SH2_struct* sh)
{
  SH2movwi(sh, 15, (4<<4)+14);
}
#define init_9_15_4_15
void FASTCALL call_9_15_4_15(SH2_struct* sh)
{
  SH2movwi(sh, 15, (4<<4)+15);
}
#define init_9_15_5_0
void FASTCALL call_9_15_5_0(SH2_struct* sh)
{
  SH2movwi(sh, 15, (5<<4)+0);
}
#define init_9_15_5_1
void FASTCALL call_9_15_5_1(SH2_struct* sh)
{
  SH2movwi(sh, 15, (5<<4)+1);
}
#define init_9_15_5_2
void FASTCALL call_9_15_5_2(SH2_struct* sh)
{
  SH2movwi(sh, 15, (5<<4)+2);
}
#define init_9_15_5_3
void FASTCALL call_9_15_5_3(SH2_struct* sh)
{
  SH2movwi(sh, 15, (5<<4)+3);
}
#define init_9_15_5_4
void FASTCALL call_9_15_5_4(SH2_struct* sh)
{
  SH2movwi(sh, 15, (5<<4)+4);
}
#define init_9_15_5_5
void FASTCALL call_9_15_5_5(SH2_struct* sh)
{
  SH2movwi(sh, 15, (5<<4)+5);
}
#define init_9_15_5_6
void FASTCALL call_9_15_5_6(SH2_struct* sh)
{
  SH2movwi(sh, 15, (5<<4)+6);
}
#define init_9_15_5_7
void FASTCALL call_9_15_5_7(SH2_struct* sh)
{
  SH2movwi(sh, 15, (5<<4)+7);
}
#define init_9_15_5_8
void FASTCALL call_9_15_5_8(SH2_struct* sh)
{
  SH2movwi(sh, 15, (5<<4)+8);
}
#define init_9_15_5_9
void FASTCALL call_9_15_5_9(SH2_struct* sh)
{
  SH2movwi(sh, 15, (5<<4)+9);
}
#define init_9_15_5_10
void FASTCALL call_9_15_5_10(SH2_struct* sh)
{
  SH2movwi(sh, 15, (5<<4)+10);
}
#define init_9_15_5_11
void FASTCALL call_9_15_5_11(SH2_struct* sh)
{
  SH2movwi(sh, 15, (5<<4)+11);
}
#define init_9_15_5_12
void FASTCALL call_9_15_5_12(SH2_struct* sh)
{
  SH2movwi(sh, 15, (5<<4)+12);
}
#define init_9_15_5_13
void FASTCALL call_9_15_5_13(SH2_struct* sh)
{
  SH2movwi(sh, 15, (5<<4)+13);
}
#define init_9_15_5_14
void FASTCALL call_9_15_5_14(SH2_struct* sh)
{
  SH2movwi(sh, 15, (5<<4)+14);
}
#define init_9_15_5_15
void FASTCALL call_9_15_5_15(SH2_struct* sh)
{
  SH2movwi(sh, 15, (5<<4)+15);
}
#define init_9_15_6_0
void FASTCALL call_9_15_6_0(SH2_struct* sh)
{
  SH2movwi(sh, 15, (6<<4)+0);
}
#define init_9_15_6_1
void FASTCALL call_9_15_6_1(SH2_struct* sh)
{
  SH2movwi(sh, 15, (6<<4)+1);
}
#define init_9_15_6_2
void FASTCALL call_9_15_6_2(SH2_struct* sh)
{
  SH2movwi(sh, 15, (6<<4)+2);
}
#define init_9_15_6_3
void FASTCALL call_9_15_6_3(SH2_struct* sh)
{
  SH2movwi(sh, 15, (6<<4)+3);
}
#define init_9_15_6_4
void FASTCALL call_9_15_6_4(SH2_struct* sh)
{
  SH2movwi(sh, 15, (6<<4)+4);
}
#define init_9_15_6_5
void FASTCALL call_9_15_6_5(SH2_struct* sh)
{
  SH2movwi(sh, 15, (6<<4)+5);
}
#define init_9_15_6_6
void FASTCALL call_9_15_6_6(SH2_struct* sh)
{
  SH2movwi(sh, 15, (6<<4)+6);
}
#define init_9_15_6_7
void FASTCALL call_9_15_6_7(SH2_struct* sh)
{
  SH2movwi(sh, 15, (6<<4)+7);
}
#define init_9_15_6_8
void FASTCALL call_9_15_6_8(SH2_struct* sh)
{
  SH2movwi(sh, 15, (6<<4)+8);
}
#define init_9_15_6_9
void FASTCALL call_9_15_6_9(SH2_struct* sh)
{
  SH2movwi(sh, 15, (6<<4)+9);
}
#define init_9_15_6_10
void FASTCALL call_9_15_6_10(SH2_struct* sh)
{
  SH2movwi(sh, 15, (6<<4)+10);
}
#define init_9_15_6_11
void FASTCALL call_9_15_6_11(SH2_struct* sh)
{
  SH2movwi(sh, 15, (6<<4)+11);
}
#define init_9_15_6_12
void FASTCALL call_9_15_6_12(SH2_struct* sh)
{
  SH2movwi(sh, 15, (6<<4)+12);
}
#define init_9_15_6_13
void FASTCALL call_9_15_6_13(SH2_struct* sh)
{
  SH2movwi(sh, 15, (6<<4)+13);
}
#define init_9_15_6_14
void FASTCALL call_9_15_6_14(SH2_struct* sh)
{
  SH2movwi(sh, 15, (6<<4)+14);
}
#define init_9_15_6_15
void FASTCALL call_9_15_6_15(SH2_struct* sh)
{
  SH2movwi(sh, 15, (6<<4)+15);
}
#define init_9_15_7_0
void FASTCALL call_9_15_7_0(SH2_struct* sh)
{
  SH2movwi(sh, 15, (7<<4)+0);
}
#define init_9_15_7_1
void FASTCALL call_9_15_7_1(SH2_struct* sh)
{
  SH2movwi(sh, 15, (7<<4)+1);
}
#define init_9_15_7_2
void FASTCALL call_9_15_7_2(SH2_struct* sh)
{
  SH2movwi(sh, 15, (7<<4)+2);
}
#define init_9_15_7_3
void FASTCALL call_9_15_7_3(SH2_struct* sh)
{
  SH2movwi(sh, 15, (7<<4)+3);
}
#define init_9_15_7_4
void FASTCALL call_9_15_7_4(SH2_struct* sh)
{
  SH2movwi(sh, 15, (7<<4)+4);
}
#define init_9_15_7_5
void FASTCALL call_9_15_7_5(SH2_struct* sh)
{
  SH2movwi(sh, 15, (7<<4)+5);
}
#define init_9_15_7_6
void FASTCALL call_9_15_7_6(SH2_struct* sh)
{
  SH2movwi(sh, 15, (7<<4)+6);
}
#define init_9_15_7_7
void FASTCALL call_9_15_7_7(SH2_struct* sh)
{
  SH2movwi(sh, 15, (7<<4)+7);
}
#define init_9_15_7_8
void FASTCALL call_9_15_7_8(SH2_struct* sh)
{
  SH2movwi(sh, 15, (7<<4)+8);
}
#define init_9_15_7_9
void FASTCALL call_9_15_7_9(SH2_struct* sh)
{
  SH2movwi(sh, 15, (7<<4)+9);
}
#define init_9_15_7_10
void FASTCALL call_9_15_7_10(SH2_struct* sh)
{
  SH2movwi(sh, 15, (7<<4)+10);
}
#define init_9_15_7_11
void FASTCALL call_9_15_7_11(SH2_struct* sh)
{
  SH2movwi(sh, 15, (7<<4)+11);
}
#define init_9_15_7_12
void FASTCALL call_9_15_7_12(SH2_struct* sh)
{
  SH2movwi(sh, 15, (7<<4)+12);
}
#define init_9_15_7_13
void FASTCALL call_9_15_7_13(SH2_struct* sh)
{
  SH2movwi(sh, 15, (7<<4)+13);
}
#define init_9_15_7_14
void FASTCALL call_9_15_7_14(SH2_struct* sh)
{
  SH2movwi(sh, 15, (7<<4)+14);
}
#define init_9_15_7_15
void FASTCALL call_9_15_7_15(SH2_struct* sh)
{
  SH2movwi(sh, 15, (7<<4)+15);
}
#define init_9_15_8_0
void FASTCALL call_9_15_8_0(SH2_struct* sh)
{
  SH2movwi(sh, 15, (8<<4)+0);
}
#define init_9_15_8_1
void FASTCALL call_9_15_8_1(SH2_struct* sh)
{
  SH2movwi(sh, 15, (8<<4)+1);
}
#define init_9_15_8_2
void FASTCALL call_9_15_8_2(SH2_struct* sh)
{
  SH2movwi(sh, 15, (8<<4)+2);
}
#define init_9_15_8_3
void FASTCALL call_9_15_8_3(SH2_struct* sh)
{
  SH2movwi(sh, 15, (8<<4)+3);
}
#define init_9_15_8_4
void FASTCALL call_9_15_8_4(SH2_struct* sh)
{
  SH2movwi(sh, 15, (8<<4)+4);
}
#define init_9_15_8_5
void FASTCALL call_9_15_8_5(SH2_struct* sh)
{
  SH2movwi(sh, 15, (8<<4)+5);
}
#define init_9_15_8_6
void FASTCALL call_9_15_8_6(SH2_struct* sh)
{
  SH2movwi(sh, 15, (8<<4)+6);
}
#define init_9_15_8_7
void FASTCALL call_9_15_8_7(SH2_struct* sh)
{
  SH2movwi(sh, 15, (8<<4)+7);
}
#define init_9_15_8_8
void FASTCALL call_9_15_8_8(SH2_struct* sh)
{
  SH2movwi(sh, 15, (8<<4)+8);
}
#define init_9_15_8_9
void FASTCALL call_9_15_8_9(SH2_struct* sh)
{
  SH2movwi(sh, 15, (8<<4)+9);
}
#define init_9_15_8_10
void FASTCALL call_9_15_8_10(SH2_struct* sh)
{
  SH2movwi(sh, 15, (8<<4)+10);
}
#define init_9_15_8_11
void FASTCALL call_9_15_8_11(SH2_struct* sh)
{
  SH2movwi(sh, 15, (8<<4)+11);
}
#define init_9_15_8_12
void FASTCALL call_9_15_8_12(SH2_struct* sh)
{
  SH2movwi(sh, 15, (8<<4)+12);
}
#define init_9_15_8_13
void FASTCALL call_9_15_8_13(SH2_struct* sh)
{
  SH2movwi(sh, 15, (8<<4)+13);
}
#define init_9_15_8_14
void FASTCALL call_9_15_8_14(SH2_struct* sh)
{
  SH2movwi(sh, 15, (8<<4)+14);
}
#define init_9_15_8_15
void FASTCALL call_9_15_8_15(SH2_struct* sh)
{
  SH2movwi(sh, 15, (8<<4)+15);
}
#define init_9_15_9_0
void FASTCALL call_9_15_9_0(SH2_struct* sh)
{
  SH2movwi(sh, 15, (9<<4)+0);
}
#define init_9_15_9_1
void FASTCALL call_9_15_9_1(SH2_struct* sh)
{
  SH2movwi(sh, 15, (9<<4)+1);
}
#define init_9_15_9_2
void FASTCALL call_9_15_9_2(SH2_struct* sh)
{
  SH2movwi(sh, 15, (9<<4)+2);
}
#define init_9_15_9_3
void FASTCALL call_9_15_9_3(SH2_struct* sh)
{
  SH2movwi(sh, 15, (9<<4)+3);
}
#define init_9_15_9_4
void FASTCALL call_9_15_9_4(SH2_struct* sh)
{
  SH2movwi(sh, 15, (9<<4)+4);
}
#define init_9_15_9_5
void FASTCALL call_9_15_9_5(SH2_struct* sh)
{
  SH2movwi(sh, 15, (9<<4)+5);
}
#define init_9_15_9_6
void FASTCALL call_9_15_9_6(SH2_struct* sh)
{
  SH2movwi(sh, 15, (9<<4)+6);
}
#define init_9_15_9_7
void FASTCALL call_9_15_9_7(SH2_struct* sh)
{
  SH2movwi(sh, 15, (9<<4)+7);
}
#define init_9_15_9_8
void FASTCALL call_9_15_9_8(SH2_struct* sh)
{
  SH2movwi(sh, 15, (9<<4)+8);
}
#define init_9_15_9_9
void FASTCALL call_9_15_9_9(SH2_struct* sh)
{
  SH2movwi(sh, 15, (9<<4)+9);
}
#define init_9_15_9_10
void FASTCALL call_9_15_9_10(SH2_struct* sh)
{
  SH2movwi(sh, 15, (9<<4)+10);
}
#define init_9_15_9_11
void FASTCALL call_9_15_9_11(SH2_struct* sh)
{
  SH2movwi(sh, 15, (9<<4)+11);
}
#define init_9_15_9_12
void FASTCALL call_9_15_9_12(SH2_struct* sh)
{
  SH2movwi(sh, 15, (9<<4)+12);
}
#define init_9_15_9_13
void FASTCALL call_9_15_9_13(SH2_struct* sh)
{
  SH2movwi(sh, 15, (9<<4)+13);
}
#define init_9_15_9_14
void FASTCALL call_9_15_9_14(SH2_struct* sh)
{
  SH2movwi(sh, 15, (9<<4)+14);
}
#define init_9_15_9_15
void FASTCALL call_9_15_9_15(SH2_struct* sh)
{
  SH2movwi(sh, 15, (9<<4)+15);
}
#define init_9_15_10_0
void FASTCALL call_9_15_10_0(SH2_struct* sh)
{
  SH2movwi(sh, 15, (10<<4)+0);
}
#define init_9_15_10_1
void FASTCALL call_9_15_10_1(SH2_struct* sh)
{
  SH2movwi(sh, 15, (10<<4)+1);
}
#define init_9_15_10_2
void FASTCALL call_9_15_10_2(SH2_struct* sh)
{
  SH2movwi(sh, 15, (10<<4)+2);
}
#define init_9_15_10_3
void FASTCALL call_9_15_10_3(SH2_struct* sh)
{
  SH2movwi(sh, 15, (10<<4)+3);
}
#define init_9_15_10_4
void FASTCALL call_9_15_10_4(SH2_struct* sh)
{
  SH2movwi(sh, 15, (10<<4)+4);
}
#define init_9_15_10_5
void FASTCALL call_9_15_10_5(SH2_struct* sh)
{
  SH2movwi(sh, 15, (10<<4)+5);
}
#define init_9_15_10_6
void FASTCALL call_9_15_10_6(SH2_struct* sh)
{
  SH2movwi(sh, 15, (10<<4)+6);
}
#define init_9_15_10_7
void FASTCALL call_9_15_10_7(SH2_struct* sh)
{
  SH2movwi(sh, 15, (10<<4)+7);
}
#define init_9_15_10_8
void FASTCALL call_9_15_10_8(SH2_struct* sh)
{
  SH2movwi(sh, 15, (10<<4)+8);
}
#define init_9_15_10_9
void FASTCALL call_9_15_10_9(SH2_struct* sh)
{
  SH2movwi(sh, 15, (10<<4)+9);
}
#define init_9_15_10_10
void FASTCALL call_9_15_10_10(SH2_struct* sh)
{
  SH2movwi(sh, 15, (10<<4)+10);
}
#define init_9_15_10_11
void FASTCALL call_9_15_10_11(SH2_struct* sh)
{
  SH2movwi(sh, 15, (10<<4)+11);
}
#define init_9_15_10_12
void FASTCALL call_9_15_10_12(SH2_struct* sh)
{
  SH2movwi(sh, 15, (10<<4)+12);
}
#define init_9_15_10_13
void FASTCALL call_9_15_10_13(SH2_struct* sh)
{
  SH2movwi(sh, 15, (10<<4)+13);
}
#define init_9_15_10_14
void FASTCALL call_9_15_10_14(SH2_struct* sh)
{
  SH2movwi(sh, 15, (10<<4)+14);
}
#define init_9_15_10_15
void FASTCALL call_9_15_10_15(SH2_struct* sh)
{
  SH2movwi(sh, 15, (10<<4)+15);
}
#define init_9_15_11_0
void FASTCALL call_9_15_11_0(SH2_struct* sh)
{
  SH2movwi(sh, 15, (11<<4)+0);
}
#define init_9_15_11_1
void FASTCALL call_9_15_11_1(SH2_struct* sh)
{
  SH2movwi(sh, 15, (11<<4)+1);
}
#define init_9_15_11_2
void FASTCALL call_9_15_11_2(SH2_struct* sh)
{
  SH2movwi(sh, 15, (11<<4)+2);
}
#define init_9_15_11_3
void FASTCALL call_9_15_11_3(SH2_struct* sh)
{
  SH2movwi(sh, 15, (11<<4)+3);
}
#define init_9_15_11_4
void FASTCALL call_9_15_11_4(SH2_struct* sh)
{
  SH2movwi(sh, 15, (11<<4)+4);
}
#define init_9_15_11_5
void FASTCALL call_9_15_11_5(SH2_struct* sh)
{
  SH2movwi(sh, 15, (11<<4)+5);
}
#define init_9_15_11_6
void FASTCALL call_9_15_11_6(SH2_struct* sh)
{
  SH2movwi(sh, 15, (11<<4)+6);
}
#define init_9_15_11_7
void FASTCALL call_9_15_11_7(SH2_struct* sh)
{
  SH2movwi(sh, 15, (11<<4)+7);
}
#define init_9_15_11_8
void FASTCALL call_9_15_11_8(SH2_struct* sh)
{
  SH2movwi(sh, 15, (11<<4)+8);
}
#define init_9_15_11_9
void FASTCALL call_9_15_11_9(SH2_struct* sh)
{
  SH2movwi(sh, 15, (11<<4)+9);
}
#define init_9_15_11_10
void FASTCALL call_9_15_11_10(SH2_struct* sh)
{
  SH2movwi(sh, 15, (11<<4)+10);
}
#define init_9_15_11_11
void FASTCALL call_9_15_11_11(SH2_struct* sh)
{
  SH2movwi(sh, 15, (11<<4)+11);
}
#define init_9_15_11_12
void FASTCALL call_9_15_11_12(SH2_struct* sh)
{
  SH2movwi(sh, 15, (11<<4)+12);
}
#define init_9_15_11_13
void FASTCALL call_9_15_11_13(SH2_struct* sh)
{
  SH2movwi(sh, 15, (11<<4)+13);
}
#define init_9_15_11_14
void FASTCALL call_9_15_11_14(SH2_struct* sh)
{
  SH2movwi(sh, 15, (11<<4)+14);
}
#define init_9_15_11_15
void FASTCALL call_9_15_11_15(SH2_struct* sh)
{
  SH2movwi(sh, 15, (11<<4)+15);
}
#define init_9_15_12_0
void FASTCALL call_9_15_12_0(SH2_struct* sh)
{
  SH2movwi(sh, 15, (12<<4)+0);
}
#define init_9_15_12_1
void FASTCALL call_9_15_12_1(SH2_struct* sh)
{
  SH2movwi(sh, 15, (12<<4)+1);
}
#define init_9_15_12_2
void FASTCALL call_9_15_12_2(SH2_struct* sh)
{
  SH2movwi(sh, 15, (12<<4)+2);
}
#define init_9_15_12_3
void FASTCALL call_9_15_12_3(SH2_struct* sh)
{
  SH2movwi(sh, 15, (12<<4)+3);
}
#define init_9_15_12_4
void FASTCALL call_9_15_12_4(SH2_struct* sh)
{
  SH2movwi(sh, 15, (12<<4)+4);
}
#define init_9_15_12_5
void FASTCALL call_9_15_12_5(SH2_struct* sh)
{
  SH2movwi(sh, 15, (12<<4)+5);
}
#define init_9_15_12_6
void FASTCALL call_9_15_12_6(SH2_struct* sh)
{
  SH2movwi(sh, 15, (12<<4)+6);
}
#define init_9_15_12_7
void FASTCALL call_9_15_12_7(SH2_struct* sh)
{
  SH2movwi(sh, 15, (12<<4)+7);
}
#define init_9_15_12_8
void FASTCALL call_9_15_12_8(SH2_struct* sh)
{
  SH2movwi(sh, 15, (12<<4)+8);
}
#define init_9_15_12_9
void FASTCALL call_9_15_12_9(SH2_struct* sh)
{
  SH2movwi(sh, 15, (12<<4)+9);
}
#define init_9_15_12_10
void FASTCALL call_9_15_12_10(SH2_struct* sh)
{
  SH2movwi(sh, 15, (12<<4)+10);
}
#define init_9_15_12_11
void FASTCALL call_9_15_12_11(SH2_struct* sh)
{
  SH2movwi(sh, 15, (12<<4)+11);
}
#define init_9_15_12_12
void FASTCALL call_9_15_12_12(SH2_struct* sh)
{
  SH2movwi(sh, 15, (12<<4)+12);
}
#define init_9_15_12_13
void FASTCALL call_9_15_12_13(SH2_struct* sh)
{
  SH2movwi(sh, 15, (12<<4)+13);
}
#define init_9_15_12_14
void FASTCALL call_9_15_12_14(SH2_struct* sh)
{
  SH2movwi(sh, 15, (12<<4)+14);
}
#define init_9_15_12_15
void FASTCALL call_9_15_12_15(SH2_struct* sh)
{
  SH2movwi(sh, 15, (12<<4)+15);
}
#define init_9_15_13_0
void FASTCALL call_9_15_13_0(SH2_struct* sh)
{
  SH2movwi(sh, 15, (13<<4)+0);
}
#define init_9_15_13_1
void FASTCALL call_9_15_13_1(SH2_struct* sh)
{
  SH2movwi(sh, 15, (13<<4)+1);
}
#define init_9_15_13_2
void FASTCALL call_9_15_13_2(SH2_struct* sh)
{
  SH2movwi(sh, 15, (13<<4)+2);
}
#define init_9_15_13_3
void FASTCALL call_9_15_13_3(SH2_struct* sh)
{
  SH2movwi(sh, 15, (13<<4)+3);
}
#define init_9_15_13_4
void FASTCALL call_9_15_13_4(SH2_struct* sh)
{
  SH2movwi(sh, 15, (13<<4)+4);
}
#define init_9_15_13_5
void FASTCALL call_9_15_13_5(SH2_struct* sh)
{
  SH2movwi(sh, 15, (13<<4)+5);
}
#define init_9_15_13_6
void FASTCALL call_9_15_13_6(SH2_struct* sh)
{
  SH2movwi(sh, 15, (13<<4)+6);
}
#define init_9_15_13_7
void FASTCALL call_9_15_13_7(SH2_struct* sh)
{
  SH2movwi(sh, 15, (13<<4)+7);
}
#define init_9_15_13_8
void FASTCALL call_9_15_13_8(SH2_struct* sh)
{
  SH2movwi(sh, 15, (13<<4)+8);
}
#define init_9_15_13_9
void FASTCALL call_9_15_13_9(SH2_struct* sh)
{
  SH2movwi(sh, 15, (13<<4)+9);
}
#define init_9_15_13_10
void FASTCALL call_9_15_13_10(SH2_struct* sh)
{
  SH2movwi(sh, 15, (13<<4)+10);
}
#define init_9_15_13_11
void FASTCALL call_9_15_13_11(SH2_struct* sh)
{
  SH2movwi(sh, 15, (13<<4)+11);
}
#define init_9_15_13_12
void FASTCALL call_9_15_13_12(SH2_struct* sh)
{
  SH2movwi(sh, 15, (13<<4)+12);
}
#define init_9_15_13_13
void FASTCALL call_9_15_13_13(SH2_struct* sh)
{
  SH2movwi(sh, 15, (13<<4)+13);
}
#define init_9_15_13_14
void FASTCALL call_9_15_13_14(SH2_struct* sh)
{
  SH2movwi(sh, 15, (13<<4)+14);
}
#define init_9_15_13_15
void FASTCALL call_9_15_13_15(SH2_struct* sh)
{
  SH2movwi(sh, 15, (13<<4)+15);
}
#define init_9_15_14_0
void FASTCALL call_9_15_14_0(SH2_struct* sh)
{
  SH2movwi(sh, 15, (14<<4)+0);
}
#define init_9_15_14_1
void FASTCALL call_9_15_14_1(SH2_struct* sh)
{
  SH2movwi(sh, 15, (14<<4)+1);
}
#define init_9_15_14_2
void FASTCALL call_9_15_14_2(SH2_struct* sh)
{
  SH2movwi(sh, 15, (14<<4)+2);
}
#define init_9_15_14_3
void FASTCALL call_9_15_14_3(SH2_struct* sh)
{
  SH2movwi(sh, 15, (14<<4)+3);
}
#define init_9_15_14_4
void FASTCALL call_9_15_14_4(SH2_struct* sh)
{
  SH2movwi(sh, 15, (14<<4)+4);
}
#define init_9_15_14_5
void FASTCALL call_9_15_14_5(SH2_struct* sh)
{
  SH2movwi(sh, 15, (14<<4)+5);
}
#define init_9_15_14_6
void FASTCALL call_9_15_14_6(SH2_struct* sh)
{
  SH2movwi(sh, 15, (14<<4)+6);
}
#define init_9_15_14_7
void FASTCALL call_9_15_14_7(SH2_struct* sh)
{
  SH2movwi(sh, 15, (14<<4)+7);
}
#define init_9_15_14_8
void FASTCALL call_9_15_14_8(SH2_struct* sh)
{
  SH2movwi(sh, 15, (14<<4)+8);
}
#define init_9_15_14_9
void FASTCALL call_9_15_14_9(SH2_struct* sh)
{
  SH2movwi(sh, 15, (14<<4)+9);
}
#define init_9_15_14_10
void FASTCALL call_9_15_14_10(SH2_struct* sh)
{
  SH2movwi(sh, 15, (14<<4)+10);
}
#define init_9_15_14_11
void FASTCALL call_9_15_14_11(SH2_struct* sh)
{
  SH2movwi(sh, 15, (14<<4)+11);
}
#define init_9_15_14_12
void FASTCALL call_9_15_14_12(SH2_struct* sh)
{
  SH2movwi(sh, 15, (14<<4)+12);
}
#define init_9_15_14_13
void FASTCALL call_9_15_14_13(SH2_struct* sh)
{
  SH2movwi(sh, 15, (14<<4)+13);
}
#define init_9_15_14_14
void FASTCALL call_9_15_14_14(SH2_struct* sh)
{
  SH2movwi(sh, 15, (14<<4)+14);
}
#define init_9_15_14_15
void FASTCALL call_9_15_14_15(SH2_struct* sh)
{
  SH2movwi(sh, 15, (14<<4)+15);
}
#define init_9_15_15_0
void FASTCALL call_9_15_15_0(SH2_struct* sh)
{
  SH2movwi(sh, 15, (15<<4)+0);
}
#define init_9_15_15_1
void FASTCALL call_9_15_15_1(SH2_struct* sh)
{
  SH2movwi(sh, 15, (15<<4)+1);
}
#define init_9_15_15_2
void FASTCALL call_9_15_15_2(SH2_struct* sh)
{
  SH2movwi(sh, 15, (15<<4)+2);
}
#define init_9_15_15_3
void FASTCALL call_9_15_15_3(SH2_struct* sh)
{
  SH2movwi(sh, 15, (15<<4)+3);
}
#define init_9_15_15_4
void FASTCALL call_9_15_15_4(SH2_struct* sh)
{
  SH2movwi(sh, 15, (15<<4)+4);
}
#define init_9_15_15_5
void FASTCALL call_9_15_15_5(SH2_struct* sh)
{
  SH2movwi(sh, 15, (15<<4)+5);
}
#define init_9_15_15_6
void FASTCALL call_9_15_15_6(SH2_struct* sh)
{
  SH2movwi(sh, 15, (15<<4)+6);
}
#define init_9_15_15_7
void FASTCALL call_9_15_15_7(SH2_struct* sh)
{
  SH2movwi(sh, 15, (15<<4)+7);
}
#define init_9_15_15_8
void FASTCALL call_9_15_15_8(SH2_struct* sh)
{
  SH2movwi(sh, 15, (15<<4)+8);
}
#define init_9_15_15_9
void FASTCALL call_9_15_15_9(SH2_struct* sh)
{
  SH2movwi(sh, 15, (15<<4)+9);
}
#define init_9_15_15_10
void FASTCALL call_9_15_15_10(SH2_struct* sh)
{
  SH2movwi(sh, 15, (15<<4)+10);
}
#define init_9_15_15_11
void FASTCALL call_9_15_15_11(SH2_struct* sh)
{
  SH2movwi(sh, 15, (15<<4)+11);
}
#define init_9_15_15_12
void FASTCALL call_9_15_15_12(SH2_struct* sh)
{
  SH2movwi(sh, 15, (15<<4)+12);
}
#define init_9_15_15_13
void FASTCALL call_9_15_15_13(SH2_struct* sh)
{
  SH2movwi(sh, 15, (15<<4)+13);
}
#define init_9_15_15_14
void FASTCALL call_9_15_15_14(SH2_struct* sh)
{
  SH2movwi(sh, 15, (15<<4)+14);
}
#define init_9_15_15_15
void FASTCALL call_9_15_15_15(SH2_struct* sh)
{
  SH2movwi(sh, 15, (15<<4)+15);
}
#define init_12_2_0_0
void FASTCALL call_12_2_0_0(SH2_struct* sh)
{
  SH2movlsg(sh,(0<<4)+0);
}
#define init_12_2_0_1
void FASTCALL call_12_2_0_1(SH2_struct* sh)
{
  SH2movlsg(sh,(0<<4)+1);
}
#define init_12_2_0_2
void FASTCALL call_12_2_0_2(SH2_struct* sh)
{
  SH2movlsg(sh,(0<<4)+2);
}
#define init_12_2_0_3
void FASTCALL call_12_2_0_3(SH2_struct* sh)
{
  SH2movlsg(sh,(0<<4)+3);
}
#define init_12_2_0_4
void FASTCALL call_12_2_0_4(SH2_struct* sh)
{
  SH2movlsg(sh,(0<<4)+4);
}
#define init_12_2_0_5
void FASTCALL call_12_2_0_5(SH2_struct* sh)
{
  SH2movlsg(sh,(0<<4)+5);
}
#define init_12_2_0_6
void FASTCALL call_12_2_0_6(SH2_struct* sh)
{
  SH2movlsg(sh,(0<<4)+6);
}
#define init_12_2_0_7
void FASTCALL call_12_2_0_7(SH2_struct* sh)
{
  SH2movlsg(sh,(0<<4)+7);
}
#define init_12_2_0_8
void FASTCALL call_12_2_0_8(SH2_struct* sh)
{
  SH2movlsg(sh,(0<<4)+8);
}
#define init_12_2_0_9
void FASTCALL call_12_2_0_9(SH2_struct* sh)
{
  SH2movlsg(sh,(0<<4)+9);
}
#define init_12_2_0_10
void FASTCALL call_12_2_0_10(SH2_struct* sh)
{
  SH2movlsg(sh,(0<<4)+10);
}
#define init_12_2_0_11
void FASTCALL call_12_2_0_11(SH2_struct* sh)
{
  SH2movlsg(sh,(0<<4)+11);
}
#define init_12_2_0_12
void FASTCALL call_12_2_0_12(SH2_struct* sh)
{
  SH2movlsg(sh,(0<<4)+12);
}
#define init_12_2_0_13
void FASTCALL call_12_2_0_13(SH2_struct* sh)
{
  SH2movlsg(sh,(0<<4)+13);
}
#define init_12_2_0_14
void FASTCALL call_12_2_0_14(SH2_struct* sh)
{
  SH2movlsg(sh,(0<<4)+14);
}
#define init_12_2_0_15
void FASTCALL call_12_2_0_15(SH2_struct* sh)
{
  SH2movlsg(sh,(0<<4)+15);
}
#define init_12_2_1_0
void FASTCALL call_12_2_1_0(SH2_struct* sh)
{
  SH2movlsg(sh,(1<<4)+0);
}
#define init_12_2_1_1
void FASTCALL call_12_2_1_1(SH2_struct* sh)
{
  SH2movlsg(sh,(1<<4)+1);
}
#define init_12_2_1_2
void FASTCALL call_12_2_1_2(SH2_struct* sh)
{
  SH2movlsg(sh,(1<<4)+2);
}
#define init_12_2_1_3
void FASTCALL call_12_2_1_3(SH2_struct* sh)
{
  SH2movlsg(sh,(1<<4)+3);
}
#define init_12_2_1_4
void FASTCALL call_12_2_1_4(SH2_struct* sh)
{
  SH2movlsg(sh,(1<<4)+4);
}
#define init_12_2_1_5
void FASTCALL call_12_2_1_5(SH2_struct* sh)
{
  SH2movlsg(sh,(1<<4)+5);
}
#define init_12_2_1_6
void FASTCALL call_12_2_1_6(SH2_struct* sh)
{
  SH2movlsg(sh,(1<<4)+6);
}
#define init_12_2_1_7
void FASTCALL call_12_2_1_7(SH2_struct* sh)
{
  SH2movlsg(sh,(1<<4)+7);
}
#define init_12_2_1_8
void FASTCALL call_12_2_1_8(SH2_struct* sh)
{
  SH2movlsg(sh,(1<<4)+8);
}
#define init_12_2_1_9
void FASTCALL call_12_2_1_9(SH2_struct* sh)
{
  SH2movlsg(sh,(1<<4)+9);
}
#define init_12_2_1_10
void FASTCALL call_12_2_1_10(SH2_struct* sh)
{
  SH2movlsg(sh,(1<<4)+10);
}
#define init_12_2_1_11
void FASTCALL call_12_2_1_11(SH2_struct* sh)
{
  SH2movlsg(sh,(1<<4)+11);
}
#define init_12_2_1_12
void FASTCALL call_12_2_1_12(SH2_struct* sh)
{
  SH2movlsg(sh,(1<<4)+12);
}
#define init_12_2_1_13
void FASTCALL call_12_2_1_13(SH2_struct* sh)
{
  SH2movlsg(sh,(1<<4)+13);
}
#define init_12_2_1_14
void FASTCALL call_12_2_1_14(SH2_struct* sh)
{
  SH2movlsg(sh,(1<<4)+14);
}
#define init_12_2_1_15
void FASTCALL call_12_2_1_15(SH2_struct* sh)
{
  SH2movlsg(sh,(1<<4)+15);
}
#define init_12_2_2_0
void FASTCALL call_12_2_2_0(SH2_struct* sh)
{
  SH2movlsg(sh,(2<<4)+0);
}
#define init_12_2_2_1
void FASTCALL call_12_2_2_1(SH2_struct* sh)
{
  SH2movlsg(sh,(2<<4)+1);
}
#define init_12_2_2_2
void FASTCALL call_12_2_2_2(SH2_struct* sh)
{
  SH2movlsg(sh,(2<<4)+2);
}
#define init_12_2_2_3
void FASTCALL call_12_2_2_3(SH2_struct* sh)
{
  SH2movlsg(sh,(2<<4)+3);
}
#define init_12_2_2_4
void FASTCALL call_12_2_2_4(SH2_struct* sh)
{
  SH2movlsg(sh,(2<<4)+4);
}
#define init_12_2_2_5
void FASTCALL call_12_2_2_5(SH2_struct* sh)
{
  SH2movlsg(sh,(2<<4)+5);
}
#define init_12_2_2_6
void FASTCALL call_12_2_2_6(SH2_struct* sh)
{
  SH2movlsg(sh,(2<<4)+6);
}
#define init_12_2_2_7
void FASTCALL call_12_2_2_7(SH2_struct* sh)
{
  SH2movlsg(sh,(2<<4)+7);
}
#define init_12_2_2_8
void FASTCALL call_12_2_2_8(SH2_struct* sh)
{
  SH2movlsg(sh,(2<<4)+8);
}
#define init_12_2_2_9
void FASTCALL call_12_2_2_9(SH2_struct* sh)
{
  SH2movlsg(sh,(2<<4)+9);
}
#define init_12_2_2_10
void FASTCALL call_12_2_2_10(SH2_struct* sh)
{
  SH2movlsg(sh,(2<<4)+10);
}
#define init_12_2_2_11
void FASTCALL call_12_2_2_11(SH2_struct* sh)
{
  SH2movlsg(sh,(2<<4)+11);
}
#define init_12_2_2_12
void FASTCALL call_12_2_2_12(SH2_struct* sh)
{
  SH2movlsg(sh,(2<<4)+12);
}
#define init_12_2_2_13
void FASTCALL call_12_2_2_13(SH2_struct* sh)
{
  SH2movlsg(sh,(2<<4)+13);
}
#define init_12_2_2_14
void FASTCALL call_12_2_2_14(SH2_struct* sh)
{
  SH2movlsg(sh,(2<<4)+14);
}
#define init_12_2_2_15
void FASTCALL call_12_2_2_15(SH2_struct* sh)
{
  SH2movlsg(sh,(2<<4)+15);
}
#define init_12_2_3_0
void FASTCALL call_12_2_3_0(SH2_struct* sh)
{
  SH2movlsg(sh,(3<<4)+0);
}
#define init_12_2_3_1
void FASTCALL call_12_2_3_1(SH2_struct* sh)
{
  SH2movlsg(sh,(3<<4)+1);
}
#define init_12_2_3_2
void FASTCALL call_12_2_3_2(SH2_struct* sh)
{
  SH2movlsg(sh,(3<<4)+2);
}
#define init_12_2_3_3
void FASTCALL call_12_2_3_3(SH2_struct* sh)
{
  SH2movlsg(sh,(3<<4)+3);
}
#define init_12_2_3_4
void FASTCALL call_12_2_3_4(SH2_struct* sh)
{
  SH2movlsg(sh,(3<<4)+4);
}
#define init_12_2_3_5
void FASTCALL call_12_2_3_5(SH2_struct* sh)
{
  SH2movlsg(sh,(3<<4)+5);
}
#define init_12_2_3_6
void FASTCALL call_12_2_3_6(SH2_struct* sh)
{
  SH2movlsg(sh,(3<<4)+6);
}
#define init_12_2_3_7
void FASTCALL call_12_2_3_7(SH2_struct* sh)
{
  SH2movlsg(sh,(3<<4)+7);
}
#define init_12_2_3_8
void FASTCALL call_12_2_3_8(SH2_struct* sh)
{
  SH2movlsg(sh,(3<<4)+8);
}
#define init_12_2_3_9
void FASTCALL call_12_2_3_9(SH2_struct* sh)
{
  SH2movlsg(sh,(3<<4)+9);
}
#define init_12_2_3_10
void FASTCALL call_12_2_3_10(SH2_struct* sh)
{
  SH2movlsg(sh,(3<<4)+10);
}
#define init_12_2_3_11
void FASTCALL call_12_2_3_11(SH2_struct* sh)
{
  SH2movlsg(sh,(3<<4)+11);
}
#define init_12_2_3_12
void FASTCALL call_12_2_3_12(SH2_struct* sh)
{
  SH2movlsg(sh,(3<<4)+12);
}
#define init_12_2_3_13
void FASTCALL call_12_2_3_13(SH2_struct* sh)
{
  SH2movlsg(sh,(3<<4)+13);
}
#define init_12_2_3_14
void FASTCALL call_12_2_3_14(SH2_struct* sh)
{
  SH2movlsg(sh,(3<<4)+14);
}
#define init_12_2_3_15
void FASTCALL call_12_2_3_15(SH2_struct* sh)
{
  SH2movlsg(sh,(3<<4)+15);
}
#define init_12_2_4_0
void FASTCALL call_12_2_4_0(SH2_struct* sh)
{
  SH2movlsg(sh,(4<<4)+0);
}
#define init_12_2_4_1
void FASTCALL call_12_2_4_1(SH2_struct* sh)
{
  SH2movlsg(sh,(4<<4)+1);
}
#define init_12_2_4_2
void FASTCALL call_12_2_4_2(SH2_struct* sh)
{
  SH2movlsg(sh,(4<<4)+2);
}
#define init_12_2_4_3
void FASTCALL call_12_2_4_3(SH2_struct* sh)
{
  SH2movlsg(sh,(4<<4)+3);
}
#define init_12_2_4_4
void FASTCALL call_12_2_4_4(SH2_struct* sh)
{
  SH2movlsg(sh,(4<<4)+4);
}
#define init_12_2_4_5
void FASTCALL call_12_2_4_5(SH2_struct* sh)
{
  SH2movlsg(sh,(4<<4)+5);
}
#define init_12_2_4_6
void FASTCALL call_12_2_4_6(SH2_struct* sh)
{
  SH2movlsg(sh,(4<<4)+6);
}
#define init_12_2_4_7
void FASTCALL call_12_2_4_7(SH2_struct* sh)
{
  SH2movlsg(sh,(4<<4)+7);
}
#define init_12_2_4_8
void FASTCALL call_12_2_4_8(SH2_struct* sh)
{
  SH2movlsg(sh,(4<<4)+8);
}
#define init_12_2_4_9
void FASTCALL call_12_2_4_9(SH2_struct* sh)
{
  SH2movlsg(sh,(4<<4)+9);
}
#define init_12_2_4_10
void FASTCALL call_12_2_4_10(SH2_struct* sh)
{
  SH2movlsg(sh,(4<<4)+10);
}
#define init_12_2_4_11
void FASTCALL call_12_2_4_11(SH2_struct* sh)
{
  SH2movlsg(sh,(4<<4)+11);
}
#define init_12_2_4_12
void FASTCALL call_12_2_4_12(SH2_struct* sh)
{
  SH2movlsg(sh,(4<<4)+12);
}
#define init_12_2_4_13
void FASTCALL call_12_2_4_13(SH2_struct* sh)
{
  SH2movlsg(sh,(4<<4)+13);
}
#define init_12_2_4_14
void FASTCALL call_12_2_4_14(SH2_struct* sh)
{
  SH2movlsg(sh,(4<<4)+14);
}
#define init_12_2_4_15
void FASTCALL call_12_2_4_15(SH2_struct* sh)
{
  SH2movlsg(sh,(4<<4)+15);
}
#define init_12_2_5_0
void FASTCALL call_12_2_5_0(SH2_struct* sh)
{
  SH2movlsg(sh,(5<<4)+0);
}
#define init_12_2_5_1
void FASTCALL call_12_2_5_1(SH2_struct* sh)
{
  SH2movlsg(sh,(5<<4)+1);
}
#define init_12_2_5_2
void FASTCALL call_12_2_5_2(SH2_struct* sh)
{
  SH2movlsg(sh,(5<<4)+2);
}
#define init_12_2_5_3
void FASTCALL call_12_2_5_3(SH2_struct* sh)
{
  SH2movlsg(sh,(5<<4)+3);
}
#define init_12_2_5_4
void FASTCALL call_12_2_5_4(SH2_struct* sh)
{
  SH2movlsg(sh,(5<<4)+4);
}
#define init_12_2_5_5
void FASTCALL call_12_2_5_5(SH2_struct* sh)
{
  SH2movlsg(sh,(5<<4)+5);
}
#define init_12_2_5_6
void FASTCALL call_12_2_5_6(SH2_struct* sh)
{
  SH2movlsg(sh,(5<<4)+6);
}
#define init_12_2_5_7
void FASTCALL call_12_2_5_7(SH2_struct* sh)
{
  SH2movlsg(sh,(5<<4)+7);
}
#define init_12_2_5_8
void FASTCALL call_12_2_5_8(SH2_struct* sh)
{
  SH2movlsg(sh,(5<<4)+8);
}
#define init_12_2_5_9
void FASTCALL call_12_2_5_9(SH2_struct* sh)
{
  SH2movlsg(sh,(5<<4)+9);
}
#define init_12_2_5_10
void FASTCALL call_12_2_5_10(SH2_struct* sh)
{
  SH2movlsg(sh,(5<<4)+10);
}
#define init_12_2_5_11
void FASTCALL call_12_2_5_11(SH2_struct* sh)
{
  SH2movlsg(sh,(5<<4)+11);
}
#define init_12_2_5_12
void FASTCALL call_12_2_5_12(SH2_struct* sh)
{
  SH2movlsg(sh,(5<<4)+12);
}
#define init_12_2_5_13
void FASTCALL call_12_2_5_13(SH2_struct* sh)
{
  SH2movlsg(sh,(5<<4)+13);
}
#define init_12_2_5_14
void FASTCALL call_12_2_5_14(SH2_struct* sh)
{
  SH2movlsg(sh,(5<<4)+14);
}
#define init_12_2_5_15
void FASTCALL call_12_2_5_15(SH2_struct* sh)
{
  SH2movlsg(sh,(5<<4)+15);
}
#define init_12_2_6_0
void FASTCALL call_12_2_6_0(SH2_struct* sh)
{
  SH2movlsg(sh,(6<<4)+0);
}
#define init_12_2_6_1
void FASTCALL call_12_2_6_1(SH2_struct* sh)
{
  SH2movlsg(sh,(6<<4)+1);
}
#define init_12_2_6_2
void FASTCALL call_12_2_6_2(SH2_struct* sh)
{
  SH2movlsg(sh,(6<<4)+2);
}
#define init_12_2_6_3
void FASTCALL call_12_2_6_3(SH2_struct* sh)
{
  SH2movlsg(sh,(6<<4)+3);
}
#define init_12_2_6_4
void FASTCALL call_12_2_6_4(SH2_struct* sh)
{
  SH2movlsg(sh,(6<<4)+4);
}
#define init_12_2_6_5
void FASTCALL call_12_2_6_5(SH2_struct* sh)
{
  SH2movlsg(sh,(6<<4)+5);
}
#define init_12_2_6_6
void FASTCALL call_12_2_6_6(SH2_struct* sh)
{
  SH2movlsg(sh,(6<<4)+6);
}
#define init_12_2_6_7
void FASTCALL call_12_2_6_7(SH2_struct* sh)
{
  SH2movlsg(sh,(6<<4)+7);
}
#define init_12_2_6_8
void FASTCALL call_12_2_6_8(SH2_struct* sh)
{
  SH2movlsg(sh,(6<<4)+8);
}
#define init_12_2_6_9
void FASTCALL call_12_2_6_9(SH2_struct* sh)
{
  SH2movlsg(sh,(6<<4)+9);
}
#define init_12_2_6_10
void FASTCALL call_12_2_6_10(SH2_struct* sh)
{
  SH2movlsg(sh,(6<<4)+10);
}
#define init_12_2_6_11
void FASTCALL call_12_2_6_11(SH2_struct* sh)
{
  SH2movlsg(sh,(6<<4)+11);
}
#define init_12_2_6_12
void FASTCALL call_12_2_6_12(SH2_struct* sh)
{
  SH2movlsg(sh,(6<<4)+12);
}
#define init_12_2_6_13
void FASTCALL call_12_2_6_13(SH2_struct* sh)
{
  SH2movlsg(sh,(6<<4)+13);
}
#define init_12_2_6_14
void FASTCALL call_12_2_6_14(SH2_struct* sh)
{
  SH2movlsg(sh,(6<<4)+14);
}
#define init_12_2_6_15
void FASTCALL call_12_2_6_15(SH2_struct* sh)
{
  SH2movlsg(sh,(6<<4)+15);
}
#define init_12_2_7_0
void FASTCALL call_12_2_7_0(SH2_struct* sh)
{
  SH2movlsg(sh,(7<<4)+0);
}
#define init_12_2_7_1
void FASTCALL call_12_2_7_1(SH2_struct* sh)
{
  SH2movlsg(sh,(7<<4)+1);
}
#define init_12_2_7_2
void FASTCALL call_12_2_7_2(SH2_struct* sh)
{
  SH2movlsg(sh,(7<<4)+2);
}
#define init_12_2_7_3
void FASTCALL call_12_2_7_3(SH2_struct* sh)
{
  SH2movlsg(sh,(7<<4)+3);
}
#define init_12_2_7_4
void FASTCALL call_12_2_7_4(SH2_struct* sh)
{
  SH2movlsg(sh,(7<<4)+4);
}
#define init_12_2_7_5
void FASTCALL call_12_2_7_5(SH2_struct* sh)
{
  SH2movlsg(sh,(7<<4)+5);
}
#define init_12_2_7_6
void FASTCALL call_12_2_7_6(SH2_struct* sh)
{
  SH2movlsg(sh,(7<<4)+6);
}
#define init_12_2_7_7
void FASTCALL call_12_2_7_7(SH2_struct* sh)
{
  SH2movlsg(sh,(7<<4)+7);
}
#define init_12_2_7_8
void FASTCALL call_12_2_7_8(SH2_struct* sh)
{
  SH2movlsg(sh,(7<<4)+8);
}
#define init_12_2_7_9
void FASTCALL call_12_2_7_9(SH2_struct* sh)
{
  SH2movlsg(sh,(7<<4)+9);
}
#define init_12_2_7_10
void FASTCALL call_12_2_7_10(SH2_struct* sh)
{
  SH2movlsg(sh,(7<<4)+10);
}
#define init_12_2_7_11
void FASTCALL call_12_2_7_11(SH2_struct* sh)
{
  SH2movlsg(sh,(7<<4)+11);
}
#define init_12_2_7_12
void FASTCALL call_12_2_7_12(SH2_struct* sh)
{
  SH2movlsg(sh,(7<<4)+12);
}
#define init_12_2_7_13
void FASTCALL call_12_2_7_13(SH2_struct* sh)
{
  SH2movlsg(sh,(7<<4)+13);
}
#define init_12_2_7_14
void FASTCALL call_12_2_7_14(SH2_struct* sh)
{
  SH2movlsg(sh,(7<<4)+14);
}
#define init_12_2_7_15
void FASTCALL call_12_2_7_15(SH2_struct* sh)
{
  SH2movlsg(sh,(7<<4)+15);
}
#define init_12_2_8_0
void FASTCALL call_12_2_8_0(SH2_struct* sh)
{
  SH2movlsg(sh,(8<<4)+0);
}
#define init_12_2_8_1
void FASTCALL call_12_2_8_1(SH2_struct* sh)
{
  SH2movlsg(sh,(8<<4)+1);
}
#define init_12_2_8_2
void FASTCALL call_12_2_8_2(SH2_struct* sh)
{
  SH2movlsg(sh,(8<<4)+2);
}
#define init_12_2_8_3
void FASTCALL call_12_2_8_3(SH2_struct* sh)
{
  SH2movlsg(sh,(8<<4)+3);
}
#define init_12_2_8_4
void FASTCALL call_12_2_8_4(SH2_struct* sh)
{
  SH2movlsg(sh,(8<<4)+4);
}
#define init_12_2_8_5
void FASTCALL call_12_2_8_5(SH2_struct* sh)
{
  SH2movlsg(sh,(8<<4)+5);
}
#define init_12_2_8_6
void FASTCALL call_12_2_8_6(SH2_struct* sh)
{
  SH2movlsg(sh,(8<<4)+6);
}
#define init_12_2_8_7
void FASTCALL call_12_2_8_7(SH2_struct* sh)
{
  SH2movlsg(sh,(8<<4)+7);
}
#define init_12_2_8_8
void FASTCALL call_12_2_8_8(SH2_struct* sh)
{
  SH2movlsg(sh,(8<<4)+8);
}
#define init_12_2_8_9
void FASTCALL call_12_2_8_9(SH2_struct* sh)
{
  SH2movlsg(sh,(8<<4)+9);
}
#define init_12_2_8_10
void FASTCALL call_12_2_8_10(SH2_struct* sh)
{
  SH2movlsg(sh,(8<<4)+10);
}
#define init_12_2_8_11
void FASTCALL call_12_2_8_11(SH2_struct* sh)
{
  SH2movlsg(sh,(8<<4)+11);
}
#define init_12_2_8_12
void FASTCALL call_12_2_8_12(SH2_struct* sh)
{
  SH2movlsg(sh,(8<<4)+12);
}
#define init_12_2_8_13
void FASTCALL call_12_2_8_13(SH2_struct* sh)
{
  SH2movlsg(sh,(8<<4)+13);
}
#define init_12_2_8_14
void FASTCALL call_12_2_8_14(SH2_struct* sh)
{
  SH2movlsg(sh,(8<<4)+14);
}
#define init_12_2_8_15
void FASTCALL call_12_2_8_15(SH2_struct* sh)
{
  SH2movlsg(sh,(8<<4)+15);
}
#define init_12_2_9_0
void FASTCALL call_12_2_9_0(SH2_struct* sh)
{
  SH2movlsg(sh,(9<<4)+0);
}
#define init_12_2_9_1
void FASTCALL call_12_2_9_1(SH2_struct* sh)
{
  SH2movlsg(sh,(9<<4)+1);
}
#define init_12_2_9_2
void FASTCALL call_12_2_9_2(SH2_struct* sh)
{
  SH2movlsg(sh,(9<<4)+2);
}
#define init_12_2_9_3
void FASTCALL call_12_2_9_3(SH2_struct* sh)
{
  SH2movlsg(sh,(9<<4)+3);
}
#define init_12_2_9_4
void FASTCALL call_12_2_9_4(SH2_struct* sh)
{
  SH2movlsg(sh,(9<<4)+4);
}
#define init_12_2_9_5
void FASTCALL call_12_2_9_5(SH2_struct* sh)
{
  SH2movlsg(sh,(9<<4)+5);
}
#define init_12_2_9_6
void FASTCALL call_12_2_9_6(SH2_struct* sh)
{
  SH2movlsg(sh,(9<<4)+6);
}
#define init_12_2_9_7
void FASTCALL call_12_2_9_7(SH2_struct* sh)
{
  SH2movlsg(sh,(9<<4)+7);
}
#define init_12_2_9_8
void FASTCALL call_12_2_9_8(SH2_struct* sh)
{
  SH2movlsg(sh,(9<<4)+8);
}
#define init_12_2_9_9
void FASTCALL call_12_2_9_9(SH2_struct* sh)
{
  SH2movlsg(sh,(9<<4)+9);
}
#define init_12_2_9_10
void FASTCALL call_12_2_9_10(SH2_struct* sh)
{
  SH2movlsg(sh,(9<<4)+10);
}
#define init_12_2_9_11
void FASTCALL call_12_2_9_11(SH2_struct* sh)
{
  SH2movlsg(sh,(9<<4)+11);
}
#define init_12_2_9_12
void FASTCALL call_12_2_9_12(SH2_struct* sh)
{
  SH2movlsg(sh,(9<<4)+12);
}
#define init_12_2_9_13
void FASTCALL call_12_2_9_13(SH2_struct* sh)
{
  SH2movlsg(sh,(9<<4)+13);
}
#define init_12_2_9_14
void FASTCALL call_12_2_9_14(SH2_struct* sh)
{
  SH2movlsg(sh,(9<<4)+14);
}
#define init_12_2_9_15
void FASTCALL call_12_2_9_15(SH2_struct* sh)
{
  SH2movlsg(sh,(9<<4)+15);
}
#define init_12_2_10_0
void FASTCALL call_12_2_10_0(SH2_struct* sh)
{
  SH2movlsg(sh,(10<<4)+0);
}
#define init_12_2_10_1
void FASTCALL call_12_2_10_1(SH2_struct* sh)
{
  SH2movlsg(sh,(10<<4)+1);
}
#define init_12_2_10_2
void FASTCALL call_12_2_10_2(SH2_struct* sh)
{
  SH2movlsg(sh,(10<<4)+2);
}
#define init_12_2_10_3
void FASTCALL call_12_2_10_3(SH2_struct* sh)
{
  SH2movlsg(sh,(10<<4)+3);
}
#define init_12_2_10_4
void FASTCALL call_12_2_10_4(SH2_struct* sh)
{
  SH2movlsg(sh,(10<<4)+4);
}
#define init_12_2_10_5
void FASTCALL call_12_2_10_5(SH2_struct* sh)
{
  SH2movlsg(sh,(10<<4)+5);
}
#define init_12_2_10_6
void FASTCALL call_12_2_10_6(SH2_struct* sh)
{
  SH2movlsg(sh,(10<<4)+6);
}
#define init_12_2_10_7
void FASTCALL call_12_2_10_7(SH2_struct* sh)
{
  SH2movlsg(sh,(10<<4)+7);
}
#define init_12_2_10_8
void FASTCALL call_12_2_10_8(SH2_struct* sh)
{
  SH2movlsg(sh,(10<<4)+8);
}
#define init_12_2_10_9
void FASTCALL call_12_2_10_9(SH2_struct* sh)
{
  SH2movlsg(sh,(10<<4)+9);
}
#define init_12_2_10_10
void FASTCALL call_12_2_10_10(SH2_struct* sh)
{
  SH2movlsg(sh,(10<<4)+10);
}
#define init_12_2_10_11
void FASTCALL call_12_2_10_11(SH2_struct* sh)
{
  SH2movlsg(sh,(10<<4)+11);
}
#define init_12_2_10_12
void FASTCALL call_12_2_10_12(SH2_struct* sh)
{
  SH2movlsg(sh,(10<<4)+12);
}
#define init_12_2_10_13
void FASTCALL call_12_2_10_13(SH2_struct* sh)
{
  SH2movlsg(sh,(10<<4)+13);
}
#define init_12_2_10_14
void FASTCALL call_12_2_10_14(SH2_struct* sh)
{
  SH2movlsg(sh,(10<<4)+14);
}
#define init_12_2_10_15
void FASTCALL call_12_2_10_15(SH2_struct* sh)
{
  SH2movlsg(sh,(10<<4)+15);
}
#define init_12_2_11_0
void FASTCALL call_12_2_11_0(SH2_struct* sh)
{
  SH2movlsg(sh,(11<<4)+0);
}
#define init_12_2_11_1
void FASTCALL call_12_2_11_1(SH2_struct* sh)
{
  SH2movlsg(sh,(11<<4)+1);
}
#define init_12_2_11_2
void FASTCALL call_12_2_11_2(SH2_struct* sh)
{
  SH2movlsg(sh,(11<<4)+2);
}
#define init_12_2_11_3
void FASTCALL call_12_2_11_3(SH2_struct* sh)
{
  SH2movlsg(sh,(11<<4)+3);
}
#define init_12_2_11_4
void FASTCALL call_12_2_11_4(SH2_struct* sh)
{
  SH2movlsg(sh,(11<<4)+4);
}
#define init_12_2_11_5
void FASTCALL call_12_2_11_5(SH2_struct* sh)
{
  SH2movlsg(sh,(11<<4)+5);
}
#define init_12_2_11_6
void FASTCALL call_12_2_11_6(SH2_struct* sh)
{
  SH2movlsg(sh,(11<<4)+6);
}
#define init_12_2_11_7
void FASTCALL call_12_2_11_7(SH2_struct* sh)
{
  SH2movlsg(sh,(11<<4)+7);
}
#define init_12_2_11_8
void FASTCALL call_12_2_11_8(SH2_struct* sh)
{
  SH2movlsg(sh,(11<<4)+8);
}
#define init_12_2_11_9
void FASTCALL call_12_2_11_9(SH2_struct* sh)
{
  SH2movlsg(sh,(11<<4)+9);
}
#define init_12_2_11_10
void FASTCALL call_12_2_11_10(SH2_struct* sh)
{
  SH2movlsg(sh,(11<<4)+10);
}
#define init_12_2_11_11
void FASTCALL call_12_2_11_11(SH2_struct* sh)
{
  SH2movlsg(sh,(11<<4)+11);
}
#define init_12_2_11_12
void FASTCALL call_12_2_11_12(SH2_struct* sh)
{
  SH2movlsg(sh,(11<<4)+12);
}
#define init_12_2_11_13
void FASTCALL call_12_2_11_13(SH2_struct* sh)
{
  SH2movlsg(sh,(11<<4)+13);
}
#define init_12_2_11_14
void FASTCALL call_12_2_11_14(SH2_struct* sh)
{
  SH2movlsg(sh,(11<<4)+14);
}
#define init_12_2_11_15
void FASTCALL call_12_2_11_15(SH2_struct* sh)
{
  SH2movlsg(sh,(11<<4)+15);
}
#define init_12_2_12_0
void FASTCALL call_12_2_12_0(SH2_struct* sh)
{
  SH2movlsg(sh,(12<<4)+0);
}
#define init_12_2_12_1
void FASTCALL call_12_2_12_1(SH2_struct* sh)
{
  SH2movlsg(sh,(12<<4)+1);
}
#define init_12_2_12_2
void FASTCALL call_12_2_12_2(SH2_struct* sh)
{
  SH2movlsg(sh,(12<<4)+2);
}
#define init_12_2_12_3
void FASTCALL call_12_2_12_3(SH2_struct* sh)
{
  SH2movlsg(sh,(12<<4)+3);
}
#define init_12_2_12_4
void FASTCALL call_12_2_12_4(SH2_struct* sh)
{
  SH2movlsg(sh,(12<<4)+4);
}
#define init_12_2_12_5
void FASTCALL call_12_2_12_5(SH2_struct* sh)
{
  SH2movlsg(sh,(12<<4)+5);
}
#define init_12_2_12_6
void FASTCALL call_12_2_12_6(SH2_struct* sh)
{
  SH2movlsg(sh,(12<<4)+6);
}
#define init_12_2_12_7
void FASTCALL call_12_2_12_7(SH2_struct* sh)
{
  SH2movlsg(sh,(12<<4)+7);
}
#define init_12_2_12_8
void FASTCALL call_12_2_12_8(SH2_struct* sh)
{
  SH2movlsg(sh,(12<<4)+8);
}
#define init_12_2_12_9
void FASTCALL call_12_2_12_9(SH2_struct* sh)
{
  SH2movlsg(sh,(12<<4)+9);
}
#define init_12_2_12_10
void FASTCALL call_12_2_12_10(SH2_struct* sh)
{
  SH2movlsg(sh,(12<<4)+10);
}
#define init_12_2_12_11
void FASTCALL call_12_2_12_11(SH2_struct* sh)
{
  SH2movlsg(sh,(12<<4)+11);
}
#define init_12_2_12_12
void FASTCALL call_12_2_12_12(SH2_struct* sh)
{
  SH2movlsg(sh,(12<<4)+12);
}
#define init_12_2_12_13
void FASTCALL call_12_2_12_13(SH2_struct* sh)
{
  SH2movlsg(sh,(12<<4)+13);
}
#define init_12_2_12_14
void FASTCALL call_12_2_12_14(SH2_struct* sh)
{
  SH2movlsg(sh,(12<<4)+14);
}
#define init_12_2_12_15
void FASTCALL call_12_2_12_15(SH2_struct* sh)
{
  SH2movlsg(sh,(12<<4)+15);
}
#define init_12_2_13_0
void FASTCALL call_12_2_13_0(SH2_struct* sh)
{
  SH2movlsg(sh,(13<<4)+0);
}
#define init_12_2_13_1
void FASTCALL call_12_2_13_1(SH2_struct* sh)
{
  SH2movlsg(sh,(13<<4)+1);
}
#define init_12_2_13_2
void FASTCALL call_12_2_13_2(SH2_struct* sh)
{
  SH2movlsg(sh,(13<<4)+2);
}
#define init_12_2_13_3
void FASTCALL call_12_2_13_3(SH2_struct* sh)
{
  SH2movlsg(sh,(13<<4)+3);
}
#define init_12_2_13_4
void FASTCALL call_12_2_13_4(SH2_struct* sh)
{
  SH2movlsg(sh,(13<<4)+4);
}
#define init_12_2_13_5
void FASTCALL call_12_2_13_5(SH2_struct* sh)
{
  SH2movlsg(sh,(13<<4)+5);
}
#define init_12_2_13_6
void FASTCALL call_12_2_13_6(SH2_struct* sh)
{
  SH2movlsg(sh,(13<<4)+6);
}
#define init_12_2_13_7
void FASTCALL call_12_2_13_7(SH2_struct* sh)
{
  SH2movlsg(sh,(13<<4)+7);
}
#define init_12_2_13_8
void FASTCALL call_12_2_13_8(SH2_struct* sh)
{
  SH2movlsg(sh,(13<<4)+8);
}
#define init_12_2_13_9
void FASTCALL call_12_2_13_9(SH2_struct* sh)
{
  SH2movlsg(sh,(13<<4)+9);
}
#define init_12_2_13_10
void FASTCALL call_12_2_13_10(SH2_struct* sh)
{
  SH2movlsg(sh,(13<<4)+10);
}
#define init_12_2_13_11
void FASTCALL call_12_2_13_11(SH2_struct* sh)
{
  SH2movlsg(sh,(13<<4)+11);
}
#define init_12_2_13_12
void FASTCALL call_12_2_13_12(SH2_struct* sh)
{
  SH2movlsg(sh,(13<<4)+12);
}
#define init_12_2_13_13
void FASTCALL call_12_2_13_13(SH2_struct* sh)
{
  SH2movlsg(sh,(13<<4)+13);
}
#define init_12_2_13_14
void FASTCALL call_12_2_13_14(SH2_struct* sh)
{
  SH2movlsg(sh,(13<<4)+14);
}
#define init_12_2_13_15
void FASTCALL call_12_2_13_15(SH2_struct* sh)
{
  SH2movlsg(sh,(13<<4)+15);
}
#define init_12_2_14_0
void FASTCALL call_12_2_14_0(SH2_struct* sh)
{
  SH2movlsg(sh,(14<<4)+0);
}
#define init_12_2_14_1
void FASTCALL call_12_2_14_1(SH2_struct* sh)
{
  SH2movlsg(sh,(14<<4)+1);
}
#define init_12_2_14_2
void FASTCALL call_12_2_14_2(SH2_struct* sh)
{
  SH2movlsg(sh,(14<<4)+2);
}
#define init_12_2_14_3
void FASTCALL call_12_2_14_3(SH2_struct* sh)
{
  SH2movlsg(sh,(14<<4)+3);
}
#define init_12_2_14_4
void FASTCALL call_12_2_14_4(SH2_struct* sh)
{
  SH2movlsg(sh,(14<<4)+4);
}
#define init_12_2_14_5
void FASTCALL call_12_2_14_5(SH2_struct* sh)
{
  SH2movlsg(sh,(14<<4)+5);
}
#define init_12_2_14_6
void FASTCALL call_12_2_14_6(SH2_struct* sh)
{
  SH2movlsg(sh,(14<<4)+6);
}
#define init_12_2_14_7
void FASTCALL call_12_2_14_7(SH2_struct* sh)
{
  SH2movlsg(sh,(14<<4)+7);
}
#define init_12_2_14_8
void FASTCALL call_12_2_14_8(SH2_struct* sh)
{
  SH2movlsg(sh,(14<<4)+8);
}
#define init_12_2_14_9
void FASTCALL call_12_2_14_9(SH2_struct* sh)
{
  SH2movlsg(sh,(14<<4)+9);
}
#define init_12_2_14_10
void FASTCALL call_12_2_14_10(SH2_struct* sh)
{
  SH2movlsg(sh,(14<<4)+10);
}
#define init_12_2_14_11
void FASTCALL call_12_2_14_11(SH2_struct* sh)
{
  SH2movlsg(sh,(14<<4)+11);
}
#define init_12_2_14_12
void FASTCALL call_12_2_14_12(SH2_struct* sh)
{
  SH2movlsg(sh,(14<<4)+12);
}
#define init_12_2_14_13
void FASTCALL call_12_2_14_13(SH2_struct* sh)
{
  SH2movlsg(sh,(14<<4)+13);
}
#define init_12_2_14_14
void FASTCALL call_12_2_14_14(SH2_struct* sh)
{
  SH2movlsg(sh,(14<<4)+14);
}
#define init_12_2_14_15
void FASTCALL call_12_2_14_15(SH2_struct* sh)
{
  SH2movlsg(sh,(14<<4)+15);
}
#define init_12_2_15_0
void FASTCALL call_12_2_15_0(SH2_struct* sh)
{
  SH2movlsg(sh,(15<<4)+0);
}
#define init_12_2_15_1
void FASTCALL call_12_2_15_1(SH2_struct* sh)
{
  SH2movlsg(sh,(15<<4)+1);
}
#define init_12_2_15_2
void FASTCALL call_12_2_15_2(SH2_struct* sh)
{
  SH2movlsg(sh,(15<<4)+2);
}
#define init_12_2_15_3
void FASTCALL call_12_2_15_3(SH2_struct* sh)
{
  SH2movlsg(sh,(15<<4)+3);
}
#define init_12_2_15_4
void FASTCALL call_12_2_15_4(SH2_struct* sh)
{
  SH2movlsg(sh,(15<<4)+4);
}
#define init_12_2_15_5
void FASTCALL call_12_2_15_5(SH2_struct* sh)
{
  SH2movlsg(sh,(15<<4)+5);
}
#define init_12_2_15_6
void FASTCALL call_12_2_15_6(SH2_struct* sh)
{
  SH2movlsg(sh,(15<<4)+6);
}
#define init_12_2_15_7
void FASTCALL call_12_2_15_7(SH2_struct* sh)
{
  SH2movlsg(sh,(15<<4)+7);
}
#define init_12_2_15_8
void FASTCALL call_12_2_15_8(SH2_struct* sh)
{
  SH2movlsg(sh,(15<<4)+8);
}
#define init_12_2_15_9
void FASTCALL call_12_2_15_9(SH2_struct* sh)
{
  SH2movlsg(sh,(15<<4)+9);
}
#define init_12_2_15_10
void FASTCALL call_12_2_15_10(SH2_struct* sh)
{
  SH2movlsg(sh,(15<<4)+10);
}
#define init_12_2_15_11
void FASTCALL call_12_2_15_11(SH2_struct* sh)
{
  SH2movlsg(sh,(15<<4)+11);
}
#define init_12_2_15_12
void FASTCALL call_12_2_15_12(SH2_struct* sh)
{
  SH2movlsg(sh,(15<<4)+12);
}
#define init_12_2_15_13
void FASTCALL call_12_2_15_13(SH2_struct* sh)
{
  SH2movlsg(sh,(15<<4)+13);
}
#define init_12_2_15_14
void FASTCALL call_12_2_15_14(SH2_struct* sh)
{
  SH2movlsg(sh,(15<<4)+14);
}
#define init_12_2_15_15
void FASTCALL call_12_2_15_15(SH2_struct* sh)
{
  SH2movlsg(sh,(15<<4)+15);
}
#define init_1_0_0_0
void FASTCALL call_1_0_0_0(SH2_struct* sh)
{
  SH2movls4(sh, 0,0, 0);
}
#define init_1_0_0_1
void FASTCALL call_1_0_0_1(SH2_struct* sh)
{
  SH2movls4(sh, 0,0, 1);
}
#define init_1_0_0_2
void FASTCALL call_1_0_0_2(SH2_struct* sh)
{
  SH2movls4(sh, 0,0, 2);
}
#define init_1_0_0_3
void FASTCALL call_1_0_0_3(SH2_struct* sh)
{
  SH2movls4(sh, 0,0, 3);
}
#define init_1_0_0_4
void FASTCALL call_1_0_0_4(SH2_struct* sh)
{
  SH2movls4(sh, 0,0, 4);
}
#define init_1_0_0_5
void FASTCALL call_1_0_0_5(SH2_struct* sh)
{
  SH2movls4(sh, 0,0, 5);
}
#define init_1_0_0_6
void FASTCALL call_1_0_0_6(SH2_struct* sh)
{
  SH2movls4(sh, 0,0, 6);
}
#define init_1_0_0_7
void FASTCALL call_1_0_0_7(SH2_struct* sh)
{
  SH2movls4(sh, 0,0, 7);
}
#define init_1_0_0_8
void FASTCALL call_1_0_0_8(SH2_struct* sh)
{
  SH2movls4(sh, 0,0, 8);
}
#define init_1_0_0_9
void FASTCALL call_1_0_0_9(SH2_struct* sh)
{
  SH2movls4(sh, 0,0, 9);
}
#define init_1_0_0_10
void FASTCALL call_1_0_0_10(SH2_struct* sh)
{
  SH2movls4(sh, 0,0, 10);
}
#define init_1_0_0_11
void FASTCALL call_1_0_0_11(SH2_struct* sh)
{
  SH2movls4(sh, 0,0, 11);
}
#define init_1_0_0_12
void FASTCALL call_1_0_0_12(SH2_struct* sh)
{
  SH2movls4(sh, 0,0, 12);
}
#define init_1_0_0_13
void FASTCALL call_1_0_0_13(SH2_struct* sh)
{
  SH2movls4(sh, 0,0, 13);
}
#define init_1_0_0_14
void FASTCALL call_1_0_0_14(SH2_struct* sh)
{
  SH2movls4(sh, 0,0, 14);
}
#define init_1_0_0_15
void FASTCALL call_1_0_0_15(SH2_struct* sh)
{
  SH2movls4(sh, 0,0, 15);
}
#define init_1_0_1_0
void FASTCALL call_1_0_1_0(SH2_struct* sh)
{
  SH2movls4(sh, 0,1, 0);
}
#define init_1_0_1_1
void FASTCALL call_1_0_1_1(SH2_struct* sh)
{
  SH2movls4(sh, 0,1, 1);
}
#define init_1_0_1_2
void FASTCALL call_1_0_1_2(SH2_struct* sh)
{
  SH2movls4(sh, 0,1, 2);
}
#define init_1_0_1_3
void FASTCALL call_1_0_1_3(SH2_struct* sh)
{
  SH2movls4(sh, 0,1, 3);
}
#define init_1_0_1_4
void FASTCALL call_1_0_1_4(SH2_struct* sh)
{
  SH2movls4(sh, 0,1, 4);
}
#define init_1_0_1_5
void FASTCALL call_1_0_1_5(SH2_struct* sh)
{
  SH2movls4(sh, 0,1, 5);
}
#define init_1_0_1_6
void FASTCALL call_1_0_1_6(SH2_struct* sh)
{
  SH2movls4(sh, 0,1, 6);
}
#define init_1_0_1_7
void FASTCALL call_1_0_1_7(SH2_struct* sh)
{
  SH2movls4(sh, 0,1, 7);
}
#define init_1_0_1_8
void FASTCALL call_1_0_1_8(SH2_struct* sh)
{
  SH2movls4(sh, 0,1, 8);
}
#define init_1_0_1_9
void FASTCALL call_1_0_1_9(SH2_struct* sh)
{
  SH2movls4(sh, 0,1, 9);
}
#define init_1_0_1_10
void FASTCALL call_1_0_1_10(SH2_struct* sh)
{
  SH2movls4(sh, 0,1, 10);
}
#define init_1_0_1_11
void FASTCALL call_1_0_1_11(SH2_struct* sh)
{
  SH2movls4(sh, 0,1, 11);
}
#define init_1_0_1_12
void FASTCALL call_1_0_1_12(SH2_struct* sh)
{
  SH2movls4(sh, 0,1, 12);
}
#define init_1_0_1_13
void FASTCALL call_1_0_1_13(SH2_struct* sh)
{
  SH2movls4(sh, 0,1, 13);
}
#define init_1_0_1_14
void FASTCALL call_1_0_1_14(SH2_struct* sh)
{
  SH2movls4(sh, 0,1, 14);
}
#define init_1_0_1_15
void FASTCALL call_1_0_1_15(SH2_struct* sh)
{
  SH2movls4(sh, 0,1, 15);
}
#define init_1_0_2_0
void FASTCALL call_1_0_2_0(SH2_struct* sh)
{
  SH2movls4(sh, 0,2, 0);
}
#define init_1_0_2_1
void FASTCALL call_1_0_2_1(SH2_struct* sh)
{
  SH2movls4(sh, 0,2, 1);
}
#define init_1_0_2_2
void FASTCALL call_1_0_2_2(SH2_struct* sh)
{
  SH2movls4(sh, 0,2, 2);
}
#define init_1_0_2_3
void FASTCALL call_1_0_2_3(SH2_struct* sh)
{
  SH2movls4(sh, 0,2, 3);
}
#define init_1_0_2_4
void FASTCALL call_1_0_2_4(SH2_struct* sh)
{
  SH2movls4(sh, 0,2, 4);
}
#define init_1_0_2_5
void FASTCALL call_1_0_2_5(SH2_struct* sh)
{
  SH2movls4(sh, 0,2, 5);
}
#define init_1_0_2_6
void FASTCALL call_1_0_2_6(SH2_struct* sh)
{
  SH2movls4(sh, 0,2, 6);
}
#define init_1_0_2_7
void FASTCALL call_1_0_2_7(SH2_struct* sh)
{
  SH2movls4(sh, 0,2, 7);
}
#define init_1_0_2_8
void FASTCALL call_1_0_2_8(SH2_struct* sh)
{
  SH2movls4(sh, 0,2, 8);
}
#define init_1_0_2_9
void FASTCALL call_1_0_2_9(SH2_struct* sh)
{
  SH2movls4(sh, 0,2, 9);
}
#define init_1_0_2_10
void FASTCALL call_1_0_2_10(SH2_struct* sh)
{
  SH2movls4(sh, 0,2, 10);
}
#define init_1_0_2_11
void FASTCALL call_1_0_2_11(SH2_struct* sh)
{
  SH2movls4(sh, 0,2, 11);
}
#define init_1_0_2_12
void FASTCALL call_1_0_2_12(SH2_struct* sh)
{
  SH2movls4(sh, 0,2, 12);
}
#define init_1_0_2_13
void FASTCALL call_1_0_2_13(SH2_struct* sh)
{
  SH2movls4(sh, 0,2, 13);
}
#define init_1_0_2_14
void FASTCALL call_1_0_2_14(SH2_struct* sh)
{
  SH2movls4(sh, 0,2, 14);
}
#define init_1_0_2_15
void FASTCALL call_1_0_2_15(SH2_struct* sh)
{
  SH2movls4(sh, 0,2, 15);
}
#define init_1_0_3_0
void FASTCALL call_1_0_3_0(SH2_struct* sh)
{
  SH2movls4(sh, 0,3, 0);
}
#define init_1_0_3_1
void FASTCALL call_1_0_3_1(SH2_struct* sh)
{
  SH2movls4(sh, 0,3, 1);
}
#define init_1_0_3_2
void FASTCALL call_1_0_3_2(SH2_struct* sh)
{
  SH2movls4(sh, 0,3, 2);
}
#define init_1_0_3_3
void FASTCALL call_1_0_3_3(SH2_struct* sh)
{
  SH2movls4(sh, 0,3, 3);
}
#define init_1_0_3_4
void FASTCALL call_1_0_3_4(SH2_struct* sh)
{
  SH2movls4(sh, 0,3, 4);
}
#define init_1_0_3_5
void FASTCALL call_1_0_3_5(SH2_struct* sh)
{
  SH2movls4(sh, 0,3, 5);
}
#define init_1_0_3_6
void FASTCALL call_1_0_3_6(SH2_struct* sh)
{
  SH2movls4(sh, 0,3, 6);
}
#define init_1_0_3_7
void FASTCALL call_1_0_3_7(SH2_struct* sh)
{
  SH2movls4(sh, 0,3, 7);
}
#define init_1_0_3_8
void FASTCALL call_1_0_3_8(SH2_struct* sh)
{
  SH2movls4(sh, 0,3, 8);
}
#define init_1_0_3_9
void FASTCALL call_1_0_3_9(SH2_struct* sh)
{
  SH2movls4(sh, 0,3, 9);
}
#define init_1_0_3_10
void FASTCALL call_1_0_3_10(SH2_struct* sh)
{
  SH2movls4(sh, 0,3, 10);
}
#define init_1_0_3_11
void FASTCALL call_1_0_3_11(SH2_struct* sh)
{
  SH2movls4(sh, 0,3, 11);
}
#define init_1_0_3_12
void FASTCALL call_1_0_3_12(SH2_struct* sh)
{
  SH2movls4(sh, 0,3, 12);
}
#define init_1_0_3_13
void FASTCALL call_1_0_3_13(SH2_struct* sh)
{
  SH2movls4(sh, 0,3, 13);
}
#define init_1_0_3_14
void FASTCALL call_1_0_3_14(SH2_struct* sh)
{
  SH2movls4(sh, 0,3, 14);
}
#define init_1_0_3_15
void FASTCALL call_1_0_3_15(SH2_struct* sh)
{
  SH2movls4(sh, 0,3, 15);
}
#define init_1_0_4_0
void FASTCALL call_1_0_4_0(SH2_struct* sh)
{
  SH2movls4(sh, 0,4, 0);
}
#define init_1_0_4_1
void FASTCALL call_1_0_4_1(SH2_struct* sh)
{
  SH2movls4(sh, 0,4, 1);
}
#define init_1_0_4_2
void FASTCALL call_1_0_4_2(SH2_struct* sh)
{
  SH2movls4(sh, 0,4, 2);
}
#define init_1_0_4_3
void FASTCALL call_1_0_4_3(SH2_struct* sh)
{
  SH2movls4(sh, 0,4, 3);
}
#define init_1_0_4_4
void FASTCALL call_1_0_4_4(SH2_struct* sh)
{
  SH2movls4(sh, 0,4, 4);
}
#define init_1_0_4_5
void FASTCALL call_1_0_4_5(SH2_struct* sh)
{
  SH2movls4(sh, 0,4, 5);
}
#define init_1_0_4_6
void FASTCALL call_1_0_4_6(SH2_struct* sh)
{
  SH2movls4(sh, 0,4, 6);
}
#define init_1_0_4_7
void FASTCALL call_1_0_4_7(SH2_struct* sh)
{
  SH2movls4(sh, 0,4, 7);
}
#define init_1_0_4_8
void FASTCALL call_1_0_4_8(SH2_struct* sh)
{
  SH2movls4(sh, 0,4, 8);
}
#define init_1_0_4_9
void FASTCALL call_1_0_4_9(SH2_struct* sh)
{
  SH2movls4(sh, 0,4, 9);
}
#define init_1_0_4_10
void FASTCALL call_1_0_4_10(SH2_struct* sh)
{
  SH2movls4(sh, 0,4, 10);
}
#define init_1_0_4_11
void FASTCALL call_1_0_4_11(SH2_struct* sh)
{
  SH2movls4(sh, 0,4, 11);
}
#define init_1_0_4_12
void FASTCALL call_1_0_4_12(SH2_struct* sh)
{
  SH2movls4(sh, 0,4, 12);
}
#define init_1_0_4_13
void FASTCALL call_1_0_4_13(SH2_struct* sh)
{
  SH2movls4(sh, 0,4, 13);
}
#define init_1_0_4_14
void FASTCALL call_1_0_4_14(SH2_struct* sh)
{
  SH2movls4(sh, 0,4, 14);
}
#define init_1_0_4_15
void FASTCALL call_1_0_4_15(SH2_struct* sh)
{
  SH2movls4(sh, 0,4, 15);
}
#define init_1_0_5_0
void FASTCALL call_1_0_5_0(SH2_struct* sh)
{
  SH2movls4(sh, 0,5, 0);
}
#define init_1_0_5_1
void FASTCALL call_1_0_5_1(SH2_struct* sh)
{
  SH2movls4(sh, 0,5, 1);
}
#define init_1_0_5_2
void FASTCALL call_1_0_5_2(SH2_struct* sh)
{
  SH2movls4(sh, 0,5, 2);
}
#define init_1_0_5_3
void FASTCALL call_1_0_5_3(SH2_struct* sh)
{
  SH2movls4(sh, 0,5, 3);
}
#define init_1_0_5_4
void FASTCALL call_1_0_5_4(SH2_struct* sh)
{
  SH2movls4(sh, 0,5, 4);
}
#define init_1_0_5_5
void FASTCALL call_1_0_5_5(SH2_struct* sh)
{
  SH2movls4(sh, 0,5, 5);
}
#define init_1_0_5_6
void FASTCALL call_1_0_5_6(SH2_struct* sh)
{
  SH2movls4(sh, 0,5, 6);
}
#define init_1_0_5_7
void FASTCALL call_1_0_5_7(SH2_struct* sh)
{
  SH2movls4(sh, 0,5, 7);
}
#define init_1_0_5_8
void FASTCALL call_1_0_5_8(SH2_struct* sh)
{
  SH2movls4(sh, 0,5, 8);
}
#define init_1_0_5_9
void FASTCALL call_1_0_5_9(SH2_struct* sh)
{
  SH2movls4(sh, 0,5, 9);
}
#define init_1_0_5_10
void FASTCALL call_1_0_5_10(SH2_struct* sh)
{
  SH2movls4(sh, 0,5, 10);
}
#define init_1_0_5_11
void FASTCALL call_1_0_5_11(SH2_struct* sh)
{
  SH2movls4(sh, 0,5, 11);
}
#define init_1_0_5_12
void FASTCALL call_1_0_5_12(SH2_struct* sh)
{
  SH2movls4(sh, 0,5, 12);
}
#define init_1_0_5_13
void FASTCALL call_1_0_5_13(SH2_struct* sh)
{
  SH2movls4(sh, 0,5, 13);
}
#define init_1_0_5_14
void FASTCALL call_1_0_5_14(SH2_struct* sh)
{
  SH2movls4(sh, 0,5, 14);
}
#define init_1_0_5_15
void FASTCALL call_1_0_5_15(SH2_struct* sh)
{
  SH2movls4(sh, 0,5, 15);
}
#define init_1_0_6_0
void FASTCALL call_1_0_6_0(SH2_struct* sh)
{
  SH2movls4(sh, 0,6, 0);
}
#define init_1_0_6_1
void FASTCALL call_1_0_6_1(SH2_struct* sh)
{
  SH2movls4(sh, 0,6, 1);
}
#define init_1_0_6_2
void FASTCALL call_1_0_6_2(SH2_struct* sh)
{
  SH2movls4(sh, 0,6, 2);
}
#define init_1_0_6_3
void FASTCALL call_1_0_6_3(SH2_struct* sh)
{
  SH2movls4(sh, 0,6, 3);
}
#define init_1_0_6_4
void FASTCALL call_1_0_6_4(SH2_struct* sh)
{
  SH2movls4(sh, 0,6, 4);
}
#define init_1_0_6_5
void FASTCALL call_1_0_6_5(SH2_struct* sh)
{
  SH2movls4(sh, 0,6, 5);
}
#define init_1_0_6_6
void FASTCALL call_1_0_6_6(SH2_struct* sh)
{
  SH2movls4(sh, 0,6, 6);
}
#define init_1_0_6_7
void FASTCALL call_1_0_6_7(SH2_struct* sh)
{
  SH2movls4(sh, 0,6, 7);
}
#define init_1_0_6_8
void FASTCALL call_1_0_6_8(SH2_struct* sh)
{
  SH2movls4(sh, 0,6, 8);
}
#define init_1_0_6_9
void FASTCALL call_1_0_6_9(SH2_struct* sh)
{
  SH2movls4(sh, 0,6, 9);
}
#define init_1_0_6_10
void FASTCALL call_1_0_6_10(SH2_struct* sh)
{
  SH2movls4(sh, 0,6, 10);
}
#define init_1_0_6_11
void FASTCALL call_1_0_6_11(SH2_struct* sh)
{
  SH2movls4(sh, 0,6, 11);
}
#define init_1_0_6_12
void FASTCALL call_1_0_6_12(SH2_struct* sh)
{
  SH2movls4(sh, 0,6, 12);
}
#define init_1_0_6_13
void FASTCALL call_1_0_6_13(SH2_struct* sh)
{
  SH2movls4(sh, 0,6, 13);
}
#define init_1_0_6_14
void FASTCALL call_1_0_6_14(SH2_struct* sh)
{
  SH2movls4(sh, 0,6, 14);
}
#define init_1_0_6_15
void FASTCALL call_1_0_6_15(SH2_struct* sh)
{
  SH2movls4(sh, 0,6, 15);
}
#define init_1_0_7_0
void FASTCALL call_1_0_7_0(SH2_struct* sh)
{
  SH2movls4(sh, 0,7, 0);
}
#define init_1_0_7_1
void FASTCALL call_1_0_7_1(SH2_struct* sh)
{
  SH2movls4(sh, 0,7, 1);
}
#define init_1_0_7_2
void FASTCALL call_1_0_7_2(SH2_struct* sh)
{
  SH2movls4(sh, 0,7, 2);
}
#define init_1_0_7_3
void FASTCALL call_1_0_7_3(SH2_struct* sh)
{
  SH2movls4(sh, 0,7, 3);
}
#define init_1_0_7_4
void FASTCALL call_1_0_7_4(SH2_struct* sh)
{
  SH2movls4(sh, 0,7, 4);
}
#define init_1_0_7_5
void FASTCALL call_1_0_7_5(SH2_struct* sh)
{
  SH2movls4(sh, 0,7, 5);
}
#define init_1_0_7_6
void FASTCALL call_1_0_7_6(SH2_struct* sh)
{
  SH2movls4(sh, 0,7, 6);
}
#define init_1_0_7_7
void FASTCALL call_1_0_7_7(SH2_struct* sh)
{
  SH2movls4(sh, 0,7, 7);
}
#define init_1_0_7_8
void FASTCALL call_1_0_7_8(SH2_struct* sh)
{
  SH2movls4(sh, 0,7, 8);
}
#define init_1_0_7_9
void FASTCALL call_1_0_7_9(SH2_struct* sh)
{
  SH2movls4(sh, 0,7, 9);
}
#define init_1_0_7_10
void FASTCALL call_1_0_7_10(SH2_struct* sh)
{
  SH2movls4(sh, 0,7, 10);
}
#define init_1_0_7_11
void FASTCALL call_1_0_7_11(SH2_struct* sh)
{
  SH2movls4(sh, 0,7, 11);
}
#define init_1_0_7_12
void FASTCALL call_1_0_7_12(SH2_struct* sh)
{
  SH2movls4(sh, 0,7, 12);
}
#define init_1_0_7_13
void FASTCALL call_1_0_7_13(SH2_struct* sh)
{
  SH2movls4(sh, 0,7, 13);
}
#define init_1_0_7_14
void FASTCALL call_1_0_7_14(SH2_struct* sh)
{
  SH2movls4(sh, 0,7, 14);
}
#define init_1_0_7_15
void FASTCALL call_1_0_7_15(SH2_struct* sh)
{
  SH2movls4(sh, 0,7, 15);
}
#define init_1_0_8_0
void FASTCALL call_1_0_8_0(SH2_struct* sh)
{
  SH2movls4(sh, 0,8, 0);
}
#define init_1_0_8_1
void FASTCALL call_1_0_8_1(SH2_struct* sh)
{
  SH2movls4(sh, 0,8, 1);
}
#define init_1_0_8_2
void FASTCALL call_1_0_8_2(SH2_struct* sh)
{
  SH2movls4(sh, 0,8, 2);
}
#define init_1_0_8_3
void FASTCALL call_1_0_8_3(SH2_struct* sh)
{
  SH2movls4(sh, 0,8, 3);
}
#define init_1_0_8_4
void FASTCALL call_1_0_8_4(SH2_struct* sh)
{
  SH2movls4(sh, 0,8, 4);
}
#define init_1_0_8_5
void FASTCALL call_1_0_8_5(SH2_struct* sh)
{
  SH2movls4(sh, 0,8, 5);
}
#define init_1_0_8_6
void FASTCALL call_1_0_8_6(SH2_struct* sh)
{
  SH2movls4(sh, 0,8, 6);
}
#define init_1_0_8_7
void FASTCALL call_1_0_8_7(SH2_struct* sh)
{
  SH2movls4(sh, 0,8, 7);
}
#define init_1_0_8_8
void FASTCALL call_1_0_8_8(SH2_struct* sh)
{
  SH2movls4(sh, 0,8, 8);
}
#define init_1_0_8_9
void FASTCALL call_1_0_8_9(SH2_struct* sh)
{
  SH2movls4(sh, 0,8, 9);
}
#define init_1_0_8_10
void FASTCALL call_1_0_8_10(SH2_struct* sh)
{
  SH2movls4(sh, 0,8, 10);
}
#define init_1_0_8_11
void FASTCALL call_1_0_8_11(SH2_struct* sh)
{
  SH2movls4(sh, 0,8, 11);
}
#define init_1_0_8_12
void FASTCALL call_1_0_8_12(SH2_struct* sh)
{
  SH2movls4(sh, 0,8, 12);
}
#define init_1_0_8_13
void FASTCALL call_1_0_8_13(SH2_struct* sh)
{
  SH2movls4(sh, 0,8, 13);
}
#define init_1_0_8_14
void FASTCALL call_1_0_8_14(SH2_struct* sh)
{
  SH2movls4(sh, 0,8, 14);
}
#define init_1_0_8_15
void FASTCALL call_1_0_8_15(SH2_struct* sh)
{
  SH2movls4(sh, 0,8, 15);
}
#define init_1_0_9_0
void FASTCALL call_1_0_9_0(SH2_struct* sh)
{
  SH2movls4(sh, 0,9, 0);
}
#define init_1_0_9_1
void FASTCALL call_1_0_9_1(SH2_struct* sh)
{
  SH2movls4(sh, 0,9, 1);
}
#define init_1_0_9_2
void FASTCALL call_1_0_9_2(SH2_struct* sh)
{
  SH2movls4(sh, 0,9, 2);
}
#define init_1_0_9_3
void FASTCALL call_1_0_9_3(SH2_struct* sh)
{
  SH2movls4(sh, 0,9, 3);
}
#define init_1_0_9_4
void FASTCALL call_1_0_9_4(SH2_struct* sh)
{
  SH2movls4(sh, 0,9, 4);
}
#define init_1_0_9_5
void FASTCALL call_1_0_9_5(SH2_struct* sh)
{
  SH2movls4(sh, 0,9, 5);
}
#define init_1_0_9_6
void FASTCALL call_1_0_9_6(SH2_struct* sh)
{
  SH2movls4(sh, 0,9, 6);
}
#define init_1_0_9_7
void FASTCALL call_1_0_9_7(SH2_struct* sh)
{
  SH2movls4(sh, 0,9, 7);
}
#define init_1_0_9_8
void FASTCALL call_1_0_9_8(SH2_struct* sh)
{
  SH2movls4(sh, 0,9, 8);
}
#define init_1_0_9_9
void FASTCALL call_1_0_9_9(SH2_struct* sh)
{
  SH2movls4(sh, 0,9, 9);
}
#define init_1_0_9_10
void FASTCALL call_1_0_9_10(SH2_struct* sh)
{
  SH2movls4(sh, 0,9, 10);
}
#define init_1_0_9_11
void FASTCALL call_1_0_9_11(SH2_struct* sh)
{
  SH2movls4(sh, 0,9, 11);
}
#define init_1_0_9_12
void FASTCALL call_1_0_9_12(SH2_struct* sh)
{
  SH2movls4(sh, 0,9, 12);
}
#define init_1_0_9_13
void FASTCALL call_1_0_9_13(SH2_struct* sh)
{
  SH2movls4(sh, 0,9, 13);
}
#define init_1_0_9_14
void FASTCALL call_1_0_9_14(SH2_struct* sh)
{
  SH2movls4(sh, 0,9, 14);
}
#define init_1_0_9_15
void FASTCALL call_1_0_9_15(SH2_struct* sh)
{
  SH2movls4(sh, 0,9, 15);
}
#define init_1_0_10_0
void FASTCALL call_1_0_10_0(SH2_struct* sh)
{
  SH2movls4(sh, 0,10, 0);
}
#define init_1_0_10_1
void FASTCALL call_1_0_10_1(SH2_struct* sh)
{
  SH2movls4(sh, 0,10, 1);
}
#define init_1_0_10_2
void FASTCALL call_1_0_10_2(SH2_struct* sh)
{
  SH2movls4(sh, 0,10, 2);
}
#define init_1_0_10_3
void FASTCALL call_1_0_10_3(SH2_struct* sh)
{
  SH2movls4(sh, 0,10, 3);
}
#define init_1_0_10_4
void FASTCALL call_1_0_10_4(SH2_struct* sh)
{
  SH2movls4(sh, 0,10, 4);
}
#define init_1_0_10_5
void FASTCALL call_1_0_10_5(SH2_struct* sh)
{
  SH2movls4(sh, 0,10, 5);
}
#define init_1_0_10_6
void FASTCALL call_1_0_10_6(SH2_struct* sh)
{
  SH2movls4(sh, 0,10, 6);
}
#define init_1_0_10_7
void FASTCALL call_1_0_10_7(SH2_struct* sh)
{
  SH2movls4(sh, 0,10, 7);
}
#define init_1_0_10_8
void FASTCALL call_1_0_10_8(SH2_struct* sh)
{
  SH2movls4(sh, 0,10, 8);
}
#define init_1_0_10_9
void FASTCALL call_1_0_10_9(SH2_struct* sh)
{
  SH2movls4(sh, 0,10, 9);
}
#define init_1_0_10_10
void FASTCALL call_1_0_10_10(SH2_struct* sh)
{
  SH2movls4(sh, 0,10, 10);
}
#define init_1_0_10_11
void FASTCALL call_1_0_10_11(SH2_struct* sh)
{
  SH2movls4(sh, 0,10, 11);
}
#define init_1_0_10_12
void FASTCALL call_1_0_10_12(SH2_struct* sh)
{
  SH2movls4(sh, 0,10, 12);
}
#define init_1_0_10_13
void FASTCALL call_1_0_10_13(SH2_struct* sh)
{
  SH2movls4(sh, 0,10, 13);
}
#define init_1_0_10_14
void FASTCALL call_1_0_10_14(SH2_struct* sh)
{
  SH2movls4(sh, 0,10, 14);
}
#define init_1_0_10_15
void FASTCALL call_1_0_10_15(SH2_struct* sh)
{
  SH2movls4(sh, 0,10, 15);
}
#define init_1_0_11_0
void FASTCALL call_1_0_11_0(SH2_struct* sh)
{
  SH2movls4(sh, 0,11, 0);
}
#define init_1_0_11_1
void FASTCALL call_1_0_11_1(SH2_struct* sh)
{
  SH2movls4(sh, 0,11, 1);
}
#define init_1_0_11_2
void FASTCALL call_1_0_11_2(SH2_struct* sh)
{
  SH2movls4(sh, 0,11, 2);
}
#define init_1_0_11_3
void FASTCALL call_1_0_11_3(SH2_struct* sh)
{
  SH2movls4(sh, 0,11, 3);
}
#define init_1_0_11_4
void FASTCALL call_1_0_11_4(SH2_struct* sh)
{
  SH2movls4(sh, 0,11, 4);
}
#define init_1_0_11_5
void FASTCALL call_1_0_11_5(SH2_struct* sh)
{
  SH2movls4(sh, 0,11, 5);
}
#define init_1_0_11_6
void FASTCALL call_1_0_11_6(SH2_struct* sh)
{
  SH2movls4(sh, 0,11, 6);
}
#define init_1_0_11_7
void FASTCALL call_1_0_11_7(SH2_struct* sh)
{
  SH2movls4(sh, 0,11, 7);
}
#define init_1_0_11_8
void FASTCALL call_1_0_11_8(SH2_struct* sh)
{
  SH2movls4(sh, 0,11, 8);
}
#define init_1_0_11_9
void FASTCALL call_1_0_11_9(SH2_struct* sh)
{
  SH2movls4(sh, 0,11, 9);
}
#define init_1_0_11_10
void FASTCALL call_1_0_11_10(SH2_struct* sh)
{
  SH2movls4(sh, 0,11, 10);
}
#define init_1_0_11_11
void FASTCALL call_1_0_11_11(SH2_struct* sh)
{
  SH2movls4(sh, 0,11, 11);
}
#define init_1_0_11_12
void FASTCALL call_1_0_11_12(SH2_struct* sh)
{
  SH2movls4(sh, 0,11, 12);
}
#define init_1_0_11_13
void FASTCALL call_1_0_11_13(SH2_struct* sh)
{
  SH2movls4(sh, 0,11, 13);
}
#define init_1_0_11_14
void FASTCALL call_1_0_11_14(SH2_struct* sh)
{
  SH2movls4(sh, 0,11, 14);
}
#define init_1_0_11_15
void FASTCALL call_1_0_11_15(SH2_struct* sh)
{
  SH2movls4(sh, 0,11, 15);
}
#define init_1_0_12_0
void FASTCALL call_1_0_12_0(SH2_struct* sh)
{
  SH2movls4(sh, 0,12, 0);
}
#define init_1_0_12_1
void FASTCALL call_1_0_12_1(SH2_struct* sh)
{
  SH2movls4(sh, 0,12, 1);
}
#define init_1_0_12_2
void FASTCALL call_1_0_12_2(SH2_struct* sh)
{
  SH2movls4(sh, 0,12, 2);
}
#define init_1_0_12_3
void FASTCALL call_1_0_12_3(SH2_struct* sh)
{
  SH2movls4(sh, 0,12, 3);
}
#define init_1_0_12_4
void FASTCALL call_1_0_12_4(SH2_struct* sh)
{
  SH2movls4(sh, 0,12, 4);
}
#define init_1_0_12_5
void FASTCALL call_1_0_12_5(SH2_struct* sh)
{
  SH2movls4(sh, 0,12, 5);
}
#define init_1_0_12_6
void FASTCALL call_1_0_12_6(SH2_struct* sh)
{
  SH2movls4(sh, 0,12, 6);
}
#define init_1_0_12_7
void FASTCALL call_1_0_12_7(SH2_struct* sh)
{
  SH2movls4(sh, 0,12, 7);
}
#define init_1_0_12_8
void FASTCALL call_1_0_12_8(SH2_struct* sh)
{
  SH2movls4(sh, 0,12, 8);
}
#define init_1_0_12_9
void FASTCALL call_1_0_12_9(SH2_struct* sh)
{
  SH2movls4(sh, 0,12, 9);
}
#define init_1_0_12_10
void FASTCALL call_1_0_12_10(SH2_struct* sh)
{
  SH2movls4(sh, 0,12, 10);
}
#define init_1_0_12_11
void FASTCALL call_1_0_12_11(SH2_struct* sh)
{
  SH2movls4(sh, 0,12, 11);
}
#define init_1_0_12_12
void FASTCALL call_1_0_12_12(SH2_struct* sh)
{
  SH2movls4(sh, 0,12, 12);
}
#define init_1_0_12_13
void FASTCALL call_1_0_12_13(SH2_struct* sh)
{
  SH2movls4(sh, 0,12, 13);
}
#define init_1_0_12_14
void FASTCALL call_1_0_12_14(SH2_struct* sh)
{
  SH2movls4(sh, 0,12, 14);
}
#define init_1_0_12_15
void FASTCALL call_1_0_12_15(SH2_struct* sh)
{
  SH2movls4(sh, 0,12, 15);
}
#define init_1_0_13_0
void FASTCALL call_1_0_13_0(SH2_struct* sh)
{
  SH2movls4(sh, 0,13, 0);
}
#define init_1_0_13_1
void FASTCALL call_1_0_13_1(SH2_struct* sh)
{
  SH2movls4(sh, 0,13, 1);
}
#define init_1_0_13_2
void FASTCALL call_1_0_13_2(SH2_struct* sh)
{
  SH2movls4(sh, 0,13, 2);
}
#define init_1_0_13_3
void FASTCALL call_1_0_13_3(SH2_struct* sh)
{
  SH2movls4(sh, 0,13, 3);
}
#define init_1_0_13_4
void FASTCALL call_1_0_13_4(SH2_struct* sh)
{
  SH2movls4(sh, 0,13, 4);
}
#define init_1_0_13_5
void FASTCALL call_1_0_13_5(SH2_struct* sh)
{
  SH2movls4(sh, 0,13, 5);
}
#define init_1_0_13_6
void FASTCALL call_1_0_13_6(SH2_struct* sh)
{
  SH2movls4(sh, 0,13, 6);
}
#define init_1_0_13_7
void FASTCALL call_1_0_13_7(SH2_struct* sh)
{
  SH2movls4(sh, 0,13, 7);
}
#define init_1_0_13_8
void FASTCALL call_1_0_13_8(SH2_struct* sh)
{
  SH2movls4(sh, 0,13, 8);
}
#define init_1_0_13_9
void FASTCALL call_1_0_13_9(SH2_struct* sh)
{
  SH2movls4(sh, 0,13, 9);
}
#define init_1_0_13_10
void FASTCALL call_1_0_13_10(SH2_struct* sh)
{
  SH2movls4(sh, 0,13, 10);
}
#define init_1_0_13_11
void FASTCALL call_1_0_13_11(SH2_struct* sh)
{
  SH2movls4(sh, 0,13, 11);
}
#define init_1_0_13_12
void FASTCALL call_1_0_13_12(SH2_struct* sh)
{
  SH2movls4(sh, 0,13, 12);
}
#define init_1_0_13_13
void FASTCALL call_1_0_13_13(SH2_struct* sh)
{
  SH2movls4(sh, 0,13, 13);
}
#define init_1_0_13_14
void FASTCALL call_1_0_13_14(SH2_struct* sh)
{
  SH2movls4(sh, 0,13, 14);
}
#define init_1_0_13_15
void FASTCALL call_1_0_13_15(SH2_struct* sh)
{
  SH2movls4(sh, 0,13, 15);
}
#define init_1_0_14_0
void FASTCALL call_1_0_14_0(SH2_struct* sh)
{
  SH2movls4(sh, 0,14, 0);
}
#define init_1_0_14_1
void FASTCALL call_1_0_14_1(SH2_struct* sh)
{
  SH2movls4(sh, 0,14, 1);
}
#define init_1_0_14_2
void FASTCALL call_1_0_14_2(SH2_struct* sh)
{
  SH2movls4(sh, 0,14, 2);
}
#define init_1_0_14_3
void FASTCALL call_1_0_14_3(SH2_struct* sh)
{
  SH2movls4(sh, 0,14, 3);
}
#define init_1_0_14_4
void FASTCALL call_1_0_14_4(SH2_struct* sh)
{
  SH2movls4(sh, 0,14, 4);
}
#define init_1_0_14_5
void FASTCALL call_1_0_14_5(SH2_struct* sh)
{
  SH2movls4(sh, 0,14, 5);
}
#define init_1_0_14_6
void FASTCALL call_1_0_14_6(SH2_struct* sh)
{
  SH2movls4(sh, 0,14, 6);
}
#define init_1_0_14_7
void FASTCALL call_1_0_14_7(SH2_struct* sh)
{
  SH2movls4(sh, 0,14, 7);
}
#define init_1_0_14_8
void FASTCALL call_1_0_14_8(SH2_struct* sh)
{
  SH2movls4(sh, 0,14, 8);
}
#define init_1_0_14_9
void FASTCALL call_1_0_14_9(SH2_struct* sh)
{
  SH2movls4(sh, 0,14, 9);
}
#define init_1_0_14_10
void FASTCALL call_1_0_14_10(SH2_struct* sh)
{
  SH2movls4(sh, 0,14, 10);
}
#define init_1_0_14_11
void FASTCALL call_1_0_14_11(SH2_struct* sh)
{
  SH2movls4(sh, 0,14, 11);
}
#define init_1_0_14_12
void FASTCALL call_1_0_14_12(SH2_struct* sh)
{
  SH2movls4(sh, 0,14, 12);
}
#define init_1_0_14_13
void FASTCALL call_1_0_14_13(SH2_struct* sh)
{
  SH2movls4(sh, 0,14, 13);
}
#define init_1_0_14_14
void FASTCALL call_1_0_14_14(SH2_struct* sh)
{
  SH2movls4(sh, 0,14, 14);
}
#define init_1_0_14_15
void FASTCALL call_1_0_14_15(SH2_struct* sh)
{
  SH2movls4(sh, 0,14, 15);
}
#define init_1_0_15_0
void FASTCALL call_1_0_15_0(SH2_struct* sh)
{
  SH2movls4(sh, 0,15, 0);
}
#define init_1_0_15_1
void FASTCALL call_1_0_15_1(SH2_struct* sh)
{
  SH2movls4(sh, 0,15, 1);
}
#define init_1_0_15_2
void FASTCALL call_1_0_15_2(SH2_struct* sh)
{
  SH2movls4(sh, 0,15, 2);
}
#define init_1_0_15_3
void FASTCALL call_1_0_15_3(SH2_struct* sh)
{
  SH2movls4(sh, 0,15, 3);
}
#define init_1_0_15_4
void FASTCALL call_1_0_15_4(SH2_struct* sh)
{
  SH2movls4(sh, 0,15, 4);
}
#define init_1_0_15_5
void FASTCALL call_1_0_15_5(SH2_struct* sh)
{
  SH2movls4(sh, 0,15, 5);
}
#define init_1_0_15_6
void FASTCALL call_1_0_15_6(SH2_struct* sh)
{
  SH2movls4(sh, 0,15, 6);
}
#define init_1_0_15_7
void FASTCALL call_1_0_15_7(SH2_struct* sh)
{
  SH2movls4(sh, 0,15, 7);
}
#define init_1_0_15_8
void FASTCALL call_1_0_15_8(SH2_struct* sh)
{
  SH2movls4(sh, 0,15, 8);
}
#define init_1_0_15_9
void FASTCALL call_1_0_15_9(SH2_struct* sh)
{
  SH2movls4(sh, 0,15, 9);
}
#define init_1_0_15_10
void FASTCALL call_1_0_15_10(SH2_struct* sh)
{
  SH2movls4(sh, 0,15, 10);
}
#define init_1_0_15_11
void FASTCALL call_1_0_15_11(SH2_struct* sh)
{
  SH2movls4(sh, 0,15, 11);
}
#define init_1_0_15_12
void FASTCALL call_1_0_15_12(SH2_struct* sh)
{
  SH2movls4(sh, 0,15, 12);
}
#define init_1_0_15_13
void FASTCALL call_1_0_15_13(SH2_struct* sh)
{
  SH2movls4(sh, 0,15, 13);
}
#define init_1_0_15_14
void FASTCALL call_1_0_15_14(SH2_struct* sh)
{
  SH2movls4(sh, 0,15, 14);
}
#define init_1_0_15_15
void FASTCALL call_1_0_15_15(SH2_struct* sh)
{
  SH2movls4(sh, 0,15, 15);
}
#define init_1_1_0_0
void FASTCALL call_1_1_0_0(SH2_struct* sh)
{
  SH2movls4(sh, 1,0, 0);
}
#define init_1_1_0_1
void FASTCALL call_1_1_0_1(SH2_struct* sh)
{
  SH2movls4(sh, 1,0, 1);
}
#define init_1_1_0_2
void FASTCALL call_1_1_0_2(SH2_struct* sh)
{
  SH2movls4(sh, 1,0, 2);
}
#define init_1_1_0_3
void FASTCALL call_1_1_0_3(SH2_struct* sh)
{
  SH2movls4(sh, 1,0, 3);
}
#define init_1_1_0_4
void FASTCALL call_1_1_0_4(SH2_struct* sh)
{
  SH2movls4(sh, 1,0, 4);
}
#define init_1_1_0_5
void FASTCALL call_1_1_0_5(SH2_struct* sh)
{
  SH2movls4(sh, 1,0, 5);
}
#define init_1_1_0_6
void FASTCALL call_1_1_0_6(SH2_struct* sh)
{
  SH2movls4(sh, 1,0, 6);
}
#define init_1_1_0_7
void FASTCALL call_1_1_0_7(SH2_struct* sh)
{
  SH2movls4(sh, 1,0, 7);
}
#define init_1_1_0_8
void FASTCALL call_1_1_0_8(SH2_struct* sh)
{
  SH2movls4(sh, 1,0, 8);
}
#define init_1_1_0_9
void FASTCALL call_1_1_0_9(SH2_struct* sh)
{
  SH2movls4(sh, 1,0, 9);
}
#define init_1_1_0_10
void FASTCALL call_1_1_0_10(SH2_struct* sh)
{
  SH2movls4(sh, 1,0, 10);
}
#define init_1_1_0_11
void FASTCALL call_1_1_0_11(SH2_struct* sh)
{
  SH2movls4(sh, 1,0, 11);
}
#define init_1_1_0_12
void FASTCALL call_1_1_0_12(SH2_struct* sh)
{
  SH2movls4(sh, 1,0, 12);
}
#define init_1_1_0_13
void FASTCALL call_1_1_0_13(SH2_struct* sh)
{
  SH2movls4(sh, 1,0, 13);
}
#define init_1_1_0_14
void FASTCALL call_1_1_0_14(SH2_struct* sh)
{
  SH2movls4(sh, 1,0, 14);
}
#define init_1_1_0_15
void FASTCALL call_1_1_0_15(SH2_struct* sh)
{
  SH2movls4(sh, 1,0, 15);
}
#define init_1_1_1_0
void FASTCALL call_1_1_1_0(SH2_struct* sh)
{
  SH2movls4(sh, 1,1, 0);
}
#define init_1_1_1_1
void FASTCALL call_1_1_1_1(SH2_struct* sh)
{
  SH2movls4(sh, 1,1, 1);
}
#define init_1_1_1_2
void FASTCALL call_1_1_1_2(SH2_struct* sh)
{
  SH2movls4(sh, 1,1, 2);
}
#define init_1_1_1_3
void FASTCALL call_1_1_1_3(SH2_struct* sh)
{
  SH2movls4(sh, 1,1, 3);
}
#define init_1_1_1_4
void FASTCALL call_1_1_1_4(SH2_struct* sh)
{
  SH2movls4(sh, 1,1, 4);
}
#define init_1_1_1_5
void FASTCALL call_1_1_1_5(SH2_struct* sh)
{
  SH2movls4(sh, 1,1, 5);
}
#define init_1_1_1_6
void FASTCALL call_1_1_1_6(SH2_struct* sh)
{
  SH2movls4(sh, 1,1, 6);
}
#define init_1_1_1_7
void FASTCALL call_1_1_1_7(SH2_struct* sh)
{
  SH2movls4(sh, 1,1, 7);
}
#define init_1_1_1_8
void FASTCALL call_1_1_1_8(SH2_struct* sh)
{
  SH2movls4(sh, 1,1, 8);
}
#define init_1_1_1_9
void FASTCALL call_1_1_1_9(SH2_struct* sh)
{
  SH2movls4(sh, 1,1, 9);
}
#define init_1_1_1_10
void FASTCALL call_1_1_1_10(SH2_struct* sh)
{
  SH2movls4(sh, 1,1, 10);
}
#define init_1_1_1_11
void FASTCALL call_1_1_1_11(SH2_struct* sh)
{
  SH2movls4(sh, 1,1, 11);
}
#define init_1_1_1_12
void FASTCALL call_1_1_1_12(SH2_struct* sh)
{
  SH2movls4(sh, 1,1, 12);
}
#define init_1_1_1_13
void FASTCALL call_1_1_1_13(SH2_struct* sh)
{
  SH2movls4(sh, 1,1, 13);
}
#define init_1_1_1_14
void FASTCALL call_1_1_1_14(SH2_struct* sh)
{
  SH2movls4(sh, 1,1, 14);
}
#define init_1_1_1_15
void FASTCALL call_1_1_1_15(SH2_struct* sh)
{
  SH2movls4(sh, 1,1, 15);
}
#define init_1_1_2_0
void FASTCALL call_1_1_2_0(SH2_struct* sh)
{
  SH2movls4(sh, 1,2, 0);
}
#define init_1_1_2_1
void FASTCALL call_1_1_2_1(SH2_struct* sh)
{
  SH2movls4(sh, 1,2, 1);
}
#define init_1_1_2_2
void FASTCALL call_1_1_2_2(SH2_struct* sh)
{
  SH2movls4(sh, 1,2, 2);
}
#define init_1_1_2_3
void FASTCALL call_1_1_2_3(SH2_struct* sh)
{
  SH2movls4(sh, 1,2, 3);
}
#define init_1_1_2_4
void FASTCALL call_1_1_2_4(SH2_struct* sh)
{
  SH2movls4(sh, 1,2, 4);
}
#define init_1_1_2_5
void FASTCALL call_1_1_2_5(SH2_struct* sh)
{
  SH2movls4(sh, 1,2, 5);
}
#define init_1_1_2_6
void FASTCALL call_1_1_2_6(SH2_struct* sh)
{
  SH2movls4(sh, 1,2, 6);
}
#define init_1_1_2_7
void FASTCALL call_1_1_2_7(SH2_struct* sh)
{
  SH2movls4(sh, 1,2, 7);
}
#define init_1_1_2_8
void FASTCALL call_1_1_2_8(SH2_struct* sh)
{
  SH2movls4(sh, 1,2, 8);
}
#define init_1_1_2_9
void FASTCALL call_1_1_2_9(SH2_struct* sh)
{
  SH2movls4(sh, 1,2, 9);
}
#define init_1_1_2_10
void FASTCALL call_1_1_2_10(SH2_struct* sh)
{
  SH2movls4(sh, 1,2, 10);
}
#define init_1_1_2_11
void FASTCALL call_1_1_2_11(SH2_struct* sh)
{
  SH2movls4(sh, 1,2, 11);
}
#define init_1_1_2_12
void FASTCALL call_1_1_2_12(SH2_struct* sh)
{
  SH2movls4(sh, 1,2, 12);
}
#define init_1_1_2_13
void FASTCALL call_1_1_2_13(SH2_struct* sh)
{
  SH2movls4(sh, 1,2, 13);
}
#define init_1_1_2_14
void FASTCALL call_1_1_2_14(SH2_struct* sh)
{
  SH2movls4(sh, 1,2, 14);
}
#define init_1_1_2_15
void FASTCALL call_1_1_2_15(SH2_struct* sh)
{
  SH2movls4(sh, 1,2, 15);
}
#define init_1_1_3_0
void FASTCALL call_1_1_3_0(SH2_struct* sh)
{
  SH2movls4(sh, 1,3, 0);
}
#define init_1_1_3_1
void FASTCALL call_1_1_3_1(SH2_struct* sh)
{
  SH2movls4(sh, 1,3, 1);
}
#define init_1_1_3_2
void FASTCALL call_1_1_3_2(SH2_struct* sh)
{
  SH2movls4(sh, 1,3, 2);
}
#define init_1_1_3_3
void FASTCALL call_1_1_3_3(SH2_struct* sh)
{
  SH2movls4(sh, 1,3, 3);
}
#define init_1_1_3_4
void FASTCALL call_1_1_3_4(SH2_struct* sh)
{
  SH2movls4(sh, 1,3, 4);
}
#define init_1_1_3_5
void FASTCALL call_1_1_3_5(SH2_struct* sh)
{
  SH2movls4(sh, 1,3, 5);
}
#define init_1_1_3_6
void FASTCALL call_1_1_3_6(SH2_struct* sh)
{
  SH2movls4(sh, 1,3, 6);
}
#define init_1_1_3_7
void FASTCALL call_1_1_3_7(SH2_struct* sh)
{
  SH2movls4(sh, 1,3, 7);
}
#define init_1_1_3_8
void FASTCALL call_1_1_3_8(SH2_struct* sh)
{
  SH2movls4(sh, 1,3, 8);
}
#define init_1_1_3_9
void FASTCALL call_1_1_3_9(SH2_struct* sh)
{
  SH2movls4(sh, 1,3, 9);
}
#define init_1_1_3_10
void FASTCALL call_1_1_3_10(SH2_struct* sh)
{
  SH2movls4(sh, 1,3, 10);
}
#define init_1_1_3_11
void FASTCALL call_1_1_3_11(SH2_struct* sh)
{
  SH2movls4(sh, 1,3, 11);
}
#define init_1_1_3_12
void FASTCALL call_1_1_3_12(SH2_struct* sh)
{
  SH2movls4(sh, 1,3, 12);
}
#define init_1_1_3_13
void FASTCALL call_1_1_3_13(SH2_struct* sh)
{
  SH2movls4(sh, 1,3, 13);
}
#define init_1_1_3_14
void FASTCALL call_1_1_3_14(SH2_struct* sh)
{
  SH2movls4(sh, 1,3, 14);
}
#define init_1_1_3_15
void FASTCALL call_1_1_3_15(SH2_struct* sh)
{
  SH2movls4(sh, 1,3, 15);
}
#define init_1_1_4_0
void FASTCALL call_1_1_4_0(SH2_struct* sh)
{
  SH2movls4(sh, 1,4, 0);
}
#define init_1_1_4_1
void FASTCALL call_1_1_4_1(SH2_struct* sh)
{
  SH2movls4(sh, 1,4, 1);
}
#define init_1_1_4_2
void FASTCALL call_1_1_4_2(SH2_struct* sh)
{
  SH2movls4(sh, 1,4, 2);
}
#define init_1_1_4_3
void FASTCALL call_1_1_4_3(SH2_struct* sh)
{
  SH2movls4(sh, 1,4, 3);
}
#define init_1_1_4_4
void FASTCALL call_1_1_4_4(SH2_struct* sh)
{
  SH2movls4(sh, 1,4, 4);
}
#define init_1_1_4_5
void FASTCALL call_1_1_4_5(SH2_struct* sh)
{
  SH2movls4(sh, 1,4, 5);
}
#define init_1_1_4_6
void FASTCALL call_1_1_4_6(SH2_struct* sh)
{
  SH2movls4(sh, 1,4, 6);
}
#define init_1_1_4_7
void FASTCALL call_1_1_4_7(SH2_struct* sh)
{
  SH2movls4(sh, 1,4, 7);
}
#define init_1_1_4_8
void FASTCALL call_1_1_4_8(SH2_struct* sh)
{
  SH2movls4(sh, 1,4, 8);
}
#define init_1_1_4_9
void FASTCALL call_1_1_4_9(SH2_struct* sh)
{
  SH2movls4(sh, 1,4, 9);
}
#define init_1_1_4_10
void FASTCALL call_1_1_4_10(SH2_struct* sh)
{
  SH2movls4(sh, 1,4, 10);
}
#define init_1_1_4_11
void FASTCALL call_1_1_4_11(SH2_struct* sh)
{
  SH2movls4(sh, 1,4, 11);
}
#define init_1_1_4_12
void FASTCALL call_1_1_4_12(SH2_struct* sh)
{
  SH2movls4(sh, 1,4, 12);
}
#define init_1_1_4_13
void FASTCALL call_1_1_4_13(SH2_struct* sh)
{
  SH2movls4(sh, 1,4, 13);
}
#define init_1_1_4_14
void FASTCALL call_1_1_4_14(SH2_struct* sh)
{
  SH2movls4(sh, 1,4, 14);
}
#define init_1_1_4_15
void FASTCALL call_1_1_4_15(SH2_struct* sh)
{
  SH2movls4(sh, 1,4, 15);
}
#define init_1_1_5_0
void FASTCALL call_1_1_5_0(SH2_struct* sh)
{
  SH2movls4(sh, 1,5, 0);
}
#define init_1_1_5_1
void FASTCALL call_1_1_5_1(SH2_struct* sh)
{
  SH2movls4(sh, 1,5, 1);
}
#define init_1_1_5_2
void FASTCALL call_1_1_5_2(SH2_struct* sh)
{
  SH2movls4(sh, 1,5, 2);
}
#define init_1_1_5_3
void FASTCALL call_1_1_5_3(SH2_struct* sh)
{
  SH2movls4(sh, 1,5, 3);
}
#define init_1_1_5_4
void FASTCALL call_1_1_5_4(SH2_struct* sh)
{
  SH2movls4(sh, 1,5, 4);
}
#define init_1_1_5_5
void FASTCALL call_1_1_5_5(SH2_struct* sh)
{
  SH2movls4(sh, 1,5, 5);
}
#define init_1_1_5_6
void FASTCALL call_1_1_5_6(SH2_struct* sh)
{
  SH2movls4(sh, 1,5, 6);
}
#define init_1_1_5_7
void FASTCALL call_1_1_5_7(SH2_struct* sh)
{
  SH2movls4(sh, 1,5, 7);
}
#define init_1_1_5_8
void FASTCALL call_1_1_5_8(SH2_struct* sh)
{
  SH2movls4(sh, 1,5, 8);
}
#define init_1_1_5_9
void FASTCALL call_1_1_5_9(SH2_struct* sh)
{
  SH2movls4(sh, 1,5, 9);
}
#define init_1_1_5_10
void FASTCALL call_1_1_5_10(SH2_struct* sh)
{
  SH2movls4(sh, 1,5, 10);
}
#define init_1_1_5_11
void FASTCALL call_1_1_5_11(SH2_struct* sh)
{
  SH2movls4(sh, 1,5, 11);
}
#define init_1_1_5_12
void FASTCALL call_1_1_5_12(SH2_struct* sh)
{
  SH2movls4(sh, 1,5, 12);
}
#define init_1_1_5_13
void FASTCALL call_1_1_5_13(SH2_struct* sh)
{
  SH2movls4(sh, 1,5, 13);
}
#define init_1_1_5_14
void FASTCALL call_1_1_5_14(SH2_struct* sh)
{
  SH2movls4(sh, 1,5, 14);
}
#define init_1_1_5_15
void FASTCALL call_1_1_5_15(SH2_struct* sh)
{
  SH2movls4(sh, 1,5, 15);
}
#define init_1_1_6_0
void FASTCALL call_1_1_6_0(SH2_struct* sh)
{
  SH2movls4(sh, 1,6, 0);
}
#define init_1_1_6_1
void FASTCALL call_1_1_6_1(SH2_struct* sh)
{
  SH2movls4(sh, 1,6, 1);
}
#define init_1_1_6_2
void FASTCALL call_1_1_6_2(SH2_struct* sh)
{
  SH2movls4(sh, 1,6, 2);
}
#define init_1_1_6_3
void FASTCALL call_1_1_6_3(SH2_struct* sh)
{
  SH2movls4(sh, 1,6, 3);
}
#define init_1_1_6_4
void FASTCALL call_1_1_6_4(SH2_struct* sh)
{
  SH2movls4(sh, 1,6, 4);
}
#define init_1_1_6_5
void FASTCALL call_1_1_6_5(SH2_struct* sh)
{
  SH2movls4(sh, 1,6, 5);
}
#define init_1_1_6_6
void FASTCALL call_1_1_6_6(SH2_struct* sh)
{
  SH2movls4(sh, 1,6, 6);
}
#define init_1_1_6_7
void FASTCALL call_1_1_6_7(SH2_struct* sh)
{
  SH2movls4(sh, 1,6, 7);
}
#define init_1_1_6_8
void FASTCALL call_1_1_6_8(SH2_struct* sh)
{
  SH2movls4(sh, 1,6, 8);
}
#define init_1_1_6_9
void FASTCALL call_1_1_6_9(SH2_struct* sh)
{
  SH2movls4(sh, 1,6, 9);
}
#define init_1_1_6_10
void FASTCALL call_1_1_6_10(SH2_struct* sh)
{
  SH2movls4(sh, 1,6, 10);
}
#define init_1_1_6_11
void FASTCALL call_1_1_6_11(SH2_struct* sh)
{
  SH2movls4(sh, 1,6, 11);
}
#define init_1_1_6_12
void FASTCALL call_1_1_6_12(SH2_struct* sh)
{
  SH2movls4(sh, 1,6, 12);
}
#define init_1_1_6_13
void FASTCALL call_1_1_6_13(SH2_struct* sh)
{
  SH2movls4(sh, 1,6, 13);
}
#define init_1_1_6_14
void FASTCALL call_1_1_6_14(SH2_struct* sh)
{
  SH2movls4(sh, 1,6, 14);
}
#define init_1_1_6_15
void FASTCALL call_1_1_6_15(SH2_struct* sh)
{
  SH2movls4(sh, 1,6, 15);
}
#define init_1_1_7_0
void FASTCALL call_1_1_7_0(SH2_struct* sh)
{
  SH2movls4(sh, 1,7, 0);
}
#define init_1_1_7_1
void FASTCALL call_1_1_7_1(SH2_struct* sh)
{
  SH2movls4(sh, 1,7, 1);
}
#define init_1_1_7_2
void FASTCALL call_1_1_7_2(SH2_struct* sh)
{
  SH2movls4(sh, 1,7, 2);
}
#define init_1_1_7_3
void FASTCALL call_1_1_7_3(SH2_struct* sh)
{
  SH2movls4(sh, 1,7, 3);
}
#define init_1_1_7_4
void FASTCALL call_1_1_7_4(SH2_struct* sh)
{
  SH2movls4(sh, 1,7, 4);
}
#define init_1_1_7_5
void FASTCALL call_1_1_7_5(SH2_struct* sh)
{
  SH2movls4(sh, 1,7, 5);
}
#define init_1_1_7_6
void FASTCALL call_1_1_7_6(SH2_struct* sh)
{
  SH2movls4(sh, 1,7, 6);
}
#define init_1_1_7_7
void FASTCALL call_1_1_7_7(SH2_struct* sh)
{
  SH2movls4(sh, 1,7, 7);
}
#define init_1_1_7_8
void FASTCALL call_1_1_7_8(SH2_struct* sh)
{
  SH2movls4(sh, 1,7, 8);
}
#define init_1_1_7_9
void FASTCALL call_1_1_7_9(SH2_struct* sh)
{
  SH2movls4(sh, 1,7, 9);
}
#define init_1_1_7_10
void FASTCALL call_1_1_7_10(SH2_struct* sh)
{
  SH2movls4(sh, 1,7, 10);
}
#define init_1_1_7_11
void FASTCALL call_1_1_7_11(SH2_struct* sh)
{
  SH2movls4(sh, 1,7, 11);
}
#define init_1_1_7_12
void FASTCALL call_1_1_7_12(SH2_struct* sh)
{
  SH2movls4(sh, 1,7, 12);
}
#define init_1_1_7_13
void FASTCALL call_1_1_7_13(SH2_struct* sh)
{
  SH2movls4(sh, 1,7, 13);
}
#define init_1_1_7_14
void FASTCALL call_1_1_7_14(SH2_struct* sh)
{
  SH2movls4(sh, 1,7, 14);
}
#define init_1_1_7_15
void FASTCALL call_1_1_7_15(SH2_struct* sh)
{
  SH2movls4(sh, 1,7, 15);
}
#define init_1_1_8_0
void FASTCALL call_1_1_8_0(SH2_struct* sh)
{
  SH2movls4(sh, 1,8, 0);
}
#define init_1_1_8_1
void FASTCALL call_1_1_8_1(SH2_struct* sh)
{
  SH2movls4(sh, 1,8, 1);
}
#define init_1_1_8_2
void FASTCALL call_1_1_8_2(SH2_struct* sh)
{
  SH2movls4(sh, 1,8, 2);
}
#define init_1_1_8_3
void FASTCALL call_1_1_8_3(SH2_struct* sh)
{
  SH2movls4(sh, 1,8, 3);
}
#define init_1_1_8_4
void FASTCALL call_1_1_8_4(SH2_struct* sh)
{
  SH2movls4(sh, 1,8, 4);
}
#define init_1_1_8_5
void FASTCALL call_1_1_8_5(SH2_struct* sh)
{
  SH2movls4(sh, 1,8, 5);
}
#define init_1_1_8_6
void FASTCALL call_1_1_8_6(SH2_struct* sh)
{
  SH2movls4(sh, 1,8, 6);
}
#define init_1_1_8_7
void FASTCALL call_1_1_8_7(SH2_struct* sh)
{
  SH2movls4(sh, 1,8, 7);
}
#define init_1_1_8_8
void FASTCALL call_1_1_8_8(SH2_struct* sh)
{
  SH2movls4(sh, 1,8, 8);
}
#define init_1_1_8_9
void FASTCALL call_1_1_8_9(SH2_struct* sh)
{
  SH2movls4(sh, 1,8, 9);
}
#define init_1_1_8_10
void FASTCALL call_1_1_8_10(SH2_struct* sh)
{
  SH2movls4(sh, 1,8, 10);
}
#define init_1_1_8_11
void FASTCALL call_1_1_8_11(SH2_struct* sh)
{
  SH2movls4(sh, 1,8, 11);
}
#define init_1_1_8_12
void FASTCALL call_1_1_8_12(SH2_struct* sh)
{
  SH2movls4(sh, 1,8, 12);
}
#define init_1_1_8_13
void FASTCALL call_1_1_8_13(SH2_struct* sh)
{
  SH2movls4(sh, 1,8, 13);
}
#define init_1_1_8_14
void FASTCALL call_1_1_8_14(SH2_struct* sh)
{
  SH2movls4(sh, 1,8, 14);
}
#define init_1_1_8_15
void FASTCALL call_1_1_8_15(SH2_struct* sh)
{
  SH2movls4(sh, 1,8, 15);
}
#define init_1_1_9_0
void FASTCALL call_1_1_9_0(SH2_struct* sh)
{
  SH2movls4(sh, 1,9, 0);
}
#define init_1_1_9_1
void FASTCALL call_1_1_9_1(SH2_struct* sh)
{
  SH2movls4(sh, 1,9, 1);
}
#define init_1_1_9_2
void FASTCALL call_1_1_9_2(SH2_struct* sh)
{
  SH2movls4(sh, 1,9, 2);
}
#define init_1_1_9_3
void FASTCALL call_1_1_9_3(SH2_struct* sh)
{
  SH2movls4(sh, 1,9, 3);
}
#define init_1_1_9_4
void FASTCALL call_1_1_9_4(SH2_struct* sh)
{
  SH2movls4(sh, 1,9, 4);
}
#define init_1_1_9_5
void FASTCALL call_1_1_9_5(SH2_struct* sh)
{
  SH2movls4(sh, 1,9, 5);
}
#define init_1_1_9_6
void FASTCALL call_1_1_9_6(SH2_struct* sh)
{
  SH2movls4(sh, 1,9, 6);
}
#define init_1_1_9_7
void FASTCALL call_1_1_9_7(SH2_struct* sh)
{
  SH2movls4(sh, 1,9, 7);
}
#define init_1_1_9_8
void FASTCALL call_1_1_9_8(SH2_struct* sh)
{
  SH2movls4(sh, 1,9, 8);
}
#define init_1_1_9_9
void FASTCALL call_1_1_9_9(SH2_struct* sh)
{
  SH2movls4(sh, 1,9, 9);
}
#define init_1_1_9_10
void FASTCALL call_1_1_9_10(SH2_struct* sh)
{
  SH2movls4(sh, 1,9, 10);
}
#define init_1_1_9_11
void FASTCALL call_1_1_9_11(SH2_struct* sh)
{
  SH2movls4(sh, 1,9, 11);
}
#define init_1_1_9_12
void FASTCALL call_1_1_9_12(SH2_struct* sh)
{
  SH2movls4(sh, 1,9, 12);
}
#define init_1_1_9_13
void FASTCALL call_1_1_9_13(SH2_struct* sh)
{
  SH2movls4(sh, 1,9, 13);
}
#define init_1_1_9_14
void FASTCALL call_1_1_9_14(SH2_struct* sh)
{
  SH2movls4(sh, 1,9, 14);
}
#define init_1_1_9_15
void FASTCALL call_1_1_9_15(SH2_struct* sh)
{
  SH2movls4(sh, 1,9, 15);
}
#define init_1_1_10_0
void FASTCALL call_1_1_10_0(SH2_struct* sh)
{
  SH2movls4(sh, 1,10, 0);
}
#define init_1_1_10_1
void FASTCALL call_1_1_10_1(SH2_struct* sh)
{
  SH2movls4(sh, 1,10, 1);
}
#define init_1_1_10_2
void FASTCALL call_1_1_10_2(SH2_struct* sh)
{
  SH2movls4(sh, 1,10, 2);
}
#define init_1_1_10_3
void FASTCALL call_1_1_10_3(SH2_struct* sh)
{
  SH2movls4(sh, 1,10, 3);
}
#define init_1_1_10_4
void FASTCALL call_1_1_10_4(SH2_struct* sh)
{
  SH2movls4(sh, 1,10, 4);
}
#define init_1_1_10_5
void FASTCALL call_1_1_10_5(SH2_struct* sh)
{
  SH2movls4(sh, 1,10, 5);
}
#define init_1_1_10_6
void FASTCALL call_1_1_10_6(SH2_struct* sh)
{
  SH2movls4(sh, 1,10, 6);
}
#define init_1_1_10_7
void FASTCALL call_1_1_10_7(SH2_struct* sh)
{
  SH2movls4(sh, 1,10, 7);
}
#define init_1_1_10_8
void FASTCALL call_1_1_10_8(SH2_struct* sh)
{
  SH2movls4(sh, 1,10, 8);
}
#define init_1_1_10_9
void FASTCALL call_1_1_10_9(SH2_struct* sh)
{
  SH2movls4(sh, 1,10, 9);
}
#define init_1_1_10_10
void FASTCALL call_1_1_10_10(SH2_struct* sh)
{
  SH2movls4(sh, 1,10, 10);
}
#define init_1_1_10_11
void FASTCALL call_1_1_10_11(SH2_struct* sh)
{
  SH2movls4(sh, 1,10, 11);
}
#define init_1_1_10_12
void FASTCALL call_1_1_10_12(SH2_struct* sh)
{
  SH2movls4(sh, 1,10, 12);
}
#define init_1_1_10_13
void FASTCALL call_1_1_10_13(SH2_struct* sh)
{
  SH2movls4(sh, 1,10, 13);
}
#define init_1_1_10_14
void FASTCALL call_1_1_10_14(SH2_struct* sh)
{
  SH2movls4(sh, 1,10, 14);
}
#define init_1_1_10_15
void FASTCALL call_1_1_10_15(SH2_struct* sh)
{
  SH2movls4(sh, 1,10, 15);
}
#define init_1_1_11_0
void FASTCALL call_1_1_11_0(SH2_struct* sh)
{
  SH2movls4(sh, 1,11, 0);
}
#define init_1_1_11_1
void FASTCALL call_1_1_11_1(SH2_struct* sh)
{
  SH2movls4(sh, 1,11, 1);
}
#define init_1_1_11_2
void FASTCALL call_1_1_11_2(SH2_struct* sh)
{
  SH2movls4(sh, 1,11, 2);
}
#define init_1_1_11_3
void FASTCALL call_1_1_11_3(SH2_struct* sh)
{
  SH2movls4(sh, 1,11, 3);
}
#define init_1_1_11_4
void FASTCALL call_1_1_11_4(SH2_struct* sh)
{
  SH2movls4(sh, 1,11, 4);
}
#define init_1_1_11_5
void FASTCALL call_1_1_11_5(SH2_struct* sh)
{
  SH2movls4(sh, 1,11, 5);
}
#define init_1_1_11_6
void FASTCALL call_1_1_11_6(SH2_struct* sh)
{
  SH2movls4(sh, 1,11, 6);
}
#define init_1_1_11_7
void FASTCALL call_1_1_11_7(SH2_struct* sh)
{
  SH2movls4(sh, 1,11, 7);
}
#define init_1_1_11_8
void FASTCALL call_1_1_11_8(SH2_struct* sh)
{
  SH2movls4(sh, 1,11, 8);
}
#define init_1_1_11_9
void FASTCALL call_1_1_11_9(SH2_struct* sh)
{
  SH2movls4(sh, 1,11, 9);
}
#define init_1_1_11_10
void FASTCALL call_1_1_11_10(SH2_struct* sh)
{
  SH2movls4(sh, 1,11, 10);
}
#define init_1_1_11_11
void FASTCALL call_1_1_11_11(SH2_struct* sh)
{
  SH2movls4(sh, 1,11, 11);
}
#define init_1_1_11_12
void FASTCALL call_1_1_11_12(SH2_struct* sh)
{
  SH2movls4(sh, 1,11, 12);
}
#define init_1_1_11_13
void FASTCALL call_1_1_11_13(SH2_struct* sh)
{
  SH2movls4(sh, 1,11, 13);
}
#define init_1_1_11_14
void FASTCALL call_1_1_11_14(SH2_struct* sh)
{
  SH2movls4(sh, 1,11, 14);
}
#define init_1_1_11_15
void FASTCALL call_1_1_11_15(SH2_struct* sh)
{
  SH2movls4(sh, 1,11, 15);
}
#define init_1_1_12_0
void FASTCALL call_1_1_12_0(SH2_struct* sh)
{
  SH2movls4(sh, 1,12, 0);
}
#define init_1_1_12_1
void FASTCALL call_1_1_12_1(SH2_struct* sh)
{
  SH2movls4(sh, 1,12, 1);
}
#define init_1_1_12_2
void FASTCALL call_1_1_12_2(SH2_struct* sh)
{
  SH2movls4(sh, 1,12, 2);
}
#define init_1_1_12_3
void FASTCALL call_1_1_12_3(SH2_struct* sh)
{
  SH2movls4(sh, 1,12, 3);
}
#define init_1_1_12_4
void FASTCALL call_1_1_12_4(SH2_struct* sh)
{
  SH2movls4(sh, 1,12, 4);
}
#define init_1_1_12_5
void FASTCALL call_1_1_12_5(SH2_struct* sh)
{
  SH2movls4(sh, 1,12, 5);
}
#define init_1_1_12_6
void FASTCALL call_1_1_12_6(SH2_struct* sh)
{
  SH2movls4(sh, 1,12, 6);
}
#define init_1_1_12_7
void FASTCALL call_1_1_12_7(SH2_struct* sh)
{
  SH2movls4(sh, 1,12, 7);
}
#define init_1_1_12_8
void FASTCALL call_1_1_12_8(SH2_struct* sh)
{
  SH2movls4(sh, 1,12, 8);
}
#define init_1_1_12_9
void FASTCALL call_1_1_12_9(SH2_struct* sh)
{
  SH2movls4(sh, 1,12, 9);
}
#define init_1_1_12_10
void FASTCALL call_1_1_12_10(SH2_struct* sh)
{
  SH2movls4(sh, 1,12, 10);
}
#define init_1_1_12_11
void FASTCALL call_1_1_12_11(SH2_struct* sh)
{
  SH2movls4(sh, 1,12, 11);
}
#define init_1_1_12_12
void FASTCALL call_1_1_12_12(SH2_struct* sh)
{
  SH2movls4(sh, 1,12, 12);
}
#define init_1_1_12_13
void FASTCALL call_1_1_12_13(SH2_struct* sh)
{
  SH2movls4(sh, 1,12, 13);
}
#define init_1_1_12_14
void FASTCALL call_1_1_12_14(SH2_struct* sh)
{
  SH2movls4(sh, 1,12, 14);
}
#define init_1_1_12_15
void FASTCALL call_1_1_12_15(SH2_struct* sh)
{
  SH2movls4(sh, 1,12, 15);
}
#define init_1_1_13_0
void FASTCALL call_1_1_13_0(SH2_struct* sh)
{
  SH2movls4(sh, 1,13, 0);
}
#define init_1_1_13_1
void FASTCALL call_1_1_13_1(SH2_struct* sh)
{
  SH2movls4(sh, 1,13, 1);
}
#define init_1_1_13_2
void FASTCALL call_1_1_13_2(SH2_struct* sh)
{
  SH2movls4(sh, 1,13, 2);
}
#define init_1_1_13_3
void FASTCALL call_1_1_13_3(SH2_struct* sh)
{
  SH2movls4(sh, 1,13, 3);
}
#define init_1_1_13_4
void FASTCALL call_1_1_13_4(SH2_struct* sh)
{
  SH2movls4(sh, 1,13, 4);
}
#define init_1_1_13_5
void FASTCALL call_1_1_13_5(SH2_struct* sh)
{
  SH2movls4(sh, 1,13, 5);
}
#define init_1_1_13_6
void FASTCALL call_1_1_13_6(SH2_struct* sh)
{
  SH2movls4(sh, 1,13, 6);
}
#define init_1_1_13_7
void FASTCALL call_1_1_13_7(SH2_struct* sh)
{
  SH2movls4(sh, 1,13, 7);
}
#define init_1_1_13_8
void FASTCALL call_1_1_13_8(SH2_struct* sh)
{
  SH2movls4(sh, 1,13, 8);
}
#define init_1_1_13_9
void FASTCALL call_1_1_13_9(SH2_struct* sh)
{
  SH2movls4(sh, 1,13, 9);
}
#define init_1_1_13_10
void FASTCALL call_1_1_13_10(SH2_struct* sh)
{
  SH2movls4(sh, 1,13, 10);
}
#define init_1_1_13_11
void FASTCALL call_1_1_13_11(SH2_struct* sh)
{
  SH2movls4(sh, 1,13, 11);
}
#define init_1_1_13_12
void FASTCALL call_1_1_13_12(SH2_struct* sh)
{
  SH2movls4(sh, 1,13, 12);
}
#define init_1_1_13_13
void FASTCALL call_1_1_13_13(SH2_struct* sh)
{
  SH2movls4(sh, 1,13, 13);
}
#define init_1_1_13_14
void FASTCALL call_1_1_13_14(SH2_struct* sh)
{
  SH2movls4(sh, 1,13, 14);
}
#define init_1_1_13_15
void FASTCALL call_1_1_13_15(SH2_struct* sh)
{
  SH2movls4(sh, 1,13, 15);
}
#define init_1_1_14_0
void FASTCALL call_1_1_14_0(SH2_struct* sh)
{
  SH2movls4(sh, 1,14, 0);
}
#define init_1_1_14_1
void FASTCALL call_1_1_14_1(SH2_struct* sh)
{
  SH2movls4(sh, 1,14, 1);
}
#define init_1_1_14_2
void FASTCALL call_1_1_14_2(SH2_struct* sh)
{
  SH2movls4(sh, 1,14, 2);
}
#define init_1_1_14_3
void FASTCALL call_1_1_14_3(SH2_struct* sh)
{
  SH2movls4(sh, 1,14, 3);
}
#define init_1_1_14_4
void FASTCALL call_1_1_14_4(SH2_struct* sh)
{
  SH2movls4(sh, 1,14, 4);
}
#define init_1_1_14_5
void FASTCALL call_1_1_14_5(SH2_struct* sh)
{
  SH2movls4(sh, 1,14, 5);
}
#define init_1_1_14_6
void FASTCALL call_1_1_14_6(SH2_struct* sh)
{
  SH2movls4(sh, 1,14, 6);
}
#define init_1_1_14_7
void FASTCALL call_1_1_14_7(SH2_struct* sh)
{
  SH2movls4(sh, 1,14, 7);
}
#define init_1_1_14_8
void FASTCALL call_1_1_14_8(SH2_struct* sh)
{
  SH2movls4(sh, 1,14, 8);
}
#define init_1_1_14_9
void FASTCALL call_1_1_14_9(SH2_struct* sh)
{
  SH2movls4(sh, 1,14, 9);
}
#define init_1_1_14_10
void FASTCALL call_1_1_14_10(SH2_struct* sh)
{
  SH2movls4(sh, 1,14, 10);
}
#define init_1_1_14_11
void FASTCALL call_1_1_14_11(SH2_struct* sh)
{
  SH2movls4(sh, 1,14, 11);
}
#define init_1_1_14_12
void FASTCALL call_1_1_14_12(SH2_struct* sh)
{
  SH2movls4(sh, 1,14, 12);
}
#define init_1_1_14_13
void FASTCALL call_1_1_14_13(SH2_struct* sh)
{
  SH2movls4(sh, 1,14, 13);
}
#define init_1_1_14_14
void FASTCALL call_1_1_14_14(SH2_struct* sh)
{
  SH2movls4(sh, 1,14, 14);
}
#define init_1_1_14_15
void FASTCALL call_1_1_14_15(SH2_struct* sh)
{
  SH2movls4(sh, 1,14, 15);
}
#define init_1_1_15_0
void FASTCALL call_1_1_15_0(SH2_struct* sh)
{
  SH2movls4(sh, 1,15, 0);
}
#define init_1_1_15_1
void FASTCALL call_1_1_15_1(SH2_struct* sh)
{
  SH2movls4(sh, 1,15, 1);
}
#define init_1_1_15_2
void FASTCALL call_1_1_15_2(SH2_struct* sh)
{
  SH2movls4(sh, 1,15, 2);
}
#define init_1_1_15_3
void FASTCALL call_1_1_15_3(SH2_struct* sh)
{
  SH2movls4(sh, 1,15, 3);
}
#define init_1_1_15_4
void FASTCALL call_1_1_15_4(SH2_struct* sh)
{
  SH2movls4(sh, 1,15, 4);
}
#define init_1_1_15_5
void FASTCALL call_1_1_15_5(SH2_struct* sh)
{
  SH2movls4(sh, 1,15, 5);
}
#define init_1_1_15_6
void FASTCALL call_1_1_15_6(SH2_struct* sh)
{
  SH2movls4(sh, 1,15, 6);
}
#define init_1_1_15_7
void FASTCALL call_1_1_15_7(SH2_struct* sh)
{
  SH2movls4(sh, 1,15, 7);
}
#define init_1_1_15_8
void FASTCALL call_1_1_15_8(SH2_struct* sh)
{
  SH2movls4(sh, 1,15, 8);
}
#define init_1_1_15_9
void FASTCALL call_1_1_15_9(SH2_struct* sh)
{
  SH2movls4(sh, 1,15, 9);
}
#define init_1_1_15_10
void FASTCALL call_1_1_15_10(SH2_struct* sh)
{
  SH2movls4(sh, 1,15, 10);
}
#define init_1_1_15_11
void FASTCALL call_1_1_15_11(SH2_struct* sh)
{
  SH2movls4(sh, 1,15, 11);
}
#define init_1_1_15_12
void FASTCALL call_1_1_15_12(SH2_struct* sh)
{
  SH2movls4(sh, 1,15, 12);
}
#define init_1_1_15_13
void FASTCALL call_1_1_15_13(SH2_struct* sh)
{
  SH2movls4(sh, 1,15, 13);
}
#define init_1_1_15_14
void FASTCALL call_1_1_15_14(SH2_struct* sh)
{
  SH2movls4(sh, 1,15, 14);
}
#define init_1_1_15_15
void FASTCALL call_1_1_15_15(SH2_struct* sh)
{
  SH2movls4(sh, 1,15, 15);
}
#define init_1_2_0_0
void FASTCALL call_1_2_0_0(SH2_struct* sh)
{
  SH2movls4(sh, 2,0, 0);
}
#define init_1_2_0_1
void FASTCALL call_1_2_0_1(SH2_struct* sh)
{
  SH2movls4(sh, 2,0, 1);
}
#define init_1_2_0_2
void FASTCALL call_1_2_0_2(SH2_struct* sh)
{
  SH2movls4(sh, 2,0, 2);
}
#define init_1_2_0_3
void FASTCALL call_1_2_0_3(SH2_struct* sh)
{
  SH2movls4(sh, 2,0, 3);
}
#define init_1_2_0_4
void FASTCALL call_1_2_0_4(SH2_struct* sh)
{
  SH2movls4(sh, 2,0, 4);
}
#define init_1_2_0_5
void FASTCALL call_1_2_0_5(SH2_struct* sh)
{
  SH2movls4(sh, 2,0, 5);
}
#define init_1_2_0_6
void FASTCALL call_1_2_0_6(SH2_struct* sh)
{
  SH2movls4(sh, 2,0, 6);
}
#define init_1_2_0_7
void FASTCALL call_1_2_0_7(SH2_struct* sh)
{
  SH2movls4(sh, 2,0, 7);
}
#define init_1_2_0_8
void FASTCALL call_1_2_0_8(SH2_struct* sh)
{
  SH2movls4(sh, 2,0, 8);
}
#define init_1_2_0_9
void FASTCALL call_1_2_0_9(SH2_struct* sh)
{
  SH2movls4(sh, 2,0, 9);
}
#define init_1_2_0_10
void FASTCALL call_1_2_0_10(SH2_struct* sh)
{
  SH2movls4(sh, 2,0, 10);
}
#define init_1_2_0_11
void FASTCALL call_1_2_0_11(SH2_struct* sh)
{
  SH2movls4(sh, 2,0, 11);
}
#define init_1_2_0_12
void FASTCALL call_1_2_0_12(SH2_struct* sh)
{
  SH2movls4(sh, 2,0, 12);
}
#define init_1_2_0_13
void FASTCALL call_1_2_0_13(SH2_struct* sh)
{
  SH2movls4(sh, 2,0, 13);
}
#define init_1_2_0_14
void FASTCALL call_1_2_0_14(SH2_struct* sh)
{
  SH2movls4(sh, 2,0, 14);
}
#define init_1_2_0_15
void FASTCALL call_1_2_0_15(SH2_struct* sh)
{
  SH2movls4(sh, 2,0, 15);
}
#define init_1_2_1_0
void FASTCALL call_1_2_1_0(SH2_struct* sh)
{
  SH2movls4(sh, 2,1, 0);
}
#define init_1_2_1_1
void FASTCALL call_1_2_1_1(SH2_struct* sh)
{
  SH2movls4(sh, 2,1, 1);
}
#define init_1_2_1_2
void FASTCALL call_1_2_1_2(SH2_struct* sh)
{
  SH2movls4(sh, 2,1, 2);
}
#define init_1_2_1_3
void FASTCALL call_1_2_1_3(SH2_struct* sh)
{
  SH2movls4(sh, 2,1, 3);
}
#define init_1_2_1_4
void FASTCALL call_1_2_1_4(SH2_struct* sh)
{
  SH2movls4(sh, 2,1, 4);
}
#define init_1_2_1_5
void FASTCALL call_1_2_1_5(SH2_struct* sh)
{
  SH2movls4(sh, 2,1, 5);
}
#define init_1_2_1_6
void FASTCALL call_1_2_1_6(SH2_struct* sh)
{
  SH2movls4(sh, 2,1, 6);
}
#define init_1_2_1_7
void FASTCALL call_1_2_1_7(SH2_struct* sh)
{
  SH2movls4(sh, 2,1, 7);
}
#define init_1_2_1_8
void FASTCALL call_1_2_1_8(SH2_struct* sh)
{
  SH2movls4(sh, 2,1, 8);
}
#define init_1_2_1_9
void FASTCALL call_1_2_1_9(SH2_struct* sh)
{
  SH2movls4(sh, 2,1, 9);
}
#define init_1_2_1_10
void FASTCALL call_1_2_1_10(SH2_struct* sh)
{
  SH2movls4(sh, 2,1, 10);
}
#define init_1_2_1_11
void FASTCALL call_1_2_1_11(SH2_struct* sh)
{
  SH2movls4(sh, 2,1, 11);
}
#define init_1_2_1_12
void FASTCALL call_1_2_1_12(SH2_struct* sh)
{
  SH2movls4(sh, 2,1, 12);
}
#define init_1_2_1_13
void FASTCALL call_1_2_1_13(SH2_struct* sh)
{
  SH2movls4(sh, 2,1, 13);
}
#define init_1_2_1_14
void FASTCALL call_1_2_1_14(SH2_struct* sh)
{
  SH2movls4(sh, 2,1, 14);
}
#define init_1_2_1_15
void FASTCALL call_1_2_1_15(SH2_struct* sh)
{
  SH2movls4(sh, 2,1, 15);
}
#define init_1_2_2_0
void FASTCALL call_1_2_2_0(SH2_struct* sh)
{
  SH2movls4(sh, 2,2, 0);
}
#define init_1_2_2_1
void FASTCALL call_1_2_2_1(SH2_struct* sh)
{
  SH2movls4(sh, 2,2, 1);
}
#define init_1_2_2_2
void FASTCALL call_1_2_2_2(SH2_struct* sh)
{
  SH2movls4(sh, 2,2, 2);
}
#define init_1_2_2_3
void FASTCALL call_1_2_2_3(SH2_struct* sh)
{
  SH2movls4(sh, 2,2, 3);
}
#define init_1_2_2_4
void FASTCALL call_1_2_2_4(SH2_struct* sh)
{
  SH2movls4(sh, 2,2, 4);
}
#define init_1_2_2_5
void FASTCALL call_1_2_2_5(SH2_struct* sh)
{
  SH2movls4(sh, 2,2, 5);
}
#define init_1_2_2_6
void FASTCALL call_1_2_2_6(SH2_struct* sh)
{
  SH2movls4(sh, 2,2, 6);
}
#define init_1_2_2_7
void FASTCALL call_1_2_2_7(SH2_struct* sh)
{
  SH2movls4(sh, 2,2, 7);
}
#define init_1_2_2_8
void FASTCALL call_1_2_2_8(SH2_struct* sh)
{
  SH2movls4(sh, 2,2, 8);
}
#define init_1_2_2_9
void FASTCALL call_1_2_2_9(SH2_struct* sh)
{
  SH2movls4(sh, 2,2, 9);
}
#define init_1_2_2_10
void FASTCALL call_1_2_2_10(SH2_struct* sh)
{
  SH2movls4(sh, 2,2, 10);
}
#define init_1_2_2_11
void FASTCALL call_1_2_2_11(SH2_struct* sh)
{
  SH2movls4(sh, 2,2, 11);
}
#define init_1_2_2_12
void FASTCALL call_1_2_2_12(SH2_struct* sh)
{
  SH2movls4(sh, 2,2, 12);
}
#define init_1_2_2_13
void FASTCALL call_1_2_2_13(SH2_struct* sh)
{
  SH2movls4(sh, 2,2, 13);
}
#define init_1_2_2_14
void FASTCALL call_1_2_2_14(SH2_struct* sh)
{
  SH2movls4(sh, 2,2, 14);
}
#define init_1_2_2_15
void FASTCALL call_1_2_2_15(SH2_struct* sh)
{
  SH2movls4(sh, 2,2, 15);
}
#define init_1_2_3_0
void FASTCALL call_1_2_3_0(SH2_struct* sh)
{
  SH2movls4(sh, 2,3, 0);
}
#define init_1_2_3_1
void FASTCALL call_1_2_3_1(SH2_struct* sh)
{
  SH2movls4(sh, 2,3, 1);
}
#define init_1_2_3_2
void FASTCALL call_1_2_3_2(SH2_struct* sh)
{
  SH2movls4(sh, 2,3, 2);
}
#define init_1_2_3_3
void FASTCALL call_1_2_3_3(SH2_struct* sh)
{
  SH2movls4(sh, 2,3, 3);
}
#define init_1_2_3_4
void FASTCALL call_1_2_3_4(SH2_struct* sh)
{
  SH2movls4(sh, 2,3, 4);
}
#define init_1_2_3_5
void FASTCALL call_1_2_3_5(SH2_struct* sh)
{
  SH2movls4(sh, 2,3, 5);
}
#define init_1_2_3_6
void FASTCALL call_1_2_3_6(SH2_struct* sh)
{
  SH2movls4(sh, 2,3, 6);
}
#define init_1_2_3_7
void FASTCALL call_1_2_3_7(SH2_struct* sh)
{
  SH2movls4(sh, 2,3, 7);
}
#define init_1_2_3_8
void FASTCALL call_1_2_3_8(SH2_struct* sh)
{
  SH2movls4(sh, 2,3, 8);
}
#define init_1_2_3_9
void FASTCALL call_1_2_3_9(SH2_struct* sh)
{
  SH2movls4(sh, 2,3, 9);
}
#define init_1_2_3_10
void FASTCALL call_1_2_3_10(SH2_struct* sh)
{
  SH2movls4(sh, 2,3, 10);
}
#define init_1_2_3_11
void FASTCALL call_1_2_3_11(SH2_struct* sh)
{
  SH2movls4(sh, 2,3, 11);
}
#define init_1_2_3_12
void FASTCALL call_1_2_3_12(SH2_struct* sh)
{
  SH2movls4(sh, 2,3, 12);
}
#define init_1_2_3_13
void FASTCALL call_1_2_3_13(SH2_struct* sh)
{
  SH2movls4(sh, 2,3, 13);
}
#define init_1_2_3_14
void FASTCALL call_1_2_3_14(SH2_struct* sh)
{
  SH2movls4(sh, 2,3, 14);
}
#define init_1_2_3_15
void FASTCALL call_1_2_3_15(SH2_struct* sh)
{
  SH2movls4(sh, 2,3, 15);
}
#define init_1_2_4_0
void FASTCALL call_1_2_4_0(SH2_struct* sh)
{
  SH2movls4(sh, 2,4, 0);
}
#define init_1_2_4_1
void FASTCALL call_1_2_4_1(SH2_struct* sh)
{
  SH2movls4(sh, 2,4, 1);
}
#define init_1_2_4_2
void FASTCALL call_1_2_4_2(SH2_struct* sh)
{
  SH2movls4(sh, 2,4, 2);
}
#define init_1_2_4_3
void FASTCALL call_1_2_4_3(SH2_struct* sh)
{
  SH2movls4(sh, 2,4, 3);
}
#define init_1_2_4_4
void FASTCALL call_1_2_4_4(SH2_struct* sh)
{
  SH2movls4(sh, 2,4, 4);
}
#define init_1_2_4_5
void FASTCALL call_1_2_4_5(SH2_struct* sh)
{
  SH2movls4(sh, 2,4, 5);
}
#define init_1_2_4_6
void FASTCALL call_1_2_4_6(SH2_struct* sh)
{
  SH2movls4(sh, 2,4, 6);
}
#define init_1_2_4_7
void FASTCALL call_1_2_4_7(SH2_struct* sh)
{
  SH2movls4(sh, 2,4, 7);
}
#define init_1_2_4_8
void FASTCALL call_1_2_4_8(SH2_struct* sh)
{
  SH2movls4(sh, 2,4, 8);
}
#define init_1_2_4_9
void FASTCALL call_1_2_4_9(SH2_struct* sh)
{
  SH2movls4(sh, 2,4, 9);
}
#define init_1_2_4_10
void FASTCALL call_1_2_4_10(SH2_struct* sh)
{
  SH2movls4(sh, 2,4, 10);
}
#define init_1_2_4_11
void FASTCALL call_1_2_4_11(SH2_struct* sh)
{
  SH2movls4(sh, 2,4, 11);
}
#define init_1_2_4_12
void FASTCALL call_1_2_4_12(SH2_struct* sh)
{
  SH2movls4(sh, 2,4, 12);
}
#define init_1_2_4_13
void FASTCALL call_1_2_4_13(SH2_struct* sh)
{
  SH2movls4(sh, 2,4, 13);
}
#define init_1_2_4_14
void FASTCALL call_1_2_4_14(SH2_struct* sh)
{
  SH2movls4(sh, 2,4, 14);
}
#define init_1_2_4_15
void FASTCALL call_1_2_4_15(SH2_struct* sh)
{
  SH2movls4(sh, 2,4, 15);
}
#define init_1_2_5_0
void FASTCALL call_1_2_5_0(SH2_struct* sh)
{
  SH2movls4(sh, 2,5, 0);
}
#define init_1_2_5_1
void FASTCALL call_1_2_5_1(SH2_struct* sh)
{
  SH2movls4(sh, 2,5, 1);
}
#define init_1_2_5_2
void FASTCALL call_1_2_5_2(SH2_struct* sh)
{
  SH2movls4(sh, 2,5, 2);
}
#define init_1_2_5_3
void FASTCALL call_1_2_5_3(SH2_struct* sh)
{
  SH2movls4(sh, 2,5, 3);
}
#define init_1_2_5_4
void FASTCALL call_1_2_5_4(SH2_struct* sh)
{
  SH2movls4(sh, 2,5, 4);
}
#define init_1_2_5_5
void FASTCALL call_1_2_5_5(SH2_struct* sh)
{
  SH2movls4(sh, 2,5, 5);
}
#define init_1_2_5_6
void FASTCALL call_1_2_5_6(SH2_struct* sh)
{
  SH2movls4(sh, 2,5, 6);
}
#define init_1_2_5_7
void FASTCALL call_1_2_5_7(SH2_struct* sh)
{
  SH2movls4(sh, 2,5, 7);
}
#define init_1_2_5_8
void FASTCALL call_1_2_5_8(SH2_struct* sh)
{
  SH2movls4(sh, 2,5, 8);
}
#define init_1_2_5_9
void FASTCALL call_1_2_5_9(SH2_struct* sh)
{
  SH2movls4(sh, 2,5, 9);
}
#define init_1_2_5_10
void FASTCALL call_1_2_5_10(SH2_struct* sh)
{
  SH2movls4(sh, 2,5, 10);
}
#define init_1_2_5_11
void FASTCALL call_1_2_5_11(SH2_struct* sh)
{
  SH2movls4(sh, 2,5, 11);
}
#define init_1_2_5_12
void FASTCALL call_1_2_5_12(SH2_struct* sh)
{
  SH2movls4(sh, 2,5, 12);
}
#define init_1_2_5_13
void FASTCALL call_1_2_5_13(SH2_struct* sh)
{
  SH2movls4(sh, 2,5, 13);
}
#define init_1_2_5_14
void FASTCALL call_1_2_5_14(SH2_struct* sh)
{
  SH2movls4(sh, 2,5, 14);
}
#define init_1_2_5_15
void FASTCALL call_1_2_5_15(SH2_struct* sh)
{
  SH2movls4(sh, 2,5, 15);
}
#define init_1_2_6_0
void FASTCALL call_1_2_6_0(SH2_struct* sh)
{
  SH2movls4(sh, 2,6, 0);
}
#define init_1_2_6_1
void FASTCALL call_1_2_6_1(SH2_struct* sh)
{
  SH2movls4(sh, 2,6, 1);
}
#define init_1_2_6_2
void FASTCALL call_1_2_6_2(SH2_struct* sh)
{
  SH2movls4(sh, 2,6, 2);
}
#define init_1_2_6_3
void FASTCALL call_1_2_6_3(SH2_struct* sh)
{
  SH2movls4(sh, 2,6, 3);
}
#define init_1_2_6_4
void FASTCALL call_1_2_6_4(SH2_struct* sh)
{
  SH2movls4(sh, 2,6, 4);
}
#define init_1_2_6_5
void FASTCALL call_1_2_6_5(SH2_struct* sh)
{
  SH2movls4(sh, 2,6, 5);
}
#define init_1_2_6_6
void FASTCALL call_1_2_6_6(SH2_struct* sh)
{
  SH2movls4(sh, 2,6, 6);
}
#define init_1_2_6_7
void FASTCALL call_1_2_6_7(SH2_struct* sh)
{
  SH2movls4(sh, 2,6, 7);
}
#define init_1_2_6_8
void FASTCALL call_1_2_6_8(SH2_struct* sh)
{
  SH2movls4(sh, 2,6, 8);
}
#define init_1_2_6_9
void FASTCALL call_1_2_6_9(SH2_struct* sh)
{
  SH2movls4(sh, 2,6, 9);
}
#define init_1_2_6_10
void FASTCALL call_1_2_6_10(SH2_struct* sh)
{
  SH2movls4(sh, 2,6, 10);
}
#define init_1_2_6_11
void FASTCALL call_1_2_6_11(SH2_struct* sh)
{
  SH2movls4(sh, 2,6, 11);
}
#define init_1_2_6_12
void FASTCALL call_1_2_6_12(SH2_struct* sh)
{
  SH2movls4(sh, 2,6, 12);
}
#define init_1_2_6_13
void FASTCALL call_1_2_6_13(SH2_struct* sh)
{
  SH2movls4(sh, 2,6, 13);
}
#define init_1_2_6_14
void FASTCALL call_1_2_6_14(SH2_struct* sh)
{
  SH2movls4(sh, 2,6, 14);
}
#define init_1_2_6_15
void FASTCALL call_1_2_6_15(SH2_struct* sh)
{
  SH2movls4(sh, 2,6, 15);
}
#define init_1_2_7_0
void FASTCALL call_1_2_7_0(SH2_struct* sh)
{
  SH2movls4(sh, 2,7, 0);
}
#define init_1_2_7_1
void FASTCALL call_1_2_7_1(SH2_struct* sh)
{
  SH2movls4(sh, 2,7, 1);
}
#define init_1_2_7_2
void FASTCALL call_1_2_7_2(SH2_struct* sh)
{
  SH2movls4(sh, 2,7, 2);
}
#define init_1_2_7_3
void FASTCALL call_1_2_7_3(SH2_struct* sh)
{
  SH2movls4(sh, 2,7, 3);
}
#define init_1_2_7_4
void FASTCALL call_1_2_7_4(SH2_struct* sh)
{
  SH2movls4(sh, 2,7, 4);
}
#define init_1_2_7_5
void FASTCALL call_1_2_7_5(SH2_struct* sh)
{
  SH2movls4(sh, 2,7, 5);
}
#define init_1_2_7_6
void FASTCALL call_1_2_7_6(SH2_struct* sh)
{
  SH2movls4(sh, 2,7, 6);
}
#define init_1_2_7_7
void FASTCALL call_1_2_7_7(SH2_struct* sh)
{
  SH2movls4(sh, 2,7, 7);
}
#define init_1_2_7_8
void FASTCALL call_1_2_7_8(SH2_struct* sh)
{
  SH2movls4(sh, 2,7, 8);
}
#define init_1_2_7_9
void FASTCALL call_1_2_7_9(SH2_struct* sh)
{
  SH2movls4(sh, 2,7, 9);
}
#define init_1_2_7_10
void FASTCALL call_1_2_7_10(SH2_struct* sh)
{
  SH2movls4(sh, 2,7, 10);
}
#define init_1_2_7_11
void FASTCALL call_1_2_7_11(SH2_struct* sh)
{
  SH2movls4(sh, 2,7, 11);
}
#define init_1_2_7_12
void FASTCALL call_1_2_7_12(SH2_struct* sh)
{
  SH2movls4(sh, 2,7, 12);
}
#define init_1_2_7_13
void FASTCALL call_1_2_7_13(SH2_struct* sh)
{
  SH2movls4(sh, 2,7, 13);
}
#define init_1_2_7_14
void FASTCALL call_1_2_7_14(SH2_struct* sh)
{
  SH2movls4(sh, 2,7, 14);
}
#define init_1_2_7_15
void FASTCALL call_1_2_7_15(SH2_struct* sh)
{
  SH2movls4(sh, 2,7, 15);
}
#define init_1_2_8_0
void FASTCALL call_1_2_8_0(SH2_struct* sh)
{
  SH2movls4(sh, 2,8, 0);
}
#define init_1_2_8_1
void FASTCALL call_1_2_8_1(SH2_struct* sh)
{
  SH2movls4(sh, 2,8, 1);
}
#define init_1_2_8_2
void FASTCALL call_1_2_8_2(SH2_struct* sh)
{
  SH2movls4(sh, 2,8, 2);
}
#define init_1_2_8_3
void FASTCALL call_1_2_8_3(SH2_struct* sh)
{
  SH2movls4(sh, 2,8, 3);
}
#define init_1_2_8_4
void FASTCALL call_1_2_8_4(SH2_struct* sh)
{
  SH2movls4(sh, 2,8, 4);
}
#define init_1_2_8_5
void FASTCALL call_1_2_8_5(SH2_struct* sh)
{
  SH2movls4(sh, 2,8, 5);
}
#define init_1_2_8_6
void FASTCALL call_1_2_8_6(SH2_struct* sh)
{
  SH2movls4(sh, 2,8, 6);
}
#define init_1_2_8_7
void FASTCALL call_1_2_8_7(SH2_struct* sh)
{
  SH2movls4(sh, 2,8, 7);
}
#define init_1_2_8_8
void FASTCALL call_1_2_8_8(SH2_struct* sh)
{
  SH2movls4(sh, 2,8, 8);
}
#define init_1_2_8_9
void FASTCALL call_1_2_8_9(SH2_struct* sh)
{
  SH2movls4(sh, 2,8, 9);
}
#define init_1_2_8_10
void FASTCALL call_1_2_8_10(SH2_struct* sh)
{
  SH2movls4(sh, 2,8, 10);
}
#define init_1_2_8_11
void FASTCALL call_1_2_8_11(SH2_struct* sh)
{
  SH2movls4(sh, 2,8, 11);
}
#define init_1_2_8_12
void FASTCALL call_1_2_8_12(SH2_struct* sh)
{
  SH2movls4(sh, 2,8, 12);
}
#define init_1_2_8_13
void FASTCALL call_1_2_8_13(SH2_struct* sh)
{
  SH2movls4(sh, 2,8, 13);
}
#define init_1_2_8_14
void FASTCALL call_1_2_8_14(SH2_struct* sh)
{
  SH2movls4(sh, 2,8, 14);
}
#define init_1_2_8_15
void FASTCALL call_1_2_8_15(SH2_struct* sh)
{
  SH2movls4(sh, 2,8, 15);
}
#define init_1_2_9_0
void FASTCALL call_1_2_9_0(SH2_struct* sh)
{
  SH2movls4(sh, 2,9, 0);
}
#define init_1_2_9_1
void FASTCALL call_1_2_9_1(SH2_struct* sh)
{
  SH2movls4(sh, 2,9, 1);
}
#define init_1_2_9_2
void FASTCALL call_1_2_9_2(SH2_struct* sh)
{
  SH2movls4(sh, 2,9, 2);
}
#define init_1_2_9_3
void FASTCALL call_1_2_9_3(SH2_struct* sh)
{
  SH2movls4(sh, 2,9, 3);
}
#define init_1_2_9_4
void FASTCALL call_1_2_9_4(SH2_struct* sh)
{
  SH2movls4(sh, 2,9, 4);
}
#define init_1_2_9_5
void FASTCALL call_1_2_9_5(SH2_struct* sh)
{
  SH2movls4(sh, 2,9, 5);
}
#define init_1_2_9_6
void FASTCALL call_1_2_9_6(SH2_struct* sh)
{
  SH2movls4(sh, 2,9, 6);
}
#define init_1_2_9_7
void FASTCALL call_1_2_9_7(SH2_struct* sh)
{
  SH2movls4(sh, 2,9, 7);
}
#define init_1_2_9_8
void FASTCALL call_1_2_9_8(SH2_struct* sh)
{
  SH2movls4(sh, 2,9, 8);
}
#define init_1_2_9_9
void FASTCALL call_1_2_9_9(SH2_struct* sh)
{
  SH2movls4(sh, 2,9, 9);
}
#define init_1_2_9_10
void FASTCALL call_1_2_9_10(SH2_struct* sh)
{
  SH2movls4(sh, 2,9, 10);
}
#define init_1_2_9_11
void FASTCALL call_1_2_9_11(SH2_struct* sh)
{
  SH2movls4(sh, 2,9, 11);
}
#define init_1_2_9_12
void FASTCALL call_1_2_9_12(SH2_struct* sh)
{
  SH2movls4(sh, 2,9, 12);
}
#define init_1_2_9_13
void FASTCALL call_1_2_9_13(SH2_struct* sh)
{
  SH2movls4(sh, 2,9, 13);
}
#define init_1_2_9_14
void FASTCALL call_1_2_9_14(SH2_struct* sh)
{
  SH2movls4(sh, 2,9, 14);
}
#define init_1_2_9_15
void FASTCALL call_1_2_9_15(SH2_struct* sh)
{
  SH2movls4(sh, 2,9, 15);
}
#define init_1_2_10_0
void FASTCALL call_1_2_10_0(SH2_struct* sh)
{
  SH2movls4(sh, 2,10, 0);
}
#define init_1_2_10_1
void FASTCALL call_1_2_10_1(SH2_struct* sh)
{
  SH2movls4(sh, 2,10, 1);
}
#define init_1_2_10_2
void FASTCALL call_1_2_10_2(SH2_struct* sh)
{
  SH2movls4(sh, 2,10, 2);
}
#define init_1_2_10_3
void FASTCALL call_1_2_10_3(SH2_struct* sh)
{
  SH2movls4(sh, 2,10, 3);
}
#define init_1_2_10_4
void FASTCALL call_1_2_10_4(SH2_struct* sh)
{
  SH2movls4(sh, 2,10, 4);
}
#define init_1_2_10_5
void FASTCALL call_1_2_10_5(SH2_struct* sh)
{
  SH2movls4(sh, 2,10, 5);
}
#define init_1_2_10_6
void FASTCALL call_1_2_10_6(SH2_struct* sh)
{
  SH2movls4(sh, 2,10, 6);
}
#define init_1_2_10_7
void FASTCALL call_1_2_10_7(SH2_struct* sh)
{
  SH2movls4(sh, 2,10, 7);
}
#define init_1_2_10_8
void FASTCALL call_1_2_10_8(SH2_struct* sh)
{
  SH2movls4(sh, 2,10, 8);
}
#define init_1_2_10_9
void FASTCALL call_1_2_10_9(SH2_struct* sh)
{
  SH2movls4(sh, 2,10, 9);
}
#define init_1_2_10_10
void FASTCALL call_1_2_10_10(SH2_struct* sh)
{
  SH2movls4(sh, 2,10, 10);
}
#define init_1_2_10_11
void FASTCALL call_1_2_10_11(SH2_struct* sh)
{
  SH2movls4(sh, 2,10, 11);
}
#define init_1_2_10_12
void FASTCALL call_1_2_10_12(SH2_struct* sh)
{
  SH2movls4(sh, 2,10, 12);
}
#define init_1_2_10_13
void FASTCALL call_1_2_10_13(SH2_struct* sh)
{
  SH2movls4(sh, 2,10, 13);
}
#define init_1_2_10_14
void FASTCALL call_1_2_10_14(SH2_struct* sh)
{
  SH2movls4(sh, 2,10, 14);
}
#define init_1_2_10_15
void FASTCALL call_1_2_10_15(SH2_struct* sh)
{
  SH2movls4(sh, 2,10, 15);
}
#define init_1_2_11_0
void FASTCALL call_1_2_11_0(SH2_struct* sh)
{
  SH2movls4(sh, 2,11, 0);
}
#define init_1_2_11_1
void FASTCALL call_1_2_11_1(SH2_struct* sh)
{
  SH2movls4(sh, 2,11, 1);
}
#define init_1_2_11_2
void FASTCALL call_1_2_11_2(SH2_struct* sh)
{
  SH2movls4(sh, 2,11, 2);
}
#define init_1_2_11_3
void FASTCALL call_1_2_11_3(SH2_struct* sh)
{
  SH2movls4(sh, 2,11, 3);
}
#define init_1_2_11_4
void FASTCALL call_1_2_11_4(SH2_struct* sh)
{
  SH2movls4(sh, 2,11, 4);
}
#define init_1_2_11_5
void FASTCALL call_1_2_11_5(SH2_struct* sh)
{
  SH2movls4(sh, 2,11, 5);
}
#define init_1_2_11_6
void FASTCALL call_1_2_11_6(SH2_struct* sh)
{
  SH2movls4(sh, 2,11, 6);
}
#define init_1_2_11_7
void FASTCALL call_1_2_11_7(SH2_struct* sh)
{
  SH2movls4(sh, 2,11, 7);
}
#define init_1_2_11_8
void FASTCALL call_1_2_11_8(SH2_struct* sh)
{
  SH2movls4(sh, 2,11, 8);
}
#define init_1_2_11_9
void FASTCALL call_1_2_11_9(SH2_struct* sh)
{
  SH2movls4(sh, 2,11, 9);
}
#define init_1_2_11_10
void FASTCALL call_1_2_11_10(SH2_struct* sh)
{
  SH2movls4(sh, 2,11, 10);
}
#define init_1_2_11_11
void FASTCALL call_1_2_11_11(SH2_struct* sh)
{
  SH2movls4(sh, 2,11, 11);
}
#define init_1_2_11_12
void FASTCALL call_1_2_11_12(SH2_struct* sh)
{
  SH2movls4(sh, 2,11, 12);
}
#define init_1_2_11_13
void FASTCALL call_1_2_11_13(SH2_struct* sh)
{
  SH2movls4(sh, 2,11, 13);
}
#define init_1_2_11_14
void FASTCALL call_1_2_11_14(SH2_struct* sh)
{
  SH2movls4(sh, 2,11, 14);
}
#define init_1_2_11_15
void FASTCALL call_1_2_11_15(SH2_struct* sh)
{
  SH2movls4(sh, 2,11, 15);
}
#define init_1_2_12_0
void FASTCALL call_1_2_12_0(SH2_struct* sh)
{
  SH2movls4(sh, 2,12, 0);
}
#define init_1_2_12_1
void FASTCALL call_1_2_12_1(SH2_struct* sh)
{
  SH2movls4(sh, 2,12, 1);
}
#define init_1_2_12_2
void FASTCALL call_1_2_12_2(SH2_struct* sh)
{
  SH2movls4(sh, 2,12, 2);
}
#define init_1_2_12_3
void FASTCALL call_1_2_12_3(SH2_struct* sh)
{
  SH2movls4(sh, 2,12, 3);
}
#define init_1_2_12_4
void FASTCALL call_1_2_12_4(SH2_struct* sh)
{
  SH2movls4(sh, 2,12, 4);
}
#define init_1_2_12_5
void FASTCALL call_1_2_12_5(SH2_struct* sh)
{
  SH2movls4(sh, 2,12, 5);
}
#define init_1_2_12_6
void FASTCALL call_1_2_12_6(SH2_struct* sh)
{
  SH2movls4(sh, 2,12, 6);
}
#define init_1_2_12_7
void FASTCALL call_1_2_12_7(SH2_struct* sh)
{
  SH2movls4(sh, 2,12, 7);
}
#define init_1_2_12_8
void FASTCALL call_1_2_12_8(SH2_struct* sh)
{
  SH2movls4(sh, 2,12, 8);
}
#define init_1_2_12_9
void FASTCALL call_1_2_12_9(SH2_struct* sh)
{
  SH2movls4(sh, 2,12, 9);
}
#define init_1_2_12_10
void FASTCALL call_1_2_12_10(SH2_struct* sh)
{
  SH2movls4(sh, 2,12, 10);
}
#define init_1_2_12_11
void FASTCALL call_1_2_12_11(SH2_struct* sh)
{
  SH2movls4(sh, 2,12, 11);
}
#define init_1_2_12_12
void FASTCALL call_1_2_12_12(SH2_struct* sh)
{
  SH2movls4(sh, 2,12, 12);
}
#define init_1_2_12_13
void FASTCALL call_1_2_12_13(SH2_struct* sh)
{
  SH2movls4(sh, 2,12, 13);
}
#define init_1_2_12_14
void FASTCALL call_1_2_12_14(SH2_struct* sh)
{
  SH2movls4(sh, 2,12, 14);
}
#define init_1_2_12_15
void FASTCALL call_1_2_12_15(SH2_struct* sh)
{
  SH2movls4(sh, 2,12, 15);
}
#define init_1_2_13_0
void FASTCALL call_1_2_13_0(SH2_struct* sh)
{
  SH2movls4(sh, 2,13, 0);
}
#define init_1_2_13_1
void FASTCALL call_1_2_13_1(SH2_struct* sh)
{
  SH2movls4(sh, 2,13, 1);
}
#define init_1_2_13_2
void FASTCALL call_1_2_13_2(SH2_struct* sh)
{
  SH2movls4(sh, 2,13, 2);
}
#define init_1_2_13_3
void FASTCALL call_1_2_13_3(SH2_struct* sh)
{
  SH2movls4(sh, 2,13, 3);
}
#define init_1_2_13_4
void FASTCALL call_1_2_13_4(SH2_struct* sh)
{
  SH2movls4(sh, 2,13, 4);
}
#define init_1_2_13_5
void FASTCALL call_1_2_13_5(SH2_struct* sh)
{
  SH2movls4(sh, 2,13, 5);
}
#define init_1_2_13_6
void FASTCALL call_1_2_13_6(SH2_struct* sh)
{
  SH2movls4(sh, 2,13, 6);
}
#define init_1_2_13_7
void FASTCALL call_1_2_13_7(SH2_struct* sh)
{
  SH2movls4(sh, 2,13, 7);
}
#define init_1_2_13_8
void FASTCALL call_1_2_13_8(SH2_struct* sh)
{
  SH2movls4(sh, 2,13, 8);
}
#define init_1_2_13_9
void FASTCALL call_1_2_13_9(SH2_struct* sh)
{
  SH2movls4(sh, 2,13, 9);
}
#define init_1_2_13_10
void FASTCALL call_1_2_13_10(SH2_struct* sh)
{
  SH2movls4(sh, 2,13, 10);
}
#define init_1_2_13_11
void FASTCALL call_1_2_13_11(SH2_struct* sh)
{
  SH2movls4(sh, 2,13, 11);
}
#define init_1_2_13_12
void FASTCALL call_1_2_13_12(SH2_struct* sh)
{
  SH2movls4(sh, 2,13, 12);
}
#define init_1_2_13_13
void FASTCALL call_1_2_13_13(SH2_struct* sh)
{
  SH2movls4(sh, 2,13, 13);
}
#define init_1_2_13_14
void FASTCALL call_1_2_13_14(SH2_struct* sh)
{
  SH2movls4(sh, 2,13, 14);
}
#define init_1_2_13_15
void FASTCALL call_1_2_13_15(SH2_struct* sh)
{
  SH2movls4(sh, 2,13, 15);
}
#define init_1_2_14_0
void FASTCALL call_1_2_14_0(SH2_struct* sh)
{
  SH2movls4(sh, 2,14, 0);
}
#define init_1_2_14_1
void FASTCALL call_1_2_14_1(SH2_struct* sh)
{
  SH2movls4(sh, 2,14, 1);
}
#define init_1_2_14_2
void FASTCALL call_1_2_14_2(SH2_struct* sh)
{
  SH2movls4(sh, 2,14, 2);
}
#define init_1_2_14_3
void FASTCALL call_1_2_14_3(SH2_struct* sh)
{
  SH2movls4(sh, 2,14, 3);
}
#define init_1_2_14_4
void FASTCALL call_1_2_14_4(SH2_struct* sh)
{
  SH2movls4(sh, 2,14, 4);
}
#define init_1_2_14_5
void FASTCALL call_1_2_14_5(SH2_struct* sh)
{
  SH2movls4(sh, 2,14, 5);
}
#define init_1_2_14_6
void FASTCALL call_1_2_14_6(SH2_struct* sh)
{
  SH2movls4(sh, 2,14, 6);
}
#define init_1_2_14_7
void FASTCALL call_1_2_14_7(SH2_struct* sh)
{
  SH2movls4(sh, 2,14, 7);
}
#define init_1_2_14_8
void FASTCALL call_1_2_14_8(SH2_struct* sh)
{
  SH2movls4(sh, 2,14, 8);
}
#define init_1_2_14_9
void FASTCALL call_1_2_14_9(SH2_struct* sh)
{
  SH2movls4(sh, 2,14, 9);
}
#define init_1_2_14_10
void FASTCALL call_1_2_14_10(SH2_struct* sh)
{
  SH2movls4(sh, 2,14, 10);
}
#define init_1_2_14_11
void FASTCALL call_1_2_14_11(SH2_struct* sh)
{
  SH2movls4(sh, 2,14, 11);
}
#define init_1_2_14_12
void FASTCALL call_1_2_14_12(SH2_struct* sh)
{
  SH2movls4(sh, 2,14, 12);
}
#define init_1_2_14_13
void FASTCALL call_1_2_14_13(SH2_struct* sh)
{
  SH2movls4(sh, 2,14, 13);
}
#define init_1_2_14_14
void FASTCALL call_1_2_14_14(SH2_struct* sh)
{
  SH2movls4(sh, 2,14, 14);
}
#define init_1_2_14_15
void FASTCALL call_1_2_14_15(SH2_struct* sh)
{
  SH2movls4(sh, 2,14, 15);
}
#define init_1_2_15_0
void FASTCALL call_1_2_15_0(SH2_struct* sh)
{
  SH2movls4(sh, 2,15, 0);
}
#define init_1_2_15_1
void FASTCALL call_1_2_15_1(SH2_struct* sh)
{
  SH2movls4(sh, 2,15, 1);
}
#define init_1_2_15_2
void FASTCALL call_1_2_15_2(SH2_struct* sh)
{
  SH2movls4(sh, 2,15, 2);
}
#define init_1_2_15_3
void FASTCALL call_1_2_15_3(SH2_struct* sh)
{
  SH2movls4(sh, 2,15, 3);
}
#define init_1_2_15_4
void FASTCALL call_1_2_15_4(SH2_struct* sh)
{
  SH2movls4(sh, 2,15, 4);
}
#define init_1_2_15_5
void FASTCALL call_1_2_15_5(SH2_struct* sh)
{
  SH2movls4(sh, 2,15, 5);
}
#define init_1_2_15_6
void FASTCALL call_1_2_15_6(SH2_struct* sh)
{
  SH2movls4(sh, 2,15, 6);
}
#define init_1_2_15_7
void FASTCALL call_1_2_15_7(SH2_struct* sh)
{
  SH2movls4(sh, 2,15, 7);
}
#define init_1_2_15_8
void FASTCALL call_1_2_15_8(SH2_struct* sh)
{
  SH2movls4(sh, 2,15, 8);
}
#define init_1_2_15_9
void FASTCALL call_1_2_15_9(SH2_struct* sh)
{
  SH2movls4(sh, 2,15, 9);
}
#define init_1_2_15_10
void FASTCALL call_1_2_15_10(SH2_struct* sh)
{
  SH2movls4(sh, 2,15, 10);
}
#define init_1_2_15_11
void FASTCALL call_1_2_15_11(SH2_struct* sh)
{
  SH2movls4(sh, 2,15, 11);
}
#define init_1_2_15_12
void FASTCALL call_1_2_15_12(SH2_struct* sh)
{
  SH2movls4(sh, 2,15, 12);
}
#define init_1_2_15_13
void FASTCALL call_1_2_15_13(SH2_struct* sh)
{
  SH2movls4(sh, 2,15, 13);
}
#define init_1_2_15_14
void FASTCALL call_1_2_15_14(SH2_struct* sh)
{
  SH2movls4(sh, 2,15, 14);
}
#define init_1_2_15_15
void FASTCALL call_1_2_15_15(SH2_struct* sh)
{
  SH2movls4(sh, 2,15, 15);
}
#define init_1_3_0_0
void FASTCALL call_1_3_0_0(SH2_struct* sh)
{
  SH2movls4(sh, 3,0, 0);
}
#define init_1_3_0_1
void FASTCALL call_1_3_0_1(SH2_struct* sh)
{
  SH2movls4(sh, 3,0, 1);
}
#define init_1_3_0_2
void FASTCALL call_1_3_0_2(SH2_struct* sh)
{
  SH2movls4(sh, 3,0, 2);
}
#define init_1_3_0_3
void FASTCALL call_1_3_0_3(SH2_struct* sh)
{
  SH2movls4(sh, 3,0, 3);
}
#define init_1_3_0_4
void FASTCALL call_1_3_0_4(SH2_struct* sh)
{
  SH2movls4(sh, 3,0, 4);
}
#define init_1_3_0_5
void FASTCALL call_1_3_0_5(SH2_struct* sh)
{
  SH2movls4(sh, 3,0, 5);
}
#define init_1_3_0_6
void FASTCALL call_1_3_0_6(SH2_struct* sh)
{
  SH2movls4(sh, 3,0, 6);
}
#define init_1_3_0_7
void FASTCALL call_1_3_0_7(SH2_struct* sh)
{
  SH2movls4(sh, 3,0, 7);
}
#define init_1_3_0_8
void FASTCALL call_1_3_0_8(SH2_struct* sh)
{
  SH2movls4(sh, 3,0, 8);
}
#define init_1_3_0_9
void FASTCALL call_1_3_0_9(SH2_struct* sh)
{
  SH2movls4(sh, 3,0, 9);
}
#define init_1_3_0_10
void FASTCALL call_1_3_0_10(SH2_struct* sh)
{
  SH2movls4(sh, 3,0, 10);
}
#define init_1_3_0_11
void FASTCALL call_1_3_0_11(SH2_struct* sh)
{
  SH2movls4(sh, 3,0, 11);
}
#define init_1_3_0_12
void FASTCALL call_1_3_0_12(SH2_struct* sh)
{
  SH2movls4(sh, 3,0, 12);
}
#define init_1_3_0_13
void FASTCALL call_1_3_0_13(SH2_struct* sh)
{
  SH2movls4(sh, 3,0, 13);
}
#define init_1_3_0_14
void FASTCALL call_1_3_0_14(SH2_struct* sh)
{
  SH2movls4(sh, 3,0, 14);
}
#define init_1_3_0_15
void FASTCALL call_1_3_0_15(SH2_struct* sh)
{
  SH2movls4(sh, 3,0, 15);
}
#define init_1_3_1_0
void FASTCALL call_1_3_1_0(SH2_struct* sh)
{
  SH2movls4(sh, 3,1, 0);
}
#define init_1_3_1_1
void FASTCALL call_1_3_1_1(SH2_struct* sh)
{
  SH2movls4(sh, 3,1, 1);
}
#define init_1_3_1_2
void FASTCALL call_1_3_1_2(SH2_struct* sh)
{
  SH2movls4(sh, 3,1, 2);
}
#define init_1_3_1_3
void FASTCALL call_1_3_1_3(SH2_struct* sh)
{
  SH2movls4(sh, 3,1, 3);
}
#define init_1_3_1_4
void FASTCALL call_1_3_1_4(SH2_struct* sh)
{
  SH2movls4(sh, 3,1, 4);
}
#define init_1_3_1_5
void FASTCALL call_1_3_1_5(SH2_struct* sh)
{
  SH2movls4(sh, 3,1, 5);
}
#define init_1_3_1_6
void FASTCALL call_1_3_1_6(SH2_struct* sh)
{
  SH2movls4(sh, 3,1, 6);
}
#define init_1_3_1_7
void FASTCALL call_1_3_1_7(SH2_struct* sh)
{
  SH2movls4(sh, 3,1, 7);
}
#define init_1_3_1_8
void FASTCALL call_1_3_1_8(SH2_struct* sh)
{
  SH2movls4(sh, 3,1, 8);
}
#define init_1_3_1_9
void FASTCALL call_1_3_1_9(SH2_struct* sh)
{
  SH2movls4(sh, 3,1, 9);
}
#define init_1_3_1_10
void FASTCALL call_1_3_1_10(SH2_struct* sh)
{
  SH2movls4(sh, 3,1, 10);
}
#define init_1_3_1_11
void FASTCALL call_1_3_1_11(SH2_struct* sh)
{
  SH2movls4(sh, 3,1, 11);
}
#define init_1_3_1_12
void FASTCALL call_1_3_1_12(SH2_struct* sh)
{
  SH2movls4(sh, 3,1, 12);
}
#define init_1_3_1_13
void FASTCALL call_1_3_1_13(SH2_struct* sh)
{
  SH2movls4(sh, 3,1, 13);
}
#define init_1_3_1_14
void FASTCALL call_1_3_1_14(SH2_struct* sh)
{
  SH2movls4(sh, 3,1, 14);
}
#define init_1_3_1_15
void FASTCALL call_1_3_1_15(SH2_struct* sh)
{
  SH2movls4(sh, 3,1, 15);
}
#define init_1_3_2_0
void FASTCALL call_1_3_2_0(SH2_struct* sh)
{
  SH2movls4(sh, 3,2, 0);
}
#define init_1_3_2_1
void FASTCALL call_1_3_2_1(SH2_struct* sh)
{
  SH2movls4(sh, 3,2, 1);
}
#define init_1_3_2_2
void FASTCALL call_1_3_2_2(SH2_struct* sh)
{
  SH2movls4(sh, 3,2, 2);
}
#define init_1_3_2_3
void FASTCALL call_1_3_2_3(SH2_struct* sh)
{
  SH2movls4(sh, 3,2, 3);
}
#define init_1_3_2_4
void FASTCALL call_1_3_2_4(SH2_struct* sh)
{
  SH2movls4(sh, 3,2, 4);
}
#define init_1_3_2_5
void FASTCALL call_1_3_2_5(SH2_struct* sh)
{
  SH2movls4(sh, 3,2, 5);
}
#define init_1_3_2_6
void FASTCALL call_1_3_2_6(SH2_struct* sh)
{
  SH2movls4(sh, 3,2, 6);
}
#define init_1_3_2_7
void FASTCALL call_1_3_2_7(SH2_struct* sh)
{
  SH2movls4(sh, 3,2, 7);
}
#define init_1_3_2_8
void FASTCALL call_1_3_2_8(SH2_struct* sh)
{
  SH2movls4(sh, 3,2, 8);
}
#define init_1_3_2_9
void FASTCALL call_1_3_2_9(SH2_struct* sh)
{
  SH2movls4(sh, 3,2, 9);
}
#define init_1_3_2_10
void FASTCALL call_1_3_2_10(SH2_struct* sh)
{
  SH2movls4(sh, 3,2, 10);
}
#define init_1_3_2_11
void FASTCALL call_1_3_2_11(SH2_struct* sh)
{
  SH2movls4(sh, 3,2, 11);
}
#define init_1_3_2_12
void FASTCALL call_1_3_2_12(SH2_struct* sh)
{
  SH2movls4(sh, 3,2, 12);
}
#define init_1_3_2_13
void FASTCALL call_1_3_2_13(SH2_struct* sh)
{
  SH2movls4(sh, 3,2, 13);
}
#define init_1_3_2_14
void FASTCALL call_1_3_2_14(SH2_struct* sh)
{
  SH2movls4(sh, 3,2, 14);
}
#define init_1_3_2_15
void FASTCALL call_1_3_2_15(SH2_struct* sh)
{
  SH2movls4(sh, 3,2, 15);
}
#define init_1_3_3_0
void FASTCALL call_1_3_3_0(SH2_struct* sh)
{
  SH2movls4(sh, 3,3, 0);
}
#define init_1_3_3_1
void FASTCALL call_1_3_3_1(SH2_struct* sh)
{
  SH2movls4(sh, 3,3, 1);
}
#define init_1_3_3_2
void FASTCALL call_1_3_3_2(SH2_struct* sh)
{
  SH2movls4(sh, 3,3, 2);
}
#define init_1_3_3_3
void FASTCALL call_1_3_3_3(SH2_struct* sh)
{
  SH2movls4(sh, 3,3, 3);
}
#define init_1_3_3_4
void FASTCALL call_1_3_3_4(SH2_struct* sh)
{
  SH2movls4(sh, 3,3, 4);
}
#define init_1_3_3_5
void FASTCALL call_1_3_3_5(SH2_struct* sh)
{
  SH2movls4(sh, 3,3, 5);
}
#define init_1_3_3_6
void FASTCALL call_1_3_3_6(SH2_struct* sh)
{
  SH2movls4(sh, 3,3, 6);
}
#define init_1_3_3_7
void FASTCALL call_1_3_3_7(SH2_struct* sh)
{
  SH2movls4(sh, 3,3, 7);
}
#define init_1_3_3_8
void FASTCALL call_1_3_3_8(SH2_struct* sh)
{
  SH2movls4(sh, 3,3, 8);
}
#define init_1_3_3_9
void FASTCALL call_1_3_3_9(SH2_struct* sh)
{
  SH2movls4(sh, 3,3, 9);
}
#define init_1_3_3_10
void FASTCALL call_1_3_3_10(SH2_struct* sh)
{
  SH2movls4(sh, 3,3, 10);
}
#define init_1_3_3_11
void FASTCALL call_1_3_3_11(SH2_struct* sh)
{
  SH2movls4(sh, 3,3, 11);
}
#define init_1_3_3_12
void FASTCALL call_1_3_3_12(SH2_struct* sh)
{
  SH2movls4(sh, 3,3, 12);
}
#define init_1_3_3_13
void FASTCALL call_1_3_3_13(SH2_struct* sh)
{
  SH2movls4(sh, 3,3, 13);
}
#define init_1_3_3_14
void FASTCALL call_1_3_3_14(SH2_struct* sh)
{
  SH2movls4(sh, 3,3, 14);
}
#define init_1_3_3_15
void FASTCALL call_1_3_3_15(SH2_struct* sh)
{
  SH2movls4(sh, 3,3, 15);
}
#define init_1_3_4_0
void FASTCALL call_1_3_4_0(SH2_struct* sh)
{
  SH2movls4(sh, 3,4, 0);
}
#define init_1_3_4_1
void FASTCALL call_1_3_4_1(SH2_struct* sh)
{
  SH2movls4(sh, 3,4, 1);
}
#define init_1_3_4_2
void FASTCALL call_1_3_4_2(SH2_struct* sh)
{
  SH2movls4(sh, 3,4, 2);
}
#define init_1_3_4_3
void FASTCALL call_1_3_4_3(SH2_struct* sh)
{
  SH2movls4(sh, 3,4, 3);
}
#define init_1_3_4_4
void FASTCALL call_1_3_4_4(SH2_struct* sh)
{
  SH2movls4(sh, 3,4, 4);
}
#define init_1_3_4_5
void FASTCALL call_1_3_4_5(SH2_struct* sh)
{
  SH2movls4(sh, 3,4, 5);
}
#define init_1_3_4_6
void FASTCALL call_1_3_4_6(SH2_struct* sh)
{
  SH2movls4(sh, 3,4, 6);
}
#define init_1_3_4_7
void FASTCALL call_1_3_4_7(SH2_struct* sh)
{
  SH2movls4(sh, 3,4, 7);
}
#define init_1_3_4_8
void FASTCALL call_1_3_4_8(SH2_struct* sh)
{
  SH2movls4(sh, 3,4, 8);
}
#define init_1_3_4_9
void FASTCALL call_1_3_4_9(SH2_struct* sh)
{
  SH2movls4(sh, 3,4, 9);
}
#define init_1_3_4_10
void FASTCALL call_1_3_4_10(SH2_struct* sh)
{
  SH2movls4(sh, 3,4, 10);
}
#define init_1_3_4_11
void FASTCALL call_1_3_4_11(SH2_struct* sh)
{
  SH2movls4(sh, 3,4, 11);
}
#define init_1_3_4_12
void FASTCALL call_1_3_4_12(SH2_struct* sh)
{
  SH2movls4(sh, 3,4, 12);
}
#define init_1_3_4_13
void FASTCALL call_1_3_4_13(SH2_struct* sh)
{
  SH2movls4(sh, 3,4, 13);
}
#define init_1_3_4_14
void FASTCALL call_1_3_4_14(SH2_struct* sh)
{
  SH2movls4(sh, 3,4, 14);
}
#define init_1_3_4_15
void FASTCALL call_1_3_4_15(SH2_struct* sh)
{
  SH2movls4(sh, 3,4, 15);
}
#define init_1_3_5_0
void FASTCALL call_1_3_5_0(SH2_struct* sh)
{
  SH2movls4(sh, 3,5, 0);
}
#define init_1_3_5_1
void FASTCALL call_1_3_5_1(SH2_struct* sh)
{
  SH2movls4(sh, 3,5, 1);
}
#define init_1_3_5_2
void FASTCALL call_1_3_5_2(SH2_struct* sh)
{
  SH2movls4(sh, 3,5, 2);
}
#define init_1_3_5_3
void FASTCALL call_1_3_5_3(SH2_struct* sh)
{
  SH2movls4(sh, 3,5, 3);
}
#define init_1_3_5_4
void FASTCALL call_1_3_5_4(SH2_struct* sh)
{
  SH2movls4(sh, 3,5, 4);
}
#define init_1_3_5_5
void FASTCALL call_1_3_5_5(SH2_struct* sh)
{
  SH2movls4(sh, 3,5, 5);
}
#define init_1_3_5_6
void FASTCALL call_1_3_5_6(SH2_struct* sh)
{
  SH2movls4(sh, 3,5, 6);
}
#define init_1_3_5_7
void FASTCALL call_1_3_5_7(SH2_struct* sh)
{
  SH2movls4(sh, 3,5, 7);
}
#define init_1_3_5_8
void FASTCALL call_1_3_5_8(SH2_struct* sh)
{
  SH2movls4(sh, 3,5, 8);
}
#define init_1_3_5_9
void FASTCALL call_1_3_5_9(SH2_struct* sh)
{
  SH2movls4(sh, 3,5, 9);
}
#define init_1_3_5_10
void FASTCALL call_1_3_5_10(SH2_struct* sh)
{
  SH2movls4(sh, 3,5, 10);
}
#define init_1_3_5_11
void FASTCALL call_1_3_5_11(SH2_struct* sh)
{
  SH2movls4(sh, 3,5, 11);
}
#define init_1_3_5_12
void FASTCALL call_1_3_5_12(SH2_struct* sh)
{
  SH2movls4(sh, 3,5, 12);
}
#define init_1_3_5_13
void FASTCALL call_1_3_5_13(SH2_struct* sh)
{
  SH2movls4(sh, 3,5, 13);
}
#define init_1_3_5_14
void FASTCALL call_1_3_5_14(SH2_struct* sh)
{
  SH2movls4(sh, 3,5, 14);
}
#define init_1_3_5_15
void FASTCALL call_1_3_5_15(SH2_struct* sh)
{
  SH2movls4(sh, 3,5, 15);
}
#define init_1_3_6_0
void FASTCALL call_1_3_6_0(SH2_struct* sh)
{
  SH2movls4(sh, 3,6, 0);
}
#define init_1_3_6_1
void FASTCALL call_1_3_6_1(SH2_struct* sh)
{
  SH2movls4(sh, 3,6, 1);
}
#define init_1_3_6_2
void FASTCALL call_1_3_6_2(SH2_struct* sh)
{
  SH2movls4(sh, 3,6, 2);
}
#define init_1_3_6_3
void FASTCALL call_1_3_6_3(SH2_struct* sh)
{
  SH2movls4(sh, 3,6, 3);
}
#define init_1_3_6_4
void FASTCALL call_1_3_6_4(SH2_struct* sh)
{
  SH2movls4(sh, 3,6, 4);
}
#define init_1_3_6_5
void FASTCALL call_1_3_6_5(SH2_struct* sh)
{
  SH2movls4(sh, 3,6, 5);
}
#define init_1_3_6_6
void FASTCALL call_1_3_6_6(SH2_struct* sh)
{
  SH2movls4(sh, 3,6, 6);
}
#define init_1_3_6_7
void FASTCALL call_1_3_6_7(SH2_struct* sh)
{
  SH2movls4(sh, 3,6, 7);
}
#define init_1_3_6_8
void FASTCALL call_1_3_6_8(SH2_struct* sh)
{
  SH2movls4(sh, 3,6, 8);
}
#define init_1_3_6_9
void FASTCALL call_1_3_6_9(SH2_struct* sh)
{
  SH2movls4(sh, 3,6, 9);
}
#define init_1_3_6_10
void FASTCALL call_1_3_6_10(SH2_struct* sh)
{
  SH2movls4(sh, 3,6, 10);
}
#define init_1_3_6_11
void FASTCALL call_1_3_6_11(SH2_struct* sh)
{
  SH2movls4(sh, 3,6, 11);
}
#define init_1_3_6_12
void FASTCALL call_1_3_6_12(SH2_struct* sh)
{
  SH2movls4(sh, 3,6, 12);
}
#define init_1_3_6_13
void FASTCALL call_1_3_6_13(SH2_struct* sh)
{
  SH2movls4(sh, 3,6, 13);
}
#define init_1_3_6_14
void FASTCALL call_1_3_6_14(SH2_struct* sh)
{
  SH2movls4(sh, 3,6, 14);
}
#define init_1_3_6_15
void FASTCALL call_1_3_6_15(SH2_struct* sh)
{
  SH2movls4(sh, 3,6, 15);
}
#define init_1_3_7_0
void FASTCALL call_1_3_7_0(SH2_struct* sh)
{
  SH2movls4(sh, 3,7, 0);
}
#define init_1_3_7_1
void FASTCALL call_1_3_7_1(SH2_struct* sh)
{
  SH2movls4(sh, 3,7, 1);
}
#define init_1_3_7_2
void FASTCALL call_1_3_7_2(SH2_struct* sh)
{
  SH2movls4(sh, 3,7, 2);
}
#define init_1_3_7_3
void FASTCALL call_1_3_7_3(SH2_struct* sh)
{
  SH2movls4(sh, 3,7, 3);
}
#define init_1_3_7_4
void FASTCALL call_1_3_7_4(SH2_struct* sh)
{
  SH2movls4(sh, 3,7, 4);
}
#define init_1_3_7_5
void FASTCALL call_1_3_7_5(SH2_struct* sh)
{
  SH2movls4(sh, 3,7, 5);
}
#define init_1_3_7_6
void FASTCALL call_1_3_7_6(SH2_struct* sh)
{
  SH2movls4(sh, 3,7, 6);
}
#define init_1_3_7_7
void FASTCALL call_1_3_7_7(SH2_struct* sh)
{
  SH2movls4(sh, 3,7, 7);
}
#define init_1_3_7_8
void FASTCALL call_1_3_7_8(SH2_struct* sh)
{
  SH2movls4(sh, 3,7, 8);
}
#define init_1_3_7_9
void FASTCALL call_1_3_7_9(SH2_struct* sh)
{
  SH2movls4(sh, 3,7, 9);
}
#define init_1_3_7_10
void FASTCALL call_1_3_7_10(SH2_struct* sh)
{
  SH2movls4(sh, 3,7, 10);
}
#define init_1_3_7_11
void FASTCALL call_1_3_7_11(SH2_struct* sh)
{
  SH2movls4(sh, 3,7, 11);
}
#define init_1_3_7_12
void FASTCALL call_1_3_7_12(SH2_struct* sh)
{
  SH2movls4(sh, 3,7, 12);
}
#define init_1_3_7_13
void FASTCALL call_1_3_7_13(SH2_struct* sh)
{
  SH2movls4(sh, 3,7, 13);
}
#define init_1_3_7_14
void FASTCALL call_1_3_7_14(SH2_struct* sh)
{
  SH2movls4(sh, 3,7, 14);
}
#define init_1_3_7_15
void FASTCALL call_1_3_7_15(SH2_struct* sh)
{
  SH2movls4(sh, 3,7, 15);
}
#define init_1_3_8_0
void FASTCALL call_1_3_8_0(SH2_struct* sh)
{
  SH2movls4(sh, 3,8, 0);
}
#define init_1_3_8_1
void FASTCALL call_1_3_8_1(SH2_struct* sh)
{
  SH2movls4(sh, 3,8, 1);
}
#define init_1_3_8_2
void FASTCALL call_1_3_8_2(SH2_struct* sh)
{
  SH2movls4(sh, 3,8, 2);
}
#define init_1_3_8_3
void FASTCALL call_1_3_8_3(SH2_struct* sh)
{
  SH2movls4(sh, 3,8, 3);
}
#define init_1_3_8_4
void FASTCALL call_1_3_8_4(SH2_struct* sh)
{
  SH2movls4(sh, 3,8, 4);
}
#define init_1_3_8_5
void FASTCALL call_1_3_8_5(SH2_struct* sh)
{
  SH2movls4(sh, 3,8, 5);
}
#define init_1_3_8_6
void FASTCALL call_1_3_8_6(SH2_struct* sh)
{
  SH2movls4(sh, 3,8, 6);
}
#define init_1_3_8_7
void FASTCALL call_1_3_8_7(SH2_struct* sh)
{
  SH2movls4(sh, 3,8, 7);
}
#define init_1_3_8_8
void FASTCALL call_1_3_8_8(SH2_struct* sh)
{
  SH2movls4(sh, 3,8, 8);
}
#define init_1_3_8_9
void FASTCALL call_1_3_8_9(SH2_struct* sh)
{
  SH2movls4(sh, 3,8, 9);
}
#define init_1_3_8_10
void FASTCALL call_1_3_8_10(SH2_struct* sh)
{
  SH2movls4(sh, 3,8, 10);
}
#define init_1_3_8_11
void FASTCALL call_1_3_8_11(SH2_struct* sh)
{
  SH2movls4(sh, 3,8, 11);
}
#define init_1_3_8_12
void FASTCALL call_1_3_8_12(SH2_struct* sh)
{
  SH2movls4(sh, 3,8, 12);
}
#define init_1_3_8_13
void FASTCALL call_1_3_8_13(SH2_struct* sh)
{
  SH2movls4(sh, 3,8, 13);
}
#define init_1_3_8_14
void FASTCALL call_1_3_8_14(SH2_struct* sh)
{
  SH2movls4(sh, 3,8, 14);
}
#define init_1_3_8_15
void FASTCALL call_1_3_8_15(SH2_struct* sh)
{
  SH2movls4(sh, 3,8, 15);
}
#define init_1_3_9_0
void FASTCALL call_1_3_9_0(SH2_struct* sh)
{
  SH2movls4(sh, 3,9, 0);
}
#define init_1_3_9_1
void FASTCALL call_1_3_9_1(SH2_struct* sh)
{
  SH2movls4(sh, 3,9, 1);
}
#define init_1_3_9_2
void FASTCALL call_1_3_9_2(SH2_struct* sh)
{
  SH2movls4(sh, 3,9, 2);
}
#define init_1_3_9_3
void FASTCALL call_1_3_9_3(SH2_struct* sh)
{
  SH2movls4(sh, 3,9, 3);
}
#define init_1_3_9_4
void FASTCALL call_1_3_9_4(SH2_struct* sh)
{
  SH2movls4(sh, 3,9, 4);
}
#define init_1_3_9_5
void FASTCALL call_1_3_9_5(SH2_struct* sh)
{
  SH2movls4(sh, 3,9, 5);
}
#define init_1_3_9_6
void FASTCALL call_1_3_9_6(SH2_struct* sh)
{
  SH2movls4(sh, 3,9, 6);
}
#define init_1_3_9_7
void FASTCALL call_1_3_9_7(SH2_struct* sh)
{
  SH2movls4(sh, 3,9, 7);
}
#define init_1_3_9_8
void FASTCALL call_1_3_9_8(SH2_struct* sh)
{
  SH2movls4(sh, 3,9, 8);
}
#define init_1_3_9_9
void FASTCALL call_1_3_9_9(SH2_struct* sh)
{
  SH2movls4(sh, 3,9, 9);
}
#define init_1_3_9_10
void FASTCALL call_1_3_9_10(SH2_struct* sh)
{
  SH2movls4(sh, 3,9, 10);
}
#define init_1_3_9_11
void FASTCALL call_1_3_9_11(SH2_struct* sh)
{
  SH2movls4(sh, 3,9, 11);
}
#define init_1_3_9_12
void FASTCALL call_1_3_9_12(SH2_struct* sh)
{
  SH2movls4(sh, 3,9, 12);
}
#define init_1_3_9_13
void FASTCALL call_1_3_9_13(SH2_struct* sh)
{
  SH2movls4(sh, 3,9, 13);
}
#define init_1_3_9_14
void FASTCALL call_1_3_9_14(SH2_struct* sh)
{
  SH2movls4(sh, 3,9, 14);
}
#define init_1_3_9_15
void FASTCALL call_1_3_9_15(SH2_struct* sh)
{
  SH2movls4(sh, 3,9, 15);
}
#define init_1_3_10_0
void FASTCALL call_1_3_10_0(SH2_struct* sh)
{
  SH2movls4(sh, 3,10, 0);
}
#define init_1_3_10_1
void FASTCALL call_1_3_10_1(SH2_struct* sh)
{
  SH2movls4(sh, 3,10, 1);
}
#define init_1_3_10_2
void FASTCALL call_1_3_10_2(SH2_struct* sh)
{
  SH2movls4(sh, 3,10, 2);
}
#define init_1_3_10_3
void FASTCALL call_1_3_10_3(SH2_struct* sh)
{
  SH2movls4(sh, 3,10, 3);
}
#define init_1_3_10_4
void FASTCALL call_1_3_10_4(SH2_struct* sh)
{
  SH2movls4(sh, 3,10, 4);
}
#define init_1_3_10_5
void FASTCALL call_1_3_10_5(SH2_struct* sh)
{
  SH2movls4(sh, 3,10, 5);
}
#define init_1_3_10_6
void FASTCALL call_1_3_10_6(SH2_struct* sh)
{
  SH2movls4(sh, 3,10, 6);
}
#define init_1_3_10_7
void FASTCALL call_1_3_10_7(SH2_struct* sh)
{
  SH2movls4(sh, 3,10, 7);
}
#define init_1_3_10_8
void FASTCALL call_1_3_10_8(SH2_struct* sh)
{
  SH2movls4(sh, 3,10, 8);
}
#define init_1_3_10_9
void FASTCALL call_1_3_10_9(SH2_struct* sh)
{
  SH2movls4(sh, 3,10, 9);
}
#define init_1_3_10_10
void FASTCALL call_1_3_10_10(SH2_struct* sh)
{
  SH2movls4(sh, 3,10, 10);
}
#define init_1_3_10_11
void FASTCALL call_1_3_10_11(SH2_struct* sh)
{
  SH2movls4(sh, 3,10, 11);
}
#define init_1_3_10_12
void FASTCALL call_1_3_10_12(SH2_struct* sh)
{
  SH2movls4(sh, 3,10, 12);
}
#define init_1_3_10_13
void FASTCALL call_1_3_10_13(SH2_struct* sh)
{
  SH2movls4(sh, 3,10, 13);
}
#define init_1_3_10_14
void FASTCALL call_1_3_10_14(SH2_struct* sh)
{
  SH2movls4(sh, 3,10, 14);
}
#define init_1_3_10_15
void FASTCALL call_1_3_10_15(SH2_struct* sh)
{
  SH2movls4(sh, 3,10, 15);
}
#define init_1_3_11_0
void FASTCALL call_1_3_11_0(SH2_struct* sh)
{
  SH2movls4(sh, 3,11, 0);
}
#define init_1_3_11_1
void FASTCALL call_1_3_11_1(SH2_struct* sh)
{
  SH2movls4(sh, 3,11, 1);
}
#define init_1_3_11_2
void FASTCALL call_1_3_11_2(SH2_struct* sh)
{
  SH2movls4(sh, 3,11, 2);
}
#define init_1_3_11_3
void FASTCALL call_1_3_11_3(SH2_struct* sh)
{
  SH2movls4(sh, 3,11, 3);
}
#define init_1_3_11_4
void FASTCALL call_1_3_11_4(SH2_struct* sh)
{
  SH2movls4(sh, 3,11, 4);
}
#define init_1_3_11_5
void FASTCALL call_1_3_11_5(SH2_struct* sh)
{
  SH2movls4(sh, 3,11, 5);
}
#define init_1_3_11_6
void FASTCALL call_1_3_11_6(SH2_struct* sh)
{
  SH2movls4(sh, 3,11, 6);
}
#define init_1_3_11_7
void FASTCALL call_1_3_11_7(SH2_struct* sh)
{
  SH2movls4(sh, 3,11, 7);
}
#define init_1_3_11_8
void FASTCALL call_1_3_11_8(SH2_struct* sh)
{
  SH2movls4(sh, 3,11, 8);
}
#define init_1_3_11_9
void FASTCALL call_1_3_11_9(SH2_struct* sh)
{
  SH2movls4(sh, 3,11, 9);
}
#define init_1_3_11_10
void FASTCALL call_1_3_11_10(SH2_struct* sh)
{
  SH2movls4(sh, 3,11, 10);
}
#define init_1_3_11_11
void FASTCALL call_1_3_11_11(SH2_struct* sh)
{
  SH2movls4(sh, 3,11, 11);
}
#define init_1_3_11_12
void FASTCALL call_1_3_11_12(SH2_struct* sh)
{
  SH2movls4(sh, 3,11, 12);
}
#define init_1_3_11_13
void FASTCALL call_1_3_11_13(SH2_struct* sh)
{
  SH2movls4(sh, 3,11, 13);
}
#define init_1_3_11_14
void FASTCALL call_1_3_11_14(SH2_struct* sh)
{
  SH2movls4(sh, 3,11, 14);
}
#define init_1_3_11_15
void FASTCALL call_1_3_11_15(SH2_struct* sh)
{
  SH2movls4(sh, 3,11, 15);
}
#define init_1_3_12_0
void FASTCALL call_1_3_12_0(SH2_struct* sh)
{
  SH2movls4(sh, 3,12, 0);
}
#define init_1_3_12_1
void FASTCALL call_1_3_12_1(SH2_struct* sh)
{
  SH2movls4(sh, 3,12, 1);
}
#define init_1_3_12_2
void FASTCALL call_1_3_12_2(SH2_struct* sh)
{
  SH2movls4(sh, 3,12, 2);
}
#define init_1_3_12_3
void FASTCALL call_1_3_12_3(SH2_struct* sh)
{
  SH2movls4(sh, 3,12, 3);
}
#define init_1_3_12_4
void FASTCALL call_1_3_12_4(SH2_struct* sh)
{
  SH2movls4(sh, 3,12, 4);
}
#define init_1_3_12_5
void FASTCALL call_1_3_12_5(SH2_struct* sh)
{
  SH2movls4(sh, 3,12, 5);
}
#define init_1_3_12_6
void FASTCALL call_1_3_12_6(SH2_struct* sh)
{
  SH2movls4(sh, 3,12, 6);
}
#define init_1_3_12_7
void FASTCALL call_1_3_12_7(SH2_struct* sh)
{
  SH2movls4(sh, 3,12, 7);
}
#define init_1_3_12_8
void FASTCALL call_1_3_12_8(SH2_struct* sh)
{
  SH2movls4(sh, 3,12, 8);
}
#define init_1_3_12_9
void FASTCALL call_1_3_12_9(SH2_struct* sh)
{
  SH2movls4(sh, 3,12, 9);
}
#define init_1_3_12_10
void FASTCALL call_1_3_12_10(SH2_struct* sh)
{
  SH2movls4(sh, 3,12, 10);
}
#define init_1_3_12_11
void FASTCALL call_1_3_12_11(SH2_struct* sh)
{
  SH2movls4(sh, 3,12, 11);
}
#define init_1_3_12_12
void FASTCALL call_1_3_12_12(SH2_struct* sh)
{
  SH2movls4(sh, 3,12, 12);
}
#define init_1_3_12_13
void FASTCALL call_1_3_12_13(SH2_struct* sh)
{
  SH2movls4(sh, 3,12, 13);
}
#define init_1_3_12_14
void FASTCALL call_1_3_12_14(SH2_struct* sh)
{
  SH2movls4(sh, 3,12, 14);
}
#define init_1_3_12_15
void FASTCALL call_1_3_12_15(SH2_struct* sh)
{
  SH2movls4(sh, 3,12, 15);
}
#define init_1_3_13_0
void FASTCALL call_1_3_13_0(SH2_struct* sh)
{
  SH2movls4(sh, 3,13, 0);
}
#define init_1_3_13_1
void FASTCALL call_1_3_13_1(SH2_struct* sh)
{
  SH2movls4(sh, 3,13, 1);
}
#define init_1_3_13_2
void FASTCALL call_1_3_13_2(SH2_struct* sh)
{
  SH2movls4(sh, 3,13, 2);
}
#define init_1_3_13_3
void FASTCALL call_1_3_13_3(SH2_struct* sh)
{
  SH2movls4(sh, 3,13, 3);
}
#define init_1_3_13_4
void FASTCALL call_1_3_13_4(SH2_struct* sh)
{
  SH2movls4(sh, 3,13, 4);
}
#define init_1_3_13_5
void FASTCALL call_1_3_13_5(SH2_struct* sh)
{
  SH2movls4(sh, 3,13, 5);
}
#define init_1_3_13_6
void FASTCALL call_1_3_13_6(SH2_struct* sh)
{
  SH2movls4(sh, 3,13, 6);
}
#define init_1_3_13_7
void FASTCALL call_1_3_13_7(SH2_struct* sh)
{
  SH2movls4(sh, 3,13, 7);
}
#define init_1_3_13_8
void FASTCALL call_1_3_13_8(SH2_struct* sh)
{
  SH2movls4(sh, 3,13, 8);
}
#define init_1_3_13_9
void FASTCALL call_1_3_13_9(SH2_struct* sh)
{
  SH2movls4(sh, 3,13, 9);
}
#define init_1_3_13_10
void FASTCALL call_1_3_13_10(SH2_struct* sh)
{
  SH2movls4(sh, 3,13, 10);
}
#define init_1_3_13_11
void FASTCALL call_1_3_13_11(SH2_struct* sh)
{
  SH2movls4(sh, 3,13, 11);
}
#define init_1_3_13_12
void FASTCALL call_1_3_13_12(SH2_struct* sh)
{
  SH2movls4(sh, 3,13, 12);
}
#define init_1_3_13_13
void FASTCALL call_1_3_13_13(SH2_struct* sh)
{
  SH2movls4(sh, 3,13, 13);
}
#define init_1_3_13_14
void FASTCALL call_1_3_13_14(SH2_struct* sh)
{
  SH2movls4(sh, 3,13, 14);
}
#define init_1_3_13_15
void FASTCALL call_1_3_13_15(SH2_struct* sh)
{
  SH2movls4(sh, 3,13, 15);
}
#define init_1_3_14_0
void FASTCALL call_1_3_14_0(SH2_struct* sh)
{
  SH2movls4(sh, 3,14, 0);
}
#define init_1_3_14_1
void FASTCALL call_1_3_14_1(SH2_struct* sh)
{
  SH2movls4(sh, 3,14, 1);
}
#define init_1_3_14_2
void FASTCALL call_1_3_14_2(SH2_struct* sh)
{
  SH2movls4(sh, 3,14, 2);
}
#define init_1_3_14_3
void FASTCALL call_1_3_14_3(SH2_struct* sh)
{
  SH2movls4(sh, 3,14, 3);
}
#define init_1_3_14_4
void FASTCALL call_1_3_14_4(SH2_struct* sh)
{
  SH2movls4(sh, 3,14, 4);
}
#define init_1_3_14_5
void FASTCALL call_1_3_14_5(SH2_struct* sh)
{
  SH2movls4(sh, 3,14, 5);
}
#define init_1_3_14_6
void FASTCALL call_1_3_14_6(SH2_struct* sh)
{
  SH2movls4(sh, 3,14, 6);
}
#define init_1_3_14_7
void FASTCALL call_1_3_14_7(SH2_struct* sh)
{
  SH2movls4(sh, 3,14, 7);
}
#define init_1_3_14_8
void FASTCALL call_1_3_14_8(SH2_struct* sh)
{
  SH2movls4(sh, 3,14, 8);
}
#define init_1_3_14_9
void FASTCALL call_1_3_14_9(SH2_struct* sh)
{
  SH2movls4(sh, 3,14, 9);
}
#define init_1_3_14_10
void FASTCALL call_1_3_14_10(SH2_struct* sh)
{
  SH2movls4(sh, 3,14, 10);
}
#define init_1_3_14_11
void FASTCALL call_1_3_14_11(SH2_struct* sh)
{
  SH2movls4(sh, 3,14, 11);
}
#define init_1_3_14_12
void FASTCALL call_1_3_14_12(SH2_struct* sh)
{
  SH2movls4(sh, 3,14, 12);
}
#define init_1_3_14_13
void FASTCALL call_1_3_14_13(SH2_struct* sh)
{
  SH2movls4(sh, 3,14, 13);
}
#define init_1_3_14_14
void FASTCALL call_1_3_14_14(SH2_struct* sh)
{
  SH2movls4(sh, 3,14, 14);
}
#define init_1_3_14_15
void FASTCALL call_1_3_14_15(SH2_struct* sh)
{
  SH2movls4(sh, 3,14, 15);
}
#define init_1_3_15_0
void FASTCALL call_1_3_15_0(SH2_struct* sh)
{
  SH2movls4(sh, 3,15, 0);
}
#define init_1_3_15_1
void FASTCALL call_1_3_15_1(SH2_struct* sh)
{
  SH2movls4(sh, 3,15, 1);
}
#define init_1_3_15_2
void FASTCALL call_1_3_15_2(SH2_struct* sh)
{
  SH2movls4(sh, 3,15, 2);
}
#define init_1_3_15_3
void FASTCALL call_1_3_15_3(SH2_struct* sh)
{
  SH2movls4(sh, 3,15, 3);
}
#define init_1_3_15_4
void FASTCALL call_1_3_15_4(SH2_struct* sh)
{
  SH2movls4(sh, 3,15, 4);
}
#define init_1_3_15_5
void FASTCALL call_1_3_15_5(SH2_struct* sh)
{
  SH2movls4(sh, 3,15, 5);
}
#define init_1_3_15_6
void FASTCALL call_1_3_15_6(SH2_struct* sh)
{
  SH2movls4(sh, 3,15, 6);
}
#define init_1_3_15_7
void FASTCALL call_1_3_15_7(SH2_struct* sh)
{
  SH2movls4(sh, 3,15, 7);
}
#define init_1_3_15_8
void FASTCALL call_1_3_15_8(SH2_struct* sh)
{
  SH2movls4(sh, 3,15, 8);
}
#define init_1_3_15_9
void FASTCALL call_1_3_15_9(SH2_struct* sh)
{
  SH2movls4(sh, 3,15, 9);
}
#define init_1_3_15_10
void FASTCALL call_1_3_15_10(SH2_struct* sh)
{
  SH2movls4(sh, 3,15, 10);
}
#define init_1_3_15_11
void FASTCALL call_1_3_15_11(SH2_struct* sh)
{
  SH2movls4(sh, 3,15, 11);
}
#define init_1_3_15_12
void FASTCALL call_1_3_15_12(SH2_struct* sh)
{
  SH2movls4(sh, 3,15, 12);
}
#define init_1_3_15_13
void FASTCALL call_1_3_15_13(SH2_struct* sh)
{
  SH2movls4(sh, 3,15, 13);
}
#define init_1_3_15_14
void FASTCALL call_1_3_15_14(SH2_struct* sh)
{
  SH2movls4(sh, 3,15, 14);
}
#define init_1_3_15_15
void FASTCALL call_1_3_15_15(SH2_struct* sh)
{
  SH2movls4(sh, 3,15, 15);
}
#define init_1_4_0_0
void FASTCALL call_1_4_0_0(SH2_struct* sh)
{
  SH2movls4(sh, 4,0, 0);
}
#define init_1_4_0_1
void FASTCALL call_1_4_0_1(SH2_struct* sh)
{
  SH2movls4(sh, 4,0, 1);
}
#define init_1_4_0_2
void FASTCALL call_1_4_0_2(SH2_struct* sh)
{
  SH2movls4(sh, 4,0, 2);
}
#define init_1_4_0_3
void FASTCALL call_1_4_0_3(SH2_struct* sh)
{
  SH2movls4(sh, 4,0, 3);
}
#define init_1_4_0_4
void FASTCALL call_1_4_0_4(SH2_struct* sh)
{
  SH2movls4(sh, 4,0, 4);
}
#define init_1_4_0_5
void FASTCALL call_1_4_0_5(SH2_struct* sh)
{
  SH2movls4(sh, 4,0, 5);
}
#define init_1_4_0_6
void FASTCALL call_1_4_0_6(SH2_struct* sh)
{
  SH2movls4(sh, 4,0, 6);
}
#define init_1_4_0_7
void FASTCALL call_1_4_0_7(SH2_struct* sh)
{
  SH2movls4(sh, 4,0, 7);
}
#define init_1_4_0_8
void FASTCALL call_1_4_0_8(SH2_struct* sh)
{
  SH2movls4(sh, 4,0, 8);
}
#define init_1_4_0_9
void FASTCALL call_1_4_0_9(SH2_struct* sh)
{
  SH2movls4(sh, 4,0, 9);
}
#define init_1_4_0_10
void FASTCALL call_1_4_0_10(SH2_struct* sh)
{
  SH2movls4(sh, 4,0, 10);
}
#define init_1_4_0_11
void FASTCALL call_1_4_0_11(SH2_struct* sh)
{
  SH2movls4(sh, 4,0, 11);
}
#define init_1_4_0_12
void FASTCALL call_1_4_0_12(SH2_struct* sh)
{
  SH2movls4(sh, 4,0, 12);
}
#define init_1_4_0_13
void FASTCALL call_1_4_0_13(SH2_struct* sh)
{
  SH2movls4(sh, 4,0, 13);
}
#define init_1_4_0_14
void FASTCALL call_1_4_0_14(SH2_struct* sh)
{
  SH2movls4(sh, 4,0, 14);
}
#define init_1_4_0_15
void FASTCALL call_1_4_0_15(SH2_struct* sh)
{
  SH2movls4(sh, 4,0, 15);
}
#define init_1_4_1_0
void FASTCALL call_1_4_1_0(SH2_struct* sh)
{
  SH2movls4(sh, 4,1, 0);
}
#define init_1_4_1_1
void FASTCALL call_1_4_1_1(SH2_struct* sh)
{
  SH2movls4(sh, 4,1, 1);
}
#define init_1_4_1_2
void FASTCALL call_1_4_1_2(SH2_struct* sh)
{
  SH2movls4(sh, 4,1, 2);
}
#define init_1_4_1_3
void FASTCALL call_1_4_1_3(SH2_struct* sh)
{
  SH2movls4(sh, 4,1, 3);
}
#define init_1_4_1_4
void FASTCALL call_1_4_1_4(SH2_struct* sh)
{
  SH2movls4(sh, 4,1, 4);
}
#define init_1_4_1_5
void FASTCALL call_1_4_1_5(SH2_struct* sh)
{
  SH2movls4(sh, 4,1, 5);
}
#define init_1_4_1_6
void FASTCALL call_1_4_1_6(SH2_struct* sh)
{
  SH2movls4(sh, 4,1, 6);
}
#define init_1_4_1_7
void FASTCALL call_1_4_1_7(SH2_struct* sh)
{
  SH2movls4(sh, 4,1, 7);
}
#define init_1_4_1_8
void FASTCALL call_1_4_1_8(SH2_struct* sh)
{
  SH2movls4(sh, 4,1, 8);
}
#define init_1_4_1_9
void FASTCALL call_1_4_1_9(SH2_struct* sh)
{
  SH2movls4(sh, 4,1, 9);
}
#define init_1_4_1_10
void FASTCALL call_1_4_1_10(SH2_struct* sh)
{
  SH2movls4(sh, 4,1, 10);
}
#define init_1_4_1_11
void FASTCALL call_1_4_1_11(SH2_struct* sh)
{
  SH2movls4(sh, 4,1, 11);
}
#define init_1_4_1_12
void FASTCALL call_1_4_1_12(SH2_struct* sh)
{
  SH2movls4(sh, 4,1, 12);
}
#define init_1_4_1_13
void FASTCALL call_1_4_1_13(SH2_struct* sh)
{
  SH2movls4(sh, 4,1, 13);
}
#define init_1_4_1_14
void FASTCALL call_1_4_1_14(SH2_struct* sh)
{
  SH2movls4(sh, 4,1, 14);
}
#define init_1_4_1_15
void FASTCALL call_1_4_1_15(SH2_struct* sh)
{
  SH2movls4(sh, 4,1, 15);
}
#define init_1_4_2_0
void FASTCALL call_1_4_2_0(SH2_struct* sh)
{
  SH2movls4(sh, 4,2, 0);
}
#define init_1_4_2_1
void FASTCALL call_1_4_2_1(SH2_struct* sh)
{
  SH2movls4(sh, 4,2, 1);
}
#define init_1_4_2_2
void FASTCALL call_1_4_2_2(SH2_struct* sh)
{
  SH2movls4(sh, 4,2, 2);
}
#define init_1_4_2_3
void FASTCALL call_1_4_2_3(SH2_struct* sh)
{
  SH2movls4(sh, 4,2, 3);
}
#define init_1_4_2_4
void FASTCALL call_1_4_2_4(SH2_struct* sh)
{
  SH2movls4(sh, 4,2, 4);
}
#define init_1_4_2_5
void FASTCALL call_1_4_2_5(SH2_struct* sh)
{
  SH2movls4(sh, 4,2, 5);
}
#define init_1_4_2_6
void FASTCALL call_1_4_2_6(SH2_struct* sh)
{
  SH2movls4(sh, 4,2, 6);
}
#define init_1_4_2_7
void FASTCALL call_1_4_2_7(SH2_struct* sh)
{
  SH2movls4(sh, 4,2, 7);
}
#define init_1_4_2_8
void FASTCALL call_1_4_2_8(SH2_struct* sh)
{
  SH2movls4(sh, 4,2, 8);
}
#define init_1_4_2_9
void FASTCALL call_1_4_2_9(SH2_struct* sh)
{
  SH2movls4(sh, 4,2, 9);
}
#define init_1_4_2_10
void FASTCALL call_1_4_2_10(SH2_struct* sh)
{
  SH2movls4(sh, 4,2, 10);
}
#define init_1_4_2_11
void FASTCALL call_1_4_2_11(SH2_struct* sh)
{
  SH2movls4(sh, 4,2, 11);
}
#define init_1_4_2_12
void FASTCALL call_1_4_2_12(SH2_struct* sh)
{
  SH2movls4(sh, 4,2, 12);
}
#define init_1_4_2_13
void FASTCALL call_1_4_2_13(SH2_struct* sh)
{
  SH2movls4(sh, 4,2, 13);
}
#define init_1_4_2_14
void FASTCALL call_1_4_2_14(SH2_struct* sh)
{
  SH2movls4(sh, 4,2, 14);
}
#define init_1_4_2_15
void FASTCALL call_1_4_2_15(SH2_struct* sh)
{
  SH2movls4(sh, 4,2, 15);
}
#define init_1_4_3_0
void FASTCALL call_1_4_3_0(SH2_struct* sh)
{
  SH2movls4(sh, 4,3, 0);
}
#define init_1_4_3_1
void FASTCALL call_1_4_3_1(SH2_struct* sh)
{
  SH2movls4(sh, 4,3, 1);
}
#define init_1_4_3_2
void FASTCALL call_1_4_3_2(SH2_struct* sh)
{
  SH2movls4(sh, 4,3, 2);
}
#define init_1_4_3_3
void FASTCALL call_1_4_3_3(SH2_struct* sh)
{
  SH2movls4(sh, 4,3, 3);
}
#define init_1_4_3_4
void FASTCALL call_1_4_3_4(SH2_struct* sh)
{
  SH2movls4(sh, 4,3, 4);
}
#define init_1_4_3_5
void FASTCALL call_1_4_3_5(SH2_struct* sh)
{
  SH2movls4(sh, 4,3, 5);
}
#define init_1_4_3_6
void FASTCALL call_1_4_3_6(SH2_struct* sh)
{
  SH2movls4(sh, 4,3, 6);
}
#define init_1_4_3_7
void FASTCALL call_1_4_3_7(SH2_struct* sh)
{
  SH2movls4(sh, 4,3, 7);
}
#define init_1_4_3_8
void FASTCALL call_1_4_3_8(SH2_struct* sh)
{
  SH2movls4(sh, 4,3, 8);
}
#define init_1_4_3_9
void FASTCALL call_1_4_3_9(SH2_struct* sh)
{
  SH2movls4(sh, 4,3, 9);
}
#define init_1_4_3_10
void FASTCALL call_1_4_3_10(SH2_struct* sh)
{
  SH2movls4(sh, 4,3, 10);
}
#define init_1_4_3_11
void FASTCALL call_1_4_3_11(SH2_struct* sh)
{
  SH2movls4(sh, 4,3, 11);
}
#define init_1_4_3_12
void FASTCALL call_1_4_3_12(SH2_struct* sh)
{
  SH2movls4(sh, 4,3, 12);
}
#define init_1_4_3_13
void FASTCALL call_1_4_3_13(SH2_struct* sh)
{
  SH2movls4(sh, 4,3, 13);
}
#define init_1_4_3_14
void FASTCALL call_1_4_3_14(SH2_struct* sh)
{
  SH2movls4(sh, 4,3, 14);
}
#define init_1_4_3_15
void FASTCALL call_1_4_3_15(SH2_struct* sh)
{
  SH2movls4(sh, 4,3, 15);
}
#define init_1_4_4_0
void FASTCALL call_1_4_4_0(SH2_struct* sh)
{
  SH2movls4(sh, 4,4, 0);
}
#define init_1_4_4_1
void FASTCALL call_1_4_4_1(SH2_struct* sh)
{
  SH2movls4(sh, 4,4, 1);
}
#define init_1_4_4_2
void FASTCALL call_1_4_4_2(SH2_struct* sh)
{
  SH2movls4(sh, 4,4, 2);
}
#define init_1_4_4_3
void FASTCALL call_1_4_4_3(SH2_struct* sh)
{
  SH2movls4(sh, 4,4, 3);
}
#define init_1_4_4_4
void FASTCALL call_1_4_4_4(SH2_struct* sh)
{
  SH2movls4(sh, 4,4, 4);
}
#define init_1_4_4_5
void FASTCALL call_1_4_4_5(SH2_struct* sh)
{
  SH2movls4(sh, 4,4, 5);
}
#define init_1_4_4_6
void FASTCALL call_1_4_4_6(SH2_struct* sh)
{
  SH2movls4(sh, 4,4, 6);
}
#define init_1_4_4_7
void FASTCALL call_1_4_4_7(SH2_struct* sh)
{
  SH2movls4(sh, 4,4, 7);
}
#define init_1_4_4_8
void FASTCALL call_1_4_4_8(SH2_struct* sh)
{
  SH2movls4(sh, 4,4, 8);
}
#define init_1_4_4_9
void FASTCALL call_1_4_4_9(SH2_struct* sh)
{
  SH2movls4(sh, 4,4, 9);
}
#define init_1_4_4_10
void FASTCALL call_1_4_4_10(SH2_struct* sh)
{
  SH2movls4(sh, 4,4, 10);
}
#define init_1_4_4_11
void FASTCALL call_1_4_4_11(SH2_struct* sh)
{
  SH2movls4(sh, 4,4, 11);
}
#define init_1_4_4_12
void FASTCALL call_1_4_4_12(SH2_struct* sh)
{
  SH2movls4(sh, 4,4, 12);
}
#define init_1_4_4_13
void FASTCALL call_1_4_4_13(SH2_struct* sh)
{
  SH2movls4(sh, 4,4, 13);
}
#define init_1_4_4_14
void FASTCALL call_1_4_4_14(SH2_struct* sh)
{
  SH2movls4(sh, 4,4, 14);
}
#define init_1_4_4_15
void FASTCALL call_1_4_4_15(SH2_struct* sh)
{
  SH2movls4(sh, 4,4, 15);
}
#define init_1_4_5_0
void FASTCALL call_1_4_5_0(SH2_struct* sh)
{
  SH2movls4(sh, 4,5, 0);
}
#define init_1_4_5_1
void FASTCALL call_1_4_5_1(SH2_struct* sh)
{
  SH2movls4(sh, 4,5, 1);
}
#define init_1_4_5_2
void FASTCALL call_1_4_5_2(SH2_struct* sh)
{
  SH2movls4(sh, 4,5, 2);
}
#define init_1_4_5_3
void FASTCALL call_1_4_5_3(SH2_struct* sh)
{
  SH2movls4(sh, 4,5, 3);
}
#define init_1_4_5_4
void FASTCALL call_1_4_5_4(SH2_struct* sh)
{
  SH2movls4(sh, 4,5, 4);
}
#define init_1_4_5_5
void FASTCALL call_1_4_5_5(SH2_struct* sh)
{
  SH2movls4(sh, 4,5, 5);
}
#define init_1_4_5_6
void FASTCALL call_1_4_5_6(SH2_struct* sh)
{
  SH2movls4(sh, 4,5, 6);
}
#define init_1_4_5_7
void FASTCALL call_1_4_5_7(SH2_struct* sh)
{
  SH2movls4(sh, 4,5, 7);
}
#define init_1_4_5_8
void FASTCALL call_1_4_5_8(SH2_struct* sh)
{
  SH2movls4(sh, 4,5, 8);
}
#define init_1_4_5_9
void FASTCALL call_1_4_5_9(SH2_struct* sh)
{
  SH2movls4(sh, 4,5, 9);
}
#define init_1_4_5_10
void FASTCALL call_1_4_5_10(SH2_struct* sh)
{
  SH2movls4(sh, 4,5, 10);
}
#define init_1_4_5_11
void FASTCALL call_1_4_5_11(SH2_struct* sh)
{
  SH2movls4(sh, 4,5, 11);
}
#define init_1_4_5_12
void FASTCALL call_1_4_5_12(SH2_struct* sh)
{
  SH2movls4(sh, 4,5, 12);
}
#define init_1_4_5_13
void FASTCALL call_1_4_5_13(SH2_struct* sh)
{
  SH2movls4(sh, 4,5, 13);
}
#define init_1_4_5_14
void FASTCALL call_1_4_5_14(SH2_struct* sh)
{
  SH2movls4(sh, 4,5, 14);
}
#define init_1_4_5_15
void FASTCALL call_1_4_5_15(SH2_struct* sh)
{
  SH2movls4(sh, 4,5, 15);
}
#define init_1_4_6_0
void FASTCALL call_1_4_6_0(SH2_struct* sh)
{
  SH2movls4(sh, 4,6, 0);
}
#define init_1_4_6_1
void FASTCALL call_1_4_6_1(SH2_struct* sh)
{
  SH2movls4(sh, 4,6, 1);
}
#define init_1_4_6_2
void FASTCALL call_1_4_6_2(SH2_struct* sh)
{
  SH2movls4(sh, 4,6, 2);
}
#define init_1_4_6_3
void FASTCALL call_1_4_6_3(SH2_struct* sh)
{
  SH2movls4(sh, 4,6, 3);
}
#define init_1_4_6_4
void FASTCALL call_1_4_6_4(SH2_struct* sh)
{
  SH2movls4(sh, 4,6, 4);
}
#define init_1_4_6_5
void FASTCALL call_1_4_6_5(SH2_struct* sh)
{
  SH2movls4(sh, 4,6, 5);
}
#define init_1_4_6_6
void FASTCALL call_1_4_6_6(SH2_struct* sh)
{
  SH2movls4(sh, 4,6, 6);
}
#define init_1_4_6_7
void FASTCALL call_1_4_6_7(SH2_struct* sh)
{
  SH2movls4(sh, 4,6, 7);
}
#define init_1_4_6_8
void FASTCALL call_1_4_6_8(SH2_struct* sh)
{
  SH2movls4(sh, 4,6, 8);
}
#define init_1_4_6_9
void FASTCALL call_1_4_6_9(SH2_struct* sh)
{
  SH2movls4(sh, 4,6, 9);
}
#define init_1_4_6_10
void FASTCALL call_1_4_6_10(SH2_struct* sh)
{
  SH2movls4(sh, 4,6, 10);
}
#define init_1_4_6_11
void FASTCALL call_1_4_6_11(SH2_struct* sh)
{
  SH2movls4(sh, 4,6, 11);
}
#define init_1_4_6_12
void FASTCALL call_1_4_6_12(SH2_struct* sh)
{
  SH2movls4(sh, 4,6, 12);
}
#define init_1_4_6_13
void FASTCALL call_1_4_6_13(SH2_struct* sh)
{
  SH2movls4(sh, 4,6, 13);
}
#define init_1_4_6_14
void FASTCALL call_1_4_6_14(SH2_struct* sh)
{
  SH2movls4(sh, 4,6, 14);
}
#define init_1_4_6_15
void FASTCALL call_1_4_6_15(SH2_struct* sh)
{
  SH2movls4(sh, 4,6, 15);
}
#define init_1_4_7_0
void FASTCALL call_1_4_7_0(SH2_struct* sh)
{
  SH2movls4(sh, 4,7, 0);
}
#define init_1_4_7_1
void FASTCALL call_1_4_7_1(SH2_struct* sh)
{
  SH2movls4(sh, 4,7, 1);
}
#define init_1_4_7_2
void FASTCALL call_1_4_7_2(SH2_struct* sh)
{
  SH2movls4(sh, 4,7, 2);
}
#define init_1_4_7_3
void FASTCALL call_1_4_7_3(SH2_struct* sh)
{
  SH2movls4(sh, 4,7, 3);
}
#define init_1_4_7_4
void FASTCALL call_1_4_7_4(SH2_struct* sh)
{
  SH2movls4(sh, 4,7, 4);
}
#define init_1_4_7_5
void FASTCALL call_1_4_7_5(SH2_struct* sh)
{
  SH2movls4(sh, 4,7, 5);
}
#define init_1_4_7_6
void FASTCALL call_1_4_7_6(SH2_struct* sh)
{
  SH2movls4(sh, 4,7, 6);
}
#define init_1_4_7_7
void FASTCALL call_1_4_7_7(SH2_struct* sh)
{
  SH2movls4(sh, 4,7, 7);
}
#define init_1_4_7_8
void FASTCALL call_1_4_7_8(SH2_struct* sh)
{
  SH2movls4(sh, 4,7, 8);
}
#define init_1_4_7_9
void FASTCALL call_1_4_7_9(SH2_struct* sh)
{
  SH2movls4(sh, 4,7, 9);
}
#define init_1_4_7_10
void FASTCALL call_1_4_7_10(SH2_struct* sh)
{
  SH2movls4(sh, 4,7, 10);
}
#define init_1_4_7_11
void FASTCALL call_1_4_7_11(SH2_struct* sh)
{
  SH2movls4(sh, 4,7, 11);
}
#define init_1_4_7_12
void FASTCALL call_1_4_7_12(SH2_struct* sh)
{
  SH2movls4(sh, 4,7, 12);
}
#define init_1_4_7_13
void FASTCALL call_1_4_7_13(SH2_struct* sh)
{
  SH2movls4(sh, 4,7, 13);
}
#define init_1_4_7_14
void FASTCALL call_1_4_7_14(SH2_struct* sh)
{
  SH2movls4(sh, 4,7, 14);
}
#define init_1_4_7_15
void FASTCALL call_1_4_7_15(SH2_struct* sh)
{
  SH2movls4(sh, 4,7, 15);
}
#define init_1_4_8_0
void FASTCALL call_1_4_8_0(SH2_struct* sh)
{
  SH2movls4(sh, 4,8, 0);
}
#define init_1_4_8_1
void FASTCALL call_1_4_8_1(SH2_struct* sh)
{
  SH2movls4(sh, 4,8, 1);
}
#define init_1_4_8_2
void FASTCALL call_1_4_8_2(SH2_struct* sh)
{
  SH2movls4(sh, 4,8, 2);
}
#define init_1_4_8_3
void FASTCALL call_1_4_8_3(SH2_struct* sh)
{
  SH2movls4(sh, 4,8, 3);
}
#define init_1_4_8_4
void FASTCALL call_1_4_8_4(SH2_struct* sh)
{
  SH2movls4(sh, 4,8, 4);
}
#define init_1_4_8_5
void FASTCALL call_1_4_8_5(SH2_struct* sh)
{
  SH2movls4(sh, 4,8, 5);
}
#define init_1_4_8_6
void FASTCALL call_1_4_8_6(SH2_struct* sh)
{
  SH2movls4(sh, 4,8, 6);
}
#define init_1_4_8_7
void FASTCALL call_1_4_8_7(SH2_struct* sh)
{
  SH2movls4(sh, 4,8, 7);
}
#define init_1_4_8_8
void FASTCALL call_1_4_8_8(SH2_struct* sh)
{
  SH2movls4(sh, 4,8, 8);
}
#define init_1_4_8_9
void FASTCALL call_1_4_8_9(SH2_struct* sh)
{
  SH2movls4(sh, 4,8, 9);
}
#define init_1_4_8_10
void FASTCALL call_1_4_8_10(SH2_struct* sh)
{
  SH2movls4(sh, 4,8, 10);
}
#define init_1_4_8_11
void FASTCALL call_1_4_8_11(SH2_struct* sh)
{
  SH2movls4(sh, 4,8, 11);
}
#define init_1_4_8_12
void FASTCALL call_1_4_8_12(SH2_struct* sh)
{
  SH2movls4(sh, 4,8, 12);
}
#define init_1_4_8_13
void FASTCALL call_1_4_8_13(SH2_struct* sh)
{
  SH2movls4(sh, 4,8, 13);
}
#define init_1_4_8_14
void FASTCALL call_1_4_8_14(SH2_struct* sh)
{
  SH2movls4(sh, 4,8, 14);
}
#define init_1_4_8_15
void FASTCALL call_1_4_8_15(SH2_struct* sh)
{
  SH2movls4(sh, 4,8, 15);
}
#define init_1_4_9_0
void FASTCALL call_1_4_9_0(SH2_struct* sh)
{
  SH2movls4(sh, 4,9, 0);
}
#define init_1_4_9_1
void FASTCALL call_1_4_9_1(SH2_struct* sh)
{
  SH2movls4(sh, 4,9, 1);
}
#define init_1_4_9_2
void FASTCALL call_1_4_9_2(SH2_struct* sh)
{
  SH2movls4(sh, 4,9, 2);
}
#define init_1_4_9_3
void FASTCALL call_1_4_9_3(SH2_struct* sh)
{
  SH2movls4(sh, 4,9, 3);
}
#define init_1_4_9_4
void FASTCALL call_1_4_9_4(SH2_struct* sh)
{
  SH2movls4(sh, 4,9, 4);
}
#define init_1_4_9_5
void FASTCALL call_1_4_9_5(SH2_struct* sh)
{
  SH2movls4(sh, 4,9, 5);
}
#define init_1_4_9_6
void FASTCALL call_1_4_9_6(SH2_struct* sh)
{
  SH2movls4(sh, 4,9, 6);
}
#define init_1_4_9_7
void FASTCALL call_1_4_9_7(SH2_struct* sh)
{
  SH2movls4(sh, 4,9, 7);
}
#define init_1_4_9_8
void FASTCALL call_1_4_9_8(SH2_struct* sh)
{
  SH2movls4(sh, 4,9, 8);
}
#define init_1_4_9_9
void FASTCALL call_1_4_9_9(SH2_struct* sh)
{
  SH2movls4(sh, 4,9, 9);
}
#define init_1_4_9_10
void FASTCALL call_1_4_9_10(SH2_struct* sh)
{
  SH2movls4(sh, 4,9, 10);
}
#define init_1_4_9_11
void FASTCALL call_1_4_9_11(SH2_struct* sh)
{
  SH2movls4(sh, 4,9, 11);
}
#define init_1_4_9_12
void FASTCALL call_1_4_9_12(SH2_struct* sh)
{
  SH2movls4(sh, 4,9, 12);
}
#define init_1_4_9_13
void FASTCALL call_1_4_9_13(SH2_struct* sh)
{
  SH2movls4(sh, 4,9, 13);
}
#define init_1_4_9_14
void FASTCALL call_1_4_9_14(SH2_struct* sh)
{
  SH2movls4(sh, 4,9, 14);
}
#define init_1_4_9_15
void FASTCALL call_1_4_9_15(SH2_struct* sh)
{
  SH2movls4(sh, 4,9, 15);
}
#define init_1_4_10_0
void FASTCALL call_1_4_10_0(SH2_struct* sh)
{
  SH2movls4(sh, 4,10, 0);
}
#define init_1_4_10_1
void FASTCALL call_1_4_10_1(SH2_struct* sh)
{
  SH2movls4(sh, 4,10, 1);
}
#define init_1_4_10_2
void FASTCALL call_1_4_10_2(SH2_struct* sh)
{
  SH2movls4(sh, 4,10, 2);
}
#define init_1_4_10_3
void FASTCALL call_1_4_10_3(SH2_struct* sh)
{
  SH2movls4(sh, 4,10, 3);
}
#define init_1_4_10_4
void FASTCALL call_1_4_10_4(SH2_struct* sh)
{
  SH2movls4(sh, 4,10, 4);
}
#define init_1_4_10_5
void FASTCALL call_1_4_10_5(SH2_struct* sh)
{
  SH2movls4(sh, 4,10, 5);
}
#define init_1_4_10_6
void FASTCALL call_1_4_10_6(SH2_struct* sh)
{
  SH2movls4(sh, 4,10, 6);
}
#define init_1_4_10_7
void FASTCALL call_1_4_10_7(SH2_struct* sh)
{
  SH2movls4(sh, 4,10, 7);
}
#define init_1_4_10_8
void FASTCALL call_1_4_10_8(SH2_struct* sh)
{
  SH2movls4(sh, 4,10, 8);
}
#define init_1_4_10_9
void FASTCALL call_1_4_10_9(SH2_struct* sh)
{
  SH2movls4(sh, 4,10, 9);
}
#define init_1_4_10_10
void FASTCALL call_1_4_10_10(SH2_struct* sh)
{
  SH2movls4(sh, 4,10, 10);
}
#define init_1_4_10_11
void FASTCALL call_1_4_10_11(SH2_struct* sh)
{
  SH2movls4(sh, 4,10, 11);
}
#define init_1_4_10_12
void FASTCALL call_1_4_10_12(SH2_struct* sh)
{
  SH2movls4(sh, 4,10, 12);
}
#define init_1_4_10_13
void FASTCALL call_1_4_10_13(SH2_struct* sh)
{
  SH2movls4(sh, 4,10, 13);
}
#define init_1_4_10_14
void FASTCALL call_1_4_10_14(SH2_struct* sh)
{
  SH2movls4(sh, 4,10, 14);
}
#define init_1_4_10_15
void FASTCALL call_1_4_10_15(SH2_struct* sh)
{
  SH2movls4(sh, 4,10, 15);
}
#define init_1_4_11_0
void FASTCALL call_1_4_11_0(SH2_struct* sh)
{
  SH2movls4(sh, 4,11, 0);
}
#define init_1_4_11_1
void FASTCALL call_1_4_11_1(SH2_struct* sh)
{
  SH2movls4(sh, 4,11, 1);
}
#define init_1_4_11_2
void FASTCALL call_1_4_11_2(SH2_struct* sh)
{
  SH2movls4(sh, 4,11, 2);
}
#define init_1_4_11_3
void FASTCALL call_1_4_11_3(SH2_struct* sh)
{
  SH2movls4(sh, 4,11, 3);
}
#define init_1_4_11_4
void FASTCALL call_1_4_11_4(SH2_struct* sh)
{
  SH2movls4(sh, 4,11, 4);
}
#define init_1_4_11_5
void FASTCALL call_1_4_11_5(SH2_struct* sh)
{
  SH2movls4(sh, 4,11, 5);
}
#define init_1_4_11_6
void FASTCALL call_1_4_11_6(SH2_struct* sh)
{
  SH2movls4(sh, 4,11, 6);
}
#define init_1_4_11_7
void FASTCALL call_1_4_11_7(SH2_struct* sh)
{
  SH2movls4(sh, 4,11, 7);
}
#define init_1_4_11_8
void FASTCALL call_1_4_11_8(SH2_struct* sh)
{
  SH2movls4(sh, 4,11, 8);
}
#define init_1_4_11_9
void FASTCALL call_1_4_11_9(SH2_struct* sh)
{
  SH2movls4(sh, 4,11, 9);
}
#define init_1_4_11_10
void FASTCALL call_1_4_11_10(SH2_struct* sh)
{
  SH2movls4(sh, 4,11, 10);
}
#define init_1_4_11_11
void FASTCALL call_1_4_11_11(SH2_struct* sh)
{
  SH2movls4(sh, 4,11, 11);
}
#define init_1_4_11_12
void FASTCALL call_1_4_11_12(SH2_struct* sh)
{
  SH2movls4(sh, 4,11, 12);
}
#define init_1_4_11_13
void FASTCALL call_1_4_11_13(SH2_struct* sh)
{
  SH2movls4(sh, 4,11, 13);
}
#define init_1_4_11_14
void FASTCALL call_1_4_11_14(SH2_struct* sh)
{
  SH2movls4(sh, 4,11, 14);
}
#define init_1_4_11_15
void FASTCALL call_1_4_11_15(SH2_struct* sh)
{
  SH2movls4(sh, 4,11, 15);
}
#define init_1_4_12_0
void FASTCALL call_1_4_12_0(SH2_struct* sh)
{
  SH2movls4(sh, 4,12, 0);
}
#define init_1_4_12_1
void FASTCALL call_1_4_12_1(SH2_struct* sh)
{
  SH2movls4(sh, 4,12, 1);
}
#define init_1_4_12_2
void FASTCALL call_1_4_12_2(SH2_struct* sh)
{
  SH2movls4(sh, 4,12, 2);
}
#define init_1_4_12_3
void FASTCALL call_1_4_12_3(SH2_struct* sh)
{
  SH2movls4(sh, 4,12, 3);
}
#define init_1_4_12_4
void FASTCALL call_1_4_12_4(SH2_struct* sh)
{
  SH2movls4(sh, 4,12, 4);
}
#define init_1_4_12_5
void FASTCALL call_1_4_12_5(SH2_struct* sh)
{
  SH2movls4(sh, 4,12, 5);
}
#define init_1_4_12_6
void FASTCALL call_1_4_12_6(SH2_struct* sh)
{
  SH2movls4(sh, 4,12, 6);
}
#define init_1_4_12_7
void FASTCALL call_1_4_12_7(SH2_struct* sh)
{
  SH2movls4(sh, 4,12, 7);
}
#define init_1_4_12_8
void FASTCALL call_1_4_12_8(SH2_struct* sh)
{
  SH2movls4(sh, 4,12, 8);
}
#define init_1_4_12_9
void FASTCALL call_1_4_12_9(SH2_struct* sh)
{
  SH2movls4(sh, 4,12, 9);
}
#define init_1_4_12_10
void FASTCALL call_1_4_12_10(SH2_struct* sh)
{
  SH2movls4(sh, 4,12, 10);
}
#define init_1_4_12_11
void FASTCALL call_1_4_12_11(SH2_struct* sh)
{
  SH2movls4(sh, 4,12, 11);
}
#define init_1_4_12_12
void FASTCALL call_1_4_12_12(SH2_struct* sh)
{
  SH2movls4(sh, 4,12, 12);
}
#define init_1_4_12_13
void FASTCALL call_1_4_12_13(SH2_struct* sh)
{
  SH2movls4(sh, 4,12, 13);
}
#define init_1_4_12_14
void FASTCALL call_1_4_12_14(SH2_struct* sh)
{
  SH2movls4(sh, 4,12, 14);
}
#define init_1_4_12_15
void FASTCALL call_1_4_12_15(SH2_struct* sh)
{
  SH2movls4(sh, 4,12, 15);
}
#define init_1_4_13_0
void FASTCALL call_1_4_13_0(SH2_struct* sh)
{
  SH2movls4(sh, 4,13, 0);
}
#define init_1_4_13_1
void FASTCALL call_1_4_13_1(SH2_struct* sh)
{
  SH2movls4(sh, 4,13, 1);
}
#define init_1_4_13_2
void FASTCALL call_1_4_13_2(SH2_struct* sh)
{
  SH2movls4(sh, 4,13, 2);
}
#define init_1_4_13_3
void FASTCALL call_1_4_13_3(SH2_struct* sh)
{
  SH2movls4(sh, 4,13, 3);
}
#define init_1_4_13_4
void FASTCALL call_1_4_13_4(SH2_struct* sh)
{
  SH2movls4(sh, 4,13, 4);
}
#define init_1_4_13_5
void FASTCALL call_1_4_13_5(SH2_struct* sh)
{
  SH2movls4(sh, 4,13, 5);
}
#define init_1_4_13_6
void FASTCALL call_1_4_13_6(SH2_struct* sh)
{
  SH2movls4(sh, 4,13, 6);
}
#define init_1_4_13_7
void FASTCALL call_1_4_13_7(SH2_struct* sh)
{
  SH2movls4(sh, 4,13, 7);
}
#define init_1_4_13_8
void FASTCALL call_1_4_13_8(SH2_struct* sh)
{
  SH2movls4(sh, 4,13, 8);
}
#define init_1_4_13_9
void FASTCALL call_1_4_13_9(SH2_struct* sh)
{
  SH2movls4(sh, 4,13, 9);
}
#define init_1_4_13_10
void FASTCALL call_1_4_13_10(SH2_struct* sh)
{
  SH2movls4(sh, 4,13, 10);
}
#define init_1_4_13_11
void FASTCALL call_1_4_13_11(SH2_struct* sh)
{
  SH2movls4(sh, 4,13, 11);
}
#define init_1_4_13_12
void FASTCALL call_1_4_13_12(SH2_struct* sh)
{
  SH2movls4(sh, 4,13, 12);
}
#define init_1_4_13_13
void FASTCALL call_1_4_13_13(SH2_struct* sh)
{
  SH2movls4(sh, 4,13, 13);
}
#define init_1_4_13_14
void FASTCALL call_1_4_13_14(SH2_struct* sh)
{
  SH2movls4(sh, 4,13, 14);
}
#define init_1_4_13_15
void FASTCALL call_1_4_13_15(SH2_struct* sh)
{
  SH2movls4(sh, 4,13, 15);
}
#define init_1_4_14_0
void FASTCALL call_1_4_14_0(SH2_struct* sh)
{
  SH2movls4(sh, 4,14, 0);
}
#define init_1_4_14_1
void FASTCALL call_1_4_14_1(SH2_struct* sh)
{
  SH2movls4(sh, 4,14, 1);
}
#define init_1_4_14_2
void FASTCALL call_1_4_14_2(SH2_struct* sh)
{
  SH2movls4(sh, 4,14, 2);
}
#define init_1_4_14_3
void FASTCALL call_1_4_14_3(SH2_struct* sh)
{
  SH2movls4(sh, 4,14, 3);
}
#define init_1_4_14_4
void FASTCALL call_1_4_14_4(SH2_struct* sh)
{
  SH2movls4(sh, 4,14, 4);
}
#define init_1_4_14_5
void FASTCALL call_1_4_14_5(SH2_struct* sh)
{
  SH2movls4(sh, 4,14, 5);
}
#define init_1_4_14_6
void FASTCALL call_1_4_14_6(SH2_struct* sh)
{
  SH2movls4(sh, 4,14, 6);
}
#define init_1_4_14_7
void FASTCALL call_1_4_14_7(SH2_struct* sh)
{
  SH2movls4(sh, 4,14, 7);
}
#define init_1_4_14_8
void FASTCALL call_1_4_14_8(SH2_struct* sh)
{
  SH2movls4(sh, 4,14, 8);
}
#define init_1_4_14_9
void FASTCALL call_1_4_14_9(SH2_struct* sh)
{
  SH2movls4(sh, 4,14, 9);
}
#define init_1_4_14_10
void FASTCALL call_1_4_14_10(SH2_struct* sh)
{
  SH2movls4(sh, 4,14, 10);
}
#define init_1_4_14_11
void FASTCALL call_1_4_14_11(SH2_struct* sh)
{
  SH2movls4(sh, 4,14, 11);
}
#define init_1_4_14_12
void FASTCALL call_1_4_14_12(SH2_struct* sh)
{
  SH2movls4(sh, 4,14, 12);
}
#define init_1_4_14_13
void FASTCALL call_1_4_14_13(SH2_struct* sh)
{
  SH2movls4(sh, 4,14, 13);
}
#define init_1_4_14_14
void FASTCALL call_1_4_14_14(SH2_struct* sh)
{
  SH2movls4(sh, 4,14, 14);
}
#define init_1_4_14_15
void FASTCALL call_1_4_14_15(SH2_struct* sh)
{
  SH2movls4(sh, 4,14, 15);
}
#define init_1_4_15_0
void FASTCALL call_1_4_15_0(SH2_struct* sh)
{
  SH2movls4(sh, 4,15, 0);
}
#define init_1_4_15_1
void FASTCALL call_1_4_15_1(SH2_struct* sh)
{
  SH2movls4(sh, 4,15, 1);
}
#define init_1_4_15_2
void FASTCALL call_1_4_15_2(SH2_struct* sh)
{
  SH2movls4(sh, 4,15, 2);
}
#define init_1_4_15_3
void FASTCALL call_1_4_15_3(SH2_struct* sh)
{
  SH2movls4(sh, 4,15, 3);
}
#define init_1_4_15_4
void FASTCALL call_1_4_15_4(SH2_struct* sh)
{
  SH2movls4(sh, 4,15, 4);
}
#define init_1_4_15_5
void FASTCALL call_1_4_15_5(SH2_struct* sh)
{
  SH2movls4(sh, 4,15, 5);
}
#define init_1_4_15_6
void FASTCALL call_1_4_15_6(SH2_struct* sh)
{
  SH2movls4(sh, 4,15, 6);
}
#define init_1_4_15_7
void FASTCALL call_1_4_15_7(SH2_struct* sh)
{
  SH2movls4(sh, 4,15, 7);
}
#define init_1_4_15_8
void FASTCALL call_1_4_15_8(SH2_struct* sh)
{
  SH2movls4(sh, 4,15, 8);
}
#define init_1_4_15_9
void FASTCALL call_1_4_15_9(SH2_struct* sh)
{
  SH2movls4(sh, 4,15, 9);
}
#define init_1_4_15_10
void FASTCALL call_1_4_15_10(SH2_struct* sh)
{
  SH2movls4(sh, 4,15, 10);
}
#define init_1_4_15_11
void FASTCALL call_1_4_15_11(SH2_struct* sh)
{
  SH2movls4(sh, 4,15, 11);
}
#define init_1_4_15_12
void FASTCALL call_1_4_15_12(SH2_struct* sh)
{
  SH2movls4(sh, 4,15, 12);
}
#define init_1_4_15_13
void FASTCALL call_1_4_15_13(SH2_struct* sh)
{
  SH2movls4(sh, 4,15, 13);
}
#define init_1_4_15_14
void FASTCALL call_1_4_15_14(SH2_struct* sh)
{
  SH2movls4(sh, 4,15, 14);
}
#define init_1_4_15_15
void FASTCALL call_1_4_15_15(SH2_struct* sh)
{
  SH2movls4(sh, 4,15, 15);
}
#define init_1_5_0_0
void FASTCALL call_1_5_0_0(SH2_struct* sh)
{
  SH2movls4(sh, 5,0, 0);
}
#define init_1_5_0_1
void FASTCALL call_1_5_0_1(SH2_struct* sh)
{
  SH2movls4(sh, 5,0, 1);
}
#define init_1_5_0_2
void FASTCALL call_1_5_0_2(SH2_struct* sh)
{
  SH2movls4(sh, 5,0, 2);
}
#define init_1_5_0_3
void FASTCALL call_1_5_0_3(SH2_struct* sh)
{
  SH2movls4(sh, 5,0, 3);
}
#define init_1_5_0_4
void FASTCALL call_1_5_0_4(SH2_struct* sh)
{
  SH2movls4(sh, 5,0, 4);
}
#define init_1_5_0_5
void FASTCALL call_1_5_0_5(SH2_struct* sh)
{
  SH2movls4(sh, 5,0, 5);
}
#define init_1_5_0_6
void FASTCALL call_1_5_0_6(SH2_struct* sh)
{
  SH2movls4(sh, 5,0, 6);
}
#define init_1_5_0_7
void FASTCALL call_1_5_0_7(SH2_struct* sh)
{
  SH2movls4(sh, 5,0, 7);
}
#define init_1_5_0_8
void FASTCALL call_1_5_0_8(SH2_struct* sh)
{
  SH2movls4(sh, 5,0, 8);
}
#define init_1_5_0_9
void FASTCALL call_1_5_0_9(SH2_struct* sh)
{
  SH2movls4(sh, 5,0, 9);
}
#define init_1_5_0_10
void FASTCALL call_1_5_0_10(SH2_struct* sh)
{
  SH2movls4(sh, 5,0, 10);
}
#define init_1_5_0_11
void FASTCALL call_1_5_0_11(SH2_struct* sh)
{
  SH2movls4(sh, 5,0, 11);
}
#define init_1_5_0_12
void FASTCALL call_1_5_0_12(SH2_struct* sh)
{
  SH2movls4(sh, 5,0, 12);
}
#define init_1_5_0_13
void FASTCALL call_1_5_0_13(SH2_struct* sh)
{
  SH2movls4(sh, 5,0, 13);
}
#define init_1_5_0_14
void FASTCALL call_1_5_0_14(SH2_struct* sh)
{
  SH2movls4(sh, 5,0, 14);
}
#define init_1_5_0_15
void FASTCALL call_1_5_0_15(SH2_struct* sh)
{
  SH2movls4(sh, 5,0, 15);
}
#define init_1_5_1_0
void FASTCALL call_1_5_1_0(SH2_struct* sh)
{
  SH2movls4(sh, 5,1, 0);
}
#define init_1_5_1_1
void FASTCALL call_1_5_1_1(SH2_struct* sh)
{
  SH2movls4(sh, 5,1, 1);
}
#define init_1_5_1_2
void FASTCALL call_1_5_1_2(SH2_struct* sh)
{
  SH2movls4(sh, 5,1, 2);
}
#define init_1_5_1_3
void FASTCALL call_1_5_1_3(SH2_struct* sh)
{
  SH2movls4(sh, 5,1, 3);
}
#define init_1_5_1_4
void FASTCALL call_1_5_1_4(SH2_struct* sh)
{
  SH2movls4(sh, 5,1, 4);
}
#define init_1_5_1_5
void FASTCALL call_1_5_1_5(SH2_struct* sh)
{
  SH2movls4(sh, 5,1, 5);
}
#define init_1_5_1_6
void FASTCALL call_1_5_1_6(SH2_struct* sh)
{
  SH2movls4(sh, 5,1, 6);
}
#define init_1_5_1_7
void FASTCALL call_1_5_1_7(SH2_struct* sh)
{
  SH2movls4(sh, 5,1, 7);
}
#define init_1_5_1_8
void FASTCALL call_1_5_1_8(SH2_struct* sh)
{
  SH2movls4(sh, 5,1, 8);
}
#define init_1_5_1_9
void FASTCALL call_1_5_1_9(SH2_struct* sh)
{
  SH2movls4(sh, 5,1, 9);
}
#define init_1_5_1_10
void FASTCALL call_1_5_1_10(SH2_struct* sh)
{
  SH2movls4(sh, 5,1, 10);
}
#define init_1_5_1_11
void FASTCALL call_1_5_1_11(SH2_struct* sh)
{
  SH2movls4(sh, 5,1, 11);
}
#define init_1_5_1_12
void FASTCALL call_1_5_1_12(SH2_struct* sh)
{
  SH2movls4(sh, 5,1, 12);
}
#define init_1_5_1_13
void FASTCALL call_1_5_1_13(SH2_struct* sh)
{
  SH2movls4(sh, 5,1, 13);
}
#define init_1_5_1_14
void FASTCALL call_1_5_1_14(SH2_struct* sh)
{
  SH2movls4(sh, 5,1, 14);
}
#define init_1_5_1_15
void FASTCALL call_1_5_1_15(SH2_struct* sh)
{
  SH2movls4(sh, 5,1, 15);
}
#define init_1_5_2_0
void FASTCALL call_1_5_2_0(SH2_struct* sh)
{
  SH2movls4(sh, 5,2, 0);
}
#define init_1_5_2_1
void FASTCALL call_1_5_2_1(SH2_struct* sh)
{
  SH2movls4(sh, 5,2, 1);
}
#define init_1_5_2_2
void FASTCALL call_1_5_2_2(SH2_struct* sh)
{
  SH2movls4(sh, 5,2, 2);
}
#define init_1_5_2_3
void FASTCALL call_1_5_2_3(SH2_struct* sh)
{
  SH2movls4(sh, 5,2, 3);
}
#define init_1_5_2_4
void FASTCALL call_1_5_2_4(SH2_struct* sh)
{
  SH2movls4(sh, 5,2, 4);
}
#define init_1_5_2_5
void FASTCALL call_1_5_2_5(SH2_struct* sh)
{
  SH2movls4(sh, 5,2, 5);
}
#define init_1_5_2_6
void FASTCALL call_1_5_2_6(SH2_struct* sh)
{
  SH2movls4(sh, 5,2, 6);
}
#define init_1_5_2_7
void FASTCALL call_1_5_2_7(SH2_struct* sh)
{
  SH2movls4(sh, 5,2, 7);
}
#define init_1_5_2_8
void FASTCALL call_1_5_2_8(SH2_struct* sh)
{
  SH2movls4(sh, 5,2, 8);
}
#define init_1_5_2_9
void FASTCALL call_1_5_2_9(SH2_struct* sh)
{
  SH2movls4(sh, 5,2, 9);
}
#define init_1_5_2_10
void FASTCALL call_1_5_2_10(SH2_struct* sh)
{
  SH2movls4(sh, 5,2, 10);
}
#define init_1_5_2_11
void FASTCALL call_1_5_2_11(SH2_struct* sh)
{
  SH2movls4(sh, 5,2, 11);
}
#define init_1_5_2_12
void FASTCALL call_1_5_2_12(SH2_struct* sh)
{
  SH2movls4(sh, 5,2, 12);
}
#define init_1_5_2_13
void FASTCALL call_1_5_2_13(SH2_struct* sh)
{
  SH2movls4(sh, 5,2, 13);
}
#define init_1_5_2_14
void FASTCALL call_1_5_2_14(SH2_struct* sh)
{
  SH2movls4(sh, 5,2, 14);
}
#define init_1_5_2_15
void FASTCALL call_1_5_2_15(SH2_struct* sh)
{
  SH2movls4(sh, 5,2, 15);
}
#define init_1_5_3_0
void FASTCALL call_1_5_3_0(SH2_struct* sh)
{
  SH2movls4(sh, 5,3, 0);
}
#define init_1_5_3_1
void FASTCALL call_1_5_3_1(SH2_struct* sh)
{
  SH2movls4(sh, 5,3, 1);
}
#define init_1_5_3_2
void FASTCALL call_1_5_3_2(SH2_struct* sh)
{
  SH2movls4(sh, 5,3, 2);
}
#define init_1_5_3_3
void FASTCALL call_1_5_3_3(SH2_struct* sh)
{
  SH2movls4(sh, 5,3, 3);
}
#define init_1_5_3_4
void FASTCALL call_1_5_3_4(SH2_struct* sh)
{
  SH2movls4(sh, 5,3, 4);
}
#define init_1_5_3_5
void FASTCALL call_1_5_3_5(SH2_struct* sh)
{
  SH2movls4(sh, 5,3, 5);
}
#define init_1_5_3_6
void FASTCALL call_1_5_3_6(SH2_struct* sh)
{
  SH2movls4(sh, 5,3, 6);
}
#define init_1_5_3_7
void FASTCALL call_1_5_3_7(SH2_struct* sh)
{
  SH2movls4(sh, 5,3, 7);
}
#define init_1_5_3_8
void FASTCALL call_1_5_3_8(SH2_struct* sh)
{
  SH2movls4(sh, 5,3, 8);
}
#define init_1_5_3_9
void FASTCALL call_1_5_3_9(SH2_struct* sh)
{
  SH2movls4(sh, 5,3, 9);
}
#define init_1_5_3_10
void FASTCALL call_1_5_3_10(SH2_struct* sh)
{
  SH2movls4(sh, 5,3, 10);
}
#define init_1_5_3_11
void FASTCALL call_1_5_3_11(SH2_struct* sh)
{
  SH2movls4(sh, 5,3, 11);
}
#define init_1_5_3_12
void FASTCALL call_1_5_3_12(SH2_struct* sh)
{
  SH2movls4(sh, 5,3, 12);
}
#define init_1_5_3_13
void FASTCALL call_1_5_3_13(SH2_struct* sh)
{
  SH2movls4(sh, 5,3, 13);
}
#define init_1_5_3_14
void FASTCALL call_1_5_3_14(SH2_struct* sh)
{
  SH2movls4(sh, 5,3, 14);
}
#define init_1_5_3_15
void FASTCALL call_1_5_3_15(SH2_struct* sh)
{
  SH2movls4(sh, 5,3, 15);
}
#define init_1_5_4_0
void FASTCALL call_1_5_4_0(SH2_struct* sh)
{
  SH2movls4(sh, 5,4, 0);
}
#define init_1_5_4_1
void FASTCALL call_1_5_4_1(SH2_struct* sh)
{
  SH2movls4(sh, 5,4, 1);
}
#define init_1_5_4_2
void FASTCALL call_1_5_4_2(SH2_struct* sh)
{
  SH2movls4(sh, 5,4, 2);
}
#define init_1_5_4_3
void FASTCALL call_1_5_4_3(SH2_struct* sh)
{
  SH2movls4(sh, 5,4, 3);
}
#define init_1_5_4_4
void FASTCALL call_1_5_4_4(SH2_struct* sh)
{
  SH2movls4(sh, 5,4, 4);
}
#define init_1_5_4_5
void FASTCALL call_1_5_4_5(SH2_struct* sh)
{
  SH2movls4(sh, 5,4, 5);
}
#define init_1_5_4_6
void FASTCALL call_1_5_4_6(SH2_struct* sh)
{
  SH2movls4(sh, 5,4, 6);
}
#define init_1_5_4_7
void FASTCALL call_1_5_4_7(SH2_struct* sh)
{
  SH2movls4(sh, 5,4, 7);
}
#define init_1_5_4_8
void FASTCALL call_1_5_4_8(SH2_struct* sh)
{
  SH2movls4(sh, 5,4, 8);
}
#define init_1_5_4_9
void FASTCALL call_1_5_4_9(SH2_struct* sh)
{
  SH2movls4(sh, 5,4, 9);
}
#define init_1_5_4_10
void FASTCALL call_1_5_4_10(SH2_struct* sh)
{
  SH2movls4(sh, 5,4, 10);
}
#define init_1_5_4_11
void FASTCALL call_1_5_4_11(SH2_struct* sh)
{
  SH2movls4(sh, 5,4, 11);
}
#define init_1_5_4_12
void FASTCALL call_1_5_4_12(SH2_struct* sh)
{
  SH2movls4(sh, 5,4, 12);
}
#define init_1_5_4_13
void FASTCALL call_1_5_4_13(SH2_struct* sh)
{
  SH2movls4(sh, 5,4, 13);
}
#define init_1_5_4_14
void FASTCALL call_1_5_4_14(SH2_struct* sh)
{
  SH2movls4(sh, 5,4, 14);
}
#define init_1_5_4_15
void FASTCALL call_1_5_4_15(SH2_struct* sh)
{
  SH2movls4(sh, 5,4, 15);
}
#define init_1_5_5_0
void FASTCALL call_1_5_5_0(SH2_struct* sh)
{
  SH2movls4(sh, 5,5, 0);
}
#define init_1_5_5_1
void FASTCALL call_1_5_5_1(SH2_struct* sh)
{
  SH2movls4(sh, 5,5, 1);
}
#define init_1_5_5_2
void FASTCALL call_1_5_5_2(SH2_struct* sh)
{
  SH2movls4(sh, 5,5, 2);
}
#define init_1_5_5_3
void FASTCALL call_1_5_5_3(SH2_struct* sh)
{
  SH2movls4(sh, 5,5, 3);
}
#define init_1_5_5_4
void FASTCALL call_1_5_5_4(SH2_struct* sh)
{
  SH2movls4(sh, 5,5, 4);
}
#define init_1_5_5_5
void FASTCALL call_1_5_5_5(SH2_struct* sh)
{
  SH2movls4(sh, 5,5, 5);
}
#define init_1_5_5_6
void FASTCALL call_1_5_5_6(SH2_struct* sh)
{
  SH2movls4(sh, 5,5, 6);
}
#define init_1_5_5_7
void FASTCALL call_1_5_5_7(SH2_struct* sh)
{
  SH2movls4(sh, 5,5, 7);
}
#define init_1_5_5_8
void FASTCALL call_1_5_5_8(SH2_struct* sh)
{
  SH2movls4(sh, 5,5, 8);
}
#define init_1_5_5_9
void FASTCALL call_1_5_5_9(SH2_struct* sh)
{
  SH2movls4(sh, 5,5, 9);
}
#define init_1_5_5_10
void FASTCALL call_1_5_5_10(SH2_struct* sh)
{
  SH2movls4(sh, 5,5, 10);
}
#define init_1_5_5_11
void FASTCALL call_1_5_5_11(SH2_struct* sh)
{
  SH2movls4(sh, 5,5, 11);
}
#define init_1_5_5_12
void FASTCALL call_1_5_5_12(SH2_struct* sh)
{
  SH2movls4(sh, 5,5, 12);
}
#define init_1_5_5_13
void FASTCALL call_1_5_5_13(SH2_struct* sh)
{
  SH2movls4(sh, 5,5, 13);
}
#define init_1_5_5_14
void FASTCALL call_1_5_5_14(SH2_struct* sh)
{
  SH2movls4(sh, 5,5, 14);
}
#define init_1_5_5_15
void FASTCALL call_1_5_5_15(SH2_struct* sh)
{
  SH2movls4(sh, 5,5, 15);
}
#define init_1_5_6_0
void FASTCALL call_1_5_6_0(SH2_struct* sh)
{
  SH2movls4(sh, 5,6, 0);
}
#define init_1_5_6_1
void FASTCALL call_1_5_6_1(SH2_struct* sh)
{
  SH2movls4(sh, 5,6, 1);
}
#define init_1_5_6_2
void FASTCALL call_1_5_6_2(SH2_struct* sh)
{
  SH2movls4(sh, 5,6, 2);
}
#define init_1_5_6_3
void FASTCALL call_1_5_6_3(SH2_struct* sh)
{
  SH2movls4(sh, 5,6, 3);
}
#define init_1_5_6_4
void FASTCALL call_1_5_6_4(SH2_struct* sh)
{
  SH2movls4(sh, 5,6, 4);
}
#define init_1_5_6_5
void FASTCALL call_1_5_6_5(SH2_struct* sh)
{
  SH2movls4(sh, 5,6, 5);
}
#define init_1_5_6_6
void FASTCALL call_1_5_6_6(SH2_struct* sh)
{
  SH2movls4(sh, 5,6, 6);
}
#define init_1_5_6_7
void FASTCALL call_1_5_6_7(SH2_struct* sh)
{
  SH2movls4(sh, 5,6, 7);
}
#define init_1_5_6_8
void FASTCALL call_1_5_6_8(SH2_struct* sh)
{
  SH2movls4(sh, 5,6, 8);
}
#define init_1_5_6_9
void FASTCALL call_1_5_6_9(SH2_struct* sh)
{
  SH2movls4(sh, 5,6, 9);
}
#define init_1_5_6_10
void FASTCALL call_1_5_6_10(SH2_struct* sh)
{
  SH2movls4(sh, 5,6, 10);
}
#define init_1_5_6_11
void FASTCALL call_1_5_6_11(SH2_struct* sh)
{
  SH2movls4(sh, 5,6, 11);
}
#define init_1_5_6_12
void FASTCALL call_1_5_6_12(SH2_struct* sh)
{
  SH2movls4(sh, 5,6, 12);
}
#define init_1_5_6_13
void FASTCALL call_1_5_6_13(SH2_struct* sh)
{
  SH2movls4(sh, 5,6, 13);
}
#define init_1_5_6_14
void FASTCALL call_1_5_6_14(SH2_struct* sh)
{
  SH2movls4(sh, 5,6, 14);
}
#define init_1_5_6_15
void FASTCALL call_1_5_6_15(SH2_struct* sh)
{
  SH2movls4(sh, 5,6, 15);
}
#define init_1_5_7_0
void FASTCALL call_1_5_7_0(SH2_struct* sh)
{
  SH2movls4(sh, 5,7, 0);
}
#define init_1_5_7_1
void FASTCALL call_1_5_7_1(SH2_struct* sh)
{
  SH2movls4(sh, 5,7, 1);
}
#define init_1_5_7_2
void FASTCALL call_1_5_7_2(SH2_struct* sh)
{
  SH2movls4(sh, 5,7, 2);
}
#define init_1_5_7_3
void FASTCALL call_1_5_7_3(SH2_struct* sh)
{
  SH2movls4(sh, 5,7, 3);
}
#define init_1_5_7_4
void FASTCALL call_1_5_7_4(SH2_struct* sh)
{
  SH2movls4(sh, 5,7, 4);
}
#define init_1_5_7_5
void FASTCALL call_1_5_7_5(SH2_struct* sh)
{
  SH2movls4(sh, 5,7, 5);
}
#define init_1_5_7_6
void FASTCALL call_1_5_7_6(SH2_struct* sh)
{
  SH2movls4(sh, 5,7, 6);
}
#define init_1_5_7_7
void FASTCALL call_1_5_7_7(SH2_struct* sh)
{
  SH2movls4(sh, 5,7, 7);
}
#define init_1_5_7_8
void FASTCALL call_1_5_7_8(SH2_struct* sh)
{
  SH2movls4(sh, 5,7, 8);
}
#define init_1_5_7_9
void FASTCALL call_1_5_7_9(SH2_struct* sh)
{
  SH2movls4(sh, 5,7, 9);
}
#define init_1_5_7_10
void FASTCALL call_1_5_7_10(SH2_struct* sh)
{
  SH2movls4(sh, 5,7, 10);
}
#define init_1_5_7_11
void FASTCALL call_1_5_7_11(SH2_struct* sh)
{
  SH2movls4(sh, 5,7, 11);
}
#define init_1_5_7_12
void FASTCALL call_1_5_7_12(SH2_struct* sh)
{
  SH2movls4(sh, 5,7, 12);
}
#define init_1_5_7_13
void FASTCALL call_1_5_7_13(SH2_struct* sh)
{
  SH2movls4(sh, 5,7, 13);
}
#define init_1_5_7_14
void FASTCALL call_1_5_7_14(SH2_struct* sh)
{
  SH2movls4(sh, 5,7, 14);
}
#define init_1_5_7_15
void FASTCALL call_1_5_7_15(SH2_struct* sh)
{
  SH2movls4(sh, 5,7, 15);
}
#define init_1_5_8_0
void FASTCALL call_1_5_8_0(SH2_struct* sh)
{
  SH2movls4(sh, 5,8, 0);
}
#define init_1_5_8_1
void FASTCALL call_1_5_8_1(SH2_struct* sh)
{
  SH2movls4(sh, 5,8, 1);
}
#define init_1_5_8_2
void FASTCALL call_1_5_8_2(SH2_struct* sh)
{
  SH2movls4(sh, 5,8, 2);
}
#define init_1_5_8_3
void FASTCALL call_1_5_8_3(SH2_struct* sh)
{
  SH2movls4(sh, 5,8, 3);
}
#define init_1_5_8_4
void FASTCALL call_1_5_8_4(SH2_struct* sh)
{
  SH2movls4(sh, 5,8, 4);
}
#define init_1_5_8_5
void FASTCALL call_1_5_8_5(SH2_struct* sh)
{
  SH2movls4(sh, 5,8, 5);
}
#define init_1_5_8_6
void FASTCALL call_1_5_8_6(SH2_struct* sh)
{
  SH2movls4(sh, 5,8, 6);
}
#define init_1_5_8_7
void FASTCALL call_1_5_8_7(SH2_struct* sh)
{
  SH2movls4(sh, 5,8, 7);
}
#define init_1_5_8_8
void FASTCALL call_1_5_8_8(SH2_struct* sh)
{
  SH2movls4(sh, 5,8, 8);
}
#define init_1_5_8_9
void FASTCALL call_1_5_8_9(SH2_struct* sh)
{
  SH2movls4(sh, 5,8, 9);
}
#define init_1_5_8_10
void FASTCALL call_1_5_8_10(SH2_struct* sh)
{
  SH2movls4(sh, 5,8, 10);
}
#define init_1_5_8_11
void FASTCALL call_1_5_8_11(SH2_struct* sh)
{
  SH2movls4(sh, 5,8, 11);
}
#define init_1_5_8_12
void FASTCALL call_1_5_8_12(SH2_struct* sh)
{
  SH2movls4(sh, 5,8, 12);
}
#define init_1_5_8_13
void FASTCALL call_1_5_8_13(SH2_struct* sh)
{
  SH2movls4(sh, 5,8, 13);
}
#define init_1_5_8_14
void FASTCALL call_1_5_8_14(SH2_struct* sh)
{
  SH2movls4(sh, 5,8, 14);
}
#define init_1_5_8_15
void FASTCALL call_1_5_8_15(SH2_struct* sh)
{
  SH2movls4(sh, 5,8, 15);
}
#define init_1_5_9_0
void FASTCALL call_1_5_9_0(SH2_struct* sh)
{
  SH2movls4(sh, 5,9, 0);
}
#define init_1_5_9_1
void FASTCALL call_1_5_9_1(SH2_struct* sh)
{
  SH2movls4(sh, 5,9, 1);
}
#define init_1_5_9_2
void FASTCALL call_1_5_9_2(SH2_struct* sh)
{
  SH2movls4(sh, 5,9, 2);
}
#define init_1_5_9_3
void FASTCALL call_1_5_9_3(SH2_struct* sh)
{
  SH2movls4(sh, 5,9, 3);
}
#define init_1_5_9_4
void FASTCALL call_1_5_9_4(SH2_struct* sh)
{
  SH2movls4(sh, 5,9, 4);
}
#define init_1_5_9_5
void FASTCALL call_1_5_9_5(SH2_struct* sh)
{
  SH2movls4(sh, 5,9, 5);
}
#define init_1_5_9_6
void FASTCALL call_1_5_9_6(SH2_struct* sh)
{
  SH2movls4(sh, 5,9, 6);
}
#define init_1_5_9_7
void FASTCALL call_1_5_9_7(SH2_struct* sh)
{
  SH2movls4(sh, 5,9, 7);
}
#define init_1_5_9_8
void FASTCALL call_1_5_9_8(SH2_struct* sh)
{
  SH2movls4(sh, 5,9, 8);
}
#define init_1_5_9_9
void FASTCALL call_1_5_9_9(SH2_struct* sh)
{
  SH2movls4(sh, 5,9, 9);
}
#define init_1_5_9_10
void FASTCALL call_1_5_9_10(SH2_struct* sh)
{
  SH2movls4(sh, 5,9, 10);
}
#define init_1_5_9_11
void FASTCALL call_1_5_9_11(SH2_struct* sh)
{
  SH2movls4(sh, 5,9, 11);
}
#define init_1_5_9_12
void FASTCALL call_1_5_9_12(SH2_struct* sh)
{
  SH2movls4(sh, 5,9, 12);
}
#define init_1_5_9_13
void FASTCALL call_1_5_9_13(SH2_struct* sh)
{
  SH2movls4(sh, 5,9, 13);
}
#define init_1_5_9_14
void FASTCALL call_1_5_9_14(SH2_struct* sh)
{
  SH2movls4(sh, 5,9, 14);
}
#define init_1_5_9_15
void FASTCALL call_1_5_9_15(SH2_struct* sh)
{
  SH2movls4(sh, 5,9, 15);
}
#define init_1_5_10_0
void FASTCALL call_1_5_10_0(SH2_struct* sh)
{
  SH2movls4(sh, 5,10, 0);
}
#define init_1_5_10_1
void FASTCALL call_1_5_10_1(SH2_struct* sh)
{
  SH2movls4(sh, 5,10, 1);
}
#define init_1_5_10_2
void FASTCALL call_1_5_10_2(SH2_struct* sh)
{
  SH2movls4(sh, 5,10, 2);
}
#define init_1_5_10_3
void FASTCALL call_1_5_10_3(SH2_struct* sh)
{
  SH2movls4(sh, 5,10, 3);
}
#define init_1_5_10_4
void FASTCALL call_1_5_10_4(SH2_struct* sh)
{
  SH2movls4(sh, 5,10, 4);
}
#define init_1_5_10_5
void FASTCALL call_1_5_10_5(SH2_struct* sh)
{
  SH2movls4(sh, 5,10, 5);
}
#define init_1_5_10_6
void FASTCALL call_1_5_10_6(SH2_struct* sh)
{
  SH2movls4(sh, 5,10, 6);
}
#define init_1_5_10_7
void FASTCALL call_1_5_10_7(SH2_struct* sh)
{
  SH2movls4(sh, 5,10, 7);
}
#define init_1_5_10_8
void FASTCALL call_1_5_10_8(SH2_struct* sh)
{
  SH2movls4(sh, 5,10, 8);
}
#define init_1_5_10_9
void FASTCALL call_1_5_10_9(SH2_struct* sh)
{
  SH2movls4(sh, 5,10, 9);
}
#define init_1_5_10_10
void FASTCALL call_1_5_10_10(SH2_struct* sh)
{
  SH2movls4(sh, 5,10, 10);
}
#define init_1_5_10_11
void FASTCALL call_1_5_10_11(SH2_struct* sh)
{
  SH2movls4(sh, 5,10, 11);
}
#define init_1_5_10_12
void FASTCALL call_1_5_10_12(SH2_struct* sh)
{
  SH2movls4(sh, 5,10, 12);
}
#define init_1_5_10_13
void FASTCALL call_1_5_10_13(SH2_struct* sh)
{
  SH2movls4(sh, 5,10, 13);
}
#define init_1_5_10_14
void FASTCALL call_1_5_10_14(SH2_struct* sh)
{
  SH2movls4(sh, 5,10, 14);
}
#define init_1_5_10_15
void FASTCALL call_1_5_10_15(SH2_struct* sh)
{
  SH2movls4(sh, 5,10, 15);
}
#define init_1_5_11_0
void FASTCALL call_1_5_11_0(SH2_struct* sh)
{
  SH2movls4(sh, 5,11, 0);
}
#define init_1_5_11_1
void FASTCALL call_1_5_11_1(SH2_struct* sh)
{
  SH2movls4(sh, 5,11, 1);
}
#define init_1_5_11_2
void FASTCALL call_1_5_11_2(SH2_struct* sh)
{
  SH2movls4(sh, 5,11, 2);
}
#define init_1_5_11_3
void FASTCALL call_1_5_11_3(SH2_struct* sh)
{
  SH2movls4(sh, 5,11, 3);
}
#define init_1_5_11_4
void FASTCALL call_1_5_11_4(SH2_struct* sh)
{
  SH2movls4(sh, 5,11, 4);
}
#define init_1_5_11_5
void FASTCALL call_1_5_11_5(SH2_struct* sh)
{
  SH2movls4(sh, 5,11, 5);
}
#define init_1_5_11_6
void FASTCALL call_1_5_11_6(SH2_struct* sh)
{
  SH2movls4(sh, 5,11, 6);
}
#define init_1_5_11_7
void FASTCALL call_1_5_11_7(SH2_struct* sh)
{
  SH2movls4(sh, 5,11, 7);
}
#define init_1_5_11_8
void FASTCALL call_1_5_11_8(SH2_struct* sh)
{
  SH2movls4(sh, 5,11, 8);
}
#define init_1_5_11_9
void FASTCALL call_1_5_11_9(SH2_struct* sh)
{
  SH2movls4(sh, 5,11, 9);
}
#define init_1_5_11_10
void FASTCALL call_1_5_11_10(SH2_struct* sh)
{
  SH2movls4(sh, 5,11, 10);
}
#define init_1_5_11_11
void FASTCALL call_1_5_11_11(SH2_struct* sh)
{
  SH2movls4(sh, 5,11, 11);
}
#define init_1_5_11_12
void FASTCALL call_1_5_11_12(SH2_struct* sh)
{
  SH2movls4(sh, 5,11, 12);
}
#define init_1_5_11_13
void FASTCALL call_1_5_11_13(SH2_struct* sh)
{
  SH2movls4(sh, 5,11, 13);
}
#define init_1_5_11_14
void FASTCALL call_1_5_11_14(SH2_struct* sh)
{
  SH2movls4(sh, 5,11, 14);
}
#define init_1_5_11_15
void FASTCALL call_1_5_11_15(SH2_struct* sh)
{
  SH2movls4(sh, 5,11, 15);
}
#define init_1_5_12_0
void FASTCALL call_1_5_12_0(SH2_struct* sh)
{
  SH2movls4(sh, 5,12, 0);
}
#define init_1_5_12_1
void FASTCALL call_1_5_12_1(SH2_struct* sh)
{
  SH2movls4(sh, 5,12, 1);
}
#define init_1_5_12_2
void FASTCALL call_1_5_12_2(SH2_struct* sh)
{
  SH2movls4(sh, 5,12, 2);
}
#define init_1_5_12_3
void FASTCALL call_1_5_12_3(SH2_struct* sh)
{
  SH2movls4(sh, 5,12, 3);
}
#define init_1_5_12_4
void FASTCALL call_1_5_12_4(SH2_struct* sh)
{
  SH2movls4(sh, 5,12, 4);
}
#define init_1_5_12_5
void FASTCALL call_1_5_12_5(SH2_struct* sh)
{
  SH2movls4(sh, 5,12, 5);
}
#define init_1_5_12_6
void FASTCALL call_1_5_12_6(SH2_struct* sh)
{
  SH2movls4(sh, 5,12, 6);
}
#define init_1_5_12_7
void FASTCALL call_1_5_12_7(SH2_struct* sh)
{
  SH2movls4(sh, 5,12, 7);
}
#define init_1_5_12_8
void FASTCALL call_1_5_12_8(SH2_struct* sh)
{
  SH2movls4(sh, 5,12, 8);
}
#define init_1_5_12_9
void FASTCALL call_1_5_12_9(SH2_struct* sh)
{
  SH2movls4(sh, 5,12, 9);
}
#define init_1_5_12_10
void FASTCALL call_1_5_12_10(SH2_struct* sh)
{
  SH2movls4(sh, 5,12, 10);
}
#define init_1_5_12_11
void FASTCALL call_1_5_12_11(SH2_struct* sh)
{
  SH2movls4(sh, 5,12, 11);
}
#define init_1_5_12_12
void FASTCALL call_1_5_12_12(SH2_struct* sh)
{
  SH2movls4(sh, 5,12, 12);
}
#define init_1_5_12_13
void FASTCALL call_1_5_12_13(SH2_struct* sh)
{
  SH2movls4(sh, 5,12, 13);
}
#define init_1_5_12_14
void FASTCALL call_1_5_12_14(SH2_struct* sh)
{
  SH2movls4(sh, 5,12, 14);
}
#define init_1_5_12_15
void FASTCALL call_1_5_12_15(SH2_struct* sh)
{
  SH2movls4(sh, 5,12, 15);
}
#define init_1_5_13_0
void FASTCALL call_1_5_13_0(SH2_struct* sh)
{
  SH2movls4(sh, 5,13, 0);
}
#define init_1_5_13_1
void FASTCALL call_1_5_13_1(SH2_struct* sh)
{
  SH2movls4(sh, 5,13, 1);
}
#define init_1_5_13_2
void FASTCALL call_1_5_13_2(SH2_struct* sh)
{
  SH2movls4(sh, 5,13, 2);
}
#define init_1_5_13_3
void FASTCALL call_1_5_13_3(SH2_struct* sh)
{
  SH2movls4(sh, 5,13, 3);
}
#define init_1_5_13_4
void FASTCALL call_1_5_13_4(SH2_struct* sh)
{
  SH2movls4(sh, 5,13, 4);
}
#define init_1_5_13_5
void FASTCALL call_1_5_13_5(SH2_struct* sh)
{
  SH2movls4(sh, 5,13, 5);
}
#define init_1_5_13_6
void FASTCALL call_1_5_13_6(SH2_struct* sh)
{
  SH2movls4(sh, 5,13, 6);
}
#define init_1_5_13_7
void FASTCALL call_1_5_13_7(SH2_struct* sh)
{
  SH2movls4(sh, 5,13, 7);
}
#define init_1_5_13_8
void FASTCALL call_1_5_13_8(SH2_struct* sh)
{
  SH2movls4(sh, 5,13, 8);
}
#define init_1_5_13_9
void FASTCALL call_1_5_13_9(SH2_struct* sh)
{
  SH2movls4(sh, 5,13, 9);
}
#define init_1_5_13_10
void FASTCALL call_1_5_13_10(SH2_struct* sh)
{
  SH2movls4(sh, 5,13, 10);
}
#define init_1_5_13_11
void FASTCALL call_1_5_13_11(SH2_struct* sh)
{
  SH2movls4(sh, 5,13, 11);
}
#define init_1_5_13_12
void FASTCALL call_1_5_13_12(SH2_struct* sh)
{
  SH2movls4(sh, 5,13, 12);
}
#define init_1_5_13_13
void FASTCALL call_1_5_13_13(SH2_struct* sh)
{
  SH2movls4(sh, 5,13, 13);
}
#define init_1_5_13_14
void FASTCALL call_1_5_13_14(SH2_struct* sh)
{
  SH2movls4(sh, 5,13, 14);
}
#define init_1_5_13_15
void FASTCALL call_1_5_13_15(SH2_struct* sh)
{
  SH2movls4(sh, 5,13, 15);
}
#define init_1_5_14_0
void FASTCALL call_1_5_14_0(SH2_struct* sh)
{
  SH2movls4(sh, 5,14, 0);
}
#define init_1_5_14_1
void FASTCALL call_1_5_14_1(SH2_struct* sh)
{
  SH2movls4(sh, 5,14, 1);
}
#define init_1_5_14_2
void FASTCALL call_1_5_14_2(SH2_struct* sh)
{
  SH2movls4(sh, 5,14, 2);
}
#define init_1_5_14_3
void FASTCALL call_1_5_14_3(SH2_struct* sh)
{
  SH2movls4(sh, 5,14, 3);
}
#define init_1_5_14_4
void FASTCALL call_1_5_14_4(SH2_struct* sh)
{
  SH2movls4(sh, 5,14, 4);
}
#define init_1_5_14_5
void FASTCALL call_1_5_14_5(SH2_struct* sh)
{
  SH2movls4(sh, 5,14, 5);
}
#define init_1_5_14_6
void FASTCALL call_1_5_14_6(SH2_struct* sh)
{
  SH2movls4(sh, 5,14, 6);
}
#define init_1_5_14_7
void FASTCALL call_1_5_14_7(SH2_struct* sh)
{
  SH2movls4(sh, 5,14, 7);
}
#define init_1_5_14_8
void FASTCALL call_1_5_14_8(SH2_struct* sh)
{
  SH2movls4(sh, 5,14, 8);
}
#define init_1_5_14_9
void FASTCALL call_1_5_14_9(SH2_struct* sh)
{
  SH2movls4(sh, 5,14, 9);
}
#define init_1_5_14_10
void FASTCALL call_1_5_14_10(SH2_struct* sh)
{
  SH2movls4(sh, 5,14, 10);
}
#define init_1_5_14_11
void FASTCALL call_1_5_14_11(SH2_struct* sh)
{
  SH2movls4(sh, 5,14, 11);
}
#define init_1_5_14_12
void FASTCALL call_1_5_14_12(SH2_struct* sh)
{
  SH2movls4(sh, 5,14, 12);
}
#define init_1_5_14_13
void FASTCALL call_1_5_14_13(SH2_struct* sh)
{
  SH2movls4(sh, 5,14, 13);
}
#define init_1_5_14_14
void FASTCALL call_1_5_14_14(SH2_struct* sh)
{
  SH2movls4(sh, 5,14, 14);
}
#define init_1_5_14_15
void FASTCALL call_1_5_14_15(SH2_struct* sh)
{
  SH2movls4(sh, 5,14, 15);
}
#define init_1_5_15_0
void FASTCALL call_1_5_15_0(SH2_struct* sh)
{
  SH2movls4(sh, 5,15, 0);
}
#define init_1_5_15_1
void FASTCALL call_1_5_15_1(SH2_struct* sh)
{
  SH2movls4(sh, 5,15, 1);
}
#define init_1_5_15_2
void FASTCALL call_1_5_15_2(SH2_struct* sh)
{
  SH2movls4(sh, 5,15, 2);
}
#define init_1_5_15_3
void FASTCALL call_1_5_15_3(SH2_struct* sh)
{
  SH2movls4(sh, 5,15, 3);
}
#define init_1_5_15_4
void FASTCALL call_1_5_15_4(SH2_struct* sh)
{
  SH2movls4(sh, 5,15, 4);
}
#define init_1_5_15_5
void FASTCALL call_1_5_15_5(SH2_struct* sh)
{
  SH2movls4(sh, 5,15, 5);
}
#define init_1_5_15_6
void FASTCALL call_1_5_15_6(SH2_struct* sh)
{
  SH2movls4(sh, 5,15, 6);
}
#define init_1_5_15_7
void FASTCALL call_1_5_15_7(SH2_struct* sh)
{
  SH2movls4(sh, 5,15, 7);
}
#define init_1_5_15_8
void FASTCALL call_1_5_15_8(SH2_struct* sh)
{
  SH2movls4(sh, 5,15, 8);
}
#define init_1_5_15_9
void FASTCALL call_1_5_15_9(SH2_struct* sh)
{
  SH2movls4(sh, 5,15, 9);
}
#define init_1_5_15_10
void FASTCALL call_1_5_15_10(SH2_struct* sh)
{
  SH2movls4(sh, 5,15, 10);
}
#define init_1_5_15_11
void FASTCALL call_1_5_15_11(SH2_struct* sh)
{
  SH2movls4(sh, 5,15, 11);
}
#define init_1_5_15_12
void FASTCALL call_1_5_15_12(SH2_struct* sh)
{
  SH2movls4(sh, 5,15, 12);
}
#define init_1_5_15_13
void FASTCALL call_1_5_15_13(SH2_struct* sh)
{
  SH2movls4(sh, 5,15, 13);
}
#define init_1_5_15_14
void FASTCALL call_1_5_15_14(SH2_struct* sh)
{
  SH2movls4(sh, 5,15, 14);
}
#define init_1_5_15_15
void FASTCALL call_1_5_15_15(SH2_struct* sh)
{
  SH2movls4(sh, 5,15, 15);
}
#define init_1_6_0_0
void FASTCALL call_1_6_0_0(SH2_struct* sh)
{
  SH2movls4(sh, 6,0, 0);
}
#define init_1_6_0_1
void FASTCALL call_1_6_0_1(SH2_struct* sh)
{
  SH2movls4(sh, 6,0, 1);
}
#define init_1_6_0_2
void FASTCALL call_1_6_0_2(SH2_struct* sh)
{
  SH2movls4(sh, 6,0, 2);
}
#define init_1_6_0_3
void FASTCALL call_1_6_0_3(SH2_struct* sh)
{
  SH2movls4(sh, 6,0, 3);
}
#define init_1_6_0_4
void FASTCALL call_1_6_0_4(SH2_struct* sh)
{
  SH2movls4(sh, 6,0, 4);
}
#define init_1_6_0_5
void FASTCALL call_1_6_0_5(SH2_struct* sh)
{
  SH2movls4(sh, 6,0, 5);
}
#define init_1_6_0_6
void FASTCALL call_1_6_0_6(SH2_struct* sh)
{
  SH2movls4(sh, 6,0, 6);
}
#define init_1_6_0_7
void FASTCALL call_1_6_0_7(SH2_struct* sh)
{
  SH2movls4(sh, 6,0, 7);
}
#define init_1_6_0_8
void FASTCALL call_1_6_0_8(SH2_struct* sh)
{
  SH2movls4(sh, 6,0, 8);
}
#define init_1_6_0_9
void FASTCALL call_1_6_0_9(SH2_struct* sh)
{
  SH2movls4(sh, 6,0, 9);
}
#define init_1_6_0_10
void FASTCALL call_1_6_0_10(SH2_struct* sh)
{
  SH2movls4(sh, 6,0, 10);
}
#define init_1_6_0_11
void FASTCALL call_1_6_0_11(SH2_struct* sh)
{
  SH2movls4(sh, 6,0, 11);
}
#define init_1_6_0_12
void FASTCALL call_1_6_0_12(SH2_struct* sh)
{
  SH2movls4(sh, 6,0, 12);
}
#define init_1_6_0_13
void FASTCALL call_1_6_0_13(SH2_struct* sh)
{
  SH2movls4(sh, 6,0, 13);
}
#define init_1_6_0_14
void FASTCALL call_1_6_0_14(SH2_struct* sh)
{
  SH2movls4(sh, 6,0, 14);
}
#define init_1_6_0_15
void FASTCALL call_1_6_0_15(SH2_struct* sh)
{
  SH2movls4(sh, 6,0, 15);
}
#define init_1_6_1_0
void FASTCALL call_1_6_1_0(SH2_struct* sh)
{
  SH2movls4(sh, 6,1, 0);
}
#define init_1_6_1_1
void FASTCALL call_1_6_1_1(SH2_struct* sh)
{
  SH2movls4(sh, 6,1, 1);
}
#define init_1_6_1_2
void FASTCALL call_1_6_1_2(SH2_struct* sh)
{
  SH2movls4(sh, 6,1, 2);
}
#define init_1_6_1_3
void FASTCALL call_1_6_1_3(SH2_struct* sh)
{
  SH2movls4(sh, 6,1, 3);
}
#define init_1_6_1_4
void FASTCALL call_1_6_1_4(SH2_struct* sh)
{
  SH2movls4(sh, 6,1, 4);
}
#define init_1_6_1_5
void FASTCALL call_1_6_1_5(SH2_struct* sh)
{
  SH2movls4(sh, 6,1, 5);
}
#define init_1_6_1_6
void FASTCALL call_1_6_1_6(SH2_struct* sh)
{
  SH2movls4(sh, 6,1, 6);
}
#define init_1_6_1_7
void FASTCALL call_1_6_1_7(SH2_struct* sh)
{
  SH2movls4(sh, 6,1, 7);
}
#define init_1_6_1_8
void FASTCALL call_1_6_1_8(SH2_struct* sh)
{
  SH2movls4(sh, 6,1, 8);
}
#define init_1_6_1_9
void FASTCALL call_1_6_1_9(SH2_struct* sh)
{
  SH2movls4(sh, 6,1, 9);
}
#define init_1_6_1_10
void FASTCALL call_1_6_1_10(SH2_struct* sh)
{
  SH2movls4(sh, 6,1, 10);
}
#define init_1_6_1_11
void FASTCALL call_1_6_1_11(SH2_struct* sh)
{
  SH2movls4(sh, 6,1, 11);
}
#define init_1_6_1_12
void FASTCALL call_1_6_1_12(SH2_struct* sh)
{
  SH2movls4(sh, 6,1, 12);
}
#define init_1_6_1_13
void FASTCALL call_1_6_1_13(SH2_struct* sh)
{
  SH2movls4(sh, 6,1, 13);
}
#define init_1_6_1_14
void FASTCALL call_1_6_1_14(SH2_struct* sh)
{
  SH2movls4(sh, 6,1, 14);
}
#define init_1_6_1_15
void FASTCALL call_1_6_1_15(SH2_struct* sh)
{
  SH2movls4(sh, 6,1, 15);
}
#define init_1_6_2_0
void FASTCALL call_1_6_2_0(SH2_struct* sh)
{
  SH2movls4(sh, 6,2, 0);
}
#define init_1_6_2_1
void FASTCALL call_1_6_2_1(SH2_struct* sh)
{
  SH2movls4(sh, 6,2, 1);
}
#define init_1_6_2_2
void FASTCALL call_1_6_2_2(SH2_struct* sh)
{
  SH2movls4(sh, 6,2, 2);
}
#define init_1_6_2_3
void FASTCALL call_1_6_2_3(SH2_struct* sh)
{
  SH2movls4(sh, 6,2, 3);
}
#define init_1_6_2_4
void FASTCALL call_1_6_2_4(SH2_struct* sh)
{
  SH2movls4(sh, 6,2, 4);
}
#define init_1_6_2_5
void FASTCALL call_1_6_2_5(SH2_struct* sh)
{
  SH2movls4(sh, 6,2, 5);
}
#define init_1_6_2_6
void FASTCALL call_1_6_2_6(SH2_struct* sh)
{
  SH2movls4(sh, 6,2, 6);
}
#define init_1_6_2_7
void FASTCALL call_1_6_2_7(SH2_struct* sh)
{
  SH2movls4(sh, 6,2, 7);
}
#define init_1_6_2_8
void FASTCALL call_1_6_2_8(SH2_struct* sh)
{
  SH2movls4(sh, 6,2, 8);
}
#define init_1_6_2_9
void FASTCALL call_1_6_2_9(SH2_struct* sh)
{
  SH2movls4(sh, 6,2, 9);
}
#define init_1_6_2_10
void FASTCALL call_1_6_2_10(SH2_struct* sh)
{
  SH2movls4(sh, 6,2, 10);
}
#define init_1_6_2_11
void FASTCALL call_1_6_2_11(SH2_struct* sh)
{
  SH2movls4(sh, 6,2, 11);
}
#define init_1_6_2_12
void FASTCALL call_1_6_2_12(SH2_struct* sh)
{
  SH2movls4(sh, 6,2, 12);
}
#define init_1_6_2_13
void FASTCALL call_1_6_2_13(SH2_struct* sh)
{
  SH2movls4(sh, 6,2, 13);
}
#define init_1_6_2_14
void FASTCALL call_1_6_2_14(SH2_struct* sh)
{
  SH2movls4(sh, 6,2, 14);
}
#define init_1_6_2_15
void FASTCALL call_1_6_2_15(SH2_struct* sh)
{
  SH2movls4(sh, 6,2, 15);
}
#define init_1_6_3_0
void FASTCALL call_1_6_3_0(SH2_struct* sh)
{
  SH2movls4(sh, 6,3, 0);
}
#define init_1_6_3_1
void FASTCALL call_1_6_3_1(SH2_struct* sh)
{
  SH2movls4(sh, 6,3, 1);
}
#define init_1_6_3_2
void FASTCALL call_1_6_3_2(SH2_struct* sh)
{
  SH2movls4(sh, 6,3, 2);
}
#define init_1_6_3_3
void FASTCALL call_1_6_3_3(SH2_struct* sh)
{
  SH2movls4(sh, 6,3, 3);
}
#define init_1_6_3_4
void FASTCALL call_1_6_3_4(SH2_struct* sh)
{
  SH2movls4(sh, 6,3, 4);
}
#define init_1_6_3_5
void FASTCALL call_1_6_3_5(SH2_struct* sh)
{
  SH2movls4(sh, 6,3, 5);
}
#define init_1_6_3_6
void FASTCALL call_1_6_3_6(SH2_struct* sh)
{
  SH2movls4(sh, 6,3, 6);
}
#define init_1_6_3_7
void FASTCALL call_1_6_3_7(SH2_struct* sh)
{
  SH2movls4(sh, 6,3, 7);
}
#define init_1_6_3_8
void FASTCALL call_1_6_3_8(SH2_struct* sh)
{
  SH2movls4(sh, 6,3, 8);
}
#define init_1_6_3_9
void FASTCALL call_1_6_3_9(SH2_struct* sh)
{
  SH2movls4(sh, 6,3, 9);
}
#define init_1_6_3_10
void FASTCALL call_1_6_3_10(SH2_struct* sh)
{
  SH2movls4(sh, 6,3, 10);
}
#define init_1_6_3_11
void FASTCALL call_1_6_3_11(SH2_struct* sh)
{
  SH2movls4(sh, 6,3, 11);
}
#define init_1_6_3_12
void FASTCALL call_1_6_3_12(SH2_struct* sh)
{
  SH2movls4(sh, 6,3, 12);
}
#define init_1_6_3_13
void FASTCALL call_1_6_3_13(SH2_struct* sh)
{
  SH2movls4(sh, 6,3, 13);
}
#define init_1_6_3_14
void FASTCALL call_1_6_3_14(SH2_struct* sh)
{
  SH2movls4(sh, 6,3, 14);
}
#define init_1_6_3_15
void FASTCALL call_1_6_3_15(SH2_struct* sh)
{
  SH2movls4(sh, 6,3, 15);
}
#define init_1_6_4_0
void FASTCALL call_1_6_4_0(SH2_struct* sh)
{
  SH2movls4(sh, 6,4, 0);
}
#define init_1_6_4_1
void FASTCALL call_1_6_4_1(SH2_struct* sh)
{
  SH2movls4(sh, 6,4, 1);
}
#define init_1_6_4_2
void FASTCALL call_1_6_4_2(SH2_struct* sh)
{
  SH2movls4(sh, 6,4, 2);
}
#define init_1_6_4_3
void FASTCALL call_1_6_4_3(SH2_struct* sh)
{
  SH2movls4(sh, 6,4, 3);
}
#define init_1_6_4_4
void FASTCALL call_1_6_4_4(SH2_struct* sh)
{
  SH2movls4(sh, 6,4, 4);
}
#define init_1_6_4_5
void FASTCALL call_1_6_4_5(SH2_struct* sh)
{
  SH2movls4(sh, 6,4, 5);
}
#define init_1_6_4_6
void FASTCALL call_1_6_4_6(SH2_struct* sh)
{
  SH2movls4(sh, 6,4, 6);
}
#define init_1_6_4_7
void FASTCALL call_1_6_4_7(SH2_struct* sh)
{
  SH2movls4(sh, 6,4, 7);
}
#define init_1_6_4_8
void FASTCALL call_1_6_4_8(SH2_struct* sh)
{
  SH2movls4(sh, 6,4, 8);
}
#define init_1_6_4_9
void FASTCALL call_1_6_4_9(SH2_struct* sh)
{
  SH2movls4(sh, 6,4, 9);
}
#define init_1_6_4_10
void FASTCALL call_1_6_4_10(SH2_struct* sh)
{
  SH2movls4(sh, 6,4, 10);
}
#define init_1_6_4_11
void FASTCALL call_1_6_4_11(SH2_struct* sh)
{
  SH2movls4(sh, 6,4, 11);
}
#define init_1_6_4_12
void FASTCALL call_1_6_4_12(SH2_struct* sh)
{
  SH2movls4(sh, 6,4, 12);
}
#define init_1_6_4_13
void FASTCALL call_1_6_4_13(SH2_struct* sh)
{
  SH2movls4(sh, 6,4, 13);
}
#define init_1_6_4_14
void FASTCALL call_1_6_4_14(SH2_struct* sh)
{
  SH2movls4(sh, 6,4, 14);
}
#define init_1_6_4_15
void FASTCALL call_1_6_4_15(SH2_struct* sh)
{
  SH2movls4(sh, 6,4, 15);
}
#define init_1_6_5_0
void FASTCALL call_1_6_5_0(SH2_struct* sh)
{
  SH2movls4(sh, 6,5, 0);
}
#define init_1_6_5_1
void FASTCALL call_1_6_5_1(SH2_struct* sh)
{
  SH2movls4(sh, 6,5, 1);
}
#define init_1_6_5_2
void FASTCALL call_1_6_5_2(SH2_struct* sh)
{
  SH2movls4(sh, 6,5, 2);
}
#define init_1_6_5_3
void FASTCALL call_1_6_5_3(SH2_struct* sh)
{
  SH2movls4(sh, 6,5, 3);
}
#define init_1_6_5_4
void FASTCALL call_1_6_5_4(SH2_struct* sh)
{
  SH2movls4(sh, 6,5, 4);
}
#define init_1_6_5_5
void FASTCALL call_1_6_5_5(SH2_struct* sh)
{
  SH2movls4(sh, 6,5, 5);
}
#define init_1_6_5_6
void FASTCALL call_1_6_5_6(SH2_struct* sh)
{
  SH2movls4(sh, 6,5, 6);
}
#define init_1_6_5_7
void FASTCALL call_1_6_5_7(SH2_struct* sh)
{
  SH2movls4(sh, 6,5, 7);
}
#define init_1_6_5_8
void FASTCALL call_1_6_5_8(SH2_struct* sh)
{
  SH2movls4(sh, 6,5, 8);
}
#define init_1_6_5_9
void FASTCALL call_1_6_5_9(SH2_struct* sh)
{
  SH2movls4(sh, 6,5, 9);
}
#define init_1_6_5_10
void FASTCALL call_1_6_5_10(SH2_struct* sh)
{
  SH2movls4(sh, 6,5, 10);
}
#define init_1_6_5_11
void FASTCALL call_1_6_5_11(SH2_struct* sh)
{
  SH2movls4(sh, 6,5, 11);
}
#define init_1_6_5_12
void FASTCALL call_1_6_5_12(SH2_struct* sh)
{
  SH2movls4(sh, 6,5, 12);
}
#define init_1_6_5_13
void FASTCALL call_1_6_5_13(SH2_struct* sh)
{
  SH2movls4(sh, 6,5, 13);
}
#define init_1_6_5_14
void FASTCALL call_1_6_5_14(SH2_struct* sh)
{
  SH2movls4(sh, 6,5, 14);
}
#define init_1_6_5_15
void FASTCALL call_1_6_5_15(SH2_struct* sh)
{
  SH2movls4(sh, 6,5, 15);
}
#define init_1_6_6_0
void FASTCALL call_1_6_6_0(SH2_struct* sh)
{
  SH2movls4(sh, 6,6, 0);
}
#define init_1_6_6_1
void FASTCALL call_1_6_6_1(SH2_struct* sh)
{
  SH2movls4(sh, 6,6, 1);
}
#define init_1_6_6_2
void FASTCALL call_1_6_6_2(SH2_struct* sh)
{
  SH2movls4(sh, 6,6, 2);
}
#define init_1_6_6_3
void FASTCALL call_1_6_6_3(SH2_struct* sh)
{
  SH2movls4(sh, 6,6, 3);
}
#define init_1_6_6_4
void FASTCALL call_1_6_6_4(SH2_struct* sh)
{
  SH2movls4(sh, 6,6, 4);
}
#define init_1_6_6_5
void FASTCALL call_1_6_6_5(SH2_struct* sh)
{
  SH2movls4(sh, 6,6, 5);
}
#define init_1_6_6_6
void FASTCALL call_1_6_6_6(SH2_struct* sh)
{
  SH2movls4(sh, 6,6, 6);
}
#define init_1_6_6_7
void FASTCALL call_1_6_6_7(SH2_struct* sh)
{
  SH2movls4(sh, 6,6, 7);
}
#define init_1_6_6_8
void FASTCALL call_1_6_6_8(SH2_struct* sh)
{
  SH2movls4(sh, 6,6, 8);
}
#define init_1_6_6_9
void FASTCALL call_1_6_6_9(SH2_struct* sh)
{
  SH2movls4(sh, 6,6, 9);
}
#define init_1_6_6_10
void FASTCALL call_1_6_6_10(SH2_struct* sh)
{
  SH2movls4(sh, 6,6, 10);
}
#define init_1_6_6_11
void FASTCALL call_1_6_6_11(SH2_struct* sh)
{
  SH2movls4(sh, 6,6, 11);
}
#define init_1_6_6_12
void FASTCALL call_1_6_6_12(SH2_struct* sh)
{
  SH2movls4(sh, 6,6, 12);
}
#define init_1_6_6_13
void FASTCALL call_1_6_6_13(SH2_struct* sh)
{
  SH2movls4(sh, 6,6, 13);
}
#define init_1_6_6_14
void FASTCALL call_1_6_6_14(SH2_struct* sh)
{
  SH2movls4(sh, 6,6, 14);
}
#define init_1_6_6_15
void FASTCALL call_1_6_6_15(SH2_struct* sh)
{
  SH2movls4(sh, 6,6, 15);
}
#define init_1_6_7_0
void FASTCALL call_1_6_7_0(SH2_struct* sh)
{
  SH2movls4(sh, 6,7, 0);
}
#define init_1_6_7_1
void FASTCALL call_1_6_7_1(SH2_struct* sh)
{
  SH2movls4(sh, 6,7, 1);
}
#define init_1_6_7_2
void FASTCALL call_1_6_7_2(SH2_struct* sh)
{
  SH2movls4(sh, 6,7, 2);
}
#define init_1_6_7_3
void FASTCALL call_1_6_7_3(SH2_struct* sh)
{
  SH2movls4(sh, 6,7, 3);
}
#define init_1_6_7_4
void FASTCALL call_1_6_7_4(SH2_struct* sh)
{
  SH2movls4(sh, 6,7, 4);
}
#define init_1_6_7_5
void FASTCALL call_1_6_7_5(SH2_struct* sh)
{
  SH2movls4(sh, 6,7, 5);
}
#define init_1_6_7_6
void FASTCALL call_1_6_7_6(SH2_struct* sh)
{
  SH2movls4(sh, 6,7, 6);
}
#define init_1_6_7_7
void FASTCALL call_1_6_7_7(SH2_struct* sh)
{
  SH2movls4(sh, 6,7, 7);
}
#define init_1_6_7_8
void FASTCALL call_1_6_7_8(SH2_struct* sh)
{
  SH2movls4(sh, 6,7, 8);
}
#define init_1_6_7_9
void FASTCALL call_1_6_7_9(SH2_struct* sh)
{
  SH2movls4(sh, 6,7, 9);
}
#define init_1_6_7_10
void FASTCALL call_1_6_7_10(SH2_struct* sh)
{
  SH2movls4(sh, 6,7, 10);
}
#define init_1_6_7_11
void FASTCALL call_1_6_7_11(SH2_struct* sh)
{
  SH2movls4(sh, 6,7, 11);
}
#define init_1_6_7_12
void FASTCALL call_1_6_7_12(SH2_struct* sh)
{
  SH2movls4(sh, 6,7, 12);
}
#define init_1_6_7_13
void FASTCALL call_1_6_7_13(SH2_struct* sh)
{
  SH2movls4(sh, 6,7, 13);
}
#define init_1_6_7_14
void FASTCALL call_1_6_7_14(SH2_struct* sh)
{
  SH2movls4(sh, 6,7, 14);
}
#define init_1_6_7_15
void FASTCALL call_1_6_7_15(SH2_struct* sh)
{
  SH2movls4(sh, 6,7, 15);
}
#define init_1_6_8_0
void FASTCALL call_1_6_8_0(SH2_struct* sh)
{
  SH2movls4(sh, 6,8, 0);
}
#define init_1_6_8_1
void FASTCALL call_1_6_8_1(SH2_struct* sh)
{
  SH2movls4(sh, 6,8, 1);
}
#define init_1_6_8_2
void FASTCALL call_1_6_8_2(SH2_struct* sh)
{
  SH2movls4(sh, 6,8, 2);
}
#define init_1_6_8_3
void FASTCALL call_1_6_8_3(SH2_struct* sh)
{
  SH2movls4(sh, 6,8, 3);
}
#define init_1_6_8_4
void FASTCALL call_1_6_8_4(SH2_struct* sh)
{
  SH2movls4(sh, 6,8, 4);
}
#define init_1_6_8_5
void FASTCALL call_1_6_8_5(SH2_struct* sh)
{
  SH2movls4(sh, 6,8, 5);
}
#define init_1_6_8_6
void FASTCALL call_1_6_8_6(SH2_struct* sh)
{
  SH2movls4(sh, 6,8, 6);
}
#define init_1_6_8_7
void FASTCALL call_1_6_8_7(SH2_struct* sh)
{
  SH2movls4(sh, 6,8, 7);
}
#define init_1_6_8_8
void FASTCALL call_1_6_8_8(SH2_struct* sh)
{
  SH2movls4(sh, 6,8, 8);
}
#define init_1_6_8_9
void FASTCALL call_1_6_8_9(SH2_struct* sh)
{
  SH2movls4(sh, 6,8, 9);
}
#define init_1_6_8_10
void FASTCALL call_1_6_8_10(SH2_struct* sh)
{
  SH2movls4(sh, 6,8, 10);
}
#define init_1_6_8_11
void FASTCALL call_1_6_8_11(SH2_struct* sh)
{
  SH2movls4(sh, 6,8, 11);
}
#define init_1_6_8_12
void FASTCALL call_1_6_8_12(SH2_struct* sh)
{
  SH2movls4(sh, 6,8, 12);
}
#define init_1_6_8_13
void FASTCALL call_1_6_8_13(SH2_struct* sh)
{
  SH2movls4(sh, 6,8, 13);
}
#define init_1_6_8_14
void FASTCALL call_1_6_8_14(SH2_struct* sh)
{
  SH2movls4(sh, 6,8, 14);
}
#define init_1_6_8_15
void FASTCALL call_1_6_8_15(SH2_struct* sh)
{
  SH2movls4(sh, 6,8, 15);
}
#define init_1_6_9_0
void FASTCALL call_1_6_9_0(SH2_struct* sh)
{
  SH2movls4(sh, 6,9, 0);
}
#define init_1_6_9_1
void FASTCALL call_1_6_9_1(SH2_struct* sh)
{
  SH2movls4(sh, 6,9, 1);
}
#define init_1_6_9_2
void FASTCALL call_1_6_9_2(SH2_struct* sh)
{
  SH2movls4(sh, 6,9, 2);
}
#define init_1_6_9_3
void FASTCALL call_1_6_9_3(SH2_struct* sh)
{
  SH2movls4(sh, 6,9, 3);
}
#define init_1_6_9_4
void FASTCALL call_1_6_9_4(SH2_struct* sh)
{
  SH2movls4(sh, 6,9, 4);
}
#define init_1_6_9_5
void FASTCALL call_1_6_9_5(SH2_struct* sh)
{
  SH2movls4(sh, 6,9, 5);
}
#define init_1_6_9_6
void FASTCALL call_1_6_9_6(SH2_struct* sh)
{
  SH2movls4(sh, 6,9, 6);
}
#define init_1_6_9_7
void FASTCALL call_1_6_9_7(SH2_struct* sh)
{
  SH2movls4(sh, 6,9, 7);
}
#define init_1_6_9_8
void FASTCALL call_1_6_9_8(SH2_struct* sh)
{
  SH2movls4(sh, 6,9, 8);
}
#define init_1_6_9_9
void FASTCALL call_1_6_9_9(SH2_struct* sh)
{
  SH2movls4(sh, 6,9, 9);
}
#define init_1_6_9_10
void FASTCALL call_1_6_9_10(SH2_struct* sh)
{
  SH2movls4(sh, 6,9, 10);
}
#define init_1_6_9_11
void FASTCALL call_1_6_9_11(SH2_struct* sh)
{
  SH2movls4(sh, 6,9, 11);
}
#define init_1_6_9_12
void FASTCALL call_1_6_9_12(SH2_struct* sh)
{
  SH2movls4(sh, 6,9, 12);
}
#define init_1_6_9_13
void FASTCALL call_1_6_9_13(SH2_struct* sh)
{
  SH2movls4(sh, 6,9, 13);
}
#define init_1_6_9_14
void FASTCALL call_1_6_9_14(SH2_struct* sh)
{
  SH2movls4(sh, 6,9, 14);
}
#define init_1_6_9_15
void FASTCALL call_1_6_9_15(SH2_struct* sh)
{
  SH2movls4(sh, 6,9, 15);
}
#define init_1_6_10_0
void FASTCALL call_1_6_10_0(SH2_struct* sh)
{
  SH2movls4(sh, 6,10, 0);
}
#define init_1_6_10_1
void FASTCALL call_1_6_10_1(SH2_struct* sh)
{
  SH2movls4(sh, 6,10, 1);
}
#define init_1_6_10_2
void FASTCALL call_1_6_10_2(SH2_struct* sh)
{
  SH2movls4(sh, 6,10, 2);
}
#define init_1_6_10_3
void FASTCALL call_1_6_10_3(SH2_struct* sh)
{
  SH2movls4(sh, 6,10, 3);
}
#define init_1_6_10_4
void FASTCALL call_1_6_10_4(SH2_struct* sh)
{
  SH2movls4(sh, 6,10, 4);
}
#define init_1_6_10_5
void FASTCALL call_1_6_10_5(SH2_struct* sh)
{
  SH2movls4(sh, 6,10, 5);
}
#define init_1_6_10_6
void FASTCALL call_1_6_10_6(SH2_struct* sh)
{
  SH2movls4(sh, 6,10, 6);
}
#define init_1_6_10_7
void FASTCALL call_1_6_10_7(SH2_struct* sh)
{
  SH2movls4(sh, 6,10, 7);
}
#define init_1_6_10_8
void FASTCALL call_1_6_10_8(SH2_struct* sh)
{
  SH2movls4(sh, 6,10, 8);
}
#define init_1_6_10_9
void FASTCALL call_1_6_10_9(SH2_struct* sh)
{
  SH2movls4(sh, 6,10, 9);
}
#define init_1_6_10_10
void FASTCALL call_1_6_10_10(SH2_struct* sh)
{
  SH2movls4(sh, 6,10, 10);
}
#define init_1_6_10_11
void FASTCALL call_1_6_10_11(SH2_struct* sh)
{
  SH2movls4(sh, 6,10, 11);
}
#define init_1_6_10_12
void FASTCALL call_1_6_10_12(SH2_struct* sh)
{
  SH2movls4(sh, 6,10, 12);
}
#define init_1_6_10_13
void FASTCALL call_1_6_10_13(SH2_struct* sh)
{
  SH2movls4(sh, 6,10, 13);
}
#define init_1_6_10_14
void FASTCALL call_1_6_10_14(SH2_struct* sh)
{
  SH2movls4(sh, 6,10, 14);
}
#define init_1_6_10_15
void FASTCALL call_1_6_10_15(SH2_struct* sh)
{
  SH2movls4(sh, 6,10, 15);
}
#define init_1_6_11_0
void FASTCALL call_1_6_11_0(SH2_struct* sh)
{
  SH2movls4(sh, 6,11, 0);
}
#define init_1_6_11_1
void FASTCALL call_1_6_11_1(SH2_struct* sh)
{
  SH2movls4(sh, 6,11, 1);
}
#define init_1_6_11_2
void FASTCALL call_1_6_11_2(SH2_struct* sh)
{
  SH2movls4(sh, 6,11, 2);
}
#define init_1_6_11_3
void FASTCALL call_1_6_11_3(SH2_struct* sh)
{
  SH2movls4(sh, 6,11, 3);
}
#define init_1_6_11_4
void FASTCALL call_1_6_11_4(SH2_struct* sh)
{
  SH2movls4(sh, 6,11, 4);
}
#define init_1_6_11_5
void FASTCALL call_1_6_11_5(SH2_struct* sh)
{
  SH2movls4(sh, 6,11, 5);
}
#define init_1_6_11_6
void FASTCALL call_1_6_11_6(SH2_struct* sh)
{
  SH2movls4(sh, 6,11, 6);
}
#define init_1_6_11_7
void FASTCALL call_1_6_11_7(SH2_struct* sh)
{
  SH2movls4(sh, 6,11, 7);
}
#define init_1_6_11_8
void FASTCALL call_1_6_11_8(SH2_struct* sh)
{
  SH2movls4(sh, 6,11, 8);
}
#define init_1_6_11_9
void FASTCALL call_1_6_11_9(SH2_struct* sh)
{
  SH2movls4(sh, 6,11, 9);
}
#define init_1_6_11_10
void FASTCALL call_1_6_11_10(SH2_struct* sh)
{
  SH2movls4(sh, 6,11, 10);
}
#define init_1_6_11_11
void FASTCALL call_1_6_11_11(SH2_struct* sh)
{
  SH2movls4(sh, 6,11, 11);
}
#define init_1_6_11_12
void FASTCALL call_1_6_11_12(SH2_struct* sh)
{
  SH2movls4(sh, 6,11, 12);
}
#define init_1_6_11_13
void FASTCALL call_1_6_11_13(SH2_struct* sh)
{
  SH2movls4(sh, 6,11, 13);
}
#define init_1_6_11_14
void FASTCALL call_1_6_11_14(SH2_struct* sh)
{
  SH2movls4(sh, 6,11, 14);
}
#define init_1_6_11_15
void FASTCALL call_1_6_11_15(SH2_struct* sh)
{
  SH2movls4(sh, 6,11, 15);
}
#define init_1_6_12_0
void FASTCALL call_1_6_12_0(SH2_struct* sh)
{
  SH2movls4(sh, 6,12, 0);
}
#define init_1_6_12_1
void FASTCALL call_1_6_12_1(SH2_struct* sh)
{
  SH2movls4(sh, 6,12, 1);
}
#define init_1_6_12_2
void FASTCALL call_1_6_12_2(SH2_struct* sh)
{
  SH2movls4(sh, 6,12, 2);
}
#define init_1_6_12_3
void FASTCALL call_1_6_12_3(SH2_struct* sh)
{
  SH2movls4(sh, 6,12, 3);
}
#define init_1_6_12_4
void FASTCALL call_1_6_12_4(SH2_struct* sh)
{
  SH2movls4(sh, 6,12, 4);
}
#define init_1_6_12_5
void FASTCALL call_1_6_12_5(SH2_struct* sh)
{
  SH2movls4(sh, 6,12, 5);
}
#define init_1_6_12_6
void FASTCALL call_1_6_12_6(SH2_struct* sh)
{
  SH2movls4(sh, 6,12, 6);
}
#define init_1_6_12_7
void FASTCALL call_1_6_12_7(SH2_struct* sh)
{
  SH2movls4(sh, 6,12, 7);
}
#define init_1_6_12_8
void FASTCALL call_1_6_12_8(SH2_struct* sh)
{
  SH2movls4(sh, 6,12, 8);
}
#define init_1_6_12_9
void FASTCALL call_1_6_12_9(SH2_struct* sh)
{
  SH2movls4(sh, 6,12, 9);
}
#define init_1_6_12_10
void FASTCALL call_1_6_12_10(SH2_struct* sh)
{
  SH2movls4(sh, 6,12, 10);
}
#define init_1_6_12_11
void FASTCALL call_1_6_12_11(SH2_struct* sh)
{
  SH2movls4(sh, 6,12, 11);
}
#define init_1_6_12_12
void FASTCALL call_1_6_12_12(SH2_struct* sh)
{
  SH2movls4(sh, 6,12, 12);
}
#define init_1_6_12_13
void FASTCALL call_1_6_12_13(SH2_struct* sh)
{
  SH2movls4(sh, 6,12, 13);
}
#define init_1_6_12_14
void FASTCALL call_1_6_12_14(SH2_struct* sh)
{
  SH2movls4(sh, 6,12, 14);
}
#define init_1_6_12_15
void FASTCALL call_1_6_12_15(SH2_struct* sh)
{
  SH2movls4(sh, 6,12, 15);
}
#define init_1_6_13_0
void FASTCALL call_1_6_13_0(SH2_struct* sh)
{
  SH2movls4(sh, 6,13, 0);
}
#define init_1_6_13_1
void FASTCALL call_1_6_13_1(SH2_struct* sh)
{
  SH2movls4(sh, 6,13, 1);
}
#define init_1_6_13_2
void FASTCALL call_1_6_13_2(SH2_struct* sh)
{
  SH2movls4(sh, 6,13, 2);
}
#define init_1_6_13_3
void FASTCALL call_1_6_13_3(SH2_struct* sh)
{
  SH2movls4(sh, 6,13, 3);
}
#define init_1_6_13_4
void FASTCALL call_1_6_13_4(SH2_struct* sh)
{
  SH2movls4(sh, 6,13, 4);
}
#define init_1_6_13_5
void FASTCALL call_1_6_13_5(SH2_struct* sh)
{
  SH2movls4(sh, 6,13, 5);
}
#define init_1_6_13_6
void FASTCALL call_1_6_13_6(SH2_struct* sh)
{
  SH2movls4(sh, 6,13, 6);
}
#define init_1_6_13_7
void FASTCALL call_1_6_13_7(SH2_struct* sh)
{
  SH2movls4(sh, 6,13, 7);
}
#define init_1_6_13_8
void FASTCALL call_1_6_13_8(SH2_struct* sh)
{
  SH2movls4(sh, 6,13, 8);
}
#define init_1_6_13_9
void FASTCALL call_1_6_13_9(SH2_struct* sh)
{
  SH2movls4(sh, 6,13, 9);
}
#define init_1_6_13_10
void FASTCALL call_1_6_13_10(SH2_struct* sh)
{
  SH2movls4(sh, 6,13, 10);
}
#define init_1_6_13_11
void FASTCALL call_1_6_13_11(SH2_struct* sh)
{
  SH2movls4(sh, 6,13, 11);
}
#define init_1_6_13_12
void FASTCALL call_1_6_13_12(SH2_struct* sh)
{
  SH2movls4(sh, 6,13, 12);
}
#define init_1_6_13_13
void FASTCALL call_1_6_13_13(SH2_struct* sh)
{
  SH2movls4(sh, 6,13, 13);
}
#define init_1_6_13_14
void FASTCALL call_1_6_13_14(SH2_struct* sh)
{
  SH2movls4(sh, 6,13, 14);
}
#define init_1_6_13_15
void FASTCALL call_1_6_13_15(SH2_struct* sh)
{
  SH2movls4(sh, 6,13, 15);
}
#define init_1_6_14_0
void FASTCALL call_1_6_14_0(SH2_struct* sh)
{
  SH2movls4(sh, 6,14, 0);
}
#define init_1_6_14_1
void FASTCALL call_1_6_14_1(SH2_struct* sh)
{
  SH2movls4(sh, 6,14, 1);
}
#define init_1_6_14_2
void FASTCALL call_1_6_14_2(SH2_struct* sh)
{
  SH2movls4(sh, 6,14, 2);
}
#define init_1_6_14_3
void FASTCALL call_1_6_14_3(SH2_struct* sh)
{
  SH2movls4(sh, 6,14, 3);
}
#define init_1_6_14_4
void FASTCALL call_1_6_14_4(SH2_struct* sh)
{
  SH2movls4(sh, 6,14, 4);
}
#define init_1_6_14_5
void FASTCALL call_1_6_14_5(SH2_struct* sh)
{
  SH2movls4(sh, 6,14, 5);
}
#define init_1_6_14_6
void FASTCALL call_1_6_14_6(SH2_struct* sh)
{
  SH2movls4(sh, 6,14, 6);
}
#define init_1_6_14_7
void FASTCALL call_1_6_14_7(SH2_struct* sh)
{
  SH2movls4(sh, 6,14, 7);
}
#define init_1_6_14_8
void FASTCALL call_1_6_14_8(SH2_struct* sh)
{
  SH2movls4(sh, 6,14, 8);
}
#define init_1_6_14_9
void FASTCALL call_1_6_14_9(SH2_struct* sh)
{
  SH2movls4(sh, 6,14, 9);
}
#define init_1_6_14_10
void FASTCALL call_1_6_14_10(SH2_struct* sh)
{
  SH2movls4(sh, 6,14, 10);
}
#define init_1_6_14_11
void FASTCALL call_1_6_14_11(SH2_struct* sh)
{
  SH2movls4(sh, 6,14, 11);
}
#define init_1_6_14_12
void FASTCALL call_1_6_14_12(SH2_struct* sh)
{
  SH2movls4(sh, 6,14, 12);
}
#define init_1_6_14_13
void FASTCALL call_1_6_14_13(SH2_struct* sh)
{
  SH2movls4(sh, 6,14, 13);
}
#define init_1_6_14_14
void FASTCALL call_1_6_14_14(SH2_struct* sh)
{
  SH2movls4(sh, 6,14, 14);
}
#define init_1_6_14_15
void FASTCALL call_1_6_14_15(SH2_struct* sh)
{
  SH2movls4(sh, 6,14, 15);
}
#define init_1_6_15_0
void FASTCALL call_1_6_15_0(SH2_struct* sh)
{
  SH2movls4(sh, 6,15, 0);
}
#define init_1_6_15_1
void FASTCALL call_1_6_15_1(SH2_struct* sh)
{
  SH2movls4(sh, 6,15, 1);
}
#define init_1_6_15_2
void FASTCALL call_1_6_15_2(SH2_struct* sh)
{
  SH2movls4(sh, 6,15, 2);
}
#define init_1_6_15_3
void FASTCALL call_1_6_15_3(SH2_struct* sh)
{
  SH2movls4(sh, 6,15, 3);
}
#define init_1_6_15_4
void FASTCALL call_1_6_15_4(SH2_struct* sh)
{
  SH2movls4(sh, 6,15, 4);
}
#define init_1_6_15_5
void FASTCALL call_1_6_15_5(SH2_struct* sh)
{
  SH2movls4(sh, 6,15, 5);
}
#define init_1_6_15_6
void FASTCALL call_1_6_15_6(SH2_struct* sh)
{
  SH2movls4(sh, 6,15, 6);
}
#define init_1_6_15_7
void FASTCALL call_1_6_15_7(SH2_struct* sh)
{
  SH2movls4(sh, 6,15, 7);
}
#define init_1_6_15_8
void FASTCALL call_1_6_15_8(SH2_struct* sh)
{
  SH2movls4(sh, 6,15, 8);
}
#define init_1_6_15_9
void FASTCALL call_1_6_15_9(SH2_struct* sh)
{
  SH2movls4(sh, 6,15, 9);
}
#define init_1_6_15_10
void FASTCALL call_1_6_15_10(SH2_struct* sh)
{
  SH2movls4(sh, 6,15, 10);
}
#define init_1_6_15_11
void FASTCALL call_1_6_15_11(SH2_struct* sh)
{
  SH2movls4(sh, 6,15, 11);
}
#define init_1_6_15_12
void FASTCALL call_1_6_15_12(SH2_struct* sh)
{
  SH2movls4(sh, 6,15, 12);
}
#define init_1_6_15_13
void FASTCALL call_1_6_15_13(SH2_struct* sh)
{
  SH2movls4(sh, 6,15, 13);
}
#define init_1_6_15_14
void FASTCALL call_1_6_15_14(SH2_struct* sh)
{
  SH2movls4(sh, 6,15, 14);
}
#define init_1_6_15_15
void FASTCALL call_1_6_15_15(SH2_struct* sh)
{
  SH2movls4(sh, 6,15, 15);
}
#define init_1_7_0_0
void FASTCALL call_1_7_0_0(SH2_struct* sh)
{
  SH2movls4(sh, 7,0, 0);
}
#define init_1_7_0_1
void FASTCALL call_1_7_0_1(SH2_struct* sh)
{
  SH2movls4(sh, 7,0, 1);
}
#define init_1_7_0_2
void FASTCALL call_1_7_0_2(SH2_struct* sh)
{
  SH2movls4(sh, 7,0, 2);
}
#define init_1_7_0_3
void FASTCALL call_1_7_0_3(SH2_struct* sh)
{
  SH2movls4(sh, 7,0, 3);
}
#define init_1_7_0_4
void FASTCALL call_1_7_0_4(SH2_struct* sh)
{
  SH2movls4(sh, 7,0, 4);
}
#define init_1_7_0_5
void FASTCALL call_1_7_0_5(SH2_struct* sh)
{
  SH2movls4(sh, 7,0, 5);
}
#define init_1_7_0_6
void FASTCALL call_1_7_0_6(SH2_struct* sh)
{
  SH2movls4(sh, 7,0, 6);
}
#define init_1_7_0_7
void FASTCALL call_1_7_0_7(SH2_struct* sh)
{
  SH2movls4(sh, 7,0, 7);
}
#define init_1_7_0_8
void FASTCALL call_1_7_0_8(SH2_struct* sh)
{
  SH2movls4(sh, 7,0, 8);
}
#define init_1_7_0_9
void FASTCALL call_1_7_0_9(SH2_struct* sh)
{
  SH2movls4(sh, 7,0, 9);
}
#define init_1_7_0_10
void FASTCALL call_1_7_0_10(SH2_struct* sh)
{
  SH2movls4(sh, 7,0, 10);
}
#define init_1_7_0_11
void FASTCALL call_1_7_0_11(SH2_struct* sh)
{
  SH2movls4(sh, 7,0, 11);
}
#define init_1_7_0_12
void FASTCALL call_1_7_0_12(SH2_struct* sh)
{
  SH2movls4(sh, 7,0, 12);
}
#define init_1_7_0_13
void FASTCALL call_1_7_0_13(SH2_struct* sh)
{
  SH2movls4(sh, 7,0, 13);
}
#define init_1_7_0_14
void FASTCALL call_1_7_0_14(SH2_struct* sh)
{
  SH2movls4(sh, 7,0, 14);
}
#define init_1_7_0_15
void FASTCALL call_1_7_0_15(SH2_struct* sh)
{
  SH2movls4(sh, 7,0, 15);
}
#define init_1_7_1_0
void FASTCALL call_1_7_1_0(SH2_struct* sh)
{
  SH2movls4(sh, 7,1, 0);
}
#define init_1_7_1_1
void FASTCALL call_1_7_1_1(SH2_struct* sh)
{
  SH2movls4(sh, 7,1, 1);
}
#define init_1_7_1_2
void FASTCALL call_1_7_1_2(SH2_struct* sh)
{
  SH2movls4(sh, 7,1, 2);
}
#define init_1_7_1_3
void FASTCALL call_1_7_1_3(SH2_struct* sh)
{
  SH2movls4(sh, 7,1, 3);
}
#define init_1_7_1_4
void FASTCALL call_1_7_1_4(SH2_struct* sh)
{
  SH2movls4(sh, 7,1, 4);
}
#define init_1_7_1_5
void FASTCALL call_1_7_1_5(SH2_struct* sh)
{
  SH2movls4(sh, 7,1, 5);
}
#define init_1_7_1_6
void FASTCALL call_1_7_1_6(SH2_struct* sh)
{
  SH2movls4(sh, 7,1, 6);
}
#define init_1_7_1_7
void FASTCALL call_1_7_1_7(SH2_struct* sh)
{
  SH2movls4(sh, 7,1, 7);
}
#define init_1_7_1_8
void FASTCALL call_1_7_1_8(SH2_struct* sh)
{
  SH2movls4(sh, 7,1, 8);
}
#define init_1_7_1_9
void FASTCALL call_1_7_1_9(SH2_struct* sh)
{
  SH2movls4(sh, 7,1, 9);
}
#define init_1_7_1_10
void FASTCALL call_1_7_1_10(SH2_struct* sh)
{
  SH2movls4(sh, 7,1, 10);
}
#define init_1_7_1_11
void FASTCALL call_1_7_1_11(SH2_struct* sh)
{
  SH2movls4(sh, 7,1, 11);
}
#define init_1_7_1_12
void FASTCALL call_1_7_1_12(SH2_struct* sh)
{
  SH2movls4(sh, 7,1, 12);
}
#define init_1_7_1_13
void FASTCALL call_1_7_1_13(SH2_struct* sh)
{
  SH2movls4(sh, 7,1, 13);
}
#define init_1_7_1_14
void FASTCALL call_1_7_1_14(SH2_struct* sh)
{
  SH2movls4(sh, 7,1, 14);
}
#define init_1_7_1_15
void FASTCALL call_1_7_1_15(SH2_struct* sh)
{
  SH2movls4(sh, 7,1, 15);
}
#define init_1_7_2_0
void FASTCALL call_1_7_2_0(SH2_struct* sh)
{
  SH2movls4(sh, 7,2, 0);
}
#define init_1_7_2_1
void FASTCALL call_1_7_2_1(SH2_struct* sh)
{
  SH2movls4(sh, 7,2, 1);
}
#define init_1_7_2_2
void FASTCALL call_1_7_2_2(SH2_struct* sh)
{
  SH2movls4(sh, 7,2, 2);
}
#define init_1_7_2_3
void FASTCALL call_1_7_2_3(SH2_struct* sh)
{
  SH2movls4(sh, 7,2, 3);
}
#define init_1_7_2_4
void FASTCALL call_1_7_2_4(SH2_struct* sh)
{
  SH2movls4(sh, 7,2, 4);
}
#define init_1_7_2_5
void FASTCALL call_1_7_2_5(SH2_struct* sh)
{
  SH2movls4(sh, 7,2, 5);
}
#define init_1_7_2_6
void FASTCALL call_1_7_2_6(SH2_struct* sh)
{
  SH2movls4(sh, 7,2, 6);
}
#define init_1_7_2_7
void FASTCALL call_1_7_2_7(SH2_struct* sh)
{
  SH2movls4(sh, 7,2, 7);
}
#define init_1_7_2_8
void FASTCALL call_1_7_2_8(SH2_struct* sh)
{
  SH2movls4(sh, 7,2, 8);
}
#define init_1_7_2_9
void FASTCALL call_1_7_2_9(SH2_struct* sh)
{
  SH2movls4(sh, 7,2, 9);
}
#define init_1_7_2_10
void FASTCALL call_1_7_2_10(SH2_struct* sh)
{
  SH2movls4(sh, 7,2, 10);
}
#define init_1_7_2_11
void FASTCALL call_1_7_2_11(SH2_struct* sh)
{
  SH2movls4(sh, 7,2, 11);
}
#define init_1_7_2_12
void FASTCALL call_1_7_2_12(SH2_struct* sh)
{
  SH2movls4(sh, 7,2, 12);
}
#define init_1_7_2_13
void FASTCALL call_1_7_2_13(SH2_struct* sh)
{
  SH2movls4(sh, 7,2, 13);
}
#define init_1_7_2_14
void FASTCALL call_1_7_2_14(SH2_struct* sh)
{
  SH2movls4(sh, 7,2, 14);
}
#define init_1_7_2_15
void FASTCALL call_1_7_2_15(SH2_struct* sh)
{
  SH2movls4(sh, 7,2, 15);
}
#define init_1_7_3_0
void FASTCALL call_1_7_3_0(SH2_struct* sh)
{
  SH2movls4(sh, 7,3, 0);
}
#define init_1_7_3_1
void FASTCALL call_1_7_3_1(SH2_struct* sh)
{
  SH2movls4(sh, 7,3, 1);
}
#define init_1_7_3_2
void FASTCALL call_1_7_3_2(SH2_struct* sh)
{
  SH2movls4(sh, 7,3, 2);
}
#define init_1_7_3_3
void FASTCALL call_1_7_3_3(SH2_struct* sh)
{
  SH2movls4(sh, 7,3, 3);
}
#define init_1_7_3_4
void FASTCALL call_1_7_3_4(SH2_struct* sh)
{
  SH2movls4(sh, 7,3, 4);
}
#define init_1_7_3_5
void FASTCALL call_1_7_3_5(SH2_struct* sh)
{
  SH2movls4(sh, 7,3, 5);
}
#define init_1_7_3_6
void FASTCALL call_1_7_3_6(SH2_struct* sh)
{
  SH2movls4(sh, 7,3, 6);
}
#define init_1_7_3_7
void FASTCALL call_1_7_3_7(SH2_struct* sh)
{
  SH2movls4(sh, 7,3, 7);
}
#define init_1_7_3_8
void FASTCALL call_1_7_3_8(SH2_struct* sh)
{
  SH2movls4(sh, 7,3, 8);
}
#define init_1_7_3_9
void FASTCALL call_1_7_3_9(SH2_struct* sh)
{
  SH2movls4(sh, 7,3, 9);
}
#define init_1_7_3_10
void FASTCALL call_1_7_3_10(SH2_struct* sh)
{
  SH2movls4(sh, 7,3, 10);
}
#define init_1_7_3_11
void FASTCALL call_1_7_3_11(SH2_struct* sh)
{
  SH2movls4(sh, 7,3, 11);
}
#define init_1_7_3_12
void FASTCALL call_1_7_3_12(SH2_struct* sh)
{
  SH2movls4(sh, 7,3, 12);
}
#define init_1_7_3_13
void FASTCALL call_1_7_3_13(SH2_struct* sh)
{
  SH2movls4(sh, 7,3, 13);
}
#define init_1_7_3_14
void FASTCALL call_1_7_3_14(SH2_struct* sh)
{
  SH2movls4(sh, 7,3, 14);
}
#define init_1_7_3_15
void FASTCALL call_1_7_3_15(SH2_struct* sh)
{
  SH2movls4(sh, 7,3, 15);
}
#define init_1_7_4_0
void FASTCALL call_1_7_4_0(SH2_struct* sh)
{
  SH2movls4(sh, 7,4, 0);
}
#define init_1_7_4_1
void FASTCALL call_1_7_4_1(SH2_struct* sh)
{
  SH2movls4(sh, 7,4, 1);
}
#define init_1_7_4_2
void FASTCALL call_1_7_4_2(SH2_struct* sh)
{
  SH2movls4(sh, 7,4, 2);
}
#define init_1_7_4_3
void FASTCALL call_1_7_4_3(SH2_struct* sh)
{
  SH2movls4(sh, 7,4, 3);
}
#define init_1_7_4_4
void FASTCALL call_1_7_4_4(SH2_struct* sh)
{
  SH2movls4(sh, 7,4, 4);
}
#define init_1_7_4_5
void FASTCALL call_1_7_4_5(SH2_struct* sh)
{
  SH2movls4(sh, 7,4, 5);
}
#define init_1_7_4_6
void FASTCALL call_1_7_4_6(SH2_struct* sh)
{
  SH2movls4(sh, 7,4, 6);
}
#define init_1_7_4_7
void FASTCALL call_1_7_4_7(SH2_struct* sh)
{
  SH2movls4(sh, 7,4, 7);
}
#define init_1_7_4_8
void FASTCALL call_1_7_4_8(SH2_struct* sh)
{
  SH2movls4(sh, 7,4, 8);
}
#define init_1_7_4_9
void FASTCALL call_1_7_4_9(SH2_struct* sh)
{
  SH2movls4(sh, 7,4, 9);
}
#define init_1_7_4_10
void FASTCALL call_1_7_4_10(SH2_struct* sh)
{
  SH2movls4(sh, 7,4, 10);
}
#define init_1_7_4_11
void FASTCALL call_1_7_4_11(SH2_struct* sh)
{
  SH2movls4(sh, 7,4, 11);
}
#define init_1_7_4_12
void FASTCALL call_1_7_4_12(SH2_struct* sh)
{
  SH2movls4(sh, 7,4, 12);
}
#define init_1_7_4_13
void FASTCALL call_1_7_4_13(SH2_struct* sh)
{
  SH2movls4(sh, 7,4, 13);
}
#define init_1_7_4_14
void FASTCALL call_1_7_4_14(SH2_struct* sh)
{
  SH2movls4(sh, 7,4, 14);
}
#define init_1_7_4_15
void FASTCALL call_1_7_4_15(SH2_struct* sh)
{
  SH2movls4(sh, 7,4, 15);
}
#define init_1_7_5_0
void FASTCALL call_1_7_5_0(SH2_struct* sh)
{
  SH2movls4(sh, 7,5, 0);
}
#define init_1_7_5_1
void FASTCALL call_1_7_5_1(SH2_struct* sh)
{
  SH2movls4(sh, 7,5, 1);
}
#define init_1_7_5_2
void FASTCALL call_1_7_5_2(SH2_struct* sh)
{
  SH2movls4(sh, 7,5, 2);
}
#define init_1_7_5_3
void FASTCALL call_1_7_5_3(SH2_struct* sh)
{
  SH2movls4(sh, 7,5, 3);
}
#define init_1_7_5_4
void FASTCALL call_1_7_5_4(SH2_struct* sh)
{
  SH2movls4(sh, 7,5, 4);
}
#define init_1_7_5_5
void FASTCALL call_1_7_5_5(SH2_struct* sh)
{
  SH2movls4(sh, 7,5, 5);
}
#define init_1_7_5_6
void FASTCALL call_1_7_5_6(SH2_struct* sh)
{
  SH2movls4(sh, 7,5, 6);
}
#define init_1_7_5_7
void FASTCALL call_1_7_5_7(SH2_struct* sh)
{
  SH2movls4(sh, 7,5, 7);
}
#define init_1_7_5_8
void FASTCALL call_1_7_5_8(SH2_struct* sh)
{
  SH2movls4(sh, 7,5, 8);
}
#define init_1_7_5_9
void FASTCALL call_1_7_5_9(SH2_struct* sh)
{
  SH2movls4(sh, 7,5, 9);
}
#define init_1_7_5_10
void FASTCALL call_1_7_5_10(SH2_struct* sh)
{
  SH2movls4(sh, 7,5, 10);
}
#define init_1_7_5_11
void FASTCALL call_1_7_5_11(SH2_struct* sh)
{
  SH2movls4(sh, 7,5, 11);
}
#define init_1_7_5_12
void FASTCALL call_1_7_5_12(SH2_struct* sh)
{
  SH2movls4(sh, 7,5, 12);
}
#define init_1_7_5_13
void FASTCALL call_1_7_5_13(SH2_struct* sh)
{
  SH2movls4(sh, 7,5, 13);
}
#define init_1_7_5_14
void FASTCALL call_1_7_5_14(SH2_struct* sh)
{
  SH2movls4(sh, 7,5, 14);
}
#define init_1_7_5_15
void FASTCALL call_1_7_5_15(SH2_struct* sh)
{
  SH2movls4(sh, 7,5, 15);
}
#define init_1_7_6_0
void FASTCALL call_1_7_6_0(SH2_struct* sh)
{
  SH2movls4(sh, 7,6, 0);
}
#define init_1_7_6_1
void FASTCALL call_1_7_6_1(SH2_struct* sh)
{
  SH2movls4(sh, 7,6, 1);
}
#define init_1_7_6_2
void FASTCALL call_1_7_6_2(SH2_struct* sh)
{
  SH2movls4(sh, 7,6, 2);
}
#define init_1_7_6_3
void FASTCALL call_1_7_6_3(SH2_struct* sh)
{
  SH2movls4(sh, 7,6, 3);
}
#define init_1_7_6_4
void FASTCALL call_1_7_6_4(SH2_struct* sh)
{
  SH2movls4(sh, 7,6, 4);
}
#define init_1_7_6_5
void FASTCALL call_1_7_6_5(SH2_struct* sh)
{
  SH2movls4(sh, 7,6, 5);
}
#define init_1_7_6_6
void FASTCALL call_1_7_6_6(SH2_struct* sh)
{
  SH2movls4(sh, 7,6, 6);
}
#define init_1_7_6_7
void FASTCALL call_1_7_6_7(SH2_struct* sh)
{
  SH2movls4(sh, 7,6, 7);
}
#define init_1_7_6_8
void FASTCALL call_1_7_6_8(SH2_struct* sh)
{
  SH2movls4(sh, 7,6, 8);
}
#define init_1_7_6_9
void FASTCALL call_1_7_6_9(SH2_struct* sh)
{
  SH2movls4(sh, 7,6, 9);
}
#define init_1_7_6_10
void FASTCALL call_1_7_6_10(SH2_struct* sh)
{
  SH2movls4(sh, 7,6, 10);
}
#define init_1_7_6_11
void FASTCALL call_1_7_6_11(SH2_struct* sh)
{
  SH2movls4(sh, 7,6, 11);
}
#define init_1_7_6_12
void FASTCALL call_1_7_6_12(SH2_struct* sh)
{
  SH2movls4(sh, 7,6, 12);
}
#define init_1_7_6_13
void FASTCALL call_1_7_6_13(SH2_struct* sh)
{
  SH2movls4(sh, 7,6, 13);
}
#define init_1_7_6_14
void FASTCALL call_1_7_6_14(SH2_struct* sh)
{
  SH2movls4(sh, 7,6, 14);
}
#define init_1_7_6_15
void FASTCALL call_1_7_6_15(SH2_struct* sh)
{
  SH2movls4(sh, 7,6, 15);
}
#define init_1_7_7_0
void FASTCALL call_1_7_7_0(SH2_struct* sh)
{
  SH2movls4(sh, 7,7, 0);
}
#define init_1_7_7_1
void FASTCALL call_1_7_7_1(SH2_struct* sh)
{
  SH2movls4(sh, 7,7, 1);
}
#define init_1_7_7_2
void FASTCALL call_1_7_7_2(SH2_struct* sh)
{
  SH2movls4(sh, 7,7, 2);
}
#define init_1_7_7_3
void FASTCALL call_1_7_7_3(SH2_struct* sh)
{
  SH2movls4(sh, 7,7, 3);
}
#define init_1_7_7_4
void FASTCALL call_1_7_7_4(SH2_struct* sh)
{
  SH2movls4(sh, 7,7, 4);
}
#define init_1_7_7_5
void FASTCALL call_1_7_7_5(SH2_struct* sh)
{
  SH2movls4(sh, 7,7, 5);
}
#define init_1_7_7_6
void FASTCALL call_1_7_7_6(SH2_struct* sh)
{
  SH2movls4(sh, 7,7, 6);
}
#define init_1_7_7_7
void FASTCALL call_1_7_7_7(SH2_struct* sh)
{
  SH2movls4(sh, 7,7, 7);
}
#define init_1_7_7_8
void FASTCALL call_1_7_7_8(SH2_struct* sh)
{
  SH2movls4(sh, 7,7, 8);
}
#define init_1_7_7_9
void FASTCALL call_1_7_7_9(SH2_struct* sh)
{
  SH2movls4(sh, 7,7, 9);
}
#define init_1_7_7_10
void FASTCALL call_1_7_7_10(SH2_struct* sh)
{
  SH2movls4(sh, 7,7, 10);
}
#define init_1_7_7_11
void FASTCALL call_1_7_7_11(SH2_struct* sh)
{
  SH2movls4(sh, 7,7, 11);
}
#define init_1_7_7_12
void FASTCALL call_1_7_7_12(SH2_struct* sh)
{
  SH2movls4(sh, 7,7, 12);
}
#define init_1_7_7_13
void FASTCALL call_1_7_7_13(SH2_struct* sh)
{
  SH2movls4(sh, 7,7, 13);
}
#define init_1_7_7_14
void FASTCALL call_1_7_7_14(SH2_struct* sh)
{
  SH2movls4(sh, 7,7, 14);
}
#define init_1_7_7_15
void FASTCALL call_1_7_7_15(SH2_struct* sh)
{
  SH2movls4(sh, 7,7, 15);
}
#define init_1_7_8_0
void FASTCALL call_1_7_8_0(SH2_struct* sh)
{
  SH2movls4(sh, 7,8, 0);
}
#define init_1_7_8_1
void FASTCALL call_1_7_8_1(SH2_struct* sh)
{
  SH2movls4(sh, 7,8, 1);
}
#define init_1_7_8_2
void FASTCALL call_1_7_8_2(SH2_struct* sh)
{
  SH2movls4(sh, 7,8, 2);
}
#define init_1_7_8_3
void FASTCALL call_1_7_8_3(SH2_struct* sh)
{
  SH2movls4(sh, 7,8, 3);
}
#define init_1_7_8_4
void FASTCALL call_1_7_8_4(SH2_struct* sh)
{
  SH2movls4(sh, 7,8, 4);
}
#define init_1_7_8_5
void FASTCALL call_1_7_8_5(SH2_struct* sh)
{
  SH2movls4(sh, 7,8, 5);
}
#define init_1_7_8_6
void FASTCALL call_1_7_8_6(SH2_struct* sh)
{
  SH2movls4(sh, 7,8, 6);
}
#define init_1_7_8_7
void FASTCALL call_1_7_8_7(SH2_struct* sh)
{
  SH2movls4(sh, 7,8, 7);
}
#define init_1_7_8_8
void FASTCALL call_1_7_8_8(SH2_struct* sh)
{
  SH2movls4(sh, 7,8, 8);
}
#define init_1_7_8_9
void FASTCALL call_1_7_8_9(SH2_struct* sh)
{
  SH2movls4(sh, 7,8, 9);
}
#define init_1_7_8_10
void FASTCALL call_1_7_8_10(SH2_struct* sh)
{
  SH2movls4(sh, 7,8, 10);
}
#define init_1_7_8_11
void FASTCALL call_1_7_8_11(SH2_struct* sh)
{
  SH2movls4(sh, 7,8, 11);
}
#define init_1_7_8_12
void FASTCALL call_1_7_8_12(SH2_struct* sh)
{
  SH2movls4(sh, 7,8, 12);
}
#define init_1_7_8_13
void FASTCALL call_1_7_8_13(SH2_struct* sh)
{
  SH2movls4(sh, 7,8, 13);
}
#define init_1_7_8_14
void FASTCALL call_1_7_8_14(SH2_struct* sh)
{
  SH2movls4(sh, 7,8, 14);
}
#define init_1_7_8_15
void FASTCALL call_1_7_8_15(SH2_struct* sh)
{
  SH2movls4(sh, 7,8, 15);
}
#define init_1_7_9_0
void FASTCALL call_1_7_9_0(SH2_struct* sh)
{
  SH2movls4(sh, 7,9, 0);
}
#define init_1_7_9_1
void FASTCALL call_1_7_9_1(SH2_struct* sh)
{
  SH2movls4(sh, 7,9, 1);
}
#define init_1_7_9_2
void FASTCALL call_1_7_9_2(SH2_struct* sh)
{
  SH2movls4(sh, 7,9, 2);
}
#define init_1_7_9_3
void FASTCALL call_1_7_9_3(SH2_struct* sh)
{
  SH2movls4(sh, 7,9, 3);
}
#define init_1_7_9_4
void FASTCALL call_1_7_9_4(SH2_struct* sh)
{
  SH2movls4(sh, 7,9, 4);
}
#define init_1_7_9_5
void FASTCALL call_1_7_9_5(SH2_struct* sh)
{
  SH2movls4(sh, 7,9, 5);
}
#define init_1_7_9_6
void FASTCALL call_1_7_9_6(SH2_struct* sh)
{
  SH2movls4(sh, 7,9, 6);
}
#define init_1_7_9_7
void FASTCALL call_1_7_9_7(SH2_struct* sh)
{
  SH2movls4(sh, 7,9, 7);
}
#define init_1_7_9_8
void FASTCALL call_1_7_9_8(SH2_struct* sh)
{
  SH2movls4(sh, 7,9, 8);
}
#define init_1_7_9_9
void FASTCALL call_1_7_9_9(SH2_struct* sh)
{
  SH2movls4(sh, 7,9, 9);
}
#define init_1_7_9_10
void FASTCALL call_1_7_9_10(SH2_struct* sh)
{
  SH2movls4(sh, 7,9, 10);
}
#define init_1_7_9_11
void FASTCALL call_1_7_9_11(SH2_struct* sh)
{
  SH2movls4(sh, 7,9, 11);
}
#define init_1_7_9_12
void FASTCALL call_1_7_9_12(SH2_struct* sh)
{
  SH2movls4(sh, 7,9, 12);
}
#define init_1_7_9_13
void FASTCALL call_1_7_9_13(SH2_struct* sh)
{
  SH2movls4(sh, 7,9, 13);
}
#define init_1_7_9_14
void FASTCALL call_1_7_9_14(SH2_struct* sh)
{
  SH2movls4(sh, 7,9, 14);
}
#define init_1_7_9_15
void FASTCALL call_1_7_9_15(SH2_struct* sh)
{
  SH2movls4(sh, 7,9, 15);
}
#define init_1_7_10_0
void FASTCALL call_1_7_10_0(SH2_struct* sh)
{
  SH2movls4(sh, 7,10, 0);
}
#define init_1_7_10_1
void FASTCALL call_1_7_10_1(SH2_struct* sh)
{
  SH2movls4(sh, 7,10, 1);
}
#define init_1_7_10_2
void FASTCALL call_1_7_10_2(SH2_struct* sh)
{
  SH2movls4(sh, 7,10, 2);
}
#define init_1_7_10_3
void FASTCALL call_1_7_10_3(SH2_struct* sh)
{
  SH2movls4(sh, 7,10, 3);
}
#define init_1_7_10_4
void FASTCALL call_1_7_10_4(SH2_struct* sh)
{
  SH2movls4(sh, 7,10, 4);
}
#define init_1_7_10_5
void FASTCALL call_1_7_10_5(SH2_struct* sh)
{
  SH2movls4(sh, 7,10, 5);
}
#define init_1_7_10_6
void FASTCALL call_1_7_10_6(SH2_struct* sh)
{
  SH2movls4(sh, 7,10, 6);
}
#define init_1_7_10_7
void FASTCALL call_1_7_10_7(SH2_struct* sh)
{
  SH2movls4(sh, 7,10, 7);
}
#define init_1_7_10_8
void FASTCALL call_1_7_10_8(SH2_struct* sh)
{
  SH2movls4(sh, 7,10, 8);
}
#define init_1_7_10_9
void FASTCALL call_1_7_10_9(SH2_struct* sh)
{
  SH2movls4(sh, 7,10, 9);
}
#define init_1_7_10_10
void FASTCALL call_1_7_10_10(SH2_struct* sh)
{
  SH2movls4(sh, 7,10, 10);
}
#define init_1_7_10_11
void FASTCALL call_1_7_10_11(SH2_struct* sh)
{
  SH2movls4(sh, 7,10, 11);
}
#define init_1_7_10_12
void FASTCALL call_1_7_10_12(SH2_struct* sh)
{
  SH2movls4(sh, 7,10, 12);
}
#define init_1_7_10_13
void FASTCALL call_1_7_10_13(SH2_struct* sh)
{
  SH2movls4(sh, 7,10, 13);
}
#define init_1_7_10_14
void FASTCALL call_1_7_10_14(SH2_struct* sh)
{
  SH2movls4(sh, 7,10, 14);
}
#define init_1_7_10_15
void FASTCALL call_1_7_10_15(SH2_struct* sh)
{
  SH2movls4(sh, 7,10, 15);
}
#define init_1_7_11_0
void FASTCALL call_1_7_11_0(SH2_struct* sh)
{
  SH2movls4(sh, 7,11, 0);
}
#define init_1_7_11_1
void FASTCALL call_1_7_11_1(SH2_struct* sh)
{
  SH2movls4(sh, 7,11, 1);
}
#define init_1_7_11_2
void FASTCALL call_1_7_11_2(SH2_struct* sh)
{
  SH2movls4(sh, 7,11, 2);
}
#define init_1_7_11_3
void FASTCALL call_1_7_11_3(SH2_struct* sh)
{
  SH2movls4(sh, 7,11, 3);
}
#define init_1_7_11_4
void FASTCALL call_1_7_11_4(SH2_struct* sh)
{
  SH2movls4(sh, 7,11, 4);
}
#define init_1_7_11_5
void FASTCALL call_1_7_11_5(SH2_struct* sh)
{
  SH2movls4(sh, 7,11, 5);
}
#define init_1_7_11_6
void FASTCALL call_1_7_11_6(SH2_struct* sh)
{
  SH2movls4(sh, 7,11, 6);
}
#define init_1_7_11_7
void FASTCALL call_1_7_11_7(SH2_struct* sh)
{
  SH2movls4(sh, 7,11, 7);
}
#define init_1_7_11_8
void FASTCALL call_1_7_11_8(SH2_struct* sh)
{
  SH2movls4(sh, 7,11, 8);
}
#define init_1_7_11_9
void FASTCALL call_1_7_11_9(SH2_struct* sh)
{
  SH2movls4(sh, 7,11, 9);
}
#define init_1_7_11_10
void FASTCALL call_1_7_11_10(SH2_struct* sh)
{
  SH2movls4(sh, 7,11, 10);
}
#define init_1_7_11_11
void FASTCALL call_1_7_11_11(SH2_struct* sh)
{
  SH2movls4(sh, 7,11, 11);
}
#define init_1_7_11_12
void FASTCALL call_1_7_11_12(SH2_struct* sh)
{
  SH2movls4(sh, 7,11, 12);
}
#define init_1_7_11_13
void FASTCALL call_1_7_11_13(SH2_struct* sh)
{
  SH2movls4(sh, 7,11, 13);
}
#define init_1_7_11_14
void FASTCALL call_1_7_11_14(SH2_struct* sh)
{
  SH2movls4(sh, 7,11, 14);
}
#define init_1_7_11_15
void FASTCALL call_1_7_11_15(SH2_struct* sh)
{
  SH2movls4(sh, 7,11, 15);
}
#define init_1_7_12_0
void FASTCALL call_1_7_12_0(SH2_struct* sh)
{
  SH2movls4(sh, 7,12, 0);
}
#define init_1_7_12_1
void FASTCALL call_1_7_12_1(SH2_struct* sh)
{
  SH2movls4(sh, 7,12, 1);
}
#define init_1_7_12_2
void FASTCALL call_1_7_12_2(SH2_struct* sh)
{
  SH2movls4(sh, 7,12, 2);
}
#define init_1_7_12_3
void FASTCALL call_1_7_12_3(SH2_struct* sh)
{
  SH2movls4(sh, 7,12, 3);
}
#define init_1_7_12_4
void FASTCALL call_1_7_12_4(SH2_struct* sh)
{
  SH2movls4(sh, 7,12, 4);
}
#define init_1_7_12_5
void FASTCALL call_1_7_12_5(SH2_struct* sh)
{
  SH2movls4(sh, 7,12, 5);
}
#define init_1_7_12_6
void FASTCALL call_1_7_12_6(SH2_struct* sh)
{
  SH2movls4(sh, 7,12, 6);
}
#define init_1_7_12_7
void FASTCALL call_1_7_12_7(SH2_struct* sh)
{
  SH2movls4(sh, 7,12, 7);
}
#define init_1_7_12_8
void FASTCALL call_1_7_12_8(SH2_struct* sh)
{
  SH2movls4(sh, 7,12, 8);
}
#define init_1_7_12_9
void FASTCALL call_1_7_12_9(SH2_struct* sh)
{
  SH2movls4(sh, 7,12, 9);
}
#define init_1_7_12_10
void FASTCALL call_1_7_12_10(SH2_struct* sh)
{
  SH2movls4(sh, 7,12, 10);
}
#define init_1_7_12_11
void FASTCALL call_1_7_12_11(SH2_struct* sh)
{
  SH2movls4(sh, 7,12, 11);
}
#define init_1_7_12_12
void FASTCALL call_1_7_12_12(SH2_struct* sh)
{
  SH2movls4(sh, 7,12, 12);
}
#define init_1_7_12_13
void FASTCALL call_1_7_12_13(SH2_struct* sh)
{
  SH2movls4(sh, 7,12, 13);
}
#define init_1_7_12_14
void FASTCALL call_1_7_12_14(SH2_struct* sh)
{
  SH2movls4(sh, 7,12, 14);
}
#define init_1_7_12_15
void FASTCALL call_1_7_12_15(SH2_struct* sh)
{
  SH2movls4(sh, 7,12, 15);
}
#define init_1_7_13_0
void FASTCALL call_1_7_13_0(SH2_struct* sh)
{
  SH2movls4(sh, 7,13, 0);
}
#define init_1_7_13_1
void FASTCALL call_1_7_13_1(SH2_struct* sh)
{
  SH2movls4(sh, 7,13, 1);
}
#define init_1_7_13_2
void FASTCALL call_1_7_13_2(SH2_struct* sh)
{
  SH2movls4(sh, 7,13, 2);
}
#define init_1_7_13_3
void FASTCALL call_1_7_13_3(SH2_struct* sh)
{
  SH2movls4(sh, 7,13, 3);
}
#define init_1_7_13_4
void FASTCALL call_1_7_13_4(SH2_struct* sh)
{
  SH2movls4(sh, 7,13, 4);
}
#define init_1_7_13_5
void FASTCALL call_1_7_13_5(SH2_struct* sh)
{
  SH2movls4(sh, 7,13, 5);
}
#define init_1_7_13_6
void FASTCALL call_1_7_13_6(SH2_struct* sh)
{
  SH2movls4(sh, 7,13, 6);
}
#define init_1_7_13_7
void FASTCALL call_1_7_13_7(SH2_struct* sh)
{
  SH2movls4(sh, 7,13, 7);
}
#define init_1_7_13_8
void FASTCALL call_1_7_13_8(SH2_struct* sh)
{
  SH2movls4(sh, 7,13, 8);
}
#define init_1_7_13_9
void FASTCALL call_1_7_13_9(SH2_struct* sh)
{
  SH2movls4(sh, 7,13, 9);
}
#define init_1_7_13_10
void FASTCALL call_1_7_13_10(SH2_struct* sh)
{
  SH2movls4(sh, 7,13, 10);
}
#define init_1_7_13_11
void FASTCALL call_1_7_13_11(SH2_struct* sh)
{
  SH2movls4(sh, 7,13, 11);
}
#define init_1_7_13_12
void FASTCALL call_1_7_13_12(SH2_struct* sh)
{
  SH2movls4(sh, 7,13, 12);
}
#define init_1_7_13_13
void FASTCALL call_1_7_13_13(SH2_struct* sh)
{
  SH2movls4(sh, 7,13, 13);
}
#define init_1_7_13_14
void FASTCALL call_1_7_13_14(SH2_struct* sh)
{
  SH2movls4(sh, 7,13, 14);
}
#define init_1_7_13_15
void FASTCALL call_1_7_13_15(SH2_struct* sh)
{
  SH2movls4(sh, 7,13, 15);
}
#define init_1_7_14_0
void FASTCALL call_1_7_14_0(SH2_struct* sh)
{
  SH2movls4(sh, 7,14, 0);
}
#define init_1_7_14_1
void FASTCALL call_1_7_14_1(SH2_struct* sh)
{
  SH2movls4(sh, 7,14, 1);
}
#define init_1_7_14_2
void FASTCALL call_1_7_14_2(SH2_struct* sh)
{
  SH2movls4(sh, 7,14, 2);
}
#define init_1_7_14_3
void FASTCALL call_1_7_14_3(SH2_struct* sh)
{
  SH2movls4(sh, 7,14, 3);
}
#define init_1_7_14_4
void FASTCALL call_1_7_14_4(SH2_struct* sh)
{
  SH2movls4(sh, 7,14, 4);
}
#define init_1_7_14_5
void FASTCALL call_1_7_14_5(SH2_struct* sh)
{
  SH2movls4(sh, 7,14, 5);
}
#define init_1_7_14_6
void FASTCALL call_1_7_14_6(SH2_struct* sh)
{
  SH2movls4(sh, 7,14, 6);
}
#define init_1_7_14_7
void FASTCALL call_1_7_14_7(SH2_struct* sh)
{
  SH2movls4(sh, 7,14, 7);
}
#define init_1_7_14_8
void FASTCALL call_1_7_14_8(SH2_struct* sh)
{
  SH2movls4(sh, 7,14, 8);
}
#define init_1_7_14_9
void FASTCALL call_1_7_14_9(SH2_struct* sh)
{
  SH2movls4(sh, 7,14, 9);
}
#define init_1_7_14_10
void FASTCALL call_1_7_14_10(SH2_struct* sh)
{
  SH2movls4(sh, 7,14, 10);
}
#define init_1_7_14_11
void FASTCALL call_1_7_14_11(SH2_struct* sh)
{
  SH2movls4(sh, 7,14, 11);
}
#define init_1_7_14_12
void FASTCALL call_1_7_14_12(SH2_struct* sh)
{
  SH2movls4(sh, 7,14, 12);
}
#define init_1_7_14_13
void FASTCALL call_1_7_14_13(SH2_struct* sh)
{
  SH2movls4(sh, 7,14, 13);
}
#define init_1_7_14_14
void FASTCALL call_1_7_14_14(SH2_struct* sh)
{
  SH2movls4(sh, 7,14, 14);
}
#define init_1_7_14_15
void FASTCALL call_1_7_14_15(SH2_struct* sh)
{
  SH2movls4(sh, 7,14, 15);
}
#define init_1_7_15_0
void FASTCALL call_1_7_15_0(SH2_struct* sh)
{
  SH2movls4(sh, 7,15, 0);
}
#define init_1_7_15_1
void FASTCALL call_1_7_15_1(SH2_struct* sh)
{
  SH2movls4(sh, 7,15, 1);
}
#define init_1_7_15_2
void FASTCALL call_1_7_15_2(SH2_struct* sh)
{
  SH2movls4(sh, 7,15, 2);
}
#define init_1_7_15_3
void FASTCALL call_1_7_15_3(SH2_struct* sh)
{
  SH2movls4(sh, 7,15, 3);
}
#define init_1_7_15_4
void FASTCALL call_1_7_15_4(SH2_struct* sh)
{
  SH2movls4(sh, 7,15, 4);
}
#define init_1_7_15_5
void FASTCALL call_1_7_15_5(SH2_struct* sh)
{
  SH2movls4(sh, 7,15, 5);
}
#define init_1_7_15_6
void FASTCALL call_1_7_15_6(SH2_struct* sh)
{
  SH2movls4(sh, 7,15, 6);
}
#define init_1_7_15_7
void FASTCALL call_1_7_15_7(SH2_struct* sh)
{
  SH2movls4(sh, 7,15, 7);
}
#define init_1_7_15_8
void FASTCALL call_1_7_15_8(SH2_struct* sh)
{
  SH2movls4(sh, 7,15, 8);
}
#define init_1_7_15_9
void FASTCALL call_1_7_15_9(SH2_struct* sh)
{
  SH2movls4(sh, 7,15, 9);
}
#define init_1_7_15_10
void FASTCALL call_1_7_15_10(SH2_struct* sh)
{
  SH2movls4(sh, 7,15, 10);
}
#define init_1_7_15_11
void FASTCALL call_1_7_15_11(SH2_struct* sh)
{
  SH2movls4(sh, 7,15, 11);
}
#define init_1_7_15_12
void FASTCALL call_1_7_15_12(SH2_struct* sh)
{
  SH2movls4(sh, 7,15, 12);
}
#define init_1_7_15_13
void FASTCALL call_1_7_15_13(SH2_struct* sh)
{
  SH2movls4(sh, 7,15, 13);
}
#define init_1_7_15_14
void FASTCALL call_1_7_15_14(SH2_struct* sh)
{
  SH2movls4(sh, 7,15, 14);
}
#define init_1_7_15_15
void FASTCALL call_1_7_15_15(SH2_struct* sh)
{
  SH2movls4(sh, 7,15, 15);
}
#define init_1_8_0_0
void FASTCALL call_1_8_0_0(SH2_struct* sh)
{
  SH2movls4(sh, 8,0, 0);
}
#define init_1_8_0_1
void FASTCALL call_1_8_0_1(SH2_struct* sh)
{
  SH2movls4(sh, 8,0, 1);
}
#define init_1_8_0_2
void FASTCALL call_1_8_0_2(SH2_struct* sh)
{
  SH2movls4(sh, 8,0, 2);
}
#define init_1_8_0_3
void FASTCALL call_1_8_0_3(SH2_struct* sh)
{
  SH2movls4(sh, 8,0, 3);
}
#define init_1_8_0_4
void FASTCALL call_1_8_0_4(SH2_struct* sh)
{
  SH2movls4(sh, 8,0, 4);
}
#define init_1_8_0_5
void FASTCALL call_1_8_0_5(SH2_struct* sh)
{
  SH2movls4(sh, 8,0, 5);
}
#define init_1_8_0_6
void FASTCALL call_1_8_0_6(SH2_struct* sh)
{
  SH2movls4(sh, 8,0, 6);
}
#define init_1_8_0_7
void FASTCALL call_1_8_0_7(SH2_struct* sh)
{
  SH2movls4(sh, 8,0, 7);
}
#define init_1_8_0_8
void FASTCALL call_1_8_0_8(SH2_struct* sh)
{
  SH2movls4(sh, 8,0, 8);
}
#define init_1_8_0_9
void FASTCALL call_1_8_0_9(SH2_struct* sh)
{
  SH2movls4(sh, 8,0, 9);
}
#define init_1_8_0_10
void FASTCALL call_1_8_0_10(SH2_struct* sh)
{
  SH2movls4(sh, 8,0, 10);
}
#define init_1_8_0_11
void FASTCALL call_1_8_0_11(SH2_struct* sh)
{
  SH2movls4(sh, 8,0, 11);
}
#define init_1_8_0_12
void FASTCALL call_1_8_0_12(SH2_struct* sh)
{
  SH2movls4(sh, 8,0, 12);
}
#define init_1_8_0_13
void FASTCALL call_1_8_0_13(SH2_struct* sh)
{
  SH2movls4(sh, 8,0, 13);
}
#define init_1_8_0_14
void FASTCALL call_1_8_0_14(SH2_struct* sh)
{
  SH2movls4(sh, 8,0, 14);
}
#define init_1_8_0_15
void FASTCALL call_1_8_0_15(SH2_struct* sh)
{
  SH2movls4(sh, 8,0, 15);
}
#define init_1_8_1_0
void FASTCALL call_1_8_1_0(SH2_struct* sh)
{
  SH2movls4(sh, 8,1, 0);
}
#define init_1_8_1_1
void FASTCALL call_1_8_1_1(SH2_struct* sh)
{
  SH2movls4(sh, 8,1, 1);
}
#define init_1_8_1_2
void FASTCALL call_1_8_1_2(SH2_struct* sh)
{
  SH2movls4(sh, 8,1, 2);
}
#define init_1_8_1_3
void FASTCALL call_1_8_1_3(SH2_struct* sh)
{
  SH2movls4(sh, 8,1, 3);
}
#define init_1_8_1_4
void FASTCALL call_1_8_1_4(SH2_struct* sh)
{
  SH2movls4(sh, 8,1, 4);
}
#define init_1_8_1_5
void FASTCALL call_1_8_1_5(SH2_struct* sh)
{
  SH2movls4(sh, 8,1, 5);
}
#define init_1_8_1_6
void FASTCALL call_1_8_1_6(SH2_struct* sh)
{
  SH2movls4(sh, 8,1, 6);
}
#define init_1_8_1_7
void FASTCALL call_1_8_1_7(SH2_struct* sh)
{
  SH2movls4(sh, 8,1, 7);
}
#define init_1_8_1_8
void FASTCALL call_1_8_1_8(SH2_struct* sh)
{
  SH2movls4(sh, 8,1, 8);
}
#define init_1_8_1_9
void FASTCALL call_1_8_1_9(SH2_struct* sh)
{
  SH2movls4(sh, 8,1, 9);
}
#define init_1_8_1_10
void FASTCALL call_1_8_1_10(SH2_struct* sh)
{
  SH2movls4(sh, 8,1, 10);
}
#define init_1_8_1_11
void FASTCALL call_1_8_1_11(SH2_struct* sh)
{
  SH2movls4(sh, 8,1, 11);
}
#define init_1_8_1_12
void FASTCALL call_1_8_1_12(SH2_struct* sh)
{
  SH2movls4(sh, 8,1, 12);
}
#define init_1_8_1_13
void FASTCALL call_1_8_1_13(SH2_struct* sh)
{
  SH2movls4(sh, 8,1, 13);
}
#define init_1_8_1_14
void FASTCALL call_1_8_1_14(SH2_struct* sh)
{
  SH2movls4(sh, 8,1, 14);
}
#define init_1_8_1_15
void FASTCALL call_1_8_1_15(SH2_struct* sh)
{
  SH2movls4(sh, 8,1, 15);
}
#define init_1_8_2_0
void FASTCALL call_1_8_2_0(SH2_struct* sh)
{
  SH2movls4(sh, 8,2, 0);
}
#define init_1_8_2_1
void FASTCALL call_1_8_2_1(SH2_struct* sh)
{
  SH2movls4(sh, 8,2, 1);
}
#define init_1_8_2_2
void FASTCALL call_1_8_2_2(SH2_struct* sh)
{
  SH2movls4(sh, 8,2, 2);
}
#define init_1_8_2_3
void FASTCALL call_1_8_2_3(SH2_struct* sh)
{
  SH2movls4(sh, 8,2, 3);
}
#define init_1_8_2_4
void FASTCALL call_1_8_2_4(SH2_struct* sh)
{
  SH2movls4(sh, 8,2, 4);
}
#define init_1_8_2_5
void FASTCALL call_1_8_2_5(SH2_struct* sh)
{
  SH2movls4(sh, 8,2, 5);
}
#define init_1_8_2_6
void FASTCALL call_1_8_2_6(SH2_struct* sh)
{
  SH2movls4(sh, 8,2, 6);
}
#define init_1_8_2_7
void FASTCALL call_1_8_2_7(SH2_struct* sh)
{
  SH2movls4(sh, 8,2, 7);
}
#define init_1_8_2_8
void FASTCALL call_1_8_2_8(SH2_struct* sh)
{
  SH2movls4(sh, 8,2, 8);
}
#define init_1_8_2_9
void FASTCALL call_1_8_2_9(SH2_struct* sh)
{
  SH2movls4(sh, 8,2, 9);
}
#define init_1_8_2_10
void FASTCALL call_1_8_2_10(SH2_struct* sh)
{
  SH2movls4(sh, 8,2, 10);
}
#define init_1_8_2_11
void FASTCALL call_1_8_2_11(SH2_struct* sh)
{
  SH2movls4(sh, 8,2, 11);
}
#define init_1_8_2_12
void FASTCALL call_1_8_2_12(SH2_struct* sh)
{
  SH2movls4(sh, 8,2, 12);
}
#define init_1_8_2_13
void FASTCALL call_1_8_2_13(SH2_struct* sh)
{
  SH2movls4(sh, 8,2, 13);
}
#define init_1_8_2_14
void FASTCALL call_1_8_2_14(SH2_struct* sh)
{
  SH2movls4(sh, 8,2, 14);
}
#define init_1_8_2_15
void FASTCALL call_1_8_2_15(SH2_struct* sh)
{
  SH2movls4(sh, 8,2, 15);
}
#define init_1_8_3_0
void FASTCALL call_1_8_3_0(SH2_struct* sh)
{
  SH2movls4(sh, 8,3, 0);
}
#define init_1_8_3_1
void FASTCALL call_1_8_3_1(SH2_struct* sh)
{
  SH2movls4(sh, 8,3, 1);
}
#define init_1_8_3_2
void FASTCALL call_1_8_3_2(SH2_struct* sh)
{
  SH2movls4(sh, 8,3, 2);
}
#define init_1_8_3_3
void FASTCALL call_1_8_3_3(SH2_struct* sh)
{
  SH2movls4(sh, 8,3, 3);
}
#define init_1_8_3_4
void FASTCALL call_1_8_3_4(SH2_struct* sh)
{
  SH2movls4(sh, 8,3, 4);
}
#define init_1_8_3_5
void FASTCALL call_1_8_3_5(SH2_struct* sh)
{
  SH2movls4(sh, 8,3, 5);
}
#define init_1_8_3_6
void FASTCALL call_1_8_3_6(SH2_struct* sh)
{
  SH2movls4(sh, 8,3, 6);
}
#define init_1_8_3_7
void FASTCALL call_1_8_3_7(SH2_struct* sh)
{
  SH2movls4(sh, 8,3, 7);
}
#define init_1_8_3_8
void FASTCALL call_1_8_3_8(SH2_struct* sh)
{
  SH2movls4(sh, 8,3, 8);
}
#define init_1_8_3_9
void FASTCALL call_1_8_3_9(SH2_struct* sh)
{
  SH2movls4(sh, 8,3, 9);
}
#define init_1_8_3_10
void FASTCALL call_1_8_3_10(SH2_struct* sh)
{
  SH2movls4(sh, 8,3, 10);
}
#define init_1_8_3_11
void FASTCALL call_1_8_3_11(SH2_struct* sh)
{
  SH2movls4(sh, 8,3, 11);
}
#define init_1_8_3_12
void FASTCALL call_1_8_3_12(SH2_struct* sh)
{
  SH2movls4(sh, 8,3, 12);
}
#define init_1_8_3_13
void FASTCALL call_1_8_3_13(SH2_struct* sh)
{
  SH2movls4(sh, 8,3, 13);
}
#define init_1_8_3_14
void FASTCALL call_1_8_3_14(SH2_struct* sh)
{
  SH2movls4(sh, 8,3, 14);
}
#define init_1_8_3_15
void FASTCALL call_1_8_3_15(SH2_struct* sh)
{
  SH2movls4(sh, 8,3, 15);
}
#define init_1_8_4_0
void FASTCALL call_1_8_4_0(SH2_struct* sh)
{
  SH2movls4(sh, 8,4, 0);
}
#define init_1_8_4_1
void FASTCALL call_1_8_4_1(SH2_struct* sh)
{
  SH2movls4(sh, 8,4, 1);
}
#define init_1_8_4_2
void FASTCALL call_1_8_4_2(SH2_struct* sh)
{
  SH2movls4(sh, 8,4, 2);
}
#define init_1_8_4_3
void FASTCALL call_1_8_4_3(SH2_struct* sh)
{
  SH2movls4(sh, 8,4, 3);
}
#define init_1_8_4_4
void FASTCALL call_1_8_4_4(SH2_struct* sh)
{
  SH2movls4(sh, 8,4, 4);
}
#define init_1_8_4_5
void FASTCALL call_1_8_4_5(SH2_struct* sh)
{
  SH2movls4(sh, 8,4, 5);
}
#define init_1_8_4_6
void FASTCALL call_1_8_4_6(SH2_struct* sh)
{
  SH2movls4(sh, 8,4, 6);
}
#define init_1_8_4_7
void FASTCALL call_1_8_4_7(SH2_struct* sh)
{
  SH2movls4(sh, 8,4, 7);
}
#define init_1_8_4_8
void FASTCALL call_1_8_4_8(SH2_struct* sh)
{
  SH2movls4(sh, 8,4, 8);
}
#define init_1_8_4_9
void FASTCALL call_1_8_4_9(SH2_struct* sh)
{
  SH2movls4(sh, 8,4, 9);
}
#define init_1_8_4_10
void FASTCALL call_1_8_4_10(SH2_struct* sh)
{
  SH2movls4(sh, 8,4, 10);
}
#define init_1_8_4_11
void FASTCALL call_1_8_4_11(SH2_struct* sh)
{
  SH2movls4(sh, 8,4, 11);
}
#define init_1_8_4_12
void FASTCALL call_1_8_4_12(SH2_struct* sh)
{
  SH2movls4(sh, 8,4, 12);
}
#define init_1_8_4_13
void FASTCALL call_1_8_4_13(SH2_struct* sh)
{
  SH2movls4(sh, 8,4, 13);
}
#define init_1_8_4_14
void FASTCALL call_1_8_4_14(SH2_struct* sh)
{
  SH2movls4(sh, 8,4, 14);
}
#define init_1_8_4_15
void FASTCALL call_1_8_4_15(SH2_struct* sh)
{
  SH2movls4(sh, 8,4, 15);
}
#define init_1_8_5_0
void FASTCALL call_1_8_5_0(SH2_struct* sh)
{
  SH2movls4(sh, 8,5, 0);
}
#define init_1_8_5_1
void FASTCALL call_1_8_5_1(SH2_struct* sh)
{
  SH2movls4(sh, 8,5, 1);
}
#define init_1_8_5_2
void FASTCALL call_1_8_5_2(SH2_struct* sh)
{
  SH2movls4(sh, 8,5, 2);
}
#define init_1_8_5_3
void FASTCALL call_1_8_5_3(SH2_struct* sh)
{
  SH2movls4(sh, 8,5, 3);
}
#define init_1_8_5_4
void FASTCALL call_1_8_5_4(SH2_struct* sh)
{
  SH2movls4(sh, 8,5, 4);
}
#define init_1_8_5_5
void FASTCALL call_1_8_5_5(SH2_struct* sh)
{
  SH2movls4(sh, 8,5, 5);
}
#define init_1_8_5_6
void FASTCALL call_1_8_5_6(SH2_struct* sh)
{
  SH2movls4(sh, 8,5, 6);
}
#define init_1_8_5_7
void FASTCALL call_1_8_5_7(SH2_struct* sh)
{
  SH2movls4(sh, 8,5, 7);
}
#define init_1_8_5_8
void FASTCALL call_1_8_5_8(SH2_struct* sh)
{
  SH2movls4(sh, 8,5, 8);
}
#define init_1_8_5_9
void FASTCALL call_1_8_5_9(SH2_struct* sh)
{
  SH2movls4(sh, 8,5, 9);
}
#define init_1_8_5_10
void FASTCALL call_1_8_5_10(SH2_struct* sh)
{
  SH2movls4(sh, 8,5, 10);
}
#define init_1_8_5_11
void FASTCALL call_1_8_5_11(SH2_struct* sh)
{
  SH2movls4(sh, 8,5, 11);
}
#define init_1_8_5_12
void FASTCALL call_1_8_5_12(SH2_struct* sh)
{
  SH2movls4(sh, 8,5, 12);
}
#define init_1_8_5_13
void FASTCALL call_1_8_5_13(SH2_struct* sh)
{
  SH2movls4(sh, 8,5, 13);
}
#define init_1_8_5_14
void FASTCALL call_1_8_5_14(SH2_struct* sh)
{
  SH2movls4(sh, 8,5, 14);
}
#define init_1_8_5_15
void FASTCALL call_1_8_5_15(SH2_struct* sh)
{
  SH2movls4(sh, 8,5, 15);
}
#define init_1_8_6_0
void FASTCALL call_1_8_6_0(SH2_struct* sh)
{
  SH2movls4(sh, 8,6, 0);
}
#define init_1_8_6_1
void FASTCALL call_1_8_6_1(SH2_struct* sh)
{
  SH2movls4(sh, 8,6, 1);
}
#define init_1_8_6_2
void FASTCALL call_1_8_6_2(SH2_struct* sh)
{
  SH2movls4(sh, 8,6, 2);
}
#define init_1_8_6_3
void FASTCALL call_1_8_6_3(SH2_struct* sh)
{
  SH2movls4(sh, 8,6, 3);
}
#define init_1_8_6_4
void FASTCALL call_1_8_6_4(SH2_struct* sh)
{
  SH2movls4(sh, 8,6, 4);
}
#define init_1_8_6_5
void FASTCALL call_1_8_6_5(SH2_struct* sh)
{
  SH2movls4(sh, 8,6, 5);
}
#define init_1_8_6_6
void FASTCALL call_1_8_6_6(SH2_struct* sh)
{
  SH2movls4(sh, 8,6, 6);
}
#define init_1_8_6_7
void FASTCALL call_1_8_6_7(SH2_struct* sh)
{
  SH2movls4(sh, 8,6, 7);
}
#define init_1_8_6_8
void FASTCALL call_1_8_6_8(SH2_struct* sh)
{
  SH2movls4(sh, 8,6, 8);
}
#define init_1_8_6_9
void FASTCALL call_1_8_6_9(SH2_struct* sh)
{
  SH2movls4(sh, 8,6, 9);
}
#define init_1_8_6_10
void FASTCALL call_1_8_6_10(SH2_struct* sh)
{
  SH2movls4(sh, 8,6, 10);
}
#define init_1_8_6_11
void FASTCALL call_1_8_6_11(SH2_struct* sh)
{
  SH2movls4(sh, 8,6, 11);
}
#define init_1_8_6_12
void FASTCALL call_1_8_6_12(SH2_struct* sh)
{
  SH2movls4(sh, 8,6, 12);
}
#define init_1_8_6_13
void FASTCALL call_1_8_6_13(SH2_struct* sh)
{
  SH2movls4(sh, 8,6, 13);
}
#define init_1_8_6_14
void FASTCALL call_1_8_6_14(SH2_struct* sh)
{
  SH2movls4(sh, 8,6, 14);
}
#define init_1_8_6_15
void FASTCALL call_1_8_6_15(SH2_struct* sh)
{
  SH2movls4(sh, 8,6, 15);
}
#define init_1_8_7_0
void FASTCALL call_1_8_7_0(SH2_struct* sh)
{
  SH2movls4(sh, 8,7, 0);
}
#define init_1_8_7_1
void FASTCALL call_1_8_7_1(SH2_struct* sh)
{
  SH2movls4(sh, 8,7, 1);
}
#define init_1_8_7_2
void FASTCALL call_1_8_7_2(SH2_struct* sh)
{
  SH2movls4(sh, 8,7, 2);
}
#define init_1_8_7_3
void FASTCALL call_1_8_7_3(SH2_struct* sh)
{
  SH2movls4(sh, 8,7, 3);
}
#define init_1_8_7_4
void FASTCALL call_1_8_7_4(SH2_struct* sh)
{
  SH2movls4(sh, 8,7, 4);
}
#define init_1_8_7_5
void FASTCALL call_1_8_7_5(SH2_struct* sh)
{
  SH2movls4(sh, 8,7, 5);
}
#define init_1_8_7_6
void FASTCALL call_1_8_7_6(SH2_struct* sh)
{
  SH2movls4(sh, 8,7, 6);
}
#define init_1_8_7_7
void FASTCALL call_1_8_7_7(SH2_struct* sh)
{
  SH2movls4(sh, 8,7, 7);
}
#define init_1_8_7_8
void FASTCALL call_1_8_7_8(SH2_struct* sh)
{
  SH2movls4(sh, 8,7, 8);
}
#define init_1_8_7_9
void FASTCALL call_1_8_7_9(SH2_struct* sh)
{
  SH2movls4(sh, 8,7, 9);
}
#define init_1_8_7_10
void FASTCALL call_1_8_7_10(SH2_struct* sh)
{
  SH2movls4(sh, 8,7, 10);
}
#define init_1_8_7_11
void FASTCALL call_1_8_7_11(SH2_struct* sh)
{
  SH2movls4(sh, 8,7, 11);
}
#define init_1_8_7_12
void FASTCALL call_1_8_7_12(SH2_struct* sh)
{
  SH2movls4(sh, 8,7, 12);
}
#define init_1_8_7_13
void FASTCALL call_1_8_7_13(SH2_struct* sh)
{
  SH2movls4(sh, 8,7, 13);
}
#define init_1_8_7_14
void FASTCALL call_1_8_7_14(SH2_struct* sh)
{
  SH2movls4(sh, 8,7, 14);
}
#define init_1_8_7_15
void FASTCALL call_1_8_7_15(SH2_struct* sh)
{
  SH2movls4(sh, 8,7, 15);
}
#define init_1_8_8_0
void FASTCALL call_1_8_8_0(SH2_struct* sh)
{
  SH2movls4(sh, 8,8, 0);
}
#define init_1_8_8_1
void FASTCALL call_1_8_8_1(SH2_struct* sh)
{
  SH2movls4(sh, 8,8, 1);
}
#define init_1_8_8_2
void FASTCALL call_1_8_8_2(SH2_struct* sh)
{
  SH2movls4(sh, 8,8, 2);
}
#define init_1_8_8_3
void FASTCALL call_1_8_8_3(SH2_struct* sh)
{
  SH2movls4(sh, 8,8, 3);
}
#define init_1_8_8_4
void FASTCALL call_1_8_8_4(SH2_struct* sh)
{
  SH2movls4(sh, 8,8, 4);
}
#define init_1_8_8_5
void FASTCALL call_1_8_8_5(SH2_struct* sh)
{
  SH2movls4(sh, 8,8, 5);
}
#define init_1_8_8_6
void FASTCALL call_1_8_8_6(SH2_struct* sh)
{
  SH2movls4(sh, 8,8, 6);
}
#define init_1_8_8_7
void FASTCALL call_1_8_8_7(SH2_struct* sh)
{
  SH2movls4(sh, 8,8, 7);
}
#define init_1_8_8_8
void FASTCALL call_1_8_8_8(SH2_struct* sh)
{
  SH2movls4(sh, 8,8, 8);
}
#define init_1_8_8_9
void FASTCALL call_1_8_8_9(SH2_struct* sh)
{
  SH2movls4(sh, 8,8, 9);
}
#define init_1_8_8_10
void FASTCALL call_1_8_8_10(SH2_struct* sh)
{
  SH2movls4(sh, 8,8, 10);
}
#define init_1_8_8_11
void FASTCALL call_1_8_8_11(SH2_struct* sh)
{
  SH2movls4(sh, 8,8, 11);
}
#define init_1_8_8_12
void FASTCALL call_1_8_8_12(SH2_struct* sh)
{
  SH2movls4(sh, 8,8, 12);
}
#define init_1_8_8_13
void FASTCALL call_1_8_8_13(SH2_struct* sh)
{
  SH2movls4(sh, 8,8, 13);
}
#define init_1_8_8_14
void FASTCALL call_1_8_8_14(SH2_struct* sh)
{
  SH2movls4(sh, 8,8, 14);
}
#define init_1_8_8_15
void FASTCALL call_1_8_8_15(SH2_struct* sh)
{
  SH2movls4(sh, 8,8, 15);
}
#define init_1_8_9_0
void FASTCALL call_1_8_9_0(SH2_struct* sh)
{
  SH2movls4(sh, 8,9, 0);
}
#define init_1_8_9_1
void FASTCALL call_1_8_9_1(SH2_struct* sh)
{
  SH2movls4(sh, 8,9, 1);
}
#define init_1_8_9_2
void FASTCALL call_1_8_9_2(SH2_struct* sh)
{
  SH2movls4(sh, 8,9, 2);
}
#define init_1_8_9_3
void FASTCALL call_1_8_9_3(SH2_struct* sh)
{
  SH2movls4(sh, 8,9, 3);
}
#define init_1_8_9_4
void FASTCALL call_1_8_9_4(SH2_struct* sh)
{
  SH2movls4(sh, 8,9, 4);
}
#define init_1_8_9_5
void FASTCALL call_1_8_9_5(SH2_struct* sh)
{
  SH2movls4(sh, 8,9, 5);
}
#define init_1_8_9_6
void FASTCALL call_1_8_9_6(SH2_struct* sh)
{
  SH2movls4(sh, 8,9, 6);
}
#define init_1_8_9_7
void FASTCALL call_1_8_9_7(SH2_struct* sh)
{
  SH2movls4(sh, 8,9, 7);
}
#define init_1_8_9_8
void FASTCALL call_1_8_9_8(SH2_struct* sh)
{
  SH2movls4(sh, 8,9, 8);
}
#define init_1_8_9_9
void FASTCALL call_1_8_9_9(SH2_struct* sh)
{
  SH2movls4(sh, 8,9, 9);
}
#define init_1_8_9_10
void FASTCALL call_1_8_9_10(SH2_struct* sh)
{
  SH2movls4(sh, 8,9, 10);
}
#define init_1_8_9_11
void FASTCALL call_1_8_9_11(SH2_struct* sh)
{
  SH2movls4(sh, 8,9, 11);
}
#define init_1_8_9_12
void FASTCALL call_1_8_9_12(SH2_struct* sh)
{
  SH2movls4(sh, 8,9, 12);
}
#define init_1_8_9_13
void FASTCALL call_1_8_9_13(SH2_struct* sh)
{
  SH2movls4(sh, 8,9, 13);
}
#define init_1_8_9_14
void FASTCALL call_1_8_9_14(SH2_struct* sh)
{
  SH2movls4(sh, 8,9, 14);
}
#define init_1_8_9_15
void FASTCALL call_1_8_9_15(SH2_struct* sh)
{
  SH2movls4(sh, 8,9, 15);
}
#define init_1_8_10_0
void FASTCALL call_1_8_10_0(SH2_struct* sh)
{
  SH2movls4(sh, 8,10, 0);
}
#define init_1_8_10_1
void FASTCALL call_1_8_10_1(SH2_struct* sh)
{
  SH2movls4(sh, 8,10, 1);
}
#define init_1_8_10_2
void FASTCALL call_1_8_10_2(SH2_struct* sh)
{
  SH2movls4(sh, 8,10, 2);
}
#define init_1_8_10_3
void FASTCALL call_1_8_10_3(SH2_struct* sh)
{
  SH2movls4(sh, 8,10, 3);
}
#define init_1_8_10_4
void FASTCALL call_1_8_10_4(SH2_struct* sh)
{
  SH2movls4(sh, 8,10, 4);
}
#define init_1_8_10_5
void FASTCALL call_1_8_10_5(SH2_struct* sh)
{
  SH2movls4(sh, 8,10, 5);
}
#define init_1_8_10_6
void FASTCALL call_1_8_10_6(SH2_struct* sh)
{
  SH2movls4(sh, 8,10, 6);
}
#define init_1_8_10_7
void FASTCALL call_1_8_10_7(SH2_struct* sh)
{
  SH2movls4(sh, 8,10, 7);
}
#define init_1_8_10_8
void FASTCALL call_1_8_10_8(SH2_struct* sh)
{
  SH2movls4(sh, 8,10, 8);
}
#define init_1_8_10_9
void FASTCALL call_1_8_10_9(SH2_struct* sh)
{
  SH2movls4(sh, 8,10, 9);
}
#define init_1_8_10_10
void FASTCALL call_1_8_10_10(SH2_struct* sh)
{
  SH2movls4(sh, 8,10, 10);
}
#define init_1_8_10_11
void FASTCALL call_1_8_10_11(SH2_struct* sh)
{
  SH2movls4(sh, 8,10, 11);
}
#define init_1_8_10_12
void FASTCALL call_1_8_10_12(SH2_struct* sh)
{
  SH2movls4(sh, 8,10, 12);
}
#define init_1_8_10_13
void FASTCALL call_1_8_10_13(SH2_struct* sh)
{
  SH2movls4(sh, 8,10, 13);
}
#define init_1_8_10_14
void FASTCALL call_1_8_10_14(SH2_struct* sh)
{
  SH2movls4(sh, 8,10, 14);
}
#define init_1_8_10_15
void FASTCALL call_1_8_10_15(SH2_struct* sh)
{
  SH2movls4(sh, 8,10, 15);
}
#define init_1_8_11_0
void FASTCALL call_1_8_11_0(SH2_struct* sh)
{
  SH2movls4(sh, 8,11, 0);
}
#define init_1_8_11_1
void FASTCALL call_1_8_11_1(SH2_struct* sh)
{
  SH2movls4(sh, 8,11, 1);
}
#define init_1_8_11_2
void FASTCALL call_1_8_11_2(SH2_struct* sh)
{
  SH2movls4(sh, 8,11, 2);
}
#define init_1_8_11_3
void FASTCALL call_1_8_11_3(SH2_struct* sh)
{
  SH2movls4(sh, 8,11, 3);
}
#define init_1_8_11_4
void FASTCALL call_1_8_11_4(SH2_struct* sh)
{
  SH2movls4(sh, 8,11, 4);
}
#define init_1_8_11_5
void FASTCALL call_1_8_11_5(SH2_struct* sh)
{
  SH2movls4(sh, 8,11, 5);
}
#define init_1_8_11_6
void FASTCALL call_1_8_11_6(SH2_struct* sh)
{
  SH2movls4(sh, 8,11, 6);
}
#define init_1_8_11_7
void FASTCALL call_1_8_11_7(SH2_struct* sh)
{
  SH2movls4(sh, 8,11, 7);
}
#define init_1_8_11_8
void FASTCALL call_1_8_11_8(SH2_struct* sh)
{
  SH2movls4(sh, 8,11, 8);
}
#define init_1_8_11_9
void FASTCALL call_1_8_11_9(SH2_struct* sh)
{
  SH2movls4(sh, 8,11, 9);
}
#define init_1_8_11_10
void FASTCALL call_1_8_11_10(SH2_struct* sh)
{
  SH2movls4(sh, 8,11, 10);
}
#define init_1_8_11_11
void FASTCALL call_1_8_11_11(SH2_struct* sh)
{
  SH2movls4(sh, 8,11, 11);
}
#define init_1_8_11_12
void FASTCALL call_1_8_11_12(SH2_struct* sh)
{
  SH2movls4(sh, 8,11, 12);
}
#define init_1_8_11_13
void FASTCALL call_1_8_11_13(SH2_struct* sh)
{
  SH2movls4(sh, 8,11, 13);
}
#define init_1_8_11_14
void FASTCALL call_1_8_11_14(SH2_struct* sh)
{
  SH2movls4(sh, 8,11, 14);
}
#define init_1_8_11_15
void FASTCALL call_1_8_11_15(SH2_struct* sh)
{
  SH2movls4(sh, 8,11, 15);
}
#define init_1_8_12_0
void FASTCALL call_1_8_12_0(SH2_struct* sh)
{
  SH2movls4(sh, 8,12, 0);
}
#define init_1_8_12_1
void FASTCALL call_1_8_12_1(SH2_struct* sh)
{
  SH2movls4(sh, 8,12, 1);
}
#define init_1_8_12_2
void FASTCALL call_1_8_12_2(SH2_struct* sh)
{
  SH2movls4(sh, 8,12, 2);
}
#define init_1_8_12_3
void FASTCALL call_1_8_12_3(SH2_struct* sh)
{
  SH2movls4(sh, 8,12, 3);
}
#define init_1_8_12_4
void FASTCALL call_1_8_12_4(SH2_struct* sh)
{
  SH2movls4(sh, 8,12, 4);
}
#define init_1_8_12_5
void FASTCALL call_1_8_12_5(SH2_struct* sh)
{
  SH2movls4(sh, 8,12, 5);
}
#define init_1_8_12_6
void FASTCALL call_1_8_12_6(SH2_struct* sh)
{
  SH2movls4(sh, 8,12, 6);
}
#define init_1_8_12_7
void FASTCALL call_1_8_12_7(SH2_struct* sh)
{
  SH2movls4(sh, 8,12, 7);
}
#define init_1_8_12_8
void FASTCALL call_1_8_12_8(SH2_struct* sh)
{
  SH2movls4(sh, 8,12, 8);
}
#define init_1_8_12_9
void FASTCALL call_1_8_12_9(SH2_struct* sh)
{
  SH2movls4(sh, 8,12, 9);
}
#define init_1_8_12_10
void FASTCALL call_1_8_12_10(SH2_struct* sh)
{
  SH2movls4(sh, 8,12, 10);
}
#define init_1_8_12_11
void FASTCALL call_1_8_12_11(SH2_struct* sh)
{
  SH2movls4(sh, 8,12, 11);
}
#define init_1_8_12_12
void FASTCALL call_1_8_12_12(SH2_struct* sh)
{
  SH2movls4(sh, 8,12, 12);
}
#define init_1_8_12_13
void FASTCALL call_1_8_12_13(SH2_struct* sh)
{
  SH2movls4(sh, 8,12, 13);
}
#define init_1_8_12_14
void FASTCALL call_1_8_12_14(SH2_struct* sh)
{
  SH2movls4(sh, 8,12, 14);
}
#define init_1_8_12_15
void FASTCALL call_1_8_12_15(SH2_struct* sh)
{
  SH2movls4(sh, 8,12, 15);
}
#define init_1_8_13_0
void FASTCALL call_1_8_13_0(SH2_struct* sh)
{
  SH2movls4(sh, 8,13, 0);
}
#define init_1_8_13_1
void FASTCALL call_1_8_13_1(SH2_struct* sh)
{
  SH2movls4(sh, 8,13, 1);
}
#define init_1_8_13_2
void FASTCALL call_1_8_13_2(SH2_struct* sh)
{
  SH2movls4(sh, 8,13, 2);
}
#define init_1_8_13_3
void FASTCALL call_1_8_13_3(SH2_struct* sh)
{
  SH2movls4(sh, 8,13, 3);
}
#define init_1_8_13_4
void FASTCALL call_1_8_13_4(SH2_struct* sh)
{
  SH2movls4(sh, 8,13, 4);
}
#define init_1_8_13_5
void FASTCALL call_1_8_13_5(SH2_struct* sh)
{
  SH2movls4(sh, 8,13, 5);
}
#define init_1_8_13_6
void FASTCALL call_1_8_13_6(SH2_struct* sh)
{
  SH2movls4(sh, 8,13, 6);
}
#define init_1_8_13_7
void FASTCALL call_1_8_13_7(SH2_struct* sh)
{
  SH2movls4(sh, 8,13, 7);
}
#define init_1_8_13_8
void FASTCALL call_1_8_13_8(SH2_struct* sh)
{
  SH2movls4(sh, 8,13, 8);
}
#define init_1_8_13_9
void FASTCALL call_1_8_13_9(SH2_struct* sh)
{
  SH2movls4(sh, 8,13, 9);
}
#define init_1_8_13_10
void FASTCALL call_1_8_13_10(SH2_struct* sh)
{
  SH2movls4(sh, 8,13, 10);
}
#define init_1_8_13_11
void FASTCALL call_1_8_13_11(SH2_struct* sh)
{
  SH2movls4(sh, 8,13, 11);
}
#define init_1_8_13_12
void FASTCALL call_1_8_13_12(SH2_struct* sh)
{
  SH2movls4(sh, 8,13, 12);
}
#define init_1_8_13_13
void FASTCALL call_1_8_13_13(SH2_struct* sh)
{
  SH2movls4(sh, 8,13, 13);
}
#define init_1_8_13_14
void FASTCALL call_1_8_13_14(SH2_struct* sh)
{
  SH2movls4(sh, 8,13, 14);
}
#define init_1_8_13_15
void FASTCALL call_1_8_13_15(SH2_struct* sh)
{
  SH2movls4(sh, 8,13, 15);
}
#define init_1_8_14_0
void FASTCALL call_1_8_14_0(SH2_struct* sh)
{
  SH2movls4(sh, 8,14, 0);
}
#define init_1_8_14_1
void FASTCALL call_1_8_14_1(SH2_struct* sh)
{
  SH2movls4(sh, 8,14, 1);
}
#define init_1_8_14_2
void FASTCALL call_1_8_14_2(SH2_struct* sh)
{
  SH2movls4(sh, 8,14, 2);
}
#define init_1_8_14_3
void FASTCALL call_1_8_14_3(SH2_struct* sh)
{
  SH2movls4(sh, 8,14, 3);
}
#define init_1_8_14_4
void FASTCALL call_1_8_14_4(SH2_struct* sh)
{
  SH2movls4(sh, 8,14, 4);
}
#define init_1_8_14_5
void FASTCALL call_1_8_14_5(SH2_struct* sh)
{
  SH2movls4(sh, 8,14, 5);
}
#define init_1_8_14_6
void FASTCALL call_1_8_14_6(SH2_struct* sh)
{
  SH2movls4(sh, 8,14, 6);
}
#define init_1_8_14_7
void FASTCALL call_1_8_14_7(SH2_struct* sh)
{
  SH2movls4(sh, 8,14, 7);
}
#define init_1_8_14_8
void FASTCALL call_1_8_14_8(SH2_struct* sh)
{
  SH2movls4(sh, 8,14, 8);
}
#define init_1_8_14_9
void FASTCALL call_1_8_14_9(SH2_struct* sh)
{
  SH2movls4(sh, 8,14, 9);
}
#define init_1_8_14_10
void FASTCALL call_1_8_14_10(SH2_struct* sh)
{
  SH2movls4(sh, 8,14, 10);
}
#define init_1_8_14_11
void FASTCALL call_1_8_14_11(SH2_struct* sh)
{
  SH2movls4(sh, 8,14, 11);
}
#define init_1_8_14_12
void FASTCALL call_1_8_14_12(SH2_struct* sh)
{
  SH2movls4(sh, 8,14, 12);
}
#define init_1_8_14_13
void FASTCALL call_1_8_14_13(SH2_struct* sh)
{
  SH2movls4(sh, 8,14, 13);
}
#define init_1_8_14_14
void FASTCALL call_1_8_14_14(SH2_struct* sh)
{
  SH2movls4(sh, 8,14, 14);
}
#define init_1_8_14_15
void FASTCALL call_1_8_14_15(SH2_struct* sh)
{
  SH2movls4(sh, 8,14, 15);
}
#define init_1_8_15_0
void FASTCALL call_1_8_15_0(SH2_struct* sh)
{
  SH2movls4(sh, 8,15, 0);
}
#define init_1_8_15_1
void FASTCALL call_1_8_15_1(SH2_struct* sh)
{
  SH2movls4(sh, 8,15, 1);
}
#define init_1_8_15_2
void FASTCALL call_1_8_15_2(SH2_struct* sh)
{
  SH2movls4(sh, 8,15, 2);
}
#define init_1_8_15_3
void FASTCALL call_1_8_15_3(SH2_struct* sh)
{
  SH2movls4(sh, 8,15, 3);
}
#define init_1_8_15_4
void FASTCALL call_1_8_15_4(SH2_struct* sh)
{
  SH2movls4(sh, 8,15, 4);
}
#define init_1_8_15_5
void FASTCALL call_1_8_15_5(SH2_struct* sh)
{
  SH2movls4(sh, 8,15, 5);
}
#define init_1_8_15_6
void FASTCALL call_1_8_15_6(SH2_struct* sh)
{
  SH2movls4(sh, 8,15, 6);
}
#define init_1_8_15_7
void FASTCALL call_1_8_15_7(SH2_struct* sh)
{
  SH2movls4(sh, 8,15, 7);
}
#define init_1_8_15_8
void FASTCALL call_1_8_15_8(SH2_struct* sh)
{
  SH2movls4(sh, 8,15, 8);
}
#define init_1_8_15_9
void FASTCALL call_1_8_15_9(SH2_struct* sh)
{
  SH2movls4(sh, 8,15, 9);
}
#define init_1_8_15_10
void FASTCALL call_1_8_15_10(SH2_struct* sh)
{
  SH2movls4(sh, 8,15, 10);
}
#define init_1_8_15_11
void FASTCALL call_1_8_15_11(SH2_struct* sh)
{
  SH2movls4(sh, 8,15, 11);
}
#define init_1_8_15_12
void FASTCALL call_1_8_15_12(SH2_struct* sh)
{
  SH2movls4(sh, 8,15, 12);
}
#define init_1_8_15_13
void FASTCALL call_1_8_15_13(SH2_struct* sh)
{
  SH2movls4(sh, 8,15, 13);
}
#define init_1_8_15_14
void FASTCALL call_1_8_15_14(SH2_struct* sh)
{
  SH2movls4(sh, 8,15, 14);
}
#define init_1_8_15_15
void FASTCALL call_1_8_15_15(SH2_struct* sh)
{
  SH2movls4(sh, 8,15, 15);
}
#define init_1_9_0_0
void FASTCALL call_1_9_0_0(SH2_struct* sh)
{
  SH2movls4(sh, 9,0, 0);
}
#define init_1_9_0_1
void FASTCALL call_1_9_0_1(SH2_struct* sh)
{
  SH2movls4(sh, 9,0, 1);
}
#define init_1_9_0_2
void FASTCALL call_1_9_0_2(SH2_struct* sh)
{
  SH2movls4(sh, 9,0, 2);
}
#define init_1_9_0_3
void FASTCALL call_1_9_0_3(SH2_struct* sh)
{
  SH2movls4(sh, 9,0, 3);
}
#define init_1_9_0_4
void FASTCALL call_1_9_0_4(SH2_struct* sh)
{
  SH2movls4(sh, 9,0, 4);
}
#define init_1_9_0_5
void FASTCALL call_1_9_0_5(SH2_struct* sh)
{
  SH2movls4(sh, 9,0, 5);
}
#define init_1_9_0_6
void FASTCALL call_1_9_0_6(SH2_struct* sh)
{
  SH2movls4(sh, 9,0, 6);
}
#define init_1_9_0_7
void FASTCALL call_1_9_0_7(SH2_struct* sh)
{
  SH2movls4(sh, 9,0, 7);
}
#define init_1_9_0_8
void FASTCALL call_1_9_0_8(SH2_struct* sh)
{
  SH2movls4(sh, 9,0, 8);
}
#define init_1_9_0_9
void FASTCALL call_1_9_0_9(SH2_struct* sh)
{
  SH2movls4(sh, 9,0, 9);
}
#define init_1_9_0_10
void FASTCALL call_1_9_0_10(SH2_struct* sh)
{
  SH2movls4(sh, 9,0, 10);
}
#define init_1_9_0_11
void FASTCALL call_1_9_0_11(SH2_struct* sh)
{
  SH2movls4(sh, 9,0, 11);
}
#define init_1_9_0_12
void FASTCALL call_1_9_0_12(SH2_struct* sh)
{
  SH2movls4(sh, 9,0, 12);
}
#define init_1_9_0_13
void FASTCALL call_1_9_0_13(SH2_struct* sh)
{
  SH2movls4(sh, 9,0, 13);
}
#define init_1_9_0_14
void FASTCALL call_1_9_0_14(SH2_struct* sh)
{
  SH2movls4(sh, 9,0, 14);
}
#define init_1_9_0_15
void FASTCALL call_1_9_0_15(SH2_struct* sh)
{
  SH2movls4(sh, 9,0, 15);
}
#define init_1_9_1_0
void FASTCALL call_1_9_1_0(SH2_struct* sh)
{
  SH2movls4(sh, 9,1, 0);
}
#define init_1_9_1_1
void FASTCALL call_1_9_1_1(SH2_struct* sh)
{
  SH2movls4(sh, 9,1, 1);
}
#define init_1_9_1_2
void FASTCALL call_1_9_1_2(SH2_struct* sh)
{
  SH2movls4(sh, 9,1, 2);
}
#define init_1_9_1_3
void FASTCALL call_1_9_1_3(SH2_struct* sh)
{
  SH2movls4(sh, 9,1, 3);
}
#define init_1_9_1_4
void FASTCALL call_1_9_1_4(SH2_struct* sh)
{
  SH2movls4(sh, 9,1, 4);
}
#define init_1_9_1_5
void FASTCALL call_1_9_1_5(SH2_struct* sh)
{
  SH2movls4(sh, 9,1, 5);
}
#define init_1_9_1_6
void FASTCALL call_1_9_1_6(SH2_struct* sh)
{
  SH2movls4(sh, 9,1, 6);
}
#define init_1_9_1_7
void FASTCALL call_1_9_1_7(SH2_struct* sh)
{
  SH2movls4(sh, 9,1, 7);
}
#define init_1_9_1_8
void FASTCALL call_1_9_1_8(SH2_struct* sh)
{
  SH2movls4(sh, 9,1, 8);
}
#define init_1_9_1_9
void FASTCALL call_1_9_1_9(SH2_struct* sh)
{
  SH2movls4(sh, 9,1, 9);
}
#define init_1_9_1_10
void FASTCALL call_1_9_1_10(SH2_struct* sh)
{
  SH2movls4(sh, 9,1, 10);
}
#define init_1_9_1_11
void FASTCALL call_1_9_1_11(SH2_struct* sh)
{
  SH2movls4(sh, 9,1, 11);
}
#define init_1_9_1_12
void FASTCALL call_1_9_1_12(SH2_struct* sh)
{
  SH2movls4(sh, 9,1, 12);
}
#define init_1_9_1_13
void FASTCALL call_1_9_1_13(SH2_struct* sh)
{
  SH2movls4(sh, 9,1, 13);
}
#define init_1_9_1_14
void FASTCALL call_1_9_1_14(SH2_struct* sh)
{
  SH2movls4(sh, 9,1, 14);
}
#define init_1_9_1_15
void FASTCALL call_1_9_1_15(SH2_struct* sh)
{
  SH2movls4(sh, 9,1, 15);
}
#define init_1_9_2_0
void FASTCALL call_1_9_2_0(SH2_struct* sh)
{
  SH2movls4(sh, 9,2, 0);
}
#define init_1_9_2_1
void FASTCALL call_1_9_2_1(SH2_struct* sh)
{
  SH2movls4(sh, 9,2, 1);
}
#define init_1_9_2_2
void FASTCALL call_1_9_2_2(SH2_struct* sh)
{
  SH2movls4(sh, 9,2, 2);
}
#define init_1_9_2_3
void FASTCALL call_1_9_2_3(SH2_struct* sh)
{
  SH2movls4(sh, 9,2, 3);
}
#define init_1_9_2_4
void FASTCALL call_1_9_2_4(SH2_struct* sh)
{
  SH2movls4(sh, 9,2, 4);
}
#define init_1_9_2_5
void FASTCALL call_1_9_2_5(SH2_struct* sh)
{
  SH2movls4(sh, 9,2, 5);
}
#define init_1_9_2_6
void FASTCALL call_1_9_2_6(SH2_struct* sh)
{
  SH2movls4(sh, 9,2, 6);
}
#define init_1_9_2_7
void FASTCALL call_1_9_2_7(SH2_struct* sh)
{
  SH2movls4(sh, 9,2, 7);
}
#define init_1_9_2_8
void FASTCALL call_1_9_2_8(SH2_struct* sh)
{
  SH2movls4(sh, 9,2, 8);
}
#define init_1_9_2_9
void FASTCALL call_1_9_2_9(SH2_struct* sh)
{
  SH2movls4(sh, 9,2, 9);
}
#define init_1_9_2_10
void FASTCALL call_1_9_2_10(SH2_struct* sh)
{
  SH2movls4(sh, 9,2, 10);
}
#define init_1_9_2_11
void FASTCALL call_1_9_2_11(SH2_struct* sh)
{
  SH2movls4(sh, 9,2, 11);
}
#define init_1_9_2_12
void FASTCALL call_1_9_2_12(SH2_struct* sh)
{
  SH2movls4(sh, 9,2, 12);
}
#define init_1_9_2_13
void FASTCALL call_1_9_2_13(SH2_struct* sh)
{
  SH2movls4(sh, 9,2, 13);
}
#define init_1_9_2_14
void FASTCALL call_1_9_2_14(SH2_struct* sh)
{
  SH2movls4(sh, 9,2, 14);
}
#define init_1_9_2_15
void FASTCALL call_1_9_2_15(SH2_struct* sh)
{
  SH2movls4(sh, 9,2, 15);
}
#define init_1_9_3_0
void FASTCALL call_1_9_3_0(SH2_struct* sh)
{
  SH2movls4(sh, 9,3, 0);
}
#define init_1_9_3_1
void FASTCALL call_1_9_3_1(SH2_struct* sh)
{
  SH2movls4(sh, 9,3, 1);
}
#define init_1_9_3_2
void FASTCALL call_1_9_3_2(SH2_struct* sh)
{
  SH2movls4(sh, 9,3, 2);
}
#define init_1_9_3_3
void FASTCALL call_1_9_3_3(SH2_struct* sh)
{
  SH2movls4(sh, 9,3, 3);
}
#define init_1_9_3_4
void FASTCALL call_1_9_3_4(SH2_struct* sh)
{
  SH2movls4(sh, 9,3, 4);
}
#define init_1_9_3_5
void FASTCALL call_1_9_3_5(SH2_struct* sh)
{
  SH2movls4(sh, 9,3, 5);
}
#define init_1_9_3_6
void FASTCALL call_1_9_3_6(SH2_struct* sh)
{
  SH2movls4(sh, 9,3, 6);
}
#define init_1_9_3_7
void FASTCALL call_1_9_3_7(SH2_struct* sh)
{
  SH2movls4(sh, 9,3, 7);
}
#define init_1_9_3_8
void FASTCALL call_1_9_3_8(SH2_struct* sh)
{
  SH2movls4(sh, 9,3, 8);
}
#define init_1_9_3_9
void FASTCALL call_1_9_3_9(SH2_struct* sh)
{
  SH2movls4(sh, 9,3, 9);
}
#define init_1_9_3_10
void FASTCALL call_1_9_3_10(SH2_struct* sh)
{
  SH2movls4(sh, 9,3, 10);
}
#define init_1_9_3_11
void FASTCALL call_1_9_3_11(SH2_struct* sh)
{
  SH2movls4(sh, 9,3, 11);
}
#define init_1_9_3_12
void FASTCALL call_1_9_3_12(SH2_struct* sh)
{
  SH2movls4(sh, 9,3, 12);
}
#define init_1_9_3_13
void FASTCALL call_1_9_3_13(SH2_struct* sh)
{
  SH2movls4(sh, 9,3, 13);
}
#define init_1_9_3_14
void FASTCALL call_1_9_3_14(SH2_struct* sh)
{
  SH2movls4(sh, 9,3, 14);
}
#define init_1_9_3_15
void FASTCALL call_1_9_3_15(SH2_struct* sh)
{
  SH2movls4(sh, 9,3, 15);
}
#define init_1_9_4_0
void FASTCALL call_1_9_4_0(SH2_struct* sh)
{
  SH2movls4(sh, 9,4, 0);
}
#define init_1_9_4_1
void FASTCALL call_1_9_4_1(SH2_struct* sh)
{
  SH2movls4(sh, 9,4, 1);
}
#define init_1_9_4_2
void FASTCALL call_1_9_4_2(SH2_struct* sh)
{
  SH2movls4(sh, 9,4, 2);
}
#define init_1_9_4_3
void FASTCALL call_1_9_4_3(SH2_struct* sh)
{
  SH2movls4(sh, 9,4, 3);
}
#define init_1_9_4_4
void FASTCALL call_1_9_4_4(SH2_struct* sh)
{
  SH2movls4(sh, 9,4, 4);
}
#define init_1_9_4_5
void FASTCALL call_1_9_4_5(SH2_struct* sh)
{
  SH2movls4(sh, 9,4, 5);
}
#define init_1_9_4_6
void FASTCALL call_1_9_4_6(SH2_struct* sh)
{
  SH2movls4(sh, 9,4, 6);
}
#define init_1_9_4_7
void FASTCALL call_1_9_4_7(SH2_struct* sh)
{
  SH2movls4(sh, 9,4, 7);
}
#define init_1_9_4_8
void FASTCALL call_1_9_4_8(SH2_struct* sh)
{
  SH2movls4(sh, 9,4, 8);
}
#define init_1_9_4_9
void FASTCALL call_1_9_4_9(SH2_struct* sh)
{
  SH2movls4(sh, 9,4, 9);
}
#define init_1_9_4_10
void FASTCALL call_1_9_4_10(SH2_struct* sh)
{
  SH2movls4(sh, 9,4, 10);
}
#define init_1_9_4_11
void FASTCALL call_1_9_4_11(SH2_struct* sh)
{
  SH2movls4(sh, 9,4, 11);
}
#define init_1_9_4_12
void FASTCALL call_1_9_4_12(SH2_struct* sh)
{
  SH2movls4(sh, 9,4, 12);
}
#define init_1_9_4_13
void FASTCALL call_1_9_4_13(SH2_struct* sh)
{
  SH2movls4(sh, 9,4, 13);
}
#define init_1_9_4_14
void FASTCALL call_1_9_4_14(SH2_struct* sh)
{
  SH2movls4(sh, 9,4, 14);
}
#define init_1_9_4_15
void FASTCALL call_1_9_4_15(SH2_struct* sh)
{
  SH2movls4(sh, 9,4, 15);
}
#define init_1_9_5_0
void FASTCALL call_1_9_5_0(SH2_struct* sh)
{
  SH2movls4(sh, 9,5, 0);
}
#define init_1_9_5_1
void FASTCALL call_1_9_5_1(SH2_struct* sh)
{
  SH2movls4(sh, 9,5, 1);
}
#define init_1_9_5_2
void FASTCALL call_1_9_5_2(SH2_struct* sh)
{
  SH2movls4(sh, 9,5, 2);
}
#define init_1_9_5_3
void FASTCALL call_1_9_5_3(SH2_struct* sh)
{
  SH2movls4(sh, 9,5, 3);
}
#define init_1_9_5_4
void FASTCALL call_1_9_5_4(SH2_struct* sh)
{
  SH2movls4(sh, 9,5, 4);
}
#define init_1_9_5_5
void FASTCALL call_1_9_5_5(SH2_struct* sh)
{
  SH2movls4(sh, 9,5, 5);
}
#define init_1_9_5_6
void FASTCALL call_1_9_5_6(SH2_struct* sh)
{
  SH2movls4(sh, 9,5, 6);
}
#define init_1_9_5_7
void FASTCALL call_1_9_5_7(SH2_struct* sh)
{
  SH2movls4(sh, 9,5, 7);
}
#define init_1_9_5_8
void FASTCALL call_1_9_5_8(SH2_struct* sh)
{
  SH2movls4(sh, 9,5, 8);
}
#define init_1_9_5_9
void FASTCALL call_1_9_5_9(SH2_struct* sh)
{
  SH2movls4(sh, 9,5, 9);
}
#define init_1_9_5_10
void FASTCALL call_1_9_5_10(SH2_struct* sh)
{
  SH2movls4(sh, 9,5, 10);
}
#define init_1_9_5_11
void FASTCALL call_1_9_5_11(SH2_struct* sh)
{
  SH2movls4(sh, 9,5, 11);
}
#define init_1_9_5_12
void FASTCALL call_1_9_5_12(SH2_struct* sh)
{
  SH2movls4(sh, 9,5, 12);
}
#define init_1_9_5_13
void FASTCALL call_1_9_5_13(SH2_struct* sh)
{
  SH2movls4(sh, 9,5, 13);
}
#define init_1_9_5_14
void FASTCALL call_1_9_5_14(SH2_struct* sh)
{
  SH2movls4(sh, 9,5, 14);
}
#define init_1_9_5_15
void FASTCALL call_1_9_5_15(SH2_struct* sh)
{
  SH2movls4(sh, 9,5, 15);
}
#define init_1_9_6_0
void FASTCALL call_1_9_6_0(SH2_struct* sh)
{
  SH2movls4(sh, 9,6, 0);
}
#define init_1_9_6_1
void FASTCALL call_1_9_6_1(SH2_struct* sh)
{
  SH2movls4(sh, 9,6, 1);
}
#define init_1_9_6_2
void FASTCALL call_1_9_6_2(SH2_struct* sh)
{
  SH2movls4(sh, 9,6, 2);
}
#define init_1_9_6_3
void FASTCALL call_1_9_6_3(SH2_struct* sh)
{
  SH2movls4(sh, 9,6, 3);
}
#define init_1_9_6_4
void FASTCALL call_1_9_6_4(SH2_struct* sh)
{
  SH2movls4(sh, 9,6, 4);
}
#define init_1_9_6_5
void FASTCALL call_1_9_6_5(SH2_struct* sh)
{
  SH2movls4(sh, 9,6, 5);
}
#define init_1_9_6_6
void FASTCALL call_1_9_6_6(SH2_struct* sh)
{
  SH2movls4(sh, 9,6, 6);
}
#define init_1_9_6_7
void FASTCALL call_1_9_6_7(SH2_struct* sh)
{
  SH2movls4(sh, 9,6, 7);
}
#define init_1_9_6_8
void FASTCALL call_1_9_6_8(SH2_struct* sh)
{
  SH2movls4(sh, 9,6, 8);
}
#define init_1_9_6_9
void FASTCALL call_1_9_6_9(SH2_struct* sh)
{
  SH2movls4(sh, 9,6, 9);
}
#define init_1_9_6_10
void FASTCALL call_1_9_6_10(SH2_struct* sh)
{
  SH2movls4(sh, 9,6, 10);
}
#define init_1_9_6_11
void FASTCALL call_1_9_6_11(SH2_struct* sh)
{
  SH2movls4(sh, 9,6, 11);
}
#define init_1_9_6_12
void FASTCALL call_1_9_6_12(SH2_struct* sh)
{
  SH2movls4(sh, 9,6, 12);
}
#define init_1_9_6_13
void FASTCALL call_1_9_6_13(SH2_struct* sh)
{
  SH2movls4(sh, 9,6, 13);
}
#define init_1_9_6_14
void FASTCALL call_1_9_6_14(SH2_struct* sh)
{
  SH2movls4(sh, 9,6, 14);
}
#define init_1_9_6_15
void FASTCALL call_1_9_6_15(SH2_struct* sh)
{
  SH2movls4(sh, 9,6, 15);
}
#define init_1_9_7_0
void FASTCALL call_1_9_7_0(SH2_struct* sh)
{
  SH2movls4(sh, 9,7, 0);
}
#define init_1_9_7_1
void FASTCALL call_1_9_7_1(SH2_struct* sh)
{
  SH2movls4(sh, 9,7, 1);
}
#define init_1_9_7_2
void FASTCALL call_1_9_7_2(SH2_struct* sh)
{
  SH2movls4(sh, 9,7, 2);
}
#define init_1_9_7_3
void FASTCALL call_1_9_7_3(SH2_struct* sh)
{
  SH2movls4(sh, 9,7, 3);
}
#define init_1_9_7_4
void FASTCALL call_1_9_7_4(SH2_struct* sh)
{
  SH2movls4(sh, 9,7, 4);
}
#define init_1_9_7_5
void FASTCALL call_1_9_7_5(SH2_struct* sh)
{
  SH2movls4(sh, 9,7, 5);
}
#define init_1_9_7_6
void FASTCALL call_1_9_7_6(SH2_struct* sh)
{
  SH2movls4(sh, 9,7, 6);
}
#define init_1_9_7_7
void FASTCALL call_1_9_7_7(SH2_struct* sh)
{
  SH2movls4(sh, 9,7, 7);
}
#define init_1_9_7_8
void FASTCALL call_1_9_7_8(SH2_struct* sh)
{
  SH2movls4(sh, 9,7, 8);
}
#define init_1_9_7_9
void FASTCALL call_1_9_7_9(SH2_struct* sh)
{
  SH2movls4(sh, 9,7, 9);
}
#define init_1_9_7_10
void FASTCALL call_1_9_7_10(SH2_struct* sh)
{
  SH2movls4(sh, 9,7, 10);
}
#define init_1_9_7_11
void FASTCALL call_1_9_7_11(SH2_struct* sh)
{
  SH2movls4(sh, 9,7, 11);
}
#define init_1_9_7_12
void FASTCALL call_1_9_7_12(SH2_struct* sh)
{
  SH2movls4(sh, 9,7, 12);
}
#define init_1_9_7_13
void FASTCALL call_1_9_7_13(SH2_struct* sh)
{
  SH2movls4(sh, 9,7, 13);
}
#define init_1_9_7_14
void FASTCALL call_1_9_7_14(SH2_struct* sh)
{
  SH2movls4(sh, 9,7, 14);
}
#define init_1_9_7_15
void FASTCALL call_1_9_7_15(SH2_struct* sh)
{
  SH2movls4(sh, 9,7, 15);
}
#define init_1_9_8_0
void FASTCALL call_1_9_8_0(SH2_struct* sh)
{
  SH2movls4(sh, 9,8, 0);
}
#define init_1_9_8_1
void FASTCALL call_1_9_8_1(SH2_struct* sh)
{
  SH2movls4(sh, 9,8, 1);
}
#define init_1_9_8_2
void FASTCALL call_1_9_8_2(SH2_struct* sh)
{
  SH2movls4(sh, 9,8, 2);
}
#define init_1_9_8_3
void FASTCALL call_1_9_8_3(SH2_struct* sh)
{
  SH2movls4(sh, 9,8, 3);
}
#define init_1_9_8_4
void FASTCALL call_1_9_8_4(SH2_struct* sh)
{
  SH2movls4(sh, 9,8, 4);
}
#define init_1_9_8_5
void FASTCALL call_1_9_8_5(SH2_struct* sh)
{
  SH2movls4(sh, 9,8, 5);
}
#define init_1_9_8_6
void FASTCALL call_1_9_8_6(SH2_struct* sh)
{
  SH2movls4(sh, 9,8, 6);
}
#define init_1_9_8_7
void FASTCALL call_1_9_8_7(SH2_struct* sh)
{
  SH2movls4(sh, 9,8, 7);
}
#define init_1_9_8_8
void FASTCALL call_1_9_8_8(SH2_struct* sh)
{
  SH2movls4(sh, 9,8, 8);
}
#define init_1_9_8_9
void FASTCALL call_1_9_8_9(SH2_struct* sh)
{
  SH2movls4(sh, 9,8, 9);
}
#define init_1_9_8_10
void FASTCALL call_1_9_8_10(SH2_struct* sh)
{
  SH2movls4(sh, 9,8, 10);
}
#define init_1_9_8_11
void FASTCALL call_1_9_8_11(SH2_struct* sh)
{
  SH2movls4(sh, 9,8, 11);
}
#define init_1_9_8_12
void FASTCALL call_1_9_8_12(SH2_struct* sh)
{
  SH2movls4(sh, 9,8, 12);
}
#define init_1_9_8_13
void FASTCALL call_1_9_8_13(SH2_struct* sh)
{
  SH2movls4(sh, 9,8, 13);
}
#define init_1_9_8_14
void FASTCALL call_1_9_8_14(SH2_struct* sh)
{
  SH2movls4(sh, 9,8, 14);
}
#define init_1_9_8_15
void FASTCALL call_1_9_8_15(SH2_struct* sh)
{
  SH2movls4(sh, 9,8, 15);
}
#define init_1_9_9_0
void FASTCALL call_1_9_9_0(SH2_struct* sh)
{
  SH2movls4(sh, 9,9, 0);
}
#define init_1_9_9_1
void FASTCALL call_1_9_9_1(SH2_struct* sh)
{
  SH2movls4(sh, 9,9, 1);
}
#define init_1_9_9_2
void FASTCALL call_1_9_9_2(SH2_struct* sh)
{
  SH2movls4(sh, 9,9, 2);
}
#define init_1_9_9_3
void FASTCALL call_1_9_9_3(SH2_struct* sh)
{
  SH2movls4(sh, 9,9, 3);
}
#define init_1_9_9_4
void FASTCALL call_1_9_9_4(SH2_struct* sh)
{
  SH2movls4(sh, 9,9, 4);
}
#define init_1_9_9_5
void FASTCALL call_1_9_9_5(SH2_struct* sh)
{
  SH2movls4(sh, 9,9, 5);
}
#define init_1_9_9_6
void FASTCALL call_1_9_9_6(SH2_struct* sh)
{
  SH2movls4(sh, 9,9, 6);
}
#define init_1_9_9_7
void FASTCALL call_1_9_9_7(SH2_struct* sh)
{
  SH2movls4(sh, 9,9, 7);
}
#define init_1_9_9_8
void FASTCALL call_1_9_9_8(SH2_struct* sh)
{
  SH2movls4(sh, 9,9, 8);
}
#define init_1_9_9_9
void FASTCALL call_1_9_9_9(SH2_struct* sh)
{
  SH2movls4(sh, 9,9, 9);
}
#define init_1_9_9_10
void FASTCALL call_1_9_9_10(SH2_struct* sh)
{
  SH2movls4(sh, 9,9, 10);
}
#define init_1_9_9_11
void FASTCALL call_1_9_9_11(SH2_struct* sh)
{
  SH2movls4(sh, 9,9, 11);
}
#define init_1_9_9_12
void FASTCALL call_1_9_9_12(SH2_struct* sh)
{
  SH2movls4(sh, 9,9, 12);
}
#define init_1_9_9_13
void FASTCALL call_1_9_9_13(SH2_struct* sh)
{
  SH2movls4(sh, 9,9, 13);
}
#define init_1_9_9_14
void FASTCALL call_1_9_9_14(SH2_struct* sh)
{
  SH2movls4(sh, 9,9, 14);
}
#define init_1_9_9_15
void FASTCALL call_1_9_9_15(SH2_struct* sh)
{
  SH2movls4(sh, 9,9, 15);
}
#define init_1_9_10_0
void FASTCALL call_1_9_10_0(SH2_struct* sh)
{
  SH2movls4(sh, 9,10, 0);
}
#define init_1_9_10_1
void FASTCALL call_1_9_10_1(SH2_struct* sh)
{
  SH2movls4(sh, 9,10, 1);
}
#define init_1_9_10_2
void FASTCALL call_1_9_10_2(SH2_struct* sh)
{
  SH2movls4(sh, 9,10, 2);
}
#define init_1_9_10_3
void FASTCALL call_1_9_10_3(SH2_struct* sh)
{
  SH2movls4(sh, 9,10, 3);
}
#define init_1_9_10_4
void FASTCALL call_1_9_10_4(SH2_struct* sh)
{
  SH2movls4(sh, 9,10, 4);
}
#define init_1_9_10_5
void FASTCALL call_1_9_10_5(SH2_struct* sh)
{
  SH2movls4(sh, 9,10, 5);
}
#define init_1_9_10_6
void FASTCALL call_1_9_10_6(SH2_struct* sh)
{
  SH2movls4(sh, 9,10, 6);
}
#define init_1_9_10_7
void FASTCALL call_1_9_10_7(SH2_struct* sh)
{
  SH2movls4(sh, 9,10, 7);
}
#define init_1_9_10_8
void FASTCALL call_1_9_10_8(SH2_struct* sh)
{
  SH2movls4(sh, 9,10, 8);
}
#define init_1_9_10_9
void FASTCALL call_1_9_10_9(SH2_struct* sh)
{
  SH2movls4(sh, 9,10, 9);
}
#define init_1_9_10_10
void FASTCALL call_1_9_10_10(SH2_struct* sh)
{
  SH2movls4(sh, 9,10, 10);
}
#define init_1_9_10_11
void FASTCALL call_1_9_10_11(SH2_struct* sh)
{
  SH2movls4(sh, 9,10, 11);
}
#define init_1_9_10_12
void FASTCALL call_1_9_10_12(SH2_struct* sh)
{
  SH2movls4(sh, 9,10, 12);
}
#define init_1_9_10_13
void FASTCALL call_1_9_10_13(SH2_struct* sh)
{
  SH2movls4(sh, 9,10, 13);
}
#define init_1_9_10_14
void FASTCALL call_1_9_10_14(SH2_struct* sh)
{
  SH2movls4(sh, 9,10, 14);
}
#define init_1_9_10_15
void FASTCALL call_1_9_10_15(SH2_struct* sh)
{
  SH2movls4(sh, 9,10, 15);
}
#define init_1_9_11_0
void FASTCALL call_1_9_11_0(SH2_struct* sh)
{
  SH2movls4(sh, 9,11, 0);
}
#define init_1_9_11_1
void FASTCALL call_1_9_11_1(SH2_struct* sh)
{
  SH2movls4(sh, 9,11, 1);
}
#define init_1_9_11_2
void FASTCALL call_1_9_11_2(SH2_struct* sh)
{
  SH2movls4(sh, 9,11, 2);
}
#define init_1_9_11_3
void FASTCALL call_1_9_11_3(SH2_struct* sh)
{
  SH2movls4(sh, 9,11, 3);
}
#define init_1_9_11_4
void FASTCALL call_1_9_11_4(SH2_struct* sh)
{
  SH2movls4(sh, 9,11, 4);
}
#define init_1_9_11_5
void FASTCALL call_1_9_11_5(SH2_struct* sh)
{
  SH2movls4(sh, 9,11, 5);
}
#define init_1_9_11_6
void FASTCALL call_1_9_11_6(SH2_struct* sh)
{
  SH2movls4(sh, 9,11, 6);
}
#define init_1_9_11_7
void FASTCALL call_1_9_11_7(SH2_struct* sh)
{
  SH2movls4(sh, 9,11, 7);
}
#define init_1_9_11_8
void FASTCALL call_1_9_11_8(SH2_struct* sh)
{
  SH2movls4(sh, 9,11, 8);
}
#define init_1_9_11_9
void FASTCALL call_1_9_11_9(SH2_struct* sh)
{
  SH2movls4(sh, 9,11, 9);
}
#define init_1_9_11_10
void FASTCALL call_1_9_11_10(SH2_struct* sh)
{
  SH2movls4(sh, 9,11, 10);
}
#define init_1_9_11_11
void FASTCALL call_1_9_11_11(SH2_struct* sh)
{
  SH2movls4(sh, 9,11, 11);
}
#define init_1_9_11_12
void FASTCALL call_1_9_11_12(SH2_struct* sh)
{
  SH2movls4(sh, 9,11, 12);
}
#define init_1_9_11_13
void FASTCALL call_1_9_11_13(SH2_struct* sh)
{
  SH2movls4(sh, 9,11, 13);
}
#define init_1_9_11_14
void FASTCALL call_1_9_11_14(SH2_struct* sh)
{
  SH2movls4(sh, 9,11, 14);
}
#define init_1_9_11_15
void FASTCALL call_1_9_11_15(SH2_struct* sh)
{
  SH2movls4(sh, 9,11, 15);
}
#define init_1_9_12_0
void FASTCALL call_1_9_12_0(SH2_struct* sh)
{
  SH2movls4(sh, 9,12, 0);
}
#define init_1_9_12_1
void FASTCALL call_1_9_12_1(SH2_struct* sh)
{
  SH2movls4(sh, 9,12, 1);
}
#define init_1_9_12_2
void FASTCALL call_1_9_12_2(SH2_struct* sh)
{
  SH2movls4(sh, 9,12, 2);
}
#define init_1_9_12_3
void FASTCALL call_1_9_12_3(SH2_struct* sh)
{
  SH2movls4(sh, 9,12, 3);
}
#define init_1_9_12_4
void FASTCALL call_1_9_12_4(SH2_struct* sh)
{
  SH2movls4(sh, 9,12, 4);
}
#define init_1_9_12_5
void FASTCALL call_1_9_12_5(SH2_struct* sh)
{
  SH2movls4(sh, 9,12, 5);
}
#define init_1_9_12_6
void FASTCALL call_1_9_12_6(SH2_struct* sh)
{
  SH2movls4(sh, 9,12, 6);
}
#define init_1_9_12_7
void FASTCALL call_1_9_12_7(SH2_struct* sh)
{
  SH2movls4(sh, 9,12, 7);
}
#define init_1_9_12_8
void FASTCALL call_1_9_12_8(SH2_struct* sh)
{
  SH2movls4(sh, 9,12, 8);
}
#define init_1_9_12_9
void FASTCALL call_1_9_12_9(SH2_struct* sh)
{
  SH2movls4(sh, 9,12, 9);
}
#define init_1_9_12_10
void FASTCALL call_1_9_12_10(SH2_struct* sh)
{
  SH2movls4(sh, 9,12, 10);
}
#define init_1_9_12_11
void FASTCALL call_1_9_12_11(SH2_struct* sh)
{
  SH2movls4(sh, 9,12, 11);
}
#define init_1_9_12_12
void FASTCALL call_1_9_12_12(SH2_struct* sh)
{
  SH2movls4(sh, 9,12, 12);
}
#define init_1_9_12_13
void FASTCALL call_1_9_12_13(SH2_struct* sh)
{
  SH2movls4(sh, 9,12, 13);
}
#define init_1_9_12_14
void FASTCALL call_1_9_12_14(SH2_struct* sh)
{
  SH2movls4(sh, 9,12, 14);
}
#define init_1_9_12_15
void FASTCALL call_1_9_12_15(SH2_struct* sh)
{
  SH2movls4(sh, 9,12, 15);
}
#define init_1_9_13_0
void FASTCALL call_1_9_13_0(SH2_struct* sh)
{
  SH2movls4(sh, 9,13, 0);
}
#define init_1_9_13_1
void FASTCALL call_1_9_13_1(SH2_struct* sh)
{
  SH2movls4(sh, 9,13, 1);
}
#define init_1_9_13_2
void FASTCALL call_1_9_13_2(SH2_struct* sh)
{
  SH2movls4(sh, 9,13, 2);
}
#define init_1_9_13_3
void FASTCALL call_1_9_13_3(SH2_struct* sh)
{
  SH2movls4(sh, 9,13, 3);
}
#define init_1_9_13_4
void FASTCALL call_1_9_13_4(SH2_struct* sh)
{
  SH2movls4(sh, 9,13, 4);
}
#define init_1_9_13_5
void FASTCALL call_1_9_13_5(SH2_struct* sh)
{
  SH2movls4(sh, 9,13, 5);
}
#define init_1_9_13_6
void FASTCALL call_1_9_13_6(SH2_struct* sh)
{
  SH2movls4(sh, 9,13, 6);
}
#define init_1_9_13_7
void FASTCALL call_1_9_13_7(SH2_struct* sh)
{
  SH2movls4(sh, 9,13, 7);
}
#define init_1_9_13_8
void FASTCALL call_1_9_13_8(SH2_struct* sh)
{
  SH2movls4(sh, 9,13, 8);
}
#define init_1_9_13_9
void FASTCALL call_1_9_13_9(SH2_struct* sh)
{
  SH2movls4(sh, 9,13, 9);
}
#define init_1_9_13_10
void FASTCALL call_1_9_13_10(SH2_struct* sh)
{
  SH2movls4(sh, 9,13, 10);
}
#define init_1_9_13_11
void FASTCALL call_1_9_13_11(SH2_struct* sh)
{
  SH2movls4(sh, 9,13, 11);
}
#define init_1_9_13_12
void FASTCALL call_1_9_13_12(SH2_struct* sh)
{
  SH2movls4(sh, 9,13, 12);
}
#define init_1_9_13_13
void FASTCALL call_1_9_13_13(SH2_struct* sh)
{
  SH2movls4(sh, 9,13, 13);
}
#define init_1_9_13_14
void FASTCALL call_1_9_13_14(SH2_struct* sh)
{
  SH2movls4(sh, 9,13, 14);
}
#define init_1_9_13_15
void FASTCALL call_1_9_13_15(SH2_struct* sh)
{
  SH2movls4(sh, 9,13, 15);
}
#define init_1_9_14_0
void FASTCALL call_1_9_14_0(SH2_struct* sh)
{
  SH2movls4(sh, 9,14, 0);
}
#define init_1_9_14_1
void FASTCALL call_1_9_14_1(SH2_struct* sh)
{
  SH2movls4(sh, 9,14, 1);
}
#define init_1_9_14_2
void FASTCALL call_1_9_14_2(SH2_struct* sh)
{
  SH2movls4(sh, 9,14, 2);
}
#define init_1_9_14_3
void FASTCALL call_1_9_14_3(SH2_struct* sh)
{
  SH2movls4(sh, 9,14, 3);
}
#define init_1_9_14_4
void FASTCALL call_1_9_14_4(SH2_struct* sh)
{
  SH2movls4(sh, 9,14, 4);
}
#define init_1_9_14_5
void FASTCALL call_1_9_14_5(SH2_struct* sh)
{
  SH2movls4(sh, 9,14, 5);
}
#define init_1_9_14_6
void FASTCALL call_1_9_14_6(SH2_struct* sh)
{
  SH2movls4(sh, 9,14, 6);
}
#define init_1_9_14_7
void FASTCALL call_1_9_14_7(SH2_struct* sh)
{
  SH2movls4(sh, 9,14, 7);
}
#define init_1_9_14_8
void FASTCALL call_1_9_14_8(SH2_struct* sh)
{
  SH2movls4(sh, 9,14, 8);
}
#define init_1_9_14_9
void FASTCALL call_1_9_14_9(SH2_struct* sh)
{
  SH2movls4(sh, 9,14, 9);
}
#define init_1_9_14_10
void FASTCALL call_1_9_14_10(SH2_struct* sh)
{
  SH2movls4(sh, 9,14, 10);
}
#define init_1_9_14_11
void FASTCALL call_1_9_14_11(SH2_struct* sh)
{
  SH2movls4(sh, 9,14, 11);
}
#define init_1_9_14_12
void FASTCALL call_1_9_14_12(SH2_struct* sh)
{
  SH2movls4(sh, 9,14, 12);
}
#define init_1_9_14_13
void FASTCALL call_1_9_14_13(SH2_struct* sh)
{
  SH2movls4(sh, 9,14, 13);
}
#define init_1_9_14_14
void FASTCALL call_1_9_14_14(SH2_struct* sh)
{
  SH2movls4(sh, 9,14, 14);
}
#define init_1_9_14_15
void FASTCALL call_1_9_14_15(SH2_struct* sh)
{
  SH2movls4(sh, 9,14, 15);
}
#define init_1_9_15_0
void FASTCALL call_1_9_15_0(SH2_struct* sh)
{
  SH2movls4(sh, 9,15, 0);
}
#define init_1_9_15_1
void FASTCALL call_1_9_15_1(SH2_struct* sh)
{
  SH2movls4(sh, 9,15, 1);
}
#define init_1_9_15_2
void FASTCALL call_1_9_15_2(SH2_struct* sh)
{
  SH2movls4(sh, 9,15, 2);
}
#define init_1_9_15_3
void FASTCALL call_1_9_15_3(SH2_struct* sh)
{
  SH2movls4(sh, 9,15, 3);
}
#define init_1_9_15_4
void FASTCALL call_1_9_15_4(SH2_struct* sh)
{
  SH2movls4(sh, 9,15, 4);
}
#define init_1_9_15_5
void FASTCALL call_1_9_15_5(SH2_struct* sh)
{
  SH2movls4(sh, 9,15, 5);
}
#define init_1_9_15_6
void FASTCALL call_1_9_15_6(SH2_struct* sh)
{
  SH2movls4(sh, 9,15, 6);
}
#define init_1_9_15_7
void FASTCALL call_1_9_15_7(SH2_struct* sh)
{
  SH2movls4(sh, 9,15, 7);
}
#define init_1_9_15_8
void FASTCALL call_1_9_15_8(SH2_struct* sh)
{
  SH2movls4(sh, 9,15, 8);
}
#define init_1_9_15_9
void FASTCALL call_1_9_15_9(SH2_struct* sh)
{
  SH2movls4(sh, 9,15, 9);
}
#define init_1_9_15_10
void FASTCALL call_1_9_15_10(SH2_struct* sh)
{
  SH2movls4(sh, 9,15, 10);
}
#define init_1_9_15_11
void FASTCALL call_1_9_15_11(SH2_struct* sh)
{
  SH2movls4(sh, 9,15, 11);
}
#define init_1_9_15_12
void FASTCALL call_1_9_15_12(SH2_struct* sh)
{
  SH2movls4(sh, 9,15, 12);
}
#define init_1_9_15_13
void FASTCALL call_1_9_15_13(SH2_struct* sh)
{
  SH2movls4(sh, 9,15, 13);
}
#define init_1_9_15_14
void FASTCALL call_1_9_15_14(SH2_struct* sh)
{
  SH2movls4(sh, 9,15, 14);
}
#define init_1_9_15_15
void FASTCALL call_1_9_15_15(SH2_struct* sh)
{
  SH2movls4(sh, 9,15, 15);
}
#define init_1_10_0_0
void FASTCALL call_1_10_0_0(SH2_struct* sh)
{
  SH2movls4(sh, 10,0, 0);
}
#define init_1_10_0_1
void FASTCALL call_1_10_0_1(SH2_struct* sh)
{
  SH2movls4(sh, 10,0, 1);
}
#define init_1_10_0_2
void FASTCALL call_1_10_0_2(SH2_struct* sh)
{
  SH2movls4(sh, 10,0, 2);
}
#define init_1_10_0_3
void FASTCALL call_1_10_0_3(SH2_struct* sh)
{
  SH2movls4(sh, 10,0, 3);
}
#define init_1_10_0_4
void FASTCALL call_1_10_0_4(SH2_struct* sh)
{
  SH2movls4(sh, 10,0, 4);
}
#define init_1_10_0_5
void FASTCALL call_1_10_0_5(SH2_struct* sh)
{
  SH2movls4(sh, 10,0, 5);
}
#define init_1_10_0_6
void FASTCALL call_1_10_0_6(SH2_struct* sh)
{
  SH2movls4(sh, 10,0, 6);
}
#define init_1_10_0_7
void FASTCALL call_1_10_0_7(SH2_struct* sh)
{
  SH2movls4(sh, 10,0, 7);
}
#define init_1_10_0_8
void FASTCALL call_1_10_0_8(SH2_struct* sh)
{
  SH2movls4(sh, 10,0, 8);
}
#define init_1_10_0_9
void FASTCALL call_1_10_0_9(SH2_struct* sh)
{
  SH2movls4(sh, 10,0, 9);
}
#define init_1_10_0_10
void FASTCALL call_1_10_0_10(SH2_struct* sh)
{
  SH2movls4(sh, 10,0, 10);
}
#define init_1_10_0_11
void FASTCALL call_1_10_0_11(SH2_struct* sh)
{
  SH2movls4(sh, 10,0, 11);
}
#define init_1_10_0_12
void FASTCALL call_1_10_0_12(SH2_struct* sh)
{
  SH2movls4(sh, 10,0, 12);
}
#define init_1_10_0_13
void FASTCALL call_1_10_0_13(SH2_struct* sh)
{
  SH2movls4(sh, 10,0, 13);
}
#define init_1_10_0_14
void FASTCALL call_1_10_0_14(SH2_struct* sh)
{
  SH2movls4(sh, 10,0, 14);
}
#define init_1_10_0_15
void FASTCALL call_1_10_0_15(SH2_struct* sh)
{
  SH2movls4(sh, 10,0, 15);
}
#define init_1_10_1_0
void FASTCALL call_1_10_1_0(SH2_struct* sh)
{
  SH2movls4(sh, 10,1, 0);
}
#define init_1_10_1_1
void FASTCALL call_1_10_1_1(SH2_struct* sh)
{
  SH2movls4(sh, 10,1, 1);
}
#define init_1_10_1_2
void FASTCALL call_1_10_1_2(SH2_struct* sh)
{
  SH2movls4(sh, 10,1, 2);
}
#define init_1_10_1_3
void FASTCALL call_1_10_1_3(SH2_struct* sh)
{
  SH2movls4(sh, 10,1, 3);
}
#define init_1_10_1_4
void FASTCALL call_1_10_1_4(SH2_struct* sh)
{
  SH2movls4(sh, 10,1, 4);
}
#define init_1_10_1_5
void FASTCALL call_1_10_1_5(SH2_struct* sh)
{
  SH2movls4(sh, 10,1, 5);
}
#define init_1_10_1_6
void FASTCALL call_1_10_1_6(SH2_struct* sh)
{
  SH2movls4(sh, 10,1, 6);
}
#define init_1_10_1_7
void FASTCALL call_1_10_1_7(SH2_struct* sh)
{
  SH2movls4(sh, 10,1, 7);
}
#define init_1_10_1_8
void FASTCALL call_1_10_1_8(SH2_struct* sh)
{
  SH2movls4(sh, 10,1, 8);
}
#define init_1_10_1_9
void FASTCALL call_1_10_1_9(SH2_struct* sh)
{
  SH2movls4(sh, 10,1, 9);
}
#define init_1_10_1_10
void FASTCALL call_1_10_1_10(SH2_struct* sh)
{
  SH2movls4(sh, 10,1, 10);
}
#define init_1_10_1_11
void FASTCALL call_1_10_1_11(SH2_struct* sh)
{
  SH2movls4(sh, 10,1, 11);
}
#define init_1_10_1_12
void FASTCALL call_1_10_1_12(SH2_struct* sh)
{
  SH2movls4(sh, 10,1, 12);
}
#define init_1_10_1_13
void FASTCALL call_1_10_1_13(SH2_struct* sh)
{
  SH2movls4(sh, 10,1, 13);
}
#define init_1_10_1_14
void FASTCALL call_1_10_1_14(SH2_struct* sh)
{
  SH2movls4(sh, 10,1, 14);
}
#define init_1_10_1_15
void FASTCALL call_1_10_1_15(SH2_struct* sh)
{
  SH2movls4(sh, 10,1, 15);
}
#define init_1_10_2_0
void FASTCALL call_1_10_2_0(SH2_struct* sh)
{
  SH2movls4(sh, 10,2, 0);
}
#define init_1_10_2_1
void FASTCALL call_1_10_2_1(SH2_struct* sh)
{
  SH2movls4(sh, 10,2, 1);
}
#define init_1_10_2_2
void FASTCALL call_1_10_2_2(SH2_struct* sh)
{
  SH2movls4(sh, 10,2, 2);
}
#define init_1_10_2_3
void FASTCALL call_1_10_2_3(SH2_struct* sh)
{
  SH2movls4(sh, 10,2, 3);
}
#define init_1_10_2_4
void FASTCALL call_1_10_2_4(SH2_struct* sh)
{
  SH2movls4(sh, 10,2, 4);
}
#define init_1_10_2_5
void FASTCALL call_1_10_2_5(SH2_struct* sh)
{
  SH2movls4(sh, 10,2, 5);
}
#define init_1_10_2_6
void FASTCALL call_1_10_2_6(SH2_struct* sh)
{
  SH2movls4(sh, 10,2, 6);
}
#define init_1_10_2_7
void FASTCALL call_1_10_2_7(SH2_struct* sh)
{
  SH2movls4(sh, 10,2, 7);
}
#define init_1_10_2_8
void FASTCALL call_1_10_2_8(SH2_struct* sh)
{
  SH2movls4(sh, 10,2, 8);
}
#define init_1_10_2_9
void FASTCALL call_1_10_2_9(SH2_struct* sh)
{
  SH2movls4(sh, 10,2, 9);
}
#define init_1_10_2_10
void FASTCALL call_1_10_2_10(SH2_struct* sh)
{
  SH2movls4(sh, 10,2, 10);
}
#define init_1_10_2_11
void FASTCALL call_1_10_2_11(SH2_struct* sh)
{
  SH2movls4(sh, 10,2, 11);
}
#define init_1_10_2_12
void FASTCALL call_1_10_2_12(SH2_struct* sh)
{
  SH2movls4(sh, 10,2, 12);
}
#define init_1_10_2_13
void FASTCALL call_1_10_2_13(SH2_struct* sh)
{
  SH2movls4(sh, 10,2, 13);
}
#define init_1_10_2_14
void FASTCALL call_1_10_2_14(SH2_struct* sh)
{
  SH2movls4(sh, 10,2, 14);
}
#define init_1_10_2_15
void FASTCALL call_1_10_2_15(SH2_struct* sh)
{
  SH2movls4(sh, 10,2, 15);
}
#define init_1_10_3_0
void FASTCALL call_1_10_3_0(SH2_struct* sh)
{
  SH2movls4(sh, 10,3, 0);
}
#define init_1_10_3_1
void FASTCALL call_1_10_3_1(SH2_struct* sh)
{
  SH2movls4(sh, 10,3, 1);
}
#define init_1_10_3_2
void FASTCALL call_1_10_3_2(SH2_struct* sh)
{
  SH2movls4(sh, 10,3, 2);
}
#define init_1_10_3_3
void FASTCALL call_1_10_3_3(SH2_struct* sh)
{
  SH2movls4(sh, 10,3, 3);
}
#define init_1_10_3_4
void FASTCALL call_1_10_3_4(SH2_struct* sh)
{
  SH2movls4(sh, 10,3, 4);
}
#define init_1_10_3_5
void FASTCALL call_1_10_3_5(SH2_struct* sh)
{
  SH2movls4(sh, 10,3, 5);
}
#define init_1_10_3_6
void FASTCALL call_1_10_3_6(SH2_struct* sh)
{
  SH2movls4(sh, 10,3, 6);
}
#define init_1_10_3_7
void FASTCALL call_1_10_3_7(SH2_struct* sh)
{
  SH2movls4(sh, 10,3, 7);
}
#define init_1_10_3_8
void FASTCALL call_1_10_3_8(SH2_struct* sh)
{
  SH2movls4(sh, 10,3, 8);
}
#define init_1_10_3_9
void FASTCALL call_1_10_3_9(SH2_struct* sh)
{
  SH2movls4(sh, 10,3, 9);
}
#define init_1_10_3_10
void FASTCALL call_1_10_3_10(SH2_struct* sh)
{
  SH2movls4(sh, 10,3, 10);
}
#define init_1_10_3_11
void FASTCALL call_1_10_3_11(SH2_struct* sh)
{
  SH2movls4(sh, 10,3, 11);
}
#define init_1_10_3_12
void FASTCALL call_1_10_3_12(SH2_struct* sh)
{
  SH2movls4(sh, 10,3, 12);
}
#define init_1_10_3_13
void FASTCALL call_1_10_3_13(SH2_struct* sh)
{
  SH2movls4(sh, 10,3, 13);
}
#define init_1_10_3_14
void FASTCALL call_1_10_3_14(SH2_struct* sh)
{
  SH2movls4(sh, 10,3, 14);
}
#define init_1_10_3_15
void FASTCALL call_1_10_3_15(SH2_struct* sh)
{
  SH2movls4(sh, 10,3, 15);
}
#define init_1_10_4_0
void FASTCALL call_1_10_4_0(SH2_struct* sh)
{
  SH2movls4(sh, 10,4, 0);
}
#define init_1_10_4_1
void FASTCALL call_1_10_4_1(SH2_struct* sh)
{
  SH2movls4(sh, 10,4, 1);
}
#define init_1_10_4_2
void FASTCALL call_1_10_4_2(SH2_struct* sh)
{
  SH2movls4(sh, 10,4, 2);
}
#define init_1_10_4_3
void FASTCALL call_1_10_4_3(SH2_struct* sh)
{
  SH2movls4(sh, 10,4, 3);
}
#define init_1_10_4_4
void FASTCALL call_1_10_4_4(SH2_struct* sh)
{
  SH2movls4(sh, 10,4, 4);
}
#define init_1_10_4_5
void FASTCALL call_1_10_4_5(SH2_struct* sh)
{
  SH2movls4(sh, 10,4, 5);
}
#define init_1_10_4_6
void FASTCALL call_1_10_4_6(SH2_struct* sh)
{
  SH2movls4(sh, 10,4, 6);
}
#define init_1_10_4_7
void FASTCALL call_1_10_4_7(SH2_struct* sh)
{
  SH2movls4(sh, 10,4, 7);
}
#define init_1_10_4_8
void FASTCALL call_1_10_4_8(SH2_struct* sh)
{
  SH2movls4(sh, 10,4, 8);
}
#define init_1_10_4_9
void FASTCALL call_1_10_4_9(SH2_struct* sh)
{
  SH2movls4(sh, 10,4, 9);
}
#define init_1_10_4_10
void FASTCALL call_1_10_4_10(SH2_struct* sh)
{
  SH2movls4(sh, 10,4, 10);
}
#define init_1_10_4_11
void FASTCALL call_1_10_4_11(SH2_struct* sh)
{
  SH2movls4(sh, 10,4, 11);
}
#define init_1_10_4_12
void FASTCALL call_1_10_4_12(SH2_struct* sh)
{
  SH2movls4(sh, 10,4, 12);
}
#define init_1_10_4_13
void FASTCALL call_1_10_4_13(SH2_struct* sh)
{
  SH2movls4(sh, 10,4, 13);
}
#define init_1_10_4_14
void FASTCALL call_1_10_4_14(SH2_struct* sh)
{
  SH2movls4(sh, 10,4, 14);
}
#define init_1_10_4_15
void FASTCALL call_1_10_4_15(SH2_struct* sh)
{
  SH2movls4(sh, 10,4, 15);
}
#define init_1_10_5_0
void FASTCALL call_1_10_5_0(SH2_struct* sh)
{
  SH2movls4(sh, 10,5, 0);
}
#define init_1_10_5_1
void FASTCALL call_1_10_5_1(SH2_struct* sh)
{
  SH2movls4(sh, 10,5, 1);
}
#define init_1_10_5_2
void FASTCALL call_1_10_5_2(SH2_struct* sh)
{
  SH2movls4(sh, 10,5, 2);
}
#define init_1_10_5_3
void FASTCALL call_1_10_5_3(SH2_struct* sh)
{
  SH2movls4(sh, 10,5, 3);
}
#define init_1_10_5_4
void FASTCALL call_1_10_5_4(SH2_struct* sh)
{
  SH2movls4(sh, 10,5, 4);
}
#define init_1_10_5_5
void FASTCALL call_1_10_5_5(SH2_struct* sh)
{
  SH2movls4(sh, 10,5, 5);
}
#define init_1_10_5_6
void FASTCALL call_1_10_5_6(SH2_struct* sh)
{
  SH2movls4(sh, 10,5, 6);
}
#define init_1_10_5_7
void FASTCALL call_1_10_5_7(SH2_struct* sh)
{
  SH2movls4(sh, 10,5, 7);
}
#define init_1_10_5_8
void FASTCALL call_1_10_5_8(SH2_struct* sh)
{
  SH2movls4(sh, 10,5, 8);
}
#define init_1_10_5_9
void FASTCALL call_1_10_5_9(SH2_struct* sh)
{
  SH2movls4(sh, 10,5, 9);
}
#define init_1_10_5_10
void FASTCALL call_1_10_5_10(SH2_struct* sh)
{
  SH2movls4(sh, 10,5, 10);
}
#define init_1_10_5_11
void FASTCALL call_1_10_5_11(SH2_struct* sh)
{
  SH2movls4(sh, 10,5, 11);
}
#define init_1_10_5_12
void FASTCALL call_1_10_5_12(SH2_struct* sh)
{
  SH2movls4(sh, 10,5, 12);
}
#define init_1_10_5_13
void FASTCALL call_1_10_5_13(SH2_struct* sh)
{
  SH2movls4(sh, 10,5, 13);
}
#define init_1_10_5_14
void FASTCALL call_1_10_5_14(SH2_struct* sh)
{
  SH2movls4(sh, 10,5, 14);
}
#define init_1_10_5_15
void FASTCALL call_1_10_5_15(SH2_struct* sh)
{
  SH2movls4(sh, 10,5, 15);
}
#define init_1_10_6_0
void FASTCALL call_1_10_6_0(SH2_struct* sh)
{
  SH2movls4(sh, 10,6, 0);
}
#define init_1_10_6_1
void FASTCALL call_1_10_6_1(SH2_struct* sh)
{
  SH2movls4(sh, 10,6, 1);
}
#define init_1_10_6_2
void FASTCALL call_1_10_6_2(SH2_struct* sh)
{
  SH2movls4(sh, 10,6, 2);
}
#define init_1_10_6_3
void FASTCALL call_1_10_6_3(SH2_struct* sh)
{
  SH2movls4(sh, 10,6, 3);
}
#define init_1_10_6_4
void FASTCALL call_1_10_6_4(SH2_struct* sh)
{
  SH2movls4(sh, 10,6, 4);
}
#define init_1_10_6_5
void FASTCALL call_1_10_6_5(SH2_struct* sh)
{
  SH2movls4(sh, 10,6, 5);
}
#define init_1_10_6_6
void FASTCALL call_1_10_6_6(SH2_struct* sh)
{
  SH2movls4(sh, 10,6, 6);
}
#define init_1_10_6_7
void FASTCALL call_1_10_6_7(SH2_struct* sh)
{
  SH2movls4(sh, 10,6, 7);
}
#define init_1_10_6_8
void FASTCALL call_1_10_6_8(SH2_struct* sh)
{
  SH2movls4(sh, 10,6, 8);
}
#define init_1_10_6_9
void FASTCALL call_1_10_6_9(SH2_struct* sh)
{
  SH2movls4(sh, 10,6, 9);
}
#define init_1_10_6_10
void FASTCALL call_1_10_6_10(SH2_struct* sh)
{
  SH2movls4(sh, 10,6, 10);
}
#define init_1_10_6_11
void FASTCALL call_1_10_6_11(SH2_struct* sh)
{
  SH2movls4(sh, 10,6, 11);
}
#define init_1_10_6_12
void FASTCALL call_1_10_6_12(SH2_struct* sh)
{
  SH2movls4(sh, 10,6, 12);
}
#define init_1_10_6_13
void FASTCALL call_1_10_6_13(SH2_struct* sh)
{
  SH2movls4(sh, 10,6, 13);
}
#define init_1_10_6_14
void FASTCALL call_1_10_6_14(SH2_struct* sh)
{
  SH2movls4(sh, 10,6, 14);
}
#define init_1_10_6_15
void FASTCALL call_1_10_6_15(SH2_struct* sh)
{
  SH2movls4(sh, 10,6, 15);
}
#define init_1_10_7_0
void FASTCALL call_1_10_7_0(SH2_struct* sh)
{
  SH2movls4(sh, 10,7, 0);
}
#define init_1_10_7_1
void FASTCALL call_1_10_7_1(SH2_struct* sh)
{
  SH2movls4(sh, 10,7, 1);
}
#define init_1_10_7_2
void FASTCALL call_1_10_7_2(SH2_struct* sh)
{
  SH2movls4(sh, 10,7, 2);
}
#define init_1_10_7_3
void FASTCALL call_1_10_7_3(SH2_struct* sh)
{
  SH2movls4(sh, 10,7, 3);
}
#define init_1_10_7_4
void FASTCALL call_1_10_7_4(SH2_struct* sh)
{
  SH2movls4(sh, 10,7, 4);
}
#define init_1_10_7_5
void FASTCALL call_1_10_7_5(SH2_struct* sh)
{
  SH2movls4(sh, 10,7, 5);
}
#define init_1_10_7_6
void FASTCALL call_1_10_7_6(SH2_struct* sh)
{
  SH2movls4(sh, 10,7, 6);
}
#define init_1_10_7_7
void FASTCALL call_1_10_7_7(SH2_struct* sh)
{
  SH2movls4(sh, 10,7, 7);
}
#define init_1_10_7_8
void FASTCALL call_1_10_7_8(SH2_struct* sh)
{
  SH2movls4(sh, 10,7, 8);
}
#define init_1_10_7_9
void FASTCALL call_1_10_7_9(SH2_struct* sh)
{
  SH2movls4(sh, 10,7, 9);
}
#define init_1_10_7_10
void FASTCALL call_1_10_7_10(SH2_struct* sh)
{
  SH2movls4(sh, 10,7, 10);
}
#define init_1_10_7_11
void FASTCALL call_1_10_7_11(SH2_struct* sh)
{
  SH2movls4(sh, 10,7, 11);
}
#define init_1_10_7_12
void FASTCALL call_1_10_7_12(SH2_struct* sh)
{
  SH2movls4(sh, 10,7, 12);
}
#define init_1_10_7_13
void FASTCALL call_1_10_7_13(SH2_struct* sh)
{
  SH2movls4(sh, 10,7, 13);
}
#define init_1_10_7_14
void FASTCALL call_1_10_7_14(SH2_struct* sh)
{
  SH2movls4(sh, 10,7, 14);
}
#define init_1_10_7_15
void FASTCALL call_1_10_7_15(SH2_struct* sh)
{
  SH2movls4(sh, 10,7, 15);
}
#define init_1_10_8_0
void FASTCALL call_1_10_8_0(SH2_struct* sh)
{
  SH2movls4(sh, 10,8, 0);
}
#define init_1_10_8_1
void FASTCALL call_1_10_8_1(SH2_struct* sh)
{
  SH2movls4(sh, 10,8, 1);
}
#define init_1_10_8_2
void FASTCALL call_1_10_8_2(SH2_struct* sh)
{
  SH2movls4(sh, 10,8, 2);
}
#define init_1_10_8_3
void FASTCALL call_1_10_8_3(SH2_struct* sh)
{
  SH2movls4(sh, 10,8, 3);
}
#define init_1_10_8_4
void FASTCALL call_1_10_8_4(SH2_struct* sh)
{
  SH2movls4(sh, 10,8, 4);
}
#define init_1_10_8_5
void FASTCALL call_1_10_8_5(SH2_struct* sh)
{
  SH2movls4(sh, 10,8, 5);
}
#define init_1_10_8_6
void FASTCALL call_1_10_8_6(SH2_struct* sh)
{
  SH2movls4(sh, 10,8, 6);
}
#define init_1_10_8_7
void FASTCALL call_1_10_8_7(SH2_struct* sh)
{
  SH2movls4(sh, 10,8, 7);
}
#define init_1_10_8_8
void FASTCALL call_1_10_8_8(SH2_struct* sh)
{
  SH2movls4(sh, 10,8, 8);
}
#define init_1_10_8_9
void FASTCALL call_1_10_8_9(SH2_struct* sh)
{
  SH2movls4(sh, 10,8, 9);
}
#define init_1_10_8_10
void FASTCALL call_1_10_8_10(SH2_struct* sh)
{
  SH2movls4(sh, 10,8, 10);
}
#define init_1_10_8_11
void FASTCALL call_1_10_8_11(SH2_struct* sh)
{
  SH2movls4(sh, 10,8, 11);
}
#define init_1_10_8_12
void FASTCALL call_1_10_8_12(SH2_struct* sh)
{
  SH2movls4(sh, 10,8, 12);
}
#define init_1_10_8_13
void FASTCALL call_1_10_8_13(SH2_struct* sh)
{
  SH2movls4(sh, 10,8, 13);
}
#define init_1_10_8_14
void FASTCALL call_1_10_8_14(SH2_struct* sh)
{
  SH2movls4(sh, 10,8, 14);
}
#define init_1_10_8_15
void FASTCALL call_1_10_8_15(SH2_struct* sh)
{
  SH2movls4(sh, 10,8, 15);
}
#define init_1_10_9_0
void FASTCALL call_1_10_9_0(SH2_struct* sh)
{
  SH2movls4(sh, 10,9, 0);
}
#define init_1_10_9_1
void FASTCALL call_1_10_9_1(SH2_struct* sh)
{
  SH2movls4(sh, 10,9, 1);
}
#define init_1_10_9_2
void FASTCALL call_1_10_9_2(SH2_struct* sh)
{
  SH2movls4(sh, 10,9, 2);
}
#define init_1_10_9_3
void FASTCALL call_1_10_9_3(SH2_struct* sh)
{
  SH2movls4(sh, 10,9, 3);
}
#define init_1_10_9_4
void FASTCALL call_1_10_9_4(SH2_struct* sh)
{
  SH2movls4(sh, 10,9, 4);
}
#define init_1_10_9_5
void FASTCALL call_1_10_9_5(SH2_struct* sh)
{
  SH2movls4(sh, 10,9, 5);
}
#define init_1_10_9_6
void FASTCALL call_1_10_9_6(SH2_struct* sh)
{
  SH2movls4(sh, 10,9, 6);
}
#define init_1_10_9_7
void FASTCALL call_1_10_9_7(SH2_struct* sh)
{
  SH2movls4(sh, 10,9, 7);
}
#define init_1_10_9_8
void FASTCALL call_1_10_9_8(SH2_struct* sh)
{
  SH2movls4(sh, 10,9, 8);
}
#define init_1_10_9_9
void FASTCALL call_1_10_9_9(SH2_struct* sh)
{
  SH2movls4(sh, 10,9, 9);
}
#define init_1_10_9_10
void FASTCALL call_1_10_9_10(SH2_struct* sh)
{
  SH2movls4(sh, 10,9, 10);
}
#define init_1_10_9_11
void FASTCALL call_1_10_9_11(SH2_struct* sh)
{
  SH2movls4(sh, 10,9, 11);
}
#define init_1_10_9_12
void FASTCALL call_1_10_9_12(SH2_struct* sh)
{
  SH2movls4(sh, 10,9, 12);
}
#define init_1_10_9_13
void FASTCALL call_1_10_9_13(SH2_struct* sh)
{
  SH2movls4(sh, 10,9, 13);
}
#define init_1_10_9_14
void FASTCALL call_1_10_9_14(SH2_struct* sh)
{
  SH2movls4(sh, 10,9, 14);
}
#define init_1_10_9_15
void FASTCALL call_1_10_9_15(SH2_struct* sh)
{
  SH2movls4(sh, 10,9, 15);
}
#define init_1_10_10_0
void FASTCALL call_1_10_10_0(SH2_struct* sh)
{
  SH2movls4(sh, 10,10, 0);
}
#define init_1_10_10_1
void FASTCALL call_1_10_10_1(SH2_struct* sh)
{
  SH2movls4(sh, 10,10, 1);
}
#define init_1_10_10_2
void FASTCALL call_1_10_10_2(SH2_struct* sh)
{
  SH2movls4(sh, 10,10, 2);
}
#define init_1_10_10_3
void FASTCALL call_1_10_10_3(SH2_struct* sh)
{
  SH2movls4(sh, 10,10, 3);
}
#define init_1_10_10_4
void FASTCALL call_1_10_10_4(SH2_struct* sh)
{
  SH2movls4(sh, 10,10, 4);
}
#define init_1_10_10_5
void FASTCALL call_1_10_10_5(SH2_struct* sh)
{
  SH2movls4(sh, 10,10, 5);
}
#define init_1_10_10_6
void FASTCALL call_1_10_10_6(SH2_struct* sh)
{
  SH2movls4(sh, 10,10, 6);
}
#define init_1_10_10_7
void FASTCALL call_1_10_10_7(SH2_struct* sh)
{
  SH2movls4(sh, 10,10, 7);
}
#define init_1_10_10_8
void FASTCALL call_1_10_10_8(SH2_struct* sh)
{
  SH2movls4(sh, 10,10, 8);
}
#define init_1_10_10_9
void FASTCALL call_1_10_10_9(SH2_struct* sh)
{
  SH2movls4(sh, 10,10, 9);
}
#define init_1_10_10_10
void FASTCALL call_1_10_10_10(SH2_struct* sh)
{
  SH2movls4(sh, 10,10, 10);
}
#define init_1_10_10_11
void FASTCALL call_1_10_10_11(SH2_struct* sh)
{
  SH2movls4(sh, 10,10, 11);
}
#define init_1_10_10_12
void FASTCALL call_1_10_10_12(SH2_struct* sh)
{
  SH2movls4(sh, 10,10, 12);
}
#define init_1_10_10_13
void FASTCALL call_1_10_10_13(SH2_struct* sh)
{
  SH2movls4(sh, 10,10, 13);
}
#define init_1_10_10_14
void FASTCALL call_1_10_10_14(SH2_struct* sh)
{
  SH2movls4(sh, 10,10, 14);
}
#define init_1_10_10_15
void FASTCALL call_1_10_10_15(SH2_struct* sh)
{
  SH2movls4(sh, 10,10, 15);
}
#define init_1_10_11_0
void FASTCALL call_1_10_11_0(SH2_struct* sh)
{
  SH2movls4(sh, 10,11, 0);
}
#define init_1_10_11_1
void FASTCALL call_1_10_11_1(SH2_struct* sh)
{
  SH2movls4(sh, 10,11, 1);
}
#define init_1_10_11_2
void FASTCALL call_1_10_11_2(SH2_struct* sh)
{
  SH2movls4(sh, 10,11, 2);
}
#define init_1_10_11_3
void FASTCALL call_1_10_11_3(SH2_struct* sh)
{
  SH2movls4(sh, 10,11, 3);
}
#define init_1_10_11_4
void FASTCALL call_1_10_11_4(SH2_struct* sh)
{
  SH2movls4(sh, 10,11, 4);
}
#define init_1_10_11_5
void FASTCALL call_1_10_11_5(SH2_struct* sh)
{
  SH2movls4(sh, 10,11, 5);
}
#define init_1_10_11_6
void FASTCALL call_1_10_11_6(SH2_struct* sh)
{
  SH2movls4(sh, 10,11, 6);
}
#define init_1_10_11_7
void FASTCALL call_1_10_11_7(SH2_struct* sh)
{
  SH2movls4(sh, 10,11, 7);
}
#define init_1_10_11_8
void FASTCALL call_1_10_11_8(SH2_struct* sh)
{
  SH2movls4(sh, 10,11, 8);
}
#define init_1_10_11_9
void FASTCALL call_1_10_11_9(SH2_struct* sh)
{
  SH2movls4(sh, 10,11, 9);
}
#define init_1_10_11_10
void FASTCALL call_1_10_11_10(SH2_struct* sh)
{
  SH2movls4(sh, 10,11, 10);
}
#define init_1_10_11_11
void FASTCALL call_1_10_11_11(SH2_struct* sh)
{
  SH2movls4(sh, 10,11, 11);
}
#define init_1_10_11_12
void FASTCALL call_1_10_11_12(SH2_struct* sh)
{
  SH2movls4(sh, 10,11, 12);
}
#define init_1_10_11_13
void FASTCALL call_1_10_11_13(SH2_struct* sh)
{
  SH2movls4(sh, 10,11, 13);
}
#define init_1_10_11_14
void FASTCALL call_1_10_11_14(SH2_struct* sh)
{
  SH2movls4(sh, 10,11, 14);
}
#define init_1_10_11_15
void FASTCALL call_1_10_11_15(SH2_struct* sh)
{
  SH2movls4(sh, 10,11, 15);
}
#define init_1_10_12_0
void FASTCALL call_1_10_12_0(SH2_struct* sh)
{
  SH2movls4(sh, 10,12, 0);
}
#define init_1_10_12_1
void FASTCALL call_1_10_12_1(SH2_struct* sh)
{
  SH2movls4(sh, 10,12, 1);
}
#define init_1_10_12_2
void FASTCALL call_1_10_12_2(SH2_struct* sh)
{
  SH2movls4(sh, 10,12, 2);
}
#define init_1_10_12_3
void FASTCALL call_1_10_12_3(SH2_struct* sh)
{
  SH2movls4(sh, 10,12, 3);
}
#define init_1_10_12_4
void FASTCALL call_1_10_12_4(SH2_struct* sh)
{
  SH2movls4(sh, 10,12, 4);
}
#define init_1_10_12_5
void FASTCALL call_1_10_12_5(SH2_struct* sh)
{
  SH2movls4(sh, 10,12, 5);
}
#define init_1_10_12_6
void FASTCALL call_1_10_12_6(SH2_struct* sh)
{
  SH2movls4(sh, 10,12, 6);
}
#define init_1_10_12_7
void FASTCALL call_1_10_12_7(SH2_struct* sh)
{
  SH2movls4(sh, 10,12, 7);
}
#define init_1_10_12_8
void FASTCALL call_1_10_12_8(SH2_struct* sh)
{
  SH2movls4(sh, 10,12, 8);
}
#define init_1_10_12_9
void FASTCALL call_1_10_12_9(SH2_struct* sh)
{
  SH2movls4(sh, 10,12, 9);
}
#define init_1_10_12_10
void FASTCALL call_1_10_12_10(SH2_struct* sh)
{
  SH2movls4(sh, 10,12, 10);
}
#define init_1_10_12_11
void FASTCALL call_1_10_12_11(SH2_struct* sh)
{
  SH2movls4(sh, 10,12, 11);
}
#define init_1_10_12_12
void FASTCALL call_1_10_12_12(SH2_struct* sh)
{
  SH2movls4(sh, 10,12, 12);
}
#define init_1_10_12_13
void FASTCALL call_1_10_12_13(SH2_struct* sh)
{
  SH2movls4(sh, 10,12, 13);
}
#define init_1_10_12_14
void FASTCALL call_1_10_12_14(SH2_struct* sh)
{
  SH2movls4(sh, 10,12, 14);
}
#define init_1_10_12_15
void FASTCALL call_1_10_12_15(SH2_struct* sh)
{
  SH2movls4(sh, 10,12, 15);
}
#define init_1_10_13_0
void FASTCALL call_1_10_13_0(SH2_struct* sh)
{
  SH2movls4(sh, 10,13, 0);
}
#define init_1_10_13_1
void FASTCALL call_1_10_13_1(SH2_struct* sh)
{
  SH2movls4(sh, 10,13, 1);
}
#define init_1_10_13_2
void FASTCALL call_1_10_13_2(SH2_struct* sh)
{
  SH2movls4(sh, 10,13, 2);
}
#define init_1_10_13_3
void FASTCALL call_1_10_13_3(SH2_struct* sh)
{
  SH2movls4(sh, 10,13, 3);
}
#define init_1_10_13_4
void FASTCALL call_1_10_13_4(SH2_struct* sh)
{
  SH2movls4(sh, 10,13, 4);
}
#define init_1_10_13_5
void FASTCALL call_1_10_13_5(SH2_struct* sh)
{
  SH2movls4(sh, 10,13, 5);
}
#define init_1_10_13_6
void FASTCALL call_1_10_13_6(SH2_struct* sh)
{
  SH2movls4(sh, 10,13, 6);
}
#define init_1_10_13_7
void FASTCALL call_1_10_13_7(SH2_struct* sh)
{
  SH2movls4(sh, 10,13, 7);
}
#define init_1_10_13_8
void FASTCALL call_1_10_13_8(SH2_struct* sh)
{
  SH2movls4(sh, 10,13, 8);
}
#define init_1_10_13_9
void FASTCALL call_1_10_13_9(SH2_struct* sh)
{
  SH2movls4(sh, 10,13, 9);
}
#define init_1_10_13_10
void FASTCALL call_1_10_13_10(SH2_struct* sh)
{
  SH2movls4(sh, 10,13, 10);
}
#define init_1_10_13_11
void FASTCALL call_1_10_13_11(SH2_struct* sh)
{
  SH2movls4(sh, 10,13, 11);
}
#define init_1_10_13_12
void FASTCALL call_1_10_13_12(SH2_struct* sh)
{
  SH2movls4(sh, 10,13, 12);
}
#define init_1_10_13_13
void FASTCALL call_1_10_13_13(SH2_struct* sh)
{
  SH2movls4(sh, 10,13, 13);
}
#define init_1_10_13_14
void FASTCALL call_1_10_13_14(SH2_struct* sh)
{
  SH2movls4(sh, 10,13, 14);
}
#define init_1_10_13_15
void FASTCALL call_1_10_13_15(SH2_struct* sh)
{
  SH2movls4(sh, 10,13, 15);
}
#define init_1_10_14_0
void FASTCALL call_1_10_14_0(SH2_struct* sh)
{
  SH2movls4(sh, 10,14, 0);
}
#define init_1_10_14_1
void FASTCALL call_1_10_14_1(SH2_struct* sh)
{
  SH2movls4(sh, 10,14, 1);
}
#define init_1_10_14_2
void FASTCALL call_1_10_14_2(SH2_struct* sh)
{
  SH2movls4(sh, 10,14, 2);
}
#define init_1_10_14_3
void FASTCALL call_1_10_14_3(SH2_struct* sh)
{
  SH2movls4(sh, 10,14, 3);
}
#define init_1_10_14_4
void FASTCALL call_1_10_14_4(SH2_struct* sh)
{
  SH2movls4(sh, 10,14, 4);
}
#define init_1_10_14_5
void FASTCALL call_1_10_14_5(SH2_struct* sh)
{
  SH2movls4(sh, 10,14, 5);
}
#define init_1_10_14_6
void FASTCALL call_1_10_14_6(SH2_struct* sh)
{
  SH2movls4(sh, 10,14, 6);
}
#define init_1_10_14_7
void FASTCALL call_1_10_14_7(SH2_struct* sh)
{
  SH2movls4(sh, 10,14, 7);
}
#define init_1_10_14_8
void FASTCALL call_1_10_14_8(SH2_struct* sh)
{
  SH2movls4(sh, 10,14, 8);
}
#define init_1_10_14_9
void FASTCALL call_1_10_14_9(SH2_struct* sh)
{
  SH2movls4(sh, 10,14, 9);
}
#define init_1_10_14_10
void FASTCALL call_1_10_14_10(SH2_struct* sh)
{
  SH2movls4(sh, 10,14, 10);
}
#define init_1_10_14_11
void FASTCALL call_1_10_14_11(SH2_struct* sh)
{
  SH2movls4(sh, 10,14, 11);
}
#define init_1_10_14_12
void FASTCALL call_1_10_14_12(SH2_struct* sh)
{
  SH2movls4(sh, 10,14, 12);
}
#define init_1_10_14_13
void FASTCALL call_1_10_14_13(SH2_struct* sh)
{
  SH2movls4(sh, 10,14, 13);
}
#define init_1_10_14_14
void FASTCALL call_1_10_14_14(SH2_struct* sh)
{
  SH2movls4(sh, 10,14, 14);
}
#define init_1_10_14_15
void FASTCALL call_1_10_14_15(SH2_struct* sh)
{
  SH2movls4(sh, 10,14, 15);
}
#define init_1_10_15_0
void FASTCALL call_1_10_15_0(SH2_struct* sh)
{
  SH2movls4(sh, 10,15, 0);
}
#define init_1_10_15_1
void FASTCALL call_1_10_15_1(SH2_struct* sh)
{
  SH2movls4(sh, 10,15, 1);
}
#define init_1_10_15_2
void FASTCALL call_1_10_15_2(SH2_struct* sh)
{
  SH2movls4(sh, 10,15, 2);
}
#define init_1_10_15_3
void FASTCALL call_1_10_15_3(SH2_struct* sh)
{
  SH2movls4(sh, 10,15, 3);
}
#define init_1_10_15_4
void FASTCALL call_1_10_15_4(SH2_struct* sh)
{
  SH2movls4(sh, 10,15, 4);
}
#define init_1_10_15_5
void FASTCALL call_1_10_15_5(SH2_struct* sh)
{
  SH2movls4(sh, 10,15, 5);
}
#define init_1_10_15_6
void FASTCALL call_1_10_15_6(SH2_struct* sh)
{
  SH2movls4(sh, 10,15, 6);
}
#define init_1_10_15_7
void FASTCALL call_1_10_15_7(SH2_struct* sh)
{
  SH2movls4(sh, 10,15, 7);
}
#define init_1_10_15_8
void FASTCALL call_1_10_15_8(SH2_struct* sh)
{
  SH2movls4(sh, 10,15, 8);
}
#define init_1_10_15_9
void FASTCALL call_1_10_15_9(SH2_struct* sh)
{
  SH2movls4(sh, 10,15, 9);
}
#define init_1_10_15_10
void FASTCALL call_1_10_15_10(SH2_struct* sh)
{
  SH2movls4(sh, 10,15, 10);
}
#define init_1_10_15_11
void FASTCALL call_1_10_15_11(SH2_struct* sh)
{
  SH2movls4(sh, 10,15, 11);
}
#define init_1_10_15_12
void FASTCALL call_1_10_15_12(SH2_struct* sh)
{
  SH2movls4(sh, 10,15, 12);
}
#define init_1_10_15_13
void FASTCALL call_1_10_15_13(SH2_struct* sh)
{
  SH2movls4(sh, 10,15, 13);
}
#define init_1_10_15_14
void FASTCALL call_1_10_15_14(SH2_struct* sh)
{
  SH2movls4(sh, 10,15, 14);
}
#define init_1_10_15_15
void FASTCALL call_1_10_15_15(SH2_struct* sh)
{
  SH2movls4(sh, 10,15, 15);
}
#define init_1_11_0_0
void FASTCALL call_1_11_0_0(SH2_struct* sh)
{
  SH2movls4(sh, 11,0, 0);
}
#define init_1_11_0_1
void FASTCALL call_1_11_0_1(SH2_struct* sh)
{
  SH2movls4(sh, 11,0, 1);
}
#define init_1_11_0_2
void FASTCALL call_1_11_0_2(SH2_struct* sh)
{
  SH2movls4(sh, 11,0, 2);
}
#define init_1_11_0_3
void FASTCALL call_1_11_0_3(SH2_struct* sh)
{
  SH2movls4(sh, 11,0, 3);
}
#define init_1_11_0_4
void FASTCALL call_1_11_0_4(SH2_struct* sh)
{
  SH2movls4(sh, 11,0, 4);
}
#define init_1_11_0_5
void FASTCALL call_1_11_0_5(SH2_struct* sh)
{
  SH2movls4(sh, 11,0, 5);
}
#define init_1_11_0_6
void FASTCALL call_1_11_0_6(SH2_struct* sh)
{
  SH2movls4(sh, 11,0, 6);
}
#define init_1_11_0_7
void FASTCALL call_1_11_0_7(SH2_struct* sh)
{
  SH2movls4(sh, 11,0, 7);
}
#define init_1_11_0_8
void FASTCALL call_1_11_0_8(SH2_struct* sh)
{
  SH2movls4(sh, 11,0, 8);
}
#define init_1_11_0_9
void FASTCALL call_1_11_0_9(SH2_struct* sh)
{
  SH2movls4(sh, 11,0, 9);
}
#define init_1_11_0_10
void FASTCALL call_1_11_0_10(SH2_struct* sh)
{
  SH2movls4(sh, 11,0, 10);
}
#define init_1_11_0_11
void FASTCALL call_1_11_0_11(SH2_struct* sh)
{
  SH2movls4(sh, 11,0, 11);
}
#define init_1_11_0_12
void FASTCALL call_1_11_0_12(SH2_struct* sh)
{
  SH2movls4(sh, 11,0, 12);
}
#define init_1_11_0_13
void FASTCALL call_1_11_0_13(SH2_struct* sh)
{
  SH2movls4(sh, 11,0, 13);
}
#define init_1_11_0_14
void FASTCALL call_1_11_0_14(SH2_struct* sh)
{
  SH2movls4(sh, 11,0, 14);
}
#define init_1_11_0_15
void FASTCALL call_1_11_0_15(SH2_struct* sh)
{
  SH2movls4(sh, 11,0, 15);
}
#define init_1_11_1_0
void FASTCALL call_1_11_1_0(SH2_struct* sh)
{
  SH2movls4(sh, 11,1, 0);
}
#define init_1_11_1_1
void FASTCALL call_1_11_1_1(SH2_struct* sh)
{
  SH2movls4(sh, 11,1, 1);
}
#define init_1_11_1_2
void FASTCALL call_1_11_1_2(SH2_struct* sh)
{
  SH2movls4(sh, 11,1, 2);
}
#define init_1_11_1_3
void FASTCALL call_1_11_1_3(SH2_struct* sh)
{
  SH2movls4(sh, 11,1, 3);
}
#define init_1_11_1_4
void FASTCALL call_1_11_1_4(SH2_struct* sh)
{
  SH2movls4(sh, 11,1, 4);
}
#define init_1_11_1_5
void FASTCALL call_1_11_1_5(SH2_struct* sh)
{
  SH2movls4(sh, 11,1, 5);
}
#define init_1_11_1_6
void FASTCALL call_1_11_1_6(SH2_struct* sh)
{
  SH2movls4(sh, 11,1, 6);
}
#define init_1_11_1_7
void FASTCALL call_1_11_1_7(SH2_struct* sh)
{
  SH2movls4(sh, 11,1, 7);
}
#define init_1_11_1_8
void FASTCALL call_1_11_1_8(SH2_struct* sh)
{
  SH2movls4(sh, 11,1, 8);
}
#define init_1_11_1_9
void FASTCALL call_1_11_1_9(SH2_struct* sh)
{
  SH2movls4(sh, 11,1, 9);
}
#define init_1_11_1_10
void FASTCALL call_1_11_1_10(SH2_struct* sh)
{
  SH2movls4(sh, 11,1, 10);
}
#define init_1_11_1_11
void FASTCALL call_1_11_1_11(SH2_struct* sh)
{
  SH2movls4(sh, 11,1, 11);
}
#define init_1_11_1_12
void FASTCALL call_1_11_1_12(SH2_struct* sh)
{
  SH2movls4(sh, 11,1, 12);
}
#define init_1_11_1_13
void FASTCALL call_1_11_1_13(SH2_struct* sh)
{
  SH2movls4(sh, 11,1, 13);
}
#define init_1_11_1_14
void FASTCALL call_1_11_1_14(SH2_struct* sh)
{
  SH2movls4(sh, 11,1, 14);
}
#define init_1_11_1_15
void FASTCALL call_1_11_1_15(SH2_struct* sh)
{
  SH2movls4(sh, 11,1, 15);
}
#define init_1_11_2_0
void FASTCALL call_1_11_2_0(SH2_struct* sh)
{
  SH2movls4(sh, 11,2, 0);
}
#define init_1_11_2_1
void FASTCALL call_1_11_2_1(SH2_struct* sh)
{
  SH2movls4(sh, 11,2, 1);
}
#define init_1_11_2_2
void FASTCALL call_1_11_2_2(SH2_struct* sh)
{
  SH2movls4(sh, 11,2, 2);
}
#define init_1_11_2_3
void FASTCALL call_1_11_2_3(SH2_struct* sh)
{
  SH2movls4(sh, 11,2, 3);
}
#define init_1_11_2_4
void FASTCALL call_1_11_2_4(SH2_struct* sh)
{
  SH2movls4(sh, 11,2, 4);
}
#define init_1_11_2_5
void FASTCALL call_1_11_2_5(SH2_struct* sh)
{
  SH2movls4(sh, 11,2, 5);
}
#define init_1_11_2_6
void FASTCALL call_1_11_2_6(SH2_struct* sh)
{
  SH2movls4(sh, 11,2, 6);
}
#define init_1_11_2_7
void FASTCALL call_1_11_2_7(SH2_struct* sh)
{
  SH2movls4(sh, 11,2, 7);
}
#define init_1_11_2_8
void FASTCALL call_1_11_2_8(SH2_struct* sh)
{
  SH2movls4(sh, 11,2, 8);
}
#define init_1_11_2_9
void FASTCALL call_1_11_2_9(SH2_struct* sh)
{
  SH2movls4(sh, 11,2, 9);
}
#define init_1_11_2_10
void FASTCALL call_1_11_2_10(SH2_struct* sh)
{
  SH2movls4(sh, 11,2, 10);
}
#define init_1_11_2_11
void FASTCALL call_1_11_2_11(SH2_struct* sh)
{
  SH2movls4(sh, 11,2, 11);
}
#define init_1_11_2_12
void FASTCALL call_1_11_2_12(SH2_struct* sh)
{
  SH2movls4(sh, 11,2, 12);
}
#define init_1_11_2_13
void FASTCALL call_1_11_2_13(SH2_struct* sh)
{
  SH2movls4(sh, 11,2, 13);
}
#define init_1_11_2_14
void FASTCALL call_1_11_2_14(SH2_struct* sh)
{
  SH2movls4(sh, 11,2, 14);
}
#define init_1_11_2_15
void FASTCALL call_1_11_2_15(SH2_struct* sh)
{
  SH2movls4(sh, 11,2, 15);
}
#define init_1_11_3_0
void FASTCALL call_1_11_3_0(SH2_struct* sh)
{
  SH2movls4(sh, 11,3, 0);
}
#define init_1_11_3_1
void FASTCALL call_1_11_3_1(SH2_struct* sh)
{
  SH2movls4(sh, 11,3, 1);
}
#define init_1_11_3_2
void FASTCALL call_1_11_3_2(SH2_struct* sh)
{
  SH2movls4(sh, 11,3, 2);
}
#define init_1_11_3_3
void FASTCALL call_1_11_3_3(SH2_struct* sh)
{
  SH2movls4(sh, 11,3, 3);
}
#define init_1_11_3_4
void FASTCALL call_1_11_3_4(SH2_struct* sh)
{
  SH2movls4(sh, 11,3, 4);
}
#define init_1_11_3_5
void FASTCALL call_1_11_3_5(SH2_struct* sh)
{
  SH2movls4(sh, 11,3, 5);
}
#define init_1_11_3_6
void FASTCALL call_1_11_3_6(SH2_struct* sh)
{
  SH2movls4(sh, 11,3, 6);
}
#define init_1_11_3_7
void FASTCALL call_1_11_3_7(SH2_struct* sh)
{
  SH2movls4(sh, 11,3, 7);
}
#define init_1_11_3_8
void FASTCALL call_1_11_3_8(SH2_struct* sh)
{
  SH2movls4(sh, 11,3, 8);
}
#define init_1_11_3_9
void FASTCALL call_1_11_3_9(SH2_struct* sh)
{
  SH2movls4(sh, 11,3, 9);
}
#define init_1_11_3_10
void FASTCALL call_1_11_3_10(SH2_struct* sh)
{
  SH2movls4(sh, 11,3, 10);
}
#define init_1_11_3_11
void FASTCALL call_1_11_3_11(SH2_struct* sh)
{
  SH2movls4(sh, 11,3, 11);
}
#define init_1_11_3_12
void FASTCALL call_1_11_3_12(SH2_struct* sh)
{
  SH2movls4(sh, 11,3, 12);
}
#define init_1_11_3_13
void FASTCALL call_1_11_3_13(SH2_struct* sh)
{
  SH2movls4(sh, 11,3, 13);
}
#define init_1_11_3_14
void FASTCALL call_1_11_3_14(SH2_struct* sh)
{
  SH2movls4(sh, 11,3, 14);
}
#define init_1_11_3_15
void FASTCALL call_1_11_3_15(SH2_struct* sh)
{
  SH2movls4(sh, 11,3, 15);
}
#define init_1_11_4_0
void FASTCALL call_1_11_4_0(SH2_struct* sh)
{
  SH2movls4(sh, 11,4, 0);
}
#define init_1_11_4_1
void FASTCALL call_1_11_4_1(SH2_struct* sh)
{
  SH2movls4(sh, 11,4, 1);
}
#define init_1_11_4_2
void FASTCALL call_1_11_4_2(SH2_struct* sh)
{
  SH2movls4(sh, 11,4, 2);
}
#define init_1_11_4_3
void FASTCALL call_1_11_4_3(SH2_struct* sh)
{
  SH2movls4(sh, 11,4, 3);
}
#define init_1_11_4_4
void FASTCALL call_1_11_4_4(SH2_struct* sh)
{
  SH2movls4(sh, 11,4, 4);
}
#define init_1_11_4_5
void FASTCALL call_1_11_4_5(SH2_struct* sh)
{
  SH2movls4(sh, 11,4, 5);
}
#define init_1_11_4_6
void FASTCALL call_1_11_4_6(SH2_struct* sh)
{
  SH2movls4(sh, 11,4, 6);
}
#define init_1_11_4_7
void FASTCALL call_1_11_4_7(SH2_struct* sh)
{
  SH2movls4(sh, 11,4, 7);
}
#define init_1_11_4_8
void FASTCALL call_1_11_4_8(SH2_struct* sh)
{
  SH2movls4(sh, 11,4, 8);
}
#define init_1_11_4_9
void FASTCALL call_1_11_4_9(SH2_struct* sh)
{
  SH2movls4(sh, 11,4, 9);
}
#define init_1_11_4_10
void FASTCALL call_1_11_4_10(SH2_struct* sh)
{
  SH2movls4(sh, 11,4, 10);
}
#define init_1_11_4_11
void FASTCALL call_1_11_4_11(SH2_struct* sh)
{
  SH2movls4(sh, 11,4, 11);
}
#define init_1_11_4_12
void FASTCALL call_1_11_4_12(SH2_struct* sh)
{
  SH2movls4(sh, 11,4, 12);
}
#define init_1_11_4_13
void FASTCALL call_1_11_4_13(SH2_struct* sh)
{
  SH2movls4(sh, 11,4, 13);
}
#define init_1_11_4_14
void FASTCALL call_1_11_4_14(SH2_struct* sh)
{
  SH2movls4(sh, 11,4, 14);
}
#define init_1_11_4_15
void FASTCALL call_1_11_4_15(SH2_struct* sh)
{
  SH2movls4(sh, 11,4, 15);
}
#define init_1_11_5_0
void FASTCALL call_1_11_5_0(SH2_struct* sh)
{
  SH2movls4(sh, 11,5, 0);
}
#define init_1_11_5_1
void FASTCALL call_1_11_5_1(SH2_struct* sh)
{
  SH2movls4(sh, 11,5, 1);
}
#define init_1_11_5_2
void FASTCALL call_1_11_5_2(SH2_struct* sh)
{
  SH2movls4(sh, 11,5, 2);
}
#define init_1_11_5_3
void FASTCALL call_1_11_5_3(SH2_struct* sh)
{
  SH2movls4(sh, 11,5, 3);
}
#define init_1_11_5_4
void FASTCALL call_1_11_5_4(SH2_struct* sh)
{
  SH2movls4(sh, 11,5, 4);
}
#define init_1_11_5_5
void FASTCALL call_1_11_5_5(SH2_struct* sh)
{
  SH2movls4(sh, 11,5, 5);
}
#define init_1_11_5_6
void FASTCALL call_1_11_5_6(SH2_struct* sh)
{
  SH2movls4(sh, 11,5, 6);
}
#define init_1_11_5_7
void FASTCALL call_1_11_5_7(SH2_struct* sh)
{
  SH2movls4(sh, 11,5, 7);
}
#define init_1_11_5_8
void FASTCALL call_1_11_5_8(SH2_struct* sh)
{
  SH2movls4(sh, 11,5, 8);
}
#define init_1_11_5_9
void FASTCALL call_1_11_5_9(SH2_struct* sh)
{
  SH2movls4(sh, 11,5, 9);
}
#define init_1_11_5_10
void FASTCALL call_1_11_5_10(SH2_struct* sh)
{
  SH2movls4(sh, 11,5, 10);
}
#define init_1_11_5_11
void FASTCALL call_1_11_5_11(SH2_struct* sh)
{
  SH2movls4(sh, 11,5, 11);
}
#define init_1_11_5_12
void FASTCALL call_1_11_5_12(SH2_struct* sh)
{
  SH2movls4(sh, 11,5, 12);
}
#define init_1_11_5_13
void FASTCALL call_1_11_5_13(SH2_struct* sh)
{
  SH2movls4(sh, 11,5, 13);
}
#define init_1_11_5_14
void FASTCALL call_1_11_5_14(SH2_struct* sh)
{
  SH2movls4(sh, 11,5, 14);
}
#define init_1_11_5_15
void FASTCALL call_1_11_5_15(SH2_struct* sh)
{
  SH2movls4(sh, 11,5, 15);
}
#define init_1_11_6_0
void FASTCALL call_1_11_6_0(SH2_struct* sh)
{
  SH2movls4(sh, 11,6, 0);
}
#define init_1_11_6_1
void FASTCALL call_1_11_6_1(SH2_struct* sh)
{
  SH2movls4(sh, 11,6, 1);
}
#define init_1_11_6_2
void FASTCALL call_1_11_6_2(SH2_struct* sh)
{
  SH2movls4(sh, 11,6, 2);
}
#define init_1_11_6_3
void FASTCALL call_1_11_6_3(SH2_struct* sh)
{
  SH2movls4(sh, 11,6, 3);
}
#define init_1_11_6_4
void FASTCALL call_1_11_6_4(SH2_struct* sh)
{
  SH2movls4(sh, 11,6, 4);
}
#define init_1_11_6_5
void FASTCALL call_1_11_6_5(SH2_struct* sh)
{
  SH2movls4(sh, 11,6, 5);
}
#define init_1_11_6_6
void FASTCALL call_1_11_6_6(SH2_struct* sh)
{
  SH2movls4(sh, 11,6, 6);
}
#define init_1_11_6_7
void FASTCALL call_1_11_6_7(SH2_struct* sh)
{
  SH2movls4(sh, 11,6, 7);
}
#define init_1_11_6_8
void FASTCALL call_1_11_6_8(SH2_struct* sh)
{
  SH2movls4(sh, 11,6, 8);
}
#define init_1_11_6_9
void FASTCALL call_1_11_6_9(SH2_struct* sh)
{
  SH2movls4(sh, 11,6, 9);
}
#define init_1_11_6_10
void FASTCALL call_1_11_6_10(SH2_struct* sh)
{
  SH2movls4(sh, 11,6, 10);
}
#define init_1_11_6_11
void FASTCALL call_1_11_6_11(SH2_struct* sh)
{
  SH2movls4(sh, 11,6, 11);
}
#define init_1_11_6_12
void FASTCALL call_1_11_6_12(SH2_struct* sh)
{
  SH2movls4(sh, 11,6, 12);
}
#define init_1_11_6_13
void FASTCALL call_1_11_6_13(SH2_struct* sh)
{
  SH2movls4(sh, 11,6, 13);
}
#define init_1_11_6_14
void FASTCALL call_1_11_6_14(SH2_struct* sh)
{
  SH2movls4(sh, 11,6, 14);
}
#define init_1_11_6_15
void FASTCALL call_1_11_6_15(SH2_struct* sh)
{
  SH2movls4(sh, 11,6, 15);
}
#define init_1_11_7_0
void FASTCALL call_1_11_7_0(SH2_struct* sh)
{
  SH2movls4(sh, 11,7, 0);
}
#define init_1_11_7_1
void FASTCALL call_1_11_7_1(SH2_struct* sh)
{
  SH2movls4(sh, 11,7, 1);
}
#define init_1_11_7_2
void FASTCALL call_1_11_7_2(SH2_struct* sh)
{
  SH2movls4(sh, 11,7, 2);
}
#define init_1_11_7_3
void FASTCALL call_1_11_7_3(SH2_struct* sh)
{
  SH2movls4(sh, 11,7, 3);
}
#define init_1_11_7_4
void FASTCALL call_1_11_7_4(SH2_struct* sh)
{
  SH2movls4(sh, 11,7, 4);
}
#define init_1_11_7_5
void FASTCALL call_1_11_7_5(SH2_struct* sh)
{
  SH2movls4(sh, 11,7, 5);
}
#define init_1_11_7_6
void FASTCALL call_1_11_7_6(SH2_struct* sh)
{
  SH2movls4(sh, 11,7, 6);
}
#define init_1_11_7_7
void FASTCALL call_1_11_7_7(SH2_struct* sh)
{
  SH2movls4(sh, 11,7, 7);
}
#define init_1_11_7_8
void FASTCALL call_1_11_7_8(SH2_struct* sh)
{
  SH2movls4(sh, 11,7, 8);
}
#define init_1_11_7_9
void FASTCALL call_1_11_7_9(SH2_struct* sh)
{
  SH2movls4(sh, 11,7, 9);
}
#define init_1_11_7_10
void FASTCALL call_1_11_7_10(SH2_struct* sh)
{
  SH2movls4(sh, 11,7, 10);
}
#define init_1_11_7_11
void FASTCALL call_1_11_7_11(SH2_struct* sh)
{
  SH2movls4(sh, 11,7, 11);
}
#define init_1_11_7_12
void FASTCALL call_1_11_7_12(SH2_struct* sh)
{
  SH2movls4(sh, 11,7, 12);
}
#define init_1_11_7_13
void FASTCALL call_1_11_7_13(SH2_struct* sh)
{
  SH2movls4(sh, 11,7, 13);
}
#define init_1_11_7_14
void FASTCALL call_1_11_7_14(SH2_struct* sh)
{
  SH2movls4(sh, 11,7, 14);
}
#define init_1_11_7_15
void FASTCALL call_1_11_7_15(SH2_struct* sh)
{
  SH2movls4(sh, 11,7, 15);
}
#define init_1_11_8_0
void FASTCALL call_1_11_8_0(SH2_struct* sh)
{
  SH2movls4(sh, 11,8, 0);
}
#define init_1_11_8_1
void FASTCALL call_1_11_8_1(SH2_struct* sh)
{
  SH2movls4(sh, 11,8, 1);
}
#define init_1_11_8_2
void FASTCALL call_1_11_8_2(SH2_struct* sh)
{
  SH2movls4(sh, 11,8, 2);
}
#define init_1_11_8_3
void FASTCALL call_1_11_8_3(SH2_struct* sh)
{
  SH2movls4(sh, 11,8, 3);
}
#define init_1_11_8_4
void FASTCALL call_1_11_8_4(SH2_struct* sh)
{
  SH2movls4(sh, 11,8, 4);
}
#define init_1_11_8_5
void FASTCALL call_1_11_8_5(SH2_struct* sh)
{
  SH2movls4(sh, 11,8, 5);
}
#define init_1_11_8_6
void FASTCALL call_1_11_8_6(SH2_struct* sh)
{
  SH2movls4(sh, 11,8, 6);
}
#define init_1_11_8_7
void FASTCALL call_1_11_8_7(SH2_struct* sh)
{
  SH2movls4(sh, 11,8, 7);
}
#define init_1_11_8_8
void FASTCALL call_1_11_8_8(SH2_struct* sh)
{
  SH2movls4(sh, 11,8, 8);
}
#define init_1_11_8_9
void FASTCALL call_1_11_8_9(SH2_struct* sh)
{
  SH2movls4(sh, 11,8, 9);
}
#define init_1_11_8_10
void FASTCALL call_1_11_8_10(SH2_struct* sh)
{
  SH2movls4(sh, 11,8, 10);
}
#define init_1_11_8_11
void FASTCALL call_1_11_8_11(SH2_struct* sh)
{
  SH2movls4(sh, 11,8, 11);
}
#define init_1_11_8_12
void FASTCALL call_1_11_8_12(SH2_struct* sh)
{
  SH2movls4(sh, 11,8, 12);
}
#define init_1_11_8_13
void FASTCALL call_1_11_8_13(SH2_struct* sh)
{
  SH2movls4(sh, 11,8, 13);
}
#define init_1_11_8_14
void FASTCALL call_1_11_8_14(SH2_struct* sh)
{
  SH2movls4(sh, 11,8, 14);
}
#define init_1_11_8_15
void FASTCALL call_1_11_8_15(SH2_struct* sh)
{
  SH2movls4(sh, 11,8, 15);
}
#define init_1_11_9_0
void FASTCALL call_1_11_9_0(SH2_struct* sh)
{
  SH2movls4(sh, 11,9, 0);
}
#define init_1_11_9_1
void FASTCALL call_1_11_9_1(SH2_struct* sh)
{
  SH2movls4(sh, 11,9, 1);
}
#define init_1_11_9_2
void FASTCALL call_1_11_9_2(SH2_struct* sh)
{
  SH2movls4(sh, 11,9, 2);
}
#define init_1_11_9_3
void FASTCALL call_1_11_9_3(SH2_struct* sh)
{
  SH2movls4(sh, 11,9, 3);
}
#define init_1_11_9_4
void FASTCALL call_1_11_9_4(SH2_struct* sh)
{
  SH2movls4(sh, 11,9, 4);
}
#define init_1_11_9_5
void FASTCALL call_1_11_9_5(SH2_struct* sh)
{
  SH2movls4(sh, 11,9, 5);
}
#define init_1_11_9_6
void FASTCALL call_1_11_9_6(SH2_struct* sh)
{
  SH2movls4(sh, 11,9, 6);
}
#define init_1_11_9_7
void FASTCALL call_1_11_9_7(SH2_struct* sh)
{
  SH2movls4(sh, 11,9, 7);
}
#define init_1_11_9_8
void FASTCALL call_1_11_9_8(SH2_struct* sh)
{
  SH2movls4(sh, 11,9, 8);
}
#define init_1_11_9_9
void FASTCALL call_1_11_9_9(SH2_struct* sh)
{
  SH2movls4(sh, 11,9, 9);
}
#define init_1_11_9_10
void FASTCALL call_1_11_9_10(SH2_struct* sh)
{
  SH2movls4(sh, 11,9, 10);
}
#define init_1_11_9_11
void FASTCALL call_1_11_9_11(SH2_struct* sh)
{
  SH2movls4(sh, 11,9, 11);
}
#define init_1_11_9_12
void FASTCALL call_1_11_9_12(SH2_struct* sh)
{
  SH2movls4(sh, 11,9, 12);
}
#define init_1_11_9_13
void FASTCALL call_1_11_9_13(SH2_struct* sh)
{
  SH2movls4(sh, 11,9, 13);
}
#define init_1_11_9_14
void FASTCALL call_1_11_9_14(SH2_struct* sh)
{
  SH2movls4(sh, 11,9, 14);
}
#define init_1_11_9_15
void FASTCALL call_1_11_9_15(SH2_struct* sh)
{
  SH2movls4(sh, 11,9, 15);
}
#define init_1_11_10_0
void FASTCALL call_1_11_10_0(SH2_struct* sh)
{
  SH2movls4(sh, 11,10, 0);
}
#define init_1_11_10_1
void FASTCALL call_1_11_10_1(SH2_struct* sh)
{
  SH2movls4(sh, 11,10, 1);
}
#define init_1_11_10_2
void FASTCALL call_1_11_10_2(SH2_struct* sh)
{
  SH2movls4(sh, 11,10, 2);
}
#define init_1_11_10_3
void FASTCALL call_1_11_10_3(SH2_struct* sh)
{
  SH2movls4(sh, 11,10, 3);
}
#define init_1_11_10_4
void FASTCALL call_1_11_10_4(SH2_struct* sh)
{
  SH2movls4(sh, 11,10, 4);
}
#define init_1_11_10_5
void FASTCALL call_1_11_10_5(SH2_struct* sh)
{
  SH2movls4(sh, 11,10, 5);
}
#define init_1_11_10_6
void FASTCALL call_1_11_10_6(SH2_struct* sh)
{
  SH2movls4(sh, 11,10, 6);
}
#define init_1_11_10_7
void FASTCALL call_1_11_10_7(SH2_struct* sh)
{
  SH2movls4(sh, 11,10, 7);
}
#define init_1_11_10_8
void FASTCALL call_1_11_10_8(SH2_struct* sh)
{
  SH2movls4(sh, 11,10, 8);
}
#define init_1_11_10_9
void FASTCALL call_1_11_10_9(SH2_struct* sh)
{
  SH2movls4(sh, 11,10, 9);
}
#define init_1_11_10_10
void FASTCALL call_1_11_10_10(SH2_struct* sh)
{
  SH2movls4(sh, 11,10, 10);
}
#define init_1_11_10_11
void FASTCALL call_1_11_10_11(SH2_struct* sh)
{
  SH2movls4(sh, 11,10, 11);
}
#define init_1_11_10_12
void FASTCALL call_1_11_10_12(SH2_struct* sh)
{
  SH2movls4(sh, 11,10, 12);
}
#define init_1_11_10_13
void FASTCALL call_1_11_10_13(SH2_struct* sh)
{
  SH2movls4(sh, 11,10, 13);
}
#define init_1_11_10_14
void FASTCALL call_1_11_10_14(SH2_struct* sh)
{
  SH2movls4(sh, 11,10, 14);
}
#define init_1_11_10_15
void FASTCALL call_1_11_10_15(SH2_struct* sh)
{
  SH2movls4(sh, 11,10, 15);
}
#define init_1_11_11_0
void FASTCALL call_1_11_11_0(SH2_struct* sh)
{
  SH2movls4(sh, 11,11, 0);
}
#define init_1_11_11_1
void FASTCALL call_1_11_11_1(SH2_struct* sh)
{
  SH2movls4(sh, 11,11, 1);
}
#define init_1_11_11_2
void FASTCALL call_1_11_11_2(SH2_struct* sh)
{
  SH2movls4(sh, 11,11, 2);
}
#define init_1_11_11_3
void FASTCALL call_1_11_11_3(SH2_struct* sh)
{
  SH2movls4(sh, 11,11, 3);
}
#define init_1_11_11_4
void FASTCALL call_1_11_11_4(SH2_struct* sh)
{
  SH2movls4(sh, 11,11, 4);
}
#define init_1_11_11_5
void FASTCALL call_1_11_11_5(SH2_struct* sh)
{
  SH2movls4(sh, 11,11, 5);
}
#define init_1_11_11_6
void FASTCALL call_1_11_11_6(SH2_struct* sh)
{
  SH2movls4(sh, 11,11, 6);
}
#define init_1_11_11_7
void FASTCALL call_1_11_11_7(SH2_struct* sh)
{
  SH2movls4(sh, 11,11, 7);
}
#define init_1_11_11_8
void FASTCALL call_1_11_11_8(SH2_struct* sh)
{
  SH2movls4(sh, 11,11, 8);
}
#define init_1_11_11_9
void FASTCALL call_1_11_11_9(SH2_struct* sh)
{
  SH2movls4(sh, 11,11, 9);
}
#define init_1_11_11_10
void FASTCALL call_1_11_11_10(SH2_struct* sh)
{
  SH2movls4(sh, 11,11, 10);
}
#define init_1_11_11_11
void FASTCALL call_1_11_11_11(SH2_struct* sh)
{
  SH2movls4(sh, 11,11, 11);
}
#define init_1_11_11_12
void FASTCALL call_1_11_11_12(SH2_struct* sh)
{
  SH2movls4(sh, 11,11, 12);
}
#define init_1_11_11_13
void FASTCALL call_1_11_11_13(SH2_struct* sh)
{
  SH2movls4(sh, 11,11, 13);
}
#define init_1_11_11_14
void FASTCALL call_1_11_11_14(SH2_struct* sh)
{
  SH2movls4(sh, 11,11, 14);
}
#define init_1_11_11_15
void FASTCALL call_1_11_11_15(SH2_struct* sh)
{
  SH2movls4(sh, 11,11, 15);
}
#define init_1_11_12_0
void FASTCALL call_1_11_12_0(SH2_struct* sh)
{
  SH2movls4(sh, 11,12, 0);
}
#define init_1_11_12_1
void FASTCALL call_1_11_12_1(SH2_struct* sh)
{
  SH2movls4(sh, 11,12, 1);
}
#define init_1_11_12_2
void FASTCALL call_1_11_12_2(SH2_struct* sh)
{
  SH2movls4(sh, 11,12, 2);
}
#define init_1_11_12_3
void FASTCALL call_1_11_12_3(SH2_struct* sh)
{
  SH2movls4(sh, 11,12, 3);
}
#define init_1_11_12_4
void FASTCALL call_1_11_12_4(SH2_struct* sh)
{
  SH2movls4(sh, 11,12, 4);
}
#define init_1_11_12_5
void FASTCALL call_1_11_12_5(SH2_struct* sh)
{
  SH2movls4(sh, 11,12, 5);
}
#define init_1_11_12_6
void FASTCALL call_1_11_12_6(SH2_struct* sh)
{
  SH2movls4(sh, 11,12, 6);
}
#define init_1_11_12_7
void FASTCALL call_1_11_12_7(SH2_struct* sh)
{
  SH2movls4(sh, 11,12, 7);
}
#define init_1_11_12_8
void FASTCALL call_1_11_12_8(SH2_struct* sh)
{
  SH2movls4(sh, 11,12, 8);
}
#define init_1_11_12_9
void FASTCALL call_1_11_12_9(SH2_struct* sh)
{
  SH2movls4(sh, 11,12, 9);
}
#define init_1_11_12_10
void FASTCALL call_1_11_12_10(SH2_struct* sh)
{
  SH2movls4(sh, 11,12, 10);
}
#define init_1_11_12_11
void FASTCALL call_1_11_12_11(SH2_struct* sh)
{
  SH2movls4(sh, 11,12, 11);
}
#define init_1_11_12_12
void FASTCALL call_1_11_12_12(SH2_struct* sh)
{
  SH2movls4(sh, 11,12, 12);
}
#define init_1_11_12_13
void FASTCALL call_1_11_12_13(SH2_struct* sh)
{
  SH2movls4(sh, 11,12, 13);
}
#define init_1_11_12_14
void FASTCALL call_1_11_12_14(SH2_struct* sh)
{
  SH2movls4(sh, 11,12, 14);
}
#define init_1_11_12_15
void FASTCALL call_1_11_12_15(SH2_struct* sh)
{
  SH2movls4(sh, 11,12, 15);
}
#define init_1_11_13_0
void FASTCALL call_1_11_13_0(SH2_struct* sh)
{
  SH2movls4(sh, 11,13, 0);
}
#define init_1_11_13_1
void FASTCALL call_1_11_13_1(SH2_struct* sh)
{
  SH2movls4(sh, 11,13, 1);
}
#define init_1_11_13_2
void FASTCALL call_1_11_13_2(SH2_struct* sh)
{
  SH2movls4(sh, 11,13, 2);
}
#define init_1_11_13_3
void FASTCALL call_1_11_13_3(SH2_struct* sh)
{
  SH2movls4(sh, 11,13, 3);
}
#define init_1_11_13_4
void FASTCALL call_1_11_13_4(SH2_struct* sh)
{
  SH2movls4(sh, 11,13, 4);
}
#define init_1_11_13_5
void FASTCALL call_1_11_13_5(SH2_struct* sh)
{
  SH2movls4(sh, 11,13, 5);
}
#define init_1_11_13_6
void FASTCALL call_1_11_13_6(SH2_struct* sh)
{
  SH2movls4(sh, 11,13, 6);
}
#define init_1_11_13_7
void FASTCALL call_1_11_13_7(SH2_struct* sh)
{
  SH2movls4(sh, 11,13, 7);
}
#define init_1_11_13_8
void FASTCALL call_1_11_13_8(SH2_struct* sh)
{
  SH2movls4(sh, 11,13, 8);
}
#define init_1_11_13_9
void FASTCALL call_1_11_13_9(SH2_struct* sh)
{
  SH2movls4(sh, 11,13, 9);
}
#define init_1_11_13_10
void FASTCALL call_1_11_13_10(SH2_struct* sh)
{
  SH2movls4(sh, 11,13, 10);
}
#define init_1_11_13_11
void FASTCALL call_1_11_13_11(SH2_struct* sh)
{
  SH2movls4(sh, 11,13, 11);
}
#define init_1_11_13_12
void FASTCALL call_1_11_13_12(SH2_struct* sh)
{
  SH2movls4(sh, 11,13, 12);
}
#define init_1_11_13_13
void FASTCALL call_1_11_13_13(SH2_struct* sh)
{
  SH2movls4(sh, 11,13, 13);
}
#define init_1_11_13_14
void FASTCALL call_1_11_13_14(SH2_struct* sh)
{
  SH2movls4(sh, 11,13, 14);
}
#define init_1_11_13_15
void FASTCALL call_1_11_13_15(SH2_struct* sh)
{
  SH2movls4(sh, 11,13, 15);
}
#define init_1_11_14_0
void FASTCALL call_1_11_14_0(SH2_struct* sh)
{
  SH2movls4(sh, 11,14, 0);
}
#define init_1_11_14_1
void FASTCALL call_1_11_14_1(SH2_struct* sh)
{
  SH2movls4(sh, 11,14, 1);
}
#define init_1_11_14_2
void FASTCALL call_1_11_14_2(SH2_struct* sh)
{
  SH2movls4(sh, 11,14, 2);
}
#define init_1_11_14_3
void FASTCALL call_1_11_14_3(SH2_struct* sh)
{
  SH2movls4(sh, 11,14, 3);
}
#define init_1_11_14_4
void FASTCALL call_1_11_14_4(SH2_struct* sh)
{
  SH2movls4(sh, 11,14, 4);
}
#define init_1_11_14_5
void FASTCALL call_1_11_14_5(SH2_struct* sh)
{
  SH2movls4(sh, 11,14, 5);
}
#define init_1_11_14_6
void FASTCALL call_1_11_14_6(SH2_struct* sh)
{
  SH2movls4(sh, 11,14, 6);
}
#define init_1_11_14_7
void FASTCALL call_1_11_14_7(SH2_struct* sh)
{
  SH2movls4(sh, 11,14, 7);
}
#define init_1_11_14_8
void FASTCALL call_1_11_14_8(SH2_struct* sh)
{
  SH2movls4(sh, 11,14, 8);
}
#define init_1_11_14_9
void FASTCALL call_1_11_14_9(SH2_struct* sh)
{
  SH2movls4(sh, 11,14, 9);
}
#define init_1_11_14_10
void FASTCALL call_1_11_14_10(SH2_struct* sh)
{
  SH2movls4(sh, 11,14, 10);
}
#define init_1_11_14_11
void FASTCALL call_1_11_14_11(SH2_struct* sh)
{
  SH2movls4(sh, 11,14, 11);
}
#define init_1_11_14_12
void FASTCALL call_1_11_14_12(SH2_struct* sh)
{
  SH2movls4(sh, 11,14, 12);
}
#define init_1_11_14_13
void FASTCALL call_1_11_14_13(SH2_struct* sh)
{
  SH2movls4(sh, 11,14, 13);
}
#define init_1_11_14_14
void FASTCALL call_1_11_14_14(SH2_struct* sh)
{
  SH2movls4(sh, 11,14, 14);
}
#define init_1_11_14_15
void FASTCALL call_1_11_14_15(SH2_struct* sh)
{
  SH2movls4(sh, 11,14, 15);
}
#define init_1_11_15_0
void FASTCALL call_1_11_15_0(SH2_struct* sh)
{
  SH2movls4(sh, 11,15, 0);
}
#define init_1_11_15_1
void FASTCALL call_1_11_15_1(SH2_struct* sh)
{
  SH2movls4(sh, 11,15, 1);
}
#define init_1_11_15_2
void FASTCALL call_1_11_15_2(SH2_struct* sh)
{
  SH2movls4(sh, 11,15, 2);
}
#define init_1_11_15_3
void FASTCALL call_1_11_15_3(SH2_struct* sh)
{
  SH2movls4(sh, 11,15, 3);
}
#define init_1_11_15_4
void FASTCALL call_1_11_15_4(SH2_struct* sh)
{
  SH2movls4(sh, 11,15, 4);
}
#define init_1_11_15_5
void FASTCALL call_1_11_15_5(SH2_struct* sh)
{
  SH2movls4(sh, 11,15, 5);
}
#define init_1_11_15_6
void FASTCALL call_1_11_15_6(SH2_struct* sh)
{
  SH2movls4(sh, 11,15, 6);
}
#define init_1_11_15_7
void FASTCALL call_1_11_15_7(SH2_struct* sh)
{
  SH2movls4(sh, 11,15, 7);
}
#define init_1_11_15_8
void FASTCALL call_1_11_15_8(SH2_struct* sh)
{
  SH2movls4(sh, 11,15, 8);
}
#define init_1_11_15_9
void FASTCALL call_1_11_15_9(SH2_struct* sh)
{
  SH2movls4(sh, 11,15, 9);
}
#define init_1_11_15_10
void FASTCALL call_1_11_15_10(SH2_struct* sh)
{
  SH2movls4(sh, 11,15, 10);
}
#define init_1_11_15_11
void FASTCALL call_1_11_15_11(SH2_struct* sh)
{
  SH2movls4(sh, 11,15, 11);
}
#define init_1_11_15_12
void FASTCALL call_1_11_15_12(SH2_struct* sh)
{
  SH2movls4(sh, 11,15, 12);
}
#define init_1_11_15_13
void FASTCALL call_1_11_15_13(SH2_struct* sh)
{
  SH2movls4(sh, 11,15, 13);
}
#define init_1_11_15_14
void FASTCALL call_1_11_15_14(SH2_struct* sh)
{
  SH2movls4(sh, 11,15, 14);
}
#define init_1_11_15_15
void FASTCALL call_1_11_15_15(SH2_struct* sh)
{
  SH2movls4(sh, 11,15, 15);
}
#define init_1_12_0_0
void FASTCALL call_1_12_0_0(SH2_struct* sh)
{
  SH2movls4(sh, 12,0, 0);
}
#define init_1_12_0_1
void FASTCALL call_1_12_0_1(SH2_struct* sh)
{
  SH2movls4(sh, 12,0, 1);
}
#define init_1_12_0_2
void FASTCALL call_1_12_0_2(SH2_struct* sh)
{
  SH2movls4(sh, 12,0, 2);
}
#define init_1_12_0_3
void FASTCALL call_1_12_0_3(SH2_struct* sh)
{
  SH2movls4(sh, 12,0, 3);
}
#define init_1_12_0_4
void FASTCALL call_1_12_0_4(SH2_struct* sh)
{
  SH2movls4(sh, 12,0, 4);
}
#define init_1_12_0_5
void FASTCALL call_1_12_0_5(SH2_struct* sh)
{
  SH2movls4(sh, 12,0, 5);
}
#define init_1_12_0_6
void FASTCALL call_1_12_0_6(SH2_struct* sh)
{
  SH2movls4(sh, 12,0, 6);
}
#define init_1_12_0_7
void FASTCALL call_1_12_0_7(SH2_struct* sh)
{
  SH2movls4(sh, 12,0, 7);
}
#define init_1_12_0_8
void FASTCALL call_1_12_0_8(SH2_struct* sh)
{
  SH2movls4(sh, 12,0, 8);
}
#define init_1_12_0_9
void FASTCALL call_1_12_0_9(SH2_struct* sh)
{
  SH2movls4(sh, 12,0, 9);
}
#define init_1_12_0_10
void FASTCALL call_1_12_0_10(SH2_struct* sh)
{
  SH2movls4(sh, 12,0, 10);
}
#define init_1_12_0_11
void FASTCALL call_1_12_0_11(SH2_struct* sh)
{
  SH2movls4(sh, 12,0, 11);
}
#define init_1_12_0_12
void FASTCALL call_1_12_0_12(SH2_struct* sh)
{
  SH2movls4(sh, 12,0, 12);
}
#define init_1_12_0_13
void FASTCALL call_1_12_0_13(SH2_struct* sh)
{
  SH2movls4(sh, 12,0, 13);
}
#define init_1_12_0_14
void FASTCALL call_1_12_0_14(SH2_struct* sh)
{
  SH2movls4(sh, 12,0, 14);
}
#define init_1_12_0_15
void FASTCALL call_1_12_0_15(SH2_struct* sh)
{
  SH2movls4(sh, 12,0, 15);
}
#define init_1_12_1_0
void FASTCALL call_1_12_1_0(SH2_struct* sh)
{
  SH2movls4(sh, 12,1, 0);
}
#define init_1_12_1_1
void FASTCALL call_1_12_1_1(SH2_struct* sh)
{
  SH2movls4(sh, 12,1, 1);
}
#define init_1_12_1_2
void FASTCALL call_1_12_1_2(SH2_struct* sh)
{
  SH2movls4(sh, 12,1, 2);
}
#define init_1_12_1_3
void FASTCALL call_1_12_1_3(SH2_struct* sh)
{
  SH2movls4(sh, 12,1, 3);
}
#define init_1_12_1_4
void FASTCALL call_1_12_1_4(SH2_struct* sh)
{
  SH2movls4(sh, 12,1, 4);
}
#define init_1_12_1_5
void FASTCALL call_1_12_1_5(SH2_struct* sh)
{
  SH2movls4(sh, 12,1, 5);
}
#define init_1_12_1_6
void FASTCALL call_1_12_1_6(SH2_struct* sh)
{
  SH2movls4(sh, 12,1, 6);
}
#define init_1_12_1_7
void FASTCALL call_1_12_1_7(SH2_struct* sh)
{
  SH2movls4(sh, 12,1, 7);
}
#define init_1_12_1_8
void FASTCALL call_1_12_1_8(SH2_struct* sh)
{
  SH2movls4(sh, 12,1, 8);
}
#define init_1_12_1_9
void FASTCALL call_1_12_1_9(SH2_struct* sh)
{
  SH2movls4(sh, 12,1, 9);
}
#define init_1_12_1_10
void FASTCALL call_1_12_1_10(SH2_struct* sh)
{
  SH2movls4(sh, 12,1, 10);
}
#define init_1_12_1_11
void FASTCALL call_1_12_1_11(SH2_struct* sh)
{
  SH2movls4(sh, 12,1, 11);
}
#define init_1_12_1_12
void FASTCALL call_1_12_1_12(SH2_struct* sh)
{
  SH2movls4(sh, 12,1, 12);
}
#define init_1_12_1_13
void FASTCALL call_1_12_1_13(SH2_struct* sh)
{
  SH2movls4(sh, 12,1, 13);
}
#define init_1_12_1_14
void FASTCALL call_1_12_1_14(SH2_struct* sh)
{
  SH2movls4(sh, 12,1, 14);
}
#define init_1_12_1_15
void FASTCALL call_1_12_1_15(SH2_struct* sh)
{
  SH2movls4(sh, 12,1, 15);
}
#define init_1_12_2_0
void FASTCALL call_1_12_2_0(SH2_struct* sh)
{
  SH2movls4(sh, 12,2, 0);
}
#define init_1_12_2_1
void FASTCALL call_1_12_2_1(SH2_struct* sh)
{
  SH2movls4(sh, 12,2, 1);
}
#define init_1_12_2_2
void FASTCALL call_1_12_2_2(SH2_struct* sh)
{
  SH2movls4(sh, 12,2, 2);
}
#define init_1_12_2_3
void FASTCALL call_1_12_2_3(SH2_struct* sh)
{
  SH2movls4(sh, 12,2, 3);
}
#define init_1_12_2_4
void FASTCALL call_1_12_2_4(SH2_struct* sh)
{
  SH2movls4(sh, 12,2, 4);
}
#define init_1_12_2_5
void FASTCALL call_1_12_2_5(SH2_struct* sh)
{
  SH2movls4(sh, 12,2, 5);
}
#define init_1_12_2_6
void FASTCALL call_1_12_2_6(SH2_struct* sh)
{
  SH2movls4(sh, 12,2, 6);
}
#define init_1_12_2_7
void FASTCALL call_1_12_2_7(SH2_struct* sh)
{
  SH2movls4(sh, 12,2, 7);
}
#define init_1_12_2_8
void FASTCALL call_1_12_2_8(SH2_struct* sh)
{
  SH2movls4(sh, 12,2, 8);
}
#define init_1_12_2_9
void FASTCALL call_1_12_2_9(SH2_struct* sh)
{
  SH2movls4(sh, 12,2, 9);
}
#define init_1_12_2_10
void FASTCALL call_1_12_2_10(SH2_struct* sh)
{
  SH2movls4(sh, 12,2, 10);
}
#define init_1_12_2_11
void FASTCALL call_1_12_2_11(SH2_struct* sh)
{
  SH2movls4(sh, 12,2, 11);
}
#define init_1_12_2_12
void FASTCALL call_1_12_2_12(SH2_struct* sh)
{
  SH2movls4(sh, 12,2, 12);
}
#define init_1_12_2_13
void FASTCALL call_1_12_2_13(SH2_struct* sh)
{
  SH2movls4(sh, 12,2, 13);
}
#define init_1_12_2_14
void FASTCALL call_1_12_2_14(SH2_struct* sh)
{
  SH2movls4(sh, 12,2, 14);
}
#define init_1_12_2_15
void FASTCALL call_1_12_2_15(SH2_struct* sh)
{
  SH2movls4(sh, 12,2, 15);
}
#define init_1_12_3_0
void FASTCALL call_1_12_3_0(SH2_struct* sh)
{
  SH2movls4(sh, 12,3, 0);
}
#define init_1_12_3_1
void FASTCALL call_1_12_3_1(SH2_struct* sh)
{
  SH2movls4(sh, 12,3, 1);
}
#define init_1_12_3_2
void FASTCALL call_1_12_3_2(SH2_struct* sh)
{
  SH2movls4(sh, 12,3, 2);
}
#define init_1_12_3_3
void FASTCALL call_1_12_3_3(SH2_struct* sh)
{
  SH2movls4(sh, 12,3, 3);
}
#define init_1_12_3_4
void FASTCALL call_1_12_3_4(SH2_struct* sh)
{
  SH2movls4(sh, 12,3, 4);
}
#define init_1_12_3_5
void FASTCALL call_1_12_3_5(SH2_struct* sh)
{
  SH2movls4(sh, 12,3, 5);
}
#define init_1_12_3_6
void FASTCALL call_1_12_3_6(SH2_struct* sh)
{
  SH2movls4(sh, 12,3, 6);
}
#define init_1_12_3_7
void FASTCALL call_1_12_3_7(SH2_struct* sh)
{
  SH2movls4(sh, 12,3, 7);
}
#define init_1_12_3_8
void FASTCALL call_1_12_3_8(SH2_struct* sh)
{
  SH2movls4(sh, 12,3, 8);
}
#define init_1_12_3_9
void FASTCALL call_1_12_3_9(SH2_struct* sh)
{
  SH2movls4(sh, 12,3, 9);
}
#define init_1_12_3_10
void FASTCALL call_1_12_3_10(SH2_struct* sh)
{
  SH2movls4(sh, 12,3, 10);
}
#define init_1_12_3_11
void FASTCALL call_1_12_3_11(SH2_struct* sh)
{
  SH2movls4(sh, 12,3, 11);
}
#define init_1_12_3_12
void FASTCALL call_1_12_3_12(SH2_struct* sh)
{
  SH2movls4(sh, 12,3, 12);
}
#define init_1_12_3_13
void FASTCALL call_1_12_3_13(SH2_struct* sh)
{
  SH2movls4(sh, 12,3, 13);
}
#define init_1_12_3_14
void FASTCALL call_1_12_3_14(SH2_struct* sh)
{
  SH2movls4(sh, 12,3, 14);
}
#define init_1_12_3_15
void FASTCALL call_1_12_3_15(SH2_struct* sh)
{
  SH2movls4(sh, 12,3, 15);
}
#define init_1_12_4_0
void FASTCALL call_1_12_4_0(SH2_struct* sh)
{
  SH2movls4(sh, 12,4, 0);
}
#define init_1_12_4_1
void FASTCALL call_1_12_4_1(SH2_struct* sh)
{
  SH2movls4(sh, 12,4, 1);
}
#define init_1_12_4_2
void FASTCALL call_1_12_4_2(SH2_struct* sh)
{
  SH2movls4(sh, 12,4, 2);
}
#define init_1_12_4_3
void FASTCALL call_1_12_4_3(SH2_struct* sh)
{
  SH2movls4(sh, 12,4, 3);
}
#define init_1_12_4_4
void FASTCALL call_1_12_4_4(SH2_struct* sh)
{
  SH2movls4(sh, 12,4, 4);
}
#define init_1_12_4_5
void FASTCALL call_1_12_4_5(SH2_struct* sh)
{
  SH2movls4(sh, 12,4, 5);
}
#define init_1_12_4_6
void FASTCALL call_1_12_4_6(SH2_struct* sh)
{
  SH2movls4(sh, 12,4, 6);
}
#define init_1_12_4_7
void FASTCALL call_1_12_4_7(SH2_struct* sh)
{
  SH2movls4(sh, 12,4, 7);
}
#define init_1_12_4_8
void FASTCALL call_1_12_4_8(SH2_struct* sh)
{
  SH2movls4(sh, 12,4, 8);
}
#define init_1_12_4_9
void FASTCALL call_1_12_4_9(SH2_struct* sh)
{
  SH2movls4(sh, 12,4, 9);
}
#define init_1_12_4_10
void FASTCALL call_1_12_4_10(SH2_struct* sh)
{
  SH2movls4(sh, 12,4, 10);
}
#define init_1_12_4_11
void FASTCALL call_1_12_4_11(SH2_struct* sh)
{
  SH2movls4(sh, 12,4, 11);
}
#define init_1_12_4_12
void FASTCALL call_1_12_4_12(SH2_struct* sh)
{
  SH2movls4(sh, 12,4, 12);
}
#define init_1_12_4_13
void FASTCALL call_1_12_4_13(SH2_struct* sh)
{
  SH2movls4(sh, 12,4, 13);
}
#define init_1_12_4_14
void FASTCALL call_1_12_4_14(SH2_struct* sh)
{
  SH2movls4(sh, 12,4, 14);
}
#define init_1_12_4_15
void FASTCALL call_1_12_4_15(SH2_struct* sh)
{
  SH2movls4(sh, 12,4, 15);
}
#define init_1_12_5_0
void FASTCALL call_1_12_5_0(SH2_struct* sh)
{
  SH2movls4(sh, 12,5, 0);
}
#define init_1_12_5_1
void FASTCALL call_1_12_5_1(SH2_struct* sh)
{
  SH2movls4(sh, 12,5, 1);
}
#define init_1_12_5_2
void FASTCALL call_1_12_5_2(SH2_struct* sh)
{
  SH2movls4(sh, 12,5, 2);
}
#define init_1_12_5_3
void FASTCALL call_1_12_5_3(SH2_struct* sh)
{
  SH2movls4(sh, 12,5, 3);
}
#define init_1_12_5_4
void FASTCALL call_1_12_5_4(SH2_struct* sh)
{
  SH2movls4(sh, 12,5, 4);
}
#define init_1_12_5_5
void FASTCALL call_1_12_5_5(SH2_struct* sh)
{
  SH2movls4(sh, 12,5, 5);
}
#define init_1_12_5_6
void FASTCALL call_1_12_5_6(SH2_struct* sh)
{
  SH2movls4(sh, 12,5, 6);
}
#define init_1_12_5_7
void FASTCALL call_1_12_5_7(SH2_struct* sh)
{
  SH2movls4(sh, 12,5, 7);
}
#define init_1_12_5_8
void FASTCALL call_1_12_5_8(SH2_struct* sh)
{
  SH2movls4(sh, 12,5, 8);
}
#define init_1_12_5_9
void FASTCALL call_1_12_5_9(SH2_struct* sh)
{
  SH2movls4(sh, 12,5, 9);
}
#define init_1_12_5_10
void FASTCALL call_1_12_5_10(SH2_struct* sh)
{
  SH2movls4(sh, 12,5, 10);
}
#define init_1_12_5_11
void FASTCALL call_1_12_5_11(SH2_struct* sh)
{
  SH2movls4(sh, 12,5, 11);
}
#define init_1_12_5_12
void FASTCALL call_1_12_5_12(SH2_struct* sh)
{
  SH2movls4(sh, 12,5, 12);
}
#define init_1_12_5_13
void FASTCALL call_1_12_5_13(SH2_struct* sh)
{
  SH2movls4(sh, 12,5, 13);
}
#define init_1_12_5_14
void FASTCALL call_1_12_5_14(SH2_struct* sh)
{
  SH2movls4(sh, 12,5, 14);
}
#define init_1_12_5_15
void FASTCALL call_1_12_5_15(SH2_struct* sh)
{
  SH2movls4(sh, 12,5, 15);
}
#define init_1_12_6_0
void FASTCALL call_1_12_6_0(SH2_struct* sh)
{
  SH2movls4(sh, 12,6, 0);
}
#define init_1_12_6_1
void FASTCALL call_1_12_6_1(SH2_struct* sh)
{
  SH2movls4(sh, 12,6, 1);
}
#define init_1_12_6_2
void FASTCALL call_1_12_6_2(SH2_struct* sh)
{
  SH2movls4(sh, 12,6, 2);
}
#define init_1_12_6_3
void FASTCALL call_1_12_6_3(SH2_struct* sh)
{
  SH2movls4(sh, 12,6, 3);
}
#define init_1_12_6_4
void FASTCALL call_1_12_6_4(SH2_struct* sh)
{
  SH2movls4(sh, 12,6, 4);
}
#define init_1_12_6_5
void FASTCALL call_1_12_6_5(SH2_struct* sh)
{
  SH2movls4(sh, 12,6, 5);
}
#define init_1_12_6_6
void FASTCALL call_1_12_6_6(SH2_struct* sh)
{
  SH2movls4(sh, 12,6, 6);
}
#define init_1_12_6_7
void FASTCALL call_1_12_6_7(SH2_struct* sh)
{
  SH2movls4(sh, 12,6, 7);
}
#define init_1_12_6_8
void FASTCALL call_1_12_6_8(SH2_struct* sh)
{
  SH2movls4(sh, 12,6, 8);
}
#define init_1_12_6_9
void FASTCALL call_1_12_6_9(SH2_struct* sh)
{
  SH2movls4(sh, 12,6, 9);
}
#define init_1_12_6_10
void FASTCALL call_1_12_6_10(SH2_struct* sh)
{
  SH2movls4(sh, 12,6, 10);
}
#define init_1_12_6_11
void FASTCALL call_1_12_6_11(SH2_struct* sh)
{
  SH2movls4(sh, 12,6, 11);
}
#define init_1_12_6_12
void FASTCALL call_1_12_6_12(SH2_struct* sh)
{
  SH2movls4(sh, 12,6, 12);
}
#define init_1_12_6_13
void FASTCALL call_1_12_6_13(SH2_struct* sh)
{
  SH2movls4(sh, 12,6, 13);
}
#define init_1_12_6_14
void FASTCALL call_1_12_6_14(SH2_struct* sh)
{
  SH2movls4(sh, 12,6, 14);
}
#define init_1_12_6_15
void FASTCALL call_1_12_6_15(SH2_struct* sh)
{
  SH2movls4(sh, 12,6, 15);
}
#define init_1_12_7_0
void FASTCALL call_1_12_7_0(SH2_struct* sh)
{
  SH2movls4(sh, 12,7, 0);
}
#define init_1_12_7_1
void FASTCALL call_1_12_7_1(SH2_struct* sh)
{
  SH2movls4(sh, 12,7, 1);
}
#define init_1_12_7_2
void FASTCALL call_1_12_7_2(SH2_struct* sh)
{
  SH2movls4(sh, 12,7, 2);
}
#define init_1_12_7_3
void FASTCALL call_1_12_7_3(SH2_struct* sh)
{
  SH2movls4(sh, 12,7, 3);
}
#define init_1_12_7_4
void FASTCALL call_1_12_7_4(SH2_struct* sh)
{
  SH2movls4(sh, 12,7, 4);
}
#define init_1_12_7_5
void FASTCALL call_1_12_7_5(SH2_struct* sh)
{
  SH2movls4(sh, 12,7, 5);
}
#define init_1_12_7_6
void FASTCALL call_1_12_7_6(SH2_struct* sh)
{
  SH2movls4(sh, 12,7, 6);
}
#define init_1_12_7_7
void FASTCALL call_1_12_7_7(SH2_struct* sh)
{
  SH2movls4(sh, 12,7, 7);
}
#define init_1_12_7_8
void FASTCALL call_1_12_7_8(SH2_struct* sh)
{
  SH2movls4(sh, 12,7, 8);
}
#define init_1_12_7_9
void FASTCALL call_1_12_7_9(SH2_struct* sh)
{
  SH2movls4(sh, 12,7, 9);
}
#define init_1_12_7_10
void FASTCALL call_1_12_7_10(SH2_struct* sh)
{
  SH2movls4(sh, 12,7, 10);
}
#define init_1_12_7_11
void FASTCALL call_1_12_7_11(SH2_struct* sh)
{
  SH2movls4(sh, 12,7, 11);
}
#define init_1_12_7_12
void FASTCALL call_1_12_7_12(SH2_struct* sh)
{
  SH2movls4(sh, 12,7, 12);
}
#define init_1_12_7_13
void FASTCALL call_1_12_7_13(SH2_struct* sh)
{
  SH2movls4(sh, 12,7, 13);
}
#define init_1_12_7_14
void FASTCALL call_1_12_7_14(SH2_struct* sh)
{
  SH2movls4(sh, 12,7, 14);
}
#define init_1_12_7_15
void FASTCALL call_1_12_7_15(SH2_struct* sh)
{
  SH2movls4(sh, 12,7, 15);
}
#define init_1_12_8_0
void FASTCALL call_1_12_8_0(SH2_struct* sh)
{
  SH2movls4(sh, 12,8, 0);
}
#define init_1_12_8_1
void FASTCALL call_1_12_8_1(SH2_struct* sh)
{
  SH2movls4(sh, 12,8, 1);
}
#define init_1_12_8_2
void FASTCALL call_1_12_8_2(SH2_struct* sh)
{
  SH2movls4(sh, 12,8, 2);
}
#define init_1_12_8_3
void FASTCALL call_1_12_8_3(SH2_struct* sh)
{
  SH2movls4(sh, 12,8, 3);
}
#define init_1_12_8_4
void FASTCALL call_1_12_8_4(SH2_struct* sh)
{
  SH2movls4(sh, 12,8, 4);
}
#define init_1_12_8_5
void FASTCALL call_1_12_8_5(SH2_struct* sh)
{
  SH2movls4(sh, 12,8, 5);
}
#define init_1_12_8_6
void FASTCALL call_1_12_8_6(SH2_struct* sh)
{
  SH2movls4(sh, 12,8, 6);
}
#define init_1_12_8_7
void FASTCALL call_1_12_8_7(SH2_struct* sh)
{
  SH2movls4(sh, 12,8, 7);
}
#define init_1_12_8_8
void FASTCALL call_1_12_8_8(SH2_struct* sh)
{
  SH2movls4(sh, 12,8, 8);
}
#define init_1_12_8_9
void FASTCALL call_1_12_8_9(SH2_struct* sh)
{
  SH2movls4(sh, 12,8, 9);
}
#define init_1_12_8_10
void FASTCALL call_1_12_8_10(SH2_struct* sh)
{
  SH2movls4(sh, 12,8, 10);
}
#define init_1_12_8_11
void FASTCALL call_1_12_8_11(SH2_struct* sh)
{
  SH2movls4(sh, 12,8, 11);
}
#define init_1_12_8_12
void FASTCALL call_1_12_8_12(SH2_struct* sh)
{
  SH2movls4(sh, 12,8, 12);
}
#define init_1_12_8_13
void FASTCALL call_1_12_8_13(SH2_struct* sh)
{
  SH2movls4(sh, 12,8, 13);
}
#define init_1_12_8_14
void FASTCALL call_1_12_8_14(SH2_struct* sh)
{
  SH2movls4(sh, 12,8, 14);
}
#define init_1_12_8_15
void FASTCALL call_1_12_8_15(SH2_struct* sh)
{
  SH2movls4(sh, 12,8, 15);
}
#define init_1_12_9_0
void FASTCALL call_1_12_9_0(SH2_struct* sh)
{
  SH2movls4(sh, 12,9, 0);
}
#define init_1_12_9_1
void FASTCALL call_1_12_9_1(SH2_struct* sh)
{
  SH2movls4(sh, 12,9, 1);
}
#define init_1_12_9_2
void FASTCALL call_1_12_9_2(SH2_struct* sh)
{
  SH2movls4(sh, 12,9, 2);
}
#define init_1_12_9_3
void FASTCALL call_1_12_9_3(SH2_struct* sh)
{
  SH2movls4(sh, 12,9, 3);
}
#define init_1_12_9_4
void FASTCALL call_1_12_9_4(SH2_struct* sh)
{
  SH2movls4(sh, 12,9, 4);
}
#define init_1_12_9_5
void FASTCALL call_1_12_9_5(SH2_struct* sh)
{
  SH2movls4(sh, 12,9, 5);
}
#define init_1_12_9_6
void FASTCALL call_1_12_9_6(SH2_struct* sh)
{
  SH2movls4(sh, 12,9, 6);
}
#define init_1_12_9_7
void FASTCALL call_1_12_9_7(SH2_struct* sh)
{
  SH2movls4(sh, 12,9, 7);
}
#define init_1_12_9_8
void FASTCALL call_1_12_9_8(SH2_struct* sh)
{
  SH2movls4(sh, 12,9, 8);
}
#define init_1_12_9_9
void FASTCALL call_1_12_9_9(SH2_struct* sh)
{
  SH2movls4(sh, 12,9, 9);
}
#define init_1_12_9_10
void FASTCALL call_1_12_9_10(SH2_struct* sh)
{
  SH2movls4(sh, 12,9, 10);
}
#define init_1_12_9_11
void FASTCALL call_1_12_9_11(SH2_struct* sh)
{
  SH2movls4(sh, 12,9, 11);
}
#define init_1_12_9_12
void FASTCALL call_1_12_9_12(SH2_struct* sh)
{
  SH2movls4(sh, 12,9, 12);
}
#define init_1_12_9_13
void FASTCALL call_1_12_9_13(SH2_struct* sh)
{
  SH2movls4(sh, 12,9, 13);
}
#define init_1_12_9_14
void FASTCALL call_1_12_9_14(SH2_struct* sh)
{
  SH2movls4(sh, 12,9, 14);
}
#define init_1_12_9_15
void FASTCALL call_1_12_9_15(SH2_struct* sh)
{
  SH2movls4(sh, 12,9, 15);
}
#define init_1_12_10_0
void FASTCALL call_1_12_10_0(SH2_struct* sh)
{
  SH2movls4(sh, 12,10, 0);
}
#define init_1_12_10_1
void FASTCALL call_1_12_10_1(SH2_struct* sh)
{
  SH2movls4(sh, 12,10, 1);
}
#define init_1_12_10_2
void FASTCALL call_1_12_10_2(SH2_struct* sh)
{
  SH2movls4(sh, 12,10, 2);
}
#define init_1_12_10_3
void FASTCALL call_1_12_10_3(SH2_struct* sh)
{
  SH2movls4(sh, 12,10, 3);
}
#define init_1_12_10_4
void FASTCALL call_1_12_10_4(SH2_struct* sh)
{
  SH2movls4(sh, 12,10, 4);
}
#define init_1_12_10_5
void FASTCALL call_1_12_10_5(SH2_struct* sh)
{
  SH2movls4(sh, 12,10, 5);
}
#define init_1_12_10_6
void FASTCALL call_1_12_10_6(SH2_struct* sh)
{
  SH2movls4(sh, 12,10, 6);
}
#define init_1_12_10_7
void FASTCALL call_1_12_10_7(SH2_struct* sh)
{
  SH2movls4(sh, 12,10, 7);
}
#define init_1_12_10_8
void FASTCALL call_1_12_10_8(SH2_struct* sh)
{
  SH2movls4(sh, 12,10, 8);
}
#define init_1_12_10_9
void FASTCALL call_1_12_10_9(SH2_struct* sh)
{
  SH2movls4(sh, 12,10, 9);
}
#define init_1_12_10_10
void FASTCALL call_1_12_10_10(SH2_struct* sh)
{
  SH2movls4(sh, 12,10, 10);
}
#define init_1_12_10_11
void FASTCALL call_1_12_10_11(SH2_struct* sh)
{
  SH2movls4(sh, 12,10, 11);
}
#define init_1_12_10_12
void FASTCALL call_1_12_10_12(SH2_struct* sh)
{
  SH2movls4(sh, 12,10, 12);
}
#define init_1_12_10_13
void FASTCALL call_1_12_10_13(SH2_struct* sh)
{
  SH2movls4(sh, 12,10, 13);
}
#define init_1_12_10_14
void FASTCALL call_1_12_10_14(SH2_struct* sh)
{
  SH2movls4(sh, 12,10, 14);
}
#define init_1_12_10_15
void FASTCALL call_1_12_10_15(SH2_struct* sh)
{
  SH2movls4(sh, 12,10, 15);
}
#define init_1_12_11_0
void FASTCALL call_1_12_11_0(SH2_struct* sh)
{
  SH2movls4(sh, 12,11, 0);
}
#define init_1_12_11_1
void FASTCALL call_1_12_11_1(SH2_struct* sh)
{
  SH2movls4(sh, 12,11, 1);
}
#define init_1_12_11_2
void FASTCALL call_1_12_11_2(SH2_struct* sh)
{
  SH2movls4(sh, 12,11, 2);
}
#define init_1_12_11_3
void FASTCALL call_1_12_11_3(SH2_struct* sh)
{
  SH2movls4(sh, 12,11, 3);
}
#define init_1_12_11_4
void FASTCALL call_1_12_11_4(SH2_struct* sh)
{
  SH2movls4(sh, 12,11, 4);
}
#define init_1_12_11_5
void FASTCALL call_1_12_11_5(SH2_struct* sh)
{
  SH2movls4(sh, 12,11, 5);
}
#define init_1_12_11_6
void FASTCALL call_1_12_11_6(SH2_struct* sh)
{
  SH2movls4(sh, 12,11, 6);
}
#define init_1_12_11_7
void FASTCALL call_1_12_11_7(SH2_struct* sh)
{
  SH2movls4(sh, 12,11, 7);
}
#define init_1_12_11_8
void FASTCALL call_1_12_11_8(SH2_struct* sh)
{
  SH2movls4(sh, 12,11, 8);
}
#define init_1_12_11_9
void FASTCALL call_1_12_11_9(SH2_struct* sh)
{
  SH2movls4(sh, 12,11, 9);
}
#define init_1_12_11_10
void FASTCALL call_1_12_11_10(SH2_struct* sh)
{
  SH2movls4(sh, 12,11, 10);
}
#define init_1_12_11_11
void FASTCALL call_1_12_11_11(SH2_struct* sh)
{
  SH2movls4(sh, 12,11, 11);
}
#define init_1_12_11_12
void FASTCALL call_1_12_11_12(SH2_struct* sh)
{
  SH2movls4(sh, 12,11, 12);
}
#define init_1_12_11_13
void FASTCALL call_1_12_11_13(SH2_struct* sh)
{
  SH2movls4(sh, 12,11, 13);
}
#define init_1_12_11_14
void FASTCALL call_1_12_11_14(SH2_struct* sh)
{
  SH2movls4(sh, 12,11, 14);
}
#define init_1_12_11_15
void FASTCALL call_1_12_11_15(SH2_struct* sh)
{
  SH2movls4(sh, 12,11, 15);
}
#define init_1_12_12_0
void FASTCALL call_1_12_12_0(SH2_struct* sh)
{
  SH2movls4(sh, 12,12, 0);
}
#define init_1_12_12_1
void FASTCALL call_1_12_12_1(SH2_struct* sh)
{
  SH2movls4(sh, 12,12, 1);
}
#define init_1_12_12_2
void FASTCALL call_1_12_12_2(SH2_struct* sh)
{
  SH2movls4(sh, 12,12, 2);
}
#define init_1_12_12_3
void FASTCALL call_1_12_12_3(SH2_struct* sh)
{
  SH2movls4(sh, 12,12, 3);
}
#define init_1_12_12_4
void FASTCALL call_1_12_12_4(SH2_struct* sh)
{
  SH2movls4(sh, 12,12, 4);
}
#define init_1_12_12_5
void FASTCALL call_1_12_12_5(SH2_struct* sh)
{
  SH2movls4(sh, 12,12, 5);
}
#define init_1_12_12_6
void FASTCALL call_1_12_12_6(SH2_struct* sh)
{
  SH2movls4(sh, 12,12, 6);
}
#define init_1_12_12_7
void FASTCALL call_1_12_12_7(SH2_struct* sh)
{
  SH2movls4(sh, 12,12, 7);
}
#define init_1_12_12_8
void FASTCALL call_1_12_12_8(SH2_struct* sh)
{
  SH2movls4(sh, 12,12, 8);
}
#define init_1_12_12_9
void FASTCALL call_1_12_12_9(SH2_struct* sh)
{
  SH2movls4(sh, 12,12, 9);
}
#define init_1_12_12_10
void FASTCALL call_1_12_12_10(SH2_struct* sh)
{
  SH2movls4(sh, 12,12, 10);
}
#define init_1_12_12_11
void FASTCALL call_1_12_12_11(SH2_struct* sh)
{
  SH2movls4(sh, 12,12, 11);
}
#define init_1_12_12_12
void FASTCALL call_1_12_12_12(SH2_struct* sh)
{
  SH2movls4(sh, 12,12, 12);
}
#define init_1_12_12_13
void FASTCALL call_1_12_12_13(SH2_struct* sh)
{
  SH2movls4(sh, 12,12, 13);
}
#define init_1_12_12_14
void FASTCALL call_1_12_12_14(SH2_struct* sh)
{
  SH2movls4(sh, 12,12, 14);
}
#define init_1_12_12_15
void FASTCALL call_1_12_12_15(SH2_struct* sh)
{
  SH2movls4(sh, 12,12, 15);
}
#define init_1_12_13_0
void FASTCALL call_1_12_13_0(SH2_struct* sh)
{
  SH2movls4(sh, 12,13, 0);
}
#define init_1_12_13_1
void FASTCALL call_1_12_13_1(SH2_struct* sh)
{
  SH2movls4(sh, 12,13, 1);
}
#define init_1_12_13_2
void FASTCALL call_1_12_13_2(SH2_struct* sh)
{
  SH2movls4(sh, 12,13, 2);
}
#define init_1_12_13_3
void FASTCALL call_1_12_13_3(SH2_struct* sh)
{
  SH2movls4(sh, 12,13, 3);
}
#define init_1_12_13_4
void FASTCALL call_1_12_13_4(SH2_struct* sh)
{
  SH2movls4(sh, 12,13, 4);
}
#define init_1_12_13_5
void FASTCALL call_1_12_13_5(SH2_struct* sh)
{
  SH2movls4(sh, 12,13, 5);
}
#define init_1_12_13_6
void FASTCALL call_1_12_13_6(SH2_struct* sh)
{
  SH2movls4(sh, 12,13, 6);
}
#define init_1_12_13_7
void FASTCALL call_1_12_13_7(SH2_struct* sh)
{
  SH2movls4(sh, 12,13, 7);
}
#define init_1_12_13_8
void FASTCALL call_1_12_13_8(SH2_struct* sh)
{
  SH2movls4(sh, 12,13, 8);
}
#define init_1_12_13_9
void FASTCALL call_1_12_13_9(SH2_struct* sh)
{
  SH2movls4(sh, 12,13, 9);
}
#define init_1_12_13_10
void FASTCALL call_1_12_13_10(SH2_struct* sh)
{
  SH2movls4(sh, 12,13, 10);
}
#define init_1_12_13_11
void FASTCALL call_1_12_13_11(SH2_struct* sh)
{
  SH2movls4(sh, 12,13, 11);
}
#define init_1_12_13_12
void FASTCALL call_1_12_13_12(SH2_struct* sh)
{
  SH2movls4(sh, 12,13, 12);
}
#define init_1_12_13_13
void FASTCALL call_1_12_13_13(SH2_struct* sh)
{
  SH2movls4(sh, 12,13, 13);
}
#define init_1_12_13_14
void FASTCALL call_1_12_13_14(SH2_struct* sh)
{
  SH2movls4(sh, 12,13, 14);
}
#define init_1_12_13_15
void FASTCALL call_1_12_13_15(SH2_struct* sh)
{
  SH2movls4(sh, 12,13, 15);
}
#define init_1_12_14_0
void FASTCALL call_1_12_14_0(SH2_struct* sh)
{
  SH2movls4(sh, 12,14, 0);
}
#define init_1_12_14_1
void FASTCALL call_1_12_14_1(SH2_struct* sh)
{
  SH2movls4(sh, 12,14, 1);
}
#define init_1_12_14_2
void FASTCALL call_1_12_14_2(SH2_struct* sh)
{
  SH2movls4(sh, 12,14, 2);
}
#define init_1_12_14_3
void FASTCALL call_1_12_14_3(SH2_struct* sh)
{
  SH2movls4(sh, 12,14, 3);
}
#define init_1_12_14_4
void FASTCALL call_1_12_14_4(SH2_struct* sh)
{
  SH2movls4(sh, 12,14, 4);
}
#define init_1_12_14_5
void FASTCALL call_1_12_14_5(SH2_struct* sh)
{
  SH2movls4(sh, 12,14, 5);
}
#define init_1_12_14_6
void FASTCALL call_1_12_14_6(SH2_struct* sh)
{
  SH2movls4(sh, 12,14, 6);
}
#define init_1_12_14_7
void FASTCALL call_1_12_14_7(SH2_struct* sh)
{
  SH2movls4(sh, 12,14, 7);
}
#define init_1_12_14_8
void FASTCALL call_1_12_14_8(SH2_struct* sh)
{
  SH2movls4(sh, 12,14, 8);
}
#define init_1_12_14_9
void FASTCALL call_1_12_14_9(SH2_struct* sh)
{
  SH2movls4(sh, 12,14, 9);
}
#define init_1_12_14_10
void FASTCALL call_1_12_14_10(SH2_struct* sh)
{
  SH2movls4(sh, 12,14, 10);
}
#define init_1_12_14_11
void FASTCALL call_1_12_14_11(SH2_struct* sh)
{
  SH2movls4(sh, 12,14, 11);
}
#define init_1_12_14_12
void FASTCALL call_1_12_14_12(SH2_struct* sh)
{
  SH2movls4(sh, 12,14, 12);
}
#define init_1_12_14_13
void FASTCALL call_1_12_14_13(SH2_struct* sh)
{
  SH2movls4(sh, 12,14, 13);
}
#define init_1_12_14_14
void FASTCALL call_1_12_14_14(SH2_struct* sh)
{
  SH2movls4(sh, 12,14, 14);
}
#define init_1_12_14_15
void FASTCALL call_1_12_14_15(SH2_struct* sh)
{
  SH2movls4(sh, 12,14, 15);
}
#define init_1_12_15_0
void FASTCALL call_1_12_15_0(SH2_struct* sh)
{
  SH2movls4(sh, 12,15, 0);
}
#define init_1_12_15_1
void FASTCALL call_1_12_15_1(SH2_struct* sh)
{
  SH2movls4(sh, 12,15, 1);
}
#define init_1_12_15_2
void FASTCALL call_1_12_15_2(SH2_struct* sh)
{
  SH2movls4(sh, 12,15, 2);
}
#define init_1_12_15_3
void FASTCALL call_1_12_15_3(SH2_struct* sh)
{
  SH2movls4(sh, 12,15, 3);
}
#define init_1_12_15_4
void FASTCALL call_1_12_15_4(SH2_struct* sh)
{
  SH2movls4(sh, 12,15, 4);
}
#define init_1_12_15_5
void FASTCALL call_1_12_15_5(SH2_struct* sh)
{
  SH2movls4(sh, 12,15, 5);
}
#define init_1_12_15_6
void FASTCALL call_1_12_15_6(SH2_struct* sh)
{
  SH2movls4(sh, 12,15, 6);
}
#define init_1_12_15_7
void FASTCALL call_1_12_15_7(SH2_struct* sh)
{
  SH2movls4(sh, 12,15, 7);
}
#define init_1_12_15_8
void FASTCALL call_1_12_15_8(SH2_struct* sh)
{
  SH2movls4(sh, 12,15, 8);
}
#define init_1_12_15_9
void FASTCALL call_1_12_15_9(SH2_struct* sh)
{
  SH2movls4(sh, 12,15, 9);
}
#define init_1_12_15_10
void FASTCALL call_1_12_15_10(SH2_struct* sh)
{
  SH2movls4(sh, 12,15, 10);
}
#define init_1_12_15_11
void FASTCALL call_1_12_15_11(SH2_struct* sh)
{
  SH2movls4(sh, 12,15, 11);
}
#define init_1_12_15_12
void FASTCALL call_1_12_15_12(SH2_struct* sh)
{
  SH2movls4(sh, 12,15, 12);
}
#define init_1_12_15_13
void FASTCALL call_1_12_15_13(SH2_struct* sh)
{
  SH2movls4(sh, 12,15, 13);
}
#define init_1_12_15_14
void FASTCALL call_1_12_15_14(SH2_struct* sh)
{
  SH2movls4(sh, 12,15, 14);
}
#define init_1_12_15_15
void FASTCALL call_1_12_15_15(SH2_struct* sh)
{
  SH2movls4(sh, 12,15, 15);
}
#define init_1_13_0_0
void FASTCALL call_1_13_0_0(SH2_struct* sh)
{
  SH2movls4(sh, 13,0, 0);
}
#define init_1_13_0_1
void FASTCALL call_1_13_0_1(SH2_struct* sh)
{
  SH2movls4(sh, 13,0, 1);
}
#define init_1_13_0_2
void FASTCALL call_1_13_0_2(SH2_struct* sh)
{
  SH2movls4(sh, 13,0, 2);
}
#define init_1_13_0_3
void FASTCALL call_1_13_0_3(SH2_struct* sh)
{
  SH2movls4(sh, 13,0, 3);
}
#define init_1_13_0_4
void FASTCALL call_1_13_0_4(SH2_struct* sh)
{
  SH2movls4(sh, 13,0, 4);
}
#define init_1_13_0_5
void FASTCALL call_1_13_0_5(SH2_struct* sh)
{
  SH2movls4(sh, 13,0, 5);
}
#define init_1_13_0_6
void FASTCALL call_1_13_0_6(SH2_struct* sh)
{
  SH2movls4(sh, 13,0, 6);
}
#define init_1_13_0_7
void FASTCALL call_1_13_0_7(SH2_struct* sh)
{
  SH2movls4(sh, 13,0, 7);
}
#define init_1_13_0_8
void FASTCALL call_1_13_0_8(SH2_struct* sh)
{
  SH2movls4(sh, 13,0, 8);
}
#define init_1_13_0_9
void FASTCALL call_1_13_0_9(SH2_struct* sh)
{
  SH2movls4(sh, 13,0, 9);
}
#define init_1_13_0_10
void FASTCALL call_1_13_0_10(SH2_struct* sh)
{
  SH2movls4(sh, 13,0, 10);
}
#define init_1_13_0_11
void FASTCALL call_1_13_0_11(SH2_struct* sh)
{
  SH2movls4(sh, 13,0, 11);
}
#define init_1_13_0_12
void FASTCALL call_1_13_0_12(SH2_struct* sh)
{
  SH2movls4(sh, 13,0, 12);
}
#define init_1_13_0_13
void FASTCALL call_1_13_0_13(SH2_struct* sh)
{
  SH2movls4(sh, 13,0, 13);
}
#define init_1_13_0_14
void FASTCALL call_1_13_0_14(SH2_struct* sh)
{
  SH2movls4(sh, 13,0, 14);
}
#define init_1_13_0_15
void FASTCALL call_1_13_0_15(SH2_struct* sh)
{
  SH2movls4(sh, 13,0, 15);
}
#define init_1_13_1_0
void FASTCALL call_1_13_1_0(SH2_struct* sh)
{
  SH2movls4(sh, 13,1, 0);
}
#define init_1_13_1_1
void FASTCALL call_1_13_1_1(SH2_struct* sh)
{
  SH2movls4(sh, 13,1, 1);
}
#define init_1_13_1_2
void FASTCALL call_1_13_1_2(SH2_struct* sh)
{
  SH2movls4(sh, 13,1, 2);
}
#define init_1_13_1_3
void FASTCALL call_1_13_1_3(SH2_struct* sh)
{
  SH2movls4(sh, 13,1, 3);
}
#define init_1_13_1_4
void FASTCALL call_1_13_1_4(SH2_struct* sh)
{
  SH2movls4(sh, 13,1, 4);
}
#define init_1_13_1_5
void FASTCALL call_1_13_1_5(SH2_struct* sh)
{
  SH2movls4(sh, 13,1, 5);
}
#define init_1_13_1_6
void FASTCALL call_1_13_1_6(SH2_struct* sh)
{
  SH2movls4(sh, 13,1, 6);
}
#define init_1_13_1_7
void FASTCALL call_1_13_1_7(SH2_struct* sh)
{
  SH2movls4(sh, 13,1, 7);
}
#define init_1_13_1_8
void FASTCALL call_1_13_1_8(SH2_struct* sh)
{
  SH2movls4(sh, 13,1, 8);
}
#define init_1_13_1_9
void FASTCALL call_1_13_1_9(SH2_struct* sh)
{
  SH2movls4(sh, 13,1, 9);
}
#define init_1_13_1_10
void FASTCALL call_1_13_1_10(SH2_struct* sh)
{
  SH2movls4(sh, 13,1, 10);
}
#define init_1_13_1_11
void FASTCALL call_1_13_1_11(SH2_struct* sh)
{
  SH2movls4(sh, 13,1, 11);
}
#define init_1_13_1_12
void FASTCALL call_1_13_1_12(SH2_struct* sh)
{
  SH2movls4(sh, 13,1, 12);
}
#define init_1_13_1_13
void FASTCALL call_1_13_1_13(SH2_struct* sh)
{
  SH2movls4(sh, 13,1, 13);
}
#define init_1_13_1_14
void FASTCALL call_1_13_1_14(SH2_struct* sh)
{
  SH2movls4(sh, 13,1, 14);
}
#define init_1_13_1_15
void FASTCALL call_1_13_1_15(SH2_struct* sh)
{
  SH2movls4(sh, 13,1, 15);
}
#define init_1_13_2_0
void FASTCALL call_1_13_2_0(SH2_struct* sh)
{
  SH2movls4(sh, 13,2, 0);
}
#define init_1_13_2_1
void FASTCALL call_1_13_2_1(SH2_struct* sh)
{
  SH2movls4(sh, 13,2, 1);
}
#define init_1_13_2_2
void FASTCALL call_1_13_2_2(SH2_struct* sh)
{
  SH2movls4(sh, 13,2, 2);
}
#define init_1_13_2_3
void FASTCALL call_1_13_2_3(SH2_struct* sh)
{
  SH2movls4(sh, 13,2, 3);
}
#define init_1_13_2_4
void FASTCALL call_1_13_2_4(SH2_struct* sh)
{
  SH2movls4(sh, 13,2, 4);
}
#define init_1_13_2_5
void FASTCALL call_1_13_2_5(SH2_struct* sh)
{
  SH2movls4(sh, 13,2, 5);
}
#define init_1_13_2_6
void FASTCALL call_1_13_2_6(SH2_struct* sh)
{
  SH2movls4(sh, 13,2, 6);
}
#define init_1_13_2_7
void FASTCALL call_1_13_2_7(SH2_struct* sh)
{
  SH2movls4(sh, 13,2, 7);
}
#define init_1_13_2_8
void FASTCALL call_1_13_2_8(SH2_struct* sh)
{
  SH2movls4(sh, 13,2, 8);
}
#define init_1_13_2_9
void FASTCALL call_1_13_2_9(SH2_struct* sh)
{
  SH2movls4(sh, 13,2, 9);
}
#define init_1_13_2_10
void FASTCALL call_1_13_2_10(SH2_struct* sh)
{
  SH2movls4(sh, 13,2, 10);
}
#define init_1_13_2_11
void FASTCALL call_1_13_2_11(SH2_struct* sh)
{
  SH2movls4(sh, 13,2, 11);
}
#define init_1_13_2_12
void FASTCALL call_1_13_2_12(SH2_struct* sh)
{
  SH2movls4(sh, 13,2, 12);
}
#define init_1_13_2_13
void FASTCALL call_1_13_2_13(SH2_struct* sh)
{
  SH2movls4(sh, 13,2, 13);
}
#define init_1_13_2_14
void FASTCALL call_1_13_2_14(SH2_struct* sh)
{
  SH2movls4(sh, 13,2, 14);
}
#define init_1_13_2_15
void FASTCALL call_1_13_2_15(SH2_struct* sh)
{
  SH2movls4(sh, 13,2, 15);
}
#define init_1_13_3_0
void FASTCALL call_1_13_3_0(SH2_struct* sh)
{
  SH2movls4(sh, 13,3, 0);
}
#define init_1_13_3_1
void FASTCALL call_1_13_3_1(SH2_struct* sh)
{
  SH2movls4(sh, 13,3, 1);
}
#define init_1_13_3_2
void FASTCALL call_1_13_3_2(SH2_struct* sh)
{
  SH2movls4(sh, 13,3, 2);
}
#define init_1_13_3_3
void FASTCALL call_1_13_3_3(SH2_struct* sh)
{
  SH2movls4(sh, 13,3, 3);
}
#define init_1_13_3_4
void FASTCALL call_1_13_3_4(SH2_struct* sh)
{
  SH2movls4(sh, 13,3, 4);
}
#define init_1_13_3_5
void FASTCALL call_1_13_3_5(SH2_struct* sh)
{
  SH2movls4(sh, 13,3, 5);
}
#define init_1_13_3_6
void FASTCALL call_1_13_3_6(SH2_struct* sh)
{
  SH2movls4(sh, 13,3, 6);
}
#define init_1_13_3_7
void FASTCALL call_1_13_3_7(SH2_struct* sh)
{
  SH2movls4(sh, 13,3, 7);
}
#define init_1_13_3_8
void FASTCALL call_1_13_3_8(SH2_struct* sh)
{
  SH2movls4(sh, 13,3, 8);
}
#define init_1_13_3_9
void FASTCALL call_1_13_3_9(SH2_struct* sh)
{
  SH2movls4(sh, 13,3, 9);
}
#define init_1_13_3_10
void FASTCALL call_1_13_3_10(SH2_struct* sh)
{
  SH2movls4(sh, 13,3, 10);
}
#define init_1_13_3_11
void FASTCALL call_1_13_3_11(SH2_struct* sh)
{
  SH2movls4(sh, 13,3, 11);
}
#define init_1_13_3_12
void FASTCALL call_1_13_3_12(SH2_struct* sh)
{
  SH2movls4(sh, 13,3, 12);
}
#define init_1_13_3_13
void FASTCALL call_1_13_3_13(SH2_struct* sh)
{
  SH2movls4(sh, 13,3, 13);
}
#define init_1_13_3_14
void FASTCALL call_1_13_3_14(SH2_struct* sh)
{
  SH2movls4(sh, 13,3, 14);
}
#define init_1_13_3_15
void FASTCALL call_1_13_3_15(SH2_struct* sh)
{
  SH2movls4(sh, 13,3, 15);
}
#define init_1_13_4_0
void FASTCALL call_1_13_4_0(SH2_struct* sh)
{
  SH2movls4(sh, 13,4, 0);
}
#define init_1_13_4_1
void FASTCALL call_1_13_4_1(SH2_struct* sh)
{
  SH2movls4(sh, 13,4, 1);
}
#define init_1_13_4_2
void FASTCALL call_1_13_4_2(SH2_struct* sh)
{
  SH2movls4(sh, 13,4, 2);
}
#define init_1_13_4_3
void FASTCALL call_1_13_4_3(SH2_struct* sh)
{
  SH2movls4(sh, 13,4, 3);
}
#define init_1_13_4_4
void FASTCALL call_1_13_4_4(SH2_struct* sh)
{
  SH2movls4(sh, 13,4, 4);
}
#define init_1_13_4_5
void FASTCALL call_1_13_4_5(SH2_struct* sh)
{
  SH2movls4(sh, 13,4, 5);
}
#define init_1_13_4_6
void FASTCALL call_1_13_4_6(SH2_struct* sh)
{
  SH2movls4(sh, 13,4, 6);
}
#define init_1_13_4_7
void FASTCALL call_1_13_4_7(SH2_struct* sh)
{
  SH2movls4(sh, 13,4, 7);
}
#define init_1_13_4_8
void FASTCALL call_1_13_4_8(SH2_struct* sh)
{
  SH2movls4(sh, 13,4, 8);
}
#define init_1_13_4_9
void FASTCALL call_1_13_4_9(SH2_struct* sh)
{
  SH2movls4(sh, 13,4, 9);
}
#define init_1_13_4_10
void FASTCALL call_1_13_4_10(SH2_struct* sh)
{
  SH2movls4(sh, 13,4, 10);
}
#define init_1_13_4_11
void FASTCALL call_1_13_4_11(SH2_struct* sh)
{
  SH2movls4(sh, 13,4, 11);
}
#define init_1_13_4_12
void FASTCALL call_1_13_4_12(SH2_struct* sh)
{
  SH2movls4(sh, 13,4, 12);
}
#define init_1_13_4_13
void FASTCALL call_1_13_4_13(SH2_struct* sh)
{
  SH2movls4(sh, 13,4, 13);
}
#define init_1_13_4_14
void FASTCALL call_1_13_4_14(SH2_struct* sh)
{
  SH2movls4(sh, 13,4, 14);
}
#define init_1_13_4_15
void FASTCALL call_1_13_4_15(SH2_struct* sh)
{
  SH2movls4(sh, 13,4, 15);
}
#define init_1_13_5_0
void FASTCALL call_1_13_5_0(SH2_struct* sh)
{
  SH2movls4(sh, 13,5, 0);
}
#define init_1_13_5_1
void FASTCALL call_1_13_5_1(SH2_struct* sh)
{
  SH2movls4(sh, 13,5, 1);
}
#define init_1_13_5_2
void FASTCALL call_1_13_5_2(SH2_struct* sh)
{
  SH2movls4(sh, 13,5, 2);
}
#define init_1_13_5_3
void FASTCALL call_1_13_5_3(SH2_struct* sh)
{
  SH2movls4(sh, 13,5, 3);
}
#define init_1_13_5_4
void FASTCALL call_1_13_5_4(SH2_struct* sh)
{
  SH2movls4(sh, 13,5, 4);
}
#define init_1_13_5_5
void FASTCALL call_1_13_5_5(SH2_struct* sh)
{
  SH2movls4(sh, 13,5, 5);
}
#define init_1_13_5_6
void FASTCALL call_1_13_5_6(SH2_struct* sh)
{
  SH2movls4(sh, 13,5, 6);
}
#define init_1_13_5_7
void FASTCALL call_1_13_5_7(SH2_struct* sh)
{
  SH2movls4(sh, 13,5, 7);
}
#define init_1_13_5_8
void FASTCALL call_1_13_5_8(SH2_struct* sh)
{
  SH2movls4(sh, 13,5, 8);
}
#define init_1_13_5_9
void FASTCALL call_1_13_5_9(SH2_struct* sh)
{
  SH2movls4(sh, 13,5, 9);
}
#define init_1_13_5_10
void FASTCALL call_1_13_5_10(SH2_struct* sh)
{
  SH2movls4(sh, 13,5, 10);
}
#define init_1_13_5_11
void FASTCALL call_1_13_5_11(SH2_struct* sh)
{
  SH2movls4(sh, 13,5, 11);
}
#define init_1_13_5_12
void FASTCALL call_1_13_5_12(SH2_struct* sh)
{
  SH2movls4(sh, 13,5, 12);
}
#define init_1_13_5_13
void FASTCALL call_1_13_5_13(SH2_struct* sh)
{
  SH2movls4(sh, 13,5, 13);
}
#define init_1_13_5_14
void FASTCALL call_1_13_5_14(SH2_struct* sh)
{
  SH2movls4(sh, 13,5, 14);
}
#define init_1_13_5_15
void FASTCALL call_1_13_5_15(SH2_struct* sh)
{
  SH2movls4(sh, 13,5, 15);
}
#define init_1_13_6_0
void FASTCALL call_1_13_6_0(SH2_struct* sh)
{
  SH2movls4(sh, 13,6, 0);
}
#define init_1_13_6_1
void FASTCALL call_1_13_6_1(SH2_struct* sh)
{
  SH2movls4(sh, 13,6, 1);
}
#define init_1_13_6_2
void FASTCALL call_1_13_6_2(SH2_struct* sh)
{
  SH2movls4(sh, 13,6, 2);
}
#define init_1_13_6_3
void FASTCALL call_1_13_6_3(SH2_struct* sh)
{
  SH2movls4(sh, 13,6, 3);
}
#define init_1_13_6_4
void FASTCALL call_1_13_6_4(SH2_struct* sh)
{
  SH2movls4(sh, 13,6, 4);
}
#define init_1_13_6_5
void FASTCALL call_1_13_6_5(SH2_struct* sh)
{
  SH2movls4(sh, 13,6, 5);
}
#define init_1_13_6_6
void FASTCALL call_1_13_6_6(SH2_struct* sh)
{
  SH2movls4(sh, 13,6, 6);
}
#define init_1_13_6_7
void FASTCALL call_1_13_6_7(SH2_struct* sh)
{
  SH2movls4(sh, 13,6, 7);
}
#define init_1_13_6_8
void FASTCALL call_1_13_6_8(SH2_struct* sh)
{
  SH2movls4(sh, 13,6, 8);
}
#define init_1_13_6_9
void FASTCALL call_1_13_6_9(SH2_struct* sh)
{
  SH2movls4(sh, 13,6, 9);
}
#define init_1_13_6_10
void FASTCALL call_1_13_6_10(SH2_struct* sh)
{
  SH2movls4(sh, 13,6, 10);
}
#define init_1_13_6_11
void FASTCALL call_1_13_6_11(SH2_struct* sh)
{
  SH2movls4(sh, 13,6, 11);
}
#define init_1_13_6_12
void FASTCALL call_1_13_6_12(SH2_struct* sh)
{
  SH2movls4(sh, 13,6, 12);
}
#define init_1_13_6_13
void FASTCALL call_1_13_6_13(SH2_struct* sh)
{
  SH2movls4(sh, 13,6, 13);
}
#define init_1_13_6_14
void FASTCALL call_1_13_6_14(SH2_struct* sh)
{
  SH2movls4(sh, 13,6, 14);
}
#define init_1_13_6_15
void FASTCALL call_1_13_6_15(SH2_struct* sh)
{
  SH2movls4(sh, 13,6, 15);
}
#define init_1_13_7_0
void FASTCALL call_1_13_7_0(SH2_struct* sh)
{
  SH2movls4(sh, 13,7, 0);
}
#define init_1_13_7_1
void FASTCALL call_1_13_7_1(SH2_struct* sh)
{
  SH2movls4(sh, 13,7, 1);
}
#define init_1_13_7_2
void FASTCALL call_1_13_7_2(SH2_struct* sh)
{
  SH2movls4(sh, 13,7, 2);
}
#define init_1_13_7_3
void FASTCALL call_1_13_7_3(SH2_struct* sh)
{
  SH2movls4(sh, 13,7, 3);
}
#define init_1_13_7_4
void FASTCALL call_1_13_7_4(SH2_struct* sh)
{
  SH2movls4(sh, 13,7, 4);
}
#define init_1_13_7_5
void FASTCALL call_1_13_7_5(SH2_struct* sh)
{
  SH2movls4(sh, 13,7, 5);
}
#define init_1_13_7_6
void FASTCALL call_1_13_7_6(SH2_struct* sh)
{
  SH2movls4(sh, 13,7, 6);
}
#define init_1_13_7_7
void FASTCALL call_1_13_7_7(SH2_struct* sh)
{
  SH2movls4(sh, 13,7, 7);
}
#define init_1_13_7_8
void FASTCALL call_1_13_7_8(SH2_struct* sh)
{
  SH2movls4(sh, 13,7, 8);
}
#define init_1_13_7_9
void FASTCALL call_1_13_7_9(SH2_struct* sh)
{
  SH2movls4(sh, 13,7, 9);
}
#define init_1_13_7_10
void FASTCALL call_1_13_7_10(SH2_struct* sh)
{
  SH2movls4(sh, 13,7, 10);
}
#define init_1_13_7_11
void FASTCALL call_1_13_7_11(SH2_struct* sh)
{
  SH2movls4(sh, 13,7, 11);
}
#define init_1_13_7_12
void FASTCALL call_1_13_7_12(SH2_struct* sh)
{
  SH2movls4(sh, 13,7, 12);
}
#define init_1_13_7_13
void FASTCALL call_1_13_7_13(SH2_struct* sh)
{
  SH2movls4(sh, 13,7, 13);
}
#define init_1_13_7_14
void FASTCALL call_1_13_7_14(SH2_struct* sh)
{
  SH2movls4(sh, 13,7, 14);
}
#define init_1_13_7_15
void FASTCALL call_1_13_7_15(SH2_struct* sh)
{
  SH2movls4(sh, 13,7, 15);
}
#define init_1_13_8_0
void FASTCALL call_1_13_8_0(SH2_struct* sh)
{
  SH2movls4(sh, 13,8, 0);
}
#define init_1_13_8_1
void FASTCALL call_1_13_8_1(SH2_struct* sh)
{
  SH2movls4(sh, 13,8, 1);
}
#define init_1_13_8_2
void FASTCALL call_1_13_8_2(SH2_struct* sh)
{
  SH2movls4(sh, 13,8, 2);
}
#define init_1_13_8_3
void FASTCALL call_1_13_8_3(SH2_struct* sh)
{
  SH2movls4(sh, 13,8, 3);
}
#define init_1_13_8_4
void FASTCALL call_1_13_8_4(SH2_struct* sh)
{
  SH2movls4(sh, 13,8, 4);
}
#define init_1_13_8_5
void FASTCALL call_1_13_8_5(SH2_struct* sh)
{
  SH2movls4(sh, 13,8, 5);
}
#define init_1_13_8_6
void FASTCALL call_1_13_8_6(SH2_struct* sh)
{
  SH2movls4(sh, 13,8, 6);
}
#define init_1_13_8_7
void FASTCALL call_1_13_8_7(SH2_struct* sh)
{
  SH2movls4(sh, 13,8, 7);
}
#define init_1_13_8_8
void FASTCALL call_1_13_8_8(SH2_struct* sh)
{
  SH2movls4(sh, 13,8, 8);
}
#define init_1_13_8_9
void FASTCALL call_1_13_8_9(SH2_struct* sh)
{
  SH2movls4(sh, 13,8, 9);
}
#define init_1_13_8_10
void FASTCALL call_1_13_8_10(SH2_struct* sh)
{
  SH2movls4(sh, 13,8, 10);
}
#define init_1_13_8_11
void FASTCALL call_1_13_8_11(SH2_struct* sh)
{
  SH2movls4(sh, 13,8, 11);
}
#define init_1_13_8_12
void FASTCALL call_1_13_8_12(SH2_struct* sh)
{
  SH2movls4(sh, 13,8, 12);
}
#define init_1_13_8_13
void FASTCALL call_1_13_8_13(SH2_struct* sh)
{
  SH2movls4(sh, 13,8, 13);
}
#define init_1_13_8_14
void FASTCALL call_1_13_8_14(SH2_struct* sh)
{
  SH2movls4(sh, 13,8, 14);
}
#define init_1_13_8_15
void FASTCALL call_1_13_8_15(SH2_struct* sh)
{
  SH2movls4(sh, 13,8, 15);
}
#define init_1_13_9_0
void FASTCALL call_1_13_9_0(SH2_struct* sh)
{
  SH2movls4(sh, 13,9, 0);
}
#define init_1_13_9_1
void FASTCALL call_1_13_9_1(SH2_struct* sh)
{
  SH2movls4(sh, 13,9, 1);
}
#define init_1_13_9_2
void FASTCALL call_1_13_9_2(SH2_struct* sh)
{
  SH2movls4(sh, 13,9, 2);
}
#define init_1_13_9_3
void FASTCALL call_1_13_9_3(SH2_struct* sh)
{
  SH2movls4(sh, 13,9, 3);
}
#define init_1_13_9_4
void FASTCALL call_1_13_9_4(SH2_struct* sh)
{
  SH2movls4(sh, 13,9, 4);
}
#define init_1_13_9_5
void FASTCALL call_1_13_9_5(SH2_struct* sh)
{
  SH2movls4(sh, 13,9, 5);
}
#define init_1_13_9_6
void FASTCALL call_1_13_9_6(SH2_struct* sh)
{
  SH2movls4(sh, 13,9, 6);
}
#define init_1_13_9_7
void FASTCALL call_1_13_9_7(SH2_struct* sh)
{
  SH2movls4(sh, 13,9, 7);
}
#define init_1_13_9_8
void FASTCALL call_1_13_9_8(SH2_struct* sh)
{
  SH2movls4(sh, 13,9, 8);
}
#define init_1_13_9_9
void FASTCALL call_1_13_9_9(SH2_struct* sh)
{
  SH2movls4(sh, 13,9, 9);
}
#define init_1_13_9_10
void FASTCALL call_1_13_9_10(SH2_struct* sh)
{
  SH2movls4(sh, 13,9, 10);
}
#define init_1_13_9_11
void FASTCALL call_1_13_9_11(SH2_struct* sh)
{
  SH2movls4(sh, 13,9, 11);
}
#define init_1_13_9_12
void FASTCALL call_1_13_9_12(SH2_struct* sh)
{
  SH2movls4(sh, 13,9, 12);
}
#define init_1_13_9_13
void FASTCALL call_1_13_9_13(SH2_struct* sh)
{
  SH2movls4(sh, 13,9, 13);
}
#define init_1_13_9_14
void FASTCALL call_1_13_9_14(SH2_struct* sh)
{
  SH2movls4(sh, 13,9, 14);
}
#define init_1_13_9_15
void FASTCALL call_1_13_9_15(SH2_struct* sh)
{
  SH2movls4(sh, 13,9, 15);
}
#define init_1_13_10_0
void FASTCALL call_1_13_10_0(SH2_struct* sh)
{
  SH2movls4(sh, 13,10, 0);
}
#define init_1_13_10_1
void FASTCALL call_1_13_10_1(SH2_struct* sh)
{
  SH2movls4(sh, 13,10, 1);
}
#define init_1_13_10_2
void FASTCALL call_1_13_10_2(SH2_struct* sh)
{
  SH2movls4(sh, 13,10, 2);
}
#define init_1_13_10_3
void FASTCALL call_1_13_10_3(SH2_struct* sh)
{
  SH2movls4(sh, 13,10, 3);
}
#define init_1_13_10_4
void FASTCALL call_1_13_10_4(SH2_struct* sh)
{
  SH2movls4(sh, 13,10, 4);
}
#define init_1_13_10_5
void FASTCALL call_1_13_10_5(SH2_struct* sh)
{
  SH2movls4(sh, 13,10, 5);
}
#define init_1_13_10_6
void FASTCALL call_1_13_10_6(SH2_struct* sh)
{
  SH2movls4(sh, 13,10, 6);
}
#define init_1_13_10_7
void FASTCALL call_1_13_10_7(SH2_struct* sh)
{
  SH2movls4(sh, 13,10, 7);
}
#define init_1_13_10_8
void FASTCALL call_1_13_10_8(SH2_struct* sh)
{
  SH2movls4(sh, 13,10, 8);
}
#define init_1_13_10_9
void FASTCALL call_1_13_10_9(SH2_struct* sh)
{
  SH2movls4(sh, 13,10, 9);
}
#define init_1_13_10_10
void FASTCALL call_1_13_10_10(SH2_struct* sh)
{
  SH2movls4(sh, 13,10, 10);
}
#define init_1_13_10_11
void FASTCALL call_1_13_10_11(SH2_struct* sh)
{
  SH2movls4(sh, 13,10, 11);
}
#define init_1_13_10_12
void FASTCALL call_1_13_10_12(SH2_struct* sh)
{
  SH2movls4(sh, 13,10, 12);
}
#define init_1_13_10_13
void FASTCALL call_1_13_10_13(SH2_struct* sh)
{
  SH2movls4(sh, 13,10, 13);
}
#define init_1_13_10_14
void FASTCALL call_1_13_10_14(SH2_struct* sh)
{
  SH2movls4(sh, 13,10, 14);
}
#define init_1_13_10_15
void FASTCALL call_1_13_10_15(SH2_struct* sh)
{
  SH2movls4(sh, 13,10, 15);
}
#define init_1_13_11_0
void FASTCALL call_1_13_11_0(SH2_struct* sh)
{
  SH2movls4(sh, 13,11, 0);
}
#define init_1_13_11_1
void FASTCALL call_1_13_11_1(SH2_struct* sh)
{
  SH2movls4(sh, 13,11, 1);
}
#define init_1_13_11_2
void FASTCALL call_1_13_11_2(SH2_struct* sh)
{
  SH2movls4(sh, 13,11, 2);
}
#define init_1_13_11_3
void FASTCALL call_1_13_11_3(SH2_struct* sh)
{
  SH2movls4(sh, 13,11, 3);
}
#define init_1_13_11_4
void FASTCALL call_1_13_11_4(SH2_struct* sh)
{
  SH2movls4(sh, 13,11, 4);
}
#define init_1_13_11_5
void FASTCALL call_1_13_11_5(SH2_struct* sh)
{
  SH2movls4(sh, 13,11, 5);
}
#define init_1_13_11_6
void FASTCALL call_1_13_11_6(SH2_struct* sh)
{
  SH2movls4(sh, 13,11, 6);
}
#define init_1_13_11_7
void FASTCALL call_1_13_11_7(SH2_struct* sh)
{
  SH2movls4(sh, 13,11, 7);
}
#define init_1_13_11_8
void FASTCALL call_1_13_11_8(SH2_struct* sh)
{
  SH2movls4(sh, 13,11, 8);
}
#define init_1_13_11_9
void FASTCALL call_1_13_11_9(SH2_struct* sh)
{
  SH2movls4(sh, 13,11, 9);
}
#define init_1_13_11_10
void FASTCALL call_1_13_11_10(SH2_struct* sh)
{
  SH2movls4(sh, 13,11, 10);
}
#define init_1_13_11_11
void FASTCALL call_1_13_11_11(SH2_struct* sh)
{
  SH2movls4(sh, 13,11, 11);
}
#define init_1_13_11_12
void FASTCALL call_1_13_11_12(SH2_struct* sh)
{
  SH2movls4(sh, 13,11, 12);
}
#define init_1_13_11_13
void FASTCALL call_1_13_11_13(SH2_struct* sh)
{
  SH2movls4(sh, 13,11, 13);
}
#define init_1_13_11_14
void FASTCALL call_1_13_11_14(SH2_struct* sh)
{
  SH2movls4(sh, 13,11, 14);
}
#define init_1_13_11_15
void FASTCALL call_1_13_11_15(SH2_struct* sh)
{
  SH2movls4(sh, 13,11, 15);
}
#define init_1_13_12_0
void FASTCALL call_1_13_12_0(SH2_struct* sh)
{
  SH2movls4(sh, 13,12, 0);
}
#define init_1_13_12_1
void FASTCALL call_1_13_12_1(SH2_struct* sh)
{
  SH2movls4(sh, 13,12, 1);
}
#define init_1_13_12_2
void FASTCALL call_1_13_12_2(SH2_struct* sh)
{
  SH2movls4(sh, 13,12, 2);
}
#define init_1_13_12_3
void FASTCALL call_1_13_12_3(SH2_struct* sh)
{
  SH2movls4(sh, 13,12, 3);
}
#define init_1_13_12_4
void FASTCALL call_1_13_12_4(SH2_struct* sh)
{
  SH2movls4(sh, 13,12, 4);
}
#define init_1_13_12_5
void FASTCALL call_1_13_12_5(SH2_struct* sh)
{
  SH2movls4(sh, 13,12, 5);
}
#define init_1_13_12_6
void FASTCALL call_1_13_12_6(SH2_struct* sh)
{
  SH2movls4(sh, 13,12, 6);
}
#define init_1_13_12_7
void FASTCALL call_1_13_12_7(SH2_struct* sh)
{
  SH2movls4(sh, 13,12, 7);
}
#define init_1_13_12_8
void FASTCALL call_1_13_12_8(SH2_struct* sh)
{
  SH2movls4(sh, 13,12, 8);
}
#define init_1_13_12_9
void FASTCALL call_1_13_12_9(SH2_struct* sh)
{
  SH2movls4(sh, 13,12, 9);
}
#define init_1_13_12_10
void FASTCALL call_1_13_12_10(SH2_struct* sh)
{
  SH2movls4(sh, 13,12, 10);
}
#define init_1_13_12_11
void FASTCALL call_1_13_12_11(SH2_struct* sh)
{
  SH2movls4(sh, 13,12, 11);
}
#define init_1_13_12_12
void FASTCALL call_1_13_12_12(SH2_struct* sh)
{
  SH2movls4(sh, 13,12, 12);
}
#define init_1_13_12_13
void FASTCALL call_1_13_12_13(SH2_struct* sh)
{
  SH2movls4(sh, 13,12, 13);
}
#define init_1_13_12_14
void FASTCALL call_1_13_12_14(SH2_struct* sh)
{
  SH2movls4(sh, 13,12, 14);
}
#define init_1_13_12_15
void FASTCALL call_1_13_12_15(SH2_struct* sh)
{
  SH2movls4(sh, 13,12, 15);
}
#define init_1_13_13_0
void FASTCALL call_1_13_13_0(SH2_struct* sh)
{
  SH2movls4(sh, 13,13, 0);
}
#define init_1_13_13_1
void FASTCALL call_1_13_13_1(SH2_struct* sh)
{
  SH2movls4(sh, 13,13, 1);
}
#define init_1_13_13_2
void FASTCALL call_1_13_13_2(SH2_struct* sh)
{
  SH2movls4(sh, 13,13, 2);
}
#define init_1_13_13_3
void FASTCALL call_1_13_13_3(SH2_struct* sh)
{
  SH2movls4(sh, 13,13, 3);
}
#define init_1_13_13_4
void FASTCALL call_1_13_13_4(SH2_struct* sh)
{
  SH2movls4(sh, 13,13, 4);
}
#define init_1_13_13_5
void FASTCALL call_1_13_13_5(SH2_struct* sh)
{
  SH2movls4(sh, 13,13, 5);
}
#define init_1_13_13_6
void FASTCALL call_1_13_13_6(SH2_struct* sh)
{
  SH2movls4(sh, 13,13, 6);
}
#define init_1_13_13_7
void FASTCALL call_1_13_13_7(SH2_struct* sh)
{
  SH2movls4(sh, 13,13, 7);
}
#define init_1_13_13_8
void FASTCALL call_1_13_13_8(SH2_struct* sh)
{
  SH2movls4(sh, 13,13, 8);
}
#define init_1_13_13_9
void FASTCALL call_1_13_13_9(SH2_struct* sh)
{
  SH2movls4(sh, 13,13, 9);
}
#define init_1_13_13_10
void FASTCALL call_1_13_13_10(SH2_struct* sh)
{
  SH2movls4(sh, 13,13, 10);
}
#define init_1_13_13_11
void FASTCALL call_1_13_13_11(SH2_struct* sh)
{
  SH2movls4(sh, 13,13, 11);
}
#define init_1_13_13_12
void FASTCALL call_1_13_13_12(SH2_struct* sh)
{
  SH2movls4(sh, 13,13, 12);
}
#define init_1_13_13_13
void FASTCALL call_1_13_13_13(SH2_struct* sh)
{
  SH2movls4(sh, 13,13, 13);
}
#define init_1_13_13_14
void FASTCALL call_1_13_13_14(SH2_struct* sh)
{
  SH2movls4(sh, 13,13, 14);
}
#define init_1_13_13_15
void FASTCALL call_1_13_13_15(SH2_struct* sh)
{
  SH2movls4(sh, 13,13, 15);
}
#define init_1_13_14_0
void FASTCALL call_1_13_14_0(SH2_struct* sh)
{
  SH2movls4(sh, 13,14, 0);
}
#define init_1_13_14_1
void FASTCALL call_1_13_14_1(SH2_struct* sh)
{
  SH2movls4(sh, 13,14, 1);
}
#define init_1_13_14_2
void FASTCALL call_1_13_14_2(SH2_struct* sh)
{
  SH2movls4(sh, 13,14, 2);
}
#define init_1_13_14_3
void FASTCALL call_1_13_14_3(SH2_struct* sh)
{
  SH2movls4(sh, 13,14, 3);
}
#define init_1_13_14_4
void FASTCALL call_1_13_14_4(SH2_struct* sh)
{
  SH2movls4(sh, 13,14, 4);
}
#define init_1_13_14_5
void FASTCALL call_1_13_14_5(SH2_struct* sh)
{
  SH2movls4(sh, 13,14, 5);
}
#define init_1_13_14_6
void FASTCALL call_1_13_14_6(SH2_struct* sh)
{
  SH2movls4(sh, 13,14, 6);
}
#define init_1_13_14_7
void FASTCALL call_1_13_14_7(SH2_struct* sh)
{
  SH2movls4(sh, 13,14, 7);
}
#define init_1_13_14_8
void FASTCALL call_1_13_14_8(SH2_struct* sh)
{
  SH2movls4(sh, 13,14, 8);
}
#define init_1_13_14_9
void FASTCALL call_1_13_14_9(SH2_struct* sh)
{
  SH2movls4(sh, 13,14, 9);
}
#define init_1_13_14_10
void FASTCALL call_1_13_14_10(SH2_struct* sh)
{
  SH2movls4(sh, 13,14, 10);
}
#define init_1_13_14_11
void FASTCALL call_1_13_14_11(SH2_struct* sh)
{
  SH2movls4(sh, 13,14, 11);
}
#define init_1_13_14_12
void FASTCALL call_1_13_14_12(SH2_struct* sh)
{
  SH2movls4(sh, 13,14, 12);
}
#define init_1_13_14_13
void FASTCALL call_1_13_14_13(SH2_struct* sh)
{
  SH2movls4(sh, 13,14, 13);
}
#define init_1_13_14_14
void FASTCALL call_1_13_14_14(SH2_struct* sh)
{
  SH2movls4(sh, 13,14, 14);
}
#define init_1_13_14_15
void FASTCALL call_1_13_14_15(SH2_struct* sh)
{
  SH2movls4(sh, 13,14, 15);
}
#define init_1_13_15_0
void FASTCALL call_1_13_15_0(SH2_struct* sh)
{
  SH2movls4(sh, 13,15, 0);
}
#define init_1_13_15_1
void FASTCALL call_1_13_15_1(SH2_struct* sh)
{
  SH2movls4(sh, 13,15, 1);
}
#define init_1_13_15_2
void FASTCALL call_1_13_15_2(SH2_struct* sh)
{
  SH2movls4(sh, 13,15, 2);
}
#define init_1_13_15_3
void FASTCALL call_1_13_15_3(SH2_struct* sh)
{
  SH2movls4(sh, 13,15, 3);
}
#define init_1_13_15_4
void FASTCALL call_1_13_15_4(SH2_struct* sh)
{
  SH2movls4(sh, 13,15, 4);
}
#define init_1_13_15_5
void FASTCALL call_1_13_15_5(SH2_struct* sh)
{
  SH2movls4(sh, 13,15, 5);
}
#define init_1_13_15_6
void FASTCALL call_1_13_15_6(SH2_struct* sh)
{
  SH2movls4(sh, 13,15, 6);
}
#define init_1_13_15_7
void FASTCALL call_1_13_15_7(SH2_struct* sh)
{
  SH2movls4(sh, 13,15, 7);
}
#define init_1_13_15_8
void FASTCALL call_1_13_15_8(SH2_struct* sh)
{
  SH2movls4(sh, 13,15, 8);
}
#define init_1_13_15_9
void FASTCALL call_1_13_15_9(SH2_struct* sh)
{
  SH2movls4(sh, 13,15, 9);
}
#define init_1_13_15_10
void FASTCALL call_1_13_15_10(SH2_struct* sh)
{
  SH2movls4(sh, 13,15, 10);
}
#define init_1_13_15_11
void FASTCALL call_1_13_15_11(SH2_struct* sh)
{
  SH2movls4(sh, 13,15, 11);
}
#define init_1_13_15_12
void FASTCALL call_1_13_15_12(SH2_struct* sh)
{
  SH2movls4(sh, 13,15, 12);
}
#define init_1_13_15_13
void FASTCALL call_1_13_15_13(SH2_struct* sh)
{
  SH2movls4(sh, 13,15, 13);
}
#define init_1_13_15_14
void FASTCALL call_1_13_15_14(SH2_struct* sh)
{
  SH2movls4(sh, 13,15, 14);
}
#define init_1_13_15_15
void FASTCALL call_1_13_15_15(SH2_struct* sh)
{
  SH2movls4(sh, 13,15, 15);
}
#define init_1_14_0_0
void FASTCALL call_1_14_0_0(SH2_struct* sh)
{
  SH2movls4(sh, 14,0, 0);
}
#define init_1_14_0_1
void FASTCALL call_1_14_0_1(SH2_struct* sh)
{
  SH2movls4(sh, 14,0, 1);
}
#define init_1_14_0_2
void FASTCALL call_1_14_0_2(SH2_struct* sh)
{
  SH2movls4(sh, 14,0, 2);
}
#define init_1_14_0_3
void FASTCALL call_1_14_0_3(SH2_struct* sh)
{
  SH2movls4(sh, 14,0, 3);
}
#define init_1_14_0_4
void FASTCALL call_1_14_0_4(SH2_struct* sh)
{
  SH2movls4(sh, 14,0, 4);
}
#define init_1_14_0_5
void FASTCALL call_1_14_0_5(SH2_struct* sh)
{
  SH2movls4(sh, 14,0, 5);
}
#define init_1_14_0_6
void FASTCALL call_1_14_0_6(SH2_struct* sh)
{
  SH2movls4(sh, 14,0, 6);
}
#define init_1_14_0_7
void FASTCALL call_1_14_0_7(SH2_struct* sh)
{
  SH2movls4(sh, 14,0, 7);
}
#define init_1_14_0_8
void FASTCALL call_1_14_0_8(SH2_struct* sh)
{
  SH2movls4(sh, 14,0, 8);
}
#define init_1_14_0_9
void FASTCALL call_1_14_0_9(SH2_struct* sh)
{
  SH2movls4(sh, 14,0, 9);
}
#define init_1_14_0_10
void FASTCALL call_1_14_0_10(SH2_struct* sh)
{
  SH2movls4(sh, 14,0, 10);
}
#define init_1_14_0_11
void FASTCALL call_1_14_0_11(SH2_struct* sh)
{
  SH2movls4(sh, 14,0, 11);
}
#define init_1_14_0_12
void FASTCALL call_1_14_0_12(SH2_struct* sh)
{
  SH2movls4(sh, 14,0, 12);
}
#define init_1_14_0_13
void FASTCALL call_1_14_0_13(SH2_struct* sh)
{
  SH2movls4(sh, 14,0, 13);
}
#define init_1_14_0_14
void FASTCALL call_1_14_0_14(SH2_struct* sh)
{
  SH2movls4(sh, 14,0, 14);
}
#define init_1_14_0_15
void FASTCALL call_1_14_0_15(SH2_struct* sh)
{
  SH2movls4(sh, 14,0, 15);
}
#define init_1_14_1_0
void FASTCALL call_1_14_1_0(SH2_struct* sh)
{
  SH2movls4(sh, 14,1, 0);
}
#define init_1_14_1_1
void FASTCALL call_1_14_1_1(SH2_struct* sh)
{
  SH2movls4(sh, 14,1, 1);
}
#define init_1_14_1_2
void FASTCALL call_1_14_1_2(SH2_struct* sh)
{
  SH2movls4(sh, 14,1, 2);
}
#define init_1_14_1_3
void FASTCALL call_1_14_1_3(SH2_struct* sh)
{
  SH2movls4(sh, 14,1, 3);
}
#define init_1_14_1_4
void FASTCALL call_1_14_1_4(SH2_struct* sh)
{
  SH2movls4(sh, 14,1, 4);
}
#define init_1_14_1_5
void FASTCALL call_1_14_1_5(SH2_struct* sh)
{
  SH2movls4(sh, 14,1, 5);
}
#define init_1_14_1_6
void FASTCALL call_1_14_1_6(SH2_struct* sh)
{
  SH2movls4(sh, 14,1, 6);
}
#define init_1_14_1_7
void FASTCALL call_1_14_1_7(SH2_struct* sh)
{
  SH2movls4(sh, 14,1, 7);
}
#define init_1_14_1_8
void FASTCALL call_1_14_1_8(SH2_struct* sh)
{
  SH2movls4(sh, 14,1, 8);
}
#define init_1_14_1_9
void FASTCALL call_1_14_1_9(SH2_struct* sh)
{
  SH2movls4(sh, 14,1, 9);
}
#define init_1_14_1_10
void FASTCALL call_1_14_1_10(SH2_struct* sh)
{
  SH2movls4(sh, 14,1, 10);
}
#define init_1_14_1_11
void FASTCALL call_1_14_1_11(SH2_struct* sh)
{
  SH2movls4(sh, 14,1, 11);
}
#define init_1_14_1_12
void FASTCALL call_1_14_1_12(SH2_struct* sh)
{
  SH2movls4(sh, 14,1, 12);
}
#define init_1_14_1_13
void FASTCALL call_1_14_1_13(SH2_struct* sh)
{
  SH2movls4(sh, 14,1, 13);
}
#define init_1_14_1_14
void FASTCALL call_1_14_1_14(SH2_struct* sh)
{
  SH2movls4(sh, 14,1, 14);
}
#define init_1_14_1_15
void FASTCALL call_1_14_1_15(SH2_struct* sh)
{
  SH2movls4(sh, 14,1, 15);
}
#define init_1_14_2_0
void FASTCALL call_1_14_2_0(SH2_struct* sh)
{
  SH2movls4(sh, 14,2, 0);
}
#define init_1_14_2_1
void FASTCALL call_1_14_2_1(SH2_struct* sh)
{
  SH2movls4(sh, 14,2, 1);
}
#define init_1_14_2_2
void FASTCALL call_1_14_2_2(SH2_struct* sh)
{
  SH2movls4(sh, 14,2, 2);
}
#define init_1_14_2_3
void FASTCALL call_1_14_2_3(SH2_struct* sh)
{
  SH2movls4(sh, 14,2, 3);
}
#define init_1_14_2_4
void FASTCALL call_1_14_2_4(SH2_struct* sh)
{
  SH2movls4(sh, 14,2, 4);
}
#define init_1_14_2_5
void FASTCALL call_1_14_2_5(SH2_struct* sh)
{
  SH2movls4(sh, 14,2, 5);
}
#define init_1_14_2_6
void FASTCALL call_1_14_2_6(SH2_struct* sh)
{
  SH2movls4(sh, 14,2, 6);
}
#define init_1_14_2_7
void FASTCALL call_1_14_2_7(SH2_struct* sh)
{
  SH2movls4(sh, 14,2, 7);
}
#define init_1_14_2_8
void FASTCALL call_1_14_2_8(SH2_struct* sh)
{
  SH2movls4(sh, 14,2, 8);
}
#define init_1_14_2_9
void FASTCALL call_1_14_2_9(SH2_struct* sh)
{
  SH2movls4(sh, 14,2, 9);
}
#define init_1_14_2_10
void FASTCALL call_1_14_2_10(SH2_struct* sh)
{
  SH2movls4(sh, 14,2, 10);
}
#define init_1_14_2_11
void FASTCALL call_1_14_2_11(SH2_struct* sh)
{
  SH2movls4(sh, 14,2, 11);
}
#define init_1_14_2_12
void FASTCALL call_1_14_2_12(SH2_struct* sh)
{
  SH2movls4(sh, 14,2, 12);
}
#define init_1_14_2_13
void FASTCALL call_1_14_2_13(SH2_struct* sh)
{
  SH2movls4(sh, 14,2, 13);
}
#define init_1_14_2_14
void FASTCALL call_1_14_2_14(SH2_struct* sh)
{
  SH2movls4(sh, 14,2, 14);
}
#define init_1_14_2_15
void FASTCALL call_1_14_2_15(SH2_struct* sh)
{
  SH2movls4(sh, 14,2, 15);
}
#define init_1_14_3_0
void FASTCALL call_1_14_3_0(SH2_struct* sh)
{
  SH2movls4(sh, 14,3, 0);
}
#define init_1_14_3_1
void FASTCALL call_1_14_3_1(SH2_struct* sh)
{
  SH2movls4(sh, 14,3, 1);
}
#define init_1_14_3_2
void FASTCALL call_1_14_3_2(SH2_struct* sh)
{
  SH2movls4(sh, 14,3, 2);
}
#define init_1_14_3_3
void FASTCALL call_1_14_3_3(SH2_struct* sh)
{
  SH2movls4(sh, 14,3, 3);
}
#define init_1_14_3_4
void FASTCALL call_1_14_3_4(SH2_struct* sh)
{
  SH2movls4(sh, 14,3, 4);
}
#define init_1_14_3_5
void FASTCALL call_1_14_3_5(SH2_struct* sh)
{
  SH2movls4(sh, 14,3, 5);
}
#define init_1_14_3_6
void FASTCALL call_1_14_3_6(SH2_struct* sh)
{
  SH2movls4(sh, 14,3, 6);
}
#define init_1_14_3_7
void FASTCALL call_1_14_3_7(SH2_struct* sh)
{
  SH2movls4(sh, 14,3, 7);
}
#define init_1_14_3_8
void FASTCALL call_1_14_3_8(SH2_struct* sh)
{
  SH2movls4(sh, 14,3, 8);
}
#define init_1_14_3_9
void FASTCALL call_1_14_3_9(SH2_struct* sh)
{
  SH2movls4(sh, 14,3, 9);
}
#define init_1_14_3_10
void FASTCALL call_1_14_3_10(SH2_struct* sh)
{
  SH2movls4(sh, 14,3, 10);
}
#define init_1_14_3_11
void FASTCALL call_1_14_3_11(SH2_struct* sh)
{
  SH2movls4(sh, 14,3, 11);
}
#define init_1_14_3_12
void FASTCALL call_1_14_3_12(SH2_struct* sh)
{
  SH2movls4(sh, 14,3, 12);
}
#define init_1_14_3_13
void FASTCALL call_1_14_3_13(SH2_struct* sh)
{
  SH2movls4(sh, 14,3, 13);
}
#define init_1_14_3_14
void FASTCALL call_1_14_3_14(SH2_struct* sh)
{
  SH2movls4(sh, 14,3, 14);
}
#define init_1_14_3_15
void FASTCALL call_1_14_3_15(SH2_struct* sh)
{
  SH2movls4(sh, 14,3, 15);
}
#define init_1_14_4_0
void FASTCALL call_1_14_4_0(SH2_struct* sh)
{
  SH2movls4(sh, 14,4, 0);
}
#define init_1_14_4_1
void FASTCALL call_1_14_4_1(SH2_struct* sh)
{
  SH2movls4(sh, 14,4, 1);
}
#define init_1_14_4_2
void FASTCALL call_1_14_4_2(SH2_struct* sh)
{
  SH2movls4(sh, 14,4, 2);
}
#define init_1_14_4_3
void FASTCALL call_1_14_4_3(SH2_struct* sh)
{
  SH2movls4(sh, 14,4, 3);
}
#define init_1_14_4_4
void FASTCALL call_1_14_4_4(SH2_struct* sh)
{
  SH2movls4(sh, 14,4, 4);
}
#define init_1_14_4_5
void FASTCALL call_1_14_4_5(SH2_struct* sh)
{
  SH2movls4(sh, 14,4, 5);
}
#define init_1_14_4_6
void FASTCALL call_1_14_4_6(SH2_struct* sh)
{
  SH2movls4(sh, 14,4, 6);
}
#define init_1_14_4_7
void FASTCALL call_1_14_4_7(SH2_struct* sh)
{
  SH2movls4(sh, 14,4, 7);
}
#define init_1_14_4_8
void FASTCALL call_1_14_4_8(SH2_struct* sh)
{
  SH2movls4(sh, 14,4, 8);
}
#define init_1_14_4_9
void FASTCALL call_1_14_4_9(SH2_struct* sh)
{
  SH2movls4(sh, 14,4, 9);
}
#define init_1_14_4_10
void FASTCALL call_1_14_4_10(SH2_struct* sh)
{
  SH2movls4(sh, 14,4, 10);
}
#define init_1_14_4_11
void FASTCALL call_1_14_4_11(SH2_struct* sh)
{
  SH2movls4(sh, 14,4, 11);
}
#define init_1_14_4_12
void FASTCALL call_1_14_4_12(SH2_struct* sh)
{
  SH2movls4(sh, 14,4, 12);
}
#define init_1_14_4_13
void FASTCALL call_1_14_4_13(SH2_struct* sh)
{
  SH2movls4(sh, 14,4, 13);
}
#define init_1_14_4_14
void FASTCALL call_1_14_4_14(SH2_struct* sh)
{
  SH2movls4(sh, 14,4, 14);
}
#define init_1_14_4_15
void FASTCALL call_1_14_4_15(SH2_struct* sh)
{
  SH2movls4(sh, 14,4, 15);
}
#define init_1_14_5_0
void FASTCALL call_1_14_5_0(SH2_struct* sh)
{
  SH2movls4(sh, 14,5, 0);
}
#define init_1_14_5_1
void FASTCALL call_1_14_5_1(SH2_struct* sh)
{
  SH2movls4(sh, 14,5, 1);
}
#define init_1_14_5_2
void FASTCALL call_1_14_5_2(SH2_struct* sh)
{
  SH2movls4(sh, 14,5, 2);
}
#define init_1_14_5_3
void FASTCALL call_1_14_5_3(SH2_struct* sh)
{
  SH2movls4(sh, 14,5, 3);
}
#define init_1_14_5_4
void FASTCALL call_1_14_5_4(SH2_struct* sh)
{
  SH2movls4(sh, 14,5, 4);
}
#define init_1_14_5_5
void FASTCALL call_1_14_5_5(SH2_struct* sh)
{
  SH2movls4(sh, 14,5, 5);
}
#define init_1_14_5_6
void FASTCALL call_1_14_5_6(SH2_struct* sh)
{
  SH2movls4(sh, 14,5, 6);
}
#define init_1_14_5_7
void FASTCALL call_1_14_5_7(SH2_struct* sh)
{
  SH2movls4(sh, 14,5, 7);
}
#define init_1_14_5_8
void FASTCALL call_1_14_5_8(SH2_struct* sh)
{
  SH2movls4(sh, 14,5, 8);
}
#define init_1_14_5_9
void FASTCALL call_1_14_5_9(SH2_struct* sh)
{
  SH2movls4(sh, 14,5, 9);
}
#define init_1_14_5_10
void FASTCALL call_1_14_5_10(SH2_struct* sh)
{
  SH2movls4(sh, 14,5, 10);
}
#define init_1_14_5_11
void FASTCALL call_1_14_5_11(SH2_struct* sh)
{
  SH2movls4(sh, 14,5, 11);
}
#define init_1_14_5_12
void FASTCALL call_1_14_5_12(SH2_struct* sh)
{
  SH2movls4(sh, 14,5, 12);
}
#define init_1_14_5_13
void FASTCALL call_1_14_5_13(SH2_struct* sh)
{
  SH2movls4(sh, 14,5, 13);
}
#define init_1_14_5_14
void FASTCALL call_1_14_5_14(SH2_struct* sh)
{
  SH2movls4(sh, 14,5, 14);
}
#define init_1_14_5_15
void FASTCALL call_1_14_5_15(SH2_struct* sh)
{
  SH2movls4(sh, 14,5, 15);
}
#define init_1_14_6_0
void FASTCALL call_1_14_6_0(SH2_struct* sh)
{
  SH2movls4(sh, 14,6, 0);
}
#define init_1_14_6_1
void FASTCALL call_1_14_6_1(SH2_struct* sh)
{
  SH2movls4(sh, 14,6, 1);
}
#define init_1_14_6_2
void FASTCALL call_1_14_6_2(SH2_struct* sh)
{
  SH2movls4(sh, 14,6, 2);
}
#define init_1_14_6_3
void FASTCALL call_1_14_6_3(SH2_struct* sh)
{
  SH2movls4(sh, 14,6, 3);
}
#define init_1_14_6_4
void FASTCALL call_1_14_6_4(SH2_struct* sh)
{
  SH2movls4(sh, 14,6, 4);
}
#define init_1_14_6_5
void FASTCALL call_1_14_6_5(SH2_struct* sh)
{
  SH2movls4(sh, 14,6, 5);
}
#define init_1_14_6_6
void FASTCALL call_1_14_6_6(SH2_struct* sh)
{
  SH2movls4(sh, 14,6, 6);
}
#define init_1_14_6_7
void FASTCALL call_1_14_6_7(SH2_struct* sh)
{
  SH2movls4(sh, 14,6, 7);
}
#define init_1_14_6_8
void FASTCALL call_1_14_6_8(SH2_struct* sh)
{
  SH2movls4(sh, 14,6, 8);
}
#define init_1_14_6_9
void FASTCALL call_1_14_6_9(SH2_struct* sh)
{
  SH2movls4(sh, 14,6, 9);
}
#define init_1_14_6_10
void FASTCALL call_1_14_6_10(SH2_struct* sh)
{
  SH2movls4(sh, 14,6, 10);
}
#define init_1_14_6_11
void FASTCALL call_1_14_6_11(SH2_struct* sh)
{
  SH2movls4(sh, 14,6, 11);
}
#define init_1_14_6_12
void FASTCALL call_1_14_6_12(SH2_struct* sh)
{
  SH2movls4(sh, 14,6, 12);
}
#define init_1_14_6_13
void FASTCALL call_1_14_6_13(SH2_struct* sh)
{
  SH2movls4(sh, 14,6, 13);
}
#define init_1_14_6_14
void FASTCALL call_1_14_6_14(SH2_struct* sh)
{
  SH2movls4(sh, 14,6, 14);
}
#define init_1_14_6_15
void FASTCALL call_1_14_6_15(SH2_struct* sh)
{
  SH2movls4(sh, 14,6, 15);
}
#define init_1_14_7_0
void FASTCALL call_1_14_7_0(SH2_struct* sh)
{
  SH2movls4(sh, 14,7, 0);
}
#define init_1_14_7_1
void FASTCALL call_1_14_7_1(SH2_struct* sh)
{
  SH2movls4(sh, 14,7, 1);
}
#define init_1_14_7_2
void FASTCALL call_1_14_7_2(SH2_struct* sh)
{
  SH2movls4(sh, 14,7, 2);
}
#define init_1_14_7_3
void FASTCALL call_1_14_7_3(SH2_struct* sh)
{
  SH2movls4(sh, 14,7, 3);
}
#define init_1_14_7_4
void FASTCALL call_1_14_7_4(SH2_struct* sh)
{
  SH2movls4(sh, 14,7, 4);
}
#define init_1_14_7_5
void FASTCALL call_1_14_7_5(SH2_struct* sh)
{
  SH2movls4(sh, 14,7, 5);
}
#define init_1_14_7_6
void FASTCALL call_1_14_7_6(SH2_struct* sh)
{
  SH2movls4(sh, 14,7, 6);
}
#define init_1_14_7_7
void FASTCALL call_1_14_7_7(SH2_struct* sh)
{
  SH2movls4(sh, 14,7, 7);
}
#define init_1_14_7_8
void FASTCALL call_1_14_7_8(SH2_struct* sh)
{
  SH2movls4(sh, 14,7, 8);
}
#define init_1_14_7_9
void FASTCALL call_1_14_7_9(SH2_struct* sh)
{
  SH2movls4(sh, 14,7, 9);
}
#define init_1_14_7_10
void FASTCALL call_1_14_7_10(SH2_struct* sh)
{
  SH2movls4(sh, 14,7, 10);
}
#define init_1_14_7_11
void FASTCALL call_1_14_7_11(SH2_struct* sh)
{
  SH2movls4(sh, 14,7, 11);
}
#define init_1_14_7_12
void FASTCALL call_1_14_7_12(SH2_struct* sh)
{
  SH2movls4(sh, 14,7, 12);
}
#define init_1_14_7_13
void FASTCALL call_1_14_7_13(SH2_struct* sh)
{
  SH2movls4(sh, 14,7, 13);
}
#define init_1_14_7_14
void FASTCALL call_1_14_7_14(SH2_struct* sh)
{
  SH2movls4(sh, 14,7, 14);
}
#define init_1_14_7_15
void FASTCALL call_1_14_7_15(SH2_struct* sh)
{
  SH2movls4(sh, 14,7, 15);
}
#define init_1_14_8_0
void FASTCALL call_1_14_8_0(SH2_struct* sh)
{
  SH2movls4(sh, 14,8, 0);
}
#define init_1_14_8_1
void FASTCALL call_1_14_8_1(SH2_struct* sh)
{
  SH2movls4(sh, 14,8, 1);
}
#define init_1_14_8_2
void FASTCALL call_1_14_8_2(SH2_struct* sh)
{
  SH2movls4(sh, 14,8, 2);
}
#define init_1_14_8_3
void FASTCALL call_1_14_8_3(SH2_struct* sh)
{
  SH2movls4(sh, 14,8, 3);
}
#define init_1_14_8_4
void FASTCALL call_1_14_8_4(SH2_struct* sh)
{
  SH2movls4(sh, 14,8, 4);
}
#define init_1_14_8_5
void FASTCALL call_1_14_8_5(SH2_struct* sh)
{
  SH2movls4(sh, 14,8, 5);
}
#define init_1_14_8_6
void FASTCALL call_1_14_8_6(SH2_struct* sh)
{
  SH2movls4(sh, 14,8, 6);
}
#define init_1_14_8_7
void FASTCALL call_1_14_8_7(SH2_struct* sh)
{
  SH2movls4(sh, 14,8, 7);
}
#define init_1_14_8_8
void FASTCALL call_1_14_8_8(SH2_struct* sh)
{
  SH2movls4(sh, 14,8, 8);
}
#define init_1_14_8_9
void FASTCALL call_1_14_8_9(SH2_struct* sh)
{
  SH2movls4(sh, 14,8, 9);
}
#define init_1_14_8_10
void FASTCALL call_1_14_8_10(SH2_struct* sh)
{
  SH2movls4(sh, 14,8, 10);
}
#define init_1_14_8_11
void FASTCALL call_1_14_8_11(SH2_struct* sh)
{
  SH2movls4(sh, 14,8, 11);
}
#define init_1_14_8_12
void FASTCALL call_1_14_8_12(SH2_struct* sh)
{
  SH2movls4(sh, 14,8, 12);
}
#define init_1_14_8_13
void FASTCALL call_1_14_8_13(SH2_struct* sh)
{
  SH2movls4(sh, 14,8, 13);
}
#define init_1_14_8_14
void FASTCALL call_1_14_8_14(SH2_struct* sh)
{
  SH2movls4(sh, 14,8, 14);
}
#define init_1_14_8_15
void FASTCALL call_1_14_8_15(SH2_struct* sh)
{
  SH2movls4(sh, 14,8, 15);
}
#define init_1_14_9_0
void FASTCALL call_1_14_9_0(SH2_struct* sh)
{
  SH2movls4(sh, 14,9, 0);
}
#define init_1_14_9_1
void FASTCALL call_1_14_9_1(SH2_struct* sh)
{
  SH2movls4(sh, 14,9, 1);
}
#define init_1_14_9_2
void FASTCALL call_1_14_9_2(SH2_struct* sh)
{
  SH2movls4(sh, 14,9, 2);
}
#define init_1_14_9_3
void FASTCALL call_1_14_9_3(SH2_struct* sh)
{
  SH2movls4(sh, 14,9, 3);
}
#define init_1_14_9_4
void FASTCALL call_1_14_9_4(SH2_struct* sh)
{
  SH2movls4(sh, 14,9, 4);
}
#define init_1_14_9_5
void FASTCALL call_1_14_9_5(SH2_struct* sh)
{
  SH2movls4(sh, 14,9, 5);
}
#define init_1_14_9_6
void FASTCALL call_1_14_9_6(SH2_struct* sh)
{
  SH2movls4(sh, 14,9, 6);
}
#define init_1_14_9_7
void FASTCALL call_1_14_9_7(SH2_struct* sh)
{
  SH2movls4(sh, 14,9, 7);
}
#define init_1_14_9_8
void FASTCALL call_1_14_9_8(SH2_struct* sh)
{
  SH2movls4(sh, 14,9, 8);
}
#define init_1_14_9_9
void FASTCALL call_1_14_9_9(SH2_struct* sh)
{
  SH2movls4(sh, 14,9, 9);
}
#define init_1_14_9_10
void FASTCALL call_1_14_9_10(SH2_struct* sh)
{
  SH2movls4(sh, 14,9, 10);
}
#define init_1_14_9_11
void FASTCALL call_1_14_9_11(SH2_struct* sh)
{
  SH2movls4(sh, 14,9, 11);
}
#define init_1_14_9_12
void FASTCALL call_1_14_9_12(SH2_struct* sh)
{
  SH2movls4(sh, 14,9, 12);
}
#define init_1_14_9_13
void FASTCALL call_1_14_9_13(SH2_struct* sh)
{
  SH2movls4(sh, 14,9, 13);
}
#define init_1_14_9_14
void FASTCALL call_1_14_9_14(SH2_struct* sh)
{
  SH2movls4(sh, 14,9, 14);
}
#define init_1_14_9_15
void FASTCALL call_1_14_9_15(SH2_struct* sh)
{
  SH2movls4(sh, 14,9, 15);
}
#define init_1_14_10_0
void FASTCALL call_1_14_10_0(SH2_struct* sh)
{
  SH2movls4(sh, 14,10, 0);
}
#define init_1_14_10_1
void FASTCALL call_1_14_10_1(SH2_struct* sh)
{
  SH2movls4(sh, 14,10, 1);
}
#define init_1_14_10_2
void FASTCALL call_1_14_10_2(SH2_struct* sh)
{
  SH2movls4(sh, 14,10, 2);
}
#define init_1_14_10_3
void FASTCALL call_1_14_10_3(SH2_struct* sh)
{
  SH2movls4(sh, 14,10, 3);
}
#define init_1_14_10_4
void FASTCALL call_1_14_10_4(SH2_struct* sh)
{
  SH2movls4(sh, 14,10, 4);
}
#define init_1_14_10_5
void FASTCALL call_1_14_10_5(SH2_struct* sh)
{
  SH2movls4(sh, 14,10, 5);
}
#define init_1_14_10_6
void FASTCALL call_1_14_10_6(SH2_struct* sh)
{
  SH2movls4(sh, 14,10, 6);
}
#define init_1_14_10_7
void FASTCALL call_1_14_10_7(SH2_struct* sh)
{
  SH2movls4(sh, 14,10, 7);
}
#define init_1_14_10_8
void FASTCALL call_1_14_10_8(SH2_struct* sh)
{
  SH2movls4(sh, 14,10, 8);
}
#define init_1_14_10_9
void FASTCALL call_1_14_10_9(SH2_struct* sh)
{
  SH2movls4(sh, 14,10, 9);
}
#define init_1_14_10_10
void FASTCALL call_1_14_10_10(SH2_struct* sh)
{
  SH2movls4(sh, 14,10, 10);
}
#define init_1_14_10_11
void FASTCALL call_1_14_10_11(SH2_struct* sh)
{
  SH2movls4(sh, 14,10, 11);
}
#define init_1_14_10_12
void FASTCALL call_1_14_10_12(SH2_struct* sh)
{
  SH2movls4(sh, 14,10, 12);
}
#define init_1_14_10_13
void FASTCALL call_1_14_10_13(SH2_struct* sh)
{
  SH2movls4(sh, 14,10, 13);
}
#define init_1_14_10_14
void FASTCALL call_1_14_10_14(SH2_struct* sh)
{
  SH2movls4(sh, 14,10, 14);
}
#define init_1_14_10_15
void FASTCALL call_1_14_10_15(SH2_struct* sh)
{
  SH2movls4(sh, 14,10, 15);
}
#define init_1_14_11_0
void FASTCALL call_1_14_11_0(SH2_struct* sh)
{
  SH2movls4(sh, 14,11, 0);
}
#define init_1_14_11_1
void FASTCALL call_1_14_11_1(SH2_struct* sh)
{
  SH2movls4(sh, 14,11, 1);
}
#define init_1_14_11_2
void FASTCALL call_1_14_11_2(SH2_struct* sh)
{
  SH2movls4(sh, 14,11, 2);
}
#define init_1_14_11_3
void FASTCALL call_1_14_11_3(SH2_struct* sh)
{
  SH2movls4(sh, 14,11, 3);
}
#define init_1_14_11_4
void FASTCALL call_1_14_11_4(SH2_struct* sh)
{
  SH2movls4(sh, 14,11, 4);
}
#define init_1_14_11_5
void FASTCALL call_1_14_11_5(SH2_struct* sh)
{
  SH2movls4(sh, 14,11, 5);
}
#define init_1_14_11_6
void FASTCALL call_1_14_11_6(SH2_struct* sh)
{
  SH2movls4(sh, 14,11, 6);
}
#define init_1_14_11_7
void FASTCALL call_1_14_11_7(SH2_struct* sh)
{
  SH2movls4(sh, 14,11, 7);
}
#define init_1_14_11_8
void FASTCALL call_1_14_11_8(SH2_struct* sh)
{
  SH2movls4(sh, 14,11, 8);
}
#define init_1_14_11_9
void FASTCALL call_1_14_11_9(SH2_struct* sh)
{
  SH2movls4(sh, 14,11, 9);
}
#define init_1_14_11_10
void FASTCALL call_1_14_11_10(SH2_struct* sh)
{
  SH2movls4(sh, 14,11, 10);
}
#define init_1_14_11_11
void FASTCALL call_1_14_11_11(SH2_struct* sh)
{
  SH2movls4(sh, 14,11, 11);
}
#define init_1_14_11_12
void FASTCALL call_1_14_11_12(SH2_struct* sh)
{
  SH2movls4(sh, 14,11, 12);
}
#define init_1_14_11_13
void FASTCALL call_1_14_11_13(SH2_struct* sh)
{
  SH2movls4(sh, 14,11, 13);
}
#define init_1_14_11_14
void FASTCALL call_1_14_11_14(SH2_struct* sh)
{
  SH2movls4(sh, 14,11, 14);
}
#define init_1_14_11_15
void FASTCALL call_1_14_11_15(SH2_struct* sh)
{
  SH2movls4(sh, 14,11, 15);
}
#define init_1_14_12_0
void FASTCALL call_1_14_12_0(SH2_struct* sh)
{
  SH2movls4(sh, 14,12, 0);
}
#define init_1_14_12_1
void FASTCALL call_1_14_12_1(SH2_struct* sh)
{
  SH2movls4(sh, 14,12, 1);
}
#define init_1_14_12_2
void FASTCALL call_1_14_12_2(SH2_struct* sh)
{
  SH2movls4(sh, 14,12, 2);
}
#define init_1_14_12_3
void FASTCALL call_1_14_12_3(SH2_struct* sh)
{
  SH2movls4(sh, 14,12, 3);
}
#define init_1_14_12_4
void FASTCALL call_1_14_12_4(SH2_struct* sh)
{
  SH2movls4(sh, 14,12, 4);
}
#define init_1_14_12_5
void FASTCALL call_1_14_12_5(SH2_struct* sh)
{
  SH2movls4(sh, 14,12, 5);
}
#define init_1_14_12_6
void FASTCALL call_1_14_12_6(SH2_struct* sh)
{
  SH2movls4(sh, 14,12, 6);
}
#define init_1_14_12_7
void FASTCALL call_1_14_12_7(SH2_struct* sh)
{
  SH2movls4(sh, 14,12, 7);
}
#define init_1_14_12_8
void FASTCALL call_1_14_12_8(SH2_struct* sh)
{
  SH2movls4(sh, 14,12, 8);
}
#define init_1_14_12_9
void FASTCALL call_1_14_12_9(SH2_struct* sh)
{
  SH2movls4(sh, 14,12, 9);
}
#define init_1_14_12_10
void FASTCALL call_1_14_12_10(SH2_struct* sh)
{
  SH2movls4(sh, 14,12, 10);
}
#define init_1_14_12_11
void FASTCALL call_1_14_12_11(SH2_struct* sh)
{
  SH2movls4(sh, 14,12, 11);
}
#define init_1_14_12_12
void FASTCALL call_1_14_12_12(SH2_struct* sh)
{
  SH2movls4(sh, 14,12, 12);
}
#define init_1_14_12_13
void FASTCALL call_1_14_12_13(SH2_struct* sh)
{
  SH2movls4(sh, 14,12, 13);
}
#define init_1_14_12_14
void FASTCALL call_1_14_12_14(SH2_struct* sh)
{
  SH2movls4(sh, 14,12, 14);
}
#define init_1_14_12_15
void FASTCALL call_1_14_12_15(SH2_struct* sh)
{
  SH2movls4(sh, 14,12, 15);
}
#define init_1_14_13_0
void FASTCALL call_1_14_13_0(SH2_struct* sh)
{
  SH2movls4(sh, 14,13, 0);
}
#define init_1_14_13_1
void FASTCALL call_1_14_13_1(SH2_struct* sh)
{
  SH2movls4(sh, 14,13, 1);
}
#define init_1_14_13_2
void FASTCALL call_1_14_13_2(SH2_struct* sh)
{
  SH2movls4(sh, 14,13, 2);
}
#define init_1_14_13_3
void FASTCALL call_1_14_13_3(SH2_struct* sh)
{
  SH2movls4(sh, 14,13, 3);
}
#define init_1_14_13_4
void FASTCALL call_1_14_13_4(SH2_struct* sh)
{
  SH2movls4(sh, 14,13, 4);
}
#define init_1_14_13_5
void FASTCALL call_1_14_13_5(SH2_struct* sh)
{
  SH2movls4(sh, 14,13, 5);
}
#define init_1_14_13_6
void FASTCALL call_1_14_13_6(SH2_struct* sh)
{
  SH2movls4(sh, 14,13, 6);
}
#define init_1_14_13_7
void FASTCALL call_1_14_13_7(SH2_struct* sh)
{
  SH2movls4(sh, 14,13, 7);
}
#define init_1_14_13_8
void FASTCALL call_1_14_13_8(SH2_struct* sh)
{
  SH2movls4(sh, 14,13, 8);
}
#define init_1_14_13_9
void FASTCALL call_1_14_13_9(SH2_struct* sh)
{
  SH2movls4(sh, 14,13, 9);
}
#define init_1_14_13_10
void FASTCALL call_1_14_13_10(SH2_struct* sh)
{
  SH2movls4(sh, 14,13, 10);
}
#define init_1_14_13_11
void FASTCALL call_1_14_13_11(SH2_struct* sh)
{
  SH2movls4(sh, 14,13, 11);
}
#define init_1_14_13_12
void FASTCALL call_1_14_13_12(SH2_struct* sh)
{
  SH2movls4(sh, 14,13, 12);
}
#define init_1_14_13_13
void FASTCALL call_1_14_13_13(SH2_struct* sh)
{
  SH2movls4(sh, 14,13, 13);
}
#define init_1_14_13_14
void FASTCALL call_1_14_13_14(SH2_struct* sh)
{
  SH2movls4(sh, 14,13, 14);
}
#define init_1_14_13_15
void FASTCALL call_1_14_13_15(SH2_struct* sh)
{
  SH2movls4(sh, 14,13, 15);
}
#define init_1_14_14_0
void FASTCALL call_1_14_14_0(SH2_struct* sh)
{
  SH2movls4(sh, 14,14, 0);
}
#define init_1_14_14_1
void FASTCALL call_1_14_14_1(SH2_struct* sh)
{
  SH2movls4(sh, 14,14, 1);
}
#define init_1_14_14_2
void FASTCALL call_1_14_14_2(SH2_struct* sh)
{
  SH2movls4(sh, 14,14, 2);
}
#define init_1_14_14_3
void FASTCALL call_1_14_14_3(SH2_struct* sh)
{
  SH2movls4(sh, 14,14, 3);
}
#define init_1_14_14_4
void FASTCALL call_1_14_14_4(SH2_struct* sh)
{
  SH2movls4(sh, 14,14, 4);
}
#define init_1_14_14_5
void FASTCALL call_1_14_14_5(SH2_struct* sh)
{
  SH2movls4(sh, 14,14, 5);
}
#define init_1_14_14_6
void FASTCALL call_1_14_14_6(SH2_struct* sh)
{
  SH2movls4(sh, 14,14, 6);
}
#define init_1_14_14_7
void FASTCALL call_1_14_14_7(SH2_struct* sh)
{
  SH2movls4(sh, 14,14, 7);
}
#define init_1_14_14_8
void FASTCALL call_1_14_14_8(SH2_struct* sh)
{
  SH2movls4(sh, 14,14, 8);
}
#define init_1_14_14_9
void FASTCALL call_1_14_14_9(SH2_struct* sh)
{
  SH2movls4(sh, 14,14, 9);
}
#define init_1_14_14_10
void FASTCALL call_1_14_14_10(SH2_struct* sh)
{
  SH2movls4(sh, 14,14, 10);
}
#define init_1_14_14_11
void FASTCALL call_1_14_14_11(SH2_struct* sh)
{
  SH2movls4(sh, 14,14, 11);
}
#define init_1_14_14_12
void FASTCALL call_1_14_14_12(SH2_struct* sh)
{
  SH2movls4(sh, 14,14, 12);
}
#define init_1_14_14_13
void FASTCALL call_1_14_14_13(SH2_struct* sh)
{
  SH2movls4(sh, 14,14, 13);
}
#define init_1_14_14_14
void FASTCALL call_1_14_14_14(SH2_struct* sh)
{
  SH2movls4(sh, 14,14, 14);
}
#define init_1_14_14_15
void FASTCALL call_1_14_14_15(SH2_struct* sh)
{
  SH2movls4(sh, 14,14, 15);
}
#define init_1_14_15_0
void FASTCALL call_1_14_15_0(SH2_struct* sh)
{
  SH2movls4(sh, 14,15, 0);
}
#define init_1_14_15_1
void FASTCALL call_1_14_15_1(SH2_struct* sh)
{
  SH2movls4(sh, 14,15, 1);
}
#define init_1_14_15_2
void FASTCALL call_1_14_15_2(SH2_struct* sh)
{
  SH2movls4(sh, 14,15, 2);
}
#define init_1_14_15_3
void FASTCALL call_1_14_15_3(SH2_struct* sh)
{
  SH2movls4(sh, 14,15, 3);
}
#define init_1_14_15_4
void FASTCALL call_1_14_15_4(SH2_struct* sh)
{
  SH2movls4(sh, 14,15, 4);
}
#define init_1_14_15_5
void FASTCALL call_1_14_15_5(SH2_struct* sh)
{
  SH2movls4(sh, 14,15, 5);
}
#define init_1_14_15_6
void FASTCALL call_1_14_15_6(SH2_struct* sh)
{
  SH2movls4(sh, 14,15, 6);
}
#define init_1_14_15_7
void FASTCALL call_1_14_15_7(SH2_struct* sh)
{
  SH2movls4(sh, 14,15, 7);
}
#define init_1_14_15_8
void FASTCALL call_1_14_15_8(SH2_struct* sh)
{
  SH2movls4(sh, 14,15, 8);
}
#define init_1_14_15_9
void FASTCALL call_1_14_15_9(SH2_struct* sh)
{
  SH2movls4(sh, 14,15, 9);
}
#define init_1_14_15_10
void FASTCALL call_1_14_15_10(SH2_struct* sh)
{
  SH2movls4(sh, 14,15, 10);
}
#define init_1_14_15_11
void FASTCALL call_1_14_15_11(SH2_struct* sh)
{
  SH2movls4(sh, 14,15, 11);
}
#define init_1_14_15_12
void FASTCALL call_1_14_15_12(SH2_struct* sh)
{
  SH2movls4(sh, 14,15, 12);
}
#define init_1_14_15_13
void FASTCALL call_1_14_15_13(SH2_struct* sh)
{
  SH2movls4(sh, 14,15, 13);
}
#define init_1_14_15_14
void FASTCALL call_1_14_15_14(SH2_struct* sh)
{
  SH2movls4(sh, 14,15, 14);
}
#define init_1_14_15_15
void FASTCALL call_1_14_15_15(SH2_struct* sh)
{
  SH2movls4(sh, 14,15, 15);
}
#define init_1_15_0_0
void FASTCALL call_1_15_0_0(SH2_struct* sh)
{
  SH2movls4(sh, 15,0, 0);
}
#define init_1_15_0_1
void FASTCALL call_1_15_0_1(SH2_struct* sh)
{
  SH2movls4(sh, 15,0, 1);
}
#define init_1_15_0_2
void FASTCALL call_1_15_0_2(SH2_struct* sh)
{
  SH2movls4(sh, 15,0, 2);
}
#define init_1_15_0_3
void FASTCALL call_1_15_0_3(SH2_struct* sh)
{
  SH2movls4(sh, 15,0, 3);
}
#define init_1_15_0_4
void FASTCALL call_1_15_0_4(SH2_struct* sh)
{
  SH2movls4(sh, 15,0, 4);
}
#define init_1_15_0_5
void FASTCALL call_1_15_0_5(SH2_struct* sh)
{
  SH2movls4(sh, 15,0, 5);
}
#define init_1_15_0_6
void FASTCALL call_1_15_0_6(SH2_struct* sh)
{
  SH2movls4(sh, 15,0, 6);
}
#define init_1_15_0_7
void FASTCALL call_1_15_0_7(SH2_struct* sh)
{
  SH2movls4(sh, 15,0, 7);
}
#define init_1_15_0_8
void FASTCALL call_1_15_0_8(SH2_struct* sh)
{
  SH2movls4(sh, 15,0, 8);
}
#define init_1_15_0_9
void FASTCALL call_1_15_0_9(SH2_struct* sh)
{
  SH2movls4(sh, 15,0, 9);
}
#define init_1_15_0_10
void FASTCALL call_1_15_0_10(SH2_struct* sh)
{
  SH2movls4(sh, 15,0, 10);
}
#define init_1_15_0_11
void FASTCALL call_1_15_0_11(SH2_struct* sh)
{
  SH2movls4(sh, 15,0, 11);
}
#define init_1_15_0_12
void FASTCALL call_1_15_0_12(SH2_struct* sh)
{
  SH2movls4(sh, 15,0, 12);
}
#define init_1_15_0_13
void FASTCALL call_1_15_0_13(SH2_struct* sh)
{
  SH2movls4(sh, 15,0, 13);
}
#define init_1_15_0_14
void FASTCALL call_1_15_0_14(SH2_struct* sh)
{
  SH2movls4(sh, 15,0, 14);
}
#define init_1_15_0_15
void FASTCALL call_1_15_0_15(SH2_struct* sh)
{
  SH2movls4(sh, 15,0, 15);
}
#define init_1_15_1_0
void FASTCALL call_1_15_1_0(SH2_struct* sh)
{
  SH2movls4(sh, 15,1, 0);
}
#define init_1_15_1_1
void FASTCALL call_1_15_1_1(SH2_struct* sh)
{
  SH2movls4(sh, 15,1, 1);
}
#define init_1_15_1_2
void FASTCALL call_1_15_1_2(SH2_struct* sh)
{
  SH2movls4(sh, 15,1, 2);
}
#define init_1_15_1_3
void FASTCALL call_1_15_1_3(SH2_struct* sh)
{
  SH2movls4(sh, 15,1, 3);
}
#define init_1_15_1_4
void FASTCALL call_1_15_1_4(SH2_struct* sh)
{
  SH2movls4(sh, 15,1, 4);
}
#define init_1_15_1_5
void FASTCALL call_1_15_1_5(SH2_struct* sh)
{
  SH2movls4(sh, 15,1, 5);
}
#define init_1_15_1_6
void FASTCALL call_1_15_1_6(SH2_struct* sh)
{
  SH2movls4(sh, 15,1, 6);
}
#define init_1_15_1_7
void FASTCALL call_1_15_1_7(SH2_struct* sh)
{
  SH2movls4(sh, 15,1, 7);
}
#define init_1_15_1_8
void FASTCALL call_1_15_1_8(SH2_struct* sh)
{
  SH2movls4(sh, 15,1, 8);
}
#define init_1_15_1_9
void FASTCALL call_1_15_1_9(SH2_struct* sh)
{
  SH2movls4(sh, 15,1, 9);
}
#define init_1_15_1_10
void FASTCALL call_1_15_1_10(SH2_struct* sh)
{
  SH2movls4(sh, 15,1, 10);
}
#define init_1_15_1_11
void FASTCALL call_1_15_1_11(SH2_struct* sh)
{
  SH2movls4(sh, 15,1, 11);
}
#define init_1_15_1_12
void FASTCALL call_1_15_1_12(SH2_struct* sh)
{
  SH2movls4(sh, 15,1, 12);
}
#define init_1_15_1_13
void FASTCALL call_1_15_1_13(SH2_struct* sh)
{
  SH2movls4(sh, 15,1, 13);
}
#define init_1_15_1_14
void FASTCALL call_1_15_1_14(SH2_struct* sh)
{
  SH2movls4(sh, 15,1, 14);
}
#define init_1_15_1_15
void FASTCALL call_1_15_1_15(SH2_struct* sh)
{
  SH2movls4(sh, 15,1, 15);
}
#define init_1_15_2_0
void FASTCALL call_1_15_2_0(SH2_struct* sh)
{
  SH2movls4(sh, 15,2, 0);
}
#define init_1_15_2_1
void FASTCALL call_1_15_2_1(SH2_struct* sh)
{
  SH2movls4(sh, 15,2, 1);
}
#define init_1_15_2_2
void FASTCALL call_1_15_2_2(SH2_struct* sh)
{
  SH2movls4(sh, 15,2, 2);
}
#define init_1_15_2_3
void FASTCALL call_1_15_2_3(SH2_struct* sh)
{
  SH2movls4(sh, 15,2, 3);
}
#define init_1_15_2_4
void FASTCALL call_1_15_2_4(SH2_struct* sh)
{
  SH2movls4(sh, 15,2, 4);
}
#define init_1_15_2_5
void FASTCALL call_1_15_2_5(SH2_struct* sh)
{
  SH2movls4(sh, 15,2, 5);
}
#define init_1_15_2_6
void FASTCALL call_1_15_2_6(SH2_struct* sh)
{
  SH2movls4(sh, 15,2, 6);
}
#define init_1_15_2_7
void FASTCALL call_1_15_2_7(SH2_struct* sh)
{
  SH2movls4(sh, 15,2, 7);
}
#define init_1_15_2_8
void FASTCALL call_1_15_2_8(SH2_struct* sh)
{
  SH2movls4(sh, 15,2, 8);
}
#define init_1_15_2_9
void FASTCALL call_1_15_2_9(SH2_struct* sh)
{
  SH2movls4(sh, 15,2, 9);
}
#define init_1_15_2_10
void FASTCALL call_1_15_2_10(SH2_struct* sh)
{
  SH2movls4(sh, 15,2, 10);
}
#define init_1_15_2_11
void FASTCALL call_1_15_2_11(SH2_struct* sh)
{
  SH2movls4(sh, 15,2, 11);
}
#define init_1_15_2_12
void FASTCALL call_1_15_2_12(SH2_struct* sh)
{
  SH2movls4(sh, 15,2, 12);
}
#define init_1_15_2_13
void FASTCALL call_1_15_2_13(SH2_struct* sh)
{
  SH2movls4(sh, 15,2, 13);
}
#define init_1_15_2_14
void FASTCALL call_1_15_2_14(SH2_struct* sh)
{
  SH2movls4(sh, 15,2, 14);
}
#define init_1_15_2_15
void FASTCALL call_1_15_2_15(SH2_struct* sh)
{
  SH2movls4(sh, 15,2, 15);
}
#define init_1_15_3_0
void FASTCALL call_1_15_3_0(SH2_struct* sh)
{
  SH2movls4(sh, 15,3, 0);
}
#define init_1_15_3_1
void FASTCALL call_1_15_3_1(SH2_struct* sh)
{
  SH2movls4(sh, 15,3, 1);
}
#define init_1_15_3_2
void FASTCALL call_1_15_3_2(SH2_struct* sh)
{
  SH2movls4(sh, 15,3, 2);
}
#define init_1_15_3_3
void FASTCALL call_1_15_3_3(SH2_struct* sh)
{
  SH2movls4(sh, 15,3, 3);
}
#define init_1_15_3_4
void FASTCALL call_1_15_3_4(SH2_struct* sh)
{
  SH2movls4(sh, 15,3, 4);
}
#define init_1_15_3_5
void FASTCALL call_1_15_3_5(SH2_struct* sh)
{
  SH2movls4(sh, 15,3, 5);
}
#define init_1_15_3_6
void FASTCALL call_1_15_3_6(SH2_struct* sh)
{
  SH2movls4(sh, 15,3, 6);
}
#define init_1_15_3_7
void FASTCALL call_1_15_3_7(SH2_struct* sh)
{
  SH2movls4(sh, 15,3, 7);
}
#define init_1_15_3_8
void FASTCALL call_1_15_3_8(SH2_struct* sh)
{
  SH2movls4(sh, 15,3, 8);
}
#define init_1_15_3_9
void FASTCALL call_1_15_3_9(SH2_struct* sh)
{
  SH2movls4(sh, 15,3, 9);
}
#define init_1_15_3_10
void FASTCALL call_1_15_3_10(SH2_struct* sh)
{
  SH2movls4(sh, 15,3, 10);
}
#define init_1_15_3_11
void FASTCALL call_1_15_3_11(SH2_struct* sh)
{
  SH2movls4(sh, 15,3, 11);
}
#define init_1_15_3_12
void FASTCALL call_1_15_3_12(SH2_struct* sh)
{
  SH2movls4(sh, 15,3, 12);
}
#define init_1_15_3_13
void FASTCALL call_1_15_3_13(SH2_struct* sh)
{
  SH2movls4(sh, 15,3, 13);
}
#define init_1_15_3_14
void FASTCALL call_1_15_3_14(SH2_struct* sh)
{
  SH2movls4(sh, 15,3, 14);
}
#define init_1_15_3_15
void FASTCALL call_1_15_3_15(SH2_struct* sh)
{
  SH2movls4(sh, 15,3, 15);
}
#define init_1_15_4_0
void FASTCALL call_1_15_4_0(SH2_struct* sh)
{
  SH2movls4(sh, 15,4, 0);
}
#define init_1_15_4_1
void FASTCALL call_1_15_4_1(SH2_struct* sh)
{
  SH2movls4(sh, 15,4, 1);
}
#define init_1_15_4_2
void FASTCALL call_1_15_4_2(SH2_struct* sh)
{
  SH2movls4(sh, 15,4, 2);
}
#define init_1_15_4_3
void FASTCALL call_1_15_4_3(SH2_struct* sh)
{
  SH2movls4(sh, 15,4, 3);
}
#define init_1_15_4_4
void FASTCALL call_1_15_4_4(SH2_struct* sh)
{
  SH2movls4(sh, 15,4, 4);
}
#define init_1_15_4_5
void FASTCALL call_1_15_4_5(SH2_struct* sh)
{
  SH2movls4(sh, 15,4, 5);
}
#define init_1_15_4_6
void FASTCALL call_1_15_4_6(SH2_struct* sh)
{
  SH2movls4(sh, 15,4, 6);
}
#define init_1_15_4_7
void FASTCALL call_1_15_4_7(SH2_struct* sh)
{
  SH2movls4(sh, 15,4, 7);
}
#define init_1_15_4_8
void FASTCALL call_1_15_4_8(SH2_struct* sh)
{
  SH2movls4(sh, 15,4, 8);
}
#define init_1_15_4_9
void FASTCALL call_1_15_4_9(SH2_struct* sh)
{
  SH2movls4(sh, 15,4, 9);
}
#define init_1_15_4_10
void FASTCALL call_1_15_4_10(SH2_struct* sh)
{
  SH2movls4(sh, 15,4, 10);
}
#define init_1_15_4_11
void FASTCALL call_1_15_4_11(SH2_struct* sh)
{
  SH2movls4(sh, 15,4, 11);
}
#define init_1_15_4_12
void FASTCALL call_1_15_4_12(SH2_struct* sh)
{
  SH2movls4(sh, 15,4, 12);
}
#define init_1_15_4_13
void FASTCALL call_1_15_4_13(SH2_struct* sh)
{
  SH2movls4(sh, 15,4, 13);
}
#define init_1_15_4_14
void FASTCALL call_1_15_4_14(SH2_struct* sh)
{
  SH2movls4(sh, 15,4, 14);
}
#define init_1_15_4_15
void FASTCALL call_1_15_4_15(SH2_struct* sh)
{
  SH2movls4(sh, 15,4, 15);
}
#define init_1_15_5_0
void FASTCALL call_1_15_5_0(SH2_struct* sh)
{
  SH2movls4(sh, 15,5, 0);
}
#define init_1_15_5_1
void FASTCALL call_1_15_5_1(SH2_struct* sh)
{
  SH2movls4(sh, 15,5, 1);
}
#define init_1_15_5_2
void FASTCALL call_1_15_5_2(SH2_struct* sh)
{
  SH2movls4(sh, 15,5, 2);
}
#define init_1_15_5_3
void FASTCALL call_1_15_5_3(SH2_struct* sh)
{
  SH2movls4(sh, 15,5, 3);
}
#define init_1_15_5_4
void FASTCALL call_1_15_5_4(SH2_struct* sh)
{
  SH2movls4(sh, 15,5, 4);
}
#define init_1_15_5_5
void FASTCALL call_1_15_5_5(SH2_struct* sh)
{
  SH2movls4(sh, 15,5, 5);
}
#define init_1_15_5_6
void FASTCALL call_1_15_5_6(SH2_struct* sh)
{
  SH2movls4(sh, 15,5, 6);
}
#define init_1_15_5_7
void FASTCALL call_1_15_5_7(SH2_struct* sh)
{
  SH2movls4(sh, 15,5, 7);
}
#define init_1_15_5_8
void FASTCALL call_1_15_5_8(SH2_struct* sh)
{
  SH2movls4(sh, 15,5, 8);
}
#define init_1_15_5_9
void FASTCALL call_1_15_5_9(SH2_struct* sh)
{
  SH2movls4(sh, 15,5, 9);
}
#define init_1_15_5_10
void FASTCALL call_1_15_5_10(SH2_struct* sh)
{
  SH2movls4(sh, 15,5, 10);
}
#define init_1_15_5_11
void FASTCALL call_1_15_5_11(SH2_struct* sh)
{
  SH2movls4(sh, 15,5, 11);
}
#define init_1_15_5_12
void FASTCALL call_1_15_5_12(SH2_struct* sh)
{
  SH2movls4(sh, 15,5, 12);
}
#define init_1_15_5_13
void FASTCALL call_1_15_5_13(SH2_struct* sh)
{
  SH2movls4(sh, 15,5, 13);
}
#define init_1_15_5_14
void FASTCALL call_1_15_5_14(SH2_struct* sh)
{
  SH2movls4(sh, 15,5, 14);
}
#define init_1_15_5_15
void FASTCALL call_1_15_5_15(SH2_struct* sh)
{
  SH2movls4(sh, 15,5, 15);
}
#define init_1_15_6_0
void FASTCALL call_1_15_6_0(SH2_struct* sh)
{
  SH2movls4(sh, 15,6, 0);
}
#define init_1_15_6_1
void FASTCALL call_1_15_6_1(SH2_struct* sh)
{
  SH2movls4(sh, 15,6, 1);
}
#define init_1_15_6_2
void FASTCALL call_1_15_6_2(SH2_struct* sh)
{
  SH2movls4(sh, 15,6, 2);
}
#define init_1_15_6_3
void FASTCALL call_1_15_6_3(SH2_struct* sh)
{
  SH2movls4(sh, 15,6, 3);
}
#define init_1_15_6_4
void FASTCALL call_1_15_6_4(SH2_struct* sh)
{
  SH2movls4(sh, 15,6, 4);
}
#define init_1_15_6_5
void FASTCALL call_1_15_6_5(SH2_struct* sh)
{
  SH2movls4(sh, 15,6, 5);
}
#define init_1_15_6_6
void FASTCALL call_1_15_6_6(SH2_struct* sh)
{
  SH2movls4(sh, 15,6, 6);
}
#define init_1_15_6_7
void FASTCALL call_1_15_6_7(SH2_struct* sh)
{
  SH2movls4(sh, 15,6, 7);
}
#define init_1_15_6_8
void FASTCALL call_1_15_6_8(SH2_struct* sh)
{
  SH2movls4(sh, 15,6, 8);
}
#define init_1_15_6_9
void FASTCALL call_1_15_6_9(SH2_struct* sh)
{
  SH2movls4(sh, 15,6, 9);
}
#define init_1_15_6_10
void FASTCALL call_1_15_6_10(SH2_struct* sh)
{
  SH2movls4(sh, 15,6, 10);
}
#define init_1_15_6_11
void FASTCALL call_1_15_6_11(SH2_struct* sh)
{
  SH2movls4(sh, 15,6, 11);
}
#define init_1_15_6_12
void FASTCALL call_1_15_6_12(SH2_struct* sh)
{
  SH2movls4(sh, 15,6, 12);
}
#define init_1_15_6_13
void FASTCALL call_1_15_6_13(SH2_struct* sh)
{
  SH2movls4(sh, 15,6, 13);
}
#define init_1_15_6_14
void FASTCALL call_1_15_6_14(SH2_struct* sh)
{
  SH2movls4(sh, 15,6, 14);
}
#define init_1_15_6_15
void FASTCALL call_1_15_6_15(SH2_struct* sh)
{
  SH2movls4(sh, 15,6, 15);
}
#define init_1_15_7_0
void FASTCALL call_1_15_7_0(SH2_struct* sh)
{
  SH2movls4(sh, 15,7, 0);
}
#define init_1_15_7_1
void FASTCALL call_1_15_7_1(SH2_struct* sh)
{
  SH2movls4(sh, 15,7, 1);
}
#define init_1_15_7_2
void FASTCALL call_1_15_7_2(SH2_struct* sh)
{
  SH2movls4(sh, 15,7, 2);
}
#define init_1_15_7_3
void FASTCALL call_1_15_7_3(SH2_struct* sh)
{
  SH2movls4(sh, 15,7, 3);
}
#define init_1_15_7_4
void FASTCALL call_1_15_7_4(SH2_struct* sh)
{
  SH2movls4(sh, 15,7, 4);
}
#define init_1_15_7_5
void FASTCALL call_1_15_7_5(SH2_struct* sh)
{
  SH2movls4(sh, 15,7, 5);
}
#define init_1_15_7_6
void FASTCALL call_1_15_7_6(SH2_struct* sh)
{
  SH2movls4(sh, 15,7, 6);
}
#define init_1_15_7_7
void FASTCALL call_1_15_7_7(SH2_struct* sh)
{
  SH2movls4(sh, 15,7, 7);
}
#define init_1_15_7_8
void FASTCALL call_1_15_7_8(SH2_struct* sh)
{
  SH2movls4(sh, 15,7, 8);
}
#define init_1_15_7_9
void FASTCALL call_1_15_7_9(SH2_struct* sh)
{
  SH2movls4(sh, 15,7, 9);
}
#define init_1_15_7_10
void FASTCALL call_1_15_7_10(SH2_struct* sh)
{
  SH2movls4(sh, 15,7, 10);
}
#define init_1_15_7_11
void FASTCALL call_1_15_7_11(SH2_struct* sh)
{
  SH2movls4(sh, 15,7, 11);
}
#define init_1_15_7_12
void FASTCALL call_1_15_7_12(SH2_struct* sh)
{
  SH2movls4(sh, 15,7, 12);
}
#define init_1_15_7_13
void FASTCALL call_1_15_7_13(SH2_struct* sh)
{
  SH2movls4(sh, 15,7, 13);
}
#define init_1_15_7_14
void FASTCALL call_1_15_7_14(SH2_struct* sh)
{
  SH2movls4(sh, 15,7, 14);
}
#define init_1_15_7_15
void FASTCALL call_1_15_7_15(SH2_struct* sh)
{
  SH2movls4(sh, 15,7, 15);
}
#define init_1_15_8_0
void FASTCALL call_1_15_8_0(SH2_struct* sh)
{
  SH2movls4(sh, 15,8, 0);
}
#define init_1_15_8_1
void FASTCALL call_1_15_8_1(SH2_struct* sh)
{
  SH2movls4(sh, 15,8, 1);
}
#define init_1_15_8_2
void FASTCALL call_1_15_8_2(SH2_struct* sh)
{
  SH2movls4(sh, 15,8, 2);
}
#define init_1_15_8_3
void FASTCALL call_1_15_8_3(SH2_struct* sh)
{
  SH2movls4(sh, 15,8, 3);
}
#define init_1_15_8_4
void FASTCALL call_1_15_8_4(SH2_struct* sh)
{
  SH2movls4(sh, 15,8, 4);
}
#define init_1_15_8_5
void FASTCALL call_1_15_8_5(SH2_struct* sh)
{
  SH2movls4(sh, 15,8, 5);
}
#define init_1_15_8_6
void FASTCALL call_1_15_8_6(SH2_struct* sh)
{
  SH2movls4(sh, 15,8, 6);
}
#define init_1_15_8_7
void FASTCALL call_1_15_8_7(SH2_struct* sh)
{
  SH2movls4(sh, 15,8, 7);
}
#define init_1_15_8_8
void FASTCALL call_1_15_8_8(SH2_struct* sh)
{
  SH2movls4(sh, 15,8, 8);
}
#define init_1_15_8_9
void FASTCALL call_1_15_8_9(SH2_struct* sh)
{
  SH2movls4(sh, 15,8, 9);
}
#define init_1_15_8_10
void FASTCALL call_1_15_8_10(SH2_struct* sh)
{
  SH2movls4(sh, 15,8, 10);
}
#define init_1_15_8_11
void FASTCALL call_1_15_8_11(SH2_struct* sh)
{
  SH2movls4(sh, 15,8, 11);
}
#define init_1_15_8_12
void FASTCALL call_1_15_8_12(SH2_struct* sh)
{
  SH2movls4(sh, 15,8, 12);
}
#define init_1_15_8_13
void FASTCALL call_1_15_8_13(SH2_struct* sh)
{
  SH2movls4(sh, 15,8, 13);
}
#define init_1_15_8_14
void FASTCALL call_1_15_8_14(SH2_struct* sh)
{
  SH2movls4(sh, 15,8, 14);
}
#define init_1_15_8_15
void FASTCALL call_1_15_8_15(SH2_struct* sh)
{
  SH2movls4(sh, 15,8, 15);
}
#define init_1_15_9_0
void FASTCALL call_1_15_9_0(SH2_struct* sh)
{
  SH2movls4(sh, 15,9, 0);
}
#define init_1_15_9_1
void FASTCALL call_1_15_9_1(SH2_struct* sh)
{
  SH2movls4(sh, 15,9, 1);
}
#define init_1_15_9_2
void FASTCALL call_1_15_9_2(SH2_struct* sh)
{
  SH2movls4(sh, 15,9, 2);
}
#define init_1_15_9_3
void FASTCALL call_1_15_9_3(SH2_struct* sh)
{
  SH2movls4(sh, 15,9, 3);
}
#define init_1_15_9_4
void FASTCALL call_1_15_9_4(SH2_struct* sh)
{
  SH2movls4(sh, 15,9, 4);
}
#define init_1_15_9_5
void FASTCALL call_1_15_9_5(SH2_struct* sh)
{
  SH2movls4(sh, 15,9, 5);
}
#define init_1_15_9_6
void FASTCALL call_1_15_9_6(SH2_struct* sh)
{
  SH2movls4(sh, 15,9, 6);
}
#define init_1_15_9_7
void FASTCALL call_1_15_9_7(SH2_struct* sh)
{
  SH2movls4(sh, 15,9, 7);
}
#define init_1_15_9_8
void FASTCALL call_1_15_9_8(SH2_struct* sh)
{
  SH2movls4(sh, 15,9, 8);
}
#define init_1_15_9_9
void FASTCALL call_1_15_9_9(SH2_struct* sh)
{
  SH2movls4(sh, 15,9, 9);
}
#define init_1_15_9_10
void FASTCALL call_1_15_9_10(SH2_struct* sh)
{
  SH2movls4(sh, 15,9, 10);
}
#define init_1_15_9_11
void FASTCALL call_1_15_9_11(SH2_struct* sh)
{
  SH2movls4(sh, 15,9, 11);
}
#define init_1_15_9_12
void FASTCALL call_1_15_9_12(SH2_struct* sh)
{
  SH2movls4(sh, 15,9, 12);
}
#define init_1_15_9_13
void FASTCALL call_1_15_9_13(SH2_struct* sh)
{
  SH2movls4(sh, 15,9, 13);
}
#define init_1_15_9_14
void FASTCALL call_1_15_9_14(SH2_struct* sh)
{
  SH2movls4(sh, 15,9, 14);
}
#define init_1_15_9_15
void FASTCALL call_1_15_9_15(SH2_struct* sh)
{
  SH2movls4(sh, 15,9, 15);
}
#define init_1_15_10_0
void FASTCALL call_1_15_10_0(SH2_struct* sh)
{
  SH2movls4(sh, 15,10, 0);
}
#define init_1_15_10_1
void FASTCALL call_1_15_10_1(SH2_struct* sh)
{
  SH2movls4(sh, 15,10, 1);
}
#define init_1_15_10_2
void FASTCALL call_1_15_10_2(SH2_struct* sh)
{
  SH2movls4(sh, 15,10, 2);
}
#define init_1_15_10_3
void FASTCALL call_1_15_10_3(SH2_struct* sh)
{
  SH2movls4(sh, 15,10, 3);
}
#define init_1_15_10_4
void FASTCALL call_1_15_10_4(SH2_struct* sh)
{
  SH2movls4(sh, 15,10, 4);
}
#define init_1_15_10_5
void FASTCALL call_1_15_10_5(SH2_struct* sh)
{
  SH2movls4(sh, 15,10, 5);
}
#define init_1_15_10_6
void FASTCALL call_1_15_10_6(SH2_struct* sh)
{
  SH2movls4(sh, 15,10, 6);
}
#define init_1_15_10_7
void FASTCALL call_1_15_10_7(SH2_struct* sh)
{
  SH2movls4(sh, 15,10, 7);
}
#define init_1_15_10_8
void FASTCALL call_1_15_10_8(SH2_struct* sh)
{
  SH2movls4(sh, 15,10, 8);
}
#define init_1_15_10_9
void FASTCALL call_1_15_10_9(SH2_struct* sh)
{
  SH2movls4(sh, 15,10, 9);
}
#define init_1_15_10_10
void FASTCALL call_1_15_10_10(SH2_struct* sh)
{
  SH2movls4(sh, 15,10, 10);
}
#define init_1_15_10_11
void FASTCALL call_1_15_10_11(SH2_struct* sh)
{
  SH2movls4(sh, 15,10, 11);
}
#define init_1_15_10_12
void FASTCALL call_1_15_10_12(SH2_struct* sh)
{
  SH2movls4(sh, 15,10, 12);
}
#define init_1_15_10_13
void FASTCALL call_1_15_10_13(SH2_struct* sh)
{
  SH2movls4(sh, 15,10, 13);
}
#define init_1_15_10_14
void FASTCALL call_1_15_10_14(SH2_struct* sh)
{
  SH2movls4(sh, 15,10, 14);
}
#define init_1_15_10_15
void FASTCALL call_1_15_10_15(SH2_struct* sh)
{
  SH2movls4(sh, 15,10, 15);
}
#define init_1_15_11_0
void FASTCALL call_1_15_11_0(SH2_struct* sh)
{
  SH2movls4(sh, 15,11, 0);
}
#define init_1_15_11_1
void FASTCALL call_1_15_11_1(SH2_struct* sh)
{
  SH2movls4(sh, 15,11, 1);
}
#define init_1_15_11_2
void FASTCALL call_1_15_11_2(SH2_struct* sh)
{
  SH2movls4(sh, 15,11, 2);
}
#define init_1_15_11_3
void FASTCALL call_1_15_11_3(SH2_struct* sh)
{
  SH2movls4(sh, 15,11, 3);
}
#define init_1_15_11_4
void FASTCALL call_1_15_11_4(SH2_struct* sh)
{
  SH2movls4(sh, 15,11, 4);
}
#define init_1_15_11_5
void FASTCALL call_1_15_11_5(SH2_struct* sh)
{
  SH2movls4(sh, 15,11, 5);
}
#define init_1_15_11_6
void FASTCALL call_1_15_11_6(SH2_struct* sh)
{
  SH2movls4(sh, 15,11, 6);
}
#define init_1_15_11_7
void FASTCALL call_1_15_11_7(SH2_struct* sh)
{
  SH2movls4(sh, 15,11, 7);
}
#define init_1_15_11_8
void FASTCALL call_1_15_11_8(SH2_struct* sh)
{
  SH2movls4(sh, 15,11, 8);
}
#define init_1_15_11_9
void FASTCALL call_1_15_11_9(SH2_struct* sh)
{
  SH2movls4(sh, 15,11, 9);
}
#define init_1_15_11_10
void FASTCALL call_1_15_11_10(SH2_struct* sh)
{
  SH2movls4(sh, 15,11, 10);
}
#define init_1_15_11_11
void FASTCALL call_1_15_11_11(SH2_struct* sh)
{
  SH2movls4(sh, 15,11, 11);
}
#define init_1_15_11_12
void FASTCALL call_1_15_11_12(SH2_struct* sh)
{
  SH2movls4(sh, 15,11, 12);
}
#define init_1_15_11_13
void FASTCALL call_1_15_11_13(SH2_struct* sh)
{
  SH2movls4(sh, 15,11, 13);
}
#define init_1_15_11_14
void FASTCALL call_1_15_11_14(SH2_struct* sh)
{
  SH2movls4(sh, 15,11, 14);
}
#define init_1_15_11_15
void FASTCALL call_1_15_11_15(SH2_struct* sh)
{
  SH2movls4(sh, 15,11, 15);
}
#define init_1_15_12_0
void FASTCALL call_1_15_12_0(SH2_struct* sh)
{
  SH2movls4(sh, 15,12, 0);
}
#define init_1_15_12_1
void FASTCALL call_1_15_12_1(SH2_struct* sh)
{
  SH2movls4(sh, 15,12, 1);
}
#define init_1_15_12_2
void FASTCALL call_1_15_12_2(SH2_struct* sh)
{
  SH2movls4(sh, 15,12, 2);
}
#define init_1_15_12_3
void FASTCALL call_1_15_12_3(SH2_struct* sh)
{
  SH2movls4(sh, 15,12, 3);
}
#define init_1_15_12_4
void FASTCALL call_1_15_12_4(SH2_struct* sh)
{
  SH2movls4(sh, 15,12, 4);
}
#define init_1_15_12_5
void FASTCALL call_1_15_12_5(SH2_struct* sh)
{
  SH2movls4(sh, 15,12, 5);
}
#define init_1_15_12_6
void FASTCALL call_1_15_12_6(SH2_struct* sh)
{
  SH2movls4(sh, 15,12, 6);
}
#define init_1_15_12_7
void FASTCALL call_1_15_12_7(SH2_struct* sh)
{
  SH2movls4(sh, 15,12, 7);
}
#define init_1_15_12_8
void FASTCALL call_1_15_12_8(SH2_struct* sh)
{
  SH2movls4(sh, 15,12, 8);
}
#define init_1_15_12_9
void FASTCALL call_1_15_12_9(SH2_struct* sh)
{
  SH2movls4(sh, 15,12, 9);
}
#define init_1_15_12_10
void FASTCALL call_1_15_12_10(SH2_struct* sh)
{
  SH2movls4(sh, 15,12, 10);
}
#define init_1_15_12_11
void FASTCALL call_1_15_12_11(SH2_struct* sh)
{
  SH2movls4(sh, 15,12, 11);
}
#define init_1_15_12_12
void FASTCALL call_1_15_12_12(SH2_struct* sh)
{
  SH2movls4(sh, 15,12, 12);
}
#define init_1_15_12_13
void FASTCALL call_1_15_12_13(SH2_struct* sh)
{
  SH2movls4(sh, 15,12, 13);
}
#define init_1_15_12_14
void FASTCALL call_1_15_12_14(SH2_struct* sh)
{
  SH2movls4(sh, 15,12, 14);
}
#define init_1_15_12_15
void FASTCALL call_1_15_12_15(SH2_struct* sh)
{
  SH2movls4(sh, 15,12, 15);
}
#define init_1_15_13_0
void FASTCALL call_1_15_13_0(SH2_struct* sh)
{
  SH2movls4(sh, 15,13, 0);
}
#define init_1_15_13_1
void FASTCALL call_1_15_13_1(SH2_struct* sh)
{
  SH2movls4(sh, 15,13, 1);
}
#define init_1_15_13_2
void FASTCALL call_1_15_13_2(SH2_struct* sh)
{
  SH2movls4(sh, 15,13, 2);
}
#define init_1_15_13_3
void FASTCALL call_1_15_13_3(SH2_struct* sh)
{
  SH2movls4(sh, 15,13, 3);
}
#define init_1_15_13_4
void FASTCALL call_1_15_13_4(SH2_struct* sh)
{
  SH2movls4(sh, 15,13, 4);
}
#define init_1_15_13_5
void FASTCALL call_1_15_13_5(SH2_struct* sh)
{
  SH2movls4(sh, 15,13, 5);
}
#define init_1_15_13_6
void FASTCALL call_1_15_13_6(SH2_struct* sh)
{
  SH2movls4(sh, 15,13, 6);
}
#define init_1_15_13_7
void FASTCALL call_1_15_13_7(SH2_struct* sh)
{
  SH2movls4(sh, 15,13, 7);
}
#define init_1_15_13_8
void FASTCALL call_1_15_13_8(SH2_struct* sh)
{
  SH2movls4(sh, 15,13, 8);
}
#define init_1_15_13_9
void FASTCALL call_1_15_13_9(SH2_struct* sh)
{
  SH2movls4(sh, 15,13, 9);
}
#define init_1_15_13_10
void FASTCALL call_1_15_13_10(SH2_struct* sh)
{
  SH2movls4(sh, 15,13, 10);
}
#define init_1_15_13_11
void FASTCALL call_1_15_13_11(SH2_struct* sh)
{
  SH2movls4(sh, 15,13, 11);
}
#define init_1_15_13_12
void FASTCALL call_1_15_13_12(SH2_struct* sh)
{
  SH2movls4(sh, 15,13, 12);
}
#define init_1_15_13_13
void FASTCALL call_1_15_13_13(SH2_struct* sh)
{
  SH2movls4(sh, 15,13, 13);
}
#define init_1_15_13_14
void FASTCALL call_1_15_13_14(SH2_struct* sh)
{
  SH2movls4(sh, 15,13, 14);
}
#define init_1_15_13_15
void FASTCALL call_1_15_13_15(SH2_struct* sh)
{
  SH2movls4(sh, 15,13, 15);
}
#define init_1_15_14_0
void FASTCALL call_1_15_14_0(SH2_struct* sh)
{
  SH2movls4(sh, 15,14, 0);
}
#define init_1_15_14_1
void FASTCALL call_1_15_14_1(SH2_struct* sh)
{
  SH2movls4(sh, 15,14, 1);
}
#define init_1_15_14_2
void FASTCALL call_1_15_14_2(SH2_struct* sh)
{
  SH2movls4(sh, 15,14, 2);
}
#define init_1_15_14_3
void FASTCALL call_1_15_14_3(SH2_struct* sh)
{
  SH2movls4(sh, 15,14, 3);
}
#define init_1_15_14_4
void FASTCALL call_1_15_14_4(SH2_struct* sh)
{
  SH2movls4(sh, 15,14, 4);
}
#define init_1_15_14_5
void FASTCALL call_1_15_14_5(SH2_struct* sh)
{
  SH2movls4(sh, 15,14, 5);
}
#define init_1_15_14_6
void FASTCALL call_1_15_14_6(SH2_struct* sh)
{
  SH2movls4(sh, 15,14, 6);
}
#define init_1_15_14_7
void FASTCALL call_1_15_14_7(SH2_struct* sh)
{
  SH2movls4(sh, 15,14, 7);
}
#define init_1_15_14_8
void FASTCALL call_1_15_14_8(SH2_struct* sh)
{
  SH2movls4(sh, 15,14, 8);
}
#define init_1_15_14_9
void FASTCALL call_1_15_14_9(SH2_struct* sh)
{
  SH2movls4(sh, 15,14, 9);
}
#define init_1_15_14_10
void FASTCALL call_1_15_14_10(SH2_struct* sh)
{
  SH2movls4(sh, 15,14, 10);
}
#define init_1_15_14_11
void FASTCALL call_1_15_14_11(SH2_struct* sh)
{
  SH2movls4(sh, 15,14, 11);
}
#define init_1_15_14_12
void FASTCALL call_1_15_14_12(SH2_struct* sh)
{
  SH2movls4(sh, 15,14, 12);
}
#define init_1_15_14_13
void FASTCALL call_1_15_14_13(SH2_struct* sh)
{
  SH2movls4(sh, 15,14, 13);
}
#define init_1_15_14_14
void FASTCALL call_1_15_14_14(SH2_struct* sh)
{
  SH2movls4(sh, 15,14, 14);
}
#define init_1_15_14_15
void FASTCALL call_1_15_14_15(SH2_struct* sh)
{
  SH2movls4(sh, 15,14, 15);
}
#define init_1_15_15_0
void FASTCALL call_1_15_15_0(SH2_struct* sh)
{
  SH2movls4(sh, 15,15, 0);
}
#define init_1_15_15_1
void FASTCALL call_1_15_15_1(SH2_struct* sh)
{
  SH2movls4(sh, 15,15, 1);
}
#define init_1_15_15_2
void FASTCALL call_1_15_15_2(SH2_struct* sh)
{
  SH2movls4(sh, 15,15, 2);
}
#define init_1_15_15_3
void FASTCALL call_1_15_15_3(SH2_struct* sh)
{
  SH2movls4(sh, 15,15, 3);
}
#define init_1_15_15_4
void FASTCALL call_1_15_15_4(SH2_struct* sh)
{
  SH2movls4(sh, 15,15, 4);
}
#define init_1_15_15_5
void FASTCALL call_1_15_15_5(SH2_struct* sh)
{
  SH2movls4(sh, 15,15, 5);
}
#define init_1_15_15_6
void FASTCALL call_1_15_15_6(SH2_struct* sh)
{
  SH2movls4(sh, 15,15, 6);
}
#define init_1_15_15_7
void FASTCALL call_1_15_15_7(SH2_struct* sh)
{
  SH2movls4(sh, 15,15, 7);
}
#define init_1_15_15_8
void FASTCALL call_1_15_15_8(SH2_struct* sh)
{
  SH2movls4(sh, 15,15, 8);
}
#define init_1_15_15_9
void FASTCALL call_1_15_15_9(SH2_struct* sh)
{
  SH2movls4(sh, 15,15, 9);
}
#define init_1_15_15_10
void FASTCALL call_1_15_15_10(SH2_struct* sh)
{
  SH2movls4(sh, 15,15, 10);
}
#define init_1_15_15_11
void FASTCALL call_1_15_15_11(SH2_struct* sh)
{
  SH2movls4(sh, 15,15, 11);
}
#define init_1_15_15_12
void FASTCALL call_1_15_15_12(SH2_struct* sh)
{
  SH2movls4(sh, 15,15, 12);
}
#define init_1_15_15_13
void FASTCALL call_1_15_15_13(SH2_struct* sh)
{
  SH2movls4(sh, 15,15, 13);
}
#define init_1_15_15_14
void FASTCALL call_1_15_15_14(SH2_struct* sh)
{
  SH2movls4(sh, 15,15, 14);
}
#define init_1_15_15_15
void FASTCALL call_1_15_15_15(SH2_struct* sh)
{
  SH2movls4(sh, 15,15, 15);
}
#define init_2_0_0_2
void FASTCALL call_2_0_0_2(SH2_struct* sh)
{
  SH2movls(sh,0,0);
}
#define init_2_0_1_2
void FASTCALL call_2_0_1_2(SH2_struct* sh)
{
  SH2movls(sh,0,1);
}
#define init_2_0_2_2
void FASTCALL call_2_0_2_2(SH2_struct* sh)
{
  SH2movls(sh,0,2);
}
#define init_2_0_3_2
void FASTCALL call_2_0_3_2(SH2_struct* sh)
{
  SH2movls(sh,0,3);
}
#define init_2_0_4_2
void FASTCALL call_2_0_4_2(SH2_struct* sh)
{
  SH2movls(sh,0,4);
}
#define init_2_0_5_2
void FASTCALL call_2_0_5_2(SH2_struct* sh)
{
  SH2movls(sh,0,5);
}
#define init_2_0_6_2
void FASTCALL call_2_0_6_2(SH2_struct* sh)
{
  SH2movls(sh,0,6);
}
#define init_2_0_7_2
void FASTCALL call_2_0_7_2(SH2_struct* sh)
{
  SH2movls(sh,0,7);
}
#define init_2_0_8_2
void FASTCALL call_2_0_8_2(SH2_struct* sh)
{
  SH2movls(sh,0,8);
}
#define init_2_0_9_2
void FASTCALL call_2_0_9_2(SH2_struct* sh)
{
  SH2movls(sh,0,9);
}
#define init_2_0_10_2
void FASTCALL call_2_0_10_2(SH2_struct* sh)
{
  SH2movls(sh,0,10);
}
#define init_2_0_11_2
void FASTCALL call_2_0_11_2(SH2_struct* sh)
{
  SH2movls(sh,0,11);
}
#define init_2_0_12_2
void FASTCALL call_2_0_12_2(SH2_struct* sh)
{
  SH2movls(sh,0,12);
}
#define init_2_0_13_2
void FASTCALL call_2_0_13_2(SH2_struct* sh)
{
  SH2movls(sh,0,13);
}
#define init_2_0_14_2
void FASTCALL call_2_0_14_2(SH2_struct* sh)
{
  SH2movls(sh,0,14);
}
#define init_2_0_15_2
void FASTCALL call_2_0_15_2(SH2_struct* sh)
{
  SH2movls(sh,0,15);
}
#define init_2_1_0_2
void FASTCALL call_2_1_0_2(SH2_struct* sh)
{
  SH2movls(sh,1,0);
}
#define init_2_1_1_2
void FASTCALL call_2_1_1_2(SH2_struct* sh)
{
  SH2movls(sh,1,1);
}
#define init_2_1_2_2
void FASTCALL call_2_1_2_2(SH2_struct* sh)
{
  SH2movls(sh,1,2);
}
#define init_2_1_3_2
void FASTCALL call_2_1_3_2(SH2_struct* sh)
{
  SH2movls(sh,1,3);
}
#define init_2_1_4_2
void FASTCALL call_2_1_4_2(SH2_struct* sh)
{
  SH2movls(sh,1,4);
}
#define init_2_1_5_2
void FASTCALL call_2_1_5_2(SH2_struct* sh)
{
  SH2movls(sh,1,5);
}
#define init_2_1_6_2
void FASTCALL call_2_1_6_2(SH2_struct* sh)
{
  SH2movls(sh,1,6);
}
#define init_2_1_7_2
void FASTCALL call_2_1_7_2(SH2_struct* sh)
{
  SH2movls(sh,1,7);
}
#define init_2_1_8_2
void FASTCALL call_2_1_8_2(SH2_struct* sh)
{
  SH2movls(sh,1,8);
}
#define init_2_1_9_2
void FASTCALL call_2_1_9_2(SH2_struct* sh)
{
  SH2movls(sh,1,9);
}
#define init_2_1_10_2
void FASTCALL call_2_1_10_2(SH2_struct* sh)
{
  SH2movls(sh,1,10);
}
#define init_2_1_11_2
void FASTCALL call_2_1_11_2(SH2_struct* sh)
{
  SH2movls(sh,1,11);
}
#define init_2_1_12_2
void FASTCALL call_2_1_12_2(SH2_struct* sh)
{
  SH2movls(sh,1,12);
}
#define init_2_1_13_2
void FASTCALL call_2_1_13_2(SH2_struct* sh)
{
  SH2movls(sh,1,13);
}
#define init_2_1_14_2
void FASTCALL call_2_1_14_2(SH2_struct* sh)
{
  SH2movls(sh,1,14);
}
#define init_2_1_15_2
void FASTCALL call_2_1_15_2(SH2_struct* sh)
{
  SH2movls(sh,1,15);
}
#define init_2_2_0_2
void FASTCALL call_2_2_0_2(SH2_struct* sh)
{
  SH2movls(sh,2,0);
}
#define init_2_2_1_2
void FASTCALL call_2_2_1_2(SH2_struct* sh)
{
  SH2movls(sh,2,1);
}
#define init_2_2_2_2
void FASTCALL call_2_2_2_2(SH2_struct* sh)
{
  SH2movls(sh,2,2);
}
#define init_2_2_3_2
void FASTCALL call_2_2_3_2(SH2_struct* sh)
{
  SH2movls(sh,2,3);
}
#define init_2_2_4_2
void FASTCALL call_2_2_4_2(SH2_struct* sh)
{
  SH2movls(sh,2,4);
}
#define init_2_2_5_2
void FASTCALL call_2_2_5_2(SH2_struct* sh)
{
  SH2movls(sh,2,5);
}
#define init_2_2_6_2
void FASTCALL call_2_2_6_2(SH2_struct* sh)
{
  SH2movls(sh,2,6);
}
#define init_2_2_7_2
void FASTCALL call_2_2_7_2(SH2_struct* sh)
{
  SH2movls(sh,2,7);
}
#define init_2_2_8_2
void FASTCALL call_2_2_8_2(SH2_struct* sh)
{
  SH2movls(sh,2,8);
}
#define init_2_2_9_2
void FASTCALL call_2_2_9_2(SH2_struct* sh)
{
  SH2movls(sh,2,9);
}
#define init_2_2_10_2
void FASTCALL call_2_2_10_2(SH2_struct* sh)
{
  SH2movls(sh,2,10);
}
#define init_2_2_11_2
void FASTCALL call_2_2_11_2(SH2_struct* sh)
{
  SH2movls(sh,2,11);
}
#define init_2_2_12_2
void FASTCALL call_2_2_12_2(SH2_struct* sh)
{
  SH2movls(sh,2,12);
}
#define init_2_2_13_2
void FASTCALL call_2_2_13_2(SH2_struct* sh)
{
  SH2movls(sh,2,13);
}
#define init_2_2_14_2
void FASTCALL call_2_2_14_2(SH2_struct* sh)
{
  SH2movls(sh,2,14);
}
#define init_2_2_15_2
void FASTCALL call_2_2_15_2(SH2_struct* sh)
{
  SH2movls(sh,2,15);
}
#define init_2_3_0_2
void FASTCALL call_2_3_0_2(SH2_struct* sh)
{
  SH2movls(sh,3,0);
}
#define init_2_3_1_2
void FASTCALL call_2_3_1_2(SH2_struct* sh)
{
  SH2movls(sh,3,1);
}
#define init_2_3_2_2
void FASTCALL call_2_3_2_2(SH2_struct* sh)
{
  SH2movls(sh,3,2);
}
#define init_2_3_3_2
void FASTCALL call_2_3_3_2(SH2_struct* sh)
{
  SH2movls(sh,3,3);
}
#define init_2_3_4_2
void FASTCALL call_2_3_4_2(SH2_struct* sh)
{
  SH2movls(sh,3,4);
}
#define init_2_3_5_2
void FASTCALL call_2_3_5_2(SH2_struct* sh)
{
  SH2movls(sh,3,5);
}
#define init_2_3_6_2
void FASTCALL call_2_3_6_2(SH2_struct* sh)
{
  SH2movls(sh,3,6);
}
#define init_2_3_7_2
void FASTCALL call_2_3_7_2(SH2_struct* sh)
{
  SH2movls(sh,3,7);
}
#define init_2_3_8_2
void FASTCALL call_2_3_8_2(SH2_struct* sh)
{
  SH2movls(sh,3,8);
}
#define init_2_3_9_2
void FASTCALL call_2_3_9_2(SH2_struct* sh)
{
  SH2movls(sh,3,9);
}
#define init_2_3_10_2
void FASTCALL call_2_3_10_2(SH2_struct* sh)
{
  SH2movls(sh,3,10);
}
#define init_2_3_11_2
void FASTCALL call_2_3_11_2(SH2_struct* sh)
{
  SH2movls(sh,3,11);
}
#define init_2_3_12_2
void FASTCALL call_2_3_12_2(SH2_struct* sh)
{
  SH2movls(sh,3,12);
}
#define init_2_3_13_2
void FASTCALL call_2_3_13_2(SH2_struct* sh)
{
  SH2movls(sh,3,13);
}
#define init_2_3_14_2
void FASTCALL call_2_3_14_2(SH2_struct* sh)
{
  SH2movls(sh,3,14);
}
#define init_2_3_15_2
void FASTCALL call_2_3_15_2(SH2_struct* sh)
{
  SH2movls(sh,3,15);
}
#define init_2_4_0_2
void FASTCALL call_2_4_0_2(SH2_struct* sh)
{
  SH2movls(sh,4,0);
}
#define init_2_4_1_2
void FASTCALL call_2_4_1_2(SH2_struct* sh)
{
  SH2movls(sh,4,1);
}
#define init_2_4_2_2
void FASTCALL call_2_4_2_2(SH2_struct* sh)
{
  SH2movls(sh,4,2);
}
#define init_2_4_3_2
void FASTCALL call_2_4_3_2(SH2_struct* sh)
{
  SH2movls(sh,4,3);
}
#define init_2_4_4_2
void FASTCALL call_2_4_4_2(SH2_struct* sh)
{
  SH2movls(sh,4,4);
}
#define init_2_4_5_2
void FASTCALL call_2_4_5_2(SH2_struct* sh)
{
  SH2movls(sh,4,5);
}
#define init_2_4_6_2
void FASTCALL call_2_4_6_2(SH2_struct* sh)
{
  SH2movls(sh,4,6);
}
#define init_2_4_7_2
void FASTCALL call_2_4_7_2(SH2_struct* sh)
{
  SH2movls(sh,4,7);
}
#define init_2_4_8_2
void FASTCALL call_2_4_8_2(SH2_struct* sh)
{
  SH2movls(sh,4,8);
}
#define init_2_4_9_2
void FASTCALL call_2_4_9_2(SH2_struct* sh)
{
  SH2movls(sh,4,9);
}
#define init_2_4_10_2
void FASTCALL call_2_4_10_2(SH2_struct* sh)
{
  SH2movls(sh,4,10);
}
#define init_2_4_11_2
void FASTCALL call_2_4_11_2(SH2_struct* sh)
{
  SH2movls(sh,4,11);
}
#define init_2_4_12_2
void FASTCALL call_2_4_12_2(SH2_struct* sh)
{
  SH2movls(sh,4,12);
}
#define init_2_4_13_2
void FASTCALL call_2_4_13_2(SH2_struct* sh)
{
  SH2movls(sh,4,13);
}
#define init_2_4_14_2
void FASTCALL call_2_4_14_2(SH2_struct* sh)
{
  SH2movls(sh,4,14);
}
#define init_2_4_15_2
void FASTCALL call_2_4_15_2(SH2_struct* sh)
{
  SH2movls(sh,4,15);
}
#define init_2_5_0_2
void FASTCALL call_2_5_0_2(SH2_struct* sh)
{
  SH2movls(sh,5,0);
}
#define init_2_5_1_2
void FASTCALL call_2_5_1_2(SH2_struct* sh)
{
  SH2movls(sh,5,1);
}
#define init_2_5_2_2
void FASTCALL call_2_5_2_2(SH2_struct* sh)
{
  SH2movls(sh,5,2);
}
#define init_2_5_3_2
void FASTCALL call_2_5_3_2(SH2_struct* sh)
{
  SH2movls(sh,5,3);
}
#define init_2_5_4_2
void FASTCALL call_2_5_4_2(SH2_struct* sh)
{
  SH2movls(sh,5,4);
}
#define init_2_5_5_2
void FASTCALL call_2_5_5_2(SH2_struct* sh)
{
  SH2movls(sh,5,5);
}
#define init_2_5_6_2
void FASTCALL call_2_5_6_2(SH2_struct* sh)
{
  SH2movls(sh,5,6);
}
#define init_2_5_7_2
void FASTCALL call_2_5_7_2(SH2_struct* sh)
{
  SH2movls(sh,5,7);
}
#define init_2_5_8_2
void FASTCALL call_2_5_8_2(SH2_struct* sh)
{
  SH2movls(sh,5,8);
}
#define init_2_5_9_2
void FASTCALL call_2_5_9_2(SH2_struct* sh)
{
  SH2movls(sh,5,9);
}
#define init_2_5_10_2
void FASTCALL call_2_5_10_2(SH2_struct* sh)
{
  SH2movls(sh,5,10);
}
#define init_2_5_11_2
void FASTCALL call_2_5_11_2(SH2_struct* sh)
{
  SH2movls(sh,5,11);
}
#define init_2_5_12_2
void FASTCALL call_2_5_12_2(SH2_struct* sh)
{
  SH2movls(sh,5,12);
}
#define init_2_5_13_2
void FASTCALL call_2_5_13_2(SH2_struct* sh)
{
  SH2movls(sh,5,13);
}
#define init_2_5_14_2
void FASTCALL call_2_5_14_2(SH2_struct* sh)
{
  SH2movls(sh,5,14);
}
#define init_2_5_15_2
void FASTCALL call_2_5_15_2(SH2_struct* sh)
{
  SH2movls(sh,5,15);
}
#define init_2_6_0_2
void FASTCALL call_2_6_0_2(SH2_struct* sh)
{
  SH2movls(sh,6,0);
}
#define init_2_6_1_2
void FASTCALL call_2_6_1_2(SH2_struct* sh)
{
  SH2movls(sh,6,1);
}
#define init_2_6_2_2
void FASTCALL call_2_6_2_2(SH2_struct* sh)
{
  SH2movls(sh,6,2);
}
#define init_2_6_3_2
void FASTCALL call_2_6_3_2(SH2_struct* sh)
{
  SH2movls(sh,6,3);
}
#define init_2_6_4_2
void FASTCALL call_2_6_4_2(SH2_struct* sh)
{
  SH2movls(sh,6,4);
}
#define init_2_6_5_2
void FASTCALL call_2_6_5_2(SH2_struct* sh)
{
  SH2movls(sh,6,5);
}
#define init_2_6_6_2
void FASTCALL call_2_6_6_2(SH2_struct* sh)
{
  SH2movls(sh,6,6);
}
#define init_2_6_7_2
void FASTCALL call_2_6_7_2(SH2_struct* sh)
{
  SH2movls(sh,6,7);
}
#define init_2_6_8_2
void FASTCALL call_2_6_8_2(SH2_struct* sh)
{
  SH2movls(sh,6,8);
}
#define init_2_6_9_2
void FASTCALL call_2_6_9_2(SH2_struct* sh)
{
  SH2movls(sh,6,9);
}
#define init_2_6_10_2
void FASTCALL call_2_6_10_2(SH2_struct* sh)
{
  SH2movls(sh,6,10);
}
#define init_2_6_11_2
void FASTCALL call_2_6_11_2(SH2_struct* sh)
{
  SH2movls(sh,6,11);
}
#define init_2_6_12_2
void FASTCALL call_2_6_12_2(SH2_struct* sh)
{
  SH2movls(sh,6,12);
}
#define init_2_6_13_2
void FASTCALL call_2_6_13_2(SH2_struct* sh)
{
  SH2movls(sh,6,13);
}
#define init_2_6_14_2
void FASTCALL call_2_6_14_2(SH2_struct* sh)
{
  SH2movls(sh,6,14);
}
#define init_2_6_15_2
void FASTCALL call_2_6_15_2(SH2_struct* sh)
{
  SH2movls(sh,6,15);
}
#define init_2_7_0_2
void FASTCALL call_2_7_0_2(SH2_struct* sh)
{
  SH2movls(sh,7,0);
}
#define init_2_7_1_2
void FASTCALL call_2_7_1_2(SH2_struct* sh)
{
  SH2movls(sh,7,1);
}
#define init_2_7_2_2
void FASTCALL call_2_7_2_2(SH2_struct* sh)
{
  SH2movls(sh,7,2);
}
#define init_2_7_3_2
void FASTCALL call_2_7_3_2(SH2_struct* sh)
{
  SH2movls(sh,7,3);
}
#define init_2_7_4_2
void FASTCALL call_2_7_4_2(SH2_struct* sh)
{
  SH2movls(sh,7,4);
}
#define init_2_7_5_2
void FASTCALL call_2_7_5_2(SH2_struct* sh)
{
  SH2movls(sh,7,5);
}
#define init_2_7_6_2
void FASTCALL call_2_7_6_2(SH2_struct* sh)
{
  SH2movls(sh,7,6);
}
#define init_2_7_7_2
void FASTCALL call_2_7_7_2(SH2_struct* sh)
{
  SH2movls(sh,7,7);
}
#define init_2_7_8_2
void FASTCALL call_2_7_8_2(SH2_struct* sh)
{
  SH2movls(sh,7,8);
}
#define init_2_7_9_2
void FASTCALL call_2_7_9_2(SH2_struct* sh)
{
  SH2movls(sh,7,9);
}
#define init_2_7_10_2
void FASTCALL call_2_7_10_2(SH2_struct* sh)
{
  SH2movls(sh,7,10);
}
#define init_2_7_11_2
void FASTCALL call_2_7_11_2(SH2_struct* sh)
{
  SH2movls(sh,7,11);
}
#define init_2_7_12_2
void FASTCALL call_2_7_12_2(SH2_struct* sh)
{
  SH2movls(sh,7,12);
}
#define init_2_7_13_2
void FASTCALL call_2_7_13_2(SH2_struct* sh)
{
  SH2movls(sh,7,13);
}
#define init_2_7_14_2
void FASTCALL call_2_7_14_2(SH2_struct* sh)
{
  SH2movls(sh,7,14);
}
#define init_2_7_15_2
void FASTCALL call_2_7_15_2(SH2_struct* sh)
{
  SH2movls(sh,7,15);
}
#define init_2_8_0_2
void FASTCALL call_2_8_0_2(SH2_struct* sh)
{
  SH2movls(sh,8,0);
}
#define init_2_8_1_2
void FASTCALL call_2_8_1_2(SH2_struct* sh)
{
  SH2movls(sh,8,1);
}
#define init_2_8_2_2
void FASTCALL call_2_8_2_2(SH2_struct* sh)
{
  SH2movls(sh,8,2);
}
#define init_2_8_3_2
void FASTCALL call_2_8_3_2(SH2_struct* sh)
{
  SH2movls(sh,8,3);
}
#define init_2_8_4_2
void FASTCALL call_2_8_4_2(SH2_struct* sh)
{
  SH2movls(sh,8,4);
}
#define init_2_8_5_2
void FASTCALL call_2_8_5_2(SH2_struct* sh)
{
  SH2movls(sh,8,5);
}
#define init_2_8_6_2
void FASTCALL call_2_8_6_2(SH2_struct* sh)
{
  SH2movls(sh,8,6);
}
#define init_2_8_7_2
void FASTCALL call_2_8_7_2(SH2_struct* sh)
{
  SH2movls(sh,8,7);
}
#define init_2_8_8_2
void FASTCALL call_2_8_8_2(SH2_struct* sh)
{
  SH2movls(sh,8,8);
}
#define init_2_8_9_2
void FASTCALL call_2_8_9_2(SH2_struct* sh)
{
  SH2movls(sh,8,9);
}
#define init_2_8_10_2
void FASTCALL call_2_8_10_2(SH2_struct* sh)
{
  SH2movls(sh,8,10);
}
#define init_2_8_11_2
void FASTCALL call_2_8_11_2(SH2_struct* sh)
{
  SH2movls(sh,8,11);
}
#define init_2_8_12_2
void FASTCALL call_2_8_12_2(SH2_struct* sh)
{
  SH2movls(sh,8,12);
}
#define init_2_8_13_2
void FASTCALL call_2_8_13_2(SH2_struct* sh)
{
  SH2movls(sh,8,13);
}
#define init_2_8_14_2
void FASTCALL call_2_8_14_2(SH2_struct* sh)
{
  SH2movls(sh,8,14);
}
#define init_2_8_15_2
void FASTCALL call_2_8_15_2(SH2_struct* sh)
{
  SH2movls(sh,8,15);
}
#define init_2_9_0_2
void FASTCALL call_2_9_0_2(SH2_struct* sh)
{
  SH2movls(sh,9,0);
}
#define init_2_9_1_2
void FASTCALL call_2_9_1_2(SH2_struct* sh)
{
  SH2movls(sh,9,1);
}
#define init_2_9_2_2
void FASTCALL call_2_9_2_2(SH2_struct* sh)
{
  SH2movls(sh,9,2);
}
#define init_2_9_3_2
void FASTCALL call_2_9_3_2(SH2_struct* sh)
{
  SH2movls(sh,9,3);
}
#define init_2_9_4_2
void FASTCALL call_2_9_4_2(SH2_struct* sh)
{
  SH2movls(sh,9,4);
}
#define init_2_9_5_2
void FASTCALL call_2_9_5_2(SH2_struct* sh)
{
  SH2movls(sh,9,5);
}
#define init_2_9_6_2
void FASTCALL call_2_9_6_2(SH2_struct* sh)
{
  SH2movls(sh,9,6);
}
#define init_2_9_7_2
void FASTCALL call_2_9_7_2(SH2_struct* sh)
{
  SH2movls(sh,9,7);
}
#define init_2_9_8_2
void FASTCALL call_2_9_8_2(SH2_struct* sh)
{
  SH2movls(sh,9,8);
}
#define init_2_9_9_2
void FASTCALL call_2_9_9_2(SH2_struct* sh)
{
  SH2movls(sh,9,9);
}
#define init_2_9_10_2
void FASTCALL call_2_9_10_2(SH2_struct* sh)
{
  SH2movls(sh,9,10);
}
#define init_2_9_11_2
void FASTCALL call_2_9_11_2(SH2_struct* sh)
{
  SH2movls(sh,9,11);
}
#define init_2_9_12_2
void FASTCALL call_2_9_12_2(SH2_struct* sh)
{
  SH2movls(sh,9,12);
}
#define init_2_9_13_2
void FASTCALL call_2_9_13_2(SH2_struct* sh)
{
  SH2movls(sh,9,13);
}
#define init_2_9_14_2
void FASTCALL call_2_9_14_2(SH2_struct* sh)
{
  SH2movls(sh,9,14);
}
#define init_2_9_15_2
void FASTCALL call_2_9_15_2(SH2_struct* sh)
{
  SH2movls(sh,9,15);
}
#define init_2_10_0_2
void FASTCALL call_2_10_0_2(SH2_struct* sh)
{
  SH2movls(sh,10,0);
}
#define init_2_10_1_2
void FASTCALL call_2_10_1_2(SH2_struct* sh)
{
  SH2movls(sh,10,1);
}
#define init_2_10_2_2
void FASTCALL call_2_10_2_2(SH2_struct* sh)
{
  SH2movls(sh,10,2);
}
#define init_2_10_3_2
void FASTCALL call_2_10_3_2(SH2_struct* sh)
{
  SH2movls(sh,10,3);
}
#define init_2_10_4_2
void FASTCALL call_2_10_4_2(SH2_struct* sh)
{
  SH2movls(sh,10,4);
}
#define init_2_10_5_2
void FASTCALL call_2_10_5_2(SH2_struct* sh)
{
  SH2movls(sh,10,5);
}
#define init_2_10_6_2
void FASTCALL call_2_10_6_2(SH2_struct* sh)
{
  SH2movls(sh,10,6);
}
#define init_2_10_7_2
void FASTCALL call_2_10_7_2(SH2_struct* sh)
{
  SH2movls(sh,10,7);
}
#define init_2_10_8_2
void FASTCALL call_2_10_8_2(SH2_struct* sh)
{
  SH2movls(sh,10,8);
}
#define init_2_10_9_2
void FASTCALL call_2_10_9_2(SH2_struct* sh)
{
  SH2movls(sh,10,9);
}
#define init_2_10_10_2
void FASTCALL call_2_10_10_2(SH2_struct* sh)
{
  SH2movls(sh,10,10);
}
#define init_2_10_11_2
void FASTCALL call_2_10_11_2(SH2_struct* sh)
{
  SH2movls(sh,10,11);
}
#define init_2_10_12_2
void FASTCALL call_2_10_12_2(SH2_struct* sh)
{
  SH2movls(sh,10,12);
}
#define init_2_10_13_2
void FASTCALL call_2_10_13_2(SH2_struct* sh)
{
  SH2movls(sh,10,13);
}
#define init_2_10_14_2
void FASTCALL call_2_10_14_2(SH2_struct* sh)
{
  SH2movls(sh,10,14);
}
#define init_2_10_15_2
void FASTCALL call_2_10_15_2(SH2_struct* sh)
{
  SH2movls(sh,10,15);
}
#define init_2_11_0_2
void FASTCALL call_2_11_0_2(SH2_struct* sh)
{
  SH2movls(sh,11,0);
}
#define init_2_11_1_2
void FASTCALL call_2_11_1_2(SH2_struct* sh)
{
  SH2movls(sh,11,1);
}
#define init_2_11_2_2
void FASTCALL call_2_11_2_2(SH2_struct* sh)
{
  SH2movls(sh,11,2);
}
#define init_2_11_3_2
void FASTCALL call_2_11_3_2(SH2_struct* sh)
{
  SH2movls(sh,11,3);
}
#define init_2_11_4_2
void FASTCALL call_2_11_4_2(SH2_struct* sh)
{
  SH2movls(sh,11,4);
}
#define init_2_11_5_2
void FASTCALL call_2_11_5_2(SH2_struct* sh)
{
  SH2movls(sh,11,5);
}
#define init_2_11_6_2
void FASTCALL call_2_11_6_2(SH2_struct* sh)
{
  SH2movls(sh,11,6);
}
#define init_2_11_7_2
void FASTCALL call_2_11_7_2(SH2_struct* sh)
{
  SH2movls(sh,11,7);
}
#define init_2_11_8_2
void FASTCALL call_2_11_8_2(SH2_struct* sh)
{
  SH2movls(sh,11,8);
}
#define init_2_11_9_2
void FASTCALL call_2_11_9_2(SH2_struct* sh)
{
  SH2movls(sh,11,9);
}
#define init_2_11_10_2
void FASTCALL call_2_11_10_2(SH2_struct* sh)
{
  SH2movls(sh,11,10);
}
#define init_2_11_11_2
void FASTCALL call_2_11_11_2(SH2_struct* sh)
{
  SH2movls(sh,11,11);
}
#define init_2_11_12_2
void FASTCALL call_2_11_12_2(SH2_struct* sh)
{
  SH2movls(sh,11,12);
}
#define init_2_11_13_2
void FASTCALL call_2_11_13_2(SH2_struct* sh)
{
  SH2movls(sh,11,13);
}
#define init_2_11_14_2
void FASTCALL call_2_11_14_2(SH2_struct* sh)
{
  SH2movls(sh,11,14);
}
#define init_2_11_15_2
void FASTCALL call_2_11_15_2(SH2_struct* sh)
{
  SH2movls(sh,11,15);
}
#define init_2_12_0_2
void FASTCALL call_2_12_0_2(SH2_struct* sh)
{
  SH2movls(sh,12,0);
}
#define init_2_12_1_2
void FASTCALL call_2_12_1_2(SH2_struct* sh)
{
  SH2movls(sh,12,1);
}
#define init_2_12_2_2
void FASTCALL call_2_12_2_2(SH2_struct* sh)
{
  SH2movls(sh,12,2);
}
#define init_2_12_3_2
void FASTCALL call_2_12_3_2(SH2_struct* sh)
{
  SH2movls(sh,12,3);
}
#define init_2_12_4_2
void FASTCALL call_2_12_4_2(SH2_struct* sh)
{
  SH2movls(sh,12,4);
}
#define init_2_12_5_2
void FASTCALL call_2_12_5_2(SH2_struct* sh)
{
  SH2movls(sh,12,5);
}
#define init_2_12_6_2
void FASTCALL call_2_12_6_2(SH2_struct* sh)
{
  SH2movls(sh,12,6);
}
#define init_2_12_7_2
void FASTCALL call_2_12_7_2(SH2_struct* sh)
{
  SH2movls(sh,12,7);
}
#define init_2_12_8_2
void FASTCALL call_2_12_8_2(SH2_struct* sh)
{
  SH2movls(sh,12,8);
}
#define init_2_12_9_2
void FASTCALL call_2_12_9_2(SH2_struct* sh)
{
  SH2movls(sh,12,9);
}
#define init_2_12_10_2
void FASTCALL call_2_12_10_2(SH2_struct* sh)
{
  SH2movls(sh,12,10);
}
#define init_2_12_11_2
void FASTCALL call_2_12_11_2(SH2_struct* sh)
{
  SH2movls(sh,12,11);
}
#define init_2_12_12_2
void FASTCALL call_2_12_12_2(SH2_struct* sh)
{
  SH2movls(sh,12,12);
}
#define init_2_12_13_2
void FASTCALL call_2_12_13_2(SH2_struct* sh)
{
  SH2movls(sh,12,13);
}
#define init_2_12_14_2
void FASTCALL call_2_12_14_2(SH2_struct* sh)
{
  SH2movls(sh,12,14);
}
#define init_2_12_15_2
void FASTCALL call_2_12_15_2(SH2_struct* sh)
{
  SH2movls(sh,12,15);
}
#define init_2_13_0_2
void FASTCALL call_2_13_0_2(SH2_struct* sh)
{
  SH2movls(sh,13,0);
}
#define init_2_13_1_2
void FASTCALL call_2_13_1_2(SH2_struct* sh)
{
  SH2movls(sh,13,1);
}
#define init_2_13_2_2
void FASTCALL call_2_13_2_2(SH2_struct* sh)
{
  SH2movls(sh,13,2);
}
#define init_2_13_3_2
void FASTCALL call_2_13_3_2(SH2_struct* sh)
{
  SH2movls(sh,13,3);
}
#define init_2_13_4_2
void FASTCALL call_2_13_4_2(SH2_struct* sh)
{
  SH2movls(sh,13,4);
}
#define init_2_13_5_2
void FASTCALL call_2_13_5_2(SH2_struct* sh)
{
  SH2movls(sh,13,5);
}
#define init_2_13_6_2
void FASTCALL call_2_13_6_2(SH2_struct* sh)
{
  SH2movls(sh,13,6);
}
#define init_2_13_7_2
void FASTCALL call_2_13_7_2(SH2_struct* sh)
{
  SH2movls(sh,13,7);
}
#define init_2_13_8_2
void FASTCALL call_2_13_8_2(SH2_struct* sh)
{
  SH2movls(sh,13,8);
}
#define init_2_13_9_2
void FASTCALL call_2_13_9_2(SH2_struct* sh)
{
  SH2movls(sh,13,9);
}
#define init_2_13_10_2
void FASTCALL call_2_13_10_2(SH2_struct* sh)
{
  SH2movls(sh,13,10);
}
#define init_2_13_11_2
void FASTCALL call_2_13_11_2(SH2_struct* sh)
{
  SH2movls(sh,13,11);
}
#define init_2_13_12_2
void FASTCALL call_2_13_12_2(SH2_struct* sh)
{
  SH2movls(sh,13,12);
}
#define init_2_13_13_2
void FASTCALL call_2_13_13_2(SH2_struct* sh)
{
  SH2movls(sh,13,13);
}
#define init_2_13_14_2
void FASTCALL call_2_13_14_2(SH2_struct* sh)
{
  SH2movls(sh,13,14);
}
#define init_2_13_15_2
void FASTCALL call_2_13_15_2(SH2_struct* sh)
{
  SH2movls(sh,13,15);
}
#define init_2_14_0_2
void FASTCALL call_2_14_0_2(SH2_struct* sh)
{
  SH2movls(sh,14,0);
}
#define init_2_14_1_2
void FASTCALL call_2_14_1_2(SH2_struct* sh)
{
  SH2movls(sh,14,1);
}
#define init_2_14_2_2
void FASTCALL call_2_14_2_2(SH2_struct* sh)
{
  SH2movls(sh,14,2);
}
#define init_2_14_3_2
void FASTCALL call_2_14_3_2(SH2_struct* sh)
{
  SH2movls(sh,14,3);
}
#define init_2_14_4_2
void FASTCALL call_2_14_4_2(SH2_struct* sh)
{
  SH2movls(sh,14,4);
}
#define init_2_14_5_2
void FASTCALL call_2_14_5_2(SH2_struct* sh)
{
  SH2movls(sh,14,5);
}
#define init_2_14_6_2
void FASTCALL call_2_14_6_2(SH2_struct* sh)
{
  SH2movls(sh,14,6);
}
#define init_2_14_7_2
void FASTCALL call_2_14_7_2(SH2_struct* sh)
{
  SH2movls(sh,14,7);
}
#define init_2_14_8_2
void FASTCALL call_2_14_8_2(SH2_struct* sh)
{
  SH2movls(sh,14,8);
}
#define init_2_14_9_2
void FASTCALL call_2_14_9_2(SH2_struct* sh)
{
  SH2movls(sh,14,9);
}
#define init_2_14_10_2
void FASTCALL call_2_14_10_2(SH2_struct* sh)
{
  SH2movls(sh,14,10);
}
#define init_2_14_11_2
void FASTCALL call_2_14_11_2(SH2_struct* sh)
{
  SH2movls(sh,14,11);
}
#define init_2_14_12_2
void FASTCALL call_2_14_12_2(SH2_struct* sh)
{
  SH2movls(sh,14,12);
}
#define init_2_14_13_2
void FASTCALL call_2_14_13_2(SH2_struct* sh)
{
  SH2movls(sh,14,13);
}
#define init_2_14_14_2
void FASTCALL call_2_14_14_2(SH2_struct* sh)
{
  SH2movls(sh,14,14);
}
#define init_2_14_15_2
void FASTCALL call_2_14_15_2(SH2_struct* sh)
{
  SH2movls(sh,14,15);
}
#define init_2_15_0_2
void FASTCALL call_2_15_0_2(SH2_struct* sh)
{
  SH2movls(sh,15,0);
}
#define init_2_15_1_2
void FASTCALL call_2_15_1_2(SH2_struct* sh)
{
  SH2movls(sh,15,1);
}
#define init_2_15_2_2
void FASTCALL call_2_15_2_2(SH2_struct* sh)
{
  SH2movls(sh,15,2);
}
#define init_2_15_3_2
void FASTCALL call_2_15_3_2(SH2_struct* sh)
{
  SH2movls(sh,15,3);
}
#define init_2_15_4_2
void FASTCALL call_2_15_4_2(SH2_struct* sh)
{
  SH2movls(sh,15,4);
}
#define init_2_15_5_2
void FASTCALL call_2_15_5_2(SH2_struct* sh)
{
  SH2movls(sh,15,5);
}
#define init_2_15_6_2
void FASTCALL call_2_15_6_2(SH2_struct* sh)
{
  SH2movls(sh,15,6);
}
#define init_2_15_7_2
void FASTCALL call_2_15_7_2(SH2_struct* sh)
{
  SH2movls(sh,15,7);
}
#define init_2_15_8_2
void FASTCALL call_2_15_8_2(SH2_struct* sh)
{
  SH2movls(sh,15,8);
}
#define init_2_15_9_2
void FASTCALL call_2_15_9_2(SH2_struct* sh)
{
  SH2movls(sh,15,9);
}
#define init_2_15_10_2
void FASTCALL call_2_15_10_2(SH2_struct* sh)
{
  SH2movls(sh,15,10);
}
#define init_2_15_11_2
void FASTCALL call_2_15_11_2(SH2_struct* sh)
{
  SH2movls(sh,15,11);
}
#define init_2_15_12_2
void FASTCALL call_2_15_12_2(SH2_struct* sh)
{
  SH2movls(sh,15,12);
}
#define init_2_15_13_2
void FASTCALL call_2_15_13_2(SH2_struct* sh)
{
  SH2movls(sh,15,13);
}
#define init_2_15_14_2
void FASTCALL call_2_15_14_2(SH2_struct* sh)
{
  SH2movls(sh,15,14);
}
#define init_2_15_15_2
void FASTCALL call_2_15_15_2(SH2_struct* sh)
{
  SH2movls(sh,15,15);
}
#define init_6_0_0_6
void FASTCALL call_6_0_0_6(SH2_struct* sh)
{
  SH2movlp(sh,0,0);
}
#define init_6_0_1_6
void FASTCALL call_6_0_1_6(SH2_struct* sh)
{
  SH2movlp(sh,0,1);
}
#define init_6_0_2_6
void FASTCALL call_6_0_2_6(SH2_struct* sh)
{
  SH2movlp(sh,0,2);
}
#define init_6_0_3_6
void FASTCALL call_6_0_3_6(SH2_struct* sh)
{
  SH2movlp(sh,0,3);
}
#define init_6_0_4_6
void FASTCALL call_6_0_4_6(SH2_struct* sh)
{
  SH2movlp(sh,0,4);
}
#define init_6_0_5_6
void FASTCALL call_6_0_5_6(SH2_struct* sh)
{
  SH2movlp(sh,0,5);
}
#define init_6_0_6_6
void FASTCALL call_6_0_6_6(SH2_struct* sh)
{
  SH2movlp(sh,0,6);
}
#define init_6_0_7_6
void FASTCALL call_6_0_7_6(SH2_struct* sh)
{
  SH2movlp(sh,0,7);
}
#define init_6_0_8_6
void FASTCALL call_6_0_8_6(SH2_struct* sh)
{
  SH2movlp(sh,0,8);
}
#define init_6_0_9_6
void FASTCALL call_6_0_9_6(SH2_struct* sh)
{
  SH2movlp(sh,0,9);
}
#define init_6_0_10_6
void FASTCALL call_6_0_10_6(SH2_struct* sh)
{
  SH2movlp(sh,0,10);
}
#define init_6_0_11_6
void FASTCALL call_6_0_11_6(SH2_struct* sh)
{
  SH2movlp(sh,0,11);
}
#define init_6_0_12_6
void FASTCALL call_6_0_12_6(SH2_struct* sh)
{
  SH2movlp(sh,0,12);
}
#define init_6_0_13_6
void FASTCALL call_6_0_13_6(SH2_struct* sh)
{
  SH2movlp(sh,0,13);
}
#define init_6_0_14_6
void FASTCALL call_6_0_14_6(SH2_struct* sh)
{
  SH2movlp(sh,0,14);
}
#define init_6_0_15_6
void FASTCALL call_6_0_15_6(SH2_struct* sh)
{
  SH2movlp(sh,0,15);
}
#define init_6_1_0_6
void FASTCALL call_6_1_0_6(SH2_struct* sh)
{
  SH2movlp(sh,1,0);
}
#define init_6_1_1_6
void FASTCALL call_6_1_1_6(SH2_struct* sh)
{
  SH2movlp(sh,1,1);
}
#define init_6_1_2_6
void FASTCALL call_6_1_2_6(SH2_struct* sh)
{
  SH2movlp(sh,1,2);
}
#define init_6_1_3_6
void FASTCALL call_6_1_3_6(SH2_struct* sh)
{
  SH2movlp(sh,1,3);
}
#define init_6_1_4_6
void FASTCALL call_6_1_4_6(SH2_struct* sh)
{
  SH2movlp(sh,1,4);
}
#define init_6_1_5_6
void FASTCALL call_6_1_5_6(SH2_struct* sh)
{
  SH2movlp(sh,1,5);
}
#define init_6_1_6_6
void FASTCALL call_6_1_6_6(SH2_struct* sh)
{
  SH2movlp(sh,1,6);
}
#define init_6_1_7_6
void FASTCALL call_6_1_7_6(SH2_struct* sh)
{
  SH2movlp(sh,1,7);
}
#define init_6_1_8_6
void FASTCALL call_6_1_8_6(SH2_struct* sh)
{
  SH2movlp(sh,1,8);
}
#define init_6_1_9_6
void FASTCALL call_6_1_9_6(SH2_struct* sh)
{
  SH2movlp(sh,1,9);
}
#define init_6_1_10_6
void FASTCALL call_6_1_10_6(SH2_struct* sh)
{
  SH2movlp(sh,1,10);
}
#define init_6_1_11_6
void FASTCALL call_6_1_11_6(SH2_struct* sh)
{
  SH2movlp(sh,1,11);
}
#define init_6_1_12_6
void FASTCALL call_6_1_12_6(SH2_struct* sh)
{
  SH2movlp(sh,1,12);
}
#define init_6_1_13_6
void FASTCALL call_6_1_13_6(SH2_struct* sh)
{
  SH2movlp(sh,1,13);
}
#define init_6_1_14_6
void FASTCALL call_6_1_14_6(SH2_struct* sh)
{
  SH2movlp(sh,1,14);
}
#define init_6_1_15_6
void FASTCALL call_6_1_15_6(SH2_struct* sh)
{
  SH2movlp(sh,1,15);
}
#define init_6_2_0_6
void FASTCALL call_6_2_0_6(SH2_struct* sh)
{
  SH2movlp(sh,2,0);
}
#define init_6_2_1_6
void FASTCALL call_6_2_1_6(SH2_struct* sh)
{
  SH2movlp(sh,2,1);
}
#define init_6_2_2_6
void FASTCALL call_6_2_2_6(SH2_struct* sh)
{
  SH2movlp(sh,2,2);
}
#define init_6_2_3_6
void FASTCALL call_6_2_3_6(SH2_struct* sh)
{
  SH2movlp(sh,2,3);
}
#define init_6_2_4_6
void FASTCALL call_6_2_4_6(SH2_struct* sh)
{
  SH2movlp(sh,2,4);
}
#define init_6_2_5_6
void FASTCALL call_6_2_5_6(SH2_struct* sh)
{
  SH2movlp(sh,2,5);
}
#define init_6_2_6_6
void FASTCALL call_6_2_6_6(SH2_struct* sh)
{
  SH2movlp(sh,2,6);
}
#define init_6_2_7_6
void FASTCALL call_6_2_7_6(SH2_struct* sh)
{
  SH2movlp(sh,2,7);
}
#define init_6_2_8_6
void FASTCALL call_6_2_8_6(SH2_struct* sh)
{
  SH2movlp(sh,2,8);
}
#define init_6_2_9_6
void FASTCALL call_6_2_9_6(SH2_struct* sh)
{
  SH2movlp(sh,2,9);
}
#define init_6_2_10_6
void FASTCALL call_6_2_10_6(SH2_struct* sh)
{
  SH2movlp(sh,2,10);
}
#define init_6_2_11_6
void FASTCALL call_6_2_11_6(SH2_struct* sh)
{
  SH2movlp(sh,2,11);
}
#define init_6_2_12_6
void FASTCALL call_6_2_12_6(SH2_struct* sh)
{
  SH2movlp(sh,2,12);
}
#define init_6_2_13_6
void FASTCALL call_6_2_13_6(SH2_struct* sh)
{
  SH2movlp(sh,2,13);
}
#define init_6_2_14_6
void FASTCALL call_6_2_14_6(SH2_struct* sh)
{
  SH2movlp(sh,2,14);
}
#define init_6_2_15_6
void FASTCALL call_6_2_15_6(SH2_struct* sh)
{
  SH2movlp(sh,2,15);
}
#define init_6_3_0_6
void FASTCALL call_6_3_0_6(SH2_struct* sh)
{
  SH2movlp(sh,3,0);
}
#define init_6_3_1_6
void FASTCALL call_6_3_1_6(SH2_struct* sh)
{
  SH2movlp(sh,3,1);
}
#define init_6_3_2_6
void FASTCALL call_6_3_2_6(SH2_struct* sh)
{
  SH2movlp(sh,3,2);
}
#define init_6_3_3_6
void FASTCALL call_6_3_3_6(SH2_struct* sh)
{
  SH2movlp(sh,3,3);
}
#define init_6_3_4_6
void FASTCALL call_6_3_4_6(SH2_struct* sh)
{
  SH2movlp(sh,3,4);
}
#define init_6_3_5_6
void FASTCALL call_6_3_5_6(SH2_struct* sh)
{
  SH2movlp(sh,3,5);
}
#define init_6_3_6_6
void FASTCALL call_6_3_6_6(SH2_struct* sh)
{
  SH2movlp(sh,3,6);
}
#define init_6_3_7_6
void FASTCALL call_6_3_7_6(SH2_struct* sh)
{
  SH2movlp(sh,3,7);
}
#define init_6_3_8_6
void FASTCALL call_6_3_8_6(SH2_struct* sh)
{
  SH2movlp(sh,3,8);
}
#define init_6_3_9_6
void FASTCALL call_6_3_9_6(SH2_struct* sh)
{
  SH2movlp(sh,3,9);
}
#define init_6_3_10_6
void FASTCALL call_6_3_10_6(SH2_struct* sh)
{
  SH2movlp(sh,3,10);
}
#define init_6_3_11_6
void FASTCALL call_6_3_11_6(SH2_struct* sh)
{
  SH2movlp(sh,3,11);
}
#define init_6_3_12_6
void FASTCALL call_6_3_12_6(SH2_struct* sh)
{
  SH2movlp(sh,3,12);
}
#define init_6_3_13_6
void FASTCALL call_6_3_13_6(SH2_struct* sh)
{
  SH2movlp(sh,3,13);
}
#define init_6_3_14_6
void FASTCALL call_6_3_14_6(SH2_struct* sh)
{
  SH2movlp(sh,3,14);
}
#define init_6_3_15_6
void FASTCALL call_6_3_15_6(SH2_struct* sh)
{
  SH2movlp(sh,3,15);
}
#define init_6_4_0_6
void FASTCALL call_6_4_0_6(SH2_struct* sh)
{
  SH2movlp(sh,4,0);
}
#define init_6_4_1_6
void FASTCALL call_6_4_1_6(SH2_struct* sh)
{
  SH2movlp(sh,4,1);
}
#define init_6_4_2_6
void FASTCALL call_6_4_2_6(SH2_struct* sh)
{
  SH2movlp(sh,4,2);
}
#define init_6_4_3_6
void FASTCALL call_6_4_3_6(SH2_struct* sh)
{
  SH2movlp(sh,4,3);
}
#define init_6_4_4_6
void FASTCALL call_6_4_4_6(SH2_struct* sh)
{
  SH2movlp(sh,4,4);
}
#define init_6_4_5_6
void FASTCALL call_6_4_5_6(SH2_struct* sh)
{
  SH2movlp(sh,4,5);
}
#define init_6_4_6_6
void FASTCALL call_6_4_6_6(SH2_struct* sh)
{
  SH2movlp(sh,4,6);
}
#define init_6_4_7_6
void FASTCALL call_6_4_7_6(SH2_struct* sh)
{
  SH2movlp(sh,4,7);
}
#define init_6_4_8_6
void FASTCALL call_6_4_8_6(SH2_struct* sh)
{
  SH2movlp(sh,4,8);
}
#define init_6_4_9_6
void FASTCALL call_6_4_9_6(SH2_struct* sh)
{
  SH2movlp(sh,4,9);
}
#define init_6_4_10_6
void FASTCALL call_6_4_10_6(SH2_struct* sh)
{
  SH2movlp(sh,4,10);
}
#define init_6_4_11_6
void FASTCALL call_6_4_11_6(SH2_struct* sh)
{
  SH2movlp(sh,4,11);
}
#define init_6_4_12_6
void FASTCALL call_6_4_12_6(SH2_struct* sh)
{
  SH2movlp(sh,4,12);
}
#define init_6_4_13_6
void FASTCALL call_6_4_13_6(SH2_struct* sh)
{
  SH2movlp(sh,4,13);
}
#define init_6_4_14_6
void FASTCALL call_6_4_14_6(SH2_struct* sh)
{
  SH2movlp(sh,4,14);
}
#define init_6_4_15_6
void FASTCALL call_6_4_15_6(SH2_struct* sh)
{
  SH2movlp(sh,4,15);
}
#define init_6_5_0_6
void FASTCALL call_6_5_0_6(SH2_struct* sh)
{
  SH2movlp(sh,5,0);
}
#define init_6_5_1_6
void FASTCALL call_6_5_1_6(SH2_struct* sh)
{
  SH2movlp(sh,5,1);
}
#define init_6_5_2_6
void FASTCALL call_6_5_2_6(SH2_struct* sh)
{
  SH2movlp(sh,5,2);
}
#define init_6_5_3_6
void FASTCALL call_6_5_3_6(SH2_struct* sh)
{
  SH2movlp(sh,5,3);
}
#define init_6_5_4_6
void FASTCALL call_6_5_4_6(SH2_struct* sh)
{
  SH2movlp(sh,5,4);
}
#define init_6_5_5_6
void FASTCALL call_6_5_5_6(SH2_struct* sh)
{
  SH2movlp(sh,5,5);
}
#define init_6_5_6_6
void FASTCALL call_6_5_6_6(SH2_struct* sh)
{
  SH2movlp(sh,5,6);
}
#define init_6_5_7_6
void FASTCALL call_6_5_7_6(SH2_struct* sh)
{
  SH2movlp(sh,5,7);
}
#define init_6_5_8_6
void FASTCALL call_6_5_8_6(SH2_struct* sh)
{
  SH2movlp(sh,5,8);
}
#define init_6_5_9_6
void FASTCALL call_6_5_9_6(SH2_struct* sh)
{
  SH2movlp(sh,5,9);
}
#define init_6_5_10_6
void FASTCALL call_6_5_10_6(SH2_struct* sh)
{
  SH2movlp(sh,5,10);
}
#define init_6_5_11_6
void FASTCALL call_6_5_11_6(SH2_struct* sh)
{
  SH2movlp(sh,5,11);
}
#define init_6_5_12_6
void FASTCALL call_6_5_12_6(SH2_struct* sh)
{
  SH2movlp(sh,5,12);
}
#define init_6_5_13_6
void FASTCALL call_6_5_13_6(SH2_struct* sh)
{
  SH2movlp(sh,5,13);
}
#define init_6_5_14_6
void FASTCALL call_6_5_14_6(SH2_struct* sh)
{
  SH2movlp(sh,5,14);
}
#define init_6_5_15_6
void FASTCALL call_6_5_15_6(SH2_struct* sh)
{
  SH2movlp(sh,5,15);
}
#define init_6_6_0_6
void FASTCALL call_6_6_0_6(SH2_struct* sh)
{
  SH2movlp(sh,6,0);
}
#define init_6_6_1_6
void FASTCALL call_6_6_1_6(SH2_struct* sh)
{
  SH2movlp(sh,6,1);
}
#define init_6_6_2_6
void FASTCALL call_6_6_2_6(SH2_struct* sh)
{
  SH2movlp(sh,6,2);
}
#define init_6_6_3_6
void FASTCALL call_6_6_3_6(SH2_struct* sh)
{
  SH2movlp(sh,6,3);
}
#define init_6_6_4_6
void FASTCALL call_6_6_4_6(SH2_struct* sh)
{
  SH2movlp(sh,6,4);
}
#define init_6_6_5_6
void FASTCALL call_6_6_5_6(SH2_struct* sh)
{
  SH2movlp(sh,6,5);
}
#define init_6_6_6_6
void FASTCALL call_6_6_6_6(SH2_struct* sh)
{
  SH2movlp(sh,6,6);
}
#define init_6_6_7_6
void FASTCALL call_6_6_7_6(SH2_struct* sh)
{
  SH2movlp(sh,6,7);
}
#define init_6_6_8_6
void FASTCALL call_6_6_8_6(SH2_struct* sh)
{
  SH2movlp(sh,6,8);
}
#define init_6_6_9_6
void FASTCALL call_6_6_9_6(SH2_struct* sh)
{
  SH2movlp(sh,6,9);
}
#define init_6_6_10_6
void FASTCALL call_6_6_10_6(SH2_struct* sh)
{
  SH2movlp(sh,6,10);
}
#define init_6_6_11_6
void FASTCALL call_6_6_11_6(SH2_struct* sh)
{
  SH2movlp(sh,6,11);
}
#define init_6_6_12_6
void FASTCALL call_6_6_12_6(SH2_struct* sh)
{
  SH2movlp(sh,6,12);
}
#define init_6_6_13_6
void FASTCALL call_6_6_13_6(SH2_struct* sh)
{
  SH2movlp(sh,6,13);
}
#define init_6_6_14_6
void FASTCALL call_6_6_14_6(SH2_struct* sh)
{
  SH2movlp(sh,6,14);
}
#define init_6_6_15_6
void FASTCALL call_6_6_15_6(SH2_struct* sh)
{
  SH2movlp(sh,6,15);
}
#define init_6_7_0_6
void FASTCALL call_6_7_0_6(SH2_struct* sh)
{
  SH2movlp(sh,7,0);
}
#define init_6_7_1_6
void FASTCALL call_6_7_1_6(SH2_struct* sh)
{
  SH2movlp(sh,7,1);
}
#define init_6_7_2_6
void FASTCALL call_6_7_2_6(SH2_struct* sh)
{
  SH2movlp(sh,7,2);
}
#define init_6_7_3_6
void FASTCALL call_6_7_3_6(SH2_struct* sh)
{
  SH2movlp(sh,7,3);
}
#define init_6_7_4_6
void FASTCALL call_6_7_4_6(SH2_struct* sh)
{
  SH2movlp(sh,7,4);
}
#define init_6_7_5_6
void FASTCALL call_6_7_5_6(SH2_struct* sh)
{
  SH2movlp(sh,7,5);
}
#define init_6_7_6_6
void FASTCALL call_6_7_6_6(SH2_struct* sh)
{
  SH2movlp(sh,7,6);
}
#define init_6_7_7_6
void FASTCALL call_6_7_7_6(SH2_struct* sh)
{
  SH2movlp(sh,7,7);
}
#define init_6_7_8_6
void FASTCALL call_6_7_8_6(SH2_struct* sh)
{
  SH2movlp(sh,7,8);
}
#define init_6_7_9_6
void FASTCALL call_6_7_9_6(SH2_struct* sh)
{
  SH2movlp(sh,7,9);
}
#define init_6_7_10_6
void FASTCALL call_6_7_10_6(SH2_struct* sh)
{
  SH2movlp(sh,7,10);
}
#define init_6_7_11_6
void FASTCALL call_6_7_11_6(SH2_struct* sh)
{
  SH2movlp(sh,7,11);
}
#define init_6_7_12_6
void FASTCALL call_6_7_12_6(SH2_struct* sh)
{
  SH2movlp(sh,7,12);
}
#define init_6_7_13_6
void FASTCALL call_6_7_13_6(SH2_struct* sh)
{
  SH2movlp(sh,7,13);
}
#define init_6_7_14_6
void FASTCALL call_6_7_14_6(SH2_struct* sh)
{
  SH2movlp(sh,7,14);
}
#define init_6_7_15_6
void FASTCALL call_6_7_15_6(SH2_struct* sh)
{
  SH2movlp(sh,7,15);
}
#define init_6_8_0_6
void FASTCALL call_6_8_0_6(SH2_struct* sh)
{
  SH2movlp(sh,8,0);
}
#define init_6_8_1_6
void FASTCALL call_6_8_1_6(SH2_struct* sh)
{
  SH2movlp(sh,8,1);
}
#define init_6_8_2_6
void FASTCALL call_6_8_2_6(SH2_struct* sh)
{
  SH2movlp(sh,8,2);
}
#define init_6_8_3_6
void FASTCALL call_6_8_3_6(SH2_struct* sh)
{
  SH2movlp(sh,8,3);
}
#define init_6_8_4_6
void FASTCALL call_6_8_4_6(SH2_struct* sh)
{
  SH2movlp(sh,8,4);
}
#define init_6_8_5_6
void FASTCALL call_6_8_5_6(SH2_struct* sh)
{
  SH2movlp(sh,8,5);
}
#define init_6_8_6_6
void FASTCALL call_6_8_6_6(SH2_struct* sh)
{
  SH2movlp(sh,8,6);
}
#define init_6_8_7_6
void FASTCALL call_6_8_7_6(SH2_struct* sh)
{
  SH2movlp(sh,8,7);
}
#define init_6_8_8_6
void FASTCALL call_6_8_8_6(SH2_struct* sh)
{
  SH2movlp(sh,8,8);
}
#define init_6_8_9_6
void FASTCALL call_6_8_9_6(SH2_struct* sh)
{
  SH2movlp(sh,8,9);
}
#define init_6_8_10_6
void FASTCALL call_6_8_10_6(SH2_struct* sh)
{
  SH2movlp(sh,8,10);
}
#define init_6_8_11_6
void FASTCALL call_6_8_11_6(SH2_struct* sh)
{
  SH2movlp(sh,8,11);
}
#define init_6_8_12_6
void FASTCALL call_6_8_12_6(SH2_struct* sh)
{
  SH2movlp(sh,8,12);
}
#define init_6_8_13_6
void FASTCALL call_6_8_13_6(SH2_struct* sh)
{
  SH2movlp(sh,8,13);
}
#define init_6_8_14_6
void FASTCALL call_6_8_14_6(SH2_struct* sh)
{
  SH2movlp(sh,8,14);
}
#define init_6_8_15_6
void FASTCALL call_6_8_15_6(SH2_struct* sh)
{
  SH2movlp(sh,8,15);
}
#define init_6_9_0_6
void FASTCALL call_6_9_0_6(SH2_struct* sh)
{
  SH2movlp(sh,9,0);
}
#define init_6_9_1_6
void FASTCALL call_6_9_1_6(SH2_struct* sh)
{
  SH2movlp(sh,9,1);
}
#define init_6_9_2_6
void FASTCALL call_6_9_2_6(SH2_struct* sh)
{
  SH2movlp(sh,9,2);
}
#define init_6_9_3_6
void FASTCALL call_6_9_3_6(SH2_struct* sh)
{
  SH2movlp(sh,9,3);
}
#define init_6_9_4_6
void FASTCALL call_6_9_4_6(SH2_struct* sh)
{
  SH2movlp(sh,9,4);
}
#define init_6_9_5_6
void FASTCALL call_6_9_5_6(SH2_struct* sh)
{
  SH2movlp(sh,9,5);
}
#define init_6_9_6_6
void FASTCALL call_6_9_6_6(SH2_struct* sh)
{
  SH2movlp(sh,9,6);
}
#define init_6_9_7_6
void FASTCALL call_6_9_7_6(SH2_struct* sh)
{
  SH2movlp(sh,9,7);
}
#define init_6_9_8_6
void FASTCALL call_6_9_8_6(SH2_struct* sh)
{
  SH2movlp(sh,9,8);
}
#define init_6_9_9_6
void FASTCALL call_6_9_9_6(SH2_struct* sh)
{
  SH2movlp(sh,9,9);
}
#define init_6_9_10_6
void FASTCALL call_6_9_10_6(SH2_struct* sh)
{
  SH2movlp(sh,9,10);
}
#define init_6_9_11_6
void FASTCALL call_6_9_11_6(SH2_struct* sh)
{
  SH2movlp(sh,9,11);
}
#define init_6_9_12_6
void FASTCALL call_6_9_12_6(SH2_struct* sh)
{
  SH2movlp(sh,9,12);
}
#define init_6_9_13_6
void FASTCALL call_6_9_13_6(SH2_struct* sh)
{
  SH2movlp(sh,9,13);
}
#define init_6_9_14_6
void FASTCALL call_6_9_14_6(SH2_struct* sh)
{
  SH2movlp(sh,9,14);
}
#define init_6_9_15_6
void FASTCALL call_6_9_15_6(SH2_struct* sh)
{
  SH2movlp(sh,9,15);
}
#define init_6_10_0_6
void FASTCALL call_6_10_0_6(SH2_struct* sh)
{
  SH2movlp(sh,10,0);
}
#define init_6_10_1_6
void FASTCALL call_6_10_1_6(SH2_struct* sh)
{
  SH2movlp(sh,10,1);
}
#define init_6_10_2_6
void FASTCALL call_6_10_2_6(SH2_struct* sh)
{
  SH2movlp(sh,10,2);
}
#define init_6_10_3_6
void FASTCALL call_6_10_3_6(SH2_struct* sh)
{
  SH2movlp(sh,10,3);
}
#define init_6_10_4_6
void FASTCALL call_6_10_4_6(SH2_struct* sh)
{
  SH2movlp(sh,10,4);
}
#define init_6_10_5_6
void FASTCALL call_6_10_5_6(SH2_struct* sh)
{
  SH2movlp(sh,10,5);
}
#define init_6_10_6_6
void FASTCALL call_6_10_6_6(SH2_struct* sh)
{
  SH2movlp(sh,10,6);
}
#define init_6_10_7_6
void FASTCALL call_6_10_7_6(SH2_struct* sh)
{
  SH2movlp(sh,10,7);
}
#define init_6_10_8_6
void FASTCALL call_6_10_8_6(SH2_struct* sh)
{
  SH2movlp(sh,10,8);
}
#define init_6_10_9_6
void FASTCALL call_6_10_9_6(SH2_struct* sh)
{
  SH2movlp(sh,10,9);
}
#define init_6_10_10_6
void FASTCALL call_6_10_10_6(SH2_struct* sh)
{
  SH2movlp(sh,10,10);
}
#define init_6_10_11_6
void FASTCALL call_6_10_11_6(SH2_struct* sh)
{
  SH2movlp(sh,10,11);
}
#define init_6_10_12_6
void FASTCALL call_6_10_12_6(SH2_struct* sh)
{
  SH2movlp(sh,10,12);
}
#define init_6_10_13_6
void FASTCALL call_6_10_13_6(SH2_struct* sh)
{
  SH2movlp(sh,10,13);
}
#define init_6_10_14_6
void FASTCALL call_6_10_14_6(SH2_struct* sh)
{
  SH2movlp(sh,10,14);
}
#define init_6_10_15_6
void FASTCALL call_6_10_15_6(SH2_struct* sh)
{
  SH2movlp(sh,10,15);
}
#define init_6_11_0_6
void FASTCALL call_6_11_0_6(SH2_struct* sh)
{
  SH2movlp(sh,11,0);
}
#define init_6_11_1_6
void FASTCALL call_6_11_1_6(SH2_struct* sh)
{
  SH2movlp(sh,11,1);
}
#define init_6_11_2_6
void FASTCALL call_6_11_2_6(SH2_struct* sh)
{
  SH2movlp(sh,11,2);
}
#define init_6_11_3_6
void FASTCALL call_6_11_3_6(SH2_struct* sh)
{
  SH2movlp(sh,11,3);
}
#define init_6_11_4_6
void FASTCALL call_6_11_4_6(SH2_struct* sh)
{
  SH2movlp(sh,11,4);
}
#define init_6_11_5_6
void FASTCALL call_6_11_5_6(SH2_struct* sh)
{
  SH2movlp(sh,11,5);
}
#define init_6_11_6_6
void FASTCALL call_6_11_6_6(SH2_struct* sh)
{
  SH2movlp(sh,11,6);
}
#define init_6_11_7_6
void FASTCALL call_6_11_7_6(SH2_struct* sh)
{
  SH2movlp(sh,11,7);
}
#define init_6_11_8_6
void FASTCALL call_6_11_8_6(SH2_struct* sh)
{
  SH2movlp(sh,11,8);
}
#define init_6_11_9_6
void FASTCALL call_6_11_9_6(SH2_struct* sh)
{
  SH2movlp(sh,11,9);
}
#define init_6_11_10_6
void FASTCALL call_6_11_10_6(SH2_struct* sh)
{
  SH2movlp(sh,11,10);
}
#define init_6_11_11_6
void FASTCALL call_6_11_11_6(SH2_struct* sh)
{
  SH2movlp(sh,11,11);
}
#define init_6_11_12_6
void FASTCALL call_6_11_12_6(SH2_struct* sh)
{
  SH2movlp(sh,11,12);
}
#define init_6_11_13_6
void FASTCALL call_6_11_13_6(SH2_struct* sh)
{
  SH2movlp(sh,11,13);
}
#define init_6_11_14_6
void FASTCALL call_6_11_14_6(SH2_struct* sh)
{
  SH2movlp(sh,11,14);
}
#define init_6_11_15_6
void FASTCALL call_6_11_15_6(SH2_struct* sh)
{
  SH2movlp(sh,11,15);
}
#define init_6_12_0_6
void FASTCALL call_6_12_0_6(SH2_struct* sh)
{
  SH2movlp(sh,12,0);
}
#define init_6_12_1_6
void FASTCALL call_6_12_1_6(SH2_struct* sh)
{
  SH2movlp(sh,12,1);
}
#define init_6_12_2_6
void FASTCALL call_6_12_2_6(SH2_struct* sh)
{
  SH2movlp(sh,12,2);
}
#define init_6_12_3_6
void FASTCALL call_6_12_3_6(SH2_struct* sh)
{
  SH2movlp(sh,12,3);
}
#define init_6_12_4_6
void FASTCALL call_6_12_4_6(SH2_struct* sh)
{
  SH2movlp(sh,12,4);
}
#define init_6_12_5_6
void FASTCALL call_6_12_5_6(SH2_struct* sh)
{
  SH2movlp(sh,12,5);
}
#define init_6_12_6_6
void FASTCALL call_6_12_6_6(SH2_struct* sh)
{
  SH2movlp(sh,12,6);
}
#define init_6_12_7_6
void FASTCALL call_6_12_7_6(SH2_struct* sh)
{
  SH2movlp(sh,12,7);
}
#define init_6_12_8_6
void FASTCALL call_6_12_8_6(SH2_struct* sh)
{
  SH2movlp(sh,12,8);
}
#define init_6_12_9_6
void FASTCALL call_6_12_9_6(SH2_struct* sh)
{
  SH2movlp(sh,12,9);
}
#define init_6_12_10_6
void FASTCALL call_6_12_10_6(SH2_struct* sh)
{
  SH2movlp(sh,12,10);
}
#define init_6_12_11_6
void FASTCALL call_6_12_11_6(SH2_struct* sh)
{
  SH2movlp(sh,12,11);
}
#define init_6_12_12_6
void FASTCALL call_6_12_12_6(SH2_struct* sh)
{
  SH2movlp(sh,12,12);
}
#define init_6_12_13_6
void FASTCALL call_6_12_13_6(SH2_struct* sh)
{
  SH2movlp(sh,12,13);
}
#define init_6_12_14_6
void FASTCALL call_6_12_14_6(SH2_struct* sh)
{
  SH2movlp(sh,12,14);
}
#define init_6_12_15_6
void FASTCALL call_6_12_15_6(SH2_struct* sh)
{
  SH2movlp(sh,12,15);
}
#define init_6_13_0_6
void FASTCALL call_6_13_0_6(SH2_struct* sh)
{
  SH2movlp(sh,13,0);
}
#define init_6_13_1_6
void FASTCALL call_6_13_1_6(SH2_struct* sh)
{
  SH2movlp(sh,13,1);
}
#define init_6_13_2_6
void FASTCALL call_6_13_2_6(SH2_struct* sh)
{
  SH2movlp(sh,13,2);
}
#define init_6_13_3_6
void FASTCALL call_6_13_3_6(SH2_struct* sh)
{
  SH2movlp(sh,13,3);
}
#define init_6_13_4_6
void FASTCALL call_6_13_4_6(SH2_struct* sh)
{
  SH2movlp(sh,13,4);
}
#define init_6_13_5_6
void FASTCALL call_6_13_5_6(SH2_struct* sh)
{
  SH2movlp(sh,13,5);
}
#define init_6_13_6_6
void FASTCALL call_6_13_6_6(SH2_struct* sh)
{
  SH2movlp(sh,13,6);
}
#define init_6_13_7_6
void FASTCALL call_6_13_7_6(SH2_struct* sh)
{
  SH2movlp(sh,13,7);
}
#define init_6_13_8_6
void FASTCALL call_6_13_8_6(SH2_struct* sh)
{
  SH2movlp(sh,13,8);
}
#define init_6_13_9_6
void FASTCALL call_6_13_9_6(SH2_struct* sh)
{
  SH2movlp(sh,13,9);
}
#define init_6_13_10_6
void FASTCALL call_6_13_10_6(SH2_struct* sh)
{
  SH2movlp(sh,13,10);
}
#define init_6_13_11_6
void FASTCALL call_6_13_11_6(SH2_struct* sh)
{
  SH2movlp(sh,13,11);
}
#define init_6_13_12_6
void FASTCALL call_6_13_12_6(SH2_struct* sh)
{
  SH2movlp(sh,13,12);
}
#define init_6_13_13_6
void FASTCALL call_6_13_13_6(SH2_struct* sh)
{
  SH2movlp(sh,13,13);
}
#define init_6_13_14_6
void FASTCALL call_6_13_14_6(SH2_struct* sh)
{
  SH2movlp(sh,13,14);
}
#define init_6_13_15_6
void FASTCALL call_6_13_15_6(SH2_struct* sh)
{
  SH2movlp(sh,13,15);
}
#define init_6_14_0_6
void FASTCALL call_6_14_0_6(SH2_struct* sh)
{
  SH2movlp(sh,14,0);
}
#define init_6_14_1_6
void FASTCALL call_6_14_1_6(SH2_struct* sh)
{
  SH2movlp(sh,14,1);
}
#define init_6_14_2_6
void FASTCALL call_6_14_2_6(SH2_struct* sh)
{
  SH2movlp(sh,14,2);
}
#define init_6_14_3_6
void FASTCALL call_6_14_3_6(SH2_struct* sh)
{
  SH2movlp(sh,14,3);
}
#define init_6_14_4_6
void FASTCALL call_6_14_4_6(SH2_struct* sh)
{
  SH2movlp(sh,14,4);
}
#define init_6_14_5_6
void FASTCALL call_6_14_5_6(SH2_struct* sh)
{
  SH2movlp(sh,14,5);
}
#define init_6_14_6_6
void FASTCALL call_6_14_6_6(SH2_struct* sh)
{
  SH2movlp(sh,14,6);
}
#define init_6_14_7_6
void FASTCALL call_6_14_7_6(SH2_struct* sh)
{
  SH2movlp(sh,14,7);
}
#define init_6_14_8_6
void FASTCALL call_6_14_8_6(SH2_struct* sh)
{
  SH2movlp(sh,14,8);
}
#define init_6_14_9_6
void FASTCALL call_6_14_9_6(SH2_struct* sh)
{
  SH2movlp(sh,14,9);
}
#define init_6_14_10_6
void FASTCALL call_6_14_10_6(SH2_struct* sh)
{
  SH2movlp(sh,14,10);
}
#define init_6_14_11_6
void FASTCALL call_6_14_11_6(SH2_struct* sh)
{
  SH2movlp(sh,14,11);
}
#define init_6_14_12_6
void FASTCALL call_6_14_12_6(SH2_struct* sh)
{
  SH2movlp(sh,14,12);
}
#define init_6_14_13_6
void FASTCALL call_6_14_13_6(SH2_struct* sh)
{
  SH2movlp(sh,14,13);
}
#define init_6_14_14_6
void FASTCALL call_6_14_14_6(SH2_struct* sh)
{
  SH2movlp(sh,14,14);
}
#define init_6_14_15_6
void FASTCALL call_6_14_15_6(SH2_struct* sh)
{
  SH2movlp(sh,14,15);
}
#define init_6_15_0_6
void FASTCALL call_6_15_0_6(SH2_struct* sh)
{
  SH2movlp(sh,15,0);
}
#define init_6_15_1_6
void FASTCALL call_6_15_1_6(SH2_struct* sh)
{
  SH2movlp(sh,15,1);
}
#define init_6_15_2_6
void FASTCALL call_6_15_2_6(SH2_struct* sh)
{
  SH2movlp(sh,15,2);
}
#define init_6_15_3_6
void FASTCALL call_6_15_3_6(SH2_struct* sh)
{
  SH2movlp(sh,15,3);
}
#define init_6_15_4_6
void FASTCALL call_6_15_4_6(SH2_struct* sh)
{
  SH2movlp(sh,15,4);
}
#define init_6_15_5_6
void FASTCALL call_6_15_5_6(SH2_struct* sh)
{
  SH2movlp(sh,15,5);
}
#define init_6_15_6_6
void FASTCALL call_6_15_6_6(SH2_struct* sh)
{
  SH2movlp(sh,15,6);
}
#define init_6_15_7_6
void FASTCALL call_6_15_7_6(SH2_struct* sh)
{
  SH2movlp(sh,15,7);
}
#define init_6_15_8_6
void FASTCALL call_6_15_8_6(SH2_struct* sh)
{
  SH2movlp(sh,15,8);
}
#define init_6_15_9_6
void FASTCALL call_6_15_9_6(SH2_struct* sh)
{
  SH2movlp(sh,15,9);
}
#define init_6_15_10_6
void FASTCALL call_6_15_10_6(SH2_struct* sh)
{
  SH2movlp(sh,15,10);
}
#define init_6_15_11_6
void FASTCALL call_6_15_11_6(SH2_struct* sh)
{
  SH2movlp(sh,15,11);
}
#define init_6_15_12_6
void FASTCALL call_6_15_12_6(SH2_struct* sh)
{
  SH2movlp(sh,15,12);
}
#define init_6_15_13_6
void FASTCALL call_6_15_13_6(SH2_struct* sh)
{
  SH2movlp(sh,15,13);
}
#define init_6_15_14_6
void FASTCALL call_6_15_14_6(SH2_struct* sh)
{
  SH2movlp(sh,15,14);
}
#define init_6_15_15_6
void FASTCALL call_6_15_15_6(SH2_struct* sh)
{
  SH2movlp(sh,15,15);
}
#define init_2_0_0_6
void FASTCALL call_2_0_0_6(SH2_struct* sh)
{
  SH2movlm(sh,0,0);
}
#define init_2_0_1_6
void FASTCALL call_2_0_1_6(SH2_struct* sh)
{
  SH2movlm(sh,0,1);
}
#define init_2_0_2_6
void FASTCALL call_2_0_2_6(SH2_struct* sh)
{
  SH2movlm(sh,0,2);
}
#define init_2_0_3_6
void FASTCALL call_2_0_3_6(SH2_struct* sh)
{
  SH2movlm(sh,0,3);
}
#define init_2_0_4_6
void FASTCALL call_2_0_4_6(SH2_struct* sh)
{
  SH2movlm(sh,0,4);
}
#define init_2_0_5_6
void FASTCALL call_2_0_5_6(SH2_struct* sh)
{
  SH2movlm(sh,0,5);
}
#define init_2_0_6_6
void FASTCALL call_2_0_6_6(SH2_struct* sh)
{
  SH2movlm(sh,0,6);
}
#define init_2_0_7_6
void FASTCALL call_2_0_7_6(SH2_struct* sh)
{
  SH2movlm(sh,0,7);
}
#define init_2_0_8_6
void FASTCALL call_2_0_8_6(SH2_struct* sh)
{
  SH2movlm(sh,0,8);
}
#define init_2_0_9_6
void FASTCALL call_2_0_9_6(SH2_struct* sh)
{
  SH2movlm(sh,0,9);
}
#define init_2_0_10_6
void FASTCALL call_2_0_10_6(SH2_struct* sh)
{
  SH2movlm(sh,0,10);
}
#define init_2_0_11_6
void FASTCALL call_2_0_11_6(SH2_struct* sh)
{
  SH2movlm(sh,0,11);
}
#define init_2_0_12_6
void FASTCALL call_2_0_12_6(SH2_struct* sh)
{
  SH2movlm(sh,0,12);
}
#define init_2_0_13_6
void FASTCALL call_2_0_13_6(SH2_struct* sh)
{
  SH2movlm(sh,0,13);
}
#define init_2_0_14_6
void FASTCALL call_2_0_14_6(SH2_struct* sh)
{
  SH2movlm(sh,0,14);
}
#define init_2_0_15_6
void FASTCALL call_2_0_15_6(SH2_struct* sh)
{
  SH2movlm(sh,0,15);
}
#define init_2_1_0_6
void FASTCALL call_2_1_0_6(SH2_struct* sh)
{
  SH2movlm(sh,1,0);
}
#define init_2_1_1_6
void FASTCALL call_2_1_1_6(SH2_struct* sh)
{
  SH2movlm(sh,1,1);
}
#define init_2_1_2_6
void FASTCALL call_2_1_2_6(SH2_struct* sh)
{
  SH2movlm(sh,1,2);
}
#define init_2_1_3_6
void FASTCALL call_2_1_3_6(SH2_struct* sh)
{
  SH2movlm(sh,1,3);
}
#define init_2_1_4_6
void FASTCALL call_2_1_4_6(SH2_struct* sh)
{
  SH2movlm(sh,1,4);
}
#define init_2_1_5_6
void FASTCALL call_2_1_5_6(SH2_struct* sh)
{
  SH2movlm(sh,1,5);
}
#define init_2_1_6_6
void FASTCALL call_2_1_6_6(SH2_struct* sh)
{
  SH2movlm(sh,1,6);
}
#define init_2_1_7_6
void FASTCALL call_2_1_7_6(SH2_struct* sh)
{
  SH2movlm(sh,1,7);
}
#define init_2_1_8_6
void FASTCALL call_2_1_8_6(SH2_struct* sh)
{
  SH2movlm(sh,1,8);
}
#define init_2_1_9_6
void FASTCALL call_2_1_9_6(SH2_struct* sh)
{
  SH2movlm(sh,1,9);
}
#define init_2_1_10_6
void FASTCALL call_2_1_10_6(SH2_struct* sh)
{
  SH2movlm(sh,1,10);
}
#define init_2_1_11_6
void FASTCALL call_2_1_11_6(SH2_struct* sh)
{
  SH2movlm(sh,1,11);
}
#define init_2_1_12_6
void FASTCALL call_2_1_12_6(SH2_struct* sh)
{
  SH2movlm(sh,1,12);
}
#define init_2_1_13_6
void FASTCALL call_2_1_13_6(SH2_struct* sh)
{
  SH2movlm(sh,1,13);
}
#define init_2_1_14_6
void FASTCALL call_2_1_14_6(SH2_struct* sh)
{
  SH2movlm(sh,1,14);
}
#define init_2_1_15_6
void FASTCALL call_2_1_15_6(SH2_struct* sh)
{
  SH2movlm(sh,1,15);
}
#define init_2_2_0_6
void FASTCALL call_2_2_0_6(SH2_struct* sh)
{
  SH2movlm(sh,2,0);
}
#define init_2_2_1_6
void FASTCALL call_2_2_1_6(SH2_struct* sh)
{
  SH2movlm(sh,2,1);
}
#define init_2_2_2_6
void FASTCALL call_2_2_2_6(SH2_struct* sh)
{
  SH2movlm(sh,2,2);
}
#define init_2_2_3_6
void FASTCALL call_2_2_3_6(SH2_struct* sh)
{
  SH2movlm(sh,2,3);
}
#define init_2_2_4_6
void FASTCALL call_2_2_4_6(SH2_struct* sh)
{
  SH2movlm(sh,2,4);
}
#define init_2_2_5_6
void FASTCALL call_2_2_5_6(SH2_struct* sh)
{
  SH2movlm(sh,2,5);
}
#define init_2_2_6_6
void FASTCALL call_2_2_6_6(SH2_struct* sh)
{
  SH2movlm(sh,2,6);
}
#define init_2_2_7_6
void FASTCALL call_2_2_7_6(SH2_struct* sh)
{
  SH2movlm(sh,2,7);
}
#define init_2_2_8_6
void FASTCALL call_2_2_8_6(SH2_struct* sh)
{
  SH2movlm(sh,2,8);
}
#define init_2_2_9_6
void FASTCALL call_2_2_9_6(SH2_struct* sh)
{
  SH2movlm(sh,2,9);
}
#define init_2_2_10_6
void FASTCALL call_2_2_10_6(SH2_struct* sh)
{
  SH2movlm(sh,2,10);
}
#define init_2_2_11_6
void FASTCALL call_2_2_11_6(SH2_struct* sh)
{
  SH2movlm(sh,2,11);
}
#define init_2_2_12_6
void FASTCALL call_2_2_12_6(SH2_struct* sh)
{
  SH2movlm(sh,2,12);
}
#define init_2_2_13_6
void FASTCALL call_2_2_13_6(SH2_struct* sh)
{
  SH2movlm(sh,2,13);
}
#define init_2_2_14_6
void FASTCALL call_2_2_14_6(SH2_struct* sh)
{
  SH2movlm(sh,2,14);
}
#define init_2_2_15_6
void FASTCALL call_2_2_15_6(SH2_struct* sh)
{
  SH2movlm(sh,2,15);
}
#define init_2_3_0_6
void FASTCALL call_2_3_0_6(SH2_struct* sh)
{
  SH2movlm(sh,3,0);
}
#define init_2_3_1_6
void FASTCALL call_2_3_1_6(SH2_struct* sh)
{
  SH2movlm(sh,3,1);
}
#define init_2_3_2_6
void FASTCALL call_2_3_2_6(SH2_struct* sh)
{
  SH2movlm(sh,3,2);
}
#define init_2_3_3_6
void FASTCALL call_2_3_3_6(SH2_struct* sh)
{
  SH2movlm(sh,3,3);
}
#define init_2_3_4_6
void FASTCALL call_2_3_4_6(SH2_struct* sh)
{
  SH2movlm(sh,3,4);
}
#define init_2_3_5_6
void FASTCALL call_2_3_5_6(SH2_struct* sh)
{
  SH2movlm(sh,3,5);
}
#define init_2_3_6_6
void FASTCALL call_2_3_6_6(SH2_struct* sh)
{
  SH2movlm(sh,3,6);
}
#define init_2_3_7_6
void FASTCALL call_2_3_7_6(SH2_struct* sh)
{
  SH2movlm(sh,3,7);
}
#define init_2_3_8_6
void FASTCALL call_2_3_8_6(SH2_struct* sh)
{
  SH2movlm(sh,3,8);
}
#define init_2_3_9_6
void FASTCALL call_2_3_9_6(SH2_struct* sh)
{
  SH2movlm(sh,3,9);
}
#define init_2_3_10_6
void FASTCALL call_2_3_10_6(SH2_struct* sh)
{
  SH2movlm(sh,3,10);
}
#define init_2_3_11_6
void FASTCALL call_2_3_11_6(SH2_struct* sh)
{
  SH2movlm(sh,3,11);
}
#define init_2_3_12_6
void FASTCALL call_2_3_12_6(SH2_struct* sh)
{
  SH2movlm(sh,3,12);
}
#define init_2_3_13_6
void FASTCALL call_2_3_13_6(SH2_struct* sh)
{
  SH2movlm(sh,3,13);
}
#define init_2_3_14_6
void FASTCALL call_2_3_14_6(SH2_struct* sh)
{
  SH2movlm(sh,3,14);
}
#define init_2_3_15_6
void FASTCALL call_2_3_15_6(SH2_struct* sh)
{
  SH2movlm(sh,3,15);
}
#define init_2_4_0_6
void FASTCALL call_2_4_0_6(SH2_struct* sh)
{
  SH2movlm(sh,4,0);
}
#define init_2_4_1_6
void FASTCALL call_2_4_1_6(SH2_struct* sh)
{
  SH2movlm(sh,4,1);
}
#define init_2_4_2_6
void FASTCALL call_2_4_2_6(SH2_struct* sh)
{
  SH2movlm(sh,4,2);
}
#define init_2_4_3_6
void FASTCALL call_2_4_3_6(SH2_struct* sh)
{
  SH2movlm(sh,4,3);
}
#define init_2_4_4_6
void FASTCALL call_2_4_4_6(SH2_struct* sh)
{
  SH2movlm(sh,4,4);
}
#define init_2_4_5_6
void FASTCALL call_2_4_5_6(SH2_struct* sh)
{
  SH2movlm(sh,4,5);
}
#define init_2_4_6_6
void FASTCALL call_2_4_6_6(SH2_struct* sh)
{
  SH2movlm(sh,4,6);
}
#define init_2_4_7_6
void FASTCALL call_2_4_7_6(SH2_struct* sh)
{
  SH2movlm(sh,4,7);
}
#define init_2_4_8_6
void FASTCALL call_2_4_8_6(SH2_struct* sh)
{
  SH2movlm(sh,4,8);
}
#define init_2_4_9_6
void FASTCALL call_2_4_9_6(SH2_struct* sh)
{
  SH2movlm(sh,4,9);
}
#define init_2_4_10_6
void FASTCALL call_2_4_10_6(SH2_struct* sh)
{
  SH2movlm(sh,4,10);
}
#define init_2_4_11_6
void FASTCALL call_2_4_11_6(SH2_struct* sh)
{
  SH2movlm(sh,4,11);
}
#define init_2_4_12_6
void FASTCALL call_2_4_12_6(SH2_struct* sh)
{
  SH2movlm(sh,4,12);
}
#define init_2_4_13_6
void FASTCALL call_2_4_13_6(SH2_struct* sh)
{
  SH2movlm(sh,4,13);
}
#define init_2_4_14_6
void FASTCALL call_2_4_14_6(SH2_struct* sh)
{
  SH2movlm(sh,4,14);
}
#define init_2_4_15_6
void FASTCALL call_2_4_15_6(SH2_struct* sh)
{
  SH2movlm(sh,4,15);
}
#define init_2_5_0_6
void FASTCALL call_2_5_0_6(SH2_struct* sh)
{
  SH2movlm(sh,5,0);
}
#define init_2_5_1_6
void FASTCALL call_2_5_1_6(SH2_struct* sh)
{
  SH2movlm(sh,5,1);
}
#define init_2_5_2_6
void FASTCALL call_2_5_2_6(SH2_struct* sh)
{
  SH2movlm(sh,5,2);
}
#define init_2_5_3_6
void FASTCALL call_2_5_3_6(SH2_struct* sh)
{
  SH2movlm(sh,5,3);
}
#define init_2_5_4_6
void FASTCALL call_2_5_4_6(SH2_struct* sh)
{
  SH2movlm(sh,5,4);
}
#define init_2_5_5_6
void FASTCALL call_2_5_5_6(SH2_struct* sh)
{
  SH2movlm(sh,5,5);
}
#define init_2_5_6_6
void FASTCALL call_2_5_6_6(SH2_struct* sh)
{
  SH2movlm(sh,5,6);
}
#define init_2_5_7_6
void FASTCALL call_2_5_7_6(SH2_struct* sh)
{
  SH2movlm(sh,5,7);
}
#define init_2_5_8_6
void FASTCALL call_2_5_8_6(SH2_struct* sh)
{
  SH2movlm(sh,5,8);
}
#define init_2_5_9_6
void FASTCALL call_2_5_9_6(SH2_struct* sh)
{
  SH2movlm(sh,5,9);
}
#define init_2_5_10_6
void FASTCALL call_2_5_10_6(SH2_struct* sh)
{
  SH2movlm(sh,5,10);
}
#define init_2_5_11_6
void FASTCALL call_2_5_11_6(SH2_struct* sh)
{
  SH2movlm(sh,5,11);
}
#define init_2_5_12_6
void FASTCALL call_2_5_12_6(SH2_struct* sh)
{
  SH2movlm(sh,5,12);
}
#define init_2_5_13_6
void FASTCALL call_2_5_13_6(SH2_struct* sh)
{
  SH2movlm(sh,5,13);
}
#define init_2_5_14_6
void FASTCALL call_2_5_14_6(SH2_struct* sh)
{
  SH2movlm(sh,5,14);
}
#define init_2_5_15_6
void FASTCALL call_2_5_15_6(SH2_struct* sh)
{
  SH2movlm(sh,5,15);
}
#define init_2_6_0_6
void FASTCALL call_2_6_0_6(SH2_struct* sh)
{
  SH2movlm(sh,6,0);
}
#define init_2_6_1_6
void FASTCALL call_2_6_1_6(SH2_struct* sh)
{
  SH2movlm(sh,6,1);
}
#define init_2_6_2_6
void FASTCALL call_2_6_2_6(SH2_struct* sh)
{
  SH2movlm(sh,6,2);
}
#define init_2_6_3_6
void FASTCALL call_2_6_3_6(SH2_struct* sh)
{
  SH2movlm(sh,6,3);
}
#define init_2_6_4_6
void FASTCALL call_2_6_4_6(SH2_struct* sh)
{
  SH2movlm(sh,6,4);
}
#define init_2_6_5_6
void FASTCALL call_2_6_5_6(SH2_struct* sh)
{
  SH2movlm(sh,6,5);
}
#define init_2_6_6_6
void FASTCALL call_2_6_6_6(SH2_struct* sh)
{
  SH2movlm(sh,6,6);
}
#define init_2_6_7_6
void FASTCALL call_2_6_7_6(SH2_struct* sh)
{
  SH2movlm(sh,6,7);
}
#define init_2_6_8_6
void FASTCALL call_2_6_8_6(SH2_struct* sh)
{
  SH2movlm(sh,6,8);
}
#define init_2_6_9_6
void FASTCALL call_2_6_9_6(SH2_struct* sh)
{
  SH2movlm(sh,6,9);
}
#define init_2_6_10_6
void FASTCALL call_2_6_10_6(SH2_struct* sh)
{
  SH2movlm(sh,6,10);
}
#define init_2_6_11_6
void FASTCALL call_2_6_11_6(SH2_struct* sh)
{
  SH2movlm(sh,6,11);
}
#define init_2_6_12_6
void FASTCALL call_2_6_12_6(SH2_struct* sh)
{
  SH2movlm(sh,6,12);
}
#define init_2_6_13_6
void FASTCALL call_2_6_13_6(SH2_struct* sh)
{
  SH2movlm(sh,6,13);
}
#define init_2_6_14_6
void FASTCALL call_2_6_14_6(SH2_struct* sh)
{
  SH2movlm(sh,6,14);
}
#define init_2_6_15_6
void FASTCALL call_2_6_15_6(SH2_struct* sh)
{
  SH2movlm(sh,6,15);
}
#define init_2_7_0_6
void FASTCALL call_2_7_0_6(SH2_struct* sh)
{
  SH2movlm(sh,7,0);
}
#define init_2_7_1_6
void FASTCALL call_2_7_1_6(SH2_struct* sh)
{
  SH2movlm(sh,7,1);
}
#define init_2_7_2_6
void FASTCALL call_2_7_2_6(SH2_struct* sh)
{
  SH2movlm(sh,7,2);
}
#define init_2_7_3_6
void FASTCALL call_2_7_3_6(SH2_struct* sh)
{
  SH2movlm(sh,7,3);
}
#define init_2_7_4_6
void FASTCALL call_2_7_4_6(SH2_struct* sh)
{
  SH2movlm(sh,7,4);
}
#define init_2_7_5_6
void FASTCALL call_2_7_5_6(SH2_struct* sh)
{
  SH2movlm(sh,7,5);
}
#define init_2_7_6_6
void FASTCALL call_2_7_6_6(SH2_struct* sh)
{
  SH2movlm(sh,7,6);
}
#define init_2_7_7_6
void FASTCALL call_2_7_7_6(SH2_struct* sh)
{
  SH2movlm(sh,7,7);
}
#define init_2_7_8_6
void FASTCALL call_2_7_8_6(SH2_struct* sh)
{
  SH2movlm(sh,7,8);
}
#define init_2_7_9_6
void FASTCALL call_2_7_9_6(SH2_struct* sh)
{
  SH2movlm(sh,7,9);
}
#define init_2_7_10_6
void FASTCALL call_2_7_10_6(SH2_struct* sh)
{
  SH2movlm(sh,7,10);
}
#define init_2_7_11_6
void FASTCALL call_2_7_11_6(SH2_struct* sh)
{
  SH2movlm(sh,7,11);
}
#define init_2_7_12_6
void FASTCALL call_2_7_12_6(SH2_struct* sh)
{
  SH2movlm(sh,7,12);
}
#define init_2_7_13_6
void FASTCALL call_2_7_13_6(SH2_struct* sh)
{
  SH2movlm(sh,7,13);
}
#define init_2_7_14_6
void FASTCALL call_2_7_14_6(SH2_struct* sh)
{
  SH2movlm(sh,7,14);
}
#define init_2_7_15_6
void FASTCALL call_2_7_15_6(SH2_struct* sh)
{
  SH2movlm(sh,7,15);
}
#define init_2_8_0_6
void FASTCALL call_2_8_0_6(SH2_struct* sh)
{
  SH2movlm(sh,8,0);
}
#define init_2_8_1_6
void FASTCALL call_2_8_1_6(SH2_struct* sh)
{
  SH2movlm(sh,8,1);
}
#define init_2_8_2_6
void FASTCALL call_2_8_2_6(SH2_struct* sh)
{
  SH2movlm(sh,8,2);
}
#define init_2_8_3_6
void FASTCALL call_2_8_3_6(SH2_struct* sh)
{
  SH2movlm(sh,8,3);
}
#define init_2_8_4_6
void FASTCALL call_2_8_4_6(SH2_struct* sh)
{
  SH2movlm(sh,8,4);
}
#define init_2_8_5_6
void FASTCALL call_2_8_5_6(SH2_struct* sh)
{
  SH2movlm(sh,8,5);
}
#define init_2_8_6_6
void FASTCALL call_2_8_6_6(SH2_struct* sh)
{
  SH2movlm(sh,8,6);
}
#define init_2_8_7_6
void FASTCALL call_2_8_7_6(SH2_struct* sh)
{
  SH2movlm(sh,8,7);
}
#define init_2_8_8_6
void FASTCALL call_2_8_8_6(SH2_struct* sh)
{
  SH2movlm(sh,8,8);
}
#define init_2_8_9_6
void FASTCALL call_2_8_9_6(SH2_struct* sh)
{
  SH2movlm(sh,8,9);
}
#define init_2_8_10_6
void FASTCALL call_2_8_10_6(SH2_struct* sh)
{
  SH2movlm(sh,8,10);
}
#define init_2_8_11_6
void FASTCALL call_2_8_11_6(SH2_struct* sh)
{
  SH2movlm(sh,8,11);
}
#define init_2_8_12_6
void FASTCALL call_2_8_12_6(SH2_struct* sh)
{
  SH2movlm(sh,8,12);
}
#define init_2_8_13_6
void FASTCALL call_2_8_13_6(SH2_struct* sh)
{
  SH2movlm(sh,8,13);
}
#define init_2_8_14_6
void FASTCALL call_2_8_14_6(SH2_struct* sh)
{
  SH2movlm(sh,8,14);
}
#define init_2_8_15_6
void FASTCALL call_2_8_15_6(SH2_struct* sh)
{
  SH2movlm(sh,8,15);
}
#define init_2_9_0_6
void FASTCALL call_2_9_0_6(SH2_struct* sh)
{
  SH2movlm(sh,9,0);
}
#define init_2_9_1_6
void FASTCALL call_2_9_1_6(SH2_struct* sh)
{
  SH2movlm(sh,9,1);
}
#define init_2_9_2_6
void FASTCALL call_2_9_2_6(SH2_struct* sh)
{
  SH2movlm(sh,9,2);
}
#define init_2_9_3_6
void FASTCALL call_2_9_3_6(SH2_struct* sh)
{
  SH2movlm(sh,9,3);
}
#define init_2_9_4_6
void FASTCALL call_2_9_4_6(SH2_struct* sh)
{
  SH2movlm(sh,9,4);
}
#define init_2_9_5_6
void FASTCALL call_2_9_5_6(SH2_struct* sh)
{
  SH2movlm(sh,9,5);
}
#define init_2_9_6_6
void FASTCALL call_2_9_6_6(SH2_struct* sh)
{
  SH2movlm(sh,9,6);
}
#define init_2_9_7_6
void FASTCALL call_2_9_7_6(SH2_struct* sh)
{
  SH2movlm(sh,9,7);
}
#define init_2_9_8_6
void FASTCALL call_2_9_8_6(SH2_struct* sh)
{
  SH2movlm(sh,9,8);
}
#define init_2_9_9_6
void FASTCALL call_2_9_9_6(SH2_struct* sh)
{
  SH2movlm(sh,9,9);
}
#define init_2_9_10_6
void FASTCALL call_2_9_10_6(SH2_struct* sh)
{
  SH2movlm(sh,9,10);
}
#define init_2_9_11_6
void FASTCALL call_2_9_11_6(SH2_struct* sh)
{
  SH2movlm(sh,9,11);
}
#define init_2_9_12_6
void FASTCALL call_2_9_12_6(SH2_struct* sh)
{
  SH2movlm(sh,9,12);
}
#define init_2_9_13_6
void FASTCALL call_2_9_13_6(SH2_struct* sh)
{
  SH2movlm(sh,9,13);
}
#define init_2_9_14_6
void FASTCALL call_2_9_14_6(SH2_struct* sh)
{
  SH2movlm(sh,9,14);
}
#define init_2_9_15_6
void FASTCALL call_2_9_15_6(SH2_struct* sh)
{
  SH2movlm(sh,9,15);
}
#define init_2_10_0_6
void FASTCALL call_2_10_0_6(SH2_struct* sh)
{
  SH2movlm(sh,10,0);
}
#define init_2_10_1_6
void FASTCALL call_2_10_1_6(SH2_struct* sh)
{
  SH2movlm(sh,10,1);
}
#define init_2_10_2_6
void FASTCALL call_2_10_2_6(SH2_struct* sh)
{
  SH2movlm(sh,10,2);
}
#define init_2_10_3_6
void FASTCALL call_2_10_3_6(SH2_struct* sh)
{
  SH2movlm(sh,10,3);
}
#define init_2_10_4_6
void FASTCALL call_2_10_4_6(SH2_struct* sh)
{
  SH2movlm(sh,10,4);
}
#define init_2_10_5_6
void FASTCALL call_2_10_5_6(SH2_struct* sh)
{
  SH2movlm(sh,10,5);
}
#define init_2_10_6_6
void FASTCALL call_2_10_6_6(SH2_struct* sh)
{
  SH2movlm(sh,10,6);
}
#define init_2_10_7_6
void FASTCALL call_2_10_7_6(SH2_struct* sh)
{
  SH2movlm(sh,10,7);
}
#define init_2_10_8_6
void FASTCALL call_2_10_8_6(SH2_struct* sh)
{
  SH2movlm(sh,10,8);
}
#define init_2_10_9_6
void FASTCALL call_2_10_9_6(SH2_struct* sh)
{
  SH2movlm(sh,10,9);
}
#define init_2_10_10_6
void FASTCALL call_2_10_10_6(SH2_struct* sh)
{
  SH2movlm(sh,10,10);
}
#define init_2_10_11_6
void FASTCALL call_2_10_11_6(SH2_struct* sh)
{
  SH2movlm(sh,10,11);
}
#define init_2_10_12_6
void FASTCALL call_2_10_12_6(SH2_struct* sh)
{
  SH2movlm(sh,10,12);
}
#define init_2_10_13_6
void FASTCALL call_2_10_13_6(SH2_struct* sh)
{
  SH2movlm(sh,10,13);
}
#define init_2_10_14_6
void FASTCALL call_2_10_14_6(SH2_struct* sh)
{
  SH2movlm(sh,10,14);
}
#define init_2_10_15_6
void FASTCALL call_2_10_15_6(SH2_struct* sh)
{
  SH2movlm(sh,10,15);
}
#define init_2_11_0_6
void FASTCALL call_2_11_0_6(SH2_struct* sh)
{
  SH2movlm(sh,11,0);
}
#define init_2_11_1_6
void FASTCALL call_2_11_1_6(SH2_struct* sh)
{
  SH2movlm(sh,11,1);
}
#define init_2_11_2_6
void FASTCALL call_2_11_2_6(SH2_struct* sh)
{
  SH2movlm(sh,11,2);
}
#define init_2_11_3_6
void FASTCALL call_2_11_3_6(SH2_struct* sh)
{
  SH2movlm(sh,11,3);
}
#define init_2_11_4_6
void FASTCALL call_2_11_4_6(SH2_struct* sh)
{
  SH2movlm(sh,11,4);
}
#define init_2_11_5_6
void FASTCALL call_2_11_5_6(SH2_struct* sh)
{
  SH2movlm(sh,11,5);
}
#define init_2_11_6_6
void FASTCALL call_2_11_6_6(SH2_struct* sh)
{
  SH2movlm(sh,11,6);
}
#define init_2_11_7_6
void FASTCALL call_2_11_7_6(SH2_struct* sh)
{
  SH2movlm(sh,11,7);
}
#define init_2_11_8_6
void FASTCALL call_2_11_8_6(SH2_struct* sh)
{
  SH2movlm(sh,11,8);
}
#define init_2_11_9_6
void FASTCALL call_2_11_9_6(SH2_struct* sh)
{
  SH2movlm(sh,11,9);
}
#define init_2_11_10_6
void FASTCALL call_2_11_10_6(SH2_struct* sh)
{
  SH2movlm(sh,11,10);
}
#define init_2_11_11_6
void FASTCALL call_2_11_11_6(SH2_struct* sh)
{
  SH2movlm(sh,11,11);
}
#define init_2_11_12_6
void FASTCALL call_2_11_12_6(SH2_struct* sh)
{
  SH2movlm(sh,11,12);
}
#define init_2_11_13_6
void FASTCALL call_2_11_13_6(SH2_struct* sh)
{
  SH2movlm(sh,11,13);
}
#define init_2_11_14_6
void FASTCALL call_2_11_14_6(SH2_struct* sh)
{
  SH2movlm(sh,11,14);
}
#define init_2_11_15_6
void FASTCALL call_2_11_15_6(SH2_struct* sh)
{
  SH2movlm(sh,11,15);
}
#define init_2_12_0_6
void FASTCALL call_2_12_0_6(SH2_struct* sh)
{
  SH2movlm(sh,12,0);
}
#define init_2_12_1_6
void FASTCALL call_2_12_1_6(SH2_struct* sh)
{
  SH2movlm(sh,12,1);
}
#define init_2_12_2_6
void FASTCALL call_2_12_2_6(SH2_struct* sh)
{
  SH2movlm(sh,12,2);
}
#define init_2_12_3_6
void FASTCALL call_2_12_3_6(SH2_struct* sh)
{
  SH2movlm(sh,12,3);
}
#define init_2_12_4_6
void FASTCALL call_2_12_4_6(SH2_struct* sh)
{
  SH2movlm(sh,12,4);
}
#define init_2_12_5_6
void FASTCALL call_2_12_5_6(SH2_struct* sh)
{
  SH2movlm(sh,12,5);
}
#define init_2_12_6_6
void FASTCALL call_2_12_6_6(SH2_struct* sh)
{
  SH2movlm(sh,12,6);
}
#define init_2_12_7_6
void FASTCALL call_2_12_7_6(SH2_struct* sh)
{
  SH2movlm(sh,12,7);
}
#define init_2_12_8_6
void FASTCALL call_2_12_8_6(SH2_struct* sh)
{
  SH2movlm(sh,12,8);
}
#define init_2_12_9_6
void FASTCALL call_2_12_9_6(SH2_struct* sh)
{
  SH2movlm(sh,12,9);
}
#define init_2_12_10_6
void FASTCALL call_2_12_10_6(SH2_struct* sh)
{
  SH2movlm(sh,12,10);
}
#define init_2_12_11_6
void FASTCALL call_2_12_11_6(SH2_struct* sh)
{
  SH2movlm(sh,12,11);
}
#define init_2_12_12_6
void FASTCALL call_2_12_12_6(SH2_struct* sh)
{
  SH2movlm(sh,12,12);
}
#define init_2_12_13_6
void FASTCALL call_2_12_13_6(SH2_struct* sh)
{
  SH2movlm(sh,12,13);
}
#define init_2_12_14_6
void FASTCALL call_2_12_14_6(SH2_struct* sh)
{
  SH2movlm(sh,12,14);
}
#define init_2_12_15_6
void FASTCALL call_2_12_15_6(SH2_struct* sh)
{
  SH2movlm(sh,12,15);
}
#define init_2_13_0_6
void FASTCALL call_2_13_0_6(SH2_struct* sh)
{
  SH2movlm(sh,13,0);
}
#define init_2_13_1_6
void FASTCALL call_2_13_1_6(SH2_struct* sh)
{
  SH2movlm(sh,13,1);
}
#define init_2_13_2_6
void FASTCALL call_2_13_2_6(SH2_struct* sh)
{
  SH2movlm(sh,13,2);
}
#define init_2_13_3_6
void FASTCALL call_2_13_3_6(SH2_struct* sh)
{
  SH2movlm(sh,13,3);
}
#define init_2_13_4_6
void FASTCALL call_2_13_4_6(SH2_struct* sh)
{
  SH2movlm(sh,13,4);
}
#define init_2_13_5_6
void FASTCALL call_2_13_5_6(SH2_struct* sh)
{
  SH2movlm(sh,13,5);
}
#define init_2_13_6_6
void FASTCALL call_2_13_6_6(SH2_struct* sh)
{
  SH2movlm(sh,13,6);
}
#define init_2_13_7_6
void FASTCALL call_2_13_7_6(SH2_struct* sh)
{
  SH2movlm(sh,13,7);
}
#define init_2_13_8_6
void FASTCALL call_2_13_8_6(SH2_struct* sh)
{
  SH2movlm(sh,13,8);
}
#define init_2_13_9_6
void FASTCALL call_2_13_9_6(SH2_struct* sh)
{
  SH2movlm(sh,13,9);
}
#define init_2_13_10_6
void FASTCALL call_2_13_10_6(SH2_struct* sh)
{
  SH2movlm(sh,13,10);
}
#define init_2_13_11_6
void FASTCALL call_2_13_11_6(SH2_struct* sh)
{
  SH2movlm(sh,13,11);
}
#define init_2_13_12_6
void FASTCALL call_2_13_12_6(SH2_struct* sh)
{
  SH2movlm(sh,13,12);
}
#define init_2_13_13_6
void FASTCALL call_2_13_13_6(SH2_struct* sh)
{
  SH2movlm(sh,13,13);
}
#define init_2_13_14_6
void FASTCALL call_2_13_14_6(SH2_struct* sh)
{
  SH2movlm(sh,13,14);
}
#define init_2_13_15_6
void FASTCALL call_2_13_15_6(SH2_struct* sh)
{
  SH2movlm(sh,13,15);
}
#define init_2_14_0_6
void FASTCALL call_2_14_0_6(SH2_struct* sh)
{
  SH2movlm(sh,14,0);
}
#define init_2_14_1_6
void FASTCALL call_2_14_1_6(SH2_struct* sh)
{
  SH2movlm(sh,14,1);
}
#define init_2_14_2_6
void FASTCALL call_2_14_2_6(SH2_struct* sh)
{
  SH2movlm(sh,14,2);
}
#define init_2_14_3_6
void FASTCALL call_2_14_3_6(SH2_struct* sh)
{
  SH2movlm(sh,14,3);
}
#define init_2_14_4_6
void FASTCALL call_2_14_4_6(SH2_struct* sh)
{
  SH2movlm(sh,14,4);
}
#define init_2_14_5_6
void FASTCALL call_2_14_5_6(SH2_struct* sh)
{
  SH2movlm(sh,14,5);
}
#define init_2_14_6_6
void FASTCALL call_2_14_6_6(SH2_struct* sh)
{
  SH2movlm(sh,14,6);
}
#define init_2_14_7_6
void FASTCALL call_2_14_7_6(SH2_struct* sh)
{
  SH2movlm(sh,14,7);
}
#define init_2_14_8_6
void FASTCALL call_2_14_8_6(SH2_struct* sh)
{
  SH2movlm(sh,14,8);
}
#define init_2_14_9_6
void FASTCALL call_2_14_9_6(SH2_struct* sh)
{
  SH2movlm(sh,14,9);
}
#define init_2_14_10_6
void FASTCALL call_2_14_10_6(SH2_struct* sh)
{
  SH2movlm(sh,14,10);
}
#define init_2_14_11_6
void FASTCALL call_2_14_11_6(SH2_struct* sh)
{
  SH2movlm(sh,14,11);
}
#define init_2_14_12_6
void FASTCALL call_2_14_12_6(SH2_struct* sh)
{
  SH2movlm(sh,14,12);
}
#define init_2_14_13_6
void FASTCALL call_2_14_13_6(SH2_struct* sh)
{
  SH2movlm(sh,14,13);
}
#define init_2_14_14_6
void FASTCALL call_2_14_14_6(SH2_struct* sh)
{
  SH2movlm(sh,14,14);
}
#define init_2_14_15_6
void FASTCALL call_2_14_15_6(SH2_struct* sh)
{
  SH2movlm(sh,14,15);
}
#define init_2_15_0_6
void FASTCALL call_2_15_0_6(SH2_struct* sh)
{
  SH2movlm(sh,15,0);
}
#define init_2_15_1_6
void FASTCALL call_2_15_1_6(SH2_struct* sh)
{
  SH2movlm(sh,15,1);
}
#define init_2_15_2_6
void FASTCALL call_2_15_2_6(SH2_struct* sh)
{
  SH2movlm(sh,15,2);
}
#define init_2_15_3_6
void FASTCALL call_2_15_3_6(SH2_struct* sh)
{
  SH2movlm(sh,15,3);
}
#define init_2_15_4_6
void FASTCALL call_2_15_4_6(SH2_struct* sh)
{
  SH2movlm(sh,15,4);
}
#define init_2_15_5_6
void FASTCALL call_2_15_5_6(SH2_struct* sh)
{
  SH2movlm(sh,15,5);
}
#define init_2_15_6_6
void FASTCALL call_2_15_6_6(SH2_struct* sh)
{
  SH2movlm(sh,15,6);
}
#define init_2_15_7_6
void FASTCALL call_2_15_7_6(SH2_struct* sh)
{
  SH2movlm(sh,15,7);
}
#define init_2_15_8_6
void FASTCALL call_2_15_8_6(SH2_struct* sh)
{
  SH2movlm(sh,15,8);
}
#define init_2_15_9_6
void FASTCALL call_2_15_9_6(SH2_struct* sh)
{
  SH2movlm(sh,15,9);
}
#define init_2_15_10_6
void FASTCALL call_2_15_10_6(SH2_struct* sh)
{
  SH2movlm(sh,15,10);
}
#define init_2_15_11_6
void FASTCALL call_2_15_11_6(SH2_struct* sh)
{
  SH2movlm(sh,15,11);
}
#define init_2_15_12_6
void FASTCALL call_2_15_12_6(SH2_struct* sh)
{
  SH2movlm(sh,15,12);
}
#define init_2_15_13_6
void FASTCALL call_2_15_13_6(SH2_struct* sh)
{
  SH2movlm(sh,15,13);
}
#define init_2_15_14_6
void FASTCALL call_2_15_14_6(SH2_struct* sh)
{
  SH2movlm(sh,15,14);
}
#define init_2_15_15_6
void FASTCALL call_2_15_15_6(SH2_struct* sh)
{
  SH2movlm(sh,15,15);
}
#define init_12_6_0_0
void FASTCALL call_12_6_0_0(SH2_struct* sh)
{
  SH2movllg(sh,(0<<4)+0);
}
#define init_12_6_0_1
void FASTCALL call_12_6_0_1(SH2_struct* sh)
{
  SH2movllg(sh,(0<<4)+1);
}
#define init_12_6_0_2
void FASTCALL call_12_6_0_2(SH2_struct* sh)
{
  SH2movllg(sh,(0<<4)+2);
}
#define init_12_6_0_3
void FASTCALL call_12_6_0_3(SH2_struct* sh)
{
  SH2movllg(sh,(0<<4)+3);
}
#define init_12_6_0_4
void FASTCALL call_12_6_0_4(SH2_struct* sh)
{
  SH2movllg(sh,(0<<4)+4);
}
#define init_12_6_0_5
void FASTCALL call_12_6_0_5(SH2_struct* sh)
{
  SH2movllg(sh,(0<<4)+5);
}
#define init_12_6_0_6
void FASTCALL call_12_6_0_6(SH2_struct* sh)
{
  SH2movllg(sh,(0<<4)+6);
}
#define init_12_6_0_7
void FASTCALL call_12_6_0_7(SH2_struct* sh)
{
  SH2movllg(sh,(0<<4)+7);
}
#define init_12_6_0_8
void FASTCALL call_12_6_0_8(SH2_struct* sh)
{
  SH2movllg(sh,(0<<4)+8);
}
#define init_12_6_0_9
void FASTCALL call_12_6_0_9(SH2_struct* sh)
{
  SH2movllg(sh,(0<<4)+9);
}
#define init_12_6_0_10
void FASTCALL call_12_6_0_10(SH2_struct* sh)
{
  SH2movllg(sh,(0<<4)+10);
}
#define init_12_6_0_11
void FASTCALL call_12_6_0_11(SH2_struct* sh)
{
  SH2movllg(sh,(0<<4)+11);
}
#define init_12_6_0_12
void FASTCALL call_12_6_0_12(SH2_struct* sh)
{
  SH2movllg(sh,(0<<4)+12);
}
#define init_12_6_0_13
void FASTCALL call_12_6_0_13(SH2_struct* sh)
{
  SH2movllg(sh,(0<<4)+13);
}
#define init_12_6_0_14
void FASTCALL call_12_6_0_14(SH2_struct* sh)
{
  SH2movllg(sh,(0<<4)+14);
}
#define init_12_6_0_15
void FASTCALL call_12_6_0_15(SH2_struct* sh)
{
  SH2movllg(sh,(0<<4)+15);
}
#define init_12_6_1_0
void FASTCALL call_12_6_1_0(SH2_struct* sh)
{
  SH2movllg(sh,(1<<4)+0);
}
#define init_12_6_1_1
void FASTCALL call_12_6_1_1(SH2_struct* sh)
{
  SH2movllg(sh,(1<<4)+1);
}
#define init_12_6_1_2
void FASTCALL call_12_6_1_2(SH2_struct* sh)
{
  SH2movllg(sh,(1<<4)+2);
}
#define init_12_6_1_3
void FASTCALL call_12_6_1_3(SH2_struct* sh)
{
  SH2movllg(sh,(1<<4)+3);
}
#define init_12_6_1_4
void FASTCALL call_12_6_1_4(SH2_struct* sh)
{
  SH2movllg(sh,(1<<4)+4);
}
#define init_12_6_1_5
void FASTCALL call_12_6_1_5(SH2_struct* sh)
{
  SH2movllg(sh,(1<<4)+5);
}
#define init_12_6_1_6
void FASTCALL call_12_6_1_6(SH2_struct* sh)
{
  SH2movllg(sh,(1<<4)+6);
}
#define init_12_6_1_7
void FASTCALL call_12_6_1_7(SH2_struct* sh)
{
  SH2movllg(sh,(1<<4)+7);
}
#define init_12_6_1_8
void FASTCALL call_12_6_1_8(SH2_struct* sh)
{
  SH2movllg(sh,(1<<4)+8);
}
#define init_12_6_1_9
void FASTCALL call_12_6_1_9(SH2_struct* sh)
{
  SH2movllg(sh,(1<<4)+9);
}
#define init_12_6_1_10
void FASTCALL call_12_6_1_10(SH2_struct* sh)
{
  SH2movllg(sh,(1<<4)+10);
}
#define init_12_6_1_11
void FASTCALL call_12_6_1_11(SH2_struct* sh)
{
  SH2movllg(sh,(1<<4)+11);
}
#define init_12_6_1_12
void FASTCALL call_12_6_1_12(SH2_struct* sh)
{
  SH2movllg(sh,(1<<4)+12);
}
#define init_12_6_1_13
void FASTCALL call_12_6_1_13(SH2_struct* sh)
{
  SH2movllg(sh,(1<<4)+13);
}
#define init_12_6_1_14
void FASTCALL call_12_6_1_14(SH2_struct* sh)
{
  SH2movllg(sh,(1<<4)+14);
}
#define init_12_6_1_15
void FASTCALL call_12_6_1_15(SH2_struct* sh)
{
  SH2movllg(sh,(1<<4)+15);
}
#define init_12_6_2_0
void FASTCALL call_12_6_2_0(SH2_struct* sh)
{
  SH2movllg(sh,(2<<4)+0);
}
#define init_12_6_2_1
void FASTCALL call_12_6_2_1(SH2_struct* sh)
{
  SH2movllg(sh,(2<<4)+1);
}
#define init_12_6_2_2
void FASTCALL call_12_6_2_2(SH2_struct* sh)
{
  SH2movllg(sh,(2<<4)+2);
}
#define init_12_6_2_3
void FASTCALL call_12_6_2_3(SH2_struct* sh)
{
  SH2movllg(sh,(2<<4)+3);
}
#define init_12_6_2_4
void FASTCALL call_12_6_2_4(SH2_struct* sh)
{
  SH2movllg(sh,(2<<4)+4);
}
#define init_12_6_2_5
void FASTCALL call_12_6_2_5(SH2_struct* sh)
{
  SH2movllg(sh,(2<<4)+5);
}
#define init_12_6_2_6
void FASTCALL call_12_6_2_6(SH2_struct* sh)
{
  SH2movllg(sh,(2<<4)+6);
}
#define init_12_6_2_7
void FASTCALL call_12_6_2_7(SH2_struct* sh)
{
  SH2movllg(sh,(2<<4)+7);
}
#define init_12_6_2_8
void FASTCALL call_12_6_2_8(SH2_struct* sh)
{
  SH2movllg(sh,(2<<4)+8);
}
#define init_12_6_2_9
void FASTCALL call_12_6_2_9(SH2_struct* sh)
{
  SH2movllg(sh,(2<<4)+9);
}
#define init_12_6_2_10
void FASTCALL call_12_6_2_10(SH2_struct* sh)
{
  SH2movllg(sh,(2<<4)+10);
}
#define init_12_6_2_11
void FASTCALL call_12_6_2_11(SH2_struct* sh)
{
  SH2movllg(sh,(2<<4)+11);
}
#define init_12_6_2_12
void FASTCALL call_12_6_2_12(SH2_struct* sh)
{
  SH2movllg(sh,(2<<4)+12);
}
#define init_12_6_2_13
void FASTCALL call_12_6_2_13(SH2_struct* sh)
{
  SH2movllg(sh,(2<<4)+13);
}
#define init_12_6_2_14
void FASTCALL call_12_6_2_14(SH2_struct* sh)
{
  SH2movllg(sh,(2<<4)+14);
}
#define init_12_6_2_15
void FASTCALL call_12_6_2_15(SH2_struct* sh)
{
  SH2movllg(sh,(2<<4)+15);
}
#define init_12_6_3_0
void FASTCALL call_12_6_3_0(SH2_struct* sh)
{
  SH2movllg(sh,(3<<4)+0);
}
#define init_12_6_3_1
void FASTCALL call_12_6_3_1(SH2_struct* sh)
{
  SH2movllg(sh,(3<<4)+1);
}
#define init_12_6_3_2
void FASTCALL call_12_6_3_2(SH2_struct* sh)
{
  SH2movllg(sh,(3<<4)+2);
}
#define init_12_6_3_3
void FASTCALL call_12_6_3_3(SH2_struct* sh)
{
  SH2movllg(sh,(3<<4)+3);
}
#define init_12_6_3_4
void FASTCALL call_12_6_3_4(SH2_struct* sh)
{
  SH2movllg(sh,(3<<4)+4);
}
#define init_12_6_3_5
void FASTCALL call_12_6_3_5(SH2_struct* sh)
{
  SH2movllg(sh,(3<<4)+5);
}
#define init_12_6_3_6
void FASTCALL call_12_6_3_6(SH2_struct* sh)
{
  SH2movllg(sh,(3<<4)+6);
}
#define init_12_6_3_7
void FASTCALL call_12_6_3_7(SH2_struct* sh)
{
  SH2movllg(sh,(3<<4)+7);
}
#define init_12_6_3_8
void FASTCALL call_12_6_3_8(SH2_struct* sh)
{
  SH2movllg(sh,(3<<4)+8);
}
#define init_12_6_3_9
void FASTCALL call_12_6_3_9(SH2_struct* sh)
{
  SH2movllg(sh,(3<<4)+9);
}
#define init_12_6_3_10
void FASTCALL call_12_6_3_10(SH2_struct* sh)
{
  SH2movllg(sh,(3<<4)+10);
}
#define init_12_6_3_11
void FASTCALL call_12_6_3_11(SH2_struct* sh)
{
  SH2movllg(sh,(3<<4)+11);
}
#define init_12_6_3_12
void FASTCALL call_12_6_3_12(SH2_struct* sh)
{
  SH2movllg(sh,(3<<4)+12);
}
#define init_12_6_3_13
void FASTCALL call_12_6_3_13(SH2_struct* sh)
{
  SH2movllg(sh,(3<<4)+13);
}
#define init_12_6_3_14
void FASTCALL call_12_6_3_14(SH2_struct* sh)
{
  SH2movllg(sh,(3<<4)+14);
}
#define init_12_6_3_15
void FASTCALL call_12_6_3_15(SH2_struct* sh)
{
  SH2movllg(sh,(3<<4)+15);
}
#define init_12_6_4_0
void FASTCALL call_12_6_4_0(SH2_struct* sh)
{
  SH2movllg(sh,(4<<4)+0);
}
#define init_12_6_4_1
void FASTCALL call_12_6_4_1(SH2_struct* sh)
{
  SH2movllg(sh,(4<<4)+1);
}
#define init_12_6_4_2
void FASTCALL call_12_6_4_2(SH2_struct* sh)
{
  SH2movllg(sh,(4<<4)+2);
}
#define init_12_6_4_3
void FASTCALL call_12_6_4_3(SH2_struct* sh)
{
  SH2movllg(sh,(4<<4)+3);
}
#define init_12_6_4_4
void FASTCALL call_12_6_4_4(SH2_struct* sh)
{
  SH2movllg(sh,(4<<4)+4);
}
#define init_12_6_4_5
void FASTCALL call_12_6_4_5(SH2_struct* sh)
{
  SH2movllg(sh,(4<<4)+5);
}
#define init_12_6_4_6
void FASTCALL call_12_6_4_6(SH2_struct* sh)
{
  SH2movllg(sh,(4<<4)+6);
}
#define init_12_6_4_7
void FASTCALL call_12_6_4_7(SH2_struct* sh)
{
  SH2movllg(sh,(4<<4)+7);
}
#define init_12_6_4_8
void FASTCALL call_12_6_4_8(SH2_struct* sh)
{
  SH2movllg(sh,(4<<4)+8);
}
#define init_12_6_4_9
void FASTCALL call_12_6_4_9(SH2_struct* sh)
{
  SH2movllg(sh,(4<<4)+9);
}
#define init_12_6_4_10
void FASTCALL call_12_6_4_10(SH2_struct* sh)
{
  SH2movllg(sh,(4<<4)+10);
}
#define init_12_6_4_11
void FASTCALL call_12_6_4_11(SH2_struct* sh)
{
  SH2movllg(sh,(4<<4)+11);
}
#define init_12_6_4_12
void FASTCALL call_12_6_4_12(SH2_struct* sh)
{
  SH2movllg(sh,(4<<4)+12);
}
#define init_12_6_4_13
void FASTCALL call_12_6_4_13(SH2_struct* sh)
{
  SH2movllg(sh,(4<<4)+13);
}
#define init_12_6_4_14
void FASTCALL call_12_6_4_14(SH2_struct* sh)
{
  SH2movllg(sh,(4<<4)+14);
}
#define init_12_6_4_15
void FASTCALL call_12_6_4_15(SH2_struct* sh)
{
  SH2movllg(sh,(4<<4)+15);
}
#define init_12_6_5_0
void FASTCALL call_12_6_5_0(SH2_struct* sh)
{
  SH2movllg(sh,(5<<4)+0);
}
#define init_12_6_5_1
void FASTCALL call_12_6_5_1(SH2_struct* sh)
{
  SH2movllg(sh,(5<<4)+1);
}
#define init_12_6_5_2
void FASTCALL call_12_6_5_2(SH2_struct* sh)
{
  SH2movllg(sh,(5<<4)+2);
}
#define init_12_6_5_3
void FASTCALL call_12_6_5_3(SH2_struct* sh)
{
  SH2movllg(sh,(5<<4)+3);
}
#define init_12_6_5_4
void FASTCALL call_12_6_5_4(SH2_struct* sh)
{
  SH2movllg(sh,(5<<4)+4);
}
#define init_12_6_5_5
void FASTCALL call_12_6_5_5(SH2_struct* sh)
{
  SH2movllg(sh,(5<<4)+5);
}
#define init_12_6_5_6
void FASTCALL call_12_6_5_6(SH2_struct* sh)
{
  SH2movllg(sh,(5<<4)+6);
}
#define init_12_6_5_7
void FASTCALL call_12_6_5_7(SH2_struct* sh)
{
  SH2movllg(sh,(5<<4)+7);
}
#define init_12_6_5_8
void FASTCALL call_12_6_5_8(SH2_struct* sh)
{
  SH2movllg(sh,(5<<4)+8);
}
#define init_12_6_5_9
void FASTCALL call_12_6_5_9(SH2_struct* sh)
{
  SH2movllg(sh,(5<<4)+9);
}
#define init_12_6_5_10
void FASTCALL call_12_6_5_10(SH2_struct* sh)
{
  SH2movllg(sh,(5<<4)+10);
}
#define init_12_6_5_11
void FASTCALL call_12_6_5_11(SH2_struct* sh)
{
  SH2movllg(sh,(5<<4)+11);
}
#define init_12_6_5_12
void FASTCALL call_12_6_5_12(SH2_struct* sh)
{
  SH2movllg(sh,(5<<4)+12);
}
#define init_12_6_5_13
void FASTCALL call_12_6_5_13(SH2_struct* sh)
{
  SH2movllg(sh,(5<<4)+13);
}
#define init_12_6_5_14
void FASTCALL call_12_6_5_14(SH2_struct* sh)
{
  SH2movllg(sh,(5<<4)+14);
}
#define init_12_6_5_15
void FASTCALL call_12_6_5_15(SH2_struct* sh)
{
  SH2movllg(sh,(5<<4)+15);
}
#define init_12_6_6_0
void FASTCALL call_12_6_6_0(SH2_struct* sh)
{
  SH2movllg(sh,(6<<4)+0);
}
#define init_12_6_6_1
void FASTCALL call_12_6_6_1(SH2_struct* sh)
{
  SH2movllg(sh,(6<<4)+1);
}
#define init_12_6_6_2
void FASTCALL call_12_6_6_2(SH2_struct* sh)
{
  SH2movllg(sh,(6<<4)+2);
}
#define init_12_6_6_3
void FASTCALL call_12_6_6_3(SH2_struct* sh)
{
  SH2movllg(sh,(6<<4)+3);
}
#define init_12_6_6_4
void FASTCALL call_12_6_6_4(SH2_struct* sh)
{
  SH2movllg(sh,(6<<4)+4);
}
#define init_12_6_6_5
void FASTCALL call_12_6_6_5(SH2_struct* sh)
{
  SH2movllg(sh,(6<<4)+5);
}
#define init_12_6_6_6
void FASTCALL call_12_6_6_6(SH2_struct* sh)
{
  SH2movllg(sh,(6<<4)+6);
}
#define init_12_6_6_7
void FASTCALL call_12_6_6_7(SH2_struct* sh)
{
  SH2movllg(sh,(6<<4)+7);
}
#define init_12_6_6_8
void FASTCALL call_12_6_6_8(SH2_struct* sh)
{
  SH2movllg(sh,(6<<4)+8);
}
#define init_12_6_6_9
void FASTCALL call_12_6_6_9(SH2_struct* sh)
{
  SH2movllg(sh,(6<<4)+9);
}
#define init_12_6_6_10
void FASTCALL call_12_6_6_10(SH2_struct* sh)
{
  SH2movllg(sh,(6<<4)+10);
}
#define init_12_6_6_11
void FASTCALL call_12_6_6_11(SH2_struct* sh)
{
  SH2movllg(sh,(6<<4)+11);
}
#define init_12_6_6_12
void FASTCALL call_12_6_6_12(SH2_struct* sh)
{
  SH2movllg(sh,(6<<4)+12);
}
#define init_12_6_6_13
void FASTCALL call_12_6_6_13(SH2_struct* sh)
{
  SH2movllg(sh,(6<<4)+13);
}
#define init_12_6_6_14
void FASTCALL call_12_6_6_14(SH2_struct* sh)
{
  SH2movllg(sh,(6<<4)+14);
}
#define init_12_6_6_15
void FASTCALL call_12_6_6_15(SH2_struct* sh)
{
  SH2movllg(sh,(6<<4)+15);
}
#define init_12_6_7_0
void FASTCALL call_12_6_7_0(SH2_struct* sh)
{
  SH2movllg(sh,(7<<4)+0);
}
#define init_12_6_7_1
void FASTCALL call_12_6_7_1(SH2_struct* sh)
{
  SH2movllg(sh,(7<<4)+1);
}
#define init_12_6_7_2
void FASTCALL call_12_6_7_2(SH2_struct* sh)
{
  SH2movllg(sh,(7<<4)+2);
}
#define init_12_6_7_3
void FASTCALL call_12_6_7_3(SH2_struct* sh)
{
  SH2movllg(sh,(7<<4)+3);
}
#define init_12_6_7_4
void FASTCALL call_12_6_7_4(SH2_struct* sh)
{
  SH2movllg(sh,(7<<4)+4);
}
#define init_12_6_7_5
void FASTCALL call_12_6_7_5(SH2_struct* sh)
{
  SH2movllg(sh,(7<<4)+5);
}
#define init_12_6_7_6
void FASTCALL call_12_6_7_6(SH2_struct* sh)
{
  SH2movllg(sh,(7<<4)+6);
}
#define init_12_6_7_7
void FASTCALL call_12_6_7_7(SH2_struct* sh)
{
  SH2movllg(sh,(7<<4)+7);
}
#define init_12_6_7_8
void FASTCALL call_12_6_7_8(SH2_struct* sh)
{
  SH2movllg(sh,(7<<4)+8);
}
#define init_12_6_7_9
void FASTCALL call_12_6_7_9(SH2_struct* sh)
{
  SH2movllg(sh,(7<<4)+9);
}
#define init_12_6_7_10
void FASTCALL call_12_6_7_10(SH2_struct* sh)
{
  SH2movllg(sh,(7<<4)+10);
}
#define init_12_6_7_11
void FASTCALL call_12_6_7_11(SH2_struct* sh)
{
  SH2movllg(sh,(7<<4)+11);
}
#define init_12_6_7_12
void FASTCALL call_12_6_7_12(SH2_struct* sh)
{
  SH2movllg(sh,(7<<4)+12);
}
#define init_12_6_7_13
void FASTCALL call_12_6_7_13(SH2_struct* sh)
{
  SH2movllg(sh,(7<<4)+13);
}
#define init_12_6_7_14
void FASTCALL call_12_6_7_14(SH2_struct* sh)
{
  SH2movllg(sh,(7<<4)+14);
}
#define init_12_6_7_15
void FASTCALL call_12_6_7_15(SH2_struct* sh)
{
  SH2movllg(sh,(7<<4)+15);
}
#define init_12_6_8_0
void FASTCALL call_12_6_8_0(SH2_struct* sh)
{
  SH2movllg(sh,(8<<4)+0);
}
#define init_12_6_8_1
void FASTCALL call_12_6_8_1(SH2_struct* sh)
{
  SH2movllg(sh,(8<<4)+1);
}
#define init_12_6_8_2
void FASTCALL call_12_6_8_2(SH2_struct* sh)
{
  SH2movllg(sh,(8<<4)+2);
}
#define init_12_6_8_3
void FASTCALL call_12_6_8_3(SH2_struct* sh)
{
  SH2movllg(sh,(8<<4)+3);
}
#define init_12_6_8_4
void FASTCALL call_12_6_8_4(SH2_struct* sh)
{
  SH2movllg(sh,(8<<4)+4);
}
#define init_12_6_8_5
void FASTCALL call_12_6_8_5(SH2_struct* sh)
{
  SH2movllg(sh,(8<<4)+5);
}
#define init_12_6_8_6
void FASTCALL call_12_6_8_6(SH2_struct* sh)
{
  SH2movllg(sh,(8<<4)+6);
}
#define init_12_6_8_7
void FASTCALL call_12_6_8_7(SH2_struct* sh)
{
  SH2movllg(sh,(8<<4)+7);
}
#define init_12_6_8_8
void FASTCALL call_12_6_8_8(SH2_struct* sh)
{
  SH2movllg(sh,(8<<4)+8);
}
#define init_12_6_8_9
void FASTCALL call_12_6_8_9(SH2_struct* sh)
{
  SH2movllg(sh,(8<<4)+9);
}
#define init_12_6_8_10
void FASTCALL call_12_6_8_10(SH2_struct* sh)
{
  SH2movllg(sh,(8<<4)+10);
}
#define init_12_6_8_11
void FASTCALL call_12_6_8_11(SH2_struct* sh)
{
  SH2movllg(sh,(8<<4)+11);
}
#define init_12_6_8_12
void FASTCALL call_12_6_8_12(SH2_struct* sh)
{
  SH2movllg(sh,(8<<4)+12);
}
#define init_12_6_8_13
void FASTCALL call_12_6_8_13(SH2_struct* sh)
{
  SH2movllg(sh,(8<<4)+13);
}
#define init_12_6_8_14
void FASTCALL call_12_6_8_14(SH2_struct* sh)
{
  SH2movllg(sh,(8<<4)+14);
}
#define init_12_6_8_15
void FASTCALL call_12_6_8_15(SH2_struct* sh)
{
  SH2movllg(sh,(8<<4)+15);
}
#define init_12_6_9_0
void FASTCALL call_12_6_9_0(SH2_struct* sh)
{
  SH2movllg(sh,(9<<4)+0);
}
#define init_12_6_9_1
void FASTCALL call_12_6_9_1(SH2_struct* sh)
{
  SH2movllg(sh,(9<<4)+1);
}
#define init_12_6_9_2
void FASTCALL call_12_6_9_2(SH2_struct* sh)
{
  SH2movllg(sh,(9<<4)+2);
}
#define init_12_6_9_3
void FASTCALL call_12_6_9_3(SH2_struct* sh)
{
  SH2movllg(sh,(9<<4)+3);
}
#define init_12_6_9_4
void FASTCALL call_12_6_9_4(SH2_struct* sh)
{
  SH2movllg(sh,(9<<4)+4);
}
#define init_12_6_9_5
void FASTCALL call_12_6_9_5(SH2_struct* sh)
{
  SH2movllg(sh,(9<<4)+5);
}
#define init_12_6_9_6
void FASTCALL call_12_6_9_6(SH2_struct* sh)
{
  SH2movllg(sh,(9<<4)+6);
}
#define init_12_6_9_7
void FASTCALL call_12_6_9_7(SH2_struct* sh)
{
  SH2movllg(sh,(9<<4)+7);
}
#define init_12_6_9_8
void FASTCALL call_12_6_9_8(SH2_struct* sh)
{
  SH2movllg(sh,(9<<4)+8);
}
#define init_12_6_9_9
void FASTCALL call_12_6_9_9(SH2_struct* sh)
{
  SH2movllg(sh,(9<<4)+9);
}
#define init_12_6_9_10
void FASTCALL call_12_6_9_10(SH2_struct* sh)
{
  SH2movllg(sh,(9<<4)+10);
}
#define init_12_6_9_11
void FASTCALL call_12_6_9_11(SH2_struct* sh)
{
  SH2movllg(sh,(9<<4)+11);
}
#define init_12_6_9_12
void FASTCALL call_12_6_9_12(SH2_struct* sh)
{
  SH2movllg(sh,(9<<4)+12);
}
#define init_12_6_9_13
void FASTCALL call_12_6_9_13(SH2_struct* sh)
{
  SH2movllg(sh,(9<<4)+13);
}
#define init_12_6_9_14
void FASTCALL call_12_6_9_14(SH2_struct* sh)
{
  SH2movllg(sh,(9<<4)+14);
}
#define init_12_6_9_15
void FASTCALL call_12_6_9_15(SH2_struct* sh)
{
  SH2movllg(sh,(9<<4)+15);
}
#define init_12_6_10_0
void FASTCALL call_12_6_10_0(SH2_struct* sh)
{
  SH2movllg(sh,(10<<4)+0);
}
#define init_12_6_10_1
void FASTCALL call_12_6_10_1(SH2_struct* sh)
{
  SH2movllg(sh,(10<<4)+1);
}
#define init_12_6_10_2
void FASTCALL call_12_6_10_2(SH2_struct* sh)
{
  SH2movllg(sh,(10<<4)+2);
}
#define init_12_6_10_3
void FASTCALL call_12_6_10_3(SH2_struct* sh)
{
  SH2movllg(sh,(10<<4)+3);
}
#define init_12_6_10_4
void FASTCALL call_12_6_10_4(SH2_struct* sh)
{
  SH2movllg(sh,(10<<4)+4);
}
#define init_12_6_10_5
void FASTCALL call_12_6_10_5(SH2_struct* sh)
{
  SH2movllg(sh,(10<<4)+5);
}
#define init_12_6_10_6
void FASTCALL call_12_6_10_6(SH2_struct* sh)
{
  SH2movllg(sh,(10<<4)+6);
}
#define init_12_6_10_7
void FASTCALL call_12_6_10_7(SH2_struct* sh)
{
  SH2movllg(sh,(10<<4)+7);
}
#define init_12_6_10_8
void FASTCALL call_12_6_10_8(SH2_struct* sh)
{
  SH2movllg(sh,(10<<4)+8);
}
#define init_12_6_10_9
void FASTCALL call_12_6_10_9(SH2_struct* sh)
{
  SH2movllg(sh,(10<<4)+9);
}
#define init_12_6_10_10
void FASTCALL call_12_6_10_10(SH2_struct* sh)
{
  SH2movllg(sh,(10<<4)+10);
}
#define init_12_6_10_11
void FASTCALL call_12_6_10_11(SH2_struct* sh)
{
  SH2movllg(sh,(10<<4)+11);
}
#define init_12_6_10_12
void FASTCALL call_12_6_10_12(SH2_struct* sh)
{
  SH2movllg(sh,(10<<4)+12);
}
#define init_12_6_10_13
void FASTCALL call_12_6_10_13(SH2_struct* sh)
{
  SH2movllg(sh,(10<<4)+13);
}
#define init_12_6_10_14
void FASTCALL call_12_6_10_14(SH2_struct* sh)
{
  SH2movllg(sh,(10<<4)+14);
}
#define init_12_6_10_15
void FASTCALL call_12_6_10_15(SH2_struct* sh)
{
  SH2movllg(sh,(10<<4)+15);
}
#define init_12_6_11_0
void FASTCALL call_12_6_11_0(SH2_struct* sh)
{
  SH2movllg(sh,(11<<4)+0);
}
#define init_12_6_11_1
void FASTCALL call_12_6_11_1(SH2_struct* sh)
{
  SH2movllg(sh,(11<<4)+1);
}
#define init_12_6_11_2
void FASTCALL call_12_6_11_2(SH2_struct* sh)
{
  SH2movllg(sh,(11<<4)+2);
}
#define init_12_6_11_3
void FASTCALL call_12_6_11_3(SH2_struct* sh)
{
  SH2movllg(sh,(11<<4)+3);
}
#define init_12_6_11_4
void FASTCALL call_12_6_11_4(SH2_struct* sh)
{
  SH2movllg(sh,(11<<4)+4);
}
#define init_12_6_11_5
void FASTCALL call_12_6_11_5(SH2_struct* sh)
{
  SH2movllg(sh,(11<<4)+5);
}
#define init_12_6_11_6
void FASTCALL call_12_6_11_6(SH2_struct* sh)
{
  SH2movllg(sh,(11<<4)+6);
}
#define init_12_6_11_7
void FASTCALL call_12_6_11_7(SH2_struct* sh)
{
  SH2movllg(sh,(11<<4)+7);
}
#define init_12_6_11_8
void FASTCALL call_12_6_11_8(SH2_struct* sh)
{
  SH2movllg(sh,(11<<4)+8);
}
#define init_12_6_11_9
void FASTCALL call_12_6_11_9(SH2_struct* sh)
{
  SH2movllg(sh,(11<<4)+9);
}
#define init_12_6_11_10
void FASTCALL call_12_6_11_10(SH2_struct* sh)
{
  SH2movllg(sh,(11<<4)+10);
}
#define init_12_6_11_11
void FASTCALL call_12_6_11_11(SH2_struct* sh)
{
  SH2movllg(sh,(11<<4)+11);
}
#define init_12_6_11_12
void FASTCALL call_12_6_11_12(SH2_struct* sh)
{
  SH2movllg(sh,(11<<4)+12);
}
#define init_12_6_11_13
void FASTCALL call_12_6_11_13(SH2_struct* sh)
{
  SH2movllg(sh,(11<<4)+13);
}
#define init_12_6_11_14
void FASTCALL call_12_6_11_14(SH2_struct* sh)
{
  SH2movllg(sh,(11<<4)+14);
}
#define init_12_6_11_15
void FASTCALL call_12_6_11_15(SH2_struct* sh)
{
  SH2movllg(sh,(11<<4)+15);
}
#define init_12_6_12_0
void FASTCALL call_12_6_12_0(SH2_struct* sh)
{
  SH2movllg(sh,(12<<4)+0);
}
#define init_12_6_12_1
void FASTCALL call_12_6_12_1(SH2_struct* sh)
{
  SH2movllg(sh,(12<<4)+1);
}
#define init_12_6_12_2
void FASTCALL call_12_6_12_2(SH2_struct* sh)
{
  SH2movllg(sh,(12<<4)+2);
}
#define init_12_6_12_3
void FASTCALL call_12_6_12_3(SH2_struct* sh)
{
  SH2movllg(sh,(12<<4)+3);
}
#define init_12_6_12_4
void FASTCALL call_12_6_12_4(SH2_struct* sh)
{
  SH2movllg(sh,(12<<4)+4);
}
#define init_12_6_12_5
void FASTCALL call_12_6_12_5(SH2_struct* sh)
{
  SH2movllg(sh,(12<<4)+5);
}
#define init_12_6_12_6
void FASTCALL call_12_6_12_6(SH2_struct* sh)
{
  SH2movllg(sh,(12<<4)+6);
}
#define init_12_6_12_7
void FASTCALL call_12_6_12_7(SH2_struct* sh)
{
  SH2movllg(sh,(12<<4)+7);
}
#define init_12_6_12_8
void FASTCALL call_12_6_12_8(SH2_struct* sh)
{
  SH2movllg(sh,(12<<4)+8);
}
#define init_12_6_12_9
void FASTCALL call_12_6_12_9(SH2_struct* sh)
{
  SH2movllg(sh,(12<<4)+9);
}
#define init_12_6_12_10
void FASTCALL call_12_6_12_10(SH2_struct* sh)
{
  SH2movllg(sh,(12<<4)+10);
}
#define init_12_6_12_11
void FASTCALL call_12_6_12_11(SH2_struct* sh)
{
  SH2movllg(sh,(12<<4)+11);
}
#define init_12_6_12_12
void FASTCALL call_12_6_12_12(SH2_struct* sh)
{
  SH2movllg(sh,(12<<4)+12);
}
#define init_12_6_12_13
void FASTCALL call_12_6_12_13(SH2_struct* sh)
{
  SH2movllg(sh,(12<<4)+13);
}
#define init_12_6_12_14
void FASTCALL call_12_6_12_14(SH2_struct* sh)
{
  SH2movllg(sh,(12<<4)+14);
}
#define init_12_6_12_15
void FASTCALL call_12_6_12_15(SH2_struct* sh)
{
  SH2movllg(sh,(12<<4)+15);
}
#define init_12_6_13_0
void FASTCALL call_12_6_13_0(SH2_struct* sh)
{
  SH2movllg(sh,(13<<4)+0);
}
#define init_12_6_13_1
void FASTCALL call_12_6_13_1(SH2_struct* sh)
{
  SH2movllg(sh,(13<<4)+1);
}
#define init_12_6_13_2
void FASTCALL call_12_6_13_2(SH2_struct* sh)
{
  SH2movllg(sh,(13<<4)+2);
}
#define init_12_6_13_3
void FASTCALL call_12_6_13_3(SH2_struct* sh)
{
  SH2movllg(sh,(13<<4)+3);
}
#define init_12_6_13_4
void FASTCALL call_12_6_13_4(SH2_struct* sh)
{
  SH2movllg(sh,(13<<4)+4);
}
#define init_12_6_13_5
void FASTCALL call_12_6_13_5(SH2_struct* sh)
{
  SH2movllg(sh,(13<<4)+5);
}
#define init_12_6_13_6
void FASTCALL call_12_6_13_6(SH2_struct* sh)
{
  SH2movllg(sh,(13<<4)+6);
}
#define init_12_6_13_7
void FASTCALL call_12_6_13_7(SH2_struct* sh)
{
  SH2movllg(sh,(13<<4)+7);
}
#define init_12_6_13_8
void FASTCALL call_12_6_13_8(SH2_struct* sh)
{
  SH2movllg(sh,(13<<4)+8);
}
#define init_12_6_13_9
void FASTCALL call_12_6_13_9(SH2_struct* sh)
{
  SH2movllg(sh,(13<<4)+9);
}
#define init_12_6_13_10
void FASTCALL call_12_6_13_10(SH2_struct* sh)
{
  SH2movllg(sh,(13<<4)+10);
}
#define init_12_6_13_11
void FASTCALL call_12_6_13_11(SH2_struct* sh)
{
  SH2movllg(sh,(13<<4)+11);
}
#define init_12_6_13_12
void FASTCALL call_12_6_13_12(SH2_struct* sh)
{
  SH2movllg(sh,(13<<4)+12);
}
#define init_12_6_13_13
void FASTCALL call_12_6_13_13(SH2_struct* sh)
{
  SH2movllg(sh,(13<<4)+13);
}
#define init_12_6_13_14
void FASTCALL call_12_6_13_14(SH2_struct* sh)
{
  SH2movllg(sh,(13<<4)+14);
}
#define init_12_6_13_15
void FASTCALL call_12_6_13_15(SH2_struct* sh)
{
  SH2movllg(sh,(13<<4)+15);
}
#define init_12_6_14_0
void FASTCALL call_12_6_14_0(SH2_struct* sh)
{
  SH2movllg(sh,(14<<4)+0);
}
#define init_12_6_14_1
void FASTCALL call_12_6_14_1(SH2_struct* sh)
{
  SH2movllg(sh,(14<<4)+1);
}
#define init_12_6_14_2
void FASTCALL call_12_6_14_2(SH2_struct* sh)
{
  SH2movllg(sh,(14<<4)+2);
}
#define init_12_6_14_3
void FASTCALL call_12_6_14_3(SH2_struct* sh)
{
  SH2movllg(sh,(14<<4)+3);
}
#define init_12_6_14_4
void FASTCALL call_12_6_14_4(SH2_struct* sh)
{
  SH2movllg(sh,(14<<4)+4);
}
#define init_12_6_14_5
void FASTCALL call_12_6_14_5(SH2_struct* sh)
{
  SH2movllg(sh,(14<<4)+5);
}
#define init_12_6_14_6
void FASTCALL call_12_6_14_6(SH2_struct* sh)
{
  SH2movllg(sh,(14<<4)+6);
}
#define init_12_6_14_7
void FASTCALL call_12_6_14_7(SH2_struct* sh)
{
  SH2movllg(sh,(14<<4)+7);
}
#define init_12_6_14_8
void FASTCALL call_12_6_14_8(SH2_struct* sh)
{
  SH2movllg(sh,(14<<4)+8);
}
#define init_12_6_14_9
void FASTCALL call_12_6_14_9(SH2_struct* sh)
{
  SH2movllg(sh,(14<<4)+9);
}
#define init_12_6_14_10
void FASTCALL call_12_6_14_10(SH2_struct* sh)
{
  SH2movllg(sh,(14<<4)+10);
}
#define init_12_6_14_11
void FASTCALL call_12_6_14_11(SH2_struct* sh)
{
  SH2movllg(sh,(14<<4)+11);
}
#define init_12_6_14_12
void FASTCALL call_12_6_14_12(SH2_struct* sh)
{
  SH2movllg(sh,(14<<4)+12);
}
#define init_12_6_14_13
void FASTCALL call_12_6_14_13(SH2_struct* sh)
{
  SH2movllg(sh,(14<<4)+13);
}
#define init_12_6_14_14
void FASTCALL call_12_6_14_14(SH2_struct* sh)
{
  SH2movllg(sh,(14<<4)+14);
}
#define init_12_6_14_15
void FASTCALL call_12_6_14_15(SH2_struct* sh)
{
  SH2movllg(sh,(14<<4)+15);
}
#define init_12_6_15_0
void FASTCALL call_12_6_15_0(SH2_struct* sh)
{
  SH2movllg(sh,(15<<4)+0);
}
#define init_12_6_15_1
void FASTCALL call_12_6_15_1(SH2_struct* sh)
{
  SH2movllg(sh,(15<<4)+1);
}
#define init_12_6_15_2
void FASTCALL call_12_6_15_2(SH2_struct* sh)
{
  SH2movllg(sh,(15<<4)+2);
}
#define init_12_6_15_3
void FASTCALL call_12_6_15_3(SH2_struct* sh)
{
  SH2movllg(sh,(15<<4)+3);
}
#define init_12_6_15_4
void FASTCALL call_12_6_15_4(SH2_struct* sh)
{
  SH2movllg(sh,(15<<4)+4);
}
#define init_12_6_15_5
void FASTCALL call_12_6_15_5(SH2_struct* sh)
{
  SH2movllg(sh,(15<<4)+5);
}
#define init_12_6_15_6
void FASTCALL call_12_6_15_6(SH2_struct* sh)
{
  SH2movllg(sh,(15<<4)+6);
}
#define init_12_6_15_7
void FASTCALL call_12_6_15_7(SH2_struct* sh)
{
  SH2movllg(sh,(15<<4)+7);
}
#define init_12_6_15_8
void FASTCALL call_12_6_15_8(SH2_struct* sh)
{
  SH2movllg(sh,(15<<4)+8);
}
#define init_12_6_15_9
void FASTCALL call_12_6_15_9(SH2_struct* sh)
{
  SH2movllg(sh,(15<<4)+9);
}
#define init_12_6_15_10
void FASTCALL call_12_6_15_10(SH2_struct* sh)
{
  SH2movllg(sh,(15<<4)+10);
}
#define init_12_6_15_11
void FASTCALL call_12_6_15_11(SH2_struct* sh)
{
  SH2movllg(sh,(15<<4)+11);
}
#define init_12_6_15_12
void FASTCALL call_12_6_15_12(SH2_struct* sh)
{
  SH2movllg(sh,(15<<4)+12);
}
#define init_12_6_15_13
void FASTCALL call_12_6_15_13(SH2_struct* sh)
{
  SH2movllg(sh,(15<<4)+13);
}
#define init_12_6_15_14
void FASTCALL call_12_6_15_14(SH2_struct* sh)
{
  SH2movllg(sh,(15<<4)+14);
}
#define init_12_6_15_15
void FASTCALL call_12_6_15_15(SH2_struct* sh)
{
  SH2movllg(sh,(15<<4)+15);
}
#define init_5_0_0_0
void FASTCALL call_5_0_0_0(SH2_struct* sh)
{
  SH2movll4(sh, 0,0, 0);
}
#define init_5_0_0_1
void FASTCALL call_5_0_0_1(SH2_struct* sh)
{
  SH2movll4(sh, 0,0, 1);
}
#define init_5_0_0_2
void FASTCALL call_5_0_0_2(SH2_struct* sh)
{
  SH2movll4(sh, 0,0, 2);
}
#define init_5_0_0_3
void FASTCALL call_5_0_0_3(SH2_struct* sh)
{
  SH2movll4(sh, 0,0, 3);
}
#define init_5_0_0_4
void FASTCALL call_5_0_0_4(SH2_struct* sh)
{
  SH2movll4(sh, 0,0, 4);
}
#define init_5_0_0_5
void FASTCALL call_5_0_0_5(SH2_struct* sh)
{
  SH2movll4(sh, 0,0, 5);
}
#define init_5_0_0_6
void FASTCALL call_5_0_0_6(SH2_struct* sh)
{
  SH2movll4(sh, 0,0, 6);
}
#define init_5_0_0_7
void FASTCALL call_5_0_0_7(SH2_struct* sh)
{
  SH2movll4(sh, 0,0, 7);
}
#define init_5_0_0_8
void FASTCALL call_5_0_0_8(SH2_struct* sh)
{
  SH2movll4(sh, 0,0, 8);
}
#define init_5_0_0_9
void FASTCALL call_5_0_0_9(SH2_struct* sh)
{
  SH2movll4(sh, 0,0, 9);
}
#define init_5_0_0_10
void FASTCALL call_5_0_0_10(SH2_struct* sh)
{
  SH2movll4(sh, 0,0, 10);
}
#define init_5_0_0_11
void FASTCALL call_5_0_0_11(SH2_struct* sh)
{
  SH2movll4(sh, 0,0, 11);
}
#define init_5_0_0_12
void FASTCALL call_5_0_0_12(SH2_struct* sh)
{
  SH2movll4(sh, 0,0, 12);
}
#define init_5_0_0_13
void FASTCALL call_5_0_0_13(SH2_struct* sh)
{
  SH2movll4(sh, 0,0, 13);
}
#define init_5_0_0_14
void FASTCALL call_5_0_0_14(SH2_struct* sh)
{
  SH2movll4(sh, 0,0, 14);
}
#define init_5_0_0_15
void FASTCALL call_5_0_0_15(SH2_struct* sh)
{
  SH2movll4(sh, 0,0, 15);
}
#define init_5_0_1_0
void FASTCALL call_5_0_1_0(SH2_struct* sh)
{
  SH2movll4(sh, 0,1, 0);
}
#define init_5_0_1_1
void FASTCALL call_5_0_1_1(SH2_struct* sh)
{
  SH2movll4(sh, 0,1, 1);
}
#define init_5_0_1_2
void FASTCALL call_5_0_1_2(SH2_struct* sh)
{
  SH2movll4(sh, 0,1, 2);
}
#define init_5_0_1_3
void FASTCALL call_5_0_1_3(SH2_struct* sh)
{
  SH2movll4(sh, 0,1, 3);
}
#define init_5_0_1_4
void FASTCALL call_5_0_1_4(SH2_struct* sh)
{
  SH2movll4(sh, 0,1, 4);
}
#define init_5_0_1_5
void FASTCALL call_5_0_1_5(SH2_struct* sh)
{
  SH2movll4(sh, 0,1, 5);
}
#define init_5_0_1_6
void FASTCALL call_5_0_1_6(SH2_struct* sh)
{
  SH2movll4(sh, 0,1, 6);
}
#define init_5_0_1_7
void FASTCALL call_5_0_1_7(SH2_struct* sh)
{
  SH2movll4(sh, 0,1, 7);
}
#define init_5_0_1_8
void FASTCALL call_5_0_1_8(SH2_struct* sh)
{
  SH2movll4(sh, 0,1, 8);
}
#define init_5_0_1_9
void FASTCALL call_5_0_1_9(SH2_struct* sh)
{
  SH2movll4(sh, 0,1, 9);
}
#define init_5_0_1_10
void FASTCALL call_5_0_1_10(SH2_struct* sh)
{
  SH2movll4(sh, 0,1, 10);
}
#define init_5_0_1_11
void FASTCALL call_5_0_1_11(SH2_struct* sh)
{
  SH2movll4(sh, 0,1, 11);
}
#define init_5_0_1_12
void FASTCALL call_5_0_1_12(SH2_struct* sh)
{
  SH2movll4(sh, 0,1, 12);
}
#define init_5_0_1_13
void FASTCALL call_5_0_1_13(SH2_struct* sh)
{
  SH2movll4(sh, 0,1, 13);
}
#define init_5_0_1_14
void FASTCALL call_5_0_1_14(SH2_struct* sh)
{
  SH2movll4(sh, 0,1, 14);
}
#define init_5_0_1_15
void FASTCALL call_5_0_1_15(SH2_struct* sh)
{
  SH2movll4(sh, 0,1, 15);
}
#define init_5_0_2_0
void FASTCALL call_5_0_2_0(SH2_struct* sh)
{
  SH2movll4(sh, 0,2, 0);
}
#define init_5_0_2_1
void FASTCALL call_5_0_2_1(SH2_struct* sh)
{
  SH2movll4(sh, 0,2, 1);
}
#define init_5_0_2_2
void FASTCALL call_5_0_2_2(SH2_struct* sh)
{
  SH2movll4(sh, 0,2, 2);
}
#define init_5_0_2_3
void FASTCALL call_5_0_2_3(SH2_struct* sh)
{
  SH2movll4(sh, 0,2, 3);
}
#define init_5_0_2_4
void FASTCALL call_5_0_2_4(SH2_struct* sh)
{
  SH2movll4(sh, 0,2, 4);
}
#define init_5_0_2_5
void FASTCALL call_5_0_2_5(SH2_struct* sh)
{
  SH2movll4(sh, 0,2, 5);
}
#define init_5_0_2_6
void FASTCALL call_5_0_2_6(SH2_struct* sh)
{
  SH2movll4(sh, 0,2, 6);
}
#define init_5_0_2_7
void FASTCALL call_5_0_2_7(SH2_struct* sh)
{
  SH2movll4(sh, 0,2, 7);
}
#define init_5_0_2_8
void FASTCALL call_5_0_2_8(SH2_struct* sh)
{
  SH2movll4(sh, 0,2, 8);
}
#define init_5_0_2_9
void FASTCALL call_5_0_2_9(SH2_struct* sh)
{
  SH2movll4(sh, 0,2, 9);
}
#define init_5_0_2_10
void FASTCALL call_5_0_2_10(SH2_struct* sh)
{
  SH2movll4(sh, 0,2, 10);
}
#define init_5_0_2_11
void FASTCALL call_5_0_2_11(SH2_struct* sh)
{
  SH2movll4(sh, 0,2, 11);
}
#define init_5_0_2_12
void FASTCALL call_5_0_2_12(SH2_struct* sh)
{
  SH2movll4(sh, 0,2, 12);
}
#define init_5_0_2_13
void FASTCALL call_5_0_2_13(SH2_struct* sh)
{
  SH2movll4(sh, 0,2, 13);
}
#define init_5_0_2_14
void FASTCALL call_5_0_2_14(SH2_struct* sh)
{
  SH2movll4(sh, 0,2, 14);
}
#define init_5_0_2_15
void FASTCALL call_5_0_2_15(SH2_struct* sh)
{
  SH2movll4(sh, 0,2, 15);
}
#define init_5_0_3_0
void FASTCALL call_5_0_3_0(SH2_struct* sh)
{
  SH2movll4(sh, 0,3, 0);
}
#define init_5_0_3_1
void FASTCALL call_5_0_3_1(SH2_struct* sh)
{
  SH2movll4(sh, 0,3, 1);
}
#define init_5_0_3_2
void FASTCALL call_5_0_3_2(SH2_struct* sh)
{
  SH2movll4(sh, 0,3, 2);
}
#define init_5_0_3_3
void FASTCALL call_5_0_3_3(SH2_struct* sh)
{
  SH2movll4(sh, 0,3, 3);
}
#define init_5_0_3_4
void FASTCALL call_5_0_3_4(SH2_struct* sh)
{
  SH2movll4(sh, 0,3, 4);
}
#define init_5_0_3_5
void FASTCALL call_5_0_3_5(SH2_struct* sh)
{
  SH2movll4(sh, 0,3, 5);
}
#define init_5_0_3_6
void FASTCALL call_5_0_3_6(SH2_struct* sh)
{
  SH2movll4(sh, 0,3, 6);
}
#define init_5_0_3_7
void FASTCALL call_5_0_3_7(SH2_struct* sh)
{
  SH2movll4(sh, 0,3, 7);
}
#define init_5_0_3_8
void FASTCALL call_5_0_3_8(SH2_struct* sh)
{
  SH2movll4(sh, 0,3, 8);
}
#define init_5_0_3_9
void FASTCALL call_5_0_3_9(SH2_struct* sh)
{
  SH2movll4(sh, 0,3, 9);
}
#define init_5_0_3_10
void FASTCALL call_5_0_3_10(SH2_struct* sh)
{
  SH2movll4(sh, 0,3, 10);
}
#define init_5_0_3_11
void FASTCALL call_5_0_3_11(SH2_struct* sh)
{
  SH2movll4(sh, 0,3, 11);
}
#define init_5_0_3_12
void FASTCALL call_5_0_3_12(SH2_struct* sh)
{
  SH2movll4(sh, 0,3, 12);
}
#define init_5_0_3_13
void FASTCALL call_5_0_3_13(SH2_struct* sh)
{
  SH2movll4(sh, 0,3, 13);
}
#define init_5_0_3_14
void FASTCALL call_5_0_3_14(SH2_struct* sh)
{
  SH2movll4(sh, 0,3, 14);
}
#define init_5_0_3_15
void FASTCALL call_5_0_3_15(SH2_struct* sh)
{
  SH2movll4(sh, 0,3, 15);
}
#define init_5_0_4_0
void FASTCALL call_5_0_4_0(SH2_struct* sh)
{
  SH2movll4(sh, 0,4, 0);
}
#define init_5_0_4_1
void FASTCALL call_5_0_4_1(SH2_struct* sh)
{
  SH2movll4(sh, 0,4, 1);
}
#define init_5_0_4_2
void FASTCALL call_5_0_4_2(SH2_struct* sh)
{
  SH2movll4(sh, 0,4, 2);
}
#define init_5_0_4_3
void FASTCALL call_5_0_4_3(SH2_struct* sh)
{
  SH2movll4(sh, 0,4, 3);
}
#define init_5_0_4_4
void FASTCALL call_5_0_4_4(SH2_struct* sh)
{
  SH2movll4(sh, 0,4, 4);
}
#define init_5_0_4_5
void FASTCALL call_5_0_4_5(SH2_struct* sh)
{
  SH2movll4(sh, 0,4, 5);
}
#define init_5_0_4_6
void FASTCALL call_5_0_4_6(SH2_struct* sh)
{
  SH2movll4(sh, 0,4, 6);
}
#define init_5_0_4_7
void FASTCALL call_5_0_4_7(SH2_struct* sh)
{
  SH2movll4(sh, 0,4, 7);
}
#define init_5_0_4_8
void FASTCALL call_5_0_4_8(SH2_struct* sh)
{
  SH2movll4(sh, 0,4, 8);
}
#define init_5_0_4_9
void FASTCALL call_5_0_4_9(SH2_struct* sh)
{
  SH2movll4(sh, 0,4, 9);
}
#define init_5_0_4_10
void FASTCALL call_5_0_4_10(SH2_struct* sh)
{
  SH2movll4(sh, 0,4, 10);
}
#define init_5_0_4_11
void FASTCALL call_5_0_4_11(SH2_struct* sh)
{
  SH2movll4(sh, 0,4, 11);
}
#define init_5_0_4_12
void FASTCALL call_5_0_4_12(SH2_struct* sh)
{
  SH2movll4(sh, 0,4, 12);
}
#define init_5_0_4_13
void FASTCALL call_5_0_4_13(SH2_struct* sh)
{
  SH2movll4(sh, 0,4, 13);
}
#define init_5_0_4_14
void FASTCALL call_5_0_4_14(SH2_struct* sh)
{
  SH2movll4(sh, 0,4, 14);
}
#define init_5_0_4_15
void FASTCALL call_5_0_4_15(SH2_struct* sh)
{
  SH2movll4(sh, 0,4, 15);
}
#define init_5_0_5_0
void FASTCALL call_5_0_5_0(SH2_struct* sh)
{
  SH2movll4(sh, 0,5, 0);
}
#define init_5_0_5_1
void FASTCALL call_5_0_5_1(SH2_struct* sh)
{
  SH2movll4(sh, 0,5, 1);
}
#define init_5_0_5_2
void FASTCALL call_5_0_5_2(SH2_struct* sh)
{
  SH2movll4(sh, 0,5, 2);
}
#define init_5_0_5_3
void FASTCALL call_5_0_5_3(SH2_struct* sh)
{
  SH2movll4(sh, 0,5, 3);
}
#define init_5_0_5_4
void FASTCALL call_5_0_5_4(SH2_struct* sh)
{
  SH2movll4(sh, 0,5, 4);
}
#define init_5_0_5_5
void FASTCALL call_5_0_5_5(SH2_struct* sh)
{
  SH2movll4(sh, 0,5, 5);
}
#define init_5_0_5_6
void FASTCALL call_5_0_5_6(SH2_struct* sh)
{
  SH2movll4(sh, 0,5, 6);
}
#define init_5_0_5_7
void FASTCALL call_5_0_5_7(SH2_struct* sh)
{
  SH2movll4(sh, 0,5, 7);
}
#define init_5_0_5_8
void FASTCALL call_5_0_5_8(SH2_struct* sh)
{
  SH2movll4(sh, 0,5, 8);
}
#define init_5_0_5_9
void FASTCALL call_5_0_5_9(SH2_struct* sh)
{
  SH2movll4(sh, 0,5, 9);
}
#define init_5_0_5_10
void FASTCALL call_5_0_5_10(SH2_struct* sh)
{
  SH2movll4(sh, 0,5, 10);
}
#define init_5_0_5_11
void FASTCALL call_5_0_5_11(SH2_struct* sh)
{
  SH2movll4(sh, 0,5, 11);
}
#define init_5_0_5_12
void FASTCALL call_5_0_5_12(SH2_struct* sh)
{
  SH2movll4(sh, 0,5, 12);
}
#define init_5_0_5_13
void FASTCALL call_5_0_5_13(SH2_struct* sh)
{
  SH2movll4(sh, 0,5, 13);
}
#define init_5_0_5_14
void FASTCALL call_5_0_5_14(SH2_struct* sh)
{
  SH2movll4(sh, 0,5, 14);
}
#define init_5_0_5_15
void FASTCALL call_5_0_5_15(SH2_struct* sh)
{
  SH2movll4(sh, 0,5, 15);
}
#define init_5_0_6_0
void FASTCALL call_5_0_6_0(SH2_struct* sh)
{
  SH2movll4(sh, 0,6, 0);
}
#define init_5_0_6_1
void FASTCALL call_5_0_6_1(SH2_struct* sh)
{
  SH2movll4(sh, 0,6, 1);
}
#define init_5_0_6_2
void FASTCALL call_5_0_6_2(SH2_struct* sh)
{
  SH2movll4(sh, 0,6, 2);
}
#define init_5_0_6_3
void FASTCALL call_5_0_6_3(SH2_struct* sh)
{
  SH2movll4(sh, 0,6, 3);
}
#define init_5_0_6_4
void FASTCALL call_5_0_6_4(SH2_struct* sh)
{
  SH2movll4(sh, 0,6, 4);
}
#define init_5_0_6_5
void FASTCALL call_5_0_6_5(SH2_struct* sh)
{
  SH2movll4(sh, 0,6, 5);
}
#define init_5_0_6_6
void FASTCALL call_5_0_6_6(SH2_struct* sh)
{
  SH2movll4(sh, 0,6, 6);
}
#define init_5_0_6_7
void FASTCALL call_5_0_6_7(SH2_struct* sh)
{
  SH2movll4(sh, 0,6, 7);
}
#define init_5_0_6_8
void FASTCALL call_5_0_6_8(SH2_struct* sh)
{
  SH2movll4(sh, 0,6, 8);
}
#define init_5_0_6_9
void FASTCALL call_5_0_6_9(SH2_struct* sh)
{
  SH2movll4(sh, 0,6, 9);
}
#define init_5_0_6_10
void FASTCALL call_5_0_6_10(SH2_struct* sh)
{
  SH2movll4(sh, 0,6, 10);
}
#define init_5_0_6_11
void FASTCALL call_5_0_6_11(SH2_struct* sh)
{
  SH2movll4(sh, 0,6, 11);
}
#define init_5_0_6_12
void FASTCALL call_5_0_6_12(SH2_struct* sh)
{
  SH2movll4(sh, 0,6, 12);
}
#define init_5_0_6_13
void FASTCALL call_5_0_6_13(SH2_struct* sh)
{
  SH2movll4(sh, 0,6, 13);
}
#define init_5_0_6_14
void FASTCALL call_5_0_6_14(SH2_struct* sh)
{
  SH2movll4(sh, 0,6, 14);
}
#define init_5_0_6_15
void FASTCALL call_5_0_6_15(SH2_struct* sh)
{
  SH2movll4(sh, 0,6, 15);
}
#define init_5_0_7_0
void FASTCALL call_5_0_7_0(SH2_struct* sh)
{
  SH2movll4(sh, 0,7, 0);
}
#define init_5_0_7_1
void FASTCALL call_5_0_7_1(SH2_struct* sh)
{
  SH2movll4(sh, 0,7, 1);
}
#define init_5_0_7_2
void FASTCALL call_5_0_7_2(SH2_struct* sh)
{
  SH2movll4(sh, 0,7, 2);
}
#define init_5_0_7_3
void FASTCALL call_5_0_7_3(SH2_struct* sh)
{
  SH2movll4(sh, 0,7, 3);
}
#define init_5_0_7_4
void FASTCALL call_5_0_7_4(SH2_struct* sh)
{
  SH2movll4(sh, 0,7, 4);
}
#define init_5_0_7_5
void FASTCALL call_5_0_7_5(SH2_struct* sh)
{
  SH2movll4(sh, 0,7, 5);
}
#define init_5_0_7_6
void FASTCALL call_5_0_7_6(SH2_struct* sh)
{
  SH2movll4(sh, 0,7, 6);
}
#define init_5_0_7_7
void FASTCALL call_5_0_7_7(SH2_struct* sh)
{
  SH2movll4(sh, 0,7, 7);
}
#define init_5_0_7_8
void FASTCALL call_5_0_7_8(SH2_struct* sh)
{
  SH2movll4(sh, 0,7, 8);
}
#define init_5_0_7_9
void FASTCALL call_5_0_7_9(SH2_struct* sh)
{
  SH2movll4(sh, 0,7, 9);
}
#define init_5_0_7_10
void FASTCALL call_5_0_7_10(SH2_struct* sh)
{
  SH2movll4(sh, 0,7, 10);
}
#define init_5_0_7_11
void FASTCALL call_5_0_7_11(SH2_struct* sh)
{
  SH2movll4(sh, 0,7, 11);
}
#define init_5_0_7_12
void FASTCALL call_5_0_7_12(SH2_struct* sh)
{
  SH2movll4(sh, 0,7, 12);
}
#define init_5_0_7_13
void FASTCALL call_5_0_7_13(SH2_struct* sh)
{
  SH2movll4(sh, 0,7, 13);
}
#define init_5_0_7_14
void FASTCALL call_5_0_7_14(SH2_struct* sh)
{
  SH2movll4(sh, 0,7, 14);
}
#define init_5_0_7_15
void FASTCALL call_5_0_7_15(SH2_struct* sh)
{
  SH2movll4(sh, 0,7, 15);
}
#define init_5_0_8_0
void FASTCALL call_5_0_8_0(SH2_struct* sh)
{
  SH2movll4(sh, 0,8, 0);
}
#define init_5_0_8_1
void FASTCALL call_5_0_8_1(SH2_struct* sh)
{
  SH2movll4(sh, 0,8, 1);
}
#define init_5_0_8_2
void FASTCALL call_5_0_8_2(SH2_struct* sh)
{
  SH2movll4(sh, 0,8, 2);
}
#define init_5_0_8_3
void FASTCALL call_5_0_8_3(SH2_struct* sh)
{
  SH2movll4(sh, 0,8, 3);
}
#define init_5_0_8_4
void FASTCALL call_5_0_8_4(SH2_struct* sh)
{
  SH2movll4(sh, 0,8, 4);
}
#define init_5_0_8_5
void FASTCALL call_5_0_8_5(SH2_struct* sh)
{
  SH2movll4(sh, 0,8, 5);
}
#define init_5_0_8_6
void FASTCALL call_5_0_8_6(SH2_struct* sh)
{
  SH2movll4(sh, 0,8, 6);
}
#define init_5_0_8_7
void FASTCALL call_5_0_8_7(SH2_struct* sh)
{
  SH2movll4(sh, 0,8, 7);
}
#define init_5_0_8_8
void FASTCALL call_5_0_8_8(SH2_struct* sh)
{
  SH2movll4(sh, 0,8, 8);
}
#define init_5_0_8_9
void FASTCALL call_5_0_8_9(SH2_struct* sh)
{
  SH2movll4(sh, 0,8, 9);
}
#define init_5_0_8_10
void FASTCALL call_5_0_8_10(SH2_struct* sh)
{
  SH2movll4(sh, 0,8, 10);
}
#define init_5_0_8_11
void FASTCALL call_5_0_8_11(SH2_struct* sh)
{
  SH2movll4(sh, 0,8, 11);
}
#define init_5_0_8_12
void FASTCALL call_5_0_8_12(SH2_struct* sh)
{
  SH2movll4(sh, 0,8, 12);
}
#define init_5_0_8_13
void FASTCALL call_5_0_8_13(SH2_struct* sh)
{
  SH2movll4(sh, 0,8, 13);
}
#define init_5_0_8_14
void FASTCALL call_5_0_8_14(SH2_struct* sh)
{
  SH2movll4(sh, 0,8, 14);
}
#define init_5_0_8_15
void FASTCALL call_5_0_8_15(SH2_struct* sh)
{
  SH2movll4(sh, 0,8, 15);
}
#define init_5_0_9_0
void FASTCALL call_5_0_9_0(SH2_struct* sh)
{
  SH2movll4(sh, 0,9, 0);
}
#define init_5_0_9_1
void FASTCALL call_5_0_9_1(SH2_struct* sh)
{
  SH2movll4(sh, 0,9, 1);
}
#define init_5_0_9_2
void FASTCALL call_5_0_9_2(SH2_struct* sh)
{
  SH2movll4(sh, 0,9, 2);
}
#define init_5_0_9_3
void FASTCALL call_5_0_9_3(SH2_struct* sh)
{
  SH2movll4(sh, 0,9, 3);
}
#define init_5_0_9_4
void FASTCALL call_5_0_9_4(SH2_struct* sh)
{
  SH2movll4(sh, 0,9, 4);
}
#define init_5_0_9_5
void FASTCALL call_5_0_9_5(SH2_struct* sh)
{
  SH2movll4(sh, 0,9, 5);
}
#define init_5_0_9_6
void FASTCALL call_5_0_9_6(SH2_struct* sh)
{
  SH2movll4(sh, 0,9, 6);
}
#define init_5_0_9_7
void FASTCALL call_5_0_9_7(SH2_struct* sh)
{
  SH2movll4(sh, 0,9, 7);
}
#define init_5_0_9_8
void FASTCALL call_5_0_9_8(SH2_struct* sh)
{
  SH2movll4(sh, 0,9, 8);
}
#define init_5_0_9_9
void FASTCALL call_5_0_9_9(SH2_struct* sh)
{
  SH2movll4(sh, 0,9, 9);
}
#define init_5_0_9_10
void FASTCALL call_5_0_9_10(SH2_struct* sh)
{
  SH2movll4(sh, 0,9, 10);
}
#define init_5_0_9_11
void FASTCALL call_5_0_9_11(SH2_struct* sh)
{
  SH2movll4(sh, 0,9, 11);
}
#define init_5_0_9_12
void FASTCALL call_5_0_9_12(SH2_struct* sh)
{
  SH2movll4(sh, 0,9, 12);
}
#define init_5_0_9_13
void FASTCALL call_5_0_9_13(SH2_struct* sh)
{
  SH2movll4(sh, 0,9, 13);
}
#define init_5_0_9_14
void FASTCALL call_5_0_9_14(SH2_struct* sh)
{
  SH2movll4(sh, 0,9, 14);
}
#define init_5_0_9_15
void FASTCALL call_5_0_9_15(SH2_struct* sh)
{
  SH2movll4(sh, 0,9, 15);
}
#define init_5_0_10_0
void FASTCALL call_5_0_10_0(SH2_struct* sh)
{
  SH2movll4(sh, 0,10, 0);
}
#define init_5_0_10_1
void FASTCALL call_5_0_10_1(SH2_struct* sh)
{
  SH2movll4(sh, 0,10, 1);
}
#define init_5_0_10_2
void FASTCALL call_5_0_10_2(SH2_struct* sh)
{
  SH2movll4(sh, 0,10, 2);
}
#define init_5_0_10_3
void FASTCALL call_5_0_10_3(SH2_struct* sh)
{
  SH2movll4(sh, 0,10, 3);
}
#define init_5_0_10_4
void FASTCALL call_5_0_10_4(SH2_struct* sh)
{
  SH2movll4(sh, 0,10, 4);
}
#define init_5_0_10_5
void FASTCALL call_5_0_10_5(SH2_struct* sh)
{
  SH2movll4(sh, 0,10, 5);
}
#define init_5_0_10_6
void FASTCALL call_5_0_10_6(SH2_struct* sh)
{
  SH2movll4(sh, 0,10, 6);
}
#define init_5_0_10_7
void FASTCALL call_5_0_10_7(SH2_struct* sh)
{
  SH2movll4(sh, 0,10, 7);
}
#define init_5_0_10_8
void FASTCALL call_5_0_10_8(SH2_struct* sh)
{
  SH2movll4(sh, 0,10, 8);
}
#define init_5_0_10_9
void FASTCALL call_5_0_10_9(SH2_struct* sh)
{
  SH2movll4(sh, 0,10, 9);
}
#define init_5_0_10_10
void FASTCALL call_5_0_10_10(SH2_struct* sh)
{
  SH2movll4(sh, 0,10, 10);
}
#define init_5_0_10_11
void FASTCALL call_5_0_10_11(SH2_struct* sh)
{
  SH2movll4(sh, 0,10, 11);
}
#define init_5_0_10_12
void FASTCALL call_5_0_10_12(SH2_struct* sh)
{
  SH2movll4(sh, 0,10, 12);
}
#define init_5_0_10_13
void FASTCALL call_5_0_10_13(SH2_struct* sh)
{
  SH2movll4(sh, 0,10, 13);
}
#define init_5_0_10_14
void FASTCALL call_5_0_10_14(SH2_struct* sh)
{
  SH2movll4(sh, 0,10, 14);
}
#define init_5_0_10_15
void FASTCALL call_5_0_10_15(SH2_struct* sh)
{
  SH2movll4(sh, 0,10, 15);
}
#define init_5_0_11_0
void FASTCALL call_5_0_11_0(SH2_struct* sh)
{
  SH2movll4(sh, 0,11, 0);
}
#define init_5_0_11_1
void FASTCALL call_5_0_11_1(SH2_struct* sh)
{
  SH2movll4(sh, 0,11, 1);
}
#define init_5_0_11_2
void FASTCALL call_5_0_11_2(SH2_struct* sh)
{
  SH2movll4(sh, 0,11, 2);
}
#define init_5_0_11_3
void FASTCALL call_5_0_11_3(SH2_struct* sh)
{
  SH2movll4(sh, 0,11, 3);
}
#define init_5_0_11_4
void FASTCALL call_5_0_11_4(SH2_struct* sh)
{
  SH2movll4(sh, 0,11, 4);
}
#define init_5_0_11_5
void FASTCALL call_5_0_11_5(SH2_struct* sh)
{
  SH2movll4(sh, 0,11, 5);
}
#define init_5_0_11_6
void FASTCALL call_5_0_11_6(SH2_struct* sh)
{
  SH2movll4(sh, 0,11, 6);
}
#define init_5_0_11_7
void FASTCALL call_5_0_11_7(SH2_struct* sh)
{
  SH2movll4(sh, 0,11, 7);
}
#define init_5_0_11_8
void FASTCALL call_5_0_11_8(SH2_struct* sh)
{
  SH2movll4(sh, 0,11, 8);
}
#define init_5_0_11_9
void FASTCALL call_5_0_11_9(SH2_struct* sh)
{
  SH2movll4(sh, 0,11, 9);
}
#define init_5_0_11_10
void FASTCALL call_5_0_11_10(SH2_struct* sh)
{
  SH2movll4(sh, 0,11, 10);
}
#define init_5_0_11_11
void FASTCALL call_5_0_11_11(SH2_struct* sh)
{
  SH2movll4(sh, 0,11, 11);
}
#define init_5_0_11_12
void FASTCALL call_5_0_11_12(SH2_struct* sh)
{
  SH2movll4(sh, 0,11, 12);
}
#define init_5_0_11_13
void FASTCALL call_5_0_11_13(SH2_struct* sh)
{
  SH2movll4(sh, 0,11, 13);
}
#define init_5_0_11_14
void FASTCALL call_5_0_11_14(SH2_struct* sh)
{
  SH2movll4(sh, 0,11, 14);
}
#define init_5_0_11_15
void FASTCALL call_5_0_11_15(SH2_struct* sh)
{
  SH2movll4(sh, 0,11, 15);
}
#define init_5_0_12_0
void FASTCALL call_5_0_12_0(SH2_struct* sh)
{
  SH2movll4(sh, 0,12, 0);
}
#define init_5_0_12_1
void FASTCALL call_5_0_12_1(SH2_struct* sh)
{
  SH2movll4(sh, 0,12, 1);
}
#define init_5_0_12_2
void FASTCALL call_5_0_12_2(SH2_struct* sh)
{
  SH2movll4(sh, 0,12, 2);
}
#define init_5_0_12_3
void FASTCALL call_5_0_12_3(SH2_struct* sh)
{
  SH2movll4(sh, 0,12, 3);
}
#define init_5_0_12_4
void FASTCALL call_5_0_12_4(SH2_struct* sh)
{
  SH2movll4(sh, 0,12, 4);
}
#define init_5_0_12_5
void FASTCALL call_5_0_12_5(SH2_struct* sh)
{
  SH2movll4(sh, 0,12, 5);
}
#define init_5_0_12_6
void FASTCALL call_5_0_12_6(SH2_struct* sh)
{
  SH2movll4(sh, 0,12, 6);
}
#define init_5_0_12_7
void FASTCALL call_5_0_12_7(SH2_struct* sh)
{
  SH2movll4(sh, 0,12, 7);
}
#define init_5_0_12_8
void FASTCALL call_5_0_12_8(SH2_struct* sh)
{
  SH2movll4(sh, 0,12, 8);
}
#define init_5_0_12_9
void FASTCALL call_5_0_12_9(SH2_struct* sh)
{
  SH2movll4(sh, 0,12, 9);
}
#define init_5_0_12_10
void FASTCALL call_5_0_12_10(SH2_struct* sh)
{
  SH2movll4(sh, 0,12, 10);
}
#define init_5_0_12_11
void FASTCALL call_5_0_12_11(SH2_struct* sh)
{
  SH2movll4(sh, 0,12, 11);
}
#define init_5_0_12_12
void FASTCALL call_5_0_12_12(SH2_struct* sh)
{
  SH2movll4(sh, 0,12, 12);
}
#define init_5_0_12_13
void FASTCALL call_5_0_12_13(SH2_struct* sh)
{
  SH2movll4(sh, 0,12, 13);
}
#define init_5_0_12_14
void FASTCALL call_5_0_12_14(SH2_struct* sh)
{
  SH2movll4(sh, 0,12, 14);
}
#define init_5_0_12_15
void FASTCALL call_5_0_12_15(SH2_struct* sh)
{
  SH2movll4(sh, 0,12, 15);
}
#define init_5_0_13_0
void FASTCALL call_5_0_13_0(SH2_struct* sh)
{
  SH2movll4(sh, 0,13, 0);
}
#define init_5_0_13_1
void FASTCALL call_5_0_13_1(SH2_struct* sh)
{
  SH2movll4(sh, 0,13, 1);
}
#define init_5_0_13_2
void FASTCALL call_5_0_13_2(SH2_struct* sh)
{
  SH2movll4(sh, 0,13, 2);
}
#define init_5_0_13_3
void FASTCALL call_5_0_13_3(SH2_struct* sh)
{
  SH2movll4(sh, 0,13, 3);
}
#define init_5_0_13_4
void FASTCALL call_5_0_13_4(SH2_struct* sh)
{
  SH2movll4(sh, 0,13, 4);
}
#define init_5_0_13_5
void FASTCALL call_5_0_13_5(SH2_struct* sh)
{
  SH2movll4(sh, 0,13, 5);
}
#define init_5_0_13_6
void FASTCALL call_5_0_13_6(SH2_struct* sh)
{
  SH2movll4(sh, 0,13, 6);
}
#define init_5_0_13_7
void FASTCALL call_5_0_13_7(SH2_struct* sh)
{
  SH2movll4(sh, 0,13, 7);
}
#define init_5_0_13_8
void FASTCALL call_5_0_13_8(SH2_struct* sh)
{
  SH2movll4(sh, 0,13, 8);
}
#define init_5_0_13_9
void FASTCALL call_5_0_13_9(SH2_struct* sh)
{
  SH2movll4(sh, 0,13, 9);
}
#define init_5_0_13_10
void FASTCALL call_5_0_13_10(SH2_struct* sh)
{
  SH2movll4(sh, 0,13, 10);
}
#define init_5_0_13_11
void FASTCALL call_5_0_13_11(SH2_struct* sh)
{
  SH2movll4(sh, 0,13, 11);
}
#define init_5_0_13_12
void FASTCALL call_5_0_13_12(SH2_struct* sh)
{
  SH2movll4(sh, 0,13, 12);
}
#define init_5_0_13_13
void FASTCALL call_5_0_13_13(SH2_struct* sh)
{
  SH2movll4(sh, 0,13, 13);
}
#define init_5_0_13_14
void FASTCALL call_5_0_13_14(SH2_struct* sh)
{
  SH2movll4(sh, 0,13, 14);
}
#define init_5_0_13_15
void FASTCALL call_5_0_13_15(SH2_struct* sh)
{
  SH2movll4(sh, 0,13, 15);
}
#define init_5_0_14_0
void FASTCALL call_5_0_14_0(SH2_struct* sh)
{
  SH2movll4(sh, 0,14, 0);
}
#define init_5_0_14_1
void FASTCALL call_5_0_14_1(SH2_struct* sh)
{
  SH2movll4(sh, 0,14, 1);
}
#define init_5_0_14_2
void FASTCALL call_5_0_14_2(SH2_struct* sh)
{
  SH2movll4(sh, 0,14, 2);
}
#define init_5_0_14_3
void FASTCALL call_5_0_14_3(SH2_struct* sh)
{
  SH2movll4(sh, 0,14, 3);
}
#define init_5_0_14_4
void FASTCALL call_5_0_14_4(SH2_struct* sh)
{
  SH2movll4(sh, 0,14, 4);
}
#define init_5_0_14_5
void FASTCALL call_5_0_14_5(SH2_struct* sh)
{
  SH2movll4(sh, 0,14, 5);
}
#define init_5_0_14_6
void FASTCALL call_5_0_14_6(SH2_struct* sh)
{
  SH2movll4(sh, 0,14, 6);
}
#define init_5_0_14_7
void FASTCALL call_5_0_14_7(SH2_struct* sh)
{
  SH2movll4(sh, 0,14, 7);
}
#define init_5_0_14_8
void FASTCALL call_5_0_14_8(SH2_struct* sh)
{
  SH2movll4(sh, 0,14, 8);
}
#define init_5_0_14_9
void FASTCALL call_5_0_14_9(SH2_struct* sh)
{
  SH2movll4(sh, 0,14, 9);
}
#define init_5_0_14_10
void FASTCALL call_5_0_14_10(SH2_struct* sh)
{
  SH2movll4(sh, 0,14, 10);
}
#define init_5_0_14_11
void FASTCALL call_5_0_14_11(SH2_struct* sh)
{
  SH2movll4(sh, 0,14, 11);
}
#define init_5_0_14_12
void FASTCALL call_5_0_14_12(SH2_struct* sh)
{
  SH2movll4(sh, 0,14, 12);
}
#define init_5_0_14_13
void FASTCALL call_5_0_14_13(SH2_struct* sh)
{
  SH2movll4(sh, 0,14, 13);
}
#define init_5_0_14_14
void FASTCALL call_5_0_14_14(SH2_struct* sh)
{
  SH2movll4(sh, 0,14, 14);
}
#define init_5_0_14_15
void FASTCALL call_5_0_14_15(SH2_struct* sh)
{
  SH2movll4(sh, 0,14, 15);
}
#define init_5_0_15_0
void FASTCALL call_5_0_15_0(SH2_struct* sh)
{
  SH2movll4(sh, 0,15, 0);
}
#define init_5_0_15_1
void FASTCALL call_5_0_15_1(SH2_struct* sh)
{
  SH2movll4(sh, 0,15, 1);
}
#define init_5_0_15_2
void FASTCALL call_5_0_15_2(SH2_struct* sh)
{
  SH2movll4(sh, 0,15, 2);
}
#define init_5_0_15_3
void FASTCALL call_5_0_15_3(SH2_struct* sh)
{
  SH2movll4(sh, 0,15, 3);
}
#define init_5_0_15_4
void FASTCALL call_5_0_15_4(SH2_struct* sh)
{
  SH2movll4(sh, 0,15, 4);
}
#define init_5_0_15_5
void FASTCALL call_5_0_15_5(SH2_struct* sh)
{
  SH2movll4(sh, 0,15, 5);
}
#define init_5_0_15_6
void FASTCALL call_5_0_15_6(SH2_struct* sh)
{
  SH2movll4(sh, 0,15, 6);
}
#define init_5_0_15_7
void FASTCALL call_5_0_15_7(SH2_struct* sh)
{
  SH2movll4(sh, 0,15, 7);
}
#define init_5_0_15_8
void FASTCALL call_5_0_15_8(SH2_struct* sh)
{
  SH2movll4(sh, 0,15, 8);
}
#define init_5_0_15_9
void FASTCALL call_5_0_15_9(SH2_struct* sh)
{
  SH2movll4(sh, 0,15, 9);
}
#define init_5_0_15_10
void FASTCALL call_5_0_15_10(SH2_struct* sh)
{
  SH2movll4(sh, 0,15, 10);
}
#define init_5_0_15_11
void FASTCALL call_5_0_15_11(SH2_struct* sh)
{
  SH2movll4(sh, 0,15, 11);
}
#define init_5_0_15_12
void FASTCALL call_5_0_15_12(SH2_struct* sh)
{
  SH2movll4(sh, 0,15, 12);
}
#define init_5_0_15_13
void FASTCALL call_5_0_15_13(SH2_struct* sh)
{
  SH2movll4(sh, 0,15, 13);
}
#define init_5_0_15_14
void FASTCALL call_5_0_15_14(SH2_struct* sh)
{
  SH2movll4(sh, 0,15, 14);
}
#define init_5_0_15_15
void FASTCALL call_5_0_15_15(SH2_struct* sh)
{
  SH2movll4(sh, 0,15, 15);
}
#define init_5_1_0_0
void FASTCALL call_5_1_0_0(SH2_struct* sh)
{
  SH2movll4(sh, 1,0, 0);
}
#define init_5_1_0_1
void FASTCALL call_5_1_0_1(SH2_struct* sh)
{
  SH2movll4(sh, 1,0, 1);
}
#define init_5_1_0_2
void FASTCALL call_5_1_0_2(SH2_struct* sh)
{
  SH2movll4(sh, 1,0, 2);
}
#define init_5_1_0_3
void FASTCALL call_5_1_0_3(SH2_struct* sh)
{
  SH2movll4(sh, 1,0, 3);
}
#define init_5_1_0_4
void FASTCALL call_5_1_0_4(SH2_struct* sh)
{
  SH2movll4(sh, 1,0, 4);
}
#define init_5_1_0_5
void FASTCALL call_5_1_0_5(SH2_struct* sh)
{
  SH2movll4(sh, 1,0, 5);
}
#define init_5_1_0_6
void FASTCALL call_5_1_0_6(SH2_struct* sh)
{
  SH2movll4(sh, 1,0, 6);
}
#define init_5_1_0_7
void FASTCALL call_5_1_0_7(SH2_struct* sh)
{
  SH2movll4(sh, 1,0, 7);
}
#define init_5_1_0_8
void FASTCALL call_5_1_0_8(SH2_struct* sh)
{
  SH2movll4(sh, 1,0, 8);
}
#define init_5_1_0_9
void FASTCALL call_5_1_0_9(SH2_struct* sh)
{
  SH2movll4(sh, 1,0, 9);
}
#define init_5_1_0_10
void FASTCALL call_5_1_0_10(SH2_struct* sh)
{
  SH2movll4(sh, 1,0, 10);
}
#define init_5_1_0_11
void FASTCALL call_5_1_0_11(SH2_struct* sh)
{
  SH2movll4(sh, 1,0, 11);
}
#define init_5_1_0_12
void FASTCALL call_5_1_0_12(SH2_struct* sh)
{
  SH2movll4(sh, 1,0, 12);
}
#define init_5_1_0_13
void FASTCALL call_5_1_0_13(SH2_struct* sh)
{
  SH2movll4(sh, 1,0, 13);
}
#define init_5_1_0_14
void FASTCALL call_5_1_0_14(SH2_struct* sh)
{
  SH2movll4(sh, 1,0, 14);
}
#define init_5_1_0_15
void FASTCALL call_5_1_0_15(SH2_struct* sh)
{
  SH2movll4(sh, 1,0, 15);
}
#define init_5_1_1_0
void FASTCALL call_5_1_1_0(SH2_struct* sh)
{
  SH2movll4(sh, 1,1, 0);
}
#define init_5_1_1_1
void FASTCALL call_5_1_1_1(SH2_struct* sh)
{
  SH2movll4(sh, 1,1, 1);
}
#define init_5_1_1_2
void FASTCALL call_5_1_1_2(SH2_struct* sh)
{
  SH2movll4(sh, 1,1, 2);
}
#define init_5_1_1_3
void FASTCALL call_5_1_1_3(SH2_struct* sh)
{
  SH2movll4(sh, 1,1, 3);
}
#define init_5_1_1_4
void FASTCALL call_5_1_1_4(SH2_struct* sh)
{
  SH2movll4(sh, 1,1, 4);
}
#define init_5_1_1_5
void FASTCALL call_5_1_1_5(SH2_struct* sh)
{
  SH2movll4(sh, 1,1, 5);
}
#define init_5_1_1_6
void FASTCALL call_5_1_1_6(SH2_struct* sh)
{
  SH2movll4(sh, 1,1, 6);
}
#define init_5_1_1_7
void FASTCALL call_5_1_1_7(SH2_struct* sh)
{
  SH2movll4(sh, 1,1, 7);
}
#define init_5_1_1_8
void FASTCALL call_5_1_1_8(SH2_struct* sh)
{
  SH2movll4(sh, 1,1, 8);
}
#define init_5_1_1_9
void FASTCALL call_5_1_1_9(SH2_struct* sh)
{
  SH2movll4(sh, 1,1, 9);
}
#define init_5_1_1_10
void FASTCALL call_5_1_1_10(SH2_struct* sh)
{
  SH2movll4(sh, 1,1, 10);
}
#define init_5_1_1_11
void FASTCALL call_5_1_1_11(SH2_struct* sh)
{
  SH2movll4(sh, 1,1, 11);
}
#define init_5_1_1_12
void FASTCALL call_5_1_1_12(SH2_struct* sh)
{
  SH2movll4(sh, 1,1, 12);
}
#define init_5_1_1_13
void FASTCALL call_5_1_1_13(SH2_struct* sh)
{
  SH2movll4(sh, 1,1, 13);
}
#define init_5_1_1_14
void FASTCALL call_5_1_1_14(SH2_struct* sh)
{
  SH2movll4(sh, 1,1, 14);
}
#define init_5_1_1_15
void FASTCALL call_5_1_1_15(SH2_struct* sh)
{
  SH2movll4(sh, 1,1, 15);
}
#define init_5_1_2_0
void FASTCALL call_5_1_2_0(SH2_struct* sh)
{
  SH2movll4(sh, 1,2, 0);
}
#define init_5_1_2_1
void FASTCALL call_5_1_2_1(SH2_struct* sh)
{
  SH2movll4(sh, 1,2, 1);
}
#define init_5_1_2_2
void FASTCALL call_5_1_2_2(SH2_struct* sh)
{
  SH2movll4(sh, 1,2, 2);
}
#define init_5_1_2_3
void FASTCALL call_5_1_2_3(SH2_struct* sh)
{
  SH2movll4(sh, 1,2, 3);
}
#define init_5_1_2_4
void FASTCALL call_5_1_2_4(SH2_struct* sh)
{
  SH2movll4(sh, 1,2, 4);
}
#define init_5_1_2_5
void FASTCALL call_5_1_2_5(SH2_struct* sh)
{
  SH2movll4(sh, 1,2, 5);
}
#define init_5_1_2_6
void FASTCALL call_5_1_2_6(SH2_struct* sh)
{
  SH2movll4(sh, 1,2, 6);
}
#define init_5_1_2_7
void FASTCALL call_5_1_2_7(SH2_struct* sh)
{
  SH2movll4(sh, 1,2, 7);
}
#define init_5_1_2_8
void FASTCALL call_5_1_2_8(SH2_struct* sh)
{
  SH2movll4(sh, 1,2, 8);
}
#define init_5_1_2_9
void FASTCALL call_5_1_2_9(SH2_struct* sh)
{
  SH2movll4(sh, 1,2, 9);
}
#define init_5_1_2_10
void FASTCALL call_5_1_2_10(SH2_struct* sh)
{
  SH2movll4(sh, 1,2, 10);
}
#define init_5_1_2_11
void FASTCALL call_5_1_2_11(SH2_struct* sh)
{
  SH2movll4(sh, 1,2, 11);
}
#define init_5_1_2_12
void FASTCALL call_5_1_2_12(SH2_struct* sh)
{
  SH2movll4(sh, 1,2, 12);
}
#define init_5_1_2_13
void FASTCALL call_5_1_2_13(SH2_struct* sh)
{
  SH2movll4(sh, 1,2, 13);
}
#define init_5_1_2_14
void FASTCALL call_5_1_2_14(SH2_struct* sh)
{
  SH2movll4(sh, 1,2, 14);
}
#define init_5_1_2_15
void FASTCALL call_5_1_2_15(SH2_struct* sh)
{
  SH2movll4(sh, 1,2, 15);
}
#define init_5_1_3_0
void FASTCALL call_5_1_3_0(SH2_struct* sh)
{
  SH2movll4(sh, 1,3, 0);
}
#define init_5_1_3_1
void FASTCALL call_5_1_3_1(SH2_struct* sh)
{
  SH2movll4(sh, 1,3, 1);
}
#define init_5_1_3_2
void FASTCALL call_5_1_3_2(SH2_struct* sh)
{
  SH2movll4(sh, 1,3, 2);
}
#define init_5_1_3_3
void FASTCALL call_5_1_3_3(SH2_struct* sh)
{
  SH2movll4(sh, 1,3, 3);
}
#define init_5_1_3_4
void FASTCALL call_5_1_3_4(SH2_struct* sh)
{
  SH2movll4(sh, 1,3, 4);
}
#define init_5_1_3_5
void FASTCALL call_5_1_3_5(SH2_struct* sh)
{
  SH2movll4(sh, 1,3, 5);
}
#define init_5_1_3_6
void FASTCALL call_5_1_3_6(SH2_struct* sh)
{
  SH2movll4(sh, 1,3, 6);
}
#define init_5_1_3_7
void FASTCALL call_5_1_3_7(SH2_struct* sh)
{
  SH2movll4(sh, 1,3, 7);
}
#define init_5_1_3_8
void FASTCALL call_5_1_3_8(SH2_struct* sh)
{
  SH2movll4(sh, 1,3, 8);
}
#define init_5_1_3_9
void FASTCALL call_5_1_3_9(SH2_struct* sh)
{
  SH2movll4(sh, 1,3, 9);
}
#define init_5_1_3_10
void FASTCALL call_5_1_3_10(SH2_struct* sh)
{
  SH2movll4(sh, 1,3, 10);
}
#define init_5_1_3_11
void FASTCALL call_5_1_3_11(SH2_struct* sh)
{
  SH2movll4(sh, 1,3, 11);
}
#define init_5_1_3_12
void FASTCALL call_5_1_3_12(SH2_struct* sh)
{
  SH2movll4(sh, 1,3, 12);
}
#define init_5_1_3_13
void FASTCALL call_5_1_3_13(SH2_struct* sh)
{
  SH2movll4(sh, 1,3, 13);
}
#define init_5_1_3_14
void FASTCALL call_5_1_3_14(SH2_struct* sh)
{
  SH2movll4(sh, 1,3, 14);
}
#define init_5_1_3_15
void FASTCALL call_5_1_3_15(SH2_struct* sh)
{
  SH2movll4(sh, 1,3, 15);
}
#define init_5_1_4_0
void FASTCALL call_5_1_4_0(SH2_struct* sh)
{
  SH2movll4(sh, 1,4, 0);
}
#define init_5_1_4_1
void FASTCALL call_5_1_4_1(SH2_struct* sh)
{
  SH2movll4(sh, 1,4, 1);
}
#define init_5_1_4_2
void FASTCALL call_5_1_4_2(SH2_struct* sh)
{
  SH2movll4(sh, 1,4, 2);
}
#define init_5_1_4_3
void FASTCALL call_5_1_4_3(SH2_struct* sh)
{
  SH2movll4(sh, 1,4, 3);
}
#define init_5_1_4_4
void FASTCALL call_5_1_4_4(SH2_struct* sh)
{
  SH2movll4(sh, 1,4, 4);
}
#define init_5_1_4_5
void FASTCALL call_5_1_4_5(SH2_struct* sh)
{
  SH2movll4(sh, 1,4, 5);
}
#define init_5_1_4_6
void FASTCALL call_5_1_4_6(SH2_struct* sh)
{
  SH2movll4(sh, 1,4, 6);
}
#define init_5_1_4_7
void FASTCALL call_5_1_4_7(SH2_struct* sh)
{
  SH2movll4(sh, 1,4, 7);
}
#define init_5_1_4_8
void FASTCALL call_5_1_4_8(SH2_struct* sh)
{
  SH2movll4(sh, 1,4, 8);
}
#define init_5_1_4_9
void FASTCALL call_5_1_4_9(SH2_struct* sh)
{
  SH2movll4(sh, 1,4, 9);
}
#define init_5_1_4_10
void FASTCALL call_5_1_4_10(SH2_struct* sh)
{
  SH2movll4(sh, 1,4, 10);
}
#define init_5_1_4_11
void FASTCALL call_5_1_4_11(SH2_struct* sh)
{
  SH2movll4(sh, 1,4, 11);
}
#define init_5_1_4_12
void FASTCALL call_5_1_4_12(SH2_struct* sh)
{
  SH2movll4(sh, 1,4, 12);
}
#define init_5_1_4_13
void FASTCALL call_5_1_4_13(SH2_struct* sh)
{
  SH2movll4(sh, 1,4, 13);
}
#define init_5_1_4_14
void FASTCALL call_5_1_4_14(SH2_struct* sh)
{
  SH2movll4(sh, 1,4, 14);
}
#define init_5_1_4_15
void FASTCALL call_5_1_4_15(SH2_struct* sh)
{
  SH2movll4(sh, 1,4, 15);
}
#define init_5_1_5_0
void FASTCALL call_5_1_5_0(SH2_struct* sh)
{
  SH2movll4(sh, 1,5, 0);
}
#define init_5_1_5_1
void FASTCALL call_5_1_5_1(SH2_struct* sh)
{
  SH2movll4(sh, 1,5, 1);
}
#define init_5_1_5_2
void FASTCALL call_5_1_5_2(SH2_struct* sh)
{
  SH2movll4(sh, 1,5, 2);
}
#define init_5_1_5_3
void FASTCALL call_5_1_5_3(SH2_struct* sh)
{
  SH2movll4(sh, 1,5, 3);
}
#define init_5_1_5_4
void FASTCALL call_5_1_5_4(SH2_struct* sh)
{
  SH2movll4(sh, 1,5, 4);
}
#define init_5_1_5_5
void FASTCALL call_5_1_5_5(SH2_struct* sh)
{
  SH2movll4(sh, 1,5, 5);
}
#define init_5_1_5_6
void FASTCALL call_5_1_5_6(SH2_struct* sh)
{
  SH2movll4(sh, 1,5, 6);
}
#define init_5_1_5_7
void FASTCALL call_5_1_5_7(SH2_struct* sh)
{
  SH2movll4(sh, 1,5, 7);
}
#define init_5_1_5_8
void FASTCALL call_5_1_5_8(SH2_struct* sh)
{
  SH2movll4(sh, 1,5, 8);
}
#define init_5_1_5_9
void FASTCALL call_5_1_5_9(SH2_struct* sh)
{
  SH2movll4(sh, 1,5, 9);
}
#define init_5_1_5_10
void FASTCALL call_5_1_5_10(SH2_struct* sh)
{
  SH2movll4(sh, 1,5, 10);
}
#define init_5_1_5_11
void FASTCALL call_5_1_5_11(SH2_struct* sh)
{
  SH2movll4(sh, 1,5, 11);
}
#define init_5_1_5_12
void FASTCALL call_5_1_5_12(SH2_struct* sh)
{
  SH2movll4(sh, 1,5, 12);
}
#define init_5_1_5_13
void FASTCALL call_5_1_5_13(SH2_struct* sh)
{
  SH2movll4(sh, 1,5, 13);
}
#define init_5_1_5_14
void FASTCALL call_5_1_5_14(SH2_struct* sh)
{
  SH2movll4(sh, 1,5, 14);
}
#define init_5_1_5_15
void FASTCALL call_5_1_5_15(SH2_struct* sh)
{
  SH2movll4(sh, 1,5, 15);
}
#define init_5_1_6_0
void FASTCALL call_5_1_6_0(SH2_struct* sh)
{
  SH2movll4(sh, 1,6, 0);
}
#define init_5_1_6_1
void FASTCALL call_5_1_6_1(SH2_struct* sh)
{
  SH2movll4(sh, 1,6, 1);
}
#define init_5_1_6_2
void FASTCALL call_5_1_6_2(SH2_struct* sh)
{
  SH2movll4(sh, 1,6, 2);
}
#define init_5_1_6_3
void FASTCALL call_5_1_6_3(SH2_struct* sh)
{
  SH2movll4(sh, 1,6, 3);
}
#define init_5_1_6_4
void FASTCALL call_5_1_6_4(SH2_struct* sh)
{
  SH2movll4(sh, 1,6, 4);
}
#define init_5_1_6_5
void FASTCALL call_5_1_6_5(SH2_struct* sh)
{
  SH2movll4(sh, 1,6, 5);
}
#define init_5_1_6_6
void FASTCALL call_5_1_6_6(SH2_struct* sh)
{
  SH2movll4(sh, 1,6, 6);
}
#define init_5_1_6_7
void FASTCALL call_5_1_6_7(SH2_struct* sh)
{
  SH2movll4(sh, 1,6, 7);
}
#define init_5_1_6_8
void FASTCALL call_5_1_6_8(SH2_struct* sh)
{
  SH2movll4(sh, 1,6, 8);
}
#define init_5_1_6_9
void FASTCALL call_5_1_6_9(SH2_struct* sh)
{
  SH2movll4(sh, 1,6, 9);
}
#define init_5_1_6_10
void FASTCALL call_5_1_6_10(SH2_struct* sh)
{
  SH2movll4(sh, 1,6, 10);
}
#define init_5_1_6_11
void FASTCALL call_5_1_6_11(SH2_struct* sh)
{
  SH2movll4(sh, 1,6, 11);
}
#define init_5_1_6_12
void FASTCALL call_5_1_6_12(SH2_struct* sh)
{
  SH2movll4(sh, 1,6, 12);
}
#define init_5_1_6_13
void FASTCALL call_5_1_6_13(SH2_struct* sh)
{
  SH2movll4(sh, 1,6, 13);
}
#define init_5_1_6_14
void FASTCALL call_5_1_6_14(SH2_struct* sh)
{
  SH2movll4(sh, 1,6, 14);
}
#define init_5_1_6_15
void FASTCALL call_5_1_6_15(SH2_struct* sh)
{
  SH2movll4(sh, 1,6, 15);
}
#define init_5_1_7_0
void FASTCALL call_5_1_7_0(SH2_struct* sh)
{
  SH2movll4(sh, 1,7, 0);
}
#define init_5_1_7_1
void FASTCALL call_5_1_7_1(SH2_struct* sh)
{
  SH2movll4(sh, 1,7, 1);
}
#define init_5_1_7_2
void FASTCALL call_5_1_7_2(SH2_struct* sh)
{
  SH2movll4(sh, 1,7, 2);
}
#define init_5_1_7_3
void FASTCALL call_5_1_7_3(SH2_struct* sh)
{
  SH2movll4(sh, 1,7, 3);
}
#define init_5_1_7_4
void FASTCALL call_5_1_7_4(SH2_struct* sh)
{
  SH2movll4(sh, 1,7, 4);
}
#define init_5_1_7_5
void FASTCALL call_5_1_7_5(SH2_struct* sh)
{
  SH2movll4(sh, 1,7, 5);
}
#define init_5_1_7_6
void FASTCALL call_5_1_7_6(SH2_struct* sh)
{
  SH2movll4(sh, 1,7, 6);
}
#define init_5_1_7_7
void FASTCALL call_5_1_7_7(SH2_struct* sh)
{
  SH2movll4(sh, 1,7, 7);
}
#define init_5_1_7_8
void FASTCALL call_5_1_7_8(SH2_struct* sh)
{
  SH2movll4(sh, 1,7, 8);
}
#define init_5_1_7_9
void FASTCALL call_5_1_7_9(SH2_struct* sh)
{
  SH2movll4(sh, 1,7, 9);
}
#define init_5_1_7_10
void FASTCALL call_5_1_7_10(SH2_struct* sh)
{
  SH2movll4(sh, 1,7, 10);
}
#define init_5_1_7_11
void FASTCALL call_5_1_7_11(SH2_struct* sh)
{
  SH2movll4(sh, 1,7, 11);
}
#define init_5_1_7_12
void FASTCALL call_5_1_7_12(SH2_struct* sh)
{
  SH2movll4(sh, 1,7, 12);
}
#define init_5_1_7_13
void FASTCALL call_5_1_7_13(SH2_struct* sh)
{
  SH2movll4(sh, 1,7, 13);
}
#define init_5_1_7_14
void FASTCALL call_5_1_7_14(SH2_struct* sh)
{
  SH2movll4(sh, 1,7, 14);
}
#define init_5_1_7_15
void FASTCALL call_5_1_7_15(SH2_struct* sh)
{
  SH2movll4(sh, 1,7, 15);
}
#define init_5_1_8_0
void FASTCALL call_5_1_8_0(SH2_struct* sh)
{
  SH2movll4(sh, 1,8, 0);
}
#define init_5_1_8_1
void FASTCALL call_5_1_8_1(SH2_struct* sh)
{
  SH2movll4(sh, 1,8, 1);
}
#define init_5_1_8_2
void FASTCALL call_5_1_8_2(SH2_struct* sh)
{
  SH2movll4(sh, 1,8, 2);
}
#define init_5_1_8_3
void FASTCALL call_5_1_8_3(SH2_struct* sh)
{
  SH2movll4(sh, 1,8, 3);
}
#define init_5_1_8_4
void FASTCALL call_5_1_8_4(SH2_struct* sh)
{
  SH2movll4(sh, 1,8, 4);
}
#define init_5_1_8_5
void FASTCALL call_5_1_8_5(SH2_struct* sh)
{
  SH2movll4(sh, 1,8, 5);
}
#define init_5_1_8_6
void FASTCALL call_5_1_8_6(SH2_struct* sh)
{
  SH2movll4(sh, 1,8, 6);
}
#define init_5_1_8_7
void FASTCALL call_5_1_8_7(SH2_struct* sh)
{
  SH2movll4(sh, 1,8, 7);
}
#define init_5_1_8_8
void FASTCALL call_5_1_8_8(SH2_struct* sh)
{
  SH2movll4(sh, 1,8, 8);
}
#define init_5_1_8_9
void FASTCALL call_5_1_8_9(SH2_struct* sh)
{
  SH2movll4(sh, 1,8, 9);
}
#define init_5_1_8_10
void FASTCALL call_5_1_8_10(SH2_struct* sh)
{
  SH2movll4(sh, 1,8, 10);
}
#define init_5_1_8_11
void FASTCALL call_5_1_8_11(SH2_struct* sh)
{
  SH2movll4(sh, 1,8, 11);
}
#define init_5_1_8_12
void FASTCALL call_5_1_8_12(SH2_struct* sh)
{
  SH2movll4(sh, 1,8, 12);
}
#define init_5_1_8_13
void FASTCALL call_5_1_8_13(SH2_struct* sh)
{
  SH2movll4(sh, 1,8, 13);
}
#define init_5_1_8_14
void FASTCALL call_5_1_8_14(SH2_struct* sh)
{
  SH2movll4(sh, 1,8, 14);
}
#define init_5_1_8_15
void FASTCALL call_5_1_8_15(SH2_struct* sh)
{
  SH2movll4(sh, 1,8, 15);
}
#define init_5_1_9_0
void FASTCALL call_5_1_9_0(SH2_struct* sh)
{
  SH2movll4(sh, 1,9, 0);
}
#define init_5_1_9_1
void FASTCALL call_5_1_9_1(SH2_struct* sh)
{
  SH2movll4(sh, 1,9, 1);
}
#define init_5_1_9_2
void FASTCALL call_5_1_9_2(SH2_struct* sh)
{
  SH2movll4(sh, 1,9, 2);
}
#define init_5_1_9_3
void FASTCALL call_5_1_9_3(SH2_struct* sh)
{
  SH2movll4(sh, 1,9, 3);
}
#define init_5_1_9_4
void FASTCALL call_5_1_9_4(SH2_struct* sh)
{
  SH2movll4(sh, 1,9, 4);
}
#define init_5_1_9_5
void FASTCALL call_5_1_9_5(SH2_struct* sh)
{
  SH2movll4(sh, 1,9, 5);
}
#define init_5_1_9_6
void FASTCALL call_5_1_9_6(SH2_struct* sh)
{
  SH2movll4(sh, 1,9, 6);
}
#define init_5_1_9_7
void FASTCALL call_5_1_9_7(SH2_struct* sh)
{
  SH2movll4(sh, 1,9, 7);
}
#define init_5_1_9_8
void FASTCALL call_5_1_9_8(SH2_struct* sh)
{
  SH2movll4(sh, 1,9, 8);
}
#define init_5_1_9_9
void FASTCALL call_5_1_9_9(SH2_struct* sh)
{
  SH2movll4(sh, 1,9, 9);
}
#define init_5_1_9_10
void FASTCALL call_5_1_9_10(SH2_struct* sh)
{
  SH2movll4(sh, 1,9, 10);
}
#define init_5_1_9_11
void FASTCALL call_5_1_9_11(SH2_struct* sh)
{
  SH2movll4(sh, 1,9, 11);
}
#define init_5_1_9_12
void FASTCALL call_5_1_9_12(SH2_struct* sh)
{
  SH2movll4(sh, 1,9, 12);
}
#define init_5_1_9_13
void FASTCALL call_5_1_9_13(SH2_struct* sh)
{
  SH2movll4(sh, 1,9, 13);
}
#define init_5_1_9_14
void FASTCALL call_5_1_9_14(SH2_struct* sh)
{
  SH2movll4(sh, 1,9, 14);
}
#define init_5_1_9_15
void FASTCALL call_5_1_9_15(SH2_struct* sh)
{
  SH2movll4(sh, 1,9, 15);
}
#define init_5_1_10_0
void FASTCALL call_5_1_10_0(SH2_struct* sh)
{
  SH2movll4(sh, 1,10, 0);
}
#define init_5_1_10_1
void FASTCALL call_5_1_10_1(SH2_struct* sh)
{
  SH2movll4(sh, 1,10, 1);
}
#define init_5_1_10_2
void FASTCALL call_5_1_10_2(SH2_struct* sh)
{
  SH2movll4(sh, 1,10, 2);
}
#define init_5_1_10_3
void FASTCALL call_5_1_10_3(SH2_struct* sh)
{
  SH2movll4(sh, 1,10, 3);
}
#define init_5_1_10_4
void FASTCALL call_5_1_10_4(SH2_struct* sh)
{
  SH2movll4(sh, 1,10, 4);
}
#define init_5_1_10_5
void FASTCALL call_5_1_10_5(SH2_struct* sh)
{
  SH2movll4(sh, 1,10, 5);
}
#define init_5_1_10_6
void FASTCALL call_5_1_10_6(SH2_struct* sh)
{
  SH2movll4(sh, 1,10, 6);
}
#define init_5_1_10_7
void FASTCALL call_5_1_10_7(SH2_struct* sh)
{
  SH2movll4(sh, 1,10, 7);
}
#define init_5_1_10_8
void FASTCALL call_5_1_10_8(SH2_struct* sh)
{
  SH2movll4(sh, 1,10, 8);
}
#define init_5_1_10_9
void FASTCALL call_5_1_10_9(SH2_struct* sh)
{
  SH2movll4(sh, 1,10, 9);
}
#define init_5_1_10_10
void FASTCALL call_5_1_10_10(SH2_struct* sh)
{
  SH2movll4(sh, 1,10, 10);
}
#define init_5_1_10_11
void FASTCALL call_5_1_10_11(SH2_struct* sh)
{
  SH2movll4(sh, 1,10, 11);
}
#define init_5_1_10_12
void FASTCALL call_5_1_10_12(SH2_struct* sh)
{
  SH2movll4(sh, 1,10, 12);
}
#define init_5_1_10_13
void FASTCALL call_5_1_10_13(SH2_struct* sh)
{
  SH2movll4(sh, 1,10, 13);
}
#define init_5_1_10_14
void FASTCALL call_5_1_10_14(SH2_struct* sh)
{
  SH2movll4(sh, 1,10, 14);
}
#define init_5_1_10_15
void FASTCALL call_5_1_10_15(SH2_struct* sh)
{
  SH2movll4(sh, 1,10, 15);
}
#define init_5_1_11_0
void FASTCALL call_5_1_11_0(SH2_struct* sh)
{
  SH2movll4(sh, 1,11, 0);
}
#define init_5_1_11_1
void FASTCALL call_5_1_11_1(SH2_struct* sh)
{
  SH2movll4(sh, 1,11, 1);
}
#define init_5_1_11_2
void FASTCALL call_5_1_11_2(SH2_struct* sh)
{
  SH2movll4(sh, 1,11, 2);
}
#define init_5_1_11_3
void FASTCALL call_5_1_11_3(SH2_struct* sh)
{
  SH2movll4(sh, 1,11, 3);
}
#define init_5_1_11_4
void FASTCALL call_5_1_11_4(SH2_struct* sh)
{
  SH2movll4(sh, 1,11, 4);
}
#define init_5_1_11_5
void FASTCALL call_5_1_11_5(SH2_struct* sh)
{
  SH2movll4(sh, 1,11, 5);
}
#define init_5_1_11_6
void FASTCALL call_5_1_11_6(SH2_struct* sh)
{
  SH2movll4(sh, 1,11, 6);
}
#define init_5_1_11_7
void FASTCALL call_5_1_11_7(SH2_struct* sh)
{
  SH2movll4(sh, 1,11, 7);
}
#define init_5_1_11_8
void FASTCALL call_5_1_11_8(SH2_struct* sh)
{
  SH2movll4(sh, 1,11, 8);
}
#define init_5_1_11_9
void FASTCALL call_5_1_11_9(SH2_struct* sh)
{
  SH2movll4(sh, 1,11, 9);
}
#define init_5_1_11_10
void FASTCALL call_5_1_11_10(SH2_struct* sh)
{
  SH2movll4(sh, 1,11, 10);
}
#define init_5_1_11_11
void FASTCALL call_5_1_11_11(SH2_struct* sh)
{
  SH2movll4(sh, 1,11, 11);
}
#define init_5_1_11_12
void FASTCALL call_5_1_11_12(SH2_struct* sh)
{
  SH2movll4(sh, 1,11, 12);
}
#define init_5_1_11_13
void FASTCALL call_5_1_11_13(SH2_struct* sh)
{
  SH2movll4(sh, 1,11, 13);
}
#define init_5_1_11_14
void FASTCALL call_5_1_11_14(SH2_struct* sh)
{
  SH2movll4(sh, 1,11, 14);
}
#define init_5_1_11_15
void FASTCALL call_5_1_11_15(SH2_struct* sh)
{
  SH2movll4(sh, 1,11, 15);
}
#define init_5_1_12_0
void FASTCALL call_5_1_12_0(SH2_struct* sh)
{
  SH2movll4(sh, 1,12, 0);
}
#define init_5_1_12_1
void FASTCALL call_5_1_12_1(SH2_struct* sh)
{
  SH2movll4(sh, 1,12, 1);
}
#define init_5_1_12_2
void FASTCALL call_5_1_12_2(SH2_struct* sh)
{
  SH2movll4(sh, 1,12, 2);
}
#define init_5_1_12_3
void FASTCALL call_5_1_12_3(SH2_struct* sh)
{
  SH2movll4(sh, 1,12, 3);
}
#define init_5_1_12_4
void FASTCALL call_5_1_12_4(SH2_struct* sh)
{
  SH2movll4(sh, 1,12, 4);
}
#define init_5_1_12_5
void FASTCALL call_5_1_12_5(SH2_struct* sh)
{
  SH2movll4(sh, 1,12, 5);
}
#define init_5_1_12_6
void FASTCALL call_5_1_12_6(SH2_struct* sh)
{
  SH2movll4(sh, 1,12, 6);
}
#define init_5_1_12_7
void FASTCALL call_5_1_12_7(SH2_struct* sh)
{
  SH2movll4(sh, 1,12, 7);
}
#define init_5_1_12_8
void FASTCALL call_5_1_12_8(SH2_struct* sh)
{
  SH2movll4(sh, 1,12, 8);
}
#define init_5_1_12_9
void FASTCALL call_5_1_12_9(SH2_struct* sh)
{
  SH2movll4(sh, 1,12, 9);
}
#define init_5_1_12_10
void FASTCALL call_5_1_12_10(SH2_struct* sh)
{
  SH2movll4(sh, 1,12, 10);
}
#define init_5_1_12_11
void FASTCALL call_5_1_12_11(SH2_struct* sh)
{
  SH2movll4(sh, 1,12, 11);
}
#define init_5_1_12_12
void FASTCALL call_5_1_12_12(SH2_struct* sh)
{
  SH2movll4(sh, 1,12, 12);
}
#define init_5_1_12_13
void FASTCALL call_5_1_12_13(SH2_struct* sh)
{
  SH2movll4(sh, 1,12, 13);
}
#define init_5_1_12_14
void FASTCALL call_5_1_12_14(SH2_struct* sh)
{
  SH2movll4(sh, 1,12, 14);
}
#define init_5_1_12_15
void FASTCALL call_5_1_12_15(SH2_struct* sh)
{
  SH2movll4(sh, 1,12, 15);
}
#define init_5_1_13_0
void FASTCALL call_5_1_13_0(SH2_struct* sh)
{
  SH2movll4(sh, 1,13, 0);
}
#define init_5_1_13_1
void FASTCALL call_5_1_13_1(SH2_struct* sh)
{
  SH2movll4(sh, 1,13, 1);
}
#define init_5_1_13_2
void FASTCALL call_5_1_13_2(SH2_struct* sh)
{
  SH2movll4(sh, 1,13, 2);
}
#define init_5_1_13_3
void FASTCALL call_5_1_13_3(SH2_struct* sh)
{
  SH2movll4(sh, 1,13, 3);
}
#define init_5_1_13_4
void FASTCALL call_5_1_13_4(SH2_struct* sh)
{
  SH2movll4(sh, 1,13, 4);
}
#define init_5_1_13_5
void FASTCALL call_5_1_13_5(SH2_struct* sh)
{
  SH2movll4(sh, 1,13, 5);
}
#define init_5_1_13_6
void FASTCALL call_5_1_13_6(SH2_struct* sh)
{
  SH2movll4(sh, 1,13, 6);
}
#define init_5_1_13_7
void FASTCALL call_5_1_13_7(SH2_struct* sh)
{
  SH2movll4(sh, 1,13, 7);
}
#define init_5_1_13_8
void FASTCALL call_5_1_13_8(SH2_struct* sh)
{
  SH2movll4(sh, 1,13, 8);
}
#define init_5_1_13_9
void FASTCALL call_5_1_13_9(SH2_struct* sh)
{
  SH2movll4(sh, 1,13, 9);
}
#define init_5_1_13_10
void FASTCALL call_5_1_13_10(SH2_struct* sh)
{
  SH2movll4(sh, 1,13, 10);
}
#define init_5_1_13_11
void FASTCALL call_5_1_13_11(SH2_struct* sh)
{
  SH2movll4(sh, 1,13, 11);
}
#define init_5_1_13_12
void FASTCALL call_5_1_13_12(SH2_struct* sh)
{
  SH2movll4(sh, 1,13, 12);
}
#define init_5_1_13_13
void FASTCALL call_5_1_13_13(SH2_struct* sh)
{
  SH2movll4(sh, 1,13, 13);
}
#define init_5_1_13_14
void FASTCALL call_5_1_13_14(SH2_struct* sh)
{
  SH2movll4(sh, 1,13, 14);
}
#define init_5_1_13_15
void FASTCALL call_5_1_13_15(SH2_struct* sh)
{
  SH2movll4(sh, 1,13, 15);
}
#define init_5_1_14_0
void FASTCALL call_5_1_14_0(SH2_struct* sh)
{
  SH2movll4(sh, 1,14, 0);
}
#define init_5_1_14_1
void FASTCALL call_5_1_14_1(SH2_struct* sh)
{
  SH2movll4(sh, 1,14, 1);
}
#define init_5_1_14_2
void FASTCALL call_5_1_14_2(SH2_struct* sh)
{
  SH2movll4(sh, 1,14, 2);
}
#define init_5_1_14_3
void FASTCALL call_5_1_14_3(SH2_struct* sh)
{
  SH2movll4(sh, 1,14, 3);
}
#define init_5_1_14_4
void FASTCALL call_5_1_14_4(SH2_struct* sh)
{
  SH2movll4(sh, 1,14, 4);
}
#define init_5_1_14_5
void FASTCALL call_5_1_14_5(SH2_struct* sh)
{
  SH2movll4(sh, 1,14, 5);
}
#define init_5_1_14_6
void FASTCALL call_5_1_14_6(SH2_struct* sh)
{
  SH2movll4(sh, 1,14, 6);
}
#define init_5_1_14_7
void FASTCALL call_5_1_14_7(SH2_struct* sh)
{
  SH2movll4(sh, 1,14, 7);
}
#define init_5_1_14_8
void FASTCALL call_5_1_14_8(SH2_struct* sh)
{
  SH2movll4(sh, 1,14, 8);
}
#define init_5_1_14_9
void FASTCALL call_5_1_14_9(SH2_struct* sh)
{
  SH2movll4(sh, 1,14, 9);
}
#define init_5_1_14_10
void FASTCALL call_5_1_14_10(SH2_struct* sh)
{
  SH2movll4(sh, 1,14, 10);
}
#define init_5_1_14_11
void FASTCALL call_5_1_14_11(SH2_struct* sh)
{
  SH2movll4(sh, 1,14, 11);
}
#define init_5_1_14_12
void FASTCALL call_5_1_14_12(SH2_struct* sh)
{
  SH2movll4(sh, 1,14, 12);
}
#define init_5_1_14_13
void FASTCALL call_5_1_14_13(SH2_struct* sh)
{
  SH2movll4(sh, 1,14, 13);
}
#define init_5_1_14_14
void FASTCALL call_5_1_14_14(SH2_struct* sh)
{
  SH2movll4(sh, 1,14, 14);
}
#define init_5_1_14_15
void FASTCALL call_5_1_14_15(SH2_struct* sh)
{
  SH2movll4(sh, 1,14, 15);
}
#define init_5_1_15_0
void FASTCALL call_5_1_15_0(SH2_struct* sh)
{
  SH2movll4(sh, 1,15, 0);
}
#define init_5_1_15_1
void FASTCALL call_5_1_15_1(SH2_struct* sh)
{
  SH2movll4(sh, 1,15, 1);
}
#define init_5_1_15_2
void FASTCALL call_5_1_15_2(SH2_struct* sh)
{
  SH2movll4(sh, 1,15, 2);
}
#define init_5_1_15_3
void FASTCALL call_5_1_15_3(SH2_struct* sh)
{
  SH2movll4(sh, 1,15, 3);
}
#define init_5_1_15_4
void FASTCALL call_5_1_15_4(SH2_struct* sh)
{
  SH2movll4(sh, 1,15, 4);
}
#define init_5_1_15_5
void FASTCALL call_5_1_15_5(SH2_struct* sh)
{
  SH2movll4(sh, 1,15, 5);
}
#define init_5_1_15_6
void FASTCALL call_5_1_15_6(SH2_struct* sh)
{
  SH2movll4(sh, 1,15, 6);
}
#define init_5_1_15_7
void FASTCALL call_5_1_15_7(SH2_struct* sh)
{
  SH2movll4(sh, 1,15, 7);
}
#define init_5_1_15_8
void FASTCALL call_5_1_15_8(SH2_struct* sh)
{
  SH2movll4(sh, 1,15, 8);
}
#define init_5_1_15_9
void FASTCALL call_5_1_15_9(SH2_struct* sh)
{
  SH2movll4(sh, 1,15, 9);
}
#define init_5_1_15_10
void FASTCALL call_5_1_15_10(SH2_struct* sh)
{
  SH2movll4(sh, 1,15, 10);
}
#define init_5_1_15_11
void FASTCALL call_5_1_15_11(SH2_struct* sh)
{
  SH2movll4(sh, 1,15, 11);
}
#define init_5_1_15_12
void FASTCALL call_5_1_15_12(SH2_struct* sh)
{
  SH2movll4(sh, 1,15, 12);
}
#define init_5_1_15_13
void FASTCALL call_5_1_15_13(SH2_struct* sh)
{
  SH2movll4(sh, 1,15, 13);
}
#define init_5_1_15_14
void FASTCALL call_5_1_15_14(SH2_struct* sh)
{
  SH2movll4(sh, 1,15, 14);
}
#define init_5_1_15_15
void FASTCALL call_5_1_15_15(SH2_struct* sh)
{
  SH2movll4(sh, 1,15, 15);
}
#define init_5_2_0_0
void FASTCALL call_5_2_0_0(SH2_struct* sh)
{
  SH2movll4(sh, 2,0, 0);
}
#define init_5_2_0_1
void FASTCALL call_5_2_0_1(SH2_struct* sh)
{
  SH2movll4(sh, 2,0, 1);
}
#define init_5_2_0_2
void FASTCALL call_5_2_0_2(SH2_struct* sh)
{
  SH2movll4(sh, 2,0, 2);
}
#define init_5_2_0_3
void FASTCALL call_5_2_0_3(SH2_struct* sh)
{
  SH2movll4(sh, 2,0, 3);
}
#define init_5_2_0_4
void FASTCALL call_5_2_0_4(SH2_struct* sh)
{
  SH2movll4(sh, 2,0, 4);
}
#define init_5_2_0_5
void FASTCALL call_5_2_0_5(SH2_struct* sh)
{
  SH2movll4(sh, 2,0, 5);
}
#define init_5_2_0_6
void FASTCALL call_5_2_0_6(SH2_struct* sh)
{
  SH2movll4(sh, 2,0, 6);
}
#define init_5_2_0_7
void FASTCALL call_5_2_0_7(SH2_struct* sh)
{
  SH2movll4(sh, 2,0, 7);
}
#define init_5_2_0_8
void FASTCALL call_5_2_0_8(SH2_struct* sh)
{
  SH2movll4(sh, 2,0, 8);
}
#define init_5_2_0_9
void FASTCALL call_5_2_0_9(SH2_struct* sh)
{
  SH2movll4(sh, 2,0, 9);
}
#define init_5_2_0_10
void FASTCALL call_5_2_0_10(SH2_struct* sh)
{
  SH2movll4(sh, 2,0, 10);
}
#define init_5_2_0_11
void FASTCALL call_5_2_0_11(SH2_struct* sh)
{
  SH2movll4(sh, 2,0, 11);
}
#define init_5_2_0_12
void FASTCALL call_5_2_0_12(SH2_struct* sh)
{
  SH2movll4(sh, 2,0, 12);
}
#define init_5_2_0_13
void FASTCALL call_5_2_0_13(SH2_struct* sh)
{
  SH2movll4(sh, 2,0, 13);
}
#define init_5_2_0_14
void FASTCALL call_5_2_0_14(SH2_struct* sh)
{
  SH2movll4(sh, 2,0, 14);
}
#define init_5_2_0_15
void FASTCALL call_5_2_0_15(SH2_struct* sh)
{
  SH2movll4(sh, 2,0, 15);
}
#define init_5_2_1_0
void FASTCALL call_5_2_1_0(SH2_struct* sh)
{
  SH2movll4(sh, 2,1, 0);
}
#define init_5_2_1_1
void FASTCALL call_5_2_1_1(SH2_struct* sh)
{
  SH2movll4(sh, 2,1, 1);
}
#define init_5_2_1_2
void FASTCALL call_5_2_1_2(SH2_struct* sh)
{
  SH2movll4(sh, 2,1, 2);
}
#define init_5_2_1_3
void FASTCALL call_5_2_1_3(SH2_struct* sh)
{
  SH2movll4(sh, 2,1, 3);
}
#define init_5_2_1_4
void FASTCALL call_5_2_1_4(SH2_struct* sh)
{
  SH2movll4(sh, 2,1, 4);
}
#define init_5_2_1_5
void FASTCALL call_5_2_1_5(SH2_struct* sh)
{
  SH2movll4(sh, 2,1, 5);
}
#define init_5_2_1_6
void FASTCALL call_5_2_1_6(SH2_struct* sh)
{
  SH2movll4(sh, 2,1, 6);
}
#define init_5_2_1_7
void FASTCALL call_5_2_1_7(SH2_struct* sh)
{
  SH2movll4(sh, 2,1, 7);
}
#define init_5_2_1_8
void FASTCALL call_5_2_1_8(SH2_struct* sh)
{
  SH2movll4(sh, 2,1, 8);
}
#define init_5_2_1_9
void FASTCALL call_5_2_1_9(SH2_struct* sh)
{
  SH2movll4(sh, 2,1, 9);
}
#define init_5_2_1_10
void FASTCALL call_5_2_1_10(SH2_struct* sh)
{
  SH2movll4(sh, 2,1, 10);
}
#define init_5_2_1_11
void FASTCALL call_5_2_1_11(SH2_struct* sh)
{
  SH2movll4(sh, 2,1, 11);
}
#define init_5_2_1_12
void FASTCALL call_5_2_1_12(SH2_struct* sh)
{
  SH2movll4(sh, 2,1, 12);
}
#define init_5_2_1_13
void FASTCALL call_5_2_1_13(SH2_struct* sh)
{
  SH2movll4(sh, 2,1, 13);
}
#define init_5_2_1_14
void FASTCALL call_5_2_1_14(SH2_struct* sh)
{
  SH2movll4(sh, 2,1, 14);
}
#define init_5_2_1_15
void FASTCALL call_5_2_1_15(SH2_struct* sh)
{
  SH2movll4(sh, 2,1, 15);
}
#define init_5_2_2_0
void FASTCALL call_5_2_2_0(SH2_struct* sh)
{
  SH2movll4(sh, 2,2, 0);
}
#define init_5_2_2_1
void FASTCALL call_5_2_2_1(SH2_struct* sh)
{
  SH2movll4(sh, 2,2, 1);
}
#define init_5_2_2_2
void FASTCALL call_5_2_2_2(SH2_struct* sh)
{
  SH2movll4(sh, 2,2, 2);
}
#define init_5_2_2_3
void FASTCALL call_5_2_2_3(SH2_struct* sh)
{
  SH2movll4(sh, 2,2, 3);
}
#define init_5_2_2_4
void FASTCALL call_5_2_2_4(SH2_struct* sh)
{
  SH2movll4(sh, 2,2, 4);
}
#define init_5_2_2_5
void FASTCALL call_5_2_2_5(SH2_struct* sh)
{
  SH2movll4(sh, 2,2, 5);
}
#define init_5_2_2_6
void FASTCALL call_5_2_2_6(SH2_struct* sh)
{
  SH2movll4(sh, 2,2, 6);
}
#define init_5_2_2_7
void FASTCALL call_5_2_2_7(SH2_struct* sh)
{
  SH2movll4(sh, 2,2, 7);
}
#define init_5_2_2_8
void FASTCALL call_5_2_2_8(SH2_struct* sh)
{
  SH2movll4(sh, 2,2, 8);
}
#define init_5_2_2_9
void FASTCALL call_5_2_2_9(SH2_struct* sh)
{
  SH2movll4(sh, 2,2, 9);
}
#define init_5_2_2_10
void FASTCALL call_5_2_2_10(SH2_struct* sh)
{
  SH2movll4(sh, 2,2, 10);
}
#define init_5_2_2_11
void FASTCALL call_5_2_2_11(SH2_struct* sh)
{
  SH2movll4(sh, 2,2, 11);
}
#define init_5_2_2_12
void FASTCALL call_5_2_2_12(SH2_struct* sh)
{
  SH2movll4(sh, 2,2, 12);
}
#define init_5_2_2_13
void FASTCALL call_5_2_2_13(SH2_struct* sh)
{
  SH2movll4(sh, 2,2, 13);
}
#define init_5_2_2_14
void FASTCALL call_5_2_2_14(SH2_struct* sh)
{
  SH2movll4(sh, 2,2, 14);
}
#define init_5_2_2_15
void FASTCALL call_5_2_2_15(SH2_struct* sh)
{
  SH2movll4(sh, 2,2, 15);
}
#define init_5_2_3_0
void FASTCALL call_5_2_3_0(SH2_struct* sh)
{
  SH2movll4(sh, 2,3, 0);
}
#define init_5_2_3_1
void FASTCALL call_5_2_3_1(SH2_struct* sh)
{
  SH2movll4(sh, 2,3, 1);
}
#define init_5_2_3_2
void FASTCALL call_5_2_3_2(SH2_struct* sh)
{
  SH2movll4(sh, 2,3, 2);
}
#define init_5_2_3_3
void FASTCALL call_5_2_3_3(SH2_struct* sh)
{
  SH2movll4(sh, 2,3, 3);
}
#define init_5_2_3_4
void FASTCALL call_5_2_3_4(SH2_struct* sh)
{
  SH2movll4(sh, 2,3, 4);
}
#define init_5_2_3_5
void FASTCALL call_5_2_3_5(SH2_struct* sh)
{
  SH2movll4(sh, 2,3, 5);
}
#define init_5_2_3_6
void FASTCALL call_5_2_3_6(SH2_struct* sh)
{
  SH2movll4(sh, 2,3, 6);
}
#define init_5_2_3_7
void FASTCALL call_5_2_3_7(SH2_struct* sh)
{
  SH2movll4(sh, 2,3, 7);
}
#define init_5_2_3_8
void FASTCALL call_5_2_3_8(SH2_struct* sh)
{
  SH2movll4(sh, 2,3, 8);
}
#define init_5_2_3_9
void FASTCALL call_5_2_3_9(SH2_struct* sh)
{
  SH2movll4(sh, 2,3, 9);
}
#define init_5_2_3_10
void FASTCALL call_5_2_3_10(SH2_struct* sh)
{
  SH2movll4(sh, 2,3, 10);
}
#define init_5_2_3_11
void FASTCALL call_5_2_3_11(SH2_struct* sh)
{
  SH2movll4(sh, 2,3, 11);
}
#define init_5_2_3_12
void FASTCALL call_5_2_3_12(SH2_struct* sh)
{
  SH2movll4(sh, 2,3, 12);
}
#define init_5_2_3_13
void FASTCALL call_5_2_3_13(SH2_struct* sh)
{
  SH2movll4(sh, 2,3, 13);
}
#define init_5_2_3_14
void FASTCALL call_5_2_3_14(SH2_struct* sh)
{
  SH2movll4(sh, 2,3, 14);
}
#define init_5_2_3_15
void FASTCALL call_5_2_3_15(SH2_struct* sh)
{
  SH2movll4(sh, 2,3, 15);
}
#define init_5_2_4_0
void FASTCALL call_5_2_4_0(SH2_struct* sh)
{
  SH2movll4(sh, 2,4, 0);
}
#define init_5_2_4_1
void FASTCALL call_5_2_4_1(SH2_struct* sh)
{
  SH2movll4(sh, 2,4, 1);
}
#define init_5_2_4_2
void FASTCALL call_5_2_4_2(SH2_struct* sh)
{
  SH2movll4(sh, 2,4, 2);
}
#define init_5_2_4_3
void FASTCALL call_5_2_4_3(SH2_struct* sh)
{
  SH2movll4(sh, 2,4, 3);
}
#define init_5_2_4_4
void FASTCALL call_5_2_4_4(SH2_struct* sh)
{
  SH2movll4(sh, 2,4, 4);
}
#define init_5_2_4_5
void FASTCALL call_5_2_4_5(SH2_struct* sh)
{
  SH2movll4(sh, 2,4, 5);
}
#define init_5_2_4_6
void FASTCALL call_5_2_4_6(SH2_struct* sh)
{
  SH2movll4(sh, 2,4, 6);
}
#define init_5_2_4_7
void FASTCALL call_5_2_4_7(SH2_struct* sh)
{
  SH2movll4(sh, 2,4, 7);
}
#define init_5_2_4_8
void FASTCALL call_5_2_4_8(SH2_struct* sh)
{
  SH2movll4(sh, 2,4, 8);
}
#define init_5_2_4_9
void FASTCALL call_5_2_4_9(SH2_struct* sh)
{
  SH2movll4(sh, 2,4, 9);
}
#define init_5_2_4_10
void FASTCALL call_5_2_4_10(SH2_struct* sh)
{
  SH2movll4(sh, 2,4, 10);
}
#define init_5_2_4_11
void FASTCALL call_5_2_4_11(SH2_struct* sh)
{
  SH2movll4(sh, 2,4, 11);
}
#define init_5_2_4_12
void FASTCALL call_5_2_4_12(SH2_struct* sh)
{
  SH2movll4(sh, 2,4, 12);
}
#define init_5_2_4_13
void FASTCALL call_5_2_4_13(SH2_struct* sh)
{
  SH2movll4(sh, 2,4, 13);
}
#define init_5_2_4_14
void FASTCALL call_5_2_4_14(SH2_struct* sh)
{
  SH2movll4(sh, 2,4, 14);
}
#define init_5_2_4_15
void FASTCALL call_5_2_4_15(SH2_struct* sh)
{
  SH2movll4(sh, 2,4, 15);
}
#define init_5_2_5_0
void FASTCALL call_5_2_5_0(SH2_struct* sh)
{
  SH2movll4(sh, 2,5, 0);
}
#define init_5_2_5_1
void FASTCALL call_5_2_5_1(SH2_struct* sh)
{
  SH2movll4(sh, 2,5, 1);
}
#define init_5_2_5_2
void FASTCALL call_5_2_5_2(SH2_struct* sh)
{
  SH2movll4(sh, 2,5, 2);
}
#define init_5_2_5_3
void FASTCALL call_5_2_5_3(SH2_struct* sh)
{
  SH2movll4(sh, 2,5, 3);
}
#define init_5_2_5_4
void FASTCALL call_5_2_5_4(SH2_struct* sh)
{
  SH2movll4(sh, 2,5, 4);
}
#define init_5_2_5_5
void FASTCALL call_5_2_5_5(SH2_struct* sh)
{
  SH2movll4(sh, 2,5, 5);
}
#define init_5_2_5_6
void FASTCALL call_5_2_5_6(SH2_struct* sh)
{
  SH2movll4(sh, 2,5, 6);
}
#define init_5_2_5_7
void FASTCALL call_5_2_5_7(SH2_struct* sh)
{
  SH2movll4(sh, 2,5, 7);
}
#define init_5_2_5_8
void FASTCALL call_5_2_5_8(SH2_struct* sh)
{
  SH2movll4(sh, 2,5, 8);
}
#define init_5_2_5_9
void FASTCALL call_5_2_5_9(SH2_struct* sh)
{
  SH2movll4(sh, 2,5, 9);
}
#define init_5_2_5_10
void FASTCALL call_5_2_5_10(SH2_struct* sh)
{
  SH2movll4(sh, 2,5, 10);
}
#define init_5_2_5_11
void FASTCALL call_5_2_5_11(SH2_struct* sh)
{
  SH2movll4(sh, 2,5, 11);
}
#define init_5_2_5_12
void FASTCALL call_5_2_5_12(SH2_struct* sh)
{
  SH2movll4(sh, 2,5, 12);
}
#define init_5_2_5_13
void FASTCALL call_5_2_5_13(SH2_struct* sh)
{
  SH2movll4(sh, 2,5, 13);
}
#define init_5_2_5_14
void FASTCALL call_5_2_5_14(SH2_struct* sh)
{
  SH2movll4(sh, 2,5, 14);
}
#define init_5_2_5_15
void FASTCALL call_5_2_5_15(SH2_struct* sh)
{
  SH2movll4(sh, 2,5, 15);
}
#define init_5_2_6_0
void FASTCALL call_5_2_6_0(SH2_struct* sh)
{
  SH2movll4(sh, 2,6, 0);
}
#define init_5_2_6_1
void FASTCALL call_5_2_6_1(SH2_struct* sh)
{
  SH2movll4(sh, 2,6, 1);
}
#define init_5_2_6_2
void FASTCALL call_5_2_6_2(SH2_struct* sh)
{
  SH2movll4(sh, 2,6, 2);
}
#define init_5_2_6_3
void FASTCALL call_5_2_6_3(SH2_struct* sh)
{
  SH2movll4(sh, 2,6, 3);
}
#define init_5_2_6_4
void FASTCALL call_5_2_6_4(SH2_struct* sh)
{
  SH2movll4(sh, 2,6, 4);
}
#define init_5_2_6_5
void FASTCALL call_5_2_6_5(SH2_struct* sh)
{
  SH2movll4(sh, 2,6, 5);
}
#define init_5_2_6_6
void FASTCALL call_5_2_6_6(SH2_struct* sh)
{
  SH2movll4(sh, 2,6, 6);
}
#define init_5_2_6_7
void FASTCALL call_5_2_6_7(SH2_struct* sh)
{
  SH2movll4(sh, 2,6, 7);
}
#define init_5_2_6_8
void FASTCALL call_5_2_6_8(SH2_struct* sh)
{
  SH2movll4(sh, 2,6, 8);
}
#define init_5_2_6_9
void FASTCALL call_5_2_6_9(SH2_struct* sh)
{
  SH2movll4(sh, 2,6, 9);
}
#define init_5_2_6_10
void FASTCALL call_5_2_6_10(SH2_struct* sh)
{
  SH2movll4(sh, 2,6, 10);
}
#define init_5_2_6_11
void FASTCALL call_5_2_6_11(SH2_struct* sh)
{
  SH2movll4(sh, 2,6, 11);
}
#define init_5_2_6_12
void FASTCALL call_5_2_6_12(SH2_struct* sh)
{
  SH2movll4(sh, 2,6, 12);
}
#define init_5_2_6_13
void FASTCALL call_5_2_6_13(SH2_struct* sh)
{
  SH2movll4(sh, 2,6, 13);
}
#define init_5_2_6_14
void FASTCALL call_5_2_6_14(SH2_struct* sh)
{
  SH2movll4(sh, 2,6, 14);
}
#define init_5_2_6_15
void FASTCALL call_5_2_6_15(SH2_struct* sh)
{
  SH2movll4(sh, 2,6, 15);
}
#define init_5_2_7_0
void FASTCALL call_5_2_7_0(SH2_struct* sh)
{
  SH2movll4(sh, 2,7, 0);
}
#define init_5_2_7_1
void FASTCALL call_5_2_7_1(SH2_struct* sh)
{
  SH2movll4(sh, 2,7, 1);
}
#define init_5_2_7_2
void FASTCALL call_5_2_7_2(SH2_struct* sh)
{
  SH2movll4(sh, 2,7, 2);
}
#define init_5_2_7_3
void FASTCALL call_5_2_7_3(SH2_struct* sh)
{
  SH2movll4(sh, 2,7, 3);
}
#define init_5_2_7_4
void FASTCALL call_5_2_7_4(SH2_struct* sh)
{
  SH2movll4(sh, 2,7, 4);
}
#define init_5_2_7_5
void FASTCALL call_5_2_7_5(SH2_struct* sh)
{
  SH2movll4(sh, 2,7, 5);
}
#define init_5_2_7_6
void FASTCALL call_5_2_7_6(SH2_struct* sh)
{
  SH2movll4(sh, 2,7, 6);
}
#define init_5_2_7_7
void FASTCALL call_5_2_7_7(SH2_struct* sh)
{
  SH2movll4(sh, 2,7, 7);
}
#define init_5_2_7_8
void FASTCALL call_5_2_7_8(SH2_struct* sh)
{
  SH2movll4(sh, 2,7, 8);
}
#define init_5_2_7_9
void FASTCALL call_5_2_7_9(SH2_struct* sh)
{
  SH2movll4(sh, 2,7, 9);
}
#define init_5_2_7_10
void FASTCALL call_5_2_7_10(SH2_struct* sh)
{
  SH2movll4(sh, 2,7, 10);
}
#define init_5_2_7_11
void FASTCALL call_5_2_7_11(SH2_struct* sh)
{
  SH2movll4(sh, 2,7, 11);
}
#define init_5_2_7_12
void FASTCALL call_5_2_7_12(SH2_struct* sh)
{
  SH2movll4(sh, 2,7, 12);
}
#define init_5_2_7_13
void FASTCALL call_5_2_7_13(SH2_struct* sh)
{
  SH2movll4(sh, 2,7, 13);
}
#define init_5_2_7_14
void FASTCALL call_5_2_7_14(SH2_struct* sh)
{
  SH2movll4(sh, 2,7, 14);
}
#define init_5_2_7_15
void FASTCALL call_5_2_7_15(SH2_struct* sh)
{
  SH2movll4(sh, 2,7, 15);
}
#define init_5_2_8_0
void FASTCALL call_5_2_8_0(SH2_struct* sh)
{
  SH2movll4(sh, 2,8, 0);
}
#define init_5_2_8_1
void FASTCALL call_5_2_8_1(SH2_struct* sh)
{
  SH2movll4(sh, 2,8, 1);
}
#define init_5_2_8_2
void FASTCALL call_5_2_8_2(SH2_struct* sh)
{
  SH2movll4(sh, 2,8, 2);
}
#define init_5_2_8_3
void FASTCALL call_5_2_8_3(SH2_struct* sh)
{
  SH2movll4(sh, 2,8, 3);
}
#define init_5_2_8_4
void FASTCALL call_5_2_8_4(SH2_struct* sh)
{
  SH2movll4(sh, 2,8, 4);
}
#define init_5_2_8_5
void FASTCALL call_5_2_8_5(SH2_struct* sh)
{
  SH2movll4(sh, 2,8, 5);
}
#define init_5_2_8_6
void FASTCALL call_5_2_8_6(SH2_struct* sh)
{
  SH2movll4(sh, 2,8, 6);
}
#define init_5_2_8_7
void FASTCALL call_5_2_8_7(SH2_struct* sh)
{
  SH2movll4(sh, 2,8, 7);
}
#define init_5_2_8_8
void FASTCALL call_5_2_8_8(SH2_struct* sh)
{
  SH2movll4(sh, 2,8, 8);
}
#define init_5_2_8_9
void FASTCALL call_5_2_8_9(SH2_struct* sh)
{
  SH2movll4(sh, 2,8, 9);
}
#define init_5_2_8_10
void FASTCALL call_5_2_8_10(SH2_struct* sh)
{
  SH2movll4(sh, 2,8, 10);
}
#define init_5_2_8_11
void FASTCALL call_5_2_8_11(SH2_struct* sh)
{
  SH2movll4(sh, 2,8, 11);
}
#define init_5_2_8_12
void FASTCALL call_5_2_8_12(SH2_struct* sh)
{
  SH2movll4(sh, 2,8, 12);
}
#define init_5_2_8_13
void FASTCALL call_5_2_8_13(SH2_struct* sh)
{
  SH2movll4(sh, 2,8, 13);
}
#define init_5_2_8_14
void FASTCALL call_5_2_8_14(SH2_struct* sh)
{
  SH2movll4(sh, 2,8, 14);
}
#define init_5_2_8_15
void FASTCALL call_5_2_8_15(SH2_struct* sh)
{
  SH2movll4(sh, 2,8, 15);
}
#define init_5_2_9_0
void FASTCALL call_5_2_9_0(SH2_struct* sh)
{
  SH2movll4(sh, 2,9, 0);
}
#define init_5_2_9_1
void FASTCALL call_5_2_9_1(SH2_struct* sh)
{
  SH2movll4(sh, 2,9, 1);
}
#define init_5_2_9_2
void FASTCALL call_5_2_9_2(SH2_struct* sh)
{
  SH2movll4(sh, 2,9, 2);
}
#define init_5_2_9_3
void FASTCALL call_5_2_9_3(SH2_struct* sh)
{
  SH2movll4(sh, 2,9, 3);
}
#define init_5_2_9_4
void FASTCALL call_5_2_9_4(SH2_struct* sh)
{
  SH2movll4(sh, 2,9, 4);
}
#define init_5_2_9_5
void FASTCALL call_5_2_9_5(SH2_struct* sh)
{
  SH2movll4(sh, 2,9, 5);
}
#define init_5_2_9_6
void FASTCALL call_5_2_9_6(SH2_struct* sh)
{
  SH2movll4(sh, 2,9, 6);
}
#define init_5_2_9_7
void FASTCALL call_5_2_9_7(SH2_struct* sh)
{
  SH2movll4(sh, 2,9, 7);
}
#define init_5_2_9_8
void FASTCALL call_5_2_9_8(SH2_struct* sh)
{
  SH2movll4(sh, 2,9, 8);
}
#define init_5_2_9_9
void FASTCALL call_5_2_9_9(SH2_struct* sh)
{
  SH2movll4(sh, 2,9, 9);
}
#define init_5_2_9_10
void FASTCALL call_5_2_9_10(SH2_struct* sh)
{
  SH2movll4(sh, 2,9, 10);
}
#define init_5_2_9_11
void FASTCALL call_5_2_9_11(SH2_struct* sh)
{
  SH2movll4(sh, 2,9, 11);
}
#define init_5_2_9_12
void FASTCALL call_5_2_9_12(SH2_struct* sh)
{
  SH2movll4(sh, 2,9, 12);
}
#define init_5_2_9_13
void FASTCALL call_5_2_9_13(SH2_struct* sh)
{
  SH2movll4(sh, 2,9, 13);
}
#define init_5_2_9_14
void FASTCALL call_5_2_9_14(SH2_struct* sh)
{
  SH2movll4(sh, 2,9, 14);
}
#define init_5_2_9_15
void FASTCALL call_5_2_9_15(SH2_struct* sh)
{
  SH2movll4(sh, 2,9, 15);
}
#define init_5_2_10_0
void FASTCALL call_5_2_10_0(SH2_struct* sh)
{
  SH2movll4(sh, 2,10, 0);
}
#define init_5_2_10_1
void FASTCALL call_5_2_10_1(SH2_struct* sh)
{
  SH2movll4(sh, 2,10, 1);
}
#define init_5_2_10_2
void FASTCALL call_5_2_10_2(SH2_struct* sh)
{
  SH2movll4(sh, 2,10, 2);
}
#define init_5_2_10_3
void FASTCALL call_5_2_10_3(SH2_struct* sh)
{
  SH2movll4(sh, 2,10, 3);
}
#define init_5_2_10_4
void FASTCALL call_5_2_10_4(SH2_struct* sh)
{
  SH2movll4(sh, 2,10, 4);
}
#define init_5_2_10_5
void FASTCALL call_5_2_10_5(SH2_struct* sh)
{
  SH2movll4(sh, 2,10, 5);
}
#define init_5_2_10_6
void FASTCALL call_5_2_10_6(SH2_struct* sh)
{
  SH2movll4(sh, 2,10, 6);
}
#define init_5_2_10_7
void FASTCALL call_5_2_10_7(SH2_struct* sh)
{
  SH2movll4(sh, 2,10, 7);
}
#define init_5_2_10_8
void FASTCALL call_5_2_10_8(SH2_struct* sh)
{
  SH2movll4(sh, 2,10, 8);
}
#define init_5_2_10_9
void FASTCALL call_5_2_10_9(SH2_struct* sh)
{
  SH2movll4(sh, 2,10, 9);
}
#define init_5_2_10_10
void FASTCALL call_5_2_10_10(SH2_struct* sh)
{
  SH2movll4(sh, 2,10, 10);
}
#define init_5_2_10_11
void FASTCALL call_5_2_10_11(SH2_struct* sh)
{
  SH2movll4(sh, 2,10, 11);
}
#define init_5_2_10_12
void FASTCALL call_5_2_10_12(SH2_struct* sh)
{
  SH2movll4(sh, 2,10, 12);
}
#define init_5_2_10_13
void FASTCALL call_5_2_10_13(SH2_struct* sh)
{
  SH2movll4(sh, 2,10, 13);
}
#define init_5_2_10_14
void FASTCALL call_5_2_10_14(SH2_struct* sh)
{
  SH2movll4(sh, 2,10, 14);
}
#define init_5_2_10_15
void FASTCALL call_5_2_10_15(SH2_struct* sh)
{
  SH2movll4(sh, 2,10, 15);
}
#define init_5_2_11_0
void FASTCALL call_5_2_11_0(SH2_struct* sh)
{
  SH2movll4(sh, 2,11, 0);
}
#define init_5_2_11_1
void FASTCALL call_5_2_11_1(SH2_struct* sh)
{
  SH2movll4(sh, 2,11, 1);
}
#define init_5_2_11_2
void FASTCALL call_5_2_11_2(SH2_struct* sh)
{
  SH2movll4(sh, 2,11, 2);
}
#define init_5_2_11_3
void FASTCALL call_5_2_11_3(SH2_struct* sh)
{
  SH2movll4(sh, 2,11, 3);
}
#define init_5_2_11_4
void FASTCALL call_5_2_11_4(SH2_struct* sh)
{
  SH2movll4(sh, 2,11, 4);
}
#define init_5_2_11_5
void FASTCALL call_5_2_11_5(SH2_struct* sh)
{
  SH2movll4(sh, 2,11, 5);
}
#define init_5_2_11_6
void FASTCALL call_5_2_11_6(SH2_struct* sh)
{
  SH2movll4(sh, 2,11, 6);
}
#define init_5_2_11_7
void FASTCALL call_5_2_11_7(SH2_struct* sh)
{
  SH2movll4(sh, 2,11, 7);
}
#define init_5_2_11_8
void FASTCALL call_5_2_11_8(SH2_struct* sh)
{
  SH2movll4(sh, 2,11, 8);
}
#define init_5_2_11_9
void FASTCALL call_5_2_11_9(SH2_struct* sh)
{
  SH2movll4(sh, 2,11, 9);
}
#define init_5_2_11_10
void FASTCALL call_5_2_11_10(SH2_struct* sh)
{
  SH2movll4(sh, 2,11, 10);
}
#define init_5_2_11_11
void FASTCALL call_5_2_11_11(SH2_struct* sh)
{
  SH2movll4(sh, 2,11, 11);
}
#define init_5_2_11_12
void FASTCALL call_5_2_11_12(SH2_struct* sh)
{
  SH2movll4(sh, 2,11, 12);
}
#define init_5_2_11_13
void FASTCALL call_5_2_11_13(SH2_struct* sh)
{
  SH2movll4(sh, 2,11, 13);
}
#define init_5_2_11_14
void FASTCALL call_5_2_11_14(SH2_struct* sh)
{
  SH2movll4(sh, 2,11, 14);
}
#define init_5_2_11_15
void FASTCALL call_5_2_11_15(SH2_struct* sh)
{
  SH2movll4(sh, 2,11, 15);
}
#define init_5_2_12_0
void FASTCALL call_5_2_12_0(SH2_struct* sh)
{
  SH2movll4(sh, 2,12, 0);
}
#define init_5_2_12_1
void FASTCALL call_5_2_12_1(SH2_struct* sh)
{
  SH2movll4(sh, 2,12, 1);
}
#define init_5_2_12_2
void FASTCALL call_5_2_12_2(SH2_struct* sh)
{
  SH2movll4(sh, 2,12, 2);
}
#define init_5_2_12_3
void FASTCALL call_5_2_12_3(SH2_struct* sh)
{
  SH2movll4(sh, 2,12, 3);
}
#define init_5_2_12_4
void FASTCALL call_5_2_12_4(SH2_struct* sh)
{
  SH2movll4(sh, 2,12, 4);
}
#define init_5_2_12_5
void FASTCALL call_5_2_12_5(SH2_struct* sh)
{
  SH2movll4(sh, 2,12, 5);
}
#define init_5_2_12_6
void FASTCALL call_5_2_12_6(SH2_struct* sh)
{
  SH2movll4(sh, 2,12, 6);
}
#define init_5_2_12_7
void FASTCALL call_5_2_12_7(SH2_struct* sh)
{
  SH2movll4(sh, 2,12, 7);
}
#define init_5_2_12_8
void FASTCALL call_5_2_12_8(SH2_struct* sh)
{
  SH2movll4(sh, 2,12, 8);
}
#define init_5_2_12_9
void FASTCALL call_5_2_12_9(SH2_struct* sh)
{
  SH2movll4(sh, 2,12, 9);
}
#define init_5_2_12_10
void FASTCALL call_5_2_12_10(SH2_struct* sh)
{
  SH2movll4(sh, 2,12, 10);
}
#define init_5_2_12_11
void FASTCALL call_5_2_12_11(SH2_struct* sh)
{
  SH2movll4(sh, 2,12, 11);
}
#define init_5_2_12_12
void FASTCALL call_5_2_12_12(SH2_struct* sh)
{
  SH2movll4(sh, 2,12, 12);
}
#define init_5_2_12_13
void FASTCALL call_5_2_12_13(SH2_struct* sh)
{
  SH2movll4(sh, 2,12, 13);
}
#define init_5_2_12_14
void FASTCALL call_5_2_12_14(SH2_struct* sh)
{
  SH2movll4(sh, 2,12, 14);
}
#define init_5_2_12_15
void FASTCALL call_5_2_12_15(SH2_struct* sh)
{
  SH2movll4(sh, 2,12, 15);
}
#define init_5_2_13_0
void FASTCALL call_5_2_13_0(SH2_struct* sh)
{
  SH2movll4(sh, 2,13, 0);
}
#define init_5_2_13_1
void FASTCALL call_5_2_13_1(SH2_struct* sh)
{
  SH2movll4(sh, 2,13, 1);
}
#define init_5_2_13_2
void FASTCALL call_5_2_13_2(SH2_struct* sh)
{
  SH2movll4(sh, 2,13, 2);
}
#define init_5_2_13_3
void FASTCALL call_5_2_13_3(SH2_struct* sh)
{
  SH2movll4(sh, 2,13, 3);
}
#define init_5_2_13_4
void FASTCALL call_5_2_13_4(SH2_struct* sh)
{
  SH2movll4(sh, 2,13, 4);
}
#define init_5_2_13_5
void FASTCALL call_5_2_13_5(SH2_struct* sh)
{
  SH2movll4(sh, 2,13, 5);
}
#define init_5_2_13_6
void FASTCALL call_5_2_13_6(SH2_struct* sh)
{
  SH2movll4(sh, 2,13, 6);
}
#define init_5_2_13_7
void FASTCALL call_5_2_13_7(SH2_struct* sh)
{
  SH2movll4(sh, 2,13, 7);
}
#define init_5_2_13_8
void FASTCALL call_5_2_13_8(SH2_struct* sh)
{
  SH2movll4(sh, 2,13, 8);
}
#define init_5_2_13_9
void FASTCALL call_5_2_13_9(SH2_struct* sh)
{
  SH2movll4(sh, 2,13, 9);
}
#define init_5_2_13_10
void FASTCALL call_5_2_13_10(SH2_struct* sh)
{
  SH2movll4(sh, 2,13, 10);
}
#define init_5_2_13_11
void FASTCALL call_5_2_13_11(SH2_struct* sh)
{
  SH2movll4(sh, 2,13, 11);
}
#define init_5_2_13_12
void FASTCALL call_5_2_13_12(SH2_struct* sh)
{
  SH2movll4(sh, 2,13, 12);
}
#define init_5_2_13_13
void FASTCALL call_5_2_13_13(SH2_struct* sh)
{
  SH2movll4(sh, 2,13, 13);
}
#define init_5_2_13_14
void FASTCALL call_5_2_13_14(SH2_struct* sh)
{
  SH2movll4(sh, 2,13, 14);
}
#define init_5_2_13_15
void FASTCALL call_5_2_13_15(SH2_struct* sh)
{
  SH2movll4(sh, 2,13, 15);
}
#define init_5_2_14_0
void FASTCALL call_5_2_14_0(SH2_struct* sh)
{
  SH2movll4(sh, 2,14, 0);
}
#define init_5_2_14_1
void FASTCALL call_5_2_14_1(SH2_struct* sh)
{
  SH2movll4(sh, 2,14, 1);
}
#define init_5_2_14_2
void FASTCALL call_5_2_14_2(SH2_struct* sh)
{
  SH2movll4(sh, 2,14, 2);
}
#define init_5_2_14_3
void FASTCALL call_5_2_14_3(SH2_struct* sh)
{
  SH2movll4(sh, 2,14, 3);
}
#define init_5_2_14_4
void FASTCALL call_5_2_14_4(SH2_struct* sh)
{
  SH2movll4(sh, 2,14, 4);
}
#define init_5_2_14_5
void FASTCALL call_5_2_14_5(SH2_struct* sh)
{
  SH2movll4(sh, 2,14, 5);
}
#define init_5_2_14_6
void FASTCALL call_5_2_14_6(SH2_struct* sh)
{
  SH2movll4(sh, 2,14, 6);
}
#define init_5_2_14_7
void FASTCALL call_5_2_14_7(SH2_struct* sh)
{
  SH2movll4(sh, 2,14, 7);
}
#define init_5_2_14_8
void FASTCALL call_5_2_14_8(SH2_struct* sh)
{
  SH2movll4(sh, 2,14, 8);
}
#define init_5_2_14_9
void FASTCALL call_5_2_14_9(SH2_struct* sh)
{
  SH2movll4(sh, 2,14, 9);
}
#define init_5_2_14_10
void FASTCALL call_5_2_14_10(SH2_struct* sh)
{
  SH2movll4(sh, 2,14, 10);
}
#define init_5_2_14_11
void FASTCALL call_5_2_14_11(SH2_struct* sh)
{
  SH2movll4(sh, 2,14, 11);
}
#define init_5_2_14_12
void FASTCALL call_5_2_14_12(SH2_struct* sh)
{
  SH2movll4(sh, 2,14, 12);
}
#define init_5_2_14_13
void FASTCALL call_5_2_14_13(SH2_struct* sh)
{
  SH2movll4(sh, 2,14, 13);
}
#define init_5_2_14_14
void FASTCALL call_5_2_14_14(SH2_struct* sh)
{
  SH2movll4(sh, 2,14, 14);
}
#define init_5_2_14_15
void FASTCALL call_5_2_14_15(SH2_struct* sh)
{
  SH2movll4(sh, 2,14, 15);
}
#define init_5_2_15_0
void FASTCALL call_5_2_15_0(SH2_struct* sh)
{
  SH2movll4(sh, 2,15, 0);
}
#define init_5_2_15_1
void FASTCALL call_5_2_15_1(SH2_struct* sh)
{
  SH2movll4(sh, 2,15, 1);
}
#define init_5_2_15_2
void FASTCALL call_5_2_15_2(SH2_struct* sh)
{
  SH2movll4(sh, 2,15, 2);
}
#define init_5_2_15_3
void FASTCALL call_5_2_15_3(SH2_struct* sh)
{
  SH2movll4(sh, 2,15, 3);
}
#define init_5_2_15_4
void FASTCALL call_5_2_15_4(SH2_struct* sh)
{
  SH2movll4(sh, 2,15, 4);
}
#define init_5_2_15_5
void FASTCALL call_5_2_15_5(SH2_struct* sh)
{
  SH2movll4(sh, 2,15, 5);
}
#define init_5_2_15_6
void FASTCALL call_5_2_15_6(SH2_struct* sh)
{
  SH2movll4(sh, 2,15, 6);
}
#define init_5_2_15_7
void FASTCALL call_5_2_15_7(SH2_struct* sh)
{
  SH2movll4(sh, 2,15, 7);
}
#define init_5_2_15_8
void FASTCALL call_5_2_15_8(SH2_struct* sh)
{
  SH2movll4(sh, 2,15, 8);
}
#define init_5_2_15_9
void FASTCALL call_5_2_15_9(SH2_struct* sh)
{
  SH2movll4(sh, 2,15, 9);
}
#define init_5_2_15_10
void FASTCALL call_5_2_15_10(SH2_struct* sh)
{
  SH2movll4(sh, 2,15, 10);
}
#define init_5_2_15_11
void FASTCALL call_5_2_15_11(SH2_struct* sh)
{
  SH2movll4(sh, 2,15, 11);
}
#define init_5_2_15_12
void FASTCALL call_5_2_15_12(SH2_struct* sh)
{
  SH2movll4(sh, 2,15, 12);
}
#define init_5_2_15_13
void FASTCALL call_5_2_15_13(SH2_struct* sh)
{
  SH2movll4(sh, 2,15, 13);
}
#define init_5_2_15_14
void FASTCALL call_5_2_15_14(SH2_struct* sh)
{
  SH2movll4(sh, 2,15, 14);
}
#define init_5_2_15_15
void FASTCALL call_5_2_15_15(SH2_struct* sh)
{
  SH2movll4(sh, 2,15, 15);
}
#define init_5_3_0_0
void FASTCALL call_5_3_0_0(SH2_struct* sh)
{
  SH2movll4(sh, 3,0, 0);
}
#define init_5_3_0_1
void FASTCALL call_5_3_0_1(SH2_struct* sh)
{
  SH2movll4(sh, 3,0, 1);
}
#define init_5_3_0_2
void FASTCALL call_5_3_0_2(SH2_struct* sh)
{
  SH2movll4(sh, 3,0, 2);
}
#define init_5_3_0_3
void FASTCALL call_5_3_0_3(SH2_struct* sh)
{
  SH2movll4(sh, 3,0, 3);
}
#define init_5_3_0_4
void FASTCALL call_5_3_0_4(SH2_struct* sh)
{
  SH2movll4(sh, 3,0, 4);
}
#define init_5_3_0_5
void FASTCALL call_5_3_0_5(SH2_struct* sh)
{
  SH2movll4(sh, 3,0, 5);
}
#define init_5_3_0_6
void FASTCALL call_5_3_0_6(SH2_struct* sh)
{
  SH2movll4(sh, 3,0, 6);
}
#define init_5_3_0_7
void FASTCALL call_5_3_0_7(SH2_struct* sh)
{
  SH2movll4(sh, 3,0, 7);
}
#define init_5_3_0_8
void FASTCALL call_5_3_0_8(SH2_struct* sh)
{
  SH2movll4(sh, 3,0, 8);
}
#define init_5_3_0_9
void FASTCALL call_5_3_0_9(SH2_struct* sh)
{
  SH2movll4(sh, 3,0, 9);
}
#define init_5_3_0_10
void FASTCALL call_5_3_0_10(SH2_struct* sh)
{
  SH2movll4(sh, 3,0, 10);
}
#define init_5_3_0_11
void FASTCALL call_5_3_0_11(SH2_struct* sh)
{
  SH2movll4(sh, 3,0, 11);
}
#define init_5_3_0_12
void FASTCALL call_5_3_0_12(SH2_struct* sh)
{
  SH2movll4(sh, 3,0, 12);
}
#define init_5_3_0_13
void FASTCALL call_5_3_0_13(SH2_struct* sh)
{
  SH2movll4(sh, 3,0, 13);
}
#define init_5_3_0_14
void FASTCALL call_5_3_0_14(SH2_struct* sh)
{
  SH2movll4(sh, 3,0, 14);
}
#define init_5_3_0_15
void FASTCALL call_5_3_0_15(SH2_struct* sh)
{
  SH2movll4(sh, 3,0, 15);
}
#define init_5_3_1_0
void FASTCALL call_5_3_1_0(SH2_struct* sh)
{
  SH2movll4(sh, 3,1, 0);
}
#define init_5_3_1_1
void FASTCALL call_5_3_1_1(SH2_struct* sh)
{
  SH2movll4(sh, 3,1, 1);
}
#define init_5_3_1_2
void FASTCALL call_5_3_1_2(SH2_struct* sh)
{
  SH2movll4(sh, 3,1, 2);
}
#define init_5_3_1_3
void FASTCALL call_5_3_1_3(SH2_struct* sh)
{
  SH2movll4(sh, 3,1, 3);
}
#define init_5_3_1_4
void FASTCALL call_5_3_1_4(SH2_struct* sh)
{
  SH2movll4(sh, 3,1, 4);
}
#define init_5_3_1_5
void FASTCALL call_5_3_1_5(SH2_struct* sh)
{
  SH2movll4(sh, 3,1, 5);
}
#define init_5_3_1_6
void FASTCALL call_5_3_1_6(SH2_struct* sh)
{
  SH2movll4(sh, 3,1, 6);
}
#define init_5_3_1_7
void FASTCALL call_5_3_1_7(SH2_struct* sh)
{
  SH2movll4(sh, 3,1, 7);
}
#define init_5_3_1_8
void FASTCALL call_5_3_1_8(SH2_struct* sh)
{
  SH2movll4(sh, 3,1, 8);
}
#define init_5_3_1_9
void FASTCALL call_5_3_1_9(SH2_struct* sh)
{
  SH2movll4(sh, 3,1, 9);
}
#define init_5_3_1_10
void FASTCALL call_5_3_1_10(SH2_struct* sh)
{
  SH2movll4(sh, 3,1, 10);
}
#define init_5_3_1_11
void FASTCALL call_5_3_1_11(SH2_struct* sh)
{
  SH2movll4(sh, 3,1, 11);
}
#define init_5_3_1_12
void FASTCALL call_5_3_1_12(SH2_struct* sh)
{
  SH2movll4(sh, 3,1, 12);
}
#define init_5_3_1_13
void FASTCALL call_5_3_1_13(SH2_struct* sh)
{
  SH2movll4(sh, 3,1, 13);
}
#define init_5_3_1_14
void FASTCALL call_5_3_1_14(SH2_struct* sh)
{
  SH2movll4(sh, 3,1, 14);
}
#define init_5_3_1_15
void FASTCALL call_5_3_1_15(SH2_struct* sh)
{
  SH2movll4(sh, 3,1, 15);
}
#define init_5_3_2_0
void FASTCALL call_5_3_2_0(SH2_struct* sh)
{
  SH2movll4(sh, 3,2, 0);
}
#define init_5_3_2_1
void FASTCALL call_5_3_2_1(SH2_struct* sh)
{
  SH2movll4(sh, 3,2, 1);
}
#define init_5_3_2_2
void FASTCALL call_5_3_2_2(SH2_struct* sh)
{
  SH2movll4(sh, 3,2, 2);
}
#define init_5_3_2_3
void FASTCALL call_5_3_2_3(SH2_struct* sh)
{
  SH2movll4(sh, 3,2, 3);
}
#define init_5_3_2_4
void FASTCALL call_5_3_2_4(SH2_struct* sh)
{
  SH2movll4(sh, 3,2, 4);
}
#define init_5_3_2_5
void FASTCALL call_5_3_2_5(SH2_struct* sh)
{
  SH2movll4(sh, 3,2, 5);
}
#define init_5_3_2_6
void FASTCALL call_5_3_2_6(SH2_struct* sh)
{
  SH2movll4(sh, 3,2, 6);
}
#define init_5_3_2_7
void FASTCALL call_5_3_2_7(SH2_struct* sh)
{
  SH2movll4(sh, 3,2, 7);
}
#define init_5_3_2_8
void FASTCALL call_5_3_2_8(SH2_struct* sh)
{
  SH2movll4(sh, 3,2, 8);
}
#define init_5_3_2_9
void FASTCALL call_5_3_2_9(SH2_struct* sh)
{
  SH2movll4(sh, 3,2, 9);
}
#define init_5_3_2_10
void FASTCALL call_5_3_2_10(SH2_struct* sh)
{
  SH2movll4(sh, 3,2, 10);
}
#define init_5_3_2_11
void FASTCALL call_5_3_2_11(SH2_struct* sh)
{
  SH2movll4(sh, 3,2, 11);
}
#define init_5_3_2_12
void FASTCALL call_5_3_2_12(SH2_struct* sh)
{
  SH2movll4(sh, 3,2, 12);
}
#define init_5_3_2_13
void FASTCALL call_5_3_2_13(SH2_struct* sh)
{
  SH2movll4(sh, 3,2, 13);
}
#define init_5_3_2_14
void FASTCALL call_5_3_2_14(SH2_struct* sh)
{
  SH2movll4(sh, 3,2, 14);
}
#define init_5_3_2_15
void FASTCALL call_5_3_2_15(SH2_struct* sh)
{
  SH2movll4(sh, 3,2, 15);
}
#define init_5_3_3_0
void FASTCALL call_5_3_3_0(SH2_struct* sh)
{
  SH2movll4(sh, 3,3, 0);
}
#define init_5_3_3_1
void FASTCALL call_5_3_3_1(SH2_struct* sh)
{
  SH2movll4(sh, 3,3, 1);
}
#define init_5_3_3_2
void FASTCALL call_5_3_3_2(SH2_struct* sh)
{
  SH2movll4(sh, 3,3, 2);
}
#define init_5_3_3_3
void FASTCALL call_5_3_3_3(SH2_struct* sh)
{
  SH2movll4(sh, 3,3, 3);
}
#define init_5_3_3_4
void FASTCALL call_5_3_3_4(SH2_struct* sh)
{
  SH2movll4(sh, 3,3, 4);
}
#define init_5_3_3_5
void FASTCALL call_5_3_3_5(SH2_struct* sh)
{
  SH2movll4(sh, 3,3, 5);
}
#define init_5_3_3_6
void FASTCALL call_5_3_3_6(SH2_struct* sh)
{
  SH2movll4(sh, 3,3, 6);
}
#define init_5_3_3_7
void FASTCALL call_5_3_3_7(SH2_struct* sh)
{
  SH2movll4(sh, 3,3, 7);
}
#define init_5_3_3_8
void FASTCALL call_5_3_3_8(SH2_struct* sh)
{
  SH2movll4(sh, 3,3, 8);
}
#define init_5_3_3_9
void FASTCALL call_5_3_3_9(SH2_struct* sh)
{
  SH2movll4(sh, 3,3, 9);
}
#define init_5_3_3_10
void FASTCALL call_5_3_3_10(SH2_struct* sh)
{
  SH2movll4(sh, 3,3, 10);
}
#define init_5_3_3_11
void FASTCALL call_5_3_3_11(SH2_struct* sh)
{
  SH2movll4(sh, 3,3, 11);
}
#define init_5_3_3_12
void FASTCALL call_5_3_3_12(SH2_struct* sh)
{
  SH2movll4(sh, 3,3, 12);
}
#define init_5_3_3_13
void FASTCALL call_5_3_3_13(SH2_struct* sh)
{
  SH2movll4(sh, 3,3, 13);
}
#define init_5_3_3_14
void FASTCALL call_5_3_3_14(SH2_struct* sh)
{
  SH2movll4(sh, 3,3, 14);
}
#define init_5_3_3_15
void FASTCALL call_5_3_3_15(SH2_struct* sh)
{
  SH2movll4(sh, 3,3, 15);
}
#define init_5_3_4_0
void FASTCALL call_5_3_4_0(SH2_struct* sh)
{
  SH2movll4(sh, 3,4, 0);
}
#define init_5_3_4_1
void FASTCALL call_5_3_4_1(SH2_struct* sh)
{
  SH2movll4(sh, 3,4, 1);
}
#define init_5_3_4_2
void FASTCALL call_5_3_4_2(SH2_struct* sh)
{
  SH2movll4(sh, 3,4, 2);
}
#define init_5_3_4_3
void FASTCALL call_5_3_4_3(SH2_struct* sh)
{
  SH2movll4(sh, 3,4, 3);
}
#define init_5_3_4_4
void FASTCALL call_5_3_4_4(SH2_struct* sh)
{
  SH2movll4(sh, 3,4, 4);
}
#define init_5_3_4_5
void FASTCALL call_5_3_4_5(SH2_struct* sh)
{
  SH2movll4(sh, 3,4, 5);
}
#define init_5_3_4_6
void FASTCALL call_5_3_4_6(SH2_struct* sh)
{
  SH2movll4(sh, 3,4, 6);
}
#define init_5_3_4_7
void FASTCALL call_5_3_4_7(SH2_struct* sh)
{
  SH2movll4(sh, 3,4, 7);
}
#define init_5_3_4_8
void FASTCALL call_5_3_4_8(SH2_struct* sh)
{
  SH2movll4(sh, 3,4, 8);
}
#define init_5_3_4_9
void FASTCALL call_5_3_4_9(SH2_struct* sh)
{
  SH2movll4(sh, 3,4, 9);
}
#define init_5_3_4_10
void FASTCALL call_5_3_4_10(SH2_struct* sh)
{
  SH2movll4(sh, 3,4, 10);
}
#define init_5_3_4_11
void FASTCALL call_5_3_4_11(SH2_struct* sh)
{
  SH2movll4(sh, 3,4, 11);
}
#define init_5_3_4_12
void FASTCALL call_5_3_4_12(SH2_struct* sh)
{
  SH2movll4(sh, 3,4, 12);
}
#define init_5_3_4_13
void FASTCALL call_5_3_4_13(SH2_struct* sh)
{
  SH2movll4(sh, 3,4, 13);
}
#define init_5_3_4_14
void FASTCALL call_5_3_4_14(SH2_struct* sh)
{
  SH2movll4(sh, 3,4, 14);
}
#define init_5_3_4_15
void FASTCALL call_5_3_4_15(SH2_struct* sh)
{
  SH2movll4(sh, 3,4, 15);
}
#define init_5_3_5_0
void FASTCALL call_5_3_5_0(SH2_struct* sh)
{
  SH2movll4(sh, 3,5, 0);
}
#define init_5_3_5_1
void FASTCALL call_5_3_5_1(SH2_struct* sh)
{
  SH2movll4(sh, 3,5, 1);
}
#define init_5_3_5_2
void FASTCALL call_5_3_5_2(SH2_struct* sh)
{
  SH2movll4(sh, 3,5, 2);
}
#define init_5_3_5_3
void FASTCALL call_5_3_5_3(SH2_struct* sh)
{
  SH2movll4(sh, 3,5, 3);
}
#define init_5_3_5_4
void FASTCALL call_5_3_5_4(SH2_struct* sh)
{
  SH2movll4(sh, 3,5, 4);
}
#define init_5_3_5_5
void FASTCALL call_5_3_5_5(SH2_struct* sh)
{
  SH2movll4(sh, 3,5, 5);
}
#define init_5_3_5_6
void FASTCALL call_5_3_5_6(SH2_struct* sh)
{
  SH2movll4(sh, 3,5, 6);
}
#define init_5_3_5_7
void FASTCALL call_5_3_5_7(SH2_struct* sh)
{
  SH2movll4(sh, 3,5, 7);
}
#define init_5_3_5_8
void FASTCALL call_5_3_5_8(SH2_struct* sh)
{
  SH2movll4(sh, 3,5, 8);
}
#define init_5_3_5_9
void FASTCALL call_5_3_5_9(SH2_struct* sh)
{
  SH2movll4(sh, 3,5, 9);
}
#define init_5_3_5_10
void FASTCALL call_5_3_5_10(SH2_struct* sh)
{
  SH2movll4(sh, 3,5, 10);
}
#define init_5_3_5_11
void FASTCALL call_5_3_5_11(SH2_struct* sh)
{
  SH2movll4(sh, 3,5, 11);
}
#define init_5_3_5_12
void FASTCALL call_5_3_5_12(SH2_struct* sh)
{
  SH2movll4(sh, 3,5, 12);
}
#define init_5_3_5_13
void FASTCALL call_5_3_5_13(SH2_struct* sh)
{
  SH2movll4(sh, 3,5, 13);
}
#define init_5_3_5_14
void FASTCALL call_5_3_5_14(SH2_struct* sh)
{
  SH2movll4(sh, 3,5, 14);
}
#define init_5_3_5_15
void FASTCALL call_5_3_5_15(SH2_struct* sh)
{
  SH2movll4(sh, 3,5, 15);
}
#define init_5_3_6_0
void FASTCALL call_5_3_6_0(SH2_struct* sh)
{
  SH2movll4(sh, 3,6, 0);
}
#define init_5_3_6_1
void FASTCALL call_5_3_6_1(SH2_struct* sh)
{
  SH2movll4(sh, 3,6, 1);
}
#define init_5_3_6_2
void FASTCALL call_5_3_6_2(SH2_struct* sh)
{
  SH2movll4(sh, 3,6, 2);
}
#define init_5_3_6_3
void FASTCALL call_5_3_6_3(SH2_struct* sh)
{
  SH2movll4(sh, 3,6, 3);
}
#define init_5_3_6_4
void FASTCALL call_5_3_6_4(SH2_struct* sh)
{
  SH2movll4(sh, 3,6, 4);
}
#define init_5_3_6_5
void FASTCALL call_5_3_6_5(SH2_struct* sh)
{
  SH2movll4(sh, 3,6, 5);
}
#define init_5_3_6_6
void FASTCALL call_5_3_6_6(SH2_struct* sh)
{
  SH2movll4(sh, 3,6, 6);
}
#define init_5_3_6_7
void FASTCALL call_5_3_6_7(SH2_struct* sh)
{
  SH2movll4(sh, 3,6, 7);
}
#define init_5_3_6_8
void FASTCALL call_5_3_6_8(SH2_struct* sh)
{
  SH2movll4(sh, 3,6, 8);
}
#define init_5_3_6_9
void FASTCALL call_5_3_6_9(SH2_struct* sh)
{
  SH2movll4(sh, 3,6, 9);
}
#define init_5_3_6_10
void FASTCALL call_5_3_6_10(SH2_struct* sh)
{
  SH2movll4(sh, 3,6, 10);
}
#define init_5_3_6_11
void FASTCALL call_5_3_6_11(SH2_struct* sh)
{
  SH2movll4(sh, 3,6, 11);
}
#define init_5_3_6_12
void FASTCALL call_5_3_6_12(SH2_struct* sh)
{
  SH2movll4(sh, 3,6, 12);
}
#define init_5_3_6_13
void FASTCALL call_5_3_6_13(SH2_struct* sh)
{
  SH2movll4(sh, 3,6, 13);
}
#define init_5_3_6_14
void FASTCALL call_5_3_6_14(SH2_struct* sh)
{
  SH2movll4(sh, 3,6, 14);
}
#define init_5_3_6_15
void FASTCALL call_5_3_6_15(SH2_struct* sh)
{
  SH2movll4(sh, 3,6, 15);
}
#define init_5_3_7_0
void FASTCALL call_5_3_7_0(SH2_struct* sh)
{
  SH2movll4(sh, 3,7, 0);
}
#define init_5_3_7_1
void FASTCALL call_5_3_7_1(SH2_struct* sh)
{
  SH2movll4(sh, 3,7, 1);
}
#define init_5_3_7_2
void FASTCALL call_5_3_7_2(SH2_struct* sh)
{
  SH2movll4(sh, 3,7, 2);
}
#define init_5_3_7_3
void FASTCALL call_5_3_7_3(SH2_struct* sh)
{
  SH2movll4(sh, 3,7, 3);
}
#define init_5_3_7_4
void FASTCALL call_5_3_7_4(SH2_struct* sh)
{
  SH2movll4(sh, 3,7, 4);
}
#define init_5_3_7_5
void FASTCALL call_5_3_7_5(SH2_struct* sh)
{
  SH2movll4(sh, 3,7, 5);
}
#define init_5_3_7_6
void FASTCALL call_5_3_7_6(SH2_struct* sh)
{
  SH2movll4(sh, 3,7, 6);
}
#define init_5_3_7_7
void FASTCALL call_5_3_7_7(SH2_struct* sh)
{
  SH2movll4(sh, 3,7, 7);
}
#define init_5_3_7_8
void FASTCALL call_5_3_7_8(SH2_struct* sh)
{
  SH2movll4(sh, 3,7, 8);
}
#define init_5_3_7_9
void FASTCALL call_5_3_7_9(SH2_struct* sh)
{
  SH2movll4(sh, 3,7, 9);
}
#define init_5_3_7_10
void FASTCALL call_5_3_7_10(SH2_struct* sh)
{
  SH2movll4(sh, 3,7, 10);
}
#define init_5_3_7_11
void FASTCALL call_5_3_7_11(SH2_struct* sh)
{
  SH2movll4(sh, 3,7, 11);
}
#define init_5_3_7_12
void FASTCALL call_5_3_7_12(SH2_struct* sh)
{
  SH2movll4(sh, 3,7, 12);
}
#define init_5_3_7_13
void FASTCALL call_5_3_7_13(SH2_struct* sh)
{
  SH2movll4(sh, 3,7, 13);
}
#define init_5_3_7_14
void FASTCALL call_5_3_7_14(SH2_struct* sh)
{
  SH2movll4(sh, 3,7, 14);
}
#define init_5_3_7_15
void FASTCALL call_5_3_7_15(SH2_struct* sh)
{
  SH2movll4(sh, 3,7, 15);
}
#define init_5_3_8_0
void FASTCALL call_5_3_8_0(SH2_struct* sh)
{
  SH2movll4(sh, 3,8, 0);
}
#define init_5_3_8_1
void FASTCALL call_5_3_8_1(SH2_struct* sh)
{
  SH2movll4(sh, 3,8, 1);
}
#define init_5_3_8_2
void FASTCALL call_5_3_8_2(SH2_struct* sh)
{
  SH2movll4(sh, 3,8, 2);
}
#define init_5_3_8_3
void FASTCALL call_5_3_8_3(SH2_struct* sh)
{
  SH2movll4(sh, 3,8, 3);
}
#define init_5_3_8_4
void FASTCALL call_5_3_8_4(SH2_struct* sh)
{
  SH2movll4(sh, 3,8, 4);
}
#define init_5_3_8_5
void FASTCALL call_5_3_8_5(SH2_struct* sh)
{
  SH2movll4(sh, 3,8, 5);
}
#define init_5_3_8_6
void FASTCALL call_5_3_8_6(SH2_struct* sh)
{
  SH2movll4(sh, 3,8, 6);
}
#define init_5_3_8_7
void FASTCALL call_5_3_8_7(SH2_struct* sh)
{
  SH2movll4(sh, 3,8, 7);
}
#define init_5_3_8_8
void FASTCALL call_5_3_8_8(SH2_struct* sh)
{
  SH2movll4(sh, 3,8, 8);
}
#define init_5_3_8_9
void FASTCALL call_5_3_8_9(SH2_struct* sh)
{
  SH2movll4(sh, 3,8, 9);
}
#define init_5_3_8_10
void FASTCALL call_5_3_8_10(SH2_struct* sh)
{
  SH2movll4(sh, 3,8, 10);
}
#define init_5_3_8_11
void FASTCALL call_5_3_8_11(SH2_struct* sh)
{
  SH2movll4(sh, 3,8, 11);
}
#define init_5_3_8_12
void FASTCALL call_5_3_8_12(SH2_struct* sh)
{
  SH2movll4(sh, 3,8, 12);
}
#define init_5_3_8_13
void FASTCALL call_5_3_8_13(SH2_struct* sh)
{
  SH2movll4(sh, 3,8, 13);
}
#define init_5_3_8_14
void FASTCALL call_5_3_8_14(SH2_struct* sh)
{
  SH2movll4(sh, 3,8, 14);
}
#define init_5_3_8_15
void FASTCALL call_5_3_8_15(SH2_struct* sh)
{
  SH2movll4(sh, 3,8, 15);
}
#define init_5_3_9_0
void FASTCALL call_5_3_9_0(SH2_struct* sh)
{
  SH2movll4(sh, 3,9, 0);
}
#define init_5_3_9_1
void FASTCALL call_5_3_9_1(SH2_struct* sh)
{
  SH2movll4(sh, 3,9, 1);
}
#define init_5_3_9_2
void FASTCALL call_5_3_9_2(SH2_struct* sh)
{
  SH2movll4(sh, 3,9, 2);
}
#define init_5_3_9_3
void FASTCALL call_5_3_9_3(SH2_struct* sh)
{
  SH2movll4(sh, 3,9, 3);
}
#define init_5_3_9_4
void FASTCALL call_5_3_9_4(SH2_struct* sh)
{
  SH2movll4(sh, 3,9, 4);
}
#define init_5_3_9_5
void FASTCALL call_5_3_9_5(SH2_struct* sh)
{
  SH2movll4(sh, 3,9, 5);
}
#define init_5_3_9_6
void FASTCALL call_5_3_9_6(SH2_struct* sh)
{
  SH2movll4(sh, 3,9, 6);
}
#define init_5_3_9_7
void FASTCALL call_5_3_9_7(SH2_struct* sh)
{
  SH2movll4(sh, 3,9, 7);
}
#define init_5_3_9_8
void FASTCALL call_5_3_9_8(SH2_struct* sh)
{
  SH2movll4(sh, 3,9, 8);
}
#define init_5_3_9_9
void FASTCALL call_5_3_9_9(SH2_struct* sh)
{
  SH2movll4(sh, 3,9, 9);
}
#define init_5_3_9_10
void FASTCALL call_5_3_9_10(SH2_struct* sh)
{
  SH2movll4(sh, 3,9, 10);
}
#define init_5_3_9_11
void FASTCALL call_5_3_9_11(SH2_struct* sh)
{
  SH2movll4(sh, 3,9, 11);
}
#define init_5_3_9_12
void FASTCALL call_5_3_9_12(SH2_struct* sh)
{
  SH2movll4(sh, 3,9, 12);
}
#define init_5_3_9_13
void FASTCALL call_5_3_9_13(SH2_struct* sh)
{
  SH2movll4(sh, 3,9, 13);
}
#define init_5_3_9_14
void FASTCALL call_5_3_9_14(SH2_struct* sh)
{
  SH2movll4(sh, 3,9, 14);
}
#define init_5_3_9_15
void FASTCALL call_5_3_9_15(SH2_struct* sh)
{
  SH2movll4(sh, 3,9, 15);
}
#define init_5_3_10_0
void FASTCALL call_5_3_10_0(SH2_struct* sh)
{
  SH2movll4(sh, 3,10, 0);
}
#define init_5_3_10_1
void FASTCALL call_5_3_10_1(SH2_struct* sh)
{
  SH2movll4(sh, 3,10, 1);
}
#define init_5_3_10_2
void FASTCALL call_5_3_10_2(SH2_struct* sh)
{
  SH2movll4(sh, 3,10, 2);
}
#define init_5_3_10_3
void FASTCALL call_5_3_10_3(SH2_struct* sh)
{
  SH2movll4(sh, 3,10, 3);
}
#define init_5_3_10_4
void FASTCALL call_5_3_10_4(SH2_struct* sh)
{
  SH2movll4(sh, 3,10, 4);
}
#define init_5_3_10_5
void FASTCALL call_5_3_10_5(SH2_struct* sh)
{
  SH2movll4(sh, 3,10, 5);
}
#define init_5_3_10_6
void FASTCALL call_5_3_10_6(SH2_struct* sh)
{
  SH2movll4(sh, 3,10, 6);
}
#define init_5_3_10_7
void FASTCALL call_5_3_10_7(SH2_struct* sh)
{
  SH2movll4(sh, 3,10, 7);
}
#define init_5_3_10_8
void FASTCALL call_5_3_10_8(SH2_struct* sh)
{
  SH2movll4(sh, 3,10, 8);
}
#define init_5_3_10_9
void FASTCALL call_5_3_10_9(SH2_struct* sh)
{
  SH2movll4(sh, 3,10, 9);
}
#define init_5_3_10_10
void FASTCALL call_5_3_10_10(SH2_struct* sh)
{
  SH2movll4(sh, 3,10, 10);
}
#define init_5_3_10_11
void FASTCALL call_5_3_10_11(SH2_struct* sh)
{
  SH2movll4(sh, 3,10, 11);
}
#define init_5_3_10_12
void FASTCALL call_5_3_10_12(SH2_struct* sh)
{
  SH2movll4(sh, 3,10, 12);
}
#define init_5_3_10_13
void FASTCALL call_5_3_10_13(SH2_struct* sh)
{
  SH2movll4(sh, 3,10, 13);
}
#define init_5_3_10_14
void FASTCALL call_5_3_10_14(SH2_struct* sh)
{
  SH2movll4(sh, 3,10, 14);
}
#define init_5_3_10_15
void FASTCALL call_5_3_10_15(SH2_struct* sh)
{
  SH2movll4(sh, 3,10, 15);
}
#define init_5_3_11_0
void FASTCALL call_5_3_11_0(SH2_struct* sh)
{
  SH2movll4(sh, 3,11, 0);
}
#define init_5_3_11_1
void FASTCALL call_5_3_11_1(SH2_struct* sh)
{
  SH2movll4(sh, 3,11, 1);
}
#define init_5_3_11_2
void FASTCALL call_5_3_11_2(SH2_struct* sh)
{
  SH2movll4(sh, 3,11, 2);
}
#define init_5_3_11_3
void FASTCALL call_5_3_11_3(SH2_struct* sh)
{
  SH2movll4(sh, 3,11, 3);
}
#define init_5_3_11_4
void FASTCALL call_5_3_11_4(SH2_struct* sh)
{
  SH2movll4(sh, 3,11, 4);
}
#define init_5_3_11_5
void FASTCALL call_5_3_11_5(SH2_struct* sh)
{
  SH2movll4(sh, 3,11, 5);
}
#define init_5_3_11_6
void FASTCALL call_5_3_11_6(SH2_struct* sh)
{
  SH2movll4(sh, 3,11, 6);
}
#define init_5_3_11_7
void FASTCALL call_5_3_11_7(SH2_struct* sh)
{
  SH2movll4(sh, 3,11, 7);
}
#define init_5_3_11_8
void FASTCALL call_5_3_11_8(SH2_struct* sh)
{
  SH2movll4(sh, 3,11, 8);
}
#define init_5_3_11_9
void FASTCALL call_5_3_11_9(SH2_struct* sh)
{
  SH2movll4(sh, 3,11, 9);
}
#define init_5_3_11_10
void FASTCALL call_5_3_11_10(SH2_struct* sh)
{
  SH2movll4(sh, 3,11, 10);
}
#define init_5_3_11_11
void FASTCALL call_5_3_11_11(SH2_struct* sh)
{
  SH2movll4(sh, 3,11, 11);
}
#define init_5_3_11_12
void FASTCALL call_5_3_11_12(SH2_struct* sh)
{
  SH2movll4(sh, 3,11, 12);
}
#define init_5_3_11_13
void FASTCALL call_5_3_11_13(SH2_struct* sh)
{
  SH2movll4(sh, 3,11, 13);
}
#define init_5_3_11_14
void FASTCALL call_5_3_11_14(SH2_struct* sh)
{
  SH2movll4(sh, 3,11, 14);
}
#define init_5_3_11_15
void FASTCALL call_5_3_11_15(SH2_struct* sh)
{
  SH2movll4(sh, 3,11, 15);
}
#define init_5_3_12_0
void FASTCALL call_5_3_12_0(SH2_struct* sh)
{
  SH2movll4(sh, 3,12, 0);
}
#define init_5_3_12_1
void FASTCALL call_5_3_12_1(SH2_struct* sh)
{
  SH2movll4(sh, 3,12, 1);
}
#define init_5_3_12_2
void FASTCALL call_5_3_12_2(SH2_struct* sh)
{
  SH2movll4(sh, 3,12, 2);
}
#define init_5_3_12_3
void FASTCALL call_5_3_12_3(SH2_struct* sh)
{
  SH2movll4(sh, 3,12, 3);
}
#define init_5_3_12_4
void FASTCALL call_5_3_12_4(SH2_struct* sh)
{
  SH2movll4(sh, 3,12, 4);
}
#define init_5_3_12_5
void FASTCALL call_5_3_12_5(SH2_struct* sh)
{
  SH2movll4(sh, 3,12, 5);
}
#define init_5_3_12_6
void FASTCALL call_5_3_12_6(SH2_struct* sh)
{
  SH2movll4(sh, 3,12, 6);
}
#define init_5_3_12_7
void FASTCALL call_5_3_12_7(SH2_struct* sh)
{
  SH2movll4(sh, 3,12, 7);
}
#define init_5_3_12_8
void FASTCALL call_5_3_12_8(SH2_struct* sh)
{
  SH2movll4(sh, 3,12, 8);
}
#define init_5_3_12_9
void FASTCALL call_5_3_12_9(SH2_struct* sh)
{
  SH2movll4(sh, 3,12, 9);
}
#define init_5_3_12_10
void FASTCALL call_5_3_12_10(SH2_struct* sh)
{
  SH2movll4(sh, 3,12, 10);
}
#define init_5_3_12_11
void FASTCALL call_5_3_12_11(SH2_struct* sh)
{
  SH2movll4(sh, 3,12, 11);
}
#define init_5_3_12_12
void FASTCALL call_5_3_12_12(SH2_struct* sh)
{
  SH2movll4(sh, 3,12, 12);
}
#define init_5_3_12_13
void FASTCALL call_5_3_12_13(SH2_struct* sh)
{
  SH2movll4(sh, 3,12, 13);
}
#define init_5_3_12_14
void FASTCALL call_5_3_12_14(SH2_struct* sh)
{
  SH2movll4(sh, 3,12, 14);
}
#define init_5_3_12_15
void FASTCALL call_5_3_12_15(SH2_struct* sh)
{
  SH2movll4(sh, 3,12, 15);
}
#define init_5_3_13_0
void FASTCALL call_5_3_13_0(SH2_struct* sh)
{
  SH2movll4(sh, 3,13, 0);
}
#define init_5_3_13_1
void FASTCALL call_5_3_13_1(SH2_struct* sh)
{
  SH2movll4(sh, 3,13, 1);
}
#define init_5_3_13_2
void FASTCALL call_5_3_13_2(SH2_struct* sh)
{
  SH2movll4(sh, 3,13, 2);
}
#define init_5_3_13_3
void FASTCALL call_5_3_13_3(SH2_struct* sh)
{
  SH2movll4(sh, 3,13, 3);
}
#define init_5_3_13_4
void FASTCALL call_5_3_13_4(SH2_struct* sh)
{
  SH2movll4(sh, 3,13, 4);
}
#define init_5_3_13_5
void FASTCALL call_5_3_13_5(SH2_struct* sh)
{
  SH2movll4(sh, 3,13, 5);
}
#define init_5_3_13_6
void FASTCALL call_5_3_13_6(SH2_struct* sh)
{
  SH2movll4(sh, 3,13, 6);
}
#define init_5_3_13_7
void FASTCALL call_5_3_13_7(SH2_struct* sh)
{
  SH2movll4(sh, 3,13, 7);
}
#define init_5_3_13_8
void FASTCALL call_5_3_13_8(SH2_struct* sh)
{
  SH2movll4(sh, 3,13, 8);
}
#define init_5_3_13_9
void FASTCALL call_5_3_13_9(SH2_struct* sh)
{
  SH2movll4(sh, 3,13, 9);
}
#define init_5_3_13_10
void FASTCALL call_5_3_13_10(SH2_struct* sh)
{
  SH2movll4(sh, 3,13, 10);
}
#define init_5_3_13_11
void FASTCALL call_5_3_13_11(SH2_struct* sh)
{
  SH2movll4(sh, 3,13, 11);
}
#define init_5_3_13_12
void FASTCALL call_5_3_13_12(SH2_struct* sh)
{
  SH2movll4(sh, 3,13, 12);
}
#define init_5_3_13_13
void FASTCALL call_5_3_13_13(SH2_struct* sh)
{
  SH2movll4(sh, 3,13, 13);
}
#define init_5_3_13_14
void FASTCALL call_5_3_13_14(SH2_struct* sh)
{
  SH2movll4(sh, 3,13, 14);
}
#define init_5_3_13_15
void FASTCALL call_5_3_13_15(SH2_struct* sh)
{
  SH2movll4(sh, 3,13, 15);
}
#define init_5_3_14_0
void FASTCALL call_5_3_14_0(SH2_struct* sh)
{
  SH2movll4(sh, 3,14, 0);
}
#define init_5_3_14_1
void FASTCALL call_5_3_14_1(SH2_struct* sh)
{
  SH2movll4(sh, 3,14, 1);
}
#define init_5_3_14_2
void FASTCALL call_5_3_14_2(SH2_struct* sh)
{
  SH2movll4(sh, 3,14, 2);
}
#define init_5_3_14_3
void FASTCALL call_5_3_14_3(SH2_struct* sh)
{
  SH2movll4(sh, 3,14, 3);
}
#define init_5_3_14_4
void FASTCALL call_5_3_14_4(SH2_struct* sh)
{
  SH2movll4(sh, 3,14, 4);
}
#define init_5_3_14_5
void FASTCALL call_5_3_14_5(SH2_struct* sh)
{
  SH2movll4(sh, 3,14, 5);
}
#define init_5_3_14_6
void FASTCALL call_5_3_14_6(SH2_struct* sh)
{
  SH2movll4(sh, 3,14, 6);
}
#define init_5_3_14_7
void FASTCALL call_5_3_14_7(SH2_struct* sh)
{
  SH2movll4(sh, 3,14, 7);
}
#define init_5_3_14_8
void FASTCALL call_5_3_14_8(SH2_struct* sh)
{
  SH2movll4(sh, 3,14, 8);
}
#define init_5_3_14_9
void FASTCALL call_5_3_14_9(SH2_struct* sh)
{
  SH2movll4(sh, 3,14, 9);
}
#define init_5_3_14_10
void FASTCALL call_5_3_14_10(SH2_struct* sh)
{
  SH2movll4(sh, 3,14, 10);
}
#define init_5_3_14_11
void FASTCALL call_5_3_14_11(SH2_struct* sh)
{
  SH2movll4(sh, 3,14, 11);
}
#define init_5_3_14_12
void FASTCALL call_5_3_14_12(SH2_struct* sh)
{
  SH2movll4(sh, 3,14, 12);
}
#define init_5_3_14_13
void FASTCALL call_5_3_14_13(SH2_struct* sh)
{
  SH2movll4(sh, 3,14, 13);
}
#define init_5_3_14_14
void FASTCALL call_5_3_14_14(SH2_struct* sh)
{
  SH2movll4(sh, 3,14, 14);
}
#define init_5_3_14_15
void FASTCALL call_5_3_14_15(SH2_struct* sh)
{
  SH2movll4(sh, 3,14, 15);
}
#define init_5_3_15_0
void FASTCALL call_5_3_15_0(SH2_struct* sh)
{
  SH2movll4(sh, 3,15, 0);
}
#define init_5_3_15_1
void FASTCALL call_5_3_15_1(SH2_struct* sh)
{
  SH2movll4(sh, 3,15, 1);
}
#define init_5_3_15_2
void FASTCALL call_5_3_15_2(SH2_struct* sh)
{
  SH2movll4(sh, 3,15, 2);
}
#define init_5_3_15_3
void FASTCALL call_5_3_15_3(SH2_struct* sh)
{
  SH2movll4(sh, 3,15, 3);
}
#define init_5_3_15_4
void FASTCALL call_5_3_15_4(SH2_struct* sh)
{
  SH2movll4(sh, 3,15, 4);
}
#define init_5_3_15_5
void FASTCALL call_5_3_15_5(SH2_struct* sh)
{
  SH2movll4(sh, 3,15, 5);
}
#define init_5_3_15_6
void FASTCALL call_5_3_15_6(SH2_struct* sh)
{
  SH2movll4(sh, 3,15, 6);
}
#define init_5_3_15_7
void FASTCALL call_5_3_15_7(SH2_struct* sh)
{
  SH2movll4(sh, 3,15, 7);
}
#define init_5_3_15_8
void FASTCALL call_5_3_15_8(SH2_struct* sh)
{
  SH2movll4(sh, 3,15, 8);
}
#define init_5_3_15_9
void FASTCALL call_5_3_15_9(SH2_struct* sh)
{
  SH2movll4(sh, 3,15, 9);
}
#define init_5_3_15_10
void FASTCALL call_5_3_15_10(SH2_struct* sh)
{
  SH2movll4(sh, 3,15, 10);
}
#define init_5_3_15_11
void FASTCALL call_5_3_15_11(SH2_struct* sh)
{
  SH2movll4(sh, 3,15, 11);
}
#define init_5_3_15_12
void FASTCALL call_5_3_15_12(SH2_struct* sh)
{
  SH2movll4(sh, 3,15, 12);
}
#define init_5_3_15_13
void FASTCALL call_5_3_15_13(SH2_struct* sh)
{
  SH2movll4(sh, 3,15, 13);
}
#define init_5_3_15_14
void FASTCALL call_5_3_15_14(SH2_struct* sh)
{
  SH2movll4(sh, 3,15, 14);
}
#define init_5_3_15_15
void FASTCALL call_5_3_15_15(SH2_struct* sh)
{
  SH2movll4(sh, 3,15, 15);
}
#define init_5_4_0_0
void FASTCALL call_5_4_0_0(SH2_struct* sh)
{
  SH2movll4(sh, 4,0, 0);
}
#define init_5_4_0_1
void FASTCALL call_5_4_0_1(SH2_struct* sh)
{
  SH2movll4(sh, 4,0, 1);
}
#define init_5_4_0_2
void FASTCALL call_5_4_0_2(SH2_struct* sh)
{
  SH2movll4(sh, 4,0, 2);
}
#define init_5_4_0_3
void FASTCALL call_5_4_0_3(SH2_struct* sh)
{
  SH2movll4(sh, 4,0, 3);
}
#define init_5_4_0_4
void FASTCALL call_5_4_0_4(SH2_struct* sh)
{
  SH2movll4(sh, 4,0, 4);
}
#define init_5_4_0_5
void FASTCALL call_5_4_0_5(SH2_struct* sh)
{
  SH2movll4(sh, 4,0, 5);
}
#define init_5_4_0_6
void FASTCALL call_5_4_0_6(SH2_struct* sh)
{
  SH2movll4(sh, 4,0, 6);
}
#define init_5_4_0_7
void FASTCALL call_5_4_0_7(SH2_struct* sh)
{
  SH2movll4(sh, 4,0, 7);
}
#define init_5_4_0_8
void FASTCALL call_5_4_0_8(SH2_struct* sh)
{
  SH2movll4(sh, 4,0, 8);
}
#define init_5_4_0_9
void FASTCALL call_5_4_0_9(SH2_struct* sh)
{
  SH2movll4(sh, 4,0, 9);
}
#define init_5_4_0_10
void FASTCALL call_5_4_0_10(SH2_struct* sh)
{
  SH2movll4(sh, 4,0, 10);
}
#define init_5_4_0_11
void FASTCALL call_5_4_0_11(SH2_struct* sh)
{
  SH2movll4(sh, 4,0, 11);
}
#define init_5_4_0_12
void FASTCALL call_5_4_0_12(SH2_struct* sh)
{
  SH2movll4(sh, 4,0, 12);
}
#define init_5_4_0_13
void FASTCALL call_5_4_0_13(SH2_struct* sh)
{
  SH2movll4(sh, 4,0, 13);
}
#define init_5_4_0_14
void FASTCALL call_5_4_0_14(SH2_struct* sh)
{
  SH2movll4(sh, 4,0, 14);
}
#define init_5_4_0_15
void FASTCALL call_5_4_0_15(SH2_struct* sh)
{
  SH2movll4(sh, 4,0, 15);
}
#define init_5_4_1_0
void FASTCALL call_5_4_1_0(SH2_struct* sh)
{
  SH2movll4(sh, 4,1, 0);
}
#define init_5_4_1_1
void FASTCALL call_5_4_1_1(SH2_struct* sh)
{
  SH2movll4(sh, 4,1, 1);
}
#define init_5_4_1_2
void FASTCALL call_5_4_1_2(SH2_struct* sh)
{
  SH2movll4(sh, 4,1, 2);
}
#define init_5_4_1_3
void FASTCALL call_5_4_1_3(SH2_struct* sh)
{
  SH2movll4(sh, 4,1, 3);
}
#define init_5_4_1_4
void FASTCALL call_5_4_1_4(SH2_struct* sh)
{
  SH2movll4(sh, 4,1, 4);
}
#define init_5_4_1_5
void FASTCALL call_5_4_1_5(SH2_struct* sh)
{
  SH2movll4(sh, 4,1, 5);
}
#define init_5_4_1_6
void FASTCALL call_5_4_1_6(SH2_struct* sh)
{
  SH2movll4(sh, 4,1, 6);
}
#define init_5_4_1_7
void FASTCALL call_5_4_1_7(SH2_struct* sh)
{
  SH2movll4(sh, 4,1, 7);
}
#define init_5_4_1_8
void FASTCALL call_5_4_1_8(SH2_struct* sh)
{
  SH2movll4(sh, 4,1, 8);
}
#define init_5_4_1_9
void FASTCALL call_5_4_1_9(SH2_struct* sh)
{
  SH2movll4(sh, 4,1, 9);
}
#define init_5_4_1_10
void FASTCALL call_5_4_1_10(SH2_struct* sh)
{
  SH2movll4(sh, 4,1, 10);
}
#define init_5_4_1_11
void FASTCALL call_5_4_1_11(SH2_struct* sh)
{
  SH2movll4(sh, 4,1, 11);
}
#define init_5_4_1_12
void FASTCALL call_5_4_1_12(SH2_struct* sh)
{
  SH2movll4(sh, 4,1, 12);
}
#define init_5_4_1_13
void FASTCALL call_5_4_1_13(SH2_struct* sh)
{
  SH2movll4(sh, 4,1, 13);
}
#define init_5_4_1_14
void FASTCALL call_5_4_1_14(SH2_struct* sh)
{
  SH2movll4(sh, 4,1, 14);
}
#define init_5_4_1_15
void FASTCALL call_5_4_1_15(SH2_struct* sh)
{
  SH2movll4(sh, 4,1, 15);
}
#define init_5_4_2_0
void FASTCALL call_5_4_2_0(SH2_struct* sh)
{
  SH2movll4(sh, 4,2, 0);
}
#define init_5_4_2_1
void FASTCALL call_5_4_2_1(SH2_struct* sh)
{
  SH2movll4(sh, 4,2, 1);
}
#define init_5_4_2_2
void FASTCALL call_5_4_2_2(SH2_struct* sh)
{
  SH2movll4(sh, 4,2, 2);
}
#define init_5_4_2_3
void FASTCALL call_5_4_2_3(SH2_struct* sh)
{
  SH2movll4(sh, 4,2, 3);
}
#define init_5_4_2_4
void FASTCALL call_5_4_2_4(SH2_struct* sh)
{
  SH2movll4(sh, 4,2, 4);
}
#define init_5_4_2_5
void FASTCALL call_5_4_2_5(SH2_struct* sh)
{
  SH2movll4(sh, 4,2, 5);
}
#define init_5_4_2_6
void FASTCALL call_5_4_2_6(SH2_struct* sh)
{
  SH2movll4(sh, 4,2, 6);
}
#define init_5_4_2_7
void FASTCALL call_5_4_2_7(SH2_struct* sh)
{
  SH2movll4(sh, 4,2, 7);
}
#define init_5_4_2_8
void FASTCALL call_5_4_2_8(SH2_struct* sh)
{
  SH2movll4(sh, 4,2, 8);
}
#define init_5_4_2_9
void FASTCALL call_5_4_2_9(SH2_struct* sh)
{
  SH2movll4(sh, 4,2, 9);
}
#define init_5_4_2_10
void FASTCALL call_5_4_2_10(SH2_struct* sh)
{
  SH2movll4(sh, 4,2, 10);
}
#define init_5_4_2_11
void FASTCALL call_5_4_2_11(SH2_struct* sh)
{
  SH2movll4(sh, 4,2, 11);
}
#define init_5_4_2_12
void FASTCALL call_5_4_2_12(SH2_struct* sh)
{
  SH2movll4(sh, 4,2, 12);
}
#define init_5_4_2_13
void FASTCALL call_5_4_2_13(SH2_struct* sh)
{
  SH2movll4(sh, 4,2, 13);
}
#define init_5_4_2_14
void FASTCALL call_5_4_2_14(SH2_struct* sh)
{
  SH2movll4(sh, 4,2, 14);
}
#define init_5_4_2_15
void FASTCALL call_5_4_2_15(SH2_struct* sh)
{
  SH2movll4(sh, 4,2, 15);
}
#define init_5_4_3_0
void FASTCALL call_5_4_3_0(SH2_struct* sh)
{
  SH2movll4(sh, 4,3, 0);
}
#define init_5_4_3_1
void FASTCALL call_5_4_3_1(SH2_struct* sh)
{
  SH2movll4(sh, 4,3, 1);
}
#define init_5_4_3_2
void FASTCALL call_5_4_3_2(SH2_struct* sh)
{
  SH2movll4(sh, 4,3, 2);
}
#define init_5_4_3_3
void FASTCALL call_5_4_3_3(SH2_struct* sh)
{
  SH2movll4(sh, 4,3, 3);
}
#define init_5_4_3_4
void FASTCALL call_5_4_3_4(SH2_struct* sh)
{
  SH2movll4(sh, 4,3, 4);
}
#define init_5_4_3_5
void FASTCALL call_5_4_3_5(SH2_struct* sh)
{
  SH2movll4(sh, 4,3, 5);
}
#define init_5_4_3_6
void FASTCALL call_5_4_3_6(SH2_struct* sh)
{
  SH2movll4(sh, 4,3, 6);
}
#define init_5_4_3_7
void FASTCALL call_5_4_3_7(SH2_struct* sh)
{
  SH2movll4(sh, 4,3, 7);
}
#define init_5_4_3_8
void FASTCALL call_5_4_3_8(SH2_struct* sh)
{
  SH2movll4(sh, 4,3, 8);
}
#define init_5_4_3_9
void FASTCALL call_5_4_3_9(SH2_struct* sh)
{
  SH2movll4(sh, 4,3, 9);
}
#define init_5_4_3_10
void FASTCALL call_5_4_3_10(SH2_struct* sh)
{
  SH2movll4(sh, 4,3, 10);
}
#define init_5_4_3_11
void FASTCALL call_5_4_3_11(SH2_struct* sh)
{
  SH2movll4(sh, 4,3, 11);
}
#define init_5_4_3_12
void FASTCALL call_5_4_3_12(SH2_struct* sh)
{
  SH2movll4(sh, 4,3, 12);
}
#define init_5_4_3_13
void FASTCALL call_5_4_3_13(SH2_struct* sh)
{
  SH2movll4(sh, 4,3, 13);
}
#define init_5_4_3_14
void FASTCALL call_5_4_3_14(SH2_struct* sh)
{
  SH2movll4(sh, 4,3, 14);
}
#define init_5_4_3_15
void FASTCALL call_5_4_3_15(SH2_struct* sh)
{
  SH2movll4(sh, 4,3, 15);
}
#define init_5_4_4_0
void FASTCALL call_5_4_4_0(SH2_struct* sh)
{
  SH2movll4(sh, 4,4, 0);
}
#define init_5_4_4_1
void FASTCALL call_5_4_4_1(SH2_struct* sh)
{
  SH2movll4(sh, 4,4, 1);
}
#define init_5_4_4_2
void FASTCALL call_5_4_4_2(SH2_struct* sh)
{
  SH2movll4(sh, 4,4, 2);
}
#define init_5_4_4_3
void FASTCALL call_5_4_4_3(SH2_struct* sh)
{
  SH2movll4(sh, 4,4, 3);
}
#define init_5_4_4_4
void FASTCALL call_5_4_4_4(SH2_struct* sh)
{
  SH2movll4(sh, 4,4, 4);
}
#define init_5_4_4_5
void FASTCALL call_5_4_4_5(SH2_struct* sh)
{
  SH2movll4(sh, 4,4, 5);
}
#define init_5_4_4_6
void FASTCALL call_5_4_4_6(SH2_struct* sh)
{
  SH2movll4(sh, 4,4, 6);
}
#define init_5_4_4_7
void FASTCALL call_5_4_4_7(SH2_struct* sh)
{
  SH2movll4(sh, 4,4, 7);
}
#define init_5_4_4_8
void FASTCALL call_5_4_4_8(SH2_struct* sh)
{
  SH2movll4(sh, 4,4, 8);
}
#define init_5_4_4_9
void FASTCALL call_5_4_4_9(SH2_struct* sh)
{
  SH2movll4(sh, 4,4, 9);
}
#define init_5_4_4_10
void FASTCALL call_5_4_4_10(SH2_struct* sh)
{
  SH2movll4(sh, 4,4, 10);
}
#define init_5_4_4_11
void FASTCALL call_5_4_4_11(SH2_struct* sh)
{
  SH2movll4(sh, 4,4, 11);
}
#define init_5_4_4_12
void FASTCALL call_5_4_4_12(SH2_struct* sh)
{
  SH2movll4(sh, 4,4, 12);
}
#define init_5_4_4_13
void FASTCALL call_5_4_4_13(SH2_struct* sh)
{
  SH2movll4(sh, 4,4, 13);
}
#define init_5_4_4_14
void FASTCALL call_5_4_4_14(SH2_struct* sh)
{
  SH2movll4(sh, 4,4, 14);
}
#define init_5_4_4_15
void FASTCALL call_5_4_4_15(SH2_struct* sh)
{
  SH2movll4(sh, 4,4, 15);
}
#define init_5_4_5_0
void FASTCALL call_5_4_5_0(SH2_struct* sh)
{
  SH2movll4(sh, 4,5, 0);
}
#define init_5_4_5_1
void FASTCALL call_5_4_5_1(SH2_struct* sh)
{
  SH2movll4(sh, 4,5, 1);
}
#define init_5_4_5_2
void FASTCALL call_5_4_5_2(SH2_struct* sh)
{
  SH2movll4(sh, 4,5, 2);
}
#define init_5_4_5_3
void FASTCALL call_5_4_5_3(SH2_struct* sh)
{
  SH2movll4(sh, 4,5, 3);
}
#define init_5_4_5_4
void FASTCALL call_5_4_5_4(SH2_struct* sh)
{
  SH2movll4(sh, 4,5, 4);
}
#define init_5_4_5_5
void FASTCALL call_5_4_5_5(SH2_struct* sh)
{
  SH2movll4(sh, 4,5, 5);
}
#define init_5_4_5_6
void FASTCALL call_5_4_5_6(SH2_struct* sh)
{
  SH2movll4(sh, 4,5, 6);
}
#define init_5_4_5_7
void FASTCALL call_5_4_5_7(SH2_struct* sh)
{
  SH2movll4(sh, 4,5, 7);
}
#define init_5_4_5_8
void FASTCALL call_5_4_5_8(SH2_struct* sh)
{
  SH2movll4(sh, 4,5, 8);
}
#define init_5_4_5_9
void FASTCALL call_5_4_5_9(SH2_struct* sh)
{
  SH2movll4(sh, 4,5, 9);
}
#define init_5_4_5_10
void FASTCALL call_5_4_5_10(SH2_struct* sh)
{
  SH2movll4(sh, 4,5, 10);
}
#define init_5_4_5_11
void FASTCALL call_5_4_5_11(SH2_struct* sh)
{
  SH2movll4(sh, 4,5, 11);
}
#define init_5_4_5_12
void FASTCALL call_5_4_5_12(SH2_struct* sh)
{
  SH2movll4(sh, 4,5, 12);
}
#define init_5_4_5_13
void FASTCALL call_5_4_5_13(SH2_struct* sh)
{
  SH2movll4(sh, 4,5, 13);
}
#define init_5_4_5_14
void FASTCALL call_5_4_5_14(SH2_struct* sh)
{
  SH2movll4(sh, 4,5, 14);
}
#define init_5_4_5_15
void FASTCALL call_5_4_5_15(SH2_struct* sh)
{
  SH2movll4(sh, 4,5, 15);
}
#define init_5_4_6_0
void FASTCALL call_5_4_6_0(SH2_struct* sh)
{
  SH2movll4(sh, 4,6, 0);
}
#define init_5_4_6_1
void FASTCALL call_5_4_6_1(SH2_struct* sh)
{
  SH2movll4(sh, 4,6, 1);
}
#define init_5_4_6_2
void FASTCALL call_5_4_6_2(SH2_struct* sh)
{
  SH2movll4(sh, 4,6, 2);
}
#define init_5_4_6_3
void FASTCALL call_5_4_6_3(SH2_struct* sh)
{
  SH2movll4(sh, 4,6, 3);
}
#define init_5_4_6_4
void FASTCALL call_5_4_6_4(SH2_struct* sh)
{
  SH2movll4(sh, 4,6, 4);
}
#define init_5_4_6_5
void FASTCALL call_5_4_6_5(SH2_struct* sh)
{
  SH2movll4(sh, 4,6, 5);
}
#define init_5_4_6_6
void FASTCALL call_5_4_6_6(SH2_struct* sh)
{
  SH2movll4(sh, 4,6, 6);
}
#define init_5_4_6_7
void FASTCALL call_5_4_6_7(SH2_struct* sh)
{
  SH2movll4(sh, 4,6, 7);
}
#define init_5_4_6_8
void FASTCALL call_5_4_6_8(SH2_struct* sh)
{
  SH2movll4(sh, 4,6, 8);
}
#define init_5_4_6_9
void FASTCALL call_5_4_6_9(SH2_struct* sh)
{
  SH2movll4(sh, 4,6, 9);
}
#define init_5_4_6_10
void FASTCALL call_5_4_6_10(SH2_struct* sh)
{
  SH2movll4(sh, 4,6, 10);
}
#define init_5_4_6_11
void FASTCALL call_5_4_6_11(SH2_struct* sh)
{
  SH2movll4(sh, 4,6, 11);
}
#define init_5_4_6_12
void FASTCALL call_5_4_6_12(SH2_struct* sh)
{
  SH2movll4(sh, 4,6, 12);
}
#define init_5_4_6_13
void FASTCALL call_5_4_6_13(SH2_struct* sh)
{
  SH2movll4(sh, 4,6, 13);
}
#define init_5_4_6_14
void FASTCALL call_5_4_6_14(SH2_struct* sh)
{
  SH2movll4(sh, 4,6, 14);
}
#define init_5_4_6_15
void FASTCALL call_5_4_6_15(SH2_struct* sh)
{
  SH2movll4(sh, 4,6, 15);
}
#define init_5_4_7_0
void FASTCALL call_5_4_7_0(SH2_struct* sh)
{
  SH2movll4(sh, 4,7, 0);
}
#define init_5_4_7_1
void FASTCALL call_5_4_7_1(SH2_struct* sh)
{
  SH2movll4(sh, 4,7, 1);
}
#define init_5_4_7_2
void FASTCALL call_5_4_7_2(SH2_struct* sh)
{
  SH2movll4(sh, 4,7, 2);
}
#define init_5_4_7_3
void FASTCALL call_5_4_7_3(SH2_struct* sh)
{
  SH2movll4(sh, 4,7, 3);
}
#define init_5_4_7_4
void FASTCALL call_5_4_7_4(SH2_struct* sh)
{
  SH2movll4(sh, 4,7, 4);
}
#define init_5_4_7_5
void FASTCALL call_5_4_7_5(SH2_struct* sh)
{
  SH2movll4(sh, 4,7, 5);
}
#define init_5_4_7_6
void FASTCALL call_5_4_7_6(SH2_struct* sh)
{
  SH2movll4(sh, 4,7, 6);
}
#define init_5_4_7_7
void FASTCALL call_5_4_7_7(SH2_struct* sh)
{
  SH2movll4(sh, 4,7, 7);
}
#define init_5_4_7_8
void FASTCALL call_5_4_7_8(SH2_struct* sh)
{
  SH2movll4(sh, 4,7, 8);
}
#define init_5_4_7_9
void FASTCALL call_5_4_7_9(SH2_struct* sh)
{
  SH2movll4(sh, 4,7, 9);
}
#define init_5_4_7_10
void FASTCALL call_5_4_7_10(SH2_struct* sh)
{
  SH2movll4(sh, 4,7, 10);
}
#define init_5_4_7_11
void FASTCALL call_5_4_7_11(SH2_struct* sh)
{
  SH2movll4(sh, 4,7, 11);
}
#define init_5_4_7_12
void FASTCALL call_5_4_7_12(SH2_struct* sh)
{
  SH2movll4(sh, 4,7, 12);
}
#define init_5_4_7_13
void FASTCALL call_5_4_7_13(SH2_struct* sh)
{
  SH2movll4(sh, 4,7, 13);
}
#define init_5_4_7_14
void FASTCALL call_5_4_7_14(SH2_struct* sh)
{
  SH2movll4(sh, 4,7, 14);
}
#define init_5_4_7_15
void FASTCALL call_5_4_7_15(SH2_struct* sh)
{
  SH2movll4(sh, 4,7, 15);
}
#define init_5_4_8_0
void FASTCALL call_5_4_8_0(SH2_struct* sh)
{
  SH2movll4(sh, 4,8, 0);
}
#define init_5_4_8_1
void FASTCALL call_5_4_8_1(SH2_struct* sh)
{
  SH2movll4(sh, 4,8, 1);
}
#define init_5_4_8_2
void FASTCALL call_5_4_8_2(SH2_struct* sh)
{
  SH2movll4(sh, 4,8, 2);
}
#define init_5_4_8_3
void FASTCALL call_5_4_8_3(SH2_struct* sh)
{
  SH2movll4(sh, 4,8, 3);
}
#define init_5_4_8_4
void FASTCALL call_5_4_8_4(SH2_struct* sh)
{
  SH2movll4(sh, 4,8, 4);
}
#define init_5_4_8_5
void FASTCALL call_5_4_8_5(SH2_struct* sh)
{
  SH2movll4(sh, 4,8, 5);
}
#define init_5_4_8_6
void FASTCALL call_5_4_8_6(SH2_struct* sh)
{
  SH2movll4(sh, 4,8, 6);
}
#define init_5_4_8_7
void FASTCALL call_5_4_8_7(SH2_struct* sh)
{
  SH2movll4(sh, 4,8, 7);
}
#define init_5_4_8_8
void FASTCALL call_5_4_8_8(SH2_struct* sh)
{
  SH2movll4(sh, 4,8, 8);
}
#define init_5_4_8_9
void FASTCALL call_5_4_8_9(SH2_struct* sh)
{
  SH2movll4(sh, 4,8, 9);
}
#define init_5_4_8_10
void FASTCALL call_5_4_8_10(SH2_struct* sh)
{
  SH2movll4(sh, 4,8, 10);
}
#define init_5_4_8_11
void FASTCALL call_5_4_8_11(SH2_struct* sh)
{
  SH2movll4(sh, 4,8, 11);
}
#define init_5_4_8_12
void FASTCALL call_5_4_8_12(SH2_struct* sh)
{
  SH2movll4(sh, 4,8, 12);
}
#define init_5_4_8_13
void FASTCALL call_5_4_8_13(SH2_struct* sh)
{
  SH2movll4(sh, 4,8, 13);
}
#define init_5_4_8_14
void FASTCALL call_5_4_8_14(SH2_struct* sh)
{
  SH2movll4(sh, 4,8, 14);
}
#define init_5_4_8_15
void FASTCALL call_5_4_8_15(SH2_struct* sh)
{
  SH2movll4(sh, 4,8, 15);
}
#define init_5_4_9_0
void FASTCALL call_5_4_9_0(SH2_struct* sh)
{
  SH2movll4(sh, 4,9, 0);
}
#define init_5_4_9_1
void FASTCALL call_5_4_9_1(SH2_struct* sh)
{
  SH2movll4(sh, 4,9, 1);
}
#define init_5_4_9_2
void FASTCALL call_5_4_9_2(SH2_struct* sh)
{
  SH2movll4(sh, 4,9, 2);
}
#define init_5_4_9_3
void FASTCALL call_5_4_9_3(SH2_struct* sh)
{
  SH2movll4(sh, 4,9, 3);
}
#define init_5_4_9_4
void FASTCALL call_5_4_9_4(SH2_struct* sh)
{
  SH2movll4(sh, 4,9, 4);
}
#define init_5_4_9_5
void FASTCALL call_5_4_9_5(SH2_struct* sh)
{
  SH2movll4(sh, 4,9, 5);
}
#define init_5_4_9_6
void FASTCALL call_5_4_9_6(SH2_struct* sh)
{
  SH2movll4(sh, 4,9, 6);
}
#define init_5_4_9_7
void FASTCALL call_5_4_9_7(SH2_struct* sh)
{
  SH2movll4(sh, 4,9, 7);
}
#define init_5_4_9_8
void FASTCALL call_5_4_9_8(SH2_struct* sh)
{
  SH2movll4(sh, 4,9, 8);
}
#define init_5_4_9_9
void FASTCALL call_5_4_9_9(SH2_struct* sh)
{
  SH2movll4(sh, 4,9, 9);
}
#define init_5_4_9_10
void FASTCALL call_5_4_9_10(SH2_struct* sh)
{
  SH2movll4(sh, 4,9, 10);
}
#define init_5_4_9_11
void FASTCALL call_5_4_9_11(SH2_struct* sh)
{
  SH2movll4(sh, 4,9, 11);
}
#define init_5_4_9_12
void FASTCALL call_5_4_9_12(SH2_struct* sh)
{
  SH2movll4(sh, 4,9, 12);
}
#define init_5_4_9_13
void FASTCALL call_5_4_9_13(SH2_struct* sh)
{
  SH2movll4(sh, 4,9, 13);
}
#define init_5_4_9_14
void FASTCALL call_5_4_9_14(SH2_struct* sh)
{
  SH2movll4(sh, 4,9, 14);
}
#define init_5_4_9_15
void FASTCALL call_5_4_9_15(SH2_struct* sh)
{
  SH2movll4(sh, 4,9, 15);
}
#define init_5_4_10_0
void FASTCALL call_5_4_10_0(SH2_struct* sh)
{
  SH2movll4(sh, 4,10, 0);
}
#define init_5_4_10_1
void FASTCALL call_5_4_10_1(SH2_struct* sh)
{
  SH2movll4(sh, 4,10, 1);
}
#define init_5_4_10_2
void FASTCALL call_5_4_10_2(SH2_struct* sh)
{
  SH2movll4(sh, 4,10, 2);
}
#define init_5_4_10_3
void FASTCALL call_5_4_10_3(SH2_struct* sh)
{
  SH2movll4(sh, 4,10, 3);
}
#define init_5_4_10_4
void FASTCALL call_5_4_10_4(SH2_struct* sh)
{
  SH2movll4(sh, 4,10, 4);
}
#define init_5_4_10_5
void FASTCALL call_5_4_10_5(SH2_struct* sh)
{
  SH2movll4(sh, 4,10, 5);
}
#define init_5_4_10_6
void FASTCALL call_5_4_10_6(SH2_struct* sh)
{
  SH2movll4(sh, 4,10, 6);
}
#define init_5_4_10_7
void FASTCALL call_5_4_10_7(SH2_struct* sh)
{
  SH2movll4(sh, 4,10, 7);
}
#define init_5_4_10_8
void FASTCALL call_5_4_10_8(SH2_struct* sh)
{
  SH2movll4(sh, 4,10, 8);
}
#define init_5_4_10_9
void FASTCALL call_5_4_10_9(SH2_struct* sh)
{
  SH2movll4(sh, 4,10, 9);
}
#define init_5_4_10_10
void FASTCALL call_5_4_10_10(SH2_struct* sh)
{
  SH2movll4(sh, 4,10, 10);
}
#define init_5_4_10_11
void FASTCALL call_5_4_10_11(SH2_struct* sh)
{
  SH2movll4(sh, 4,10, 11);
}
#define init_5_4_10_12
void FASTCALL call_5_4_10_12(SH2_struct* sh)
{
  SH2movll4(sh, 4,10, 12);
}
#define init_5_4_10_13
void FASTCALL call_5_4_10_13(SH2_struct* sh)
{
  SH2movll4(sh, 4,10, 13);
}
#define init_5_4_10_14
void FASTCALL call_5_4_10_14(SH2_struct* sh)
{
  SH2movll4(sh, 4,10, 14);
}
#define init_5_4_10_15
void FASTCALL call_5_4_10_15(SH2_struct* sh)
{
  SH2movll4(sh, 4,10, 15);
}
#define init_5_4_11_0
void FASTCALL call_5_4_11_0(SH2_struct* sh)
{
  SH2movll4(sh, 4,11, 0);
}
#define init_5_4_11_1
void FASTCALL call_5_4_11_1(SH2_struct* sh)
{
  SH2movll4(sh, 4,11, 1);
}
#define init_5_4_11_2
void FASTCALL call_5_4_11_2(SH2_struct* sh)
{
  SH2movll4(sh, 4,11, 2);
}
#define init_5_4_11_3
void FASTCALL call_5_4_11_3(SH2_struct* sh)
{
  SH2movll4(sh, 4,11, 3);
}
#define init_5_4_11_4
void FASTCALL call_5_4_11_4(SH2_struct* sh)
{
  SH2movll4(sh, 4,11, 4);
}
#define init_5_4_11_5
void FASTCALL call_5_4_11_5(SH2_struct* sh)
{
  SH2movll4(sh, 4,11, 5);
}
#define init_5_4_11_6
void FASTCALL call_5_4_11_6(SH2_struct* sh)
{
  SH2movll4(sh, 4,11, 6);
}
#define init_5_4_11_7
void FASTCALL call_5_4_11_7(SH2_struct* sh)
{
  SH2movll4(sh, 4,11, 7);
}
#define init_5_4_11_8
void FASTCALL call_5_4_11_8(SH2_struct* sh)
{
  SH2movll4(sh, 4,11, 8);
}
#define init_5_4_11_9
void FASTCALL call_5_4_11_9(SH2_struct* sh)
{
  SH2movll4(sh, 4,11, 9);
}
#define init_5_4_11_10
void FASTCALL call_5_4_11_10(SH2_struct* sh)
{
  SH2movll4(sh, 4,11, 10);
}
#define init_5_4_11_11
void FASTCALL call_5_4_11_11(SH2_struct* sh)
{
  SH2movll4(sh, 4,11, 11);
}
#define init_5_4_11_12
void FASTCALL call_5_4_11_12(SH2_struct* sh)
{
  SH2movll4(sh, 4,11, 12);
}
#define init_5_4_11_13
void FASTCALL call_5_4_11_13(SH2_struct* sh)
{
  SH2movll4(sh, 4,11, 13);
}
#define init_5_4_11_14
void FASTCALL call_5_4_11_14(SH2_struct* sh)
{
  SH2movll4(sh, 4,11, 14);
}
#define init_5_4_11_15
void FASTCALL call_5_4_11_15(SH2_struct* sh)
{
  SH2movll4(sh, 4,11, 15);
}
#define init_5_4_12_0
void FASTCALL call_5_4_12_0(SH2_struct* sh)
{
  SH2movll4(sh, 4,12, 0);
}
#define init_5_4_12_1
void FASTCALL call_5_4_12_1(SH2_struct* sh)
{
  SH2movll4(sh, 4,12, 1);
}
#define init_5_4_12_2
void FASTCALL call_5_4_12_2(SH2_struct* sh)
{
  SH2movll4(sh, 4,12, 2);
}
#define init_5_4_12_3
void FASTCALL call_5_4_12_3(SH2_struct* sh)
{
  SH2movll4(sh, 4,12, 3);
}
#define init_5_4_12_4
void FASTCALL call_5_4_12_4(SH2_struct* sh)
{
  SH2movll4(sh, 4,12, 4);
}
#define init_5_4_12_5
void FASTCALL call_5_4_12_5(SH2_struct* sh)
{
  SH2movll4(sh, 4,12, 5);
}
#define init_5_4_12_6
void FASTCALL call_5_4_12_6(SH2_struct* sh)
{
  SH2movll4(sh, 4,12, 6);
}
#define init_5_4_12_7
void FASTCALL call_5_4_12_7(SH2_struct* sh)
{
  SH2movll4(sh, 4,12, 7);
}
#define init_5_4_12_8
void FASTCALL call_5_4_12_8(SH2_struct* sh)
{
  SH2movll4(sh, 4,12, 8);
}
#define init_5_4_12_9
void FASTCALL call_5_4_12_9(SH2_struct* sh)
{
  SH2movll4(sh, 4,12, 9);
}
#define init_5_4_12_10
void FASTCALL call_5_4_12_10(SH2_struct* sh)
{
  SH2movll4(sh, 4,12, 10);
}
#define init_5_4_12_11
void FASTCALL call_5_4_12_11(SH2_struct* sh)
{
  SH2movll4(sh, 4,12, 11);
}
#define init_5_4_12_12
void FASTCALL call_5_4_12_12(SH2_struct* sh)
{
  SH2movll4(sh, 4,12, 12);
}
#define init_5_4_12_13
void FASTCALL call_5_4_12_13(SH2_struct* sh)
{
  SH2movll4(sh, 4,12, 13);
}
#define init_5_4_12_14
void FASTCALL call_5_4_12_14(SH2_struct* sh)
{
  SH2movll4(sh, 4,12, 14);
}
#define init_5_4_12_15
void FASTCALL call_5_4_12_15(SH2_struct* sh)
{
  SH2movll4(sh, 4,12, 15);
}
#define init_5_4_13_0
void FASTCALL call_5_4_13_0(SH2_struct* sh)
{
  SH2movll4(sh, 4,13, 0);
}
#define init_5_4_13_1
void FASTCALL call_5_4_13_1(SH2_struct* sh)
{
  SH2movll4(sh, 4,13, 1);
}
#define init_5_4_13_2
void FASTCALL call_5_4_13_2(SH2_struct* sh)
{
  SH2movll4(sh, 4,13, 2);
}
#define init_5_4_13_3
void FASTCALL call_5_4_13_3(SH2_struct* sh)
{
  SH2movll4(sh, 4,13, 3);
}
#define init_5_4_13_4
void FASTCALL call_5_4_13_4(SH2_struct* sh)
{
  SH2movll4(sh, 4,13, 4);
}
#define init_5_4_13_5
void FASTCALL call_5_4_13_5(SH2_struct* sh)
{
  SH2movll4(sh, 4,13, 5);
}
#define init_5_4_13_6
void FASTCALL call_5_4_13_6(SH2_struct* sh)
{
  SH2movll4(sh, 4,13, 6);
}
#define init_5_4_13_7
void FASTCALL call_5_4_13_7(SH2_struct* sh)
{
  SH2movll4(sh, 4,13, 7);
}
#define init_5_4_13_8
void FASTCALL call_5_4_13_8(SH2_struct* sh)
{
  SH2movll4(sh, 4,13, 8);
}
#define init_5_4_13_9
void FASTCALL call_5_4_13_9(SH2_struct* sh)
{
  SH2movll4(sh, 4,13, 9);
}
#define init_5_4_13_10
void FASTCALL call_5_4_13_10(SH2_struct* sh)
{
  SH2movll4(sh, 4,13, 10);
}
#define init_5_4_13_11
void FASTCALL call_5_4_13_11(SH2_struct* sh)
{
  SH2movll4(sh, 4,13, 11);
}
#define init_5_4_13_12
void FASTCALL call_5_4_13_12(SH2_struct* sh)
{
  SH2movll4(sh, 4,13, 12);
}
#define init_5_4_13_13
void FASTCALL call_5_4_13_13(SH2_struct* sh)
{
  SH2movll4(sh, 4,13, 13);
}
#define init_5_4_13_14
void FASTCALL call_5_4_13_14(SH2_struct* sh)
{
  SH2movll4(sh, 4,13, 14);
}
#define init_5_4_13_15
void FASTCALL call_5_4_13_15(SH2_struct* sh)
{
  SH2movll4(sh, 4,13, 15);
}
#define init_5_4_14_0
void FASTCALL call_5_4_14_0(SH2_struct* sh)
{
  SH2movll4(sh, 4,14, 0);
}
#define init_5_4_14_1
void FASTCALL call_5_4_14_1(SH2_struct* sh)
{
  SH2movll4(sh, 4,14, 1);
}
#define init_5_4_14_2
void FASTCALL call_5_4_14_2(SH2_struct* sh)
{
  SH2movll4(sh, 4,14, 2);
}
#define init_5_4_14_3
void FASTCALL call_5_4_14_3(SH2_struct* sh)
{
  SH2movll4(sh, 4,14, 3);
}
#define init_5_4_14_4
void FASTCALL call_5_4_14_4(SH2_struct* sh)
{
  SH2movll4(sh, 4,14, 4);
}
#define init_5_4_14_5
void FASTCALL call_5_4_14_5(SH2_struct* sh)
{
  SH2movll4(sh, 4,14, 5);
}
#define init_5_4_14_6
void FASTCALL call_5_4_14_6(SH2_struct* sh)
{
  SH2movll4(sh, 4,14, 6);
}
#define init_5_4_14_7
void FASTCALL call_5_4_14_7(SH2_struct* sh)
{
  SH2movll4(sh, 4,14, 7);
}
#define init_5_4_14_8
void FASTCALL call_5_4_14_8(SH2_struct* sh)
{
  SH2movll4(sh, 4,14, 8);
}
#define init_5_4_14_9
void FASTCALL call_5_4_14_9(SH2_struct* sh)
{
  SH2movll4(sh, 4,14, 9);
}
#define init_5_4_14_10
void FASTCALL call_5_4_14_10(SH2_struct* sh)
{
  SH2movll4(sh, 4,14, 10);
}
#define init_5_4_14_11
void FASTCALL call_5_4_14_11(SH2_struct* sh)
{
  SH2movll4(sh, 4,14, 11);
}
#define init_5_4_14_12
void FASTCALL call_5_4_14_12(SH2_struct* sh)
{
  SH2movll4(sh, 4,14, 12);
}
#define init_5_4_14_13
void FASTCALL call_5_4_14_13(SH2_struct* sh)
{
  SH2movll4(sh, 4,14, 13);
}
#define init_5_4_14_14
void FASTCALL call_5_4_14_14(SH2_struct* sh)
{
  SH2movll4(sh, 4,14, 14);
}
#define init_5_4_14_15
void FASTCALL call_5_4_14_15(SH2_struct* sh)
{
  SH2movll4(sh, 4,14, 15);
}
#define init_5_4_15_0
void FASTCALL call_5_4_15_0(SH2_struct* sh)
{
  SH2movll4(sh, 4,15, 0);
}
#define init_5_4_15_1
void FASTCALL call_5_4_15_1(SH2_struct* sh)
{
  SH2movll4(sh, 4,15, 1);
}
#define init_5_4_15_2
void FASTCALL call_5_4_15_2(SH2_struct* sh)
{
  SH2movll4(sh, 4,15, 2);
}
#define init_5_4_15_3
void FASTCALL call_5_4_15_3(SH2_struct* sh)
{
  SH2movll4(sh, 4,15, 3);
}
#define init_5_4_15_4
void FASTCALL call_5_4_15_4(SH2_struct* sh)
{
  SH2movll4(sh, 4,15, 4);
}
#define init_5_4_15_5
void FASTCALL call_5_4_15_5(SH2_struct* sh)
{
  SH2movll4(sh, 4,15, 5);
}
#define init_5_4_15_6
void FASTCALL call_5_4_15_6(SH2_struct* sh)
{
  SH2movll4(sh, 4,15, 6);
}
#define init_5_4_15_7
void FASTCALL call_5_4_15_7(SH2_struct* sh)
{
  SH2movll4(sh, 4,15, 7);
}
#define init_5_4_15_8
void FASTCALL call_5_4_15_8(SH2_struct* sh)
{
  SH2movll4(sh, 4,15, 8);
}
#define init_5_4_15_9
void FASTCALL call_5_4_15_9(SH2_struct* sh)
{
  SH2movll4(sh, 4,15, 9);
}
#define init_5_4_15_10
void FASTCALL call_5_4_15_10(SH2_struct* sh)
{
  SH2movll4(sh, 4,15, 10);
}
#define init_5_4_15_11
void FASTCALL call_5_4_15_11(SH2_struct* sh)
{
  SH2movll4(sh, 4,15, 11);
}
#define init_5_4_15_12
void FASTCALL call_5_4_15_12(SH2_struct* sh)
{
  SH2movll4(sh, 4,15, 12);
}
#define init_5_4_15_13
void FASTCALL call_5_4_15_13(SH2_struct* sh)
{
  SH2movll4(sh, 4,15, 13);
}
#define init_5_4_15_14
void FASTCALL call_5_4_15_14(SH2_struct* sh)
{
  SH2movll4(sh, 4,15, 14);
}
#define init_5_4_15_15
void FASTCALL call_5_4_15_15(SH2_struct* sh)
{
  SH2movll4(sh, 4,15, 15);
}
#define init_5_5_0_0
void FASTCALL call_5_5_0_0(SH2_struct* sh)
{
  SH2movll4(sh, 5,0, 0);
}
#define init_5_5_0_1
void FASTCALL call_5_5_0_1(SH2_struct* sh)
{
  SH2movll4(sh, 5,0, 1);
}
#define init_5_5_0_2
void FASTCALL call_5_5_0_2(SH2_struct* sh)
{
  SH2movll4(sh, 5,0, 2);
}
#define init_5_5_0_3
void FASTCALL call_5_5_0_3(SH2_struct* sh)
{
  SH2movll4(sh, 5,0, 3);
}
#define init_5_5_0_4
void FASTCALL call_5_5_0_4(SH2_struct* sh)
{
  SH2movll4(sh, 5,0, 4);
}
#define init_5_5_0_5
void FASTCALL call_5_5_0_5(SH2_struct* sh)
{
  SH2movll4(sh, 5,0, 5);
}
#define init_5_5_0_6
void FASTCALL call_5_5_0_6(SH2_struct* sh)
{
  SH2movll4(sh, 5,0, 6);
}
#define init_5_5_0_7
void FASTCALL call_5_5_0_7(SH2_struct* sh)
{
  SH2movll4(sh, 5,0, 7);
}
#define init_5_5_0_8
void FASTCALL call_5_5_0_8(SH2_struct* sh)
{
  SH2movll4(sh, 5,0, 8);
}
#define init_5_5_0_9
void FASTCALL call_5_5_0_9(SH2_struct* sh)
{
  SH2movll4(sh, 5,0, 9);
}
#define init_5_5_0_10
void FASTCALL call_5_5_0_10(SH2_struct* sh)
{
  SH2movll4(sh, 5,0, 10);
}
#define init_5_5_0_11
void FASTCALL call_5_5_0_11(SH2_struct* sh)
{
  SH2movll4(sh, 5,0, 11);
}
#define init_5_5_0_12
void FASTCALL call_5_5_0_12(SH2_struct* sh)
{
  SH2movll4(sh, 5,0, 12);
}
#define init_5_5_0_13
void FASTCALL call_5_5_0_13(SH2_struct* sh)
{
  SH2movll4(sh, 5,0, 13);
}
#define init_5_5_0_14
void FASTCALL call_5_5_0_14(SH2_struct* sh)
{
  SH2movll4(sh, 5,0, 14);
}
#define init_5_5_0_15
void FASTCALL call_5_5_0_15(SH2_struct* sh)
{
  SH2movll4(sh, 5,0, 15);
}
#define init_5_5_1_0
void FASTCALL call_5_5_1_0(SH2_struct* sh)
{
  SH2movll4(sh, 5,1, 0);
}
#define init_5_5_1_1
void FASTCALL call_5_5_1_1(SH2_struct* sh)
{
  SH2movll4(sh, 5,1, 1);
}
#define init_5_5_1_2
void FASTCALL call_5_5_1_2(SH2_struct* sh)
{
  SH2movll4(sh, 5,1, 2);
}
#define init_5_5_1_3
void FASTCALL call_5_5_1_3(SH2_struct* sh)
{
  SH2movll4(sh, 5,1, 3);
}
#define init_5_5_1_4
void FASTCALL call_5_5_1_4(SH2_struct* sh)
{
  SH2movll4(sh, 5,1, 4);
}
#define init_5_5_1_5
void FASTCALL call_5_5_1_5(SH2_struct* sh)
{
  SH2movll4(sh, 5,1, 5);
}
#define init_5_5_1_6
void FASTCALL call_5_5_1_6(SH2_struct* sh)
{
  SH2movll4(sh, 5,1, 6);
}
#define init_5_5_1_7
void FASTCALL call_5_5_1_7(SH2_struct* sh)
{
  SH2movll4(sh, 5,1, 7);
}
#define init_5_5_1_8
void FASTCALL call_5_5_1_8(SH2_struct* sh)
{
  SH2movll4(sh, 5,1, 8);
}
#define init_5_5_1_9
void FASTCALL call_5_5_1_9(SH2_struct* sh)
{
  SH2movll4(sh, 5,1, 9);
}
#define init_5_5_1_10
void FASTCALL call_5_5_1_10(SH2_struct* sh)
{
  SH2movll4(sh, 5,1, 10);
}
#define init_5_5_1_11
void FASTCALL call_5_5_1_11(SH2_struct* sh)
{
  SH2movll4(sh, 5,1, 11);
}
#define init_5_5_1_12
void FASTCALL call_5_5_1_12(SH2_struct* sh)
{
  SH2movll4(sh, 5,1, 12);
}
#define init_5_5_1_13
void FASTCALL call_5_5_1_13(SH2_struct* sh)
{
  SH2movll4(sh, 5,1, 13);
}
#define init_5_5_1_14
void FASTCALL call_5_5_1_14(SH2_struct* sh)
{
  SH2movll4(sh, 5,1, 14);
}
#define init_5_5_1_15
void FASTCALL call_5_5_1_15(SH2_struct* sh)
{
  SH2movll4(sh, 5,1, 15);
}
#define init_5_5_2_0
void FASTCALL call_5_5_2_0(SH2_struct* sh)
{
  SH2movll4(sh, 5,2, 0);
}
#define init_5_5_2_1
void FASTCALL call_5_5_2_1(SH2_struct* sh)
{
  SH2movll4(sh, 5,2, 1);
}
#define init_5_5_2_2
void FASTCALL call_5_5_2_2(SH2_struct* sh)
{
  SH2movll4(sh, 5,2, 2);
}
#define init_5_5_2_3
void FASTCALL call_5_5_2_3(SH2_struct* sh)
{
  SH2movll4(sh, 5,2, 3);
}
#define init_5_5_2_4
void FASTCALL call_5_5_2_4(SH2_struct* sh)
{
  SH2movll4(sh, 5,2, 4);
}
#define init_5_5_2_5
void FASTCALL call_5_5_2_5(SH2_struct* sh)
{
  SH2movll4(sh, 5,2, 5);
}
#define init_5_5_2_6
void FASTCALL call_5_5_2_6(SH2_struct* sh)
{
  SH2movll4(sh, 5,2, 6);
}
#define init_5_5_2_7
void FASTCALL call_5_5_2_7(SH2_struct* sh)
{
  SH2movll4(sh, 5,2, 7);
}
#define init_5_5_2_8
void FASTCALL call_5_5_2_8(SH2_struct* sh)
{
  SH2movll4(sh, 5,2, 8);
}
#define init_5_5_2_9
void FASTCALL call_5_5_2_9(SH2_struct* sh)
{
  SH2movll4(sh, 5,2, 9);
}
#define init_5_5_2_10
void FASTCALL call_5_5_2_10(SH2_struct* sh)
{
  SH2movll4(sh, 5,2, 10);
}
#define init_5_5_2_11
void FASTCALL call_5_5_2_11(SH2_struct* sh)
{
  SH2movll4(sh, 5,2, 11);
}
#define init_5_5_2_12
void FASTCALL call_5_5_2_12(SH2_struct* sh)
{
  SH2movll4(sh, 5,2, 12);
}
#define init_5_5_2_13
void FASTCALL call_5_5_2_13(SH2_struct* sh)
{
  SH2movll4(sh, 5,2, 13);
}
#define init_5_5_2_14
void FASTCALL call_5_5_2_14(SH2_struct* sh)
{
  SH2movll4(sh, 5,2, 14);
}
#define init_5_5_2_15
void FASTCALL call_5_5_2_15(SH2_struct* sh)
{
  SH2movll4(sh, 5,2, 15);
}
#define init_5_5_3_0
void FASTCALL call_5_5_3_0(SH2_struct* sh)
{
  SH2movll4(sh, 5,3, 0);
}
#define init_5_5_3_1
void FASTCALL call_5_5_3_1(SH2_struct* sh)
{
  SH2movll4(sh, 5,3, 1);
}
#define init_5_5_3_2
void FASTCALL call_5_5_3_2(SH2_struct* sh)
{
  SH2movll4(sh, 5,3, 2);
}
#define init_5_5_3_3
void FASTCALL call_5_5_3_3(SH2_struct* sh)
{
  SH2movll4(sh, 5,3, 3);
}
#define init_5_5_3_4
void FASTCALL call_5_5_3_4(SH2_struct* sh)
{
  SH2movll4(sh, 5,3, 4);
}
#define init_5_5_3_5
void FASTCALL call_5_5_3_5(SH2_struct* sh)
{
  SH2movll4(sh, 5,3, 5);
}
#define init_5_5_3_6
void FASTCALL call_5_5_3_6(SH2_struct* sh)
{
  SH2movll4(sh, 5,3, 6);
}
#define init_5_5_3_7
void FASTCALL call_5_5_3_7(SH2_struct* sh)
{
  SH2movll4(sh, 5,3, 7);
}
#define init_5_5_3_8
void FASTCALL call_5_5_3_8(SH2_struct* sh)
{
  SH2movll4(sh, 5,3, 8);
}
#define init_5_5_3_9
void FASTCALL call_5_5_3_9(SH2_struct* sh)
{
  SH2movll4(sh, 5,3, 9);
}
#define init_5_5_3_10
void FASTCALL call_5_5_3_10(SH2_struct* sh)
{
  SH2movll4(sh, 5,3, 10);
}
#define init_5_5_3_11
void FASTCALL call_5_5_3_11(SH2_struct* sh)
{
  SH2movll4(sh, 5,3, 11);
}
#define init_5_5_3_12
void FASTCALL call_5_5_3_12(SH2_struct* sh)
{
  SH2movll4(sh, 5,3, 12);
}
#define init_5_5_3_13
void FASTCALL call_5_5_3_13(SH2_struct* sh)
{
  SH2movll4(sh, 5,3, 13);
}
#define init_5_5_3_14
void FASTCALL call_5_5_3_14(SH2_struct* sh)
{
  SH2movll4(sh, 5,3, 14);
}
#define init_5_5_3_15
void FASTCALL call_5_5_3_15(SH2_struct* sh)
{
  SH2movll4(sh, 5,3, 15);
}
#define init_5_5_4_0
void FASTCALL call_5_5_4_0(SH2_struct* sh)
{
  SH2movll4(sh, 5,4, 0);
}
#define init_5_5_4_1
void FASTCALL call_5_5_4_1(SH2_struct* sh)
{
  SH2movll4(sh, 5,4, 1);
}
#define init_5_5_4_2
void FASTCALL call_5_5_4_2(SH2_struct* sh)
{
  SH2movll4(sh, 5,4, 2);
}
#define init_5_5_4_3
void FASTCALL call_5_5_4_3(SH2_struct* sh)
{
  SH2movll4(sh, 5,4, 3);
}
#define init_5_5_4_4
void FASTCALL call_5_5_4_4(SH2_struct* sh)
{
  SH2movll4(sh, 5,4, 4);
}
#define init_5_5_4_5
void FASTCALL call_5_5_4_5(SH2_struct* sh)
{
  SH2movll4(sh, 5,4, 5);
}
#define init_5_5_4_6
void FASTCALL call_5_5_4_6(SH2_struct* sh)
{
  SH2movll4(sh, 5,4, 6);
}
#define init_5_5_4_7
void FASTCALL call_5_5_4_7(SH2_struct* sh)
{
  SH2movll4(sh, 5,4, 7);
}
#define init_5_5_4_8
void FASTCALL call_5_5_4_8(SH2_struct* sh)
{
  SH2movll4(sh, 5,4, 8);
}
#define init_5_5_4_9
void FASTCALL call_5_5_4_9(SH2_struct* sh)
{
  SH2movll4(sh, 5,4, 9);
}
#define init_5_5_4_10
void FASTCALL call_5_5_4_10(SH2_struct* sh)
{
  SH2movll4(sh, 5,4, 10);
}
#define init_5_5_4_11
void FASTCALL call_5_5_4_11(SH2_struct* sh)
{
  SH2movll4(sh, 5,4, 11);
}
#define init_5_5_4_12
void FASTCALL call_5_5_4_12(SH2_struct* sh)
{
  SH2movll4(sh, 5,4, 12);
}
#define init_5_5_4_13
void FASTCALL call_5_5_4_13(SH2_struct* sh)
{
  SH2movll4(sh, 5,4, 13);
}
#define init_5_5_4_14
void FASTCALL call_5_5_4_14(SH2_struct* sh)
{
  SH2movll4(sh, 5,4, 14);
}
#define init_5_5_4_15
void FASTCALL call_5_5_4_15(SH2_struct* sh)
{
  SH2movll4(sh, 5,4, 15);
}
#define init_5_5_5_0
void FASTCALL call_5_5_5_0(SH2_struct* sh)
{
  SH2movll4(sh, 5,5, 0);
}
#define init_5_5_5_1
void FASTCALL call_5_5_5_1(SH2_struct* sh)
{
  SH2movll4(sh, 5,5, 1);
}
#define init_5_5_5_2
void FASTCALL call_5_5_5_2(SH2_struct* sh)
{
  SH2movll4(sh, 5,5, 2);
}
#define init_5_5_5_3
void FASTCALL call_5_5_5_3(SH2_struct* sh)
{
  SH2movll4(sh, 5,5, 3);
}
#define init_5_5_5_4
void FASTCALL call_5_5_5_4(SH2_struct* sh)
{
  SH2movll4(sh, 5,5, 4);
}
#define init_5_5_5_5
void FASTCALL call_5_5_5_5(SH2_struct* sh)
{
  SH2movll4(sh, 5,5, 5);
}
#define init_5_5_5_6
void FASTCALL call_5_5_5_6(SH2_struct* sh)
{
  SH2movll4(sh, 5,5, 6);
}
#define init_5_5_5_7
void FASTCALL call_5_5_5_7(SH2_struct* sh)
{
  SH2movll4(sh, 5,5, 7);
}
#define init_5_5_5_8
void FASTCALL call_5_5_5_8(SH2_struct* sh)
{
  SH2movll4(sh, 5,5, 8);
}
#define init_5_5_5_9
void FASTCALL call_5_5_5_9(SH2_struct* sh)
{
  SH2movll4(sh, 5,5, 9);
}
#define init_5_5_5_10
void FASTCALL call_5_5_5_10(SH2_struct* sh)
{
  SH2movll4(sh, 5,5, 10);
}
#define init_5_5_5_11
void FASTCALL call_5_5_5_11(SH2_struct* sh)
{
  SH2movll4(sh, 5,5, 11);
}
#define init_5_5_5_12
void FASTCALL call_5_5_5_12(SH2_struct* sh)
{
  SH2movll4(sh, 5,5, 12);
}
#define init_5_5_5_13
void FASTCALL call_5_5_5_13(SH2_struct* sh)
{
  SH2movll4(sh, 5,5, 13);
}
#define init_5_5_5_14
void FASTCALL call_5_5_5_14(SH2_struct* sh)
{
  SH2movll4(sh, 5,5, 14);
}
#define init_5_5_5_15
void FASTCALL call_5_5_5_15(SH2_struct* sh)
{
  SH2movll4(sh, 5,5, 15);
}
#define init_5_5_6_0
void FASTCALL call_5_5_6_0(SH2_struct* sh)
{
  SH2movll4(sh, 5,6, 0);
}
#define init_5_5_6_1
void FASTCALL call_5_5_6_1(SH2_struct* sh)
{
  SH2movll4(sh, 5,6, 1);
}
#define init_5_5_6_2
void FASTCALL call_5_5_6_2(SH2_struct* sh)
{
  SH2movll4(sh, 5,6, 2);
}
#define init_5_5_6_3
void FASTCALL call_5_5_6_3(SH2_struct* sh)
{
  SH2movll4(sh, 5,6, 3);
}
#define init_5_5_6_4
void FASTCALL call_5_5_6_4(SH2_struct* sh)
{
  SH2movll4(sh, 5,6, 4);
}
#define init_5_5_6_5
void FASTCALL call_5_5_6_5(SH2_struct* sh)
{
  SH2movll4(sh, 5,6, 5);
}
#define init_5_5_6_6
void FASTCALL call_5_5_6_6(SH2_struct* sh)
{
  SH2movll4(sh, 5,6, 6);
}
#define init_5_5_6_7
void FASTCALL call_5_5_6_7(SH2_struct* sh)
{
  SH2movll4(sh, 5,6, 7);
}
#define init_5_5_6_8
void FASTCALL call_5_5_6_8(SH2_struct* sh)
{
  SH2movll4(sh, 5,6, 8);
}
#define init_5_5_6_9
void FASTCALL call_5_5_6_9(SH2_struct* sh)
{
  SH2movll4(sh, 5,6, 9);
}
#define init_5_5_6_10
void FASTCALL call_5_5_6_10(SH2_struct* sh)
{
  SH2movll4(sh, 5,6, 10);
}
#define init_5_5_6_11
void FASTCALL call_5_5_6_11(SH2_struct* sh)
{
  SH2movll4(sh, 5,6, 11);
}
#define init_5_5_6_12
void FASTCALL call_5_5_6_12(SH2_struct* sh)
{
  SH2movll4(sh, 5,6, 12);
}
#define init_5_5_6_13
void FASTCALL call_5_5_6_13(SH2_struct* sh)
{
  SH2movll4(sh, 5,6, 13);
}
#define init_5_5_6_14
void FASTCALL call_5_5_6_14(SH2_struct* sh)
{
  SH2movll4(sh, 5,6, 14);
}
#define init_5_5_6_15
void FASTCALL call_5_5_6_15(SH2_struct* sh)
{
  SH2movll4(sh, 5,6, 15);
}
#define init_5_5_7_0
void FASTCALL call_5_5_7_0(SH2_struct* sh)
{
  SH2movll4(sh, 5,7, 0);
}
#define init_5_5_7_1
void FASTCALL call_5_5_7_1(SH2_struct* sh)
{
  SH2movll4(sh, 5,7, 1);
}
#define init_5_5_7_2
void FASTCALL call_5_5_7_2(SH2_struct* sh)
{
  SH2movll4(sh, 5,7, 2);
}
#define init_5_5_7_3
void FASTCALL call_5_5_7_3(SH2_struct* sh)
{
  SH2movll4(sh, 5,7, 3);
}
#define init_5_5_7_4
void FASTCALL call_5_5_7_4(SH2_struct* sh)
{
  SH2movll4(sh, 5,7, 4);
}
#define init_5_5_7_5
void FASTCALL call_5_5_7_5(SH2_struct* sh)
{
  SH2movll4(sh, 5,7, 5);
}
#define init_5_5_7_6
void FASTCALL call_5_5_7_6(SH2_struct* sh)
{
  SH2movll4(sh, 5,7, 6);
}
#define init_5_5_7_7
void FASTCALL call_5_5_7_7(SH2_struct* sh)
{
  SH2movll4(sh, 5,7, 7);
}
#define init_5_5_7_8
void FASTCALL call_5_5_7_8(SH2_struct* sh)
{
  SH2movll4(sh, 5,7, 8);
}
#define init_5_5_7_9
void FASTCALL call_5_5_7_9(SH2_struct* sh)
{
  SH2movll4(sh, 5,7, 9);
}
#define init_5_5_7_10
void FASTCALL call_5_5_7_10(SH2_struct* sh)
{
  SH2movll4(sh, 5,7, 10);
}
#define init_5_5_7_11
void FASTCALL call_5_5_7_11(SH2_struct* sh)
{
  SH2movll4(sh, 5,7, 11);
}
#define init_5_5_7_12
void FASTCALL call_5_5_7_12(SH2_struct* sh)
{
  SH2movll4(sh, 5,7, 12);
}
#define init_5_5_7_13
void FASTCALL call_5_5_7_13(SH2_struct* sh)
{
  SH2movll4(sh, 5,7, 13);
}
#define init_5_5_7_14
void FASTCALL call_5_5_7_14(SH2_struct* sh)
{
  SH2movll4(sh, 5,7, 14);
}
#define init_5_5_7_15
void FASTCALL call_5_5_7_15(SH2_struct* sh)
{
  SH2movll4(sh, 5,7, 15);
}
#define init_5_5_8_0
void FASTCALL call_5_5_8_0(SH2_struct* sh)
{
  SH2movll4(sh, 5,8, 0);
}
#define init_5_5_8_1
void FASTCALL call_5_5_8_1(SH2_struct* sh)
{
  SH2movll4(sh, 5,8, 1);
}
#define init_5_5_8_2
void FASTCALL call_5_5_8_2(SH2_struct* sh)
{
  SH2movll4(sh, 5,8, 2);
}
#define init_5_5_8_3
void FASTCALL call_5_5_8_3(SH2_struct* sh)
{
  SH2movll4(sh, 5,8, 3);
}
#define init_5_5_8_4
void FASTCALL call_5_5_8_4(SH2_struct* sh)
{
  SH2movll4(sh, 5,8, 4);
}
#define init_5_5_8_5
void FASTCALL call_5_5_8_5(SH2_struct* sh)
{
  SH2movll4(sh, 5,8, 5);
}
#define init_5_5_8_6
void FASTCALL call_5_5_8_6(SH2_struct* sh)
{
  SH2movll4(sh, 5,8, 6);
}
#define init_5_5_8_7
void FASTCALL call_5_5_8_7(SH2_struct* sh)
{
  SH2movll4(sh, 5,8, 7);
}
#define init_5_5_8_8
void FASTCALL call_5_5_8_8(SH2_struct* sh)
{
  SH2movll4(sh, 5,8, 8);
}
#define init_5_5_8_9
void FASTCALL call_5_5_8_9(SH2_struct* sh)
{
  SH2movll4(sh, 5,8, 9);
}
#define init_5_5_8_10
void FASTCALL call_5_5_8_10(SH2_struct* sh)
{
  SH2movll4(sh, 5,8, 10);
}
#define init_5_5_8_11
void FASTCALL call_5_5_8_11(SH2_struct* sh)
{
  SH2movll4(sh, 5,8, 11);
}
#define init_5_5_8_12
void FASTCALL call_5_5_8_12(SH2_struct* sh)
{
  SH2movll4(sh, 5,8, 12);
}
#define init_5_5_8_13
void FASTCALL call_5_5_8_13(SH2_struct* sh)
{
  SH2movll4(sh, 5,8, 13);
}
#define init_5_5_8_14
void FASTCALL call_5_5_8_14(SH2_struct* sh)
{
  SH2movll4(sh, 5,8, 14);
}
#define init_5_5_8_15
void FASTCALL call_5_5_8_15(SH2_struct* sh)
{
  SH2movll4(sh, 5,8, 15);
}
#define init_5_5_9_0
void FASTCALL call_5_5_9_0(SH2_struct* sh)
{
  SH2movll4(sh, 5,9, 0);
}
#define init_5_5_9_1
void FASTCALL call_5_5_9_1(SH2_struct* sh)
{
  SH2movll4(sh, 5,9, 1);
}
#define init_5_5_9_2
void FASTCALL call_5_5_9_2(SH2_struct* sh)
{
  SH2movll4(sh, 5,9, 2);
}
#define init_5_5_9_3
void FASTCALL call_5_5_9_3(SH2_struct* sh)
{
  SH2movll4(sh, 5,9, 3);
}
#define init_5_5_9_4
void FASTCALL call_5_5_9_4(SH2_struct* sh)
{
  SH2movll4(sh, 5,9, 4);
}
#define init_5_5_9_5
void FASTCALL call_5_5_9_5(SH2_struct* sh)
{
  SH2movll4(sh, 5,9, 5);
}
#define init_5_5_9_6
void FASTCALL call_5_5_9_6(SH2_struct* sh)
{
  SH2movll4(sh, 5,9, 6);
}
#define init_5_5_9_7
void FASTCALL call_5_5_9_7(SH2_struct* sh)
{
  SH2movll4(sh, 5,9, 7);
}
#define init_5_5_9_8
void FASTCALL call_5_5_9_8(SH2_struct* sh)
{
  SH2movll4(sh, 5,9, 8);
}
#define init_5_5_9_9
void FASTCALL call_5_5_9_9(SH2_struct* sh)
{
  SH2movll4(sh, 5,9, 9);
}
#define init_5_5_9_10
void FASTCALL call_5_5_9_10(SH2_struct* sh)
{
  SH2movll4(sh, 5,9, 10);
}
#define init_5_5_9_11
void FASTCALL call_5_5_9_11(SH2_struct* sh)
{
  SH2movll4(sh, 5,9, 11);
}
#define init_5_5_9_12
void FASTCALL call_5_5_9_12(SH2_struct* sh)
{
  SH2movll4(sh, 5,9, 12);
}
#define init_5_5_9_13
void FASTCALL call_5_5_9_13(SH2_struct* sh)
{
  SH2movll4(sh, 5,9, 13);
}
#define init_5_5_9_14
void FASTCALL call_5_5_9_14(SH2_struct* sh)
{
  SH2movll4(sh, 5,9, 14);
}
#define init_5_5_9_15
void FASTCALL call_5_5_9_15(SH2_struct* sh)
{
  SH2movll4(sh, 5,9, 15);
}
#define init_5_5_10_0
void FASTCALL call_5_5_10_0(SH2_struct* sh)
{
  SH2movll4(sh, 5,10, 0);
}
#define init_5_5_10_1
void FASTCALL call_5_5_10_1(SH2_struct* sh)
{
  SH2movll4(sh, 5,10, 1);
}
#define init_5_5_10_2
void FASTCALL call_5_5_10_2(SH2_struct* sh)
{
  SH2movll4(sh, 5,10, 2);
}
#define init_5_5_10_3
void FASTCALL call_5_5_10_3(SH2_struct* sh)
{
  SH2movll4(sh, 5,10, 3);
}
#define init_5_5_10_4
void FASTCALL call_5_5_10_4(SH2_struct* sh)
{
  SH2movll4(sh, 5,10, 4);
}
#define init_5_5_10_5
void FASTCALL call_5_5_10_5(SH2_struct* sh)
{
  SH2movll4(sh, 5,10, 5);
}
#define init_5_5_10_6
void FASTCALL call_5_5_10_6(SH2_struct* sh)
{
  SH2movll4(sh, 5,10, 6);
}
#define init_5_5_10_7
void FASTCALL call_5_5_10_7(SH2_struct* sh)
{
  SH2movll4(sh, 5,10, 7);
}
#define init_5_5_10_8
void FASTCALL call_5_5_10_8(SH2_struct* sh)
{
  SH2movll4(sh, 5,10, 8);
}
#define init_5_5_10_9
void FASTCALL call_5_5_10_9(SH2_struct* sh)
{
  SH2movll4(sh, 5,10, 9);
}
#define init_5_5_10_10
void FASTCALL call_5_5_10_10(SH2_struct* sh)
{
  SH2movll4(sh, 5,10, 10);
}
#define init_5_5_10_11
void FASTCALL call_5_5_10_11(SH2_struct* sh)
{
  SH2movll4(sh, 5,10, 11);
}
#define init_5_5_10_12
void FASTCALL call_5_5_10_12(SH2_struct* sh)
{
  SH2movll4(sh, 5,10, 12);
}
#define init_5_5_10_13
void FASTCALL call_5_5_10_13(SH2_struct* sh)
{
  SH2movll4(sh, 5,10, 13);
}
#define init_5_5_10_14
void FASTCALL call_5_5_10_14(SH2_struct* sh)
{
  SH2movll4(sh, 5,10, 14);
}
#define init_5_5_10_15
void FASTCALL call_5_5_10_15(SH2_struct* sh)
{
  SH2movll4(sh, 5,10, 15);
}
#define init_5_5_11_0
void FASTCALL call_5_5_11_0(SH2_struct* sh)
{
  SH2movll4(sh, 5,11, 0);
}
#define init_5_5_11_1
void FASTCALL call_5_5_11_1(SH2_struct* sh)
{
  SH2movll4(sh, 5,11, 1);
}
#define init_5_5_11_2
void FASTCALL call_5_5_11_2(SH2_struct* sh)
{
  SH2movll4(sh, 5,11, 2);
}
#define init_5_5_11_3
void FASTCALL call_5_5_11_3(SH2_struct* sh)
{
  SH2movll4(sh, 5,11, 3);
}
#define init_5_5_11_4
void FASTCALL call_5_5_11_4(SH2_struct* sh)
{
  SH2movll4(sh, 5,11, 4);
}
#define init_5_5_11_5
void FASTCALL call_5_5_11_5(SH2_struct* sh)
{
  SH2movll4(sh, 5,11, 5);
}
#define init_5_5_11_6
void FASTCALL call_5_5_11_6(SH2_struct* sh)
{
  SH2movll4(sh, 5,11, 6);
}
#define init_5_5_11_7
void FASTCALL call_5_5_11_7(SH2_struct* sh)
{
  SH2movll4(sh, 5,11, 7);
}
#define init_5_5_11_8
void FASTCALL call_5_5_11_8(SH2_struct* sh)
{
  SH2movll4(sh, 5,11, 8);
}
#define init_5_5_11_9
void FASTCALL call_5_5_11_9(SH2_struct* sh)
{
  SH2movll4(sh, 5,11, 9);
}
#define init_5_5_11_10
void FASTCALL call_5_5_11_10(SH2_struct* sh)
{
  SH2movll4(sh, 5,11, 10);
}
#define init_5_5_11_11
void FASTCALL call_5_5_11_11(SH2_struct* sh)
{
  SH2movll4(sh, 5,11, 11);
}
#define init_5_5_11_12
void FASTCALL call_5_5_11_12(SH2_struct* sh)
{
  SH2movll4(sh, 5,11, 12);
}
#define init_5_5_11_13
void FASTCALL call_5_5_11_13(SH2_struct* sh)
{
  SH2movll4(sh, 5,11, 13);
}
#define init_5_5_11_14
void FASTCALL call_5_5_11_14(SH2_struct* sh)
{
  SH2movll4(sh, 5,11, 14);
}
#define init_5_5_11_15
void FASTCALL call_5_5_11_15(SH2_struct* sh)
{
  SH2movll4(sh, 5,11, 15);
}
#define init_5_5_12_0
void FASTCALL call_5_5_12_0(SH2_struct* sh)
{
  SH2movll4(sh, 5,12, 0);
}
#define init_5_5_12_1
void FASTCALL call_5_5_12_1(SH2_struct* sh)
{
  SH2movll4(sh, 5,12, 1);
}
#define init_5_5_12_2
void FASTCALL call_5_5_12_2(SH2_struct* sh)
{
  SH2movll4(sh, 5,12, 2);
}
#define init_5_5_12_3
void FASTCALL call_5_5_12_3(SH2_struct* sh)
{
  SH2movll4(sh, 5,12, 3);
}
#define init_5_5_12_4
void FASTCALL call_5_5_12_4(SH2_struct* sh)
{
  SH2movll4(sh, 5,12, 4);
}
#define init_5_5_12_5
void FASTCALL call_5_5_12_5(SH2_struct* sh)
{
  SH2movll4(sh, 5,12, 5);
}
#define init_5_5_12_6
void FASTCALL call_5_5_12_6(SH2_struct* sh)
{
  SH2movll4(sh, 5,12, 6);
}
#define init_5_5_12_7
void FASTCALL call_5_5_12_7(SH2_struct* sh)
{
  SH2movll4(sh, 5,12, 7);
}
#define init_5_5_12_8
void FASTCALL call_5_5_12_8(SH2_struct* sh)
{
  SH2movll4(sh, 5,12, 8);
}
#define init_5_5_12_9
void FASTCALL call_5_5_12_9(SH2_struct* sh)
{
  SH2movll4(sh, 5,12, 9);
}
#define init_5_5_12_10
void FASTCALL call_5_5_12_10(SH2_struct* sh)
{
  SH2movll4(sh, 5,12, 10);
}
#define init_5_5_12_11
void FASTCALL call_5_5_12_11(SH2_struct* sh)
{
  SH2movll4(sh, 5,12, 11);
}
#define init_5_5_12_12
void FASTCALL call_5_5_12_12(SH2_struct* sh)
{
  SH2movll4(sh, 5,12, 12);
}
#define init_5_5_12_13
void FASTCALL call_5_5_12_13(SH2_struct* sh)
{
  SH2movll4(sh, 5,12, 13);
}
#define init_5_5_12_14
void FASTCALL call_5_5_12_14(SH2_struct* sh)
{
  SH2movll4(sh, 5,12, 14);
}
#define init_5_5_12_15
void FASTCALL call_5_5_12_15(SH2_struct* sh)
{
  SH2movll4(sh, 5,12, 15);
}
#define init_5_5_13_0
void FASTCALL call_5_5_13_0(SH2_struct* sh)
{
  SH2movll4(sh, 5,13, 0);
}
#define init_5_5_13_1
void FASTCALL call_5_5_13_1(SH2_struct* sh)
{
  SH2movll4(sh, 5,13, 1);
}
#define init_5_5_13_2
void FASTCALL call_5_5_13_2(SH2_struct* sh)
{
  SH2movll4(sh, 5,13, 2);
}
#define init_5_5_13_3
void FASTCALL call_5_5_13_3(SH2_struct* sh)
{
  SH2movll4(sh, 5,13, 3);
}
#define init_5_5_13_4
void FASTCALL call_5_5_13_4(SH2_struct* sh)
{
  SH2movll4(sh, 5,13, 4);
}
#define init_5_5_13_5
void FASTCALL call_5_5_13_5(SH2_struct* sh)
{
  SH2movll4(sh, 5,13, 5);
}
#define init_5_5_13_6
void FASTCALL call_5_5_13_6(SH2_struct* sh)
{
  SH2movll4(sh, 5,13, 6);
}
#define init_5_5_13_7
void FASTCALL call_5_5_13_7(SH2_struct* sh)
{
  SH2movll4(sh, 5,13, 7);
}
#define init_5_5_13_8
void FASTCALL call_5_5_13_8(SH2_struct* sh)
{
  SH2movll4(sh, 5,13, 8);
}
#define init_5_5_13_9
void FASTCALL call_5_5_13_9(SH2_struct* sh)
{
  SH2movll4(sh, 5,13, 9);
}
#define init_5_5_13_10
void FASTCALL call_5_5_13_10(SH2_struct* sh)
{
  SH2movll4(sh, 5,13, 10);
}
#define init_5_5_13_11
void FASTCALL call_5_5_13_11(SH2_struct* sh)
{
  SH2movll4(sh, 5,13, 11);
}
#define init_5_5_13_12
void FASTCALL call_5_5_13_12(SH2_struct* sh)
{
  SH2movll4(sh, 5,13, 12);
}
#define init_5_5_13_13
void FASTCALL call_5_5_13_13(SH2_struct* sh)
{
  SH2movll4(sh, 5,13, 13);
}
#define init_5_5_13_14
void FASTCALL call_5_5_13_14(SH2_struct* sh)
{
  SH2movll4(sh, 5,13, 14);
}
#define init_5_5_13_15
void FASTCALL call_5_5_13_15(SH2_struct* sh)
{
  SH2movll4(sh, 5,13, 15);
}
#define init_5_5_14_0
void FASTCALL call_5_5_14_0(SH2_struct* sh)
{
  SH2movll4(sh, 5,14, 0);
}
#define init_5_5_14_1
void FASTCALL call_5_5_14_1(SH2_struct* sh)
{
  SH2movll4(sh, 5,14, 1);
}
#define init_5_5_14_2
void FASTCALL call_5_5_14_2(SH2_struct* sh)
{
  SH2movll4(sh, 5,14, 2);
}
#define init_5_5_14_3
void FASTCALL call_5_5_14_3(SH2_struct* sh)
{
  SH2movll4(sh, 5,14, 3);
}
#define init_5_5_14_4
void FASTCALL call_5_5_14_4(SH2_struct* sh)
{
  SH2movll4(sh, 5,14, 4);
}
#define init_5_5_14_5
void FASTCALL call_5_5_14_5(SH2_struct* sh)
{
  SH2movll4(sh, 5,14, 5);
}
#define init_5_5_14_6
void FASTCALL call_5_5_14_6(SH2_struct* sh)
{
  SH2movll4(sh, 5,14, 6);
}
#define init_5_5_14_7
void FASTCALL call_5_5_14_7(SH2_struct* sh)
{
  SH2movll4(sh, 5,14, 7);
}
#define init_5_5_14_8
void FASTCALL call_5_5_14_8(SH2_struct* sh)
{
  SH2movll4(sh, 5,14, 8);
}
#define init_5_5_14_9
void FASTCALL call_5_5_14_9(SH2_struct* sh)
{
  SH2movll4(sh, 5,14, 9);
}
#define init_5_5_14_10
void FASTCALL call_5_5_14_10(SH2_struct* sh)
{
  SH2movll4(sh, 5,14, 10);
}
#define init_5_5_14_11
void FASTCALL call_5_5_14_11(SH2_struct* sh)
{
  SH2movll4(sh, 5,14, 11);
}
#define init_5_5_14_12
void FASTCALL call_5_5_14_12(SH2_struct* sh)
{
  SH2movll4(sh, 5,14, 12);
}
#define init_5_5_14_13
void FASTCALL call_5_5_14_13(SH2_struct* sh)
{
  SH2movll4(sh, 5,14, 13);
}
#define init_5_5_14_14
void FASTCALL call_5_5_14_14(SH2_struct* sh)
{
  SH2movll4(sh, 5,14, 14);
}
#define init_5_5_14_15
void FASTCALL call_5_5_14_15(SH2_struct* sh)
{
  SH2movll4(sh, 5,14, 15);
}
#define init_5_5_15_0
void FASTCALL call_5_5_15_0(SH2_struct* sh)
{
  SH2movll4(sh, 5,15, 0);
}
#define init_5_5_15_1
void FASTCALL call_5_5_15_1(SH2_struct* sh)
{
  SH2movll4(sh, 5,15, 1);
}
#define init_5_5_15_2
void FASTCALL call_5_5_15_2(SH2_struct* sh)
{
  SH2movll4(sh, 5,15, 2);
}
#define init_5_5_15_3
void FASTCALL call_5_5_15_3(SH2_struct* sh)
{
  SH2movll4(sh, 5,15, 3);
}
#define init_5_5_15_4
void FASTCALL call_5_5_15_4(SH2_struct* sh)
{
  SH2movll4(sh, 5,15, 4);
}
#define init_5_5_15_5
void FASTCALL call_5_5_15_5(SH2_struct* sh)
{
  SH2movll4(sh, 5,15, 5);
}
#define init_5_5_15_6
void FASTCALL call_5_5_15_6(SH2_struct* sh)
{
  SH2movll4(sh, 5,15, 6);
}
#define init_5_5_15_7
void FASTCALL call_5_5_15_7(SH2_struct* sh)
{
  SH2movll4(sh, 5,15, 7);
}
#define init_5_5_15_8
void FASTCALL call_5_5_15_8(SH2_struct* sh)
{
  SH2movll4(sh, 5,15, 8);
}
#define init_5_5_15_9
void FASTCALL call_5_5_15_9(SH2_struct* sh)
{
  SH2movll4(sh, 5,15, 9);
}
#define init_5_5_15_10
void FASTCALL call_5_5_15_10(SH2_struct* sh)
{
  SH2movll4(sh, 5,15, 10);
}
#define init_5_5_15_11
void FASTCALL call_5_5_15_11(SH2_struct* sh)
{
  SH2movll4(sh, 5,15, 11);
}
#define init_5_5_15_12
void FASTCALL call_5_5_15_12(SH2_struct* sh)
{
  SH2movll4(sh, 5,15, 12);
}
#define init_5_5_15_13
void FASTCALL call_5_5_15_13(SH2_struct* sh)
{
  SH2movll4(sh, 5,15, 13);
}
#define init_5_5_15_14
void FASTCALL call_5_5_15_14(SH2_struct* sh)
{
  SH2movll4(sh, 5,15, 14);
}
#define init_5_5_15_15
void FASTCALL call_5_5_15_15(SH2_struct* sh)
{
  SH2movll4(sh, 5,15, 15);
}
#define init_5_6_0_0
void FASTCALL call_5_6_0_0(SH2_struct* sh)
{
  SH2movll4(sh, 6,0, 0);
}
#define init_5_6_0_1
void FASTCALL call_5_6_0_1(SH2_struct* sh)
{
  SH2movll4(sh, 6,0, 1);
}
#define init_5_6_0_2
void FASTCALL call_5_6_0_2(SH2_struct* sh)
{
  SH2movll4(sh, 6,0, 2);
}
#define init_5_6_0_3
void FASTCALL call_5_6_0_3(SH2_struct* sh)
{
  SH2movll4(sh, 6,0, 3);
}
#define init_5_6_0_4
void FASTCALL call_5_6_0_4(SH2_struct* sh)
{
  SH2movll4(sh, 6,0, 4);
}
#define init_5_6_0_5
void FASTCALL call_5_6_0_5(SH2_struct* sh)
{
  SH2movll4(sh, 6,0, 5);
}
#define init_5_6_0_6
void FASTCALL call_5_6_0_6(SH2_struct* sh)
{
  SH2movll4(sh, 6,0, 6);
}
#define init_5_6_0_7
void FASTCALL call_5_6_0_7(SH2_struct* sh)
{
  SH2movll4(sh, 6,0, 7);
}
#define init_5_6_0_8
void FASTCALL call_5_6_0_8(SH2_struct* sh)
{
  SH2movll4(sh, 6,0, 8);
}
#define init_5_6_0_9
void FASTCALL call_5_6_0_9(SH2_struct* sh)
{
  SH2movll4(sh, 6,0, 9);
}
#define init_5_6_0_10
void FASTCALL call_5_6_0_10(SH2_struct* sh)
{
  SH2movll4(sh, 6,0, 10);
}
#define init_5_6_0_11
void FASTCALL call_5_6_0_11(SH2_struct* sh)
{
  SH2movll4(sh, 6,0, 11);
}
#define init_5_6_0_12
void FASTCALL call_5_6_0_12(SH2_struct* sh)
{
  SH2movll4(sh, 6,0, 12);
}
#define init_5_6_0_13
void FASTCALL call_5_6_0_13(SH2_struct* sh)
{
  SH2movll4(sh, 6,0, 13);
}
#define init_5_6_0_14
void FASTCALL call_5_6_0_14(SH2_struct* sh)
{
  SH2movll4(sh, 6,0, 14);
}
#define init_5_6_0_15
void FASTCALL call_5_6_0_15(SH2_struct* sh)
{
  SH2movll4(sh, 6,0, 15);
}
#define init_5_6_1_0
void FASTCALL call_5_6_1_0(SH2_struct* sh)
{
  SH2movll4(sh, 6,1, 0);
}
#define init_5_6_1_1
void FASTCALL call_5_6_1_1(SH2_struct* sh)
{
  SH2movll4(sh, 6,1, 1);
}
#define init_5_6_1_2
void FASTCALL call_5_6_1_2(SH2_struct* sh)
{
  SH2movll4(sh, 6,1, 2);
}
#define init_5_6_1_3
void FASTCALL call_5_6_1_3(SH2_struct* sh)
{
  SH2movll4(sh, 6,1, 3);
}
#define init_5_6_1_4
void FASTCALL call_5_6_1_4(SH2_struct* sh)
{
  SH2movll4(sh, 6,1, 4);
}
#define init_5_6_1_5
void FASTCALL call_5_6_1_5(SH2_struct* sh)
{
  SH2movll4(sh, 6,1, 5);
}
#define init_5_6_1_6
void FASTCALL call_5_6_1_6(SH2_struct* sh)
{
  SH2movll4(sh, 6,1, 6);
}
#define init_5_6_1_7
void FASTCALL call_5_6_1_7(SH2_struct* sh)
{
  SH2movll4(sh, 6,1, 7);
}
#define init_5_6_1_8
void FASTCALL call_5_6_1_8(SH2_struct* sh)
{
  SH2movll4(sh, 6,1, 8);
}
#define init_5_6_1_9
void FASTCALL call_5_6_1_9(SH2_struct* sh)
{
  SH2movll4(sh, 6,1, 9);
}
#define init_5_6_1_10
void FASTCALL call_5_6_1_10(SH2_struct* sh)
{
  SH2movll4(sh, 6,1, 10);
}
#define init_5_6_1_11
void FASTCALL call_5_6_1_11(SH2_struct* sh)
{
  SH2movll4(sh, 6,1, 11);
}
#define init_5_6_1_12
void FASTCALL call_5_6_1_12(SH2_struct* sh)
{
  SH2movll4(sh, 6,1, 12);
}
#define init_5_6_1_13
void FASTCALL call_5_6_1_13(SH2_struct* sh)
{
  SH2movll4(sh, 6,1, 13);
}
#define init_5_6_1_14
void FASTCALL call_5_6_1_14(SH2_struct* sh)
{
  SH2movll4(sh, 6,1, 14);
}
#define init_5_6_1_15
void FASTCALL call_5_6_1_15(SH2_struct* sh)
{
  SH2movll4(sh, 6,1, 15);
}
#define init_5_6_2_0
void FASTCALL call_5_6_2_0(SH2_struct* sh)
{
  SH2movll4(sh, 6,2, 0);
}
#define init_5_6_2_1
void FASTCALL call_5_6_2_1(SH2_struct* sh)
{
  SH2movll4(sh, 6,2, 1);
}
#define init_5_6_2_2
void FASTCALL call_5_6_2_2(SH2_struct* sh)
{
  SH2movll4(sh, 6,2, 2);
}
#define init_5_6_2_3
void FASTCALL call_5_6_2_3(SH2_struct* sh)
{
  SH2movll4(sh, 6,2, 3);
}
#define init_5_6_2_4
void FASTCALL call_5_6_2_4(SH2_struct* sh)
{
  SH2movll4(sh, 6,2, 4);
}
#define init_5_6_2_5
void FASTCALL call_5_6_2_5(SH2_struct* sh)
{
  SH2movll4(sh, 6,2, 5);
}
#define init_5_6_2_6
void FASTCALL call_5_6_2_6(SH2_struct* sh)
{
  SH2movll4(sh, 6,2, 6);
}
#define init_5_6_2_7
void FASTCALL call_5_6_2_7(SH2_struct* sh)
{
  SH2movll4(sh, 6,2, 7);
}
#define init_5_6_2_8
void FASTCALL call_5_6_2_8(SH2_struct* sh)
{
  SH2movll4(sh, 6,2, 8);
}
#define init_5_6_2_9
void FASTCALL call_5_6_2_9(SH2_struct* sh)
{
  SH2movll4(sh, 6,2, 9);
}
#define init_5_6_2_10
void FASTCALL call_5_6_2_10(SH2_struct* sh)
{
  SH2movll4(sh, 6,2, 10);
}
#define init_5_6_2_11
void FASTCALL call_5_6_2_11(SH2_struct* sh)
{
  SH2movll4(sh, 6,2, 11);
}
#define init_5_6_2_12
void FASTCALL call_5_6_2_12(SH2_struct* sh)
{
  SH2movll4(sh, 6,2, 12);
}
#define init_5_6_2_13
void FASTCALL call_5_6_2_13(SH2_struct* sh)
{
  SH2movll4(sh, 6,2, 13);
}
#define init_5_6_2_14
void FASTCALL call_5_6_2_14(SH2_struct* sh)
{
  SH2movll4(sh, 6,2, 14);
}
#define init_5_6_2_15
void FASTCALL call_5_6_2_15(SH2_struct* sh)
{
  SH2movll4(sh, 6,2, 15);
}
#define init_5_6_3_0
void FASTCALL call_5_6_3_0(SH2_struct* sh)
{
  SH2movll4(sh, 6,3, 0);
}
#define init_5_6_3_1
void FASTCALL call_5_6_3_1(SH2_struct* sh)
{
  SH2movll4(sh, 6,3, 1);
}
#define init_5_6_3_2
void FASTCALL call_5_6_3_2(SH2_struct* sh)
{
  SH2movll4(sh, 6,3, 2);
}
#define init_5_6_3_3
void FASTCALL call_5_6_3_3(SH2_struct* sh)
{
  SH2movll4(sh, 6,3, 3);
}
#define init_5_6_3_4
void FASTCALL call_5_6_3_4(SH2_struct* sh)
{
  SH2movll4(sh, 6,3, 4);
}
#define init_5_6_3_5
void FASTCALL call_5_6_3_5(SH2_struct* sh)
{
  SH2movll4(sh, 6,3, 5);
}
#define init_5_6_3_6
void FASTCALL call_5_6_3_6(SH2_struct* sh)
{
  SH2movll4(sh, 6,3, 6);
}
#define init_5_6_3_7
void FASTCALL call_5_6_3_7(SH2_struct* sh)
{
  SH2movll4(sh, 6,3, 7);
}
#define init_5_6_3_8
void FASTCALL call_5_6_3_8(SH2_struct* sh)
{
  SH2movll4(sh, 6,3, 8);
}
#define init_5_6_3_9
void FASTCALL call_5_6_3_9(SH2_struct* sh)
{
  SH2movll4(sh, 6,3, 9);
}
#define init_5_6_3_10
void FASTCALL call_5_6_3_10(SH2_struct* sh)
{
  SH2movll4(sh, 6,3, 10);
}
#define init_5_6_3_11
void FASTCALL call_5_6_3_11(SH2_struct* sh)
{
  SH2movll4(sh, 6,3, 11);
}
#define init_5_6_3_12
void FASTCALL call_5_6_3_12(SH2_struct* sh)
{
  SH2movll4(sh, 6,3, 12);
}
#define init_5_6_3_13
void FASTCALL call_5_6_3_13(SH2_struct* sh)
{
  SH2movll4(sh, 6,3, 13);
}
#define init_5_6_3_14
void FASTCALL call_5_6_3_14(SH2_struct* sh)
{
  SH2movll4(sh, 6,3, 14);
}
#define init_5_6_3_15
void FASTCALL call_5_6_3_15(SH2_struct* sh)
{
  SH2movll4(sh, 6,3, 15);
}
#define init_5_6_4_0
void FASTCALL call_5_6_4_0(SH2_struct* sh)
{
  SH2movll4(sh, 6,4, 0);
}
#define init_5_6_4_1
void FASTCALL call_5_6_4_1(SH2_struct* sh)
{
  SH2movll4(sh, 6,4, 1);
}
#define init_5_6_4_2
void FASTCALL call_5_6_4_2(SH2_struct* sh)
{
  SH2movll4(sh, 6,4, 2);
}
#define init_5_6_4_3
void FASTCALL call_5_6_4_3(SH2_struct* sh)
{
  SH2movll4(sh, 6,4, 3);
}
#define init_5_6_4_4
void FASTCALL call_5_6_4_4(SH2_struct* sh)
{
  SH2movll4(sh, 6,4, 4);
}
#define init_5_6_4_5
void FASTCALL call_5_6_4_5(SH2_struct* sh)
{
  SH2movll4(sh, 6,4, 5);
}
#define init_5_6_4_6
void FASTCALL call_5_6_4_6(SH2_struct* sh)
{
  SH2movll4(sh, 6,4, 6);
}
#define init_5_6_4_7
void FASTCALL call_5_6_4_7(SH2_struct* sh)
{
  SH2movll4(sh, 6,4, 7);
}
#define init_5_6_4_8
void FASTCALL call_5_6_4_8(SH2_struct* sh)
{
  SH2movll4(sh, 6,4, 8);
}
#define init_5_6_4_9
void FASTCALL call_5_6_4_9(SH2_struct* sh)
{
  SH2movll4(sh, 6,4, 9);
}
#define init_5_6_4_10
void FASTCALL call_5_6_4_10(SH2_struct* sh)
{
  SH2movll4(sh, 6,4, 10);
}
#define init_5_6_4_11
void FASTCALL call_5_6_4_11(SH2_struct* sh)
{
  SH2movll4(sh, 6,4, 11);
}
#define init_5_6_4_12
void FASTCALL call_5_6_4_12(SH2_struct* sh)
{
  SH2movll4(sh, 6,4, 12);
}
#define init_5_6_4_13
void FASTCALL call_5_6_4_13(SH2_struct* sh)
{
  SH2movll4(sh, 6,4, 13);
}
#define init_5_6_4_14
void FASTCALL call_5_6_4_14(SH2_struct* sh)
{
  SH2movll4(sh, 6,4, 14);
}
#define init_5_6_4_15
void FASTCALL call_5_6_4_15(SH2_struct* sh)
{
  SH2movll4(sh, 6,4, 15);
}
#define init_5_6_5_0
void FASTCALL call_5_6_5_0(SH2_struct* sh)
{
  SH2movll4(sh, 6,5, 0);
}
#define init_5_6_5_1
void FASTCALL call_5_6_5_1(SH2_struct* sh)
{
  SH2movll4(sh, 6,5, 1);
}
#define init_5_6_5_2
void FASTCALL call_5_6_5_2(SH2_struct* sh)
{
  SH2movll4(sh, 6,5, 2);
}
#define init_5_6_5_3
void FASTCALL call_5_6_5_3(SH2_struct* sh)
{
  SH2movll4(sh, 6,5, 3);
}
#define init_5_6_5_4
void FASTCALL call_5_6_5_4(SH2_struct* sh)
{
  SH2movll4(sh, 6,5, 4);
}
#define init_5_6_5_5
void FASTCALL call_5_6_5_5(SH2_struct* sh)
{
  SH2movll4(sh, 6,5, 5);
}
#define init_5_6_5_6
void FASTCALL call_5_6_5_6(SH2_struct* sh)
{
  SH2movll4(sh, 6,5, 6);
}
#define init_5_6_5_7
void FASTCALL call_5_6_5_7(SH2_struct* sh)
{
  SH2movll4(sh, 6,5, 7);
}
#define init_5_6_5_8
void FASTCALL call_5_6_5_8(SH2_struct* sh)
{
  SH2movll4(sh, 6,5, 8);
}
#define init_5_6_5_9
void FASTCALL call_5_6_5_9(SH2_struct* sh)
{
  SH2movll4(sh, 6,5, 9);
}
#define init_5_6_5_10
void FASTCALL call_5_6_5_10(SH2_struct* sh)
{
  SH2movll4(sh, 6,5, 10);
}
#define init_5_6_5_11
void FASTCALL call_5_6_5_11(SH2_struct* sh)
{
  SH2movll4(sh, 6,5, 11);
}
#define init_5_6_5_12
void FASTCALL call_5_6_5_12(SH2_struct* sh)
{
  SH2movll4(sh, 6,5, 12);
}
#define init_5_6_5_13
void FASTCALL call_5_6_5_13(SH2_struct* sh)
{
  SH2movll4(sh, 6,5, 13);
}
#define init_5_6_5_14
void FASTCALL call_5_6_5_14(SH2_struct* sh)
{
  SH2movll4(sh, 6,5, 14);
}
#define init_5_6_5_15
void FASTCALL call_5_6_5_15(SH2_struct* sh)
{
  SH2movll4(sh, 6,5, 15);
}
#define init_5_6_6_0
void FASTCALL call_5_6_6_0(SH2_struct* sh)
{
  SH2movll4(sh, 6,6, 0);
}
#define init_5_6_6_1
void FASTCALL call_5_6_6_1(SH2_struct* sh)
{
  SH2movll4(sh, 6,6, 1);
}
#define init_5_6_6_2
void FASTCALL call_5_6_6_2(SH2_struct* sh)
{
  SH2movll4(sh, 6,6, 2);
}
#define init_5_6_6_3
void FASTCALL call_5_6_6_3(SH2_struct* sh)
{
  SH2movll4(sh, 6,6, 3);
}
#define init_5_6_6_4
void FASTCALL call_5_6_6_4(SH2_struct* sh)
{
  SH2movll4(sh, 6,6, 4);
}
#define init_5_6_6_5
void FASTCALL call_5_6_6_5(SH2_struct* sh)
{
  SH2movll4(sh, 6,6, 5);
}
#define init_5_6_6_6
void FASTCALL call_5_6_6_6(SH2_struct* sh)
{
  SH2movll4(sh, 6,6, 6);
}
#define init_5_6_6_7
void FASTCALL call_5_6_6_7(SH2_struct* sh)
{
  SH2movll4(sh, 6,6, 7);
}
#define init_5_6_6_8
void FASTCALL call_5_6_6_8(SH2_struct* sh)
{
  SH2movll4(sh, 6,6, 8);
}
#define init_5_6_6_9
void FASTCALL call_5_6_6_9(SH2_struct* sh)
{
  SH2movll4(sh, 6,6, 9);
}
#define init_5_6_6_10
void FASTCALL call_5_6_6_10(SH2_struct* sh)
{
  SH2movll4(sh, 6,6, 10);
}
#define init_5_6_6_11
void FASTCALL call_5_6_6_11(SH2_struct* sh)
{
  SH2movll4(sh, 6,6, 11);
}
#define init_5_6_6_12
void FASTCALL call_5_6_6_12(SH2_struct* sh)
{
  SH2movll4(sh, 6,6, 12);
}
#define init_5_6_6_13
void FASTCALL call_5_6_6_13(SH2_struct* sh)
{
  SH2movll4(sh, 6,6, 13);
}
#define init_5_6_6_14
void FASTCALL call_5_6_6_14(SH2_struct* sh)
{
  SH2movll4(sh, 6,6, 14);
}
#define init_5_6_6_15
void FASTCALL call_5_6_6_15(SH2_struct* sh)
{
  SH2movll4(sh, 6,6, 15);
}
#define init_5_6_7_0
void FASTCALL call_5_6_7_0(SH2_struct* sh)
{
  SH2movll4(sh, 6,7, 0);
}
#define init_5_6_7_1
void FASTCALL call_5_6_7_1(SH2_struct* sh)
{
  SH2movll4(sh, 6,7, 1);
}
#define init_5_6_7_2
void FASTCALL call_5_6_7_2(SH2_struct* sh)
{
  SH2movll4(sh, 6,7, 2);
}
#define init_5_6_7_3
void FASTCALL call_5_6_7_3(SH2_struct* sh)
{
  SH2movll4(sh, 6,7, 3);
}
#define init_5_6_7_4
void FASTCALL call_5_6_7_4(SH2_struct* sh)
{
  SH2movll4(sh, 6,7, 4);
}
#define init_5_6_7_5
void FASTCALL call_5_6_7_5(SH2_struct* sh)
{
  SH2movll4(sh, 6,7, 5);
}
#define init_5_6_7_6
void FASTCALL call_5_6_7_6(SH2_struct* sh)
{
  SH2movll4(sh, 6,7, 6);
}
#define init_5_6_7_7
void FASTCALL call_5_6_7_7(SH2_struct* sh)
{
  SH2movll4(sh, 6,7, 7);
}
#define init_5_6_7_8
void FASTCALL call_5_6_7_8(SH2_struct* sh)
{
  SH2movll4(sh, 6,7, 8);
}
#define init_5_6_7_9
void FASTCALL call_5_6_7_9(SH2_struct* sh)
{
  SH2movll4(sh, 6,7, 9);
}
#define init_5_6_7_10
void FASTCALL call_5_6_7_10(SH2_struct* sh)
{
  SH2movll4(sh, 6,7, 10);
}
#define init_5_6_7_11
void FASTCALL call_5_6_7_11(SH2_struct* sh)
{
  SH2movll4(sh, 6,7, 11);
}
#define init_5_6_7_12
void FASTCALL call_5_6_7_12(SH2_struct* sh)
{
  SH2movll4(sh, 6,7, 12);
}
#define init_5_6_7_13
void FASTCALL call_5_6_7_13(SH2_struct* sh)
{
  SH2movll4(sh, 6,7, 13);
}
#define init_5_6_7_14
void FASTCALL call_5_6_7_14(SH2_struct* sh)
{
  SH2movll4(sh, 6,7, 14);
}
#define init_5_6_7_15
void FASTCALL call_5_6_7_15(SH2_struct* sh)
{
  SH2movll4(sh, 6,7, 15);
}
#define init_5_6_8_0
void FASTCALL call_5_6_8_0(SH2_struct* sh)
{
  SH2movll4(sh, 6,8, 0);
}
#define init_5_6_8_1
void FASTCALL call_5_6_8_1(SH2_struct* sh)
{
  SH2movll4(sh, 6,8, 1);
}
#define init_5_6_8_2
void FASTCALL call_5_6_8_2(SH2_struct* sh)
{
  SH2movll4(sh, 6,8, 2);
}
#define init_5_6_8_3
void FASTCALL call_5_6_8_3(SH2_struct* sh)
{
  SH2movll4(sh, 6,8, 3);
}
#define init_5_6_8_4
void FASTCALL call_5_6_8_4(SH2_struct* sh)
{
  SH2movll4(sh, 6,8, 4);
}
#define init_5_6_8_5
void FASTCALL call_5_6_8_5(SH2_struct* sh)
{
  SH2movll4(sh, 6,8, 5);
}
#define init_5_6_8_6
void FASTCALL call_5_6_8_6(SH2_struct* sh)
{
  SH2movll4(sh, 6,8, 6);
}
#define init_5_6_8_7
void FASTCALL call_5_6_8_7(SH2_struct* sh)
{
  SH2movll4(sh, 6,8, 7);
}
#define init_5_6_8_8
void FASTCALL call_5_6_8_8(SH2_struct* sh)
{
  SH2movll4(sh, 6,8, 8);
}
#define init_5_6_8_9
void FASTCALL call_5_6_8_9(SH2_struct* sh)
{
  SH2movll4(sh, 6,8, 9);
}
#define init_5_6_8_10
void FASTCALL call_5_6_8_10(SH2_struct* sh)
{
  SH2movll4(sh, 6,8, 10);
}
#define init_5_6_8_11
void FASTCALL call_5_6_8_11(SH2_struct* sh)
{
  SH2movll4(sh, 6,8, 11);
}
#define init_5_6_8_12
void FASTCALL call_5_6_8_12(SH2_struct* sh)
{
  SH2movll4(sh, 6,8, 12);
}
#define init_5_6_8_13
void FASTCALL call_5_6_8_13(SH2_struct* sh)
{
  SH2movll4(sh, 6,8, 13);
}
#define init_5_6_8_14
void FASTCALL call_5_6_8_14(SH2_struct* sh)
{
  SH2movll4(sh, 6,8, 14);
}
#define init_5_6_8_15
void FASTCALL call_5_6_8_15(SH2_struct* sh)
{
  SH2movll4(sh, 6,8, 15);
}
#define init_5_6_9_0
void FASTCALL call_5_6_9_0(SH2_struct* sh)
{
  SH2movll4(sh, 6,9, 0);
}
#define init_5_6_9_1
void FASTCALL call_5_6_9_1(SH2_struct* sh)
{
  SH2movll4(sh, 6,9, 1);
}
#define init_5_6_9_2
void FASTCALL call_5_6_9_2(SH2_struct* sh)
{
  SH2movll4(sh, 6,9, 2);
}
#define init_5_6_9_3
void FASTCALL call_5_6_9_3(SH2_struct* sh)
{
  SH2movll4(sh, 6,9, 3);
}
#define init_5_6_9_4
void FASTCALL call_5_6_9_4(SH2_struct* sh)
{
  SH2movll4(sh, 6,9, 4);
}
#define init_5_6_9_5
void FASTCALL call_5_6_9_5(SH2_struct* sh)
{
  SH2movll4(sh, 6,9, 5);
}
#define init_5_6_9_6
void FASTCALL call_5_6_9_6(SH2_struct* sh)
{
  SH2movll4(sh, 6,9, 6);
}
#define init_5_6_9_7
void FASTCALL call_5_6_9_7(SH2_struct* sh)
{
  SH2movll4(sh, 6,9, 7);
}
#define init_5_6_9_8
void FASTCALL call_5_6_9_8(SH2_struct* sh)
{
  SH2movll4(sh, 6,9, 8);
}
#define init_5_6_9_9
void FASTCALL call_5_6_9_9(SH2_struct* sh)
{
  SH2movll4(sh, 6,9, 9);
}
#define init_5_6_9_10
void FASTCALL call_5_6_9_10(SH2_struct* sh)
{
  SH2movll4(sh, 6,9, 10);
}
#define init_5_6_9_11
void FASTCALL call_5_6_9_11(SH2_struct* sh)
{
  SH2movll4(sh, 6,9, 11);
}
#define init_5_6_9_12
void FASTCALL call_5_6_9_12(SH2_struct* sh)
{
  SH2movll4(sh, 6,9, 12);
}
#define init_5_6_9_13
void FASTCALL call_5_6_9_13(SH2_struct* sh)
{
  SH2movll4(sh, 6,9, 13);
}
#define init_5_6_9_14
void FASTCALL call_5_6_9_14(SH2_struct* sh)
{
  SH2movll4(sh, 6,9, 14);
}
#define init_5_6_9_15
void FASTCALL call_5_6_9_15(SH2_struct* sh)
{
  SH2movll4(sh, 6,9, 15);
}
#define init_5_6_10_0
void FASTCALL call_5_6_10_0(SH2_struct* sh)
{
  SH2movll4(sh, 6,10, 0);
}
#define init_5_6_10_1
void FASTCALL call_5_6_10_1(SH2_struct* sh)
{
  SH2movll4(sh, 6,10, 1);
}
#define init_5_6_10_2
void FASTCALL call_5_6_10_2(SH2_struct* sh)
{
  SH2movll4(sh, 6,10, 2);
}
#define init_5_6_10_3
void FASTCALL call_5_6_10_3(SH2_struct* sh)
{
  SH2movll4(sh, 6,10, 3);
}
#define init_5_6_10_4
void FASTCALL call_5_6_10_4(SH2_struct* sh)
{
  SH2movll4(sh, 6,10, 4);
}
#define init_5_6_10_5
void FASTCALL call_5_6_10_5(SH2_struct* sh)
{
  SH2movll4(sh, 6,10, 5);
}
#define init_5_6_10_6
void FASTCALL call_5_6_10_6(SH2_struct* sh)
{
  SH2movll4(sh, 6,10, 6);
}
#define init_5_6_10_7
void FASTCALL call_5_6_10_7(SH2_struct* sh)
{
  SH2movll4(sh, 6,10, 7);
}
#define init_5_6_10_8
void FASTCALL call_5_6_10_8(SH2_struct* sh)
{
  SH2movll4(sh, 6,10, 8);
}
#define init_5_6_10_9
void FASTCALL call_5_6_10_9(SH2_struct* sh)
{
  SH2movll4(sh, 6,10, 9);
}
#define init_5_6_10_10
void FASTCALL call_5_6_10_10(SH2_struct* sh)
{
  SH2movll4(sh, 6,10, 10);
}
#define init_5_6_10_11
void FASTCALL call_5_6_10_11(SH2_struct* sh)
{
  SH2movll4(sh, 6,10, 11);
}
#define init_5_6_10_12
void FASTCALL call_5_6_10_12(SH2_struct* sh)
{
  SH2movll4(sh, 6,10, 12);
}
#define init_5_6_10_13
void FASTCALL call_5_6_10_13(SH2_struct* sh)
{
  SH2movll4(sh, 6,10, 13);
}
#define init_5_6_10_14
void FASTCALL call_5_6_10_14(SH2_struct* sh)
{
  SH2movll4(sh, 6,10, 14);
}
#define init_5_6_10_15
void FASTCALL call_5_6_10_15(SH2_struct* sh)
{
  SH2movll4(sh, 6,10, 15);
}
#define init_5_6_11_0
void FASTCALL call_5_6_11_0(SH2_struct* sh)
{
  SH2movll4(sh, 6,11, 0);
}
#define init_5_6_11_1
void FASTCALL call_5_6_11_1(SH2_struct* sh)
{
  SH2movll4(sh, 6,11, 1);
}
#define init_5_6_11_2
void FASTCALL call_5_6_11_2(SH2_struct* sh)
{
  SH2movll4(sh, 6,11, 2);
}
#define init_5_6_11_3
void FASTCALL call_5_6_11_3(SH2_struct* sh)
{
  SH2movll4(sh, 6,11, 3);
}
#define init_5_6_11_4
void FASTCALL call_5_6_11_4(SH2_struct* sh)
{
  SH2movll4(sh, 6,11, 4);
}
#define init_5_6_11_5
void FASTCALL call_5_6_11_5(SH2_struct* sh)
{
  SH2movll4(sh, 6,11, 5);
}
#define init_5_6_11_6
void FASTCALL call_5_6_11_6(SH2_struct* sh)
{
  SH2movll4(sh, 6,11, 6);
}
#define init_5_6_11_7
void FASTCALL call_5_6_11_7(SH2_struct* sh)
{
  SH2movll4(sh, 6,11, 7);
}
#define init_5_6_11_8
void FASTCALL call_5_6_11_8(SH2_struct* sh)
{
  SH2movll4(sh, 6,11, 8);
}
#define init_5_6_11_9
void FASTCALL call_5_6_11_9(SH2_struct* sh)
{
  SH2movll4(sh, 6,11, 9);
}
#define init_5_6_11_10
void FASTCALL call_5_6_11_10(SH2_struct* sh)
{
  SH2movll4(sh, 6,11, 10);
}
#define init_5_6_11_11
void FASTCALL call_5_6_11_11(SH2_struct* sh)
{
  SH2movll4(sh, 6,11, 11);
}
#define init_5_6_11_12
void FASTCALL call_5_6_11_12(SH2_struct* sh)
{
  SH2movll4(sh, 6,11, 12);
}
#define init_5_6_11_13
void FASTCALL call_5_6_11_13(SH2_struct* sh)
{
  SH2movll4(sh, 6,11, 13);
}
#define init_5_6_11_14
void FASTCALL call_5_6_11_14(SH2_struct* sh)
{
  SH2movll4(sh, 6,11, 14);
}
#define init_5_6_11_15
void FASTCALL call_5_6_11_15(SH2_struct* sh)
{
  SH2movll4(sh, 6,11, 15);
}
#define init_5_6_12_0
void FASTCALL call_5_6_12_0(SH2_struct* sh)
{
  SH2movll4(sh, 6,12, 0);
}
#define init_5_6_12_1
void FASTCALL call_5_6_12_1(SH2_struct* sh)
{
  SH2movll4(sh, 6,12, 1);
}
#define init_5_6_12_2
void FASTCALL call_5_6_12_2(SH2_struct* sh)
{
  SH2movll4(sh, 6,12, 2);
}
#define init_5_6_12_3
void FASTCALL call_5_6_12_3(SH2_struct* sh)
{
  SH2movll4(sh, 6,12, 3);
}
#define init_5_6_12_4
void FASTCALL call_5_6_12_4(SH2_struct* sh)
{
  SH2movll4(sh, 6,12, 4);
}
#define init_5_6_12_5
void FASTCALL call_5_6_12_5(SH2_struct* sh)
{
  SH2movll4(sh, 6,12, 5);
}
#define init_5_6_12_6
void FASTCALL call_5_6_12_6(SH2_struct* sh)
{
  SH2movll4(sh, 6,12, 6);
}
#define init_5_6_12_7
void FASTCALL call_5_6_12_7(SH2_struct* sh)
{
  SH2movll4(sh, 6,12, 7);
}
#define init_5_6_12_8
void FASTCALL call_5_6_12_8(SH2_struct* sh)
{
  SH2movll4(sh, 6,12, 8);
}
#define init_5_6_12_9
void FASTCALL call_5_6_12_9(SH2_struct* sh)
{
  SH2movll4(sh, 6,12, 9);
}
#define init_5_6_12_10
void FASTCALL call_5_6_12_10(SH2_struct* sh)
{
  SH2movll4(sh, 6,12, 10);
}
#define init_5_6_12_11
void FASTCALL call_5_6_12_11(SH2_struct* sh)
{
  SH2movll4(sh, 6,12, 11);
}
#define init_5_6_12_12
void FASTCALL call_5_6_12_12(SH2_struct* sh)
{
  SH2movll4(sh, 6,12, 12);
}
#define init_5_6_12_13
void FASTCALL call_5_6_12_13(SH2_struct* sh)
{
  SH2movll4(sh, 6,12, 13);
}
#define init_5_6_12_14
void FASTCALL call_5_6_12_14(SH2_struct* sh)
{
  SH2movll4(sh, 6,12, 14);
}
#define init_5_6_12_15
void FASTCALL call_5_6_12_15(SH2_struct* sh)
{
  SH2movll4(sh, 6,12, 15);
}
#define init_5_6_13_0
void FASTCALL call_5_6_13_0(SH2_struct* sh)
{
  SH2movll4(sh, 6,13, 0);
}
#define init_5_6_13_1
void FASTCALL call_5_6_13_1(SH2_struct* sh)
{
  SH2movll4(sh, 6,13, 1);
}
#define init_5_6_13_2
void FASTCALL call_5_6_13_2(SH2_struct* sh)
{
  SH2movll4(sh, 6,13, 2);
}
#define init_5_6_13_3
void FASTCALL call_5_6_13_3(SH2_struct* sh)
{
  SH2movll4(sh, 6,13, 3);
}
#define init_5_6_13_4
void FASTCALL call_5_6_13_4(SH2_struct* sh)
{
  SH2movll4(sh, 6,13, 4);
}
#define init_5_6_13_5
void FASTCALL call_5_6_13_5(SH2_struct* sh)
{
  SH2movll4(sh, 6,13, 5);
}
#define init_5_6_13_6
void FASTCALL call_5_6_13_6(SH2_struct* sh)
{
  SH2movll4(sh, 6,13, 6);
}
#define init_5_6_13_7
void FASTCALL call_5_6_13_7(SH2_struct* sh)
{
  SH2movll4(sh, 6,13, 7);
}
#define init_5_6_13_8
void FASTCALL call_5_6_13_8(SH2_struct* sh)
{
  SH2movll4(sh, 6,13, 8);
}
#define init_5_6_13_9
void FASTCALL call_5_6_13_9(SH2_struct* sh)
{
  SH2movll4(sh, 6,13, 9);
}
#define init_5_6_13_10
void FASTCALL call_5_6_13_10(SH2_struct* sh)
{
  SH2movll4(sh, 6,13, 10);
}
#define init_5_6_13_11
void FASTCALL call_5_6_13_11(SH2_struct* sh)
{
  SH2movll4(sh, 6,13, 11);
}
#define init_5_6_13_12
void FASTCALL call_5_6_13_12(SH2_struct* sh)
{
  SH2movll4(sh, 6,13, 12);
}
#define init_5_6_13_13
void FASTCALL call_5_6_13_13(SH2_struct* sh)
{
  SH2movll4(sh, 6,13, 13);
}
#define init_5_6_13_14
void FASTCALL call_5_6_13_14(SH2_struct* sh)
{
  SH2movll4(sh, 6,13, 14);
}
#define init_5_6_13_15
void FASTCALL call_5_6_13_15(SH2_struct* sh)
{
  SH2movll4(sh, 6,13, 15);
}
#define init_5_6_14_0
void FASTCALL call_5_6_14_0(SH2_struct* sh)
{
  SH2movll4(sh, 6,14, 0);
}
#define init_5_6_14_1
void FASTCALL call_5_6_14_1(SH2_struct* sh)
{
  SH2movll4(sh, 6,14, 1);
}
#define init_5_6_14_2
void FASTCALL call_5_6_14_2(SH2_struct* sh)
{
  SH2movll4(sh, 6,14, 2);
}
#define init_5_6_14_3
void FASTCALL call_5_6_14_3(SH2_struct* sh)
{
  SH2movll4(sh, 6,14, 3);
}
#define init_5_6_14_4
void FASTCALL call_5_6_14_4(SH2_struct* sh)
{
  SH2movll4(sh, 6,14, 4);
}
#define init_5_6_14_5
void FASTCALL call_5_6_14_5(SH2_struct* sh)
{
  SH2movll4(sh, 6,14, 5);
}
#define init_5_6_14_6
void FASTCALL call_5_6_14_6(SH2_struct* sh)
{
  SH2movll4(sh, 6,14, 6);
}
#define init_5_6_14_7
void FASTCALL call_5_6_14_7(SH2_struct* sh)
{
  SH2movll4(sh, 6,14, 7);
}
#define init_5_6_14_8
void FASTCALL call_5_6_14_8(SH2_struct* sh)
{
  SH2movll4(sh, 6,14, 8);
}
#define init_5_6_14_9
void FASTCALL call_5_6_14_9(SH2_struct* sh)
{
  SH2movll4(sh, 6,14, 9);
}
#define init_5_6_14_10
void FASTCALL call_5_6_14_10(SH2_struct* sh)
{
  SH2movll4(sh, 6,14, 10);
}
#define init_5_6_14_11
void FASTCALL call_5_6_14_11(SH2_struct* sh)
{
  SH2movll4(sh, 6,14, 11);
}
#define init_5_6_14_12
void FASTCALL call_5_6_14_12(SH2_struct* sh)
{
  SH2movll4(sh, 6,14, 12);
}
#define init_5_6_14_13
void FASTCALL call_5_6_14_13(SH2_struct* sh)
{
  SH2movll4(sh, 6,14, 13);
}
#define init_5_6_14_14
void FASTCALL call_5_6_14_14(SH2_struct* sh)
{
  SH2movll4(sh, 6,14, 14);
}
#define init_5_6_14_15
void FASTCALL call_5_6_14_15(SH2_struct* sh)
{
  SH2movll4(sh, 6,14, 15);
}
#define init_5_6_15_0
void FASTCALL call_5_6_15_0(SH2_struct* sh)
{
  SH2movll4(sh, 6,15, 0);
}
#define init_5_6_15_1
void FASTCALL call_5_6_15_1(SH2_struct* sh)
{
  SH2movll4(sh, 6,15, 1);
}
#define init_5_6_15_2
void FASTCALL call_5_6_15_2(SH2_struct* sh)
{
  SH2movll4(sh, 6,15, 2);
}
#define init_5_6_15_3
void FASTCALL call_5_6_15_3(SH2_struct* sh)
{
  SH2movll4(sh, 6,15, 3);
}
#define init_5_6_15_4
void FASTCALL call_5_6_15_4(SH2_struct* sh)
{
  SH2movll4(sh, 6,15, 4);
}
#define init_5_6_15_5
void FASTCALL call_5_6_15_5(SH2_struct* sh)
{
  SH2movll4(sh, 6,15, 5);
}
#define init_5_6_15_6
void FASTCALL call_5_6_15_6(SH2_struct* sh)
{
  SH2movll4(sh, 6,15, 6);
}
#define init_5_6_15_7
void FASTCALL call_5_6_15_7(SH2_struct* sh)
{
  SH2movll4(sh, 6,15, 7);
}
#define init_5_6_15_8
void FASTCALL call_5_6_15_8(SH2_struct* sh)
{
  SH2movll4(sh, 6,15, 8);
}
#define init_5_6_15_9
void FASTCALL call_5_6_15_9(SH2_struct* sh)
{
  SH2movll4(sh, 6,15, 9);
}
#define init_5_6_15_10
void FASTCALL call_5_6_15_10(SH2_struct* sh)
{
  SH2movll4(sh, 6,15, 10);
}
#define init_5_6_15_11
void FASTCALL call_5_6_15_11(SH2_struct* sh)
{
  SH2movll4(sh, 6,15, 11);
}
#define init_5_6_15_12
void FASTCALL call_5_6_15_12(SH2_struct* sh)
{
  SH2movll4(sh, 6,15, 12);
}
#define init_5_6_15_13
void FASTCALL call_5_6_15_13(SH2_struct* sh)
{
  SH2movll4(sh, 6,15, 13);
}
#define init_5_6_15_14
void FASTCALL call_5_6_15_14(SH2_struct* sh)
{
  SH2movll4(sh, 6,15, 14);
}
#define init_5_6_15_15
void FASTCALL call_5_6_15_15(SH2_struct* sh)
{
  SH2movll4(sh, 6,15, 15);
}
#define init_5_7_0_0
void FASTCALL call_5_7_0_0(SH2_struct* sh)
{
  SH2movll4(sh, 7,0, 0);
}
#define init_5_7_0_1
void FASTCALL call_5_7_0_1(SH2_struct* sh)
{
  SH2movll4(sh, 7,0, 1);
}
#define init_5_7_0_2
void FASTCALL call_5_7_0_2(SH2_struct* sh)
{
  SH2movll4(sh, 7,0, 2);
}
#define init_5_7_0_3
void FASTCALL call_5_7_0_3(SH2_struct* sh)
{
  SH2movll4(sh, 7,0, 3);
}
#define init_5_7_0_4
void FASTCALL call_5_7_0_4(SH2_struct* sh)
{
  SH2movll4(sh, 7,0, 4);
}
#define init_5_7_0_5
void FASTCALL call_5_7_0_5(SH2_struct* sh)
{
  SH2movll4(sh, 7,0, 5);
}
#define init_5_7_0_6
void FASTCALL call_5_7_0_6(SH2_struct* sh)
{
  SH2movll4(sh, 7,0, 6);
}
#define init_5_7_0_7
void FASTCALL call_5_7_0_7(SH2_struct* sh)
{
  SH2movll4(sh, 7,0, 7);
}
#define init_5_7_0_8
void FASTCALL call_5_7_0_8(SH2_struct* sh)
{
  SH2movll4(sh, 7,0, 8);
}
#define init_5_7_0_9
void FASTCALL call_5_7_0_9(SH2_struct* sh)
{
  SH2movll4(sh, 7,0, 9);
}
#define init_5_7_0_10
void FASTCALL call_5_7_0_10(SH2_struct* sh)
{
  SH2movll4(sh, 7,0, 10);
}
#define init_5_7_0_11
void FASTCALL call_5_7_0_11(SH2_struct* sh)
{
  SH2movll4(sh, 7,0, 11);
}
#define init_5_7_0_12
void FASTCALL call_5_7_0_12(SH2_struct* sh)
{
  SH2movll4(sh, 7,0, 12);
}
#define init_5_7_0_13
void FASTCALL call_5_7_0_13(SH2_struct* sh)
{
  SH2movll4(sh, 7,0, 13);
}
#define init_5_7_0_14
void FASTCALL call_5_7_0_14(SH2_struct* sh)
{
  SH2movll4(sh, 7,0, 14);
}
#define init_5_7_0_15
void FASTCALL call_5_7_0_15(SH2_struct* sh)
{
  SH2movll4(sh, 7,0, 15);
}
#define init_5_7_1_0
void FASTCALL call_5_7_1_0(SH2_struct* sh)
{
  SH2movll4(sh, 7,1, 0);
}
#define init_5_7_1_1
void FASTCALL call_5_7_1_1(SH2_struct* sh)
{
  SH2movll4(sh, 7,1, 1);
}
#define init_5_7_1_2
void FASTCALL call_5_7_1_2(SH2_struct* sh)
{
  SH2movll4(sh, 7,1, 2);
}
#define init_5_7_1_3
void FASTCALL call_5_7_1_3(SH2_struct* sh)
{
  SH2movll4(sh, 7,1, 3);
}
#define init_5_7_1_4
void FASTCALL call_5_7_1_4(SH2_struct* sh)
{
  SH2movll4(sh, 7,1, 4);
}
#define init_5_7_1_5
void FASTCALL call_5_7_1_5(SH2_struct* sh)
{
  SH2movll4(sh, 7,1, 5);
}
#define init_5_7_1_6
void FASTCALL call_5_7_1_6(SH2_struct* sh)
{
  SH2movll4(sh, 7,1, 6);
}
#define init_5_7_1_7
void FASTCALL call_5_7_1_7(SH2_struct* sh)
{
  SH2movll4(sh, 7,1, 7);
}
#define init_5_7_1_8
void FASTCALL call_5_7_1_8(SH2_struct* sh)
{
  SH2movll4(sh, 7,1, 8);
}
#define init_5_7_1_9
void FASTCALL call_5_7_1_9(SH2_struct* sh)
{
  SH2movll4(sh, 7,1, 9);
}
#define init_5_7_1_10
void FASTCALL call_5_7_1_10(SH2_struct* sh)
{
  SH2movll4(sh, 7,1, 10);
}
#define init_5_7_1_11
void FASTCALL call_5_7_1_11(SH2_struct* sh)
{
  SH2movll4(sh, 7,1, 11);
}
#define init_5_7_1_12
void FASTCALL call_5_7_1_12(SH2_struct* sh)
{
  SH2movll4(sh, 7,1, 12);
}
#define init_5_7_1_13
void FASTCALL call_5_7_1_13(SH2_struct* sh)
{
  SH2movll4(sh, 7,1, 13);
}
#define init_5_7_1_14
void FASTCALL call_5_7_1_14(SH2_struct* sh)
{
  SH2movll4(sh, 7,1, 14);
}
#define init_5_7_1_15
void FASTCALL call_5_7_1_15(SH2_struct* sh)
{
  SH2movll4(sh, 7,1, 15);
}
#define init_5_7_2_0
void FASTCALL call_5_7_2_0(SH2_struct* sh)
{
  SH2movll4(sh, 7,2, 0);
}
#define init_5_7_2_1
void FASTCALL call_5_7_2_1(SH2_struct* sh)
{
  SH2movll4(sh, 7,2, 1);
}
#define init_5_7_2_2
void FASTCALL call_5_7_2_2(SH2_struct* sh)
{
  SH2movll4(sh, 7,2, 2);
}
#define init_5_7_2_3
void FASTCALL call_5_7_2_3(SH2_struct* sh)
{
  SH2movll4(sh, 7,2, 3);
}
#define init_5_7_2_4
void FASTCALL call_5_7_2_4(SH2_struct* sh)
{
  SH2movll4(sh, 7,2, 4);
}
#define init_5_7_2_5
void FASTCALL call_5_7_2_5(SH2_struct* sh)
{
  SH2movll4(sh, 7,2, 5);
}
#define init_5_7_2_6
void FASTCALL call_5_7_2_6(SH2_struct* sh)
{
  SH2movll4(sh, 7,2, 6);
}
#define init_5_7_2_7
void FASTCALL call_5_7_2_7(SH2_struct* sh)
{
  SH2movll4(sh, 7,2, 7);
}
#define init_5_7_2_8
void FASTCALL call_5_7_2_8(SH2_struct* sh)
{
  SH2movll4(sh, 7,2, 8);
}
#define init_5_7_2_9
void FASTCALL call_5_7_2_9(SH2_struct* sh)
{
  SH2movll4(sh, 7,2, 9);
}
#define init_5_7_2_10
void FASTCALL call_5_7_2_10(SH2_struct* sh)
{
  SH2movll4(sh, 7,2, 10);
}
#define init_5_7_2_11
void FASTCALL call_5_7_2_11(SH2_struct* sh)
{
  SH2movll4(sh, 7,2, 11);
}
#define init_5_7_2_12
void FASTCALL call_5_7_2_12(SH2_struct* sh)
{
  SH2movll4(sh, 7,2, 12);
}
#define init_5_7_2_13
void FASTCALL call_5_7_2_13(SH2_struct* sh)
{
  SH2movll4(sh, 7,2, 13);
}
#define init_5_7_2_14
void FASTCALL call_5_7_2_14(SH2_struct* sh)
{
  SH2movll4(sh, 7,2, 14);
}
#define init_5_7_2_15
void FASTCALL call_5_7_2_15(SH2_struct* sh)
{
  SH2movll4(sh, 7,2, 15);
}
#define init_5_7_3_0
void FASTCALL call_5_7_3_0(SH2_struct* sh)
{
  SH2movll4(sh, 7,3, 0);
}
#define init_5_7_3_1
void FASTCALL call_5_7_3_1(SH2_struct* sh)
{
  SH2movll4(sh, 7,3, 1);
}
#define init_5_7_3_2
void FASTCALL call_5_7_3_2(SH2_struct* sh)
{
  SH2movll4(sh, 7,3, 2);
}
#define init_5_7_3_3
void FASTCALL call_5_7_3_3(SH2_struct* sh)
{
  SH2movll4(sh, 7,3, 3);
}
#define init_5_7_3_4
void FASTCALL call_5_7_3_4(SH2_struct* sh)
{
  SH2movll4(sh, 7,3, 4);
}
#define init_5_7_3_5
void FASTCALL call_5_7_3_5(SH2_struct* sh)
{
  SH2movll4(sh, 7,3, 5);
}
#define init_5_7_3_6
void FASTCALL call_5_7_3_6(SH2_struct* sh)
{
  SH2movll4(sh, 7,3, 6);
}
#define init_5_7_3_7
void FASTCALL call_5_7_3_7(SH2_struct* sh)
{
  SH2movll4(sh, 7,3, 7);
}
#define init_5_7_3_8
void FASTCALL call_5_7_3_8(SH2_struct* sh)
{
  SH2movll4(sh, 7,3, 8);
}
#define init_5_7_3_9
void FASTCALL call_5_7_3_9(SH2_struct* sh)
{
  SH2movll4(sh, 7,3, 9);
}
#define init_5_7_3_10
void FASTCALL call_5_7_3_10(SH2_struct* sh)
{
  SH2movll4(sh, 7,3, 10);
}
#define init_5_7_3_11
void FASTCALL call_5_7_3_11(SH2_struct* sh)
{
  SH2movll4(sh, 7,3, 11);
}
#define init_5_7_3_12
void FASTCALL call_5_7_3_12(SH2_struct* sh)
{
  SH2movll4(sh, 7,3, 12);
}
#define init_5_7_3_13
void FASTCALL call_5_7_3_13(SH2_struct* sh)
{
  SH2movll4(sh, 7,3, 13);
}
#define init_5_7_3_14
void FASTCALL call_5_7_3_14(SH2_struct* sh)
{
  SH2movll4(sh, 7,3, 14);
}
#define init_5_7_3_15
void FASTCALL call_5_7_3_15(SH2_struct* sh)
{
  SH2movll4(sh, 7,3, 15);
}
#define init_5_7_4_0
void FASTCALL call_5_7_4_0(SH2_struct* sh)
{
  SH2movll4(sh, 7,4, 0);
}
#define init_5_7_4_1
void FASTCALL call_5_7_4_1(SH2_struct* sh)
{
  SH2movll4(sh, 7,4, 1);
}
#define init_5_7_4_2
void FASTCALL call_5_7_4_2(SH2_struct* sh)
{
  SH2movll4(sh, 7,4, 2);
}
#define init_5_7_4_3
void FASTCALL call_5_7_4_3(SH2_struct* sh)
{
  SH2movll4(sh, 7,4, 3);
}
#define init_5_7_4_4
void FASTCALL call_5_7_4_4(SH2_struct* sh)
{
  SH2movll4(sh, 7,4, 4);
}
#define init_5_7_4_5
void FASTCALL call_5_7_4_5(SH2_struct* sh)
{
  SH2movll4(sh, 7,4, 5);
}
#define init_5_7_4_6
void FASTCALL call_5_7_4_6(SH2_struct* sh)
{
  SH2movll4(sh, 7,4, 6);
}
#define init_5_7_4_7
void FASTCALL call_5_7_4_7(SH2_struct* sh)
{
  SH2movll4(sh, 7,4, 7);
}
#define init_5_7_4_8
void FASTCALL call_5_7_4_8(SH2_struct* sh)
{
  SH2movll4(sh, 7,4, 8);
}
#define init_5_7_4_9
void FASTCALL call_5_7_4_9(SH2_struct* sh)
{
  SH2movll4(sh, 7,4, 9);
}
#define init_5_7_4_10
void FASTCALL call_5_7_4_10(SH2_struct* sh)
{
  SH2movll4(sh, 7,4, 10);
}
#define init_5_7_4_11
void FASTCALL call_5_7_4_11(SH2_struct* sh)
{
  SH2movll4(sh, 7,4, 11);
}
#define init_5_7_4_12
void FASTCALL call_5_7_4_12(SH2_struct* sh)
{
  SH2movll4(sh, 7,4, 12);
}
#define init_5_7_4_13
void FASTCALL call_5_7_4_13(SH2_struct* sh)
{
  SH2movll4(sh, 7,4, 13);
}
#define init_5_7_4_14
void FASTCALL call_5_7_4_14(SH2_struct* sh)
{
  SH2movll4(sh, 7,4, 14);
}
#define init_5_7_4_15
void FASTCALL call_5_7_4_15(SH2_struct* sh)
{
  SH2movll4(sh, 7,4, 15);
}
#define init_5_7_5_0
void FASTCALL call_5_7_5_0(SH2_struct* sh)
{
  SH2movll4(sh, 7,5, 0);
}
#define init_5_7_5_1
void FASTCALL call_5_7_5_1(SH2_struct* sh)
{
  SH2movll4(sh, 7,5, 1);
}
#define init_5_7_5_2
void FASTCALL call_5_7_5_2(SH2_struct* sh)
{
  SH2movll4(sh, 7,5, 2);
}
#define init_5_7_5_3
void FASTCALL call_5_7_5_3(SH2_struct* sh)
{
  SH2movll4(sh, 7,5, 3);
}
#define init_5_7_5_4
void FASTCALL call_5_7_5_4(SH2_struct* sh)
{
  SH2movll4(sh, 7,5, 4);
}
#define init_5_7_5_5
void FASTCALL call_5_7_5_5(SH2_struct* sh)
{
  SH2movll4(sh, 7,5, 5);
}
#define init_5_7_5_6
void FASTCALL call_5_7_5_6(SH2_struct* sh)
{
  SH2movll4(sh, 7,5, 6);
}
#define init_5_7_5_7
void FASTCALL call_5_7_5_7(SH2_struct* sh)
{
  SH2movll4(sh, 7,5, 7);
}
#define init_5_7_5_8
void FASTCALL call_5_7_5_8(SH2_struct* sh)
{
  SH2movll4(sh, 7,5, 8);
}
#define init_5_7_5_9
void FASTCALL call_5_7_5_9(SH2_struct* sh)
{
  SH2movll4(sh, 7,5, 9);
}
#define init_5_7_5_10
void FASTCALL call_5_7_5_10(SH2_struct* sh)
{
  SH2movll4(sh, 7,5, 10);
}
#define init_5_7_5_11
void FASTCALL call_5_7_5_11(SH2_struct* sh)
{
  SH2movll4(sh, 7,5, 11);
}
#define init_5_7_5_12
void FASTCALL call_5_7_5_12(SH2_struct* sh)
{
  SH2movll4(sh, 7,5, 12);
}
#define init_5_7_5_13
void FASTCALL call_5_7_5_13(SH2_struct* sh)
{
  SH2movll4(sh, 7,5, 13);
}
#define init_5_7_5_14
void FASTCALL call_5_7_5_14(SH2_struct* sh)
{
  SH2movll4(sh, 7,5, 14);
}
#define init_5_7_5_15
void FASTCALL call_5_7_5_15(SH2_struct* sh)
{
  SH2movll4(sh, 7,5, 15);
}
#define init_5_7_6_0
void FASTCALL call_5_7_6_0(SH2_struct* sh)
{
  SH2movll4(sh, 7,6, 0);
}
#define init_5_7_6_1
void FASTCALL call_5_7_6_1(SH2_struct* sh)
{
  SH2movll4(sh, 7,6, 1);
}
#define init_5_7_6_2
void FASTCALL call_5_7_6_2(SH2_struct* sh)
{
  SH2movll4(sh, 7,6, 2);
}
#define init_5_7_6_3
void FASTCALL call_5_7_6_3(SH2_struct* sh)
{
  SH2movll4(sh, 7,6, 3);
}
#define init_5_7_6_4
void FASTCALL call_5_7_6_4(SH2_struct* sh)
{
  SH2movll4(sh, 7,6, 4);
}
#define init_5_7_6_5
void FASTCALL call_5_7_6_5(SH2_struct* sh)
{
  SH2movll4(sh, 7,6, 5);
}
#define init_5_7_6_6
void FASTCALL call_5_7_6_6(SH2_struct* sh)
{
  SH2movll4(sh, 7,6, 6);
}
#define init_5_7_6_7
void FASTCALL call_5_7_6_7(SH2_struct* sh)
{
  SH2movll4(sh, 7,6, 7);
}
#define init_5_7_6_8
void FASTCALL call_5_7_6_8(SH2_struct* sh)
{
  SH2movll4(sh, 7,6, 8);
}
#define init_5_7_6_9
void FASTCALL call_5_7_6_9(SH2_struct* sh)
{
  SH2movll4(sh, 7,6, 9);
}
#define init_5_7_6_10
void FASTCALL call_5_7_6_10(SH2_struct* sh)
{
  SH2movll4(sh, 7,6, 10);
}
#define init_5_7_6_11
void FASTCALL call_5_7_6_11(SH2_struct* sh)
{
  SH2movll4(sh, 7,6, 11);
}
#define init_5_7_6_12
void FASTCALL call_5_7_6_12(SH2_struct* sh)
{
  SH2movll4(sh, 7,6, 12);
}
#define init_5_7_6_13
void FASTCALL call_5_7_6_13(SH2_struct* sh)
{
  SH2movll4(sh, 7,6, 13);
}
#define init_5_7_6_14
void FASTCALL call_5_7_6_14(SH2_struct* sh)
{
  SH2movll4(sh, 7,6, 14);
}
#define init_5_7_6_15
void FASTCALL call_5_7_6_15(SH2_struct* sh)
{
  SH2movll4(sh, 7,6, 15);
}
#define init_5_7_7_0
void FASTCALL call_5_7_7_0(SH2_struct* sh)
{
  SH2movll4(sh, 7,7, 0);
}
#define init_5_7_7_1
void FASTCALL call_5_7_7_1(SH2_struct* sh)
{
  SH2movll4(sh, 7,7, 1);
}
#define init_5_7_7_2
void FASTCALL call_5_7_7_2(SH2_struct* sh)
{
  SH2movll4(sh, 7,7, 2);
}
#define init_5_7_7_3
void FASTCALL call_5_7_7_3(SH2_struct* sh)
{
  SH2movll4(sh, 7,7, 3);
}
#define init_5_7_7_4
void FASTCALL call_5_7_7_4(SH2_struct* sh)
{
  SH2movll4(sh, 7,7, 4);
}
#define init_5_7_7_5
void FASTCALL call_5_7_7_5(SH2_struct* sh)
{
  SH2movll4(sh, 7,7, 5);
}
#define init_5_7_7_6
void FASTCALL call_5_7_7_6(SH2_struct* sh)
{
  SH2movll4(sh, 7,7, 6);
}
#define init_5_7_7_7
void FASTCALL call_5_7_7_7(SH2_struct* sh)
{
  SH2movll4(sh, 7,7, 7);
}
#define init_5_7_7_8
void FASTCALL call_5_7_7_8(SH2_struct* sh)
{
  SH2movll4(sh, 7,7, 8);
}
#define init_5_7_7_9
void FASTCALL call_5_7_7_9(SH2_struct* sh)
{
  SH2movll4(sh, 7,7, 9);
}
#define init_5_7_7_10
void FASTCALL call_5_7_7_10(SH2_struct* sh)
{
  SH2movll4(sh, 7,7, 10);
}
#define init_5_7_7_11
void FASTCALL call_5_7_7_11(SH2_struct* sh)
{
  SH2movll4(sh, 7,7, 11);
}
#define init_5_7_7_12
void FASTCALL call_5_7_7_12(SH2_struct* sh)
{
  SH2movll4(sh, 7,7, 12);
}
#define init_5_7_7_13
void FASTCALL call_5_7_7_13(SH2_struct* sh)
{
  SH2movll4(sh, 7,7, 13);
}
#define init_5_7_7_14
void FASTCALL call_5_7_7_14(SH2_struct* sh)
{
  SH2movll4(sh, 7,7, 14);
}
#define init_5_7_7_15
void FASTCALL call_5_7_7_15(SH2_struct* sh)
{
  SH2movll4(sh, 7,7, 15);
}
#define init_5_7_8_0
void FASTCALL call_5_7_8_0(SH2_struct* sh)
{
  SH2movll4(sh, 7,8, 0);
}
#define init_5_7_8_1
void FASTCALL call_5_7_8_1(SH2_struct* sh)
{
  SH2movll4(sh, 7,8, 1);
}
#define init_5_7_8_2
void FASTCALL call_5_7_8_2(SH2_struct* sh)
{
  SH2movll4(sh, 7,8, 2);
}
#define init_5_7_8_3
void FASTCALL call_5_7_8_3(SH2_struct* sh)
{
  SH2movll4(sh, 7,8, 3);
}
#define init_5_7_8_4
void FASTCALL call_5_7_8_4(SH2_struct* sh)
{
  SH2movll4(sh, 7,8, 4);
}
#define init_5_7_8_5
void FASTCALL call_5_7_8_5(SH2_struct* sh)
{
  SH2movll4(sh, 7,8, 5);
}
#define init_5_7_8_6
void FASTCALL call_5_7_8_6(SH2_struct* sh)
{
  SH2movll4(sh, 7,8, 6);
}
#define init_5_7_8_7
void FASTCALL call_5_7_8_7(SH2_struct* sh)
{
  SH2movll4(sh, 7,8, 7);
}
#define init_5_7_8_8
void FASTCALL call_5_7_8_8(SH2_struct* sh)
{
  SH2movll4(sh, 7,8, 8);
}
#define init_5_7_8_9
void FASTCALL call_5_7_8_9(SH2_struct* sh)
{
  SH2movll4(sh, 7,8, 9);
}
#define init_5_7_8_10
void FASTCALL call_5_7_8_10(SH2_struct* sh)
{
  SH2movll4(sh, 7,8, 10);
}
#define init_5_7_8_11
void FASTCALL call_5_7_8_11(SH2_struct* sh)
{
  SH2movll4(sh, 7,8, 11);
}
#define init_5_7_8_12
void FASTCALL call_5_7_8_12(SH2_struct* sh)
{
  SH2movll4(sh, 7,8, 12);
}
#define init_5_7_8_13
void FASTCALL call_5_7_8_13(SH2_struct* sh)
{
  SH2movll4(sh, 7,8, 13);
}
#define init_5_7_8_14
void FASTCALL call_5_7_8_14(SH2_struct* sh)
{
  SH2movll4(sh, 7,8, 14);
}
#define init_5_7_8_15
void FASTCALL call_5_7_8_15(SH2_struct* sh)
{
  SH2movll4(sh, 7,8, 15);
}
#define init_5_7_9_0
void FASTCALL call_5_7_9_0(SH2_struct* sh)
{
  SH2movll4(sh, 7,9, 0);
}
#define init_5_7_9_1
void FASTCALL call_5_7_9_1(SH2_struct* sh)
{
  SH2movll4(sh, 7,9, 1);
}
#define init_5_7_9_2
void FASTCALL call_5_7_9_2(SH2_struct* sh)
{
  SH2movll4(sh, 7,9, 2);
}
#define init_5_7_9_3
void FASTCALL call_5_7_9_3(SH2_struct* sh)
{
  SH2movll4(sh, 7,9, 3);
}
#define init_5_7_9_4
void FASTCALL call_5_7_9_4(SH2_struct* sh)
{
  SH2movll4(sh, 7,9, 4);
}
#define init_5_7_9_5
void FASTCALL call_5_7_9_5(SH2_struct* sh)
{
  SH2movll4(sh, 7,9, 5);
}
#define init_5_7_9_6
void FASTCALL call_5_7_9_6(SH2_struct* sh)
{
  SH2movll4(sh, 7,9, 6);
}
#define init_5_7_9_7
void FASTCALL call_5_7_9_7(SH2_struct* sh)
{
  SH2movll4(sh, 7,9, 7);
}
#define init_5_7_9_8
void FASTCALL call_5_7_9_8(SH2_struct* sh)
{
  SH2movll4(sh, 7,9, 8);
}
#define init_5_7_9_9
void FASTCALL call_5_7_9_9(SH2_struct* sh)
{
  SH2movll4(sh, 7,9, 9);
}
#define init_5_7_9_10
void FASTCALL call_5_7_9_10(SH2_struct* sh)
{
  SH2movll4(sh, 7,9, 10);
}
#define init_5_7_9_11
void FASTCALL call_5_7_9_11(SH2_struct* sh)
{
  SH2movll4(sh, 7,9, 11);
}
#define init_5_7_9_12
void FASTCALL call_5_7_9_12(SH2_struct* sh)
{
  SH2movll4(sh, 7,9, 12);
}
#define init_5_7_9_13
void FASTCALL call_5_7_9_13(SH2_struct* sh)
{
  SH2movll4(sh, 7,9, 13);
}
#define init_5_7_9_14
void FASTCALL call_5_7_9_14(SH2_struct* sh)
{
  SH2movll4(sh, 7,9, 14);
}
#define init_5_7_9_15
void FASTCALL call_5_7_9_15(SH2_struct* sh)
{
  SH2movll4(sh, 7,9, 15);
}
#define init_5_7_10_0
void FASTCALL call_5_7_10_0(SH2_struct* sh)
{
  SH2movll4(sh, 7,10, 0);
}
#define init_5_7_10_1
void FASTCALL call_5_7_10_1(SH2_struct* sh)
{
  SH2movll4(sh, 7,10, 1);
}
#define init_5_7_10_2
void FASTCALL call_5_7_10_2(SH2_struct* sh)
{
  SH2movll4(sh, 7,10, 2);
}
#define init_5_7_10_3
void FASTCALL call_5_7_10_3(SH2_struct* sh)
{
  SH2movll4(sh, 7,10, 3);
}
#define init_5_7_10_4
void FASTCALL call_5_7_10_4(SH2_struct* sh)
{
  SH2movll4(sh, 7,10, 4);
}
#define init_5_7_10_5
void FASTCALL call_5_7_10_5(SH2_struct* sh)
{
  SH2movll4(sh, 7,10, 5);
}
#define init_5_7_10_6
void FASTCALL call_5_7_10_6(SH2_struct* sh)
{
  SH2movll4(sh, 7,10, 6);
}
#define init_5_7_10_7
void FASTCALL call_5_7_10_7(SH2_struct* sh)
{
  SH2movll4(sh, 7,10, 7);
}
#define init_5_7_10_8
void FASTCALL call_5_7_10_8(SH2_struct* sh)
{
  SH2movll4(sh, 7,10, 8);
}
#define init_5_7_10_9
void FASTCALL call_5_7_10_9(SH2_struct* sh)
{
  SH2movll4(sh, 7,10, 9);
}
#define init_5_7_10_10
void FASTCALL call_5_7_10_10(SH2_struct* sh)
{
  SH2movll4(sh, 7,10, 10);
}
#define init_5_7_10_11
void FASTCALL call_5_7_10_11(SH2_struct* sh)
{
  SH2movll4(sh, 7,10, 11);
}
#define init_5_7_10_12
void FASTCALL call_5_7_10_12(SH2_struct* sh)
{
  SH2movll4(sh, 7,10, 12);
}
#define init_5_7_10_13
void FASTCALL call_5_7_10_13(SH2_struct* sh)
{
  SH2movll4(sh, 7,10, 13);
}
#define init_5_7_10_14
void FASTCALL call_5_7_10_14(SH2_struct* sh)
{
  SH2movll4(sh, 7,10, 14);
}
#define init_5_7_10_15
void FASTCALL call_5_7_10_15(SH2_struct* sh)
{
  SH2movll4(sh, 7,10, 15);
}
#define init_5_7_11_0
void FASTCALL call_5_7_11_0(SH2_struct* sh)
{
  SH2movll4(sh, 7,11, 0);
}
#define init_5_7_11_1
void FASTCALL call_5_7_11_1(SH2_struct* sh)
{
  SH2movll4(sh, 7,11, 1);
}
#define init_5_7_11_2
void FASTCALL call_5_7_11_2(SH2_struct* sh)
{
  SH2movll4(sh, 7,11, 2);
}
#define init_5_7_11_3
void FASTCALL call_5_7_11_3(SH2_struct* sh)
{
  SH2movll4(sh, 7,11, 3);
}
#define init_5_7_11_4
void FASTCALL call_5_7_11_4(SH2_struct* sh)
{
  SH2movll4(sh, 7,11, 4);
}
#define init_5_7_11_5
void FASTCALL call_5_7_11_5(SH2_struct* sh)
{
  SH2movll4(sh, 7,11, 5);
}
#define init_5_7_11_6
void FASTCALL call_5_7_11_6(SH2_struct* sh)
{
  SH2movll4(sh, 7,11, 6);
}
#define init_5_7_11_7
void FASTCALL call_5_7_11_7(SH2_struct* sh)
{
  SH2movll4(sh, 7,11, 7);
}
#define init_5_7_11_8
void FASTCALL call_5_7_11_8(SH2_struct* sh)
{
  SH2movll4(sh, 7,11, 8);
}
#define init_5_7_11_9
void FASTCALL call_5_7_11_9(SH2_struct* sh)
{
  SH2movll4(sh, 7,11, 9);
}
#define init_5_7_11_10
void FASTCALL call_5_7_11_10(SH2_struct* sh)
{
  SH2movll4(sh, 7,11, 10);
}
#define init_5_7_11_11
void FASTCALL call_5_7_11_11(SH2_struct* sh)
{
  SH2movll4(sh, 7,11, 11);
}
#define init_5_7_11_12
void FASTCALL call_5_7_11_12(SH2_struct* sh)
{
  SH2movll4(sh, 7,11, 12);
}
#define init_5_7_11_13
void FASTCALL call_5_7_11_13(SH2_struct* sh)
{
  SH2movll4(sh, 7,11, 13);
}
#define init_5_7_11_14
void FASTCALL call_5_7_11_14(SH2_struct* sh)
{
  SH2movll4(sh, 7,11, 14);
}
#define init_5_7_11_15
void FASTCALL call_5_7_11_15(SH2_struct* sh)
{
  SH2movll4(sh, 7,11, 15);
}
#define init_5_7_12_0
void FASTCALL call_5_7_12_0(SH2_struct* sh)
{
  SH2movll4(sh, 7,12, 0);
}
#define init_5_7_12_1
void FASTCALL call_5_7_12_1(SH2_struct* sh)
{
  SH2movll4(sh, 7,12, 1);
}
#define init_5_7_12_2
void FASTCALL call_5_7_12_2(SH2_struct* sh)
{
  SH2movll4(sh, 7,12, 2);
}
#define init_5_7_12_3
void FASTCALL call_5_7_12_3(SH2_struct* sh)
{
  SH2movll4(sh, 7,12, 3);
}
#define init_5_7_12_4
void FASTCALL call_5_7_12_4(SH2_struct* sh)
{
  SH2movll4(sh, 7,12, 4);
}
#define init_5_7_12_5
void FASTCALL call_5_7_12_5(SH2_struct* sh)
{
  SH2movll4(sh, 7,12, 5);
}
#define init_5_7_12_6
void FASTCALL call_5_7_12_6(SH2_struct* sh)
{
  SH2movll4(sh, 7,12, 6);
}
#define init_5_7_12_7
void FASTCALL call_5_7_12_7(SH2_struct* sh)
{
  SH2movll4(sh, 7,12, 7);
}
#define init_5_7_12_8
void FASTCALL call_5_7_12_8(SH2_struct* sh)
{
  SH2movll4(sh, 7,12, 8);
}
#define init_5_7_12_9
void FASTCALL call_5_7_12_9(SH2_struct* sh)
{
  SH2movll4(sh, 7,12, 9);
}
#define init_5_7_12_10
void FASTCALL call_5_7_12_10(SH2_struct* sh)
{
  SH2movll4(sh, 7,12, 10);
}
#define init_5_7_12_11
void FASTCALL call_5_7_12_11(SH2_struct* sh)
{
  SH2movll4(sh, 7,12, 11);
}
#define init_5_7_12_12
void FASTCALL call_5_7_12_12(SH2_struct* sh)
{
  SH2movll4(sh, 7,12, 12);
}
#define init_5_7_12_13
void FASTCALL call_5_7_12_13(SH2_struct* sh)
{
  SH2movll4(sh, 7,12, 13);
}
#define init_5_7_12_14
void FASTCALL call_5_7_12_14(SH2_struct* sh)
{
  SH2movll4(sh, 7,12, 14);
}
#define init_5_7_12_15
void FASTCALL call_5_7_12_15(SH2_struct* sh)
{
  SH2movll4(sh, 7,12, 15);
}
#define init_5_7_13_0
void FASTCALL call_5_7_13_0(SH2_struct* sh)
{
  SH2movll4(sh, 7,13, 0);
}
#define init_5_7_13_1
void FASTCALL call_5_7_13_1(SH2_struct* sh)
{
  SH2movll4(sh, 7,13, 1);
}
#define init_5_7_13_2
void FASTCALL call_5_7_13_2(SH2_struct* sh)
{
  SH2movll4(sh, 7,13, 2);
}
#define init_5_7_13_3
void FASTCALL call_5_7_13_3(SH2_struct* sh)
{
  SH2movll4(sh, 7,13, 3);
}
#define init_5_7_13_4
void FASTCALL call_5_7_13_4(SH2_struct* sh)
{
  SH2movll4(sh, 7,13, 4);
}
#define init_5_7_13_5
void FASTCALL call_5_7_13_5(SH2_struct* sh)
{
  SH2movll4(sh, 7,13, 5);
}
#define init_5_7_13_6
void FASTCALL call_5_7_13_6(SH2_struct* sh)
{
  SH2movll4(sh, 7,13, 6);
}
#define init_5_7_13_7
void FASTCALL call_5_7_13_7(SH2_struct* sh)
{
  SH2movll4(sh, 7,13, 7);
}
#define init_5_7_13_8
void FASTCALL call_5_7_13_8(SH2_struct* sh)
{
  SH2movll4(sh, 7,13, 8);
}
#define init_5_7_13_9
void FASTCALL call_5_7_13_9(SH2_struct* sh)
{
  SH2movll4(sh, 7,13, 9);
}
#define init_5_7_13_10
void FASTCALL call_5_7_13_10(SH2_struct* sh)
{
  SH2movll4(sh, 7,13, 10);
}
#define init_5_7_13_11
void FASTCALL call_5_7_13_11(SH2_struct* sh)
{
  SH2movll4(sh, 7,13, 11);
}
#define init_5_7_13_12
void FASTCALL call_5_7_13_12(SH2_struct* sh)
{
  SH2movll4(sh, 7,13, 12);
}
#define init_5_7_13_13
void FASTCALL call_5_7_13_13(SH2_struct* sh)
{
  SH2movll4(sh, 7,13, 13);
}
#define init_5_7_13_14
void FASTCALL call_5_7_13_14(SH2_struct* sh)
{
  SH2movll4(sh, 7,13, 14);
}
#define init_5_7_13_15
void FASTCALL call_5_7_13_15(SH2_struct* sh)
{
  SH2movll4(sh, 7,13, 15);
}
#define init_5_7_14_0
void FASTCALL call_5_7_14_0(SH2_struct* sh)
{
  SH2movll4(sh, 7,14, 0);
}
#define init_5_7_14_1
void FASTCALL call_5_7_14_1(SH2_struct* sh)
{
  SH2movll4(sh, 7,14, 1);
}
#define init_5_7_14_2
void FASTCALL call_5_7_14_2(SH2_struct* sh)
{
  SH2movll4(sh, 7,14, 2);
}
#define init_5_7_14_3
void FASTCALL call_5_7_14_3(SH2_struct* sh)
{
  SH2movll4(sh, 7,14, 3);
}
#define init_5_7_14_4
void FASTCALL call_5_7_14_4(SH2_struct* sh)
{
  SH2movll4(sh, 7,14, 4);
}
#define init_5_7_14_5
void FASTCALL call_5_7_14_5(SH2_struct* sh)
{
  SH2movll4(sh, 7,14, 5);
}
#define init_5_7_14_6
void FASTCALL call_5_7_14_6(SH2_struct* sh)
{
  SH2movll4(sh, 7,14, 6);
}
#define init_5_7_14_7
void FASTCALL call_5_7_14_7(SH2_struct* sh)
{
  SH2movll4(sh, 7,14, 7);
}
#define init_5_7_14_8
void FASTCALL call_5_7_14_8(SH2_struct* sh)
{
  SH2movll4(sh, 7,14, 8);
}
#define init_5_7_14_9
void FASTCALL call_5_7_14_9(SH2_struct* sh)
{
  SH2movll4(sh, 7,14, 9);
}
#define init_5_7_14_10
void FASTCALL call_5_7_14_10(SH2_struct* sh)
{
  SH2movll4(sh, 7,14, 10);
}
#define init_5_7_14_11
void FASTCALL call_5_7_14_11(SH2_struct* sh)
{
  SH2movll4(sh, 7,14, 11);
}
#define init_5_7_14_12
void FASTCALL call_5_7_14_12(SH2_struct* sh)
{
  SH2movll4(sh, 7,14, 12);
}
#define init_5_7_14_13
void FASTCALL call_5_7_14_13(SH2_struct* sh)
{
  SH2movll4(sh, 7,14, 13);
}
#define init_5_7_14_14
void FASTCALL call_5_7_14_14(SH2_struct* sh)
{
  SH2movll4(sh, 7,14, 14);
}
#define init_5_7_14_15
void FASTCALL call_5_7_14_15(SH2_struct* sh)
{
  SH2movll4(sh, 7,14, 15);
}
#define init_5_7_15_0
void FASTCALL call_5_7_15_0(SH2_struct* sh)
{
  SH2movll4(sh, 7,15, 0);
}
#define init_5_7_15_1
void FASTCALL call_5_7_15_1(SH2_struct* sh)
{
  SH2movll4(sh, 7,15, 1);
}
#define init_5_7_15_2
void FASTCALL call_5_7_15_2(SH2_struct* sh)
{
  SH2movll4(sh, 7,15, 2);
}
#define init_5_7_15_3
void FASTCALL call_5_7_15_3(SH2_struct* sh)
{
  SH2movll4(sh, 7,15, 3);
}
#define init_5_7_15_4
void FASTCALL call_5_7_15_4(SH2_struct* sh)
{
  SH2movll4(sh, 7,15, 4);
}
#define init_5_7_15_5
void FASTCALL call_5_7_15_5(SH2_struct* sh)
{
  SH2movll4(sh, 7,15, 5);
}
#define init_5_7_15_6
void FASTCALL call_5_7_15_6(SH2_struct* sh)
{
  SH2movll4(sh, 7,15, 6);
}
#define init_5_7_15_7
void FASTCALL call_5_7_15_7(SH2_struct* sh)
{
  SH2movll4(sh, 7,15, 7);
}
#define init_5_7_15_8
void FASTCALL call_5_7_15_8(SH2_struct* sh)
{
  SH2movll4(sh, 7,15, 8);
}
#define init_5_7_15_9
void FASTCALL call_5_7_15_9(SH2_struct* sh)
{
  SH2movll4(sh, 7,15, 9);
}
#define init_5_7_15_10
void FASTCALL call_5_7_15_10(SH2_struct* sh)
{
  SH2movll4(sh, 7,15, 10);
}
#define init_5_7_15_11
void FASTCALL call_5_7_15_11(SH2_struct* sh)
{
  SH2movll4(sh, 7,15, 11);
}
#define init_5_7_15_12
void FASTCALL call_5_7_15_12(SH2_struct* sh)
{
  SH2movll4(sh, 7,15, 12);
}
#define init_5_7_15_13
void FASTCALL call_5_7_15_13(SH2_struct* sh)
{
  SH2movll4(sh, 7,15, 13);
}
#define init_5_7_15_14
void FASTCALL call_5_7_15_14(SH2_struct* sh)
{
  SH2movll4(sh, 7,15, 14);
}
#define init_5_7_15_15
void FASTCALL call_5_7_15_15(SH2_struct* sh)
{
  SH2movll4(sh, 7,15, 15);
}
#define init_5_8_0_0
void FASTCALL call_5_8_0_0(SH2_struct* sh)
{
  SH2movll4(sh, 8,0, 0);
}
#define init_5_8_0_1
void FASTCALL call_5_8_0_1(SH2_struct* sh)
{
  SH2movll4(sh, 8,0, 1);
}
#define init_5_8_0_2
void FASTCALL call_5_8_0_2(SH2_struct* sh)
{
  SH2movll4(sh, 8,0, 2);
}
#define init_5_8_0_3
void FASTCALL call_5_8_0_3(SH2_struct* sh)
{
  SH2movll4(sh, 8,0, 3);
}
#define init_5_8_0_4
void FASTCALL call_5_8_0_4(SH2_struct* sh)
{
  SH2movll4(sh, 8,0, 4);
}
#define init_5_8_0_5
void FASTCALL call_5_8_0_5(SH2_struct* sh)
{
  SH2movll4(sh, 8,0, 5);
}
#define init_5_8_0_6
void FASTCALL call_5_8_0_6(SH2_struct* sh)
{
  SH2movll4(sh, 8,0, 6);
}
#define init_5_8_0_7
void FASTCALL call_5_8_0_7(SH2_struct* sh)
{
  SH2movll4(sh, 8,0, 7);
}
#define init_5_8_0_8
void FASTCALL call_5_8_0_8(SH2_struct* sh)
{
  SH2movll4(sh, 8,0, 8);
}
#define init_5_8_0_9
void FASTCALL call_5_8_0_9(SH2_struct* sh)
{
  SH2movll4(sh, 8,0, 9);
}
#define init_5_8_0_10
void FASTCALL call_5_8_0_10(SH2_struct* sh)
{
  SH2movll4(sh, 8,0, 10);
}
#define init_5_8_0_11
void FASTCALL call_5_8_0_11(SH2_struct* sh)
{
  SH2movll4(sh, 8,0, 11);
}
#define init_5_8_0_12
void FASTCALL call_5_8_0_12(SH2_struct* sh)
{
  SH2movll4(sh, 8,0, 12);
}
#define init_5_8_0_13
void FASTCALL call_5_8_0_13(SH2_struct* sh)
{
  SH2movll4(sh, 8,0, 13);
}
#define init_5_8_0_14
void FASTCALL call_5_8_0_14(SH2_struct* sh)
{
  SH2movll4(sh, 8,0, 14);
}
#define init_5_8_0_15
void FASTCALL call_5_8_0_15(SH2_struct* sh)
{
  SH2movll4(sh, 8,0, 15);
}
#define init_5_8_1_0
void FASTCALL call_5_8_1_0(SH2_struct* sh)
{
  SH2movll4(sh, 8,1, 0);
}
#define init_5_8_1_1
void FASTCALL call_5_8_1_1(SH2_struct* sh)
{
  SH2movll4(sh, 8,1, 1);
}
#define init_5_8_1_2
void FASTCALL call_5_8_1_2(SH2_struct* sh)
{
  SH2movll4(sh, 8,1, 2);
}
#define init_5_8_1_3
void FASTCALL call_5_8_1_3(SH2_struct* sh)
{
  SH2movll4(sh, 8,1, 3);
}
#define init_5_8_1_4
void FASTCALL call_5_8_1_4(SH2_struct* sh)
{
  SH2movll4(sh, 8,1, 4);
}
#define init_5_8_1_5
void FASTCALL call_5_8_1_5(SH2_struct* sh)
{
  SH2movll4(sh, 8,1, 5);
}
#define init_5_8_1_6
void FASTCALL call_5_8_1_6(SH2_struct* sh)
{
  SH2movll4(sh, 8,1, 6);
}
#define init_5_8_1_7
void FASTCALL call_5_8_1_7(SH2_struct* sh)
{
  SH2movll4(sh, 8,1, 7);
}
#define init_5_8_1_8
void FASTCALL call_5_8_1_8(SH2_struct* sh)
{
  SH2movll4(sh, 8,1, 8);
}
#define init_5_8_1_9
void FASTCALL call_5_8_1_9(SH2_struct* sh)
{
  SH2movll4(sh, 8,1, 9);
}
#define init_5_8_1_10
void FASTCALL call_5_8_1_10(SH2_struct* sh)
{
  SH2movll4(sh, 8,1, 10);
}
#define init_5_8_1_11
void FASTCALL call_5_8_1_11(SH2_struct* sh)
{
  SH2movll4(sh, 8,1, 11);
}
#define init_5_8_1_12
void FASTCALL call_5_8_1_12(SH2_struct* sh)
{
  SH2movll4(sh, 8,1, 12);
}
#define init_5_8_1_13
void FASTCALL call_5_8_1_13(SH2_struct* sh)
{
  SH2movll4(sh, 8,1, 13);
}
#define init_5_8_1_14
void FASTCALL call_5_8_1_14(SH2_struct* sh)
{
  SH2movll4(sh, 8,1, 14);
}
#define init_5_8_1_15
void FASTCALL call_5_8_1_15(SH2_struct* sh)
{
  SH2movll4(sh, 8,1, 15);
}
#define init_5_8_2_0
void FASTCALL call_5_8_2_0(SH2_struct* sh)
{
  SH2movll4(sh, 8,2, 0);
}
#define init_5_8_2_1
void FASTCALL call_5_8_2_1(SH2_struct* sh)
{
  SH2movll4(sh, 8,2, 1);
}
#define init_5_8_2_2
void FASTCALL call_5_8_2_2(SH2_struct* sh)
{
  SH2movll4(sh, 8,2, 2);
}
#define init_5_8_2_3
void FASTCALL call_5_8_2_3(SH2_struct* sh)
{
  SH2movll4(sh, 8,2, 3);
}
#define init_5_8_2_4
void FASTCALL call_5_8_2_4(SH2_struct* sh)
{
  SH2movll4(sh, 8,2, 4);
}
#define init_5_8_2_5
void FASTCALL call_5_8_2_5(SH2_struct* sh)
{
  SH2movll4(sh, 8,2, 5);
}
#define init_5_8_2_6
void FASTCALL call_5_8_2_6(SH2_struct* sh)
{
  SH2movll4(sh, 8,2, 6);
}
#define init_5_8_2_7
void FASTCALL call_5_8_2_7(SH2_struct* sh)
{
  SH2movll4(sh, 8,2, 7);
}
#define init_5_8_2_8
void FASTCALL call_5_8_2_8(SH2_struct* sh)
{
  SH2movll4(sh, 8,2, 8);
}
#define init_5_8_2_9
void FASTCALL call_5_8_2_9(SH2_struct* sh)
{
  SH2movll4(sh, 8,2, 9);
}
#define init_5_8_2_10
void FASTCALL call_5_8_2_10(SH2_struct* sh)
{
  SH2movll4(sh, 8,2, 10);
}
#define init_5_8_2_11
void FASTCALL call_5_8_2_11(SH2_struct* sh)
{
  SH2movll4(sh, 8,2, 11);
}
#define init_5_8_2_12
void FASTCALL call_5_8_2_12(SH2_struct* sh)
{
  SH2movll4(sh, 8,2, 12);
}
#define init_5_8_2_13
void FASTCALL call_5_8_2_13(SH2_struct* sh)
{
  SH2movll4(sh, 8,2, 13);
}
#define init_5_8_2_14
void FASTCALL call_5_8_2_14(SH2_struct* sh)
{
  SH2movll4(sh, 8,2, 14);
}
#define init_5_8_2_15
void FASTCALL call_5_8_2_15(SH2_struct* sh)
{
  SH2movll4(sh, 8,2, 15);
}
#define init_5_8_3_0
void FASTCALL call_5_8_3_0(SH2_struct* sh)
{
  SH2movll4(sh, 8,3, 0);
}
#define init_5_8_3_1
void FASTCALL call_5_8_3_1(SH2_struct* sh)
{
  SH2movll4(sh, 8,3, 1);
}
#define init_5_8_3_2
void FASTCALL call_5_8_3_2(SH2_struct* sh)
{
  SH2movll4(sh, 8,3, 2);
}
#define init_5_8_3_3
void FASTCALL call_5_8_3_3(SH2_struct* sh)
{
  SH2movll4(sh, 8,3, 3);
}
#define init_5_8_3_4
void FASTCALL call_5_8_3_4(SH2_struct* sh)
{
  SH2movll4(sh, 8,3, 4);
}
#define init_5_8_3_5
void FASTCALL call_5_8_3_5(SH2_struct* sh)
{
  SH2movll4(sh, 8,3, 5);
}
#define init_5_8_3_6
void FASTCALL call_5_8_3_6(SH2_struct* sh)
{
  SH2movll4(sh, 8,3, 6);
}
#define init_5_8_3_7
void FASTCALL call_5_8_3_7(SH2_struct* sh)
{
  SH2movll4(sh, 8,3, 7);
}
#define init_5_8_3_8
void FASTCALL call_5_8_3_8(SH2_struct* sh)
{
  SH2movll4(sh, 8,3, 8);
}
#define init_5_8_3_9
void FASTCALL call_5_8_3_9(SH2_struct* sh)
{
  SH2movll4(sh, 8,3, 9);
}
#define init_5_8_3_10
void FASTCALL call_5_8_3_10(SH2_struct* sh)
{
  SH2movll4(sh, 8,3, 10);
}
#define init_5_8_3_11
void FASTCALL call_5_8_3_11(SH2_struct* sh)
{
  SH2movll4(sh, 8,3, 11);
}
#define init_5_8_3_12
void FASTCALL call_5_8_3_12(SH2_struct* sh)
{
  SH2movll4(sh, 8,3, 12);
}
#define init_5_8_3_13
void FASTCALL call_5_8_3_13(SH2_struct* sh)
{
  SH2movll4(sh, 8,3, 13);
}
#define init_5_8_3_14
void FASTCALL call_5_8_3_14(SH2_struct* sh)
{
  SH2movll4(sh, 8,3, 14);
}
#define init_5_8_3_15
void FASTCALL call_5_8_3_15(SH2_struct* sh)
{
  SH2movll4(sh, 8,3, 15);
}
#define init_5_8_4_0
void FASTCALL call_5_8_4_0(SH2_struct* sh)
{
  SH2movll4(sh, 8,4, 0);
}
#define init_5_8_4_1
void FASTCALL call_5_8_4_1(SH2_struct* sh)
{
  SH2movll4(sh, 8,4, 1);
}
#define init_5_8_4_2
void FASTCALL call_5_8_4_2(SH2_struct* sh)
{
  SH2movll4(sh, 8,4, 2);
}
#define init_5_8_4_3
void FASTCALL call_5_8_4_3(SH2_struct* sh)
{
  SH2movll4(sh, 8,4, 3);
}
#define init_5_8_4_4
void FASTCALL call_5_8_4_4(SH2_struct* sh)
{
  SH2movll4(sh, 8,4, 4);
}
#define init_5_8_4_5
void FASTCALL call_5_8_4_5(SH2_struct* sh)
{
  SH2movll4(sh, 8,4, 5);
}
#define init_5_8_4_6
void FASTCALL call_5_8_4_6(SH2_struct* sh)
{
  SH2movll4(sh, 8,4, 6);
}
#define init_5_8_4_7
void FASTCALL call_5_8_4_7(SH2_struct* sh)
{
  SH2movll4(sh, 8,4, 7);
}
#define init_5_8_4_8
void FASTCALL call_5_8_4_8(SH2_struct* sh)
{
  SH2movll4(sh, 8,4, 8);
}
#define init_5_8_4_9
void FASTCALL call_5_8_4_9(SH2_struct* sh)
{
  SH2movll4(sh, 8,4, 9);
}
#define init_5_8_4_10
void FASTCALL call_5_8_4_10(SH2_struct* sh)
{
  SH2movll4(sh, 8,4, 10);
}
#define init_5_8_4_11
void FASTCALL call_5_8_4_11(SH2_struct* sh)
{
  SH2movll4(sh, 8,4, 11);
}
#define init_5_8_4_12
void FASTCALL call_5_8_4_12(SH2_struct* sh)
{
  SH2movll4(sh, 8,4, 12);
}
#define init_5_8_4_13
void FASTCALL call_5_8_4_13(SH2_struct* sh)
{
  SH2movll4(sh, 8,4, 13);
}
#define init_5_8_4_14
void FASTCALL call_5_8_4_14(SH2_struct* sh)
{
  SH2movll4(sh, 8,4, 14);
}
#define init_5_8_4_15
void FASTCALL call_5_8_4_15(SH2_struct* sh)
{
  SH2movll4(sh, 8,4, 15);
}
#define init_5_8_5_0
void FASTCALL call_5_8_5_0(SH2_struct* sh)
{
  SH2movll4(sh, 8,5, 0);
}
#define init_5_8_5_1
void FASTCALL call_5_8_5_1(SH2_struct* sh)
{
  SH2movll4(sh, 8,5, 1);
}
#define init_5_8_5_2
void FASTCALL call_5_8_5_2(SH2_struct* sh)
{
  SH2movll4(sh, 8,5, 2);
}
#define init_5_8_5_3
void FASTCALL call_5_8_5_3(SH2_struct* sh)
{
  SH2movll4(sh, 8,5, 3);
}
#define init_5_8_5_4
void FASTCALL call_5_8_5_4(SH2_struct* sh)
{
  SH2movll4(sh, 8,5, 4);
}
#define init_5_8_5_5
void FASTCALL call_5_8_5_5(SH2_struct* sh)
{
  SH2movll4(sh, 8,5, 5);
}
#define init_5_8_5_6
void FASTCALL call_5_8_5_6(SH2_struct* sh)
{
  SH2movll4(sh, 8,5, 6);
}
#define init_5_8_5_7
void FASTCALL call_5_8_5_7(SH2_struct* sh)
{
  SH2movll4(sh, 8,5, 7);
}
#define init_5_8_5_8
void FASTCALL call_5_8_5_8(SH2_struct* sh)
{
  SH2movll4(sh, 8,5, 8);
}
#define init_5_8_5_9
void FASTCALL call_5_8_5_9(SH2_struct* sh)
{
  SH2movll4(sh, 8,5, 9);
}
#define init_5_8_5_10
void FASTCALL call_5_8_5_10(SH2_struct* sh)
{
  SH2movll4(sh, 8,5, 10);
}
#define init_5_8_5_11
void FASTCALL call_5_8_5_11(SH2_struct* sh)
{
  SH2movll4(sh, 8,5, 11);
}
#define init_5_8_5_12
void FASTCALL call_5_8_5_12(SH2_struct* sh)
{
  SH2movll4(sh, 8,5, 12);
}
#define init_5_8_5_13
void FASTCALL call_5_8_5_13(SH2_struct* sh)
{
  SH2movll4(sh, 8,5, 13);
}
#define init_5_8_5_14
void FASTCALL call_5_8_5_14(SH2_struct* sh)
{
  SH2movll4(sh, 8,5, 14);
}
#define init_5_8_5_15
void FASTCALL call_5_8_5_15(SH2_struct* sh)
{
  SH2movll4(sh, 8,5, 15);
}
#define init_5_8_6_0
void FASTCALL call_5_8_6_0(SH2_struct* sh)
{
  SH2movll4(sh, 8,6, 0);
}
#define init_5_8_6_1
void FASTCALL call_5_8_6_1(SH2_struct* sh)
{
  SH2movll4(sh, 8,6, 1);
}
#define init_5_8_6_2
void FASTCALL call_5_8_6_2(SH2_struct* sh)
{
  SH2movll4(sh, 8,6, 2);
}
#define init_5_8_6_3
void FASTCALL call_5_8_6_3(SH2_struct* sh)
{
  SH2movll4(sh, 8,6, 3);
}
#define init_5_8_6_4
void FASTCALL call_5_8_6_4(SH2_struct* sh)
{
  SH2movll4(sh, 8,6, 4);
}
#define init_5_8_6_5
void FASTCALL call_5_8_6_5(SH2_struct* sh)
{
  SH2movll4(sh, 8,6, 5);
}
#define init_5_8_6_6
void FASTCALL call_5_8_6_6(SH2_struct* sh)
{
  SH2movll4(sh, 8,6, 6);
}
#define init_5_8_6_7
void FASTCALL call_5_8_6_7(SH2_struct* sh)
{
  SH2movll4(sh, 8,6, 7);
}
#define init_5_8_6_8
void FASTCALL call_5_8_6_8(SH2_struct* sh)
{
  SH2movll4(sh, 8,6, 8);
}
#define init_5_8_6_9
void FASTCALL call_5_8_6_9(SH2_struct* sh)
{
  SH2movll4(sh, 8,6, 9);
}
#define init_5_8_6_10
void FASTCALL call_5_8_6_10(SH2_struct* sh)
{
  SH2movll4(sh, 8,6, 10);
}
#define init_5_8_6_11
void FASTCALL call_5_8_6_11(SH2_struct* sh)
{
  SH2movll4(sh, 8,6, 11);
}
#define init_5_8_6_12
void FASTCALL call_5_8_6_12(SH2_struct* sh)
{
  SH2movll4(sh, 8,6, 12);
}
#define init_5_8_6_13
void FASTCALL call_5_8_6_13(SH2_struct* sh)
{
  SH2movll4(sh, 8,6, 13);
}
#define init_5_8_6_14
void FASTCALL call_5_8_6_14(SH2_struct* sh)
{
  SH2movll4(sh, 8,6, 14);
}
#define init_5_8_6_15
void FASTCALL call_5_8_6_15(SH2_struct* sh)
{
  SH2movll4(sh, 8,6, 15);
}
#define init_5_8_7_0
void FASTCALL call_5_8_7_0(SH2_struct* sh)
{
  SH2movll4(sh, 8,7, 0);
}
#define init_5_8_7_1
void FASTCALL call_5_8_7_1(SH2_struct* sh)
{
  SH2movll4(sh, 8,7, 1);
}
#define init_5_8_7_2
void FASTCALL call_5_8_7_2(SH2_struct* sh)
{
  SH2movll4(sh, 8,7, 2);
}
#define init_5_8_7_3
void FASTCALL call_5_8_7_3(SH2_struct* sh)
{
  SH2movll4(sh, 8,7, 3);
}
#define init_5_8_7_4
void FASTCALL call_5_8_7_4(SH2_struct* sh)
{
  SH2movll4(sh, 8,7, 4);
}
#define init_5_8_7_5
void FASTCALL call_5_8_7_5(SH2_struct* sh)
{
  SH2movll4(sh, 8,7, 5);
}
#define init_5_8_7_6
void FASTCALL call_5_8_7_6(SH2_struct* sh)
{
  SH2movll4(sh, 8,7, 6);
}
#define init_5_8_7_7
void FASTCALL call_5_8_7_7(SH2_struct* sh)
{
  SH2movll4(sh, 8,7, 7);
}
#define init_5_8_7_8
void FASTCALL call_5_8_7_8(SH2_struct* sh)
{
  SH2movll4(sh, 8,7, 8);
}
#define init_5_8_7_9
void FASTCALL call_5_8_7_9(SH2_struct* sh)
{
  SH2movll4(sh, 8,7, 9);
}
#define init_5_8_7_10
void FASTCALL call_5_8_7_10(SH2_struct* sh)
{
  SH2movll4(sh, 8,7, 10);
}
#define init_5_8_7_11
void FASTCALL call_5_8_7_11(SH2_struct* sh)
{
  SH2movll4(sh, 8,7, 11);
}
#define init_5_8_7_12
void FASTCALL call_5_8_7_12(SH2_struct* sh)
{
  SH2movll4(sh, 8,7, 12);
}
#define init_5_8_7_13
void FASTCALL call_5_8_7_13(SH2_struct* sh)
{
  SH2movll4(sh, 8,7, 13);
}
#define init_5_8_7_14
void FASTCALL call_5_8_7_14(SH2_struct* sh)
{
  SH2movll4(sh, 8,7, 14);
}
#define init_5_8_7_15
void FASTCALL call_5_8_7_15(SH2_struct* sh)
{
  SH2movll4(sh, 8,7, 15);
}
#define init_5_8_8_0
void FASTCALL call_5_8_8_0(SH2_struct* sh)
{
  SH2movll4(sh, 8,8, 0);
}
#define init_5_8_8_1
void FASTCALL call_5_8_8_1(SH2_struct* sh)
{
  SH2movll4(sh, 8,8, 1);
}
#define init_5_8_8_2
void FASTCALL call_5_8_8_2(SH2_struct* sh)
{
  SH2movll4(sh, 8,8, 2);
}
#define init_5_8_8_3
void FASTCALL call_5_8_8_3(SH2_struct* sh)
{
  SH2movll4(sh, 8,8, 3);
}
#define init_5_8_8_4
void FASTCALL call_5_8_8_4(SH2_struct* sh)
{
  SH2movll4(sh, 8,8, 4);
}
#define init_5_8_8_5
void FASTCALL call_5_8_8_5(SH2_struct* sh)
{
  SH2movll4(sh, 8,8, 5);
}
#define init_5_8_8_6
void FASTCALL call_5_8_8_6(SH2_struct* sh)
{
  SH2movll4(sh, 8,8, 6);
}
#define init_5_8_8_7
void FASTCALL call_5_8_8_7(SH2_struct* sh)
{
  SH2movll4(sh, 8,8, 7);
}
#define init_5_8_8_8
void FASTCALL call_5_8_8_8(SH2_struct* sh)
{
  SH2movll4(sh, 8,8, 8);
}
#define init_5_8_8_9
void FASTCALL call_5_8_8_9(SH2_struct* sh)
{
  SH2movll4(sh, 8,8, 9);
}
#define init_5_8_8_10
void FASTCALL call_5_8_8_10(SH2_struct* sh)
{
  SH2movll4(sh, 8,8, 10);
}
#define init_5_8_8_11
void FASTCALL call_5_8_8_11(SH2_struct* sh)
{
  SH2movll4(sh, 8,8, 11);
}
#define init_5_8_8_12
void FASTCALL call_5_8_8_12(SH2_struct* sh)
{
  SH2movll4(sh, 8,8, 12);
}
#define init_5_8_8_13
void FASTCALL call_5_8_8_13(SH2_struct* sh)
{
  SH2movll4(sh, 8,8, 13);
}
#define init_5_8_8_14
void FASTCALL call_5_8_8_14(SH2_struct* sh)
{
  SH2movll4(sh, 8,8, 14);
}
#define init_5_8_8_15
void FASTCALL call_5_8_8_15(SH2_struct* sh)
{
  SH2movll4(sh, 8,8, 15);
}
#define init_5_8_9_0
void FASTCALL call_5_8_9_0(SH2_struct* sh)
{
  SH2movll4(sh, 8,9, 0);
}
#define init_5_8_9_1
void FASTCALL call_5_8_9_1(SH2_struct* sh)
{
  SH2movll4(sh, 8,9, 1);
}
#define init_5_8_9_2
void FASTCALL call_5_8_9_2(SH2_struct* sh)
{
  SH2movll4(sh, 8,9, 2);
}
#define init_5_8_9_3
void FASTCALL call_5_8_9_3(SH2_struct* sh)
{
  SH2movll4(sh, 8,9, 3);
}
#define init_5_8_9_4
void FASTCALL call_5_8_9_4(SH2_struct* sh)
{
  SH2movll4(sh, 8,9, 4);
}
#define init_5_8_9_5
void FASTCALL call_5_8_9_5(SH2_struct* sh)
{
  SH2movll4(sh, 8,9, 5);
}
#define init_5_8_9_6
void FASTCALL call_5_8_9_6(SH2_struct* sh)
{
  SH2movll4(sh, 8,9, 6);
}
#define init_5_8_9_7
void FASTCALL call_5_8_9_7(SH2_struct* sh)
{
  SH2movll4(sh, 8,9, 7);
}
#define init_5_8_9_8
void FASTCALL call_5_8_9_8(SH2_struct* sh)
{
  SH2movll4(sh, 8,9, 8);
}
#define init_5_8_9_9
void FASTCALL call_5_8_9_9(SH2_struct* sh)
{
  SH2movll4(sh, 8,9, 9);
}
#define init_5_8_9_10
void FASTCALL call_5_8_9_10(SH2_struct* sh)
{
  SH2movll4(sh, 8,9, 10);
}
#define init_5_8_9_11
void FASTCALL call_5_8_9_11(SH2_struct* sh)
{
  SH2movll4(sh, 8,9, 11);
}
#define init_5_8_9_12
void FASTCALL call_5_8_9_12(SH2_struct* sh)
{
  SH2movll4(sh, 8,9, 12);
}
#define init_5_8_9_13
void FASTCALL call_5_8_9_13(SH2_struct* sh)
{
  SH2movll4(sh, 8,9, 13);
}
#define init_5_8_9_14
void FASTCALL call_5_8_9_14(SH2_struct* sh)
{
  SH2movll4(sh, 8,9, 14);
}
#define init_5_8_9_15
void FASTCALL call_5_8_9_15(SH2_struct* sh)
{
  SH2movll4(sh, 8,9, 15);
}
#define init_5_8_10_0
void FASTCALL call_5_8_10_0(SH2_struct* sh)
{
  SH2movll4(sh, 8,10, 0);
}
#define init_5_8_10_1
void FASTCALL call_5_8_10_1(SH2_struct* sh)
{
  SH2movll4(sh, 8,10, 1);
}
#define init_5_8_10_2
void FASTCALL call_5_8_10_2(SH2_struct* sh)
{
  SH2movll4(sh, 8,10, 2);
}
#define init_5_8_10_3
void FASTCALL call_5_8_10_3(SH2_struct* sh)
{
  SH2movll4(sh, 8,10, 3);
}
#define init_5_8_10_4
void FASTCALL call_5_8_10_4(SH2_struct* sh)
{
  SH2movll4(sh, 8,10, 4);
}
#define init_5_8_10_5
void FASTCALL call_5_8_10_5(SH2_struct* sh)
{
  SH2movll4(sh, 8,10, 5);
}
#define init_5_8_10_6
void FASTCALL call_5_8_10_6(SH2_struct* sh)
{
  SH2movll4(sh, 8,10, 6);
}
#define init_5_8_10_7
void FASTCALL call_5_8_10_7(SH2_struct* sh)
{
  SH2movll4(sh, 8,10, 7);
}
#define init_5_8_10_8
void FASTCALL call_5_8_10_8(SH2_struct* sh)
{
  SH2movll4(sh, 8,10, 8);
}
#define init_5_8_10_9
void FASTCALL call_5_8_10_9(SH2_struct* sh)
{
  SH2movll4(sh, 8,10, 9);
}
#define init_5_8_10_10
void FASTCALL call_5_8_10_10(SH2_struct* sh)
{
  SH2movll4(sh, 8,10, 10);
}
#define init_5_8_10_11
void FASTCALL call_5_8_10_11(SH2_struct* sh)
{
  SH2movll4(sh, 8,10, 11);
}
#define init_5_8_10_12
void FASTCALL call_5_8_10_12(SH2_struct* sh)
{
  SH2movll4(sh, 8,10, 12);
}
#define init_5_8_10_13
void FASTCALL call_5_8_10_13(SH2_struct* sh)
{
  SH2movll4(sh, 8,10, 13);
}
#define init_5_8_10_14
void FASTCALL call_5_8_10_14(SH2_struct* sh)
{
  SH2movll4(sh, 8,10, 14);
}
#define init_5_8_10_15
void FASTCALL call_5_8_10_15(SH2_struct* sh)
{
  SH2movll4(sh, 8,10, 15);
}
#define init_5_8_11_0
void FASTCALL call_5_8_11_0(SH2_struct* sh)
{
  SH2movll4(sh, 8,11, 0);
}
#define init_5_8_11_1
void FASTCALL call_5_8_11_1(SH2_struct* sh)
{
  SH2movll4(sh, 8,11, 1);
}
#define init_5_8_11_2
void FASTCALL call_5_8_11_2(SH2_struct* sh)
{
  SH2movll4(sh, 8,11, 2);
}
#define init_5_8_11_3
void FASTCALL call_5_8_11_3(SH2_struct* sh)
{
  SH2movll4(sh, 8,11, 3);
}
#define init_5_8_11_4
void FASTCALL call_5_8_11_4(SH2_struct* sh)
{
  SH2movll4(sh, 8,11, 4);
}
#define init_5_8_11_5
void FASTCALL call_5_8_11_5(SH2_struct* sh)
{
  SH2movll4(sh, 8,11, 5);
}
#define init_5_8_11_6
void FASTCALL call_5_8_11_6(SH2_struct* sh)
{
  SH2movll4(sh, 8,11, 6);
}
#define init_5_8_11_7
void FASTCALL call_5_8_11_7(SH2_struct* sh)
{
  SH2movll4(sh, 8,11, 7);
}
#define init_5_8_11_8
void FASTCALL call_5_8_11_8(SH2_struct* sh)
{
  SH2movll4(sh, 8,11, 8);
}
#define init_5_8_11_9
void FASTCALL call_5_8_11_9(SH2_struct* sh)
{
  SH2movll4(sh, 8,11, 9);
}
#define init_5_8_11_10
void FASTCALL call_5_8_11_10(SH2_struct* sh)
{
  SH2movll4(sh, 8,11, 10);
}
#define init_5_8_11_11
void FASTCALL call_5_8_11_11(SH2_struct* sh)
{
  SH2movll4(sh, 8,11, 11);
}
#define init_5_8_11_12
void FASTCALL call_5_8_11_12(SH2_struct* sh)
{
  SH2movll4(sh, 8,11, 12);
}
#define init_5_8_11_13
void FASTCALL call_5_8_11_13(SH2_struct* sh)
{
  SH2movll4(sh, 8,11, 13);
}
#define init_5_8_11_14
void FASTCALL call_5_8_11_14(SH2_struct* sh)
{
  SH2movll4(sh, 8,11, 14);
}
#define init_5_8_11_15
void FASTCALL call_5_8_11_15(SH2_struct* sh)
{
  SH2movll4(sh, 8,11, 15);
}
#define init_5_8_12_0
void FASTCALL call_5_8_12_0(SH2_struct* sh)
{
  SH2movll4(sh, 8,12, 0);
}
#define init_5_8_12_1
void FASTCALL call_5_8_12_1(SH2_struct* sh)
{
  SH2movll4(sh, 8,12, 1);
}
#define init_5_8_12_2
void FASTCALL call_5_8_12_2(SH2_struct* sh)
{
  SH2movll4(sh, 8,12, 2);
}
#define init_5_8_12_3
void FASTCALL call_5_8_12_3(SH2_struct* sh)
{
  SH2movll4(sh, 8,12, 3);
}
#define init_5_8_12_4
void FASTCALL call_5_8_12_4(SH2_struct* sh)
{
  SH2movll4(sh, 8,12, 4);
}
#define init_5_8_12_5
void FASTCALL call_5_8_12_5(SH2_struct* sh)
{
  SH2movll4(sh, 8,12, 5);
}
#define init_5_8_12_6
void FASTCALL call_5_8_12_6(SH2_struct* sh)
{
  SH2movll4(sh, 8,12, 6);
}
#define init_5_8_12_7
void FASTCALL call_5_8_12_7(SH2_struct* sh)
{
  SH2movll4(sh, 8,12, 7);
}
#define init_5_8_12_8
void FASTCALL call_5_8_12_8(SH2_struct* sh)
{
  SH2movll4(sh, 8,12, 8);
}
#define init_5_8_12_9
void FASTCALL call_5_8_12_9(SH2_struct* sh)
{
  SH2movll4(sh, 8,12, 9);
}
#define init_5_8_12_10
void FASTCALL call_5_8_12_10(SH2_struct* sh)
{
  SH2movll4(sh, 8,12, 10);
}
#define init_5_8_12_11
void FASTCALL call_5_8_12_11(SH2_struct* sh)
{
  SH2movll4(sh, 8,12, 11);
}
#define init_5_8_12_12
void FASTCALL call_5_8_12_12(SH2_struct* sh)
{
  SH2movll4(sh, 8,12, 12);
}
#define init_5_8_12_13
void FASTCALL call_5_8_12_13(SH2_struct* sh)
{
  SH2movll4(sh, 8,12, 13);
}
#define init_5_8_12_14
void FASTCALL call_5_8_12_14(SH2_struct* sh)
{
  SH2movll4(sh, 8,12, 14);
}
#define init_5_8_12_15
void FASTCALL call_5_8_12_15(SH2_struct* sh)
{
  SH2movll4(sh, 8,12, 15);
}
#define init_5_8_13_0
void FASTCALL call_5_8_13_0(SH2_struct* sh)
{
  SH2movll4(sh, 8,13, 0);
}
#define init_5_8_13_1
void FASTCALL call_5_8_13_1(SH2_struct* sh)
{
  SH2movll4(sh, 8,13, 1);
}
#define init_5_8_13_2
void FASTCALL call_5_8_13_2(SH2_struct* sh)
{
  SH2movll4(sh, 8,13, 2);
}
#define init_5_8_13_3
void FASTCALL call_5_8_13_3(SH2_struct* sh)
{
  SH2movll4(sh, 8,13, 3);
}
#define init_5_8_13_4
void FASTCALL call_5_8_13_4(SH2_struct* sh)
{
  SH2movll4(sh, 8,13, 4);
}
#define init_5_8_13_5
void FASTCALL call_5_8_13_5(SH2_struct* sh)
{
  SH2movll4(sh, 8,13, 5);
}
#define init_5_8_13_6
void FASTCALL call_5_8_13_6(SH2_struct* sh)
{
  SH2movll4(sh, 8,13, 6);
}
#define init_5_8_13_7
void FASTCALL call_5_8_13_7(SH2_struct* sh)
{
  SH2movll4(sh, 8,13, 7);
}
#define init_5_8_13_8
void FASTCALL call_5_8_13_8(SH2_struct* sh)
{
  SH2movll4(sh, 8,13, 8);
}
#define init_5_8_13_9
void FASTCALL call_5_8_13_9(SH2_struct* sh)
{
  SH2movll4(sh, 8,13, 9);
}
#define init_5_8_13_10
void FASTCALL call_5_8_13_10(SH2_struct* sh)
{
  SH2movll4(sh, 8,13, 10);
}
#define init_5_8_13_11
void FASTCALL call_5_8_13_11(SH2_struct* sh)
{
  SH2movll4(sh, 8,13, 11);
}
#define init_5_8_13_12
void FASTCALL call_5_8_13_12(SH2_struct* sh)
{
  SH2movll4(sh, 8,13, 12);
}
#define init_5_8_13_13
void FASTCALL call_5_8_13_13(SH2_struct* sh)
{
  SH2movll4(sh, 8,13, 13);
}
#define init_5_8_13_14
void FASTCALL call_5_8_13_14(SH2_struct* sh)
{
  SH2movll4(sh, 8,13, 14);
}
#define init_5_8_13_15
void FASTCALL call_5_8_13_15(SH2_struct* sh)
{
  SH2movll4(sh, 8,13, 15);
}
#define init_5_8_14_0
void FASTCALL call_5_8_14_0(SH2_struct* sh)
{
  SH2movll4(sh, 8,14, 0);
}
#define init_5_8_14_1
void FASTCALL call_5_8_14_1(SH2_struct* sh)
{
  SH2movll4(sh, 8,14, 1);
}
#define init_5_8_14_2
void FASTCALL call_5_8_14_2(SH2_struct* sh)
{
  SH2movll4(sh, 8,14, 2);
}
#define init_5_8_14_3
void FASTCALL call_5_8_14_3(SH2_struct* sh)
{
  SH2movll4(sh, 8,14, 3);
}
#define init_5_8_14_4
void FASTCALL call_5_8_14_4(SH2_struct* sh)
{
  SH2movll4(sh, 8,14, 4);
}
#define init_5_8_14_5
void FASTCALL call_5_8_14_5(SH2_struct* sh)
{
  SH2movll4(sh, 8,14, 5);
}
#define init_5_8_14_6
void FASTCALL call_5_8_14_6(SH2_struct* sh)
{
  SH2movll4(sh, 8,14, 6);
}
#define init_5_8_14_7
void FASTCALL call_5_8_14_7(SH2_struct* sh)
{
  SH2movll4(sh, 8,14, 7);
}
#define init_5_8_14_8
void FASTCALL call_5_8_14_8(SH2_struct* sh)
{
  SH2movll4(sh, 8,14, 8);
}
#define init_5_8_14_9
void FASTCALL call_5_8_14_9(SH2_struct* sh)
{
  SH2movll4(sh, 8,14, 9);
}
#define init_5_8_14_10
void FASTCALL call_5_8_14_10(SH2_struct* sh)
{
  SH2movll4(sh, 8,14, 10);
}
#define init_5_8_14_11
void FASTCALL call_5_8_14_11(SH2_struct* sh)
{
  SH2movll4(sh, 8,14, 11);
}
#define init_5_8_14_12
void FASTCALL call_5_8_14_12(SH2_struct* sh)
{
  SH2movll4(sh, 8,14, 12);
}
#define init_5_8_14_13
void FASTCALL call_5_8_14_13(SH2_struct* sh)
{
  SH2movll4(sh, 8,14, 13);
}
#define init_5_8_14_14
void FASTCALL call_5_8_14_14(SH2_struct* sh)
{
  SH2movll4(sh, 8,14, 14);
}
#define init_5_8_14_15
void FASTCALL call_5_8_14_15(SH2_struct* sh)
{
  SH2movll4(sh, 8,14, 15);
}
#define init_5_8_15_0
void FASTCALL call_5_8_15_0(SH2_struct* sh)
{
  SH2movll4(sh, 8,15, 0);
}
#define init_5_8_15_1
void FASTCALL call_5_8_15_1(SH2_struct* sh)
{
  SH2movll4(sh, 8,15, 1);
}
#define init_5_8_15_2
void FASTCALL call_5_8_15_2(SH2_struct* sh)
{
  SH2movll4(sh, 8,15, 2);
}
#define init_5_8_15_3
void FASTCALL call_5_8_15_3(SH2_struct* sh)
{
  SH2movll4(sh, 8,15, 3);
}
#define init_5_8_15_4
void FASTCALL call_5_8_15_4(SH2_struct* sh)
{
  SH2movll4(sh, 8,15, 4);
}
#define init_5_8_15_5
void FASTCALL call_5_8_15_5(SH2_struct* sh)
{
  SH2movll4(sh, 8,15, 5);
}
#define init_5_8_15_6
void FASTCALL call_5_8_15_6(SH2_struct* sh)
{
  SH2movll4(sh, 8,15, 6);
}
#define init_5_8_15_7
void FASTCALL call_5_8_15_7(SH2_struct* sh)
{
  SH2movll4(sh, 8,15, 7);
}
#define init_5_8_15_8
void FASTCALL call_5_8_15_8(SH2_struct* sh)
{
  SH2movll4(sh, 8,15, 8);
}
#define init_5_8_15_9
void FASTCALL call_5_8_15_9(SH2_struct* sh)
{
  SH2movll4(sh, 8,15, 9);
}
#define init_5_8_15_10
void FASTCALL call_5_8_15_10(SH2_struct* sh)
{
  SH2movll4(sh, 8,15, 10);
}
#define init_5_8_15_11
void FASTCALL call_5_8_15_11(SH2_struct* sh)
{
  SH2movll4(sh, 8,15, 11);
}
#define init_5_8_15_12
void FASTCALL call_5_8_15_12(SH2_struct* sh)
{
  SH2movll4(sh, 8,15, 12);
}
#define init_5_8_15_13
void FASTCALL call_5_8_15_13(SH2_struct* sh)
{
  SH2movll4(sh, 8,15, 13);
}
#define init_5_8_15_14
void FASTCALL call_5_8_15_14(SH2_struct* sh)
{
  SH2movll4(sh, 8,15, 14);
}
#define init_5_8_15_15
void FASTCALL call_5_8_15_15(SH2_struct* sh)
{
  SH2movll4(sh, 8,15, 15);
}
#define init_5_9_0_0
void FASTCALL call_5_9_0_0(SH2_struct* sh)
{
  SH2movll4(sh, 9,0, 0);
}
#define init_5_9_0_1
void FASTCALL call_5_9_0_1(SH2_struct* sh)
{
  SH2movll4(sh, 9,0, 1);
}
#define init_5_9_0_2
void FASTCALL call_5_9_0_2(SH2_struct* sh)
{
  SH2movll4(sh, 9,0, 2);
}
#define init_5_9_0_3
void FASTCALL call_5_9_0_3(SH2_struct* sh)
{
  SH2movll4(sh, 9,0, 3);
}
#define init_5_9_0_4
void FASTCALL call_5_9_0_4(SH2_struct* sh)
{
  SH2movll4(sh, 9,0, 4);
}
#define init_5_9_0_5
void FASTCALL call_5_9_0_5(SH2_struct* sh)
{
  SH2movll4(sh, 9,0, 5);
}
#define init_5_9_0_6
void FASTCALL call_5_9_0_6(SH2_struct* sh)
{
  SH2movll4(sh, 9,0, 6);
}
#define init_5_9_0_7
void FASTCALL call_5_9_0_7(SH2_struct* sh)
{
  SH2movll4(sh, 9,0, 7);
}
#define init_5_9_0_8
void FASTCALL call_5_9_0_8(SH2_struct* sh)
{
  SH2movll4(sh, 9,0, 8);
}
#define init_5_9_0_9
void FASTCALL call_5_9_0_9(SH2_struct* sh)
{
  SH2movll4(sh, 9,0, 9);
}
#define init_5_9_0_10
void FASTCALL call_5_9_0_10(SH2_struct* sh)
{
  SH2movll4(sh, 9,0, 10);
}
#define init_5_9_0_11
void FASTCALL call_5_9_0_11(SH2_struct* sh)
{
  SH2movll4(sh, 9,0, 11);
}
#define init_5_9_0_12
void FASTCALL call_5_9_0_12(SH2_struct* sh)
{
  SH2movll4(sh, 9,0, 12);
}
#define init_5_9_0_13
void FASTCALL call_5_9_0_13(SH2_struct* sh)
{
  SH2movll4(sh, 9,0, 13);
}
#define init_5_9_0_14
void FASTCALL call_5_9_0_14(SH2_struct* sh)
{
  SH2movll4(sh, 9,0, 14);
}
#define init_5_9_0_15
void FASTCALL call_5_9_0_15(SH2_struct* sh)
{
  SH2movll4(sh, 9,0, 15);
}
#define init_5_9_1_0
void FASTCALL call_5_9_1_0(SH2_struct* sh)
{
  SH2movll4(sh, 9,1, 0);
}
#define init_5_9_1_1
void FASTCALL call_5_9_1_1(SH2_struct* sh)
{
  SH2movll4(sh, 9,1, 1);
}
#define init_5_9_1_2
void FASTCALL call_5_9_1_2(SH2_struct* sh)
{
  SH2movll4(sh, 9,1, 2);
}
#define init_5_9_1_3
void FASTCALL call_5_9_1_3(SH2_struct* sh)
{
  SH2movll4(sh, 9,1, 3);
}
#define init_5_9_1_4
void FASTCALL call_5_9_1_4(SH2_struct* sh)
{
  SH2movll4(sh, 9,1, 4);
}
#define init_5_9_1_5
void FASTCALL call_5_9_1_5(SH2_struct* sh)
{
  SH2movll4(sh, 9,1, 5);
}
#define init_5_9_1_6
void FASTCALL call_5_9_1_6(SH2_struct* sh)
{
  SH2movll4(sh, 9,1, 6);
}
#define init_5_9_1_7
void FASTCALL call_5_9_1_7(SH2_struct* sh)
{
  SH2movll4(sh, 9,1, 7);
}
#define init_5_9_1_8
void FASTCALL call_5_9_1_8(SH2_struct* sh)
{
  SH2movll4(sh, 9,1, 8);
}
#define init_5_9_1_9
void FASTCALL call_5_9_1_9(SH2_struct* sh)
{
  SH2movll4(sh, 9,1, 9);
}
#define init_5_9_1_10
void FASTCALL call_5_9_1_10(SH2_struct* sh)
{
  SH2movll4(sh, 9,1, 10);
}
#define init_5_9_1_11
void FASTCALL call_5_9_1_11(SH2_struct* sh)
{
  SH2movll4(sh, 9,1, 11);
}
#define init_5_9_1_12
void FASTCALL call_5_9_1_12(SH2_struct* sh)
{
  SH2movll4(sh, 9,1, 12);
}
#define init_5_9_1_13
void FASTCALL call_5_9_1_13(SH2_struct* sh)
{
  SH2movll4(sh, 9,1, 13);
}
#define init_5_9_1_14
void FASTCALL call_5_9_1_14(SH2_struct* sh)
{
  SH2movll4(sh, 9,1, 14);
}
#define init_5_9_1_15
void FASTCALL call_5_9_1_15(SH2_struct* sh)
{
  SH2movll4(sh, 9,1, 15);
}
#define init_5_9_2_0
void FASTCALL call_5_9_2_0(SH2_struct* sh)
{
  SH2movll4(sh, 9,2, 0);
}
#define init_5_9_2_1
void FASTCALL call_5_9_2_1(SH2_struct* sh)
{
  SH2movll4(sh, 9,2, 1);
}
#define init_5_9_2_2
void FASTCALL call_5_9_2_2(SH2_struct* sh)
{
  SH2movll4(sh, 9,2, 2);
}
#define init_5_9_2_3
void FASTCALL call_5_9_2_3(SH2_struct* sh)
{
  SH2movll4(sh, 9,2, 3);
}
#define init_5_9_2_4
void FASTCALL call_5_9_2_4(SH2_struct* sh)
{
  SH2movll4(sh, 9,2, 4);
}
#define init_5_9_2_5
void FASTCALL call_5_9_2_5(SH2_struct* sh)
{
  SH2movll4(sh, 9,2, 5);
}
#define init_5_9_2_6
void FASTCALL call_5_9_2_6(SH2_struct* sh)
{
  SH2movll4(sh, 9,2, 6);
}
#define init_5_9_2_7
void FASTCALL call_5_9_2_7(SH2_struct* sh)
{
  SH2movll4(sh, 9,2, 7);
}
#define init_5_9_2_8
void FASTCALL call_5_9_2_8(SH2_struct* sh)
{
  SH2movll4(sh, 9,2, 8);
}
#define init_5_9_2_9
void FASTCALL call_5_9_2_9(SH2_struct* sh)
{
  SH2movll4(sh, 9,2, 9);
}
#define init_5_9_2_10
void FASTCALL call_5_9_2_10(SH2_struct* sh)
{
  SH2movll4(sh, 9,2, 10);
}
#define init_5_9_2_11
void FASTCALL call_5_9_2_11(SH2_struct* sh)
{
  SH2movll4(sh, 9,2, 11);
}
#define init_5_9_2_12
void FASTCALL call_5_9_2_12(SH2_struct* sh)
{
  SH2movll4(sh, 9,2, 12);
}
#define init_5_9_2_13
void FASTCALL call_5_9_2_13(SH2_struct* sh)
{
  SH2movll4(sh, 9,2, 13);
}
#define init_5_9_2_14
void FASTCALL call_5_9_2_14(SH2_struct* sh)
{
  SH2movll4(sh, 9,2, 14);
}
#define init_5_9_2_15
void FASTCALL call_5_9_2_15(SH2_struct* sh)
{
  SH2movll4(sh, 9,2, 15);
}
#define init_5_9_3_0
void FASTCALL call_5_9_3_0(SH2_struct* sh)
{
  SH2movll4(sh, 9,3, 0);
}
#define init_5_9_3_1
void FASTCALL call_5_9_3_1(SH2_struct* sh)
{
  SH2movll4(sh, 9,3, 1);
}
#define init_5_9_3_2
void FASTCALL call_5_9_3_2(SH2_struct* sh)
{
  SH2movll4(sh, 9,3, 2);
}
#define init_5_9_3_3
void FASTCALL call_5_9_3_3(SH2_struct* sh)
{
  SH2movll4(sh, 9,3, 3);
}
#define init_5_9_3_4
void FASTCALL call_5_9_3_4(SH2_struct* sh)
{
  SH2movll4(sh, 9,3, 4);
}
#define init_5_9_3_5
void FASTCALL call_5_9_3_5(SH2_struct* sh)
{
  SH2movll4(sh, 9,3, 5);
}
#define init_5_9_3_6
void FASTCALL call_5_9_3_6(SH2_struct* sh)
{
  SH2movll4(sh, 9,3, 6);
}
#define init_5_9_3_7
void FASTCALL call_5_9_3_7(SH2_struct* sh)
{
  SH2movll4(sh, 9,3, 7);
}
#define init_5_9_3_8
void FASTCALL call_5_9_3_8(SH2_struct* sh)
{
  SH2movll4(sh, 9,3, 8);
}
#define init_5_9_3_9
void FASTCALL call_5_9_3_9(SH2_struct* sh)
{
  SH2movll4(sh, 9,3, 9);
}
#define init_5_9_3_10
void FASTCALL call_5_9_3_10(SH2_struct* sh)
{
  SH2movll4(sh, 9,3, 10);
}
#define init_5_9_3_11
void FASTCALL call_5_9_3_11(SH2_struct* sh)
{
  SH2movll4(sh, 9,3, 11);
}
#define init_5_9_3_12
void FASTCALL call_5_9_3_12(SH2_struct* sh)
{
  SH2movll4(sh, 9,3, 12);
}
#define init_5_9_3_13
void FASTCALL call_5_9_3_13(SH2_struct* sh)
{
  SH2movll4(sh, 9,3, 13);
}
#define init_5_9_3_14
void FASTCALL call_5_9_3_14(SH2_struct* sh)
{
  SH2movll4(sh, 9,3, 14);
}
#define init_5_9_3_15
void FASTCALL call_5_9_3_15(SH2_struct* sh)
{
  SH2movll4(sh, 9,3, 15);
}
#define init_5_9_4_0
void FASTCALL call_5_9_4_0(SH2_struct* sh)
{
  SH2movll4(sh, 9,4, 0);
}
#define init_5_9_4_1
void FASTCALL call_5_9_4_1(SH2_struct* sh)
{
  SH2movll4(sh, 9,4, 1);
}
#define init_5_9_4_2
void FASTCALL call_5_9_4_2(SH2_struct* sh)
{
  SH2movll4(sh, 9,4, 2);
}
#define init_5_9_4_3
void FASTCALL call_5_9_4_3(SH2_struct* sh)
{
  SH2movll4(sh, 9,4, 3);
}
#define init_5_9_4_4
void FASTCALL call_5_9_4_4(SH2_struct* sh)
{
  SH2movll4(sh, 9,4, 4);
}
#define init_5_9_4_5
void FASTCALL call_5_9_4_5(SH2_struct* sh)
{
  SH2movll4(sh, 9,4, 5);
}
#define init_5_9_4_6
void FASTCALL call_5_9_4_6(SH2_struct* sh)
{
  SH2movll4(sh, 9,4, 6);
}
#define init_5_9_4_7
void FASTCALL call_5_9_4_7(SH2_struct* sh)
{
  SH2movll4(sh, 9,4, 7);
}
#define init_5_9_4_8
void FASTCALL call_5_9_4_8(SH2_struct* sh)
{
  SH2movll4(sh, 9,4, 8);
}
#define init_5_9_4_9
void FASTCALL call_5_9_4_9(SH2_struct* sh)
{
  SH2movll4(sh, 9,4, 9);
}
#define init_5_9_4_10
void FASTCALL call_5_9_4_10(SH2_struct* sh)
{
  SH2movll4(sh, 9,4, 10);
}
#define init_5_9_4_11
void FASTCALL call_5_9_4_11(SH2_struct* sh)
{
  SH2movll4(sh, 9,4, 11);
}
#define init_5_9_4_12
void FASTCALL call_5_9_4_12(SH2_struct* sh)
{
  SH2movll4(sh, 9,4, 12);
}
#define init_5_9_4_13
void FASTCALL call_5_9_4_13(SH2_struct* sh)
{
  SH2movll4(sh, 9,4, 13);
}
#define init_5_9_4_14
void FASTCALL call_5_9_4_14(SH2_struct* sh)
{
  SH2movll4(sh, 9,4, 14);
}
#define init_5_9_4_15
void FASTCALL call_5_9_4_15(SH2_struct* sh)
{
  SH2movll4(sh, 9,4, 15);
}
#define init_5_9_5_0
void FASTCALL call_5_9_5_0(SH2_struct* sh)
{
  SH2movll4(sh, 9,5, 0);
}
#define init_5_9_5_1
void FASTCALL call_5_9_5_1(SH2_struct* sh)
{
  SH2movll4(sh, 9,5, 1);
}
#define init_5_9_5_2
void FASTCALL call_5_9_5_2(SH2_struct* sh)
{
  SH2movll4(sh, 9,5, 2);
}
#define init_5_9_5_3
void FASTCALL call_5_9_5_3(SH2_struct* sh)
{
  SH2movll4(sh, 9,5, 3);
}
#define init_5_9_5_4
void FASTCALL call_5_9_5_4(SH2_struct* sh)
{
  SH2movll4(sh, 9,5, 4);
}
#define init_5_9_5_5
void FASTCALL call_5_9_5_5(SH2_struct* sh)
{
  SH2movll4(sh, 9,5, 5);
}
#define init_5_9_5_6
void FASTCALL call_5_9_5_6(SH2_struct* sh)
{
  SH2movll4(sh, 9,5, 6);
}
#define init_5_9_5_7
void FASTCALL call_5_9_5_7(SH2_struct* sh)
{
  SH2movll4(sh, 9,5, 7);
}
#define init_5_9_5_8
void FASTCALL call_5_9_5_8(SH2_struct* sh)
{
  SH2movll4(sh, 9,5, 8);
}
#define init_5_9_5_9
void FASTCALL call_5_9_5_9(SH2_struct* sh)
{
  SH2movll4(sh, 9,5, 9);
}
#define init_5_9_5_10
void FASTCALL call_5_9_5_10(SH2_struct* sh)
{
  SH2movll4(sh, 9,5, 10);
}
#define init_5_9_5_11
void FASTCALL call_5_9_5_11(SH2_struct* sh)
{
  SH2movll4(sh, 9,5, 11);
}
#define init_5_9_5_12
void FASTCALL call_5_9_5_12(SH2_struct* sh)
{
  SH2movll4(sh, 9,5, 12);
}
#define init_5_9_5_13
void FASTCALL call_5_9_5_13(SH2_struct* sh)
{
  SH2movll4(sh, 9,5, 13);
}
#define init_5_9_5_14
void FASTCALL call_5_9_5_14(SH2_struct* sh)
{
  SH2movll4(sh, 9,5, 14);
}
#define init_5_9_5_15
void FASTCALL call_5_9_5_15(SH2_struct* sh)
{
  SH2movll4(sh, 9,5, 15);
}
#define init_5_9_6_0
void FASTCALL call_5_9_6_0(SH2_struct* sh)
{
  SH2movll4(sh, 9,6, 0);
}
#define init_5_9_6_1
void FASTCALL call_5_9_6_1(SH2_struct* sh)
{
  SH2movll4(sh, 9,6, 1);
}
#define init_5_9_6_2
void FASTCALL call_5_9_6_2(SH2_struct* sh)
{
  SH2movll4(sh, 9,6, 2);
}
#define init_5_9_6_3
void FASTCALL call_5_9_6_3(SH2_struct* sh)
{
  SH2movll4(sh, 9,6, 3);
}
#define init_5_9_6_4
void FASTCALL call_5_9_6_4(SH2_struct* sh)
{
  SH2movll4(sh, 9,6, 4);
}
#define init_5_9_6_5
void FASTCALL call_5_9_6_5(SH2_struct* sh)
{
  SH2movll4(sh, 9,6, 5);
}
#define init_5_9_6_6
void FASTCALL call_5_9_6_6(SH2_struct* sh)
{
  SH2movll4(sh, 9,6, 6);
}
#define init_5_9_6_7
void FASTCALL call_5_9_6_7(SH2_struct* sh)
{
  SH2movll4(sh, 9,6, 7);
}
#define init_5_9_6_8
void FASTCALL call_5_9_6_8(SH2_struct* sh)
{
  SH2movll4(sh, 9,6, 8);
}
#define init_5_9_6_9
void FASTCALL call_5_9_6_9(SH2_struct* sh)
{
  SH2movll4(sh, 9,6, 9);
}
#define init_5_9_6_10
void FASTCALL call_5_9_6_10(SH2_struct* sh)
{
  SH2movll4(sh, 9,6, 10);
}
#define init_5_9_6_11
void FASTCALL call_5_9_6_11(SH2_struct* sh)
{
  SH2movll4(sh, 9,6, 11);
}
#define init_5_9_6_12
void FASTCALL call_5_9_6_12(SH2_struct* sh)
{
  SH2movll4(sh, 9,6, 12);
}
#define init_5_9_6_13
void FASTCALL call_5_9_6_13(SH2_struct* sh)
{
  SH2movll4(sh, 9,6, 13);
}
#define init_5_9_6_14
void FASTCALL call_5_9_6_14(SH2_struct* sh)
{
  SH2movll4(sh, 9,6, 14);
}
#define init_5_9_6_15
void FASTCALL call_5_9_6_15(SH2_struct* sh)
{
  SH2movll4(sh, 9,6, 15);
}
#define init_5_9_7_0
void FASTCALL call_5_9_7_0(SH2_struct* sh)
{
  SH2movll4(sh, 9,7, 0);
}
#define init_5_9_7_1
void FASTCALL call_5_9_7_1(SH2_struct* sh)
{
  SH2movll4(sh, 9,7, 1);
}
#define init_5_9_7_2
void FASTCALL call_5_9_7_2(SH2_struct* sh)
{
  SH2movll4(sh, 9,7, 2);
}
#define init_5_9_7_3
void FASTCALL call_5_9_7_3(SH2_struct* sh)
{
  SH2movll4(sh, 9,7, 3);
}
#define init_5_9_7_4
void FASTCALL call_5_9_7_4(SH2_struct* sh)
{
  SH2movll4(sh, 9,7, 4);
}
#define init_5_9_7_5
void FASTCALL call_5_9_7_5(SH2_struct* sh)
{
  SH2movll4(sh, 9,7, 5);
}
#define init_5_9_7_6
void FASTCALL call_5_9_7_6(SH2_struct* sh)
{
  SH2movll4(sh, 9,7, 6);
}
#define init_5_9_7_7
void FASTCALL call_5_9_7_7(SH2_struct* sh)
{
  SH2movll4(sh, 9,7, 7);
}
#define init_5_9_7_8
void FASTCALL call_5_9_7_8(SH2_struct* sh)
{
  SH2movll4(sh, 9,7, 8);
}
#define init_5_9_7_9
void FASTCALL call_5_9_7_9(SH2_struct* sh)
{
  SH2movll4(sh, 9,7, 9);
}
#define init_5_9_7_10
void FASTCALL call_5_9_7_10(SH2_struct* sh)
{
  SH2movll4(sh, 9,7, 10);
}
#define init_5_9_7_11
void FASTCALL call_5_9_7_11(SH2_struct* sh)
{
  SH2movll4(sh, 9,7, 11);
}
#define init_5_9_7_12
void FASTCALL call_5_9_7_12(SH2_struct* sh)
{
  SH2movll4(sh, 9,7, 12);
}
#define init_5_9_7_13
void FASTCALL call_5_9_7_13(SH2_struct* sh)
{
  SH2movll4(sh, 9,7, 13);
}
#define init_5_9_7_14
void FASTCALL call_5_9_7_14(SH2_struct* sh)
{
  SH2movll4(sh, 9,7, 14);
}
#define init_5_9_7_15
void FASTCALL call_5_9_7_15(SH2_struct* sh)
{
  SH2movll4(sh, 9,7, 15);
}
#define init_5_9_8_0
void FASTCALL call_5_9_8_0(SH2_struct* sh)
{
  SH2movll4(sh, 9,8, 0);
}
#define init_5_9_8_1
void FASTCALL call_5_9_8_1(SH2_struct* sh)
{
  SH2movll4(sh, 9,8, 1);
}
#define init_5_9_8_2
void FASTCALL call_5_9_8_2(SH2_struct* sh)
{
  SH2movll4(sh, 9,8, 2);
}
#define init_5_9_8_3
void FASTCALL call_5_9_8_3(SH2_struct* sh)
{
  SH2movll4(sh, 9,8, 3);
}
#define init_5_9_8_4
void FASTCALL call_5_9_8_4(SH2_struct* sh)
{
  SH2movll4(sh, 9,8, 4);
}
#define init_5_9_8_5
void FASTCALL call_5_9_8_5(SH2_struct* sh)
{
  SH2movll4(sh, 9,8, 5);
}
#define init_5_9_8_6
void FASTCALL call_5_9_8_6(SH2_struct* sh)
{
  SH2movll4(sh, 9,8, 6);
}
#define init_5_9_8_7
void FASTCALL call_5_9_8_7(SH2_struct* sh)
{
  SH2movll4(sh, 9,8, 7);
}
#define init_5_9_8_8
void FASTCALL call_5_9_8_8(SH2_struct* sh)
{
  SH2movll4(sh, 9,8, 8);
}
#define init_5_9_8_9
void FASTCALL call_5_9_8_9(SH2_struct* sh)
{
  SH2movll4(sh, 9,8, 9);
}
#define init_5_9_8_10
void FASTCALL call_5_9_8_10(SH2_struct* sh)
{
  SH2movll4(sh, 9,8, 10);
}
#define init_5_9_8_11
void FASTCALL call_5_9_8_11(SH2_struct* sh)
{
  SH2movll4(sh, 9,8, 11);
}
#define init_5_9_8_12
void FASTCALL call_5_9_8_12(SH2_struct* sh)
{
  SH2movll4(sh, 9,8, 12);
}
#define init_5_9_8_13
void FASTCALL call_5_9_8_13(SH2_struct* sh)
{
  SH2movll4(sh, 9,8, 13);
}
#define init_5_9_8_14
void FASTCALL call_5_9_8_14(SH2_struct* sh)
{
  SH2movll4(sh, 9,8, 14);
}
#define init_5_9_8_15
void FASTCALL call_5_9_8_15(SH2_struct* sh)
{
  SH2movll4(sh, 9,8, 15);
}
#define init_5_9_9_0
void FASTCALL call_5_9_9_0(SH2_struct* sh)
{
  SH2movll4(sh, 9,9, 0);
}
#define init_5_9_9_1
void FASTCALL call_5_9_9_1(SH2_struct* sh)
{
  SH2movll4(sh, 9,9, 1);
}
#define init_5_9_9_2
void FASTCALL call_5_9_9_2(SH2_struct* sh)
{
  SH2movll4(sh, 9,9, 2);
}
#define init_5_9_9_3
void FASTCALL call_5_9_9_3(SH2_struct* sh)
{
  SH2movll4(sh, 9,9, 3);
}
#define init_5_9_9_4
void FASTCALL call_5_9_9_4(SH2_struct* sh)
{
  SH2movll4(sh, 9,9, 4);
}
#define init_5_9_9_5
void FASTCALL call_5_9_9_5(SH2_struct* sh)
{
  SH2movll4(sh, 9,9, 5);
}
#define init_5_9_9_6
void FASTCALL call_5_9_9_6(SH2_struct* sh)
{
  SH2movll4(sh, 9,9, 6);
}
#define init_5_9_9_7
void FASTCALL call_5_9_9_7(SH2_struct* sh)
{
  SH2movll4(sh, 9,9, 7);
}
#define init_5_9_9_8
void FASTCALL call_5_9_9_8(SH2_struct* sh)
{
  SH2movll4(sh, 9,9, 8);
}
#define init_5_9_9_9
void FASTCALL call_5_9_9_9(SH2_struct* sh)
{
  SH2movll4(sh, 9,9, 9);
}
#define init_5_9_9_10
void FASTCALL call_5_9_9_10(SH2_struct* sh)
{
  SH2movll4(sh, 9,9, 10);
}
#define init_5_9_9_11
void FASTCALL call_5_9_9_11(SH2_struct* sh)
{
  SH2movll4(sh, 9,9, 11);
}
#define init_5_9_9_12
void FASTCALL call_5_9_9_12(SH2_struct* sh)
{
  SH2movll4(sh, 9,9, 12);
}
#define init_5_9_9_13
void FASTCALL call_5_9_9_13(SH2_struct* sh)
{
  SH2movll4(sh, 9,9, 13);
}
#define init_5_9_9_14
void FASTCALL call_5_9_9_14(SH2_struct* sh)
{
  SH2movll4(sh, 9,9, 14);
}
#define init_5_9_9_15
void FASTCALL call_5_9_9_15(SH2_struct* sh)
{
  SH2movll4(sh, 9,9, 15);
}
#define init_5_9_10_0
void FASTCALL call_5_9_10_0(SH2_struct* sh)
{
  SH2movll4(sh, 9,10, 0);
}
#define init_5_9_10_1
void FASTCALL call_5_9_10_1(SH2_struct* sh)
{
  SH2movll4(sh, 9,10, 1);
}
#define init_5_9_10_2
void FASTCALL call_5_9_10_2(SH2_struct* sh)
{
  SH2movll4(sh, 9,10, 2);
}
#define init_5_9_10_3
void FASTCALL call_5_9_10_3(SH2_struct* sh)
{
  SH2movll4(sh, 9,10, 3);
}
#define init_5_9_10_4
void FASTCALL call_5_9_10_4(SH2_struct* sh)
{
  SH2movll4(sh, 9,10, 4);
}
#define init_5_9_10_5
void FASTCALL call_5_9_10_5(SH2_struct* sh)
{
  SH2movll4(sh, 9,10, 5);
}
#define init_5_9_10_6
void FASTCALL call_5_9_10_6(SH2_struct* sh)
{
  SH2movll4(sh, 9,10, 6);
}
#define init_5_9_10_7
void FASTCALL call_5_9_10_7(SH2_struct* sh)
{
  SH2movll4(sh, 9,10, 7);
}
#define init_5_9_10_8
void FASTCALL call_5_9_10_8(SH2_struct* sh)
{
  SH2movll4(sh, 9,10, 8);
}
#define init_5_9_10_9
void FASTCALL call_5_9_10_9(SH2_struct* sh)
{
  SH2movll4(sh, 9,10, 9);
}
#define init_5_9_10_10
void FASTCALL call_5_9_10_10(SH2_struct* sh)
{
  SH2movll4(sh, 9,10, 10);
}
#define init_5_9_10_11
void FASTCALL call_5_9_10_11(SH2_struct* sh)
{
  SH2movll4(sh, 9,10, 11);
}
#define init_5_9_10_12
void FASTCALL call_5_9_10_12(SH2_struct* sh)
{
  SH2movll4(sh, 9,10, 12);
}
#define init_5_9_10_13
void FASTCALL call_5_9_10_13(SH2_struct* sh)
{
  SH2movll4(sh, 9,10, 13);
}
#define init_5_9_10_14
void FASTCALL call_5_9_10_14(SH2_struct* sh)
{
  SH2movll4(sh, 9,10, 14);
}
#define init_5_9_10_15
void FASTCALL call_5_9_10_15(SH2_struct* sh)
{
  SH2movll4(sh, 9,10, 15);
}
#define init_5_9_11_0
void FASTCALL call_5_9_11_0(SH2_struct* sh)
{
  SH2movll4(sh, 9,11, 0);
}
#define init_5_9_11_1
void FASTCALL call_5_9_11_1(SH2_struct* sh)
{
  SH2movll4(sh, 9,11, 1);
}
#define init_5_9_11_2
void FASTCALL call_5_9_11_2(SH2_struct* sh)
{
  SH2movll4(sh, 9,11, 2);
}
#define init_5_9_11_3
void FASTCALL call_5_9_11_3(SH2_struct* sh)
{
  SH2movll4(sh, 9,11, 3);
}
#define init_5_9_11_4
void FASTCALL call_5_9_11_4(SH2_struct* sh)
{
  SH2movll4(sh, 9,11, 4);
}
#define init_5_9_11_5
void FASTCALL call_5_9_11_5(SH2_struct* sh)
{
  SH2movll4(sh, 9,11, 5);
}
#define init_5_9_11_6
void FASTCALL call_5_9_11_6(SH2_struct* sh)
{
  SH2movll4(sh, 9,11, 6);
}
#define init_5_9_11_7
void FASTCALL call_5_9_11_7(SH2_struct* sh)
{
  SH2movll4(sh, 9,11, 7);
}
#define init_5_9_11_8
void FASTCALL call_5_9_11_8(SH2_struct* sh)
{
  SH2movll4(sh, 9,11, 8);
}
#define init_5_9_11_9
void FASTCALL call_5_9_11_9(SH2_struct* sh)
{
  SH2movll4(sh, 9,11, 9);
}
#define init_5_9_11_10
void FASTCALL call_5_9_11_10(SH2_struct* sh)
{
  SH2movll4(sh, 9,11, 10);
}
#define init_5_9_11_11
void FASTCALL call_5_9_11_11(SH2_struct* sh)
{
  SH2movll4(sh, 9,11, 11);
}
#define init_5_9_11_12
void FASTCALL call_5_9_11_12(SH2_struct* sh)
{
  SH2movll4(sh, 9,11, 12);
}
#define init_5_9_11_13
void FASTCALL call_5_9_11_13(SH2_struct* sh)
{
  SH2movll4(sh, 9,11, 13);
}
#define init_5_9_11_14
void FASTCALL call_5_9_11_14(SH2_struct* sh)
{
  SH2movll4(sh, 9,11, 14);
}
#define init_5_9_11_15
void FASTCALL call_5_9_11_15(SH2_struct* sh)
{
  SH2movll4(sh, 9,11, 15);
}
#define init_5_9_12_0
void FASTCALL call_5_9_12_0(SH2_struct* sh)
{
  SH2movll4(sh, 9,12, 0);
}
#define init_5_9_12_1
void FASTCALL call_5_9_12_1(SH2_struct* sh)
{
  SH2movll4(sh, 9,12, 1);
}
#define init_5_9_12_2
void FASTCALL call_5_9_12_2(SH2_struct* sh)
{
  SH2movll4(sh, 9,12, 2);
}
#define init_5_9_12_3
void FASTCALL call_5_9_12_3(SH2_struct* sh)
{
  SH2movll4(sh, 9,12, 3);
}
#define init_5_9_12_4
void FASTCALL call_5_9_12_4(SH2_struct* sh)
{
  SH2movll4(sh, 9,12, 4);
}
#define init_5_9_12_5
void FASTCALL call_5_9_12_5(SH2_struct* sh)
{
  SH2movll4(sh, 9,12, 5);
}
#define init_5_9_12_6
void FASTCALL call_5_9_12_6(SH2_struct* sh)
{
  SH2movll4(sh, 9,12, 6);
}
#define init_5_9_12_7
void FASTCALL call_5_9_12_7(SH2_struct* sh)
{
  SH2movll4(sh, 9,12, 7);
}
#define init_5_9_12_8
void FASTCALL call_5_9_12_8(SH2_struct* sh)
{
  SH2movll4(sh, 9,12, 8);
}
#define init_5_9_12_9
void FASTCALL call_5_9_12_9(SH2_struct* sh)
{
  SH2movll4(sh, 9,12, 9);
}
#define init_5_9_12_10
void FASTCALL call_5_9_12_10(SH2_struct* sh)
{
  SH2movll4(sh, 9,12, 10);
}
#define init_5_9_12_11
void FASTCALL call_5_9_12_11(SH2_struct* sh)
{
  SH2movll4(sh, 9,12, 11);
}
#define init_5_9_12_12
void FASTCALL call_5_9_12_12(SH2_struct* sh)
{
  SH2movll4(sh, 9,12, 12);
}
#define init_5_9_12_13
void FASTCALL call_5_9_12_13(SH2_struct* sh)
{
  SH2movll4(sh, 9,12, 13);
}
#define init_5_9_12_14
void FASTCALL call_5_9_12_14(SH2_struct* sh)
{
  SH2movll4(sh, 9,12, 14);
}
#define init_5_9_12_15
void FASTCALL call_5_9_12_15(SH2_struct* sh)
{
  SH2movll4(sh, 9,12, 15);
}
#define init_5_9_13_0
void FASTCALL call_5_9_13_0(SH2_struct* sh)
{
  SH2movll4(sh, 9,13, 0);
}
#define init_5_9_13_1
void FASTCALL call_5_9_13_1(SH2_struct* sh)
{
  SH2movll4(sh, 9,13, 1);
}
#define init_5_9_13_2
void FASTCALL call_5_9_13_2(SH2_struct* sh)
{
  SH2movll4(sh, 9,13, 2);
}
#define init_5_9_13_3
void FASTCALL call_5_9_13_3(SH2_struct* sh)
{
  SH2movll4(sh, 9,13, 3);
}
#define init_5_9_13_4
void FASTCALL call_5_9_13_4(SH2_struct* sh)
{
  SH2movll4(sh, 9,13, 4);
}
#define init_5_9_13_5
void FASTCALL call_5_9_13_5(SH2_struct* sh)
{
  SH2movll4(sh, 9,13, 5);
}
#define init_5_9_13_6
void FASTCALL call_5_9_13_6(SH2_struct* sh)
{
  SH2movll4(sh, 9,13, 6);
}
#define init_5_9_13_7
void FASTCALL call_5_9_13_7(SH2_struct* sh)
{
  SH2movll4(sh, 9,13, 7);
}
#define init_5_9_13_8
void FASTCALL call_5_9_13_8(SH2_struct* sh)
{
  SH2movll4(sh, 9,13, 8);
}
#define init_5_9_13_9
void FASTCALL call_5_9_13_9(SH2_struct* sh)
{
  SH2movll4(sh, 9,13, 9);
}
#define init_5_9_13_10
void FASTCALL call_5_9_13_10(SH2_struct* sh)
{
  SH2movll4(sh, 9,13, 10);
}
#define init_5_9_13_11
void FASTCALL call_5_9_13_11(SH2_struct* sh)
{
  SH2movll4(sh, 9,13, 11);
}
#define init_5_9_13_12
void FASTCALL call_5_9_13_12(SH2_struct* sh)
{
  SH2movll4(sh, 9,13, 12);
}
#define init_5_9_13_13
void FASTCALL call_5_9_13_13(SH2_struct* sh)
{
  SH2movll4(sh, 9,13, 13);
}
#define init_5_9_13_14
void FASTCALL call_5_9_13_14(SH2_struct* sh)
{
  SH2movll4(sh, 9,13, 14);
}
#define init_5_9_13_15
void FASTCALL call_5_9_13_15(SH2_struct* sh)
{
  SH2movll4(sh, 9,13, 15);
}
#define init_5_9_14_0
void FASTCALL call_5_9_14_0(SH2_struct* sh)
{
  SH2movll4(sh, 9,14, 0);
}
#define init_5_9_14_1
void FASTCALL call_5_9_14_1(SH2_struct* sh)
{
  SH2movll4(sh, 9,14, 1);
}
#define init_5_9_14_2
void FASTCALL call_5_9_14_2(SH2_struct* sh)
{
  SH2movll4(sh, 9,14, 2);
}
#define init_5_9_14_3
void FASTCALL call_5_9_14_3(SH2_struct* sh)
{
  SH2movll4(sh, 9,14, 3);
}
#define init_5_9_14_4
void FASTCALL call_5_9_14_4(SH2_struct* sh)
{
  SH2movll4(sh, 9,14, 4);
}
#define init_5_9_14_5
void FASTCALL call_5_9_14_5(SH2_struct* sh)
{
  SH2movll4(sh, 9,14, 5);
}
#define init_5_9_14_6
void FASTCALL call_5_9_14_6(SH2_struct* sh)
{
  SH2movll4(sh, 9,14, 6);
}
#define init_5_9_14_7
void FASTCALL call_5_9_14_7(SH2_struct* sh)
{
  SH2movll4(sh, 9,14, 7);
}
#define init_5_9_14_8
void FASTCALL call_5_9_14_8(SH2_struct* sh)
{
  SH2movll4(sh, 9,14, 8);
}
#define init_5_9_14_9
void FASTCALL call_5_9_14_9(SH2_struct* sh)
{
  SH2movll4(sh, 9,14, 9);
}
#define init_5_9_14_10
void FASTCALL call_5_9_14_10(SH2_struct* sh)
{
  SH2movll4(sh, 9,14, 10);
}
#define init_5_9_14_11
void FASTCALL call_5_9_14_11(SH2_struct* sh)
{
  SH2movll4(sh, 9,14, 11);
}
#define init_5_9_14_12
void FASTCALL call_5_9_14_12(SH2_struct* sh)
{
  SH2movll4(sh, 9,14, 12);
}
#define init_5_9_14_13
void FASTCALL call_5_9_14_13(SH2_struct* sh)
{
  SH2movll4(sh, 9,14, 13);
}
#define init_5_9_14_14
void FASTCALL call_5_9_14_14(SH2_struct* sh)
{
  SH2movll4(sh, 9,14, 14);
}
#define init_5_9_14_15
void FASTCALL call_5_9_14_15(SH2_struct* sh)
{
  SH2movll4(sh, 9,14, 15);
}
#define init_5_9_15_0
void FASTCALL call_5_9_15_0(SH2_struct* sh)
{
  SH2movll4(sh, 9,15, 0);
}
#define init_5_9_15_1
void FASTCALL call_5_9_15_1(SH2_struct* sh)
{
  SH2movll4(sh, 9,15, 1);
}
#define init_5_9_15_2
void FASTCALL call_5_9_15_2(SH2_struct* sh)
{
  SH2movll4(sh, 9,15, 2);
}
#define init_5_9_15_3
void FASTCALL call_5_9_15_3(SH2_struct* sh)
{
  SH2movll4(sh, 9,15, 3);
}
#define init_5_9_15_4
void FASTCALL call_5_9_15_4(SH2_struct* sh)
{
  SH2movll4(sh, 9,15, 4);
}
#define init_5_9_15_5
void FASTCALL call_5_9_15_5(SH2_struct* sh)
{
  SH2movll4(sh, 9,15, 5);
}
#define init_5_9_15_6
void FASTCALL call_5_9_15_6(SH2_struct* sh)
{
  SH2movll4(sh, 9,15, 6);
}
#define init_5_9_15_7
void FASTCALL call_5_9_15_7(SH2_struct* sh)
{
  SH2movll4(sh, 9,15, 7);
}
#define init_5_9_15_8
void FASTCALL call_5_9_15_8(SH2_struct* sh)
{
  SH2movll4(sh, 9,15, 8);
}
#define init_5_9_15_9
void FASTCALL call_5_9_15_9(SH2_struct* sh)
{
  SH2movll4(sh, 9,15, 9);
}
#define init_5_9_15_10
void FASTCALL call_5_9_15_10(SH2_struct* sh)
{
  SH2movll4(sh, 9,15, 10);
}
#define init_5_9_15_11
void FASTCALL call_5_9_15_11(SH2_struct* sh)
{
  SH2movll4(sh, 9,15, 11);
}
#define init_5_9_15_12
void FASTCALL call_5_9_15_12(SH2_struct* sh)
{
  SH2movll4(sh, 9,15, 12);
}
#define init_5_9_15_13
void FASTCALL call_5_9_15_13(SH2_struct* sh)
{
  SH2movll4(sh, 9,15, 13);
}
#define init_5_9_15_14
void FASTCALL call_5_9_15_14(SH2_struct* sh)
{
  SH2movll4(sh, 9,15, 14);
}
#define init_5_9_15_15
void FASTCALL call_5_9_15_15(SH2_struct* sh)
{
  SH2movll4(sh, 9,15, 15);
}
#define init_5_10_0_0
void FASTCALL call_5_10_0_0(SH2_struct* sh)
{
  SH2movll4(sh, 10,0, 0);
}
#define init_5_10_0_1
void FASTCALL call_5_10_0_1(SH2_struct* sh)
{
  SH2movll4(sh, 10,0, 1);
}
#define init_5_10_0_2
void FASTCALL call_5_10_0_2(SH2_struct* sh)
{
  SH2movll4(sh, 10,0, 2);
}
#define init_5_10_0_3
void FASTCALL call_5_10_0_3(SH2_struct* sh)
{
  SH2movll4(sh, 10,0, 3);
}
#define init_5_10_0_4
void FASTCALL call_5_10_0_4(SH2_struct* sh)
{
  SH2movll4(sh, 10,0, 4);
}
#define init_5_10_0_5
void FASTCALL call_5_10_0_5(SH2_struct* sh)
{
  SH2movll4(sh, 10,0, 5);
}
#define init_5_10_0_6
void FASTCALL call_5_10_0_6(SH2_struct* sh)
{
  SH2movll4(sh, 10,0, 6);
}
#define init_5_10_0_7
void FASTCALL call_5_10_0_7(SH2_struct* sh)
{
  SH2movll4(sh, 10,0, 7);
}
#define init_5_10_0_8
void FASTCALL call_5_10_0_8(SH2_struct* sh)
{
  SH2movll4(sh, 10,0, 8);
}
#define init_5_10_0_9
void FASTCALL call_5_10_0_9(SH2_struct* sh)
{
  SH2movll4(sh, 10,0, 9);
}
#define init_5_10_0_10
void FASTCALL call_5_10_0_10(SH2_struct* sh)
{
  SH2movll4(sh, 10,0, 10);
}
#define init_5_10_0_11
void FASTCALL call_5_10_0_11(SH2_struct* sh)
{
  SH2movll4(sh, 10,0, 11);
}
#define init_5_10_0_12
void FASTCALL call_5_10_0_12(SH2_struct* sh)
{
  SH2movll4(sh, 10,0, 12);
}
#define init_5_10_0_13
void FASTCALL call_5_10_0_13(SH2_struct* sh)
{
  SH2movll4(sh, 10,0, 13);
}
#define init_5_10_0_14
void FASTCALL call_5_10_0_14(SH2_struct* sh)
{
  SH2movll4(sh, 10,0, 14);
}
#define init_5_10_0_15
void FASTCALL call_5_10_0_15(SH2_struct* sh)
{
  SH2movll4(sh, 10,0, 15);
}
#define init_5_10_1_0
void FASTCALL call_5_10_1_0(SH2_struct* sh)
{
  SH2movll4(sh, 10,1, 0);
}
#define init_5_10_1_1
void FASTCALL call_5_10_1_1(SH2_struct* sh)
{
  SH2movll4(sh, 10,1, 1);
}
#define init_5_10_1_2
void FASTCALL call_5_10_1_2(SH2_struct* sh)
{
  SH2movll4(sh, 10,1, 2);
}
#define init_5_10_1_3
void FASTCALL call_5_10_1_3(SH2_struct* sh)
{
  SH2movll4(sh, 10,1, 3);
}
#define init_5_10_1_4
void FASTCALL call_5_10_1_4(SH2_struct* sh)
{
  SH2movll4(sh, 10,1, 4);
}
#define init_5_10_1_5
void FASTCALL call_5_10_1_5(SH2_struct* sh)
{
  SH2movll4(sh, 10,1, 5);
}
#define init_5_10_1_6
void FASTCALL call_5_10_1_6(SH2_struct* sh)
{
  SH2movll4(sh, 10,1, 6);
}
#define init_5_10_1_7
void FASTCALL call_5_10_1_7(SH2_struct* sh)
{
  SH2movll4(sh, 10,1, 7);
}
#define init_5_10_1_8
void FASTCALL call_5_10_1_8(SH2_struct* sh)
{
  SH2movll4(sh, 10,1, 8);
}
#define init_5_10_1_9
void FASTCALL call_5_10_1_9(SH2_struct* sh)
{
  SH2movll4(sh, 10,1, 9);
}
#define init_5_10_1_10
void FASTCALL call_5_10_1_10(SH2_struct* sh)
{
  SH2movll4(sh, 10,1, 10);
}
#define init_5_10_1_11
void FASTCALL call_5_10_1_11(SH2_struct* sh)
{
  SH2movll4(sh, 10,1, 11);
}
#define init_5_10_1_12
void FASTCALL call_5_10_1_12(SH2_struct* sh)
{
  SH2movll4(sh, 10,1, 12);
}
#define init_5_10_1_13
void FASTCALL call_5_10_1_13(SH2_struct* sh)
{
  SH2movll4(sh, 10,1, 13);
}
#define init_5_10_1_14
void FASTCALL call_5_10_1_14(SH2_struct* sh)
{
  SH2movll4(sh, 10,1, 14);
}
#define init_5_10_1_15
void FASTCALL call_5_10_1_15(SH2_struct* sh)
{
  SH2movll4(sh, 10,1, 15);
}
#define init_5_10_2_0
void FASTCALL call_5_10_2_0(SH2_struct* sh)
{
  SH2movll4(sh, 10,2, 0);
}
#define init_5_10_2_1
void FASTCALL call_5_10_2_1(SH2_struct* sh)
{
  SH2movll4(sh, 10,2, 1);
}
#define init_5_10_2_2
void FASTCALL call_5_10_2_2(SH2_struct* sh)
{
  SH2movll4(sh, 10,2, 2);
}
#define init_5_10_2_3
void FASTCALL call_5_10_2_3(SH2_struct* sh)
{
  SH2movll4(sh, 10,2, 3);
}
#define init_5_10_2_4
void FASTCALL call_5_10_2_4(SH2_struct* sh)
{
  SH2movll4(sh, 10,2, 4);
}
#define init_5_10_2_5
void FASTCALL call_5_10_2_5(SH2_struct* sh)
{
  SH2movll4(sh, 10,2, 5);
}
#define init_5_10_2_6
void FASTCALL call_5_10_2_6(SH2_struct* sh)
{
  SH2movll4(sh, 10,2, 6);
}
#define init_5_10_2_7
void FASTCALL call_5_10_2_7(SH2_struct* sh)
{
  SH2movll4(sh, 10,2, 7);
}
#define init_5_10_2_8
void FASTCALL call_5_10_2_8(SH2_struct* sh)
{
  SH2movll4(sh, 10,2, 8);
}
#define init_5_10_2_9
void FASTCALL call_5_10_2_9(SH2_struct* sh)
{
  SH2movll4(sh, 10,2, 9);
}
#define init_5_10_2_10
void FASTCALL call_5_10_2_10(SH2_struct* sh)
{
  SH2movll4(sh, 10,2, 10);
}
#define init_5_10_2_11
void FASTCALL call_5_10_2_11(SH2_struct* sh)
{
  SH2movll4(sh, 10,2, 11);
}
#define init_5_10_2_12
void FASTCALL call_5_10_2_12(SH2_struct* sh)
{
  SH2movll4(sh, 10,2, 12);
}
#define init_5_10_2_13
void FASTCALL call_5_10_2_13(SH2_struct* sh)
{
  SH2movll4(sh, 10,2, 13);
}
#define init_5_10_2_14
void FASTCALL call_5_10_2_14(SH2_struct* sh)
{
  SH2movll4(sh, 10,2, 14);
}
#define init_5_10_2_15
void FASTCALL call_5_10_2_15(SH2_struct* sh)
{
  SH2movll4(sh, 10,2, 15);
}
#define init_5_10_3_0
void FASTCALL call_5_10_3_0(SH2_struct* sh)
{
  SH2movll4(sh, 10,3, 0);
}
#define init_5_10_3_1
void FASTCALL call_5_10_3_1(SH2_struct* sh)
{
  SH2movll4(sh, 10,3, 1);
}
#define init_5_10_3_2
void FASTCALL call_5_10_3_2(SH2_struct* sh)
{
  SH2movll4(sh, 10,3, 2);
}
#define init_5_10_3_3
void FASTCALL call_5_10_3_3(SH2_struct* sh)
{
  SH2movll4(sh, 10,3, 3);
}
#define init_5_10_3_4
void FASTCALL call_5_10_3_4(SH2_struct* sh)
{
  SH2movll4(sh, 10,3, 4);
}
#define init_5_10_3_5
void FASTCALL call_5_10_3_5(SH2_struct* sh)
{
  SH2movll4(sh, 10,3, 5);
}
#define init_5_10_3_6
void FASTCALL call_5_10_3_6(SH2_struct* sh)
{
  SH2movll4(sh, 10,3, 6);
}
#define init_5_10_3_7
void FASTCALL call_5_10_3_7(SH2_struct* sh)
{
  SH2movll4(sh, 10,3, 7);
}
#define init_5_10_3_8
void FASTCALL call_5_10_3_8(SH2_struct* sh)
{
  SH2movll4(sh, 10,3, 8);
}
#define init_5_10_3_9
void FASTCALL call_5_10_3_9(SH2_struct* sh)
{
  SH2movll4(sh, 10,3, 9);
}
#define init_5_10_3_10
void FASTCALL call_5_10_3_10(SH2_struct* sh)
{
  SH2movll4(sh, 10,3, 10);
}
#define init_5_10_3_11
void FASTCALL call_5_10_3_11(SH2_struct* sh)
{
  SH2movll4(sh, 10,3, 11);
}
#define init_5_10_3_12
void FASTCALL call_5_10_3_12(SH2_struct* sh)
{
  SH2movll4(sh, 10,3, 12);
}
#define init_5_10_3_13
void FASTCALL call_5_10_3_13(SH2_struct* sh)
{
  SH2movll4(sh, 10,3, 13);
}
#define init_5_10_3_14
void FASTCALL call_5_10_3_14(SH2_struct* sh)
{
  SH2movll4(sh, 10,3, 14);
}
#define init_5_10_3_15
void FASTCALL call_5_10_3_15(SH2_struct* sh)
{
  SH2movll4(sh, 10,3, 15);
}
#define init_5_10_4_0
void FASTCALL call_5_10_4_0(SH2_struct* sh)
{
  SH2movll4(sh, 10,4, 0);
}
#define init_5_10_4_1
void FASTCALL call_5_10_4_1(SH2_struct* sh)
{
  SH2movll4(sh, 10,4, 1);
}
#define init_5_10_4_2
void FASTCALL call_5_10_4_2(SH2_struct* sh)
{
  SH2movll4(sh, 10,4, 2);
}
#define init_5_10_4_3
void FASTCALL call_5_10_4_3(SH2_struct* sh)
{
  SH2movll4(sh, 10,4, 3);
}
#define init_5_10_4_4
void FASTCALL call_5_10_4_4(SH2_struct* sh)
{
  SH2movll4(sh, 10,4, 4);
}
#define init_5_10_4_5
void FASTCALL call_5_10_4_5(SH2_struct* sh)
{
  SH2movll4(sh, 10,4, 5);
}
#define init_5_10_4_6
void FASTCALL call_5_10_4_6(SH2_struct* sh)
{
  SH2movll4(sh, 10,4, 6);
}
#define init_5_10_4_7
void FASTCALL call_5_10_4_7(SH2_struct* sh)
{
  SH2movll4(sh, 10,4, 7);
}
#define init_5_10_4_8
void FASTCALL call_5_10_4_8(SH2_struct* sh)
{
  SH2movll4(sh, 10,4, 8);
}
#define init_5_10_4_9
void FASTCALL call_5_10_4_9(SH2_struct* sh)
{
  SH2movll4(sh, 10,4, 9);
}
#define init_5_10_4_10
void FASTCALL call_5_10_4_10(SH2_struct* sh)
{
  SH2movll4(sh, 10,4, 10);
}
#define init_5_10_4_11
void FASTCALL call_5_10_4_11(SH2_struct* sh)
{
  SH2movll4(sh, 10,4, 11);
}
#define init_5_10_4_12
void FASTCALL call_5_10_4_12(SH2_struct* sh)
{
  SH2movll4(sh, 10,4, 12);
}
#define init_5_10_4_13
void FASTCALL call_5_10_4_13(SH2_struct* sh)
{
  SH2movll4(sh, 10,4, 13);
}
#define init_5_10_4_14
void FASTCALL call_5_10_4_14(SH2_struct* sh)
{
  SH2movll4(sh, 10,4, 14);
}
#define init_5_10_4_15
void FASTCALL call_5_10_4_15(SH2_struct* sh)
{
  SH2movll4(sh, 10,4, 15);
}
#define init_5_10_5_0
void FASTCALL call_5_10_5_0(SH2_struct* sh)
{
  SH2movll4(sh, 10,5, 0);
}
#define init_5_10_5_1
void FASTCALL call_5_10_5_1(SH2_struct* sh)
{
  SH2movll4(sh, 10,5, 1);
}
#define init_5_10_5_2
void FASTCALL call_5_10_5_2(SH2_struct* sh)
{
  SH2movll4(sh, 10,5, 2);
}
#define init_5_10_5_3
void FASTCALL call_5_10_5_3(SH2_struct* sh)
{
  SH2movll4(sh, 10,5, 3);
}
#define init_5_10_5_4
void FASTCALL call_5_10_5_4(SH2_struct* sh)
{
  SH2movll4(sh, 10,5, 4);
}
#define init_5_10_5_5
void FASTCALL call_5_10_5_5(SH2_struct* sh)
{
  SH2movll4(sh, 10,5, 5);
}
#define init_5_10_5_6
void FASTCALL call_5_10_5_6(SH2_struct* sh)
{
  SH2movll4(sh, 10,5, 6);
}
#define init_5_10_5_7
void FASTCALL call_5_10_5_7(SH2_struct* sh)
{
  SH2movll4(sh, 10,5, 7);
}
#define init_5_10_5_8
void FASTCALL call_5_10_5_8(SH2_struct* sh)
{
  SH2movll4(sh, 10,5, 8);
}
#define init_5_10_5_9
void FASTCALL call_5_10_5_9(SH2_struct* sh)
{
  SH2movll4(sh, 10,5, 9);
}
#define init_5_10_5_10
void FASTCALL call_5_10_5_10(SH2_struct* sh)
{
  SH2movll4(sh, 10,5, 10);
}
#define init_5_10_5_11
void FASTCALL call_5_10_5_11(SH2_struct* sh)
{
  SH2movll4(sh, 10,5, 11);
}
#define init_5_10_5_12
void FASTCALL call_5_10_5_12(SH2_struct* sh)
{
  SH2movll4(sh, 10,5, 12);
}
#define init_5_10_5_13
void FASTCALL call_5_10_5_13(SH2_struct* sh)
{
  SH2movll4(sh, 10,5, 13);
}
#define init_5_10_5_14
void FASTCALL call_5_10_5_14(SH2_struct* sh)
{
  SH2movll4(sh, 10,5, 14);
}
#define init_5_10_5_15
void FASTCALL call_5_10_5_15(SH2_struct* sh)
{
  SH2movll4(sh, 10,5, 15);
}
#define init_5_10_6_0
void FASTCALL call_5_10_6_0(SH2_struct* sh)
{
  SH2movll4(sh, 10,6, 0);
}
#define init_5_10_6_1
void FASTCALL call_5_10_6_1(SH2_struct* sh)
{
  SH2movll4(sh, 10,6, 1);
}
#define init_5_10_6_2
void FASTCALL call_5_10_6_2(SH2_struct* sh)
{
  SH2movll4(sh, 10,6, 2);
}
#define init_5_10_6_3
void FASTCALL call_5_10_6_3(SH2_struct* sh)
{
  SH2movll4(sh, 10,6, 3);
}
#define init_5_10_6_4
void FASTCALL call_5_10_6_4(SH2_struct* sh)
{
  SH2movll4(sh, 10,6, 4);
}
#define init_5_10_6_5
void FASTCALL call_5_10_6_5(SH2_struct* sh)
{
  SH2movll4(sh, 10,6, 5);
}
#define init_5_10_6_6
void FASTCALL call_5_10_6_6(SH2_struct* sh)
{
  SH2movll4(sh, 10,6, 6);
}
#define init_5_10_6_7
void FASTCALL call_5_10_6_7(SH2_struct* sh)
{
  SH2movll4(sh, 10,6, 7);
}
#define init_5_10_6_8
void FASTCALL call_5_10_6_8(SH2_struct* sh)
{
  SH2movll4(sh, 10,6, 8);
}
#define init_5_10_6_9
void FASTCALL call_5_10_6_9(SH2_struct* sh)
{
  SH2movll4(sh, 10,6, 9);
}
#define init_5_10_6_10
void FASTCALL call_5_10_6_10(SH2_struct* sh)
{
  SH2movll4(sh, 10,6, 10);
}
#define init_5_10_6_11
void FASTCALL call_5_10_6_11(SH2_struct* sh)
{
  SH2movll4(sh, 10,6, 11);
}
#define init_5_10_6_12
void FASTCALL call_5_10_6_12(SH2_struct* sh)
{
  SH2movll4(sh, 10,6, 12);
}
#define init_5_10_6_13
void FASTCALL call_5_10_6_13(SH2_struct* sh)
{
  SH2movll4(sh, 10,6, 13);
}
#define init_5_10_6_14
void FASTCALL call_5_10_6_14(SH2_struct* sh)
{
  SH2movll4(sh, 10,6, 14);
}
#define init_5_10_6_15
void FASTCALL call_5_10_6_15(SH2_struct* sh)
{
  SH2movll4(sh, 10,6, 15);
}
#define init_5_10_7_0
void FASTCALL call_5_10_7_0(SH2_struct* sh)
{
  SH2movll4(sh, 10,7, 0);
}
#define init_5_10_7_1
void FASTCALL call_5_10_7_1(SH2_struct* sh)
{
  SH2movll4(sh, 10,7, 1);
}
#define init_5_10_7_2
void FASTCALL call_5_10_7_2(SH2_struct* sh)
{
  SH2movll4(sh, 10,7, 2);
}
#define init_5_10_7_3
void FASTCALL call_5_10_7_3(SH2_struct* sh)
{
  SH2movll4(sh, 10,7, 3);
}
#define init_5_10_7_4
void FASTCALL call_5_10_7_4(SH2_struct* sh)
{
  SH2movll4(sh, 10,7, 4);
}
#define init_5_10_7_5
void FASTCALL call_5_10_7_5(SH2_struct* sh)
{
  SH2movll4(sh, 10,7, 5);
}
#define init_5_10_7_6
void FASTCALL call_5_10_7_6(SH2_struct* sh)
{
  SH2movll4(sh, 10,7, 6);
}
#define init_5_10_7_7
void FASTCALL call_5_10_7_7(SH2_struct* sh)
{
  SH2movll4(sh, 10,7, 7);
}
#define init_5_10_7_8
void FASTCALL call_5_10_7_8(SH2_struct* sh)
{
  SH2movll4(sh, 10,7, 8);
}
#define init_5_10_7_9
void FASTCALL call_5_10_7_9(SH2_struct* sh)
{
  SH2movll4(sh, 10,7, 9);
}
#define init_5_10_7_10
void FASTCALL call_5_10_7_10(SH2_struct* sh)
{
  SH2movll4(sh, 10,7, 10);
}
#define init_5_10_7_11
void FASTCALL call_5_10_7_11(SH2_struct* sh)
{
  SH2movll4(sh, 10,7, 11);
}
#define init_5_10_7_12
void FASTCALL call_5_10_7_12(SH2_struct* sh)
{
  SH2movll4(sh, 10,7, 12);
}
#define init_5_10_7_13
void FASTCALL call_5_10_7_13(SH2_struct* sh)
{
  SH2movll4(sh, 10,7, 13);
}
#define init_5_10_7_14
void FASTCALL call_5_10_7_14(SH2_struct* sh)
{
  SH2movll4(sh, 10,7, 14);
}
#define init_5_10_7_15
void FASTCALL call_5_10_7_15(SH2_struct* sh)
{
  SH2movll4(sh, 10,7, 15);
}
#define init_5_10_8_0
void FASTCALL call_5_10_8_0(SH2_struct* sh)
{
  SH2movll4(sh, 10,8, 0);
}
#define init_5_10_8_1
void FASTCALL call_5_10_8_1(SH2_struct* sh)
{
  SH2movll4(sh, 10,8, 1);
}
#define init_5_10_8_2
void FASTCALL call_5_10_8_2(SH2_struct* sh)
{
  SH2movll4(sh, 10,8, 2);
}
#define init_5_10_8_3
void FASTCALL call_5_10_8_3(SH2_struct* sh)
{
  SH2movll4(sh, 10,8, 3);
}
#define init_5_10_8_4
void FASTCALL call_5_10_8_4(SH2_struct* sh)
{
  SH2movll4(sh, 10,8, 4);
}
#define init_5_10_8_5
void FASTCALL call_5_10_8_5(SH2_struct* sh)
{
  SH2movll4(sh, 10,8, 5);
}
#define init_5_10_8_6
void FASTCALL call_5_10_8_6(SH2_struct* sh)
{
  SH2movll4(sh, 10,8, 6);
}
#define init_5_10_8_7
void FASTCALL call_5_10_8_7(SH2_struct* sh)
{
  SH2movll4(sh, 10,8, 7);
}
#define init_5_10_8_8
void FASTCALL call_5_10_8_8(SH2_struct* sh)
{
  SH2movll4(sh, 10,8, 8);
}
#define init_5_10_8_9
void FASTCALL call_5_10_8_9(SH2_struct* sh)
{
  SH2movll4(sh, 10,8, 9);
}
#define init_5_10_8_10
void FASTCALL call_5_10_8_10(SH2_struct* sh)
{
  SH2movll4(sh, 10,8, 10);
}
#define init_5_10_8_11
void FASTCALL call_5_10_8_11(SH2_struct* sh)
{
  SH2movll4(sh, 10,8, 11);
}
#define init_5_10_8_12
void FASTCALL call_5_10_8_12(SH2_struct* sh)
{
  SH2movll4(sh, 10,8, 12);
}
#define init_5_10_8_13
void FASTCALL call_5_10_8_13(SH2_struct* sh)
{
  SH2movll4(sh, 10,8, 13);
}
#define init_5_10_8_14
void FASTCALL call_5_10_8_14(SH2_struct* sh)
{
  SH2movll4(sh, 10,8, 14);
}
#define init_5_10_8_15
void FASTCALL call_5_10_8_15(SH2_struct* sh)
{
  SH2movll4(sh, 10,8, 15);
}
#define init_5_10_9_0
void FASTCALL call_5_10_9_0(SH2_struct* sh)
{
  SH2movll4(sh, 10,9, 0);
}
#define init_5_10_9_1
void FASTCALL call_5_10_9_1(SH2_struct* sh)
{
  SH2movll4(sh, 10,9, 1);
}
#define init_5_10_9_2
void FASTCALL call_5_10_9_2(SH2_struct* sh)
{
  SH2movll4(sh, 10,9, 2);
}
#define init_5_10_9_3
void FASTCALL call_5_10_9_3(SH2_struct* sh)
{
  SH2movll4(sh, 10,9, 3);
}
#define init_5_10_9_4
void FASTCALL call_5_10_9_4(SH2_struct* sh)
{
  SH2movll4(sh, 10,9, 4);
}
#define init_5_10_9_5
void FASTCALL call_5_10_9_5(SH2_struct* sh)
{
  SH2movll4(sh, 10,9, 5);
}
#define init_5_10_9_6
void FASTCALL call_5_10_9_6(SH2_struct* sh)
{
  SH2movll4(sh, 10,9, 6);
}
#define init_5_10_9_7
void FASTCALL call_5_10_9_7(SH2_struct* sh)
{
  SH2movll4(sh, 10,9, 7);
}
#define init_5_10_9_8
void FASTCALL call_5_10_9_8(SH2_struct* sh)
{
  SH2movll4(sh, 10,9, 8);
}
#define init_5_10_9_9
void FASTCALL call_5_10_9_9(SH2_struct* sh)
{
  SH2movll4(sh, 10,9, 9);
}
#define init_5_10_9_10
void FASTCALL call_5_10_9_10(SH2_struct* sh)
{
  SH2movll4(sh, 10,9, 10);
}
#define init_5_10_9_11
void FASTCALL call_5_10_9_11(SH2_struct* sh)
{
  SH2movll4(sh, 10,9, 11);
}
#define init_5_10_9_12
void FASTCALL call_5_10_9_12(SH2_struct* sh)
{
  SH2movll4(sh, 10,9, 12);
}
#define init_5_10_9_13
void FASTCALL call_5_10_9_13(SH2_struct* sh)
{
  SH2movll4(sh, 10,9, 13);
}
#define init_5_10_9_14
void FASTCALL call_5_10_9_14(SH2_struct* sh)
{
  SH2movll4(sh, 10,9, 14);
}
#define init_5_10_9_15
void FASTCALL call_5_10_9_15(SH2_struct* sh)
{
  SH2movll4(sh, 10,9, 15);
}
#define init_5_10_10_0
void FASTCALL call_5_10_10_0(SH2_struct* sh)
{
  SH2movll4(sh, 10,10, 0);
}
#define init_5_10_10_1
void FASTCALL call_5_10_10_1(SH2_struct* sh)
{
  SH2movll4(sh, 10,10, 1);
}
#define init_5_10_10_2
void FASTCALL call_5_10_10_2(SH2_struct* sh)
{
  SH2movll4(sh, 10,10, 2);
}
#define init_5_10_10_3
void FASTCALL call_5_10_10_3(SH2_struct* sh)
{
  SH2movll4(sh, 10,10, 3);
}
#define init_5_10_10_4
void FASTCALL call_5_10_10_4(SH2_struct* sh)
{
  SH2movll4(sh, 10,10, 4);
}
#define init_5_10_10_5
void FASTCALL call_5_10_10_5(SH2_struct* sh)
{
  SH2movll4(sh, 10,10, 5);
}
#define init_5_10_10_6
void FASTCALL call_5_10_10_6(SH2_struct* sh)
{
  SH2movll4(sh, 10,10, 6);
}
#define init_5_10_10_7
void FASTCALL call_5_10_10_7(SH2_struct* sh)
{
  SH2movll4(sh, 10,10, 7);
}
#define init_5_10_10_8
void FASTCALL call_5_10_10_8(SH2_struct* sh)
{
  SH2movll4(sh, 10,10, 8);
}
#define init_5_10_10_9
void FASTCALL call_5_10_10_9(SH2_struct* sh)
{
  SH2movll4(sh, 10,10, 9);
}
#define init_5_10_10_10
void FASTCALL call_5_10_10_10(SH2_struct* sh)
{
  SH2movll4(sh, 10,10, 10);
}
#define init_5_10_10_11
void FASTCALL call_5_10_10_11(SH2_struct* sh)
{
  SH2movll4(sh, 10,10, 11);
}
#define init_5_10_10_12
void FASTCALL call_5_10_10_12(SH2_struct* sh)
{
  SH2movll4(sh, 10,10, 12);
}
#define init_5_10_10_13
void FASTCALL call_5_10_10_13(SH2_struct* sh)
{
  SH2movll4(sh, 10,10, 13);
}
#define init_5_10_10_14
void FASTCALL call_5_10_10_14(SH2_struct* sh)
{
  SH2movll4(sh, 10,10, 14);
}
#define init_5_10_10_15
void FASTCALL call_5_10_10_15(SH2_struct* sh)
{
  SH2movll4(sh, 10,10, 15);
}
#define init_5_10_11_0
void FASTCALL call_5_10_11_0(SH2_struct* sh)
{
  SH2movll4(sh, 10,11, 0);
}
#define init_5_10_11_1
void FASTCALL call_5_10_11_1(SH2_struct* sh)
{
  SH2movll4(sh, 10,11, 1);
}
#define init_5_10_11_2
void FASTCALL call_5_10_11_2(SH2_struct* sh)
{
  SH2movll4(sh, 10,11, 2);
}
#define init_5_10_11_3
void FASTCALL call_5_10_11_3(SH2_struct* sh)
{
  SH2movll4(sh, 10,11, 3);
}
#define init_5_10_11_4
void FASTCALL call_5_10_11_4(SH2_struct* sh)
{
  SH2movll4(sh, 10,11, 4);
}
#define init_5_10_11_5
void FASTCALL call_5_10_11_5(SH2_struct* sh)
{
  SH2movll4(sh, 10,11, 5);
}
#define init_5_10_11_6
void FASTCALL call_5_10_11_6(SH2_struct* sh)
{
  SH2movll4(sh, 10,11, 6);
}
#define init_5_10_11_7
void FASTCALL call_5_10_11_7(SH2_struct* sh)
{
  SH2movll4(sh, 10,11, 7);
}
#define init_5_10_11_8
void FASTCALL call_5_10_11_8(SH2_struct* sh)
{
  SH2movll4(sh, 10,11, 8);
}
#define init_5_10_11_9
void FASTCALL call_5_10_11_9(SH2_struct* sh)
{
  SH2movll4(sh, 10,11, 9);
}
#define init_5_10_11_10
void FASTCALL call_5_10_11_10(SH2_struct* sh)
{
  SH2movll4(sh, 10,11, 10);
}
#define init_5_10_11_11
void FASTCALL call_5_10_11_11(SH2_struct* sh)
{
  SH2movll4(sh, 10,11, 11);
}
#define init_5_10_11_12
void FASTCALL call_5_10_11_12(SH2_struct* sh)
{
  SH2movll4(sh, 10,11, 12);
}
#define init_5_10_11_13
void FASTCALL call_5_10_11_13(SH2_struct* sh)
{
  SH2movll4(sh, 10,11, 13);
}
#define init_5_10_11_14
void FASTCALL call_5_10_11_14(SH2_struct* sh)
{
  SH2movll4(sh, 10,11, 14);
}
#define init_5_10_11_15
void FASTCALL call_5_10_11_15(SH2_struct* sh)
{
  SH2movll4(sh, 10,11, 15);
}
#define init_5_10_12_0
void FASTCALL call_5_10_12_0(SH2_struct* sh)
{
  SH2movll4(sh, 10,12, 0);
}
#define init_5_10_12_1
void FASTCALL call_5_10_12_1(SH2_struct* sh)
{
  SH2movll4(sh, 10,12, 1);
}
#define init_5_10_12_2
void FASTCALL call_5_10_12_2(SH2_struct* sh)
{
  SH2movll4(sh, 10,12, 2);
}
#define init_5_10_12_3
void FASTCALL call_5_10_12_3(SH2_struct* sh)
{
  SH2movll4(sh, 10,12, 3);
}
#define init_5_10_12_4
void FASTCALL call_5_10_12_4(SH2_struct* sh)
{
  SH2movll4(sh, 10,12, 4);
}
#define init_5_10_12_5
void FASTCALL call_5_10_12_5(SH2_struct* sh)
{
  SH2movll4(sh, 10,12, 5);
}
#define init_5_10_12_6
void FASTCALL call_5_10_12_6(SH2_struct* sh)
{
  SH2movll4(sh, 10,12, 6);
}
#define init_5_10_12_7
void FASTCALL call_5_10_12_7(SH2_struct* sh)
{
  SH2movll4(sh, 10,12, 7);
}
#define init_5_10_12_8
void FASTCALL call_5_10_12_8(SH2_struct* sh)
{
  SH2movll4(sh, 10,12, 8);
}
#define init_5_10_12_9
void FASTCALL call_5_10_12_9(SH2_struct* sh)
{
  SH2movll4(sh, 10,12, 9);
}
#define init_5_10_12_10
void FASTCALL call_5_10_12_10(SH2_struct* sh)
{
  SH2movll4(sh, 10,12, 10);
}
#define init_5_10_12_11
void FASTCALL call_5_10_12_11(SH2_struct* sh)
{
  SH2movll4(sh, 10,12, 11);
}
#define init_5_10_12_12
void FASTCALL call_5_10_12_12(SH2_struct* sh)
{
  SH2movll4(sh, 10,12, 12);
}
#define init_5_10_12_13
void FASTCALL call_5_10_12_13(SH2_struct* sh)
{
  SH2movll4(sh, 10,12, 13);
}
#define init_5_10_12_14
void FASTCALL call_5_10_12_14(SH2_struct* sh)
{
  SH2movll4(sh, 10,12, 14);
}
#define init_5_10_12_15
void FASTCALL call_5_10_12_15(SH2_struct* sh)
{
  SH2movll4(sh, 10,12, 15);
}
#define init_5_10_13_0
void FASTCALL call_5_10_13_0(SH2_struct* sh)
{
  SH2movll4(sh, 10,13, 0);
}
#define init_5_10_13_1
void FASTCALL call_5_10_13_1(SH2_struct* sh)
{
  SH2movll4(sh, 10,13, 1);
}
#define init_5_10_13_2
void FASTCALL call_5_10_13_2(SH2_struct* sh)
{
  SH2movll4(sh, 10,13, 2);
}
#define init_5_10_13_3
void FASTCALL call_5_10_13_3(SH2_struct* sh)
{
  SH2movll4(sh, 10,13, 3);
}
#define init_5_10_13_4
void FASTCALL call_5_10_13_4(SH2_struct* sh)
{
  SH2movll4(sh, 10,13, 4);
}
#define init_5_10_13_5
void FASTCALL call_5_10_13_5(SH2_struct* sh)
{
  SH2movll4(sh, 10,13, 5);
}
#define init_5_10_13_6
void FASTCALL call_5_10_13_6(SH2_struct* sh)
{
  SH2movll4(sh, 10,13, 6);
}
#define init_5_10_13_7
void FASTCALL call_5_10_13_7(SH2_struct* sh)
{
  SH2movll4(sh, 10,13, 7);
}
#define init_5_10_13_8
void FASTCALL call_5_10_13_8(SH2_struct* sh)
{
  SH2movll4(sh, 10,13, 8);
}
#define init_5_10_13_9
void FASTCALL call_5_10_13_9(SH2_struct* sh)
{
  SH2movll4(sh, 10,13, 9);
}
#define init_5_10_13_10
void FASTCALL call_5_10_13_10(SH2_struct* sh)
{
  SH2movll4(sh, 10,13, 10);
}
#define init_5_10_13_11
void FASTCALL call_5_10_13_11(SH2_struct* sh)
{
  SH2movll4(sh, 10,13, 11);
}
#define init_5_10_13_12
void FASTCALL call_5_10_13_12(SH2_struct* sh)
{
  SH2movll4(sh, 10,13, 12);
}
#define init_5_10_13_13
void FASTCALL call_5_10_13_13(SH2_struct* sh)
{
  SH2movll4(sh, 10,13, 13);
}
#define init_5_10_13_14
void FASTCALL call_5_10_13_14(SH2_struct* sh)
{
  SH2movll4(sh, 10,13, 14);
}
#define init_5_10_13_15
void FASTCALL call_5_10_13_15(SH2_struct* sh)
{
  SH2movll4(sh, 10,13, 15);
}
#define init_5_10_14_0
void FASTCALL call_5_10_14_0(SH2_struct* sh)
{
  SH2movll4(sh, 10,14, 0);
}
#define init_5_10_14_1
void FASTCALL call_5_10_14_1(SH2_struct* sh)
{
  SH2movll4(sh, 10,14, 1);
}
#define init_5_10_14_2
void FASTCALL call_5_10_14_2(SH2_struct* sh)
{
  SH2movll4(sh, 10,14, 2);
}
#define init_5_10_14_3
void FASTCALL call_5_10_14_3(SH2_struct* sh)
{
  SH2movll4(sh, 10,14, 3);
}
#define init_5_10_14_4
void FASTCALL call_5_10_14_4(SH2_struct* sh)
{
  SH2movll4(sh, 10,14, 4);
}
#define init_5_10_14_5
void FASTCALL call_5_10_14_5(SH2_struct* sh)
{
  SH2movll4(sh, 10,14, 5);
}
#define init_5_10_14_6
void FASTCALL call_5_10_14_6(SH2_struct* sh)
{
  SH2movll4(sh, 10,14, 6);
}
#define init_5_10_14_7
void FASTCALL call_5_10_14_7(SH2_struct* sh)
{
  SH2movll4(sh, 10,14, 7);
}
#define init_5_10_14_8
void FASTCALL call_5_10_14_8(SH2_struct* sh)
{
  SH2movll4(sh, 10,14, 8);
}
#define init_5_10_14_9
void FASTCALL call_5_10_14_9(SH2_struct* sh)
{
  SH2movll4(sh, 10,14, 9);
}
#define init_5_10_14_10
void FASTCALL call_5_10_14_10(SH2_struct* sh)
{
  SH2movll4(sh, 10,14, 10);
}
#define init_5_10_14_11
void FASTCALL call_5_10_14_11(SH2_struct* sh)
{
  SH2movll4(sh, 10,14, 11);
}
#define init_5_10_14_12
void FASTCALL call_5_10_14_12(SH2_struct* sh)
{
  SH2movll4(sh, 10,14, 12);
}
#define init_5_10_14_13
void FASTCALL call_5_10_14_13(SH2_struct* sh)
{
  SH2movll4(sh, 10,14, 13);
}
#define init_5_10_14_14
void FASTCALL call_5_10_14_14(SH2_struct* sh)
{
  SH2movll4(sh, 10,14, 14);
}
#define init_5_10_14_15
void FASTCALL call_5_10_14_15(SH2_struct* sh)
{
  SH2movll4(sh, 10,14, 15);
}
#define init_5_10_15_0
void FASTCALL call_5_10_15_0(SH2_struct* sh)
{
  SH2movll4(sh, 10,15, 0);
}
#define init_5_10_15_1
void FASTCALL call_5_10_15_1(SH2_struct* sh)
{
  SH2movll4(sh, 10,15, 1);
}
#define init_5_10_15_2
void FASTCALL call_5_10_15_2(SH2_struct* sh)
{
  SH2movll4(sh, 10,15, 2);
}
#define init_5_10_15_3
void FASTCALL call_5_10_15_3(SH2_struct* sh)
{
  SH2movll4(sh, 10,15, 3);
}
#define init_5_10_15_4
void FASTCALL call_5_10_15_4(SH2_struct* sh)
{
  SH2movll4(sh, 10,15, 4);
}
#define init_5_10_15_5
void FASTCALL call_5_10_15_5(SH2_struct* sh)
{
  SH2movll4(sh, 10,15, 5);
}
#define init_5_10_15_6
void FASTCALL call_5_10_15_6(SH2_struct* sh)
{
  SH2movll4(sh, 10,15, 6);
}
#define init_5_10_15_7
void FASTCALL call_5_10_15_7(SH2_struct* sh)
{
  SH2movll4(sh, 10,15, 7);
}
#define init_5_10_15_8
void FASTCALL call_5_10_15_8(SH2_struct* sh)
{
  SH2movll4(sh, 10,15, 8);
}
#define init_5_10_15_9
void FASTCALL call_5_10_15_9(SH2_struct* sh)
{
  SH2movll4(sh, 10,15, 9);
}
#define init_5_10_15_10
void FASTCALL call_5_10_15_10(SH2_struct* sh)
{
  SH2movll4(sh, 10,15, 10);
}
#define init_5_10_15_11
void FASTCALL call_5_10_15_11(SH2_struct* sh)
{
  SH2movll4(sh, 10,15, 11);
}
#define init_5_10_15_12
void FASTCALL call_5_10_15_12(SH2_struct* sh)
{
  SH2movll4(sh, 10,15, 12);
}
#define init_5_10_15_13
void FASTCALL call_5_10_15_13(SH2_struct* sh)
{
  SH2movll4(sh, 10,15, 13);
}
#define init_5_10_15_14
void FASTCALL call_5_10_15_14(SH2_struct* sh)
{
  SH2movll4(sh, 10,15, 14);
}
#define init_5_10_15_15
void FASTCALL call_5_10_15_15(SH2_struct* sh)
{
  SH2movll4(sh, 10,15, 15);
}
#define init_5_11_0_0
void FASTCALL call_5_11_0_0(SH2_struct* sh)
{
  SH2movll4(sh, 11,0, 0);
}
#define init_5_11_0_1
void FASTCALL call_5_11_0_1(SH2_struct* sh)
{
  SH2movll4(sh, 11,0, 1);
}
#define init_5_11_0_2
void FASTCALL call_5_11_0_2(SH2_struct* sh)
{
  SH2movll4(sh, 11,0, 2);
}
#define init_5_11_0_3
void FASTCALL call_5_11_0_3(SH2_struct* sh)
{
  SH2movll4(sh, 11,0, 3);
}
#define init_5_11_0_4
void FASTCALL call_5_11_0_4(SH2_struct* sh)
{
  SH2movll4(sh, 11,0, 4);
}
#define init_5_11_0_5
void FASTCALL call_5_11_0_5(SH2_struct* sh)
{
  SH2movll4(sh, 11,0, 5);
}
#define init_5_11_0_6
void FASTCALL call_5_11_0_6(SH2_struct* sh)
{
  SH2movll4(sh, 11,0, 6);
}
#define init_5_11_0_7
void FASTCALL call_5_11_0_7(SH2_struct* sh)
{
  SH2movll4(sh, 11,0, 7);
}
#define init_5_11_0_8
void FASTCALL call_5_11_0_8(SH2_struct* sh)
{
  SH2movll4(sh, 11,0, 8);
}
#define init_5_11_0_9
void FASTCALL call_5_11_0_9(SH2_struct* sh)
{
  SH2movll4(sh, 11,0, 9);
}
#define init_5_11_0_10
void FASTCALL call_5_11_0_10(SH2_struct* sh)
{
  SH2movll4(sh, 11,0, 10);
}
#define init_5_11_0_11
void FASTCALL call_5_11_0_11(SH2_struct* sh)
{
  SH2movll4(sh, 11,0, 11);
}
#define init_5_11_0_12
void FASTCALL call_5_11_0_12(SH2_struct* sh)
{
  SH2movll4(sh, 11,0, 12);
}
#define init_5_11_0_13
void FASTCALL call_5_11_0_13(SH2_struct* sh)
{
  SH2movll4(sh, 11,0, 13);
}
#define init_5_11_0_14
void FASTCALL call_5_11_0_14(SH2_struct* sh)
{
  SH2movll4(sh, 11,0, 14);
}
#define init_5_11_0_15
void FASTCALL call_5_11_0_15(SH2_struct* sh)
{
  SH2movll4(sh, 11,0, 15);
}
#define init_5_11_1_0
void FASTCALL call_5_11_1_0(SH2_struct* sh)
{
  SH2movll4(sh, 11,1, 0);
}
#define init_5_11_1_1
void FASTCALL call_5_11_1_1(SH2_struct* sh)
{
  SH2movll4(sh, 11,1, 1);
}
#define init_5_11_1_2
void FASTCALL call_5_11_1_2(SH2_struct* sh)
{
  SH2movll4(sh, 11,1, 2);
}
#define init_5_11_1_3
void FASTCALL call_5_11_1_3(SH2_struct* sh)
{
  SH2movll4(sh, 11,1, 3);
}
#define init_5_11_1_4
void FASTCALL call_5_11_1_4(SH2_struct* sh)
{
  SH2movll4(sh, 11,1, 4);
}
#define init_5_11_1_5
void FASTCALL call_5_11_1_5(SH2_struct* sh)
{
  SH2movll4(sh, 11,1, 5);
}
#define init_5_11_1_6
void FASTCALL call_5_11_1_6(SH2_struct* sh)
{
  SH2movll4(sh, 11,1, 6);
}
#define init_5_11_1_7
void FASTCALL call_5_11_1_7(SH2_struct* sh)
{
  SH2movll4(sh, 11,1, 7);
}
#define init_5_11_1_8
void FASTCALL call_5_11_1_8(SH2_struct* sh)
{
  SH2movll4(sh, 11,1, 8);
}
#define init_5_11_1_9
void FASTCALL call_5_11_1_9(SH2_struct* sh)
{
  SH2movll4(sh, 11,1, 9);
}
#define init_5_11_1_10
void FASTCALL call_5_11_1_10(SH2_struct* sh)
{
  SH2movll4(sh, 11,1, 10);
}
#define init_5_11_1_11
void FASTCALL call_5_11_1_11(SH2_struct* sh)
{
  SH2movll4(sh, 11,1, 11);
}
#define init_5_11_1_12
void FASTCALL call_5_11_1_12(SH2_struct* sh)
{
  SH2movll4(sh, 11,1, 12);
}
#define init_5_11_1_13
void FASTCALL call_5_11_1_13(SH2_struct* sh)
{
  SH2movll4(sh, 11,1, 13);
}
#define init_5_11_1_14
void FASTCALL call_5_11_1_14(SH2_struct* sh)
{
  SH2movll4(sh, 11,1, 14);
}
#define init_5_11_1_15
void FASTCALL call_5_11_1_15(SH2_struct* sh)
{
  SH2movll4(sh, 11,1, 15);
}
#define init_5_11_2_0
void FASTCALL call_5_11_2_0(SH2_struct* sh)
{
  SH2movll4(sh, 11,2, 0);
}
#define init_5_11_2_1
void FASTCALL call_5_11_2_1(SH2_struct* sh)
{
  SH2movll4(sh, 11,2, 1);
}
#define init_5_11_2_2
void FASTCALL call_5_11_2_2(SH2_struct* sh)
{
  SH2movll4(sh, 11,2, 2);
}
#define init_5_11_2_3
void FASTCALL call_5_11_2_3(SH2_struct* sh)
{
  SH2movll4(sh, 11,2, 3);
}
#define init_5_11_2_4
void FASTCALL call_5_11_2_4(SH2_struct* sh)
{
  SH2movll4(sh, 11,2, 4);
}
#define init_5_11_2_5
void FASTCALL call_5_11_2_5(SH2_struct* sh)
{
  SH2movll4(sh, 11,2, 5);
}
#define init_5_11_2_6
void FASTCALL call_5_11_2_6(SH2_struct* sh)
{
  SH2movll4(sh, 11,2, 6);
}
#define init_5_11_2_7
void FASTCALL call_5_11_2_7(SH2_struct* sh)
{
  SH2movll4(sh, 11,2, 7);
}
#define init_5_11_2_8
void FASTCALL call_5_11_2_8(SH2_struct* sh)
{
  SH2movll4(sh, 11,2, 8);
}
#define init_5_11_2_9
void FASTCALL call_5_11_2_9(SH2_struct* sh)
{
  SH2movll4(sh, 11,2, 9);
}
#define init_5_11_2_10
void FASTCALL call_5_11_2_10(SH2_struct* sh)
{
  SH2movll4(sh, 11,2, 10);
}
#define init_5_11_2_11
void FASTCALL call_5_11_2_11(SH2_struct* sh)
{
  SH2movll4(sh, 11,2, 11);
}
#define init_5_11_2_12
void FASTCALL call_5_11_2_12(SH2_struct* sh)
{
  SH2movll4(sh, 11,2, 12);
}
#define init_5_11_2_13
void FASTCALL call_5_11_2_13(SH2_struct* sh)
{
  SH2movll4(sh, 11,2, 13);
}
#define init_5_11_2_14
void FASTCALL call_5_11_2_14(SH2_struct* sh)
{
  SH2movll4(sh, 11,2, 14);
}
#define init_5_11_2_15
void FASTCALL call_5_11_2_15(SH2_struct* sh)
{
  SH2movll4(sh, 11,2, 15);
}
#define init_5_11_3_0
void FASTCALL call_5_11_3_0(SH2_struct* sh)
{
  SH2movll4(sh, 11,3, 0);
}
#define init_5_11_3_1
void FASTCALL call_5_11_3_1(SH2_struct* sh)
{
  SH2movll4(sh, 11,3, 1);
}
#define init_5_11_3_2
void FASTCALL call_5_11_3_2(SH2_struct* sh)
{
  SH2movll4(sh, 11,3, 2);
}
#define init_5_11_3_3
void FASTCALL call_5_11_3_3(SH2_struct* sh)
{
  SH2movll4(sh, 11,3, 3);
}
#define init_5_11_3_4
void FASTCALL call_5_11_3_4(SH2_struct* sh)
{
  SH2movll4(sh, 11,3, 4);
}
#define init_5_11_3_5
void FASTCALL call_5_11_3_5(SH2_struct* sh)
{
  SH2movll4(sh, 11,3, 5);
}
#define init_5_11_3_6
void FASTCALL call_5_11_3_6(SH2_struct* sh)
{
  SH2movll4(sh, 11,3, 6);
}
#define init_5_11_3_7
void FASTCALL call_5_11_3_7(SH2_struct* sh)
{
  SH2movll4(sh, 11,3, 7);
}
#define init_5_11_3_8
void FASTCALL call_5_11_3_8(SH2_struct* sh)
{
  SH2movll4(sh, 11,3, 8);
}
#define init_5_11_3_9
void FASTCALL call_5_11_3_9(SH2_struct* sh)
{
  SH2movll4(sh, 11,3, 9);
}
#define init_5_11_3_10
void FASTCALL call_5_11_3_10(SH2_struct* sh)
{
  SH2movll4(sh, 11,3, 10);
}
#define init_5_11_3_11
void FASTCALL call_5_11_3_11(SH2_struct* sh)
{
  SH2movll4(sh, 11,3, 11);
}
#define init_5_11_3_12
void FASTCALL call_5_11_3_12(SH2_struct* sh)
{
  SH2movll4(sh, 11,3, 12);
}
#define init_5_11_3_13
void FASTCALL call_5_11_3_13(SH2_struct* sh)
{
  SH2movll4(sh, 11,3, 13);
}
#define init_5_11_3_14
void FASTCALL call_5_11_3_14(SH2_struct* sh)
{
  SH2movll4(sh, 11,3, 14);
}
#define init_5_11_3_15
void FASTCALL call_5_11_3_15(SH2_struct* sh)
{
  SH2movll4(sh, 11,3, 15);
}
#define init_5_11_4_0
void FASTCALL call_5_11_4_0(SH2_struct* sh)
{
  SH2movll4(sh, 11,4, 0);
}
#define init_5_11_4_1
void FASTCALL call_5_11_4_1(SH2_struct* sh)
{
  SH2movll4(sh, 11,4, 1);
}
#define init_5_11_4_2
void FASTCALL call_5_11_4_2(SH2_struct* sh)
{
  SH2movll4(sh, 11,4, 2);
}
#define init_5_11_4_3
void FASTCALL call_5_11_4_3(SH2_struct* sh)
{
  SH2movll4(sh, 11,4, 3);
}
#define init_5_11_4_4
void FASTCALL call_5_11_4_4(SH2_struct* sh)
{
  SH2movll4(sh, 11,4, 4);
}
#define init_5_11_4_5
void FASTCALL call_5_11_4_5(SH2_struct* sh)
{
  SH2movll4(sh, 11,4, 5);
}
#define init_5_11_4_6
void FASTCALL call_5_11_4_6(SH2_struct* sh)
{
  SH2movll4(sh, 11,4, 6);
}
#define init_5_11_4_7
void FASTCALL call_5_11_4_7(SH2_struct* sh)
{
  SH2movll4(sh, 11,4, 7);
}
#define init_5_11_4_8
void FASTCALL call_5_11_4_8(SH2_struct* sh)
{
  SH2movll4(sh, 11,4, 8);
}
#define init_5_11_4_9
void FASTCALL call_5_11_4_9(SH2_struct* sh)
{
  SH2movll4(sh, 11,4, 9);
}
#define init_5_11_4_10
void FASTCALL call_5_11_4_10(SH2_struct* sh)
{
  SH2movll4(sh, 11,4, 10);
}
#define init_5_11_4_11
void FASTCALL call_5_11_4_11(SH2_struct* sh)
{
  SH2movll4(sh, 11,4, 11);
}
#define init_5_11_4_12
void FASTCALL call_5_11_4_12(SH2_struct* sh)
{
  SH2movll4(sh, 11,4, 12);
}
#define init_5_11_4_13
void FASTCALL call_5_11_4_13(SH2_struct* sh)
{
  SH2movll4(sh, 11,4, 13);
}
#define init_5_11_4_14
void FASTCALL call_5_11_4_14(SH2_struct* sh)
{
  SH2movll4(sh, 11,4, 14);
}
#define init_5_11_4_15
void FASTCALL call_5_11_4_15(SH2_struct* sh)
{
  SH2movll4(sh, 11,4, 15);
}
#define init_5_11_5_0
void FASTCALL call_5_11_5_0(SH2_struct* sh)
{
  SH2movll4(sh, 11,5, 0);
}
#define init_5_11_5_1
void FASTCALL call_5_11_5_1(SH2_struct* sh)
{
  SH2movll4(sh, 11,5, 1);
}
#define init_5_11_5_2
void FASTCALL call_5_11_5_2(SH2_struct* sh)
{
  SH2movll4(sh, 11,5, 2);
}
#define init_5_11_5_3
void FASTCALL call_5_11_5_3(SH2_struct* sh)
{
  SH2movll4(sh, 11,5, 3);
}
#define init_5_11_5_4
void FASTCALL call_5_11_5_4(SH2_struct* sh)
{
  SH2movll4(sh, 11,5, 4);
}
#define init_5_11_5_5
void FASTCALL call_5_11_5_5(SH2_struct* sh)
{
  SH2movll4(sh, 11,5, 5);
}
#define init_5_11_5_6
void FASTCALL call_5_11_5_6(SH2_struct* sh)
{
  SH2movll4(sh, 11,5, 6);
}
#define init_5_11_5_7
void FASTCALL call_5_11_5_7(SH2_struct* sh)
{
  SH2movll4(sh, 11,5, 7);
}
#define init_5_11_5_8
void FASTCALL call_5_11_5_8(SH2_struct* sh)
{
  SH2movll4(sh, 11,5, 8);
}
#define init_5_11_5_9
void FASTCALL call_5_11_5_9(SH2_struct* sh)
{
  SH2movll4(sh, 11,5, 9);
}
#define init_5_11_5_10
void FASTCALL call_5_11_5_10(SH2_struct* sh)
{
  SH2movll4(sh, 11,5, 10);
}
#define init_5_11_5_11
void FASTCALL call_5_11_5_11(SH2_struct* sh)
{
  SH2movll4(sh, 11,5, 11);
}
#define init_5_11_5_12
void FASTCALL call_5_11_5_12(SH2_struct* sh)
{
  SH2movll4(sh, 11,5, 12);
}
#define init_5_11_5_13
void FASTCALL call_5_11_5_13(SH2_struct* sh)
{
  SH2movll4(sh, 11,5, 13);
}
#define init_5_11_5_14
void FASTCALL call_5_11_5_14(SH2_struct* sh)
{
  SH2movll4(sh, 11,5, 14);
}
#define init_5_11_5_15
void FASTCALL call_5_11_5_15(SH2_struct* sh)
{
  SH2movll4(sh, 11,5, 15);
}
#define init_5_11_6_0
void FASTCALL call_5_11_6_0(SH2_struct* sh)
{
  SH2movll4(sh, 11,6, 0);
}
#define init_5_11_6_1
void FASTCALL call_5_11_6_1(SH2_struct* sh)
{
  SH2movll4(sh, 11,6, 1);
}
#define init_5_11_6_2
void FASTCALL call_5_11_6_2(SH2_struct* sh)
{
  SH2movll4(sh, 11,6, 2);
}
#define init_5_11_6_3
void FASTCALL call_5_11_6_3(SH2_struct* sh)
{
  SH2movll4(sh, 11,6, 3);
}
#define init_5_11_6_4
void FASTCALL call_5_11_6_4(SH2_struct* sh)
{
  SH2movll4(sh, 11,6, 4);
}
#define init_5_11_6_5
void FASTCALL call_5_11_6_5(SH2_struct* sh)
{
  SH2movll4(sh, 11,6, 5);
}
#define init_5_11_6_6
void FASTCALL call_5_11_6_6(SH2_struct* sh)
{
  SH2movll4(sh, 11,6, 6);
}
#define init_5_11_6_7
void FASTCALL call_5_11_6_7(SH2_struct* sh)
{
  SH2movll4(sh, 11,6, 7);
}
#define init_5_11_6_8
void FASTCALL call_5_11_6_8(SH2_struct* sh)
{
  SH2movll4(sh, 11,6, 8);
}
#define init_5_11_6_9
void FASTCALL call_5_11_6_9(SH2_struct* sh)
{
  SH2movll4(sh, 11,6, 9);
}
#define init_5_11_6_10
void FASTCALL call_5_11_6_10(SH2_struct* sh)
{
  SH2movll4(sh, 11,6, 10);
}
#define init_5_11_6_11
void FASTCALL call_5_11_6_11(SH2_struct* sh)
{
  SH2movll4(sh, 11,6, 11);
}
#define init_5_11_6_12
void FASTCALL call_5_11_6_12(SH2_struct* sh)
{
  SH2movll4(sh, 11,6, 12);
}
#define init_5_11_6_13
void FASTCALL call_5_11_6_13(SH2_struct* sh)
{
  SH2movll4(sh, 11,6, 13);
}
#define init_5_11_6_14
void FASTCALL call_5_11_6_14(SH2_struct* sh)
{
  SH2movll4(sh, 11,6, 14);
}
#define init_5_11_6_15
void FASTCALL call_5_11_6_15(SH2_struct* sh)
{
  SH2movll4(sh, 11,6, 15);
}
#define init_5_11_7_0
void FASTCALL call_5_11_7_0(SH2_struct* sh)
{
  SH2movll4(sh, 11,7, 0);
}
#define init_5_11_7_1
void FASTCALL call_5_11_7_1(SH2_struct* sh)
{
  SH2movll4(sh, 11,7, 1);
}
#define init_5_11_7_2
void FASTCALL call_5_11_7_2(SH2_struct* sh)
{
  SH2movll4(sh, 11,7, 2);
}
#define init_5_11_7_3
void FASTCALL call_5_11_7_3(SH2_struct* sh)
{
  SH2movll4(sh, 11,7, 3);
}
#define init_5_11_7_4
void FASTCALL call_5_11_7_4(SH2_struct* sh)
{
  SH2movll4(sh, 11,7, 4);
}
#define init_5_11_7_5
void FASTCALL call_5_11_7_5(SH2_struct* sh)
{
  SH2movll4(sh, 11,7, 5);
}
#define init_5_11_7_6
void FASTCALL call_5_11_7_6(SH2_struct* sh)
{
  SH2movll4(sh, 11,7, 6);
}
#define init_5_11_7_7
void FASTCALL call_5_11_7_7(SH2_struct* sh)
{
  SH2movll4(sh, 11,7, 7);
}
#define init_5_11_7_8
void FASTCALL call_5_11_7_8(SH2_struct* sh)
{
  SH2movll4(sh, 11,7, 8);
}
#define init_5_11_7_9
void FASTCALL call_5_11_7_9(SH2_struct* sh)
{
  SH2movll4(sh, 11,7, 9);
}
#define init_5_11_7_10
void FASTCALL call_5_11_7_10(SH2_struct* sh)
{
  SH2movll4(sh, 11,7, 10);
}
#define init_5_11_7_11
void FASTCALL call_5_11_7_11(SH2_struct* sh)
{
  SH2movll4(sh, 11,7, 11);
}
#define init_5_11_7_12
void FASTCALL call_5_11_7_12(SH2_struct* sh)
{
  SH2movll4(sh, 11,7, 12);
}
#define init_5_11_7_13
void FASTCALL call_5_11_7_13(SH2_struct* sh)
{
  SH2movll4(sh, 11,7, 13);
}
#define init_5_11_7_14
void FASTCALL call_5_11_7_14(SH2_struct* sh)
{
  SH2movll4(sh, 11,7, 14);
}
#define init_5_11_7_15
void FASTCALL call_5_11_7_15(SH2_struct* sh)
{
  SH2movll4(sh, 11,7, 15);
}
#define init_5_11_8_0
void FASTCALL call_5_11_8_0(SH2_struct* sh)
{
  SH2movll4(sh, 11,8, 0);
}
#define init_5_11_8_1
void FASTCALL call_5_11_8_1(SH2_struct* sh)
{
  SH2movll4(sh, 11,8, 1);
}
#define init_5_11_8_2
void FASTCALL call_5_11_8_2(SH2_struct* sh)
{
  SH2movll4(sh, 11,8, 2);
}
#define init_5_11_8_3
void FASTCALL call_5_11_8_3(SH2_struct* sh)
{
  SH2movll4(sh, 11,8, 3);
}
#define init_5_11_8_4
void FASTCALL call_5_11_8_4(SH2_struct* sh)
{
  SH2movll4(sh, 11,8, 4);
}
#define init_5_11_8_5
void FASTCALL call_5_11_8_5(SH2_struct* sh)
{
  SH2movll4(sh, 11,8, 5);
}
#define init_5_11_8_6
void FASTCALL call_5_11_8_6(SH2_struct* sh)
{
  SH2movll4(sh, 11,8, 6);
}
#define init_5_11_8_7
void FASTCALL call_5_11_8_7(SH2_struct* sh)
{
  SH2movll4(sh, 11,8, 7);
}
#define init_5_11_8_8
void FASTCALL call_5_11_8_8(SH2_struct* sh)
{
  SH2movll4(sh, 11,8, 8);
}
#define init_5_11_8_9
void FASTCALL call_5_11_8_9(SH2_struct* sh)
{
  SH2movll4(sh, 11,8, 9);
}
#define init_5_11_8_10
void FASTCALL call_5_11_8_10(SH2_struct* sh)
{
  SH2movll4(sh, 11,8, 10);
}
#define init_5_11_8_11
void FASTCALL call_5_11_8_11(SH2_struct* sh)
{
  SH2movll4(sh, 11,8, 11);
}
#define init_5_11_8_12
void FASTCALL call_5_11_8_12(SH2_struct* sh)
{
  SH2movll4(sh, 11,8, 12);
}
#define init_5_11_8_13
void FASTCALL call_5_11_8_13(SH2_struct* sh)
{
  SH2movll4(sh, 11,8, 13);
}
#define init_5_11_8_14
void FASTCALL call_5_11_8_14(SH2_struct* sh)
{
  SH2movll4(sh, 11,8, 14);
}
#define init_5_11_8_15
void FASTCALL call_5_11_8_15(SH2_struct* sh)
{
  SH2movll4(sh, 11,8, 15);
}
#define init_5_11_9_0
void FASTCALL call_5_11_9_0(SH2_struct* sh)
{
  SH2movll4(sh, 11,9, 0);
}
#define init_5_11_9_1
void FASTCALL call_5_11_9_1(SH2_struct* sh)
{
  SH2movll4(sh, 11,9, 1);
}
#define init_5_11_9_2
void FASTCALL call_5_11_9_2(SH2_struct* sh)
{
  SH2movll4(sh, 11,9, 2);
}
#define init_5_11_9_3
void FASTCALL call_5_11_9_3(SH2_struct* sh)
{
  SH2movll4(sh, 11,9, 3);
}
#define init_5_11_9_4
void FASTCALL call_5_11_9_4(SH2_struct* sh)
{
  SH2movll4(sh, 11,9, 4);
}
#define init_5_11_9_5
void FASTCALL call_5_11_9_5(SH2_struct* sh)
{
  SH2movll4(sh, 11,9, 5);
}
#define init_5_11_9_6
void FASTCALL call_5_11_9_6(SH2_struct* sh)
{
  SH2movll4(sh, 11,9, 6);
}
#define init_5_11_9_7
void FASTCALL call_5_11_9_7(SH2_struct* sh)
{
  SH2movll4(sh, 11,9, 7);
}
#define init_5_11_9_8
void FASTCALL call_5_11_9_8(SH2_struct* sh)
{
  SH2movll4(sh, 11,9, 8);
}
#define init_5_11_9_9
void FASTCALL call_5_11_9_9(SH2_struct* sh)
{
  SH2movll4(sh, 11,9, 9);
}
#define init_5_11_9_10
void FASTCALL call_5_11_9_10(SH2_struct* sh)
{
  SH2movll4(sh, 11,9, 10);
}
#define init_5_11_9_11
void FASTCALL call_5_11_9_11(SH2_struct* sh)
{
  SH2movll4(sh, 11,9, 11);
}
#define init_5_11_9_12
void FASTCALL call_5_11_9_12(SH2_struct* sh)
{
  SH2movll4(sh, 11,9, 12);
}
#define init_5_11_9_13
void FASTCALL call_5_11_9_13(SH2_struct* sh)
{
  SH2movll4(sh, 11,9, 13);
}
#define init_5_11_9_14
void FASTCALL call_5_11_9_14(SH2_struct* sh)
{
  SH2movll4(sh, 11,9, 14);
}
#define init_5_11_9_15
void FASTCALL call_5_11_9_15(SH2_struct* sh)
{
  SH2movll4(sh, 11,9, 15);
}
#define init_5_11_10_0
void FASTCALL call_5_11_10_0(SH2_struct* sh)
{
  SH2movll4(sh, 11,10, 0);
}
#define init_5_11_10_1
void FASTCALL call_5_11_10_1(SH2_struct* sh)
{
  SH2movll4(sh, 11,10, 1);
}
#define init_5_11_10_2
void FASTCALL call_5_11_10_2(SH2_struct* sh)
{
  SH2movll4(sh, 11,10, 2);
}
#define init_5_11_10_3
void FASTCALL call_5_11_10_3(SH2_struct* sh)
{
  SH2movll4(sh, 11,10, 3);
}
#define init_5_11_10_4
void FASTCALL call_5_11_10_4(SH2_struct* sh)
{
  SH2movll4(sh, 11,10, 4);
}
#define init_5_11_10_5
void FASTCALL call_5_11_10_5(SH2_struct* sh)
{
  SH2movll4(sh, 11,10, 5);
}
#define init_5_11_10_6
void FASTCALL call_5_11_10_6(SH2_struct* sh)
{
  SH2movll4(sh, 11,10, 6);
}
#define init_5_11_10_7
void FASTCALL call_5_11_10_7(SH2_struct* sh)
{
  SH2movll4(sh, 11,10, 7);
}
#define init_5_11_10_8
void FASTCALL call_5_11_10_8(SH2_struct* sh)
{
  SH2movll4(sh, 11,10, 8);
}
#define init_5_11_10_9
void FASTCALL call_5_11_10_9(SH2_struct* sh)
{
  SH2movll4(sh, 11,10, 9);
}
#define init_5_11_10_10
void FASTCALL call_5_11_10_10(SH2_struct* sh)
{
  SH2movll4(sh, 11,10, 10);
}
#define init_5_11_10_11
void FASTCALL call_5_11_10_11(SH2_struct* sh)
{
  SH2movll4(sh, 11,10, 11);
}
#define init_5_11_10_12
void FASTCALL call_5_11_10_12(SH2_struct* sh)
{
  SH2movll4(sh, 11,10, 12);
}
#define init_5_11_10_13
void FASTCALL call_5_11_10_13(SH2_struct* sh)
{
  SH2movll4(sh, 11,10, 13);
}
#define init_5_11_10_14
void FASTCALL call_5_11_10_14(SH2_struct* sh)
{
  SH2movll4(sh, 11,10, 14);
}
#define init_5_11_10_15
void FASTCALL call_5_11_10_15(SH2_struct* sh)
{
  SH2movll4(sh, 11,10, 15);
}
#define init_5_11_11_0
void FASTCALL call_5_11_11_0(SH2_struct* sh)
{
  SH2movll4(sh, 11,11, 0);
}
#define init_5_11_11_1
void FASTCALL call_5_11_11_1(SH2_struct* sh)
{
  SH2movll4(sh, 11,11, 1);
}
#define init_5_11_11_2
void FASTCALL call_5_11_11_2(SH2_struct* sh)
{
  SH2movll4(sh, 11,11, 2);
}
#define init_5_11_11_3
void FASTCALL call_5_11_11_3(SH2_struct* sh)
{
  SH2movll4(sh, 11,11, 3);
}
#define init_5_11_11_4
void FASTCALL call_5_11_11_4(SH2_struct* sh)
{
  SH2movll4(sh, 11,11, 4);
}
#define init_5_11_11_5
void FASTCALL call_5_11_11_5(SH2_struct* sh)
{
  SH2movll4(sh, 11,11, 5);
}
#define init_5_11_11_6
void FASTCALL call_5_11_11_6(SH2_struct* sh)
{
  SH2movll4(sh, 11,11, 6);
}
#define init_5_11_11_7
void FASTCALL call_5_11_11_7(SH2_struct* sh)
{
  SH2movll4(sh, 11,11, 7);
}
#define init_5_11_11_8
void FASTCALL call_5_11_11_8(SH2_struct* sh)
{
  SH2movll4(sh, 11,11, 8);
}
#define init_5_11_11_9
void FASTCALL call_5_11_11_9(SH2_struct* sh)
{
  SH2movll4(sh, 11,11, 9);
}
#define init_5_11_11_10
void FASTCALL call_5_11_11_10(SH2_struct* sh)
{
  SH2movll4(sh, 11,11, 10);
}
#define init_5_11_11_11
void FASTCALL call_5_11_11_11(SH2_struct* sh)
{
  SH2movll4(sh, 11,11, 11);
}
#define init_5_11_11_12
void FASTCALL call_5_11_11_12(SH2_struct* sh)
{
  SH2movll4(sh, 11,11, 12);
}
#define init_5_11_11_13
void FASTCALL call_5_11_11_13(SH2_struct* sh)
{
  SH2movll4(sh, 11,11, 13);
}
#define init_5_11_11_14
void FASTCALL call_5_11_11_14(SH2_struct* sh)
{
  SH2movll4(sh, 11,11, 14);
}
#define init_5_11_11_15
void FASTCALL call_5_11_11_15(SH2_struct* sh)
{
  SH2movll4(sh, 11,11, 15);
}
#define init_5_11_12_0
void FASTCALL call_5_11_12_0(SH2_struct* sh)
{
  SH2movll4(sh, 11,12, 0);
}
#define init_5_11_12_1
void FASTCALL call_5_11_12_1(SH2_struct* sh)
{
  SH2movll4(sh, 11,12, 1);
}
#define init_5_11_12_2
void FASTCALL call_5_11_12_2(SH2_struct* sh)
{
  SH2movll4(sh, 11,12, 2);
}
#define init_5_11_12_3
void FASTCALL call_5_11_12_3(SH2_struct* sh)
{
  SH2movll4(sh, 11,12, 3);
}
#define init_5_11_12_4
void FASTCALL call_5_11_12_4(SH2_struct* sh)
{
  SH2movll4(sh, 11,12, 4);
}
#define init_5_11_12_5
void FASTCALL call_5_11_12_5(SH2_struct* sh)
{
  SH2movll4(sh, 11,12, 5);
}
#define init_5_11_12_6
void FASTCALL call_5_11_12_6(SH2_struct* sh)
{
  SH2movll4(sh, 11,12, 6);
}
#define init_5_11_12_7
void FASTCALL call_5_11_12_7(SH2_struct* sh)
{
  SH2movll4(sh, 11,12, 7);
}
#define init_5_11_12_8
void FASTCALL call_5_11_12_8(SH2_struct* sh)
{
  SH2movll4(sh, 11,12, 8);
}
#define init_5_11_12_9
void FASTCALL call_5_11_12_9(SH2_struct* sh)
{
  SH2movll4(sh, 11,12, 9);
}
#define init_5_11_12_10
void FASTCALL call_5_11_12_10(SH2_struct* sh)
{
  SH2movll4(sh, 11,12, 10);
}
#define init_5_11_12_11
void FASTCALL call_5_11_12_11(SH2_struct* sh)
{
  SH2movll4(sh, 11,12, 11);
}
#define init_5_11_12_12
void FASTCALL call_5_11_12_12(SH2_struct* sh)
{
  SH2movll4(sh, 11,12, 12);
}
#define init_5_11_12_13
void FASTCALL call_5_11_12_13(SH2_struct* sh)
{
  SH2movll4(sh, 11,12, 13);
}
#define init_5_11_12_14
void FASTCALL call_5_11_12_14(SH2_struct* sh)
{
  SH2movll4(sh, 11,12, 14);
}
#define init_5_11_12_15
void FASTCALL call_5_11_12_15(SH2_struct* sh)
{
  SH2movll4(sh, 11,12, 15);
}
#define init_5_11_13_0
void FASTCALL call_5_11_13_0(SH2_struct* sh)
{
  SH2movll4(sh, 11,13, 0);
}
#define init_5_11_13_1
void FASTCALL call_5_11_13_1(SH2_struct* sh)
{
  SH2movll4(sh, 11,13, 1);
}
#define init_5_11_13_2
void FASTCALL call_5_11_13_2(SH2_struct* sh)
{
  SH2movll4(sh, 11,13, 2);
}
#define init_5_11_13_3
void FASTCALL call_5_11_13_3(SH2_struct* sh)
{
  SH2movll4(sh, 11,13, 3);
}
#define init_5_11_13_4
void FASTCALL call_5_11_13_4(SH2_struct* sh)
{
  SH2movll4(sh, 11,13, 4);
}
#define init_5_11_13_5
void FASTCALL call_5_11_13_5(SH2_struct* sh)
{
  SH2movll4(sh, 11,13, 5);
}
#define init_5_11_13_6
void FASTCALL call_5_11_13_6(SH2_struct* sh)
{
  SH2movll4(sh, 11,13, 6);
}
#define init_5_11_13_7
void FASTCALL call_5_11_13_7(SH2_struct* sh)
{
  SH2movll4(sh, 11,13, 7);
}
#define init_5_11_13_8
void FASTCALL call_5_11_13_8(SH2_struct* sh)
{
  SH2movll4(sh, 11,13, 8);
}
#define init_5_11_13_9
void FASTCALL call_5_11_13_9(SH2_struct* sh)
{
  SH2movll4(sh, 11,13, 9);
}
#define init_5_11_13_10
void FASTCALL call_5_11_13_10(SH2_struct* sh)
{
  SH2movll4(sh, 11,13, 10);
}
#define init_5_11_13_11
void FASTCALL call_5_11_13_11(SH2_struct* sh)
{
  SH2movll4(sh, 11,13, 11);
}
#define init_5_11_13_12
void FASTCALL call_5_11_13_12(SH2_struct* sh)
{
  SH2movll4(sh, 11,13, 12);
}
#define init_5_11_13_13
void FASTCALL call_5_11_13_13(SH2_struct* sh)
{
  SH2movll4(sh, 11,13, 13);
}
#define init_5_11_13_14
void FASTCALL call_5_11_13_14(SH2_struct* sh)
{
  SH2movll4(sh, 11,13, 14);
}
#define init_5_11_13_15
void FASTCALL call_5_11_13_15(SH2_struct* sh)
{
  SH2movll4(sh, 11,13, 15);
}
#define init_5_11_14_0
void FASTCALL call_5_11_14_0(SH2_struct* sh)
{
  SH2movll4(sh, 11,14, 0);
}
#define init_5_11_14_1
void FASTCALL call_5_11_14_1(SH2_struct* sh)
{
  SH2movll4(sh, 11,14, 1);
}
#define init_5_11_14_2
void FASTCALL call_5_11_14_2(SH2_struct* sh)
{
  SH2movll4(sh, 11,14, 2);
}
#define init_5_11_14_3
void FASTCALL call_5_11_14_3(SH2_struct* sh)
{
  SH2movll4(sh, 11,14, 3);
}
#define init_5_11_14_4
void FASTCALL call_5_11_14_4(SH2_struct* sh)
{
  SH2movll4(sh, 11,14, 4);
}
#define init_5_11_14_5
void FASTCALL call_5_11_14_5(SH2_struct* sh)
{
  SH2movll4(sh, 11,14, 5);
}
#define init_5_11_14_6
void FASTCALL call_5_11_14_6(SH2_struct* sh)
{
  SH2movll4(sh, 11,14, 6);
}
#define init_5_11_14_7
void FASTCALL call_5_11_14_7(SH2_struct* sh)
{
  SH2movll4(sh, 11,14, 7);
}
#define init_5_11_14_8
void FASTCALL call_5_11_14_8(SH2_struct* sh)
{
  SH2movll4(sh, 11,14, 8);
}
#define init_5_11_14_9
void FASTCALL call_5_11_14_9(SH2_struct* sh)
{
  SH2movll4(sh, 11,14, 9);
}
#define init_5_11_14_10
void FASTCALL call_5_11_14_10(SH2_struct* sh)
{
  SH2movll4(sh, 11,14, 10);
}
#define init_5_11_14_11
void FASTCALL call_5_11_14_11(SH2_struct* sh)
{
  SH2movll4(sh, 11,14, 11);
}
#define init_5_11_14_12
void FASTCALL call_5_11_14_12(SH2_struct* sh)
{
  SH2movll4(sh, 11,14, 12);
}
#define init_5_11_14_13
void FASTCALL call_5_11_14_13(SH2_struct* sh)
{
  SH2movll4(sh, 11,14, 13);
}
#define init_5_11_14_14
void FASTCALL call_5_11_14_14(SH2_struct* sh)
{
  SH2movll4(sh, 11,14, 14);
}
#define init_5_11_14_15
void FASTCALL call_5_11_14_15(SH2_struct* sh)
{
  SH2movll4(sh, 11,14, 15);
}
#define init_5_11_15_0
void FASTCALL call_5_11_15_0(SH2_struct* sh)
{
  SH2movll4(sh, 11,15, 0);
}
#define init_5_11_15_1
void FASTCALL call_5_11_15_1(SH2_struct* sh)
{
  SH2movll4(sh, 11,15, 1);
}
#define init_5_11_15_2
void FASTCALL call_5_11_15_2(SH2_struct* sh)
{
  SH2movll4(sh, 11,15, 2);
}
#define init_5_11_15_3
void FASTCALL call_5_11_15_3(SH2_struct* sh)
{
  SH2movll4(sh, 11,15, 3);
}
#define init_5_11_15_4
void FASTCALL call_5_11_15_4(SH2_struct* sh)
{
  SH2movll4(sh, 11,15, 4);
}
#define init_5_11_15_5
void FASTCALL call_5_11_15_5(SH2_struct* sh)
{
  SH2movll4(sh, 11,15, 5);
}
#define init_5_11_15_6
void FASTCALL call_5_11_15_6(SH2_struct* sh)
{
  SH2movll4(sh, 11,15, 6);
}
#define init_5_11_15_7
void FASTCALL call_5_11_15_7(SH2_struct* sh)
{
  SH2movll4(sh, 11,15, 7);
}
#define init_5_11_15_8
void FASTCALL call_5_11_15_8(SH2_struct* sh)
{
  SH2movll4(sh, 11,15, 8);
}
#define init_5_11_15_9
void FASTCALL call_5_11_15_9(SH2_struct* sh)
{
  SH2movll4(sh, 11,15, 9);
}
#define init_5_11_15_10
void FASTCALL call_5_11_15_10(SH2_struct* sh)
{
  SH2movll4(sh, 11,15, 10);
}
#define init_5_11_15_11
void FASTCALL call_5_11_15_11(SH2_struct* sh)
{
  SH2movll4(sh, 11,15, 11);
}
#define init_5_11_15_12
void FASTCALL call_5_11_15_12(SH2_struct* sh)
{
  SH2movll4(sh, 11,15, 12);
}
#define init_5_11_15_13
void FASTCALL call_5_11_15_13(SH2_struct* sh)
{
  SH2movll4(sh, 11,15, 13);
}
#define init_5_11_15_14
void FASTCALL call_5_11_15_14(SH2_struct* sh)
{
  SH2movll4(sh, 11,15, 14);
}
#define init_5_11_15_15
void FASTCALL call_5_11_15_15(SH2_struct* sh)
{
  SH2movll4(sh, 11,15, 15);
}
#define init_5_12_0_0
void FASTCALL call_5_12_0_0(SH2_struct* sh)
{
  SH2movll4(sh, 12,0, 0);
}
#define init_5_12_0_1
void FASTCALL call_5_12_0_1(SH2_struct* sh)
{
  SH2movll4(sh, 12,0, 1);
}
#define init_5_12_0_2
void FASTCALL call_5_12_0_2(SH2_struct* sh)
{
  SH2movll4(sh, 12,0, 2);
}
#define init_5_12_0_3
void FASTCALL call_5_12_0_3(SH2_struct* sh)
{
  SH2movll4(sh, 12,0, 3);
}
#define init_5_12_0_4
void FASTCALL call_5_12_0_4(SH2_struct* sh)
{
  SH2movll4(sh, 12,0, 4);
}
#define init_5_12_0_5
void FASTCALL call_5_12_0_5(SH2_struct* sh)
{
  SH2movll4(sh, 12,0, 5);
}
#define init_5_12_0_6
void FASTCALL call_5_12_0_6(SH2_struct* sh)
{
  SH2movll4(sh, 12,0, 6);
}
#define init_5_12_0_7
void FASTCALL call_5_12_0_7(SH2_struct* sh)
{
  SH2movll4(sh, 12,0, 7);
}
#define init_5_12_0_8
void FASTCALL call_5_12_0_8(SH2_struct* sh)
{
  SH2movll4(sh, 12,0, 8);
}
#define init_5_12_0_9
void FASTCALL call_5_12_0_9(SH2_struct* sh)
{
  SH2movll4(sh, 12,0, 9);
}
#define init_5_12_0_10
void FASTCALL call_5_12_0_10(SH2_struct* sh)
{
  SH2movll4(sh, 12,0, 10);
}
#define init_5_12_0_11
void FASTCALL call_5_12_0_11(SH2_struct* sh)
{
  SH2movll4(sh, 12,0, 11);
}
#define init_5_12_0_12
void FASTCALL call_5_12_0_12(SH2_struct* sh)
{
  SH2movll4(sh, 12,0, 12);
}
#define init_5_12_0_13
void FASTCALL call_5_12_0_13(SH2_struct* sh)
{
  SH2movll4(sh, 12,0, 13);
}
#define init_5_12_0_14
void FASTCALL call_5_12_0_14(SH2_struct* sh)
{
  SH2movll4(sh, 12,0, 14);
}
#define init_5_12_0_15
void FASTCALL call_5_12_0_15(SH2_struct* sh)
{
  SH2movll4(sh, 12,0, 15);
}
#define init_5_12_1_0
void FASTCALL call_5_12_1_0(SH2_struct* sh)
{
  SH2movll4(sh, 12,1, 0);
}
#define init_5_12_1_1
void FASTCALL call_5_12_1_1(SH2_struct* sh)
{
  SH2movll4(sh, 12,1, 1);
}
#define init_5_12_1_2
void FASTCALL call_5_12_1_2(SH2_struct* sh)
{
  SH2movll4(sh, 12,1, 2);
}
#define init_5_12_1_3
void FASTCALL call_5_12_1_3(SH2_struct* sh)
{
  SH2movll4(sh, 12,1, 3);
}
#define init_5_12_1_4
void FASTCALL call_5_12_1_4(SH2_struct* sh)
{
  SH2movll4(sh, 12,1, 4);
}
#define init_5_12_1_5
void FASTCALL call_5_12_1_5(SH2_struct* sh)
{
  SH2movll4(sh, 12,1, 5);
}
#define init_5_12_1_6
void FASTCALL call_5_12_1_6(SH2_struct* sh)
{
  SH2movll4(sh, 12,1, 6);
}
#define init_5_12_1_7
void FASTCALL call_5_12_1_7(SH2_struct* sh)
{
  SH2movll4(sh, 12,1, 7);
}
#define init_5_12_1_8
void FASTCALL call_5_12_1_8(SH2_struct* sh)
{
  SH2movll4(sh, 12,1, 8);
}
#define init_5_12_1_9
void FASTCALL call_5_12_1_9(SH2_struct* sh)
{
  SH2movll4(sh, 12,1, 9);
}
#define init_5_12_1_10
void FASTCALL call_5_12_1_10(SH2_struct* sh)
{
  SH2movll4(sh, 12,1, 10);
}
#define init_5_12_1_11
void FASTCALL call_5_12_1_11(SH2_struct* sh)
{
  SH2movll4(sh, 12,1, 11);
}
#define init_5_12_1_12
void FASTCALL call_5_12_1_12(SH2_struct* sh)
{
  SH2movll4(sh, 12,1, 12);
}
#define init_5_12_1_13
void FASTCALL call_5_12_1_13(SH2_struct* sh)
{
  SH2movll4(sh, 12,1, 13);
}
#define init_5_12_1_14
void FASTCALL call_5_12_1_14(SH2_struct* sh)
{
  SH2movll4(sh, 12,1, 14);
}
#define init_5_12_1_15
void FASTCALL call_5_12_1_15(SH2_struct* sh)
{
  SH2movll4(sh, 12,1, 15);
}
#define init_5_12_2_0
void FASTCALL call_5_12_2_0(SH2_struct* sh)
{
  SH2movll4(sh, 12,2, 0);
}
#define init_5_12_2_1
void FASTCALL call_5_12_2_1(SH2_struct* sh)
{
  SH2movll4(sh, 12,2, 1);
}
#define init_5_12_2_2
void FASTCALL call_5_12_2_2(SH2_struct* sh)
{
  SH2movll4(sh, 12,2, 2);
}
#define init_5_12_2_3
void FASTCALL call_5_12_2_3(SH2_struct* sh)
{
  SH2movll4(sh, 12,2, 3);
}
#define init_5_12_2_4
void FASTCALL call_5_12_2_4(SH2_struct* sh)
{
  SH2movll4(sh, 12,2, 4);
}
#define init_5_12_2_5
void FASTCALL call_5_12_2_5(SH2_struct* sh)
{
  SH2movll4(sh, 12,2, 5);
}
#define init_5_12_2_6
void FASTCALL call_5_12_2_6(SH2_struct* sh)
{
  SH2movll4(sh, 12,2, 6);
}
#define init_5_12_2_7
void FASTCALL call_5_12_2_7(SH2_struct* sh)
{
  SH2movll4(sh, 12,2, 7);
}
#define init_5_12_2_8
void FASTCALL call_5_12_2_8(SH2_struct* sh)
{
  SH2movll4(sh, 12,2, 8);
}
#define init_5_12_2_9
void FASTCALL call_5_12_2_9(SH2_struct* sh)
{
  SH2movll4(sh, 12,2, 9);
}
#define init_5_12_2_10
void FASTCALL call_5_12_2_10(SH2_struct* sh)
{
  SH2movll4(sh, 12,2, 10);
}
#define init_5_12_2_11
void FASTCALL call_5_12_2_11(SH2_struct* sh)
{
  SH2movll4(sh, 12,2, 11);
}
#define init_5_12_2_12
void FASTCALL call_5_12_2_12(SH2_struct* sh)
{
  SH2movll4(sh, 12,2, 12);
}
#define init_5_12_2_13
void FASTCALL call_5_12_2_13(SH2_struct* sh)
{
  SH2movll4(sh, 12,2, 13);
}
#define init_5_12_2_14
void FASTCALL call_5_12_2_14(SH2_struct* sh)
{
  SH2movll4(sh, 12,2, 14);
}
#define init_5_12_2_15
void FASTCALL call_5_12_2_15(SH2_struct* sh)
{
  SH2movll4(sh, 12,2, 15);
}
#define init_5_12_3_0
void FASTCALL call_5_12_3_0(SH2_struct* sh)
{
  SH2movll4(sh, 12,3, 0);
}
#define init_5_12_3_1
void FASTCALL call_5_12_3_1(SH2_struct* sh)
{
  SH2movll4(sh, 12,3, 1);
}
#define init_5_12_3_2
void FASTCALL call_5_12_3_2(SH2_struct* sh)
{
  SH2movll4(sh, 12,3, 2);
}
#define init_5_12_3_3
void FASTCALL call_5_12_3_3(SH2_struct* sh)
{
  SH2movll4(sh, 12,3, 3);
}
#define init_5_12_3_4
void FASTCALL call_5_12_3_4(SH2_struct* sh)
{
  SH2movll4(sh, 12,3, 4);
}
#define init_5_12_3_5
void FASTCALL call_5_12_3_5(SH2_struct* sh)
{
  SH2movll4(sh, 12,3, 5);
}
#define init_5_12_3_6
void FASTCALL call_5_12_3_6(SH2_struct* sh)
{
  SH2movll4(sh, 12,3, 6);
}
#define init_5_12_3_7
void FASTCALL call_5_12_3_7(SH2_struct* sh)
{
  SH2movll4(sh, 12,3, 7);
}
#define init_5_12_3_8
void FASTCALL call_5_12_3_8(SH2_struct* sh)
{
  SH2movll4(sh, 12,3, 8);
}
#define init_5_12_3_9
void FASTCALL call_5_12_3_9(SH2_struct* sh)
{
  SH2movll4(sh, 12,3, 9);
}
#define init_5_12_3_10
void FASTCALL call_5_12_3_10(SH2_struct* sh)
{
  SH2movll4(sh, 12,3, 10);
}
#define init_5_12_3_11
void FASTCALL call_5_12_3_11(SH2_struct* sh)
{
  SH2movll4(sh, 12,3, 11);
}
#define init_5_12_3_12
void FASTCALL call_5_12_3_12(SH2_struct* sh)
{
  SH2movll4(sh, 12,3, 12);
}
#define init_5_12_3_13
void FASTCALL call_5_12_3_13(SH2_struct* sh)
{
  SH2movll4(sh, 12,3, 13);
}
#define init_5_12_3_14
void FASTCALL call_5_12_3_14(SH2_struct* sh)
{
  SH2movll4(sh, 12,3, 14);
}
#define init_5_12_3_15
void FASTCALL call_5_12_3_15(SH2_struct* sh)
{
  SH2movll4(sh, 12,3, 15);
}
#define init_5_12_4_0
void FASTCALL call_5_12_4_0(SH2_struct* sh)
{
  SH2movll4(sh, 12,4, 0);
}
#define init_5_12_4_1
void FASTCALL call_5_12_4_1(SH2_struct* sh)
{
  SH2movll4(sh, 12,4, 1);
}
#define init_5_12_4_2
void FASTCALL call_5_12_4_2(SH2_struct* sh)
{
  SH2movll4(sh, 12,4, 2);
}
#define init_5_12_4_3
void FASTCALL call_5_12_4_3(SH2_struct* sh)
{
  SH2movll4(sh, 12,4, 3);
}
#define init_5_12_4_4
void FASTCALL call_5_12_4_4(SH2_struct* sh)
{
  SH2movll4(sh, 12,4, 4);
}
#define init_5_12_4_5
void FASTCALL call_5_12_4_5(SH2_struct* sh)
{
  SH2movll4(sh, 12,4, 5);
}
#define init_5_12_4_6
void FASTCALL call_5_12_4_6(SH2_struct* sh)
{
  SH2movll4(sh, 12,4, 6);
}
#define init_5_12_4_7
void FASTCALL call_5_12_4_7(SH2_struct* sh)
{
  SH2movll4(sh, 12,4, 7);
}
#define init_5_12_4_8
void FASTCALL call_5_12_4_8(SH2_struct* sh)
{
  SH2movll4(sh, 12,4, 8);
}
#define init_5_12_4_9
void FASTCALL call_5_12_4_9(SH2_struct* sh)
{
  SH2movll4(sh, 12,4, 9);
}
#define init_5_12_4_10
void FASTCALL call_5_12_4_10(SH2_struct* sh)
{
  SH2movll4(sh, 12,4, 10);
}
#define init_5_12_4_11
void FASTCALL call_5_12_4_11(SH2_struct* sh)
{
  SH2movll4(sh, 12,4, 11);
}
#define init_5_12_4_12
void FASTCALL call_5_12_4_12(SH2_struct* sh)
{
  SH2movll4(sh, 12,4, 12);
}
#define init_5_12_4_13
void FASTCALL call_5_12_4_13(SH2_struct* sh)
{
  SH2movll4(sh, 12,4, 13);
}
#define init_5_12_4_14
void FASTCALL call_5_12_4_14(SH2_struct* sh)
{
  SH2movll4(sh, 12,4, 14);
}
#define init_5_12_4_15
void FASTCALL call_5_12_4_15(SH2_struct* sh)
{
  SH2movll4(sh, 12,4, 15);
}
#define init_5_12_5_0
void FASTCALL call_5_12_5_0(SH2_struct* sh)
{
  SH2movll4(sh, 12,5, 0);
}
#define init_5_12_5_1
void FASTCALL call_5_12_5_1(SH2_struct* sh)
{
  SH2movll4(sh, 12,5, 1);
}
#define init_5_12_5_2
void FASTCALL call_5_12_5_2(SH2_struct* sh)
{
  SH2movll4(sh, 12,5, 2);
}
#define init_5_12_5_3
void FASTCALL call_5_12_5_3(SH2_struct* sh)
{
  SH2movll4(sh, 12,5, 3);
}
#define init_5_12_5_4
void FASTCALL call_5_12_5_4(SH2_struct* sh)
{
  SH2movll4(sh, 12,5, 4);
}
#define init_5_12_5_5
void FASTCALL call_5_12_5_5(SH2_struct* sh)
{
  SH2movll4(sh, 12,5, 5);
}
#define init_5_12_5_6
void FASTCALL call_5_12_5_6(SH2_struct* sh)
{
  SH2movll4(sh, 12,5, 6);
}
#define init_5_12_5_7
void FASTCALL call_5_12_5_7(SH2_struct* sh)
{
  SH2movll4(sh, 12,5, 7);
}
#define init_5_12_5_8
void FASTCALL call_5_12_5_8(SH2_struct* sh)
{
  SH2movll4(sh, 12,5, 8);
}
#define init_5_12_5_9
void FASTCALL call_5_12_5_9(SH2_struct* sh)
{
  SH2movll4(sh, 12,5, 9);
}
#define init_5_12_5_10
void FASTCALL call_5_12_5_10(SH2_struct* sh)
{
  SH2movll4(sh, 12,5, 10);
}
#define init_5_12_5_11
void FASTCALL call_5_12_5_11(SH2_struct* sh)
{
  SH2movll4(sh, 12,5, 11);
}
#define init_5_12_5_12
void FASTCALL call_5_12_5_12(SH2_struct* sh)
{
  SH2movll4(sh, 12,5, 12);
}
#define init_5_12_5_13
void FASTCALL call_5_12_5_13(SH2_struct* sh)
{
  SH2movll4(sh, 12,5, 13);
}
#define init_5_12_5_14
void FASTCALL call_5_12_5_14(SH2_struct* sh)
{
  SH2movll4(sh, 12,5, 14);
}
#define init_5_12_5_15
void FASTCALL call_5_12_5_15(SH2_struct* sh)
{
  SH2movll4(sh, 12,5, 15);
}
#define init_5_12_6_0
void FASTCALL call_5_12_6_0(SH2_struct* sh)
{
  SH2movll4(sh, 12,6, 0);
}
#define init_5_12_6_1
void FASTCALL call_5_12_6_1(SH2_struct* sh)
{
  SH2movll4(sh, 12,6, 1);
}
#define init_5_12_6_2
void FASTCALL call_5_12_6_2(SH2_struct* sh)
{
  SH2movll4(sh, 12,6, 2);
}
#define init_5_12_6_3
void FASTCALL call_5_12_6_3(SH2_struct* sh)
{
  SH2movll4(sh, 12,6, 3);
}
#define init_5_12_6_4
void FASTCALL call_5_12_6_4(SH2_struct* sh)
{
  SH2movll4(sh, 12,6, 4);
}
#define init_5_12_6_5
void FASTCALL call_5_12_6_5(SH2_struct* sh)
{
  SH2movll4(sh, 12,6, 5);
}
#define init_5_12_6_6
void FASTCALL call_5_12_6_6(SH2_struct* sh)
{
  SH2movll4(sh, 12,6, 6);
}
#define init_5_12_6_7
void FASTCALL call_5_12_6_7(SH2_struct* sh)
{
  SH2movll4(sh, 12,6, 7);
}
#define init_5_12_6_8
void FASTCALL call_5_12_6_8(SH2_struct* sh)
{
  SH2movll4(sh, 12,6, 8);
}
#define init_5_12_6_9
void FASTCALL call_5_12_6_9(SH2_struct* sh)
{
  SH2movll4(sh, 12,6, 9);
}
#define init_5_12_6_10
void FASTCALL call_5_12_6_10(SH2_struct* sh)
{
  SH2movll4(sh, 12,6, 10);
}
#define init_5_12_6_11
void FASTCALL call_5_12_6_11(SH2_struct* sh)
{
  SH2movll4(sh, 12,6, 11);
}
#define init_5_12_6_12
void FASTCALL call_5_12_6_12(SH2_struct* sh)
{
  SH2movll4(sh, 12,6, 12);
}
#define init_5_12_6_13
void FASTCALL call_5_12_6_13(SH2_struct* sh)
{
  SH2movll4(sh, 12,6, 13);
}
#define init_5_12_6_14
void FASTCALL call_5_12_6_14(SH2_struct* sh)
{
  SH2movll4(sh, 12,6, 14);
}
#define init_5_12_6_15
void FASTCALL call_5_12_6_15(SH2_struct* sh)
{
  SH2movll4(sh, 12,6, 15);
}
#define init_5_12_7_0
void FASTCALL call_5_12_7_0(SH2_struct* sh)
{
  SH2movll4(sh, 12,7, 0);
}
#define init_5_12_7_1
void FASTCALL call_5_12_7_1(SH2_struct* sh)
{
  SH2movll4(sh, 12,7, 1);
}
#define init_5_12_7_2
void FASTCALL call_5_12_7_2(SH2_struct* sh)
{
  SH2movll4(sh, 12,7, 2);
}
#define init_5_12_7_3
void FASTCALL call_5_12_7_3(SH2_struct* sh)
{
  SH2movll4(sh, 12,7, 3);
}
#define init_5_12_7_4
void FASTCALL call_5_12_7_4(SH2_struct* sh)
{
  SH2movll4(sh, 12,7, 4);
}
#define init_5_12_7_5
void FASTCALL call_5_12_7_5(SH2_struct* sh)
{
  SH2movll4(sh, 12,7, 5);
}
#define init_5_12_7_6
void FASTCALL call_5_12_7_6(SH2_struct* sh)
{
  SH2movll4(sh, 12,7, 6);
}
#define init_5_12_7_7
void FASTCALL call_5_12_7_7(SH2_struct* sh)
{
  SH2movll4(sh, 12,7, 7);
}
#define init_5_12_7_8
void FASTCALL call_5_12_7_8(SH2_struct* sh)
{
  SH2movll4(sh, 12,7, 8);
}
#define init_5_12_7_9
void FASTCALL call_5_12_7_9(SH2_struct* sh)
{
  SH2movll4(sh, 12,7, 9);
}
#define init_5_12_7_10
void FASTCALL call_5_12_7_10(SH2_struct* sh)
{
  SH2movll4(sh, 12,7, 10);
}
#define init_5_12_7_11
void FASTCALL call_5_12_7_11(SH2_struct* sh)
{
  SH2movll4(sh, 12,7, 11);
}
#define init_5_12_7_12
void FASTCALL call_5_12_7_12(SH2_struct* sh)
{
  SH2movll4(sh, 12,7, 12);
}
#define init_5_12_7_13
void FASTCALL call_5_12_7_13(SH2_struct* sh)
{
  SH2movll4(sh, 12,7, 13);
}
#define init_5_12_7_14
void FASTCALL call_5_12_7_14(SH2_struct* sh)
{
  SH2movll4(sh, 12,7, 14);
}
#define init_5_12_7_15
void FASTCALL call_5_12_7_15(SH2_struct* sh)
{
  SH2movll4(sh, 12,7, 15);
}
#define init_5_12_8_0
void FASTCALL call_5_12_8_0(SH2_struct* sh)
{
  SH2movll4(sh, 12,8, 0);
}
#define init_5_12_8_1
void FASTCALL call_5_12_8_1(SH2_struct* sh)
{
  SH2movll4(sh, 12,8, 1);
}
#define init_5_12_8_2
void FASTCALL call_5_12_8_2(SH2_struct* sh)
{
  SH2movll4(sh, 12,8, 2);
}
#define init_5_12_8_3
void FASTCALL call_5_12_8_3(SH2_struct* sh)
{
  SH2movll4(sh, 12,8, 3);
}
#define init_5_12_8_4
void FASTCALL call_5_12_8_4(SH2_struct* sh)
{
  SH2movll4(sh, 12,8, 4);
}
#define init_5_12_8_5
void FASTCALL call_5_12_8_5(SH2_struct* sh)
{
  SH2movll4(sh, 12,8, 5);
}
#define init_5_12_8_6
void FASTCALL call_5_12_8_6(SH2_struct* sh)
{
  SH2movll4(sh, 12,8, 6);
}
#define init_5_12_8_7
void FASTCALL call_5_12_8_7(SH2_struct* sh)
{
  SH2movll4(sh, 12,8, 7);
}
#define init_5_12_8_8
void FASTCALL call_5_12_8_8(SH2_struct* sh)
{
  SH2movll4(sh, 12,8, 8);
}
#define init_5_12_8_9
void FASTCALL call_5_12_8_9(SH2_struct* sh)
{
  SH2movll4(sh, 12,8, 9);
}
#define init_5_12_8_10
void FASTCALL call_5_12_8_10(SH2_struct* sh)
{
  SH2movll4(sh, 12,8, 10);
}
#define init_5_12_8_11
void FASTCALL call_5_12_8_11(SH2_struct* sh)
{
  SH2movll4(sh, 12,8, 11);
}
#define init_5_12_8_12
void FASTCALL call_5_12_8_12(SH2_struct* sh)
{
  SH2movll4(sh, 12,8, 12);
}
#define init_5_12_8_13
void FASTCALL call_5_12_8_13(SH2_struct* sh)
{
  SH2movll4(sh, 12,8, 13);
}
#define init_5_12_8_14
void FASTCALL call_5_12_8_14(SH2_struct* sh)
{
  SH2movll4(sh, 12,8, 14);
}
#define init_5_12_8_15
void FASTCALL call_5_12_8_15(SH2_struct* sh)
{
  SH2movll4(sh, 12,8, 15);
}
#define init_5_12_9_0
void FASTCALL call_5_12_9_0(SH2_struct* sh)
{
  SH2movll4(sh, 12,9, 0);
}
#define init_5_12_9_1
void FASTCALL call_5_12_9_1(SH2_struct* sh)
{
  SH2movll4(sh, 12,9, 1);
}
#define init_5_12_9_2
void FASTCALL call_5_12_9_2(SH2_struct* sh)
{
  SH2movll4(sh, 12,9, 2);
}
#define init_5_12_9_3
void FASTCALL call_5_12_9_3(SH2_struct* sh)
{
  SH2movll4(sh, 12,9, 3);
}
#define init_5_12_9_4
void FASTCALL call_5_12_9_4(SH2_struct* sh)
{
  SH2movll4(sh, 12,9, 4);
}
#define init_5_12_9_5
void FASTCALL call_5_12_9_5(SH2_struct* sh)
{
  SH2movll4(sh, 12,9, 5);
}
#define init_5_12_9_6
void FASTCALL call_5_12_9_6(SH2_struct* sh)
{
  SH2movll4(sh, 12,9, 6);
}
#define init_5_12_9_7
void FASTCALL call_5_12_9_7(SH2_struct* sh)
{
  SH2movll4(sh, 12,9, 7);
}
#define init_5_12_9_8
void FASTCALL call_5_12_9_8(SH2_struct* sh)
{
  SH2movll4(sh, 12,9, 8);
}
#define init_5_12_9_9
void FASTCALL call_5_12_9_9(SH2_struct* sh)
{
  SH2movll4(sh, 12,9, 9);
}
#define init_5_12_9_10
void FASTCALL call_5_12_9_10(SH2_struct* sh)
{
  SH2movll4(sh, 12,9, 10);
}
#define init_5_12_9_11
void FASTCALL call_5_12_9_11(SH2_struct* sh)
{
  SH2movll4(sh, 12,9, 11);
}
#define init_5_12_9_12
void FASTCALL call_5_12_9_12(SH2_struct* sh)
{
  SH2movll4(sh, 12,9, 12);
}
#define init_5_12_9_13
void FASTCALL call_5_12_9_13(SH2_struct* sh)
{
  SH2movll4(sh, 12,9, 13);
}
#define init_5_12_9_14
void FASTCALL call_5_12_9_14(SH2_struct* sh)
{
  SH2movll4(sh, 12,9, 14);
}
#define init_5_12_9_15
void FASTCALL call_5_12_9_15(SH2_struct* sh)
{
  SH2movll4(sh, 12,9, 15);
}
#define init_5_12_10_0
void FASTCALL call_5_12_10_0(SH2_struct* sh)
{
  SH2movll4(sh, 12,10, 0);
}
#define init_5_12_10_1
void FASTCALL call_5_12_10_1(SH2_struct* sh)
{
  SH2movll4(sh, 12,10, 1);
}
#define init_5_12_10_2
void FASTCALL call_5_12_10_2(SH2_struct* sh)
{
  SH2movll4(sh, 12,10, 2);
}
#define init_5_12_10_3
void FASTCALL call_5_12_10_3(SH2_struct* sh)
{
  SH2movll4(sh, 12,10, 3);
}
#define init_5_12_10_4
void FASTCALL call_5_12_10_4(SH2_struct* sh)
{
  SH2movll4(sh, 12,10, 4);
}
#define init_5_12_10_5
void FASTCALL call_5_12_10_5(SH2_struct* sh)
{
  SH2movll4(sh, 12,10, 5);
}
#define init_5_12_10_6
void FASTCALL call_5_12_10_6(SH2_struct* sh)
{
  SH2movll4(sh, 12,10, 6);
}
#define init_5_12_10_7
void FASTCALL call_5_12_10_7(SH2_struct* sh)
{
  SH2movll4(sh, 12,10, 7);
}
#define init_5_12_10_8
void FASTCALL call_5_12_10_8(SH2_struct* sh)
{
  SH2movll4(sh, 12,10, 8);
}
#define init_5_12_10_9
void FASTCALL call_5_12_10_9(SH2_struct* sh)
{
  SH2movll4(sh, 12,10, 9);
}
#define init_5_12_10_10
void FASTCALL call_5_12_10_10(SH2_struct* sh)
{
  SH2movll4(sh, 12,10, 10);
}
#define init_5_12_10_11
void FASTCALL call_5_12_10_11(SH2_struct* sh)
{
  SH2movll4(sh, 12,10, 11);
}
#define init_5_12_10_12
void FASTCALL call_5_12_10_12(SH2_struct* sh)
{
  SH2movll4(sh, 12,10, 12);
}
#define init_5_12_10_13
void FASTCALL call_5_12_10_13(SH2_struct* sh)
{
  SH2movll4(sh, 12,10, 13);
}
#define init_5_12_10_14
void FASTCALL call_5_12_10_14(SH2_struct* sh)
{
  SH2movll4(sh, 12,10, 14);
}
#define init_5_12_10_15
void FASTCALL call_5_12_10_15(SH2_struct* sh)
{
  SH2movll4(sh, 12,10, 15);
}
#define init_5_12_11_0
void FASTCALL call_5_12_11_0(SH2_struct* sh)
{
  SH2movll4(sh, 12,11, 0);
}
#define init_5_12_11_1
void FASTCALL call_5_12_11_1(SH2_struct* sh)
{
  SH2movll4(sh, 12,11, 1);
}
#define init_5_12_11_2
void FASTCALL call_5_12_11_2(SH2_struct* sh)
{
  SH2movll4(sh, 12,11, 2);
}
#define init_5_12_11_3
void FASTCALL call_5_12_11_3(SH2_struct* sh)
{
  SH2movll4(sh, 12,11, 3);
}
#define init_5_12_11_4
void FASTCALL call_5_12_11_4(SH2_struct* sh)
{
  SH2movll4(sh, 12,11, 4);
}
#define init_5_12_11_5
void FASTCALL call_5_12_11_5(SH2_struct* sh)
{
  SH2movll4(sh, 12,11, 5);
}
#define init_5_12_11_6
void FASTCALL call_5_12_11_6(SH2_struct* sh)
{
  SH2movll4(sh, 12,11, 6);
}
#define init_5_12_11_7
void FASTCALL call_5_12_11_7(SH2_struct* sh)
{
  SH2movll4(sh, 12,11, 7);
}
#define init_5_12_11_8
void FASTCALL call_5_12_11_8(SH2_struct* sh)
{
  SH2movll4(sh, 12,11, 8);
}
#define init_5_12_11_9
void FASTCALL call_5_12_11_9(SH2_struct* sh)
{
  SH2movll4(sh, 12,11, 9);
}
#define init_5_12_11_10
void FASTCALL call_5_12_11_10(SH2_struct* sh)
{
  SH2movll4(sh, 12,11, 10);
}
#define init_5_12_11_11
void FASTCALL call_5_12_11_11(SH2_struct* sh)
{
  SH2movll4(sh, 12,11, 11);
}
#define init_5_12_11_12
void FASTCALL call_5_12_11_12(SH2_struct* sh)
{
  SH2movll4(sh, 12,11, 12);
}
#define init_5_12_11_13
void FASTCALL call_5_12_11_13(SH2_struct* sh)
{
  SH2movll4(sh, 12,11, 13);
}
#define init_5_12_11_14
void FASTCALL call_5_12_11_14(SH2_struct* sh)
{
  SH2movll4(sh, 12,11, 14);
}
#define init_5_12_11_15
void FASTCALL call_5_12_11_15(SH2_struct* sh)
{
  SH2movll4(sh, 12,11, 15);
}
#define init_5_12_12_0
void FASTCALL call_5_12_12_0(SH2_struct* sh)
{
  SH2movll4(sh, 12,12, 0);
}
#define init_5_12_12_1
void FASTCALL call_5_12_12_1(SH2_struct* sh)
{
  SH2movll4(sh, 12,12, 1);
}
#define init_5_12_12_2
void FASTCALL call_5_12_12_2(SH2_struct* sh)
{
  SH2movll4(sh, 12,12, 2);
}
#define init_5_12_12_3
void FASTCALL call_5_12_12_3(SH2_struct* sh)
{
  SH2movll4(sh, 12,12, 3);
}
#define init_5_12_12_4
void FASTCALL call_5_12_12_4(SH2_struct* sh)
{
  SH2movll4(sh, 12,12, 4);
}
#define init_5_12_12_5
void FASTCALL call_5_12_12_5(SH2_struct* sh)
{
  SH2movll4(sh, 12,12, 5);
}
#define init_5_12_12_6
void FASTCALL call_5_12_12_6(SH2_struct* sh)
{
  SH2movll4(sh, 12,12, 6);
}
#define init_5_12_12_7
void FASTCALL call_5_12_12_7(SH2_struct* sh)
{
  SH2movll4(sh, 12,12, 7);
}
#define init_5_12_12_8
void FASTCALL call_5_12_12_8(SH2_struct* sh)
{
  SH2movll4(sh, 12,12, 8);
}
#define init_5_12_12_9
void FASTCALL call_5_12_12_9(SH2_struct* sh)
{
  SH2movll4(sh, 12,12, 9);
}
#define init_5_12_12_10
void FASTCALL call_5_12_12_10(SH2_struct* sh)
{
  SH2movll4(sh, 12,12, 10);
}
#define init_5_12_12_11
void FASTCALL call_5_12_12_11(SH2_struct* sh)
{
  SH2movll4(sh, 12,12, 11);
}
#define init_5_12_12_12
void FASTCALL call_5_12_12_12(SH2_struct* sh)
{
  SH2movll4(sh, 12,12, 12);
}
#define init_5_12_12_13
void FASTCALL call_5_12_12_13(SH2_struct* sh)
{
  SH2movll4(sh, 12,12, 13);
}
#define init_5_12_12_14
void FASTCALL call_5_12_12_14(SH2_struct* sh)
{
  SH2movll4(sh, 12,12, 14);
}
#define init_5_12_12_15
void FASTCALL call_5_12_12_15(SH2_struct* sh)
{
  SH2movll4(sh, 12,12, 15);
}
#define init_5_12_13_0
void FASTCALL call_5_12_13_0(SH2_struct* sh)
{
  SH2movll4(sh, 12,13, 0);
}
#define init_5_12_13_1
void FASTCALL call_5_12_13_1(SH2_struct* sh)
{
  SH2movll4(sh, 12,13, 1);
}
#define init_5_12_13_2
void FASTCALL call_5_12_13_2(SH2_struct* sh)
{
  SH2movll4(sh, 12,13, 2);
}
#define init_5_12_13_3
void FASTCALL call_5_12_13_3(SH2_struct* sh)
{
  SH2movll4(sh, 12,13, 3);
}
#define init_5_12_13_4
void FASTCALL call_5_12_13_4(SH2_struct* sh)
{
  SH2movll4(sh, 12,13, 4);
}
#define init_5_12_13_5
void FASTCALL call_5_12_13_5(SH2_struct* sh)
{
  SH2movll4(sh, 12,13, 5);
}
#define init_5_12_13_6
void FASTCALL call_5_12_13_6(SH2_struct* sh)
{
  SH2movll4(sh, 12,13, 6);
}
#define init_5_12_13_7
void FASTCALL call_5_12_13_7(SH2_struct* sh)
{
  SH2movll4(sh, 12,13, 7);
}
#define init_5_12_13_8
void FASTCALL call_5_12_13_8(SH2_struct* sh)
{
  SH2movll4(sh, 12,13, 8);
}
#define init_5_12_13_9
void FASTCALL call_5_12_13_9(SH2_struct* sh)
{
  SH2movll4(sh, 12,13, 9);
}
#define init_5_12_13_10
void FASTCALL call_5_12_13_10(SH2_struct* sh)
{
  SH2movll4(sh, 12,13, 10);
}
#define init_5_12_13_11
void FASTCALL call_5_12_13_11(SH2_struct* sh)
{
  SH2movll4(sh, 12,13, 11);
}
#define init_5_12_13_12
void FASTCALL call_5_12_13_12(SH2_struct* sh)
{
  SH2movll4(sh, 12,13, 12);
}
#define init_5_12_13_13
void FASTCALL call_5_12_13_13(SH2_struct* sh)
{
  SH2movll4(sh, 12,13, 13);
}
#define init_5_12_13_14
void FASTCALL call_5_12_13_14(SH2_struct* sh)
{
  SH2movll4(sh, 12,13, 14);
}
#define init_5_12_13_15
void FASTCALL call_5_12_13_15(SH2_struct* sh)
{
  SH2movll4(sh, 12,13, 15);
}
#define init_5_12_14_0
void FASTCALL call_5_12_14_0(SH2_struct* sh)
{
  SH2movll4(sh, 12,14, 0);
}
#define init_5_12_14_1
void FASTCALL call_5_12_14_1(SH2_struct* sh)
{
  SH2movll4(sh, 12,14, 1);
}
#define init_5_12_14_2
void FASTCALL call_5_12_14_2(SH2_struct* sh)
{
  SH2movll4(sh, 12,14, 2);
}
#define init_5_12_14_3
void FASTCALL call_5_12_14_3(SH2_struct* sh)
{
  SH2movll4(sh, 12,14, 3);
}
#define init_5_12_14_4
void FASTCALL call_5_12_14_4(SH2_struct* sh)
{
  SH2movll4(sh, 12,14, 4);
}
#define init_5_12_14_5
void FASTCALL call_5_12_14_5(SH2_struct* sh)
{
  SH2movll4(sh, 12,14, 5);
}
#define init_5_12_14_6
void FASTCALL call_5_12_14_6(SH2_struct* sh)
{
  SH2movll4(sh, 12,14, 6);
}
#define init_5_12_14_7
void FASTCALL call_5_12_14_7(SH2_struct* sh)
{
  SH2movll4(sh, 12,14, 7);
}
#define init_5_12_14_8
void FASTCALL call_5_12_14_8(SH2_struct* sh)
{
  SH2movll4(sh, 12,14, 8);
}
#define init_5_12_14_9
void FASTCALL call_5_12_14_9(SH2_struct* sh)
{
  SH2movll4(sh, 12,14, 9);
}
#define init_5_12_14_10
void FASTCALL call_5_12_14_10(SH2_struct* sh)
{
  SH2movll4(sh, 12,14, 10);
}
#define init_5_12_14_11
void FASTCALL call_5_12_14_11(SH2_struct* sh)
{
  SH2movll4(sh, 12,14, 11);
}
#define init_5_12_14_12
void FASTCALL call_5_12_14_12(SH2_struct* sh)
{
  SH2movll4(sh, 12,14, 12);
}
#define init_5_12_14_13
void FASTCALL call_5_12_14_13(SH2_struct* sh)
{
  SH2movll4(sh, 12,14, 13);
}
#define init_5_12_14_14
void FASTCALL call_5_12_14_14(SH2_struct* sh)
{
  SH2movll4(sh, 12,14, 14);
}
#define init_5_12_14_15
void FASTCALL call_5_12_14_15(SH2_struct* sh)
{
  SH2movll4(sh, 12,14, 15);
}
#define init_5_12_15_0
void FASTCALL call_5_12_15_0(SH2_struct* sh)
{
  SH2movll4(sh, 12,15, 0);
}
#define init_5_12_15_1
void FASTCALL call_5_12_15_1(SH2_struct* sh)
{
  SH2movll4(sh, 12,15, 1);
}
#define init_5_12_15_2
void FASTCALL call_5_12_15_2(SH2_struct* sh)
{
  SH2movll4(sh, 12,15, 2);
}
#define init_5_12_15_3
void FASTCALL call_5_12_15_3(SH2_struct* sh)
{
  SH2movll4(sh, 12,15, 3);
}
#define init_5_12_15_4
void FASTCALL call_5_12_15_4(SH2_struct* sh)
{
  SH2movll4(sh, 12,15, 4);
}
#define init_5_12_15_5
void FASTCALL call_5_12_15_5(SH2_struct* sh)
{
  SH2movll4(sh, 12,15, 5);
}
#define init_5_12_15_6
void FASTCALL call_5_12_15_6(SH2_struct* sh)
{
  SH2movll4(sh, 12,15, 6);
}
#define init_5_12_15_7
void FASTCALL call_5_12_15_7(SH2_struct* sh)
{
  SH2movll4(sh, 12,15, 7);
}
#define init_5_12_15_8
void FASTCALL call_5_12_15_8(SH2_struct* sh)
{
  SH2movll4(sh, 12,15, 8);
}
#define init_5_12_15_9
void FASTCALL call_5_12_15_9(SH2_struct* sh)
{
  SH2movll4(sh, 12,15, 9);
}
#define init_5_12_15_10
void FASTCALL call_5_12_15_10(SH2_struct* sh)
{
  SH2movll4(sh, 12,15, 10);
}
#define init_5_12_15_11
void FASTCALL call_5_12_15_11(SH2_struct* sh)
{
  SH2movll4(sh, 12,15, 11);
}
#define init_5_12_15_12
void FASTCALL call_5_12_15_12(SH2_struct* sh)
{
  SH2movll4(sh, 12,15, 12);
}
#define init_5_12_15_13
void FASTCALL call_5_12_15_13(SH2_struct* sh)
{
  SH2movll4(sh, 12,15, 13);
}
#define init_5_12_15_14
void FASTCALL call_5_12_15_14(SH2_struct* sh)
{
  SH2movll4(sh, 12,15, 14);
}
#define init_5_12_15_15
void FASTCALL call_5_12_15_15(SH2_struct* sh)
{
  SH2movll4(sh, 12,15, 15);
}
#define init_5_13_0_0
void FASTCALL call_5_13_0_0(SH2_struct* sh)
{
  SH2movll4(sh, 13,0, 0);
}
#define init_5_13_0_1
void FASTCALL call_5_13_0_1(SH2_struct* sh)
{
  SH2movll4(sh, 13,0, 1);
}
#define init_5_13_0_2
void FASTCALL call_5_13_0_2(SH2_struct* sh)
{
  SH2movll4(sh, 13,0, 2);
}
#define init_5_13_0_3
void FASTCALL call_5_13_0_3(SH2_struct* sh)
{
  SH2movll4(sh, 13,0, 3);
}
#define init_5_13_0_4
void FASTCALL call_5_13_0_4(SH2_struct* sh)
{
  SH2movll4(sh, 13,0, 4);
}
#define init_5_13_0_5
void FASTCALL call_5_13_0_5(SH2_struct* sh)
{
  SH2movll4(sh, 13,0, 5);
}
#define init_5_13_0_6
void FASTCALL call_5_13_0_6(SH2_struct* sh)
{
  SH2movll4(sh, 13,0, 6);
}
#define init_5_13_0_7
void FASTCALL call_5_13_0_7(SH2_struct* sh)
{
  SH2movll4(sh, 13,0, 7);
}
#define init_5_13_0_8
void FASTCALL call_5_13_0_8(SH2_struct* sh)
{
  SH2movll4(sh, 13,0, 8);
}
#define init_5_13_0_9
void FASTCALL call_5_13_0_9(SH2_struct* sh)
{
  SH2movll4(sh, 13,0, 9);
}
#define init_5_13_0_10
void FASTCALL call_5_13_0_10(SH2_struct* sh)
{
  SH2movll4(sh, 13,0, 10);
}
#define init_5_13_0_11
void FASTCALL call_5_13_0_11(SH2_struct* sh)
{
  SH2movll4(sh, 13,0, 11);
}
#define init_5_13_0_12
void FASTCALL call_5_13_0_12(SH2_struct* sh)
{
  SH2movll4(sh, 13,0, 12);
}
#define init_5_13_0_13
void FASTCALL call_5_13_0_13(SH2_struct* sh)
{
  SH2movll4(sh, 13,0, 13);
}
#define init_5_13_0_14
void FASTCALL call_5_13_0_14(SH2_struct* sh)
{
  SH2movll4(sh, 13,0, 14);
}
#define init_5_13_0_15
void FASTCALL call_5_13_0_15(SH2_struct* sh)
{
  SH2movll4(sh, 13,0, 15);
}
#define init_5_13_1_0
void FASTCALL call_5_13_1_0(SH2_struct* sh)
{
  SH2movll4(sh, 13,1, 0);
}
#define init_5_13_1_1
void FASTCALL call_5_13_1_1(SH2_struct* sh)
{
  SH2movll4(sh, 13,1, 1);
}
#define init_5_13_1_2
void FASTCALL call_5_13_1_2(SH2_struct* sh)
{
  SH2movll4(sh, 13,1, 2);
}
#define init_5_13_1_3
void FASTCALL call_5_13_1_3(SH2_struct* sh)
{
  SH2movll4(sh, 13,1, 3);
}
#define init_5_13_1_4
void FASTCALL call_5_13_1_4(SH2_struct* sh)
{
  SH2movll4(sh, 13,1, 4);
}
#define init_5_13_1_5
void FASTCALL call_5_13_1_5(SH2_struct* sh)
{
  SH2movll4(sh, 13,1, 5);
}
#define init_5_13_1_6
void FASTCALL call_5_13_1_6(SH2_struct* sh)
{
  SH2movll4(sh, 13,1, 6);
}
#define init_5_13_1_7
void FASTCALL call_5_13_1_7(SH2_struct* sh)
{
  SH2movll4(sh, 13,1, 7);
}
#define init_5_13_1_8
void FASTCALL call_5_13_1_8(SH2_struct* sh)
{
  SH2movll4(sh, 13,1, 8);
}
#define init_5_13_1_9
void FASTCALL call_5_13_1_9(SH2_struct* sh)
{
  SH2movll4(sh, 13,1, 9);
}
#define init_5_13_1_10
void FASTCALL call_5_13_1_10(SH2_struct* sh)
{
  SH2movll4(sh, 13,1, 10);
}
#define init_5_13_1_11
void FASTCALL call_5_13_1_11(SH2_struct* sh)
{
  SH2movll4(sh, 13,1, 11);
}
#define init_5_13_1_12
void FASTCALL call_5_13_1_12(SH2_struct* sh)
{
  SH2movll4(sh, 13,1, 12);
}
#define init_5_13_1_13
void FASTCALL call_5_13_1_13(SH2_struct* sh)
{
  SH2movll4(sh, 13,1, 13);
}
#define init_5_13_1_14
void FASTCALL call_5_13_1_14(SH2_struct* sh)
{
  SH2movll4(sh, 13,1, 14);
}
#define init_5_13_1_15
void FASTCALL call_5_13_1_15(SH2_struct* sh)
{
  SH2movll4(sh, 13,1, 15);
}
#define init_5_13_2_0
void FASTCALL call_5_13_2_0(SH2_struct* sh)
{
  SH2movll4(sh, 13,2, 0);
}
#define init_5_13_2_1
void FASTCALL call_5_13_2_1(SH2_struct* sh)
{
  SH2movll4(sh, 13,2, 1);
}
#define init_5_13_2_2
void FASTCALL call_5_13_2_2(SH2_struct* sh)
{
  SH2movll4(sh, 13,2, 2);
}
#define init_5_13_2_3
void FASTCALL call_5_13_2_3(SH2_struct* sh)
{
  SH2movll4(sh, 13,2, 3);
}
#define init_5_13_2_4
void FASTCALL call_5_13_2_4(SH2_struct* sh)
{
  SH2movll4(sh, 13,2, 4);
}
#define init_5_13_2_5
void FASTCALL call_5_13_2_5(SH2_struct* sh)
{
  SH2movll4(sh, 13,2, 5);
}
#define init_5_13_2_6
void FASTCALL call_5_13_2_6(SH2_struct* sh)
{
  SH2movll4(sh, 13,2, 6);
}
#define init_5_13_2_7
void FASTCALL call_5_13_2_7(SH2_struct* sh)
{
  SH2movll4(sh, 13,2, 7);
}
#define init_5_13_2_8
void FASTCALL call_5_13_2_8(SH2_struct* sh)
{
  SH2movll4(sh, 13,2, 8);
}
#define init_5_13_2_9
void FASTCALL call_5_13_2_9(SH2_struct* sh)
{
  SH2movll4(sh, 13,2, 9);
}
#define init_5_13_2_10
void FASTCALL call_5_13_2_10(SH2_struct* sh)
{
  SH2movll4(sh, 13,2, 10);
}
#define init_5_13_2_11
void FASTCALL call_5_13_2_11(SH2_struct* sh)
{
  SH2movll4(sh, 13,2, 11);
}
#define init_5_13_2_12
void FASTCALL call_5_13_2_12(SH2_struct* sh)
{
  SH2movll4(sh, 13,2, 12);
}
#define init_5_13_2_13
void FASTCALL call_5_13_2_13(SH2_struct* sh)
{
  SH2movll4(sh, 13,2, 13);
}
#define init_5_13_2_14
void FASTCALL call_5_13_2_14(SH2_struct* sh)
{
  SH2movll4(sh, 13,2, 14);
}
#define init_5_13_2_15
void FASTCALL call_5_13_2_15(SH2_struct* sh)
{
  SH2movll4(sh, 13,2, 15);
}
#define init_5_13_3_0
void FASTCALL call_5_13_3_0(SH2_struct* sh)
{
  SH2movll4(sh, 13,3, 0);
}
#define init_5_13_3_1
void FASTCALL call_5_13_3_1(SH2_struct* sh)
{
  SH2movll4(sh, 13,3, 1);
}
#define init_5_13_3_2
void FASTCALL call_5_13_3_2(SH2_struct* sh)
{
  SH2movll4(sh, 13,3, 2);
}
#define init_5_13_3_3
void FASTCALL call_5_13_3_3(SH2_struct* sh)
{
  SH2movll4(sh, 13,3, 3);
}
#define init_5_13_3_4
void FASTCALL call_5_13_3_4(SH2_struct* sh)
{
  SH2movll4(sh, 13,3, 4);
}
#define init_5_13_3_5
void FASTCALL call_5_13_3_5(SH2_struct* sh)
{
  SH2movll4(sh, 13,3, 5);
}
#define init_5_13_3_6
void FASTCALL call_5_13_3_6(SH2_struct* sh)
{
  SH2movll4(sh, 13,3, 6);
}
#define init_5_13_3_7
void FASTCALL call_5_13_3_7(SH2_struct* sh)
{
  SH2movll4(sh, 13,3, 7);
}
#define init_5_13_3_8
void FASTCALL call_5_13_3_8(SH2_struct* sh)
{
  SH2movll4(sh, 13,3, 8);
}
#define init_5_13_3_9
void FASTCALL call_5_13_3_9(SH2_struct* sh)
{
  SH2movll4(sh, 13,3, 9);
}
#define init_5_13_3_10
void FASTCALL call_5_13_3_10(SH2_struct* sh)
{
  SH2movll4(sh, 13,3, 10);
}
#define init_5_13_3_11
void FASTCALL call_5_13_3_11(SH2_struct* sh)
{
  SH2movll4(sh, 13,3, 11);
}
#define init_5_13_3_12
void FASTCALL call_5_13_3_12(SH2_struct* sh)
{
  SH2movll4(sh, 13,3, 12);
}
#define init_5_13_3_13
void FASTCALL call_5_13_3_13(SH2_struct* sh)
{
  SH2movll4(sh, 13,3, 13);
}
#define init_5_13_3_14
void FASTCALL call_5_13_3_14(SH2_struct* sh)
{
  SH2movll4(sh, 13,3, 14);
}
#define init_5_13_3_15
void FASTCALL call_5_13_3_15(SH2_struct* sh)
{
  SH2movll4(sh, 13,3, 15);
}
#define init_5_13_4_0
void FASTCALL call_5_13_4_0(SH2_struct* sh)
{
  SH2movll4(sh, 13,4, 0);
}
#define init_5_13_4_1
void FASTCALL call_5_13_4_1(SH2_struct* sh)
{
  SH2movll4(sh, 13,4, 1);
}
#define init_5_13_4_2
void FASTCALL call_5_13_4_2(SH2_struct* sh)
{
  SH2movll4(sh, 13,4, 2);
}
#define init_5_13_4_3
void FASTCALL call_5_13_4_3(SH2_struct* sh)
{
  SH2movll4(sh, 13,4, 3);
}
#define init_5_13_4_4
void FASTCALL call_5_13_4_4(SH2_struct* sh)
{
  SH2movll4(sh, 13,4, 4);
}
#define init_5_13_4_5
void FASTCALL call_5_13_4_5(SH2_struct* sh)
{
  SH2movll4(sh, 13,4, 5);
}
#define init_5_13_4_6
void FASTCALL call_5_13_4_6(SH2_struct* sh)
{
  SH2movll4(sh, 13,4, 6);
}
#define init_5_13_4_7
void FASTCALL call_5_13_4_7(SH2_struct* sh)
{
  SH2movll4(sh, 13,4, 7);
}
#define init_5_13_4_8
void FASTCALL call_5_13_4_8(SH2_struct* sh)
{
  SH2movll4(sh, 13,4, 8);
}
#define init_5_13_4_9
void FASTCALL call_5_13_4_9(SH2_struct* sh)
{
  SH2movll4(sh, 13,4, 9);
}
#define init_5_13_4_10
void FASTCALL call_5_13_4_10(SH2_struct* sh)
{
  SH2movll4(sh, 13,4, 10);
}
#define init_5_13_4_11
void FASTCALL call_5_13_4_11(SH2_struct* sh)
{
  SH2movll4(sh, 13,4, 11);
}
#define init_5_13_4_12
void FASTCALL call_5_13_4_12(SH2_struct* sh)
{
  SH2movll4(sh, 13,4, 12);
}
#define init_5_13_4_13
void FASTCALL call_5_13_4_13(SH2_struct* sh)
{
  SH2movll4(sh, 13,4, 13);
}
#define init_5_13_4_14
void FASTCALL call_5_13_4_14(SH2_struct* sh)
{
  SH2movll4(sh, 13,4, 14);
}
#define init_5_13_4_15
void FASTCALL call_5_13_4_15(SH2_struct* sh)
{
  SH2movll4(sh, 13,4, 15);
}
#define init_5_13_5_0
void FASTCALL call_5_13_5_0(SH2_struct* sh)
{
  SH2movll4(sh, 13,5, 0);
}
#define init_5_13_5_1
void FASTCALL call_5_13_5_1(SH2_struct* sh)
{
  SH2movll4(sh, 13,5, 1);
}
#define init_5_13_5_2
void FASTCALL call_5_13_5_2(SH2_struct* sh)
{
  SH2movll4(sh, 13,5, 2);
}
#define init_5_13_5_3
void FASTCALL call_5_13_5_3(SH2_struct* sh)
{
  SH2movll4(sh, 13,5, 3);
}
#define init_5_13_5_4
void FASTCALL call_5_13_5_4(SH2_struct* sh)
{
  SH2movll4(sh, 13,5, 4);
}
#define init_5_13_5_5
void FASTCALL call_5_13_5_5(SH2_struct* sh)
{
  SH2movll4(sh, 13,5, 5);
}
#define init_5_13_5_6
void FASTCALL call_5_13_5_6(SH2_struct* sh)
{
  SH2movll4(sh, 13,5, 6);
}
#define init_5_13_5_7
void FASTCALL call_5_13_5_7(SH2_struct* sh)
{
  SH2movll4(sh, 13,5, 7);
}
#define init_5_13_5_8
void FASTCALL call_5_13_5_8(SH2_struct* sh)
{
  SH2movll4(sh, 13,5, 8);
}
#define init_5_13_5_9
void FASTCALL call_5_13_5_9(SH2_struct* sh)
{
  SH2movll4(sh, 13,5, 9);
}
#define init_5_13_5_10
void FASTCALL call_5_13_5_10(SH2_struct* sh)
{
  SH2movll4(sh, 13,5, 10);
}
#define init_5_13_5_11
void FASTCALL call_5_13_5_11(SH2_struct* sh)
{
  SH2movll4(sh, 13,5, 11);
}
#define init_5_13_5_12
void FASTCALL call_5_13_5_12(SH2_struct* sh)
{
  SH2movll4(sh, 13,5, 12);
}
#define init_5_13_5_13
void FASTCALL call_5_13_5_13(SH2_struct* sh)
{
  SH2movll4(sh, 13,5, 13);
}
#define init_5_13_5_14
void FASTCALL call_5_13_5_14(SH2_struct* sh)
{
  SH2movll4(sh, 13,5, 14);
}
#define init_5_13_5_15
void FASTCALL call_5_13_5_15(SH2_struct* sh)
{
  SH2movll4(sh, 13,5, 15);
}
#define init_5_13_6_0
void FASTCALL call_5_13_6_0(SH2_struct* sh)
{
  SH2movll4(sh, 13,6, 0);
}
#define init_5_13_6_1
void FASTCALL call_5_13_6_1(SH2_struct* sh)
{
  SH2movll4(sh, 13,6, 1);
}
#define init_5_13_6_2
void FASTCALL call_5_13_6_2(SH2_struct* sh)
{
  SH2movll4(sh, 13,6, 2);
}
#define init_5_13_6_3
void FASTCALL call_5_13_6_3(SH2_struct* sh)
{
  SH2movll4(sh, 13,6, 3);
}
#define init_5_13_6_4
void FASTCALL call_5_13_6_4(SH2_struct* sh)
{
  SH2movll4(sh, 13,6, 4);
}
#define init_5_13_6_5
void FASTCALL call_5_13_6_5(SH2_struct* sh)
{
  SH2movll4(sh, 13,6, 5);
}
#define init_5_13_6_6
void FASTCALL call_5_13_6_6(SH2_struct* sh)
{
  SH2movll4(sh, 13,6, 6);
}
#define init_5_13_6_7
void FASTCALL call_5_13_6_7(SH2_struct* sh)
{
  SH2movll4(sh, 13,6, 7);
}
#define init_5_13_6_8
void FASTCALL call_5_13_6_8(SH2_struct* sh)
{
  SH2movll4(sh, 13,6, 8);
}
#define init_5_13_6_9
void FASTCALL call_5_13_6_9(SH2_struct* sh)
{
  SH2movll4(sh, 13,6, 9);
}
#define init_5_13_6_10
void FASTCALL call_5_13_6_10(SH2_struct* sh)
{
  SH2movll4(sh, 13,6, 10);
}
#define init_5_13_6_11
void FASTCALL call_5_13_6_11(SH2_struct* sh)
{
  SH2movll4(sh, 13,6, 11);
}
#define init_5_13_6_12
void FASTCALL call_5_13_6_12(SH2_struct* sh)
{
  SH2movll4(sh, 13,6, 12);
}
#define init_5_13_6_13
void FASTCALL call_5_13_6_13(SH2_struct* sh)
{
  SH2movll4(sh, 13,6, 13);
}
#define init_5_13_6_14
void FASTCALL call_5_13_6_14(SH2_struct* sh)
{
  SH2movll4(sh, 13,6, 14);
}
#define init_5_13_6_15
void FASTCALL call_5_13_6_15(SH2_struct* sh)
{
  SH2movll4(sh, 13,6, 15);
}
#define init_5_13_7_0
void FASTCALL call_5_13_7_0(SH2_struct* sh)
{
  SH2movll4(sh, 13,7, 0);
}
#define init_5_13_7_1
void FASTCALL call_5_13_7_1(SH2_struct* sh)
{
  SH2movll4(sh, 13,7, 1);
}
#define init_5_13_7_2
void FASTCALL call_5_13_7_2(SH2_struct* sh)
{
  SH2movll4(sh, 13,7, 2);
}
#define init_5_13_7_3
void FASTCALL call_5_13_7_3(SH2_struct* sh)
{
  SH2movll4(sh, 13,7, 3);
}
#define init_5_13_7_4
void FASTCALL call_5_13_7_4(SH2_struct* sh)
{
  SH2movll4(sh, 13,7, 4);
}
#define init_5_13_7_5
void FASTCALL call_5_13_7_5(SH2_struct* sh)
{
  SH2movll4(sh, 13,7, 5);
}
#define init_5_13_7_6
void FASTCALL call_5_13_7_6(SH2_struct* sh)
{
  SH2movll4(sh, 13,7, 6);
}
#define init_5_13_7_7
void FASTCALL call_5_13_7_7(SH2_struct* sh)
{
  SH2movll4(sh, 13,7, 7);
}
#define init_5_13_7_8
void FASTCALL call_5_13_7_8(SH2_struct* sh)
{
  SH2movll4(sh, 13,7, 8);
}
#define init_5_13_7_9
void FASTCALL call_5_13_7_9(SH2_struct* sh)
{
  SH2movll4(sh, 13,7, 9);
}
#define init_5_13_7_10
void FASTCALL call_5_13_7_10(SH2_struct* sh)
{
  SH2movll4(sh, 13,7, 10);
}
#define init_5_13_7_11
void FASTCALL call_5_13_7_11(SH2_struct* sh)
{
  SH2movll4(sh, 13,7, 11);
}
#define init_5_13_7_12
void FASTCALL call_5_13_7_12(SH2_struct* sh)
{
  SH2movll4(sh, 13,7, 12);
}
#define init_5_13_7_13
void FASTCALL call_5_13_7_13(SH2_struct* sh)
{
  SH2movll4(sh, 13,7, 13);
}
#define init_5_13_7_14
void FASTCALL call_5_13_7_14(SH2_struct* sh)
{
  SH2movll4(sh, 13,7, 14);
}
#define init_5_13_7_15
void FASTCALL call_5_13_7_15(SH2_struct* sh)
{
  SH2movll4(sh, 13,7, 15);
}
#define init_5_13_8_0
void FASTCALL call_5_13_8_0(SH2_struct* sh)
{
  SH2movll4(sh, 13,8, 0);
}
#define init_5_13_8_1
void FASTCALL call_5_13_8_1(SH2_struct* sh)
{
  SH2movll4(sh, 13,8, 1);
}
#define init_5_13_8_2
void FASTCALL call_5_13_8_2(SH2_struct* sh)
{
  SH2movll4(sh, 13,8, 2);
}
#define init_5_13_8_3
void FASTCALL call_5_13_8_3(SH2_struct* sh)
{
  SH2movll4(sh, 13,8, 3);
}
#define init_5_13_8_4
void FASTCALL call_5_13_8_4(SH2_struct* sh)
{
  SH2movll4(sh, 13,8, 4);
}
#define init_5_13_8_5
void FASTCALL call_5_13_8_5(SH2_struct* sh)
{
  SH2movll4(sh, 13,8, 5);
}
#define init_5_13_8_6
void FASTCALL call_5_13_8_6(SH2_struct* sh)
{
  SH2movll4(sh, 13,8, 6);
}
#define init_5_13_8_7
void FASTCALL call_5_13_8_7(SH2_struct* sh)
{
  SH2movll4(sh, 13,8, 7);
}
#define init_5_13_8_8
void FASTCALL call_5_13_8_8(SH2_struct* sh)
{
  SH2movll4(sh, 13,8, 8);
}
#define init_5_13_8_9
void FASTCALL call_5_13_8_9(SH2_struct* sh)
{
  SH2movll4(sh, 13,8, 9);
}
#define init_5_13_8_10
void FASTCALL call_5_13_8_10(SH2_struct* sh)
{
  SH2movll4(sh, 13,8, 10);
}
#define init_5_13_8_11
void FASTCALL call_5_13_8_11(SH2_struct* sh)
{
  SH2movll4(sh, 13,8, 11);
}
#define init_5_13_8_12
void FASTCALL call_5_13_8_12(SH2_struct* sh)
{
  SH2movll4(sh, 13,8, 12);
}
#define init_5_13_8_13
void FASTCALL call_5_13_8_13(SH2_struct* sh)
{
  SH2movll4(sh, 13,8, 13);
}
#define init_5_13_8_14
void FASTCALL call_5_13_8_14(SH2_struct* sh)
{
  SH2movll4(sh, 13,8, 14);
}
#define init_5_13_8_15
void FASTCALL call_5_13_8_15(SH2_struct* sh)
{
  SH2movll4(sh, 13,8, 15);
}
#define init_5_13_9_0
void FASTCALL call_5_13_9_0(SH2_struct* sh)
{
  SH2movll4(sh, 13,9, 0);
}
#define init_5_13_9_1
void FASTCALL call_5_13_9_1(SH2_struct* sh)
{
  SH2movll4(sh, 13,9, 1);
}
#define init_5_13_9_2
void FASTCALL call_5_13_9_2(SH2_struct* sh)
{
  SH2movll4(sh, 13,9, 2);
}
#define init_5_13_9_3
void FASTCALL call_5_13_9_3(SH2_struct* sh)
{
  SH2movll4(sh, 13,9, 3);
}
#define init_5_13_9_4
void FASTCALL call_5_13_9_4(SH2_struct* sh)
{
  SH2movll4(sh, 13,9, 4);
}
#define init_5_13_9_5
void FASTCALL call_5_13_9_5(SH2_struct* sh)
{
  SH2movll4(sh, 13,9, 5);
}
#define init_5_13_9_6
void FASTCALL call_5_13_9_6(SH2_struct* sh)
{
  SH2movll4(sh, 13,9, 6);
}
#define init_5_13_9_7
void FASTCALL call_5_13_9_7(SH2_struct* sh)
{
  SH2movll4(sh, 13,9, 7);
}
#define init_5_13_9_8
void FASTCALL call_5_13_9_8(SH2_struct* sh)
{
  SH2movll4(sh, 13,9, 8);
}
#define init_5_13_9_9
void FASTCALL call_5_13_9_9(SH2_struct* sh)
{
  SH2movll4(sh, 13,9, 9);
}
#define init_5_13_9_10
void FASTCALL call_5_13_9_10(SH2_struct* sh)
{
  SH2movll4(sh, 13,9, 10);
}
#define init_5_13_9_11
void FASTCALL call_5_13_9_11(SH2_struct* sh)
{
  SH2movll4(sh, 13,9, 11);
}
#define init_5_13_9_12
void FASTCALL call_5_13_9_12(SH2_struct* sh)
{
  SH2movll4(sh, 13,9, 12);
}
#define init_5_13_9_13
void FASTCALL call_5_13_9_13(SH2_struct* sh)
{
  SH2movll4(sh, 13,9, 13);
}
#define init_5_13_9_14
void FASTCALL call_5_13_9_14(SH2_struct* sh)
{
  SH2movll4(sh, 13,9, 14);
}
#define init_5_13_9_15
void FASTCALL call_5_13_9_15(SH2_struct* sh)
{
  SH2movll4(sh, 13,9, 15);
}
#define init_5_13_10_0
void FASTCALL call_5_13_10_0(SH2_struct* sh)
{
  SH2movll4(sh, 13,10, 0);
}
#define init_5_13_10_1
void FASTCALL call_5_13_10_1(SH2_struct* sh)
{
  SH2movll4(sh, 13,10, 1);
}
#define init_5_13_10_2
void FASTCALL call_5_13_10_2(SH2_struct* sh)
{
  SH2movll4(sh, 13,10, 2);
}
#define init_5_13_10_3
void FASTCALL call_5_13_10_3(SH2_struct* sh)
{
  SH2movll4(sh, 13,10, 3);
}
#define init_5_13_10_4
void FASTCALL call_5_13_10_4(SH2_struct* sh)
{
  SH2movll4(sh, 13,10, 4);
}
#define init_5_13_10_5
void FASTCALL call_5_13_10_5(SH2_struct* sh)
{
  SH2movll4(sh, 13,10, 5);
}
#define init_5_13_10_6
void FASTCALL call_5_13_10_6(SH2_struct* sh)
{
  SH2movll4(sh, 13,10, 6);
}
#define init_5_13_10_7
void FASTCALL call_5_13_10_7(SH2_struct* sh)
{
  SH2movll4(sh, 13,10, 7);
}
#define init_5_13_10_8
void FASTCALL call_5_13_10_8(SH2_struct* sh)
{
  SH2movll4(sh, 13,10, 8);
}
#define init_5_13_10_9
void FASTCALL call_5_13_10_9(SH2_struct* sh)
{
  SH2movll4(sh, 13,10, 9);
}
#define init_5_13_10_10
void FASTCALL call_5_13_10_10(SH2_struct* sh)
{
  SH2movll4(sh, 13,10, 10);
}
#define init_5_13_10_11
void FASTCALL call_5_13_10_11(SH2_struct* sh)
{
  SH2movll4(sh, 13,10, 11);
}
#define init_5_13_10_12
void FASTCALL call_5_13_10_12(SH2_struct* sh)
{
  SH2movll4(sh, 13,10, 12);
}
#define init_5_13_10_13
void FASTCALL call_5_13_10_13(SH2_struct* sh)
{
  SH2movll4(sh, 13,10, 13);
}
#define init_5_13_10_14
void FASTCALL call_5_13_10_14(SH2_struct* sh)
{
  SH2movll4(sh, 13,10, 14);
}
#define init_5_13_10_15
void FASTCALL call_5_13_10_15(SH2_struct* sh)
{
  SH2movll4(sh, 13,10, 15);
}
#define init_5_13_11_0
void FASTCALL call_5_13_11_0(SH2_struct* sh)
{
  SH2movll4(sh, 13,11, 0);
}
#define init_5_13_11_1
void FASTCALL call_5_13_11_1(SH2_struct* sh)
{
  SH2movll4(sh, 13,11, 1);
}
#define init_5_13_11_2
void FASTCALL call_5_13_11_2(SH2_struct* sh)
{
  SH2movll4(sh, 13,11, 2);
}
#define init_5_13_11_3
void FASTCALL call_5_13_11_3(SH2_struct* sh)
{
  SH2movll4(sh, 13,11, 3);
}
#define init_5_13_11_4
void FASTCALL call_5_13_11_4(SH2_struct* sh)
{
  SH2movll4(sh, 13,11, 4);
}
#define init_5_13_11_5
void FASTCALL call_5_13_11_5(SH2_struct* sh)
{
  SH2movll4(sh, 13,11, 5);
}
#define init_5_13_11_6
void FASTCALL call_5_13_11_6(SH2_struct* sh)
{
  SH2movll4(sh, 13,11, 6);
}
#define init_5_13_11_7
void FASTCALL call_5_13_11_7(SH2_struct* sh)
{
  SH2movll4(sh, 13,11, 7);
}
#define init_5_13_11_8
void FASTCALL call_5_13_11_8(SH2_struct* sh)
{
  SH2movll4(sh, 13,11, 8);
}
#define init_5_13_11_9
void FASTCALL call_5_13_11_9(SH2_struct* sh)
{
  SH2movll4(sh, 13,11, 9);
}
#define init_5_13_11_10
void FASTCALL call_5_13_11_10(SH2_struct* sh)
{
  SH2movll4(sh, 13,11, 10);
}
#define init_5_13_11_11
void FASTCALL call_5_13_11_11(SH2_struct* sh)
{
  SH2movll4(sh, 13,11, 11);
}
#define init_5_13_11_12
void FASTCALL call_5_13_11_12(SH2_struct* sh)
{
  SH2movll4(sh, 13,11, 12);
}
#define init_5_13_11_13
void FASTCALL call_5_13_11_13(SH2_struct* sh)
{
  SH2movll4(sh, 13,11, 13);
}
#define init_5_13_11_14
void FASTCALL call_5_13_11_14(SH2_struct* sh)
{
  SH2movll4(sh, 13,11, 14);
}
#define init_5_13_11_15
void FASTCALL call_5_13_11_15(SH2_struct* sh)
{
  SH2movll4(sh, 13,11, 15);
}
#define init_5_13_12_0
void FASTCALL call_5_13_12_0(SH2_struct* sh)
{
  SH2movll4(sh, 13,12, 0);
}
#define init_5_13_12_1
void FASTCALL call_5_13_12_1(SH2_struct* sh)
{
  SH2movll4(sh, 13,12, 1);
}
#define init_5_13_12_2
void FASTCALL call_5_13_12_2(SH2_struct* sh)
{
  SH2movll4(sh, 13,12, 2);
}
#define init_5_13_12_3
void FASTCALL call_5_13_12_3(SH2_struct* sh)
{
  SH2movll4(sh, 13,12, 3);
}
#define init_5_13_12_4
void FASTCALL call_5_13_12_4(SH2_struct* sh)
{
  SH2movll4(sh, 13,12, 4);
}
#define init_5_13_12_5
void FASTCALL call_5_13_12_5(SH2_struct* sh)
{
  SH2movll4(sh, 13,12, 5);
}
#define init_5_13_12_6
void FASTCALL call_5_13_12_6(SH2_struct* sh)
{
  SH2movll4(sh, 13,12, 6);
}
#define init_5_13_12_7
void FASTCALL call_5_13_12_7(SH2_struct* sh)
{
  SH2movll4(sh, 13,12, 7);
}
#define init_5_13_12_8
void FASTCALL call_5_13_12_8(SH2_struct* sh)
{
  SH2movll4(sh, 13,12, 8);
}
#define init_5_13_12_9
void FASTCALL call_5_13_12_9(SH2_struct* sh)
{
  SH2movll4(sh, 13,12, 9);
}
#define init_5_13_12_10
void FASTCALL call_5_13_12_10(SH2_struct* sh)
{
  SH2movll4(sh, 13,12, 10);
}
#define init_5_13_12_11
void FASTCALL call_5_13_12_11(SH2_struct* sh)
{
  SH2movll4(sh, 13,12, 11);
}
#define init_5_13_12_12
void FASTCALL call_5_13_12_12(SH2_struct* sh)
{
  SH2movll4(sh, 13,12, 12);
}
#define init_5_13_12_13
void FASTCALL call_5_13_12_13(SH2_struct* sh)
{
  SH2movll4(sh, 13,12, 13);
}
#define init_5_13_12_14
void FASTCALL call_5_13_12_14(SH2_struct* sh)
{
  SH2movll4(sh, 13,12, 14);
}
#define init_5_13_12_15
void FASTCALL call_5_13_12_15(SH2_struct* sh)
{
  SH2movll4(sh, 13,12, 15);
}
#define init_5_13_13_0
void FASTCALL call_5_13_13_0(SH2_struct* sh)
{
  SH2movll4(sh, 13,13, 0);
}
#define init_5_13_13_1
void FASTCALL call_5_13_13_1(SH2_struct* sh)
{
  SH2movll4(sh, 13,13, 1);
}
#define init_5_13_13_2
void FASTCALL call_5_13_13_2(SH2_struct* sh)
{
  SH2movll4(sh, 13,13, 2);
}
#define init_5_13_13_3
void FASTCALL call_5_13_13_3(SH2_struct* sh)
{
  SH2movll4(sh, 13,13, 3);
}
#define init_5_13_13_4
void FASTCALL call_5_13_13_4(SH2_struct* sh)
{
  SH2movll4(sh, 13,13, 4);
}
#define init_5_13_13_5
void FASTCALL call_5_13_13_5(SH2_struct* sh)
{
  SH2movll4(sh, 13,13, 5);
}
#define init_5_13_13_6
void FASTCALL call_5_13_13_6(SH2_struct* sh)
{
  SH2movll4(sh, 13,13, 6);
}
#define init_5_13_13_7
void FASTCALL call_5_13_13_7(SH2_struct* sh)
{
  SH2movll4(sh, 13,13, 7);
}
#define init_5_13_13_8
void FASTCALL call_5_13_13_8(SH2_struct* sh)
{
  SH2movll4(sh, 13,13, 8);
}
#define init_5_13_13_9
void FASTCALL call_5_13_13_9(SH2_struct* sh)
{
  SH2movll4(sh, 13,13, 9);
}
#define init_5_13_13_10
void FASTCALL call_5_13_13_10(SH2_struct* sh)
{
  SH2movll4(sh, 13,13, 10);
}
#define init_5_13_13_11
void FASTCALL call_5_13_13_11(SH2_struct* sh)
{
  SH2movll4(sh, 13,13, 11);
}
#define init_5_13_13_12
void FASTCALL call_5_13_13_12(SH2_struct* sh)
{
  SH2movll4(sh, 13,13, 12);
}
#define init_5_13_13_13
void FASTCALL call_5_13_13_13(SH2_struct* sh)
{
  SH2movll4(sh, 13,13, 13);
}
#define init_5_13_13_14
void FASTCALL call_5_13_13_14(SH2_struct* sh)
{
  SH2movll4(sh, 13,13, 14);
}
#define init_5_13_13_15
void FASTCALL call_5_13_13_15(SH2_struct* sh)
{
  SH2movll4(sh, 13,13, 15);
}
#define init_5_13_14_0
void FASTCALL call_5_13_14_0(SH2_struct* sh)
{
  SH2movll4(sh, 13,14, 0);
}
#define init_5_13_14_1
void FASTCALL call_5_13_14_1(SH2_struct* sh)
{
  SH2movll4(sh, 13,14, 1);
}
#define init_5_13_14_2
void FASTCALL call_5_13_14_2(SH2_struct* sh)
{
  SH2movll4(sh, 13,14, 2);
}
#define init_5_13_14_3
void FASTCALL call_5_13_14_3(SH2_struct* sh)
{
  SH2movll4(sh, 13,14, 3);
}
#define init_5_13_14_4
void FASTCALL call_5_13_14_4(SH2_struct* sh)
{
  SH2movll4(sh, 13,14, 4);
}
#define init_5_13_14_5
void FASTCALL call_5_13_14_5(SH2_struct* sh)
{
  SH2movll4(sh, 13,14, 5);
}
#define init_5_13_14_6
void FASTCALL call_5_13_14_6(SH2_struct* sh)
{
  SH2movll4(sh, 13,14, 6);
}
#define init_5_13_14_7
void FASTCALL call_5_13_14_7(SH2_struct* sh)
{
  SH2movll4(sh, 13,14, 7);
}
#define init_5_13_14_8
void FASTCALL call_5_13_14_8(SH2_struct* sh)
{
  SH2movll4(sh, 13,14, 8);
}
#define init_5_13_14_9
void FASTCALL call_5_13_14_9(SH2_struct* sh)
{
  SH2movll4(sh, 13,14, 9);
}
#define init_5_13_14_10
void FASTCALL call_5_13_14_10(SH2_struct* sh)
{
  SH2movll4(sh, 13,14, 10);
}
#define init_5_13_14_11
void FASTCALL call_5_13_14_11(SH2_struct* sh)
{
  SH2movll4(sh, 13,14, 11);
}
#define init_5_13_14_12
void FASTCALL call_5_13_14_12(SH2_struct* sh)
{
  SH2movll4(sh, 13,14, 12);
}
#define init_5_13_14_13
void FASTCALL call_5_13_14_13(SH2_struct* sh)
{
  SH2movll4(sh, 13,14, 13);
}
#define init_5_13_14_14
void FASTCALL call_5_13_14_14(SH2_struct* sh)
{
  SH2movll4(sh, 13,14, 14);
}
#define init_5_13_14_15
void FASTCALL call_5_13_14_15(SH2_struct* sh)
{
  SH2movll4(sh, 13,14, 15);
}
#define init_5_13_15_0
void FASTCALL call_5_13_15_0(SH2_struct* sh)
{
  SH2movll4(sh, 13,15, 0);
}
#define init_5_13_15_1
void FASTCALL call_5_13_15_1(SH2_struct* sh)
{
  SH2movll4(sh, 13,15, 1);
}
#define init_5_13_15_2
void FASTCALL call_5_13_15_2(SH2_struct* sh)
{
  SH2movll4(sh, 13,15, 2);
}
#define init_5_13_15_3
void FASTCALL call_5_13_15_3(SH2_struct* sh)
{
  SH2movll4(sh, 13,15, 3);
}
#define init_5_13_15_4
void FASTCALL call_5_13_15_4(SH2_struct* sh)
{
  SH2movll4(sh, 13,15, 4);
}
#define init_5_13_15_5
void FASTCALL call_5_13_15_5(SH2_struct* sh)
{
  SH2movll4(sh, 13,15, 5);
}
#define init_5_13_15_6
void FASTCALL call_5_13_15_6(SH2_struct* sh)
{
  SH2movll4(sh, 13,15, 6);
}
#define init_5_13_15_7
void FASTCALL call_5_13_15_7(SH2_struct* sh)
{
  SH2movll4(sh, 13,15, 7);
}
#define init_5_13_15_8
void FASTCALL call_5_13_15_8(SH2_struct* sh)
{
  SH2movll4(sh, 13,15, 8);
}
#define init_5_13_15_9
void FASTCALL call_5_13_15_9(SH2_struct* sh)
{
  SH2movll4(sh, 13,15, 9);
}
#define init_5_13_15_10
void FASTCALL call_5_13_15_10(SH2_struct* sh)
{
  SH2movll4(sh, 13,15, 10);
}
#define init_5_13_15_11
void FASTCALL call_5_13_15_11(SH2_struct* sh)
{
  SH2movll4(sh, 13,15, 11);
}
#define init_5_13_15_12
void FASTCALL call_5_13_15_12(SH2_struct* sh)
{
  SH2movll4(sh, 13,15, 12);
}
#define init_5_13_15_13
void FASTCALL call_5_13_15_13(SH2_struct* sh)
{
  SH2movll4(sh, 13,15, 13);
}
#define init_5_13_15_14
void FASTCALL call_5_13_15_14(SH2_struct* sh)
{
  SH2movll4(sh, 13,15, 14);
}
#define init_5_13_15_15
void FASTCALL call_5_13_15_15(SH2_struct* sh)
{
  SH2movll4(sh, 13,15, 15);
}
#define init_5_14_0_0
void FASTCALL call_5_14_0_0(SH2_struct* sh)
{
  SH2movll4(sh, 14,0, 0);
}
#define init_5_14_0_1
void FASTCALL call_5_14_0_1(SH2_struct* sh)
{
  SH2movll4(sh, 14,0, 1);
}
#define init_5_14_0_2
void FASTCALL call_5_14_0_2(SH2_struct* sh)
{
  SH2movll4(sh, 14,0, 2);
}
#define init_5_14_0_3
void FASTCALL call_5_14_0_3(SH2_struct* sh)
{
  SH2movll4(sh, 14,0, 3);
}
#define init_5_14_0_4
void FASTCALL call_5_14_0_4(SH2_struct* sh)
{
  SH2movll4(sh, 14,0, 4);
}
#define init_5_14_0_5
void FASTCALL call_5_14_0_5(SH2_struct* sh)
{
  SH2movll4(sh, 14,0, 5);
}
#define init_5_14_0_6
void FASTCALL call_5_14_0_6(SH2_struct* sh)
{
  SH2movll4(sh, 14,0, 6);
}
#define init_5_14_0_7
void FASTCALL call_5_14_0_7(SH2_struct* sh)
{
  SH2movll4(sh, 14,0, 7);
}
#define init_5_14_0_8
void FASTCALL call_5_14_0_8(SH2_struct* sh)
{
  SH2movll4(sh, 14,0, 8);
}
#define init_5_14_0_9
void FASTCALL call_5_14_0_9(SH2_struct* sh)
{
  SH2movll4(sh, 14,0, 9);
}
#define init_5_14_0_10
void FASTCALL call_5_14_0_10(SH2_struct* sh)
{
  SH2movll4(sh, 14,0, 10);
}
#define init_5_14_0_11
void FASTCALL call_5_14_0_11(SH2_struct* sh)
{
  SH2movll4(sh, 14,0, 11);
}
#define init_5_14_0_12
void FASTCALL call_5_14_0_12(SH2_struct* sh)
{
  SH2movll4(sh, 14,0, 12);
}
#define init_5_14_0_13
void FASTCALL call_5_14_0_13(SH2_struct* sh)
{
  SH2movll4(sh, 14,0, 13);
}
#define init_5_14_0_14
void FASTCALL call_5_14_0_14(SH2_struct* sh)
{
  SH2movll4(sh, 14,0, 14);
}
#define init_5_14_0_15
void FASTCALL call_5_14_0_15(SH2_struct* sh)
{
  SH2movll4(sh, 14,0, 15);
}
#define init_5_14_1_0
void FASTCALL call_5_14_1_0(SH2_struct* sh)
{
  SH2movll4(sh, 14,1, 0);
}
#define init_5_14_1_1
void FASTCALL call_5_14_1_1(SH2_struct* sh)
{
  SH2movll4(sh, 14,1, 1);
}
#define init_5_14_1_2
void FASTCALL call_5_14_1_2(SH2_struct* sh)
{
  SH2movll4(sh, 14,1, 2);
}
#define init_5_14_1_3
void FASTCALL call_5_14_1_3(SH2_struct* sh)
{
  SH2movll4(sh, 14,1, 3);
}
#define init_5_14_1_4
void FASTCALL call_5_14_1_4(SH2_struct* sh)
{
  SH2movll4(sh, 14,1, 4);
}
#define init_5_14_1_5
void FASTCALL call_5_14_1_5(SH2_struct* sh)
{
  SH2movll4(sh, 14,1, 5);
}
#define init_5_14_1_6
void FASTCALL call_5_14_1_6(SH2_struct* sh)
{
  SH2movll4(sh, 14,1, 6);
}
#define init_5_14_1_7
void FASTCALL call_5_14_1_7(SH2_struct* sh)
{
  SH2movll4(sh, 14,1, 7);
}
#define init_5_14_1_8
void FASTCALL call_5_14_1_8(SH2_struct* sh)
{
  SH2movll4(sh, 14,1, 8);
}
#define init_5_14_1_9
void FASTCALL call_5_14_1_9(SH2_struct* sh)
{
  SH2movll4(sh, 14,1, 9);
}
#define init_5_14_1_10
void FASTCALL call_5_14_1_10(SH2_struct* sh)
{
  SH2movll4(sh, 14,1, 10);
}
#define init_5_14_1_11
void FASTCALL call_5_14_1_11(SH2_struct* sh)
{
  SH2movll4(sh, 14,1, 11);
}
#define init_5_14_1_12
void FASTCALL call_5_14_1_12(SH2_struct* sh)
{
  SH2movll4(sh, 14,1, 12);
}
#define init_5_14_1_13
void FASTCALL call_5_14_1_13(SH2_struct* sh)
{
  SH2movll4(sh, 14,1, 13);
}
#define init_5_14_1_14
void FASTCALL call_5_14_1_14(SH2_struct* sh)
{
  SH2movll4(sh, 14,1, 14);
}
#define init_5_14_1_15
void FASTCALL call_5_14_1_15(SH2_struct* sh)
{
  SH2movll4(sh, 14,1, 15);
}
#define init_5_14_2_0
void FASTCALL call_5_14_2_0(SH2_struct* sh)
{
  SH2movll4(sh, 14,2, 0);
}
#define init_5_14_2_1
void FASTCALL call_5_14_2_1(SH2_struct* sh)
{
  SH2movll4(sh, 14,2, 1);
}
#define init_5_14_2_2
void FASTCALL call_5_14_2_2(SH2_struct* sh)
{
  SH2movll4(sh, 14,2, 2);
}
#define init_5_14_2_3
void FASTCALL call_5_14_2_3(SH2_struct* sh)
{
  SH2movll4(sh, 14,2, 3);
}
#define init_5_14_2_4
void FASTCALL call_5_14_2_4(SH2_struct* sh)
{
  SH2movll4(sh, 14,2, 4);
}
#define init_5_14_2_5
void FASTCALL call_5_14_2_5(SH2_struct* sh)
{
  SH2movll4(sh, 14,2, 5);
}
#define init_5_14_2_6
void FASTCALL call_5_14_2_6(SH2_struct* sh)
{
  SH2movll4(sh, 14,2, 6);
}
#define init_5_14_2_7
void FASTCALL call_5_14_2_7(SH2_struct* sh)
{
  SH2movll4(sh, 14,2, 7);
}
#define init_5_14_2_8
void FASTCALL call_5_14_2_8(SH2_struct* sh)
{
  SH2movll4(sh, 14,2, 8);
}
#define init_5_14_2_9
void FASTCALL call_5_14_2_9(SH2_struct* sh)
{
  SH2movll4(sh, 14,2, 9);
}
#define init_5_14_2_10
void FASTCALL call_5_14_2_10(SH2_struct* sh)
{
  SH2movll4(sh, 14,2, 10);
}
#define init_5_14_2_11
void FASTCALL call_5_14_2_11(SH2_struct* sh)
{
  SH2movll4(sh, 14,2, 11);
}
#define init_5_14_2_12
void FASTCALL call_5_14_2_12(SH2_struct* sh)
{
  SH2movll4(sh, 14,2, 12);
}
#define init_5_14_2_13
void FASTCALL call_5_14_2_13(SH2_struct* sh)
{
  SH2movll4(sh, 14,2, 13);
}
#define init_5_14_2_14
void FASTCALL call_5_14_2_14(SH2_struct* sh)
{
  SH2movll4(sh, 14,2, 14);
}
#define init_5_14_2_15
void FASTCALL call_5_14_2_15(SH2_struct* sh)
{
  SH2movll4(sh, 14,2, 15);
}
#define init_5_14_3_0
void FASTCALL call_5_14_3_0(SH2_struct* sh)
{
  SH2movll4(sh, 14,3, 0);
}
#define init_5_14_3_1
void FASTCALL call_5_14_3_1(SH2_struct* sh)
{
  SH2movll4(sh, 14,3, 1);
}
#define init_5_14_3_2
void FASTCALL call_5_14_3_2(SH2_struct* sh)
{
  SH2movll4(sh, 14,3, 2);
}
#define init_5_14_3_3
void FASTCALL call_5_14_3_3(SH2_struct* sh)
{
  SH2movll4(sh, 14,3, 3);
}
#define init_5_14_3_4
void FASTCALL call_5_14_3_4(SH2_struct* sh)
{
  SH2movll4(sh, 14,3, 4);
}
#define init_5_14_3_5
void FASTCALL call_5_14_3_5(SH2_struct* sh)
{
  SH2movll4(sh, 14,3, 5);
}
#define init_5_14_3_6
void FASTCALL call_5_14_3_6(SH2_struct* sh)
{
  SH2movll4(sh, 14,3, 6);
}
#define init_5_14_3_7
void FASTCALL call_5_14_3_7(SH2_struct* sh)
{
  SH2movll4(sh, 14,3, 7);
}
#define init_5_14_3_8
void FASTCALL call_5_14_3_8(SH2_struct* sh)
{
  SH2movll4(sh, 14,3, 8);
}
#define init_5_14_3_9
void FASTCALL call_5_14_3_9(SH2_struct* sh)
{
  SH2movll4(sh, 14,3, 9);
}
#define init_5_14_3_10
void FASTCALL call_5_14_3_10(SH2_struct* sh)
{
  SH2movll4(sh, 14,3, 10);
}
#define init_5_14_3_11
void FASTCALL call_5_14_3_11(SH2_struct* sh)
{
  SH2movll4(sh, 14,3, 11);
}
#define init_5_14_3_12
void FASTCALL call_5_14_3_12(SH2_struct* sh)
{
  SH2movll4(sh, 14,3, 12);
}
#define init_5_14_3_13
void FASTCALL call_5_14_3_13(SH2_struct* sh)
{
  SH2movll4(sh, 14,3, 13);
}
#define init_5_14_3_14
void FASTCALL call_5_14_3_14(SH2_struct* sh)
{
  SH2movll4(sh, 14,3, 14);
}
#define init_5_14_3_15
void FASTCALL call_5_14_3_15(SH2_struct* sh)
{
  SH2movll4(sh, 14,3, 15);
}
#define init_5_14_4_0
void FASTCALL call_5_14_4_0(SH2_struct* sh)
{
  SH2movll4(sh, 14,4, 0);
}
#define init_5_14_4_1
void FASTCALL call_5_14_4_1(SH2_struct* sh)
{
  SH2movll4(sh, 14,4, 1);
}
#define init_5_14_4_2
void FASTCALL call_5_14_4_2(SH2_struct* sh)
{
  SH2movll4(sh, 14,4, 2);
}
#define init_5_14_4_3
void FASTCALL call_5_14_4_3(SH2_struct* sh)
{
  SH2movll4(sh, 14,4, 3);
}
#define init_5_14_4_4
void FASTCALL call_5_14_4_4(SH2_struct* sh)
{
  SH2movll4(sh, 14,4, 4);
}
#define init_5_14_4_5
void FASTCALL call_5_14_4_5(SH2_struct* sh)
{
  SH2movll4(sh, 14,4, 5);
}
#define init_5_14_4_6
void FASTCALL call_5_14_4_6(SH2_struct* sh)
{
  SH2movll4(sh, 14,4, 6);
}
#define init_5_14_4_7
void FASTCALL call_5_14_4_7(SH2_struct* sh)
{
  SH2movll4(sh, 14,4, 7);
}
#define init_5_14_4_8
void FASTCALL call_5_14_4_8(SH2_struct* sh)
{
  SH2movll4(sh, 14,4, 8);
}
#define init_5_14_4_9
void FASTCALL call_5_14_4_9(SH2_struct* sh)
{
  SH2movll4(sh, 14,4, 9);
}
#define init_5_14_4_10
void FASTCALL call_5_14_4_10(SH2_struct* sh)
{
  SH2movll4(sh, 14,4, 10);
}
#define init_5_14_4_11
void FASTCALL call_5_14_4_11(SH2_struct* sh)
{
  SH2movll4(sh, 14,4, 11);
}
#define init_5_14_4_12
void FASTCALL call_5_14_4_12(SH2_struct* sh)
{
  SH2movll4(sh, 14,4, 12);
}
#define init_5_14_4_13
void FASTCALL call_5_14_4_13(SH2_struct* sh)
{
  SH2movll4(sh, 14,4, 13);
}
#define init_5_14_4_14
void FASTCALL call_5_14_4_14(SH2_struct* sh)
{
  SH2movll4(sh, 14,4, 14);
}
#define init_5_14_4_15
void FASTCALL call_5_14_4_15(SH2_struct* sh)
{
  SH2movll4(sh, 14,4, 15);
}
#define init_5_14_5_0
void FASTCALL call_5_14_5_0(SH2_struct* sh)
{
  SH2movll4(sh, 14,5, 0);
}
#define init_5_14_5_1
void FASTCALL call_5_14_5_1(SH2_struct* sh)
{
  SH2movll4(sh, 14,5, 1);
}
#define init_5_14_5_2
void FASTCALL call_5_14_5_2(SH2_struct* sh)
{
  SH2movll4(sh, 14,5, 2);
}
#define init_5_14_5_3
void FASTCALL call_5_14_5_3(SH2_struct* sh)
{
  SH2movll4(sh, 14,5, 3);
}
#define init_5_14_5_4
void FASTCALL call_5_14_5_4(SH2_struct* sh)
{
  SH2movll4(sh, 14,5, 4);
}
#define init_5_14_5_5
void FASTCALL call_5_14_5_5(SH2_struct* sh)
{
  SH2movll4(sh, 14,5, 5);
}
#define init_5_14_5_6
void FASTCALL call_5_14_5_6(SH2_struct* sh)
{
  SH2movll4(sh, 14,5, 6);
}
#define init_5_14_5_7
void FASTCALL call_5_14_5_7(SH2_struct* sh)
{
  SH2movll4(sh, 14,5, 7);
}
#define init_5_14_5_8
void FASTCALL call_5_14_5_8(SH2_struct* sh)
{
  SH2movll4(sh, 14,5, 8);
}
#define init_5_14_5_9
void FASTCALL call_5_14_5_9(SH2_struct* sh)
{
  SH2movll4(sh, 14,5, 9);
}
#define init_5_14_5_10
void FASTCALL call_5_14_5_10(SH2_struct* sh)
{
  SH2movll4(sh, 14,5, 10);
}
#define init_5_14_5_11
void FASTCALL call_5_14_5_11(SH2_struct* sh)
{
  SH2movll4(sh, 14,5, 11);
}
#define init_5_14_5_12
void FASTCALL call_5_14_5_12(SH2_struct* sh)
{
  SH2movll4(sh, 14,5, 12);
}
#define init_5_14_5_13
void FASTCALL call_5_14_5_13(SH2_struct* sh)
{
  SH2movll4(sh, 14,5, 13);
}
#define init_5_14_5_14
void FASTCALL call_5_14_5_14(SH2_struct* sh)
{
  SH2movll4(sh, 14,5, 14);
}
#define init_5_14_5_15
void FASTCALL call_5_14_5_15(SH2_struct* sh)
{
  SH2movll4(sh, 14,5, 15);
}
#define init_5_14_6_0
void FASTCALL call_5_14_6_0(SH2_struct* sh)
{
  SH2movll4(sh, 14,6, 0);
}
#define init_5_14_6_1
void FASTCALL call_5_14_6_1(SH2_struct* sh)
{
  SH2movll4(sh, 14,6, 1);
}
#define init_5_14_6_2
void FASTCALL call_5_14_6_2(SH2_struct* sh)
{
  SH2movll4(sh, 14,6, 2);
}
#define init_5_14_6_3
void FASTCALL call_5_14_6_3(SH2_struct* sh)
{
  SH2movll4(sh, 14,6, 3);
}
#define init_5_14_6_4
void FASTCALL call_5_14_6_4(SH2_struct* sh)
{
  SH2movll4(sh, 14,6, 4);
}
#define init_5_14_6_5
void FASTCALL call_5_14_6_5(SH2_struct* sh)
{
  SH2movll4(sh, 14,6, 5);
}
#define init_5_14_6_6
void FASTCALL call_5_14_6_6(SH2_struct* sh)
{
  SH2movll4(sh, 14,6, 6);
}
#define init_5_14_6_7
void FASTCALL call_5_14_6_7(SH2_struct* sh)
{
  SH2movll4(sh, 14,6, 7);
}
#define init_5_14_6_8
void FASTCALL call_5_14_6_8(SH2_struct* sh)
{
  SH2movll4(sh, 14,6, 8);
}
#define init_5_14_6_9
void FASTCALL call_5_14_6_9(SH2_struct* sh)
{
  SH2movll4(sh, 14,6, 9);
}
#define init_5_14_6_10
void FASTCALL call_5_14_6_10(SH2_struct* sh)
{
  SH2movll4(sh, 14,6, 10);
}
#define init_5_14_6_11
void FASTCALL call_5_14_6_11(SH2_struct* sh)
{
  SH2movll4(sh, 14,6, 11);
}
#define init_5_14_6_12
void FASTCALL call_5_14_6_12(SH2_struct* sh)
{
  SH2movll4(sh, 14,6, 12);
}
#define init_5_14_6_13
void FASTCALL call_5_14_6_13(SH2_struct* sh)
{
  SH2movll4(sh, 14,6, 13);
}
#define init_5_14_6_14
void FASTCALL call_5_14_6_14(SH2_struct* sh)
{
  SH2movll4(sh, 14,6, 14);
}
#define init_5_14_6_15
void FASTCALL call_5_14_6_15(SH2_struct* sh)
{
  SH2movll4(sh, 14,6, 15);
}
#define init_5_14_7_0
void FASTCALL call_5_14_7_0(SH2_struct* sh)
{
  SH2movll4(sh, 14,7, 0);
}
#define init_5_14_7_1
void FASTCALL call_5_14_7_1(SH2_struct* sh)
{
  SH2movll4(sh, 14,7, 1);
}
#define init_5_14_7_2
void FASTCALL call_5_14_7_2(SH2_struct* sh)
{
  SH2movll4(sh, 14,7, 2);
}
#define init_5_14_7_3
void FASTCALL call_5_14_7_3(SH2_struct* sh)
{
  SH2movll4(sh, 14,7, 3);
}
#define init_5_14_7_4
void FASTCALL call_5_14_7_4(SH2_struct* sh)
{
  SH2movll4(sh, 14,7, 4);
}
#define init_5_14_7_5
void FASTCALL call_5_14_7_5(SH2_struct* sh)
{
  SH2movll4(sh, 14,7, 5);
}
#define init_5_14_7_6
void FASTCALL call_5_14_7_6(SH2_struct* sh)
{
  SH2movll4(sh, 14,7, 6);
}
#define init_5_14_7_7
void FASTCALL call_5_14_7_7(SH2_struct* sh)
{
  SH2movll4(sh, 14,7, 7);
}
#define init_5_14_7_8
void FASTCALL call_5_14_7_8(SH2_struct* sh)
{
  SH2movll4(sh, 14,7, 8);
}
#define init_5_14_7_9
void FASTCALL call_5_14_7_9(SH2_struct* sh)
{
  SH2movll4(sh, 14,7, 9);
}
#define init_5_14_7_10
void FASTCALL call_5_14_7_10(SH2_struct* sh)
{
  SH2movll4(sh, 14,7, 10);
}
#define init_5_14_7_11
void FASTCALL call_5_14_7_11(SH2_struct* sh)
{
  SH2movll4(sh, 14,7, 11);
}
#define init_5_14_7_12
void FASTCALL call_5_14_7_12(SH2_struct* sh)
{
  SH2movll4(sh, 14,7, 12);
}
#define init_5_14_7_13
void FASTCALL call_5_14_7_13(SH2_struct* sh)
{
  SH2movll4(sh, 14,7, 13);
}
#define init_5_14_7_14
void FASTCALL call_5_14_7_14(SH2_struct* sh)
{
  SH2movll4(sh, 14,7, 14);
}
#define init_5_14_7_15
void FASTCALL call_5_14_7_15(SH2_struct* sh)
{
  SH2movll4(sh, 14,7, 15);
}
#define init_5_14_8_0
void FASTCALL call_5_14_8_0(SH2_struct* sh)
{
  SH2movll4(sh, 14,8, 0);
}
#define init_5_14_8_1
void FASTCALL call_5_14_8_1(SH2_struct* sh)
{
  SH2movll4(sh, 14,8, 1);
}
#define init_5_14_8_2
void FASTCALL call_5_14_8_2(SH2_struct* sh)
{
  SH2movll4(sh, 14,8, 2);
}
#define init_5_14_8_3
void FASTCALL call_5_14_8_3(SH2_struct* sh)
{
  SH2movll4(sh, 14,8, 3);
}
#define init_5_14_8_4
void FASTCALL call_5_14_8_4(SH2_struct* sh)
{
  SH2movll4(sh, 14,8, 4);
}
#define init_5_14_8_5
void FASTCALL call_5_14_8_5(SH2_struct* sh)
{
  SH2movll4(sh, 14,8, 5);
}
#define init_5_14_8_6
void FASTCALL call_5_14_8_6(SH2_struct* sh)
{
  SH2movll4(sh, 14,8, 6);
}
#define init_5_14_8_7
void FASTCALL call_5_14_8_7(SH2_struct* sh)
{
  SH2movll4(sh, 14,8, 7);
}
#define init_5_14_8_8
void FASTCALL call_5_14_8_8(SH2_struct* sh)
{
  SH2movll4(sh, 14,8, 8);
}
#define init_5_14_8_9
void FASTCALL call_5_14_8_9(SH2_struct* sh)
{
  SH2movll4(sh, 14,8, 9);
}
#define init_5_14_8_10
void FASTCALL call_5_14_8_10(SH2_struct* sh)
{
  SH2movll4(sh, 14,8, 10);
}
#define init_5_14_8_11
void FASTCALL call_5_14_8_11(SH2_struct* sh)
{
  SH2movll4(sh, 14,8, 11);
}
#define init_5_14_8_12
void FASTCALL call_5_14_8_12(SH2_struct* sh)
{
  SH2movll4(sh, 14,8, 12);
}
#define init_5_14_8_13
void FASTCALL call_5_14_8_13(SH2_struct* sh)
{
  SH2movll4(sh, 14,8, 13);
}
#define init_5_14_8_14
void FASTCALL call_5_14_8_14(SH2_struct* sh)
{
  SH2movll4(sh, 14,8, 14);
}
#define init_5_14_8_15
void FASTCALL call_5_14_8_15(SH2_struct* sh)
{
  SH2movll4(sh, 14,8, 15);
}
#define init_5_14_9_0
void FASTCALL call_5_14_9_0(SH2_struct* sh)
{
  SH2movll4(sh, 14,9, 0);
}
#define init_5_14_9_1
void FASTCALL call_5_14_9_1(SH2_struct* sh)
{
  SH2movll4(sh, 14,9, 1);
}
#define init_5_14_9_2
void FASTCALL call_5_14_9_2(SH2_struct* sh)
{
  SH2movll4(sh, 14,9, 2);
}
#define init_5_14_9_3
void FASTCALL call_5_14_9_3(SH2_struct* sh)
{
  SH2movll4(sh, 14,9, 3);
}
#define init_5_14_9_4
void FASTCALL call_5_14_9_4(SH2_struct* sh)
{
  SH2movll4(sh, 14,9, 4);
}
#define init_5_14_9_5
void FASTCALL call_5_14_9_5(SH2_struct* sh)
{
  SH2movll4(sh, 14,9, 5);
}
#define init_5_14_9_6
void FASTCALL call_5_14_9_6(SH2_struct* sh)
{
  SH2movll4(sh, 14,9, 6);
}
#define init_5_14_9_7
void FASTCALL call_5_14_9_7(SH2_struct* sh)
{
  SH2movll4(sh, 14,9, 7);
}
#define init_5_14_9_8
void FASTCALL call_5_14_9_8(SH2_struct* sh)
{
  SH2movll4(sh, 14,9, 8);
}
#define init_5_14_9_9
void FASTCALL call_5_14_9_9(SH2_struct* sh)
{
  SH2movll4(sh, 14,9, 9);
}
#define init_5_14_9_10
void FASTCALL call_5_14_9_10(SH2_struct* sh)
{
  SH2movll4(sh, 14,9, 10);
}
#define init_5_14_9_11
void FASTCALL call_5_14_9_11(SH2_struct* sh)
{
  SH2movll4(sh, 14,9, 11);
}
#define init_5_14_9_12
void FASTCALL call_5_14_9_12(SH2_struct* sh)
{
  SH2movll4(sh, 14,9, 12);
}
#define init_5_14_9_13
void FASTCALL call_5_14_9_13(SH2_struct* sh)
{
  SH2movll4(sh, 14,9, 13);
}
#define init_5_14_9_14
void FASTCALL call_5_14_9_14(SH2_struct* sh)
{
  SH2movll4(sh, 14,9, 14);
}
#define init_5_14_9_15
void FASTCALL call_5_14_9_15(SH2_struct* sh)
{
  SH2movll4(sh, 14,9, 15);
}
#define init_5_14_10_0
void FASTCALL call_5_14_10_0(SH2_struct* sh)
{
  SH2movll4(sh, 14,10, 0);
}
#define init_5_14_10_1
void FASTCALL call_5_14_10_1(SH2_struct* sh)
{
  SH2movll4(sh, 14,10, 1);
}
#define init_5_14_10_2
void FASTCALL call_5_14_10_2(SH2_struct* sh)
{
  SH2movll4(sh, 14,10, 2);
}
#define init_5_14_10_3
void FASTCALL call_5_14_10_3(SH2_struct* sh)
{
  SH2movll4(sh, 14,10, 3);
}
#define init_5_14_10_4
void FASTCALL call_5_14_10_4(SH2_struct* sh)
{
  SH2movll4(sh, 14,10, 4);
}
#define init_5_14_10_5
void FASTCALL call_5_14_10_5(SH2_struct* sh)
{
  SH2movll4(sh, 14,10, 5);
}
#define init_5_14_10_6
void FASTCALL call_5_14_10_6(SH2_struct* sh)
{
  SH2movll4(sh, 14,10, 6);
}
#define init_5_14_10_7
void FASTCALL call_5_14_10_7(SH2_struct* sh)
{
  SH2movll4(sh, 14,10, 7);
}
#define init_5_14_10_8
void FASTCALL call_5_14_10_8(SH2_struct* sh)
{
  SH2movll4(sh, 14,10, 8);
}
#define init_5_14_10_9
void FASTCALL call_5_14_10_9(SH2_struct* sh)
{
  SH2movll4(sh, 14,10, 9);
}
#define init_5_14_10_10
void FASTCALL call_5_14_10_10(SH2_struct* sh)
{
  SH2movll4(sh, 14,10, 10);
}
#define init_5_14_10_11
void FASTCALL call_5_14_10_11(SH2_struct* sh)
{
  SH2movll4(sh, 14,10, 11);
}
#define init_5_14_10_12
void FASTCALL call_5_14_10_12(SH2_struct* sh)
{
  SH2movll4(sh, 14,10, 12);
}
#define init_5_14_10_13
void FASTCALL call_5_14_10_13(SH2_struct* sh)
{
  SH2movll4(sh, 14,10, 13);
}
#define init_5_14_10_14
void FASTCALL call_5_14_10_14(SH2_struct* sh)
{
  SH2movll4(sh, 14,10, 14);
}
#define init_5_14_10_15
void FASTCALL call_5_14_10_15(SH2_struct* sh)
{
  SH2movll4(sh, 14,10, 15);
}
#define init_5_14_11_0
void FASTCALL call_5_14_11_0(SH2_struct* sh)
{
  SH2movll4(sh, 14,11, 0);
}
#define init_5_14_11_1
void FASTCALL call_5_14_11_1(SH2_struct* sh)
{
  SH2movll4(sh, 14,11, 1);
}
#define init_5_14_11_2
void FASTCALL call_5_14_11_2(SH2_struct* sh)
{
  SH2movll4(sh, 14,11, 2);
}
#define init_5_14_11_3
void FASTCALL call_5_14_11_3(SH2_struct* sh)
{
  SH2movll4(sh, 14,11, 3);
}
#define init_5_14_11_4
void FASTCALL call_5_14_11_4(SH2_struct* sh)
{
  SH2movll4(sh, 14,11, 4);
}
#define init_5_14_11_5
void FASTCALL call_5_14_11_5(SH2_struct* sh)
{
  SH2movll4(sh, 14,11, 5);
}
#define init_5_14_11_6
void FASTCALL call_5_14_11_6(SH2_struct* sh)
{
  SH2movll4(sh, 14,11, 6);
}
#define init_5_14_11_7
void FASTCALL call_5_14_11_7(SH2_struct* sh)
{
  SH2movll4(sh, 14,11, 7);
}
#define init_5_14_11_8
void FASTCALL call_5_14_11_8(SH2_struct* sh)
{
  SH2movll4(sh, 14,11, 8);
}
#define init_5_14_11_9
void FASTCALL call_5_14_11_9(SH2_struct* sh)
{
  SH2movll4(sh, 14,11, 9);
}
#define init_5_14_11_10
void FASTCALL call_5_14_11_10(SH2_struct* sh)
{
  SH2movll4(sh, 14,11, 10);
}
#define init_5_14_11_11
void FASTCALL call_5_14_11_11(SH2_struct* sh)
{
  SH2movll4(sh, 14,11, 11);
}
#define init_5_14_11_12
void FASTCALL call_5_14_11_12(SH2_struct* sh)
{
  SH2movll4(sh, 14,11, 12);
}
#define init_5_14_11_13
void FASTCALL call_5_14_11_13(SH2_struct* sh)
{
  SH2movll4(sh, 14,11, 13);
}
#define init_5_14_11_14
void FASTCALL call_5_14_11_14(SH2_struct* sh)
{
  SH2movll4(sh, 14,11, 14);
}
#define init_5_14_11_15
void FASTCALL call_5_14_11_15(SH2_struct* sh)
{
  SH2movll4(sh, 14,11, 15);
}
#define init_5_14_12_0
void FASTCALL call_5_14_12_0(SH2_struct* sh)
{
  SH2movll4(sh, 14,12, 0);
}
#define init_5_14_12_1
void FASTCALL call_5_14_12_1(SH2_struct* sh)
{
  SH2movll4(sh, 14,12, 1);
}
#define init_5_14_12_2
void FASTCALL call_5_14_12_2(SH2_struct* sh)
{
  SH2movll4(sh, 14,12, 2);
}
#define init_5_14_12_3
void FASTCALL call_5_14_12_3(SH2_struct* sh)
{
  SH2movll4(sh, 14,12, 3);
}
#define init_5_14_12_4
void FASTCALL call_5_14_12_4(SH2_struct* sh)
{
  SH2movll4(sh, 14,12, 4);
}
#define init_5_14_12_5
void FASTCALL call_5_14_12_5(SH2_struct* sh)
{
  SH2movll4(sh, 14,12, 5);
}
#define init_5_14_12_6
void FASTCALL call_5_14_12_6(SH2_struct* sh)
{
  SH2movll4(sh, 14,12, 6);
}
#define init_5_14_12_7
void FASTCALL call_5_14_12_7(SH2_struct* sh)
{
  SH2movll4(sh, 14,12, 7);
}
#define init_5_14_12_8
void FASTCALL call_5_14_12_8(SH2_struct* sh)
{
  SH2movll4(sh, 14,12, 8);
}
#define init_5_14_12_9
void FASTCALL call_5_14_12_9(SH2_struct* sh)
{
  SH2movll4(sh, 14,12, 9);
}
#define init_5_14_12_10
void FASTCALL call_5_14_12_10(SH2_struct* sh)
{
  SH2movll4(sh, 14,12, 10);
}
#define init_5_14_12_11
void FASTCALL call_5_14_12_11(SH2_struct* sh)
{
  SH2movll4(sh, 14,12, 11);
}
#define init_5_14_12_12
void FASTCALL call_5_14_12_12(SH2_struct* sh)
{
  SH2movll4(sh, 14,12, 12);
}
#define init_5_14_12_13
void FASTCALL call_5_14_12_13(SH2_struct* sh)
{
  SH2movll4(sh, 14,12, 13);
}
#define init_5_14_12_14
void FASTCALL call_5_14_12_14(SH2_struct* sh)
{
  SH2movll4(sh, 14,12, 14);
}
#define init_5_14_12_15
void FASTCALL call_5_14_12_15(SH2_struct* sh)
{
  SH2movll4(sh, 14,12, 15);
}
#define init_5_14_13_0
void FASTCALL call_5_14_13_0(SH2_struct* sh)
{
  SH2movll4(sh, 14,13, 0);
}
#define init_5_14_13_1
void FASTCALL call_5_14_13_1(SH2_struct* sh)
{
  SH2movll4(sh, 14,13, 1);
}
#define init_5_14_13_2
void FASTCALL call_5_14_13_2(SH2_struct* sh)
{
  SH2movll4(sh, 14,13, 2);
}
#define init_5_14_13_3
void FASTCALL call_5_14_13_3(SH2_struct* sh)
{
  SH2movll4(sh, 14,13, 3);
}
#define init_5_14_13_4
void FASTCALL call_5_14_13_4(SH2_struct* sh)
{
  SH2movll4(sh, 14,13, 4);
}
#define init_5_14_13_5
void FASTCALL call_5_14_13_5(SH2_struct* sh)
{
  SH2movll4(sh, 14,13, 5);
}
#define init_5_14_13_6
void FASTCALL call_5_14_13_6(SH2_struct* sh)
{
  SH2movll4(sh, 14,13, 6);
}
#define init_5_14_13_7
void FASTCALL call_5_14_13_7(SH2_struct* sh)
{
  SH2movll4(sh, 14,13, 7);
}
#define init_5_14_13_8
void FASTCALL call_5_14_13_8(SH2_struct* sh)
{
  SH2movll4(sh, 14,13, 8);
}
#define init_5_14_13_9
void FASTCALL call_5_14_13_9(SH2_struct* sh)
{
  SH2movll4(sh, 14,13, 9);
}
#define init_5_14_13_10
void FASTCALL call_5_14_13_10(SH2_struct* sh)
{
  SH2movll4(sh, 14,13, 10);
}
#define init_5_14_13_11
void FASTCALL call_5_14_13_11(SH2_struct* sh)
{
  SH2movll4(sh, 14,13, 11);
}
#define init_5_14_13_12
void FASTCALL call_5_14_13_12(SH2_struct* sh)
{
  SH2movll4(sh, 14,13, 12);
}
#define init_5_14_13_13
void FASTCALL call_5_14_13_13(SH2_struct* sh)
{
  SH2movll4(sh, 14,13, 13);
}
#define init_5_14_13_14
void FASTCALL call_5_14_13_14(SH2_struct* sh)
{
  SH2movll4(sh, 14,13, 14);
}
#define init_5_14_13_15
void FASTCALL call_5_14_13_15(SH2_struct* sh)
{
  SH2movll4(sh, 14,13, 15);
}
#define init_5_14_14_0
void FASTCALL call_5_14_14_0(SH2_struct* sh)
{
  SH2movll4(sh, 14,14, 0);
}
#define init_5_14_14_1
void FASTCALL call_5_14_14_1(SH2_struct* sh)
{
  SH2movll4(sh, 14,14, 1);
}
#define init_5_14_14_2
void FASTCALL call_5_14_14_2(SH2_struct* sh)
{
  SH2movll4(sh, 14,14, 2);
}
#define init_5_14_14_3
void FASTCALL call_5_14_14_3(SH2_struct* sh)
{
  SH2movll4(sh, 14,14, 3);
}
#define init_5_14_14_4
void FASTCALL call_5_14_14_4(SH2_struct* sh)
{
  SH2movll4(sh, 14,14, 4);
}
#define init_5_14_14_5
void FASTCALL call_5_14_14_5(SH2_struct* sh)
{
  SH2movll4(sh, 14,14, 5);
}
#define init_5_14_14_6
void FASTCALL call_5_14_14_6(SH2_struct* sh)
{
  SH2movll4(sh, 14,14, 6);
}
#define init_5_14_14_7
void FASTCALL call_5_14_14_7(SH2_struct* sh)
{
  SH2movll4(sh, 14,14, 7);
}
#define init_5_14_14_8
void FASTCALL call_5_14_14_8(SH2_struct* sh)
{
  SH2movll4(sh, 14,14, 8);
}
#define init_5_14_14_9
void FASTCALL call_5_14_14_9(SH2_struct* sh)
{
  SH2movll4(sh, 14,14, 9);
}
#define init_5_14_14_10
void FASTCALL call_5_14_14_10(SH2_struct* sh)
{
  SH2movll4(sh, 14,14, 10);
}
#define init_5_14_14_11
void FASTCALL call_5_14_14_11(SH2_struct* sh)
{
  SH2movll4(sh, 14,14, 11);
}
#define init_5_14_14_12
void FASTCALL call_5_14_14_12(SH2_struct* sh)
{
  SH2movll4(sh, 14,14, 12);
}
#define init_5_14_14_13
void FASTCALL call_5_14_14_13(SH2_struct* sh)
{
  SH2movll4(sh, 14,14, 13);
}
#define init_5_14_14_14
void FASTCALL call_5_14_14_14(SH2_struct* sh)
{
  SH2movll4(sh, 14,14, 14);
}
#define init_5_14_14_15
void FASTCALL call_5_14_14_15(SH2_struct* sh)
{
  SH2movll4(sh, 14,14, 15);
}
#define init_5_14_15_0
void FASTCALL call_5_14_15_0(SH2_struct* sh)
{
  SH2movll4(sh, 14,15, 0);
}
#define init_5_14_15_1
void FASTCALL call_5_14_15_1(SH2_struct* sh)
{
  SH2movll4(sh, 14,15, 1);
}
#define init_5_14_15_2
void FASTCALL call_5_14_15_2(SH2_struct* sh)
{
  SH2movll4(sh, 14,15, 2);
}
#define init_5_14_15_3
void FASTCALL call_5_14_15_3(SH2_struct* sh)
{
  SH2movll4(sh, 14,15, 3);
}
#define init_5_14_15_4
void FASTCALL call_5_14_15_4(SH2_struct* sh)
{
  SH2movll4(sh, 14,15, 4);
}
#define init_5_14_15_5
void FASTCALL call_5_14_15_5(SH2_struct* sh)
{
  SH2movll4(sh, 14,15, 5);
}
#define init_5_14_15_6
void FASTCALL call_5_14_15_6(SH2_struct* sh)
{
  SH2movll4(sh, 14,15, 6);
}
#define init_5_14_15_7
void FASTCALL call_5_14_15_7(SH2_struct* sh)
{
  SH2movll4(sh, 14,15, 7);
}
#define init_5_14_15_8
void FASTCALL call_5_14_15_8(SH2_struct* sh)
{
  SH2movll4(sh, 14,15, 8);
}
#define init_5_14_15_9
void FASTCALL call_5_14_15_9(SH2_struct* sh)
{
  SH2movll4(sh, 14,15, 9);
}
#define init_5_14_15_10
void FASTCALL call_5_14_15_10(SH2_struct* sh)
{
  SH2movll4(sh, 14,15, 10);
}
#define init_5_14_15_11
void FASTCALL call_5_14_15_11(SH2_struct* sh)
{
  SH2movll4(sh, 14,15, 11);
}
#define init_5_14_15_12
void FASTCALL call_5_14_15_12(SH2_struct* sh)
{
  SH2movll4(sh, 14,15, 12);
}
#define init_5_14_15_13
void FASTCALL call_5_14_15_13(SH2_struct* sh)
{
  SH2movll4(sh, 14,15, 13);
}
#define init_5_14_15_14
void FASTCALL call_5_14_15_14(SH2_struct* sh)
{
  SH2movll4(sh, 14,15, 14);
}
#define init_5_14_15_15
void FASTCALL call_5_14_15_15(SH2_struct* sh)
{
  SH2movll4(sh, 14,15, 15);
}
#define init_5_15_0_0
void FASTCALL call_5_15_0_0(SH2_struct* sh)
{
  SH2movll4(sh, 15,0, 0);
}
#define init_5_15_0_1
void FASTCALL call_5_15_0_1(SH2_struct* sh)
{
  SH2movll4(sh, 15,0, 1);
}
#define init_5_15_0_2
void FASTCALL call_5_15_0_2(SH2_struct* sh)
{
  SH2movll4(sh, 15,0, 2);
}
#define init_5_15_0_3
void FASTCALL call_5_15_0_3(SH2_struct* sh)
{
  SH2movll4(sh, 15,0, 3);
}
#define init_5_15_0_4
void FASTCALL call_5_15_0_4(SH2_struct* sh)
{
  SH2movll4(sh, 15,0, 4);
}
#define init_5_15_0_5
void FASTCALL call_5_15_0_5(SH2_struct* sh)
{
  SH2movll4(sh, 15,0, 5);
}
#define init_5_15_0_6
void FASTCALL call_5_15_0_6(SH2_struct* sh)
{
  SH2movll4(sh, 15,0, 6);
}
#define init_5_15_0_7
void FASTCALL call_5_15_0_7(SH2_struct* sh)
{
  SH2movll4(sh, 15,0, 7);
}
#define init_5_15_0_8
void FASTCALL call_5_15_0_8(SH2_struct* sh)
{
  SH2movll4(sh, 15,0, 8);
}
#define init_5_15_0_9
void FASTCALL call_5_15_0_9(SH2_struct* sh)
{
  SH2movll4(sh, 15,0, 9);
}
#define init_5_15_0_10
void FASTCALL call_5_15_0_10(SH2_struct* sh)
{
  SH2movll4(sh, 15,0, 10);
}
#define init_5_15_0_11
void FASTCALL call_5_15_0_11(SH2_struct* sh)
{
  SH2movll4(sh, 15,0, 11);
}
#define init_5_15_0_12
void FASTCALL call_5_15_0_12(SH2_struct* sh)
{
  SH2movll4(sh, 15,0, 12);
}
#define init_5_15_0_13
void FASTCALL call_5_15_0_13(SH2_struct* sh)
{
  SH2movll4(sh, 15,0, 13);
}
#define init_5_15_0_14
void FASTCALL call_5_15_0_14(SH2_struct* sh)
{
  SH2movll4(sh, 15,0, 14);
}
#define init_5_15_0_15
void FASTCALL call_5_15_0_15(SH2_struct* sh)
{
  SH2movll4(sh, 15,0, 15);
}
#define init_5_15_1_0
void FASTCALL call_5_15_1_0(SH2_struct* sh)
{
  SH2movll4(sh, 15,1, 0);
}
#define init_5_15_1_1
void FASTCALL call_5_15_1_1(SH2_struct* sh)
{
  SH2movll4(sh, 15,1, 1);
}
#define init_5_15_1_2
void FASTCALL call_5_15_1_2(SH2_struct* sh)
{
  SH2movll4(sh, 15,1, 2);
}
#define init_5_15_1_3
void FASTCALL call_5_15_1_3(SH2_struct* sh)
{
  SH2movll4(sh, 15,1, 3);
}
#define init_5_15_1_4
void FASTCALL call_5_15_1_4(SH2_struct* sh)
{
  SH2movll4(sh, 15,1, 4);
}
#define init_5_15_1_5
void FASTCALL call_5_15_1_5(SH2_struct* sh)
{
  SH2movll4(sh, 15,1, 5);
}
#define init_5_15_1_6
void FASTCALL call_5_15_1_6(SH2_struct* sh)
{
  SH2movll4(sh, 15,1, 6);
}
#define init_5_15_1_7
void FASTCALL call_5_15_1_7(SH2_struct* sh)
{
  SH2movll4(sh, 15,1, 7);
}
#define init_5_15_1_8
void FASTCALL call_5_15_1_8(SH2_struct* sh)
{
  SH2movll4(sh, 15,1, 8);
}
#define init_5_15_1_9
void FASTCALL call_5_15_1_9(SH2_struct* sh)
{
  SH2movll4(sh, 15,1, 9);
}
#define init_5_15_1_10
void FASTCALL call_5_15_1_10(SH2_struct* sh)
{
  SH2movll4(sh, 15,1, 10);
}
#define init_5_15_1_11
void FASTCALL call_5_15_1_11(SH2_struct* sh)
{
  SH2movll4(sh, 15,1, 11);
}
#define init_5_15_1_12
void FASTCALL call_5_15_1_12(SH2_struct* sh)
{
  SH2movll4(sh, 15,1, 12);
}
#define init_5_15_1_13
void FASTCALL call_5_15_1_13(SH2_struct* sh)
{
  SH2movll4(sh, 15,1, 13);
}
#define init_5_15_1_14
void FASTCALL call_5_15_1_14(SH2_struct* sh)
{
  SH2movll4(sh, 15,1, 14);
}
#define init_5_15_1_15
void FASTCALL call_5_15_1_15(SH2_struct* sh)
{
  SH2movll4(sh, 15,1, 15);
}
#define init_5_15_2_0
void FASTCALL call_5_15_2_0(SH2_struct* sh)
{
  SH2movll4(sh, 15,2, 0);
}
#define init_5_15_2_1
void FASTCALL call_5_15_2_1(SH2_struct* sh)
{
  SH2movll4(sh, 15,2, 1);
}
#define init_5_15_2_2
void FASTCALL call_5_15_2_2(SH2_struct* sh)
{
  SH2movll4(sh, 15,2, 2);
}
#define init_5_15_2_3
void FASTCALL call_5_15_2_3(SH2_struct* sh)
{
  SH2movll4(sh, 15,2, 3);
}
#define init_5_15_2_4
void FASTCALL call_5_15_2_4(SH2_struct* sh)
{
  SH2movll4(sh, 15,2, 4);
}
#define init_5_15_2_5
void FASTCALL call_5_15_2_5(SH2_struct* sh)
{
  SH2movll4(sh, 15,2, 5);
}
#define init_5_15_2_6
void FASTCALL call_5_15_2_6(SH2_struct* sh)
{
  SH2movll4(sh, 15,2, 6);
}
#define init_5_15_2_7
void FASTCALL call_5_15_2_7(SH2_struct* sh)
{
  SH2movll4(sh, 15,2, 7);
}
#define init_5_15_2_8
void FASTCALL call_5_15_2_8(SH2_struct* sh)
{
  SH2movll4(sh, 15,2, 8);
}
#define init_5_15_2_9
void FASTCALL call_5_15_2_9(SH2_struct* sh)
{
  SH2movll4(sh, 15,2, 9);
}
#define init_5_15_2_10
void FASTCALL call_5_15_2_10(SH2_struct* sh)
{
  SH2movll4(sh, 15,2, 10);
}
#define init_5_15_2_11
void FASTCALL call_5_15_2_11(SH2_struct* sh)
{
  SH2movll4(sh, 15,2, 11);
}
#define init_5_15_2_12
void FASTCALL call_5_15_2_12(SH2_struct* sh)
{
  SH2movll4(sh, 15,2, 12);
}
#define init_5_15_2_13
void FASTCALL call_5_15_2_13(SH2_struct* sh)
{
  SH2movll4(sh, 15,2, 13);
}
#define init_5_15_2_14
void FASTCALL call_5_15_2_14(SH2_struct* sh)
{
  SH2movll4(sh, 15,2, 14);
}
#define init_5_15_2_15
void FASTCALL call_5_15_2_15(SH2_struct* sh)
{
  SH2movll4(sh, 15,2, 15);
}
#define init_5_15_3_0
void FASTCALL call_5_15_3_0(SH2_struct* sh)
{
  SH2movll4(sh, 15,3, 0);
}
#define init_5_15_3_1
void FASTCALL call_5_15_3_1(SH2_struct* sh)
{
  SH2movll4(sh, 15,3, 1);
}
#define init_5_15_3_2
void FASTCALL call_5_15_3_2(SH2_struct* sh)
{
  SH2movll4(sh, 15,3, 2);
}
#define init_5_15_3_3
void FASTCALL call_5_15_3_3(SH2_struct* sh)
{
  SH2movll4(sh, 15,3, 3);
}
#define init_5_15_3_4
void FASTCALL call_5_15_3_4(SH2_struct* sh)
{
  SH2movll4(sh, 15,3, 4);
}
#define init_5_15_3_5
void FASTCALL call_5_15_3_5(SH2_struct* sh)
{
  SH2movll4(sh, 15,3, 5);
}
#define init_5_15_3_6
void FASTCALL call_5_15_3_6(SH2_struct* sh)
{
  SH2movll4(sh, 15,3, 6);
}
#define init_5_15_3_7
void FASTCALL call_5_15_3_7(SH2_struct* sh)
{
  SH2movll4(sh, 15,3, 7);
}
#define init_5_15_3_8
void FASTCALL call_5_15_3_8(SH2_struct* sh)
{
  SH2movll4(sh, 15,3, 8);
}
#define init_5_15_3_9
void FASTCALL call_5_15_3_9(SH2_struct* sh)
{
  SH2movll4(sh, 15,3, 9);
}
#define init_5_15_3_10
void FASTCALL call_5_15_3_10(SH2_struct* sh)
{
  SH2movll4(sh, 15,3, 10);
}
#define init_5_15_3_11
void FASTCALL call_5_15_3_11(SH2_struct* sh)
{
  SH2movll4(sh, 15,3, 11);
}
#define init_5_15_3_12
void FASTCALL call_5_15_3_12(SH2_struct* sh)
{
  SH2movll4(sh, 15,3, 12);
}
#define init_5_15_3_13
void FASTCALL call_5_15_3_13(SH2_struct* sh)
{
  SH2movll4(sh, 15,3, 13);
}
#define init_5_15_3_14
void FASTCALL call_5_15_3_14(SH2_struct* sh)
{
  SH2movll4(sh, 15,3, 14);
}
#define init_5_15_3_15
void FASTCALL call_5_15_3_15(SH2_struct* sh)
{
  SH2movll4(sh, 15,3, 15);
}
#define init_5_15_4_0
void FASTCALL call_5_15_4_0(SH2_struct* sh)
{
  SH2movll4(sh, 15,4, 0);
}
#define init_5_15_4_1
void FASTCALL call_5_15_4_1(SH2_struct* sh)
{
  SH2movll4(sh, 15,4, 1);
}
#define init_5_15_4_2
void FASTCALL call_5_15_4_2(SH2_struct* sh)
{
  SH2movll4(sh, 15,4, 2);
}
#define init_5_15_4_3
void FASTCALL call_5_15_4_3(SH2_struct* sh)
{
  SH2movll4(sh, 15,4, 3);
}
#define init_5_15_4_4
void FASTCALL call_5_15_4_4(SH2_struct* sh)
{
  SH2movll4(sh, 15,4, 4);
}
#define init_5_15_4_5
void FASTCALL call_5_15_4_5(SH2_struct* sh)
{
  SH2movll4(sh, 15,4, 5);
}
#define init_5_15_4_6
void FASTCALL call_5_15_4_6(SH2_struct* sh)
{
  SH2movll4(sh, 15,4, 6);
}
#define init_5_15_4_7
void FASTCALL call_5_15_4_7(SH2_struct* sh)
{
  SH2movll4(sh, 15,4, 7);
}
#define init_5_15_4_8
void FASTCALL call_5_15_4_8(SH2_struct* sh)
{
  SH2movll4(sh, 15,4, 8);
}
#define init_5_15_4_9
void FASTCALL call_5_15_4_9(SH2_struct* sh)
{
  SH2movll4(sh, 15,4, 9);
}
#define init_5_15_4_10
void FASTCALL call_5_15_4_10(SH2_struct* sh)
{
  SH2movll4(sh, 15,4, 10);
}
#define init_5_15_4_11
void FASTCALL call_5_15_4_11(SH2_struct* sh)
{
  SH2movll4(sh, 15,4, 11);
}
#define init_5_15_4_12
void FASTCALL call_5_15_4_12(SH2_struct* sh)
{
  SH2movll4(sh, 15,4, 12);
}
#define init_5_15_4_13
void FASTCALL call_5_15_4_13(SH2_struct* sh)
{
  SH2movll4(sh, 15,4, 13);
}
#define init_5_15_4_14
void FASTCALL call_5_15_4_14(SH2_struct* sh)
{
  SH2movll4(sh, 15,4, 14);
}
#define init_5_15_4_15
void FASTCALL call_5_15_4_15(SH2_struct* sh)
{
  SH2movll4(sh, 15,4, 15);
}
#define init_5_15_5_0
void FASTCALL call_5_15_5_0(SH2_struct* sh)
{
  SH2movll4(sh, 15,5, 0);
}
#define init_5_15_5_1
void FASTCALL call_5_15_5_1(SH2_struct* sh)
{
  SH2movll4(sh, 15,5, 1);
}
#define init_5_15_5_2
void FASTCALL call_5_15_5_2(SH2_struct* sh)
{
  SH2movll4(sh, 15,5, 2);
}
#define init_5_15_5_3
void FASTCALL call_5_15_5_3(SH2_struct* sh)
{
  SH2movll4(sh, 15,5, 3);
}
#define init_5_15_5_4
void FASTCALL call_5_15_5_4(SH2_struct* sh)
{
  SH2movll4(sh, 15,5, 4);
}
#define init_5_15_5_5
void FASTCALL call_5_15_5_5(SH2_struct* sh)
{
  SH2movll4(sh, 15,5, 5);
}
#define init_5_15_5_6
void FASTCALL call_5_15_5_6(SH2_struct* sh)
{
  SH2movll4(sh, 15,5, 6);
}
#define init_5_15_5_7
void FASTCALL call_5_15_5_7(SH2_struct* sh)
{
  SH2movll4(sh, 15,5, 7);
}
#define init_5_15_5_8
void FASTCALL call_5_15_5_8(SH2_struct* sh)
{
  SH2movll4(sh, 15,5, 8);
}
#define init_5_15_5_9
void FASTCALL call_5_15_5_9(SH2_struct* sh)
{
  SH2movll4(sh, 15,5, 9);
}
#define init_5_15_5_10
void FASTCALL call_5_15_5_10(SH2_struct* sh)
{
  SH2movll4(sh, 15,5, 10);
}
#define init_5_15_5_11
void FASTCALL call_5_15_5_11(SH2_struct* sh)
{
  SH2movll4(sh, 15,5, 11);
}
#define init_5_15_5_12
void FASTCALL call_5_15_5_12(SH2_struct* sh)
{
  SH2movll4(sh, 15,5, 12);
}
#define init_5_15_5_13
void FASTCALL call_5_15_5_13(SH2_struct* sh)
{
  SH2movll4(sh, 15,5, 13);
}
#define init_5_15_5_14
void FASTCALL call_5_15_5_14(SH2_struct* sh)
{
  SH2movll4(sh, 15,5, 14);
}
#define init_5_15_5_15
void FASTCALL call_5_15_5_15(SH2_struct* sh)
{
  SH2movll4(sh, 15,5, 15);
}
#define init_5_15_6_0
void FASTCALL call_5_15_6_0(SH2_struct* sh)
{
  SH2movll4(sh, 15,6, 0);
}
#define init_5_15_6_1
void FASTCALL call_5_15_6_1(SH2_struct* sh)
{
  SH2movll4(sh, 15,6, 1);
}
#define init_5_15_6_2
void FASTCALL call_5_15_6_2(SH2_struct* sh)
{
  SH2movll4(sh, 15,6, 2);
}
#define init_5_15_6_3
void FASTCALL call_5_15_6_3(SH2_struct* sh)
{
  SH2movll4(sh, 15,6, 3);
}
#define init_5_15_6_4
void FASTCALL call_5_15_6_4(SH2_struct* sh)
{
  SH2movll4(sh, 15,6, 4);
}
#define init_5_15_6_5
void FASTCALL call_5_15_6_5(SH2_struct* sh)
{
  SH2movll4(sh, 15,6, 5);
}
#define init_5_15_6_6
void FASTCALL call_5_15_6_6(SH2_struct* sh)
{
  SH2movll4(sh, 15,6, 6);
}
#define init_5_15_6_7
void FASTCALL call_5_15_6_7(SH2_struct* sh)
{
  SH2movll4(sh, 15,6, 7);
}
#define init_5_15_6_8
void FASTCALL call_5_15_6_8(SH2_struct* sh)
{
  SH2movll4(sh, 15,6, 8);
}
#define init_5_15_6_9
void FASTCALL call_5_15_6_9(SH2_struct* sh)
{
  SH2movll4(sh, 15,6, 9);
}
#define init_5_15_6_10
void FASTCALL call_5_15_6_10(SH2_struct* sh)
{
  SH2movll4(sh, 15,6, 10);
}
#define init_5_15_6_11
void FASTCALL call_5_15_6_11(SH2_struct* sh)
{
  SH2movll4(sh, 15,6, 11);
}
#define init_5_15_6_12
void FASTCALL call_5_15_6_12(SH2_struct* sh)
{
  SH2movll4(sh, 15,6, 12);
}
#define init_5_15_6_13
void FASTCALL call_5_15_6_13(SH2_struct* sh)
{
  SH2movll4(sh, 15,6, 13);
}
#define init_5_15_6_14
void FASTCALL call_5_15_6_14(SH2_struct* sh)
{
  SH2movll4(sh, 15,6, 14);
}
#define init_5_15_6_15
void FASTCALL call_5_15_6_15(SH2_struct* sh)
{
  SH2movll4(sh, 15,6, 15);
}
#define init_5_15_7_0
void FASTCALL call_5_15_7_0(SH2_struct* sh)
{
  SH2movll4(sh, 15,7, 0);
}
#define init_5_15_7_1
void FASTCALL call_5_15_7_1(SH2_struct* sh)
{
  SH2movll4(sh, 15,7, 1);
}
#define init_5_15_7_2
void FASTCALL call_5_15_7_2(SH2_struct* sh)
{
  SH2movll4(sh, 15,7, 2);
}
#define init_5_15_7_3
void FASTCALL call_5_15_7_3(SH2_struct* sh)
{
  SH2movll4(sh, 15,7, 3);
}
#define init_5_15_7_4
void FASTCALL call_5_15_7_4(SH2_struct* sh)
{
  SH2movll4(sh, 15,7, 4);
}
#define init_5_15_7_5
void FASTCALL call_5_15_7_5(SH2_struct* sh)
{
  SH2movll4(sh, 15,7, 5);
}
#define init_5_15_7_6
void FASTCALL call_5_15_7_6(SH2_struct* sh)
{
  SH2movll4(sh, 15,7, 6);
}
#define init_5_15_7_7
void FASTCALL call_5_15_7_7(SH2_struct* sh)
{
  SH2movll4(sh, 15,7, 7);
}
#define init_5_15_7_8
void FASTCALL call_5_15_7_8(SH2_struct* sh)
{
  SH2movll4(sh, 15,7, 8);
}
#define init_5_15_7_9
void FASTCALL call_5_15_7_9(SH2_struct* sh)
{
  SH2movll4(sh, 15,7, 9);
}
#define init_5_15_7_10
void FASTCALL call_5_15_7_10(SH2_struct* sh)
{
  SH2movll4(sh, 15,7, 10);
}
#define init_5_15_7_11
void FASTCALL call_5_15_7_11(SH2_struct* sh)
{
  SH2movll4(sh, 15,7, 11);
}
#define init_5_15_7_12
void FASTCALL call_5_15_7_12(SH2_struct* sh)
{
  SH2movll4(sh, 15,7, 12);
}
#define init_5_15_7_13
void FASTCALL call_5_15_7_13(SH2_struct* sh)
{
  SH2movll4(sh, 15,7, 13);
}
#define init_5_15_7_14
void FASTCALL call_5_15_7_14(SH2_struct* sh)
{
  SH2movll4(sh, 15,7, 14);
}
#define init_5_15_7_15
void FASTCALL call_5_15_7_15(SH2_struct* sh)
{
  SH2movll4(sh, 15,7, 15);
}
#define init_5_15_8_0
void FASTCALL call_5_15_8_0(SH2_struct* sh)
{
  SH2movll4(sh, 15,8, 0);
}
#define init_5_15_8_1
void FASTCALL call_5_15_8_1(SH2_struct* sh)
{
  SH2movll4(sh, 15,8, 1);
}
#define init_5_15_8_2
void FASTCALL call_5_15_8_2(SH2_struct* sh)
{
  SH2movll4(sh, 15,8, 2);
}
#define init_5_15_8_3
void FASTCALL call_5_15_8_3(SH2_struct* sh)
{
  SH2movll4(sh, 15,8, 3);
}
#define init_5_15_8_4
void FASTCALL call_5_15_8_4(SH2_struct* sh)
{
  SH2movll4(sh, 15,8, 4);
}
#define init_5_15_8_5
void FASTCALL call_5_15_8_5(SH2_struct* sh)
{
  SH2movll4(sh, 15,8, 5);
}
#define init_5_15_8_6
void FASTCALL call_5_15_8_6(SH2_struct* sh)
{
  SH2movll4(sh, 15,8, 6);
}
#define init_5_15_8_7
void FASTCALL call_5_15_8_7(SH2_struct* sh)
{
  SH2movll4(sh, 15,8, 7);
}
#define init_5_15_8_8
void FASTCALL call_5_15_8_8(SH2_struct* sh)
{
  SH2movll4(sh, 15,8, 8);
}
#define init_5_15_8_9
void FASTCALL call_5_15_8_9(SH2_struct* sh)
{
  SH2movll4(sh, 15,8, 9);
}
#define init_5_15_8_10
void FASTCALL call_5_15_8_10(SH2_struct* sh)
{
  SH2movll4(sh, 15,8, 10);
}
#define init_5_15_8_11
void FASTCALL call_5_15_8_11(SH2_struct* sh)
{
  SH2movll4(sh, 15,8, 11);
}
#define init_5_15_8_12
void FASTCALL call_5_15_8_12(SH2_struct* sh)
{
  SH2movll4(sh, 15,8, 12);
}
#define init_5_15_8_13
void FASTCALL call_5_15_8_13(SH2_struct* sh)
{
  SH2movll4(sh, 15,8, 13);
}
#define init_5_15_8_14
void FASTCALL call_5_15_8_14(SH2_struct* sh)
{
  SH2movll4(sh, 15,8, 14);
}
#define init_5_15_8_15
void FASTCALL call_5_15_8_15(SH2_struct* sh)
{
  SH2movll4(sh, 15,8, 15);
}
#define init_5_15_9_0
void FASTCALL call_5_15_9_0(SH2_struct* sh)
{
  SH2movll4(sh, 15,9, 0);
}
#define init_5_15_9_1
void FASTCALL call_5_15_9_1(SH2_struct* sh)
{
  SH2movll4(sh, 15,9, 1);
}
#define init_5_15_9_2
void FASTCALL call_5_15_9_2(SH2_struct* sh)
{
  SH2movll4(sh, 15,9, 2);
}
#define init_5_15_9_3
void FASTCALL call_5_15_9_3(SH2_struct* sh)
{
  SH2movll4(sh, 15,9, 3);
}
#define init_5_15_9_4
void FASTCALL call_5_15_9_4(SH2_struct* sh)
{
  SH2movll4(sh, 15,9, 4);
}
#define init_5_15_9_5
void FASTCALL call_5_15_9_5(SH2_struct* sh)
{
  SH2movll4(sh, 15,9, 5);
}
#define init_5_15_9_6
void FASTCALL call_5_15_9_6(SH2_struct* sh)
{
  SH2movll4(sh, 15,9, 6);
}
#define init_5_15_9_7
void FASTCALL call_5_15_9_7(SH2_struct* sh)
{
  SH2movll4(sh, 15,9, 7);
}
#define init_5_15_9_8
void FASTCALL call_5_15_9_8(SH2_struct* sh)
{
  SH2movll4(sh, 15,9, 8);
}
#define init_5_15_9_9
void FASTCALL call_5_15_9_9(SH2_struct* sh)
{
  SH2movll4(sh, 15,9, 9);
}
#define init_5_15_9_10
void FASTCALL call_5_15_9_10(SH2_struct* sh)
{
  SH2movll4(sh, 15,9, 10);
}
#define init_5_15_9_11
void FASTCALL call_5_15_9_11(SH2_struct* sh)
{
  SH2movll4(sh, 15,9, 11);
}
#define init_5_15_9_12
void FASTCALL call_5_15_9_12(SH2_struct* sh)
{
  SH2movll4(sh, 15,9, 12);
}
#define init_5_15_9_13
void FASTCALL call_5_15_9_13(SH2_struct* sh)
{
  SH2movll4(sh, 15,9, 13);
}
#define init_5_15_9_14
void FASTCALL call_5_15_9_14(SH2_struct* sh)
{
  SH2movll4(sh, 15,9, 14);
}
#define init_5_15_9_15
void FASTCALL call_5_15_9_15(SH2_struct* sh)
{
  SH2movll4(sh, 15,9, 15);
}
#define init_5_15_10_0
void FASTCALL call_5_15_10_0(SH2_struct* sh)
{
  SH2movll4(sh, 15,10, 0);
}
#define init_5_15_10_1
void FASTCALL call_5_15_10_1(SH2_struct* sh)
{
  SH2movll4(sh, 15,10, 1);
}
#define init_5_15_10_2
void FASTCALL call_5_15_10_2(SH2_struct* sh)
{
  SH2movll4(sh, 15,10, 2);
}
#define init_5_15_10_3
void FASTCALL call_5_15_10_3(SH2_struct* sh)
{
  SH2movll4(sh, 15,10, 3);
}
#define init_5_15_10_4
void FASTCALL call_5_15_10_4(SH2_struct* sh)
{
  SH2movll4(sh, 15,10, 4);
}
#define init_5_15_10_5
void FASTCALL call_5_15_10_5(SH2_struct* sh)
{
  SH2movll4(sh, 15,10, 5);
}
#define init_5_15_10_6
void FASTCALL call_5_15_10_6(SH2_struct* sh)
{
  SH2movll4(sh, 15,10, 6);
}
#define init_5_15_10_7
void FASTCALL call_5_15_10_7(SH2_struct* sh)
{
  SH2movll4(sh, 15,10, 7);
}
#define init_5_15_10_8
void FASTCALL call_5_15_10_8(SH2_struct* sh)
{
  SH2movll4(sh, 15,10, 8);
}
#define init_5_15_10_9
void FASTCALL call_5_15_10_9(SH2_struct* sh)
{
  SH2movll4(sh, 15,10, 9);
}
#define init_5_15_10_10
void FASTCALL call_5_15_10_10(SH2_struct* sh)
{
  SH2movll4(sh, 15,10, 10);
}
#define init_5_15_10_11
void FASTCALL call_5_15_10_11(SH2_struct* sh)
{
  SH2movll4(sh, 15,10, 11);
}
#define init_5_15_10_12
void FASTCALL call_5_15_10_12(SH2_struct* sh)
{
  SH2movll4(sh, 15,10, 12);
}
#define init_5_15_10_13
void FASTCALL call_5_15_10_13(SH2_struct* sh)
{
  SH2movll4(sh, 15,10, 13);
}
#define init_5_15_10_14
void FASTCALL call_5_15_10_14(SH2_struct* sh)
{
  SH2movll4(sh, 15,10, 14);
}
#define init_5_15_10_15
void FASTCALL call_5_15_10_15(SH2_struct* sh)
{
  SH2movll4(sh, 15,10, 15);
}
#define init_5_15_11_0
void FASTCALL call_5_15_11_0(SH2_struct* sh)
{
  SH2movll4(sh, 15,11, 0);
}
#define init_5_15_11_1
void FASTCALL call_5_15_11_1(SH2_struct* sh)
{
  SH2movll4(sh, 15,11, 1);
}
#define init_5_15_11_2
void FASTCALL call_5_15_11_2(SH2_struct* sh)
{
  SH2movll4(sh, 15,11, 2);
}
#define init_5_15_11_3
void FASTCALL call_5_15_11_3(SH2_struct* sh)
{
  SH2movll4(sh, 15,11, 3);
}
#define init_5_15_11_4
void FASTCALL call_5_15_11_4(SH2_struct* sh)
{
  SH2movll4(sh, 15,11, 4);
}
#define init_5_15_11_5
void FASTCALL call_5_15_11_5(SH2_struct* sh)
{
  SH2movll4(sh, 15,11, 5);
}
#define init_5_15_11_6
void FASTCALL call_5_15_11_6(SH2_struct* sh)
{
  SH2movll4(sh, 15,11, 6);
}
#define init_5_15_11_7
void FASTCALL call_5_15_11_7(SH2_struct* sh)
{
  SH2movll4(sh, 15,11, 7);
}
#define init_5_15_11_8
void FASTCALL call_5_15_11_8(SH2_struct* sh)
{
  SH2movll4(sh, 15,11, 8);
}
#define init_5_15_11_9
void FASTCALL call_5_15_11_9(SH2_struct* sh)
{
  SH2movll4(sh, 15,11, 9);
}
#define init_5_15_11_10
void FASTCALL call_5_15_11_10(SH2_struct* sh)
{
  SH2movll4(sh, 15,11, 10);
}
#define init_5_15_11_11
void FASTCALL call_5_15_11_11(SH2_struct* sh)
{
  SH2movll4(sh, 15,11, 11);
}
#define init_5_15_11_12
void FASTCALL call_5_15_11_12(SH2_struct* sh)
{
  SH2movll4(sh, 15,11, 12);
}
#define init_5_15_11_13
void FASTCALL call_5_15_11_13(SH2_struct* sh)
{
  SH2movll4(sh, 15,11, 13);
}
#define init_5_15_11_14
void FASTCALL call_5_15_11_14(SH2_struct* sh)
{
  SH2movll4(sh, 15,11, 14);
}
#define init_5_15_11_15
void FASTCALL call_5_15_11_15(SH2_struct* sh)
{
  SH2movll4(sh, 15,11, 15);
}
#define init_5_15_12_0
void FASTCALL call_5_15_12_0(SH2_struct* sh)
{
  SH2movll4(sh, 15,12, 0);
}
#define init_5_15_12_1
void FASTCALL call_5_15_12_1(SH2_struct* sh)
{
  SH2movll4(sh, 15,12, 1);
}
#define init_5_15_12_2
void FASTCALL call_5_15_12_2(SH2_struct* sh)
{
  SH2movll4(sh, 15,12, 2);
}
#define init_5_15_12_3
void FASTCALL call_5_15_12_3(SH2_struct* sh)
{
  SH2movll4(sh, 15,12, 3);
}
#define init_5_15_12_4
void FASTCALL call_5_15_12_4(SH2_struct* sh)
{
  SH2movll4(sh, 15,12, 4);
}
#define init_5_15_12_5
void FASTCALL call_5_15_12_5(SH2_struct* sh)
{
  SH2movll4(sh, 15,12, 5);
}
#define init_5_15_12_6
void FASTCALL call_5_15_12_6(SH2_struct* sh)
{
  SH2movll4(sh, 15,12, 6);
}
#define init_5_15_12_7
void FASTCALL call_5_15_12_7(SH2_struct* sh)
{
  SH2movll4(sh, 15,12, 7);
}
#define init_5_15_12_8
void FASTCALL call_5_15_12_8(SH2_struct* sh)
{
  SH2movll4(sh, 15,12, 8);
}
#define init_5_15_12_9
void FASTCALL call_5_15_12_9(SH2_struct* sh)
{
  SH2movll4(sh, 15,12, 9);
}
#define init_5_15_12_10
void FASTCALL call_5_15_12_10(SH2_struct* sh)
{
  SH2movll4(sh, 15,12, 10);
}
#define init_5_15_12_11
void FASTCALL call_5_15_12_11(SH2_struct* sh)
{
  SH2movll4(sh, 15,12, 11);
}
#define init_5_15_12_12
void FASTCALL call_5_15_12_12(SH2_struct* sh)
{
  SH2movll4(sh, 15,12, 12);
}
#define init_5_15_12_13
void FASTCALL call_5_15_12_13(SH2_struct* sh)
{
  SH2movll4(sh, 15,12, 13);
}
#define init_5_15_12_14
void FASTCALL call_5_15_12_14(SH2_struct* sh)
{
  SH2movll4(sh, 15,12, 14);
}
#define init_5_15_12_15
void FASTCALL call_5_15_12_15(SH2_struct* sh)
{
  SH2movll4(sh, 15,12, 15);
}
#define init_5_15_13_0
void FASTCALL call_5_15_13_0(SH2_struct* sh)
{
  SH2movll4(sh, 15,13, 0);
}
#define init_5_15_13_1
void FASTCALL call_5_15_13_1(SH2_struct* sh)
{
  SH2movll4(sh, 15,13, 1);
}
#define init_5_15_13_2
void FASTCALL call_5_15_13_2(SH2_struct* sh)
{
  SH2movll4(sh, 15,13, 2);
}
#define init_5_15_13_3
void FASTCALL call_5_15_13_3(SH2_struct* sh)
{
  SH2movll4(sh, 15,13, 3);
}
#define init_5_15_13_4
void FASTCALL call_5_15_13_4(SH2_struct* sh)
{
  SH2movll4(sh, 15,13, 4);
}
#define init_5_15_13_5
void FASTCALL call_5_15_13_5(SH2_struct* sh)
{
  SH2movll4(sh, 15,13, 5);
}
#define init_5_15_13_6
void FASTCALL call_5_15_13_6(SH2_struct* sh)
{
  SH2movll4(sh, 15,13, 6);
}
#define init_5_15_13_7
void FASTCALL call_5_15_13_7(SH2_struct* sh)
{
  SH2movll4(sh, 15,13, 7);
}
#define init_5_15_13_8
void FASTCALL call_5_15_13_8(SH2_struct* sh)
{
  SH2movll4(sh, 15,13, 8);
}
#define init_5_15_13_9
void FASTCALL call_5_15_13_9(SH2_struct* sh)
{
  SH2movll4(sh, 15,13, 9);
}
#define init_5_15_13_10
void FASTCALL call_5_15_13_10(SH2_struct* sh)
{
  SH2movll4(sh, 15,13, 10);
}
#define init_5_15_13_11
void FASTCALL call_5_15_13_11(SH2_struct* sh)
{
  SH2movll4(sh, 15,13, 11);
}
#define init_5_15_13_12
void FASTCALL call_5_15_13_12(SH2_struct* sh)
{
  SH2movll4(sh, 15,13, 12);
}
#define init_5_15_13_13
void FASTCALL call_5_15_13_13(SH2_struct* sh)
{
  SH2movll4(sh, 15,13, 13);
}
#define init_5_15_13_14
void FASTCALL call_5_15_13_14(SH2_struct* sh)
{
  SH2movll4(sh, 15,13, 14);
}
#define init_5_15_13_15
void FASTCALL call_5_15_13_15(SH2_struct* sh)
{
  SH2movll4(sh, 15,13, 15);
}
#define init_5_15_14_0
void FASTCALL call_5_15_14_0(SH2_struct* sh)
{
  SH2movll4(sh, 15,14, 0);
}
#define init_5_15_14_1
void FASTCALL call_5_15_14_1(SH2_struct* sh)
{
  SH2movll4(sh, 15,14, 1);
}
#define init_5_15_14_2
void FASTCALL call_5_15_14_2(SH2_struct* sh)
{
  SH2movll4(sh, 15,14, 2);
}
#define init_5_15_14_3
void FASTCALL call_5_15_14_3(SH2_struct* sh)
{
  SH2movll4(sh, 15,14, 3);
}
#define init_5_15_14_4
void FASTCALL call_5_15_14_4(SH2_struct* sh)
{
  SH2movll4(sh, 15,14, 4);
}
#define init_5_15_14_5
void FASTCALL call_5_15_14_5(SH2_struct* sh)
{
  SH2movll4(sh, 15,14, 5);
}
#define init_5_15_14_6
void FASTCALL call_5_15_14_6(SH2_struct* sh)
{
  SH2movll4(sh, 15,14, 6);
}
#define init_5_15_14_7
void FASTCALL call_5_15_14_7(SH2_struct* sh)
{
  SH2movll4(sh, 15,14, 7);
}
#define init_5_15_14_8
void FASTCALL call_5_15_14_8(SH2_struct* sh)
{
  SH2movll4(sh, 15,14, 8);
}
#define init_5_15_14_9
void FASTCALL call_5_15_14_9(SH2_struct* sh)
{
  SH2movll4(sh, 15,14, 9);
}
#define init_5_15_14_10
void FASTCALL call_5_15_14_10(SH2_struct* sh)
{
  SH2movll4(sh, 15,14, 10);
}
#define init_5_15_14_11
void FASTCALL call_5_15_14_11(SH2_struct* sh)
{
  SH2movll4(sh, 15,14, 11);
}
#define init_5_15_14_12
void FASTCALL call_5_15_14_12(SH2_struct* sh)
{
  SH2movll4(sh, 15,14, 12);
}
#define init_5_15_14_13
void FASTCALL call_5_15_14_13(SH2_struct* sh)
{
  SH2movll4(sh, 15,14, 13);
}
#define init_5_15_14_14
void FASTCALL call_5_15_14_14(SH2_struct* sh)
{
  SH2movll4(sh, 15,14, 14);
}
#define init_5_15_14_15
void FASTCALL call_5_15_14_15(SH2_struct* sh)
{
  SH2movll4(sh, 15,14, 15);
}
#define init_5_15_15_0
void FASTCALL call_5_15_15_0(SH2_struct* sh)
{
  SH2movll4(sh, 15,15, 0);
}
#define init_5_15_15_1
void FASTCALL call_5_15_15_1(SH2_struct* sh)
{
  SH2movll4(sh, 15,15, 1);
}
#define init_5_15_15_2
void FASTCALL call_5_15_15_2(SH2_struct* sh)
{
  SH2movll4(sh, 15,15, 2);
}
#define init_5_15_15_3
void FASTCALL call_5_15_15_3(SH2_struct* sh)
{
  SH2movll4(sh, 15,15, 3);
}
#define init_5_15_15_4
void FASTCALL call_5_15_15_4(SH2_struct* sh)
{
  SH2movll4(sh, 15,15, 4);
}
#define init_5_15_15_5
void FASTCALL call_5_15_15_5(SH2_struct* sh)
{
  SH2movll4(sh, 15,15, 5);
}
#define init_5_15_15_6
void FASTCALL call_5_15_15_6(SH2_struct* sh)
{
  SH2movll4(sh, 15,15, 6);
}
#define init_5_15_15_7
void FASTCALL call_5_15_15_7(SH2_struct* sh)
{
  SH2movll4(sh, 15,15, 7);
}
#define init_5_15_15_8
void FASTCALL call_5_15_15_8(SH2_struct* sh)
{
  SH2movll4(sh, 15,15, 8);
}
#define init_5_15_15_9
void FASTCALL call_5_15_15_9(SH2_struct* sh)
{
  SH2movll4(sh, 15,15, 9);
}
#define init_5_15_15_10
void FASTCALL call_5_15_15_10(SH2_struct* sh)
{
  SH2movll4(sh, 15,15, 10);
}
#define init_5_15_15_11
void FASTCALL call_5_15_15_11(SH2_struct* sh)
{
  SH2movll4(sh, 15,15, 11);
}
#define init_5_15_15_12
void FASTCALL call_5_15_15_12(SH2_struct* sh)
{
  SH2movll4(sh, 15,15, 12);
}
#define init_5_15_15_13
void FASTCALL call_5_15_15_13(SH2_struct* sh)
{
  SH2movll4(sh, 15,15, 13);
}
#define init_5_15_15_14
void FASTCALL call_5_15_15_14(SH2_struct* sh)
{
  SH2movll4(sh, 15,15, 14);
}
#define init_5_15_15_15
void FASTCALL call_5_15_15_15(SH2_struct* sh)
{
  SH2movll4(sh, 15,15, 15);
}
#define init_6_0_0_2
void FASTCALL call_6_0_0_2(SH2_struct* sh)
{
  SH2movll(sh,0,0);
}
#define init_6_0_1_2
void FASTCALL call_6_0_1_2(SH2_struct* sh)
{
  SH2movll(sh,0,1);
}
#define init_6_0_2_2
void FASTCALL call_6_0_2_2(SH2_struct* sh)
{
  SH2movll(sh,0,2);
}
#define init_6_0_3_2
void FASTCALL call_6_0_3_2(SH2_struct* sh)
{
  SH2movll(sh,0,3);
}
#define init_6_0_4_2
void FASTCALL call_6_0_4_2(SH2_struct* sh)
{
  SH2movll(sh,0,4);
}
#define init_6_0_5_2
void FASTCALL call_6_0_5_2(SH2_struct* sh)
{
  SH2movll(sh,0,5);
}
#define init_6_0_6_2
void FASTCALL call_6_0_6_2(SH2_struct* sh)
{
  SH2movll(sh,0,6);
}
#define init_6_0_7_2
void FASTCALL call_6_0_7_2(SH2_struct* sh)
{
  SH2movll(sh,0,7);
}
#define init_6_0_8_2
void FASTCALL call_6_0_8_2(SH2_struct* sh)
{
  SH2movll(sh,0,8);
}
#define init_6_0_9_2
void FASTCALL call_6_0_9_2(SH2_struct* sh)
{
  SH2movll(sh,0,9);
}
#define init_6_0_10_2
void FASTCALL call_6_0_10_2(SH2_struct* sh)
{
  SH2movll(sh,0,10);
}
#define init_6_0_11_2
void FASTCALL call_6_0_11_2(SH2_struct* sh)
{
  SH2movll(sh,0,11);
}
#define init_6_0_12_2
void FASTCALL call_6_0_12_2(SH2_struct* sh)
{
  SH2movll(sh,0,12);
}
#define init_6_0_13_2
void FASTCALL call_6_0_13_2(SH2_struct* sh)
{
  SH2movll(sh,0,13);
}
#define init_6_0_14_2
void FASTCALL call_6_0_14_2(SH2_struct* sh)
{
  SH2movll(sh,0,14);
}
#define init_6_0_15_2
void FASTCALL call_6_0_15_2(SH2_struct* sh)
{
  SH2movll(sh,0,15);
}
#define init_6_1_0_2
void FASTCALL call_6_1_0_2(SH2_struct* sh)
{
  SH2movll(sh,1,0);
}
#define init_6_1_1_2
void FASTCALL call_6_1_1_2(SH2_struct* sh)
{
  SH2movll(sh,1,1);
}
#define init_6_1_2_2
void FASTCALL call_6_1_2_2(SH2_struct* sh)
{
  SH2movll(sh,1,2);
}
#define init_6_1_3_2
void FASTCALL call_6_1_3_2(SH2_struct* sh)
{
  SH2movll(sh,1,3);
}
#define init_6_1_4_2
void FASTCALL call_6_1_4_2(SH2_struct* sh)
{
  SH2movll(sh,1,4);
}
#define init_6_1_5_2
void FASTCALL call_6_1_5_2(SH2_struct* sh)
{
  SH2movll(sh,1,5);
}
#define init_6_1_6_2
void FASTCALL call_6_1_6_2(SH2_struct* sh)
{
  SH2movll(sh,1,6);
}
#define init_6_1_7_2
void FASTCALL call_6_1_7_2(SH2_struct* sh)
{
  SH2movll(sh,1,7);
}
#define init_6_1_8_2
void FASTCALL call_6_1_8_2(SH2_struct* sh)
{
  SH2movll(sh,1,8);
}
#define init_6_1_9_2
void FASTCALL call_6_1_9_2(SH2_struct* sh)
{
  SH2movll(sh,1,9);
}
#define init_6_1_10_2
void FASTCALL call_6_1_10_2(SH2_struct* sh)
{
  SH2movll(sh,1,10);
}
#define init_6_1_11_2
void FASTCALL call_6_1_11_2(SH2_struct* sh)
{
  SH2movll(sh,1,11);
}
#define init_6_1_12_2
void FASTCALL call_6_1_12_2(SH2_struct* sh)
{
  SH2movll(sh,1,12);
}
#define init_6_1_13_2
void FASTCALL call_6_1_13_2(SH2_struct* sh)
{
  SH2movll(sh,1,13);
}
#define init_6_1_14_2
void FASTCALL call_6_1_14_2(SH2_struct* sh)
{
  SH2movll(sh,1,14);
}
#define init_6_1_15_2
void FASTCALL call_6_1_15_2(SH2_struct* sh)
{
  SH2movll(sh,1,15);
}
#define init_6_2_0_2
void FASTCALL call_6_2_0_2(SH2_struct* sh)
{
  SH2movll(sh,2,0);
}
#define init_6_2_1_2
void FASTCALL call_6_2_1_2(SH2_struct* sh)
{
  SH2movll(sh,2,1);
}
#define init_6_2_2_2
void FASTCALL call_6_2_2_2(SH2_struct* sh)
{
  SH2movll(sh,2,2);
}
#define init_6_2_3_2
void FASTCALL call_6_2_3_2(SH2_struct* sh)
{
  SH2movll(sh,2,3);
}
#define init_6_2_4_2
void FASTCALL call_6_2_4_2(SH2_struct* sh)
{
  SH2movll(sh,2,4);
}
#define init_6_2_5_2
void FASTCALL call_6_2_5_2(SH2_struct* sh)
{
  SH2movll(sh,2,5);
}
#define init_6_2_6_2
void FASTCALL call_6_2_6_2(SH2_struct* sh)
{
  SH2movll(sh,2,6);
}
#define init_6_2_7_2
void FASTCALL call_6_2_7_2(SH2_struct* sh)
{
  SH2movll(sh,2,7);
}
#define init_6_2_8_2
void FASTCALL call_6_2_8_2(SH2_struct* sh)
{
  SH2movll(sh,2,8);
}
#define init_6_2_9_2
void FASTCALL call_6_2_9_2(SH2_struct* sh)
{
  SH2movll(sh,2,9);
}
#define init_6_2_10_2
void FASTCALL call_6_2_10_2(SH2_struct* sh)
{
  SH2movll(sh,2,10);
}
#define init_6_2_11_2
void FASTCALL call_6_2_11_2(SH2_struct* sh)
{
  SH2movll(sh,2,11);
}
#define init_6_2_12_2
void FASTCALL call_6_2_12_2(SH2_struct* sh)
{
  SH2movll(sh,2,12);
}
#define init_6_2_13_2
void FASTCALL call_6_2_13_2(SH2_struct* sh)
{
  SH2movll(sh,2,13);
}
#define init_6_2_14_2
void FASTCALL call_6_2_14_2(SH2_struct* sh)
{
  SH2movll(sh,2,14);
}
#define init_6_2_15_2
void FASTCALL call_6_2_15_2(SH2_struct* sh)
{
  SH2movll(sh,2,15);
}
#define init_6_3_0_2
void FASTCALL call_6_3_0_2(SH2_struct* sh)
{
  SH2movll(sh,3,0);
}
#define init_6_3_1_2
void FASTCALL call_6_3_1_2(SH2_struct* sh)
{
  SH2movll(sh,3,1);
}
#define init_6_3_2_2
void FASTCALL call_6_3_2_2(SH2_struct* sh)
{
  SH2movll(sh,3,2);
}
#define init_6_3_3_2
void FASTCALL call_6_3_3_2(SH2_struct* sh)
{
  SH2movll(sh,3,3);
}
#define init_6_3_4_2
void FASTCALL call_6_3_4_2(SH2_struct* sh)
{
  SH2movll(sh,3,4);
}
#define init_6_3_5_2
void FASTCALL call_6_3_5_2(SH2_struct* sh)
{
  SH2movll(sh,3,5);
}
#define init_6_3_6_2
void FASTCALL call_6_3_6_2(SH2_struct* sh)
{
  SH2movll(sh,3,6);
}
#define init_6_3_7_2
void FASTCALL call_6_3_7_2(SH2_struct* sh)
{
  SH2movll(sh,3,7);
}
#define init_6_3_8_2
void FASTCALL call_6_3_8_2(SH2_struct* sh)
{
  SH2movll(sh,3,8);
}
#define init_6_3_9_2
void FASTCALL call_6_3_9_2(SH2_struct* sh)
{
  SH2movll(sh,3,9);
}
#define init_6_3_10_2
void FASTCALL call_6_3_10_2(SH2_struct* sh)
{
  SH2movll(sh,3,10);
}
#define init_6_3_11_2
void FASTCALL call_6_3_11_2(SH2_struct* sh)
{
  SH2movll(sh,3,11);
}
#define init_6_3_12_2
void FASTCALL call_6_3_12_2(SH2_struct* sh)
{
  SH2movll(sh,3,12);
}
#define init_6_3_13_2
void FASTCALL call_6_3_13_2(SH2_struct* sh)
{
  SH2movll(sh,3,13);
}
#define init_6_3_14_2
void FASTCALL call_6_3_14_2(SH2_struct* sh)
{
  SH2movll(sh,3,14);
}
#define init_6_3_15_2
void FASTCALL call_6_3_15_2(SH2_struct* sh)
{
  SH2movll(sh,3,15);
}
#define init_6_4_0_2
void FASTCALL call_6_4_0_2(SH2_struct* sh)
{
  SH2movll(sh,4,0);
}
#define init_6_4_1_2
void FASTCALL call_6_4_1_2(SH2_struct* sh)
{
  SH2movll(sh,4,1);
}
#define init_6_4_2_2
void FASTCALL call_6_4_2_2(SH2_struct* sh)
{
  SH2movll(sh,4,2);
}
#define init_6_4_3_2
void FASTCALL call_6_4_3_2(SH2_struct* sh)
{
  SH2movll(sh,4,3);
}
#define init_6_4_4_2
void FASTCALL call_6_4_4_2(SH2_struct* sh)
{
  SH2movll(sh,4,4);
}
#define init_6_4_5_2
void FASTCALL call_6_4_5_2(SH2_struct* sh)
{
  SH2movll(sh,4,5);
}
#define init_6_4_6_2
void FASTCALL call_6_4_6_2(SH2_struct* sh)
{
  SH2movll(sh,4,6);
}
#define init_6_4_7_2
void FASTCALL call_6_4_7_2(SH2_struct* sh)
{
  SH2movll(sh,4,7);
}
#define init_6_4_8_2
void FASTCALL call_6_4_8_2(SH2_struct* sh)
{
  SH2movll(sh,4,8);
}
#define init_6_4_9_2
void FASTCALL call_6_4_9_2(SH2_struct* sh)
{
  SH2movll(sh,4,9);
}
#define init_6_4_10_2
void FASTCALL call_6_4_10_2(SH2_struct* sh)
{
  SH2movll(sh,4,10);
}
#define init_6_4_11_2
void FASTCALL call_6_4_11_2(SH2_struct* sh)
{
  SH2movll(sh,4,11);
}
#define init_6_4_12_2
void FASTCALL call_6_4_12_2(SH2_struct* sh)
{
  SH2movll(sh,4,12);
}
#define init_6_4_13_2
void FASTCALL call_6_4_13_2(SH2_struct* sh)
{
  SH2movll(sh,4,13);
}
#define init_6_4_14_2
void FASTCALL call_6_4_14_2(SH2_struct* sh)
{
  SH2movll(sh,4,14);
}
#define init_6_4_15_2
void FASTCALL call_6_4_15_2(SH2_struct* sh)
{
  SH2movll(sh,4,15);
}
#define init_6_5_0_2
void FASTCALL call_6_5_0_2(SH2_struct* sh)
{
  SH2movll(sh,5,0);
}
#define init_6_5_1_2
void FASTCALL call_6_5_1_2(SH2_struct* sh)
{
  SH2movll(sh,5,1);
}
#define init_6_5_2_2
void FASTCALL call_6_5_2_2(SH2_struct* sh)
{
  SH2movll(sh,5,2);
}
#define init_6_5_3_2
void FASTCALL call_6_5_3_2(SH2_struct* sh)
{
  SH2movll(sh,5,3);
}
#define init_6_5_4_2
void FASTCALL call_6_5_4_2(SH2_struct* sh)
{
  SH2movll(sh,5,4);
}
#define init_6_5_5_2
void FASTCALL call_6_5_5_2(SH2_struct* sh)
{
  SH2movll(sh,5,5);
}
#define init_6_5_6_2
void FASTCALL call_6_5_6_2(SH2_struct* sh)
{
  SH2movll(sh,5,6);
}
#define init_6_5_7_2
void FASTCALL call_6_5_7_2(SH2_struct* sh)
{
  SH2movll(sh,5,7);
}
#define init_6_5_8_2
void FASTCALL call_6_5_8_2(SH2_struct* sh)
{
  SH2movll(sh,5,8);
}
#define init_6_5_9_2
void FASTCALL call_6_5_9_2(SH2_struct* sh)
{
  SH2movll(sh,5,9);
}
#define init_6_5_10_2
void FASTCALL call_6_5_10_2(SH2_struct* sh)
{
  SH2movll(sh,5,10);
}
#define init_6_5_11_2
void FASTCALL call_6_5_11_2(SH2_struct* sh)
{
  SH2movll(sh,5,11);
}
#define init_6_5_12_2
void FASTCALL call_6_5_12_2(SH2_struct* sh)
{
  SH2movll(sh,5,12);
}
#define init_6_5_13_2
void FASTCALL call_6_5_13_2(SH2_struct* sh)
{
  SH2movll(sh,5,13);
}
#define init_6_5_14_2
void FASTCALL call_6_5_14_2(SH2_struct* sh)
{
  SH2movll(sh,5,14);
}
#define init_6_5_15_2
void FASTCALL call_6_5_15_2(SH2_struct* sh)
{
  SH2movll(sh,5,15);
}
#define init_6_6_0_2
void FASTCALL call_6_6_0_2(SH2_struct* sh)
{
  SH2movll(sh,6,0);
}
#define init_6_6_1_2
void FASTCALL call_6_6_1_2(SH2_struct* sh)
{
  SH2movll(sh,6,1);
}
#define init_6_6_2_2
void FASTCALL call_6_6_2_2(SH2_struct* sh)
{
  SH2movll(sh,6,2);
}
#define init_6_6_3_2
void FASTCALL call_6_6_3_2(SH2_struct* sh)
{
  SH2movll(sh,6,3);
}
#define init_6_6_4_2
void FASTCALL call_6_6_4_2(SH2_struct* sh)
{
  SH2movll(sh,6,4);
}
#define init_6_6_5_2
void FASTCALL call_6_6_5_2(SH2_struct* sh)
{
  SH2movll(sh,6,5);
}
#define init_6_6_6_2
void FASTCALL call_6_6_6_2(SH2_struct* sh)
{
  SH2movll(sh,6,6);
}
#define init_6_6_7_2
void FASTCALL call_6_6_7_2(SH2_struct* sh)
{
  SH2movll(sh,6,7);
}
#define init_6_6_8_2
void FASTCALL call_6_6_8_2(SH2_struct* sh)
{
  SH2movll(sh,6,8);
}
#define init_6_6_9_2
void FASTCALL call_6_6_9_2(SH2_struct* sh)
{
  SH2movll(sh,6,9);
}
#define init_6_6_10_2
void FASTCALL call_6_6_10_2(SH2_struct* sh)
{
  SH2movll(sh,6,10);
}
#define init_6_6_11_2
void FASTCALL call_6_6_11_2(SH2_struct* sh)
{
  SH2movll(sh,6,11);
}
#define init_6_6_12_2
void FASTCALL call_6_6_12_2(SH2_struct* sh)
{
  SH2movll(sh,6,12);
}
#define init_6_6_13_2
void FASTCALL call_6_6_13_2(SH2_struct* sh)
{
  SH2movll(sh,6,13);
}
#define init_6_6_14_2
void FASTCALL call_6_6_14_2(SH2_struct* sh)
{
  SH2movll(sh,6,14);
}
#define init_6_6_15_2
void FASTCALL call_6_6_15_2(SH2_struct* sh)
{
  SH2movll(sh,6,15);
}
#define init_6_7_0_2
void FASTCALL call_6_7_0_2(SH2_struct* sh)
{
  SH2movll(sh,7,0);
}
#define init_6_7_1_2
void FASTCALL call_6_7_1_2(SH2_struct* sh)
{
  SH2movll(sh,7,1);
}
#define init_6_7_2_2
void FASTCALL call_6_7_2_2(SH2_struct* sh)
{
  SH2movll(sh,7,2);
}
#define init_6_7_3_2
void FASTCALL call_6_7_3_2(SH2_struct* sh)
{
  SH2movll(sh,7,3);
}
#define init_6_7_4_2
void FASTCALL call_6_7_4_2(SH2_struct* sh)
{
  SH2movll(sh,7,4);
}
#define init_6_7_5_2
void FASTCALL call_6_7_5_2(SH2_struct* sh)
{
  SH2movll(sh,7,5);
}
#define init_6_7_6_2
void FASTCALL call_6_7_6_2(SH2_struct* sh)
{
  SH2movll(sh,7,6);
}
#define init_6_7_7_2
void FASTCALL call_6_7_7_2(SH2_struct* sh)
{
  SH2movll(sh,7,7);
}
#define init_6_7_8_2
void FASTCALL call_6_7_8_2(SH2_struct* sh)
{
  SH2movll(sh,7,8);
}
#define init_6_7_9_2
void FASTCALL call_6_7_9_2(SH2_struct* sh)
{
  SH2movll(sh,7,9);
}
#define init_6_7_10_2
void FASTCALL call_6_7_10_2(SH2_struct* sh)
{
  SH2movll(sh,7,10);
}
#define init_6_7_11_2
void FASTCALL call_6_7_11_2(SH2_struct* sh)
{
  SH2movll(sh,7,11);
}
#define init_6_7_12_2
void FASTCALL call_6_7_12_2(SH2_struct* sh)
{
  SH2movll(sh,7,12);
}
#define init_6_7_13_2
void FASTCALL call_6_7_13_2(SH2_struct* sh)
{
  SH2movll(sh,7,13);
}
#define init_6_7_14_2
void FASTCALL call_6_7_14_2(SH2_struct* sh)
{
  SH2movll(sh,7,14);
}
#define init_6_7_15_2
void FASTCALL call_6_7_15_2(SH2_struct* sh)
{
  SH2movll(sh,7,15);
}
#define init_6_8_0_2
void FASTCALL call_6_8_0_2(SH2_struct* sh)
{
  SH2movll(sh,8,0);
}
#define init_6_8_1_2
void FASTCALL call_6_8_1_2(SH2_struct* sh)
{
  SH2movll(sh,8,1);
}
#define init_6_8_2_2
void FASTCALL call_6_8_2_2(SH2_struct* sh)
{
  SH2movll(sh,8,2);
}
#define init_6_8_3_2
void FASTCALL call_6_8_3_2(SH2_struct* sh)
{
  SH2movll(sh,8,3);
}
#define init_6_8_4_2
void FASTCALL call_6_8_4_2(SH2_struct* sh)
{
  SH2movll(sh,8,4);
}
#define init_6_8_5_2
void FASTCALL call_6_8_5_2(SH2_struct* sh)
{
  SH2movll(sh,8,5);
}
#define init_6_8_6_2
void FASTCALL call_6_8_6_2(SH2_struct* sh)
{
  SH2movll(sh,8,6);
}
#define init_6_8_7_2
void FASTCALL call_6_8_7_2(SH2_struct* sh)
{
  SH2movll(sh,8,7);
}
#define init_6_8_8_2
void FASTCALL call_6_8_8_2(SH2_struct* sh)
{
  SH2movll(sh,8,8);
}
#define init_6_8_9_2
void FASTCALL call_6_8_9_2(SH2_struct* sh)
{
  SH2movll(sh,8,9);
}
#define init_6_8_10_2
void FASTCALL call_6_8_10_2(SH2_struct* sh)
{
  SH2movll(sh,8,10);
}
#define init_6_8_11_2
void FASTCALL call_6_8_11_2(SH2_struct* sh)
{
  SH2movll(sh,8,11);
}
#define init_6_8_12_2
void FASTCALL call_6_8_12_2(SH2_struct* sh)
{
  SH2movll(sh,8,12);
}
#define init_6_8_13_2
void FASTCALL call_6_8_13_2(SH2_struct* sh)
{
  SH2movll(sh,8,13);
}
#define init_6_8_14_2
void FASTCALL call_6_8_14_2(SH2_struct* sh)
{
  SH2movll(sh,8,14);
}
#define init_6_8_15_2
void FASTCALL call_6_8_15_2(SH2_struct* sh)
{
  SH2movll(sh,8,15);
}
#define init_6_9_0_2
void FASTCALL call_6_9_0_2(SH2_struct* sh)
{
  SH2movll(sh,9,0);
}
#define init_6_9_1_2
void FASTCALL call_6_9_1_2(SH2_struct* sh)
{
  SH2movll(sh,9,1);
}
#define init_6_9_2_2
void FASTCALL call_6_9_2_2(SH2_struct* sh)
{
  SH2movll(sh,9,2);
}
#define init_6_9_3_2
void FASTCALL call_6_9_3_2(SH2_struct* sh)
{
  SH2movll(sh,9,3);
}
#define init_6_9_4_2
void FASTCALL call_6_9_4_2(SH2_struct* sh)
{
  SH2movll(sh,9,4);
}
#define init_6_9_5_2
void FASTCALL call_6_9_5_2(SH2_struct* sh)
{
  SH2movll(sh,9,5);
}
#define init_6_9_6_2
void FASTCALL call_6_9_6_2(SH2_struct* sh)
{
  SH2movll(sh,9,6);
}
#define init_6_9_7_2
void FASTCALL call_6_9_7_2(SH2_struct* sh)
{
  SH2movll(sh,9,7);
}
#define init_6_9_8_2
void FASTCALL call_6_9_8_2(SH2_struct* sh)
{
  SH2movll(sh,9,8);
}
#define init_6_9_9_2
void FASTCALL call_6_9_9_2(SH2_struct* sh)
{
  SH2movll(sh,9,9);
}
#define init_6_9_10_2
void FASTCALL call_6_9_10_2(SH2_struct* sh)
{
  SH2movll(sh,9,10);
}
#define init_6_9_11_2
void FASTCALL call_6_9_11_2(SH2_struct* sh)
{
  SH2movll(sh,9,11);
}
#define init_6_9_12_2
void FASTCALL call_6_9_12_2(SH2_struct* sh)
{
  SH2movll(sh,9,12);
}
#define init_6_9_13_2
void FASTCALL call_6_9_13_2(SH2_struct* sh)
{
  SH2movll(sh,9,13);
}
#define init_6_9_14_2
void FASTCALL call_6_9_14_2(SH2_struct* sh)
{
  SH2movll(sh,9,14);
}
#define init_6_9_15_2
void FASTCALL call_6_9_15_2(SH2_struct* sh)
{
  SH2movll(sh,9,15);
}
#define init_6_10_0_2
void FASTCALL call_6_10_0_2(SH2_struct* sh)
{
  SH2movll(sh,10,0);
}
#define init_6_10_1_2
void FASTCALL call_6_10_1_2(SH2_struct* sh)
{
  SH2movll(sh,10,1);
}
#define init_6_10_2_2
void FASTCALL call_6_10_2_2(SH2_struct* sh)
{
  SH2movll(sh,10,2);
}
#define init_6_10_3_2
void FASTCALL call_6_10_3_2(SH2_struct* sh)
{
  SH2movll(sh,10,3);
}
#define init_6_10_4_2
void FASTCALL call_6_10_4_2(SH2_struct* sh)
{
  SH2movll(sh,10,4);
}
#define init_6_10_5_2
void FASTCALL call_6_10_5_2(SH2_struct* sh)
{
  SH2movll(sh,10,5);
}
#define init_6_10_6_2
void FASTCALL call_6_10_6_2(SH2_struct* sh)
{
  SH2movll(sh,10,6);
}
#define init_6_10_7_2
void FASTCALL call_6_10_7_2(SH2_struct* sh)
{
  SH2movll(sh,10,7);
}
#define init_6_10_8_2
void FASTCALL call_6_10_8_2(SH2_struct* sh)
{
  SH2movll(sh,10,8);
}
#define init_6_10_9_2
void FASTCALL call_6_10_9_2(SH2_struct* sh)
{
  SH2movll(sh,10,9);
}
#define init_6_10_10_2
void FASTCALL call_6_10_10_2(SH2_struct* sh)
{
  SH2movll(sh,10,10);
}
#define init_6_10_11_2
void FASTCALL call_6_10_11_2(SH2_struct* sh)
{
  SH2movll(sh,10,11);
}
#define init_6_10_12_2
void FASTCALL call_6_10_12_2(SH2_struct* sh)
{
  SH2movll(sh,10,12);
}
#define init_6_10_13_2
void FASTCALL call_6_10_13_2(SH2_struct* sh)
{
  SH2movll(sh,10,13);
}
#define init_6_10_14_2
void FASTCALL call_6_10_14_2(SH2_struct* sh)
{
  SH2movll(sh,10,14);
}
#define init_6_10_15_2
void FASTCALL call_6_10_15_2(SH2_struct* sh)
{
  SH2movll(sh,10,15);
}
#define init_6_11_0_2
void FASTCALL call_6_11_0_2(SH2_struct* sh)
{
  SH2movll(sh,11,0);
}
#define init_6_11_1_2
void FASTCALL call_6_11_1_2(SH2_struct* sh)
{
  SH2movll(sh,11,1);
}
#define init_6_11_2_2
void FASTCALL call_6_11_2_2(SH2_struct* sh)
{
  SH2movll(sh,11,2);
}
#define init_6_11_3_2
void FASTCALL call_6_11_3_2(SH2_struct* sh)
{
  SH2movll(sh,11,3);
}
#define init_6_11_4_2
void FASTCALL call_6_11_4_2(SH2_struct* sh)
{
  SH2movll(sh,11,4);
}
#define init_6_11_5_2
void FASTCALL call_6_11_5_2(SH2_struct* sh)
{
  SH2movll(sh,11,5);
}
#define init_6_11_6_2
void FASTCALL call_6_11_6_2(SH2_struct* sh)
{
  SH2movll(sh,11,6);
}
#define init_6_11_7_2
void FASTCALL call_6_11_7_2(SH2_struct* sh)
{
  SH2movll(sh,11,7);
}
#define init_6_11_8_2
void FASTCALL call_6_11_8_2(SH2_struct* sh)
{
  SH2movll(sh,11,8);
}
#define init_6_11_9_2
void FASTCALL call_6_11_9_2(SH2_struct* sh)
{
  SH2movll(sh,11,9);
}
#define init_6_11_10_2
void FASTCALL call_6_11_10_2(SH2_struct* sh)
{
  SH2movll(sh,11,10);
}
#define init_6_11_11_2
void FASTCALL call_6_11_11_2(SH2_struct* sh)
{
  SH2movll(sh,11,11);
}
#define init_6_11_12_2
void FASTCALL call_6_11_12_2(SH2_struct* sh)
{
  SH2movll(sh,11,12);
}
#define init_6_11_13_2
void FASTCALL call_6_11_13_2(SH2_struct* sh)
{
  SH2movll(sh,11,13);
}
#define init_6_11_14_2
void FASTCALL call_6_11_14_2(SH2_struct* sh)
{
  SH2movll(sh,11,14);
}
#define init_6_11_15_2
void FASTCALL call_6_11_15_2(SH2_struct* sh)
{
  SH2movll(sh,11,15);
}
#define init_6_12_0_2
void FASTCALL call_6_12_0_2(SH2_struct* sh)
{
  SH2movll(sh,12,0);
}
#define init_6_12_1_2
void FASTCALL call_6_12_1_2(SH2_struct* sh)
{
  SH2movll(sh,12,1);
}
#define init_6_12_2_2
void FASTCALL call_6_12_2_2(SH2_struct* sh)
{
  SH2movll(sh,12,2);
}
#define init_6_12_3_2
void FASTCALL call_6_12_3_2(SH2_struct* sh)
{
  SH2movll(sh,12,3);
}
#define init_6_12_4_2
void FASTCALL call_6_12_4_2(SH2_struct* sh)
{
  SH2movll(sh,12,4);
}
#define init_6_12_5_2
void FASTCALL call_6_12_5_2(SH2_struct* sh)
{
  SH2movll(sh,12,5);
}
#define init_6_12_6_2
void FASTCALL call_6_12_6_2(SH2_struct* sh)
{
  SH2movll(sh,12,6);
}
#define init_6_12_7_2
void FASTCALL call_6_12_7_2(SH2_struct* sh)
{
  SH2movll(sh,12,7);
}
#define init_6_12_8_2
void FASTCALL call_6_12_8_2(SH2_struct* sh)
{
  SH2movll(sh,12,8);
}
#define init_6_12_9_2
void FASTCALL call_6_12_9_2(SH2_struct* sh)
{
  SH2movll(sh,12,9);
}
#define init_6_12_10_2
void FASTCALL call_6_12_10_2(SH2_struct* sh)
{
  SH2movll(sh,12,10);
}
#define init_6_12_11_2
void FASTCALL call_6_12_11_2(SH2_struct* sh)
{
  SH2movll(sh,12,11);
}
#define init_6_12_12_2
void FASTCALL call_6_12_12_2(SH2_struct* sh)
{
  SH2movll(sh,12,12);
}
#define init_6_12_13_2
void FASTCALL call_6_12_13_2(SH2_struct* sh)
{
  SH2movll(sh,12,13);
}
#define init_6_12_14_2
void FASTCALL call_6_12_14_2(SH2_struct* sh)
{
  SH2movll(sh,12,14);
}
#define init_6_12_15_2
void FASTCALL call_6_12_15_2(SH2_struct* sh)
{
  SH2movll(sh,12,15);
}
#define init_6_13_0_2
void FASTCALL call_6_13_0_2(SH2_struct* sh)
{
  SH2movll(sh,13,0);
}
#define init_6_13_1_2
void FASTCALL call_6_13_1_2(SH2_struct* sh)
{
  SH2movll(sh,13,1);
}
#define init_6_13_2_2
void FASTCALL call_6_13_2_2(SH2_struct* sh)
{
  SH2movll(sh,13,2);
}
#define init_6_13_3_2
void FASTCALL call_6_13_3_2(SH2_struct* sh)
{
  SH2movll(sh,13,3);
}
#define init_6_13_4_2
void FASTCALL call_6_13_4_2(SH2_struct* sh)
{
  SH2movll(sh,13,4);
}
#define init_6_13_5_2
void FASTCALL call_6_13_5_2(SH2_struct* sh)
{
  SH2movll(sh,13,5);
}
#define init_6_13_6_2
void FASTCALL call_6_13_6_2(SH2_struct* sh)
{
  SH2movll(sh,13,6);
}
#define init_6_13_7_2
void FASTCALL call_6_13_7_2(SH2_struct* sh)
{
  SH2movll(sh,13,7);
}
#define init_6_13_8_2
void FASTCALL call_6_13_8_2(SH2_struct* sh)
{
  SH2movll(sh,13,8);
}
#define init_6_13_9_2
void FASTCALL call_6_13_9_2(SH2_struct* sh)
{
  SH2movll(sh,13,9);
}
#define init_6_13_10_2
void FASTCALL call_6_13_10_2(SH2_struct* sh)
{
  SH2movll(sh,13,10);
}
#define init_6_13_11_2
void FASTCALL call_6_13_11_2(SH2_struct* sh)
{
  SH2movll(sh,13,11);
}
#define init_6_13_12_2
void FASTCALL call_6_13_12_2(SH2_struct* sh)
{
  SH2movll(sh,13,12);
}
#define init_6_13_13_2
void FASTCALL call_6_13_13_2(SH2_struct* sh)
{
  SH2movll(sh,13,13);
}
#define init_6_13_14_2
void FASTCALL call_6_13_14_2(SH2_struct* sh)
{
  SH2movll(sh,13,14);
}
#define init_6_13_15_2
void FASTCALL call_6_13_15_2(SH2_struct* sh)
{
  SH2movll(sh,13,15);
}
#define init_6_14_0_2
void FASTCALL call_6_14_0_2(SH2_struct* sh)
{
  SH2movll(sh,14,0);
}
#define init_6_14_1_2
void FASTCALL call_6_14_1_2(SH2_struct* sh)
{
  SH2movll(sh,14,1);
}
#define init_6_14_2_2
void FASTCALL call_6_14_2_2(SH2_struct* sh)
{
  SH2movll(sh,14,2);
}
#define init_6_14_3_2
void FASTCALL call_6_14_3_2(SH2_struct* sh)
{
  SH2movll(sh,14,3);
}
#define init_6_14_4_2
void FASTCALL call_6_14_4_2(SH2_struct* sh)
{
  SH2movll(sh,14,4);
}
#define init_6_14_5_2
void FASTCALL call_6_14_5_2(SH2_struct* sh)
{
  SH2movll(sh,14,5);
}
#define init_6_14_6_2
void FASTCALL call_6_14_6_2(SH2_struct* sh)
{
  SH2movll(sh,14,6);
}
#define init_6_14_7_2
void FASTCALL call_6_14_7_2(SH2_struct* sh)
{
  SH2movll(sh,14,7);
}
#define init_6_14_8_2
void FASTCALL call_6_14_8_2(SH2_struct* sh)
{
  SH2movll(sh,14,8);
}
#define init_6_14_9_2
void FASTCALL call_6_14_9_2(SH2_struct* sh)
{
  SH2movll(sh,14,9);
}
#define init_6_14_10_2
void FASTCALL call_6_14_10_2(SH2_struct* sh)
{
  SH2movll(sh,14,10);
}
#define init_6_14_11_2
void FASTCALL call_6_14_11_2(SH2_struct* sh)
{
  SH2movll(sh,14,11);
}
#define init_6_14_12_2
void FASTCALL call_6_14_12_2(SH2_struct* sh)
{
  SH2movll(sh,14,12);
}
#define init_6_14_13_2
void FASTCALL call_6_14_13_2(SH2_struct* sh)
{
  SH2movll(sh,14,13);
}
#define init_6_14_14_2
void FASTCALL call_6_14_14_2(SH2_struct* sh)
{
  SH2movll(sh,14,14);
}
#define init_6_14_15_2
void FASTCALL call_6_14_15_2(SH2_struct* sh)
{
  SH2movll(sh,14,15);
}
#define init_6_15_0_2
void FASTCALL call_6_15_0_2(SH2_struct* sh)
{
  SH2movll(sh,15,0);
}
#define init_6_15_1_2
void FASTCALL call_6_15_1_2(SH2_struct* sh)
{
  SH2movll(sh,15,1);
}
#define init_6_15_2_2
void FASTCALL call_6_15_2_2(SH2_struct* sh)
{
  SH2movll(sh,15,2);
}
#define init_6_15_3_2
void FASTCALL call_6_15_3_2(SH2_struct* sh)
{
  SH2movll(sh,15,3);
}
#define init_6_15_4_2
void FASTCALL call_6_15_4_2(SH2_struct* sh)
{
  SH2movll(sh,15,4);
}
#define init_6_15_5_2
void FASTCALL call_6_15_5_2(SH2_struct* sh)
{
  SH2movll(sh,15,5);
}
#define init_6_15_6_2
void FASTCALL call_6_15_6_2(SH2_struct* sh)
{
  SH2movll(sh,15,6);
}
#define init_6_15_7_2
void FASTCALL call_6_15_7_2(SH2_struct* sh)
{
  SH2movll(sh,15,7);
}
#define init_6_15_8_2
void FASTCALL call_6_15_8_2(SH2_struct* sh)
{
  SH2movll(sh,15,8);
}
#define init_6_15_9_2
void FASTCALL call_6_15_9_2(SH2_struct* sh)
{
  SH2movll(sh,15,9);
}
#define init_6_15_10_2
void FASTCALL call_6_15_10_2(SH2_struct* sh)
{
  SH2movll(sh,15,10);
}
#define init_6_15_11_2
void FASTCALL call_6_15_11_2(SH2_struct* sh)
{
  SH2movll(sh,15,11);
}
#define init_6_15_12_2
void FASTCALL call_6_15_12_2(SH2_struct* sh)
{
  SH2movll(sh,15,12);
}
#define init_6_15_13_2
void FASTCALL call_6_15_13_2(SH2_struct* sh)
{
  SH2movll(sh,15,13);
}
#define init_6_15_14_2
void FASTCALL call_6_15_14_2(SH2_struct* sh)
{
  SH2movll(sh,15,14);
}
#define init_6_15_15_2
void FASTCALL call_6_15_15_2(SH2_struct* sh)
{
  SH2movll(sh,15,15);
}
#define init_13_0_0_0
void FASTCALL call_13_0_0_0(SH2_struct* sh)
{
  SH2movli(sh, 0, (0<<4)+0);
}
#define init_13_0_0_1
void FASTCALL call_13_0_0_1(SH2_struct* sh)
{
  SH2movli(sh, 0, (0<<4)+1);
}
#define init_13_0_0_2
void FASTCALL call_13_0_0_2(SH2_struct* sh)
{
  SH2movli(sh, 0, (0<<4)+2);
}
#define init_13_0_0_3
void FASTCALL call_13_0_0_3(SH2_struct* sh)
{
  SH2movli(sh, 0, (0<<4)+3);
}
#define init_13_0_0_4
void FASTCALL call_13_0_0_4(SH2_struct* sh)
{
  SH2movli(sh, 0, (0<<4)+4);
}
#define init_13_0_0_5
void FASTCALL call_13_0_0_5(SH2_struct* sh)
{
  SH2movli(sh, 0, (0<<4)+5);
}
#define init_13_0_0_6
void FASTCALL call_13_0_0_6(SH2_struct* sh)
{
  SH2movli(sh, 0, (0<<4)+6);
}
#define init_13_0_0_7
void FASTCALL call_13_0_0_7(SH2_struct* sh)
{
  SH2movli(sh, 0, (0<<4)+7);
}
#define init_13_0_0_8
void FASTCALL call_13_0_0_8(SH2_struct* sh)
{
  SH2movli(sh, 0, (0<<4)+8);
}
#define init_13_0_0_9
void FASTCALL call_13_0_0_9(SH2_struct* sh)
{
  SH2movli(sh, 0, (0<<4)+9);
}
#define init_13_0_0_10
void FASTCALL call_13_0_0_10(SH2_struct* sh)
{
  SH2movli(sh, 0, (0<<4)+10);
}
#define init_13_0_0_11
void FASTCALL call_13_0_0_11(SH2_struct* sh)
{
  SH2movli(sh, 0, (0<<4)+11);
}
#define init_13_0_0_12
void FASTCALL call_13_0_0_12(SH2_struct* sh)
{
  SH2movli(sh, 0, (0<<4)+12);
}
#define init_13_0_0_13
void FASTCALL call_13_0_0_13(SH2_struct* sh)
{
  SH2movli(sh, 0, (0<<4)+13);
}
#define init_13_0_0_14
void FASTCALL call_13_0_0_14(SH2_struct* sh)
{
  SH2movli(sh, 0, (0<<4)+14);
}
#define init_13_0_0_15
void FASTCALL call_13_0_0_15(SH2_struct* sh)
{
  SH2movli(sh, 0, (0<<4)+15);
}
#define init_13_0_1_0
void FASTCALL call_13_0_1_0(SH2_struct* sh)
{
  SH2movli(sh, 0, (1<<4)+0);
}
#define init_13_0_1_1
void FASTCALL call_13_0_1_1(SH2_struct* sh)
{
  SH2movli(sh, 0, (1<<4)+1);
}
#define init_13_0_1_2
void FASTCALL call_13_0_1_2(SH2_struct* sh)
{
  SH2movli(sh, 0, (1<<4)+2);
}
#define init_13_0_1_3
void FASTCALL call_13_0_1_3(SH2_struct* sh)
{
  SH2movli(sh, 0, (1<<4)+3);
}
#define init_13_0_1_4
void FASTCALL call_13_0_1_4(SH2_struct* sh)
{
  SH2movli(sh, 0, (1<<4)+4);
}
#define init_13_0_1_5
void FASTCALL call_13_0_1_5(SH2_struct* sh)
{
  SH2movli(sh, 0, (1<<4)+5);
}
#define init_13_0_1_6
void FASTCALL call_13_0_1_6(SH2_struct* sh)
{
  SH2movli(sh, 0, (1<<4)+6);
}
#define init_13_0_1_7
void FASTCALL call_13_0_1_7(SH2_struct* sh)
{
  SH2movli(sh, 0, (1<<4)+7);
}
#define init_13_0_1_8
void FASTCALL call_13_0_1_8(SH2_struct* sh)
{
  SH2movli(sh, 0, (1<<4)+8);
}
#define init_13_0_1_9
void FASTCALL call_13_0_1_9(SH2_struct* sh)
{
  SH2movli(sh, 0, (1<<4)+9);
}
#define init_13_0_1_10
void FASTCALL call_13_0_1_10(SH2_struct* sh)
{
  SH2movli(sh, 0, (1<<4)+10);
}
#define init_13_0_1_11
void FASTCALL call_13_0_1_11(SH2_struct* sh)
{
  SH2movli(sh, 0, (1<<4)+11);
}
#define init_13_0_1_12
void FASTCALL call_13_0_1_12(SH2_struct* sh)
{
  SH2movli(sh, 0, (1<<4)+12);
}
#define init_13_0_1_13
void FASTCALL call_13_0_1_13(SH2_struct* sh)
{
  SH2movli(sh, 0, (1<<4)+13);
}
#define init_13_0_1_14
void FASTCALL call_13_0_1_14(SH2_struct* sh)
{
  SH2movli(sh, 0, (1<<4)+14);
}
#define init_13_0_1_15
void FASTCALL call_13_0_1_15(SH2_struct* sh)
{
  SH2movli(sh, 0, (1<<4)+15);
}
#define init_13_0_2_0
void FASTCALL call_13_0_2_0(SH2_struct* sh)
{
  SH2movli(sh, 0, (2<<4)+0);
}
#define init_13_0_2_1
void FASTCALL call_13_0_2_1(SH2_struct* sh)
{
  SH2movli(sh, 0, (2<<4)+1);
}
#define init_13_0_2_2
void FASTCALL call_13_0_2_2(SH2_struct* sh)
{
  SH2movli(sh, 0, (2<<4)+2);
}
#define init_13_0_2_3
void FASTCALL call_13_0_2_3(SH2_struct* sh)
{
  SH2movli(sh, 0, (2<<4)+3);
}
#define init_13_0_2_4
void FASTCALL call_13_0_2_4(SH2_struct* sh)
{
  SH2movli(sh, 0, (2<<4)+4);
}
#define init_13_0_2_5
void FASTCALL call_13_0_2_5(SH2_struct* sh)
{
  SH2movli(sh, 0, (2<<4)+5);
}
#define init_13_0_2_6
void FASTCALL call_13_0_2_6(SH2_struct* sh)
{
  SH2movli(sh, 0, (2<<4)+6);
}
#define init_13_0_2_7
void FASTCALL call_13_0_2_7(SH2_struct* sh)
{
  SH2movli(sh, 0, (2<<4)+7);
}
#define init_13_0_2_8
void FASTCALL call_13_0_2_8(SH2_struct* sh)
{
  SH2movli(sh, 0, (2<<4)+8);
}
#define init_13_0_2_9
void FASTCALL call_13_0_2_9(SH2_struct* sh)
{
  SH2movli(sh, 0, (2<<4)+9);
}
#define init_13_0_2_10
void FASTCALL call_13_0_2_10(SH2_struct* sh)
{
  SH2movli(sh, 0, (2<<4)+10);
}
#define init_13_0_2_11
void FASTCALL call_13_0_2_11(SH2_struct* sh)
{
  SH2movli(sh, 0, (2<<4)+11);
}
#define init_13_0_2_12
void FASTCALL call_13_0_2_12(SH2_struct* sh)
{
  SH2movli(sh, 0, (2<<4)+12);
}
#define init_13_0_2_13
void FASTCALL call_13_0_2_13(SH2_struct* sh)
{
  SH2movli(sh, 0, (2<<4)+13);
}
#define init_13_0_2_14
void FASTCALL call_13_0_2_14(SH2_struct* sh)
{
  SH2movli(sh, 0, (2<<4)+14);
}
#define init_13_0_2_15
void FASTCALL call_13_0_2_15(SH2_struct* sh)
{
  SH2movli(sh, 0, (2<<4)+15);
}
#define init_13_0_3_0
void FASTCALL call_13_0_3_0(SH2_struct* sh)
{
  SH2movli(sh, 0, (3<<4)+0);
}
#define init_13_0_3_1
void FASTCALL call_13_0_3_1(SH2_struct* sh)
{
  SH2movli(sh, 0, (3<<4)+1);
}
#define init_13_0_3_2
void FASTCALL call_13_0_3_2(SH2_struct* sh)
{
  SH2movli(sh, 0, (3<<4)+2);
}
#define init_13_0_3_3
void FASTCALL call_13_0_3_3(SH2_struct* sh)
{
  SH2movli(sh, 0, (3<<4)+3);
}
#define init_13_0_3_4
void FASTCALL call_13_0_3_4(SH2_struct* sh)
{
  SH2movli(sh, 0, (3<<4)+4);
}
#define init_13_0_3_5
void FASTCALL call_13_0_3_5(SH2_struct* sh)
{
  SH2movli(sh, 0, (3<<4)+5);
}
#define init_13_0_3_6
void FASTCALL call_13_0_3_6(SH2_struct* sh)
{
  SH2movli(sh, 0, (3<<4)+6);
}
#define init_13_0_3_7
void FASTCALL call_13_0_3_7(SH2_struct* sh)
{
  SH2movli(sh, 0, (3<<4)+7);
}
#define init_13_0_3_8
void FASTCALL call_13_0_3_8(SH2_struct* sh)
{
  SH2movli(sh, 0, (3<<4)+8);
}
#define init_13_0_3_9
void FASTCALL call_13_0_3_9(SH2_struct* sh)
{
  SH2movli(sh, 0, (3<<4)+9);
}
#define init_13_0_3_10
void FASTCALL call_13_0_3_10(SH2_struct* sh)
{
  SH2movli(sh, 0, (3<<4)+10);
}
#define init_13_0_3_11
void FASTCALL call_13_0_3_11(SH2_struct* sh)
{
  SH2movli(sh, 0, (3<<4)+11);
}
#define init_13_0_3_12
void FASTCALL call_13_0_3_12(SH2_struct* sh)
{
  SH2movli(sh, 0, (3<<4)+12);
}
#define init_13_0_3_13
void FASTCALL call_13_0_3_13(SH2_struct* sh)
{
  SH2movli(sh, 0, (3<<4)+13);
}
#define init_13_0_3_14
void FASTCALL call_13_0_3_14(SH2_struct* sh)
{
  SH2movli(sh, 0, (3<<4)+14);
}
#define init_13_0_3_15
void FASTCALL call_13_0_3_15(SH2_struct* sh)
{
  SH2movli(sh, 0, (3<<4)+15);
}
#define init_13_0_4_0
void FASTCALL call_13_0_4_0(SH2_struct* sh)
{
  SH2movli(sh, 0, (4<<4)+0);
}
#define init_13_0_4_1
void FASTCALL call_13_0_4_1(SH2_struct* sh)
{
  SH2movli(sh, 0, (4<<4)+1);
}
#define init_13_0_4_2
void FASTCALL call_13_0_4_2(SH2_struct* sh)
{
  SH2movli(sh, 0, (4<<4)+2);
}
#define init_13_0_4_3
void FASTCALL call_13_0_4_3(SH2_struct* sh)
{
  SH2movli(sh, 0, (4<<4)+3);
}
#define init_13_0_4_4
void FASTCALL call_13_0_4_4(SH2_struct* sh)
{
  SH2movli(sh, 0, (4<<4)+4);
}
#define init_13_0_4_5
void FASTCALL call_13_0_4_5(SH2_struct* sh)
{
  SH2movli(sh, 0, (4<<4)+5);
}
#define init_13_0_4_6
void FASTCALL call_13_0_4_6(SH2_struct* sh)
{
  SH2movli(sh, 0, (4<<4)+6);
}
#define init_13_0_4_7
void FASTCALL call_13_0_4_7(SH2_struct* sh)
{
  SH2movli(sh, 0, (4<<4)+7);
}
#define init_13_0_4_8
void FASTCALL call_13_0_4_8(SH2_struct* sh)
{
  SH2movli(sh, 0, (4<<4)+8);
}
#define init_13_0_4_9
void FASTCALL call_13_0_4_9(SH2_struct* sh)
{
  SH2movli(sh, 0, (4<<4)+9);
}
#define init_13_0_4_10
void FASTCALL call_13_0_4_10(SH2_struct* sh)
{
  SH2movli(sh, 0, (4<<4)+10);
}
#define init_13_0_4_11
void FASTCALL call_13_0_4_11(SH2_struct* sh)
{
  SH2movli(sh, 0, (4<<4)+11);
}
#define init_13_0_4_12
void FASTCALL call_13_0_4_12(SH2_struct* sh)
{
  SH2movli(sh, 0, (4<<4)+12);
}
#define init_13_0_4_13
void FASTCALL call_13_0_4_13(SH2_struct* sh)
{
  SH2movli(sh, 0, (4<<4)+13);
}
#define init_13_0_4_14
void FASTCALL call_13_0_4_14(SH2_struct* sh)
{
  SH2movli(sh, 0, (4<<4)+14);
}
#define init_13_0_4_15
void FASTCALL call_13_0_4_15(SH2_struct* sh)
{
  SH2movli(sh, 0, (4<<4)+15);
}
#define init_13_0_5_0
void FASTCALL call_13_0_5_0(SH2_struct* sh)
{
  SH2movli(sh, 0, (5<<4)+0);
}
#define init_13_0_5_1
void FASTCALL call_13_0_5_1(SH2_struct* sh)
{
  SH2movli(sh, 0, (5<<4)+1);
}
#define init_13_0_5_2
void FASTCALL call_13_0_5_2(SH2_struct* sh)
{
  SH2movli(sh, 0, (5<<4)+2);
}
#define init_13_0_5_3
void FASTCALL call_13_0_5_3(SH2_struct* sh)
{
  SH2movli(sh, 0, (5<<4)+3);
}
#define init_13_0_5_4
void FASTCALL call_13_0_5_4(SH2_struct* sh)
{
  SH2movli(sh, 0, (5<<4)+4);
}
#define init_13_0_5_5
void FASTCALL call_13_0_5_5(SH2_struct* sh)
{
  SH2movli(sh, 0, (5<<4)+5);
}
#define init_13_0_5_6
void FASTCALL call_13_0_5_6(SH2_struct* sh)
{
  SH2movli(sh, 0, (5<<4)+6);
}
#define init_13_0_5_7
void FASTCALL call_13_0_5_7(SH2_struct* sh)
{
  SH2movli(sh, 0, (5<<4)+7);
}
#define init_13_0_5_8
void FASTCALL call_13_0_5_8(SH2_struct* sh)
{
  SH2movli(sh, 0, (5<<4)+8);
}
#define init_13_0_5_9
void FASTCALL call_13_0_5_9(SH2_struct* sh)
{
  SH2movli(sh, 0, (5<<4)+9);
}
#define init_13_0_5_10
void FASTCALL call_13_0_5_10(SH2_struct* sh)
{
  SH2movli(sh, 0, (5<<4)+10);
}
#define init_13_0_5_11
void FASTCALL call_13_0_5_11(SH2_struct* sh)
{
  SH2movli(sh, 0, (5<<4)+11);
}
#define init_13_0_5_12
void FASTCALL call_13_0_5_12(SH2_struct* sh)
{
  SH2movli(sh, 0, (5<<4)+12);
}
#define init_13_0_5_13
void FASTCALL call_13_0_5_13(SH2_struct* sh)
{
  SH2movli(sh, 0, (5<<4)+13);
}
#define init_13_0_5_14
void FASTCALL call_13_0_5_14(SH2_struct* sh)
{
  SH2movli(sh, 0, (5<<4)+14);
}
#define init_13_0_5_15
void FASTCALL call_13_0_5_15(SH2_struct* sh)
{
  SH2movli(sh, 0, (5<<4)+15);
}
#define init_13_0_6_0
void FASTCALL call_13_0_6_0(SH2_struct* sh)
{
  SH2movli(sh, 0, (6<<4)+0);
}
#define init_13_0_6_1
void FASTCALL call_13_0_6_1(SH2_struct* sh)
{
  SH2movli(sh, 0, (6<<4)+1);
}
#define init_13_0_6_2
void FASTCALL call_13_0_6_2(SH2_struct* sh)
{
  SH2movli(sh, 0, (6<<4)+2);
}
#define init_13_0_6_3
void FASTCALL call_13_0_6_3(SH2_struct* sh)
{
  SH2movli(sh, 0, (6<<4)+3);
}
#define init_13_0_6_4
void FASTCALL call_13_0_6_4(SH2_struct* sh)
{
  SH2movli(sh, 0, (6<<4)+4);
}
#define init_13_0_6_5
void FASTCALL call_13_0_6_5(SH2_struct* sh)
{
  SH2movli(sh, 0, (6<<4)+5);
}
#define init_13_0_6_6
void FASTCALL call_13_0_6_6(SH2_struct* sh)
{
  SH2movli(sh, 0, (6<<4)+6);
}
#define init_13_0_6_7
void FASTCALL call_13_0_6_7(SH2_struct* sh)
{
  SH2movli(sh, 0, (6<<4)+7);
}
#define init_13_0_6_8
void FASTCALL call_13_0_6_8(SH2_struct* sh)
{
  SH2movli(sh, 0, (6<<4)+8);
}
#define init_13_0_6_9
void FASTCALL call_13_0_6_9(SH2_struct* sh)
{
  SH2movli(sh, 0, (6<<4)+9);
}
#define init_13_0_6_10
void FASTCALL call_13_0_6_10(SH2_struct* sh)
{
  SH2movli(sh, 0, (6<<4)+10);
}
#define init_13_0_6_11
void FASTCALL call_13_0_6_11(SH2_struct* sh)
{
  SH2movli(sh, 0, (6<<4)+11);
}
#define init_13_0_6_12
void FASTCALL call_13_0_6_12(SH2_struct* sh)
{
  SH2movli(sh, 0, (6<<4)+12);
}
#define init_13_0_6_13
void FASTCALL call_13_0_6_13(SH2_struct* sh)
{
  SH2movli(sh, 0, (6<<4)+13);
}
#define init_13_0_6_14
void FASTCALL call_13_0_6_14(SH2_struct* sh)
{
  SH2movli(sh, 0, (6<<4)+14);
}
#define init_13_0_6_15
void FASTCALL call_13_0_6_15(SH2_struct* sh)
{
  SH2movli(sh, 0, (6<<4)+15);
}
#define init_13_0_7_0
void FASTCALL call_13_0_7_0(SH2_struct* sh)
{
  SH2movli(sh, 0, (7<<4)+0);
}
#define init_13_0_7_1
void FASTCALL call_13_0_7_1(SH2_struct* sh)
{
  SH2movli(sh, 0, (7<<4)+1);
}
#define init_13_0_7_2
void FASTCALL call_13_0_7_2(SH2_struct* sh)
{
  SH2movli(sh, 0, (7<<4)+2);
}
#define init_13_0_7_3
void FASTCALL call_13_0_7_3(SH2_struct* sh)
{
  SH2movli(sh, 0, (7<<4)+3);
}
#define init_13_0_7_4
void FASTCALL call_13_0_7_4(SH2_struct* sh)
{
  SH2movli(sh, 0, (7<<4)+4);
}
#define init_13_0_7_5
void FASTCALL call_13_0_7_5(SH2_struct* sh)
{
  SH2movli(sh, 0, (7<<4)+5);
}
#define init_13_0_7_6
void FASTCALL call_13_0_7_6(SH2_struct* sh)
{
  SH2movli(sh, 0, (7<<4)+6);
}
#define init_13_0_7_7
void FASTCALL call_13_0_7_7(SH2_struct* sh)
{
  SH2movli(sh, 0, (7<<4)+7);
}
#define init_13_0_7_8
void FASTCALL call_13_0_7_8(SH2_struct* sh)
{
  SH2movli(sh, 0, (7<<4)+8);
}
#define init_13_0_7_9
void FASTCALL call_13_0_7_9(SH2_struct* sh)
{
  SH2movli(sh, 0, (7<<4)+9);
}
#define init_13_0_7_10
void FASTCALL call_13_0_7_10(SH2_struct* sh)
{
  SH2movli(sh, 0, (7<<4)+10);
}
#define init_13_0_7_11
void FASTCALL call_13_0_7_11(SH2_struct* sh)
{
  SH2movli(sh, 0, (7<<4)+11);
}
#define init_13_0_7_12
void FASTCALL call_13_0_7_12(SH2_struct* sh)
{
  SH2movli(sh, 0, (7<<4)+12);
}
#define init_13_0_7_13
void FASTCALL call_13_0_7_13(SH2_struct* sh)
{
  SH2movli(sh, 0, (7<<4)+13);
}
#define init_13_0_7_14
void FASTCALL call_13_0_7_14(SH2_struct* sh)
{
  SH2movli(sh, 0, (7<<4)+14);
}
#define init_13_0_7_15
void FASTCALL call_13_0_7_15(SH2_struct* sh)
{
  SH2movli(sh, 0, (7<<4)+15);
}
#define init_13_0_8_0
void FASTCALL call_13_0_8_0(SH2_struct* sh)
{
  SH2movli(sh, 0, (8<<4)+0);
}
#define init_13_0_8_1
void FASTCALL call_13_0_8_1(SH2_struct* sh)
{
  SH2movli(sh, 0, (8<<4)+1);
}
#define init_13_0_8_2
void FASTCALL call_13_0_8_2(SH2_struct* sh)
{
  SH2movli(sh, 0, (8<<4)+2);
}
#define init_13_0_8_3
void FASTCALL call_13_0_8_3(SH2_struct* sh)
{
  SH2movli(sh, 0, (8<<4)+3);
}
#define init_13_0_8_4
void FASTCALL call_13_0_8_4(SH2_struct* sh)
{
  SH2movli(sh, 0, (8<<4)+4);
}
#define init_13_0_8_5
void FASTCALL call_13_0_8_5(SH2_struct* sh)
{
  SH2movli(sh, 0, (8<<4)+5);
}
#define init_13_0_8_6
void FASTCALL call_13_0_8_6(SH2_struct* sh)
{
  SH2movli(sh, 0, (8<<4)+6);
}
#define init_13_0_8_7
void FASTCALL call_13_0_8_7(SH2_struct* sh)
{
  SH2movli(sh, 0, (8<<4)+7);
}
#define init_13_0_8_8
void FASTCALL call_13_0_8_8(SH2_struct* sh)
{
  SH2movli(sh, 0, (8<<4)+8);
}
#define init_13_0_8_9
void FASTCALL call_13_0_8_9(SH2_struct* sh)
{
  SH2movli(sh, 0, (8<<4)+9);
}
#define init_13_0_8_10
void FASTCALL call_13_0_8_10(SH2_struct* sh)
{
  SH2movli(sh, 0, (8<<4)+10);
}
#define init_13_0_8_11
void FASTCALL call_13_0_8_11(SH2_struct* sh)
{
  SH2movli(sh, 0, (8<<4)+11);
}
#define init_13_0_8_12
void FASTCALL call_13_0_8_12(SH2_struct* sh)
{
  SH2movli(sh, 0, (8<<4)+12);
}
#define init_13_0_8_13
void FASTCALL call_13_0_8_13(SH2_struct* sh)
{
  SH2movli(sh, 0, (8<<4)+13);
}
#define init_13_0_8_14
void FASTCALL call_13_0_8_14(SH2_struct* sh)
{
  SH2movli(sh, 0, (8<<4)+14);
}
#define init_13_0_8_15
void FASTCALL call_13_0_8_15(SH2_struct* sh)
{
  SH2movli(sh, 0, (8<<4)+15);
}
#define init_13_0_9_0
void FASTCALL call_13_0_9_0(SH2_struct* sh)
{
  SH2movli(sh, 0, (9<<4)+0);
}
#define init_13_0_9_1
void FASTCALL call_13_0_9_1(SH2_struct* sh)
{
  SH2movli(sh, 0, (9<<4)+1);
}
#define init_13_0_9_2
void FASTCALL call_13_0_9_2(SH2_struct* sh)
{
  SH2movli(sh, 0, (9<<4)+2);
}
#define init_13_0_9_3
void FASTCALL call_13_0_9_3(SH2_struct* sh)
{
  SH2movli(sh, 0, (9<<4)+3);
}
#define init_13_0_9_4
void FASTCALL call_13_0_9_4(SH2_struct* sh)
{
  SH2movli(sh, 0, (9<<4)+4);
}
#define init_13_0_9_5
void FASTCALL call_13_0_9_5(SH2_struct* sh)
{
  SH2movli(sh, 0, (9<<4)+5);
}
#define init_13_0_9_6
void FASTCALL call_13_0_9_6(SH2_struct* sh)
{
  SH2movli(sh, 0, (9<<4)+6);
}
#define init_13_0_9_7
void FASTCALL call_13_0_9_7(SH2_struct* sh)
{
  SH2movli(sh, 0, (9<<4)+7);
}
#define init_13_0_9_8
void FASTCALL call_13_0_9_8(SH2_struct* sh)
{
  SH2movli(sh, 0, (9<<4)+8);
}
#define init_13_0_9_9
void FASTCALL call_13_0_9_9(SH2_struct* sh)
{
  SH2movli(sh, 0, (9<<4)+9);
}
#define init_13_0_9_10
void FASTCALL call_13_0_9_10(SH2_struct* sh)
{
  SH2movli(sh, 0, (9<<4)+10);
}
#define init_13_0_9_11
void FASTCALL call_13_0_9_11(SH2_struct* sh)
{
  SH2movli(sh, 0, (9<<4)+11);
}
#define init_13_0_9_12
void FASTCALL call_13_0_9_12(SH2_struct* sh)
{
  SH2movli(sh, 0, (9<<4)+12);
}
#define init_13_0_9_13
void FASTCALL call_13_0_9_13(SH2_struct* sh)
{
  SH2movli(sh, 0, (9<<4)+13);
}
#define init_13_0_9_14
void FASTCALL call_13_0_9_14(SH2_struct* sh)
{
  SH2movli(sh, 0, (9<<4)+14);
}
#define init_13_0_9_15
void FASTCALL call_13_0_9_15(SH2_struct* sh)
{
  SH2movli(sh, 0, (9<<4)+15);
}
#define init_13_0_10_0
void FASTCALL call_13_0_10_0(SH2_struct* sh)
{
  SH2movli(sh, 0, (10<<4)+0);
}
#define init_13_0_10_1
void FASTCALL call_13_0_10_1(SH2_struct* sh)
{
  SH2movli(sh, 0, (10<<4)+1);
}
#define init_13_0_10_2
void FASTCALL call_13_0_10_2(SH2_struct* sh)
{
  SH2movli(sh, 0, (10<<4)+2);
}
#define init_13_0_10_3
void FASTCALL call_13_0_10_3(SH2_struct* sh)
{
  SH2movli(sh, 0, (10<<4)+3);
}
#define init_13_0_10_4
void FASTCALL call_13_0_10_4(SH2_struct* sh)
{
  SH2movli(sh, 0, (10<<4)+4);
}
#define init_13_0_10_5
void FASTCALL call_13_0_10_5(SH2_struct* sh)
{
  SH2movli(sh, 0, (10<<4)+5);
}
#define init_13_0_10_6
void FASTCALL call_13_0_10_6(SH2_struct* sh)
{
  SH2movli(sh, 0, (10<<4)+6);
}
#define init_13_0_10_7
void FASTCALL call_13_0_10_7(SH2_struct* sh)
{
  SH2movli(sh, 0, (10<<4)+7);
}
#define init_13_0_10_8
void FASTCALL call_13_0_10_8(SH2_struct* sh)
{
  SH2movli(sh, 0, (10<<4)+8);
}
#define init_13_0_10_9
void FASTCALL call_13_0_10_9(SH2_struct* sh)
{
  SH2movli(sh, 0, (10<<4)+9);
}
#define init_13_0_10_10
void FASTCALL call_13_0_10_10(SH2_struct* sh)
{
  SH2movli(sh, 0, (10<<4)+10);
}
#define init_13_0_10_11
void FASTCALL call_13_0_10_11(SH2_struct* sh)
{
  SH2movli(sh, 0, (10<<4)+11);
}
#define init_13_0_10_12
void FASTCALL call_13_0_10_12(SH2_struct* sh)
{
  SH2movli(sh, 0, (10<<4)+12);
}
#define init_13_0_10_13
void FASTCALL call_13_0_10_13(SH2_struct* sh)
{
  SH2movli(sh, 0, (10<<4)+13);
}
#define init_13_0_10_14
void FASTCALL call_13_0_10_14(SH2_struct* sh)
{
  SH2movli(sh, 0, (10<<4)+14);
}
#define init_13_0_10_15
void FASTCALL call_13_0_10_15(SH2_struct* sh)
{
  SH2movli(sh, 0, (10<<4)+15);
}
#define init_13_0_11_0
void FASTCALL call_13_0_11_0(SH2_struct* sh)
{
  SH2movli(sh, 0, (11<<4)+0);
}
#define init_13_0_11_1
void FASTCALL call_13_0_11_1(SH2_struct* sh)
{
  SH2movli(sh, 0, (11<<4)+1);
}
#define init_13_0_11_2
void FASTCALL call_13_0_11_2(SH2_struct* sh)
{
  SH2movli(sh, 0, (11<<4)+2);
}
#define init_13_0_11_3
void FASTCALL call_13_0_11_3(SH2_struct* sh)
{
  SH2movli(sh, 0, (11<<4)+3);
}
#define init_13_0_11_4
void FASTCALL call_13_0_11_4(SH2_struct* sh)
{
  SH2movli(sh, 0, (11<<4)+4);
}
#define init_13_0_11_5
void FASTCALL call_13_0_11_5(SH2_struct* sh)
{
  SH2movli(sh, 0, (11<<4)+5);
}
#define init_13_0_11_6
void FASTCALL call_13_0_11_6(SH2_struct* sh)
{
  SH2movli(sh, 0, (11<<4)+6);
}
#define init_13_0_11_7
void FASTCALL call_13_0_11_7(SH2_struct* sh)
{
  SH2movli(sh, 0, (11<<4)+7);
}
#define init_13_0_11_8
void FASTCALL call_13_0_11_8(SH2_struct* sh)
{
  SH2movli(sh, 0, (11<<4)+8);
}
#define init_13_0_11_9
void FASTCALL call_13_0_11_9(SH2_struct* sh)
{
  SH2movli(sh, 0, (11<<4)+9);
}
#define init_13_0_11_10
void FASTCALL call_13_0_11_10(SH2_struct* sh)
{
  SH2movli(sh, 0, (11<<4)+10);
}
#define init_13_0_11_11
void FASTCALL call_13_0_11_11(SH2_struct* sh)
{
  SH2movli(sh, 0, (11<<4)+11);
}
#define init_13_0_11_12
void FASTCALL call_13_0_11_12(SH2_struct* sh)
{
  SH2movli(sh, 0, (11<<4)+12);
}
#define init_13_0_11_13
void FASTCALL call_13_0_11_13(SH2_struct* sh)
{
  SH2movli(sh, 0, (11<<4)+13);
}
#define init_13_0_11_14
void FASTCALL call_13_0_11_14(SH2_struct* sh)
{
  SH2movli(sh, 0, (11<<4)+14);
}
#define init_13_0_11_15
void FASTCALL call_13_0_11_15(SH2_struct* sh)
{
  SH2movli(sh, 0, (11<<4)+15);
}
#define init_13_0_12_0
void FASTCALL call_13_0_12_0(SH2_struct* sh)
{
  SH2movli(sh, 0, (12<<4)+0);
}
#define init_13_0_12_1
void FASTCALL call_13_0_12_1(SH2_struct* sh)
{
  SH2movli(sh, 0, (12<<4)+1);
}
#define init_13_0_12_2
void FASTCALL call_13_0_12_2(SH2_struct* sh)
{
  SH2movli(sh, 0, (12<<4)+2);
}
#define init_13_0_12_3
void FASTCALL call_13_0_12_3(SH2_struct* sh)
{
  SH2movli(sh, 0, (12<<4)+3);
}
#define init_13_0_12_4
void FASTCALL call_13_0_12_4(SH2_struct* sh)
{
  SH2movli(sh, 0, (12<<4)+4);
}
#define init_13_0_12_5
void FASTCALL call_13_0_12_5(SH2_struct* sh)
{
  SH2movli(sh, 0, (12<<4)+5);
}
#define init_13_0_12_6
void FASTCALL call_13_0_12_6(SH2_struct* sh)
{
  SH2movli(sh, 0, (12<<4)+6);
}
#define init_13_0_12_7
void FASTCALL call_13_0_12_7(SH2_struct* sh)
{
  SH2movli(sh, 0, (12<<4)+7);
}
#define init_13_0_12_8
void FASTCALL call_13_0_12_8(SH2_struct* sh)
{
  SH2movli(sh, 0, (12<<4)+8);
}
#define init_13_0_12_9
void FASTCALL call_13_0_12_9(SH2_struct* sh)
{
  SH2movli(sh, 0, (12<<4)+9);
}
#define init_13_0_12_10
void FASTCALL call_13_0_12_10(SH2_struct* sh)
{
  SH2movli(sh, 0, (12<<4)+10);
}
#define init_13_0_12_11
void FASTCALL call_13_0_12_11(SH2_struct* sh)
{
  SH2movli(sh, 0, (12<<4)+11);
}
#define init_13_0_12_12
void FASTCALL call_13_0_12_12(SH2_struct* sh)
{
  SH2movli(sh, 0, (12<<4)+12);
}
#define init_13_0_12_13
void FASTCALL call_13_0_12_13(SH2_struct* sh)
{
  SH2movli(sh, 0, (12<<4)+13);
}
#define init_13_0_12_14
void FASTCALL call_13_0_12_14(SH2_struct* sh)
{
  SH2movli(sh, 0, (12<<4)+14);
}
#define init_13_0_12_15
void FASTCALL call_13_0_12_15(SH2_struct* sh)
{
  SH2movli(sh, 0, (12<<4)+15);
}
#define init_13_0_13_0
void FASTCALL call_13_0_13_0(SH2_struct* sh)
{
  SH2movli(sh, 0, (13<<4)+0);
}
#define init_13_0_13_1
void FASTCALL call_13_0_13_1(SH2_struct* sh)
{
  SH2movli(sh, 0, (13<<4)+1);
}
#define init_13_0_13_2
void FASTCALL call_13_0_13_2(SH2_struct* sh)
{
  SH2movli(sh, 0, (13<<4)+2);
}
#define init_13_0_13_3
void FASTCALL call_13_0_13_3(SH2_struct* sh)
{
  SH2movli(sh, 0, (13<<4)+3);
}
#define init_13_0_13_4
void FASTCALL call_13_0_13_4(SH2_struct* sh)
{
  SH2movli(sh, 0, (13<<4)+4);
}
#define init_13_0_13_5
void FASTCALL call_13_0_13_5(SH2_struct* sh)
{
  SH2movli(sh, 0, (13<<4)+5);
}
#define init_13_0_13_6
void FASTCALL call_13_0_13_6(SH2_struct* sh)
{
  SH2movli(sh, 0, (13<<4)+6);
}
#define init_13_0_13_7
void FASTCALL call_13_0_13_7(SH2_struct* sh)
{
  SH2movli(sh, 0, (13<<4)+7);
}
#define init_13_0_13_8
void FASTCALL call_13_0_13_8(SH2_struct* sh)
{
  SH2movli(sh, 0, (13<<4)+8);
}
#define init_13_0_13_9
void FASTCALL call_13_0_13_9(SH2_struct* sh)
{
  SH2movli(sh, 0, (13<<4)+9);
}
#define init_13_0_13_10
void FASTCALL call_13_0_13_10(SH2_struct* sh)
{
  SH2movli(sh, 0, (13<<4)+10);
}
#define init_13_0_13_11
void FASTCALL call_13_0_13_11(SH2_struct* sh)
{
  SH2movli(sh, 0, (13<<4)+11);
}
#define init_13_0_13_12
void FASTCALL call_13_0_13_12(SH2_struct* sh)
{
  SH2movli(sh, 0, (13<<4)+12);
}
#define init_13_0_13_13
void FASTCALL call_13_0_13_13(SH2_struct* sh)
{
  SH2movli(sh, 0, (13<<4)+13);
}
#define init_13_0_13_14
void FASTCALL call_13_0_13_14(SH2_struct* sh)
{
  SH2movli(sh, 0, (13<<4)+14);
}
#define init_13_0_13_15
void FASTCALL call_13_0_13_15(SH2_struct* sh)
{
  SH2movli(sh, 0, (13<<4)+15);
}
#define init_13_0_14_0
void FASTCALL call_13_0_14_0(SH2_struct* sh)
{
  SH2movli(sh, 0, (14<<4)+0);
}
#define init_13_0_14_1
void FASTCALL call_13_0_14_1(SH2_struct* sh)
{
  SH2movli(sh, 0, (14<<4)+1);
}
#define init_13_0_14_2
void FASTCALL call_13_0_14_2(SH2_struct* sh)
{
  SH2movli(sh, 0, (14<<4)+2);
}
#define init_13_0_14_3
void FASTCALL call_13_0_14_3(SH2_struct* sh)
{
  SH2movli(sh, 0, (14<<4)+3);
}
#define init_13_0_14_4
void FASTCALL call_13_0_14_4(SH2_struct* sh)
{
  SH2movli(sh, 0, (14<<4)+4);
}
#define init_13_0_14_5
void FASTCALL call_13_0_14_5(SH2_struct* sh)
{
  SH2movli(sh, 0, (14<<4)+5);
}
#define init_13_0_14_6
void FASTCALL call_13_0_14_6(SH2_struct* sh)
{
  SH2movli(sh, 0, (14<<4)+6);
}
#define init_13_0_14_7
void FASTCALL call_13_0_14_7(SH2_struct* sh)
{
  SH2movli(sh, 0, (14<<4)+7);
}
#define init_13_0_14_8
void FASTCALL call_13_0_14_8(SH2_struct* sh)
{
  SH2movli(sh, 0, (14<<4)+8);
}
#define init_13_0_14_9
void FASTCALL call_13_0_14_9(SH2_struct* sh)
{
  SH2movli(sh, 0, (14<<4)+9);
}
#define init_13_0_14_10
void FASTCALL call_13_0_14_10(SH2_struct* sh)
{
  SH2movli(sh, 0, (14<<4)+10);
}
#define init_13_0_14_11
void FASTCALL call_13_0_14_11(SH2_struct* sh)
{
  SH2movli(sh, 0, (14<<4)+11);
}
#define init_13_0_14_12
void FASTCALL call_13_0_14_12(SH2_struct* sh)
{
  SH2movli(sh, 0, (14<<4)+12);
}
#define init_13_0_14_13
void FASTCALL call_13_0_14_13(SH2_struct* sh)
{
  SH2movli(sh, 0, (14<<4)+13);
}
#define init_13_0_14_14
void FASTCALL call_13_0_14_14(SH2_struct* sh)
{
  SH2movli(sh, 0, (14<<4)+14);
}
#define init_13_0_14_15
void FASTCALL call_13_0_14_15(SH2_struct* sh)
{
  SH2movli(sh, 0, (14<<4)+15);
}
#define init_13_0_15_0
void FASTCALL call_13_0_15_0(SH2_struct* sh)
{
  SH2movli(sh, 0, (15<<4)+0);
}
#define init_13_0_15_1
void FASTCALL call_13_0_15_1(SH2_struct* sh)
{
  SH2movli(sh, 0, (15<<4)+1);
}
#define init_13_0_15_2
void FASTCALL call_13_0_15_2(SH2_struct* sh)
{
  SH2movli(sh, 0, (15<<4)+2);
}
#define init_13_0_15_3
void FASTCALL call_13_0_15_3(SH2_struct* sh)
{
  SH2movli(sh, 0, (15<<4)+3);
}
#define init_13_0_15_4
void FASTCALL call_13_0_15_4(SH2_struct* sh)
{
  SH2movli(sh, 0, (15<<4)+4);
}
#define init_13_0_15_5
void FASTCALL call_13_0_15_5(SH2_struct* sh)
{
  SH2movli(sh, 0, (15<<4)+5);
}
#define init_13_0_15_6
void FASTCALL call_13_0_15_6(SH2_struct* sh)
{
  SH2movli(sh, 0, (15<<4)+6);
}
#define init_13_0_15_7
void FASTCALL call_13_0_15_7(SH2_struct* sh)
{
  SH2movli(sh, 0, (15<<4)+7);
}
#define init_13_0_15_8
void FASTCALL call_13_0_15_8(SH2_struct* sh)
{
  SH2movli(sh, 0, (15<<4)+8);
}
#define init_13_0_15_9
void FASTCALL call_13_0_15_9(SH2_struct* sh)
{
  SH2movli(sh, 0, (15<<4)+9);
}
#define init_13_0_15_10
void FASTCALL call_13_0_15_10(SH2_struct* sh)
{
  SH2movli(sh, 0, (15<<4)+10);
}
#define init_13_0_15_11
void FASTCALL call_13_0_15_11(SH2_struct* sh)
{
  SH2movli(sh, 0, (15<<4)+11);
}
#define init_13_0_15_12
void FASTCALL call_13_0_15_12(SH2_struct* sh)
{
  SH2movli(sh, 0, (15<<4)+12);
}
#define init_13_0_15_13
void FASTCALL call_13_0_15_13(SH2_struct* sh)
{
  SH2movli(sh, 0, (15<<4)+13);
}
#define init_13_0_15_14
void FASTCALL call_13_0_15_14(SH2_struct* sh)
{
  SH2movli(sh, 0, (15<<4)+14);
}
#define init_13_0_15_15
void FASTCALL call_13_0_15_15(SH2_struct* sh)
{
  SH2movli(sh, 0, (15<<4)+15);
}
#define init_13_1_0_0
void FASTCALL call_13_1_0_0(SH2_struct* sh)
{
  SH2movli(sh, 1, (0<<4)+0);
}
#define init_13_1_0_1
void FASTCALL call_13_1_0_1(SH2_struct* sh)
{
  SH2movli(sh, 1, (0<<4)+1);
}
#define init_13_1_0_2
void FASTCALL call_13_1_0_2(SH2_struct* sh)
{
  SH2movli(sh, 1, (0<<4)+2);
}
#define init_13_1_0_3
void FASTCALL call_13_1_0_3(SH2_struct* sh)
{
  SH2movli(sh, 1, (0<<4)+3);
}
#define init_13_1_0_4
void FASTCALL call_13_1_0_4(SH2_struct* sh)
{
  SH2movli(sh, 1, (0<<4)+4);
}
#define init_13_1_0_5
void FASTCALL call_13_1_0_5(SH2_struct* sh)
{
  SH2movli(sh, 1, (0<<4)+5);
}
#define init_13_1_0_6
void FASTCALL call_13_1_0_6(SH2_struct* sh)
{
  SH2movli(sh, 1, (0<<4)+6);
}
#define init_13_1_0_7
void FASTCALL call_13_1_0_7(SH2_struct* sh)
{
  SH2movli(sh, 1, (0<<4)+7);
}
#define init_13_1_0_8
void FASTCALL call_13_1_0_8(SH2_struct* sh)
{
  SH2movli(sh, 1, (0<<4)+8);
}
#define init_13_1_0_9
void FASTCALL call_13_1_0_9(SH2_struct* sh)
{
  SH2movli(sh, 1, (0<<4)+9);
}
#define init_13_1_0_10
void FASTCALL call_13_1_0_10(SH2_struct* sh)
{
  SH2movli(sh, 1, (0<<4)+10);
}
#define init_13_1_0_11
void FASTCALL call_13_1_0_11(SH2_struct* sh)
{
  SH2movli(sh, 1, (0<<4)+11);
}
#define init_13_1_0_12
void FASTCALL call_13_1_0_12(SH2_struct* sh)
{
  SH2movli(sh, 1, (0<<4)+12);
}
#define init_13_1_0_13
void FASTCALL call_13_1_0_13(SH2_struct* sh)
{
  SH2movli(sh, 1, (0<<4)+13);
}
#define init_13_1_0_14
void FASTCALL call_13_1_0_14(SH2_struct* sh)
{
  SH2movli(sh, 1, (0<<4)+14);
}
#define init_13_1_0_15
void FASTCALL call_13_1_0_15(SH2_struct* sh)
{
  SH2movli(sh, 1, (0<<4)+15);
}
#define init_13_1_1_0
void FASTCALL call_13_1_1_0(SH2_struct* sh)
{
  SH2movli(sh, 1, (1<<4)+0);
}
#define init_13_1_1_1
void FASTCALL call_13_1_1_1(SH2_struct* sh)
{
  SH2movli(sh, 1, (1<<4)+1);
}
#define init_13_1_1_2
void FASTCALL call_13_1_1_2(SH2_struct* sh)
{
  SH2movli(sh, 1, (1<<4)+2);
}
#define init_13_1_1_3
void FASTCALL call_13_1_1_3(SH2_struct* sh)
{
  SH2movli(sh, 1, (1<<4)+3);
}
#define init_13_1_1_4
void FASTCALL call_13_1_1_4(SH2_struct* sh)
{
  SH2movli(sh, 1, (1<<4)+4);
}
#define init_13_1_1_5
void FASTCALL call_13_1_1_5(SH2_struct* sh)
{
  SH2movli(sh, 1, (1<<4)+5);
}
#define init_13_1_1_6
void FASTCALL call_13_1_1_6(SH2_struct* sh)
{
  SH2movli(sh, 1, (1<<4)+6);
}
#define init_13_1_1_7
void FASTCALL call_13_1_1_7(SH2_struct* sh)
{
  SH2movli(sh, 1, (1<<4)+7);
}
#define init_13_1_1_8
void FASTCALL call_13_1_1_8(SH2_struct* sh)
{
  SH2movli(sh, 1, (1<<4)+8);
}
#define init_13_1_1_9
void FASTCALL call_13_1_1_9(SH2_struct* sh)
{
  SH2movli(sh, 1, (1<<4)+9);
}
#define init_13_1_1_10
void FASTCALL call_13_1_1_10(SH2_struct* sh)
{
  SH2movli(sh, 1, (1<<4)+10);
}
#define init_13_1_1_11
void FASTCALL call_13_1_1_11(SH2_struct* sh)
{
  SH2movli(sh, 1, (1<<4)+11);
}
#define init_13_1_1_12
void FASTCALL call_13_1_1_12(SH2_struct* sh)
{
  SH2movli(sh, 1, (1<<4)+12);
}
#define init_13_1_1_13
void FASTCALL call_13_1_1_13(SH2_struct* sh)
{
  SH2movli(sh, 1, (1<<4)+13);
}
#define init_13_1_1_14
void FASTCALL call_13_1_1_14(SH2_struct* sh)
{
  SH2movli(sh, 1, (1<<4)+14);
}
#define init_13_1_1_15
void FASTCALL call_13_1_1_15(SH2_struct* sh)
{
  SH2movli(sh, 1, (1<<4)+15);
}
#define init_13_1_2_0
void FASTCALL call_13_1_2_0(SH2_struct* sh)
{
  SH2movli(sh, 1, (2<<4)+0);
}
#define init_13_1_2_1
void FASTCALL call_13_1_2_1(SH2_struct* sh)
{
  SH2movli(sh, 1, (2<<4)+1);
}
#define init_13_1_2_2
void FASTCALL call_13_1_2_2(SH2_struct* sh)
{
  SH2movli(sh, 1, (2<<4)+2);
}
#define init_13_1_2_3
void FASTCALL call_13_1_2_3(SH2_struct* sh)
{
  SH2movli(sh, 1, (2<<4)+3);
}
#define init_13_1_2_4
void FASTCALL call_13_1_2_4(SH2_struct* sh)
{
  SH2movli(sh, 1, (2<<4)+4);
}
#define init_13_1_2_5
void FASTCALL call_13_1_2_5(SH2_struct* sh)
{
  SH2movli(sh, 1, (2<<4)+5);
}
#define init_13_1_2_6
void FASTCALL call_13_1_2_6(SH2_struct* sh)
{
  SH2movli(sh, 1, (2<<4)+6);
}
#define init_13_1_2_7
void FASTCALL call_13_1_2_7(SH2_struct* sh)
{
  SH2movli(sh, 1, (2<<4)+7);
}
#define init_13_1_2_8
void FASTCALL call_13_1_2_8(SH2_struct* sh)
{
  SH2movli(sh, 1, (2<<4)+8);
}
#define init_13_1_2_9
void FASTCALL call_13_1_2_9(SH2_struct* sh)
{
  SH2movli(sh, 1, (2<<4)+9);
}
#define init_13_1_2_10
void FASTCALL call_13_1_2_10(SH2_struct* sh)
{
  SH2movli(sh, 1, (2<<4)+10);
}
#define init_13_1_2_11
void FASTCALL call_13_1_2_11(SH2_struct* sh)
{
  SH2movli(sh, 1, (2<<4)+11);
}
#define init_13_1_2_12
void FASTCALL call_13_1_2_12(SH2_struct* sh)
{
  SH2movli(sh, 1, (2<<4)+12);
}
#define init_13_1_2_13
void FASTCALL call_13_1_2_13(SH2_struct* sh)
{
  SH2movli(sh, 1, (2<<4)+13);
}
#define init_13_1_2_14
void FASTCALL call_13_1_2_14(SH2_struct* sh)
{
  SH2movli(sh, 1, (2<<4)+14);
}
#define init_13_1_2_15
void FASTCALL call_13_1_2_15(SH2_struct* sh)
{
  SH2movli(sh, 1, (2<<4)+15);
}
#define init_13_1_3_0
void FASTCALL call_13_1_3_0(SH2_struct* sh)
{
  SH2movli(sh, 1, (3<<4)+0);
}
#define init_13_1_3_1
void FASTCALL call_13_1_3_1(SH2_struct* sh)
{
  SH2movli(sh, 1, (3<<4)+1);
}
#define init_13_1_3_2
void FASTCALL call_13_1_3_2(SH2_struct* sh)
{
  SH2movli(sh, 1, (3<<4)+2);
}
#define init_13_1_3_3
void FASTCALL call_13_1_3_3(SH2_struct* sh)
{
  SH2movli(sh, 1, (3<<4)+3);
}
#define init_13_1_3_4
void FASTCALL call_13_1_3_4(SH2_struct* sh)
{
  SH2movli(sh, 1, (3<<4)+4);
}
#define init_13_1_3_5
void FASTCALL call_13_1_3_5(SH2_struct* sh)
{
  SH2movli(sh, 1, (3<<4)+5);
}
#define init_13_1_3_6
void FASTCALL call_13_1_3_6(SH2_struct* sh)
{
  SH2movli(sh, 1, (3<<4)+6);
}
#define init_13_1_3_7
void FASTCALL call_13_1_3_7(SH2_struct* sh)
{
  SH2movli(sh, 1, (3<<4)+7);
}
#define init_13_1_3_8
void FASTCALL call_13_1_3_8(SH2_struct* sh)
{
  SH2movli(sh, 1, (3<<4)+8);
}
#define init_13_1_3_9
void FASTCALL call_13_1_3_9(SH2_struct* sh)
{
  SH2movli(sh, 1, (3<<4)+9);
}
#define init_13_1_3_10
void FASTCALL call_13_1_3_10(SH2_struct* sh)
{
  SH2movli(sh, 1, (3<<4)+10);
}
#define init_13_1_3_11
void FASTCALL call_13_1_3_11(SH2_struct* sh)
{
  SH2movli(sh, 1, (3<<4)+11);
}
#define init_13_1_3_12
void FASTCALL call_13_1_3_12(SH2_struct* sh)
{
  SH2movli(sh, 1, (3<<4)+12);
}
#define init_13_1_3_13
void FASTCALL call_13_1_3_13(SH2_struct* sh)
{
  SH2movli(sh, 1, (3<<4)+13);
}
#define init_13_1_3_14
void FASTCALL call_13_1_3_14(SH2_struct* sh)
{
  SH2movli(sh, 1, (3<<4)+14);
}
#define init_13_1_3_15
void FASTCALL call_13_1_3_15(SH2_struct* sh)
{
  SH2movli(sh, 1, (3<<4)+15);
}
#define init_13_1_4_0
void FASTCALL call_13_1_4_0(SH2_struct* sh)
{
  SH2movli(sh, 1, (4<<4)+0);
}
#define init_13_1_4_1
void FASTCALL call_13_1_4_1(SH2_struct* sh)
{
  SH2movli(sh, 1, (4<<4)+1);
}
#define init_13_1_4_2
void FASTCALL call_13_1_4_2(SH2_struct* sh)
{
  SH2movli(sh, 1, (4<<4)+2);
}
#define init_13_1_4_3
void FASTCALL call_13_1_4_3(SH2_struct* sh)
{
  SH2movli(sh, 1, (4<<4)+3);
}
#define init_13_1_4_4
void FASTCALL call_13_1_4_4(SH2_struct* sh)
{
  SH2movli(sh, 1, (4<<4)+4);
}
#define init_13_1_4_5
void FASTCALL call_13_1_4_5(SH2_struct* sh)
{
  SH2movli(sh, 1, (4<<4)+5);
}
#define init_13_1_4_6
void FASTCALL call_13_1_4_6(SH2_struct* sh)
{
  SH2movli(sh, 1, (4<<4)+6);
}
#define init_13_1_4_7
void FASTCALL call_13_1_4_7(SH2_struct* sh)
{
  SH2movli(sh, 1, (4<<4)+7);
}
#define init_13_1_4_8
void FASTCALL call_13_1_4_8(SH2_struct* sh)
{
  SH2movli(sh, 1, (4<<4)+8);
}
#define init_13_1_4_9
void FASTCALL call_13_1_4_9(SH2_struct* sh)
{
  SH2movli(sh, 1, (4<<4)+9);
}
#define init_13_1_4_10
void FASTCALL call_13_1_4_10(SH2_struct* sh)
{
  SH2movli(sh, 1, (4<<4)+10);
}
#define init_13_1_4_11
void FASTCALL call_13_1_4_11(SH2_struct* sh)
{
  SH2movli(sh, 1, (4<<4)+11);
}
#define init_13_1_4_12
void FASTCALL call_13_1_4_12(SH2_struct* sh)
{
  SH2movli(sh, 1, (4<<4)+12);
}
#define init_13_1_4_13
void FASTCALL call_13_1_4_13(SH2_struct* sh)
{
  SH2movli(sh, 1, (4<<4)+13);
}
#define init_13_1_4_14
void FASTCALL call_13_1_4_14(SH2_struct* sh)
{
  SH2movli(sh, 1, (4<<4)+14);
}
#define init_13_1_4_15
void FASTCALL call_13_1_4_15(SH2_struct* sh)
{
  SH2movli(sh, 1, (4<<4)+15);
}
#define init_13_1_5_0
void FASTCALL call_13_1_5_0(SH2_struct* sh)
{
  SH2movli(sh, 1, (5<<4)+0);
}
#define init_13_1_5_1
void FASTCALL call_13_1_5_1(SH2_struct* sh)
{
  SH2movli(sh, 1, (5<<4)+1);
}
#define init_13_1_5_2
void FASTCALL call_13_1_5_2(SH2_struct* sh)
{
  SH2movli(sh, 1, (5<<4)+2);
}
#define init_13_1_5_3
void FASTCALL call_13_1_5_3(SH2_struct* sh)
{
  SH2movli(sh, 1, (5<<4)+3);
}
#define init_13_1_5_4
void FASTCALL call_13_1_5_4(SH2_struct* sh)
{
  SH2movli(sh, 1, (5<<4)+4);
}
#define init_13_1_5_5
void FASTCALL call_13_1_5_5(SH2_struct* sh)
{
  SH2movli(sh, 1, (5<<4)+5);
}
#define init_13_1_5_6
void FASTCALL call_13_1_5_6(SH2_struct* sh)
{
  SH2movli(sh, 1, (5<<4)+6);
}
#define init_13_1_5_7
void FASTCALL call_13_1_5_7(SH2_struct* sh)
{
  SH2movli(sh, 1, (5<<4)+7);
}
#define init_13_1_5_8
void FASTCALL call_13_1_5_8(SH2_struct* sh)
{
  SH2movli(sh, 1, (5<<4)+8);
}
#define init_13_1_5_9
void FASTCALL call_13_1_5_9(SH2_struct* sh)
{
  SH2movli(sh, 1, (5<<4)+9);
}
#define init_13_1_5_10
void FASTCALL call_13_1_5_10(SH2_struct* sh)
{
  SH2movli(sh, 1, (5<<4)+10);
}
#define init_13_1_5_11
void FASTCALL call_13_1_5_11(SH2_struct* sh)
{
  SH2movli(sh, 1, (5<<4)+11);
}
#define init_13_1_5_12
void FASTCALL call_13_1_5_12(SH2_struct* sh)
{
  SH2movli(sh, 1, (5<<4)+12);
}
#define init_13_1_5_13
void FASTCALL call_13_1_5_13(SH2_struct* sh)
{
  SH2movli(sh, 1, (5<<4)+13);
}
#define init_13_1_5_14
void FASTCALL call_13_1_5_14(SH2_struct* sh)
{
  SH2movli(sh, 1, (5<<4)+14);
}
#define init_13_1_5_15
void FASTCALL call_13_1_5_15(SH2_struct* sh)
{
  SH2movli(sh, 1, (5<<4)+15);
}
#define init_13_1_6_0
void FASTCALL call_13_1_6_0(SH2_struct* sh)
{
  SH2movli(sh, 1, (6<<4)+0);
}
#define init_13_1_6_1
void FASTCALL call_13_1_6_1(SH2_struct* sh)
{
  SH2movli(sh, 1, (6<<4)+1);
}
#define init_13_1_6_2
void FASTCALL call_13_1_6_2(SH2_struct* sh)
{
  SH2movli(sh, 1, (6<<4)+2);
}
#define init_13_1_6_3
void FASTCALL call_13_1_6_3(SH2_struct* sh)
{
  SH2movli(sh, 1, (6<<4)+3);
}
#define init_13_1_6_4
void FASTCALL call_13_1_6_4(SH2_struct* sh)
{
  SH2movli(sh, 1, (6<<4)+4);
}
#define init_13_1_6_5
void FASTCALL call_13_1_6_5(SH2_struct* sh)
{
  SH2movli(sh, 1, (6<<4)+5);
}
#define init_13_1_6_6
void FASTCALL call_13_1_6_6(SH2_struct* sh)
{
  SH2movli(sh, 1, (6<<4)+6);
}
#define init_13_1_6_7
void FASTCALL call_13_1_6_7(SH2_struct* sh)
{
  SH2movli(sh, 1, (6<<4)+7);
}
#define init_13_1_6_8
void FASTCALL call_13_1_6_8(SH2_struct* sh)
{
  SH2movli(sh, 1, (6<<4)+8);
}
#define init_13_1_6_9
void FASTCALL call_13_1_6_9(SH2_struct* sh)
{
  SH2movli(sh, 1, (6<<4)+9);
}
#define init_13_1_6_10
void FASTCALL call_13_1_6_10(SH2_struct* sh)
{
  SH2movli(sh, 1, (6<<4)+10);
}
#define init_13_1_6_11
void FASTCALL call_13_1_6_11(SH2_struct* sh)
{
  SH2movli(sh, 1, (6<<4)+11);
}
#define init_13_1_6_12
void FASTCALL call_13_1_6_12(SH2_struct* sh)
{
  SH2movli(sh, 1, (6<<4)+12);
}
#define init_13_1_6_13
void FASTCALL call_13_1_6_13(SH2_struct* sh)
{
  SH2movli(sh, 1, (6<<4)+13);
}
#define init_13_1_6_14
void FASTCALL call_13_1_6_14(SH2_struct* sh)
{
  SH2movli(sh, 1, (6<<4)+14);
}
#define init_13_1_6_15
void FASTCALL call_13_1_6_15(SH2_struct* sh)
{
  SH2movli(sh, 1, (6<<4)+15);
}
#define init_13_1_7_0
void FASTCALL call_13_1_7_0(SH2_struct* sh)
{
  SH2movli(sh, 1, (7<<4)+0);
}
#define init_13_1_7_1
void FASTCALL call_13_1_7_1(SH2_struct* sh)
{
  SH2movli(sh, 1, (7<<4)+1);
}
#define init_13_1_7_2
void FASTCALL call_13_1_7_2(SH2_struct* sh)
{
  SH2movli(sh, 1, (7<<4)+2);
}
#define init_13_1_7_3
void FASTCALL call_13_1_7_3(SH2_struct* sh)
{
  SH2movli(sh, 1, (7<<4)+3);
}
#define init_13_1_7_4
void FASTCALL call_13_1_7_4(SH2_struct* sh)
{
  SH2movli(sh, 1, (7<<4)+4);
}
#define init_13_1_7_5
void FASTCALL call_13_1_7_5(SH2_struct* sh)
{
  SH2movli(sh, 1, (7<<4)+5);
}
#define init_13_1_7_6
void FASTCALL call_13_1_7_6(SH2_struct* sh)
{
  SH2movli(sh, 1, (7<<4)+6);
}
#define init_13_1_7_7
void FASTCALL call_13_1_7_7(SH2_struct* sh)
{
  SH2movli(sh, 1, (7<<4)+7);
}
#define init_13_1_7_8
void FASTCALL call_13_1_7_8(SH2_struct* sh)
{
  SH2movli(sh, 1, (7<<4)+8);
}
#define init_13_1_7_9
void FASTCALL call_13_1_7_9(SH2_struct* sh)
{
  SH2movli(sh, 1, (7<<4)+9);
}
#define init_13_1_7_10
void FASTCALL call_13_1_7_10(SH2_struct* sh)
{
  SH2movli(sh, 1, (7<<4)+10);
}
#define init_13_1_7_11
void FASTCALL call_13_1_7_11(SH2_struct* sh)
{
  SH2movli(sh, 1, (7<<4)+11);
}
#define init_13_1_7_12
void FASTCALL call_13_1_7_12(SH2_struct* sh)
{
  SH2movli(sh, 1, (7<<4)+12);
}
#define init_13_1_7_13
void FASTCALL call_13_1_7_13(SH2_struct* sh)
{
  SH2movli(sh, 1, (7<<4)+13);
}
#define init_13_1_7_14
void FASTCALL call_13_1_7_14(SH2_struct* sh)
{
  SH2movli(sh, 1, (7<<4)+14);
}
#define init_13_1_7_15
void FASTCALL call_13_1_7_15(SH2_struct* sh)
{
  SH2movli(sh, 1, (7<<4)+15);
}
#define init_13_1_8_0
void FASTCALL call_13_1_8_0(SH2_struct* sh)
{
  SH2movli(sh, 1, (8<<4)+0);
}
#define init_13_1_8_1
void FASTCALL call_13_1_8_1(SH2_struct* sh)
{
  SH2movli(sh, 1, (8<<4)+1);
}
#define init_13_1_8_2
void FASTCALL call_13_1_8_2(SH2_struct* sh)
{
  SH2movli(sh, 1, (8<<4)+2);
}
#define init_13_1_8_3
void FASTCALL call_13_1_8_3(SH2_struct* sh)
{
  SH2movli(sh, 1, (8<<4)+3);
}
#define init_13_1_8_4
void FASTCALL call_13_1_8_4(SH2_struct* sh)
{
  SH2movli(sh, 1, (8<<4)+4);
}
#define init_13_1_8_5
void FASTCALL call_13_1_8_5(SH2_struct* sh)
{
  SH2movli(sh, 1, (8<<4)+5);
}
#define init_13_1_8_6
void FASTCALL call_13_1_8_6(SH2_struct* sh)
{
  SH2movli(sh, 1, (8<<4)+6);
}
#define init_13_1_8_7
void FASTCALL call_13_1_8_7(SH2_struct* sh)
{
  SH2movli(sh, 1, (8<<4)+7);
}
#define init_13_1_8_8
void FASTCALL call_13_1_8_8(SH2_struct* sh)
{
  SH2movli(sh, 1, (8<<4)+8);
}
#define init_13_1_8_9
void FASTCALL call_13_1_8_9(SH2_struct* sh)
{
  SH2movli(sh, 1, (8<<4)+9);
}
#define init_13_1_8_10
void FASTCALL call_13_1_8_10(SH2_struct* sh)
{
  SH2movli(sh, 1, (8<<4)+10);
}
#define init_13_1_8_11
void FASTCALL call_13_1_8_11(SH2_struct* sh)
{
  SH2movli(sh, 1, (8<<4)+11);
}
#define init_13_1_8_12
void FASTCALL call_13_1_8_12(SH2_struct* sh)
{
  SH2movli(sh, 1, (8<<4)+12);
}
#define init_13_1_8_13
void FASTCALL call_13_1_8_13(SH2_struct* sh)
{
  SH2movli(sh, 1, (8<<4)+13);
}
#define init_13_1_8_14
void FASTCALL call_13_1_8_14(SH2_struct* sh)
{
  SH2movli(sh, 1, (8<<4)+14);
}
#define init_13_1_8_15
void FASTCALL call_13_1_8_15(SH2_struct* sh)
{
  SH2movli(sh, 1, (8<<4)+15);
}
#define init_13_1_9_0
void FASTCALL call_13_1_9_0(SH2_struct* sh)
{
  SH2movli(sh, 1, (9<<4)+0);
}
#define init_13_1_9_1
void FASTCALL call_13_1_9_1(SH2_struct* sh)
{
  SH2movli(sh, 1, (9<<4)+1);
}
#define init_13_1_9_2
void FASTCALL call_13_1_9_2(SH2_struct* sh)
{
  SH2movli(sh, 1, (9<<4)+2);
}
#define init_13_1_9_3
void FASTCALL call_13_1_9_3(SH2_struct* sh)
{
  SH2movli(sh, 1, (9<<4)+3);
}
#define init_13_1_9_4
void FASTCALL call_13_1_9_4(SH2_struct* sh)
{
  SH2movli(sh, 1, (9<<4)+4);
}
#define init_13_1_9_5
void FASTCALL call_13_1_9_5(SH2_struct* sh)
{
  SH2movli(sh, 1, (9<<4)+5);
}
#define init_13_1_9_6
void FASTCALL call_13_1_9_6(SH2_struct* sh)
{
  SH2movli(sh, 1, (9<<4)+6);
}
#define init_13_1_9_7
void FASTCALL call_13_1_9_7(SH2_struct* sh)
{
  SH2movli(sh, 1, (9<<4)+7);
}
#define init_13_1_9_8
void FASTCALL call_13_1_9_8(SH2_struct* sh)
{
  SH2movli(sh, 1, (9<<4)+8);
}
#define init_13_1_9_9
void FASTCALL call_13_1_9_9(SH2_struct* sh)
{
  SH2movli(sh, 1, (9<<4)+9);
}
#define init_13_1_9_10
void FASTCALL call_13_1_9_10(SH2_struct* sh)
{
  SH2movli(sh, 1, (9<<4)+10);
}
#define init_13_1_9_11
void FASTCALL call_13_1_9_11(SH2_struct* sh)
{
  SH2movli(sh, 1, (9<<4)+11);
}
#define init_13_1_9_12
void FASTCALL call_13_1_9_12(SH2_struct* sh)
{
  SH2movli(sh, 1, (9<<4)+12);
}
#define init_13_1_9_13
void FASTCALL call_13_1_9_13(SH2_struct* sh)
{
  SH2movli(sh, 1, (9<<4)+13);
}
#define init_13_1_9_14
void FASTCALL call_13_1_9_14(SH2_struct* sh)
{
  SH2movli(sh, 1, (9<<4)+14);
}
#define init_13_1_9_15
void FASTCALL call_13_1_9_15(SH2_struct* sh)
{
  SH2movli(sh, 1, (9<<4)+15);
}
#define init_13_1_10_0
void FASTCALL call_13_1_10_0(SH2_struct* sh)
{
  SH2movli(sh, 1, (10<<4)+0);
}
#define init_13_1_10_1
void FASTCALL call_13_1_10_1(SH2_struct* sh)
{
  SH2movli(sh, 1, (10<<4)+1);
}
#define init_13_1_10_2
void FASTCALL call_13_1_10_2(SH2_struct* sh)
{
  SH2movli(sh, 1, (10<<4)+2);
}
#define init_13_1_10_3
void FASTCALL call_13_1_10_3(SH2_struct* sh)
{
  SH2movli(sh, 1, (10<<4)+3);
}
#define init_13_1_10_4
void FASTCALL call_13_1_10_4(SH2_struct* sh)
{
  SH2movli(sh, 1, (10<<4)+4);
}
#define init_13_1_10_5
void FASTCALL call_13_1_10_5(SH2_struct* sh)
{
  SH2movli(sh, 1, (10<<4)+5);
}
#define init_13_1_10_6
void FASTCALL call_13_1_10_6(SH2_struct* sh)
{
  SH2movli(sh, 1, (10<<4)+6);
}
#define init_13_1_10_7
void FASTCALL call_13_1_10_7(SH2_struct* sh)
{
  SH2movli(sh, 1, (10<<4)+7);
}
#define init_13_1_10_8
void FASTCALL call_13_1_10_8(SH2_struct* sh)
{
  SH2movli(sh, 1, (10<<4)+8);
}
#define init_13_1_10_9
void FASTCALL call_13_1_10_9(SH2_struct* sh)
{
  SH2movli(sh, 1, (10<<4)+9);
}
#define init_13_1_10_10
void FASTCALL call_13_1_10_10(SH2_struct* sh)
{
  SH2movli(sh, 1, (10<<4)+10);
}
#define init_13_1_10_11
void FASTCALL call_13_1_10_11(SH2_struct* sh)
{
  SH2movli(sh, 1, (10<<4)+11);
}
#define init_13_1_10_12
void FASTCALL call_13_1_10_12(SH2_struct* sh)
{
  SH2movli(sh, 1, (10<<4)+12);
}
#define init_13_1_10_13
void FASTCALL call_13_1_10_13(SH2_struct* sh)
{
  SH2movli(sh, 1, (10<<4)+13);
}
#define init_13_1_10_14
void FASTCALL call_13_1_10_14(SH2_struct* sh)
{
  SH2movli(sh, 1, (10<<4)+14);
}
#define init_13_1_10_15
void FASTCALL call_13_1_10_15(SH2_struct* sh)
{
  SH2movli(sh, 1, (10<<4)+15);
}
#define init_13_1_11_0
void FASTCALL call_13_1_11_0(SH2_struct* sh)
{
  SH2movli(sh, 1, (11<<4)+0);
}
#define init_13_1_11_1
void FASTCALL call_13_1_11_1(SH2_struct* sh)
{
  SH2movli(sh, 1, (11<<4)+1);
}
#define init_13_1_11_2
void FASTCALL call_13_1_11_2(SH2_struct* sh)
{
  SH2movli(sh, 1, (11<<4)+2);
}
#define init_13_1_11_3
void FASTCALL call_13_1_11_3(SH2_struct* sh)
{
  SH2movli(sh, 1, (11<<4)+3);
}
#define init_13_1_11_4
void FASTCALL call_13_1_11_4(SH2_struct* sh)
{
  SH2movli(sh, 1, (11<<4)+4);
}
#define init_13_1_11_5
void FASTCALL call_13_1_11_5(SH2_struct* sh)
{
  SH2movli(sh, 1, (11<<4)+5);
}
#define init_13_1_11_6
void FASTCALL call_13_1_11_6(SH2_struct* sh)
{
  SH2movli(sh, 1, (11<<4)+6);
}
#define init_13_1_11_7
void FASTCALL call_13_1_11_7(SH2_struct* sh)
{
  SH2movli(sh, 1, (11<<4)+7);
}
#define init_13_1_11_8
void FASTCALL call_13_1_11_8(SH2_struct* sh)
{
  SH2movli(sh, 1, (11<<4)+8);
}
#define init_13_1_11_9
void FASTCALL call_13_1_11_9(SH2_struct* sh)
{
  SH2movli(sh, 1, (11<<4)+9);
}
#define init_13_1_11_10
void FASTCALL call_13_1_11_10(SH2_struct* sh)
{
  SH2movli(sh, 1, (11<<4)+10);
}
#define init_13_1_11_11
void FASTCALL call_13_1_11_11(SH2_struct* sh)
{
  SH2movli(sh, 1, (11<<4)+11);
}
#define init_13_1_11_12
void FASTCALL call_13_1_11_12(SH2_struct* sh)
{
  SH2movli(sh, 1, (11<<4)+12);
}
#define init_13_1_11_13
void FASTCALL call_13_1_11_13(SH2_struct* sh)
{
  SH2movli(sh, 1, (11<<4)+13);
}
#define init_13_1_11_14
void FASTCALL call_13_1_11_14(SH2_struct* sh)
{
  SH2movli(sh, 1, (11<<4)+14);
}
#define init_13_1_11_15
void FASTCALL call_13_1_11_15(SH2_struct* sh)
{
  SH2movli(sh, 1, (11<<4)+15);
}
#define init_13_1_12_0
void FASTCALL call_13_1_12_0(SH2_struct* sh)
{
  SH2movli(sh, 1, (12<<4)+0);
}
#define init_13_1_12_1
void FASTCALL call_13_1_12_1(SH2_struct* sh)
{
  SH2movli(sh, 1, (12<<4)+1);
}
#define init_13_1_12_2
void FASTCALL call_13_1_12_2(SH2_struct* sh)
{
  SH2movli(sh, 1, (12<<4)+2);
}
#define init_13_1_12_3
void FASTCALL call_13_1_12_3(SH2_struct* sh)
{
  SH2movli(sh, 1, (12<<4)+3);
}
#define init_13_1_12_4
void FASTCALL call_13_1_12_4(SH2_struct* sh)
{
  SH2movli(sh, 1, (12<<4)+4);
}
#define init_13_1_12_5
void FASTCALL call_13_1_12_5(SH2_struct* sh)
{
  SH2movli(sh, 1, (12<<4)+5);
}
#define init_13_1_12_6
void FASTCALL call_13_1_12_6(SH2_struct* sh)
{
  SH2movli(sh, 1, (12<<4)+6);
}
#define init_13_1_12_7
void FASTCALL call_13_1_12_7(SH2_struct* sh)
{
  SH2movli(sh, 1, (12<<4)+7);
}
#define init_13_1_12_8
void FASTCALL call_13_1_12_8(SH2_struct* sh)
{
  SH2movli(sh, 1, (12<<4)+8);
}
#define init_13_1_12_9
void FASTCALL call_13_1_12_9(SH2_struct* sh)
{
  SH2movli(sh, 1, (12<<4)+9);
}
#define init_13_1_12_10
void FASTCALL call_13_1_12_10(SH2_struct* sh)
{
  SH2movli(sh, 1, (12<<4)+10);
}
#define init_13_1_12_11
void FASTCALL call_13_1_12_11(SH2_struct* sh)
{
  SH2movli(sh, 1, (12<<4)+11);
}
#define init_13_1_12_12
void FASTCALL call_13_1_12_12(SH2_struct* sh)
{
  SH2movli(sh, 1, (12<<4)+12);
}
#define init_13_1_12_13
void FASTCALL call_13_1_12_13(SH2_struct* sh)
{
  SH2movli(sh, 1, (12<<4)+13);
}
#define init_13_1_12_14
void FASTCALL call_13_1_12_14(SH2_struct* sh)
{
  SH2movli(sh, 1, (12<<4)+14);
}
#define init_13_1_12_15
void FASTCALL call_13_1_12_15(SH2_struct* sh)
{
  SH2movli(sh, 1, (12<<4)+15);
}
#define init_13_1_13_0
void FASTCALL call_13_1_13_0(SH2_struct* sh)
{
  SH2movli(sh, 1, (13<<4)+0);
}
#define init_13_1_13_1
void FASTCALL call_13_1_13_1(SH2_struct* sh)
{
  SH2movli(sh, 1, (13<<4)+1);
}
#define init_13_1_13_2
void FASTCALL call_13_1_13_2(SH2_struct* sh)
{
  SH2movli(sh, 1, (13<<4)+2);
}
#define init_13_1_13_3
void FASTCALL call_13_1_13_3(SH2_struct* sh)
{
  SH2movli(sh, 1, (13<<4)+3);
}
#define init_13_1_13_4
void FASTCALL call_13_1_13_4(SH2_struct* sh)
{
  SH2movli(sh, 1, (13<<4)+4);
}
#define init_13_1_13_5
void FASTCALL call_13_1_13_5(SH2_struct* sh)
{
  SH2movli(sh, 1, (13<<4)+5);
}
#define init_13_1_13_6
void FASTCALL call_13_1_13_6(SH2_struct* sh)
{
  SH2movli(sh, 1, (13<<4)+6);
}
#define init_13_1_13_7
void FASTCALL call_13_1_13_7(SH2_struct* sh)
{
  SH2movli(sh, 1, (13<<4)+7);
}
#define init_13_1_13_8
void FASTCALL call_13_1_13_8(SH2_struct* sh)
{
  SH2movli(sh, 1, (13<<4)+8);
}
#define init_13_1_13_9
void FASTCALL call_13_1_13_9(SH2_struct* sh)
{
  SH2movli(sh, 1, (13<<4)+9);
}
#define init_13_1_13_10
void FASTCALL call_13_1_13_10(SH2_struct* sh)
{
  SH2movli(sh, 1, (13<<4)+10);
}
#define init_13_1_13_11
void FASTCALL call_13_1_13_11(SH2_struct* sh)
{
  SH2movli(sh, 1, (13<<4)+11);
}
#define init_13_1_13_12
void FASTCALL call_13_1_13_12(SH2_struct* sh)
{
  SH2movli(sh, 1, (13<<4)+12);
}
#define init_13_1_13_13
void FASTCALL call_13_1_13_13(SH2_struct* sh)
{
  SH2movli(sh, 1, (13<<4)+13);
}
#define init_13_1_13_14
void FASTCALL call_13_1_13_14(SH2_struct* sh)
{
  SH2movli(sh, 1, (13<<4)+14);
}
#define init_13_1_13_15
void FASTCALL call_13_1_13_15(SH2_struct* sh)
{
  SH2movli(sh, 1, (13<<4)+15);
}
#define init_13_1_14_0
void FASTCALL call_13_1_14_0(SH2_struct* sh)
{
  SH2movli(sh, 1, (14<<4)+0);
}
#define init_13_1_14_1
void FASTCALL call_13_1_14_1(SH2_struct* sh)
{
  SH2movli(sh, 1, (14<<4)+1);
}
#define init_13_1_14_2
void FASTCALL call_13_1_14_2(SH2_struct* sh)
{
  SH2movli(sh, 1, (14<<4)+2);
}
#define init_13_1_14_3
void FASTCALL call_13_1_14_3(SH2_struct* sh)
{
  SH2movli(sh, 1, (14<<4)+3);
}
#define init_13_1_14_4
void FASTCALL call_13_1_14_4(SH2_struct* sh)
{
  SH2movli(sh, 1, (14<<4)+4);
}
#define init_13_1_14_5
void FASTCALL call_13_1_14_5(SH2_struct* sh)
{
  SH2movli(sh, 1, (14<<4)+5);
}
#define init_13_1_14_6
void FASTCALL call_13_1_14_6(SH2_struct* sh)
{
  SH2movli(sh, 1, (14<<4)+6);
}
#define init_13_1_14_7
void FASTCALL call_13_1_14_7(SH2_struct* sh)
{
  SH2movli(sh, 1, (14<<4)+7);
}
#define init_13_1_14_8
void FASTCALL call_13_1_14_8(SH2_struct* sh)
{
  SH2movli(sh, 1, (14<<4)+8);
}
#define init_13_1_14_9
void FASTCALL call_13_1_14_9(SH2_struct* sh)
{
  SH2movli(sh, 1, (14<<4)+9);
}
#define init_13_1_14_10
void FASTCALL call_13_1_14_10(SH2_struct* sh)
{
  SH2movli(sh, 1, (14<<4)+10);
}
#define init_13_1_14_11
void FASTCALL call_13_1_14_11(SH2_struct* sh)
{
  SH2movli(sh, 1, (14<<4)+11);
}
#define init_13_1_14_12
void FASTCALL call_13_1_14_12(SH2_struct* sh)
{
  SH2movli(sh, 1, (14<<4)+12);
}
#define init_13_1_14_13
void FASTCALL call_13_1_14_13(SH2_struct* sh)
{
  SH2movli(sh, 1, (14<<4)+13);
}
#define init_13_1_14_14
void FASTCALL call_13_1_14_14(SH2_struct* sh)
{
  SH2movli(sh, 1, (14<<4)+14);
}
#define init_13_1_14_15
void FASTCALL call_13_1_14_15(SH2_struct* sh)
{
  SH2movli(sh, 1, (14<<4)+15);
}
#define init_13_1_15_0
void FASTCALL call_13_1_15_0(SH2_struct* sh)
{
  SH2movli(sh, 1, (15<<4)+0);
}
#define init_13_1_15_1
void FASTCALL call_13_1_15_1(SH2_struct* sh)
{
  SH2movli(sh, 1, (15<<4)+1);
}
#define init_13_1_15_2
void FASTCALL call_13_1_15_2(SH2_struct* sh)
{
  SH2movli(sh, 1, (15<<4)+2);
}
#define init_13_1_15_3
void FASTCALL call_13_1_15_3(SH2_struct* sh)
{
  SH2movli(sh, 1, (15<<4)+3);
}
#define init_13_1_15_4
void FASTCALL call_13_1_15_4(SH2_struct* sh)
{
  SH2movli(sh, 1, (15<<4)+4);
}
#define init_13_1_15_5
void FASTCALL call_13_1_15_5(SH2_struct* sh)
{
  SH2movli(sh, 1, (15<<4)+5);
}
#define init_13_1_15_6
void FASTCALL call_13_1_15_6(SH2_struct* sh)
{
  SH2movli(sh, 1, (15<<4)+6);
}
#define init_13_1_15_7
void FASTCALL call_13_1_15_7(SH2_struct* sh)
{
  SH2movli(sh, 1, (15<<4)+7);
}
#define init_13_1_15_8
void FASTCALL call_13_1_15_8(SH2_struct* sh)
{
  SH2movli(sh, 1, (15<<4)+8);
}
#define init_13_1_15_9
void FASTCALL call_13_1_15_9(SH2_struct* sh)
{
  SH2movli(sh, 1, (15<<4)+9);
}
#define init_13_1_15_10
void FASTCALL call_13_1_15_10(SH2_struct* sh)
{
  SH2movli(sh, 1, (15<<4)+10);
}
#define init_13_1_15_11
void FASTCALL call_13_1_15_11(SH2_struct* sh)
{
  SH2movli(sh, 1, (15<<4)+11);
}
#define init_13_1_15_12
void FASTCALL call_13_1_15_12(SH2_struct* sh)
{
  SH2movli(sh, 1, (15<<4)+12);
}
#define init_13_1_15_13
void FASTCALL call_13_1_15_13(SH2_struct* sh)
{
  SH2movli(sh, 1, (15<<4)+13);
}
#define init_13_1_15_14
void FASTCALL call_13_1_15_14(SH2_struct* sh)
{
  SH2movli(sh, 1, (15<<4)+14);
}
#define init_13_1_15_15
void FASTCALL call_13_1_15_15(SH2_struct* sh)
{
  SH2movli(sh, 1, (15<<4)+15);
}
#define init_13_2_0_0
void FASTCALL call_13_2_0_0(SH2_struct* sh)
{
  SH2movli(sh, 2, (0<<4)+0);
}
#define init_13_2_0_1
void FASTCALL call_13_2_0_1(SH2_struct* sh)
{
  SH2movli(sh, 2, (0<<4)+1);
}
#define init_13_2_0_2
void FASTCALL call_13_2_0_2(SH2_struct* sh)
{
  SH2movli(sh, 2, (0<<4)+2);
}
#define init_13_2_0_3
void FASTCALL call_13_2_0_3(SH2_struct* sh)
{
  SH2movli(sh, 2, (0<<4)+3);
}
#define init_13_2_0_4
void FASTCALL call_13_2_0_4(SH2_struct* sh)
{
  SH2movli(sh, 2, (0<<4)+4);
}
#define init_13_2_0_5
void FASTCALL call_13_2_0_5(SH2_struct* sh)
{
  SH2movli(sh, 2, (0<<4)+5);
}
#define init_13_2_0_6
void FASTCALL call_13_2_0_6(SH2_struct* sh)
{
  SH2movli(sh, 2, (0<<4)+6);
}
#define init_13_2_0_7
void FASTCALL call_13_2_0_7(SH2_struct* sh)
{
  SH2movli(sh, 2, (0<<4)+7);
}
#define init_13_2_0_8
void FASTCALL call_13_2_0_8(SH2_struct* sh)
{
  SH2movli(sh, 2, (0<<4)+8);
}
#define init_13_2_0_9
void FASTCALL call_13_2_0_9(SH2_struct* sh)
{
  SH2movli(sh, 2, (0<<4)+9);
}
#define init_13_2_0_10
void FASTCALL call_13_2_0_10(SH2_struct* sh)
{
  SH2movli(sh, 2, (0<<4)+10);
}
#define init_13_2_0_11
void FASTCALL call_13_2_0_11(SH2_struct* sh)
{
  SH2movli(sh, 2, (0<<4)+11);
}
#define init_13_2_0_12
void FASTCALL call_13_2_0_12(SH2_struct* sh)
{
  SH2movli(sh, 2, (0<<4)+12);
}
#define init_13_2_0_13
void FASTCALL call_13_2_0_13(SH2_struct* sh)
{
  SH2movli(sh, 2, (0<<4)+13);
}
#define init_13_2_0_14
void FASTCALL call_13_2_0_14(SH2_struct* sh)
{
  SH2movli(sh, 2, (0<<4)+14);
}
#define init_13_2_0_15
void FASTCALL call_13_2_0_15(SH2_struct* sh)
{
  SH2movli(sh, 2, (0<<4)+15);
}
#define init_13_2_1_0
void FASTCALL call_13_2_1_0(SH2_struct* sh)
{
  SH2movli(sh, 2, (1<<4)+0);
}
#define init_13_2_1_1
void FASTCALL call_13_2_1_1(SH2_struct* sh)
{
  SH2movli(sh, 2, (1<<4)+1);
}
#define init_13_2_1_2
void FASTCALL call_13_2_1_2(SH2_struct* sh)
{
  SH2movli(sh, 2, (1<<4)+2);
}
#define init_13_2_1_3
void FASTCALL call_13_2_1_3(SH2_struct* sh)
{
  SH2movli(sh, 2, (1<<4)+3);
}
#define init_13_2_1_4
void FASTCALL call_13_2_1_4(SH2_struct* sh)
{
  SH2movli(sh, 2, (1<<4)+4);
}
#define init_13_2_1_5
void FASTCALL call_13_2_1_5(SH2_struct* sh)
{
  SH2movli(sh, 2, (1<<4)+5);
}
#define init_13_2_1_6
void FASTCALL call_13_2_1_6(SH2_struct* sh)
{
  SH2movli(sh, 2, (1<<4)+6);
}
#define init_13_2_1_7
void FASTCALL call_13_2_1_7(SH2_struct* sh)
{
  SH2movli(sh, 2, (1<<4)+7);
}
#define init_13_2_1_8
void FASTCALL call_13_2_1_8(SH2_struct* sh)
{
  SH2movli(sh, 2, (1<<4)+8);
}
#define init_13_2_1_9
void FASTCALL call_13_2_1_9(SH2_struct* sh)
{
  SH2movli(sh, 2, (1<<4)+9);
}
#define init_13_2_1_10
void FASTCALL call_13_2_1_10(SH2_struct* sh)
{
  SH2movli(sh, 2, (1<<4)+10);
}
#define init_13_2_1_11
void FASTCALL call_13_2_1_11(SH2_struct* sh)
{
  SH2movli(sh, 2, (1<<4)+11);
}
#define init_13_2_1_12
void FASTCALL call_13_2_1_12(SH2_struct* sh)
{
  SH2movli(sh, 2, (1<<4)+12);
}
#define init_13_2_1_13
void FASTCALL call_13_2_1_13(SH2_struct* sh)
{
  SH2movli(sh, 2, (1<<4)+13);
}
#define init_13_2_1_14
void FASTCALL call_13_2_1_14(SH2_struct* sh)
{
  SH2movli(sh, 2, (1<<4)+14);
}
#define init_13_2_1_15
void FASTCALL call_13_2_1_15(SH2_struct* sh)
{
  SH2movli(sh, 2, (1<<4)+15);
}
#define init_13_2_2_0
void FASTCALL call_13_2_2_0(SH2_struct* sh)
{
  SH2movli(sh, 2, (2<<4)+0);
}
#define init_13_2_2_1
void FASTCALL call_13_2_2_1(SH2_struct* sh)
{
  SH2movli(sh, 2, (2<<4)+1);
}
#define init_13_2_2_2
void FASTCALL call_13_2_2_2(SH2_struct* sh)
{
  SH2movli(sh, 2, (2<<4)+2);
}
#define init_13_2_2_3
void FASTCALL call_13_2_2_3(SH2_struct* sh)
{
  SH2movli(sh, 2, (2<<4)+3);
}
#define init_13_2_2_4
void FASTCALL call_13_2_2_4(SH2_struct* sh)
{
  SH2movli(sh, 2, (2<<4)+4);
}
#define init_13_2_2_5
void FASTCALL call_13_2_2_5(SH2_struct* sh)
{
  SH2movli(sh, 2, (2<<4)+5);
}
#define init_13_2_2_6
void FASTCALL call_13_2_2_6(SH2_struct* sh)
{
  SH2movli(sh, 2, (2<<4)+6);
}
#define init_13_2_2_7
void FASTCALL call_13_2_2_7(SH2_struct* sh)
{
  SH2movli(sh, 2, (2<<4)+7);
}
#define init_13_2_2_8
void FASTCALL call_13_2_2_8(SH2_struct* sh)
{
  SH2movli(sh, 2, (2<<4)+8);
}
#define init_13_2_2_9
void FASTCALL call_13_2_2_9(SH2_struct* sh)
{
  SH2movli(sh, 2, (2<<4)+9);
}
#define init_13_2_2_10
void FASTCALL call_13_2_2_10(SH2_struct* sh)
{
  SH2movli(sh, 2, (2<<4)+10);
}
#define init_13_2_2_11
void FASTCALL call_13_2_2_11(SH2_struct* sh)
{
  SH2movli(sh, 2, (2<<4)+11);
}
#define init_13_2_2_12
void FASTCALL call_13_2_2_12(SH2_struct* sh)
{
  SH2movli(sh, 2, (2<<4)+12);
}
#define init_13_2_2_13
void FASTCALL call_13_2_2_13(SH2_struct* sh)
{
  SH2movli(sh, 2, (2<<4)+13);
}
#define init_13_2_2_14
void FASTCALL call_13_2_2_14(SH2_struct* sh)
{
  SH2movli(sh, 2, (2<<4)+14);
}
#define init_13_2_2_15
void FASTCALL call_13_2_2_15(SH2_struct* sh)
{
  SH2movli(sh, 2, (2<<4)+15);
}
#define init_13_2_3_0
void FASTCALL call_13_2_3_0(SH2_struct* sh)
{
  SH2movli(sh, 2, (3<<4)+0);
}
#define init_13_2_3_1
void FASTCALL call_13_2_3_1(SH2_struct* sh)
{
  SH2movli(sh, 2, (3<<4)+1);
}
#define init_13_2_3_2
void FASTCALL call_13_2_3_2(SH2_struct* sh)
{
  SH2movli(sh, 2, (3<<4)+2);
}
#define init_13_2_3_3
void FASTCALL call_13_2_3_3(SH2_struct* sh)
{
  SH2movli(sh, 2, (3<<4)+3);
}
#define init_13_2_3_4
void FASTCALL call_13_2_3_4(SH2_struct* sh)
{
  SH2movli(sh, 2, (3<<4)+4);
}
#define init_13_2_3_5
void FASTCALL call_13_2_3_5(SH2_struct* sh)
{
  SH2movli(sh, 2, (3<<4)+5);
}
#define init_13_2_3_6
void FASTCALL call_13_2_3_6(SH2_struct* sh)
{
  SH2movli(sh, 2, (3<<4)+6);
}
#define init_13_2_3_7
void FASTCALL call_13_2_3_7(SH2_struct* sh)
{
  SH2movli(sh, 2, (3<<4)+7);
}
#define init_13_2_3_8
void FASTCALL call_13_2_3_8(SH2_struct* sh)
{
  SH2movli(sh, 2, (3<<4)+8);
}
#define init_13_2_3_9
void FASTCALL call_13_2_3_9(SH2_struct* sh)
{
  SH2movli(sh, 2, (3<<4)+9);
}
#define init_13_2_3_10
void FASTCALL call_13_2_3_10(SH2_struct* sh)
{
  SH2movli(sh, 2, (3<<4)+10);
}
#define init_13_2_3_11
void FASTCALL call_13_2_3_11(SH2_struct* sh)
{
  SH2movli(sh, 2, (3<<4)+11);
}
#define init_13_2_3_12
void FASTCALL call_13_2_3_12(SH2_struct* sh)
{
  SH2movli(sh, 2, (3<<4)+12);
}
#define init_13_2_3_13
void FASTCALL call_13_2_3_13(SH2_struct* sh)
{
  SH2movli(sh, 2, (3<<4)+13);
}
#define init_13_2_3_14
void FASTCALL call_13_2_3_14(SH2_struct* sh)
{
  SH2movli(sh, 2, (3<<4)+14);
}
#define init_13_2_3_15
void FASTCALL call_13_2_3_15(SH2_struct* sh)
{
  SH2movli(sh, 2, (3<<4)+15);
}
#define init_13_2_4_0
void FASTCALL call_13_2_4_0(SH2_struct* sh)
{
  SH2movli(sh, 2, (4<<4)+0);
}
#define init_13_2_4_1
void FASTCALL call_13_2_4_1(SH2_struct* sh)
{
  SH2movli(sh, 2, (4<<4)+1);
}
#define init_13_2_4_2
void FASTCALL call_13_2_4_2(SH2_struct* sh)
{
  SH2movli(sh, 2, (4<<4)+2);
}
#define init_13_2_4_3
void FASTCALL call_13_2_4_3(SH2_struct* sh)
{
  SH2movli(sh, 2, (4<<4)+3);
}
#define init_13_2_4_4
void FASTCALL call_13_2_4_4(SH2_struct* sh)
{
  SH2movli(sh, 2, (4<<4)+4);
}
#define init_13_2_4_5
void FASTCALL call_13_2_4_5(SH2_struct* sh)
{
  SH2movli(sh, 2, (4<<4)+5);
}
#define init_13_2_4_6
void FASTCALL call_13_2_4_6(SH2_struct* sh)
{
  SH2movli(sh, 2, (4<<4)+6);
}
#define init_13_2_4_7
void FASTCALL call_13_2_4_7(SH2_struct* sh)
{
  SH2movli(sh, 2, (4<<4)+7);
}
#define init_13_2_4_8
void FASTCALL call_13_2_4_8(SH2_struct* sh)
{
  SH2movli(sh, 2, (4<<4)+8);
}
#define init_13_2_4_9
void FASTCALL call_13_2_4_9(SH2_struct* sh)
{
  SH2movli(sh, 2, (4<<4)+9);
}
#define init_13_2_4_10
void FASTCALL call_13_2_4_10(SH2_struct* sh)
{
  SH2movli(sh, 2, (4<<4)+10);
}
#define init_13_2_4_11
void FASTCALL call_13_2_4_11(SH2_struct* sh)
{
  SH2movli(sh, 2, (4<<4)+11);
}
#define init_13_2_4_12
void FASTCALL call_13_2_4_12(SH2_struct* sh)
{
  SH2movli(sh, 2, (4<<4)+12);
}
#define init_13_2_4_13
void FASTCALL call_13_2_4_13(SH2_struct* sh)
{
  SH2movli(sh, 2, (4<<4)+13);
}
#define init_13_2_4_14
void FASTCALL call_13_2_4_14(SH2_struct* sh)
{
  SH2movli(sh, 2, (4<<4)+14);
}
#define init_13_2_4_15
void FASTCALL call_13_2_4_15(SH2_struct* sh)
{
  SH2movli(sh, 2, (4<<4)+15);
}
#define init_13_2_5_0
void FASTCALL call_13_2_5_0(SH2_struct* sh)
{
  SH2movli(sh, 2, (5<<4)+0);
}
#define init_13_2_5_1
void FASTCALL call_13_2_5_1(SH2_struct* sh)
{
  SH2movli(sh, 2, (5<<4)+1);
}
#define init_13_2_5_2
void FASTCALL call_13_2_5_2(SH2_struct* sh)
{
  SH2movli(sh, 2, (5<<4)+2);
}
#define init_13_2_5_3
void FASTCALL call_13_2_5_3(SH2_struct* sh)
{
  SH2movli(sh, 2, (5<<4)+3);
}
#define init_13_2_5_4
void FASTCALL call_13_2_5_4(SH2_struct* sh)
{
  SH2movli(sh, 2, (5<<4)+4);
}
#define init_13_2_5_5
void FASTCALL call_13_2_5_5(SH2_struct* sh)
{
  SH2movli(sh, 2, (5<<4)+5);
}
#define init_13_2_5_6
void FASTCALL call_13_2_5_6(SH2_struct* sh)
{
  SH2movli(sh, 2, (5<<4)+6);
}
#define init_13_2_5_7
void FASTCALL call_13_2_5_7(SH2_struct* sh)
{
  SH2movli(sh, 2, (5<<4)+7);
}
#define init_13_2_5_8
void FASTCALL call_13_2_5_8(SH2_struct* sh)
{
  SH2movli(sh, 2, (5<<4)+8);
}
#define init_13_2_5_9
void FASTCALL call_13_2_5_9(SH2_struct* sh)
{
  SH2movli(sh, 2, (5<<4)+9);
}
#define init_13_2_5_10
void FASTCALL call_13_2_5_10(SH2_struct* sh)
{
  SH2movli(sh, 2, (5<<4)+10);
}
#define init_13_2_5_11
void FASTCALL call_13_2_5_11(SH2_struct* sh)
{
  SH2movli(sh, 2, (5<<4)+11);
}
#define init_13_2_5_12
void FASTCALL call_13_2_5_12(SH2_struct* sh)
{
  SH2movli(sh, 2, (5<<4)+12);
}
#define init_13_2_5_13
void FASTCALL call_13_2_5_13(SH2_struct* sh)
{
  SH2movli(sh, 2, (5<<4)+13);
}
#define init_13_2_5_14
void FASTCALL call_13_2_5_14(SH2_struct* sh)
{
  SH2movli(sh, 2, (5<<4)+14);
}
#define init_13_2_5_15
void FASTCALL call_13_2_5_15(SH2_struct* sh)
{
  SH2movli(sh, 2, (5<<4)+15);
}
#define init_13_2_6_0
void FASTCALL call_13_2_6_0(SH2_struct* sh)
{
  SH2movli(sh, 2, (6<<4)+0);
}
#define init_13_2_6_1
void FASTCALL call_13_2_6_1(SH2_struct* sh)
{
  SH2movli(sh, 2, (6<<4)+1);
}
#define init_13_2_6_2
void FASTCALL call_13_2_6_2(SH2_struct* sh)
{
  SH2movli(sh, 2, (6<<4)+2);
}
#define init_13_2_6_3
void FASTCALL call_13_2_6_3(SH2_struct* sh)
{
  SH2movli(sh, 2, (6<<4)+3);
}
#define init_13_2_6_4
void FASTCALL call_13_2_6_4(SH2_struct* sh)
{
  SH2movli(sh, 2, (6<<4)+4);
}
#define init_13_2_6_5
void FASTCALL call_13_2_6_5(SH2_struct* sh)
{
  SH2movli(sh, 2, (6<<4)+5);
}
#define init_13_2_6_6
void FASTCALL call_13_2_6_6(SH2_struct* sh)
{
  SH2movli(sh, 2, (6<<4)+6);
}
#define init_13_2_6_7
void FASTCALL call_13_2_6_7(SH2_struct* sh)
{
  SH2movli(sh, 2, (6<<4)+7);
}
#define init_13_2_6_8
void FASTCALL call_13_2_6_8(SH2_struct* sh)
{
  SH2movli(sh, 2, (6<<4)+8);
}
#define init_13_2_6_9
void FASTCALL call_13_2_6_9(SH2_struct* sh)
{
  SH2movli(sh, 2, (6<<4)+9);
}
#define init_13_2_6_10
void FASTCALL call_13_2_6_10(SH2_struct* sh)
{
  SH2movli(sh, 2, (6<<4)+10);
}
#define init_13_2_6_11
void FASTCALL call_13_2_6_11(SH2_struct* sh)
{
  SH2movli(sh, 2, (6<<4)+11);
}
#define init_13_2_6_12
void FASTCALL call_13_2_6_12(SH2_struct* sh)
{
  SH2movli(sh, 2, (6<<4)+12);
}
#define init_13_2_6_13
void FASTCALL call_13_2_6_13(SH2_struct* sh)
{
  SH2movli(sh, 2, (6<<4)+13);
}
#define init_13_2_6_14
void FASTCALL call_13_2_6_14(SH2_struct* sh)
{
  SH2movli(sh, 2, (6<<4)+14);
}
#define init_13_2_6_15
void FASTCALL call_13_2_6_15(SH2_struct* sh)
{
  SH2movli(sh, 2, (6<<4)+15);
}
#define init_13_2_7_0
void FASTCALL call_13_2_7_0(SH2_struct* sh)
{
  SH2movli(sh, 2, (7<<4)+0);
}
#define init_13_2_7_1
void FASTCALL call_13_2_7_1(SH2_struct* sh)
{
  SH2movli(sh, 2, (7<<4)+1);
}
#define init_13_2_7_2
void FASTCALL call_13_2_7_2(SH2_struct* sh)
{
  SH2movli(sh, 2, (7<<4)+2);
}
#define init_13_2_7_3
void FASTCALL call_13_2_7_3(SH2_struct* sh)
{
  SH2movli(sh, 2, (7<<4)+3);
}
#define init_13_2_7_4
void FASTCALL call_13_2_7_4(SH2_struct* sh)
{
  SH2movli(sh, 2, (7<<4)+4);
}
#define init_13_2_7_5
void FASTCALL call_13_2_7_5(SH2_struct* sh)
{
  SH2movli(sh, 2, (7<<4)+5);
}
#define init_13_2_7_6
void FASTCALL call_13_2_7_6(SH2_struct* sh)
{
  SH2movli(sh, 2, (7<<4)+6);
}
#define init_13_2_7_7
void FASTCALL call_13_2_7_7(SH2_struct* sh)
{
  SH2movli(sh, 2, (7<<4)+7);
}
#define init_13_2_7_8
void FASTCALL call_13_2_7_8(SH2_struct* sh)
{
  SH2movli(sh, 2, (7<<4)+8);
}
#define init_13_2_7_9
void FASTCALL call_13_2_7_9(SH2_struct* sh)
{
  SH2movli(sh, 2, (7<<4)+9);
}
#define init_13_2_7_10
void FASTCALL call_13_2_7_10(SH2_struct* sh)
{
  SH2movli(sh, 2, (7<<4)+10);
}
#define init_13_2_7_11
void FASTCALL call_13_2_7_11(SH2_struct* sh)
{
  SH2movli(sh, 2, (7<<4)+11);
}
#define init_13_2_7_12
void FASTCALL call_13_2_7_12(SH2_struct* sh)
{
  SH2movli(sh, 2, (7<<4)+12);
}
#define init_13_2_7_13
void FASTCALL call_13_2_7_13(SH2_struct* sh)
{
  SH2movli(sh, 2, (7<<4)+13);
}
#define init_13_2_7_14
void FASTCALL call_13_2_7_14(SH2_struct* sh)
{
  SH2movli(sh, 2, (7<<4)+14);
}
#define init_13_2_7_15
void FASTCALL call_13_2_7_15(SH2_struct* sh)
{
  SH2movli(sh, 2, (7<<4)+15);
}
#define init_13_2_8_0
void FASTCALL call_13_2_8_0(SH2_struct* sh)
{
  SH2movli(sh, 2, (8<<4)+0);
}
#define init_13_2_8_1
void FASTCALL call_13_2_8_1(SH2_struct* sh)
{
  SH2movli(sh, 2, (8<<4)+1);
}
#define init_13_2_8_2
void FASTCALL call_13_2_8_2(SH2_struct* sh)
{
  SH2movli(sh, 2, (8<<4)+2);
}
#define init_13_2_8_3
void FASTCALL call_13_2_8_3(SH2_struct* sh)
{
  SH2movli(sh, 2, (8<<4)+3);
}
#define init_13_2_8_4
void FASTCALL call_13_2_8_4(SH2_struct* sh)
{
  SH2movli(sh, 2, (8<<4)+4);
}
#define init_13_2_8_5
void FASTCALL call_13_2_8_5(SH2_struct* sh)
{
  SH2movli(sh, 2, (8<<4)+5);
}
#define init_13_2_8_6
void FASTCALL call_13_2_8_6(SH2_struct* sh)
{
  SH2movli(sh, 2, (8<<4)+6);
}
#define init_13_2_8_7
void FASTCALL call_13_2_8_7(SH2_struct* sh)
{
  SH2movli(sh, 2, (8<<4)+7);
}
#define init_13_2_8_8
void FASTCALL call_13_2_8_8(SH2_struct* sh)
{
  SH2movli(sh, 2, (8<<4)+8);
}
#define init_13_2_8_9
void FASTCALL call_13_2_8_9(SH2_struct* sh)
{
  SH2movli(sh, 2, (8<<4)+9);
}
#define init_13_2_8_10
void FASTCALL call_13_2_8_10(SH2_struct* sh)
{
  SH2movli(sh, 2, (8<<4)+10);
}
#define init_13_2_8_11
void FASTCALL call_13_2_8_11(SH2_struct* sh)
{
  SH2movli(sh, 2, (8<<4)+11);
}
#define init_13_2_8_12
void FASTCALL call_13_2_8_12(SH2_struct* sh)
{
  SH2movli(sh, 2, (8<<4)+12);
}
#define init_13_2_8_13
void FASTCALL call_13_2_8_13(SH2_struct* sh)
{
  SH2movli(sh, 2, (8<<4)+13);
}
#define init_13_2_8_14
void FASTCALL call_13_2_8_14(SH2_struct* sh)
{
  SH2movli(sh, 2, (8<<4)+14);
}
#define init_13_2_8_15
void FASTCALL call_13_2_8_15(SH2_struct* sh)
{
  SH2movli(sh, 2, (8<<4)+15);
}
#define init_13_2_9_0
void FASTCALL call_13_2_9_0(SH2_struct* sh)
{
  SH2movli(sh, 2, (9<<4)+0);
}
#define init_13_2_9_1
void FASTCALL call_13_2_9_1(SH2_struct* sh)
{
  SH2movli(sh, 2, (9<<4)+1);
}
#define init_13_2_9_2
void FASTCALL call_13_2_9_2(SH2_struct* sh)
{
  SH2movli(sh, 2, (9<<4)+2);
}
#define init_13_2_9_3
void FASTCALL call_13_2_9_3(SH2_struct* sh)
{
  SH2movli(sh, 2, (9<<4)+3);
}
#define init_13_2_9_4
void FASTCALL call_13_2_9_4(SH2_struct* sh)
{
  SH2movli(sh, 2, (9<<4)+4);
}
#define init_13_2_9_5
void FASTCALL call_13_2_9_5(SH2_struct* sh)
{
  SH2movli(sh, 2, (9<<4)+5);
}
#define init_13_2_9_6
void FASTCALL call_13_2_9_6(SH2_struct* sh)
{
  SH2movli(sh, 2, (9<<4)+6);
}
#define init_13_2_9_7
void FASTCALL call_13_2_9_7(SH2_struct* sh)
{
  SH2movli(sh, 2, (9<<4)+7);
}
#define init_13_2_9_8
void FASTCALL call_13_2_9_8(SH2_struct* sh)
{
  SH2movli(sh, 2, (9<<4)+8);
}
#define init_13_2_9_9
void FASTCALL call_13_2_9_9(SH2_struct* sh)
{
  SH2movli(sh, 2, (9<<4)+9);
}
#define init_13_2_9_10
void FASTCALL call_13_2_9_10(SH2_struct* sh)
{
  SH2movli(sh, 2, (9<<4)+10);
}
#define init_13_2_9_11
void FASTCALL call_13_2_9_11(SH2_struct* sh)
{
  SH2movli(sh, 2, (9<<4)+11);
}
#define init_13_2_9_12
void FASTCALL call_13_2_9_12(SH2_struct* sh)
{
  SH2movli(sh, 2, (9<<4)+12);
}
#define init_13_2_9_13
void FASTCALL call_13_2_9_13(SH2_struct* sh)
{
  SH2movli(sh, 2, (9<<4)+13);
}
#define init_13_2_9_14
void FASTCALL call_13_2_9_14(SH2_struct* sh)
{
  SH2movli(sh, 2, (9<<4)+14);
}
#define init_13_2_9_15
void FASTCALL call_13_2_9_15(SH2_struct* sh)
{
  SH2movli(sh, 2, (9<<4)+15);
}
#define init_13_2_10_0
void FASTCALL call_13_2_10_0(SH2_struct* sh)
{
  SH2movli(sh, 2, (10<<4)+0);
}
#define init_13_2_10_1
void FASTCALL call_13_2_10_1(SH2_struct* sh)
{
  SH2movli(sh, 2, (10<<4)+1);
}
#define init_13_2_10_2
void FASTCALL call_13_2_10_2(SH2_struct* sh)
{
  SH2movli(sh, 2, (10<<4)+2);
}
#define init_13_2_10_3
void FASTCALL call_13_2_10_3(SH2_struct* sh)
{
  SH2movli(sh, 2, (10<<4)+3);
}
#define init_13_2_10_4
void FASTCALL call_13_2_10_4(SH2_struct* sh)
{
  SH2movli(sh, 2, (10<<4)+4);
}
#define init_13_2_10_5
void FASTCALL call_13_2_10_5(SH2_struct* sh)
{
  SH2movli(sh, 2, (10<<4)+5);
}
#define init_13_2_10_6
void FASTCALL call_13_2_10_6(SH2_struct* sh)
{
  SH2movli(sh, 2, (10<<4)+6);
}
#define init_13_2_10_7
void FASTCALL call_13_2_10_7(SH2_struct* sh)
{
  SH2movli(sh, 2, (10<<4)+7);
}
#define init_13_2_10_8
void FASTCALL call_13_2_10_8(SH2_struct* sh)
{
  SH2movli(sh, 2, (10<<4)+8);
}
#define init_13_2_10_9
void FASTCALL call_13_2_10_9(SH2_struct* sh)
{
  SH2movli(sh, 2, (10<<4)+9);
}
#define init_13_2_10_10
void FASTCALL call_13_2_10_10(SH2_struct* sh)
{
  SH2movli(sh, 2, (10<<4)+10);
}
#define init_13_2_10_11
void FASTCALL call_13_2_10_11(SH2_struct* sh)
{
  SH2movli(sh, 2, (10<<4)+11);
}
#define init_13_2_10_12
void FASTCALL call_13_2_10_12(SH2_struct* sh)
{
  SH2movli(sh, 2, (10<<4)+12);
}
#define init_13_2_10_13
void FASTCALL call_13_2_10_13(SH2_struct* sh)
{
  SH2movli(sh, 2, (10<<4)+13);
}
#define init_13_2_10_14
void FASTCALL call_13_2_10_14(SH2_struct* sh)
{
  SH2movli(sh, 2, (10<<4)+14);
}
#define init_13_2_10_15
void FASTCALL call_13_2_10_15(SH2_struct* sh)
{
  SH2movli(sh, 2, (10<<4)+15);
}
#define init_13_2_11_0
void FASTCALL call_13_2_11_0(SH2_struct* sh)
{
  SH2movli(sh, 2, (11<<4)+0);
}
#define init_13_2_11_1
void FASTCALL call_13_2_11_1(SH2_struct* sh)
{
  SH2movli(sh, 2, (11<<4)+1);
}
#define init_13_2_11_2
void FASTCALL call_13_2_11_2(SH2_struct* sh)
{
  SH2movli(sh, 2, (11<<4)+2);
}
#define init_13_2_11_3
void FASTCALL call_13_2_11_3(SH2_struct* sh)
{
  SH2movli(sh, 2, (11<<4)+3);
}
#define init_13_2_11_4
void FASTCALL call_13_2_11_4(SH2_struct* sh)
{
  SH2movli(sh, 2, (11<<4)+4);
}
#define init_13_2_11_5
void FASTCALL call_13_2_11_5(SH2_struct* sh)
{
  SH2movli(sh, 2, (11<<4)+5);
}
#define init_13_2_11_6
void FASTCALL call_13_2_11_6(SH2_struct* sh)
{
  SH2movli(sh, 2, (11<<4)+6);
}
#define init_13_2_11_7
void FASTCALL call_13_2_11_7(SH2_struct* sh)
{
  SH2movli(sh, 2, (11<<4)+7);
}
#define init_13_2_11_8
void FASTCALL call_13_2_11_8(SH2_struct* sh)
{
  SH2movli(sh, 2, (11<<4)+8);
}
#define init_13_2_11_9
void FASTCALL call_13_2_11_9(SH2_struct* sh)
{
  SH2movli(sh, 2, (11<<4)+9);
}
#define init_13_2_11_10
void FASTCALL call_13_2_11_10(SH2_struct* sh)
{
  SH2movli(sh, 2, (11<<4)+10);
}
#define init_13_2_11_11
void FASTCALL call_13_2_11_11(SH2_struct* sh)
{
  SH2movli(sh, 2, (11<<4)+11);
}
#define init_13_2_11_12
void FASTCALL call_13_2_11_12(SH2_struct* sh)
{
  SH2movli(sh, 2, (11<<4)+12);
}
#define init_13_2_11_13
void FASTCALL call_13_2_11_13(SH2_struct* sh)
{
  SH2movli(sh, 2, (11<<4)+13);
}
#define init_13_2_11_14
void FASTCALL call_13_2_11_14(SH2_struct* sh)
{
  SH2movli(sh, 2, (11<<4)+14);
}
#define init_13_2_11_15
void FASTCALL call_13_2_11_15(SH2_struct* sh)
{
  SH2movli(sh, 2, (11<<4)+15);
}
#define init_13_2_12_0
void FASTCALL call_13_2_12_0(SH2_struct* sh)
{
  SH2movli(sh, 2, (12<<4)+0);
}
#define init_13_2_12_1
void FASTCALL call_13_2_12_1(SH2_struct* sh)
{
  SH2movli(sh, 2, (12<<4)+1);
}
#define init_13_2_12_2
void FASTCALL call_13_2_12_2(SH2_struct* sh)
{
  SH2movli(sh, 2, (12<<4)+2);
}
#define init_13_2_12_3
void FASTCALL call_13_2_12_3(SH2_struct* sh)
{
  SH2movli(sh, 2, (12<<4)+3);
}
#define init_13_2_12_4
void FASTCALL call_13_2_12_4(SH2_struct* sh)
{
  SH2movli(sh, 2, (12<<4)+4);
}
#define init_13_2_12_5
void FASTCALL call_13_2_12_5(SH2_struct* sh)
{
  SH2movli(sh, 2, (12<<4)+5);
}
#define init_13_2_12_6
void FASTCALL call_13_2_12_6(SH2_struct* sh)
{
  SH2movli(sh, 2, (12<<4)+6);
}
#define init_13_2_12_7
void FASTCALL call_13_2_12_7(SH2_struct* sh)
{
  SH2movli(sh, 2, (12<<4)+7);
}
#define init_13_2_12_8
void FASTCALL call_13_2_12_8(SH2_struct* sh)
{
  SH2movli(sh, 2, (12<<4)+8);
}
#define init_13_2_12_9
void FASTCALL call_13_2_12_9(SH2_struct* sh)
{
  SH2movli(sh, 2, (12<<4)+9);
}
#define init_13_2_12_10
void FASTCALL call_13_2_12_10(SH2_struct* sh)
{
  SH2movli(sh, 2, (12<<4)+10);
}
#define init_13_2_12_11
void FASTCALL call_13_2_12_11(SH2_struct* sh)
{
  SH2movli(sh, 2, (12<<4)+11);
}
#define init_13_2_12_12
void FASTCALL call_13_2_12_12(SH2_struct* sh)
{
  SH2movli(sh, 2, (12<<4)+12);
}
#define init_13_2_12_13
void FASTCALL call_13_2_12_13(SH2_struct* sh)
{
  SH2movli(sh, 2, (12<<4)+13);
}
#define init_13_2_12_14
void FASTCALL call_13_2_12_14(SH2_struct* sh)
{
  SH2movli(sh, 2, (12<<4)+14);
}
#define init_13_2_12_15
void FASTCALL call_13_2_12_15(SH2_struct* sh)
{
  SH2movli(sh, 2, (12<<4)+15);
}
#define init_13_2_13_0
void FASTCALL call_13_2_13_0(SH2_struct* sh)
{
  SH2movli(sh, 2, (13<<4)+0);
}
#define init_13_2_13_1
void FASTCALL call_13_2_13_1(SH2_struct* sh)
{
  SH2movli(sh, 2, (13<<4)+1);
}
#define init_13_2_13_2
void FASTCALL call_13_2_13_2(SH2_struct* sh)
{
  SH2movli(sh, 2, (13<<4)+2);
}
#define init_13_2_13_3
void FASTCALL call_13_2_13_3(SH2_struct* sh)
{
  SH2movli(sh, 2, (13<<4)+3);
}
#define init_13_2_13_4
void FASTCALL call_13_2_13_4(SH2_struct* sh)
{
  SH2movli(sh, 2, (13<<4)+4);
}
#define init_13_2_13_5
void FASTCALL call_13_2_13_5(SH2_struct* sh)
{
  SH2movli(sh, 2, (13<<4)+5);
}
#define init_13_2_13_6
void FASTCALL call_13_2_13_6(SH2_struct* sh)
{
  SH2movli(sh, 2, (13<<4)+6);
}
#define init_13_2_13_7
void FASTCALL call_13_2_13_7(SH2_struct* sh)
{
  SH2movli(sh, 2, (13<<4)+7);
}
#define init_13_2_13_8
void FASTCALL call_13_2_13_8(SH2_struct* sh)
{
  SH2movli(sh, 2, (13<<4)+8);
}
#define init_13_2_13_9
void FASTCALL call_13_2_13_9(SH2_struct* sh)
{
  SH2movli(sh, 2, (13<<4)+9);
}
#define init_13_2_13_10
void FASTCALL call_13_2_13_10(SH2_struct* sh)
{
  SH2movli(sh, 2, (13<<4)+10);
}
#define init_13_2_13_11
void FASTCALL call_13_2_13_11(SH2_struct* sh)
{
  SH2movli(sh, 2, (13<<4)+11);
}
#define init_13_2_13_12
void FASTCALL call_13_2_13_12(SH2_struct* sh)
{
  SH2movli(sh, 2, (13<<4)+12);
}
#define init_13_2_13_13
void FASTCALL call_13_2_13_13(SH2_struct* sh)
{
  SH2movli(sh, 2, (13<<4)+13);
}
#define init_13_2_13_14
void FASTCALL call_13_2_13_14(SH2_struct* sh)
{
  SH2movli(sh, 2, (13<<4)+14);
}
#define init_13_2_13_15
void FASTCALL call_13_2_13_15(SH2_struct* sh)
{
  SH2movli(sh, 2, (13<<4)+15);
}
#define init_13_2_14_0
void FASTCALL call_13_2_14_0(SH2_struct* sh)
{
  SH2movli(sh, 2, (14<<4)+0);
}
#define init_13_2_14_1
void FASTCALL call_13_2_14_1(SH2_struct* sh)
{
  SH2movli(sh, 2, (14<<4)+1);
}
#define init_13_2_14_2
void FASTCALL call_13_2_14_2(SH2_struct* sh)
{
  SH2movli(sh, 2, (14<<4)+2);
}
#define init_13_2_14_3
void FASTCALL call_13_2_14_3(SH2_struct* sh)
{
  SH2movli(sh, 2, (14<<4)+3);
}
#define init_13_2_14_4
void FASTCALL call_13_2_14_4(SH2_struct* sh)
{
  SH2movli(sh, 2, (14<<4)+4);
}
#define init_13_2_14_5
void FASTCALL call_13_2_14_5(SH2_struct* sh)
{
  SH2movli(sh, 2, (14<<4)+5);
}
#define init_13_2_14_6
void FASTCALL call_13_2_14_6(SH2_struct* sh)
{
  SH2movli(sh, 2, (14<<4)+6);
}
#define init_13_2_14_7
void FASTCALL call_13_2_14_7(SH2_struct* sh)
{
  SH2movli(sh, 2, (14<<4)+7);
}
#define init_13_2_14_8
void FASTCALL call_13_2_14_8(SH2_struct* sh)
{
  SH2movli(sh, 2, (14<<4)+8);
}
#define init_13_2_14_9
void FASTCALL call_13_2_14_9(SH2_struct* sh)
{
  SH2movli(sh, 2, (14<<4)+9);
}
#define init_13_2_14_10
void FASTCALL call_13_2_14_10(SH2_struct* sh)
{
  SH2movli(sh, 2, (14<<4)+10);
}
#define init_13_2_14_11
void FASTCALL call_13_2_14_11(SH2_struct* sh)
{
  SH2movli(sh, 2, (14<<4)+11);
}
#define init_13_2_14_12
void FASTCALL call_13_2_14_12(SH2_struct* sh)
{
  SH2movli(sh, 2, (14<<4)+12);
}
#define init_13_2_14_13
void FASTCALL call_13_2_14_13(SH2_struct* sh)
{
  SH2movli(sh, 2, (14<<4)+13);
}
#define init_13_2_14_14
void FASTCALL call_13_2_14_14(SH2_struct* sh)
{
  SH2movli(sh, 2, (14<<4)+14);
}
#define init_13_2_14_15
void FASTCALL call_13_2_14_15(SH2_struct* sh)
{
  SH2movli(sh, 2, (14<<4)+15);
}
#define init_13_2_15_0
void FASTCALL call_13_2_15_0(SH2_struct* sh)
{
  SH2movli(sh, 2, (15<<4)+0);
}
#define init_13_2_15_1
void FASTCALL call_13_2_15_1(SH2_struct* sh)
{
  SH2movli(sh, 2, (15<<4)+1);
}
#define init_13_2_15_2
void FASTCALL call_13_2_15_2(SH2_struct* sh)
{
  SH2movli(sh, 2, (15<<4)+2);
}
#define init_13_2_15_3
void FASTCALL call_13_2_15_3(SH2_struct* sh)
{
  SH2movli(sh, 2, (15<<4)+3);
}
#define init_13_2_15_4
void FASTCALL call_13_2_15_4(SH2_struct* sh)
{
  SH2movli(sh, 2, (15<<4)+4);
}
#define init_13_2_15_5
void FASTCALL call_13_2_15_5(SH2_struct* sh)
{
  SH2movli(sh, 2, (15<<4)+5);
}
#define init_13_2_15_6
void FASTCALL call_13_2_15_6(SH2_struct* sh)
{
  SH2movli(sh, 2, (15<<4)+6);
}
#define init_13_2_15_7
void FASTCALL call_13_2_15_7(SH2_struct* sh)
{
  SH2movli(sh, 2, (15<<4)+7);
}
#define init_13_2_15_8
void FASTCALL call_13_2_15_8(SH2_struct* sh)
{
  SH2movli(sh, 2, (15<<4)+8);
}
#define init_13_2_15_9
void FASTCALL call_13_2_15_9(SH2_struct* sh)
{
  SH2movli(sh, 2, (15<<4)+9);
}
#define init_13_2_15_10
void FASTCALL call_13_2_15_10(SH2_struct* sh)
{
  SH2movli(sh, 2, (15<<4)+10);
}
#define init_13_2_15_11
void FASTCALL call_13_2_15_11(SH2_struct* sh)
{
  SH2movli(sh, 2, (15<<4)+11);
}
#define init_13_2_15_12
void FASTCALL call_13_2_15_12(SH2_struct* sh)
{
  SH2movli(sh, 2, (15<<4)+12);
}
#define init_13_2_15_13
void FASTCALL call_13_2_15_13(SH2_struct* sh)
{
  SH2movli(sh, 2, (15<<4)+13);
}
#define init_13_2_15_14
void FASTCALL call_13_2_15_14(SH2_struct* sh)
{
  SH2movli(sh, 2, (15<<4)+14);
}
#define init_13_2_15_15
void FASTCALL call_13_2_15_15(SH2_struct* sh)
{
  SH2movli(sh, 2, (15<<4)+15);
}
#define init_13_3_0_0
void FASTCALL call_13_3_0_0(SH2_struct* sh)
{
  SH2movli(sh, 3, (0<<4)+0);
}
#define init_13_3_0_1
void FASTCALL call_13_3_0_1(SH2_struct* sh)
{
  SH2movli(sh, 3, (0<<4)+1);
}
#define init_13_3_0_2
void FASTCALL call_13_3_0_2(SH2_struct* sh)
{
  SH2movli(sh, 3, (0<<4)+2);
}
#define init_13_3_0_3
void FASTCALL call_13_3_0_3(SH2_struct* sh)
{
  SH2movli(sh, 3, (0<<4)+3);
}
#define init_13_3_0_4
void FASTCALL call_13_3_0_4(SH2_struct* sh)
{
  SH2movli(sh, 3, (0<<4)+4);
}
#define init_13_3_0_5
void FASTCALL call_13_3_0_5(SH2_struct* sh)
{
  SH2movli(sh, 3, (0<<4)+5);
}
#define init_13_3_0_6
void FASTCALL call_13_3_0_6(SH2_struct* sh)
{
  SH2movli(sh, 3, (0<<4)+6);
}
#define init_13_3_0_7
void FASTCALL call_13_3_0_7(SH2_struct* sh)
{
  SH2movli(sh, 3, (0<<4)+7);
}
#define init_13_3_0_8
void FASTCALL call_13_3_0_8(SH2_struct* sh)
{
  SH2movli(sh, 3, (0<<4)+8);
}
#define init_13_3_0_9
void FASTCALL call_13_3_0_9(SH2_struct* sh)
{
  SH2movli(sh, 3, (0<<4)+9);
}
#define init_13_3_0_10
void FASTCALL call_13_3_0_10(SH2_struct* sh)
{
  SH2movli(sh, 3, (0<<4)+10);
}
#define init_13_3_0_11
void FASTCALL call_13_3_0_11(SH2_struct* sh)
{
  SH2movli(sh, 3, (0<<4)+11);
}
#define init_13_3_0_12
void FASTCALL call_13_3_0_12(SH2_struct* sh)
{
  SH2movli(sh, 3, (0<<4)+12);
}
#define init_13_3_0_13
void FASTCALL call_13_3_0_13(SH2_struct* sh)
{
  SH2movli(sh, 3, (0<<4)+13);
}
#define init_13_3_0_14
void FASTCALL call_13_3_0_14(SH2_struct* sh)
{
  SH2movli(sh, 3, (0<<4)+14);
}
#define init_13_3_0_15
void FASTCALL call_13_3_0_15(SH2_struct* sh)
{
  SH2movli(sh, 3, (0<<4)+15);
}
#define init_13_3_1_0
void FASTCALL call_13_3_1_0(SH2_struct* sh)
{
  SH2movli(sh, 3, (1<<4)+0);
}
#define init_13_3_1_1
void FASTCALL call_13_3_1_1(SH2_struct* sh)
{
  SH2movli(sh, 3, (1<<4)+1);
}
#define init_13_3_1_2
void FASTCALL call_13_3_1_2(SH2_struct* sh)
{
  SH2movli(sh, 3, (1<<4)+2);
}
#define init_13_3_1_3
void FASTCALL call_13_3_1_3(SH2_struct* sh)
{
  SH2movli(sh, 3, (1<<4)+3);
}
#define init_13_3_1_4
void FASTCALL call_13_3_1_4(SH2_struct* sh)
{
  SH2movli(sh, 3, (1<<4)+4);
}
#define init_13_3_1_5
void FASTCALL call_13_3_1_5(SH2_struct* sh)
{
  SH2movli(sh, 3, (1<<4)+5);
}
#define init_13_3_1_6
void FASTCALL call_13_3_1_6(SH2_struct* sh)
{
  SH2movli(sh, 3, (1<<4)+6);
}
#define init_13_3_1_7
void FASTCALL call_13_3_1_7(SH2_struct* sh)
{
  SH2movli(sh, 3, (1<<4)+7);
}
#define init_13_3_1_8
void FASTCALL call_13_3_1_8(SH2_struct* sh)
{
  SH2movli(sh, 3, (1<<4)+8);
}
#define init_13_3_1_9
void FASTCALL call_13_3_1_9(SH2_struct* sh)
{
  SH2movli(sh, 3, (1<<4)+9);
}
#define init_13_3_1_10
void FASTCALL call_13_3_1_10(SH2_struct* sh)
{
  SH2movli(sh, 3, (1<<4)+10);
}
#define init_13_3_1_11
void FASTCALL call_13_3_1_11(SH2_struct* sh)
{
  SH2movli(sh, 3, (1<<4)+11);
}
#define init_13_3_1_12
void FASTCALL call_13_3_1_12(SH2_struct* sh)
{
  SH2movli(sh, 3, (1<<4)+12);
}
#define init_13_3_1_13
void FASTCALL call_13_3_1_13(SH2_struct* sh)
{
  SH2movli(sh, 3, (1<<4)+13);
}
#define init_13_3_1_14
void FASTCALL call_13_3_1_14(SH2_struct* sh)
{
  SH2movli(sh, 3, (1<<4)+14);
}
#define init_13_3_1_15
void FASTCALL call_13_3_1_15(SH2_struct* sh)
{
  SH2movli(sh, 3, (1<<4)+15);
}
#define init_13_3_2_0
void FASTCALL call_13_3_2_0(SH2_struct* sh)
{
  SH2movli(sh, 3, (2<<4)+0);
}
#define init_13_3_2_1
void FASTCALL call_13_3_2_1(SH2_struct* sh)
{
  SH2movli(sh, 3, (2<<4)+1);
}
#define init_13_3_2_2
void FASTCALL call_13_3_2_2(SH2_struct* sh)
{
  SH2movli(sh, 3, (2<<4)+2);
}
#define init_13_3_2_3
void FASTCALL call_13_3_2_3(SH2_struct* sh)
{
  SH2movli(sh, 3, (2<<4)+3);
}
#define init_13_3_2_4
void FASTCALL call_13_3_2_4(SH2_struct* sh)
{
  SH2movli(sh, 3, (2<<4)+4);
}
#define init_13_3_2_5
void FASTCALL call_13_3_2_5(SH2_struct* sh)
{
  SH2movli(sh, 3, (2<<4)+5);
}
#define init_13_3_2_6
void FASTCALL call_13_3_2_6(SH2_struct* sh)
{
  SH2movli(sh, 3, (2<<4)+6);
}
#define init_13_3_2_7
void FASTCALL call_13_3_2_7(SH2_struct* sh)
{
  SH2movli(sh, 3, (2<<4)+7);
}
#define init_13_3_2_8
void FASTCALL call_13_3_2_8(SH2_struct* sh)
{
  SH2movli(sh, 3, (2<<4)+8);
}
#define init_13_3_2_9
void FASTCALL call_13_3_2_9(SH2_struct* sh)
{
  SH2movli(sh, 3, (2<<4)+9);
}
#define init_13_3_2_10
void FASTCALL call_13_3_2_10(SH2_struct* sh)
{
  SH2movli(sh, 3, (2<<4)+10);
}
#define init_13_3_2_11
void FASTCALL call_13_3_2_11(SH2_struct* sh)
{
  SH2movli(sh, 3, (2<<4)+11);
}
#define init_13_3_2_12
void FASTCALL call_13_3_2_12(SH2_struct* sh)
{
  SH2movli(sh, 3, (2<<4)+12);
}
#define init_13_3_2_13
void FASTCALL call_13_3_2_13(SH2_struct* sh)
{
  SH2movli(sh, 3, (2<<4)+13);
}
#define init_13_3_2_14
void FASTCALL call_13_3_2_14(SH2_struct* sh)
{
  SH2movli(sh, 3, (2<<4)+14);
}
#define init_13_3_2_15
void FASTCALL call_13_3_2_15(SH2_struct* sh)
{
  SH2movli(sh, 3, (2<<4)+15);
}
#define init_13_3_3_0
void FASTCALL call_13_3_3_0(SH2_struct* sh)
{
  SH2movli(sh, 3, (3<<4)+0);
}
#define init_13_3_3_1
void FASTCALL call_13_3_3_1(SH2_struct* sh)
{
  SH2movli(sh, 3, (3<<4)+1);
}
#define init_13_3_3_2
void FASTCALL call_13_3_3_2(SH2_struct* sh)
{
  SH2movli(sh, 3, (3<<4)+2);
}
#define init_13_3_3_3
void FASTCALL call_13_3_3_3(SH2_struct* sh)
{
  SH2movli(sh, 3, (3<<4)+3);
}
#define init_13_3_3_4
void FASTCALL call_13_3_3_4(SH2_struct* sh)
{
  SH2movli(sh, 3, (3<<4)+4);
}
#define init_13_3_3_5
void FASTCALL call_13_3_3_5(SH2_struct* sh)
{
  SH2movli(sh, 3, (3<<4)+5);
}
#define init_13_3_3_6
void FASTCALL call_13_3_3_6(SH2_struct* sh)
{
  SH2movli(sh, 3, (3<<4)+6);
}
#define init_13_3_3_7
void FASTCALL call_13_3_3_7(SH2_struct* sh)
{
  SH2movli(sh, 3, (3<<4)+7);
}
#define init_13_3_3_8
void FASTCALL call_13_3_3_8(SH2_struct* sh)
{
  SH2movli(sh, 3, (3<<4)+8);
}
#define init_13_3_3_9
void FASTCALL call_13_3_3_9(SH2_struct* sh)
{
  SH2movli(sh, 3, (3<<4)+9);
}
#define init_13_3_3_10
void FASTCALL call_13_3_3_10(SH2_struct* sh)
{
  SH2movli(sh, 3, (3<<4)+10);
}
#define init_13_3_3_11
void FASTCALL call_13_3_3_11(SH2_struct* sh)
{
  SH2movli(sh, 3, (3<<4)+11);
}
#define init_13_3_3_12
void FASTCALL call_13_3_3_12(SH2_struct* sh)
{
  SH2movli(sh, 3, (3<<4)+12);
}
#define init_13_3_3_13
void FASTCALL call_13_3_3_13(SH2_struct* sh)
{
  SH2movli(sh, 3, (3<<4)+13);
}
#define init_13_3_3_14
void FASTCALL call_13_3_3_14(SH2_struct* sh)
{
  SH2movli(sh, 3, (3<<4)+14);
}
#define init_13_3_3_15
void FASTCALL call_13_3_3_15(SH2_struct* sh)
{
  SH2movli(sh, 3, (3<<4)+15);
}
#define init_13_3_4_0
void FASTCALL call_13_3_4_0(SH2_struct* sh)
{
  SH2movli(sh, 3, (4<<4)+0);
}
#define init_13_3_4_1
void FASTCALL call_13_3_4_1(SH2_struct* sh)
{
  SH2movli(sh, 3, (4<<4)+1);
}
#define init_13_3_4_2
void FASTCALL call_13_3_4_2(SH2_struct* sh)
{
  SH2movli(sh, 3, (4<<4)+2);
}
#define init_13_3_4_3
void FASTCALL call_13_3_4_3(SH2_struct* sh)
{
  SH2movli(sh, 3, (4<<4)+3);
}
#define init_13_3_4_4
void FASTCALL call_13_3_4_4(SH2_struct* sh)
{
  SH2movli(sh, 3, (4<<4)+4);
}
#define init_13_3_4_5
void FASTCALL call_13_3_4_5(SH2_struct* sh)
{
  SH2movli(sh, 3, (4<<4)+5);
}
#define init_13_3_4_6
void FASTCALL call_13_3_4_6(SH2_struct* sh)
{
  SH2movli(sh, 3, (4<<4)+6);
}
#define init_13_3_4_7
void FASTCALL call_13_3_4_7(SH2_struct* sh)
{
  SH2movli(sh, 3, (4<<4)+7);
}
#define init_13_3_4_8
void FASTCALL call_13_3_4_8(SH2_struct* sh)
{
  SH2movli(sh, 3, (4<<4)+8);
}
#define init_13_3_4_9
void FASTCALL call_13_3_4_9(SH2_struct* sh)
{
  SH2movli(sh, 3, (4<<4)+9);
}
#define init_13_3_4_10
void FASTCALL call_13_3_4_10(SH2_struct* sh)
{
  SH2movli(sh, 3, (4<<4)+10);
}
#define init_13_3_4_11
void FASTCALL call_13_3_4_11(SH2_struct* sh)
{
  SH2movli(sh, 3, (4<<4)+11);
}
#define init_13_3_4_12
void FASTCALL call_13_3_4_12(SH2_struct* sh)
{
  SH2movli(sh, 3, (4<<4)+12);
}
#define init_13_3_4_13
void FASTCALL call_13_3_4_13(SH2_struct* sh)
{
  SH2movli(sh, 3, (4<<4)+13);
}
#define init_13_3_4_14
void FASTCALL call_13_3_4_14(SH2_struct* sh)
{
  SH2movli(sh, 3, (4<<4)+14);
}
#define init_13_3_4_15
void FASTCALL call_13_3_4_15(SH2_struct* sh)
{
  SH2movli(sh, 3, (4<<4)+15);
}
#define init_13_3_5_0
void FASTCALL call_13_3_5_0(SH2_struct* sh)
{
  SH2movli(sh, 3, (5<<4)+0);
}
#define init_13_3_5_1
void FASTCALL call_13_3_5_1(SH2_struct* sh)
{
  SH2movli(sh, 3, (5<<4)+1);
}
#define init_13_3_5_2
void FASTCALL call_13_3_5_2(SH2_struct* sh)
{
  SH2movli(sh, 3, (5<<4)+2);
}
#define init_13_3_5_3
void FASTCALL call_13_3_5_3(SH2_struct* sh)
{
  SH2movli(sh, 3, (5<<4)+3);
}
#define init_13_3_5_4
void FASTCALL call_13_3_5_4(SH2_struct* sh)
{
  SH2movli(sh, 3, (5<<4)+4);
}
#define init_13_3_5_5
void FASTCALL call_13_3_5_5(SH2_struct* sh)
{
  SH2movli(sh, 3, (5<<4)+5);
}
#define init_13_3_5_6
void FASTCALL call_13_3_5_6(SH2_struct* sh)
{
  SH2movli(sh, 3, (5<<4)+6);
}
#define init_13_3_5_7
void FASTCALL call_13_3_5_7(SH2_struct* sh)
{
  SH2movli(sh, 3, (5<<4)+7);
}
#define init_13_3_5_8
void FASTCALL call_13_3_5_8(SH2_struct* sh)
{
  SH2movli(sh, 3, (5<<4)+8);
}
#define init_13_3_5_9
void FASTCALL call_13_3_5_9(SH2_struct* sh)
{
  SH2movli(sh, 3, (5<<4)+9);
}
#define init_13_3_5_10
void FASTCALL call_13_3_5_10(SH2_struct* sh)
{
  SH2movli(sh, 3, (5<<4)+10);
}
#define init_13_3_5_11
void FASTCALL call_13_3_5_11(SH2_struct* sh)
{
  SH2movli(sh, 3, (5<<4)+11);
}
#define init_13_3_5_12
void FASTCALL call_13_3_5_12(SH2_struct* sh)
{
  SH2movli(sh, 3, (5<<4)+12);
}
#define init_13_3_5_13
void FASTCALL call_13_3_5_13(SH2_struct* sh)
{
  SH2movli(sh, 3, (5<<4)+13);
}
#define init_13_3_5_14
void FASTCALL call_13_3_5_14(SH2_struct* sh)
{
  SH2movli(sh, 3, (5<<4)+14);
}
#define init_13_3_5_15
void FASTCALL call_13_3_5_15(SH2_struct* sh)
{
  SH2movli(sh, 3, (5<<4)+15);
}
#define init_13_3_6_0
void FASTCALL call_13_3_6_0(SH2_struct* sh)
{
  SH2movli(sh, 3, (6<<4)+0);
}
#define init_13_3_6_1
void FASTCALL call_13_3_6_1(SH2_struct* sh)
{
  SH2movli(sh, 3, (6<<4)+1);
}
#define init_13_3_6_2
void FASTCALL call_13_3_6_2(SH2_struct* sh)
{
  SH2movli(sh, 3, (6<<4)+2);
}
#define init_13_3_6_3
void FASTCALL call_13_3_6_3(SH2_struct* sh)
{
  SH2movli(sh, 3, (6<<4)+3);
}
#define init_13_3_6_4
void FASTCALL call_13_3_6_4(SH2_struct* sh)
{
  SH2movli(sh, 3, (6<<4)+4);
}
#define init_13_3_6_5
void FASTCALL call_13_3_6_5(SH2_struct* sh)
{
  SH2movli(sh, 3, (6<<4)+5);
}
#define init_13_3_6_6
void FASTCALL call_13_3_6_6(SH2_struct* sh)
{
  SH2movli(sh, 3, (6<<4)+6);
}
#define init_13_3_6_7
void FASTCALL call_13_3_6_7(SH2_struct* sh)
{
  SH2movli(sh, 3, (6<<4)+7);
}
#define init_13_3_6_8
void FASTCALL call_13_3_6_8(SH2_struct* sh)
{
  SH2movli(sh, 3, (6<<4)+8);
}
#define init_13_3_6_9
void FASTCALL call_13_3_6_9(SH2_struct* sh)
{
  SH2movli(sh, 3, (6<<4)+9);
}
#define init_13_3_6_10
void FASTCALL call_13_3_6_10(SH2_struct* sh)
{
  SH2movli(sh, 3, (6<<4)+10);
}
#define init_13_3_6_11
void FASTCALL call_13_3_6_11(SH2_struct* sh)
{
  SH2movli(sh, 3, (6<<4)+11);
}
#define init_13_3_6_12
void FASTCALL call_13_3_6_12(SH2_struct* sh)
{
  SH2movli(sh, 3, (6<<4)+12);
}
#define init_13_3_6_13
void FASTCALL call_13_3_6_13(SH2_struct* sh)
{
  SH2movli(sh, 3, (6<<4)+13);
}
#define init_13_3_6_14
void FASTCALL call_13_3_6_14(SH2_struct* sh)
{
  SH2movli(sh, 3, (6<<4)+14);
}
#define init_13_3_6_15
void FASTCALL call_13_3_6_15(SH2_struct* sh)
{
  SH2movli(sh, 3, (6<<4)+15);
}
#define init_13_3_7_0
void FASTCALL call_13_3_7_0(SH2_struct* sh)
{
  SH2movli(sh, 3, (7<<4)+0);
}
#define init_13_3_7_1
void FASTCALL call_13_3_7_1(SH2_struct* sh)
{
  SH2movli(sh, 3, (7<<4)+1);
}
#define init_13_3_7_2
void FASTCALL call_13_3_7_2(SH2_struct* sh)
{
  SH2movli(sh, 3, (7<<4)+2);
}
#define init_13_3_7_3
void FASTCALL call_13_3_7_3(SH2_struct* sh)
{
  SH2movli(sh, 3, (7<<4)+3);
}
#define init_13_3_7_4
void FASTCALL call_13_3_7_4(SH2_struct* sh)
{
  SH2movli(sh, 3, (7<<4)+4);
}
#define init_13_3_7_5
void FASTCALL call_13_3_7_5(SH2_struct* sh)
{
  SH2movli(sh, 3, (7<<4)+5);
}
#define init_13_3_7_6
void FASTCALL call_13_3_7_6(SH2_struct* sh)
{
  SH2movli(sh, 3, (7<<4)+6);
}
#define init_13_3_7_7
void FASTCALL call_13_3_7_7(SH2_struct* sh)
{
  SH2movli(sh, 3, (7<<4)+7);
}
#define init_13_3_7_8
void FASTCALL call_13_3_7_8(SH2_struct* sh)
{
  SH2movli(sh, 3, (7<<4)+8);
}
#define init_13_3_7_9
void FASTCALL call_13_3_7_9(SH2_struct* sh)
{
  SH2movli(sh, 3, (7<<4)+9);
}
#define init_13_3_7_10
void FASTCALL call_13_3_7_10(SH2_struct* sh)
{
  SH2movli(sh, 3, (7<<4)+10);
}
#define init_13_3_7_11
void FASTCALL call_13_3_7_11(SH2_struct* sh)
{
  SH2movli(sh, 3, (7<<4)+11);
}
#define init_13_3_7_12
void FASTCALL call_13_3_7_12(SH2_struct* sh)
{
  SH2movli(sh, 3, (7<<4)+12);
}
#define init_13_3_7_13
void FASTCALL call_13_3_7_13(SH2_struct* sh)
{
  SH2movli(sh, 3, (7<<4)+13);
}
#define init_13_3_7_14
void FASTCALL call_13_3_7_14(SH2_struct* sh)
{
  SH2movli(sh, 3, (7<<4)+14);
}
#define init_13_3_7_15
void FASTCALL call_13_3_7_15(SH2_struct* sh)
{
  SH2movli(sh, 3, (7<<4)+15);
}
#define init_13_3_8_0
void FASTCALL call_13_3_8_0(SH2_struct* sh)
{
  SH2movli(sh, 3, (8<<4)+0);
}
#define init_13_3_8_1
void FASTCALL call_13_3_8_1(SH2_struct* sh)
{
  SH2movli(sh, 3, (8<<4)+1);
}
#define init_13_3_8_2
void FASTCALL call_13_3_8_2(SH2_struct* sh)
{
  SH2movli(sh, 3, (8<<4)+2);
}
#define init_13_3_8_3
void FASTCALL call_13_3_8_3(SH2_struct* sh)
{
  SH2movli(sh, 3, (8<<4)+3);
}
#define init_13_3_8_4
void FASTCALL call_13_3_8_4(SH2_struct* sh)
{
  SH2movli(sh, 3, (8<<4)+4);
}
#define init_13_3_8_5
void FASTCALL call_13_3_8_5(SH2_struct* sh)
{
  SH2movli(sh, 3, (8<<4)+5);
}
#define init_13_3_8_6
void FASTCALL call_13_3_8_6(SH2_struct* sh)
{
  SH2movli(sh, 3, (8<<4)+6);
}
#define init_13_3_8_7
void FASTCALL call_13_3_8_7(SH2_struct* sh)
{
  SH2movli(sh, 3, (8<<4)+7);
}
#define init_13_3_8_8
void FASTCALL call_13_3_8_8(SH2_struct* sh)
{
  SH2movli(sh, 3, (8<<4)+8);
}
#define init_13_3_8_9
void FASTCALL call_13_3_8_9(SH2_struct* sh)
{
  SH2movli(sh, 3, (8<<4)+9);
}
#define init_13_3_8_10
void FASTCALL call_13_3_8_10(SH2_struct* sh)
{
  SH2movli(sh, 3, (8<<4)+10);
}
#define init_13_3_8_11
void FASTCALL call_13_3_8_11(SH2_struct* sh)
{
  SH2movli(sh, 3, (8<<4)+11);
}
#define init_13_3_8_12
void FASTCALL call_13_3_8_12(SH2_struct* sh)
{
  SH2movli(sh, 3, (8<<4)+12);
}
#define init_13_3_8_13
void FASTCALL call_13_3_8_13(SH2_struct* sh)
{
  SH2movli(sh, 3, (8<<4)+13);
}
#define init_13_3_8_14
void FASTCALL call_13_3_8_14(SH2_struct* sh)
{
  SH2movli(sh, 3, (8<<4)+14);
}
#define init_13_3_8_15
void FASTCALL call_13_3_8_15(SH2_struct* sh)
{
  SH2movli(sh, 3, (8<<4)+15);
}
#define init_13_3_9_0
void FASTCALL call_13_3_9_0(SH2_struct* sh)
{
  SH2movli(sh, 3, (9<<4)+0);
}
#define init_13_3_9_1
void FASTCALL call_13_3_9_1(SH2_struct* sh)
{
  SH2movli(sh, 3, (9<<4)+1);
}
#define init_13_3_9_2
void FASTCALL call_13_3_9_2(SH2_struct* sh)
{
  SH2movli(sh, 3, (9<<4)+2);
}
#define init_13_3_9_3
void FASTCALL call_13_3_9_3(SH2_struct* sh)
{
  SH2movli(sh, 3, (9<<4)+3);
}
#define init_13_3_9_4
void FASTCALL call_13_3_9_4(SH2_struct* sh)
{
  SH2movli(sh, 3, (9<<4)+4);
}
#define init_13_3_9_5
void FASTCALL call_13_3_9_5(SH2_struct* sh)
{
  SH2movli(sh, 3, (9<<4)+5);
}
#define init_13_3_9_6
void FASTCALL call_13_3_9_6(SH2_struct* sh)
{
  SH2movli(sh, 3, (9<<4)+6);
}
#define init_13_3_9_7
void FASTCALL call_13_3_9_7(SH2_struct* sh)
{
  SH2movli(sh, 3, (9<<4)+7);
}
#define init_13_3_9_8
void FASTCALL call_13_3_9_8(SH2_struct* sh)
{
  SH2movli(sh, 3, (9<<4)+8);
}
#define init_13_3_9_9
void FASTCALL call_13_3_9_9(SH2_struct* sh)
{
  SH2movli(sh, 3, (9<<4)+9);
}
#define init_13_3_9_10
void FASTCALL call_13_3_9_10(SH2_struct* sh)
{
  SH2movli(sh, 3, (9<<4)+10);
}
#define init_13_3_9_11
void FASTCALL call_13_3_9_11(SH2_struct* sh)
{
  SH2movli(sh, 3, (9<<4)+11);
}
#define init_13_3_9_12
void FASTCALL call_13_3_9_12(SH2_struct* sh)
{
  SH2movli(sh, 3, (9<<4)+12);
}
#define init_13_3_9_13
void FASTCALL call_13_3_9_13(SH2_struct* sh)
{
  SH2movli(sh, 3, (9<<4)+13);
}
#define init_13_3_9_14
void FASTCALL call_13_3_9_14(SH2_struct* sh)
{
  SH2movli(sh, 3, (9<<4)+14);
}
#define init_13_3_9_15
void FASTCALL call_13_3_9_15(SH2_struct* sh)
{
  SH2movli(sh, 3, (9<<4)+15);
}
#define init_13_3_10_0
void FASTCALL call_13_3_10_0(SH2_struct* sh)
{
  SH2movli(sh, 3, (10<<4)+0);
}
#define init_13_3_10_1
void FASTCALL call_13_3_10_1(SH2_struct* sh)
{
  SH2movli(sh, 3, (10<<4)+1);
}
#define init_13_3_10_2
void FASTCALL call_13_3_10_2(SH2_struct* sh)
{
  SH2movli(sh, 3, (10<<4)+2);
}
#define init_13_3_10_3
void FASTCALL call_13_3_10_3(SH2_struct* sh)
{
  SH2movli(sh, 3, (10<<4)+3);
}
#define init_13_3_10_4
void FASTCALL call_13_3_10_4(SH2_struct* sh)
{
  SH2movli(sh, 3, (10<<4)+4);
}
#define init_13_3_10_5
void FASTCALL call_13_3_10_5(SH2_struct* sh)
{
  SH2movli(sh, 3, (10<<4)+5);
}
#define init_13_3_10_6
void FASTCALL call_13_3_10_6(SH2_struct* sh)
{
  SH2movli(sh, 3, (10<<4)+6);
}
#define init_13_3_10_7
void FASTCALL call_13_3_10_7(SH2_struct* sh)
{
  SH2movli(sh, 3, (10<<4)+7);
}
#define init_13_3_10_8
void FASTCALL call_13_3_10_8(SH2_struct* sh)
{
  SH2movli(sh, 3, (10<<4)+8);
}
#define init_13_3_10_9
void FASTCALL call_13_3_10_9(SH2_struct* sh)
{
  SH2movli(sh, 3, (10<<4)+9);
}
#define init_13_3_10_10
void FASTCALL call_13_3_10_10(SH2_struct* sh)
{
  SH2movli(sh, 3, (10<<4)+10);
}
#define init_13_3_10_11
void FASTCALL call_13_3_10_11(SH2_struct* sh)
{
  SH2movli(sh, 3, (10<<4)+11);
}
#define init_13_3_10_12
void FASTCALL call_13_3_10_12(SH2_struct* sh)
{
  SH2movli(sh, 3, (10<<4)+12);
}
#define init_13_3_10_13
void FASTCALL call_13_3_10_13(SH2_struct* sh)
{
  SH2movli(sh, 3, (10<<4)+13);
}
#define init_13_3_10_14
void FASTCALL call_13_3_10_14(SH2_struct* sh)
{
  SH2movli(sh, 3, (10<<4)+14);
}
#define init_13_3_10_15
void FASTCALL call_13_3_10_15(SH2_struct* sh)
{
  SH2movli(sh, 3, (10<<4)+15);
}
#define init_13_3_11_0
void FASTCALL call_13_3_11_0(SH2_struct* sh)
{
  SH2movli(sh, 3, (11<<4)+0);
}
#define init_13_3_11_1
void FASTCALL call_13_3_11_1(SH2_struct* sh)
{
  SH2movli(sh, 3, (11<<4)+1);
}
#define init_13_3_11_2
void FASTCALL call_13_3_11_2(SH2_struct* sh)
{
  SH2movli(sh, 3, (11<<4)+2);
}
#define init_13_3_11_3
void FASTCALL call_13_3_11_3(SH2_struct* sh)
{
  SH2movli(sh, 3, (11<<4)+3);
}
#define init_13_3_11_4
void FASTCALL call_13_3_11_4(SH2_struct* sh)
{
  SH2movli(sh, 3, (11<<4)+4);
}
#define init_13_3_11_5
void FASTCALL call_13_3_11_5(SH2_struct* sh)
{
  SH2movli(sh, 3, (11<<4)+5);
}
#define init_13_3_11_6
void FASTCALL call_13_3_11_6(SH2_struct* sh)
{
  SH2movli(sh, 3, (11<<4)+6);
}
#define init_13_3_11_7
void FASTCALL call_13_3_11_7(SH2_struct* sh)
{
  SH2movli(sh, 3, (11<<4)+7);
}
#define init_13_3_11_8
void FASTCALL call_13_3_11_8(SH2_struct* sh)
{
  SH2movli(sh, 3, (11<<4)+8);
}
#define init_13_3_11_9
void FASTCALL call_13_3_11_9(SH2_struct* sh)
{
  SH2movli(sh, 3, (11<<4)+9);
}
#define init_13_3_11_10
void FASTCALL call_13_3_11_10(SH2_struct* sh)
{
  SH2movli(sh, 3, (11<<4)+10);
}
#define init_13_3_11_11
void FASTCALL call_13_3_11_11(SH2_struct* sh)
{
  SH2movli(sh, 3, (11<<4)+11);
}
#define init_13_3_11_12
void FASTCALL call_13_3_11_12(SH2_struct* sh)
{
  SH2movli(sh, 3, (11<<4)+12);
}
#define init_13_3_11_13
void FASTCALL call_13_3_11_13(SH2_struct* sh)
{
  SH2movli(sh, 3, (11<<4)+13);
}
#define init_13_3_11_14
void FASTCALL call_13_3_11_14(SH2_struct* sh)
{
  SH2movli(sh, 3, (11<<4)+14);
}
#define init_13_3_11_15
void FASTCALL call_13_3_11_15(SH2_struct* sh)
{
  SH2movli(sh, 3, (11<<4)+15);
}
#define init_13_3_12_0
void FASTCALL call_13_3_12_0(SH2_struct* sh)
{
  SH2movli(sh, 3, (12<<4)+0);
}
#define init_13_3_12_1
void FASTCALL call_13_3_12_1(SH2_struct* sh)
{
  SH2movli(sh, 3, (12<<4)+1);
}
#define init_13_3_12_2
void FASTCALL call_13_3_12_2(SH2_struct* sh)
{
  SH2movli(sh, 3, (12<<4)+2);
}
#define init_13_3_12_3
void FASTCALL call_13_3_12_3(SH2_struct* sh)
{
  SH2movli(sh, 3, (12<<4)+3);
}
#define init_13_3_12_4
void FASTCALL call_13_3_12_4(SH2_struct* sh)
{
  SH2movli(sh, 3, (12<<4)+4);
}
#define init_13_3_12_5
void FASTCALL call_13_3_12_5(SH2_struct* sh)
{
  SH2movli(sh, 3, (12<<4)+5);
}
#define init_13_3_12_6
void FASTCALL call_13_3_12_6(SH2_struct* sh)
{
  SH2movli(sh, 3, (12<<4)+6);
}
#define init_13_3_12_7
void FASTCALL call_13_3_12_7(SH2_struct* sh)
{
  SH2movli(sh, 3, (12<<4)+7);
}
#define init_13_3_12_8
void FASTCALL call_13_3_12_8(SH2_struct* sh)
{
  SH2movli(sh, 3, (12<<4)+8);
}
#define init_13_3_12_9
void FASTCALL call_13_3_12_9(SH2_struct* sh)
{
  SH2movli(sh, 3, (12<<4)+9);
}
#define init_13_3_12_10
void FASTCALL call_13_3_12_10(SH2_struct* sh)
{
  SH2movli(sh, 3, (12<<4)+10);
}
#define init_13_3_12_11
void FASTCALL call_13_3_12_11(SH2_struct* sh)
{
  SH2movli(sh, 3, (12<<4)+11);
}
#define init_13_3_12_12
void FASTCALL call_13_3_12_12(SH2_struct* sh)
{
  SH2movli(sh, 3, (12<<4)+12);
}
#define init_13_3_12_13
void FASTCALL call_13_3_12_13(SH2_struct* sh)
{
  SH2movli(sh, 3, (12<<4)+13);
}
#define init_13_3_12_14
void FASTCALL call_13_3_12_14(SH2_struct* sh)
{
  SH2movli(sh, 3, (12<<4)+14);
}
#define init_13_3_12_15
void FASTCALL call_13_3_12_15(SH2_struct* sh)
{
  SH2movli(sh, 3, (12<<4)+15);
}
#define init_13_3_13_0
void FASTCALL call_13_3_13_0(SH2_struct* sh)
{
  SH2movli(sh, 3, (13<<4)+0);
}
#define init_13_3_13_1
void FASTCALL call_13_3_13_1(SH2_struct* sh)
{
  SH2movli(sh, 3, (13<<4)+1);
}
#define init_13_3_13_2
void FASTCALL call_13_3_13_2(SH2_struct* sh)
{
  SH2movli(sh, 3, (13<<4)+2);
}
#define init_13_3_13_3
void FASTCALL call_13_3_13_3(SH2_struct* sh)
{
  SH2movli(sh, 3, (13<<4)+3);
}
#define init_13_3_13_4
void FASTCALL call_13_3_13_4(SH2_struct* sh)
{
  SH2movli(sh, 3, (13<<4)+4);
}
#define init_13_3_13_5
void FASTCALL call_13_3_13_5(SH2_struct* sh)
{
  SH2movli(sh, 3, (13<<4)+5);
}
#define init_13_3_13_6
void FASTCALL call_13_3_13_6(SH2_struct* sh)
{
  SH2movli(sh, 3, (13<<4)+6);
}
#define init_13_3_13_7
void FASTCALL call_13_3_13_7(SH2_struct* sh)
{
  SH2movli(sh, 3, (13<<4)+7);
}
#define init_13_3_13_8
void FASTCALL call_13_3_13_8(SH2_struct* sh)
{
  SH2movli(sh, 3, (13<<4)+8);
}
#define init_13_3_13_9
void FASTCALL call_13_3_13_9(SH2_struct* sh)
{
  SH2movli(sh, 3, (13<<4)+9);
}
#define init_13_3_13_10
void FASTCALL call_13_3_13_10(SH2_struct* sh)
{
  SH2movli(sh, 3, (13<<4)+10);
}
#define init_13_3_13_11
void FASTCALL call_13_3_13_11(SH2_struct* sh)
{
  SH2movli(sh, 3, (13<<4)+11);
}
#define init_13_3_13_12
void FASTCALL call_13_3_13_12(SH2_struct* sh)
{
  SH2movli(sh, 3, (13<<4)+12);
}
#define init_13_3_13_13
void FASTCALL call_13_3_13_13(SH2_struct* sh)
{
  SH2movli(sh, 3, (13<<4)+13);
}
#define init_13_3_13_14
void FASTCALL call_13_3_13_14(SH2_struct* sh)
{
  SH2movli(sh, 3, (13<<4)+14);
}
#define init_13_3_13_15
void FASTCALL call_13_3_13_15(SH2_struct* sh)
{
  SH2movli(sh, 3, (13<<4)+15);
}
#define init_13_3_14_0
void FASTCALL call_13_3_14_0(SH2_struct* sh)
{
  SH2movli(sh, 3, (14<<4)+0);
}
#define init_13_3_14_1
void FASTCALL call_13_3_14_1(SH2_struct* sh)
{
  SH2movli(sh, 3, (14<<4)+1);
}
#define init_13_3_14_2
void FASTCALL call_13_3_14_2(SH2_struct* sh)
{
  SH2movli(sh, 3, (14<<4)+2);
}
#define init_13_3_14_3
void FASTCALL call_13_3_14_3(SH2_struct* sh)
{
  SH2movli(sh, 3, (14<<4)+3);
}
#define init_13_3_14_4
void FASTCALL call_13_3_14_4(SH2_struct* sh)
{
  SH2movli(sh, 3, (14<<4)+4);
}
#define init_13_3_14_5
void FASTCALL call_13_3_14_5(SH2_struct* sh)
{
  SH2movli(sh, 3, (14<<4)+5);
}
#define init_13_3_14_6
void FASTCALL call_13_3_14_6(SH2_struct* sh)
{
  SH2movli(sh, 3, (14<<4)+6);
}
#define init_13_3_14_7
void FASTCALL call_13_3_14_7(SH2_struct* sh)
{
  SH2movli(sh, 3, (14<<4)+7);
}
#define init_13_3_14_8
void FASTCALL call_13_3_14_8(SH2_struct* sh)
{
  SH2movli(sh, 3, (14<<4)+8);
}
#define init_13_3_14_9
void FASTCALL call_13_3_14_9(SH2_struct* sh)
{
  SH2movli(sh, 3, (14<<4)+9);
}
#define init_13_3_14_10
void FASTCALL call_13_3_14_10(SH2_struct* sh)
{
  SH2movli(sh, 3, (14<<4)+10);
}
#define init_13_3_14_11
void FASTCALL call_13_3_14_11(SH2_struct* sh)
{
  SH2movli(sh, 3, (14<<4)+11);
}
#define init_13_3_14_12
void FASTCALL call_13_3_14_12(SH2_struct* sh)
{
  SH2movli(sh, 3, (14<<4)+12);
}
#define init_13_3_14_13
void FASTCALL call_13_3_14_13(SH2_struct* sh)
{
  SH2movli(sh, 3, (14<<4)+13);
}
#define init_13_3_14_14
void FASTCALL call_13_3_14_14(SH2_struct* sh)
{
  SH2movli(sh, 3, (14<<4)+14);
}
#define init_13_3_14_15
void FASTCALL call_13_3_14_15(SH2_struct* sh)
{
  SH2movli(sh, 3, (14<<4)+15);
}
#define init_13_3_15_0
void FASTCALL call_13_3_15_0(SH2_struct* sh)
{
  SH2movli(sh, 3, (15<<4)+0);
}
#define init_13_3_15_1
void FASTCALL call_13_3_15_1(SH2_struct* sh)
{
  SH2movli(sh, 3, (15<<4)+1);
}
#define init_13_3_15_2
void FASTCALL call_13_3_15_2(SH2_struct* sh)
{
  SH2movli(sh, 3, (15<<4)+2);
}
#define init_13_3_15_3
void FASTCALL call_13_3_15_3(SH2_struct* sh)
{
  SH2movli(sh, 3, (15<<4)+3);
}
#define init_13_3_15_4
void FASTCALL call_13_3_15_4(SH2_struct* sh)
{
  SH2movli(sh, 3, (15<<4)+4);
}
#define init_13_3_15_5
void FASTCALL call_13_3_15_5(SH2_struct* sh)
{
  SH2movli(sh, 3, (15<<4)+5);
}
#define init_13_3_15_6
void FASTCALL call_13_3_15_6(SH2_struct* sh)
{
  SH2movli(sh, 3, (15<<4)+6);
}
#define init_13_3_15_7
void FASTCALL call_13_3_15_7(SH2_struct* sh)
{
  SH2movli(sh, 3, (15<<4)+7);
}
#define init_13_3_15_8
void FASTCALL call_13_3_15_8(SH2_struct* sh)
{
  SH2movli(sh, 3, (15<<4)+8);
}
#define init_13_3_15_9
void FASTCALL call_13_3_15_9(SH2_struct* sh)
{
  SH2movli(sh, 3, (15<<4)+9);
}
#define init_13_3_15_10
void FASTCALL call_13_3_15_10(SH2_struct* sh)
{
  SH2movli(sh, 3, (15<<4)+10);
}
#define init_13_3_15_11
void FASTCALL call_13_3_15_11(SH2_struct* sh)
{
  SH2movli(sh, 3, (15<<4)+11);
}
#define init_13_3_15_12
void FASTCALL call_13_3_15_12(SH2_struct* sh)
{
  SH2movli(sh, 3, (15<<4)+12);
}
#define init_13_3_15_13
void FASTCALL call_13_3_15_13(SH2_struct* sh)
{
  SH2movli(sh, 3, (15<<4)+13);
}
#define init_13_3_15_14
void FASTCALL call_13_3_15_14(SH2_struct* sh)
{
  SH2movli(sh, 3, (15<<4)+14);
}
#define init_13_3_15_15
void FASTCALL call_13_3_15_15(SH2_struct* sh)
{
  SH2movli(sh, 3, (15<<4)+15);
}
#define init_13_4_0_0
void FASTCALL call_13_4_0_0(SH2_struct* sh)
{
  SH2movli(sh, 4, (0<<4)+0);
}
#define init_13_4_0_1
void FASTCALL call_13_4_0_1(SH2_struct* sh)
{
  SH2movli(sh, 4, (0<<4)+1);
}
#define init_13_4_0_2
void FASTCALL call_13_4_0_2(SH2_struct* sh)
{
  SH2movli(sh, 4, (0<<4)+2);
}
#define init_13_4_0_3
void FASTCALL call_13_4_0_3(SH2_struct* sh)
{
  SH2movli(sh, 4, (0<<4)+3);
}
#define init_13_4_0_4
void FASTCALL call_13_4_0_4(SH2_struct* sh)
{
  SH2movli(sh, 4, (0<<4)+4);
}
#define init_13_4_0_5
void FASTCALL call_13_4_0_5(SH2_struct* sh)
{
  SH2movli(sh, 4, (0<<4)+5);
}
#define init_13_4_0_6
void FASTCALL call_13_4_0_6(SH2_struct* sh)
{
  SH2movli(sh, 4, (0<<4)+6);
}
#define init_13_4_0_7
void FASTCALL call_13_4_0_7(SH2_struct* sh)
{
  SH2movli(sh, 4, (0<<4)+7);
}
#define init_13_4_0_8
void FASTCALL call_13_4_0_8(SH2_struct* sh)
{
  SH2movli(sh, 4, (0<<4)+8);
}
#define init_13_4_0_9
void FASTCALL call_13_4_0_9(SH2_struct* sh)
{
  SH2movli(sh, 4, (0<<4)+9);
}
#define init_13_4_0_10
void FASTCALL call_13_4_0_10(SH2_struct* sh)
{
  SH2movli(sh, 4, (0<<4)+10);
}
#define init_13_4_0_11
void FASTCALL call_13_4_0_11(SH2_struct* sh)
{
  SH2movli(sh, 4, (0<<4)+11);
}
#define init_13_4_0_12
void FASTCALL call_13_4_0_12(SH2_struct* sh)
{
  SH2movli(sh, 4, (0<<4)+12);
}
#define init_13_4_0_13
void FASTCALL call_13_4_0_13(SH2_struct* sh)
{
  SH2movli(sh, 4, (0<<4)+13);
}
#define init_13_4_0_14
void FASTCALL call_13_4_0_14(SH2_struct* sh)
{
  SH2movli(sh, 4, (0<<4)+14);
}
#define init_13_4_0_15
void FASTCALL call_13_4_0_15(SH2_struct* sh)
{
  SH2movli(sh, 4, (0<<4)+15);
}
#define init_13_4_1_0
void FASTCALL call_13_4_1_0(SH2_struct* sh)
{
  SH2movli(sh, 4, (1<<4)+0);
}
#define init_13_4_1_1
void FASTCALL call_13_4_1_1(SH2_struct* sh)
{
  SH2movli(sh, 4, (1<<4)+1);
}
#define init_13_4_1_2
void FASTCALL call_13_4_1_2(SH2_struct* sh)
{
  SH2movli(sh, 4, (1<<4)+2);
}
#define init_13_4_1_3
void FASTCALL call_13_4_1_3(SH2_struct* sh)
{
  SH2movli(sh, 4, (1<<4)+3);
}
#define init_13_4_1_4
void FASTCALL call_13_4_1_4(SH2_struct* sh)
{
  SH2movli(sh, 4, (1<<4)+4);
}
#define init_13_4_1_5
void FASTCALL call_13_4_1_5(SH2_struct* sh)
{
  SH2movli(sh, 4, (1<<4)+5);
}
#define init_13_4_1_6
void FASTCALL call_13_4_1_6(SH2_struct* sh)
{
  SH2movli(sh, 4, (1<<4)+6);
}
#define init_13_4_1_7
void FASTCALL call_13_4_1_7(SH2_struct* sh)
{
  SH2movli(sh, 4, (1<<4)+7);
}
#define init_13_4_1_8
void FASTCALL call_13_4_1_8(SH2_struct* sh)
{
  SH2movli(sh, 4, (1<<4)+8);
}
#define init_13_4_1_9
void FASTCALL call_13_4_1_9(SH2_struct* sh)
{
  SH2movli(sh, 4, (1<<4)+9);
}
#define init_13_4_1_10
void FASTCALL call_13_4_1_10(SH2_struct* sh)
{
  SH2movli(sh, 4, (1<<4)+10);
}
#define init_13_4_1_11
void FASTCALL call_13_4_1_11(SH2_struct* sh)
{
  SH2movli(sh, 4, (1<<4)+11);
}
#define init_13_4_1_12
void FASTCALL call_13_4_1_12(SH2_struct* sh)
{
  SH2movli(sh, 4, (1<<4)+12);
}
#define init_13_4_1_13
void FASTCALL call_13_4_1_13(SH2_struct* sh)
{
  SH2movli(sh, 4, (1<<4)+13);
}
#define init_13_4_1_14
void FASTCALL call_13_4_1_14(SH2_struct* sh)
{
  SH2movli(sh, 4, (1<<4)+14);
}
#define init_13_4_1_15
void FASTCALL call_13_4_1_15(SH2_struct* sh)
{
  SH2movli(sh, 4, (1<<4)+15);
}
#define init_13_4_2_0
void FASTCALL call_13_4_2_0(SH2_struct* sh)
{
  SH2movli(sh, 4, (2<<4)+0);
}
#define init_13_4_2_1
void FASTCALL call_13_4_2_1(SH2_struct* sh)
{
  SH2movli(sh, 4, (2<<4)+1);
}
#define init_13_4_2_2
void FASTCALL call_13_4_2_2(SH2_struct* sh)
{
  SH2movli(sh, 4, (2<<4)+2);
}
#define init_13_4_2_3
void FASTCALL call_13_4_2_3(SH2_struct* sh)
{
  SH2movli(sh, 4, (2<<4)+3);
}
#define init_13_4_2_4
void FASTCALL call_13_4_2_4(SH2_struct* sh)
{
  SH2movli(sh, 4, (2<<4)+4);
}
#define init_13_4_2_5
void FASTCALL call_13_4_2_5(SH2_struct* sh)
{
  SH2movli(sh, 4, (2<<4)+5);
}
#define init_13_4_2_6
void FASTCALL call_13_4_2_6(SH2_struct* sh)
{
  SH2movli(sh, 4, (2<<4)+6);
}
#define init_13_4_2_7
void FASTCALL call_13_4_2_7(SH2_struct* sh)
{
  SH2movli(sh, 4, (2<<4)+7);
}
#define init_13_4_2_8
void FASTCALL call_13_4_2_8(SH2_struct* sh)
{
  SH2movli(sh, 4, (2<<4)+8);
}
#define init_13_4_2_9
void FASTCALL call_13_4_2_9(SH2_struct* sh)
{
  SH2movli(sh, 4, (2<<4)+9);
}
#define init_13_4_2_10
void FASTCALL call_13_4_2_10(SH2_struct* sh)
{
  SH2movli(sh, 4, (2<<4)+10);
}
#define init_13_4_2_11
void FASTCALL call_13_4_2_11(SH2_struct* sh)
{
  SH2movli(sh, 4, (2<<4)+11);
}
#define init_13_4_2_12
void FASTCALL call_13_4_2_12(SH2_struct* sh)
{
  SH2movli(sh, 4, (2<<4)+12);
}
#define init_13_4_2_13
void FASTCALL call_13_4_2_13(SH2_struct* sh)
{
  SH2movli(sh, 4, (2<<4)+13);
}
#define init_13_4_2_14
void FASTCALL call_13_4_2_14(SH2_struct* sh)
{
  SH2movli(sh, 4, (2<<4)+14);
}
#define init_13_4_2_15
void FASTCALL call_13_4_2_15(SH2_struct* sh)
{
  SH2movli(sh, 4, (2<<4)+15);
}
#define init_13_4_3_0
void FASTCALL call_13_4_3_0(SH2_struct* sh)
{
  SH2movli(sh, 4, (3<<4)+0);
}
#define init_13_4_3_1
void FASTCALL call_13_4_3_1(SH2_struct* sh)
{
  SH2movli(sh, 4, (3<<4)+1);
}
#define init_13_4_3_2
void FASTCALL call_13_4_3_2(SH2_struct* sh)
{
  SH2movli(sh, 4, (3<<4)+2);
}
#define init_13_4_3_3
void FASTCALL call_13_4_3_3(SH2_struct* sh)
{
  SH2movli(sh, 4, (3<<4)+3);
}
#define init_13_4_3_4
void FASTCALL call_13_4_3_4(SH2_struct* sh)
{
  SH2movli(sh, 4, (3<<4)+4);
}
#define init_13_4_3_5
void FASTCALL call_13_4_3_5(SH2_struct* sh)
{
  SH2movli(sh, 4, (3<<4)+5);
}
#define init_13_4_3_6
void FASTCALL call_13_4_3_6(SH2_struct* sh)
{
  SH2movli(sh, 4, (3<<4)+6);
}
#define init_13_4_3_7
void FASTCALL call_13_4_3_7(SH2_struct* sh)
{
  SH2movli(sh, 4, (3<<4)+7);
}
#define init_13_4_3_8
void FASTCALL call_13_4_3_8(SH2_struct* sh)
{
  SH2movli(sh, 4, (3<<4)+8);
}
#define init_13_4_3_9
void FASTCALL call_13_4_3_9(SH2_struct* sh)
{
  SH2movli(sh, 4, (3<<4)+9);
}
#define init_13_4_3_10
void FASTCALL call_13_4_3_10(SH2_struct* sh)
{
  SH2movli(sh, 4, (3<<4)+10);
}
#define init_13_4_3_11
void FASTCALL call_13_4_3_11(SH2_struct* sh)
{
  SH2movli(sh, 4, (3<<4)+11);
}
#define init_13_4_3_12
void FASTCALL call_13_4_3_12(SH2_struct* sh)
{
  SH2movli(sh, 4, (3<<4)+12);
}
#define init_13_4_3_13
void FASTCALL call_13_4_3_13(SH2_struct* sh)
{
  SH2movli(sh, 4, (3<<4)+13);
}
#define init_13_4_3_14
void FASTCALL call_13_4_3_14(SH2_struct* sh)
{
  SH2movli(sh, 4, (3<<4)+14);
}
#define init_13_4_3_15
void FASTCALL call_13_4_3_15(SH2_struct* sh)
{
  SH2movli(sh, 4, (3<<4)+15);
}
#define init_13_4_4_0
void FASTCALL call_13_4_4_0(SH2_struct* sh)
{
  SH2movli(sh, 4, (4<<4)+0);
}
#define init_13_4_4_1
void FASTCALL call_13_4_4_1(SH2_struct* sh)
{
  SH2movli(sh, 4, (4<<4)+1);
}
#define init_13_4_4_2
void FASTCALL call_13_4_4_2(SH2_struct* sh)
{
  SH2movli(sh, 4, (4<<4)+2);
}
#define init_13_4_4_3
void FASTCALL call_13_4_4_3(SH2_struct* sh)
{
  SH2movli(sh, 4, (4<<4)+3);
}
#define init_13_4_4_4
void FASTCALL call_13_4_4_4(SH2_struct* sh)
{
  SH2movli(sh, 4, (4<<4)+4);
}
#define init_13_4_4_5
void FASTCALL call_13_4_4_5(SH2_struct* sh)
{
  SH2movli(sh, 4, (4<<4)+5);
}
#define init_13_4_4_6
void FASTCALL call_13_4_4_6(SH2_struct* sh)
{
  SH2movli(sh, 4, (4<<4)+6);
}
#define init_13_4_4_7
void FASTCALL call_13_4_4_7(SH2_struct* sh)
{
  SH2movli(sh, 4, (4<<4)+7);
}
#define init_13_4_4_8
void FASTCALL call_13_4_4_8(SH2_struct* sh)
{
  SH2movli(sh, 4, (4<<4)+8);
}
#define init_13_4_4_9
void FASTCALL call_13_4_4_9(SH2_struct* sh)
{
  SH2movli(sh, 4, (4<<4)+9);
}
#define init_13_4_4_10
void FASTCALL call_13_4_4_10(SH2_struct* sh)
{
  SH2movli(sh, 4, (4<<4)+10);
}
#define init_13_4_4_11
void FASTCALL call_13_4_4_11(SH2_struct* sh)
{
  SH2movli(sh, 4, (4<<4)+11);
}
#define init_13_4_4_12
void FASTCALL call_13_4_4_12(SH2_struct* sh)
{
  SH2movli(sh, 4, (4<<4)+12);
}
#define init_13_4_4_13
void FASTCALL call_13_4_4_13(SH2_struct* sh)
{
  SH2movli(sh, 4, (4<<4)+13);
}
#define init_13_4_4_14
void FASTCALL call_13_4_4_14(SH2_struct* sh)
{
  SH2movli(sh, 4, (4<<4)+14);
}
#define init_13_4_4_15
void FASTCALL call_13_4_4_15(SH2_struct* sh)
{
  SH2movli(sh, 4, (4<<4)+15);
}
#define init_13_4_5_0
void FASTCALL call_13_4_5_0(SH2_struct* sh)
{
  SH2movli(sh, 4, (5<<4)+0);
}
#define init_13_4_5_1
void FASTCALL call_13_4_5_1(SH2_struct* sh)
{
  SH2movli(sh, 4, (5<<4)+1);
}
#define init_13_4_5_2
void FASTCALL call_13_4_5_2(SH2_struct* sh)
{
  SH2movli(sh, 4, (5<<4)+2);
}
#define init_13_4_5_3
void FASTCALL call_13_4_5_3(SH2_struct* sh)
{
  SH2movli(sh, 4, (5<<4)+3);
}
#define init_13_4_5_4
void FASTCALL call_13_4_5_4(SH2_struct* sh)
{
  SH2movli(sh, 4, (5<<4)+4);
}
#define init_13_4_5_5
void FASTCALL call_13_4_5_5(SH2_struct* sh)
{
  SH2movli(sh, 4, (5<<4)+5);
}
#define init_13_4_5_6
void FASTCALL call_13_4_5_6(SH2_struct* sh)
{
  SH2movli(sh, 4, (5<<4)+6);
}
#define init_13_4_5_7
void FASTCALL call_13_4_5_7(SH2_struct* sh)
{
  SH2movli(sh, 4, (5<<4)+7);
}
#define init_13_4_5_8
void FASTCALL call_13_4_5_8(SH2_struct* sh)
{
  SH2movli(sh, 4, (5<<4)+8);
}
#define init_13_4_5_9
void FASTCALL call_13_4_5_9(SH2_struct* sh)
{
  SH2movli(sh, 4, (5<<4)+9);
}
#define init_13_4_5_10
void FASTCALL call_13_4_5_10(SH2_struct* sh)
{
  SH2movli(sh, 4, (5<<4)+10);
}
#define init_13_4_5_11
void FASTCALL call_13_4_5_11(SH2_struct* sh)
{
  SH2movli(sh, 4, (5<<4)+11);
}
#define init_13_4_5_12
void FASTCALL call_13_4_5_12(SH2_struct* sh)
{
  SH2movli(sh, 4, (5<<4)+12);
}
#define init_13_4_5_13
void FASTCALL call_13_4_5_13(SH2_struct* sh)
{
  SH2movli(sh, 4, (5<<4)+13);
}
#define init_13_4_5_14
void FASTCALL call_13_4_5_14(SH2_struct* sh)
{
  SH2movli(sh, 4, (5<<4)+14);
}
#define init_13_4_5_15
void FASTCALL call_13_4_5_15(SH2_struct* sh)
{
  SH2movli(sh, 4, (5<<4)+15);
}
#define init_13_4_6_0
void FASTCALL call_13_4_6_0(SH2_struct* sh)
{
  SH2movli(sh, 4, (6<<4)+0);
}
#define init_13_4_6_1
void FASTCALL call_13_4_6_1(SH2_struct* sh)
{
  SH2movli(sh, 4, (6<<4)+1);
}
#define init_13_4_6_2
void FASTCALL call_13_4_6_2(SH2_struct* sh)
{
  SH2movli(sh, 4, (6<<4)+2);
}
#define init_13_4_6_3
void FASTCALL call_13_4_6_3(SH2_struct* sh)
{
  SH2movli(sh, 4, (6<<4)+3);
}
#define init_13_4_6_4
void FASTCALL call_13_4_6_4(SH2_struct* sh)
{
  SH2movli(sh, 4, (6<<4)+4);
}
#define init_13_4_6_5
void FASTCALL call_13_4_6_5(SH2_struct* sh)
{
  SH2movli(sh, 4, (6<<4)+5);
}
#define init_13_4_6_6
void FASTCALL call_13_4_6_6(SH2_struct* sh)
{
  SH2movli(sh, 4, (6<<4)+6);
}
#define init_13_4_6_7
void FASTCALL call_13_4_6_7(SH2_struct* sh)
{
  SH2movli(sh, 4, (6<<4)+7);
}
#define init_13_4_6_8
void FASTCALL call_13_4_6_8(SH2_struct* sh)
{
  SH2movli(sh, 4, (6<<4)+8);
}
#define init_13_4_6_9
void FASTCALL call_13_4_6_9(SH2_struct* sh)
{
  SH2movli(sh, 4, (6<<4)+9);
}
#define init_13_4_6_10
void FASTCALL call_13_4_6_10(SH2_struct* sh)
{
  SH2movli(sh, 4, (6<<4)+10);
}
#define init_13_4_6_11
void FASTCALL call_13_4_6_11(SH2_struct* sh)
{
  SH2movli(sh, 4, (6<<4)+11);
}
#define init_13_4_6_12
void FASTCALL call_13_4_6_12(SH2_struct* sh)
{
  SH2movli(sh, 4, (6<<4)+12);
}
#define init_13_4_6_13
void FASTCALL call_13_4_6_13(SH2_struct* sh)
{
  SH2movli(sh, 4, (6<<4)+13);
}
#define init_13_4_6_14
void FASTCALL call_13_4_6_14(SH2_struct* sh)
{
  SH2movli(sh, 4, (6<<4)+14);
}
#define init_13_4_6_15
void FASTCALL call_13_4_6_15(SH2_struct* sh)
{
  SH2movli(sh, 4, (6<<4)+15);
}
#define init_13_4_7_0
void FASTCALL call_13_4_7_0(SH2_struct* sh)
{
  SH2movli(sh, 4, (7<<4)+0);
}
#define init_13_4_7_1
void FASTCALL call_13_4_7_1(SH2_struct* sh)
{
  SH2movli(sh, 4, (7<<4)+1);
}
#define init_13_4_7_2
void FASTCALL call_13_4_7_2(SH2_struct* sh)
{
  SH2movli(sh, 4, (7<<4)+2);
}
#define init_13_4_7_3
void FASTCALL call_13_4_7_3(SH2_struct* sh)
{
  SH2movli(sh, 4, (7<<4)+3);
}
#define init_13_4_7_4
void FASTCALL call_13_4_7_4(SH2_struct* sh)
{
  SH2movli(sh, 4, (7<<4)+4);
}
#define init_13_4_7_5
void FASTCALL call_13_4_7_5(SH2_struct* sh)
{
  SH2movli(sh, 4, (7<<4)+5);
}
#define init_13_4_7_6
void FASTCALL call_13_4_7_6(SH2_struct* sh)
{
  SH2movli(sh, 4, (7<<4)+6);
}
#define init_13_4_7_7
void FASTCALL call_13_4_7_7(SH2_struct* sh)
{
  SH2movli(sh, 4, (7<<4)+7);
}
#define init_13_4_7_8
void FASTCALL call_13_4_7_8(SH2_struct* sh)
{
  SH2movli(sh, 4, (7<<4)+8);
}
#define init_13_4_7_9
void FASTCALL call_13_4_7_9(SH2_struct* sh)
{
  SH2movli(sh, 4, (7<<4)+9);
}
#define init_13_4_7_10
void FASTCALL call_13_4_7_10(SH2_struct* sh)
{
  SH2movli(sh, 4, (7<<4)+10);
}
#define init_13_4_7_11
void FASTCALL call_13_4_7_11(SH2_struct* sh)
{
  SH2movli(sh, 4, (7<<4)+11);
}
#define init_13_4_7_12
void FASTCALL call_13_4_7_12(SH2_struct* sh)
{
  SH2movli(sh, 4, (7<<4)+12);
}
#define init_13_4_7_13
void FASTCALL call_13_4_7_13(SH2_struct* sh)
{
  SH2movli(sh, 4, (7<<4)+13);
}
#define init_13_4_7_14
void FASTCALL call_13_4_7_14(SH2_struct* sh)
{
  SH2movli(sh, 4, (7<<4)+14);
}
#define init_13_4_7_15
void FASTCALL call_13_4_7_15(SH2_struct* sh)
{
  SH2movli(sh, 4, (7<<4)+15);
}
#define init_13_4_8_0
void FASTCALL call_13_4_8_0(SH2_struct* sh)
{
  SH2movli(sh, 4, (8<<4)+0);
}
#define init_13_4_8_1
void FASTCALL call_13_4_8_1(SH2_struct* sh)
{
  SH2movli(sh, 4, (8<<4)+1);
}
#define init_13_4_8_2
void FASTCALL call_13_4_8_2(SH2_struct* sh)
{
  SH2movli(sh, 4, (8<<4)+2);
}
#define init_13_4_8_3
void FASTCALL call_13_4_8_3(SH2_struct* sh)
{
  SH2movli(sh, 4, (8<<4)+3);
}
#define init_13_4_8_4
void FASTCALL call_13_4_8_4(SH2_struct* sh)
{
  SH2movli(sh, 4, (8<<4)+4);
}
#define init_13_4_8_5
void FASTCALL call_13_4_8_5(SH2_struct* sh)
{
  SH2movli(sh, 4, (8<<4)+5);
}
#define init_13_4_8_6
void FASTCALL call_13_4_8_6(SH2_struct* sh)
{
  SH2movli(sh, 4, (8<<4)+6);
}
#define init_13_4_8_7
void FASTCALL call_13_4_8_7(SH2_struct* sh)
{
  SH2movli(sh, 4, (8<<4)+7);
}
#define init_13_4_8_8
void FASTCALL call_13_4_8_8(SH2_struct* sh)
{
  SH2movli(sh, 4, (8<<4)+8);
}
#define init_13_4_8_9
void FASTCALL call_13_4_8_9(SH2_struct* sh)
{
  SH2movli(sh, 4, (8<<4)+9);
}
#define init_13_4_8_10
void FASTCALL call_13_4_8_10(SH2_struct* sh)
{
  SH2movli(sh, 4, (8<<4)+10);
}
#define init_13_4_8_11
void FASTCALL call_13_4_8_11(SH2_struct* sh)
{
  SH2movli(sh, 4, (8<<4)+11);
}
#define init_13_4_8_12
void FASTCALL call_13_4_8_12(SH2_struct* sh)
{
  SH2movli(sh, 4, (8<<4)+12);
}
#define init_13_4_8_13
void FASTCALL call_13_4_8_13(SH2_struct* sh)
{
  SH2movli(sh, 4, (8<<4)+13);
}
#define init_13_4_8_14
void FASTCALL call_13_4_8_14(SH2_struct* sh)
{
  SH2movli(sh, 4, (8<<4)+14);
}
#define init_13_4_8_15
void FASTCALL call_13_4_8_15(SH2_struct* sh)
{
  SH2movli(sh, 4, (8<<4)+15);
}
#define init_13_4_9_0
void FASTCALL call_13_4_9_0(SH2_struct* sh)
{
  SH2movli(sh, 4, (9<<4)+0);
}
#define init_13_4_9_1
void FASTCALL call_13_4_9_1(SH2_struct* sh)
{
  SH2movli(sh, 4, (9<<4)+1);
}
#define init_13_4_9_2
void FASTCALL call_13_4_9_2(SH2_struct* sh)
{
  SH2movli(sh, 4, (9<<4)+2);
}
#define init_13_4_9_3
void FASTCALL call_13_4_9_3(SH2_struct* sh)
{
  SH2movli(sh, 4, (9<<4)+3);
}
#define init_13_4_9_4
void FASTCALL call_13_4_9_4(SH2_struct* sh)
{
  SH2movli(sh, 4, (9<<4)+4);
}
#define init_13_4_9_5
void FASTCALL call_13_4_9_5(SH2_struct* sh)
{
  SH2movli(sh, 4, (9<<4)+5);
}
#define init_13_4_9_6
void FASTCALL call_13_4_9_6(SH2_struct* sh)
{
  SH2movli(sh, 4, (9<<4)+6);
}
#define init_13_4_9_7
void FASTCALL call_13_4_9_7(SH2_struct* sh)
{
  SH2movli(sh, 4, (9<<4)+7);
}
#define init_13_4_9_8
void FASTCALL call_13_4_9_8(SH2_struct* sh)
{
  SH2movli(sh, 4, (9<<4)+8);
}
#define init_13_4_9_9
void FASTCALL call_13_4_9_9(SH2_struct* sh)
{
  SH2movli(sh, 4, (9<<4)+9);
}
#define init_13_4_9_10
void FASTCALL call_13_4_9_10(SH2_struct* sh)
{
  SH2movli(sh, 4, (9<<4)+10);
}
#define init_13_4_9_11
void FASTCALL call_13_4_9_11(SH2_struct* sh)
{
  SH2movli(sh, 4, (9<<4)+11);
}
#define init_13_4_9_12
void FASTCALL call_13_4_9_12(SH2_struct* sh)
{
  SH2movli(sh, 4, (9<<4)+12);
}
#define init_13_4_9_13
void FASTCALL call_13_4_9_13(SH2_struct* sh)
{
  SH2movli(sh, 4, (9<<4)+13);
}
#define init_13_4_9_14
void FASTCALL call_13_4_9_14(SH2_struct* sh)
{
  SH2movli(sh, 4, (9<<4)+14);
}
#define init_13_4_9_15
void FASTCALL call_13_4_9_15(SH2_struct* sh)
{
  SH2movli(sh, 4, (9<<4)+15);
}
#define init_13_4_10_0
void FASTCALL call_13_4_10_0(SH2_struct* sh)
{
  SH2movli(sh, 4, (10<<4)+0);
}
#define init_13_4_10_1
void FASTCALL call_13_4_10_1(SH2_struct* sh)
{
  SH2movli(sh, 4, (10<<4)+1);
}
#define init_13_4_10_2
void FASTCALL call_13_4_10_2(SH2_struct* sh)
{
  SH2movli(sh, 4, (10<<4)+2);
}
#define init_13_4_10_3
void FASTCALL call_13_4_10_3(SH2_struct* sh)
{
  SH2movli(sh, 4, (10<<4)+3);
}
#define init_13_4_10_4
void FASTCALL call_13_4_10_4(SH2_struct* sh)
{
  SH2movli(sh, 4, (10<<4)+4);
}
#define init_13_4_10_5
void FASTCALL call_13_4_10_5(SH2_struct* sh)
{
  SH2movli(sh, 4, (10<<4)+5);
}
#define init_13_4_10_6
void FASTCALL call_13_4_10_6(SH2_struct* sh)
{
  SH2movli(sh, 4, (10<<4)+6);
}
#define init_13_4_10_7
void FASTCALL call_13_4_10_7(SH2_struct* sh)
{
  SH2movli(sh, 4, (10<<4)+7);
}
#define init_13_4_10_8
void FASTCALL call_13_4_10_8(SH2_struct* sh)
{
  SH2movli(sh, 4, (10<<4)+8);
}
#define init_13_4_10_9
void FASTCALL call_13_4_10_9(SH2_struct* sh)
{
  SH2movli(sh, 4, (10<<4)+9);
}
#define init_13_4_10_10
void FASTCALL call_13_4_10_10(SH2_struct* sh)
{
  SH2movli(sh, 4, (10<<4)+10);
}
#define init_13_4_10_11
void FASTCALL call_13_4_10_11(SH2_struct* sh)
{
  SH2movli(sh, 4, (10<<4)+11);
}
#define init_13_4_10_12
void FASTCALL call_13_4_10_12(SH2_struct* sh)
{
  SH2movli(sh, 4, (10<<4)+12);
}
#define init_13_4_10_13
void FASTCALL call_13_4_10_13(SH2_struct* sh)
{
  SH2movli(sh, 4, (10<<4)+13);
}
#define init_13_4_10_14
void FASTCALL call_13_4_10_14(SH2_struct* sh)
{
  SH2movli(sh, 4, (10<<4)+14);
}
#define init_13_4_10_15
void FASTCALL call_13_4_10_15(SH2_struct* sh)
{
  SH2movli(sh, 4, (10<<4)+15);
}
#define init_13_4_11_0
void FASTCALL call_13_4_11_0(SH2_struct* sh)
{
  SH2movli(sh, 4, (11<<4)+0);
}
#define init_13_4_11_1
void FASTCALL call_13_4_11_1(SH2_struct* sh)
{
  SH2movli(sh, 4, (11<<4)+1);
}
#define init_13_4_11_2
void FASTCALL call_13_4_11_2(SH2_struct* sh)
{
  SH2movli(sh, 4, (11<<4)+2);
}
#define init_13_4_11_3
void FASTCALL call_13_4_11_3(SH2_struct* sh)
{
  SH2movli(sh, 4, (11<<4)+3);
}
#define init_13_4_11_4
void FASTCALL call_13_4_11_4(SH2_struct* sh)
{
  SH2movli(sh, 4, (11<<4)+4);
}
#define init_13_4_11_5
void FASTCALL call_13_4_11_5(SH2_struct* sh)
{
  SH2movli(sh, 4, (11<<4)+5);
}
#define init_13_4_11_6
void FASTCALL call_13_4_11_6(SH2_struct* sh)
{
  SH2movli(sh, 4, (11<<4)+6);
}
#define init_13_4_11_7
void FASTCALL call_13_4_11_7(SH2_struct* sh)
{
  SH2movli(sh, 4, (11<<4)+7);
}
#define init_13_4_11_8
void FASTCALL call_13_4_11_8(SH2_struct* sh)
{
  SH2movli(sh, 4, (11<<4)+8);
}
#define init_13_4_11_9
void FASTCALL call_13_4_11_9(SH2_struct* sh)
{
  SH2movli(sh, 4, (11<<4)+9);
}
#define init_13_4_11_10
void FASTCALL call_13_4_11_10(SH2_struct* sh)
{
  SH2movli(sh, 4, (11<<4)+10);
}
#define init_13_4_11_11
void FASTCALL call_13_4_11_11(SH2_struct* sh)
{
  SH2movli(sh, 4, (11<<4)+11);
}
#define init_13_4_11_12
void FASTCALL call_13_4_11_12(SH2_struct* sh)
{
  SH2movli(sh, 4, (11<<4)+12);
}
#define init_13_4_11_13
void FASTCALL call_13_4_11_13(SH2_struct* sh)
{
  SH2movli(sh, 4, (11<<4)+13);
}
#define init_13_4_11_14
void FASTCALL call_13_4_11_14(SH2_struct* sh)
{
  SH2movli(sh, 4, (11<<4)+14);
}
#define init_13_4_11_15
void FASTCALL call_13_4_11_15(SH2_struct* sh)
{
  SH2movli(sh, 4, (11<<4)+15);
}
#define init_13_4_12_0
void FASTCALL call_13_4_12_0(SH2_struct* sh)
{
  SH2movli(sh, 4, (12<<4)+0);
}
#define init_13_4_12_1
void FASTCALL call_13_4_12_1(SH2_struct* sh)
{
  SH2movli(sh, 4, (12<<4)+1);
}
#define init_13_4_12_2
void FASTCALL call_13_4_12_2(SH2_struct* sh)
{
  SH2movli(sh, 4, (12<<4)+2);
}
#define init_13_4_12_3
void FASTCALL call_13_4_12_3(SH2_struct* sh)
{
  SH2movli(sh, 4, (12<<4)+3);
}
#define init_13_4_12_4
void FASTCALL call_13_4_12_4(SH2_struct* sh)
{
  SH2movli(sh, 4, (12<<4)+4);
}
#define init_13_4_12_5
void FASTCALL call_13_4_12_5(SH2_struct* sh)
{
  SH2movli(sh, 4, (12<<4)+5);
}
#define init_13_4_12_6
void FASTCALL call_13_4_12_6(SH2_struct* sh)
{
  SH2movli(sh, 4, (12<<4)+6);
}
#define init_13_4_12_7
void FASTCALL call_13_4_12_7(SH2_struct* sh)
{
  SH2movli(sh, 4, (12<<4)+7);
}
#define init_13_4_12_8
void FASTCALL call_13_4_12_8(SH2_struct* sh)
{
  SH2movli(sh, 4, (12<<4)+8);
}
#define init_13_4_12_9
void FASTCALL call_13_4_12_9(SH2_struct* sh)
{
  SH2movli(sh, 4, (12<<4)+9);
}
#define init_13_4_12_10
void FASTCALL call_13_4_12_10(SH2_struct* sh)
{
  SH2movli(sh, 4, (12<<4)+10);
}
#define init_13_4_12_11
void FASTCALL call_13_4_12_11(SH2_struct* sh)
{
  SH2movli(sh, 4, (12<<4)+11);
}
#define init_13_4_12_12
void FASTCALL call_13_4_12_12(SH2_struct* sh)
{
  SH2movli(sh, 4, (12<<4)+12);
}
#define init_13_4_12_13
void FASTCALL call_13_4_12_13(SH2_struct* sh)
{
  SH2movli(sh, 4, (12<<4)+13);
}
#define init_13_4_12_14
void FASTCALL call_13_4_12_14(SH2_struct* sh)
{
  SH2movli(sh, 4, (12<<4)+14);
}
#define init_13_4_12_15
void FASTCALL call_13_4_12_15(SH2_struct* sh)
{
  SH2movli(sh, 4, (12<<4)+15);
}
#define init_13_4_13_0
void FASTCALL call_13_4_13_0(SH2_struct* sh)
{
  SH2movli(sh, 4, (13<<4)+0);
}
#define init_13_4_13_1
void FASTCALL call_13_4_13_1(SH2_struct* sh)
{
  SH2movli(sh, 4, (13<<4)+1);
}
#define init_13_4_13_2
void FASTCALL call_13_4_13_2(SH2_struct* sh)
{
  SH2movli(sh, 4, (13<<4)+2);
}
#define init_13_4_13_3
void FASTCALL call_13_4_13_3(SH2_struct* sh)
{
  SH2movli(sh, 4, (13<<4)+3);
}
#define init_13_4_13_4
void FASTCALL call_13_4_13_4(SH2_struct* sh)
{
  SH2movli(sh, 4, (13<<4)+4);
}
#define init_13_4_13_5
void FASTCALL call_13_4_13_5(SH2_struct* sh)
{
  SH2movli(sh, 4, (13<<4)+5);
}
#define init_13_4_13_6
void FASTCALL call_13_4_13_6(SH2_struct* sh)
{
  SH2movli(sh, 4, (13<<4)+6);
}
#define init_13_4_13_7
void FASTCALL call_13_4_13_7(SH2_struct* sh)
{
  SH2movli(sh, 4, (13<<4)+7);
}
#define init_13_4_13_8
void FASTCALL call_13_4_13_8(SH2_struct* sh)
{
  SH2movli(sh, 4, (13<<4)+8);
}
#define init_13_4_13_9
void FASTCALL call_13_4_13_9(SH2_struct* sh)
{
  SH2movli(sh, 4, (13<<4)+9);
}
#define init_13_4_13_10
void FASTCALL call_13_4_13_10(SH2_struct* sh)
{
  SH2movli(sh, 4, (13<<4)+10);
}
#define init_13_4_13_11
void FASTCALL call_13_4_13_11(SH2_struct* sh)
{
  SH2movli(sh, 4, (13<<4)+11);
}
#define init_13_4_13_12
void FASTCALL call_13_4_13_12(SH2_struct* sh)
{
  SH2movli(sh, 4, (13<<4)+12);
}
#define init_13_4_13_13
void FASTCALL call_13_4_13_13(SH2_struct* sh)
{
  SH2movli(sh, 4, (13<<4)+13);
}
#define init_13_4_13_14
void FASTCALL call_13_4_13_14(SH2_struct* sh)
{
  SH2movli(sh, 4, (13<<4)+14);
}
#define init_13_4_13_15
void FASTCALL call_13_4_13_15(SH2_struct* sh)
{
  SH2movli(sh, 4, (13<<4)+15);
}
#define init_13_4_14_0
void FASTCALL call_13_4_14_0(SH2_struct* sh)
{
  SH2movli(sh, 4, (14<<4)+0);
}
#define init_13_4_14_1
void FASTCALL call_13_4_14_1(SH2_struct* sh)
{
  SH2movli(sh, 4, (14<<4)+1);
}
#define init_13_4_14_2
void FASTCALL call_13_4_14_2(SH2_struct* sh)
{
  SH2movli(sh, 4, (14<<4)+2);
}
#define init_13_4_14_3
void FASTCALL call_13_4_14_3(SH2_struct* sh)
{
  SH2movli(sh, 4, (14<<4)+3);
}
#define init_13_4_14_4
void FASTCALL call_13_4_14_4(SH2_struct* sh)
{
  SH2movli(sh, 4, (14<<4)+4);
}
#define init_13_4_14_5
void FASTCALL call_13_4_14_5(SH2_struct* sh)
{
  SH2movli(sh, 4, (14<<4)+5);
}
#define init_13_4_14_6
void FASTCALL call_13_4_14_6(SH2_struct* sh)
{
  SH2movli(sh, 4, (14<<4)+6);
}
#define init_13_4_14_7
void FASTCALL call_13_4_14_7(SH2_struct* sh)
{
  SH2movli(sh, 4, (14<<4)+7);
}
#define init_13_4_14_8
void FASTCALL call_13_4_14_8(SH2_struct* sh)
{
  SH2movli(sh, 4, (14<<4)+8);
}
#define init_13_4_14_9
void FASTCALL call_13_4_14_9(SH2_struct* sh)
{
  SH2movli(sh, 4, (14<<4)+9);
}
#define init_13_4_14_10
void FASTCALL call_13_4_14_10(SH2_struct* sh)
{
  SH2movli(sh, 4, (14<<4)+10);
}
#define init_13_4_14_11
void FASTCALL call_13_4_14_11(SH2_struct* sh)
{
  SH2movli(sh, 4, (14<<4)+11);
}
#define init_13_4_14_12
void FASTCALL call_13_4_14_12(SH2_struct* sh)
{
  SH2movli(sh, 4, (14<<4)+12);
}
#define init_13_4_14_13
void FASTCALL call_13_4_14_13(SH2_struct* sh)
{
  SH2movli(sh, 4, (14<<4)+13);
}
#define init_13_4_14_14
void FASTCALL call_13_4_14_14(SH2_struct* sh)
{
  SH2movli(sh, 4, (14<<4)+14);
}
#define init_13_4_14_15
void FASTCALL call_13_4_14_15(SH2_struct* sh)
{
  SH2movli(sh, 4, (14<<4)+15);
}
#define init_13_4_15_0
void FASTCALL call_13_4_15_0(SH2_struct* sh)
{
  SH2movli(sh, 4, (15<<4)+0);
}
#define init_13_4_15_1
void FASTCALL call_13_4_15_1(SH2_struct* sh)
{
  SH2movli(sh, 4, (15<<4)+1);
}
#define init_13_4_15_2
void FASTCALL call_13_4_15_2(SH2_struct* sh)
{
  SH2movli(sh, 4, (15<<4)+2);
}
#define init_13_4_15_3
void FASTCALL call_13_4_15_3(SH2_struct* sh)
{
  SH2movli(sh, 4, (15<<4)+3);
}
#define init_13_4_15_4
void FASTCALL call_13_4_15_4(SH2_struct* sh)
{
  SH2movli(sh, 4, (15<<4)+4);
}
#define init_13_4_15_5
void FASTCALL call_13_4_15_5(SH2_struct* sh)
{
  SH2movli(sh, 4, (15<<4)+5);
}
#define init_13_4_15_6
void FASTCALL call_13_4_15_6(SH2_struct* sh)
{
  SH2movli(sh, 4, (15<<4)+6);
}
#define init_13_4_15_7
void FASTCALL call_13_4_15_7(SH2_struct* sh)
{
  SH2movli(sh, 4, (15<<4)+7);
}
#define init_13_4_15_8
void FASTCALL call_13_4_15_8(SH2_struct* sh)
{
  SH2movli(sh, 4, (15<<4)+8);
}
#define init_13_4_15_9
void FASTCALL call_13_4_15_9(SH2_struct* sh)
{
  SH2movli(sh, 4, (15<<4)+9);
}
#define init_13_4_15_10
void FASTCALL call_13_4_15_10(SH2_struct* sh)
{
  SH2movli(sh, 4, (15<<4)+10);
}
#define init_13_4_15_11
void FASTCALL call_13_4_15_11(SH2_struct* sh)
{
  SH2movli(sh, 4, (15<<4)+11);
}
#define init_13_4_15_12
void FASTCALL call_13_4_15_12(SH2_struct* sh)
{
  SH2movli(sh, 4, (15<<4)+12);
}
#define init_13_4_15_13
void FASTCALL call_13_4_15_13(SH2_struct* sh)
{
  SH2movli(sh, 4, (15<<4)+13);
}
#define init_13_4_15_14
void FASTCALL call_13_4_15_14(SH2_struct* sh)
{
  SH2movli(sh, 4, (15<<4)+14);
}
#define init_13_4_15_15
void FASTCALL call_13_4_15_15(SH2_struct* sh)
{
  SH2movli(sh, 4, (15<<4)+15);
}
#define init_13_5_0_0
void FASTCALL call_13_5_0_0(SH2_struct* sh)
{
  SH2movli(sh, 5, (0<<4)+0);
}
#define init_13_5_0_1
void FASTCALL call_13_5_0_1(SH2_struct* sh)
{
  SH2movli(sh, 5, (0<<4)+1);
}
#define init_13_5_0_2
void FASTCALL call_13_5_0_2(SH2_struct* sh)
{
  SH2movli(sh, 5, (0<<4)+2);
}
#define init_13_5_0_3
void FASTCALL call_13_5_0_3(SH2_struct* sh)
{
  SH2movli(sh, 5, (0<<4)+3);
}
#define init_13_5_0_4
void FASTCALL call_13_5_0_4(SH2_struct* sh)
{
  SH2movli(sh, 5, (0<<4)+4);
}
#define init_13_5_0_5
void FASTCALL call_13_5_0_5(SH2_struct* sh)
{
  SH2movli(sh, 5, (0<<4)+5);
}
#define init_13_5_0_6
void FASTCALL call_13_5_0_6(SH2_struct* sh)
{
  SH2movli(sh, 5, (0<<4)+6);
}
#define init_13_5_0_7
void FASTCALL call_13_5_0_7(SH2_struct* sh)
{
  SH2movli(sh, 5, (0<<4)+7);
}
#define init_13_5_0_8
void FASTCALL call_13_5_0_8(SH2_struct* sh)
{
  SH2movli(sh, 5, (0<<4)+8);
}
#define init_13_5_0_9
void FASTCALL call_13_5_0_9(SH2_struct* sh)
{
  SH2movli(sh, 5, (0<<4)+9);
}
#define init_13_5_0_10
void FASTCALL call_13_5_0_10(SH2_struct* sh)
{
  SH2movli(sh, 5, (0<<4)+10);
}
#define init_13_5_0_11
void FASTCALL call_13_5_0_11(SH2_struct* sh)
{
  SH2movli(sh, 5, (0<<4)+11);
}
#define init_13_5_0_12
void FASTCALL call_13_5_0_12(SH2_struct* sh)
{
  SH2movli(sh, 5, (0<<4)+12);
}
#define init_13_5_0_13
void FASTCALL call_13_5_0_13(SH2_struct* sh)
{
  SH2movli(sh, 5, (0<<4)+13);
}
#define init_13_5_0_14
void FASTCALL call_13_5_0_14(SH2_struct* sh)
{
  SH2movli(sh, 5, (0<<4)+14);
}
#define init_13_5_0_15
void FASTCALL call_13_5_0_15(SH2_struct* sh)
{
  SH2movli(sh, 5, (0<<4)+15);
}
#define init_13_5_1_0
void FASTCALL call_13_5_1_0(SH2_struct* sh)
{
  SH2movli(sh, 5, (1<<4)+0);
}
#define init_13_5_1_1
void FASTCALL call_13_5_1_1(SH2_struct* sh)
{
  SH2movli(sh, 5, (1<<4)+1);
}
#define init_13_5_1_2
void FASTCALL call_13_5_1_2(SH2_struct* sh)
{
  SH2movli(sh, 5, (1<<4)+2);
}
#define init_13_5_1_3
void FASTCALL call_13_5_1_3(SH2_struct* sh)
{
  SH2movli(sh, 5, (1<<4)+3);
}
#define init_13_5_1_4
void FASTCALL call_13_5_1_4(SH2_struct* sh)
{
  SH2movli(sh, 5, (1<<4)+4);
}
#define init_13_5_1_5
void FASTCALL call_13_5_1_5(SH2_struct* sh)
{
  SH2movli(sh, 5, (1<<4)+5);
}
#define init_13_5_1_6
void FASTCALL call_13_5_1_6(SH2_struct* sh)
{
  SH2movli(sh, 5, (1<<4)+6);
}
#define init_13_5_1_7
void FASTCALL call_13_5_1_7(SH2_struct* sh)
{
  SH2movli(sh, 5, (1<<4)+7);
}
#define init_13_5_1_8
void FASTCALL call_13_5_1_8(SH2_struct* sh)
{
  SH2movli(sh, 5, (1<<4)+8);
}
#define init_13_5_1_9
void FASTCALL call_13_5_1_9(SH2_struct* sh)
{
  SH2movli(sh, 5, (1<<4)+9);
}
#define init_13_5_1_10
void FASTCALL call_13_5_1_10(SH2_struct* sh)
{
  SH2movli(sh, 5, (1<<4)+10);
}
#define init_13_5_1_11
void FASTCALL call_13_5_1_11(SH2_struct* sh)
{
  SH2movli(sh, 5, (1<<4)+11);
}
#define init_13_5_1_12
void FASTCALL call_13_5_1_12(SH2_struct* sh)
{
  SH2movli(sh, 5, (1<<4)+12);
}
#define init_13_5_1_13
void FASTCALL call_13_5_1_13(SH2_struct* sh)
{
  SH2movli(sh, 5, (1<<4)+13);
}
#define init_13_5_1_14
void FASTCALL call_13_5_1_14(SH2_struct* sh)
{
  SH2movli(sh, 5, (1<<4)+14);
}
#define init_13_5_1_15
void FASTCALL call_13_5_1_15(SH2_struct* sh)
{
  SH2movli(sh, 5, (1<<4)+15);
}
#define init_13_5_2_0
void FASTCALL call_13_5_2_0(SH2_struct* sh)
{
  SH2movli(sh, 5, (2<<4)+0);
}
#define init_13_5_2_1
void FASTCALL call_13_5_2_1(SH2_struct* sh)
{
  SH2movli(sh, 5, (2<<4)+1);
}
#define init_13_5_2_2
void FASTCALL call_13_5_2_2(SH2_struct* sh)
{
  SH2movli(sh, 5, (2<<4)+2);
}
#define init_13_5_2_3
void FASTCALL call_13_5_2_3(SH2_struct* sh)
{
  SH2movli(sh, 5, (2<<4)+3);
}
#define init_13_5_2_4
void FASTCALL call_13_5_2_4(SH2_struct* sh)
{
  SH2movli(sh, 5, (2<<4)+4);
}
#define init_13_5_2_5
void FASTCALL call_13_5_2_5(SH2_struct* sh)
{
  SH2movli(sh, 5, (2<<4)+5);
}
#define init_13_5_2_6
void FASTCALL call_13_5_2_6(SH2_struct* sh)
{
  SH2movli(sh, 5, (2<<4)+6);
}
#define init_13_5_2_7
void FASTCALL call_13_5_2_7(SH2_struct* sh)
{
  SH2movli(sh, 5, (2<<4)+7);
}
#define init_13_5_2_8
void FASTCALL call_13_5_2_8(SH2_struct* sh)
{
  SH2movli(sh, 5, (2<<4)+8);
}
#define init_13_5_2_9
void FASTCALL call_13_5_2_9(SH2_struct* sh)
{
  SH2movli(sh, 5, (2<<4)+9);
}
#define init_13_5_2_10
void FASTCALL call_13_5_2_10(SH2_struct* sh)
{
  SH2movli(sh, 5, (2<<4)+10);
}
#define init_13_5_2_11
void FASTCALL call_13_5_2_11(SH2_struct* sh)
{
  SH2movli(sh, 5, (2<<4)+11);
}
#define init_13_5_2_12
void FASTCALL call_13_5_2_12(SH2_struct* sh)
{
  SH2movli(sh, 5, (2<<4)+12);
}
#define init_13_5_2_13
void FASTCALL call_13_5_2_13(SH2_struct* sh)
{
  SH2movli(sh, 5, (2<<4)+13);
}
#define init_13_5_2_14
void FASTCALL call_13_5_2_14(SH2_struct* sh)
{
  SH2movli(sh, 5, (2<<4)+14);
}
#define init_13_5_2_15
void FASTCALL call_13_5_2_15(SH2_struct* sh)
{
  SH2movli(sh, 5, (2<<4)+15);
}
#define init_13_5_3_0
void FASTCALL call_13_5_3_0(SH2_struct* sh)
{
  SH2movli(sh, 5, (3<<4)+0);
}
#define init_13_5_3_1
void FASTCALL call_13_5_3_1(SH2_struct* sh)
{
  SH2movli(sh, 5, (3<<4)+1);
}
#define init_13_5_3_2
void FASTCALL call_13_5_3_2(SH2_struct* sh)
{
  SH2movli(sh, 5, (3<<4)+2);
}
#define init_13_5_3_3
void FASTCALL call_13_5_3_3(SH2_struct* sh)
{
  SH2movli(sh, 5, (3<<4)+3);
}
#define init_13_5_3_4
void FASTCALL call_13_5_3_4(SH2_struct* sh)
{
  SH2movli(sh, 5, (3<<4)+4);
}
#define init_13_5_3_5
void FASTCALL call_13_5_3_5(SH2_struct* sh)
{
  SH2movli(sh, 5, (3<<4)+5);
}
#define init_13_5_3_6
void FASTCALL call_13_5_3_6(SH2_struct* sh)
{
  SH2movli(sh, 5, (3<<4)+6);
}
#define init_13_5_3_7
void FASTCALL call_13_5_3_7(SH2_struct* sh)
{
  SH2movli(sh, 5, (3<<4)+7);
}
#define init_13_5_3_8
void FASTCALL call_13_5_3_8(SH2_struct* sh)
{
  SH2movli(sh, 5, (3<<4)+8);
}
#define init_13_5_3_9
void FASTCALL call_13_5_3_9(SH2_struct* sh)
{
  SH2movli(sh, 5, (3<<4)+9);
}
#define init_13_5_3_10
void FASTCALL call_13_5_3_10(SH2_struct* sh)
{
  SH2movli(sh, 5, (3<<4)+10);
}
#define init_13_5_3_11
void FASTCALL call_13_5_3_11(SH2_struct* sh)
{
  SH2movli(sh, 5, (3<<4)+11);
}
#define init_13_5_3_12
void FASTCALL call_13_5_3_12(SH2_struct* sh)
{
  SH2movli(sh, 5, (3<<4)+12);
}
#define init_13_5_3_13
void FASTCALL call_13_5_3_13(SH2_struct* sh)
{
  SH2movli(sh, 5, (3<<4)+13);
}
#define init_13_5_3_14
void FASTCALL call_13_5_3_14(SH2_struct* sh)
{
  SH2movli(sh, 5, (3<<4)+14);
}
#define init_13_5_3_15
void FASTCALL call_13_5_3_15(SH2_struct* sh)
{
  SH2movli(sh, 5, (3<<4)+15);
}
#define init_13_5_4_0
void FASTCALL call_13_5_4_0(SH2_struct* sh)
{
  SH2movli(sh, 5, (4<<4)+0);
}
#define init_13_5_4_1
void FASTCALL call_13_5_4_1(SH2_struct* sh)
{
  SH2movli(sh, 5, (4<<4)+1);
}
#define init_13_5_4_2
void FASTCALL call_13_5_4_2(SH2_struct* sh)
{
  SH2movli(sh, 5, (4<<4)+2);
}
#define init_13_5_4_3
void FASTCALL call_13_5_4_3(SH2_struct* sh)
{
  SH2movli(sh, 5, (4<<4)+3);
}
#define init_13_5_4_4
void FASTCALL call_13_5_4_4(SH2_struct* sh)
{
  SH2movli(sh, 5, (4<<4)+4);
}
#define init_13_5_4_5
void FASTCALL call_13_5_4_5(SH2_struct* sh)
{
  SH2movli(sh, 5, (4<<4)+5);
}
#define init_13_5_4_6
void FASTCALL call_13_5_4_6(SH2_struct* sh)
{
  SH2movli(sh, 5, (4<<4)+6);
}
#define init_13_5_4_7
void FASTCALL call_13_5_4_7(SH2_struct* sh)
{
  SH2movli(sh, 5, (4<<4)+7);
}
#define init_13_5_4_8
void FASTCALL call_13_5_4_8(SH2_struct* sh)
{
  SH2movli(sh, 5, (4<<4)+8);
}
#define init_13_5_4_9
void FASTCALL call_13_5_4_9(SH2_struct* sh)
{
  SH2movli(sh, 5, (4<<4)+9);
}
#define init_13_5_4_10
void FASTCALL call_13_5_4_10(SH2_struct* sh)
{
  SH2movli(sh, 5, (4<<4)+10);
}
#define init_13_5_4_11
void FASTCALL call_13_5_4_11(SH2_struct* sh)
{
  SH2movli(sh, 5, (4<<4)+11);
}
#define init_13_5_4_12
void FASTCALL call_13_5_4_12(SH2_struct* sh)
{
  SH2movli(sh, 5, (4<<4)+12);
}
#define init_13_5_4_13
void FASTCALL call_13_5_4_13(SH2_struct* sh)
{
  SH2movli(sh, 5, (4<<4)+13);
}
#define init_13_5_4_14
void FASTCALL call_13_5_4_14(SH2_struct* sh)
{
  SH2movli(sh, 5, (4<<4)+14);
}
#define init_13_5_4_15
void FASTCALL call_13_5_4_15(SH2_struct* sh)
{
  SH2movli(sh, 5, (4<<4)+15);
}
#define init_13_5_5_0
void FASTCALL call_13_5_5_0(SH2_struct* sh)
{
  SH2movli(sh, 5, (5<<4)+0);
}
#define init_13_5_5_1
void FASTCALL call_13_5_5_1(SH2_struct* sh)
{
  SH2movli(sh, 5, (5<<4)+1);
}
#define init_13_5_5_2
void FASTCALL call_13_5_5_2(SH2_struct* sh)
{
  SH2movli(sh, 5, (5<<4)+2);
}
#define init_13_5_5_3
void FASTCALL call_13_5_5_3(SH2_struct* sh)
{
  SH2movli(sh, 5, (5<<4)+3);
}
#define init_13_5_5_4
void FASTCALL call_13_5_5_4(SH2_struct* sh)
{
  SH2movli(sh, 5, (5<<4)+4);
}
#define init_13_5_5_5
void FASTCALL call_13_5_5_5(SH2_struct* sh)
{
  SH2movli(sh, 5, (5<<4)+5);
}
#define init_13_5_5_6
void FASTCALL call_13_5_5_6(SH2_struct* sh)
{
  SH2movli(sh, 5, (5<<4)+6);
}
#define init_13_5_5_7
void FASTCALL call_13_5_5_7(SH2_struct* sh)
{
  SH2movli(sh, 5, (5<<4)+7);
}
#define init_13_5_5_8
void FASTCALL call_13_5_5_8(SH2_struct* sh)
{
  SH2movli(sh, 5, (5<<4)+8);
}
#define init_13_5_5_9
void FASTCALL call_13_5_5_9(SH2_struct* sh)
{
  SH2movli(sh, 5, (5<<4)+9);
}
#define init_13_5_5_10
void FASTCALL call_13_5_5_10(SH2_struct* sh)
{
  SH2movli(sh, 5, (5<<4)+10);
}
#define init_13_5_5_11
void FASTCALL call_13_5_5_11(SH2_struct* sh)
{
  SH2movli(sh, 5, (5<<4)+11);
}
#define init_13_5_5_12
void FASTCALL call_13_5_5_12(SH2_struct* sh)
{
  SH2movli(sh, 5, (5<<4)+12);
}
#define init_13_5_5_13
void FASTCALL call_13_5_5_13(SH2_struct* sh)
{
  SH2movli(sh, 5, (5<<4)+13);
}
#define init_13_5_5_14
void FASTCALL call_13_5_5_14(SH2_struct* sh)
{
  SH2movli(sh, 5, (5<<4)+14);
}
#define init_13_5_5_15
void FASTCALL call_13_5_5_15(SH2_struct* sh)
{
  SH2movli(sh, 5, (5<<4)+15);
}
#define init_13_5_6_0
void FASTCALL call_13_5_6_0(SH2_struct* sh)
{
  SH2movli(sh, 5, (6<<4)+0);
}
#define init_13_5_6_1
void FASTCALL call_13_5_6_1(SH2_struct* sh)
{
  SH2movli(sh, 5, (6<<4)+1);
}
#define init_13_5_6_2
void FASTCALL call_13_5_6_2(SH2_struct* sh)
{
  SH2movli(sh, 5, (6<<4)+2);
}
#define init_13_5_6_3
void FASTCALL call_13_5_6_3(SH2_struct* sh)
{
  SH2movli(sh, 5, (6<<4)+3);
}
#define init_13_5_6_4
void FASTCALL call_13_5_6_4(SH2_struct* sh)
{
  SH2movli(sh, 5, (6<<4)+4);
}
#define init_13_5_6_5
void FASTCALL call_13_5_6_5(SH2_struct* sh)
{
  SH2movli(sh, 5, (6<<4)+5);
}
#define init_13_5_6_6
void FASTCALL call_13_5_6_6(SH2_struct* sh)
{
  SH2movli(sh, 5, (6<<4)+6);
}
#define init_13_5_6_7
void FASTCALL call_13_5_6_7(SH2_struct* sh)
{
  SH2movli(sh, 5, (6<<4)+7);
}
#define init_13_5_6_8
void FASTCALL call_13_5_6_8(SH2_struct* sh)
{
  SH2movli(sh, 5, (6<<4)+8);
}
#define init_13_5_6_9
void FASTCALL call_13_5_6_9(SH2_struct* sh)
{
  SH2movli(sh, 5, (6<<4)+9);
}
#define init_13_5_6_10
void FASTCALL call_13_5_6_10(SH2_struct* sh)
{
  SH2movli(sh, 5, (6<<4)+10);
}
#define init_13_5_6_11
void FASTCALL call_13_5_6_11(SH2_struct* sh)
{
  SH2movli(sh, 5, (6<<4)+11);
}
#define init_13_5_6_12
void FASTCALL call_13_5_6_12(SH2_struct* sh)
{
  SH2movli(sh, 5, (6<<4)+12);
}
#define init_13_5_6_13
void FASTCALL call_13_5_6_13(SH2_struct* sh)
{
  SH2movli(sh, 5, (6<<4)+13);
}
#define init_13_5_6_14
void FASTCALL call_13_5_6_14(SH2_struct* sh)
{
  SH2movli(sh, 5, (6<<4)+14);
}
#define init_13_5_6_15
void FASTCALL call_13_5_6_15(SH2_struct* sh)
{
  SH2movli(sh, 5, (6<<4)+15);
}
#define init_13_5_7_0
void FASTCALL call_13_5_7_0(SH2_struct* sh)
{
  SH2movli(sh, 5, (7<<4)+0);
}
#define init_13_5_7_1
void FASTCALL call_13_5_7_1(SH2_struct* sh)
{
  SH2movli(sh, 5, (7<<4)+1);
}
#define init_13_5_7_2
void FASTCALL call_13_5_7_2(SH2_struct* sh)
{
  SH2movli(sh, 5, (7<<4)+2);
}
#define init_13_5_7_3
void FASTCALL call_13_5_7_3(SH2_struct* sh)
{
  SH2movli(sh, 5, (7<<4)+3);
}
#define init_13_5_7_4
void FASTCALL call_13_5_7_4(SH2_struct* sh)
{
  SH2movli(sh, 5, (7<<4)+4);
}
#define init_13_5_7_5
void FASTCALL call_13_5_7_5(SH2_struct* sh)
{
  SH2movli(sh, 5, (7<<4)+5);
}
#define init_13_5_7_6
void FASTCALL call_13_5_7_6(SH2_struct* sh)
{
  SH2movli(sh, 5, (7<<4)+6);
}
#define init_13_5_7_7
void FASTCALL call_13_5_7_7(SH2_struct* sh)
{
  SH2movli(sh, 5, (7<<4)+7);
}
#define init_13_5_7_8
void FASTCALL call_13_5_7_8(SH2_struct* sh)
{
  SH2movli(sh, 5, (7<<4)+8);
}
#define init_13_5_7_9
void FASTCALL call_13_5_7_9(SH2_struct* sh)
{
  SH2movli(sh, 5, (7<<4)+9);
}
#define init_13_5_7_10
void FASTCALL call_13_5_7_10(SH2_struct* sh)
{
  SH2movli(sh, 5, (7<<4)+10);
}
#define init_13_5_7_11
void FASTCALL call_13_5_7_11(SH2_struct* sh)
{
  SH2movli(sh, 5, (7<<4)+11);
}
#define init_13_5_7_12
void FASTCALL call_13_5_7_12(SH2_struct* sh)
{
  SH2movli(sh, 5, (7<<4)+12);
}
#define init_13_5_7_13
void FASTCALL call_13_5_7_13(SH2_struct* sh)
{
  SH2movli(sh, 5, (7<<4)+13);
}
#define init_13_5_7_14
void FASTCALL call_13_5_7_14(SH2_struct* sh)
{
  SH2movli(sh, 5, (7<<4)+14);
}
#define init_13_5_7_15
void FASTCALL call_13_5_7_15(SH2_struct* sh)
{
  SH2movli(sh, 5, (7<<4)+15);
}
#define init_13_5_8_0
void FASTCALL call_13_5_8_0(SH2_struct* sh)
{
  SH2movli(sh, 5, (8<<4)+0);
}
#define init_13_5_8_1
void FASTCALL call_13_5_8_1(SH2_struct* sh)
{
  SH2movli(sh, 5, (8<<4)+1);
}
#define init_13_5_8_2
void FASTCALL call_13_5_8_2(SH2_struct* sh)
{
  SH2movli(sh, 5, (8<<4)+2);
}
#define init_13_5_8_3
void FASTCALL call_13_5_8_3(SH2_struct* sh)
{
  SH2movli(sh, 5, (8<<4)+3);
}
#define init_13_5_8_4
void FASTCALL call_13_5_8_4(SH2_struct* sh)
{
  SH2movli(sh, 5, (8<<4)+4);
}
#define init_13_5_8_5
void FASTCALL call_13_5_8_5(SH2_struct* sh)
{
  SH2movli(sh, 5, (8<<4)+5);
}
#define init_13_5_8_6
void FASTCALL call_13_5_8_6(SH2_struct* sh)
{
  SH2movli(sh, 5, (8<<4)+6);
}
#define init_13_5_8_7
void FASTCALL call_13_5_8_7(SH2_struct* sh)
{
  SH2movli(sh, 5, (8<<4)+7);
}
#define init_13_5_8_8
void FASTCALL call_13_5_8_8(SH2_struct* sh)
{
  SH2movli(sh, 5, (8<<4)+8);
}
#define init_13_5_8_9
void FASTCALL call_13_5_8_9(SH2_struct* sh)
{
  SH2movli(sh, 5, (8<<4)+9);
}
#define init_13_5_8_10
void FASTCALL call_13_5_8_10(SH2_struct* sh)
{
  SH2movli(sh, 5, (8<<4)+10);
}
#define init_13_5_8_11
void FASTCALL call_13_5_8_11(SH2_struct* sh)
{
  SH2movli(sh, 5, (8<<4)+11);
}
#define init_13_5_8_12
void FASTCALL call_13_5_8_12(SH2_struct* sh)
{
  SH2movli(sh, 5, (8<<4)+12);
}
#define init_13_5_8_13
void FASTCALL call_13_5_8_13(SH2_struct* sh)
{
  SH2movli(sh, 5, (8<<4)+13);
}
#define init_13_5_8_14
void FASTCALL call_13_5_8_14(SH2_struct* sh)
{
  SH2movli(sh, 5, (8<<4)+14);
}
#define init_13_5_8_15
void FASTCALL call_13_5_8_15(SH2_struct* sh)
{
  SH2movli(sh, 5, (8<<4)+15);
}
#define init_13_5_9_0
void FASTCALL call_13_5_9_0(SH2_struct* sh)
{
  SH2movli(sh, 5, (9<<4)+0);
}
#define init_13_5_9_1
void FASTCALL call_13_5_9_1(SH2_struct* sh)
{
  SH2movli(sh, 5, (9<<4)+1);
}
#define init_13_5_9_2
void FASTCALL call_13_5_9_2(SH2_struct* sh)
{
  SH2movli(sh, 5, (9<<4)+2);
}
#define init_13_5_9_3
void FASTCALL call_13_5_9_3(SH2_struct* sh)
{
  SH2movli(sh, 5, (9<<4)+3);
}
#define init_13_5_9_4
void FASTCALL call_13_5_9_4(SH2_struct* sh)
{
  SH2movli(sh, 5, (9<<4)+4);
}
#define init_13_5_9_5
void FASTCALL call_13_5_9_5(SH2_struct* sh)
{
  SH2movli(sh, 5, (9<<4)+5);
}
#define init_13_5_9_6
void FASTCALL call_13_5_9_6(SH2_struct* sh)
{
  SH2movli(sh, 5, (9<<4)+6);
}
#define init_13_5_9_7
void FASTCALL call_13_5_9_7(SH2_struct* sh)
{
  SH2movli(sh, 5, (9<<4)+7);
}
#define init_13_5_9_8
void FASTCALL call_13_5_9_8(SH2_struct* sh)
{
  SH2movli(sh, 5, (9<<4)+8);
}
#define init_13_5_9_9
void FASTCALL call_13_5_9_9(SH2_struct* sh)
{
  SH2movli(sh, 5, (9<<4)+9);
}
#define init_13_5_9_10
void FASTCALL call_13_5_9_10(SH2_struct* sh)
{
  SH2movli(sh, 5, (9<<4)+10);
}
#define init_13_5_9_11
void FASTCALL call_13_5_9_11(SH2_struct* sh)
{
  SH2movli(sh, 5, (9<<4)+11);
}
#define init_13_5_9_12
void FASTCALL call_13_5_9_12(SH2_struct* sh)
{
  SH2movli(sh, 5, (9<<4)+12);
}
#define init_13_5_9_13
void FASTCALL call_13_5_9_13(SH2_struct* sh)
{
  SH2movli(sh, 5, (9<<4)+13);
}
#define init_13_5_9_14
void FASTCALL call_13_5_9_14(SH2_struct* sh)
{
  SH2movli(sh, 5, (9<<4)+14);
}
#define init_13_5_9_15
void FASTCALL call_13_5_9_15(SH2_struct* sh)
{
  SH2movli(sh, 5, (9<<4)+15);
}
#define init_13_5_10_0
void FASTCALL call_13_5_10_0(SH2_struct* sh)
{
  SH2movli(sh, 5, (10<<4)+0);
}
#define init_13_5_10_1
void FASTCALL call_13_5_10_1(SH2_struct* sh)
{
  SH2movli(sh, 5, (10<<4)+1);
}
#define init_13_5_10_2
void FASTCALL call_13_5_10_2(SH2_struct* sh)
{
  SH2movli(sh, 5, (10<<4)+2);
}
#define init_13_5_10_3
void FASTCALL call_13_5_10_3(SH2_struct* sh)
{
  SH2movli(sh, 5, (10<<4)+3);
}
#define init_13_5_10_4
void FASTCALL call_13_5_10_4(SH2_struct* sh)
{
  SH2movli(sh, 5, (10<<4)+4);
}
#define init_13_5_10_5
void FASTCALL call_13_5_10_5(SH2_struct* sh)
{
  SH2movli(sh, 5, (10<<4)+5);
}
#define init_13_5_10_6
void FASTCALL call_13_5_10_6(SH2_struct* sh)
{
  SH2movli(sh, 5, (10<<4)+6);
}
#define init_13_5_10_7
void FASTCALL call_13_5_10_7(SH2_struct* sh)
{
  SH2movli(sh, 5, (10<<4)+7);
}
#define init_13_5_10_8
void FASTCALL call_13_5_10_8(SH2_struct* sh)
{
  SH2movli(sh, 5, (10<<4)+8);
}
#define init_13_5_10_9
void FASTCALL call_13_5_10_9(SH2_struct* sh)
{
  SH2movli(sh, 5, (10<<4)+9);
}
#define init_13_5_10_10
void FASTCALL call_13_5_10_10(SH2_struct* sh)
{
  SH2movli(sh, 5, (10<<4)+10);
}
#define init_13_5_10_11
void FASTCALL call_13_5_10_11(SH2_struct* sh)
{
  SH2movli(sh, 5, (10<<4)+11);
}
#define init_13_5_10_12
void FASTCALL call_13_5_10_12(SH2_struct* sh)
{
  SH2movli(sh, 5, (10<<4)+12);
}
#define init_13_5_10_13
void FASTCALL call_13_5_10_13(SH2_struct* sh)
{
  SH2movli(sh, 5, (10<<4)+13);
}
#define init_13_5_10_14
void FASTCALL call_13_5_10_14(SH2_struct* sh)
{
  SH2movli(sh, 5, (10<<4)+14);
}
#define init_13_5_10_15
void FASTCALL call_13_5_10_15(SH2_struct* sh)
{
  SH2movli(sh, 5, (10<<4)+15);
}
#define init_13_5_11_0
void FASTCALL call_13_5_11_0(SH2_struct* sh)
{
  SH2movli(sh, 5, (11<<4)+0);
}
#define init_13_5_11_1
void FASTCALL call_13_5_11_1(SH2_struct* sh)
{
  SH2movli(sh, 5, (11<<4)+1);
}
#define init_13_5_11_2
void FASTCALL call_13_5_11_2(SH2_struct* sh)
{
  SH2movli(sh, 5, (11<<4)+2);
}
#define init_13_5_11_3
void FASTCALL call_13_5_11_3(SH2_struct* sh)
{
  SH2movli(sh, 5, (11<<4)+3);
}
#define init_13_5_11_4
void FASTCALL call_13_5_11_4(SH2_struct* sh)
{
  SH2movli(sh, 5, (11<<4)+4);
}
#define init_13_5_11_5
void FASTCALL call_13_5_11_5(SH2_struct* sh)
{
  SH2movli(sh, 5, (11<<4)+5);
}
#define init_13_5_11_6
void FASTCALL call_13_5_11_6(SH2_struct* sh)
{
  SH2movli(sh, 5, (11<<4)+6);
}
#define init_13_5_11_7
void FASTCALL call_13_5_11_7(SH2_struct* sh)
{
  SH2movli(sh, 5, (11<<4)+7);
}
#define init_13_5_11_8
void FASTCALL call_13_5_11_8(SH2_struct* sh)
{
  SH2movli(sh, 5, (11<<4)+8);
}
#define init_13_5_11_9
void FASTCALL call_13_5_11_9(SH2_struct* sh)
{
  SH2movli(sh, 5, (11<<4)+9);
}
#define init_13_5_11_10
void FASTCALL call_13_5_11_10(SH2_struct* sh)
{
  SH2movli(sh, 5, (11<<4)+10);
}
#define init_13_5_11_11
void FASTCALL call_13_5_11_11(SH2_struct* sh)
{
  SH2movli(sh, 5, (11<<4)+11);
}
#define init_13_5_11_12
void FASTCALL call_13_5_11_12(SH2_struct* sh)
{
  SH2movli(sh, 5, (11<<4)+12);
}
#define init_13_5_11_13
void FASTCALL call_13_5_11_13(SH2_struct* sh)
{
  SH2movli(sh, 5, (11<<4)+13);
}
#define init_13_5_11_14
void FASTCALL call_13_5_11_14(SH2_struct* sh)
{
  SH2movli(sh, 5, (11<<4)+14);
}
#define init_13_5_11_15
void FASTCALL call_13_5_11_15(SH2_struct* sh)
{
  SH2movli(sh, 5, (11<<4)+15);
}
#define init_13_5_12_0
void FASTCALL call_13_5_12_0(SH2_struct* sh)
{
  SH2movli(sh, 5, (12<<4)+0);
}
#define init_13_5_12_1
void FASTCALL call_13_5_12_1(SH2_struct* sh)
{
  SH2movli(sh, 5, (12<<4)+1);
}
#define init_13_5_12_2
void FASTCALL call_13_5_12_2(SH2_struct* sh)
{
  SH2movli(sh, 5, (12<<4)+2);
}
#define init_13_5_12_3
void FASTCALL call_13_5_12_3(SH2_struct* sh)
{
  SH2movli(sh, 5, (12<<4)+3);
}
#define init_13_5_12_4
void FASTCALL call_13_5_12_4(SH2_struct* sh)
{
  SH2movli(sh, 5, (12<<4)+4);
}
#define init_13_5_12_5
void FASTCALL call_13_5_12_5(SH2_struct* sh)
{
  SH2movli(sh, 5, (12<<4)+5);
}
#define init_13_5_12_6
void FASTCALL call_13_5_12_6(SH2_struct* sh)
{
  SH2movli(sh, 5, (12<<4)+6);
}
#define init_13_5_12_7
void FASTCALL call_13_5_12_7(SH2_struct* sh)
{
  SH2movli(sh, 5, (12<<4)+7);
}
#define init_13_5_12_8
void FASTCALL call_13_5_12_8(SH2_struct* sh)
{
  SH2movli(sh, 5, (12<<4)+8);
}
#define init_13_5_12_9
void FASTCALL call_13_5_12_9(SH2_struct* sh)
{
  SH2movli(sh, 5, (12<<4)+9);
}
#define init_13_5_12_10
void FASTCALL call_13_5_12_10(SH2_struct* sh)
{
  SH2movli(sh, 5, (12<<4)+10);
}
#define init_13_5_12_11
void FASTCALL call_13_5_12_11(SH2_struct* sh)
{
  SH2movli(sh, 5, (12<<4)+11);
}
#define init_13_5_12_12
void FASTCALL call_13_5_12_12(SH2_struct* sh)
{
  SH2movli(sh, 5, (12<<4)+12);
}
#define init_13_5_12_13
void FASTCALL call_13_5_12_13(SH2_struct* sh)
{
  SH2movli(sh, 5, (12<<4)+13);
}
#define init_13_5_12_14
void FASTCALL call_13_5_12_14(SH2_struct* sh)
{
  SH2movli(sh, 5, (12<<4)+14);
}
#define init_13_5_12_15
void FASTCALL call_13_5_12_15(SH2_struct* sh)
{
  SH2movli(sh, 5, (12<<4)+15);
}
#define init_13_5_13_0
void FASTCALL call_13_5_13_0(SH2_struct* sh)
{
  SH2movli(sh, 5, (13<<4)+0);
}
#define init_13_5_13_1
void FASTCALL call_13_5_13_1(SH2_struct* sh)
{
  SH2movli(sh, 5, (13<<4)+1);
}
#define init_13_5_13_2
void FASTCALL call_13_5_13_2(SH2_struct* sh)
{
  SH2movli(sh, 5, (13<<4)+2);
}
#define init_13_5_13_3
void FASTCALL call_13_5_13_3(SH2_struct* sh)
{
  SH2movli(sh, 5, (13<<4)+3);
}
#define init_13_5_13_4
void FASTCALL call_13_5_13_4(SH2_struct* sh)
{
  SH2movli(sh, 5, (13<<4)+4);
}
#define init_13_5_13_5
void FASTCALL call_13_5_13_5(SH2_struct* sh)
{
  SH2movli(sh, 5, (13<<4)+5);
}
#define init_13_5_13_6
void FASTCALL call_13_5_13_6(SH2_struct* sh)
{
  SH2movli(sh, 5, (13<<4)+6);
}
#define init_13_5_13_7
void FASTCALL call_13_5_13_7(SH2_struct* sh)
{
  SH2movli(sh, 5, (13<<4)+7);
}
#define init_13_5_13_8
void FASTCALL call_13_5_13_8(SH2_struct* sh)
{
  SH2movli(sh, 5, (13<<4)+8);
}
#define init_13_5_13_9
void FASTCALL call_13_5_13_9(SH2_struct* sh)
{
  SH2movli(sh, 5, (13<<4)+9);
}
#define init_13_5_13_10
void FASTCALL call_13_5_13_10(SH2_struct* sh)
{
  SH2movli(sh, 5, (13<<4)+10);
}
#define init_13_5_13_11
void FASTCALL call_13_5_13_11(SH2_struct* sh)
{
  SH2movli(sh, 5, (13<<4)+11);
}
#define init_13_5_13_12
void FASTCALL call_13_5_13_12(SH2_struct* sh)
{
  SH2movli(sh, 5, (13<<4)+12);
}
#define init_13_5_13_13
void FASTCALL call_13_5_13_13(SH2_struct* sh)
{
  SH2movli(sh, 5, (13<<4)+13);
}
#define init_13_5_13_14
void FASTCALL call_13_5_13_14(SH2_struct* sh)
{
  SH2movli(sh, 5, (13<<4)+14);
}
#define init_13_5_13_15
void FASTCALL call_13_5_13_15(SH2_struct* sh)
{
  SH2movli(sh, 5, (13<<4)+15);
}
#define init_13_5_14_0
void FASTCALL call_13_5_14_0(SH2_struct* sh)
{
  SH2movli(sh, 5, (14<<4)+0);
}
#define init_13_5_14_1
void FASTCALL call_13_5_14_1(SH2_struct* sh)
{
  SH2movli(sh, 5, (14<<4)+1);
}
#define init_13_5_14_2
void FASTCALL call_13_5_14_2(SH2_struct* sh)
{
  SH2movli(sh, 5, (14<<4)+2);
}
#define init_13_5_14_3
void FASTCALL call_13_5_14_3(SH2_struct* sh)
{
  SH2movli(sh, 5, (14<<4)+3);
}
#define init_13_5_14_4
void FASTCALL call_13_5_14_4(SH2_struct* sh)
{
  SH2movli(sh, 5, (14<<4)+4);
}
#define init_13_5_14_5
void FASTCALL call_13_5_14_5(SH2_struct* sh)
{
  SH2movli(sh, 5, (14<<4)+5);
}
#define init_13_5_14_6
void FASTCALL call_13_5_14_6(SH2_struct* sh)
{
  SH2movli(sh, 5, (14<<4)+6);
}
#define init_13_5_14_7
void FASTCALL call_13_5_14_7(SH2_struct* sh)
{
  SH2movli(sh, 5, (14<<4)+7);
}
#define init_13_5_14_8
void FASTCALL call_13_5_14_8(SH2_struct* sh)
{
  SH2movli(sh, 5, (14<<4)+8);
}
#define init_13_5_14_9
void FASTCALL call_13_5_14_9(SH2_struct* sh)
{
  SH2movli(sh, 5, (14<<4)+9);
}
#define init_13_5_14_10
void FASTCALL call_13_5_14_10(SH2_struct* sh)
{
  SH2movli(sh, 5, (14<<4)+10);
}
#define init_13_5_14_11
void FASTCALL call_13_5_14_11(SH2_struct* sh)
{
  SH2movli(sh, 5, (14<<4)+11);
}
#define init_13_5_14_12
void FASTCALL call_13_5_14_12(SH2_struct* sh)
{
  SH2movli(sh, 5, (14<<4)+12);
}
#define init_13_5_14_13
void FASTCALL call_13_5_14_13(SH2_struct* sh)
{
  SH2movli(sh, 5, (14<<4)+13);
}
#define init_13_5_14_14
void FASTCALL call_13_5_14_14(SH2_struct* sh)
{
  SH2movli(sh, 5, (14<<4)+14);
}
#define init_13_5_14_15
void FASTCALL call_13_5_14_15(SH2_struct* sh)
{
  SH2movli(sh, 5, (14<<4)+15);
}
#define init_13_5_15_0
void FASTCALL call_13_5_15_0(SH2_struct* sh)
{
  SH2movli(sh, 5, (15<<4)+0);
}
#define init_13_5_15_1
void FASTCALL call_13_5_15_1(SH2_struct* sh)
{
  SH2movli(sh, 5, (15<<4)+1);
}
#define init_13_5_15_2
void FASTCALL call_13_5_15_2(SH2_struct* sh)
{
  SH2movli(sh, 5, (15<<4)+2);
}
#define init_13_5_15_3
void FASTCALL call_13_5_15_3(SH2_struct* sh)
{
  SH2movli(sh, 5, (15<<4)+3);
}
#define init_13_5_15_4
void FASTCALL call_13_5_15_4(SH2_struct* sh)
{
  SH2movli(sh, 5, (15<<4)+4);
}
#define init_13_5_15_5
void FASTCALL call_13_5_15_5(SH2_struct* sh)
{
  SH2movli(sh, 5, (15<<4)+5);
}
#define init_13_5_15_6
void FASTCALL call_13_5_15_6(SH2_struct* sh)
{
  SH2movli(sh, 5, (15<<4)+6);
}
#define init_13_5_15_7
void FASTCALL call_13_5_15_7(SH2_struct* sh)
{
  SH2movli(sh, 5, (15<<4)+7);
}
#define init_13_5_15_8
void FASTCALL call_13_5_15_8(SH2_struct* sh)
{
  SH2movli(sh, 5, (15<<4)+8);
}
#define init_13_5_15_9
void FASTCALL call_13_5_15_9(SH2_struct* sh)
{
  SH2movli(sh, 5, (15<<4)+9);
}
#define init_13_5_15_10
void FASTCALL call_13_5_15_10(SH2_struct* sh)
{
  SH2movli(sh, 5, (15<<4)+10);
}
#define init_13_5_15_11
void FASTCALL call_13_5_15_11(SH2_struct* sh)
{
  SH2movli(sh, 5, (15<<4)+11);
}
#define init_13_5_15_12
void FASTCALL call_13_5_15_12(SH2_struct* sh)
{
  SH2movli(sh, 5, (15<<4)+12);
}
#define init_13_5_15_13
void FASTCALL call_13_5_15_13(SH2_struct* sh)
{
  SH2movli(sh, 5, (15<<4)+13);
}
#define init_13_5_15_14
void FASTCALL call_13_5_15_14(SH2_struct* sh)
{
  SH2movli(sh, 5, (15<<4)+14);
}
#define init_13_5_15_15
void FASTCALL call_13_5_15_15(SH2_struct* sh)
{
  SH2movli(sh, 5, (15<<4)+15);
}
#define init_13_6_0_0
void FASTCALL call_13_6_0_0(SH2_struct* sh)
{
  SH2movli(sh, 6, (0<<4)+0);
}
#define init_13_6_0_1
void FASTCALL call_13_6_0_1(SH2_struct* sh)
{
  SH2movli(sh, 6, (0<<4)+1);
}
#define init_13_6_0_2
void FASTCALL call_13_6_0_2(SH2_struct* sh)
{
  SH2movli(sh, 6, (0<<4)+2);
}
#define init_13_6_0_3
void FASTCALL call_13_6_0_3(SH2_struct* sh)
{
  SH2movli(sh, 6, (0<<4)+3);
}
#define init_13_6_0_4
void FASTCALL call_13_6_0_4(SH2_struct* sh)
{
  SH2movli(sh, 6, (0<<4)+4);
}
#define init_13_6_0_5
void FASTCALL call_13_6_0_5(SH2_struct* sh)
{
  SH2movli(sh, 6, (0<<4)+5);
}
#define init_13_6_0_6
void FASTCALL call_13_6_0_6(SH2_struct* sh)
{
  SH2movli(sh, 6, (0<<4)+6);
}
#define init_13_6_0_7
void FASTCALL call_13_6_0_7(SH2_struct* sh)
{
  SH2movli(sh, 6, (0<<4)+7);
}
#define init_13_6_0_8
void FASTCALL call_13_6_0_8(SH2_struct* sh)
{
  SH2movli(sh, 6, (0<<4)+8);
}
#define init_13_6_0_9
void FASTCALL call_13_6_0_9(SH2_struct* sh)
{
  SH2movli(sh, 6, (0<<4)+9);
}
#define init_13_6_0_10
void FASTCALL call_13_6_0_10(SH2_struct* sh)
{
  SH2movli(sh, 6, (0<<4)+10);
}
#define init_13_6_0_11
void FASTCALL call_13_6_0_11(SH2_struct* sh)
{
  SH2movli(sh, 6, (0<<4)+11);
}
#define init_13_6_0_12
void FASTCALL call_13_6_0_12(SH2_struct* sh)
{
  SH2movli(sh, 6, (0<<4)+12);
}
#define init_13_6_0_13
void FASTCALL call_13_6_0_13(SH2_struct* sh)
{
  SH2movli(sh, 6, (0<<4)+13);
}
#define init_13_6_0_14
void FASTCALL call_13_6_0_14(SH2_struct* sh)
{
  SH2movli(sh, 6, (0<<4)+14);
}
#define init_13_6_0_15
void FASTCALL call_13_6_0_15(SH2_struct* sh)
{
  SH2movli(sh, 6, (0<<4)+15);
}
#define init_13_6_1_0
void FASTCALL call_13_6_1_0(SH2_struct* sh)
{
  SH2movli(sh, 6, (1<<4)+0);
}
#define init_13_6_1_1
void FASTCALL call_13_6_1_1(SH2_struct* sh)
{
  SH2movli(sh, 6, (1<<4)+1);
}
#define init_13_6_1_2
void FASTCALL call_13_6_1_2(SH2_struct* sh)
{
  SH2movli(sh, 6, (1<<4)+2);
}
#define init_13_6_1_3
void FASTCALL call_13_6_1_3(SH2_struct* sh)
{
  SH2movli(sh, 6, (1<<4)+3);
}
#define init_13_6_1_4
void FASTCALL call_13_6_1_4(SH2_struct* sh)
{
  SH2movli(sh, 6, (1<<4)+4);
}
#define init_13_6_1_5
void FASTCALL call_13_6_1_5(SH2_struct* sh)
{
  SH2movli(sh, 6, (1<<4)+5);
}
#define init_13_6_1_6
void FASTCALL call_13_6_1_6(SH2_struct* sh)
{
  SH2movli(sh, 6, (1<<4)+6);
}
#define init_13_6_1_7
void FASTCALL call_13_6_1_7(SH2_struct* sh)
{
  SH2movli(sh, 6, (1<<4)+7);
}
#define init_13_6_1_8
void FASTCALL call_13_6_1_8(SH2_struct* sh)
{
  SH2movli(sh, 6, (1<<4)+8);
}
#define init_13_6_1_9
void FASTCALL call_13_6_1_9(SH2_struct* sh)
{
  SH2movli(sh, 6, (1<<4)+9);
}
#define init_13_6_1_10
void FASTCALL call_13_6_1_10(SH2_struct* sh)
{
  SH2movli(sh, 6, (1<<4)+10);
}
#define init_13_6_1_11
void FASTCALL call_13_6_1_11(SH2_struct* sh)
{
  SH2movli(sh, 6, (1<<4)+11);
}
#define init_13_6_1_12
void FASTCALL call_13_6_1_12(SH2_struct* sh)
{
  SH2movli(sh, 6, (1<<4)+12);
}
#define init_13_6_1_13
void FASTCALL call_13_6_1_13(SH2_struct* sh)
{
  SH2movli(sh, 6, (1<<4)+13);
}
#define init_13_6_1_14
void FASTCALL call_13_6_1_14(SH2_struct* sh)
{
  SH2movli(sh, 6, (1<<4)+14);
}
#define init_13_6_1_15
void FASTCALL call_13_6_1_15(SH2_struct* sh)
{
  SH2movli(sh, 6, (1<<4)+15);
}
#define init_13_6_2_0
void FASTCALL call_13_6_2_0(SH2_struct* sh)
{
  SH2movli(sh, 6, (2<<4)+0);
}
#define init_13_6_2_1
void FASTCALL call_13_6_2_1(SH2_struct* sh)
{
  SH2movli(sh, 6, (2<<4)+1);
}
#define init_13_6_2_2
void FASTCALL call_13_6_2_2(SH2_struct* sh)
{
  SH2movli(sh, 6, (2<<4)+2);
}
#define init_13_6_2_3
void FASTCALL call_13_6_2_3(SH2_struct* sh)
{
  SH2movli(sh, 6, (2<<4)+3);
}
#define init_13_6_2_4
void FASTCALL call_13_6_2_4(SH2_struct* sh)
{
  SH2movli(sh, 6, (2<<4)+4);
}
#define init_13_6_2_5
void FASTCALL call_13_6_2_5(SH2_struct* sh)
{
  SH2movli(sh, 6, (2<<4)+5);
}
#define init_13_6_2_6
void FASTCALL call_13_6_2_6(SH2_struct* sh)
{
  SH2movli(sh, 6, (2<<4)+6);
}
#define init_13_6_2_7
void FASTCALL call_13_6_2_7(SH2_struct* sh)
{
  SH2movli(sh, 6, (2<<4)+7);
}
#define init_13_6_2_8
void FASTCALL call_13_6_2_8(SH2_struct* sh)
{
  SH2movli(sh, 6, (2<<4)+8);
}
#define init_13_6_2_9
void FASTCALL call_13_6_2_9(SH2_struct* sh)
{
  SH2movli(sh, 6, (2<<4)+9);
}
#define init_13_6_2_10
void FASTCALL call_13_6_2_10(SH2_struct* sh)
{
  SH2movli(sh, 6, (2<<4)+10);
}
#define init_13_6_2_11
void FASTCALL call_13_6_2_11(SH2_struct* sh)
{
  SH2movli(sh, 6, (2<<4)+11);
}
#define init_13_6_2_12
void FASTCALL call_13_6_2_12(SH2_struct* sh)
{
  SH2movli(sh, 6, (2<<4)+12);
}
#define init_13_6_2_13
void FASTCALL call_13_6_2_13(SH2_struct* sh)
{
  SH2movli(sh, 6, (2<<4)+13);
}
#define init_13_6_2_14
void FASTCALL call_13_6_2_14(SH2_struct* sh)
{
  SH2movli(sh, 6, (2<<4)+14);
}
#define init_13_6_2_15
void FASTCALL call_13_6_2_15(SH2_struct* sh)
{
  SH2movli(sh, 6, (2<<4)+15);
}
#define init_13_6_3_0
void FASTCALL call_13_6_3_0(SH2_struct* sh)
{
  SH2movli(sh, 6, (3<<4)+0);
}
#define init_13_6_3_1
void FASTCALL call_13_6_3_1(SH2_struct* sh)
{
  SH2movli(sh, 6, (3<<4)+1);
}
#define init_13_6_3_2
void FASTCALL call_13_6_3_2(SH2_struct* sh)
{
  SH2movli(sh, 6, (3<<4)+2);
}
#define init_13_6_3_3
void FASTCALL call_13_6_3_3(SH2_struct* sh)
{
  SH2movli(sh, 6, (3<<4)+3);
}
#define init_13_6_3_4
void FASTCALL call_13_6_3_4(SH2_struct* sh)
{
  SH2movli(sh, 6, (3<<4)+4);
}
#define init_13_6_3_5
void FASTCALL call_13_6_3_5(SH2_struct* sh)
{
  SH2movli(sh, 6, (3<<4)+5);
}
#define init_13_6_3_6
void FASTCALL call_13_6_3_6(SH2_struct* sh)
{
  SH2movli(sh, 6, (3<<4)+6);
}
#define init_13_6_3_7
void FASTCALL call_13_6_3_7(SH2_struct* sh)
{
  SH2movli(sh, 6, (3<<4)+7);
}
#define init_13_6_3_8
void FASTCALL call_13_6_3_8(SH2_struct* sh)
{
  SH2movli(sh, 6, (3<<4)+8);
}
#define init_13_6_3_9
void FASTCALL call_13_6_3_9(SH2_struct* sh)
{
  SH2movli(sh, 6, (3<<4)+9);
}
#define init_13_6_3_10
void FASTCALL call_13_6_3_10(SH2_struct* sh)
{
  SH2movli(sh, 6, (3<<4)+10);
}
#define init_13_6_3_11
void FASTCALL call_13_6_3_11(SH2_struct* sh)
{
  SH2movli(sh, 6, (3<<4)+11);
}
#define init_13_6_3_12
void FASTCALL call_13_6_3_12(SH2_struct* sh)
{
  SH2movli(sh, 6, (3<<4)+12);
}
#define init_13_6_3_13
void FASTCALL call_13_6_3_13(SH2_struct* sh)
{
  SH2movli(sh, 6, (3<<4)+13);
}
#define init_13_6_3_14
void FASTCALL call_13_6_3_14(SH2_struct* sh)
{
  SH2movli(sh, 6, (3<<4)+14);
}
#define init_13_6_3_15
void FASTCALL call_13_6_3_15(SH2_struct* sh)
{
  SH2movli(sh, 6, (3<<4)+15);
}
#define init_13_6_4_0
void FASTCALL call_13_6_4_0(SH2_struct* sh)
{
  SH2movli(sh, 6, (4<<4)+0);
}
#define init_13_6_4_1
void FASTCALL call_13_6_4_1(SH2_struct* sh)
{
  SH2movli(sh, 6, (4<<4)+1);
}
#define init_13_6_4_2
void FASTCALL call_13_6_4_2(SH2_struct* sh)
{
  SH2movli(sh, 6, (4<<4)+2);
}
#define init_13_6_4_3
void FASTCALL call_13_6_4_3(SH2_struct* sh)
{
  SH2movli(sh, 6, (4<<4)+3);
}
#define init_13_6_4_4
void FASTCALL call_13_6_4_4(SH2_struct* sh)
{
  SH2movli(sh, 6, (4<<4)+4);
}
#define init_13_6_4_5
void FASTCALL call_13_6_4_5(SH2_struct* sh)
{
  SH2movli(sh, 6, (4<<4)+5);
}
#define init_13_6_4_6
void FASTCALL call_13_6_4_6(SH2_struct* sh)
{
  SH2movli(sh, 6, (4<<4)+6);
}
#define init_13_6_4_7
void FASTCALL call_13_6_4_7(SH2_struct* sh)
{
  SH2movli(sh, 6, (4<<4)+7);
}
#define init_13_6_4_8
void FASTCALL call_13_6_4_8(SH2_struct* sh)
{
  SH2movli(sh, 6, (4<<4)+8);
}
#define init_13_6_4_9
void FASTCALL call_13_6_4_9(SH2_struct* sh)
{
  SH2movli(sh, 6, (4<<4)+9);
}
#define init_13_6_4_10
void FASTCALL call_13_6_4_10(SH2_struct* sh)
{
  SH2movli(sh, 6, (4<<4)+10);
}
#define init_13_6_4_11
void FASTCALL call_13_6_4_11(SH2_struct* sh)
{
  SH2movli(sh, 6, (4<<4)+11);
}
#define init_13_6_4_12
void FASTCALL call_13_6_4_12(SH2_struct* sh)
{
  SH2movli(sh, 6, (4<<4)+12);
}
#define init_13_6_4_13
void FASTCALL call_13_6_4_13(SH2_struct* sh)
{
  SH2movli(sh, 6, (4<<4)+13);
}
#define init_13_6_4_14
void FASTCALL call_13_6_4_14(SH2_struct* sh)
{
  SH2movli(sh, 6, (4<<4)+14);
}
#define init_13_6_4_15
void FASTCALL call_13_6_4_15(SH2_struct* sh)
{
  SH2movli(sh, 6, (4<<4)+15);
}
#define init_13_6_5_0
void FASTCALL call_13_6_5_0(SH2_struct* sh)
{
  SH2movli(sh, 6, (5<<4)+0);
}
#define init_13_6_5_1
void FASTCALL call_13_6_5_1(SH2_struct* sh)
{
  SH2movli(sh, 6, (5<<4)+1);
}
#define init_13_6_5_2
void FASTCALL call_13_6_5_2(SH2_struct* sh)
{
  SH2movli(sh, 6, (5<<4)+2);
}
#define init_13_6_5_3
void FASTCALL call_13_6_5_3(SH2_struct* sh)
{
  SH2movli(sh, 6, (5<<4)+3);
}
#define init_13_6_5_4
void FASTCALL call_13_6_5_4(SH2_struct* sh)
{
  SH2movli(sh, 6, (5<<4)+4);
}
#define init_13_6_5_5
void FASTCALL call_13_6_5_5(SH2_struct* sh)
{
  SH2movli(sh, 6, (5<<4)+5);
}
#define init_13_6_5_6
void FASTCALL call_13_6_5_6(SH2_struct* sh)
{
  SH2movli(sh, 6, (5<<4)+6);
}
#define init_13_6_5_7
void FASTCALL call_13_6_5_7(SH2_struct* sh)
{
  SH2movli(sh, 6, (5<<4)+7);
}
#define init_13_6_5_8
void FASTCALL call_13_6_5_8(SH2_struct* sh)
{
  SH2movli(sh, 6, (5<<4)+8);
}
#define init_13_6_5_9
void FASTCALL call_13_6_5_9(SH2_struct* sh)
{
  SH2movli(sh, 6, (5<<4)+9);
}
#define init_13_6_5_10
void FASTCALL call_13_6_5_10(SH2_struct* sh)
{
  SH2movli(sh, 6, (5<<4)+10);
}
#define init_13_6_5_11
void FASTCALL call_13_6_5_11(SH2_struct* sh)
{
  SH2movli(sh, 6, (5<<4)+11);
}
#define init_13_6_5_12
void FASTCALL call_13_6_5_12(SH2_struct* sh)
{
  SH2movli(sh, 6, (5<<4)+12);
}
#define init_13_6_5_13
void FASTCALL call_13_6_5_13(SH2_struct* sh)
{
  SH2movli(sh, 6, (5<<4)+13);
}
#define init_13_6_5_14
void FASTCALL call_13_6_5_14(SH2_struct* sh)
{
  SH2movli(sh, 6, (5<<4)+14);
}
#define init_13_6_5_15
void FASTCALL call_13_6_5_15(SH2_struct* sh)
{
  SH2movli(sh, 6, (5<<4)+15);
}
#define init_13_6_6_0
void FASTCALL call_13_6_6_0(SH2_struct* sh)
{
  SH2movli(sh, 6, (6<<4)+0);
}
#define init_13_6_6_1
void FASTCALL call_13_6_6_1(SH2_struct* sh)
{
  SH2movli(sh, 6, (6<<4)+1);
}
#define init_13_6_6_2
void FASTCALL call_13_6_6_2(SH2_struct* sh)
{
  SH2movli(sh, 6, (6<<4)+2);
}
#define init_13_6_6_3
void FASTCALL call_13_6_6_3(SH2_struct* sh)
{
  SH2movli(sh, 6, (6<<4)+3);
}
#define init_13_6_6_4
void FASTCALL call_13_6_6_4(SH2_struct* sh)
{
  SH2movli(sh, 6, (6<<4)+4);
}
#define init_13_6_6_5
void FASTCALL call_13_6_6_5(SH2_struct* sh)
{
  SH2movli(sh, 6, (6<<4)+5);
}
#define init_13_6_6_6
void FASTCALL call_13_6_6_6(SH2_struct* sh)
{
  SH2movli(sh, 6, (6<<4)+6);
}
#define init_13_6_6_7
void FASTCALL call_13_6_6_7(SH2_struct* sh)
{
  SH2movli(sh, 6, (6<<4)+7);
}
#define init_13_6_6_8
void FASTCALL call_13_6_6_8(SH2_struct* sh)
{
  SH2movli(sh, 6, (6<<4)+8);
}
#define init_13_6_6_9
void FASTCALL call_13_6_6_9(SH2_struct* sh)
{
  SH2movli(sh, 6, (6<<4)+9);
}
#define init_13_6_6_10
void FASTCALL call_13_6_6_10(SH2_struct* sh)
{
  SH2movli(sh, 6, (6<<4)+10);
}
#define init_13_6_6_11
void FASTCALL call_13_6_6_11(SH2_struct* sh)
{
  SH2movli(sh, 6, (6<<4)+11);
}
#define init_13_6_6_12
void FASTCALL call_13_6_6_12(SH2_struct* sh)
{
  SH2movli(sh, 6, (6<<4)+12);
}
#define init_13_6_6_13
void FASTCALL call_13_6_6_13(SH2_struct* sh)
{
  SH2movli(sh, 6, (6<<4)+13);
}
#define init_13_6_6_14
void FASTCALL call_13_6_6_14(SH2_struct* sh)
{
  SH2movli(sh, 6, (6<<4)+14);
}
#define init_13_6_6_15
void FASTCALL call_13_6_6_15(SH2_struct* sh)
{
  SH2movli(sh, 6, (6<<4)+15);
}
#define init_13_6_7_0
void FASTCALL call_13_6_7_0(SH2_struct* sh)
{
  SH2movli(sh, 6, (7<<4)+0);
}
#define init_13_6_7_1
void FASTCALL call_13_6_7_1(SH2_struct* sh)
{
  SH2movli(sh, 6, (7<<4)+1);
}
#define init_13_6_7_2
void FASTCALL call_13_6_7_2(SH2_struct* sh)
{
  SH2movli(sh, 6, (7<<4)+2);
}
#define init_13_6_7_3
void FASTCALL call_13_6_7_3(SH2_struct* sh)
{
  SH2movli(sh, 6, (7<<4)+3);
}
#define init_13_6_7_4
void FASTCALL call_13_6_7_4(SH2_struct* sh)
{
  SH2movli(sh, 6, (7<<4)+4);
}
#define init_13_6_7_5
void FASTCALL call_13_6_7_5(SH2_struct* sh)
{
  SH2movli(sh, 6, (7<<4)+5);
}
#define init_13_6_7_6
void FASTCALL call_13_6_7_6(SH2_struct* sh)
{
  SH2movli(sh, 6, (7<<4)+6);
}
#define init_13_6_7_7
void FASTCALL call_13_6_7_7(SH2_struct* sh)
{
  SH2movli(sh, 6, (7<<4)+7);
}
#define init_13_6_7_8
void FASTCALL call_13_6_7_8(SH2_struct* sh)
{
  SH2movli(sh, 6, (7<<4)+8);
}
#define init_13_6_7_9
void FASTCALL call_13_6_7_9(SH2_struct* sh)
{
  SH2movli(sh, 6, (7<<4)+9);
}
#define init_13_6_7_10
void FASTCALL call_13_6_7_10(SH2_struct* sh)
{
  SH2movli(sh, 6, (7<<4)+10);
}
#define init_13_6_7_11
void FASTCALL call_13_6_7_11(SH2_struct* sh)
{
  SH2movli(sh, 6, (7<<4)+11);
}
#define init_13_6_7_12
void FASTCALL call_13_6_7_12(SH2_struct* sh)
{
  SH2movli(sh, 6, (7<<4)+12);
}
#define init_13_6_7_13
void FASTCALL call_13_6_7_13(SH2_struct* sh)
{
  SH2movli(sh, 6, (7<<4)+13);
}
#define init_13_6_7_14
void FASTCALL call_13_6_7_14(SH2_struct* sh)
{
  SH2movli(sh, 6, (7<<4)+14);
}
#define init_13_6_7_15
void FASTCALL call_13_6_7_15(SH2_struct* sh)
{
  SH2movli(sh, 6, (7<<4)+15);
}
#define init_13_6_8_0
void FASTCALL call_13_6_8_0(SH2_struct* sh)
{
  SH2movli(sh, 6, (8<<4)+0);
}
#define init_13_6_8_1
void FASTCALL call_13_6_8_1(SH2_struct* sh)
{
  SH2movli(sh, 6, (8<<4)+1);
}
#define init_13_6_8_2
void FASTCALL call_13_6_8_2(SH2_struct* sh)
{
  SH2movli(sh, 6, (8<<4)+2);
}
#define init_13_6_8_3
void FASTCALL call_13_6_8_3(SH2_struct* sh)
{
  SH2movli(sh, 6, (8<<4)+3);
}
#define init_13_6_8_4
void FASTCALL call_13_6_8_4(SH2_struct* sh)
{
  SH2movli(sh, 6, (8<<4)+4);
}
#define init_13_6_8_5
void FASTCALL call_13_6_8_5(SH2_struct* sh)
{
  SH2movli(sh, 6, (8<<4)+5);
}
#define init_13_6_8_6
void FASTCALL call_13_6_8_6(SH2_struct* sh)
{
  SH2movli(sh, 6, (8<<4)+6);
}
#define init_13_6_8_7
void FASTCALL call_13_6_8_7(SH2_struct* sh)
{
  SH2movli(sh, 6, (8<<4)+7);
}
#define init_13_6_8_8
void FASTCALL call_13_6_8_8(SH2_struct* sh)
{
  SH2movli(sh, 6, (8<<4)+8);
}
#define init_13_6_8_9
void FASTCALL call_13_6_8_9(SH2_struct* sh)
{
  SH2movli(sh, 6, (8<<4)+9);
}
#define init_13_6_8_10
void FASTCALL call_13_6_8_10(SH2_struct* sh)
{
  SH2movli(sh, 6, (8<<4)+10);
}
#define init_13_6_8_11
void FASTCALL call_13_6_8_11(SH2_struct* sh)
{
  SH2movli(sh, 6, (8<<4)+11);
}
#define init_13_6_8_12
void FASTCALL call_13_6_8_12(SH2_struct* sh)
{
  SH2movli(sh, 6, (8<<4)+12);
}
#define init_13_6_8_13
void FASTCALL call_13_6_8_13(SH2_struct* sh)
{
  SH2movli(sh, 6, (8<<4)+13);
}
#define init_13_6_8_14
void FASTCALL call_13_6_8_14(SH2_struct* sh)
{
  SH2movli(sh, 6, (8<<4)+14);
}
#define init_13_6_8_15
void FASTCALL call_13_6_8_15(SH2_struct* sh)
{
  SH2movli(sh, 6, (8<<4)+15);
}
#define init_13_6_9_0
void FASTCALL call_13_6_9_0(SH2_struct* sh)
{
  SH2movli(sh, 6, (9<<4)+0);
}
#define init_13_6_9_1
void FASTCALL call_13_6_9_1(SH2_struct* sh)
{
  SH2movli(sh, 6, (9<<4)+1);
}
#define init_13_6_9_2
void FASTCALL call_13_6_9_2(SH2_struct* sh)
{
  SH2movli(sh, 6, (9<<4)+2);
}
#define init_13_6_9_3
void FASTCALL call_13_6_9_3(SH2_struct* sh)
{
  SH2movli(sh, 6, (9<<4)+3);
}
#define init_13_6_9_4
void FASTCALL call_13_6_9_4(SH2_struct* sh)
{
  SH2movli(sh, 6, (9<<4)+4);
}
#define init_13_6_9_5
void FASTCALL call_13_6_9_5(SH2_struct* sh)
{
  SH2movli(sh, 6, (9<<4)+5);
}
#define init_13_6_9_6
void FASTCALL call_13_6_9_6(SH2_struct* sh)
{
  SH2movli(sh, 6, (9<<4)+6);
}
#define init_13_6_9_7
void FASTCALL call_13_6_9_7(SH2_struct* sh)
{
  SH2movli(sh, 6, (9<<4)+7);
}
#define init_13_6_9_8
void FASTCALL call_13_6_9_8(SH2_struct* sh)
{
  SH2movli(sh, 6, (9<<4)+8);
}
#define init_13_6_9_9
void FASTCALL call_13_6_9_9(SH2_struct* sh)
{
  SH2movli(sh, 6, (9<<4)+9);
}
#define init_13_6_9_10
void FASTCALL call_13_6_9_10(SH2_struct* sh)
{
  SH2movli(sh, 6, (9<<4)+10);
}
#define init_13_6_9_11
void FASTCALL call_13_6_9_11(SH2_struct* sh)
{
  SH2movli(sh, 6, (9<<4)+11);
}
#define init_13_6_9_12
void FASTCALL call_13_6_9_12(SH2_struct* sh)
{
  SH2movli(sh, 6, (9<<4)+12);
}
#define init_13_6_9_13
void FASTCALL call_13_6_9_13(SH2_struct* sh)
{
  SH2movli(sh, 6, (9<<4)+13);
}
#define init_13_6_9_14
void FASTCALL call_13_6_9_14(SH2_struct* sh)
{
  SH2movli(sh, 6, (9<<4)+14);
}
#define init_13_6_9_15
void FASTCALL call_13_6_9_15(SH2_struct* sh)
{
  SH2movli(sh, 6, (9<<4)+15);
}
#define init_13_6_10_0
void FASTCALL call_13_6_10_0(SH2_struct* sh)
{
  SH2movli(sh, 6, (10<<4)+0);
}
#define init_13_6_10_1
void FASTCALL call_13_6_10_1(SH2_struct* sh)
{
  SH2movli(sh, 6, (10<<4)+1);
}
#define init_13_6_10_2
void FASTCALL call_13_6_10_2(SH2_struct* sh)
{
  SH2movli(sh, 6, (10<<4)+2);
}
#define init_13_6_10_3
void FASTCALL call_13_6_10_3(SH2_struct* sh)
{
  SH2movli(sh, 6, (10<<4)+3);
}
#define init_13_6_10_4
void FASTCALL call_13_6_10_4(SH2_struct* sh)
{
  SH2movli(sh, 6, (10<<4)+4);
}
#define init_13_6_10_5
void FASTCALL call_13_6_10_5(SH2_struct* sh)
{
  SH2movli(sh, 6, (10<<4)+5);
}
#define init_13_6_10_6
void FASTCALL call_13_6_10_6(SH2_struct* sh)
{
  SH2movli(sh, 6, (10<<4)+6);
}
#define init_13_6_10_7
void FASTCALL call_13_6_10_7(SH2_struct* sh)
{
  SH2movli(sh, 6, (10<<4)+7);
}
#define init_13_6_10_8
void FASTCALL call_13_6_10_8(SH2_struct* sh)
{
  SH2movli(sh, 6, (10<<4)+8);
}
#define init_13_6_10_9
void FASTCALL call_13_6_10_9(SH2_struct* sh)
{
  SH2movli(sh, 6, (10<<4)+9);
}
#define init_13_6_10_10
void FASTCALL call_13_6_10_10(SH2_struct* sh)
{
  SH2movli(sh, 6, (10<<4)+10);
}
#define init_13_6_10_11
void FASTCALL call_13_6_10_11(SH2_struct* sh)
{
  SH2movli(sh, 6, (10<<4)+11);
}
#define init_13_6_10_12
void FASTCALL call_13_6_10_12(SH2_struct* sh)
{
  SH2movli(sh, 6, (10<<4)+12);
}
#define init_13_6_10_13
void FASTCALL call_13_6_10_13(SH2_struct* sh)
{
  SH2movli(sh, 6, (10<<4)+13);
}
#define init_13_6_10_14
void FASTCALL call_13_6_10_14(SH2_struct* sh)
{
  SH2movli(sh, 6, (10<<4)+14);
}
#define init_13_6_10_15
void FASTCALL call_13_6_10_15(SH2_struct* sh)
{
  SH2movli(sh, 6, (10<<4)+15);
}
#define init_13_6_11_0
void FASTCALL call_13_6_11_0(SH2_struct* sh)
{
  SH2movli(sh, 6, (11<<4)+0);
}
#define init_13_6_11_1
void FASTCALL call_13_6_11_1(SH2_struct* sh)
{
  SH2movli(sh, 6, (11<<4)+1);
}
#define init_13_6_11_2
void FASTCALL call_13_6_11_2(SH2_struct* sh)
{
  SH2movli(sh, 6, (11<<4)+2);
}
#define init_13_6_11_3
void FASTCALL call_13_6_11_3(SH2_struct* sh)
{
  SH2movli(sh, 6, (11<<4)+3);
}
#define init_13_6_11_4
void FASTCALL call_13_6_11_4(SH2_struct* sh)
{
  SH2movli(sh, 6, (11<<4)+4);
}
#define init_13_6_11_5
void FASTCALL call_13_6_11_5(SH2_struct* sh)
{
  SH2movli(sh, 6, (11<<4)+5);
}
#define init_13_6_11_6
void FASTCALL call_13_6_11_6(SH2_struct* sh)
{
  SH2movli(sh, 6, (11<<4)+6);
}
#define init_13_6_11_7
void FASTCALL call_13_6_11_7(SH2_struct* sh)
{
  SH2movli(sh, 6, (11<<4)+7);
}
#define init_13_6_11_8
void FASTCALL call_13_6_11_8(SH2_struct* sh)
{
  SH2movli(sh, 6, (11<<4)+8);
}
#define init_13_6_11_9
void FASTCALL call_13_6_11_9(SH2_struct* sh)
{
  SH2movli(sh, 6, (11<<4)+9);
}
#define init_13_6_11_10
void FASTCALL call_13_6_11_10(SH2_struct* sh)
{
  SH2movli(sh, 6, (11<<4)+10);
}
#define init_13_6_11_11
void FASTCALL call_13_6_11_11(SH2_struct* sh)
{
  SH2movli(sh, 6, (11<<4)+11);
}
#define init_13_6_11_12
void FASTCALL call_13_6_11_12(SH2_struct* sh)
{
  SH2movli(sh, 6, (11<<4)+12);
}
#define init_13_6_11_13
void FASTCALL call_13_6_11_13(SH2_struct* sh)
{
  SH2movli(sh, 6, (11<<4)+13);
}
#define init_13_6_11_14
void FASTCALL call_13_6_11_14(SH2_struct* sh)
{
  SH2movli(sh, 6, (11<<4)+14);
}
#define init_13_6_11_15
void FASTCALL call_13_6_11_15(SH2_struct* sh)
{
  SH2movli(sh, 6, (11<<4)+15);
}
#define init_13_6_12_0
void FASTCALL call_13_6_12_0(SH2_struct* sh)
{
  SH2movli(sh, 6, (12<<4)+0);
}
#define init_13_6_12_1
void FASTCALL call_13_6_12_1(SH2_struct* sh)
{
  SH2movli(sh, 6, (12<<4)+1);
}
#define init_13_6_12_2
void FASTCALL call_13_6_12_2(SH2_struct* sh)
{
  SH2movli(sh, 6, (12<<4)+2);
}
#define init_13_6_12_3
void FASTCALL call_13_6_12_3(SH2_struct* sh)
{
  SH2movli(sh, 6, (12<<4)+3);
}
#define init_13_6_12_4
void FASTCALL call_13_6_12_4(SH2_struct* sh)
{
  SH2movli(sh, 6, (12<<4)+4);
}
#define init_13_6_12_5
void FASTCALL call_13_6_12_5(SH2_struct* sh)
{
  SH2movli(sh, 6, (12<<4)+5);
}
#define init_13_6_12_6
void FASTCALL call_13_6_12_6(SH2_struct* sh)
{
  SH2movli(sh, 6, (12<<4)+6);
}
#define init_13_6_12_7
void FASTCALL call_13_6_12_7(SH2_struct* sh)
{
  SH2movli(sh, 6, (12<<4)+7);
}
#define init_13_6_12_8
void FASTCALL call_13_6_12_8(SH2_struct* sh)
{
  SH2movli(sh, 6, (12<<4)+8);
}
#define init_13_6_12_9
void FASTCALL call_13_6_12_9(SH2_struct* sh)
{
  SH2movli(sh, 6, (12<<4)+9);
}
#define init_13_6_12_10
void FASTCALL call_13_6_12_10(SH2_struct* sh)
{
  SH2movli(sh, 6, (12<<4)+10);
}
#define init_13_6_12_11
void FASTCALL call_13_6_12_11(SH2_struct* sh)
{
  SH2movli(sh, 6, (12<<4)+11);
}
#define init_13_6_12_12
void FASTCALL call_13_6_12_12(SH2_struct* sh)
{
  SH2movli(sh, 6, (12<<4)+12);
}
#define init_13_6_12_13
void FASTCALL call_13_6_12_13(SH2_struct* sh)
{
  SH2movli(sh, 6, (12<<4)+13);
}
#define init_13_6_12_14
void FASTCALL call_13_6_12_14(SH2_struct* sh)
{
  SH2movli(sh, 6, (12<<4)+14);
}
#define init_13_6_12_15
void FASTCALL call_13_6_12_15(SH2_struct* sh)
{
  SH2movli(sh, 6, (12<<4)+15);
}
#define init_13_6_13_0
void FASTCALL call_13_6_13_0(SH2_struct* sh)
{
  SH2movli(sh, 6, (13<<4)+0);
}
#define init_13_6_13_1
void FASTCALL call_13_6_13_1(SH2_struct* sh)
{
  SH2movli(sh, 6, (13<<4)+1);
}
#define init_13_6_13_2
void FASTCALL call_13_6_13_2(SH2_struct* sh)
{
  SH2movli(sh, 6, (13<<4)+2);
}
#define init_13_6_13_3
void FASTCALL call_13_6_13_3(SH2_struct* sh)
{
  SH2movli(sh, 6, (13<<4)+3);
}
#define init_13_6_13_4
void FASTCALL call_13_6_13_4(SH2_struct* sh)
{
  SH2movli(sh, 6, (13<<4)+4);
}
#define init_13_6_13_5
void FASTCALL call_13_6_13_5(SH2_struct* sh)
{
  SH2movli(sh, 6, (13<<4)+5);
}
#define init_13_6_13_6
void FASTCALL call_13_6_13_6(SH2_struct* sh)
{
  SH2movli(sh, 6, (13<<4)+6);
}
#define init_13_6_13_7
void FASTCALL call_13_6_13_7(SH2_struct* sh)
{
  SH2movli(sh, 6, (13<<4)+7);
}
#define init_13_6_13_8
void FASTCALL call_13_6_13_8(SH2_struct* sh)
{
  SH2movli(sh, 6, (13<<4)+8);
}
#define init_13_6_13_9
void FASTCALL call_13_6_13_9(SH2_struct* sh)
{
  SH2movli(sh, 6, (13<<4)+9);
}
#define init_13_6_13_10
void FASTCALL call_13_6_13_10(SH2_struct* sh)
{
  SH2movli(sh, 6, (13<<4)+10);
}
#define init_13_6_13_11
void FASTCALL call_13_6_13_11(SH2_struct* sh)
{
  SH2movli(sh, 6, (13<<4)+11);
}
#define init_13_6_13_12
void FASTCALL call_13_6_13_12(SH2_struct* sh)
{
  SH2movli(sh, 6, (13<<4)+12);
}
#define init_13_6_13_13
void FASTCALL call_13_6_13_13(SH2_struct* sh)
{
  SH2movli(sh, 6, (13<<4)+13);
}
#define init_13_6_13_14
void FASTCALL call_13_6_13_14(SH2_struct* sh)
{
  SH2movli(sh, 6, (13<<4)+14);
}
#define init_13_6_13_15
void FASTCALL call_13_6_13_15(SH2_struct* sh)
{
  SH2movli(sh, 6, (13<<4)+15);
}
#define init_13_6_14_0
void FASTCALL call_13_6_14_0(SH2_struct* sh)
{
  SH2movli(sh, 6, (14<<4)+0);
}
#define init_13_6_14_1
void FASTCALL call_13_6_14_1(SH2_struct* sh)
{
  SH2movli(sh, 6, (14<<4)+1);
}
#define init_13_6_14_2
void FASTCALL call_13_6_14_2(SH2_struct* sh)
{
  SH2movli(sh, 6, (14<<4)+2);
}
#define init_13_6_14_3
void FASTCALL call_13_6_14_3(SH2_struct* sh)
{
  SH2movli(sh, 6, (14<<4)+3);
}
#define init_13_6_14_4
void FASTCALL call_13_6_14_4(SH2_struct* sh)
{
  SH2movli(sh, 6, (14<<4)+4);
}
#define init_13_6_14_5
void FASTCALL call_13_6_14_5(SH2_struct* sh)
{
  SH2movli(sh, 6, (14<<4)+5);
}
#define init_13_6_14_6
void FASTCALL call_13_6_14_6(SH2_struct* sh)
{
  SH2movli(sh, 6, (14<<4)+6);
}
#define init_13_6_14_7
void FASTCALL call_13_6_14_7(SH2_struct* sh)
{
  SH2movli(sh, 6, (14<<4)+7);
}
#define init_13_6_14_8
void FASTCALL call_13_6_14_8(SH2_struct* sh)
{
  SH2movli(sh, 6, (14<<4)+8);
}
#define init_13_6_14_9
void FASTCALL call_13_6_14_9(SH2_struct* sh)
{
  SH2movli(sh, 6, (14<<4)+9);
}
#define init_13_6_14_10
void FASTCALL call_13_6_14_10(SH2_struct* sh)
{
  SH2movli(sh, 6, (14<<4)+10);
}
#define init_13_6_14_11
void FASTCALL call_13_6_14_11(SH2_struct* sh)
{
  SH2movli(sh, 6, (14<<4)+11);
}
#define init_13_6_14_12
void FASTCALL call_13_6_14_12(SH2_struct* sh)
{
  SH2movli(sh, 6, (14<<4)+12);
}
#define init_13_6_14_13
void FASTCALL call_13_6_14_13(SH2_struct* sh)
{
  SH2movli(sh, 6, (14<<4)+13);
}
#define init_13_6_14_14
void FASTCALL call_13_6_14_14(SH2_struct* sh)
{
  SH2movli(sh, 6, (14<<4)+14);
}
#define init_13_6_14_15
void FASTCALL call_13_6_14_15(SH2_struct* sh)
{
  SH2movli(sh, 6, (14<<4)+15);
}
#define init_13_6_15_0
void FASTCALL call_13_6_15_0(SH2_struct* sh)
{
  SH2movli(sh, 6, (15<<4)+0);
}
#define init_13_6_15_1
void FASTCALL call_13_6_15_1(SH2_struct* sh)
{
  SH2movli(sh, 6, (15<<4)+1);
}
#define init_13_6_15_2
void FASTCALL call_13_6_15_2(SH2_struct* sh)
{
  SH2movli(sh, 6, (15<<4)+2);
}
#define init_13_6_15_3
void FASTCALL call_13_6_15_3(SH2_struct* sh)
{
  SH2movli(sh, 6, (15<<4)+3);
}
#define init_13_6_15_4
void FASTCALL call_13_6_15_4(SH2_struct* sh)
{
  SH2movli(sh, 6, (15<<4)+4);
}
#define init_13_6_15_5
void FASTCALL call_13_6_15_5(SH2_struct* sh)
{
  SH2movli(sh, 6, (15<<4)+5);
}
#define init_13_6_15_6
void FASTCALL call_13_6_15_6(SH2_struct* sh)
{
  SH2movli(sh, 6, (15<<4)+6);
}
#define init_13_6_15_7
void FASTCALL call_13_6_15_7(SH2_struct* sh)
{
  SH2movli(sh, 6, (15<<4)+7);
}
#define init_13_6_15_8
void FASTCALL call_13_6_15_8(SH2_struct* sh)
{
  SH2movli(sh, 6, (15<<4)+8);
}
#define init_13_6_15_9
void FASTCALL call_13_6_15_9(SH2_struct* sh)
{
  SH2movli(sh, 6, (15<<4)+9);
}
#define init_13_6_15_10
void FASTCALL call_13_6_15_10(SH2_struct* sh)
{
  SH2movli(sh, 6, (15<<4)+10);
}
#define init_13_6_15_11
void FASTCALL call_13_6_15_11(SH2_struct* sh)
{
  SH2movli(sh, 6, (15<<4)+11);
}
#define init_13_6_15_12
void FASTCALL call_13_6_15_12(SH2_struct* sh)
{
  SH2movli(sh, 6, (15<<4)+12);
}
#define init_13_6_15_13
void FASTCALL call_13_6_15_13(SH2_struct* sh)
{
  SH2movli(sh, 6, (15<<4)+13);
}
#define init_13_6_15_14
void FASTCALL call_13_6_15_14(SH2_struct* sh)
{
  SH2movli(sh, 6, (15<<4)+14);
}
#define init_13_6_15_15
void FASTCALL call_13_6_15_15(SH2_struct* sh)
{
  SH2movli(sh, 6, (15<<4)+15);
}
#define init_13_7_0_0
void FASTCALL call_13_7_0_0(SH2_struct* sh)
{
  SH2movli(sh, 7, (0<<4)+0);
}
#define init_13_7_0_1
void FASTCALL call_13_7_0_1(SH2_struct* sh)
{
  SH2movli(sh, 7, (0<<4)+1);
}
#define init_13_7_0_2
void FASTCALL call_13_7_0_2(SH2_struct* sh)
{
  SH2movli(sh, 7, (0<<4)+2);
}
#define init_13_7_0_3
void FASTCALL call_13_7_0_3(SH2_struct* sh)
{
  SH2movli(sh, 7, (0<<4)+3);
}
#define init_13_7_0_4
void FASTCALL call_13_7_0_4(SH2_struct* sh)
{
  SH2movli(sh, 7, (0<<4)+4);
}
#define init_13_7_0_5
void FASTCALL call_13_7_0_5(SH2_struct* sh)
{
  SH2movli(sh, 7, (0<<4)+5);
}
#define init_13_7_0_6
void FASTCALL call_13_7_0_6(SH2_struct* sh)
{
  SH2movli(sh, 7, (0<<4)+6);
}
#define init_13_7_0_7
void FASTCALL call_13_7_0_7(SH2_struct* sh)
{
  SH2movli(sh, 7, (0<<4)+7);
}
#define init_13_7_0_8
void FASTCALL call_13_7_0_8(SH2_struct* sh)
{
  SH2movli(sh, 7, (0<<4)+8);
}
#define init_13_7_0_9
void FASTCALL call_13_7_0_9(SH2_struct* sh)
{
  SH2movli(sh, 7, (0<<4)+9);
}
#define init_13_7_0_10
void FASTCALL call_13_7_0_10(SH2_struct* sh)
{
  SH2movli(sh, 7, (0<<4)+10);
}
#define init_13_7_0_11
void FASTCALL call_13_7_0_11(SH2_struct* sh)
{
  SH2movli(sh, 7, (0<<4)+11);
}
#define init_13_7_0_12
void FASTCALL call_13_7_0_12(SH2_struct* sh)
{
  SH2movli(sh, 7, (0<<4)+12);
}
#define init_13_7_0_13
void FASTCALL call_13_7_0_13(SH2_struct* sh)
{
  SH2movli(sh, 7, (0<<4)+13);
}
#define init_13_7_0_14
void FASTCALL call_13_7_0_14(SH2_struct* sh)
{
  SH2movli(sh, 7, (0<<4)+14);
}
#define init_13_7_0_15
void FASTCALL call_13_7_0_15(SH2_struct* sh)
{
  SH2movli(sh, 7, (0<<4)+15);
}
#define init_13_7_1_0
void FASTCALL call_13_7_1_0(SH2_struct* sh)
{
  SH2movli(sh, 7, (1<<4)+0);
}
#define init_13_7_1_1
void FASTCALL call_13_7_1_1(SH2_struct* sh)
{
  SH2movli(sh, 7, (1<<4)+1);
}
#define init_13_7_1_2
void FASTCALL call_13_7_1_2(SH2_struct* sh)
{
  SH2movli(sh, 7, (1<<4)+2);
}
#define init_13_7_1_3
void FASTCALL call_13_7_1_3(SH2_struct* sh)
{
  SH2movli(sh, 7, (1<<4)+3);
}
#define init_13_7_1_4
void FASTCALL call_13_7_1_4(SH2_struct* sh)
{
  SH2movli(sh, 7, (1<<4)+4);
}
#define init_13_7_1_5
void FASTCALL call_13_7_1_5(SH2_struct* sh)
{
  SH2movli(sh, 7, (1<<4)+5);
}
#define init_13_7_1_6
void FASTCALL call_13_7_1_6(SH2_struct* sh)
{
  SH2movli(sh, 7, (1<<4)+6);
}
#define init_13_7_1_7
void FASTCALL call_13_7_1_7(SH2_struct* sh)
{
  SH2movli(sh, 7, (1<<4)+7);
}
#define init_13_7_1_8
void FASTCALL call_13_7_1_8(SH2_struct* sh)
{
  SH2movli(sh, 7, (1<<4)+8);
}
#define init_13_7_1_9
void FASTCALL call_13_7_1_9(SH2_struct* sh)
{
  SH2movli(sh, 7, (1<<4)+9);
}
#define init_13_7_1_10
void FASTCALL call_13_7_1_10(SH2_struct* sh)
{
  SH2movli(sh, 7, (1<<4)+10);
}
#define init_13_7_1_11
void FASTCALL call_13_7_1_11(SH2_struct* sh)
{
  SH2movli(sh, 7, (1<<4)+11);
}
#define init_13_7_1_12
void FASTCALL call_13_7_1_12(SH2_struct* sh)
{
  SH2movli(sh, 7, (1<<4)+12);
}
#define init_13_7_1_13
void FASTCALL call_13_7_1_13(SH2_struct* sh)
{
  SH2movli(sh, 7, (1<<4)+13);
}
#define init_13_7_1_14
void FASTCALL call_13_7_1_14(SH2_struct* sh)
{
  SH2movli(sh, 7, (1<<4)+14);
}
#define init_13_7_1_15
void FASTCALL call_13_7_1_15(SH2_struct* sh)
{
  SH2movli(sh, 7, (1<<4)+15);
}
#define init_13_7_2_0
void FASTCALL call_13_7_2_0(SH2_struct* sh)
{
  SH2movli(sh, 7, (2<<4)+0);
}
#define init_13_7_2_1
void FASTCALL call_13_7_2_1(SH2_struct* sh)
{
  SH2movli(sh, 7, (2<<4)+1);
}
#define init_13_7_2_2
void FASTCALL call_13_7_2_2(SH2_struct* sh)
{
  SH2movli(sh, 7, (2<<4)+2);
}
#define init_13_7_2_3
void FASTCALL call_13_7_2_3(SH2_struct* sh)
{
  SH2movli(sh, 7, (2<<4)+3);
}
#define init_13_7_2_4
void FASTCALL call_13_7_2_4(SH2_struct* sh)
{
  SH2movli(sh, 7, (2<<4)+4);
}
#define init_13_7_2_5
void FASTCALL call_13_7_2_5(SH2_struct* sh)
{
  SH2movli(sh, 7, (2<<4)+5);
}
#define init_13_7_2_6
void FASTCALL call_13_7_2_6(SH2_struct* sh)
{
  SH2movli(sh, 7, (2<<4)+6);
}
#define init_13_7_2_7
void FASTCALL call_13_7_2_7(SH2_struct* sh)
{
  SH2movli(sh, 7, (2<<4)+7);
}
#define init_13_7_2_8
void FASTCALL call_13_7_2_8(SH2_struct* sh)
{
  SH2movli(sh, 7, (2<<4)+8);
}
#define init_13_7_2_9
void FASTCALL call_13_7_2_9(SH2_struct* sh)
{
  SH2movli(sh, 7, (2<<4)+9);
}
#define init_13_7_2_10
void FASTCALL call_13_7_2_10(SH2_struct* sh)
{
  SH2movli(sh, 7, (2<<4)+10);
}
#define init_13_7_2_11
void FASTCALL call_13_7_2_11(SH2_struct* sh)
{
  SH2movli(sh, 7, (2<<4)+11);
}
#define init_13_7_2_12
void FASTCALL call_13_7_2_12(SH2_struct* sh)
{
  SH2movli(sh, 7, (2<<4)+12);
}
#define init_13_7_2_13
void FASTCALL call_13_7_2_13(SH2_struct* sh)
{
  SH2movli(sh, 7, (2<<4)+13);
}
#define init_13_7_2_14
void FASTCALL call_13_7_2_14(SH2_struct* sh)
{
  SH2movli(sh, 7, (2<<4)+14);
}
#define init_13_7_2_15
void FASTCALL call_13_7_2_15(SH2_struct* sh)
{
  SH2movli(sh, 7, (2<<4)+15);
}
#define init_13_7_3_0
void FASTCALL call_13_7_3_0(SH2_struct* sh)
{
  SH2movli(sh, 7, (3<<4)+0);
}
#define init_13_7_3_1
void FASTCALL call_13_7_3_1(SH2_struct* sh)
{
  SH2movli(sh, 7, (3<<4)+1);
}
#define init_13_7_3_2
void FASTCALL call_13_7_3_2(SH2_struct* sh)
{
  SH2movli(sh, 7, (3<<4)+2);
}
#define init_13_7_3_3
void FASTCALL call_13_7_3_3(SH2_struct* sh)
{
  SH2movli(sh, 7, (3<<4)+3);
}
#define init_13_7_3_4
void FASTCALL call_13_7_3_4(SH2_struct* sh)
{
  SH2movli(sh, 7, (3<<4)+4);
}
#define init_13_7_3_5
void FASTCALL call_13_7_3_5(SH2_struct* sh)
{
  SH2movli(sh, 7, (3<<4)+5);
}
#define init_13_7_3_6
void FASTCALL call_13_7_3_6(SH2_struct* sh)
{
  SH2movli(sh, 7, (3<<4)+6);
}
#define init_13_7_3_7
void FASTCALL call_13_7_3_7(SH2_struct* sh)
{
  SH2movli(sh, 7, (3<<4)+7);
}
#define init_13_7_3_8
void FASTCALL call_13_7_3_8(SH2_struct* sh)
{
  SH2movli(sh, 7, (3<<4)+8);
}
#define init_13_7_3_9
void FASTCALL call_13_7_3_9(SH2_struct* sh)
{
  SH2movli(sh, 7, (3<<4)+9);
}
#define init_13_7_3_10
void FASTCALL call_13_7_3_10(SH2_struct* sh)
{
  SH2movli(sh, 7, (3<<4)+10);
}
#define init_13_7_3_11
void FASTCALL call_13_7_3_11(SH2_struct* sh)
{
  SH2movli(sh, 7, (3<<4)+11);
}
#define init_13_7_3_12
void FASTCALL call_13_7_3_12(SH2_struct* sh)
{
  SH2movli(sh, 7, (3<<4)+12);
}
#define init_13_7_3_13
void FASTCALL call_13_7_3_13(SH2_struct* sh)
{
  SH2movli(sh, 7, (3<<4)+13);
}
#define init_13_7_3_14
void FASTCALL call_13_7_3_14(SH2_struct* sh)
{
  SH2movli(sh, 7, (3<<4)+14);
}
#define init_13_7_3_15
void FASTCALL call_13_7_3_15(SH2_struct* sh)
{
  SH2movli(sh, 7, (3<<4)+15);
}
#define init_13_7_4_0
void FASTCALL call_13_7_4_0(SH2_struct* sh)
{
  SH2movli(sh, 7, (4<<4)+0);
}
#define init_13_7_4_1
void FASTCALL call_13_7_4_1(SH2_struct* sh)
{
  SH2movli(sh, 7, (4<<4)+1);
}
#define init_13_7_4_2
void FASTCALL call_13_7_4_2(SH2_struct* sh)
{
  SH2movli(sh, 7, (4<<4)+2);
}
#define init_13_7_4_3
void FASTCALL call_13_7_4_3(SH2_struct* sh)
{
  SH2movli(sh, 7, (4<<4)+3);
}
#define init_13_7_4_4
void FASTCALL call_13_7_4_4(SH2_struct* sh)
{
  SH2movli(sh, 7, (4<<4)+4);
}
#define init_13_7_4_5
void FASTCALL call_13_7_4_5(SH2_struct* sh)
{
  SH2movli(sh, 7, (4<<4)+5);
}
#define init_13_7_4_6
void FASTCALL call_13_7_4_6(SH2_struct* sh)
{
  SH2movli(sh, 7, (4<<4)+6);
}
#define init_13_7_4_7
void FASTCALL call_13_7_4_7(SH2_struct* sh)
{
  SH2movli(sh, 7, (4<<4)+7);
}
#define init_13_7_4_8
void FASTCALL call_13_7_4_8(SH2_struct* sh)
{
  SH2movli(sh, 7, (4<<4)+8);
}
#define init_13_7_4_9
void FASTCALL call_13_7_4_9(SH2_struct* sh)
{
  SH2movli(sh, 7, (4<<4)+9);
}
#define init_13_7_4_10
void FASTCALL call_13_7_4_10(SH2_struct* sh)
{
  SH2movli(sh, 7, (4<<4)+10);
}
#define init_13_7_4_11
void FASTCALL call_13_7_4_11(SH2_struct* sh)
{
  SH2movli(sh, 7, (4<<4)+11);
}
#define init_13_7_4_12
void FASTCALL call_13_7_4_12(SH2_struct* sh)
{
  SH2movli(sh, 7, (4<<4)+12);
}
#define init_13_7_4_13
void FASTCALL call_13_7_4_13(SH2_struct* sh)
{
  SH2movli(sh, 7, (4<<4)+13);
}
#define init_13_7_4_14
void FASTCALL call_13_7_4_14(SH2_struct* sh)
{
  SH2movli(sh, 7, (4<<4)+14);
}
#define init_13_7_4_15
void FASTCALL call_13_7_4_15(SH2_struct* sh)
{
  SH2movli(sh, 7, (4<<4)+15);
}
#define init_13_7_5_0
void FASTCALL call_13_7_5_0(SH2_struct* sh)
{
  SH2movli(sh, 7, (5<<4)+0);
}
#define init_13_7_5_1
void FASTCALL call_13_7_5_1(SH2_struct* sh)
{
  SH2movli(sh, 7, (5<<4)+1);
}
#define init_13_7_5_2
void FASTCALL call_13_7_5_2(SH2_struct* sh)
{
  SH2movli(sh, 7, (5<<4)+2);
}
#define init_13_7_5_3
void FASTCALL call_13_7_5_3(SH2_struct* sh)
{
  SH2movli(sh, 7, (5<<4)+3);
}
#define init_13_7_5_4
void FASTCALL call_13_7_5_4(SH2_struct* sh)
{
  SH2movli(sh, 7, (5<<4)+4);
}
#define init_13_7_5_5
void FASTCALL call_13_7_5_5(SH2_struct* sh)
{
  SH2movli(sh, 7, (5<<4)+5);
}
#define init_13_7_5_6
void FASTCALL call_13_7_5_6(SH2_struct* sh)
{
  SH2movli(sh, 7, (5<<4)+6);
}
#define init_13_7_5_7
void FASTCALL call_13_7_5_7(SH2_struct* sh)
{
  SH2movli(sh, 7, (5<<4)+7);
}
#define init_13_7_5_8
void FASTCALL call_13_7_5_8(SH2_struct* sh)
{
  SH2movli(sh, 7, (5<<4)+8);
}
#define init_13_7_5_9
void FASTCALL call_13_7_5_9(SH2_struct* sh)
{
  SH2movli(sh, 7, (5<<4)+9);
}
#define init_13_7_5_10
void FASTCALL call_13_7_5_10(SH2_struct* sh)
{
  SH2movli(sh, 7, (5<<4)+10);
}
#define init_13_7_5_11
void FASTCALL call_13_7_5_11(SH2_struct* sh)
{
  SH2movli(sh, 7, (5<<4)+11);
}
#define init_13_7_5_12
void FASTCALL call_13_7_5_12(SH2_struct* sh)
{
  SH2movli(sh, 7, (5<<4)+12);
}
#define init_13_7_5_13
void FASTCALL call_13_7_5_13(SH2_struct* sh)
{
  SH2movli(sh, 7, (5<<4)+13);
}
#define init_13_7_5_14
void FASTCALL call_13_7_5_14(SH2_struct* sh)
{
  SH2movli(sh, 7, (5<<4)+14);
}
#define init_13_7_5_15
void FASTCALL call_13_7_5_15(SH2_struct* sh)
{
  SH2movli(sh, 7, (5<<4)+15);
}
#define init_13_7_6_0
void FASTCALL call_13_7_6_0(SH2_struct* sh)
{
  SH2movli(sh, 7, (6<<4)+0);
}
#define init_13_7_6_1
void FASTCALL call_13_7_6_1(SH2_struct* sh)
{
  SH2movli(sh, 7, (6<<4)+1);
}
#define init_13_7_6_2
void FASTCALL call_13_7_6_2(SH2_struct* sh)
{
  SH2movli(sh, 7, (6<<4)+2);
}
#define init_13_7_6_3
void FASTCALL call_13_7_6_3(SH2_struct* sh)
{
  SH2movli(sh, 7, (6<<4)+3);
}
#define init_13_7_6_4
void FASTCALL call_13_7_6_4(SH2_struct* sh)
{
  SH2movli(sh, 7, (6<<4)+4);
}
#define init_13_7_6_5
void FASTCALL call_13_7_6_5(SH2_struct* sh)
{
  SH2movli(sh, 7, (6<<4)+5);
}
#define init_13_7_6_6
void FASTCALL call_13_7_6_6(SH2_struct* sh)
{
  SH2movli(sh, 7, (6<<4)+6);
}
#define init_13_7_6_7
void FASTCALL call_13_7_6_7(SH2_struct* sh)
{
  SH2movli(sh, 7, (6<<4)+7);
}
#define init_13_7_6_8
void FASTCALL call_13_7_6_8(SH2_struct* sh)
{
  SH2movli(sh, 7, (6<<4)+8);
}
#define init_13_7_6_9
void FASTCALL call_13_7_6_9(SH2_struct* sh)
{
  SH2movli(sh, 7, (6<<4)+9);
}
#define init_13_7_6_10
void FASTCALL call_13_7_6_10(SH2_struct* sh)
{
  SH2movli(sh, 7, (6<<4)+10);
}
#define init_13_7_6_11
void FASTCALL call_13_7_6_11(SH2_struct* sh)
{
  SH2movli(sh, 7, (6<<4)+11);
}
#define init_13_7_6_12
void FASTCALL call_13_7_6_12(SH2_struct* sh)
{
  SH2movli(sh, 7, (6<<4)+12);
}
#define init_13_7_6_13
void FASTCALL call_13_7_6_13(SH2_struct* sh)
{
  SH2movli(sh, 7, (6<<4)+13);
}
#define init_13_7_6_14
void FASTCALL call_13_7_6_14(SH2_struct* sh)
{
  SH2movli(sh, 7, (6<<4)+14);
}
#define init_13_7_6_15
void FASTCALL call_13_7_6_15(SH2_struct* sh)
{
  SH2movli(sh, 7, (6<<4)+15);
}
#define init_13_7_7_0
void FASTCALL call_13_7_7_0(SH2_struct* sh)
{
  SH2movli(sh, 7, (7<<4)+0);
}
#define init_13_7_7_1
void FASTCALL call_13_7_7_1(SH2_struct* sh)
{
  SH2movli(sh, 7, (7<<4)+1);
}
#define init_13_7_7_2
void FASTCALL call_13_7_7_2(SH2_struct* sh)
{
  SH2movli(sh, 7, (7<<4)+2);
}
#define init_13_7_7_3
void FASTCALL call_13_7_7_3(SH2_struct* sh)
{
  SH2movli(sh, 7, (7<<4)+3);
}
#define init_13_7_7_4
void FASTCALL call_13_7_7_4(SH2_struct* sh)
{
  SH2movli(sh, 7, (7<<4)+4);
}
#define init_13_7_7_5
void FASTCALL call_13_7_7_5(SH2_struct* sh)
{
  SH2movli(sh, 7, (7<<4)+5);
}
#define init_13_7_7_6
void FASTCALL call_13_7_7_6(SH2_struct* sh)
{
  SH2movli(sh, 7, (7<<4)+6);
}
#define init_13_7_7_7
void FASTCALL call_13_7_7_7(SH2_struct* sh)
{
  SH2movli(sh, 7, (7<<4)+7);
}
#define init_13_7_7_8
void FASTCALL call_13_7_7_8(SH2_struct* sh)
{
  SH2movli(sh, 7, (7<<4)+8);
}
#define init_13_7_7_9
void FASTCALL call_13_7_7_9(SH2_struct* sh)
{
  SH2movli(sh, 7, (7<<4)+9);
}
#define init_13_7_7_10
void FASTCALL call_13_7_7_10(SH2_struct* sh)
{
  SH2movli(sh, 7, (7<<4)+10);
}
#define init_13_7_7_11
void FASTCALL call_13_7_7_11(SH2_struct* sh)
{
  SH2movli(sh, 7, (7<<4)+11);
}
#define init_13_7_7_12
void FASTCALL call_13_7_7_12(SH2_struct* sh)
{
  SH2movli(sh, 7, (7<<4)+12);
}
#define init_13_7_7_13
void FASTCALL call_13_7_7_13(SH2_struct* sh)
{
  SH2movli(sh, 7, (7<<4)+13);
}
#define init_13_7_7_14
void FASTCALL call_13_7_7_14(SH2_struct* sh)
{
  SH2movli(sh, 7, (7<<4)+14);
}
#define init_13_7_7_15
void FASTCALL call_13_7_7_15(SH2_struct* sh)
{
  SH2movli(sh, 7, (7<<4)+15);
}
#define init_13_7_8_0
void FASTCALL call_13_7_8_0(SH2_struct* sh)
{
  SH2movli(sh, 7, (8<<4)+0);
}
#define init_13_7_8_1
void FASTCALL call_13_7_8_1(SH2_struct* sh)
{
  SH2movli(sh, 7, (8<<4)+1);
}
#define init_13_7_8_2
void FASTCALL call_13_7_8_2(SH2_struct* sh)
{
  SH2movli(sh, 7, (8<<4)+2);
}
#define init_13_7_8_3
void FASTCALL call_13_7_8_3(SH2_struct* sh)
{
  SH2movli(sh, 7, (8<<4)+3);
}
#define init_13_7_8_4
void FASTCALL call_13_7_8_4(SH2_struct* sh)
{
  SH2movli(sh, 7, (8<<4)+4);
}
#define init_13_7_8_5
void FASTCALL call_13_7_8_5(SH2_struct* sh)
{
  SH2movli(sh, 7, (8<<4)+5);
}
#define init_13_7_8_6
void FASTCALL call_13_7_8_6(SH2_struct* sh)
{
  SH2movli(sh, 7, (8<<4)+6);
}
#define init_13_7_8_7
void FASTCALL call_13_7_8_7(SH2_struct* sh)
{
  SH2movli(sh, 7, (8<<4)+7);
}
#define init_13_7_8_8
void FASTCALL call_13_7_8_8(SH2_struct* sh)
{
  SH2movli(sh, 7, (8<<4)+8);
}
#define init_13_7_8_9
void FASTCALL call_13_7_8_9(SH2_struct* sh)
{
  SH2movli(sh, 7, (8<<4)+9);
}
#define init_13_7_8_10
void FASTCALL call_13_7_8_10(SH2_struct* sh)
{
  SH2movli(sh, 7, (8<<4)+10);
}
#define init_13_7_8_11
void FASTCALL call_13_7_8_11(SH2_struct* sh)
{
  SH2movli(sh, 7, (8<<4)+11);
}
#define init_13_7_8_12
void FASTCALL call_13_7_8_12(SH2_struct* sh)
{
  SH2movli(sh, 7, (8<<4)+12);
}
#define init_13_7_8_13
void FASTCALL call_13_7_8_13(SH2_struct* sh)
{
  SH2movli(sh, 7, (8<<4)+13);
}
#define init_13_7_8_14
void FASTCALL call_13_7_8_14(SH2_struct* sh)
{
  SH2movli(sh, 7, (8<<4)+14);
}
#define init_13_7_8_15
void FASTCALL call_13_7_8_15(SH2_struct* sh)
{
  SH2movli(sh, 7, (8<<4)+15);
}
#define init_13_7_9_0
void FASTCALL call_13_7_9_0(SH2_struct* sh)
{
  SH2movli(sh, 7, (9<<4)+0);
}
#define init_13_7_9_1
void FASTCALL call_13_7_9_1(SH2_struct* sh)
{
  SH2movli(sh, 7, (9<<4)+1);
}
#define init_13_7_9_2
void FASTCALL call_13_7_9_2(SH2_struct* sh)
{
  SH2movli(sh, 7, (9<<4)+2);
}
#define init_13_7_9_3
void FASTCALL call_13_7_9_3(SH2_struct* sh)
{
  SH2movli(sh, 7, (9<<4)+3);
}
#define init_13_7_9_4
void FASTCALL call_13_7_9_4(SH2_struct* sh)
{
  SH2movli(sh, 7, (9<<4)+4);
}
#define init_13_7_9_5
void FASTCALL call_13_7_9_5(SH2_struct* sh)
{
  SH2movli(sh, 7, (9<<4)+5);
}
#define init_13_7_9_6
void FASTCALL call_13_7_9_6(SH2_struct* sh)
{
  SH2movli(sh, 7, (9<<4)+6);
}
#define init_13_7_9_7
void FASTCALL call_13_7_9_7(SH2_struct* sh)
{
  SH2movli(sh, 7, (9<<4)+7);
}
#define init_13_7_9_8
void FASTCALL call_13_7_9_8(SH2_struct* sh)
{
  SH2movli(sh, 7, (9<<4)+8);
}
#define init_13_7_9_9
void FASTCALL call_13_7_9_9(SH2_struct* sh)
{
  SH2movli(sh, 7, (9<<4)+9);
}
#define init_13_7_9_10
void FASTCALL call_13_7_9_10(SH2_struct* sh)
{
  SH2movli(sh, 7, (9<<4)+10);
}
#define init_13_7_9_11
void FASTCALL call_13_7_9_11(SH2_struct* sh)
{
  SH2movli(sh, 7, (9<<4)+11);
}
#define init_13_7_9_12
void FASTCALL call_13_7_9_12(SH2_struct* sh)
{
  SH2movli(sh, 7, (9<<4)+12);
}
#define init_13_7_9_13
void FASTCALL call_13_7_9_13(SH2_struct* sh)
{
  SH2movli(sh, 7, (9<<4)+13);
}
#define init_13_7_9_14
void FASTCALL call_13_7_9_14(SH2_struct* sh)
{
  SH2movli(sh, 7, (9<<4)+14);
}
#define init_13_7_9_15
void FASTCALL call_13_7_9_15(SH2_struct* sh)
{
  SH2movli(sh, 7, (9<<4)+15);
}
#define init_13_7_10_0
void FASTCALL call_13_7_10_0(SH2_struct* sh)
{
  SH2movli(sh, 7, (10<<4)+0);
}
#define init_13_7_10_1
void FASTCALL call_13_7_10_1(SH2_struct* sh)
{
  SH2movli(sh, 7, (10<<4)+1);
}
#define init_13_7_10_2
void FASTCALL call_13_7_10_2(SH2_struct* sh)
{
  SH2movli(sh, 7, (10<<4)+2);
}
#define init_13_7_10_3
void FASTCALL call_13_7_10_3(SH2_struct* sh)
{
  SH2movli(sh, 7, (10<<4)+3);
}
#define init_13_7_10_4
void FASTCALL call_13_7_10_4(SH2_struct* sh)
{
  SH2movli(sh, 7, (10<<4)+4);
}
#define init_13_7_10_5
void FASTCALL call_13_7_10_5(SH2_struct* sh)
{
  SH2movli(sh, 7, (10<<4)+5);
}
#define init_13_7_10_6
void FASTCALL call_13_7_10_6(SH2_struct* sh)
{
  SH2movli(sh, 7, (10<<4)+6);
}
#define init_13_7_10_7
void FASTCALL call_13_7_10_7(SH2_struct* sh)
{
  SH2movli(sh, 7, (10<<4)+7);
}
#define init_13_7_10_8
void FASTCALL call_13_7_10_8(SH2_struct* sh)
{
  SH2movli(sh, 7, (10<<4)+8);
}
#define init_13_7_10_9
void FASTCALL call_13_7_10_9(SH2_struct* sh)
{
  SH2movli(sh, 7, (10<<4)+9);
}
#define init_13_7_10_10
void FASTCALL call_13_7_10_10(SH2_struct* sh)
{
  SH2movli(sh, 7, (10<<4)+10);
}
#define init_13_7_10_11
void FASTCALL call_13_7_10_11(SH2_struct* sh)
{
  SH2movli(sh, 7, (10<<4)+11);
}
#define init_13_7_10_12
void FASTCALL call_13_7_10_12(SH2_struct* sh)
{
  SH2movli(sh, 7, (10<<4)+12);
}
#define init_13_7_10_13
void FASTCALL call_13_7_10_13(SH2_struct* sh)
{
  SH2movli(sh, 7, (10<<4)+13);
}
#define init_13_7_10_14
void FASTCALL call_13_7_10_14(SH2_struct* sh)
{
  SH2movli(sh, 7, (10<<4)+14);
}
#define init_13_7_10_15
void FASTCALL call_13_7_10_15(SH2_struct* sh)
{
  SH2movli(sh, 7, (10<<4)+15);
}
#define init_13_7_11_0
void FASTCALL call_13_7_11_0(SH2_struct* sh)
{
  SH2movli(sh, 7, (11<<4)+0);
}
#define init_13_7_11_1
void FASTCALL call_13_7_11_1(SH2_struct* sh)
{
  SH2movli(sh, 7, (11<<4)+1);
}
#define init_13_7_11_2
void FASTCALL call_13_7_11_2(SH2_struct* sh)
{
  SH2movli(sh, 7, (11<<4)+2);
}
#define init_13_7_11_3
void FASTCALL call_13_7_11_3(SH2_struct* sh)
{
  SH2movli(sh, 7, (11<<4)+3);
}
#define init_13_7_11_4
void FASTCALL call_13_7_11_4(SH2_struct* sh)
{
  SH2movli(sh, 7, (11<<4)+4);
}
#define init_13_7_11_5
void FASTCALL call_13_7_11_5(SH2_struct* sh)
{
  SH2movli(sh, 7, (11<<4)+5);
}
#define init_13_7_11_6
void FASTCALL call_13_7_11_6(SH2_struct* sh)
{
  SH2movli(sh, 7, (11<<4)+6);
}
#define init_13_7_11_7
void FASTCALL call_13_7_11_7(SH2_struct* sh)
{
  SH2movli(sh, 7, (11<<4)+7);
}
#define init_13_7_11_8
void FASTCALL call_13_7_11_8(SH2_struct* sh)
{
  SH2movli(sh, 7, (11<<4)+8);
}
#define init_13_7_11_9
void FASTCALL call_13_7_11_9(SH2_struct* sh)
{
  SH2movli(sh, 7, (11<<4)+9);
}
#define init_13_7_11_10
void FASTCALL call_13_7_11_10(SH2_struct* sh)
{
  SH2movli(sh, 7, (11<<4)+10);
}
#define init_13_7_11_11
void FASTCALL call_13_7_11_11(SH2_struct* sh)
{
  SH2movli(sh, 7, (11<<4)+11);
}
#define init_13_7_11_12
void FASTCALL call_13_7_11_12(SH2_struct* sh)
{
  SH2movli(sh, 7, (11<<4)+12);
}
#define init_13_7_11_13
void FASTCALL call_13_7_11_13(SH2_struct* sh)
{
  SH2movli(sh, 7, (11<<4)+13);
}
#define init_13_7_11_14
void FASTCALL call_13_7_11_14(SH2_struct* sh)
{
  SH2movli(sh, 7, (11<<4)+14);
}
#define init_13_7_11_15
void FASTCALL call_13_7_11_15(SH2_struct* sh)
{
  SH2movli(sh, 7, (11<<4)+15);
}
#define init_13_7_12_0
void FASTCALL call_13_7_12_0(SH2_struct* sh)
{
  SH2movli(sh, 7, (12<<4)+0);
}
#define init_13_7_12_1
void FASTCALL call_13_7_12_1(SH2_struct* sh)
{
  SH2movli(sh, 7, (12<<4)+1);
}
#define init_13_7_12_2
void FASTCALL call_13_7_12_2(SH2_struct* sh)
{
  SH2movli(sh, 7, (12<<4)+2);
}
#define init_13_7_12_3
void FASTCALL call_13_7_12_3(SH2_struct* sh)
{
  SH2movli(sh, 7, (12<<4)+3);
}
#define init_13_7_12_4
void FASTCALL call_13_7_12_4(SH2_struct* sh)
{
  SH2movli(sh, 7, (12<<4)+4);
}
#define init_13_7_12_5
void FASTCALL call_13_7_12_5(SH2_struct* sh)
{
  SH2movli(sh, 7, (12<<4)+5);
}
#define init_13_7_12_6
void FASTCALL call_13_7_12_6(SH2_struct* sh)
{
  SH2movli(sh, 7, (12<<4)+6);
}
#define init_13_7_12_7
void FASTCALL call_13_7_12_7(SH2_struct* sh)
{
  SH2movli(sh, 7, (12<<4)+7);
}
#define init_13_7_12_8
void FASTCALL call_13_7_12_8(SH2_struct* sh)
{
  SH2movli(sh, 7, (12<<4)+8);
}
#define init_13_7_12_9
void FASTCALL call_13_7_12_9(SH2_struct* sh)
{
  SH2movli(sh, 7, (12<<4)+9);
}
#define init_13_7_12_10
void FASTCALL call_13_7_12_10(SH2_struct* sh)
{
  SH2movli(sh, 7, (12<<4)+10);
}
#define init_13_7_12_11
void FASTCALL call_13_7_12_11(SH2_struct* sh)
{
  SH2movli(sh, 7, (12<<4)+11);
}
#define init_13_7_12_12
void FASTCALL call_13_7_12_12(SH2_struct* sh)
{
  SH2movli(sh, 7, (12<<4)+12);
}
#define init_13_7_12_13
void FASTCALL call_13_7_12_13(SH2_struct* sh)
{
  SH2movli(sh, 7, (12<<4)+13);
}
#define init_13_7_12_14
void FASTCALL call_13_7_12_14(SH2_struct* sh)
{
  SH2movli(sh, 7, (12<<4)+14);
}
#define init_13_7_12_15
void FASTCALL call_13_7_12_15(SH2_struct* sh)
{
  SH2movli(sh, 7, (12<<4)+15);
}
#define init_13_7_13_0
void FASTCALL call_13_7_13_0(SH2_struct* sh)
{
  SH2movli(sh, 7, (13<<4)+0);
}
#define init_13_7_13_1
void FASTCALL call_13_7_13_1(SH2_struct* sh)
{
  SH2movli(sh, 7, (13<<4)+1);
}
#define init_13_7_13_2
void FASTCALL call_13_7_13_2(SH2_struct* sh)
{
  SH2movli(sh, 7, (13<<4)+2);
}
#define init_13_7_13_3
void FASTCALL call_13_7_13_3(SH2_struct* sh)
{
  SH2movli(sh, 7, (13<<4)+3);
}
#define init_13_7_13_4
void FASTCALL call_13_7_13_4(SH2_struct* sh)
{
  SH2movli(sh, 7, (13<<4)+4);
}
#define init_13_7_13_5
void FASTCALL call_13_7_13_5(SH2_struct* sh)
{
  SH2movli(sh, 7, (13<<4)+5);
}
#define init_13_7_13_6
void FASTCALL call_13_7_13_6(SH2_struct* sh)
{
  SH2movli(sh, 7, (13<<4)+6);
}
#define init_13_7_13_7
void FASTCALL call_13_7_13_7(SH2_struct* sh)
{
  SH2movli(sh, 7, (13<<4)+7);
}
#define init_13_7_13_8
void FASTCALL call_13_7_13_8(SH2_struct* sh)
{
  SH2movli(sh, 7, (13<<4)+8);
}
#define init_13_7_13_9
void FASTCALL call_13_7_13_9(SH2_struct* sh)
{
  SH2movli(sh, 7, (13<<4)+9);
}
#define init_13_7_13_10
void FASTCALL call_13_7_13_10(SH2_struct* sh)
{
  SH2movli(sh, 7, (13<<4)+10);
}
#define init_13_7_13_11
void FASTCALL call_13_7_13_11(SH2_struct* sh)
{
  SH2movli(sh, 7, (13<<4)+11);
}
#define init_13_7_13_12
void FASTCALL call_13_7_13_12(SH2_struct* sh)
{
  SH2movli(sh, 7, (13<<4)+12);
}
#define init_13_7_13_13
void FASTCALL call_13_7_13_13(SH2_struct* sh)
{
  SH2movli(sh, 7, (13<<4)+13);
}
#define init_13_7_13_14
void FASTCALL call_13_7_13_14(SH2_struct* sh)
{
  SH2movli(sh, 7, (13<<4)+14);
}
#define init_13_7_13_15
void FASTCALL call_13_7_13_15(SH2_struct* sh)
{
  SH2movli(sh, 7, (13<<4)+15);
}
#define init_13_7_14_0
void FASTCALL call_13_7_14_0(SH2_struct* sh)
{
  SH2movli(sh, 7, (14<<4)+0);
}
#define init_13_7_14_1
void FASTCALL call_13_7_14_1(SH2_struct* sh)
{
  SH2movli(sh, 7, (14<<4)+1);
}
#define init_13_7_14_2
void FASTCALL call_13_7_14_2(SH2_struct* sh)
{
  SH2movli(sh, 7, (14<<4)+2);
}
#define init_13_7_14_3
void FASTCALL call_13_7_14_3(SH2_struct* sh)
{
  SH2movli(sh, 7, (14<<4)+3);
}
#define init_13_7_14_4
void FASTCALL call_13_7_14_4(SH2_struct* sh)
{
  SH2movli(sh, 7, (14<<4)+4);
}
#define init_13_7_14_5
void FASTCALL call_13_7_14_5(SH2_struct* sh)
{
  SH2movli(sh, 7, (14<<4)+5);
}
#define init_13_7_14_6
void FASTCALL call_13_7_14_6(SH2_struct* sh)
{
  SH2movli(sh, 7, (14<<4)+6);
}
#define init_13_7_14_7
void FASTCALL call_13_7_14_7(SH2_struct* sh)
{
  SH2movli(sh, 7, (14<<4)+7);
}
#define init_13_7_14_8
void FASTCALL call_13_7_14_8(SH2_struct* sh)
{
  SH2movli(sh, 7, (14<<4)+8);
}
#define init_13_7_14_9
void FASTCALL call_13_7_14_9(SH2_struct* sh)
{
  SH2movli(sh, 7, (14<<4)+9);
}
#define init_13_7_14_10
void FASTCALL call_13_7_14_10(SH2_struct* sh)
{
  SH2movli(sh, 7, (14<<4)+10);
}
#define init_13_7_14_11
void FASTCALL call_13_7_14_11(SH2_struct* sh)
{
  SH2movli(sh, 7, (14<<4)+11);
}
#define init_13_7_14_12
void FASTCALL call_13_7_14_12(SH2_struct* sh)
{
  SH2movli(sh, 7, (14<<4)+12);
}
#define init_13_7_14_13
void FASTCALL call_13_7_14_13(SH2_struct* sh)
{
  SH2movli(sh, 7, (14<<4)+13);
}
#define init_13_7_14_14
void FASTCALL call_13_7_14_14(SH2_struct* sh)
{
  SH2movli(sh, 7, (14<<4)+14);
}
#define init_13_7_14_15
void FASTCALL call_13_7_14_15(SH2_struct* sh)
{
  SH2movli(sh, 7, (14<<4)+15);
}
#define init_13_7_15_0
void FASTCALL call_13_7_15_0(SH2_struct* sh)
{
  SH2movli(sh, 7, (15<<4)+0);
}
#define init_13_7_15_1
void FASTCALL call_13_7_15_1(SH2_struct* sh)
{
  SH2movli(sh, 7, (15<<4)+1);
}
#define init_13_7_15_2
void FASTCALL call_13_7_15_2(SH2_struct* sh)
{
  SH2movli(sh, 7, (15<<4)+2);
}
#define init_13_7_15_3
void FASTCALL call_13_7_15_3(SH2_struct* sh)
{
  SH2movli(sh, 7, (15<<4)+3);
}
#define init_13_7_15_4
void FASTCALL call_13_7_15_4(SH2_struct* sh)
{
  SH2movli(sh, 7, (15<<4)+4);
}
#define init_13_7_15_5
void FASTCALL call_13_7_15_5(SH2_struct* sh)
{
  SH2movli(sh, 7, (15<<4)+5);
}
#define init_13_7_15_6
void FASTCALL call_13_7_15_6(SH2_struct* sh)
{
  SH2movli(sh, 7, (15<<4)+6);
}
#define init_13_7_15_7
void FASTCALL call_13_7_15_7(SH2_struct* sh)
{
  SH2movli(sh, 7, (15<<4)+7);
}
#define init_13_7_15_8
void FASTCALL call_13_7_15_8(SH2_struct* sh)
{
  SH2movli(sh, 7, (15<<4)+8);
}
#define init_13_7_15_9
void FASTCALL call_13_7_15_9(SH2_struct* sh)
{
  SH2movli(sh, 7, (15<<4)+9);
}
#define init_13_7_15_10
void FASTCALL call_13_7_15_10(SH2_struct* sh)
{
  SH2movli(sh, 7, (15<<4)+10);
}
#define init_13_7_15_11
void FASTCALL call_13_7_15_11(SH2_struct* sh)
{
  SH2movli(sh, 7, (15<<4)+11);
}
#define init_13_7_15_12
void FASTCALL call_13_7_15_12(SH2_struct* sh)
{
  SH2movli(sh, 7, (15<<4)+12);
}
#define init_13_7_15_13
void FASTCALL call_13_7_15_13(SH2_struct* sh)
{
  SH2movli(sh, 7, (15<<4)+13);
}
#define init_13_7_15_14
void FASTCALL call_13_7_15_14(SH2_struct* sh)
{
  SH2movli(sh, 7, (15<<4)+14);
}
#define init_13_7_15_15
void FASTCALL call_13_7_15_15(SH2_struct* sh)
{
  SH2movli(sh, 7, (15<<4)+15);
}
#define init_13_8_0_0
void FASTCALL call_13_8_0_0(SH2_struct* sh)
{
  SH2movli(sh, 8, (0<<4)+0);
}
#define init_13_8_0_1
void FASTCALL call_13_8_0_1(SH2_struct* sh)
{
  SH2movli(sh, 8, (0<<4)+1);
}
#define init_13_8_0_2
void FASTCALL call_13_8_0_2(SH2_struct* sh)
{
  SH2movli(sh, 8, (0<<4)+2);
}
#define init_13_8_0_3
void FASTCALL call_13_8_0_3(SH2_struct* sh)
{
  SH2movli(sh, 8, (0<<4)+3);
}
#define init_13_8_0_4
void FASTCALL call_13_8_0_4(SH2_struct* sh)
{
  SH2movli(sh, 8, (0<<4)+4);
}
#define init_13_8_0_5
void FASTCALL call_13_8_0_5(SH2_struct* sh)
{
  SH2movli(sh, 8, (0<<4)+5);
}
#define init_13_8_0_6
void FASTCALL call_13_8_0_6(SH2_struct* sh)
{
  SH2movli(sh, 8, (0<<4)+6);
}
#define init_13_8_0_7
void FASTCALL call_13_8_0_7(SH2_struct* sh)
{
  SH2movli(sh, 8, (0<<4)+7);
}
#define init_13_8_0_8
void FASTCALL call_13_8_0_8(SH2_struct* sh)
{
  SH2movli(sh, 8, (0<<4)+8);
}
#define init_13_8_0_9
void FASTCALL call_13_8_0_9(SH2_struct* sh)
{
  SH2movli(sh, 8, (0<<4)+9);
}
#define init_13_8_0_10
void FASTCALL call_13_8_0_10(SH2_struct* sh)
{
  SH2movli(sh, 8, (0<<4)+10);
}
#define init_13_8_0_11
void FASTCALL call_13_8_0_11(SH2_struct* sh)
{
  SH2movli(sh, 8, (0<<4)+11);
}
#define init_13_8_0_12
void FASTCALL call_13_8_0_12(SH2_struct* sh)
{
  SH2movli(sh, 8, (0<<4)+12);
}
#define init_13_8_0_13
void FASTCALL call_13_8_0_13(SH2_struct* sh)
{
  SH2movli(sh, 8, (0<<4)+13);
}
#define init_13_8_0_14
void FASTCALL call_13_8_0_14(SH2_struct* sh)
{
  SH2movli(sh, 8, (0<<4)+14);
}
#define init_13_8_0_15
void FASTCALL call_13_8_0_15(SH2_struct* sh)
{
  SH2movli(sh, 8, (0<<4)+15);
}
#define init_13_8_1_0
void FASTCALL call_13_8_1_0(SH2_struct* sh)
{
  SH2movli(sh, 8, (1<<4)+0);
}
#define init_13_8_1_1
void FASTCALL call_13_8_1_1(SH2_struct* sh)
{
  SH2movli(sh, 8, (1<<4)+1);
}
#define init_13_8_1_2
void FASTCALL call_13_8_1_2(SH2_struct* sh)
{
  SH2movli(sh, 8, (1<<4)+2);
}
#define init_13_8_1_3
void FASTCALL call_13_8_1_3(SH2_struct* sh)
{
  SH2movli(sh, 8, (1<<4)+3);
}
#define init_13_8_1_4
void FASTCALL call_13_8_1_4(SH2_struct* sh)
{
  SH2movli(sh, 8, (1<<4)+4);
}
#define init_13_8_1_5
void FASTCALL call_13_8_1_5(SH2_struct* sh)
{
  SH2movli(sh, 8, (1<<4)+5);
}
#define init_13_8_1_6
void FASTCALL call_13_8_1_6(SH2_struct* sh)
{
  SH2movli(sh, 8, (1<<4)+6);
}
#define init_13_8_1_7
void FASTCALL call_13_8_1_7(SH2_struct* sh)
{
  SH2movli(sh, 8, (1<<4)+7);
}
#define init_13_8_1_8
void FASTCALL call_13_8_1_8(SH2_struct* sh)
{
  SH2movli(sh, 8, (1<<4)+8);
}
#define init_13_8_1_9
void FASTCALL call_13_8_1_9(SH2_struct* sh)
{
  SH2movli(sh, 8, (1<<4)+9);
}
#define init_13_8_1_10
void FASTCALL call_13_8_1_10(SH2_struct* sh)
{
  SH2movli(sh, 8, (1<<4)+10);
}
#define init_13_8_1_11
void FASTCALL call_13_8_1_11(SH2_struct* sh)
{
  SH2movli(sh, 8, (1<<4)+11);
}
#define init_13_8_1_12
void FASTCALL call_13_8_1_12(SH2_struct* sh)
{
  SH2movli(sh, 8, (1<<4)+12);
}
#define init_13_8_1_13
void FASTCALL call_13_8_1_13(SH2_struct* sh)
{
  SH2movli(sh, 8, (1<<4)+13);
}
#define init_13_8_1_14
void FASTCALL call_13_8_1_14(SH2_struct* sh)
{
  SH2movli(sh, 8, (1<<4)+14);
}
#define init_13_8_1_15
void FASTCALL call_13_8_1_15(SH2_struct* sh)
{
  SH2movli(sh, 8, (1<<4)+15);
}
#define init_13_8_2_0
void FASTCALL call_13_8_2_0(SH2_struct* sh)
{
  SH2movli(sh, 8, (2<<4)+0);
}
#define init_13_8_2_1
void FASTCALL call_13_8_2_1(SH2_struct* sh)
{
  SH2movli(sh, 8, (2<<4)+1);
}
#define init_13_8_2_2
void FASTCALL call_13_8_2_2(SH2_struct* sh)
{
  SH2movli(sh, 8, (2<<4)+2);
}
#define init_13_8_2_3
void FASTCALL call_13_8_2_3(SH2_struct* sh)
{
  SH2movli(sh, 8, (2<<4)+3);
}
#define init_13_8_2_4
void FASTCALL call_13_8_2_4(SH2_struct* sh)
{
  SH2movli(sh, 8, (2<<4)+4);
}
#define init_13_8_2_5
void FASTCALL call_13_8_2_5(SH2_struct* sh)
{
  SH2movli(sh, 8, (2<<4)+5);
}
#define init_13_8_2_6
void FASTCALL call_13_8_2_6(SH2_struct* sh)
{
  SH2movli(sh, 8, (2<<4)+6);
}
#define init_13_8_2_7
void FASTCALL call_13_8_2_7(SH2_struct* sh)
{
  SH2movli(sh, 8, (2<<4)+7);
}
#define init_13_8_2_8
void FASTCALL call_13_8_2_8(SH2_struct* sh)
{
  SH2movli(sh, 8, (2<<4)+8);
}
#define init_13_8_2_9
void FASTCALL call_13_8_2_9(SH2_struct* sh)
{
  SH2movli(sh, 8, (2<<4)+9);
}
#define init_13_8_2_10
void FASTCALL call_13_8_2_10(SH2_struct* sh)
{
  SH2movli(sh, 8, (2<<4)+10);
}
#define init_13_8_2_11
void FASTCALL call_13_8_2_11(SH2_struct* sh)
{
  SH2movli(sh, 8, (2<<4)+11);
}
#define init_13_8_2_12
void FASTCALL call_13_8_2_12(SH2_struct* sh)
{
  SH2movli(sh, 8, (2<<4)+12);
}
#define init_13_8_2_13
void FASTCALL call_13_8_2_13(SH2_struct* sh)
{
  SH2movli(sh, 8, (2<<4)+13);
}
#define init_13_8_2_14
void FASTCALL call_13_8_2_14(SH2_struct* sh)
{
  SH2movli(sh, 8, (2<<4)+14);
}
#define init_13_8_2_15
void FASTCALL call_13_8_2_15(SH2_struct* sh)
{
  SH2movli(sh, 8, (2<<4)+15);
}
#define init_13_8_3_0
void FASTCALL call_13_8_3_0(SH2_struct* sh)
{
  SH2movli(sh, 8, (3<<4)+0);
}
#define init_13_8_3_1
void FASTCALL call_13_8_3_1(SH2_struct* sh)
{
  SH2movli(sh, 8, (3<<4)+1);
}
#define init_13_8_3_2
void FASTCALL call_13_8_3_2(SH2_struct* sh)
{
  SH2movli(sh, 8, (3<<4)+2);
}
#define init_13_8_3_3
void FASTCALL call_13_8_3_3(SH2_struct* sh)
{
  SH2movli(sh, 8, (3<<4)+3);
}
#define init_13_8_3_4
void FASTCALL call_13_8_3_4(SH2_struct* sh)
{
  SH2movli(sh, 8, (3<<4)+4);
}
#define init_13_8_3_5
void FASTCALL call_13_8_3_5(SH2_struct* sh)
{
  SH2movli(sh, 8, (3<<4)+5);
}
#define init_13_8_3_6
void FASTCALL call_13_8_3_6(SH2_struct* sh)
{
  SH2movli(sh, 8, (3<<4)+6);
}
#define init_13_8_3_7
void FASTCALL call_13_8_3_7(SH2_struct* sh)
{
  SH2movli(sh, 8, (3<<4)+7);
}
#define init_13_8_3_8
void FASTCALL call_13_8_3_8(SH2_struct* sh)
{
  SH2movli(sh, 8, (3<<4)+8);
}
#define init_13_8_3_9
void FASTCALL call_13_8_3_9(SH2_struct* sh)
{
  SH2movli(sh, 8, (3<<4)+9);
}
#define init_13_8_3_10
void FASTCALL call_13_8_3_10(SH2_struct* sh)
{
  SH2movli(sh, 8, (3<<4)+10);
}
#define init_13_8_3_11
void FASTCALL call_13_8_3_11(SH2_struct* sh)
{
  SH2movli(sh, 8, (3<<4)+11);
}
#define init_13_8_3_12
void FASTCALL call_13_8_3_12(SH2_struct* sh)
{
  SH2movli(sh, 8, (3<<4)+12);
}
#define init_13_8_3_13
void FASTCALL call_13_8_3_13(SH2_struct* sh)
{
  SH2movli(sh, 8, (3<<4)+13);
}
#define init_13_8_3_14
void FASTCALL call_13_8_3_14(SH2_struct* sh)
{
  SH2movli(sh, 8, (3<<4)+14);
}
#define init_13_8_3_15
void FASTCALL call_13_8_3_15(SH2_struct* sh)
{
  SH2movli(sh, 8, (3<<4)+15);
}
#define init_13_8_4_0
void FASTCALL call_13_8_4_0(SH2_struct* sh)
{
  SH2movli(sh, 8, (4<<4)+0);
}
#define init_13_8_4_1
void FASTCALL call_13_8_4_1(SH2_struct* sh)
{
  SH2movli(sh, 8, (4<<4)+1);
}
#define init_13_8_4_2
void FASTCALL call_13_8_4_2(SH2_struct* sh)
{
  SH2movli(sh, 8, (4<<4)+2);
}
#define init_13_8_4_3
void FASTCALL call_13_8_4_3(SH2_struct* sh)
{
  SH2movli(sh, 8, (4<<4)+3);
}
#define init_13_8_4_4
void FASTCALL call_13_8_4_4(SH2_struct* sh)
{
  SH2movli(sh, 8, (4<<4)+4);
}
#define init_13_8_4_5
void FASTCALL call_13_8_4_5(SH2_struct* sh)
{
  SH2movli(sh, 8, (4<<4)+5);
}
#define init_13_8_4_6
void FASTCALL call_13_8_4_6(SH2_struct* sh)
{
  SH2movli(sh, 8, (4<<4)+6);
}
#define init_13_8_4_7
void FASTCALL call_13_8_4_7(SH2_struct* sh)
{
  SH2movli(sh, 8, (4<<4)+7);
}
#define init_13_8_4_8
void FASTCALL call_13_8_4_8(SH2_struct* sh)
{
  SH2movli(sh, 8, (4<<4)+8);
}
#define init_13_8_4_9
void FASTCALL call_13_8_4_9(SH2_struct* sh)
{
  SH2movli(sh, 8, (4<<4)+9);
}
#define init_13_8_4_10
void FASTCALL call_13_8_4_10(SH2_struct* sh)
{
  SH2movli(sh, 8, (4<<4)+10);
}
#define init_13_8_4_11
void FASTCALL call_13_8_4_11(SH2_struct* sh)
{
  SH2movli(sh, 8, (4<<4)+11);
}
#define init_13_8_4_12
void FASTCALL call_13_8_4_12(SH2_struct* sh)
{
  SH2movli(sh, 8, (4<<4)+12);
}
#define init_13_8_4_13
void FASTCALL call_13_8_4_13(SH2_struct* sh)
{
  SH2movli(sh, 8, (4<<4)+13);
}
#define init_13_8_4_14
void FASTCALL call_13_8_4_14(SH2_struct* sh)
{
  SH2movli(sh, 8, (4<<4)+14);
}
#define init_13_8_4_15
void FASTCALL call_13_8_4_15(SH2_struct* sh)
{
  SH2movli(sh, 8, (4<<4)+15);
}
#define init_13_8_5_0
void FASTCALL call_13_8_5_0(SH2_struct* sh)
{
  SH2movli(sh, 8, (5<<4)+0);
}
#define init_13_8_5_1
void FASTCALL call_13_8_5_1(SH2_struct* sh)
{
  SH2movli(sh, 8, (5<<4)+1);
}
#define init_13_8_5_2
void FASTCALL call_13_8_5_2(SH2_struct* sh)
{
  SH2movli(sh, 8, (5<<4)+2);
}
#define init_13_8_5_3
void FASTCALL call_13_8_5_3(SH2_struct* sh)
{
  SH2movli(sh, 8, (5<<4)+3);
}
#define init_13_8_5_4
void FASTCALL call_13_8_5_4(SH2_struct* sh)
{
  SH2movli(sh, 8, (5<<4)+4);
}
#define init_13_8_5_5
void FASTCALL call_13_8_5_5(SH2_struct* sh)
{
  SH2movli(sh, 8, (5<<4)+5);
}
#define init_13_8_5_6
void FASTCALL call_13_8_5_6(SH2_struct* sh)
{
  SH2movli(sh, 8, (5<<4)+6);
}
#define init_13_8_5_7
void FASTCALL call_13_8_5_7(SH2_struct* sh)
{
  SH2movli(sh, 8, (5<<4)+7);
}
#define init_13_8_5_8
void FASTCALL call_13_8_5_8(SH2_struct* sh)
{
  SH2movli(sh, 8, (5<<4)+8);
}
#define init_13_8_5_9
void FASTCALL call_13_8_5_9(SH2_struct* sh)
{
  SH2movli(sh, 8, (5<<4)+9);
}
#define init_13_8_5_10
void FASTCALL call_13_8_5_10(SH2_struct* sh)
{
  SH2movli(sh, 8, (5<<4)+10);
}
#define init_13_8_5_11
void FASTCALL call_13_8_5_11(SH2_struct* sh)
{
  SH2movli(sh, 8, (5<<4)+11);
}
#define init_13_8_5_12
void FASTCALL call_13_8_5_12(SH2_struct* sh)
{
  SH2movli(sh, 8, (5<<4)+12);
}
#define init_13_8_5_13
void FASTCALL call_13_8_5_13(SH2_struct* sh)
{
  SH2movli(sh, 8, (5<<4)+13);
}
#define init_13_8_5_14
void FASTCALL call_13_8_5_14(SH2_struct* sh)
{
  SH2movli(sh, 8, (5<<4)+14);
}
#define init_13_8_5_15
void FASTCALL call_13_8_5_15(SH2_struct* sh)
{
  SH2movli(sh, 8, (5<<4)+15);
}
#define init_13_8_6_0
void FASTCALL call_13_8_6_0(SH2_struct* sh)
{
  SH2movli(sh, 8, (6<<4)+0);
}
#define init_13_8_6_1
void FASTCALL call_13_8_6_1(SH2_struct* sh)
{
  SH2movli(sh, 8, (6<<4)+1);
}
#define init_13_8_6_2
void FASTCALL call_13_8_6_2(SH2_struct* sh)
{
  SH2movli(sh, 8, (6<<4)+2);
}
#define init_13_8_6_3
void FASTCALL call_13_8_6_3(SH2_struct* sh)
{
  SH2movli(sh, 8, (6<<4)+3);
}
#define init_13_8_6_4
void FASTCALL call_13_8_6_4(SH2_struct* sh)
{
  SH2movli(sh, 8, (6<<4)+4);
}
#define init_13_8_6_5
void FASTCALL call_13_8_6_5(SH2_struct* sh)
{
  SH2movli(sh, 8, (6<<4)+5);
}
#define init_13_8_6_6
void FASTCALL call_13_8_6_6(SH2_struct* sh)
{
  SH2movli(sh, 8, (6<<4)+6);
}
#define init_13_8_6_7
void FASTCALL call_13_8_6_7(SH2_struct* sh)
{
  SH2movli(sh, 8, (6<<4)+7);
}
#define init_13_8_6_8
void FASTCALL call_13_8_6_8(SH2_struct* sh)
{
  SH2movli(sh, 8, (6<<4)+8);
}
#define init_13_8_6_9
void FASTCALL call_13_8_6_9(SH2_struct* sh)
{
  SH2movli(sh, 8, (6<<4)+9);
}
#define init_13_8_6_10
void FASTCALL call_13_8_6_10(SH2_struct* sh)
{
  SH2movli(sh, 8, (6<<4)+10);
}
#define init_13_8_6_11
void FASTCALL call_13_8_6_11(SH2_struct* sh)
{
  SH2movli(sh, 8, (6<<4)+11);
}
#define init_13_8_6_12
void FASTCALL call_13_8_6_12(SH2_struct* sh)
{
  SH2movli(sh, 8, (6<<4)+12);
}
#define init_13_8_6_13
void FASTCALL call_13_8_6_13(SH2_struct* sh)
{
  SH2movli(sh, 8, (6<<4)+13);
}
#define init_13_8_6_14
void FASTCALL call_13_8_6_14(SH2_struct* sh)
{
  SH2movli(sh, 8, (6<<4)+14);
}
#define init_13_8_6_15
void FASTCALL call_13_8_6_15(SH2_struct* sh)
{
  SH2movli(sh, 8, (6<<4)+15);
}
#define init_13_8_7_0
void FASTCALL call_13_8_7_0(SH2_struct* sh)
{
  SH2movli(sh, 8, (7<<4)+0);
}
#define init_13_8_7_1
void FASTCALL call_13_8_7_1(SH2_struct* sh)
{
  SH2movli(sh, 8, (7<<4)+1);
}
#define init_13_8_7_2
void FASTCALL call_13_8_7_2(SH2_struct* sh)
{
  SH2movli(sh, 8, (7<<4)+2);
}
#define init_13_8_7_3
void FASTCALL call_13_8_7_3(SH2_struct* sh)
{
  SH2movli(sh, 8, (7<<4)+3);
}
#define init_13_8_7_4
void FASTCALL call_13_8_7_4(SH2_struct* sh)
{
  SH2movli(sh, 8, (7<<4)+4);
}
#define init_13_8_7_5
void FASTCALL call_13_8_7_5(SH2_struct* sh)
{
  SH2movli(sh, 8, (7<<4)+5);
}
#define init_13_8_7_6
void FASTCALL call_13_8_7_6(SH2_struct* sh)
{
  SH2movli(sh, 8, (7<<4)+6);
}
#define init_13_8_7_7
void FASTCALL call_13_8_7_7(SH2_struct* sh)
{
  SH2movli(sh, 8, (7<<4)+7);
}
#define init_13_8_7_8
void FASTCALL call_13_8_7_8(SH2_struct* sh)
{
  SH2movli(sh, 8, (7<<4)+8);
}
#define init_13_8_7_9
void FASTCALL call_13_8_7_9(SH2_struct* sh)
{
  SH2movli(sh, 8, (7<<4)+9);
}
#define init_13_8_7_10
void FASTCALL call_13_8_7_10(SH2_struct* sh)
{
  SH2movli(sh, 8, (7<<4)+10);
}
#define init_13_8_7_11
void FASTCALL call_13_8_7_11(SH2_struct* sh)
{
  SH2movli(sh, 8, (7<<4)+11);
}
#define init_13_8_7_12
void FASTCALL call_13_8_7_12(SH2_struct* sh)
{
  SH2movli(sh, 8, (7<<4)+12);
}
#define init_13_8_7_13
void FASTCALL call_13_8_7_13(SH2_struct* sh)
{
  SH2movli(sh, 8, (7<<4)+13);
}
#define init_13_8_7_14
void FASTCALL call_13_8_7_14(SH2_struct* sh)
{
  SH2movli(sh, 8, (7<<4)+14);
}
#define init_13_8_7_15
void FASTCALL call_13_8_7_15(SH2_struct* sh)
{
  SH2movli(sh, 8, (7<<4)+15);
}
#define init_13_8_8_0
void FASTCALL call_13_8_8_0(SH2_struct* sh)
{
  SH2movli(sh, 8, (8<<4)+0);
}
#define init_13_8_8_1
void FASTCALL call_13_8_8_1(SH2_struct* sh)
{
  SH2movli(sh, 8, (8<<4)+1);
}
#define init_13_8_8_2
void FASTCALL call_13_8_8_2(SH2_struct* sh)
{
  SH2movli(sh, 8, (8<<4)+2);
}
#define init_13_8_8_3
void FASTCALL call_13_8_8_3(SH2_struct* sh)
{
  SH2movli(sh, 8, (8<<4)+3);
}
#define init_13_8_8_4
void FASTCALL call_13_8_8_4(SH2_struct* sh)
{
  SH2movli(sh, 8, (8<<4)+4);
}
#define init_13_8_8_5
void FASTCALL call_13_8_8_5(SH2_struct* sh)
{
  SH2movli(sh, 8, (8<<4)+5);
}
#define init_13_8_8_6
void FASTCALL call_13_8_8_6(SH2_struct* sh)
{
  SH2movli(sh, 8, (8<<4)+6);
}
#define init_13_8_8_7
void FASTCALL call_13_8_8_7(SH2_struct* sh)
{
  SH2movli(sh, 8, (8<<4)+7);
}
#define init_13_8_8_8
void FASTCALL call_13_8_8_8(SH2_struct* sh)
{
  SH2movli(sh, 8, (8<<4)+8);
}
#define init_13_8_8_9
void FASTCALL call_13_8_8_9(SH2_struct* sh)
{
  SH2movli(sh, 8, (8<<4)+9);
}
#define init_13_8_8_10
void FASTCALL call_13_8_8_10(SH2_struct* sh)
{
  SH2movli(sh, 8, (8<<4)+10);
}
#define init_13_8_8_11
void FASTCALL call_13_8_8_11(SH2_struct* sh)
{
  SH2movli(sh, 8, (8<<4)+11);
}
#define init_13_8_8_12
void FASTCALL call_13_8_8_12(SH2_struct* sh)
{
  SH2movli(sh, 8, (8<<4)+12);
}
#define init_13_8_8_13
void FASTCALL call_13_8_8_13(SH2_struct* sh)
{
  SH2movli(sh, 8, (8<<4)+13);
}
#define init_13_8_8_14
void FASTCALL call_13_8_8_14(SH2_struct* sh)
{
  SH2movli(sh, 8, (8<<4)+14);
}
#define init_13_8_8_15
void FASTCALL call_13_8_8_15(SH2_struct* sh)
{
  SH2movli(sh, 8, (8<<4)+15);
}
#define init_13_8_9_0
void FASTCALL call_13_8_9_0(SH2_struct* sh)
{
  SH2movli(sh, 8, (9<<4)+0);
}
#define init_13_8_9_1
void FASTCALL call_13_8_9_1(SH2_struct* sh)
{
  SH2movli(sh, 8, (9<<4)+1);
}
#define init_13_8_9_2
void FASTCALL call_13_8_9_2(SH2_struct* sh)
{
  SH2movli(sh, 8, (9<<4)+2);
}
#define init_13_8_9_3
void FASTCALL call_13_8_9_3(SH2_struct* sh)
{
  SH2movli(sh, 8, (9<<4)+3);
}
#define init_13_8_9_4
void FASTCALL call_13_8_9_4(SH2_struct* sh)
{
  SH2movli(sh, 8, (9<<4)+4);
}
#define init_13_8_9_5
void FASTCALL call_13_8_9_5(SH2_struct* sh)
{
  SH2movli(sh, 8, (9<<4)+5);
}
#define init_13_8_9_6
void FASTCALL call_13_8_9_6(SH2_struct* sh)
{
  SH2movli(sh, 8, (9<<4)+6);
}
#define init_13_8_9_7
void FASTCALL call_13_8_9_7(SH2_struct* sh)
{
  SH2movli(sh, 8, (9<<4)+7);
}
#define init_13_8_9_8
void FASTCALL call_13_8_9_8(SH2_struct* sh)
{
  SH2movli(sh, 8, (9<<4)+8);
}
#define init_13_8_9_9
void FASTCALL call_13_8_9_9(SH2_struct* sh)
{
  SH2movli(sh, 8, (9<<4)+9);
}
#define init_13_8_9_10
void FASTCALL call_13_8_9_10(SH2_struct* sh)
{
  SH2movli(sh, 8, (9<<4)+10);
}
#define init_13_8_9_11
void FASTCALL call_13_8_9_11(SH2_struct* sh)
{
  SH2movli(sh, 8, (9<<4)+11);
}
#define init_13_8_9_12
void FASTCALL call_13_8_9_12(SH2_struct* sh)
{
  SH2movli(sh, 8, (9<<4)+12);
}
#define init_13_8_9_13
void FASTCALL call_13_8_9_13(SH2_struct* sh)
{
  SH2movli(sh, 8, (9<<4)+13);
}
#define init_13_8_9_14
void FASTCALL call_13_8_9_14(SH2_struct* sh)
{
  SH2movli(sh, 8, (9<<4)+14);
}
#define init_13_8_9_15
void FASTCALL call_13_8_9_15(SH2_struct* sh)
{
  SH2movli(sh, 8, (9<<4)+15);
}
#define init_13_8_10_0
void FASTCALL call_13_8_10_0(SH2_struct* sh)
{
  SH2movli(sh, 8, (10<<4)+0);
}
#define init_13_8_10_1
void FASTCALL call_13_8_10_1(SH2_struct* sh)
{
  SH2movli(sh, 8, (10<<4)+1);
}
#define init_13_8_10_2
void FASTCALL call_13_8_10_2(SH2_struct* sh)
{
  SH2movli(sh, 8, (10<<4)+2);
}
#define init_13_8_10_3
void FASTCALL call_13_8_10_3(SH2_struct* sh)
{
  SH2movli(sh, 8, (10<<4)+3);
}
#define init_13_8_10_4
void FASTCALL call_13_8_10_4(SH2_struct* sh)
{
  SH2movli(sh, 8, (10<<4)+4);
}
#define init_13_8_10_5
void FASTCALL call_13_8_10_5(SH2_struct* sh)
{
  SH2movli(sh, 8, (10<<4)+5);
}
#define init_13_8_10_6
void FASTCALL call_13_8_10_6(SH2_struct* sh)
{
  SH2movli(sh, 8, (10<<4)+6);
}
#define init_13_8_10_7
void FASTCALL call_13_8_10_7(SH2_struct* sh)
{
  SH2movli(sh, 8, (10<<4)+7);
}
#define init_13_8_10_8
void FASTCALL call_13_8_10_8(SH2_struct* sh)
{
  SH2movli(sh, 8, (10<<4)+8);
}
#define init_13_8_10_9
void FASTCALL call_13_8_10_9(SH2_struct* sh)
{
  SH2movli(sh, 8, (10<<4)+9);
}
#define init_13_8_10_10
void FASTCALL call_13_8_10_10(SH2_struct* sh)
{
  SH2movli(sh, 8, (10<<4)+10);
}
#define init_13_8_10_11
void FASTCALL call_13_8_10_11(SH2_struct* sh)
{
  SH2movli(sh, 8, (10<<4)+11);
}
#define init_13_8_10_12
void FASTCALL call_13_8_10_12(SH2_struct* sh)
{
  SH2movli(sh, 8, (10<<4)+12);
}
#define init_13_8_10_13
void FASTCALL call_13_8_10_13(SH2_struct* sh)
{
  SH2movli(sh, 8, (10<<4)+13);
}
#define init_13_8_10_14
void FASTCALL call_13_8_10_14(SH2_struct* sh)
{
  SH2movli(sh, 8, (10<<4)+14);
}
#define init_13_8_10_15
void FASTCALL call_13_8_10_15(SH2_struct* sh)
{
  SH2movli(sh, 8, (10<<4)+15);
}
#define init_13_8_11_0
void FASTCALL call_13_8_11_0(SH2_struct* sh)
{
  SH2movli(sh, 8, (11<<4)+0);
}
#define init_13_8_11_1
void FASTCALL call_13_8_11_1(SH2_struct* sh)
{
  SH2movli(sh, 8, (11<<4)+1);
}
#define init_13_8_11_2
void FASTCALL call_13_8_11_2(SH2_struct* sh)
{
  SH2movli(sh, 8, (11<<4)+2);
}
#define init_13_8_11_3
void FASTCALL call_13_8_11_3(SH2_struct* sh)
{
  SH2movli(sh, 8, (11<<4)+3);
}
#define init_13_8_11_4
void FASTCALL call_13_8_11_4(SH2_struct* sh)
{
  SH2movli(sh, 8, (11<<4)+4);
}
#define init_13_8_11_5
void FASTCALL call_13_8_11_5(SH2_struct* sh)
{
  SH2movli(sh, 8, (11<<4)+5);
}
#define init_13_8_11_6
void FASTCALL call_13_8_11_6(SH2_struct* sh)
{
  SH2movli(sh, 8, (11<<4)+6);
}
#define init_13_8_11_7
void FASTCALL call_13_8_11_7(SH2_struct* sh)
{
  SH2movli(sh, 8, (11<<4)+7);
}
#define init_13_8_11_8
void FASTCALL call_13_8_11_8(SH2_struct* sh)
{
  SH2movli(sh, 8, (11<<4)+8);
}
#define init_13_8_11_9
void FASTCALL call_13_8_11_9(SH2_struct* sh)
{
  SH2movli(sh, 8, (11<<4)+9);
}
#define init_13_8_11_10
void FASTCALL call_13_8_11_10(SH2_struct* sh)
{
  SH2movli(sh, 8, (11<<4)+10);
}
#define init_13_8_11_11
void FASTCALL call_13_8_11_11(SH2_struct* sh)
{
  SH2movli(sh, 8, (11<<4)+11);
}
#define init_13_8_11_12
void FASTCALL call_13_8_11_12(SH2_struct* sh)
{
  SH2movli(sh, 8, (11<<4)+12);
}
#define init_13_8_11_13
void FASTCALL call_13_8_11_13(SH2_struct* sh)
{
  SH2movli(sh, 8, (11<<4)+13);
}
#define init_13_8_11_14
void FASTCALL call_13_8_11_14(SH2_struct* sh)
{
  SH2movli(sh, 8, (11<<4)+14);
}
#define init_13_8_11_15
void FASTCALL call_13_8_11_15(SH2_struct* sh)
{
  SH2movli(sh, 8, (11<<4)+15);
}
#define init_13_8_12_0
void FASTCALL call_13_8_12_0(SH2_struct* sh)
{
  SH2movli(sh, 8, (12<<4)+0);
}
#define init_13_8_12_1
void FASTCALL call_13_8_12_1(SH2_struct* sh)
{
  SH2movli(sh, 8, (12<<4)+1);
}
#define init_13_8_12_2
void FASTCALL call_13_8_12_2(SH2_struct* sh)
{
  SH2movli(sh, 8, (12<<4)+2);
}
#define init_13_8_12_3
void FASTCALL call_13_8_12_3(SH2_struct* sh)
{
  SH2movli(sh, 8, (12<<4)+3);
}
#define init_13_8_12_4
void FASTCALL call_13_8_12_4(SH2_struct* sh)
{
  SH2movli(sh, 8, (12<<4)+4);
}
#define init_13_8_12_5
void FASTCALL call_13_8_12_5(SH2_struct* sh)
{
  SH2movli(sh, 8, (12<<4)+5);
}
#define init_13_8_12_6
void FASTCALL call_13_8_12_6(SH2_struct* sh)
{
  SH2movli(sh, 8, (12<<4)+6);
}
#define init_13_8_12_7
void FASTCALL call_13_8_12_7(SH2_struct* sh)
{
  SH2movli(sh, 8, (12<<4)+7);
}
#define init_13_8_12_8
void FASTCALL call_13_8_12_8(SH2_struct* sh)
{
  SH2movli(sh, 8, (12<<4)+8);
}
#define init_13_8_12_9
void FASTCALL call_13_8_12_9(SH2_struct* sh)
{
  SH2movli(sh, 8, (12<<4)+9);
}
#define init_13_8_12_10
void FASTCALL call_13_8_12_10(SH2_struct* sh)
{
  SH2movli(sh, 8, (12<<4)+10);
}
#define init_13_8_12_11
void FASTCALL call_13_8_12_11(SH2_struct* sh)
{
  SH2movli(sh, 8, (12<<4)+11);
}
#define init_13_8_12_12
void FASTCALL call_13_8_12_12(SH2_struct* sh)
{
  SH2movli(sh, 8, (12<<4)+12);
}
#define init_13_8_12_13
void FASTCALL call_13_8_12_13(SH2_struct* sh)
{
  SH2movli(sh, 8, (12<<4)+13);
}
#define init_13_8_12_14
void FASTCALL call_13_8_12_14(SH2_struct* sh)
{
  SH2movli(sh, 8, (12<<4)+14);
}
#define init_13_8_12_15
void FASTCALL call_13_8_12_15(SH2_struct* sh)
{
  SH2movli(sh, 8, (12<<4)+15);
}
#define init_13_8_13_0
void FASTCALL call_13_8_13_0(SH2_struct* sh)
{
  SH2movli(sh, 8, (13<<4)+0);
}
#define init_13_8_13_1
void FASTCALL call_13_8_13_1(SH2_struct* sh)
{
  SH2movli(sh, 8, (13<<4)+1);
}
#define init_13_8_13_2
void FASTCALL call_13_8_13_2(SH2_struct* sh)
{
  SH2movli(sh, 8, (13<<4)+2);
}
#define init_13_8_13_3
void FASTCALL call_13_8_13_3(SH2_struct* sh)
{
  SH2movli(sh, 8, (13<<4)+3);
}
#define init_13_8_13_4
void FASTCALL call_13_8_13_4(SH2_struct* sh)
{
  SH2movli(sh, 8, (13<<4)+4);
}
#define init_13_8_13_5
void FASTCALL call_13_8_13_5(SH2_struct* sh)
{
  SH2movli(sh, 8, (13<<4)+5);
}
#define init_13_8_13_6
void FASTCALL call_13_8_13_6(SH2_struct* sh)
{
  SH2movli(sh, 8, (13<<4)+6);
}
#define init_13_8_13_7
void FASTCALL call_13_8_13_7(SH2_struct* sh)
{
  SH2movli(sh, 8, (13<<4)+7);
}
#define init_13_8_13_8
void FASTCALL call_13_8_13_8(SH2_struct* sh)
{
  SH2movli(sh, 8, (13<<4)+8);
}
#define init_13_8_13_9
void FASTCALL call_13_8_13_9(SH2_struct* sh)
{
  SH2movli(sh, 8, (13<<4)+9);
}
#define init_13_8_13_10
void FASTCALL call_13_8_13_10(SH2_struct* sh)
{
  SH2movli(sh, 8, (13<<4)+10);
}
#define init_13_8_13_11
void FASTCALL call_13_8_13_11(SH2_struct* sh)
{
  SH2movli(sh, 8, (13<<4)+11);
}
#define init_13_8_13_12
void FASTCALL call_13_8_13_12(SH2_struct* sh)
{
  SH2movli(sh, 8, (13<<4)+12);
}
#define init_13_8_13_13
void FASTCALL call_13_8_13_13(SH2_struct* sh)
{
  SH2movli(sh, 8, (13<<4)+13);
}
#define init_13_8_13_14
void FASTCALL call_13_8_13_14(SH2_struct* sh)
{
  SH2movli(sh, 8, (13<<4)+14);
}
#define init_13_8_13_15
void FASTCALL call_13_8_13_15(SH2_struct* sh)
{
  SH2movli(sh, 8, (13<<4)+15);
}
#define init_13_8_14_0
void FASTCALL call_13_8_14_0(SH2_struct* sh)
{
  SH2movli(sh, 8, (14<<4)+0);
}
#define init_13_8_14_1
void FASTCALL call_13_8_14_1(SH2_struct* sh)
{
  SH2movli(sh, 8, (14<<4)+1);
}
#define init_13_8_14_2
void FASTCALL call_13_8_14_2(SH2_struct* sh)
{
  SH2movli(sh, 8, (14<<4)+2);
}
#define init_13_8_14_3
void FASTCALL call_13_8_14_3(SH2_struct* sh)
{
  SH2movli(sh, 8, (14<<4)+3);
}
#define init_13_8_14_4
void FASTCALL call_13_8_14_4(SH2_struct* sh)
{
  SH2movli(sh, 8, (14<<4)+4);
}
#define init_13_8_14_5
void FASTCALL call_13_8_14_5(SH2_struct* sh)
{
  SH2movli(sh, 8, (14<<4)+5);
}
#define init_13_8_14_6
void FASTCALL call_13_8_14_6(SH2_struct* sh)
{
  SH2movli(sh, 8, (14<<4)+6);
}
#define init_13_8_14_7
void FASTCALL call_13_8_14_7(SH2_struct* sh)
{
  SH2movli(sh, 8, (14<<4)+7);
}
#define init_13_8_14_8
void FASTCALL call_13_8_14_8(SH2_struct* sh)
{
  SH2movli(sh, 8, (14<<4)+8);
}
#define init_13_8_14_9
void FASTCALL call_13_8_14_9(SH2_struct* sh)
{
  SH2movli(sh, 8, (14<<4)+9);
}
#define init_13_8_14_10
void FASTCALL call_13_8_14_10(SH2_struct* sh)
{
  SH2movli(sh, 8, (14<<4)+10);
}
#define init_13_8_14_11
void FASTCALL call_13_8_14_11(SH2_struct* sh)
{
  SH2movli(sh, 8, (14<<4)+11);
}
#define init_13_8_14_12
void FASTCALL call_13_8_14_12(SH2_struct* sh)
{
  SH2movli(sh, 8, (14<<4)+12);
}
#define init_13_8_14_13
void FASTCALL call_13_8_14_13(SH2_struct* sh)
{
  SH2movli(sh, 8, (14<<4)+13);
}
#define init_13_8_14_14
void FASTCALL call_13_8_14_14(SH2_struct* sh)
{
  SH2movli(sh, 8, (14<<4)+14);
}
#define init_13_8_14_15
void FASTCALL call_13_8_14_15(SH2_struct* sh)
{
  SH2movli(sh, 8, (14<<4)+15);
}
#define init_13_8_15_0
void FASTCALL call_13_8_15_0(SH2_struct* sh)
{
  SH2movli(sh, 8, (15<<4)+0);
}
#define init_13_8_15_1
void FASTCALL call_13_8_15_1(SH2_struct* sh)
{
  SH2movli(sh, 8, (15<<4)+1);
}
#define init_13_8_15_2
void FASTCALL call_13_8_15_2(SH2_struct* sh)
{
  SH2movli(sh, 8, (15<<4)+2);
}
#define init_13_8_15_3
void FASTCALL call_13_8_15_3(SH2_struct* sh)
{
  SH2movli(sh, 8, (15<<4)+3);
}
#define init_13_8_15_4
void FASTCALL call_13_8_15_4(SH2_struct* sh)
{
  SH2movli(sh, 8, (15<<4)+4);
}
#define init_13_8_15_5
void FASTCALL call_13_8_15_5(SH2_struct* sh)
{
  SH2movli(sh, 8, (15<<4)+5);
}
#define init_13_8_15_6
void FASTCALL call_13_8_15_6(SH2_struct* sh)
{
  SH2movli(sh, 8, (15<<4)+6);
}
#define init_13_8_15_7
void FASTCALL call_13_8_15_7(SH2_struct* sh)
{
  SH2movli(sh, 8, (15<<4)+7);
}
#define init_13_8_15_8
void FASTCALL call_13_8_15_8(SH2_struct* sh)
{
  SH2movli(sh, 8, (15<<4)+8);
}
#define init_13_8_15_9
void FASTCALL call_13_8_15_9(SH2_struct* sh)
{
  SH2movli(sh, 8, (15<<4)+9);
}
#define init_13_8_15_10
void FASTCALL call_13_8_15_10(SH2_struct* sh)
{
  SH2movli(sh, 8, (15<<4)+10);
}
#define init_13_8_15_11
void FASTCALL call_13_8_15_11(SH2_struct* sh)
{
  SH2movli(sh, 8, (15<<4)+11);
}
#define init_13_8_15_12
void FASTCALL call_13_8_15_12(SH2_struct* sh)
{
  SH2movli(sh, 8, (15<<4)+12);
}
#define init_13_8_15_13
void FASTCALL call_13_8_15_13(SH2_struct* sh)
{
  SH2movli(sh, 8, (15<<4)+13);
}
#define init_13_8_15_14
void FASTCALL call_13_8_15_14(SH2_struct* sh)
{
  SH2movli(sh, 8, (15<<4)+14);
}
#define init_13_8_15_15
void FASTCALL call_13_8_15_15(SH2_struct* sh)
{
  SH2movli(sh, 8, (15<<4)+15);
}
#define init_13_9_0_0
void FASTCALL call_13_9_0_0(SH2_struct* sh)
{
  SH2movli(sh, 9, (0<<4)+0);
}
#define init_13_9_0_1
void FASTCALL call_13_9_0_1(SH2_struct* sh)
{
  SH2movli(sh, 9, (0<<4)+1);
}
#define init_13_9_0_2
void FASTCALL call_13_9_0_2(SH2_struct* sh)
{
  SH2movli(sh, 9, (0<<4)+2);
}
#define init_13_9_0_3
void FASTCALL call_13_9_0_3(SH2_struct* sh)
{
  SH2movli(sh, 9, (0<<4)+3);
}
#define init_13_9_0_4
void FASTCALL call_13_9_0_4(SH2_struct* sh)
{
  SH2movli(sh, 9, (0<<4)+4);
}
#define init_13_9_0_5
void FASTCALL call_13_9_0_5(SH2_struct* sh)
{
  SH2movli(sh, 9, (0<<4)+5);
}
#define init_13_9_0_6
void FASTCALL call_13_9_0_6(SH2_struct* sh)
{
  SH2movli(sh, 9, (0<<4)+6);
}
#define init_13_9_0_7
void FASTCALL call_13_9_0_7(SH2_struct* sh)
{
  SH2movli(sh, 9, (0<<4)+7);
}
#define init_13_9_0_8
void FASTCALL call_13_9_0_8(SH2_struct* sh)
{
  SH2movli(sh, 9, (0<<4)+8);
}
#define init_13_9_0_9
void FASTCALL call_13_9_0_9(SH2_struct* sh)
{
  SH2movli(sh, 9, (0<<4)+9);
}
#define init_13_9_0_10
void FASTCALL call_13_9_0_10(SH2_struct* sh)
{
  SH2movli(sh, 9, (0<<4)+10);
}
#define init_13_9_0_11
void FASTCALL call_13_9_0_11(SH2_struct* sh)
{
  SH2movli(sh, 9, (0<<4)+11);
}
#define init_13_9_0_12
void FASTCALL call_13_9_0_12(SH2_struct* sh)
{
  SH2movli(sh, 9, (0<<4)+12);
}
#define init_13_9_0_13
void FASTCALL call_13_9_0_13(SH2_struct* sh)
{
  SH2movli(sh, 9, (0<<4)+13);
}
#define init_13_9_0_14
void FASTCALL call_13_9_0_14(SH2_struct* sh)
{
  SH2movli(sh, 9, (0<<4)+14);
}
#define init_13_9_0_15
void FASTCALL call_13_9_0_15(SH2_struct* sh)
{
  SH2movli(sh, 9, (0<<4)+15);
}
#define init_13_9_1_0
void FASTCALL call_13_9_1_0(SH2_struct* sh)
{
  SH2movli(sh, 9, (1<<4)+0);
}
#define init_13_9_1_1
void FASTCALL call_13_9_1_1(SH2_struct* sh)
{
  SH2movli(sh, 9, (1<<4)+1);
}
#define init_13_9_1_2
void FASTCALL call_13_9_1_2(SH2_struct* sh)
{
  SH2movli(sh, 9, (1<<4)+2);
}
#define init_13_9_1_3
void FASTCALL call_13_9_1_3(SH2_struct* sh)
{
  SH2movli(sh, 9, (1<<4)+3);
}
#define init_13_9_1_4
void FASTCALL call_13_9_1_4(SH2_struct* sh)
{
  SH2movli(sh, 9, (1<<4)+4);
}
#define init_13_9_1_5
void FASTCALL call_13_9_1_5(SH2_struct* sh)
{
  SH2movli(sh, 9, (1<<4)+5);
}
#define init_13_9_1_6
void FASTCALL call_13_9_1_6(SH2_struct* sh)
{
  SH2movli(sh, 9, (1<<4)+6);
}
#define init_13_9_1_7
void FASTCALL call_13_9_1_7(SH2_struct* sh)
{
  SH2movli(sh, 9, (1<<4)+7);
}
#define init_13_9_1_8
void FASTCALL call_13_9_1_8(SH2_struct* sh)
{
  SH2movli(sh, 9, (1<<4)+8);
}
#define init_13_9_1_9
void FASTCALL call_13_9_1_9(SH2_struct* sh)
{
  SH2movli(sh, 9, (1<<4)+9);
}
#define init_13_9_1_10
void FASTCALL call_13_9_1_10(SH2_struct* sh)
{
  SH2movli(sh, 9, (1<<4)+10);
}
#define init_13_9_1_11
void FASTCALL call_13_9_1_11(SH2_struct* sh)
{
  SH2movli(sh, 9, (1<<4)+11);
}
#define init_13_9_1_12
void FASTCALL call_13_9_1_12(SH2_struct* sh)
{
  SH2movli(sh, 9, (1<<4)+12);
}
#define init_13_9_1_13
void FASTCALL call_13_9_1_13(SH2_struct* sh)
{
  SH2movli(sh, 9, (1<<4)+13);
}
#define init_13_9_1_14
void FASTCALL call_13_9_1_14(SH2_struct* sh)
{
  SH2movli(sh, 9, (1<<4)+14);
}
#define init_13_9_1_15
void FASTCALL call_13_9_1_15(SH2_struct* sh)
{
  SH2movli(sh, 9, (1<<4)+15);
}
#define init_13_9_2_0
void FASTCALL call_13_9_2_0(SH2_struct* sh)
{
  SH2movli(sh, 9, (2<<4)+0);
}
#define init_13_9_2_1
void FASTCALL call_13_9_2_1(SH2_struct* sh)
{
  SH2movli(sh, 9, (2<<4)+1);
}
#define init_13_9_2_2
void FASTCALL call_13_9_2_2(SH2_struct* sh)
{
  SH2movli(sh, 9, (2<<4)+2);
}
#define init_13_9_2_3
void FASTCALL call_13_9_2_3(SH2_struct* sh)
{
  SH2movli(sh, 9, (2<<4)+3);
}
#define init_13_9_2_4
void FASTCALL call_13_9_2_4(SH2_struct* sh)
{
  SH2movli(sh, 9, (2<<4)+4);
}
#define init_13_9_2_5
void FASTCALL call_13_9_2_5(SH2_struct* sh)
{
  SH2movli(sh, 9, (2<<4)+5);
}
#define init_13_9_2_6
void FASTCALL call_13_9_2_6(SH2_struct* sh)
{
  SH2movli(sh, 9, (2<<4)+6);
}
#define init_13_9_2_7
void FASTCALL call_13_9_2_7(SH2_struct* sh)
{
  SH2movli(sh, 9, (2<<4)+7);
}
#define init_13_9_2_8
void FASTCALL call_13_9_2_8(SH2_struct* sh)
{
  SH2movli(sh, 9, (2<<4)+8);
}
#define init_13_9_2_9
void FASTCALL call_13_9_2_9(SH2_struct* sh)
{
  SH2movli(sh, 9, (2<<4)+9);
}
#define init_13_9_2_10
void FASTCALL call_13_9_2_10(SH2_struct* sh)
{
  SH2movli(sh, 9, (2<<4)+10);
}
#define init_13_9_2_11
void FASTCALL call_13_9_2_11(SH2_struct* sh)
{
  SH2movli(sh, 9, (2<<4)+11);
}
#define init_13_9_2_12
void FASTCALL call_13_9_2_12(SH2_struct* sh)
{
  SH2movli(sh, 9, (2<<4)+12);
}
#define init_13_9_2_13
void FASTCALL call_13_9_2_13(SH2_struct* sh)
{
  SH2movli(sh, 9, (2<<4)+13);
}
#define init_13_9_2_14
void FASTCALL call_13_9_2_14(SH2_struct* sh)
{
  SH2movli(sh, 9, (2<<4)+14);
}
#define init_13_9_2_15
void FASTCALL call_13_9_2_15(SH2_struct* sh)
{
  SH2movli(sh, 9, (2<<4)+15);
}
#define init_13_9_3_0
void FASTCALL call_13_9_3_0(SH2_struct* sh)
{
  SH2movli(sh, 9, (3<<4)+0);
}
#define init_13_9_3_1
void FASTCALL call_13_9_3_1(SH2_struct* sh)
{
  SH2movli(sh, 9, (3<<4)+1);
}
#define init_13_9_3_2
void FASTCALL call_13_9_3_2(SH2_struct* sh)
{
  SH2movli(sh, 9, (3<<4)+2);
}
#define init_13_9_3_3
void FASTCALL call_13_9_3_3(SH2_struct* sh)
{
  SH2movli(sh, 9, (3<<4)+3);
}
#define init_13_9_3_4
void FASTCALL call_13_9_3_4(SH2_struct* sh)
{
  SH2movli(sh, 9, (3<<4)+4);
}
#define init_13_9_3_5
void FASTCALL call_13_9_3_5(SH2_struct* sh)
{
  SH2movli(sh, 9, (3<<4)+5);
}
#define init_13_9_3_6
void FASTCALL call_13_9_3_6(SH2_struct* sh)
{
  SH2movli(sh, 9, (3<<4)+6);
}
#define init_13_9_3_7
void FASTCALL call_13_9_3_7(SH2_struct* sh)
{
  SH2movli(sh, 9, (3<<4)+7);
}
#define init_13_9_3_8
void FASTCALL call_13_9_3_8(SH2_struct* sh)
{
  SH2movli(sh, 9, (3<<4)+8);
}
#define init_13_9_3_9
void FASTCALL call_13_9_3_9(SH2_struct* sh)
{
  SH2movli(sh, 9, (3<<4)+9);
}
#define init_13_9_3_10
void FASTCALL call_13_9_3_10(SH2_struct* sh)
{
  SH2movli(sh, 9, (3<<4)+10);
}
#define init_13_9_3_11
void FASTCALL call_13_9_3_11(SH2_struct* sh)
{
  SH2movli(sh, 9, (3<<4)+11);
}
#define init_13_9_3_12
void FASTCALL call_13_9_3_12(SH2_struct* sh)
{
  SH2movli(sh, 9, (3<<4)+12);
}
#define init_13_9_3_13
void FASTCALL call_13_9_3_13(SH2_struct* sh)
{
  SH2movli(sh, 9, (3<<4)+13);
}
#define init_13_9_3_14
void FASTCALL call_13_9_3_14(SH2_struct* sh)
{
  SH2movli(sh, 9, (3<<4)+14);
}
#define init_13_9_3_15
void FASTCALL call_13_9_3_15(SH2_struct* sh)
{
  SH2movli(sh, 9, (3<<4)+15);
}
#define init_13_9_4_0
void FASTCALL call_13_9_4_0(SH2_struct* sh)
{
  SH2movli(sh, 9, (4<<4)+0);
}
#define init_13_9_4_1
void FASTCALL call_13_9_4_1(SH2_struct* sh)
{
  SH2movli(sh, 9, (4<<4)+1);
}
#define init_13_9_4_2
void FASTCALL call_13_9_4_2(SH2_struct* sh)
{
  SH2movli(sh, 9, (4<<4)+2);
}
#define init_13_9_4_3
void FASTCALL call_13_9_4_3(SH2_struct* sh)
{
  SH2movli(sh, 9, (4<<4)+3);
}
#define init_13_9_4_4
void FASTCALL call_13_9_4_4(SH2_struct* sh)
{
  SH2movli(sh, 9, (4<<4)+4);
}
#define init_13_9_4_5
void FASTCALL call_13_9_4_5(SH2_struct* sh)
{
  SH2movli(sh, 9, (4<<4)+5);
}
#define init_13_9_4_6
void FASTCALL call_13_9_4_6(SH2_struct* sh)
{
  SH2movli(sh, 9, (4<<4)+6);
}
#define init_13_9_4_7
void FASTCALL call_13_9_4_7(SH2_struct* sh)
{
  SH2movli(sh, 9, (4<<4)+7);
}
#define init_13_9_4_8
void FASTCALL call_13_9_4_8(SH2_struct* sh)
{
  SH2movli(sh, 9, (4<<4)+8);
}
#define init_13_9_4_9
void FASTCALL call_13_9_4_9(SH2_struct* sh)
{
  SH2movli(sh, 9, (4<<4)+9);
}
#define init_13_9_4_10
void FASTCALL call_13_9_4_10(SH2_struct* sh)
{
  SH2movli(sh, 9, (4<<4)+10);
}
#define init_13_9_4_11
void FASTCALL call_13_9_4_11(SH2_struct* sh)
{
  SH2movli(sh, 9, (4<<4)+11);
}
#define init_13_9_4_12
void FASTCALL call_13_9_4_12(SH2_struct* sh)
{
  SH2movli(sh, 9, (4<<4)+12);
}
#define init_13_9_4_13
void FASTCALL call_13_9_4_13(SH2_struct* sh)
{
  SH2movli(sh, 9, (4<<4)+13);
}
#define init_13_9_4_14
void FASTCALL call_13_9_4_14(SH2_struct* sh)
{
  SH2movli(sh, 9, (4<<4)+14);
}
#define init_13_9_4_15
void FASTCALL call_13_9_4_15(SH2_struct* sh)
{
  SH2movli(sh, 9, (4<<4)+15);
}
#define init_13_9_5_0
void FASTCALL call_13_9_5_0(SH2_struct* sh)
{
  SH2movli(sh, 9, (5<<4)+0);
}
#define init_13_9_5_1
void FASTCALL call_13_9_5_1(SH2_struct* sh)
{
  SH2movli(sh, 9, (5<<4)+1);
}
#define init_13_9_5_2
void FASTCALL call_13_9_5_2(SH2_struct* sh)
{
  SH2movli(sh, 9, (5<<4)+2);
}
#define init_13_9_5_3
void FASTCALL call_13_9_5_3(SH2_struct* sh)
{
  SH2movli(sh, 9, (5<<4)+3);
}
#define init_13_9_5_4
void FASTCALL call_13_9_5_4(SH2_struct* sh)
{
  SH2movli(sh, 9, (5<<4)+4);
}
#define init_13_9_5_5
void FASTCALL call_13_9_5_5(SH2_struct* sh)
{
  SH2movli(sh, 9, (5<<4)+5);
}
#define init_13_9_5_6
void FASTCALL call_13_9_5_6(SH2_struct* sh)
{
  SH2movli(sh, 9, (5<<4)+6);
}
#define init_13_9_5_7
void FASTCALL call_13_9_5_7(SH2_struct* sh)
{
  SH2movli(sh, 9, (5<<4)+7);
}
#define init_13_9_5_8
void FASTCALL call_13_9_5_8(SH2_struct* sh)
{
  SH2movli(sh, 9, (5<<4)+8);
}
#define init_13_9_5_9
void FASTCALL call_13_9_5_9(SH2_struct* sh)
{
  SH2movli(sh, 9, (5<<4)+9);
}
#define init_13_9_5_10
void FASTCALL call_13_9_5_10(SH2_struct* sh)
{
  SH2movli(sh, 9, (5<<4)+10);
}
#define init_13_9_5_11
void FASTCALL call_13_9_5_11(SH2_struct* sh)
{
  SH2movli(sh, 9, (5<<4)+11);
}
#define init_13_9_5_12
void FASTCALL call_13_9_5_12(SH2_struct* sh)
{
  SH2movli(sh, 9, (5<<4)+12);
}
#define init_13_9_5_13
void FASTCALL call_13_9_5_13(SH2_struct* sh)
{
  SH2movli(sh, 9, (5<<4)+13);
}
#define init_13_9_5_14
void FASTCALL call_13_9_5_14(SH2_struct* sh)
{
  SH2movli(sh, 9, (5<<4)+14);
}
#define init_13_9_5_15
void FASTCALL call_13_9_5_15(SH2_struct* sh)
{
  SH2movli(sh, 9, (5<<4)+15);
}
#define init_13_9_6_0
void FASTCALL call_13_9_6_0(SH2_struct* sh)
{
  SH2movli(sh, 9, (6<<4)+0);
}
#define init_13_9_6_1
void FASTCALL call_13_9_6_1(SH2_struct* sh)
{
  SH2movli(sh, 9, (6<<4)+1);
}
#define init_13_9_6_2
void FASTCALL call_13_9_6_2(SH2_struct* sh)
{
  SH2movli(sh, 9, (6<<4)+2);
}
#define init_13_9_6_3
void FASTCALL call_13_9_6_3(SH2_struct* sh)
{
  SH2movli(sh, 9, (6<<4)+3);
}
#define init_13_9_6_4
void FASTCALL call_13_9_6_4(SH2_struct* sh)
{
  SH2movli(sh, 9, (6<<4)+4);
}
#define init_13_9_6_5
void FASTCALL call_13_9_6_5(SH2_struct* sh)
{
  SH2movli(sh, 9, (6<<4)+5);
}
#define init_13_9_6_6
void FASTCALL call_13_9_6_6(SH2_struct* sh)
{
  SH2movli(sh, 9, (6<<4)+6);
}
#define init_13_9_6_7
void FASTCALL call_13_9_6_7(SH2_struct* sh)
{
  SH2movli(sh, 9, (6<<4)+7);
}
#define init_13_9_6_8
void FASTCALL call_13_9_6_8(SH2_struct* sh)
{
  SH2movli(sh, 9, (6<<4)+8);
}
#define init_13_9_6_9
void FASTCALL call_13_9_6_9(SH2_struct* sh)
{
  SH2movli(sh, 9, (6<<4)+9);
}
#define init_13_9_6_10
void FASTCALL call_13_9_6_10(SH2_struct* sh)
{
  SH2movli(sh, 9, (6<<4)+10);
}
#define init_13_9_6_11
void FASTCALL call_13_9_6_11(SH2_struct* sh)
{
  SH2movli(sh, 9, (6<<4)+11);
}
#define init_13_9_6_12
void FASTCALL call_13_9_6_12(SH2_struct* sh)
{
  SH2movli(sh, 9, (6<<4)+12);
}
#define init_13_9_6_13
void FASTCALL call_13_9_6_13(SH2_struct* sh)
{
  SH2movli(sh, 9, (6<<4)+13);
}
#define init_13_9_6_14
void FASTCALL call_13_9_6_14(SH2_struct* sh)
{
  SH2movli(sh, 9, (6<<4)+14);
}
#define init_13_9_6_15
void FASTCALL call_13_9_6_15(SH2_struct* sh)
{
  SH2movli(sh, 9, (6<<4)+15);
}
#define init_13_9_7_0
void FASTCALL call_13_9_7_0(SH2_struct* sh)
{
  SH2movli(sh, 9, (7<<4)+0);
}
#define init_13_9_7_1
void FASTCALL call_13_9_7_1(SH2_struct* sh)
{
  SH2movli(sh, 9, (7<<4)+1);
}
#define init_13_9_7_2
void FASTCALL call_13_9_7_2(SH2_struct* sh)
{
  SH2movli(sh, 9, (7<<4)+2);
}
#define init_13_9_7_3
void FASTCALL call_13_9_7_3(SH2_struct* sh)
{
  SH2movli(sh, 9, (7<<4)+3);
}
#define init_13_9_7_4
void FASTCALL call_13_9_7_4(SH2_struct* sh)
{
  SH2movli(sh, 9, (7<<4)+4);
}
#define init_13_9_7_5
void FASTCALL call_13_9_7_5(SH2_struct* sh)
{
  SH2movli(sh, 9, (7<<4)+5);
}
#define init_13_9_7_6
void FASTCALL call_13_9_7_6(SH2_struct* sh)
{
  SH2movli(sh, 9, (7<<4)+6);
}
#define init_13_9_7_7
void FASTCALL call_13_9_7_7(SH2_struct* sh)
{
  SH2movli(sh, 9, (7<<4)+7);
}
#define init_13_9_7_8
void FASTCALL call_13_9_7_8(SH2_struct* sh)
{
  SH2movli(sh, 9, (7<<4)+8);
}
#define init_13_9_7_9
void FASTCALL call_13_9_7_9(SH2_struct* sh)
{
  SH2movli(sh, 9, (7<<4)+9);
}
#define init_13_9_7_10
void FASTCALL call_13_9_7_10(SH2_struct* sh)
{
  SH2movli(sh, 9, (7<<4)+10);
}
#define init_13_9_7_11
void FASTCALL call_13_9_7_11(SH2_struct* sh)
{
  SH2movli(sh, 9, (7<<4)+11);
}
#define init_13_9_7_12
void FASTCALL call_13_9_7_12(SH2_struct* sh)
{
  SH2movli(sh, 9, (7<<4)+12);
}
#define init_13_9_7_13
void FASTCALL call_13_9_7_13(SH2_struct* sh)
{
  SH2movli(sh, 9, (7<<4)+13);
}
#define init_13_9_7_14
void FASTCALL call_13_9_7_14(SH2_struct* sh)
{
  SH2movli(sh, 9, (7<<4)+14);
}
#define init_13_9_7_15
void FASTCALL call_13_9_7_15(SH2_struct* sh)
{
  SH2movli(sh, 9, (7<<4)+15);
}
#define init_13_9_8_0
void FASTCALL call_13_9_8_0(SH2_struct* sh)
{
  SH2movli(sh, 9, (8<<4)+0);
}
#define init_13_9_8_1
void FASTCALL call_13_9_8_1(SH2_struct* sh)
{
  SH2movli(sh, 9, (8<<4)+1);
}
#define init_13_9_8_2
void FASTCALL call_13_9_8_2(SH2_struct* sh)
{
  SH2movli(sh, 9, (8<<4)+2);
}
#define init_13_9_8_3
void FASTCALL call_13_9_8_3(SH2_struct* sh)
{
  SH2movli(sh, 9, (8<<4)+3);
}
#define init_13_9_8_4
void FASTCALL call_13_9_8_4(SH2_struct* sh)
{
  SH2movli(sh, 9, (8<<4)+4);
}
#define init_13_9_8_5
void FASTCALL call_13_9_8_5(SH2_struct* sh)
{
  SH2movli(sh, 9, (8<<4)+5);
}
#define init_13_9_8_6
void FASTCALL call_13_9_8_6(SH2_struct* sh)
{
  SH2movli(sh, 9, (8<<4)+6);
}
#define init_13_9_8_7
void FASTCALL call_13_9_8_7(SH2_struct* sh)
{
  SH2movli(sh, 9, (8<<4)+7);
}
#define init_13_9_8_8
void FASTCALL call_13_9_8_8(SH2_struct* sh)
{
  SH2movli(sh, 9, (8<<4)+8);
}
#define init_13_9_8_9
void FASTCALL call_13_9_8_9(SH2_struct* sh)
{
  SH2movli(sh, 9, (8<<4)+9);
}
#define init_13_9_8_10
void FASTCALL call_13_9_8_10(SH2_struct* sh)
{
  SH2movli(sh, 9, (8<<4)+10);
}
#define init_13_9_8_11
void FASTCALL call_13_9_8_11(SH2_struct* sh)
{
  SH2movli(sh, 9, (8<<4)+11);
}
#define init_13_9_8_12
void FASTCALL call_13_9_8_12(SH2_struct* sh)
{
  SH2movli(sh, 9, (8<<4)+12);
}
#define init_13_9_8_13
void FASTCALL call_13_9_8_13(SH2_struct* sh)
{
  SH2movli(sh, 9, (8<<4)+13);
}
#define init_13_9_8_14
void FASTCALL call_13_9_8_14(SH2_struct* sh)
{
  SH2movli(sh, 9, (8<<4)+14);
}
#define init_13_9_8_15
void FASTCALL call_13_9_8_15(SH2_struct* sh)
{
  SH2movli(sh, 9, (8<<4)+15);
}
#define init_13_9_9_0
void FASTCALL call_13_9_9_0(SH2_struct* sh)
{
  SH2movli(sh, 9, (9<<4)+0);
}
#define init_13_9_9_1
void FASTCALL call_13_9_9_1(SH2_struct* sh)
{
  SH2movli(sh, 9, (9<<4)+1);
}
#define init_13_9_9_2
void FASTCALL call_13_9_9_2(SH2_struct* sh)
{
  SH2movli(sh, 9, (9<<4)+2);
}
#define init_13_9_9_3
void FASTCALL call_13_9_9_3(SH2_struct* sh)
{
  SH2movli(sh, 9, (9<<4)+3);
}
#define init_13_9_9_4
void FASTCALL call_13_9_9_4(SH2_struct* sh)
{
  SH2movli(sh, 9, (9<<4)+4);
}
#define init_13_9_9_5
void FASTCALL call_13_9_9_5(SH2_struct* sh)
{
  SH2movli(sh, 9, (9<<4)+5);
}
#define init_13_9_9_6
void FASTCALL call_13_9_9_6(SH2_struct* sh)
{
  SH2movli(sh, 9, (9<<4)+6);
}
#define init_13_9_9_7
void FASTCALL call_13_9_9_7(SH2_struct* sh)
{
  SH2movli(sh, 9, (9<<4)+7);
}
#define init_13_9_9_8
void FASTCALL call_13_9_9_8(SH2_struct* sh)
{
  SH2movli(sh, 9, (9<<4)+8);
}
#define init_13_9_9_9
void FASTCALL call_13_9_9_9(SH2_struct* sh)
{
  SH2movli(sh, 9, (9<<4)+9);
}
#define init_13_9_9_10
void FASTCALL call_13_9_9_10(SH2_struct* sh)
{
  SH2movli(sh, 9, (9<<4)+10);
}
#define init_13_9_9_11
void FASTCALL call_13_9_9_11(SH2_struct* sh)
{
  SH2movli(sh, 9, (9<<4)+11);
}
#define init_13_9_9_12
void FASTCALL call_13_9_9_12(SH2_struct* sh)
{
  SH2movli(sh, 9, (9<<4)+12);
}
#define init_13_9_9_13
void FASTCALL call_13_9_9_13(SH2_struct* sh)
{
  SH2movli(sh, 9, (9<<4)+13);
}
#define init_13_9_9_14
void FASTCALL call_13_9_9_14(SH2_struct* sh)
{
  SH2movli(sh, 9, (9<<4)+14);
}
#define init_13_9_9_15
void FASTCALL call_13_9_9_15(SH2_struct* sh)
{
  SH2movli(sh, 9, (9<<4)+15);
}
#define init_13_9_10_0
void FASTCALL call_13_9_10_0(SH2_struct* sh)
{
  SH2movli(sh, 9, (10<<4)+0);
}
#define init_13_9_10_1
void FASTCALL call_13_9_10_1(SH2_struct* sh)
{
  SH2movli(sh, 9, (10<<4)+1);
}
#define init_13_9_10_2
void FASTCALL call_13_9_10_2(SH2_struct* sh)
{
  SH2movli(sh, 9, (10<<4)+2);
}
#define init_13_9_10_3
void FASTCALL call_13_9_10_3(SH2_struct* sh)
{
  SH2movli(sh, 9, (10<<4)+3);
}
#define init_13_9_10_4
void FASTCALL call_13_9_10_4(SH2_struct* sh)
{
  SH2movli(sh, 9, (10<<4)+4);
}
#define init_13_9_10_5
void FASTCALL call_13_9_10_5(SH2_struct* sh)
{
  SH2movli(sh, 9, (10<<4)+5);
}
#define init_13_9_10_6
void FASTCALL call_13_9_10_6(SH2_struct* sh)
{
  SH2movli(sh, 9, (10<<4)+6);
}
#define init_13_9_10_7
void FASTCALL call_13_9_10_7(SH2_struct* sh)
{
  SH2movli(sh, 9, (10<<4)+7);
}
#define init_13_9_10_8
void FASTCALL call_13_9_10_8(SH2_struct* sh)
{
  SH2movli(sh, 9, (10<<4)+8);
}
#define init_13_9_10_9
void FASTCALL call_13_9_10_9(SH2_struct* sh)
{
  SH2movli(sh, 9, (10<<4)+9);
}
#define init_13_9_10_10
void FASTCALL call_13_9_10_10(SH2_struct* sh)
{
  SH2movli(sh, 9, (10<<4)+10);
}
#define init_13_9_10_11
void FASTCALL call_13_9_10_11(SH2_struct* sh)
{
  SH2movli(sh, 9, (10<<4)+11);
}
#define init_13_9_10_12
void FASTCALL call_13_9_10_12(SH2_struct* sh)
{
  SH2movli(sh, 9, (10<<4)+12);
}
#define init_13_9_10_13
void FASTCALL call_13_9_10_13(SH2_struct* sh)
{
  SH2movli(sh, 9, (10<<4)+13);
}
#define init_13_9_10_14
void FASTCALL call_13_9_10_14(SH2_struct* sh)
{
  SH2movli(sh, 9, (10<<4)+14);
}
#define init_13_9_10_15
void FASTCALL call_13_9_10_15(SH2_struct* sh)
{
  SH2movli(sh, 9, (10<<4)+15);
}
#define init_13_9_11_0
void FASTCALL call_13_9_11_0(SH2_struct* sh)
{
  SH2movli(sh, 9, (11<<4)+0);
}
#define init_13_9_11_1
void FASTCALL call_13_9_11_1(SH2_struct* sh)
{
  SH2movli(sh, 9, (11<<4)+1);
}
#define init_13_9_11_2
void FASTCALL call_13_9_11_2(SH2_struct* sh)
{
  SH2movli(sh, 9, (11<<4)+2);
}
#define init_13_9_11_3
void FASTCALL call_13_9_11_3(SH2_struct* sh)
{
  SH2movli(sh, 9, (11<<4)+3);
}
#define init_13_9_11_4
void FASTCALL call_13_9_11_4(SH2_struct* sh)
{
  SH2movli(sh, 9, (11<<4)+4);
}
#define init_13_9_11_5
void FASTCALL call_13_9_11_5(SH2_struct* sh)
{
  SH2movli(sh, 9, (11<<4)+5);
}
#define init_13_9_11_6
void FASTCALL call_13_9_11_6(SH2_struct* sh)
{
  SH2movli(sh, 9, (11<<4)+6);
}
#define init_13_9_11_7
void FASTCALL call_13_9_11_7(SH2_struct* sh)
{
  SH2movli(sh, 9, (11<<4)+7);
}
#define init_13_9_11_8
void FASTCALL call_13_9_11_8(SH2_struct* sh)
{
  SH2movli(sh, 9, (11<<4)+8);
}
#define init_13_9_11_9
void FASTCALL call_13_9_11_9(SH2_struct* sh)
{
  SH2movli(sh, 9, (11<<4)+9);
}
#define init_13_9_11_10
void FASTCALL call_13_9_11_10(SH2_struct* sh)
{
  SH2movli(sh, 9, (11<<4)+10);
}
#define init_13_9_11_11
void FASTCALL call_13_9_11_11(SH2_struct* sh)
{
  SH2movli(sh, 9, (11<<4)+11);
}
#define init_13_9_11_12
void FASTCALL call_13_9_11_12(SH2_struct* sh)
{
  SH2movli(sh, 9, (11<<4)+12);
}
#define init_13_9_11_13
void FASTCALL call_13_9_11_13(SH2_struct* sh)
{
  SH2movli(sh, 9, (11<<4)+13);
}
#define init_13_9_11_14
void FASTCALL call_13_9_11_14(SH2_struct* sh)
{
  SH2movli(sh, 9, (11<<4)+14);
}
#define init_13_9_11_15
void FASTCALL call_13_9_11_15(SH2_struct* sh)
{
  SH2movli(sh, 9, (11<<4)+15);
}
#define init_13_9_12_0
void FASTCALL call_13_9_12_0(SH2_struct* sh)
{
  SH2movli(sh, 9, (12<<4)+0);
}
#define init_13_9_12_1
void FASTCALL call_13_9_12_1(SH2_struct* sh)
{
  SH2movli(sh, 9, (12<<4)+1);
}
#define init_13_9_12_2
void FASTCALL call_13_9_12_2(SH2_struct* sh)
{
  SH2movli(sh, 9, (12<<4)+2);
}
#define init_13_9_12_3
void FASTCALL call_13_9_12_3(SH2_struct* sh)
{
  SH2movli(sh, 9, (12<<4)+3);
}
#define init_13_9_12_4
void FASTCALL call_13_9_12_4(SH2_struct* sh)
{
  SH2movli(sh, 9, (12<<4)+4);
}
#define init_13_9_12_5
void FASTCALL call_13_9_12_5(SH2_struct* sh)
{
  SH2movli(sh, 9, (12<<4)+5);
}
#define init_13_9_12_6
void FASTCALL call_13_9_12_6(SH2_struct* sh)
{
  SH2movli(sh, 9, (12<<4)+6);
}
#define init_13_9_12_7
void FASTCALL call_13_9_12_7(SH2_struct* sh)
{
  SH2movli(sh, 9, (12<<4)+7);
}
#define init_13_9_12_8
void FASTCALL call_13_9_12_8(SH2_struct* sh)
{
  SH2movli(sh, 9, (12<<4)+8);
}
#define init_13_9_12_9
void FASTCALL call_13_9_12_9(SH2_struct* sh)
{
  SH2movli(sh, 9, (12<<4)+9);
}
#define init_13_9_12_10
void FASTCALL call_13_9_12_10(SH2_struct* sh)
{
  SH2movli(sh, 9, (12<<4)+10);
}
#define init_13_9_12_11
void FASTCALL call_13_9_12_11(SH2_struct* sh)
{
  SH2movli(sh, 9, (12<<4)+11);
}
#define init_13_9_12_12
void FASTCALL call_13_9_12_12(SH2_struct* sh)
{
  SH2movli(sh, 9, (12<<4)+12);
}
#define init_13_9_12_13
void FASTCALL call_13_9_12_13(SH2_struct* sh)
{
  SH2movli(sh, 9, (12<<4)+13);
}
#define init_13_9_12_14
void FASTCALL call_13_9_12_14(SH2_struct* sh)
{
  SH2movli(sh, 9, (12<<4)+14);
}
#define init_13_9_12_15
void FASTCALL call_13_9_12_15(SH2_struct* sh)
{
  SH2movli(sh, 9, (12<<4)+15);
}
#define init_13_9_13_0
void FASTCALL call_13_9_13_0(SH2_struct* sh)
{
  SH2movli(sh, 9, (13<<4)+0);
}
#define init_13_9_13_1
void FASTCALL call_13_9_13_1(SH2_struct* sh)
{
  SH2movli(sh, 9, (13<<4)+1);
}
#define init_13_9_13_2
void FASTCALL call_13_9_13_2(SH2_struct* sh)
{
  SH2movli(sh, 9, (13<<4)+2);
}
#define init_13_9_13_3
void FASTCALL call_13_9_13_3(SH2_struct* sh)
{
  SH2movli(sh, 9, (13<<4)+3);
}
#define init_13_9_13_4
void FASTCALL call_13_9_13_4(SH2_struct* sh)
{
  SH2movli(sh, 9, (13<<4)+4);
}
#define init_13_9_13_5
void FASTCALL call_13_9_13_5(SH2_struct* sh)
{
  SH2movli(sh, 9, (13<<4)+5);
}
#define init_13_9_13_6
void FASTCALL call_13_9_13_6(SH2_struct* sh)
{
  SH2movli(sh, 9, (13<<4)+6);
}
#define init_13_9_13_7
void FASTCALL call_13_9_13_7(SH2_struct* sh)
{
  SH2movli(sh, 9, (13<<4)+7);
}
#define init_13_9_13_8
void FASTCALL call_13_9_13_8(SH2_struct* sh)
{
  SH2movli(sh, 9, (13<<4)+8);
}
#define init_13_9_13_9
void FASTCALL call_13_9_13_9(SH2_struct* sh)
{
  SH2movli(sh, 9, (13<<4)+9);
}
#define init_13_9_13_10
void FASTCALL call_13_9_13_10(SH2_struct* sh)
{
  SH2movli(sh, 9, (13<<4)+10);
}
#define init_13_9_13_11
void FASTCALL call_13_9_13_11(SH2_struct* sh)
{
  SH2movli(sh, 9, (13<<4)+11);
}
#define init_13_9_13_12
void FASTCALL call_13_9_13_12(SH2_struct* sh)
{
  SH2movli(sh, 9, (13<<4)+12);
}
#define init_13_9_13_13
void FASTCALL call_13_9_13_13(SH2_struct* sh)
{
  SH2movli(sh, 9, (13<<4)+13);
}
#define init_13_9_13_14
void FASTCALL call_13_9_13_14(SH2_struct* sh)
{
  SH2movli(sh, 9, (13<<4)+14);
}
#define init_13_9_13_15
void FASTCALL call_13_9_13_15(SH2_struct* sh)
{
  SH2movli(sh, 9, (13<<4)+15);
}
#define init_13_9_14_0
void FASTCALL call_13_9_14_0(SH2_struct* sh)
{
  SH2movli(sh, 9, (14<<4)+0);
}
#define init_13_9_14_1
void FASTCALL call_13_9_14_1(SH2_struct* sh)
{
  SH2movli(sh, 9, (14<<4)+1);
}
#define init_13_9_14_2
void FASTCALL call_13_9_14_2(SH2_struct* sh)
{
  SH2movli(sh, 9, (14<<4)+2);
}
#define init_13_9_14_3
void FASTCALL call_13_9_14_3(SH2_struct* sh)
{
  SH2movli(sh, 9, (14<<4)+3);
}
#define init_13_9_14_4
void FASTCALL call_13_9_14_4(SH2_struct* sh)
{
  SH2movli(sh, 9, (14<<4)+4);
}
#define init_13_9_14_5
void FASTCALL call_13_9_14_5(SH2_struct* sh)
{
  SH2movli(sh, 9, (14<<4)+5);
}
#define init_13_9_14_6
void FASTCALL call_13_9_14_6(SH2_struct* sh)
{
  SH2movli(sh, 9, (14<<4)+6);
}
#define init_13_9_14_7
void FASTCALL call_13_9_14_7(SH2_struct* sh)
{
  SH2movli(sh, 9, (14<<4)+7);
}
#define init_13_9_14_8
void FASTCALL call_13_9_14_8(SH2_struct* sh)
{
  SH2movli(sh, 9, (14<<4)+8);
}
#define init_13_9_14_9
void FASTCALL call_13_9_14_9(SH2_struct* sh)
{
  SH2movli(sh, 9, (14<<4)+9);
}
#define init_13_9_14_10
void FASTCALL call_13_9_14_10(SH2_struct* sh)
{
  SH2movli(sh, 9, (14<<4)+10);
}
#define init_13_9_14_11
void FASTCALL call_13_9_14_11(SH2_struct* sh)
{
  SH2movli(sh, 9, (14<<4)+11);
}
#define init_13_9_14_12
void FASTCALL call_13_9_14_12(SH2_struct* sh)
{
  SH2movli(sh, 9, (14<<4)+12);
}
#define init_13_9_14_13
void FASTCALL call_13_9_14_13(SH2_struct* sh)
{
  SH2movli(sh, 9, (14<<4)+13);
}
#define init_13_9_14_14
void FASTCALL call_13_9_14_14(SH2_struct* sh)
{
  SH2movli(sh, 9, (14<<4)+14);
}
#define init_13_9_14_15
void FASTCALL call_13_9_14_15(SH2_struct* sh)
{
  SH2movli(sh, 9, (14<<4)+15);
}
#define init_13_9_15_0
void FASTCALL call_13_9_15_0(SH2_struct* sh)
{
  SH2movli(sh, 9, (15<<4)+0);
}
#define init_13_9_15_1
void FASTCALL call_13_9_15_1(SH2_struct* sh)
{
  SH2movli(sh, 9, (15<<4)+1);
}
#define init_13_9_15_2
void FASTCALL call_13_9_15_2(SH2_struct* sh)
{
  SH2movli(sh, 9, (15<<4)+2);
}
#define init_13_9_15_3
void FASTCALL call_13_9_15_3(SH2_struct* sh)
{
  SH2movli(sh, 9, (15<<4)+3);
}
#define init_13_9_15_4
void FASTCALL call_13_9_15_4(SH2_struct* sh)
{
  SH2movli(sh, 9, (15<<4)+4);
}
#define init_13_9_15_5
void FASTCALL call_13_9_15_5(SH2_struct* sh)
{
  SH2movli(sh, 9, (15<<4)+5);
}
#define init_13_9_15_6
void FASTCALL call_13_9_15_6(SH2_struct* sh)
{
  SH2movli(sh, 9, (15<<4)+6);
}
#define init_13_9_15_7
void FASTCALL call_13_9_15_7(SH2_struct* sh)
{
  SH2movli(sh, 9, (15<<4)+7);
}
#define init_13_9_15_8
void FASTCALL call_13_9_15_8(SH2_struct* sh)
{
  SH2movli(sh, 9, (15<<4)+8);
}
#define init_13_9_15_9
void FASTCALL call_13_9_15_9(SH2_struct* sh)
{
  SH2movli(sh, 9, (15<<4)+9);
}
#define init_13_9_15_10
void FASTCALL call_13_9_15_10(SH2_struct* sh)
{
  SH2movli(sh, 9, (15<<4)+10);
}
#define init_13_9_15_11
void FASTCALL call_13_9_15_11(SH2_struct* sh)
{
  SH2movli(sh, 9, (15<<4)+11);
}
#define init_13_9_15_12
void FASTCALL call_13_9_15_12(SH2_struct* sh)
{
  SH2movli(sh, 9, (15<<4)+12);
}
#define init_13_9_15_13
void FASTCALL call_13_9_15_13(SH2_struct* sh)
{
  SH2movli(sh, 9, (15<<4)+13);
}
#define init_13_9_15_14
void FASTCALL call_13_9_15_14(SH2_struct* sh)
{
  SH2movli(sh, 9, (15<<4)+14);
}
#define init_13_9_15_15
void FASTCALL call_13_9_15_15(SH2_struct* sh)
{
  SH2movli(sh, 9, (15<<4)+15);
}
#define init_13_10_0_0
void FASTCALL call_13_10_0_0(SH2_struct* sh)
{
  SH2movli(sh, 10, (0<<4)+0);
}
#define init_13_10_0_1
void FASTCALL call_13_10_0_1(SH2_struct* sh)
{
  SH2movli(sh, 10, (0<<4)+1);
}
#define init_13_10_0_2
void FASTCALL call_13_10_0_2(SH2_struct* sh)
{
  SH2movli(sh, 10, (0<<4)+2);
}
#define init_13_10_0_3
void FASTCALL call_13_10_0_3(SH2_struct* sh)
{
  SH2movli(sh, 10, (0<<4)+3);
}
#define init_13_10_0_4
void FASTCALL call_13_10_0_4(SH2_struct* sh)
{
  SH2movli(sh, 10, (0<<4)+4);
}
#define init_13_10_0_5
void FASTCALL call_13_10_0_5(SH2_struct* sh)
{
  SH2movli(sh, 10, (0<<4)+5);
}
#define init_13_10_0_6
void FASTCALL call_13_10_0_6(SH2_struct* sh)
{
  SH2movli(sh, 10, (0<<4)+6);
}
#define init_13_10_0_7
void FASTCALL call_13_10_0_7(SH2_struct* sh)
{
  SH2movli(sh, 10, (0<<4)+7);
}
#define init_13_10_0_8
void FASTCALL call_13_10_0_8(SH2_struct* sh)
{
  SH2movli(sh, 10, (0<<4)+8);
}
#define init_13_10_0_9
void FASTCALL call_13_10_0_9(SH2_struct* sh)
{
  SH2movli(sh, 10, (0<<4)+9);
}
#define init_13_10_0_10
void FASTCALL call_13_10_0_10(SH2_struct* sh)
{
  SH2movli(sh, 10, (0<<4)+10);
}
#define init_13_10_0_11
void FASTCALL call_13_10_0_11(SH2_struct* sh)
{
  SH2movli(sh, 10, (0<<4)+11);
}
#define init_13_10_0_12
void FASTCALL call_13_10_0_12(SH2_struct* sh)
{
  SH2movli(sh, 10, (0<<4)+12);
}
#define init_13_10_0_13
void FASTCALL call_13_10_0_13(SH2_struct* sh)
{
  SH2movli(sh, 10, (0<<4)+13);
}
#define init_13_10_0_14
void FASTCALL call_13_10_0_14(SH2_struct* sh)
{
  SH2movli(sh, 10, (0<<4)+14);
}
#define init_13_10_0_15
void FASTCALL call_13_10_0_15(SH2_struct* sh)
{
  SH2movli(sh, 10, (0<<4)+15);
}
#define init_13_10_1_0
void FASTCALL call_13_10_1_0(SH2_struct* sh)
{
  SH2movli(sh, 10, (1<<4)+0);
}
#define init_13_10_1_1
void FASTCALL call_13_10_1_1(SH2_struct* sh)
{
  SH2movli(sh, 10, (1<<4)+1);
}
#define init_13_10_1_2
void FASTCALL call_13_10_1_2(SH2_struct* sh)
{
  SH2movli(sh, 10, (1<<4)+2);
}
#define init_13_10_1_3
void FASTCALL call_13_10_1_3(SH2_struct* sh)
{
  SH2movli(sh, 10, (1<<4)+3);
}
#define init_13_10_1_4
void FASTCALL call_13_10_1_4(SH2_struct* sh)
{
  SH2movli(sh, 10, (1<<4)+4);
}
#define init_13_10_1_5
void FASTCALL call_13_10_1_5(SH2_struct* sh)
{
  SH2movli(sh, 10, (1<<4)+5);
}
#define init_13_10_1_6
void FASTCALL call_13_10_1_6(SH2_struct* sh)
{
  SH2movli(sh, 10, (1<<4)+6);
}
#define init_13_10_1_7
void FASTCALL call_13_10_1_7(SH2_struct* sh)
{
  SH2movli(sh, 10, (1<<4)+7);
}
#define init_13_10_1_8
void FASTCALL call_13_10_1_8(SH2_struct* sh)
{
  SH2movli(sh, 10, (1<<4)+8);
}
#define init_13_10_1_9
void FASTCALL call_13_10_1_9(SH2_struct* sh)
{
  SH2movli(sh, 10, (1<<4)+9);
}
#define init_13_10_1_10
void FASTCALL call_13_10_1_10(SH2_struct* sh)
{
  SH2movli(sh, 10, (1<<4)+10);
}
#define init_13_10_1_11
void FASTCALL call_13_10_1_11(SH2_struct* sh)
{
  SH2movli(sh, 10, (1<<4)+11);
}
#define init_13_10_1_12
void FASTCALL call_13_10_1_12(SH2_struct* sh)
{
  SH2movli(sh, 10, (1<<4)+12);
}
#define init_13_10_1_13
void FASTCALL call_13_10_1_13(SH2_struct* sh)
{
  SH2movli(sh, 10, (1<<4)+13);
}
#define init_13_10_1_14
void FASTCALL call_13_10_1_14(SH2_struct* sh)
{
  SH2movli(sh, 10, (1<<4)+14);
}
#define init_13_10_1_15
void FASTCALL call_13_10_1_15(SH2_struct* sh)
{
  SH2movli(sh, 10, (1<<4)+15);
}
#define init_13_10_2_0
void FASTCALL call_13_10_2_0(SH2_struct* sh)
{
  SH2movli(sh, 10, (2<<4)+0);
}
#define init_13_10_2_1
void FASTCALL call_13_10_2_1(SH2_struct* sh)
{
  SH2movli(sh, 10, (2<<4)+1);
}
#define init_13_10_2_2
void FASTCALL call_13_10_2_2(SH2_struct* sh)
{
  SH2movli(sh, 10, (2<<4)+2);
}
#define init_13_10_2_3
void FASTCALL call_13_10_2_3(SH2_struct* sh)
{
  SH2movli(sh, 10, (2<<4)+3);
}
#define init_13_10_2_4
void FASTCALL call_13_10_2_4(SH2_struct* sh)
{
  SH2movli(sh, 10, (2<<4)+4);
}
#define init_13_10_2_5
void FASTCALL call_13_10_2_5(SH2_struct* sh)
{
  SH2movli(sh, 10, (2<<4)+5);
}
#define init_13_10_2_6
void FASTCALL call_13_10_2_6(SH2_struct* sh)
{
  SH2movli(sh, 10, (2<<4)+6);
}
#define init_13_10_2_7
void FASTCALL call_13_10_2_7(SH2_struct* sh)
{
  SH2movli(sh, 10, (2<<4)+7);
}
#define init_13_10_2_8
void FASTCALL call_13_10_2_8(SH2_struct* sh)
{
  SH2movli(sh, 10, (2<<4)+8);
}
#define init_13_10_2_9
void FASTCALL call_13_10_2_9(SH2_struct* sh)
{
  SH2movli(sh, 10, (2<<4)+9);
}
#define init_13_10_2_10
void FASTCALL call_13_10_2_10(SH2_struct* sh)
{
  SH2movli(sh, 10, (2<<4)+10);
}
#define init_13_10_2_11
void FASTCALL call_13_10_2_11(SH2_struct* sh)
{
  SH2movli(sh, 10, (2<<4)+11);
}
#define init_13_10_2_12
void FASTCALL call_13_10_2_12(SH2_struct* sh)
{
  SH2movli(sh, 10, (2<<4)+12);
}
#define init_13_10_2_13
void FASTCALL call_13_10_2_13(SH2_struct* sh)
{
  SH2movli(sh, 10, (2<<4)+13);
}
#define init_13_10_2_14
void FASTCALL call_13_10_2_14(SH2_struct* sh)
{
  SH2movli(sh, 10, (2<<4)+14);
}
#define init_13_10_2_15
void FASTCALL call_13_10_2_15(SH2_struct* sh)
{
  SH2movli(sh, 10, (2<<4)+15);
}
#define init_13_10_3_0
void FASTCALL call_13_10_3_0(SH2_struct* sh)
{
  SH2movli(sh, 10, (3<<4)+0);
}
#define init_13_10_3_1
void FASTCALL call_13_10_3_1(SH2_struct* sh)
{
  SH2movli(sh, 10, (3<<4)+1);
}
#define init_13_10_3_2
void FASTCALL call_13_10_3_2(SH2_struct* sh)
{
  SH2movli(sh, 10, (3<<4)+2);
}
#define init_13_10_3_3
void FASTCALL call_13_10_3_3(SH2_struct* sh)
{
  SH2movli(sh, 10, (3<<4)+3);
}
#define init_13_10_3_4
void FASTCALL call_13_10_3_4(SH2_struct* sh)
{
  SH2movli(sh, 10, (3<<4)+4);
}
#define init_13_10_3_5
void FASTCALL call_13_10_3_5(SH2_struct* sh)
{
  SH2movli(sh, 10, (3<<4)+5);
}
#define init_13_10_3_6
void FASTCALL call_13_10_3_6(SH2_struct* sh)
{
  SH2movli(sh, 10, (3<<4)+6);
}
#define init_13_10_3_7
void FASTCALL call_13_10_3_7(SH2_struct* sh)
{
  SH2movli(sh, 10, (3<<4)+7);
}
#define init_13_10_3_8
void FASTCALL call_13_10_3_8(SH2_struct* sh)
{
  SH2movli(sh, 10, (3<<4)+8);
}
#define init_13_10_3_9
void FASTCALL call_13_10_3_9(SH2_struct* sh)
{
  SH2movli(sh, 10, (3<<4)+9);
}
#define init_13_10_3_10
void FASTCALL call_13_10_3_10(SH2_struct* sh)
{
  SH2movli(sh, 10, (3<<4)+10);
}
#define init_13_10_3_11
void FASTCALL call_13_10_3_11(SH2_struct* sh)
{
  SH2movli(sh, 10, (3<<4)+11);
}
#define init_13_10_3_12
void FASTCALL call_13_10_3_12(SH2_struct* sh)
{
  SH2movli(sh, 10, (3<<4)+12);
}
#define init_13_10_3_13
void FASTCALL call_13_10_3_13(SH2_struct* sh)
{
  SH2movli(sh, 10, (3<<4)+13);
}
#define init_13_10_3_14
void FASTCALL call_13_10_3_14(SH2_struct* sh)
{
  SH2movli(sh, 10, (3<<4)+14);
}
#define init_13_10_3_15
void FASTCALL call_13_10_3_15(SH2_struct* sh)
{
  SH2movli(sh, 10, (3<<4)+15);
}
#define init_13_10_4_0
void FASTCALL call_13_10_4_0(SH2_struct* sh)
{
  SH2movli(sh, 10, (4<<4)+0);
}
#define init_13_10_4_1
void FASTCALL call_13_10_4_1(SH2_struct* sh)
{
  SH2movli(sh, 10, (4<<4)+1);
}
#define init_13_10_4_2
void FASTCALL call_13_10_4_2(SH2_struct* sh)
{
  SH2movli(sh, 10, (4<<4)+2);
}
#define init_13_10_4_3
void FASTCALL call_13_10_4_3(SH2_struct* sh)
{
  SH2movli(sh, 10, (4<<4)+3);
}
#define init_13_10_4_4
void FASTCALL call_13_10_4_4(SH2_struct* sh)
{
  SH2movli(sh, 10, (4<<4)+4);
}
#define init_13_10_4_5
void FASTCALL call_13_10_4_5(SH2_struct* sh)
{
  SH2movli(sh, 10, (4<<4)+5);
}
#define init_13_10_4_6
void FASTCALL call_13_10_4_6(SH2_struct* sh)
{
  SH2movli(sh, 10, (4<<4)+6);
}
#define init_13_10_4_7
void FASTCALL call_13_10_4_7(SH2_struct* sh)
{
  SH2movli(sh, 10, (4<<4)+7);
}
#define init_13_10_4_8
void FASTCALL call_13_10_4_8(SH2_struct* sh)
{
  SH2movli(sh, 10, (4<<4)+8);
}
#define init_13_10_4_9
void FASTCALL call_13_10_4_9(SH2_struct* sh)
{
  SH2movli(sh, 10, (4<<4)+9);
}
#define init_13_10_4_10
void FASTCALL call_13_10_4_10(SH2_struct* sh)
{
  SH2movli(sh, 10, (4<<4)+10);
}
#define init_13_10_4_11
void FASTCALL call_13_10_4_11(SH2_struct* sh)
{
  SH2movli(sh, 10, (4<<4)+11);
}
#define init_13_10_4_12
void FASTCALL call_13_10_4_12(SH2_struct* sh)
{
  SH2movli(sh, 10, (4<<4)+12);
}
#define init_13_10_4_13
void FASTCALL call_13_10_4_13(SH2_struct* sh)
{
  SH2movli(sh, 10, (4<<4)+13);
}
#define init_13_10_4_14
void FASTCALL call_13_10_4_14(SH2_struct* sh)
{
  SH2movli(sh, 10, (4<<4)+14);
}
#define init_13_10_4_15
void FASTCALL call_13_10_4_15(SH2_struct* sh)
{
  SH2movli(sh, 10, (4<<4)+15);
}
#define init_13_10_5_0
void FASTCALL call_13_10_5_0(SH2_struct* sh)
{
  SH2movli(sh, 10, (5<<4)+0);
}
#define init_13_10_5_1
void FASTCALL call_13_10_5_1(SH2_struct* sh)
{
  SH2movli(sh, 10, (5<<4)+1);
}
#define init_13_10_5_2
void FASTCALL call_13_10_5_2(SH2_struct* sh)
{
  SH2movli(sh, 10, (5<<4)+2);
}
#define init_13_10_5_3
void FASTCALL call_13_10_5_3(SH2_struct* sh)
{
  SH2movli(sh, 10, (5<<4)+3);
}
#define init_13_10_5_4
void FASTCALL call_13_10_5_4(SH2_struct* sh)
{
  SH2movli(sh, 10, (5<<4)+4);
}
#define init_13_10_5_5
void FASTCALL call_13_10_5_5(SH2_struct* sh)
{
  SH2movli(sh, 10, (5<<4)+5);
}
#define init_13_10_5_6
void FASTCALL call_13_10_5_6(SH2_struct* sh)
{
  SH2movli(sh, 10, (5<<4)+6);
}
#define init_13_10_5_7
void FASTCALL call_13_10_5_7(SH2_struct* sh)
{
  SH2movli(sh, 10, (5<<4)+7);
}
#define init_13_10_5_8
void FASTCALL call_13_10_5_8(SH2_struct* sh)
{
  SH2movli(sh, 10, (5<<4)+8);
}
#define init_13_10_5_9
void FASTCALL call_13_10_5_9(SH2_struct* sh)
{
  SH2movli(sh, 10, (5<<4)+9);
}
#define init_13_10_5_10
void FASTCALL call_13_10_5_10(SH2_struct* sh)
{
  SH2movli(sh, 10, (5<<4)+10);
}
#define init_13_10_5_11
void FASTCALL call_13_10_5_11(SH2_struct* sh)
{
  SH2movli(sh, 10, (5<<4)+11);
}
#define init_13_10_5_12
void FASTCALL call_13_10_5_12(SH2_struct* sh)
{
  SH2movli(sh, 10, (5<<4)+12);
}
#define init_13_10_5_13
void FASTCALL call_13_10_5_13(SH2_struct* sh)
{
  SH2movli(sh, 10, (5<<4)+13);
}
#define init_13_10_5_14
void FASTCALL call_13_10_5_14(SH2_struct* sh)
{
  SH2movli(sh, 10, (5<<4)+14);
}
#define init_13_10_5_15
void FASTCALL call_13_10_5_15(SH2_struct* sh)
{
  SH2movli(sh, 10, (5<<4)+15);
}
#define init_13_10_6_0
void FASTCALL call_13_10_6_0(SH2_struct* sh)
{
  SH2movli(sh, 10, (6<<4)+0);
}
#define init_13_10_6_1
void FASTCALL call_13_10_6_1(SH2_struct* sh)
{
  SH2movli(sh, 10, (6<<4)+1);
}
#define init_13_10_6_2
void FASTCALL call_13_10_6_2(SH2_struct* sh)
{
  SH2movli(sh, 10, (6<<4)+2);
}
#define init_13_10_6_3
void FASTCALL call_13_10_6_3(SH2_struct* sh)
{
  SH2movli(sh, 10, (6<<4)+3);
}
#define init_13_10_6_4
void FASTCALL call_13_10_6_4(SH2_struct* sh)
{
  SH2movli(sh, 10, (6<<4)+4);
}
#define init_13_10_6_5
void FASTCALL call_13_10_6_5(SH2_struct* sh)
{
  SH2movli(sh, 10, (6<<4)+5);
}
#define init_13_10_6_6
void FASTCALL call_13_10_6_6(SH2_struct* sh)
{
  SH2movli(sh, 10, (6<<4)+6);
}
#define init_13_10_6_7
void FASTCALL call_13_10_6_7(SH2_struct* sh)
{
  SH2movli(sh, 10, (6<<4)+7);
}
#define init_13_10_6_8
void FASTCALL call_13_10_6_8(SH2_struct* sh)
{
  SH2movli(sh, 10, (6<<4)+8);
}
#define init_13_10_6_9
void FASTCALL call_13_10_6_9(SH2_struct* sh)
{
  SH2movli(sh, 10, (6<<4)+9);
}
#define init_13_10_6_10
void FASTCALL call_13_10_6_10(SH2_struct* sh)
{
  SH2movli(sh, 10, (6<<4)+10);
}
#define init_13_10_6_11
void FASTCALL call_13_10_6_11(SH2_struct* sh)
{
  SH2movli(sh, 10, (6<<4)+11);
}
#define init_13_10_6_12
void FASTCALL call_13_10_6_12(SH2_struct* sh)
{
  SH2movli(sh, 10, (6<<4)+12);
}
#define init_13_10_6_13
void FASTCALL call_13_10_6_13(SH2_struct* sh)
{
  SH2movli(sh, 10, (6<<4)+13);
}
#define init_13_10_6_14
void FASTCALL call_13_10_6_14(SH2_struct* sh)
{
  SH2movli(sh, 10, (6<<4)+14);
}
#define init_13_10_6_15
void FASTCALL call_13_10_6_15(SH2_struct* sh)
{
  SH2movli(sh, 10, (6<<4)+15);
}
#define init_13_10_7_0
void FASTCALL call_13_10_7_0(SH2_struct* sh)
{
  SH2movli(sh, 10, (7<<4)+0);
}
#define init_13_10_7_1
void FASTCALL call_13_10_7_1(SH2_struct* sh)
{
  SH2movli(sh, 10, (7<<4)+1);
}
#define init_13_10_7_2
void FASTCALL call_13_10_7_2(SH2_struct* sh)
{
  SH2movli(sh, 10, (7<<4)+2);
}
#define init_13_10_7_3
void FASTCALL call_13_10_7_3(SH2_struct* sh)
{
  SH2movli(sh, 10, (7<<4)+3);
}
#define init_13_10_7_4
void FASTCALL call_13_10_7_4(SH2_struct* sh)
{
  SH2movli(sh, 10, (7<<4)+4);
}
#define init_13_10_7_5
void FASTCALL call_13_10_7_5(SH2_struct* sh)
{
  SH2movli(sh, 10, (7<<4)+5);
}
#define init_13_10_7_6
void FASTCALL call_13_10_7_6(SH2_struct* sh)
{
  SH2movli(sh, 10, (7<<4)+6);
}
#define init_13_10_7_7
void FASTCALL call_13_10_7_7(SH2_struct* sh)
{
  SH2movli(sh, 10, (7<<4)+7);
}
#define init_13_10_7_8
void FASTCALL call_13_10_7_8(SH2_struct* sh)
{
  SH2movli(sh, 10, (7<<4)+8);
}
#define init_13_10_7_9
void FASTCALL call_13_10_7_9(SH2_struct* sh)
{
  SH2movli(sh, 10, (7<<4)+9);
}
#define init_13_10_7_10
void FASTCALL call_13_10_7_10(SH2_struct* sh)
{
  SH2movli(sh, 10, (7<<4)+10);
}
#define init_13_10_7_11
void FASTCALL call_13_10_7_11(SH2_struct* sh)
{
  SH2movli(sh, 10, (7<<4)+11);
}
#define init_13_10_7_12
void FASTCALL call_13_10_7_12(SH2_struct* sh)
{
  SH2movli(sh, 10, (7<<4)+12);
}
#define init_13_10_7_13
void FASTCALL call_13_10_7_13(SH2_struct* sh)
{
  SH2movli(sh, 10, (7<<4)+13);
}
#define init_13_10_7_14
void FASTCALL call_13_10_7_14(SH2_struct* sh)
{
  SH2movli(sh, 10, (7<<4)+14);
}
#define init_13_10_7_15
void FASTCALL call_13_10_7_15(SH2_struct* sh)
{
  SH2movli(sh, 10, (7<<4)+15);
}
#define init_13_10_8_0
void FASTCALL call_13_10_8_0(SH2_struct* sh)
{
  SH2movli(sh, 10, (8<<4)+0);
}
#define init_13_10_8_1
void FASTCALL call_13_10_8_1(SH2_struct* sh)
{
  SH2movli(sh, 10, (8<<4)+1);
}
#define init_13_10_8_2
void FASTCALL call_13_10_8_2(SH2_struct* sh)
{
  SH2movli(sh, 10, (8<<4)+2);
}
#define init_13_10_8_3
void FASTCALL call_13_10_8_3(SH2_struct* sh)
{
  SH2movli(sh, 10, (8<<4)+3);
}
#define init_13_10_8_4
void FASTCALL call_13_10_8_4(SH2_struct* sh)
{
  SH2movli(sh, 10, (8<<4)+4);
}
#define init_13_10_8_5
void FASTCALL call_13_10_8_5(SH2_struct* sh)
{
  SH2movli(sh, 10, (8<<4)+5);
}
#define init_13_10_8_6
void FASTCALL call_13_10_8_6(SH2_struct* sh)
{
  SH2movli(sh, 10, (8<<4)+6);
}
#define init_13_10_8_7
void FASTCALL call_13_10_8_7(SH2_struct* sh)
{
  SH2movli(sh, 10, (8<<4)+7);
}
#define init_13_10_8_8
void FASTCALL call_13_10_8_8(SH2_struct* sh)
{
  SH2movli(sh, 10, (8<<4)+8);
}
#define init_13_10_8_9
void FASTCALL call_13_10_8_9(SH2_struct* sh)
{
  SH2movli(sh, 10, (8<<4)+9);
}
#define init_13_10_8_10
void FASTCALL call_13_10_8_10(SH2_struct* sh)
{
  SH2movli(sh, 10, (8<<4)+10);
}
#define init_13_10_8_11
void FASTCALL call_13_10_8_11(SH2_struct* sh)
{
  SH2movli(sh, 10, (8<<4)+11);
}
#define init_13_10_8_12
void FASTCALL call_13_10_8_12(SH2_struct* sh)
{
  SH2movli(sh, 10, (8<<4)+12);
}
#define init_13_10_8_13
void FASTCALL call_13_10_8_13(SH2_struct* sh)
{
  SH2movli(sh, 10, (8<<4)+13);
}
#define init_13_10_8_14
void FASTCALL call_13_10_8_14(SH2_struct* sh)
{
  SH2movli(sh, 10, (8<<4)+14);
}
#define init_13_10_8_15
void FASTCALL call_13_10_8_15(SH2_struct* sh)
{
  SH2movli(sh, 10, (8<<4)+15);
}
#define init_13_10_9_0
void FASTCALL call_13_10_9_0(SH2_struct* sh)
{
  SH2movli(sh, 10, (9<<4)+0);
}
#define init_13_10_9_1
void FASTCALL call_13_10_9_1(SH2_struct* sh)
{
  SH2movli(sh, 10, (9<<4)+1);
}
#define init_13_10_9_2
void FASTCALL call_13_10_9_2(SH2_struct* sh)
{
  SH2movli(sh, 10, (9<<4)+2);
}
#define init_13_10_9_3
void FASTCALL call_13_10_9_3(SH2_struct* sh)
{
  SH2movli(sh, 10, (9<<4)+3);
}
#define init_13_10_9_4
void FASTCALL call_13_10_9_4(SH2_struct* sh)
{
  SH2movli(sh, 10, (9<<4)+4);
}
#define init_13_10_9_5
void FASTCALL call_13_10_9_5(SH2_struct* sh)
{
  SH2movli(sh, 10, (9<<4)+5);
}
#define init_13_10_9_6
void FASTCALL call_13_10_9_6(SH2_struct* sh)
{
  SH2movli(sh, 10, (9<<4)+6);
}
#define init_13_10_9_7
void FASTCALL call_13_10_9_7(SH2_struct* sh)
{
  SH2movli(sh, 10, (9<<4)+7);
}
#define init_13_10_9_8
void FASTCALL call_13_10_9_8(SH2_struct* sh)
{
  SH2movli(sh, 10, (9<<4)+8);
}
#define init_13_10_9_9
void FASTCALL call_13_10_9_9(SH2_struct* sh)
{
  SH2movli(sh, 10, (9<<4)+9);
}
#define init_13_10_9_10
void FASTCALL call_13_10_9_10(SH2_struct* sh)
{
  SH2movli(sh, 10, (9<<4)+10);
}
#define init_13_10_9_11
void FASTCALL call_13_10_9_11(SH2_struct* sh)
{
  SH2movli(sh, 10, (9<<4)+11);
}
#define init_13_10_9_12
void FASTCALL call_13_10_9_12(SH2_struct* sh)
{
  SH2movli(sh, 10, (9<<4)+12);
}
#define init_13_10_9_13
void FASTCALL call_13_10_9_13(SH2_struct* sh)
{
  SH2movli(sh, 10, (9<<4)+13);
}
#define init_13_10_9_14
void FASTCALL call_13_10_9_14(SH2_struct* sh)
{
  SH2movli(sh, 10, (9<<4)+14);
}
#define init_13_10_9_15
void FASTCALL call_13_10_9_15(SH2_struct* sh)
{
  SH2movli(sh, 10, (9<<4)+15);
}
#define init_13_10_10_0
void FASTCALL call_13_10_10_0(SH2_struct* sh)
{
  SH2movli(sh, 10, (10<<4)+0);
}
#define init_13_10_10_1
void FASTCALL call_13_10_10_1(SH2_struct* sh)
{
  SH2movli(sh, 10, (10<<4)+1);
}
#define init_13_10_10_2
void FASTCALL call_13_10_10_2(SH2_struct* sh)
{
  SH2movli(sh, 10, (10<<4)+2);
}
#define init_13_10_10_3
void FASTCALL call_13_10_10_3(SH2_struct* sh)
{
  SH2movli(sh, 10, (10<<4)+3);
}
#define init_13_10_10_4
void FASTCALL call_13_10_10_4(SH2_struct* sh)
{
  SH2movli(sh, 10, (10<<4)+4);
}
#define init_13_10_10_5
void FASTCALL call_13_10_10_5(SH2_struct* sh)
{
  SH2movli(sh, 10, (10<<4)+5);
}
#define init_13_10_10_6
void FASTCALL call_13_10_10_6(SH2_struct* sh)
{
  SH2movli(sh, 10, (10<<4)+6);
}
#define init_13_10_10_7
void FASTCALL call_13_10_10_7(SH2_struct* sh)
{
  SH2movli(sh, 10, (10<<4)+7);
}
#define init_13_10_10_8
void FASTCALL call_13_10_10_8(SH2_struct* sh)
{
  SH2movli(sh, 10, (10<<4)+8);
}
#define init_13_10_10_9
void FASTCALL call_13_10_10_9(SH2_struct* sh)
{
  SH2movli(sh, 10, (10<<4)+9);
}
#define init_13_10_10_10
void FASTCALL call_13_10_10_10(SH2_struct* sh)
{
  SH2movli(sh, 10, (10<<4)+10);
}
#define init_13_10_10_11
void FASTCALL call_13_10_10_11(SH2_struct* sh)
{
  SH2movli(sh, 10, (10<<4)+11);
}
#define init_13_10_10_12
void FASTCALL call_13_10_10_12(SH2_struct* sh)
{
  SH2movli(sh, 10, (10<<4)+12);
}
#define init_13_10_10_13
void FASTCALL call_13_10_10_13(SH2_struct* sh)
{
  SH2movli(sh, 10, (10<<4)+13);
}
#define init_13_10_10_14
void FASTCALL call_13_10_10_14(SH2_struct* sh)
{
  SH2movli(sh, 10, (10<<4)+14);
}
#define init_13_10_10_15
void FASTCALL call_13_10_10_15(SH2_struct* sh)
{
  SH2movli(sh, 10, (10<<4)+15);
}
#define init_13_10_11_0
void FASTCALL call_13_10_11_0(SH2_struct* sh)
{
  SH2movli(sh, 10, (11<<4)+0);
}
#define init_13_10_11_1
void FASTCALL call_13_10_11_1(SH2_struct* sh)
{
  SH2movli(sh, 10, (11<<4)+1);
}
#define init_13_10_11_2
void FASTCALL call_13_10_11_2(SH2_struct* sh)
{
  SH2movli(sh, 10, (11<<4)+2);
}
#define init_13_10_11_3
void FASTCALL call_13_10_11_3(SH2_struct* sh)
{
  SH2movli(sh, 10, (11<<4)+3);
}
#define init_13_10_11_4
void FASTCALL call_13_10_11_4(SH2_struct* sh)
{
  SH2movli(sh, 10, (11<<4)+4);
}
#define init_13_10_11_5
void FASTCALL call_13_10_11_5(SH2_struct* sh)
{
  SH2movli(sh, 10, (11<<4)+5);
}
#define init_13_10_11_6
void FASTCALL call_13_10_11_6(SH2_struct* sh)
{
  SH2movli(sh, 10, (11<<4)+6);
}
#define init_13_10_11_7
void FASTCALL call_13_10_11_7(SH2_struct* sh)
{
  SH2movli(sh, 10, (11<<4)+7);
}
#define init_13_10_11_8
void FASTCALL call_13_10_11_8(SH2_struct* sh)
{
  SH2movli(sh, 10, (11<<4)+8);
}
#define init_13_10_11_9
void FASTCALL call_13_10_11_9(SH2_struct* sh)
{
  SH2movli(sh, 10, (11<<4)+9);
}
#define init_13_10_11_10
void FASTCALL call_13_10_11_10(SH2_struct* sh)
{
  SH2movli(sh, 10, (11<<4)+10);
}
#define init_13_10_11_11
void FASTCALL call_13_10_11_11(SH2_struct* sh)
{
  SH2movli(sh, 10, (11<<4)+11);
}
#define init_13_10_11_12
void FASTCALL call_13_10_11_12(SH2_struct* sh)
{
  SH2movli(sh, 10, (11<<4)+12);
}
#define init_13_10_11_13
void FASTCALL call_13_10_11_13(SH2_struct* sh)
{
  SH2movli(sh, 10, (11<<4)+13);
}
#define init_13_10_11_14
void FASTCALL call_13_10_11_14(SH2_struct* sh)
{
  SH2movli(sh, 10, (11<<4)+14);
}
#define init_13_10_11_15
void FASTCALL call_13_10_11_15(SH2_struct* sh)
{
  SH2movli(sh, 10, (11<<4)+15);
}
#define init_13_10_12_0
void FASTCALL call_13_10_12_0(SH2_struct* sh)
{
  SH2movli(sh, 10, (12<<4)+0);
}
#define init_13_10_12_1
void FASTCALL call_13_10_12_1(SH2_struct* sh)
{
  SH2movli(sh, 10, (12<<4)+1);
}
#define init_13_10_12_2
void FASTCALL call_13_10_12_2(SH2_struct* sh)
{
  SH2movli(sh, 10, (12<<4)+2);
}
#define init_13_10_12_3
void FASTCALL call_13_10_12_3(SH2_struct* sh)
{
  SH2movli(sh, 10, (12<<4)+3);
}
#define init_13_10_12_4
void FASTCALL call_13_10_12_4(SH2_struct* sh)
{
  SH2movli(sh, 10, (12<<4)+4);
}
#define init_13_10_12_5
void FASTCALL call_13_10_12_5(SH2_struct* sh)
{
  SH2movli(sh, 10, (12<<4)+5);
}
#define init_13_10_12_6
void FASTCALL call_13_10_12_6(SH2_struct* sh)
{
  SH2movli(sh, 10, (12<<4)+6);
}
#define init_13_10_12_7
void FASTCALL call_13_10_12_7(SH2_struct* sh)
{
  SH2movli(sh, 10, (12<<4)+7);
}
#define init_13_10_12_8
void FASTCALL call_13_10_12_8(SH2_struct* sh)
{
  SH2movli(sh, 10, (12<<4)+8);
}
#define init_13_10_12_9
void FASTCALL call_13_10_12_9(SH2_struct* sh)
{
  SH2movli(sh, 10, (12<<4)+9);
}
#define init_13_10_12_10
void FASTCALL call_13_10_12_10(SH2_struct* sh)
{
  SH2movli(sh, 10, (12<<4)+10);
}
#define init_13_10_12_11
void FASTCALL call_13_10_12_11(SH2_struct* sh)
{
  SH2movli(sh, 10, (12<<4)+11);
}
#define init_13_10_12_12
void FASTCALL call_13_10_12_12(SH2_struct* sh)
{
  SH2movli(sh, 10, (12<<4)+12);
}
#define init_13_10_12_13
void FASTCALL call_13_10_12_13(SH2_struct* sh)
{
  SH2movli(sh, 10, (12<<4)+13);
}
#define init_13_10_12_14
void FASTCALL call_13_10_12_14(SH2_struct* sh)
{
  SH2movli(sh, 10, (12<<4)+14);
}
#define init_13_10_12_15
void FASTCALL call_13_10_12_15(SH2_struct* sh)
{
  SH2movli(sh, 10, (12<<4)+15);
}
#define init_13_10_13_0
void FASTCALL call_13_10_13_0(SH2_struct* sh)
{
  SH2movli(sh, 10, (13<<4)+0);
}
#define init_13_10_13_1
void FASTCALL call_13_10_13_1(SH2_struct* sh)
{
  SH2movli(sh, 10, (13<<4)+1);
}
#define init_13_10_13_2
void FASTCALL call_13_10_13_2(SH2_struct* sh)
{
  SH2movli(sh, 10, (13<<4)+2);
}
#define init_13_10_13_3
void FASTCALL call_13_10_13_3(SH2_struct* sh)
{
  SH2movli(sh, 10, (13<<4)+3);
}
#define init_13_10_13_4
void FASTCALL call_13_10_13_4(SH2_struct* sh)
{
  SH2movli(sh, 10, (13<<4)+4);
}
#define init_13_10_13_5
void FASTCALL call_13_10_13_5(SH2_struct* sh)
{
  SH2movli(sh, 10, (13<<4)+5);
}
#define init_13_10_13_6
void FASTCALL call_13_10_13_6(SH2_struct* sh)
{
  SH2movli(sh, 10, (13<<4)+6);
}
#define init_13_10_13_7
void FASTCALL call_13_10_13_7(SH2_struct* sh)
{
  SH2movli(sh, 10, (13<<4)+7);
}
#define init_13_10_13_8
void FASTCALL call_13_10_13_8(SH2_struct* sh)
{
  SH2movli(sh, 10, (13<<4)+8);
}
#define init_13_10_13_9
void FASTCALL call_13_10_13_9(SH2_struct* sh)
{
  SH2movli(sh, 10, (13<<4)+9);
}
#define init_13_10_13_10
void FASTCALL call_13_10_13_10(SH2_struct* sh)
{
  SH2movli(sh, 10, (13<<4)+10);
}
#define init_13_10_13_11
void FASTCALL call_13_10_13_11(SH2_struct* sh)
{
  SH2movli(sh, 10, (13<<4)+11);
}
#define init_13_10_13_12
void FASTCALL call_13_10_13_12(SH2_struct* sh)
{
  SH2movli(sh, 10, (13<<4)+12);
}
#define init_13_10_13_13
void FASTCALL call_13_10_13_13(SH2_struct* sh)
{
  SH2movli(sh, 10, (13<<4)+13);
}
#define init_13_10_13_14
void FASTCALL call_13_10_13_14(SH2_struct* sh)
{
  SH2movli(sh, 10, (13<<4)+14);
}
#define init_13_10_13_15
void FASTCALL call_13_10_13_15(SH2_struct* sh)
{
  SH2movli(sh, 10, (13<<4)+15);
}
#define init_13_10_14_0
void FASTCALL call_13_10_14_0(SH2_struct* sh)
{
  SH2movli(sh, 10, (14<<4)+0);
}
#define init_13_10_14_1
void FASTCALL call_13_10_14_1(SH2_struct* sh)
{
  SH2movli(sh, 10, (14<<4)+1);
}
#define init_13_10_14_2
void FASTCALL call_13_10_14_2(SH2_struct* sh)
{
  SH2movli(sh, 10, (14<<4)+2);
}
#define init_13_10_14_3
void FASTCALL call_13_10_14_3(SH2_struct* sh)
{
  SH2movli(sh, 10, (14<<4)+3);
}
#define init_13_10_14_4
void FASTCALL call_13_10_14_4(SH2_struct* sh)
{
  SH2movli(sh, 10, (14<<4)+4);
}
#define init_13_10_14_5
void FASTCALL call_13_10_14_5(SH2_struct* sh)
{
  SH2movli(sh, 10, (14<<4)+5);
}
#define init_13_10_14_6
void FASTCALL call_13_10_14_6(SH2_struct* sh)
{
  SH2movli(sh, 10, (14<<4)+6);
}
#define init_13_10_14_7
void FASTCALL call_13_10_14_7(SH2_struct* sh)
{
  SH2movli(sh, 10, (14<<4)+7);
}
#define init_13_10_14_8
void FASTCALL call_13_10_14_8(SH2_struct* sh)
{
  SH2movli(sh, 10, (14<<4)+8);
}
#define init_13_10_14_9
void FASTCALL call_13_10_14_9(SH2_struct* sh)
{
  SH2movli(sh, 10, (14<<4)+9);
}
#define init_13_10_14_10
void FASTCALL call_13_10_14_10(SH2_struct* sh)
{
  SH2movli(sh, 10, (14<<4)+10);
}
#define init_13_10_14_11
void FASTCALL call_13_10_14_11(SH2_struct* sh)
{
  SH2movli(sh, 10, (14<<4)+11);
}
#define init_13_10_14_12
void FASTCALL call_13_10_14_12(SH2_struct* sh)
{
  SH2movli(sh, 10, (14<<4)+12);
}
#define init_13_10_14_13
void FASTCALL call_13_10_14_13(SH2_struct* sh)
{
  SH2movli(sh, 10, (14<<4)+13);
}
#define init_13_10_14_14
void FASTCALL call_13_10_14_14(SH2_struct* sh)
{
  SH2movli(sh, 10, (14<<4)+14);
}
#define init_13_10_14_15
void FASTCALL call_13_10_14_15(SH2_struct* sh)
{
  SH2movli(sh, 10, (14<<4)+15);
}
#define init_13_10_15_0
void FASTCALL call_13_10_15_0(SH2_struct* sh)
{
  SH2movli(sh, 10, (15<<4)+0);
}
#define init_13_10_15_1
void FASTCALL call_13_10_15_1(SH2_struct* sh)
{
  SH2movli(sh, 10, (15<<4)+1);
}
#define init_13_10_15_2
void FASTCALL call_13_10_15_2(SH2_struct* sh)
{
  SH2movli(sh, 10, (15<<4)+2);
}
#define init_13_10_15_3
void FASTCALL call_13_10_15_3(SH2_struct* sh)
{
  SH2movli(sh, 10, (15<<4)+3);
}
#define init_13_10_15_4
void FASTCALL call_13_10_15_4(SH2_struct* sh)
{
  SH2movli(sh, 10, (15<<4)+4);
}
#define init_13_10_15_5
void FASTCALL call_13_10_15_5(SH2_struct* sh)
{
  SH2movli(sh, 10, (15<<4)+5);
}
#define init_13_10_15_6
void FASTCALL call_13_10_15_6(SH2_struct* sh)
{
  SH2movli(sh, 10, (15<<4)+6);
}
#define init_13_10_15_7
void FASTCALL call_13_10_15_7(SH2_struct* sh)
{
  SH2movli(sh, 10, (15<<4)+7);
}
#define init_13_10_15_8
void FASTCALL call_13_10_15_8(SH2_struct* sh)
{
  SH2movli(sh, 10, (15<<4)+8);
}
#define init_13_10_15_9
void FASTCALL call_13_10_15_9(SH2_struct* sh)
{
  SH2movli(sh, 10, (15<<4)+9);
}
#define init_13_10_15_10
void FASTCALL call_13_10_15_10(SH2_struct* sh)
{
  SH2movli(sh, 10, (15<<4)+10);
}
#define init_13_10_15_11
void FASTCALL call_13_10_15_11(SH2_struct* sh)
{
  SH2movli(sh, 10, (15<<4)+11);
}
#define init_13_10_15_12
void FASTCALL call_13_10_15_12(SH2_struct* sh)
{
  SH2movli(sh, 10, (15<<4)+12);
}
#define init_13_10_15_13
void FASTCALL call_13_10_15_13(SH2_struct* sh)
{
  SH2movli(sh, 10, (15<<4)+13);
}
#define init_13_10_15_14
void FASTCALL call_13_10_15_14(SH2_struct* sh)
{
  SH2movli(sh, 10, (15<<4)+14);
}
#define init_13_10_15_15
void FASTCALL call_13_10_15_15(SH2_struct* sh)
{
  SH2movli(sh, 10, (15<<4)+15);
}
#define init_13_11_0_0
void FASTCALL call_13_11_0_0(SH2_struct* sh)
{
  SH2movli(sh, 11, (0<<4)+0);
}
#define init_13_11_0_1
void FASTCALL call_13_11_0_1(SH2_struct* sh)
{
  SH2movli(sh, 11, (0<<4)+1);
}
#define init_13_11_0_2
void FASTCALL call_13_11_0_2(SH2_struct* sh)
{
  SH2movli(sh, 11, (0<<4)+2);
}
#define init_13_11_0_3
void FASTCALL call_13_11_0_3(SH2_struct* sh)
{
  SH2movli(sh, 11, (0<<4)+3);
}
#define init_13_11_0_4
void FASTCALL call_13_11_0_4(SH2_struct* sh)
{
  SH2movli(sh, 11, (0<<4)+4);
}
#define init_13_11_0_5
void FASTCALL call_13_11_0_5(SH2_struct* sh)
{
  SH2movli(sh, 11, (0<<4)+5);
}
#define init_13_11_0_6
void FASTCALL call_13_11_0_6(SH2_struct* sh)
{
  SH2movli(sh, 11, (0<<4)+6);
}
#define init_13_11_0_7
void FASTCALL call_13_11_0_7(SH2_struct* sh)
{
  SH2movli(sh, 11, (0<<4)+7);
}
#define init_13_11_0_8
void FASTCALL call_13_11_0_8(SH2_struct* sh)
{
  SH2movli(sh, 11, (0<<4)+8);
}
#define init_13_11_0_9
void FASTCALL call_13_11_0_9(SH2_struct* sh)
{
  SH2movli(sh, 11, (0<<4)+9);
}
#define init_13_11_0_10
void FASTCALL call_13_11_0_10(SH2_struct* sh)
{
  SH2movli(sh, 11, (0<<4)+10);
}
#define init_13_11_0_11
void FASTCALL call_13_11_0_11(SH2_struct* sh)
{
  SH2movli(sh, 11, (0<<4)+11);
}
#define init_13_11_0_12
void FASTCALL call_13_11_0_12(SH2_struct* sh)
{
  SH2movli(sh, 11, (0<<4)+12);
}
#define init_13_11_0_13
void FASTCALL call_13_11_0_13(SH2_struct* sh)
{
  SH2movli(sh, 11, (0<<4)+13);
}
#define init_13_11_0_14
void FASTCALL call_13_11_0_14(SH2_struct* sh)
{
  SH2movli(sh, 11, (0<<4)+14);
}
#define init_13_11_0_15
void FASTCALL call_13_11_0_15(SH2_struct* sh)
{
  SH2movli(sh, 11, (0<<4)+15);
}
#define init_13_11_1_0
void FASTCALL call_13_11_1_0(SH2_struct* sh)
{
  SH2movli(sh, 11, (1<<4)+0);
}
#define init_13_11_1_1
void FASTCALL call_13_11_1_1(SH2_struct* sh)
{
  SH2movli(sh, 11, (1<<4)+1);
}
#define init_13_11_1_2
void FASTCALL call_13_11_1_2(SH2_struct* sh)
{
  SH2movli(sh, 11, (1<<4)+2);
}
#define init_13_11_1_3
void FASTCALL call_13_11_1_3(SH2_struct* sh)
{
  SH2movli(sh, 11, (1<<4)+3);
}
#define init_13_11_1_4
void FASTCALL call_13_11_1_4(SH2_struct* sh)
{
  SH2movli(sh, 11, (1<<4)+4);
}
#define init_13_11_1_5
void FASTCALL call_13_11_1_5(SH2_struct* sh)
{
  SH2movli(sh, 11, (1<<4)+5);
}
#define init_13_11_1_6
void FASTCALL call_13_11_1_6(SH2_struct* sh)
{
  SH2movli(sh, 11, (1<<4)+6);
}
#define init_13_11_1_7
void FASTCALL call_13_11_1_7(SH2_struct* sh)
{
  SH2movli(sh, 11, (1<<4)+7);
}
#define init_13_11_1_8
void FASTCALL call_13_11_1_8(SH2_struct* sh)
{
  SH2movli(sh, 11, (1<<4)+8);
}
#define init_13_11_1_9
void FASTCALL call_13_11_1_9(SH2_struct* sh)
{
  SH2movli(sh, 11, (1<<4)+9);
}
#define init_13_11_1_10
void FASTCALL call_13_11_1_10(SH2_struct* sh)
{
  SH2movli(sh, 11, (1<<4)+10);
}
#define init_13_11_1_11
void FASTCALL call_13_11_1_11(SH2_struct* sh)
{
  SH2movli(sh, 11, (1<<4)+11);
}
#define init_13_11_1_12
void FASTCALL call_13_11_1_12(SH2_struct* sh)
{
  SH2movli(sh, 11, (1<<4)+12);
}
#define init_13_11_1_13
void FASTCALL call_13_11_1_13(SH2_struct* sh)
{
  SH2movli(sh, 11, (1<<4)+13);
}
#define init_13_11_1_14
void FASTCALL call_13_11_1_14(SH2_struct* sh)
{
  SH2movli(sh, 11, (1<<4)+14);
}
#define init_13_11_1_15
void FASTCALL call_13_11_1_15(SH2_struct* sh)
{
  SH2movli(sh, 11, (1<<4)+15);
}
#define init_13_11_2_0
void FASTCALL call_13_11_2_0(SH2_struct* sh)
{
  SH2movli(sh, 11, (2<<4)+0);
}
#define init_13_11_2_1
void FASTCALL call_13_11_2_1(SH2_struct* sh)
{
  SH2movli(sh, 11, (2<<4)+1);
}
#define init_13_11_2_2
void FASTCALL call_13_11_2_2(SH2_struct* sh)
{
  SH2movli(sh, 11, (2<<4)+2);
}
#define init_13_11_2_3
void FASTCALL call_13_11_2_3(SH2_struct* sh)
{
  SH2movli(sh, 11, (2<<4)+3);
}
#define init_13_11_2_4
void FASTCALL call_13_11_2_4(SH2_struct* sh)
{
  SH2movli(sh, 11, (2<<4)+4);
}
#define init_13_11_2_5
void FASTCALL call_13_11_2_5(SH2_struct* sh)
{
  SH2movli(sh, 11, (2<<4)+5);
}
#define init_13_11_2_6
void FASTCALL call_13_11_2_6(SH2_struct* sh)
{
  SH2movli(sh, 11, (2<<4)+6);
}
#define init_13_11_2_7
void FASTCALL call_13_11_2_7(SH2_struct* sh)
{
  SH2movli(sh, 11, (2<<4)+7);
}
#define init_13_11_2_8
void FASTCALL call_13_11_2_8(SH2_struct* sh)
{
  SH2movli(sh, 11, (2<<4)+8);
}
#define init_13_11_2_9
void FASTCALL call_13_11_2_9(SH2_struct* sh)
{
  SH2movli(sh, 11, (2<<4)+9);
}
#define init_13_11_2_10
void FASTCALL call_13_11_2_10(SH2_struct* sh)
{
  SH2movli(sh, 11, (2<<4)+10);
}
#define init_13_11_2_11
void FASTCALL call_13_11_2_11(SH2_struct* sh)
{
  SH2movli(sh, 11, (2<<4)+11);
}
#define init_13_11_2_12
void FASTCALL call_13_11_2_12(SH2_struct* sh)
{
  SH2movli(sh, 11, (2<<4)+12);
}
#define init_13_11_2_13
void FASTCALL call_13_11_2_13(SH2_struct* sh)
{
  SH2movli(sh, 11, (2<<4)+13);
}
#define init_13_11_2_14
void FASTCALL call_13_11_2_14(SH2_struct* sh)
{
  SH2movli(sh, 11, (2<<4)+14);
}
#define init_13_11_2_15
void FASTCALL call_13_11_2_15(SH2_struct* sh)
{
  SH2movli(sh, 11, (2<<4)+15);
}
#define init_13_11_3_0
void FASTCALL call_13_11_3_0(SH2_struct* sh)
{
  SH2movli(sh, 11, (3<<4)+0);
}
#define init_13_11_3_1
void FASTCALL call_13_11_3_1(SH2_struct* sh)
{
  SH2movli(sh, 11, (3<<4)+1);
}
#define init_13_11_3_2
void FASTCALL call_13_11_3_2(SH2_struct* sh)
{
  SH2movli(sh, 11, (3<<4)+2);
}
#define init_13_11_3_3
void FASTCALL call_13_11_3_3(SH2_struct* sh)
{
  SH2movli(sh, 11, (3<<4)+3);
}
#define init_13_11_3_4
void FASTCALL call_13_11_3_4(SH2_struct* sh)
{
  SH2movli(sh, 11, (3<<4)+4);
}
#define init_13_11_3_5
void FASTCALL call_13_11_3_5(SH2_struct* sh)
{
  SH2movli(sh, 11, (3<<4)+5);
}
#define init_13_11_3_6
void FASTCALL call_13_11_3_6(SH2_struct* sh)
{
  SH2movli(sh, 11, (3<<4)+6);
}
#define init_13_11_3_7
void FASTCALL call_13_11_3_7(SH2_struct* sh)
{
  SH2movli(sh, 11, (3<<4)+7);
}
#define init_13_11_3_8
void FASTCALL call_13_11_3_8(SH2_struct* sh)
{
  SH2movli(sh, 11, (3<<4)+8);
}
#define init_13_11_3_9
void FASTCALL call_13_11_3_9(SH2_struct* sh)
{
  SH2movli(sh, 11, (3<<4)+9);
}
#define init_13_11_3_10
void FASTCALL call_13_11_3_10(SH2_struct* sh)
{
  SH2movli(sh, 11, (3<<4)+10);
}
#define init_13_11_3_11
void FASTCALL call_13_11_3_11(SH2_struct* sh)
{
  SH2movli(sh, 11, (3<<4)+11);
}
#define init_13_11_3_12
void FASTCALL call_13_11_3_12(SH2_struct* sh)
{
  SH2movli(sh, 11, (3<<4)+12);
}
#define init_13_11_3_13
void FASTCALL call_13_11_3_13(SH2_struct* sh)
{
  SH2movli(sh, 11, (3<<4)+13);
}
#define init_13_11_3_14
void FASTCALL call_13_11_3_14(SH2_struct* sh)
{
  SH2movli(sh, 11, (3<<4)+14);
}
#define init_13_11_3_15
void FASTCALL call_13_11_3_15(SH2_struct* sh)
{
  SH2movli(sh, 11, (3<<4)+15);
}
#define init_13_11_4_0
void FASTCALL call_13_11_4_0(SH2_struct* sh)
{
  SH2movli(sh, 11, (4<<4)+0);
}
#define init_13_11_4_1
void FASTCALL call_13_11_4_1(SH2_struct* sh)
{
  SH2movli(sh, 11, (4<<4)+1);
}
#define init_13_11_4_2
void FASTCALL call_13_11_4_2(SH2_struct* sh)
{
  SH2movli(sh, 11, (4<<4)+2);
}
#define init_13_11_4_3
void FASTCALL call_13_11_4_3(SH2_struct* sh)
{
  SH2movli(sh, 11, (4<<4)+3);
}
#define init_13_11_4_4
void FASTCALL call_13_11_4_4(SH2_struct* sh)
{
  SH2movli(sh, 11, (4<<4)+4);
}
#define init_13_11_4_5
void FASTCALL call_13_11_4_5(SH2_struct* sh)
{
  SH2movli(sh, 11, (4<<4)+5);
}
#define init_13_11_4_6
void FASTCALL call_13_11_4_6(SH2_struct* sh)
{
  SH2movli(sh, 11, (4<<4)+6);
}
#define init_13_11_4_7
void FASTCALL call_13_11_4_7(SH2_struct* sh)
{
  SH2movli(sh, 11, (4<<4)+7);
}
#define init_13_11_4_8
void FASTCALL call_13_11_4_8(SH2_struct* sh)
{
  SH2movli(sh, 11, (4<<4)+8);
}
#define init_13_11_4_9
void FASTCALL call_13_11_4_9(SH2_struct* sh)
{
  SH2movli(sh, 11, (4<<4)+9);
}
#define init_13_11_4_10
void FASTCALL call_13_11_4_10(SH2_struct* sh)
{
  SH2movli(sh, 11, (4<<4)+10);
}
#define init_13_11_4_11
void FASTCALL call_13_11_4_11(SH2_struct* sh)
{
  SH2movli(sh, 11, (4<<4)+11);
}
#define init_13_11_4_12
void FASTCALL call_13_11_4_12(SH2_struct* sh)
{
  SH2movli(sh, 11, (4<<4)+12);
}
#define init_13_11_4_13
void FASTCALL call_13_11_4_13(SH2_struct* sh)
{
  SH2movli(sh, 11, (4<<4)+13);
}
#define init_13_11_4_14
void FASTCALL call_13_11_4_14(SH2_struct* sh)
{
  SH2movli(sh, 11, (4<<4)+14);
}
#define init_13_11_4_15
void FASTCALL call_13_11_4_15(SH2_struct* sh)
{
  SH2movli(sh, 11, (4<<4)+15);
}
#define init_13_11_5_0
void FASTCALL call_13_11_5_0(SH2_struct* sh)
{
  SH2movli(sh, 11, (5<<4)+0);
}
#define init_13_11_5_1
void FASTCALL call_13_11_5_1(SH2_struct* sh)
{
  SH2movli(sh, 11, (5<<4)+1);
}
#define init_13_11_5_2
void FASTCALL call_13_11_5_2(SH2_struct* sh)
{
  SH2movli(sh, 11, (5<<4)+2);
}
#define init_13_11_5_3
void FASTCALL call_13_11_5_3(SH2_struct* sh)
{
  SH2movli(sh, 11, (5<<4)+3);
}
#define init_13_11_5_4
void FASTCALL call_13_11_5_4(SH2_struct* sh)
{
  SH2movli(sh, 11, (5<<4)+4);
}
#define init_13_11_5_5
void FASTCALL call_13_11_5_5(SH2_struct* sh)
{
  SH2movli(sh, 11, (5<<4)+5);
}
#define init_13_11_5_6
void FASTCALL call_13_11_5_6(SH2_struct* sh)
{
  SH2movli(sh, 11, (5<<4)+6);
}
#define init_13_11_5_7
void FASTCALL call_13_11_5_7(SH2_struct* sh)
{
  SH2movli(sh, 11, (5<<4)+7);
}
#define init_13_11_5_8
void FASTCALL call_13_11_5_8(SH2_struct* sh)
{
  SH2movli(sh, 11, (5<<4)+8);
}
#define init_13_11_5_9
void FASTCALL call_13_11_5_9(SH2_struct* sh)
{
  SH2movli(sh, 11, (5<<4)+9);
}
#define init_13_11_5_10
void FASTCALL call_13_11_5_10(SH2_struct* sh)
{
  SH2movli(sh, 11, (5<<4)+10);
}
#define init_13_11_5_11
void FASTCALL call_13_11_5_11(SH2_struct* sh)
{
  SH2movli(sh, 11, (5<<4)+11);
}
#define init_13_11_5_12
void FASTCALL call_13_11_5_12(SH2_struct* sh)
{
  SH2movli(sh, 11, (5<<4)+12);
}
#define init_13_11_5_13
void FASTCALL call_13_11_5_13(SH2_struct* sh)
{
  SH2movli(sh, 11, (5<<4)+13);
}
#define init_13_11_5_14
void FASTCALL call_13_11_5_14(SH2_struct* sh)
{
  SH2movli(sh, 11, (5<<4)+14);
}
#define init_13_11_5_15
void FASTCALL call_13_11_5_15(SH2_struct* sh)
{
  SH2movli(sh, 11, (5<<4)+15);
}
#define init_13_11_6_0
void FASTCALL call_13_11_6_0(SH2_struct* sh)
{
  SH2movli(sh, 11, (6<<4)+0);
}
#define init_13_11_6_1
void FASTCALL call_13_11_6_1(SH2_struct* sh)
{
  SH2movli(sh, 11, (6<<4)+1);
}
#define init_13_11_6_2
void FASTCALL call_13_11_6_2(SH2_struct* sh)
{
  SH2movli(sh, 11, (6<<4)+2);
}
#define init_13_11_6_3
void FASTCALL call_13_11_6_3(SH2_struct* sh)
{
  SH2movli(sh, 11, (6<<4)+3);
}
#define init_13_11_6_4
void FASTCALL call_13_11_6_4(SH2_struct* sh)
{
  SH2movli(sh, 11, (6<<4)+4);
}
#define init_13_11_6_5
void FASTCALL call_13_11_6_5(SH2_struct* sh)
{
  SH2movli(sh, 11, (6<<4)+5);
}
#define init_13_11_6_6
void FASTCALL call_13_11_6_6(SH2_struct* sh)
{
  SH2movli(sh, 11, (6<<4)+6);
}
#define init_13_11_6_7
void FASTCALL call_13_11_6_7(SH2_struct* sh)
{
  SH2movli(sh, 11, (6<<4)+7);
}
#define init_13_11_6_8
void FASTCALL call_13_11_6_8(SH2_struct* sh)
{
  SH2movli(sh, 11, (6<<4)+8);
}
#define init_13_11_6_9
void FASTCALL call_13_11_6_9(SH2_struct* sh)
{
  SH2movli(sh, 11, (6<<4)+9);
}
#define init_13_11_6_10
void FASTCALL call_13_11_6_10(SH2_struct* sh)
{
  SH2movli(sh, 11, (6<<4)+10);
}
#define init_13_11_6_11
void FASTCALL call_13_11_6_11(SH2_struct* sh)
{
  SH2movli(sh, 11, (6<<4)+11);
}
#define init_13_11_6_12
void FASTCALL call_13_11_6_12(SH2_struct* sh)
{
  SH2movli(sh, 11, (6<<4)+12);
}
#define init_13_11_6_13
void FASTCALL call_13_11_6_13(SH2_struct* sh)
{
  SH2movli(sh, 11, (6<<4)+13);
}
#define init_13_11_6_14
void FASTCALL call_13_11_6_14(SH2_struct* sh)
{
  SH2movli(sh, 11, (6<<4)+14);
}
#define init_13_11_6_15
void FASTCALL call_13_11_6_15(SH2_struct* sh)
{
  SH2movli(sh, 11, (6<<4)+15);
}
#define init_13_11_7_0
void FASTCALL call_13_11_7_0(SH2_struct* sh)
{
  SH2movli(sh, 11, (7<<4)+0);
}
#define init_13_11_7_1
void FASTCALL call_13_11_7_1(SH2_struct* sh)
{
  SH2movli(sh, 11, (7<<4)+1);
}
#define init_13_11_7_2
void FASTCALL call_13_11_7_2(SH2_struct* sh)
{
  SH2movli(sh, 11, (7<<4)+2);
}
#define init_13_11_7_3
void FASTCALL call_13_11_7_3(SH2_struct* sh)
{
  SH2movli(sh, 11, (7<<4)+3);
}
#define init_13_11_7_4
void FASTCALL call_13_11_7_4(SH2_struct* sh)
{
  SH2movli(sh, 11, (7<<4)+4);
}
#define init_13_11_7_5
void FASTCALL call_13_11_7_5(SH2_struct* sh)
{
  SH2movli(sh, 11, (7<<4)+5);
}
#define init_13_11_7_6
void FASTCALL call_13_11_7_6(SH2_struct* sh)
{
  SH2movli(sh, 11, (7<<4)+6);
}
#define init_13_11_7_7
void FASTCALL call_13_11_7_7(SH2_struct* sh)
{
  SH2movli(sh, 11, (7<<4)+7);
}
#define init_13_11_7_8
void FASTCALL call_13_11_7_8(SH2_struct* sh)
{
  SH2movli(sh, 11, (7<<4)+8);
}
#define init_13_11_7_9
void FASTCALL call_13_11_7_9(SH2_struct* sh)
{
  SH2movli(sh, 11, (7<<4)+9);
}
#define init_13_11_7_10
void FASTCALL call_13_11_7_10(SH2_struct* sh)
{
  SH2movli(sh, 11, (7<<4)+10);
}
#define init_13_11_7_11
void FASTCALL call_13_11_7_11(SH2_struct* sh)
{
  SH2movli(sh, 11, (7<<4)+11);
}
#define init_13_11_7_12
void FASTCALL call_13_11_7_12(SH2_struct* sh)
{
  SH2movli(sh, 11, (7<<4)+12);
}
#define init_13_11_7_13
void FASTCALL call_13_11_7_13(SH2_struct* sh)
{
  SH2movli(sh, 11, (7<<4)+13);
}
#define init_13_11_7_14
void FASTCALL call_13_11_7_14(SH2_struct* sh)
{
  SH2movli(sh, 11, (7<<4)+14);
}
#define init_13_11_7_15
void FASTCALL call_13_11_7_15(SH2_struct* sh)
{
  SH2movli(sh, 11, (7<<4)+15);
}
#define init_13_11_8_0
void FASTCALL call_13_11_8_0(SH2_struct* sh)
{
  SH2movli(sh, 11, (8<<4)+0);
}
#define init_13_11_8_1
void FASTCALL call_13_11_8_1(SH2_struct* sh)
{
  SH2movli(sh, 11, (8<<4)+1);
}
#define init_13_11_8_2
void FASTCALL call_13_11_8_2(SH2_struct* sh)
{
  SH2movli(sh, 11, (8<<4)+2);
}
#define init_13_11_8_3
void FASTCALL call_13_11_8_3(SH2_struct* sh)
{
  SH2movli(sh, 11, (8<<4)+3);
}
#define init_13_11_8_4
void FASTCALL call_13_11_8_4(SH2_struct* sh)
{
  SH2movli(sh, 11, (8<<4)+4);
}
#define init_13_11_8_5
void FASTCALL call_13_11_8_5(SH2_struct* sh)
{
  SH2movli(sh, 11, (8<<4)+5);
}
#define init_13_11_8_6
void FASTCALL call_13_11_8_6(SH2_struct* sh)
{
  SH2movli(sh, 11, (8<<4)+6);
}
#define init_13_11_8_7
void FASTCALL call_13_11_8_7(SH2_struct* sh)
{
  SH2movli(sh, 11, (8<<4)+7);
}
#define init_13_11_8_8
void FASTCALL call_13_11_8_8(SH2_struct* sh)
{
  SH2movli(sh, 11, (8<<4)+8);
}
#define init_13_11_8_9
void FASTCALL call_13_11_8_9(SH2_struct* sh)
{
  SH2movli(sh, 11, (8<<4)+9);
}
#define init_13_11_8_10
void FASTCALL call_13_11_8_10(SH2_struct* sh)
{
  SH2movli(sh, 11, (8<<4)+10);
}
#define init_13_11_8_11
void FASTCALL call_13_11_8_11(SH2_struct* sh)
{
  SH2movli(sh, 11, (8<<4)+11);
}
#define init_13_11_8_12
void FASTCALL call_13_11_8_12(SH2_struct* sh)
{
  SH2movli(sh, 11, (8<<4)+12);
}
#define init_13_11_8_13
void FASTCALL call_13_11_8_13(SH2_struct* sh)
{
  SH2movli(sh, 11, (8<<4)+13);
}
#define init_13_11_8_14
void FASTCALL call_13_11_8_14(SH2_struct* sh)
{
  SH2movli(sh, 11, (8<<4)+14);
}
#define init_13_11_8_15
void FASTCALL call_13_11_8_15(SH2_struct* sh)
{
  SH2movli(sh, 11, (8<<4)+15);
}
#define init_13_11_9_0
void FASTCALL call_13_11_9_0(SH2_struct* sh)
{
  SH2movli(sh, 11, (9<<4)+0);
}
#define init_13_11_9_1
void FASTCALL call_13_11_9_1(SH2_struct* sh)
{
  SH2movli(sh, 11, (9<<4)+1);
}
#define init_13_11_9_2
void FASTCALL call_13_11_9_2(SH2_struct* sh)
{
  SH2movli(sh, 11, (9<<4)+2);
}
#define init_13_11_9_3
void FASTCALL call_13_11_9_3(SH2_struct* sh)
{
  SH2movli(sh, 11, (9<<4)+3);
}
#define init_13_11_9_4
void FASTCALL call_13_11_9_4(SH2_struct* sh)
{
  SH2movli(sh, 11, (9<<4)+4);
}
#define init_13_11_9_5
void FASTCALL call_13_11_9_5(SH2_struct* sh)
{
  SH2movli(sh, 11, (9<<4)+5);
}
#define init_13_11_9_6
void FASTCALL call_13_11_9_6(SH2_struct* sh)
{
  SH2movli(sh, 11, (9<<4)+6);
}
#define init_13_11_9_7
void FASTCALL call_13_11_9_7(SH2_struct* sh)
{
  SH2movli(sh, 11, (9<<4)+7);
}
#define init_13_11_9_8
void FASTCALL call_13_11_9_8(SH2_struct* sh)
{
  SH2movli(sh, 11, (9<<4)+8);
}
#define init_13_11_9_9
void FASTCALL call_13_11_9_9(SH2_struct* sh)
{
  SH2movli(sh, 11, (9<<4)+9);
}
#define init_13_11_9_10
void FASTCALL call_13_11_9_10(SH2_struct* sh)
{
  SH2movli(sh, 11, (9<<4)+10);
}
#define init_13_11_9_11
void FASTCALL call_13_11_9_11(SH2_struct* sh)
{
  SH2movli(sh, 11, (9<<4)+11);
}
#define init_13_11_9_12
void FASTCALL call_13_11_9_12(SH2_struct* sh)
{
  SH2movli(sh, 11, (9<<4)+12);
}
#define init_13_11_9_13
void FASTCALL call_13_11_9_13(SH2_struct* sh)
{
  SH2movli(sh, 11, (9<<4)+13);
}
#define init_13_11_9_14
void FASTCALL call_13_11_9_14(SH2_struct* sh)
{
  SH2movli(sh, 11, (9<<4)+14);
}
#define init_13_11_9_15
void FASTCALL call_13_11_9_15(SH2_struct* sh)
{
  SH2movli(sh, 11, (9<<4)+15);
}
#define init_13_11_10_0
void FASTCALL call_13_11_10_0(SH2_struct* sh)
{
  SH2movli(sh, 11, (10<<4)+0);
}
#define init_13_11_10_1
void FASTCALL call_13_11_10_1(SH2_struct* sh)
{
  SH2movli(sh, 11, (10<<4)+1);
}
#define init_13_11_10_2
void FASTCALL call_13_11_10_2(SH2_struct* sh)
{
  SH2movli(sh, 11, (10<<4)+2);
}
#define init_13_11_10_3
void FASTCALL call_13_11_10_3(SH2_struct* sh)
{
  SH2movli(sh, 11, (10<<4)+3);
}
#define init_13_11_10_4
void FASTCALL call_13_11_10_4(SH2_struct* sh)
{
  SH2movli(sh, 11, (10<<4)+4);
}
#define init_13_11_10_5
void FASTCALL call_13_11_10_5(SH2_struct* sh)
{
  SH2movli(sh, 11, (10<<4)+5);
}
#define init_13_11_10_6
void FASTCALL call_13_11_10_6(SH2_struct* sh)
{
  SH2movli(sh, 11, (10<<4)+6);
}
#define init_13_11_10_7
void FASTCALL call_13_11_10_7(SH2_struct* sh)
{
  SH2movli(sh, 11, (10<<4)+7);
}
#define init_13_11_10_8
void FASTCALL call_13_11_10_8(SH2_struct* sh)
{
  SH2movli(sh, 11, (10<<4)+8);
}
#define init_13_11_10_9
void FASTCALL call_13_11_10_9(SH2_struct* sh)
{
  SH2movli(sh, 11, (10<<4)+9);
}
#define init_13_11_10_10
void FASTCALL call_13_11_10_10(SH2_struct* sh)
{
  SH2movli(sh, 11, (10<<4)+10);
}
#define init_13_11_10_11
void FASTCALL call_13_11_10_11(SH2_struct* sh)
{
  SH2movli(sh, 11, (10<<4)+11);
}
#define init_13_11_10_12
void FASTCALL call_13_11_10_12(SH2_struct* sh)
{
  SH2movli(sh, 11, (10<<4)+12);
}
#define init_13_11_10_13
void FASTCALL call_13_11_10_13(SH2_struct* sh)
{
  SH2movli(sh, 11, (10<<4)+13);
}
#define init_13_11_10_14
void FASTCALL call_13_11_10_14(SH2_struct* sh)
{
  SH2movli(sh, 11, (10<<4)+14);
}
#define init_13_11_10_15
void FASTCALL call_13_11_10_15(SH2_struct* sh)
{
  SH2movli(sh, 11, (10<<4)+15);
}
#define init_13_11_11_0
void FASTCALL call_13_11_11_0(SH2_struct* sh)
{
  SH2movli(sh, 11, (11<<4)+0);
}
#define init_13_11_11_1
void FASTCALL call_13_11_11_1(SH2_struct* sh)
{
  SH2movli(sh, 11, (11<<4)+1);
}
#define init_13_11_11_2
void FASTCALL call_13_11_11_2(SH2_struct* sh)
{
  SH2movli(sh, 11, (11<<4)+2);
}
#define init_13_11_11_3
void FASTCALL call_13_11_11_3(SH2_struct* sh)
{
  SH2movli(sh, 11, (11<<4)+3);
}
#define init_13_11_11_4
void FASTCALL call_13_11_11_4(SH2_struct* sh)
{
  SH2movli(sh, 11, (11<<4)+4);
}
#define init_13_11_11_5
void FASTCALL call_13_11_11_5(SH2_struct* sh)
{
  SH2movli(sh, 11, (11<<4)+5);
}
#define init_13_11_11_6
void FASTCALL call_13_11_11_6(SH2_struct* sh)
{
  SH2movli(sh, 11, (11<<4)+6);
}
#define init_13_11_11_7
void FASTCALL call_13_11_11_7(SH2_struct* sh)
{
  SH2movli(sh, 11, (11<<4)+7);
}
#define init_13_11_11_8
void FASTCALL call_13_11_11_8(SH2_struct* sh)
{
  SH2movli(sh, 11, (11<<4)+8);
}
#define init_13_11_11_9
void FASTCALL call_13_11_11_9(SH2_struct* sh)
{
  SH2movli(sh, 11, (11<<4)+9);
}
#define init_13_11_11_10
void FASTCALL call_13_11_11_10(SH2_struct* sh)
{
  SH2movli(sh, 11, (11<<4)+10);
}
#define init_13_11_11_11
void FASTCALL call_13_11_11_11(SH2_struct* sh)
{
  SH2movli(sh, 11, (11<<4)+11);
}
#define init_13_11_11_12
void FASTCALL call_13_11_11_12(SH2_struct* sh)
{
  SH2movli(sh, 11, (11<<4)+12);
}
#define init_13_11_11_13
void FASTCALL call_13_11_11_13(SH2_struct* sh)
{
  SH2movli(sh, 11, (11<<4)+13);
}
#define init_13_11_11_14
void FASTCALL call_13_11_11_14(SH2_struct* sh)
{
  SH2movli(sh, 11, (11<<4)+14);
}
#define init_13_11_11_15
void FASTCALL call_13_11_11_15(SH2_struct* sh)
{
  SH2movli(sh, 11, (11<<4)+15);
}
#define init_13_11_12_0
void FASTCALL call_13_11_12_0(SH2_struct* sh)
{
  SH2movli(sh, 11, (12<<4)+0);
}
#define init_13_11_12_1
void FASTCALL call_13_11_12_1(SH2_struct* sh)
{
  SH2movli(sh, 11, (12<<4)+1);
}
#define init_13_11_12_2
void FASTCALL call_13_11_12_2(SH2_struct* sh)
{
  SH2movli(sh, 11, (12<<4)+2);
}
#define init_13_11_12_3
void FASTCALL call_13_11_12_3(SH2_struct* sh)
{
  SH2movli(sh, 11, (12<<4)+3);
}
#define init_13_11_12_4
void FASTCALL call_13_11_12_4(SH2_struct* sh)
{
  SH2movli(sh, 11, (12<<4)+4);
}
#define init_13_11_12_5
void FASTCALL call_13_11_12_5(SH2_struct* sh)
{
  SH2movli(sh, 11, (12<<4)+5);
}
#define init_13_11_12_6
void FASTCALL call_13_11_12_6(SH2_struct* sh)
{
  SH2movli(sh, 11, (12<<4)+6);
}
#define init_13_11_12_7
void FASTCALL call_13_11_12_7(SH2_struct* sh)
{
  SH2movli(sh, 11, (12<<4)+7);
}
#define init_13_11_12_8
void FASTCALL call_13_11_12_8(SH2_struct* sh)
{
  SH2movli(sh, 11, (12<<4)+8);
}
#define init_13_11_12_9
void FASTCALL call_13_11_12_9(SH2_struct* sh)
{
  SH2movli(sh, 11, (12<<4)+9);
}
#define init_13_11_12_10
void FASTCALL call_13_11_12_10(SH2_struct* sh)
{
  SH2movli(sh, 11, (12<<4)+10);
}
#define init_13_11_12_11
void FASTCALL call_13_11_12_11(SH2_struct* sh)
{
  SH2movli(sh, 11, (12<<4)+11);
}
#define init_13_11_12_12
void FASTCALL call_13_11_12_12(SH2_struct* sh)
{
  SH2movli(sh, 11, (12<<4)+12);
}
#define init_13_11_12_13
void FASTCALL call_13_11_12_13(SH2_struct* sh)
{
  SH2movli(sh, 11, (12<<4)+13);
}
#define init_13_11_12_14
void FASTCALL call_13_11_12_14(SH2_struct* sh)
{
  SH2movli(sh, 11, (12<<4)+14);
}
#define init_13_11_12_15
void FASTCALL call_13_11_12_15(SH2_struct* sh)
{
  SH2movli(sh, 11, (12<<4)+15);
}
#define init_13_11_13_0
void FASTCALL call_13_11_13_0(SH2_struct* sh)
{
  SH2movli(sh, 11, (13<<4)+0);
}
#define init_13_11_13_1
void FASTCALL call_13_11_13_1(SH2_struct* sh)
{
  SH2movli(sh, 11, (13<<4)+1);
}
#define init_13_11_13_2
void FASTCALL call_13_11_13_2(SH2_struct* sh)
{
  SH2movli(sh, 11, (13<<4)+2);
}
#define init_13_11_13_3
void FASTCALL call_13_11_13_3(SH2_struct* sh)
{
  SH2movli(sh, 11, (13<<4)+3);
}
#define init_13_11_13_4
void FASTCALL call_13_11_13_4(SH2_struct* sh)
{
  SH2movli(sh, 11, (13<<4)+4);
}
#define init_13_11_13_5
void FASTCALL call_13_11_13_5(SH2_struct* sh)
{
  SH2movli(sh, 11, (13<<4)+5);
}
#define init_13_11_13_6
void FASTCALL call_13_11_13_6(SH2_struct* sh)
{
  SH2movli(sh, 11, (13<<4)+6);
}
#define init_13_11_13_7
void FASTCALL call_13_11_13_7(SH2_struct* sh)
{
  SH2movli(sh, 11, (13<<4)+7);
}
#define init_13_11_13_8
void FASTCALL call_13_11_13_8(SH2_struct* sh)
{
  SH2movli(sh, 11, (13<<4)+8);
}
#define init_13_11_13_9
void FASTCALL call_13_11_13_9(SH2_struct* sh)
{
  SH2movli(sh, 11, (13<<4)+9);
}
#define init_13_11_13_10
void FASTCALL call_13_11_13_10(SH2_struct* sh)
{
  SH2movli(sh, 11, (13<<4)+10);
}
#define init_13_11_13_11
void FASTCALL call_13_11_13_11(SH2_struct* sh)
{
  SH2movli(sh, 11, (13<<4)+11);
}
#define init_13_11_13_12
void FASTCALL call_13_11_13_12(SH2_struct* sh)
{
  SH2movli(sh, 11, (13<<4)+12);
}
#define init_13_11_13_13
void FASTCALL call_13_11_13_13(SH2_struct* sh)
{
  SH2movli(sh, 11, (13<<4)+13);
}
#define init_13_11_13_14
void FASTCALL call_13_11_13_14(SH2_struct* sh)
{
  SH2movli(sh, 11, (13<<4)+14);
}
#define init_13_11_13_15
void FASTCALL call_13_11_13_15(SH2_struct* sh)
{
  SH2movli(sh, 11, (13<<4)+15);
}
#define init_13_11_14_0
void FASTCALL call_13_11_14_0(SH2_struct* sh)
{
  SH2movli(sh, 11, (14<<4)+0);
}
#define init_13_11_14_1
void FASTCALL call_13_11_14_1(SH2_struct* sh)
{
  SH2movli(sh, 11, (14<<4)+1);
}
#define init_13_11_14_2
void FASTCALL call_13_11_14_2(SH2_struct* sh)
{
  SH2movli(sh, 11, (14<<4)+2);
}
#define init_13_11_14_3
void FASTCALL call_13_11_14_3(SH2_struct* sh)
{
  SH2movli(sh, 11, (14<<4)+3);
}
#define init_13_11_14_4
void FASTCALL call_13_11_14_4(SH2_struct* sh)
{
  SH2movli(sh, 11, (14<<4)+4);
}
#define init_13_11_14_5
void FASTCALL call_13_11_14_5(SH2_struct* sh)
{
  SH2movli(sh, 11, (14<<4)+5);
}
#define init_13_11_14_6
void FASTCALL call_13_11_14_6(SH2_struct* sh)
{
  SH2movli(sh, 11, (14<<4)+6);
}
#define init_13_11_14_7
void FASTCALL call_13_11_14_7(SH2_struct* sh)
{
  SH2movli(sh, 11, (14<<4)+7);
}
#define init_13_11_14_8
void FASTCALL call_13_11_14_8(SH2_struct* sh)
{
  SH2movli(sh, 11, (14<<4)+8);
}
#define init_13_11_14_9
void FASTCALL call_13_11_14_9(SH2_struct* sh)
{
  SH2movli(sh, 11, (14<<4)+9);
}
#define init_13_11_14_10
void FASTCALL call_13_11_14_10(SH2_struct* sh)
{
  SH2movli(sh, 11, (14<<4)+10);
}
#define init_13_11_14_11
void FASTCALL call_13_11_14_11(SH2_struct* sh)
{
  SH2movli(sh, 11, (14<<4)+11);
}
#define init_13_11_14_12
void FASTCALL call_13_11_14_12(SH2_struct* sh)
{
  SH2movli(sh, 11, (14<<4)+12);
}
#define init_13_11_14_13
void FASTCALL call_13_11_14_13(SH2_struct* sh)
{
  SH2movli(sh, 11, (14<<4)+13);
}
#define init_13_11_14_14
void FASTCALL call_13_11_14_14(SH2_struct* sh)
{
  SH2movli(sh, 11, (14<<4)+14);
}
#define init_13_11_14_15
void FASTCALL call_13_11_14_15(SH2_struct* sh)
{
  SH2movli(sh, 11, (14<<4)+15);
}
#define init_13_11_15_0
void FASTCALL call_13_11_15_0(SH2_struct* sh)
{
  SH2movli(sh, 11, (15<<4)+0);
}
#define init_13_11_15_1
void FASTCALL call_13_11_15_1(SH2_struct* sh)
{
  SH2movli(sh, 11, (15<<4)+1);
}
#define init_13_11_15_2
void FASTCALL call_13_11_15_2(SH2_struct* sh)
{
  SH2movli(sh, 11, (15<<4)+2);
}
#define init_13_11_15_3
void FASTCALL call_13_11_15_3(SH2_struct* sh)
{
  SH2movli(sh, 11, (15<<4)+3);
}
#define init_13_11_15_4
void FASTCALL call_13_11_15_4(SH2_struct* sh)
{
  SH2movli(sh, 11, (15<<4)+4);
}
#define init_13_11_15_5
void FASTCALL call_13_11_15_5(SH2_struct* sh)
{
  SH2movli(sh, 11, (15<<4)+5);
}
#define init_13_11_15_6
void FASTCALL call_13_11_15_6(SH2_struct* sh)
{
  SH2movli(sh, 11, (15<<4)+6);
}
#define init_13_11_15_7
void FASTCALL call_13_11_15_7(SH2_struct* sh)
{
  SH2movli(sh, 11, (15<<4)+7);
}
#define init_13_11_15_8
void FASTCALL call_13_11_15_8(SH2_struct* sh)
{
  SH2movli(sh, 11, (15<<4)+8);
}
#define init_13_11_15_9
void FASTCALL call_13_11_15_9(SH2_struct* sh)
{
  SH2movli(sh, 11, (15<<4)+9);
}
#define init_13_11_15_10
void FASTCALL call_13_11_15_10(SH2_struct* sh)
{
  SH2movli(sh, 11, (15<<4)+10);
}
#define init_13_11_15_11
void FASTCALL call_13_11_15_11(SH2_struct* sh)
{
  SH2movli(sh, 11, (15<<4)+11);
}
#define init_13_11_15_12
void FASTCALL call_13_11_15_12(SH2_struct* sh)
{
  SH2movli(sh, 11, (15<<4)+12);
}
#define init_13_11_15_13
void FASTCALL call_13_11_15_13(SH2_struct* sh)
{
  SH2movli(sh, 11, (15<<4)+13);
}
#define init_13_11_15_14
void FASTCALL call_13_11_15_14(SH2_struct* sh)
{
  SH2movli(sh, 11, (15<<4)+14);
}
#define init_13_11_15_15
void FASTCALL call_13_11_15_15(SH2_struct* sh)
{
  SH2movli(sh, 11, (15<<4)+15);
}
#define init_13_12_0_0
void FASTCALL call_13_12_0_0(SH2_struct* sh)
{
  SH2movli(sh, 12, (0<<4)+0);
}
#define init_13_12_0_1
void FASTCALL call_13_12_0_1(SH2_struct* sh)
{
  SH2movli(sh, 12, (0<<4)+1);
}
#define init_13_12_0_2
void FASTCALL call_13_12_0_2(SH2_struct* sh)
{
  SH2movli(sh, 12, (0<<4)+2);
}
#define init_13_12_0_3
void FASTCALL call_13_12_0_3(SH2_struct* sh)
{
  SH2movli(sh, 12, (0<<4)+3);
}
#define init_13_12_0_4
void FASTCALL call_13_12_0_4(SH2_struct* sh)
{
  SH2movli(sh, 12, (0<<4)+4);
}
#define init_13_12_0_5
void FASTCALL call_13_12_0_5(SH2_struct* sh)
{
  SH2movli(sh, 12, (0<<4)+5);
}
#define init_13_12_0_6
void FASTCALL call_13_12_0_6(SH2_struct* sh)
{
  SH2movli(sh, 12, (0<<4)+6);
}
#define init_13_12_0_7
void FASTCALL call_13_12_0_7(SH2_struct* sh)
{
  SH2movli(sh, 12, (0<<4)+7);
}
#define init_13_12_0_8
void FASTCALL call_13_12_0_8(SH2_struct* sh)
{
  SH2movli(sh, 12, (0<<4)+8);
}
#define init_13_12_0_9
void FASTCALL call_13_12_0_9(SH2_struct* sh)
{
  SH2movli(sh, 12, (0<<4)+9);
}
#define init_13_12_0_10
void FASTCALL call_13_12_0_10(SH2_struct* sh)
{
  SH2movli(sh, 12, (0<<4)+10);
}
#define init_13_12_0_11
void FASTCALL call_13_12_0_11(SH2_struct* sh)
{
  SH2movli(sh, 12, (0<<4)+11);
}
#define init_13_12_0_12
void FASTCALL call_13_12_0_12(SH2_struct* sh)
{
  SH2movli(sh, 12, (0<<4)+12);
}
#define init_13_12_0_13
void FASTCALL call_13_12_0_13(SH2_struct* sh)
{
  SH2movli(sh, 12, (0<<4)+13);
}
#define init_13_12_0_14
void FASTCALL call_13_12_0_14(SH2_struct* sh)
{
  SH2movli(sh, 12, (0<<4)+14);
}
#define init_13_12_0_15
void FASTCALL call_13_12_0_15(SH2_struct* sh)
{
  SH2movli(sh, 12, (0<<4)+15);
}
#define init_13_12_1_0
void FASTCALL call_13_12_1_0(SH2_struct* sh)
{
  SH2movli(sh, 12, (1<<4)+0);
}
#define init_13_12_1_1
void FASTCALL call_13_12_1_1(SH2_struct* sh)
{
  SH2movli(sh, 12, (1<<4)+1);
}
#define init_13_12_1_2
void FASTCALL call_13_12_1_2(SH2_struct* sh)
{
  SH2movli(sh, 12, (1<<4)+2);
}
#define init_13_12_1_3
void FASTCALL call_13_12_1_3(SH2_struct* sh)
{
  SH2movli(sh, 12, (1<<4)+3);
}
#define init_13_12_1_4
void FASTCALL call_13_12_1_4(SH2_struct* sh)
{
  SH2movli(sh, 12, (1<<4)+4);
}
#define init_13_12_1_5
void FASTCALL call_13_12_1_5(SH2_struct* sh)
{
  SH2movli(sh, 12, (1<<4)+5);
}
#define init_13_12_1_6
void FASTCALL call_13_12_1_6(SH2_struct* sh)
{
  SH2movli(sh, 12, (1<<4)+6);
}
#define init_13_12_1_7
void FASTCALL call_13_12_1_7(SH2_struct* sh)
{
  SH2movli(sh, 12, (1<<4)+7);
}
#define init_13_12_1_8
void FASTCALL call_13_12_1_8(SH2_struct* sh)
{
  SH2movli(sh, 12, (1<<4)+8);
}
#define init_13_12_1_9
void FASTCALL call_13_12_1_9(SH2_struct* sh)
{
  SH2movli(sh, 12, (1<<4)+9);
}
#define init_13_12_1_10
void FASTCALL call_13_12_1_10(SH2_struct* sh)
{
  SH2movli(sh, 12, (1<<4)+10);
}
#define init_13_12_1_11
void FASTCALL call_13_12_1_11(SH2_struct* sh)
{
  SH2movli(sh, 12, (1<<4)+11);
}
#define init_13_12_1_12
void FASTCALL call_13_12_1_12(SH2_struct* sh)
{
  SH2movli(sh, 12, (1<<4)+12);
}
#define init_13_12_1_13
void FASTCALL call_13_12_1_13(SH2_struct* sh)
{
  SH2movli(sh, 12, (1<<4)+13);
}
#define init_13_12_1_14
void FASTCALL call_13_12_1_14(SH2_struct* sh)
{
  SH2movli(sh, 12, (1<<4)+14);
}
#define init_13_12_1_15
void FASTCALL call_13_12_1_15(SH2_struct* sh)
{
  SH2movli(sh, 12, (1<<4)+15);
}
#define init_13_12_2_0
void FASTCALL call_13_12_2_0(SH2_struct* sh)
{
  SH2movli(sh, 12, (2<<4)+0);
}
#define init_13_12_2_1
void FASTCALL call_13_12_2_1(SH2_struct* sh)
{
  SH2movli(sh, 12, (2<<4)+1);
}
#define init_13_12_2_2
void FASTCALL call_13_12_2_2(SH2_struct* sh)
{
  SH2movli(sh, 12, (2<<4)+2);
}
#define init_13_12_2_3
void FASTCALL call_13_12_2_3(SH2_struct* sh)
{
  SH2movli(sh, 12, (2<<4)+3);
}
#define init_13_12_2_4
void FASTCALL call_13_12_2_4(SH2_struct* sh)
{
  SH2movli(sh, 12, (2<<4)+4);
}
#define init_13_12_2_5
void FASTCALL call_13_12_2_5(SH2_struct* sh)
{
  SH2movli(sh, 12, (2<<4)+5);
}
#define init_13_12_2_6
void FASTCALL call_13_12_2_6(SH2_struct* sh)
{
  SH2movli(sh, 12, (2<<4)+6);
}
#define init_13_12_2_7
void FASTCALL call_13_12_2_7(SH2_struct* sh)
{
  SH2movli(sh, 12, (2<<4)+7);
}
#define init_13_12_2_8
void FASTCALL call_13_12_2_8(SH2_struct* sh)
{
  SH2movli(sh, 12, (2<<4)+8);
}
#define init_13_12_2_9
void FASTCALL call_13_12_2_9(SH2_struct* sh)
{
  SH2movli(sh, 12, (2<<4)+9);
}
#define init_13_12_2_10
void FASTCALL call_13_12_2_10(SH2_struct* sh)
{
  SH2movli(sh, 12, (2<<4)+10);
}
#define init_13_12_2_11
void FASTCALL call_13_12_2_11(SH2_struct* sh)
{
  SH2movli(sh, 12, (2<<4)+11);
}
#define init_13_12_2_12
void FASTCALL call_13_12_2_12(SH2_struct* sh)
{
  SH2movli(sh, 12, (2<<4)+12);
}
#define init_13_12_2_13
void FASTCALL call_13_12_2_13(SH2_struct* sh)
{
  SH2movli(sh, 12, (2<<4)+13);
}
#define init_13_12_2_14
void FASTCALL call_13_12_2_14(SH2_struct* sh)
{
  SH2movli(sh, 12, (2<<4)+14);
}
#define init_13_12_2_15
void FASTCALL call_13_12_2_15(SH2_struct* sh)
{
  SH2movli(sh, 12, (2<<4)+15);
}
#define init_13_12_3_0
void FASTCALL call_13_12_3_0(SH2_struct* sh)
{
  SH2movli(sh, 12, (3<<4)+0);
}
#define init_13_12_3_1
void FASTCALL call_13_12_3_1(SH2_struct* sh)
{
  SH2movli(sh, 12, (3<<4)+1);
}
#define init_13_12_3_2
void FASTCALL call_13_12_3_2(SH2_struct* sh)
{
  SH2movli(sh, 12, (3<<4)+2);
}
#define init_13_12_3_3
void FASTCALL call_13_12_3_3(SH2_struct* sh)
{
  SH2movli(sh, 12, (3<<4)+3);
}
#define init_13_12_3_4
void FASTCALL call_13_12_3_4(SH2_struct* sh)
{
  SH2movli(sh, 12, (3<<4)+4);
}
#define init_13_12_3_5
void FASTCALL call_13_12_3_5(SH2_struct* sh)
{
  SH2movli(sh, 12, (3<<4)+5);
}
#define init_13_12_3_6
void FASTCALL call_13_12_3_6(SH2_struct* sh)
{
  SH2movli(sh, 12, (3<<4)+6);
}
#define init_13_12_3_7
void FASTCALL call_13_12_3_7(SH2_struct* sh)
{
  SH2movli(sh, 12, (3<<4)+7);
}
#define init_13_12_3_8
void FASTCALL call_13_12_3_8(SH2_struct* sh)
{
  SH2movli(sh, 12, (3<<4)+8);
}
#define init_13_12_3_9
void FASTCALL call_13_12_3_9(SH2_struct* sh)
{
  SH2movli(sh, 12, (3<<4)+9);
}
#define init_13_12_3_10
void FASTCALL call_13_12_3_10(SH2_struct* sh)
{
  SH2movli(sh, 12, (3<<4)+10);
}
#define init_13_12_3_11
void FASTCALL call_13_12_3_11(SH2_struct* sh)
{
  SH2movli(sh, 12, (3<<4)+11);
}
#define init_13_12_3_12
void FASTCALL call_13_12_3_12(SH2_struct* sh)
{
  SH2movli(sh, 12, (3<<4)+12);
}
#define init_13_12_3_13
void FASTCALL call_13_12_3_13(SH2_struct* sh)
{
  SH2movli(sh, 12, (3<<4)+13);
}
#define init_13_12_3_14
void FASTCALL call_13_12_3_14(SH2_struct* sh)
{
  SH2movli(sh, 12, (3<<4)+14);
}
#define init_13_12_3_15
void FASTCALL call_13_12_3_15(SH2_struct* sh)
{
  SH2movli(sh, 12, (3<<4)+15);
}
#define init_13_12_4_0
void FASTCALL call_13_12_4_0(SH2_struct* sh)
{
  SH2movli(sh, 12, (4<<4)+0);
}
#define init_13_12_4_1
void FASTCALL call_13_12_4_1(SH2_struct* sh)
{
  SH2movli(sh, 12, (4<<4)+1);
}
#define init_13_12_4_2
void FASTCALL call_13_12_4_2(SH2_struct* sh)
{
  SH2movli(sh, 12, (4<<4)+2);
}
#define init_13_12_4_3
void FASTCALL call_13_12_4_3(SH2_struct* sh)
{
  SH2movli(sh, 12, (4<<4)+3);
}
#define init_13_12_4_4
void FASTCALL call_13_12_4_4(SH2_struct* sh)
{
  SH2movli(sh, 12, (4<<4)+4);
}
#define init_13_12_4_5
void FASTCALL call_13_12_4_5(SH2_struct* sh)
{
  SH2movli(sh, 12, (4<<4)+5);
}
#define init_13_12_4_6
void FASTCALL call_13_12_4_6(SH2_struct* sh)
{
  SH2movli(sh, 12, (4<<4)+6);
}
#define init_13_12_4_7
void FASTCALL call_13_12_4_7(SH2_struct* sh)
{
  SH2movli(sh, 12, (4<<4)+7);
}
#define init_13_12_4_8
void FASTCALL call_13_12_4_8(SH2_struct* sh)
{
  SH2movli(sh, 12, (4<<4)+8);
}
#define init_13_12_4_9
void FASTCALL call_13_12_4_9(SH2_struct* sh)
{
  SH2movli(sh, 12, (4<<4)+9);
}
#define init_13_12_4_10
void FASTCALL call_13_12_4_10(SH2_struct* sh)
{
  SH2movli(sh, 12, (4<<4)+10);
}
#define init_13_12_4_11
void FASTCALL call_13_12_4_11(SH2_struct* sh)
{
  SH2movli(sh, 12, (4<<4)+11);
}
#define init_13_12_4_12
void FASTCALL call_13_12_4_12(SH2_struct* sh)
{
  SH2movli(sh, 12, (4<<4)+12);
}
#define init_13_12_4_13
void FASTCALL call_13_12_4_13(SH2_struct* sh)
{
  SH2movli(sh, 12, (4<<4)+13);
}
#define init_13_12_4_14
void FASTCALL call_13_12_4_14(SH2_struct* sh)
{
  SH2movli(sh, 12, (4<<4)+14);
}
#define init_13_12_4_15
void FASTCALL call_13_12_4_15(SH2_struct* sh)
{
  SH2movli(sh, 12, (4<<4)+15);
}
#define init_13_12_5_0
void FASTCALL call_13_12_5_0(SH2_struct* sh)
{
  SH2movli(sh, 12, (5<<4)+0);
}
#define init_13_12_5_1
void FASTCALL call_13_12_5_1(SH2_struct* sh)
{
  SH2movli(sh, 12, (5<<4)+1);
}
#define init_13_12_5_2
void FASTCALL call_13_12_5_2(SH2_struct* sh)
{
  SH2movli(sh, 12, (5<<4)+2);
}
#define init_13_12_5_3
void FASTCALL call_13_12_5_3(SH2_struct* sh)
{
  SH2movli(sh, 12, (5<<4)+3);
}
#define init_13_12_5_4
void FASTCALL call_13_12_5_4(SH2_struct* sh)
{
  SH2movli(sh, 12, (5<<4)+4);
}
#define init_13_12_5_5
void FASTCALL call_13_12_5_5(SH2_struct* sh)
{
  SH2movli(sh, 12, (5<<4)+5);
}
#define init_13_12_5_6
void FASTCALL call_13_12_5_6(SH2_struct* sh)
{
  SH2movli(sh, 12, (5<<4)+6);
}
#define init_13_12_5_7
void FASTCALL call_13_12_5_7(SH2_struct* sh)
{
  SH2movli(sh, 12, (5<<4)+7);
}
#define init_13_12_5_8
void FASTCALL call_13_12_5_8(SH2_struct* sh)
{
  SH2movli(sh, 12, (5<<4)+8);
}
#define init_13_12_5_9
void FASTCALL call_13_12_5_9(SH2_struct* sh)
{
  SH2movli(sh, 12, (5<<4)+9);
}
#define init_13_12_5_10
void FASTCALL call_13_12_5_10(SH2_struct* sh)
{
  SH2movli(sh, 12, (5<<4)+10);
}
#define init_13_12_5_11
void FASTCALL call_13_12_5_11(SH2_struct* sh)
{
  SH2movli(sh, 12, (5<<4)+11);
}
#define init_13_12_5_12
void FASTCALL call_13_12_5_12(SH2_struct* sh)
{
  SH2movli(sh, 12, (5<<4)+12);
}
#define init_13_12_5_13
void FASTCALL call_13_12_5_13(SH2_struct* sh)
{
  SH2movli(sh, 12, (5<<4)+13);
}
#define init_13_12_5_14
void FASTCALL call_13_12_5_14(SH2_struct* sh)
{
  SH2movli(sh, 12, (5<<4)+14);
}
#define init_13_12_5_15
void FASTCALL call_13_12_5_15(SH2_struct* sh)
{
  SH2movli(sh, 12, (5<<4)+15);
}
#define init_13_12_6_0
void FASTCALL call_13_12_6_0(SH2_struct* sh)
{
  SH2movli(sh, 12, (6<<4)+0);
}
#define init_13_12_6_1
void FASTCALL call_13_12_6_1(SH2_struct* sh)
{
  SH2movli(sh, 12, (6<<4)+1);
}
#define init_13_12_6_2
void FASTCALL call_13_12_6_2(SH2_struct* sh)
{
  SH2movli(sh, 12, (6<<4)+2);
}
#define init_13_12_6_3
void FASTCALL call_13_12_6_3(SH2_struct* sh)
{
  SH2movli(sh, 12, (6<<4)+3);
}
#define init_13_12_6_4
void FASTCALL call_13_12_6_4(SH2_struct* sh)
{
  SH2movli(sh, 12, (6<<4)+4);
}
#define init_13_12_6_5
void FASTCALL call_13_12_6_5(SH2_struct* sh)
{
  SH2movli(sh, 12, (6<<4)+5);
}
#define init_13_12_6_6
void FASTCALL call_13_12_6_6(SH2_struct* sh)
{
  SH2movli(sh, 12, (6<<4)+6);
}
#define init_13_12_6_7
void FASTCALL call_13_12_6_7(SH2_struct* sh)
{
  SH2movli(sh, 12, (6<<4)+7);
}
#define init_13_12_6_8
void FASTCALL call_13_12_6_8(SH2_struct* sh)
{
  SH2movli(sh, 12, (6<<4)+8);
}
#define init_13_12_6_9
void FASTCALL call_13_12_6_9(SH2_struct* sh)
{
  SH2movli(sh, 12, (6<<4)+9);
}
#define init_13_12_6_10
void FASTCALL call_13_12_6_10(SH2_struct* sh)
{
  SH2movli(sh, 12, (6<<4)+10);
}
#define init_13_12_6_11
void FASTCALL call_13_12_6_11(SH2_struct* sh)
{
  SH2movli(sh, 12, (6<<4)+11);
}
#define init_13_12_6_12
void FASTCALL call_13_12_6_12(SH2_struct* sh)
{
  SH2movli(sh, 12, (6<<4)+12);
}
#define init_13_12_6_13
void FASTCALL call_13_12_6_13(SH2_struct* sh)
{
  SH2movli(sh, 12, (6<<4)+13);
}
#define init_13_12_6_14
void FASTCALL call_13_12_6_14(SH2_struct* sh)
{
  SH2movli(sh, 12, (6<<4)+14);
}
#define init_13_12_6_15
void FASTCALL call_13_12_6_15(SH2_struct* sh)
{
  SH2movli(sh, 12, (6<<4)+15);
}
#define init_13_12_7_0
void FASTCALL call_13_12_7_0(SH2_struct* sh)
{
  SH2movli(sh, 12, (7<<4)+0);
}
#define init_13_12_7_1
void FASTCALL call_13_12_7_1(SH2_struct* sh)
{
  SH2movli(sh, 12, (7<<4)+1);
}
#define init_13_12_7_2
void FASTCALL call_13_12_7_2(SH2_struct* sh)
{
  SH2movli(sh, 12, (7<<4)+2);
}
#define init_13_12_7_3
void FASTCALL call_13_12_7_3(SH2_struct* sh)
{
  SH2movli(sh, 12, (7<<4)+3);
}
#define init_13_12_7_4
void FASTCALL call_13_12_7_4(SH2_struct* sh)
{
  SH2movli(sh, 12, (7<<4)+4);
}
#define init_13_12_7_5
void FASTCALL call_13_12_7_5(SH2_struct* sh)
{
  SH2movli(sh, 12, (7<<4)+5);
}
#define init_13_12_7_6
void FASTCALL call_13_12_7_6(SH2_struct* sh)
{
  SH2movli(sh, 12, (7<<4)+6);
}
#define init_13_12_7_7
void FASTCALL call_13_12_7_7(SH2_struct* sh)
{
  SH2movli(sh, 12, (7<<4)+7);
}
#define init_13_12_7_8
void FASTCALL call_13_12_7_8(SH2_struct* sh)
{
  SH2movli(sh, 12, (7<<4)+8);
}
#define init_13_12_7_9
void FASTCALL call_13_12_7_9(SH2_struct* sh)
{
  SH2movli(sh, 12, (7<<4)+9);
}
#define init_13_12_7_10
void FASTCALL call_13_12_7_10(SH2_struct* sh)
{
  SH2movli(sh, 12, (7<<4)+10);
}
#define init_13_12_7_11
void FASTCALL call_13_12_7_11(SH2_struct* sh)
{
  SH2movli(sh, 12, (7<<4)+11);
}
#define init_13_12_7_12
void FASTCALL call_13_12_7_12(SH2_struct* sh)
{
  SH2movli(sh, 12, (7<<4)+12);
}
#define init_13_12_7_13
void FASTCALL call_13_12_7_13(SH2_struct* sh)
{
  SH2movli(sh, 12, (7<<4)+13);
}
#define init_13_12_7_14
void FASTCALL call_13_12_7_14(SH2_struct* sh)
{
  SH2movli(sh, 12, (7<<4)+14);
}
#define init_13_12_7_15
void FASTCALL call_13_12_7_15(SH2_struct* sh)
{
  SH2movli(sh, 12, (7<<4)+15);
}
#define init_13_12_8_0
void FASTCALL call_13_12_8_0(SH2_struct* sh)
{
  SH2movli(sh, 12, (8<<4)+0);
}
#define init_13_12_8_1
void FASTCALL call_13_12_8_1(SH2_struct* sh)
{
  SH2movli(sh, 12, (8<<4)+1);
}
#define init_13_12_8_2
void FASTCALL call_13_12_8_2(SH2_struct* sh)
{
  SH2movli(sh, 12, (8<<4)+2);
}
#define init_13_12_8_3
void FASTCALL call_13_12_8_3(SH2_struct* sh)
{
  SH2movli(sh, 12, (8<<4)+3);
}
#define init_13_12_8_4
void FASTCALL call_13_12_8_4(SH2_struct* sh)
{
  SH2movli(sh, 12, (8<<4)+4);
}
#define init_13_12_8_5
void FASTCALL call_13_12_8_5(SH2_struct* sh)
{
  SH2movli(sh, 12, (8<<4)+5);
}
#define init_13_12_8_6
void FASTCALL call_13_12_8_6(SH2_struct* sh)
{
  SH2movli(sh, 12, (8<<4)+6);
}
#define init_13_12_8_7
void FASTCALL call_13_12_8_7(SH2_struct* sh)
{
  SH2movli(sh, 12, (8<<4)+7);
}
#define init_13_12_8_8
void FASTCALL call_13_12_8_8(SH2_struct* sh)
{
  SH2movli(sh, 12, (8<<4)+8);
}
#define init_13_12_8_9
void FASTCALL call_13_12_8_9(SH2_struct* sh)
{
  SH2movli(sh, 12, (8<<4)+9);
}
#define init_13_12_8_10
void FASTCALL call_13_12_8_10(SH2_struct* sh)
{
  SH2movli(sh, 12, (8<<4)+10);
}
#define init_13_12_8_11
void FASTCALL call_13_12_8_11(SH2_struct* sh)
{
  SH2movli(sh, 12, (8<<4)+11);
}
#define init_13_12_8_12
void FASTCALL call_13_12_8_12(SH2_struct* sh)
{
  SH2movli(sh, 12, (8<<4)+12);
}
#define init_13_12_8_13
void FASTCALL call_13_12_8_13(SH2_struct* sh)
{
  SH2movli(sh, 12, (8<<4)+13);
}
#define init_13_12_8_14
void FASTCALL call_13_12_8_14(SH2_struct* sh)
{
  SH2movli(sh, 12, (8<<4)+14);
}
#define init_13_12_8_15
void FASTCALL call_13_12_8_15(SH2_struct* sh)
{
  SH2movli(sh, 12, (8<<4)+15);
}
#define init_13_12_9_0
void FASTCALL call_13_12_9_0(SH2_struct* sh)
{
  SH2movli(sh, 12, (9<<4)+0);
}
#define init_13_12_9_1
void FASTCALL call_13_12_9_1(SH2_struct* sh)
{
  SH2movli(sh, 12, (9<<4)+1);
}
#define init_13_12_9_2
void FASTCALL call_13_12_9_2(SH2_struct* sh)
{
  SH2movli(sh, 12, (9<<4)+2);
}
#define init_13_12_9_3
void FASTCALL call_13_12_9_3(SH2_struct* sh)
{
  SH2movli(sh, 12, (9<<4)+3);
}
#define init_13_12_9_4
void FASTCALL call_13_12_9_4(SH2_struct* sh)
{
  SH2movli(sh, 12, (9<<4)+4);
}
#define init_13_12_9_5
void FASTCALL call_13_12_9_5(SH2_struct* sh)
{
  SH2movli(sh, 12, (9<<4)+5);
}
#define init_13_12_9_6
void FASTCALL call_13_12_9_6(SH2_struct* sh)
{
  SH2movli(sh, 12, (9<<4)+6);
}
#define init_13_12_9_7
void FASTCALL call_13_12_9_7(SH2_struct* sh)
{
  SH2movli(sh, 12, (9<<4)+7);
}
#define init_13_12_9_8
void FASTCALL call_13_12_9_8(SH2_struct* sh)
{
  SH2movli(sh, 12, (9<<4)+8);
}
#define init_13_12_9_9
void FASTCALL call_13_12_9_9(SH2_struct* sh)
{
  SH2movli(sh, 12, (9<<4)+9);
}
#define init_13_12_9_10
void FASTCALL call_13_12_9_10(SH2_struct* sh)
{
  SH2movli(sh, 12, (9<<4)+10);
}
#define init_13_12_9_11
void FASTCALL call_13_12_9_11(SH2_struct* sh)
{
  SH2movli(sh, 12, (9<<4)+11);
}
#define init_13_12_9_12
void FASTCALL call_13_12_9_12(SH2_struct* sh)
{
  SH2movli(sh, 12, (9<<4)+12);
}
#define init_13_12_9_13
void FASTCALL call_13_12_9_13(SH2_struct* sh)
{
  SH2movli(sh, 12, (9<<4)+13);
}
#define init_13_12_9_14
void FASTCALL call_13_12_9_14(SH2_struct* sh)
{
  SH2movli(sh, 12, (9<<4)+14);
}
#define init_13_12_9_15
void FASTCALL call_13_12_9_15(SH2_struct* sh)
{
  SH2movli(sh, 12, (9<<4)+15);
}
#define init_13_12_10_0
void FASTCALL call_13_12_10_0(SH2_struct* sh)
{
  SH2movli(sh, 12, (10<<4)+0);
}
#define init_13_12_10_1
void FASTCALL call_13_12_10_1(SH2_struct* sh)
{
  SH2movli(sh, 12, (10<<4)+1);
}
#define init_13_12_10_2
void FASTCALL call_13_12_10_2(SH2_struct* sh)
{
  SH2movli(sh, 12, (10<<4)+2);
}
#define init_13_12_10_3
void FASTCALL call_13_12_10_3(SH2_struct* sh)
{
  SH2movli(sh, 12, (10<<4)+3);
}
#define init_13_12_10_4
void FASTCALL call_13_12_10_4(SH2_struct* sh)
{
  SH2movli(sh, 12, (10<<4)+4);
}
#define init_13_12_10_5
void FASTCALL call_13_12_10_5(SH2_struct* sh)
{
  SH2movli(sh, 12, (10<<4)+5);
}
#define init_13_12_10_6
void FASTCALL call_13_12_10_6(SH2_struct* sh)
{
  SH2movli(sh, 12, (10<<4)+6);
}
#define init_13_12_10_7
void FASTCALL call_13_12_10_7(SH2_struct* sh)
{
  SH2movli(sh, 12, (10<<4)+7);
}
#define init_13_12_10_8
void FASTCALL call_13_12_10_8(SH2_struct* sh)
{
  SH2movli(sh, 12, (10<<4)+8);
}
#define init_13_12_10_9
void FASTCALL call_13_12_10_9(SH2_struct* sh)
{
  SH2movli(sh, 12, (10<<4)+9);
}
#define init_13_12_10_10
void FASTCALL call_13_12_10_10(SH2_struct* sh)
{
  SH2movli(sh, 12, (10<<4)+10);
}
#define init_13_12_10_11
void FASTCALL call_13_12_10_11(SH2_struct* sh)
{
  SH2movli(sh, 12, (10<<4)+11);
}
#define init_13_12_10_12
void FASTCALL call_13_12_10_12(SH2_struct* sh)
{
  SH2movli(sh, 12, (10<<4)+12);
}
#define init_13_12_10_13
void FASTCALL call_13_12_10_13(SH2_struct* sh)
{
  SH2movli(sh, 12, (10<<4)+13);
}
#define init_13_12_10_14
void FASTCALL call_13_12_10_14(SH2_struct* sh)
{
  SH2movli(sh, 12, (10<<4)+14);
}
#define init_13_12_10_15
void FASTCALL call_13_12_10_15(SH2_struct* sh)
{
  SH2movli(sh, 12, (10<<4)+15);
}
#define init_13_12_11_0
void FASTCALL call_13_12_11_0(SH2_struct* sh)
{
  SH2movli(sh, 12, (11<<4)+0);
}
#define init_13_12_11_1
void FASTCALL call_13_12_11_1(SH2_struct* sh)
{
  SH2movli(sh, 12, (11<<4)+1);
}
#define init_13_12_11_2
void FASTCALL call_13_12_11_2(SH2_struct* sh)
{
  SH2movli(sh, 12, (11<<4)+2);
}
#define init_13_12_11_3
void FASTCALL call_13_12_11_3(SH2_struct* sh)
{
  SH2movli(sh, 12, (11<<4)+3);
}
#define init_13_12_11_4
void FASTCALL call_13_12_11_4(SH2_struct* sh)
{
  SH2movli(sh, 12, (11<<4)+4);
}
#define init_13_12_11_5
void FASTCALL call_13_12_11_5(SH2_struct* sh)
{
  SH2movli(sh, 12, (11<<4)+5);
}
#define init_13_12_11_6
void FASTCALL call_13_12_11_6(SH2_struct* sh)
{
  SH2movli(sh, 12, (11<<4)+6);
}
#define init_13_12_11_7
void FASTCALL call_13_12_11_7(SH2_struct* sh)
{
  SH2movli(sh, 12, (11<<4)+7);
}
#define init_13_12_11_8
void FASTCALL call_13_12_11_8(SH2_struct* sh)
{
  SH2movli(sh, 12, (11<<4)+8);
}
#define init_13_12_11_9
void FASTCALL call_13_12_11_9(SH2_struct* sh)
{
  SH2movli(sh, 12, (11<<4)+9);
}
#define init_13_12_11_10
void FASTCALL call_13_12_11_10(SH2_struct* sh)
{
  SH2movli(sh, 12, (11<<4)+10);
}
#define init_13_12_11_11
void FASTCALL call_13_12_11_11(SH2_struct* sh)
{
  SH2movli(sh, 12, (11<<4)+11);
}
#define init_13_12_11_12
void FASTCALL call_13_12_11_12(SH2_struct* sh)
{
  SH2movli(sh, 12, (11<<4)+12);
}
#define init_13_12_11_13
void FASTCALL call_13_12_11_13(SH2_struct* sh)
{
  SH2movli(sh, 12, (11<<4)+13);
}
#define init_13_12_11_14
void FASTCALL call_13_12_11_14(SH2_struct* sh)
{
  SH2movli(sh, 12, (11<<4)+14);
}
#define init_13_12_11_15
void FASTCALL call_13_12_11_15(SH2_struct* sh)
{
  SH2movli(sh, 12, (11<<4)+15);
}
#define init_13_12_12_0
void FASTCALL call_13_12_12_0(SH2_struct* sh)
{
  SH2movli(sh, 12, (12<<4)+0);
}
#define init_13_12_12_1
void FASTCALL call_13_12_12_1(SH2_struct* sh)
{
  SH2movli(sh, 12, (12<<4)+1);
}
#define init_13_12_12_2
void FASTCALL call_13_12_12_2(SH2_struct* sh)
{
  SH2movli(sh, 12, (12<<4)+2);
}
#define init_13_12_12_3
void FASTCALL call_13_12_12_3(SH2_struct* sh)
{
  SH2movli(sh, 12, (12<<4)+3);
}
#define init_13_12_12_4
void FASTCALL call_13_12_12_4(SH2_struct* sh)
{
  SH2movli(sh, 12, (12<<4)+4);
}
#define init_13_12_12_5
void FASTCALL call_13_12_12_5(SH2_struct* sh)
{
  SH2movli(sh, 12, (12<<4)+5);
}
#define init_13_12_12_6
void FASTCALL call_13_12_12_6(SH2_struct* sh)
{
  SH2movli(sh, 12, (12<<4)+6);
}
#define init_13_12_12_7
void FASTCALL call_13_12_12_7(SH2_struct* sh)
{
  SH2movli(sh, 12, (12<<4)+7);
}
#define init_13_12_12_8
void FASTCALL call_13_12_12_8(SH2_struct* sh)
{
  SH2movli(sh, 12, (12<<4)+8);
}
#define init_13_12_12_9
void FASTCALL call_13_12_12_9(SH2_struct* sh)
{
  SH2movli(sh, 12, (12<<4)+9);
}
#define init_13_12_12_10
void FASTCALL call_13_12_12_10(SH2_struct* sh)
{
  SH2movli(sh, 12, (12<<4)+10);
}
#define init_13_12_12_11
void FASTCALL call_13_12_12_11(SH2_struct* sh)
{
  SH2movli(sh, 12, (12<<4)+11);
}
#define init_13_12_12_12
void FASTCALL call_13_12_12_12(SH2_struct* sh)
{
  SH2movli(sh, 12, (12<<4)+12);
}
#define init_13_12_12_13
void FASTCALL call_13_12_12_13(SH2_struct* sh)
{
  SH2movli(sh, 12, (12<<4)+13);
}
#define init_13_12_12_14
void FASTCALL call_13_12_12_14(SH2_struct* sh)
{
  SH2movli(sh, 12, (12<<4)+14);
}
#define init_13_12_12_15
void FASTCALL call_13_12_12_15(SH2_struct* sh)
{
  SH2movli(sh, 12, (12<<4)+15);
}
#define init_13_12_13_0
void FASTCALL call_13_12_13_0(SH2_struct* sh)
{
  SH2movli(sh, 12, (13<<4)+0);
}
#define init_13_12_13_1
void FASTCALL call_13_12_13_1(SH2_struct* sh)
{
  SH2movli(sh, 12, (13<<4)+1);
}
#define init_13_12_13_2
void FASTCALL call_13_12_13_2(SH2_struct* sh)
{
  SH2movli(sh, 12, (13<<4)+2);
}
#define init_13_12_13_3
void FASTCALL call_13_12_13_3(SH2_struct* sh)
{
  SH2movli(sh, 12, (13<<4)+3);
}
#define init_13_12_13_4
void FASTCALL call_13_12_13_4(SH2_struct* sh)
{
  SH2movli(sh, 12, (13<<4)+4);
}
#define init_13_12_13_5
void FASTCALL call_13_12_13_5(SH2_struct* sh)
{
  SH2movli(sh, 12, (13<<4)+5);
}
#define init_13_12_13_6
void FASTCALL call_13_12_13_6(SH2_struct* sh)
{
  SH2movli(sh, 12, (13<<4)+6);
}
#define init_13_12_13_7
void FASTCALL call_13_12_13_7(SH2_struct* sh)
{
  SH2movli(sh, 12, (13<<4)+7);
}
#define init_13_12_13_8
void FASTCALL call_13_12_13_8(SH2_struct* sh)
{
  SH2movli(sh, 12, (13<<4)+8);
}
#define init_13_12_13_9
void FASTCALL call_13_12_13_9(SH2_struct* sh)
{
  SH2movli(sh, 12, (13<<4)+9);
}
#define init_13_12_13_10
void FASTCALL call_13_12_13_10(SH2_struct* sh)
{
  SH2movli(sh, 12, (13<<4)+10);
}
#define init_13_12_13_11
void FASTCALL call_13_12_13_11(SH2_struct* sh)
{
  SH2movli(sh, 12, (13<<4)+11);
}
#define init_13_12_13_12
void FASTCALL call_13_12_13_12(SH2_struct* sh)
{
  SH2movli(sh, 12, (13<<4)+12);
}
#define init_13_12_13_13
void FASTCALL call_13_12_13_13(SH2_struct* sh)
{
  SH2movli(sh, 12, (13<<4)+13);
}
#define init_13_12_13_14
void FASTCALL call_13_12_13_14(SH2_struct* sh)
{
  SH2movli(sh, 12, (13<<4)+14);
}
#define init_13_12_13_15
void FASTCALL call_13_12_13_15(SH2_struct* sh)
{
  SH2movli(sh, 12, (13<<4)+15);
}
#define init_13_12_14_0
void FASTCALL call_13_12_14_0(SH2_struct* sh)
{
  SH2movli(sh, 12, (14<<4)+0);
}
#define init_13_12_14_1
void FASTCALL call_13_12_14_1(SH2_struct* sh)
{
  SH2movli(sh, 12, (14<<4)+1);
}
#define init_13_12_14_2
void FASTCALL call_13_12_14_2(SH2_struct* sh)
{
  SH2movli(sh, 12, (14<<4)+2);
}
#define init_13_12_14_3
void FASTCALL call_13_12_14_3(SH2_struct* sh)
{
  SH2movli(sh, 12, (14<<4)+3);
}
#define init_13_12_14_4
void FASTCALL call_13_12_14_4(SH2_struct* sh)
{
  SH2movli(sh, 12, (14<<4)+4);
}
#define init_13_12_14_5
void FASTCALL call_13_12_14_5(SH2_struct* sh)
{
  SH2movli(sh, 12, (14<<4)+5);
}
#define init_13_12_14_6
void FASTCALL call_13_12_14_6(SH2_struct* sh)
{
  SH2movli(sh, 12, (14<<4)+6);
}
#define init_13_12_14_7
void FASTCALL call_13_12_14_7(SH2_struct* sh)
{
  SH2movli(sh, 12, (14<<4)+7);
}
#define init_13_12_14_8
void FASTCALL call_13_12_14_8(SH2_struct* sh)
{
  SH2movli(sh, 12, (14<<4)+8);
}
#define init_13_12_14_9
void FASTCALL call_13_12_14_9(SH2_struct* sh)
{
  SH2movli(sh, 12, (14<<4)+9);
}
#define init_13_12_14_10
void FASTCALL call_13_12_14_10(SH2_struct* sh)
{
  SH2movli(sh, 12, (14<<4)+10);
}
#define init_13_12_14_11
void FASTCALL call_13_12_14_11(SH2_struct* sh)
{
  SH2movli(sh, 12, (14<<4)+11);
}
#define init_13_12_14_12
void FASTCALL call_13_12_14_12(SH2_struct* sh)
{
  SH2movli(sh, 12, (14<<4)+12);
}
#define init_13_12_14_13
void FASTCALL call_13_12_14_13(SH2_struct* sh)
{
  SH2movli(sh, 12, (14<<4)+13);
}
#define init_13_12_14_14
void FASTCALL call_13_12_14_14(SH2_struct* sh)
{
  SH2movli(sh, 12, (14<<4)+14);
}
#define init_13_12_14_15
void FASTCALL call_13_12_14_15(SH2_struct* sh)
{
  SH2movli(sh, 12, (14<<4)+15);
}
#define init_13_12_15_0
void FASTCALL call_13_12_15_0(SH2_struct* sh)
{
  SH2movli(sh, 12, (15<<4)+0);
}
#define init_13_12_15_1
void FASTCALL call_13_12_15_1(SH2_struct* sh)
{
  SH2movli(sh, 12, (15<<4)+1);
}
#define init_13_12_15_2
void FASTCALL call_13_12_15_2(SH2_struct* sh)
{
  SH2movli(sh, 12, (15<<4)+2);
}
#define init_13_12_15_3
void FASTCALL call_13_12_15_3(SH2_struct* sh)
{
  SH2movli(sh, 12, (15<<4)+3);
}
#define init_13_12_15_4
void FASTCALL call_13_12_15_4(SH2_struct* sh)
{
  SH2movli(sh, 12, (15<<4)+4);
}
#define init_13_12_15_5
void FASTCALL call_13_12_15_5(SH2_struct* sh)
{
  SH2movli(sh, 12, (15<<4)+5);
}
#define init_13_12_15_6
void FASTCALL call_13_12_15_6(SH2_struct* sh)
{
  SH2movli(sh, 12, (15<<4)+6);
}
#define init_13_12_15_7
void FASTCALL call_13_12_15_7(SH2_struct* sh)
{
  SH2movli(sh, 12, (15<<4)+7);
}
#define init_13_12_15_8
void FASTCALL call_13_12_15_8(SH2_struct* sh)
{
  SH2movli(sh, 12, (15<<4)+8);
}
#define init_13_12_15_9
void FASTCALL call_13_12_15_9(SH2_struct* sh)
{
  SH2movli(sh, 12, (15<<4)+9);
}
#define init_13_12_15_10
void FASTCALL call_13_12_15_10(SH2_struct* sh)
{
  SH2movli(sh, 12, (15<<4)+10);
}
#define init_13_12_15_11
void FASTCALL call_13_12_15_11(SH2_struct* sh)
{
  SH2movli(sh, 12, (15<<4)+11);
}
#define init_13_12_15_12
void FASTCALL call_13_12_15_12(SH2_struct* sh)
{
  SH2movli(sh, 12, (15<<4)+12);
}
#define init_13_12_15_13
void FASTCALL call_13_12_15_13(SH2_struct* sh)
{
  SH2movli(sh, 12, (15<<4)+13);
}
#define init_13_12_15_14
void FASTCALL call_13_12_15_14(SH2_struct* sh)
{
  SH2movli(sh, 12, (15<<4)+14);
}
#define init_13_12_15_15
void FASTCALL call_13_12_15_15(SH2_struct* sh)
{
  SH2movli(sh, 12, (15<<4)+15);
}
#define init_13_13_0_0
void FASTCALL call_13_13_0_0(SH2_struct* sh)
{
  SH2movli(sh, 13, (0<<4)+0);
}
#define init_13_13_0_1
void FASTCALL call_13_13_0_1(SH2_struct* sh)
{
  SH2movli(sh, 13, (0<<4)+1);
}
#define init_13_13_0_2
void FASTCALL call_13_13_0_2(SH2_struct* sh)
{
  SH2movli(sh, 13, (0<<4)+2);
}
#define init_13_13_0_3
void FASTCALL call_13_13_0_3(SH2_struct* sh)
{
  SH2movli(sh, 13, (0<<4)+3);
}
#define init_13_13_0_4
void FASTCALL call_13_13_0_4(SH2_struct* sh)
{
  SH2movli(sh, 13, (0<<4)+4);
}
#define init_13_13_0_5
void FASTCALL call_13_13_0_5(SH2_struct* sh)
{
  SH2movli(sh, 13, (0<<4)+5);
}
#define init_13_13_0_6
void FASTCALL call_13_13_0_6(SH2_struct* sh)
{
  SH2movli(sh, 13, (0<<4)+6);
}
#define init_13_13_0_7
void FASTCALL call_13_13_0_7(SH2_struct* sh)
{
  SH2movli(sh, 13, (0<<4)+7);
}
#define init_13_13_0_8
void FASTCALL call_13_13_0_8(SH2_struct* sh)
{
  SH2movli(sh, 13, (0<<4)+8);
}
#define init_13_13_0_9
void FASTCALL call_13_13_0_9(SH2_struct* sh)
{
  SH2movli(sh, 13, (0<<4)+9);
}
#define init_13_13_0_10
void FASTCALL call_13_13_0_10(SH2_struct* sh)
{
  SH2movli(sh, 13, (0<<4)+10);
}
#define init_13_13_0_11
void FASTCALL call_13_13_0_11(SH2_struct* sh)
{
  SH2movli(sh, 13, (0<<4)+11);
}
#define init_13_13_0_12
void FASTCALL call_13_13_0_12(SH2_struct* sh)
{
  SH2movli(sh, 13, (0<<4)+12);
}
#define init_13_13_0_13
void FASTCALL call_13_13_0_13(SH2_struct* sh)
{
  SH2movli(sh, 13, (0<<4)+13);
}
#define init_13_13_0_14
void FASTCALL call_13_13_0_14(SH2_struct* sh)
{
  SH2movli(sh, 13, (0<<4)+14);
}
#define init_13_13_0_15
void FASTCALL call_13_13_0_15(SH2_struct* sh)
{
  SH2movli(sh, 13, (0<<4)+15);
}
#define init_13_13_1_0
void FASTCALL call_13_13_1_0(SH2_struct* sh)
{
  SH2movli(sh, 13, (1<<4)+0);
}
#define init_13_13_1_1
void FASTCALL call_13_13_1_1(SH2_struct* sh)
{
  SH2movli(sh, 13, (1<<4)+1);
}
#define init_13_13_1_2
void FASTCALL call_13_13_1_2(SH2_struct* sh)
{
  SH2movli(sh, 13, (1<<4)+2);
}
#define init_13_13_1_3
void FASTCALL call_13_13_1_3(SH2_struct* sh)
{
  SH2movli(sh, 13, (1<<4)+3);
}
#define init_13_13_1_4
void FASTCALL call_13_13_1_4(SH2_struct* sh)
{
  SH2movli(sh, 13, (1<<4)+4);
}
#define init_13_13_1_5
void FASTCALL call_13_13_1_5(SH2_struct* sh)
{
  SH2movli(sh, 13, (1<<4)+5);
}
#define init_13_13_1_6
void FASTCALL call_13_13_1_6(SH2_struct* sh)
{
  SH2movli(sh, 13, (1<<4)+6);
}
#define init_13_13_1_7
void FASTCALL call_13_13_1_7(SH2_struct* sh)
{
  SH2movli(sh, 13, (1<<4)+7);
}
#define init_13_13_1_8
void FASTCALL call_13_13_1_8(SH2_struct* sh)
{
  SH2movli(sh, 13, (1<<4)+8);
}
#define init_13_13_1_9
void FASTCALL call_13_13_1_9(SH2_struct* sh)
{
  SH2movli(sh, 13, (1<<4)+9);
}
#define init_13_13_1_10
void FASTCALL call_13_13_1_10(SH2_struct* sh)
{
  SH2movli(sh, 13, (1<<4)+10);
}
#define init_13_13_1_11
void FASTCALL call_13_13_1_11(SH2_struct* sh)
{
  SH2movli(sh, 13, (1<<4)+11);
}
#define init_13_13_1_12
void FASTCALL call_13_13_1_12(SH2_struct* sh)
{
  SH2movli(sh, 13, (1<<4)+12);
}
#define init_13_13_1_13
void FASTCALL call_13_13_1_13(SH2_struct* sh)
{
  SH2movli(sh, 13, (1<<4)+13);
}
#define init_13_13_1_14
void FASTCALL call_13_13_1_14(SH2_struct* sh)
{
  SH2movli(sh, 13, (1<<4)+14);
}
#define init_13_13_1_15
void FASTCALL call_13_13_1_15(SH2_struct* sh)
{
  SH2movli(sh, 13, (1<<4)+15);
}
#define init_13_13_2_0
void FASTCALL call_13_13_2_0(SH2_struct* sh)
{
  SH2movli(sh, 13, (2<<4)+0);
}
#define init_13_13_2_1
void FASTCALL call_13_13_2_1(SH2_struct* sh)
{
  SH2movli(sh, 13, (2<<4)+1);
}
#define init_13_13_2_2
void FASTCALL call_13_13_2_2(SH2_struct* sh)
{
  SH2movli(sh, 13, (2<<4)+2);
}
#define init_13_13_2_3
void FASTCALL call_13_13_2_3(SH2_struct* sh)
{
  SH2movli(sh, 13, (2<<4)+3);
}
#define init_13_13_2_4
void FASTCALL call_13_13_2_4(SH2_struct* sh)
{
  SH2movli(sh, 13, (2<<4)+4);
}
#define init_13_13_2_5
void FASTCALL call_13_13_2_5(SH2_struct* sh)
{
  SH2movli(sh, 13, (2<<4)+5);
}
#define init_13_13_2_6
void FASTCALL call_13_13_2_6(SH2_struct* sh)
{
  SH2movli(sh, 13, (2<<4)+6);
}
#define init_13_13_2_7
void FASTCALL call_13_13_2_7(SH2_struct* sh)
{
  SH2movli(sh, 13, (2<<4)+7);
}
#define init_13_13_2_8
void FASTCALL call_13_13_2_8(SH2_struct* sh)
{
  SH2movli(sh, 13, (2<<4)+8);
}
#define init_13_13_2_9
void FASTCALL call_13_13_2_9(SH2_struct* sh)
{
  SH2movli(sh, 13, (2<<4)+9);
}
#define init_13_13_2_10
void FASTCALL call_13_13_2_10(SH2_struct* sh)
{
  SH2movli(sh, 13, (2<<4)+10);
}
#define init_13_13_2_11
void FASTCALL call_13_13_2_11(SH2_struct* sh)
{
  SH2movli(sh, 13, (2<<4)+11);
}
#define init_13_13_2_12
void FASTCALL call_13_13_2_12(SH2_struct* sh)
{
  SH2movli(sh, 13, (2<<4)+12);
}
#define init_13_13_2_13
void FASTCALL call_13_13_2_13(SH2_struct* sh)
{
  SH2movli(sh, 13, (2<<4)+13);
}
#define init_13_13_2_14
void FASTCALL call_13_13_2_14(SH2_struct* sh)
{
  SH2movli(sh, 13, (2<<4)+14);
}
#define init_13_13_2_15
void FASTCALL call_13_13_2_15(SH2_struct* sh)
{
  SH2movli(sh, 13, (2<<4)+15);
}
#define init_13_13_3_0
void FASTCALL call_13_13_3_0(SH2_struct* sh)
{
  SH2movli(sh, 13, (3<<4)+0);
}
#define init_13_13_3_1
void FASTCALL call_13_13_3_1(SH2_struct* sh)
{
  SH2movli(sh, 13, (3<<4)+1);
}
#define init_13_13_3_2
void FASTCALL call_13_13_3_2(SH2_struct* sh)
{
  SH2movli(sh, 13, (3<<4)+2);
}
#define init_13_13_3_3
void FASTCALL call_13_13_3_3(SH2_struct* sh)
{
  SH2movli(sh, 13, (3<<4)+3);
}
#define init_13_13_3_4
void FASTCALL call_13_13_3_4(SH2_struct* sh)
{
  SH2movli(sh, 13, (3<<4)+4);
}
#define init_13_13_3_5
void FASTCALL call_13_13_3_5(SH2_struct* sh)
{
  SH2movli(sh, 13, (3<<4)+5);
}
#define init_13_13_3_6
void FASTCALL call_13_13_3_6(SH2_struct* sh)
{
  SH2movli(sh, 13, (3<<4)+6);
}
#define init_13_13_3_7
void FASTCALL call_13_13_3_7(SH2_struct* sh)
{
  SH2movli(sh, 13, (3<<4)+7);
}
#define init_13_13_3_8
void FASTCALL call_13_13_3_8(SH2_struct* sh)
{
  SH2movli(sh, 13, (3<<4)+8);
}
#define init_13_13_3_9
void FASTCALL call_13_13_3_9(SH2_struct* sh)
{
  SH2movli(sh, 13, (3<<4)+9);
}
#define init_13_13_3_10
void FASTCALL call_13_13_3_10(SH2_struct* sh)
{
  SH2movli(sh, 13, (3<<4)+10);
}
#define init_13_13_3_11
void FASTCALL call_13_13_3_11(SH2_struct* sh)
{
  SH2movli(sh, 13, (3<<4)+11);
}
#define init_13_13_3_12
void FASTCALL call_13_13_3_12(SH2_struct* sh)
{
  SH2movli(sh, 13, (3<<4)+12);
}
#define init_13_13_3_13
void FASTCALL call_13_13_3_13(SH2_struct* sh)
{
  SH2movli(sh, 13, (3<<4)+13);
}
#define init_13_13_3_14
void FASTCALL call_13_13_3_14(SH2_struct* sh)
{
  SH2movli(sh, 13, (3<<4)+14);
}
#define init_13_13_3_15
void FASTCALL call_13_13_3_15(SH2_struct* sh)
{
  SH2movli(sh, 13, (3<<4)+15);
}
#define init_13_13_4_0
void FASTCALL call_13_13_4_0(SH2_struct* sh)
{
  SH2movli(sh, 13, (4<<4)+0);
}
#define init_13_13_4_1
void FASTCALL call_13_13_4_1(SH2_struct* sh)
{
  SH2movli(sh, 13, (4<<4)+1);
}
#define init_13_13_4_2
void FASTCALL call_13_13_4_2(SH2_struct* sh)
{
  SH2movli(sh, 13, (4<<4)+2);
}
#define init_13_13_4_3
void FASTCALL call_13_13_4_3(SH2_struct* sh)
{
  SH2movli(sh, 13, (4<<4)+3);
}
#define init_13_13_4_4
void FASTCALL call_13_13_4_4(SH2_struct* sh)
{
  SH2movli(sh, 13, (4<<4)+4);
}
#define init_13_13_4_5
void FASTCALL call_13_13_4_5(SH2_struct* sh)
{
  SH2movli(sh, 13, (4<<4)+5);
}
#define init_13_13_4_6
void FASTCALL call_13_13_4_6(SH2_struct* sh)
{
  SH2movli(sh, 13, (4<<4)+6);
}
#define init_13_13_4_7
void FASTCALL call_13_13_4_7(SH2_struct* sh)
{
  SH2movli(sh, 13, (4<<4)+7);
}
#define init_13_13_4_8
void FASTCALL call_13_13_4_8(SH2_struct* sh)
{
  SH2movli(sh, 13, (4<<4)+8);
}
#define init_13_13_4_9
void FASTCALL call_13_13_4_9(SH2_struct* sh)
{
  SH2movli(sh, 13, (4<<4)+9);
}
#define init_13_13_4_10
void FASTCALL call_13_13_4_10(SH2_struct* sh)
{
  SH2movli(sh, 13, (4<<4)+10);
}
#define init_13_13_4_11
void FASTCALL call_13_13_4_11(SH2_struct* sh)
{
  SH2movli(sh, 13, (4<<4)+11);
}
#define init_13_13_4_12
void FASTCALL call_13_13_4_12(SH2_struct* sh)
{
  SH2movli(sh, 13, (4<<4)+12);
}
#define init_13_13_4_13
void FASTCALL call_13_13_4_13(SH2_struct* sh)
{
  SH2movli(sh, 13, (4<<4)+13);
}
#define init_13_13_4_14
void FASTCALL call_13_13_4_14(SH2_struct* sh)
{
  SH2movli(sh, 13, (4<<4)+14);
}
#define init_13_13_4_15
void FASTCALL call_13_13_4_15(SH2_struct* sh)
{
  SH2movli(sh, 13, (4<<4)+15);
}
#define init_13_13_5_0
void FASTCALL call_13_13_5_0(SH2_struct* sh)
{
  SH2movli(sh, 13, (5<<4)+0);
}
#define init_13_13_5_1
void FASTCALL call_13_13_5_1(SH2_struct* sh)
{
  SH2movli(sh, 13, (5<<4)+1);
}
#define init_13_13_5_2
void FASTCALL call_13_13_5_2(SH2_struct* sh)
{
  SH2movli(sh, 13, (5<<4)+2);
}
#define init_13_13_5_3
void FASTCALL call_13_13_5_3(SH2_struct* sh)
{
  SH2movli(sh, 13, (5<<4)+3);
}
#define init_13_13_5_4
void FASTCALL call_13_13_5_4(SH2_struct* sh)
{
  SH2movli(sh, 13, (5<<4)+4);
}
#define init_13_13_5_5
void FASTCALL call_13_13_5_5(SH2_struct* sh)
{
  SH2movli(sh, 13, (5<<4)+5);
}
#define init_13_13_5_6
void FASTCALL call_13_13_5_6(SH2_struct* sh)
{
  SH2movli(sh, 13, (5<<4)+6);
}
#define init_13_13_5_7
void FASTCALL call_13_13_5_7(SH2_struct* sh)
{
  SH2movli(sh, 13, (5<<4)+7);
}
#define init_13_13_5_8
void FASTCALL call_13_13_5_8(SH2_struct* sh)
{
  SH2movli(sh, 13, (5<<4)+8);
}
#define init_13_13_5_9
void FASTCALL call_13_13_5_9(SH2_struct* sh)
{
  SH2movli(sh, 13, (5<<4)+9);
}
#define init_13_13_5_10
void FASTCALL call_13_13_5_10(SH2_struct* sh)
{
  SH2movli(sh, 13, (5<<4)+10);
}
#define init_13_13_5_11
void FASTCALL call_13_13_5_11(SH2_struct* sh)
{
  SH2movli(sh, 13, (5<<4)+11);
}
#define init_13_13_5_12
void FASTCALL call_13_13_5_12(SH2_struct* sh)
{
  SH2movli(sh, 13, (5<<4)+12);
}
#define init_13_13_5_13
void FASTCALL call_13_13_5_13(SH2_struct* sh)
{
  SH2movli(sh, 13, (5<<4)+13);
}
#define init_13_13_5_14
void FASTCALL call_13_13_5_14(SH2_struct* sh)
{
  SH2movli(sh, 13, (5<<4)+14);
}
#define init_13_13_5_15
void FASTCALL call_13_13_5_15(SH2_struct* sh)
{
  SH2movli(sh, 13, (5<<4)+15);
}
#define init_13_13_6_0
void FASTCALL call_13_13_6_0(SH2_struct* sh)
{
  SH2movli(sh, 13, (6<<4)+0);
}
#define init_13_13_6_1
void FASTCALL call_13_13_6_1(SH2_struct* sh)
{
  SH2movli(sh, 13, (6<<4)+1);
}
#define init_13_13_6_2
void FASTCALL call_13_13_6_2(SH2_struct* sh)
{
  SH2movli(sh, 13, (6<<4)+2);
}
#define init_13_13_6_3
void FASTCALL call_13_13_6_3(SH2_struct* sh)
{
  SH2movli(sh, 13, (6<<4)+3);
}
#define init_13_13_6_4
void FASTCALL call_13_13_6_4(SH2_struct* sh)
{
  SH2movli(sh, 13, (6<<4)+4);
}
#define init_13_13_6_5
void FASTCALL call_13_13_6_5(SH2_struct* sh)
{
  SH2movli(sh, 13, (6<<4)+5);
}
#define init_13_13_6_6
void FASTCALL call_13_13_6_6(SH2_struct* sh)
{
  SH2movli(sh, 13, (6<<4)+6);
}
#define init_13_13_6_7
void FASTCALL call_13_13_6_7(SH2_struct* sh)
{
  SH2movli(sh, 13, (6<<4)+7);
}
#define init_13_13_6_8
void FASTCALL call_13_13_6_8(SH2_struct* sh)
{
  SH2movli(sh, 13, (6<<4)+8);
}
#define init_13_13_6_9
void FASTCALL call_13_13_6_9(SH2_struct* sh)
{
  SH2movli(sh, 13, (6<<4)+9);
}
#define init_13_13_6_10
void FASTCALL call_13_13_6_10(SH2_struct* sh)
{
  SH2movli(sh, 13, (6<<4)+10);
}
#define init_13_13_6_11
void FASTCALL call_13_13_6_11(SH2_struct* sh)
{
  SH2movli(sh, 13, (6<<4)+11);
}
#define init_13_13_6_12
void FASTCALL call_13_13_6_12(SH2_struct* sh)
{
  SH2movli(sh, 13, (6<<4)+12);
}
#define init_13_13_6_13
void FASTCALL call_13_13_6_13(SH2_struct* sh)
{
  SH2movli(sh, 13, (6<<4)+13);
}
#define init_13_13_6_14
void FASTCALL call_13_13_6_14(SH2_struct* sh)
{
  SH2movli(sh, 13, (6<<4)+14);
}
#define init_13_13_6_15
void FASTCALL call_13_13_6_15(SH2_struct* sh)
{
  SH2movli(sh, 13, (6<<4)+15);
}
#define init_13_13_7_0
void FASTCALL call_13_13_7_0(SH2_struct* sh)
{
  SH2movli(sh, 13, (7<<4)+0);
}
#define init_13_13_7_1
void FASTCALL call_13_13_7_1(SH2_struct* sh)
{
  SH2movli(sh, 13, (7<<4)+1);
}
#define init_13_13_7_2
void FASTCALL call_13_13_7_2(SH2_struct* sh)
{
  SH2movli(sh, 13, (7<<4)+2);
}
#define init_13_13_7_3
void FASTCALL call_13_13_7_3(SH2_struct* sh)
{
  SH2movli(sh, 13, (7<<4)+3);
}
#define init_13_13_7_4
void FASTCALL call_13_13_7_4(SH2_struct* sh)
{
  SH2movli(sh, 13, (7<<4)+4);
}
#define init_13_13_7_5
void FASTCALL call_13_13_7_5(SH2_struct* sh)
{
  SH2movli(sh, 13, (7<<4)+5);
}
#define init_13_13_7_6
void FASTCALL call_13_13_7_6(SH2_struct* sh)
{
  SH2movli(sh, 13, (7<<4)+6);
}
#define init_13_13_7_7
void FASTCALL call_13_13_7_7(SH2_struct* sh)
{
  SH2movli(sh, 13, (7<<4)+7);
}
#define init_13_13_7_8
void FASTCALL call_13_13_7_8(SH2_struct* sh)
{
  SH2movli(sh, 13, (7<<4)+8);
}
#define init_13_13_7_9
void FASTCALL call_13_13_7_9(SH2_struct* sh)
{
  SH2movli(sh, 13, (7<<4)+9);
}
#define init_13_13_7_10
void FASTCALL call_13_13_7_10(SH2_struct* sh)
{
  SH2movli(sh, 13, (7<<4)+10);
}
#define init_13_13_7_11
void FASTCALL call_13_13_7_11(SH2_struct* sh)
{
  SH2movli(sh, 13, (7<<4)+11);
}
#define init_13_13_7_12
void FASTCALL call_13_13_7_12(SH2_struct* sh)
{
  SH2movli(sh, 13, (7<<4)+12);
}
#define init_13_13_7_13
void FASTCALL call_13_13_7_13(SH2_struct* sh)
{
  SH2movli(sh, 13, (7<<4)+13);
}
#define init_13_13_7_14
void FASTCALL call_13_13_7_14(SH2_struct* sh)
{
  SH2movli(sh, 13, (7<<4)+14);
}
#define init_13_13_7_15
void FASTCALL call_13_13_7_15(SH2_struct* sh)
{
  SH2movli(sh, 13, (7<<4)+15);
}
#define init_13_13_8_0
void FASTCALL call_13_13_8_0(SH2_struct* sh)
{
  SH2movli(sh, 13, (8<<4)+0);
}
#define init_13_13_8_1
void FASTCALL call_13_13_8_1(SH2_struct* sh)
{
  SH2movli(sh, 13, (8<<4)+1);
}
#define init_13_13_8_2
void FASTCALL call_13_13_8_2(SH2_struct* sh)
{
  SH2movli(sh, 13, (8<<4)+2);
}
#define init_13_13_8_3
void FASTCALL call_13_13_8_3(SH2_struct* sh)
{
  SH2movli(sh, 13, (8<<4)+3);
}
#define init_13_13_8_4
void FASTCALL call_13_13_8_4(SH2_struct* sh)
{
  SH2movli(sh, 13, (8<<4)+4);
}
#define init_13_13_8_5
void FASTCALL call_13_13_8_5(SH2_struct* sh)
{
  SH2movli(sh, 13, (8<<4)+5);
}
#define init_13_13_8_6
void FASTCALL call_13_13_8_6(SH2_struct* sh)
{
  SH2movli(sh, 13, (8<<4)+6);
}
#define init_13_13_8_7
void FASTCALL call_13_13_8_7(SH2_struct* sh)
{
  SH2movli(sh, 13, (8<<4)+7);
}
#define init_13_13_8_8
void FASTCALL call_13_13_8_8(SH2_struct* sh)
{
  SH2movli(sh, 13, (8<<4)+8);
}
#define init_13_13_8_9
void FASTCALL call_13_13_8_9(SH2_struct* sh)
{
  SH2movli(sh, 13, (8<<4)+9);
}
#define init_13_13_8_10
void FASTCALL call_13_13_8_10(SH2_struct* sh)
{
  SH2movli(sh, 13, (8<<4)+10);
}
#define init_13_13_8_11
void FASTCALL call_13_13_8_11(SH2_struct* sh)
{
  SH2movli(sh, 13, (8<<4)+11);
}
#define init_13_13_8_12
void FASTCALL call_13_13_8_12(SH2_struct* sh)
{
  SH2movli(sh, 13, (8<<4)+12);
}
#define init_13_13_8_13
void FASTCALL call_13_13_8_13(SH2_struct* sh)
{
  SH2movli(sh, 13, (8<<4)+13);
}
#define init_13_13_8_14
void FASTCALL call_13_13_8_14(SH2_struct* sh)
{
  SH2movli(sh, 13, (8<<4)+14);
}
#define init_13_13_8_15
void FASTCALL call_13_13_8_15(SH2_struct* sh)
{
  SH2movli(sh, 13, (8<<4)+15);
}
#define init_13_13_9_0
void FASTCALL call_13_13_9_0(SH2_struct* sh)
{
  SH2movli(sh, 13, (9<<4)+0);
}
#define init_13_13_9_1
void FASTCALL call_13_13_9_1(SH2_struct* sh)
{
  SH2movli(sh, 13, (9<<4)+1);
}
#define init_13_13_9_2
void FASTCALL call_13_13_9_2(SH2_struct* sh)
{
  SH2movli(sh, 13, (9<<4)+2);
}
#define init_13_13_9_3
void FASTCALL call_13_13_9_3(SH2_struct* sh)
{
  SH2movli(sh, 13, (9<<4)+3);
}
#define init_13_13_9_4
void FASTCALL call_13_13_9_4(SH2_struct* sh)
{
  SH2movli(sh, 13, (9<<4)+4);
}
#define init_13_13_9_5
void FASTCALL call_13_13_9_5(SH2_struct* sh)
{
  SH2movli(sh, 13, (9<<4)+5);
}
#define init_13_13_9_6
void FASTCALL call_13_13_9_6(SH2_struct* sh)
{
  SH2movli(sh, 13, (9<<4)+6);
}
#define init_13_13_9_7
void FASTCALL call_13_13_9_7(SH2_struct* sh)
{
  SH2movli(sh, 13, (9<<4)+7);
}
#define init_13_13_9_8
void FASTCALL call_13_13_9_8(SH2_struct* sh)
{
  SH2movli(sh, 13, (9<<4)+8);
}
#define init_13_13_9_9
void FASTCALL call_13_13_9_9(SH2_struct* sh)
{
  SH2movli(sh, 13, (9<<4)+9);
}
#define init_13_13_9_10
void FASTCALL call_13_13_9_10(SH2_struct* sh)
{
  SH2movli(sh, 13, (9<<4)+10);
}
#define init_13_13_9_11
void FASTCALL call_13_13_9_11(SH2_struct* sh)
{
  SH2movli(sh, 13, (9<<4)+11);
}
#define init_13_13_9_12
void FASTCALL call_13_13_9_12(SH2_struct* sh)
{
  SH2movli(sh, 13, (9<<4)+12);
}
#define init_13_13_9_13
void FASTCALL call_13_13_9_13(SH2_struct* sh)
{
  SH2movli(sh, 13, (9<<4)+13);
}
#define init_13_13_9_14
void FASTCALL call_13_13_9_14(SH2_struct* sh)
{
  SH2movli(sh, 13, (9<<4)+14);
}
#define init_13_13_9_15
void FASTCALL call_13_13_9_15(SH2_struct* sh)
{
  SH2movli(sh, 13, (9<<4)+15);
}
#define init_13_13_10_0
void FASTCALL call_13_13_10_0(SH2_struct* sh)
{
  SH2movli(sh, 13, (10<<4)+0);
}
#define init_13_13_10_1
void FASTCALL call_13_13_10_1(SH2_struct* sh)
{
  SH2movli(sh, 13, (10<<4)+1);
}
#define init_13_13_10_2
void FASTCALL call_13_13_10_2(SH2_struct* sh)
{
  SH2movli(sh, 13, (10<<4)+2);
}
#define init_13_13_10_3
void FASTCALL call_13_13_10_3(SH2_struct* sh)
{
  SH2movli(sh, 13, (10<<4)+3);
}
#define init_13_13_10_4
void FASTCALL call_13_13_10_4(SH2_struct* sh)
{
  SH2movli(sh, 13, (10<<4)+4);
}
#define init_13_13_10_5
void FASTCALL call_13_13_10_5(SH2_struct* sh)
{
  SH2movli(sh, 13, (10<<4)+5);
}
#define init_13_13_10_6
void FASTCALL call_13_13_10_6(SH2_struct* sh)
{
  SH2movli(sh, 13, (10<<4)+6);
}
#define init_13_13_10_7
void FASTCALL call_13_13_10_7(SH2_struct* sh)
{
  SH2movli(sh, 13, (10<<4)+7);
}
#define init_13_13_10_8
void FASTCALL call_13_13_10_8(SH2_struct* sh)
{
  SH2movli(sh, 13, (10<<4)+8);
}
#define init_13_13_10_9
void FASTCALL call_13_13_10_9(SH2_struct* sh)
{
  SH2movli(sh, 13, (10<<4)+9);
}
#define init_13_13_10_10
void FASTCALL call_13_13_10_10(SH2_struct* sh)
{
  SH2movli(sh, 13, (10<<4)+10);
}
#define init_13_13_10_11
void FASTCALL call_13_13_10_11(SH2_struct* sh)
{
  SH2movli(sh, 13, (10<<4)+11);
}
#define init_13_13_10_12
void FASTCALL call_13_13_10_12(SH2_struct* sh)
{
  SH2movli(sh, 13, (10<<4)+12);
}
#define init_13_13_10_13
void FASTCALL call_13_13_10_13(SH2_struct* sh)
{
  SH2movli(sh, 13, (10<<4)+13);
}
#define init_13_13_10_14
void FASTCALL call_13_13_10_14(SH2_struct* sh)
{
  SH2movli(sh, 13, (10<<4)+14);
}
#define init_13_13_10_15
void FASTCALL call_13_13_10_15(SH2_struct* sh)
{
  SH2movli(sh, 13, (10<<4)+15);
}
#define init_13_13_11_0
void FASTCALL call_13_13_11_0(SH2_struct* sh)
{
  SH2movli(sh, 13, (11<<4)+0);
}
#define init_13_13_11_1
void FASTCALL call_13_13_11_1(SH2_struct* sh)
{
  SH2movli(sh, 13, (11<<4)+1);
}
#define init_13_13_11_2
void FASTCALL call_13_13_11_2(SH2_struct* sh)
{
  SH2movli(sh, 13, (11<<4)+2);
}
#define init_13_13_11_3
void FASTCALL call_13_13_11_3(SH2_struct* sh)
{
  SH2movli(sh, 13, (11<<4)+3);
}
#define init_13_13_11_4
void FASTCALL call_13_13_11_4(SH2_struct* sh)
{
  SH2movli(sh, 13, (11<<4)+4);
}
#define init_13_13_11_5
void FASTCALL call_13_13_11_5(SH2_struct* sh)
{
  SH2movli(sh, 13, (11<<4)+5);
}
#define init_13_13_11_6
void FASTCALL call_13_13_11_6(SH2_struct* sh)
{
  SH2movli(sh, 13, (11<<4)+6);
}
#define init_13_13_11_7
void FASTCALL call_13_13_11_7(SH2_struct* sh)
{
  SH2movli(sh, 13, (11<<4)+7);
}
#define init_13_13_11_8
void FASTCALL call_13_13_11_8(SH2_struct* sh)
{
  SH2movli(sh, 13, (11<<4)+8);
}
#define init_13_13_11_9
void FASTCALL call_13_13_11_9(SH2_struct* sh)
{
  SH2movli(sh, 13, (11<<4)+9);
}
#define init_13_13_11_10
void FASTCALL call_13_13_11_10(SH2_struct* sh)
{
  SH2movli(sh, 13, (11<<4)+10);
}
#define init_13_13_11_11
void FASTCALL call_13_13_11_11(SH2_struct* sh)
{
  SH2movli(sh, 13, (11<<4)+11);
}
#define init_13_13_11_12
void FASTCALL call_13_13_11_12(SH2_struct* sh)
{
  SH2movli(sh, 13, (11<<4)+12);
}
#define init_13_13_11_13
void FASTCALL call_13_13_11_13(SH2_struct* sh)
{
  SH2movli(sh, 13, (11<<4)+13);
}
#define init_13_13_11_14
void FASTCALL call_13_13_11_14(SH2_struct* sh)
{
  SH2movli(sh, 13, (11<<4)+14);
}
#define init_13_13_11_15
void FASTCALL call_13_13_11_15(SH2_struct* sh)
{
  SH2movli(sh, 13, (11<<4)+15);
}
#define init_13_13_12_0
void FASTCALL call_13_13_12_0(SH2_struct* sh)
{
  SH2movli(sh, 13, (12<<4)+0);
}
#define init_13_13_12_1
void FASTCALL call_13_13_12_1(SH2_struct* sh)
{
  SH2movli(sh, 13, (12<<4)+1);
}
#define init_13_13_12_2
void FASTCALL call_13_13_12_2(SH2_struct* sh)
{
  SH2movli(sh, 13, (12<<4)+2);
}
#define init_13_13_12_3
void FASTCALL call_13_13_12_3(SH2_struct* sh)
{
  SH2movli(sh, 13, (12<<4)+3);
}
#define init_13_13_12_4
void FASTCALL call_13_13_12_4(SH2_struct* sh)
{
  SH2movli(sh, 13, (12<<4)+4);
}
#define init_13_13_12_5
void FASTCALL call_13_13_12_5(SH2_struct* sh)
{
  SH2movli(sh, 13, (12<<4)+5);
}
#define init_13_13_12_6
void FASTCALL call_13_13_12_6(SH2_struct* sh)
{
  SH2movli(sh, 13, (12<<4)+6);
}
#define init_13_13_12_7
void FASTCALL call_13_13_12_7(SH2_struct* sh)
{
  SH2movli(sh, 13, (12<<4)+7);
}
#define init_13_13_12_8
void FASTCALL call_13_13_12_8(SH2_struct* sh)
{
  SH2movli(sh, 13, (12<<4)+8);
}
#define init_13_13_12_9
void FASTCALL call_13_13_12_9(SH2_struct* sh)
{
  SH2movli(sh, 13, (12<<4)+9);
}
#define init_13_13_12_10
void FASTCALL call_13_13_12_10(SH2_struct* sh)
{
  SH2movli(sh, 13, (12<<4)+10);
}
#define init_13_13_12_11
void FASTCALL call_13_13_12_11(SH2_struct* sh)
{
  SH2movli(sh, 13, (12<<4)+11);
}
#define init_13_13_12_12
void FASTCALL call_13_13_12_12(SH2_struct* sh)
{
  SH2movli(sh, 13, (12<<4)+12);
}
#define init_13_13_12_13
void FASTCALL call_13_13_12_13(SH2_struct* sh)
{
  SH2movli(sh, 13, (12<<4)+13);
}
#define init_13_13_12_14
void FASTCALL call_13_13_12_14(SH2_struct* sh)
{
  SH2movli(sh, 13, (12<<4)+14);
}
#define init_13_13_12_15
void FASTCALL call_13_13_12_15(SH2_struct* sh)
{
  SH2movli(sh, 13, (12<<4)+15);
}
#define init_13_13_13_0
void FASTCALL call_13_13_13_0(SH2_struct* sh)
{
  SH2movli(sh, 13, (13<<4)+0);
}
#define init_13_13_13_1
void FASTCALL call_13_13_13_1(SH2_struct* sh)
{
  SH2movli(sh, 13, (13<<4)+1);
}
#define init_13_13_13_2
void FASTCALL call_13_13_13_2(SH2_struct* sh)
{
  SH2movli(sh, 13, (13<<4)+2);
}
#define init_13_13_13_3
void FASTCALL call_13_13_13_3(SH2_struct* sh)
{
  SH2movli(sh, 13, (13<<4)+3);
}
#define init_13_13_13_4
void FASTCALL call_13_13_13_4(SH2_struct* sh)
{
  SH2movli(sh, 13, (13<<4)+4);
}
#define init_13_13_13_5
void FASTCALL call_13_13_13_5(SH2_struct* sh)
{
  SH2movli(sh, 13, (13<<4)+5);
}
#define init_13_13_13_6
void FASTCALL call_13_13_13_6(SH2_struct* sh)
{
  SH2movli(sh, 13, (13<<4)+6);
}
#define init_13_13_13_7
void FASTCALL call_13_13_13_7(SH2_struct* sh)
{
  SH2movli(sh, 13, (13<<4)+7);
}
#define init_13_13_13_8
void FASTCALL call_13_13_13_8(SH2_struct* sh)
{
  SH2movli(sh, 13, (13<<4)+8);
}
#define init_13_13_13_9
void FASTCALL call_13_13_13_9(SH2_struct* sh)
{
  SH2movli(sh, 13, (13<<4)+9);
}
#define init_13_13_13_10
void FASTCALL call_13_13_13_10(SH2_struct* sh)
{
  SH2movli(sh, 13, (13<<4)+10);
}
#define init_13_13_13_11
void FASTCALL call_13_13_13_11(SH2_struct* sh)
{
  SH2movli(sh, 13, (13<<4)+11);
}
#define init_13_13_13_12
void FASTCALL call_13_13_13_12(SH2_struct* sh)
{
  SH2movli(sh, 13, (13<<4)+12);
}
#define init_13_13_13_13
void FASTCALL call_13_13_13_13(SH2_struct* sh)
{
  SH2movli(sh, 13, (13<<4)+13);
}
#define init_13_13_13_14
void FASTCALL call_13_13_13_14(SH2_struct* sh)
{
  SH2movli(sh, 13, (13<<4)+14);
}
#define init_13_13_13_15
void FASTCALL call_13_13_13_15(SH2_struct* sh)
{
  SH2movli(sh, 13, (13<<4)+15);
}
#define init_13_13_14_0
void FASTCALL call_13_13_14_0(SH2_struct* sh)
{
  SH2movli(sh, 13, (14<<4)+0);
}
#define init_13_13_14_1
void FASTCALL call_13_13_14_1(SH2_struct* sh)
{
  SH2movli(sh, 13, (14<<4)+1);
}
#define init_13_13_14_2
void FASTCALL call_13_13_14_2(SH2_struct* sh)
{
  SH2movli(sh, 13, (14<<4)+2);
}
#define init_13_13_14_3
void FASTCALL call_13_13_14_3(SH2_struct* sh)
{
  SH2movli(sh, 13, (14<<4)+3);
}
#define init_13_13_14_4
void FASTCALL call_13_13_14_4(SH2_struct* sh)
{
  SH2movli(sh, 13, (14<<4)+4);
}
#define init_13_13_14_5
void FASTCALL call_13_13_14_5(SH2_struct* sh)
{
  SH2movli(sh, 13, (14<<4)+5);
}
#define init_13_13_14_6
void FASTCALL call_13_13_14_6(SH2_struct* sh)
{
  SH2movli(sh, 13, (14<<4)+6);
}
#define init_13_13_14_7
void FASTCALL call_13_13_14_7(SH2_struct* sh)
{
  SH2movli(sh, 13, (14<<4)+7);
}
#define init_13_13_14_8
void FASTCALL call_13_13_14_8(SH2_struct* sh)
{
  SH2movli(sh, 13, (14<<4)+8);
}
#define init_13_13_14_9
void FASTCALL call_13_13_14_9(SH2_struct* sh)
{
  SH2movli(sh, 13, (14<<4)+9);
}
#define init_13_13_14_10
void FASTCALL call_13_13_14_10(SH2_struct* sh)
{
  SH2movli(sh, 13, (14<<4)+10);
}
#define init_13_13_14_11
void FASTCALL call_13_13_14_11(SH2_struct* sh)
{
  SH2movli(sh, 13, (14<<4)+11);
}
#define init_13_13_14_12
void FASTCALL call_13_13_14_12(SH2_struct* sh)
{
  SH2movli(sh, 13, (14<<4)+12);
}
#define init_13_13_14_13
void FASTCALL call_13_13_14_13(SH2_struct* sh)
{
  SH2movli(sh, 13, (14<<4)+13);
}
#define init_13_13_14_14
void FASTCALL call_13_13_14_14(SH2_struct* sh)
{
  SH2movli(sh, 13, (14<<4)+14);
}
#define init_13_13_14_15
void FASTCALL call_13_13_14_15(SH2_struct* sh)
{
  SH2movli(sh, 13, (14<<4)+15);
}
#define init_13_13_15_0
void FASTCALL call_13_13_15_0(SH2_struct* sh)
{
  SH2movli(sh, 13, (15<<4)+0);
}
#define init_13_13_15_1
void FASTCALL call_13_13_15_1(SH2_struct* sh)
{
  SH2movli(sh, 13, (15<<4)+1);
}
#define init_13_13_15_2
void FASTCALL call_13_13_15_2(SH2_struct* sh)
{
  SH2movli(sh, 13, (15<<4)+2);
}
#define init_13_13_15_3
void FASTCALL call_13_13_15_3(SH2_struct* sh)
{
  SH2movli(sh, 13, (15<<4)+3);
}
#define init_13_13_15_4
void FASTCALL call_13_13_15_4(SH2_struct* sh)
{
  SH2movli(sh, 13, (15<<4)+4);
}
#define init_13_13_15_5
void FASTCALL call_13_13_15_5(SH2_struct* sh)
{
  SH2movli(sh, 13, (15<<4)+5);
}
#define init_13_13_15_6
void FASTCALL call_13_13_15_6(SH2_struct* sh)
{
  SH2movli(sh, 13, (15<<4)+6);
}
#define init_13_13_15_7
void FASTCALL call_13_13_15_7(SH2_struct* sh)
{
  SH2movli(sh, 13, (15<<4)+7);
}
#define init_13_13_15_8
void FASTCALL call_13_13_15_8(SH2_struct* sh)
{
  SH2movli(sh, 13, (15<<4)+8);
}
#define init_13_13_15_9
void FASTCALL call_13_13_15_9(SH2_struct* sh)
{
  SH2movli(sh, 13, (15<<4)+9);
}
#define init_13_13_15_10
void FASTCALL call_13_13_15_10(SH2_struct* sh)
{
  SH2movli(sh, 13, (15<<4)+10);
}
#define init_13_13_15_11
void FASTCALL call_13_13_15_11(SH2_struct* sh)
{
  SH2movli(sh, 13, (15<<4)+11);
}
#define init_13_13_15_12
void FASTCALL call_13_13_15_12(SH2_struct* sh)
{
  SH2movli(sh, 13, (15<<4)+12);
}
#define init_13_13_15_13
void FASTCALL call_13_13_15_13(SH2_struct* sh)
{
  SH2movli(sh, 13, (15<<4)+13);
}
#define init_13_13_15_14
void FASTCALL call_13_13_15_14(SH2_struct* sh)
{
  SH2movli(sh, 13, (15<<4)+14);
}
#define init_13_13_15_15
void FASTCALL call_13_13_15_15(SH2_struct* sh)
{
  SH2movli(sh, 13, (15<<4)+15);
}
#define init_13_14_0_0
void FASTCALL call_13_14_0_0(SH2_struct* sh)
{
  SH2movli(sh, 14, (0<<4)+0);
}
#define init_13_14_0_1
void FASTCALL call_13_14_0_1(SH2_struct* sh)
{
  SH2movli(sh, 14, (0<<4)+1);
}
#define init_13_14_0_2
void FASTCALL call_13_14_0_2(SH2_struct* sh)
{
  SH2movli(sh, 14, (0<<4)+2);
}
#define init_13_14_0_3
void FASTCALL call_13_14_0_3(SH2_struct* sh)
{
  SH2movli(sh, 14, (0<<4)+3);
}
#define init_13_14_0_4
void FASTCALL call_13_14_0_4(SH2_struct* sh)
{
  SH2movli(sh, 14, (0<<4)+4);
}
#define init_13_14_0_5
void FASTCALL call_13_14_0_5(SH2_struct* sh)
{
  SH2movli(sh, 14, (0<<4)+5);
}
#define init_13_14_0_6
void FASTCALL call_13_14_0_6(SH2_struct* sh)
{
  SH2movli(sh, 14, (0<<4)+6);
}
#define init_13_14_0_7
void FASTCALL call_13_14_0_7(SH2_struct* sh)
{
  SH2movli(sh, 14, (0<<4)+7);
}
#define init_13_14_0_8
void FASTCALL call_13_14_0_8(SH2_struct* sh)
{
  SH2movli(sh, 14, (0<<4)+8);
}
#define init_13_14_0_9
void FASTCALL call_13_14_0_9(SH2_struct* sh)
{
  SH2movli(sh, 14, (0<<4)+9);
}
#define init_13_14_0_10
void FASTCALL call_13_14_0_10(SH2_struct* sh)
{
  SH2movli(sh, 14, (0<<4)+10);
}
#define init_13_14_0_11
void FASTCALL call_13_14_0_11(SH2_struct* sh)
{
  SH2movli(sh, 14, (0<<4)+11);
}
#define init_13_14_0_12
void FASTCALL call_13_14_0_12(SH2_struct* sh)
{
  SH2movli(sh, 14, (0<<4)+12);
}
#define init_13_14_0_13
void FASTCALL call_13_14_0_13(SH2_struct* sh)
{
  SH2movli(sh, 14, (0<<4)+13);
}
#define init_13_14_0_14
void FASTCALL call_13_14_0_14(SH2_struct* sh)
{
  SH2movli(sh, 14, (0<<4)+14);
}
#define init_13_14_0_15
void FASTCALL call_13_14_0_15(SH2_struct* sh)
{
  SH2movli(sh, 14, (0<<4)+15);
}
#define init_13_14_1_0
void FASTCALL call_13_14_1_0(SH2_struct* sh)
{
  SH2movli(sh, 14, (1<<4)+0);
}
#define init_13_14_1_1
void FASTCALL call_13_14_1_1(SH2_struct* sh)
{
  SH2movli(sh, 14, (1<<4)+1);
}
#define init_13_14_1_2
void FASTCALL call_13_14_1_2(SH2_struct* sh)
{
  SH2movli(sh, 14, (1<<4)+2);
}
#define init_13_14_1_3
void FASTCALL call_13_14_1_3(SH2_struct* sh)
{
  SH2movli(sh, 14, (1<<4)+3);
}
#define init_13_14_1_4
void FASTCALL call_13_14_1_4(SH2_struct* sh)
{
  SH2movli(sh, 14, (1<<4)+4);
}
#define init_13_14_1_5
void FASTCALL call_13_14_1_5(SH2_struct* sh)
{
  SH2movli(sh, 14, (1<<4)+5);
}
#define init_13_14_1_6
void FASTCALL call_13_14_1_6(SH2_struct* sh)
{
  SH2movli(sh, 14, (1<<4)+6);
}
#define init_13_14_1_7
void FASTCALL call_13_14_1_7(SH2_struct* sh)
{
  SH2movli(sh, 14, (1<<4)+7);
}
#define init_13_14_1_8
void FASTCALL call_13_14_1_8(SH2_struct* sh)
{
  SH2movli(sh, 14, (1<<4)+8);
}
#define init_13_14_1_9
void FASTCALL call_13_14_1_9(SH2_struct* sh)
{
  SH2movli(sh, 14, (1<<4)+9);
}
#define init_13_14_1_10
void FASTCALL call_13_14_1_10(SH2_struct* sh)
{
  SH2movli(sh, 14, (1<<4)+10);
}
#define init_13_14_1_11
void FASTCALL call_13_14_1_11(SH2_struct* sh)
{
  SH2movli(sh, 14, (1<<4)+11);
}
#define init_13_14_1_12
void FASTCALL call_13_14_1_12(SH2_struct* sh)
{
  SH2movli(sh, 14, (1<<4)+12);
}
#define init_13_14_1_13
void FASTCALL call_13_14_1_13(SH2_struct* sh)
{
  SH2movli(sh, 14, (1<<4)+13);
}
#define init_13_14_1_14
void FASTCALL call_13_14_1_14(SH2_struct* sh)
{
  SH2movli(sh, 14, (1<<4)+14);
}
#define init_13_14_1_15
void FASTCALL call_13_14_1_15(SH2_struct* sh)
{
  SH2movli(sh, 14, (1<<4)+15);
}
#define init_13_14_2_0
void FASTCALL call_13_14_2_0(SH2_struct* sh)
{
  SH2movli(sh, 14, (2<<4)+0);
}
#define init_13_14_2_1
void FASTCALL call_13_14_2_1(SH2_struct* sh)
{
  SH2movli(sh, 14, (2<<4)+1);
}
#define init_13_14_2_2
void FASTCALL call_13_14_2_2(SH2_struct* sh)
{
  SH2movli(sh, 14, (2<<4)+2);
}
#define init_13_14_2_3
void FASTCALL call_13_14_2_3(SH2_struct* sh)
{
  SH2movli(sh, 14, (2<<4)+3);
}
#define init_13_14_2_4
void FASTCALL call_13_14_2_4(SH2_struct* sh)
{
  SH2movli(sh, 14, (2<<4)+4);
}
#define init_13_14_2_5
void FASTCALL call_13_14_2_5(SH2_struct* sh)
{
  SH2movli(sh, 14, (2<<4)+5);
}
#define init_13_14_2_6
void FASTCALL call_13_14_2_6(SH2_struct* sh)
{
  SH2movli(sh, 14, (2<<4)+6);
}
#define init_13_14_2_7
void FASTCALL call_13_14_2_7(SH2_struct* sh)
{
  SH2movli(sh, 14, (2<<4)+7);
}
#define init_13_14_2_8
void FASTCALL call_13_14_2_8(SH2_struct* sh)
{
  SH2movli(sh, 14, (2<<4)+8);
}
#define init_13_14_2_9
void FASTCALL call_13_14_2_9(SH2_struct* sh)
{
  SH2movli(sh, 14, (2<<4)+9);
}
#define init_13_14_2_10
void FASTCALL call_13_14_2_10(SH2_struct* sh)
{
  SH2movli(sh, 14, (2<<4)+10);
}
#define init_13_14_2_11
void FASTCALL call_13_14_2_11(SH2_struct* sh)
{
  SH2movli(sh, 14, (2<<4)+11);
}
#define init_13_14_2_12
void FASTCALL call_13_14_2_12(SH2_struct* sh)
{
  SH2movli(sh, 14, (2<<4)+12);
}
#define init_13_14_2_13
void FASTCALL call_13_14_2_13(SH2_struct* sh)
{
  SH2movli(sh, 14, (2<<4)+13);
}
#define init_13_14_2_14
void FASTCALL call_13_14_2_14(SH2_struct* sh)
{
  SH2movli(sh, 14, (2<<4)+14);
}
#define init_13_14_2_15
void FASTCALL call_13_14_2_15(SH2_struct* sh)
{
  SH2movli(sh, 14, (2<<4)+15);
}
#define init_13_14_3_0
void FASTCALL call_13_14_3_0(SH2_struct* sh)
{
  SH2movli(sh, 14, (3<<4)+0);
}
#define init_13_14_3_1
void FASTCALL call_13_14_3_1(SH2_struct* sh)
{
  SH2movli(sh, 14, (3<<4)+1);
}
#define init_13_14_3_2
void FASTCALL call_13_14_3_2(SH2_struct* sh)
{
  SH2movli(sh, 14, (3<<4)+2);
}
#define init_13_14_3_3
void FASTCALL call_13_14_3_3(SH2_struct* sh)
{
  SH2movli(sh, 14, (3<<4)+3);
}
#define init_13_14_3_4
void FASTCALL call_13_14_3_4(SH2_struct* sh)
{
  SH2movli(sh, 14, (3<<4)+4);
}
#define init_13_14_3_5
void FASTCALL call_13_14_3_5(SH2_struct* sh)
{
  SH2movli(sh, 14, (3<<4)+5);
}
#define init_13_14_3_6
void FASTCALL call_13_14_3_6(SH2_struct* sh)
{
  SH2movli(sh, 14, (3<<4)+6);
}
#define init_13_14_3_7
void FASTCALL call_13_14_3_7(SH2_struct* sh)
{
  SH2movli(sh, 14, (3<<4)+7);
}
#define init_13_14_3_8
void FASTCALL call_13_14_3_8(SH2_struct* sh)
{
  SH2movli(sh, 14, (3<<4)+8);
}
#define init_13_14_3_9
void FASTCALL call_13_14_3_9(SH2_struct* sh)
{
  SH2movli(sh, 14, (3<<4)+9);
}
#define init_13_14_3_10
void FASTCALL call_13_14_3_10(SH2_struct* sh)
{
  SH2movli(sh, 14, (3<<4)+10);
}
#define init_13_14_3_11
void FASTCALL call_13_14_3_11(SH2_struct* sh)
{
  SH2movli(sh, 14, (3<<4)+11);
}
#define init_13_14_3_12
void FASTCALL call_13_14_3_12(SH2_struct* sh)
{
  SH2movli(sh, 14, (3<<4)+12);
}
#define init_13_14_3_13
void FASTCALL call_13_14_3_13(SH2_struct* sh)
{
  SH2movli(sh, 14, (3<<4)+13);
}
#define init_13_14_3_14
void FASTCALL call_13_14_3_14(SH2_struct* sh)
{
  SH2movli(sh, 14, (3<<4)+14);
}
#define init_13_14_3_15
void FASTCALL call_13_14_3_15(SH2_struct* sh)
{
  SH2movli(sh, 14, (3<<4)+15);
}
#define init_13_14_4_0
void FASTCALL call_13_14_4_0(SH2_struct* sh)
{
  SH2movli(sh, 14, (4<<4)+0);
}
#define init_13_14_4_1
void FASTCALL call_13_14_4_1(SH2_struct* sh)
{
  SH2movli(sh, 14, (4<<4)+1);
}
#define init_13_14_4_2
void FASTCALL call_13_14_4_2(SH2_struct* sh)
{
  SH2movli(sh, 14, (4<<4)+2);
}
#define init_13_14_4_3
void FASTCALL call_13_14_4_3(SH2_struct* sh)
{
  SH2movli(sh, 14, (4<<4)+3);
}
#define init_13_14_4_4
void FASTCALL call_13_14_4_4(SH2_struct* sh)
{
  SH2movli(sh, 14, (4<<4)+4);
}
#define init_13_14_4_5
void FASTCALL call_13_14_4_5(SH2_struct* sh)
{
  SH2movli(sh, 14, (4<<4)+5);
}
#define init_13_14_4_6
void FASTCALL call_13_14_4_6(SH2_struct* sh)
{
  SH2movli(sh, 14, (4<<4)+6);
}
#define init_13_14_4_7
void FASTCALL call_13_14_4_7(SH2_struct* sh)
{
  SH2movli(sh, 14, (4<<4)+7);
}
#define init_13_14_4_8
void FASTCALL call_13_14_4_8(SH2_struct* sh)
{
  SH2movli(sh, 14, (4<<4)+8);
}
#define init_13_14_4_9
void FASTCALL call_13_14_4_9(SH2_struct* sh)
{
  SH2movli(sh, 14, (4<<4)+9);
}
#define init_13_14_4_10
void FASTCALL call_13_14_4_10(SH2_struct* sh)
{
  SH2movli(sh, 14, (4<<4)+10);
}
#define init_13_14_4_11
void FASTCALL call_13_14_4_11(SH2_struct* sh)
{
  SH2movli(sh, 14, (4<<4)+11);
}
#define init_13_14_4_12
void FASTCALL call_13_14_4_12(SH2_struct* sh)
{
  SH2movli(sh, 14, (4<<4)+12);
}
#define init_13_14_4_13
void FASTCALL call_13_14_4_13(SH2_struct* sh)
{
  SH2movli(sh, 14, (4<<4)+13);
}
#define init_13_14_4_14
void FASTCALL call_13_14_4_14(SH2_struct* sh)
{
  SH2movli(sh, 14, (4<<4)+14);
}
#define init_13_14_4_15
void FASTCALL call_13_14_4_15(SH2_struct* sh)
{
  SH2movli(sh, 14, (4<<4)+15);
}
#define init_13_14_5_0
void FASTCALL call_13_14_5_0(SH2_struct* sh)
{
  SH2movli(sh, 14, (5<<4)+0);
}
#define init_13_14_5_1
void FASTCALL call_13_14_5_1(SH2_struct* sh)
{
  SH2movli(sh, 14, (5<<4)+1);
}
#define init_13_14_5_2
void FASTCALL call_13_14_5_2(SH2_struct* sh)
{
  SH2movli(sh, 14, (5<<4)+2);
}
#define init_13_14_5_3
void FASTCALL call_13_14_5_3(SH2_struct* sh)
{
  SH2movli(sh, 14, (5<<4)+3);
}
#define init_13_14_5_4
void FASTCALL call_13_14_5_4(SH2_struct* sh)
{
  SH2movli(sh, 14, (5<<4)+4);
}
#define init_13_14_5_5
void FASTCALL call_13_14_5_5(SH2_struct* sh)
{
  SH2movli(sh, 14, (5<<4)+5);
}
#define init_13_14_5_6
void FASTCALL call_13_14_5_6(SH2_struct* sh)
{
  SH2movli(sh, 14, (5<<4)+6);
}
#define init_13_14_5_7
void FASTCALL call_13_14_5_7(SH2_struct* sh)
{
  SH2movli(sh, 14, (5<<4)+7);
}
#define init_13_14_5_8
void FASTCALL call_13_14_5_8(SH2_struct* sh)
{
  SH2movli(sh, 14, (5<<4)+8);
}
#define init_13_14_5_9
void FASTCALL call_13_14_5_9(SH2_struct* sh)
{
  SH2movli(sh, 14, (5<<4)+9);
}
#define init_13_14_5_10
void FASTCALL call_13_14_5_10(SH2_struct* sh)
{
  SH2movli(sh, 14, (5<<4)+10);
}
#define init_13_14_5_11
void FASTCALL call_13_14_5_11(SH2_struct* sh)
{
  SH2movli(sh, 14, (5<<4)+11);
}
#define init_13_14_5_12
void FASTCALL call_13_14_5_12(SH2_struct* sh)
{
  SH2movli(sh, 14, (5<<4)+12);
}
#define init_13_14_5_13
void FASTCALL call_13_14_5_13(SH2_struct* sh)
{
  SH2movli(sh, 14, (5<<4)+13);
}
#define init_13_14_5_14
void FASTCALL call_13_14_5_14(SH2_struct* sh)
{
  SH2movli(sh, 14, (5<<4)+14);
}
#define init_13_14_5_15
void FASTCALL call_13_14_5_15(SH2_struct* sh)
{
  SH2movli(sh, 14, (5<<4)+15);
}
#define init_13_14_6_0
void FASTCALL call_13_14_6_0(SH2_struct* sh)
{
  SH2movli(sh, 14, (6<<4)+0);
}
#define init_13_14_6_1
void FASTCALL call_13_14_6_1(SH2_struct* sh)
{
  SH2movli(sh, 14, (6<<4)+1);
}
#define init_13_14_6_2
void FASTCALL call_13_14_6_2(SH2_struct* sh)
{
  SH2movli(sh, 14, (6<<4)+2);
}
#define init_13_14_6_3
void FASTCALL call_13_14_6_3(SH2_struct* sh)
{
  SH2movli(sh, 14, (6<<4)+3);
}
#define init_13_14_6_4
void FASTCALL call_13_14_6_4(SH2_struct* sh)
{
  SH2movli(sh, 14, (6<<4)+4);
}
#define init_13_14_6_5
void FASTCALL call_13_14_6_5(SH2_struct* sh)
{
  SH2movli(sh, 14, (6<<4)+5);
}
#define init_13_14_6_6
void FASTCALL call_13_14_6_6(SH2_struct* sh)
{
  SH2movli(sh, 14, (6<<4)+6);
}
#define init_13_14_6_7
void FASTCALL call_13_14_6_7(SH2_struct* sh)
{
  SH2movli(sh, 14, (6<<4)+7);
}
#define init_13_14_6_8
void FASTCALL call_13_14_6_8(SH2_struct* sh)
{
  SH2movli(sh, 14, (6<<4)+8);
}
#define init_13_14_6_9
void FASTCALL call_13_14_6_9(SH2_struct* sh)
{
  SH2movli(sh, 14, (6<<4)+9);
}
#define init_13_14_6_10
void FASTCALL call_13_14_6_10(SH2_struct* sh)
{
  SH2movli(sh, 14, (6<<4)+10);
}
#define init_13_14_6_11
void FASTCALL call_13_14_6_11(SH2_struct* sh)
{
  SH2movli(sh, 14, (6<<4)+11);
}
#define init_13_14_6_12
void FASTCALL call_13_14_6_12(SH2_struct* sh)
{
  SH2movli(sh, 14, (6<<4)+12);
}
#define init_13_14_6_13
void FASTCALL call_13_14_6_13(SH2_struct* sh)
{
  SH2movli(sh, 14, (6<<4)+13);
}
#define init_13_14_6_14
void FASTCALL call_13_14_6_14(SH2_struct* sh)
{
  SH2movli(sh, 14, (6<<4)+14);
}
#define init_13_14_6_15
void FASTCALL call_13_14_6_15(SH2_struct* sh)
{
  SH2movli(sh, 14, (6<<4)+15);
}
#define init_13_14_7_0
void FASTCALL call_13_14_7_0(SH2_struct* sh)
{
  SH2movli(sh, 14, (7<<4)+0);
}
#define init_13_14_7_1
void FASTCALL call_13_14_7_1(SH2_struct* sh)
{
  SH2movli(sh, 14, (7<<4)+1);
}
#define init_13_14_7_2
void FASTCALL call_13_14_7_2(SH2_struct* sh)
{
  SH2movli(sh, 14, (7<<4)+2);
}
#define init_13_14_7_3
void FASTCALL call_13_14_7_3(SH2_struct* sh)
{
  SH2movli(sh, 14, (7<<4)+3);
}
#define init_13_14_7_4
void FASTCALL call_13_14_7_4(SH2_struct* sh)
{
  SH2movli(sh, 14, (7<<4)+4);
}
#define init_13_14_7_5
void FASTCALL call_13_14_7_5(SH2_struct* sh)
{
  SH2movli(sh, 14, (7<<4)+5);
}
#define init_13_14_7_6
void FASTCALL call_13_14_7_6(SH2_struct* sh)
{
  SH2movli(sh, 14, (7<<4)+6);
}
#define init_13_14_7_7
void FASTCALL call_13_14_7_7(SH2_struct* sh)
{
  SH2movli(sh, 14, (7<<4)+7);
}
#define init_13_14_7_8
void FASTCALL call_13_14_7_8(SH2_struct* sh)
{
  SH2movli(sh, 14, (7<<4)+8);
}
#define init_13_14_7_9
void FASTCALL call_13_14_7_9(SH2_struct* sh)
{
  SH2movli(sh, 14, (7<<4)+9);
}
#define init_13_14_7_10
void FASTCALL call_13_14_7_10(SH2_struct* sh)
{
  SH2movli(sh, 14, (7<<4)+10);
}
#define init_13_14_7_11
void FASTCALL call_13_14_7_11(SH2_struct* sh)
{
  SH2movli(sh, 14, (7<<4)+11);
}
#define init_13_14_7_12
void FASTCALL call_13_14_7_12(SH2_struct* sh)
{
  SH2movli(sh, 14, (7<<4)+12);
}
#define init_13_14_7_13
void FASTCALL call_13_14_7_13(SH2_struct* sh)
{
  SH2movli(sh, 14, (7<<4)+13);
}
#define init_13_14_7_14
void FASTCALL call_13_14_7_14(SH2_struct* sh)
{
  SH2movli(sh, 14, (7<<4)+14);
}
#define init_13_14_7_15
void FASTCALL call_13_14_7_15(SH2_struct* sh)
{
  SH2movli(sh, 14, (7<<4)+15);
}
#define init_13_14_8_0
void FASTCALL call_13_14_8_0(SH2_struct* sh)
{
  SH2movli(sh, 14, (8<<4)+0);
}
#define init_13_14_8_1
void FASTCALL call_13_14_8_1(SH2_struct* sh)
{
  SH2movli(sh, 14, (8<<4)+1);
}
#define init_13_14_8_2
void FASTCALL call_13_14_8_2(SH2_struct* sh)
{
  SH2movli(sh, 14, (8<<4)+2);
}
#define init_13_14_8_3
void FASTCALL call_13_14_8_3(SH2_struct* sh)
{
  SH2movli(sh, 14, (8<<4)+3);
}
#define init_13_14_8_4
void FASTCALL call_13_14_8_4(SH2_struct* sh)
{
  SH2movli(sh, 14, (8<<4)+4);
}
#define init_13_14_8_5
void FASTCALL call_13_14_8_5(SH2_struct* sh)
{
  SH2movli(sh, 14, (8<<4)+5);
}
#define init_13_14_8_6
void FASTCALL call_13_14_8_6(SH2_struct* sh)
{
  SH2movli(sh, 14, (8<<4)+6);
}
#define init_13_14_8_7
void FASTCALL call_13_14_8_7(SH2_struct* sh)
{
  SH2movli(sh, 14, (8<<4)+7);
}
#define init_13_14_8_8
void FASTCALL call_13_14_8_8(SH2_struct* sh)
{
  SH2movli(sh, 14, (8<<4)+8);
}
#define init_13_14_8_9
void FASTCALL call_13_14_8_9(SH2_struct* sh)
{
  SH2movli(sh, 14, (8<<4)+9);
}
#define init_13_14_8_10
void FASTCALL call_13_14_8_10(SH2_struct* sh)
{
  SH2movli(sh, 14, (8<<4)+10);
}
#define init_13_14_8_11
void FASTCALL call_13_14_8_11(SH2_struct* sh)
{
  SH2movli(sh, 14, (8<<4)+11);
}
#define init_13_14_8_12
void FASTCALL call_13_14_8_12(SH2_struct* sh)
{
  SH2movli(sh, 14, (8<<4)+12);
}
#define init_13_14_8_13
void FASTCALL call_13_14_8_13(SH2_struct* sh)
{
  SH2movli(sh, 14, (8<<4)+13);
}
#define init_13_14_8_14
void FASTCALL call_13_14_8_14(SH2_struct* sh)
{
  SH2movli(sh, 14, (8<<4)+14);
}
#define init_13_14_8_15
void FASTCALL call_13_14_8_15(SH2_struct* sh)
{
  SH2movli(sh, 14, (8<<4)+15);
}
#define init_13_14_9_0
void FASTCALL call_13_14_9_0(SH2_struct* sh)
{
  SH2movli(sh, 14, (9<<4)+0);
}
#define init_13_14_9_1
void FASTCALL call_13_14_9_1(SH2_struct* sh)
{
  SH2movli(sh, 14, (9<<4)+1);
}
#define init_13_14_9_2
void FASTCALL call_13_14_9_2(SH2_struct* sh)
{
  SH2movli(sh, 14, (9<<4)+2);
}
#define init_13_14_9_3
void FASTCALL call_13_14_9_3(SH2_struct* sh)
{
  SH2movli(sh, 14, (9<<4)+3);
}
#define init_13_14_9_4
void FASTCALL call_13_14_9_4(SH2_struct* sh)
{
  SH2movli(sh, 14, (9<<4)+4);
}
#define init_13_14_9_5
void FASTCALL call_13_14_9_5(SH2_struct* sh)
{
  SH2movli(sh, 14, (9<<4)+5);
}
#define init_13_14_9_6
void FASTCALL call_13_14_9_6(SH2_struct* sh)
{
  SH2movli(sh, 14, (9<<4)+6);
}
#define init_13_14_9_7
void FASTCALL call_13_14_9_7(SH2_struct* sh)
{
  SH2movli(sh, 14, (9<<4)+7);
}
#define init_13_14_9_8
void FASTCALL call_13_14_9_8(SH2_struct* sh)
{
  SH2movli(sh, 14, (9<<4)+8);
}
#define init_13_14_9_9
void FASTCALL call_13_14_9_9(SH2_struct* sh)
{
  SH2movli(sh, 14, (9<<4)+9);
}
#define init_13_14_9_10
void FASTCALL call_13_14_9_10(SH2_struct* sh)
{
  SH2movli(sh, 14, (9<<4)+10);
}
#define init_13_14_9_11
void FASTCALL call_13_14_9_11(SH2_struct* sh)
{
  SH2movli(sh, 14, (9<<4)+11);
}
#define init_13_14_9_12
void FASTCALL call_13_14_9_12(SH2_struct* sh)
{
  SH2movli(sh, 14, (9<<4)+12);
}
#define init_13_14_9_13
void FASTCALL call_13_14_9_13(SH2_struct* sh)
{
  SH2movli(sh, 14, (9<<4)+13);
}
#define init_13_14_9_14
void FASTCALL call_13_14_9_14(SH2_struct* sh)
{
  SH2movli(sh, 14, (9<<4)+14);
}
#define init_13_14_9_15
void FASTCALL call_13_14_9_15(SH2_struct* sh)
{
  SH2movli(sh, 14, (9<<4)+15);
}
#define init_13_14_10_0
void FASTCALL call_13_14_10_0(SH2_struct* sh)
{
  SH2movli(sh, 14, (10<<4)+0);
}
#define init_13_14_10_1
void FASTCALL call_13_14_10_1(SH2_struct* sh)
{
  SH2movli(sh, 14, (10<<4)+1);
}
#define init_13_14_10_2
void FASTCALL call_13_14_10_2(SH2_struct* sh)
{
  SH2movli(sh, 14, (10<<4)+2);
}
#define init_13_14_10_3
void FASTCALL call_13_14_10_3(SH2_struct* sh)
{
  SH2movli(sh, 14, (10<<4)+3);
}
#define init_13_14_10_4
void FASTCALL call_13_14_10_4(SH2_struct* sh)
{
  SH2movli(sh, 14, (10<<4)+4);
}
#define init_13_14_10_5
void FASTCALL call_13_14_10_5(SH2_struct* sh)
{
  SH2movli(sh, 14, (10<<4)+5);
}
#define init_13_14_10_6
void FASTCALL call_13_14_10_6(SH2_struct* sh)
{
  SH2movli(sh, 14, (10<<4)+6);
}
#define init_13_14_10_7
void FASTCALL call_13_14_10_7(SH2_struct* sh)
{
  SH2movli(sh, 14, (10<<4)+7);
}
#define init_13_14_10_8
void FASTCALL call_13_14_10_8(SH2_struct* sh)
{
  SH2movli(sh, 14, (10<<4)+8);
}
#define init_13_14_10_9
void FASTCALL call_13_14_10_9(SH2_struct* sh)
{
  SH2movli(sh, 14, (10<<4)+9);
}
#define init_13_14_10_10
void FASTCALL call_13_14_10_10(SH2_struct* sh)
{
  SH2movli(sh, 14, (10<<4)+10);
}
#define init_13_14_10_11
void FASTCALL call_13_14_10_11(SH2_struct* sh)
{
  SH2movli(sh, 14, (10<<4)+11);
}
#define init_13_14_10_12
void FASTCALL call_13_14_10_12(SH2_struct* sh)
{
  SH2movli(sh, 14, (10<<4)+12);
}
#define init_13_14_10_13
void FASTCALL call_13_14_10_13(SH2_struct* sh)
{
  SH2movli(sh, 14, (10<<4)+13);
}
#define init_13_14_10_14
void FASTCALL call_13_14_10_14(SH2_struct* sh)
{
  SH2movli(sh, 14, (10<<4)+14);
}
#define init_13_14_10_15
void FASTCALL call_13_14_10_15(SH2_struct* sh)
{
  SH2movli(sh, 14, (10<<4)+15);
}
#define init_13_14_11_0
void FASTCALL call_13_14_11_0(SH2_struct* sh)
{
  SH2movli(sh, 14, (11<<4)+0);
}
#define init_13_14_11_1
void FASTCALL call_13_14_11_1(SH2_struct* sh)
{
  SH2movli(sh, 14, (11<<4)+1);
}
#define init_13_14_11_2
void FASTCALL call_13_14_11_2(SH2_struct* sh)
{
  SH2movli(sh, 14, (11<<4)+2);
}
#define init_13_14_11_3
void FASTCALL call_13_14_11_3(SH2_struct* sh)
{
  SH2movli(sh, 14, (11<<4)+3);
}
#define init_13_14_11_4
void FASTCALL call_13_14_11_4(SH2_struct* sh)
{
  SH2movli(sh, 14, (11<<4)+4);
}
#define init_13_14_11_5
void FASTCALL call_13_14_11_5(SH2_struct* sh)
{
  SH2movli(sh, 14, (11<<4)+5);
}
#define init_13_14_11_6
void FASTCALL call_13_14_11_6(SH2_struct* sh)
{
  SH2movli(sh, 14, (11<<4)+6);
}
#define init_13_14_11_7
void FASTCALL call_13_14_11_7(SH2_struct* sh)
{
  SH2movli(sh, 14, (11<<4)+7);
}
#define init_13_14_11_8
void FASTCALL call_13_14_11_8(SH2_struct* sh)
{
  SH2movli(sh, 14, (11<<4)+8);
}
#define init_13_14_11_9
void FASTCALL call_13_14_11_9(SH2_struct* sh)
{
  SH2movli(sh, 14, (11<<4)+9);
}
#define init_13_14_11_10
void FASTCALL call_13_14_11_10(SH2_struct* sh)
{
  SH2movli(sh, 14, (11<<4)+10);
}
#define init_13_14_11_11
void FASTCALL call_13_14_11_11(SH2_struct* sh)
{
  SH2movli(sh, 14, (11<<4)+11);
}
#define init_13_14_11_12
void FASTCALL call_13_14_11_12(SH2_struct* sh)
{
  SH2movli(sh, 14, (11<<4)+12);
}
#define init_13_14_11_13
void FASTCALL call_13_14_11_13(SH2_struct* sh)
{
  SH2movli(sh, 14, (11<<4)+13);
}
#define init_13_14_11_14
void FASTCALL call_13_14_11_14(SH2_struct* sh)
{
  SH2movli(sh, 14, (11<<4)+14);
}
#define init_13_14_11_15
void FASTCALL call_13_14_11_15(SH2_struct* sh)
{
  SH2movli(sh, 14, (11<<4)+15);
}
#define init_13_14_12_0
void FASTCALL call_13_14_12_0(SH2_struct* sh)
{
  SH2movli(sh, 14, (12<<4)+0);
}
#define init_13_14_12_1
void FASTCALL call_13_14_12_1(SH2_struct* sh)
{
  SH2movli(sh, 14, (12<<4)+1);
}
#define init_13_14_12_2
void FASTCALL call_13_14_12_2(SH2_struct* sh)
{
  SH2movli(sh, 14, (12<<4)+2);
}
#define init_13_14_12_3
void FASTCALL call_13_14_12_3(SH2_struct* sh)
{
  SH2movli(sh, 14, (12<<4)+3);
}
#define init_13_14_12_4
void FASTCALL call_13_14_12_4(SH2_struct* sh)
{
  SH2movli(sh, 14, (12<<4)+4);
}
#define init_13_14_12_5
void FASTCALL call_13_14_12_5(SH2_struct* sh)
{
  SH2movli(sh, 14, (12<<4)+5);
}
#define init_13_14_12_6
void FASTCALL call_13_14_12_6(SH2_struct* sh)
{
  SH2movli(sh, 14, (12<<4)+6);
}
#define init_13_14_12_7
void FASTCALL call_13_14_12_7(SH2_struct* sh)
{
  SH2movli(sh, 14, (12<<4)+7);
}
#define init_13_14_12_8
void FASTCALL call_13_14_12_8(SH2_struct* sh)
{
  SH2movli(sh, 14, (12<<4)+8);
}
#define init_13_14_12_9
void FASTCALL call_13_14_12_9(SH2_struct* sh)
{
  SH2movli(sh, 14, (12<<4)+9);
}
#define init_13_14_12_10
void FASTCALL call_13_14_12_10(SH2_struct* sh)
{
  SH2movli(sh, 14, (12<<4)+10);
}
#define init_13_14_12_11
void FASTCALL call_13_14_12_11(SH2_struct* sh)
{
  SH2movli(sh, 14, (12<<4)+11);
}
#define init_13_14_12_12
void FASTCALL call_13_14_12_12(SH2_struct* sh)
{
  SH2movli(sh, 14, (12<<4)+12);
}
#define init_13_14_12_13
void FASTCALL call_13_14_12_13(SH2_struct* sh)
{
  SH2movli(sh, 14, (12<<4)+13);
}
#define init_13_14_12_14
void FASTCALL call_13_14_12_14(SH2_struct* sh)
{
  SH2movli(sh, 14, (12<<4)+14);
}
#define init_13_14_12_15
void FASTCALL call_13_14_12_15(SH2_struct* sh)
{
  SH2movli(sh, 14, (12<<4)+15);
}
#define init_13_14_13_0
void FASTCALL call_13_14_13_0(SH2_struct* sh)
{
  SH2movli(sh, 14, (13<<4)+0);
}
#define init_13_14_13_1
void FASTCALL call_13_14_13_1(SH2_struct* sh)
{
  SH2movli(sh, 14, (13<<4)+1);
}
#define init_13_14_13_2
void FASTCALL call_13_14_13_2(SH2_struct* sh)
{
  SH2movli(sh, 14, (13<<4)+2);
}
#define init_13_14_13_3
void FASTCALL call_13_14_13_3(SH2_struct* sh)
{
  SH2movli(sh, 14, (13<<4)+3);
}
#define init_13_14_13_4
void FASTCALL call_13_14_13_4(SH2_struct* sh)
{
  SH2movli(sh, 14, (13<<4)+4);
}
#define init_13_14_13_5
void FASTCALL call_13_14_13_5(SH2_struct* sh)
{
  SH2movli(sh, 14, (13<<4)+5);
}
#define init_13_14_13_6
void FASTCALL call_13_14_13_6(SH2_struct* sh)
{
  SH2movli(sh, 14, (13<<4)+6);
}
#define init_13_14_13_7
void FASTCALL call_13_14_13_7(SH2_struct* sh)
{
  SH2movli(sh, 14, (13<<4)+7);
}
#define init_13_14_13_8
void FASTCALL call_13_14_13_8(SH2_struct* sh)
{
  SH2movli(sh, 14, (13<<4)+8);
}
#define init_13_14_13_9
void FASTCALL call_13_14_13_9(SH2_struct* sh)
{
  SH2movli(sh, 14, (13<<4)+9);
}
#define init_13_14_13_10
void FASTCALL call_13_14_13_10(SH2_struct* sh)
{
  SH2movli(sh, 14, (13<<4)+10);
}
#define init_13_14_13_11
void FASTCALL call_13_14_13_11(SH2_struct* sh)
{
  SH2movli(sh, 14, (13<<4)+11);
}
#define init_13_14_13_12
void FASTCALL call_13_14_13_12(SH2_struct* sh)
{
  SH2movli(sh, 14, (13<<4)+12);
}
#define init_13_14_13_13
void FASTCALL call_13_14_13_13(SH2_struct* sh)
{
  SH2movli(sh, 14, (13<<4)+13);
}
#define init_13_14_13_14
void FASTCALL call_13_14_13_14(SH2_struct* sh)
{
  SH2movli(sh, 14, (13<<4)+14);
}
#define init_13_14_13_15
void FASTCALL call_13_14_13_15(SH2_struct* sh)
{
  SH2movli(sh, 14, (13<<4)+15);
}
#define init_13_14_14_0
void FASTCALL call_13_14_14_0(SH2_struct* sh)
{
  SH2movli(sh, 14, (14<<4)+0);
}
#define init_13_14_14_1
void FASTCALL call_13_14_14_1(SH2_struct* sh)
{
  SH2movli(sh, 14, (14<<4)+1);
}
#define init_13_14_14_2
void FASTCALL call_13_14_14_2(SH2_struct* sh)
{
  SH2movli(sh, 14, (14<<4)+2);
}
#define init_13_14_14_3
void FASTCALL call_13_14_14_3(SH2_struct* sh)
{
  SH2movli(sh, 14, (14<<4)+3);
}
#define init_13_14_14_4
void FASTCALL call_13_14_14_4(SH2_struct* sh)
{
  SH2movli(sh, 14, (14<<4)+4);
}
#define init_13_14_14_5
void FASTCALL call_13_14_14_5(SH2_struct* sh)
{
  SH2movli(sh, 14, (14<<4)+5);
}
#define init_13_14_14_6
void FASTCALL call_13_14_14_6(SH2_struct* sh)
{
  SH2movli(sh, 14, (14<<4)+6);
}
#define init_13_14_14_7
void FASTCALL call_13_14_14_7(SH2_struct* sh)
{
  SH2movli(sh, 14, (14<<4)+7);
}
#define init_13_14_14_8
void FASTCALL call_13_14_14_8(SH2_struct* sh)
{
  SH2movli(sh, 14, (14<<4)+8);
}
#define init_13_14_14_9
void FASTCALL call_13_14_14_9(SH2_struct* sh)
{
  SH2movli(sh, 14, (14<<4)+9);
}
#define init_13_14_14_10
void FASTCALL call_13_14_14_10(SH2_struct* sh)
{
  SH2movli(sh, 14, (14<<4)+10);
}
#define init_13_14_14_11
void FASTCALL call_13_14_14_11(SH2_struct* sh)
{
  SH2movli(sh, 14, (14<<4)+11);
}
#define init_13_14_14_12
void FASTCALL call_13_14_14_12(SH2_struct* sh)
{
  SH2movli(sh, 14, (14<<4)+12);
}
#define init_13_14_14_13
void FASTCALL call_13_14_14_13(SH2_struct* sh)
{
  SH2movli(sh, 14, (14<<4)+13);
}
#define init_13_14_14_14
void FASTCALL call_13_14_14_14(SH2_struct* sh)
{
  SH2movli(sh, 14, (14<<4)+14);
}
#define init_13_14_14_15
void FASTCALL call_13_14_14_15(SH2_struct* sh)
{
  SH2movli(sh, 14, (14<<4)+15);
}
#define init_13_14_15_0
void FASTCALL call_13_14_15_0(SH2_struct* sh)
{
  SH2movli(sh, 14, (15<<4)+0);
}
#define init_13_14_15_1
void FASTCALL call_13_14_15_1(SH2_struct* sh)
{
  SH2movli(sh, 14, (15<<4)+1);
}
#define init_13_14_15_2
void FASTCALL call_13_14_15_2(SH2_struct* sh)
{
  SH2movli(sh, 14, (15<<4)+2);
}
#define init_13_14_15_3
void FASTCALL call_13_14_15_3(SH2_struct* sh)
{
  SH2movli(sh, 14, (15<<4)+3);
}
#define init_13_14_15_4
void FASTCALL call_13_14_15_4(SH2_struct* sh)
{
  SH2movli(sh, 14, (15<<4)+4);
}
#define init_13_14_15_5
void FASTCALL call_13_14_15_5(SH2_struct* sh)
{
  SH2movli(sh, 14, (15<<4)+5);
}
#define init_13_14_15_6
void FASTCALL call_13_14_15_6(SH2_struct* sh)
{
  SH2movli(sh, 14, (15<<4)+6);
}
#define init_13_14_15_7
void FASTCALL call_13_14_15_7(SH2_struct* sh)
{
  SH2movli(sh, 14, (15<<4)+7);
}
#define init_13_14_15_8
void FASTCALL call_13_14_15_8(SH2_struct* sh)
{
  SH2movli(sh, 14, (15<<4)+8);
}
#define init_13_14_15_9
void FASTCALL call_13_14_15_9(SH2_struct* sh)
{
  SH2movli(sh, 14, (15<<4)+9);
}
#define init_13_14_15_10
void FASTCALL call_13_14_15_10(SH2_struct* sh)
{
  SH2movli(sh, 14, (15<<4)+10);
}
#define init_13_14_15_11
void FASTCALL call_13_14_15_11(SH2_struct* sh)
{
  SH2movli(sh, 14, (15<<4)+11);
}
#define init_13_14_15_12
void FASTCALL call_13_14_15_12(SH2_struct* sh)
{
  SH2movli(sh, 14, (15<<4)+12);
}
#define init_13_14_15_13
void FASTCALL call_13_14_15_13(SH2_struct* sh)
{
  SH2movli(sh, 14, (15<<4)+13);
}
#define init_13_14_15_14
void FASTCALL call_13_14_15_14(SH2_struct* sh)
{
  SH2movli(sh, 14, (15<<4)+14);
}
#define init_13_14_15_15
void FASTCALL call_13_14_15_15(SH2_struct* sh)
{
  SH2movli(sh, 14, (15<<4)+15);
}
#define init_13_15_0_0
void FASTCALL call_13_15_0_0(SH2_struct* sh)
{
  SH2movli(sh, 15, (0<<4)+0);
}
#define init_13_15_0_1
void FASTCALL call_13_15_0_1(SH2_struct* sh)
{
  SH2movli(sh, 15, (0<<4)+1);
}
#define init_13_15_0_2
void FASTCALL call_13_15_0_2(SH2_struct* sh)
{
  SH2movli(sh, 15, (0<<4)+2);
}
#define init_13_15_0_3
void FASTCALL call_13_15_0_3(SH2_struct* sh)
{
  SH2movli(sh, 15, (0<<4)+3);
}
#define init_13_15_0_4
void FASTCALL call_13_15_0_4(SH2_struct* sh)
{
  SH2movli(sh, 15, (0<<4)+4);
}
#define init_13_15_0_5
void FASTCALL call_13_15_0_5(SH2_struct* sh)
{
  SH2movli(sh, 15, (0<<4)+5);
}
#define init_13_15_0_6
void FASTCALL call_13_15_0_6(SH2_struct* sh)
{
  SH2movli(sh, 15, (0<<4)+6);
}
#define init_13_15_0_7
void FASTCALL call_13_15_0_7(SH2_struct* sh)
{
  SH2movli(sh, 15, (0<<4)+7);
}
#define init_13_15_0_8
void FASTCALL call_13_15_0_8(SH2_struct* sh)
{
  SH2movli(sh, 15, (0<<4)+8);
}
#define init_13_15_0_9
void FASTCALL call_13_15_0_9(SH2_struct* sh)
{
  SH2movli(sh, 15, (0<<4)+9);
}
#define init_13_15_0_10
void FASTCALL call_13_15_0_10(SH2_struct* sh)
{
  SH2movli(sh, 15, (0<<4)+10);
}
#define init_13_15_0_11
void FASTCALL call_13_15_0_11(SH2_struct* sh)
{
  SH2movli(sh, 15, (0<<4)+11);
}
#define init_13_15_0_12
void FASTCALL call_13_15_0_12(SH2_struct* sh)
{
  SH2movli(sh, 15, (0<<4)+12);
}
#define init_13_15_0_13
void FASTCALL call_13_15_0_13(SH2_struct* sh)
{
  SH2movli(sh, 15, (0<<4)+13);
}
#define init_13_15_0_14
void FASTCALL call_13_15_0_14(SH2_struct* sh)
{
  SH2movli(sh, 15, (0<<4)+14);
}
#define init_13_15_0_15
void FASTCALL call_13_15_0_15(SH2_struct* sh)
{
  SH2movli(sh, 15, (0<<4)+15);
}
#define init_13_15_1_0
void FASTCALL call_13_15_1_0(SH2_struct* sh)
{
  SH2movli(sh, 15, (1<<4)+0);
}
#define init_13_15_1_1
void FASTCALL call_13_15_1_1(SH2_struct* sh)
{
  SH2movli(sh, 15, (1<<4)+1);
}
#define init_13_15_1_2
void FASTCALL call_13_15_1_2(SH2_struct* sh)
{
  SH2movli(sh, 15, (1<<4)+2);
}
#define init_13_15_1_3
void FASTCALL call_13_15_1_3(SH2_struct* sh)
{
  SH2movli(sh, 15, (1<<4)+3);
}
#define init_13_15_1_4
void FASTCALL call_13_15_1_4(SH2_struct* sh)
{
  SH2movli(sh, 15, (1<<4)+4);
}
#define init_13_15_1_5
void FASTCALL call_13_15_1_5(SH2_struct* sh)
{
  SH2movli(sh, 15, (1<<4)+5);
}
#define init_13_15_1_6
void FASTCALL call_13_15_1_6(SH2_struct* sh)
{
  SH2movli(sh, 15, (1<<4)+6);
}
#define init_13_15_1_7
void FASTCALL call_13_15_1_7(SH2_struct* sh)
{
  SH2movli(sh, 15, (1<<4)+7);
}
#define init_13_15_1_8
void FASTCALL call_13_15_1_8(SH2_struct* sh)
{
  SH2movli(sh, 15, (1<<4)+8);
}
#define init_13_15_1_9
void FASTCALL call_13_15_1_9(SH2_struct* sh)
{
  SH2movli(sh, 15, (1<<4)+9);
}
#define init_13_15_1_10
void FASTCALL call_13_15_1_10(SH2_struct* sh)
{
  SH2movli(sh, 15, (1<<4)+10);
}
#define init_13_15_1_11
void FASTCALL call_13_15_1_11(SH2_struct* sh)
{
  SH2movli(sh, 15, (1<<4)+11);
}
#define init_13_15_1_12
void FASTCALL call_13_15_1_12(SH2_struct* sh)
{
  SH2movli(sh, 15, (1<<4)+12);
}
#define init_13_15_1_13
void FASTCALL call_13_15_1_13(SH2_struct* sh)
{
  SH2movli(sh, 15, (1<<4)+13);
}
#define init_13_15_1_14
void FASTCALL call_13_15_1_14(SH2_struct* sh)
{
  SH2movli(sh, 15, (1<<4)+14);
}
#define init_13_15_1_15
void FASTCALL call_13_15_1_15(SH2_struct* sh)
{
  SH2movli(sh, 15, (1<<4)+15);
}
#define init_13_15_2_0
void FASTCALL call_13_15_2_0(SH2_struct* sh)
{
  SH2movli(sh, 15, (2<<4)+0);
}
#define init_13_15_2_1
void FASTCALL call_13_15_2_1(SH2_struct* sh)
{
  SH2movli(sh, 15, (2<<4)+1);
}
#define init_13_15_2_2
void FASTCALL call_13_15_2_2(SH2_struct* sh)
{
  SH2movli(sh, 15, (2<<4)+2);
}
#define init_13_15_2_3
void FASTCALL call_13_15_2_3(SH2_struct* sh)
{
  SH2movli(sh, 15, (2<<4)+3);
}
#define init_13_15_2_4
void FASTCALL call_13_15_2_4(SH2_struct* sh)
{
  SH2movli(sh, 15, (2<<4)+4);
}
#define init_13_15_2_5
void FASTCALL call_13_15_2_5(SH2_struct* sh)
{
  SH2movli(sh, 15, (2<<4)+5);
}
#define init_13_15_2_6
void FASTCALL call_13_15_2_6(SH2_struct* sh)
{
  SH2movli(sh, 15, (2<<4)+6);
}
#define init_13_15_2_7
void FASTCALL call_13_15_2_7(SH2_struct* sh)
{
  SH2movli(sh, 15, (2<<4)+7);
}
#define init_13_15_2_8
void FASTCALL call_13_15_2_8(SH2_struct* sh)
{
  SH2movli(sh, 15, (2<<4)+8);
}
#define init_13_15_2_9
void FASTCALL call_13_15_2_9(SH2_struct* sh)
{
  SH2movli(sh, 15, (2<<4)+9);
}
#define init_13_15_2_10
void FASTCALL call_13_15_2_10(SH2_struct* sh)
{
  SH2movli(sh, 15, (2<<4)+10);
}
#define init_13_15_2_11
void FASTCALL call_13_15_2_11(SH2_struct* sh)
{
  SH2movli(sh, 15, (2<<4)+11);
}
#define init_13_15_2_12
void FASTCALL call_13_15_2_12(SH2_struct* sh)
{
  SH2movli(sh, 15, (2<<4)+12);
}
#define init_13_15_2_13
void FASTCALL call_13_15_2_13(SH2_struct* sh)
{
  SH2movli(sh, 15, (2<<4)+13);
}
#define init_13_15_2_14
void FASTCALL call_13_15_2_14(SH2_struct* sh)
{
  SH2movli(sh, 15, (2<<4)+14);
}
#define init_13_15_2_15
void FASTCALL call_13_15_2_15(SH2_struct* sh)
{
  SH2movli(sh, 15, (2<<4)+15);
}
#define init_13_15_3_0
void FASTCALL call_13_15_3_0(SH2_struct* sh)
{
  SH2movli(sh, 15, (3<<4)+0);
}
#define init_13_15_3_1
void FASTCALL call_13_15_3_1(SH2_struct* sh)
{
  SH2movli(sh, 15, (3<<4)+1);
}
#define init_13_15_3_2
void FASTCALL call_13_15_3_2(SH2_struct* sh)
{
  SH2movli(sh, 15, (3<<4)+2);
}
#define init_13_15_3_3
void FASTCALL call_13_15_3_3(SH2_struct* sh)
{
  SH2movli(sh, 15, (3<<4)+3);
}
#define init_13_15_3_4
void FASTCALL call_13_15_3_4(SH2_struct* sh)
{
  SH2movli(sh, 15, (3<<4)+4);
}
#define init_13_15_3_5
void FASTCALL call_13_15_3_5(SH2_struct* sh)
{
  SH2movli(sh, 15, (3<<4)+5);
}
#define init_13_15_3_6
void FASTCALL call_13_15_3_6(SH2_struct* sh)
{
  SH2movli(sh, 15, (3<<4)+6);
}
#define init_13_15_3_7
void FASTCALL call_13_15_3_7(SH2_struct* sh)
{
  SH2movli(sh, 15, (3<<4)+7);
}
#define init_13_15_3_8
void FASTCALL call_13_15_3_8(SH2_struct* sh)
{
  SH2movli(sh, 15, (3<<4)+8);
}
#define init_13_15_3_9
void FASTCALL call_13_15_3_9(SH2_struct* sh)
{
  SH2movli(sh, 15, (3<<4)+9);
}
#define init_13_15_3_10
void FASTCALL call_13_15_3_10(SH2_struct* sh)
{
  SH2movli(sh, 15, (3<<4)+10);
}
#define init_13_15_3_11
void FASTCALL call_13_15_3_11(SH2_struct* sh)
{
  SH2movli(sh, 15, (3<<4)+11);
}
#define init_13_15_3_12
void FASTCALL call_13_15_3_12(SH2_struct* sh)
{
  SH2movli(sh, 15, (3<<4)+12);
}
#define init_13_15_3_13
void FASTCALL call_13_15_3_13(SH2_struct* sh)
{
  SH2movli(sh, 15, (3<<4)+13);
}
#define init_13_15_3_14
void FASTCALL call_13_15_3_14(SH2_struct* sh)
{
  SH2movli(sh, 15, (3<<4)+14);
}
#define init_13_15_3_15
void FASTCALL call_13_15_3_15(SH2_struct* sh)
{
  SH2movli(sh, 15, (3<<4)+15);
}
#define init_13_15_4_0
void FASTCALL call_13_15_4_0(SH2_struct* sh)
{
  SH2movli(sh, 15, (4<<4)+0);
}
#define init_13_15_4_1
void FASTCALL call_13_15_4_1(SH2_struct* sh)
{
  SH2movli(sh, 15, (4<<4)+1);
}
#define init_13_15_4_2
void FASTCALL call_13_15_4_2(SH2_struct* sh)
{
  SH2movli(sh, 15, (4<<4)+2);
}
#define init_13_15_4_3
void FASTCALL call_13_15_4_3(SH2_struct* sh)
{
  SH2movli(sh, 15, (4<<4)+3);
}
#define init_13_15_4_4
void FASTCALL call_13_15_4_4(SH2_struct* sh)
{
  SH2movli(sh, 15, (4<<4)+4);
}
#define init_13_15_4_5
void FASTCALL call_13_15_4_5(SH2_struct* sh)
{
  SH2movli(sh, 15, (4<<4)+5);
}
#define init_13_15_4_6
void FASTCALL call_13_15_4_6(SH2_struct* sh)
{
  SH2movli(sh, 15, (4<<4)+6);
}
#define init_13_15_4_7
void FASTCALL call_13_15_4_7(SH2_struct* sh)
{
  SH2movli(sh, 15, (4<<4)+7);
}
#define init_13_15_4_8
void FASTCALL call_13_15_4_8(SH2_struct* sh)
{
  SH2movli(sh, 15, (4<<4)+8);
}
#define init_13_15_4_9
void FASTCALL call_13_15_4_9(SH2_struct* sh)
{
  SH2movli(sh, 15, (4<<4)+9);
}
#define init_13_15_4_10
void FASTCALL call_13_15_4_10(SH2_struct* sh)
{
  SH2movli(sh, 15, (4<<4)+10);
}
#define init_13_15_4_11
void FASTCALL call_13_15_4_11(SH2_struct* sh)
{
  SH2movli(sh, 15, (4<<4)+11);
}
#define init_13_15_4_12
void FASTCALL call_13_15_4_12(SH2_struct* sh)
{
  SH2movli(sh, 15, (4<<4)+12);
}
#define init_13_15_4_13
void FASTCALL call_13_15_4_13(SH2_struct* sh)
{
  SH2movli(sh, 15, (4<<4)+13);
}
#define init_13_15_4_14
void FASTCALL call_13_15_4_14(SH2_struct* sh)
{
  SH2movli(sh, 15, (4<<4)+14);
}
#define init_13_15_4_15
void FASTCALL call_13_15_4_15(SH2_struct* sh)
{
  SH2movli(sh, 15, (4<<4)+15);
}
#define init_13_15_5_0
void FASTCALL call_13_15_5_0(SH2_struct* sh)
{
  SH2movli(sh, 15, (5<<4)+0);
}
#define init_13_15_5_1
void FASTCALL call_13_15_5_1(SH2_struct* sh)
{
  SH2movli(sh, 15, (5<<4)+1);
}
#define init_13_15_5_2
void FASTCALL call_13_15_5_2(SH2_struct* sh)
{
  SH2movli(sh, 15, (5<<4)+2);
}
#define init_13_15_5_3
void FASTCALL call_13_15_5_3(SH2_struct* sh)
{
  SH2movli(sh, 15, (5<<4)+3);
}
#define init_13_15_5_4
void FASTCALL call_13_15_5_4(SH2_struct* sh)
{
  SH2movli(sh, 15, (5<<4)+4);
}
#define init_13_15_5_5
void FASTCALL call_13_15_5_5(SH2_struct* sh)
{
  SH2movli(sh, 15, (5<<4)+5);
}
#define init_13_15_5_6
void FASTCALL call_13_15_5_6(SH2_struct* sh)
{
  SH2movli(sh, 15, (5<<4)+6);
}
#define init_13_15_5_7
void FASTCALL call_13_15_5_7(SH2_struct* sh)
{
  SH2movli(sh, 15, (5<<4)+7);
}
#define init_13_15_5_8
void FASTCALL call_13_15_5_8(SH2_struct* sh)
{
  SH2movli(sh, 15, (5<<4)+8);
}
#define init_13_15_5_9
void FASTCALL call_13_15_5_9(SH2_struct* sh)
{
  SH2movli(sh, 15, (5<<4)+9);
}
#define init_13_15_5_10
void FASTCALL call_13_15_5_10(SH2_struct* sh)
{
  SH2movli(sh, 15, (5<<4)+10);
}
#define init_13_15_5_11
void FASTCALL call_13_15_5_11(SH2_struct* sh)
{
  SH2movli(sh, 15, (5<<4)+11);
}
#define init_13_15_5_12
void FASTCALL call_13_15_5_12(SH2_struct* sh)
{
  SH2movli(sh, 15, (5<<4)+12);
}
#define init_13_15_5_13
void FASTCALL call_13_15_5_13(SH2_struct* sh)
{
  SH2movli(sh, 15, (5<<4)+13);
}
#define init_13_15_5_14
void FASTCALL call_13_15_5_14(SH2_struct* sh)
{
  SH2movli(sh, 15, (5<<4)+14);
}
#define init_13_15_5_15
void FASTCALL call_13_15_5_15(SH2_struct* sh)
{
  SH2movli(sh, 15, (5<<4)+15);
}
#define init_13_15_6_0
void FASTCALL call_13_15_6_0(SH2_struct* sh)
{
  SH2movli(sh, 15, (6<<4)+0);
}
#define init_13_15_6_1
void FASTCALL call_13_15_6_1(SH2_struct* sh)
{
  SH2movli(sh, 15, (6<<4)+1);
}
#define init_13_15_6_2
void FASTCALL call_13_15_6_2(SH2_struct* sh)
{
  SH2movli(sh, 15, (6<<4)+2);
}
#define init_13_15_6_3
void FASTCALL call_13_15_6_3(SH2_struct* sh)
{
  SH2movli(sh, 15, (6<<4)+3);
}
#define init_13_15_6_4
void FASTCALL call_13_15_6_4(SH2_struct* sh)
{
  SH2movli(sh, 15, (6<<4)+4);
}
#define init_13_15_6_5
void FASTCALL call_13_15_6_5(SH2_struct* sh)
{
  SH2movli(sh, 15, (6<<4)+5);
}
#define init_13_15_6_6
void FASTCALL call_13_15_6_6(SH2_struct* sh)
{
  SH2movli(sh, 15, (6<<4)+6);
}
#define init_13_15_6_7
void FASTCALL call_13_15_6_7(SH2_struct* sh)
{
  SH2movli(sh, 15, (6<<4)+7);
}
#define init_13_15_6_8
void FASTCALL call_13_15_6_8(SH2_struct* sh)
{
  SH2movli(sh, 15, (6<<4)+8);
}
#define init_13_15_6_9
void FASTCALL call_13_15_6_9(SH2_struct* sh)
{
  SH2movli(sh, 15, (6<<4)+9);
}
#define init_13_15_6_10
void FASTCALL call_13_15_6_10(SH2_struct* sh)
{
  SH2movli(sh, 15, (6<<4)+10);
}
#define init_13_15_6_11
void FASTCALL call_13_15_6_11(SH2_struct* sh)
{
  SH2movli(sh, 15, (6<<4)+11);
}
#define init_13_15_6_12
void FASTCALL call_13_15_6_12(SH2_struct* sh)
{
  SH2movli(sh, 15, (6<<4)+12);
}
#define init_13_15_6_13
void FASTCALL call_13_15_6_13(SH2_struct* sh)
{
  SH2movli(sh, 15, (6<<4)+13);
}
#define init_13_15_6_14
void FASTCALL call_13_15_6_14(SH2_struct* sh)
{
  SH2movli(sh, 15, (6<<4)+14);
}
#define init_13_15_6_15
void FASTCALL call_13_15_6_15(SH2_struct* sh)
{
  SH2movli(sh, 15, (6<<4)+15);
}
#define init_13_15_7_0
void FASTCALL call_13_15_7_0(SH2_struct* sh)
{
  SH2movli(sh, 15, (7<<4)+0);
}
#define init_13_15_7_1
void FASTCALL call_13_15_7_1(SH2_struct* sh)
{
  SH2movli(sh, 15, (7<<4)+1);
}
#define init_13_15_7_2
void FASTCALL call_13_15_7_2(SH2_struct* sh)
{
  SH2movli(sh, 15, (7<<4)+2);
}
#define init_13_15_7_3
void FASTCALL call_13_15_7_3(SH2_struct* sh)
{
  SH2movli(sh, 15, (7<<4)+3);
}
#define init_13_15_7_4
void FASTCALL call_13_15_7_4(SH2_struct* sh)
{
  SH2movli(sh, 15, (7<<4)+4);
}
#define init_13_15_7_5
void FASTCALL call_13_15_7_5(SH2_struct* sh)
{
  SH2movli(sh, 15, (7<<4)+5);
}
#define init_13_15_7_6
void FASTCALL call_13_15_7_6(SH2_struct* sh)
{
  SH2movli(sh, 15, (7<<4)+6);
}
#define init_13_15_7_7
void FASTCALL call_13_15_7_7(SH2_struct* sh)
{
  SH2movli(sh, 15, (7<<4)+7);
}
#define init_13_15_7_8
void FASTCALL call_13_15_7_8(SH2_struct* sh)
{
  SH2movli(sh, 15, (7<<4)+8);
}
#define init_13_15_7_9
void FASTCALL call_13_15_7_9(SH2_struct* sh)
{
  SH2movli(sh, 15, (7<<4)+9);
}
#define init_13_15_7_10
void FASTCALL call_13_15_7_10(SH2_struct* sh)
{
  SH2movli(sh, 15, (7<<4)+10);
}
#define init_13_15_7_11
void FASTCALL call_13_15_7_11(SH2_struct* sh)
{
  SH2movli(sh, 15, (7<<4)+11);
}
#define init_13_15_7_12
void FASTCALL call_13_15_7_12(SH2_struct* sh)
{
  SH2movli(sh, 15, (7<<4)+12);
}
#define init_13_15_7_13
void FASTCALL call_13_15_7_13(SH2_struct* sh)
{
  SH2movli(sh, 15, (7<<4)+13);
}
#define init_13_15_7_14
void FASTCALL call_13_15_7_14(SH2_struct* sh)
{
  SH2movli(sh, 15, (7<<4)+14);
}
#define init_13_15_7_15
void FASTCALL call_13_15_7_15(SH2_struct* sh)
{
  SH2movli(sh, 15, (7<<4)+15);
}
#define init_13_15_8_0
void FASTCALL call_13_15_8_0(SH2_struct* sh)
{
  SH2movli(sh, 15, (8<<4)+0);
}
#define init_13_15_8_1
void FASTCALL call_13_15_8_1(SH2_struct* sh)
{
  SH2movli(sh, 15, (8<<4)+1);
}
#define init_13_15_8_2
void FASTCALL call_13_15_8_2(SH2_struct* sh)
{
  SH2movli(sh, 15, (8<<4)+2);
}
#define init_13_15_8_3
void FASTCALL call_13_15_8_3(SH2_struct* sh)
{
  SH2movli(sh, 15, (8<<4)+3);
}
#define init_13_15_8_4
void FASTCALL call_13_15_8_4(SH2_struct* sh)
{
  SH2movli(sh, 15, (8<<4)+4);
}
#define init_13_15_8_5
void FASTCALL call_13_15_8_5(SH2_struct* sh)
{
  SH2movli(sh, 15, (8<<4)+5);
}
#define init_13_15_8_6
void FASTCALL call_13_15_8_6(SH2_struct* sh)
{
  SH2movli(sh, 15, (8<<4)+6);
}
#define init_13_15_8_7
void FASTCALL call_13_15_8_7(SH2_struct* sh)
{
  SH2movli(sh, 15, (8<<4)+7);
}
#define init_13_15_8_8
void FASTCALL call_13_15_8_8(SH2_struct* sh)
{
  SH2movli(sh, 15, (8<<4)+8);
}
#define init_13_15_8_9
void FASTCALL call_13_15_8_9(SH2_struct* sh)
{
  SH2movli(sh, 15, (8<<4)+9);
}
#define init_13_15_8_10
void FASTCALL call_13_15_8_10(SH2_struct* sh)
{
  SH2movli(sh, 15, (8<<4)+10);
}
#define init_13_15_8_11
void FASTCALL call_13_15_8_11(SH2_struct* sh)
{
  SH2movli(sh, 15, (8<<4)+11);
}
#define init_13_15_8_12
void FASTCALL call_13_15_8_12(SH2_struct* sh)
{
  SH2movli(sh, 15, (8<<4)+12);
}
#define init_13_15_8_13
void FASTCALL call_13_15_8_13(SH2_struct* sh)
{
  SH2movli(sh, 15, (8<<4)+13);
}
#define init_13_15_8_14
void FASTCALL call_13_15_8_14(SH2_struct* sh)
{
  SH2movli(sh, 15, (8<<4)+14);
}
#define init_13_15_8_15
void FASTCALL call_13_15_8_15(SH2_struct* sh)
{
  SH2movli(sh, 15, (8<<4)+15);
}
#define init_13_15_9_0
void FASTCALL call_13_15_9_0(SH2_struct* sh)
{
  SH2movli(sh, 15, (9<<4)+0);
}
#define init_13_15_9_1
void FASTCALL call_13_15_9_1(SH2_struct* sh)
{
  SH2movli(sh, 15, (9<<4)+1);
}
#define init_13_15_9_2
void FASTCALL call_13_15_9_2(SH2_struct* sh)
{
  SH2movli(sh, 15, (9<<4)+2);
}
#define init_13_15_9_3
void FASTCALL call_13_15_9_3(SH2_struct* sh)
{
  SH2movli(sh, 15, (9<<4)+3);
}
#define init_13_15_9_4
void FASTCALL call_13_15_9_4(SH2_struct* sh)
{
  SH2movli(sh, 15, (9<<4)+4);
}
#define init_13_15_9_5
void FASTCALL call_13_15_9_5(SH2_struct* sh)
{
  SH2movli(sh, 15, (9<<4)+5);
}
#define init_13_15_9_6
void FASTCALL call_13_15_9_6(SH2_struct* sh)
{
  SH2movli(sh, 15, (9<<4)+6);
}
#define init_13_15_9_7
void FASTCALL call_13_15_9_7(SH2_struct* sh)
{
  SH2movli(sh, 15, (9<<4)+7);
}
#define init_13_15_9_8
void FASTCALL call_13_15_9_8(SH2_struct* sh)
{
  SH2movli(sh, 15, (9<<4)+8);
}
#define init_13_15_9_9
void FASTCALL call_13_15_9_9(SH2_struct* sh)
{
  SH2movli(sh, 15, (9<<4)+9);
}
#define init_13_15_9_10
void FASTCALL call_13_15_9_10(SH2_struct* sh)
{
  SH2movli(sh, 15, (9<<4)+10);
}
#define init_13_15_9_11
void FASTCALL call_13_15_9_11(SH2_struct* sh)
{
  SH2movli(sh, 15, (9<<4)+11);
}
#define init_13_15_9_12
void FASTCALL call_13_15_9_12(SH2_struct* sh)
{
  SH2movli(sh, 15, (9<<4)+12);
}
#define init_13_15_9_13
void FASTCALL call_13_15_9_13(SH2_struct* sh)
{
  SH2movli(sh, 15, (9<<4)+13);
}
#define init_13_15_9_14
void FASTCALL call_13_15_9_14(SH2_struct* sh)
{
  SH2movli(sh, 15, (9<<4)+14);
}
#define init_13_15_9_15
void FASTCALL call_13_15_9_15(SH2_struct* sh)
{
  SH2movli(sh, 15, (9<<4)+15);
}
#define init_13_15_10_0
void FASTCALL call_13_15_10_0(SH2_struct* sh)
{
  SH2movli(sh, 15, (10<<4)+0);
}
#define init_13_15_10_1
void FASTCALL call_13_15_10_1(SH2_struct* sh)
{
  SH2movli(sh, 15, (10<<4)+1);
}
#define init_13_15_10_2
void FASTCALL call_13_15_10_2(SH2_struct* sh)
{
  SH2movli(sh, 15, (10<<4)+2);
}
#define init_13_15_10_3
void FASTCALL call_13_15_10_3(SH2_struct* sh)
{
  SH2movli(sh, 15, (10<<4)+3);
}
#define init_13_15_10_4
void FASTCALL call_13_15_10_4(SH2_struct* sh)
{
  SH2movli(sh, 15, (10<<4)+4);
}
#define init_13_15_10_5
void FASTCALL call_13_15_10_5(SH2_struct* sh)
{
  SH2movli(sh, 15, (10<<4)+5);
}
#define init_13_15_10_6
void FASTCALL call_13_15_10_6(SH2_struct* sh)
{
  SH2movli(sh, 15, (10<<4)+6);
}
#define init_13_15_10_7
void FASTCALL call_13_15_10_7(SH2_struct* sh)
{
  SH2movli(sh, 15, (10<<4)+7);
}
#define init_13_15_10_8
void FASTCALL call_13_15_10_8(SH2_struct* sh)
{
  SH2movli(sh, 15, (10<<4)+8);
}
#define init_13_15_10_9
void FASTCALL call_13_15_10_9(SH2_struct* sh)
{
  SH2movli(sh, 15, (10<<4)+9);
}
#define init_13_15_10_10
void FASTCALL call_13_15_10_10(SH2_struct* sh)
{
  SH2movli(sh, 15, (10<<4)+10);
}
#define init_13_15_10_11
void FASTCALL call_13_15_10_11(SH2_struct* sh)
{
  SH2movli(sh, 15, (10<<4)+11);
}
#define init_13_15_10_12
void FASTCALL call_13_15_10_12(SH2_struct* sh)
{
  SH2movli(sh, 15, (10<<4)+12);
}
#define init_13_15_10_13
void FASTCALL call_13_15_10_13(SH2_struct* sh)
{
  SH2movli(sh, 15, (10<<4)+13);
}
#define init_13_15_10_14
void FASTCALL call_13_15_10_14(SH2_struct* sh)
{
  SH2movli(sh, 15, (10<<4)+14);
}
#define init_13_15_10_15
void FASTCALL call_13_15_10_15(SH2_struct* sh)
{
  SH2movli(sh, 15, (10<<4)+15);
}
#define init_13_15_11_0
void FASTCALL call_13_15_11_0(SH2_struct* sh)
{
  SH2movli(sh, 15, (11<<4)+0);
}
#define init_13_15_11_1
void FASTCALL call_13_15_11_1(SH2_struct* sh)
{
  SH2movli(sh, 15, (11<<4)+1);
}
#define init_13_15_11_2
void FASTCALL call_13_15_11_2(SH2_struct* sh)
{
  SH2movli(sh, 15, (11<<4)+2);
}
#define init_13_15_11_3
void FASTCALL call_13_15_11_3(SH2_struct* sh)
{
  SH2movli(sh, 15, (11<<4)+3);
}
#define init_13_15_11_4
void FASTCALL call_13_15_11_4(SH2_struct* sh)
{
  SH2movli(sh, 15, (11<<4)+4);
}
#define init_13_15_11_5
void FASTCALL call_13_15_11_5(SH2_struct* sh)
{
  SH2movli(sh, 15, (11<<4)+5);
}
#define init_13_15_11_6
void FASTCALL call_13_15_11_6(SH2_struct* sh)
{
  SH2movli(sh, 15, (11<<4)+6);
}
#define init_13_15_11_7
void FASTCALL call_13_15_11_7(SH2_struct* sh)
{
  SH2movli(sh, 15, (11<<4)+7);
}
#define init_13_15_11_8
void FASTCALL call_13_15_11_8(SH2_struct* sh)
{
  SH2movli(sh, 15, (11<<4)+8);
}
#define init_13_15_11_9
void FASTCALL call_13_15_11_9(SH2_struct* sh)
{
  SH2movli(sh, 15, (11<<4)+9);
}
#define init_13_15_11_10
void FASTCALL call_13_15_11_10(SH2_struct* sh)
{
  SH2movli(sh, 15, (11<<4)+10);
}
#define init_13_15_11_11
void FASTCALL call_13_15_11_11(SH2_struct* sh)
{
  SH2movli(sh, 15, (11<<4)+11);
}
#define init_13_15_11_12
void FASTCALL call_13_15_11_12(SH2_struct* sh)
{
  SH2movli(sh, 15, (11<<4)+12);
}
#define init_13_15_11_13
void FASTCALL call_13_15_11_13(SH2_struct* sh)
{
  SH2movli(sh, 15, (11<<4)+13);
}
#define init_13_15_11_14
void FASTCALL call_13_15_11_14(SH2_struct* sh)
{
  SH2movli(sh, 15, (11<<4)+14);
}
#define init_13_15_11_15
void FASTCALL call_13_15_11_15(SH2_struct* sh)
{
  SH2movli(sh, 15, (11<<4)+15);
}
#define init_13_15_12_0
void FASTCALL call_13_15_12_0(SH2_struct* sh)
{
  SH2movli(sh, 15, (12<<4)+0);
}
#define init_13_15_12_1
void FASTCALL call_13_15_12_1(SH2_struct* sh)
{
  SH2movli(sh, 15, (12<<4)+1);
}
#define init_13_15_12_2
void FASTCALL call_13_15_12_2(SH2_struct* sh)
{
  SH2movli(sh, 15, (12<<4)+2);
}
#define init_13_15_12_3
void FASTCALL call_13_15_12_3(SH2_struct* sh)
{
  SH2movli(sh, 15, (12<<4)+3);
}
#define init_13_15_12_4
void FASTCALL call_13_15_12_4(SH2_struct* sh)
{
  SH2movli(sh, 15, (12<<4)+4);
}
#define init_13_15_12_5
void FASTCALL call_13_15_12_5(SH2_struct* sh)
{
  SH2movli(sh, 15, (12<<4)+5);
}
#define init_13_15_12_6
void FASTCALL call_13_15_12_6(SH2_struct* sh)
{
  SH2movli(sh, 15, (12<<4)+6);
}
#define init_13_15_12_7
void FASTCALL call_13_15_12_7(SH2_struct* sh)
{
  SH2movli(sh, 15, (12<<4)+7);
}
#define init_13_15_12_8
void FASTCALL call_13_15_12_8(SH2_struct* sh)
{
  SH2movli(sh, 15, (12<<4)+8);
}
#define init_13_15_12_9
void FASTCALL call_13_15_12_9(SH2_struct* sh)
{
  SH2movli(sh, 15, (12<<4)+9);
}
#define init_13_15_12_10
void FASTCALL call_13_15_12_10(SH2_struct* sh)
{
  SH2movli(sh, 15, (12<<4)+10);
}
#define init_13_15_12_11
void FASTCALL call_13_15_12_11(SH2_struct* sh)
{
  SH2movli(sh, 15, (12<<4)+11);
}
#define init_13_15_12_12
void FASTCALL call_13_15_12_12(SH2_struct* sh)
{
  SH2movli(sh, 15, (12<<4)+12);
}
#define init_13_15_12_13
void FASTCALL call_13_15_12_13(SH2_struct* sh)
{
  SH2movli(sh, 15, (12<<4)+13);
}
#define init_13_15_12_14
void FASTCALL call_13_15_12_14(SH2_struct* sh)
{
  SH2movli(sh, 15, (12<<4)+14);
}
#define init_13_15_12_15
void FASTCALL call_13_15_12_15(SH2_struct* sh)
{
  SH2movli(sh, 15, (12<<4)+15);
}
#define init_13_15_13_0
void FASTCALL call_13_15_13_0(SH2_struct* sh)
{
  SH2movli(sh, 15, (13<<4)+0);
}
#define init_13_15_13_1
void FASTCALL call_13_15_13_1(SH2_struct* sh)
{
  SH2movli(sh, 15, (13<<4)+1);
}
#define init_13_15_13_2
void FASTCALL call_13_15_13_2(SH2_struct* sh)
{
  SH2movli(sh, 15, (13<<4)+2);
}
#define init_13_15_13_3
void FASTCALL call_13_15_13_3(SH2_struct* sh)
{
  SH2movli(sh, 15, (13<<4)+3);
}
#define init_13_15_13_4
void FASTCALL call_13_15_13_4(SH2_struct* sh)
{
  SH2movli(sh, 15, (13<<4)+4);
}
#define init_13_15_13_5
void FASTCALL call_13_15_13_5(SH2_struct* sh)
{
  SH2movli(sh, 15, (13<<4)+5);
}
#define init_13_15_13_6
void FASTCALL call_13_15_13_6(SH2_struct* sh)
{
  SH2movli(sh, 15, (13<<4)+6);
}
#define init_13_15_13_7
void FASTCALL call_13_15_13_7(SH2_struct* sh)
{
  SH2movli(sh, 15, (13<<4)+7);
}
#define init_13_15_13_8
void FASTCALL call_13_15_13_8(SH2_struct* sh)
{
  SH2movli(sh, 15, (13<<4)+8);
}
#define init_13_15_13_9
void FASTCALL call_13_15_13_9(SH2_struct* sh)
{
  SH2movli(sh, 15, (13<<4)+9);
}
#define init_13_15_13_10
void FASTCALL call_13_15_13_10(SH2_struct* sh)
{
  SH2movli(sh, 15, (13<<4)+10);
}
#define init_13_15_13_11
void FASTCALL call_13_15_13_11(SH2_struct* sh)
{
  SH2movli(sh, 15, (13<<4)+11);
}
#define init_13_15_13_12
void FASTCALL call_13_15_13_12(SH2_struct* sh)
{
  SH2movli(sh, 15, (13<<4)+12);
}
#define init_13_15_13_13
void FASTCALL call_13_15_13_13(SH2_struct* sh)
{
  SH2movli(sh, 15, (13<<4)+13);
}
#define init_13_15_13_14
void FASTCALL call_13_15_13_14(SH2_struct* sh)
{
  SH2movli(sh, 15, (13<<4)+14);
}
#define init_13_15_13_15
void FASTCALL call_13_15_13_15(SH2_struct* sh)
{
  SH2movli(sh, 15, (13<<4)+15);
}
#define init_13_15_14_0
void FASTCALL call_13_15_14_0(SH2_struct* sh)
{
  SH2movli(sh, 15, (14<<4)+0);
}
#define init_13_15_14_1
void FASTCALL call_13_15_14_1(SH2_struct* sh)
{
  SH2movli(sh, 15, (14<<4)+1);
}
#define init_13_15_14_2
void FASTCALL call_13_15_14_2(SH2_struct* sh)
{
  SH2movli(sh, 15, (14<<4)+2);
}
#define init_13_15_14_3
void FASTCALL call_13_15_14_3(SH2_struct* sh)
{
  SH2movli(sh, 15, (14<<4)+3);
}
#define init_13_15_14_4
void FASTCALL call_13_15_14_4(SH2_struct* sh)
{
  SH2movli(sh, 15, (14<<4)+4);
}
#define init_13_15_14_5
void FASTCALL call_13_15_14_5(SH2_struct* sh)
{
  SH2movli(sh, 15, (14<<4)+5);
}
#define init_13_15_14_6
void FASTCALL call_13_15_14_6(SH2_struct* sh)
{
  SH2movli(sh, 15, (14<<4)+6);
}
#define init_13_15_14_7
void FASTCALL call_13_15_14_7(SH2_struct* sh)
{
  SH2movli(sh, 15, (14<<4)+7);
}
#define init_13_15_14_8
void FASTCALL call_13_15_14_8(SH2_struct* sh)
{
  SH2movli(sh, 15, (14<<4)+8);
}
#define init_13_15_14_9
void FASTCALL call_13_15_14_9(SH2_struct* sh)
{
  SH2movli(sh, 15, (14<<4)+9);
}
#define init_13_15_14_10
void FASTCALL call_13_15_14_10(SH2_struct* sh)
{
  SH2movli(sh, 15, (14<<4)+10);
}
#define init_13_15_14_11
void FASTCALL call_13_15_14_11(SH2_struct* sh)
{
  SH2movli(sh, 15, (14<<4)+11);
}
#define init_13_15_14_12
void FASTCALL call_13_15_14_12(SH2_struct* sh)
{
  SH2movli(sh, 15, (14<<4)+12);
}
#define init_13_15_14_13
void FASTCALL call_13_15_14_13(SH2_struct* sh)
{
  SH2movli(sh, 15, (14<<4)+13);
}
#define init_13_15_14_14
void FASTCALL call_13_15_14_14(SH2_struct* sh)
{
  SH2movli(sh, 15, (14<<4)+14);
}
#define init_13_15_14_15
void FASTCALL call_13_15_14_15(SH2_struct* sh)
{
  SH2movli(sh, 15, (14<<4)+15);
}
#define init_13_15_15_0
void FASTCALL call_13_15_15_0(SH2_struct* sh)
{
  SH2movli(sh, 15, (15<<4)+0);
}
#define init_13_15_15_1
void FASTCALL call_13_15_15_1(SH2_struct* sh)
{
  SH2movli(sh, 15, (15<<4)+1);
}
#define init_13_15_15_2
void FASTCALL call_13_15_15_2(SH2_struct* sh)
{
  SH2movli(sh, 15, (15<<4)+2);
}
#define init_13_15_15_3
void FASTCALL call_13_15_15_3(SH2_struct* sh)
{
  SH2movli(sh, 15, (15<<4)+3);
}
#define init_13_15_15_4
void FASTCALL call_13_15_15_4(SH2_struct* sh)
{
  SH2movli(sh, 15, (15<<4)+4);
}
#define init_13_15_15_5
void FASTCALL call_13_15_15_5(SH2_struct* sh)
{
  SH2movli(sh, 15, (15<<4)+5);
}
#define init_13_15_15_6
void FASTCALL call_13_15_15_6(SH2_struct* sh)
{
  SH2movli(sh, 15, (15<<4)+6);
}
#define init_13_15_15_7
void FASTCALL call_13_15_15_7(SH2_struct* sh)
{
  SH2movli(sh, 15, (15<<4)+7);
}
#define init_13_15_15_8
void FASTCALL call_13_15_15_8(SH2_struct* sh)
{
  SH2movli(sh, 15, (15<<4)+8);
}
#define init_13_15_15_9
void FASTCALL call_13_15_15_9(SH2_struct* sh)
{
  SH2movli(sh, 15, (15<<4)+9);
}
#define init_13_15_15_10
void FASTCALL call_13_15_15_10(SH2_struct* sh)
{
  SH2movli(sh, 15, (15<<4)+10);
}
#define init_13_15_15_11
void FASTCALL call_13_15_15_11(SH2_struct* sh)
{
  SH2movli(sh, 15, (15<<4)+11);
}
#define init_13_15_15_12
void FASTCALL call_13_15_15_12(SH2_struct* sh)
{
  SH2movli(sh, 15, (15<<4)+12);
}
#define init_13_15_15_13
void FASTCALL call_13_15_15_13(SH2_struct* sh)
{
  SH2movli(sh, 15, (15<<4)+13);
}
#define init_13_15_15_14
void FASTCALL call_13_15_15_14(SH2_struct* sh)
{
  SH2movli(sh, 15, (15<<4)+14);
}
#define init_13_15_15_15
void FASTCALL call_13_15_15_15(SH2_struct* sh)
{
  SH2movli(sh, 15, (15<<4)+15);
}
#define init_14_0_0_0
void FASTCALL call_14_0_0_0(SH2_struct* sh)
{
  SH2movi(sh, 0, (0<<4)+0);
}
#define init_14_0_0_1
void FASTCALL call_14_0_0_1(SH2_struct* sh)
{
  SH2movi(sh, 0, (0<<4)+1);
}
#define init_14_0_0_2
void FASTCALL call_14_0_0_2(SH2_struct* sh)
{
  SH2movi(sh, 0, (0<<4)+2);
}
#define init_14_0_0_3
void FASTCALL call_14_0_0_3(SH2_struct* sh)
{
  SH2movi(sh, 0, (0<<4)+3);
}
#define init_14_0_0_4
void FASTCALL call_14_0_0_4(SH2_struct* sh)
{
  SH2movi(sh, 0, (0<<4)+4);
}
#define init_14_0_0_5
void FASTCALL call_14_0_0_5(SH2_struct* sh)
{
  SH2movi(sh, 0, (0<<4)+5);
}
#define init_14_0_0_6
void FASTCALL call_14_0_0_6(SH2_struct* sh)
{
  SH2movi(sh, 0, (0<<4)+6);
}
#define init_14_0_0_7
void FASTCALL call_14_0_0_7(SH2_struct* sh)
{
  SH2movi(sh, 0, (0<<4)+7);
}
#define init_14_0_0_8
void FASTCALL call_14_0_0_8(SH2_struct* sh)
{
  SH2movi(sh, 0, (0<<4)+8);
}
#define init_14_0_0_9
void FASTCALL call_14_0_0_9(SH2_struct* sh)
{
  SH2movi(sh, 0, (0<<4)+9);
}
#define init_14_0_0_10
void FASTCALL call_14_0_0_10(SH2_struct* sh)
{
  SH2movi(sh, 0, (0<<4)+10);
}
#define init_14_0_0_11
void FASTCALL call_14_0_0_11(SH2_struct* sh)
{
  SH2movi(sh, 0, (0<<4)+11);
}
#define init_14_0_0_12
void FASTCALL call_14_0_0_12(SH2_struct* sh)
{
  SH2movi(sh, 0, (0<<4)+12);
}
#define init_14_0_0_13
void FASTCALL call_14_0_0_13(SH2_struct* sh)
{
  SH2movi(sh, 0, (0<<4)+13);
}
#define init_14_0_0_14
void FASTCALL call_14_0_0_14(SH2_struct* sh)
{
  SH2movi(sh, 0, (0<<4)+14);
}
#define init_14_0_0_15
void FASTCALL call_14_0_0_15(SH2_struct* sh)
{
  SH2movi(sh, 0, (0<<4)+15);
}
#define init_14_0_1_0
void FASTCALL call_14_0_1_0(SH2_struct* sh)
{
  SH2movi(sh, 0, (1<<4)+0);
}
#define init_14_0_1_1
void FASTCALL call_14_0_1_1(SH2_struct* sh)
{
  SH2movi(sh, 0, (1<<4)+1);
}
#define init_14_0_1_2
void FASTCALL call_14_0_1_2(SH2_struct* sh)
{
  SH2movi(sh, 0, (1<<4)+2);
}
#define init_14_0_1_3
void FASTCALL call_14_0_1_3(SH2_struct* sh)
{
  SH2movi(sh, 0, (1<<4)+3);
}
#define init_14_0_1_4
void FASTCALL call_14_0_1_4(SH2_struct* sh)
{
  SH2movi(sh, 0, (1<<4)+4);
}
#define init_14_0_1_5
void FASTCALL call_14_0_1_5(SH2_struct* sh)
{
  SH2movi(sh, 0, (1<<4)+5);
}
#define init_14_0_1_6
void FASTCALL call_14_0_1_6(SH2_struct* sh)
{
  SH2movi(sh, 0, (1<<4)+6);
}
#define init_14_0_1_7
void FASTCALL call_14_0_1_7(SH2_struct* sh)
{
  SH2movi(sh, 0, (1<<4)+7);
}
#define init_14_0_1_8
void FASTCALL call_14_0_1_8(SH2_struct* sh)
{
  SH2movi(sh, 0, (1<<4)+8);
}
#define init_14_0_1_9
void FASTCALL call_14_0_1_9(SH2_struct* sh)
{
  SH2movi(sh, 0, (1<<4)+9);
}
#define init_14_0_1_10
void FASTCALL call_14_0_1_10(SH2_struct* sh)
{
  SH2movi(sh, 0, (1<<4)+10);
}
#define init_14_0_1_11
void FASTCALL call_14_0_1_11(SH2_struct* sh)
{
  SH2movi(sh, 0, (1<<4)+11);
}
#define init_14_0_1_12
void FASTCALL call_14_0_1_12(SH2_struct* sh)
{
  SH2movi(sh, 0, (1<<4)+12);
}
#define init_14_0_1_13
void FASTCALL call_14_0_1_13(SH2_struct* sh)
{
  SH2movi(sh, 0, (1<<4)+13);
}
#define init_14_0_1_14
void FASTCALL call_14_0_1_14(SH2_struct* sh)
{
  SH2movi(sh, 0, (1<<4)+14);
}
#define init_14_0_1_15
void FASTCALL call_14_0_1_15(SH2_struct* sh)
{
  SH2movi(sh, 0, (1<<4)+15);
}
#define init_14_0_2_0
void FASTCALL call_14_0_2_0(SH2_struct* sh)
{
  SH2movi(sh, 0, (2<<4)+0);
}
#define init_14_0_2_1
void FASTCALL call_14_0_2_1(SH2_struct* sh)
{
  SH2movi(sh, 0, (2<<4)+1);
}
#define init_14_0_2_2
void FASTCALL call_14_0_2_2(SH2_struct* sh)
{
  SH2movi(sh, 0, (2<<4)+2);
}
#define init_14_0_2_3
void FASTCALL call_14_0_2_3(SH2_struct* sh)
{
  SH2movi(sh, 0, (2<<4)+3);
}
#define init_14_0_2_4
void FASTCALL call_14_0_2_4(SH2_struct* sh)
{
  SH2movi(sh, 0, (2<<4)+4);
}
#define init_14_0_2_5
void FASTCALL call_14_0_2_5(SH2_struct* sh)
{
  SH2movi(sh, 0, (2<<4)+5);
}
#define init_14_0_2_6
void FASTCALL call_14_0_2_6(SH2_struct* sh)
{
  SH2movi(sh, 0, (2<<4)+6);
}
#define init_14_0_2_7
void FASTCALL call_14_0_2_7(SH2_struct* sh)
{
  SH2movi(sh, 0, (2<<4)+7);
}
#define init_14_0_2_8
void FASTCALL call_14_0_2_8(SH2_struct* sh)
{
  SH2movi(sh, 0, (2<<4)+8);
}
#define init_14_0_2_9
void FASTCALL call_14_0_2_9(SH2_struct* sh)
{
  SH2movi(sh, 0, (2<<4)+9);
}
#define init_14_0_2_10
void FASTCALL call_14_0_2_10(SH2_struct* sh)
{
  SH2movi(sh, 0, (2<<4)+10);
}
#define init_14_0_2_11
void FASTCALL call_14_0_2_11(SH2_struct* sh)
{
  SH2movi(sh, 0, (2<<4)+11);
}
#define init_14_0_2_12
void FASTCALL call_14_0_2_12(SH2_struct* sh)
{
  SH2movi(sh, 0, (2<<4)+12);
}
#define init_14_0_2_13
void FASTCALL call_14_0_2_13(SH2_struct* sh)
{
  SH2movi(sh, 0, (2<<4)+13);
}
#define init_14_0_2_14
void FASTCALL call_14_0_2_14(SH2_struct* sh)
{
  SH2movi(sh, 0, (2<<4)+14);
}
#define init_14_0_2_15
void FASTCALL call_14_0_2_15(SH2_struct* sh)
{
  SH2movi(sh, 0, (2<<4)+15);
}
#define init_14_0_3_0
void FASTCALL call_14_0_3_0(SH2_struct* sh)
{
  SH2movi(sh, 0, (3<<4)+0);
}
#define init_14_0_3_1
void FASTCALL call_14_0_3_1(SH2_struct* sh)
{
  SH2movi(sh, 0, (3<<4)+1);
}
#define init_14_0_3_2
void FASTCALL call_14_0_3_2(SH2_struct* sh)
{
  SH2movi(sh, 0, (3<<4)+2);
}
#define init_14_0_3_3
void FASTCALL call_14_0_3_3(SH2_struct* sh)
{
  SH2movi(sh, 0, (3<<4)+3);
}
#define init_14_0_3_4
void FASTCALL call_14_0_3_4(SH2_struct* sh)
{
  SH2movi(sh, 0, (3<<4)+4);
}
#define init_14_0_3_5
void FASTCALL call_14_0_3_5(SH2_struct* sh)
{
  SH2movi(sh, 0, (3<<4)+5);
}
#define init_14_0_3_6
void FASTCALL call_14_0_3_6(SH2_struct* sh)
{
  SH2movi(sh, 0, (3<<4)+6);
}
#define init_14_0_3_7
void FASTCALL call_14_0_3_7(SH2_struct* sh)
{
  SH2movi(sh, 0, (3<<4)+7);
}
#define init_14_0_3_8
void FASTCALL call_14_0_3_8(SH2_struct* sh)
{
  SH2movi(sh, 0, (3<<4)+8);
}
#define init_14_0_3_9
void FASTCALL call_14_0_3_9(SH2_struct* sh)
{
  SH2movi(sh, 0, (3<<4)+9);
}
#define init_14_0_3_10
void FASTCALL call_14_0_3_10(SH2_struct* sh)
{
  SH2movi(sh, 0, (3<<4)+10);
}
#define init_14_0_3_11
void FASTCALL call_14_0_3_11(SH2_struct* sh)
{
  SH2movi(sh, 0, (3<<4)+11);
}
#define init_14_0_3_12
void FASTCALL call_14_0_3_12(SH2_struct* sh)
{
  SH2movi(sh, 0, (3<<4)+12);
}
#define init_14_0_3_13
void FASTCALL call_14_0_3_13(SH2_struct* sh)
{
  SH2movi(sh, 0, (3<<4)+13);
}
#define init_14_0_3_14
void FASTCALL call_14_0_3_14(SH2_struct* sh)
{
  SH2movi(sh, 0, (3<<4)+14);
}
#define init_14_0_3_15
void FASTCALL call_14_0_3_15(SH2_struct* sh)
{
  SH2movi(sh, 0, (3<<4)+15);
}
#define init_14_0_4_0
void FASTCALL call_14_0_4_0(SH2_struct* sh)
{
  SH2movi(sh, 0, (4<<4)+0);
}
#define init_14_0_4_1
void FASTCALL call_14_0_4_1(SH2_struct* sh)
{
  SH2movi(sh, 0, (4<<4)+1);
}
#define init_14_0_4_2
void FASTCALL call_14_0_4_2(SH2_struct* sh)
{
  SH2movi(sh, 0, (4<<4)+2);
}
#define init_14_0_4_3
void FASTCALL call_14_0_4_3(SH2_struct* sh)
{
  SH2movi(sh, 0, (4<<4)+3);
}
#define init_14_0_4_4
void FASTCALL call_14_0_4_4(SH2_struct* sh)
{
  SH2movi(sh, 0, (4<<4)+4);
}
#define init_14_0_4_5
void FASTCALL call_14_0_4_5(SH2_struct* sh)
{
  SH2movi(sh, 0, (4<<4)+5);
}
#define init_14_0_4_6
void FASTCALL call_14_0_4_6(SH2_struct* sh)
{
  SH2movi(sh, 0, (4<<4)+6);
}
#define init_14_0_4_7
void FASTCALL call_14_0_4_7(SH2_struct* sh)
{
  SH2movi(sh, 0, (4<<4)+7);
}
#define init_14_0_4_8
void FASTCALL call_14_0_4_8(SH2_struct* sh)
{
  SH2movi(sh, 0, (4<<4)+8);
}
#define init_14_0_4_9
void FASTCALL call_14_0_4_9(SH2_struct* sh)
{
  SH2movi(sh, 0, (4<<4)+9);
}
#define init_14_0_4_10
void FASTCALL call_14_0_4_10(SH2_struct* sh)
{
  SH2movi(sh, 0, (4<<4)+10);
}
#define init_14_0_4_11
void FASTCALL call_14_0_4_11(SH2_struct* sh)
{
  SH2movi(sh, 0, (4<<4)+11);
}
#define init_14_0_4_12
void FASTCALL call_14_0_4_12(SH2_struct* sh)
{
  SH2movi(sh, 0, (4<<4)+12);
}
#define init_14_0_4_13
void FASTCALL call_14_0_4_13(SH2_struct* sh)
{
  SH2movi(sh, 0, (4<<4)+13);
}
#define init_14_0_4_14
void FASTCALL call_14_0_4_14(SH2_struct* sh)
{
  SH2movi(sh, 0, (4<<4)+14);
}
#define init_14_0_4_15
void FASTCALL call_14_0_4_15(SH2_struct* sh)
{
  SH2movi(sh, 0, (4<<4)+15);
}
#define init_14_0_5_0
void FASTCALL call_14_0_5_0(SH2_struct* sh)
{
  SH2movi(sh, 0, (5<<4)+0);
}
#define init_14_0_5_1
void FASTCALL call_14_0_5_1(SH2_struct* sh)
{
  SH2movi(sh, 0, (5<<4)+1);
}
#define init_14_0_5_2
void FASTCALL call_14_0_5_2(SH2_struct* sh)
{
  SH2movi(sh, 0, (5<<4)+2);
}
#define init_14_0_5_3
void FASTCALL call_14_0_5_3(SH2_struct* sh)
{
  SH2movi(sh, 0, (5<<4)+3);
}
#define init_14_0_5_4
void FASTCALL call_14_0_5_4(SH2_struct* sh)
{
  SH2movi(sh, 0, (5<<4)+4);
}
#define init_14_0_5_5
void FASTCALL call_14_0_5_5(SH2_struct* sh)
{
  SH2movi(sh, 0, (5<<4)+5);
}
#define init_14_0_5_6
void FASTCALL call_14_0_5_6(SH2_struct* sh)
{
  SH2movi(sh, 0, (5<<4)+6);
}
#define init_14_0_5_7
void FASTCALL call_14_0_5_7(SH2_struct* sh)
{
  SH2movi(sh, 0, (5<<4)+7);
}
#define init_14_0_5_8
void FASTCALL call_14_0_5_8(SH2_struct* sh)
{
  SH2movi(sh, 0, (5<<4)+8);
}
#define init_14_0_5_9
void FASTCALL call_14_0_5_9(SH2_struct* sh)
{
  SH2movi(sh, 0, (5<<4)+9);
}
#define init_14_0_5_10
void FASTCALL call_14_0_5_10(SH2_struct* sh)
{
  SH2movi(sh, 0, (5<<4)+10);
}
#define init_14_0_5_11
void FASTCALL call_14_0_5_11(SH2_struct* sh)
{
  SH2movi(sh, 0, (5<<4)+11);
}
#define init_14_0_5_12
void FASTCALL call_14_0_5_12(SH2_struct* sh)
{
  SH2movi(sh, 0, (5<<4)+12);
}
#define init_14_0_5_13
void FASTCALL call_14_0_5_13(SH2_struct* sh)
{
  SH2movi(sh, 0, (5<<4)+13);
}
#define init_14_0_5_14
void FASTCALL call_14_0_5_14(SH2_struct* sh)
{
  SH2movi(sh, 0, (5<<4)+14);
}
#define init_14_0_5_15
void FASTCALL call_14_0_5_15(SH2_struct* sh)
{
  SH2movi(sh, 0, (5<<4)+15);
}
#define init_14_0_6_0
void FASTCALL call_14_0_6_0(SH2_struct* sh)
{
  SH2movi(sh, 0, (6<<4)+0);
}
#define init_14_0_6_1
void FASTCALL call_14_0_6_1(SH2_struct* sh)
{
  SH2movi(sh, 0, (6<<4)+1);
}
#define init_14_0_6_2
void FASTCALL call_14_0_6_2(SH2_struct* sh)
{
  SH2movi(sh, 0, (6<<4)+2);
}
#define init_14_0_6_3
void FASTCALL call_14_0_6_3(SH2_struct* sh)
{
  SH2movi(sh, 0, (6<<4)+3);
}
#define init_14_0_6_4
void FASTCALL call_14_0_6_4(SH2_struct* sh)
{
  SH2movi(sh, 0, (6<<4)+4);
}
#define init_14_0_6_5
void FASTCALL call_14_0_6_5(SH2_struct* sh)
{
  SH2movi(sh, 0, (6<<4)+5);
}
#define init_14_0_6_6
void FASTCALL call_14_0_6_6(SH2_struct* sh)
{
  SH2movi(sh, 0, (6<<4)+6);
}
#define init_14_0_6_7
void FASTCALL call_14_0_6_7(SH2_struct* sh)
{
  SH2movi(sh, 0, (6<<4)+7);
}
#define init_14_0_6_8
void FASTCALL call_14_0_6_8(SH2_struct* sh)
{
  SH2movi(sh, 0, (6<<4)+8);
}
#define init_14_0_6_9
void FASTCALL call_14_0_6_9(SH2_struct* sh)
{
  SH2movi(sh, 0, (6<<4)+9);
}
#define init_14_0_6_10
void FASTCALL call_14_0_6_10(SH2_struct* sh)
{
  SH2movi(sh, 0, (6<<4)+10);
}
#define init_14_0_6_11
void FASTCALL call_14_0_6_11(SH2_struct* sh)
{
  SH2movi(sh, 0, (6<<4)+11);
}
#define init_14_0_6_12
void FASTCALL call_14_0_6_12(SH2_struct* sh)
{
  SH2movi(sh, 0, (6<<4)+12);
}
#define init_14_0_6_13
void FASTCALL call_14_0_6_13(SH2_struct* sh)
{
  SH2movi(sh, 0, (6<<4)+13);
}
#define init_14_0_6_14
void FASTCALL call_14_0_6_14(SH2_struct* sh)
{
  SH2movi(sh, 0, (6<<4)+14);
}
#define init_14_0_6_15
void FASTCALL call_14_0_6_15(SH2_struct* sh)
{
  SH2movi(sh, 0, (6<<4)+15);
}
#define init_14_0_7_0
void FASTCALL call_14_0_7_0(SH2_struct* sh)
{
  SH2movi(sh, 0, (7<<4)+0);
}
#define init_14_0_7_1
void FASTCALL call_14_0_7_1(SH2_struct* sh)
{
  SH2movi(sh, 0, (7<<4)+1);
}
#define init_14_0_7_2
void FASTCALL call_14_0_7_2(SH2_struct* sh)
{
  SH2movi(sh, 0, (7<<4)+2);
}
#define init_14_0_7_3
void FASTCALL call_14_0_7_3(SH2_struct* sh)
{
  SH2movi(sh, 0, (7<<4)+3);
}
#define init_14_0_7_4
void FASTCALL call_14_0_7_4(SH2_struct* sh)
{
  SH2movi(sh, 0, (7<<4)+4);
}
#define init_14_0_7_5
void FASTCALL call_14_0_7_5(SH2_struct* sh)
{
  SH2movi(sh, 0, (7<<4)+5);
}
#define init_14_0_7_6
void FASTCALL call_14_0_7_6(SH2_struct* sh)
{
  SH2movi(sh, 0, (7<<4)+6);
}
#define init_14_0_7_7
void FASTCALL call_14_0_7_7(SH2_struct* sh)
{
  SH2movi(sh, 0, (7<<4)+7);
}
#define init_14_0_7_8
void FASTCALL call_14_0_7_8(SH2_struct* sh)
{
  SH2movi(sh, 0, (7<<4)+8);
}
#define init_14_0_7_9
void FASTCALL call_14_0_7_9(SH2_struct* sh)
{
  SH2movi(sh, 0, (7<<4)+9);
}
#define init_14_0_7_10
void FASTCALL call_14_0_7_10(SH2_struct* sh)
{
  SH2movi(sh, 0, (7<<4)+10);
}
#define init_14_0_7_11
void FASTCALL call_14_0_7_11(SH2_struct* sh)
{
  SH2movi(sh, 0, (7<<4)+11);
}
#define init_14_0_7_12
void FASTCALL call_14_0_7_12(SH2_struct* sh)
{
  SH2movi(sh, 0, (7<<4)+12);
}
#define init_14_0_7_13
void FASTCALL call_14_0_7_13(SH2_struct* sh)
{
  SH2movi(sh, 0, (7<<4)+13);
}
#define init_14_0_7_14
void FASTCALL call_14_0_7_14(SH2_struct* sh)
{
  SH2movi(sh, 0, (7<<4)+14);
}
#define init_14_0_7_15
void FASTCALL call_14_0_7_15(SH2_struct* sh)
{
  SH2movi(sh, 0, (7<<4)+15);
}
#define init_14_0_8_0
void FASTCALL call_14_0_8_0(SH2_struct* sh)
{
  SH2movi(sh, 0, (8<<4)+0);
}
#define init_14_0_8_1
void FASTCALL call_14_0_8_1(SH2_struct* sh)
{
  SH2movi(sh, 0, (8<<4)+1);
}
#define init_14_0_8_2
void FASTCALL call_14_0_8_2(SH2_struct* sh)
{
  SH2movi(sh, 0, (8<<4)+2);
}
#define init_14_0_8_3
void FASTCALL call_14_0_8_3(SH2_struct* sh)
{
  SH2movi(sh, 0, (8<<4)+3);
}
#define init_14_0_8_4
void FASTCALL call_14_0_8_4(SH2_struct* sh)
{
  SH2movi(sh, 0, (8<<4)+4);
}
#define init_14_0_8_5
void FASTCALL call_14_0_8_5(SH2_struct* sh)
{
  SH2movi(sh, 0, (8<<4)+5);
}
#define init_14_0_8_6
void FASTCALL call_14_0_8_6(SH2_struct* sh)
{
  SH2movi(sh, 0, (8<<4)+6);
}
#define init_14_0_8_7
void FASTCALL call_14_0_8_7(SH2_struct* sh)
{
  SH2movi(sh, 0, (8<<4)+7);
}
#define init_14_0_8_8
void FASTCALL call_14_0_8_8(SH2_struct* sh)
{
  SH2movi(sh, 0, (8<<4)+8);
}
#define init_14_0_8_9
void FASTCALL call_14_0_8_9(SH2_struct* sh)
{
  SH2movi(sh, 0, (8<<4)+9);
}
#define init_14_0_8_10
void FASTCALL call_14_0_8_10(SH2_struct* sh)
{
  SH2movi(sh, 0, (8<<4)+10);
}
#define init_14_0_8_11
void FASTCALL call_14_0_8_11(SH2_struct* sh)
{
  SH2movi(sh, 0, (8<<4)+11);
}
#define init_14_0_8_12
void FASTCALL call_14_0_8_12(SH2_struct* sh)
{
  SH2movi(sh, 0, (8<<4)+12);
}
#define init_14_0_8_13
void FASTCALL call_14_0_8_13(SH2_struct* sh)
{
  SH2movi(sh, 0, (8<<4)+13);
}
#define init_14_0_8_14
void FASTCALL call_14_0_8_14(SH2_struct* sh)
{
  SH2movi(sh, 0, (8<<4)+14);
}
#define init_14_0_8_15
void FASTCALL call_14_0_8_15(SH2_struct* sh)
{
  SH2movi(sh, 0, (8<<4)+15);
}
#define init_14_0_9_0
void FASTCALL call_14_0_9_0(SH2_struct* sh)
{
  SH2movi(sh, 0, (9<<4)+0);
}
#define init_14_0_9_1
void FASTCALL call_14_0_9_1(SH2_struct* sh)
{
  SH2movi(sh, 0, (9<<4)+1);
}
#define init_14_0_9_2
void FASTCALL call_14_0_9_2(SH2_struct* sh)
{
  SH2movi(sh, 0, (9<<4)+2);
}
#define init_14_0_9_3
void FASTCALL call_14_0_9_3(SH2_struct* sh)
{
  SH2movi(sh, 0, (9<<4)+3);
}
#define init_14_0_9_4
void FASTCALL call_14_0_9_4(SH2_struct* sh)
{
  SH2movi(sh, 0, (9<<4)+4);
}
#define init_14_0_9_5
void FASTCALL call_14_0_9_5(SH2_struct* sh)
{
  SH2movi(sh, 0, (9<<4)+5);
}
#define init_14_0_9_6
void FASTCALL call_14_0_9_6(SH2_struct* sh)
{
  SH2movi(sh, 0, (9<<4)+6);
}
#define init_14_0_9_7
void FASTCALL call_14_0_9_7(SH2_struct* sh)
{
  SH2movi(sh, 0, (9<<4)+7);
}
#define init_14_0_9_8
void FASTCALL call_14_0_9_8(SH2_struct* sh)
{
  SH2movi(sh, 0, (9<<4)+8);
}
#define init_14_0_9_9
void FASTCALL call_14_0_9_9(SH2_struct* sh)
{
  SH2movi(sh, 0, (9<<4)+9);
}
#define init_14_0_9_10
void FASTCALL call_14_0_9_10(SH2_struct* sh)
{
  SH2movi(sh, 0, (9<<4)+10);
}
#define init_14_0_9_11
void FASTCALL call_14_0_9_11(SH2_struct* sh)
{
  SH2movi(sh, 0, (9<<4)+11);
}
#define init_14_0_9_12
void FASTCALL call_14_0_9_12(SH2_struct* sh)
{
  SH2movi(sh, 0, (9<<4)+12);
}
#define init_14_0_9_13
void FASTCALL call_14_0_9_13(SH2_struct* sh)
{
  SH2movi(sh, 0, (9<<4)+13);
}
#define init_14_0_9_14
void FASTCALL call_14_0_9_14(SH2_struct* sh)
{
  SH2movi(sh, 0, (9<<4)+14);
}
#define init_14_0_9_15
void FASTCALL call_14_0_9_15(SH2_struct* sh)
{
  SH2movi(sh, 0, (9<<4)+15);
}
#define init_14_0_10_0
void FASTCALL call_14_0_10_0(SH2_struct* sh)
{
  SH2movi(sh, 0, (10<<4)+0);
}
#define init_14_0_10_1
void FASTCALL call_14_0_10_1(SH2_struct* sh)
{
  SH2movi(sh, 0, (10<<4)+1);
}
#define init_14_0_10_2
void FASTCALL call_14_0_10_2(SH2_struct* sh)
{
  SH2movi(sh, 0, (10<<4)+2);
}
#define init_14_0_10_3
void FASTCALL call_14_0_10_3(SH2_struct* sh)
{
  SH2movi(sh, 0, (10<<4)+3);
}
#define init_14_0_10_4
void FASTCALL call_14_0_10_4(SH2_struct* sh)
{
  SH2movi(sh, 0, (10<<4)+4);
}
#define init_14_0_10_5
void FASTCALL call_14_0_10_5(SH2_struct* sh)
{
  SH2movi(sh, 0, (10<<4)+5);
}
#define init_14_0_10_6
void FASTCALL call_14_0_10_6(SH2_struct* sh)
{
  SH2movi(sh, 0, (10<<4)+6);
}
#define init_14_0_10_7
void FASTCALL call_14_0_10_7(SH2_struct* sh)
{
  SH2movi(sh, 0, (10<<4)+7);
}
#define init_14_0_10_8
void FASTCALL call_14_0_10_8(SH2_struct* sh)
{
  SH2movi(sh, 0, (10<<4)+8);
}
#define init_14_0_10_9
void FASTCALL call_14_0_10_9(SH2_struct* sh)
{
  SH2movi(sh, 0, (10<<4)+9);
}
#define init_14_0_10_10
void FASTCALL call_14_0_10_10(SH2_struct* sh)
{
  SH2movi(sh, 0, (10<<4)+10);
}
#define init_14_0_10_11
void FASTCALL call_14_0_10_11(SH2_struct* sh)
{
  SH2movi(sh, 0, (10<<4)+11);
}
#define init_14_0_10_12
void FASTCALL call_14_0_10_12(SH2_struct* sh)
{
  SH2movi(sh, 0, (10<<4)+12);
}
#define init_14_0_10_13
void FASTCALL call_14_0_10_13(SH2_struct* sh)
{
  SH2movi(sh, 0, (10<<4)+13);
}
#define init_14_0_10_14
void FASTCALL call_14_0_10_14(SH2_struct* sh)
{
  SH2movi(sh, 0, (10<<4)+14);
}
#define init_14_0_10_15
void FASTCALL call_14_0_10_15(SH2_struct* sh)
{
  SH2movi(sh, 0, (10<<4)+15);
}
#define init_14_0_11_0
void FASTCALL call_14_0_11_0(SH2_struct* sh)
{
  SH2movi(sh, 0, (11<<4)+0);
}
#define init_14_0_11_1
void FASTCALL call_14_0_11_1(SH2_struct* sh)
{
  SH2movi(sh, 0, (11<<4)+1);
}
#define init_14_0_11_2
void FASTCALL call_14_0_11_2(SH2_struct* sh)
{
  SH2movi(sh, 0, (11<<4)+2);
}
#define init_14_0_11_3
void FASTCALL call_14_0_11_3(SH2_struct* sh)
{
  SH2movi(sh, 0, (11<<4)+3);
}
#define init_14_0_11_4
void FASTCALL call_14_0_11_4(SH2_struct* sh)
{
  SH2movi(sh, 0, (11<<4)+4);
}
#define init_14_0_11_5
void FASTCALL call_14_0_11_5(SH2_struct* sh)
{
  SH2movi(sh, 0, (11<<4)+5);
}
#define init_14_0_11_6
void FASTCALL call_14_0_11_6(SH2_struct* sh)
{
  SH2movi(sh, 0, (11<<4)+6);
}
#define init_14_0_11_7
void FASTCALL call_14_0_11_7(SH2_struct* sh)
{
  SH2movi(sh, 0, (11<<4)+7);
}
#define init_14_0_11_8
void FASTCALL call_14_0_11_8(SH2_struct* sh)
{
  SH2movi(sh, 0, (11<<4)+8);
}
#define init_14_0_11_9
void FASTCALL call_14_0_11_9(SH2_struct* sh)
{
  SH2movi(sh, 0, (11<<4)+9);
}
#define init_14_0_11_10
void FASTCALL call_14_0_11_10(SH2_struct* sh)
{
  SH2movi(sh, 0, (11<<4)+10);
}
#define init_14_0_11_11
void FASTCALL call_14_0_11_11(SH2_struct* sh)
{
  SH2movi(sh, 0, (11<<4)+11);
}
#define init_14_0_11_12
void FASTCALL call_14_0_11_12(SH2_struct* sh)
{
  SH2movi(sh, 0, (11<<4)+12);
}
#define init_14_0_11_13
void FASTCALL call_14_0_11_13(SH2_struct* sh)
{
  SH2movi(sh, 0, (11<<4)+13);
}
#define init_14_0_11_14
void FASTCALL call_14_0_11_14(SH2_struct* sh)
{
  SH2movi(sh, 0, (11<<4)+14);
}
#define init_14_0_11_15
void FASTCALL call_14_0_11_15(SH2_struct* sh)
{
  SH2movi(sh, 0, (11<<4)+15);
}
#define init_14_0_12_0
void FASTCALL call_14_0_12_0(SH2_struct* sh)
{
  SH2movi(sh, 0, (12<<4)+0);
}
#define init_14_0_12_1
void FASTCALL call_14_0_12_1(SH2_struct* sh)
{
  SH2movi(sh, 0, (12<<4)+1);
}
#define init_14_0_12_2
void FASTCALL call_14_0_12_2(SH2_struct* sh)
{
  SH2movi(sh, 0, (12<<4)+2);
}
#define init_14_0_12_3
void FASTCALL call_14_0_12_3(SH2_struct* sh)
{
  SH2movi(sh, 0, (12<<4)+3);
}
#define init_14_0_12_4
void FASTCALL call_14_0_12_4(SH2_struct* sh)
{
  SH2movi(sh, 0, (12<<4)+4);
}
#define init_14_0_12_5
void FASTCALL call_14_0_12_5(SH2_struct* sh)
{
  SH2movi(sh, 0, (12<<4)+5);
}
#define init_14_0_12_6
void FASTCALL call_14_0_12_6(SH2_struct* sh)
{
  SH2movi(sh, 0, (12<<4)+6);
}
#define init_14_0_12_7
void FASTCALL call_14_0_12_7(SH2_struct* sh)
{
  SH2movi(sh, 0, (12<<4)+7);
}
#define init_14_0_12_8
void FASTCALL call_14_0_12_8(SH2_struct* sh)
{
  SH2movi(sh, 0, (12<<4)+8);
}
#define init_14_0_12_9
void FASTCALL call_14_0_12_9(SH2_struct* sh)
{
  SH2movi(sh, 0, (12<<4)+9);
}
#define init_14_0_12_10
void FASTCALL call_14_0_12_10(SH2_struct* sh)
{
  SH2movi(sh, 0, (12<<4)+10);
}
#define init_14_0_12_11
void FASTCALL call_14_0_12_11(SH2_struct* sh)
{
  SH2movi(sh, 0, (12<<4)+11);
}
#define init_14_0_12_12
void FASTCALL call_14_0_12_12(SH2_struct* sh)
{
  SH2movi(sh, 0, (12<<4)+12);
}
#define init_14_0_12_13
void FASTCALL call_14_0_12_13(SH2_struct* sh)
{
  SH2movi(sh, 0, (12<<4)+13);
}
#define init_14_0_12_14
void FASTCALL call_14_0_12_14(SH2_struct* sh)
{
  SH2movi(sh, 0, (12<<4)+14);
}
#define init_14_0_12_15
void FASTCALL call_14_0_12_15(SH2_struct* sh)
{
  SH2movi(sh, 0, (12<<4)+15);
}
#define init_14_0_13_0
void FASTCALL call_14_0_13_0(SH2_struct* sh)
{
  SH2movi(sh, 0, (13<<4)+0);
}
#define init_14_0_13_1
void FASTCALL call_14_0_13_1(SH2_struct* sh)
{
  SH2movi(sh, 0, (13<<4)+1);
}
#define init_14_0_13_2
void FASTCALL call_14_0_13_2(SH2_struct* sh)
{
  SH2movi(sh, 0, (13<<4)+2);
}
#define init_14_0_13_3
void FASTCALL call_14_0_13_3(SH2_struct* sh)
{
  SH2movi(sh, 0, (13<<4)+3);
}
#define init_14_0_13_4
void FASTCALL call_14_0_13_4(SH2_struct* sh)
{
  SH2movi(sh, 0, (13<<4)+4);
}
#define init_14_0_13_5
void FASTCALL call_14_0_13_5(SH2_struct* sh)
{
  SH2movi(sh, 0, (13<<4)+5);
}
#define init_14_0_13_6
void FASTCALL call_14_0_13_6(SH2_struct* sh)
{
  SH2movi(sh, 0, (13<<4)+6);
}
#define init_14_0_13_7
void FASTCALL call_14_0_13_7(SH2_struct* sh)
{
  SH2movi(sh, 0, (13<<4)+7);
}
#define init_14_0_13_8
void FASTCALL call_14_0_13_8(SH2_struct* sh)
{
  SH2movi(sh, 0, (13<<4)+8);
}
#define init_14_0_13_9
void FASTCALL call_14_0_13_9(SH2_struct* sh)
{
  SH2movi(sh, 0, (13<<4)+9);
}
#define init_14_0_13_10
void FASTCALL call_14_0_13_10(SH2_struct* sh)
{
  SH2movi(sh, 0, (13<<4)+10);
}
#define init_14_0_13_11
void FASTCALL call_14_0_13_11(SH2_struct* sh)
{
  SH2movi(sh, 0, (13<<4)+11);
}
#define init_14_0_13_12
void FASTCALL call_14_0_13_12(SH2_struct* sh)
{
  SH2movi(sh, 0, (13<<4)+12);
}
#define init_14_0_13_13
void FASTCALL call_14_0_13_13(SH2_struct* sh)
{
  SH2movi(sh, 0, (13<<4)+13);
}
#define init_14_0_13_14
void FASTCALL call_14_0_13_14(SH2_struct* sh)
{
  SH2movi(sh, 0, (13<<4)+14);
}
#define init_14_0_13_15
void FASTCALL call_14_0_13_15(SH2_struct* sh)
{
  SH2movi(sh, 0, (13<<4)+15);
}
#define init_14_0_14_0
void FASTCALL call_14_0_14_0(SH2_struct* sh)
{
  SH2movi(sh, 0, (14<<4)+0);
}
#define init_14_0_14_1
void FASTCALL call_14_0_14_1(SH2_struct* sh)
{
  SH2movi(sh, 0, (14<<4)+1);
}
#define init_14_0_14_2
void FASTCALL call_14_0_14_2(SH2_struct* sh)
{
  SH2movi(sh, 0, (14<<4)+2);
}
#define init_14_0_14_3
void FASTCALL call_14_0_14_3(SH2_struct* sh)
{
  SH2movi(sh, 0, (14<<4)+3);
}
#define init_14_0_14_4
void FASTCALL call_14_0_14_4(SH2_struct* sh)
{
  SH2movi(sh, 0, (14<<4)+4);
}
#define init_14_0_14_5
void FASTCALL call_14_0_14_5(SH2_struct* sh)
{
  SH2movi(sh, 0, (14<<4)+5);
}
#define init_14_0_14_6
void FASTCALL call_14_0_14_6(SH2_struct* sh)
{
  SH2movi(sh, 0, (14<<4)+6);
}
#define init_14_0_14_7
void FASTCALL call_14_0_14_7(SH2_struct* sh)
{
  SH2movi(sh, 0, (14<<4)+7);
}
#define init_14_0_14_8
void FASTCALL call_14_0_14_8(SH2_struct* sh)
{
  SH2movi(sh, 0, (14<<4)+8);
}
#define init_14_0_14_9
void FASTCALL call_14_0_14_9(SH2_struct* sh)
{
  SH2movi(sh, 0, (14<<4)+9);
}
#define init_14_0_14_10
void FASTCALL call_14_0_14_10(SH2_struct* sh)
{
  SH2movi(sh, 0, (14<<4)+10);
}
#define init_14_0_14_11
void FASTCALL call_14_0_14_11(SH2_struct* sh)
{
  SH2movi(sh, 0, (14<<4)+11);
}
#define init_14_0_14_12
void FASTCALL call_14_0_14_12(SH2_struct* sh)
{
  SH2movi(sh, 0, (14<<4)+12);
}
#define init_14_0_14_13
void FASTCALL call_14_0_14_13(SH2_struct* sh)
{
  SH2movi(sh, 0, (14<<4)+13);
}
#define init_14_0_14_14
void FASTCALL call_14_0_14_14(SH2_struct* sh)
{
  SH2movi(sh, 0, (14<<4)+14);
}
#define init_14_0_14_15
void FASTCALL call_14_0_14_15(SH2_struct* sh)
{
  SH2movi(sh, 0, (14<<4)+15);
}
#define init_14_0_15_0
void FASTCALL call_14_0_15_0(SH2_struct* sh)
{
  SH2movi(sh, 0, (15<<4)+0);
}
#define init_14_0_15_1
void FASTCALL call_14_0_15_1(SH2_struct* sh)
{
  SH2movi(sh, 0, (15<<4)+1);
}
#define init_14_0_15_2
void FASTCALL call_14_0_15_2(SH2_struct* sh)
{
  SH2movi(sh, 0, (15<<4)+2);
}
#define init_14_0_15_3
void FASTCALL call_14_0_15_3(SH2_struct* sh)
{
  SH2movi(sh, 0, (15<<4)+3);
}
#define init_14_0_15_4
void FASTCALL call_14_0_15_4(SH2_struct* sh)
{
  SH2movi(sh, 0, (15<<4)+4);
}
#define init_14_0_15_5
void FASTCALL call_14_0_15_5(SH2_struct* sh)
{
  SH2movi(sh, 0, (15<<4)+5);
}
#define init_14_0_15_6
void FASTCALL call_14_0_15_6(SH2_struct* sh)
{
  SH2movi(sh, 0, (15<<4)+6);
}
#define init_14_0_15_7
void FASTCALL call_14_0_15_7(SH2_struct* sh)
{
  SH2movi(sh, 0, (15<<4)+7);
}
#define init_14_0_15_8
void FASTCALL call_14_0_15_8(SH2_struct* sh)
{
  SH2movi(sh, 0, (15<<4)+8);
}
#define init_14_0_15_9
void FASTCALL call_14_0_15_9(SH2_struct* sh)
{
  SH2movi(sh, 0, (15<<4)+9);
}
#define init_14_0_15_10
void FASTCALL call_14_0_15_10(SH2_struct* sh)
{
  SH2movi(sh, 0, (15<<4)+10);
}
#define init_14_0_15_11
void FASTCALL call_14_0_15_11(SH2_struct* sh)
{
  SH2movi(sh, 0, (15<<4)+11);
}
#define init_14_0_15_12
void FASTCALL call_14_0_15_12(SH2_struct* sh)
{
  SH2movi(sh, 0, (15<<4)+12);
}
#define init_14_0_15_13
void FASTCALL call_14_0_15_13(SH2_struct* sh)
{
  SH2movi(sh, 0, (15<<4)+13);
}
#define init_14_0_15_14
void FASTCALL call_14_0_15_14(SH2_struct* sh)
{
  SH2movi(sh, 0, (15<<4)+14);
}
#define init_14_0_15_15
void FASTCALL call_14_0_15_15(SH2_struct* sh)
{
  SH2movi(sh, 0, (15<<4)+15);
}
#define init_14_1_0_0
void FASTCALL call_14_1_0_0(SH2_struct* sh)
{
  SH2movi(sh, 1, (0<<4)+0);
}
#define init_14_1_0_1
void FASTCALL call_14_1_0_1(SH2_struct* sh)
{
  SH2movi(sh, 1, (0<<4)+1);
}
#define init_14_1_0_2
void FASTCALL call_14_1_0_2(SH2_struct* sh)
{
  SH2movi(sh, 1, (0<<4)+2);
}
#define init_14_1_0_3
void FASTCALL call_14_1_0_3(SH2_struct* sh)
{
  SH2movi(sh, 1, (0<<4)+3);
}
#define init_14_1_0_4
void FASTCALL call_14_1_0_4(SH2_struct* sh)
{
  SH2movi(sh, 1, (0<<4)+4);
}
#define init_14_1_0_5
void FASTCALL call_14_1_0_5(SH2_struct* sh)
{
  SH2movi(sh, 1, (0<<4)+5);
}
#define init_14_1_0_6
void FASTCALL call_14_1_0_6(SH2_struct* sh)
{
  SH2movi(sh, 1, (0<<4)+6);
}
#define init_14_1_0_7
void FASTCALL call_14_1_0_7(SH2_struct* sh)
{
  SH2movi(sh, 1, (0<<4)+7);
}
#define init_14_1_0_8
void FASTCALL call_14_1_0_8(SH2_struct* sh)
{
  SH2movi(sh, 1, (0<<4)+8);
}
#define init_14_1_0_9
void FASTCALL call_14_1_0_9(SH2_struct* sh)
{
  SH2movi(sh, 1, (0<<4)+9);
}
#define init_14_1_0_10
void FASTCALL call_14_1_0_10(SH2_struct* sh)
{
  SH2movi(sh, 1, (0<<4)+10);
}
#define init_14_1_0_11
void FASTCALL call_14_1_0_11(SH2_struct* sh)
{
  SH2movi(sh, 1, (0<<4)+11);
}
#define init_14_1_0_12
void FASTCALL call_14_1_0_12(SH2_struct* sh)
{
  SH2movi(sh, 1, (0<<4)+12);
}
#define init_14_1_0_13
void FASTCALL call_14_1_0_13(SH2_struct* sh)
{
  SH2movi(sh, 1, (0<<4)+13);
}
#define init_14_1_0_14
void FASTCALL call_14_1_0_14(SH2_struct* sh)
{
  SH2movi(sh, 1, (0<<4)+14);
}
#define init_14_1_0_15
void FASTCALL call_14_1_0_15(SH2_struct* sh)
{
  SH2movi(sh, 1, (0<<4)+15);
}
#define init_14_1_1_0
void FASTCALL call_14_1_1_0(SH2_struct* sh)
{
  SH2movi(sh, 1, (1<<4)+0);
}
#define init_14_1_1_1
void FASTCALL call_14_1_1_1(SH2_struct* sh)
{
  SH2movi(sh, 1, (1<<4)+1);
}
#define init_14_1_1_2
void FASTCALL call_14_1_1_2(SH2_struct* sh)
{
  SH2movi(sh, 1, (1<<4)+2);
}
#define init_14_1_1_3
void FASTCALL call_14_1_1_3(SH2_struct* sh)
{
  SH2movi(sh, 1, (1<<4)+3);
}
#define init_14_1_1_4
void FASTCALL call_14_1_1_4(SH2_struct* sh)
{
  SH2movi(sh, 1, (1<<4)+4);
}
#define init_14_1_1_5
void FASTCALL call_14_1_1_5(SH2_struct* sh)
{
  SH2movi(sh, 1, (1<<4)+5);
}
#define init_14_1_1_6
void FASTCALL call_14_1_1_6(SH2_struct* sh)
{
  SH2movi(sh, 1, (1<<4)+6);
}
#define init_14_1_1_7
void FASTCALL call_14_1_1_7(SH2_struct* sh)
{
  SH2movi(sh, 1, (1<<4)+7);
}
#define init_14_1_1_8
void FASTCALL call_14_1_1_8(SH2_struct* sh)
{
  SH2movi(sh, 1, (1<<4)+8);
}
#define init_14_1_1_9
void FASTCALL call_14_1_1_9(SH2_struct* sh)
{
  SH2movi(sh, 1, (1<<4)+9);
}
#define init_14_1_1_10
void FASTCALL call_14_1_1_10(SH2_struct* sh)
{
  SH2movi(sh, 1, (1<<4)+10);
}
#define init_14_1_1_11
void FASTCALL call_14_1_1_11(SH2_struct* sh)
{
  SH2movi(sh, 1, (1<<4)+11);
}
#define init_14_1_1_12
void FASTCALL call_14_1_1_12(SH2_struct* sh)
{
  SH2movi(sh, 1, (1<<4)+12);
}
#define init_14_1_1_13
void FASTCALL call_14_1_1_13(SH2_struct* sh)
{
  SH2movi(sh, 1, (1<<4)+13);
}
#define init_14_1_1_14
void FASTCALL call_14_1_1_14(SH2_struct* sh)
{
  SH2movi(sh, 1, (1<<4)+14);
}
#define init_14_1_1_15
void FASTCALL call_14_1_1_15(SH2_struct* sh)
{
  SH2movi(sh, 1, (1<<4)+15);
}
#define init_14_1_2_0
void FASTCALL call_14_1_2_0(SH2_struct* sh)
{
  SH2movi(sh, 1, (2<<4)+0);
}
#define init_14_1_2_1
void FASTCALL call_14_1_2_1(SH2_struct* sh)
{
  SH2movi(sh, 1, (2<<4)+1);
}
#define init_14_1_2_2
void FASTCALL call_14_1_2_2(SH2_struct* sh)
{
  SH2movi(sh, 1, (2<<4)+2);
}
#define init_14_1_2_3
void FASTCALL call_14_1_2_3(SH2_struct* sh)
{
  SH2movi(sh, 1, (2<<4)+3);
}
#define init_14_1_2_4
void FASTCALL call_14_1_2_4(SH2_struct* sh)
{
  SH2movi(sh, 1, (2<<4)+4);
}
#define init_14_1_2_5
void FASTCALL call_14_1_2_5(SH2_struct* sh)
{
  SH2movi(sh, 1, (2<<4)+5);
}
#define init_14_1_2_6
void FASTCALL call_14_1_2_6(SH2_struct* sh)
{
  SH2movi(sh, 1, (2<<4)+6);
}
#define init_14_1_2_7
void FASTCALL call_14_1_2_7(SH2_struct* sh)
{
  SH2movi(sh, 1, (2<<4)+7);
}
#define init_14_1_2_8
void FASTCALL call_14_1_2_8(SH2_struct* sh)
{
  SH2movi(sh, 1, (2<<4)+8);
}
#define init_14_1_2_9
void FASTCALL call_14_1_2_9(SH2_struct* sh)
{
  SH2movi(sh, 1, (2<<4)+9);
}
#define init_14_1_2_10
void FASTCALL call_14_1_2_10(SH2_struct* sh)
{
  SH2movi(sh, 1, (2<<4)+10);
}
#define init_14_1_2_11
void FASTCALL call_14_1_2_11(SH2_struct* sh)
{
  SH2movi(sh, 1, (2<<4)+11);
}
#define init_14_1_2_12
void FASTCALL call_14_1_2_12(SH2_struct* sh)
{
  SH2movi(sh, 1, (2<<4)+12);
}
#define init_14_1_2_13
void FASTCALL call_14_1_2_13(SH2_struct* sh)
{
  SH2movi(sh, 1, (2<<4)+13);
}
#define init_14_1_2_14
void FASTCALL call_14_1_2_14(SH2_struct* sh)
{
  SH2movi(sh, 1, (2<<4)+14);
}
#define init_14_1_2_15
void FASTCALL call_14_1_2_15(SH2_struct* sh)
{
  SH2movi(sh, 1, (2<<4)+15);
}
#define init_14_1_3_0
void FASTCALL call_14_1_3_0(SH2_struct* sh)
{
  SH2movi(sh, 1, (3<<4)+0);
}
#define init_14_1_3_1
void FASTCALL call_14_1_3_1(SH2_struct* sh)
{
  SH2movi(sh, 1, (3<<4)+1);
}
#define init_14_1_3_2
void FASTCALL call_14_1_3_2(SH2_struct* sh)
{
  SH2movi(sh, 1, (3<<4)+2);
}
#define init_14_1_3_3
void FASTCALL call_14_1_3_3(SH2_struct* sh)
{
  SH2movi(sh, 1, (3<<4)+3);
}
#define init_14_1_3_4
void FASTCALL call_14_1_3_4(SH2_struct* sh)
{
  SH2movi(sh, 1, (3<<4)+4);
}
#define init_14_1_3_5
void FASTCALL call_14_1_3_5(SH2_struct* sh)
{
  SH2movi(sh, 1, (3<<4)+5);
}
#define init_14_1_3_6
void FASTCALL call_14_1_3_6(SH2_struct* sh)
{
  SH2movi(sh, 1, (3<<4)+6);
}
#define init_14_1_3_7
void FASTCALL call_14_1_3_7(SH2_struct* sh)
{
  SH2movi(sh, 1, (3<<4)+7);
}
#define init_14_1_3_8
void FASTCALL call_14_1_3_8(SH2_struct* sh)
{
  SH2movi(sh, 1, (3<<4)+8);
}
#define init_14_1_3_9
void FASTCALL call_14_1_3_9(SH2_struct* sh)
{
  SH2movi(sh, 1, (3<<4)+9);
}
#define init_14_1_3_10
void FASTCALL call_14_1_3_10(SH2_struct* sh)
{
  SH2movi(sh, 1, (3<<4)+10);
}
#define init_14_1_3_11
void FASTCALL call_14_1_3_11(SH2_struct* sh)
{
  SH2movi(sh, 1, (3<<4)+11);
}
#define init_14_1_3_12
void FASTCALL call_14_1_3_12(SH2_struct* sh)
{
  SH2movi(sh, 1, (3<<4)+12);
}
#define init_14_1_3_13
void FASTCALL call_14_1_3_13(SH2_struct* sh)
{
  SH2movi(sh, 1, (3<<4)+13);
}
#define init_14_1_3_14
void FASTCALL call_14_1_3_14(SH2_struct* sh)
{
  SH2movi(sh, 1, (3<<4)+14);
}
#define init_14_1_3_15
void FASTCALL call_14_1_3_15(SH2_struct* sh)
{
  SH2movi(sh, 1, (3<<4)+15);
}
#define init_14_1_4_0
void FASTCALL call_14_1_4_0(SH2_struct* sh)
{
  SH2movi(sh, 1, (4<<4)+0);
}
#define init_14_1_4_1
void FASTCALL call_14_1_4_1(SH2_struct* sh)
{
  SH2movi(sh, 1, (4<<4)+1);
}
#define init_14_1_4_2
void FASTCALL call_14_1_4_2(SH2_struct* sh)
{
  SH2movi(sh, 1, (4<<4)+2);
}
#define init_14_1_4_3
void FASTCALL call_14_1_4_3(SH2_struct* sh)
{
  SH2movi(sh, 1, (4<<4)+3);
}
#define init_14_1_4_4
void FASTCALL call_14_1_4_4(SH2_struct* sh)
{
  SH2movi(sh, 1, (4<<4)+4);
}
#define init_14_1_4_5
void FASTCALL call_14_1_4_5(SH2_struct* sh)
{
  SH2movi(sh, 1, (4<<4)+5);
}
#define init_14_1_4_6
void FASTCALL call_14_1_4_6(SH2_struct* sh)
{
  SH2movi(sh, 1, (4<<4)+6);
}
#define init_14_1_4_7
void FASTCALL call_14_1_4_7(SH2_struct* sh)
{
  SH2movi(sh, 1, (4<<4)+7);
}
#define init_14_1_4_8
void FASTCALL call_14_1_4_8(SH2_struct* sh)
{
  SH2movi(sh, 1, (4<<4)+8);
}
#define init_14_1_4_9
void FASTCALL call_14_1_4_9(SH2_struct* sh)
{
  SH2movi(sh, 1, (4<<4)+9);
}
#define init_14_1_4_10
void FASTCALL call_14_1_4_10(SH2_struct* sh)
{
  SH2movi(sh, 1, (4<<4)+10);
}
#define init_14_1_4_11
void FASTCALL call_14_1_4_11(SH2_struct* sh)
{
  SH2movi(sh, 1, (4<<4)+11);
}
#define init_14_1_4_12
void FASTCALL call_14_1_4_12(SH2_struct* sh)
{
  SH2movi(sh, 1, (4<<4)+12);
}
#define init_14_1_4_13
void FASTCALL call_14_1_4_13(SH2_struct* sh)
{
  SH2movi(sh, 1, (4<<4)+13);
}
#define init_14_1_4_14
void FASTCALL call_14_1_4_14(SH2_struct* sh)
{
  SH2movi(sh, 1, (4<<4)+14);
}
#define init_14_1_4_15
void FASTCALL call_14_1_4_15(SH2_struct* sh)
{
  SH2movi(sh, 1, (4<<4)+15);
}
#define init_14_1_5_0
void FASTCALL call_14_1_5_0(SH2_struct* sh)
{
  SH2movi(sh, 1, (5<<4)+0);
}
#define init_14_1_5_1
void FASTCALL call_14_1_5_1(SH2_struct* sh)
{
  SH2movi(sh, 1, (5<<4)+1);
}
#define init_14_1_5_2
void FASTCALL call_14_1_5_2(SH2_struct* sh)
{
  SH2movi(sh, 1, (5<<4)+2);
}
#define init_14_1_5_3
void FASTCALL call_14_1_5_3(SH2_struct* sh)
{
  SH2movi(sh, 1, (5<<4)+3);
}
#define init_14_1_5_4
void FASTCALL call_14_1_5_4(SH2_struct* sh)
{
  SH2movi(sh, 1, (5<<4)+4);
}
#define init_14_1_5_5
void FASTCALL call_14_1_5_5(SH2_struct* sh)
{
  SH2movi(sh, 1, (5<<4)+5);
}
#define init_14_1_5_6
void FASTCALL call_14_1_5_6(SH2_struct* sh)
{
  SH2movi(sh, 1, (5<<4)+6);
}
#define init_14_1_5_7
void FASTCALL call_14_1_5_7(SH2_struct* sh)
{
  SH2movi(sh, 1, (5<<4)+7);
}
#define init_14_1_5_8
void FASTCALL call_14_1_5_8(SH2_struct* sh)
{
  SH2movi(sh, 1, (5<<4)+8);
}
#define init_14_1_5_9
void FASTCALL call_14_1_5_9(SH2_struct* sh)
{
  SH2movi(sh, 1, (5<<4)+9);
}
#define init_14_1_5_10
void FASTCALL call_14_1_5_10(SH2_struct* sh)
{
  SH2movi(sh, 1, (5<<4)+10);
}
#define init_14_1_5_11
void FASTCALL call_14_1_5_11(SH2_struct* sh)
{
  SH2movi(sh, 1, (5<<4)+11);
}
#define init_14_1_5_12
void FASTCALL call_14_1_5_12(SH2_struct* sh)
{
  SH2movi(sh, 1, (5<<4)+12);
}
#define init_14_1_5_13
void FASTCALL call_14_1_5_13(SH2_struct* sh)
{
  SH2movi(sh, 1, (5<<4)+13);
}
#define init_14_1_5_14
void FASTCALL call_14_1_5_14(SH2_struct* sh)
{
  SH2movi(sh, 1, (5<<4)+14);
}
#define init_14_1_5_15
void FASTCALL call_14_1_5_15(SH2_struct* sh)
{
  SH2movi(sh, 1, (5<<4)+15);
}
#define init_14_1_6_0
void FASTCALL call_14_1_6_0(SH2_struct* sh)
{
  SH2movi(sh, 1, (6<<4)+0);
}
#define init_14_1_6_1
void FASTCALL call_14_1_6_1(SH2_struct* sh)
{
  SH2movi(sh, 1, (6<<4)+1);
}
#define init_14_1_6_2
void FASTCALL call_14_1_6_2(SH2_struct* sh)
{
  SH2movi(sh, 1, (6<<4)+2);
}
#define init_14_1_6_3
void FASTCALL call_14_1_6_3(SH2_struct* sh)
{
  SH2movi(sh, 1, (6<<4)+3);
}
#define init_14_1_6_4
void FASTCALL call_14_1_6_4(SH2_struct* sh)
{
  SH2movi(sh, 1, (6<<4)+4);
}
#define init_14_1_6_5
void FASTCALL call_14_1_6_5(SH2_struct* sh)
{
  SH2movi(sh, 1, (6<<4)+5);
}
#define init_14_1_6_6
void FASTCALL call_14_1_6_6(SH2_struct* sh)
{
  SH2movi(sh, 1, (6<<4)+6);
}
#define init_14_1_6_7
void FASTCALL call_14_1_6_7(SH2_struct* sh)
{
  SH2movi(sh, 1, (6<<4)+7);
}
#define init_14_1_6_8
void FASTCALL call_14_1_6_8(SH2_struct* sh)
{
  SH2movi(sh, 1, (6<<4)+8);
}
#define init_14_1_6_9
void FASTCALL call_14_1_6_9(SH2_struct* sh)
{
  SH2movi(sh, 1, (6<<4)+9);
}
#define init_14_1_6_10
void FASTCALL call_14_1_6_10(SH2_struct* sh)
{
  SH2movi(sh, 1, (6<<4)+10);
}
#define init_14_1_6_11
void FASTCALL call_14_1_6_11(SH2_struct* sh)
{
  SH2movi(sh, 1, (6<<4)+11);
}
#define init_14_1_6_12
void FASTCALL call_14_1_6_12(SH2_struct* sh)
{
  SH2movi(sh, 1, (6<<4)+12);
}
#define init_14_1_6_13
void FASTCALL call_14_1_6_13(SH2_struct* sh)
{
  SH2movi(sh, 1, (6<<4)+13);
}
#define init_14_1_6_14
void FASTCALL call_14_1_6_14(SH2_struct* sh)
{
  SH2movi(sh, 1, (6<<4)+14);
}
#define init_14_1_6_15
void FASTCALL call_14_1_6_15(SH2_struct* sh)
{
  SH2movi(sh, 1, (6<<4)+15);
}
#define init_14_1_7_0
void FASTCALL call_14_1_7_0(SH2_struct* sh)
{
  SH2movi(sh, 1, (7<<4)+0);
}
#define init_14_1_7_1
void FASTCALL call_14_1_7_1(SH2_struct* sh)
{
  SH2movi(sh, 1, (7<<4)+1);
}
#define init_14_1_7_2
void FASTCALL call_14_1_7_2(SH2_struct* sh)
{
  SH2movi(sh, 1, (7<<4)+2);
}
#define init_14_1_7_3
void FASTCALL call_14_1_7_3(SH2_struct* sh)
{
  SH2movi(sh, 1, (7<<4)+3);
}
#define init_14_1_7_4
void FASTCALL call_14_1_7_4(SH2_struct* sh)
{
  SH2movi(sh, 1, (7<<4)+4);
}
#define init_14_1_7_5
void FASTCALL call_14_1_7_5(SH2_struct* sh)
{
  SH2movi(sh, 1, (7<<4)+5);
}
#define init_14_1_7_6
void FASTCALL call_14_1_7_6(SH2_struct* sh)
{
  SH2movi(sh, 1, (7<<4)+6);
}
#define init_14_1_7_7
void FASTCALL call_14_1_7_7(SH2_struct* sh)
{
  SH2movi(sh, 1, (7<<4)+7);
}
#define init_14_1_7_8
void FASTCALL call_14_1_7_8(SH2_struct* sh)
{
  SH2movi(sh, 1, (7<<4)+8);
}
#define init_14_1_7_9
void FASTCALL call_14_1_7_9(SH2_struct* sh)
{
  SH2movi(sh, 1, (7<<4)+9);
}
#define init_14_1_7_10
void FASTCALL call_14_1_7_10(SH2_struct* sh)
{
  SH2movi(sh, 1, (7<<4)+10);
}
#define init_14_1_7_11
void FASTCALL call_14_1_7_11(SH2_struct* sh)
{
  SH2movi(sh, 1, (7<<4)+11);
}
#define init_14_1_7_12
void FASTCALL call_14_1_7_12(SH2_struct* sh)
{
  SH2movi(sh, 1, (7<<4)+12);
}
#define init_14_1_7_13
void FASTCALL call_14_1_7_13(SH2_struct* sh)
{
  SH2movi(sh, 1, (7<<4)+13);
}
#define init_14_1_7_14
void FASTCALL call_14_1_7_14(SH2_struct* sh)
{
  SH2movi(sh, 1, (7<<4)+14);
}
#define init_14_1_7_15
void FASTCALL call_14_1_7_15(SH2_struct* sh)
{
  SH2movi(sh, 1, (7<<4)+15);
}
#define init_14_1_8_0
void FASTCALL call_14_1_8_0(SH2_struct* sh)
{
  SH2movi(sh, 1, (8<<4)+0);
}
#define init_14_1_8_1
void FASTCALL call_14_1_8_1(SH2_struct* sh)
{
  SH2movi(sh, 1, (8<<4)+1);
}
#define init_14_1_8_2
void FASTCALL call_14_1_8_2(SH2_struct* sh)
{
  SH2movi(sh, 1, (8<<4)+2);
}
#define init_14_1_8_3
void FASTCALL call_14_1_8_3(SH2_struct* sh)
{
  SH2movi(sh, 1, (8<<4)+3);
}
#define init_14_1_8_4
void FASTCALL call_14_1_8_4(SH2_struct* sh)
{
  SH2movi(sh, 1, (8<<4)+4);
}
#define init_14_1_8_5
void FASTCALL call_14_1_8_5(SH2_struct* sh)
{
  SH2movi(sh, 1, (8<<4)+5);
}
#define init_14_1_8_6
void FASTCALL call_14_1_8_6(SH2_struct* sh)
{
  SH2movi(sh, 1, (8<<4)+6);
}
#define init_14_1_8_7
void FASTCALL call_14_1_8_7(SH2_struct* sh)
{
  SH2movi(sh, 1, (8<<4)+7);
}
#define init_14_1_8_8
void FASTCALL call_14_1_8_8(SH2_struct* sh)
{
  SH2movi(sh, 1, (8<<4)+8);
}
#define init_14_1_8_9
void FASTCALL call_14_1_8_9(SH2_struct* sh)
{
  SH2movi(sh, 1, (8<<4)+9);
}
#define init_14_1_8_10
void FASTCALL call_14_1_8_10(SH2_struct* sh)
{
  SH2movi(sh, 1, (8<<4)+10);
}
#define init_14_1_8_11
void FASTCALL call_14_1_8_11(SH2_struct* sh)
{
  SH2movi(sh, 1, (8<<4)+11);
}
#define init_14_1_8_12
void FASTCALL call_14_1_8_12(SH2_struct* sh)
{
  SH2movi(sh, 1, (8<<4)+12);
}
#define init_14_1_8_13
void FASTCALL call_14_1_8_13(SH2_struct* sh)
{
  SH2movi(sh, 1, (8<<4)+13);
}
#define init_14_1_8_14
void FASTCALL call_14_1_8_14(SH2_struct* sh)
{
  SH2movi(sh, 1, (8<<4)+14);
}
#define init_14_1_8_15
void FASTCALL call_14_1_8_15(SH2_struct* sh)
{
  SH2movi(sh, 1, (8<<4)+15);
}
#define init_14_1_9_0
void FASTCALL call_14_1_9_0(SH2_struct* sh)
{
  SH2movi(sh, 1, (9<<4)+0);
}
#define init_14_1_9_1
void FASTCALL call_14_1_9_1(SH2_struct* sh)
{
  SH2movi(sh, 1, (9<<4)+1);
}
#define init_14_1_9_2
void FASTCALL call_14_1_9_2(SH2_struct* sh)
{
  SH2movi(sh, 1, (9<<4)+2);
}
#define init_14_1_9_3
void FASTCALL call_14_1_9_3(SH2_struct* sh)
{
  SH2movi(sh, 1, (9<<4)+3);
}
#define init_14_1_9_4
void FASTCALL call_14_1_9_4(SH2_struct* sh)
{
  SH2movi(sh, 1, (9<<4)+4);
}
#define init_14_1_9_5
void FASTCALL call_14_1_9_5(SH2_struct* sh)
{
  SH2movi(sh, 1, (9<<4)+5);
}
#define init_14_1_9_6
void FASTCALL call_14_1_9_6(SH2_struct* sh)
{
  SH2movi(sh, 1, (9<<4)+6);
}
#define init_14_1_9_7
void FASTCALL call_14_1_9_7(SH2_struct* sh)
{
  SH2movi(sh, 1, (9<<4)+7);
}
#define init_14_1_9_8
void FASTCALL call_14_1_9_8(SH2_struct* sh)
{
  SH2movi(sh, 1, (9<<4)+8);
}
#define init_14_1_9_9
void FASTCALL call_14_1_9_9(SH2_struct* sh)
{
  SH2movi(sh, 1, (9<<4)+9);
}
#define init_14_1_9_10
void FASTCALL call_14_1_9_10(SH2_struct* sh)
{
  SH2movi(sh, 1, (9<<4)+10);
}
#define init_14_1_9_11
void FASTCALL call_14_1_9_11(SH2_struct* sh)
{
  SH2movi(sh, 1, (9<<4)+11);
}
#define init_14_1_9_12
void FASTCALL call_14_1_9_12(SH2_struct* sh)
{
  SH2movi(sh, 1, (9<<4)+12);
}
#define init_14_1_9_13
void FASTCALL call_14_1_9_13(SH2_struct* sh)
{
  SH2movi(sh, 1, (9<<4)+13);
}
#define init_14_1_9_14
void FASTCALL call_14_1_9_14(SH2_struct* sh)
{
  SH2movi(sh, 1, (9<<4)+14);
}
#define init_14_1_9_15
void FASTCALL call_14_1_9_15(SH2_struct* sh)
{
  SH2movi(sh, 1, (9<<4)+15);
}
#define init_14_1_10_0
void FASTCALL call_14_1_10_0(SH2_struct* sh)
{
  SH2movi(sh, 1, (10<<4)+0);
}
#define init_14_1_10_1
void FASTCALL call_14_1_10_1(SH2_struct* sh)
{
  SH2movi(sh, 1, (10<<4)+1);
}
#define init_14_1_10_2
void FASTCALL call_14_1_10_2(SH2_struct* sh)
{
  SH2movi(sh, 1, (10<<4)+2);
}
#define init_14_1_10_3
void FASTCALL call_14_1_10_3(SH2_struct* sh)
{
  SH2movi(sh, 1, (10<<4)+3);
}
#define init_14_1_10_4
void FASTCALL call_14_1_10_4(SH2_struct* sh)
{
  SH2movi(sh, 1, (10<<4)+4);
}
#define init_14_1_10_5
void FASTCALL call_14_1_10_5(SH2_struct* sh)
{
  SH2movi(sh, 1, (10<<4)+5);
}
#define init_14_1_10_6
void FASTCALL call_14_1_10_6(SH2_struct* sh)
{
  SH2movi(sh, 1, (10<<4)+6);
}
#define init_14_1_10_7
void FASTCALL call_14_1_10_7(SH2_struct* sh)
{
  SH2movi(sh, 1, (10<<4)+7);
}
#define init_14_1_10_8
void FASTCALL call_14_1_10_8(SH2_struct* sh)
{
  SH2movi(sh, 1, (10<<4)+8);
}
#define init_14_1_10_9
void FASTCALL call_14_1_10_9(SH2_struct* sh)
{
  SH2movi(sh, 1, (10<<4)+9);
}
#define init_14_1_10_10
void FASTCALL call_14_1_10_10(SH2_struct* sh)
{
  SH2movi(sh, 1, (10<<4)+10);
}
#define init_14_1_10_11
void FASTCALL call_14_1_10_11(SH2_struct* sh)
{
  SH2movi(sh, 1, (10<<4)+11);
}
#define init_14_1_10_12
void FASTCALL call_14_1_10_12(SH2_struct* sh)
{
  SH2movi(sh, 1, (10<<4)+12);
}
#define init_14_1_10_13
void FASTCALL call_14_1_10_13(SH2_struct* sh)
{
  SH2movi(sh, 1, (10<<4)+13);
}
#define init_14_1_10_14
void FASTCALL call_14_1_10_14(SH2_struct* sh)
{
  SH2movi(sh, 1, (10<<4)+14);
}
#define init_14_1_10_15
void FASTCALL call_14_1_10_15(SH2_struct* sh)
{
  SH2movi(sh, 1, (10<<4)+15);
}
#define init_14_1_11_0
void FASTCALL call_14_1_11_0(SH2_struct* sh)
{
  SH2movi(sh, 1, (11<<4)+0);
}
#define init_14_1_11_1
void FASTCALL call_14_1_11_1(SH2_struct* sh)
{
  SH2movi(sh, 1, (11<<4)+1);
}
#define init_14_1_11_2
void FASTCALL call_14_1_11_2(SH2_struct* sh)
{
  SH2movi(sh, 1, (11<<4)+2);
}
#define init_14_1_11_3
void FASTCALL call_14_1_11_3(SH2_struct* sh)
{
  SH2movi(sh, 1, (11<<4)+3);
}
#define init_14_1_11_4
void FASTCALL call_14_1_11_4(SH2_struct* sh)
{
  SH2movi(sh, 1, (11<<4)+4);
}
#define init_14_1_11_5
void FASTCALL call_14_1_11_5(SH2_struct* sh)
{
  SH2movi(sh, 1, (11<<4)+5);
}
#define init_14_1_11_6
void FASTCALL call_14_1_11_6(SH2_struct* sh)
{
  SH2movi(sh, 1, (11<<4)+6);
}
#define init_14_1_11_7
void FASTCALL call_14_1_11_7(SH2_struct* sh)
{
  SH2movi(sh, 1, (11<<4)+7);
}
#define init_14_1_11_8
void FASTCALL call_14_1_11_8(SH2_struct* sh)
{
  SH2movi(sh, 1, (11<<4)+8);
}
#define init_14_1_11_9
void FASTCALL call_14_1_11_9(SH2_struct* sh)
{
  SH2movi(sh, 1, (11<<4)+9);
}
#define init_14_1_11_10
void FASTCALL call_14_1_11_10(SH2_struct* sh)
{
  SH2movi(sh, 1, (11<<4)+10);
}
#define init_14_1_11_11
void FASTCALL call_14_1_11_11(SH2_struct* sh)
{
  SH2movi(sh, 1, (11<<4)+11);
}
#define init_14_1_11_12
void FASTCALL call_14_1_11_12(SH2_struct* sh)
{
  SH2movi(sh, 1, (11<<4)+12);
}
#define init_14_1_11_13
void FASTCALL call_14_1_11_13(SH2_struct* sh)
{
  SH2movi(sh, 1, (11<<4)+13);
}
#define init_14_1_11_14
void FASTCALL call_14_1_11_14(SH2_struct* sh)
{
  SH2movi(sh, 1, (11<<4)+14);
}
#define init_14_1_11_15
void FASTCALL call_14_1_11_15(SH2_struct* sh)
{
  SH2movi(sh, 1, (11<<4)+15);
}
#define init_14_1_12_0
void FASTCALL call_14_1_12_0(SH2_struct* sh)
{
  SH2movi(sh, 1, (12<<4)+0);
}
#define init_14_1_12_1
void FASTCALL call_14_1_12_1(SH2_struct* sh)
{
  SH2movi(sh, 1, (12<<4)+1);
}
#define init_14_1_12_2
void FASTCALL call_14_1_12_2(SH2_struct* sh)
{
  SH2movi(sh, 1, (12<<4)+2);
}
#define init_14_1_12_3
void FASTCALL call_14_1_12_3(SH2_struct* sh)
{
  SH2movi(sh, 1, (12<<4)+3);
}
#define init_14_1_12_4
void FASTCALL call_14_1_12_4(SH2_struct* sh)
{
  SH2movi(sh, 1, (12<<4)+4);
}
#define init_14_1_12_5
void FASTCALL call_14_1_12_5(SH2_struct* sh)
{
  SH2movi(sh, 1, (12<<4)+5);
}
#define init_14_1_12_6
void FASTCALL call_14_1_12_6(SH2_struct* sh)
{
  SH2movi(sh, 1, (12<<4)+6);
}
#define init_14_1_12_7
void FASTCALL call_14_1_12_7(SH2_struct* sh)
{
  SH2movi(sh, 1, (12<<4)+7);
}
#define init_14_1_12_8
void FASTCALL call_14_1_12_8(SH2_struct* sh)
{
  SH2movi(sh, 1, (12<<4)+8);
}
#define init_14_1_12_9
void FASTCALL call_14_1_12_9(SH2_struct* sh)
{
  SH2movi(sh, 1, (12<<4)+9);
}
#define init_14_1_12_10
void FASTCALL call_14_1_12_10(SH2_struct* sh)
{
  SH2movi(sh, 1, (12<<4)+10);
}
#define init_14_1_12_11
void FASTCALL call_14_1_12_11(SH2_struct* sh)
{
  SH2movi(sh, 1, (12<<4)+11);
}
#define init_14_1_12_12
void FASTCALL call_14_1_12_12(SH2_struct* sh)
{
  SH2movi(sh, 1, (12<<4)+12);
}
#define init_14_1_12_13
void FASTCALL call_14_1_12_13(SH2_struct* sh)
{
  SH2movi(sh, 1, (12<<4)+13);
}
#define init_14_1_12_14
void FASTCALL call_14_1_12_14(SH2_struct* sh)
{
  SH2movi(sh, 1, (12<<4)+14);
}
#define init_14_1_12_15
void FASTCALL call_14_1_12_15(SH2_struct* sh)
{
  SH2movi(sh, 1, (12<<4)+15);
}
#define init_14_1_13_0
void FASTCALL call_14_1_13_0(SH2_struct* sh)
{
  SH2movi(sh, 1, (13<<4)+0);
}
#define init_14_1_13_1
void FASTCALL call_14_1_13_1(SH2_struct* sh)
{
  SH2movi(sh, 1, (13<<4)+1);
}
#define init_14_1_13_2
void FASTCALL call_14_1_13_2(SH2_struct* sh)
{
  SH2movi(sh, 1, (13<<4)+2);
}
#define init_14_1_13_3
void FASTCALL call_14_1_13_3(SH2_struct* sh)
{
  SH2movi(sh, 1, (13<<4)+3);
}
#define init_14_1_13_4
void FASTCALL call_14_1_13_4(SH2_struct* sh)
{
  SH2movi(sh, 1, (13<<4)+4);
}
#define init_14_1_13_5
void FASTCALL call_14_1_13_5(SH2_struct* sh)
{
  SH2movi(sh, 1, (13<<4)+5);
}
#define init_14_1_13_6
void FASTCALL call_14_1_13_6(SH2_struct* sh)
{
  SH2movi(sh, 1, (13<<4)+6);
}
#define init_14_1_13_7
void FASTCALL call_14_1_13_7(SH2_struct* sh)
{
  SH2movi(sh, 1, (13<<4)+7);
}
#define init_14_1_13_8
void FASTCALL call_14_1_13_8(SH2_struct* sh)
{
  SH2movi(sh, 1, (13<<4)+8);
}
#define init_14_1_13_9
void FASTCALL call_14_1_13_9(SH2_struct* sh)
{
  SH2movi(sh, 1, (13<<4)+9);
}
#define init_14_1_13_10
void FASTCALL call_14_1_13_10(SH2_struct* sh)
{
  SH2movi(sh, 1, (13<<4)+10);
}
#define init_14_1_13_11
void FASTCALL call_14_1_13_11(SH2_struct* sh)
{
  SH2movi(sh, 1, (13<<4)+11);
}
#define init_14_1_13_12
void FASTCALL call_14_1_13_12(SH2_struct* sh)
{
  SH2movi(sh, 1, (13<<4)+12);
}
#define init_14_1_13_13
void FASTCALL call_14_1_13_13(SH2_struct* sh)
{
  SH2movi(sh, 1, (13<<4)+13);
}
#define init_14_1_13_14
void FASTCALL call_14_1_13_14(SH2_struct* sh)
{
  SH2movi(sh, 1, (13<<4)+14);
}
#define init_14_1_13_15
void FASTCALL call_14_1_13_15(SH2_struct* sh)
{
  SH2movi(sh, 1, (13<<4)+15);
}
#define init_14_1_14_0
void FASTCALL call_14_1_14_0(SH2_struct* sh)
{
  SH2movi(sh, 1, (14<<4)+0);
}
#define init_14_1_14_1
void FASTCALL call_14_1_14_1(SH2_struct* sh)
{
  SH2movi(sh, 1, (14<<4)+1);
}
#define init_14_1_14_2
void FASTCALL call_14_1_14_2(SH2_struct* sh)
{
  SH2movi(sh, 1, (14<<4)+2);
}
#define init_14_1_14_3
void FASTCALL call_14_1_14_3(SH2_struct* sh)
{
  SH2movi(sh, 1, (14<<4)+3);
}
#define init_14_1_14_4
void FASTCALL call_14_1_14_4(SH2_struct* sh)
{
  SH2movi(sh, 1, (14<<4)+4);
}
#define init_14_1_14_5
void FASTCALL call_14_1_14_5(SH2_struct* sh)
{
  SH2movi(sh, 1, (14<<4)+5);
}
#define init_14_1_14_6
void FASTCALL call_14_1_14_6(SH2_struct* sh)
{
  SH2movi(sh, 1, (14<<4)+6);
}
#define init_14_1_14_7
void FASTCALL call_14_1_14_7(SH2_struct* sh)
{
  SH2movi(sh, 1, (14<<4)+7);
}
#define init_14_1_14_8
void FASTCALL call_14_1_14_8(SH2_struct* sh)
{
  SH2movi(sh, 1, (14<<4)+8);
}
#define init_14_1_14_9
void FASTCALL call_14_1_14_9(SH2_struct* sh)
{
  SH2movi(sh, 1, (14<<4)+9);
}
#define init_14_1_14_10
void FASTCALL call_14_1_14_10(SH2_struct* sh)
{
  SH2movi(sh, 1, (14<<4)+10);
}
#define init_14_1_14_11
void FASTCALL call_14_1_14_11(SH2_struct* sh)
{
  SH2movi(sh, 1, (14<<4)+11);
}
#define init_14_1_14_12
void FASTCALL call_14_1_14_12(SH2_struct* sh)
{
  SH2movi(sh, 1, (14<<4)+12);
}
#define init_14_1_14_13
void FASTCALL call_14_1_14_13(SH2_struct* sh)
{
  SH2movi(sh, 1, (14<<4)+13);
}
#define init_14_1_14_14
void FASTCALL call_14_1_14_14(SH2_struct* sh)
{
  SH2movi(sh, 1, (14<<4)+14);
}
#define init_14_1_14_15
void FASTCALL call_14_1_14_15(SH2_struct* sh)
{
  SH2movi(sh, 1, (14<<4)+15);
}
#define init_14_1_15_0
void FASTCALL call_14_1_15_0(SH2_struct* sh)
{
  SH2movi(sh, 1, (15<<4)+0);
}
#define init_14_1_15_1
void FASTCALL call_14_1_15_1(SH2_struct* sh)
{
  SH2movi(sh, 1, (15<<4)+1);
}
#define init_14_1_15_2
void FASTCALL call_14_1_15_2(SH2_struct* sh)
{
  SH2movi(sh, 1, (15<<4)+2);
}
#define init_14_1_15_3
void FASTCALL call_14_1_15_3(SH2_struct* sh)
{
  SH2movi(sh, 1, (15<<4)+3);
}
#define init_14_1_15_4
void FASTCALL call_14_1_15_4(SH2_struct* sh)
{
  SH2movi(sh, 1, (15<<4)+4);
}
#define init_14_1_15_5
void FASTCALL call_14_1_15_5(SH2_struct* sh)
{
  SH2movi(sh, 1, (15<<4)+5);
}
#define init_14_1_15_6
void FASTCALL call_14_1_15_6(SH2_struct* sh)
{
  SH2movi(sh, 1, (15<<4)+6);
}
#define init_14_1_15_7
void FASTCALL call_14_1_15_7(SH2_struct* sh)
{
  SH2movi(sh, 1, (15<<4)+7);
}
#define init_14_1_15_8
void FASTCALL call_14_1_15_8(SH2_struct* sh)
{
  SH2movi(sh, 1, (15<<4)+8);
}
#define init_14_1_15_9
void FASTCALL call_14_1_15_9(SH2_struct* sh)
{
  SH2movi(sh, 1, (15<<4)+9);
}
#define init_14_1_15_10
void FASTCALL call_14_1_15_10(SH2_struct* sh)
{
  SH2movi(sh, 1, (15<<4)+10);
}
#define init_14_1_15_11
void FASTCALL call_14_1_15_11(SH2_struct* sh)
{
  SH2movi(sh, 1, (15<<4)+11);
}
#define init_14_1_15_12
void FASTCALL call_14_1_15_12(SH2_struct* sh)
{
  SH2movi(sh, 1, (15<<4)+12);
}
#define init_14_1_15_13
void FASTCALL call_14_1_15_13(SH2_struct* sh)
{
  SH2movi(sh, 1, (15<<4)+13);
}
#define init_14_1_15_14
void FASTCALL call_14_1_15_14(SH2_struct* sh)
{
  SH2movi(sh, 1, (15<<4)+14);
}
#define init_14_1_15_15
void FASTCALL call_14_1_15_15(SH2_struct* sh)
{
  SH2movi(sh, 1, (15<<4)+15);
}
#define init_14_2_0_0
void FASTCALL call_14_2_0_0(SH2_struct* sh)
{
  SH2movi(sh, 2, (0<<4)+0);
}
#define init_14_2_0_1
void FASTCALL call_14_2_0_1(SH2_struct* sh)
{
  SH2movi(sh, 2, (0<<4)+1);
}
#define init_14_2_0_2
void FASTCALL call_14_2_0_2(SH2_struct* sh)
{
  SH2movi(sh, 2, (0<<4)+2);
}
#define init_14_2_0_3
void FASTCALL call_14_2_0_3(SH2_struct* sh)
{
  SH2movi(sh, 2, (0<<4)+3);
}
#define init_14_2_0_4
void FASTCALL call_14_2_0_4(SH2_struct* sh)
{
  SH2movi(sh, 2, (0<<4)+4);
}
#define init_14_2_0_5
void FASTCALL call_14_2_0_5(SH2_struct* sh)
{
  SH2movi(sh, 2, (0<<4)+5);
}
#define init_14_2_0_6
void FASTCALL call_14_2_0_6(SH2_struct* sh)
{
  SH2movi(sh, 2, (0<<4)+6);
}
#define init_14_2_0_7
void FASTCALL call_14_2_0_7(SH2_struct* sh)
{
  SH2movi(sh, 2, (0<<4)+7);
}
#define init_14_2_0_8
void FASTCALL call_14_2_0_8(SH2_struct* sh)
{
  SH2movi(sh, 2, (0<<4)+8);
}
#define init_14_2_0_9
void FASTCALL call_14_2_0_9(SH2_struct* sh)
{
  SH2movi(sh, 2, (0<<4)+9);
}
#define init_14_2_0_10
void FASTCALL call_14_2_0_10(SH2_struct* sh)
{
  SH2movi(sh, 2, (0<<4)+10);
}
#define init_14_2_0_11
void FASTCALL call_14_2_0_11(SH2_struct* sh)
{
  SH2movi(sh, 2, (0<<4)+11);
}
#define init_14_2_0_12
void FASTCALL call_14_2_0_12(SH2_struct* sh)
{
  SH2movi(sh, 2, (0<<4)+12);
}
#define init_14_2_0_13
void FASTCALL call_14_2_0_13(SH2_struct* sh)
{
  SH2movi(sh, 2, (0<<4)+13);
}
#define init_14_2_0_14
void FASTCALL call_14_2_0_14(SH2_struct* sh)
{
  SH2movi(sh, 2, (0<<4)+14);
}
#define init_14_2_0_15
void FASTCALL call_14_2_0_15(SH2_struct* sh)
{
  SH2movi(sh, 2, (0<<4)+15);
}
#define init_14_2_1_0
void FASTCALL call_14_2_1_0(SH2_struct* sh)
{
  SH2movi(sh, 2, (1<<4)+0);
}
#define init_14_2_1_1
void FASTCALL call_14_2_1_1(SH2_struct* sh)
{
  SH2movi(sh, 2, (1<<4)+1);
}
#define init_14_2_1_2
void FASTCALL call_14_2_1_2(SH2_struct* sh)
{
  SH2movi(sh, 2, (1<<4)+2);
}
#define init_14_2_1_3
void FASTCALL call_14_2_1_3(SH2_struct* sh)
{
  SH2movi(sh, 2, (1<<4)+3);
}
#define init_14_2_1_4
void FASTCALL call_14_2_1_4(SH2_struct* sh)
{
  SH2movi(sh, 2, (1<<4)+4);
}
#define init_14_2_1_5
void FASTCALL call_14_2_1_5(SH2_struct* sh)
{
  SH2movi(sh, 2, (1<<4)+5);
}
#define init_14_2_1_6
void FASTCALL call_14_2_1_6(SH2_struct* sh)
{
  SH2movi(sh, 2, (1<<4)+6);
}
#define init_14_2_1_7
void FASTCALL call_14_2_1_7(SH2_struct* sh)
{
  SH2movi(sh, 2, (1<<4)+7);
}
#define init_14_2_1_8
void FASTCALL call_14_2_1_8(SH2_struct* sh)
{
  SH2movi(sh, 2, (1<<4)+8);
}
#define init_14_2_1_9
void FASTCALL call_14_2_1_9(SH2_struct* sh)
{
  SH2movi(sh, 2, (1<<4)+9);
}
#define init_14_2_1_10
void FASTCALL call_14_2_1_10(SH2_struct* sh)
{
  SH2movi(sh, 2, (1<<4)+10);
}
#define init_14_2_1_11
void FASTCALL call_14_2_1_11(SH2_struct* sh)
{
  SH2movi(sh, 2, (1<<4)+11);
}
#define init_14_2_1_12
void FASTCALL call_14_2_1_12(SH2_struct* sh)
{
  SH2movi(sh, 2, (1<<4)+12);
}
#define init_14_2_1_13
void FASTCALL call_14_2_1_13(SH2_struct* sh)
{
  SH2movi(sh, 2, (1<<4)+13);
}
#define init_14_2_1_14
void FASTCALL call_14_2_1_14(SH2_struct* sh)
{
  SH2movi(sh, 2, (1<<4)+14);
}
#define init_14_2_1_15
void FASTCALL call_14_2_1_15(SH2_struct* sh)
{
  SH2movi(sh, 2, (1<<4)+15);
}
#define init_14_2_2_0
void FASTCALL call_14_2_2_0(SH2_struct* sh)
{
  SH2movi(sh, 2, (2<<4)+0);
}
#define init_14_2_2_1
void FASTCALL call_14_2_2_1(SH2_struct* sh)
{
  SH2movi(sh, 2, (2<<4)+1);
}
#define init_14_2_2_2
void FASTCALL call_14_2_2_2(SH2_struct* sh)
{
  SH2movi(sh, 2, (2<<4)+2);
}
#define init_14_2_2_3
void FASTCALL call_14_2_2_3(SH2_struct* sh)
{
  SH2movi(sh, 2, (2<<4)+3);
}
#define init_14_2_2_4
void FASTCALL call_14_2_2_4(SH2_struct* sh)
{
  SH2movi(sh, 2, (2<<4)+4);
}
#define init_14_2_2_5
void FASTCALL call_14_2_2_5(SH2_struct* sh)
{
  SH2movi(sh, 2, (2<<4)+5);
}
#define init_14_2_2_6
void FASTCALL call_14_2_2_6(SH2_struct* sh)
{
  SH2movi(sh, 2, (2<<4)+6);
}
#define init_14_2_2_7
void FASTCALL call_14_2_2_7(SH2_struct* sh)
{
  SH2movi(sh, 2, (2<<4)+7);
}
#define init_14_2_2_8
void FASTCALL call_14_2_2_8(SH2_struct* sh)
{
  SH2movi(sh, 2, (2<<4)+8);
}
#define init_14_2_2_9
void FASTCALL call_14_2_2_9(SH2_struct* sh)
{
  SH2movi(sh, 2, (2<<4)+9);
}
#define init_14_2_2_10
void FASTCALL call_14_2_2_10(SH2_struct* sh)
{
  SH2movi(sh, 2, (2<<4)+10);
}
#define init_14_2_2_11
void FASTCALL call_14_2_2_11(SH2_struct* sh)
{
  SH2movi(sh, 2, (2<<4)+11);
}
#define init_14_2_2_12
void FASTCALL call_14_2_2_12(SH2_struct* sh)
{
  SH2movi(sh, 2, (2<<4)+12);
}
#define init_14_2_2_13
void FASTCALL call_14_2_2_13(SH2_struct* sh)
{
  SH2movi(sh, 2, (2<<4)+13);
}
#define init_14_2_2_14
void FASTCALL call_14_2_2_14(SH2_struct* sh)
{
  SH2movi(sh, 2, (2<<4)+14);
}
#define init_14_2_2_15
void FASTCALL call_14_2_2_15(SH2_struct* sh)
{
  SH2movi(sh, 2, (2<<4)+15);
}
#define init_14_2_3_0
void FASTCALL call_14_2_3_0(SH2_struct* sh)
{
  SH2movi(sh, 2, (3<<4)+0);
}
#define init_14_2_3_1
void FASTCALL call_14_2_3_1(SH2_struct* sh)
{
  SH2movi(sh, 2, (3<<4)+1);
}
#define init_14_2_3_2
void FASTCALL call_14_2_3_2(SH2_struct* sh)
{
  SH2movi(sh, 2, (3<<4)+2);
}
#define init_14_2_3_3
void FASTCALL call_14_2_3_3(SH2_struct* sh)
{
  SH2movi(sh, 2, (3<<4)+3);
}
#define init_14_2_3_4
void FASTCALL call_14_2_3_4(SH2_struct* sh)
{
  SH2movi(sh, 2, (3<<4)+4);
}
#define init_14_2_3_5
void FASTCALL call_14_2_3_5(SH2_struct* sh)
{
  SH2movi(sh, 2, (3<<4)+5);
}
#define init_14_2_3_6
void FASTCALL call_14_2_3_6(SH2_struct* sh)
{
  SH2movi(sh, 2, (3<<4)+6);
}
#define init_14_2_3_7
void FASTCALL call_14_2_3_7(SH2_struct* sh)
{
  SH2movi(sh, 2, (3<<4)+7);
}
#define init_14_2_3_8
void FASTCALL call_14_2_3_8(SH2_struct* sh)
{
  SH2movi(sh, 2, (3<<4)+8);
}
#define init_14_2_3_9
void FASTCALL call_14_2_3_9(SH2_struct* sh)
{
  SH2movi(sh, 2, (3<<4)+9);
}
#define init_14_2_3_10
void FASTCALL call_14_2_3_10(SH2_struct* sh)
{
  SH2movi(sh, 2, (3<<4)+10);
}
#define init_14_2_3_11
void FASTCALL call_14_2_3_11(SH2_struct* sh)
{
  SH2movi(sh, 2, (3<<4)+11);
}
#define init_14_2_3_12
void FASTCALL call_14_2_3_12(SH2_struct* sh)
{
  SH2movi(sh, 2, (3<<4)+12);
}
#define init_14_2_3_13
void FASTCALL call_14_2_3_13(SH2_struct* sh)
{
  SH2movi(sh, 2, (3<<4)+13);
}
#define init_14_2_3_14
void FASTCALL call_14_2_3_14(SH2_struct* sh)
{
  SH2movi(sh, 2, (3<<4)+14);
}
#define init_14_2_3_15
void FASTCALL call_14_2_3_15(SH2_struct* sh)
{
  SH2movi(sh, 2, (3<<4)+15);
}
#define init_14_2_4_0
void FASTCALL call_14_2_4_0(SH2_struct* sh)
{
  SH2movi(sh, 2, (4<<4)+0);
}
#define init_14_2_4_1
void FASTCALL call_14_2_4_1(SH2_struct* sh)
{
  SH2movi(sh, 2, (4<<4)+1);
}
#define init_14_2_4_2
void FASTCALL call_14_2_4_2(SH2_struct* sh)
{
  SH2movi(sh, 2, (4<<4)+2);
}
#define init_14_2_4_3
void FASTCALL call_14_2_4_3(SH2_struct* sh)
{
  SH2movi(sh, 2, (4<<4)+3);
}
#define init_14_2_4_4
void FASTCALL call_14_2_4_4(SH2_struct* sh)
{
  SH2movi(sh, 2, (4<<4)+4);
}
#define init_14_2_4_5
void FASTCALL call_14_2_4_5(SH2_struct* sh)
{
  SH2movi(sh, 2, (4<<4)+5);
}
#define init_14_2_4_6
void FASTCALL call_14_2_4_6(SH2_struct* sh)
{
  SH2movi(sh, 2, (4<<4)+6);
}
#define init_14_2_4_7
void FASTCALL call_14_2_4_7(SH2_struct* sh)
{
  SH2movi(sh, 2, (4<<4)+7);
}
#define init_14_2_4_8
void FASTCALL call_14_2_4_8(SH2_struct* sh)
{
  SH2movi(sh, 2, (4<<4)+8);
}
#define init_14_2_4_9
void FASTCALL call_14_2_4_9(SH2_struct* sh)
{
  SH2movi(sh, 2, (4<<4)+9);
}
#define init_14_2_4_10
void FASTCALL call_14_2_4_10(SH2_struct* sh)
{
  SH2movi(sh, 2, (4<<4)+10);
}
#define init_14_2_4_11
void FASTCALL call_14_2_4_11(SH2_struct* sh)
{
  SH2movi(sh, 2, (4<<4)+11);
}
#define init_14_2_4_12
void FASTCALL call_14_2_4_12(SH2_struct* sh)
{
  SH2movi(sh, 2, (4<<4)+12);
}
#define init_14_2_4_13
void FASTCALL call_14_2_4_13(SH2_struct* sh)
{
  SH2movi(sh, 2, (4<<4)+13);
}
#define init_14_2_4_14
void FASTCALL call_14_2_4_14(SH2_struct* sh)
{
  SH2movi(sh, 2, (4<<4)+14);
}
#define init_14_2_4_15
void FASTCALL call_14_2_4_15(SH2_struct* sh)
{
  SH2movi(sh, 2, (4<<4)+15);
}
#define init_14_2_5_0
void FASTCALL call_14_2_5_0(SH2_struct* sh)
{
  SH2movi(sh, 2, (5<<4)+0);
}
#define init_14_2_5_1
void FASTCALL call_14_2_5_1(SH2_struct* sh)
{
  SH2movi(sh, 2, (5<<4)+1);
}
#define init_14_2_5_2
void FASTCALL call_14_2_5_2(SH2_struct* sh)
{
  SH2movi(sh, 2, (5<<4)+2);
}
#define init_14_2_5_3
void FASTCALL call_14_2_5_3(SH2_struct* sh)
{
  SH2movi(sh, 2, (5<<4)+3);
}
#define init_14_2_5_4
void FASTCALL call_14_2_5_4(SH2_struct* sh)
{
  SH2movi(sh, 2, (5<<4)+4);
}
#define init_14_2_5_5
void FASTCALL call_14_2_5_5(SH2_struct* sh)
{
  SH2movi(sh, 2, (5<<4)+5);
}
#define init_14_2_5_6
void FASTCALL call_14_2_5_6(SH2_struct* sh)
{
  SH2movi(sh, 2, (5<<4)+6);
}
#define init_14_2_5_7
void FASTCALL call_14_2_5_7(SH2_struct* sh)
{
  SH2movi(sh, 2, (5<<4)+7);
}
#define init_14_2_5_8
void FASTCALL call_14_2_5_8(SH2_struct* sh)
{
  SH2movi(sh, 2, (5<<4)+8);
}
#define init_14_2_5_9
void FASTCALL call_14_2_5_9(SH2_struct* sh)
{
  SH2movi(sh, 2, (5<<4)+9);
}
#define init_14_2_5_10
void FASTCALL call_14_2_5_10(SH2_struct* sh)
{
  SH2movi(sh, 2, (5<<4)+10);
}
#define init_14_2_5_11
void FASTCALL call_14_2_5_11(SH2_struct* sh)
{
  SH2movi(sh, 2, (5<<4)+11);
}
#define init_14_2_5_12
void FASTCALL call_14_2_5_12(SH2_struct* sh)
{
  SH2movi(sh, 2, (5<<4)+12);
}
#define init_14_2_5_13
void FASTCALL call_14_2_5_13(SH2_struct* sh)
{
  SH2movi(sh, 2, (5<<4)+13);
}
#define init_14_2_5_14
void FASTCALL call_14_2_5_14(SH2_struct* sh)
{
  SH2movi(sh, 2, (5<<4)+14);
}
#define init_14_2_5_15
void FASTCALL call_14_2_5_15(SH2_struct* sh)
{
  SH2movi(sh, 2, (5<<4)+15);
}
#define init_14_2_6_0
void FASTCALL call_14_2_6_0(SH2_struct* sh)
{
  SH2movi(sh, 2, (6<<4)+0);
}
#define init_14_2_6_1
void FASTCALL call_14_2_6_1(SH2_struct* sh)
{
  SH2movi(sh, 2, (6<<4)+1);
}
#define init_14_2_6_2
void FASTCALL call_14_2_6_2(SH2_struct* sh)
{
  SH2movi(sh, 2, (6<<4)+2);
}
#define init_14_2_6_3
void FASTCALL call_14_2_6_3(SH2_struct* sh)
{
  SH2movi(sh, 2, (6<<4)+3);
}
#define init_14_2_6_4
void FASTCALL call_14_2_6_4(SH2_struct* sh)
{
  SH2movi(sh, 2, (6<<4)+4);
}
#define init_14_2_6_5
void FASTCALL call_14_2_6_5(SH2_struct* sh)
{
  SH2movi(sh, 2, (6<<4)+5);
}
#define init_14_2_6_6
void FASTCALL call_14_2_6_6(SH2_struct* sh)
{
  SH2movi(sh, 2, (6<<4)+6);
}
#define init_14_2_6_7
void FASTCALL call_14_2_6_7(SH2_struct* sh)
{
  SH2movi(sh, 2, (6<<4)+7);
}
#define init_14_2_6_8
void FASTCALL call_14_2_6_8(SH2_struct* sh)
{
  SH2movi(sh, 2, (6<<4)+8);
}
#define init_14_2_6_9
void FASTCALL call_14_2_6_9(SH2_struct* sh)
{
  SH2movi(sh, 2, (6<<4)+9);
}
#define init_14_2_6_10
void FASTCALL call_14_2_6_10(SH2_struct* sh)
{
  SH2movi(sh, 2, (6<<4)+10);
}
#define init_14_2_6_11
void FASTCALL call_14_2_6_11(SH2_struct* sh)
{
  SH2movi(sh, 2, (6<<4)+11);
}
#define init_14_2_6_12
void FASTCALL call_14_2_6_12(SH2_struct* sh)
{
  SH2movi(sh, 2, (6<<4)+12);
}
#define init_14_2_6_13
void FASTCALL call_14_2_6_13(SH2_struct* sh)
{
  SH2movi(sh, 2, (6<<4)+13);
}
#define init_14_2_6_14
void FASTCALL call_14_2_6_14(SH2_struct* sh)
{
  SH2movi(sh, 2, (6<<4)+14);
}
#define init_14_2_6_15
void FASTCALL call_14_2_6_15(SH2_struct* sh)
{
  SH2movi(sh, 2, (6<<4)+15);
}
#define init_14_2_7_0
void FASTCALL call_14_2_7_0(SH2_struct* sh)
{
  SH2movi(sh, 2, (7<<4)+0);
}
#define init_14_2_7_1
void FASTCALL call_14_2_7_1(SH2_struct* sh)
{
  SH2movi(sh, 2, (7<<4)+1);
}
#define init_14_2_7_2
void FASTCALL call_14_2_7_2(SH2_struct* sh)
{
  SH2movi(sh, 2, (7<<4)+2);
}
#define init_14_2_7_3
void FASTCALL call_14_2_7_3(SH2_struct* sh)
{
  SH2movi(sh, 2, (7<<4)+3);
}
#define init_14_2_7_4
void FASTCALL call_14_2_7_4(SH2_struct* sh)
{
  SH2movi(sh, 2, (7<<4)+4);
}
#define init_14_2_7_5
void FASTCALL call_14_2_7_5(SH2_struct* sh)
{
  SH2movi(sh, 2, (7<<4)+5);
}
#define init_14_2_7_6
void FASTCALL call_14_2_7_6(SH2_struct* sh)
{
  SH2movi(sh, 2, (7<<4)+6);
}
#define init_14_2_7_7
void FASTCALL call_14_2_7_7(SH2_struct* sh)
{
  SH2movi(sh, 2, (7<<4)+7);
}
#define init_14_2_7_8
void FASTCALL call_14_2_7_8(SH2_struct* sh)
{
  SH2movi(sh, 2, (7<<4)+8);
}
#define init_14_2_7_9
void FASTCALL call_14_2_7_9(SH2_struct* sh)
{
  SH2movi(sh, 2, (7<<4)+9);
}
#define init_14_2_7_10
void FASTCALL call_14_2_7_10(SH2_struct* sh)
{
  SH2movi(sh, 2, (7<<4)+10);
}
#define init_14_2_7_11
void FASTCALL call_14_2_7_11(SH2_struct* sh)
{
  SH2movi(sh, 2, (7<<4)+11);
}
#define init_14_2_7_12
void FASTCALL call_14_2_7_12(SH2_struct* sh)
{
  SH2movi(sh, 2, (7<<4)+12);
}
#define init_14_2_7_13
void FASTCALL call_14_2_7_13(SH2_struct* sh)
{
  SH2movi(sh, 2, (7<<4)+13);
}
#define init_14_2_7_14
void FASTCALL call_14_2_7_14(SH2_struct* sh)
{
  SH2movi(sh, 2, (7<<4)+14);
}
#define init_14_2_7_15
void FASTCALL call_14_2_7_15(SH2_struct* sh)
{
  SH2movi(sh, 2, (7<<4)+15);
}
#define init_14_2_8_0
void FASTCALL call_14_2_8_0(SH2_struct* sh)
{
  SH2movi(sh, 2, (8<<4)+0);
}
#define init_14_2_8_1
void FASTCALL call_14_2_8_1(SH2_struct* sh)
{
  SH2movi(sh, 2, (8<<4)+1);
}
#define init_14_2_8_2
void FASTCALL call_14_2_8_2(SH2_struct* sh)
{
  SH2movi(sh, 2, (8<<4)+2);
}
#define init_14_2_8_3
void FASTCALL call_14_2_8_3(SH2_struct* sh)
{
  SH2movi(sh, 2, (8<<4)+3);
}
#define init_14_2_8_4
void FASTCALL call_14_2_8_4(SH2_struct* sh)
{
  SH2movi(sh, 2, (8<<4)+4);
}
#define init_14_2_8_5
void FASTCALL call_14_2_8_5(SH2_struct* sh)
{
  SH2movi(sh, 2, (8<<4)+5);
}
#define init_14_2_8_6
void FASTCALL call_14_2_8_6(SH2_struct* sh)
{
  SH2movi(sh, 2, (8<<4)+6);
}
#define init_14_2_8_7
void FASTCALL call_14_2_8_7(SH2_struct* sh)
{
  SH2movi(sh, 2, (8<<4)+7);
}
#define init_14_2_8_8
void FASTCALL call_14_2_8_8(SH2_struct* sh)
{
  SH2movi(sh, 2, (8<<4)+8);
}
#define init_14_2_8_9
void FASTCALL call_14_2_8_9(SH2_struct* sh)
{
  SH2movi(sh, 2, (8<<4)+9);
}
#define init_14_2_8_10
void FASTCALL call_14_2_8_10(SH2_struct* sh)
{
  SH2movi(sh, 2, (8<<4)+10);
}
#define init_14_2_8_11
void FASTCALL call_14_2_8_11(SH2_struct* sh)
{
  SH2movi(sh, 2, (8<<4)+11);
}
#define init_14_2_8_12
void FASTCALL call_14_2_8_12(SH2_struct* sh)
{
  SH2movi(sh, 2, (8<<4)+12);
}
#define init_14_2_8_13
void FASTCALL call_14_2_8_13(SH2_struct* sh)
{
  SH2movi(sh, 2, (8<<4)+13);
}
#define init_14_2_8_14
void FASTCALL call_14_2_8_14(SH2_struct* sh)
{
  SH2movi(sh, 2, (8<<4)+14);
}
#define init_14_2_8_15
void FASTCALL call_14_2_8_15(SH2_struct* sh)
{
  SH2movi(sh, 2, (8<<4)+15);
}
#define init_14_2_9_0
void FASTCALL call_14_2_9_0(SH2_struct* sh)
{
  SH2movi(sh, 2, (9<<4)+0);
}
#define init_14_2_9_1
void FASTCALL call_14_2_9_1(SH2_struct* sh)
{
  SH2movi(sh, 2, (9<<4)+1);
}
#define init_14_2_9_2
void FASTCALL call_14_2_9_2(SH2_struct* sh)
{
  SH2movi(sh, 2, (9<<4)+2);
}
#define init_14_2_9_3
void FASTCALL call_14_2_9_3(SH2_struct* sh)
{
  SH2movi(sh, 2, (9<<4)+3);
}
#define init_14_2_9_4
void FASTCALL call_14_2_9_4(SH2_struct* sh)
{
  SH2movi(sh, 2, (9<<4)+4);
}
#define init_14_2_9_5
void FASTCALL call_14_2_9_5(SH2_struct* sh)
{
  SH2movi(sh, 2, (9<<4)+5);
}
#define init_14_2_9_6
void FASTCALL call_14_2_9_6(SH2_struct* sh)
{
  SH2movi(sh, 2, (9<<4)+6);
}
#define init_14_2_9_7
void FASTCALL call_14_2_9_7(SH2_struct* sh)
{
  SH2movi(sh, 2, (9<<4)+7);
}
#define init_14_2_9_8
void FASTCALL call_14_2_9_8(SH2_struct* sh)
{
  SH2movi(sh, 2, (9<<4)+8);
}
#define init_14_2_9_9
void FASTCALL call_14_2_9_9(SH2_struct* sh)
{
  SH2movi(sh, 2, (9<<4)+9);
}
#define init_14_2_9_10
void FASTCALL call_14_2_9_10(SH2_struct* sh)
{
  SH2movi(sh, 2, (9<<4)+10);
}
#define init_14_2_9_11
void FASTCALL call_14_2_9_11(SH2_struct* sh)
{
  SH2movi(sh, 2, (9<<4)+11);
}
#define init_14_2_9_12
void FASTCALL call_14_2_9_12(SH2_struct* sh)
{
  SH2movi(sh, 2, (9<<4)+12);
}
#define init_14_2_9_13
void FASTCALL call_14_2_9_13(SH2_struct* sh)
{
  SH2movi(sh, 2, (9<<4)+13);
}
#define init_14_2_9_14
void FASTCALL call_14_2_9_14(SH2_struct* sh)
{
  SH2movi(sh, 2, (9<<4)+14);
}
#define init_14_2_9_15
void FASTCALL call_14_2_9_15(SH2_struct* sh)
{
  SH2movi(sh, 2, (9<<4)+15);
}
#define init_14_2_10_0
void FASTCALL call_14_2_10_0(SH2_struct* sh)
{
  SH2movi(sh, 2, (10<<4)+0);
}
#define init_14_2_10_1
void FASTCALL call_14_2_10_1(SH2_struct* sh)
{
  SH2movi(sh, 2, (10<<4)+1);
}
#define init_14_2_10_2
void FASTCALL call_14_2_10_2(SH2_struct* sh)
{
  SH2movi(sh, 2, (10<<4)+2);
}
#define init_14_2_10_3
void FASTCALL call_14_2_10_3(SH2_struct* sh)
{
  SH2movi(sh, 2, (10<<4)+3);
}
#define init_14_2_10_4
void FASTCALL call_14_2_10_4(SH2_struct* sh)
{
  SH2movi(sh, 2, (10<<4)+4);
}
#define init_14_2_10_5
void FASTCALL call_14_2_10_5(SH2_struct* sh)
{
  SH2movi(sh, 2, (10<<4)+5);
}
#define init_14_2_10_6
void FASTCALL call_14_2_10_6(SH2_struct* sh)
{
  SH2movi(sh, 2, (10<<4)+6);
}
#define init_14_2_10_7
void FASTCALL call_14_2_10_7(SH2_struct* sh)
{
  SH2movi(sh, 2, (10<<4)+7);
}
#define init_14_2_10_8
void FASTCALL call_14_2_10_8(SH2_struct* sh)
{
  SH2movi(sh, 2, (10<<4)+8);
}
#define init_14_2_10_9
void FASTCALL call_14_2_10_9(SH2_struct* sh)
{
  SH2movi(sh, 2, (10<<4)+9);
}
#define init_14_2_10_10
void FASTCALL call_14_2_10_10(SH2_struct* sh)
{
  SH2movi(sh, 2, (10<<4)+10);
}
#define init_14_2_10_11
void FASTCALL call_14_2_10_11(SH2_struct* sh)
{
  SH2movi(sh, 2, (10<<4)+11);
}
#define init_14_2_10_12
void FASTCALL call_14_2_10_12(SH2_struct* sh)
{
  SH2movi(sh, 2, (10<<4)+12);
}
#define init_14_2_10_13
void FASTCALL call_14_2_10_13(SH2_struct* sh)
{
  SH2movi(sh, 2, (10<<4)+13);
}
#define init_14_2_10_14
void FASTCALL call_14_2_10_14(SH2_struct* sh)
{
  SH2movi(sh, 2, (10<<4)+14);
}
#define init_14_2_10_15
void FASTCALL call_14_2_10_15(SH2_struct* sh)
{
  SH2movi(sh, 2, (10<<4)+15);
}
#define init_14_2_11_0
void FASTCALL call_14_2_11_0(SH2_struct* sh)
{
  SH2movi(sh, 2, (11<<4)+0);
}
#define init_14_2_11_1
void FASTCALL call_14_2_11_1(SH2_struct* sh)
{
  SH2movi(sh, 2, (11<<4)+1);
}
#define init_14_2_11_2
void FASTCALL call_14_2_11_2(SH2_struct* sh)
{
  SH2movi(sh, 2, (11<<4)+2);
}
#define init_14_2_11_3
void FASTCALL call_14_2_11_3(SH2_struct* sh)
{
  SH2movi(sh, 2, (11<<4)+3);
}
#define init_14_2_11_4
void FASTCALL call_14_2_11_4(SH2_struct* sh)
{
  SH2movi(sh, 2, (11<<4)+4);
}
#define init_14_2_11_5
void FASTCALL call_14_2_11_5(SH2_struct* sh)
{
  SH2movi(sh, 2, (11<<4)+5);
}
#define init_14_2_11_6
void FASTCALL call_14_2_11_6(SH2_struct* sh)
{
  SH2movi(sh, 2, (11<<4)+6);
}
#define init_14_2_11_7
void FASTCALL call_14_2_11_7(SH2_struct* sh)
{
  SH2movi(sh, 2, (11<<4)+7);
}
#define init_14_2_11_8
void FASTCALL call_14_2_11_8(SH2_struct* sh)
{
  SH2movi(sh, 2, (11<<4)+8);
}
#define init_14_2_11_9
void FASTCALL call_14_2_11_9(SH2_struct* sh)
{
  SH2movi(sh, 2, (11<<4)+9);
}
#define init_14_2_11_10
void FASTCALL call_14_2_11_10(SH2_struct* sh)
{
  SH2movi(sh, 2, (11<<4)+10);
}
#define init_14_2_11_11
void FASTCALL call_14_2_11_11(SH2_struct* sh)
{
  SH2movi(sh, 2, (11<<4)+11);
}
#define init_14_2_11_12
void FASTCALL call_14_2_11_12(SH2_struct* sh)
{
  SH2movi(sh, 2, (11<<4)+12);
}
#define init_14_2_11_13
void FASTCALL call_14_2_11_13(SH2_struct* sh)
{
  SH2movi(sh, 2, (11<<4)+13);
}
#define init_14_2_11_14
void FASTCALL call_14_2_11_14(SH2_struct* sh)
{
  SH2movi(sh, 2, (11<<4)+14);
}
#define init_14_2_11_15
void FASTCALL call_14_2_11_15(SH2_struct* sh)
{
  SH2movi(sh, 2, (11<<4)+15);
}
#define init_14_2_12_0
void FASTCALL call_14_2_12_0(SH2_struct* sh)
{
  SH2movi(sh, 2, (12<<4)+0);
}
#define init_14_2_12_1
void FASTCALL call_14_2_12_1(SH2_struct* sh)
{
  SH2movi(sh, 2, (12<<4)+1);
}
#define init_14_2_12_2
void FASTCALL call_14_2_12_2(SH2_struct* sh)
{
  SH2movi(sh, 2, (12<<4)+2);
}
#define init_14_2_12_3
void FASTCALL call_14_2_12_3(SH2_struct* sh)
{
  SH2movi(sh, 2, (12<<4)+3);
}
#define init_14_2_12_4
void FASTCALL call_14_2_12_4(SH2_struct* sh)
{
  SH2movi(sh, 2, (12<<4)+4);
}
#define init_14_2_12_5
void FASTCALL call_14_2_12_5(SH2_struct* sh)
{
  SH2movi(sh, 2, (12<<4)+5);
}
#define init_14_2_12_6
void FASTCALL call_14_2_12_6(SH2_struct* sh)
{
  SH2movi(sh, 2, (12<<4)+6);
}
#define init_14_2_12_7
void FASTCALL call_14_2_12_7(SH2_struct* sh)
{
  SH2movi(sh, 2, (12<<4)+7);
}
#define init_14_2_12_8
void FASTCALL call_14_2_12_8(SH2_struct* sh)
{
  SH2movi(sh, 2, (12<<4)+8);
}
#define init_14_2_12_9
void FASTCALL call_14_2_12_9(SH2_struct* sh)
{
  SH2movi(sh, 2, (12<<4)+9);
}
#define init_14_2_12_10
void FASTCALL call_14_2_12_10(SH2_struct* sh)
{
  SH2movi(sh, 2, (12<<4)+10);
}
#define init_14_2_12_11
void FASTCALL call_14_2_12_11(SH2_struct* sh)
{
  SH2movi(sh, 2, (12<<4)+11);
}
#define init_14_2_12_12
void FASTCALL call_14_2_12_12(SH2_struct* sh)
{
  SH2movi(sh, 2, (12<<4)+12);
}
#define init_14_2_12_13
void FASTCALL call_14_2_12_13(SH2_struct* sh)
{
  SH2movi(sh, 2, (12<<4)+13);
}
#define init_14_2_12_14
void FASTCALL call_14_2_12_14(SH2_struct* sh)
{
  SH2movi(sh, 2, (12<<4)+14);
}
#define init_14_2_12_15
void FASTCALL call_14_2_12_15(SH2_struct* sh)
{
  SH2movi(sh, 2, (12<<4)+15);
}
#define init_14_2_13_0
void FASTCALL call_14_2_13_0(SH2_struct* sh)
{
  SH2movi(sh, 2, (13<<4)+0);
}
#define init_14_2_13_1
void FASTCALL call_14_2_13_1(SH2_struct* sh)
{
  SH2movi(sh, 2, (13<<4)+1);
}
#define init_14_2_13_2
void FASTCALL call_14_2_13_2(SH2_struct* sh)
{
  SH2movi(sh, 2, (13<<4)+2);
}
#define init_14_2_13_3
void FASTCALL call_14_2_13_3(SH2_struct* sh)
{
  SH2movi(sh, 2, (13<<4)+3);
}
#define init_14_2_13_4
void FASTCALL call_14_2_13_4(SH2_struct* sh)
{
  SH2movi(sh, 2, (13<<4)+4);
}
#define init_14_2_13_5
void FASTCALL call_14_2_13_5(SH2_struct* sh)
{
  SH2movi(sh, 2, (13<<4)+5);
}
#define init_14_2_13_6
void FASTCALL call_14_2_13_6(SH2_struct* sh)
{
  SH2movi(sh, 2, (13<<4)+6);
}
#define init_14_2_13_7
void FASTCALL call_14_2_13_7(SH2_struct* sh)
{
  SH2movi(sh, 2, (13<<4)+7);
}
#define init_14_2_13_8
void FASTCALL call_14_2_13_8(SH2_struct* sh)
{
  SH2movi(sh, 2, (13<<4)+8);
}
#define init_14_2_13_9
void FASTCALL call_14_2_13_9(SH2_struct* sh)
{
  SH2movi(sh, 2, (13<<4)+9);
}
#define init_14_2_13_10
void FASTCALL call_14_2_13_10(SH2_struct* sh)
{
  SH2movi(sh, 2, (13<<4)+10);
}
#define init_14_2_13_11
void FASTCALL call_14_2_13_11(SH2_struct* sh)
{
  SH2movi(sh, 2, (13<<4)+11);
}
#define init_14_2_13_12
void FASTCALL call_14_2_13_12(SH2_struct* sh)
{
  SH2movi(sh, 2, (13<<4)+12);
}
#define init_14_2_13_13
void FASTCALL call_14_2_13_13(SH2_struct* sh)
{
  SH2movi(sh, 2, (13<<4)+13);
}
#define init_14_2_13_14
void FASTCALL call_14_2_13_14(SH2_struct* sh)
{
  SH2movi(sh, 2, (13<<4)+14);
}
#define init_14_2_13_15
void FASTCALL call_14_2_13_15(SH2_struct* sh)
{
  SH2movi(sh, 2, (13<<4)+15);
}
#define init_14_2_14_0
void FASTCALL call_14_2_14_0(SH2_struct* sh)
{
  SH2movi(sh, 2, (14<<4)+0);
}
#define init_14_2_14_1
void FASTCALL call_14_2_14_1(SH2_struct* sh)
{
  SH2movi(sh, 2, (14<<4)+1);
}
#define init_14_2_14_2
void FASTCALL call_14_2_14_2(SH2_struct* sh)
{
  SH2movi(sh, 2, (14<<4)+2);
}
#define init_14_2_14_3
void FASTCALL call_14_2_14_3(SH2_struct* sh)
{
  SH2movi(sh, 2, (14<<4)+3);
}
#define init_14_2_14_4
void FASTCALL call_14_2_14_4(SH2_struct* sh)
{
  SH2movi(sh, 2, (14<<4)+4);
}
#define init_14_2_14_5
void FASTCALL call_14_2_14_5(SH2_struct* sh)
{
  SH2movi(sh, 2, (14<<4)+5);
}
#define init_14_2_14_6
void FASTCALL call_14_2_14_6(SH2_struct* sh)
{
  SH2movi(sh, 2, (14<<4)+6);
}
#define init_14_2_14_7
void FASTCALL call_14_2_14_7(SH2_struct* sh)
{
  SH2movi(sh, 2, (14<<4)+7);
}
#define init_14_2_14_8
void FASTCALL call_14_2_14_8(SH2_struct* sh)
{
  SH2movi(sh, 2, (14<<4)+8);
}
#define init_14_2_14_9
void FASTCALL call_14_2_14_9(SH2_struct* sh)
{
  SH2movi(sh, 2, (14<<4)+9);
}
#define init_14_2_14_10
void FASTCALL call_14_2_14_10(SH2_struct* sh)
{
  SH2movi(sh, 2, (14<<4)+10);
}
#define init_14_2_14_11
void FASTCALL call_14_2_14_11(SH2_struct* sh)
{
  SH2movi(sh, 2, (14<<4)+11);
}
#define init_14_2_14_12
void FASTCALL call_14_2_14_12(SH2_struct* sh)
{
  SH2movi(sh, 2, (14<<4)+12);
}
#define init_14_2_14_13
void FASTCALL call_14_2_14_13(SH2_struct* sh)
{
  SH2movi(sh, 2, (14<<4)+13);
}
#define init_14_2_14_14
void FASTCALL call_14_2_14_14(SH2_struct* sh)
{
  SH2movi(sh, 2, (14<<4)+14);
}
#define init_14_2_14_15
void FASTCALL call_14_2_14_15(SH2_struct* sh)
{
  SH2movi(sh, 2, (14<<4)+15);
}
#define init_14_2_15_0
void FASTCALL call_14_2_15_0(SH2_struct* sh)
{
  SH2movi(sh, 2, (15<<4)+0);
}
#define init_14_2_15_1
void FASTCALL call_14_2_15_1(SH2_struct* sh)
{
  SH2movi(sh, 2, (15<<4)+1);
}
#define init_14_2_15_2
void FASTCALL call_14_2_15_2(SH2_struct* sh)
{
  SH2movi(sh, 2, (15<<4)+2);
}
#define init_14_2_15_3
void FASTCALL call_14_2_15_3(SH2_struct* sh)
{
  SH2movi(sh, 2, (15<<4)+3);
}
#define init_14_2_15_4
void FASTCALL call_14_2_15_4(SH2_struct* sh)
{
  SH2movi(sh, 2, (15<<4)+4);
}
#define init_14_2_15_5
void FASTCALL call_14_2_15_5(SH2_struct* sh)
{
  SH2movi(sh, 2, (15<<4)+5);
}
#define init_14_2_15_6
void FASTCALL call_14_2_15_6(SH2_struct* sh)
{
  SH2movi(sh, 2, (15<<4)+6);
}
#define init_14_2_15_7
void FASTCALL call_14_2_15_7(SH2_struct* sh)
{
  SH2movi(sh, 2, (15<<4)+7);
}
#define init_14_2_15_8
void FASTCALL call_14_2_15_8(SH2_struct* sh)
{
  SH2movi(sh, 2, (15<<4)+8);
}
#define init_14_2_15_9
void FASTCALL call_14_2_15_9(SH2_struct* sh)
{
  SH2movi(sh, 2, (15<<4)+9);
}
#define init_14_2_15_10
void FASTCALL call_14_2_15_10(SH2_struct* sh)
{
  SH2movi(sh, 2, (15<<4)+10);
}
#define init_14_2_15_11
void FASTCALL call_14_2_15_11(SH2_struct* sh)
{
  SH2movi(sh, 2, (15<<4)+11);
}
#define init_14_2_15_12
void FASTCALL call_14_2_15_12(SH2_struct* sh)
{
  SH2movi(sh, 2, (15<<4)+12);
}
#define init_14_2_15_13
void FASTCALL call_14_2_15_13(SH2_struct* sh)
{
  SH2movi(sh, 2, (15<<4)+13);
}
#define init_14_2_15_14
void FASTCALL call_14_2_15_14(SH2_struct* sh)
{
  SH2movi(sh, 2, (15<<4)+14);
}
#define init_14_2_15_15
void FASTCALL call_14_2_15_15(SH2_struct* sh)
{
  SH2movi(sh, 2, (15<<4)+15);
}
#define init_14_3_0_0
void FASTCALL call_14_3_0_0(SH2_struct* sh)
{
  SH2movi(sh, 3, (0<<4)+0);
}
#define init_14_3_0_1
void FASTCALL call_14_3_0_1(SH2_struct* sh)
{
  SH2movi(sh, 3, (0<<4)+1);
}
#define init_14_3_0_2
void FASTCALL call_14_3_0_2(SH2_struct* sh)
{
  SH2movi(sh, 3, (0<<4)+2);
}
#define init_14_3_0_3
void FASTCALL call_14_3_0_3(SH2_struct* sh)
{
  SH2movi(sh, 3, (0<<4)+3);
}
#define init_14_3_0_4
void FASTCALL call_14_3_0_4(SH2_struct* sh)
{
  SH2movi(sh, 3, (0<<4)+4);
}
#define init_14_3_0_5
void FASTCALL call_14_3_0_5(SH2_struct* sh)
{
  SH2movi(sh, 3, (0<<4)+5);
}
#define init_14_3_0_6
void FASTCALL call_14_3_0_6(SH2_struct* sh)
{
  SH2movi(sh, 3, (0<<4)+6);
}
#define init_14_3_0_7
void FASTCALL call_14_3_0_7(SH2_struct* sh)
{
  SH2movi(sh, 3, (0<<4)+7);
}
#define init_14_3_0_8
void FASTCALL call_14_3_0_8(SH2_struct* sh)
{
  SH2movi(sh, 3, (0<<4)+8);
}
#define init_14_3_0_9
void FASTCALL call_14_3_0_9(SH2_struct* sh)
{
  SH2movi(sh, 3, (0<<4)+9);
}
#define init_14_3_0_10
void FASTCALL call_14_3_0_10(SH2_struct* sh)
{
  SH2movi(sh, 3, (0<<4)+10);
}
#define init_14_3_0_11
void FASTCALL call_14_3_0_11(SH2_struct* sh)
{
  SH2movi(sh, 3, (0<<4)+11);
}
#define init_14_3_0_12
void FASTCALL call_14_3_0_12(SH2_struct* sh)
{
  SH2movi(sh, 3, (0<<4)+12);
}
#define init_14_3_0_13
void FASTCALL call_14_3_0_13(SH2_struct* sh)
{
  SH2movi(sh, 3, (0<<4)+13);
}
#define init_14_3_0_14
void FASTCALL call_14_3_0_14(SH2_struct* sh)
{
  SH2movi(sh, 3, (0<<4)+14);
}
#define init_14_3_0_15
void FASTCALL call_14_3_0_15(SH2_struct* sh)
{
  SH2movi(sh, 3, (0<<4)+15);
}
#define init_14_3_1_0
void FASTCALL call_14_3_1_0(SH2_struct* sh)
{
  SH2movi(sh, 3, (1<<4)+0);
}
#define init_14_3_1_1
void FASTCALL call_14_3_1_1(SH2_struct* sh)
{
  SH2movi(sh, 3, (1<<4)+1);
}
#define init_14_3_1_2
void FASTCALL call_14_3_1_2(SH2_struct* sh)
{
  SH2movi(sh, 3, (1<<4)+2);
}
#define init_14_3_1_3
void FASTCALL call_14_3_1_3(SH2_struct* sh)
{
  SH2movi(sh, 3, (1<<4)+3);
}
#define init_14_3_1_4
void FASTCALL call_14_3_1_4(SH2_struct* sh)
{
  SH2movi(sh, 3, (1<<4)+4);
}
#define init_14_3_1_5
void FASTCALL call_14_3_1_5(SH2_struct* sh)
{
  SH2movi(sh, 3, (1<<4)+5);
}
#define init_14_3_1_6
void FASTCALL call_14_3_1_6(SH2_struct* sh)
{
  SH2movi(sh, 3, (1<<4)+6);
}
#define init_14_3_1_7
void FASTCALL call_14_3_1_7(SH2_struct* sh)
{
  SH2movi(sh, 3, (1<<4)+7);
}
#define init_14_3_1_8
void FASTCALL call_14_3_1_8(SH2_struct* sh)
{
  SH2movi(sh, 3, (1<<4)+8);
}
#define init_14_3_1_9
void FASTCALL call_14_3_1_9(SH2_struct* sh)
{
  SH2movi(sh, 3, (1<<4)+9);
}
#define init_14_3_1_10
void FASTCALL call_14_3_1_10(SH2_struct* sh)
{
  SH2movi(sh, 3, (1<<4)+10);
}
#define init_14_3_1_11
void FASTCALL call_14_3_1_11(SH2_struct* sh)
{
  SH2movi(sh, 3, (1<<4)+11);
}
#define init_14_3_1_12
void FASTCALL call_14_3_1_12(SH2_struct* sh)
{
  SH2movi(sh, 3, (1<<4)+12);
}
#define init_14_3_1_13
void FASTCALL call_14_3_1_13(SH2_struct* sh)
{
  SH2movi(sh, 3, (1<<4)+13);
}
#define init_14_3_1_14
void FASTCALL call_14_3_1_14(SH2_struct* sh)
{
  SH2movi(sh, 3, (1<<4)+14);
}
#define init_14_3_1_15
void FASTCALL call_14_3_1_15(SH2_struct* sh)
{
  SH2movi(sh, 3, (1<<4)+15);
}
#define init_14_3_2_0
void FASTCALL call_14_3_2_0(SH2_struct* sh)
{
  SH2movi(sh, 3, (2<<4)+0);
}
#define init_14_3_2_1
void FASTCALL call_14_3_2_1(SH2_struct* sh)
{
  SH2movi(sh, 3, (2<<4)+1);
}
#define init_14_3_2_2
void FASTCALL call_14_3_2_2(SH2_struct* sh)
{
  SH2movi(sh, 3, (2<<4)+2);
}
#define init_14_3_2_3
void FASTCALL call_14_3_2_3(SH2_struct* sh)
{
  SH2movi(sh, 3, (2<<4)+3);
}
#define init_14_3_2_4
void FASTCALL call_14_3_2_4(SH2_struct* sh)
{
  SH2movi(sh, 3, (2<<4)+4);
}
#define init_14_3_2_5
void FASTCALL call_14_3_2_5(SH2_struct* sh)
{
  SH2movi(sh, 3, (2<<4)+5);
}
#define init_14_3_2_6
void FASTCALL call_14_3_2_6(SH2_struct* sh)
{
  SH2movi(sh, 3, (2<<4)+6);
}
#define init_14_3_2_7
void FASTCALL call_14_3_2_7(SH2_struct* sh)
{
  SH2movi(sh, 3, (2<<4)+7);
}
#define init_14_3_2_8
void FASTCALL call_14_3_2_8(SH2_struct* sh)
{
  SH2movi(sh, 3, (2<<4)+8);
}
#define init_14_3_2_9
void FASTCALL call_14_3_2_9(SH2_struct* sh)
{
  SH2movi(sh, 3, (2<<4)+9);
}
#define init_14_3_2_10
void FASTCALL call_14_3_2_10(SH2_struct* sh)
{
  SH2movi(sh, 3, (2<<4)+10);
}
#define init_14_3_2_11
void FASTCALL call_14_3_2_11(SH2_struct* sh)
{
  SH2movi(sh, 3, (2<<4)+11);
}
#define init_14_3_2_12
void FASTCALL call_14_3_2_12(SH2_struct* sh)
{
  SH2movi(sh, 3, (2<<4)+12);
}
#define init_14_3_2_13
void FASTCALL call_14_3_2_13(SH2_struct* sh)
{
  SH2movi(sh, 3, (2<<4)+13);
}
#define init_14_3_2_14
void FASTCALL call_14_3_2_14(SH2_struct* sh)
{
  SH2movi(sh, 3, (2<<4)+14);
}
#define init_14_3_2_15
void FASTCALL call_14_3_2_15(SH2_struct* sh)
{
  SH2movi(sh, 3, (2<<4)+15);
}
#define init_14_3_3_0
void FASTCALL call_14_3_3_0(SH2_struct* sh)
{
  SH2movi(sh, 3, (3<<4)+0);
}
#define init_14_3_3_1
void FASTCALL call_14_3_3_1(SH2_struct* sh)
{
  SH2movi(sh, 3, (3<<4)+1);
}
#define init_14_3_3_2
void FASTCALL call_14_3_3_2(SH2_struct* sh)
{
  SH2movi(sh, 3, (3<<4)+2);
}
#define init_14_3_3_3
void FASTCALL call_14_3_3_3(SH2_struct* sh)
{
  SH2movi(sh, 3, (3<<4)+3);
}
#define init_14_3_3_4
void FASTCALL call_14_3_3_4(SH2_struct* sh)
{
  SH2movi(sh, 3, (3<<4)+4);
}
#define init_14_3_3_5
void FASTCALL call_14_3_3_5(SH2_struct* sh)
{
  SH2movi(sh, 3, (3<<4)+5);
}
#define init_14_3_3_6
void FASTCALL call_14_3_3_6(SH2_struct* sh)
{
  SH2movi(sh, 3, (3<<4)+6);
}
#define init_14_3_3_7
void FASTCALL call_14_3_3_7(SH2_struct* sh)
{
  SH2movi(sh, 3, (3<<4)+7);
}
#define init_14_3_3_8
void FASTCALL call_14_3_3_8(SH2_struct* sh)
{
  SH2movi(sh, 3, (3<<4)+8);
}
#define init_14_3_3_9
void FASTCALL call_14_3_3_9(SH2_struct* sh)
{
  SH2movi(sh, 3, (3<<4)+9);
}
#define init_14_3_3_10
void FASTCALL call_14_3_3_10(SH2_struct* sh)
{
  SH2movi(sh, 3, (3<<4)+10);
}
#define init_14_3_3_11
void FASTCALL call_14_3_3_11(SH2_struct* sh)
{
  SH2movi(sh, 3, (3<<4)+11);
}
#define init_14_3_3_12
void FASTCALL call_14_3_3_12(SH2_struct* sh)
{
  SH2movi(sh, 3, (3<<4)+12);
}
#define init_14_3_3_13
void FASTCALL call_14_3_3_13(SH2_struct* sh)
{
  SH2movi(sh, 3, (3<<4)+13);
}
#define init_14_3_3_14
void FASTCALL call_14_3_3_14(SH2_struct* sh)
{
  SH2movi(sh, 3, (3<<4)+14);
}
#define init_14_3_3_15
void FASTCALL call_14_3_3_15(SH2_struct* sh)
{
  SH2movi(sh, 3, (3<<4)+15);
}
#define init_14_3_4_0
void FASTCALL call_14_3_4_0(SH2_struct* sh)
{
  SH2movi(sh, 3, (4<<4)+0);
}
#define init_14_3_4_1
void FASTCALL call_14_3_4_1(SH2_struct* sh)
{
  SH2movi(sh, 3, (4<<4)+1);
}
#define init_14_3_4_2
void FASTCALL call_14_3_4_2(SH2_struct* sh)
{
  SH2movi(sh, 3, (4<<4)+2);
}
#define init_14_3_4_3
void FASTCALL call_14_3_4_3(SH2_struct* sh)
{
  SH2movi(sh, 3, (4<<4)+3);
}
#define init_14_3_4_4
void FASTCALL call_14_3_4_4(SH2_struct* sh)
{
  SH2movi(sh, 3, (4<<4)+4);
}
#define init_14_3_4_5
void FASTCALL call_14_3_4_5(SH2_struct* sh)
{
  SH2movi(sh, 3, (4<<4)+5);
}
#define init_14_3_4_6
void FASTCALL call_14_3_4_6(SH2_struct* sh)
{
  SH2movi(sh, 3, (4<<4)+6);
}
#define init_14_3_4_7
void FASTCALL call_14_3_4_7(SH2_struct* sh)
{
  SH2movi(sh, 3, (4<<4)+7);
}
#define init_14_3_4_8
void FASTCALL call_14_3_4_8(SH2_struct* sh)
{
  SH2movi(sh, 3, (4<<4)+8);
}
#define init_14_3_4_9
void FASTCALL call_14_3_4_9(SH2_struct* sh)
{
  SH2movi(sh, 3, (4<<4)+9);
}
#define init_14_3_4_10
void FASTCALL call_14_3_4_10(SH2_struct* sh)
{
  SH2movi(sh, 3, (4<<4)+10);
}
#define init_14_3_4_11
void FASTCALL call_14_3_4_11(SH2_struct* sh)
{
  SH2movi(sh, 3, (4<<4)+11);
}
#define init_14_3_4_12
void FASTCALL call_14_3_4_12(SH2_struct* sh)
{
  SH2movi(sh, 3, (4<<4)+12);
}
#define init_14_3_4_13
void FASTCALL call_14_3_4_13(SH2_struct* sh)
{
  SH2movi(sh, 3, (4<<4)+13);
}
#define init_14_3_4_14
void FASTCALL call_14_3_4_14(SH2_struct* sh)
{
  SH2movi(sh, 3, (4<<4)+14);
}
#define init_14_3_4_15
void FASTCALL call_14_3_4_15(SH2_struct* sh)
{
  SH2movi(sh, 3, (4<<4)+15);
}
#define init_14_3_5_0
void FASTCALL call_14_3_5_0(SH2_struct* sh)
{
  SH2movi(sh, 3, (5<<4)+0);
}
#define init_14_3_5_1
void FASTCALL call_14_3_5_1(SH2_struct* sh)
{
  SH2movi(sh, 3, (5<<4)+1);
}
#define init_14_3_5_2
void FASTCALL call_14_3_5_2(SH2_struct* sh)
{
  SH2movi(sh, 3, (5<<4)+2);
}
#define init_14_3_5_3
void FASTCALL call_14_3_5_3(SH2_struct* sh)
{
  SH2movi(sh, 3, (5<<4)+3);
}
#define init_14_3_5_4
void FASTCALL call_14_3_5_4(SH2_struct* sh)
{
  SH2movi(sh, 3, (5<<4)+4);
}
#define init_14_3_5_5
void FASTCALL call_14_3_5_5(SH2_struct* sh)
{
  SH2movi(sh, 3, (5<<4)+5);
}
#define init_14_3_5_6
void FASTCALL call_14_3_5_6(SH2_struct* sh)
{
  SH2movi(sh, 3, (5<<4)+6);
}
#define init_14_3_5_7
void FASTCALL call_14_3_5_7(SH2_struct* sh)
{
  SH2movi(sh, 3, (5<<4)+7);
}
#define init_14_3_5_8
void FASTCALL call_14_3_5_8(SH2_struct* sh)
{
  SH2movi(sh, 3, (5<<4)+8);
}
#define init_14_3_5_9
void FASTCALL call_14_3_5_9(SH2_struct* sh)
{
  SH2movi(sh, 3, (5<<4)+9);
}
#define init_14_3_5_10
void FASTCALL call_14_3_5_10(SH2_struct* sh)
{
  SH2movi(sh, 3, (5<<4)+10);
}
#define init_14_3_5_11
void FASTCALL call_14_3_5_11(SH2_struct* sh)
{
  SH2movi(sh, 3, (5<<4)+11);
}
#define init_14_3_5_12
void FASTCALL call_14_3_5_12(SH2_struct* sh)
{
  SH2movi(sh, 3, (5<<4)+12);
}
#define init_14_3_5_13
void FASTCALL call_14_3_5_13(SH2_struct* sh)
{
  SH2movi(sh, 3, (5<<4)+13);
}
#define init_14_3_5_14
void FASTCALL call_14_3_5_14(SH2_struct* sh)
{
  SH2movi(sh, 3, (5<<4)+14);
}
#define init_14_3_5_15
void FASTCALL call_14_3_5_15(SH2_struct* sh)
{
  SH2movi(sh, 3, (5<<4)+15);
}
#define init_14_3_6_0
void FASTCALL call_14_3_6_0(SH2_struct* sh)
{
  SH2movi(sh, 3, (6<<4)+0);
}
#define init_14_3_6_1
void FASTCALL call_14_3_6_1(SH2_struct* sh)
{
  SH2movi(sh, 3, (6<<4)+1);
}
#define init_14_3_6_2
void FASTCALL call_14_3_6_2(SH2_struct* sh)
{
  SH2movi(sh, 3, (6<<4)+2);
}
#define init_14_3_6_3
void FASTCALL call_14_3_6_3(SH2_struct* sh)
{
  SH2movi(sh, 3, (6<<4)+3);
}
#define init_14_3_6_4
void FASTCALL call_14_3_6_4(SH2_struct* sh)
{
  SH2movi(sh, 3, (6<<4)+4);
}
#define init_14_3_6_5
void FASTCALL call_14_3_6_5(SH2_struct* sh)
{
  SH2movi(sh, 3, (6<<4)+5);
}
#define init_14_3_6_6
void FASTCALL call_14_3_6_6(SH2_struct* sh)
{
  SH2movi(sh, 3, (6<<4)+6);
}
#define init_14_3_6_7
void FASTCALL call_14_3_6_7(SH2_struct* sh)
{
  SH2movi(sh, 3, (6<<4)+7);
}
#define init_14_3_6_8
void FASTCALL call_14_3_6_8(SH2_struct* sh)
{
  SH2movi(sh, 3, (6<<4)+8);
}
#define init_14_3_6_9
void FASTCALL call_14_3_6_9(SH2_struct* sh)
{
  SH2movi(sh, 3, (6<<4)+9);
}
#define init_14_3_6_10
void FASTCALL call_14_3_6_10(SH2_struct* sh)
{
  SH2movi(sh, 3, (6<<4)+10);
}
#define init_14_3_6_11
void FASTCALL call_14_3_6_11(SH2_struct* sh)
{
  SH2movi(sh, 3, (6<<4)+11);
}
#define init_14_3_6_12
void FASTCALL call_14_3_6_12(SH2_struct* sh)
{
  SH2movi(sh, 3, (6<<4)+12);
}
#define init_14_3_6_13
void FASTCALL call_14_3_6_13(SH2_struct* sh)
{
  SH2movi(sh, 3, (6<<4)+13);
}
#define init_14_3_6_14
void FASTCALL call_14_3_6_14(SH2_struct* sh)
{
  SH2movi(sh, 3, (6<<4)+14);
}
#define init_14_3_6_15
void FASTCALL call_14_3_6_15(SH2_struct* sh)
{
  SH2movi(sh, 3, (6<<4)+15);
}
#define init_14_3_7_0
void FASTCALL call_14_3_7_0(SH2_struct* sh)
{
  SH2movi(sh, 3, (7<<4)+0);
}
#define init_14_3_7_1
void FASTCALL call_14_3_7_1(SH2_struct* sh)
{
  SH2movi(sh, 3, (7<<4)+1);
}
#define init_14_3_7_2
void FASTCALL call_14_3_7_2(SH2_struct* sh)
{
  SH2movi(sh, 3, (7<<4)+2);
}
#define init_14_3_7_3
void FASTCALL call_14_3_7_3(SH2_struct* sh)
{
  SH2movi(sh, 3, (7<<4)+3);
}
#define init_14_3_7_4
void FASTCALL call_14_3_7_4(SH2_struct* sh)
{
  SH2movi(sh, 3, (7<<4)+4);
}
#define init_14_3_7_5
void FASTCALL call_14_3_7_5(SH2_struct* sh)
{
  SH2movi(sh, 3, (7<<4)+5);
}
#define init_14_3_7_6
void FASTCALL call_14_3_7_6(SH2_struct* sh)
{
  SH2movi(sh, 3, (7<<4)+6);
}
#define init_14_3_7_7
void FASTCALL call_14_3_7_7(SH2_struct* sh)
{
  SH2movi(sh, 3, (7<<4)+7);
}
#define init_14_3_7_8
void FASTCALL call_14_3_7_8(SH2_struct* sh)
{
  SH2movi(sh, 3, (7<<4)+8);
}
#define init_14_3_7_9
void FASTCALL call_14_3_7_9(SH2_struct* sh)
{
  SH2movi(sh, 3, (7<<4)+9);
}
#define init_14_3_7_10
void FASTCALL call_14_3_7_10(SH2_struct* sh)
{
  SH2movi(sh, 3, (7<<4)+10);
}
#define init_14_3_7_11
void FASTCALL call_14_3_7_11(SH2_struct* sh)
{
  SH2movi(sh, 3, (7<<4)+11);
}
#define init_14_3_7_12
void FASTCALL call_14_3_7_12(SH2_struct* sh)
{
  SH2movi(sh, 3, (7<<4)+12);
}
#define init_14_3_7_13
void FASTCALL call_14_3_7_13(SH2_struct* sh)
{
  SH2movi(sh, 3, (7<<4)+13);
}
#define init_14_3_7_14
void FASTCALL call_14_3_7_14(SH2_struct* sh)
{
  SH2movi(sh, 3, (7<<4)+14);
}
#define init_14_3_7_15
void FASTCALL call_14_3_7_15(SH2_struct* sh)
{
  SH2movi(sh, 3, (7<<4)+15);
}
#define init_14_3_8_0
void FASTCALL call_14_3_8_0(SH2_struct* sh)
{
  SH2movi(sh, 3, (8<<4)+0);
}
#define init_14_3_8_1
void FASTCALL call_14_3_8_1(SH2_struct* sh)
{
  SH2movi(sh, 3, (8<<4)+1);
}
#define init_14_3_8_2
void FASTCALL call_14_3_8_2(SH2_struct* sh)
{
  SH2movi(sh, 3, (8<<4)+2);
}
#define init_14_3_8_3
void FASTCALL call_14_3_8_3(SH2_struct* sh)
{
  SH2movi(sh, 3, (8<<4)+3);
}
#define init_14_3_8_4
void FASTCALL call_14_3_8_4(SH2_struct* sh)
{
  SH2movi(sh, 3, (8<<4)+4);
}
#define init_14_3_8_5
void FASTCALL call_14_3_8_5(SH2_struct* sh)
{
  SH2movi(sh, 3, (8<<4)+5);
}
#define init_14_3_8_6
void FASTCALL call_14_3_8_6(SH2_struct* sh)
{
  SH2movi(sh, 3, (8<<4)+6);
}
#define init_14_3_8_7
void FASTCALL call_14_3_8_7(SH2_struct* sh)
{
  SH2movi(sh, 3, (8<<4)+7);
}
#define init_14_3_8_8
void FASTCALL call_14_3_8_8(SH2_struct* sh)
{
  SH2movi(sh, 3, (8<<4)+8);
}
#define init_14_3_8_9
void FASTCALL call_14_3_8_9(SH2_struct* sh)
{
  SH2movi(sh, 3, (8<<4)+9);
}
#define init_14_3_8_10
void FASTCALL call_14_3_8_10(SH2_struct* sh)
{
  SH2movi(sh, 3, (8<<4)+10);
}
#define init_14_3_8_11
void FASTCALL call_14_3_8_11(SH2_struct* sh)
{
  SH2movi(sh, 3, (8<<4)+11);
}
#define init_14_3_8_12
void FASTCALL call_14_3_8_12(SH2_struct* sh)
{
  SH2movi(sh, 3, (8<<4)+12);
}
#define init_14_3_8_13
void FASTCALL call_14_3_8_13(SH2_struct* sh)
{
  SH2movi(sh, 3, (8<<4)+13);
}
#define init_14_3_8_14
void FASTCALL call_14_3_8_14(SH2_struct* sh)
{
  SH2movi(sh, 3, (8<<4)+14);
}
#define init_14_3_8_15
void FASTCALL call_14_3_8_15(SH2_struct* sh)
{
  SH2movi(sh, 3, (8<<4)+15);
}
#define init_14_3_9_0
void FASTCALL call_14_3_9_0(SH2_struct* sh)
{
  SH2movi(sh, 3, (9<<4)+0);
}
#define init_14_3_9_1
void FASTCALL call_14_3_9_1(SH2_struct* sh)
{
  SH2movi(sh, 3, (9<<4)+1);
}
#define init_14_3_9_2
void FASTCALL call_14_3_9_2(SH2_struct* sh)
{
  SH2movi(sh, 3, (9<<4)+2);
}
#define init_14_3_9_3
void FASTCALL call_14_3_9_3(SH2_struct* sh)
{
  SH2movi(sh, 3, (9<<4)+3);
}
#define init_14_3_9_4
void FASTCALL call_14_3_9_4(SH2_struct* sh)
{
  SH2movi(sh, 3, (9<<4)+4);
}
#define init_14_3_9_5
void FASTCALL call_14_3_9_5(SH2_struct* sh)
{
  SH2movi(sh, 3, (9<<4)+5);
}
#define init_14_3_9_6
void FASTCALL call_14_3_9_6(SH2_struct* sh)
{
  SH2movi(sh, 3, (9<<4)+6);
}
#define init_14_3_9_7
void FASTCALL call_14_3_9_7(SH2_struct* sh)
{
  SH2movi(sh, 3, (9<<4)+7);
}
#define init_14_3_9_8
void FASTCALL call_14_3_9_8(SH2_struct* sh)
{
  SH2movi(sh, 3, (9<<4)+8);
}
#define init_14_3_9_9
void FASTCALL call_14_3_9_9(SH2_struct* sh)
{
  SH2movi(sh, 3, (9<<4)+9);
}
#define init_14_3_9_10
void FASTCALL call_14_3_9_10(SH2_struct* sh)
{
  SH2movi(sh, 3, (9<<4)+10);
}
#define init_14_3_9_11
void FASTCALL call_14_3_9_11(SH2_struct* sh)
{
  SH2movi(sh, 3, (9<<4)+11);
}
#define init_14_3_9_12
void FASTCALL call_14_3_9_12(SH2_struct* sh)
{
  SH2movi(sh, 3, (9<<4)+12);
}
#define init_14_3_9_13
void FASTCALL call_14_3_9_13(SH2_struct* sh)
{
  SH2movi(sh, 3, (9<<4)+13);
}
#define init_14_3_9_14
void FASTCALL call_14_3_9_14(SH2_struct* sh)
{
  SH2movi(sh, 3, (9<<4)+14);
}
#define init_14_3_9_15
void FASTCALL call_14_3_9_15(SH2_struct* sh)
{
  SH2movi(sh, 3, (9<<4)+15);
}
#define init_14_3_10_0
void FASTCALL call_14_3_10_0(SH2_struct* sh)
{
  SH2movi(sh, 3, (10<<4)+0);
}
#define init_14_3_10_1
void FASTCALL call_14_3_10_1(SH2_struct* sh)
{
  SH2movi(sh, 3, (10<<4)+1);
}
#define init_14_3_10_2
void FASTCALL call_14_3_10_2(SH2_struct* sh)
{
  SH2movi(sh, 3, (10<<4)+2);
}
#define init_14_3_10_3
void FASTCALL call_14_3_10_3(SH2_struct* sh)
{
  SH2movi(sh, 3, (10<<4)+3);
}
#define init_14_3_10_4
void FASTCALL call_14_3_10_4(SH2_struct* sh)
{
  SH2movi(sh, 3, (10<<4)+4);
}
#define init_14_3_10_5
void FASTCALL call_14_3_10_5(SH2_struct* sh)
{
  SH2movi(sh, 3, (10<<4)+5);
}
#define init_14_3_10_6
void FASTCALL call_14_3_10_6(SH2_struct* sh)
{
  SH2movi(sh, 3, (10<<4)+6);
}
#define init_14_3_10_7
void FASTCALL call_14_3_10_7(SH2_struct* sh)
{
  SH2movi(sh, 3, (10<<4)+7);
}
#define init_14_3_10_8
void FASTCALL call_14_3_10_8(SH2_struct* sh)
{
  SH2movi(sh, 3, (10<<4)+8);
}
#define init_14_3_10_9
void FASTCALL call_14_3_10_9(SH2_struct* sh)
{
  SH2movi(sh, 3, (10<<4)+9);
}
#define init_14_3_10_10
void FASTCALL call_14_3_10_10(SH2_struct* sh)
{
  SH2movi(sh, 3, (10<<4)+10);
}
#define init_14_3_10_11
void FASTCALL call_14_3_10_11(SH2_struct* sh)
{
  SH2movi(sh, 3, (10<<4)+11);
}
#define init_14_3_10_12
void FASTCALL call_14_3_10_12(SH2_struct* sh)
{
  SH2movi(sh, 3, (10<<4)+12);
}
#define init_14_3_10_13
void FASTCALL call_14_3_10_13(SH2_struct* sh)
{
  SH2movi(sh, 3, (10<<4)+13);
}
#define init_14_3_10_14
void FASTCALL call_14_3_10_14(SH2_struct* sh)
{
  SH2movi(sh, 3, (10<<4)+14);
}
#define init_14_3_10_15
void FASTCALL call_14_3_10_15(SH2_struct* sh)
{
  SH2movi(sh, 3, (10<<4)+15);
}
#define init_14_3_11_0
void FASTCALL call_14_3_11_0(SH2_struct* sh)
{
  SH2movi(sh, 3, (11<<4)+0);
}
#define init_14_3_11_1
void FASTCALL call_14_3_11_1(SH2_struct* sh)
{
  SH2movi(sh, 3, (11<<4)+1);
}
#define init_14_3_11_2
void FASTCALL call_14_3_11_2(SH2_struct* sh)
{
  SH2movi(sh, 3, (11<<4)+2);
}
#define init_14_3_11_3
void FASTCALL call_14_3_11_3(SH2_struct* sh)
{
  SH2movi(sh, 3, (11<<4)+3);
}
#define init_14_3_11_4
void FASTCALL call_14_3_11_4(SH2_struct* sh)
{
  SH2movi(sh, 3, (11<<4)+4);
}
#define init_14_3_11_5
void FASTCALL call_14_3_11_5(SH2_struct* sh)
{
  SH2movi(sh, 3, (11<<4)+5);
}
#define init_14_3_11_6
void FASTCALL call_14_3_11_6(SH2_struct* sh)
{
  SH2movi(sh, 3, (11<<4)+6);
}
#define init_14_3_11_7
void FASTCALL call_14_3_11_7(SH2_struct* sh)
{
  SH2movi(sh, 3, (11<<4)+7);
}
#define init_14_3_11_8
void FASTCALL call_14_3_11_8(SH2_struct* sh)
{
  SH2movi(sh, 3, (11<<4)+8);
}
#define init_14_3_11_9
void FASTCALL call_14_3_11_9(SH2_struct* sh)
{
  SH2movi(sh, 3, (11<<4)+9);
}
#define init_14_3_11_10
void FASTCALL call_14_3_11_10(SH2_struct* sh)
{
  SH2movi(sh, 3, (11<<4)+10);
}
#define init_14_3_11_11
void FASTCALL call_14_3_11_11(SH2_struct* sh)
{
  SH2movi(sh, 3, (11<<4)+11);
}
#define init_14_3_11_12
void FASTCALL call_14_3_11_12(SH2_struct* sh)
{
  SH2movi(sh, 3, (11<<4)+12);
}
#define init_14_3_11_13
void FASTCALL call_14_3_11_13(SH2_struct* sh)
{
  SH2movi(sh, 3, (11<<4)+13);
}
#define init_14_3_11_14
void FASTCALL call_14_3_11_14(SH2_struct* sh)
{
  SH2movi(sh, 3, (11<<4)+14);
}
#define init_14_3_11_15
void FASTCALL call_14_3_11_15(SH2_struct* sh)
{
  SH2movi(sh, 3, (11<<4)+15);
}
#define init_14_3_12_0
void FASTCALL call_14_3_12_0(SH2_struct* sh)
{
  SH2movi(sh, 3, (12<<4)+0);
}
#define init_14_3_12_1
void FASTCALL call_14_3_12_1(SH2_struct* sh)
{
  SH2movi(sh, 3, (12<<4)+1);
}
#define init_14_3_12_2
void FASTCALL call_14_3_12_2(SH2_struct* sh)
{
  SH2movi(sh, 3, (12<<4)+2);
}
#define init_14_3_12_3
void FASTCALL call_14_3_12_3(SH2_struct* sh)
{
  SH2movi(sh, 3, (12<<4)+3);
}
#define init_14_3_12_4
void FASTCALL call_14_3_12_4(SH2_struct* sh)
{
  SH2movi(sh, 3, (12<<4)+4);
}
#define init_14_3_12_5
void FASTCALL call_14_3_12_5(SH2_struct* sh)
{
  SH2movi(sh, 3, (12<<4)+5);
}
#define init_14_3_12_6
void FASTCALL call_14_3_12_6(SH2_struct* sh)
{
  SH2movi(sh, 3, (12<<4)+6);
}
#define init_14_3_12_7
void FASTCALL call_14_3_12_7(SH2_struct* sh)
{
  SH2movi(sh, 3, (12<<4)+7);
}
#define init_14_3_12_8
void FASTCALL call_14_3_12_8(SH2_struct* sh)
{
  SH2movi(sh, 3, (12<<4)+8);
}
#define init_14_3_12_9
void FASTCALL call_14_3_12_9(SH2_struct* sh)
{
  SH2movi(sh, 3, (12<<4)+9);
}
#define init_14_3_12_10
void FASTCALL call_14_3_12_10(SH2_struct* sh)
{
  SH2movi(sh, 3, (12<<4)+10);
}
#define init_14_3_12_11
void FASTCALL call_14_3_12_11(SH2_struct* sh)
{
  SH2movi(sh, 3, (12<<4)+11);
}
#define init_14_3_12_12
void FASTCALL call_14_3_12_12(SH2_struct* sh)
{
  SH2movi(sh, 3, (12<<4)+12);
}
#define init_14_3_12_13
void FASTCALL call_14_3_12_13(SH2_struct* sh)
{
  SH2movi(sh, 3, (12<<4)+13);
}
#define init_14_3_12_14
void FASTCALL call_14_3_12_14(SH2_struct* sh)
{
  SH2movi(sh, 3, (12<<4)+14);
}
#define init_14_3_12_15
void FASTCALL call_14_3_12_15(SH2_struct* sh)
{
  SH2movi(sh, 3, (12<<4)+15);
}
#define init_14_3_13_0
void FASTCALL call_14_3_13_0(SH2_struct* sh)
{
  SH2movi(sh, 3, (13<<4)+0);
}
#define init_14_3_13_1
void FASTCALL call_14_3_13_1(SH2_struct* sh)
{
  SH2movi(sh, 3, (13<<4)+1);
}
#define init_14_3_13_2
void FASTCALL call_14_3_13_2(SH2_struct* sh)
{
  SH2movi(sh, 3, (13<<4)+2);
}
#define init_14_3_13_3
void FASTCALL call_14_3_13_3(SH2_struct* sh)
{
  SH2movi(sh, 3, (13<<4)+3);
}
#define init_14_3_13_4
void FASTCALL call_14_3_13_4(SH2_struct* sh)
{
  SH2movi(sh, 3, (13<<4)+4);
}
#define init_14_3_13_5
void FASTCALL call_14_3_13_5(SH2_struct* sh)
{
  SH2movi(sh, 3, (13<<4)+5);
}
#define init_14_3_13_6
void FASTCALL call_14_3_13_6(SH2_struct* sh)
{
  SH2movi(sh, 3, (13<<4)+6);
}
#define init_14_3_13_7
void FASTCALL call_14_3_13_7(SH2_struct* sh)
{
  SH2movi(sh, 3, (13<<4)+7);
}
#define init_14_3_13_8
void FASTCALL call_14_3_13_8(SH2_struct* sh)
{
  SH2movi(sh, 3, (13<<4)+8);
}
#define init_14_3_13_9
void FASTCALL call_14_3_13_9(SH2_struct* sh)
{
  SH2movi(sh, 3, (13<<4)+9);
}
#define init_14_3_13_10
void FASTCALL call_14_3_13_10(SH2_struct* sh)
{
  SH2movi(sh, 3, (13<<4)+10);
}
#define init_14_3_13_11
void FASTCALL call_14_3_13_11(SH2_struct* sh)
{
  SH2movi(sh, 3, (13<<4)+11);
}
#define init_14_3_13_12
void FASTCALL call_14_3_13_12(SH2_struct* sh)
{
  SH2movi(sh, 3, (13<<4)+12);
}
#define init_14_3_13_13
void FASTCALL call_14_3_13_13(SH2_struct* sh)
{
  SH2movi(sh, 3, (13<<4)+13);
}
#define init_14_3_13_14
void FASTCALL call_14_3_13_14(SH2_struct* sh)
{
  SH2movi(sh, 3, (13<<4)+14);
}
#define init_14_3_13_15
void FASTCALL call_14_3_13_15(SH2_struct* sh)
{
  SH2movi(sh, 3, (13<<4)+15);
}
#define init_14_3_14_0
void FASTCALL call_14_3_14_0(SH2_struct* sh)
{
  SH2movi(sh, 3, (14<<4)+0);
}
#define init_14_3_14_1
void FASTCALL call_14_3_14_1(SH2_struct* sh)
{
  SH2movi(sh, 3, (14<<4)+1);
}
#define init_14_3_14_2
void FASTCALL call_14_3_14_2(SH2_struct* sh)
{
  SH2movi(sh, 3, (14<<4)+2);
}
#define init_14_3_14_3
void FASTCALL call_14_3_14_3(SH2_struct* sh)
{
  SH2movi(sh, 3, (14<<4)+3);
}
#define init_14_3_14_4
void FASTCALL call_14_3_14_4(SH2_struct* sh)
{
  SH2movi(sh, 3, (14<<4)+4);
}
#define init_14_3_14_5
void FASTCALL call_14_3_14_5(SH2_struct* sh)
{
  SH2movi(sh, 3, (14<<4)+5);
}
#define init_14_3_14_6
void FASTCALL call_14_3_14_6(SH2_struct* sh)
{
  SH2movi(sh, 3, (14<<4)+6);
}
#define init_14_3_14_7
void FASTCALL call_14_3_14_7(SH2_struct* sh)
{
  SH2movi(sh, 3, (14<<4)+7);
}
#define init_14_3_14_8
void FASTCALL call_14_3_14_8(SH2_struct* sh)
{
  SH2movi(sh, 3, (14<<4)+8);
}
#define init_14_3_14_9
void FASTCALL call_14_3_14_9(SH2_struct* sh)
{
  SH2movi(sh, 3, (14<<4)+9);
}
#define init_14_3_14_10
void FASTCALL call_14_3_14_10(SH2_struct* sh)
{
  SH2movi(sh, 3, (14<<4)+10);
}
#define init_14_3_14_11
void FASTCALL call_14_3_14_11(SH2_struct* sh)
{
  SH2movi(sh, 3, (14<<4)+11);
}
#define init_14_3_14_12
void FASTCALL call_14_3_14_12(SH2_struct* sh)
{
  SH2movi(sh, 3, (14<<4)+12);
}
#define init_14_3_14_13
void FASTCALL call_14_3_14_13(SH2_struct* sh)
{
  SH2movi(sh, 3, (14<<4)+13);
}
#define init_14_3_14_14
void FASTCALL call_14_3_14_14(SH2_struct* sh)
{
  SH2movi(sh, 3, (14<<4)+14);
}
#define init_14_3_14_15
void FASTCALL call_14_3_14_15(SH2_struct* sh)
{
  SH2movi(sh, 3, (14<<4)+15);
}
#define init_14_3_15_0
void FASTCALL call_14_3_15_0(SH2_struct* sh)
{
  SH2movi(sh, 3, (15<<4)+0);
}
#define init_14_3_15_1
void FASTCALL call_14_3_15_1(SH2_struct* sh)
{
  SH2movi(sh, 3, (15<<4)+1);
}
#define init_14_3_15_2
void FASTCALL call_14_3_15_2(SH2_struct* sh)
{
  SH2movi(sh, 3, (15<<4)+2);
}
#define init_14_3_15_3
void FASTCALL call_14_3_15_3(SH2_struct* sh)
{
  SH2movi(sh, 3, (15<<4)+3);
}
#define init_14_3_15_4
void FASTCALL call_14_3_15_4(SH2_struct* sh)
{
  SH2movi(sh, 3, (15<<4)+4);
}
#define init_14_3_15_5
void FASTCALL call_14_3_15_5(SH2_struct* sh)
{
  SH2movi(sh, 3, (15<<4)+5);
}
#define init_14_3_15_6
void FASTCALL call_14_3_15_6(SH2_struct* sh)
{
  SH2movi(sh, 3, (15<<4)+6);
}
#define init_14_3_15_7
void FASTCALL call_14_3_15_7(SH2_struct* sh)
{
  SH2movi(sh, 3, (15<<4)+7);
}
#define init_14_3_15_8
void FASTCALL call_14_3_15_8(SH2_struct* sh)
{
  SH2movi(sh, 3, (15<<4)+8);
}
#define init_14_3_15_9
void FASTCALL call_14_3_15_9(SH2_struct* sh)
{
  SH2movi(sh, 3, (15<<4)+9);
}
#define init_14_3_15_10
void FASTCALL call_14_3_15_10(SH2_struct* sh)
{
  SH2movi(sh, 3, (15<<4)+10);
}
#define init_14_3_15_11
void FASTCALL call_14_3_15_11(SH2_struct* sh)
{
  SH2movi(sh, 3, (15<<4)+11);
}
#define init_14_3_15_12
void FASTCALL call_14_3_15_12(SH2_struct* sh)
{
  SH2movi(sh, 3, (15<<4)+12);
}
#define init_14_3_15_13
void FASTCALL call_14_3_15_13(SH2_struct* sh)
{
  SH2movi(sh, 3, (15<<4)+13);
}
#define init_14_3_15_14
void FASTCALL call_14_3_15_14(SH2_struct* sh)
{
  SH2movi(sh, 3, (15<<4)+14);
}
#define init_14_3_15_15
void FASTCALL call_14_3_15_15(SH2_struct* sh)
{
  SH2movi(sh, 3, (15<<4)+15);
}
#define init_14_4_0_0
void FASTCALL call_14_4_0_0(SH2_struct* sh)
{
  SH2movi(sh, 4, (0<<4)+0);
}
#define init_14_4_0_1
void FASTCALL call_14_4_0_1(SH2_struct* sh)
{
  SH2movi(sh, 4, (0<<4)+1);
}
#define init_14_4_0_2
void FASTCALL call_14_4_0_2(SH2_struct* sh)
{
  SH2movi(sh, 4, (0<<4)+2);
}
#define init_14_4_0_3
void FASTCALL call_14_4_0_3(SH2_struct* sh)
{
  SH2movi(sh, 4, (0<<4)+3);
}
#define init_14_4_0_4
void FASTCALL call_14_4_0_4(SH2_struct* sh)
{
  SH2movi(sh, 4, (0<<4)+4);
}
#define init_14_4_0_5
void FASTCALL call_14_4_0_5(SH2_struct* sh)
{
  SH2movi(sh, 4, (0<<4)+5);
}
#define init_14_4_0_6
void FASTCALL call_14_4_0_6(SH2_struct* sh)
{
  SH2movi(sh, 4, (0<<4)+6);
}
#define init_14_4_0_7
void FASTCALL call_14_4_0_7(SH2_struct* sh)
{
  SH2movi(sh, 4, (0<<4)+7);
}
#define init_14_4_0_8
void FASTCALL call_14_4_0_8(SH2_struct* sh)
{
  SH2movi(sh, 4, (0<<4)+8);
}
#define init_14_4_0_9
void FASTCALL call_14_4_0_9(SH2_struct* sh)
{
  SH2movi(sh, 4, (0<<4)+9);
}
#define init_14_4_0_10
void FASTCALL call_14_4_0_10(SH2_struct* sh)
{
  SH2movi(sh, 4, (0<<4)+10);
}
#define init_14_4_0_11
void FASTCALL call_14_4_0_11(SH2_struct* sh)
{
  SH2movi(sh, 4, (0<<4)+11);
}
#define init_14_4_0_12
void FASTCALL call_14_4_0_12(SH2_struct* sh)
{
  SH2movi(sh, 4, (0<<4)+12);
}
#define init_14_4_0_13
void FASTCALL call_14_4_0_13(SH2_struct* sh)
{
  SH2movi(sh, 4, (0<<4)+13);
}
#define init_14_4_0_14
void FASTCALL call_14_4_0_14(SH2_struct* sh)
{
  SH2movi(sh, 4, (0<<4)+14);
}
#define init_14_4_0_15
void FASTCALL call_14_4_0_15(SH2_struct* sh)
{
  SH2movi(sh, 4, (0<<4)+15);
}
#define init_14_4_1_0
void FASTCALL call_14_4_1_0(SH2_struct* sh)
{
  SH2movi(sh, 4, (1<<4)+0);
}
#define init_14_4_1_1
void FASTCALL call_14_4_1_1(SH2_struct* sh)
{
  SH2movi(sh, 4, (1<<4)+1);
}
#define init_14_4_1_2
void FASTCALL call_14_4_1_2(SH2_struct* sh)
{
  SH2movi(sh, 4, (1<<4)+2);
}
#define init_14_4_1_3
void FASTCALL call_14_4_1_3(SH2_struct* sh)
{
  SH2movi(sh, 4, (1<<4)+3);
}
#define init_14_4_1_4
void FASTCALL call_14_4_1_4(SH2_struct* sh)
{
  SH2movi(sh, 4, (1<<4)+4);
}
#define init_14_4_1_5
void FASTCALL call_14_4_1_5(SH2_struct* sh)
{
  SH2movi(sh, 4, (1<<4)+5);
}
#define init_14_4_1_6
void FASTCALL call_14_4_1_6(SH2_struct* sh)
{
  SH2movi(sh, 4, (1<<4)+6);
}
#define init_14_4_1_7
void FASTCALL call_14_4_1_7(SH2_struct* sh)
{
  SH2movi(sh, 4, (1<<4)+7);
}
#define init_14_4_1_8
void FASTCALL call_14_4_1_8(SH2_struct* sh)
{
  SH2movi(sh, 4, (1<<4)+8);
}
#define init_14_4_1_9
void FASTCALL call_14_4_1_9(SH2_struct* sh)
{
  SH2movi(sh, 4, (1<<4)+9);
}
#define init_14_4_1_10
void FASTCALL call_14_4_1_10(SH2_struct* sh)
{
  SH2movi(sh, 4, (1<<4)+10);
}
#define init_14_4_1_11
void FASTCALL call_14_4_1_11(SH2_struct* sh)
{
  SH2movi(sh, 4, (1<<4)+11);
}
#define init_14_4_1_12
void FASTCALL call_14_4_1_12(SH2_struct* sh)
{
  SH2movi(sh, 4, (1<<4)+12);
}
#define init_14_4_1_13
void FASTCALL call_14_4_1_13(SH2_struct* sh)
{
  SH2movi(sh, 4, (1<<4)+13);
}
#define init_14_4_1_14
void FASTCALL call_14_4_1_14(SH2_struct* sh)
{
  SH2movi(sh, 4, (1<<4)+14);
}
#define init_14_4_1_15
void FASTCALL call_14_4_1_15(SH2_struct* sh)
{
  SH2movi(sh, 4, (1<<4)+15);
}
#define init_14_4_2_0
void FASTCALL call_14_4_2_0(SH2_struct* sh)
{
  SH2movi(sh, 4, (2<<4)+0);
}
#define init_14_4_2_1
void FASTCALL call_14_4_2_1(SH2_struct* sh)
{
  SH2movi(sh, 4, (2<<4)+1);
}
#define init_14_4_2_2
void FASTCALL call_14_4_2_2(SH2_struct* sh)
{
  SH2movi(sh, 4, (2<<4)+2);
}
#define init_14_4_2_3
void FASTCALL call_14_4_2_3(SH2_struct* sh)
{
  SH2movi(sh, 4, (2<<4)+3);
}
#define init_14_4_2_4
void FASTCALL call_14_4_2_4(SH2_struct* sh)
{
  SH2movi(sh, 4, (2<<4)+4);
}
#define init_14_4_2_5
void FASTCALL call_14_4_2_5(SH2_struct* sh)
{
  SH2movi(sh, 4, (2<<4)+5);
}
#define init_14_4_2_6
void FASTCALL call_14_4_2_6(SH2_struct* sh)
{
  SH2movi(sh, 4, (2<<4)+6);
}
#define init_14_4_2_7
void FASTCALL call_14_4_2_7(SH2_struct* sh)
{
  SH2movi(sh, 4, (2<<4)+7);
}
#define init_14_4_2_8
void FASTCALL call_14_4_2_8(SH2_struct* sh)
{
  SH2movi(sh, 4, (2<<4)+8);
}
#define init_14_4_2_9
void FASTCALL call_14_4_2_9(SH2_struct* sh)
{
  SH2movi(sh, 4, (2<<4)+9);
}
#define init_14_4_2_10
void FASTCALL call_14_4_2_10(SH2_struct* sh)
{
  SH2movi(sh, 4, (2<<4)+10);
}
#define init_14_4_2_11
void FASTCALL call_14_4_2_11(SH2_struct* sh)
{
  SH2movi(sh, 4, (2<<4)+11);
}
#define init_14_4_2_12
void FASTCALL call_14_4_2_12(SH2_struct* sh)
{
  SH2movi(sh, 4, (2<<4)+12);
}
#define init_14_4_2_13
void FASTCALL call_14_4_2_13(SH2_struct* sh)
{
  SH2movi(sh, 4, (2<<4)+13);
}
#define init_14_4_2_14
void FASTCALL call_14_4_2_14(SH2_struct* sh)
{
  SH2movi(sh, 4, (2<<4)+14);
}
#define init_14_4_2_15
void FASTCALL call_14_4_2_15(SH2_struct* sh)
{
  SH2movi(sh, 4, (2<<4)+15);
}
#define init_14_4_3_0
void FASTCALL call_14_4_3_0(SH2_struct* sh)
{
  SH2movi(sh, 4, (3<<4)+0);
}
#define init_14_4_3_1
void FASTCALL call_14_4_3_1(SH2_struct* sh)
{
  SH2movi(sh, 4, (3<<4)+1);
}
#define init_14_4_3_2
void FASTCALL call_14_4_3_2(SH2_struct* sh)
{
  SH2movi(sh, 4, (3<<4)+2);
}
#define init_14_4_3_3
void FASTCALL call_14_4_3_3(SH2_struct* sh)
{
  SH2movi(sh, 4, (3<<4)+3);
}
#define init_14_4_3_4
void FASTCALL call_14_4_3_4(SH2_struct* sh)
{
  SH2movi(sh, 4, (3<<4)+4);
}
#define init_14_4_3_5
void FASTCALL call_14_4_3_5(SH2_struct* sh)
{
  SH2movi(sh, 4, (3<<4)+5);
}
#define init_14_4_3_6
void FASTCALL call_14_4_3_6(SH2_struct* sh)
{
  SH2movi(sh, 4, (3<<4)+6);
}
#define init_14_4_3_7
void FASTCALL call_14_4_3_7(SH2_struct* sh)
{
  SH2movi(sh, 4, (3<<4)+7);
}
#define init_14_4_3_8
void FASTCALL call_14_4_3_8(SH2_struct* sh)
{
  SH2movi(sh, 4, (3<<4)+8);
}
#define init_14_4_3_9
void FASTCALL call_14_4_3_9(SH2_struct* sh)
{
  SH2movi(sh, 4, (3<<4)+9);
}
#define init_14_4_3_10
void FASTCALL call_14_4_3_10(SH2_struct* sh)
{
  SH2movi(sh, 4, (3<<4)+10);
}
#define init_14_4_3_11
void FASTCALL call_14_4_3_11(SH2_struct* sh)
{
  SH2movi(sh, 4, (3<<4)+11);
}
#define init_14_4_3_12
void FASTCALL call_14_4_3_12(SH2_struct* sh)
{
  SH2movi(sh, 4, (3<<4)+12);
}
#define init_14_4_3_13
void FASTCALL call_14_4_3_13(SH2_struct* sh)
{
  SH2movi(sh, 4, (3<<4)+13);
}
#define init_14_4_3_14
void FASTCALL call_14_4_3_14(SH2_struct* sh)
{
  SH2movi(sh, 4, (3<<4)+14);
}
#define init_14_4_3_15
void FASTCALL call_14_4_3_15(SH2_struct* sh)
{
  SH2movi(sh, 4, (3<<4)+15);
}
#define init_14_4_4_0
void FASTCALL call_14_4_4_0(SH2_struct* sh)
{
  SH2movi(sh, 4, (4<<4)+0);
}
#define init_14_4_4_1
void FASTCALL call_14_4_4_1(SH2_struct* sh)
{
  SH2movi(sh, 4, (4<<4)+1);
}
#define init_14_4_4_2
void FASTCALL call_14_4_4_2(SH2_struct* sh)
{
  SH2movi(sh, 4, (4<<4)+2);
}
#define init_14_4_4_3
void FASTCALL call_14_4_4_3(SH2_struct* sh)
{
  SH2movi(sh, 4, (4<<4)+3);
}
#define init_14_4_4_4
void FASTCALL call_14_4_4_4(SH2_struct* sh)
{
  SH2movi(sh, 4, (4<<4)+4);
}
#define init_14_4_4_5
void FASTCALL call_14_4_4_5(SH2_struct* sh)
{
  SH2movi(sh, 4, (4<<4)+5);
}
#define init_14_4_4_6
void FASTCALL call_14_4_4_6(SH2_struct* sh)
{
  SH2movi(sh, 4, (4<<4)+6);
}
#define init_14_4_4_7
void FASTCALL call_14_4_4_7(SH2_struct* sh)
{
  SH2movi(sh, 4, (4<<4)+7);
}
#define init_14_4_4_8
void FASTCALL call_14_4_4_8(SH2_struct* sh)
{
  SH2movi(sh, 4, (4<<4)+8);
}
#define init_14_4_4_9
void FASTCALL call_14_4_4_9(SH2_struct* sh)
{
  SH2movi(sh, 4, (4<<4)+9);
}
#define init_14_4_4_10
void FASTCALL call_14_4_4_10(SH2_struct* sh)
{
  SH2movi(sh, 4, (4<<4)+10);
}
#define init_14_4_4_11
void FASTCALL call_14_4_4_11(SH2_struct* sh)
{
  SH2movi(sh, 4, (4<<4)+11);
}
#define init_14_4_4_12
void FASTCALL call_14_4_4_12(SH2_struct* sh)
{
  SH2movi(sh, 4, (4<<4)+12);
}
#define init_14_4_4_13
void FASTCALL call_14_4_4_13(SH2_struct* sh)
{
  SH2movi(sh, 4, (4<<4)+13);
}
#define init_14_4_4_14
void FASTCALL call_14_4_4_14(SH2_struct* sh)
{
  SH2movi(sh, 4, (4<<4)+14);
}
#define init_14_4_4_15
void FASTCALL call_14_4_4_15(SH2_struct* sh)
{
  SH2movi(sh, 4, (4<<4)+15);
}
#define init_14_4_5_0
void FASTCALL call_14_4_5_0(SH2_struct* sh)
{
  SH2movi(sh, 4, (5<<4)+0);
}
#define init_14_4_5_1
void FASTCALL call_14_4_5_1(SH2_struct* sh)
{
  SH2movi(sh, 4, (5<<4)+1);
}
#define init_14_4_5_2
void FASTCALL call_14_4_5_2(SH2_struct* sh)
{
  SH2movi(sh, 4, (5<<4)+2);
}
#define init_14_4_5_3
void FASTCALL call_14_4_5_3(SH2_struct* sh)
{
  SH2movi(sh, 4, (5<<4)+3);
}
#define init_14_4_5_4
void FASTCALL call_14_4_5_4(SH2_struct* sh)
{
  SH2movi(sh, 4, (5<<4)+4);
}
#define init_14_4_5_5
void FASTCALL call_14_4_5_5(SH2_struct* sh)
{
  SH2movi(sh, 4, (5<<4)+5);
}
#define init_14_4_5_6
void FASTCALL call_14_4_5_6(SH2_struct* sh)
{
  SH2movi(sh, 4, (5<<4)+6);
}
#define init_14_4_5_7
void FASTCALL call_14_4_5_7(SH2_struct* sh)
{
  SH2movi(sh, 4, (5<<4)+7);
}
#define init_14_4_5_8
void FASTCALL call_14_4_5_8(SH2_struct* sh)
{
  SH2movi(sh, 4, (5<<4)+8);
}
#define init_14_4_5_9
void FASTCALL call_14_4_5_9(SH2_struct* sh)
{
  SH2movi(sh, 4, (5<<4)+9);
}
#define init_14_4_5_10
void FASTCALL call_14_4_5_10(SH2_struct* sh)
{
  SH2movi(sh, 4, (5<<4)+10);
}
#define init_14_4_5_11
void FASTCALL call_14_4_5_11(SH2_struct* sh)
{
  SH2movi(sh, 4, (5<<4)+11);
}
#define init_14_4_5_12
void FASTCALL call_14_4_5_12(SH2_struct* sh)
{
  SH2movi(sh, 4, (5<<4)+12);
}
#define init_14_4_5_13
void FASTCALL call_14_4_5_13(SH2_struct* sh)
{
  SH2movi(sh, 4, (5<<4)+13);
}
#define init_14_4_5_14
void FASTCALL call_14_4_5_14(SH2_struct* sh)
{
  SH2movi(sh, 4, (5<<4)+14);
}
#define init_14_4_5_15
void FASTCALL call_14_4_5_15(SH2_struct* sh)
{
  SH2movi(sh, 4, (5<<4)+15);
}
#define init_14_4_6_0
void FASTCALL call_14_4_6_0(SH2_struct* sh)
{
  SH2movi(sh, 4, (6<<4)+0);
}
#define init_14_4_6_1
void FASTCALL call_14_4_6_1(SH2_struct* sh)
{
  SH2movi(sh, 4, (6<<4)+1);
}
#define init_14_4_6_2
void FASTCALL call_14_4_6_2(SH2_struct* sh)
{
  SH2movi(sh, 4, (6<<4)+2);
}
#define init_14_4_6_3
void FASTCALL call_14_4_6_3(SH2_struct* sh)
{
  SH2movi(sh, 4, (6<<4)+3);
}
#define init_14_4_6_4
void FASTCALL call_14_4_6_4(SH2_struct* sh)
{
  SH2movi(sh, 4, (6<<4)+4);
}
#define init_14_4_6_5
void FASTCALL call_14_4_6_5(SH2_struct* sh)
{
  SH2movi(sh, 4, (6<<4)+5);
}
#define init_14_4_6_6
void FASTCALL call_14_4_6_6(SH2_struct* sh)
{
  SH2movi(sh, 4, (6<<4)+6);
}
#define init_14_4_6_7
void FASTCALL call_14_4_6_7(SH2_struct* sh)
{
  SH2movi(sh, 4, (6<<4)+7);
}
#define init_14_4_6_8
void FASTCALL call_14_4_6_8(SH2_struct* sh)
{
  SH2movi(sh, 4, (6<<4)+8);
}
#define init_14_4_6_9
void FASTCALL call_14_4_6_9(SH2_struct* sh)
{
  SH2movi(sh, 4, (6<<4)+9);
}
#define init_14_4_6_10
void FASTCALL call_14_4_6_10(SH2_struct* sh)
{
  SH2movi(sh, 4, (6<<4)+10);
}
#define init_14_4_6_11
void FASTCALL call_14_4_6_11(SH2_struct* sh)
{
  SH2movi(sh, 4, (6<<4)+11);
}
#define init_14_4_6_12
void FASTCALL call_14_4_6_12(SH2_struct* sh)
{
  SH2movi(sh, 4, (6<<4)+12);
}
#define init_14_4_6_13
void FASTCALL call_14_4_6_13(SH2_struct* sh)
{
  SH2movi(sh, 4, (6<<4)+13);
}
#define init_14_4_6_14
void FASTCALL call_14_4_6_14(SH2_struct* sh)
{
  SH2movi(sh, 4, (6<<4)+14);
}
#define init_14_4_6_15
void FASTCALL call_14_4_6_15(SH2_struct* sh)
{
  SH2movi(sh, 4, (6<<4)+15);
}
#define init_14_4_7_0
void FASTCALL call_14_4_7_0(SH2_struct* sh)
{
  SH2movi(sh, 4, (7<<4)+0);
}
#define init_14_4_7_1
void FASTCALL call_14_4_7_1(SH2_struct* sh)
{
  SH2movi(sh, 4, (7<<4)+1);
}
#define init_14_4_7_2
void FASTCALL call_14_4_7_2(SH2_struct* sh)
{
  SH2movi(sh, 4, (7<<4)+2);
}
#define init_14_4_7_3
void FASTCALL call_14_4_7_3(SH2_struct* sh)
{
  SH2movi(sh, 4, (7<<4)+3);
}
#define init_14_4_7_4
void FASTCALL call_14_4_7_4(SH2_struct* sh)
{
  SH2movi(sh, 4, (7<<4)+4);
}
#define init_14_4_7_5
void FASTCALL call_14_4_7_5(SH2_struct* sh)
{
  SH2movi(sh, 4, (7<<4)+5);
}
#define init_14_4_7_6
void FASTCALL call_14_4_7_6(SH2_struct* sh)
{
  SH2movi(sh, 4, (7<<4)+6);
}
#define init_14_4_7_7
void FASTCALL call_14_4_7_7(SH2_struct* sh)
{
  SH2movi(sh, 4, (7<<4)+7);
}
#define init_14_4_7_8
void FASTCALL call_14_4_7_8(SH2_struct* sh)
{
  SH2movi(sh, 4, (7<<4)+8);
}
#define init_14_4_7_9
void FASTCALL call_14_4_7_9(SH2_struct* sh)
{
  SH2movi(sh, 4, (7<<4)+9);
}
#define init_14_4_7_10
void FASTCALL call_14_4_7_10(SH2_struct* sh)
{
  SH2movi(sh, 4, (7<<4)+10);
}
#define init_14_4_7_11
void FASTCALL call_14_4_7_11(SH2_struct* sh)
{
  SH2movi(sh, 4, (7<<4)+11);
}
#define init_14_4_7_12
void FASTCALL call_14_4_7_12(SH2_struct* sh)
{
  SH2movi(sh, 4, (7<<4)+12);
}
#define init_14_4_7_13
void FASTCALL call_14_4_7_13(SH2_struct* sh)
{
  SH2movi(sh, 4, (7<<4)+13);
}
#define init_14_4_7_14
void FASTCALL call_14_4_7_14(SH2_struct* sh)
{
  SH2movi(sh, 4, (7<<4)+14);
}
#define init_14_4_7_15
void FASTCALL call_14_4_7_15(SH2_struct* sh)
{
  SH2movi(sh, 4, (7<<4)+15);
}
#define init_14_4_8_0
void FASTCALL call_14_4_8_0(SH2_struct* sh)
{
  SH2movi(sh, 4, (8<<4)+0);
}
#define init_14_4_8_1
void FASTCALL call_14_4_8_1(SH2_struct* sh)
{
  SH2movi(sh, 4, (8<<4)+1);
}
#define init_14_4_8_2
void FASTCALL call_14_4_8_2(SH2_struct* sh)
{
  SH2movi(sh, 4, (8<<4)+2);
}
#define init_14_4_8_3
void FASTCALL call_14_4_8_3(SH2_struct* sh)
{
  SH2movi(sh, 4, (8<<4)+3);
}
#define init_14_4_8_4
void FASTCALL call_14_4_8_4(SH2_struct* sh)
{
  SH2movi(sh, 4, (8<<4)+4);
}
#define init_14_4_8_5
void FASTCALL call_14_4_8_5(SH2_struct* sh)
{
  SH2movi(sh, 4, (8<<4)+5);
}
#define init_14_4_8_6
void FASTCALL call_14_4_8_6(SH2_struct* sh)
{
  SH2movi(sh, 4, (8<<4)+6);
}
#define init_14_4_8_7
void FASTCALL call_14_4_8_7(SH2_struct* sh)
{
  SH2movi(sh, 4, (8<<4)+7);
}
#define init_14_4_8_8
void FASTCALL call_14_4_8_8(SH2_struct* sh)
{
  SH2movi(sh, 4, (8<<4)+8);
}
#define init_14_4_8_9
void FASTCALL call_14_4_8_9(SH2_struct* sh)
{
  SH2movi(sh, 4, (8<<4)+9);
}
#define init_14_4_8_10
void FASTCALL call_14_4_8_10(SH2_struct* sh)
{
  SH2movi(sh, 4, (8<<4)+10);
}
#define init_14_4_8_11
void FASTCALL call_14_4_8_11(SH2_struct* sh)
{
  SH2movi(sh, 4, (8<<4)+11);
}
#define init_14_4_8_12
void FASTCALL call_14_4_8_12(SH2_struct* sh)
{
  SH2movi(sh, 4, (8<<4)+12);
}
#define init_14_4_8_13
void FASTCALL call_14_4_8_13(SH2_struct* sh)
{
  SH2movi(sh, 4, (8<<4)+13);
}
#define init_14_4_8_14
void FASTCALL call_14_4_8_14(SH2_struct* sh)
{
  SH2movi(sh, 4, (8<<4)+14);
}
#define init_14_4_8_15
void FASTCALL call_14_4_8_15(SH2_struct* sh)
{
  SH2movi(sh, 4, (8<<4)+15);
}
#define init_14_4_9_0
void FASTCALL call_14_4_9_0(SH2_struct* sh)
{
  SH2movi(sh, 4, (9<<4)+0);
}
#define init_14_4_9_1
void FASTCALL call_14_4_9_1(SH2_struct* sh)
{
  SH2movi(sh, 4, (9<<4)+1);
}
#define init_14_4_9_2
void FASTCALL call_14_4_9_2(SH2_struct* sh)
{
  SH2movi(sh, 4, (9<<4)+2);
}
#define init_14_4_9_3
void FASTCALL call_14_4_9_3(SH2_struct* sh)
{
  SH2movi(sh, 4, (9<<4)+3);
}
#define init_14_4_9_4
void FASTCALL call_14_4_9_4(SH2_struct* sh)
{
  SH2movi(sh, 4, (9<<4)+4);
}
#define init_14_4_9_5
void FASTCALL call_14_4_9_5(SH2_struct* sh)
{
  SH2movi(sh, 4, (9<<4)+5);
}
#define init_14_4_9_6
void FASTCALL call_14_4_9_6(SH2_struct* sh)
{
  SH2movi(sh, 4, (9<<4)+6);
}
#define init_14_4_9_7
void FASTCALL call_14_4_9_7(SH2_struct* sh)
{
  SH2movi(sh, 4, (9<<4)+7);
}
#define init_14_4_9_8
void FASTCALL call_14_4_9_8(SH2_struct* sh)
{
  SH2movi(sh, 4, (9<<4)+8);
}
#define init_14_4_9_9
void FASTCALL call_14_4_9_9(SH2_struct* sh)
{
  SH2movi(sh, 4, (9<<4)+9);
}
#define init_14_4_9_10
void FASTCALL call_14_4_9_10(SH2_struct* sh)
{
  SH2movi(sh, 4, (9<<4)+10);
}
#define init_14_4_9_11
void FASTCALL call_14_4_9_11(SH2_struct* sh)
{
  SH2movi(sh, 4, (9<<4)+11);
}
#define init_14_4_9_12
void FASTCALL call_14_4_9_12(SH2_struct* sh)
{
  SH2movi(sh, 4, (9<<4)+12);
}
#define init_14_4_9_13
void FASTCALL call_14_4_9_13(SH2_struct* sh)
{
  SH2movi(sh, 4, (9<<4)+13);
}
#define init_14_4_9_14
void FASTCALL call_14_4_9_14(SH2_struct* sh)
{
  SH2movi(sh, 4, (9<<4)+14);
}
#define init_14_4_9_15
void FASTCALL call_14_4_9_15(SH2_struct* sh)
{
  SH2movi(sh, 4, (9<<4)+15);
}
#define init_14_4_10_0
void FASTCALL call_14_4_10_0(SH2_struct* sh)
{
  SH2movi(sh, 4, (10<<4)+0);
}
#define init_14_4_10_1
void FASTCALL call_14_4_10_1(SH2_struct* sh)
{
  SH2movi(sh, 4, (10<<4)+1);
}
#define init_14_4_10_2
void FASTCALL call_14_4_10_2(SH2_struct* sh)
{
  SH2movi(sh, 4, (10<<4)+2);
}
#define init_14_4_10_3
void FASTCALL call_14_4_10_3(SH2_struct* sh)
{
  SH2movi(sh, 4, (10<<4)+3);
}
#define init_14_4_10_4
void FASTCALL call_14_4_10_4(SH2_struct* sh)
{
  SH2movi(sh, 4, (10<<4)+4);
}
#define init_14_4_10_5
void FASTCALL call_14_4_10_5(SH2_struct* sh)
{
  SH2movi(sh, 4, (10<<4)+5);
}
#define init_14_4_10_6
void FASTCALL call_14_4_10_6(SH2_struct* sh)
{
  SH2movi(sh, 4, (10<<4)+6);
}
#define init_14_4_10_7
void FASTCALL call_14_4_10_7(SH2_struct* sh)
{
  SH2movi(sh, 4, (10<<4)+7);
}
#define init_14_4_10_8
void FASTCALL call_14_4_10_8(SH2_struct* sh)
{
  SH2movi(sh, 4, (10<<4)+8);
}
#define init_14_4_10_9
void FASTCALL call_14_4_10_9(SH2_struct* sh)
{
  SH2movi(sh, 4, (10<<4)+9);
}
#define init_14_4_10_10
void FASTCALL call_14_4_10_10(SH2_struct* sh)
{
  SH2movi(sh, 4, (10<<4)+10);
}
#define init_14_4_10_11
void FASTCALL call_14_4_10_11(SH2_struct* sh)
{
  SH2movi(sh, 4, (10<<4)+11);
}
#define init_14_4_10_12
void FASTCALL call_14_4_10_12(SH2_struct* sh)
{
  SH2movi(sh, 4, (10<<4)+12);
}
#define init_14_4_10_13
void FASTCALL call_14_4_10_13(SH2_struct* sh)
{
  SH2movi(sh, 4, (10<<4)+13);
}
#define init_14_4_10_14
void FASTCALL call_14_4_10_14(SH2_struct* sh)
{
  SH2movi(sh, 4, (10<<4)+14);
}
#define init_14_4_10_15
void FASTCALL call_14_4_10_15(SH2_struct* sh)
{
  SH2movi(sh, 4, (10<<4)+15);
}
#define init_14_4_11_0
void FASTCALL call_14_4_11_0(SH2_struct* sh)
{
  SH2movi(sh, 4, (11<<4)+0);
}
#define init_14_4_11_1
void FASTCALL call_14_4_11_1(SH2_struct* sh)
{
  SH2movi(sh, 4, (11<<4)+1);
}
#define init_14_4_11_2
void FASTCALL call_14_4_11_2(SH2_struct* sh)
{
  SH2movi(sh, 4, (11<<4)+2);
}
#define init_14_4_11_3
void FASTCALL call_14_4_11_3(SH2_struct* sh)
{
  SH2movi(sh, 4, (11<<4)+3);
}
#define init_14_4_11_4
void FASTCALL call_14_4_11_4(SH2_struct* sh)
{
  SH2movi(sh, 4, (11<<4)+4);
}
#define init_14_4_11_5
void FASTCALL call_14_4_11_5(SH2_struct* sh)
{
  SH2movi(sh, 4, (11<<4)+5);
}
#define init_14_4_11_6
void FASTCALL call_14_4_11_6(SH2_struct* sh)
{
  SH2movi(sh, 4, (11<<4)+6);
}
#define init_14_4_11_7
void FASTCALL call_14_4_11_7(SH2_struct* sh)
{
  SH2movi(sh, 4, (11<<4)+7);
}
#define init_14_4_11_8
void FASTCALL call_14_4_11_8(SH2_struct* sh)
{
  SH2movi(sh, 4, (11<<4)+8);
}
#define init_14_4_11_9
void FASTCALL call_14_4_11_9(SH2_struct* sh)
{
  SH2movi(sh, 4, (11<<4)+9);
}
#define init_14_4_11_10
void FASTCALL call_14_4_11_10(SH2_struct* sh)
{
  SH2movi(sh, 4, (11<<4)+10);
}
#define init_14_4_11_11
void FASTCALL call_14_4_11_11(SH2_struct* sh)
{
  SH2movi(sh, 4, (11<<4)+11);
}
#define init_14_4_11_12
void FASTCALL call_14_4_11_12(SH2_struct* sh)
{
  SH2movi(sh, 4, (11<<4)+12);
}
#define init_14_4_11_13
void FASTCALL call_14_4_11_13(SH2_struct* sh)
{
  SH2movi(sh, 4, (11<<4)+13);
}
#define init_14_4_11_14
void FASTCALL call_14_4_11_14(SH2_struct* sh)
{
  SH2movi(sh, 4, (11<<4)+14);
}
#define init_14_4_11_15
void FASTCALL call_14_4_11_15(SH2_struct* sh)
{
  SH2movi(sh, 4, (11<<4)+15);
}
#define init_14_4_12_0
void FASTCALL call_14_4_12_0(SH2_struct* sh)
{
  SH2movi(sh, 4, (12<<4)+0);
}
#define init_14_4_12_1
void FASTCALL call_14_4_12_1(SH2_struct* sh)
{
  SH2movi(sh, 4, (12<<4)+1);
}
#define init_14_4_12_2
void FASTCALL call_14_4_12_2(SH2_struct* sh)
{
  SH2movi(sh, 4, (12<<4)+2);
}
#define init_14_4_12_3
void FASTCALL call_14_4_12_3(SH2_struct* sh)
{
  SH2movi(sh, 4, (12<<4)+3);
}
#define init_14_4_12_4
void FASTCALL call_14_4_12_4(SH2_struct* sh)
{
  SH2movi(sh, 4, (12<<4)+4);
}
#define init_14_4_12_5
void FASTCALL call_14_4_12_5(SH2_struct* sh)
{
  SH2movi(sh, 4, (12<<4)+5);
}
#define init_14_4_12_6
void FASTCALL call_14_4_12_6(SH2_struct* sh)
{
  SH2movi(sh, 4, (12<<4)+6);
}
#define init_14_4_12_7
void FASTCALL call_14_4_12_7(SH2_struct* sh)
{
  SH2movi(sh, 4, (12<<4)+7);
}
#define init_14_4_12_8
void FASTCALL call_14_4_12_8(SH2_struct* sh)
{
  SH2movi(sh, 4, (12<<4)+8);
}
#define init_14_4_12_9
void FASTCALL call_14_4_12_9(SH2_struct* sh)
{
  SH2movi(sh, 4, (12<<4)+9);
}
#define init_14_4_12_10
void FASTCALL call_14_4_12_10(SH2_struct* sh)
{
  SH2movi(sh, 4, (12<<4)+10);
}
#define init_14_4_12_11
void FASTCALL call_14_4_12_11(SH2_struct* sh)
{
  SH2movi(sh, 4, (12<<4)+11);
}
#define init_14_4_12_12
void FASTCALL call_14_4_12_12(SH2_struct* sh)
{
  SH2movi(sh, 4, (12<<4)+12);
}
#define init_14_4_12_13
void FASTCALL call_14_4_12_13(SH2_struct* sh)
{
  SH2movi(sh, 4, (12<<4)+13);
}
#define init_14_4_12_14
void FASTCALL call_14_4_12_14(SH2_struct* sh)
{
  SH2movi(sh, 4, (12<<4)+14);
}
#define init_14_4_12_15
void FASTCALL call_14_4_12_15(SH2_struct* sh)
{
  SH2movi(sh, 4, (12<<4)+15);
}
#define init_14_4_13_0
void FASTCALL call_14_4_13_0(SH2_struct* sh)
{
  SH2movi(sh, 4, (13<<4)+0);
}
#define init_14_4_13_1
void FASTCALL call_14_4_13_1(SH2_struct* sh)
{
  SH2movi(sh, 4, (13<<4)+1);
}
#define init_14_4_13_2
void FASTCALL call_14_4_13_2(SH2_struct* sh)
{
  SH2movi(sh, 4, (13<<4)+2);
}
#define init_14_4_13_3
void FASTCALL call_14_4_13_3(SH2_struct* sh)
{
  SH2movi(sh, 4, (13<<4)+3);
}
#define init_14_4_13_4
void FASTCALL call_14_4_13_4(SH2_struct* sh)
{
  SH2movi(sh, 4, (13<<4)+4);
}
#define init_14_4_13_5
void FASTCALL call_14_4_13_5(SH2_struct* sh)
{
  SH2movi(sh, 4, (13<<4)+5);
}
#define init_14_4_13_6
void FASTCALL call_14_4_13_6(SH2_struct* sh)
{
  SH2movi(sh, 4, (13<<4)+6);
}
#define init_14_4_13_7
void FASTCALL call_14_4_13_7(SH2_struct* sh)
{
  SH2movi(sh, 4, (13<<4)+7);
}
#define init_14_4_13_8
void FASTCALL call_14_4_13_8(SH2_struct* sh)
{
  SH2movi(sh, 4, (13<<4)+8);
}
#define init_14_4_13_9
void FASTCALL call_14_4_13_9(SH2_struct* sh)
{
  SH2movi(sh, 4, (13<<4)+9);
}
#define init_14_4_13_10
void FASTCALL call_14_4_13_10(SH2_struct* sh)
{
  SH2movi(sh, 4, (13<<4)+10);
}
#define init_14_4_13_11
void FASTCALL call_14_4_13_11(SH2_struct* sh)
{
  SH2movi(sh, 4, (13<<4)+11);
}
#define init_14_4_13_12
void FASTCALL call_14_4_13_12(SH2_struct* sh)
{
  SH2movi(sh, 4, (13<<4)+12);
}
#define init_14_4_13_13
void FASTCALL call_14_4_13_13(SH2_struct* sh)
{
  SH2movi(sh, 4, (13<<4)+13);
}
#define init_14_4_13_14
void FASTCALL call_14_4_13_14(SH2_struct* sh)
{
  SH2movi(sh, 4, (13<<4)+14);
}
#define init_14_4_13_15
void FASTCALL call_14_4_13_15(SH2_struct* sh)
{
  SH2movi(sh, 4, (13<<4)+15);
}
#define init_14_4_14_0
void FASTCALL call_14_4_14_0(SH2_struct* sh)
{
  SH2movi(sh, 4, (14<<4)+0);
}
#define init_14_4_14_1
void FASTCALL call_14_4_14_1(SH2_struct* sh)
{
  SH2movi(sh, 4, (14<<4)+1);
}
#define init_14_4_14_2
void FASTCALL call_14_4_14_2(SH2_struct* sh)
{
  SH2movi(sh, 4, (14<<4)+2);
}
#define init_14_4_14_3
void FASTCALL call_14_4_14_3(SH2_struct* sh)
{
  SH2movi(sh, 4, (14<<4)+3);
}
#define init_14_4_14_4
void FASTCALL call_14_4_14_4(SH2_struct* sh)
{
  SH2movi(sh, 4, (14<<4)+4);
}
#define init_14_4_14_5
void FASTCALL call_14_4_14_5(SH2_struct* sh)
{
  SH2movi(sh, 4, (14<<4)+5);
}
#define init_14_4_14_6
void FASTCALL call_14_4_14_6(SH2_struct* sh)
{
  SH2movi(sh, 4, (14<<4)+6);
}
#define init_14_4_14_7
void FASTCALL call_14_4_14_7(SH2_struct* sh)
{
  SH2movi(sh, 4, (14<<4)+7);
}
#define init_14_4_14_8
void FASTCALL call_14_4_14_8(SH2_struct* sh)
{
  SH2movi(sh, 4, (14<<4)+8);
}
#define init_14_4_14_9
void FASTCALL call_14_4_14_9(SH2_struct* sh)
{
  SH2movi(sh, 4, (14<<4)+9);
}
#define init_14_4_14_10
void FASTCALL call_14_4_14_10(SH2_struct* sh)
{
  SH2movi(sh, 4, (14<<4)+10);
}
#define init_14_4_14_11
void FASTCALL call_14_4_14_11(SH2_struct* sh)
{
  SH2movi(sh, 4, (14<<4)+11);
}
#define init_14_4_14_12
void FASTCALL call_14_4_14_12(SH2_struct* sh)
{
  SH2movi(sh, 4, (14<<4)+12);
}
#define init_14_4_14_13
void FASTCALL call_14_4_14_13(SH2_struct* sh)
{
  SH2movi(sh, 4, (14<<4)+13);
}
#define init_14_4_14_14
void FASTCALL call_14_4_14_14(SH2_struct* sh)
{
  SH2movi(sh, 4, (14<<4)+14);
}
#define init_14_4_14_15
void FASTCALL call_14_4_14_15(SH2_struct* sh)
{
  SH2movi(sh, 4, (14<<4)+15);
}
#define init_14_4_15_0
void FASTCALL call_14_4_15_0(SH2_struct* sh)
{
  SH2movi(sh, 4, (15<<4)+0);
}
#define init_14_4_15_1
void FASTCALL call_14_4_15_1(SH2_struct* sh)
{
  SH2movi(sh, 4, (15<<4)+1);
}
#define init_14_4_15_2
void FASTCALL call_14_4_15_2(SH2_struct* sh)
{
  SH2movi(sh, 4, (15<<4)+2);
}
#define init_14_4_15_3
void FASTCALL call_14_4_15_3(SH2_struct* sh)
{
  SH2movi(sh, 4, (15<<4)+3);
}
#define init_14_4_15_4
void FASTCALL call_14_4_15_4(SH2_struct* sh)
{
  SH2movi(sh, 4, (15<<4)+4);
}
#define init_14_4_15_5
void FASTCALL call_14_4_15_5(SH2_struct* sh)
{
  SH2movi(sh, 4, (15<<4)+5);
}
#define init_14_4_15_6
void FASTCALL call_14_4_15_6(SH2_struct* sh)
{
  SH2movi(sh, 4, (15<<4)+6);
}
#define init_14_4_15_7
void FASTCALL call_14_4_15_7(SH2_struct* sh)
{
  SH2movi(sh, 4, (15<<4)+7);
}
#define init_14_4_15_8
void FASTCALL call_14_4_15_8(SH2_struct* sh)
{
  SH2movi(sh, 4, (15<<4)+8);
}
#define init_14_4_15_9
void FASTCALL call_14_4_15_9(SH2_struct* sh)
{
  SH2movi(sh, 4, (15<<4)+9);
}
#define init_14_4_15_10
void FASTCALL call_14_4_15_10(SH2_struct* sh)
{
  SH2movi(sh, 4, (15<<4)+10);
}
#define init_14_4_15_11
void FASTCALL call_14_4_15_11(SH2_struct* sh)
{
  SH2movi(sh, 4, (15<<4)+11);
}
#define init_14_4_15_12
void FASTCALL call_14_4_15_12(SH2_struct* sh)
{
  SH2movi(sh, 4, (15<<4)+12);
}
#define init_14_4_15_13
void FASTCALL call_14_4_15_13(SH2_struct* sh)
{
  SH2movi(sh, 4, (15<<4)+13);
}
#define init_14_4_15_14
void FASTCALL call_14_4_15_14(SH2_struct* sh)
{
  SH2movi(sh, 4, (15<<4)+14);
}
#define init_14_4_15_15
void FASTCALL call_14_4_15_15(SH2_struct* sh)
{
  SH2movi(sh, 4, (15<<4)+15);
}
#define init_14_5_0_0
void FASTCALL call_14_5_0_0(SH2_struct* sh)
{
  SH2movi(sh, 5, (0<<4)+0);
}
#define init_14_5_0_1
void FASTCALL call_14_5_0_1(SH2_struct* sh)
{
  SH2movi(sh, 5, (0<<4)+1);
}
#define init_14_5_0_2
void FASTCALL call_14_5_0_2(SH2_struct* sh)
{
  SH2movi(sh, 5, (0<<4)+2);
}
#define init_14_5_0_3
void FASTCALL call_14_5_0_3(SH2_struct* sh)
{
  SH2movi(sh, 5, (0<<4)+3);
}
#define init_14_5_0_4
void FASTCALL call_14_5_0_4(SH2_struct* sh)
{
  SH2movi(sh, 5, (0<<4)+4);
}
#define init_14_5_0_5
void FASTCALL call_14_5_0_5(SH2_struct* sh)
{
  SH2movi(sh, 5, (0<<4)+5);
}
#define init_14_5_0_6
void FASTCALL call_14_5_0_6(SH2_struct* sh)
{
  SH2movi(sh, 5, (0<<4)+6);
}
#define init_14_5_0_7
void FASTCALL call_14_5_0_7(SH2_struct* sh)
{
  SH2movi(sh, 5, (0<<4)+7);
}
#define init_14_5_0_8
void FASTCALL call_14_5_0_8(SH2_struct* sh)
{
  SH2movi(sh, 5, (0<<4)+8);
}
#define init_14_5_0_9
void FASTCALL call_14_5_0_9(SH2_struct* sh)
{
  SH2movi(sh, 5, (0<<4)+9);
}
#define init_14_5_0_10
void FASTCALL call_14_5_0_10(SH2_struct* sh)
{
  SH2movi(sh, 5, (0<<4)+10);
}
#define init_14_5_0_11
void FASTCALL call_14_5_0_11(SH2_struct* sh)
{
  SH2movi(sh, 5, (0<<4)+11);
}
#define init_14_5_0_12
void FASTCALL call_14_5_0_12(SH2_struct* sh)
{
  SH2movi(sh, 5, (0<<4)+12);
}
#define init_14_5_0_13
void FASTCALL call_14_5_0_13(SH2_struct* sh)
{
  SH2movi(sh, 5, (0<<4)+13);
}
#define init_14_5_0_14
void FASTCALL call_14_5_0_14(SH2_struct* sh)
{
  SH2movi(sh, 5, (0<<4)+14);
}
#define init_14_5_0_15
void FASTCALL call_14_5_0_15(SH2_struct* sh)
{
  SH2movi(sh, 5, (0<<4)+15);
}
#define init_14_5_1_0
void FASTCALL call_14_5_1_0(SH2_struct* sh)
{
  SH2movi(sh, 5, (1<<4)+0);
}
#define init_14_5_1_1
void FASTCALL call_14_5_1_1(SH2_struct* sh)
{
  SH2movi(sh, 5, (1<<4)+1);
}
#define init_14_5_1_2
void FASTCALL call_14_5_1_2(SH2_struct* sh)
{
  SH2movi(sh, 5, (1<<4)+2);
}
#define init_14_5_1_3
void FASTCALL call_14_5_1_3(SH2_struct* sh)
{
  SH2movi(sh, 5, (1<<4)+3);
}
#define init_14_5_1_4
void FASTCALL call_14_5_1_4(SH2_struct* sh)
{
  SH2movi(sh, 5, (1<<4)+4);
}
#define init_14_5_1_5
void FASTCALL call_14_5_1_5(SH2_struct* sh)
{
  SH2movi(sh, 5, (1<<4)+5);
}
#define init_14_5_1_6
void FASTCALL call_14_5_1_6(SH2_struct* sh)
{
  SH2movi(sh, 5, (1<<4)+6);
}
#define init_14_5_1_7
void FASTCALL call_14_5_1_7(SH2_struct* sh)
{
  SH2movi(sh, 5, (1<<4)+7);
}
#define init_14_5_1_8
void FASTCALL call_14_5_1_8(SH2_struct* sh)
{
  SH2movi(sh, 5, (1<<4)+8);
}
#define init_14_5_1_9
void FASTCALL call_14_5_1_9(SH2_struct* sh)
{
  SH2movi(sh, 5, (1<<4)+9);
}
#define init_14_5_1_10
void FASTCALL call_14_5_1_10(SH2_struct* sh)
{
  SH2movi(sh, 5, (1<<4)+10);
}
#define init_14_5_1_11
void FASTCALL call_14_5_1_11(SH2_struct* sh)
{
  SH2movi(sh, 5, (1<<4)+11);
}
#define init_14_5_1_12
void FASTCALL call_14_5_1_12(SH2_struct* sh)
{
  SH2movi(sh, 5, (1<<4)+12);
}
#define init_14_5_1_13
void FASTCALL call_14_5_1_13(SH2_struct* sh)
{
  SH2movi(sh, 5, (1<<4)+13);
}
#define init_14_5_1_14
void FASTCALL call_14_5_1_14(SH2_struct* sh)
{
  SH2movi(sh, 5, (1<<4)+14);
}
#define init_14_5_1_15
void FASTCALL call_14_5_1_15(SH2_struct* sh)
{
  SH2movi(sh, 5, (1<<4)+15);
}
#define init_14_5_2_0
void FASTCALL call_14_5_2_0(SH2_struct* sh)
{
  SH2movi(sh, 5, (2<<4)+0);
}
#define init_14_5_2_1
void FASTCALL call_14_5_2_1(SH2_struct* sh)
{
  SH2movi(sh, 5, (2<<4)+1);
}
#define init_14_5_2_2
void FASTCALL call_14_5_2_2(SH2_struct* sh)
{
  SH2movi(sh, 5, (2<<4)+2);
}
#define init_14_5_2_3
void FASTCALL call_14_5_2_3(SH2_struct* sh)
{
  SH2movi(sh, 5, (2<<4)+3);
}
#define init_14_5_2_4
void FASTCALL call_14_5_2_4(SH2_struct* sh)
{
  SH2movi(sh, 5, (2<<4)+4);
}
#define init_14_5_2_5
void FASTCALL call_14_5_2_5(SH2_struct* sh)
{
  SH2movi(sh, 5, (2<<4)+5);
}
#define init_14_5_2_6
void FASTCALL call_14_5_2_6(SH2_struct* sh)
{
  SH2movi(sh, 5, (2<<4)+6);
}
#define init_14_5_2_7
void FASTCALL call_14_5_2_7(SH2_struct* sh)
{
  SH2movi(sh, 5, (2<<4)+7);
}
#define init_14_5_2_8
void FASTCALL call_14_5_2_8(SH2_struct* sh)
{
  SH2movi(sh, 5, (2<<4)+8);
}
#define init_14_5_2_9
void FASTCALL call_14_5_2_9(SH2_struct* sh)
{
  SH2movi(sh, 5, (2<<4)+9);
}
#define init_14_5_2_10
void FASTCALL call_14_5_2_10(SH2_struct* sh)
{
  SH2movi(sh, 5, (2<<4)+10);
}
#define init_14_5_2_11
void FASTCALL call_14_5_2_11(SH2_struct* sh)
{
  SH2movi(sh, 5, (2<<4)+11);
}
#define init_14_5_2_12
void FASTCALL call_14_5_2_12(SH2_struct* sh)
{
  SH2movi(sh, 5, (2<<4)+12);
}
#define init_14_5_2_13
void FASTCALL call_14_5_2_13(SH2_struct* sh)
{
  SH2movi(sh, 5, (2<<4)+13);
}
#define init_14_5_2_14
void FASTCALL call_14_5_2_14(SH2_struct* sh)
{
  SH2movi(sh, 5, (2<<4)+14);
}
#define init_14_5_2_15
void FASTCALL call_14_5_2_15(SH2_struct* sh)
{
  SH2movi(sh, 5, (2<<4)+15);
}
#define init_14_5_3_0
void FASTCALL call_14_5_3_0(SH2_struct* sh)
{
  SH2movi(sh, 5, (3<<4)+0);
}
#define init_14_5_3_1
void FASTCALL call_14_5_3_1(SH2_struct* sh)
{
  SH2movi(sh, 5, (3<<4)+1);
}
#define init_14_5_3_2
void FASTCALL call_14_5_3_2(SH2_struct* sh)
{
  SH2movi(sh, 5, (3<<4)+2);
}
#define init_14_5_3_3
void FASTCALL call_14_5_3_3(SH2_struct* sh)
{
  SH2movi(sh, 5, (3<<4)+3);
}
#define init_14_5_3_4
void FASTCALL call_14_5_3_4(SH2_struct* sh)
{
  SH2movi(sh, 5, (3<<4)+4);
}
#define init_14_5_3_5
void FASTCALL call_14_5_3_5(SH2_struct* sh)
{
  SH2movi(sh, 5, (3<<4)+5);
}
#define init_14_5_3_6
void FASTCALL call_14_5_3_6(SH2_struct* sh)
{
  SH2movi(sh, 5, (3<<4)+6);
}
#define init_14_5_3_7
void FASTCALL call_14_5_3_7(SH2_struct* sh)
{
  SH2movi(sh, 5, (3<<4)+7);
}
#define init_14_5_3_8
void FASTCALL call_14_5_3_8(SH2_struct* sh)
{
  SH2movi(sh, 5, (3<<4)+8);
}
#define init_14_5_3_9
void FASTCALL call_14_5_3_9(SH2_struct* sh)
{
  SH2movi(sh, 5, (3<<4)+9);
}
#define init_14_5_3_10
void FASTCALL call_14_5_3_10(SH2_struct* sh)
{
  SH2movi(sh, 5, (3<<4)+10);
}
#define init_14_5_3_11
void FASTCALL call_14_5_3_11(SH2_struct* sh)
{
  SH2movi(sh, 5, (3<<4)+11);
}
#define init_14_5_3_12
void FASTCALL call_14_5_3_12(SH2_struct* sh)
{
  SH2movi(sh, 5, (3<<4)+12);
}
#define init_14_5_3_13
void FASTCALL call_14_5_3_13(SH2_struct* sh)
{
  SH2movi(sh, 5, (3<<4)+13);
}
#define init_14_5_3_14
void FASTCALL call_14_5_3_14(SH2_struct* sh)
{
  SH2movi(sh, 5, (3<<4)+14);
}
#define init_14_5_3_15
void FASTCALL call_14_5_3_15(SH2_struct* sh)
{
  SH2movi(sh, 5, (3<<4)+15);
}
#define init_14_5_4_0
void FASTCALL call_14_5_4_0(SH2_struct* sh)
{
  SH2movi(sh, 5, (4<<4)+0);
}
#define init_14_5_4_1
void FASTCALL call_14_5_4_1(SH2_struct* sh)
{
  SH2movi(sh, 5, (4<<4)+1);
}
#define init_14_5_4_2
void FASTCALL call_14_5_4_2(SH2_struct* sh)
{
  SH2movi(sh, 5, (4<<4)+2);
}
#define init_14_5_4_3
void FASTCALL call_14_5_4_3(SH2_struct* sh)
{
  SH2movi(sh, 5, (4<<4)+3);
}
#define init_14_5_4_4
void FASTCALL call_14_5_4_4(SH2_struct* sh)
{
  SH2movi(sh, 5, (4<<4)+4);
}
#define init_14_5_4_5
void FASTCALL call_14_5_4_5(SH2_struct* sh)
{
  SH2movi(sh, 5, (4<<4)+5);
}
#define init_14_5_4_6
void FASTCALL call_14_5_4_6(SH2_struct* sh)
{
  SH2movi(sh, 5, (4<<4)+6);
}
#define init_14_5_4_7
void FASTCALL call_14_5_4_7(SH2_struct* sh)
{
  SH2movi(sh, 5, (4<<4)+7);
}
#define init_14_5_4_8
void FASTCALL call_14_5_4_8(SH2_struct* sh)
{
  SH2movi(sh, 5, (4<<4)+8);
}
#define init_14_5_4_9
void FASTCALL call_14_5_4_9(SH2_struct* sh)
{
  SH2movi(sh, 5, (4<<4)+9);
}
#define init_14_5_4_10
void FASTCALL call_14_5_4_10(SH2_struct* sh)
{
  SH2movi(sh, 5, (4<<4)+10);
}
#define init_14_5_4_11
void FASTCALL call_14_5_4_11(SH2_struct* sh)
{
  SH2movi(sh, 5, (4<<4)+11);
}
#define init_14_5_4_12
void FASTCALL call_14_5_4_12(SH2_struct* sh)
{
  SH2movi(sh, 5, (4<<4)+12);
}
#define init_14_5_4_13
void FASTCALL call_14_5_4_13(SH2_struct* sh)
{
  SH2movi(sh, 5, (4<<4)+13);
}
#define init_14_5_4_14
void FASTCALL call_14_5_4_14(SH2_struct* sh)
{
  SH2movi(sh, 5, (4<<4)+14);
}
#define init_14_5_4_15
void FASTCALL call_14_5_4_15(SH2_struct* sh)
{
  SH2movi(sh, 5, (4<<4)+15);
}
#define init_14_5_5_0
void FASTCALL call_14_5_5_0(SH2_struct* sh)
{
  SH2movi(sh, 5, (5<<4)+0);
}
#define init_14_5_5_1
void FASTCALL call_14_5_5_1(SH2_struct* sh)
{
  SH2movi(sh, 5, (5<<4)+1);
}
#define init_14_5_5_2
void FASTCALL call_14_5_5_2(SH2_struct* sh)
{
  SH2movi(sh, 5, (5<<4)+2);
}
#define init_14_5_5_3
void FASTCALL call_14_5_5_3(SH2_struct* sh)
{
  SH2movi(sh, 5, (5<<4)+3);
}
#define init_14_5_5_4
void FASTCALL call_14_5_5_4(SH2_struct* sh)
{
  SH2movi(sh, 5, (5<<4)+4);
}
#define init_14_5_5_5
void FASTCALL call_14_5_5_5(SH2_struct* sh)
{
  SH2movi(sh, 5, (5<<4)+5);
}
#define init_14_5_5_6
void FASTCALL call_14_5_5_6(SH2_struct* sh)
{
  SH2movi(sh, 5, (5<<4)+6);
}
#define init_14_5_5_7
void FASTCALL call_14_5_5_7(SH2_struct* sh)
{
  SH2movi(sh, 5, (5<<4)+7);
}
#define init_14_5_5_8
void FASTCALL call_14_5_5_8(SH2_struct* sh)
{
  SH2movi(sh, 5, (5<<4)+8);
}
#define init_14_5_5_9
void FASTCALL call_14_5_5_9(SH2_struct* sh)
{
  SH2movi(sh, 5, (5<<4)+9);
}
#define init_14_5_5_10
void FASTCALL call_14_5_5_10(SH2_struct* sh)
{
  SH2movi(sh, 5, (5<<4)+10);
}
#define init_14_5_5_11
void FASTCALL call_14_5_5_11(SH2_struct* sh)
{
  SH2movi(sh, 5, (5<<4)+11);
}
#define init_14_5_5_12
void FASTCALL call_14_5_5_12(SH2_struct* sh)
{
  SH2movi(sh, 5, (5<<4)+12);
}
#define init_14_5_5_13
void FASTCALL call_14_5_5_13(SH2_struct* sh)
{
  SH2movi(sh, 5, (5<<4)+13);
}
#define init_14_5_5_14
void FASTCALL call_14_5_5_14(SH2_struct* sh)
{
  SH2movi(sh, 5, (5<<4)+14);
}
#define init_14_5_5_15
void FASTCALL call_14_5_5_15(SH2_struct* sh)
{
  SH2movi(sh, 5, (5<<4)+15);
}
#define init_14_5_6_0
void FASTCALL call_14_5_6_0(SH2_struct* sh)
{
  SH2movi(sh, 5, (6<<4)+0);
}
#define init_14_5_6_1
void FASTCALL call_14_5_6_1(SH2_struct* sh)
{
  SH2movi(sh, 5, (6<<4)+1);
}
#define init_14_5_6_2
void FASTCALL call_14_5_6_2(SH2_struct* sh)
{
  SH2movi(sh, 5, (6<<4)+2);
}
#define init_14_5_6_3
void FASTCALL call_14_5_6_3(SH2_struct* sh)
{
  SH2movi(sh, 5, (6<<4)+3);
}
#define init_14_5_6_4
void FASTCALL call_14_5_6_4(SH2_struct* sh)
{
  SH2movi(sh, 5, (6<<4)+4);
}
#define init_14_5_6_5
void FASTCALL call_14_5_6_5(SH2_struct* sh)
{
  SH2movi(sh, 5, (6<<4)+5);
}
#define init_14_5_6_6
void FASTCALL call_14_5_6_6(SH2_struct* sh)
{
  SH2movi(sh, 5, (6<<4)+6);
}
#define init_14_5_6_7
void FASTCALL call_14_5_6_7(SH2_struct* sh)
{
  SH2movi(sh, 5, (6<<4)+7);
}
#define init_14_5_6_8
void FASTCALL call_14_5_6_8(SH2_struct* sh)
{
  SH2movi(sh, 5, (6<<4)+8);
}
#define init_14_5_6_9
void FASTCALL call_14_5_6_9(SH2_struct* sh)
{
  SH2movi(sh, 5, (6<<4)+9);
}
#define init_14_5_6_10
void FASTCALL call_14_5_6_10(SH2_struct* sh)
{
  SH2movi(sh, 5, (6<<4)+10);
}
#define init_14_5_6_11
void FASTCALL call_14_5_6_11(SH2_struct* sh)
{
  SH2movi(sh, 5, (6<<4)+11);
}
#define init_14_5_6_12
void FASTCALL call_14_5_6_12(SH2_struct* sh)
{
  SH2movi(sh, 5, (6<<4)+12);
}
#define init_14_5_6_13
void FASTCALL call_14_5_6_13(SH2_struct* sh)
{
  SH2movi(sh, 5, (6<<4)+13);
}
#define init_14_5_6_14
void FASTCALL call_14_5_6_14(SH2_struct* sh)
{
  SH2movi(sh, 5, (6<<4)+14);
}
#define init_14_5_6_15
void FASTCALL call_14_5_6_15(SH2_struct* sh)
{
  SH2movi(sh, 5, (6<<4)+15);
}
#define init_14_5_7_0
void FASTCALL call_14_5_7_0(SH2_struct* sh)
{
  SH2movi(sh, 5, (7<<4)+0);
}
#define init_14_5_7_1
void FASTCALL call_14_5_7_1(SH2_struct* sh)
{
  SH2movi(sh, 5, (7<<4)+1);
}
#define init_14_5_7_2
void FASTCALL call_14_5_7_2(SH2_struct* sh)
{
  SH2movi(sh, 5, (7<<4)+2);
}
#define init_14_5_7_3
void FASTCALL call_14_5_7_3(SH2_struct* sh)
{
  SH2movi(sh, 5, (7<<4)+3);
}
#define init_14_5_7_4
void FASTCALL call_14_5_7_4(SH2_struct* sh)
{
  SH2movi(sh, 5, (7<<4)+4);
}
#define init_14_5_7_5
void FASTCALL call_14_5_7_5(SH2_struct* sh)
{
  SH2movi(sh, 5, (7<<4)+5);
}
#define init_14_5_7_6
void FASTCALL call_14_5_7_6(SH2_struct* sh)
{
  SH2movi(sh, 5, (7<<4)+6);
}
#define init_14_5_7_7
void FASTCALL call_14_5_7_7(SH2_struct* sh)
{
  SH2movi(sh, 5, (7<<4)+7);
}
#define init_14_5_7_8
void FASTCALL call_14_5_7_8(SH2_struct* sh)
{
  SH2movi(sh, 5, (7<<4)+8);
}
#define init_14_5_7_9
void FASTCALL call_14_5_7_9(SH2_struct* sh)
{
  SH2movi(sh, 5, (7<<4)+9);
}
#define init_14_5_7_10
void FASTCALL call_14_5_7_10(SH2_struct* sh)
{
  SH2movi(sh, 5, (7<<4)+10);
}
#define init_14_5_7_11
void FASTCALL call_14_5_7_11(SH2_struct* sh)
{
  SH2movi(sh, 5, (7<<4)+11);
}
#define init_14_5_7_12
void FASTCALL call_14_5_7_12(SH2_struct* sh)
{
  SH2movi(sh, 5, (7<<4)+12);
}
#define init_14_5_7_13
void FASTCALL call_14_5_7_13(SH2_struct* sh)
{
  SH2movi(sh, 5, (7<<4)+13);
}
#define init_14_5_7_14
void FASTCALL call_14_5_7_14(SH2_struct* sh)
{
  SH2movi(sh, 5, (7<<4)+14);
}
#define init_14_5_7_15
void FASTCALL call_14_5_7_15(SH2_struct* sh)
{
  SH2movi(sh, 5, (7<<4)+15);
}
#define init_14_5_8_0
void FASTCALL call_14_5_8_0(SH2_struct* sh)
{
  SH2movi(sh, 5, (8<<4)+0);
}
#define init_14_5_8_1
void FASTCALL call_14_5_8_1(SH2_struct* sh)
{
  SH2movi(sh, 5, (8<<4)+1);
}
#define init_14_5_8_2
void FASTCALL call_14_5_8_2(SH2_struct* sh)
{
  SH2movi(sh, 5, (8<<4)+2);
}
#define init_14_5_8_3
void FASTCALL call_14_5_8_3(SH2_struct* sh)
{
  SH2movi(sh, 5, (8<<4)+3);
}
#define init_14_5_8_4
void FASTCALL call_14_5_8_4(SH2_struct* sh)
{
  SH2movi(sh, 5, (8<<4)+4);
}
#define init_14_5_8_5
void FASTCALL call_14_5_8_5(SH2_struct* sh)
{
  SH2movi(sh, 5, (8<<4)+5);
}
#define init_14_5_8_6
void FASTCALL call_14_5_8_6(SH2_struct* sh)
{
  SH2movi(sh, 5, (8<<4)+6);
}
#define init_14_5_8_7
void FASTCALL call_14_5_8_7(SH2_struct* sh)
{
  SH2movi(sh, 5, (8<<4)+7);
}
#define init_14_5_8_8
void FASTCALL call_14_5_8_8(SH2_struct* sh)
{
  SH2movi(sh, 5, (8<<4)+8);
}
#define init_14_5_8_9
void FASTCALL call_14_5_8_9(SH2_struct* sh)
{
  SH2movi(sh, 5, (8<<4)+9);
}
#define init_14_5_8_10
void FASTCALL call_14_5_8_10(SH2_struct* sh)
{
  SH2movi(sh, 5, (8<<4)+10);
}
#define init_14_5_8_11
void FASTCALL call_14_5_8_11(SH2_struct* sh)
{
  SH2movi(sh, 5, (8<<4)+11);
}
#define init_14_5_8_12
void FASTCALL call_14_5_8_12(SH2_struct* sh)
{
  SH2movi(sh, 5, (8<<4)+12);
}
#define init_14_5_8_13
void FASTCALL call_14_5_8_13(SH2_struct* sh)
{
  SH2movi(sh, 5, (8<<4)+13);
}
#define init_14_5_8_14
void FASTCALL call_14_5_8_14(SH2_struct* sh)
{
  SH2movi(sh, 5, (8<<4)+14);
}
#define init_14_5_8_15
void FASTCALL call_14_5_8_15(SH2_struct* sh)
{
  SH2movi(sh, 5, (8<<4)+15);
}
#define init_14_5_9_0
void FASTCALL call_14_5_9_0(SH2_struct* sh)
{
  SH2movi(sh, 5, (9<<4)+0);
}
#define init_14_5_9_1
void FASTCALL call_14_5_9_1(SH2_struct* sh)
{
  SH2movi(sh, 5, (9<<4)+1);
}
#define init_14_5_9_2
void FASTCALL call_14_5_9_2(SH2_struct* sh)
{
  SH2movi(sh, 5, (9<<4)+2);
}
#define init_14_5_9_3
void FASTCALL call_14_5_9_3(SH2_struct* sh)
{
  SH2movi(sh, 5, (9<<4)+3);
}
#define init_14_5_9_4
void FASTCALL call_14_5_9_4(SH2_struct* sh)
{
  SH2movi(sh, 5, (9<<4)+4);
}
#define init_14_5_9_5
void FASTCALL call_14_5_9_5(SH2_struct* sh)
{
  SH2movi(sh, 5, (9<<4)+5);
}
#define init_14_5_9_6
void FASTCALL call_14_5_9_6(SH2_struct* sh)
{
  SH2movi(sh, 5, (9<<4)+6);
}
#define init_14_5_9_7
void FASTCALL call_14_5_9_7(SH2_struct* sh)
{
  SH2movi(sh, 5, (9<<4)+7);
}
#define init_14_5_9_8
void FASTCALL call_14_5_9_8(SH2_struct* sh)
{
  SH2movi(sh, 5, (9<<4)+8);
}
#define init_14_5_9_9
void FASTCALL call_14_5_9_9(SH2_struct* sh)
{
  SH2movi(sh, 5, (9<<4)+9);
}
#define init_14_5_9_10
void FASTCALL call_14_5_9_10(SH2_struct* sh)
{
  SH2movi(sh, 5, (9<<4)+10);
}
#define init_14_5_9_11
void FASTCALL call_14_5_9_11(SH2_struct* sh)
{
  SH2movi(sh, 5, (9<<4)+11);
}
#define init_14_5_9_12
void FASTCALL call_14_5_9_12(SH2_struct* sh)
{
  SH2movi(sh, 5, (9<<4)+12);
}
#define init_14_5_9_13
void FASTCALL call_14_5_9_13(SH2_struct* sh)
{
  SH2movi(sh, 5, (9<<4)+13);
}
#define init_14_5_9_14
void FASTCALL call_14_5_9_14(SH2_struct* sh)
{
  SH2movi(sh, 5, (9<<4)+14);
}
#define init_14_5_9_15
void FASTCALL call_14_5_9_15(SH2_struct* sh)
{
  SH2movi(sh, 5, (9<<4)+15);
}
#define init_14_5_10_0
void FASTCALL call_14_5_10_0(SH2_struct* sh)
{
  SH2movi(sh, 5, (10<<4)+0);
}
#define init_14_5_10_1
void FASTCALL call_14_5_10_1(SH2_struct* sh)
{
  SH2movi(sh, 5, (10<<4)+1);
}
#define init_14_5_10_2
void FASTCALL call_14_5_10_2(SH2_struct* sh)
{
  SH2movi(sh, 5, (10<<4)+2);
}
#define init_14_5_10_3
void FASTCALL call_14_5_10_3(SH2_struct* sh)
{
  SH2movi(sh, 5, (10<<4)+3);
}
#define init_14_5_10_4
void FASTCALL call_14_5_10_4(SH2_struct* sh)
{
  SH2movi(sh, 5, (10<<4)+4);
}
#define init_14_5_10_5
void FASTCALL call_14_5_10_5(SH2_struct* sh)
{
  SH2movi(sh, 5, (10<<4)+5);
}
#define init_14_5_10_6
void FASTCALL call_14_5_10_6(SH2_struct* sh)
{
  SH2movi(sh, 5, (10<<4)+6);
}
#define init_14_5_10_7
void FASTCALL call_14_5_10_7(SH2_struct* sh)
{
  SH2movi(sh, 5, (10<<4)+7);
}
#define init_14_5_10_8
void FASTCALL call_14_5_10_8(SH2_struct* sh)
{
  SH2movi(sh, 5, (10<<4)+8);
}
#define init_14_5_10_9
void FASTCALL call_14_5_10_9(SH2_struct* sh)
{
  SH2movi(sh, 5, (10<<4)+9);
}
#define init_14_5_10_10
void FASTCALL call_14_5_10_10(SH2_struct* sh)
{
  SH2movi(sh, 5, (10<<4)+10);
}
#define init_14_5_10_11
void FASTCALL call_14_5_10_11(SH2_struct* sh)
{
  SH2movi(sh, 5, (10<<4)+11);
}
#define init_14_5_10_12
void FASTCALL call_14_5_10_12(SH2_struct* sh)
{
  SH2movi(sh, 5, (10<<4)+12);
}
#define init_14_5_10_13
void FASTCALL call_14_5_10_13(SH2_struct* sh)
{
  SH2movi(sh, 5, (10<<4)+13);
}
#define init_14_5_10_14
void FASTCALL call_14_5_10_14(SH2_struct* sh)
{
  SH2movi(sh, 5, (10<<4)+14);
}
#define init_14_5_10_15
void FASTCALL call_14_5_10_15(SH2_struct* sh)
{
  SH2movi(sh, 5, (10<<4)+15);
}
#define init_14_5_11_0
void FASTCALL call_14_5_11_0(SH2_struct* sh)
{
  SH2movi(sh, 5, (11<<4)+0);
}
#define init_14_5_11_1
void FASTCALL call_14_5_11_1(SH2_struct* sh)
{
  SH2movi(sh, 5, (11<<4)+1);
}
#define init_14_5_11_2
void FASTCALL call_14_5_11_2(SH2_struct* sh)
{
  SH2movi(sh, 5, (11<<4)+2);
}
#define init_14_5_11_3
void FASTCALL call_14_5_11_3(SH2_struct* sh)
{
  SH2movi(sh, 5, (11<<4)+3);
}
#define init_14_5_11_4
void FASTCALL call_14_5_11_4(SH2_struct* sh)
{
  SH2movi(sh, 5, (11<<4)+4);
}
#define init_14_5_11_5
void FASTCALL call_14_5_11_5(SH2_struct* sh)
{
  SH2movi(sh, 5, (11<<4)+5);
}
#define init_14_5_11_6
void FASTCALL call_14_5_11_6(SH2_struct* sh)
{
  SH2movi(sh, 5, (11<<4)+6);
}
#define init_14_5_11_7
void FASTCALL call_14_5_11_7(SH2_struct* sh)
{
  SH2movi(sh, 5, (11<<4)+7);
}
#define init_14_5_11_8
void FASTCALL call_14_5_11_8(SH2_struct* sh)
{
  SH2movi(sh, 5, (11<<4)+8);
}
#define init_14_5_11_9
void FASTCALL call_14_5_11_9(SH2_struct* sh)
{
  SH2movi(sh, 5, (11<<4)+9);
}
#define init_14_5_11_10
void FASTCALL call_14_5_11_10(SH2_struct* sh)
{
  SH2movi(sh, 5, (11<<4)+10);
}
#define init_14_5_11_11
void FASTCALL call_14_5_11_11(SH2_struct* sh)
{
  SH2movi(sh, 5, (11<<4)+11);
}
#define init_14_5_11_12
void FASTCALL call_14_5_11_12(SH2_struct* sh)
{
  SH2movi(sh, 5, (11<<4)+12);
}
#define init_14_5_11_13
void FASTCALL call_14_5_11_13(SH2_struct* sh)
{
  SH2movi(sh, 5, (11<<4)+13);
}
#define init_14_5_11_14
void FASTCALL call_14_5_11_14(SH2_struct* sh)
{
  SH2movi(sh, 5, (11<<4)+14);
}
#define init_14_5_11_15
void FASTCALL call_14_5_11_15(SH2_struct* sh)
{
  SH2movi(sh, 5, (11<<4)+15);
}
#define init_14_5_12_0
void FASTCALL call_14_5_12_0(SH2_struct* sh)
{
  SH2movi(sh, 5, (12<<4)+0);
}
#define init_14_5_12_1
void FASTCALL call_14_5_12_1(SH2_struct* sh)
{
  SH2movi(sh, 5, (12<<4)+1);
}
#define init_14_5_12_2
void FASTCALL call_14_5_12_2(SH2_struct* sh)
{
  SH2movi(sh, 5, (12<<4)+2);
}
#define init_14_5_12_3
void FASTCALL call_14_5_12_3(SH2_struct* sh)
{
  SH2movi(sh, 5, (12<<4)+3);
}
#define init_14_5_12_4
void FASTCALL call_14_5_12_4(SH2_struct* sh)
{
  SH2movi(sh, 5, (12<<4)+4);
}
#define init_14_5_12_5
void FASTCALL call_14_5_12_5(SH2_struct* sh)
{
  SH2movi(sh, 5, (12<<4)+5);
}
#define init_14_5_12_6
void FASTCALL call_14_5_12_6(SH2_struct* sh)
{
  SH2movi(sh, 5, (12<<4)+6);
}
#define init_14_5_12_7
void FASTCALL call_14_5_12_7(SH2_struct* sh)
{
  SH2movi(sh, 5, (12<<4)+7);
}
#define init_14_5_12_8
void FASTCALL call_14_5_12_8(SH2_struct* sh)
{
  SH2movi(sh, 5, (12<<4)+8);
}
#define init_14_5_12_9
void FASTCALL call_14_5_12_9(SH2_struct* sh)
{
  SH2movi(sh, 5, (12<<4)+9);
}
#define init_14_5_12_10
void FASTCALL call_14_5_12_10(SH2_struct* sh)
{
  SH2movi(sh, 5, (12<<4)+10);
}
#define init_14_5_12_11
void FASTCALL call_14_5_12_11(SH2_struct* sh)
{
  SH2movi(sh, 5, (12<<4)+11);
}
#define init_14_5_12_12
void FASTCALL call_14_5_12_12(SH2_struct* sh)
{
  SH2movi(sh, 5, (12<<4)+12);
}
#define init_14_5_12_13
void FASTCALL call_14_5_12_13(SH2_struct* sh)
{
  SH2movi(sh, 5, (12<<4)+13);
}
#define init_14_5_12_14
void FASTCALL call_14_5_12_14(SH2_struct* sh)
{
  SH2movi(sh, 5, (12<<4)+14);
}
#define init_14_5_12_15
void FASTCALL call_14_5_12_15(SH2_struct* sh)
{
  SH2movi(sh, 5, (12<<4)+15);
}
#define init_14_5_13_0
void FASTCALL call_14_5_13_0(SH2_struct* sh)
{
  SH2movi(sh, 5, (13<<4)+0);
}
#define init_14_5_13_1
void FASTCALL call_14_5_13_1(SH2_struct* sh)
{
  SH2movi(sh, 5, (13<<4)+1);
}
#define init_14_5_13_2
void FASTCALL call_14_5_13_2(SH2_struct* sh)
{
  SH2movi(sh, 5, (13<<4)+2);
}
#define init_14_5_13_3
void FASTCALL call_14_5_13_3(SH2_struct* sh)
{
  SH2movi(sh, 5, (13<<4)+3);
}
#define init_14_5_13_4
void FASTCALL call_14_5_13_4(SH2_struct* sh)
{
  SH2movi(sh, 5, (13<<4)+4);
}
#define init_14_5_13_5
void FASTCALL call_14_5_13_5(SH2_struct* sh)
{
  SH2movi(sh, 5, (13<<4)+5);
}
#define init_14_5_13_6
void FASTCALL call_14_5_13_6(SH2_struct* sh)
{
  SH2movi(sh, 5, (13<<4)+6);
}
#define init_14_5_13_7
void FASTCALL call_14_5_13_7(SH2_struct* sh)
{
  SH2movi(sh, 5, (13<<4)+7);
}
#define init_14_5_13_8
void FASTCALL call_14_5_13_8(SH2_struct* sh)
{
  SH2movi(sh, 5, (13<<4)+8);
}
#define init_14_5_13_9
void FASTCALL call_14_5_13_9(SH2_struct* sh)
{
  SH2movi(sh, 5, (13<<4)+9);
}
#define init_14_5_13_10
void FASTCALL call_14_5_13_10(SH2_struct* sh)
{
  SH2movi(sh, 5, (13<<4)+10);
}
#define init_14_5_13_11
void FASTCALL call_14_5_13_11(SH2_struct* sh)
{
  SH2movi(sh, 5, (13<<4)+11);
}
#define init_14_5_13_12
void FASTCALL call_14_5_13_12(SH2_struct* sh)
{
  SH2movi(sh, 5, (13<<4)+12);
}
#define init_14_5_13_13
void FASTCALL call_14_5_13_13(SH2_struct* sh)
{
  SH2movi(sh, 5, (13<<4)+13);
}
#define init_14_5_13_14
void FASTCALL call_14_5_13_14(SH2_struct* sh)
{
  SH2movi(sh, 5, (13<<4)+14);
}
#define init_14_5_13_15
void FASTCALL call_14_5_13_15(SH2_struct* sh)
{
  SH2movi(sh, 5, (13<<4)+15);
}
#define init_14_5_14_0
void FASTCALL call_14_5_14_0(SH2_struct* sh)
{
  SH2movi(sh, 5, (14<<4)+0);
}
#define init_14_5_14_1
void FASTCALL call_14_5_14_1(SH2_struct* sh)
{
  SH2movi(sh, 5, (14<<4)+1);
}
#define init_14_5_14_2
void FASTCALL call_14_5_14_2(SH2_struct* sh)
{
  SH2movi(sh, 5, (14<<4)+2);
}
#define init_14_5_14_3
void FASTCALL call_14_5_14_3(SH2_struct* sh)
{
  SH2movi(sh, 5, (14<<4)+3);
}
#define init_14_5_14_4
void FASTCALL call_14_5_14_4(SH2_struct* sh)
{
  SH2movi(sh, 5, (14<<4)+4);
}
#define init_14_5_14_5
void FASTCALL call_14_5_14_5(SH2_struct* sh)
{
  SH2movi(sh, 5, (14<<4)+5);
}
#define init_14_5_14_6
void FASTCALL call_14_5_14_6(SH2_struct* sh)
{
  SH2movi(sh, 5, (14<<4)+6);
}
#define init_14_5_14_7
void FASTCALL call_14_5_14_7(SH2_struct* sh)
{
  SH2movi(sh, 5, (14<<4)+7);
}
#define init_14_5_14_8
void FASTCALL call_14_5_14_8(SH2_struct* sh)
{
  SH2movi(sh, 5, (14<<4)+8);
}
#define init_14_5_14_9
void FASTCALL call_14_5_14_9(SH2_struct* sh)
{
  SH2movi(sh, 5, (14<<4)+9);
}
#define init_14_5_14_10
void FASTCALL call_14_5_14_10(SH2_struct* sh)
{
  SH2movi(sh, 5, (14<<4)+10);
}
#define init_14_5_14_11
void FASTCALL call_14_5_14_11(SH2_struct* sh)
{
  SH2movi(sh, 5, (14<<4)+11);
}
#define init_14_5_14_12
void FASTCALL call_14_5_14_12(SH2_struct* sh)
{
  SH2movi(sh, 5, (14<<4)+12);
}
#define init_14_5_14_13
void FASTCALL call_14_5_14_13(SH2_struct* sh)
{
  SH2movi(sh, 5, (14<<4)+13);
}
#define init_14_5_14_14
void FASTCALL call_14_5_14_14(SH2_struct* sh)
{
  SH2movi(sh, 5, (14<<4)+14);
}
#define init_14_5_14_15
void FASTCALL call_14_5_14_15(SH2_struct* sh)
{
  SH2movi(sh, 5, (14<<4)+15);
}
#define init_14_5_15_0
void FASTCALL call_14_5_15_0(SH2_struct* sh)
{
  SH2movi(sh, 5, (15<<4)+0);
}
#define init_14_5_15_1
void FASTCALL call_14_5_15_1(SH2_struct* sh)
{
  SH2movi(sh, 5, (15<<4)+1);
}
#define init_14_5_15_2
void FASTCALL call_14_5_15_2(SH2_struct* sh)
{
  SH2movi(sh, 5, (15<<4)+2);
}
#define init_14_5_15_3
void FASTCALL call_14_5_15_3(SH2_struct* sh)
{
  SH2movi(sh, 5, (15<<4)+3);
}
#define init_14_5_15_4
void FASTCALL call_14_5_15_4(SH2_struct* sh)
{
  SH2movi(sh, 5, (15<<4)+4);
}
#define init_14_5_15_5
void FASTCALL call_14_5_15_5(SH2_struct* sh)
{
  SH2movi(sh, 5, (15<<4)+5);
}
#define init_14_5_15_6
void FASTCALL call_14_5_15_6(SH2_struct* sh)
{
  SH2movi(sh, 5, (15<<4)+6);
}
#define init_14_5_15_7
void FASTCALL call_14_5_15_7(SH2_struct* sh)
{
  SH2movi(sh, 5, (15<<4)+7);
}
#define init_14_5_15_8
void FASTCALL call_14_5_15_8(SH2_struct* sh)
{
  SH2movi(sh, 5, (15<<4)+8);
}
#define init_14_5_15_9
void FASTCALL call_14_5_15_9(SH2_struct* sh)
{
  SH2movi(sh, 5, (15<<4)+9);
}
#define init_14_5_15_10
void FASTCALL call_14_5_15_10(SH2_struct* sh)
{
  SH2movi(sh, 5, (15<<4)+10);
}
#define init_14_5_15_11
void FASTCALL call_14_5_15_11(SH2_struct* sh)
{
  SH2movi(sh, 5, (15<<4)+11);
}
#define init_14_5_15_12
void FASTCALL call_14_5_15_12(SH2_struct* sh)
{
  SH2movi(sh, 5, (15<<4)+12);
}
#define init_14_5_15_13
void FASTCALL call_14_5_15_13(SH2_struct* sh)
{
  SH2movi(sh, 5, (15<<4)+13);
}
#define init_14_5_15_14
void FASTCALL call_14_5_15_14(SH2_struct* sh)
{
  SH2movi(sh, 5, (15<<4)+14);
}
#define init_14_5_15_15
void FASTCALL call_14_5_15_15(SH2_struct* sh)
{
  SH2movi(sh, 5, (15<<4)+15);
}
#define init_14_6_0_0
void FASTCALL call_14_6_0_0(SH2_struct* sh)
{
  SH2movi(sh, 6, (0<<4)+0);
}
#define init_14_6_0_1
void FASTCALL call_14_6_0_1(SH2_struct* sh)
{
  SH2movi(sh, 6, (0<<4)+1);
}
#define init_14_6_0_2
void FASTCALL call_14_6_0_2(SH2_struct* sh)
{
  SH2movi(sh, 6, (0<<4)+2);
}
#define init_14_6_0_3
void FASTCALL call_14_6_0_3(SH2_struct* sh)
{
  SH2movi(sh, 6, (0<<4)+3);
}
#define init_14_6_0_4
void FASTCALL call_14_6_0_4(SH2_struct* sh)
{
  SH2movi(sh, 6, (0<<4)+4);
}
#define init_14_6_0_5
void FASTCALL call_14_6_0_5(SH2_struct* sh)
{
  SH2movi(sh, 6, (0<<4)+5);
}
#define init_14_6_0_6
void FASTCALL call_14_6_0_6(SH2_struct* sh)
{
  SH2movi(sh, 6, (0<<4)+6);
}
#define init_14_6_0_7
void FASTCALL call_14_6_0_7(SH2_struct* sh)
{
  SH2movi(sh, 6, (0<<4)+7);
}
#define init_14_6_0_8
void FASTCALL call_14_6_0_8(SH2_struct* sh)
{
  SH2movi(sh, 6, (0<<4)+8);
}
#define init_14_6_0_9
void FASTCALL call_14_6_0_9(SH2_struct* sh)
{
  SH2movi(sh, 6, (0<<4)+9);
}
#define init_14_6_0_10
void FASTCALL call_14_6_0_10(SH2_struct* sh)
{
  SH2movi(sh, 6, (0<<4)+10);
}
#define init_14_6_0_11
void FASTCALL call_14_6_0_11(SH2_struct* sh)
{
  SH2movi(sh, 6, (0<<4)+11);
}
#define init_14_6_0_12
void FASTCALL call_14_6_0_12(SH2_struct* sh)
{
  SH2movi(sh, 6, (0<<4)+12);
}
#define init_14_6_0_13
void FASTCALL call_14_6_0_13(SH2_struct* sh)
{
  SH2movi(sh, 6, (0<<4)+13);
}
#define init_14_6_0_14
void FASTCALL call_14_6_0_14(SH2_struct* sh)
{
  SH2movi(sh, 6, (0<<4)+14);
}
#define init_14_6_0_15
void FASTCALL call_14_6_0_15(SH2_struct* sh)
{
  SH2movi(sh, 6, (0<<4)+15);
}
#define init_14_6_1_0
void FASTCALL call_14_6_1_0(SH2_struct* sh)
{
  SH2movi(sh, 6, (1<<4)+0);
}
#define init_14_6_1_1
void FASTCALL call_14_6_1_1(SH2_struct* sh)
{
  SH2movi(sh, 6, (1<<4)+1);
}
#define init_14_6_1_2
void FASTCALL call_14_6_1_2(SH2_struct* sh)
{
  SH2movi(sh, 6, (1<<4)+2);
}
#define init_14_6_1_3
void FASTCALL call_14_6_1_3(SH2_struct* sh)
{
  SH2movi(sh, 6, (1<<4)+3);
}
#define init_14_6_1_4
void FASTCALL call_14_6_1_4(SH2_struct* sh)
{
  SH2movi(sh, 6, (1<<4)+4);
}
#define init_14_6_1_5
void FASTCALL call_14_6_1_5(SH2_struct* sh)
{
  SH2movi(sh, 6, (1<<4)+5);
}
#define init_14_6_1_6
void FASTCALL call_14_6_1_6(SH2_struct* sh)
{
  SH2movi(sh, 6, (1<<4)+6);
}
#define init_14_6_1_7
void FASTCALL call_14_6_1_7(SH2_struct* sh)
{
  SH2movi(sh, 6, (1<<4)+7);
}
#define init_14_6_1_8
void FASTCALL call_14_6_1_8(SH2_struct* sh)
{
  SH2movi(sh, 6, (1<<4)+8);
}
#define init_14_6_1_9
void FASTCALL call_14_6_1_9(SH2_struct* sh)
{
  SH2movi(sh, 6, (1<<4)+9);
}
#define init_14_6_1_10
void FASTCALL call_14_6_1_10(SH2_struct* sh)
{
  SH2movi(sh, 6, (1<<4)+10);
}
#define init_14_6_1_11
void FASTCALL call_14_6_1_11(SH2_struct* sh)
{
  SH2movi(sh, 6, (1<<4)+11);
}
#define init_14_6_1_12
void FASTCALL call_14_6_1_12(SH2_struct* sh)
{
  SH2movi(sh, 6, (1<<4)+12);
}
#define init_14_6_1_13
void FASTCALL call_14_6_1_13(SH2_struct* sh)
{
  SH2movi(sh, 6, (1<<4)+13);
}
#define init_14_6_1_14
void FASTCALL call_14_6_1_14(SH2_struct* sh)
{
  SH2movi(sh, 6, (1<<4)+14);
}
#define init_14_6_1_15
void FASTCALL call_14_6_1_15(SH2_struct* sh)
{
  SH2movi(sh, 6, (1<<4)+15);
}
#define init_14_6_2_0
void FASTCALL call_14_6_2_0(SH2_struct* sh)
{
  SH2movi(sh, 6, (2<<4)+0);
}
#define init_14_6_2_1
void FASTCALL call_14_6_2_1(SH2_struct* sh)
{
  SH2movi(sh, 6, (2<<4)+1);
}
#define init_14_6_2_2
void FASTCALL call_14_6_2_2(SH2_struct* sh)
{
  SH2movi(sh, 6, (2<<4)+2);
}
#define init_14_6_2_3
void FASTCALL call_14_6_2_3(SH2_struct* sh)
{
  SH2movi(sh, 6, (2<<4)+3);
}
#define init_14_6_2_4
void FASTCALL call_14_6_2_4(SH2_struct* sh)
{
  SH2movi(sh, 6, (2<<4)+4);
}
#define init_14_6_2_5
void FASTCALL call_14_6_2_5(SH2_struct* sh)
{
  SH2movi(sh, 6, (2<<4)+5);
}
#define init_14_6_2_6
void FASTCALL call_14_6_2_6(SH2_struct* sh)
{
  SH2movi(sh, 6, (2<<4)+6);
}
#define init_14_6_2_7
void FASTCALL call_14_6_2_7(SH2_struct* sh)
{
  SH2movi(sh, 6, (2<<4)+7);
}
#define init_14_6_2_8
void FASTCALL call_14_6_2_8(SH2_struct* sh)
{
  SH2movi(sh, 6, (2<<4)+8);
}
#define init_14_6_2_9
void FASTCALL call_14_6_2_9(SH2_struct* sh)
{
  SH2movi(sh, 6, (2<<4)+9);
}
#define init_14_6_2_10
void FASTCALL call_14_6_2_10(SH2_struct* sh)
{
  SH2movi(sh, 6, (2<<4)+10);
}
#define init_14_6_2_11
void FASTCALL call_14_6_2_11(SH2_struct* sh)
{
  SH2movi(sh, 6, (2<<4)+11);
}
#define init_14_6_2_12
void FASTCALL call_14_6_2_12(SH2_struct* sh)
{
  SH2movi(sh, 6, (2<<4)+12);
}
#define init_14_6_2_13
void FASTCALL call_14_6_2_13(SH2_struct* sh)
{
  SH2movi(sh, 6, (2<<4)+13);
}
#define init_14_6_2_14
void FASTCALL call_14_6_2_14(SH2_struct* sh)
{
  SH2movi(sh, 6, (2<<4)+14);
}
#define init_14_6_2_15
void FASTCALL call_14_6_2_15(SH2_struct* sh)
{
  SH2movi(sh, 6, (2<<4)+15);
}
#define init_14_6_3_0
void FASTCALL call_14_6_3_0(SH2_struct* sh)
{
  SH2movi(sh, 6, (3<<4)+0);
}
#define init_14_6_3_1
void FASTCALL call_14_6_3_1(SH2_struct* sh)
{
  SH2movi(sh, 6, (3<<4)+1);
}
#define init_14_6_3_2
void FASTCALL call_14_6_3_2(SH2_struct* sh)
{
  SH2movi(sh, 6, (3<<4)+2);
}
#define init_14_6_3_3
void FASTCALL call_14_6_3_3(SH2_struct* sh)
{
  SH2movi(sh, 6, (3<<4)+3);
}
#define init_14_6_3_4
void FASTCALL call_14_6_3_4(SH2_struct* sh)
{
  SH2movi(sh, 6, (3<<4)+4);
}
#define init_14_6_3_5
void FASTCALL call_14_6_3_5(SH2_struct* sh)
{
  SH2movi(sh, 6, (3<<4)+5);
}
#define init_14_6_3_6
void FASTCALL call_14_6_3_6(SH2_struct* sh)
{
  SH2movi(sh, 6, (3<<4)+6);
}
#define init_14_6_3_7
void FASTCALL call_14_6_3_7(SH2_struct* sh)
{
  SH2movi(sh, 6, (3<<4)+7);
}
#define init_14_6_3_8
void FASTCALL call_14_6_3_8(SH2_struct* sh)
{
  SH2movi(sh, 6, (3<<4)+8);
}
#define init_14_6_3_9
void FASTCALL call_14_6_3_9(SH2_struct* sh)
{
  SH2movi(sh, 6, (3<<4)+9);
}
#define init_14_6_3_10
void FASTCALL call_14_6_3_10(SH2_struct* sh)
{
  SH2movi(sh, 6, (3<<4)+10);
}
#define init_14_6_3_11
void FASTCALL call_14_6_3_11(SH2_struct* sh)
{
  SH2movi(sh, 6, (3<<4)+11);
}
#define init_14_6_3_12
void FASTCALL call_14_6_3_12(SH2_struct* sh)
{
  SH2movi(sh, 6, (3<<4)+12);
}
#define init_14_6_3_13
void FASTCALL call_14_6_3_13(SH2_struct* sh)
{
  SH2movi(sh, 6, (3<<4)+13);
}
#define init_14_6_3_14
void FASTCALL call_14_6_3_14(SH2_struct* sh)
{
  SH2movi(sh, 6, (3<<4)+14);
}
#define init_14_6_3_15
void FASTCALL call_14_6_3_15(SH2_struct* sh)
{
  SH2movi(sh, 6, (3<<4)+15);
}
#define init_14_6_4_0
void FASTCALL call_14_6_4_0(SH2_struct* sh)
{
  SH2movi(sh, 6, (4<<4)+0);
}
#define init_14_6_4_1
void FASTCALL call_14_6_4_1(SH2_struct* sh)
{
  SH2movi(sh, 6, (4<<4)+1);
}
#define init_14_6_4_2
void FASTCALL call_14_6_4_2(SH2_struct* sh)
{
  SH2movi(sh, 6, (4<<4)+2);
}
#define init_14_6_4_3
void FASTCALL call_14_6_4_3(SH2_struct* sh)
{
  SH2movi(sh, 6, (4<<4)+3);
}
#define init_14_6_4_4
void FASTCALL call_14_6_4_4(SH2_struct* sh)
{
  SH2movi(sh, 6, (4<<4)+4);
}
#define init_14_6_4_5
void FASTCALL call_14_6_4_5(SH2_struct* sh)
{
  SH2movi(sh, 6, (4<<4)+5);
}
#define init_14_6_4_6
void FASTCALL call_14_6_4_6(SH2_struct* sh)
{
  SH2movi(sh, 6, (4<<4)+6);
}
#define init_14_6_4_7
void FASTCALL call_14_6_4_7(SH2_struct* sh)
{
  SH2movi(sh, 6, (4<<4)+7);
}
#define init_14_6_4_8
void FASTCALL call_14_6_4_8(SH2_struct* sh)
{
  SH2movi(sh, 6, (4<<4)+8);
}
#define init_14_6_4_9
void FASTCALL call_14_6_4_9(SH2_struct* sh)
{
  SH2movi(sh, 6, (4<<4)+9);
}
#define init_14_6_4_10
void FASTCALL call_14_6_4_10(SH2_struct* sh)
{
  SH2movi(sh, 6, (4<<4)+10);
}
#define init_14_6_4_11
void FASTCALL call_14_6_4_11(SH2_struct* sh)
{
  SH2movi(sh, 6, (4<<4)+11);
}
#define init_14_6_4_12
void FASTCALL call_14_6_4_12(SH2_struct* sh)
{
  SH2movi(sh, 6, (4<<4)+12);
}
#define init_14_6_4_13
void FASTCALL call_14_6_4_13(SH2_struct* sh)
{
  SH2movi(sh, 6, (4<<4)+13);
}
#define init_14_6_4_14
void FASTCALL call_14_6_4_14(SH2_struct* sh)
{
  SH2movi(sh, 6, (4<<4)+14);
}
#define init_14_6_4_15
void FASTCALL call_14_6_4_15(SH2_struct* sh)
{
  SH2movi(sh, 6, (4<<4)+15);
}
#define init_14_6_5_0
void FASTCALL call_14_6_5_0(SH2_struct* sh)
{
  SH2movi(sh, 6, (5<<4)+0);
}
#define init_14_6_5_1
void FASTCALL call_14_6_5_1(SH2_struct* sh)
{
  SH2movi(sh, 6, (5<<4)+1);
}
#define init_14_6_5_2
void FASTCALL call_14_6_5_2(SH2_struct* sh)
{
  SH2movi(sh, 6, (5<<4)+2);
}
#define init_14_6_5_3
void FASTCALL call_14_6_5_3(SH2_struct* sh)
{
  SH2movi(sh, 6, (5<<4)+3);
}
#define init_14_6_5_4
void FASTCALL call_14_6_5_4(SH2_struct* sh)
{
  SH2movi(sh, 6, (5<<4)+4);
}
#define init_14_6_5_5
void FASTCALL call_14_6_5_5(SH2_struct* sh)
{
  SH2movi(sh, 6, (5<<4)+5);
}
#define init_14_6_5_6
void FASTCALL call_14_6_5_6(SH2_struct* sh)
{
  SH2movi(sh, 6, (5<<4)+6);
}
#define init_14_6_5_7
void FASTCALL call_14_6_5_7(SH2_struct* sh)
{
  SH2movi(sh, 6, (5<<4)+7);
}
#define init_14_6_5_8
void FASTCALL call_14_6_5_8(SH2_struct* sh)
{
  SH2movi(sh, 6, (5<<4)+8);
}
#define init_14_6_5_9
void FASTCALL call_14_6_5_9(SH2_struct* sh)
{
  SH2movi(sh, 6, (5<<4)+9);
}
#define init_14_6_5_10
void FASTCALL call_14_6_5_10(SH2_struct* sh)
{
  SH2movi(sh, 6, (5<<4)+10);
}
#define init_14_6_5_11
void FASTCALL call_14_6_5_11(SH2_struct* sh)
{
  SH2movi(sh, 6, (5<<4)+11);
}
#define init_14_6_5_12
void FASTCALL call_14_6_5_12(SH2_struct* sh)
{
  SH2movi(sh, 6, (5<<4)+12);
}
#define init_14_6_5_13
void FASTCALL call_14_6_5_13(SH2_struct* sh)
{
  SH2movi(sh, 6, (5<<4)+13);
}
#define init_14_6_5_14
void FASTCALL call_14_6_5_14(SH2_struct* sh)
{
  SH2movi(sh, 6, (5<<4)+14);
}
#define init_14_6_5_15
void FASTCALL call_14_6_5_15(SH2_struct* sh)
{
  SH2movi(sh, 6, (5<<4)+15);
}
#define init_14_6_6_0
void FASTCALL call_14_6_6_0(SH2_struct* sh)
{
  SH2movi(sh, 6, (6<<4)+0);
}
#define init_14_6_6_1
void FASTCALL call_14_6_6_1(SH2_struct* sh)
{
  SH2movi(sh, 6, (6<<4)+1);
}
#define init_14_6_6_2
void FASTCALL call_14_6_6_2(SH2_struct* sh)
{
  SH2movi(sh, 6, (6<<4)+2);
}
#define init_14_6_6_3
void FASTCALL call_14_6_6_3(SH2_struct* sh)
{
  SH2movi(sh, 6, (6<<4)+3);
}
#define init_14_6_6_4
void FASTCALL call_14_6_6_4(SH2_struct* sh)
{
  SH2movi(sh, 6, (6<<4)+4);
}
#define init_14_6_6_5
void FASTCALL call_14_6_6_5(SH2_struct* sh)
{
  SH2movi(sh, 6, (6<<4)+5);
}
#define init_14_6_6_6
void FASTCALL call_14_6_6_6(SH2_struct* sh)
{
  SH2movi(sh, 6, (6<<4)+6);
}
#define init_14_6_6_7
void FASTCALL call_14_6_6_7(SH2_struct* sh)
{
  SH2movi(sh, 6, (6<<4)+7);
}
#define init_14_6_6_8
void FASTCALL call_14_6_6_8(SH2_struct* sh)
{
  SH2movi(sh, 6, (6<<4)+8);
}
#define init_14_6_6_9
void FASTCALL call_14_6_6_9(SH2_struct* sh)
{
  SH2movi(sh, 6, (6<<4)+9);
}
#define init_14_6_6_10
void FASTCALL call_14_6_6_10(SH2_struct* sh)
{
  SH2movi(sh, 6, (6<<4)+10);
}
#define init_14_6_6_11
void FASTCALL call_14_6_6_11(SH2_struct* sh)
{
  SH2movi(sh, 6, (6<<4)+11);
}
#define init_14_6_6_12
void FASTCALL call_14_6_6_12(SH2_struct* sh)
{
  SH2movi(sh, 6, (6<<4)+12);
}
#define init_14_6_6_13
void FASTCALL call_14_6_6_13(SH2_struct* sh)
{
  SH2movi(sh, 6, (6<<4)+13);
}
#define init_14_6_6_14
void FASTCALL call_14_6_6_14(SH2_struct* sh)
{
  SH2movi(sh, 6, (6<<4)+14);
}
#define init_14_6_6_15
void FASTCALL call_14_6_6_15(SH2_struct* sh)
{
  SH2movi(sh, 6, (6<<4)+15);
}
#define init_14_6_7_0
void FASTCALL call_14_6_7_0(SH2_struct* sh)
{
  SH2movi(sh, 6, (7<<4)+0);
}
#define init_14_6_7_1
void FASTCALL call_14_6_7_1(SH2_struct* sh)
{
  SH2movi(sh, 6, (7<<4)+1);
}
#define init_14_6_7_2
void FASTCALL call_14_6_7_2(SH2_struct* sh)
{
  SH2movi(sh, 6, (7<<4)+2);
}
#define init_14_6_7_3
void FASTCALL call_14_6_7_3(SH2_struct* sh)
{
  SH2movi(sh, 6, (7<<4)+3);
}
#define init_14_6_7_4
void FASTCALL call_14_6_7_4(SH2_struct* sh)
{
  SH2movi(sh, 6, (7<<4)+4);
}
#define init_14_6_7_5
void FASTCALL call_14_6_7_5(SH2_struct* sh)
{
  SH2movi(sh, 6, (7<<4)+5);
}
#define init_14_6_7_6
void FASTCALL call_14_6_7_6(SH2_struct* sh)
{
  SH2movi(sh, 6, (7<<4)+6);
}
#define init_14_6_7_7
void FASTCALL call_14_6_7_7(SH2_struct* sh)
{
  SH2movi(sh, 6, (7<<4)+7);
}
#define init_14_6_7_8
void FASTCALL call_14_6_7_8(SH2_struct* sh)
{
  SH2movi(sh, 6, (7<<4)+8);
}
#define init_14_6_7_9
void FASTCALL call_14_6_7_9(SH2_struct* sh)
{
  SH2movi(sh, 6, (7<<4)+9);
}
#define init_14_6_7_10
void FASTCALL call_14_6_7_10(SH2_struct* sh)
{
  SH2movi(sh, 6, (7<<4)+10);
}
#define init_14_6_7_11
void FASTCALL call_14_6_7_11(SH2_struct* sh)
{
  SH2movi(sh, 6, (7<<4)+11);
}
#define init_14_6_7_12
void FASTCALL call_14_6_7_12(SH2_struct* sh)
{
  SH2movi(sh, 6, (7<<4)+12);
}
#define init_14_6_7_13
void FASTCALL call_14_6_7_13(SH2_struct* sh)
{
  SH2movi(sh, 6, (7<<4)+13);
}
#define init_14_6_7_14
void FASTCALL call_14_6_7_14(SH2_struct* sh)
{
  SH2movi(sh, 6, (7<<4)+14);
}
#define init_14_6_7_15
void FASTCALL call_14_6_7_15(SH2_struct* sh)
{
  SH2movi(sh, 6, (7<<4)+15);
}
#define init_14_6_8_0
void FASTCALL call_14_6_8_0(SH2_struct* sh)
{
  SH2movi(sh, 6, (8<<4)+0);
}
#define init_14_6_8_1
void FASTCALL call_14_6_8_1(SH2_struct* sh)
{
  SH2movi(sh, 6, (8<<4)+1);
}
#define init_14_6_8_2
void FASTCALL call_14_6_8_2(SH2_struct* sh)
{
  SH2movi(sh, 6, (8<<4)+2);
}
#define init_14_6_8_3
void FASTCALL call_14_6_8_3(SH2_struct* sh)
{
  SH2movi(sh, 6, (8<<4)+3);
}
#define init_14_6_8_4
void FASTCALL call_14_6_8_4(SH2_struct* sh)
{
  SH2movi(sh, 6, (8<<4)+4);
}
#define init_14_6_8_5
void FASTCALL call_14_6_8_5(SH2_struct* sh)
{
  SH2movi(sh, 6, (8<<4)+5);
}
#define init_14_6_8_6
void FASTCALL call_14_6_8_6(SH2_struct* sh)
{
  SH2movi(sh, 6, (8<<4)+6);
}
#define init_14_6_8_7
void FASTCALL call_14_6_8_7(SH2_struct* sh)
{
  SH2movi(sh, 6, (8<<4)+7);
}
#define init_14_6_8_8
void FASTCALL call_14_6_8_8(SH2_struct* sh)
{
  SH2movi(sh, 6, (8<<4)+8);
}
#define init_14_6_8_9
void FASTCALL call_14_6_8_9(SH2_struct* sh)
{
  SH2movi(sh, 6, (8<<4)+9);
}
#define init_14_6_8_10
void FASTCALL call_14_6_8_10(SH2_struct* sh)
{
  SH2movi(sh, 6, (8<<4)+10);
}
#define init_14_6_8_11
void FASTCALL call_14_6_8_11(SH2_struct* sh)
{
  SH2movi(sh, 6, (8<<4)+11);
}
#define init_14_6_8_12
void FASTCALL call_14_6_8_12(SH2_struct* sh)
{
  SH2movi(sh, 6, (8<<4)+12);
}
#define init_14_6_8_13
void FASTCALL call_14_6_8_13(SH2_struct* sh)
{
  SH2movi(sh, 6, (8<<4)+13);
}
#define init_14_6_8_14
void FASTCALL call_14_6_8_14(SH2_struct* sh)
{
  SH2movi(sh, 6, (8<<4)+14);
}
#define init_14_6_8_15
void FASTCALL call_14_6_8_15(SH2_struct* sh)
{
  SH2movi(sh, 6, (8<<4)+15);
}
#define init_14_6_9_0
void FASTCALL call_14_6_9_0(SH2_struct* sh)
{
  SH2movi(sh, 6, (9<<4)+0);
}
#define init_14_6_9_1
void FASTCALL call_14_6_9_1(SH2_struct* sh)
{
  SH2movi(sh, 6, (9<<4)+1);
}
#define init_14_6_9_2
void FASTCALL call_14_6_9_2(SH2_struct* sh)
{
  SH2movi(sh, 6, (9<<4)+2);
}
#define init_14_6_9_3
void FASTCALL call_14_6_9_3(SH2_struct* sh)
{
  SH2movi(sh, 6, (9<<4)+3);
}
#define init_14_6_9_4
void FASTCALL call_14_6_9_4(SH2_struct* sh)
{
  SH2movi(sh, 6, (9<<4)+4);
}
#define init_14_6_9_5
void FASTCALL call_14_6_9_5(SH2_struct* sh)
{
  SH2movi(sh, 6, (9<<4)+5);
}
#define init_14_6_9_6
void FASTCALL call_14_6_9_6(SH2_struct* sh)
{
  SH2movi(sh, 6, (9<<4)+6);
}
#define init_14_6_9_7
void FASTCALL call_14_6_9_7(SH2_struct* sh)
{
  SH2movi(sh, 6, (9<<4)+7);
}
#define init_14_6_9_8
void FASTCALL call_14_6_9_8(SH2_struct* sh)
{
  SH2movi(sh, 6, (9<<4)+8);
}
#define init_14_6_9_9
void FASTCALL call_14_6_9_9(SH2_struct* sh)
{
  SH2movi(sh, 6, (9<<4)+9);
}
#define init_14_6_9_10
void FASTCALL call_14_6_9_10(SH2_struct* sh)
{
  SH2movi(sh, 6, (9<<4)+10);
}
#define init_14_6_9_11
void FASTCALL call_14_6_9_11(SH2_struct* sh)
{
  SH2movi(sh, 6, (9<<4)+11);
}
#define init_14_6_9_12
void FASTCALL call_14_6_9_12(SH2_struct* sh)
{
  SH2movi(sh, 6, (9<<4)+12);
}
#define init_14_6_9_13
void FASTCALL call_14_6_9_13(SH2_struct* sh)
{
  SH2movi(sh, 6, (9<<4)+13);
}
#define init_14_6_9_14
void FASTCALL call_14_6_9_14(SH2_struct* sh)
{
  SH2movi(sh, 6, (9<<4)+14);
}
#define init_14_6_9_15
void FASTCALL call_14_6_9_15(SH2_struct* sh)
{
  SH2movi(sh, 6, (9<<4)+15);
}
#define init_14_6_10_0
void FASTCALL call_14_6_10_0(SH2_struct* sh)
{
  SH2movi(sh, 6, (10<<4)+0);
}
#define init_14_6_10_1
void FASTCALL call_14_6_10_1(SH2_struct* sh)
{
  SH2movi(sh, 6, (10<<4)+1);
}
#define init_14_6_10_2
void FASTCALL call_14_6_10_2(SH2_struct* sh)
{
  SH2movi(sh, 6, (10<<4)+2);
}
#define init_14_6_10_3
void FASTCALL call_14_6_10_3(SH2_struct* sh)
{
  SH2movi(sh, 6, (10<<4)+3);
}
#define init_14_6_10_4
void FASTCALL call_14_6_10_4(SH2_struct* sh)
{
  SH2movi(sh, 6, (10<<4)+4);
}
#define init_14_6_10_5
void FASTCALL call_14_6_10_5(SH2_struct* sh)
{
  SH2movi(sh, 6, (10<<4)+5);
}
#define init_14_6_10_6
void FASTCALL call_14_6_10_6(SH2_struct* sh)
{
  SH2movi(sh, 6, (10<<4)+6);
}
#define init_14_6_10_7
void FASTCALL call_14_6_10_7(SH2_struct* sh)
{
  SH2movi(sh, 6, (10<<4)+7);
}
#define init_14_6_10_8
void FASTCALL call_14_6_10_8(SH2_struct* sh)
{
  SH2movi(sh, 6, (10<<4)+8);
}
#define init_14_6_10_9
void FASTCALL call_14_6_10_9(SH2_struct* sh)
{
  SH2movi(sh, 6, (10<<4)+9);
}
#define init_14_6_10_10
void FASTCALL call_14_6_10_10(SH2_struct* sh)
{
  SH2movi(sh, 6, (10<<4)+10);
}
#define init_14_6_10_11
void FASTCALL call_14_6_10_11(SH2_struct* sh)
{
  SH2movi(sh, 6, (10<<4)+11);
}
#define init_14_6_10_12
void FASTCALL call_14_6_10_12(SH2_struct* sh)
{
  SH2movi(sh, 6, (10<<4)+12);
}
#define init_14_6_10_13
void FASTCALL call_14_6_10_13(SH2_struct* sh)
{
  SH2movi(sh, 6, (10<<4)+13);
}
#define init_14_6_10_14
void FASTCALL call_14_6_10_14(SH2_struct* sh)
{
  SH2movi(sh, 6, (10<<4)+14);
}
#define init_14_6_10_15
void FASTCALL call_14_6_10_15(SH2_struct* sh)
{
  SH2movi(sh, 6, (10<<4)+15);
}
#define init_14_6_11_0
void FASTCALL call_14_6_11_0(SH2_struct* sh)
{
  SH2movi(sh, 6, (11<<4)+0);
}
#define init_14_6_11_1
void FASTCALL call_14_6_11_1(SH2_struct* sh)
{
  SH2movi(sh, 6, (11<<4)+1);
}
#define init_14_6_11_2
void FASTCALL call_14_6_11_2(SH2_struct* sh)
{
  SH2movi(sh, 6, (11<<4)+2);
}
#define init_14_6_11_3
void FASTCALL call_14_6_11_3(SH2_struct* sh)
{
  SH2movi(sh, 6, (11<<4)+3);
}
#define init_14_6_11_4
void FASTCALL call_14_6_11_4(SH2_struct* sh)
{
  SH2movi(sh, 6, (11<<4)+4);
}
#define init_14_6_11_5
void FASTCALL call_14_6_11_5(SH2_struct* sh)
{
  SH2movi(sh, 6, (11<<4)+5);
}
#define init_14_6_11_6
void FASTCALL call_14_6_11_6(SH2_struct* sh)
{
  SH2movi(sh, 6, (11<<4)+6);
}
#define init_14_6_11_7
void FASTCALL call_14_6_11_7(SH2_struct* sh)
{
  SH2movi(sh, 6, (11<<4)+7);
}
#define init_14_6_11_8
void FASTCALL call_14_6_11_8(SH2_struct* sh)
{
  SH2movi(sh, 6, (11<<4)+8);
}
#define init_14_6_11_9
void FASTCALL call_14_6_11_9(SH2_struct* sh)
{
  SH2movi(sh, 6, (11<<4)+9);
}
#define init_14_6_11_10
void FASTCALL call_14_6_11_10(SH2_struct* sh)
{
  SH2movi(sh, 6, (11<<4)+10);
}
#define init_14_6_11_11
void FASTCALL call_14_6_11_11(SH2_struct* sh)
{
  SH2movi(sh, 6, (11<<4)+11);
}
#define init_14_6_11_12
void FASTCALL call_14_6_11_12(SH2_struct* sh)
{
  SH2movi(sh, 6, (11<<4)+12);
}
#define init_14_6_11_13
void FASTCALL call_14_6_11_13(SH2_struct* sh)
{
  SH2movi(sh, 6, (11<<4)+13);
}
#define init_14_6_11_14
void FASTCALL call_14_6_11_14(SH2_struct* sh)
{
  SH2movi(sh, 6, (11<<4)+14);
}
#define init_14_6_11_15
void FASTCALL call_14_6_11_15(SH2_struct* sh)
{
  SH2movi(sh, 6, (11<<4)+15);
}
#define init_14_6_12_0
void FASTCALL call_14_6_12_0(SH2_struct* sh)
{
  SH2movi(sh, 6, (12<<4)+0);
}
#define init_14_6_12_1
void FASTCALL call_14_6_12_1(SH2_struct* sh)
{
  SH2movi(sh, 6, (12<<4)+1);
}
#define init_14_6_12_2
void FASTCALL call_14_6_12_2(SH2_struct* sh)
{
  SH2movi(sh, 6, (12<<4)+2);
}
#define init_14_6_12_3
void FASTCALL call_14_6_12_3(SH2_struct* sh)
{
  SH2movi(sh, 6, (12<<4)+3);
}
#define init_14_6_12_4
void FASTCALL call_14_6_12_4(SH2_struct* sh)
{
  SH2movi(sh, 6, (12<<4)+4);
}
#define init_14_6_12_5
void FASTCALL call_14_6_12_5(SH2_struct* sh)
{
  SH2movi(sh, 6, (12<<4)+5);
}
#define init_14_6_12_6
void FASTCALL call_14_6_12_6(SH2_struct* sh)
{
  SH2movi(sh, 6, (12<<4)+6);
}
#define init_14_6_12_7
void FASTCALL call_14_6_12_7(SH2_struct* sh)
{
  SH2movi(sh, 6, (12<<4)+7);
}
#define init_14_6_12_8
void FASTCALL call_14_6_12_8(SH2_struct* sh)
{
  SH2movi(sh, 6, (12<<4)+8);
}
#define init_14_6_12_9
void FASTCALL call_14_6_12_9(SH2_struct* sh)
{
  SH2movi(sh, 6, (12<<4)+9);
}
#define init_14_6_12_10
void FASTCALL call_14_6_12_10(SH2_struct* sh)
{
  SH2movi(sh, 6, (12<<4)+10);
}
#define init_14_6_12_11
void FASTCALL call_14_6_12_11(SH2_struct* sh)
{
  SH2movi(sh, 6, (12<<4)+11);
}
#define init_14_6_12_12
void FASTCALL call_14_6_12_12(SH2_struct* sh)
{
  SH2movi(sh, 6, (12<<4)+12);
}
#define init_14_6_12_13
void FASTCALL call_14_6_12_13(SH2_struct* sh)
{
  SH2movi(sh, 6, (12<<4)+13);
}
#define init_14_6_12_14
void FASTCALL call_14_6_12_14(SH2_struct* sh)
{
  SH2movi(sh, 6, (12<<4)+14);
}
#define init_14_6_12_15
void FASTCALL call_14_6_12_15(SH2_struct* sh)
{
  SH2movi(sh, 6, (12<<4)+15);
}
#define init_14_6_13_0
void FASTCALL call_14_6_13_0(SH2_struct* sh)
{
  SH2movi(sh, 6, (13<<4)+0);
}
#define init_14_6_13_1
void FASTCALL call_14_6_13_1(SH2_struct* sh)
{
  SH2movi(sh, 6, (13<<4)+1);
}
#define init_14_6_13_2
void FASTCALL call_14_6_13_2(SH2_struct* sh)
{
  SH2movi(sh, 6, (13<<4)+2);
}
#define init_14_6_13_3
void FASTCALL call_14_6_13_3(SH2_struct* sh)
{
  SH2movi(sh, 6, (13<<4)+3);
}
#define init_14_6_13_4
void FASTCALL call_14_6_13_4(SH2_struct* sh)
{
  SH2movi(sh, 6, (13<<4)+4);
}
#define init_14_6_13_5
void FASTCALL call_14_6_13_5(SH2_struct* sh)
{
  SH2movi(sh, 6, (13<<4)+5);
}
#define init_14_6_13_6
void FASTCALL call_14_6_13_6(SH2_struct* sh)
{
  SH2movi(sh, 6, (13<<4)+6);
}
#define init_14_6_13_7
void FASTCALL call_14_6_13_7(SH2_struct* sh)
{
  SH2movi(sh, 6, (13<<4)+7);
}
#define init_14_6_13_8
void FASTCALL call_14_6_13_8(SH2_struct* sh)
{
  SH2movi(sh, 6, (13<<4)+8);
}
#define init_14_6_13_9
void FASTCALL call_14_6_13_9(SH2_struct* sh)
{
  SH2movi(sh, 6, (13<<4)+9);
}
#define init_14_6_13_10
void FASTCALL call_14_6_13_10(SH2_struct* sh)
{
  SH2movi(sh, 6, (13<<4)+10);
}
#define init_14_6_13_11
void FASTCALL call_14_6_13_11(SH2_struct* sh)
{
  SH2movi(sh, 6, (13<<4)+11);
}
#define init_14_6_13_12
void FASTCALL call_14_6_13_12(SH2_struct* sh)
{
  SH2movi(sh, 6, (13<<4)+12);
}
#define init_14_6_13_13
void FASTCALL call_14_6_13_13(SH2_struct* sh)
{
  SH2movi(sh, 6, (13<<4)+13);
}
#define init_14_6_13_14
void FASTCALL call_14_6_13_14(SH2_struct* sh)
{
  SH2movi(sh, 6, (13<<4)+14);
}
#define init_14_6_13_15
void FASTCALL call_14_6_13_15(SH2_struct* sh)
{
  SH2movi(sh, 6, (13<<4)+15);
}
#define init_14_6_14_0
void FASTCALL call_14_6_14_0(SH2_struct* sh)
{
  SH2movi(sh, 6, (14<<4)+0);
}
#define init_14_6_14_1
void FASTCALL call_14_6_14_1(SH2_struct* sh)
{
  SH2movi(sh, 6, (14<<4)+1);
}
#define init_14_6_14_2
void FASTCALL call_14_6_14_2(SH2_struct* sh)
{
  SH2movi(sh, 6, (14<<4)+2);
}
#define init_14_6_14_3
void FASTCALL call_14_6_14_3(SH2_struct* sh)
{
  SH2movi(sh, 6, (14<<4)+3);
}
#define init_14_6_14_4
void FASTCALL call_14_6_14_4(SH2_struct* sh)
{
  SH2movi(sh, 6, (14<<4)+4);
}
#define init_14_6_14_5
void FASTCALL call_14_6_14_5(SH2_struct* sh)
{
  SH2movi(sh, 6, (14<<4)+5);
}
#define init_14_6_14_6
void FASTCALL call_14_6_14_6(SH2_struct* sh)
{
  SH2movi(sh, 6, (14<<4)+6);
}
#define init_14_6_14_7
void FASTCALL call_14_6_14_7(SH2_struct* sh)
{
  SH2movi(sh, 6, (14<<4)+7);
}
#define init_14_6_14_8
void FASTCALL call_14_6_14_8(SH2_struct* sh)
{
  SH2movi(sh, 6, (14<<4)+8);
}
#define init_14_6_14_9
void FASTCALL call_14_6_14_9(SH2_struct* sh)
{
  SH2movi(sh, 6, (14<<4)+9);
}
#define init_14_6_14_10
void FASTCALL call_14_6_14_10(SH2_struct* sh)
{
  SH2movi(sh, 6, (14<<4)+10);
}
#define init_14_6_14_11
void FASTCALL call_14_6_14_11(SH2_struct* sh)
{
  SH2movi(sh, 6, (14<<4)+11);
}
#define init_14_6_14_12
void FASTCALL call_14_6_14_12(SH2_struct* sh)
{
  SH2movi(sh, 6, (14<<4)+12);
}
#define init_14_6_14_13
void FASTCALL call_14_6_14_13(SH2_struct* sh)
{
  SH2movi(sh, 6, (14<<4)+13);
}
#define init_14_6_14_14
void FASTCALL call_14_6_14_14(SH2_struct* sh)
{
  SH2movi(sh, 6, (14<<4)+14);
}
#define init_14_6_14_15
void FASTCALL call_14_6_14_15(SH2_struct* sh)
{
  SH2movi(sh, 6, (14<<4)+15);
}
#define init_14_6_15_0
void FASTCALL call_14_6_15_0(SH2_struct* sh)
{
  SH2movi(sh, 6, (15<<4)+0);
}
#define init_14_6_15_1
void FASTCALL call_14_6_15_1(SH2_struct* sh)
{
  SH2movi(sh, 6, (15<<4)+1);
}
#define init_14_6_15_2
void FASTCALL call_14_6_15_2(SH2_struct* sh)
{
  SH2movi(sh, 6, (15<<4)+2);
}
#define init_14_6_15_3
void FASTCALL call_14_6_15_3(SH2_struct* sh)
{
  SH2movi(sh, 6, (15<<4)+3);
}
#define init_14_6_15_4
void FASTCALL call_14_6_15_4(SH2_struct* sh)
{
  SH2movi(sh, 6, (15<<4)+4);
}
#define init_14_6_15_5
void FASTCALL call_14_6_15_5(SH2_struct* sh)
{
  SH2movi(sh, 6, (15<<4)+5);
}
#define init_14_6_15_6
void FASTCALL call_14_6_15_6(SH2_struct* sh)
{
  SH2movi(sh, 6, (15<<4)+6);
}
#define init_14_6_15_7
void FASTCALL call_14_6_15_7(SH2_struct* sh)
{
  SH2movi(sh, 6, (15<<4)+7);
}
#define init_14_6_15_8
void FASTCALL call_14_6_15_8(SH2_struct* sh)
{
  SH2movi(sh, 6, (15<<4)+8);
}
#define init_14_6_15_9
void FASTCALL call_14_6_15_9(SH2_struct* sh)
{
  SH2movi(sh, 6, (15<<4)+9);
}
#define init_14_6_15_10
void FASTCALL call_14_6_15_10(SH2_struct* sh)
{
  SH2movi(sh, 6, (15<<4)+10);
}
#define init_14_6_15_11
void FASTCALL call_14_6_15_11(SH2_struct* sh)
{
  SH2movi(sh, 6, (15<<4)+11);
}
#define init_14_6_15_12
void FASTCALL call_14_6_15_12(SH2_struct* sh)
{
  SH2movi(sh, 6, (15<<4)+12);
}
#define init_14_6_15_13
void FASTCALL call_14_6_15_13(SH2_struct* sh)
{
  SH2movi(sh, 6, (15<<4)+13);
}
#define init_14_6_15_14
void FASTCALL call_14_6_15_14(SH2_struct* sh)
{
  SH2movi(sh, 6, (15<<4)+14);
}
#define init_14_6_15_15
void FASTCALL call_14_6_15_15(SH2_struct* sh)
{
  SH2movi(sh, 6, (15<<4)+15);
}
#define init_14_7_0_0
void FASTCALL call_14_7_0_0(SH2_struct* sh)
{
  SH2movi(sh, 7, (0<<4)+0);
}
#define init_14_7_0_1
void FASTCALL call_14_7_0_1(SH2_struct* sh)
{
  SH2movi(sh, 7, (0<<4)+1);
}
#define init_14_7_0_2
void FASTCALL call_14_7_0_2(SH2_struct* sh)
{
  SH2movi(sh, 7, (0<<4)+2);
}
#define init_14_7_0_3
void FASTCALL call_14_7_0_3(SH2_struct* sh)
{
  SH2movi(sh, 7, (0<<4)+3);
}
#define init_14_7_0_4
void FASTCALL call_14_7_0_4(SH2_struct* sh)
{
  SH2movi(sh, 7, (0<<4)+4);
}
#define init_14_7_0_5
void FASTCALL call_14_7_0_5(SH2_struct* sh)
{
  SH2movi(sh, 7, (0<<4)+5);
}
#define init_14_7_0_6
void FASTCALL call_14_7_0_6(SH2_struct* sh)
{
  SH2movi(sh, 7, (0<<4)+6);
}
#define init_14_7_0_7
void FASTCALL call_14_7_0_7(SH2_struct* sh)
{
  SH2movi(sh, 7, (0<<4)+7);
}
#define init_14_7_0_8
void FASTCALL call_14_7_0_8(SH2_struct* sh)
{
  SH2movi(sh, 7, (0<<4)+8);
}
#define init_14_7_0_9
void FASTCALL call_14_7_0_9(SH2_struct* sh)
{
  SH2movi(sh, 7, (0<<4)+9);
}
#define init_14_7_0_10
void FASTCALL call_14_7_0_10(SH2_struct* sh)
{
  SH2movi(sh, 7, (0<<4)+10);
}
#define init_14_7_0_11
void FASTCALL call_14_7_0_11(SH2_struct* sh)
{
  SH2movi(sh, 7, (0<<4)+11);
}
#define init_14_7_0_12
void FASTCALL call_14_7_0_12(SH2_struct* sh)
{
  SH2movi(sh, 7, (0<<4)+12);
}
#define init_14_7_0_13
void FASTCALL call_14_7_0_13(SH2_struct* sh)
{
  SH2movi(sh, 7, (0<<4)+13);
}
#define init_14_7_0_14
void FASTCALL call_14_7_0_14(SH2_struct* sh)
{
  SH2movi(sh, 7, (0<<4)+14);
}
#define init_14_7_0_15
void FASTCALL call_14_7_0_15(SH2_struct* sh)
{
  SH2movi(sh, 7, (0<<4)+15);
}
#define init_14_7_1_0
void FASTCALL call_14_7_1_0(SH2_struct* sh)
{
  SH2movi(sh, 7, (1<<4)+0);
}
#define init_14_7_1_1
void FASTCALL call_14_7_1_1(SH2_struct* sh)
{
  SH2movi(sh, 7, (1<<4)+1);
}
#define init_14_7_1_2
void FASTCALL call_14_7_1_2(SH2_struct* sh)
{
  SH2movi(sh, 7, (1<<4)+2);
}
#define init_14_7_1_3
void FASTCALL call_14_7_1_3(SH2_struct* sh)
{
  SH2movi(sh, 7, (1<<4)+3);
}
#define init_14_7_1_4
void FASTCALL call_14_7_1_4(SH2_struct* sh)
{
  SH2movi(sh, 7, (1<<4)+4);
}
#define init_14_7_1_5
void FASTCALL call_14_7_1_5(SH2_struct* sh)
{
  SH2movi(sh, 7, (1<<4)+5);
}
#define init_14_7_1_6
void FASTCALL call_14_7_1_6(SH2_struct* sh)
{
  SH2movi(sh, 7, (1<<4)+6);
}
#define init_14_7_1_7
void FASTCALL call_14_7_1_7(SH2_struct* sh)
{
  SH2movi(sh, 7, (1<<4)+7);
}
#define init_14_7_1_8
void FASTCALL call_14_7_1_8(SH2_struct* sh)
{
  SH2movi(sh, 7, (1<<4)+8);
}
#define init_14_7_1_9
void FASTCALL call_14_7_1_9(SH2_struct* sh)
{
  SH2movi(sh, 7, (1<<4)+9);
}
#define init_14_7_1_10
void FASTCALL call_14_7_1_10(SH2_struct* sh)
{
  SH2movi(sh, 7, (1<<4)+10);
}
#define init_14_7_1_11
void FASTCALL call_14_7_1_11(SH2_struct* sh)
{
  SH2movi(sh, 7, (1<<4)+11);
}
#define init_14_7_1_12
void FASTCALL call_14_7_1_12(SH2_struct* sh)
{
  SH2movi(sh, 7, (1<<4)+12);
}
#define init_14_7_1_13
void FASTCALL call_14_7_1_13(SH2_struct* sh)
{
  SH2movi(sh, 7, (1<<4)+13);
}
#define init_14_7_1_14
void FASTCALL call_14_7_1_14(SH2_struct* sh)
{
  SH2movi(sh, 7, (1<<4)+14);
}
#define init_14_7_1_15
void FASTCALL call_14_7_1_15(SH2_struct* sh)
{
  SH2movi(sh, 7, (1<<4)+15);
}
#define init_14_7_2_0
void FASTCALL call_14_7_2_0(SH2_struct* sh)
{
  SH2movi(sh, 7, (2<<4)+0);
}
#define init_14_7_2_1
void FASTCALL call_14_7_2_1(SH2_struct* sh)
{
  SH2movi(sh, 7, (2<<4)+1);
}
#define init_14_7_2_2
void FASTCALL call_14_7_2_2(SH2_struct* sh)
{
  SH2movi(sh, 7, (2<<4)+2);
}
#define init_14_7_2_3
void FASTCALL call_14_7_2_3(SH2_struct* sh)
{
  SH2movi(sh, 7, (2<<4)+3);
}
#define init_14_7_2_4
void FASTCALL call_14_7_2_4(SH2_struct* sh)
{
  SH2movi(sh, 7, (2<<4)+4);
}
#define init_14_7_2_5
void FASTCALL call_14_7_2_5(SH2_struct* sh)
{
  SH2movi(sh, 7, (2<<4)+5);
}
#define init_14_7_2_6
void FASTCALL call_14_7_2_6(SH2_struct* sh)
{
  SH2movi(sh, 7, (2<<4)+6);
}
#define init_14_7_2_7
void FASTCALL call_14_7_2_7(SH2_struct* sh)
{
  SH2movi(sh, 7, (2<<4)+7);
}
#define init_14_7_2_8
void FASTCALL call_14_7_2_8(SH2_struct* sh)
{
  SH2movi(sh, 7, (2<<4)+8);
}
#define init_14_7_2_9
void FASTCALL call_14_7_2_9(SH2_struct* sh)
{
  SH2movi(sh, 7, (2<<4)+9);
}
#define init_14_7_2_10
void FASTCALL call_14_7_2_10(SH2_struct* sh)
{
  SH2movi(sh, 7, (2<<4)+10);
}
#define init_14_7_2_11
void FASTCALL call_14_7_2_11(SH2_struct* sh)
{
  SH2movi(sh, 7, (2<<4)+11);
}
#define init_14_7_2_12
void FASTCALL call_14_7_2_12(SH2_struct* sh)
{
  SH2movi(sh, 7, (2<<4)+12);
}
#define init_14_7_2_13
void FASTCALL call_14_7_2_13(SH2_struct* sh)
{
  SH2movi(sh, 7, (2<<4)+13);
}
#define init_14_7_2_14
void FASTCALL call_14_7_2_14(SH2_struct* sh)
{
  SH2movi(sh, 7, (2<<4)+14);
}
#define init_14_7_2_15
void FASTCALL call_14_7_2_15(SH2_struct* sh)
{
  SH2movi(sh, 7, (2<<4)+15);
}
#define init_14_7_3_0
void FASTCALL call_14_7_3_0(SH2_struct* sh)
{
  SH2movi(sh, 7, (3<<4)+0);
}
#define init_14_7_3_1
void FASTCALL call_14_7_3_1(SH2_struct* sh)
{
  SH2movi(sh, 7, (3<<4)+1);
}
#define init_14_7_3_2
void FASTCALL call_14_7_3_2(SH2_struct* sh)
{
  SH2movi(sh, 7, (3<<4)+2);
}
#define init_14_7_3_3
void FASTCALL call_14_7_3_3(SH2_struct* sh)
{
  SH2movi(sh, 7, (3<<4)+3);
}
#define init_14_7_3_4
void FASTCALL call_14_7_3_4(SH2_struct* sh)
{
  SH2movi(sh, 7, (3<<4)+4);
}
#define init_14_7_3_5
void FASTCALL call_14_7_3_5(SH2_struct* sh)
{
  SH2movi(sh, 7, (3<<4)+5);
}
#define init_14_7_3_6
void FASTCALL call_14_7_3_6(SH2_struct* sh)
{
  SH2movi(sh, 7, (3<<4)+6);
}
#define init_14_7_3_7
void FASTCALL call_14_7_3_7(SH2_struct* sh)
{
  SH2movi(sh, 7, (3<<4)+7);
}
#define init_14_7_3_8
void FASTCALL call_14_7_3_8(SH2_struct* sh)
{
  SH2movi(sh, 7, (3<<4)+8);
}
#define init_14_7_3_9
void FASTCALL call_14_7_3_9(SH2_struct* sh)
{
  SH2movi(sh, 7, (3<<4)+9);
}
#define init_14_7_3_10
void FASTCALL call_14_7_3_10(SH2_struct* sh)
{
  SH2movi(sh, 7, (3<<4)+10);
}
#define init_14_7_3_11
void FASTCALL call_14_7_3_11(SH2_struct* sh)
{
  SH2movi(sh, 7, (3<<4)+11);
}
#define init_14_7_3_12
void FASTCALL call_14_7_3_12(SH2_struct* sh)
{
  SH2movi(sh, 7, (3<<4)+12);
}
#define init_14_7_3_13
void FASTCALL call_14_7_3_13(SH2_struct* sh)
{
  SH2movi(sh, 7, (3<<4)+13);
}
#define init_14_7_3_14
void FASTCALL call_14_7_3_14(SH2_struct* sh)
{
  SH2movi(sh, 7, (3<<4)+14);
}
#define init_14_7_3_15
void FASTCALL call_14_7_3_15(SH2_struct* sh)
{
  SH2movi(sh, 7, (3<<4)+15);
}
#define init_14_7_4_0
void FASTCALL call_14_7_4_0(SH2_struct* sh)
{
  SH2movi(sh, 7, (4<<4)+0);
}
#define init_14_7_4_1
void FASTCALL call_14_7_4_1(SH2_struct* sh)
{
  SH2movi(sh, 7, (4<<4)+1);
}
#define init_14_7_4_2
void FASTCALL call_14_7_4_2(SH2_struct* sh)
{
  SH2movi(sh, 7, (4<<4)+2);
}
#define init_14_7_4_3
void FASTCALL call_14_7_4_3(SH2_struct* sh)
{
  SH2movi(sh, 7, (4<<4)+3);
}
#define init_14_7_4_4
void FASTCALL call_14_7_4_4(SH2_struct* sh)
{
  SH2movi(sh, 7, (4<<4)+4);
}
#define init_14_7_4_5
void FASTCALL call_14_7_4_5(SH2_struct* sh)
{
  SH2movi(sh, 7, (4<<4)+5);
}
#define init_14_7_4_6
void FASTCALL call_14_7_4_6(SH2_struct* sh)
{
  SH2movi(sh, 7, (4<<4)+6);
}
#define init_14_7_4_7
void FASTCALL call_14_7_4_7(SH2_struct* sh)
{
  SH2movi(sh, 7, (4<<4)+7);
}
#define init_14_7_4_8
void FASTCALL call_14_7_4_8(SH2_struct* sh)
{
  SH2movi(sh, 7, (4<<4)+8);
}
#define init_14_7_4_9
void FASTCALL call_14_7_4_9(SH2_struct* sh)
{
  SH2movi(sh, 7, (4<<4)+9);
}
#define init_14_7_4_10
void FASTCALL call_14_7_4_10(SH2_struct* sh)
{
  SH2movi(sh, 7, (4<<4)+10);
}
#define init_14_7_4_11
void FASTCALL call_14_7_4_11(SH2_struct* sh)
{
  SH2movi(sh, 7, (4<<4)+11);
}
#define init_14_7_4_12
void FASTCALL call_14_7_4_12(SH2_struct* sh)
{
  SH2movi(sh, 7, (4<<4)+12);
}
#define init_14_7_4_13
void FASTCALL call_14_7_4_13(SH2_struct* sh)
{
  SH2movi(sh, 7, (4<<4)+13);
}
#define init_14_7_4_14
void FASTCALL call_14_7_4_14(SH2_struct* sh)
{
  SH2movi(sh, 7, (4<<4)+14);
}
#define init_14_7_4_15
void FASTCALL call_14_7_4_15(SH2_struct* sh)
{
  SH2movi(sh, 7, (4<<4)+15);
}
#define init_14_7_5_0
void FASTCALL call_14_7_5_0(SH2_struct* sh)
{
  SH2movi(sh, 7, (5<<4)+0);
}
#define init_14_7_5_1
void FASTCALL call_14_7_5_1(SH2_struct* sh)
{
  SH2movi(sh, 7, (5<<4)+1);
}
#define init_14_7_5_2
void FASTCALL call_14_7_5_2(SH2_struct* sh)
{
  SH2movi(sh, 7, (5<<4)+2);
}
#define init_14_7_5_3
void FASTCALL call_14_7_5_3(SH2_struct* sh)
{
  SH2movi(sh, 7, (5<<4)+3);
}
#define init_14_7_5_4
void FASTCALL call_14_7_5_4(SH2_struct* sh)
{
  SH2movi(sh, 7, (5<<4)+4);
}
#define init_14_7_5_5
void FASTCALL call_14_7_5_5(SH2_struct* sh)
{
  SH2movi(sh, 7, (5<<4)+5);
}
#define init_14_7_5_6
void FASTCALL call_14_7_5_6(SH2_struct* sh)
{
  SH2movi(sh, 7, (5<<4)+6);
}
#define init_14_7_5_7
void FASTCALL call_14_7_5_7(SH2_struct* sh)
{
  SH2movi(sh, 7, (5<<4)+7);
}
#define init_14_7_5_8
void FASTCALL call_14_7_5_8(SH2_struct* sh)
{
  SH2movi(sh, 7, (5<<4)+8);
}
#define init_14_7_5_9
void FASTCALL call_14_7_5_9(SH2_struct* sh)
{
  SH2movi(sh, 7, (5<<4)+9);
}
#define init_14_7_5_10
void FASTCALL call_14_7_5_10(SH2_struct* sh)
{
  SH2movi(sh, 7, (5<<4)+10);
}
#define init_14_7_5_11
void FASTCALL call_14_7_5_11(SH2_struct* sh)
{
  SH2movi(sh, 7, (5<<4)+11);
}
#define init_14_7_5_12
void FASTCALL call_14_7_5_12(SH2_struct* sh)
{
  SH2movi(sh, 7, (5<<4)+12);
}
#define init_14_7_5_13
void FASTCALL call_14_7_5_13(SH2_struct* sh)
{
  SH2movi(sh, 7, (5<<4)+13);
}
#define init_14_7_5_14
void FASTCALL call_14_7_5_14(SH2_struct* sh)
{
  SH2movi(sh, 7, (5<<4)+14);
}
#define init_14_7_5_15
void FASTCALL call_14_7_5_15(SH2_struct* sh)
{
  SH2movi(sh, 7, (5<<4)+15);
}
#define init_14_7_6_0
void FASTCALL call_14_7_6_0(SH2_struct* sh)
{
  SH2movi(sh, 7, (6<<4)+0);
}
#define init_14_7_6_1
void FASTCALL call_14_7_6_1(SH2_struct* sh)
{
  SH2movi(sh, 7, (6<<4)+1);
}
#define init_14_7_6_2
void FASTCALL call_14_7_6_2(SH2_struct* sh)
{
  SH2movi(sh, 7, (6<<4)+2);
}
#define init_14_7_6_3
void FASTCALL call_14_7_6_3(SH2_struct* sh)
{
  SH2movi(sh, 7, (6<<4)+3);
}
#define init_14_7_6_4
void FASTCALL call_14_7_6_4(SH2_struct* sh)
{
  SH2movi(sh, 7, (6<<4)+4);
}
#define init_14_7_6_5
void FASTCALL call_14_7_6_5(SH2_struct* sh)
{
  SH2movi(sh, 7, (6<<4)+5);
}
#define init_14_7_6_6
void FASTCALL call_14_7_6_6(SH2_struct* sh)
{
  SH2movi(sh, 7, (6<<4)+6);
}
#define init_14_7_6_7
void FASTCALL call_14_7_6_7(SH2_struct* sh)
{
  SH2movi(sh, 7, (6<<4)+7);
}
#define init_14_7_6_8
void FASTCALL call_14_7_6_8(SH2_struct* sh)
{
  SH2movi(sh, 7, (6<<4)+8);
}
#define init_14_7_6_9
void FASTCALL call_14_7_6_9(SH2_struct* sh)
{
  SH2movi(sh, 7, (6<<4)+9);
}
#define init_14_7_6_10
void FASTCALL call_14_7_6_10(SH2_struct* sh)
{
  SH2movi(sh, 7, (6<<4)+10);
}
#define init_14_7_6_11
void FASTCALL call_14_7_6_11(SH2_struct* sh)
{
  SH2movi(sh, 7, (6<<4)+11);
}
#define init_14_7_6_12
void FASTCALL call_14_7_6_12(SH2_struct* sh)
{
  SH2movi(sh, 7, (6<<4)+12);
}
#define init_14_7_6_13
void FASTCALL call_14_7_6_13(SH2_struct* sh)
{
  SH2movi(sh, 7, (6<<4)+13);
}
#define init_14_7_6_14
void FASTCALL call_14_7_6_14(SH2_struct* sh)
{
  SH2movi(sh, 7, (6<<4)+14);
}
#define init_14_7_6_15
void FASTCALL call_14_7_6_15(SH2_struct* sh)
{
  SH2movi(sh, 7, (6<<4)+15);
}
#define init_14_7_7_0
void FASTCALL call_14_7_7_0(SH2_struct* sh)
{
  SH2movi(sh, 7, (7<<4)+0);
}
#define init_14_7_7_1
void FASTCALL call_14_7_7_1(SH2_struct* sh)
{
  SH2movi(sh, 7, (7<<4)+1);
}
#define init_14_7_7_2
void FASTCALL call_14_7_7_2(SH2_struct* sh)
{
  SH2movi(sh, 7, (7<<4)+2);
}
#define init_14_7_7_3
void FASTCALL call_14_7_7_3(SH2_struct* sh)
{
  SH2movi(sh, 7, (7<<4)+3);
}
#define init_14_7_7_4
void FASTCALL call_14_7_7_4(SH2_struct* sh)
{
  SH2movi(sh, 7, (7<<4)+4);
}
#define init_14_7_7_5
void FASTCALL call_14_7_7_5(SH2_struct* sh)
{
  SH2movi(sh, 7, (7<<4)+5);
}
#define init_14_7_7_6
void FASTCALL call_14_7_7_6(SH2_struct* sh)
{
  SH2movi(sh, 7, (7<<4)+6);
}
#define init_14_7_7_7
void FASTCALL call_14_7_7_7(SH2_struct* sh)
{
  SH2movi(sh, 7, (7<<4)+7);
}
#define init_14_7_7_8
void FASTCALL call_14_7_7_8(SH2_struct* sh)
{
  SH2movi(sh, 7, (7<<4)+8);
}
#define init_14_7_7_9
void FASTCALL call_14_7_7_9(SH2_struct* sh)
{
  SH2movi(sh, 7, (7<<4)+9);
}
#define init_14_7_7_10
void FASTCALL call_14_7_7_10(SH2_struct* sh)
{
  SH2movi(sh, 7, (7<<4)+10);
}
#define init_14_7_7_11
void FASTCALL call_14_7_7_11(SH2_struct* sh)
{
  SH2movi(sh, 7, (7<<4)+11);
}
#define init_14_7_7_12
void FASTCALL call_14_7_7_12(SH2_struct* sh)
{
  SH2movi(sh, 7, (7<<4)+12);
}
#define init_14_7_7_13
void FASTCALL call_14_7_7_13(SH2_struct* sh)
{
  SH2movi(sh, 7, (7<<4)+13);
}
#define init_14_7_7_14
void FASTCALL call_14_7_7_14(SH2_struct* sh)
{
  SH2movi(sh, 7, (7<<4)+14);
}
#define init_14_7_7_15
void FASTCALL call_14_7_7_15(SH2_struct* sh)
{
  SH2movi(sh, 7, (7<<4)+15);
}
#define init_14_7_8_0
void FASTCALL call_14_7_8_0(SH2_struct* sh)
{
  SH2movi(sh, 7, (8<<4)+0);
}
#define init_14_7_8_1
void FASTCALL call_14_7_8_1(SH2_struct* sh)
{
  SH2movi(sh, 7, (8<<4)+1);
}
#define init_14_7_8_2
void FASTCALL call_14_7_8_2(SH2_struct* sh)
{
  SH2movi(sh, 7, (8<<4)+2);
}
#define init_14_7_8_3
void FASTCALL call_14_7_8_3(SH2_struct* sh)
{
  SH2movi(sh, 7, (8<<4)+3);
}
#define init_14_7_8_4
void FASTCALL call_14_7_8_4(SH2_struct* sh)
{
  SH2movi(sh, 7, (8<<4)+4);
}
#define init_14_7_8_5
void FASTCALL call_14_7_8_5(SH2_struct* sh)
{
  SH2movi(sh, 7, (8<<4)+5);
}
#define init_14_7_8_6
void FASTCALL call_14_7_8_6(SH2_struct* sh)
{
  SH2movi(sh, 7, (8<<4)+6);
}
#define init_14_7_8_7
void FASTCALL call_14_7_8_7(SH2_struct* sh)
{
  SH2movi(sh, 7, (8<<4)+7);
}
#define init_14_7_8_8
void FASTCALL call_14_7_8_8(SH2_struct* sh)
{
  SH2movi(sh, 7, (8<<4)+8);
}
#define init_14_7_8_9
void FASTCALL call_14_7_8_9(SH2_struct* sh)
{
  SH2movi(sh, 7, (8<<4)+9);
}
#define init_14_7_8_10
void FASTCALL call_14_7_8_10(SH2_struct* sh)
{
  SH2movi(sh, 7, (8<<4)+10);
}
#define init_14_7_8_11
void FASTCALL call_14_7_8_11(SH2_struct* sh)
{
  SH2movi(sh, 7, (8<<4)+11);
}
#define init_14_7_8_12
void FASTCALL call_14_7_8_12(SH2_struct* sh)
{
  SH2movi(sh, 7, (8<<4)+12);
}
#define init_14_7_8_13
void FASTCALL call_14_7_8_13(SH2_struct* sh)
{
  SH2movi(sh, 7, (8<<4)+13);
}
#define init_14_7_8_14
void FASTCALL call_14_7_8_14(SH2_struct* sh)
{
  SH2movi(sh, 7, (8<<4)+14);
}
#define init_14_7_8_15
void FASTCALL call_14_7_8_15(SH2_struct* sh)
{
  SH2movi(sh, 7, (8<<4)+15);
}
#define init_14_7_9_0
void FASTCALL call_14_7_9_0(SH2_struct* sh)
{
  SH2movi(sh, 7, (9<<4)+0);
}
#define init_14_7_9_1
void FASTCALL call_14_7_9_1(SH2_struct* sh)
{
  SH2movi(sh, 7, (9<<4)+1);
}
#define init_14_7_9_2
void FASTCALL call_14_7_9_2(SH2_struct* sh)
{
  SH2movi(sh, 7, (9<<4)+2);
}
#define init_14_7_9_3
void FASTCALL call_14_7_9_3(SH2_struct* sh)
{
  SH2movi(sh, 7, (9<<4)+3);
}
#define init_14_7_9_4
void FASTCALL call_14_7_9_4(SH2_struct* sh)
{
  SH2movi(sh, 7, (9<<4)+4);
}
#define init_14_7_9_5
void FASTCALL call_14_7_9_5(SH2_struct* sh)
{
  SH2movi(sh, 7, (9<<4)+5);
}
#define init_14_7_9_6
void FASTCALL call_14_7_9_6(SH2_struct* sh)
{
  SH2movi(sh, 7, (9<<4)+6);
}
#define init_14_7_9_7
void FASTCALL call_14_7_9_7(SH2_struct* sh)
{
  SH2movi(sh, 7, (9<<4)+7);
}
#define init_14_7_9_8
void FASTCALL call_14_7_9_8(SH2_struct* sh)
{
  SH2movi(sh, 7, (9<<4)+8);
}
#define init_14_7_9_9
void FASTCALL call_14_7_9_9(SH2_struct* sh)
{
  SH2movi(sh, 7, (9<<4)+9);
}
#define init_14_7_9_10
void FASTCALL call_14_7_9_10(SH2_struct* sh)
{
  SH2movi(sh, 7, (9<<4)+10);
}
#define init_14_7_9_11
void FASTCALL call_14_7_9_11(SH2_struct* sh)
{
  SH2movi(sh, 7, (9<<4)+11);
}
#define init_14_7_9_12
void FASTCALL call_14_7_9_12(SH2_struct* sh)
{
  SH2movi(sh, 7, (9<<4)+12);
}
#define init_14_7_9_13
void FASTCALL call_14_7_9_13(SH2_struct* sh)
{
  SH2movi(sh, 7, (9<<4)+13);
}
#define init_14_7_9_14
void FASTCALL call_14_7_9_14(SH2_struct* sh)
{
  SH2movi(sh, 7, (9<<4)+14);
}
#define init_14_7_9_15
void FASTCALL call_14_7_9_15(SH2_struct* sh)
{
  SH2movi(sh, 7, (9<<4)+15);
}
#define init_14_7_10_0
void FASTCALL call_14_7_10_0(SH2_struct* sh)
{
  SH2movi(sh, 7, (10<<4)+0);
}
#define init_14_7_10_1
void FASTCALL call_14_7_10_1(SH2_struct* sh)
{
  SH2movi(sh, 7, (10<<4)+1);
}
#define init_14_7_10_2
void FASTCALL call_14_7_10_2(SH2_struct* sh)
{
  SH2movi(sh, 7, (10<<4)+2);
}
#define init_14_7_10_3
void FASTCALL call_14_7_10_3(SH2_struct* sh)
{
  SH2movi(sh, 7, (10<<4)+3);
}
#define init_14_7_10_4
void FASTCALL call_14_7_10_4(SH2_struct* sh)
{
  SH2movi(sh, 7, (10<<4)+4);
}
#define init_14_7_10_5
void FASTCALL call_14_7_10_5(SH2_struct* sh)
{
  SH2movi(sh, 7, (10<<4)+5);
}
#define init_14_7_10_6
void FASTCALL call_14_7_10_6(SH2_struct* sh)
{
  SH2movi(sh, 7, (10<<4)+6);
}
#define init_14_7_10_7
void FASTCALL call_14_7_10_7(SH2_struct* sh)
{
  SH2movi(sh, 7, (10<<4)+7);
}
#define init_14_7_10_8
void FASTCALL call_14_7_10_8(SH2_struct* sh)
{
  SH2movi(sh, 7, (10<<4)+8);
}
#define init_14_7_10_9
void FASTCALL call_14_7_10_9(SH2_struct* sh)
{
  SH2movi(sh, 7, (10<<4)+9);
}
#define init_14_7_10_10
void FASTCALL call_14_7_10_10(SH2_struct* sh)
{
  SH2movi(sh, 7, (10<<4)+10);
}
#define init_14_7_10_11
void FASTCALL call_14_7_10_11(SH2_struct* sh)
{
  SH2movi(sh, 7, (10<<4)+11);
}
#define init_14_7_10_12
void FASTCALL call_14_7_10_12(SH2_struct* sh)
{
  SH2movi(sh, 7, (10<<4)+12);
}
#define init_14_7_10_13
void FASTCALL call_14_7_10_13(SH2_struct* sh)
{
  SH2movi(sh, 7, (10<<4)+13);
}
#define init_14_7_10_14
void FASTCALL call_14_7_10_14(SH2_struct* sh)
{
  SH2movi(sh, 7, (10<<4)+14);
}
#define init_14_7_10_15
void FASTCALL call_14_7_10_15(SH2_struct* sh)
{
  SH2movi(sh, 7, (10<<4)+15);
}
#define init_14_7_11_0
void FASTCALL call_14_7_11_0(SH2_struct* sh)
{
  SH2movi(sh, 7, (11<<4)+0);
}
#define init_14_7_11_1
void FASTCALL call_14_7_11_1(SH2_struct* sh)
{
  SH2movi(sh, 7, (11<<4)+1);
}
#define init_14_7_11_2
void FASTCALL call_14_7_11_2(SH2_struct* sh)
{
  SH2movi(sh, 7, (11<<4)+2);
}
#define init_14_7_11_3
void FASTCALL call_14_7_11_3(SH2_struct* sh)
{
  SH2movi(sh, 7, (11<<4)+3);
}
#define init_14_7_11_4
void FASTCALL call_14_7_11_4(SH2_struct* sh)
{
  SH2movi(sh, 7, (11<<4)+4);
}
#define init_14_7_11_5
void FASTCALL call_14_7_11_5(SH2_struct* sh)
{
  SH2movi(sh, 7, (11<<4)+5);
}
#define init_14_7_11_6
void FASTCALL call_14_7_11_6(SH2_struct* sh)
{
  SH2movi(sh, 7, (11<<4)+6);
}
#define init_14_7_11_7
void FASTCALL call_14_7_11_7(SH2_struct* sh)
{
  SH2movi(sh, 7, (11<<4)+7);
}
#define init_14_7_11_8
void FASTCALL call_14_7_11_8(SH2_struct* sh)
{
  SH2movi(sh, 7, (11<<4)+8);
}
#define init_14_7_11_9
void FASTCALL call_14_7_11_9(SH2_struct* sh)
{
  SH2movi(sh, 7, (11<<4)+9);
}
#define init_14_7_11_10
void FASTCALL call_14_7_11_10(SH2_struct* sh)
{
  SH2movi(sh, 7, (11<<4)+10);
}
#define init_14_7_11_11
void FASTCALL call_14_7_11_11(SH2_struct* sh)
{
  SH2movi(sh, 7, (11<<4)+11);
}
#define init_14_7_11_12
void FASTCALL call_14_7_11_12(SH2_struct* sh)
{
  SH2movi(sh, 7, (11<<4)+12);
}
#define init_14_7_11_13
void FASTCALL call_14_7_11_13(SH2_struct* sh)
{
  SH2movi(sh, 7, (11<<4)+13);
}
#define init_14_7_11_14
void FASTCALL call_14_7_11_14(SH2_struct* sh)
{
  SH2movi(sh, 7, (11<<4)+14);
}
#define init_14_7_11_15
void FASTCALL call_14_7_11_15(SH2_struct* sh)
{
  SH2movi(sh, 7, (11<<4)+15);
}
#define init_14_7_12_0
void FASTCALL call_14_7_12_0(SH2_struct* sh)
{
  SH2movi(sh, 7, (12<<4)+0);
}
#define init_14_7_12_1
void FASTCALL call_14_7_12_1(SH2_struct* sh)
{
  SH2movi(sh, 7, (12<<4)+1);
}
#define init_14_7_12_2
void FASTCALL call_14_7_12_2(SH2_struct* sh)
{
  SH2movi(sh, 7, (12<<4)+2);
}
#define init_14_7_12_3
void FASTCALL call_14_7_12_3(SH2_struct* sh)
{
  SH2movi(sh, 7, (12<<4)+3);
}
#define init_14_7_12_4
void FASTCALL call_14_7_12_4(SH2_struct* sh)
{
  SH2movi(sh, 7, (12<<4)+4);
}
#define init_14_7_12_5
void FASTCALL call_14_7_12_5(SH2_struct* sh)
{
  SH2movi(sh, 7, (12<<4)+5);
}
#define init_14_7_12_6
void FASTCALL call_14_7_12_6(SH2_struct* sh)
{
  SH2movi(sh, 7, (12<<4)+6);
}
#define init_14_7_12_7
void FASTCALL call_14_7_12_7(SH2_struct* sh)
{
  SH2movi(sh, 7, (12<<4)+7);
}
#define init_14_7_12_8
void FASTCALL call_14_7_12_8(SH2_struct* sh)
{
  SH2movi(sh, 7, (12<<4)+8);
}
#define init_14_7_12_9
void FASTCALL call_14_7_12_9(SH2_struct* sh)
{
  SH2movi(sh, 7, (12<<4)+9);
}
#define init_14_7_12_10
void FASTCALL call_14_7_12_10(SH2_struct* sh)
{
  SH2movi(sh, 7, (12<<4)+10);
}
#define init_14_7_12_11
void FASTCALL call_14_7_12_11(SH2_struct* sh)
{
  SH2movi(sh, 7, (12<<4)+11);
}
#define init_14_7_12_12
void FASTCALL call_14_7_12_12(SH2_struct* sh)
{
  SH2movi(sh, 7, (12<<4)+12);
}
#define init_14_7_12_13
void FASTCALL call_14_7_12_13(SH2_struct* sh)
{
  SH2movi(sh, 7, (12<<4)+13);
}
#define init_14_7_12_14
void FASTCALL call_14_7_12_14(SH2_struct* sh)
{
  SH2movi(sh, 7, (12<<4)+14);
}
#define init_14_7_12_15
void FASTCALL call_14_7_12_15(SH2_struct* sh)
{
  SH2movi(sh, 7, (12<<4)+15);
}
#define init_14_7_13_0
void FASTCALL call_14_7_13_0(SH2_struct* sh)
{
  SH2movi(sh, 7, (13<<4)+0);
}
#define init_14_7_13_1
void FASTCALL call_14_7_13_1(SH2_struct* sh)
{
  SH2movi(sh, 7, (13<<4)+1);
}
#define init_14_7_13_2
void FASTCALL call_14_7_13_2(SH2_struct* sh)
{
  SH2movi(sh, 7, (13<<4)+2);
}
#define init_14_7_13_3
void FASTCALL call_14_7_13_3(SH2_struct* sh)
{
  SH2movi(sh, 7, (13<<4)+3);
}
#define init_14_7_13_4
void FASTCALL call_14_7_13_4(SH2_struct* sh)
{
  SH2movi(sh, 7, (13<<4)+4);
}
#define init_14_7_13_5
void FASTCALL call_14_7_13_5(SH2_struct* sh)
{
  SH2movi(sh, 7, (13<<4)+5);
}
#define init_14_7_13_6
void FASTCALL call_14_7_13_6(SH2_struct* sh)
{
  SH2movi(sh, 7, (13<<4)+6);
}
#define init_14_7_13_7
void FASTCALL call_14_7_13_7(SH2_struct* sh)
{
  SH2movi(sh, 7, (13<<4)+7);
}
#define init_14_7_13_8
void FASTCALL call_14_7_13_8(SH2_struct* sh)
{
  SH2movi(sh, 7, (13<<4)+8);
}
#define init_14_7_13_9
void FASTCALL call_14_7_13_9(SH2_struct* sh)
{
  SH2movi(sh, 7, (13<<4)+9);
}
#define init_14_7_13_10
void FASTCALL call_14_7_13_10(SH2_struct* sh)
{
  SH2movi(sh, 7, (13<<4)+10);
}
#define init_14_7_13_11
void FASTCALL call_14_7_13_11(SH2_struct* sh)
{
  SH2movi(sh, 7, (13<<4)+11);
}
#define init_14_7_13_12
void FASTCALL call_14_7_13_12(SH2_struct* sh)
{
  SH2movi(sh, 7, (13<<4)+12);
}
#define init_14_7_13_13
void FASTCALL call_14_7_13_13(SH2_struct* sh)
{
  SH2movi(sh, 7, (13<<4)+13);
}
#define init_14_7_13_14
void FASTCALL call_14_7_13_14(SH2_struct* sh)
{
  SH2movi(sh, 7, (13<<4)+14);
}
#define init_14_7_13_15
void FASTCALL call_14_7_13_15(SH2_struct* sh)
{
  SH2movi(sh, 7, (13<<4)+15);
}
#define init_14_7_14_0
void FASTCALL call_14_7_14_0(SH2_struct* sh)
{
  SH2movi(sh, 7, (14<<4)+0);
}
#define init_14_7_14_1
void FASTCALL call_14_7_14_1(SH2_struct* sh)
{
  SH2movi(sh, 7, (14<<4)+1);
}
#define init_14_7_14_2
void FASTCALL call_14_7_14_2(SH2_struct* sh)
{
  SH2movi(sh, 7, (14<<4)+2);
}
#define init_14_7_14_3
void FASTCALL call_14_7_14_3(SH2_struct* sh)
{
  SH2movi(sh, 7, (14<<4)+3);
}
#define init_14_7_14_4
void FASTCALL call_14_7_14_4(SH2_struct* sh)
{
  SH2movi(sh, 7, (14<<4)+4);
}
#define init_14_7_14_5
void FASTCALL call_14_7_14_5(SH2_struct* sh)
{
  SH2movi(sh, 7, (14<<4)+5);
}
#define init_14_7_14_6
void FASTCALL call_14_7_14_6(SH2_struct* sh)
{
  SH2movi(sh, 7, (14<<4)+6);
}
#define init_14_7_14_7
void FASTCALL call_14_7_14_7(SH2_struct* sh)
{
  SH2movi(sh, 7, (14<<4)+7);
}
#define init_14_7_14_8
void FASTCALL call_14_7_14_8(SH2_struct* sh)
{
  SH2movi(sh, 7, (14<<4)+8);
}
#define init_14_7_14_9
void FASTCALL call_14_7_14_9(SH2_struct* sh)
{
  SH2movi(sh, 7, (14<<4)+9);
}
#define init_14_7_14_10
void FASTCALL call_14_7_14_10(SH2_struct* sh)
{
  SH2movi(sh, 7, (14<<4)+10);
}
#define init_14_7_14_11
void FASTCALL call_14_7_14_11(SH2_struct* sh)
{
  SH2movi(sh, 7, (14<<4)+11);
}
#define init_14_7_14_12
void FASTCALL call_14_7_14_12(SH2_struct* sh)
{
  SH2movi(sh, 7, (14<<4)+12);
}
#define init_14_7_14_13
void FASTCALL call_14_7_14_13(SH2_struct* sh)
{
  SH2movi(sh, 7, (14<<4)+13);
}
#define init_14_7_14_14
void FASTCALL call_14_7_14_14(SH2_struct* sh)
{
  SH2movi(sh, 7, (14<<4)+14);
}
#define init_14_7_14_15
void FASTCALL call_14_7_14_15(SH2_struct* sh)
{
  SH2movi(sh, 7, (14<<4)+15);
}
#define init_14_7_15_0
void FASTCALL call_14_7_15_0(SH2_struct* sh)
{
  SH2movi(sh, 7, (15<<4)+0);
}
#define init_14_7_15_1
void FASTCALL call_14_7_15_1(SH2_struct* sh)
{
  SH2movi(sh, 7, (15<<4)+1);
}
#define init_14_7_15_2
void FASTCALL call_14_7_15_2(SH2_struct* sh)
{
  SH2movi(sh, 7, (15<<4)+2);
}
#define init_14_7_15_3
void FASTCALL call_14_7_15_3(SH2_struct* sh)
{
  SH2movi(sh, 7, (15<<4)+3);
}
#define init_14_7_15_4
void FASTCALL call_14_7_15_4(SH2_struct* sh)
{
  SH2movi(sh, 7, (15<<4)+4);
}
#define init_14_7_15_5
void FASTCALL call_14_7_15_5(SH2_struct* sh)
{
  SH2movi(sh, 7, (15<<4)+5);
}
#define init_14_7_15_6
void FASTCALL call_14_7_15_6(SH2_struct* sh)
{
  SH2movi(sh, 7, (15<<4)+6);
}
#define init_14_7_15_7
void FASTCALL call_14_7_15_7(SH2_struct* sh)
{
  SH2movi(sh, 7, (15<<4)+7);
}
#define init_14_7_15_8
void FASTCALL call_14_7_15_8(SH2_struct* sh)
{
  SH2movi(sh, 7, (15<<4)+8);
}
#define init_14_7_15_9
void FASTCALL call_14_7_15_9(SH2_struct* sh)
{
  SH2movi(sh, 7, (15<<4)+9);
}
#define init_14_7_15_10
void FASTCALL call_14_7_15_10(SH2_struct* sh)
{
  SH2movi(sh, 7, (15<<4)+10);
}
#define init_14_7_15_11
void FASTCALL call_14_7_15_11(SH2_struct* sh)
{
  SH2movi(sh, 7, (15<<4)+11);
}
#define init_14_7_15_12
void FASTCALL call_14_7_15_12(SH2_struct* sh)
{
  SH2movi(sh, 7, (15<<4)+12);
}
#define init_14_7_15_13
void FASTCALL call_14_7_15_13(SH2_struct* sh)
{
  SH2movi(sh, 7, (15<<4)+13);
}
#define init_14_7_15_14
void FASTCALL call_14_7_15_14(SH2_struct* sh)
{
  SH2movi(sh, 7, (15<<4)+14);
}
#define init_14_7_15_15
void FASTCALL call_14_7_15_15(SH2_struct* sh)
{
  SH2movi(sh, 7, (15<<4)+15);
}
#define init_14_8_0_0
void FASTCALL call_14_8_0_0(SH2_struct* sh)
{
  SH2movi(sh, 8, (0<<4)+0);
}
#define init_14_8_0_1
void FASTCALL call_14_8_0_1(SH2_struct* sh)
{
  SH2movi(sh, 8, (0<<4)+1);
}
#define init_14_8_0_2
void FASTCALL call_14_8_0_2(SH2_struct* sh)
{
  SH2movi(sh, 8, (0<<4)+2);
}
#define init_14_8_0_3
void FASTCALL call_14_8_0_3(SH2_struct* sh)
{
  SH2movi(sh, 8, (0<<4)+3);
}
#define init_14_8_0_4
void FASTCALL call_14_8_0_4(SH2_struct* sh)
{
  SH2movi(sh, 8, (0<<4)+4);
}
#define init_14_8_0_5
void FASTCALL call_14_8_0_5(SH2_struct* sh)
{
  SH2movi(sh, 8, (0<<4)+5);
}
#define init_14_8_0_6
void FASTCALL call_14_8_0_6(SH2_struct* sh)
{
  SH2movi(sh, 8, (0<<4)+6);
}
#define init_14_8_0_7
void FASTCALL call_14_8_0_7(SH2_struct* sh)
{
  SH2movi(sh, 8, (0<<4)+7);
}
#define init_14_8_0_8
void FASTCALL call_14_8_0_8(SH2_struct* sh)
{
  SH2movi(sh, 8, (0<<4)+8);
}
#define init_14_8_0_9
void FASTCALL call_14_8_0_9(SH2_struct* sh)
{
  SH2movi(sh, 8, (0<<4)+9);
}
#define init_14_8_0_10
void FASTCALL call_14_8_0_10(SH2_struct* sh)
{
  SH2movi(sh, 8, (0<<4)+10);
}
#define init_14_8_0_11
void FASTCALL call_14_8_0_11(SH2_struct* sh)
{
  SH2movi(sh, 8, (0<<4)+11);
}
#define init_14_8_0_12
void FASTCALL call_14_8_0_12(SH2_struct* sh)
{
  SH2movi(sh, 8, (0<<4)+12);
}
#define init_14_8_0_13
void FASTCALL call_14_8_0_13(SH2_struct* sh)
{
  SH2movi(sh, 8, (0<<4)+13);
}
#define init_14_8_0_14
void FASTCALL call_14_8_0_14(SH2_struct* sh)
{
  SH2movi(sh, 8, (0<<4)+14);
}
#define init_14_8_0_15
void FASTCALL call_14_8_0_15(SH2_struct* sh)
{
  SH2movi(sh, 8, (0<<4)+15);
}
#define init_14_8_1_0
void FASTCALL call_14_8_1_0(SH2_struct* sh)
{
  SH2movi(sh, 8, (1<<4)+0);
}
#define init_14_8_1_1
void FASTCALL call_14_8_1_1(SH2_struct* sh)
{
  SH2movi(sh, 8, (1<<4)+1);
}
#define init_14_8_1_2
void FASTCALL call_14_8_1_2(SH2_struct* sh)
{
  SH2movi(sh, 8, (1<<4)+2);
}
#define init_14_8_1_3
void FASTCALL call_14_8_1_3(SH2_struct* sh)
{
  SH2movi(sh, 8, (1<<4)+3);
}
#define init_14_8_1_4
void FASTCALL call_14_8_1_4(SH2_struct* sh)
{
  SH2movi(sh, 8, (1<<4)+4);
}
#define init_14_8_1_5
void FASTCALL call_14_8_1_5(SH2_struct* sh)
{
  SH2movi(sh, 8, (1<<4)+5);
}
#define init_14_8_1_6
void FASTCALL call_14_8_1_6(SH2_struct* sh)
{
  SH2movi(sh, 8, (1<<4)+6);
}
#define init_14_8_1_7
void FASTCALL call_14_8_1_7(SH2_struct* sh)
{
  SH2movi(sh, 8, (1<<4)+7);
}
#define init_14_8_1_8
void FASTCALL call_14_8_1_8(SH2_struct* sh)
{
  SH2movi(sh, 8, (1<<4)+8);
}
#define init_14_8_1_9
void FASTCALL call_14_8_1_9(SH2_struct* sh)
{
  SH2movi(sh, 8, (1<<4)+9);
}
#define init_14_8_1_10
void FASTCALL call_14_8_1_10(SH2_struct* sh)
{
  SH2movi(sh, 8, (1<<4)+10);
}
#define init_14_8_1_11
void FASTCALL call_14_8_1_11(SH2_struct* sh)
{
  SH2movi(sh, 8, (1<<4)+11);
}
#define init_14_8_1_12
void FASTCALL call_14_8_1_12(SH2_struct* sh)
{
  SH2movi(sh, 8, (1<<4)+12);
}
#define init_14_8_1_13
void FASTCALL call_14_8_1_13(SH2_struct* sh)
{
  SH2movi(sh, 8, (1<<4)+13);
}
#define init_14_8_1_14
void FASTCALL call_14_8_1_14(SH2_struct* sh)
{
  SH2movi(sh, 8, (1<<4)+14);
}
#define init_14_8_1_15
void FASTCALL call_14_8_1_15(SH2_struct* sh)
{
  SH2movi(sh, 8, (1<<4)+15);
}
#define init_14_8_2_0
void FASTCALL call_14_8_2_0(SH2_struct* sh)
{
  SH2movi(sh, 8, (2<<4)+0);
}
#define init_14_8_2_1
void FASTCALL call_14_8_2_1(SH2_struct* sh)
{
  SH2movi(sh, 8, (2<<4)+1);
}
#define init_14_8_2_2
void FASTCALL call_14_8_2_2(SH2_struct* sh)
{
  SH2movi(sh, 8, (2<<4)+2);
}
#define init_14_8_2_3
void FASTCALL call_14_8_2_3(SH2_struct* sh)
{
  SH2movi(sh, 8, (2<<4)+3);
}
#define init_14_8_2_4
void FASTCALL call_14_8_2_4(SH2_struct* sh)
{
  SH2movi(sh, 8, (2<<4)+4);
}
#define init_14_8_2_5
void FASTCALL call_14_8_2_5(SH2_struct* sh)
{
  SH2movi(sh, 8, (2<<4)+5);
}
#define init_14_8_2_6
void FASTCALL call_14_8_2_6(SH2_struct* sh)
{
  SH2movi(sh, 8, (2<<4)+6);
}
#define init_14_8_2_7
void FASTCALL call_14_8_2_7(SH2_struct* sh)
{
  SH2movi(sh, 8, (2<<4)+7);
}
#define init_14_8_2_8
void FASTCALL call_14_8_2_8(SH2_struct* sh)
{
  SH2movi(sh, 8, (2<<4)+8);
}
#define init_14_8_2_9
void FASTCALL call_14_8_2_9(SH2_struct* sh)
{
  SH2movi(sh, 8, (2<<4)+9);
}
#define init_14_8_2_10
void FASTCALL call_14_8_2_10(SH2_struct* sh)
{
  SH2movi(sh, 8, (2<<4)+10);
}
#define init_14_8_2_11
void FASTCALL call_14_8_2_11(SH2_struct* sh)
{
  SH2movi(sh, 8, (2<<4)+11);
}
#define init_14_8_2_12
void FASTCALL call_14_8_2_12(SH2_struct* sh)
{
  SH2movi(sh, 8, (2<<4)+12);
}
#define init_14_8_2_13
void FASTCALL call_14_8_2_13(SH2_struct* sh)
{
  SH2movi(sh, 8, (2<<4)+13);
}
#define init_14_8_2_14
void FASTCALL call_14_8_2_14(SH2_struct* sh)
{
  SH2movi(sh, 8, (2<<4)+14);
}
#define init_14_8_2_15
void FASTCALL call_14_8_2_15(SH2_struct* sh)
{
  SH2movi(sh, 8, (2<<4)+15);
}
#define init_14_8_3_0
void FASTCALL call_14_8_3_0(SH2_struct* sh)
{
  SH2movi(sh, 8, (3<<4)+0);
}
#define init_14_8_3_1
void FASTCALL call_14_8_3_1(SH2_struct* sh)
{
  SH2movi(sh, 8, (3<<4)+1);
}
#define init_14_8_3_2
void FASTCALL call_14_8_3_2(SH2_struct* sh)
{
  SH2movi(sh, 8, (3<<4)+2);
}
#define init_14_8_3_3
void FASTCALL call_14_8_3_3(SH2_struct* sh)
{
  SH2movi(sh, 8, (3<<4)+3);
}
#define init_14_8_3_4
void FASTCALL call_14_8_3_4(SH2_struct* sh)
{
  SH2movi(sh, 8, (3<<4)+4);
}
#define init_14_8_3_5
void FASTCALL call_14_8_3_5(SH2_struct* sh)
{
  SH2movi(sh, 8, (3<<4)+5);
}
#define init_14_8_3_6
void FASTCALL call_14_8_3_6(SH2_struct* sh)
{
  SH2movi(sh, 8, (3<<4)+6);
}
#define init_14_8_3_7
void FASTCALL call_14_8_3_7(SH2_struct* sh)
{
  SH2movi(sh, 8, (3<<4)+7);
}
#define init_14_8_3_8
void FASTCALL call_14_8_3_8(SH2_struct* sh)
{
  SH2movi(sh, 8, (3<<4)+8);
}
#define init_14_8_3_9
void FASTCALL call_14_8_3_9(SH2_struct* sh)
{
  SH2movi(sh, 8, (3<<4)+9);
}
#define init_14_8_3_10
void FASTCALL call_14_8_3_10(SH2_struct* sh)
{
  SH2movi(sh, 8, (3<<4)+10);
}
#define init_14_8_3_11
void FASTCALL call_14_8_3_11(SH2_struct* sh)
{
  SH2movi(sh, 8, (3<<4)+11);
}
#define init_14_8_3_12
void FASTCALL call_14_8_3_12(SH2_struct* sh)
{
  SH2movi(sh, 8, (3<<4)+12);
}
#define init_14_8_3_13
void FASTCALL call_14_8_3_13(SH2_struct* sh)
{
  SH2movi(sh, 8, (3<<4)+13);
}
#define init_14_8_3_14
void FASTCALL call_14_8_3_14(SH2_struct* sh)
{
  SH2movi(sh, 8, (3<<4)+14);
}
#define init_14_8_3_15
void FASTCALL call_14_8_3_15(SH2_struct* sh)
{
  SH2movi(sh, 8, (3<<4)+15);
}
#define init_14_8_4_0
void FASTCALL call_14_8_4_0(SH2_struct* sh)
{
  SH2movi(sh, 8, (4<<4)+0);
}
#define init_14_8_4_1
void FASTCALL call_14_8_4_1(SH2_struct* sh)
{
  SH2movi(sh, 8, (4<<4)+1);
}
#define init_14_8_4_2
void FASTCALL call_14_8_4_2(SH2_struct* sh)
{
  SH2movi(sh, 8, (4<<4)+2);
}
#define init_14_8_4_3
void FASTCALL call_14_8_4_3(SH2_struct* sh)
{
  SH2movi(sh, 8, (4<<4)+3);
}
#define init_14_8_4_4
void FASTCALL call_14_8_4_4(SH2_struct* sh)
{
  SH2movi(sh, 8, (4<<4)+4);
}
#define init_14_8_4_5
void FASTCALL call_14_8_4_5(SH2_struct* sh)
{
  SH2movi(sh, 8, (4<<4)+5);
}
#define init_14_8_4_6
void FASTCALL call_14_8_4_6(SH2_struct* sh)
{
  SH2movi(sh, 8, (4<<4)+6);
}
#define init_14_8_4_7
void FASTCALL call_14_8_4_7(SH2_struct* sh)
{
  SH2movi(sh, 8, (4<<4)+7);
}
#define init_14_8_4_8
void FASTCALL call_14_8_4_8(SH2_struct* sh)
{
  SH2movi(sh, 8, (4<<4)+8);
}
#define init_14_8_4_9
void FASTCALL call_14_8_4_9(SH2_struct* sh)
{
  SH2movi(sh, 8, (4<<4)+9);
}
#define init_14_8_4_10
void FASTCALL call_14_8_4_10(SH2_struct* sh)
{
  SH2movi(sh, 8, (4<<4)+10);
}
#define init_14_8_4_11
void FASTCALL call_14_8_4_11(SH2_struct* sh)
{
  SH2movi(sh, 8, (4<<4)+11);
}
#define init_14_8_4_12
void FASTCALL call_14_8_4_12(SH2_struct* sh)
{
  SH2movi(sh, 8, (4<<4)+12);
}
#define init_14_8_4_13
void FASTCALL call_14_8_4_13(SH2_struct* sh)
{
  SH2movi(sh, 8, (4<<4)+13);
}
#define init_14_8_4_14
void FASTCALL call_14_8_4_14(SH2_struct* sh)
{
  SH2movi(sh, 8, (4<<4)+14);
}
#define init_14_8_4_15
void FASTCALL call_14_8_4_15(SH2_struct* sh)
{
  SH2movi(sh, 8, (4<<4)+15);
}
#define init_14_8_5_0
void FASTCALL call_14_8_5_0(SH2_struct* sh)
{
  SH2movi(sh, 8, (5<<4)+0);
}
#define init_14_8_5_1
void FASTCALL call_14_8_5_1(SH2_struct* sh)
{
  SH2movi(sh, 8, (5<<4)+1);
}
#define init_14_8_5_2
void FASTCALL call_14_8_5_2(SH2_struct* sh)
{
  SH2movi(sh, 8, (5<<4)+2);
}
#define init_14_8_5_3
void FASTCALL call_14_8_5_3(SH2_struct* sh)
{
  SH2movi(sh, 8, (5<<4)+3);
}
#define init_14_8_5_4
void FASTCALL call_14_8_5_4(SH2_struct* sh)
{
  SH2movi(sh, 8, (5<<4)+4);
}
#define init_14_8_5_5
void FASTCALL call_14_8_5_5(SH2_struct* sh)
{
  SH2movi(sh, 8, (5<<4)+5);
}
#define init_14_8_5_6
void FASTCALL call_14_8_5_6(SH2_struct* sh)
{
  SH2movi(sh, 8, (5<<4)+6);
}
#define init_14_8_5_7
void FASTCALL call_14_8_5_7(SH2_struct* sh)
{
  SH2movi(sh, 8, (5<<4)+7);
}
#define init_14_8_5_8
void FASTCALL call_14_8_5_8(SH2_struct* sh)
{
  SH2movi(sh, 8, (5<<4)+8);
}
#define init_14_8_5_9
void FASTCALL call_14_8_5_9(SH2_struct* sh)
{
  SH2movi(sh, 8, (5<<4)+9);
}
#define init_14_8_5_10
void FASTCALL call_14_8_5_10(SH2_struct* sh)
{
  SH2movi(sh, 8, (5<<4)+10);
}
#define init_14_8_5_11
void FASTCALL call_14_8_5_11(SH2_struct* sh)
{
  SH2movi(sh, 8, (5<<4)+11);
}
#define init_14_8_5_12
void FASTCALL call_14_8_5_12(SH2_struct* sh)
{
  SH2movi(sh, 8, (5<<4)+12);
}
#define init_14_8_5_13
void FASTCALL call_14_8_5_13(SH2_struct* sh)
{
  SH2movi(sh, 8, (5<<4)+13);
}
#define init_14_8_5_14
void FASTCALL call_14_8_5_14(SH2_struct* sh)
{
  SH2movi(sh, 8, (5<<4)+14);
}
#define init_14_8_5_15
void FASTCALL call_14_8_5_15(SH2_struct* sh)
{
  SH2movi(sh, 8, (5<<4)+15);
}
#define init_14_8_6_0
void FASTCALL call_14_8_6_0(SH2_struct* sh)
{
  SH2movi(sh, 8, (6<<4)+0);
}
#define init_14_8_6_1
void FASTCALL call_14_8_6_1(SH2_struct* sh)
{
  SH2movi(sh, 8, (6<<4)+1);
}
#define init_14_8_6_2
void FASTCALL call_14_8_6_2(SH2_struct* sh)
{
  SH2movi(sh, 8, (6<<4)+2);
}
#define init_14_8_6_3
void FASTCALL call_14_8_6_3(SH2_struct* sh)
{
  SH2movi(sh, 8, (6<<4)+3);
}
#define init_14_8_6_4
void FASTCALL call_14_8_6_4(SH2_struct* sh)
{
  SH2movi(sh, 8, (6<<4)+4);
}
#define init_14_8_6_5
void FASTCALL call_14_8_6_5(SH2_struct* sh)
{
  SH2movi(sh, 8, (6<<4)+5);
}
#define init_14_8_6_6
void FASTCALL call_14_8_6_6(SH2_struct* sh)
{
  SH2movi(sh, 8, (6<<4)+6);
}
#define init_14_8_6_7
void FASTCALL call_14_8_6_7(SH2_struct* sh)
{
  SH2movi(sh, 8, (6<<4)+7);
}
#define init_14_8_6_8
void FASTCALL call_14_8_6_8(SH2_struct* sh)
{
  SH2movi(sh, 8, (6<<4)+8);
}
#define init_14_8_6_9
void FASTCALL call_14_8_6_9(SH2_struct* sh)
{
  SH2movi(sh, 8, (6<<4)+9);
}
#define init_14_8_6_10
void FASTCALL call_14_8_6_10(SH2_struct* sh)
{
  SH2movi(sh, 8, (6<<4)+10);
}
#define init_14_8_6_11
void FASTCALL call_14_8_6_11(SH2_struct* sh)
{
  SH2movi(sh, 8, (6<<4)+11);
}
#define init_14_8_6_12
void FASTCALL call_14_8_6_12(SH2_struct* sh)
{
  SH2movi(sh, 8, (6<<4)+12);
}
#define init_14_8_6_13
void FASTCALL call_14_8_6_13(SH2_struct* sh)
{
  SH2movi(sh, 8, (6<<4)+13);
}
#define init_14_8_6_14
void FASTCALL call_14_8_6_14(SH2_struct* sh)
{
  SH2movi(sh, 8, (6<<4)+14);
}
#define init_14_8_6_15
void FASTCALL call_14_8_6_15(SH2_struct* sh)
{
  SH2movi(sh, 8, (6<<4)+15);
}
#define init_14_8_7_0
void FASTCALL call_14_8_7_0(SH2_struct* sh)
{
  SH2movi(sh, 8, (7<<4)+0);
}
#define init_14_8_7_1
void FASTCALL call_14_8_7_1(SH2_struct* sh)
{
  SH2movi(sh, 8, (7<<4)+1);
}
#define init_14_8_7_2
void FASTCALL call_14_8_7_2(SH2_struct* sh)
{
  SH2movi(sh, 8, (7<<4)+2);
}
#define init_14_8_7_3
void FASTCALL call_14_8_7_3(SH2_struct* sh)
{
  SH2movi(sh, 8, (7<<4)+3);
}
#define init_14_8_7_4
void FASTCALL call_14_8_7_4(SH2_struct* sh)
{
  SH2movi(sh, 8, (7<<4)+4);
}
#define init_14_8_7_5
void FASTCALL call_14_8_7_5(SH2_struct* sh)
{
  SH2movi(sh, 8, (7<<4)+5);
}
#define init_14_8_7_6
void FASTCALL call_14_8_7_6(SH2_struct* sh)
{
  SH2movi(sh, 8, (7<<4)+6);
}
#define init_14_8_7_7
void FASTCALL call_14_8_7_7(SH2_struct* sh)
{
  SH2movi(sh, 8, (7<<4)+7);
}
#define init_14_8_7_8
void FASTCALL call_14_8_7_8(SH2_struct* sh)
{
  SH2movi(sh, 8, (7<<4)+8);
}
#define init_14_8_7_9
void FASTCALL call_14_8_7_9(SH2_struct* sh)
{
  SH2movi(sh, 8, (7<<4)+9);
}
#define init_14_8_7_10
void FASTCALL call_14_8_7_10(SH2_struct* sh)
{
  SH2movi(sh, 8, (7<<4)+10);
}
#define init_14_8_7_11
void FASTCALL call_14_8_7_11(SH2_struct* sh)
{
  SH2movi(sh, 8, (7<<4)+11);
}
#define init_14_8_7_12
void FASTCALL call_14_8_7_12(SH2_struct* sh)
{
  SH2movi(sh, 8, (7<<4)+12);
}
#define init_14_8_7_13
void FASTCALL call_14_8_7_13(SH2_struct* sh)
{
  SH2movi(sh, 8, (7<<4)+13);
}
#define init_14_8_7_14
void FASTCALL call_14_8_7_14(SH2_struct* sh)
{
  SH2movi(sh, 8, (7<<4)+14);
}
#define init_14_8_7_15
void FASTCALL call_14_8_7_15(SH2_struct* sh)
{
  SH2movi(sh, 8, (7<<4)+15);
}
#define init_14_8_8_0
void FASTCALL call_14_8_8_0(SH2_struct* sh)
{
  SH2movi(sh, 8, (8<<4)+0);
}
#define init_14_8_8_1
void FASTCALL call_14_8_8_1(SH2_struct* sh)
{
  SH2movi(sh, 8, (8<<4)+1);
}
#define init_14_8_8_2
void FASTCALL call_14_8_8_2(SH2_struct* sh)
{
  SH2movi(sh, 8, (8<<4)+2);
}
#define init_14_8_8_3
void FASTCALL call_14_8_8_3(SH2_struct* sh)
{
  SH2movi(sh, 8, (8<<4)+3);
}
#define init_14_8_8_4
void FASTCALL call_14_8_8_4(SH2_struct* sh)
{
  SH2movi(sh, 8, (8<<4)+4);
}
#define init_14_8_8_5
void FASTCALL call_14_8_8_5(SH2_struct* sh)
{
  SH2movi(sh, 8, (8<<4)+5);
}
#define init_14_8_8_6
void FASTCALL call_14_8_8_6(SH2_struct* sh)
{
  SH2movi(sh, 8, (8<<4)+6);
}
#define init_14_8_8_7
void FASTCALL call_14_8_8_7(SH2_struct* sh)
{
  SH2movi(sh, 8, (8<<4)+7);
}
#define init_14_8_8_8
void FASTCALL call_14_8_8_8(SH2_struct* sh)
{
  SH2movi(sh, 8, (8<<4)+8);
}
#define init_14_8_8_9
void FASTCALL call_14_8_8_9(SH2_struct* sh)
{
  SH2movi(sh, 8, (8<<4)+9);
}
#define init_14_8_8_10
void FASTCALL call_14_8_8_10(SH2_struct* sh)
{
  SH2movi(sh, 8, (8<<4)+10);
}
#define init_14_8_8_11
void FASTCALL call_14_8_8_11(SH2_struct* sh)
{
  SH2movi(sh, 8, (8<<4)+11);
}
#define init_14_8_8_12
void FASTCALL call_14_8_8_12(SH2_struct* sh)
{
  SH2movi(sh, 8, (8<<4)+12);
}
#define init_14_8_8_13
void FASTCALL call_14_8_8_13(SH2_struct* sh)
{
  SH2movi(sh, 8, (8<<4)+13);
}
#define init_14_8_8_14
void FASTCALL call_14_8_8_14(SH2_struct* sh)
{
  SH2movi(sh, 8, (8<<4)+14);
}
#define init_14_8_8_15
void FASTCALL call_14_8_8_15(SH2_struct* sh)
{
  SH2movi(sh, 8, (8<<4)+15);
}
#define init_14_8_9_0
void FASTCALL call_14_8_9_0(SH2_struct* sh)
{
  SH2movi(sh, 8, (9<<4)+0);
}
#define init_14_8_9_1
void FASTCALL call_14_8_9_1(SH2_struct* sh)
{
  SH2movi(sh, 8, (9<<4)+1);
}
#define init_14_8_9_2
void FASTCALL call_14_8_9_2(SH2_struct* sh)
{
  SH2movi(sh, 8, (9<<4)+2);
}
#define init_14_8_9_3
void FASTCALL call_14_8_9_3(SH2_struct* sh)
{
  SH2movi(sh, 8, (9<<4)+3);
}
#define init_14_8_9_4
void FASTCALL call_14_8_9_4(SH2_struct* sh)
{
  SH2movi(sh, 8, (9<<4)+4);
}
#define init_14_8_9_5
void FASTCALL call_14_8_9_5(SH2_struct* sh)
{
  SH2movi(sh, 8, (9<<4)+5);
}
#define init_14_8_9_6
void FASTCALL call_14_8_9_6(SH2_struct* sh)
{
  SH2movi(sh, 8, (9<<4)+6);
}
#define init_14_8_9_7
void FASTCALL call_14_8_9_7(SH2_struct* sh)
{
  SH2movi(sh, 8, (9<<4)+7);
}
#define init_14_8_9_8
void FASTCALL call_14_8_9_8(SH2_struct* sh)
{
  SH2movi(sh, 8, (9<<4)+8);
}
#define init_14_8_9_9
void FASTCALL call_14_8_9_9(SH2_struct* sh)
{
  SH2movi(sh, 8, (9<<4)+9);
}
#define init_14_8_9_10
void FASTCALL call_14_8_9_10(SH2_struct* sh)
{
  SH2movi(sh, 8, (9<<4)+10);
}
#define init_14_8_9_11
void FASTCALL call_14_8_9_11(SH2_struct* sh)
{
  SH2movi(sh, 8, (9<<4)+11);
}
#define init_14_8_9_12
void FASTCALL call_14_8_9_12(SH2_struct* sh)
{
  SH2movi(sh, 8, (9<<4)+12);
}
#define init_14_8_9_13
void FASTCALL call_14_8_9_13(SH2_struct* sh)
{
  SH2movi(sh, 8, (9<<4)+13);
}
#define init_14_8_9_14
void FASTCALL call_14_8_9_14(SH2_struct* sh)
{
  SH2movi(sh, 8, (9<<4)+14);
}
#define init_14_8_9_15
void FASTCALL call_14_8_9_15(SH2_struct* sh)
{
  SH2movi(sh, 8, (9<<4)+15);
}
#define init_14_8_10_0
void FASTCALL call_14_8_10_0(SH2_struct* sh)
{
  SH2movi(sh, 8, (10<<4)+0);
}
#define init_14_8_10_1
void FASTCALL call_14_8_10_1(SH2_struct* sh)
{
  SH2movi(sh, 8, (10<<4)+1);
}
#define init_14_8_10_2
void FASTCALL call_14_8_10_2(SH2_struct* sh)
{
  SH2movi(sh, 8, (10<<4)+2);
}
#define init_14_8_10_3
void FASTCALL call_14_8_10_3(SH2_struct* sh)
{
  SH2movi(sh, 8, (10<<4)+3);
}
#define init_14_8_10_4
void FASTCALL call_14_8_10_4(SH2_struct* sh)
{
  SH2movi(sh, 8, (10<<4)+4);
}
#define init_14_8_10_5
void FASTCALL call_14_8_10_5(SH2_struct* sh)
{
  SH2movi(sh, 8, (10<<4)+5);
}
#define init_14_8_10_6
void FASTCALL call_14_8_10_6(SH2_struct* sh)
{
  SH2movi(sh, 8, (10<<4)+6);
}
#define init_14_8_10_7
void FASTCALL call_14_8_10_7(SH2_struct* sh)
{
  SH2movi(sh, 8, (10<<4)+7);
}
#define init_14_8_10_8
void FASTCALL call_14_8_10_8(SH2_struct* sh)
{
  SH2movi(sh, 8, (10<<4)+8);
}
#define init_14_8_10_9
void FASTCALL call_14_8_10_9(SH2_struct* sh)
{
  SH2movi(sh, 8, (10<<4)+9);
}
#define init_14_8_10_10
void FASTCALL call_14_8_10_10(SH2_struct* sh)
{
  SH2movi(sh, 8, (10<<4)+10);
}
#define init_14_8_10_11
void FASTCALL call_14_8_10_11(SH2_struct* sh)
{
  SH2movi(sh, 8, (10<<4)+11);
}
#define init_14_8_10_12
void FASTCALL call_14_8_10_12(SH2_struct* sh)
{
  SH2movi(sh, 8, (10<<4)+12);
}
#define init_14_8_10_13
void FASTCALL call_14_8_10_13(SH2_struct* sh)
{
  SH2movi(sh, 8, (10<<4)+13);
}
#define init_14_8_10_14
void FASTCALL call_14_8_10_14(SH2_struct* sh)
{
  SH2movi(sh, 8, (10<<4)+14);
}
#define init_14_8_10_15
void FASTCALL call_14_8_10_15(SH2_struct* sh)
{
  SH2movi(sh, 8, (10<<4)+15);
}
#define init_14_8_11_0
void FASTCALL call_14_8_11_0(SH2_struct* sh)
{
  SH2movi(sh, 8, (11<<4)+0);
}
#define init_14_8_11_1
void FASTCALL call_14_8_11_1(SH2_struct* sh)
{
  SH2movi(sh, 8, (11<<4)+1);
}
#define init_14_8_11_2
void FASTCALL call_14_8_11_2(SH2_struct* sh)
{
  SH2movi(sh, 8, (11<<4)+2);
}
#define init_14_8_11_3
void FASTCALL call_14_8_11_3(SH2_struct* sh)
{
  SH2movi(sh, 8, (11<<4)+3);
}
#define init_14_8_11_4
void FASTCALL call_14_8_11_4(SH2_struct* sh)
{
  SH2movi(sh, 8, (11<<4)+4);
}
#define init_14_8_11_5
void FASTCALL call_14_8_11_5(SH2_struct* sh)
{
  SH2movi(sh, 8, (11<<4)+5);
}
#define init_14_8_11_6
void FASTCALL call_14_8_11_6(SH2_struct* sh)
{
  SH2movi(sh, 8, (11<<4)+6);
}
#define init_14_8_11_7
void FASTCALL call_14_8_11_7(SH2_struct* sh)
{
  SH2movi(sh, 8, (11<<4)+7);
}
#define init_14_8_11_8
void FASTCALL call_14_8_11_8(SH2_struct* sh)
{
  SH2movi(sh, 8, (11<<4)+8);
}
#define init_14_8_11_9
void FASTCALL call_14_8_11_9(SH2_struct* sh)
{
  SH2movi(sh, 8, (11<<4)+9);
}
#define init_14_8_11_10
void FASTCALL call_14_8_11_10(SH2_struct* sh)
{
  SH2movi(sh, 8, (11<<4)+10);
}
#define init_14_8_11_11
void FASTCALL call_14_8_11_11(SH2_struct* sh)
{
  SH2movi(sh, 8, (11<<4)+11);
}
#define init_14_8_11_12
void FASTCALL call_14_8_11_12(SH2_struct* sh)
{
  SH2movi(sh, 8, (11<<4)+12);
}
#define init_14_8_11_13
void FASTCALL call_14_8_11_13(SH2_struct* sh)
{
  SH2movi(sh, 8, (11<<4)+13);
}
#define init_14_8_11_14
void FASTCALL call_14_8_11_14(SH2_struct* sh)
{
  SH2movi(sh, 8, (11<<4)+14);
}
#define init_14_8_11_15
void FASTCALL call_14_8_11_15(SH2_struct* sh)
{
  SH2movi(sh, 8, (11<<4)+15);
}
#define init_14_8_12_0
void FASTCALL call_14_8_12_0(SH2_struct* sh)
{
  SH2movi(sh, 8, (12<<4)+0);
}
#define init_14_8_12_1
void FASTCALL call_14_8_12_1(SH2_struct* sh)
{
  SH2movi(sh, 8, (12<<4)+1);
}
#define init_14_8_12_2
void FASTCALL call_14_8_12_2(SH2_struct* sh)
{
  SH2movi(sh, 8, (12<<4)+2);
}
#define init_14_8_12_3
void FASTCALL call_14_8_12_3(SH2_struct* sh)
{
  SH2movi(sh, 8, (12<<4)+3);
}
#define init_14_8_12_4
void FASTCALL call_14_8_12_4(SH2_struct* sh)
{
  SH2movi(sh, 8, (12<<4)+4);
}
#define init_14_8_12_5
void FASTCALL call_14_8_12_5(SH2_struct* sh)
{
  SH2movi(sh, 8, (12<<4)+5);
}
#define init_14_8_12_6
void FASTCALL call_14_8_12_6(SH2_struct* sh)
{
  SH2movi(sh, 8, (12<<4)+6);
}
#define init_14_8_12_7
void FASTCALL call_14_8_12_7(SH2_struct* sh)
{
  SH2movi(sh, 8, (12<<4)+7);
}
#define init_14_8_12_8
void FASTCALL call_14_8_12_8(SH2_struct* sh)
{
  SH2movi(sh, 8, (12<<4)+8);
}
#define init_14_8_12_9
void FASTCALL call_14_8_12_9(SH2_struct* sh)
{
  SH2movi(sh, 8, (12<<4)+9);
}
#define init_14_8_12_10
void FASTCALL call_14_8_12_10(SH2_struct* sh)
{
  SH2movi(sh, 8, (12<<4)+10);
}
#define init_14_8_12_11
void FASTCALL call_14_8_12_11(SH2_struct* sh)
{
  SH2movi(sh, 8, (12<<4)+11);
}
#define init_14_8_12_12
void FASTCALL call_14_8_12_12(SH2_struct* sh)
{
  SH2movi(sh, 8, (12<<4)+12);
}
#define init_14_8_12_13
void FASTCALL call_14_8_12_13(SH2_struct* sh)
{
  SH2movi(sh, 8, (12<<4)+13);
}
#define init_14_8_12_14
void FASTCALL call_14_8_12_14(SH2_struct* sh)
{
  SH2movi(sh, 8, (12<<4)+14);
}
#define init_14_8_12_15
void FASTCALL call_14_8_12_15(SH2_struct* sh)
{
  SH2movi(sh, 8, (12<<4)+15);
}
#define init_14_8_13_0
void FASTCALL call_14_8_13_0(SH2_struct* sh)
{
  SH2movi(sh, 8, (13<<4)+0);
}
#define init_14_8_13_1
void FASTCALL call_14_8_13_1(SH2_struct* sh)
{
  SH2movi(sh, 8, (13<<4)+1);
}
#define init_14_8_13_2
void FASTCALL call_14_8_13_2(SH2_struct* sh)
{
  SH2movi(sh, 8, (13<<4)+2);
}
#define init_14_8_13_3
void FASTCALL call_14_8_13_3(SH2_struct* sh)
{
  SH2movi(sh, 8, (13<<4)+3);
}
#define init_14_8_13_4
void FASTCALL call_14_8_13_4(SH2_struct* sh)
{
  SH2movi(sh, 8, (13<<4)+4);
}
#define init_14_8_13_5
void FASTCALL call_14_8_13_5(SH2_struct* sh)
{
  SH2movi(sh, 8, (13<<4)+5);
}
#define init_14_8_13_6
void FASTCALL call_14_8_13_6(SH2_struct* sh)
{
  SH2movi(sh, 8, (13<<4)+6);
}
#define init_14_8_13_7
void FASTCALL call_14_8_13_7(SH2_struct* sh)
{
  SH2movi(sh, 8, (13<<4)+7);
}
#define init_14_8_13_8
void FASTCALL call_14_8_13_8(SH2_struct* sh)
{
  SH2movi(sh, 8, (13<<4)+8);
}
#define init_14_8_13_9
void FASTCALL call_14_8_13_9(SH2_struct* sh)
{
  SH2movi(sh, 8, (13<<4)+9);
}
#define init_14_8_13_10
void FASTCALL call_14_8_13_10(SH2_struct* sh)
{
  SH2movi(sh, 8, (13<<4)+10);
}
#define init_14_8_13_11
void FASTCALL call_14_8_13_11(SH2_struct* sh)
{
  SH2movi(sh, 8, (13<<4)+11);
}
#define init_14_8_13_12
void FASTCALL call_14_8_13_12(SH2_struct* sh)
{
  SH2movi(sh, 8, (13<<4)+12);
}
#define init_14_8_13_13
void FASTCALL call_14_8_13_13(SH2_struct* sh)
{
  SH2movi(sh, 8, (13<<4)+13);
}
#define init_14_8_13_14
void FASTCALL call_14_8_13_14(SH2_struct* sh)
{
  SH2movi(sh, 8, (13<<4)+14);
}
#define init_14_8_13_15
void FASTCALL call_14_8_13_15(SH2_struct* sh)
{
  SH2movi(sh, 8, (13<<4)+15);
}
#define init_14_8_14_0
void FASTCALL call_14_8_14_0(SH2_struct* sh)
{
  SH2movi(sh, 8, (14<<4)+0);
}
#define init_14_8_14_1
void FASTCALL call_14_8_14_1(SH2_struct* sh)
{
  SH2movi(sh, 8, (14<<4)+1);
}
#define init_14_8_14_2
void FASTCALL call_14_8_14_2(SH2_struct* sh)
{
  SH2movi(sh, 8, (14<<4)+2);
}
#define init_14_8_14_3
void FASTCALL call_14_8_14_3(SH2_struct* sh)
{
  SH2movi(sh, 8, (14<<4)+3);
}
#define init_14_8_14_4
void FASTCALL call_14_8_14_4(SH2_struct* sh)
{
  SH2movi(sh, 8, (14<<4)+4);
}
#define init_14_8_14_5
void FASTCALL call_14_8_14_5(SH2_struct* sh)
{
  SH2movi(sh, 8, (14<<4)+5);
}
#define init_14_8_14_6
void FASTCALL call_14_8_14_6(SH2_struct* sh)
{
  SH2movi(sh, 8, (14<<4)+6);
}
#define init_14_8_14_7
void FASTCALL call_14_8_14_7(SH2_struct* sh)
{
  SH2movi(sh, 8, (14<<4)+7);
}
#define init_14_8_14_8
void FASTCALL call_14_8_14_8(SH2_struct* sh)
{
  SH2movi(sh, 8, (14<<4)+8);
}
#define init_14_8_14_9
void FASTCALL call_14_8_14_9(SH2_struct* sh)
{
  SH2movi(sh, 8, (14<<4)+9);
}
#define init_14_8_14_10
void FASTCALL call_14_8_14_10(SH2_struct* sh)
{
  SH2movi(sh, 8, (14<<4)+10);
}
#define init_14_8_14_11
void FASTCALL call_14_8_14_11(SH2_struct* sh)
{
  SH2movi(sh, 8, (14<<4)+11);
}
#define init_14_8_14_12
void FASTCALL call_14_8_14_12(SH2_struct* sh)
{
  SH2movi(sh, 8, (14<<4)+12);
}
#define init_14_8_14_13
void FASTCALL call_14_8_14_13(SH2_struct* sh)
{
  SH2movi(sh, 8, (14<<4)+13);
}
#define init_14_8_14_14
void FASTCALL call_14_8_14_14(SH2_struct* sh)
{
  SH2movi(sh, 8, (14<<4)+14);
}
#define init_14_8_14_15
void FASTCALL call_14_8_14_15(SH2_struct* sh)
{
  SH2movi(sh, 8, (14<<4)+15);
}
#define init_14_8_15_0
void FASTCALL call_14_8_15_0(SH2_struct* sh)
{
  SH2movi(sh, 8, (15<<4)+0);
}
#define init_14_8_15_1
void FASTCALL call_14_8_15_1(SH2_struct* sh)
{
  SH2movi(sh, 8, (15<<4)+1);
}
#define init_14_8_15_2
void FASTCALL call_14_8_15_2(SH2_struct* sh)
{
  SH2movi(sh, 8, (15<<4)+2);
}
#define init_14_8_15_3
void FASTCALL call_14_8_15_3(SH2_struct* sh)
{
  SH2movi(sh, 8, (15<<4)+3);
}
#define init_14_8_15_4
void FASTCALL call_14_8_15_4(SH2_struct* sh)
{
  SH2movi(sh, 8, (15<<4)+4);
}
#define init_14_8_15_5
void FASTCALL call_14_8_15_5(SH2_struct* sh)
{
  SH2movi(sh, 8, (15<<4)+5);
}
#define init_14_8_15_6
void FASTCALL call_14_8_15_6(SH2_struct* sh)
{
  SH2movi(sh, 8, (15<<4)+6);
}
#define init_14_8_15_7
void FASTCALL call_14_8_15_7(SH2_struct* sh)
{
  SH2movi(sh, 8, (15<<4)+7);
}
#define init_14_8_15_8
void FASTCALL call_14_8_15_8(SH2_struct* sh)
{
  SH2movi(sh, 8, (15<<4)+8);
}
#define init_14_8_15_9
void FASTCALL call_14_8_15_9(SH2_struct* sh)
{
  SH2movi(sh, 8, (15<<4)+9);
}
#define init_14_8_15_10
void FASTCALL call_14_8_15_10(SH2_struct* sh)
{
  SH2movi(sh, 8, (15<<4)+10);
}
#define init_14_8_15_11
void FASTCALL call_14_8_15_11(SH2_struct* sh)
{
  SH2movi(sh, 8, (15<<4)+11);
}
#define init_14_8_15_12
void FASTCALL call_14_8_15_12(SH2_struct* sh)
{
  SH2movi(sh, 8, (15<<4)+12);
}
#define init_14_8_15_13
void FASTCALL call_14_8_15_13(SH2_struct* sh)
{
  SH2movi(sh, 8, (15<<4)+13);
}
#define init_14_8_15_14
void FASTCALL call_14_8_15_14(SH2_struct* sh)
{
  SH2movi(sh, 8, (15<<4)+14);
}
#define init_14_8_15_15
void FASTCALL call_14_8_15_15(SH2_struct* sh)
{
  SH2movi(sh, 8, (15<<4)+15);
}
#define init_14_9_0_0
void FASTCALL call_14_9_0_0(SH2_struct* sh)
{
  SH2movi(sh, 9, (0<<4)+0);
}
#define init_14_9_0_1
void FASTCALL call_14_9_0_1(SH2_struct* sh)
{
  SH2movi(sh, 9, (0<<4)+1);
}
#define init_14_9_0_2
void FASTCALL call_14_9_0_2(SH2_struct* sh)
{
  SH2movi(sh, 9, (0<<4)+2);
}
#define init_14_9_0_3
void FASTCALL call_14_9_0_3(SH2_struct* sh)
{
  SH2movi(sh, 9, (0<<4)+3);
}
#define init_14_9_0_4
void FASTCALL call_14_9_0_4(SH2_struct* sh)
{
  SH2movi(sh, 9, (0<<4)+4);
}
#define init_14_9_0_5
void FASTCALL call_14_9_0_5(SH2_struct* sh)
{
  SH2movi(sh, 9, (0<<4)+5);
}
#define init_14_9_0_6
void FASTCALL call_14_9_0_6(SH2_struct* sh)
{
  SH2movi(sh, 9, (0<<4)+6);
}
#define init_14_9_0_7
void FASTCALL call_14_9_0_7(SH2_struct* sh)
{
  SH2movi(sh, 9, (0<<4)+7);
}
#define init_14_9_0_8
void FASTCALL call_14_9_0_8(SH2_struct* sh)
{
  SH2movi(sh, 9, (0<<4)+8);
}
#define init_14_9_0_9
void FASTCALL call_14_9_0_9(SH2_struct* sh)
{
  SH2movi(sh, 9, (0<<4)+9);
}
#define init_14_9_0_10
void FASTCALL call_14_9_0_10(SH2_struct* sh)
{
  SH2movi(sh, 9, (0<<4)+10);
}
#define init_14_9_0_11
void FASTCALL call_14_9_0_11(SH2_struct* sh)
{
  SH2movi(sh, 9, (0<<4)+11);
}
#define init_14_9_0_12
void FASTCALL call_14_9_0_12(SH2_struct* sh)
{
  SH2movi(sh, 9, (0<<4)+12);
}
#define init_14_9_0_13
void FASTCALL call_14_9_0_13(SH2_struct* sh)
{
  SH2movi(sh, 9, (0<<4)+13);
}
#define init_14_9_0_14
void FASTCALL call_14_9_0_14(SH2_struct* sh)
{
  SH2movi(sh, 9, (0<<4)+14);
}
#define init_14_9_0_15
void FASTCALL call_14_9_0_15(SH2_struct* sh)
{
  SH2movi(sh, 9, (0<<4)+15);
}
#define init_14_9_1_0
void FASTCALL call_14_9_1_0(SH2_struct* sh)
{
  SH2movi(sh, 9, (1<<4)+0);
}
#define init_14_9_1_1
void FASTCALL call_14_9_1_1(SH2_struct* sh)
{
  SH2movi(sh, 9, (1<<4)+1);
}
#define init_14_9_1_2
void FASTCALL call_14_9_1_2(SH2_struct* sh)
{
  SH2movi(sh, 9, (1<<4)+2);
}
#define init_14_9_1_3
void FASTCALL call_14_9_1_3(SH2_struct* sh)
{
  SH2movi(sh, 9, (1<<4)+3);
}
#define init_14_9_1_4
void FASTCALL call_14_9_1_4(SH2_struct* sh)
{
  SH2movi(sh, 9, (1<<4)+4);
}
#define init_14_9_1_5
void FASTCALL call_14_9_1_5(SH2_struct* sh)
{
  SH2movi(sh, 9, (1<<4)+5);
}
#define init_14_9_1_6
void FASTCALL call_14_9_1_6(SH2_struct* sh)
{
  SH2movi(sh, 9, (1<<4)+6);
}
#define init_14_9_1_7
void FASTCALL call_14_9_1_7(SH2_struct* sh)
{
  SH2movi(sh, 9, (1<<4)+7);
}
#define init_14_9_1_8
void FASTCALL call_14_9_1_8(SH2_struct* sh)
{
  SH2movi(sh, 9, (1<<4)+8);
}
#define init_14_9_1_9
void FASTCALL call_14_9_1_9(SH2_struct* sh)
{
  SH2movi(sh, 9, (1<<4)+9);
}
#define init_14_9_1_10
void FASTCALL call_14_9_1_10(SH2_struct* sh)
{
  SH2movi(sh, 9, (1<<4)+10);
}
#define init_14_9_1_11
void FASTCALL call_14_9_1_11(SH2_struct* sh)
{
  SH2movi(sh, 9, (1<<4)+11);
}
#define init_14_9_1_12
void FASTCALL call_14_9_1_12(SH2_struct* sh)
{
  SH2movi(sh, 9, (1<<4)+12);
}
#define init_14_9_1_13
void FASTCALL call_14_9_1_13(SH2_struct* sh)
{
  SH2movi(sh, 9, (1<<4)+13);
}
#define init_14_9_1_14
void FASTCALL call_14_9_1_14(SH2_struct* sh)
{
  SH2movi(sh, 9, (1<<4)+14);
}
#define init_14_9_1_15
void FASTCALL call_14_9_1_15(SH2_struct* sh)
{
  SH2movi(sh, 9, (1<<4)+15);
}
#define init_14_9_2_0
void FASTCALL call_14_9_2_0(SH2_struct* sh)
{
  SH2movi(sh, 9, (2<<4)+0);
}
#define init_14_9_2_1
void FASTCALL call_14_9_2_1(SH2_struct* sh)
{
  SH2movi(sh, 9, (2<<4)+1);
}
#define init_14_9_2_2
void FASTCALL call_14_9_2_2(SH2_struct* sh)
{
  SH2movi(sh, 9, (2<<4)+2);
}
#define init_14_9_2_3
void FASTCALL call_14_9_2_3(SH2_struct* sh)
{
  SH2movi(sh, 9, (2<<4)+3);
}
#define init_14_9_2_4
void FASTCALL call_14_9_2_4(SH2_struct* sh)
{
  SH2movi(sh, 9, (2<<4)+4);
}
#define init_14_9_2_5
void FASTCALL call_14_9_2_5(SH2_struct* sh)
{
  SH2movi(sh, 9, (2<<4)+5);
}
#define init_14_9_2_6
void FASTCALL call_14_9_2_6(SH2_struct* sh)
{
  SH2movi(sh, 9, (2<<4)+6);
}
#define init_14_9_2_7
void FASTCALL call_14_9_2_7(SH2_struct* sh)
{
  SH2movi(sh, 9, (2<<4)+7);
}
#define init_14_9_2_8
void FASTCALL call_14_9_2_8(SH2_struct* sh)
{
  SH2movi(sh, 9, (2<<4)+8);
}
#define init_14_9_2_9
void FASTCALL call_14_9_2_9(SH2_struct* sh)
{
  SH2movi(sh, 9, (2<<4)+9);
}
#define init_14_9_2_10
void FASTCALL call_14_9_2_10(SH2_struct* sh)
{
  SH2movi(sh, 9, (2<<4)+10);
}
#define init_14_9_2_11
void FASTCALL call_14_9_2_11(SH2_struct* sh)
{
  SH2movi(sh, 9, (2<<4)+11);
}
#define init_14_9_2_12
void FASTCALL call_14_9_2_12(SH2_struct* sh)
{
  SH2movi(sh, 9, (2<<4)+12);
}
#define init_14_9_2_13
void FASTCALL call_14_9_2_13(SH2_struct* sh)
{
  SH2movi(sh, 9, (2<<4)+13);
}
#define init_14_9_2_14
void FASTCALL call_14_9_2_14(SH2_struct* sh)
{
  SH2movi(sh, 9, (2<<4)+14);
}
#define init_14_9_2_15
void FASTCALL call_14_9_2_15(SH2_struct* sh)
{
  SH2movi(sh, 9, (2<<4)+15);
}
#define init_14_9_3_0
void FASTCALL call_14_9_3_0(SH2_struct* sh)
{
  SH2movi(sh, 9, (3<<4)+0);
}
#define init_14_9_3_1
void FASTCALL call_14_9_3_1(SH2_struct* sh)
{
  SH2movi(sh, 9, (3<<4)+1);
}
#define init_14_9_3_2
void FASTCALL call_14_9_3_2(SH2_struct* sh)
{
  SH2movi(sh, 9, (3<<4)+2);
}
#define init_14_9_3_3
void FASTCALL call_14_9_3_3(SH2_struct* sh)
{
  SH2movi(sh, 9, (3<<4)+3);
}
#define init_14_9_3_4
void FASTCALL call_14_9_3_4(SH2_struct* sh)
{
  SH2movi(sh, 9, (3<<4)+4);
}
#define init_14_9_3_5
void FASTCALL call_14_9_3_5(SH2_struct* sh)
{
  SH2movi(sh, 9, (3<<4)+5);
}
#define init_14_9_3_6
void FASTCALL call_14_9_3_6(SH2_struct* sh)
{
  SH2movi(sh, 9, (3<<4)+6);
}
#define init_14_9_3_7
void FASTCALL call_14_9_3_7(SH2_struct* sh)
{
  SH2movi(sh, 9, (3<<4)+7);
}
#define init_14_9_3_8
void FASTCALL call_14_9_3_8(SH2_struct* sh)
{
  SH2movi(sh, 9, (3<<4)+8);
}
#define init_14_9_3_9
void FASTCALL call_14_9_3_9(SH2_struct* sh)
{
  SH2movi(sh, 9, (3<<4)+9);
}
#define init_14_9_3_10
void FASTCALL call_14_9_3_10(SH2_struct* sh)
{
  SH2movi(sh, 9, (3<<4)+10);
}
#define init_14_9_3_11
void FASTCALL call_14_9_3_11(SH2_struct* sh)
{
  SH2movi(sh, 9, (3<<4)+11);
}
#define init_14_9_3_12
void FASTCALL call_14_9_3_12(SH2_struct* sh)
{
  SH2movi(sh, 9, (3<<4)+12);
}
#define init_14_9_3_13
void FASTCALL call_14_9_3_13(SH2_struct* sh)
{
  SH2movi(sh, 9, (3<<4)+13);
}
#define init_14_9_3_14
void FASTCALL call_14_9_3_14(SH2_struct* sh)
{
  SH2movi(sh, 9, (3<<4)+14);
}
#define init_14_9_3_15
void FASTCALL call_14_9_3_15(SH2_struct* sh)
{
  SH2movi(sh, 9, (3<<4)+15);
}
#define init_14_9_4_0
void FASTCALL call_14_9_4_0(SH2_struct* sh)
{
  SH2movi(sh, 9, (4<<4)+0);
}
#define init_14_9_4_1
void FASTCALL call_14_9_4_1(SH2_struct* sh)
{
  SH2movi(sh, 9, (4<<4)+1);
}
#define init_14_9_4_2
void FASTCALL call_14_9_4_2(SH2_struct* sh)
{
  SH2movi(sh, 9, (4<<4)+2);
}
#define init_14_9_4_3
void FASTCALL call_14_9_4_3(SH2_struct* sh)
{
  SH2movi(sh, 9, (4<<4)+3);
}
#define init_14_9_4_4
void FASTCALL call_14_9_4_4(SH2_struct* sh)
{
  SH2movi(sh, 9, (4<<4)+4);
}
#define init_14_9_4_5
void FASTCALL call_14_9_4_5(SH2_struct* sh)
{
  SH2movi(sh, 9, (4<<4)+5);
}
#define init_14_9_4_6
void FASTCALL call_14_9_4_6(SH2_struct* sh)
{
  SH2movi(sh, 9, (4<<4)+6);
}
#define init_14_9_4_7
void FASTCALL call_14_9_4_7(SH2_struct* sh)
{
  SH2movi(sh, 9, (4<<4)+7);
}
#define init_14_9_4_8
void FASTCALL call_14_9_4_8(SH2_struct* sh)
{
  SH2movi(sh, 9, (4<<4)+8);
}
#define init_14_9_4_9
void FASTCALL call_14_9_4_9(SH2_struct* sh)
{
  SH2movi(sh, 9, (4<<4)+9);
}
#define init_14_9_4_10
void FASTCALL call_14_9_4_10(SH2_struct* sh)
{
  SH2movi(sh, 9, (4<<4)+10);
}
#define init_14_9_4_11
void FASTCALL call_14_9_4_11(SH2_struct* sh)
{
  SH2movi(sh, 9, (4<<4)+11);
}
#define init_14_9_4_12
void FASTCALL call_14_9_4_12(SH2_struct* sh)
{
  SH2movi(sh, 9, (4<<4)+12);
}
#define init_14_9_4_13
void FASTCALL call_14_9_4_13(SH2_struct* sh)
{
  SH2movi(sh, 9, (4<<4)+13);
}
#define init_14_9_4_14
void FASTCALL call_14_9_4_14(SH2_struct* sh)
{
  SH2movi(sh, 9, (4<<4)+14);
}
#define init_14_9_4_15
void FASTCALL call_14_9_4_15(SH2_struct* sh)
{
  SH2movi(sh, 9, (4<<4)+15);
}
#define init_14_9_5_0
void FASTCALL call_14_9_5_0(SH2_struct* sh)
{
  SH2movi(sh, 9, (5<<4)+0);
}
#define init_14_9_5_1
void FASTCALL call_14_9_5_1(SH2_struct* sh)
{
  SH2movi(sh, 9, (5<<4)+1);
}
#define init_14_9_5_2
void FASTCALL call_14_9_5_2(SH2_struct* sh)
{
  SH2movi(sh, 9, (5<<4)+2);
}
#define init_14_9_5_3
void FASTCALL call_14_9_5_3(SH2_struct* sh)
{
  SH2movi(sh, 9, (5<<4)+3);
}
#define init_14_9_5_4
void FASTCALL call_14_9_5_4(SH2_struct* sh)
{
  SH2movi(sh, 9, (5<<4)+4);
}
#define init_14_9_5_5
void FASTCALL call_14_9_5_5(SH2_struct* sh)
{
  SH2movi(sh, 9, (5<<4)+5);
}
#define init_14_9_5_6
void FASTCALL call_14_9_5_6(SH2_struct* sh)
{
  SH2movi(sh, 9, (5<<4)+6);
}
#define init_14_9_5_7
void FASTCALL call_14_9_5_7(SH2_struct* sh)
{
  SH2movi(sh, 9, (5<<4)+7);
}
#define init_14_9_5_8
void FASTCALL call_14_9_5_8(SH2_struct* sh)
{
  SH2movi(sh, 9, (5<<4)+8);
}
#define init_14_9_5_9
void FASTCALL call_14_9_5_9(SH2_struct* sh)
{
  SH2movi(sh, 9, (5<<4)+9);
}
#define init_14_9_5_10
void FASTCALL call_14_9_5_10(SH2_struct* sh)
{
  SH2movi(sh, 9, (5<<4)+10);
}
#define init_14_9_5_11
void FASTCALL call_14_9_5_11(SH2_struct* sh)
{
  SH2movi(sh, 9, (5<<4)+11);
}
#define init_14_9_5_12
void FASTCALL call_14_9_5_12(SH2_struct* sh)
{
  SH2movi(sh, 9, (5<<4)+12);
}
#define init_14_9_5_13
void FASTCALL call_14_9_5_13(SH2_struct* sh)
{
  SH2movi(sh, 9, (5<<4)+13);
}
#define init_14_9_5_14
void FASTCALL call_14_9_5_14(SH2_struct* sh)
{
  SH2movi(sh, 9, (5<<4)+14);
}
#define init_14_9_5_15
void FASTCALL call_14_9_5_15(SH2_struct* sh)
{
  SH2movi(sh, 9, (5<<4)+15);
}
#define init_14_9_6_0
void FASTCALL call_14_9_6_0(SH2_struct* sh)
{
  SH2movi(sh, 9, (6<<4)+0);
}
#define init_14_9_6_1
void FASTCALL call_14_9_6_1(SH2_struct* sh)
{
  SH2movi(sh, 9, (6<<4)+1);
}
#define init_14_9_6_2
void FASTCALL call_14_9_6_2(SH2_struct* sh)
{
  SH2movi(sh, 9, (6<<4)+2);
}
#define init_14_9_6_3
void FASTCALL call_14_9_6_3(SH2_struct* sh)
{
  SH2movi(sh, 9, (6<<4)+3);
}
#define init_14_9_6_4
void FASTCALL call_14_9_6_4(SH2_struct* sh)
{
  SH2movi(sh, 9, (6<<4)+4);
}
#define init_14_9_6_5
void FASTCALL call_14_9_6_5(SH2_struct* sh)
{
  SH2movi(sh, 9, (6<<4)+5);
}
#define init_14_9_6_6
void FASTCALL call_14_9_6_6(SH2_struct* sh)
{
  SH2movi(sh, 9, (6<<4)+6);
}
#define init_14_9_6_7
void FASTCALL call_14_9_6_7(SH2_struct* sh)
{
  SH2movi(sh, 9, (6<<4)+7);
}
#define init_14_9_6_8
void FASTCALL call_14_9_6_8(SH2_struct* sh)
{
  SH2movi(sh, 9, (6<<4)+8);
}
#define init_14_9_6_9
void FASTCALL call_14_9_6_9(SH2_struct* sh)
{
  SH2movi(sh, 9, (6<<4)+9);
}
#define init_14_9_6_10
void FASTCALL call_14_9_6_10(SH2_struct* sh)
{
  SH2movi(sh, 9, (6<<4)+10);
}
#define init_14_9_6_11
void FASTCALL call_14_9_6_11(SH2_struct* sh)
{
  SH2movi(sh, 9, (6<<4)+11);
}
#define init_14_9_6_12
void FASTCALL call_14_9_6_12(SH2_struct* sh)
{
  SH2movi(sh, 9, (6<<4)+12);
}
#define init_14_9_6_13
void FASTCALL call_14_9_6_13(SH2_struct* sh)
{
  SH2movi(sh, 9, (6<<4)+13);
}
#define init_14_9_6_14
void FASTCALL call_14_9_6_14(SH2_struct* sh)
{
  SH2movi(sh, 9, (6<<4)+14);
}
#define init_14_9_6_15
void FASTCALL call_14_9_6_15(SH2_struct* sh)
{
  SH2movi(sh, 9, (6<<4)+15);
}
#define init_14_9_7_0
void FASTCALL call_14_9_7_0(SH2_struct* sh)
{
  SH2movi(sh, 9, (7<<4)+0);
}
#define init_14_9_7_1
void FASTCALL call_14_9_7_1(SH2_struct* sh)
{
  SH2movi(sh, 9, (7<<4)+1);
}
#define init_14_9_7_2
void FASTCALL call_14_9_7_2(SH2_struct* sh)
{
  SH2movi(sh, 9, (7<<4)+2);
}
#define init_14_9_7_3
void FASTCALL call_14_9_7_3(SH2_struct* sh)
{
  SH2movi(sh, 9, (7<<4)+3);
}
#define init_14_9_7_4
void FASTCALL call_14_9_7_4(SH2_struct* sh)
{
  SH2movi(sh, 9, (7<<4)+4);
}
#define init_14_9_7_5
void FASTCALL call_14_9_7_5(SH2_struct* sh)
{
  SH2movi(sh, 9, (7<<4)+5);
}
#define init_14_9_7_6
void FASTCALL call_14_9_7_6(SH2_struct* sh)
{
  SH2movi(sh, 9, (7<<4)+6);
}
#define init_14_9_7_7
void FASTCALL call_14_9_7_7(SH2_struct* sh)
{
  SH2movi(sh, 9, (7<<4)+7);
}
#define init_14_9_7_8
void FASTCALL call_14_9_7_8(SH2_struct* sh)
{
  SH2movi(sh, 9, (7<<4)+8);
}
#define init_14_9_7_9
void FASTCALL call_14_9_7_9(SH2_struct* sh)
{
  SH2movi(sh, 9, (7<<4)+9);
}
#define init_14_9_7_10
void FASTCALL call_14_9_7_10(SH2_struct* sh)
{
  SH2movi(sh, 9, (7<<4)+10);
}
#define init_14_9_7_11
void FASTCALL call_14_9_7_11(SH2_struct* sh)
{
  SH2movi(sh, 9, (7<<4)+11);
}
#define init_14_9_7_12
void FASTCALL call_14_9_7_12(SH2_struct* sh)
{
  SH2movi(sh, 9, (7<<4)+12);
}
#define init_14_9_7_13
void FASTCALL call_14_9_7_13(SH2_struct* sh)
{
  SH2movi(sh, 9, (7<<4)+13);
}
#define init_14_9_7_14
void FASTCALL call_14_9_7_14(SH2_struct* sh)
{
  SH2movi(sh, 9, (7<<4)+14);
}
#define init_14_9_7_15
void FASTCALL call_14_9_7_15(SH2_struct* sh)
{
  SH2movi(sh, 9, (7<<4)+15);
}
#define init_14_9_8_0
void FASTCALL call_14_9_8_0(SH2_struct* sh)
{
  SH2movi(sh, 9, (8<<4)+0);
}
#define init_14_9_8_1
void FASTCALL call_14_9_8_1(SH2_struct* sh)
{
  SH2movi(sh, 9, (8<<4)+1);
}
#define init_14_9_8_2
void FASTCALL call_14_9_8_2(SH2_struct* sh)
{
  SH2movi(sh, 9, (8<<4)+2);
}
#define init_14_9_8_3
void FASTCALL call_14_9_8_3(SH2_struct* sh)
{
  SH2movi(sh, 9, (8<<4)+3);
}
#define init_14_9_8_4
void FASTCALL call_14_9_8_4(SH2_struct* sh)
{
  SH2movi(sh, 9, (8<<4)+4);
}
#define init_14_9_8_5
void FASTCALL call_14_9_8_5(SH2_struct* sh)
{
  SH2movi(sh, 9, (8<<4)+5);
}
#define init_14_9_8_6
void FASTCALL call_14_9_8_6(SH2_struct* sh)
{
  SH2movi(sh, 9, (8<<4)+6);
}
#define init_14_9_8_7
void FASTCALL call_14_9_8_7(SH2_struct* sh)
{
  SH2movi(sh, 9, (8<<4)+7);
}
#define init_14_9_8_8
void FASTCALL call_14_9_8_8(SH2_struct* sh)
{
  SH2movi(sh, 9, (8<<4)+8);
}
#define init_14_9_8_9
void FASTCALL call_14_9_8_9(SH2_struct* sh)
{
  SH2movi(sh, 9, (8<<4)+9);
}
#define init_14_9_8_10
void FASTCALL call_14_9_8_10(SH2_struct* sh)
{
  SH2movi(sh, 9, (8<<4)+10);
}
#define init_14_9_8_11
void FASTCALL call_14_9_8_11(SH2_struct* sh)
{
  SH2movi(sh, 9, (8<<4)+11);
}
#define init_14_9_8_12
void FASTCALL call_14_9_8_12(SH2_struct* sh)
{
  SH2movi(sh, 9, (8<<4)+12);
}
#define init_14_9_8_13
void FASTCALL call_14_9_8_13(SH2_struct* sh)
{
  SH2movi(sh, 9, (8<<4)+13);
}
#define init_14_9_8_14
void FASTCALL call_14_9_8_14(SH2_struct* sh)
{
  SH2movi(sh, 9, (8<<4)+14);
}
#define init_14_9_8_15
void FASTCALL call_14_9_8_15(SH2_struct* sh)
{
  SH2movi(sh, 9, (8<<4)+15);
}
#define init_14_9_9_0
void FASTCALL call_14_9_9_0(SH2_struct* sh)
{
  SH2movi(sh, 9, (9<<4)+0);
}
#define init_14_9_9_1
void FASTCALL call_14_9_9_1(SH2_struct* sh)
{
  SH2movi(sh, 9, (9<<4)+1);
}
#define init_14_9_9_2
void FASTCALL call_14_9_9_2(SH2_struct* sh)
{
  SH2movi(sh, 9, (9<<4)+2);
}
#define init_14_9_9_3
void FASTCALL call_14_9_9_3(SH2_struct* sh)
{
  SH2movi(sh, 9, (9<<4)+3);
}
#define init_14_9_9_4
void FASTCALL call_14_9_9_4(SH2_struct* sh)
{
  SH2movi(sh, 9, (9<<4)+4);
}
#define init_14_9_9_5
void FASTCALL call_14_9_9_5(SH2_struct* sh)
{
  SH2movi(sh, 9, (9<<4)+5);
}
#define init_14_9_9_6
void FASTCALL call_14_9_9_6(SH2_struct* sh)
{
  SH2movi(sh, 9, (9<<4)+6);
}
#define init_14_9_9_7
void FASTCALL call_14_9_9_7(SH2_struct* sh)
{
  SH2movi(sh, 9, (9<<4)+7);
}
#define init_14_9_9_8
void FASTCALL call_14_9_9_8(SH2_struct* sh)
{
  SH2movi(sh, 9, (9<<4)+8);
}
#define init_14_9_9_9
void FASTCALL call_14_9_9_9(SH2_struct* sh)
{
  SH2movi(sh, 9, (9<<4)+9);
}
#define init_14_9_9_10
void FASTCALL call_14_9_9_10(SH2_struct* sh)
{
  SH2movi(sh, 9, (9<<4)+10);
}
#define init_14_9_9_11
void FASTCALL call_14_9_9_11(SH2_struct* sh)
{
  SH2movi(sh, 9, (9<<4)+11);
}
#define init_14_9_9_12
void FASTCALL call_14_9_9_12(SH2_struct* sh)
{
  SH2movi(sh, 9, (9<<4)+12);
}
#define init_14_9_9_13
void FASTCALL call_14_9_9_13(SH2_struct* sh)
{
  SH2movi(sh, 9, (9<<4)+13);
}
#define init_14_9_9_14
void FASTCALL call_14_9_9_14(SH2_struct* sh)
{
  SH2movi(sh, 9, (9<<4)+14);
}
#define init_14_9_9_15
void FASTCALL call_14_9_9_15(SH2_struct* sh)
{
  SH2movi(sh, 9, (9<<4)+15);
}
#define init_14_9_10_0
void FASTCALL call_14_9_10_0(SH2_struct* sh)
{
  SH2movi(sh, 9, (10<<4)+0);
}
#define init_14_9_10_1
void FASTCALL call_14_9_10_1(SH2_struct* sh)
{
  SH2movi(sh, 9, (10<<4)+1);
}
#define init_14_9_10_2
void FASTCALL call_14_9_10_2(SH2_struct* sh)
{
  SH2movi(sh, 9, (10<<4)+2);
}
#define init_14_9_10_3
void FASTCALL call_14_9_10_3(SH2_struct* sh)
{
  SH2movi(sh, 9, (10<<4)+3);
}
#define init_14_9_10_4
void FASTCALL call_14_9_10_4(SH2_struct* sh)
{
  SH2movi(sh, 9, (10<<4)+4);
}
#define init_14_9_10_5
void FASTCALL call_14_9_10_5(SH2_struct* sh)
{
  SH2movi(sh, 9, (10<<4)+5);
}
#define init_14_9_10_6
void FASTCALL call_14_9_10_6(SH2_struct* sh)
{
  SH2movi(sh, 9, (10<<4)+6);
}
#define init_14_9_10_7
void FASTCALL call_14_9_10_7(SH2_struct* sh)
{
  SH2movi(sh, 9, (10<<4)+7);
}
#define init_14_9_10_8
void FASTCALL call_14_9_10_8(SH2_struct* sh)
{
  SH2movi(sh, 9, (10<<4)+8);
}
#define init_14_9_10_9
void FASTCALL call_14_9_10_9(SH2_struct* sh)
{
  SH2movi(sh, 9, (10<<4)+9);
}
#define init_14_9_10_10
void FASTCALL call_14_9_10_10(SH2_struct* sh)
{
  SH2movi(sh, 9, (10<<4)+10);
}
#define init_14_9_10_11
void FASTCALL call_14_9_10_11(SH2_struct* sh)
{
  SH2movi(sh, 9, (10<<4)+11);
}
#define init_14_9_10_12
void FASTCALL call_14_9_10_12(SH2_struct* sh)
{
  SH2movi(sh, 9, (10<<4)+12);
}
#define init_14_9_10_13
void FASTCALL call_14_9_10_13(SH2_struct* sh)
{
  SH2movi(sh, 9, (10<<4)+13);
}
#define init_14_9_10_14
void FASTCALL call_14_9_10_14(SH2_struct* sh)
{
  SH2movi(sh, 9, (10<<4)+14);
}
#define init_14_9_10_15
void FASTCALL call_14_9_10_15(SH2_struct* sh)
{
  SH2movi(sh, 9, (10<<4)+15);
}
#define init_14_9_11_0
void FASTCALL call_14_9_11_0(SH2_struct* sh)
{
  SH2movi(sh, 9, (11<<4)+0);
}
#define init_14_9_11_1
void FASTCALL call_14_9_11_1(SH2_struct* sh)
{
  SH2movi(sh, 9, (11<<4)+1);
}
#define init_14_9_11_2
void FASTCALL call_14_9_11_2(SH2_struct* sh)
{
  SH2movi(sh, 9, (11<<4)+2);
}
#define init_14_9_11_3
void FASTCALL call_14_9_11_3(SH2_struct* sh)
{
  SH2movi(sh, 9, (11<<4)+3);
}
#define init_14_9_11_4
void FASTCALL call_14_9_11_4(SH2_struct* sh)
{
  SH2movi(sh, 9, (11<<4)+4);
}
#define init_14_9_11_5
void FASTCALL call_14_9_11_5(SH2_struct* sh)
{
  SH2movi(sh, 9, (11<<4)+5);
}
#define init_14_9_11_6
void FASTCALL call_14_9_11_6(SH2_struct* sh)
{
  SH2movi(sh, 9, (11<<4)+6);
}
#define init_14_9_11_7
void FASTCALL call_14_9_11_7(SH2_struct* sh)
{
  SH2movi(sh, 9, (11<<4)+7);
}
#define init_14_9_11_8
void FASTCALL call_14_9_11_8(SH2_struct* sh)
{
  SH2movi(sh, 9, (11<<4)+8);
}
#define init_14_9_11_9
void FASTCALL call_14_9_11_9(SH2_struct* sh)
{
  SH2movi(sh, 9, (11<<4)+9);
}
#define init_14_9_11_10
void FASTCALL call_14_9_11_10(SH2_struct* sh)
{
  SH2movi(sh, 9, (11<<4)+10);
}
#define init_14_9_11_11
void FASTCALL call_14_9_11_11(SH2_struct* sh)
{
  SH2movi(sh, 9, (11<<4)+11);
}
#define init_14_9_11_12
void FASTCALL call_14_9_11_12(SH2_struct* sh)
{
  SH2movi(sh, 9, (11<<4)+12);
}
#define init_14_9_11_13
void FASTCALL call_14_9_11_13(SH2_struct* sh)
{
  SH2movi(sh, 9, (11<<4)+13);
}
#define init_14_9_11_14
void FASTCALL call_14_9_11_14(SH2_struct* sh)
{
  SH2movi(sh, 9, (11<<4)+14);
}
#define init_14_9_11_15
void FASTCALL call_14_9_11_15(SH2_struct* sh)
{
  SH2movi(sh, 9, (11<<4)+15);
}
#define init_14_9_12_0
void FASTCALL call_14_9_12_0(SH2_struct* sh)
{
  SH2movi(sh, 9, (12<<4)+0);
}
#define init_14_9_12_1
void FASTCALL call_14_9_12_1(SH2_struct* sh)
{
  SH2movi(sh, 9, (12<<4)+1);
}
#define init_14_9_12_2
void FASTCALL call_14_9_12_2(SH2_struct* sh)
{
  SH2movi(sh, 9, (12<<4)+2);
}
#define init_14_9_12_3
void FASTCALL call_14_9_12_3(SH2_struct* sh)
{
  SH2movi(sh, 9, (12<<4)+3);
}
#define init_14_9_12_4
void FASTCALL call_14_9_12_4(SH2_struct* sh)
{
  SH2movi(sh, 9, (12<<4)+4);
}
#define init_14_9_12_5
void FASTCALL call_14_9_12_5(SH2_struct* sh)
{
  SH2movi(sh, 9, (12<<4)+5);
}
#define init_14_9_12_6
void FASTCALL call_14_9_12_6(SH2_struct* sh)
{
  SH2movi(sh, 9, (12<<4)+6);
}
#define init_14_9_12_7
void FASTCALL call_14_9_12_7(SH2_struct* sh)
{
  SH2movi(sh, 9, (12<<4)+7);
}
#define init_14_9_12_8
void FASTCALL call_14_9_12_8(SH2_struct* sh)
{
  SH2movi(sh, 9, (12<<4)+8);
}
#define init_14_9_12_9
void FASTCALL call_14_9_12_9(SH2_struct* sh)
{
  SH2movi(sh, 9, (12<<4)+9);
}
#define init_14_9_12_10
void FASTCALL call_14_9_12_10(SH2_struct* sh)
{
  SH2movi(sh, 9, (12<<4)+10);
}
#define init_14_9_12_11
void FASTCALL call_14_9_12_11(SH2_struct* sh)
{
  SH2movi(sh, 9, (12<<4)+11);
}
#define init_14_9_12_12
void FASTCALL call_14_9_12_12(SH2_struct* sh)
{
  SH2movi(sh, 9, (12<<4)+12);
}
#define init_14_9_12_13
void FASTCALL call_14_9_12_13(SH2_struct* sh)
{
  SH2movi(sh, 9, (12<<4)+13);
}
#define init_14_9_12_14
void FASTCALL call_14_9_12_14(SH2_struct* sh)
{
  SH2movi(sh, 9, (12<<4)+14);
}
#define init_14_9_12_15
void FASTCALL call_14_9_12_15(SH2_struct* sh)
{
  SH2movi(sh, 9, (12<<4)+15);
}
#define init_14_9_13_0
void FASTCALL call_14_9_13_0(SH2_struct* sh)
{
  SH2movi(sh, 9, (13<<4)+0);
}
#define init_14_9_13_1
void FASTCALL call_14_9_13_1(SH2_struct* sh)
{
  SH2movi(sh, 9, (13<<4)+1);
}
#define init_14_9_13_2
void FASTCALL call_14_9_13_2(SH2_struct* sh)
{
  SH2movi(sh, 9, (13<<4)+2);
}
#define init_14_9_13_3
void FASTCALL call_14_9_13_3(SH2_struct* sh)
{
  SH2movi(sh, 9, (13<<4)+3);
}
#define init_14_9_13_4
void FASTCALL call_14_9_13_4(SH2_struct* sh)
{
  SH2movi(sh, 9, (13<<4)+4);
}
#define init_14_9_13_5
void FASTCALL call_14_9_13_5(SH2_struct* sh)
{
  SH2movi(sh, 9, (13<<4)+5);
}
#define init_14_9_13_6
void FASTCALL call_14_9_13_6(SH2_struct* sh)
{
  SH2movi(sh, 9, (13<<4)+6);
}
#define init_14_9_13_7
void FASTCALL call_14_9_13_7(SH2_struct* sh)
{
  SH2movi(sh, 9, (13<<4)+7);
}
#define init_14_9_13_8
void FASTCALL call_14_9_13_8(SH2_struct* sh)
{
  SH2movi(sh, 9, (13<<4)+8);
}
#define init_14_9_13_9
void FASTCALL call_14_9_13_9(SH2_struct* sh)
{
  SH2movi(sh, 9, (13<<4)+9);
}
#define init_14_9_13_10
void FASTCALL call_14_9_13_10(SH2_struct* sh)
{
  SH2movi(sh, 9, (13<<4)+10);
}
#define init_14_9_13_11
void FASTCALL call_14_9_13_11(SH2_struct* sh)
{
  SH2movi(sh, 9, (13<<4)+11);
}
#define init_14_9_13_12
void FASTCALL call_14_9_13_12(SH2_struct* sh)
{
  SH2movi(sh, 9, (13<<4)+12);
}
#define init_14_9_13_13
void FASTCALL call_14_9_13_13(SH2_struct* sh)
{
  SH2movi(sh, 9, (13<<4)+13);
}
#define init_14_9_13_14
void FASTCALL call_14_9_13_14(SH2_struct* sh)
{
  SH2movi(sh, 9, (13<<4)+14);
}
#define init_14_9_13_15
void FASTCALL call_14_9_13_15(SH2_struct* sh)
{
  SH2movi(sh, 9, (13<<4)+15);
}
#define init_14_9_14_0
void FASTCALL call_14_9_14_0(SH2_struct* sh)
{
  SH2movi(sh, 9, (14<<4)+0);
}
#define init_14_9_14_1
void FASTCALL call_14_9_14_1(SH2_struct* sh)
{
  SH2movi(sh, 9, (14<<4)+1);
}
#define init_14_9_14_2
void FASTCALL call_14_9_14_2(SH2_struct* sh)
{
  SH2movi(sh, 9, (14<<4)+2);
}
#define init_14_9_14_3
void FASTCALL call_14_9_14_3(SH2_struct* sh)
{
  SH2movi(sh, 9, (14<<4)+3);
}
#define init_14_9_14_4
void FASTCALL call_14_9_14_4(SH2_struct* sh)
{
  SH2movi(sh, 9, (14<<4)+4);
}
#define init_14_9_14_5
void FASTCALL call_14_9_14_5(SH2_struct* sh)
{
  SH2movi(sh, 9, (14<<4)+5);
}
#define init_14_9_14_6
void FASTCALL call_14_9_14_6(SH2_struct* sh)
{
  SH2movi(sh, 9, (14<<4)+6);
}
#define init_14_9_14_7
void FASTCALL call_14_9_14_7(SH2_struct* sh)
{
  SH2movi(sh, 9, (14<<4)+7);
}
#define init_14_9_14_8
void FASTCALL call_14_9_14_8(SH2_struct* sh)
{
  SH2movi(sh, 9, (14<<4)+8);
}
#define init_14_9_14_9
void FASTCALL call_14_9_14_9(SH2_struct* sh)
{
  SH2movi(sh, 9, (14<<4)+9);
}
#define init_14_9_14_10
void FASTCALL call_14_9_14_10(SH2_struct* sh)
{
  SH2movi(sh, 9, (14<<4)+10);
}
#define init_14_9_14_11
void FASTCALL call_14_9_14_11(SH2_struct* sh)
{
  SH2movi(sh, 9, (14<<4)+11);
}
#define init_14_9_14_12
void FASTCALL call_14_9_14_12(SH2_struct* sh)
{
  SH2movi(sh, 9, (14<<4)+12);
}
#define init_14_9_14_13
void FASTCALL call_14_9_14_13(SH2_struct* sh)
{
  SH2movi(sh, 9, (14<<4)+13);
}
#define init_14_9_14_14
void FASTCALL call_14_9_14_14(SH2_struct* sh)
{
  SH2movi(sh, 9, (14<<4)+14);
}
#define init_14_9_14_15
void FASTCALL call_14_9_14_15(SH2_struct* sh)
{
  SH2movi(sh, 9, (14<<4)+15);
}
#define init_14_9_15_0
void FASTCALL call_14_9_15_0(SH2_struct* sh)
{
  SH2movi(sh, 9, (15<<4)+0);
}
#define init_14_9_15_1
void FASTCALL call_14_9_15_1(SH2_struct* sh)
{
  SH2movi(sh, 9, (15<<4)+1);
}
#define init_14_9_15_2
void FASTCALL call_14_9_15_2(SH2_struct* sh)
{
  SH2movi(sh, 9, (15<<4)+2);
}
#define init_14_9_15_3
void FASTCALL call_14_9_15_3(SH2_struct* sh)
{
  SH2movi(sh, 9, (15<<4)+3);
}
#define init_14_9_15_4
void FASTCALL call_14_9_15_4(SH2_struct* sh)
{
  SH2movi(sh, 9, (15<<4)+4);
}
#define init_14_9_15_5
void FASTCALL call_14_9_15_5(SH2_struct* sh)
{
  SH2movi(sh, 9, (15<<4)+5);
}
#define init_14_9_15_6
void FASTCALL call_14_9_15_6(SH2_struct* sh)
{
  SH2movi(sh, 9, (15<<4)+6);
}
#define init_14_9_15_7
void FASTCALL call_14_9_15_7(SH2_struct* sh)
{
  SH2movi(sh, 9, (15<<4)+7);
}
#define init_14_9_15_8
void FASTCALL call_14_9_15_8(SH2_struct* sh)
{
  SH2movi(sh, 9, (15<<4)+8);
}
#define init_14_9_15_9
void FASTCALL call_14_9_15_9(SH2_struct* sh)
{
  SH2movi(sh, 9, (15<<4)+9);
}
#define init_14_9_15_10
void FASTCALL call_14_9_15_10(SH2_struct* sh)
{
  SH2movi(sh, 9, (15<<4)+10);
}
#define init_14_9_15_11
void FASTCALL call_14_9_15_11(SH2_struct* sh)
{
  SH2movi(sh, 9, (15<<4)+11);
}
#define init_14_9_15_12
void FASTCALL call_14_9_15_12(SH2_struct* sh)
{
  SH2movi(sh, 9, (15<<4)+12);
}
#define init_14_9_15_13
void FASTCALL call_14_9_15_13(SH2_struct* sh)
{
  SH2movi(sh, 9, (15<<4)+13);
}
#define init_14_9_15_14
void FASTCALL call_14_9_15_14(SH2_struct* sh)
{
  SH2movi(sh, 9, (15<<4)+14);
}
#define init_14_9_15_15
void FASTCALL call_14_9_15_15(SH2_struct* sh)
{
  SH2movi(sh, 9, (15<<4)+15);
}
#define init_14_10_0_0
void FASTCALL call_14_10_0_0(SH2_struct* sh)
{
  SH2movi(sh, 10, (0<<4)+0);
}
#define init_14_10_0_1
void FASTCALL call_14_10_0_1(SH2_struct* sh)
{
  SH2movi(sh, 10, (0<<4)+1);
}
#define init_14_10_0_2
void FASTCALL call_14_10_0_2(SH2_struct* sh)
{
  SH2movi(sh, 10, (0<<4)+2);
}
#define init_14_10_0_3
void FASTCALL call_14_10_0_3(SH2_struct* sh)
{
  SH2movi(sh, 10, (0<<4)+3);
}
#define init_14_10_0_4
void FASTCALL call_14_10_0_4(SH2_struct* sh)
{
  SH2movi(sh, 10, (0<<4)+4);
}
#define init_14_10_0_5
void FASTCALL call_14_10_0_5(SH2_struct* sh)
{
  SH2movi(sh, 10, (0<<4)+5);
}
#define init_14_10_0_6
void FASTCALL call_14_10_0_6(SH2_struct* sh)
{
  SH2movi(sh, 10, (0<<4)+6);
}
#define init_14_10_0_7
void FASTCALL call_14_10_0_7(SH2_struct* sh)
{
  SH2movi(sh, 10, (0<<4)+7);
}
#define init_14_10_0_8
void FASTCALL call_14_10_0_8(SH2_struct* sh)
{
  SH2movi(sh, 10, (0<<4)+8);
}
#define init_14_10_0_9
void FASTCALL call_14_10_0_9(SH2_struct* sh)
{
  SH2movi(sh, 10, (0<<4)+9);
}
#define init_14_10_0_10
void FASTCALL call_14_10_0_10(SH2_struct* sh)
{
  SH2movi(sh, 10, (0<<4)+10);
}
#define init_14_10_0_11
void FASTCALL call_14_10_0_11(SH2_struct* sh)
{
  SH2movi(sh, 10, (0<<4)+11);
}
#define init_14_10_0_12
void FASTCALL call_14_10_0_12(SH2_struct* sh)
{
  SH2movi(sh, 10, (0<<4)+12);
}
#define init_14_10_0_13
void FASTCALL call_14_10_0_13(SH2_struct* sh)
{
  SH2movi(sh, 10, (0<<4)+13);
}
#define init_14_10_0_14
void FASTCALL call_14_10_0_14(SH2_struct* sh)
{
  SH2movi(sh, 10, (0<<4)+14);
}
#define init_14_10_0_15
void FASTCALL call_14_10_0_15(SH2_struct* sh)
{
  SH2movi(sh, 10, (0<<4)+15);
}
#define init_14_10_1_0
void FASTCALL call_14_10_1_0(SH2_struct* sh)
{
  SH2movi(sh, 10, (1<<4)+0);
}
#define init_14_10_1_1
void FASTCALL call_14_10_1_1(SH2_struct* sh)
{
  SH2movi(sh, 10, (1<<4)+1);
}
#define init_14_10_1_2
void FASTCALL call_14_10_1_2(SH2_struct* sh)
{
  SH2movi(sh, 10, (1<<4)+2);
}
#define init_14_10_1_3
void FASTCALL call_14_10_1_3(SH2_struct* sh)
{
  SH2movi(sh, 10, (1<<4)+3);
}
#define init_14_10_1_4
void FASTCALL call_14_10_1_4(SH2_struct* sh)
{
  SH2movi(sh, 10, (1<<4)+4);
}
#define init_14_10_1_5
void FASTCALL call_14_10_1_5(SH2_struct* sh)
{
  SH2movi(sh, 10, (1<<4)+5);
}
#define init_14_10_1_6
void FASTCALL call_14_10_1_6(SH2_struct* sh)
{
  SH2movi(sh, 10, (1<<4)+6);
}
#define init_14_10_1_7
void FASTCALL call_14_10_1_7(SH2_struct* sh)
{
  SH2movi(sh, 10, (1<<4)+7);
}
#define init_14_10_1_8
void FASTCALL call_14_10_1_8(SH2_struct* sh)
{
  SH2movi(sh, 10, (1<<4)+8);
}
#define init_14_10_1_9
void FASTCALL call_14_10_1_9(SH2_struct* sh)
{
  SH2movi(sh, 10, (1<<4)+9);
}
#define init_14_10_1_10
void FASTCALL call_14_10_1_10(SH2_struct* sh)
{
  SH2movi(sh, 10, (1<<4)+10);
}
#define init_14_10_1_11
void FASTCALL call_14_10_1_11(SH2_struct* sh)
{
  SH2movi(sh, 10, (1<<4)+11);
}
#define init_14_10_1_12
void FASTCALL call_14_10_1_12(SH2_struct* sh)
{
  SH2movi(sh, 10, (1<<4)+12);
}
#define init_14_10_1_13
void FASTCALL call_14_10_1_13(SH2_struct* sh)
{
  SH2movi(sh, 10, (1<<4)+13);
}
#define init_14_10_1_14
void FASTCALL call_14_10_1_14(SH2_struct* sh)
{
  SH2movi(sh, 10, (1<<4)+14);
}
#define init_14_10_1_15
void FASTCALL call_14_10_1_15(SH2_struct* sh)
{
  SH2movi(sh, 10, (1<<4)+15);
}
#define init_14_10_2_0
void FASTCALL call_14_10_2_0(SH2_struct* sh)
{
  SH2movi(sh, 10, (2<<4)+0);
}
#define init_14_10_2_1
void FASTCALL call_14_10_2_1(SH2_struct* sh)
{
  SH2movi(sh, 10, (2<<4)+1);
}
#define init_14_10_2_2
void FASTCALL call_14_10_2_2(SH2_struct* sh)
{
  SH2movi(sh, 10, (2<<4)+2);
}
#define init_14_10_2_3
void FASTCALL call_14_10_2_3(SH2_struct* sh)
{
  SH2movi(sh, 10, (2<<4)+3);
}
#define init_14_10_2_4
void FASTCALL call_14_10_2_4(SH2_struct* sh)
{
  SH2movi(sh, 10, (2<<4)+4);
}
#define init_14_10_2_5
void FASTCALL call_14_10_2_5(SH2_struct* sh)
{
  SH2movi(sh, 10, (2<<4)+5);
}
#define init_14_10_2_6
void FASTCALL call_14_10_2_6(SH2_struct* sh)
{
  SH2movi(sh, 10, (2<<4)+6);
}
#define init_14_10_2_7
void FASTCALL call_14_10_2_7(SH2_struct* sh)
{
  SH2movi(sh, 10, (2<<4)+7);
}
#define init_14_10_2_8
void FASTCALL call_14_10_2_8(SH2_struct* sh)
{
  SH2movi(sh, 10, (2<<4)+8);
}
#define init_14_10_2_9
void FASTCALL call_14_10_2_9(SH2_struct* sh)
{
  SH2movi(sh, 10, (2<<4)+9);
}
#define init_14_10_2_10
void FASTCALL call_14_10_2_10(SH2_struct* sh)
{
  SH2movi(sh, 10, (2<<4)+10);
}
#define init_14_10_2_11
void FASTCALL call_14_10_2_11(SH2_struct* sh)
{
  SH2movi(sh, 10, (2<<4)+11);
}
#define init_14_10_2_12
void FASTCALL call_14_10_2_12(SH2_struct* sh)
{
  SH2movi(sh, 10, (2<<4)+12);
}
#define init_14_10_2_13
void FASTCALL call_14_10_2_13(SH2_struct* sh)
{
  SH2movi(sh, 10, (2<<4)+13);
}
#define init_14_10_2_14
void FASTCALL call_14_10_2_14(SH2_struct* sh)
{
  SH2movi(sh, 10, (2<<4)+14);
}
#define init_14_10_2_15
void FASTCALL call_14_10_2_15(SH2_struct* sh)
{
  SH2movi(sh, 10, (2<<4)+15);
}
#define init_14_10_3_0
void FASTCALL call_14_10_3_0(SH2_struct* sh)
{
  SH2movi(sh, 10, (3<<4)+0);
}
#define init_14_10_3_1
void FASTCALL call_14_10_3_1(SH2_struct* sh)
{
  SH2movi(sh, 10, (3<<4)+1);
}
#define init_14_10_3_2
void FASTCALL call_14_10_3_2(SH2_struct* sh)
{
  SH2movi(sh, 10, (3<<4)+2);
}
#define init_14_10_3_3
void FASTCALL call_14_10_3_3(SH2_struct* sh)
{
  SH2movi(sh, 10, (3<<4)+3);
}
#define init_14_10_3_4
void FASTCALL call_14_10_3_4(SH2_struct* sh)
{
  SH2movi(sh, 10, (3<<4)+4);
}
#define init_14_10_3_5
void FASTCALL call_14_10_3_5(SH2_struct* sh)
{
  SH2movi(sh, 10, (3<<4)+5);
}
#define init_14_10_3_6
void FASTCALL call_14_10_3_6(SH2_struct* sh)
{
  SH2movi(sh, 10, (3<<4)+6);
}
#define init_14_10_3_7
void FASTCALL call_14_10_3_7(SH2_struct* sh)
{
  SH2movi(sh, 10, (3<<4)+7);
}
#define init_14_10_3_8
void FASTCALL call_14_10_3_8(SH2_struct* sh)
{
  SH2movi(sh, 10, (3<<4)+8);
}
#define init_14_10_3_9
void FASTCALL call_14_10_3_9(SH2_struct* sh)
{
  SH2movi(sh, 10, (3<<4)+9);
}
#define init_14_10_3_10
void FASTCALL call_14_10_3_10(SH2_struct* sh)
{
  SH2movi(sh, 10, (3<<4)+10);
}
#define init_14_10_3_11
void FASTCALL call_14_10_3_11(SH2_struct* sh)
{
  SH2movi(sh, 10, (3<<4)+11);
}
#define init_14_10_3_12
void FASTCALL call_14_10_3_12(SH2_struct* sh)
{
  SH2movi(sh, 10, (3<<4)+12);
}
#define init_14_10_3_13
void FASTCALL call_14_10_3_13(SH2_struct* sh)
{
  SH2movi(sh, 10, (3<<4)+13);
}
#define init_14_10_3_14
void FASTCALL call_14_10_3_14(SH2_struct* sh)
{
  SH2movi(sh, 10, (3<<4)+14);
}
#define init_14_10_3_15
void FASTCALL call_14_10_3_15(SH2_struct* sh)
{
  SH2movi(sh, 10, (3<<4)+15);
}
#define init_14_10_4_0
void FASTCALL call_14_10_4_0(SH2_struct* sh)
{
  SH2movi(sh, 10, (4<<4)+0);
}
#define init_14_10_4_1
void FASTCALL call_14_10_4_1(SH2_struct* sh)
{
  SH2movi(sh, 10, (4<<4)+1);
}
#define init_14_10_4_2
void FASTCALL call_14_10_4_2(SH2_struct* sh)
{
  SH2movi(sh, 10, (4<<4)+2);
}
#define init_14_10_4_3
void FASTCALL call_14_10_4_3(SH2_struct* sh)
{
  SH2movi(sh, 10, (4<<4)+3);
}
#define init_14_10_4_4
void FASTCALL call_14_10_4_4(SH2_struct* sh)
{
  SH2movi(sh, 10, (4<<4)+4);
}
#define init_14_10_4_5
void FASTCALL call_14_10_4_5(SH2_struct* sh)
{
  SH2movi(sh, 10, (4<<4)+5);
}
#define init_14_10_4_6
void FASTCALL call_14_10_4_6(SH2_struct* sh)
{
  SH2movi(sh, 10, (4<<4)+6);
}
#define init_14_10_4_7
void FASTCALL call_14_10_4_7(SH2_struct* sh)
{
  SH2movi(sh, 10, (4<<4)+7);
}
#define init_14_10_4_8
void FASTCALL call_14_10_4_8(SH2_struct* sh)
{
  SH2movi(sh, 10, (4<<4)+8);
}
#define init_14_10_4_9
void FASTCALL call_14_10_4_9(SH2_struct* sh)
{
  SH2movi(sh, 10, (4<<4)+9);
}
#define init_14_10_4_10
void FASTCALL call_14_10_4_10(SH2_struct* sh)
{
  SH2movi(sh, 10, (4<<4)+10);
}
#define init_14_10_4_11
void FASTCALL call_14_10_4_11(SH2_struct* sh)
{
  SH2movi(sh, 10, (4<<4)+11);
}
#define init_14_10_4_12
void FASTCALL call_14_10_4_12(SH2_struct* sh)
{
  SH2movi(sh, 10, (4<<4)+12);
}
#define init_14_10_4_13
void FASTCALL call_14_10_4_13(SH2_struct* sh)
{
  SH2movi(sh, 10, (4<<4)+13);
}
#define init_14_10_4_14
void FASTCALL call_14_10_4_14(SH2_struct* sh)
{
  SH2movi(sh, 10, (4<<4)+14);
}
#define init_14_10_4_15
void FASTCALL call_14_10_4_15(SH2_struct* sh)
{
  SH2movi(sh, 10, (4<<4)+15);
}
#define init_14_10_5_0
void FASTCALL call_14_10_5_0(SH2_struct* sh)
{
  SH2movi(sh, 10, (5<<4)+0);
}
#define init_14_10_5_1
void FASTCALL call_14_10_5_1(SH2_struct* sh)
{
  SH2movi(sh, 10, (5<<4)+1);
}
#define init_14_10_5_2
void FASTCALL call_14_10_5_2(SH2_struct* sh)
{
  SH2movi(sh, 10, (5<<4)+2);
}
#define init_14_10_5_3
void FASTCALL call_14_10_5_3(SH2_struct* sh)
{
  SH2movi(sh, 10, (5<<4)+3);
}
#define init_14_10_5_4
void FASTCALL call_14_10_5_4(SH2_struct* sh)
{
  SH2movi(sh, 10, (5<<4)+4);
}
#define init_14_10_5_5
void FASTCALL call_14_10_5_5(SH2_struct* sh)
{
  SH2movi(sh, 10, (5<<4)+5);
}
#define init_14_10_5_6
void FASTCALL call_14_10_5_6(SH2_struct* sh)
{
  SH2movi(sh, 10, (5<<4)+6);
}
#define init_14_10_5_7
void FASTCALL call_14_10_5_7(SH2_struct* sh)
{
  SH2movi(sh, 10, (5<<4)+7);
}
#define init_14_10_5_8
void FASTCALL call_14_10_5_8(SH2_struct* sh)
{
  SH2movi(sh, 10, (5<<4)+8);
}
#define init_14_10_5_9
void FASTCALL call_14_10_5_9(SH2_struct* sh)
{
  SH2movi(sh, 10, (5<<4)+9);
}
#define init_14_10_5_10
void FASTCALL call_14_10_5_10(SH2_struct* sh)
{
  SH2movi(sh, 10, (5<<4)+10);
}
#define init_14_10_5_11
void FASTCALL call_14_10_5_11(SH2_struct* sh)
{
  SH2movi(sh, 10, (5<<4)+11);
}
#define init_14_10_5_12
void FASTCALL call_14_10_5_12(SH2_struct* sh)
{
  SH2movi(sh, 10, (5<<4)+12);
}
#define init_14_10_5_13
void FASTCALL call_14_10_5_13(SH2_struct* sh)
{
  SH2movi(sh, 10, (5<<4)+13);
}
#define init_14_10_5_14
void FASTCALL call_14_10_5_14(SH2_struct* sh)
{
  SH2movi(sh, 10, (5<<4)+14);
}
#define init_14_10_5_15
void FASTCALL call_14_10_5_15(SH2_struct* sh)
{
  SH2movi(sh, 10, (5<<4)+15);
}
#define init_14_10_6_0
void FASTCALL call_14_10_6_0(SH2_struct* sh)
{
  SH2movi(sh, 10, (6<<4)+0);
}
#define init_14_10_6_1
void FASTCALL call_14_10_6_1(SH2_struct* sh)
{
  SH2movi(sh, 10, (6<<4)+1);
}
#define init_14_10_6_2
void FASTCALL call_14_10_6_2(SH2_struct* sh)
{
  SH2movi(sh, 10, (6<<4)+2);
}
#define init_14_10_6_3
void FASTCALL call_14_10_6_3(SH2_struct* sh)
{
  SH2movi(sh, 10, (6<<4)+3);
}
#define init_14_10_6_4
void FASTCALL call_14_10_6_4(SH2_struct* sh)
{
  SH2movi(sh, 10, (6<<4)+4);
}
#define init_14_10_6_5
void FASTCALL call_14_10_6_5(SH2_struct* sh)
{
  SH2movi(sh, 10, (6<<4)+5);
}
#define init_14_10_6_6
void FASTCALL call_14_10_6_6(SH2_struct* sh)
{
  SH2movi(sh, 10, (6<<4)+6);
}
#define init_14_10_6_7
void FASTCALL call_14_10_6_7(SH2_struct* sh)
{
  SH2movi(sh, 10, (6<<4)+7);
}
#define init_14_10_6_8
void FASTCALL call_14_10_6_8(SH2_struct* sh)
{
  SH2movi(sh, 10, (6<<4)+8);
}
#define init_14_10_6_9
void FASTCALL call_14_10_6_9(SH2_struct* sh)
{
  SH2movi(sh, 10, (6<<4)+9);
}
#define init_14_10_6_10
void FASTCALL call_14_10_6_10(SH2_struct* sh)
{
  SH2movi(sh, 10, (6<<4)+10);
}
#define init_14_10_6_11
void FASTCALL call_14_10_6_11(SH2_struct* sh)
{
  SH2movi(sh, 10, (6<<4)+11);
}
#define init_14_10_6_12
void FASTCALL call_14_10_6_12(SH2_struct* sh)
{
  SH2movi(sh, 10, (6<<4)+12);
}
#define init_14_10_6_13
void FASTCALL call_14_10_6_13(SH2_struct* sh)
{
  SH2movi(sh, 10, (6<<4)+13);
}
#define init_14_10_6_14
void FASTCALL call_14_10_6_14(SH2_struct* sh)
{
  SH2movi(sh, 10, (6<<4)+14);
}
#define init_14_10_6_15
void FASTCALL call_14_10_6_15(SH2_struct* sh)
{
  SH2movi(sh, 10, (6<<4)+15);
}
#define init_14_10_7_0
void FASTCALL call_14_10_7_0(SH2_struct* sh)
{
  SH2movi(sh, 10, (7<<4)+0);
}
#define init_14_10_7_1
void FASTCALL call_14_10_7_1(SH2_struct* sh)
{
  SH2movi(sh, 10, (7<<4)+1);
}
#define init_14_10_7_2
void FASTCALL call_14_10_7_2(SH2_struct* sh)
{
  SH2movi(sh, 10, (7<<4)+2);
}
#define init_14_10_7_3
void FASTCALL call_14_10_7_3(SH2_struct* sh)
{
  SH2movi(sh, 10, (7<<4)+3);
}
#define init_14_10_7_4
void FASTCALL call_14_10_7_4(SH2_struct* sh)
{
  SH2movi(sh, 10, (7<<4)+4);
}
#define init_14_10_7_5
void FASTCALL call_14_10_7_5(SH2_struct* sh)
{
  SH2movi(sh, 10, (7<<4)+5);
}
#define init_14_10_7_6
void FASTCALL call_14_10_7_6(SH2_struct* sh)
{
  SH2movi(sh, 10, (7<<4)+6);
}
#define init_14_10_7_7
void FASTCALL call_14_10_7_7(SH2_struct* sh)
{
  SH2movi(sh, 10, (7<<4)+7);
}
#define init_14_10_7_8
void FASTCALL call_14_10_7_8(SH2_struct* sh)
{
  SH2movi(sh, 10, (7<<4)+8);
}
#define init_14_10_7_9
void FASTCALL call_14_10_7_9(SH2_struct* sh)
{
  SH2movi(sh, 10, (7<<4)+9);
}
#define init_14_10_7_10
void FASTCALL call_14_10_7_10(SH2_struct* sh)
{
  SH2movi(sh, 10, (7<<4)+10);
}
#define init_14_10_7_11
void FASTCALL call_14_10_7_11(SH2_struct* sh)
{
  SH2movi(sh, 10, (7<<4)+11);
}
#define init_14_10_7_12
void FASTCALL call_14_10_7_12(SH2_struct* sh)
{
  SH2movi(sh, 10, (7<<4)+12);
}
#define init_14_10_7_13
void FASTCALL call_14_10_7_13(SH2_struct* sh)
{
  SH2movi(sh, 10, (7<<4)+13);
}
#define init_14_10_7_14
void FASTCALL call_14_10_7_14(SH2_struct* sh)
{
  SH2movi(sh, 10, (7<<4)+14);
}
#define init_14_10_7_15
void FASTCALL call_14_10_7_15(SH2_struct* sh)
{
  SH2movi(sh, 10, (7<<4)+15);
}
#define init_14_10_8_0
void FASTCALL call_14_10_8_0(SH2_struct* sh)
{
  SH2movi(sh, 10, (8<<4)+0);
}
#define init_14_10_8_1
void FASTCALL call_14_10_8_1(SH2_struct* sh)
{
  SH2movi(sh, 10, (8<<4)+1);
}
#define init_14_10_8_2
void FASTCALL call_14_10_8_2(SH2_struct* sh)
{
  SH2movi(sh, 10, (8<<4)+2);
}
#define init_14_10_8_3
void FASTCALL call_14_10_8_3(SH2_struct* sh)
{
  SH2movi(sh, 10, (8<<4)+3);
}
#define init_14_10_8_4
void FASTCALL call_14_10_8_4(SH2_struct* sh)
{
  SH2movi(sh, 10, (8<<4)+4);
}
#define init_14_10_8_5
void FASTCALL call_14_10_8_5(SH2_struct* sh)
{
  SH2movi(sh, 10, (8<<4)+5);
}
#define init_14_10_8_6
void FASTCALL call_14_10_8_6(SH2_struct* sh)
{
  SH2movi(sh, 10, (8<<4)+6);
}
#define init_14_10_8_7
void FASTCALL call_14_10_8_7(SH2_struct* sh)
{
  SH2movi(sh, 10, (8<<4)+7);
}
#define init_14_10_8_8
void FASTCALL call_14_10_8_8(SH2_struct* sh)
{
  SH2movi(sh, 10, (8<<4)+8);
}
#define init_14_10_8_9
void FASTCALL call_14_10_8_9(SH2_struct* sh)
{
  SH2movi(sh, 10, (8<<4)+9);
}
#define init_14_10_8_10
void FASTCALL call_14_10_8_10(SH2_struct* sh)
{
  SH2movi(sh, 10, (8<<4)+10);
}
#define init_14_10_8_11
void FASTCALL call_14_10_8_11(SH2_struct* sh)
{
  SH2movi(sh, 10, (8<<4)+11);
}
#define init_14_10_8_12
void FASTCALL call_14_10_8_12(SH2_struct* sh)
{
  SH2movi(sh, 10, (8<<4)+12);
}
#define init_14_10_8_13
void FASTCALL call_14_10_8_13(SH2_struct* sh)
{
  SH2movi(sh, 10, (8<<4)+13);
}
#define init_14_10_8_14
void FASTCALL call_14_10_8_14(SH2_struct* sh)
{
  SH2movi(sh, 10, (8<<4)+14);
}
#define init_14_10_8_15
void FASTCALL call_14_10_8_15(SH2_struct* sh)
{
  SH2movi(sh, 10, (8<<4)+15);
}
#define init_14_10_9_0
void FASTCALL call_14_10_9_0(SH2_struct* sh)
{
  SH2movi(sh, 10, (9<<4)+0);
}
#define init_14_10_9_1
void FASTCALL call_14_10_9_1(SH2_struct* sh)
{
  SH2movi(sh, 10, (9<<4)+1);
}
#define init_14_10_9_2
void FASTCALL call_14_10_9_2(SH2_struct* sh)
{
  SH2movi(sh, 10, (9<<4)+2);
}
#define init_14_10_9_3
void FASTCALL call_14_10_9_3(SH2_struct* sh)
{
  SH2movi(sh, 10, (9<<4)+3);
}
#define init_14_10_9_4
void FASTCALL call_14_10_9_4(SH2_struct* sh)
{
  SH2movi(sh, 10, (9<<4)+4);
}
#define init_14_10_9_5
void FASTCALL call_14_10_9_5(SH2_struct* sh)
{
  SH2movi(sh, 10, (9<<4)+5);
}
#define init_14_10_9_6
void FASTCALL call_14_10_9_6(SH2_struct* sh)
{
  SH2movi(sh, 10, (9<<4)+6);
}
#define init_14_10_9_7
void FASTCALL call_14_10_9_7(SH2_struct* sh)
{
  SH2movi(sh, 10, (9<<4)+7);
}
#define init_14_10_9_8
void FASTCALL call_14_10_9_8(SH2_struct* sh)
{
  SH2movi(sh, 10, (9<<4)+8);
}
#define init_14_10_9_9
void FASTCALL call_14_10_9_9(SH2_struct* sh)
{
  SH2movi(sh, 10, (9<<4)+9);
}
#define init_14_10_9_10
void FASTCALL call_14_10_9_10(SH2_struct* sh)
{
  SH2movi(sh, 10, (9<<4)+10);
}
#define init_14_10_9_11
void FASTCALL call_14_10_9_11(SH2_struct* sh)
{
  SH2movi(sh, 10, (9<<4)+11);
}
#define init_14_10_9_12
void FASTCALL call_14_10_9_12(SH2_struct* sh)
{
  SH2movi(sh, 10, (9<<4)+12);
}
#define init_14_10_9_13
void FASTCALL call_14_10_9_13(SH2_struct* sh)
{
  SH2movi(sh, 10, (9<<4)+13);
}
#define init_14_10_9_14
void FASTCALL call_14_10_9_14(SH2_struct* sh)
{
  SH2movi(sh, 10, (9<<4)+14);
}
#define init_14_10_9_15
void FASTCALL call_14_10_9_15(SH2_struct* sh)
{
  SH2movi(sh, 10, (9<<4)+15);
}
#define init_14_10_10_0
void FASTCALL call_14_10_10_0(SH2_struct* sh)
{
  SH2movi(sh, 10, (10<<4)+0);
}
#define init_14_10_10_1
void FASTCALL call_14_10_10_1(SH2_struct* sh)
{
  SH2movi(sh, 10, (10<<4)+1);
}
#define init_14_10_10_2
void FASTCALL call_14_10_10_2(SH2_struct* sh)
{
  SH2movi(sh, 10, (10<<4)+2);
}
#define init_14_10_10_3
void FASTCALL call_14_10_10_3(SH2_struct* sh)
{
  SH2movi(sh, 10, (10<<4)+3);
}
#define init_14_10_10_4
void FASTCALL call_14_10_10_4(SH2_struct* sh)
{
  SH2movi(sh, 10, (10<<4)+4);
}
#define init_14_10_10_5
void FASTCALL call_14_10_10_5(SH2_struct* sh)
{
  SH2movi(sh, 10, (10<<4)+5);
}
#define init_14_10_10_6
void FASTCALL call_14_10_10_6(SH2_struct* sh)
{
  SH2movi(sh, 10, (10<<4)+6);
}
#define init_14_10_10_7
void FASTCALL call_14_10_10_7(SH2_struct* sh)
{
  SH2movi(sh, 10, (10<<4)+7);
}
#define init_14_10_10_8
void FASTCALL call_14_10_10_8(SH2_struct* sh)
{
  SH2movi(sh, 10, (10<<4)+8);
}
#define init_14_10_10_9
void FASTCALL call_14_10_10_9(SH2_struct* sh)
{
  SH2movi(sh, 10, (10<<4)+9);
}
#define init_14_10_10_10
void FASTCALL call_14_10_10_10(SH2_struct* sh)
{
  SH2movi(sh, 10, (10<<4)+10);
}
#define init_14_10_10_11
void FASTCALL call_14_10_10_11(SH2_struct* sh)
{
  SH2movi(sh, 10, (10<<4)+11);
}
#define init_14_10_10_12
void FASTCALL call_14_10_10_12(SH2_struct* sh)
{
  SH2movi(sh, 10, (10<<4)+12);
}
#define init_14_10_10_13
void FASTCALL call_14_10_10_13(SH2_struct* sh)
{
  SH2movi(sh, 10, (10<<4)+13);
}
#define init_14_10_10_14
void FASTCALL call_14_10_10_14(SH2_struct* sh)
{
  SH2movi(sh, 10, (10<<4)+14);
}
#define init_14_10_10_15
void FASTCALL call_14_10_10_15(SH2_struct* sh)
{
  SH2movi(sh, 10, (10<<4)+15);
}
#define init_14_10_11_0
void FASTCALL call_14_10_11_0(SH2_struct* sh)
{
  SH2movi(sh, 10, (11<<4)+0);
}
#define init_14_10_11_1
void FASTCALL call_14_10_11_1(SH2_struct* sh)
{
  SH2movi(sh, 10, (11<<4)+1);
}
#define init_14_10_11_2
void FASTCALL call_14_10_11_2(SH2_struct* sh)
{
  SH2movi(sh, 10, (11<<4)+2);
}
#define init_14_10_11_3
void FASTCALL call_14_10_11_3(SH2_struct* sh)
{
  SH2movi(sh, 10, (11<<4)+3);
}
#define init_14_10_11_4
void FASTCALL call_14_10_11_4(SH2_struct* sh)
{
  SH2movi(sh, 10, (11<<4)+4);
}
#define init_14_10_11_5
void FASTCALL call_14_10_11_5(SH2_struct* sh)
{
  SH2movi(sh, 10, (11<<4)+5);
}
#define init_14_10_11_6
void FASTCALL call_14_10_11_6(SH2_struct* sh)
{
  SH2movi(sh, 10, (11<<4)+6);
}
#define init_14_10_11_7
void FASTCALL call_14_10_11_7(SH2_struct* sh)
{
  SH2movi(sh, 10, (11<<4)+7);
}
#define init_14_10_11_8
void FASTCALL call_14_10_11_8(SH2_struct* sh)
{
  SH2movi(sh, 10, (11<<4)+8);
}
#define init_14_10_11_9
void FASTCALL call_14_10_11_9(SH2_struct* sh)
{
  SH2movi(sh, 10, (11<<4)+9);
}
#define init_14_10_11_10
void FASTCALL call_14_10_11_10(SH2_struct* sh)
{
  SH2movi(sh, 10, (11<<4)+10);
}
#define init_14_10_11_11
void FASTCALL call_14_10_11_11(SH2_struct* sh)
{
  SH2movi(sh, 10, (11<<4)+11);
}
#define init_14_10_11_12
void FASTCALL call_14_10_11_12(SH2_struct* sh)
{
  SH2movi(sh, 10, (11<<4)+12);
}
#define init_14_10_11_13
void FASTCALL call_14_10_11_13(SH2_struct* sh)
{
  SH2movi(sh, 10, (11<<4)+13);
}
#define init_14_10_11_14
void FASTCALL call_14_10_11_14(SH2_struct* sh)
{
  SH2movi(sh, 10, (11<<4)+14);
}
#define init_14_10_11_15
void FASTCALL call_14_10_11_15(SH2_struct* sh)
{
  SH2movi(sh, 10, (11<<4)+15);
}
#define init_14_10_12_0
void FASTCALL call_14_10_12_0(SH2_struct* sh)
{
  SH2movi(sh, 10, (12<<4)+0);
}
#define init_14_10_12_1
void FASTCALL call_14_10_12_1(SH2_struct* sh)
{
  SH2movi(sh, 10, (12<<4)+1);
}
#define init_14_10_12_2
void FASTCALL call_14_10_12_2(SH2_struct* sh)
{
  SH2movi(sh, 10, (12<<4)+2);
}
#define init_14_10_12_3
void FASTCALL call_14_10_12_3(SH2_struct* sh)
{
  SH2movi(sh, 10, (12<<4)+3);
}
#define init_14_10_12_4
void FASTCALL call_14_10_12_4(SH2_struct* sh)
{
  SH2movi(sh, 10, (12<<4)+4);
}
#define init_14_10_12_5
void FASTCALL call_14_10_12_5(SH2_struct* sh)
{
  SH2movi(sh, 10, (12<<4)+5);
}
#define init_14_10_12_6
void FASTCALL call_14_10_12_6(SH2_struct* sh)
{
  SH2movi(sh, 10, (12<<4)+6);
}
#define init_14_10_12_7
void FASTCALL call_14_10_12_7(SH2_struct* sh)
{
  SH2movi(sh, 10, (12<<4)+7);
}
#define init_14_10_12_8
void FASTCALL call_14_10_12_8(SH2_struct* sh)
{
  SH2movi(sh, 10, (12<<4)+8);
}
#define init_14_10_12_9
void FASTCALL call_14_10_12_9(SH2_struct* sh)
{
  SH2movi(sh, 10, (12<<4)+9);
}
#define init_14_10_12_10
void FASTCALL call_14_10_12_10(SH2_struct* sh)
{
  SH2movi(sh, 10, (12<<4)+10);
}
#define init_14_10_12_11
void FASTCALL call_14_10_12_11(SH2_struct* sh)
{
  SH2movi(sh, 10, (12<<4)+11);
}
#define init_14_10_12_12
void FASTCALL call_14_10_12_12(SH2_struct* sh)
{
  SH2movi(sh, 10, (12<<4)+12);
}
#define init_14_10_12_13
void FASTCALL call_14_10_12_13(SH2_struct* sh)
{
  SH2movi(sh, 10, (12<<4)+13);
}
#define init_14_10_12_14
void FASTCALL call_14_10_12_14(SH2_struct* sh)
{
  SH2movi(sh, 10, (12<<4)+14);
}
#define init_14_10_12_15
void FASTCALL call_14_10_12_15(SH2_struct* sh)
{
  SH2movi(sh, 10, (12<<4)+15);
}
#define init_14_10_13_0
void FASTCALL call_14_10_13_0(SH2_struct* sh)
{
  SH2movi(sh, 10, (13<<4)+0);
}
#define init_14_10_13_1
void FASTCALL call_14_10_13_1(SH2_struct* sh)
{
  SH2movi(sh, 10, (13<<4)+1);
}
#define init_14_10_13_2
void FASTCALL call_14_10_13_2(SH2_struct* sh)
{
  SH2movi(sh, 10, (13<<4)+2);
}
#define init_14_10_13_3
void FASTCALL call_14_10_13_3(SH2_struct* sh)
{
  SH2movi(sh, 10, (13<<4)+3);
}
#define init_14_10_13_4
void FASTCALL call_14_10_13_4(SH2_struct* sh)
{
  SH2movi(sh, 10, (13<<4)+4);
}
#define init_14_10_13_5
void FASTCALL call_14_10_13_5(SH2_struct* sh)
{
  SH2movi(sh, 10, (13<<4)+5);
}
#define init_14_10_13_6
void FASTCALL call_14_10_13_6(SH2_struct* sh)
{
  SH2movi(sh, 10, (13<<4)+6);
}
#define init_14_10_13_7
void FASTCALL call_14_10_13_7(SH2_struct* sh)
{
  SH2movi(sh, 10, (13<<4)+7);
}
#define init_14_10_13_8
void FASTCALL call_14_10_13_8(SH2_struct* sh)
{
  SH2movi(sh, 10, (13<<4)+8);
}
#define init_14_10_13_9
void FASTCALL call_14_10_13_9(SH2_struct* sh)
{
  SH2movi(sh, 10, (13<<4)+9);
}
#define init_14_10_13_10
void FASTCALL call_14_10_13_10(SH2_struct* sh)
{
  SH2movi(sh, 10, (13<<4)+10);
}
#define init_14_10_13_11
void FASTCALL call_14_10_13_11(SH2_struct* sh)
{
  SH2movi(sh, 10, (13<<4)+11);
}
#define init_14_10_13_12
void FASTCALL call_14_10_13_12(SH2_struct* sh)
{
  SH2movi(sh, 10, (13<<4)+12);
}
#define init_14_10_13_13
void FASTCALL call_14_10_13_13(SH2_struct* sh)
{
  SH2movi(sh, 10, (13<<4)+13);
}
#define init_14_10_13_14
void FASTCALL call_14_10_13_14(SH2_struct* sh)
{
  SH2movi(sh, 10, (13<<4)+14);
}
#define init_14_10_13_15
void FASTCALL call_14_10_13_15(SH2_struct* sh)
{
  SH2movi(sh, 10, (13<<4)+15);
}
#define init_14_10_14_0
void FASTCALL call_14_10_14_0(SH2_struct* sh)
{
  SH2movi(sh, 10, (14<<4)+0);
}
#define init_14_10_14_1
void FASTCALL call_14_10_14_1(SH2_struct* sh)
{
  SH2movi(sh, 10, (14<<4)+1);
}
#define init_14_10_14_2
void FASTCALL call_14_10_14_2(SH2_struct* sh)
{
  SH2movi(sh, 10, (14<<4)+2);
}
#define init_14_10_14_3
void FASTCALL call_14_10_14_3(SH2_struct* sh)
{
  SH2movi(sh, 10, (14<<4)+3);
}
#define init_14_10_14_4
void FASTCALL call_14_10_14_4(SH2_struct* sh)
{
  SH2movi(sh, 10, (14<<4)+4);
}
#define init_14_10_14_5
void FASTCALL call_14_10_14_5(SH2_struct* sh)
{
  SH2movi(sh, 10, (14<<4)+5);
}
#define init_14_10_14_6
void FASTCALL call_14_10_14_6(SH2_struct* sh)
{
  SH2movi(sh, 10, (14<<4)+6);
}
#define init_14_10_14_7
void FASTCALL call_14_10_14_7(SH2_struct* sh)
{
  SH2movi(sh, 10, (14<<4)+7);
}
#define init_14_10_14_8
void FASTCALL call_14_10_14_8(SH2_struct* sh)
{
  SH2movi(sh, 10, (14<<4)+8);
}
#define init_14_10_14_9
void FASTCALL call_14_10_14_9(SH2_struct* sh)
{
  SH2movi(sh, 10, (14<<4)+9);
}
#define init_14_10_14_10
void FASTCALL call_14_10_14_10(SH2_struct* sh)
{
  SH2movi(sh, 10, (14<<4)+10);
}
#define init_14_10_14_11
void FASTCALL call_14_10_14_11(SH2_struct* sh)
{
  SH2movi(sh, 10, (14<<4)+11);
}
#define init_14_10_14_12
void FASTCALL call_14_10_14_12(SH2_struct* sh)
{
  SH2movi(sh, 10, (14<<4)+12);
}
#define init_14_10_14_13
void FASTCALL call_14_10_14_13(SH2_struct* sh)
{
  SH2movi(sh, 10, (14<<4)+13);
}
#define init_14_10_14_14
void FASTCALL call_14_10_14_14(SH2_struct* sh)
{
  SH2movi(sh, 10, (14<<4)+14);
}
#define init_14_10_14_15
void FASTCALL call_14_10_14_15(SH2_struct* sh)
{
  SH2movi(sh, 10, (14<<4)+15);
}
#define init_14_10_15_0
void FASTCALL call_14_10_15_0(SH2_struct* sh)
{
  SH2movi(sh, 10, (15<<4)+0);
}
#define init_14_10_15_1
void FASTCALL call_14_10_15_1(SH2_struct* sh)
{
  SH2movi(sh, 10, (15<<4)+1);
}
#define init_14_10_15_2
void FASTCALL call_14_10_15_2(SH2_struct* sh)
{
  SH2movi(sh, 10, (15<<4)+2);
}
#define init_14_10_15_3
void FASTCALL call_14_10_15_3(SH2_struct* sh)
{
  SH2movi(sh, 10, (15<<4)+3);
}
#define init_14_10_15_4
void FASTCALL call_14_10_15_4(SH2_struct* sh)
{
  SH2movi(sh, 10, (15<<4)+4);
}
#define init_14_10_15_5
void FASTCALL call_14_10_15_5(SH2_struct* sh)
{
  SH2movi(sh, 10, (15<<4)+5);
}
#define init_14_10_15_6
void FASTCALL call_14_10_15_6(SH2_struct* sh)
{
  SH2movi(sh, 10, (15<<4)+6);
}
#define init_14_10_15_7
void FASTCALL call_14_10_15_7(SH2_struct* sh)
{
  SH2movi(sh, 10, (15<<4)+7);
}
#define init_14_10_15_8
void FASTCALL call_14_10_15_8(SH2_struct* sh)
{
  SH2movi(sh, 10, (15<<4)+8);
}
#define init_14_10_15_9
void FASTCALL call_14_10_15_9(SH2_struct* sh)
{
  SH2movi(sh, 10, (15<<4)+9);
}
#define init_14_10_15_10
void FASTCALL call_14_10_15_10(SH2_struct* sh)
{
  SH2movi(sh, 10, (15<<4)+10);
}
#define init_14_10_15_11
void FASTCALL call_14_10_15_11(SH2_struct* sh)
{
  SH2movi(sh, 10, (15<<4)+11);
}
#define init_14_10_15_12
void FASTCALL call_14_10_15_12(SH2_struct* sh)
{
  SH2movi(sh, 10, (15<<4)+12);
}
#define init_14_10_15_13
void FASTCALL call_14_10_15_13(SH2_struct* sh)
{
  SH2movi(sh, 10, (15<<4)+13);
}
#define init_14_10_15_14
void FASTCALL call_14_10_15_14(SH2_struct* sh)
{
  SH2movi(sh, 10, (15<<4)+14);
}
#define init_14_10_15_15
void FASTCALL call_14_10_15_15(SH2_struct* sh)
{
  SH2movi(sh, 10, (15<<4)+15);
}
#define init_14_11_0_0
void FASTCALL call_14_11_0_0(SH2_struct* sh)
{
  SH2movi(sh, 11, (0<<4)+0);
}
#define init_14_11_0_1
void FASTCALL call_14_11_0_1(SH2_struct* sh)
{
  SH2movi(sh, 11, (0<<4)+1);
}
#define init_14_11_0_2
void FASTCALL call_14_11_0_2(SH2_struct* sh)
{
  SH2movi(sh, 11, (0<<4)+2);
}
#define init_14_11_0_3
void FASTCALL call_14_11_0_3(SH2_struct* sh)
{
  SH2movi(sh, 11, (0<<4)+3);
}
#define init_14_11_0_4
void FASTCALL call_14_11_0_4(SH2_struct* sh)
{
  SH2movi(sh, 11, (0<<4)+4);
}
#define init_14_11_0_5
void FASTCALL call_14_11_0_5(SH2_struct* sh)
{
  SH2movi(sh, 11, (0<<4)+5);
}
#define init_14_11_0_6
void FASTCALL call_14_11_0_6(SH2_struct* sh)
{
  SH2movi(sh, 11, (0<<4)+6);
}
#define init_14_11_0_7
void FASTCALL call_14_11_0_7(SH2_struct* sh)
{
  SH2movi(sh, 11, (0<<4)+7);
}
#define init_14_11_0_8
void FASTCALL call_14_11_0_8(SH2_struct* sh)
{
  SH2movi(sh, 11, (0<<4)+8);
}
#define init_14_11_0_9
void FASTCALL call_14_11_0_9(SH2_struct* sh)
{
  SH2movi(sh, 11, (0<<4)+9);
}
#define init_14_11_0_10
void FASTCALL call_14_11_0_10(SH2_struct* sh)
{
  SH2movi(sh, 11, (0<<4)+10);
}
#define init_14_11_0_11
void FASTCALL call_14_11_0_11(SH2_struct* sh)
{
  SH2movi(sh, 11, (0<<4)+11);
}
#define init_14_11_0_12
void FASTCALL call_14_11_0_12(SH2_struct* sh)
{
  SH2movi(sh, 11, (0<<4)+12);
}
#define init_14_11_0_13
void FASTCALL call_14_11_0_13(SH2_struct* sh)
{
  SH2movi(sh, 11, (0<<4)+13);
}
#define init_14_11_0_14
void FASTCALL call_14_11_0_14(SH2_struct* sh)
{
  SH2movi(sh, 11, (0<<4)+14);
}
#define init_14_11_0_15
void FASTCALL call_14_11_0_15(SH2_struct* sh)
{
  SH2movi(sh, 11, (0<<4)+15);
}
#define init_14_11_1_0
void FASTCALL call_14_11_1_0(SH2_struct* sh)
{
  SH2movi(sh, 11, (1<<4)+0);
}
#define init_14_11_1_1
void FASTCALL call_14_11_1_1(SH2_struct* sh)
{
  SH2movi(sh, 11, (1<<4)+1);
}
#define init_14_11_1_2
void FASTCALL call_14_11_1_2(SH2_struct* sh)
{
  SH2movi(sh, 11, (1<<4)+2);
}
#define init_14_11_1_3
void FASTCALL call_14_11_1_3(SH2_struct* sh)
{
  SH2movi(sh, 11, (1<<4)+3);
}
#define init_14_11_1_4
void FASTCALL call_14_11_1_4(SH2_struct* sh)
{
  SH2movi(sh, 11, (1<<4)+4);
}
#define init_14_11_1_5
void FASTCALL call_14_11_1_5(SH2_struct* sh)
{
  SH2movi(sh, 11, (1<<4)+5);
}
#define init_14_11_1_6
void FASTCALL call_14_11_1_6(SH2_struct* sh)
{
  SH2movi(sh, 11, (1<<4)+6);
}
#define init_14_11_1_7
void FASTCALL call_14_11_1_7(SH2_struct* sh)
{
  SH2movi(sh, 11, (1<<4)+7);
}
#define init_14_11_1_8
void FASTCALL call_14_11_1_8(SH2_struct* sh)
{
  SH2movi(sh, 11, (1<<4)+8);
}
#define init_14_11_1_9
void FASTCALL call_14_11_1_9(SH2_struct* sh)
{
  SH2movi(sh, 11, (1<<4)+9);
}
#define init_14_11_1_10
void FASTCALL call_14_11_1_10(SH2_struct* sh)
{
  SH2movi(sh, 11, (1<<4)+10);
}
#define init_14_11_1_11
void FASTCALL call_14_11_1_11(SH2_struct* sh)
{
  SH2movi(sh, 11, (1<<4)+11);
}
#define init_14_11_1_12
void FASTCALL call_14_11_1_12(SH2_struct* sh)
{
  SH2movi(sh, 11, (1<<4)+12);
}
#define init_14_11_1_13
void FASTCALL call_14_11_1_13(SH2_struct* sh)
{
  SH2movi(sh, 11, (1<<4)+13);
}
#define init_14_11_1_14
void FASTCALL call_14_11_1_14(SH2_struct* sh)
{
  SH2movi(sh, 11, (1<<4)+14);
}
#define init_14_11_1_15
void FASTCALL call_14_11_1_15(SH2_struct* sh)
{
  SH2movi(sh, 11, (1<<4)+15);
}
#define init_14_11_2_0
void FASTCALL call_14_11_2_0(SH2_struct* sh)
{
  SH2movi(sh, 11, (2<<4)+0);
}
#define init_14_11_2_1
void FASTCALL call_14_11_2_1(SH2_struct* sh)
{
  SH2movi(sh, 11, (2<<4)+1);
}
#define init_14_11_2_2
void FASTCALL call_14_11_2_2(SH2_struct* sh)
{
  SH2movi(sh, 11, (2<<4)+2);
}
#define init_14_11_2_3
void FASTCALL call_14_11_2_3(SH2_struct* sh)
{
  SH2movi(sh, 11, (2<<4)+3);
}
#define init_14_11_2_4
void FASTCALL call_14_11_2_4(SH2_struct* sh)
{
  SH2movi(sh, 11, (2<<4)+4);
}
#define init_14_11_2_5
void FASTCALL call_14_11_2_5(SH2_struct* sh)
{
  SH2movi(sh, 11, (2<<4)+5);
}
#define init_14_11_2_6
void FASTCALL call_14_11_2_6(SH2_struct* sh)
{
  SH2movi(sh, 11, (2<<4)+6);
}
#define init_14_11_2_7
void FASTCALL call_14_11_2_7(SH2_struct* sh)
{
  SH2movi(sh, 11, (2<<4)+7);
}
#define init_14_11_2_8
void FASTCALL call_14_11_2_8(SH2_struct* sh)
{
  SH2movi(sh, 11, (2<<4)+8);
}
#define init_14_11_2_9
void FASTCALL call_14_11_2_9(SH2_struct* sh)
{
  SH2movi(sh, 11, (2<<4)+9);
}
#define init_14_11_2_10
void FASTCALL call_14_11_2_10(SH2_struct* sh)
{
  SH2movi(sh, 11, (2<<4)+10);
}
#define init_14_11_2_11
void FASTCALL call_14_11_2_11(SH2_struct* sh)
{
  SH2movi(sh, 11, (2<<4)+11);
}
#define init_14_11_2_12
void FASTCALL call_14_11_2_12(SH2_struct* sh)
{
  SH2movi(sh, 11, (2<<4)+12);
}
#define init_14_11_2_13
void FASTCALL call_14_11_2_13(SH2_struct* sh)
{
  SH2movi(sh, 11, (2<<4)+13);
}
#define init_14_11_2_14
void FASTCALL call_14_11_2_14(SH2_struct* sh)
{
  SH2movi(sh, 11, (2<<4)+14);
}
#define init_14_11_2_15
void FASTCALL call_14_11_2_15(SH2_struct* sh)
{
  SH2movi(sh, 11, (2<<4)+15);
}
#define init_14_11_3_0
void FASTCALL call_14_11_3_0(SH2_struct* sh)
{
  SH2movi(sh, 11, (3<<4)+0);
}
#define init_14_11_3_1
void FASTCALL call_14_11_3_1(SH2_struct* sh)
{
  SH2movi(sh, 11, (3<<4)+1);
}
#define init_14_11_3_2
void FASTCALL call_14_11_3_2(SH2_struct* sh)
{
  SH2movi(sh, 11, (3<<4)+2);
}
#define init_14_11_3_3
void FASTCALL call_14_11_3_3(SH2_struct* sh)
{
  SH2movi(sh, 11, (3<<4)+3);
}
#define init_14_11_3_4
void FASTCALL call_14_11_3_4(SH2_struct* sh)
{
  SH2movi(sh, 11, (3<<4)+4);
}
#define init_14_11_3_5
void FASTCALL call_14_11_3_5(SH2_struct* sh)
{
  SH2movi(sh, 11, (3<<4)+5);
}
#define init_14_11_3_6
void FASTCALL call_14_11_3_6(SH2_struct* sh)
{
  SH2movi(sh, 11, (3<<4)+6);
}
#define init_14_11_3_7
void FASTCALL call_14_11_3_7(SH2_struct* sh)
{
  SH2movi(sh, 11, (3<<4)+7);
}
#define init_14_11_3_8
void FASTCALL call_14_11_3_8(SH2_struct* sh)
{
  SH2movi(sh, 11, (3<<4)+8);
}
#define init_14_11_3_9
void FASTCALL call_14_11_3_9(SH2_struct* sh)
{
  SH2movi(sh, 11, (3<<4)+9);
}
#define init_14_11_3_10
void FASTCALL call_14_11_3_10(SH2_struct* sh)
{
  SH2movi(sh, 11, (3<<4)+10);
}
#define init_14_11_3_11
void FASTCALL call_14_11_3_11(SH2_struct* sh)
{
  SH2movi(sh, 11, (3<<4)+11);
}
#define init_14_11_3_12
void FASTCALL call_14_11_3_12(SH2_struct* sh)
{
  SH2movi(sh, 11, (3<<4)+12);
}
#define init_14_11_3_13
void FASTCALL call_14_11_3_13(SH2_struct* sh)
{
  SH2movi(sh, 11, (3<<4)+13);
}
#define init_14_11_3_14
void FASTCALL call_14_11_3_14(SH2_struct* sh)
{
  SH2movi(sh, 11, (3<<4)+14);
}
#define init_14_11_3_15
void FASTCALL call_14_11_3_15(SH2_struct* sh)
{
  SH2movi(sh, 11, (3<<4)+15);
}
#define init_14_11_4_0
void FASTCALL call_14_11_4_0(SH2_struct* sh)
{
  SH2movi(sh, 11, (4<<4)+0);
}
#define init_14_11_4_1
void FASTCALL call_14_11_4_1(SH2_struct* sh)
{
  SH2movi(sh, 11, (4<<4)+1);
}
#define init_14_11_4_2
void FASTCALL call_14_11_4_2(SH2_struct* sh)
{
  SH2movi(sh, 11, (4<<4)+2);
}
#define init_14_11_4_3
void FASTCALL call_14_11_4_3(SH2_struct* sh)
{
  SH2movi(sh, 11, (4<<4)+3);
}
#define init_14_11_4_4
void FASTCALL call_14_11_4_4(SH2_struct* sh)
{
  SH2movi(sh, 11, (4<<4)+4);
}
#define init_14_11_4_5
void FASTCALL call_14_11_4_5(SH2_struct* sh)
{
  SH2movi(sh, 11, (4<<4)+5);
}
#define init_14_11_4_6
void FASTCALL call_14_11_4_6(SH2_struct* sh)
{
  SH2movi(sh, 11, (4<<4)+6);
}
#define init_14_11_4_7
void FASTCALL call_14_11_4_7(SH2_struct* sh)
{
  SH2movi(sh, 11, (4<<4)+7);
}
#define init_14_11_4_8
void FASTCALL call_14_11_4_8(SH2_struct* sh)
{
  SH2movi(sh, 11, (4<<4)+8);
}
#define init_14_11_4_9
void FASTCALL call_14_11_4_9(SH2_struct* sh)
{
  SH2movi(sh, 11, (4<<4)+9);
}
#define init_14_11_4_10
void FASTCALL call_14_11_4_10(SH2_struct* sh)
{
  SH2movi(sh, 11, (4<<4)+10);
}
#define init_14_11_4_11
void FASTCALL call_14_11_4_11(SH2_struct* sh)
{
  SH2movi(sh, 11, (4<<4)+11);
}
#define init_14_11_4_12
void FASTCALL call_14_11_4_12(SH2_struct* sh)
{
  SH2movi(sh, 11, (4<<4)+12);
}
#define init_14_11_4_13
void FASTCALL call_14_11_4_13(SH2_struct* sh)
{
  SH2movi(sh, 11, (4<<4)+13);
}
#define init_14_11_4_14
void FASTCALL call_14_11_4_14(SH2_struct* sh)
{
  SH2movi(sh, 11, (4<<4)+14);
}
#define init_14_11_4_15
void FASTCALL call_14_11_4_15(SH2_struct* sh)
{
  SH2movi(sh, 11, (4<<4)+15);
}
#define init_14_11_5_0
void FASTCALL call_14_11_5_0(SH2_struct* sh)
{
  SH2movi(sh, 11, (5<<4)+0);
}
#define init_14_11_5_1
void FASTCALL call_14_11_5_1(SH2_struct* sh)
{
  SH2movi(sh, 11, (5<<4)+1);
}
#define init_14_11_5_2
void FASTCALL call_14_11_5_2(SH2_struct* sh)
{
  SH2movi(sh, 11, (5<<4)+2);
}
#define init_14_11_5_3
void FASTCALL call_14_11_5_3(SH2_struct* sh)
{
  SH2movi(sh, 11, (5<<4)+3);
}
#define init_14_11_5_4
void FASTCALL call_14_11_5_4(SH2_struct* sh)
{
  SH2movi(sh, 11, (5<<4)+4);
}
#define init_14_11_5_5
void FASTCALL call_14_11_5_5(SH2_struct* sh)
{
  SH2movi(sh, 11, (5<<4)+5);
}
#define init_14_11_5_6
void FASTCALL call_14_11_5_6(SH2_struct* sh)
{
  SH2movi(sh, 11, (5<<4)+6);
}
#define init_14_11_5_7
void FASTCALL call_14_11_5_7(SH2_struct* sh)
{
  SH2movi(sh, 11, (5<<4)+7);
}
#define init_14_11_5_8
void FASTCALL call_14_11_5_8(SH2_struct* sh)
{
  SH2movi(sh, 11, (5<<4)+8);
}
#define init_14_11_5_9
void FASTCALL call_14_11_5_9(SH2_struct* sh)
{
  SH2movi(sh, 11, (5<<4)+9);
}
#define init_14_11_5_10
void FASTCALL call_14_11_5_10(SH2_struct* sh)
{
  SH2movi(sh, 11, (5<<4)+10);
}
#define init_14_11_5_11
void FASTCALL call_14_11_5_11(SH2_struct* sh)
{
  SH2movi(sh, 11, (5<<4)+11);
}
#define init_14_11_5_12
void FASTCALL call_14_11_5_12(SH2_struct* sh)
{
  SH2movi(sh, 11, (5<<4)+12);
}
#define init_14_11_5_13
void FASTCALL call_14_11_5_13(SH2_struct* sh)
{
  SH2movi(sh, 11, (5<<4)+13);
}
#define init_14_11_5_14
void FASTCALL call_14_11_5_14(SH2_struct* sh)
{
  SH2movi(sh, 11, (5<<4)+14);
}
#define init_14_11_5_15
void FASTCALL call_14_11_5_15(SH2_struct* sh)
{
  SH2movi(sh, 11, (5<<4)+15);
}
#define init_14_11_6_0
void FASTCALL call_14_11_6_0(SH2_struct* sh)
{
  SH2movi(sh, 11, (6<<4)+0);
}
#define init_14_11_6_1
void FASTCALL call_14_11_6_1(SH2_struct* sh)
{
  SH2movi(sh, 11, (6<<4)+1);
}
#define init_14_11_6_2
void FASTCALL call_14_11_6_2(SH2_struct* sh)
{
  SH2movi(sh, 11, (6<<4)+2);
}
#define init_14_11_6_3
void FASTCALL call_14_11_6_3(SH2_struct* sh)
{
  SH2movi(sh, 11, (6<<4)+3);
}
#define init_14_11_6_4
void FASTCALL call_14_11_6_4(SH2_struct* sh)
{
  SH2movi(sh, 11, (6<<4)+4);
}
#define init_14_11_6_5
void FASTCALL call_14_11_6_5(SH2_struct* sh)
{
  SH2movi(sh, 11, (6<<4)+5);
}
#define init_14_11_6_6
void FASTCALL call_14_11_6_6(SH2_struct* sh)
{
  SH2movi(sh, 11, (6<<4)+6);
}
#define init_14_11_6_7
void FASTCALL call_14_11_6_7(SH2_struct* sh)
{
  SH2movi(sh, 11, (6<<4)+7);
}
#define init_14_11_6_8
void FASTCALL call_14_11_6_8(SH2_struct* sh)
{
  SH2movi(sh, 11, (6<<4)+8);
}
#define init_14_11_6_9
void FASTCALL call_14_11_6_9(SH2_struct* sh)
{
  SH2movi(sh, 11, (6<<4)+9);
}
#define init_14_11_6_10
void FASTCALL call_14_11_6_10(SH2_struct* sh)
{
  SH2movi(sh, 11, (6<<4)+10);
}
#define init_14_11_6_11
void FASTCALL call_14_11_6_11(SH2_struct* sh)
{
  SH2movi(sh, 11, (6<<4)+11);
}
#define init_14_11_6_12
void FASTCALL call_14_11_6_12(SH2_struct* sh)
{
  SH2movi(sh, 11, (6<<4)+12);
}
#define init_14_11_6_13
void FASTCALL call_14_11_6_13(SH2_struct* sh)
{
  SH2movi(sh, 11, (6<<4)+13);
}
#define init_14_11_6_14
void FASTCALL call_14_11_6_14(SH2_struct* sh)
{
  SH2movi(sh, 11, (6<<4)+14);
}
#define init_14_11_6_15
void FASTCALL call_14_11_6_15(SH2_struct* sh)
{
  SH2movi(sh, 11, (6<<4)+15);
}
#define init_14_11_7_0
void FASTCALL call_14_11_7_0(SH2_struct* sh)
{
  SH2movi(sh, 11, (7<<4)+0);
}
#define init_14_11_7_1
void FASTCALL call_14_11_7_1(SH2_struct* sh)
{
  SH2movi(sh, 11, (7<<4)+1);
}
#define init_14_11_7_2
void FASTCALL call_14_11_7_2(SH2_struct* sh)
{
  SH2movi(sh, 11, (7<<4)+2);
}
#define init_14_11_7_3
void FASTCALL call_14_11_7_3(SH2_struct* sh)
{
  SH2movi(sh, 11, (7<<4)+3);
}
#define init_14_11_7_4
void FASTCALL call_14_11_7_4(SH2_struct* sh)
{
  SH2movi(sh, 11, (7<<4)+4);
}
#define init_14_11_7_5
void FASTCALL call_14_11_7_5(SH2_struct* sh)
{
  SH2movi(sh, 11, (7<<4)+5);
}
#define init_14_11_7_6
void FASTCALL call_14_11_7_6(SH2_struct* sh)
{
  SH2movi(sh, 11, (7<<4)+6);
}
#define init_14_11_7_7
void FASTCALL call_14_11_7_7(SH2_struct* sh)
{
  SH2movi(sh, 11, (7<<4)+7);
}
#define init_14_11_7_8
void FASTCALL call_14_11_7_8(SH2_struct* sh)
{
  SH2movi(sh, 11, (7<<4)+8);
}
#define init_14_11_7_9
void FASTCALL call_14_11_7_9(SH2_struct* sh)
{
  SH2movi(sh, 11, (7<<4)+9);
}
#define init_14_11_7_10
void FASTCALL call_14_11_7_10(SH2_struct* sh)
{
  SH2movi(sh, 11, (7<<4)+10);
}
#define init_14_11_7_11
void FASTCALL call_14_11_7_11(SH2_struct* sh)
{
  SH2movi(sh, 11, (7<<4)+11);
}
#define init_14_11_7_12
void FASTCALL call_14_11_7_12(SH2_struct* sh)
{
  SH2movi(sh, 11, (7<<4)+12);
}
#define init_14_11_7_13
void FASTCALL call_14_11_7_13(SH2_struct* sh)
{
  SH2movi(sh, 11, (7<<4)+13);
}
#define init_14_11_7_14
void FASTCALL call_14_11_7_14(SH2_struct* sh)
{
  SH2movi(sh, 11, (7<<4)+14);
}
#define init_14_11_7_15
void FASTCALL call_14_11_7_15(SH2_struct* sh)
{
  SH2movi(sh, 11, (7<<4)+15);
}
#define init_14_11_8_0
void FASTCALL call_14_11_8_0(SH2_struct* sh)
{
  SH2movi(sh, 11, (8<<4)+0);
}
#define init_14_11_8_1
void FASTCALL call_14_11_8_1(SH2_struct* sh)
{
  SH2movi(sh, 11, (8<<4)+1);
}
#define init_14_11_8_2
void FASTCALL call_14_11_8_2(SH2_struct* sh)
{
  SH2movi(sh, 11, (8<<4)+2);
}
#define init_14_11_8_3
void FASTCALL call_14_11_8_3(SH2_struct* sh)
{
  SH2movi(sh, 11, (8<<4)+3);
}
#define init_14_11_8_4
void FASTCALL call_14_11_8_4(SH2_struct* sh)
{
  SH2movi(sh, 11, (8<<4)+4);
}
#define init_14_11_8_5
void FASTCALL call_14_11_8_5(SH2_struct* sh)
{
  SH2movi(sh, 11, (8<<4)+5);
}
#define init_14_11_8_6
void FASTCALL call_14_11_8_6(SH2_struct* sh)
{
  SH2movi(sh, 11, (8<<4)+6);
}
#define init_14_11_8_7
void FASTCALL call_14_11_8_7(SH2_struct* sh)
{
  SH2movi(sh, 11, (8<<4)+7);
}
#define init_14_11_8_8
void FASTCALL call_14_11_8_8(SH2_struct* sh)
{
  SH2movi(sh, 11, (8<<4)+8);
}
#define init_14_11_8_9
void FASTCALL call_14_11_8_9(SH2_struct* sh)
{
  SH2movi(sh, 11, (8<<4)+9);
}
#define init_14_11_8_10
void FASTCALL call_14_11_8_10(SH2_struct* sh)
{
  SH2movi(sh, 11, (8<<4)+10);
}
#define init_14_11_8_11
void FASTCALL call_14_11_8_11(SH2_struct* sh)
{
  SH2movi(sh, 11, (8<<4)+11);
}
#define init_14_11_8_12
void FASTCALL call_14_11_8_12(SH2_struct* sh)
{
  SH2movi(sh, 11, (8<<4)+12);
}
#define init_14_11_8_13
void FASTCALL call_14_11_8_13(SH2_struct* sh)
{
  SH2movi(sh, 11, (8<<4)+13);
}
#define init_14_11_8_14
void FASTCALL call_14_11_8_14(SH2_struct* sh)
{
  SH2movi(sh, 11, (8<<4)+14);
}
#define init_14_11_8_15
void FASTCALL call_14_11_8_15(SH2_struct* sh)
{
  SH2movi(sh, 11, (8<<4)+15);
}
#define init_14_11_9_0
void FASTCALL call_14_11_9_0(SH2_struct* sh)
{
  SH2movi(sh, 11, (9<<4)+0);
}
#define init_14_11_9_1
void FASTCALL call_14_11_9_1(SH2_struct* sh)
{
  SH2movi(sh, 11, (9<<4)+1);
}
#define init_14_11_9_2
void FASTCALL call_14_11_9_2(SH2_struct* sh)
{
  SH2movi(sh, 11, (9<<4)+2);
}
#define init_14_11_9_3
void FASTCALL call_14_11_9_3(SH2_struct* sh)
{
  SH2movi(sh, 11, (9<<4)+3);
}
#define init_14_11_9_4
void FASTCALL call_14_11_9_4(SH2_struct* sh)
{
  SH2movi(sh, 11, (9<<4)+4);
}
#define init_14_11_9_5
void FASTCALL call_14_11_9_5(SH2_struct* sh)
{
  SH2movi(sh, 11, (9<<4)+5);
}
#define init_14_11_9_6
void FASTCALL call_14_11_9_6(SH2_struct* sh)
{
  SH2movi(sh, 11, (9<<4)+6);
}
#define init_14_11_9_7
void FASTCALL call_14_11_9_7(SH2_struct* sh)
{
  SH2movi(sh, 11, (9<<4)+7);
}
#define init_14_11_9_8
void FASTCALL call_14_11_9_8(SH2_struct* sh)
{
  SH2movi(sh, 11, (9<<4)+8);
}
#define init_14_11_9_9
void FASTCALL call_14_11_9_9(SH2_struct* sh)
{
  SH2movi(sh, 11, (9<<4)+9);
}
#define init_14_11_9_10
void FASTCALL call_14_11_9_10(SH2_struct* sh)
{
  SH2movi(sh, 11, (9<<4)+10);
}
#define init_14_11_9_11
void FASTCALL call_14_11_9_11(SH2_struct* sh)
{
  SH2movi(sh, 11, (9<<4)+11);
}
#define init_14_11_9_12
void FASTCALL call_14_11_9_12(SH2_struct* sh)
{
  SH2movi(sh, 11, (9<<4)+12);
}
#define init_14_11_9_13
void FASTCALL call_14_11_9_13(SH2_struct* sh)
{
  SH2movi(sh, 11, (9<<4)+13);
}
#define init_14_11_9_14
void FASTCALL call_14_11_9_14(SH2_struct* sh)
{
  SH2movi(sh, 11, (9<<4)+14);
}
#define init_14_11_9_15
void FASTCALL call_14_11_9_15(SH2_struct* sh)
{
  SH2movi(sh, 11, (9<<4)+15);
}
#define init_14_11_10_0
void FASTCALL call_14_11_10_0(SH2_struct* sh)
{
  SH2movi(sh, 11, (10<<4)+0);
}
#define init_14_11_10_1
void FASTCALL call_14_11_10_1(SH2_struct* sh)
{
  SH2movi(sh, 11, (10<<4)+1);
}
#define init_14_11_10_2
void FASTCALL call_14_11_10_2(SH2_struct* sh)
{
  SH2movi(sh, 11, (10<<4)+2);
}
#define init_14_11_10_3
void FASTCALL call_14_11_10_3(SH2_struct* sh)
{
  SH2movi(sh, 11, (10<<4)+3);
}
#define init_14_11_10_4
void FASTCALL call_14_11_10_4(SH2_struct* sh)
{
  SH2movi(sh, 11, (10<<4)+4);
}
#define init_14_11_10_5
void FASTCALL call_14_11_10_5(SH2_struct* sh)
{
  SH2movi(sh, 11, (10<<4)+5);
}
#define init_14_11_10_6
void FASTCALL call_14_11_10_6(SH2_struct* sh)
{
  SH2movi(sh, 11, (10<<4)+6);
}
#define init_14_11_10_7
void FASTCALL call_14_11_10_7(SH2_struct* sh)
{
  SH2movi(sh, 11, (10<<4)+7);
}
#define init_14_11_10_8
void FASTCALL call_14_11_10_8(SH2_struct* sh)
{
  SH2movi(sh, 11, (10<<4)+8);
}
#define init_14_11_10_9
void FASTCALL call_14_11_10_9(SH2_struct* sh)
{
  SH2movi(sh, 11, (10<<4)+9);
}
#define init_14_11_10_10
void FASTCALL call_14_11_10_10(SH2_struct* sh)
{
  SH2movi(sh, 11, (10<<4)+10);
}
#define init_14_11_10_11
void FASTCALL call_14_11_10_11(SH2_struct* sh)
{
  SH2movi(sh, 11, (10<<4)+11);
}
#define init_14_11_10_12
void FASTCALL call_14_11_10_12(SH2_struct* sh)
{
  SH2movi(sh, 11, (10<<4)+12);
}
#define init_14_11_10_13
void FASTCALL call_14_11_10_13(SH2_struct* sh)
{
  SH2movi(sh, 11, (10<<4)+13);
}
#define init_14_11_10_14
void FASTCALL call_14_11_10_14(SH2_struct* sh)
{
  SH2movi(sh, 11, (10<<4)+14);
}
#define init_14_11_10_15
void FASTCALL call_14_11_10_15(SH2_struct* sh)
{
  SH2movi(sh, 11, (10<<4)+15);
}
#define init_14_11_11_0
void FASTCALL call_14_11_11_0(SH2_struct* sh)
{
  SH2movi(sh, 11, (11<<4)+0);
}
#define init_14_11_11_1
void FASTCALL call_14_11_11_1(SH2_struct* sh)
{
  SH2movi(sh, 11, (11<<4)+1);
}
#define init_14_11_11_2
void FASTCALL call_14_11_11_2(SH2_struct* sh)
{
  SH2movi(sh, 11, (11<<4)+2);
}
#define init_14_11_11_3
void FASTCALL call_14_11_11_3(SH2_struct* sh)
{
  SH2movi(sh, 11, (11<<4)+3);
}
#define init_14_11_11_4
void FASTCALL call_14_11_11_4(SH2_struct* sh)
{
  SH2movi(sh, 11, (11<<4)+4);
}
#define init_14_11_11_5
void FASTCALL call_14_11_11_5(SH2_struct* sh)
{
  SH2movi(sh, 11, (11<<4)+5);
}
#define init_14_11_11_6
void FASTCALL call_14_11_11_6(SH2_struct* sh)
{
  SH2movi(sh, 11, (11<<4)+6);
}
#define init_14_11_11_7
void FASTCALL call_14_11_11_7(SH2_struct* sh)
{
  SH2movi(sh, 11, (11<<4)+7);
}
#define init_14_11_11_8
void FASTCALL call_14_11_11_8(SH2_struct* sh)
{
  SH2movi(sh, 11, (11<<4)+8);
}
#define init_14_11_11_9
void FASTCALL call_14_11_11_9(SH2_struct* sh)
{
  SH2movi(sh, 11, (11<<4)+9);
}
#define init_14_11_11_10
void FASTCALL call_14_11_11_10(SH2_struct* sh)
{
  SH2movi(sh, 11, (11<<4)+10);
}
#define init_14_11_11_11
void FASTCALL call_14_11_11_11(SH2_struct* sh)
{
  SH2movi(sh, 11, (11<<4)+11);
}
#define init_14_11_11_12
void FASTCALL call_14_11_11_12(SH2_struct* sh)
{
  SH2movi(sh, 11, (11<<4)+12);
}
#define init_14_11_11_13
void FASTCALL call_14_11_11_13(SH2_struct* sh)
{
  SH2movi(sh, 11, (11<<4)+13);
}
#define init_14_11_11_14
void FASTCALL call_14_11_11_14(SH2_struct* sh)
{
  SH2movi(sh, 11, (11<<4)+14);
}
#define init_14_11_11_15
void FASTCALL call_14_11_11_15(SH2_struct* sh)
{
  SH2movi(sh, 11, (11<<4)+15);
}
#define init_14_11_12_0
void FASTCALL call_14_11_12_0(SH2_struct* sh)
{
  SH2movi(sh, 11, (12<<4)+0);
}
#define init_14_11_12_1
void FASTCALL call_14_11_12_1(SH2_struct* sh)
{
  SH2movi(sh, 11, (12<<4)+1);
}
#define init_14_11_12_2
void FASTCALL call_14_11_12_2(SH2_struct* sh)
{
  SH2movi(sh, 11, (12<<4)+2);
}
#define init_14_11_12_3
void FASTCALL call_14_11_12_3(SH2_struct* sh)
{
  SH2movi(sh, 11, (12<<4)+3);
}
#define init_14_11_12_4
void FASTCALL call_14_11_12_4(SH2_struct* sh)
{
  SH2movi(sh, 11, (12<<4)+4);
}
#define init_14_11_12_5
void FASTCALL call_14_11_12_5(SH2_struct* sh)
{
  SH2movi(sh, 11, (12<<4)+5);
}
#define init_14_11_12_6
void FASTCALL call_14_11_12_6(SH2_struct* sh)
{
  SH2movi(sh, 11, (12<<4)+6);
}
#define init_14_11_12_7
void FASTCALL call_14_11_12_7(SH2_struct* sh)
{
  SH2movi(sh, 11, (12<<4)+7);
}
#define init_14_11_12_8
void FASTCALL call_14_11_12_8(SH2_struct* sh)
{
  SH2movi(sh, 11, (12<<4)+8);
}
#define init_14_11_12_9
void FASTCALL call_14_11_12_9(SH2_struct* sh)
{
  SH2movi(sh, 11, (12<<4)+9);
}
#define init_14_11_12_10
void FASTCALL call_14_11_12_10(SH2_struct* sh)
{
  SH2movi(sh, 11, (12<<4)+10);
}
#define init_14_11_12_11
void FASTCALL call_14_11_12_11(SH2_struct* sh)
{
  SH2movi(sh, 11, (12<<4)+11);
}
#define init_14_11_12_12
void FASTCALL call_14_11_12_12(SH2_struct* sh)
{
  SH2movi(sh, 11, (12<<4)+12);
}
#define init_14_11_12_13
void FASTCALL call_14_11_12_13(SH2_struct* sh)
{
  SH2movi(sh, 11, (12<<4)+13);
}
#define init_14_11_12_14
void FASTCALL call_14_11_12_14(SH2_struct* sh)
{
  SH2movi(sh, 11, (12<<4)+14);
}
#define init_14_11_12_15
void FASTCALL call_14_11_12_15(SH2_struct* sh)
{
  SH2movi(sh, 11, (12<<4)+15);
}
#define init_14_11_13_0
void FASTCALL call_14_11_13_0(SH2_struct* sh)
{
  SH2movi(sh, 11, (13<<4)+0);
}
#define init_14_11_13_1
void FASTCALL call_14_11_13_1(SH2_struct* sh)
{
  SH2movi(sh, 11, (13<<4)+1);
}
#define init_14_11_13_2
void FASTCALL call_14_11_13_2(SH2_struct* sh)
{
  SH2movi(sh, 11, (13<<4)+2);
}
#define init_14_11_13_3
void FASTCALL call_14_11_13_3(SH2_struct* sh)
{
  SH2movi(sh, 11, (13<<4)+3);
}
#define init_14_11_13_4
void FASTCALL call_14_11_13_4(SH2_struct* sh)
{
  SH2movi(sh, 11, (13<<4)+4);
}
#define init_14_11_13_5
void FASTCALL call_14_11_13_5(SH2_struct* sh)
{
  SH2movi(sh, 11, (13<<4)+5);
}
#define init_14_11_13_6
void FASTCALL call_14_11_13_6(SH2_struct* sh)
{
  SH2movi(sh, 11, (13<<4)+6);
}
#define init_14_11_13_7
void FASTCALL call_14_11_13_7(SH2_struct* sh)
{
  SH2movi(sh, 11, (13<<4)+7);
}
#define init_14_11_13_8
void FASTCALL call_14_11_13_8(SH2_struct* sh)
{
  SH2movi(sh, 11, (13<<4)+8);
}
#define init_14_11_13_9
void FASTCALL call_14_11_13_9(SH2_struct* sh)
{
  SH2movi(sh, 11, (13<<4)+9);
}
#define init_14_11_13_10
void FASTCALL call_14_11_13_10(SH2_struct* sh)
{
  SH2movi(sh, 11, (13<<4)+10);
}
#define init_14_11_13_11
void FASTCALL call_14_11_13_11(SH2_struct* sh)
{
  SH2movi(sh, 11, (13<<4)+11);
}
#define init_14_11_13_12
void FASTCALL call_14_11_13_12(SH2_struct* sh)
{
  SH2movi(sh, 11, (13<<4)+12);
}
#define init_14_11_13_13
void FASTCALL call_14_11_13_13(SH2_struct* sh)
{
  SH2movi(sh, 11, (13<<4)+13);
}
#define init_14_11_13_14
void FASTCALL call_14_11_13_14(SH2_struct* sh)
{
  SH2movi(sh, 11, (13<<4)+14);
}
#define init_14_11_13_15
void FASTCALL call_14_11_13_15(SH2_struct* sh)
{
  SH2movi(sh, 11, (13<<4)+15);
}
#define init_14_11_14_0
void FASTCALL call_14_11_14_0(SH2_struct* sh)
{
  SH2movi(sh, 11, (14<<4)+0);
}
#define init_14_11_14_1
void FASTCALL call_14_11_14_1(SH2_struct* sh)
{
  SH2movi(sh, 11, (14<<4)+1);
}
#define init_14_11_14_2
void FASTCALL call_14_11_14_2(SH2_struct* sh)
{
  SH2movi(sh, 11, (14<<4)+2);
}
#define init_14_11_14_3
void FASTCALL call_14_11_14_3(SH2_struct* sh)
{
  SH2movi(sh, 11, (14<<4)+3);
}
#define init_14_11_14_4
void FASTCALL call_14_11_14_4(SH2_struct* sh)
{
  SH2movi(sh, 11, (14<<4)+4);
}
#define init_14_11_14_5
void FASTCALL call_14_11_14_5(SH2_struct* sh)
{
  SH2movi(sh, 11, (14<<4)+5);
}
#define init_14_11_14_6
void FASTCALL call_14_11_14_6(SH2_struct* sh)
{
  SH2movi(sh, 11, (14<<4)+6);
}
#define init_14_11_14_7
void FASTCALL call_14_11_14_7(SH2_struct* sh)
{
  SH2movi(sh, 11, (14<<4)+7);
}
#define init_14_11_14_8
void FASTCALL call_14_11_14_8(SH2_struct* sh)
{
  SH2movi(sh, 11, (14<<4)+8);
}
#define init_14_11_14_9
void FASTCALL call_14_11_14_9(SH2_struct* sh)
{
  SH2movi(sh, 11, (14<<4)+9);
}
#define init_14_11_14_10
void FASTCALL call_14_11_14_10(SH2_struct* sh)
{
  SH2movi(sh, 11, (14<<4)+10);
}
#define init_14_11_14_11
void FASTCALL call_14_11_14_11(SH2_struct* sh)
{
  SH2movi(sh, 11, (14<<4)+11);
}
#define init_14_11_14_12
void FASTCALL call_14_11_14_12(SH2_struct* sh)
{
  SH2movi(sh, 11, (14<<4)+12);
}
#define init_14_11_14_13
void FASTCALL call_14_11_14_13(SH2_struct* sh)
{
  SH2movi(sh, 11, (14<<4)+13);
}
#define init_14_11_14_14
void FASTCALL call_14_11_14_14(SH2_struct* sh)
{
  SH2movi(sh, 11, (14<<4)+14);
}
#define init_14_11_14_15
void FASTCALL call_14_11_14_15(SH2_struct* sh)
{
  SH2movi(sh, 11, (14<<4)+15);
}
#define init_14_11_15_0
void FASTCALL call_14_11_15_0(SH2_struct* sh)
{
  SH2movi(sh, 11, (15<<4)+0);
}
#define init_14_11_15_1
void FASTCALL call_14_11_15_1(SH2_struct* sh)
{
  SH2movi(sh, 11, (15<<4)+1);
}
#define init_14_11_15_2
void FASTCALL call_14_11_15_2(SH2_struct* sh)
{
  SH2movi(sh, 11, (15<<4)+2);
}
#define init_14_11_15_3
void FASTCALL call_14_11_15_3(SH2_struct* sh)
{
  SH2movi(sh, 11, (15<<4)+3);
}
#define init_14_11_15_4
void FASTCALL call_14_11_15_4(SH2_struct* sh)
{
  SH2movi(sh, 11, (15<<4)+4);
}
#define init_14_11_15_5
void FASTCALL call_14_11_15_5(SH2_struct* sh)
{
  SH2movi(sh, 11, (15<<4)+5);
}
#define init_14_11_15_6
void FASTCALL call_14_11_15_6(SH2_struct* sh)
{
  SH2movi(sh, 11, (15<<4)+6);
}
#define init_14_11_15_7
void FASTCALL call_14_11_15_7(SH2_struct* sh)
{
  SH2movi(sh, 11, (15<<4)+7);
}
#define init_14_11_15_8
void FASTCALL call_14_11_15_8(SH2_struct* sh)
{
  SH2movi(sh, 11, (15<<4)+8);
}
#define init_14_11_15_9
void FASTCALL call_14_11_15_9(SH2_struct* sh)
{
  SH2movi(sh, 11, (15<<4)+9);
}
#define init_14_11_15_10
void FASTCALL call_14_11_15_10(SH2_struct* sh)
{
  SH2movi(sh, 11, (15<<4)+10);
}
#define init_14_11_15_11
void FASTCALL call_14_11_15_11(SH2_struct* sh)
{
  SH2movi(sh, 11, (15<<4)+11);
}
#define init_14_11_15_12
void FASTCALL call_14_11_15_12(SH2_struct* sh)
{
  SH2movi(sh, 11, (15<<4)+12);
}
#define init_14_11_15_13
void FASTCALL call_14_11_15_13(SH2_struct* sh)
{
  SH2movi(sh, 11, (15<<4)+13);
}
#define init_14_11_15_14
void FASTCALL call_14_11_15_14(SH2_struct* sh)
{
  SH2movi(sh, 11, (15<<4)+14);
}
#define init_14_11_15_15
void FASTCALL call_14_11_15_15(SH2_struct* sh)
{
  SH2movi(sh, 11, (15<<4)+15);
}
#define init_14_12_0_0
void FASTCALL call_14_12_0_0(SH2_struct* sh)
{
  SH2movi(sh, 12, (0<<4)+0);
}
#define init_14_12_0_1
void FASTCALL call_14_12_0_1(SH2_struct* sh)
{
  SH2movi(sh, 12, (0<<4)+1);
}
#define init_14_12_0_2
void FASTCALL call_14_12_0_2(SH2_struct* sh)
{
  SH2movi(sh, 12, (0<<4)+2);
}
#define init_14_12_0_3
void FASTCALL call_14_12_0_3(SH2_struct* sh)
{
  SH2movi(sh, 12, (0<<4)+3);
}
#define init_14_12_0_4
void FASTCALL call_14_12_0_4(SH2_struct* sh)
{
  SH2movi(sh, 12, (0<<4)+4);
}
#define init_14_12_0_5
void FASTCALL call_14_12_0_5(SH2_struct* sh)
{
  SH2movi(sh, 12, (0<<4)+5);
}
#define init_14_12_0_6
void FASTCALL call_14_12_0_6(SH2_struct* sh)
{
  SH2movi(sh, 12, (0<<4)+6);
}
#define init_14_12_0_7
void FASTCALL call_14_12_0_7(SH2_struct* sh)
{
  SH2movi(sh, 12, (0<<4)+7);
}
#define init_14_12_0_8
void FASTCALL call_14_12_0_8(SH2_struct* sh)
{
  SH2movi(sh, 12, (0<<4)+8);
}
#define init_14_12_0_9
void FASTCALL call_14_12_0_9(SH2_struct* sh)
{
  SH2movi(sh, 12, (0<<4)+9);
}
#define init_14_12_0_10
void FASTCALL call_14_12_0_10(SH2_struct* sh)
{
  SH2movi(sh, 12, (0<<4)+10);
}
#define init_14_12_0_11
void FASTCALL call_14_12_0_11(SH2_struct* sh)
{
  SH2movi(sh, 12, (0<<4)+11);
}
#define init_14_12_0_12
void FASTCALL call_14_12_0_12(SH2_struct* sh)
{
  SH2movi(sh, 12, (0<<4)+12);
}
#define init_14_12_0_13
void FASTCALL call_14_12_0_13(SH2_struct* sh)
{
  SH2movi(sh, 12, (0<<4)+13);
}
#define init_14_12_0_14
void FASTCALL call_14_12_0_14(SH2_struct* sh)
{
  SH2movi(sh, 12, (0<<4)+14);
}
#define init_14_12_0_15
void FASTCALL call_14_12_0_15(SH2_struct* sh)
{
  SH2movi(sh, 12, (0<<4)+15);
}
#define init_14_12_1_0
void FASTCALL call_14_12_1_0(SH2_struct* sh)
{
  SH2movi(sh, 12, (1<<4)+0);
}
#define init_14_12_1_1
void FASTCALL call_14_12_1_1(SH2_struct* sh)
{
  SH2movi(sh, 12, (1<<4)+1);
}
#define init_14_12_1_2
void FASTCALL call_14_12_1_2(SH2_struct* sh)
{
  SH2movi(sh, 12, (1<<4)+2);
}
#define init_14_12_1_3
void FASTCALL call_14_12_1_3(SH2_struct* sh)
{
  SH2movi(sh, 12, (1<<4)+3);
}
#define init_14_12_1_4
void FASTCALL call_14_12_1_4(SH2_struct* sh)
{
  SH2movi(sh, 12, (1<<4)+4);
}
#define init_14_12_1_5
void FASTCALL call_14_12_1_5(SH2_struct* sh)
{
  SH2movi(sh, 12, (1<<4)+5);
}
#define init_14_12_1_6
void FASTCALL call_14_12_1_6(SH2_struct* sh)
{
  SH2movi(sh, 12, (1<<4)+6);
}
#define init_14_12_1_7
void FASTCALL call_14_12_1_7(SH2_struct* sh)
{
  SH2movi(sh, 12, (1<<4)+7);
}
#define init_14_12_1_8
void FASTCALL call_14_12_1_8(SH2_struct* sh)
{
  SH2movi(sh, 12, (1<<4)+8);
}
#define init_14_12_1_9
void FASTCALL call_14_12_1_9(SH2_struct* sh)
{
  SH2movi(sh, 12, (1<<4)+9);
}
#define init_14_12_1_10
void FASTCALL call_14_12_1_10(SH2_struct* sh)
{
  SH2movi(sh, 12, (1<<4)+10);
}
#define init_14_12_1_11
void FASTCALL call_14_12_1_11(SH2_struct* sh)
{
  SH2movi(sh, 12, (1<<4)+11);
}
#define init_14_12_1_12
void FASTCALL call_14_12_1_12(SH2_struct* sh)
{
  SH2movi(sh, 12, (1<<4)+12);
}
#define init_14_12_1_13
void FASTCALL call_14_12_1_13(SH2_struct* sh)
{
  SH2movi(sh, 12, (1<<4)+13);
}
#define init_14_12_1_14
void FASTCALL call_14_12_1_14(SH2_struct* sh)
{
  SH2movi(sh, 12, (1<<4)+14);
}
#define init_14_12_1_15
void FASTCALL call_14_12_1_15(SH2_struct* sh)
{
  SH2movi(sh, 12, (1<<4)+15);
}
#define init_14_12_2_0
void FASTCALL call_14_12_2_0(SH2_struct* sh)
{
  SH2movi(sh, 12, (2<<4)+0);
}
#define init_14_12_2_1
void FASTCALL call_14_12_2_1(SH2_struct* sh)
{
  SH2movi(sh, 12, (2<<4)+1);
}
#define init_14_12_2_2
void FASTCALL call_14_12_2_2(SH2_struct* sh)
{
  SH2movi(sh, 12, (2<<4)+2);
}
#define init_14_12_2_3
void FASTCALL call_14_12_2_3(SH2_struct* sh)
{
  SH2movi(sh, 12, (2<<4)+3);
}
#define init_14_12_2_4
void FASTCALL call_14_12_2_4(SH2_struct* sh)
{
  SH2movi(sh, 12, (2<<4)+4);
}
#define init_14_12_2_5
void FASTCALL call_14_12_2_5(SH2_struct* sh)
{
  SH2movi(sh, 12, (2<<4)+5);
}
#define init_14_12_2_6
void FASTCALL call_14_12_2_6(SH2_struct* sh)
{
  SH2movi(sh, 12, (2<<4)+6);
}
#define init_14_12_2_7
void FASTCALL call_14_12_2_7(SH2_struct* sh)
{
  SH2movi(sh, 12, (2<<4)+7);
}
#define init_14_12_2_8
void FASTCALL call_14_12_2_8(SH2_struct* sh)
{
  SH2movi(sh, 12, (2<<4)+8);
}
#define init_14_12_2_9
void FASTCALL call_14_12_2_9(SH2_struct* sh)
{
  SH2movi(sh, 12, (2<<4)+9);
}
#define init_14_12_2_10
void FASTCALL call_14_12_2_10(SH2_struct* sh)
{
  SH2movi(sh, 12, (2<<4)+10);
}
#define init_14_12_2_11
void FASTCALL call_14_12_2_11(SH2_struct* sh)
{
  SH2movi(sh, 12, (2<<4)+11);
}
#define init_14_12_2_12
void FASTCALL call_14_12_2_12(SH2_struct* sh)
{
  SH2movi(sh, 12, (2<<4)+12);
}
#define init_14_12_2_13
void FASTCALL call_14_12_2_13(SH2_struct* sh)
{
  SH2movi(sh, 12, (2<<4)+13);
}
#define init_14_12_2_14
void FASTCALL call_14_12_2_14(SH2_struct* sh)
{
  SH2movi(sh, 12, (2<<4)+14);
}
#define init_14_12_2_15
void FASTCALL call_14_12_2_15(SH2_struct* sh)
{
  SH2movi(sh, 12, (2<<4)+15);
}
#define init_14_12_3_0
void FASTCALL call_14_12_3_0(SH2_struct* sh)
{
  SH2movi(sh, 12, (3<<4)+0);
}
#define init_14_12_3_1
void FASTCALL call_14_12_3_1(SH2_struct* sh)
{
  SH2movi(sh, 12, (3<<4)+1);
}
#define init_14_12_3_2
void FASTCALL call_14_12_3_2(SH2_struct* sh)
{
  SH2movi(sh, 12, (3<<4)+2);
}
#define init_14_12_3_3
void FASTCALL call_14_12_3_3(SH2_struct* sh)
{
  SH2movi(sh, 12, (3<<4)+3);
}
#define init_14_12_3_4
void FASTCALL call_14_12_3_4(SH2_struct* sh)
{
  SH2movi(sh, 12, (3<<4)+4);
}
#define init_14_12_3_5
void FASTCALL call_14_12_3_5(SH2_struct* sh)
{
  SH2movi(sh, 12, (3<<4)+5);
}
#define init_14_12_3_6
void FASTCALL call_14_12_3_6(SH2_struct* sh)
{
  SH2movi(sh, 12, (3<<4)+6);
}
#define init_14_12_3_7
void FASTCALL call_14_12_3_7(SH2_struct* sh)
{
  SH2movi(sh, 12, (3<<4)+7);
}
#define init_14_12_3_8
void FASTCALL call_14_12_3_8(SH2_struct* sh)
{
  SH2movi(sh, 12, (3<<4)+8);
}
#define init_14_12_3_9
void FASTCALL call_14_12_3_9(SH2_struct* sh)
{
  SH2movi(sh, 12, (3<<4)+9);
}
#define init_14_12_3_10
void FASTCALL call_14_12_3_10(SH2_struct* sh)
{
  SH2movi(sh, 12, (3<<4)+10);
}
#define init_14_12_3_11
void FASTCALL call_14_12_3_11(SH2_struct* sh)
{
  SH2movi(sh, 12, (3<<4)+11);
}
#define init_14_12_3_12
void FASTCALL call_14_12_3_12(SH2_struct* sh)
{
  SH2movi(sh, 12, (3<<4)+12);
}
#define init_14_12_3_13
void FASTCALL call_14_12_3_13(SH2_struct* sh)
{
  SH2movi(sh, 12, (3<<4)+13);
}
#define init_14_12_3_14
void FASTCALL call_14_12_3_14(SH2_struct* sh)
{
  SH2movi(sh, 12, (3<<4)+14);
}
#define init_14_12_3_15
void FASTCALL call_14_12_3_15(SH2_struct* sh)
{
  SH2movi(sh, 12, (3<<4)+15);
}
#define init_14_12_4_0
void FASTCALL call_14_12_4_0(SH2_struct* sh)
{
  SH2movi(sh, 12, (4<<4)+0);
}
#define init_14_12_4_1
void FASTCALL call_14_12_4_1(SH2_struct* sh)
{
  SH2movi(sh, 12, (4<<4)+1);
}
#define init_14_12_4_2
void FASTCALL call_14_12_4_2(SH2_struct* sh)
{
  SH2movi(sh, 12, (4<<4)+2);
}
#define init_14_12_4_3
void FASTCALL call_14_12_4_3(SH2_struct* sh)
{
  SH2movi(sh, 12, (4<<4)+3);
}
#define init_14_12_4_4
void FASTCALL call_14_12_4_4(SH2_struct* sh)
{
  SH2movi(sh, 12, (4<<4)+4);
}
#define init_14_12_4_5
void FASTCALL call_14_12_4_5(SH2_struct* sh)
{
  SH2movi(sh, 12, (4<<4)+5);
}
#define init_14_12_4_6
void FASTCALL call_14_12_4_6(SH2_struct* sh)
{
  SH2movi(sh, 12, (4<<4)+6);
}
#define init_14_12_4_7
void FASTCALL call_14_12_4_7(SH2_struct* sh)
{
  SH2movi(sh, 12, (4<<4)+7);
}
#define init_14_12_4_8
void FASTCALL call_14_12_4_8(SH2_struct* sh)
{
  SH2movi(sh, 12, (4<<4)+8);
}
#define init_14_12_4_9
void FASTCALL call_14_12_4_9(SH2_struct* sh)
{
  SH2movi(sh, 12, (4<<4)+9);
}
#define init_14_12_4_10
void FASTCALL call_14_12_4_10(SH2_struct* sh)
{
  SH2movi(sh, 12, (4<<4)+10);
}
#define init_14_12_4_11
void FASTCALL call_14_12_4_11(SH2_struct* sh)
{
  SH2movi(sh, 12, (4<<4)+11);
}
#define init_14_12_4_12
void FASTCALL call_14_12_4_12(SH2_struct* sh)
{
  SH2movi(sh, 12, (4<<4)+12);
}
#define init_14_12_4_13
void FASTCALL call_14_12_4_13(SH2_struct* sh)
{
  SH2movi(sh, 12, (4<<4)+13);
}
#define init_14_12_4_14
void FASTCALL call_14_12_4_14(SH2_struct* sh)
{
  SH2movi(sh, 12, (4<<4)+14);
}
#define init_14_12_4_15
void FASTCALL call_14_12_4_15(SH2_struct* sh)
{
  SH2movi(sh, 12, (4<<4)+15);
}
#define init_14_12_5_0
void FASTCALL call_14_12_5_0(SH2_struct* sh)
{
  SH2movi(sh, 12, (5<<4)+0);
}
#define init_14_12_5_1
void FASTCALL call_14_12_5_1(SH2_struct* sh)
{
  SH2movi(sh, 12, (5<<4)+1);
}
#define init_14_12_5_2
void FASTCALL call_14_12_5_2(SH2_struct* sh)
{
  SH2movi(sh, 12, (5<<4)+2);
}
#define init_14_12_5_3
void FASTCALL call_14_12_5_3(SH2_struct* sh)
{
  SH2movi(sh, 12, (5<<4)+3);
}
#define init_14_12_5_4
void FASTCALL call_14_12_5_4(SH2_struct* sh)
{
  SH2movi(sh, 12, (5<<4)+4);
}
#define init_14_12_5_5
void FASTCALL call_14_12_5_5(SH2_struct* sh)
{
  SH2movi(sh, 12, (5<<4)+5);
}
#define init_14_12_5_6
void FASTCALL call_14_12_5_6(SH2_struct* sh)
{
  SH2movi(sh, 12, (5<<4)+6);
}
#define init_14_12_5_7
void FASTCALL call_14_12_5_7(SH2_struct* sh)
{
  SH2movi(sh, 12, (5<<4)+7);
}
#define init_14_12_5_8
void FASTCALL call_14_12_5_8(SH2_struct* sh)
{
  SH2movi(sh, 12, (5<<4)+8);
}
#define init_14_12_5_9
void FASTCALL call_14_12_5_9(SH2_struct* sh)
{
  SH2movi(sh, 12, (5<<4)+9);
}
#define init_14_12_5_10
void FASTCALL call_14_12_5_10(SH2_struct* sh)
{
  SH2movi(sh, 12, (5<<4)+10);
}
#define init_14_12_5_11
void FASTCALL call_14_12_5_11(SH2_struct* sh)
{
  SH2movi(sh, 12, (5<<4)+11);
}
#define init_14_12_5_12
void FASTCALL call_14_12_5_12(SH2_struct* sh)
{
  SH2movi(sh, 12, (5<<4)+12);
}
#define init_14_12_5_13
void FASTCALL call_14_12_5_13(SH2_struct* sh)
{
  SH2movi(sh, 12, (5<<4)+13);
}
#define init_14_12_5_14
void FASTCALL call_14_12_5_14(SH2_struct* sh)
{
  SH2movi(sh, 12, (5<<4)+14);
}
#define init_14_12_5_15
void FASTCALL call_14_12_5_15(SH2_struct* sh)
{
  SH2movi(sh, 12, (5<<4)+15);
}
#define init_14_12_6_0
void FASTCALL call_14_12_6_0(SH2_struct* sh)
{
  SH2movi(sh, 12, (6<<4)+0);
}
#define init_14_12_6_1
void FASTCALL call_14_12_6_1(SH2_struct* sh)
{
  SH2movi(sh, 12, (6<<4)+1);
}
#define init_14_12_6_2
void FASTCALL call_14_12_6_2(SH2_struct* sh)
{
  SH2movi(sh, 12, (6<<4)+2);
}
#define init_14_12_6_3
void FASTCALL call_14_12_6_3(SH2_struct* sh)
{
  SH2movi(sh, 12, (6<<4)+3);
}
#define init_14_12_6_4
void FASTCALL call_14_12_6_4(SH2_struct* sh)
{
  SH2movi(sh, 12, (6<<4)+4);
}
#define init_14_12_6_5
void FASTCALL call_14_12_6_5(SH2_struct* sh)
{
  SH2movi(sh, 12, (6<<4)+5);
}
#define init_14_12_6_6
void FASTCALL call_14_12_6_6(SH2_struct* sh)
{
  SH2movi(sh, 12, (6<<4)+6);
}
#define init_14_12_6_7
void FASTCALL call_14_12_6_7(SH2_struct* sh)
{
  SH2movi(sh, 12, (6<<4)+7);
}
#define init_14_12_6_8
void FASTCALL call_14_12_6_8(SH2_struct* sh)
{
  SH2movi(sh, 12, (6<<4)+8);
}
#define init_14_12_6_9
void FASTCALL call_14_12_6_9(SH2_struct* sh)
{
  SH2movi(sh, 12, (6<<4)+9);
}
#define init_14_12_6_10
void FASTCALL call_14_12_6_10(SH2_struct* sh)
{
  SH2movi(sh, 12, (6<<4)+10);
}
#define init_14_12_6_11
void FASTCALL call_14_12_6_11(SH2_struct* sh)
{
  SH2movi(sh, 12, (6<<4)+11);
}
#define init_14_12_6_12
void FASTCALL call_14_12_6_12(SH2_struct* sh)
{
  SH2movi(sh, 12, (6<<4)+12);
}
#define init_14_12_6_13
void FASTCALL call_14_12_6_13(SH2_struct* sh)
{
  SH2movi(sh, 12, (6<<4)+13);
}
#define init_14_12_6_14
void FASTCALL call_14_12_6_14(SH2_struct* sh)
{
  SH2movi(sh, 12, (6<<4)+14);
}
#define init_14_12_6_15
void FASTCALL call_14_12_6_15(SH2_struct* sh)
{
  SH2movi(sh, 12, (6<<4)+15);
}
#define init_14_12_7_0
void FASTCALL call_14_12_7_0(SH2_struct* sh)
{
  SH2movi(sh, 12, (7<<4)+0);
}
#define init_14_12_7_1
void FASTCALL call_14_12_7_1(SH2_struct* sh)
{
  SH2movi(sh, 12, (7<<4)+1);
}
#define init_14_12_7_2
void FASTCALL call_14_12_7_2(SH2_struct* sh)
{
  SH2movi(sh, 12, (7<<4)+2);
}
#define init_14_12_7_3
void FASTCALL call_14_12_7_3(SH2_struct* sh)
{
  SH2movi(sh, 12, (7<<4)+3);
}
#define init_14_12_7_4
void FASTCALL call_14_12_7_4(SH2_struct* sh)
{
  SH2movi(sh, 12, (7<<4)+4);
}
#define init_14_12_7_5
void FASTCALL call_14_12_7_5(SH2_struct* sh)
{
  SH2movi(sh, 12, (7<<4)+5);
}
#define init_14_12_7_6
void FASTCALL call_14_12_7_6(SH2_struct* sh)
{
  SH2movi(sh, 12, (7<<4)+6);
}
#define init_14_12_7_7
void FASTCALL call_14_12_7_7(SH2_struct* sh)
{
  SH2movi(sh, 12, (7<<4)+7);
}
#define init_14_12_7_8
void FASTCALL call_14_12_7_8(SH2_struct* sh)
{
  SH2movi(sh, 12, (7<<4)+8);
}
#define init_14_12_7_9
void FASTCALL call_14_12_7_9(SH2_struct* sh)
{
  SH2movi(sh, 12, (7<<4)+9);
}
#define init_14_12_7_10
void FASTCALL call_14_12_7_10(SH2_struct* sh)
{
  SH2movi(sh, 12, (7<<4)+10);
}
#define init_14_12_7_11
void FASTCALL call_14_12_7_11(SH2_struct* sh)
{
  SH2movi(sh, 12, (7<<4)+11);
}
#define init_14_12_7_12
void FASTCALL call_14_12_7_12(SH2_struct* sh)
{
  SH2movi(sh, 12, (7<<4)+12);
}
#define init_14_12_7_13
void FASTCALL call_14_12_7_13(SH2_struct* sh)
{
  SH2movi(sh, 12, (7<<4)+13);
}
#define init_14_12_7_14
void FASTCALL call_14_12_7_14(SH2_struct* sh)
{
  SH2movi(sh, 12, (7<<4)+14);
}
#define init_14_12_7_15
void FASTCALL call_14_12_7_15(SH2_struct* sh)
{
  SH2movi(sh, 12, (7<<4)+15);
}
#define init_14_12_8_0
void FASTCALL call_14_12_8_0(SH2_struct* sh)
{
  SH2movi(sh, 12, (8<<4)+0);
}
#define init_14_12_8_1
void FASTCALL call_14_12_8_1(SH2_struct* sh)
{
  SH2movi(sh, 12, (8<<4)+1);
}
#define init_14_12_8_2
void FASTCALL call_14_12_8_2(SH2_struct* sh)
{
  SH2movi(sh, 12, (8<<4)+2);
}
#define init_14_12_8_3
void FASTCALL call_14_12_8_3(SH2_struct* sh)
{
  SH2movi(sh, 12, (8<<4)+3);
}
#define init_14_12_8_4
void FASTCALL call_14_12_8_4(SH2_struct* sh)
{
  SH2movi(sh, 12, (8<<4)+4);
}
#define init_14_12_8_5
void FASTCALL call_14_12_8_5(SH2_struct* sh)
{
  SH2movi(sh, 12, (8<<4)+5);
}
#define init_14_12_8_6
void FASTCALL call_14_12_8_6(SH2_struct* sh)
{
  SH2movi(sh, 12, (8<<4)+6);
}
#define init_14_12_8_7
void FASTCALL call_14_12_8_7(SH2_struct* sh)
{
  SH2movi(sh, 12, (8<<4)+7);
}
#define init_14_12_8_8
void FASTCALL call_14_12_8_8(SH2_struct* sh)
{
  SH2movi(sh, 12, (8<<4)+8);
}
#define init_14_12_8_9
void FASTCALL call_14_12_8_9(SH2_struct* sh)
{
  SH2movi(sh, 12, (8<<4)+9);
}
#define init_14_12_8_10
void FASTCALL call_14_12_8_10(SH2_struct* sh)
{
  SH2movi(sh, 12, (8<<4)+10);
}
#define init_14_12_8_11
void FASTCALL call_14_12_8_11(SH2_struct* sh)
{
  SH2movi(sh, 12, (8<<4)+11);
}
#define init_14_12_8_12
void FASTCALL call_14_12_8_12(SH2_struct* sh)
{
  SH2movi(sh, 12, (8<<4)+12);
}
#define init_14_12_8_13
void FASTCALL call_14_12_8_13(SH2_struct* sh)
{
  SH2movi(sh, 12, (8<<4)+13);
}
#define init_14_12_8_14
void FASTCALL call_14_12_8_14(SH2_struct* sh)
{
  SH2movi(sh, 12, (8<<4)+14);
}
#define init_14_12_8_15
void FASTCALL call_14_12_8_15(SH2_struct* sh)
{
  SH2movi(sh, 12, (8<<4)+15);
}
#define init_14_12_9_0
void FASTCALL call_14_12_9_0(SH2_struct* sh)
{
  SH2movi(sh, 12, (9<<4)+0);
}
#define init_14_12_9_1
void FASTCALL call_14_12_9_1(SH2_struct* sh)
{
  SH2movi(sh, 12, (9<<4)+1);
}
#define init_14_12_9_2
void FASTCALL call_14_12_9_2(SH2_struct* sh)
{
  SH2movi(sh, 12, (9<<4)+2);
}
#define init_14_12_9_3
void FASTCALL call_14_12_9_3(SH2_struct* sh)
{
  SH2movi(sh, 12, (9<<4)+3);
}
#define init_14_12_9_4
void FASTCALL call_14_12_9_4(SH2_struct* sh)
{
  SH2movi(sh, 12, (9<<4)+4);
}
#define init_14_12_9_5
void FASTCALL call_14_12_9_5(SH2_struct* sh)
{
  SH2movi(sh, 12, (9<<4)+5);
}
#define init_14_12_9_6
void FASTCALL call_14_12_9_6(SH2_struct* sh)
{
  SH2movi(sh, 12, (9<<4)+6);
}
#define init_14_12_9_7
void FASTCALL call_14_12_9_7(SH2_struct* sh)
{
  SH2movi(sh, 12, (9<<4)+7);
}
#define init_14_12_9_8
void FASTCALL call_14_12_9_8(SH2_struct* sh)
{
  SH2movi(sh, 12, (9<<4)+8);
}
#define init_14_12_9_9
void FASTCALL call_14_12_9_9(SH2_struct* sh)
{
  SH2movi(sh, 12, (9<<4)+9);
}
#define init_14_12_9_10
void FASTCALL call_14_12_9_10(SH2_struct* sh)
{
  SH2movi(sh, 12, (9<<4)+10);
}
#define init_14_12_9_11
void FASTCALL call_14_12_9_11(SH2_struct* sh)
{
  SH2movi(sh, 12, (9<<4)+11);
}
#define init_14_12_9_12
void FASTCALL call_14_12_9_12(SH2_struct* sh)
{
  SH2movi(sh, 12, (9<<4)+12);
}
#define init_14_12_9_13
void FASTCALL call_14_12_9_13(SH2_struct* sh)
{
  SH2movi(sh, 12, (9<<4)+13);
}
#define init_14_12_9_14
void FASTCALL call_14_12_9_14(SH2_struct* sh)
{
  SH2movi(sh, 12, (9<<4)+14);
}
#define init_14_12_9_15
void FASTCALL call_14_12_9_15(SH2_struct* sh)
{
  SH2movi(sh, 12, (9<<4)+15);
}
#define init_14_12_10_0
void FASTCALL call_14_12_10_0(SH2_struct* sh)
{
  SH2movi(sh, 12, (10<<4)+0);
}
#define init_14_12_10_1
void FASTCALL call_14_12_10_1(SH2_struct* sh)
{
  SH2movi(sh, 12, (10<<4)+1);
}
#define init_14_12_10_2
void FASTCALL call_14_12_10_2(SH2_struct* sh)
{
  SH2movi(sh, 12, (10<<4)+2);
}
#define init_14_12_10_3
void FASTCALL call_14_12_10_3(SH2_struct* sh)
{
  SH2movi(sh, 12, (10<<4)+3);
}
#define init_14_12_10_4
void FASTCALL call_14_12_10_4(SH2_struct* sh)
{
  SH2movi(sh, 12, (10<<4)+4);
}
#define init_14_12_10_5
void FASTCALL call_14_12_10_5(SH2_struct* sh)
{
  SH2movi(sh, 12, (10<<4)+5);
}
#define init_14_12_10_6
void FASTCALL call_14_12_10_6(SH2_struct* sh)
{
  SH2movi(sh, 12, (10<<4)+6);
}
#define init_14_12_10_7
void FASTCALL call_14_12_10_7(SH2_struct* sh)
{
  SH2movi(sh, 12, (10<<4)+7);
}
#define init_14_12_10_8
void FASTCALL call_14_12_10_8(SH2_struct* sh)
{
  SH2movi(sh, 12, (10<<4)+8);
}
#define init_14_12_10_9
void FASTCALL call_14_12_10_9(SH2_struct* sh)
{
  SH2movi(sh, 12, (10<<4)+9);
}
#define init_14_12_10_10
void FASTCALL call_14_12_10_10(SH2_struct* sh)
{
  SH2movi(sh, 12, (10<<4)+10);
}
#define init_14_12_10_11
void FASTCALL call_14_12_10_11(SH2_struct* sh)
{
  SH2movi(sh, 12, (10<<4)+11);
}
#define init_14_12_10_12
void FASTCALL call_14_12_10_12(SH2_struct* sh)
{
  SH2movi(sh, 12, (10<<4)+12);
}
#define init_14_12_10_13
void FASTCALL call_14_12_10_13(SH2_struct* sh)
{
  SH2movi(sh, 12, (10<<4)+13);
}
#define init_14_12_10_14
void FASTCALL call_14_12_10_14(SH2_struct* sh)
{
  SH2movi(sh, 12, (10<<4)+14);
}
#define init_14_12_10_15
void FASTCALL call_14_12_10_15(SH2_struct* sh)
{
  SH2movi(sh, 12, (10<<4)+15);
}
#define init_14_12_11_0
void FASTCALL call_14_12_11_0(SH2_struct* sh)
{
  SH2movi(sh, 12, (11<<4)+0);
}
#define init_14_12_11_1
void FASTCALL call_14_12_11_1(SH2_struct* sh)
{
  SH2movi(sh, 12, (11<<4)+1);
}
#define init_14_12_11_2
void FASTCALL call_14_12_11_2(SH2_struct* sh)
{
  SH2movi(sh, 12, (11<<4)+2);
}
#define init_14_12_11_3
void FASTCALL call_14_12_11_3(SH2_struct* sh)
{
  SH2movi(sh, 12, (11<<4)+3);
}
#define init_14_12_11_4
void FASTCALL call_14_12_11_4(SH2_struct* sh)
{
  SH2movi(sh, 12, (11<<4)+4);
}
#define init_14_12_11_5
void FASTCALL call_14_12_11_5(SH2_struct* sh)
{
  SH2movi(sh, 12, (11<<4)+5);
}
#define init_14_12_11_6
void FASTCALL call_14_12_11_6(SH2_struct* sh)
{
  SH2movi(sh, 12, (11<<4)+6);
}
#define init_14_12_11_7
void FASTCALL call_14_12_11_7(SH2_struct* sh)
{
  SH2movi(sh, 12, (11<<4)+7);
}
#define init_14_12_11_8
void FASTCALL call_14_12_11_8(SH2_struct* sh)
{
  SH2movi(sh, 12, (11<<4)+8);
}
#define init_14_12_11_9
void FASTCALL call_14_12_11_9(SH2_struct* sh)
{
  SH2movi(sh, 12, (11<<4)+9);
}
#define init_14_12_11_10
void FASTCALL call_14_12_11_10(SH2_struct* sh)
{
  SH2movi(sh, 12, (11<<4)+10);
}
#define init_14_12_11_11
void FASTCALL call_14_12_11_11(SH2_struct* sh)
{
  SH2movi(sh, 12, (11<<4)+11);
}
#define init_14_12_11_12
void FASTCALL call_14_12_11_12(SH2_struct* sh)
{
  SH2movi(sh, 12, (11<<4)+12);
}
#define init_14_12_11_13
void FASTCALL call_14_12_11_13(SH2_struct* sh)
{
  SH2movi(sh, 12, (11<<4)+13);
}
#define init_14_12_11_14
void FASTCALL call_14_12_11_14(SH2_struct* sh)
{
  SH2movi(sh, 12, (11<<4)+14);
}
#define init_14_12_11_15
void FASTCALL call_14_12_11_15(SH2_struct* sh)
{
  SH2movi(sh, 12, (11<<4)+15);
}
#define init_14_12_12_0
void FASTCALL call_14_12_12_0(SH2_struct* sh)
{
  SH2movi(sh, 12, (12<<4)+0);
}
#define init_14_12_12_1
void FASTCALL call_14_12_12_1(SH2_struct* sh)
{
  SH2movi(sh, 12, (12<<4)+1);
}
#define init_14_12_12_2
void FASTCALL call_14_12_12_2(SH2_struct* sh)
{
  SH2movi(sh, 12, (12<<4)+2);
}
#define init_14_12_12_3
void FASTCALL call_14_12_12_3(SH2_struct* sh)
{
  SH2movi(sh, 12, (12<<4)+3);
}
#define init_14_12_12_4
void FASTCALL call_14_12_12_4(SH2_struct* sh)
{
  SH2movi(sh, 12, (12<<4)+4);
}
#define init_14_12_12_5
void FASTCALL call_14_12_12_5(SH2_struct* sh)
{
  SH2movi(sh, 12, (12<<4)+5);
}
#define init_14_12_12_6
void FASTCALL call_14_12_12_6(SH2_struct* sh)
{
  SH2movi(sh, 12, (12<<4)+6);
}
#define init_14_12_12_7
void FASTCALL call_14_12_12_7(SH2_struct* sh)
{
  SH2movi(sh, 12, (12<<4)+7);
}
#define init_14_12_12_8
void FASTCALL call_14_12_12_8(SH2_struct* sh)
{
  SH2movi(sh, 12, (12<<4)+8);
}
#define init_14_12_12_9
void FASTCALL call_14_12_12_9(SH2_struct* sh)
{
  SH2movi(sh, 12, (12<<4)+9);
}
#define init_14_12_12_10
void FASTCALL call_14_12_12_10(SH2_struct* sh)
{
  SH2movi(sh, 12, (12<<4)+10);
}
#define init_14_12_12_11
void FASTCALL call_14_12_12_11(SH2_struct* sh)
{
  SH2movi(sh, 12, (12<<4)+11);
}
#define init_14_12_12_12
void FASTCALL call_14_12_12_12(SH2_struct* sh)
{
  SH2movi(sh, 12, (12<<4)+12);
}
#define init_14_12_12_13
void FASTCALL call_14_12_12_13(SH2_struct* sh)
{
  SH2movi(sh, 12, (12<<4)+13);
}
#define init_14_12_12_14
void FASTCALL call_14_12_12_14(SH2_struct* sh)
{
  SH2movi(sh, 12, (12<<4)+14);
}
#define init_14_12_12_15
void FASTCALL call_14_12_12_15(SH2_struct* sh)
{
  SH2movi(sh, 12, (12<<4)+15);
}
#define init_14_12_13_0
void FASTCALL call_14_12_13_0(SH2_struct* sh)
{
  SH2movi(sh, 12, (13<<4)+0);
}
#define init_14_12_13_1
void FASTCALL call_14_12_13_1(SH2_struct* sh)
{
  SH2movi(sh, 12, (13<<4)+1);
}
#define init_14_12_13_2
void FASTCALL call_14_12_13_2(SH2_struct* sh)
{
  SH2movi(sh, 12, (13<<4)+2);
}
#define init_14_12_13_3
void FASTCALL call_14_12_13_3(SH2_struct* sh)
{
  SH2movi(sh, 12, (13<<4)+3);
}
#define init_14_12_13_4
void FASTCALL call_14_12_13_4(SH2_struct* sh)
{
  SH2movi(sh, 12, (13<<4)+4);
}
#define init_14_12_13_5
void FASTCALL call_14_12_13_5(SH2_struct* sh)
{
  SH2movi(sh, 12, (13<<4)+5);
}
#define init_14_12_13_6
void FASTCALL call_14_12_13_6(SH2_struct* sh)
{
  SH2movi(sh, 12, (13<<4)+6);
}
#define init_14_12_13_7
void FASTCALL call_14_12_13_7(SH2_struct* sh)
{
  SH2movi(sh, 12, (13<<4)+7);
}
#define init_14_12_13_8
void FASTCALL call_14_12_13_8(SH2_struct* sh)
{
  SH2movi(sh, 12, (13<<4)+8);
}
#define init_14_12_13_9
void FASTCALL call_14_12_13_9(SH2_struct* sh)
{
  SH2movi(sh, 12, (13<<4)+9);
}
#define init_14_12_13_10
void FASTCALL call_14_12_13_10(SH2_struct* sh)
{
  SH2movi(sh, 12, (13<<4)+10);
}
#define init_14_12_13_11
void FASTCALL call_14_12_13_11(SH2_struct* sh)
{
  SH2movi(sh, 12, (13<<4)+11);
}
#define init_14_12_13_12
void FASTCALL call_14_12_13_12(SH2_struct* sh)
{
  SH2movi(sh, 12, (13<<4)+12);
}
#define init_14_12_13_13
void FASTCALL call_14_12_13_13(SH2_struct* sh)
{
  SH2movi(sh, 12, (13<<4)+13);
}
#define init_14_12_13_14
void FASTCALL call_14_12_13_14(SH2_struct* sh)
{
  SH2movi(sh, 12, (13<<4)+14);
}
#define init_14_12_13_15
void FASTCALL call_14_12_13_15(SH2_struct* sh)
{
  SH2movi(sh, 12, (13<<4)+15);
}
#define init_14_12_14_0
void FASTCALL call_14_12_14_0(SH2_struct* sh)
{
  SH2movi(sh, 12, (14<<4)+0);
}
#define init_14_12_14_1
void FASTCALL call_14_12_14_1(SH2_struct* sh)
{
  SH2movi(sh, 12, (14<<4)+1);
}
#define init_14_12_14_2
void FASTCALL call_14_12_14_2(SH2_struct* sh)
{
  SH2movi(sh, 12, (14<<4)+2);
}
#define init_14_12_14_3
void FASTCALL call_14_12_14_3(SH2_struct* sh)
{
  SH2movi(sh, 12, (14<<4)+3);
}
#define init_14_12_14_4
void FASTCALL call_14_12_14_4(SH2_struct* sh)
{
  SH2movi(sh, 12, (14<<4)+4);
}
#define init_14_12_14_5
void FASTCALL call_14_12_14_5(SH2_struct* sh)
{
  SH2movi(sh, 12, (14<<4)+5);
}
#define init_14_12_14_6
void FASTCALL call_14_12_14_6(SH2_struct* sh)
{
  SH2movi(sh, 12, (14<<4)+6);
}
#define init_14_12_14_7
void FASTCALL call_14_12_14_7(SH2_struct* sh)
{
  SH2movi(sh, 12, (14<<4)+7);
}
#define init_14_12_14_8
void FASTCALL call_14_12_14_8(SH2_struct* sh)
{
  SH2movi(sh, 12, (14<<4)+8);
}
#define init_14_12_14_9
void FASTCALL call_14_12_14_9(SH2_struct* sh)
{
  SH2movi(sh, 12, (14<<4)+9);
}
#define init_14_12_14_10
void FASTCALL call_14_12_14_10(SH2_struct* sh)
{
  SH2movi(sh, 12, (14<<4)+10);
}
#define init_14_12_14_11
void FASTCALL call_14_12_14_11(SH2_struct* sh)
{
  SH2movi(sh, 12, (14<<4)+11);
}
#define init_14_12_14_12
void FASTCALL call_14_12_14_12(SH2_struct* sh)
{
  SH2movi(sh, 12, (14<<4)+12);
}
#define init_14_12_14_13
void FASTCALL call_14_12_14_13(SH2_struct* sh)
{
  SH2movi(sh, 12, (14<<4)+13);
}
#define init_14_12_14_14
void FASTCALL call_14_12_14_14(SH2_struct* sh)
{
  SH2movi(sh, 12, (14<<4)+14);
}
#define init_14_12_14_15
void FASTCALL call_14_12_14_15(SH2_struct* sh)
{
  SH2movi(sh, 12, (14<<4)+15);
}
#define init_14_12_15_0
void FASTCALL call_14_12_15_0(SH2_struct* sh)
{
  SH2movi(sh, 12, (15<<4)+0);
}
#define init_14_12_15_1
void FASTCALL call_14_12_15_1(SH2_struct* sh)
{
  SH2movi(sh, 12, (15<<4)+1);
}
#define init_14_12_15_2
void FASTCALL call_14_12_15_2(SH2_struct* sh)
{
  SH2movi(sh, 12, (15<<4)+2);
}
#define init_14_12_15_3
void FASTCALL call_14_12_15_3(SH2_struct* sh)
{
  SH2movi(sh, 12, (15<<4)+3);
}
#define init_14_12_15_4
void FASTCALL call_14_12_15_4(SH2_struct* sh)
{
  SH2movi(sh, 12, (15<<4)+4);
}
#define init_14_12_15_5
void FASTCALL call_14_12_15_5(SH2_struct* sh)
{
  SH2movi(sh, 12, (15<<4)+5);
}
#define init_14_12_15_6
void FASTCALL call_14_12_15_6(SH2_struct* sh)
{
  SH2movi(sh, 12, (15<<4)+6);
}
#define init_14_12_15_7
void FASTCALL call_14_12_15_7(SH2_struct* sh)
{
  SH2movi(sh, 12, (15<<4)+7);
}
#define init_14_12_15_8
void FASTCALL call_14_12_15_8(SH2_struct* sh)
{
  SH2movi(sh, 12, (15<<4)+8);
}
#define init_14_12_15_9
void FASTCALL call_14_12_15_9(SH2_struct* sh)
{
  SH2movi(sh, 12, (15<<4)+9);
}
#define init_14_12_15_10
void FASTCALL call_14_12_15_10(SH2_struct* sh)
{
  SH2movi(sh, 12, (15<<4)+10);
}
#define init_14_12_15_11
void FASTCALL call_14_12_15_11(SH2_struct* sh)
{
  SH2movi(sh, 12, (15<<4)+11);
}
#define init_14_12_15_12
void FASTCALL call_14_12_15_12(SH2_struct* sh)
{
  SH2movi(sh, 12, (15<<4)+12);
}
#define init_14_12_15_13
void FASTCALL call_14_12_15_13(SH2_struct* sh)
{
  SH2movi(sh, 12, (15<<4)+13);
}
#define init_14_12_15_14
void FASTCALL call_14_12_15_14(SH2_struct* sh)
{
  SH2movi(sh, 12, (15<<4)+14);
}
#define init_14_12_15_15
void FASTCALL call_14_12_15_15(SH2_struct* sh)
{
  SH2movi(sh, 12, (15<<4)+15);
}
#define init_14_13_0_0
void FASTCALL call_14_13_0_0(SH2_struct* sh)
{
  SH2movi(sh, 13, (0<<4)+0);
}
#define init_14_13_0_1
void FASTCALL call_14_13_0_1(SH2_struct* sh)
{
  SH2movi(sh, 13, (0<<4)+1);
}
#define init_14_13_0_2
void FASTCALL call_14_13_0_2(SH2_struct* sh)
{
  SH2movi(sh, 13, (0<<4)+2);
}
#define init_14_13_0_3
void FASTCALL call_14_13_0_3(SH2_struct* sh)
{
  SH2movi(sh, 13, (0<<4)+3);
}
#define init_14_13_0_4
void FASTCALL call_14_13_0_4(SH2_struct* sh)
{
  SH2movi(sh, 13, (0<<4)+4);
}
#define init_14_13_0_5
void FASTCALL call_14_13_0_5(SH2_struct* sh)
{
  SH2movi(sh, 13, (0<<4)+5);
}
#define init_14_13_0_6
void FASTCALL call_14_13_0_6(SH2_struct* sh)
{
  SH2movi(sh, 13, (0<<4)+6);
}
#define init_14_13_0_7
void FASTCALL call_14_13_0_7(SH2_struct* sh)
{
  SH2movi(sh, 13, (0<<4)+7);
}
#define init_14_13_0_8
void FASTCALL call_14_13_0_8(SH2_struct* sh)
{
  SH2movi(sh, 13, (0<<4)+8);
}
#define init_14_13_0_9
void FASTCALL call_14_13_0_9(SH2_struct* sh)
{
  SH2movi(sh, 13, (0<<4)+9);
}
#define init_14_13_0_10
void FASTCALL call_14_13_0_10(SH2_struct* sh)
{
  SH2movi(sh, 13, (0<<4)+10);
}
#define init_14_13_0_11
void FASTCALL call_14_13_0_11(SH2_struct* sh)
{
  SH2movi(sh, 13, (0<<4)+11);
}
#define init_14_13_0_12
void FASTCALL call_14_13_0_12(SH2_struct* sh)
{
  SH2movi(sh, 13, (0<<4)+12);
}
#define init_14_13_0_13
void FASTCALL call_14_13_0_13(SH2_struct* sh)
{
  SH2movi(sh, 13, (0<<4)+13);
}
#define init_14_13_0_14
void FASTCALL call_14_13_0_14(SH2_struct* sh)
{
  SH2movi(sh, 13, (0<<4)+14);
}
#define init_14_13_0_15
void FASTCALL call_14_13_0_15(SH2_struct* sh)
{
  SH2movi(sh, 13, (0<<4)+15);
}
#define init_14_13_1_0
void FASTCALL call_14_13_1_0(SH2_struct* sh)
{
  SH2movi(sh, 13, (1<<4)+0);
}
#define init_14_13_1_1
void FASTCALL call_14_13_1_1(SH2_struct* sh)
{
  SH2movi(sh, 13, (1<<4)+1);
}
#define init_14_13_1_2
void FASTCALL call_14_13_1_2(SH2_struct* sh)
{
  SH2movi(sh, 13, (1<<4)+2);
}
#define init_14_13_1_3
void FASTCALL call_14_13_1_3(SH2_struct* sh)
{
  SH2movi(sh, 13, (1<<4)+3);
}
#define init_14_13_1_4
void FASTCALL call_14_13_1_4(SH2_struct* sh)
{
  SH2movi(sh, 13, (1<<4)+4);
}
#define init_14_13_1_5
void FASTCALL call_14_13_1_5(SH2_struct* sh)
{
  SH2movi(sh, 13, (1<<4)+5);
}
#define init_14_13_1_6
void FASTCALL call_14_13_1_6(SH2_struct* sh)
{
  SH2movi(sh, 13, (1<<4)+6);
}
#define init_14_13_1_7
void FASTCALL call_14_13_1_7(SH2_struct* sh)
{
  SH2movi(sh, 13, (1<<4)+7);
}
#define init_14_13_1_8
void FASTCALL call_14_13_1_8(SH2_struct* sh)
{
  SH2movi(sh, 13, (1<<4)+8);
}
#define init_14_13_1_9
void FASTCALL call_14_13_1_9(SH2_struct* sh)
{
  SH2movi(sh, 13, (1<<4)+9);
}
#define init_14_13_1_10
void FASTCALL call_14_13_1_10(SH2_struct* sh)
{
  SH2movi(sh, 13, (1<<4)+10);
}
#define init_14_13_1_11
void FASTCALL call_14_13_1_11(SH2_struct* sh)
{
  SH2movi(sh, 13, (1<<4)+11);
}
#define init_14_13_1_12
void FASTCALL call_14_13_1_12(SH2_struct* sh)
{
  SH2movi(sh, 13, (1<<4)+12);
}
#define init_14_13_1_13
void FASTCALL call_14_13_1_13(SH2_struct* sh)
{
  SH2movi(sh, 13, (1<<4)+13);
}
#define init_14_13_1_14
void FASTCALL call_14_13_1_14(SH2_struct* sh)
{
  SH2movi(sh, 13, (1<<4)+14);
}
#define init_14_13_1_15
void FASTCALL call_14_13_1_15(SH2_struct* sh)
{
  SH2movi(sh, 13, (1<<4)+15);
}
#define init_14_13_2_0
void FASTCALL call_14_13_2_0(SH2_struct* sh)
{
  SH2movi(sh, 13, (2<<4)+0);
}
#define init_14_13_2_1
void FASTCALL call_14_13_2_1(SH2_struct* sh)
{
  SH2movi(sh, 13, (2<<4)+1);
}
#define init_14_13_2_2
void FASTCALL call_14_13_2_2(SH2_struct* sh)
{
  SH2movi(sh, 13, (2<<4)+2);
}
#define init_14_13_2_3
void FASTCALL call_14_13_2_3(SH2_struct* sh)
{
  SH2movi(sh, 13, (2<<4)+3);
}
#define init_14_13_2_4
void FASTCALL call_14_13_2_4(SH2_struct* sh)
{
  SH2movi(sh, 13, (2<<4)+4);
}
#define init_14_13_2_5
void FASTCALL call_14_13_2_5(SH2_struct* sh)
{
  SH2movi(sh, 13, (2<<4)+5);
}
#define init_14_13_2_6
void FASTCALL call_14_13_2_6(SH2_struct* sh)
{
  SH2movi(sh, 13, (2<<4)+6);
}
#define init_14_13_2_7
void FASTCALL call_14_13_2_7(SH2_struct* sh)
{
  SH2movi(sh, 13, (2<<4)+7);
}
#define init_14_13_2_8
void FASTCALL call_14_13_2_8(SH2_struct* sh)
{
  SH2movi(sh, 13, (2<<4)+8);
}
#define init_14_13_2_9
void FASTCALL call_14_13_2_9(SH2_struct* sh)
{
  SH2movi(sh, 13, (2<<4)+9);
}
#define init_14_13_2_10
void FASTCALL call_14_13_2_10(SH2_struct* sh)
{
  SH2movi(sh, 13, (2<<4)+10);
}
#define init_14_13_2_11
void FASTCALL call_14_13_2_11(SH2_struct* sh)
{
  SH2movi(sh, 13, (2<<4)+11);
}
#define init_14_13_2_12
void FASTCALL call_14_13_2_12(SH2_struct* sh)
{
  SH2movi(sh, 13, (2<<4)+12);
}
#define init_14_13_2_13
void FASTCALL call_14_13_2_13(SH2_struct* sh)
{
  SH2movi(sh, 13, (2<<4)+13);
}
#define init_14_13_2_14
void FASTCALL call_14_13_2_14(SH2_struct* sh)
{
  SH2movi(sh, 13, (2<<4)+14);
}
#define init_14_13_2_15
void FASTCALL call_14_13_2_15(SH2_struct* sh)
{
  SH2movi(sh, 13, (2<<4)+15);
}
#define init_14_13_3_0
void FASTCALL call_14_13_3_0(SH2_struct* sh)
{
  SH2movi(sh, 13, (3<<4)+0);
}
#define init_14_13_3_1
void FASTCALL call_14_13_3_1(SH2_struct* sh)
{
  SH2movi(sh, 13, (3<<4)+1);
}
#define init_14_13_3_2
void FASTCALL call_14_13_3_2(SH2_struct* sh)
{
  SH2movi(sh, 13, (3<<4)+2);
}
#define init_14_13_3_3
void FASTCALL call_14_13_3_3(SH2_struct* sh)
{
  SH2movi(sh, 13, (3<<4)+3);
}
#define init_14_13_3_4
void FASTCALL call_14_13_3_4(SH2_struct* sh)
{
  SH2movi(sh, 13, (3<<4)+4);
}
#define init_14_13_3_5
void FASTCALL call_14_13_3_5(SH2_struct* sh)
{
  SH2movi(sh, 13, (3<<4)+5);
}
#define init_14_13_3_6
void FASTCALL call_14_13_3_6(SH2_struct* sh)
{
  SH2movi(sh, 13, (3<<4)+6);
}
#define init_14_13_3_7
void FASTCALL call_14_13_3_7(SH2_struct* sh)
{
  SH2movi(sh, 13, (3<<4)+7);
}
#define init_14_13_3_8
void FASTCALL call_14_13_3_8(SH2_struct* sh)
{
  SH2movi(sh, 13, (3<<4)+8);
}
#define init_14_13_3_9
void FASTCALL call_14_13_3_9(SH2_struct* sh)
{
  SH2movi(sh, 13, (3<<4)+9);
}
#define init_14_13_3_10
void FASTCALL call_14_13_3_10(SH2_struct* sh)
{
  SH2movi(sh, 13, (3<<4)+10);
}
#define init_14_13_3_11
void FASTCALL call_14_13_3_11(SH2_struct* sh)
{
  SH2movi(sh, 13, (3<<4)+11);
}
#define init_14_13_3_12
void FASTCALL call_14_13_3_12(SH2_struct* sh)
{
  SH2movi(sh, 13, (3<<4)+12);
}
#define init_14_13_3_13
void FASTCALL call_14_13_3_13(SH2_struct* sh)
{
  SH2movi(sh, 13, (3<<4)+13);
}
#define init_14_13_3_14
void FASTCALL call_14_13_3_14(SH2_struct* sh)
{
  SH2movi(sh, 13, (3<<4)+14);
}
#define init_14_13_3_15
void FASTCALL call_14_13_3_15(SH2_struct* sh)
{
  SH2movi(sh, 13, (3<<4)+15);
}
#define init_14_13_4_0
void FASTCALL call_14_13_4_0(SH2_struct* sh)
{
  SH2movi(sh, 13, (4<<4)+0);
}
#define init_14_13_4_1
void FASTCALL call_14_13_4_1(SH2_struct* sh)
{
  SH2movi(sh, 13, (4<<4)+1);
}
#define init_14_13_4_2
void FASTCALL call_14_13_4_2(SH2_struct* sh)
{
  SH2movi(sh, 13, (4<<4)+2);
}
#define init_14_13_4_3
void FASTCALL call_14_13_4_3(SH2_struct* sh)
{
  SH2movi(sh, 13, (4<<4)+3);
}
#define init_14_13_4_4
void FASTCALL call_14_13_4_4(SH2_struct* sh)
{
  SH2movi(sh, 13, (4<<4)+4);
}
#define init_14_13_4_5
void FASTCALL call_14_13_4_5(SH2_struct* sh)
{
  SH2movi(sh, 13, (4<<4)+5);
}
#define init_14_13_4_6
void FASTCALL call_14_13_4_6(SH2_struct* sh)
{
  SH2movi(sh, 13, (4<<4)+6);
}
#define init_14_13_4_7
void FASTCALL call_14_13_4_7(SH2_struct* sh)
{
  SH2movi(sh, 13, (4<<4)+7);
}
#define init_14_13_4_8
void FASTCALL call_14_13_4_8(SH2_struct* sh)
{
  SH2movi(sh, 13, (4<<4)+8);
}
#define init_14_13_4_9
void FASTCALL call_14_13_4_9(SH2_struct* sh)
{
  SH2movi(sh, 13, (4<<4)+9);
}
#define init_14_13_4_10
void FASTCALL call_14_13_4_10(SH2_struct* sh)
{
  SH2movi(sh, 13, (4<<4)+10);
}
#define init_14_13_4_11
void FASTCALL call_14_13_4_11(SH2_struct* sh)
{
  SH2movi(sh, 13, (4<<4)+11);
}
#define init_14_13_4_12
void FASTCALL call_14_13_4_12(SH2_struct* sh)
{
  SH2movi(sh, 13, (4<<4)+12);
}
#define init_14_13_4_13
void FASTCALL call_14_13_4_13(SH2_struct* sh)
{
  SH2movi(sh, 13, (4<<4)+13);
}
#define init_14_13_4_14
void FASTCALL call_14_13_4_14(SH2_struct* sh)
{
  SH2movi(sh, 13, (4<<4)+14);
}
#define init_14_13_4_15
void FASTCALL call_14_13_4_15(SH2_struct* sh)
{
  SH2movi(sh, 13, (4<<4)+15);
}
#define init_14_13_5_0
void FASTCALL call_14_13_5_0(SH2_struct* sh)
{
  SH2movi(sh, 13, (5<<4)+0);
}
#define init_14_13_5_1
void FASTCALL call_14_13_5_1(SH2_struct* sh)
{
  SH2movi(sh, 13, (5<<4)+1);
}
#define init_14_13_5_2
void FASTCALL call_14_13_5_2(SH2_struct* sh)
{
  SH2movi(sh, 13, (5<<4)+2);
}
#define init_14_13_5_3
void FASTCALL call_14_13_5_3(SH2_struct* sh)
{
  SH2movi(sh, 13, (5<<4)+3);
}
#define init_14_13_5_4
void FASTCALL call_14_13_5_4(SH2_struct* sh)
{
  SH2movi(sh, 13, (5<<4)+4);
}
#define init_14_13_5_5
void FASTCALL call_14_13_5_5(SH2_struct* sh)
{
  SH2movi(sh, 13, (5<<4)+5);
}
#define init_14_13_5_6
void FASTCALL call_14_13_5_6(SH2_struct* sh)
{
  SH2movi(sh, 13, (5<<4)+6);
}
#define init_14_13_5_7
void FASTCALL call_14_13_5_7(SH2_struct* sh)
{
  SH2movi(sh, 13, (5<<4)+7);
}
#define init_14_13_5_8
void FASTCALL call_14_13_5_8(SH2_struct* sh)
{
  SH2movi(sh, 13, (5<<4)+8);
}
#define init_14_13_5_9
void FASTCALL call_14_13_5_9(SH2_struct* sh)
{
  SH2movi(sh, 13, (5<<4)+9);
}
#define init_14_13_5_10
void FASTCALL call_14_13_5_10(SH2_struct* sh)
{
  SH2movi(sh, 13, (5<<4)+10);
}
#define init_14_13_5_11
void FASTCALL call_14_13_5_11(SH2_struct* sh)
{
  SH2movi(sh, 13, (5<<4)+11);
}
#define init_14_13_5_12
void FASTCALL call_14_13_5_12(SH2_struct* sh)
{
  SH2movi(sh, 13, (5<<4)+12);
}
#define init_14_13_5_13
void FASTCALL call_14_13_5_13(SH2_struct* sh)
{
  SH2movi(sh, 13, (5<<4)+13);
}
#define init_14_13_5_14
void FASTCALL call_14_13_5_14(SH2_struct* sh)
{
  SH2movi(sh, 13, (5<<4)+14);
}
#define init_14_13_5_15
void FASTCALL call_14_13_5_15(SH2_struct* sh)
{
  SH2movi(sh, 13, (5<<4)+15);
}
#define init_14_13_6_0
void FASTCALL call_14_13_6_0(SH2_struct* sh)
{
  SH2movi(sh, 13, (6<<4)+0);
}
#define init_14_13_6_1
void FASTCALL call_14_13_6_1(SH2_struct* sh)
{
  SH2movi(sh, 13, (6<<4)+1);
}
#define init_14_13_6_2
void FASTCALL call_14_13_6_2(SH2_struct* sh)
{
  SH2movi(sh, 13, (6<<4)+2);
}
#define init_14_13_6_3
void FASTCALL call_14_13_6_3(SH2_struct* sh)
{
  SH2movi(sh, 13, (6<<4)+3);
}
#define init_14_13_6_4
void FASTCALL call_14_13_6_4(SH2_struct* sh)
{
  SH2movi(sh, 13, (6<<4)+4);
}
#define init_14_13_6_5
void FASTCALL call_14_13_6_5(SH2_struct* sh)
{
  SH2movi(sh, 13, (6<<4)+5);
}
#define init_14_13_6_6
void FASTCALL call_14_13_6_6(SH2_struct* sh)
{
  SH2movi(sh, 13, (6<<4)+6);
}
#define init_14_13_6_7
void FASTCALL call_14_13_6_7(SH2_struct* sh)
{
  SH2movi(sh, 13, (6<<4)+7);
}
#define init_14_13_6_8
void FASTCALL call_14_13_6_8(SH2_struct* sh)
{
  SH2movi(sh, 13, (6<<4)+8);
}
#define init_14_13_6_9
void FASTCALL call_14_13_6_9(SH2_struct* sh)
{
  SH2movi(sh, 13, (6<<4)+9);
}
#define init_14_13_6_10
void FASTCALL call_14_13_6_10(SH2_struct* sh)
{
  SH2movi(sh, 13, (6<<4)+10);
}
#define init_14_13_6_11
void FASTCALL call_14_13_6_11(SH2_struct* sh)
{
  SH2movi(sh, 13, (6<<4)+11);
}
#define init_14_13_6_12
void FASTCALL call_14_13_6_12(SH2_struct* sh)
{
  SH2movi(sh, 13, (6<<4)+12);
}
#define init_14_13_6_13
void FASTCALL call_14_13_6_13(SH2_struct* sh)
{
  SH2movi(sh, 13, (6<<4)+13);
}
#define init_14_13_6_14
void FASTCALL call_14_13_6_14(SH2_struct* sh)
{
  SH2movi(sh, 13, (6<<4)+14);
}
#define init_14_13_6_15
void FASTCALL call_14_13_6_15(SH2_struct* sh)
{
  SH2movi(sh, 13, (6<<4)+15);
}
#define init_14_13_7_0
void FASTCALL call_14_13_7_0(SH2_struct* sh)
{
  SH2movi(sh, 13, (7<<4)+0);
}
#define init_14_13_7_1
void FASTCALL call_14_13_7_1(SH2_struct* sh)
{
  SH2movi(sh, 13, (7<<4)+1);
}
#define init_14_13_7_2
void FASTCALL call_14_13_7_2(SH2_struct* sh)
{
  SH2movi(sh, 13, (7<<4)+2);
}
#define init_14_13_7_3
void FASTCALL call_14_13_7_3(SH2_struct* sh)
{
  SH2movi(sh, 13, (7<<4)+3);
}
#define init_14_13_7_4
void FASTCALL call_14_13_7_4(SH2_struct* sh)
{
  SH2movi(sh, 13, (7<<4)+4);
}
#define init_14_13_7_5
void FASTCALL call_14_13_7_5(SH2_struct* sh)
{
  SH2movi(sh, 13, (7<<4)+5);
}
#define init_14_13_7_6
void FASTCALL call_14_13_7_6(SH2_struct* sh)
{
  SH2movi(sh, 13, (7<<4)+6);
}
#define init_14_13_7_7
void FASTCALL call_14_13_7_7(SH2_struct* sh)
{
  SH2movi(sh, 13, (7<<4)+7);
}
#define init_14_13_7_8
void FASTCALL call_14_13_7_8(SH2_struct* sh)
{
  SH2movi(sh, 13, (7<<4)+8);
}
#define init_14_13_7_9
void FASTCALL call_14_13_7_9(SH2_struct* sh)
{
  SH2movi(sh, 13, (7<<4)+9);
}
#define init_14_13_7_10
void FASTCALL call_14_13_7_10(SH2_struct* sh)
{
  SH2movi(sh, 13, (7<<4)+10);
}
#define init_14_13_7_11
void FASTCALL call_14_13_7_11(SH2_struct* sh)
{
  SH2movi(sh, 13, (7<<4)+11);
}
#define init_14_13_7_12
void FASTCALL call_14_13_7_12(SH2_struct* sh)
{
  SH2movi(sh, 13, (7<<4)+12);
}
#define init_14_13_7_13
void FASTCALL call_14_13_7_13(SH2_struct* sh)
{
  SH2movi(sh, 13, (7<<4)+13);
}
#define init_14_13_7_14
void FASTCALL call_14_13_7_14(SH2_struct* sh)
{
  SH2movi(sh, 13, (7<<4)+14);
}
#define init_14_13_7_15
void FASTCALL call_14_13_7_15(SH2_struct* sh)
{
  SH2movi(sh, 13, (7<<4)+15);
}
#define init_14_13_8_0
void FASTCALL call_14_13_8_0(SH2_struct* sh)
{
  SH2movi(sh, 13, (8<<4)+0);
}
#define init_14_13_8_1
void FASTCALL call_14_13_8_1(SH2_struct* sh)
{
  SH2movi(sh, 13, (8<<4)+1);
}
#define init_14_13_8_2
void FASTCALL call_14_13_8_2(SH2_struct* sh)
{
  SH2movi(sh, 13, (8<<4)+2);
}
#define init_14_13_8_3
void FASTCALL call_14_13_8_3(SH2_struct* sh)
{
  SH2movi(sh, 13, (8<<4)+3);
}
#define init_14_13_8_4
void FASTCALL call_14_13_8_4(SH2_struct* sh)
{
  SH2movi(sh, 13, (8<<4)+4);
}
#define init_14_13_8_5
void FASTCALL call_14_13_8_5(SH2_struct* sh)
{
  SH2movi(sh, 13, (8<<4)+5);
}
#define init_14_13_8_6
void FASTCALL call_14_13_8_6(SH2_struct* sh)
{
  SH2movi(sh, 13, (8<<4)+6);
}
#define init_14_13_8_7
void FASTCALL call_14_13_8_7(SH2_struct* sh)
{
  SH2movi(sh, 13, (8<<4)+7);
}
#define init_14_13_8_8
void FASTCALL call_14_13_8_8(SH2_struct* sh)
{
  SH2movi(sh, 13, (8<<4)+8);
}
#define init_14_13_8_9
void FASTCALL call_14_13_8_9(SH2_struct* sh)
{
  SH2movi(sh, 13, (8<<4)+9);
}
#define init_14_13_8_10
void FASTCALL call_14_13_8_10(SH2_struct* sh)
{
  SH2movi(sh, 13, (8<<4)+10);
}
#define init_14_13_8_11
void FASTCALL call_14_13_8_11(SH2_struct* sh)
{
  SH2movi(sh, 13, (8<<4)+11);
}
#define init_14_13_8_12
void FASTCALL call_14_13_8_12(SH2_struct* sh)
{
  SH2movi(sh, 13, (8<<4)+12);
}
#define init_14_13_8_13
void FASTCALL call_14_13_8_13(SH2_struct* sh)
{
  SH2movi(sh, 13, (8<<4)+13);
}
#define init_14_13_8_14
void FASTCALL call_14_13_8_14(SH2_struct* sh)
{
  SH2movi(sh, 13, (8<<4)+14);
}
#define init_14_13_8_15
void FASTCALL call_14_13_8_15(SH2_struct* sh)
{
  SH2movi(sh, 13, (8<<4)+15);
}
#define init_14_13_9_0
void FASTCALL call_14_13_9_0(SH2_struct* sh)
{
  SH2movi(sh, 13, (9<<4)+0);
}
#define init_14_13_9_1
void FASTCALL call_14_13_9_1(SH2_struct* sh)
{
  SH2movi(sh, 13, (9<<4)+1);
}
#define init_14_13_9_2
void FASTCALL call_14_13_9_2(SH2_struct* sh)
{
  SH2movi(sh, 13, (9<<4)+2);
}
#define init_14_13_9_3
void FASTCALL call_14_13_9_3(SH2_struct* sh)
{
  SH2movi(sh, 13, (9<<4)+3);
}
#define init_14_13_9_4
void FASTCALL call_14_13_9_4(SH2_struct* sh)
{
  SH2movi(sh, 13, (9<<4)+4);
}
#define init_14_13_9_5
void FASTCALL call_14_13_9_5(SH2_struct* sh)
{
  SH2movi(sh, 13, (9<<4)+5);
}
#define init_14_13_9_6
void FASTCALL call_14_13_9_6(SH2_struct* sh)
{
  SH2movi(sh, 13, (9<<4)+6);
}
#define init_14_13_9_7
void FASTCALL call_14_13_9_7(SH2_struct* sh)
{
  SH2movi(sh, 13, (9<<4)+7);
}
#define init_14_13_9_8
void FASTCALL call_14_13_9_8(SH2_struct* sh)
{
  SH2movi(sh, 13, (9<<4)+8);
}
#define init_14_13_9_9
void FASTCALL call_14_13_9_9(SH2_struct* sh)
{
  SH2movi(sh, 13, (9<<4)+9);
}
#define init_14_13_9_10
void FASTCALL call_14_13_9_10(SH2_struct* sh)
{
  SH2movi(sh, 13, (9<<4)+10);
}
#define init_14_13_9_11
void FASTCALL call_14_13_9_11(SH2_struct* sh)
{
  SH2movi(sh, 13, (9<<4)+11);
}
#define init_14_13_9_12
void FASTCALL call_14_13_9_12(SH2_struct* sh)
{
  SH2movi(sh, 13, (9<<4)+12);
}
#define init_14_13_9_13
void FASTCALL call_14_13_9_13(SH2_struct* sh)
{
  SH2movi(sh, 13, (9<<4)+13);
}
#define init_14_13_9_14
void FASTCALL call_14_13_9_14(SH2_struct* sh)
{
  SH2movi(sh, 13, (9<<4)+14);
}
#define init_14_13_9_15
void FASTCALL call_14_13_9_15(SH2_struct* sh)
{
  SH2movi(sh, 13, (9<<4)+15);
}
#define init_14_13_10_0
void FASTCALL call_14_13_10_0(SH2_struct* sh)
{
  SH2movi(sh, 13, (10<<4)+0);
}
#define init_14_13_10_1
void FASTCALL call_14_13_10_1(SH2_struct* sh)
{
  SH2movi(sh, 13, (10<<4)+1);
}
#define init_14_13_10_2
void FASTCALL call_14_13_10_2(SH2_struct* sh)
{
  SH2movi(sh, 13, (10<<4)+2);
}
#define init_14_13_10_3
void FASTCALL call_14_13_10_3(SH2_struct* sh)
{
  SH2movi(sh, 13, (10<<4)+3);
}
#define init_14_13_10_4
void FASTCALL call_14_13_10_4(SH2_struct* sh)
{
  SH2movi(sh, 13, (10<<4)+4);
}
#define init_14_13_10_5
void FASTCALL call_14_13_10_5(SH2_struct* sh)
{
  SH2movi(sh, 13, (10<<4)+5);
}
#define init_14_13_10_6
void FASTCALL call_14_13_10_6(SH2_struct* sh)
{
  SH2movi(sh, 13, (10<<4)+6);
}
#define init_14_13_10_7
void FASTCALL call_14_13_10_7(SH2_struct* sh)
{
  SH2movi(sh, 13, (10<<4)+7);
}
#define init_14_13_10_8
void FASTCALL call_14_13_10_8(SH2_struct* sh)
{
  SH2movi(sh, 13, (10<<4)+8);
}
#define init_14_13_10_9
void FASTCALL call_14_13_10_9(SH2_struct* sh)
{
  SH2movi(sh, 13, (10<<4)+9);
}
#define init_14_13_10_10
void FASTCALL call_14_13_10_10(SH2_struct* sh)
{
  SH2movi(sh, 13, (10<<4)+10);
}
#define init_14_13_10_11
void FASTCALL call_14_13_10_11(SH2_struct* sh)
{
  SH2movi(sh, 13, (10<<4)+11);
}
#define init_14_13_10_12
void FASTCALL call_14_13_10_12(SH2_struct* sh)
{
  SH2movi(sh, 13, (10<<4)+12);
}
#define init_14_13_10_13
void FASTCALL call_14_13_10_13(SH2_struct* sh)
{
  SH2movi(sh, 13, (10<<4)+13);
}
#define init_14_13_10_14
void FASTCALL call_14_13_10_14(SH2_struct* sh)
{
  SH2movi(sh, 13, (10<<4)+14);
}
#define init_14_13_10_15
void FASTCALL call_14_13_10_15(SH2_struct* sh)
{
  SH2movi(sh, 13, (10<<4)+15);
}
#define init_14_13_11_0
void FASTCALL call_14_13_11_0(SH2_struct* sh)
{
  SH2movi(sh, 13, (11<<4)+0);
}
#define init_14_13_11_1
void FASTCALL call_14_13_11_1(SH2_struct* sh)
{
  SH2movi(sh, 13, (11<<4)+1);
}
#define init_14_13_11_2
void FASTCALL call_14_13_11_2(SH2_struct* sh)
{
  SH2movi(sh, 13, (11<<4)+2);
}
#define init_14_13_11_3
void FASTCALL call_14_13_11_3(SH2_struct* sh)
{
  SH2movi(sh, 13, (11<<4)+3);
}
#define init_14_13_11_4
void FASTCALL call_14_13_11_4(SH2_struct* sh)
{
  SH2movi(sh, 13, (11<<4)+4);
}
#define init_14_13_11_5
void FASTCALL call_14_13_11_5(SH2_struct* sh)
{
  SH2movi(sh, 13, (11<<4)+5);
}
#define init_14_13_11_6
void FASTCALL call_14_13_11_6(SH2_struct* sh)
{
  SH2movi(sh, 13, (11<<4)+6);
}
#define init_14_13_11_7
void FASTCALL call_14_13_11_7(SH2_struct* sh)
{
  SH2movi(sh, 13, (11<<4)+7);
}
#define init_14_13_11_8
void FASTCALL call_14_13_11_8(SH2_struct* sh)
{
  SH2movi(sh, 13, (11<<4)+8);
}
#define init_14_13_11_9
void FASTCALL call_14_13_11_9(SH2_struct* sh)
{
  SH2movi(sh, 13, (11<<4)+9);
}
#define init_14_13_11_10
void FASTCALL call_14_13_11_10(SH2_struct* sh)
{
  SH2movi(sh, 13, (11<<4)+10);
}
#define init_14_13_11_11
void FASTCALL call_14_13_11_11(SH2_struct* sh)
{
  SH2movi(sh, 13, (11<<4)+11);
}
#define init_14_13_11_12
void FASTCALL call_14_13_11_12(SH2_struct* sh)
{
  SH2movi(sh, 13, (11<<4)+12);
}
#define init_14_13_11_13
void FASTCALL call_14_13_11_13(SH2_struct* sh)
{
  SH2movi(sh, 13, (11<<4)+13);
}
#define init_14_13_11_14
void FASTCALL call_14_13_11_14(SH2_struct* sh)
{
  SH2movi(sh, 13, (11<<4)+14);
}
#define init_14_13_11_15
void FASTCALL call_14_13_11_15(SH2_struct* sh)
{
  SH2movi(sh, 13, (11<<4)+15);
}
#define init_14_13_12_0
void FASTCALL call_14_13_12_0(SH2_struct* sh)
{
  SH2movi(sh, 13, (12<<4)+0);
}
#define init_14_13_12_1
void FASTCALL call_14_13_12_1(SH2_struct* sh)
{
  SH2movi(sh, 13, (12<<4)+1);
}
#define init_14_13_12_2
void FASTCALL call_14_13_12_2(SH2_struct* sh)
{
  SH2movi(sh, 13, (12<<4)+2);
}
#define init_14_13_12_3
void FASTCALL call_14_13_12_3(SH2_struct* sh)
{
  SH2movi(sh, 13, (12<<4)+3);
}
#define init_14_13_12_4
void FASTCALL call_14_13_12_4(SH2_struct* sh)
{
  SH2movi(sh, 13, (12<<4)+4);
}
#define init_14_13_12_5
void FASTCALL call_14_13_12_5(SH2_struct* sh)
{
  SH2movi(sh, 13, (12<<4)+5);
}
#define init_14_13_12_6
void FASTCALL call_14_13_12_6(SH2_struct* sh)
{
  SH2movi(sh, 13, (12<<4)+6);
}
#define init_14_13_12_7
void FASTCALL call_14_13_12_7(SH2_struct* sh)
{
  SH2movi(sh, 13, (12<<4)+7);
}
#define init_14_13_12_8
void FASTCALL call_14_13_12_8(SH2_struct* sh)
{
  SH2movi(sh, 13, (12<<4)+8);
}
#define init_14_13_12_9
void FASTCALL call_14_13_12_9(SH2_struct* sh)
{
  SH2movi(sh, 13, (12<<4)+9);
}
#define init_14_13_12_10
void FASTCALL call_14_13_12_10(SH2_struct* sh)
{
  SH2movi(sh, 13, (12<<4)+10);
}
#define init_14_13_12_11
void FASTCALL call_14_13_12_11(SH2_struct* sh)
{
  SH2movi(sh, 13, (12<<4)+11);
}
#define init_14_13_12_12
void FASTCALL call_14_13_12_12(SH2_struct* sh)
{
  SH2movi(sh, 13, (12<<4)+12);
}
#define init_14_13_12_13
void FASTCALL call_14_13_12_13(SH2_struct* sh)
{
  SH2movi(sh, 13, (12<<4)+13);
}
#define init_14_13_12_14
void FASTCALL call_14_13_12_14(SH2_struct* sh)
{
  SH2movi(sh, 13, (12<<4)+14);
}
#define init_14_13_12_15
void FASTCALL call_14_13_12_15(SH2_struct* sh)
{
  SH2movi(sh, 13, (12<<4)+15);
}
#define init_14_13_13_0
void FASTCALL call_14_13_13_0(SH2_struct* sh)
{
  SH2movi(sh, 13, (13<<4)+0);
}
#define init_14_13_13_1
void FASTCALL call_14_13_13_1(SH2_struct* sh)
{
  SH2movi(sh, 13, (13<<4)+1);
}
#define init_14_13_13_2
void FASTCALL call_14_13_13_2(SH2_struct* sh)
{
  SH2movi(sh, 13, (13<<4)+2);
}
#define init_14_13_13_3
void FASTCALL call_14_13_13_3(SH2_struct* sh)
{
  SH2movi(sh, 13, (13<<4)+3);
}
#define init_14_13_13_4
void FASTCALL call_14_13_13_4(SH2_struct* sh)
{
  SH2movi(sh, 13, (13<<4)+4);
}
#define init_14_13_13_5
void FASTCALL call_14_13_13_5(SH2_struct* sh)
{
  SH2movi(sh, 13, (13<<4)+5);
}
#define init_14_13_13_6
void FASTCALL call_14_13_13_6(SH2_struct* sh)
{
  SH2movi(sh, 13, (13<<4)+6);
}
#define init_14_13_13_7
void FASTCALL call_14_13_13_7(SH2_struct* sh)
{
  SH2movi(sh, 13, (13<<4)+7);
}
#define init_14_13_13_8
void FASTCALL call_14_13_13_8(SH2_struct* sh)
{
  SH2movi(sh, 13, (13<<4)+8);
}
#define init_14_13_13_9
void FASTCALL call_14_13_13_9(SH2_struct* sh)
{
  SH2movi(sh, 13, (13<<4)+9);
}
#define init_14_13_13_10
void FASTCALL call_14_13_13_10(SH2_struct* sh)
{
  SH2movi(sh, 13, (13<<4)+10);
}
#define init_14_13_13_11
void FASTCALL call_14_13_13_11(SH2_struct* sh)
{
  SH2movi(sh, 13, (13<<4)+11);
}
#define init_14_13_13_12
void FASTCALL call_14_13_13_12(SH2_struct* sh)
{
  SH2movi(sh, 13, (13<<4)+12);
}
#define init_14_13_13_13
void FASTCALL call_14_13_13_13(SH2_struct* sh)
{
  SH2movi(sh, 13, (13<<4)+13);
}
#define init_14_13_13_14
void FASTCALL call_14_13_13_14(SH2_struct* sh)
{
  SH2movi(sh, 13, (13<<4)+14);
}
#define init_14_13_13_15
void FASTCALL call_14_13_13_15(SH2_struct* sh)
{
  SH2movi(sh, 13, (13<<4)+15);
}
#define init_14_13_14_0
void FASTCALL call_14_13_14_0(SH2_struct* sh)
{
  SH2movi(sh, 13, (14<<4)+0);
}
#define init_14_13_14_1
void FASTCALL call_14_13_14_1(SH2_struct* sh)
{
  SH2movi(sh, 13, (14<<4)+1);
}
#define init_14_13_14_2
void FASTCALL call_14_13_14_2(SH2_struct* sh)
{
  SH2movi(sh, 13, (14<<4)+2);
}
#define init_14_13_14_3
void FASTCALL call_14_13_14_3(SH2_struct* sh)
{
  SH2movi(sh, 13, (14<<4)+3);
}
#define init_14_13_14_4
void FASTCALL call_14_13_14_4(SH2_struct* sh)
{
  SH2movi(sh, 13, (14<<4)+4);
}
#define init_14_13_14_5
void FASTCALL call_14_13_14_5(SH2_struct* sh)
{
  SH2movi(sh, 13, (14<<4)+5);
}
#define init_14_13_14_6
void FASTCALL call_14_13_14_6(SH2_struct* sh)
{
  SH2movi(sh, 13, (14<<4)+6);
}
#define init_14_13_14_7
void FASTCALL call_14_13_14_7(SH2_struct* sh)
{
  SH2movi(sh, 13, (14<<4)+7);
}
#define init_14_13_14_8
void FASTCALL call_14_13_14_8(SH2_struct* sh)
{
  SH2movi(sh, 13, (14<<4)+8);
}
#define init_14_13_14_9
void FASTCALL call_14_13_14_9(SH2_struct* sh)
{
  SH2movi(sh, 13, (14<<4)+9);
}
#define init_14_13_14_10
void FASTCALL call_14_13_14_10(SH2_struct* sh)
{
  SH2movi(sh, 13, (14<<4)+10);
}
#define init_14_13_14_11
void FASTCALL call_14_13_14_11(SH2_struct* sh)
{
  SH2movi(sh, 13, (14<<4)+11);
}
#define init_14_13_14_12
void FASTCALL call_14_13_14_12(SH2_struct* sh)
{
  SH2movi(sh, 13, (14<<4)+12);
}
#define init_14_13_14_13
void FASTCALL call_14_13_14_13(SH2_struct* sh)
{
  SH2movi(sh, 13, (14<<4)+13);
}
#define init_14_13_14_14
void FASTCALL call_14_13_14_14(SH2_struct* sh)
{
  SH2movi(sh, 13, (14<<4)+14);
}
#define init_14_13_14_15
void FASTCALL call_14_13_14_15(SH2_struct* sh)
{
  SH2movi(sh, 13, (14<<4)+15);
}
#define init_14_13_15_0
void FASTCALL call_14_13_15_0(SH2_struct* sh)
{
  SH2movi(sh, 13, (15<<4)+0);
}
#define init_14_13_15_1
void FASTCALL call_14_13_15_1(SH2_struct* sh)
{
  SH2movi(sh, 13, (15<<4)+1);
}
#define init_14_13_15_2
void FASTCALL call_14_13_15_2(SH2_struct* sh)
{
  SH2movi(sh, 13, (15<<4)+2);
}
#define init_14_13_15_3
void FASTCALL call_14_13_15_3(SH2_struct* sh)
{
  SH2movi(sh, 13, (15<<4)+3);
}
#define init_14_13_15_4
void FASTCALL call_14_13_15_4(SH2_struct* sh)
{
  SH2movi(sh, 13, (15<<4)+4);
}
#define init_14_13_15_5
void FASTCALL call_14_13_15_5(SH2_struct* sh)
{
  SH2movi(sh, 13, (15<<4)+5);
}
#define init_14_13_15_6
void FASTCALL call_14_13_15_6(SH2_struct* sh)
{
  SH2movi(sh, 13, (15<<4)+6);
}
#define init_14_13_15_7
void FASTCALL call_14_13_15_7(SH2_struct* sh)
{
  SH2movi(sh, 13, (15<<4)+7);
}
#define init_14_13_15_8
void FASTCALL call_14_13_15_8(SH2_struct* sh)
{
  SH2movi(sh, 13, (15<<4)+8);
}
#define init_14_13_15_9
void FASTCALL call_14_13_15_9(SH2_struct* sh)
{
  SH2movi(sh, 13, (15<<4)+9);
}
#define init_14_13_15_10
void FASTCALL call_14_13_15_10(SH2_struct* sh)
{
  SH2movi(sh, 13, (15<<4)+10);
}
#define init_14_13_15_11
void FASTCALL call_14_13_15_11(SH2_struct* sh)
{
  SH2movi(sh, 13, (15<<4)+11);
}
#define init_14_13_15_12
void FASTCALL call_14_13_15_12(SH2_struct* sh)
{
  SH2movi(sh, 13, (15<<4)+12);
}
#define init_14_13_15_13
void FASTCALL call_14_13_15_13(SH2_struct* sh)
{
  SH2movi(sh, 13, (15<<4)+13);
}
#define init_14_13_15_14
void FASTCALL call_14_13_15_14(SH2_struct* sh)
{
  SH2movi(sh, 13, (15<<4)+14);
}
#define init_14_13_15_15
void FASTCALL call_14_13_15_15(SH2_struct* sh)
{
  SH2movi(sh, 13, (15<<4)+15);
}
#define init_14_14_0_0
void FASTCALL call_14_14_0_0(SH2_struct* sh)
{
  SH2movi(sh, 14, (0<<4)+0);
}
#define init_14_14_0_1
void FASTCALL call_14_14_0_1(SH2_struct* sh)
{
  SH2movi(sh, 14, (0<<4)+1);
}
#define init_14_14_0_2
void FASTCALL call_14_14_0_2(SH2_struct* sh)
{
  SH2movi(sh, 14, (0<<4)+2);
}
#define init_14_14_0_3
void FASTCALL call_14_14_0_3(SH2_struct* sh)
{
  SH2movi(sh, 14, (0<<4)+3);
}
#define init_14_14_0_4
void FASTCALL call_14_14_0_4(SH2_struct* sh)
{
  SH2movi(sh, 14, (0<<4)+4);
}
#define init_14_14_0_5
void FASTCALL call_14_14_0_5(SH2_struct* sh)
{
  SH2movi(sh, 14, (0<<4)+5);
}
#define init_14_14_0_6
void FASTCALL call_14_14_0_6(SH2_struct* sh)
{
  SH2movi(sh, 14, (0<<4)+6);
}
#define init_14_14_0_7
void FASTCALL call_14_14_0_7(SH2_struct* sh)
{
  SH2movi(sh, 14, (0<<4)+7);
}
#define init_14_14_0_8
void FASTCALL call_14_14_0_8(SH2_struct* sh)
{
  SH2movi(sh, 14, (0<<4)+8);
}
#define init_14_14_0_9
void FASTCALL call_14_14_0_9(SH2_struct* sh)
{
  SH2movi(sh, 14, (0<<4)+9);
}
#define init_14_14_0_10
void FASTCALL call_14_14_0_10(SH2_struct* sh)
{
  SH2movi(sh, 14, (0<<4)+10);
}
#define init_14_14_0_11
void FASTCALL call_14_14_0_11(SH2_struct* sh)
{
  SH2movi(sh, 14, (0<<4)+11);
}
#define init_14_14_0_12
void FASTCALL call_14_14_0_12(SH2_struct* sh)
{
  SH2movi(sh, 14, (0<<4)+12);
}
#define init_14_14_0_13
void FASTCALL call_14_14_0_13(SH2_struct* sh)
{
  SH2movi(sh, 14, (0<<4)+13);
}
#define init_14_14_0_14
void FASTCALL call_14_14_0_14(SH2_struct* sh)
{
  SH2movi(sh, 14, (0<<4)+14);
}
#define init_14_14_0_15
void FASTCALL call_14_14_0_15(SH2_struct* sh)
{
  SH2movi(sh, 14, (0<<4)+15);
}
#define init_14_14_1_0
void FASTCALL call_14_14_1_0(SH2_struct* sh)
{
  SH2movi(sh, 14, (1<<4)+0);
}
#define init_14_14_1_1
void FASTCALL call_14_14_1_1(SH2_struct* sh)
{
  SH2movi(sh, 14, (1<<4)+1);
}
#define init_14_14_1_2
void FASTCALL call_14_14_1_2(SH2_struct* sh)
{
  SH2movi(sh, 14, (1<<4)+2);
}
#define init_14_14_1_3
void FASTCALL call_14_14_1_3(SH2_struct* sh)
{
  SH2movi(sh, 14, (1<<4)+3);
}
#define init_14_14_1_4
void FASTCALL call_14_14_1_4(SH2_struct* sh)
{
  SH2movi(sh, 14, (1<<4)+4);
}
#define init_14_14_1_5
void FASTCALL call_14_14_1_5(SH2_struct* sh)
{
  SH2movi(sh, 14, (1<<4)+5);
}
#define init_14_14_1_6
void FASTCALL call_14_14_1_6(SH2_struct* sh)
{
  SH2movi(sh, 14, (1<<4)+6);
}
#define init_14_14_1_7
void FASTCALL call_14_14_1_7(SH2_struct* sh)
{
  SH2movi(sh, 14, (1<<4)+7);
}
#define init_14_14_1_8
void FASTCALL call_14_14_1_8(SH2_struct* sh)
{
  SH2movi(sh, 14, (1<<4)+8);
}
#define init_14_14_1_9
void FASTCALL call_14_14_1_9(SH2_struct* sh)
{
  SH2movi(sh, 14, (1<<4)+9);
}
#define init_14_14_1_10
void FASTCALL call_14_14_1_10(SH2_struct* sh)
{
  SH2movi(sh, 14, (1<<4)+10);
}
#define init_14_14_1_11
void FASTCALL call_14_14_1_11(SH2_struct* sh)
{
  SH2movi(sh, 14, (1<<4)+11);
}
#define init_14_14_1_12
void FASTCALL call_14_14_1_12(SH2_struct* sh)
{
  SH2movi(sh, 14, (1<<4)+12);
}
#define init_14_14_1_13
void FASTCALL call_14_14_1_13(SH2_struct* sh)
{
  SH2movi(sh, 14, (1<<4)+13);
}
#define init_14_14_1_14
void FASTCALL call_14_14_1_14(SH2_struct* sh)
{
  SH2movi(sh, 14, (1<<4)+14);
}
#define init_14_14_1_15
void FASTCALL call_14_14_1_15(SH2_struct* sh)
{
  SH2movi(sh, 14, (1<<4)+15);
}
#define init_14_14_2_0
void FASTCALL call_14_14_2_0(SH2_struct* sh)
{
  SH2movi(sh, 14, (2<<4)+0);
}
#define init_14_14_2_1
void FASTCALL call_14_14_2_1(SH2_struct* sh)
{
  SH2movi(sh, 14, (2<<4)+1);
}
#define init_14_14_2_2
void FASTCALL call_14_14_2_2(SH2_struct* sh)
{
  SH2movi(sh, 14, (2<<4)+2);
}
#define init_14_14_2_3
void FASTCALL call_14_14_2_3(SH2_struct* sh)
{
  SH2movi(sh, 14, (2<<4)+3);
}
#define init_14_14_2_4
void FASTCALL call_14_14_2_4(SH2_struct* sh)
{
  SH2movi(sh, 14, (2<<4)+4);
}
#define init_14_14_2_5
void FASTCALL call_14_14_2_5(SH2_struct* sh)
{
  SH2movi(sh, 14, (2<<4)+5);
}
#define init_14_14_2_6
void FASTCALL call_14_14_2_6(SH2_struct* sh)
{
  SH2movi(sh, 14, (2<<4)+6);
}
#define init_14_14_2_7
void FASTCALL call_14_14_2_7(SH2_struct* sh)
{
  SH2movi(sh, 14, (2<<4)+7);
}
#define init_14_14_2_8
void FASTCALL call_14_14_2_8(SH2_struct* sh)
{
  SH2movi(sh, 14, (2<<4)+8);
}
#define init_14_14_2_9
void FASTCALL call_14_14_2_9(SH2_struct* sh)
{
  SH2movi(sh, 14, (2<<4)+9);
}
#define init_14_14_2_10
void FASTCALL call_14_14_2_10(SH2_struct* sh)
{
  SH2movi(sh, 14, (2<<4)+10);
}
#define init_14_14_2_11
void FASTCALL call_14_14_2_11(SH2_struct* sh)
{
  SH2movi(sh, 14, (2<<4)+11);
}
#define init_14_14_2_12
void FASTCALL call_14_14_2_12(SH2_struct* sh)
{
  SH2movi(sh, 14, (2<<4)+12);
}
#define init_14_14_2_13
void FASTCALL call_14_14_2_13(SH2_struct* sh)
{
  SH2movi(sh, 14, (2<<4)+13);
}
#define init_14_14_2_14
void FASTCALL call_14_14_2_14(SH2_struct* sh)
{
  SH2movi(sh, 14, (2<<4)+14);
}
#define init_14_14_2_15
void FASTCALL call_14_14_2_15(SH2_struct* sh)
{
  SH2movi(sh, 14, (2<<4)+15);
}
#define init_14_14_3_0
void FASTCALL call_14_14_3_0(SH2_struct* sh)
{
  SH2movi(sh, 14, (3<<4)+0);
}
#define init_14_14_3_1
void FASTCALL call_14_14_3_1(SH2_struct* sh)
{
  SH2movi(sh, 14, (3<<4)+1);
}
#define init_14_14_3_2
void FASTCALL call_14_14_3_2(SH2_struct* sh)
{
  SH2movi(sh, 14, (3<<4)+2);
}
#define init_14_14_3_3
void FASTCALL call_14_14_3_3(SH2_struct* sh)
{
  SH2movi(sh, 14, (3<<4)+3);
}
#define init_14_14_3_4
void FASTCALL call_14_14_3_4(SH2_struct* sh)
{
  SH2movi(sh, 14, (3<<4)+4);
}
#define init_14_14_3_5
void FASTCALL call_14_14_3_5(SH2_struct* sh)
{
  SH2movi(sh, 14, (3<<4)+5);
}
#define init_14_14_3_6
void FASTCALL call_14_14_3_6(SH2_struct* sh)
{
  SH2movi(sh, 14, (3<<4)+6);
}
#define init_14_14_3_7
void FASTCALL call_14_14_3_7(SH2_struct* sh)
{
  SH2movi(sh, 14, (3<<4)+7);
}
#define init_14_14_3_8
void FASTCALL call_14_14_3_8(SH2_struct* sh)
{
  SH2movi(sh, 14, (3<<4)+8);
}
#define init_14_14_3_9
void FASTCALL call_14_14_3_9(SH2_struct* sh)
{
  SH2movi(sh, 14, (3<<4)+9);
}
#define init_14_14_3_10
void FASTCALL call_14_14_3_10(SH2_struct* sh)
{
  SH2movi(sh, 14, (3<<4)+10);
}
#define init_14_14_3_11
void FASTCALL call_14_14_3_11(SH2_struct* sh)
{
  SH2movi(sh, 14, (3<<4)+11);
}
#define init_14_14_3_12
void FASTCALL call_14_14_3_12(SH2_struct* sh)
{
  SH2movi(sh, 14, (3<<4)+12);
}
#define init_14_14_3_13
void FASTCALL call_14_14_3_13(SH2_struct* sh)
{
  SH2movi(sh, 14, (3<<4)+13);
}
#define init_14_14_3_14
void FASTCALL call_14_14_3_14(SH2_struct* sh)
{
  SH2movi(sh, 14, (3<<4)+14);
}
#define init_14_14_3_15
void FASTCALL call_14_14_3_15(SH2_struct* sh)
{
  SH2movi(sh, 14, (3<<4)+15);
}
#define init_14_14_4_0
void FASTCALL call_14_14_4_0(SH2_struct* sh)
{
  SH2movi(sh, 14, (4<<4)+0);
}
#define init_14_14_4_1
void FASTCALL call_14_14_4_1(SH2_struct* sh)
{
  SH2movi(sh, 14, (4<<4)+1);
}
#define init_14_14_4_2
void FASTCALL call_14_14_4_2(SH2_struct* sh)
{
  SH2movi(sh, 14, (4<<4)+2);
}
#define init_14_14_4_3
void FASTCALL call_14_14_4_3(SH2_struct* sh)
{
  SH2movi(sh, 14, (4<<4)+3);
}
#define init_14_14_4_4
void FASTCALL call_14_14_4_4(SH2_struct* sh)
{
  SH2movi(sh, 14, (4<<4)+4);
}
#define init_14_14_4_5
void FASTCALL call_14_14_4_5(SH2_struct* sh)
{
  SH2movi(sh, 14, (4<<4)+5);
}
#define init_14_14_4_6
void FASTCALL call_14_14_4_6(SH2_struct* sh)
{
  SH2movi(sh, 14, (4<<4)+6);
}
#define init_14_14_4_7
void FASTCALL call_14_14_4_7(SH2_struct* sh)
{
  SH2movi(sh, 14, (4<<4)+7);
}
#define init_14_14_4_8
void FASTCALL call_14_14_4_8(SH2_struct* sh)
{
  SH2movi(sh, 14, (4<<4)+8);
}
#define init_14_14_4_9
void FASTCALL call_14_14_4_9(SH2_struct* sh)
{
  SH2movi(sh, 14, (4<<4)+9);
}
#define init_14_14_4_10
void FASTCALL call_14_14_4_10(SH2_struct* sh)
{
  SH2movi(sh, 14, (4<<4)+10);
}
#define init_14_14_4_11
void FASTCALL call_14_14_4_11(SH2_struct* sh)
{
  SH2movi(sh, 14, (4<<4)+11);
}
#define init_14_14_4_12
void FASTCALL call_14_14_4_12(SH2_struct* sh)
{
  SH2movi(sh, 14, (4<<4)+12);
}
#define init_14_14_4_13
void FASTCALL call_14_14_4_13(SH2_struct* sh)
{
  SH2movi(sh, 14, (4<<4)+13);
}
#define init_14_14_4_14
void FASTCALL call_14_14_4_14(SH2_struct* sh)
{
  SH2movi(sh, 14, (4<<4)+14);
}
#define init_14_14_4_15
void FASTCALL call_14_14_4_15(SH2_struct* sh)
{
  SH2movi(sh, 14, (4<<4)+15);
}
#define init_14_14_5_0
void FASTCALL call_14_14_5_0(SH2_struct* sh)
{
  SH2movi(sh, 14, (5<<4)+0);
}
#define init_14_14_5_1
void FASTCALL call_14_14_5_1(SH2_struct* sh)
{
  SH2movi(sh, 14, (5<<4)+1);
}
#define init_14_14_5_2
void FASTCALL call_14_14_5_2(SH2_struct* sh)
{
  SH2movi(sh, 14, (5<<4)+2);
}
#define init_14_14_5_3
void FASTCALL call_14_14_5_3(SH2_struct* sh)
{
  SH2movi(sh, 14, (5<<4)+3);
}
#define init_14_14_5_4
void FASTCALL call_14_14_5_4(SH2_struct* sh)
{
  SH2movi(sh, 14, (5<<4)+4);
}
#define init_14_14_5_5
void FASTCALL call_14_14_5_5(SH2_struct* sh)
{
  SH2movi(sh, 14, (5<<4)+5);
}
#define init_14_14_5_6
void FASTCALL call_14_14_5_6(SH2_struct* sh)
{
  SH2movi(sh, 14, (5<<4)+6);
}
#define init_14_14_5_7
void FASTCALL call_14_14_5_7(SH2_struct* sh)
{
  SH2movi(sh, 14, (5<<4)+7);
}
#define init_14_14_5_8
void FASTCALL call_14_14_5_8(SH2_struct* sh)
{
  SH2movi(sh, 14, (5<<4)+8);
}
#define init_14_14_5_9
void FASTCALL call_14_14_5_9(SH2_struct* sh)
{
  SH2movi(sh, 14, (5<<4)+9);
}
#define init_14_14_5_10
void FASTCALL call_14_14_5_10(SH2_struct* sh)
{
  SH2movi(sh, 14, (5<<4)+10);
}
#define init_14_14_5_11
void FASTCALL call_14_14_5_11(SH2_struct* sh)
{
  SH2movi(sh, 14, (5<<4)+11);
}
#define init_14_14_5_12
void FASTCALL call_14_14_5_12(SH2_struct* sh)
{
  SH2movi(sh, 14, (5<<4)+12);
}
#define init_14_14_5_13
void FASTCALL call_14_14_5_13(SH2_struct* sh)
{
  SH2movi(sh, 14, (5<<4)+13);
}
#define init_14_14_5_14
void FASTCALL call_14_14_5_14(SH2_struct* sh)
{
  SH2movi(sh, 14, (5<<4)+14);
}
#define init_14_14_5_15
void FASTCALL call_14_14_5_15(SH2_struct* sh)
{
  SH2movi(sh, 14, (5<<4)+15);
}
#define init_14_14_6_0
void FASTCALL call_14_14_6_0(SH2_struct* sh)
{
  SH2movi(sh, 14, (6<<4)+0);
}
#define init_14_14_6_1
void FASTCALL call_14_14_6_1(SH2_struct* sh)
{
  SH2movi(sh, 14, (6<<4)+1);
}
#define init_14_14_6_2
void FASTCALL call_14_14_6_2(SH2_struct* sh)
{
  SH2movi(sh, 14, (6<<4)+2);
}
#define init_14_14_6_3
void FASTCALL call_14_14_6_3(SH2_struct* sh)
{
  SH2movi(sh, 14, (6<<4)+3);
}
#define init_14_14_6_4
void FASTCALL call_14_14_6_4(SH2_struct* sh)
{
  SH2movi(sh, 14, (6<<4)+4);
}
#define init_14_14_6_5
void FASTCALL call_14_14_6_5(SH2_struct* sh)
{
  SH2movi(sh, 14, (6<<4)+5);
}
#define init_14_14_6_6
void FASTCALL call_14_14_6_6(SH2_struct* sh)
{
  SH2movi(sh, 14, (6<<4)+6);
}
#define init_14_14_6_7
void FASTCALL call_14_14_6_7(SH2_struct* sh)
{
  SH2movi(sh, 14, (6<<4)+7);
}
#define init_14_14_6_8
void FASTCALL call_14_14_6_8(SH2_struct* sh)
{
  SH2movi(sh, 14, (6<<4)+8);
}
#define init_14_14_6_9
void FASTCALL call_14_14_6_9(SH2_struct* sh)
{
  SH2movi(sh, 14, (6<<4)+9);
}
#define init_14_14_6_10
void FASTCALL call_14_14_6_10(SH2_struct* sh)
{
  SH2movi(sh, 14, (6<<4)+10);
}
#define init_14_14_6_11
void FASTCALL call_14_14_6_11(SH2_struct* sh)
{
  SH2movi(sh, 14, (6<<4)+11);
}
#define init_14_14_6_12
void FASTCALL call_14_14_6_12(SH2_struct* sh)
{
  SH2movi(sh, 14, (6<<4)+12);
}
#define init_14_14_6_13
void FASTCALL call_14_14_6_13(SH2_struct* sh)
{
  SH2movi(sh, 14, (6<<4)+13);
}
#define init_14_14_6_14
void FASTCALL call_14_14_6_14(SH2_struct* sh)
{
  SH2movi(sh, 14, (6<<4)+14);
}
#define init_14_14_6_15
void FASTCALL call_14_14_6_15(SH2_struct* sh)
{
  SH2movi(sh, 14, (6<<4)+15);
}
#define init_14_14_7_0
void FASTCALL call_14_14_7_0(SH2_struct* sh)
{
  SH2movi(sh, 14, (7<<4)+0);
}
#define init_14_14_7_1
void FASTCALL call_14_14_7_1(SH2_struct* sh)
{
  SH2movi(sh, 14, (7<<4)+1);
}
#define init_14_14_7_2
void FASTCALL call_14_14_7_2(SH2_struct* sh)
{
  SH2movi(sh, 14, (7<<4)+2);
}
#define init_14_14_7_3
void FASTCALL call_14_14_7_3(SH2_struct* sh)
{
  SH2movi(sh, 14, (7<<4)+3);
}
#define init_14_14_7_4
void FASTCALL call_14_14_7_4(SH2_struct* sh)
{
  SH2movi(sh, 14, (7<<4)+4);
}
#define init_14_14_7_5
void FASTCALL call_14_14_7_5(SH2_struct* sh)
{
  SH2movi(sh, 14, (7<<4)+5);
}
#define init_14_14_7_6
void FASTCALL call_14_14_7_6(SH2_struct* sh)
{
  SH2movi(sh, 14, (7<<4)+6);
}
#define init_14_14_7_7
void FASTCALL call_14_14_7_7(SH2_struct* sh)
{
  SH2movi(sh, 14, (7<<4)+7);
}
#define init_14_14_7_8
void FASTCALL call_14_14_7_8(SH2_struct* sh)
{
  SH2movi(sh, 14, (7<<4)+8);
}
#define init_14_14_7_9
void FASTCALL call_14_14_7_9(SH2_struct* sh)
{
  SH2movi(sh, 14, (7<<4)+9);
}
#define init_14_14_7_10
void FASTCALL call_14_14_7_10(SH2_struct* sh)
{
  SH2movi(sh, 14, (7<<4)+10);
}
#define init_14_14_7_11
void FASTCALL call_14_14_7_11(SH2_struct* sh)
{
  SH2movi(sh, 14, (7<<4)+11);
}
#define init_14_14_7_12
void FASTCALL call_14_14_7_12(SH2_struct* sh)
{
  SH2movi(sh, 14, (7<<4)+12);
}
#define init_14_14_7_13
void FASTCALL call_14_14_7_13(SH2_struct* sh)
{
  SH2movi(sh, 14, (7<<4)+13);
}
#define init_14_14_7_14
void FASTCALL call_14_14_7_14(SH2_struct* sh)
{
  SH2movi(sh, 14, (7<<4)+14);
}
#define init_14_14_7_15
void FASTCALL call_14_14_7_15(SH2_struct* sh)
{
  SH2movi(sh, 14, (7<<4)+15);
}
#define init_14_14_8_0
void FASTCALL call_14_14_8_0(SH2_struct* sh)
{
  SH2movi(sh, 14, (8<<4)+0);
}
#define init_14_14_8_1
void FASTCALL call_14_14_8_1(SH2_struct* sh)
{
  SH2movi(sh, 14, (8<<4)+1);
}
#define init_14_14_8_2
void FASTCALL call_14_14_8_2(SH2_struct* sh)
{
  SH2movi(sh, 14, (8<<4)+2);
}
#define init_14_14_8_3
void FASTCALL call_14_14_8_3(SH2_struct* sh)
{
  SH2movi(sh, 14, (8<<4)+3);
}
#define init_14_14_8_4
void FASTCALL call_14_14_8_4(SH2_struct* sh)
{
  SH2movi(sh, 14, (8<<4)+4);
}
#define init_14_14_8_5
void FASTCALL call_14_14_8_5(SH2_struct* sh)
{
  SH2movi(sh, 14, (8<<4)+5);
}
#define init_14_14_8_6
void FASTCALL call_14_14_8_6(SH2_struct* sh)
{
  SH2movi(sh, 14, (8<<4)+6);
}
#define init_14_14_8_7
void FASTCALL call_14_14_8_7(SH2_struct* sh)
{
  SH2movi(sh, 14, (8<<4)+7);
}
#define init_14_14_8_8
void FASTCALL call_14_14_8_8(SH2_struct* sh)
{
  SH2movi(sh, 14, (8<<4)+8);
}
#define init_14_14_8_9
void FASTCALL call_14_14_8_9(SH2_struct* sh)
{
  SH2movi(sh, 14, (8<<4)+9);
}
#define init_14_14_8_10
void FASTCALL call_14_14_8_10(SH2_struct* sh)
{
  SH2movi(sh, 14, (8<<4)+10);
}
#define init_14_14_8_11
void FASTCALL call_14_14_8_11(SH2_struct* sh)
{
  SH2movi(sh, 14, (8<<4)+11);
}
#define init_14_14_8_12
void FASTCALL call_14_14_8_12(SH2_struct* sh)
{
  SH2movi(sh, 14, (8<<4)+12);
}
#define init_14_14_8_13
void FASTCALL call_14_14_8_13(SH2_struct* sh)
{
  SH2movi(sh, 14, (8<<4)+13);
}
#define init_14_14_8_14
void FASTCALL call_14_14_8_14(SH2_struct* sh)
{
  SH2movi(sh, 14, (8<<4)+14);
}
#define init_14_14_8_15
void FASTCALL call_14_14_8_15(SH2_struct* sh)
{
  SH2movi(sh, 14, (8<<4)+15);
}
#define init_14_14_9_0
void FASTCALL call_14_14_9_0(SH2_struct* sh)
{
  SH2movi(sh, 14, (9<<4)+0);
}
#define init_14_14_9_1
void FASTCALL call_14_14_9_1(SH2_struct* sh)
{
  SH2movi(sh, 14, (9<<4)+1);
}
#define init_14_14_9_2
void FASTCALL call_14_14_9_2(SH2_struct* sh)
{
  SH2movi(sh, 14, (9<<4)+2);
}
#define init_14_14_9_3
void FASTCALL call_14_14_9_3(SH2_struct* sh)
{
  SH2movi(sh, 14, (9<<4)+3);
}
#define init_14_14_9_4
void FASTCALL call_14_14_9_4(SH2_struct* sh)
{
  SH2movi(sh, 14, (9<<4)+4);
}
#define init_14_14_9_5
void FASTCALL call_14_14_9_5(SH2_struct* sh)
{
  SH2movi(sh, 14, (9<<4)+5);
}
#define init_14_14_9_6
void FASTCALL call_14_14_9_6(SH2_struct* sh)
{
  SH2movi(sh, 14, (9<<4)+6);
}
#define init_14_14_9_7
void FASTCALL call_14_14_9_7(SH2_struct* sh)
{
  SH2movi(sh, 14, (9<<4)+7);
}
#define init_14_14_9_8
void FASTCALL call_14_14_9_8(SH2_struct* sh)
{
  SH2movi(sh, 14, (9<<4)+8);
}
#define init_14_14_9_9
void FASTCALL call_14_14_9_9(SH2_struct* sh)
{
  SH2movi(sh, 14, (9<<4)+9);
}
#define init_14_14_9_10
void FASTCALL call_14_14_9_10(SH2_struct* sh)
{
  SH2movi(sh, 14, (9<<4)+10);
}
#define init_14_14_9_11
void FASTCALL call_14_14_9_11(SH2_struct* sh)
{
  SH2movi(sh, 14, (9<<4)+11);
}
#define init_14_14_9_12
void FASTCALL call_14_14_9_12(SH2_struct* sh)
{
  SH2movi(sh, 14, (9<<4)+12);
}
#define init_14_14_9_13
void FASTCALL call_14_14_9_13(SH2_struct* sh)
{
  SH2movi(sh, 14, (9<<4)+13);
}
#define init_14_14_9_14
void FASTCALL call_14_14_9_14(SH2_struct* sh)
{
  SH2movi(sh, 14, (9<<4)+14);
}
#define init_14_14_9_15
void FASTCALL call_14_14_9_15(SH2_struct* sh)
{
  SH2movi(sh, 14, (9<<4)+15);
}
#define init_14_14_10_0
void FASTCALL call_14_14_10_0(SH2_struct* sh)
{
  SH2movi(sh, 14, (10<<4)+0);
}
#define init_14_14_10_1
void FASTCALL call_14_14_10_1(SH2_struct* sh)
{
  SH2movi(sh, 14, (10<<4)+1);
}
#define init_14_14_10_2
void FASTCALL call_14_14_10_2(SH2_struct* sh)
{
  SH2movi(sh, 14, (10<<4)+2);
}
#define init_14_14_10_3
void FASTCALL call_14_14_10_3(SH2_struct* sh)
{
  SH2movi(sh, 14, (10<<4)+3);
}
#define init_14_14_10_4
void FASTCALL call_14_14_10_4(SH2_struct* sh)
{
  SH2movi(sh, 14, (10<<4)+4);
}
#define init_14_14_10_5
void FASTCALL call_14_14_10_5(SH2_struct* sh)
{
  SH2movi(sh, 14, (10<<4)+5);
}
#define init_14_14_10_6
void FASTCALL call_14_14_10_6(SH2_struct* sh)
{
  SH2movi(sh, 14, (10<<4)+6);
}
#define init_14_14_10_7
void FASTCALL call_14_14_10_7(SH2_struct* sh)
{
  SH2movi(sh, 14, (10<<4)+7);
}
#define init_14_14_10_8
void FASTCALL call_14_14_10_8(SH2_struct* sh)
{
  SH2movi(sh, 14, (10<<4)+8);
}
#define init_14_14_10_9
void FASTCALL call_14_14_10_9(SH2_struct* sh)
{
  SH2movi(sh, 14, (10<<4)+9);
}
#define init_14_14_10_10
void FASTCALL call_14_14_10_10(SH2_struct* sh)
{
  SH2movi(sh, 14, (10<<4)+10);
}
#define init_14_14_10_11
void FASTCALL call_14_14_10_11(SH2_struct* sh)
{
  SH2movi(sh, 14, (10<<4)+11);
}
#define init_14_14_10_12
void FASTCALL call_14_14_10_12(SH2_struct* sh)
{
  SH2movi(sh, 14, (10<<4)+12);
}
#define init_14_14_10_13
void FASTCALL call_14_14_10_13(SH2_struct* sh)
{
  SH2movi(sh, 14, (10<<4)+13);
}
#define init_14_14_10_14
void FASTCALL call_14_14_10_14(SH2_struct* sh)
{
  SH2movi(sh, 14, (10<<4)+14);
}
#define init_14_14_10_15
void FASTCALL call_14_14_10_15(SH2_struct* sh)
{
  SH2movi(sh, 14, (10<<4)+15);
}
#define init_14_14_11_0
void FASTCALL call_14_14_11_0(SH2_struct* sh)
{
  SH2movi(sh, 14, (11<<4)+0);
}
#define init_14_14_11_1
void FASTCALL call_14_14_11_1(SH2_struct* sh)
{
  SH2movi(sh, 14, (11<<4)+1);
}
#define init_14_14_11_2
void FASTCALL call_14_14_11_2(SH2_struct* sh)
{
  SH2movi(sh, 14, (11<<4)+2);
}
#define init_14_14_11_3
void FASTCALL call_14_14_11_3(SH2_struct* sh)
{
  SH2movi(sh, 14, (11<<4)+3);
}
#define init_14_14_11_4
void FASTCALL call_14_14_11_4(SH2_struct* sh)
{
  SH2movi(sh, 14, (11<<4)+4);
}
#define init_14_14_11_5
void FASTCALL call_14_14_11_5(SH2_struct* sh)
{
  SH2movi(sh, 14, (11<<4)+5);
}
#define init_14_14_11_6
void FASTCALL call_14_14_11_6(SH2_struct* sh)
{
  SH2movi(sh, 14, (11<<4)+6);
}
#define init_14_14_11_7
void FASTCALL call_14_14_11_7(SH2_struct* sh)
{
  SH2movi(sh, 14, (11<<4)+7);
}
#define init_14_14_11_8
void FASTCALL call_14_14_11_8(SH2_struct* sh)
{
  SH2movi(sh, 14, (11<<4)+8);
}
#define init_14_14_11_9
void FASTCALL call_14_14_11_9(SH2_struct* sh)
{
  SH2movi(sh, 14, (11<<4)+9);
}
#define init_14_14_11_10
void FASTCALL call_14_14_11_10(SH2_struct* sh)
{
  SH2movi(sh, 14, (11<<4)+10);
}
#define init_14_14_11_11
void FASTCALL call_14_14_11_11(SH2_struct* sh)
{
  SH2movi(sh, 14, (11<<4)+11);
}
#define init_14_14_11_12
void FASTCALL call_14_14_11_12(SH2_struct* sh)
{
  SH2movi(sh, 14, (11<<4)+12);
}
#define init_14_14_11_13
void FASTCALL call_14_14_11_13(SH2_struct* sh)
{
  SH2movi(sh, 14, (11<<4)+13);
}
#define init_14_14_11_14
void FASTCALL call_14_14_11_14(SH2_struct* sh)
{
  SH2movi(sh, 14, (11<<4)+14);
}
#define init_14_14_11_15
void FASTCALL call_14_14_11_15(SH2_struct* sh)
{
  SH2movi(sh, 14, (11<<4)+15);
}
#define init_14_14_12_0
void FASTCALL call_14_14_12_0(SH2_struct* sh)
{
  SH2movi(sh, 14, (12<<4)+0);
}
#define init_14_14_12_1
void FASTCALL call_14_14_12_1(SH2_struct* sh)
{
  SH2movi(sh, 14, (12<<4)+1);
}
#define init_14_14_12_2
void FASTCALL call_14_14_12_2(SH2_struct* sh)
{
  SH2movi(sh, 14, (12<<4)+2);
}
#define init_14_14_12_3
void FASTCALL call_14_14_12_3(SH2_struct* sh)
{
  SH2movi(sh, 14, (12<<4)+3);
}
#define init_14_14_12_4
void FASTCALL call_14_14_12_4(SH2_struct* sh)
{
  SH2movi(sh, 14, (12<<4)+4);
}
#define init_14_14_12_5
void FASTCALL call_14_14_12_5(SH2_struct* sh)
{
  SH2movi(sh, 14, (12<<4)+5);
}
#define init_14_14_12_6
void FASTCALL call_14_14_12_6(SH2_struct* sh)
{
  SH2movi(sh, 14, (12<<4)+6);
}
#define init_14_14_12_7
void FASTCALL call_14_14_12_7(SH2_struct* sh)
{
  SH2movi(sh, 14, (12<<4)+7);
}
#define init_14_14_12_8
void FASTCALL call_14_14_12_8(SH2_struct* sh)
{
  SH2movi(sh, 14, (12<<4)+8);
}
#define init_14_14_12_9
void FASTCALL call_14_14_12_9(SH2_struct* sh)
{
  SH2movi(sh, 14, (12<<4)+9);
}
#define init_14_14_12_10
void FASTCALL call_14_14_12_10(SH2_struct* sh)
{
  SH2movi(sh, 14, (12<<4)+10);
}
#define init_14_14_12_11
void FASTCALL call_14_14_12_11(SH2_struct* sh)
{
  SH2movi(sh, 14, (12<<4)+11);
}
#define init_14_14_12_12
void FASTCALL call_14_14_12_12(SH2_struct* sh)
{
  SH2movi(sh, 14, (12<<4)+12);
}
#define init_14_14_12_13
void FASTCALL call_14_14_12_13(SH2_struct* sh)
{
  SH2movi(sh, 14, (12<<4)+13);
}
#define init_14_14_12_14
void FASTCALL call_14_14_12_14(SH2_struct* sh)
{
  SH2movi(sh, 14, (12<<4)+14);
}
#define init_14_14_12_15
void FASTCALL call_14_14_12_15(SH2_struct* sh)
{
  SH2movi(sh, 14, (12<<4)+15);
}
#define init_14_14_13_0
void FASTCALL call_14_14_13_0(SH2_struct* sh)
{
  SH2movi(sh, 14, (13<<4)+0);
}
#define init_14_14_13_1
void FASTCALL call_14_14_13_1(SH2_struct* sh)
{
  SH2movi(sh, 14, (13<<4)+1);
}
#define init_14_14_13_2
void FASTCALL call_14_14_13_2(SH2_struct* sh)
{
  SH2movi(sh, 14, (13<<4)+2);
}
#define init_14_14_13_3
void FASTCALL call_14_14_13_3(SH2_struct* sh)
{
  SH2movi(sh, 14, (13<<4)+3);
}
#define init_14_14_13_4
void FASTCALL call_14_14_13_4(SH2_struct* sh)
{
  SH2movi(sh, 14, (13<<4)+4);
}
#define init_14_14_13_5
void FASTCALL call_14_14_13_5(SH2_struct* sh)
{
  SH2movi(sh, 14, (13<<4)+5);
}
#define init_14_14_13_6
void FASTCALL call_14_14_13_6(SH2_struct* sh)
{
  SH2movi(sh, 14, (13<<4)+6);
}
#define init_14_14_13_7
void FASTCALL call_14_14_13_7(SH2_struct* sh)
{
  SH2movi(sh, 14, (13<<4)+7);
}
#define init_14_14_13_8
void FASTCALL call_14_14_13_8(SH2_struct* sh)
{
  SH2movi(sh, 14, (13<<4)+8);
}
#define init_14_14_13_9
void FASTCALL call_14_14_13_9(SH2_struct* sh)
{
  SH2movi(sh, 14, (13<<4)+9);
}
#define init_14_14_13_10
void FASTCALL call_14_14_13_10(SH2_struct* sh)
{
  SH2movi(sh, 14, (13<<4)+10);
}
#define init_14_14_13_11
void FASTCALL call_14_14_13_11(SH2_struct* sh)
{
  SH2movi(sh, 14, (13<<4)+11);
}
#define init_14_14_13_12
void FASTCALL call_14_14_13_12(SH2_struct* sh)
{
  SH2movi(sh, 14, (13<<4)+12);
}
#define init_14_14_13_13
void FASTCALL call_14_14_13_13(SH2_struct* sh)
{
  SH2movi(sh, 14, (13<<4)+13);
}
#define init_14_14_13_14
void FASTCALL call_14_14_13_14(SH2_struct* sh)
{
  SH2movi(sh, 14, (13<<4)+14);
}
#define init_14_14_13_15
void FASTCALL call_14_14_13_15(SH2_struct* sh)
{
  SH2movi(sh, 14, (13<<4)+15);
}
#define init_14_14_14_0
void FASTCALL call_14_14_14_0(SH2_struct* sh)
{
  SH2movi(sh, 14, (14<<4)+0);
}
#define init_14_14_14_1
void FASTCALL call_14_14_14_1(SH2_struct* sh)
{
  SH2movi(sh, 14, (14<<4)+1);
}
#define init_14_14_14_2
void FASTCALL call_14_14_14_2(SH2_struct* sh)
{
  SH2movi(sh, 14, (14<<4)+2);
}
#define init_14_14_14_3
void FASTCALL call_14_14_14_3(SH2_struct* sh)
{
  SH2movi(sh, 14, (14<<4)+3);
}
#define init_14_14_14_4
void FASTCALL call_14_14_14_4(SH2_struct* sh)
{
  SH2movi(sh, 14, (14<<4)+4);
}
#define init_14_14_14_5
void FASTCALL call_14_14_14_5(SH2_struct* sh)
{
  SH2movi(sh, 14, (14<<4)+5);
}
#define init_14_14_14_6
void FASTCALL call_14_14_14_6(SH2_struct* sh)
{
  SH2movi(sh, 14, (14<<4)+6);
}
#define init_14_14_14_7
void FASTCALL call_14_14_14_7(SH2_struct* sh)
{
  SH2movi(sh, 14, (14<<4)+7);
}
#define init_14_14_14_8
void FASTCALL call_14_14_14_8(SH2_struct* sh)
{
  SH2movi(sh, 14, (14<<4)+8);
}
#define init_14_14_14_9
void FASTCALL call_14_14_14_9(SH2_struct* sh)
{
  SH2movi(sh, 14, (14<<4)+9);
}
#define init_14_14_14_10
void FASTCALL call_14_14_14_10(SH2_struct* sh)
{
  SH2movi(sh, 14, (14<<4)+10);
}
#define init_14_14_14_11
void FASTCALL call_14_14_14_11(SH2_struct* sh)
{
  SH2movi(sh, 14, (14<<4)+11);
}
#define init_14_14_14_12
void FASTCALL call_14_14_14_12(SH2_struct* sh)
{
  SH2movi(sh, 14, (14<<4)+12);
}
#define init_14_14_14_13
void FASTCALL call_14_14_14_13(SH2_struct* sh)
{
  SH2movi(sh, 14, (14<<4)+13);
}
#define init_14_14_14_14
void FASTCALL call_14_14_14_14(SH2_struct* sh)
{
  SH2movi(sh, 14, (14<<4)+14);
}
#define init_14_14_14_15
void FASTCALL call_14_14_14_15(SH2_struct* sh)
{
  SH2movi(sh, 14, (14<<4)+15);
}
#define init_14_14_15_0
void FASTCALL call_14_14_15_0(SH2_struct* sh)
{
  SH2movi(sh, 14, (15<<4)+0);
}
#define init_14_14_15_1
void FASTCALL call_14_14_15_1(SH2_struct* sh)
{
  SH2movi(sh, 14, (15<<4)+1);
}
#define init_14_14_15_2
void FASTCALL call_14_14_15_2(SH2_struct* sh)
{
  SH2movi(sh, 14, (15<<4)+2);
}
#define init_14_14_15_3
void FASTCALL call_14_14_15_3(SH2_struct* sh)
{
  SH2movi(sh, 14, (15<<4)+3);
}
#define init_14_14_15_4
void FASTCALL call_14_14_15_4(SH2_struct* sh)
{
  SH2movi(sh, 14, (15<<4)+4);
}
#define init_14_14_15_5
void FASTCALL call_14_14_15_5(SH2_struct* sh)
{
  SH2movi(sh, 14, (15<<4)+5);
}
#define init_14_14_15_6
void FASTCALL call_14_14_15_6(SH2_struct* sh)
{
  SH2movi(sh, 14, (15<<4)+6);
}
#define init_14_14_15_7
void FASTCALL call_14_14_15_7(SH2_struct* sh)
{
  SH2movi(sh, 14, (15<<4)+7);
}
#define init_14_14_15_8
void FASTCALL call_14_14_15_8(SH2_struct* sh)
{
  SH2movi(sh, 14, (15<<4)+8);
}
#define init_14_14_15_9
void FASTCALL call_14_14_15_9(SH2_struct* sh)
{
  SH2movi(sh, 14, (15<<4)+9);
}
#define init_14_14_15_10
void FASTCALL call_14_14_15_10(SH2_struct* sh)
{
  SH2movi(sh, 14, (15<<4)+10);
}
#define init_14_14_15_11
void FASTCALL call_14_14_15_11(SH2_struct* sh)
{
  SH2movi(sh, 14, (15<<4)+11);
}
#define init_14_14_15_12
void FASTCALL call_14_14_15_12(SH2_struct* sh)
{
  SH2movi(sh, 14, (15<<4)+12);
}
#define init_14_14_15_13
void FASTCALL call_14_14_15_13(SH2_struct* sh)
{
  SH2movi(sh, 14, (15<<4)+13);
}
#define init_14_14_15_14
void FASTCALL call_14_14_15_14(SH2_struct* sh)
{
  SH2movi(sh, 14, (15<<4)+14);
}
#define init_14_14_15_15
void FASTCALL call_14_14_15_15(SH2_struct* sh)
{
  SH2movi(sh, 14, (15<<4)+15);
}
#define init_14_15_0_0
void FASTCALL call_14_15_0_0(SH2_struct* sh)
{
  SH2movi(sh, 15, (0<<4)+0);
}
#define init_14_15_0_1
void FASTCALL call_14_15_0_1(SH2_struct* sh)
{
  SH2movi(sh, 15, (0<<4)+1);
}
#define init_14_15_0_2
void FASTCALL call_14_15_0_2(SH2_struct* sh)
{
  SH2movi(sh, 15, (0<<4)+2);
}
#define init_14_15_0_3
void FASTCALL call_14_15_0_3(SH2_struct* sh)
{
  SH2movi(sh, 15, (0<<4)+3);
}
#define init_14_15_0_4
void FASTCALL call_14_15_0_4(SH2_struct* sh)
{
  SH2movi(sh, 15, (0<<4)+4);
}
#define init_14_15_0_5
void FASTCALL call_14_15_0_5(SH2_struct* sh)
{
  SH2movi(sh, 15, (0<<4)+5);
}
#define init_14_15_0_6
void FASTCALL call_14_15_0_6(SH2_struct* sh)
{
  SH2movi(sh, 15, (0<<4)+6);
}
#define init_14_15_0_7
void FASTCALL call_14_15_0_7(SH2_struct* sh)
{
  SH2movi(sh, 15, (0<<4)+7);
}
#define init_14_15_0_8
void FASTCALL call_14_15_0_8(SH2_struct* sh)
{
  SH2movi(sh, 15, (0<<4)+8);
}
#define init_14_15_0_9
void FASTCALL call_14_15_0_9(SH2_struct* sh)
{
  SH2movi(sh, 15, (0<<4)+9);
}
#define init_14_15_0_10
void FASTCALL call_14_15_0_10(SH2_struct* sh)
{
  SH2movi(sh, 15, (0<<4)+10);
}
#define init_14_15_0_11
void FASTCALL call_14_15_0_11(SH2_struct* sh)
{
  SH2movi(sh, 15, (0<<4)+11);
}
#define init_14_15_0_12
void FASTCALL call_14_15_0_12(SH2_struct* sh)
{
  SH2movi(sh, 15, (0<<4)+12);
}
#define init_14_15_0_13
void FASTCALL call_14_15_0_13(SH2_struct* sh)
{
  SH2movi(sh, 15, (0<<4)+13);
}
#define init_14_15_0_14
void FASTCALL call_14_15_0_14(SH2_struct* sh)
{
  SH2movi(sh, 15, (0<<4)+14);
}
#define init_14_15_0_15
void FASTCALL call_14_15_0_15(SH2_struct* sh)
{
  SH2movi(sh, 15, (0<<4)+15);
}
#define init_14_15_1_0
void FASTCALL call_14_15_1_0(SH2_struct* sh)
{
  SH2movi(sh, 15, (1<<4)+0);
}
#define init_14_15_1_1
void FASTCALL call_14_15_1_1(SH2_struct* sh)
{
  SH2movi(sh, 15, (1<<4)+1);
}
#define init_14_15_1_2
void FASTCALL call_14_15_1_2(SH2_struct* sh)
{
  SH2movi(sh, 15, (1<<4)+2);
}
#define init_14_15_1_3
void FASTCALL call_14_15_1_3(SH2_struct* sh)
{
  SH2movi(sh, 15, (1<<4)+3);
}
#define init_14_15_1_4
void FASTCALL call_14_15_1_4(SH2_struct* sh)
{
  SH2movi(sh, 15, (1<<4)+4);
}
#define init_14_15_1_5
void FASTCALL call_14_15_1_5(SH2_struct* sh)
{
  SH2movi(sh, 15, (1<<4)+5);
}
#define init_14_15_1_6
void FASTCALL call_14_15_1_6(SH2_struct* sh)
{
  SH2movi(sh, 15, (1<<4)+6);
}
#define init_14_15_1_7
void FASTCALL call_14_15_1_7(SH2_struct* sh)
{
  SH2movi(sh, 15, (1<<4)+7);
}
#define init_14_15_1_8
void FASTCALL call_14_15_1_8(SH2_struct* sh)
{
  SH2movi(sh, 15, (1<<4)+8);
}
#define init_14_15_1_9
void FASTCALL call_14_15_1_9(SH2_struct* sh)
{
  SH2movi(sh, 15, (1<<4)+9);
}
#define init_14_15_1_10
void FASTCALL call_14_15_1_10(SH2_struct* sh)
{
  SH2movi(sh, 15, (1<<4)+10);
}
#define init_14_15_1_11
void FASTCALL call_14_15_1_11(SH2_struct* sh)
{
  SH2movi(sh, 15, (1<<4)+11);
}
#define init_14_15_1_12
void FASTCALL call_14_15_1_12(SH2_struct* sh)
{
  SH2movi(sh, 15, (1<<4)+12);
}
#define init_14_15_1_13
void FASTCALL call_14_15_1_13(SH2_struct* sh)
{
  SH2movi(sh, 15, (1<<4)+13);
}
#define init_14_15_1_14
void FASTCALL call_14_15_1_14(SH2_struct* sh)
{
  SH2movi(sh, 15, (1<<4)+14);
}
#define init_14_15_1_15
void FASTCALL call_14_15_1_15(SH2_struct* sh)
{
  SH2movi(sh, 15, (1<<4)+15);
}
#define init_14_15_2_0
void FASTCALL call_14_15_2_0(SH2_struct* sh)
{
  SH2movi(sh, 15, (2<<4)+0);
}
#define init_14_15_2_1
void FASTCALL call_14_15_2_1(SH2_struct* sh)
{
  SH2movi(sh, 15, (2<<4)+1);
}
#define init_14_15_2_2
void FASTCALL call_14_15_2_2(SH2_struct* sh)
{
  SH2movi(sh, 15, (2<<4)+2);
}
#define init_14_15_2_3
void FASTCALL call_14_15_2_3(SH2_struct* sh)
{
  SH2movi(sh, 15, (2<<4)+3);
}
#define init_14_15_2_4
void FASTCALL call_14_15_2_4(SH2_struct* sh)
{
  SH2movi(sh, 15, (2<<4)+4);
}
#define init_14_15_2_5
void FASTCALL call_14_15_2_5(SH2_struct* sh)
{
  SH2movi(sh, 15, (2<<4)+5);
}
#define init_14_15_2_6
void FASTCALL call_14_15_2_6(SH2_struct* sh)
{
  SH2movi(sh, 15, (2<<4)+6);
}
#define init_14_15_2_7
void FASTCALL call_14_15_2_7(SH2_struct* sh)
{
  SH2movi(sh, 15, (2<<4)+7);
}
#define init_14_15_2_8
void FASTCALL call_14_15_2_8(SH2_struct* sh)
{
  SH2movi(sh, 15, (2<<4)+8);
}
#define init_14_15_2_9
void FASTCALL call_14_15_2_9(SH2_struct* sh)
{
  SH2movi(sh, 15, (2<<4)+9);
}
#define init_14_15_2_10
void FASTCALL call_14_15_2_10(SH2_struct* sh)
{
  SH2movi(sh, 15, (2<<4)+10);
}
#define init_14_15_2_11
void FASTCALL call_14_15_2_11(SH2_struct* sh)
{
  SH2movi(sh, 15, (2<<4)+11);
}
#define init_14_15_2_12
void FASTCALL call_14_15_2_12(SH2_struct* sh)
{
  SH2movi(sh, 15, (2<<4)+12);
}
#define init_14_15_2_13
void FASTCALL call_14_15_2_13(SH2_struct* sh)
{
  SH2movi(sh, 15, (2<<4)+13);
}
#define init_14_15_2_14
void FASTCALL call_14_15_2_14(SH2_struct* sh)
{
  SH2movi(sh, 15, (2<<4)+14);
}
#define init_14_15_2_15
void FASTCALL call_14_15_2_15(SH2_struct* sh)
{
  SH2movi(sh, 15, (2<<4)+15);
}
#define init_14_15_3_0
void FASTCALL call_14_15_3_0(SH2_struct* sh)
{
  SH2movi(sh, 15, (3<<4)+0);
}
#define init_14_15_3_1
void FASTCALL call_14_15_3_1(SH2_struct* sh)
{
  SH2movi(sh, 15, (3<<4)+1);
}
#define init_14_15_3_2
void FASTCALL call_14_15_3_2(SH2_struct* sh)
{
  SH2movi(sh, 15, (3<<4)+2);
}
#define init_14_15_3_3
void FASTCALL call_14_15_3_3(SH2_struct* sh)
{
  SH2movi(sh, 15, (3<<4)+3);
}
#define init_14_15_3_4
void FASTCALL call_14_15_3_4(SH2_struct* sh)
{
  SH2movi(sh, 15, (3<<4)+4);
}
#define init_14_15_3_5
void FASTCALL call_14_15_3_5(SH2_struct* sh)
{
  SH2movi(sh, 15, (3<<4)+5);
}
#define init_14_15_3_6
void FASTCALL call_14_15_3_6(SH2_struct* sh)
{
  SH2movi(sh, 15, (3<<4)+6);
}
#define init_14_15_3_7
void FASTCALL call_14_15_3_7(SH2_struct* sh)
{
  SH2movi(sh, 15, (3<<4)+7);
}
#define init_14_15_3_8
void FASTCALL call_14_15_3_8(SH2_struct* sh)
{
  SH2movi(sh, 15, (3<<4)+8);
}
#define init_14_15_3_9
void FASTCALL call_14_15_3_9(SH2_struct* sh)
{
  SH2movi(sh, 15, (3<<4)+9);
}
#define init_14_15_3_10
void FASTCALL call_14_15_3_10(SH2_struct* sh)
{
  SH2movi(sh, 15, (3<<4)+10);
}
#define init_14_15_3_11
void FASTCALL call_14_15_3_11(SH2_struct* sh)
{
  SH2movi(sh, 15, (3<<4)+11);
}
#define init_14_15_3_12
void FASTCALL call_14_15_3_12(SH2_struct* sh)
{
  SH2movi(sh, 15, (3<<4)+12);
}
#define init_14_15_3_13
void FASTCALL call_14_15_3_13(SH2_struct* sh)
{
  SH2movi(sh, 15, (3<<4)+13);
}
#define init_14_15_3_14
void FASTCALL call_14_15_3_14(SH2_struct* sh)
{
  SH2movi(sh, 15, (3<<4)+14);
}
#define init_14_15_3_15
void FASTCALL call_14_15_3_15(SH2_struct* sh)
{
  SH2movi(sh, 15, (3<<4)+15);
}
#define init_14_15_4_0
void FASTCALL call_14_15_4_0(SH2_struct* sh)
{
  SH2movi(sh, 15, (4<<4)+0);
}
#define init_14_15_4_1
void FASTCALL call_14_15_4_1(SH2_struct* sh)
{
  SH2movi(sh, 15, (4<<4)+1);
}
#define init_14_15_4_2
void FASTCALL call_14_15_4_2(SH2_struct* sh)
{
  SH2movi(sh, 15, (4<<4)+2);
}
#define init_14_15_4_3
void FASTCALL call_14_15_4_3(SH2_struct* sh)
{
  SH2movi(sh, 15, (4<<4)+3);
}
#define init_14_15_4_4
void FASTCALL call_14_15_4_4(SH2_struct* sh)
{
  SH2movi(sh, 15, (4<<4)+4);
}
#define init_14_15_4_5
void FASTCALL call_14_15_4_5(SH2_struct* sh)
{
  SH2movi(sh, 15, (4<<4)+5);
}
#define init_14_15_4_6
void FASTCALL call_14_15_4_6(SH2_struct* sh)
{
  SH2movi(sh, 15, (4<<4)+6);
}
#define init_14_15_4_7
void FASTCALL call_14_15_4_7(SH2_struct* sh)
{
  SH2movi(sh, 15, (4<<4)+7);
}
#define init_14_15_4_8
void FASTCALL call_14_15_4_8(SH2_struct* sh)
{
  SH2movi(sh, 15, (4<<4)+8);
}
#define init_14_15_4_9
void FASTCALL call_14_15_4_9(SH2_struct* sh)
{
  SH2movi(sh, 15, (4<<4)+9);
}
#define init_14_15_4_10
void FASTCALL call_14_15_4_10(SH2_struct* sh)
{
  SH2movi(sh, 15, (4<<4)+10);
}
#define init_14_15_4_11
void FASTCALL call_14_15_4_11(SH2_struct* sh)
{
  SH2movi(sh, 15, (4<<4)+11);
}
#define init_14_15_4_12
void FASTCALL call_14_15_4_12(SH2_struct* sh)
{
  SH2movi(sh, 15, (4<<4)+12);
}
#define init_14_15_4_13
void FASTCALL call_14_15_4_13(SH2_struct* sh)
{
  SH2movi(sh, 15, (4<<4)+13);
}
#define init_14_15_4_14
void FASTCALL call_14_15_4_14(SH2_struct* sh)
{
  SH2movi(sh, 15, (4<<4)+14);
}
#define init_14_15_4_15
void FASTCALL call_14_15_4_15(SH2_struct* sh)
{
  SH2movi(sh, 15, (4<<4)+15);
}
#define init_14_15_5_0
void FASTCALL call_14_15_5_0(SH2_struct* sh)
{
  SH2movi(sh, 15, (5<<4)+0);
}
#define init_14_15_5_1
void FASTCALL call_14_15_5_1(SH2_struct* sh)
{
  SH2movi(sh, 15, (5<<4)+1);
}
#define init_14_15_5_2
void FASTCALL call_14_15_5_2(SH2_struct* sh)
{
  SH2movi(sh, 15, (5<<4)+2);
}
#define init_14_15_5_3
void FASTCALL call_14_15_5_3(SH2_struct* sh)
{
  SH2movi(sh, 15, (5<<4)+3);
}
#define init_14_15_5_4
void FASTCALL call_14_15_5_4(SH2_struct* sh)
{
  SH2movi(sh, 15, (5<<4)+4);
}
#define init_14_15_5_5
void FASTCALL call_14_15_5_5(SH2_struct* sh)
{
  SH2movi(sh, 15, (5<<4)+5);
}
#define init_14_15_5_6
void FASTCALL call_14_15_5_6(SH2_struct* sh)
{
  SH2movi(sh, 15, (5<<4)+6);
}
#define init_14_15_5_7
void FASTCALL call_14_15_5_7(SH2_struct* sh)
{
  SH2movi(sh, 15, (5<<4)+7);
}
#define init_14_15_5_8
void FASTCALL call_14_15_5_8(SH2_struct* sh)
{
  SH2movi(sh, 15, (5<<4)+8);
}
#define init_14_15_5_9
void FASTCALL call_14_15_5_9(SH2_struct* sh)
{
  SH2movi(sh, 15, (5<<4)+9);
}
#define init_14_15_5_10
void FASTCALL call_14_15_5_10(SH2_struct* sh)
{
  SH2movi(sh, 15, (5<<4)+10);
}
#define init_14_15_5_11
void FASTCALL call_14_15_5_11(SH2_struct* sh)
{
  SH2movi(sh, 15, (5<<4)+11);
}
#define init_14_15_5_12
void FASTCALL call_14_15_5_12(SH2_struct* sh)
{
  SH2movi(sh, 15, (5<<4)+12);
}
#define init_14_15_5_13
void FASTCALL call_14_15_5_13(SH2_struct* sh)
{
  SH2movi(sh, 15, (5<<4)+13);
}
#define init_14_15_5_14
void FASTCALL call_14_15_5_14(SH2_struct* sh)
{
  SH2movi(sh, 15, (5<<4)+14);
}
#define init_14_15_5_15
void FASTCALL call_14_15_5_15(SH2_struct* sh)
{
  SH2movi(sh, 15, (5<<4)+15);
}
#define init_14_15_6_0
void FASTCALL call_14_15_6_0(SH2_struct* sh)
{
  SH2movi(sh, 15, (6<<4)+0);
}
#define init_14_15_6_1
void FASTCALL call_14_15_6_1(SH2_struct* sh)
{
  SH2movi(sh, 15, (6<<4)+1);
}
#define init_14_15_6_2
void FASTCALL call_14_15_6_2(SH2_struct* sh)
{
  SH2movi(sh, 15, (6<<4)+2);
}
#define init_14_15_6_3
void FASTCALL call_14_15_6_3(SH2_struct* sh)
{
  SH2movi(sh, 15, (6<<4)+3);
}
#define init_14_15_6_4
void FASTCALL call_14_15_6_4(SH2_struct* sh)
{
  SH2movi(sh, 15, (6<<4)+4);
}
#define init_14_15_6_5
void FASTCALL call_14_15_6_5(SH2_struct* sh)
{
  SH2movi(sh, 15, (6<<4)+5);
}
#define init_14_15_6_6
void FASTCALL call_14_15_6_6(SH2_struct* sh)
{
  SH2movi(sh, 15, (6<<4)+6);
}
#define init_14_15_6_7
void FASTCALL call_14_15_6_7(SH2_struct* sh)
{
  SH2movi(sh, 15, (6<<4)+7);
}
#define init_14_15_6_8
void FASTCALL call_14_15_6_8(SH2_struct* sh)
{
  SH2movi(sh, 15, (6<<4)+8);
}
#define init_14_15_6_9
void FASTCALL call_14_15_6_9(SH2_struct* sh)
{
  SH2movi(sh, 15, (6<<4)+9);
}
#define init_14_15_6_10
void FASTCALL call_14_15_6_10(SH2_struct* sh)
{
  SH2movi(sh, 15, (6<<4)+10);
}
#define init_14_15_6_11
void FASTCALL call_14_15_6_11(SH2_struct* sh)
{
  SH2movi(sh, 15, (6<<4)+11);
}
#define init_14_15_6_12
void FASTCALL call_14_15_6_12(SH2_struct* sh)
{
  SH2movi(sh, 15, (6<<4)+12);
}
#define init_14_15_6_13
void FASTCALL call_14_15_6_13(SH2_struct* sh)
{
  SH2movi(sh, 15, (6<<4)+13);
}
#define init_14_15_6_14
void FASTCALL call_14_15_6_14(SH2_struct* sh)
{
  SH2movi(sh, 15, (6<<4)+14);
}
#define init_14_15_6_15
void FASTCALL call_14_15_6_15(SH2_struct* sh)
{
  SH2movi(sh, 15, (6<<4)+15);
}
#define init_14_15_7_0
void FASTCALL call_14_15_7_0(SH2_struct* sh)
{
  SH2movi(sh, 15, (7<<4)+0);
}
#define init_14_15_7_1
void FASTCALL call_14_15_7_1(SH2_struct* sh)
{
  SH2movi(sh, 15, (7<<4)+1);
}
#define init_14_15_7_2
void FASTCALL call_14_15_7_2(SH2_struct* sh)
{
  SH2movi(sh, 15, (7<<4)+2);
}
#define init_14_15_7_3
void FASTCALL call_14_15_7_3(SH2_struct* sh)
{
  SH2movi(sh, 15, (7<<4)+3);
}
#define init_14_15_7_4
void FASTCALL call_14_15_7_4(SH2_struct* sh)
{
  SH2movi(sh, 15, (7<<4)+4);
}
#define init_14_15_7_5
void FASTCALL call_14_15_7_5(SH2_struct* sh)
{
  SH2movi(sh, 15, (7<<4)+5);
}
#define init_14_15_7_6
void FASTCALL call_14_15_7_6(SH2_struct* sh)
{
  SH2movi(sh, 15, (7<<4)+6);
}
#define init_14_15_7_7
void FASTCALL call_14_15_7_7(SH2_struct* sh)
{
  SH2movi(sh, 15, (7<<4)+7);
}
#define init_14_15_7_8
void FASTCALL call_14_15_7_8(SH2_struct* sh)
{
  SH2movi(sh, 15, (7<<4)+8);
}
#define init_14_15_7_9
void FASTCALL call_14_15_7_9(SH2_struct* sh)
{
  SH2movi(sh, 15, (7<<4)+9);
}
#define init_14_15_7_10
void FASTCALL call_14_15_7_10(SH2_struct* sh)
{
  SH2movi(sh, 15, (7<<4)+10);
}
#define init_14_15_7_11
void FASTCALL call_14_15_7_11(SH2_struct* sh)
{
  SH2movi(sh, 15, (7<<4)+11);
}
#define init_14_15_7_12
void FASTCALL call_14_15_7_12(SH2_struct* sh)
{
  SH2movi(sh, 15, (7<<4)+12);
}
#define init_14_15_7_13
void FASTCALL call_14_15_7_13(SH2_struct* sh)
{
  SH2movi(sh, 15, (7<<4)+13);
}
#define init_14_15_7_14
void FASTCALL call_14_15_7_14(SH2_struct* sh)
{
  SH2movi(sh, 15, (7<<4)+14);
}
#define init_14_15_7_15
void FASTCALL call_14_15_7_15(SH2_struct* sh)
{
  SH2movi(sh, 15, (7<<4)+15);
}
#define init_14_15_8_0
void FASTCALL call_14_15_8_0(SH2_struct* sh)
{
  SH2movi(sh, 15, (8<<4)+0);
}
#define init_14_15_8_1
void FASTCALL call_14_15_8_1(SH2_struct* sh)
{
  SH2movi(sh, 15, (8<<4)+1);
}
#define init_14_15_8_2
void FASTCALL call_14_15_8_2(SH2_struct* sh)
{
  SH2movi(sh, 15, (8<<4)+2);
}
#define init_14_15_8_3
void FASTCALL call_14_15_8_3(SH2_struct* sh)
{
  SH2movi(sh, 15, (8<<4)+3);
}
#define init_14_15_8_4
void FASTCALL call_14_15_8_4(SH2_struct* sh)
{
  SH2movi(sh, 15, (8<<4)+4);
}
#define init_14_15_8_5
void FASTCALL call_14_15_8_5(SH2_struct* sh)
{
  SH2movi(sh, 15, (8<<4)+5);
}
#define init_14_15_8_6
void FASTCALL call_14_15_8_6(SH2_struct* sh)
{
  SH2movi(sh, 15, (8<<4)+6);
}
#define init_14_15_8_7
void FASTCALL call_14_15_8_7(SH2_struct* sh)
{
  SH2movi(sh, 15, (8<<4)+7);
}
#define init_14_15_8_8
void FASTCALL call_14_15_8_8(SH2_struct* sh)
{
  SH2movi(sh, 15, (8<<4)+8);
}
#define init_14_15_8_9
void FASTCALL call_14_15_8_9(SH2_struct* sh)
{
  SH2movi(sh, 15, (8<<4)+9);
}
#define init_14_15_8_10
void FASTCALL call_14_15_8_10(SH2_struct* sh)
{
  SH2movi(sh, 15, (8<<4)+10);
}
#define init_14_15_8_11
void FASTCALL call_14_15_8_11(SH2_struct* sh)
{
  SH2movi(sh, 15, (8<<4)+11);
}
#define init_14_15_8_12
void FASTCALL call_14_15_8_12(SH2_struct* sh)
{
  SH2movi(sh, 15, (8<<4)+12);
}
#define init_14_15_8_13
void FASTCALL call_14_15_8_13(SH2_struct* sh)
{
  SH2movi(sh, 15, (8<<4)+13);
}
#define init_14_15_8_14
void FASTCALL call_14_15_8_14(SH2_struct* sh)
{
  SH2movi(sh, 15, (8<<4)+14);
}
#define init_14_15_8_15
void FASTCALL call_14_15_8_15(SH2_struct* sh)
{
  SH2movi(sh, 15, (8<<4)+15);
}
#define init_14_15_9_0
void FASTCALL call_14_15_9_0(SH2_struct* sh)
{
  SH2movi(sh, 15, (9<<4)+0);
}
#define init_14_15_9_1
void FASTCALL call_14_15_9_1(SH2_struct* sh)
{
  SH2movi(sh, 15, (9<<4)+1);
}
#define init_14_15_9_2
void FASTCALL call_14_15_9_2(SH2_struct* sh)
{
  SH2movi(sh, 15, (9<<4)+2);
}
#define init_14_15_9_3
void FASTCALL call_14_15_9_3(SH2_struct* sh)
{
  SH2movi(sh, 15, (9<<4)+3);
}
#define init_14_15_9_4
void FASTCALL call_14_15_9_4(SH2_struct* sh)
{
  SH2movi(sh, 15, (9<<4)+4);
}
#define init_14_15_9_5
void FASTCALL call_14_15_9_5(SH2_struct* sh)
{
  SH2movi(sh, 15, (9<<4)+5);
}
#define init_14_15_9_6
void FASTCALL call_14_15_9_6(SH2_struct* sh)
{
  SH2movi(sh, 15, (9<<4)+6);
}
#define init_14_15_9_7
void FASTCALL call_14_15_9_7(SH2_struct* sh)
{
  SH2movi(sh, 15, (9<<4)+7);
}
#define init_14_15_9_8
void FASTCALL call_14_15_9_8(SH2_struct* sh)
{
  SH2movi(sh, 15, (9<<4)+8);
}
#define init_14_15_9_9
void FASTCALL call_14_15_9_9(SH2_struct* sh)
{
  SH2movi(sh, 15, (9<<4)+9);
}
#define init_14_15_9_10
void FASTCALL call_14_15_9_10(SH2_struct* sh)
{
  SH2movi(sh, 15, (9<<4)+10);
}
#define init_14_15_9_11
void FASTCALL call_14_15_9_11(SH2_struct* sh)
{
  SH2movi(sh, 15, (9<<4)+11);
}
#define init_14_15_9_12
void FASTCALL call_14_15_9_12(SH2_struct* sh)
{
  SH2movi(sh, 15, (9<<4)+12);
}
#define init_14_15_9_13
void FASTCALL call_14_15_9_13(SH2_struct* sh)
{
  SH2movi(sh, 15, (9<<4)+13);
}
#define init_14_15_9_14
void FASTCALL call_14_15_9_14(SH2_struct* sh)
{
  SH2movi(sh, 15, (9<<4)+14);
}
#define init_14_15_9_15
void FASTCALL call_14_15_9_15(SH2_struct* sh)
{
  SH2movi(sh, 15, (9<<4)+15);
}
#define init_14_15_10_0
void FASTCALL call_14_15_10_0(SH2_struct* sh)
{
  SH2movi(sh, 15, (10<<4)+0);
}
#define init_14_15_10_1
void FASTCALL call_14_15_10_1(SH2_struct* sh)
{
  SH2movi(sh, 15, (10<<4)+1);
}
#define init_14_15_10_2
void FASTCALL call_14_15_10_2(SH2_struct* sh)
{
  SH2movi(sh, 15, (10<<4)+2);
}
#define init_14_15_10_3
void FASTCALL call_14_15_10_3(SH2_struct* sh)
{
  SH2movi(sh, 15, (10<<4)+3);
}
#define init_14_15_10_4
void FASTCALL call_14_15_10_4(SH2_struct* sh)
{
  SH2movi(sh, 15, (10<<4)+4);
}
#define init_14_15_10_5
void FASTCALL call_14_15_10_5(SH2_struct* sh)
{
  SH2movi(sh, 15, (10<<4)+5);
}
#define init_14_15_10_6
void FASTCALL call_14_15_10_6(SH2_struct* sh)
{
  SH2movi(sh, 15, (10<<4)+6);
}
#define init_14_15_10_7
void FASTCALL call_14_15_10_7(SH2_struct* sh)
{
  SH2movi(sh, 15, (10<<4)+7);
}
#define init_14_15_10_8
void FASTCALL call_14_15_10_8(SH2_struct* sh)
{
  SH2movi(sh, 15, (10<<4)+8);
}
#define init_14_15_10_9
void FASTCALL call_14_15_10_9(SH2_struct* sh)
{
  SH2movi(sh, 15, (10<<4)+9);
}
#define init_14_15_10_10
void FASTCALL call_14_15_10_10(SH2_struct* sh)
{
  SH2movi(sh, 15, (10<<4)+10);
}
#define init_14_15_10_11
void FASTCALL call_14_15_10_11(SH2_struct* sh)
{
  SH2movi(sh, 15, (10<<4)+11);
}
#define init_14_15_10_12
void FASTCALL call_14_15_10_12(SH2_struct* sh)
{
  SH2movi(sh, 15, (10<<4)+12);
}
#define init_14_15_10_13
void FASTCALL call_14_15_10_13(SH2_struct* sh)
{
  SH2movi(sh, 15, (10<<4)+13);
}
#define init_14_15_10_14
void FASTCALL call_14_15_10_14(SH2_struct* sh)
{
  SH2movi(sh, 15, (10<<4)+14);
}
#define init_14_15_10_15
void FASTCALL call_14_15_10_15(SH2_struct* sh)
{
  SH2movi(sh, 15, (10<<4)+15);
}
#define init_14_15_11_0
void FASTCALL call_14_15_11_0(SH2_struct* sh)
{
  SH2movi(sh, 15, (11<<4)+0);
}
#define init_14_15_11_1
void FASTCALL call_14_15_11_1(SH2_struct* sh)
{
  SH2movi(sh, 15, (11<<4)+1);
}
#define init_14_15_11_2
void FASTCALL call_14_15_11_2(SH2_struct* sh)
{
  SH2movi(sh, 15, (11<<4)+2);
}
#define init_14_15_11_3
void FASTCALL call_14_15_11_3(SH2_struct* sh)
{
  SH2movi(sh, 15, (11<<4)+3);
}
#define init_14_15_11_4
void FASTCALL call_14_15_11_4(SH2_struct* sh)
{
  SH2movi(sh, 15, (11<<4)+4);
}
#define init_14_15_11_5
void FASTCALL call_14_15_11_5(SH2_struct* sh)
{
  SH2movi(sh, 15, (11<<4)+5);
}
#define init_14_15_11_6
void FASTCALL call_14_15_11_6(SH2_struct* sh)
{
  SH2movi(sh, 15, (11<<4)+6);
}
#define init_14_15_11_7
void FASTCALL call_14_15_11_7(SH2_struct* sh)
{
  SH2movi(sh, 15, (11<<4)+7);
}
#define init_14_15_11_8
void FASTCALL call_14_15_11_8(SH2_struct* sh)
{
  SH2movi(sh, 15, (11<<4)+8);
}
#define init_14_15_11_9
void FASTCALL call_14_15_11_9(SH2_struct* sh)
{
  SH2movi(sh, 15, (11<<4)+9);
}
#define init_14_15_11_10
void FASTCALL call_14_15_11_10(SH2_struct* sh)
{
  SH2movi(sh, 15, (11<<4)+10);
}
#define init_14_15_11_11
void FASTCALL call_14_15_11_11(SH2_struct* sh)
{
  SH2movi(sh, 15, (11<<4)+11);
}
#define init_14_15_11_12
void FASTCALL call_14_15_11_12(SH2_struct* sh)
{
  SH2movi(sh, 15, (11<<4)+12);
}
#define init_14_15_11_13
void FASTCALL call_14_15_11_13(SH2_struct* sh)
{
  SH2movi(sh, 15, (11<<4)+13);
}
#define init_14_15_11_14
void FASTCALL call_14_15_11_14(SH2_struct* sh)
{
  SH2movi(sh, 15, (11<<4)+14);
}
#define init_14_15_11_15
void FASTCALL call_14_15_11_15(SH2_struct* sh)
{
  SH2movi(sh, 15, (11<<4)+15);
}
#define init_14_15_12_0
void FASTCALL call_14_15_12_0(SH2_struct* sh)
{
  SH2movi(sh, 15, (12<<4)+0);
}
#define init_14_15_12_1
void FASTCALL call_14_15_12_1(SH2_struct* sh)
{
  SH2movi(sh, 15, (12<<4)+1);
}
#define init_14_15_12_2
void FASTCALL call_14_15_12_2(SH2_struct* sh)
{
  SH2movi(sh, 15, (12<<4)+2);
}
#define init_14_15_12_3
void FASTCALL call_14_15_12_3(SH2_struct* sh)
{
  SH2movi(sh, 15, (12<<4)+3);
}
#define init_14_15_12_4
void FASTCALL call_14_15_12_4(SH2_struct* sh)
{
  SH2movi(sh, 15, (12<<4)+4);
}
#define init_14_15_12_5
void FASTCALL call_14_15_12_5(SH2_struct* sh)
{
  SH2movi(sh, 15, (12<<4)+5);
}
#define init_14_15_12_6
void FASTCALL call_14_15_12_6(SH2_struct* sh)
{
  SH2movi(sh, 15, (12<<4)+6);
}
#define init_14_15_12_7
void FASTCALL call_14_15_12_7(SH2_struct* sh)
{
  SH2movi(sh, 15, (12<<4)+7);
}
#define init_14_15_12_8
void FASTCALL call_14_15_12_8(SH2_struct* sh)
{
  SH2movi(sh, 15, (12<<4)+8);
}
#define init_14_15_12_9
void FASTCALL call_14_15_12_9(SH2_struct* sh)
{
  SH2movi(sh, 15, (12<<4)+9);
}
#define init_14_15_12_10
void FASTCALL call_14_15_12_10(SH2_struct* sh)
{
  SH2movi(sh, 15, (12<<4)+10);
}
#define init_14_15_12_11
void FASTCALL call_14_15_12_11(SH2_struct* sh)
{
  SH2movi(sh, 15, (12<<4)+11);
}
#define init_14_15_12_12
void FASTCALL call_14_15_12_12(SH2_struct* sh)
{
  SH2movi(sh, 15, (12<<4)+12);
}
#define init_14_15_12_13
void FASTCALL call_14_15_12_13(SH2_struct* sh)
{
  SH2movi(sh, 15, (12<<4)+13);
}
#define init_14_15_12_14
void FASTCALL call_14_15_12_14(SH2_struct* sh)
{
  SH2movi(sh, 15, (12<<4)+14);
}
#define init_14_15_12_15
void FASTCALL call_14_15_12_15(SH2_struct* sh)
{
  SH2movi(sh, 15, (12<<4)+15);
}
#define init_14_15_13_0
void FASTCALL call_14_15_13_0(SH2_struct* sh)
{
  SH2movi(sh, 15, (13<<4)+0);
}
#define init_14_15_13_1
void FASTCALL call_14_15_13_1(SH2_struct* sh)
{
  SH2movi(sh, 15, (13<<4)+1);
}
#define init_14_15_13_2
void FASTCALL call_14_15_13_2(SH2_struct* sh)
{
  SH2movi(sh, 15, (13<<4)+2);
}
#define init_14_15_13_3
void FASTCALL call_14_15_13_3(SH2_struct* sh)
{
  SH2movi(sh, 15, (13<<4)+3);
}
#define init_14_15_13_4
void FASTCALL call_14_15_13_4(SH2_struct* sh)
{
  SH2movi(sh, 15, (13<<4)+4);
}
#define init_14_15_13_5
void FASTCALL call_14_15_13_5(SH2_struct* sh)
{
  SH2movi(sh, 15, (13<<4)+5);
}
#define init_14_15_13_6
void FASTCALL call_14_15_13_6(SH2_struct* sh)
{
  SH2movi(sh, 15, (13<<4)+6);
}
#define init_14_15_13_7
void FASTCALL call_14_15_13_7(SH2_struct* sh)
{
  SH2movi(sh, 15, (13<<4)+7);
}
#define init_14_15_13_8
void FASTCALL call_14_15_13_8(SH2_struct* sh)
{
  SH2movi(sh, 15, (13<<4)+8);
}
#define init_14_15_13_9
void FASTCALL call_14_15_13_9(SH2_struct* sh)
{
  SH2movi(sh, 15, (13<<4)+9);
}
#define init_14_15_13_10
void FASTCALL call_14_15_13_10(SH2_struct* sh)
{
  SH2movi(sh, 15, (13<<4)+10);
}
#define init_14_15_13_11
void FASTCALL call_14_15_13_11(SH2_struct* sh)
{
  SH2movi(sh, 15, (13<<4)+11);
}
#define init_14_15_13_12
void FASTCALL call_14_15_13_12(SH2_struct* sh)
{
  SH2movi(sh, 15, (13<<4)+12);
}
#define init_14_15_13_13
void FASTCALL call_14_15_13_13(SH2_struct* sh)
{
  SH2movi(sh, 15, (13<<4)+13);
}
#define init_14_15_13_14
void FASTCALL call_14_15_13_14(SH2_struct* sh)
{
  SH2movi(sh, 15, (13<<4)+14);
}
#define init_14_15_13_15
void FASTCALL call_14_15_13_15(SH2_struct* sh)
{
  SH2movi(sh, 15, (13<<4)+15);
}
#define init_14_15_14_0
void FASTCALL call_14_15_14_0(SH2_struct* sh)
{
  SH2movi(sh, 15, (14<<4)+0);
}
#define init_14_15_14_1
void FASTCALL call_14_15_14_1(SH2_struct* sh)
{
  SH2movi(sh, 15, (14<<4)+1);
}
#define init_14_15_14_2
void FASTCALL call_14_15_14_2(SH2_struct* sh)
{
  SH2movi(sh, 15, (14<<4)+2);
}
#define init_14_15_14_3
void FASTCALL call_14_15_14_3(SH2_struct* sh)
{
  SH2movi(sh, 15, (14<<4)+3);
}
#define init_14_15_14_4
void FASTCALL call_14_15_14_4(SH2_struct* sh)
{
  SH2movi(sh, 15, (14<<4)+4);
}
#define init_14_15_14_5
void FASTCALL call_14_15_14_5(SH2_struct* sh)
{
  SH2movi(sh, 15, (14<<4)+5);
}
#define init_14_15_14_6
void FASTCALL call_14_15_14_6(SH2_struct* sh)
{
  SH2movi(sh, 15, (14<<4)+6);
}
#define init_14_15_14_7
void FASTCALL call_14_15_14_7(SH2_struct* sh)
{
  SH2movi(sh, 15, (14<<4)+7);
}
#define init_14_15_14_8
void FASTCALL call_14_15_14_8(SH2_struct* sh)
{
  SH2movi(sh, 15, (14<<4)+8);
}
#define init_14_15_14_9
void FASTCALL call_14_15_14_9(SH2_struct* sh)
{
  SH2movi(sh, 15, (14<<4)+9);
}
#define init_14_15_14_10
void FASTCALL call_14_15_14_10(SH2_struct* sh)
{
  SH2movi(sh, 15, (14<<4)+10);
}
#define init_14_15_14_11
void FASTCALL call_14_15_14_11(SH2_struct* sh)
{
  SH2movi(sh, 15, (14<<4)+11);
}
#define init_14_15_14_12
void FASTCALL call_14_15_14_12(SH2_struct* sh)
{
  SH2movi(sh, 15, (14<<4)+12);
}
#define init_14_15_14_13
void FASTCALL call_14_15_14_13(SH2_struct* sh)
{
  SH2movi(sh, 15, (14<<4)+13);
}
#define init_14_15_14_14
void FASTCALL call_14_15_14_14(SH2_struct* sh)
{
  SH2movi(sh, 15, (14<<4)+14);
}
#define init_14_15_14_15
void FASTCALL call_14_15_14_15(SH2_struct* sh)
{
  SH2movi(sh, 15, (14<<4)+15);
}
#define init_14_15_15_0
void FASTCALL call_14_15_15_0(SH2_struct* sh)
{
  SH2movi(sh, 15, (15<<4)+0);
}
#define init_14_15_15_1
void FASTCALL call_14_15_15_1(SH2_struct* sh)
{
  SH2movi(sh, 15, (15<<4)+1);
}
#define init_14_15_15_2
void FASTCALL call_14_15_15_2(SH2_struct* sh)
{
  SH2movi(sh, 15, (15<<4)+2);
}
#define init_14_15_15_3
void FASTCALL call_14_15_15_3(SH2_struct* sh)
{
  SH2movi(sh, 15, (15<<4)+3);
}
#define init_14_15_15_4
void FASTCALL call_14_15_15_4(SH2_struct* sh)
{
  SH2movi(sh, 15, (15<<4)+4);
}
#define init_14_15_15_5
void FASTCALL call_14_15_15_5(SH2_struct* sh)
{
  SH2movi(sh, 15, (15<<4)+5);
}
#define init_14_15_15_6
void FASTCALL call_14_15_15_6(SH2_struct* sh)
{
  SH2movi(sh, 15, (15<<4)+6);
}
#define init_14_15_15_7
void FASTCALL call_14_15_15_7(SH2_struct* sh)
{
  SH2movi(sh, 15, (15<<4)+7);
}
#define init_14_15_15_8
void FASTCALL call_14_15_15_8(SH2_struct* sh)
{
  SH2movi(sh, 15, (15<<4)+8);
}
#define init_14_15_15_9
void FASTCALL call_14_15_15_9(SH2_struct* sh)
{
  SH2movi(sh, 15, (15<<4)+9);
}
#define init_14_15_15_10
void FASTCALL call_14_15_15_10(SH2_struct* sh)
{
  SH2movi(sh, 15, (15<<4)+10);
}
#define init_14_15_15_11
void FASTCALL call_14_15_15_11(SH2_struct* sh)
{
  SH2movi(sh, 15, (15<<4)+11);
}
#define init_14_15_15_12
void FASTCALL call_14_15_15_12(SH2_struct* sh)
{
  SH2movi(sh, 15, (15<<4)+12);
}
#define init_14_15_15_13
void FASTCALL call_14_15_15_13(SH2_struct* sh)
{
  SH2movi(sh, 15, (15<<4)+13);
}
#define init_14_15_15_14
void FASTCALL call_14_15_15_14(SH2_struct* sh)
{
  SH2movi(sh, 15, (15<<4)+14);
}
#define init_14_15_15_15
void FASTCALL call_14_15_15_15(SH2_struct* sh)
{
  SH2movi(sh, 15, (15<<4)+15);
}
#define init_12_0_0_0
void FASTCALL call_12_0_0_0(SH2_struct* sh)
{
  SH2movbsg(sh,(0<<4)+0);
}
#define init_12_0_0_1
void FASTCALL call_12_0_0_1(SH2_struct* sh)
{
  SH2movbsg(sh,(0<<4)+1);
}
#define init_12_0_0_2
void FASTCALL call_12_0_0_2(SH2_struct* sh)
{
  SH2movbsg(sh,(0<<4)+2);
}
#define init_12_0_0_3
void FASTCALL call_12_0_0_3(SH2_struct* sh)
{
  SH2movbsg(sh,(0<<4)+3);
}
#define init_12_0_0_4
void FASTCALL call_12_0_0_4(SH2_struct* sh)
{
  SH2movbsg(sh,(0<<4)+4);
}
#define init_12_0_0_5
void FASTCALL call_12_0_0_5(SH2_struct* sh)
{
  SH2movbsg(sh,(0<<4)+5);
}
#define init_12_0_0_6
void FASTCALL call_12_0_0_6(SH2_struct* sh)
{
  SH2movbsg(sh,(0<<4)+6);
}
#define init_12_0_0_7
void FASTCALL call_12_0_0_7(SH2_struct* sh)
{
  SH2movbsg(sh,(0<<4)+7);
}
#define init_12_0_0_8
void FASTCALL call_12_0_0_8(SH2_struct* sh)
{
  SH2movbsg(sh,(0<<4)+8);
}
#define init_12_0_0_9
void FASTCALL call_12_0_0_9(SH2_struct* sh)
{
  SH2movbsg(sh,(0<<4)+9);
}
#define init_12_0_0_10
void FASTCALL call_12_0_0_10(SH2_struct* sh)
{
  SH2movbsg(sh,(0<<4)+10);
}
#define init_12_0_0_11
void FASTCALL call_12_0_0_11(SH2_struct* sh)
{
  SH2movbsg(sh,(0<<4)+11);
}
#define init_12_0_0_12
void FASTCALL call_12_0_0_12(SH2_struct* sh)
{
  SH2movbsg(sh,(0<<4)+12);
}
#define init_12_0_0_13
void FASTCALL call_12_0_0_13(SH2_struct* sh)
{
  SH2movbsg(sh,(0<<4)+13);
}
#define init_12_0_0_14
void FASTCALL call_12_0_0_14(SH2_struct* sh)
{
  SH2movbsg(sh,(0<<4)+14);
}
#define init_12_0_0_15
void FASTCALL call_12_0_0_15(SH2_struct* sh)
{
  SH2movbsg(sh,(0<<4)+15);
}
#define init_12_0_1_0
void FASTCALL call_12_0_1_0(SH2_struct* sh)
{
  SH2movbsg(sh,(1<<4)+0);
}
#define init_12_0_1_1
void FASTCALL call_12_0_1_1(SH2_struct* sh)
{
  SH2movbsg(sh,(1<<4)+1);
}
#define init_12_0_1_2
void FASTCALL call_12_0_1_2(SH2_struct* sh)
{
  SH2movbsg(sh,(1<<4)+2);
}
#define init_12_0_1_3
void FASTCALL call_12_0_1_3(SH2_struct* sh)
{
  SH2movbsg(sh,(1<<4)+3);
}
#define init_12_0_1_4
void FASTCALL call_12_0_1_4(SH2_struct* sh)
{
  SH2movbsg(sh,(1<<4)+4);
}
#define init_12_0_1_5
void FASTCALL call_12_0_1_5(SH2_struct* sh)
{
  SH2movbsg(sh,(1<<4)+5);
}
#define init_12_0_1_6
void FASTCALL call_12_0_1_6(SH2_struct* sh)
{
  SH2movbsg(sh,(1<<4)+6);
}
#define init_12_0_1_7
void FASTCALL call_12_0_1_7(SH2_struct* sh)
{
  SH2movbsg(sh,(1<<4)+7);
}
#define init_12_0_1_8
void FASTCALL call_12_0_1_8(SH2_struct* sh)
{
  SH2movbsg(sh,(1<<4)+8);
}
#define init_12_0_1_9
void FASTCALL call_12_0_1_9(SH2_struct* sh)
{
  SH2movbsg(sh,(1<<4)+9);
}
#define init_12_0_1_10
void FASTCALL call_12_0_1_10(SH2_struct* sh)
{
  SH2movbsg(sh,(1<<4)+10);
}
#define init_12_0_1_11
void FASTCALL call_12_0_1_11(SH2_struct* sh)
{
  SH2movbsg(sh,(1<<4)+11);
}
#define init_12_0_1_12
void FASTCALL call_12_0_1_12(SH2_struct* sh)
{
  SH2movbsg(sh,(1<<4)+12);
}
#define init_12_0_1_13
void FASTCALL call_12_0_1_13(SH2_struct* sh)
{
  SH2movbsg(sh,(1<<4)+13);
}
#define init_12_0_1_14
void FASTCALL call_12_0_1_14(SH2_struct* sh)
{
  SH2movbsg(sh,(1<<4)+14);
}
#define init_12_0_1_15
void FASTCALL call_12_0_1_15(SH2_struct* sh)
{
  SH2movbsg(sh,(1<<4)+15);
}
#define init_12_0_2_0
void FASTCALL call_12_0_2_0(SH2_struct* sh)
{
  SH2movbsg(sh,(2<<4)+0);
}
#define init_12_0_2_1
void FASTCALL call_12_0_2_1(SH2_struct* sh)
{
  SH2movbsg(sh,(2<<4)+1);
}
#define init_12_0_2_2
void FASTCALL call_12_0_2_2(SH2_struct* sh)
{
  SH2movbsg(sh,(2<<4)+2);
}
#define init_12_0_2_3
void FASTCALL call_12_0_2_3(SH2_struct* sh)
{
  SH2movbsg(sh,(2<<4)+3);
}
#define init_12_0_2_4
void FASTCALL call_12_0_2_4(SH2_struct* sh)
{
  SH2movbsg(sh,(2<<4)+4);
}
#define init_12_0_2_5
void FASTCALL call_12_0_2_5(SH2_struct* sh)
{
  SH2movbsg(sh,(2<<4)+5);
}
#define init_12_0_2_6
void FASTCALL call_12_0_2_6(SH2_struct* sh)
{
  SH2movbsg(sh,(2<<4)+6);
}
#define init_12_0_2_7
void FASTCALL call_12_0_2_7(SH2_struct* sh)
{
  SH2movbsg(sh,(2<<4)+7);
}
#define init_12_0_2_8
void FASTCALL call_12_0_2_8(SH2_struct* sh)
{
  SH2movbsg(sh,(2<<4)+8);
}
#define init_12_0_2_9
void FASTCALL call_12_0_2_9(SH2_struct* sh)
{
  SH2movbsg(sh,(2<<4)+9);
}
#define init_12_0_2_10
void FASTCALL call_12_0_2_10(SH2_struct* sh)
{
  SH2movbsg(sh,(2<<4)+10);
}
#define init_12_0_2_11
void FASTCALL call_12_0_2_11(SH2_struct* sh)
{
  SH2movbsg(sh,(2<<4)+11);
}
#define init_12_0_2_12
void FASTCALL call_12_0_2_12(SH2_struct* sh)
{
  SH2movbsg(sh,(2<<4)+12);
}
#define init_12_0_2_13
void FASTCALL call_12_0_2_13(SH2_struct* sh)
{
  SH2movbsg(sh,(2<<4)+13);
}
#define init_12_0_2_14
void FASTCALL call_12_0_2_14(SH2_struct* sh)
{
  SH2movbsg(sh,(2<<4)+14);
}
#define init_12_0_2_15
void FASTCALL call_12_0_2_15(SH2_struct* sh)
{
  SH2movbsg(sh,(2<<4)+15);
}
#define init_12_0_3_0
void FASTCALL call_12_0_3_0(SH2_struct* sh)
{
  SH2movbsg(sh,(3<<4)+0);
}
#define init_12_0_3_1
void FASTCALL call_12_0_3_1(SH2_struct* sh)
{
  SH2movbsg(sh,(3<<4)+1);
}
#define init_12_0_3_2
void FASTCALL call_12_0_3_2(SH2_struct* sh)
{
  SH2movbsg(sh,(3<<4)+2);
}
#define init_12_0_3_3
void FASTCALL call_12_0_3_3(SH2_struct* sh)
{
  SH2movbsg(sh,(3<<4)+3);
}
#define init_12_0_3_4
void FASTCALL call_12_0_3_4(SH2_struct* sh)
{
  SH2movbsg(sh,(3<<4)+4);
}
#define init_12_0_3_5
void FASTCALL call_12_0_3_5(SH2_struct* sh)
{
  SH2movbsg(sh,(3<<4)+5);
}
#define init_12_0_3_6
void FASTCALL call_12_0_3_6(SH2_struct* sh)
{
  SH2movbsg(sh,(3<<4)+6);
}
#define init_12_0_3_7
void FASTCALL call_12_0_3_7(SH2_struct* sh)
{
  SH2movbsg(sh,(3<<4)+7);
}
#define init_12_0_3_8
void FASTCALL call_12_0_3_8(SH2_struct* sh)
{
  SH2movbsg(sh,(3<<4)+8);
}
#define init_12_0_3_9
void FASTCALL call_12_0_3_9(SH2_struct* sh)
{
  SH2movbsg(sh,(3<<4)+9);
}
#define init_12_0_3_10
void FASTCALL call_12_0_3_10(SH2_struct* sh)
{
  SH2movbsg(sh,(3<<4)+10);
}
#define init_12_0_3_11
void FASTCALL call_12_0_3_11(SH2_struct* sh)
{
  SH2movbsg(sh,(3<<4)+11);
}
#define init_12_0_3_12
void FASTCALL call_12_0_3_12(SH2_struct* sh)
{
  SH2movbsg(sh,(3<<4)+12);
}
#define init_12_0_3_13
void FASTCALL call_12_0_3_13(SH2_struct* sh)
{
  SH2movbsg(sh,(3<<4)+13);
}
#define init_12_0_3_14
void FASTCALL call_12_0_3_14(SH2_struct* sh)
{
  SH2movbsg(sh,(3<<4)+14);
}
#define init_12_0_3_15
void FASTCALL call_12_0_3_15(SH2_struct* sh)
{
  SH2movbsg(sh,(3<<4)+15);
}
#define init_12_0_4_0
void FASTCALL call_12_0_4_0(SH2_struct* sh)
{
  SH2movbsg(sh,(4<<4)+0);
}
#define init_12_0_4_1
void FASTCALL call_12_0_4_1(SH2_struct* sh)
{
  SH2movbsg(sh,(4<<4)+1);
}
#define init_12_0_4_2
void FASTCALL call_12_0_4_2(SH2_struct* sh)
{
  SH2movbsg(sh,(4<<4)+2);
}
#define init_12_0_4_3
void FASTCALL call_12_0_4_3(SH2_struct* sh)
{
  SH2movbsg(sh,(4<<4)+3);
}
#define init_12_0_4_4
void FASTCALL call_12_0_4_4(SH2_struct* sh)
{
  SH2movbsg(sh,(4<<4)+4);
}
#define init_12_0_4_5
void FASTCALL call_12_0_4_5(SH2_struct* sh)
{
  SH2movbsg(sh,(4<<4)+5);
}
#define init_12_0_4_6
void FASTCALL call_12_0_4_6(SH2_struct* sh)
{
  SH2movbsg(sh,(4<<4)+6);
}
#define init_12_0_4_7
void FASTCALL call_12_0_4_7(SH2_struct* sh)
{
  SH2movbsg(sh,(4<<4)+7);
}
#define init_12_0_4_8
void FASTCALL call_12_0_4_8(SH2_struct* sh)
{
  SH2movbsg(sh,(4<<4)+8);
}
#define init_12_0_4_9
void FASTCALL call_12_0_4_9(SH2_struct* sh)
{
  SH2movbsg(sh,(4<<4)+9);
}
#define init_12_0_4_10
void FASTCALL call_12_0_4_10(SH2_struct* sh)
{
  SH2movbsg(sh,(4<<4)+10);
}
#define init_12_0_4_11
void FASTCALL call_12_0_4_11(SH2_struct* sh)
{
  SH2movbsg(sh,(4<<4)+11);
}
#define init_12_0_4_12
void FASTCALL call_12_0_4_12(SH2_struct* sh)
{
  SH2movbsg(sh,(4<<4)+12);
}
#define init_12_0_4_13
void FASTCALL call_12_0_4_13(SH2_struct* sh)
{
  SH2movbsg(sh,(4<<4)+13);
}
#define init_12_0_4_14
void FASTCALL call_12_0_4_14(SH2_struct* sh)
{
  SH2movbsg(sh,(4<<4)+14);
}
#define init_12_0_4_15
void FASTCALL call_12_0_4_15(SH2_struct* sh)
{
  SH2movbsg(sh,(4<<4)+15);
}
#define init_12_0_5_0
void FASTCALL call_12_0_5_0(SH2_struct* sh)
{
  SH2movbsg(sh,(5<<4)+0);
}
#define init_12_0_5_1
void FASTCALL call_12_0_5_1(SH2_struct* sh)
{
  SH2movbsg(sh,(5<<4)+1);
}
#define init_12_0_5_2
void FASTCALL call_12_0_5_2(SH2_struct* sh)
{
  SH2movbsg(sh,(5<<4)+2);
}
#define init_12_0_5_3
void FASTCALL call_12_0_5_3(SH2_struct* sh)
{
  SH2movbsg(sh,(5<<4)+3);
}
#define init_12_0_5_4
void FASTCALL call_12_0_5_4(SH2_struct* sh)
{
  SH2movbsg(sh,(5<<4)+4);
}
#define init_12_0_5_5
void FASTCALL call_12_0_5_5(SH2_struct* sh)
{
  SH2movbsg(sh,(5<<4)+5);
}
#define init_12_0_5_6
void FASTCALL call_12_0_5_6(SH2_struct* sh)
{
  SH2movbsg(sh,(5<<4)+6);
}
#define init_12_0_5_7
void FASTCALL call_12_0_5_7(SH2_struct* sh)
{
  SH2movbsg(sh,(5<<4)+7);
}
#define init_12_0_5_8
void FASTCALL call_12_0_5_8(SH2_struct* sh)
{
  SH2movbsg(sh,(5<<4)+8);
}
#define init_12_0_5_9
void FASTCALL call_12_0_5_9(SH2_struct* sh)
{
  SH2movbsg(sh,(5<<4)+9);
}
#define init_12_0_5_10
void FASTCALL call_12_0_5_10(SH2_struct* sh)
{
  SH2movbsg(sh,(5<<4)+10);
}
#define init_12_0_5_11
void FASTCALL call_12_0_5_11(SH2_struct* sh)
{
  SH2movbsg(sh,(5<<4)+11);
}
#define init_12_0_5_12
void FASTCALL call_12_0_5_12(SH2_struct* sh)
{
  SH2movbsg(sh,(5<<4)+12);
}
#define init_12_0_5_13
void FASTCALL call_12_0_5_13(SH2_struct* sh)
{
  SH2movbsg(sh,(5<<4)+13);
}
#define init_12_0_5_14
void FASTCALL call_12_0_5_14(SH2_struct* sh)
{
  SH2movbsg(sh,(5<<4)+14);
}
#define init_12_0_5_15
void FASTCALL call_12_0_5_15(SH2_struct* sh)
{
  SH2movbsg(sh,(5<<4)+15);
}
#define init_12_0_6_0
void FASTCALL call_12_0_6_0(SH2_struct* sh)
{
  SH2movbsg(sh,(6<<4)+0);
}
#define init_12_0_6_1
void FASTCALL call_12_0_6_1(SH2_struct* sh)
{
  SH2movbsg(sh,(6<<4)+1);
}
#define init_12_0_6_2
void FASTCALL call_12_0_6_2(SH2_struct* sh)
{
  SH2movbsg(sh,(6<<4)+2);
}
#define init_12_0_6_3
void FASTCALL call_12_0_6_3(SH2_struct* sh)
{
  SH2movbsg(sh,(6<<4)+3);
}
#define init_12_0_6_4
void FASTCALL call_12_0_6_4(SH2_struct* sh)
{
  SH2movbsg(sh,(6<<4)+4);
}
#define init_12_0_6_5
void FASTCALL call_12_0_6_5(SH2_struct* sh)
{
  SH2movbsg(sh,(6<<4)+5);
}
#define init_12_0_6_6
void FASTCALL call_12_0_6_6(SH2_struct* sh)
{
  SH2movbsg(sh,(6<<4)+6);
}
#define init_12_0_6_7
void FASTCALL call_12_0_6_7(SH2_struct* sh)
{
  SH2movbsg(sh,(6<<4)+7);
}
#define init_12_0_6_8
void FASTCALL call_12_0_6_8(SH2_struct* sh)
{
  SH2movbsg(sh,(6<<4)+8);
}
#define init_12_0_6_9
void FASTCALL call_12_0_6_9(SH2_struct* sh)
{
  SH2movbsg(sh,(6<<4)+9);
}
#define init_12_0_6_10
void FASTCALL call_12_0_6_10(SH2_struct* sh)
{
  SH2movbsg(sh,(6<<4)+10);
}
#define init_12_0_6_11
void FASTCALL call_12_0_6_11(SH2_struct* sh)
{
  SH2movbsg(sh,(6<<4)+11);
}
#define init_12_0_6_12
void FASTCALL call_12_0_6_12(SH2_struct* sh)
{
  SH2movbsg(sh,(6<<4)+12);
}
#define init_12_0_6_13
void FASTCALL call_12_0_6_13(SH2_struct* sh)
{
  SH2movbsg(sh,(6<<4)+13);
}
#define init_12_0_6_14
void FASTCALL call_12_0_6_14(SH2_struct* sh)
{
  SH2movbsg(sh,(6<<4)+14);
}
#define init_12_0_6_15
void FASTCALL call_12_0_6_15(SH2_struct* sh)
{
  SH2movbsg(sh,(6<<4)+15);
}
#define init_12_0_7_0
void FASTCALL call_12_0_7_0(SH2_struct* sh)
{
  SH2movbsg(sh,(7<<4)+0);
}
#define init_12_0_7_1
void FASTCALL call_12_0_7_1(SH2_struct* sh)
{
  SH2movbsg(sh,(7<<4)+1);
}
#define init_12_0_7_2
void FASTCALL call_12_0_7_2(SH2_struct* sh)
{
  SH2movbsg(sh,(7<<4)+2);
}
#define init_12_0_7_3
void FASTCALL call_12_0_7_3(SH2_struct* sh)
{
  SH2movbsg(sh,(7<<4)+3);
}
#define init_12_0_7_4
void FASTCALL call_12_0_7_4(SH2_struct* sh)
{
  SH2movbsg(sh,(7<<4)+4);
}
#define init_12_0_7_5
void FASTCALL call_12_0_7_5(SH2_struct* sh)
{
  SH2movbsg(sh,(7<<4)+5);
}
#define init_12_0_7_6
void FASTCALL call_12_0_7_6(SH2_struct* sh)
{
  SH2movbsg(sh,(7<<4)+6);
}
#define init_12_0_7_7
void FASTCALL call_12_0_7_7(SH2_struct* sh)
{
  SH2movbsg(sh,(7<<4)+7);
}
#define init_12_0_7_8
void FASTCALL call_12_0_7_8(SH2_struct* sh)
{
  SH2movbsg(sh,(7<<4)+8);
}
#define init_12_0_7_9
void FASTCALL call_12_0_7_9(SH2_struct* sh)
{
  SH2movbsg(sh,(7<<4)+9);
}
#define init_12_0_7_10
void FASTCALL call_12_0_7_10(SH2_struct* sh)
{
  SH2movbsg(sh,(7<<4)+10);
}
#define init_12_0_7_11
void FASTCALL call_12_0_7_11(SH2_struct* sh)
{
  SH2movbsg(sh,(7<<4)+11);
}
#define init_12_0_7_12
void FASTCALL call_12_0_7_12(SH2_struct* sh)
{
  SH2movbsg(sh,(7<<4)+12);
}
#define init_12_0_7_13
void FASTCALL call_12_0_7_13(SH2_struct* sh)
{
  SH2movbsg(sh,(7<<4)+13);
}
#define init_12_0_7_14
void FASTCALL call_12_0_7_14(SH2_struct* sh)
{
  SH2movbsg(sh,(7<<4)+14);
}
#define init_12_0_7_15
void FASTCALL call_12_0_7_15(SH2_struct* sh)
{
  SH2movbsg(sh,(7<<4)+15);
}
#define init_12_0_8_0
void FASTCALL call_12_0_8_0(SH2_struct* sh)
{
  SH2movbsg(sh,(8<<4)+0);
}
#define init_12_0_8_1
void FASTCALL call_12_0_8_1(SH2_struct* sh)
{
  SH2movbsg(sh,(8<<4)+1);
}
#define init_12_0_8_2
void FASTCALL call_12_0_8_2(SH2_struct* sh)
{
  SH2movbsg(sh,(8<<4)+2);
}
#define init_12_0_8_3
void FASTCALL call_12_0_8_3(SH2_struct* sh)
{
  SH2movbsg(sh,(8<<4)+3);
}
#define init_12_0_8_4
void FASTCALL call_12_0_8_4(SH2_struct* sh)
{
  SH2movbsg(sh,(8<<4)+4);
}
#define init_12_0_8_5
void FASTCALL call_12_0_8_5(SH2_struct* sh)
{
  SH2movbsg(sh,(8<<4)+5);
}
#define init_12_0_8_6
void FASTCALL call_12_0_8_6(SH2_struct* sh)
{
  SH2movbsg(sh,(8<<4)+6);
}
#define init_12_0_8_7
void FASTCALL call_12_0_8_7(SH2_struct* sh)
{
  SH2movbsg(sh,(8<<4)+7);
}
#define init_12_0_8_8
void FASTCALL call_12_0_8_8(SH2_struct* sh)
{
  SH2movbsg(sh,(8<<4)+8);
}
#define init_12_0_8_9
void FASTCALL call_12_0_8_9(SH2_struct* sh)
{
  SH2movbsg(sh,(8<<4)+9);
}
#define init_12_0_8_10
void FASTCALL call_12_0_8_10(SH2_struct* sh)
{
  SH2movbsg(sh,(8<<4)+10);
}
#define init_12_0_8_11
void FASTCALL call_12_0_8_11(SH2_struct* sh)
{
  SH2movbsg(sh,(8<<4)+11);
}
#define init_12_0_8_12
void FASTCALL call_12_0_8_12(SH2_struct* sh)
{
  SH2movbsg(sh,(8<<4)+12);
}
#define init_12_0_8_13
void FASTCALL call_12_0_8_13(SH2_struct* sh)
{
  SH2movbsg(sh,(8<<4)+13);
}
#define init_12_0_8_14
void FASTCALL call_12_0_8_14(SH2_struct* sh)
{
  SH2movbsg(sh,(8<<4)+14);
}
#define init_12_0_8_15
void FASTCALL call_12_0_8_15(SH2_struct* sh)
{
  SH2movbsg(sh,(8<<4)+15);
}
#define init_12_0_9_0
void FASTCALL call_12_0_9_0(SH2_struct* sh)
{
  SH2movbsg(sh,(9<<4)+0);
}
#define init_12_0_9_1
void FASTCALL call_12_0_9_1(SH2_struct* sh)
{
  SH2movbsg(sh,(9<<4)+1);
}
#define init_12_0_9_2
void FASTCALL call_12_0_9_2(SH2_struct* sh)
{
  SH2movbsg(sh,(9<<4)+2);
}
#define init_12_0_9_3
void FASTCALL call_12_0_9_3(SH2_struct* sh)
{
  SH2movbsg(sh,(9<<4)+3);
}
#define init_12_0_9_4
void FASTCALL call_12_0_9_4(SH2_struct* sh)
{
  SH2movbsg(sh,(9<<4)+4);
}
#define init_12_0_9_5
void FASTCALL call_12_0_9_5(SH2_struct* sh)
{
  SH2movbsg(sh,(9<<4)+5);
}
#define init_12_0_9_6
void FASTCALL call_12_0_9_6(SH2_struct* sh)
{
  SH2movbsg(sh,(9<<4)+6);
}
#define init_12_0_9_7
void FASTCALL call_12_0_9_7(SH2_struct* sh)
{
  SH2movbsg(sh,(9<<4)+7);
}
#define init_12_0_9_8
void FASTCALL call_12_0_9_8(SH2_struct* sh)
{
  SH2movbsg(sh,(9<<4)+8);
}
#define init_12_0_9_9
void FASTCALL call_12_0_9_9(SH2_struct* sh)
{
  SH2movbsg(sh,(9<<4)+9);
}
#define init_12_0_9_10
void FASTCALL call_12_0_9_10(SH2_struct* sh)
{
  SH2movbsg(sh,(9<<4)+10);
}
#define init_12_0_9_11
void FASTCALL call_12_0_9_11(SH2_struct* sh)
{
  SH2movbsg(sh,(9<<4)+11);
}
#define init_12_0_9_12
void FASTCALL call_12_0_9_12(SH2_struct* sh)
{
  SH2movbsg(sh,(9<<4)+12);
}
#define init_12_0_9_13
void FASTCALL call_12_0_9_13(SH2_struct* sh)
{
  SH2movbsg(sh,(9<<4)+13);
}
#define init_12_0_9_14
void FASTCALL call_12_0_9_14(SH2_struct* sh)
{
  SH2movbsg(sh,(9<<4)+14);
}
#define init_12_0_9_15
void FASTCALL call_12_0_9_15(SH2_struct* sh)
{
  SH2movbsg(sh,(9<<4)+15);
}
#define init_12_0_10_0
void FASTCALL call_12_0_10_0(SH2_struct* sh)
{
  SH2movbsg(sh,(10<<4)+0);
}
#define init_12_0_10_1
void FASTCALL call_12_0_10_1(SH2_struct* sh)
{
  SH2movbsg(sh,(10<<4)+1);
}
#define init_12_0_10_2
void FASTCALL call_12_0_10_2(SH2_struct* sh)
{
  SH2movbsg(sh,(10<<4)+2);
}
#define init_12_0_10_3
void FASTCALL call_12_0_10_3(SH2_struct* sh)
{
  SH2movbsg(sh,(10<<4)+3);
}
#define init_12_0_10_4
void FASTCALL call_12_0_10_4(SH2_struct* sh)
{
  SH2movbsg(sh,(10<<4)+4);
}
#define init_12_0_10_5
void FASTCALL call_12_0_10_5(SH2_struct* sh)
{
  SH2movbsg(sh,(10<<4)+5);
}
#define init_12_0_10_6
void FASTCALL call_12_0_10_6(SH2_struct* sh)
{
  SH2movbsg(sh,(10<<4)+6);
}
#define init_12_0_10_7
void FASTCALL call_12_0_10_7(SH2_struct* sh)
{
  SH2movbsg(sh,(10<<4)+7);
}
#define init_12_0_10_8
void FASTCALL call_12_0_10_8(SH2_struct* sh)
{
  SH2movbsg(sh,(10<<4)+8);
}
#define init_12_0_10_9
void FASTCALL call_12_0_10_9(SH2_struct* sh)
{
  SH2movbsg(sh,(10<<4)+9);
}
#define init_12_0_10_10
void FASTCALL call_12_0_10_10(SH2_struct* sh)
{
  SH2movbsg(sh,(10<<4)+10);
}
#define init_12_0_10_11
void FASTCALL call_12_0_10_11(SH2_struct* sh)
{
  SH2movbsg(sh,(10<<4)+11);
}
#define init_12_0_10_12
void FASTCALL call_12_0_10_12(SH2_struct* sh)
{
  SH2movbsg(sh,(10<<4)+12);
}
#define init_12_0_10_13
void FASTCALL call_12_0_10_13(SH2_struct* sh)
{
  SH2movbsg(sh,(10<<4)+13);
}
#define init_12_0_10_14
void FASTCALL call_12_0_10_14(SH2_struct* sh)
{
  SH2movbsg(sh,(10<<4)+14);
}
#define init_12_0_10_15
void FASTCALL call_12_0_10_15(SH2_struct* sh)
{
  SH2movbsg(sh,(10<<4)+15);
}
#define init_12_0_11_0
void FASTCALL call_12_0_11_0(SH2_struct* sh)
{
  SH2movbsg(sh,(11<<4)+0);
}
#define init_12_0_11_1
void FASTCALL call_12_0_11_1(SH2_struct* sh)
{
  SH2movbsg(sh,(11<<4)+1);
}
#define init_12_0_11_2
void FASTCALL call_12_0_11_2(SH2_struct* sh)
{
  SH2movbsg(sh,(11<<4)+2);
}
#define init_12_0_11_3
void FASTCALL call_12_0_11_3(SH2_struct* sh)
{
  SH2movbsg(sh,(11<<4)+3);
}
#define init_12_0_11_4
void FASTCALL call_12_0_11_4(SH2_struct* sh)
{
  SH2movbsg(sh,(11<<4)+4);
}
#define init_12_0_11_5
void FASTCALL call_12_0_11_5(SH2_struct* sh)
{
  SH2movbsg(sh,(11<<4)+5);
}
#define init_12_0_11_6
void FASTCALL call_12_0_11_6(SH2_struct* sh)
{
  SH2movbsg(sh,(11<<4)+6);
}
#define init_12_0_11_7
void FASTCALL call_12_0_11_7(SH2_struct* sh)
{
  SH2movbsg(sh,(11<<4)+7);
}
#define init_12_0_11_8
void FASTCALL call_12_0_11_8(SH2_struct* sh)
{
  SH2movbsg(sh,(11<<4)+8);
}
#define init_12_0_11_9
void FASTCALL call_12_0_11_9(SH2_struct* sh)
{
  SH2movbsg(sh,(11<<4)+9);
}
#define init_12_0_11_10
void FASTCALL call_12_0_11_10(SH2_struct* sh)
{
  SH2movbsg(sh,(11<<4)+10);
}
#define init_12_0_11_11
void FASTCALL call_12_0_11_11(SH2_struct* sh)
{
  SH2movbsg(sh,(11<<4)+11);
}
#define init_12_0_11_12
void FASTCALL call_12_0_11_12(SH2_struct* sh)
{
  SH2movbsg(sh,(11<<4)+12);
}
#define init_12_0_11_13
void FASTCALL call_12_0_11_13(SH2_struct* sh)
{
  SH2movbsg(sh,(11<<4)+13);
}
#define init_12_0_11_14
void FASTCALL call_12_0_11_14(SH2_struct* sh)
{
  SH2movbsg(sh,(11<<4)+14);
}
#define init_12_0_11_15
void FASTCALL call_12_0_11_15(SH2_struct* sh)
{
  SH2movbsg(sh,(11<<4)+15);
}
#define init_12_0_12_0
void FASTCALL call_12_0_12_0(SH2_struct* sh)
{
  SH2movbsg(sh,(12<<4)+0);
}
#define init_12_0_12_1
void FASTCALL call_12_0_12_1(SH2_struct* sh)
{
  SH2movbsg(sh,(12<<4)+1);
}
#define init_12_0_12_2
void FASTCALL call_12_0_12_2(SH2_struct* sh)
{
  SH2movbsg(sh,(12<<4)+2);
}
#define init_12_0_12_3
void FASTCALL call_12_0_12_3(SH2_struct* sh)
{
  SH2movbsg(sh,(12<<4)+3);
}
#define init_12_0_12_4
void FASTCALL call_12_0_12_4(SH2_struct* sh)
{
  SH2movbsg(sh,(12<<4)+4);
}
#define init_12_0_12_5
void FASTCALL call_12_0_12_5(SH2_struct* sh)
{
  SH2movbsg(sh,(12<<4)+5);
}
#define init_12_0_12_6
void FASTCALL call_12_0_12_6(SH2_struct* sh)
{
  SH2movbsg(sh,(12<<4)+6);
}
#define init_12_0_12_7
void FASTCALL call_12_0_12_7(SH2_struct* sh)
{
  SH2movbsg(sh,(12<<4)+7);
}
#define init_12_0_12_8
void FASTCALL call_12_0_12_8(SH2_struct* sh)
{
  SH2movbsg(sh,(12<<4)+8);
}
#define init_12_0_12_9
void FASTCALL call_12_0_12_9(SH2_struct* sh)
{
  SH2movbsg(sh,(12<<4)+9);
}
#define init_12_0_12_10
void FASTCALL call_12_0_12_10(SH2_struct* sh)
{
  SH2movbsg(sh,(12<<4)+10);
}
#define init_12_0_12_11
void FASTCALL call_12_0_12_11(SH2_struct* sh)
{
  SH2movbsg(sh,(12<<4)+11);
}
#define init_12_0_12_12
void FASTCALL call_12_0_12_12(SH2_struct* sh)
{
  SH2movbsg(sh,(12<<4)+12);
}
#define init_12_0_12_13
void FASTCALL call_12_0_12_13(SH2_struct* sh)
{
  SH2movbsg(sh,(12<<4)+13);
}
#define init_12_0_12_14
void FASTCALL call_12_0_12_14(SH2_struct* sh)
{
  SH2movbsg(sh,(12<<4)+14);
}
#define init_12_0_12_15
void FASTCALL call_12_0_12_15(SH2_struct* sh)
{
  SH2movbsg(sh,(12<<4)+15);
}
#define init_12_0_13_0
void FASTCALL call_12_0_13_0(SH2_struct* sh)
{
  SH2movbsg(sh,(13<<4)+0);
}
#define init_12_0_13_1
void FASTCALL call_12_0_13_1(SH2_struct* sh)
{
  SH2movbsg(sh,(13<<4)+1);
}
#define init_12_0_13_2
void FASTCALL call_12_0_13_2(SH2_struct* sh)
{
  SH2movbsg(sh,(13<<4)+2);
}
#define init_12_0_13_3
void FASTCALL call_12_0_13_3(SH2_struct* sh)
{
  SH2movbsg(sh,(13<<4)+3);
}
#define init_12_0_13_4
void FASTCALL call_12_0_13_4(SH2_struct* sh)
{
  SH2movbsg(sh,(13<<4)+4);
}
#define init_12_0_13_5
void FASTCALL call_12_0_13_5(SH2_struct* sh)
{
  SH2movbsg(sh,(13<<4)+5);
}
#define init_12_0_13_6
void FASTCALL call_12_0_13_6(SH2_struct* sh)
{
  SH2movbsg(sh,(13<<4)+6);
}
#define init_12_0_13_7
void FASTCALL call_12_0_13_7(SH2_struct* sh)
{
  SH2movbsg(sh,(13<<4)+7);
}
#define init_12_0_13_8
void FASTCALL call_12_0_13_8(SH2_struct* sh)
{
  SH2movbsg(sh,(13<<4)+8);
}
#define init_12_0_13_9
void FASTCALL call_12_0_13_9(SH2_struct* sh)
{
  SH2movbsg(sh,(13<<4)+9);
}
#define init_12_0_13_10
void FASTCALL call_12_0_13_10(SH2_struct* sh)
{
  SH2movbsg(sh,(13<<4)+10);
}
#define init_12_0_13_11
void FASTCALL call_12_0_13_11(SH2_struct* sh)
{
  SH2movbsg(sh,(13<<4)+11);
}
#define init_12_0_13_12
void FASTCALL call_12_0_13_12(SH2_struct* sh)
{
  SH2movbsg(sh,(13<<4)+12);
}
#define init_12_0_13_13
void FASTCALL call_12_0_13_13(SH2_struct* sh)
{
  SH2movbsg(sh,(13<<4)+13);
}
#define init_12_0_13_14
void FASTCALL call_12_0_13_14(SH2_struct* sh)
{
  SH2movbsg(sh,(13<<4)+14);
}
#define init_12_0_13_15
void FASTCALL call_12_0_13_15(SH2_struct* sh)
{
  SH2movbsg(sh,(13<<4)+15);
}
#define init_12_0_14_0
void FASTCALL call_12_0_14_0(SH2_struct* sh)
{
  SH2movbsg(sh,(14<<4)+0);
}
#define init_12_0_14_1
void FASTCALL call_12_0_14_1(SH2_struct* sh)
{
  SH2movbsg(sh,(14<<4)+1);
}
#define init_12_0_14_2
void FASTCALL call_12_0_14_2(SH2_struct* sh)
{
  SH2movbsg(sh,(14<<4)+2);
}
#define init_12_0_14_3
void FASTCALL call_12_0_14_3(SH2_struct* sh)
{
  SH2movbsg(sh,(14<<4)+3);
}
#define init_12_0_14_4
void FASTCALL call_12_0_14_4(SH2_struct* sh)
{
  SH2movbsg(sh,(14<<4)+4);
}
#define init_12_0_14_5
void FASTCALL call_12_0_14_5(SH2_struct* sh)
{
  SH2movbsg(sh,(14<<4)+5);
}
#define init_12_0_14_6
void FASTCALL call_12_0_14_6(SH2_struct* sh)
{
  SH2movbsg(sh,(14<<4)+6);
}
#define init_12_0_14_7
void FASTCALL call_12_0_14_7(SH2_struct* sh)
{
  SH2movbsg(sh,(14<<4)+7);
}
#define init_12_0_14_8
void FASTCALL call_12_0_14_8(SH2_struct* sh)
{
  SH2movbsg(sh,(14<<4)+8);
}
#define init_12_0_14_9
void FASTCALL call_12_0_14_9(SH2_struct* sh)
{
  SH2movbsg(sh,(14<<4)+9);
}
#define init_12_0_14_10
void FASTCALL call_12_0_14_10(SH2_struct* sh)
{
  SH2movbsg(sh,(14<<4)+10);
}
#define init_12_0_14_11
void FASTCALL call_12_0_14_11(SH2_struct* sh)
{
  SH2movbsg(sh,(14<<4)+11);
}
#define init_12_0_14_12
void FASTCALL call_12_0_14_12(SH2_struct* sh)
{
  SH2movbsg(sh,(14<<4)+12);
}
#define init_12_0_14_13
void FASTCALL call_12_0_14_13(SH2_struct* sh)
{
  SH2movbsg(sh,(14<<4)+13);
}
#define init_12_0_14_14
void FASTCALL call_12_0_14_14(SH2_struct* sh)
{
  SH2movbsg(sh,(14<<4)+14);
}
#define init_12_0_14_15
void FASTCALL call_12_0_14_15(SH2_struct* sh)
{
  SH2movbsg(sh,(14<<4)+15);
}
#define init_12_0_15_0
void FASTCALL call_12_0_15_0(SH2_struct* sh)
{
  SH2movbsg(sh,(15<<4)+0);
}
#define init_12_0_15_1
void FASTCALL call_12_0_15_1(SH2_struct* sh)
{
  SH2movbsg(sh,(15<<4)+1);
}
#define init_12_0_15_2
void FASTCALL call_12_0_15_2(SH2_struct* sh)
{
  SH2movbsg(sh,(15<<4)+2);
}
#define init_12_0_15_3
void FASTCALL call_12_0_15_3(SH2_struct* sh)
{
  SH2movbsg(sh,(15<<4)+3);
}
#define init_12_0_15_4
void FASTCALL call_12_0_15_4(SH2_struct* sh)
{
  SH2movbsg(sh,(15<<4)+4);
}
#define init_12_0_15_5
void FASTCALL call_12_0_15_5(SH2_struct* sh)
{
  SH2movbsg(sh,(15<<4)+5);
}
#define init_12_0_15_6
void FASTCALL call_12_0_15_6(SH2_struct* sh)
{
  SH2movbsg(sh,(15<<4)+6);
}
#define init_12_0_15_7
void FASTCALL call_12_0_15_7(SH2_struct* sh)
{
  SH2movbsg(sh,(15<<4)+7);
}
#define init_12_0_15_8
void FASTCALL call_12_0_15_8(SH2_struct* sh)
{
  SH2movbsg(sh,(15<<4)+8);
}
#define init_12_0_15_9
void FASTCALL call_12_0_15_9(SH2_struct* sh)
{
  SH2movbsg(sh,(15<<4)+9);
}
#define init_12_0_15_10
void FASTCALL call_12_0_15_10(SH2_struct* sh)
{
  SH2movbsg(sh,(15<<4)+10);
}
#define init_12_0_15_11
void FASTCALL call_12_0_15_11(SH2_struct* sh)
{
  SH2movbsg(sh,(15<<4)+11);
}
#define init_12_0_15_12
void FASTCALL call_12_0_15_12(SH2_struct* sh)
{
  SH2movbsg(sh,(15<<4)+12);
}
#define init_12_0_15_13
void FASTCALL call_12_0_15_13(SH2_struct* sh)
{
  SH2movbsg(sh,(15<<4)+13);
}
#define init_12_0_15_14
void FASTCALL call_12_0_15_14(SH2_struct* sh)
{
  SH2movbsg(sh,(15<<4)+14);
}
#define init_12_0_15_15
void FASTCALL call_12_0_15_15(SH2_struct* sh)
{
  SH2movbsg(sh,(15<<4)+15);
}
#define init_8_0_0_0
void FASTCALL call_8_0_0_0(SH2_struct* sh)
{
  SH2movbs4(sh,0,0);
}
#define init_8_0_0_1
void FASTCALL call_8_0_0_1(SH2_struct* sh)
{
  SH2movbs4(sh,0,1);
}
#define init_8_0_0_2
void FASTCALL call_8_0_0_2(SH2_struct* sh)
{
  SH2movbs4(sh,0,2);
}
#define init_8_0_0_3
void FASTCALL call_8_0_0_3(SH2_struct* sh)
{
  SH2movbs4(sh,0,3);
}
#define init_8_0_0_4
void FASTCALL call_8_0_0_4(SH2_struct* sh)
{
  SH2movbs4(sh,0,4);
}
#define init_8_0_0_5
void FASTCALL call_8_0_0_5(SH2_struct* sh)
{
  SH2movbs4(sh,0,5);
}
#define init_8_0_0_6
void FASTCALL call_8_0_0_6(SH2_struct* sh)
{
  SH2movbs4(sh,0,6);
}
#define init_8_0_0_7
void FASTCALL call_8_0_0_7(SH2_struct* sh)
{
  SH2movbs4(sh,0,7);
}
#define init_8_0_0_8
void FASTCALL call_8_0_0_8(SH2_struct* sh)
{
  SH2movbs4(sh,0,8);
}
#define init_8_0_0_9
void FASTCALL call_8_0_0_9(SH2_struct* sh)
{
  SH2movbs4(sh,0,9);
}
#define init_8_0_0_10
void FASTCALL call_8_0_0_10(SH2_struct* sh)
{
  SH2movbs4(sh,0,10);
}
#define init_8_0_0_11
void FASTCALL call_8_0_0_11(SH2_struct* sh)
{
  SH2movbs4(sh,0,11);
}
#define init_8_0_0_12
void FASTCALL call_8_0_0_12(SH2_struct* sh)
{
  SH2movbs4(sh,0,12);
}
#define init_8_0_0_13
void FASTCALL call_8_0_0_13(SH2_struct* sh)
{
  SH2movbs4(sh,0,13);
}
#define init_8_0_0_14
void FASTCALL call_8_0_0_14(SH2_struct* sh)
{
  SH2movbs4(sh,0,14);
}
#define init_8_0_0_15
void FASTCALL call_8_0_0_15(SH2_struct* sh)
{
  SH2movbs4(sh,0,15);
}
#define init_8_0_1_0
void FASTCALL call_8_0_1_0(SH2_struct* sh)
{
  SH2movbs4(sh,1,0);
}
#define init_8_0_1_1
void FASTCALL call_8_0_1_1(SH2_struct* sh)
{
  SH2movbs4(sh,1,1);
}
#define init_8_0_1_2
void FASTCALL call_8_0_1_2(SH2_struct* sh)
{
  SH2movbs4(sh,1,2);
}
#define init_8_0_1_3
void FASTCALL call_8_0_1_3(SH2_struct* sh)
{
  SH2movbs4(sh,1,3);
}
#define init_8_0_1_4
void FASTCALL call_8_0_1_4(SH2_struct* sh)
{
  SH2movbs4(sh,1,4);
}
#define init_8_0_1_5
void FASTCALL call_8_0_1_5(SH2_struct* sh)
{
  SH2movbs4(sh,1,5);
}
#define init_8_0_1_6
void FASTCALL call_8_0_1_6(SH2_struct* sh)
{
  SH2movbs4(sh,1,6);
}
#define init_8_0_1_7
void FASTCALL call_8_0_1_7(SH2_struct* sh)
{
  SH2movbs4(sh,1,7);
}
#define init_8_0_1_8
void FASTCALL call_8_0_1_8(SH2_struct* sh)
{
  SH2movbs4(sh,1,8);
}
#define init_8_0_1_9
void FASTCALL call_8_0_1_9(SH2_struct* sh)
{
  SH2movbs4(sh,1,9);
}
#define init_8_0_1_10
void FASTCALL call_8_0_1_10(SH2_struct* sh)
{
  SH2movbs4(sh,1,10);
}
#define init_8_0_1_11
void FASTCALL call_8_0_1_11(SH2_struct* sh)
{
  SH2movbs4(sh,1,11);
}
#define init_8_0_1_12
void FASTCALL call_8_0_1_12(SH2_struct* sh)
{
  SH2movbs4(sh,1,12);
}
#define init_8_0_1_13
void FASTCALL call_8_0_1_13(SH2_struct* sh)
{
  SH2movbs4(sh,1,13);
}
#define init_8_0_1_14
void FASTCALL call_8_0_1_14(SH2_struct* sh)
{
  SH2movbs4(sh,1,14);
}
#define init_8_0_1_15
void FASTCALL call_8_0_1_15(SH2_struct* sh)
{
  SH2movbs4(sh,1,15);
}
#define init_8_0_2_0
void FASTCALL call_8_0_2_0(SH2_struct* sh)
{
  SH2movbs4(sh,2,0);
}
#define init_8_0_2_1
void FASTCALL call_8_0_2_1(SH2_struct* sh)
{
  SH2movbs4(sh,2,1);
}
#define init_8_0_2_2
void FASTCALL call_8_0_2_2(SH2_struct* sh)
{
  SH2movbs4(sh,2,2);
}
#define init_8_0_2_3
void FASTCALL call_8_0_2_3(SH2_struct* sh)
{
  SH2movbs4(sh,2,3);
}
#define init_8_0_2_4
void FASTCALL call_8_0_2_4(SH2_struct* sh)
{
  SH2movbs4(sh,2,4);
}
#define init_8_0_2_5
void FASTCALL call_8_0_2_5(SH2_struct* sh)
{
  SH2movbs4(sh,2,5);
}
#define init_8_0_2_6
void FASTCALL call_8_0_2_6(SH2_struct* sh)
{
  SH2movbs4(sh,2,6);
}
#define init_8_0_2_7
void FASTCALL call_8_0_2_7(SH2_struct* sh)
{
  SH2movbs4(sh,2,7);
}
#define init_8_0_2_8
void FASTCALL call_8_0_2_8(SH2_struct* sh)
{
  SH2movbs4(sh,2,8);
}
#define init_8_0_2_9
void FASTCALL call_8_0_2_9(SH2_struct* sh)
{
  SH2movbs4(sh,2,9);
}
#define init_8_0_2_10
void FASTCALL call_8_0_2_10(SH2_struct* sh)
{
  SH2movbs4(sh,2,10);
}
#define init_8_0_2_11
void FASTCALL call_8_0_2_11(SH2_struct* sh)
{
  SH2movbs4(sh,2,11);
}
#define init_8_0_2_12
void FASTCALL call_8_0_2_12(SH2_struct* sh)
{
  SH2movbs4(sh,2,12);
}
#define init_8_0_2_13
void FASTCALL call_8_0_2_13(SH2_struct* sh)
{
  SH2movbs4(sh,2,13);
}
#define init_8_0_2_14
void FASTCALL call_8_0_2_14(SH2_struct* sh)
{
  SH2movbs4(sh,2,14);
}
#define init_8_0_2_15
void FASTCALL call_8_0_2_15(SH2_struct* sh)
{
  SH2movbs4(sh,2,15);
}
#define init_8_0_3_0
void FASTCALL call_8_0_3_0(SH2_struct* sh)
{
  SH2movbs4(sh,3,0);
}
#define init_8_0_3_1
void FASTCALL call_8_0_3_1(SH2_struct* sh)
{
  SH2movbs4(sh,3,1);
}
#define init_8_0_3_2
void FASTCALL call_8_0_3_2(SH2_struct* sh)
{
  SH2movbs4(sh,3,2);
}
#define init_8_0_3_3
void FASTCALL call_8_0_3_3(SH2_struct* sh)
{
  SH2movbs4(sh,3,3);
}
#define init_8_0_3_4
void FASTCALL call_8_0_3_4(SH2_struct* sh)
{
  SH2movbs4(sh,3,4);
}
#define init_8_0_3_5
void FASTCALL call_8_0_3_5(SH2_struct* sh)
{
  SH2movbs4(sh,3,5);
}
#define init_8_0_3_6
void FASTCALL call_8_0_3_6(SH2_struct* sh)
{
  SH2movbs4(sh,3,6);
}
#define init_8_0_3_7
void FASTCALL call_8_0_3_7(SH2_struct* sh)
{
  SH2movbs4(sh,3,7);
}
#define init_8_0_3_8
void FASTCALL call_8_0_3_8(SH2_struct* sh)
{
  SH2movbs4(sh,3,8);
}
#define init_8_0_3_9
void FASTCALL call_8_0_3_9(SH2_struct* sh)
{
  SH2movbs4(sh,3,9);
}
#define init_8_0_3_10
void FASTCALL call_8_0_3_10(SH2_struct* sh)
{
  SH2movbs4(sh,3,10);
}
#define init_8_0_3_11
void FASTCALL call_8_0_3_11(SH2_struct* sh)
{
  SH2movbs4(sh,3,11);
}
#define init_8_0_3_12
void FASTCALL call_8_0_3_12(SH2_struct* sh)
{
  SH2movbs4(sh,3,12);
}
#define init_8_0_3_13
void FASTCALL call_8_0_3_13(SH2_struct* sh)
{
  SH2movbs4(sh,3,13);
}
#define init_8_0_3_14
void FASTCALL call_8_0_3_14(SH2_struct* sh)
{
  SH2movbs4(sh,3,14);
}
#define init_8_0_3_15
void FASTCALL call_8_0_3_15(SH2_struct* sh)
{
  SH2movbs4(sh,3,15);
}
#define init_8_0_4_0
void FASTCALL call_8_0_4_0(SH2_struct* sh)
{
  SH2movbs4(sh,4,0);
}
#define init_8_0_4_1
void FASTCALL call_8_0_4_1(SH2_struct* sh)
{
  SH2movbs4(sh,4,1);
}
#define init_8_0_4_2
void FASTCALL call_8_0_4_2(SH2_struct* sh)
{
  SH2movbs4(sh,4,2);
}
#define init_8_0_4_3
void FASTCALL call_8_0_4_3(SH2_struct* sh)
{
  SH2movbs4(sh,4,3);
}
#define init_8_0_4_4
void FASTCALL call_8_0_4_4(SH2_struct* sh)
{
  SH2movbs4(sh,4,4);
}
#define init_8_0_4_5
void FASTCALL call_8_0_4_5(SH2_struct* sh)
{
  SH2movbs4(sh,4,5);
}
#define init_8_0_4_6
void FASTCALL call_8_0_4_6(SH2_struct* sh)
{
  SH2movbs4(sh,4,6);
}
#define init_8_0_4_7
void FASTCALL call_8_0_4_7(SH2_struct* sh)
{
  SH2movbs4(sh,4,7);
}
#define init_8_0_4_8
void FASTCALL call_8_0_4_8(SH2_struct* sh)
{
  SH2movbs4(sh,4,8);
}
#define init_8_0_4_9
void FASTCALL call_8_0_4_9(SH2_struct* sh)
{
  SH2movbs4(sh,4,9);
}
#define init_8_0_4_10
void FASTCALL call_8_0_4_10(SH2_struct* sh)
{
  SH2movbs4(sh,4,10);
}
#define init_8_0_4_11
void FASTCALL call_8_0_4_11(SH2_struct* sh)
{
  SH2movbs4(sh,4,11);
}
#define init_8_0_4_12
void FASTCALL call_8_0_4_12(SH2_struct* sh)
{
  SH2movbs4(sh,4,12);
}
#define init_8_0_4_13
void FASTCALL call_8_0_4_13(SH2_struct* sh)
{
  SH2movbs4(sh,4,13);
}
#define init_8_0_4_14
void FASTCALL call_8_0_4_14(SH2_struct* sh)
{
  SH2movbs4(sh,4,14);
}
#define init_8_0_4_15
void FASTCALL call_8_0_4_15(SH2_struct* sh)
{
  SH2movbs4(sh,4,15);
}
#define init_8_0_5_0
void FASTCALL call_8_0_5_0(SH2_struct* sh)
{
  SH2movbs4(sh,5,0);
}
#define init_8_0_5_1
void FASTCALL call_8_0_5_1(SH2_struct* sh)
{
  SH2movbs4(sh,5,1);
}
#define init_8_0_5_2
void FASTCALL call_8_0_5_2(SH2_struct* sh)
{
  SH2movbs4(sh,5,2);
}
#define init_8_0_5_3
void FASTCALL call_8_0_5_3(SH2_struct* sh)
{
  SH2movbs4(sh,5,3);
}
#define init_8_0_5_4
void FASTCALL call_8_0_5_4(SH2_struct* sh)
{
  SH2movbs4(sh,5,4);
}
#define init_8_0_5_5
void FASTCALL call_8_0_5_5(SH2_struct* sh)
{
  SH2movbs4(sh,5,5);
}
#define init_8_0_5_6
void FASTCALL call_8_0_5_6(SH2_struct* sh)
{
  SH2movbs4(sh,5,6);
}
#define init_8_0_5_7
void FASTCALL call_8_0_5_7(SH2_struct* sh)
{
  SH2movbs4(sh,5,7);
}
#define init_8_0_5_8
void FASTCALL call_8_0_5_8(SH2_struct* sh)
{
  SH2movbs4(sh,5,8);
}
#define init_8_0_5_9
void FASTCALL call_8_0_5_9(SH2_struct* sh)
{
  SH2movbs4(sh,5,9);
}
#define init_8_0_5_10
void FASTCALL call_8_0_5_10(SH2_struct* sh)
{
  SH2movbs4(sh,5,10);
}
#define init_8_0_5_11
void FASTCALL call_8_0_5_11(SH2_struct* sh)
{
  SH2movbs4(sh,5,11);
}
#define init_8_0_5_12
void FASTCALL call_8_0_5_12(SH2_struct* sh)
{
  SH2movbs4(sh,5,12);
}
#define init_8_0_5_13
void FASTCALL call_8_0_5_13(SH2_struct* sh)
{
  SH2movbs4(sh,5,13);
}
#define init_8_0_5_14
void FASTCALL call_8_0_5_14(SH2_struct* sh)
{
  SH2movbs4(sh,5,14);
}
#define init_8_0_5_15
void FASTCALL call_8_0_5_15(SH2_struct* sh)
{
  SH2movbs4(sh,5,15);
}
#define init_8_0_6_0
void FASTCALL call_8_0_6_0(SH2_struct* sh)
{
  SH2movbs4(sh,6,0);
}
#define init_8_0_6_1
void FASTCALL call_8_0_6_1(SH2_struct* sh)
{
  SH2movbs4(sh,6,1);
}
#define init_8_0_6_2
void FASTCALL call_8_0_6_2(SH2_struct* sh)
{
  SH2movbs4(sh,6,2);
}
#define init_8_0_6_3
void FASTCALL call_8_0_6_3(SH2_struct* sh)
{
  SH2movbs4(sh,6,3);
}
#define init_8_0_6_4
void FASTCALL call_8_0_6_4(SH2_struct* sh)
{
  SH2movbs4(sh,6,4);
}
#define init_8_0_6_5
void FASTCALL call_8_0_6_5(SH2_struct* sh)
{
  SH2movbs4(sh,6,5);
}
#define init_8_0_6_6
void FASTCALL call_8_0_6_6(SH2_struct* sh)
{
  SH2movbs4(sh,6,6);
}
#define init_8_0_6_7
void FASTCALL call_8_0_6_7(SH2_struct* sh)
{
  SH2movbs4(sh,6,7);
}
#define init_8_0_6_8
void FASTCALL call_8_0_6_8(SH2_struct* sh)
{
  SH2movbs4(sh,6,8);
}
#define init_8_0_6_9
void FASTCALL call_8_0_6_9(SH2_struct* sh)
{
  SH2movbs4(sh,6,9);
}
#define init_8_0_6_10
void FASTCALL call_8_0_6_10(SH2_struct* sh)
{
  SH2movbs4(sh,6,10);
}
#define init_8_0_6_11
void FASTCALL call_8_0_6_11(SH2_struct* sh)
{
  SH2movbs4(sh,6,11);
}
#define init_8_0_6_12
void FASTCALL call_8_0_6_12(SH2_struct* sh)
{
  SH2movbs4(sh,6,12);
}
#define init_8_0_6_13
void FASTCALL call_8_0_6_13(SH2_struct* sh)
{
  SH2movbs4(sh,6,13);
}
#define init_8_0_6_14
void FASTCALL call_8_0_6_14(SH2_struct* sh)
{
  SH2movbs4(sh,6,14);
}
#define init_8_0_6_15
void FASTCALL call_8_0_6_15(SH2_struct* sh)
{
  SH2movbs4(sh,6,15);
}
#define init_8_0_7_0
void FASTCALL call_8_0_7_0(SH2_struct* sh)
{
  SH2movbs4(sh,7,0);
}
#define init_8_0_7_1
void FASTCALL call_8_0_7_1(SH2_struct* sh)
{
  SH2movbs4(sh,7,1);
}
#define init_8_0_7_2
void FASTCALL call_8_0_7_2(SH2_struct* sh)
{
  SH2movbs4(sh,7,2);
}
#define init_8_0_7_3
void FASTCALL call_8_0_7_3(SH2_struct* sh)
{
  SH2movbs4(sh,7,3);
}
#define init_8_0_7_4
void FASTCALL call_8_0_7_4(SH2_struct* sh)
{
  SH2movbs4(sh,7,4);
}
#define init_8_0_7_5
void FASTCALL call_8_0_7_5(SH2_struct* sh)
{
  SH2movbs4(sh,7,5);
}
#define init_8_0_7_6
void FASTCALL call_8_0_7_6(SH2_struct* sh)
{
  SH2movbs4(sh,7,6);
}
#define init_8_0_7_7
void FASTCALL call_8_0_7_7(SH2_struct* sh)
{
  SH2movbs4(sh,7,7);
}
#define init_8_0_7_8
void FASTCALL call_8_0_7_8(SH2_struct* sh)
{
  SH2movbs4(sh,7,8);
}
#define init_8_0_7_9
void FASTCALL call_8_0_7_9(SH2_struct* sh)
{
  SH2movbs4(sh,7,9);
}
#define init_8_0_7_10
void FASTCALL call_8_0_7_10(SH2_struct* sh)
{
  SH2movbs4(sh,7,10);
}
#define init_8_0_7_11
void FASTCALL call_8_0_7_11(SH2_struct* sh)
{
  SH2movbs4(sh,7,11);
}
#define init_8_0_7_12
void FASTCALL call_8_0_7_12(SH2_struct* sh)
{
  SH2movbs4(sh,7,12);
}
#define init_8_0_7_13
void FASTCALL call_8_0_7_13(SH2_struct* sh)
{
  SH2movbs4(sh,7,13);
}
#define init_8_0_7_14
void FASTCALL call_8_0_7_14(SH2_struct* sh)
{
  SH2movbs4(sh,7,14);
}
#define init_8_0_7_15
void FASTCALL call_8_0_7_15(SH2_struct* sh)
{
  SH2movbs4(sh,7,15);
}
#define init_8_0_8_0
void FASTCALL call_8_0_8_0(SH2_struct* sh)
{
  SH2movbs4(sh,8,0);
}
#define init_8_0_8_1
void FASTCALL call_8_0_8_1(SH2_struct* sh)
{
  SH2movbs4(sh,8,1);
}
#define init_8_0_8_2
void FASTCALL call_8_0_8_2(SH2_struct* sh)
{
  SH2movbs4(sh,8,2);
}
#define init_8_0_8_3
void FASTCALL call_8_0_8_3(SH2_struct* sh)
{
  SH2movbs4(sh,8,3);
}
#define init_8_0_8_4
void FASTCALL call_8_0_8_4(SH2_struct* sh)
{
  SH2movbs4(sh,8,4);
}
#define init_8_0_8_5
void FASTCALL call_8_0_8_5(SH2_struct* sh)
{
  SH2movbs4(sh,8,5);
}
#define init_8_0_8_6
void FASTCALL call_8_0_8_6(SH2_struct* sh)
{
  SH2movbs4(sh,8,6);
}
#define init_8_0_8_7
void FASTCALL call_8_0_8_7(SH2_struct* sh)
{
  SH2movbs4(sh,8,7);
}
#define init_8_0_8_8
void FASTCALL call_8_0_8_8(SH2_struct* sh)
{
  SH2movbs4(sh,8,8);
}
#define init_8_0_8_9
void FASTCALL call_8_0_8_9(SH2_struct* sh)
{
  SH2movbs4(sh,8,9);
}
#define init_8_0_8_10
void FASTCALL call_8_0_8_10(SH2_struct* sh)
{
  SH2movbs4(sh,8,10);
}
#define init_8_0_8_11
void FASTCALL call_8_0_8_11(SH2_struct* sh)
{
  SH2movbs4(sh,8,11);
}
#define init_8_0_8_12
void FASTCALL call_8_0_8_12(SH2_struct* sh)
{
  SH2movbs4(sh,8,12);
}
#define init_8_0_8_13
void FASTCALL call_8_0_8_13(SH2_struct* sh)
{
  SH2movbs4(sh,8,13);
}
#define init_8_0_8_14
void FASTCALL call_8_0_8_14(SH2_struct* sh)
{
  SH2movbs4(sh,8,14);
}
#define init_8_0_8_15
void FASTCALL call_8_0_8_15(SH2_struct* sh)
{
  SH2movbs4(sh,8,15);
}
#define init_8_0_9_0
void FASTCALL call_8_0_9_0(SH2_struct* sh)
{
  SH2movbs4(sh,9,0);
}
#define init_8_0_9_1
void FASTCALL call_8_0_9_1(SH2_struct* sh)
{
  SH2movbs4(sh,9,1);
}
#define init_8_0_9_2
void FASTCALL call_8_0_9_2(SH2_struct* sh)
{
  SH2movbs4(sh,9,2);
}
#define init_8_0_9_3
void FASTCALL call_8_0_9_3(SH2_struct* sh)
{
  SH2movbs4(sh,9,3);
}
#define init_8_0_9_4
void FASTCALL call_8_0_9_4(SH2_struct* sh)
{
  SH2movbs4(sh,9,4);
}
#define init_8_0_9_5
void FASTCALL call_8_0_9_5(SH2_struct* sh)
{
  SH2movbs4(sh,9,5);
}
#define init_8_0_9_6
void FASTCALL call_8_0_9_6(SH2_struct* sh)
{
  SH2movbs4(sh,9,6);
}
#define init_8_0_9_7
void FASTCALL call_8_0_9_7(SH2_struct* sh)
{
  SH2movbs4(sh,9,7);
}
#define init_8_0_9_8
void FASTCALL call_8_0_9_8(SH2_struct* sh)
{
  SH2movbs4(sh,9,8);
}
#define init_8_0_9_9
void FASTCALL call_8_0_9_9(SH2_struct* sh)
{
  SH2movbs4(sh,9,9);
}
#define init_8_0_9_10
void FASTCALL call_8_0_9_10(SH2_struct* sh)
{
  SH2movbs4(sh,9,10);
}
#define init_8_0_9_11
void FASTCALL call_8_0_9_11(SH2_struct* sh)
{
  SH2movbs4(sh,9,11);
}
#define init_8_0_9_12
void FASTCALL call_8_0_9_12(SH2_struct* sh)
{
  SH2movbs4(sh,9,12);
}
#define init_8_0_9_13
void FASTCALL call_8_0_9_13(SH2_struct* sh)
{
  SH2movbs4(sh,9,13);
}
#define init_8_0_9_14
void FASTCALL call_8_0_9_14(SH2_struct* sh)
{
  SH2movbs4(sh,9,14);
}
#define init_8_0_9_15
void FASTCALL call_8_0_9_15(SH2_struct* sh)
{
  SH2movbs4(sh,9,15);
}
#define init_8_0_10_0
void FASTCALL call_8_0_10_0(SH2_struct* sh)
{
  SH2movbs4(sh,10,0);
}
#define init_8_0_10_1
void FASTCALL call_8_0_10_1(SH2_struct* sh)
{
  SH2movbs4(sh,10,1);
}
#define init_8_0_10_2
void FASTCALL call_8_0_10_2(SH2_struct* sh)
{
  SH2movbs4(sh,10,2);
}
#define init_8_0_10_3
void FASTCALL call_8_0_10_3(SH2_struct* sh)
{
  SH2movbs4(sh,10,3);
}
#define init_8_0_10_4
void FASTCALL call_8_0_10_4(SH2_struct* sh)
{
  SH2movbs4(sh,10,4);
}
#define init_8_0_10_5
void FASTCALL call_8_0_10_5(SH2_struct* sh)
{
  SH2movbs4(sh,10,5);
}
#define init_8_0_10_6
void FASTCALL call_8_0_10_6(SH2_struct* sh)
{
  SH2movbs4(sh,10,6);
}
#define init_8_0_10_7
void FASTCALL call_8_0_10_7(SH2_struct* sh)
{
  SH2movbs4(sh,10,7);
}
#define init_8_0_10_8
void FASTCALL call_8_0_10_8(SH2_struct* sh)
{
  SH2movbs4(sh,10,8);
}
#define init_8_0_10_9
void FASTCALL call_8_0_10_9(SH2_struct* sh)
{
  SH2movbs4(sh,10,9);
}
#define init_8_0_10_10
void FASTCALL call_8_0_10_10(SH2_struct* sh)
{
  SH2movbs4(sh,10,10);
}
#define init_8_0_10_11
void FASTCALL call_8_0_10_11(SH2_struct* sh)
{
  SH2movbs4(sh,10,11);
}
#define init_8_0_10_12
void FASTCALL call_8_0_10_12(SH2_struct* sh)
{
  SH2movbs4(sh,10,12);
}
#define init_8_0_10_13
void FASTCALL call_8_0_10_13(SH2_struct* sh)
{
  SH2movbs4(sh,10,13);
}
#define init_8_0_10_14
void FASTCALL call_8_0_10_14(SH2_struct* sh)
{
  SH2movbs4(sh,10,14);
}
#define init_8_0_10_15
void FASTCALL call_8_0_10_15(SH2_struct* sh)
{
  SH2movbs4(sh,10,15);
}
#define init_8_0_11_0
void FASTCALL call_8_0_11_0(SH2_struct* sh)
{
  SH2movbs4(sh,11,0);
}
#define init_8_0_11_1
void FASTCALL call_8_0_11_1(SH2_struct* sh)
{
  SH2movbs4(sh,11,1);
}
#define init_8_0_11_2
void FASTCALL call_8_0_11_2(SH2_struct* sh)
{
  SH2movbs4(sh,11,2);
}
#define init_8_0_11_3
void FASTCALL call_8_0_11_3(SH2_struct* sh)
{
  SH2movbs4(sh,11,3);
}
#define init_8_0_11_4
void FASTCALL call_8_0_11_4(SH2_struct* sh)
{
  SH2movbs4(sh,11,4);
}
#define init_8_0_11_5
void FASTCALL call_8_0_11_5(SH2_struct* sh)
{
  SH2movbs4(sh,11,5);
}
#define init_8_0_11_6
void FASTCALL call_8_0_11_6(SH2_struct* sh)
{
  SH2movbs4(sh,11,6);
}
#define init_8_0_11_7
void FASTCALL call_8_0_11_7(SH2_struct* sh)
{
  SH2movbs4(sh,11,7);
}
#define init_8_0_11_8
void FASTCALL call_8_0_11_8(SH2_struct* sh)
{
  SH2movbs4(sh,11,8);
}
#define init_8_0_11_9
void FASTCALL call_8_0_11_9(SH2_struct* sh)
{
  SH2movbs4(sh,11,9);
}
#define init_8_0_11_10
void FASTCALL call_8_0_11_10(SH2_struct* sh)
{
  SH2movbs4(sh,11,10);
}
#define init_8_0_11_11
void FASTCALL call_8_0_11_11(SH2_struct* sh)
{
  SH2movbs4(sh,11,11);
}
#define init_8_0_11_12
void FASTCALL call_8_0_11_12(SH2_struct* sh)
{
  SH2movbs4(sh,11,12);
}
#define init_8_0_11_13
void FASTCALL call_8_0_11_13(SH2_struct* sh)
{
  SH2movbs4(sh,11,13);
}
#define init_8_0_11_14
void FASTCALL call_8_0_11_14(SH2_struct* sh)
{
  SH2movbs4(sh,11,14);
}
#define init_8_0_11_15
void FASTCALL call_8_0_11_15(SH2_struct* sh)
{
  SH2movbs4(sh,11,15);
}
#define init_8_0_12_0
void FASTCALL call_8_0_12_0(SH2_struct* sh)
{
  SH2movbs4(sh,12,0);
}
#define init_8_0_12_1
void FASTCALL call_8_0_12_1(SH2_struct* sh)
{
  SH2movbs4(sh,12,1);
}
#define init_8_0_12_2
void FASTCALL call_8_0_12_2(SH2_struct* sh)
{
  SH2movbs4(sh,12,2);
}
#define init_8_0_12_3
void FASTCALL call_8_0_12_3(SH2_struct* sh)
{
  SH2movbs4(sh,12,3);
}
#define init_8_0_12_4
void FASTCALL call_8_0_12_4(SH2_struct* sh)
{
  SH2movbs4(sh,12,4);
}
#define init_8_0_12_5
void FASTCALL call_8_0_12_5(SH2_struct* sh)
{
  SH2movbs4(sh,12,5);
}
#define init_8_0_12_6
void FASTCALL call_8_0_12_6(SH2_struct* sh)
{
  SH2movbs4(sh,12,6);
}
#define init_8_0_12_7
void FASTCALL call_8_0_12_7(SH2_struct* sh)
{
  SH2movbs4(sh,12,7);
}
#define init_8_0_12_8
void FASTCALL call_8_0_12_8(SH2_struct* sh)
{
  SH2movbs4(sh,12,8);
}
#define init_8_0_12_9
void FASTCALL call_8_0_12_9(SH2_struct* sh)
{
  SH2movbs4(sh,12,9);
}
#define init_8_0_12_10
void FASTCALL call_8_0_12_10(SH2_struct* sh)
{
  SH2movbs4(sh,12,10);
}
#define init_8_0_12_11
void FASTCALL call_8_0_12_11(SH2_struct* sh)
{
  SH2movbs4(sh,12,11);
}
#define init_8_0_12_12
void FASTCALL call_8_0_12_12(SH2_struct* sh)
{
  SH2movbs4(sh,12,12);
}
#define init_8_0_12_13
void FASTCALL call_8_0_12_13(SH2_struct* sh)
{
  SH2movbs4(sh,12,13);
}
#define init_8_0_12_14
void FASTCALL call_8_0_12_14(SH2_struct* sh)
{
  SH2movbs4(sh,12,14);
}
#define init_8_0_12_15
void FASTCALL call_8_0_12_15(SH2_struct* sh)
{
  SH2movbs4(sh,12,15);
}
#define init_8_0_13_0
void FASTCALL call_8_0_13_0(SH2_struct* sh)
{
  SH2movbs4(sh,13,0);
}
#define init_8_0_13_1
void FASTCALL call_8_0_13_1(SH2_struct* sh)
{
  SH2movbs4(sh,13,1);
}
#define init_8_0_13_2
void FASTCALL call_8_0_13_2(SH2_struct* sh)
{
  SH2movbs4(sh,13,2);
}
#define init_8_0_13_3
void FASTCALL call_8_0_13_3(SH2_struct* sh)
{
  SH2movbs4(sh,13,3);
}
#define init_8_0_13_4
void FASTCALL call_8_0_13_4(SH2_struct* sh)
{
  SH2movbs4(sh,13,4);
}
#define init_8_0_13_5
void FASTCALL call_8_0_13_5(SH2_struct* sh)
{
  SH2movbs4(sh,13,5);
}
#define init_8_0_13_6
void FASTCALL call_8_0_13_6(SH2_struct* sh)
{
  SH2movbs4(sh,13,6);
}
#define init_8_0_13_7
void FASTCALL call_8_0_13_7(SH2_struct* sh)
{
  SH2movbs4(sh,13,7);
}
#define init_8_0_13_8
void FASTCALL call_8_0_13_8(SH2_struct* sh)
{
  SH2movbs4(sh,13,8);
}
#define init_8_0_13_9
void FASTCALL call_8_0_13_9(SH2_struct* sh)
{
  SH2movbs4(sh,13,9);
}
#define init_8_0_13_10
void FASTCALL call_8_0_13_10(SH2_struct* sh)
{
  SH2movbs4(sh,13,10);
}
#define init_8_0_13_11
void FASTCALL call_8_0_13_11(SH2_struct* sh)
{
  SH2movbs4(sh,13,11);
}
#define init_8_0_13_12
void FASTCALL call_8_0_13_12(SH2_struct* sh)
{
  SH2movbs4(sh,13,12);
}
#define init_8_0_13_13
void FASTCALL call_8_0_13_13(SH2_struct* sh)
{
  SH2movbs4(sh,13,13);
}
#define init_8_0_13_14
void FASTCALL call_8_0_13_14(SH2_struct* sh)
{
  SH2movbs4(sh,13,14);
}
#define init_8_0_13_15
void FASTCALL call_8_0_13_15(SH2_struct* sh)
{
  SH2movbs4(sh,13,15);
}
#define init_8_0_14_0
void FASTCALL call_8_0_14_0(SH2_struct* sh)
{
  SH2movbs4(sh,14,0);
}
#define init_8_0_14_1
void FASTCALL call_8_0_14_1(SH2_struct* sh)
{
  SH2movbs4(sh,14,1);
}
#define init_8_0_14_2
void FASTCALL call_8_0_14_2(SH2_struct* sh)
{
  SH2movbs4(sh,14,2);
}
#define init_8_0_14_3
void FASTCALL call_8_0_14_3(SH2_struct* sh)
{
  SH2movbs4(sh,14,3);
}
#define init_8_0_14_4
void FASTCALL call_8_0_14_4(SH2_struct* sh)
{
  SH2movbs4(sh,14,4);
}
#define init_8_0_14_5
void FASTCALL call_8_0_14_5(SH2_struct* sh)
{
  SH2movbs4(sh,14,5);
}
#define init_8_0_14_6
void FASTCALL call_8_0_14_6(SH2_struct* sh)
{
  SH2movbs4(sh,14,6);
}
#define init_8_0_14_7
void FASTCALL call_8_0_14_7(SH2_struct* sh)
{
  SH2movbs4(sh,14,7);
}
#define init_8_0_14_8
void FASTCALL call_8_0_14_8(SH2_struct* sh)
{
  SH2movbs4(sh,14,8);
}
#define init_8_0_14_9
void FASTCALL call_8_0_14_9(SH2_struct* sh)
{
  SH2movbs4(sh,14,9);
}
#define init_8_0_14_10
void FASTCALL call_8_0_14_10(SH2_struct* sh)
{
  SH2movbs4(sh,14,10);
}
#define init_8_0_14_11
void FASTCALL call_8_0_14_11(SH2_struct* sh)
{
  SH2movbs4(sh,14,11);
}
#define init_8_0_14_12
void FASTCALL call_8_0_14_12(SH2_struct* sh)
{
  SH2movbs4(sh,14,12);
}
#define init_8_0_14_13
void FASTCALL call_8_0_14_13(SH2_struct* sh)
{
  SH2movbs4(sh,14,13);
}
#define init_8_0_14_14
void FASTCALL call_8_0_14_14(SH2_struct* sh)
{
  SH2movbs4(sh,14,14);
}
#define init_8_0_14_15
void FASTCALL call_8_0_14_15(SH2_struct* sh)
{
  SH2movbs4(sh,14,15);
}
#define init_8_0_15_0
void FASTCALL call_8_0_15_0(SH2_struct* sh)
{
  SH2movbs4(sh,15,0);
}
#define init_8_0_15_1
void FASTCALL call_8_0_15_1(SH2_struct* sh)
{
  SH2movbs4(sh,15,1);
}
#define init_8_0_15_2
void FASTCALL call_8_0_15_2(SH2_struct* sh)
{
  SH2movbs4(sh,15,2);
}
#define init_8_0_15_3
void FASTCALL call_8_0_15_3(SH2_struct* sh)
{
  SH2movbs4(sh,15,3);
}
#define init_8_0_15_4
void FASTCALL call_8_0_15_4(SH2_struct* sh)
{
  SH2movbs4(sh,15,4);
}
#define init_8_0_15_5
void FASTCALL call_8_0_15_5(SH2_struct* sh)
{
  SH2movbs4(sh,15,5);
}
#define init_8_0_15_6
void FASTCALL call_8_0_15_6(SH2_struct* sh)
{
  SH2movbs4(sh,15,6);
}
#define init_8_0_15_7
void FASTCALL call_8_0_15_7(SH2_struct* sh)
{
  SH2movbs4(sh,15,7);
}
#define init_8_0_15_8
void FASTCALL call_8_0_15_8(SH2_struct* sh)
{
  SH2movbs4(sh,15,8);
}
#define init_8_0_15_9
void FASTCALL call_8_0_15_9(SH2_struct* sh)
{
  SH2movbs4(sh,15,9);
}
#define init_8_0_15_10
void FASTCALL call_8_0_15_10(SH2_struct* sh)
{
  SH2movbs4(sh,15,10);
}
#define init_8_0_15_11
void FASTCALL call_8_0_15_11(SH2_struct* sh)
{
  SH2movbs4(sh,15,11);
}
#define init_8_0_15_12
void FASTCALL call_8_0_15_12(SH2_struct* sh)
{
  SH2movbs4(sh,15,12);
}
#define init_8_0_15_13
void FASTCALL call_8_0_15_13(SH2_struct* sh)
{
  SH2movbs4(sh,15,13);
}
#define init_8_0_15_14
void FASTCALL call_8_0_15_14(SH2_struct* sh)
{
  SH2movbs4(sh,15,14);
}
#define init_8_0_15_15
void FASTCALL call_8_0_15_15(SH2_struct* sh)
{
  SH2movbs4(sh,15,15);
}
#define init_2_0_0_0
void FASTCALL call_2_0_0_0(SH2_struct* sh)
{
  SH2movbs(sh,0,0);
}
#define init_2_0_1_0
void FASTCALL call_2_0_1_0(SH2_struct* sh)
{
  SH2movbs(sh,0,1);
}
#define init_2_0_2_0
void FASTCALL call_2_0_2_0(SH2_struct* sh)
{
  SH2movbs(sh,0,2);
}
#define init_2_0_3_0
void FASTCALL call_2_0_3_0(SH2_struct* sh)
{
  SH2movbs(sh,0,3);
}
#define init_2_0_4_0
void FASTCALL call_2_0_4_0(SH2_struct* sh)
{
  SH2movbs(sh,0,4);
}
#define init_2_0_5_0
void FASTCALL call_2_0_5_0(SH2_struct* sh)
{
  SH2movbs(sh,0,5);
}
#define init_2_0_6_0
void FASTCALL call_2_0_6_0(SH2_struct* sh)
{
  SH2movbs(sh,0,6);
}
#define init_2_0_7_0
void FASTCALL call_2_0_7_0(SH2_struct* sh)
{
  SH2movbs(sh,0,7);
}
#define init_2_0_8_0
void FASTCALL call_2_0_8_0(SH2_struct* sh)
{
  SH2movbs(sh,0,8);
}
#define init_2_0_9_0
void FASTCALL call_2_0_9_0(SH2_struct* sh)
{
  SH2movbs(sh,0,9);
}
#define init_2_0_10_0
void FASTCALL call_2_0_10_0(SH2_struct* sh)
{
  SH2movbs(sh,0,10);
}
#define init_2_0_11_0
void FASTCALL call_2_0_11_0(SH2_struct* sh)
{
  SH2movbs(sh,0,11);
}
#define init_2_0_12_0
void FASTCALL call_2_0_12_0(SH2_struct* sh)
{
  SH2movbs(sh,0,12);
}
#define init_2_0_13_0
void FASTCALL call_2_0_13_0(SH2_struct* sh)
{
  SH2movbs(sh,0,13);
}
#define init_2_0_14_0
void FASTCALL call_2_0_14_0(SH2_struct* sh)
{
  SH2movbs(sh,0,14);
}
#define init_2_0_15_0
void FASTCALL call_2_0_15_0(SH2_struct* sh)
{
  SH2movbs(sh,0,15);
}
#define init_2_1_0_0
void FASTCALL call_2_1_0_0(SH2_struct* sh)
{
  SH2movbs(sh,1,0);
}
#define init_2_1_1_0
void FASTCALL call_2_1_1_0(SH2_struct* sh)
{
  SH2movbs(sh,1,1);
}
#define init_2_1_2_0
void FASTCALL call_2_1_2_0(SH2_struct* sh)
{
  SH2movbs(sh,1,2);
}
#define init_2_1_3_0
void FASTCALL call_2_1_3_0(SH2_struct* sh)
{
  SH2movbs(sh,1,3);
}
#define init_2_1_4_0
void FASTCALL call_2_1_4_0(SH2_struct* sh)
{
  SH2movbs(sh,1,4);
}
#define init_2_1_5_0
void FASTCALL call_2_1_5_0(SH2_struct* sh)
{
  SH2movbs(sh,1,5);
}
#define init_2_1_6_0
void FASTCALL call_2_1_6_0(SH2_struct* sh)
{
  SH2movbs(sh,1,6);
}
#define init_2_1_7_0
void FASTCALL call_2_1_7_0(SH2_struct* sh)
{
  SH2movbs(sh,1,7);
}
#define init_2_1_8_0
void FASTCALL call_2_1_8_0(SH2_struct* sh)
{
  SH2movbs(sh,1,8);
}
#define init_2_1_9_0
void FASTCALL call_2_1_9_0(SH2_struct* sh)
{
  SH2movbs(sh,1,9);
}
#define init_2_1_10_0
void FASTCALL call_2_1_10_0(SH2_struct* sh)
{
  SH2movbs(sh,1,10);
}
#define init_2_1_11_0
void FASTCALL call_2_1_11_0(SH2_struct* sh)
{
  SH2movbs(sh,1,11);
}
#define init_2_1_12_0
void FASTCALL call_2_1_12_0(SH2_struct* sh)
{
  SH2movbs(sh,1,12);
}
#define init_2_1_13_0
void FASTCALL call_2_1_13_0(SH2_struct* sh)
{
  SH2movbs(sh,1,13);
}
#define init_2_1_14_0
void FASTCALL call_2_1_14_0(SH2_struct* sh)
{
  SH2movbs(sh,1,14);
}
#define init_2_1_15_0
void FASTCALL call_2_1_15_0(SH2_struct* sh)
{
  SH2movbs(sh,1,15);
}
#define init_2_2_0_0
void FASTCALL call_2_2_0_0(SH2_struct* sh)
{
  SH2movbs(sh,2,0);
}
#define init_2_2_1_0
void FASTCALL call_2_2_1_0(SH2_struct* sh)
{
  SH2movbs(sh,2,1);
}
#define init_2_2_2_0
void FASTCALL call_2_2_2_0(SH2_struct* sh)
{
  SH2movbs(sh,2,2);
}
#define init_2_2_3_0
void FASTCALL call_2_2_3_0(SH2_struct* sh)
{
  SH2movbs(sh,2,3);
}
#define init_2_2_4_0
void FASTCALL call_2_2_4_0(SH2_struct* sh)
{
  SH2movbs(sh,2,4);
}
#define init_2_2_5_0
void FASTCALL call_2_2_5_0(SH2_struct* sh)
{
  SH2movbs(sh,2,5);
}
#define init_2_2_6_0
void FASTCALL call_2_2_6_0(SH2_struct* sh)
{
  SH2movbs(sh,2,6);
}
#define init_2_2_7_0
void FASTCALL call_2_2_7_0(SH2_struct* sh)
{
  SH2movbs(sh,2,7);
}
#define init_2_2_8_0
void FASTCALL call_2_2_8_0(SH2_struct* sh)
{
  SH2movbs(sh,2,8);
}
#define init_2_2_9_0
void FASTCALL call_2_2_9_0(SH2_struct* sh)
{
  SH2movbs(sh,2,9);
}
#define init_2_2_10_0
void FASTCALL call_2_2_10_0(SH2_struct* sh)
{
  SH2movbs(sh,2,10);
}
#define init_2_2_11_0
void FASTCALL call_2_2_11_0(SH2_struct* sh)
{
  SH2movbs(sh,2,11);
}
#define init_2_2_12_0
void FASTCALL call_2_2_12_0(SH2_struct* sh)
{
  SH2movbs(sh,2,12);
}
#define init_2_2_13_0
void FASTCALL call_2_2_13_0(SH2_struct* sh)
{
  SH2movbs(sh,2,13);
}
#define init_2_2_14_0
void FASTCALL call_2_2_14_0(SH2_struct* sh)
{
  SH2movbs(sh,2,14);
}
#define init_2_2_15_0
void FASTCALL call_2_2_15_0(SH2_struct* sh)
{
  SH2movbs(sh,2,15);
}
#define init_2_3_0_0
void FASTCALL call_2_3_0_0(SH2_struct* sh)
{
  SH2movbs(sh,3,0);
}
#define init_2_3_1_0
void FASTCALL call_2_3_1_0(SH2_struct* sh)
{
  SH2movbs(sh,3,1);
}
#define init_2_3_2_0
void FASTCALL call_2_3_2_0(SH2_struct* sh)
{
  SH2movbs(sh,3,2);
}
#define init_2_3_3_0
void FASTCALL call_2_3_3_0(SH2_struct* sh)
{
  SH2movbs(sh,3,3);
}
#define init_2_3_4_0
void FASTCALL call_2_3_4_0(SH2_struct* sh)
{
  SH2movbs(sh,3,4);
}
#define init_2_3_5_0
void FASTCALL call_2_3_5_0(SH2_struct* sh)
{
  SH2movbs(sh,3,5);
}
#define init_2_3_6_0
void FASTCALL call_2_3_6_0(SH2_struct* sh)
{
  SH2movbs(sh,3,6);
}
#define init_2_3_7_0
void FASTCALL call_2_3_7_0(SH2_struct* sh)
{
  SH2movbs(sh,3,7);
}
#define init_2_3_8_0
void FASTCALL call_2_3_8_0(SH2_struct* sh)
{
  SH2movbs(sh,3,8);
}
#define init_2_3_9_0
void FASTCALL call_2_3_9_0(SH2_struct* sh)
{
  SH2movbs(sh,3,9);
}
#define init_2_3_10_0
void FASTCALL call_2_3_10_0(SH2_struct* sh)
{
  SH2movbs(sh,3,10);
}
#define init_2_3_11_0
void FASTCALL call_2_3_11_0(SH2_struct* sh)
{
  SH2movbs(sh,3,11);
}
#define init_2_3_12_0
void FASTCALL call_2_3_12_0(SH2_struct* sh)
{
  SH2movbs(sh,3,12);
}
#define init_2_3_13_0
void FASTCALL call_2_3_13_0(SH2_struct* sh)
{
  SH2movbs(sh,3,13);
}
#define init_2_3_14_0
void FASTCALL call_2_3_14_0(SH2_struct* sh)
{
  SH2movbs(sh,3,14);
}
#define init_2_3_15_0
void FASTCALL call_2_3_15_0(SH2_struct* sh)
{
  SH2movbs(sh,3,15);
}
#define init_2_4_0_0
void FASTCALL call_2_4_0_0(SH2_struct* sh)
{
  SH2movbs(sh,4,0);
}
#define init_2_4_1_0
void FASTCALL call_2_4_1_0(SH2_struct* sh)
{
  SH2movbs(sh,4,1);
}
#define init_2_4_2_0
void FASTCALL call_2_4_2_0(SH2_struct* sh)
{
  SH2movbs(sh,4,2);
}
#define init_2_4_3_0
void FASTCALL call_2_4_3_0(SH2_struct* sh)
{
  SH2movbs(sh,4,3);
}
#define init_2_4_4_0
void FASTCALL call_2_4_4_0(SH2_struct* sh)
{
  SH2movbs(sh,4,4);
}
#define init_2_4_5_0
void FASTCALL call_2_4_5_0(SH2_struct* sh)
{
  SH2movbs(sh,4,5);
}
#define init_2_4_6_0
void FASTCALL call_2_4_6_0(SH2_struct* sh)
{
  SH2movbs(sh,4,6);
}
#define init_2_4_7_0
void FASTCALL call_2_4_7_0(SH2_struct* sh)
{
  SH2movbs(sh,4,7);
}
#define init_2_4_8_0
void FASTCALL call_2_4_8_0(SH2_struct* sh)
{
  SH2movbs(sh,4,8);
}
#define init_2_4_9_0
void FASTCALL call_2_4_9_0(SH2_struct* sh)
{
  SH2movbs(sh,4,9);
}
#define init_2_4_10_0
void FASTCALL call_2_4_10_0(SH2_struct* sh)
{
  SH2movbs(sh,4,10);
}
#define init_2_4_11_0
void FASTCALL call_2_4_11_0(SH2_struct* sh)
{
  SH2movbs(sh,4,11);
}
#define init_2_4_12_0
void FASTCALL call_2_4_12_0(SH2_struct* sh)
{
  SH2movbs(sh,4,12);
}
#define init_2_4_13_0
void FASTCALL call_2_4_13_0(SH2_struct* sh)
{
  SH2movbs(sh,4,13);
}
#define init_2_4_14_0
void FASTCALL call_2_4_14_0(SH2_struct* sh)
{
  SH2movbs(sh,4,14);
}
#define init_2_4_15_0
void FASTCALL call_2_4_15_0(SH2_struct* sh)
{
  SH2movbs(sh,4,15);
}
#define init_2_5_0_0
void FASTCALL call_2_5_0_0(SH2_struct* sh)
{
  SH2movbs(sh,5,0);
}
#define init_2_5_1_0
void FASTCALL call_2_5_1_0(SH2_struct* sh)
{
  SH2movbs(sh,5,1);
}
#define init_2_5_2_0
void FASTCALL call_2_5_2_0(SH2_struct* sh)
{
  SH2movbs(sh,5,2);
}
#define init_2_5_3_0
void FASTCALL call_2_5_3_0(SH2_struct* sh)
{
  SH2movbs(sh,5,3);
}
#define init_2_5_4_0
void FASTCALL call_2_5_4_0(SH2_struct* sh)
{
  SH2movbs(sh,5,4);
}
#define init_2_5_5_0
void FASTCALL call_2_5_5_0(SH2_struct* sh)
{
  SH2movbs(sh,5,5);
}
#define init_2_5_6_0
void FASTCALL call_2_5_6_0(SH2_struct* sh)
{
  SH2movbs(sh,5,6);
}
#define init_2_5_7_0
void FASTCALL call_2_5_7_0(SH2_struct* sh)
{
  SH2movbs(sh,5,7);
}
#define init_2_5_8_0
void FASTCALL call_2_5_8_0(SH2_struct* sh)
{
  SH2movbs(sh,5,8);
}
#define init_2_5_9_0
void FASTCALL call_2_5_9_0(SH2_struct* sh)
{
  SH2movbs(sh,5,9);
}
#define init_2_5_10_0
void FASTCALL call_2_5_10_0(SH2_struct* sh)
{
  SH2movbs(sh,5,10);
}
#define init_2_5_11_0
void FASTCALL call_2_5_11_0(SH2_struct* sh)
{
  SH2movbs(sh,5,11);
}
#define init_2_5_12_0
void FASTCALL call_2_5_12_0(SH2_struct* sh)
{
  SH2movbs(sh,5,12);
}
#define init_2_5_13_0
void FASTCALL call_2_5_13_0(SH2_struct* sh)
{
  SH2movbs(sh,5,13);
}
#define init_2_5_14_0
void FASTCALL call_2_5_14_0(SH2_struct* sh)
{
  SH2movbs(sh,5,14);
}
#define init_2_5_15_0
void FASTCALL call_2_5_15_0(SH2_struct* sh)
{
  SH2movbs(sh,5,15);
}
#define init_2_6_0_0
void FASTCALL call_2_6_0_0(SH2_struct* sh)
{
  SH2movbs(sh,6,0);
}
#define init_2_6_1_0
void FASTCALL call_2_6_1_0(SH2_struct* sh)
{
  SH2movbs(sh,6,1);
}
#define init_2_6_2_0
void FASTCALL call_2_6_2_0(SH2_struct* sh)
{
  SH2movbs(sh,6,2);
}
#define init_2_6_3_0
void FASTCALL call_2_6_3_0(SH2_struct* sh)
{
  SH2movbs(sh,6,3);
}
#define init_2_6_4_0
void FASTCALL call_2_6_4_0(SH2_struct* sh)
{
  SH2movbs(sh,6,4);
}
#define init_2_6_5_0
void FASTCALL call_2_6_5_0(SH2_struct* sh)
{
  SH2movbs(sh,6,5);
}
#define init_2_6_6_0
void FASTCALL call_2_6_6_0(SH2_struct* sh)
{
  SH2movbs(sh,6,6);
}
#define init_2_6_7_0
void FASTCALL call_2_6_7_0(SH2_struct* sh)
{
  SH2movbs(sh,6,7);
}
#define init_2_6_8_0
void FASTCALL call_2_6_8_0(SH2_struct* sh)
{
  SH2movbs(sh,6,8);
}
#define init_2_6_9_0
void FASTCALL call_2_6_9_0(SH2_struct* sh)
{
  SH2movbs(sh,6,9);
}
#define init_2_6_10_0
void FASTCALL call_2_6_10_0(SH2_struct* sh)
{
  SH2movbs(sh,6,10);
}
#define init_2_6_11_0
void FASTCALL call_2_6_11_0(SH2_struct* sh)
{
  SH2movbs(sh,6,11);
}
#define init_2_6_12_0
void FASTCALL call_2_6_12_0(SH2_struct* sh)
{
  SH2movbs(sh,6,12);
}
#define init_2_6_13_0
void FASTCALL call_2_6_13_0(SH2_struct* sh)
{
  SH2movbs(sh,6,13);
}
#define init_2_6_14_0
void FASTCALL call_2_6_14_0(SH2_struct* sh)
{
  SH2movbs(sh,6,14);
}
#define init_2_6_15_0
void FASTCALL call_2_6_15_0(SH2_struct* sh)
{
  SH2movbs(sh,6,15);
}
#define init_2_7_0_0
void FASTCALL call_2_7_0_0(SH2_struct* sh)
{
  SH2movbs(sh,7,0);
}
#define init_2_7_1_0
void FASTCALL call_2_7_1_0(SH2_struct* sh)
{
  SH2movbs(sh,7,1);
}
#define init_2_7_2_0
void FASTCALL call_2_7_2_0(SH2_struct* sh)
{
  SH2movbs(sh,7,2);
}
#define init_2_7_3_0
void FASTCALL call_2_7_3_0(SH2_struct* sh)
{
  SH2movbs(sh,7,3);
}
#define init_2_7_4_0
void FASTCALL call_2_7_4_0(SH2_struct* sh)
{
  SH2movbs(sh,7,4);
}
#define init_2_7_5_0
void FASTCALL call_2_7_5_0(SH2_struct* sh)
{
  SH2movbs(sh,7,5);
}
#define init_2_7_6_0
void FASTCALL call_2_7_6_0(SH2_struct* sh)
{
  SH2movbs(sh,7,6);
}
#define init_2_7_7_0
void FASTCALL call_2_7_7_0(SH2_struct* sh)
{
  SH2movbs(sh,7,7);
}
#define init_2_7_8_0
void FASTCALL call_2_7_8_0(SH2_struct* sh)
{
  SH2movbs(sh,7,8);
}
#define init_2_7_9_0
void FASTCALL call_2_7_9_0(SH2_struct* sh)
{
  SH2movbs(sh,7,9);
}
#define init_2_7_10_0
void FASTCALL call_2_7_10_0(SH2_struct* sh)
{
  SH2movbs(sh,7,10);
}
#define init_2_7_11_0
void FASTCALL call_2_7_11_0(SH2_struct* sh)
{
  SH2movbs(sh,7,11);
}
#define init_2_7_12_0
void FASTCALL call_2_7_12_0(SH2_struct* sh)
{
  SH2movbs(sh,7,12);
}
#define init_2_7_13_0
void FASTCALL call_2_7_13_0(SH2_struct* sh)
{
  SH2movbs(sh,7,13);
}
#define init_2_7_14_0
void FASTCALL call_2_7_14_0(SH2_struct* sh)
{
  SH2movbs(sh,7,14);
}
#define init_2_7_15_0
void FASTCALL call_2_7_15_0(SH2_struct* sh)
{
  SH2movbs(sh,7,15);
}
#define init_2_8_0_0
void FASTCALL call_2_8_0_0(SH2_struct* sh)
{
  SH2movbs(sh,8,0);
}
#define init_2_8_1_0
void FASTCALL call_2_8_1_0(SH2_struct* sh)
{
  SH2movbs(sh,8,1);
}
#define init_2_8_2_0
void FASTCALL call_2_8_2_0(SH2_struct* sh)
{
  SH2movbs(sh,8,2);
}
#define init_2_8_3_0
void FASTCALL call_2_8_3_0(SH2_struct* sh)
{
  SH2movbs(sh,8,3);
}
#define init_2_8_4_0
void FASTCALL call_2_8_4_0(SH2_struct* sh)
{
  SH2movbs(sh,8,4);
}
#define init_2_8_5_0
void FASTCALL call_2_8_5_0(SH2_struct* sh)
{
  SH2movbs(sh,8,5);
}
#define init_2_8_6_0
void FASTCALL call_2_8_6_0(SH2_struct* sh)
{
  SH2movbs(sh,8,6);
}
#define init_2_8_7_0
void FASTCALL call_2_8_7_0(SH2_struct* sh)
{
  SH2movbs(sh,8,7);
}
#define init_2_8_8_0
void FASTCALL call_2_8_8_0(SH2_struct* sh)
{
  SH2movbs(sh,8,8);
}
#define init_2_8_9_0
void FASTCALL call_2_8_9_0(SH2_struct* sh)
{
  SH2movbs(sh,8,9);
}
#define init_2_8_10_0
void FASTCALL call_2_8_10_0(SH2_struct* sh)
{
  SH2movbs(sh,8,10);
}
#define init_2_8_11_0
void FASTCALL call_2_8_11_0(SH2_struct* sh)
{
  SH2movbs(sh,8,11);
}
#define init_2_8_12_0
void FASTCALL call_2_8_12_0(SH2_struct* sh)
{
  SH2movbs(sh,8,12);
}
#define init_2_8_13_0
void FASTCALL call_2_8_13_0(SH2_struct* sh)
{
  SH2movbs(sh,8,13);
}
#define init_2_8_14_0
void FASTCALL call_2_8_14_0(SH2_struct* sh)
{
  SH2movbs(sh,8,14);
}
#define init_2_8_15_0
void FASTCALL call_2_8_15_0(SH2_struct* sh)
{
  SH2movbs(sh,8,15);
}
#define init_2_9_0_0
void FASTCALL call_2_9_0_0(SH2_struct* sh)
{
  SH2movbs(sh,9,0);
}
#define init_2_9_1_0
void FASTCALL call_2_9_1_0(SH2_struct* sh)
{
  SH2movbs(sh,9,1);
}
#define init_2_9_2_0
void FASTCALL call_2_9_2_0(SH2_struct* sh)
{
  SH2movbs(sh,9,2);
}
#define init_2_9_3_0
void FASTCALL call_2_9_3_0(SH2_struct* sh)
{
  SH2movbs(sh,9,3);
}
#define init_2_9_4_0
void FASTCALL call_2_9_4_0(SH2_struct* sh)
{
  SH2movbs(sh,9,4);
}
#define init_2_9_5_0
void FASTCALL call_2_9_5_0(SH2_struct* sh)
{
  SH2movbs(sh,9,5);
}
#define init_2_9_6_0
void FASTCALL call_2_9_6_0(SH2_struct* sh)
{
  SH2movbs(sh,9,6);
}
#define init_2_9_7_0
void FASTCALL call_2_9_7_0(SH2_struct* sh)
{
  SH2movbs(sh,9,7);
}
#define init_2_9_8_0
void FASTCALL call_2_9_8_0(SH2_struct* sh)
{
  SH2movbs(sh,9,8);
}
#define init_2_9_9_0
void FASTCALL call_2_9_9_0(SH2_struct* sh)
{
  SH2movbs(sh,9,9);
}
#define init_2_9_10_0
void FASTCALL call_2_9_10_0(SH2_struct* sh)
{
  SH2movbs(sh,9,10);
}
#define init_2_9_11_0
void FASTCALL call_2_9_11_0(SH2_struct* sh)
{
  SH2movbs(sh,9,11);
}
#define init_2_9_12_0
void FASTCALL call_2_9_12_0(SH2_struct* sh)
{
  SH2movbs(sh,9,12);
}
#define init_2_9_13_0
void FASTCALL call_2_9_13_0(SH2_struct* sh)
{
  SH2movbs(sh,9,13);
}
#define init_2_9_14_0
void FASTCALL call_2_9_14_0(SH2_struct* sh)
{
  SH2movbs(sh,9,14);
}
#define init_2_9_15_0
void FASTCALL call_2_9_15_0(SH2_struct* sh)
{
  SH2movbs(sh,9,15);
}
#define init_2_10_0_0
void FASTCALL call_2_10_0_0(SH2_struct* sh)
{
  SH2movbs(sh,10,0);
}
#define init_2_10_1_0
void FASTCALL call_2_10_1_0(SH2_struct* sh)
{
  SH2movbs(sh,10,1);
}
#define init_2_10_2_0
void FASTCALL call_2_10_2_0(SH2_struct* sh)
{
  SH2movbs(sh,10,2);
}
#define init_2_10_3_0
void FASTCALL call_2_10_3_0(SH2_struct* sh)
{
  SH2movbs(sh,10,3);
}
#define init_2_10_4_0
void FASTCALL call_2_10_4_0(SH2_struct* sh)
{
  SH2movbs(sh,10,4);
}
#define init_2_10_5_0
void FASTCALL call_2_10_5_0(SH2_struct* sh)
{
  SH2movbs(sh,10,5);
}
#define init_2_10_6_0
void FASTCALL call_2_10_6_0(SH2_struct* sh)
{
  SH2movbs(sh,10,6);
}
#define init_2_10_7_0
void FASTCALL call_2_10_7_0(SH2_struct* sh)
{
  SH2movbs(sh,10,7);
}
#define init_2_10_8_0
void FASTCALL call_2_10_8_0(SH2_struct* sh)
{
  SH2movbs(sh,10,8);
}
#define init_2_10_9_0
void FASTCALL call_2_10_9_0(SH2_struct* sh)
{
  SH2movbs(sh,10,9);
}
#define init_2_10_10_0
void FASTCALL call_2_10_10_0(SH2_struct* sh)
{
  SH2movbs(sh,10,10);
}
#define init_2_10_11_0
void FASTCALL call_2_10_11_0(SH2_struct* sh)
{
  SH2movbs(sh,10,11);
}
#define init_2_10_12_0
void FASTCALL call_2_10_12_0(SH2_struct* sh)
{
  SH2movbs(sh,10,12);
}
#define init_2_10_13_0
void FASTCALL call_2_10_13_0(SH2_struct* sh)
{
  SH2movbs(sh,10,13);
}
#define init_2_10_14_0
void FASTCALL call_2_10_14_0(SH2_struct* sh)
{
  SH2movbs(sh,10,14);
}
#define init_2_10_15_0
void FASTCALL call_2_10_15_0(SH2_struct* sh)
{
  SH2movbs(sh,10,15);
}
#define init_2_11_0_0
void FASTCALL call_2_11_0_0(SH2_struct* sh)
{
  SH2movbs(sh,11,0);
}
#define init_2_11_1_0
void FASTCALL call_2_11_1_0(SH2_struct* sh)
{
  SH2movbs(sh,11,1);
}
#define init_2_11_2_0
void FASTCALL call_2_11_2_0(SH2_struct* sh)
{
  SH2movbs(sh,11,2);
}
#define init_2_11_3_0
void FASTCALL call_2_11_3_0(SH2_struct* sh)
{
  SH2movbs(sh,11,3);
}
#define init_2_11_4_0
void FASTCALL call_2_11_4_0(SH2_struct* sh)
{
  SH2movbs(sh,11,4);
}
#define init_2_11_5_0
void FASTCALL call_2_11_5_0(SH2_struct* sh)
{
  SH2movbs(sh,11,5);
}
#define init_2_11_6_0
void FASTCALL call_2_11_6_0(SH2_struct* sh)
{
  SH2movbs(sh,11,6);
}
#define init_2_11_7_0
void FASTCALL call_2_11_7_0(SH2_struct* sh)
{
  SH2movbs(sh,11,7);
}
#define init_2_11_8_0
void FASTCALL call_2_11_8_0(SH2_struct* sh)
{
  SH2movbs(sh,11,8);
}
#define init_2_11_9_0
void FASTCALL call_2_11_9_0(SH2_struct* sh)
{
  SH2movbs(sh,11,9);
}
#define init_2_11_10_0
void FASTCALL call_2_11_10_0(SH2_struct* sh)
{
  SH2movbs(sh,11,10);
}
#define init_2_11_11_0
void FASTCALL call_2_11_11_0(SH2_struct* sh)
{
  SH2movbs(sh,11,11);
}
#define init_2_11_12_0
void FASTCALL call_2_11_12_0(SH2_struct* sh)
{
  SH2movbs(sh,11,12);
}
#define init_2_11_13_0
void FASTCALL call_2_11_13_0(SH2_struct* sh)
{
  SH2movbs(sh,11,13);
}
#define init_2_11_14_0
void FASTCALL call_2_11_14_0(SH2_struct* sh)
{
  SH2movbs(sh,11,14);
}
#define init_2_11_15_0
void FASTCALL call_2_11_15_0(SH2_struct* sh)
{
  SH2movbs(sh,11,15);
}
#define init_2_12_0_0
void FASTCALL call_2_12_0_0(SH2_struct* sh)
{
  SH2movbs(sh,12,0);
}
#define init_2_12_1_0
void FASTCALL call_2_12_1_0(SH2_struct* sh)
{
  SH2movbs(sh,12,1);
}
#define init_2_12_2_0
void FASTCALL call_2_12_2_0(SH2_struct* sh)
{
  SH2movbs(sh,12,2);
}
#define init_2_12_3_0
void FASTCALL call_2_12_3_0(SH2_struct* sh)
{
  SH2movbs(sh,12,3);
}
#define init_2_12_4_0
void FASTCALL call_2_12_4_0(SH2_struct* sh)
{
  SH2movbs(sh,12,4);
}
#define init_2_12_5_0
void FASTCALL call_2_12_5_0(SH2_struct* sh)
{
  SH2movbs(sh,12,5);
}
#define init_2_12_6_0
void FASTCALL call_2_12_6_0(SH2_struct* sh)
{
  SH2movbs(sh,12,6);
}
#define init_2_12_7_0
void FASTCALL call_2_12_7_0(SH2_struct* sh)
{
  SH2movbs(sh,12,7);
}
#define init_2_12_8_0
void FASTCALL call_2_12_8_0(SH2_struct* sh)
{
  SH2movbs(sh,12,8);
}
#define init_2_12_9_0
void FASTCALL call_2_12_9_0(SH2_struct* sh)
{
  SH2movbs(sh,12,9);
}
#define init_2_12_10_0
void FASTCALL call_2_12_10_0(SH2_struct* sh)
{
  SH2movbs(sh,12,10);
}
#define init_2_12_11_0
void FASTCALL call_2_12_11_0(SH2_struct* sh)
{
  SH2movbs(sh,12,11);
}
#define init_2_12_12_0
void FASTCALL call_2_12_12_0(SH2_struct* sh)
{
  SH2movbs(sh,12,12);
}
#define init_2_12_13_0
void FASTCALL call_2_12_13_0(SH2_struct* sh)
{
  SH2movbs(sh,12,13);
}
#define init_2_12_14_0
void FASTCALL call_2_12_14_0(SH2_struct* sh)
{
  SH2movbs(sh,12,14);
}
#define init_2_12_15_0
void FASTCALL call_2_12_15_0(SH2_struct* sh)
{
  SH2movbs(sh,12,15);
}
#define init_2_13_0_0
void FASTCALL call_2_13_0_0(SH2_struct* sh)
{
  SH2movbs(sh,13,0);
}
#define init_2_13_1_0
void FASTCALL call_2_13_1_0(SH2_struct* sh)
{
  SH2movbs(sh,13,1);
}
#define init_2_13_2_0
void FASTCALL call_2_13_2_0(SH2_struct* sh)
{
  SH2movbs(sh,13,2);
}
#define init_2_13_3_0
void FASTCALL call_2_13_3_0(SH2_struct* sh)
{
  SH2movbs(sh,13,3);
}
#define init_2_13_4_0
void FASTCALL call_2_13_4_0(SH2_struct* sh)
{
  SH2movbs(sh,13,4);
}
#define init_2_13_5_0
void FASTCALL call_2_13_5_0(SH2_struct* sh)
{
  SH2movbs(sh,13,5);
}
#define init_2_13_6_0
void FASTCALL call_2_13_6_0(SH2_struct* sh)
{
  SH2movbs(sh,13,6);
}
#define init_2_13_7_0
void FASTCALL call_2_13_7_0(SH2_struct* sh)
{
  SH2movbs(sh,13,7);
}
#define init_2_13_8_0
void FASTCALL call_2_13_8_0(SH2_struct* sh)
{
  SH2movbs(sh,13,8);
}
#define init_2_13_9_0
void FASTCALL call_2_13_9_0(SH2_struct* sh)
{
  SH2movbs(sh,13,9);
}
#define init_2_13_10_0
void FASTCALL call_2_13_10_0(SH2_struct* sh)
{
  SH2movbs(sh,13,10);
}
#define init_2_13_11_0
void FASTCALL call_2_13_11_0(SH2_struct* sh)
{
  SH2movbs(sh,13,11);
}
#define init_2_13_12_0
void FASTCALL call_2_13_12_0(SH2_struct* sh)
{
  SH2movbs(sh,13,12);
}
#define init_2_13_13_0
void FASTCALL call_2_13_13_0(SH2_struct* sh)
{
  SH2movbs(sh,13,13);
}
#define init_2_13_14_0
void FASTCALL call_2_13_14_0(SH2_struct* sh)
{
  SH2movbs(sh,13,14);
}
#define init_2_13_15_0
void FASTCALL call_2_13_15_0(SH2_struct* sh)
{
  SH2movbs(sh,13,15);
}
#define init_2_14_0_0
void FASTCALL call_2_14_0_0(SH2_struct* sh)
{
  SH2movbs(sh,14,0);
}
#define init_2_14_1_0
void FASTCALL call_2_14_1_0(SH2_struct* sh)
{
  SH2movbs(sh,14,1);
}
#define init_2_14_2_0
void FASTCALL call_2_14_2_0(SH2_struct* sh)
{
  SH2movbs(sh,14,2);
}
#define init_2_14_3_0
void FASTCALL call_2_14_3_0(SH2_struct* sh)
{
  SH2movbs(sh,14,3);
}
#define init_2_14_4_0
void FASTCALL call_2_14_4_0(SH2_struct* sh)
{
  SH2movbs(sh,14,4);
}
#define init_2_14_5_0
void FASTCALL call_2_14_5_0(SH2_struct* sh)
{
  SH2movbs(sh,14,5);
}
#define init_2_14_6_0
void FASTCALL call_2_14_6_0(SH2_struct* sh)
{
  SH2movbs(sh,14,6);
}
#define init_2_14_7_0
void FASTCALL call_2_14_7_0(SH2_struct* sh)
{
  SH2movbs(sh,14,7);
}
#define init_2_14_8_0
void FASTCALL call_2_14_8_0(SH2_struct* sh)
{
  SH2movbs(sh,14,8);
}
#define init_2_14_9_0
void FASTCALL call_2_14_9_0(SH2_struct* sh)
{
  SH2movbs(sh,14,9);
}
#define init_2_14_10_0
void FASTCALL call_2_14_10_0(SH2_struct* sh)
{
  SH2movbs(sh,14,10);
}
#define init_2_14_11_0
void FASTCALL call_2_14_11_0(SH2_struct* sh)
{
  SH2movbs(sh,14,11);
}
#define init_2_14_12_0
void FASTCALL call_2_14_12_0(SH2_struct* sh)
{
  SH2movbs(sh,14,12);
}
#define init_2_14_13_0
void FASTCALL call_2_14_13_0(SH2_struct* sh)
{
  SH2movbs(sh,14,13);
}
#define init_2_14_14_0
void FASTCALL call_2_14_14_0(SH2_struct* sh)
{
  SH2movbs(sh,14,14);
}
#define init_2_14_15_0
void FASTCALL call_2_14_15_0(SH2_struct* sh)
{
  SH2movbs(sh,14,15);
}
#define init_2_15_0_0
void FASTCALL call_2_15_0_0(SH2_struct* sh)
{
  SH2movbs(sh,15,0);
}
#define init_2_15_1_0
void FASTCALL call_2_15_1_0(SH2_struct* sh)
{
  SH2movbs(sh,15,1);
}
#define init_2_15_2_0
void FASTCALL call_2_15_2_0(SH2_struct* sh)
{
  SH2movbs(sh,15,2);
}
#define init_2_15_3_0
void FASTCALL call_2_15_3_0(SH2_struct* sh)
{
  SH2movbs(sh,15,3);
}
#define init_2_15_4_0
void FASTCALL call_2_15_4_0(SH2_struct* sh)
{
  SH2movbs(sh,15,4);
}
#define init_2_15_5_0
void FASTCALL call_2_15_5_0(SH2_struct* sh)
{
  SH2movbs(sh,15,5);
}
#define init_2_15_6_0
void FASTCALL call_2_15_6_0(SH2_struct* sh)
{
  SH2movbs(sh,15,6);
}
#define init_2_15_7_0
void FASTCALL call_2_15_7_0(SH2_struct* sh)
{
  SH2movbs(sh,15,7);
}
#define init_2_15_8_0
void FASTCALL call_2_15_8_0(SH2_struct* sh)
{
  SH2movbs(sh,15,8);
}
#define init_2_15_9_0
void FASTCALL call_2_15_9_0(SH2_struct* sh)
{
  SH2movbs(sh,15,9);
}
#define init_2_15_10_0
void FASTCALL call_2_15_10_0(SH2_struct* sh)
{
  SH2movbs(sh,15,10);
}
#define init_2_15_11_0
void FASTCALL call_2_15_11_0(SH2_struct* sh)
{
  SH2movbs(sh,15,11);
}
#define init_2_15_12_0
void FASTCALL call_2_15_12_0(SH2_struct* sh)
{
  SH2movbs(sh,15,12);
}
#define init_2_15_13_0
void FASTCALL call_2_15_13_0(SH2_struct* sh)
{
  SH2movbs(sh,15,13);
}
#define init_2_15_14_0
void FASTCALL call_2_15_14_0(SH2_struct* sh)
{
  SH2movbs(sh,15,14);
}
#define init_2_15_15_0
void FASTCALL call_2_15_15_0(SH2_struct* sh)
{
  SH2movbs(sh,15,15);
}
#define init_6_0_0_4
void FASTCALL call_6_0_0_4(SH2_struct* sh)
{
  SH2movbp(sh,0,0);
}
#define init_6_0_1_4
void FASTCALL call_6_0_1_4(SH2_struct* sh)
{
  SH2movbp(sh,0,1);
}
#define init_6_0_2_4
void FASTCALL call_6_0_2_4(SH2_struct* sh)
{
  SH2movbp(sh,0,2);
}
#define init_6_0_3_4
void FASTCALL call_6_0_3_4(SH2_struct* sh)
{
  SH2movbp(sh,0,3);
}
#define init_6_0_4_4
void FASTCALL call_6_0_4_4(SH2_struct* sh)
{
  SH2movbp(sh,0,4);
}
#define init_6_0_5_4
void FASTCALL call_6_0_5_4(SH2_struct* sh)
{
  SH2movbp(sh,0,5);
}
#define init_6_0_6_4
void FASTCALL call_6_0_6_4(SH2_struct* sh)
{
  SH2movbp(sh,0,6);
}
#define init_6_0_7_4
void FASTCALL call_6_0_7_4(SH2_struct* sh)
{
  SH2movbp(sh,0,7);
}
#define init_6_0_8_4
void FASTCALL call_6_0_8_4(SH2_struct* sh)
{
  SH2movbp(sh,0,8);
}
#define init_6_0_9_4
void FASTCALL call_6_0_9_4(SH2_struct* sh)
{
  SH2movbp(sh,0,9);
}
#define init_6_0_10_4
void FASTCALL call_6_0_10_4(SH2_struct* sh)
{
  SH2movbp(sh,0,10);
}
#define init_6_0_11_4
void FASTCALL call_6_0_11_4(SH2_struct* sh)
{
  SH2movbp(sh,0,11);
}
#define init_6_0_12_4
void FASTCALL call_6_0_12_4(SH2_struct* sh)
{
  SH2movbp(sh,0,12);
}
#define init_6_0_13_4
void FASTCALL call_6_0_13_4(SH2_struct* sh)
{
  SH2movbp(sh,0,13);
}
#define init_6_0_14_4
void FASTCALL call_6_0_14_4(SH2_struct* sh)
{
  SH2movbp(sh,0,14);
}
#define init_6_0_15_4
void FASTCALL call_6_0_15_4(SH2_struct* sh)
{
  SH2movbp(sh,0,15);
}
#define init_6_1_0_4
void FASTCALL call_6_1_0_4(SH2_struct* sh)
{
  SH2movbp(sh,1,0);
}
#define init_6_1_1_4
void FASTCALL call_6_1_1_4(SH2_struct* sh)
{
  SH2movbp(sh,1,1);
}
#define init_6_1_2_4
void FASTCALL call_6_1_2_4(SH2_struct* sh)
{
  SH2movbp(sh,1,2);
}
#define init_6_1_3_4
void FASTCALL call_6_1_3_4(SH2_struct* sh)
{
  SH2movbp(sh,1,3);
}
#define init_6_1_4_4
void FASTCALL call_6_1_4_4(SH2_struct* sh)
{
  SH2movbp(sh,1,4);
}
#define init_6_1_5_4
void FASTCALL call_6_1_5_4(SH2_struct* sh)
{
  SH2movbp(sh,1,5);
}
#define init_6_1_6_4
void FASTCALL call_6_1_6_4(SH2_struct* sh)
{
  SH2movbp(sh,1,6);
}
#define init_6_1_7_4
void FASTCALL call_6_1_7_4(SH2_struct* sh)
{
  SH2movbp(sh,1,7);
}
#define init_6_1_8_4
void FASTCALL call_6_1_8_4(SH2_struct* sh)
{
  SH2movbp(sh,1,8);
}
#define init_6_1_9_4
void FASTCALL call_6_1_9_4(SH2_struct* sh)
{
  SH2movbp(sh,1,9);
}
#define init_6_1_10_4
void FASTCALL call_6_1_10_4(SH2_struct* sh)
{
  SH2movbp(sh,1,10);
}
#define init_6_1_11_4
void FASTCALL call_6_1_11_4(SH2_struct* sh)
{
  SH2movbp(sh,1,11);
}
#define init_6_1_12_4
void FASTCALL call_6_1_12_4(SH2_struct* sh)
{
  SH2movbp(sh,1,12);
}
#define init_6_1_13_4
void FASTCALL call_6_1_13_4(SH2_struct* sh)
{
  SH2movbp(sh,1,13);
}
#define init_6_1_14_4
void FASTCALL call_6_1_14_4(SH2_struct* sh)
{
  SH2movbp(sh,1,14);
}
#define init_6_1_15_4
void FASTCALL call_6_1_15_4(SH2_struct* sh)
{
  SH2movbp(sh,1,15);
}
#define init_6_2_0_4
void FASTCALL call_6_2_0_4(SH2_struct* sh)
{
  SH2movbp(sh,2,0);
}
#define init_6_2_1_4
void FASTCALL call_6_2_1_4(SH2_struct* sh)
{
  SH2movbp(sh,2,1);
}
#define init_6_2_2_4
void FASTCALL call_6_2_2_4(SH2_struct* sh)
{
  SH2movbp(sh,2,2);
}
#define init_6_2_3_4
void FASTCALL call_6_2_3_4(SH2_struct* sh)
{
  SH2movbp(sh,2,3);
}
#define init_6_2_4_4
void FASTCALL call_6_2_4_4(SH2_struct* sh)
{
  SH2movbp(sh,2,4);
}
#define init_6_2_5_4
void FASTCALL call_6_2_5_4(SH2_struct* sh)
{
  SH2movbp(sh,2,5);
}
#define init_6_2_6_4
void FASTCALL call_6_2_6_4(SH2_struct* sh)
{
  SH2movbp(sh,2,6);
}
#define init_6_2_7_4
void FASTCALL call_6_2_7_4(SH2_struct* sh)
{
  SH2movbp(sh,2,7);
}
#define init_6_2_8_4
void FASTCALL call_6_2_8_4(SH2_struct* sh)
{
  SH2movbp(sh,2,8);
}
#define init_6_2_9_4
void FASTCALL call_6_2_9_4(SH2_struct* sh)
{
  SH2movbp(sh,2,9);
}
#define init_6_2_10_4
void FASTCALL call_6_2_10_4(SH2_struct* sh)
{
  SH2movbp(sh,2,10);
}
#define init_6_2_11_4
void FASTCALL call_6_2_11_4(SH2_struct* sh)
{
  SH2movbp(sh,2,11);
}
#define init_6_2_12_4
void FASTCALL call_6_2_12_4(SH2_struct* sh)
{
  SH2movbp(sh,2,12);
}
#define init_6_2_13_4
void FASTCALL call_6_2_13_4(SH2_struct* sh)
{
  SH2movbp(sh,2,13);
}
#define init_6_2_14_4
void FASTCALL call_6_2_14_4(SH2_struct* sh)
{
  SH2movbp(sh,2,14);
}
#define init_6_2_15_4
void FASTCALL call_6_2_15_4(SH2_struct* sh)
{
  SH2movbp(sh,2,15);
}
#define init_6_3_0_4
void FASTCALL call_6_3_0_4(SH2_struct* sh)
{
  SH2movbp(sh,3,0);
}
#define init_6_3_1_4
void FASTCALL call_6_3_1_4(SH2_struct* sh)
{
  SH2movbp(sh,3,1);
}
#define init_6_3_2_4
void FASTCALL call_6_3_2_4(SH2_struct* sh)
{
  SH2movbp(sh,3,2);
}
#define init_6_3_3_4
void FASTCALL call_6_3_3_4(SH2_struct* sh)
{
  SH2movbp(sh,3,3);
}
#define init_6_3_4_4
void FASTCALL call_6_3_4_4(SH2_struct* sh)
{
  SH2movbp(sh,3,4);
}
#define init_6_3_5_4
void FASTCALL call_6_3_5_4(SH2_struct* sh)
{
  SH2movbp(sh,3,5);
}
#define init_6_3_6_4
void FASTCALL call_6_3_6_4(SH2_struct* sh)
{
  SH2movbp(sh,3,6);
}
#define init_6_3_7_4
void FASTCALL call_6_3_7_4(SH2_struct* sh)
{
  SH2movbp(sh,3,7);
}
#define init_6_3_8_4
void FASTCALL call_6_3_8_4(SH2_struct* sh)
{
  SH2movbp(sh,3,8);
}
#define init_6_3_9_4
void FASTCALL call_6_3_9_4(SH2_struct* sh)
{
  SH2movbp(sh,3,9);
}
#define init_6_3_10_4
void FASTCALL call_6_3_10_4(SH2_struct* sh)
{
  SH2movbp(sh,3,10);
}
#define init_6_3_11_4
void FASTCALL call_6_3_11_4(SH2_struct* sh)
{
  SH2movbp(sh,3,11);
}
#define init_6_3_12_4
void FASTCALL call_6_3_12_4(SH2_struct* sh)
{
  SH2movbp(sh,3,12);
}
#define init_6_3_13_4
void FASTCALL call_6_3_13_4(SH2_struct* sh)
{
  SH2movbp(sh,3,13);
}
#define init_6_3_14_4
void FASTCALL call_6_3_14_4(SH2_struct* sh)
{
  SH2movbp(sh,3,14);
}
#define init_6_3_15_4
void FASTCALL call_6_3_15_4(SH2_struct* sh)
{
  SH2movbp(sh,3,15);
}
#define init_6_4_0_4
void FASTCALL call_6_4_0_4(SH2_struct* sh)
{
  SH2movbp(sh,4,0);
}
#define init_6_4_1_4
void FASTCALL call_6_4_1_4(SH2_struct* sh)
{
  SH2movbp(sh,4,1);
}
#define init_6_4_2_4
void FASTCALL call_6_4_2_4(SH2_struct* sh)
{
  SH2movbp(sh,4,2);
}
#define init_6_4_3_4
void FASTCALL call_6_4_3_4(SH2_struct* sh)
{
  SH2movbp(sh,4,3);
}
#define init_6_4_4_4
void FASTCALL call_6_4_4_4(SH2_struct* sh)
{
  SH2movbp(sh,4,4);
}
#define init_6_4_5_4
void FASTCALL call_6_4_5_4(SH2_struct* sh)
{
  SH2movbp(sh,4,5);
}
#define init_6_4_6_4
void FASTCALL call_6_4_6_4(SH2_struct* sh)
{
  SH2movbp(sh,4,6);
}
#define init_6_4_7_4
void FASTCALL call_6_4_7_4(SH2_struct* sh)
{
  SH2movbp(sh,4,7);
}
#define init_6_4_8_4
void FASTCALL call_6_4_8_4(SH2_struct* sh)
{
  SH2movbp(sh,4,8);
}
#define init_6_4_9_4
void FASTCALL call_6_4_9_4(SH2_struct* sh)
{
  SH2movbp(sh,4,9);
}
#define init_6_4_10_4
void FASTCALL call_6_4_10_4(SH2_struct* sh)
{
  SH2movbp(sh,4,10);
}
#define init_6_4_11_4
void FASTCALL call_6_4_11_4(SH2_struct* sh)
{
  SH2movbp(sh,4,11);
}
#define init_6_4_12_4
void FASTCALL call_6_4_12_4(SH2_struct* sh)
{
  SH2movbp(sh,4,12);
}
#define init_6_4_13_4
void FASTCALL call_6_4_13_4(SH2_struct* sh)
{
  SH2movbp(sh,4,13);
}
#define init_6_4_14_4
void FASTCALL call_6_4_14_4(SH2_struct* sh)
{
  SH2movbp(sh,4,14);
}
#define init_6_4_15_4
void FASTCALL call_6_4_15_4(SH2_struct* sh)
{
  SH2movbp(sh,4,15);
}
#define init_6_5_0_4
void FASTCALL call_6_5_0_4(SH2_struct* sh)
{
  SH2movbp(sh,5,0);
}
#define init_6_5_1_4
void FASTCALL call_6_5_1_4(SH2_struct* sh)
{
  SH2movbp(sh,5,1);
}
#define init_6_5_2_4
void FASTCALL call_6_5_2_4(SH2_struct* sh)
{
  SH2movbp(sh,5,2);
}
#define init_6_5_3_4
void FASTCALL call_6_5_3_4(SH2_struct* sh)
{
  SH2movbp(sh,5,3);
}
#define init_6_5_4_4
void FASTCALL call_6_5_4_4(SH2_struct* sh)
{
  SH2movbp(sh,5,4);
}
#define init_6_5_5_4
void FASTCALL call_6_5_5_4(SH2_struct* sh)
{
  SH2movbp(sh,5,5);
}
#define init_6_5_6_4
void FASTCALL call_6_5_6_4(SH2_struct* sh)
{
  SH2movbp(sh,5,6);
}
#define init_6_5_7_4
void FASTCALL call_6_5_7_4(SH2_struct* sh)
{
  SH2movbp(sh,5,7);
}
#define init_6_5_8_4
void FASTCALL call_6_5_8_4(SH2_struct* sh)
{
  SH2movbp(sh,5,8);
}
#define init_6_5_9_4
void FASTCALL call_6_5_9_4(SH2_struct* sh)
{
  SH2movbp(sh,5,9);
}
#define init_6_5_10_4
void FASTCALL call_6_5_10_4(SH2_struct* sh)
{
  SH2movbp(sh,5,10);
}
#define init_6_5_11_4
void FASTCALL call_6_5_11_4(SH2_struct* sh)
{
  SH2movbp(sh,5,11);
}
#define init_6_5_12_4
void FASTCALL call_6_5_12_4(SH2_struct* sh)
{
  SH2movbp(sh,5,12);
}
#define init_6_5_13_4
void FASTCALL call_6_5_13_4(SH2_struct* sh)
{
  SH2movbp(sh,5,13);
}
#define init_6_5_14_4
void FASTCALL call_6_5_14_4(SH2_struct* sh)
{
  SH2movbp(sh,5,14);
}
#define init_6_5_15_4
void FASTCALL call_6_5_15_4(SH2_struct* sh)
{
  SH2movbp(sh,5,15);
}
#define init_6_6_0_4
void FASTCALL call_6_6_0_4(SH2_struct* sh)
{
  SH2movbp(sh,6,0);
}
#define init_6_6_1_4
void FASTCALL call_6_6_1_4(SH2_struct* sh)
{
  SH2movbp(sh,6,1);
}
#define init_6_6_2_4
void FASTCALL call_6_6_2_4(SH2_struct* sh)
{
  SH2movbp(sh,6,2);
}
#define init_6_6_3_4
void FASTCALL call_6_6_3_4(SH2_struct* sh)
{
  SH2movbp(sh,6,3);
}
#define init_6_6_4_4
void FASTCALL call_6_6_4_4(SH2_struct* sh)
{
  SH2movbp(sh,6,4);
}
#define init_6_6_5_4
void FASTCALL call_6_6_5_4(SH2_struct* sh)
{
  SH2movbp(sh,6,5);
}
#define init_6_6_6_4
void FASTCALL call_6_6_6_4(SH2_struct* sh)
{
  SH2movbp(sh,6,6);
}
#define init_6_6_7_4
void FASTCALL call_6_6_7_4(SH2_struct* sh)
{
  SH2movbp(sh,6,7);
}
#define init_6_6_8_4
void FASTCALL call_6_6_8_4(SH2_struct* sh)
{
  SH2movbp(sh,6,8);
}
#define init_6_6_9_4
void FASTCALL call_6_6_9_4(SH2_struct* sh)
{
  SH2movbp(sh,6,9);
}
#define init_6_6_10_4
void FASTCALL call_6_6_10_4(SH2_struct* sh)
{
  SH2movbp(sh,6,10);
}
#define init_6_6_11_4
void FASTCALL call_6_6_11_4(SH2_struct* sh)
{
  SH2movbp(sh,6,11);
}
#define init_6_6_12_4
void FASTCALL call_6_6_12_4(SH2_struct* sh)
{
  SH2movbp(sh,6,12);
}
#define init_6_6_13_4
void FASTCALL call_6_6_13_4(SH2_struct* sh)
{
  SH2movbp(sh,6,13);
}
#define init_6_6_14_4
void FASTCALL call_6_6_14_4(SH2_struct* sh)
{
  SH2movbp(sh,6,14);
}
#define init_6_6_15_4
void FASTCALL call_6_6_15_4(SH2_struct* sh)
{
  SH2movbp(sh,6,15);
}
#define init_6_7_0_4
void FASTCALL call_6_7_0_4(SH2_struct* sh)
{
  SH2movbp(sh,7,0);
}
#define init_6_7_1_4
void FASTCALL call_6_7_1_4(SH2_struct* sh)
{
  SH2movbp(sh,7,1);
}
#define init_6_7_2_4
void FASTCALL call_6_7_2_4(SH2_struct* sh)
{
  SH2movbp(sh,7,2);
}
#define init_6_7_3_4
void FASTCALL call_6_7_3_4(SH2_struct* sh)
{
  SH2movbp(sh,7,3);
}
#define init_6_7_4_4
void FASTCALL call_6_7_4_4(SH2_struct* sh)
{
  SH2movbp(sh,7,4);
}
#define init_6_7_5_4
void FASTCALL call_6_7_5_4(SH2_struct* sh)
{
  SH2movbp(sh,7,5);
}
#define init_6_7_6_4
void FASTCALL call_6_7_6_4(SH2_struct* sh)
{
  SH2movbp(sh,7,6);
}
#define init_6_7_7_4
void FASTCALL call_6_7_7_4(SH2_struct* sh)
{
  SH2movbp(sh,7,7);
}
#define init_6_7_8_4
void FASTCALL call_6_7_8_4(SH2_struct* sh)
{
  SH2movbp(sh,7,8);
}
#define init_6_7_9_4
void FASTCALL call_6_7_9_4(SH2_struct* sh)
{
  SH2movbp(sh,7,9);
}
#define init_6_7_10_4
void FASTCALL call_6_7_10_4(SH2_struct* sh)
{
  SH2movbp(sh,7,10);
}
#define init_6_7_11_4
void FASTCALL call_6_7_11_4(SH2_struct* sh)
{
  SH2movbp(sh,7,11);
}
#define init_6_7_12_4
void FASTCALL call_6_7_12_4(SH2_struct* sh)
{
  SH2movbp(sh,7,12);
}
#define init_6_7_13_4
void FASTCALL call_6_7_13_4(SH2_struct* sh)
{
  SH2movbp(sh,7,13);
}
#define init_6_7_14_4
void FASTCALL call_6_7_14_4(SH2_struct* sh)
{
  SH2movbp(sh,7,14);
}
#define init_6_7_15_4
void FASTCALL call_6_7_15_4(SH2_struct* sh)
{
  SH2movbp(sh,7,15);
}
#define init_6_8_0_4
void FASTCALL call_6_8_0_4(SH2_struct* sh)
{
  SH2movbp(sh,8,0);
}
#define init_6_8_1_4
void FASTCALL call_6_8_1_4(SH2_struct* sh)
{
  SH2movbp(sh,8,1);
}
#define init_6_8_2_4
void FASTCALL call_6_8_2_4(SH2_struct* sh)
{
  SH2movbp(sh,8,2);
}
#define init_6_8_3_4
void FASTCALL call_6_8_3_4(SH2_struct* sh)
{
  SH2movbp(sh,8,3);
}
#define init_6_8_4_4
void FASTCALL call_6_8_4_4(SH2_struct* sh)
{
  SH2movbp(sh,8,4);
}
#define init_6_8_5_4
void FASTCALL call_6_8_5_4(SH2_struct* sh)
{
  SH2movbp(sh,8,5);
}
#define init_6_8_6_4
void FASTCALL call_6_8_6_4(SH2_struct* sh)
{
  SH2movbp(sh,8,6);
}
#define init_6_8_7_4
void FASTCALL call_6_8_7_4(SH2_struct* sh)
{
  SH2movbp(sh,8,7);
}
#define init_6_8_8_4
void FASTCALL call_6_8_8_4(SH2_struct* sh)
{
  SH2movbp(sh,8,8);
}
#define init_6_8_9_4
void FASTCALL call_6_8_9_4(SH2_struct* sh)
{
  SH2movbp(sh,8,9);
}
#define init_6_8_10_4
void FASTCALL call_6_8_10_4(SH2_struct* sh)
{
  SH2movbp(sh,8,10);
}
#define init_6_8_11_4
void FASTCALL call_6_8_11_4(SH2_struct* sh)
{
  SH2movbp(sh,8,11);
}
#define init_6_8_12_4
void FASTCALL call_6_8_12_4(SH2_struct* sh)
{
  SH2movbp(sh,8,12);
}
#define init_6_8_13_4
void FASTCALL call_6_8_13_4(SH2_struct* sh)
{
  SH2movbp(sh,8,13);
}
#define init_6_8_14_4
void FASTCALL call_6_8_14_4(SH2_struct* sh)
{
  SH2movbp(sh,8,14);
}
#define init_6_8_15_4
void FASTCALL call_6_8_15_4(SH2_struct* sh)
{
  SH2movbp(sh,8,15);
}
#define init_6_9_0_4
void FASTCALL call_6_9_0_4(SH2_struct* sh)
{
  SH2movbp(sh,9,0);
}
#define init_6_9_1_4
void FASTCALL call_6_9_1_4(SH2_struct* sh)
{
  SH2movbp(sh,9,1);
}
#define init_6_9_2_4
void FASTCALL call_6_9_2_4(SH2_struct* sh)
{
  SH2movbp(sh,9,2);
}
#define init_6_9_3_4
void FASTCALL call_6_9_3_4(SH2_struct* sh)
{
  SH2movbp(sh,9,3);
}
#define init_6_9_4_4
void FASTCALL call_6_9_4_4(SH2_struct* sh)
{
  SH2movbp(sh,9,4);
}
#define init_6_9_5_4
void FASTCALL call_6_9_5_4(SH2_struct* sh)
{
  SH2movbp(sh,9,5);
}
#define init_6_9_6_4
void FASTCALL call_6_9_6_4(SH2_struct* sh)
{
  SH2movbp(sh,9,6);
}
#define init_6_9_7_4
void FASTCALL call_6_9_7_4(SH2_struct* sh)
{
  SH2movbp(sh,9,7);
}
#define init_6_9_8_4
void FASTCALL call_6_9_8_4(SH2_struct* sh)
{
  SH2movbp(sh,9,8);
}
#define init_6_9_9_4
void FASTCALL call_6_9_9_4(SH2_struct* sh)
{
  SH2movbp(sh,9,9);
}
#define init_6_9_10_4
void FASTCALL call_6_9_10_4(SH2_struct* sh)
{
  SH2movbp(sh,9,10);
}
#define init_6_9_11_4
void FASTCALL call_6_9_11_4(SH2_struct* sh)
{
  SH2movbp(sh,9,11);
}
#define init_6_9_12_4
void FASTCALL call_6_9_12_4(SH2_struct* sh)
{
  SH2movbp(sh,9,12);
}
#define init_6_9_13_4
void FASTCALL call_6_9_13_4(SH2_struct* sh)
{
  SH2movbp(sh,9,13);
}
#define init_6_9_14_4
void FASTCALL call_6_9_14_4(SH2_struct* sh)
{
  SH2movbp(sh,9,14);
}
#define init_6_9_15_4
void FASTCALL call_6_9_15_4(SH2_struct* sh)
{
  SH2movbp(sh,9,15);
}
#define init_6_10_0_4
void FASTCALL call_6_10_0_4(SH2_struct* sh)
{
  SH2movbp(sh,10,0);
}
#define init_6_10_1_4
void FASTCALL call_6_10_1_4(SH2_struct* sh)
{
  SH2movbp(sh,10,1);
}
#define init_6_10_2_4
void FASTCALL call_6_10_2_4(SH2_struct* sh)
{
  SH2movbp(sh,10,2);
}
#define init_6_10_3_4
void FASTCALL call_6_10_3_4(SH2_struct* sh)
{
  SH2movbp(sh,10,3);
}
#define init_6_10_4_4
void FASTCALL call_6_10_4_4(SH2_struct* sh)
{
  SH2movbp(sh,10,4);
}
#define init_6_10_5_4
void FASTCALL call_6_10_5_4(SH2_struct* sh)
{
  SH2movbp(sh,10,5);
}
#define init_6_10_6_4
void FASTCALL call_6_10_6_4(SH2_struct* sh)
{
  SH2movbp(sh,10,6);
}
#define init_6_10_7_4
void FASTCALL call_6_10_7_4(SH2_struct* sh)
{
  SH2movbp(sh,10,7);
}
#define init_6_10_8_4
void FASTCALL call_6_10_8_4(SH2_struct* sh)
{
  SH2movbp(sh,10,8);
}
#define init_6_10_9_4
void FASTCALL call_6_10_9_4(SH2_struct* sh)
{
  SH2movbp(sh,10,9);
}
#define init_6_10_10_4
void FASTCALL call_6_10_10_4(SH2_struct* sh)
{
  SH2movbp(sh,10,10);
}
#define init_6_10_11_4
void FASTCALL call_6_10_11_4(SH2_struct* sh)
{
  SH2movbp(sh,10,11);
}
#define init_6_10_12_4
void FASTCALL call_6_10_12_4(SH2_struct* sh)
{
  SH2movbp(sh,10,12);
}
#define init_6_10_13_4
void FASTCALL call_6_10_13_4(SH2_struct* sh)
{
  SH2movbp(sh,10,13);
}
#define init_6_10_14_4
void FASTCALL call_6_10_14_4(SH2_struct* sh)
{
  SH2movbp(sh,10,14);
}
#define init_6_10_15_4
void FASTCALL call_6_10_15_4(SH2_struct* sh)
{
  SH2movbp(sh,10,15);
}
#define init_6_11_0_4
void FASTCALL call_6_11_0_4(SH2_struct* sh)
{
  SH2movbp(sh,11,0);
}
#define init_6_11_1_4
void FASTCALL call_6_11_1_4(SH2_struct* sh)
{
  SH2movbp(sh,11,1);
}
#define init_6_11_2_4
void FASTCALL call_6_11_2_4(SH2_struct* sh)
{
  SH2movbp(sh,11,2);
}
#define init_6_11_3_4
void FASTCALL call_6_11_3_4(SH2_struct* sh)
{
  SH2movbp(sh,11,3);
}
#define init_6_11_4_4
void FASTCALL call_6_11_4_4(SH2_struct* sh)
{
  SH2movbp(sh,11,4);
}
#define init_6_11_5_4
void FASTCALL call_6_11_5_4(SH2_struct* sh)
{
  SH2movbp(sh,11,5);
}
#define init_6_11_6_4
void FASTCALL call_6_11_6_4(SH2_struct* sh)
{
  SH2movbp(sh,11,6);
}
#define init_6_11_7_4
void FASTCALL call_6_11_7_4(SH2_struct* sh)
{
  SH2movbp(sh,11,7);
}
#define init_6_11_8_4
void FASTCALL call_6_11_8_4(SH2_struct* sh)
{
  SH2movbp(sh,11,8);
}
#define init_6_11_9_4
void FASTCALL call_6_11_9_4(SH2_struct* sh)
{
  SH2movbp(sh,11,9);
}
#define init_6_11_10_4
void FASTCALL call_6_11_10_4(SH2_struct* sh)
{
  SH2movbp(sh,11,10);
}
#define init_6_11_11_4
void FASTCALL call_6_11_11_4(SH2_struct* sh)
{
  SH2movbp(sh,11,11);
}
#define init_6_11_12_4
void FASTCALL call_6_11_12_4(SH2_struct* sh)
{
  SH2movbp(sh,11,12);
}
#define init_6_11_13_4
void FASTCALL call_6_11_13_4(SH2_struct* sh)
{
  SH2movbp(sh,11,13);
}
#define init_6_11_14_4
void FASTCALL call_6_11_14_4(SH2_struct* sh)
{
  SH2movbp(sh,11,14);
}
#define init_6_11_15_4
void FASTCALL call_6_11_15_4(SH2_struct* sh)
{
  SH2movbp(sh,11,15);
}
#define init_6_12_0_4
void FASTCALL call_6_12_0_4(SH2_struct* sh)
{
  SH2movbp(sh,12,0);
}
#define init_6_12_1_4
void FASTCALL call_6_12_1_4(SH2_struct* sh)
{
  SH2movbp(sh,12,1);
}
#define init_6_12_2_4
void FASTCALL call_6_12_2_4(SH2_struct* sh)
{
  SH2movbp(sh,12,2);
}
#define init_6_12_3_4
void FASTCALL call_6_12_3_4(SH2_struct* sh)
{
  SH2movbp(sh,12,3);
}
#define init_6_12_4_4
void FASTCALL call_6_12_4_4(SH2_struct* sh)
{
  SH2movbp(sh,12,4);
}
#define init_6_12_5_4
void FASTCALL call_6_12_5_4(SH2_struct* sh)
{
  SH2movbp(sh,12,5);
}
#define init_6_12_6_4
void FASTCALL call_6_12_6_4(SH2_struct* sh)
{
  SH2movbp(sh,12,6);
}
#define init_6_12_7_4
void FASTCALL call_6_12_7_4(SH2_struct* sh)
{
  SH2movbp(sh,12,7);
}
#define init_6_12_8_4
void FASTCALL call_6_12_8_4(SH2_struct* sh)
{
  SH2movbp(sh,12,8);
}
#define init_6_12_9_4
void FASTCALL call_6_12_9_4(SH2_struct* sh)
{
  SH2movbp(sh,12,9);
}
#define init_6_12_10_4
void FASTCALL call_6_12_10_4(SH2_struct* sh)
{
  SH2movbp(sh,12,10);
}
#define init_6_12_11_4
void FASTCALL call_6_12_11_4(SH2_struct* sh)
{
  SH2movbp(sh,12,11);
}
#define init_6_12_12_4
void FASTCALL call_6_12_12_4(SH2_struct* sh)
{
  SH2movbp(sh,12,12);
}
#define init_6_12_13_4
void FASTCALL call_6_12_13_4(SH2_struct* sh)
{
  SH2movbp(sh,12,13);
}
#define init_6_12_14_4
void FASTCALL call_6_12_14_4(SH2_struct* sh)
{
  SH2movbp(sh,12,14);
}
#define init_6_12_15_4
void FASTCALL call_6_12_15_4(SH2_struct* sh)
{
  SH2movbp(sh,12,15);
}
#define init_6_13_0_4
void FASTCALL call_6_13_0_4(SH2_struct* sh)
{
  SH2movbp(sh,13,0);
}
#define init_6_13_1_4
void FASTCALL call_6_13_1_4(SH2_struct* sh)
{
  SH2movbp(sh,13,1);
}
#define init_6_13_2_4
void FASTCALL call_6_13_2_4(SH2_struct* sh)
{
  SH2movbp(sh,13,2);
}
#define init_6_13_3_4
void FASTCALL call_6_13_3_4(SH2_struct* sh)
{
  SH2movbp(sh,13,3);
}
#define init_6_13_4_4
void FASTCALL call_6_13_4_4(SH2_struct* sh)
{
  SH2movbp(sh,13,4);
}
#define init_6_13_5_4
void FASTCALL call_6_13_5_4(SH2_struct* sh)
{
  SH2movbp(sh,13,5);
}
#define init_6_13_6_4
void FASTCALL call_6_13_6_4(SH2_struct* sh)
{
  SH2movbp(sh,13,6);
}
#define init_6_13_7_4
void FASTCALL call_6_13_7_4(SH2_struct* sh)
{
  SH2movbp(sh,13,7);
}
#define init_6_13_8_4
void FASTCALL call_6_13_8_4(SH2_struct* sh)
{
  SH2movbp(sh,13,8);
}
#define init_6_13_9_4
void FASTCALL call_6_13_9_4(SH2_struct* sh)
{
  SH2movbp(sh,13,9);
}
#define init_6_13_10_4
void FASTCALL call_6_13_10_4(SH2_struct* sh)
{
  SH2movbp(sh,13,10);
}
#define init_6_13_11_4
void FASTCALL call_6_13_11_4(SH2_struct* sh)
{
  SH2movbp(sh,13,11);
}
#define init_6_13_12_4
void FASTCALL call_6_13_12_4(SH2_struct* sh)
{
  SH2movbp(sh,13,12);
}
#define init_6_13_13_4
void FASTCALL call_6_13_13_4(SH2_struct* sh)
{
  SH2movbp(sh,13,13);
}
#define init_6_13_14_4
void FASTCALL call_6_13_14_4(SH2_struct* sh)
{
  SH2movbp(sh,13,14);
}
#define init_6_13_15_4
void FASTCALL call_6_13_15_4(SH2_struct* sh)
{
  SH2movbp(sh,13,15);
}
#define init_6_14_0_4
void FASTCALL call_6_14_0_4(SH2_struct* sh)
{
  SH2movbp(sh,14,0);
}
#define init_6_14_1_4
void FASTCALL call_6_14_1_4(SH2_struct* sh)
{
  SH2movbp(sh,14,1);
}
#define init_6_14_2_4
void FASTCALL call_6_14_2_4(SH2_struct* sh)
{
  SH2movbp(sh,14,2);
}
#define init_6_14_3_4
void FASTCALL call_6_14_3_4(SH2_struct* sh)
{
  SH2movbp(sh,14,3);
}
#define init_6_14_4_4
void FASTCALL call_6_14_4_4(SH2_struct* sh)
{
  SH2movbp(sh,14,4);
}
#define init_6_14_5_4
void FASTCALL call_6_14_5_4(SH2_struct* sh)
{
  SH2movbp(sh,14,5);
}
#define init_6_14_6_4
void FASTCALL call_6_14_6_4(SH2_struct* sh)
{
  SH2movbp(sh,14,6);
}
#define init_6_14_7_4
void FASTCALL call_6_14_7_4(SH2_struct* sh)
{
  SH2movbp(sh,14,7);
}
#define init_6_14_8_4
void FASTCALL call_6_14_8_4(SH2_struct* sh)
{
  SH2movbp(sh,14,8);
}
#define init_6_14_9_4
void FASTCALL call_6_14_9_4(SH2_struct* sh)
{
  SH2movbp(sh,14,9);
}
#define init_6_14_10_4
void FASTCALL call_6_14_10_4(SH2_struct* sh)
{
  SH2movbp(sh,14,10);
}
#define init_6_14_11_4
void FASTCALL call_6_14_11_4(SH2_struct* sh)
{
  SH2movbp(sh,14,11);
}
#define init_6_14_12_4
void FASTCALL call_6_14_12_4(SH2_struct* sh)
{
  SH2movbp(sh,14,12);
}
#define init_6_14_13_4
void FASTCALL call_6_14_13_4(SH2_struct* sh)
{
  SH2movbp(sh,14,13);
}
#define init_6_14_14_4
void FASTCALL call_6_14_14_4(SH2_struct* sh)
{
  SH2movbp(sh,14,14);
}
#define init_6_14_15_4
void FASTCALL call_6_14_15_4(SH2_struct* sh)
{
  SH2movbp(sh,14,15);
}
#define init_6_15_0_4
void FASTCALL call_6_15_0_4(SH2_struct* sh)
{
  SH2movbp(sh,15,0);
}
#define init_6_15_1_4
void FASTCALL call_6_15_1_4(SH2_struct* sh)
{
  SH2movbp(sh,15,1);
}
#define init_6_15_2_4
void FASTCALL call_6_15_2_4(SH2_struct* sh)
{
  SH2movbp(sh,15,2);
}
#define init_6_15_3_4
void FASTCALL call_6_15_3_4(SH2_struct* sh)
{
  SH2movbp(sh,15,3);
}
#define init_6_15_4_4
void FASTCALL call_6_15_4_4(SH2_struct* sh)
{
  SH2movbp(sh,15,4);
}
#define init_6_15_5_4
void FASTCALL call_6_15_5_4(SH2_struct* sh)
{
  SH2movbp(sh,15,5);
}
#define init_6_15_6_4
void FASTCALL call_6_15_6_4(SH2_struct* sh)
{
  SH2movbp(sh,15,6);
}
#define init_6_15_7_4
void FASTCALL call_6_15_7_4(SH2_struct* sh)
{
  SH2movbp(sh,15,7);
}
#define init_6_15_8_4
void FASTCALL call_6_15_8_4(SH2_struct* sh)
{
  SH2movbp(sh,15,8);
}
#define init_6_15_9_4
void FASTCALL call_6_15_9_4(SH2_struct* sh)
{
  SH2movbp(sh,15,9);
}
#define init_6_15_10_4
void FASTCALL call_6_15_10_4(SH2_struct* sh)
{
  SH2movbp(sh,15,10);
}
#define init_6_15_11_4
void FASTCALL call_6_15_11_4(SH2_struct* sh)
{
  SH2movbp(sh,15,11);
}
#define init_6_15_12_4
void FASTCALL call_6_15_12_4(SH2_struct* sh)
{
  SH2movbp(sh,15,12);
}
#define init_6_15_13_4
void FASTCALL call_6_15_13_4(SH2_struct* sh)
{
  SH2movbp(sh,15,13);
}
#define init_6_15_14_4
void FASTCALL call_6_15_14_4(SH2_struct* sh)
{
  SH2movbp(sh,15,14);
}
#define init_6_15_15_4
void FASTCALL call_6_15_15_4(SH2_struct* sh)
{
  SH2movbp(sh,15,15);
}
#define init_2_0_0_4
void FASTCALL call_2_0_0_4(SH2_struct* sh)
{
  SH2movbm(sh,0,0);
}
#define init_2_0_1_4
void FASTCALL call_2_0_1_4(SH2_struct* sh)
{
  SH2movbm(sh,0,1);
}
#define init_2_0_2_4
void FASTCALL call_2_0_2_4(SH2_struct* sh)
{
  SH2movbm(sh,0,2);
}
#define init_2_0_3_4
void FASTCALL call_2_0_3_4(SH2_struct* sh)
{
  SH2movbm(sh,0,3);
}
#define init_2_0_4_4
void FASTCALL call_2_0_4_4(SH2_struct* sh)
{
  SH2movbm(sh,0,4);
}
#define init_2_0_5_4
void FASTCALL call_2_0_5_4(SH2_struct* sh)
{
  SH2movbm(sh,0,5);
}
#define init_2_0_6_4
void FASTCALL call_2_0_6_4(SH2_struct* sh)
{
  SH2movbm(sh,0,6);
}
#define init_2_0_7_4
void FASTCALL call_2_0_7_4(SH2_struct* sh)
{
  SH2movbm(sh,0,7);
}
#define init_2_0_8_4
void FASTCALL call_2_0_8_4(SH2_struct* sh)
{
  SH2movbm(sh,0,8);
}
#define init_2_0_9_4
void FASTCALL call_2_0_9_4(SH2_struct* sh)
{
  SH2movbm(sh,0,9);
}
#define init_2_0_10_4
void FASTCALL call_2_0_10_4(SH2_struct* sh)
{
  SH2movbm(sh,0,10);
}
#define init_2_0_11_4
void FASTCALL call_2_0_11_4(SH2_struct* sh)
{
  SH2movbm(sh,0,11);
}
#define init_2_0_12_4
void FASTCALL call_2_0_12_4(SH2_struct* sh)
{
  SH2movbm(sh,0,12);
}
#define init_2_0_13_4
void FASTCALL call_2_0_13_4(SH2_struct* sh)
{
  SH2movbm(sh,0,13);
}
#define init_2_0_14_4
void FASTCALL call_2_0_14_4(SH2_struct* sh)
{
  SH2movbm(sh,0,14);
}
#define init_2_0_15_4
void FASTCALL call_2_0_15_4(SH2_struct* sh)
{
  SH2movbm(sh,0,15);
}
#define init_2_1_0_4
void FASTCALL call_2_1_0_4(SH2_struct* sh)
{
  SH2movbm(sh,1,0);
}
#define init_2_1_1_4
void FASTCALL call_2_1_1_4(SH2_struct* sh)
{
  SH2movbm(sh,1,1);
}
#define init_2_1_2_4
void FASTCALL call_2_1_2_4(SH2_struct* sh)
{
  SH2movbm(sh,1,2);
}
#define init_2_1_3_4
void FASTCALL call_2_1_3_4(SH2_struct* sh)
{
  SH2movbm(sh,1,3);
}
#define init_2_1_4_4
void FASTCALL call_2_1_4_4(SH2_struct* sh)
{
  SH2movbm(sh,1,4);
}
#define init_2_1_5_4
void FASTCALL call_2_1_5_4(SH2_struct* sh)
{
  SH2movbm(sh,1,5);
}
#define init_2_1_6_4
void FASTCALL call_2_1_6_4(SH2_struct* sh)
{
  SH2movbm(sh,1,6);
}
#define init_2_1_7_4
void FASTCALL call_2_1_7_4(SH2_struct* sh)
{
  SH2movbm(sh,1,7);
}
#define init_2_1_8_4
void FASTCALL call_2_1_8_4(SH2_struct* sh)
{
  SH2movbm(sh,1,8);
}
#define init_2_1_9_4
void FASTCALL call_2_1_9_4(SH2_struct* sh)
{
  SH2movbm(sh,1,9);
}
#define init_2_1_10_4
void FASTCALL call_2_1_10_4(SH2_struct* sh)
{
  SH2movbm(sh,1,10);
}
#define init_2_1_11_4
void FASTCALL call_2_1_11_4(SH2_struct* sh)
{
  SH2movbm(sh,1,11);
}
#define init_2_1_12_4
void FASTCALL call_2_1_12_4(SH2_struct* sh)
{
  SH2movbm(sh,1,12);
}
#define init_2_1_13_4
void FASTCALL call_2_1_13_4(SH2_struct* sh)
{
  SH2movbm(sh,1,13);
}
#define init_2_1_14_4
void FASTCALL call_2_1_14_4(SH2_struct* sh)
{
  SH2movbm(sh,1,14);
}
#define init_2_1_15_4
void FASTCALL call_2_1_15_4(SH2_struct* sh)
{
  SH2movbm(sh,1,15);
}
#define init_2_2_0_4
void FASTCALL call_2_2_0_4(SH2_struct* sh)
{
  SH2movbm(sh,2,0);
}
#define init_2_2_1_4
void FASTCALL call_2_2_1_4(SH2_struct* sh)
{
  SH2movbm(sh,2,1);
}
#define init_2_2_2_4
void FASTCALL call_2_2_2_4(SH2_struct* sh)
{
  SH2movbm(sh,2,2);
}
#define init_2_2_3_4
void FASTCALL call_2_2_3_4(SH2_struct* sh)
{
  SH2movbm(sh,2,3);
}
#define init_2_2_4_4
void FASTCALL call_2_2_4_4(SH2_struct* sh)
{
  SH2movbm(sh,2,4);
}
#define init_2_2_5_4
void FASTCALL call_2_2_5_4(SH2_struct* sh)
{
  SH2movbm(sh,2,5);
}
#define init_2_2_6_4
void FASTCALL call_2_2_6_4(SH2_struct* sh)
{
  SH2movbm(sh,2,6);
}
#define init_2_2_7_4
void FASTCALL call_2_2_7_4(SH2_struct* sh)
{
  SH2movbm(sh,2,7);
}
#define init_2_2_8_4
void FASTCALL call_2_2_8_4(SH2_struct* sh)
{
  SH2movbm(sh,2,8);
}
#define init_2_2_9_4
void FASTCALL call_2_2_9_4(SH2_struct* sh)
{
  SH2movbm(sh,2,9);
}
#define init_2_2_10_4
void FASTCALL call_2_2_10_4(SH2_struct* sh)
{
  SH2movbm(sh,2,10);
}
#define init_2_2_11_4
void FASTCALL call_2_2_11_4(SH2_struct* sh)
{
  SH2movbm(sh,2,11);
}
#define init_2_2_12_4
void FASTCALL call_2_2_12_4(SH2_struct* sh)
{
  SH2movbm(sh,2,12);
}
#define init_2_2_13_4
void FASTCALL call_2_2_13_4(SH2_struct* sh)
{
  SH2movbm(sh,2,13);
}
#define init_2_2_14_4
void FASTCALL call_2_2_14_4(SH2_struct* sh)
{
  SH2movbm(sh,2,14);
}
#define init_2_2_15_4
void FASTCALL call_2_2_15_4(SH2_struct* sh)
{
  SH2movbm(sh,2,15);
}
#define init_2_3_0_4
void FASTCALL call_2_3_0_4(SH2_struct* sh)
{
  SH2movbm(sh,3,0);
}
#define init_2_3_1_4
void FASTCALL call_2_3_1_4(SH2_struct* sh)
{
  SH2movbm(sh,3,1);
}
#define init_2_3_2_4
void FASTCALL call_2_3_2_4(SH2_struct* sh)
{
  SH2movbm(sh,3,2);
}
#define init_2_3_3_4
void FASTCALL call_2_3_3_4(SH2_struct* sh)
{
  SH2movbm(sh,3,3);
}
#define init_2_3_4_4
void FASTCALL call_2_3_4_4(SH2_struct* sh)
{
  SH2movbm(sh,3,4);
}
#define init_2_3_5_4
void FASTCALL call_2_3_5_4(SH2_struct* sh)
{
  SH2movbm(sh,3,5);
}
#define init_2_3_6_4
void FASTCALL call_2_3_6_4(SH2_struct* sh)
{
  SH2movbm(sh,3,6);
}
#define init_2_3_7_4
void FASTCALL call_2_3_7_4(SH2_struct* sh)
{
  SH2movbm(sh,3,7);
}
#define init_2_3_8_4
void FASTCALL call_2_3_8_4(SH2_struct* sh)
{
  SH2movbm(sh,3,8);
}
#define init_2_3_9_4
void FASTCALL call_2_3_9_4(SH2_struct* sh)
{
  SH2movbm(sh,3,9);
}
#define init_2_3_10_4
void FASTCALL call_2_3_10_4(SH2_struct* sh)
{
  SH2movbm(sh,3,10);
}
#define init_2_3_11_4
void FASTCALL call_2_3_11_4(SH2_struct* sh)
{
  SH2movbm(sh,3,11);
}
#define init_2_3_12_4
void FASTCALL call_2_3_12_4(SH2_struct* sh)
{
  SH2movbm(sh,3,12);
}
#define init_2_3_13_4
void FASTCALL call_2_3_13_4(SH2_struct* sh)
{
  SH2movbm(sh,3,13);
}
#define init_2_3_14_4
void FASTCALL call_2_3_14_4(SH2_struct* sh)
{
  SH2movbm(sh,3,14);
}
#define init_2_3_15_4
void FASTCALL call_2_3_15_4(SH2_struct* sh)
{
  SH2movbm(sh,3,15);
}
#define init_2_4_0_4
void FASTCALL call_2_4_0_4(SH2_struct* sh)
{
  SH2movbm(sh,4,0);
}
#define init_2_4_1_4
void FASTCALL call_2_4_1_4(SH2_struct* sh)
{
  SH2movbm(sh,4,1);
}
#define init_2_4_2_4
void FASTCALL call_2_4_2_4(SH2_struct* sh)
{
  SH2movbm(sh,4,2);
}
#define init_2_4_3_4
void FASTCALL call_2_4_3_4(SH2_struct* sh)
{
  SH2movbm(sh,4,3);
}
#define init_2_4_4_4
void FASTCALL call_2_4_4_4(SH2_struct* sh)
{
  SH2movbm(sh,4,4);
}
#define init_2_4_5_4
void FASTCALL call_2_4_5_4(SH2_struct* sh)
{
  SH2movbm(sh,4,5);
}
#define init_2_4_6_4
void FASTCALL call_2_4_6_4(SH2_struct* sh)
{
  SH2movbm(sh,4,6);
}
#define init_2_4_7_4
void FASTCALL call_2_4_7_4(SH2_struct* sh)
{
  SH2movbm(sh,4,7);
}
#define init_2_4_8_4
void FASTCALL call_2_4_8_4(SH2_struct* sh)
{
  SH2movbm(sh,4,8);
}
#define init_2_4_9_4
void FASTCALL call_2_4_9_4(SH2_struct* sh)
{
  SH2movbm(sh,4,9);
}
#define init_2_4_10_4
void FASTCALL call_2_4_10_4(SH2_struct* sh)
{
  SH2movbm(sh,4,10);
}
#define init_2_4_11_4
void FASTCALL call_2_4_11_4(SH2_struct* sh)
{
  SH2movbm(sh,4,11);
}
#define init_2_4_12_4
void FASTCALL call_2_4_12_4(SH2_struct* sh)
{
  SH2movbm(sh,4,12);
}
#define init_2_4_13_4
void FASTCALL call_2_4_13_4(SH2_struct* sh)
{
  SH2movbm(sh,4,13);
}
#define init_2_4_14_4
void FASTCALL call_2_4_14_4(SH2_struct* sh)
{
  SH2movbm(sh,4,14);
}
#define init_2_4_15_4
void FASTCALL call_2_4_15_4(SH2_struct* sh)
{
  SH2movbm(sh,4,15);
}
#define init_2_5_0_4
void FASTCALL call_2_5_0_4(SH2_struct* sh)
{
  SH2movbm(sh,5,0);
}
#define init_2_5_1_4
void FASTCALL call_2_5_1_4(SH2_struct* sh)
{
  SH2movbm(sh,5,1);
}
#define init_2_5_2_4
void FASTCALL call_2_5_2_4(SH2_struct* sh)
{
  SH2movbm(sh,5,2);
}
#define init_2_5_3_4
void FASTCALL call_2_5_3_4(SH2_struct* sh)
{
  SH2movbm(sh,5,3);
}
#define init_2_5_4_4
void FASTCALL call_2_5_4_4(SH2_struct* sh)
{
  SH2movbm(sh,5,4);
}
#define init_2_5_5_4
void FASTCALL call_2_5_5_4(SH2_struct* sh)
{
  SH2movbm(sh,5,5);
}
#define init_2_5_6_4
void FASTCALL call_2_5_6_4(SH2_struct* sh)
{
  SH2movbm(sh,5,6);
}
#define init_2_5_7_4
void FASTCALL call_2_5_7_4(SH2_struct* sh)
{
  SH2movbm(sh,5,7);
}
#define init_2_5_8_4
void FASTCALL call_2_5_8_4(SH2_struct* sh)
{
  SH2movbm(sh,5,8);
}
#define init_2_5_9_4
void FASTCALL call_2_5_9_4(SH2_struct* sh)
{
  SH2movbm(sh,5,9);
}
#define init_2_5_10_4
void FASTCALL call_2_5_10_4(SH2_struct* sh)
{
  SH2movbm(sh,5,10);
}
#define init_2_5_11_4
void FASTCALL call_2_5_11_4(SH2_struct* sh)
{
  SH2movbm(sh,5,11);
}
#define init_2_5_12_4
void FASTCALL call_2_5_12_4(SH2_struct* sh)
{
  SH2movbm(sh,5,12);
}
#define init_2_5_13_4
void FASTCALL call_2_5_13_4(SH2_struct* sh)
{
  SH2movbm(sh,5,13);
}
#define init_2_5_14_4
void FASTCALL call_2_5_14_4(SH2_struct* sh)
{
  SH2movbm(sh,5,14);
}
#define init_2_5_15_4
void FASTCALL call_2_5_15_4(SH2_struct* sh)
{
  SH2movbm(sh,5,15);
}
#define init_2_6_0_4
void FASTCALL call_2_6_0_4(SH2_struct* sh)
{
  SH2movbm(sh,6,0);
}
#define init_2_6_1_4
void FASTCALL call_2_6_1_4(SH2_struct* sh)
{
  SH2movbm(sh,6,1);
}
#define init_2_6_2_4
void FASTCALL call_2_6_2_4(SH2_struct* sh)
{
  SH2movbm(sh,6,2);
}
#define init_2_6_3_4
void FASTCALL call_2_6_3_4(SH2_struct* sh)
{
  SH2movbm(sh,6,3);
}
#define init_2_6_4_4
void FASTCALL call_2_6_4_4(SH2_struct* sh)
{
  SH2movbm(sh,6,4);
}
#define init_2_6_5_4
void FASTCALL call_2_6_5_4(SH2_struct* sh)
{
  SH2movbm(sh,6,5);
}
#define init_2_6_6_4
void FASTCALL call_2_6_6_4(SH2_struct* sh)
{
  SH2movbm(sh,6,6);
}
#define init_2_6_7_4
void FASTCALL call_2_6_7_4(SH2_struct* sh)
{
  SH2movbm(sh,6,7);
}
#define init_2_6_8_4
void FASTCALL call_2_6_8_4(SH2_struct* sh)
{
  SH2movbm(sh,6,8);
}
#define init_2_6_9_4
void FASTCALL call_2_6_9_4(SH2_struct* sh)
{
  SH2movbm(sh,6,9);
}
#define init_2_6_10_4
void FASTCALL call_2_6_10_4(SH2_struct* sh)
{
  SH2movbm(sh,6,10);
}
#define init_2_6_11_4
void FASTCALL call_2_6_11_4(SH2_struct* sh)
{
  SH2movbm(sh,6,11);
}
#define init_2_6_12_4
void FASTCALL call_2_6_12_4(SH2_struct* sh)
{
  SH2movbm(sh,6,12);
}
#define init_2_6_13_4
void FASTCALL call_2_6_13_4(SH2_struct* sh)
{
  SH2movbm(sh,6,13);
}
#define init_2_6_14_4
void FASTCALL call_2_6_14_4(SH2_struct* sh)
{
  SH2movbm(sh,6,14);
}
#define init_2_6_15_4
void FASTCALL call_2_6_15_4(SH2_struct* sh)
{
  SH2movbm(sh,6,15);
}
#define init_2_7_0_4
void FASTCALL call_2_7_0_4(SH2_struct* sh)
{
  SH2movbm(sh,7,0);
}
#define init_2_7_1_4
void FASTCALL call_2_7_1_4(SH2_struct* sh)
{
  SH2movbm(sh,7,1);
}
#define init_2_7_2_4
void FASTCALL call_2_7_2_4(SH2_struct* sh)
{
  SH2movbm(sh,7,2);
}
#define init_2_7_3_4
void FASTCALL call_2_7_3_4(SH2_struct* sh)
{
  SH2movbm(sh,7,3);
}
#define init_2_7_4_4
void FASTCALL call_2_7_4_4(SH2_struct* sh)
{
  SH2movbm(sh,7,4);
}
#define init_2_7_5_4
void FASTCALL call_2_7_5_4(SH2_struct* sh)
{
  SH2movbm(sh,7,5);
}
#define init_2_7_6_4
void FASTCALL call_2_7_6_4(SH2_struct* sh)
{
  SH2movbm(sh,7,6);
}
#define init_2_7_7_4
void FASTCALL call_2_7_7_4(SH2_struct* sh)
{
  SH2movbm(sh,7,7);
}
#define init_2_7_8_4
void FASTCALL call_2_7_8_4(SH2_struct* sh)
{
  SH2movbm(sh,7,8);
}
#define init_2_7_9_4
void FASTCALL call_2_7_9_4(SH2_struct* sh)
{
  SH2movbm(sh,7,9);
}
#define init_2_7_10_4
void FASTCALL call_2_7_10_4(SH2_struct* sh)
{
  SH2movbm(sh,7,10);
}
#define init_2_7_11_4
void FASTCALL call_2_7_11_4(SH2_struct* sh)
{
  SH2movbm(sh,7,11);
}
#define init_2_7_12_4
void FASTCALL call_2_7_12_4(SH2_struct* sh)
{
  SH2movbm(sh,7,12);
}
#define init_2_7_13_4
void FASTCALL call_2_7_13_4(SH2_struct* sh)
{
  SH2movbm(sh,7,13);
}
#define init_2_7_14_4
void FASTCALL call_2_7_14_4(SH2_struct* sh)
{
  SH2movbm(sh,7,14);
}
#define init_2_7_15_4
void FASTCALL call_2_7_15_4(SH2_struct* sh)
{
  SH2movbm(sh,7,15);
}
#define init_2_8_0_4
void FASTCALL call_2_8_0_4(SH2_struct* sh)
{
  SH2movbm(sh,8,0);
}
#define init_2_8_1_4
void FASTCALL call_2_8_1_4(SH2_struct* sh)
{
  SH2movbm(sh,8,1);
}
#define init_2_8_2_4
void FASTCALL call_2_8_2_4(SH2_struct* sh)
{
  SH2movbm(sh,8,2);
}
#define init_2_8_3_4
void FASTCALL call_2_8_3_4(SH2_struct* sh)
{
  SH2movbm(sh,8,3);
}
#define init_2_8_4_4
void FASTCALL call_2_8_4_4(SH2_struct* sh)
{
  SH2movbm(sh,8,4);
}
#define init_2_8_5_4
void FASTCALL call_2_8_5_4(SH2_struct* sh)
{
  SH2movbm(sh,8,5);
}
#define init_2_8_6_4
void FASTCALL call_2_8_6_4(SH2_struct* sh)
{
  SH2movbm(sh,8,6);
}
#define init_2_8_7_4
void FASTCALL call_2_8_7_4(SH2_struct* sh)
{
  SH2movbm(sh,8,7);
}
#define init_2_8_8_4
void FASTCALL call_2_8_8_4(SH2_struct* sh)
{
  SH2movbm(sh,8,8);
}
#define init_2_8_9_4
void FASTCALL call_2_8_9_4(SH2_struct* sh)
{
  SH2movbm(sh,8,9);
}
#define init_2_8_10_4
void FASTCALL call_2_8_10_4(SH2_struct* sh)
{
  SH2movbm(sh,8,10);
}
#define init_2_8_11_4
void FASTCALL call_2_8_11_4(SH2_struct* sh)
{
  SH2movbm(sh,8,11);
}
#define init_2_8_12_4
void FASTCALL call_2_8_12_4(SH2_struct* sh)
{
  SH2movbm(sh,8,12);
}
#define init_2_8_13_4
void FASTCALL call_2_8_13_4(SH2_struct* sh)
{
  SH2movbm(sh,8,13);
}
#define init_2_8_14_4
void FASTCALL call_2_8_14_4(SH2_struct* sh)
{
  SH2movbm(sh,8,14);
}
#define init_2_8_15_4
void FASTCALL call_2_8_15_4(SH2_struct* sh)
{
  SH2movbm(sh,8,15);
}
#define init_2_9_0_4
void FASTCALL call_2_9_0_4(SH2_struct* sh)
{
  SH2movbm(sh,9,0);
}
#define init_2_9_1_4
void FASTCALL call_2_9_1_4(SH2_struct* sh)
{
  SH2movbm(sh,9,1);
}
#define init_2_9_2_4
void FASTCALL call_2_9_2_4(SH2_struct* sh)
{
  SH2movbm(sh,9,2);
}
#define init_2_9_3_4
void FASTCALL call_2_9_3_4(SH2_struct* sh)
{
  SH2movbm(sh,9,3);
}
#define init_2_9_4_4
void FASTCALL call_2_9_4_4(SH2_struct* sh)
{
  SH2movbm(sh,9,4);
}
#define init_2_9_5_4
void FASTCALL call_2_9_5_4(SH2_struct* sh)
{
  SH2movbm(sh,9,5);
}
#define init_2_9_6_4
void FASTCALL call_2_9_6_4(SH2_struct* sh)
{
  SH2movbm(sh,9,6);
}
#define init_2_9_7_4
void FASTCALL call_2_9_7_4(SH2_struct* sh)
{
  SH2movbm(sh,9,7);
}
#define init_2_9_8_4
void FASTCALL call_2_9_8_4(SH2_struct* sh)
{
  SH2movbm(sh,9,8);
}
#define init_2_9_9_4
void FASTCALL call_2_9_9_4(SH2_struct* sh)
{
  SH2movbm(sh,9,9);
}
#define init_2_9_10_4
void FASTCALL call_2_9_10_4(SH2_struct* sh)
{
  SH2movbm(sh,9,10);
}
#define init_2_9_11_4
void FASTCALL call_2_9_11_4(SH2_struct* sh)
{
  SH2movbm(sh,9,11);
}
#define init_2_9_12_4
void FASTCALL call_2_9_12_4(SH2_struct* sh)
{
  SH2movbm(sh,9,12);
}
#define init_2_9_13_4
void FASTCALL call_2_9_13_4(SH2_struct* sh)
{
  SH2movbm(sh,9,13);
}
#define init_2_9_14_4
void FASTCALL call_2_9_14_4(SH2_struct* sh)
{
  SH2movbm(sh,9,14);
}
#define init_2_9_15_4
void FASTCALL call_2_9_15_4(SH2_struct* sh)
{
  SH2movbm(sh,9,15);
}
#define init_2_10_0_4
void FASTCALL call_2_10_0_4(SH2_struct* sh)
{
  SH2movbm(sh,10,0);
}
#define init_2_10_1_4
void FASTCALL call_2_10_1_4(SH2_struct* sh)
{
  SH2movbm(sh,10,1);
}
#define init_2_10_2_4
void FASTCALL call_2_10_2_4(SH2_struct* sh)
{
  SH2movbm(sh,10,2);
}
#define init_2_10_3_4
void FASTCALL call_2_10_3_4(SH2_struct* sh)
{
  SH2movbm(sh,10,3);
}
#define init_2_10_4_4
void FASTCALL call_2_10_4_4(SH2_struct* sh)
{
  SH2movbm(sh,10,4);
}
#define init_2_10_5_4
void FASTCALL call_2_10_5_4(SH2_struct* sh)
{
  SH2movbm(sh,10,5);
}
#define init_2_10_6_4
void FASTCALL call_2_10_6_4(SH2_struct* sh)
{
  SH2movbm(sh,10,6);
}
#define init_2_10_7_4
void FASTCALL call_2_10_7_4(SH2_struct* sh)
{
  SH2movbm(sh,10,7);
}
#define init_2_10_8_4
void FASTCALL call_2_10_8_4(SH2_struct* sh)
{
  SH2movbm(sh,10,8);
}
#define init_2_10_9_4
void FASTCALL call_2_10_9_4(SH2_struct* sh)
{
  SH2movbm(sh,10,9);
}
#define init_2_10_10_4
void FASTCALL call_2_10_10_4(SH2_struct* sh)
{
  SH2movbm(sh,10,10);
}
#define init_2_10_11_4
void FASTCALL call_2_10_11_4(SH2_struct* sh)
{
  SH2movbm(sh,10,11);
}
#define init_2_10_12_4
void FASTCALL call_2_10_12_4(SH2_struct* sh)
{
  SH2movbm(sh,10,12);
}
#define init_2_10_13_4
void FASTCALL call_2_10_13_4(SH2_struct* sh)
{
  SH2movbm(sh,10,13);
}
#define init_2_10_14_4
void FASTCALL call_2_10_14_4(SH2_struct* sh)
{
  SH2movbm(sh,10,14);
}
#define init_2_10_15_4
void FASTCALL call_2_10_15_4(SH2_struct* sh)
{
  SH2movbm(sh,10,15);
}
#define init_2_11_0_4
void FASTCALL call_2_11_0_4(SH2_struct* sh)
{
  SH2movbm(sh,11,0);
}
#define init_2_11_1_4
void FASTCALL call_2_11_1_4(SH2_struct* sh)
{
  SH2movbm(sh,11,1);
}
#define init_2_11_2_4
void FASTCALL call_2_11_2_4(SH2_struct* sh)
{
  SH2movbm(sh,11,2);
}
#define init_2_11_3_4
void FASTCALL call_2_11_3_4(SH2_struct* sh)
{
  SH2movbm(sh,11,3);
}
#define init_2_11_4_4
void FASTCALL call_2_11_4_4(SH2_struct* sh)
{
  SH2movbm(sh,11,4);
}
#define init_2_11_5_4
void FASTCALL call_2_11_5_4(SH2_struct* sh)
{
  SH2movbm(sh,11,5);
}
#define init_2_11_6_4
void FASTCALL call_2_11_6_4(SH2_struct* sh)
{
  SH2movbm(sh,11,6);
}
#define init_2_11_7_4
void FASTCALL call_2_11_7_4(SH2_struct* sh)
{
  SH2movbm(sh,11,7);
}
#define init_2_11_8_4
void FASTCALL call_2_11_8_4(SH2_struct* sh)
{
  SH2movbm(sh,11,8);
}
#define init_2_11_9_4
void FASTCALL call_2_11_9_4(SH2_struct* sh)
{
  SH2movbm(sh,11,9);
}
#define init_2_11_10_4
void FASTCALL call_2_11_10_4(SH2_struct* sh)
{
  SH2movbm(sh,11,10);
}
#define init_2_11_11_4
void FASTCALL call_2_11_11_4(SH2_struct* sh)
{
  SH2movbm(sh,11,11);
}
#define init_2_11_12_4
void FASTCALL call_2_11_12_4(SH2_struct* sh)
{
  SH2movbm(sh,11,12);
}
#define init_2_11_13_4
void FASTCALL call_2_11_13_4(SH2_struct* sh)
{
  SH2movbm(sh,11,13);
}
#define init_2_11_14_4
void FASTCALL call_2_11_14_4(SH2_struct* sh)
{
  SH2movbm(sh,11,14);
}
#define init_2_11_15_4
void FASTCALL call_2_11_15_4(SH2_struct* sh)
{
  SH2movbm(sh,11,15);
}
#define init_2_12_0_4
void FASTCALL call_2_12_0_4(SH2_struct* sh)
{
  SH2movbm(sh,12,0);
}
#define init_2_12_1_4
void FASTCALL call_2_12_1_4(SH2_struct* sh)
{
  SH2movbm(sh,12,1);
}
#define init_2_12_2_4
void FASTCALL call_2_12_2_4(SH2_struct* sh)
{
  SH2movbm(sh,12,2);
}
#define init_2_12_3_4
void FASTCALL call_2_12_3_4(SH2_struct* sh)
{
  SH2movbm(sh,12,3);
}
#define init_2_12_4_4
void FASTCALL call_2_12_4_4(SH2_struct* sh)
{
  SH2movbm(sh,12,4);
}
#define init_2_12_5_4
void FASTCALL call_2_12_5_4(SH2_struct* sh)
{
  SH2movbm(sh,12,5);
}
#define init_2_12_6_4
void FASTCALL call_2_12_6_4(SH2_struct* sh)
{
  SH2movbm(sh,12,6);
}
#define init_2_12_7_4
void FASTCALL call_2_12_7_4(SH2_struct* sh)
{
  SH2movbm(sh,12,7);
}
#define init_2_12_8_4
void FASTCALL call_2_12_8_4(SH2_struct* sh)
{
  SH2movbm(sh,12,8);
}
#define init_2_12_9_4
void FASTCALL call_2_12_9_4(SH2_struct* sh)
{
  SH2movbm(sh,12,9);
}
#define init_2_12_10_4
void FASTCALL call_2_12_10_4(SH2_struct* sh)
{
  SH2movbm(sh,12,10);
}
#define init_2_12_11_4
void FASTCALL call_2_12_11_4(SH2_struct* sh)
{
  SH2movbm(sh,12,11);
}
#define init_2_12_12_4
void FASTCALL call_2_12_12_4(SH2_struct* sh)
{
  SH2movbm(sh,12,12);
}
#define init_2_12_13_4
void FASTCALL call_2_12_13_4(SH2_struct* sh)
{
  SH2movbm(sh,12,13);
}
#define init_2_12_14_4
void FASTCALL call_2_12_14_4(SH2_struct* sh)
{
  SH2movbm(sh,12,14);
}
#define init_2_12_15_4
void FASTCALL call_2_12_15_4(SH2_struct* sh)
{
  SH2movbm(sh,12,15);
}
#define init_2_13_0_4
void FASTCALL call_2_13_0_4(SH2_struct* sh)
{
  SH2movbm(sh,13,0);
}
#define init_2_13_1_4
void FASTCALL call_2_13_1_4(SH2_struct* sh)
{
  SH2movbm(sh,13,1);
}
#define init_2_13_2_4
void FASTCALL call_2_13_2_4(SH2_struct* sh)
{
  SH2movbm(sh,13,2);
}
#define init_2_13_3_4
void FASTCALL call_2_13_3_4(SH2_struct* sh)
{
  SH2movbm(sh,13,3);
}
#define init_2_13_4_4
void FASTCALL call_2_13_4_4(SH2_struct* sh)
{
  SH2movbm(sh,13,4);
}
#define init_2_13_5_4
void FASTCALL call_2_13_5_4(SH2_struct* sh)
{
  SH2movbm(sh,13,5);
}
#define init_2_13_6_4
void FASTCALL call_2_13_6_4(SH2_struct* sh)
{
  SH2movbm(sh,13,6);
}
#define init_2_13_7_4
void FASTCALL call_2_13_7_4(SH2_struct* sh)
{
  SH2movbm(sh,13,7);
}
#define init_2_13_8_4
void FASTCALL call_2_13_8_4(SH2_struct* sh)
{
  SH2movbm(sh,13,8);
}
#define init_2_13_9_4
void FASTCALL call_2_13_9_4(SH2_struct* sh)
{
  SH2movbm(sh,13,9);
}
#define init_2_13_10_4
void FASTCALL call_2_13_10_4(SH2_struct* sh)
{
  SH2movbm(sh,13,10);
}
#define init_2_13_11_4
void FASTCALL call_2_13_11_4(SH2_struct* sh)
{
  SH2movbm(sh,13,11);
}
#define init_2_13_12_4
void FASTCALL call_2_13_12_4(SH2_struct* sh)
{
  SH2movbm(sh,13,12);
}
#define init_2_13_13_4
void FASTCALL call_2_13_13_4(SH2_struct* sh)
{
  SH2movbm(sh,13,13);
}
#define init_2_13_14_4
void FASTCALL call_2_13_14_4(SH2_struct* sh)
{
  SH2movbm(sh,13,14);
}
#define init_2_13_15_4
void FASTCALL call_2_13_15_4(SH2_struct* sh)
{
  SH2movbm(sh,13,15);
}
#define init_2_14_0_4
void FASTCALL call_2_14_0_4(SH2_struct* sh)
{
  SH2movbm(sh,14,0);
}
#define init_2_14_1_4
void FASTCALL call_2_14_1_4(SH2_struct* sh)
{
  SH2movbm(sh,14,1);
}
#define init_2_14_2_4
void FASTCALL call_2_14_2_4(SH2_struct* sh)
{
  SH2movbm(sh,14,2);
}
#define init_2_14_3_4
void FASTCALL call_2_14_3_4(SH2_struct* sh)
{
  SH2movbm(sh,14,3);
}
#define init_2_14_4_4
void FASTCALL call_2_14_4_4(SH2_struct* sh)
{
  SH2movbm(sh,14,4);
}
#define init_2_14_5_4
void FASTCALL call_2_14_5_4(SH2_struct* sh)
{
  SH2movbm(sh,14,5);
}
#define init_2_14_6_4
void FASTCALL call_2_14_6_4(SH2_struct* sh)
{
  SH2movbm(sh,14,6);
}
#define init_2_14_7_4
void FASTCALL call_2_14_7_4(SH2_struct* sh)
{
  SH2movbm(sh,14,7);
}
#define init_2_14_8_4
void FASTCALL call_2_14_8_4(SH2_struct* sh)
{
  SH2movbm(sh,14,8);
}
#define init_2_14_9_4
void FASTCALL call_2_14_9_4(SH2_struct* sh)
{
  SH2movbm(sh,14,9);
}
#define init_2_14_10_4
void FASTCALL call_2_14_10_4(SH2_struct* sh)
{
  SH2movbm(sh,14,10);
}
#define init_2_14_11_4
void FASTCALL call_2_14_11_4(SH2_struct* sh)
{
  SH2movbm(sh,14,11);
}
#define init_2_14_12_4
void FASTCALL call_2_14_12_4(SH2_struct* sh)
{
  SH2movbm(sh,14,12);
}
#define init_2_14_13_4
void FASTCALL call_2_14_13_4(SH2_struct* sh)
{
  SH2movbm(sh,14,13);
}
#define init_2_14_14_4
void FASTCALL call_2_14_14_4(SH2_struct* sh)
{
  SH2movbm(sh,14,14);
}
#define init_2_14_15_4
void FASTCALL call_2_14_15_4(SH2_struct* sh)
{
  SH2movbm(sh,14,15);
}
#define init_2_15_0_4
void FASTCALL call_2_15_0_4(SH2_struct* sh)
{
  SH2movbm(sh,15,0);
}
#define init_2_15_1_4
void FASTCALL call_2_15_1_4(SH2_struct* sh)
{
  SH2movbm(sh,15,1);
}
#define init_2_15_2_4
void FASTCALL call_2_15_2_4(SH2_struct* sh)
{
  SH2movbm(sh,15,2);
}
#define init_2_15_3_4
void FASTCALL call_2_15_3_4(SH2_struct* sh)
{
  SH2movbm(sh,15,3);
}
#define init_2_15_4_4
void FASTCALL call_2_15_4_4(SH2_struct* sh)
{
  SH2movbm(sh,15,4);
}
#define init_2_15_5_4
void FASTCALL call_2_15_5_4(SH2_struct* sh)
{
  SH2movbm(sh,15,5);
}
#define init_2_15_6_4
void FASTCALL call_2_15_6_4(SH2_struct* sh)
{
  SH2movbm(sh,15,6);
}
#define init_2_15_7_4
void FASTCALL call_2_15_7_4(SH2_struct* sh)
{
  SH2movbm(sh,15,7);
}
#define init_2_15_8_4
void FASTCALL call_2_15_8_4(SH2_struct* sh)
{
  SH2movbm(sh,15,8);
}
#define init_2_15_9_4
void FASTCALL call_2_15_9_4(SH2_struct* sh)
{
  SH2movbm(sh,15,9);
}
#define init_2_15_10_4
void FASTCALL call_2_15_10_4(SH2_struct* sh)
{
  SH2movbm(sh,15,10);
}
#define init_2_15_11_4
void FASTCALL call_2_15_11_4(SH2_struct* sh)
{
  SH2movbm(sh,15,11);
}
#define init_2_15_12_4
void FASTCALL call_2_15_12_4(SH2_struct* sh)
{
  SH2movbm(sh,15,12);
}
#define init_2_15_13_4
void FASTCALL call_2_15_13_4(SH2_struct* sh)
{
  SH2movbm(sh,15,13);
}
#define init_2_15_14_4
void FASTCALL call_2_15_14_4(SH2_struct* sh)
{
  SH2movbm(sh,15,14);
}
#define init_2_15_15_4
void FASTCALL call_2_15_15_4(SH2_struct* sh)
{
  SH2movbm(sh,15,15);
}
#define init_12_4_0_0
void FASTCALL call_12_4_0_0(SH2_struct* sh)
{
  SH2movblg(sh,(0<<4)+0);
}
#define init_12_4_0_1
void FASTCALL call_12_4_0_1(SH2_struct* sh)
{
  SH2movblg(sh,(0<<4)+1);
}
#define init_12_4_0_2
void FASTCALL call_12_4_0_2(SH2_struct* sh)
{
  SH2movblg(sh,(0<<4)+2);
}
#define init_12_4_0_3
void FASTCALL call_12_4_0_3(SH2_struct* sh)
{
  SH2movblg(sh,(0<<4)+3);
}
#define init_12_4_0_4
void FASTCALL call_12_4_0_4(SH2_struct* sh)
{
  SH2movblg(sh,(0<<4)+4);
}
#define init_12_4_0_5
void FASTCALL call_12_4_0_5(SH2_struct* sh)
{
  SH2movblg(sh,(0<<4)+5);
}
#define init_12_4_0_6
void FASTCALL call_12_4_0_6(SH2_struct* sh)
{
  SH2movblg(sh,(0<<4)+6);
}
#define init_12_4_0_7
void FASTCALL call_12_4_0_7(SH2_struct* sh)
{
  SH2movblg(sh,(0<<4)+7);
}
#define init_12_4_0_8
void FASTCALL call_12_4_0_8(SH2_struct* sh)
{
  SH2movblg(sh,(0<<4)+8);
}
#define init_12_4_0_9
void FASTCALL call_12_4_0_9(SH2_struct* sh)
{
  SH2movblg(sh,(0<<4)+9);
}
#define init_12_4_0_10
void FASTCALL call_12_4_0_10(SH2_struct* sh)
{
  SH2movblg(sh,(0<<4)+10);
}
#define init_12_4_0_11
void FASTCALL call_12_4_0_11(SH2_struct* sh)
{
  SH2movblg(sh,(0<<4)+11);
}
#define init_12_4_0_12
void FASTCALL call_12_4_0_12(SH2_struct* sh)
{
  SH2movblg(sh,(0<<4)+12);
}
#define init_12_4_0_13
void FASTCALL call_12_4_0_13(SH2_struct* sh)
{
  SH2movblg(sh,(0<<4)+13);
}
#define init_12_4_0_14
void FASTCALL call_12_4_0_14(SH2_struct* sh)
{
  SH2movblg(sh,(0<<4)+14);
}
#define init_12_4_0_15
void FASTCALL call_12_4_0_15(SH2_struct* sh)
{
  SH2movblg(sh,(0<<4)+15);
}
#define init_12_4_1_0
void FASTCALL call_12_4_1_0(SH2_struct* sh)
{
  SH2movblg(sh,(1<<4)+0);
}
#define init_12_4_1_1
void FASTCALL call_12_4_1_1(SH2_struct* sh)
{
  SH2movblg(sh,(1<<4)+1);
}
#define init_12_4_1_2
void FASTCALL call_12_4_1_2(SH2_struct* sh)
{
  SH2movblg(sh,(1<<4)+2);
}
#define init_12_4_1_3
void FASTCALL call_12_4_1_3(SH2_struct* sh)
{
  SH2movblg(sh,(1<<4)+3);
}
#define init_12_4_1_4
void FASTCALL call_12_4_1_4(SH2_struct* sh)
{
  SH2movblg(sh,(1<<4)+4);
}
#define init_12_4_1_5
void FASTCALL call_12_4_1_5(SH2_struct* sh)
{
  SH2movblg(sh,(1<<4)+5);
}
#define init_12_4_1_6
void FASTCALL call_12_4_1_6(SH2_struct* sh)
{
  SH2movblg(sh,(1<<4)+6);
}
#define init_12_4_1_7
void FASTCALL call_12_4_1_7(SH2_struct* sh)
{
  SH2movblg(sh,(1<<4)+7);
}
#define init_12_4_1_8
void FASTCALL call_12_4_1_8(SH2_struct* sh)
{
  SH2movblg(sh,(1<<4)+8);
}
#define init_12_4_1_9
void FASTCALL call_12_4_1_9(SH2_struct* sh)
{
  SH2movblg(sh,(1<<4)+9);
}
#define init_12_4_1_10
void FASTCALL call_12_4_1_10(SH2_struct* sh)
{
  SH2movblg(sh,(1<<4)+10);
}
#define init_12_4_1_11
void FASTCALL call_12_4_1_11(SH2_struct* sh)
{
  SH2movblg(sh,(1<<4)+11);
}
#define init_12_4_1_12
void FASTCALL call_12_4_1_12(SH2_struct* sh)
{
  SH2movblg(sh,(1<<4)+12);
}
#define init_12_4_1_13
void FASTCALL call_12_4_1_13(SH2_struct* sh)
{
  SH2movblg(sh,(1<<4)+13);
}
#define init_12_4_1_14
void FASTCALL call_12_4_1_14(SH2_struct* sh)
{
  SH2movblg(sh,(1<<4)+14);
}
#define init_12_4_1_15
void FASTCALL call_12_4_1_15(SH2_struct* sh)
{
  SH2movblg(sh,(1<<4)+15);
}
#define init_12_4_2_0
void FASTCALL call_12_4_2_0(SH2_struct* sh)
{
  SH2movblg(sh,(2<<4)+0);
}
#define init_12_4_2_1
void FASTCALL call_12_4_2_1(SH2_struct* sh)
{
  SH2movblg(sh,(2<<4)+1);
}
#define init_12_4_2_2
void FASTCALL call_12_4_2_2(SH2_struct* sh)
{
  SH2movblg(sh,(2<<4)+2);
}
#define init_12_4_2_3
void FASTCALL call_12_4_2_3(SH2_struct* sh)
{
  SH2movblg(sh,(2<<4)+3);
}
#define init_12_4_2_4
void FASTCALL call_12_4_2_4(SH2_struct* sh)
{
  SH2movblg(sh,(2<<4)+4);
}
#define init_12_4_2_5
void FASTCALL call_12_4_2_5(SH2_struct* sh)
{
  SH2movblg(sh,(2<<4)+5);
}
#define init_12_4_2_6
void FASTCALL call_12_4_2_6(SH2_struct* sh)
{
  SH2movblg(sh,(2<<4)+6);
}
#define init_12_4_2_7
void FASTCALL call_12_4_2_7(SH2_struct* sh)
{
  SH2movblg(sh,(2<<4)+7);
}
#define init_12_4_2_8
void FASTCALL call_12_4_2_8(SH2_struct* sh)
{
  SH2movblg(sh,(2<<4)+8);
}
#define init_12_4_2_9
void FASTCALL call_12_4_2_9(SH2_struct* sh)
{
  SH2movblg(sh,(2<<4)+9);
}
#define init_12_4_2_10
void FASTCALL call_12_4_2_10(SH2_struct* sh)
{
  SH2movblg(sh,(2<<4)+10);
}
#define init_12_4_2_11
void FASTCALL call_12_4_2_11(SH2_struct* sh)
{
  SH2movblg(sh,(2<<4)+11);
}
#define init_12_4_2_12
void FASTCALL call_12_4_2_12(SH2_struct* sh)
{
  SH2movblg(sh,(2<<4)+12);
}
#define init_12_4_2_13
void FASTCALL call_12_4_2_13(SH2_struct* sh)
{
  SH2movblg(sh,(2<<4)+13);
}
#define init_12_4_2_14
void FASTCALL call_12_4_2_14(SH2_struct* sh)
{
  SH2movblg(sh,(2<<4)+14);
}
#define init_12_4_2_15
void FASTCALL call_12_4_2_15(SH2_struct* sh)
{
  SH2movblg(sh,(2<<4)+15);
}
#define init_12_4_3_0
void FASTCALL call_12_4_3_0(SH2_struct* sh)
{
  SH2movblg(sh,(3<<4)+0);
}
#define init_12_4_3_1
void FASTCALL call_12_4_3_1(SH2_struct* sh)
{
  SH2movblg(sh,(3<<4)+1);
}
#define init_12_4_3_2
void FASTCALL call_12_4_3_2(SH2_struct* sh)
{
  SH2movblg(sh,(3<<4)+2);
}
#define init_12_4_3_3
void FASTCALL call_12_4_3_3(SH2_struct* sh)
{
  SH2movblg(sh,(3<<4)+3);
}
#define init_12_4_3_4
void FASTCALL call_12_4_3_4(SH2_struct* sh)
{
  SH2movblg(sh,(3<<4)+4);
}
#define init_12_4_3_5
void FASTCALL call_12_4_3_5(SH2_struct* sh)
{
  SH2movblg(sh,(3<<4)+5);
}
#define init_12_4_3_6
void FASTCALL call_12_4_3_6(SH2_struct* sh)
{
  SH2movblg(sh,(3<<4)+6);
}
#define init_12_4_3_7
void FASTCALL call_12_4_3_7(SH2_struct* sh)
{
  SH2movblg(sh,(3<<4)+7);
}
#define init_12_4_3_8
void FASTCALL call_12_4_3_8(SH2_struct* sh)
{
  SH2movblg(sh,(3<<4)+8);
}
#define init_12_4_3_9
void FASTCALL call_12_4_3_9(SH2_struct* sh)
{
  SH2movblg(sh,(3<<4)+9);
}
#define init_12_4_3_10
void FASTCALL call_12_4_3_10(SH2_struct* sh)
{
  SH2movblg(sh,(3<<4)+10);
}
#define init_12_4_3_11
void FASTCALL call_12_4_3_11(SH2_struct* sh)
{
  SH2movblg(sh,(3<<4)+11);
}
#define init_12_4_3_12
void FASTCALL call_12_4_3_12(SH2_struct* sh)
{
  SH2movblg(sh,(3<<4)+12);
}
#define init_12_4_3_13
void FASTCALL call_12_4_3_13(SH2_struct* sh)
{
  SH2movblg(sh,(3<<4)+13);
}
#define init_12_4_3_14
void FASTCALL call_12_4_3_14(SH2_struct* sh)
{
  SH2movblg(sh,(3<<4)+14);
}
#define init_12_4_3_15
void FASTCALL call_12_4_3_15(SH2_struct* sh)
{
  SH2movblg(sh,(3<<4)+15);
}
#define init_12_4_4_0
void FASTCALL call_12_4_4_0(SH2_struct* sh)
{
  SH2movblg(sh,(4<<4)+0);
}
#define init_12_4_4_1
void FASTCALL call_12_4_4_1(SH2_struct* sh)
{
  SH2movblg(sh,(4<<4)+1);
}
#define init_12_4_4_2
void FASTCALL call_12_4_4_2(SH2_struct* sh)
{
  SH2movblg(sh,(4<<4)+2);
}
#define init_12_4_4_3
void FASTCALL call_12_4_4_3(SH2_struct* sh)
{
  SH2movblg(sh,(4<<4)+3);
}
#define init_12_4_4_4
void FASTCALL call_12_4_4_4(SH2_struct* sh)
{
  SH2movblg(sh,(4<<4)+4);
}
#define init_12_4_4_5
void FASTCALL call_12_4_4_5(SH2_struct* sh)
{
  SH2movblg(sh,(4<<4)+5);
}
#define init_12_4_4_6
void FASTCALL call_12_4_4_6(SH2_struct* sh)
{
  SH2movblg(sh,(4<<4)+6);
}
#define init_12_4_4_7
void FASTCALL call_12_4_4_7(SH2_struct* sh)
{
  SH2movblg(sh,(4<<4)+7);
}
#define init_12_4_4_8
void FASTCALL call_12_4_4_8(SH2_struct* sh)
{
  SH2movblg(sh,(4<<4)+8);
}
#define init_12_4_4_9
void FASTCALL call_12_4_4_9(SH2_struct* sh)
{
  SH2movblg(sh,(4<<4)+9);
}
#define init_12_4_4_10
void FASTCALL call_12_4_4_10(SH2_struct* sh)
{
  SH2movblg(sh,(4<<4)+10);
}
#define init_12_4_4_11
void FASTCALL call_12_4_4_11(SH2_struct* sh)
{
  SH2movblg(sh,(4<<4)+11);
}
#define init_12_4_4_12
void FASTCALL call_12_4_4_12(SH2_struct* sh)
{
  SH2movblg(sh,(4<<4)+12);
}
#define init_12_4_4_13
void FASTCALL call_12_4_4_13(SH2_struct* sh)
{
  SH2movblg(sh,(4<<4)+13);
}
#define init_12_4_4_14
void FASTCALL call_12_4_4_14(SH2_struct* sh)
{
  SH2movblg(sh,(4<<4)+14);
}
#define init_12_4_4_15
void FASTCALL call_12_4_4_15(SH2_struct* sh)
{
  SH2movblg(sh,(4<<4)+15);
}
#define init_12_4_5_0
void FASTCALL call_12_4_5_0(SH2_struct* sh)
{
  SH2movblg(sh,(5<<4)+0);
}
#define init_12_4_5_1
void FASTCALL call_12_4_5_1(SH2_struct* sh)
{
  SH2movblg(sh,(5<<4)+1);
}
#define init_12_4_5_2
void FASTCALL call_12_4_5_2(SH2_struct* sh)
{
  SH2movblg(sh,(5<<4)+2);
}
#define init_12_4_5_3
void FASTCALL call_12_4_5_3(SH2_struct* sh)
{
  SH2movblg(sh,(5<<4)+3);
}
#define init_12_4_5_4
void FASTCALL call_12_4_5_4(SH2_struct* sh)
{
  SH2movblg(sh,(5<<4)+4);
}
#define init_12_4_5_5
void FASTCALL call_12_4_5_5(SH2_struct* sh)
{
  SH2movblg(sh,(5<<4)+5);
}
#define init_12_4_5_6
void FASTCALL call_12_4_5_6(SH2_struct* sh)
{
  SH2movblg(sh,(5<<4)+6);
}
#define init_12_4_5_7
void FASTCALL call_12_4_5_7(SH2_struct* sh)
{
  SH2movblg(sh,(5<<4)+7);
}
#define init_12_4_5_8
void FASTCALL call_12_4_5_8(SH2_struct* sh)
{
  SH2movblg(sh,(5<<4)+8);
}
#define init_12_4_5_9
void FASTCALL call_12_4_5_9(SH2_struct* sh)
{
  SH2movblg(sh,(5<<4)+9);
}
#define init_12_4_5_10
void FASTCALL call_12_4_5_10(SH2_struct* sh)
{
  SH2movblg(sh,(5<<4)+10);
}
#define init_12_4_5_11
void FASTCALL call_12_4_5_11(SH2_struct* sh)
{
  SH2movblg(sh,(5<<4)+11);
}
#define init_12_4_5_12
void FASTCALL call_12_4_5_12(SH2_struct* sh)
{
  SH2movblg(sh,(5<<4)+12);
}
#define init_12_4_5_13
void FASTCALL call_12_4_5_13(SH2_struct* sh)
{
  SH2movblg(sh,(5<<4)+13);
}
#define init_12_4_5_14
void FASTCALL call_12_4_5_14(SH2_struct* sh)
{
  SH2movblg(sh,(5<<4)+14);
}
#define init_12_4_5_15
void FASTCALL call_12_4_5_15(SH2_struct* sh)
{
  SH2movblg(sh,(5<<4)+15);
}
#define init_12_4_6_0
void FASTCALL call_12_4_6_0(SH2_struct* sh)
{
  SH2movblg(sh,(6<<4)+0);
}
#define init_12_4_6_1
void FASTCALL call_12_4_6_1(SH2_struct* sh)
{
  SH2movblg(sh,(6<<4)+1);
}
#define init_12_4_6_2
void FASTCALL call_12_4_6_2(SH2_struct* sh)
{
  SH2movblg(sh,(6<<4)+2);
}
#define init_12_4_6_3
void FASTCALL call_12_4_6_3(SH2_struct* sh)
{
  SH2movblg(sh,(6<<4)+3);
}
#define init_12_4_6_4
void FASTCALL call_12_4_6_4(SH2_struct* sh)
{
  SH2movblg(sh,(6<<4)+4);
}
#define init_12_4_6_5
void FASTCALL call_12_4_6_5(SH2_struct* sh)
{
  SH2movblg(sh,(6<<4)+5);
}
#define init_12_4_6_6
void FASTCALL call_12_4_6_6(SH2_struct* sh)
{
  SH2movblg(sh,(6<<4)+6);
}
#define init_12_4_6_7
void FASTCALL call_12_4_6_7(SH2_struct* sh)
{
  SH2movblg(sh,(6<<4)+7);
}
#define init_12_4_6_8
void FASTCALL call_12_4_6_8(SH2_struct* sh)
{
  SH2movblg(sh,(6<<4)+8);
}
#define init_12_4_6_9
void FASTCALL call_12_4_6_9(SH2_struct* sh)
{
  SH2movblg(sh,(6<<4)+9);
}
#define init_12_4_6_10
void FASTCALL call_12_4_6_10(SH2_struct* sh)
{
  SH2movblg(sh,(6<<4)+10);
}
#define init_12_4_6_11
void FASTCALL call_12_4_6_11(SH2_struct* sh)
{
  SH2movblg(sh,(6<<4)+11);
}
#define init_12_4_6_12
void FASTCALL call_12_4_6_12(SH2_struct* sh)
{
  SH2movblg(sh,(6<<4)+12);
}
#define init_12_4_6_13
void FASTCALL call_12_4_6_13(SH2_struct* sh)
{
  SH2movblg(sh,(6<<4)+13);
}
#define init_12_4_6_14
void FASTCALL call_12_4_6_14(SH2_struct* sh)
{
  SH2movblg(sh,(6<<4)+14);
}
#define init_12_4_6_15
void FASTCALL call_12_4_6_15(SH2_struct* sh)
{
  SH2movblg(sh,(6<<4)+15);
}
#define init_12_4_7_0
void FASTCALL call_12_4_7_0(SH2_struct* sh)
{
  SH2movblg(sh,(7<<4)+0);
}
#define init_12_4_7_1
void FASTCALL call_12_4_7_1(SH2_struct* sh)
{
  SH2movblg(sh,(7<<4)+1);
}
#define init_12_4_7_2
void FASTCALL call_12_4_7_2(SH2_struct* sh)
{
  SH2movblg(sh,(7<<4)+2);
}
#define init_12_4_7_3
void FASTCALL call_12_4_7_3(SH2_struct* sh)
{
  SH2movblg(sh,(7<<4)+3);
}
#define init_12_4_7_4
void FASTCALL call_12_4_7_4(SH2_struct* sh)
{
  SH2movblg(sh,(7<<4)+4);
}
#define init_12_4_7_5
void FASTCALL call_12_4_7_5(SH2_struct* sh)
{
  SH2movblg(sh,(7<<4)+5);
}
#define init_12_4_7_6
void FASTCALL call_12_4_7_6(SH2_struct* sh)
{
  SH2movblg(sh,(7<<4)+6);
}
#define init_12_4_7_7
void FASTCALL call_12_4_7_7(SH2_struct* sh)
{
  SH2movblg(sh,(7<<4)+7);
}
#define init_12_4_7_8
void FASTCALL call_12_4_7_8(SH2_struct* sh)
{
  SH2movblg(sh,(7<<4)+8);
}
#define init_12_4_7_9
void FASTCALL call_12_4_7_9(SH2_struct* sh)
{
  SH2movblg(sh,(7<<4)+9);
}
#define init_12_4_7_10
void FASTCALL call_12_4_7_10(SH2_struct* sh)
{
  SH2movblg(sh,(7<<4)+10);
}
#define init_12_4_7_11
void FASTCALL call_12_4_7_11(SH2_struct* sh)
{
  SH2movblg(sh,(7<<4)+11);
}
#define init_12_4_7_12
void FASTCALL call_12_4_7_12(SH2_struct* sh)
{
  SH2movblg(sh,(7<<4)+12);
}
#define init_12_4_7_13
void FASTCALL call_12_4_7_13(SH2_struct* sh)
{
  SH2movblg(sh,(7<<4)+13);
}
#define init_12_4_7_14
void FASTCALL call_12_4_7_14(SH2_struct* sh)
{
  SH2movblg(sh,(7<<4)+14);
}
#define init_12_4_7_15
void FASTCALL call_12_4_7_15(SH2_struct* sh)
{
  SH2movblg(sh,(7<<4)+15);
}
#define init_12_4_8_0
void FASTCALL call_12_4_8_0(SH2_struct* sh)
{
  SH2movblg(sh,(8<<4)+0);
}
#define init_12_4_8_1
void FASTCALL call_12_4_8_1(SH2_struct* sh)
{
  SH2movblg(sh,(8<<4)+1);
}
#define init_12_4_8_2
void FASTCALL call_12_4_8_2(SH2_struct* sh)
{
  SH2movblg(sh,(8<<4)+2);
}
#define init_12_4_8_3
void FASTCALL call_12_4_8_3(SH2_struct* sh)
{
  SH2movblg(sh,(8<<4)+3);
}
#define init_12_4_8_4
void FASTCALL call_12_4_8_4(SH2_struct* sh)
{
  SH2movblg(sh,(8<<4)+4);
}
#define init_12_4_8_5
void FASTCALL call_12_4_8_5(SH2_struct* sh)
{
  SH2movblg(sh,(8<<4)+5);
}
#define init_12_4_8_6
void FASTCALL call_12_4_8_6(SH2_struct* sh)
{
  SH2movblg(sh,(8<<4)+6);
}
#define init_12_4_8_7
void FASTCALL call_12_4_8_7(SH2_struct* sh)
{
  SH2movblg(sh,(8<<4)+7);
}
#define init_12_4_8_8
void FASTCALL call_12_4_8_8(SH2_struct* sh)
{
  SH2movblg(sh,(8<<4)+8);
}
#define init_12_4_8_9
void FASTCALL call_12_4_8_9(SH2_struct* sh)
{
  SH2movblg(sh,(8<<4)+9);
}
#define init_12_4_8_10
void FASTCALL call_12_4_8_10(SH2_struct* sh)
{
  SH2movblg(sh,(8<<4)+10);
}
#define init_12_4_8_11
void FASTCALL call_12_4_8_11(SH2_struct* sh)
{
  SH2movblg(sh,(8<<4)+11);
}
#define init_12_4_8_12
void FASTCALL call_12_4_8_12(SH2_struct* sh)
{
  SH2movblg(sh,(8<<4)+12);
}
#define init_12_4_8_13
void FASTCALL call_12_4_8_13(SH2_struct* sh)
{
  SH2movblg(sh,(8<<4)+13);
}
#define init_12_4_8_14
void FASTCALL call_12_4_8_14(SH2_struct* sh)
{
  SH2movblg(sh,(8<<4)+14);
}
#define init_12_4_8_15
void FASTCALL call_12_4_8_15(SH2_struct* sh)
{
  SH2movblg(sh,(8<<4)+15);
}
#define init_12_4_9_0
void FASTCALL call_12_4_9_0(SH2_struct* sh)
{
  SH2movblg(sh,(9<<4)+0);
}
#define init_12_4_9_1
void FASTCALL call_12_4_9_1(SH2_struct* sh)
{
  SH2movblg(sh,(9<<4)+1);
}
#define init_12_4_9_2
void FASTCALL call_12_4_9_2(SH2_struct* sh)
{
  SH2movblg(sh,(9<<4)+2);
}
#define init_12_4_9_3
void FASTCALL call_12_4_9_3(SH2_struct* sh)
{
  SH2movblg(sh,(9<<4)+3);
}
#define init_12_4_9_4
void FASTCALL call_12_4_9_4(SH2_struct* sh)
{
  SH2movblg(sh,(9<<4)+4);
}
#define init_12_4_9_5
void FASTCALL call_12_4_9_5(SH2_struct* sh)
{
  SH2movblg(sh,(9<<4)+5);
}
#define init_12_4_9_6
void FASTCALL call_12_4_9_6(SH2_struct* sh)
{
  SH2movblg(sh,(9<<4)+6);
}
#define init_12_4_9_7
void FASTCALL call_12_4_9_7(SH2_struct* sh)
{
  SH2movblg(sh,(9<<4)+7);
}
#define init_12_4_9_8
void FASTCALL call_12_4_9_8(SH2_struct* sh)
{
  SH2movblg(sh,(9<<4)+8);
}
#define init_12_4_9_9
void FASTCALL call_12_4_9_9(SH2_struct* sh)
{
  SH2movblg(sh,(9<<4)+9);
}
#define init_12_4_9_10
void FASTCALL call_12_4_9_10(SH2_struct* sh)
{
  SH2movblg(sh,(9<<4)+10);
}
#define init_12_4_9_11
void FASTCALL call_12_4_9_11(SH2_struct* sh)
{
  SH2movblg(sh,(9<<4)+11);
}
#define init_12_4_9_12
void FASTCALL call_12_4_9_12(SH2_struct* sh)
{
  SH2movblg(sh,(9<<4)+12);
}
#define init_12_4_9_13
void FASTCALL call_12_4_9_13(SH2_struct* sh)
{
  SH2movblg(sh,(9<<4)+13);
}
#define init_12_4_9_14
void FASTCALL call_12_4_9_14(SH2_struct* sh)
{
  SH2movblg(sh,(9<<4)+14);
}
#define init_12_4_9_15
void FASTCALL call_12_4_9_15(SH2_struct* sh)
{
  SH2movblg(sh,(9<<4)+15);
}
#define init_12_4_10_0
void FASTCALL call_12_4_10_0(SH2_struct* sh)
{
  SH2movblg(sh,(10<<4)+0);
}
#define init_12_4_10_1
void FASTCALL call_12_4_10_1(SH2_struct* sh)
{
  SH2movblg(sh,(10<<4)+1);
}
#define init_12_4_10_2
void FASTCALL call_12_4_10_2(SH2_struct* sh)
{
  SH2movblg(sh,(10<<4)+2);
}
#define init_12_4_10_3
void FASTCALL call_12_4_10_3(SH2_struct* sh)
{
  SH2movblg(sh,(10<<4)+3);
}
#define init_12_4_10_4
void FASTCALL call_12_4_10_4(SH2_struct* sh)
{
  SH2movblg(sh,(10<<4)+4);
}
#define init_12_4_10_5
void FASTCALL call_12_4_10_5(SH2_struct* sh)
{
  SH2movblg(sh,(10<<4)+5);
}
#define init_12_4_10_6
void FASTCALL call_12_4_10_6(SH2_struct* sh)
{
  SH2movblg(sh,(10<<4)+6);
}
#define init_12_4_10_7
void FASTCALL call_12_4_10_7(SH2_struct* sh)
{
  SH2movblg(sh,(10<<4)+7);
}
#define init_12_4_10_8
void FASTCALL call_12_4_10_8(SH2_struct* sh)
{
  SH2movblg(sh,(10<<4)+8);
}
#define init_12_4_10_9
void FASTCALL call_12_4_10_9(SH2_struct* sh)
{
  SH2movblg(sh,(10<<4)+9);
}
#define init_12_4_10_10
void FASTCALL call_12_4_10_10(SH2_struct* sh)
{
  SH2movblg(sh,(10<<4)+10);
}
#define init_12_4_10_11
void FASTCALL call_12_4_10_11(SH2_struct* sh)
{
  SH2movblg(sh,(10<<4)+11);
}
#define init_12_4_10_12
void FASTCALL call_12_4_10_12(SH2_struct* sh)
{
  SH2movblg(sh,(10<<4)+12);
}
#define init_12_4_10_13
void FASTCALL call_12_4_10_13(SH2_struct* sh)
{
  SH2movblg(sh,(10<<4)+13);
}
#define init_12_4_10_14
void FASTCALL call_12_4_10_14(SH2_struct* sh)
{
  SH2movblg(sh,(10<<4)+14);
}
#define init_12_4_10_15
void FASTCALL call_12_4_10_15(SH2_struct* sh)
{
  SH2movblg(sh,(10<<4)+15);
}
#define init_12_4_11_0
void FASTCALL call_12_4_11_0(SH2_struct* sh)
{
  SH2movblg(sh,(11<<4)+0);
}
#define init_12_4_11_1
void FASTCALL call_12_4_11_1(SH2_struct* sh)
{
  SH2movblg(sh,(11<<4)+1);
}
#define init_12_4_11_2
void FASTCALL call_12_4_11_2(SH2_struct* sh)
{
  SH2movblg(sh,(11<<4)+2);
}
#define init_12_4_11_3
void FASTCALL call_12_4_11_3(SH2_struct* sh)
{
  SH2movblg(sh,(11<<4)+3);
}
#define init_12_4_11_4
void FASTCALL call_12_4_11_4(SH2_struct* sh)
{
  SH2movblg(sh,(11<<4)+4);
}
#define init_12_4_11_5
void FASTCALL call_12_4_11_5(SH2_struct* sh)
{
  SH2movblg(sh,(11<<4)+5);
}
#define init_12_4_11_6
void FASTCALL call_12_4_11_6(SH2_struct* sh)
{
  SH2movblg(sh,(11<<4)+6);
}
#define init_12_4_11_7
void FASTCALL call_12_4_11_7(SH2_struct* sh)
{
  SH2movblg(sh,(11<<4)+7);
}
#define init_12_4_11_8
void FASTCALL call_12_4_11_8(SH2_struct* sh)
{
  SH2movblg(sh,(11<<4)+8);
}
#define init_12_4_11_9
void FASTCALL call_12_4_11_9(SH2_struct* sh)
{
  SH2movblg(sh,(11<<4)+9);
}
#define init_12_4_11_10
void FASTCALL call_12_4_11_10(SH2_struct* sh)
{
  SH2movblg(sh,(11<<4)+10);
}
#define init_12_4_11_11
void FASTCALL call_12_4_11_11(SH2_struct* sh)
{
  SH2movblg(sh,(11<<4)+11);
}
#define init_12_4_11_12
void FASTCALL call_12_4_11_12(SH2_struct* sh)
{
  SH2movblg(sh,(11<<4)+12);
}
#define init_12_4_11_13
void FASTCALL call_12_4_11_13(SH2_struct* sh)
{
  SH2movblg(sh,(11<<4)+13);
}
#define init_12_4_11_14
void FASTCALL call_12_4_11_14(SH2_struct* sh)
{
  SH2movblg(sh,(11<<4)+14);
}
#define init_12_4_11_15
void FASTCALL call_12_4_11_15(SH2_struct* sh)
{
  SH2movblg(sh,(11<<4)+15);
}
#define init_12_4_12_0
void FASTCALL call_12_4_12_0(SH2_struct* sh)
{
  SH2movblg(sh,(12<<4)+0);
}
#define init_12_4_12_1
void FASTCALL call_12_4_12_1(SH2_struct* sh)
{
  SH2movblg(sh,(12<<4)+1);
}
#define init_12_4_12_2
void FASTCALL call_12_4_12_2(SH2_struct* sh)
{
  SH2movblg(sh,(12<<4)+2);
}
#define init_12_4_12_3
void FASTCALL call_12_4_12_3(SH2_struct* sh)
{
  SH2movblg(sh,(12<<4)+3);
}
#define init_12_4_12_4
void FASTCALL call_12_4_12_4(SH2_struct* sh)
{
  SH2movblg(sh,(12<<4)+4);
}
#define init_12_4_12_5
void FASTCALL call_12_4_12_5(SH2_struct* sh)
{
  SH2movblg(sh,(12<<4)+5);
}
#define init_12_4_12_6
void FASTCALL call_12_4_12_6(SH2_struct* sh)
{
  SH2movblg(sh,(12<<4)+6);
}
#define init_12_4_12_7
void FASTCALL call_12_4_12_7(SH2_struct* sh)
{
  SH2movblg(sh,(12<<4)+7);
}
#define init_12_4_12_8
void FASTCALL call_12_4_12_8(SH2_struct* sh)
{
  SH2movblg(sh,(12<<4)+8);
}
#define init_12_4_12_9
void FASTCALL call_12_4_12_9(SH2_struct* sh)
{
  SH2movblg(sh,(12<<4)+9);
}
#define init_12_4_12_10
void FASTCALL call_12_4_12_10(SH2_struct* sh)
{
  SH2movblg(sh,(12<<4)+10);
}
#define init_12_4_12_11
void FASTCALL call_12_4_12_11(SH2_struct* sh)
{
  SH2movblg(sh,(12<<4)+11);
}
#define init_12_4_12_12
void FASTCALL call_12_4_12_12(SH2_struct* sh)
{
  SH2movblg(sh,(12<<4)+12);
}
#define init_12_4_12_13
void FASTCALL call_12_4_12_13(SH2_struct* sh)
{
  SH2movblg(sh,(12<<4)+13);
}
#define init_12_4_12_14
void FASTCALL call_12_4_12_14(SH2_struct* sh)
{
  SH2movblg(sh,(12<<4)+14);
}
#define init_12_4_12_15
void FASTCALL call_12_4_12_15(SH2_struct* sh)
{
  SH2movblg(sh,(12<<4)+15);
}
#define init_12_4_13_0
void FASTCALL call_12_4_13_0(SH2_struct* sh)
{
  SH2movblg(sh,(13<<4)+0);
}
#define init_12_4_13_1
void FASTCALL call_12_4_13_1(SH2_struct* sh)
{
  SH2movblg(sh,(13<<4)+1);
}
#define init_12_4_13_2
void FASTCALL call_12_4_13_2(SH2_struct* sh)
{
  SH2movblg(sh,(13<<4)+2);
}
#define init_12_4_13_3
void FASTCALL call_12_4_13_3(SH2_struct* sh)
{
  SH2movblg(sh,(13<<4)+3);
}
#define init_12_4_13_4
void FASTCALL call_12_4_13_4(SH2_struct* sh)
{
  SH2movblg(sh,(13<<4)+4);
}
#define init_12_4_13_5
void FASTCALL call_12_4_13_5(SH2_struct* sh)
{
  SH2movblg(sh,(13<<4)+5);
}
#define init_12_4_13_6
void FASTCALL call_12_4_13_6(SH2_struct* sh)
{
  SH2movblg(sh,(13<<4)+6);
}
#define init_12_4_13_7
void FASTCALL call_12_4_13_7(SH2_struct* sh)
{
  SH2movblg(sh,(13<<4)+7);
}
#define init_12_4_13_8
void FASTCALL call_12_4_13_8(SH2_struct* sh)
{
  SH2movblg(sh,(13<<4)+8);
}
#define init_12_4_13_9
void FASTCALL call_12_4_13_9(SH2_struct* sh)
{
  SH2movblg(sh,(13<<4)+9);
}
#define init_12_4_13_10
void FASTCALL call_12_4_13_10(SH2_struct* sh)
{
  SH2movblg(sh,(13<<4)+10);
}
#define init_12_4_13_11
void FASTCALL call_12_4_13_11(SH2_struct* sh)
{
  SH2movblg(sh,(13<<4)+11);
}
#define init_12_4_13_12
void FASTCALL call_12_4_13_12(SH2_struct* sh)
{
  SH2movblg(sh,(13<<4)+12);
}
#define init_12_4_13_13
void FASTCALL call_12_4_13_13(SH2_struct* sh)
{
  SH2movblg(sh,(13<<4)+13);
}
#define init_12_4_13_14
void FASTCALL call_12_4_13_14(SH2_struct* sh)
{
  SH2movblg(sh,(13<<4)+14);
}
#define init_12_4_13_15
void FASTCALL call_12_4_13_15(SH2_struct* sh)
{
  SH2movblg(sh,(13<<4)+15);
}
#define init_12_4_14_0
void FASTCALL call_12_4_14_0(SH2_struct* sh)
{
  SH2movblg(sh,(14<<4)+0);
}
#define init_12_4_14_1
void FASTCALL call_12_4_14_1(SH2_struct* sh)
{
  SH2movblg(sh,(14<<4)+1);
}
#define init_12_4_14_2
void FASTCALL call_12_4_14_2(SH2_struct* sh)
{
  SH2movblg(sh,(14<<4)+2);
}
#define init_12_4_14_3
void FASTCALL call_12_4_14_3(SH2_struct* sh)
{
  SH2movblg(sh,(14<<4)+3);
}
#define init_12_4_14_4
void FASTCALL call_12_4_14_4(SH2_struct* sh)
{
  SH2movblg(sh,(14<<4)+4);
}
#define init_12_4_14_5
void FASTCALL call_12_4_14_5(SH2_struct* sh)
{
  SH2movblg(sh,(14<<4)+5);
}
#define init_12_4_14_6
void FASTCALL call_12_4_14_6(SH2_struct* sh)
{
  SH2movblg(sh,(14<<4)+6);
}
#define init_12_4_14_7
void FASTCALL call_12_4_14_7(SH2_struct* sh)
{
  SH2movblg(sh,(14<<4)+7);
}
#define init_12_4_14_8
void FASTCALL call_12_4_14_8(SH2_struct* sh)
{
  SH2movblg(sh,(14<<4)+8);
}
#define init_12_4_14_9
void FASTCALL call_12_4_14_9(SH2_struct* sh)
{
  SH2movblg(sh,(14<<4)+9);
}
#define init_12_4_14_10
void FASTCALL call_12_4_14_10(SH2_struct* sh)
{
  SH2movblg(sh,(14<<4)+10);
}
#define init_12_4_14_11
void FASTCALL call_12_4_14_11(SH2_struct* sh)
{
  SH2movblg(sh,(14<<4)+11);
}
#define init_12_4_14_12
void FASTCALL call_12_4_14_12(SH2_struct* sh)
{
  SH2movblg(sh,(14<<4)+12);
}
#define init_12_4_14_13
void FASTCALL call_12_4_14_13(SH2_struct* sh)
{
  SH2movblg(sh,(14<<4)+13);
}
#define init_12_4_14_14
void FASTCALL call_12_4_14_14(SH2_struct* sh)
{
  SH2movblg(sh,(14<<4)+14);
}
#define init_12_4_14_15
void FASTCALL call_12_4_14_15(SH2_struct* sh)
{
  SH2movblg(sh,(14<<4)+15);
}
#define init_12_4_15_0
void FASTCALL call_12_4_15_0(SH2_struct* sh)
{
  SH2movblg(sh,(15<<4)+0);
}
#define init_12_4_15_1
void FASTCALL call_12_4_15_1(SH2_struct* sh)
{
  SH2movblg(sh,(15<<4)+1);
}
#define init_12_4_15_2
void FASTCALL call_12_4_15_2(SH2_struct* sh)
{
  SH2movblg(sh,(15<<4)+2);
}
#define init_12_4_15_3
void FASTCALL call_12_4_15_3(SH2_struct* sh)
{
  SH2movblg(sh,(15<<4)+3);
}
#define init_12_4_15_4
void FASTCALL call_12_4_15_4(SH2_struct* sh)
{
  SH2movblg(sh,(15<<4)+4);
}
#define init_12_4_15_5
void FASTCALL call_12_4_15_5(SH2_struct* sh)
{
  SH2movblg(sh,(15<<4)+5);
}
#define init_12_4_15_6
void FASTCALL call_12_4_15_6(SH2_struct* sh)
{
  SH2movblg(sh,(15<<4)+6);
}
#define init_12_4_15_7
void FASTCALL call_12_4_15_7(SH2_struct* sh)
{
  SH2movblg(sh,(15<<4)+7);
}
#define init_12_4_15_8
void FASTCALL call_12_4_15_8(SH2_struct* sh)
{
  SH2movblg(sh,(15<<4)+8);
}
#define init_12_4_15_9
void FASTCALL call_12_4_15_9(SH2_struct* sh)
{
  SH2movblg(sh,(15<<4)+9);
}
#define init_12_4_15_10
void FASTCALL call_12_4_15_10(SH2_struct* sh)
{
  SH2movblg(sh,(15<<4)+10);
}
#define init_12_4_15_11
void FASTCALL call_12_4_15_11(SH2_struct* sh)
{
  SH2movblg(sh,(15<<4)+11);
}
#define init_12_4_15_12
void FASTCALL call_12_4_15_12(SH2_struct* sh)
{
  SH2movblg(sh,(15<<4)+12);
}
#define init_12_4_15_13
void FASTCALL call_12_4_15_13(SH2_struct* sh)
{
  SH2movblg(sh,(15<<4)+13);
}
#define init_12_4_15_14
void FASTCALL call_12_4_15_14(SH2_struct* sh)
{
  SH2movblg(sh,(15<<4)+14);
}
#define init_12_4_15_15
void FASTCALL call_12_4_15_15(SH2_struct* sh)
{
  SH2movblg(sh,(15<<4)+15);
}
#define init_8_4_0_0
void FASTCALL call_8_4_0_0(SH2_struct* sh)
{
  SH2movbl4(sh,0,0);
}
#define init_8_4_0_1
void FASTCALL call_8_4_0_1(SH2_struct* sh)
{
  SH2movbl4(sh,0,1);
}
#define init_8_4_0_2
void FASTCALL call_8_4_0_2(SH2_struct* sh)
{
  SH2movbl4(sh,0,2);
}
#define init_8_4_0_3
void FASTCALL call_8_4_0_3(SH2_struct* sh)
{
  SH2movbl4(sh,0,3);
}
#define init_8_4_0_4
void FASTCALL call_8_4_0_4(SH2_struct* sh)
{
  SH2movbl4(sh,0,4);
}
#define init_8_4_0_5
void FASTCALL call_8_4_0_5(SH2_struct* sh)
{
  SH2movbl4(sh,0,5);
}
#define init_8_4_0_6
void FASTCALL call_8_4_0_6(SH2_struct* sh)
{
  SH2movbl4(sh,0,6);
}
#define init_8_4_0_7
void FASTCALL call_8_4_0_7(SH2_struct* sh)
{
  SH2movbl4(sh,0,7);
}
#define init_8_4_0_8
void FASTCALL call_8_4_0_8(SH2_struct* sh)
{
  SH2movbl4(sh,0,8);
}
#define init_8_4_0_9
void FASTCALL call_8_4_0_9(SH2_struct* sh)
{
  SH2movbl4(sh,0,9);
}
#define init_8_4_0_10
void FASTCALL call_8_4_0_10(SH2_struct* sh)
{
  SH2movbl4(sh,0,10);
}
#define init_8_4_0_11
void FASTCALL call_8_4_0_11(SH2_struct* sh)
{
  SH2movbl4(sh,0,11);
}
#define init_8_4_0_12
void FASTCALL call_8_4_0_12(SH2_struct* sh)
{
  SH2movbl4(sh,0,12);
}
#define init_8_4_0_13
void FASTCALL call_8_4_0_13(SH2_struct* sh)
{
  SH2movbl4(sh,0,13);
}
#define init_8_4_0_14
void FASTCALL call_8_4_0_14(SH2_struct* sh)
{
  SH2movbl4(sh,0,14);
}
#define init_8_4_0_15
void FASTCALL call_8_4_0_15(SH2_struct* sh)
{
  SH2movbl4(sh,0,15);
}
#define init_8_4_1_0
void FASTCALL call_8_4_1_0(SH2_struct* sh)
{
  SH2movbl4(sh,1,0);
}
#define init_8_4_1_1
void FASTCALL call_8_4_1_1(SH2_struct* sh)
{
  SH2movbl4(sh,1,1);
}
#define init_8_4_1_2
void FASTCALL call_8_4_1_2(SH2_struct* sh)
{
  SH2movbl4(sh,1,2);
}
#define init_8_4_1_3
void FASTCALL call_8_4_1_3(SH2_struct* sh)
{
  SH2movbl4(sh,1,3);
}
#define init_8_4_1_4
void FASTCALL call_8_4_1_4(SH2_struct* sh)
{
  SH2movbl4(sh,1,4);
}
#define init_8_4_1_5
void FASTCALL call_8_4_1_5(SH2_struct* sh)
{
  SH2movbl4(sh,1,5);
}
#define init_8_4_1_6
void FASTCALL call_8_4_1_6(SH2_struct* sh)
{
  SH2movbl4(sh,1,6);
}
#define init_8_4_1_7
void FASTCALL call_8_4_1_7(SH2_struct* sh)
{
  SH2movbl4(sh,1,7);
}
#define init_8_4_1_8
void FASTCALL call_8_4_1_8(SH2_struct* sh)
{
  SH2movbl4(sh,1,8);
}
#define init_8_4_1_9
void FASTCALL call_8_4_1_9(SH2_struct* sh)
{
  SH2movbl4(sh,1,9);
}
#define init_8_4_1_10
void FASTCALL call_8_4_1_10(SH2_struct* sh)
{
  SH2movbl4(sh,1,10);
}
#define init_8_4_1_11
void FASTCALL call_8_4_1_11(SH2_struct* sh)
{
  SH2movbl4(sh,1,11);
}
#define init_8_4_1_12
void FASTCALL call_8_4_1_12(SH2_struct* sh)
{
  SH2movbl4(sh,1,12);
}
#define init_8_4_1_13
void FASTCALL call_8_4_1_13(SH2_struct* sh)
{
  SH2movbl4(sh,1,13);
}
#define init_8_4_1_14
void FASTCALL call_8_4_1_14(SH2_struct* sh)
{
  SH2movbl4(sh,1,14);
}
#define init_8_4_1_15
void FASTCALL call_8_4_1_15(SH2_struct* sh)
{
  SH2movbl4(sh,1,15);
}
#define init_8_4_2_0
void FASTCALL call_8_4_2_0(SH2_struct* sh)
{
  SH2movbl4(sh,2,0);
}
#define init_8_4_2_1
void FASTCALL call_8_4_2_1(SH2_struct* sh)
{
  SH2movbl4(sh,2,1);
}
#define init_8_4_2_2
void FASTCALL call_8_4_2_2(SH2_struct* sh)
{
  SH2movbl4(sh,2,2);
}
#define init_8_4_2_3
void FASTCALL call_8_4_2_3(SH2_struct* sh)
{
  SH2movbl4(sh,2,3);
}
#define init_8_4_2_4
void FASTCALL call_8_4_2_4(SH2_struct* sh)
{
  SH2movbl4(sh,2,4);
}
#define init_8_4_2_5
void FASTCALL call_8_4_2_5(SH2_struct* sh)
{
  SH2movbl4(sh,2,5);
}
#define init_8_4_2_6
void FASTCALL call_8_4_2_6(SH2_struct* sh)
{
  SH2movbl4(sh,2,6);
}
#define init_8_4_2_7
void FASTCALL call_8_4_2_7(SH2_struct* sh)
{
  SH2movbl4(sh,2,7);
}
#define init_8_4_2_8
void FASTCALL call_8_4_2_8(SH2_struct* sh)
{
  SH2movbl4(sh,2,8);
}
#define init_8_4_2_9
void FASTCALL call_8_4_2_9(SH2_struct* sh)
{
  SH2movbl4(sh,2,9);
}
#define init_8_4_2_10
void FASTCALL call_8_4_2_10(SH2_struct* sh)
{
  SH2movbl4(sh,2,10);
}
#define init_8_4_2_11
void FASTCALL call_8_4_2_11(SH2_struct* sh)
{
  SH2movbl4(sh,2,11);
}
#define init_8_4_2_12
void FASTCALL call_8_4_2_12(SH2_struct* sh)
{
  SH2movbl4(sh,2,12);
}
#define init_8_4_2_13
void FASTCALL call_8_4_2_13(SH2_struct* sh)
{
  SH2movbl4(sh,2,13);
}
#define init_8_4_2_14
void FASTCALL call_8_4_2_14(SH2_struct* sh)
{
  SH2movbl4(sh,2,14);
}
#define init_8_4_2_15
void FASTCALL call_8_4_2_15(SH2_struct* sh)
{
  SH2movbl4(sh,2,15);
}
#define init_8_4_3_0
void FASTCALL call_8_4_3_0(SH2_struct* sh)
{
  SH2movbl4(sh,3,0);
}
#define init_8_4_3_1
void FASTCALL call_8_4_3_1(SH2_struct* sh)
{
  SH2movbl4(sh,3,1);
}
#define init_8_4_3_2
void FASTCALL call_8_4_3_2(SH2_struct* sh)
{
  SH2movbl4(sh,3,2);
}
#define init_8_4_3_3
void FASTCALL call_8_4_3_3(SH2_struct* sh)
{
  SH2movbl4(sh,3,3);
}
#define init_8_4_3_4
void FASTCALL call_8_4_3_4(SH2_struct* sh)
{
  SH2movbl4(sh,3,4);
}
#define init_8_4_3_5
void FASTCALL call_8_4_3_5(SH2_struct* sh)
{
  SH2movbl4(sh,3,5);
}
#define init_8_4_3_6
void FASTCALL call_8_4_3_6(SH2_struct* sh)
{
  SH2movbl4(sh,3,6);
}
#define init_8_4_3_7
void FASTCALL call_8_4_3_7(SH2_struct* sh)
{
  SH2movbl4(sh,3,7);
}
#define init_8_4_3_8
void FASTCALL call_8_4_3_8(SH2_struct* sh)
{
  SH2movbl4(sh,3,8);
}
#define init_8_4_3_9
void FASTCALL call_8_4_3_9(SH2_struct* sh)
{
  SH2movbl4(sh,3,9);
}
#define init_8_4_3_10
void FASTCALL call_8_4_3_10(SH2_struct* sh)
{
  SH2movbl4(sh,3,10);
}
#define init_8_4_3_11
void FASTCALL call_8_4_3_11(SH2_struct* sh)
{
  SH2movbl4(sh,3,11);
}
#define init_8_4_3_12
void FASTCALL call_8_4_3_12(SH2_struct* sh)
{
  SH2movbl4(sh,3,12);
}
#define init_8_4_3_13
void FASTCALL call_8_4_3_13(SH2_struct* sh)
{
  SH2movbl4(sh,3,13);
}
#define init_8_4_3_14
void FASTCALL call_8_4_3_14(SH2_struct* sh)
{
  SH2movbl4(sh,3,14);
}
#define init_8_4_3_15
void FASTCALL call_8_4_3_15(SH2_struct* sh)
{
  SH2movbl4(sh,3,15);
}
#define init_8_4_4_0
void FASTCALL call_8_4_4_0(SH2_struct* sh)
{
  SH2movbl4(sh,4,0);
}
#define init_8_4_4_1
void FASTCALL call_8_4_4_1(SH2_struct* sh)
{
  SH2movbl4(sh,4,1);
}
#define init_8_4_4_2
void FASTCALL call_8_4_4_2(SH2_struct* sh)
{
  SH2movbl4(sh,4,2);
}
#define init_8_4_4_3
void FASTCALL call_8_4_4_3(SH2_struct* sh)
{
  SH2movbl4(sh,4,3);
}
#define init_8_4_4_4
void FASTCALL call_8_4_4_4(SH2_struct* sh)
{
  SH2movbl4(sh,4,4);
}
#define init_8_4_4_5
void FASTCALL call_8_4_4_5(SH2_struct* sh)
{
  SH2movbl4(sh,4,5);
}
#define init_8_4_4_6
void FASTCALL call_8_4_4_6(SH2_struct* sh)
{
  SH2movbl4(sh,4,6);
}
#define init_8_4_4_7
void FASTCALL call_8_4_4_7(SH2_struct* sh)
{
  SH2movbl4(sh,4,7);
}
#define init_8_4_4_8
void FASTCALL call_8_4_4_8(SH2_struct* sh)
{
  SH2movbl4(sh,4,8);
}
#define init_8_4_4_9
void FASTCALL call_8_4_4_9(SH2_struct* sh)
{
  SH2movbl4(sh,4,9);
}
#define init_8_4_4_10
void FASTCALL call_8_4_4_10(SH2_struct* sh)
{
  SH2movbl4(sh,4,10);
}
#define init_8_4_4_11
void FASTCALL call_8_4_4_11(SH2_struct* sh)
{
  SH2movbl4(sh,4,11);
}
#define init_8_4_4_12
void FASTCALL call_8_4_4_12(SH2_struct* sh)
{
  SH2movbl4(sh,4,12);
}
#define init_8_4_4_13
void FASTCALL call_8_4_4_13(SH2_struct* sh)
{
  SH2movbl4(sh,4,13);
}
#define init_8_4_4_14
void FASTCALL call_8_4_4_14(SH2_struct* sh)
{
  SH2movbl4(sh,4,14);
}
#define init_8_4_4_15
void FASTCALL call_8_4_4_15(SH2_struct* sh)
{
  SH2movbl4(sh,4,15);
}
#define init_8_4_5_0
void FASTCALL call_8_4_5_0(SH2_struct* sh)
{
  SH2movbl4(sh,5,0);
}
#define init_8_4_5_1
void FASTCALL call_8_4_5_1(SH2_struct* sh)
{
  SH2movbl4(sh,5,1);
}
#define init_8_4_5_2
void FASTCALL call_8_4_5_2(SH2_struct* sh)
{
  SH2movbl4(sh,5,2);
}
#define init_8_4_5_3
void FASTCALL call_8_4_5_3(SH2_struct* sh)
{
  SH2movbl4(sh,5,3);
}
#define init_8_4_5_4
void FASTCALL call_8_4_5_4(SH2_struct* sh)
{
  SH2movbl4(sh,5,4);
}
#define init_8_4_5_5
void FASTCALL call_8_4_5_5(SH2_struct* sh)
{
  SH2movbl4(sh,5,5);
}
#define init_8_4_5_6
void FASTCALL call_8_4_5_6(SH2_struct* sh)
{
  SH2movbl4(sh,5,6);
}
#define init_8_4_5_7
void FASTCALL call_8_4_5_7(SH2_struct* sh)
{
  SH2movbl4(sh,5,7);
}
#define init_8_4_5_8
void FASTCALL call_8_4_5_8(SH2_struct* sh)
{
  SH2movbl4(sh,5,8);
}
#define init_8_4_5_9
void FASTCALL call_8_4_5_9(SH2_struct* sh)
{
  SH2movbl4(sh,5,9);
}
#define init_8_4_5_10
void FASTCALL call_8_4_5_10(SH2_struct* sh)
{
  SH2movbl4(sh,5,10);
}
#define init_8_4_5_11
void FASTCALL call_8_4_5_11(SH2_struct* sh)
{
  SH2movbl4(sh,5,11);
}
#define init_8_4_5_12
void FASTCALL call_8_4_5_12(SH2_struct* sh)
{
  SH2movbl4(sh,5,12);
}
#define init_8_4_5_13
void FASTCALL call_8_4_5_13(SH2_struct* sh)
{
  SH2movbl4(sh,5,13);
}
#define init_8_4_5_14
void FASTCALL call_8_4_5_14(SH2_struct* sh)
{
  SH2movbl4(sh,5,14);
}
#define init_8_4_5_15
void FASTCALL call_8_4_5_15(SH2_struct* sh)
{
  SH2movbl4(sh,5,15);
}
#define init_8_4_6_0
void FASTCALL call_8_4_6_0(SH2_struct* sh)
{
  SH2movbl4(sh,6,0);
}
#define init_8_4_6_1
void FASTCALL call_8_4_6_1(SH2_struct* sh)
{
  SH2movbl4(sh,6,1);
}
#define init_8_4_6_2
void FASTCALL call_8_4_6_2(SH2_struct* sh)
{
  SH2movbl4(sh,6,2);
}
#define init_8_4_6_3
void FASTCALL call_8_4_6_3(SH2_struct* sh)
{
  SH2movbl4(sh,6,3);
}
#define init_8_4_6_4
void FASTCALL call_8_4_6_4(SH2_struct* sh)
{
  SH2movbl4(sh,6,4);
}
#define init_8_4_6_5
void FASTCALL call_8_4_6_5(SH2_struct* sh)
{
  SH2movbl4(sh,6,5);
}
#define init_8_4_6_6
void FASTCALL call_8_4_6_6(SH2_struct* sh)
{
  SH2movbl4(sh,6,6);
}
#define init_8_4_6_7
void FASTCALL call_8_4_6_7(SH2_struct* sh)
{
  SH2movbl4(sh,6,7);
}
#define init_8_4_6_8
void FASTCALL call_8_4_6_8(SH2_struct* sh)
{
  SH2movbl4(sh,6,8);
}
#define init_8_4_6_9
void FASTCALL call_8_4_6_9(SH2_struct* sh)
{
  SH2movbl4(sh,6,9);
}
#define init_8_4_6_10
void FASTCALL call_8_4_6_10(SH2_struct* sh)
{
  SH2movbl4(sh,6,10);
}
#define init_8_4_6_11
void FASTCALL call_8_4_6_11(SH2_struct* sh)
{
  SH2movbl4(sh,6,11);
}
#define init_8_4_6_12
void FASTCALL call_8_4_6_12(SH2_struct* sh)
{
  SH2movbl4(sh,6,12);
}
#define init_8_4_6_13
void FASTCALL call_8_4_6_13(SH2_struct* sh)
{
  SH2movbl4(sh,6,13);
}
#define init_8_4_6_14
void FASTCALL call_8_4_6_14(SH2_struct* sh)
{
  SH2movbl4(sh,6,14);
}
#define init_8_4_6_15
void FASTCALL call_8_4_6_15(SH2_struct* sh)
{
  SH2movbl4(sh,6,15);
}
#define init_8_4_7_0
void FASTCALL call_8_4_7_0(SH2_struct* sh)
{
  SH2movbl4(sh,7,0);
}
#define init_8_4_7_1
void FASTCALL call_8_4_7_1(SH2_struct* sh)
{
  SH2movbl4(sh,7,1);
}
#define init_8_4_7_2
void FASTCALL call_8_4_7_2(SH2_struct* sh)
{
  SH2movbl4(sh,7,2);
}
#define init_8_4_7_3
void FASTCALL call_8_4_7_3(SH2_struct* sh)
{
  SH2movbl4(sh,7,3);
}
#define init_8_4_7_4
void FASTCALL call_8_4_7_4(SH2_struct* sh)
{
  SH2movbl4(sh,7,4);
}
#define init_8_4_7_5
void FASTCALL call_8_4_7_5(SH2_struct* sh)
{
  SH2movbl4(sh,7,5);
}
#define init_8_4_7_6
void FASTCALL call_8_4_7_6(SH2_struct* sh)
{
  SH2movbl4(sh,7,6);
}
#define init_8_4_7_7
void FASTCALL call_8_4_7_7(SH2_struct* sh)
{
  SH2movbl4(sh,7,7);
}
#define init_8_4_7_8
void FASTCALL call_8_4_7_8(SH2_struct* sh)
{
  SH2movbl4(sh,7,8);
}
#define init_8_4_7_9
void FASTCALL call_8_4_7_9(SH2_struct* sh)
{
  SH2movbl4(sh,7,9);
}
#define init_8_4_7_10
void FASTCALL call_8_4_7_10(SH2_struct* sh)
{
  SH2movbl4(sh,7,10);
}
#define init_8_4_7_11
void FASTCALL call_8_4_7_11(SH2_struct* sh)
{
  SH2movbl4(sh,7,11);
}
#define init_8_4_7_12
void FASTCALL call_8_4_7_12(SH2_struct* sh)
{
  SH2movbl4(sh,7,12);
}
#define init_8_4_7_13
void FASTCALL call_8_4_7_13(SH2_struct* sh)
{
  SH2movbl4(sh,7,13);
}
#define init_8_4_7_14
void FASTCALL call_8_4_7_14(SH2_struct* sh)
{
  SH2movbl4(sh,7,14);
}
#define init_8_4_7_15
void FASTCALL call_8_4_7_15(SH2_struct* sh)
{
  SH2movbl4(sh,7,15);
}
#define init_8_4_8_0
void FASTCALL call_8_4_8_0(SH2_struct* sh)
{
  SH2movbl4(sh,8,0);
}
#define init_8_4_8_1
void FASTCALL call_8_4_8_1(SH2_struct* sh)
{
  SH2movbl4(sh,8,1);
}
#define init_8_4_8_2
void FASTCALL call_8_4_8_2(SH2_struct* sh)
{
  SH2movbl4(sh,8,2);
}
#define init_8_4_8_3
void FASTCALL call_8_4_8_3(SH2_struct* sh)
{
  SH2movbl4(sh,8,3);
}
#define init_8_4_8_4
void FASTCALL call_8_4_8_4(SH2_struct* sh)
{
  SH2movbl4(sh,8,4);
}
#define init_8_4_8_5
void FASTCALL call_8_4_8_5(SH2_struct* sh)
{
  SH2movbl4(sh,8,5);
}
#define init_8_4_8_6
void FASTCALL call_8_4_8_6(SH2_struct* sh)
{
  SH2movbl4(sh,8,6);
}
#define init_8_4_8_7
void FASTCALL call_8_4_8_7(SH2_struct* sh)
{
  SH2movbl4(sh,8,7);
}
#define init_8_4_8_8
void FASTCALL call_8_4_8_8(SH2_struct* sh)
{
  SH2movbl4(sh,8,8);
}
#define init_8_4_8_9
void FASTCALL call_8_4_8_9(SH2_struct* sh)
{
  SH2movbl4(sh,8,9);
}
#define init_8_4_8_10
void FASTCALL call_8_4_8_10(SH2_struct* sh)
{
  SH2movbl4(sh,8,10);
}
#define init_8_4_8_11
void FASTCALL call_8_4_8_11(SH2_struct* sh)
{
  SH2movbl4(sh,8,11);
}
#define init_8_4_8_12
void FASTCALL call_8_4_8_12(SH2_struct* sh)
{
  SH2movbl4(sh,8,12);
}
#define init_8_4_8_13
void FASTCALL call_8_4_8_13(SH2_struct* sh)
{
  SH2movbl4(sh,8,13);
}
#define init_8_4_8_14
void FASTCALL call_8_4_8_14(SH2_struct* sh)
{
  SH2movbl4(sh,8,14);
}
#define init_8_4_8_15
void FASTCALL call_8_4_8_15(SH2_struct* sh)
{
  SH2movbl4(sh,8,15);
}
#define init_8_4_9_0
void FASTCALL call_8_4_9_0(SH2_struct* sh)
{
  SH2movbl4(sh,9,0);
}
#define init_8_4_9_1
void FASTCALL call_8_4_9_1(SH2_struct* sh)
{
  SH2movbl4(sh,9,1);
}
#define init_8_4_9_2
void FASTCALL call_8_4_9_2(SH2_struct* sh)
{
  SH2movbl4(sh,9,2);
}
#define init_8_4_9_3
void FASTCALL call_8_4_9_3(SH2_struct* sh)
{
  SH2movbl4(sh,9,3);
}
#define init_8_4_9_4
void FASTCALL call_8_4_9_4(SH2_struct* sh)
{
  SH2movbl4(sh,9,4);
}
#define init_8_4_9_5
void FASTCALL call_8_4_9_5(SH2_struct* sh)
{
  SH2movbl4(sh,9,5);
}
#define init_8_4_9_6
void FASTCALL call_8_4_9_6(SH2_struct* sh)
{
  SH2movbl4(sh,9,6);
}
#define init_8_4_9_7
void FASTCALL call_8_4_9_7(SH2_struct* sh)
{
  SH2movbl4(sh,9,7);
}
#define init_8_4_9_8
void FASTCALL call_8_4_9_8(SH2_struct* sh)
{
  SH2movbl4(sh,9,8);
}
#define init_8_4_9_9
void FASTCALL call_8_4_9_9(SH2_struct* sh)
{
  SH2movbl4(sh,9,9);
}
#define init_8_4_9_10
void FASTCALL call_8_4_9_10(SH2_struct* sh)
{
  SH2movbl4(sh,9,10);
}
#define init_8_4_9_11
void FASTCALL call_8_4_9_11(SH2_struct* sh)
{
  SH2movbl4(sh,9,11);
}
#define init_8_4_9_12
void FASTCALL call_8_4_9_12(SH2_struct* sh)
{
  SH2movbl4(sh,9,12);
}
#define init_8_4_9_13
void FASTCALL call_8_4_9_13(SH2_struct* sh)
{
  SH2movbl4(sh,9,13);
}
#define init_8_4_9_14
void FASTCALL call_8_4_9_14(SH2_struct* sh)
{
  SH2movbl4(sh,9,14);
}
#define init_8_4_9_15
void FASTCALL call_8_4_9_15(SH2_struct* sh)
{
  SH2movbl4(sh,9,15);
}
#define init_8_4_10_0
void FASTCALL call_8_4_10_0(SH2_struct* sh)
{
  SH2movbl4(sh,10,0);
}
#define init_8_4_10_1
void FASTCALL call_8_4_10_1(SH2_struct* sh)
{
  SH2movbl4(sh,10,1);
}
#define init_8_4_10_2
void FASTCALL call_8_4_10_2(SH2_struct* sh)
{
  SH2movbl4(sh,10,2);
}
#define init_8_4_10_3
void FASTCALL call_8_4_10_3(SH2_struct* sh)
{
  SH2movbl4(sh,10,3);
}
#define init_8_4_10_4
void FASTCALL call_8_4_10_4(SH2_struct* sh)
{
  SH2movbl4(sh,10,4);
}
#define init_8_4_10_5
void FASTCALL call_8_4_10_5(SH2_struct* sh)
{
  SH2movbl4(sh,10,5);
}
#define init_8_4_10_6
void FASTCALL call_8_4_10_6(SH2_struct* sh)
{
  SH2movbl4(sh,10,6);
}
#define init_8_4_10_7
void FASTCALL call_8_4_10_7(SH2_struct* sh)
{
  SH2movbl4(sh,10,7);
}
#define init_8_4_10_8
void FASTCALL call_8_4_10_8(SH2_struct* sh)
{
  SH2movbl4(sh,10,8);
}
#define init_8_4_10_9
void FASTCALL call_8_4_10_9(SH2_struct* sh)
{
  SH2movbl4(sh,10,9);
}
#define init_8_4_10_10
void FASTCALL call_8_4_10_10(SH2_struct* sh)
{
  SH2movbl4(sh,10,10);
}
#define init_8_4_10_11
void FASTCALL call_8_4_10_11(SH2_struct* sh)
{
  SH2movbl4(sh,10,11);
}
#define init_8_4_10_12
void FASTCALL call_8_4_10_12(SH2_struct* sh)
{
  SH2movbl4(sh,10,12);
}
#define init_8_4_10_13
void FASTCALL call_8_4_10_13(SH2_struct* sh)
{
  SH2movbl4(sh,10,13);
}
#define init_8_4_10_14
void FASTCALL call_8_4_10_14(SH2_struct* sh)
{
  SH2movbl4(sh,10,14);
}
#define init_8_4_10_15
void FASTCALL call_8_4_10_15(SH2_struct* sh)
{
  SH2movbl4(sh,10,15);
}
#define init_8_4_11_0
void FASTCALL call_8_4_11_0(SH2_struct* sh)
{
  SH2movbl4(sh,11,0);
}
#define init_8_4_11_1
void FASTCALL call_8_4_11_1(SH2_struct* sh)
{
  SH2movbl4(sh,11,1);
}
#define init_8_4_11_2
void FASTCALL call_8_4_11_2(SH2_struct* sh)
{
  SH2movbl4(sh,11,2);
}
#define init_8_4_11_3
void FASTCALL call_8_4_11_3(SH2_struct* sh)
{
  SH2movbl4(sh,11,3);
}
#define init_8_4_11_4
void FASTCALL call_8_4_11_4(SH2_struct* sh)
{
  SH2movbl4(sh,11,4);
}
#define init_8_4_11_5
void FASTCALL call_8_4_11_5(SH2_struct* sh)
{
  SH2movbl4(sh,11,5);
}
#define init_8_4_11_6
void FASTCALL call_8_4_11_6(SH2_struct* sh)
{
  SH2movbl4(sh,11,6);
}
#define init_8_4_11_7
void FASTCALL call_8_4_11_7(SH2_struct* sh)
{
  SH2movbl4(sh,11,7);
}
#define init_8_4_11_8
void FASTCALL call_8_4_11_8(SH2_struct* sh)
{
  SH2movbl4(sh,11,8);
}
#define init_8_4_11_9
void FASTCALL call_8_4_11_9(SH2_struct* sh)
{
  SH2movbl4(sh,11,9);
}
#define init_8_4_11_10
void FASTCALL call_8_4_11_10(SH2_struct* sh)
{
  SH2movbl4(sh,11,10);
}
#define init_8_4_11_11
void FASTCALL call_8_4_11_11(SH2_struct* sh)
{
  SH2movbl4(sh,11,11);
}
#define init_8_4_11_12
void FASTCALL call_8_4_11_12(SH2_struct* sh)
{
  SH2movbl4(sh,11,12);
}
#define init_8_4_11_13
void FASTCALL call_8_4_11_13(SH2_struct* sh)
{
  SH2movbl4(sh,11,13);
}
#define init_8_4_11_14
void FASTCALL call_8_4_11_14(SH2_struct* sh)
{
  SH2movbl4(sh,11,14);
}
#define init_8_4_11_15
void FASTCALL call_8_4_11_15(SH2_struct* sh)
{
  SH2movbl4(sh,11,15);
}
#define init_8_4_12_0
void FASTCALL call_8_4_12_0(SH2_struct* sh)
{
  SH2movbl4(sh,12,0);
}
#define init_8_4_12_1
void FASTCALL call_8_4_12_1(SH2_struct* sh)
{
  SH2movbl4(sh,12,1);
}
#define init_8_4_12_2
void FASTCALL call_8_4_12_2(SH2_struct* sh)
{
  SH2movbl4(sh,12,2);
}
#define init_8_4_12_3
void FASTCALL call_8_4_12_3(SH2_struct* sh)
{
  SH2movbl4(sh,12,3);
}
#define init_8_4_12_4
void FASTCALL call_8_4_12_4(SH2_struct* sh)
{
  SH2movbl4(sh,12,4);
}
#define init_8_4_12_5
void FASTCALL call_8_4_12_5(SH2_struct* sh)
{
  SH2movbl4(sh,12,5);
}
#define init_8_4_12_6
void FASTCALL call_8_4_12_6(SH2_struct* sh)
{
  SH2movbl4(sh,12,6);
}
#define init_8_4_12_7
void FASTCALL call_8_4_12_7(SH2_struct* sh)
{
  SH2movbl4(sh,12,7);
}
#define init_8_4_12_8
void FASTCALL call_8_4_12_8(SH2_struct* sh)
{
  SH2movbl4(sh,12,8);
}
#define init_8_4_12_9
void FASTCALL call_8_4_12_9(SH2_struct* sh)
{
  SH2movbl4(sh,12,9);
}
#define init_8_4_12_10
void FASTCALL call_8_4_12_10(SH2_struct* sh)
{
  SH2movbl4(sh,12,10);
}
#define init_8_4_12_11
void FASTCALL call_8_4_12_11(SH2_struct* sh)
{
  SH2movbl4(sh,12,11);
}
#define init_8_4_12_12
void FASTCALL call_8_4_12_12(SH2_struct* sh)
{
  SH2movbl4(sh,12,12);
}
#define init_8_4_12_13
void FASTCALL call_8_4_12_13(SH2_struct* sh)
{
  SH2movbl4(sh,12,13);
}
#define init_8_4_12_14
void FASTCALL call_8_4_12_14(SH2_struct* sh)
{
  SH2movbl4(sh,12,14);
}
#define init_8_4_12_15
void FASTCALL call_8_4_12_15(SH2_struct* sh)
{
  SH2movbl4(sh,12,15);
}
#define init_8_4_13_0
void FASTCALL call_8_4_13_0(SH2_struct* sh)
{
  SH2movbl4(sh,13,0);
}
#define init_8_4_13_1
void FASTCALL call_8_4_13_1(SH2_struct* sh)
{
  SH2movbl4(sh,13,1);
}
#define init_8_4_13_2
void FASTCALL call_8_4_13_2(SH2_struct* sh)
{
  SH2movbl4(sh,13,2);
}
#define init_8_4_13_3
void FASTCALL call_8_4_13_3(SH2_struct* sh)
{
  SH2movbl4(sh,13,3);
}
#define init_8_4_13_4
void FASTCALL call_8_4_13_4(SH2_struct* sh)
{
  SH2movbl4(sh,13,4);
}
#define init_8_4_13_5
void FASTCALL call_8_4_13_5(SH2_struct* sh)
{
  SH2movbl4(sh,13,5);
}
#define init_8_4_13_6
void FASTCALL call_8_4_13_6(SH2_struct* sh)
{
  SH2movbl4(sh,13,6);
}
#define init_8_4_13_7
void FASTCALL call_8_4_13_7(SH2_struct* sh)
{
  SH2movbl4(sh,13,7);
}
#define init_8_4_13_8
void FASTCALL call_8_4_13_8(SH2_struct* sh)
{
  SH2movbl4(sh,13,8);
}
#define init_8_4_13_9
void FASTCALL call_8_4_13_9(SH2_struct* sh)
{
  SH2movbl4(sh,13,9);
}
#define init_8_4_13_10
void FASTCALL call_8_4_13_10(SH2_struct* sh)
{
  SH2movbl4(sh,13,10);
}
#define init_8_4_13_11
void FASTCALL call_8_4_13_11(SH2_struct* sh)
{
  SH2movbl4(sh,13,11);
}
#define init_8_4_13_12
void FASTCALL call_8_4_13_12(SH2_struct* sh)
{
  SH2movbl4(sh,13,12);
}
#define init_8_4_13_13
void FASTCALL call_8_4_13_13(SH2_struct* sh)
{
  SH2movbl4(sh,13,13);
}
#define init_8_4_13_14
void FASTCALL call_8_4_13_14(SH2_struct* sh)
{
  SH2movbl4(sh,13,14);
}
#define init_8_4_13_15
void FASTCALL call_8_4_13_15(SH2_struct* sh)
{
  SH2movbl4(sh,13,15);
}
#define init_8_4_14_0
void FASTCALL call_8_4_14_0(SH2_struct* sh)
{
  SH2movbl4(sh,14,0);
}
#define init_8_4_14_1
void FASTCALL call_8_4_14_1(SH2_struct* sh)
{
  SH2movbl4(sh,14,1);
}
#define init_8_4_14_2
void FASTCALL call_8_4_14_2(SH2_struct* sh)
{
  SH2movbl4(sh,14,2);
}
#define init_8_4_14_3
void FASTCALL call_8_4_14_3(SH2_struct* sh)
{
  SH2movbl4(sh,14,3);
}
#define init_8_4_14_4
void FASTCALL call_8_4_14_4(SH2_struct* sh)
{
  SH2movbl4(sh,14,4);
}
#define init_8_4_14_5
void FASTCALL call_8_4_14_5(SH2_struct* sh)
{
  SH2movbl4(sh,14,5);
}
#define init_8_4_14_6
void FASTCALL call_8_4_14_6(SH2_struct* sh)
{
  SH2movbl4(sh,14,6);
}
#define init_8_4_14_7
void FASTCALL call_8_4_14_7(SH2_struct* sh)
{
  SH2movbl4(sh,14,7);
}
#define init_8_4_14_8
void FASTCALL call_8_4_14_8(SH2_struct* sh)
{
  SH2movbl4(sh,14,8);
}
#define init_8_4_14_9
void FASTCALL call_8_4_14_9(SH2_struct* sh)
{
  SH2movbl4(sh,14,9);
}
#define init_8_4_14_10
void FASTCALL call_8_4_14_10(SH2_struct* sh)
{
  SH2movbl4(sh,14,10);
}
#define init_8_4_14_11
void FASTCALL call_8_4_14_11(SH2_struct* sh)
{
  SH2movbl4(sh,14,11);
}
#define init_8_4_14_12
void FASTCALL call_8_4_14_12(SH2_struct* sh)
{
  SH2movbl4(sh,14,12);
}
#define init_8_4_14_13
void FASTCALL call_8_4_14_13(SH2_struct* sh)
{
  SH2movbl4(sh,14,13);
}
#define init_8_4_14_14
void FASTCALL call_8_4_14_14(SH2_struct* sh)
{
  SH2movbl4(sh,14,14);
}
#define init_8_4_14_15
void FASTCALL call_8_4_14_15(SH2_struct* sh)
{
  SH2movbl4(sh,14,15);
}
#define init_8_4_15_0
void FASTCALL call_8_4_15_0(SH2_struct* sh)
{
  SH2movbl4(sh,15,0);
}
#define init_8_4_15_1
void FASTCALL call_8_4_15_1(SH2_struct* sh)
{
  SH2movbl4(sh,15,1);
}
#define init_8_4_15_2
void FASTCALL call_8_4_15_2(SH2_struct* sh)
{
  SH2movbl4(sh,15,2);
}
#define init_8_4_15_3
void FASTCALL call_8_4_15_3(SH2_struct* sh)
{
  SH2movbl4(sh,15,3);
}
#define init_8_4_15_4
void FASTCALL call_8_4_15_4(SH2_struct* sh)
{
  SH2movbl4(sh,15,4);
}
#define init_8_4_15_5
void FASTCALL call_8_4_15_5(SH2_struct* sh)
{
  SH2movbl4(sh,15,5);
}
#define init_8_4_15_6
void FASTCALL call_8_4_15_6(SH2_struct* sh)
{
  SH2movbl4(sh,15,6);
}
#define init_8_4_15_7
void FASTCALL call_8_4_15_7(SH2_struct* sh)
{
  SH2movbl4(sh,15,7);
}
#define init_8_4_15_8
void FASTCALL call_8_4_15_8(SH2_struct* sh)
{
  SH2movbl4(sh,15,8);
}
#define init_8_4_15_9
void FASTCALL call_8_4_15_9(SH2_struct* sh)
{
  SH2movbl4(sh,15,9);
}
#define init_8_4_15_10
void FASTCALL call_8_4_15_10(SH2_struct* sh)
{
  SH2movbl4(sh,15,10);
}
#define init_8_4_15_11
void FASTCALL call_8_4_15_11(SH2_struct* sh)
{
  SH2movbl4(sh,15,11);
}
#define init_8_4_15_12
void FASTCALL call_8_4_15_12(SH2_struct* sh)
{
  SH2movbl4(sh,15,12);
}
#define init_8_4_15_13
void FASTCALL call_8_4_15_13(SH2_struct* sh)
{
  SH2movbl4(sh,15,13);
}
#define init_8_4_15_14
void FASTCALL call_8_4_15_14(SH2_struct* sh)
{
  SH2movbl4(sh,15,14);
}
#define init_8_4_15_15
void FASTCALL call_8_4_15_15(SH2_struct* sh)
{
  SH2movbl4(sh,15,15);
}
#define init_6_0_0_0
void FASTCALL call_6_0_0_0(SH2_struct* sh)
{
  SH2movbl(sh,0,0);
}
#define init_6_0_1_0
void FASTCALL call_6_0_1_0(SH2_struct* sh)
{
  SH2movbl(sh,0,1);
}
#define init_6_0_2_0
void FASTCALL call_6_0_2_0(SH2_struct* sh)
{
  SH2movbl(sh,0,2);
}
#define init_6_0_3_0
void FASTCALL call_6_0_3_0(SH2_struct* sh)
{
  SH2movbl(sh,0,3);
}
#define init_6_0_4_0
void FASTCALL call_6_0_4_0(SH2_struct* sh)
{
  SH2movbl(sh,0,4);
}
#define init_6_0_5_0
void FASTCALL call_6_0_5_0(SH2_struct* sh)
{
  SH2movbl(sh,0,5);
}
#define init_6_0_6_0
void FASTCALL call_6_0_6_0(SH2_struct* sh)
{
  SH2movbl(sh,0,6);
}
#define init_6_0_7_0
void FASTCALL call_6_0_7_0(SH2_struct* sh)
{
  SH2movbl(sh,0,7);
}
#define init_6_0_8_0
void FASTCALL call_6_0_8_0(SH2_struct* sh)
{
  SH2movbl(sh,0,8);
}
#define init_6_0_9_0
void FASTCALL call_6_0_9_0(SH2_struct* sh)
{
  SH2movbl(sh,0,9);
}
#define init_6_0_10_0
void FASTCALL call_6_0_10_0(SH2_struct* sh)
{
  SH2movbl(sh,0,10);
}
#define init_6_0_11_0
void FASTCALL call_6_0_11_0(SH2_struct* sh)
{
  SH2movbl(sh,0,11);
}
#define init_6_0_12_0
void FASTCALL call_6_0_12_0(SH2_struct* sh)
{
  SH2movbl(sh,0,12);
}
#define init_6_0_13_0
void FASTCALL call_6_0_13_0(SH2_struct* sh)
{
  SH2movbl(sh,0,13);
}
#define init_6_0_14_0
void FASTCALL call_6_0_14_0(SH2_struct* sh)
{
  SH2movbl(sh,0,14);
}
#define init_6_0_15_0
void FASTCALL call_6_0_15_0(SH2_struct* sh)
{
  SH2movbl(sh,0,15);
}
#define init_6_1_0_0
void FASTCALL call_6_1_0_0(SH2_struct* sh)
{
  SH2movbl(sh,1,0);
}
#define init_6_1_1_0
void FASTCALL call_6_1_1_0(SH2_struct* sh)
{
  SH2movbl(sh,1,1);
}
#define init_6_1_2_0
void FASTCALL call_6_1_2_0(SH2_struct* sh)
{
  SH2movbl(sh,1,2);
}
#define init_6_1_3_0
void FASTCALL call_6_1_3_0(SH2_struct* sh)
{
  SH2movbl(sh,1,3);
}
#define init_6_1_4_0
void FASTCALL call_6_1_4_0(SH2_struct* sh)
{
  SH2movbl(sh,1,4);
}
#define init_6_1_5_0
void FASTCALL call_6_1_5_0(SH2_struct* sh)
{
  SH2movbl(sh,1,5);
}
#define init_6_1_6_0
void FASTCALL call_6_1_6_0(SH2_struct* sh)
{
  SH2movbl(sh,1,6);
}
#define init_6_1_7_0
void FASTCALL call_6_1_7_0(SH2_struct* sh)
{
  SH2movbl(sh,1,7);
}
#define init_6_1_8_0
void FASTCALL call_6_1_8_0(SH2_struct* sh)
{
  SH2movbl(sh,1,8);
}
#define init_6_1_9_0
void FASTCALL call_6_1_9_0(SH2_struct* sh)
{
  SH2movbl(sh,1,9);
}
#define init_6_1_10_0
void FASTCALL call_6_1_10_0(SH2_struct* sh)
{
  SH2movbl(sh,1,10);
}
#define init_6_1_11_0
void FASTCALL call_6_1_11_0(SH2_struct* sh)
{
  SH2movbl(sh,1,11);
}
#define init_6_1_12_0
void FASTCALL call_6_1_12_0(SH2_struct* sh)
{
  SH2movbl(sh,1,12);
}
#define init_6_1_13_0
void FASTCALL call_6_1_13_0(SH2_struct* sh)
{
  SH2movbl(sh,1,13);
}
#define init_6_1_14_0
void FASTCALL call_6_1_14_0(SH2_struct* sh)
{
  SH2movbl(sh,1,14);
}
#define init_6_1_15_0
void FASTCALL call_6_1_15_0(SH2_struct* sh)
{
  SH2movbl(sh,1,15);
}
#define init_6_2_0_0
void FASTCALL call_6_2_0_0(SH2_struct* sh)
{
  SH2movbl(sh,2,0);
}
#define init_6_2_1_0
void FASTCALL call_6_2_1_0(SH2_struct* sh)
{
  SH2movbl(sh,2,1);
}
#define init_6_2_2_0
void FASTCALL call_6_2_2_0(SH2_struct* sh)
{
  SH2movbl(sh,2,2);
}
#define init_6_2_3_0
void FASTCALL call_6_2_3_0(SH2_struct* sh)
{
  SH2movbl(sh,2,3);
}
#define init_6_2_4_0
void FASTCALL call_6_2_4_0(SH2_struct* sh)
{
  SH2movbl(sh,2,4);
}
#define init_6_2_5_0
void FASTCALL call_6_2_5_0(SH2_struct* sh)
{
  SH2movbl(sh,2,5);
}
#define init_6_2_6_0
void FASTCALL call_6_2_6_0(SH2_struct* sh)
{
  SH2movbl(sh,2,6);
}
#define init_6_2_7_0
void FASTCALL call_6_2_7_0(SH2_struct* sh)
{
  SH2movbl(sh,2,7);
}
#define init_6_2_8_0
void FASTCALL call_6_2_8_0(SH2_struct* sh)
{
  SH2movbl(sh,2,8);
}
#define init_6_2_9_0
void FASTCALL call_6_2_9_0(SH2_struct* sh)
{
  SH2movbl(sh,2,9);
}
#define init_6_2_10_0
void FASTCALL call_6_2_10_0(SH2_struct* sh)
{
  SH2movbl(sh,2,10);
}
#define init_6_2_11_0
void FASTCALL call_6_2_11_0(SH2_struct* sh)
{
  SH2movbl(sh,2,11);
}
#define init_6_2_12_0
void FASTCALL call_6_2_12_0(SH2_struct* sh)
{
  SH2movbl(sh,2,12);
}
#define init_6_2_13_0
void FASTCALL call_6_2_13_0(SH2_struct* sh)
{
  SH2movbl(sh,2,13);
}
#define init_6_2_14_0
void FASTCALL call_6_2_14_0(SH2_struct* sh)
{
  SH2movbl(sh,2,14);
}
#define init_6_2_15_0
void FASTCALL call_6_2_15_0(SH2_struct* sh)
{
  SH2movbl(sh,2,15);
}
#define init_6_3_0_0
void FASTCALL call_6_3_0_0(SH2_struct* sh)
{
  SH2movbl(sh,3,0);
}
#define init_6_3_1_0
void FASTCALL call_6_3_1_0(SH2_struct* sh)
{
  SH2movbl(sh,3,1);
}
#define init_6_3_2_0
void FASTCALL call_6_3_2_0(SH2_struct* sh)
{
  SH2movbl(sh,3,2);
}
#define init_6_3_3_0
void FASTCALL call_6_3_3_0(SH2_struct* sh)
{
  SH2movbl(sh,3,3);
}
#define init_6_3_4_0
void FASTCALL call_6_3_4_0(SH2_struct* sh)
{
  SH2movbl(sh,3,4);
}
#define init_6_3_5_0
void FASTCALL call_6_3_5_0(SH2_struct* sh)
{
  SH2movbl(sh,3,5);
}
#define init_6_3_6_0
void FASTCALL call_6_3_6_0(SH2_struct* sh)
{
  SH2movbl(sh,3,6);
}
#define init_6_3_7_0
void FASTCALL call_6_3_7_0(SH2_struct* sh)
{
  SH2movbl(sh,3,7);
}
#define init_6_3_8_0
void FASTCALL call_6_3_8_0(SH2_struct* sh)
{
  SH2movbl(sh,3,8);
}
#define init_6_3_9_0
void FASTCALL call_6_3_9_0(SH2_struct* sh)
{
  SH2movbl(sh,3,9);
}
#define init_6_3_10_0
void FASTCALL call_6_3_10_0(SH2_struct* sh)
{
  SH2movbl(sh,3,10);
}
#define init_6_3_11_0
void FASTCALL call_6_3_11_0(SH2_struct* sh)
{
  SH2movbl(sh,3,11);
}
#define init_6_3_12_0
void FASTCALL call_6_3_12_0(SH2_struct* sh)
{
  SH2movbl(sh,3,12);
}
#define init_6_3_13_0
void FASTCALL call_6_3_13_0(SH2_struct* sh)
{
  SH2movbl(sh,3,13);
}
#define init_6_3_14_0
void FASTCALL call_6_3_14_0(SH2_struct* sh)
{
  SH2movbl(sh,3,14);
}
#define init_6_3_15_0
void FASTCALL call_6_3_15_0(SH2_struct* sh)
{
  SH2movbl(sh,3,15);
}
#define init_6_4_0_0
void FASTCALL call_6_4_0_0(SH2_struct* sh)
{
  SH2movbl(sh,4,0);
}
#define init_6_4_1_0
void FASTCALL call_6_4_1_0(SH2_struct* sh)
{
  SH2movbl(sh,4,1);
}
#define init_6_4_2_0
void FASTCALL call_6_4_2_0(SH2_struct* sh)
{
  SH2movbl(sh,4,2);
}
#define init_6_4_3_0
void FASTCALL call_6_4_3_0(SH2_struct* sh)
{
  SH2movbl(sh,4,3);
}
#define init_6_4_4_0
void FASTCALL call_6_4_4_0(SH2_struct* sh)
{
  SH2movbl(sh,4,4);
}
#define init_6_4_5_0
void FASTCALL call_6_4_5_0(SH2_struct* sh)
{
  SH2movbl(sh,4,5);
}
#define init_6_4_6_0
void FASTCALL call_6_4_6_0(SH2_struct* sh)
{
  SH2movbl(sh,4,6);
}
#define init_6_4_7_0
void FASTCALL call_6_4_7_0(SH2_struct* sh)
{
  SH2movbl(sh,4,7);
}
#define init_6_4_8_0
void FASTCALL call_6_4_8_0(SH2_struct* sh)
{
  SH2movbl(sh,4,8);
}
#define init_6_4_9_0
void FASTCALL call_6_4_9_0(SH2_struct* sh)
{
  SH2movbl(sh,4,9);
}
#define init_6_4_10_0
void FASTCALL call_6_4_10_0(SH2_struct* sh)
{
  SH2movbl(sh,4,10);
}
#define init_6_4_11_0
void FASTCALL call_6_4_11_0(SH2_struct* sh)
{
  SH2movbl(sh,4,11);
}
#define init_6_4_12_0
void FASTCALL call_6_4_12_0(SH2_struct* sh)
{
  SH2movbl(sh,4,12);
}
#define init_6_4_13_0
void FASTCALL call_6_4_13_0(SH2_struct* sh)
{
  SH2movbl(sh,4,13);
}
#define init_6_4_14_0
void FASTCALL call_6_4_14_0(SH2_struct* sh)
{
  SH2movbl(sh,4,14);
}
#define init_6_4_15_0
void FASTCALL call_6_4_15_0(SH2_struct* sh)
{
  SH2movbl(sh,4,15);
}
#define init_6_5_0_0
void FASTCALL call_6_5_0_0(SH2_struct* sh)
{
  SH2movbl(sh,5,0);
}
#define init_6_5_1_0
void FASTCALL call_6_5_1_0(SH2_struct* sh)
{
  SH2movbl(sh,5,1);
}
#define init_6_5_2_0
void FASTCALL call_6_5_2_0(SH2_struct* sh)
{
  SH2movbl(sh,5,2);
}
#define init_6_5_3_0
void FASTCALL call_6_5_3_0(SH2_struct* sh)
{
  SH2movbl(sh,5,3);
}
#define init_6_5_4_0
void FASTCALL call_6_5_4_0(SH2_struct* sh)
{
  SH2movbl(sh,5,4);
}
#define init_6_5_5_0
void FASTCALL call_6_5_5_0(SH2_struct* sh)
{
  SH2movbl(sh,5,5);
}
#define init_6_5_6_0
void FASTCALL call_6_5_6_0(SH2_struct* sh)
{
  SH2movbl(sh,5,6);
}
#define init_6_5_7_0
void FASTCALL call_6_5_7_0(SH2_struct* sh)
{
  SH2movbl(sh,5,7);
}
#define init_6_5_8_0
void FASTCALL call_6_5_8_0(SH2_struct* sh)
{
  SH2movbl(sh,5,8);
}
#define init_6_5_9_0
void FASTCALL call_6_5_9_0(SH2_struct* sh)
{
  SH2movbl(sh,5,9);
}
#define init_6_5_10_0
void FASTCALL call_6_5_10_0(SH2_struct* sh)
{
  SH2movbl(sh,5,10);
}
#define init_6_5_11_0
void FASTCALL call_6_5_11_0(SH2_struct* sh)
{
  SH2movbl(sh,5,11);
}
#define init_6_5_12_0
void FASTCALL call_6_5_12_0(SH2_struct* sh)
{
  SH2movbl(sh,5,12);
}
#define init_6_5_13_0
void FASTCALL call_6_5_13_0(SH2_struct* sh)
{
  SH2movbl(sh,5,13);
}
#define init_6_5_14_0
void FASTCALL call_6_5_14_0(SH2_struct* sh)
{
  SH2movbl(sh,5,14);
}
#define init_6_5_15_0
void FASTCALL call_6_5_15_0(SH2_struct* sh)
{
  SH2movbl(sh,5,15);
}
#define init_6_6_0_0
void FASTCALL call_6_6_0_0(SH2_struct* sh)
{
  SH2movbl(sh,6,0);
}
#define init_6_6_1_0
void FASTCALL call_6_6_1_0(SH2_struct* sh)
{
  SH2movbl(sh,6,1);
}
#define init_6_6_2_0
void FASTCALL call_6_6_2_0(SH2_struct* sh)
{
  SH2movbl(sh,6,2);
}
#define init_6_6_3_0
void FASTCALL call_6_6_3_0(SH2_struct* sh)
{
  SH2movbl(sh,6,3);
}
#define init_6_6_4_0
void FASTCALL call_6_6_4_0(SH2_struct* sh)
{
  SH2movbl(sh,6,4);
}
#define init_6_6_5_0
void FASTCALL call_6_6_5_0(SH2_struct* sh)
{
  SH2movbl(sh,6,5);
}
#define init_6_6_6_0
void FASTCALL call_6_6_6_0(SH2_struct* sh)
{
  SH2movbl(sh,6,6);
}
#define init_6_6_7_0
void FASTCALL call_6_6_7_0(SH2_struct* sh)
{
  SH2movbl(sh,6,7);
}
#define init_6_6_8_0
void FASTCALL call_6_6_8_0(SH2_struct* sh)
{
  SH2movbl(sh,6,8);
}
#define init_6_6_9_0
void FASTCALL call_6_6_9_0(SH2_struct* sh)
{
  SH2movbl(sh,6,9);
}
#define init_6_6_10_0
void FASTCALL call_6_6_10_0(SH2_struct* sh)
{
  SH2movbl(sh,6,10);
}
#define init_6_6_11_0
void FASTCALL call_6_6_11_0(SH2_struct* sh)
{
  SH2movbl(sh,6,11);
}
#define init_6_6_12_0
void FASTCALL call_6_6_12_0(SH2_struct* sh)
{
  SH2movbl(sh,6,12);
}
#define init_6_6_13_0
void FASTCALL call_6_6_13_0(SH2_struct* sh)
{
  SH2movbl(sh,6,13);
}
#define init_6_6_14_0
void FASTCALL call_6_6_14_0(SH2_struct* sh)
{
  SH2movbl(sh,6,14);
}
#define init_6_6_15_0
void FASTCALL call_6_6_15_0(SH2_struct* sh)
{
  SH2movbl(sh,6,15);
}
#define init_6_7_0_0
void FASTCALL call_6_7_0_0(SH2_struct* sh)
{
  SH2movbl(sh,7,0);
}
#define init_6_7_1_0
void FASTCALL call_6_7_1_0(SH2_struct* sh)
{
  SH2movbl(sh,7,1);
}
#define init_6_7_2_0
void FASTCALL call_6_7_2_0(SH2_struct* sh)
{
  SH2movbl(sh,7,2);
}
#define init_6_7_3_0
void FASTCALL call_6_7_3_0(SH2_struct* sh)
{
  SH2movbl(sh,7,3);
}
#define init_6_7_4_0
void FASTCALL call_6_7_4_0(SH2_struct* sh)
{
  SH2movbl(sh,7,4);
}
#define init_6_7_5_0
void FASTCALL call_6_7_5_0(SH2_struct* sh)
{
  SH2movbl(sh,7,5);
}
#define init_6_7_6_0
void FASTCALL call_6_7_6_0(SH2_struct* sh)
{
  SH2movbl(sh,7,6);
}
#define init_6_7_7_0
void FASTCALL call_6_7_7_0(SH2_struct* sh)
{
  SH2movbl(sh,7,7);
}
#define init_6_7_8_0
void FASTCALL call_6_7_8_0(SH2_struct* sh)
{
  SH2movbl(sh,7,8);
}
#define init_6_7_9_0
void FASTCALL call_6_7_9_0(SH2_struct* sh)
{
  SH2movbl(sh,7,9);
}
#define init_6_7_10_0
void FASTCALL call_6_7_10_0(SH2_struct* sh)
{
  SH2movbl(sh,7,10);
}
#define init_6_7_11_0
void FASTCALL call_6_7_11_0(SH2_struct* sh)
{
  SH2movbl(sh,7,11);
}
#define init_6_7_12_0
void FASTCALL call_6_7_12_0(SH2_struct* sh)
{
  SH2movbl(sh,7,12);
}
#define init_6_7_13_0
void FASTCALL call_6_7_13_0(SH2_struct* sh)
{
  SH2movbl(sh,7,13);
}
#define init_6_7_14_0
void FASTCALL call_6_7_14_0(SH2_struct* sh)
{
  SH2movbl(sh,7,14);
}
#define init_6_7_15_0
void FASTCALL call_6_7_15_0(SH2_struct* sh)
{
  SH2movbl(sh,7,15);
}
#define init_6_8_0_0
void FASTCALL call_6_8_0_0(SH2_struct* sh)
{
  SH2movbl(sh,8,0);
}
#define init_6_8_1_0
void FASTCALL call_6_8_1_0(SH2_struct* sh)
{
  SH2movbl(sh,8,1);
}
#define init_6_8_2_0
void FASTCALL call_6_8_2_0(SH2_struct* sh)
{
  SH2movbl(sh,8,2);
}
#define init_6_8_3_0
void FASTCALL call_6_8_3_0(SH2_struct* sh)
{
  SH2movbl(sh,8,3);
}
#define init_6_8_4_0
void FASTCALL call_6_8_4_0(SH2_struct* sh)
{
  SH2movbl(sh,8,4);
}
#define init_6_8_5_0
void FASTCALL call_6_8_5_0(SH2_struct* sh)
{
  SH2movbl(sh,8,5);
}
#define init_6_8_6_0
void FASTCALL call_6_8_6_0(SH2_struct* sh)
{
  SH2movbl(sh,8,6);
}
#define init_6_8_7_0
void FASTCALL call_6_8_7_0(SH2_struct* sh)
{
  SH2movbl(sh,8,7);
}
#define init_6_8_8_0
void FASTCALL call_6_8_8_0(SH2_struct* sh)
{
  SH2movbl(sh,8,8);
}
#define init_6_8_9_0
void FASTCALL call_6_8_9_0(SH2_struct* sh)
{
  SH2movbl(sh,8,9);
}
#define init_6_8_10_0
void FASTCALL call_6_8_10_0(SH2_struct* sh)
{
  SH2movbl(sh,8,10);
}
#define init_6_8_11_0
void FASTCALL call_6_8_11_0(SH2_struct* sh)
{
  SH2movbl(sh,8,11);
}
#define init_6_8_12_0
void FASTCALL call_6_8_12_0(SH2_struct* sh)
{
  SH2movbl(sh,8,12);
}
#define init_6_8_13_0
void FASTCALL call_6_8_13_0(SH2_struct* sh)
{
  SH2movbl(sh,8,13);
}
#define init_6_8_14_0
void FASTCALL call_6_8_14_0(SH2_struct* sh)
{
  SH2movbl(sh,8,14);
}
#define init_6_8_15_0
void FASTCALL call_6_8_15_0(SH2_struct* sh)
{
  SH2movbl(sh,8,15);
}
#define init_6_9_0_0
void FASTCALL call_6_9_0_0(SH2_struct* sh)
{
  SH2movbl(sh,9,0);
}
#define init_6_9_1_0
void FASTCALL call_6_9_1_0(SH2_struct* sh)
{
  SH2movbl(sh,9,1);
}
#define init_6_9_2_0
void FASTCALL call_6_9_2_0(SH2_struct* sh)
{
  SH2movbl(sh,9,2);
}
#define init_6_9_3_0
void FASTCALL call_6_9_3_0(SH2_struct* sh)
{
  SH2movbl(sh,9,3);
}
#define init_6_9_4_0
void FASTCALL call_6_9_4_0(SH2_struct* sh)
{
  SH2movbl(sh,9,4);
}
#define init_6_9_5_0
void FASTCALL call_6_9_5_0(SH2_struct* sh)
{
  SH2movbl(sh,9,5);
}
#define init_6_9_6_0
void FASTCALL call_6_9_6_0(SH2_struct* sh)
{
  SH2movbl(sh,9,6);
}
#define init_6_9_7_0
void FASTCALL call_6_9_7_0(SH2_struct* sh)
{
  SH2movbl(sh,9,7);
}
#define init_6_9_8_0
void FASTCALL call_6_9_8_0(SH2_struct* sh)
{
  SH2movbl(sh,9,8);
}
#define init_6_9_9_0
void FASTCALL call_6_9_9_0(SH2_struct* sh)
{
  SH2movbl(sh,9,9);
}
#define init_6_9_10_0
void FASTCALL call_6_9_10_0(SH2_struct* sh)
{
  SH2movbl(sh,9,10);
}
#define init_6_9_11_0
void FASTCALL call_6_9_11_0(SH2_struct* sh)
{
  SH2movbl(sh,9,11);
}
#define init_6_9_12_0
void FASTCALL call_6_9_12_0(SH2_struct* sh)
{
  SH2movbl(sh,9,12);
}
#define init_6_9_13_0
void FASTCALL call_6_9_13_0(SH2_struct* sh)
{
  SH2movbl(sh,9,13);
}
#define init_6_9_14_0
void FASTCALL call_6_9_14_0(SH2_struct* sh)
{
  SH2movbl(sh,9,14);
}
#define init_6_9_15_0
void FASTCALL call_6_9_15_0(SH2_struct* sh)
{
  SH2movbl(sh,9,15);
}
#define init_6_10_0_0
void FASTCALL call_6_10_0_0(SH2_struct* sh)
{
  SH2movbl(sh,10,0);
}
#define init_6_10_1_0
void FASTCALL call_6_10_1_0(SH2_struct* sh)
{
  SH2movbl(sh,10,1);
}
#define init_6_10_2_0
void FASTCALL call_6_10_2_0(SH2_struct* sh)
{
  SH2movbl(sh,10,2);
}
#define init_6_10_3_0
void FASTCALL call_6_10_3_0(SH2_struct* sh)
{
  SH2movbl(sh,10,3);
}
#define init_6_10_4_0
void FASTCALL call_6_10_4_0(SH2_struct* sh)
{
  SH2movbl(sh,10,4);
}
#define init_6_10_5_0
void FASTCALL call_6_10_5_0(SH2_struct* sh)
{
  SH2movbl(sh,10,5);
}
#define init_6_10_6_0
void FASTCALL call_6_10_6_0(SH2_struct* sh)
{
  SH2movbl(sh,10,6);
}
#define init_6_10_7_0
void FASTCALL call_6_10_7_0(SH2_struct* sh)
{
  SH2movbl(sh,10,7);
}
#define init_6_10_8_0
void FASTCALL call_6_10_8_0(SH2_struct* sh)
{
  SH2movbl(sh,10,8);
}
#define init_6_10_9_0
void FASTCALL call_6_10_9_0(SH2_struct* sh)
{
  SH2movbl(sh,10,9);
}
#define init_6_10_10_0
void FASTCALL call_6_10_10_0(SH2_struct* sh)
{
  SH2movbl(sh,10,10);
}
#define init_6_10_11_0
void FASTCALL call_6_10_11_0(SH2_struct* sh)
{
  SH2movbl(sh,10,11);
}
#define init_6_10_12_0
void FASTCALL call_6_10_12_0(SH2_struct* sh)
{
  SH2movbl(sh,10,12);
}
#define init_6_10_13_0
void FASTCALL call_6_10_13_0(SH2_struct* sh)
{
  SH2movbl(sh,10,13);
}
#define init_6_10_14_0
void FASTCALL call_6_10_14_0(SH2_struct* sh)
{
  SH2movbl(sh,10,14);
}
#define init_6_10_15_0
void FASTCALL call_6_10_15_0(SH2_struct* sh)
{
  SH2movbl(sh,10,15);
}
#define init_6_11_0_0
void FASTCALL call_6_11_0_0(SH2_struct* sh)
{
  SH2movbl(sh,11,0);
}
#define init_6_11_1_0
void FASTCALL call_6_11_1_0(SH2_struct* sh)
{
  SH2movbl(sh,11,1);
}
#define init_6_11_2_0
void FASTCALL call_6_11_2_0(SH2_struct* sh)
{
  SH2movbl(sh,11,2);
}
#define init_6_11_3_0
void FASTCALL call_6_11_3_0(SH2_struct* sh)
{
  SH2movbl(sh,11,3);
}
#define init_6_11_4_0
void FASTCALL call_6_11_4_0(SH2_struct* sh)
{
  SH2movbl(sh,11,4);
}
#define init_6_11_5_0
void FASTCALL call_6_11_5_0(SH2_struct* sh)
{
  SH2movbl(sh,11,5);
}
#define init_6_11_6_0
void FASTCALL call_6_11_6_0(SH2_struct* sh)
{
  SH2movbl(sh,11,6);
}
#define init_6_11_7_0
void FASTCALL call_6_11_7_0(SH2_struct* sh)
{
  SH2movbl(sh,11,7);
}
#define init_6_11_8_0
void FASTCALL call_6_11_8_0(SH2_struct* sh)
{
  SH2movbl(sh,11,8);
}
#define init_6_11_9_0
void FASTCALL call_6_11_9_0(SH2_struct* sh)
{
  SH2movbl(sh,11,9);
}
#define init_6_11_10_0
void FASTCALL call_6_11_10_0(SH2_struct* sh)
{
  SH2movbl(sh,11,10);
}
#define init_6_11_11_0
void FASTCALL call_6_11_11_0(SH2_struct* sh)
{
  SH2movbl(sh,11,11);
}
#define init_6_11_12_0
void FASTCALL call_6_11_12_0(SH2_struct* sh)
{
  SH2movbl(sh,11,12);
}
#define init_6_11_13_0
void FASTCALL call_6_11_13_0(SH2_struct* sh)
{
  SH2movbl(sh,11,13);
}
#define init_6_11_14_0
void FASTCALL call_6_11_14_0(SH2_struct* sh)
{
  SH2movbl(sh,11,14);
}
#define init_6_11_15_0
void FASTCALL call_6_11_15_0(SH2_struct* sh)
{
  SH2movbl(sh,11,15);
}
#define init_6_12_0_0
void FASTCALL call_6_12_0_0(SH2_struct* sh)
{
  SH2movbl(sh,12,0);
}
#define init_6_12_1_0
void FASTCALL call_6_12_1_0(SH2_struct* sh)
{
  SH2movbl(sh,12,1);
}
#define init_6_12_2_0
void FASTCALL call_6_12_2_0(SH2_struct* sh)
{
  SH2movbl(sh,12,2);
}
#define init_6_12_3_0
void FASTCALL call_6_12_3_0(SH2_struct* sh)
{
  SH2movbl(sh,12,3);
}
#define init_6_12_4_0
void FASTCALL call_6_12_4_0(SH2_struct* sh)
{
  SH2movbl(sh,12,4);
}
#define init_6_12_5_0
void FASTCALL call_6_12_5_0(SH2_struct* sh)
{
  SH2movbl(sh,12,5);
}
#define init_6_12_6_0
void FASTCALL call_6_12_6_0(SH2_struct* sh)
{
  SH2movbl(sh,12,6);
}
#define init_6_12_7_0
void FASTCALL call_6_12_7_0(SH2_struct* sh)
{
  SH2movbl(sh,12,7);
}
#define init_6_12_8_0
void FASTCALL call_6_12_8_0(SH2_struct* sh)
{
  SH2movbl(sh,12,8);
}
#define init_6_12_9_0
void FASTCALL call_6_12_9_0(SH2_struct* sh)
{
  SH2movbl(sh,12,9);
}
#define init_6_12_10_0
void FASTCALL call_6_12_10_0(SH2_struct* sh)
{
  SH2movbl(sh,12,10);
}
#define init_6_12_11_0
void FASTCALL call_6_12_11_0(SH2_struct* sh)
{
  SH2movbl(sh,12,11);
}
#define init_6_12_12_0
void FASTCALL call_6_12_12_0(SH2_struct* sh)
{
  SH2movbl(sh,12,12);
}
#define init_6_12_13_0
void FASTCALL call_6_12_13_0(SH2_struct* sh)
{
  SH2movbl(sh,12,13);
}
#define init_6_12_14_0
void FASTCALL call_6_12_14_0(SH2_struct* sh)
{
  SH2movbl(sh,12,14);
}
#define init_6_12_15_0
void FASTCALL call_6_12_15_0(SH2_struct* sh)
{
  SH2movbl(sh,12,15);
}
#define init_6_13_0_0
void FASTCALL call_6_13_0_0(SH2_struct* sh)
{
  SH2movbl(sh,13,0);
}
#define init_6_13_1_0
void FASTCALL call_6_13_1_0(SH2_struct* sh)
{
  SH2movbl(sh,13,1);
}
#define init_6_13_2_0
void FASTCALL call_6_13_2_0(SH2_struct* sh)
{
  SH2movbl(sh,13,2);
}
#define init_6_13_3_0
void FASTCALL call_6_13_3_0(SH2_struct* sh)
{
  SH2movbl(sh,13,3);
}
#define init_6_13_4_0
void FASTCALL call_6_13_4_0(SH2_struct* sh)
{
  SH2movbl(sh,13,4);
}
#define init_6_13_5_0
void FASTCALL call_6_13_5_0(SH2_struct* sh)
{
  SH2movbl(sh,13,5);
}
#define init_6_13_6_0
void FASTCALL call_6_13_6_0(SH2_struct* sh)
{
  SH2movbl(sh,13,6);
}
#define init_6_13_7_0
void FASTCALL call_6_13_7_0(SH2_struct* sh)
{
  SH2movbl(sh,13,7);
}
#define init_6_13_8_0
void FASTCALL call_6_13_8_0(SH2_struct* sh)
{
  SH2movbl(sh,13,8);
}
#define init_6_13_9_0
void FASTCALL call_6_13_9_0(SH2_struct* sh)
{
  SH2movbl(sh,13,9);
}
#define init_6_13_10_0
void FASTCALL call_6_13_10_0(SH2_struct* sh)
{
  SH2movbl(sh,13,10);
}
#define init_6_13_11_0
void FASTCALL call_6_13_11_0(SH2_struct* sh)
{
  SH2movbl(sh,13,11);
}
#define init_6_13_12_0
void FASTCALL call_6_13_12_0(SH2_struct* sh)
{
  SH2movbl(sh,13,12);
}
#define init_6_13_13_0
void FASTCALL call_6_13_13_0(SH2_struct* sh)
{
  SH2movbl(sh,13,13);
}
#define init_6_13_14_0
void FASTCALL call_6_13_14_0(SH2_struct* sh)
{
  SH2movbl(sh,13,14);
}
#define init_6_13_15_0
void FASTCALL call_6_13_15_0(SH2_struct* sh)
{
  SH2movbl(sh,13,15);
}
#define init_6_14_0_0
void FASTCALL call_6_14_0_0(SH2_struct* sh)
{
  SH2movbl(sh,14,0);
}
#define init_6_14_1_0
void FASTCALL call_6_14_1_0(SH2_struct* sh)
{
  SH2movbl(sh,14,1);
}
#define init_6_14_2_0
void FASTCALL call_6_14_2_0(SH2_struct* sh)
{
  SH2movbl(sh,14,2);
}
#define init_6_14_3_0
void FASTCALL call_6_14_3_0(SH2_struct* sh)
{
  SH2movbl(sh,14,3);
}
#define init_6_14_4_0
void FASTCALL call_6_14_4_0(SH2_struct* sh)
{
  SH2movbl(sh,14,4);
}
#define init_6_14_5_0
void FASTCALL call_6_14_5_0(SH2_struct* sh)
{
  SH2movbl(sh,14,5);
}
#define init_6_14_6_0
void FASTCALL call_6_14_6_0(SH2_struct* sh)
{
  SH2movbl(sh,14,6);
}
#define init_6_14_7_0
void FASTCALL call_6_14_7_0(SH2_struct* sh)
{
  SH2movbl(sh,14,7);
}
#define init_6_14_8_0
void FASTCALL call_6_14_8_0(SH2_struct* sh)
{
  SH2movbl(sh,14,8);
}
#define init_6_14_9_0
void FASTCALL call_6_14_9_0(SH2_struct* sh)
{
  SH2movbl(sh,14,9);
}
#define init_6_14_10_0
void FASTCALL call_6_14_10_0(SH2_struct* sh)
{
  SH2movbl(sh,14,10);
}
#define init_6_14_11_0
void FASTCALL call_6_14_11_0(SH2_struct* sh)
{
  SH2movbl(sh,14,11);
}
#define init_6_14_12_0
void FASTCALL call_6_14_12_0(SH2_struct* sh)
{
  SH2movbl(sh,14,12);
}
#define init_6_14_13_0
void FASTCALL call_6_14_13_0(SH2_struct* sh)
{
  SH2movbl(sh,14,13);
}
#define init_6_14_14_0
void FASTCALL call_6_14_14_0(SH2_struct* sh)
{
  SH2movbl(sh,14,14);
}
#define init_6_14_15_0
void FASTCALL call_6_14_15_0(SH2_struct* sh)
{
  SH2movbl(sh,14,15);
}
#define init_6_15_0_0
void FASTCALL call_6_15_0_0(SH2_struct* sh)
{
  SH2movbl(sh,15,0);
}
#define init_6_15_1_0
void FASTCALL call_6_15_1_0(SH2_struct* sh)
{
  SH2movbl(sh,15,1);
}
#define init_6_15_2_0
void FASTCALL call_6_15_2_0(SH2_struct* sh)
{
  SH2movbl(sh,15,2);
}
#define init_6_15_3_0
void FASTCALL call_6_15_3_0(SH2_struct* sh)
{
  SH2movbl(sh,15,3);
}
#define init_6_15_4_0
void FASTCALL call_6_15_4_0(SH2_struct* sh)
{
  SH2movbl(sh,15,4);
}
#define init_6_15_5_0
void FASTCALL call_6_15_5_0(SH2_struct* sh)
{
  SH2movbl(sh,15,5);
}
#define init_6_15_6_0
void FASTCALL call_6_15_6_0(SH2_struct* sh)
{
  SH2movbl(sh,15,6);
}
#define init_6_15_7_0
void FASTCALL call_6_15_7_0(SH2_struct* sh)
{
  SH2movbl(sh,15,7);
}
#define init_6_15_8_0
void FASTCALL call_6_15_8_0(SH2_struct* sh)
{
  SH2movbl(sh,15,8);
}
#define init_6_15_9_0
void FASTCALL call_6_15_9_0(SH2_struct* sh)
{
  SH2movbl(sh,15,9);
}
#define init_6_15_10_0
void FASTCALL call_6_15_10_0(SH2_struct* sh)
{
  SH2movbl(sh,15,10);
}
#define init_6_15_11_0
void FASTCALL call_6_15_11_0(SH2_struct* sh)
{
  SH2movbl(sh,15,11);
}
#define init_6_15_12_0
void FASTCALL call_6_15_12_0(SH2_struct* sh)
{
  SH2movbl(sh,15,12);
}
#define init_6_15_13_0
void FASTCALL call_6_15_13_0(SH2_struct* sh)
{
  SH2movbl(sh,15,13);
}
#define init_6_15_14_0
void FASTCALL call_6_15_14_0(SH2_struct* sh)
{
  SH2movbl(sh,15,14);
}
#define init_6_15_15_0
void FASTCALL call_6_15_15_0(SH2_struct* sh)
{
  SH2movbl(sh,15,15);
}
#define init_12_7_0_0
void FASTCALL call_12_7_0_0(SH2_struct* sh)
{
  SH2mova(sh,(0<<4)+0);
}
#define init_12_7_0_1
void FASTCALL call_12_7_0_1(SH2_struct* sh)
{
  SH2mova(sh,(0<<4)+1);
}
#define init_12_7_0_2
void FASTCALL call_12_7_0_2(SH2_struct* sh)
{
  SH2mova(sh,(0<<4)+2);
}
#define init_12_7_0_3
void FASTCALL call_12_7_0_3(SH2_struct* sh)
{
  SH2mova(sh,(0<<4)+3);
}
#define init_12_7_0_4
void FASTCALL call_12_7_0_4(SH2_struct* sh)
{
  SH2mova(sh,(0<<4)+4);
}
#define init_12_7_0_5
void FASTCALL call_12_7_0_5(SH2_struct* sh)
{
  SH2mova(sh,(0<<4)+5);
}
#define init_12_7_0_6
void FASTCALL call_12_7_0_6(SH2_struct* sh)
{
  SH2mova(sh,(0<<4)+6);
}
#define init_12_7_0_7
void FASTCALL call_12_7_0_7(SH2_struct* sh)
{
  SH2mova(sh,(0<<4)+7);
}
#define init_12_7_0_8
void FASTCALL call_12_7_0_8(SH2_struct* sh)
{
  SH2mova(sh,(0<<4)+8);
}
#define init_12_7_0_9
void FASTCALL call_12_7_0_9(SH2_struct* sh)
{
  SH2mova(sh,(0<<4)+9);
}
#define init_12_7_0_10
void FASTCALL call_12_7_0_10(SH2_struct* sh)
{
  SH2mova(sh,(0<<4)+10);
}
#define init_12_7_0_11
void FASTCALL call_12_7_0_11(SH2_struct* sh)
{
  SH2mova(sh,(0<<4)+11);
}
#define init_12_7_0_12
void FASTCALL call_12_7_0_12(SH2_struct* sh)
{
  SH2mova(sh,(0<<4)+12);
}
#define init_12_7_0_13
void FASTCALL call_12_7_0_13(SH2_struct* sh)
{
  SH2mova(sh,(0<<4)+13);
}
#define init_12_7_0_14
void FASTCALL call_12_7_0_14(SH2_struct* sh)
{
  SH2mova(sh,(0<<4)+14);
}
#define init_12_7_0_15
void FASTCALL call_12_7_0_15(SH2_struct* sh)
{
  SH2mova(sh,(0<<4)+15);
}
#define init_12_7_1_0
void FASTCALL call_12_7_1_0(SH2_struct* sh)
{
  SH2mova(sh,(1<<4)+0);
}
#define init_12_7_1_1
void FASTCALL call_12_7_1_1(SH2_struct* sh)
{
  SH2mova(sh,(1<<4)+1);
}
#define init_12_7_1_2
void FASTCALL call_12_7_1_2(SH2_struct* sh)
{
  SH2mova(sh,(1<<4)+2);
}
#define init_12_7_1_3
void FASTCALL call_12_7_1_3(SH2_struct* sh)
{
  SH2mova(sh,(1<<4)+3);
}
#define init_12_7_1_4
void FASTCALL call_12_7_1_4(SH2_struct* sh)
{
  SH2mova(sh,(1<<4)+4);
}
#define init_12_7_1_5
void FASTCALL call_12_7_1_5(SH2_struct* sh)
{
  SH2mova(sh,(1<<4)+5);
}
#define init_12_7_1_6
void FASTCALL call_12_7_1_6(SH2_struct* sh)
{
  SH2mova(sh,(1<<4)+6);
}
#define init_12_7_1_7
void FASTCALL call_12_7_1_7(SH2_struct* sh)
{
  SH2mova(sh,(1<<4)+7);
}
#define init_12_7_1_8
void FASTCALL call_12_7_1_8(SH2_struct* sh)
{
  SH2mova(sh,(1<<4)+8);
}
#define init_12_7_1_9
void FASTCALL call_12_7_1_9(SH2_struct* sh)
{
  SH2mova(sh,(1<<4)+9);
}
#define init_12_7_1_10
void FASTCALL call_12_7_1_10(SH2_struct* sh)
{
  SH2mova(sh,(1<<4)+10);
}
#define init_12_7_1_11
void FASTCALL call_12_7_1_11(SH2_struct* sh)
{
  SH2mova(sh,(1<<4)+11);
}
#define init_12_7_1_12
void FASTCALL call_12_7_1_12(SH2_struct* sh)
{
  SH2mova(sh,(1<<4)+12);
}
#define init_12_7_1_13
void FASTCALL call_12_7_1_13(SH2_struct* sh)
{
  SH2mova(sh,(1<<4)+13);
}
#define init_12_7_1_14
void FASTCALL call_12_7_1_14(SH2_struct* sh)
{
  SH2mova(sh,(1<<4)+14);
}
#define init_12_7_1_15
void FASTCALL call_12_7_1_15(SH2_struct* sh)
{
  SH2mova(sh,(1<<4)+15);
}
#define init_12_7_2_0
void FASTCALL call_12_7_2_0(SH2_struct* sh)
{
  SH2mova(sh,(2<<4)+0);
}
#define init_12_7_2_1
void FASTCALL call_12_7_2_1(SH2_struct* sh)
{
  SH2mova(sh,(2<<4)+1);
}
#define init_12_7_2_2
void FASTCALL call_12_7_2_2(SH2_struct* sh)
{
  SH2mova(sh,(2<<4)+2);
}
#define init_12_7_2_3
void FASTCALL call_12_7_2_3(SH2_struct* sh)
{
  SH2mova(sh,(2<<4)+3);
}
#define init_12_7_2_4
void FASTCALL call_12_7_2_4(SH2_struct* sh)
{
  SH2mova(sh,(2<<4)+4);
}
#define init_12_7_2_5
void FASTCALL call_12_7_2_5(SH2_struct* sh)
{
  SH2mova(sh,(2<<4)+5);
}
#define init_12_7_2_6
void FASTCALL call_12_7_2_6(SH2_struct* sh)
{
  SH2mova(sh,(2<<4)+6);
}
#define init_12_7_2_7
void FASTCALL call_12_7_2_7(SH2_struct* sh)
{
  SH2mova(sh,(2<<4)+7);
}
#define init_12_7_2_8
void FASTCALL call_12_7_2_8(SH2_struct* sh)
{
  SH2mova(sh,(2<<4)+8);
}
#define init_12_7_2_9
void FASTCALL call_12_7_2_9(SH2_struct* sh)
{
  SH2mova(sh,(2<<4)+9);
}
#define init_12_7_2_10
void FASTCALL call_12_7_2_10(SH2_struct* sh)
{
  SH2mova(sh,(2<<4)+10);
}
#define init_12_7_2_11
void FASTCALL call_12_7_2_11(SH2_struct* sh)
{
  SH2mova(sh,(2<<4)+11);
}
#define init_12_7_2_12
void FASTCALL call_12_7_2_12(SH2_struct* sh)
{
  SH2mova(sh,(2<<4)+12);
}
#define init_12_7_2_13
void FASTCALL call_12_7_2_13(SH2_struct* sh)
{
  SH2mova(sh,(2<<4)+13);
}
#define init_12_7_2_14
void FASTCALL call_12_7_2_14(SH2_struct* sh)
{
  SH2mova(sh,(2<<4)+14);
}
#define init_12_7_2_15
void FASTCALL call_12_7_2_15(SH2_struct* sh)
{
  SH2mova(sh,(2<<4)+15);
}
#define init_12_7_3_0
void FASTCALL call_12_7_3_0(SH2_struct* sh)
{
  SH2mova(sh,(3<<4)+0);
}
#define init_12_7_3_1
void FASTCALL call_12_7_3_1(SH2_struct* sh)
{
  SH2mova(sh,(3<<4)+1);
}
#define init_12_7_3_2
void FASTCALL call_12_7_3_2(SH2_struct* sh)
{
  SH2mova(sh,(3<<4)+2);
}
#define init_12_7_3_3
void FASTCALL call_12_7_3_3(SH2_struct* sh)
{
  SH2mova(sh,(3<<4)+3);
}
#define init_12_7_3_4
void FASTCALL call_12_7_3_4(SH2_struct* sh)
{
  SH2mova(sh,(3<<4)+4);
}
#define init_12_7_3_5
void FASTCALL call_12_7_3_5(SH2_struct* sh)
{
  SH2mova(sh,(3<<4)+5);
}
#define init_12_7_3_6
void FASTCALL call_12_7_3_6(SH2_struct* sh)
{
  SH2mova(sh,(3<<4)+6);
}
#define init_12_7_3_7
void FASTCALL call_12_7_3_7(SH2_struct* sh)
{
  SH2mova(sh,(3<<4)+7);
}
#define init_12_7_3_8
void FASTCALL call_12_7_3_8(SH2_struct* sh)
{
  SH2mova(sh,(3<<4)+8);
}
#define init_12_7_3_9
void FASTCALL call_12_7_3_9(SH2_struct* sh)
{
  SH2mova(sh,(3<<4)+9);
}
#define init_12_7_3_10
void FASTCALL call_12_7_3_10(SH2_struct* sh)
{
  SH2mova(sh,(3<<4)+10);
}
#define init_12_7_3_11
void FASTCALL call_12_7_3_11(SH2_struct* sh)
{
  SH2mova(sh,(3<<4)+11);
}
#define init_12_7_3_12
void FASTCALL call_12_7_3_12(SH2_struct* sh)
{
  SH2mova(sh,(3<<4)+12);
}
#define init_12_7_3_13
void FASTCALL call_12_7_3_13(SH2_struct* sh)
{
  SH2mova(sh,(3<<4)+13);
}
#define init_12_7_3_14
void FASTCALL call_12_7_3_14(SH2_struct* sh)
{
  SH2mova(sh,(3<<4)+14);
}
#define init_12_7_3_15
void FASTCALL call_12_7_3_15(SH2_struct* sh)
{
  SH2mova(sh,(3<<4)+15);
}
#define init_12_7_4_0
void FASTCALL call_12_7_4_0(SH2_struct* sh)
{
  SH2mova(sh,(4<<4)+0);
}
#define init_12_7_4_1
void FASTCALL call_12_7_4_1(SH2_struct* sh)
{
  SH2mova(sh,(4<<4)+1);
}
#define init_12_7_4_2
void FASTCALL call_12_7_4_2(SH2_struct* sh)
{
  SH2mova(sh,(4<<4)+2);
}
#define init_12_7_4_3
void FASTCALL call_12_7_4_3(SH2_struct* sh)
{
  SH2mova(sh,(4<<4)+3);
}
#define init_12_7_4_4
void FASTCALL call_12_7_4_4(SH2_struct* sh)
{
  SH2mova(sh,(4<<4)+4);
}
#define init_12_7_4_5
void FASTCALL call_12_7_4_5(SH2_struct* sh)
{
  SH2mova(sh,(4<<4)+5);
}
#define init_12_7_4_6
void FASTCALL call_12_7_4_6(SH2_struct* sh)
{
  SH2mova(sh,(4<<4)+6);
}
#define init_12_7_4_7
void FASTCALL call_12_7_4_7(SH2_struct* sh)
{
  SH2mova(sh,(4<<4)+7);
}
#define init_12_7_4_8
void FASTCALL call_12_7_4_8(SH2_struct* sh)
{
  SH2mova(sh,(4<<4)+8);
}
#define init_12_7_4_9
void FASTCALL call_12_7_4_9(SH2_struct* sh)
{
  SH2mova(sh,(4<<4)+9);
}
#define init_12_7_4_10
void FASTCALL call_12_7_4_10(SH2_struct* sh)
{
  SH2mova(sh,(4<<4)+10);
}
#define init_12_7_4_11
void FASTCALL call_12_7_4_11(SH2_struct* sh)
{
  SH2mova(sh,(4<<4)+11);
}
#define init_12_7_4_12
void FASTCALL call_12_7_4_12(SH2_struct* sh)
{
  SH2mova(sh,(4<<4)+12);
}
#define init_12_7_4_13
void FASTCALL call_12_7_4_13(SH2_struct* sh)
{
  SH2mova(sh,(4<<4)+13);
}
#define init_12_7_4_14
void FASTCALL call_12_7_4_14(SH2_struct* sh)
{
  SH2mova(sh,(4<<4)+14);
}
#define init_12_7_4_15
void FASTCALL call_12_7_4_15(SH2_struct* sh)
{
  SH2mova(sh,(4<<4)+15);
}
#define init_12_7_5_0
void FASTCALL call_12_7_5_0(SH2_struct* sh)
{
  SH2mova(sh,(5<<4)+0);
}
#define init_12_7_5_1
void FASTCALL call_12_7_5_1(SH2_struct* sh)
{
  SH2mova(sh,(5<<4)+1);
}
#define init_12_7_5_2
void FASTCALL call_12_7_5_2(SH2_struct* sh)
{
  SH2mova(sh,(5<<4)+2);
}
#define init_12_7_5_3
void FASTCALL call_12_7_5_3(SH2_struct* sh)
{
  SH2mova(sh,(5<<4)+3);
}
#define init_12_7_5_4
void FASTCALL call_12_7_5_4(SH2_struct* sh)
{
  SH2mova(sh,(5<<4)+4);
}
#define init_12_7_5_5
void FASTCALL call_12_7_5_5(SH2_struct* sh)
{
  SH2mova(sh,(5<<4)+5);
}
#define init_12_7_5_6
void FASTCALL call_12_7_5_6(SH2_struct* sh)
{
  SH2mova(sh,(5<<4)+6);
}
#define init_12_7_5_7
void FASTCALL call_12_7_5_7(SH2_struct* sh)
{
  SH2mova(sh,(5<<4)+7);
}
#define init_12_7_5_8
void FASTCALL call_12_7_5_8(SH2_struct* sh)
{
  SH2mova(sh,(5<<4)+8);
}
#define init_12_7_5_9
void FASTCALL call_12_7_5_9(SH2_struct* sh)
{
  SH2mova(sh,(5<<4)+9);
}
#define init_12_7_5_10
void FASTCALL call_12_7_5_10(SH2_struct* sh)
{
  SH2mova(sh,(5<<4)+10);
}
#define init_12_7_5_11
void FASTCALL call_12_7_5_11(SH2_struct* sh)
{
  SH2mova(sh,(5<<4)+11);
}
#define init_12_7_5_12
void FASTCALL call_12_7_5_12(SH2_struct* sh)
{
  SH2mova(sh,(5<<4)+12);
}
#define init_12_7_5_13
void FASTCALL call_12_7_5_13(SH2_struct* sh)
{
  SH2mova(sh,(5<<4)+13);
}
#define init_12_7_5_14
void FASTCALL call_12_7_5_14(SH2_struct* sh)
{
  SH2mova(sh,(5<<4)+14);
}
#define init_12_7_5_15
void FASTCALL call_12_7_5_15(SH2_struct* sh)
{
  SH2mova(sh,(5<<4)+15);
}
#define init_12_7_6_0
void FASTCALL call_12_7_6_0(SH2_struct* sh)
{
  SH2mova(sh,(6<<4)+0);
}
#define init_12_7_6_1
void FASTCALL call_12_7_6_1(SH2_struct* sh)
{
  SH2mova(sh,(6<<4)+1);
}
#define init_12_7_6_2
void FASTCALL call_12_7_6_2(SH2_struct* sh)
{
  SH2mova(sh,(6<<4)+2);
}
#define init_12_7_6_3
void FASTCALL call_12_7_6_3(SH2_struct* sh)
{
  SH2mova(sh,(6<<4)+3);
}
#define init_12_7_6_4
void FASTCALL call_12_7_6_4(SH2_struct* sh)
{
  SH2mova(sh,(6<<4)+4);
}
#define init_12_7_6_5
void FASTCALL call_12_7_6_5(SH2_struct* sh)
{
  SH2mova(sh,(6<<4)+5);
}
#define init_12_7_6_6
void FASTCALL call_12_7_6_6(SH2_struct* sh)
{
  SH2mova(sh,(6<<4)+6);
}
#define init_12_7_6_7
void FASTCALL call_12_7_6_7(SH2_struct* sh)
{
  SH2mova(sh,(6<<4)+7);
}
#define init_12_7_6_8
void FASTCALL call_12_7_6_8(SH2_struct* sh)
{
  SH2mova(sh,(6<<4)+8);
}
#define init_12_7_6_9
void FASTCALL call_12_7_6_9(SH2_struct* sh)
{
  SH2mova(sh,(6<<4)+9);
}
#define init_12_7_6_10
void FASTCALL call_12_7_6_10(SH2_struct* sh)
{
  SH2mova(sh,(6<<4)+10);
}
#define init_12_7_6_11
void FASTCALL call_12_7_6_11(SH2_struct* sh)
{
  SH2mova(sh,(6<<4)+11);
}
#define init_12_7_6_12
void FASTCALL call_12_7_6_12(SH2_struct* sh)
{
  SH2mova(sh,(6<<4)+12);
}
#define init_12_7_6_13
void FASTCALL call_12_7_6_13(SH2_struct* sh)
{
  SH2mova(sh,(6<<4)+13);
}
#define init_12_7_6_14
void FASTCALL call_12_7_6_14(SH2_struct* sh)
{
  SH2mova(sh,(6<<4)+14);
}
#define init_12_7_6_15
void FASTCALL call_12_7_6_15(SH2_struct* sh)
{
  SH2mova(sh,(6<<4)+15);
}
#define init_12_7_7_0
void FASTCALL call_12_7_7_0(SH2_struct* sh)
{
  SH2mova(sh,(7<<4)+0);
}
#define init_12_7_7_1
void FASTCALL call_12_7_7_1(SH2_struct* sh)
{
  SH2mova(sh,(7<<4)+1);
}
#define init_12_7_7_2
void FASTCALL call_12_7_7_2(SH2_struct* sh)
{
  SH2mova(sh,(7<<4)+2);
}
#define init_12_7_7_3
void FASTCALL call_12_7_7_3(SH2_struct* sh)
{
  SH2mova(sh,(7<<4)+3);
}
#define init_12_7_7_4
void FASTCALL call_12_7_7_4(SH2_struct* sh)
{
  SH2mova(sh,(7<<4)+4);
}
#define init_12_7_7_5
void FASTCALL call_12_7_7_5(SH2_struct* sh)
{
  SH2mova(sh,(7<<4)+5);
}
#define init_12_7_7_6
void FASTCALL call_12_7_7_6(SH2_struct* sh)
{
  SH2mova(sh,(7<<4)+6);
}
#define init_12_7_7_7
void FASTCALL call_12_7_7_7(SH2_struct* sh)
{
  SH2mova(sh,(7<<4)+7);
}
#define init_12_7_7_8
void FASTCALL call_12_7_7_8(SH2_struct* sh)
{
  SH2mova(sh,(7<<4)+8);
}
#define init_12_7_7_9
void FASTCALL call_12_7_7_9(SH2_struct* sh)
{
  SH2mova(sh,(7<<4)+9);
}
#define init_12_7_7_10
void FASTCALL call_12_7_7_10(SH2_struct* sh)
{
  SH2mova(sh,(7<<4)+10);
}
#define init_12_7_7_11
void FASTCALL call_12_7_7_11(SH2_struct* sh)
{
  SH2mova(sh,(7<<4)+11);
}
#define init_12_7_7_12
void FASTCALL call_12_7_7_12(SH2_struct* sh)
{
  SH2mova(sh,(7<<4)+12);
}
#define init_12_7_7_13
void FASTCALL call_12_7_7_13(SH2_struct* sh)
{
  SH2mova(sh,(7<<4)+13);
}
#define init_12_7_7_14
void FASTCALL call_12_7_7_14(SH2_struct* sh)
{
  SH2mova(sh,(7<<4)+14);
}
#define init_12_7_7_15
void FASTCALL call_12_7_7_15(SH2_struct* sh)
{
  SH2mova(sh,(7<<4)+15);
}
#define init_12_7_8_0
void FASTCALL call_12_7_8_0(SH2_struct* sh)
{
  SH2mova(sh,(8<<4)+0);
}
#define init_12_7_8_1
void FASTCALL call_12_7_8_1(SH2_struct* sh)
{
  SH2mova(sh,(8<<4)+1);
}
#define init_12_7_8_2
void FASTCALL call_12_7_8_2(SH2_struct* sh)
{
  SH2mova(sh,(8<<4)+2);
}
#define init_12_7_8_3
void FASTCALL call_12_7_8_3(SH2_struct* sh)
{
  SH2mova(sh,(8<<4)+3);
}
#define init_12_7_8_4
void FASTCALL call_12_7_8_4(SH2_struct* sh)
{
  SH2mova(sh,(8<<4)+4);
}
#define init_12_7_8_5
void FASTCALL call_12_7_8_5(SH2_struct* sh)
{
  SH2mova(sh,(8<<4)+5);
}
#define init_12_7_8_6
void FASTCALL call_12_7_8_6(SH2_struct* sh)
{
  SH2mova(sh,(8<<4)+6);
}
#define init_12_7_8_7
void FASTCALL call_12_7_8_7(SH2_struct* sh)
{
  SH2mova(sh,(8<<4)+7);
}
#define init_12_7_8_8
void FASTCALL call_12_7_8_8(SH2_struct* sh)
{
  SH2mova(sh,(8<<4)+8);
}
#define init_12_7_8_9
void FASTCALL call_12_7_8_9(SH2_struct* sh)
{
  SH2mova(sh,(8<<4)+9);
}
#define init_12_7_8_10
void FASTCALL call_12_7_8_10(SH2_struct* sh)
{
  SH2mova(sh,(8<<4)+10);
}
#define init_12_7_8_11
void FASTCALL call_12_7_8_11(SH2_struct* sh)
{
  SH2mova(sh,(8<<4)+11);
}
#define init_12_7_8_12
void FASTCALL call_12_7_8_12(SH2_struct* sh)
{
  SH2mova(sh,(8<<4)+12);
}
#define init_12_7_8_13
void FASTCALL call_12_7_8_13(SH2_struct* sh)
{
  SH2mova(sh,(8<<4)+13);
}
#define init_12_7_8_14
void FASTCALL call_12_7_8_14(SH2_struct* sh)
{
  SH2mova(sh,(8<<4)+14);
}
#define init_12_7_8_15
void FASTCALL call_12_7_8_15(SH2_struct* sh)
{
  SH2mova(sh,(8<<4)+15);
}
#define init_12_7_9_0
void FASTCALL call_12_7_9_0(SH2_struct* sh)
{
  SH2mova(sh,(9<<4)+0);
}
#define init_12_7_9_1
void FASTCALL call_12_7_9_1(SH2_struct* sh)
{
  SH2mova(sh,(9<<4)+1);
}
#define init_12_7_9_2
void FASTCALL call_12_7_9_2(SH2_struct* sh)
{
  SH2mova(sh,(9<<4)+2);
}
#define init_12_7_9_3
void FASTCALL call_12_7_9_3(SH2_struct* sh)
{
  SH2mova(sh,(9<<4)+3);
}
#define init_12_7_9_4
void FASTCALL call_12_7_9_4(SH2_struct* sh)
{
  SH2mova(sh,(9<<4)+4);
}
#define init_12_7_9_5
void FASTCALL call_12_7_9_5(SH2_struct* sh)
{
  SH2mova(sh,(9<<4)+5);
}
#define init_12_7_9_6
void FASTCALL call_12_7_9_6(SH2_struct* sh)
{
  SH2mova(sh,(9<<4)+6);
}
#define init_12_7_9_7
void FASTCALL call_12_7_9_7(SH2_struct* sh)
{
  SH2mova(sh,(9<<4)+7);
}
#define init_12_7_9_8
void FASTCALL call_12_7_9_8(SH2_struct* sh)
{
  SH2mova(sh,(9<<4)+8);
}
#define init_12_7_9_9
void FASTCALL call_12_7_9_9(SH2_struct* sh)
{
  SH2mova(sh,(9<<4)+9);
}
#define init_12_7_9_10
void FASTCALL call_12_7_9_10(SH2_struct* sh)
{
  SH2mova(sh,(9<<4)+10);
}
#define init_12_7_9_11
void FASTCALL call_12_7_9_11(SH2_struct* sh)
{
  SH2mova(sh,(9<<4)+11);
}
#define init_12_7_9_12
void FASTCALL call_12_7_9_12(SH2_struct* sh)
{
  SH2mova(sh,(9<<4)+12);
}
#define init_12_7_9_13
void FASTCALL call_12_7_9_13(SH2_struct* sh)
{
  SH2mova(sh,(9<<4)+13);
}
#define init_12_7_9_14
void FASTCALL call_12_7_9_14(SH2_struct* sh)
{
  SH2mova(sh,(9<<4)+14);
}
#define init_12_7_9_15
void FASTCALL call_12_7_9_15(SH2_struct* sh)
{
  SH2mova(sh,(9<<4)+15);
}
#define init_12_7_10_0
void FASTCALL call_12_7_10_0(SH2_struct* sh)
{
  SH2mova(sh,(10<<4)+0);
}
#define init_12_7_10_1
void FASTCALL call_12_7_10_1(SH2_struct* sh)
{
  SH2mova(sh,(10<<4)+1);
}
#define init_12_7_10_2
void FASTCALL call_12_7_10_2(SH2_struct* sh)
{
  SH2mova(sh,(10<<4)+2);
}
#define init_12_7_10_3
void FASTCALL call_12_7_10_3(SH2_struct* sh)
{
  SH2mova(sh,(10<<4)+3);
}
#define init_12_7_10_4
void FASTCALL call_12_7_10_4(SH2_struct* sh)
{
  SH2mova(sh,(10<<4)+4);
}
#define init_12_7_10_5
void FASTCALL call_12_7_10_5(SH2_struct* sh)
{
  SH2mova(sh,(10<<4)+5);
}
#define init_12_7_10_6
void FASTCALL call_12_7_10_6(SH2_struct* sh)
{
  SH2mova(sh,(10<<4)+6);
}
#define init_12_7_10_7
void FASTCALL call_12_7_10_7(SH2_struct* sh)
{
  SH2mova(sh,(10<<4)+7);
}
#define init_12_7_10_8
void FASTCALL call_12_7_10_8(SH2_struct* sh)
{
  SH2mova(sh,(10<<4)+8);
}
#define init_12_7_10_9
void FASTCALL call_12_7_10_9(SH2_struct* sh)
{
  SH2mova(sh,(10<<4)+9);
}
#define init_12_7_10_10
void FASTCALL call_12_7_10_10(SH2_struct* sh)
{
  SH2mova(sh,(10<<4)+10);
}
#define init_12_7_10_11
void FASTCALL call_12_7_10_11(SH2_struct* sh)
{
  SH2mova(sh,(10<<4)+11);
}
#define init_12_7_10_12
void FASTCALL call_12_7_10_12(SH2_struct* sh)
{
  SH2mova(sh,(10<<4)+12);
}
#define init_12_7_10_13
void FASTCALL call_12_7_10_13(SH2_struct* sh)
{
  SH2mova(sh,(10<<4)+13);
}
#define init_12_7_10_14
void FASTCALL call_12_7_10_14(SH2_struct* sh)
{
  SH2mova(sh,(10<<4)+14);
}
#define init_12_7_10_15
void FASTCALL call_12_7_10_15(SH2_struct* sh)
{
  SH2mova(sh,(10<<4)+15);
}
#define init_12_7_11_0
void FASTCALL call_12_7_11_0(SH2_struct* sh)
{
  SH2mova(sh,(11<<4)+0);
}
#define init_12_7_11_1
void FASTCALL call_12_7_11_1(SH2_struct* sh)
{
  SH2mova(sh,(11<<4)+1);
}
#define init_12_7_11_2
void FASTCALL call_12_7_11_2(SH2_struct* sh)
{
  SH2mova(sh,(11<<4)+2);
}
#define init_12_7_11_3
void FASTCALL call_12_7_11_3(SH2_struct* sh)
{
  SH2mova(sh,(11<<4)+3);
}
#define init_12_7_11_4
void FASTCALL call_12_7_11_4(SH2_struct* sh)
{
  SH2mova(sh,(11<<4)+4);
}
#define init_12_7_11_5
void FASTCALL call_12_7_11_5(SH2_struct* sh)
{
  SH2mova(sh,(11<<4)+5);
}
#define init_12_7_11_6
void FASTCALL call_12_7_11_6(SH2_struct* sh)
{
  SH2mova(sh,(11<<4)+6);
}
#define init_12_7_11_7
void FASTCALL call_12_7_11_7(SH2_struct* sh)
{
  SH2mova(sh,(11<<4)+7);
}
#define init_12_7_11_8
void FASTCALL call_12_7_11_8(SH2_struct* sh)
{
  SH2mova(sh,(11<<4)+8);
}
#define init_12_7_11_9
void FASTCALL call_12_7_11_9(SH2_struct* sh)
{
  SH2mova(sh,(11<<4)+9);
}
#define init_12_7_11_10
void FASTCALL call_12_7_11_10(SH2_struct* sh)
{
  SH2mova(sh,(11<<4)+10);
}
#define init_12_7_11_11
void FASTCALL call_12_7_11_11(SH2_struct* sh)
{
  SH2mova(sh,(11<<4)+11);
}
#define init_12_7_11_12
void FASTCALL call_12_7_11_12(SH2_struct* sh)
{
  SH2mova(sh,(11<<4)+12);
}
#define init_12_7_11_13
void FASTCALL call_12_7_11_13(SH2_struct* sh)
{
  SH2mova(sh,(11<<4)+13);
}
#define init_12_7_11_14
void FASTCALL call_12_7_11_14(SH2_struct* sh)
{
  SH2mova(sh,(11<<4)+14);
}
#define init_12_7_11_15
void FASTCALL call_12_7_11_15(SH2_struct* sh)
{
  SH2mova(sh,(11<<4)+15);
}
#define init_12_7_12_0
void FASTCALL call_12_7_12_0(SH2_struct* sh)
{
  SH2mova(sh,(12<<4)+0);
}
#define init_12_7_12_1
void FASTCALL call_12_7_12_1(SH2_struct* sh)
{
  SH2mova(sh,(12<<4)+1);
}
#define init_12_7_12_2
void FASTCALL call_12_7_12_2(SH2_struct* sh)
{
  SH2mova(sh,(12<<4)+2);
}
#define init_12_7_12_3
void FASTCALL call_12_7_12_3(SH2_struct* sh)
{
  SH2mova(sh,(12<<4)+3);
}
#define init_12_7_12_4
void FASTCALL call_12_7_12_4(SH2_struct* sh)
{
  SH2mova(sh,(12<<4)+4);
}
#define init_12_7_12_5
void FASTCALL call_12_7_12_5(SH2_struct* sh)
{
  SH2mova(sh,(12<<4)+5);
}
#define init_12_7_12_6
void FASTCALL call_12_7_12_6(SH2_struct* sh)
{
  SH2mova(sh,(12<<4)+6);
}
#define init_12_7_12_7
void FASTCALL call_12_7_12_7(SH2_struct* sh)
{
  SH2mova(sh,(12<<4)+7);
}
#define init_12_7_12_8
void FASTCALL call_12_7_12_8(SH2_struct* sh)
{
  SH2mova(sh,(12<<4)+8);
}
#define init_12_7_12_9
void FASTCALL call_12_7_12_9(SH2_struct* sh)
{
  SH2mova(sh,(12<<4)+9);
}
#define init_12_7_12_10
void FASTCALL call_12_7_12_10(SH2_struct* sh)
{
  SH2mova(sh,(12<<4)+10);
}
#define init_12_7_12_11
void FASTCALL call_12_7_12_11(SH2_struct* sh)
{
  SH2mova(sh,(12<<4)+11);
}
#define init_12_7_12_12
void FASTCALL call_12_7_12_12(SH2_struct* sh)
{
  SH2mova(sh,(12<<4)+12);
}
#define init_12_7_12_13
void FASTCALL call_12_7_12_13(SH2_struct* sh)
{
  SH2mova(sh,(12<<4)+13);
}
#define init_12_7_12_14
void FASTCALL call_12_7_12_14(SH2_struct* sh)
{
  SH2mova(sh,(12<<4)+14);
}
#define init_12_7_12_15
void FASTCALL call_12_7_12_15(SH2_struct* sh)
{
  SH2mova(sh,(12<<4)+15);
}
#define init_12_7_13_0
void FASTCALL call_12_7_13_0(SH2_struct* sh)
{
  SH2mova(sh,(13<<4)+0);
}
#define init_12_7_13_1
void FASTCALL call_12_7_13_1(SH2_struct* sh)
{
  SH2mova(sh,(13<<4)+1);
}
#define init_12_7_13_2
void FASTCALL call_12_7_13_2(SH2_struct* sh)
{
  SH2mova(sh,(13<<4)+2);
}
#define init_12_7_13_3
void FASTCALL call_12_7_13_3(SH2_struct* sh)
{
  SH2mova(sh,(13<<4)+3);
}
#define init_12_7_13_4
void FASTCALL call_12_7_13_4(SH2_struct* sh)
{
  SH2mova(sh,(13<<4)+4);
}
#define init_12_7_13_5
void FASTCALL call_12_7_13_5(SH2_struct* sh)
{
  SH2mova(sh,(13<<4)+5);
}
#define init_12_7_13_6
void FASTCALL call_12_7_13_6(SH2_struct* sh)
{
  SH2mova(sh,(13<<4)+6);
}
#define init_12_7_13_7
void FASTCALL call_12_7_13_7(SH2_struct* sh)
{
  SH2mova(sh,(13<<4)+7);
}
#define init_12_7_13_8
void FASTCALL call_12_7_13_8(SH2_struct* sh)
{
  SH2mova(sh,(13<<4)+8);
}
#define init_12_7_13_9
void FASTCALL call_12_7_13_9(SH2_struct* sh)
{
  SH2mova(sh,(13<<4)+9);
}
#define init_12_7_13_10
void FASTCALL call_12_7_13_10(SH2_struct* sh)
{
  SH2mova(sh,(13<<4)+10);
}
#define init_12_7_13_11
void FASTCALL call_12_7_13_11(SH2_struct* sh)
{
  SH2mova(sh,(13<<4)+11);
}
#define init_12_7_13_12
void FASTCALL call_12_7_13_12(SH2_struct* sh)
{
  SH2mova(sh,(13<<4)+12);
}
#define init_12_7_13_13
void FASTCALL call_12_7_13_13(SH2_struct* sh)
{
  SH2mova(sh,(13<<4)+13);
}
#define init_12_7_13_14
void FASTCALL call_12_7_13_14(SH2_struct* sh)
{
  SH2mova(sh,(13<<4)+14);
}
#define init_12_7_13_15
void FASTCALL call_12_7_13_15(SH2_struct* sh)
{
  SH2mova(sh,(13<<4)+15);
}
#define init_12_7_14_0
void FASTCALL call_12_7_14_0(SH2_struct* sh)
{
  SH2mova(sh,(14<<4)+0);
}
#define init_12_7_14_1
void FASTCALL call_12_7_14_1(SH2_struct* sh)
{
  SH2mova(sh,(14<<4)+1);
}
#define init_12_7_14_2
void FASTCALL call_12_7_14_2(SH2_struct* sh)
{
  SH2mova(sh,(14<<4)+2);
}
#define init_12_7_14_3
void FASTCALL call_12_7_14_3(SH2_struct* sh)
{
  SH2mova(sh,(14<<4)+3);
}
#define init_12_7_14_4
void FASTCALL call_12_7_14_4(SH2_struct* sh)
{
  SH2mova(sh,(14<<4)+4);
}
#define init_12_7_14_5
void FASTCALL call_12_7_14_5(SH2_struct* sh)
{
  SH2mova(sh,(14<<4)+5);
}
#define init_12_7_14_6
void FASTCALL call_12_7_14_6(SH2_struct* sh)
{
  SH2mova(sh,(14<<4)+6);
}
#define init_12_7_14_7
void FASTCALL call_12_7_14_7(SH2_struct* sh)
{
  SH2mova(sh,(14<<4)+7);
}
#define init_12_7_14_8
void FASTCALL call_12_7_14_8(SH2_struct* sh)
{
  SH2mova(sh,(14<<4)+8);
}
#define init_12_7_14_9
void FASTCALL call_12_7_14_9(SH2_struct* sh)
{
  SH2mova(sh,(14<<4)+9);
}
#define init_12_7_14_10
void FASTCALL call_12_7_14_10(SH2_struct* sh)
{
  SH2mova(sh,(14<<4)+10);
}
#define init_12_7_14_11
void FASTCALL call_12_7_14_11(SH2_struct* sh)
{
  SH2mova(sh,(14<<4)+11);
}
#define init_12_7_14_12
void FASTCALL call_12_7_14_12(SH2_struct* sh)
{
  SH2mova(sh,(14<<4)+12);
}
#define init_12_7_14_13
void FASTCALL call_12_7_14_13(SH2_struct* sh)
{
  SH2mova(sh,(14<<4)+13);
}
#define init_12_7_14_14
void FASTCALL call_12_7_14_14(SH2_struct* sh)
{
  SH2mova(sh,(14<<4)+14);
}
#define init_12_7_14_15
void FASTCALL call_12_7_14_15(SH2_struct* sh)
{
  SH2mova(sh,(14<<4)+15);
}
#define init_12_7_15_0
void FASTCALL call_12_7_15_0(SH2_struct* sh)
{
  SH2mova(sh,(15<<4)+0);
}
#define init_12_7_15_1
void FASTCALL call_12_7_15_1(SH2_struct* sh)
{
  SH2mova(sh,(15<<4)+1);
}
#define init_12_7_15_2
void FASTCALL call_12_7_15_2(SH2_struct* sh)
{
  SH2mova(sh,(15<<4)+2);
}
#define init_12_7_15_3
void FASTCALL call_12_7_15_3(SH2_struct* sh)
{
  SH2mova(sh,(15<<4)+3);
}
#define init_12_7_15_4
void FASTCALL call_12_7_15_4(SH2_struct* sh)
{
  SH2mova(sh,(15<<4)+4);
}
#define init_12_7_15_5
void FASTCALL call_12_7_15_5(SH2_struct* sh)
{
  SH2mova(sh,(15<<4)+5);
}
#define init_12_7_15_6
void FASTCALL call_12_7_15_6(SH2_struct* sh)
{
  SH2mova(sh,(15<<4)+6);
}
#define init_12_7_15_7
void FASTCALL call_12_7_15_7(SH2_struct* sh)
{
  SH2mova(sh,(15<<4)+7);
}
#define init_12_7_15_8
void FASTCALL call_12_7_15_8(SH2_struct* sh)
{
  SH2mova(sh,(15<<4)+8);
}
#define init_12_7_15_9
void FASTCALL call_12_7_15_9(SH2_struct* sh)
{
  SH2mova(sh,(15<<4)+9);
}
#define init_12_7_15_10
void FASTCALL call_12_7_15_10(SH2_struct* sh)
{
  SH2mova(sh,(15<<4)+10);
}
#define init_12_7_15_11
void FASTCALL call_12_7_15_11(SH2_struct* sh)
{
  SH2mova(sh,(15<<4)+11);
}
#define init_12_7_15_12
void FASTCALL call_12_7_15_12(SH2_struct* sh)
{
  SH2mova(sh,(15<<4)+12);
}
#define init_12_7_15_13
void FASTCALL call_12_7_15_13(SH2_struct* sh)
{
  SH2mova(sh,(15<<4)+13);
}
#define init_12_7_15_14
void FASTCALL call_12_7_15_14(SH2_struct* sh)
{
  SH2mova(sh,(15<<4)+14);
}
#define init_12_7_15_15
void FASTCALL call_12_7_15_15(SH2_struct* sh)
{
  SH2mova(sh,(15<<4)+15);
}
#define init_6_0_0_3
void FASTCALL call_6_0_0_3(SH2_struct* sh)
{
  SH2mov(sh,0,0);
}
#define init_6_0_1_3
void FASTCALL call_6_0_1_3(SH2_struct* sh)
{
  SH2mov(sh,0,1);
}
#define init_6_0_2_3
void FASTCALL call_6_0_2_3(SH2_struct* sh)
{
  SH2mov(sh,0,2);
}
#define init_6_0_3_3
void FASTCALL call_6_0_3_3(SH2_struct* sh)
{
  SH2mov(sh,0,3);
}
#define init_6_0_4_3
void FASTCALL call_6_0_4_3(SH2_struct* sh)
{
  SH2mov(sh,0,4);
}
#define init_6_0_5_3
void FASTCALL call_6_0_5_3(SH2_struct* sh)
{
  SH2mov(sh,0,5);
}
#define init_6_0_6_3
void FASTCALL call_6_0_6_3(SH2_struct* sh)
{
  SH2mov(sh,0,6);
}
#define init_6_0_7_3
void FASTCALL call_6_0_7_3(SH2_struct* sh)
{
  SH2mov(sh,0,7);
}
#define init_6_0_8_3
void FASTCALL call_6_0_8_3(SH2_struct* sh)
{
  SH2mov(sh,0,8);
}
#define init_6_0_9_3
void FASTCALL call_6_0_9_3(SH2_struct* sh)
{
  SH2mov(sh,0,9);
}
#define init_6_0_10_3
void FASTCALL call_6_0_10_3(SH2_struct* sh)
{
  SH2mov(sh,0,10);
}
#define init_6_0_11_3
void FASTCALL call_6_0_11_3(SH2_struct* sh)
{
  SH2mov(sh,0,11);
}
#define init_6_0_12_3
void FASTCALL call_6_0_12_3(SH2_struct* sh)
{
  SH2mov(sh,0,12);
}
#define init_6_0_13_3
void FASTCALL call_6_0_13_3(SH2_struct* sh)
{
  SH2mov(sh,0,13);
}
#define init_6_0_14_3
void FASTCALL call_6_0_14_3(SH2_struct* sh)
{
  SH2mov(sh,0,14);
}
#define init_6_0_15_3
void FASTCALL call_6_0_15_3(SH2_struct* sh)
{
  SH2mov(sh,0,15);
}
#define init_6_1_0_3
void FASTCALL call_6_1_0_3(SH2_struct* sh)
{
  SH2mov(sh,1,0);
}
#define init_6_1_1_3
void FASTCALL call_6_1_1_3(SH2_struct* sh)
{
  SH2mov(sh,1,1);
}
#define init_6_1_2_3
void FASTCALL call_6_1_2_3(SH2_struct* sh)
{
  SH2mov(sh,1,2);
}
#define init_6_1_3_3
void FASTCALL call_6_1_3_3(SH2_struct* sh)
{
  SH2mov(sh,1,3);
}
#define init_6_1_4_3
void FASTCALL call_6_1_4_3(SH2_struct* sh)
{
  SH2mov(sh,1,4);
}
#define init_6_1_5_3
void FASTCALL call_6_1_5_3(SH2_struct* sh)
{
  SH2mov(sh,1,5);
}
#define init_6_1_6_3
void FASTCALL call_6_1_6_3(SH2_struct* sh)
{
  SH2mov(sh,1,6);
}
#define init_6_1_7_3
void FASTCALL call_6_1_7_3(SH2_struct* sh)
{
  SH2mov(sh,1,7);
}
#define init_6_1_8_3
void FASTCALL call_6_1_8_3(SH2_struct* sh)
{
  SH2mov(sh,1,8);
}
#define init_6_1_9_3
void FASTCALL call_6_1_9_3(SH2_struct* sh)
{
  SH2mov(sh,1,9);
}
#define init_6_1_10_3
void FASTCALL call_6_1_10_3(SH2_struct* sh)
{
  SH2mov(sh,1,10);
}
#define init_6_1_11_3
void FASTCALL call_6_1_11_3(SH2_struct* sh)
{
  SH2mov(sh,1,11);
}
#define init_6_1_12_3
void FASTCALL call_6_1_12_3(SH2_struct* sh)
{
  SH2mov(sh,1,12);
}
#define init_6_1_13_3
void FASTCALL call_6_1_13_3(SH2_struct* sh)
{
  SH2mov(sh,1,13);
}
#define init_6_1_14_3
void FASTCALL call_6_1_14_3(SH2_struct* sh)
{
  SH2mov(sh,1,14);
}
#define init_6_1_15_3
void FASTCALL call_6_1_15_3(SH2_struct* sh)
{
  SH2mov(sh,1,15);
}
#define init_6_2_0_3
void FASTCALL call_6_2_0_3(SH2_struct* sh)
{
  SH2mov(sh,2,0);
}
#define init_6_2_1_3
void FASTCALL call_6_2_1_3(SH2_struct* sh)
{
  SH2mov(sh,2,1);
}
#define init_6_2_2_3
void FASTCALL call_6_2_2_3(SH2_struct* sh)
{
  SH2mov(sh,2,2);
}
#define init_6_2_3_3
void FASTCALL call_6_2_3_3(SH2_struct* sh)
{
  SH2mov(sh,2,3);
}
#define init_6_2_4_3
void FASTCALL call_6_2_4_3(SH2_struct* sh)
{
  SH2mov(sh,2,4);
}
#define init_6_2_5_3
void FASTCALL call_6_2_5_3(SH2_struct* sh)
{
  SH2mov(sh,2,5);
}
#define init_6_2_6_3
void FASTCALL call_6_2_6_3(SH2_struct* sh)
{
  SH2mov(sh,2,6);
}
#define init_6_2_7_3
void FASTCALL call_6_2_7_3(SH2_struct* sh)
{
  SH2mov(sh,2,7);
}
#define init_6_2_8_3
void FASTCALL call_6_2_8_3(SH2_struct* sh)
{
  SH2mov(sh,2,8);
}
#define init_6_2_9_3
void FASTCALL call_6_2_9_3(SH2_struct* sh)
{
  SH2mov(sh,2,9);
}
#define init_6_2_10_3
void FASTCALL call_6_2_10_3(SH2_struct* sh)
{
  SH2mov(sh,2,10);
}
#define init_6_2_11_3
void FASTCALL call_6_2_11_3(SH2_struct* sh)
{
  SH2mov(sh,2,11);
}
#define init_6_2_12_3
void FASTCALL call_6_2_12_3(SH2_struct* sh)
{
  SH2mov(sh,2,12);
}
#define init_6_2_13_3
void FASTCALL call_6_2_13_3(SH2_struct* sh)
{
  SH2mov(sh,2,13);
}
#define init_6_2_14_3
void FASTCALL call_6_2_14_3(SH2_struct* sh)
{
  SH2mov(sh,2,14);
}
#define init_6_2_15_3
void FASTCALL call_6_2_15_3(SH2_struct* sh)
{
  SH2mov(sh,2,15);
}
#define init_6_3_0_3
void FASTCALL call_6_3_0_3(SH2_struct* sh)
{
  SH2mov(sh,3,0);
}
#define init_6_3_1_3
void FASTCALL call_6_3_1_3(SH2_struct* sh)
{
  SH2mov(sh,3,1);
}
#define init_6_3_2_3
void FASTCALL call_6_3_2_3(SH2_struct* sh)
{
  SH2mov(sh,3,2);
}
#define init_6_3_3_3
void FASTCALL call_6_3_3_3(SH2_struct* sh)
{
  SH2mov(sh,3,3);
}
#define init_6_3_4_3
void FASTCALL call_6_3_4_3(SH2_struct* sh)
{
  SH2mov(sh,3,4);
}
#define init_6_3_5_3
void FASTCALL call_6_3_5_3(SH2_struct* sh)
{
  SH2mov(sh,3,5);
}
#define init_6_3_6_3
void FASTCALL call_6_3_6_3(SH2_struct* sh)
{
  SH2mov(sh,3,6);
}
#define init_6_3_7_3
void FASTCALL call_6_3_7_3(SH2_struct* sh)
{
  SH2mov(sh,3,7);
}
#define init_6_3_8_3
void FASTCALL call_6_3_8_3(SH2_struct* sh)
{
  SH2mov(sh,3,8);
}
#define init_6_3_9_3
void FASTCALL call_6_3_9_3(SH2_struct* sh)
{
  SH2mov(sh,3,9);
}
#define init_6_3_10_3
void FASTCALL call_6_3_10_3(SH2_struct* sh)
{
  SH2mov(sh,3,10);
}
#define init_6_3_11_3
void FASTCALL call_6_3_11_3(SH2_struct* sh)
{
  SH2mov(sh,3,11);
}
#define init_6_3_12_3
void FASTCALL call_6_3_12_3(SH2_struct* sh)
{
  SH2mov(sh,3,12);
}
#define init_6_3_13_3
void FASTCALL call_6_3_13_3(SH2_struct* sh)
{
  SH2mov(sh,3,13);
}
#define init_6_3_14_3
void FASTCALL call_6_3_14_3(SH2_struct* sh)
{
  SH2mov(sh,3,14);
}
#define init_6_3_15_3
void FASTCALL call_6_3_15_3(SH2_struct* sh)
{
  SH2mov(sh,3,15);
}
#define init_6_4_0_3
void FASTCALL call_6_4_0_3(SH2_struct* sh)
{
  SH2mov(sh,4,0);
}
#define init_6_4_1_3
void FASTCALL call_6_4_1_3(SH2_struct* sh)
{
  SH2mov(sh,4,1);
}
#define init_6_4_2_3
void FASTCALL call_6_4_2_3(SH2_struct* sh)
{
  SH2mov(sh,4,2);
}
#define init_6_4_3_3
void FASTCALL call_6_4_3_3(SH2_struct* sh)
{
  SH2mov(sh,4,3);
}
#define init_6_4_4_3
void FASTCALL call_6_4_4_3(SH2_struct* sh)
{
  SH2mov(sh,4,4);
}
#define init_6_4_5_3
void FASTCALL call_6_4_5_3(SH2_struct* sh)
{
  SH2mov(sh,4,5);
}
#define init_6_4_6_3
void FASTCALL call_6_4_6_3(SH2_struct* sh)
{
  SH2mov(sh,4,6);
}
#define init_6_4_7_3
void FASTCALL call_6_4_7_3(SH2_struct* sh)
{
  SH2mov(sh,4,7);
}
#define init_6_4_8_3
void FASTCALL call_6_4_8_3(SH2_struct* sh)
{
  SH2mov(sh,4,8);
}
#define init_6_4_9_3
void FASTCALL call_6_4_9_3(SH2_struct* sh)
{
  SH2mov(sh,4,9);
}
#define init_6_4_10_3
void FASTCALL call_6_4_10_3(SH2_struct* sh)
{
  SH2mov(sh,4,10);
}
#define init_6_4_11_3
void FASTCALL call_6_4_11_3(SH2_struct* sh)
{
  SH2mov(sh,4,11);
}
#define init_6_4_12_3
void FASTCALL call_6_4_12_3(SH2_struct* sh)
{
  SH2mov(sh,4,12);
}
#define init_6_4_13_3
void FASTCALL call_6_4_13_3(SH2_struct* sh)
{
  SH2mov(sh,4,13);
}
#define init_6_4_14_3
void FASTCALL call_6_4_14_3(SH2_struct* sh)
{
  SH2mov(sh,4,14);
}
#define init_6_4_15_3
void FASTCALL call_6_4_15_3(SH2_struct* sh)
{
  SH2mov(sh,4,15);
}
#define init_6_5_0_3
void FASTCALL call_6_5_0_3(SH2_struct* sh)
{
  SH2mov(sh,5,0);
}
#define init_6_5_1_3
void FASTCALL call_6_5_1_3(SH2_struct* sh)
{
  SH2mov(sh,5,1);
}
#define init_6_5_2_3
void FASTCALL call_6_5_2_3(SH2_struct* sh)
{
  SH2mov(sh,5,2);
}
#define init_6_5_3_3
void FASTCALL call_6_5_3_3(SH2_struct* sh)
{
  SH2mov(sh,5,3);
}
#define init_6_5_4_3
void FASTCALL call_6_5_4_3(SH2_struct* sh)
{
  SH2mov(sh,5,4);
}
#define init_6_5_5_3
void FASTCALL call_6_5_5_3(SH2_struct* sh)
{
  SH2mov(sh,5,5);
}
#define init_6_5_6_3
void FASTCALL call_6_5_6_3(SH2_struct* sh)
{
  SH2mov(sh,5,6);
}
#define init_6_5_7_3
void FASTCALL call_6_5_7_3(SH2_struct* sh)
{
  SH2mov(sh,5,7);
}
#define init_6_5_8_3
void FASTCALL call_6_5_8_3(SH2_struct* sh)
{
  SH2mov(sh,5,8);
}
#define init_6_5_9_3
void FASTCALL call_6_5_9_3(SH2_struct* sh)
{
  SH2mov(sh,5,9);
}
#define init_6_5_10_3
void FASTCALL call_6_5_10_3(SH2_struct* sh)
{
  SH2mov(sh,5,10);
}
#define init_6_5_11_3
void FASTCALL call_6_5_11_3(SH2_struct* sh)
{
  SH2mov(sh,5,11);
}
#define init_6_5_12_3
void FASTCALL call_6_5_12_3(SH2_struct* sh)
{
  SH2mov(sh,5,12);
}
#define init_6_5_13_3
void FASTCALL call_6_5_13_3(SH2_struct* sh)
{
  SH2mov(sh,5,13);
}
#define init_6_5_14_3
void FASTCALL call_6_5_14_3(SH2_struct* sh)
{
  SH2mov(sh,5,14);
}
#define init_6_5_15_3
void FASTCALL call_6_5_15_3(SH2_struct* sh)
{
  SH2mov(sh,5,15);
}
#define init_6_6_0_3
void FASTCALL call_6_6_0_3(SH2_struct* sh)
{
  SH2mov(sh,6,0);
}
#define init_6_6_1_3
void FASTCALL call_6_6_1_3(SH2_struct* sh)
{
  SH2mov(sh,6,1);
}
#define init_6_6_2_3
void FASTCALL call_6_6_2_3(SH2_struct* sh)
{
  SH2mov(sh,6,2);
}
#define init_6_6_3_3
void FASTCALL call_6_6_3_3(SH2_struct* sh)
{
  SH2mov(sh,6,3);
}
#define init_6_6_4_3
void FASTCALL call_6_6_4_3(SH2_struct* sh)
{
  SH2mov(sh,6,4);
}
#define init_6_6_5_3
void FASTCALL call_6_6_5_3(SH2_struct* sh)
{
  SH2mov(sh,6,5);
}
#define init_6_6_6_3
void FASTCALL call_6_6_6_3(SH2_struct* sh)
{
  SH2mov(sh,6,6);
}
#define init_6_6_7_3
void FASTCALL call_6_6_7_3(SH2_struct* sh)
{
  SH2mov(sh,6,7);
}
#define init_6_6_8_3
void FASTCALL call_6_6_8_3(SH2_struct* sh)
{
  SH2mov(sh,6,8);
}
#define init_6_6_9_3
void FASTCALL call_6_6_9_3(SH2_struct* sh)
{
  SH2mov(sh,6,9);
}
#define init_6_6_10_3
void FASTCALL call_6_6_10_3(SH2_struct* sh)
{
  SH2mov(sh,6,10);
}
#define init_6_6_11_3
void FASTCALL call_6_6_11_3(SH2_struct* sh)
{
  SH2mov(sh,6,11);
}
#define init_6_6_12_3
void FASTCALL call_6_6_12_3(SH2_struct* sh)
{
  SH2mov(sh,6,12);
}
#define init_6_6_13_3
void FASTCALL call_6_6_13_3(SH2_struct* sh)
{
  SH2mov(sh,6,13);
}
#define init_6_6_14_3
void FASTCALL call_6_6_14_3(SH2_struct* sh)
{
  SH2mov(sh,6,14);
}
#define init_6_6_15_3
void FASTCALL call_6_6_15_3(SH2_struct* sh)
{
  SH2mov(sh,6,15);
}
#define init_6_7_0_3
void FASTCALL call_6_7_0_3(SH2_struct* sh)
{
  SH2mov(sh,7,0);
}
#define init_6_7_1_3
void FASTCALL call_6_7_1_3(SH2_struct* sh)
{
  SH2mov(sh,7,1);
}
#define init_6_7_2_3
void FASTCALL call_6_7_2_3(SH2_struct* sh)
{
  SH2mov(sh,7,2);
}
#define init_6_7_3_3
void FASTCALL call_6_7_3_3(SH2_struct* sh)
{
  SH2mov(sh,7,3);
}
#define init_6_7_4_3
void FASTCALL call_6_7_4_3(SH2_struct* sh)
{
  SH2mov(sh,7,4);
}
#define init_6_7_5_3
void FASTCALL call_6_7_5_3(SH2_struct* sh)
{
  SH2mov(sh,7,5);
}
#define init_6_7_6_3
void FASTCALL call_6_7_6_3(SH2_struct* sh)
{
  SH2mov(sh,7,6);
}
#define init_6_7_7_3
void FASTCALL call_6_7_7_3(SH2_struct* sh)
{
  SH2mov(sh,7,7);
}
#define init_6_7_8_3
void FASTCALL call_6_7_8_3(SH2_struct* sh)
{
  SH2mov(sh,7,8);
}
#define init_6_7_9_3
void FASTCALL call_6_7_9_3(SH2_struct* sh)
{
  SH2mov(sh,7,9);
}
#define init_6_7_10_3
void FASTCALL call_6_7_10_3(SH2_struct* sh)
{
  SH2mov(sh,7,10);
}
#define init_6_7_11_3
void FASTCALL call_6_7_11_3(SH2_struct* sh)
{
  SH2mov(sh,7,11);
}
#define init_6_7_12_3
void FASTCALL call_6_7_12_3(SH2_struct* sh)
{
  SH2mov(sh,7,12);
}
#define init_6_7_13_3
void FASTCALL call_6_7_13_3(SH2_struct* sh)
{
  SH2mov(sh,7,13);
}
#define init_6_7_14_3
void FASTCALL call_6_7_14_3(SH2_struct* sh)
{
  SH2mov(sh,7,14);
}
#define init_6_7_15_3
void FASTCALL call_6_7_15_3(SH2_struct* sh)
{
  SH2mov(sh,7,15);
}
#define init_6_8_0_3
void FASTCALL call_6_8_0_3(SH2_struct* sh)
{
  SH2mov(sh,8,0);
}
#define init_6_8_1_3
void FASTCALL call_6_8_1_3(SH2_struct* sh)
{
  SH2mov(sh,8,1);
}
#define init_6_8_2_3
void FASTCALL call_6_8_2_3(SH2_struct* sh)
{
  SH2mov(sh,8,2);
}
#define init_6_8_3_3
void FASTCALL call_6_8_3_3(SH2_struct* sh)
{
  SH2mov(sh,8,3);
}
#define init_6_8_4_3
void FASTCALL call_6_8_4_3(SH2_struct* sh)
{
  SH2mov(sh,8,4);
}
#define init_6_8_5_3
void FASTCALL call_6_8_5_3(SH2_struct* sh)
{
  SH2mov(sh,8,5);
}
#define init_6_8_6_3
void FASTCALL call_6_8_6_3(SH2_struct* sh)
{
  SH2mov(sh,8,6);
}
#define init_6_8_7_3
void FASTCALL call_6_8_7_3(SH2_struct* sh)
{
  SH2mov(sh,8,7);
}
#define init_6_8_8_3
void FASTCALL call_6_8_8_3(SH2_struct* sh)
{
  SH2mov(sh,8,8);
}
#define init_6_8_9_3
void FASTCALL call_6_8_9_3(SH2_struct* sh)
{
  SH2mov(sh,8,9);
}
#define init_6_8_10_3
void FASTCALL call_6_8_10_3(SH2_struct* sh)
{
  SH2mov(sh,8,10);
}
#define init_6_8_11_3
void FASTCALL call_6_8_11_3(SH2_struct* sh)
{
  SH2mov(sh,8,11);
}
#define init_6_8_12_3
void FASTCALL call_6_8_12_3(SH2_struct* sh)
{
  SH2mov(sh,8,12);
}
#define init_6_8_13_3
void FASTCALL call_6_8_13_3(SH2_struct* sh)
{
  SH2mov(sh,8,13);
}
#define init_6_8_14_3
void FASTCALL call_6_8_14_3(SH2_struct* sh)
{
  SH2mov(sh,8,14);
}
#define init_6_8_15_3
void FASTCALL call_6_8_15_3(SH2_struct* sh)
{
  SH2mov(sh,8,15);
}
#define init_6_9_0_3
void FASTCALL call_6_9_0_3(SH2_struct* sh)
{
  SH2mov(sh,9,0);
}
#define init_6_9_1_3
void FASTCALL call_6_9_1_3(SH2_struct* sh)
{
  SH2mov(sh,9,1);
}
#define init_6_9_2_3
void FASTCALL call_6_9_2_3(SH2_struct* sh)
{
  SH2mov(sh,9,2);
}
#define init_6_9_3_3
void FASTCALL call_6_9_3_3(SH2_struct* sh)
{
  SH2mov(sh,9,3);
}
#define init_6_9_4_3
void FASTCALL call_6_9_4_3(SH2_struct* sh)
{
  SH2mov(sh,9,4);
}
#define init_6_9_5_3
void FASTCALL call_6_9_5_3(SH2_struct* sh)
{
  SH2mov(sh,9,5);
}
#define init_6_9_6_3
void FASTCALL call_6_9_6_3(SH2_struct* sh)
{
  SH2mov(sh,9,6);
}
#define init_6_9_7_3
void FASTCALL call_6_9_7_3(SH2_struct* sh)
{
  SH2mov(sh,9,7);
}
#define init_6_9_8_3
void FASTCALL call_6_9_8_3(SH2_struct* sh)
{
  SH2mov(sh,9,8);
}
#define init_6_9_9_3
void FASTCALL call_6_9_9_3(SH2_struct* sh)
{
  SH2mov(sh,9,9);
}
#define init_6_9_10_3
void FASTCALL call_6_9_10_3(SH2_struct* sh)
{
  SH2mov(sh,9,10);
}
#define init_6_9_11_3
void FASTCALL call_6_9_11_3(SH2_struct* sh)
{
  SH2mov(sh,9,11);
}
#define init_6_9_12_3
void FASTCALL call_6_9_12_3(SH2_struct* sh)
{
  SH2mov(sh,9,12);
}
#define init_6_9_13_3
void FASTCALL call_6_9_13_3(SH2_struct* sh)
{
  SH2mov(sh,9,13);
}
#define init_6_9_14_3
void FASTCALL call_6_9_14_3(SH2_struct* sh)
{
  SH2mov(sh,9,14);
}
#define init_6_9_15_3
void FASTCALL call_6_9_15_3(SH2_struct* sh)
{
  SH2mov(sh,9,15);
}
#define init_6_10_0_3
void FASTCALL call_6_10_0_3(SH2_struct* sh)
{
  SH2mov(sh,10,0);
}
#define init_6_10_1_3
void FASTCALL call_6_10_1_3(SH2_struct* sh)
{
  SH2mov(sh,10,1);
}
#define init_6_10_2_3
void FASTCALL call_6_10_2_3(SH2_struct* sh)
{
  SH2mov(sh,10,2);
}
#define init_6_10_3_3
void FASTCALL call_6_10_3_3(SH2_struct* sh)
{
  SH2mov(sh,10,3);
}
#define init_6_10_4_3
void FASTCALL call_6_10_4_3(SH2_struct* sh)
{
  SH2mov(sh,10,4);
}
#define init_6_10_5_3
void FASTCALL call_6_10_5_3(SH2_struct* sh)
{
  SH2mov(sh,10,5);
}
#define init_6_10_6_3
void FASTCALL call_6_10_6_3(SH2_struct* sh)
{
  SH2mov(sh,10,6);
}
#define init_6_10_7_3
void FASTCALL call_6_10_7_3(SH2_struct* sh)
{
  SH2mov(sh,10,7);
}
#define init_6_10_8_3
void FASTCALL call_6_10_8_3(SH2_struct* sh)
{
  SH2mov(sh,10,8);
}
#define init_6_10_9_3
void FASTCALL call_6_10_9_3(SH2_struct* sh)
{
  SH2mov(sh,10,9);
}
#define init_6_10_10_3
void FASTCALL call_6_10_10_3(SH2_struct* sh)
{
  SH2mov(sh,10,10);
}
#define init_6_10_11_3
void FASTCALL call_6_10_11_3(SH2_struct* sh)
{
  SH2mov(sh,10,11);
}
#define init_6_10_12_3
void FASTCALL call_6_10_12_3(SH2_struct* sh)
{
  SH2mov(sh,10,12);
}
#define init_6_10_13_3
void FASTCALL call_6_10_13_3(SH2_struct* sh)
{
  SH2mov(sh,10,13);
}
#define init_6_10_14_3
void FASTCALL call_6_10_14_3(SH2_struct* sh)
{
  SH2mov(sh,10,14);
}
#define init_6_10_15_3
void FASTCALL call_6_10_15_3(SH2_struct* sh)
{
  SH2mov(sh,10,15);
}
#define init_6_11_0_3
void FASTCALL call_6_11_0_3(SH2_struct* sh)
{
  SH2mov(sh,11,0);
}
#define init_6_11_1_3
void FASTCALL call_6_11_1_3(SH2_struct* sh)
{
  SH2mov(sh,11,1);
}
#define init_6_11_2_3
void FASTCALL call_6_11_2_3(SH2_struct* sh)
{
  SH2mov(sh,11,2);
}
#define init_6_11_3_3
void FASTCALL call_6_11_3_3(SH2_struct* sh)
{
  SH2mov(sh,11,3);
}
#define init_6_11_4_3
void FASTCALL call_6_11_4_3(SH2_struct* sh)
{
  SH2mov(sh,11,4);
}
#define init_6_11_5_3
void FASTCALL call_6_11_5_3(SH2_struct* sh)
{
  SH2mov(sh,11,5);
}
#define init_6_11_6_3
void FASTCALL call_6_11_6_3(SH2_struct* sh)
{
  SH2mov(sh,11,6);
}
#define init_6_11_7_3
void FASTCALL call_6_11_7_3(SH2_struct* sh)
{
  SH2mov(sh,11,7);
}
#define init_6_11_8_3
void FASTCALL call_6_11_8_3(SH2_struct* sh)
{
  SH2mov(sh,11,8);
}
#define init_6_11_9_3
void FASTCALL call_6_11_9_3(SH2_struct* sh)
{
  SH2mov(sh,11,9);
}
#define init_6_11_10_3
void FASTCALL call_6_11_10_3(SH2_struct* sh)
{
  SH2mov(sh,11,10);
}
#define init_6_11_11_3
void FASTCALL call_6_11_11_3(SH2_struct* sh)
{
  SH2mov(sh,11,11);
}
#define init_6_11_12_3
void FASTCALL call_6_11_12_3(SH2_struct* sh)
{
  SH2mov(sh,11,12);
}
#define init_6_11_13_3
void FASTCALL call_6_11_13_3(SH2_struct* sh)
{
  SH2mov(sh,11,13);
}
#define init_6_11_14_3
void FASTCALL call_6_11_14_3(SH2_struct* sh)
{
  SH2mov(sh,11,14);
}
#define init_6_11_15_3
void FASTCALL call_6_11_15_3(SH2_struct* sh)
{
  SH2mov(sh,11,15);
}
#define init_6_12_0_3
void FASTCALL call_6_12_0_3(SH2_struct* sh)
{
  SH2mov(sh,12,0);
}
#define init_6_12_1_3
void FASTCALL call_6_12_1_3(SH2_struct* sh)
{
  SH2mov(sh,12,1);
}
#define init_6_12_2_3
void FASTCALL call_6_12_2_3(SH2_struct* sh)
{
  SH2mov(sh,12,2);
}
#define init_6_12_3_3
void FASTCALL call_6_12_3_3(SH2_struct* sh)
{
  SH2mov(sh,12,3);
}
#define init_6_12_4_3
void FASTCALL call_6_12_4_3(SH2_struct* sh)
{
  SH2mov(sh,12,4);
}
#define init_6_12_5_3
void FASTCALL call_6_12_5_3(SH2_struct* sh)
{
  SH2mov(sh,12,5);
}
#define init_6_12_6_3
void FASTCALL call_6_12_6_3(SH2_struct* sh)
{
  SH2mov(sh,12,6);
}
#define init_6_12_7_3
void FASTCALL call_6_12_7_3(SH2_struct* sh)
{
  SH2mov(sh,12,7);
}
#define init_6_12_8_3
void FASTCALL call_6_12_8_3(SH2_struct* sh)
{
  SH2mov(sh,12,8);
}
#define init_6_12_9_3
void FASTCALL call_6_12_9_3(SH2_struct* sh)
{
  SH2mov(sh,12,9);
}
#define init_6_12_10_3
void FASTCALL call_6_12_10_3(SH2_struct* sh)
{
  SH2mov(sh,12,10);
}
#define init_6_12_11_3
void FASTCALL call_6_12_11_3(SH2_struct* sh)
{
  SH2mov(sh,12,11);
}
#define init_6_12_12_3
void FASTCALL call_6_12_12_3(SH2_struct* sh)
{
  SH2mov(sh,12,12);
}
#define init_6_12_13_3
void FASTCALL call_6_12_13_3(SH2_struct* sh)
{
  SH2mov(sh,12,13);
}
#define init_6_12_14_3
void FASTCALL call_6_12_14_3(SH2_struct* sh)
{
  SH2mov(sh,12,14);
}
#define init_6_12_15_3
void FASTCALL call_6_12_15_3(SH2_struct* sh)
{
  SH2mov(sh,12,15);
}
#define init_6_13_0_3
void FASTCALL call_6_13_0_3(SH2_struct* sh)
{
  SH2mov(sh,13,0);
}
#define init_6_13_1_3
void FASTCALL call_6_13_1_3(SH2_struct* sh)
{
  SH2mov(sh,13,1);
}
#define init_6_13_2_3
void FASTCALL call_6_13_2_3(SH2_struct* sh)
{
  SH2mov(sh,13,2);
}
#define init_6_13_3_3
void FASTCALL call_6_13_3_3(SH2_struct* sh)
{
  SH2mov(sh,13,3);
}
#define init_6_13_4_3
void FASTCALL call_6_13_4_3(SH2_struct* sh)
{
  SH2mov(sh,13,4);
}
#define init_6_13_5_3
void FASTCALL call_6_13_5_3(SH2_struct* sh)
{
  SH2mov(sh,13,5);
}
#define init_6_13_6_3
void FASTCALL call_6_13_6_3(SH2_struct* sh)
{
  SH2mov(sh,13,6);
}
#define init_6_13_7_3
void FASTCALL call_6_13_7_3(SH2_struct* sh)
{
  SH2mov(sh,13,7);
}
#define init_6_13_8_3
void FASTCALL call_6_13_8_3(SH2_struct* sh)
{
  SH2mov(sh,13,8);
}
#define init_6_13_9_3
void FASTCALL call_6_13_9_3(SH2_struct* sh)
{
  SH2mov(sh,13,9);
}
#define init_6_13_10_3
void FASTCALL call_6_13_10_3(SH2_struct* sh)
{
  SH2mov(sh,13,10);
}
#define init_6_13_11_3
void FASTCALL call_6_13_11_3(SH2_struct* sh)
{
  SH2mov(sh,13,11);
}
#define init_6_13_12_3
void FASTCALL call_6_13_12_3(SH2_struct* sh)
{
  SH2mov(sh,13,12);
}
#define init_6_13_13_3
void FASTCALL call_6_13_13_3(SH2_struct* sh)
{
  SH2mov(sh,13,13);
}
#define init_6_13_14_3
void FASTCALL call_6_13_14_3(SH2_struct* sh)
{
  SH2mov(sh,13,14);
}
#define init_6_13_15_3
void FASTCALL call_6_13_15_3(SH2_struct* sh)
{
  SH2mov(sh,13,15);
}
#define init_6_14_0_3
void FASTCALL call_6_14_0_3(SH2_struct* sh)
{
  SH2mov(sh,14,0);
}
#define init_6_14_1_3
void FASTCALL call_6_14_1_3(SH2_struct* sh)
{
  SH2mov(sh,14,1);
}
#define init_6_14_2_3
void FASTCALL call_6_14_2_3(SH2_struct* sh)
{
  SH2mov(sh,14,2);
}
#define init_6_14_3_3
void FASTCALL call_6_14_3_3(SH2_struct* sh)
{
  SH2mov(sh,14,3);
}
#define init_6_14_4_3
void FASTCALL call_6_14_4_3(SH2_struct* sh)
{
  SH2mov(sh,14,4);
}
#define init_6_14_5_3
void FASTCALL call_6_14_5_3(SH2_struct* sh)
{
  SH2mov(sh,14,5);
}
#define init_6_14_6_3
void FASTCALL call_6_14_6_3(SH2_struct* sh)
{
  SH2mov(sh,14,6);
}
#define init_6_14_7_3
void FASTCALL call_6_14_7_3(SH2_struct* sh)
{
  SH2mov(sh,14,7);
}
#define init_6_14_8_3
void FASTCALL call_6_14_8_3(SH2_struct* sh)
{
  SH2mov(sh,14,8);
}
#define init_6_14_9_3
void FASTCALL call_6_14_9_3(SH2_struct* sh)
{
  SH2mov(sh,14,9);
}
#define init_6_14_10_3
void FASTCALL call_6_14_10_3(SH2_struct* sh)
{
  SH2mov(sh,14,10);
}
#define init_6_14_11_3
void FASTCALL call_6_14_11_3(SH2_struct* sh)
{
  SH2mov(sh,14,11);
}
#define init_6_14_12_3
void FASTCALL call_6_14_12_3(SH2_struct* sh)
{
  SH2mov(sh,14,12);
}
#define init_6_14_13_3
void FASTCALL call_6_14_13_3(SH2_struct* sh)
{
  SH2mov(sh,14,13);
}
#define init_6_14_14_3
void FASTCALL call_6_14_14_3(SH2_struct* sh)
{
  SH2mov(sh,14,14);
}
#define init_6_14_15_3
void FASTCALL call_6_14_15_3(SH2_struct* sh)
{
  SH2mov(sh,14,15);
}
#define init_6_15_0_3
void FASTCALL call_6_15_0_3(SH2_struct* sh)
{
  SH2mov(sh,15,0);
}
#define init_6_15_1_3
void FASTCALL call_6_15_1_3(SH2_struct* sh)
{
  SH2mov(sh,15,1);
}
#define init_6_15_2_3
void FASTCALL call_6_15_2_3(SH2_struct* sh)
{
  SH2mov(sh,15,2);
}
#define init_6_15_3_3
void FASTCALL call_6_15_3_3(SH2_struct* sh)
{
  SH2mov(sh,15,3);
}
#define init_6_15_4_3
void FASTCALL call_6_15_4_3(SH2_struct* sh)
{
  SH2mov(sh,15,4);
}
#define init_6_15_5_3
void FASTCALL call_6_15_5_3(SH2_struct* sh)
{
  SH2mov(sh,15,5);
}
#define init_6_15_6_3
void FASTCALL call_6_15_6_3(SH2_struct* sh)
{
  SH2mov(sh,15,6);
}
#define init_6_15_7_3
void FASTCALL call_6_15_7_3(SH2_struct* sh)
{
  SH2mov(sh,15,7);
}
#define init_6_15_8_3
void FASTCALL call_6_15_8_3(SH2_struct* sh)
{
  SH2mov(sh,15,8);
}
#define init_6_15_9_3
void FASTCALL call_6_15_9_3(SH2_struct* sh)
{
  SH2mov(sh,15,9);
}
#define init_6_15_10_3
void FASTCALL call_6_15_10_3(SH2_struct* sh)
{
  SH2mov(sh,15,10);
}
#define init_6_15_11_3
void FASTCALL call_6_15_11_3(SH2_struct* sh)
{
  SH2mov(sh,15,11);
}
#define init_6_15_12_3
void FASTCALL call_6_15_12_3(SH2_struct* sh)
{
  SH2mov(sh,15,12);
}
#define init_6_15_13_3
void FASTCALL call_6_15_13_3(SH2_struct* sh)
{
  SH2mov(sh,15,13);
}
#define init_6_15_14_3
void FASTCALL call_6_15_14_3(SH2_struct* sh)
{
  SH2mov(sh,15,14);
}
#define init_6_15_15_3
void FASTCALL call_6_15_15_3(SH2_struct* sh)
{
  SH2mov(sh,15,15);
}
#define init_4_0_0_15
void FASTCALL call_4_0_0_15(SH2_struct* sh)
{
  SH2macw(sh,0,0);
}
#define init_4_0_1_15
void FASTCALL call_4_0_1_15(SH2_struct* sh)
{
  SH2macw(sh,0,1);
}
#define init_4_0_2_15
void FASTCALL call_4_0_2_15(SH2_struct* sh)
{
  SH2macw(sh,0,2);
}
#define init_4_0_3_15
void FASTCALL call_4_0_3_15(SH2_struct* sh)
{
  SH2macw(sh,0,3);
}
#define init_4_0_4_15
void FASTCALL call_4_0_4_15(SH2_struct* sh)
{
  SH2macw(sh,0,4);
}
#define init_4_0_5_15
void FASTCALL call_4_0_5_15(SH2_struct* sh)
{
  SH2macw(sh,0,5);
}
#define init_4_0_6_15
void FASTCALL call_4_0_6_15(SH2_struct* sh)
{
  SH2macw(sh,0,6);
}
#define init_4_0_7_15
void FASTCALL call_4_0_7_15(SH2_struct* sh)
{
  SH2macw(sh,0,7);
}
#define init_4_0_8_15
void FASTCALL call_4_0_8_15(SH2_struct* sh)
{
  SH2macw(sh,0,8);
}
#define init_4_0_9_15
void FASTCALL call_4_0_9_15(SH2_struct* sh)
{
  SH2macw(sh,0,9);
}
#define init_4_0_10_15
void FASTCALL call_4_0_10_15(SH2_struct* sh)
{
  SH2macw(sh,0,10);
}
#define init_4_0_11_15
void FASTCALL call_4_0_11_15(SH2_struct* sh)
{
  SH2macw(sh,0,11);
}
#define init_4_0_12_15
void FASTCALL call_4_0_12_15(SH2_struct* sh)
{
  SH2macw(sh,0,12);
}
#define init_4_0_13_15
void FASTCALL call_4_0_13_15(SH2_struct* sh)
{
  SH2macw(sh,0,13);
}
#define init_4_0_14_15
void FASTCALL call_4_0_14_15(SH2_struct* sh)
{
  SH2macw(sh,0,14);
}
#define init_4_0_15_15
void FASTCALL call_4_0_15_15(SH2_struct* sh)
{
  SH2macw(sh,0,15);
}
#define init_4_1_0_15
void FASTCALL call_4_1_0_15(SH2_struct* sh)
{
  SH2macw(sh,1,0);
}
#define init_4_1_1_15
void FASTCALL call_4_1_1_15(SH2_struct* sh)
{
  SH2macw(sh,1,1);
}
#define init_4_1_2_15
void FASTCALL call_4_1_2_15(SH2_struct* sh)
{
  SH2macw(sh,1,2);
}
#define init_4_1_3_15
void FASTCALL call_4_1_3_15(SH2_struct* sh)
{
  SH2macw(sh,1,3);
}
#define init_4_1_4_15
void FASTCALL call_4_1_4_15(SH2_struct* sh)
{
  SH2macw(sh,1,4);
}
#define init_4_1_5_15
void FASTCALL call_4_1_5_15(SH2_struct* sh)
{
  SH2macw(sh,1,5);
}
#define init_4_1_6_15
void FASTCALL call_4_1_6_15(SH2_struct* sh)
{
  SH2macw(sh,1,6);
}
#define init_4_1_7_15
void FASTCALL call_4_1_7_15(SH2_struct* sh)
{
  SH2macw(sh,1,7);
}
#define init_4_1_8_15
void FASTCALL call_4_1_8_15(SH2_struct* sh)
{
  SH2macw(sh,1,8);
}
#define init_4_1_9_15
void FASTCALL call_4_1_9_15(SH2_struct* sh)
{
  SH2macw(sh,1,9);
}
#define init_4_1_10_15
void FASTCALL call_4_1_10_15(SH2_struct* sh)
{
  SH2macw(sh,1,10);
}
#define init_4_1_11_15
void FASTCALL call_4_1_11_15(SH2_struct* sh)
{
  SH2macw(sh,1,11);
}
#define init_4_1_12_15
void FASTCALL call_4_1_12_15(SH2_struct* sh)
{
  SH2macw(sh,1,12);
}
#define init_4_1_13_15
void FASTCALL call_4_1_13_15(SH2_struct* sh)
{
  SH2macw(sh,1,13);
}
#define init_4_1_14_15
void FASTCALL call_4_1_14_15(SH2_struct* sh)
{
  SH2macw(sh,1,14);
}
#define init_4_1_15_15
void FASTCALL call_4_1_15_15(SH2_struct* sh)
{
  SH2macw(sh,1,15);
}
#define init_4_2_0_15
void FASTCALL call_4_2_0_15(SH2_struct* sh)
{
  SH2macw(sh,2,0);
}
#define init_4_2_1_15
void FASTCALL call_4_2_1_15(SH2_struct* sh)
{
  SH2macw(sh,2,1);
}
#define init_4_2_2_15
void FASTCALL call_4_2_2_15(SH2_struct* sh)
{
  SH2macw(sh,2,2);
}
#define init_4_2_3_15
void FASTCALL call_4_2_3_15(SH2_struct* sh)
{
  SH2macw(sh,2,3);
}
#define init_4_2_4_15
void FASTCALL call_4_2_4_15(SH2_struct* sh)
{
  SH2macw(sh,2,4);
}
#define init_4_2_5_15
void FASTCALL call_4_2_5_15(SH2_struct* sh)
{
  SH2macw(sh,2,5);
}
#define init_4_2_6_15
void FASTCALL call_4_2_6_15(SH2_struct* sh)
{
  SH2macw(sh,2,6);
}
#define init_4_2_7_15
void FASTCALL call_4_2_7_15(SH2_struct* sh)
{
  SH2macw(sh,2,7);
}
#define init_4_2_8_15
void FASTCALL call_4_2_8_15(SH2_struct* sh)
{
  SH2macw(sh,2,8);
}
#define init_4_2_9_15
void FASTCALL call_4_2_9_15(SH2_struct* sh)
{
  SH2macw(sh,2,9);
}
#define init_4_2_10_15
void FASTCALL call_4_2_10_15(SH2_struct* sh)
{
  SH2macw(sh,2,10);
}
#define init_4_2_11_15
void FASTCALL call_4_2_11_15(SH2_struct* sh)
{
  SH2macw(sh,2,11);
}
#define init_4_2_12_15
void FASTCALL call_4_2_12_15(SH2_struct* sh)
{
  SH2macw(sh,2,12);
}
#define init_4_2_13_15
void FASTCALL call_4_2_13_15(SH2_struct* sh)
{
  SH2macw(sh,2,13);
}
#define init_4_2_14_15
void FASTCALL call_4_2_14_15(SH2_struct* sh)
{
  SH2macw(sh,2,14);
}
#define init_4_2_15_15
void FASTCALL call_4_2_15_15(SH2_struct* sh)
{
  SH2macw(sh,2,15);
}
#define init_4_3_0_15
void FASTCALL call_4_3_0_15(SH2_struct* sh)
{
  SH2macw(sh,3,0);
}
#define init_4_3_1_15
void FASTCALL call_4_3_1_15(SH2_struct* sh)
{
  SH2macw(sh,3,1);
}
#define init_4_3_2_15
void FASTCALL call_4_3_2_15(SH2_struct* sh)
{
  SH2macw(sh,3,2);
}
#define init_4_3_3_15
void FASTCALL call_4_3_3_15(SH2_struct* sh)
{
  SH2macw(sh,3,3);
}
#define init_4_3_4_15
void FASTCALL call_4_3_4_15(SH2_struct* sh)
{
  SH2macw(sh,3,4);
}
#define init_4_3_5_15
void FASTCALL call_4_3_5_15(SH2_struct* sh)
{
  SH2macw(sh,3,5);
}
#define init_4_3_6_15
void FASTCALL call_4_3_6_15(SH2_struct* sh)
{
  SH2macw(sh,3,6);
}
#define init_4_3_7_15
void FASTCALL call_4_3_7_15(SH2_struct* sh)
{
  SH2macw(sh,3,7);
}
#define init_4_3_8_15
void FASTCALL call_4_3_8_15(SH2_struct* sh)
{
  SH2macw(sh,3,8);
}
#define init_4_3_9_15
void FASTCALL call_4_3_9_15(SH2_struct* sh)
{
  SH2macw(sh,3,9);
}
#define init_4_3_10_15
void FASTCALL call_4_3_10_15(SH2_struct* sh)
{
  SH2macw(sh,3,10);
}
#define init_4_3_11_15
void FASTCALL call_4_3_11_15(SH2_struct* sh)
{
  SH2macw(sh,3,11);
}
#define init_4_3_12_15
void FASTCALL call_4_3_12_15(SH2_struct* sh)
{
  SH2macw(sh,3,12);
}
#define init_4_3_13_15
void FASTCALL call_4_3_13_15(SH2_struct* sh)
{
  SH2macw(sh,3,13);
}
#define init_4_3_14_15
void FASTCALL call_4_3_14_15(SH2_struct* sh)
{
  SH2macw(sh,3,14);
}
#define init_4_3_15_15
void FASTCALL call_4_3_15_15(SH2_struct* sh)
{
  SH2macw(sh,3,15);
}
#define init_4_4_0_15
void FASTCALL call_4_4_0_15(SH2_struct* sh)
{
  SH2macw(sh,4,0);
}
#define init_4_4_1_15
void FASTCALL call_4_4_1_15(SH2_struct* sh)
{
  SH2macw(sh,4,1);
}
#define init_4_4_2_15
void FASTCALL call_4_4_2_15(SH2_struct* sh)
{
  SH2macw(sh,4,2);
}
#define init_4_4_3_15
void FASTCALL call_4_4_3_15(SH2_struct* sh)
{
  SH2macw(sh,4,3);
}
#define init_4_4_4_15
void FASTCALL call_4_4_4_15(SH2_struct* sh)
{
  SH2macw(sh,4,4);
}
#define init_4_4_5_15
void FASTCALL call_4_4_5_15(SH2_struct* sh)
{
  SH2macw(sh,4,5);
}
#define init_4_4_6_15
void FASTCALL call_4_4_6_15(SH2_struct* sh)
{
  SH2macw(sh,4,6);
}
#define init_4_4_7_15
void FASTCALL call_4_4_7_15(SH2_struct* sh)
{
  SH2macw(sh,4,7);
}
#define init_4_4_8_15
void FASTCALL call_4_4_8_15(SH2_struct* sh)
{
  SH2macw(sh,4,8);
}
#define init_4_4_9_15
void FASTCALL call_4_4_9_15(SH2_struct* sh)
{
  SH2macw(sh,4,9);
}
#define init_4_4_10_15
void FASTCALL call_4_4_10_15(SH2_struct* sh)
{
  SH2macw(sh,4,10);
}
#define init_4_4_11_15
void FASTCALL call_4_4_11_15(SH2_struct* sh)
{
  SH2macw(sh,4,11);
}
#define init_4_4_12_15
void FASTCALL call_4_4_12_15(SH2_struct* sh)
{
  SH2macw(sh,4,12);
}
#define init_4_4_13_15
void FASTCALL call_4_4_13_15(SH2_struct* sh)
{
  SH2macw(sh,4,13);
}
#define init_4_4_14_15
void FASTCALL call_4_4_14_15(SH2_struct* sh)
{
  SH2macw(sh,4,14);
}
#define init_4_4_15_15
void FASTCALL call_4_4_15_15(SH2_struct* sh)
{
  SH2macw(sh,4,15);
}
#define init_4_5_0_15
void FASTCALL call_4_5_0_15(SH2_struct* sh)
{
  SH2macw(sh,5,0);
}
#define init_4_5_1_15
void FASTCALL call_4_5_1_15(SH2_struct* sh)
{
  SH2macw(sh,5,1);
}
#define init_4_5_2_15
void FASTCALL call_4_5_2_15(SH2_struct* sh)
{
  SH2macw(sh,5,2);
}
#define init_4_5_3_15
void FASTCALL call_4_5_3_15(SH2_struct* sh)
{
  SH2macw(sh,5,3);
}
#define init_4_5_4_15
void FASTCALL call_4_5_4_15(SH2_struct* sh)
{
  SH2macw(sh,5,4);
}
#define init_4_5_5_15
void FASTCALL call_4_5_5_15(SH2_struct* sh)
{
  SH2macw(sh,5,5);
}
#define init_4_5_6_15
void FASTCALL call_4_5_6_15(SH2_struct* sh)
{
  SH2macw(sh,5,6);
}
#define init_4_5_7_15
void FASTCALL call_4_5_7_15(SH2_struct* sh)
{
  SH2macw(sh,5,7);
}
#define init_4_5_8_15
void FASTCALL call_4_5_8_15(SH2_struct* sh)
{
  SH2macw(sh,5,8);
}
#define init_4_5_9_15
void FASTCALL call_4_5_9_15(SH2_struct* sh)
{
  SH2macw(sh,5,9);
}
#define init_4_5_10_15
void FASTCALL call_4_5_10_15(SH2_struct* sh)
{
  SH2macw(sh,5,10);
}
#define init_4_5_11_15
void FASTCALL call_4_5_11_15(SH2_struct* sh)
{
  SH2macw(sh,5,11);
}
#define init_4_5_12_15
void FASTCALL call_4_5_12_15(SH2_struct* sh)
{
  SH2macw(sh,5,12);
}
#define init_4_5_13_15
void FASTCALL call_4_5_13_15(SH2_struct* sh)
{
  SH2macw(sh,5,13);
}
#define init_4_5_14_15
void FASTCALL call_4_5_14_15(SH2_struct* sh)
{
  SH2macw(sh,5,14);
}
#define init_4_5_15_15
void FASTCALL call_4_5_15_15(SH2_struct* sh)
{
  SH2macw(sh,5,15);
}
#define init_4_6_0_15
void FASTCALL call_4_6_0_15(SH2_struct* sh)
{
  SH2macw(sh,6,0);
}
#define init_4_6_1_15
void FASTCALL call_4_6_1_15(SH2_struct* sh)
{
  SH2macw(sh,6,1);
}
#define init_4_6_2_15
void FASTCALL call_4_6_2_15(SH2_struct* sh)
{
  SH2macw(sh,6,2);
}
#define init_4_6_3_15
void FASTCALL call_4_6_3_15(SH2_struct* sh)
{
  SH2macw(sh,6,3);
}
#define init_4_6_4_15
void FASTCALL call_4_6_4_15(SH2_struct* sh)
{
  SH2macw(sh,6,4);
}
#define init_4_6_5_15
void FASTCALL call_4_6_5_15(SH2_struct* sh)
{
  SH2macw(sh,6,5);
}
#define init_4_6_6_15
void FASTCALL call_4_6_6_15(SH2_struct* sh)
{
  SH2macw(sh,6,6);
}
#define init_4_6_7_15
void FASTCALL call_4_6_7_15(SH2_struct* sh)
{
  SH2macw(sh,6,7);
}
#define init_4_6_8_15
void FASTCALL call_4_6_8_15(SH2_struct* sh)
{
  SH2macw(sh,6,8);
}
#define init_4_6_9_15
void FASTCALL call_4_6_9_15(SH2_struct* sh)
{
  SH2macw(sh,6,9);
}
#define init_4_6_10_15
void FASTCALL call_4_6_10_15(SH2_struct* sh)
{
  SH2macw(sh,6,10);
}
#define init_4_6_11_15
void FASTCALL call_4_6_11_15(SH2_struct* sh)
{
  SH2macw(sh,6,11);
}
#define init_4_6_12_15
void FASTCALL call_4_6_12_15(SH2_struct* sh)
{
  SH2macw(sh,6,12);
}
#define init_4_6_13_15
void FASTCALL call_4_6_13_15(SH2_struct* sh)
{
  SH2macw(sh,6,13);
}
#define init_4_6_14_15
void FASTCALL call_4_6_14_15(SH2_struct* sh)
{
  SH2macw(sh,6,14);
}
#define init_4_6_15_15
void FASTCALL call_4_6_15_15(SH2_struct* sh)
{
  SH2macw(sh,6,15);
}
#define init_4_7_0_15
void FASTCALL call_4_7_0_15(SH2_struct* sh)
{
  SH2macw(sh,7,0);
}
#define init_4_7_1_15
void FASTCALL call_4_7_1_15(SH2_struct* sh)
{
  SH2macw(sh,7,1);
}
#define init_4_7_2_15
void FASTCALL call_4_7_2_15(SH2_struct* sh)
{
  SH2macw(sh,7,2);
}
#define init_4_7_3_15
void FASTCALL call_4_7_3_15(SH2_struct* sh)
{
  SH2macw(sh,7,3);
}
#define init_4_7_4_15
void FASTCALL call_4_7_4_15(SH2_struct* sh)
{
  SH2macw(sh,7,4);
}
#define init_4_7_5_15
void FASTCALL call_4_7_5_15(SH2_struct* sh)
{
  SH2macw(sh,7,5);
}
#define init_4_7_6_15
void FASTCALL call_4_7_6_15(SH2_struct* sh)
{
  SH2macw(sh,7,6);
}
#define init_4_7_7_15
void FASTCALL call_4_7_7_15(SH2_struct* sh)
{
  SH2macw(sh,7,7);
}
#define init_4_7_8_15
void FASTCALL call_4_7_8_15(SH2_struct* sh)
{
  SH2macw(sh,7,8);
}
#define init_4_7_9_15
void FASTCALL call_4_7_9_15(SH2_struct* sh)
{
  SH2macw(sh,7,9);
}
#define init_4_7_10_15
void FASTCALL call_4_7_10_15(SH2_struct* sh)
{
  SH2macw(sh,7,10);
}
#define init_4_7_11_15
void FASTCALL call_4_7_11_15(SH2_struct* sh)
{
  SH2macw(sh,7,11);
}
#define init_4_7_12_15
void FASTCALL call_4_7_12_15(SH2_struct* sh)
{
  SH2macw(sh,7,12);
}
#define init_4_7_13_15
void FASTCALL call_4_7_13_15(SH2_struct* sh)
{
  SH2macw(sh,7,13);
}
#define init_4_7_14_15
void FASTCALL call_4_7_14_15(SH2_struct* sh)
{
  SH2macw(sh,7,14);
}
#define init_4_7_15_15
void FASTCALL call_4_7_15_15(SH2_struct* sh)
{
  SH2macw(sh,7,15);
}
#define init_4_8_0_15
void FASTCALL call_4_8_0_15(SH2_struct* sh)
{
  SH2macw(sh,8,0);
}
#define init_4_8_1_15
void FASTCALL call_4_8_1_15(SH2_struct* sh)
{
  SH2macw(sh,8,1);
}
#define init_4_8_2_15
void FASTCALL call_4_8_2_15(SH2_struct* sh)
{
  SH2macw(sh,8,2);
}
#define init_4_8_3_15
void FASTCALL call_4_8_3_15(SH2_struct* sh)
{
  SH2macw(sh,8,3);
}
#define init_4_8_4_15
void FASTCALL call_4_8_4_15(SH2_struct* sh)
{
  SH2macw(sh,8,4);
}
#define init_4_8_5_15
void FASTCALL call_4_8_5_15(SH2_struct* sh)
{
  SH2macw(sh,8,5);
}
#define init_4_8_6_15
void FASTCALL call_4_8_6_15(SH2_struct* sh)
{
  SH2macw(sh,8,6);
}
#define init_4_8_7_15
void FASTCALL call_4_8_7_15(SH2_struct* sh)
{
  SH2macw(sh,8,7);
}
#define init_4_8_8_15
void FASTCALL call_4_8_8_15(SH2_struct* sh)
{
  SH2macw(sh,8,8);
}
#define init_4_8_9_15
void FASTCALL call_4_8_9_15(SH2_struct* sh)
{
  SH2macw(sh,8,9);
}
#define init_4_8_10_15
void FASTCALL call_4_8_10_15(SH2_struct* sh)
{
  SH2macw(sh,8,10);
}
#define init_4_8_11_15
void FASTCALL call_4_8_11_15(SH2_struct* sh)
{
  SH2macw(sh,8,11);
}
#define init_4_8_12_15
void FASTCALL call_4_8_12_15(SH2_struct* sh)
{
  SH2macw(sh,8,12);
}
#define init_4_8_13_15
void FASTCALL call_4_8_13_15(SH2_struct* sh)
{
  SH2macw(sh,8,13);
}
#define init_4_8_14_15
void FASTCALL call_4_8_14_15(SH2_struct* sh)
{
  SH2macw(sh,8,14);
}
#define init_4_8_15_15
void FASTCALL call_4_8_15_15(SH2_struct* sh)
{
  SH2macw(sh,8,15);
}
#define init_4_9_0_15
void FASTCALL call_4_9_0_15(SH2_struct* sh)
{
  SH2macw(sh,9,0);
}
#define init_4_9_1_15
void FASTCALL call_4_9_1_15(SH2_struct* sh)
{
  SH2macw(sh,9,1);
}
#define init_4_9_2_15
void FASTCALL call_4_9_2_15(SH2_struct* sh)
{
  SH2macw(sh,9,2);
}
#define init_4_9_3_15
void FASTCALL call_4_9_3_15(SH2_struct* sh)
{
  SH2macw(sh,9,3);
}
#define init_4_9_4_15
void FASTCALL call_4_9_4_15(SH2_struct* sh)
{
  SH2macw(sh,9,4);
}
#define init_4_9_5_15
void FASTCALL call_4_9_5_15(SH2_struct* sh)
{
  SH2macw(sh,9,5);
}
#define init_4_9_6_15
void FASTCALL call_4_9_6_15(SH2_struct* sh)
{
  SH2macw(sh,9,6);
}
#define init_4_9_7_15
void FASTCALL call_4_9_7_15(SH2_struct* sh)
{
  SH2macw(sh,9,7);
}
#define init_4_9_8_15
void FASTCALL call_4_9_8_15(SH2_struct* sh)
{
  SH2macw(sh,9,8);
}
#define init_4_9_9_15
void FASTCALL call_4_9_9_15(SH2_struct* sh)
{
  SH2macw(sh,9,9);
}
#define init_4_9_10_15
void FASTCALL call_4_9_10_15(SH2_struct* sh)
{
  SH2macw(sh,9,10);
}
#define init_4_9_11_15
void FASTCALL call_4_9_11_15(SH2_struct* sh)
{
  SH2macw(sh,9,11);
}
#define init_4_9_12_15
void FASTCALL call_4_9_12_15(SH2_struct* sh)
{
  SH2macw(sh,9,12);
}
#define init_4_9_13_15
void FASTCALL call_4_9_13_15(SH2_struct* sh)
{
  SH2macw(sh,9,13);
}
#define init_4_9_14_15
void FASTCALL call_4_9_14_15(SH2_struct* sh)
{
  SH2macw(sh,9,14);
}
#define init_4_9_15_15
void FASTCALL call_4_9_15_15(SH2_struct* sh)
{
  SH2macw(sh,9,15);
}
#define init_4_10_0_15
void FASTCALL call_4_10_0_15(SH2_struct* sh)
{
  SH2macw(sh,10,0);
}
#define init_4_10_1_15
void FASTCALL call_4_10_1_15(SH2_struct* sh)
{
  SH2macw(sh,10,1);
}
#define init_4_10_2_15
void FASTCALL call_4_10_2_15(SH2_struct* sh)
{
  SH2macw(sh,10,2);
}
#define init_4_10_3_15
void FASTCALL call_4_10_3_15(SH2_struct* sh)
{
  SH2macw(sh,10,3);
}
#define init_4_10_4_15
void FASTCALL call_4_10_4_15(SH2_struct* sh)
{
  SH2macw(sh,10,4);
}
#define init_4_10_5_15
void FASTCALL call_4_10_5_15(SH2_struct* sh)
{
  SH2macw(sh,10,5);
}
#define init_4_10_6_15
void FASTCALL call_4_10_6_15(SH2_struct* sh)
{
  SH2macw(sh,10,6);
}
#define init_4_10_7_15
void FASTCALL call_4_10_7_15(SH2_struct* sh)
{
  SH2macw(sh,10,7);
}
#define init_4_10_8_15
void FASTCALL call_4_10_8_15(SH2_struct* sh)
{
  SH2macw(sh,10,8);
}
#define init_4_10_9_15
void FASTCALL call_4_10_9_15(SH2_struct* sh)
{
  SH2macw(sh,10,9);
}
#define init_4_10_10_15
void FASTCALL call_4_10_10_15(SH2_struct* sh)
{
  SH2macw(sh,10,10);
}
#define init_4_10_11_15
void FASTCALL call_4_10_11_15(SH2_struct* sh)
{
  SH2macw(sh,10,11);
}
#define init_4_10_12_15
void FASTCALL call_4_10_12_15(SH2_struct* sh)
{
  SH2macw(sh,10,12);
}
#define init_4_10_13_15
void FASTCALL call_4_10_13_15(SH2_struct* sh)
{
  SH2macw(sh,10,13);
}
#define init_4_10_14_15
void FASTCALL call_4_10_14_15(SH2_struct* sh)
{
  SH2macw(sh,10,14);
}
#define init_4_10_15_15
void FASTCALL call_4_10_15_15(SH2_struct* sh)
{
  SH2macw(sh,10,15);
}
#define init_4_11_0_15
void FASTCALL call_4_11_0_15(SH2_struct* sh)
{
  SH2macw(sh,11,0);
}
#define init_4_11_1_15
void FASTCALL call_4_11_1_15(SH2_struct* sh)
{
  SH2macw(sh,11,1);
}
#define init_4_11_2_15
void FASTCALL call_4_11_2_15(SH2_struct* sh)
{
  SH2macw(sh,11,2);
}
#define init_4_11_3_15
void FASTCALL call_4_11_3_15(SH2_struct* sh)
{
  SH2macw(sh,11,3);
}
#define init_4_11_4_15
void FASTCALL call_4_11_4_15(SH2_struct* sh)
{
  SH2macw(sh,11,4);
}
#define init_4_11_5_15
void FASTCALL call_4_11_5_15(SH2_struct* sh)
{
  SH2macw(sh,11,5);
}
#define init_4_11_6_15
void FASTCALL call_4_11_6_15(SH2_struct* sh)
{
  SH2macw(sh,11,6);
}
#define init_4_11_7_15
void FASTCALL call_4_11_7_15(SH2_struct* sh)
{
  SH2macw(sh,11,7);
}
#define init_4_11_8_15
void FASTCALL call_4_11_8_15(SH2_struct* sh)
{
  SH2macw(sh,11,8);
}
#define init_4_11_9_15
void FASTCALL call_4_11_9_15(SH2_struct* sh)
{
  SH2macw(sh,11,9);
}
#define init_4_11_10_15
void FASTCALL call_4_11_10_15(SH2_struct* sh)
{
  SH2macw(sh,11,10);
}
#define init_4_11_11_15
void FASTCALL call_4_11_11_15(SH2_struct* sh)
{
  SH2macw(sh,11,11);
}
#define init_4_11_12_15
void FASTCALL call_4_11_12_15(SH2_struct* sh)
{
  SH2macw(sh,11,12);
}
#define init_4_11_13_15
void FASTCALL call_4_11_13_15(SH2_struct* sh)
{
  SH2macw(sh,11,13);
}
#define init_4_11_14_15
void FASTCALL call_4_11_14_15(SH2_struct* sh)
{
  SH2macw(sh,11,14);
}
#define init_4_11_15_15
void FASTCALL call_4_11_15_15(SH2_struct* sh)
{
  SH2macw(sh,11,15);
}
#define init_4_12_0_15
void FASTCALL call_4_12_0_15(SH2_struct* sh)
{
  SH2macw(sh,12,0);
}
#define init_4_12_1_15
void FASTCALL call_4_12_1_15(SH2_struct* sh)
{
  SH2macw(sh,12,1);
}
#define init_4_12_2_15
void FASTCALL call_4_12_2_15(SH2_struct* sh)
{
  SH2macw(sh,12,2);
}
#define init_4_12_3_15
void FASTCALL call_4_12_3_15(SH2_struct* sh)
{
  SH2macw(sh,12,3);
}
#define init_4_12_4_15
void FASTCALL call_4_12_4_15(SH2_struct* sh)
{
  SH2macw(sh,12,4);
}
#define init_4_12_5_15
void FASTCALL call_4_12_5_15(SH2_struct* sh)
{
  SH2macw(sh,12,5);
}
#define init_4_12_6_15
void FASTCALL call_4_12_6_15(SH2_struct* sh)
{
  SH2macw(sh,12,6);
}
#define init_4_12_7_15
void FASTCALL call_4_12_7_15(SH2_struct* sh)
{
  SH2macw(sh,12,7);
}
#define init_4_12_8_15
void FASTCALL call_4_12_8_15(SH2_struct* sh)
{
  SH2macw(sh,12,8);
}
#define init_4_12_9_15
void FASTCALL call_4_12_9_15(SH2_struct* sh)
{
  SH2macw(sh,12,9);
}
#define init_4_12_10_15
void FASTCALL call_4_12_10_15(SH2_struct* sh)
{
  SH2macw(sh,12,10);
}
#define init_4_12_11_15
void FASTCALL call_4_12_11_15(SH2_struct* sh)
{
  SH2macw(sh,12,11);
}
#define init_4_12_12_15
void FASTCALL call_4_12_12_15(SH2_struct* sh)
{
  SH2macw(sh,12,12);
}
#define init_4_12_13_15
void FASTCALL call_4_12_13_15(SH2_struct* sh)
{
  SH2macw(sh,12,13);
}
#define init_4_12_14_15
void FASTCALL call_4_12_14_15(SH2_struct* sh)
{
  SH2macw(sh,12,14);
}
#define init_4_12_15_15
void FASTCALL call_4_12_15_15(SH2_struct* sh)
{
  SH2macw(sh,12,15);
}
#define init_4_13_0_15
void FASTCALL call_4_13_0_15(SH2_struct* sh)
{
  SH2macw(sh,13,0);
}
#define init_4_13_1_15
void FASTCALL call_4_13_1_15(SH2_struct* sh)
{
  SH2macw(sh,13,1);
}
#define init_4_13_2_15
void FASTCALL call_4_13_2_15(SH2_struct* sh)
{
  SH2macw(sh,13,2);
}
#define init_4_13_3_15
void FASTCALL call_4_13_3_15(SH2_struct* sh)
{
  SH2macw(sh,13,3);
}
#define init_4_13_4_15
void FASTCALL call_4_13_4_15(SH2_struct* sh)
{
  SH2macw(sh,13,4);
}
#define init_4_13_5_15
void FASTCALL call_4_13_5_15(SH2_struct* sh)
{
  SH2macw(sh,13,5);
}
#define init_4_13_6_15
void FASTCALL call_4_13_6_15(SH2_struct* sh)
{
  SH2macw(sh,13,6);
}
#define init_4_13_7_15
void FASTCALL call_4_13_7_15(SH2_struct* sh)
{
  SH2macw(sh,13,7);
}
#define init_4_13_8_15
void FASTCALL call_4_13_8_15(SH2_struct* sh)
{
  SH2macw(sh,13,8);
}
#define init_4_13_9_15
void FASTCALL call_4_13_9_15(SH2_struct* sh)
{
  SH2macw(sh,13,9);
}
#define init_4_13_10_15
void FASTCALL call_4_13_10_15(SH2_struct* sh)
{
  SH2macw(sh,13,10);
}
#define init_4_13_11_15
void FASTCALL call_4_13_11_15(SH2_struct* sh)
{
  SH2macw(sh,13,11);
}
#define init_4_13_12_15
void FASTCALL call_4_13_12_15(SH2_struct* sh)
{
  SH2macw(sh,13,12);
}
#define init_4_13_13_15
void FASTCALL call_4_13_13_15(SH2_struct* sh)
{
  SH2macw(sh,13,13);
}
#define init_4_13_14_15
void FASTCALL call_4_13_14_15(SH2_struct* sh)
{
  SH2macw(sh,13,14);
}
#define init_4_13_15_15
void FASTCALL call_4_13_15_15(SH2_struct* sh)
{
  SH2macw(sh,13,15);
}
#define init_4_14_0_15
void FASTCALL call_4_14_0_15(SH2_struct* sh)
{
  SH2macw(sh,14,0);
}
#define init_4_14_1_15
void FASTCALL call_4_14_1_15(SH2_struct* sh)
{
  SH2macw(sh,14,1);
}
#define init_4_14_2_15
void FASTCALL call_4_14_2_15(SH2_struct* sh)
{
  SH2macw(sh,14,2);
}
#define init_4_14_3_15
void FASTCALL call_4_14_3_15(SH2_struct* sh)
{
  SH2macw(sh,14,3);
}
#define init_4_14_4_15
void FASTCALL call_4_14_4_15(SH2_struct* sh)
{
  SH2macw(sh,14,4);
}
#define init_4_14_5_15
void FASTCALL call_4_14_5_15(SH2_struct* sh)
{
  SH2macw(sh,14,5);
}
#define init_4_14_6_15
void FASTCALL call_4_14_6_15(SH2_struct* sh)
{
  SH2macw(sh,14,6);
}
#define init_4_14_7_15
void FASTCALL call_4_14_7_15(SH2_struct* sh)
{
  SH2macw(sh,14,7);
}
#define init_4_14_8_15
void FASTCALL call_4_14_8_15(SH2_struct* sh)
{
  SH2macw(sh,14,8);
}
#define init_4_14_9_15
void FASTCALL call_4_14_9_15(SH2_struct* sh)
{
  SH2macw(sh,14,9);
}
#define init_4_14_10_15
void FASTCALL call_4_14_10_15(SH2_struct* sh)
{
  SH2macw(sh,14,10);
}
#define init_4_14_11_15
void FASTCALL call_4_14_11_15(SH2_struct* sh)
{
  SH2macw(sh,14,11);
}
#define init_4_14_12_15
void FASTCALL call_4_14_12_15(SH2_struct* sh)
{
  SH2macw(sh,14,12);
}
#define init_4_14_13_15
void FASTCALL call_4_14_13_15(SH2_struct* sh)
{
  SH2macw(sh,14,13);
}
#define init_4_14_14_15
void FASTCALL call_4_14_14_15(SH2_struct* sh)
{
  SH2macw(sh,14,14);
}
#define init_4_14_15_15
void FASTCALL call_4_14_15_15(SH2_struct* sh)
{
  SH2macw(sh,14,15);
}
#define init_4_15_0_15
void FASTCALL call_4_15_0_15(SH2_struct* sh)
{
  SH2macw(sh,15,0);
}
#define init_4_15_1_15
void FASTCALL call_4_15_1_15(SH2_struct* sh)
{
  SH2macw(sh,15,1);
}
#define init_4_15_2_15
void FASTCALL call_4_15_2_15(SH2_struct* sh)
{
  SH2macw(sh,15,2);
}
#define init_4_15_3_15
void FASTCALL call_4_15_3_15(SH2_struct* sh)
{
  SH2macw(sh,15,3);
}
#define init_4_15_4_15
void FASTCALL call_4_15_4_15(SH2_struct* sh)
{
  SH2macw(sh,15,4);
}
#define init_4_15_5_15
void FASTCALL call_4_15_5_15(SH2_struct* sh)
{
  SH2macw(sh,15,5);
}
#define init_4_15_6_15
void FASTCALL call_4_15_6_15(SH2_struct* sh)
{
  SH2macw(sh,15,6);
}
#define init_4_15_7_15
void FASTCALL call_4_15_7_15(SH2_struct* sh)
{
  SH2macw(sh,15,7);
}
#define init_4_15_8_15
void FASTCALL call_4_15_8_15(SH2_struct* sh)
{
  SH2macw(sh,15,8);
}
#define init_4_15_9_15
void FASTCALL call_4_15_9_15(SH2_struct* sh)
{
  SH2macw(sh,15,9);
}
#define init_4_15_10_15
void FASTCALL call_4_15_10_15(SH2_struct* sh)
{
  SH2macw(sh,15,10);
}
#define init_4_15_11_15
void FASTCALL call_4_15_11_15(SH2_struct* sh)
{
  SH2macw(sh,15,11);
}
#define init_4_15_12_15
void FASTCALL call_4_15_12_15(SH2_struct* sh)
{
  SH2macw(sh,15,12);
}
#define init_4_15_13_15
void FASTCALL call_4_15_13_15(SH2_struct* sh)
{
  SH2macw(sh,15,13);
}
#define init_4_15_14_15
void FASTCALL call_4_15_14_15(SH2_struct* sh)
{
  SH2macw(sh,15,14);
}
#define init_4_15_15_15
void FASTCALL call_4_15_15_15(SH2_struct* sh)
{
  SH2macw(sh,15,15);
}
#define init_4_0_2_10
void FASTCALL call_4_0_2_10(SH2_struct* sh)
{
  SH2ldspr(sh,0);
}
#define init_4_1_2_10
void FASTCALL call_4_1_2_10(SH2_struct* sh)
{
  SH2ldspr(sh,1);
}
#define init_4_2_2_10
void FASTCALL call_4_2_2_10(SH2_struct* sh)
{
  SH2ldspr(sh,2);
}
#define init_4_3_2_10
void FASTCALL call_4_3_2_10(SH2_struct* sh)
{
  SH2ldspr(sh,3);
}
#define init_4_4_2_10
void FASTCALL call_4_4_2_10(SH2_struct* sh)
{
  SH2ldspr(sh,4);
}
#define init_4_5_2_10
void FASTCALL call_4_5_2_10(SH2_struct* sh)
{
  SH2ldspr(sh,5);
}
#define init_4_6_2_10
void FASTCALL call_4_6_2_10(SH2_struct* sh)
{
  SH2ldspr(sh,6);
}
#define init_4_7_2_10
void FASTCALL call_4_7_2_10(SH2_struct* sh)
{
  SH2ldspr(sh,7);
}
#define init_4_8_2_10
void FASTCALL call_4_8_2_10(SH2_struct* sh)
{
  SH2ldspr(sh,8);
}
#define init_4_9_2_10
void FASTCALL call_4_9_2_10(SH2_struct* sh)
{
  SH2ldspr(sh,9);
}
#define init_4_10_2_10
void FASTCALL call_4_10_2_10(SH2_struct* sh)
{
  SH2ldspr(sh,10);
}
#define init_4_11_2_10
void FASTCALL call_4_11_2_10(SH2_struct* sh)
{
  SH2ldspr(sh,11);
}
#define init_4_12_2_10
void FASTCALL call_4_12_2_10(SH2_struct* sh)
{
  SH2ldspr(sh,12);
}
#define init_4_13_2_10
void FASTCALL call_4_13_2_10(SH2_struct* sh)
{
  SH2ldspr(sh,13);
}
#define init_4_14_2_10
void FASTCALL call_4_14_2_10(SH2_struct* sh)
{
  SH2ldspr(sh,14);
}
#define init_4_15_2_10
void FASTCALL call_4_15_2_10(SH2_struct* sh)
{
  SH2ldspr(sh,15);
}
#define init_4_0_2_6
void FASTCALL call_4_0_2_6(SH2_struct* sh)
{
  SH2ldsmpr(sh,0);
}
#define init_4_1_2_6
void FASTCALL call_4_1_2_6(SH2_struct* sh)
{
  SH2ldsmpr(sh,1);
}
#define init_4_2_2_6
void FASTCALL call_4_2_2_6(SH2_struct* sh)
{
  SH2ldsmpr(sh,2);
}
#define init_4_3_2_6
void FASTCALL call_4_3_2_6(SH2_struct* sh)
{
  SH2ldsmpr(sh,3);
}
#define init_4_4_2_6
void FASTCALL call_4_4_2_6(SH2_struct* sh)
{
  SH2ldsmpr(sh,4);
}
#define init_4_5_2_6
void FASTCALL call_4_5_2_6(SH2_struct* sh)
{
  SH2ldsmpr(sh,5);
}
#define init_4_6_2_6
void FASTCALL call_4_6_2_6(SH2_struct* sh)
{
  SH2ldsmpr(sh,6);
}
#define init_4_7_2_6
void FASTCALL call_4_7_2_6(SH2_struct* sh)
{
  SH2ldsmpr(sh,7);
}
#define init_4_8_2_6
void FASTCALL call_4_8_2_6(SH2_struct* sh)
{
  SH2ldsmpr(sh,8);
}
#define init_4_9_2_6
void FASTCALL call_4_9_2_6(SH2_struct* sh)
{
  SH2ldsmpr(sh,9);
}
#define init_4_10_2_6
void FASTCALL call_4_10_2_6(SH2_struct* sh)
{
  SH2ldsmpr(sh,10);
}
#define init_4_11_2_6
void FASTCALL call_4_11_2_6(SH2_struct* sh)
{
  SH2ldsmpr(sh,11);
}
#define init_4_12_2_6
void FASTCALL call_4_12_2_6(SH2_struct* sh)
{
  SH2ldsmpr(sh,12);
}
#define init_4_13_2_6
void FASTCALL call_4_13_2_6(SH2_struct* sh)
{
  SH2ldsmpr(sh,13);
}
#define init_4_14_2_6
void FASTCALL call_4_14_2_6(SH2_struct* sh)
{
  SH2ldsmpr(sh,14);
}
#define init_4_15_2_6
void FASTCALL call_4_15_2_6(SH2_struct* sh)
{
  SH2ldsmpr(sh,15);
}
#define init_4_0_1_6
void FASTCALL call_4_0_1_6(SH2_struct* sh)
{
  SH2ldsmmacl(sh,0);
}
#define init_4_1_1_6
void FASTCALL call_4_1_1_6(SH2_struct* sh)
{
  SH2ldsmmacl(sh,1);
}
#define init_4_2_1_6
void FASTCALL call_4_2_1_6(SH2_struct* sh)
{
  SH2ldsmmacl(sh,2);
}
#define init_4_3_1_6
void FASTCALL call_4_3_1_6(SH2_struct* sh)
{
  SH2ldsmmacl(sh,3);
}
#define init_4_4_1_6
void FASTCALL call_4_4_1_6(SH2_struct* sh)
{
  SH2ldsmmacl(sh,4);
}
#define init_4_5_1_6
void FASTCALL call_4_5_1_6(SH2_struct* sh)
{
  SH2ldsmmacl(sh,5);
}
#define init_4_6_1_6
void FASTCALL call_4_6_1_6(SH2_struct* sh)
{
  SH2ldsmmacl(sh,6);
}
#define init_4_7_1_6
void FASTCALL call_4_7_1_6(SH2_struct* sh)
{
  SH2ldsmmacl(sh,7);
}
#define init_4_8_1_6
void FASTCALL call_4_8_1_6(SH2_struct* sh)
{
  SH2ldsmmacl(sh,8);
}
#define init_4_9_1_6
void FASTCALL call_4_9_1_6(SH2_struct* sh)
{
  SH2ldsmmacl(sh,9);
}
#define init_4_10_1_6
void FASTCALL call_4_10_1_6(SH2_struct* sh)
{
  SH2ldsmmacl(sh,10);
}
#define init_4_11_1_6
void FASTCALL call_4_11_1_6(SH2_struct* sh)
{
  SH2ldsmmacl(sh,11);
}
#define init_4_12_1_6
void FASTCALL call_4_12_1_6(SH2_struct* sh)
{
  SH2ldsmmacl(sh,12);
}
#define init_4_13_1_6
void FASTCALL call_4_13_1_6(SH2_struct* sh)
{
  SH2ldsmmacl(sh,13);
}
#define init_4_14_1_6
void FASTCALL call_4_14_1_6(SH2_struct* sh)
{
  SH2ldsmmacl(sh,14);
}
#define init_4_15_1_6
void FASTCALL call_4_15_1_6(SH2_struct* sh)
{
  SH2ldsmmacl(sh,15);
}
#define init_4_0_0_6
void FASTCALL call_4_0_0_6(SH2_struct* sh)
{
  SH2ldsmmach(sh,0);
}
#define init_4_1_0_6
void FASTCALL call_4_1_0_6(SH2_struct* sh)
{
  SH2ldsmmach(sh,1);
}
#define init_4_2_0_6
void FASTCALL call_4_2_0_6(SH2_struct* sh)
{
  SH2ldsmmach(sh,2);
}
#define init_4_3_0_6
void FASTCALL call_4_3_0_6(SH2_struct* sh)
{
  SH2ldsmmach(sh,3);
}
#define init_4_4_0_6
void FASTCALL call_4_4_0_6(SH2_struct* sh)
{
  SH2ldsmmach(sh,4);
}
#define init_4_5_0_6
void FASTCALL call_4_5_0_6(SH2_struct* sh)
{
  SH2ldsmmach(sh,5);
}
#define init_4_6_0_6
void FASTCALL call_4_6_0_6(SH2_struct* sh)
{
  SH2ldsmmach(sh,6);
}
#define init_4_7_0_6
void FASTCALL call_4_7_0_6(SH2_struct* sh)
{
  SH2ldsmmach(sh,7);
}
#define init_4_8_0_6
void FASTCALL call_4_8_0_6(SH2_struct* sh)
{
  SH2ldsmmach(sh,8);
}
#define init_4_9_0_6
void FASTCALL call_4_9_0_6(SH2_struct* sh)
{
  SH2ldsmmach(sh,9);
}
#define init_4_10_0_6
void FASTCALL call_4_10_0_6(SH2_struct* sh)
{
  SH2ldsmmach(sh,10);
}
#define init_4_11_0_6
void FASTCALL call_4_11_0_6(SH2_struct* sh)
{
  SH2ldsmmach(sh,11);
}
#define init_4_12_0_6
void FASTCALL call_4_12_0_6(SH2_struct* sh)
{
  SH2ldsmmach(sh,12);
}
#define init_4_13_0_6
void FASTCALL call_4_13_0_6(SH2_struct* sh)
{
  SH2ldsmmach(sh,13);
}
#define init_4_14_0_6
void FASTCALL call_4_14_0_6(SH2_struct* sh)
{
  SH2ldsmmach(sh,14);
}
#define init_4_15_0_6
void FASTCALL call_4_15_0_6(SH2_struct* sh)
{
  SH2ldsmmach(sh,15);
}
#define init_4_0_1_10
void FASTCALL call_4_0_1_10(SH2_struct* sh)
{
  SH2ldsmacl(sh,0);
}
#define init_4_1_1_10
void FASTCALL call_4_1_1_10(SH2_struct* sh)
{
  SH2ldsmacl(sh,1);
}
#define init_4_2_1_10
void FASTCALL call_4_2_1_10(SH2_struct* sh)
{
  SH2ldsmacl(sh,2);
}
#define init_4_3_1_10
void FASTCALL call_4_3_1_10(SH2_struct* sh)
{
  SH2ldsmacl(sh,3);
}
#define init_4_4_1_10
void FASTCALL call_4_4_1_10(SH2_struct* sh)
{
  SH2ldsmacl(sh,4);
}
#define init_4_5_1_10
void FASTCALL call_4_5_1_10(SH2_struct* sh)
{
  SH2ldsmacl(sh,5);
}
#define init_4_6_1_10
void FASTCALL call_4_6_1_10(SH2_struct* sh)
{
  SH2ldsmacl(sh,6);
}
#define init_4_7_1_10
void FASTCALL call_4_7_1_10(SH2_struct* sh)
{
  SH2ldsmacl(sh,7);
}
#define init_4_8_1_10
void FASTCALL call_4_8_1_10(SH2_struct* sh)
{
  SH2ldsmacl(sh,8);
}
#define init_4_9_1_10
void FASTCALL call_4_9_1_10(SH2_struct* sh)
{
  SH2ldsmacl(sh,9);
}
#define init_4_10_1_10
void FASTCALL call_4_10_1_10(SH2_struct* sh)
{
  SH2ldsmacl(sh,10);
}
#define init_4_11_1_10
void FASTCALL call_4_11_1_10(SH2_struct* sh)
{
  SH2ldsmacl(sh,11);
}
#define init_4_12_1_10
void FASTCALL call_4_12_1_10(SH2_struct* sh)
{
  SH2ldsmacl(sh,12);
}
#define init_4_13_1_10
void FASTCALL call_4_13_1_10(SH2_struct* sh)
{
  SH2ldsmacl(sh,13);
}
#define init_4_14_1_10
void FASTCALL call_4_14_1_10(SH2_struct* sh)
{
  SH2ldsmacl(sh,14);
}
#define init_4_15_1_10
void FASTCALL call_4_15_1_10(SH2_struct* sh)
{
  SH2ldsmacl(sh,15);
}
#define init_4_0_0_10
void FASTCALL call_4_0_0_10(SH2_struct* sh)
{
  SH2ldsmach(sh,0);
}
#define init_4_1_0_10
void FASTCALL call_4_1_0_10(SH2_struct* sh)
{
  SH2ldsmach(sh,1);
}
#define init_4_2_0_10
void FASTCALL call_4_2_0_10(SH2_struct* sh)
{
  SH2ldsmach(sh,2);
}
#define init_4_3_0_10
void FASTCALL call_4_3_0_10(SH2_struct* sh)
{
  SH2ldsmach(sh,3);
}
#define init_4_4_0_10
void FASTCALL call_4_4_0_10(SH2_struct* sh)
{
  SH2ldsmach(sh,4);
}
#define init_4_5_0_10
void FASTCALL call_4_5_0_10(SH2_struct* sh)
{
  SH2ldsmach(sh,5);
}
#define init_4_6_0_10
void FASTCALL call_4_6_0_10(SH2_struct* sh)
{
  SH2ldsmach(sh,6);
}
#define init_4_7_0_10
void FASTCALL call_4_7_0_10(SH2_struct* sh)
{
  SH2ldsmach(sh,7);
}
#define init_4_8_0_10
void FASTCALL call_4_8_0_10(SH2_struct* sh)
{
  SH2ldsmach(sh,8);
}
#define init_4_9_0_10
void FASTCALL call_4_9_0_10(SH2_struct* sh)
{
  SH2ldsmach(sh,9);
}
#define init_4_10_0_10
void FASTCALL call_4_10_0_10(SH2_struct* sh)
{
  SH2ldsmach(sh,10);
}
#define init_4_11_0_10
void FASTCALL call_4_11_0_10(SH2_struct* sh)
{
  SH2ldsmach(sh,11);
}
#define init_4_12_0_10
void FASTCALL call_4_12_0_10(SH2_struct* sh)
{
  SH2ldsmach(sh,12);
}
#define init_4_13_0_10
void FASTCALL call_4_13_0_10(SH2_struct* sh)
{
  SH2ldsmach(sh,13);
}
#define init_4_14_0_10
void FASTCALL call_4_14_0_10(SH2_struct* sh)
{
  SH2ldsmach(sh,14);
}
#define init_4_15_0_10
void FASTCALL call_4_15_0_10(SH2_struct* sh)
{
  SH2ldsmach(sh,15);
}
#define init_4_0_2_14
void FASTCALL call_4_0_2_14(SH2_struct* sh)
{
  SH2ldcvbr(sh,0);
}
#define init_4_1_2_14
void FASTCALL call_4_1_2_14(SH2_struct* sh)
{
  SH2ldcvbr(sh,1);
}
#define init_4_2_2_14
void FASTCALL call_4_2_2_14(SH2_struct* sh)
{
  SH2ldcvbr(sh,2);
}
#define init_4_3_2_14
void FASTCALL call_4_3_2_14(SH2_struct* sh)
{
  SH2ldcvbr(sh,3);
}
#define init_4_4_2_14
void FASTCALL call_4_4_2_14(SH2_struct* sh)
{
  SH2ldcvbr(sh,4);
}
#define init_4_5_2_14
void FASTCALL call_4_5_2_14(SH2_struct* sh)
{
  SH2ldcvbr(sh,5);
}
#define init_4_6_2_14
void FASTCALL call_4_6_2_14(SH2_struct* sh)
{
  SH2ldcvbr(sh,6);
}
#define init_4_7_2_14
void FASTCALL call_4_7_2_14(SH2_struct* sh)
{
  SH2ldcvbr(sh,7);
}
#define init_4_8_2_14
void FASTCALL call_4_8_2_14(SH2_struct* sh)
{
  SH2ldcvbr(sh,8);
}
#define init_4_9_2_14
void FASTCALL call_4_9_2_14(SH2_struct* sh)
{
  SH2ldcvbr(sh,9);
}
#define init_4_10_2_14
void FASTCALL call_4_10_2_14(SH2_struct* sh)
{
  SH2ldcvbr(sh,10);
}
#define init_4_11_2_14
void FASTCALL call_4_11_2_14(SH2_struct* sh)
{
  SH2ldcvbr(sh,11);
}
#define init_4_12_2_14
void FASTCALL call_4_12_2_14(SH2_struct* sh)
{
  SH2ldcvbr(sh,12);
}
#define init_4_13_2_14
void FASTCALL call_4_13_2_14(SH2_struct* sh)
{
  SH2ldcvbr(sh,13);
}
#define init_4_14_2_14
void FASTCALL call_4_14_2_14(SH2_struct* sh)
{
  SH2ldcvbr(sh,14);
}
#define init_4_15_2_14
void FASTCALL call_4_15_2_14(SH2_struct* sh)
{
  SH2ldcvbr(sh,15);
}
#define init_4_0_0_14
void FASTCALL call_4_0_0_14(SH2_struct* sh)
{
  SH2ldcsr(sh,0);
}
#define init_4_1_0_14
void FASTCALL call_4_1_0_14(SH2_struct* sh)
{
  SH2ldcsr(sh,1);
}
#define init_4_2_0_14
void FASTCALL call_4_2_0_14(SH2_struct* sh)
{
  SH2ldcsr(sh,2);
}
#define init_4_3_0_14
void FASTCALL call_4_3_0_14(SH2_struct* sh)
{
  SH2ldcsr(sh,3);
}
#define init_4_4_0_14
void FASTCALL call_4_4_0_14(SH2_struct* sh)
{
  SH2ldcsr(sh,4);
}
#define init_4_5_0_14
void FASTCALL call_4_5_0_14(SH2_struct* sh)
{
  SH2ldcsr(sh,5);
}
#define init_4_6_0_14
void FASTCALL call_4_6_0_14(SH2_struct* sh)
{
  SH2ldcsr(sh,6);
}
#define init_4_7_0_14
void FASTCALL call_4_7_0_14(SH2_struct* sh)
{
  SH2ldcsr(sh,7);
}
#define init_4_8_0_14
void FASTCALL call_4_8_0_14(SH2_struct* sh)
{
  SH2ldcsr(sh,8);
}
#define init_4_9_0_14
void FASTCALL call_4_9_0_14(SH2_struct* sh)
{
  SH2ldcsr(sh,9);
}
#define init_4_10_0_14
void FASTCALL call_4_10_0_14(SH2_struct* sh)
{
  SH2ldcsr(sh,10);
}
#define init_4_11_0_14
void FASTCALL call_4_11_0_14(SH2_struct* sh)
{
  SH2ldcsr(sh,11);
}
#define init_4_12_0_14
void FASTCALL call_4_12_0_14(SH2_struct* sh)
{
  SH2ldcsr(sh,12);
}
#define init_4_13_0_14
void FASTCALL call_4_13_0_14(SH2_struct* sh)
{
  SH2ldcsr(sh,13);
}
#define init_4_14_0_14
void FASTCALL call_4_14_0_14(SH2_struct* sh)
{
  SH2ldcsr(sh,14);
}
#define init_4_15_0_14
void FASTCALL call_4_15_0_14(SH2_struct* sh)
{
  SH2ldcsr(sh,15);
}
#define init_4_0_2_7
void FASTCALL call_4_0_2_7(SH2_struct* sh)
{
  SH2ldcmvbr(sh,0);
}
#define init_4_1_2_7
void FASTCALL call_4_1_2_7(SH2_struct* sh)
{
  SH2ldcmvbr(sh,1);
}
#define init_4_2_2_7
void FASTCALL call_4_2_2_7(SH2_struct* sh)
{
  SH2ldcmvbr(sh,2);
}
#define init_4_3_2_7
void FASTCALL call_4_3_2_7(SH2_struct* sh)
{
  SH2ldcmvbr(sh,3);
}
#define init_4_4_2_7
void FASTCALL call_4_4_2_7(SH2_struct* sh)
{
  SH2ldcmvbr(sh,4);
}
#define init_4_5_2_7
void FASTCALL call_4_5_2_7(SH2_struct* sh)
{
  SH2ldcmvbr(sh,5);
}
#define init_4_6_2_7
void FASTCALL call_4_6_2_7(SH2_struct* sh)
{
  SH2ldcmvbr(sh,6);
}
#define init_4_7_2_7
void FASTCALL call_4_7_2_7(SH2_struct* sh)
{
  SH2ldcmvbr(sh,7);
}
#define init_4_8_2_7
void FASTCALL call_4_8_2_7(SH2_struct* sh)
{
  SH2ldcmvbr(sh,8);
}
#define init_4_9_2_7
void FASTCALL call_4_9_2_7(SH2_struct* sh)
{
  SH2ldcmvbr(sh,9);
}
#define init_4_10_2_7
void FASTCALL call_4_10_2_7(SH2_struct* sh)
{
  SH2ldcmvbr(sh,10);
}
#define init_4_11_2_7
void FASTCALL call_4_11_2_7(SH2_struct* sh)
{
  SH2ldcmvbr(sh,11);
}
#define init_4_12_2_7
void FASTCALL call_4_12_2_7(SH2_struct* sh)
{
  SH2ldcmvbr(sh,12);
}
#define init_4_13_2_7
void FASTCALL call_4_13_2_7(SH2_struct* sh)
{
  SH2ldcmvbr(sh,13);
}
#define init_4_14_2_7
void FASTCALL call_4_14_2_7(SH2_struct* sh)
{
  SH2ldcmvbr(sh,14);
}
#define init_4_15_2_7
void FASTCALL call_4_15_2_7(SH2_struct* sh)
{
  SH2ldcmvbr(sh,15);
}
#define init_4_0_0_7
void FASTCALL call_4_0_0_7(SH2_struct* sh)
{
  SH2ldcmsr(sh,0);
}
#define init_4_1_0_7
void FASTCALL call_4_1_0_7(SH2_struct* sh)
{
  SH2ldcmsr(sh,1);
}
#define init_4_2_0_7
void FASTCALL call_4_2_0_7(SH2_struct* sh)
{
  SH2ldcmsr(sh,2);
}
#define init_4_3_0_7
void FASTCALL call_4_3_0_7(SH2_struct* sh)
{
  SH2ldcmsr(sh,3);
}
#define init_4_4_0_7
void FASTCALL call_4_4_0_7(SH2_struct* sh)
{
  SH2ldcmsr(sh,4);
}
#define init_4_5_0_7
void FASTCALL call_4_5_0_7(SH2_struct* sh)
{
  SH2ldcmsr(sh,5);
}
#define init_4_6_0_7
void FASTCALL call_4_6_0_7(SH2_struct* sh)
{
  SH2ldcmsr(sh,6);
}
#define init_4_7_0_7
void FASTCALL call_4_7_0_7(SH2_struct* sh)
{
  SH2ldcmsr(sh,7);
}
#define init_4_8_0_7
void FASTCALL call_4_8_0_7(SH2_struct* sh)
{
  SH2ldcmsr(sh,8);
}
#define init_4_9_0_7
void FASTCALL call_4_9_0_7(SH2_struct* sh)
{
  SH2ldcmsr(sh,9);
}
#define init_4_10_0_7
void FASTCALL call_4_10_0_7(SH2_struct* sh)
{
  SH2ldcmsr(sh,10);
}
#define init_4_11_0_7
void FASTCALL call_4_11_0_7(SH2_struct* sh)
{
  SH2ldcmsr(sh,11);
}
#define init_4_12_0_7
void FASTCALL call_4_12_0_7(SH2_struct* sh)
{
  SH2ldcmsr(sh,12);
}
#define init_4_13_0_7
void FASTCALL call_4_13_0_7(SH2_struct* sh)
{
  SH2ldcmsr(sh,13);
}
#define init_4_14_0_7
void FASTCALL call_4_14_0_7(SH2_struct* sh)
{
  SH2ldcmsr(sh,14);
}
#define init_4_15_0_7
void FASTCALL call_4_15_0_7(SH2_struct* sh)
{
  SH2ldcmsr(sh,15);
}
#define init_4_0_1_7
void FASTCALL call_4_0_1_7(SH2_struct* sh)
{
  SH2ldcmgbr(sh,0);
}
#define init_4_1_1_7
void FASTCALL call_4_1_1_7(SH2_struct* sh)
{
  SH2ldcmgbr(sh,1);
}
#define init_4_2_1_7
void FASTCALL call_4_2_1_7(SH2_struct* sh)
{
  SH2ldcmgbr(sh,2);
}
#define init_4_3_1_7
void FASTCALL call_4_3_1_7(SH2_struct* sh)
{
  SH2ldcmgbr(sh,3);
}
#define init_4_4_1_7
void FASTCALL call_4_4_1_7(SH2_struct* sh)
{
  SH2ldcmgbr(sh,4);
}
#define init_4_5_1_7
void FASTCALL call_4_5_1_7(SH2_struct* sh)
{
  SH2ldcmgbr(sh,5);
}
#define init_4_6_1_7
void FASTCALL call_4_6_1_7(SH2_struct* sh)
{
  SH2ldcmgbr(sh,6);
}
#define init_4_7_1_7
void FASTCALL call_4_7_1_7(SH2_struct* sh)
{
  SH2ldcmgbr(sh,7);
}
#define init_4_8_1_7
void FASTCALL call_4_8_1_7(SH2_struct* sh)
{
  SH2ldcmgbr(sh,8);
}
#define init_4_9_1_7
void FASTCALL call_4_9_1_7(SH2_struct* sh)
{
  SH2ldcmgbr(sh,9);
}
#define init_4_10_1_7
void FASTCALL call_4_10_1_7(SH2_struct* sh)
{
  SH2ldcmgbr(sh,10);
}
#define init_4_11_1_7
void FASTCALL call_4_11_1_7(SH2_struct* sh)
{
  SH2ldcmgbr(sh,11);
}
#define init_4_12_1_7
void FASTCALL call_4_12_1_7(SH2_struct* sh)
{
  SH2ldcmgbr(sh,12);
}
#define init_4_13_1_7
void FASTCALL call_4_13_1_7(SH2_struct* sh)
{
  SH2ldcmgbr(sh,13);
}
#define init_4_14_1_7
void FASTCALL call_4_14_1_7(SH2_struct* sh)
{
  SH2ldcmgbr(sh,14);
}
#define init_4_15_1_7
void FASTCALL call_4_15_1_7(SH2_struct* sh)
{
  SH2ldcmgbr(sh,15);
}
#define init_4_0_1_14
void FASTCALL call_4_0_1_14(SH2_struct* sh)
{
  SH2ldcgbr(sh,0);
}
#define init_4_1_1_14
void FASTCALL call_4_1_1_14(SH2_struct* sh)
{
  SH2ldcgbr(sh,1);
}
#define init_4_2_1_14
void FASTCALL call_4_2_1_14(SH2_struct* sh)
{
  SH2ldcgbr(sh,2);
}
#define init_4_3_1_14
void FASTCALL call_4_3_1_14(SH2_struct* sh)
{
  SH2ldcgbr(sh,3);
}
#define init_4_4_1_14
void FASTCALL call_4_4_1_14(SH2_struct* sh)
{
  SH2ldcgbr(sh,4);
}
#define init_4_5_1_14
void FASTCALL call_4_5_1_14(SH2_struct* sh)
{
  SH2ldcgbr(sh,5);
}
#define init_4_6_1_14
void FASTCALL call_4_6_1_14(SH2_struct* sh)
{
  SH2ldcgbr(sh,6);
}
#define init_4_7_1_14
void FASTCALL call_4_7_1_14(SH2_struct* sh)
{
  SH2ldcgbr(sh,7);
}
#define init_4_8_1_14
void FASTCALL call_4_8_1_14(SH2_struct* sh)
{
  SH2ldcgbr(sh,8);
}
#define init_4_9_1_14
void FASTCALL call_4_9_1_14(SH2_struct* sh)
{
  SH2ldcgbr(sh,9);
}
#define init_4_10_1_14
void FASTCALL call_4_10_1_14(SH2_struct* sh)
{
  SH2ldcgbr(sh,10);
}
#define init_4_11_1_14
void FASTCALL call_4_11_1_14(SH2_struct* sh)
{
  SH2ldcgbr(sh,11);
}
#define init_4_12_1_14
void FASTCALL call_4_12_1_14(SH2_struct* sh)
{
  SH2ldcgbr(sh,12);
}
#define init_4_13_1_14
void FASTCALL call_4_13_1_14(SH2_struct* sh)
{
  SH2ldcgbr(sh,13);
}
#define init_4_14_1_14
void FASTCALL call_4_14_1_14(SH2_struct* sh)
{
  SH2ldcgbr(sh,14);
}
#define init_4_15_1_14
void FASTCALL call_4_15_1_14(SH2_struct* sh)
{
  SH2ldcgbr(sh,15);
}
#define init_4_0_0_11
void FASTCALL call_4_0_0_11(SH2_struct* sh)
{
  SH2jsr(sh,0);
}
#define init_4_1_0_11
void FASTCALL call_4_1_0_11(SH2_struct* sh)
{
  SH2jsr(sh,1);
}
#define init_4_2_0_11
void FASTCALL call_4_2_0_11(SH2_struct* sh)
{
  SH2jsr(sh,2);
}
#define init_4_3_0_11
void FASTCALL call_4_3_0_11(SH2_struct* sh)
{
  SH2jsr(sh,3);
}
#define init_4_4_0_11
void FASTCALL call_4_4_0_11(SH2_struct* sh)
{
  SH2jsr(sh,4);
}
#define init_4_5_0_11
void FASTCALL call_4_5_0_11(SH2_struct* sh)
{
  SH2jsr(sh,5);
}
#define init_4_6_0_11
void FASTCALL call_4_6_0_11(SH2_struct* sh)
{
  SH2jsr(sh,6);
}
#define init_4_7_0_11
void FASTCALL call_4_7_0_11(SH2_struct* sh)
{
  SH2jsr(sh,7);
}
#define init_4_8_0_11
void FASTCALL call_4_8_0_11(SH2_struct* sh)
{
  SH2jsr(sh,8);
}
#define init_4_9_0_11
void FASTCALL call_4_9_0_11(SH2_struct* sh)
{
  SH2jsr(sh,9);
}
#define init_4_10_0_11
void FASTCALL call_4_10_0_11(SH2_struct* sh)
{
  SH2jsr(sh,10);
}
#define init_4_11_0_11
void FASTCALL call_4_11_0_11(SH2_struct* sh)
{
  SH2jsr(sh,11);
}
#define init_4_12_0_11
void FASTCALL call_4_12_0_11(SH2_struct* sh)
{
  SH2jsr(sh,12);
}
#define init_4_13_0_11
void FASTCALL call_4_13_0_11(SH2_struct* sh)
{
  SH2jsr(sh,13);
}
#define init_4_14_0_11
void FASTCALL call_4_14_0_11(SH2_struct* sh)
{
  SH2jsr(sh,14);
}
#define init_4_15_0_11
void FASTCALL call_4_15_0_11(SH2_struct* sh)
{
  SH2jsr(sh,15);
}
#define init_4_0_2_11
void FASTCALL call_4_0_2_11(SH2_struct* sh)
{
  SH2jmp(sh,0);
}
#define init_4_1_2_11
void FASTCALL call_4_1_2_11(SH2_struct* sh)
{
  SH2jmp(sh,1);
}
#define init_4_2_2_11
void FASTCALL call_4_2_2_11(SH2_struct* sh)
{
  SH2jmp(sh,2);
}
#define init_4_3_2_11
void FASTCALL call_4_3_2_11(SH2_struct* sh)
{
  SH2jmp(sh,3);
}
#define init_4_4_2_11
void FASTCALL call_4_4_2_11(SH2_struct* sh)
{
  SH2jmp(sh,4);
}
#define init_4_5_2_11
void FASTCALL call_4_5_2_11(SH2_struct* sh)
{
  SH2jmp(sh,5);
}
#define init_4_6_2_11
void FASTCALL call_4_6_2_11(SH2_struct* sh)
{
  SH2jmp(sh,6);
}
#define init_4_7_2_11
void FASTCALL call_4_7_2_11(SH2_struct* sh)
{
  SH2jmp(sh,7);
}
#define init_4_8_2_11
void FASTCALL call_4_8_2_11(SH2_struct* sh)
{
  SH2jmp(sh,8);
}
#define init_4_9_2_11
void FASTCALL call_4_9_2_11(SH2_struct* sh)
{
  SH2jmp(sh,9);
}
#define init_4_10_2_11
void FASTCALL call_4_10_2_11(SH2_struct* sh)
{
  SH2jmp(sh,10);
}
#define init_4_11_2_11
void FASTCALL call_4_11_2_11(SH2_struct* sh)
{
  SH2jmp(sh,11);
}
#define init_4_12_2_11
void FASTCALL call_4_12_2_11(SH2_struct* sh)
{
  SH2jmp(sh,12);
}
#define init_4_13_2_11
void FASTCALL call_4_13_2_11(SH2_struct* sh)
{
  SH2jmp(sh,13);
}
#define init_4_14_2_11
void FASTCALL call_4_14_2_11(SH2_struct* sh)
{
  SH2jmp(sh,14);
}
#define init_4_15_2_11
void FASTCALL call_4_15_2_11(SH2_struct* sh)
{
  SH2jmp(sh,15);
}
#define init_6_0_0_13
void FASTCALL call_6_0_0_13(SH2_struct* sh)
{
  SH2extuw(sh,0,0);
}
#define init_6_0_1_13
void FASTCALL call_6_0_1_13(SH2_struct* sh)
{
  SH2extuw(sh,0,1);
}
#define init_6_0_2_13
void FASTCALL call_6_0_2_13(SH2_struct* sh)
{
  SH2extuw(sh,0,2);
}
#define init_6_0_3_13
void FASTCALL call_6_0_3_13(SH2_struct* sh)
{
  SH2extuw(sh,0,3);
}
#define init_6_0_4_13
void FASTCALL call_6_0_4_13(SH2_struct* sh)
{
  SH2extuw(sh,0,4);
}
#define init_6_0_5_13
void FASTCALL call_6_0_5_13(SH2_struct* sh)
{
  SH2extuw(sh,0,5);
}
#define init_6_0_6_13
void FASTCALL call_6_0_6_13(SH2_struct* sh)
{
  SH2extuw(sh,0,6);
}
#define init_6_0_7_13
void FASTCALL call_6_0_7_13(SH2_struct* sh)
{
  SH2extuw(sh,0,7);
}
#define init_6_0_8_13
void FASTCALL call_6_0_8_13(SH2_struct* sh)
{
  SH2extuw(sh,0,8);
}
#define init_6_0_9_13
void FASTCALL call_6_0_9_13(SH2_struct* sh)
{
  SH2extuw(sh,0,9);
}
#define init_6_0_10_13
void FASTCALL call_6_0_10_13(SH2_struct* sh)
{
  SH2extuw(sh,0,10);
}
#define init_6_0_11_13
void FASTCALL call_6_0_11_13(SH2_struct* sh)
{
  SH2extuw(sh,0,11);
}
#define init_6_0_12_13
void FASTCALL call_6_0_12_13(SH2_struct* sh)
{
  SH2extuw(sh,0,12);
}
#define init_6_0_13_13
void FASTCALL call_6_0_13_13(SH2_struct* sh)
{
  SH2extuw(sh,0,13);
}
#define init_6_0_14_13
void FASTCALL call_6_0_14_13(SH2_struct* sh)
{
  SH2extuw(sh,0,14);
}
#define init_6_0_15_13
void FASTCALL call_6_0_15_13(SH2_struct* sh)
{
  SH2extuw(sh,0,15);
}
#define init_6_1_0_13
void FASTCALL call_6_1_0_13(SH2_struct* sh)
{
  SH2extuw(sh,1,0);
}
#define init_6_1_1_13
void FASTCALL call_6_1_1_13(SH2_struct* sh)
{
  SH2extuw(sh,1,1);
}
#define init_6_1_2_13
void FASTCALL call_6_1_2_13(SH2_struct* sh)
{
  SH2extuw(sh,1,2);
}
#define init_6_1_3_13
void FASTCALL call_6_1_3_13(SH2_struct* sh)
{
  SH2extuw(sh,1,3);
}
#define init_6_1_4_13
void FASTCALL call_6_1_4_13(SH2_struct* sh)
{
  SH2extuw(sh,1,4);
}
#define init_6_1_5_13
void FASTCALL call_6_1_5_13(SH2_struct* sh)
{
  SH2extuw(sh,1,5);
}
#define init_6_1_6_13
void FASTCALL call_6_1_6_13(SH2_struct* sh)
{
  SH2extuw(sh,1,6);
}
#define init_6_1_7_13
void FASTCALL call_6_1_7_13(SH2_struct* sh)
{
  SH2extuw(sh,1,7);
}
#define init_6_1_8_13
void FASTCALL call_6_1_8_13(SH2_struct* sh)
{
  SH2extuw(sh,1,8);
}
#define init_6_1_9_13
void FASTCALL call_6_1_9_13(SH2_struct* sh)
{
  SH2extuw(sh,1,9);
}
#define init_6_1_10_13
void FASTCALL call_6_1_10_13(SH2_struct* sh)
{
  SH2extuw(sh,1,10);
}
#define init_6_1_11_13
void FASTCALL call_6_1_11_13(SH2_struct* sh)
{
  SH2extuw(sh,1,11);
}
#define init_6_1_12_13
void FASTCALL call_6_1_12_13(SH2_struct* sh)
{
  SH2extuw(sh,1,12);
}
#define init_6_1_13_13
void FASTCALL call_6_1_13_13(SH2_struct* sh)
{
  SH2extuw(sh,1,13);
}
#define init_6_1_14_13
void FASTCALL call_6_1_14_13(SH2_struct* sh)
{
  SH2extuw(sh,1,14);
}
#define init_6_1_15_13
void FASTCALL call_6_1_15_13(SH2_struct* sh)
{
  SH2extuw(sh,1,15);
}
#define init_6_2_0_13
void FASTCALL call_6_2_0_13(SH2_struct* sh)
{
  SH2extuw(sh,2,0);
}
#define init_6_2_1_13
void FASTCALL call_6_2_1_13(SH2_struct* sh)
{
  SH2extuw(sh,2,1);
}
#define init_6_2_2_13
void FASTCALL call_6_2_2_13(SH2_struct* sh)
{
  SH2extuw(sh,2,2);
}
#define init_6_2_3_13
void FASTCALL call_6_2_3_13(SH2_struct* sh)
{
  SH2extuw(sh,2,3);
}
#define init_6_2_4_13
void FASTCALL call_6_2_4_13(SH2_struct* sh)
{
  SH2extuw(sh,2,4);
}
#define init_6_2_5_13
void FASTCALL call_6_2_5_13(SH2_struct* sh)
{
  SH2extuw(sh,2,5);
}
#define init_6_2_6_13
void FASTCALL call_6_2_6_13(SH2_struct* sh)
{
  SH2extuw(sh,2,6);
}
#define init_6_2_7_13
void FASTCALL call_6_2_7_13(SH2_struct* sh)
{
  SH2extuw(sh,2,7);
}
#define init_6_2_8_13
void FASTCALL call_6_2_8_13(SH2_struct* sh)
{
  SH2extuw(sh,2,8);
}
#define init_6_2_9_13
void FASTCALL call_6_2_9_13(SH2_struct* sh)
{
  SH2extuw(sh,2,9);
}
#define init_6_2_10_13
void FASTCALL call_6_2_10_13(SH2_struct* sh)
{
  SH2extuw(sh,2,10);
}
#define init_6_2_11_13
void FASTCALL call_6_2_11_13(SH2_struct* sh)
{
  SH2extuw(sh,2,11);
}
#define init_6_2_12_13
void FASTCALL call_6_2_12_13(SH2_struct* sh)
{
  SH2extuw(sh,2,12);
}
#define init_6_2_13_13
void FASTCALL call_6_2_13_13(SH2_struct* sh)
{
  SH2extuw(sh,2,13);
}
#define init_6_2_14_13
void FASTCALL call_6_2_14_13(SH2_struct* sh)
{
  SH2extuw(sh,2,14);
}
#define init_6_2_15_13
void FASTCALL call_6_2_15_13(SH2_struct* sh)
{
  SH2extuw(sh,2,15);
}
#define init_6_3_0_13
void FASTCALL call_6_3_0_13(SH2_struct* sh)
{
  SH2extuw(sh,3,0);
}
#define init_6_3_1_13
void FASTCALL call_6_3_1_13(SH2_struct* sh)
{
  SH2extuw(sh,3,1);
}
#define init_6_3_2_13
void FASTCALL call_6_3_2_13(SH2_struct* sh)
{
  SH2extuw(sh,3,2);
}
#define init_6_3_3_13
void FASTCALL call_6_3_3_13(SH2_struct* sh)
{
  SH2extuw(sh,3,3);
}
#define init_6_3_4_13
void FASTCALL call_6_3_4_13(SH2_struct* sh)
{
  SH2extuw(sh,3,4);
}
#define init_6_3_5_13
void FASTCALL call_6_3_5_13(SH2_struct* sh)
{
  SH2extuw(sh,3,5);
}
#define init_6_3_6_13
void FASTCALL call_6_3_6_13(SH2_struct* sh)
{
  SH2extuw(sh,3,6);
}
#define init_6_3_7_13
void FASTCALL call_6_3_7_13(SH2_struct* sh)
{
  SH2extuw(sh,3,7);
}
#define init_6_3_8_13
void FASTCALL call_6_3_8_13(SH2_struct* sh)
{
  SH2extuw(sh,3,8);
}
#define init_6_3_9_13
void FASTCALL call_6_3_9_13(SH2_struct* sh)
{
  SH2extuw(sh,3,9);
}
#define init_6_3_10_13
void FASTCALL call_6_3_10_13(SH2_struct* sh)
{
  SH2extuw(sh,3,10);
}
#define init_6_3_11_13
void FASTCALL call_6_3_11_13(SH2_struct* sh)
{
  SH2extuw(sh,3,11);
}
#define init_6_3_12_13
void FASTCALL call_6_3_12_13(SH2_struct* sh)
{
  SH2extuw(sh,3,12);
}
#define init_6_3_13_13
void FASTCALL call_6_3_13_13(SH2_struct* sh)
{
  SH2extuw(sh,3,13);
}
#define init_6_3_14_13
void FASTCALL call_6_3_14_13(SH2_struct* sh)
{
  SH2extuw(sh,3,14);
}
#define init_6_3_15_13
void FASTCALL call_6_3_15_13(SH2_struct* sh)
{
  SH2extuw(sh,3,15);
}
#define init_6_4_0_13
void FASTCALL call_6_4_0_13(SH2_struct* sh)
{
  SH2extuw(sh,4,0);
}
#define init_6_4_1_13
void FASTCALL call_6_4_1_13(SH2_struct* sh)
{
  SH2extuw(sh,4,1);
}
#define init_6_4_2_13
void FASTCALL call_6_4_2_13(SH2_struct* sh)
{
  SH2extuw(sh,4,2);
}
#define init_6_4_3_13
void FASTCALL call_6_4_3_13(SH2_struct* sh)
{
  SH2extuw(sh,4,3);
}
#define init_6_4_4_13
void FASTCALL call_6_4_4_13(SH2_struct* sh)
{
  SH2extuw(sh,4,4);
}
#define init_6_4_5_13
void FASTCALL call_6_4_5_13(SH2_struct* sh)
{
  SH2extuw(sh,4,5);
}
#define init_6_4_6_13
void FASTCALL call_6_4_6_13(SH2_struct* sh)
{
  SH2extuw(sh,4,6);
}
#define init_6_4_7_13
void FASTCALL call_6_4_7_13(SH2_struct* sh)
{
  SH2extuw(sh,4,7);
}
#define init_6_4_8_13
void FASTCALL call_6_4_8_13(SH2_struct* sh)
{
  SH2extuw(sh,4,8);
}
#define init_6_4_9_13
void FASTCALL call_6_4_9_13(SH2_struct* sh)
{
  SH2extuw(sh,4,9);
}
#define init_6_4_10_13
void FASTCALL call_6_4_10_13(SH2_struct* sh)
{
  SH2extuw(sh,4,10);
}
#define init_6_4_11_13
void FASTCALL call_6_4_11_13(SH2_struct* sh)
{
  SH2extuw(sh,4,11);
}
#define init_6_4_12_13
void FASTCALL call_6_4_12_13(SH2_struct* sh)
{
  SH2extuw(sh,4,12);
}
#define init_6_4_13_13
void FASTCALL call_6_4_13_13(SH2_struct* sh)
{
  SH2extuw(sh,4,13);
}
#define init_6_4_14_13
void FASTCALL call_6_4_14_13(SH2_struct* sh)
{
  SH2extuw(sh,4,14);
}
#define init_6_4_15_13
void FASTCALL call_6_4_15_13(SH2_struct* sh)
{
  SH2extuw(sh,4,15);
}
#define init_6_5_0_13
void FASTCALL call_6_5_0_13(SH2_struct* sh)
{
  SH2extuw(sh,5,0);
}
#define init_6_5_1_13
void FASTCALL call_6_5_1_13(SH2_struct* sh)
{
  SH2extuw(sh,5,1);
}
#define init_6_5_2_13
void FASTCALL call_6_5_2_13(SH2_struct* sh)
{
  SH2extuw(sh,5,2);
}
#define init_6_5_3_13
void FASTCALL call_6_5_3_13(SH2_struct* sh)
{
  SH2extuw(sh,5,3);
}
#define init_6_5_4_13
void FASTCALL call_6_5_4_13(SH2_struct* sh)
{
  SH2extuw(sh,5,4);
}
#define init_6_5_5_13
void FASTCALL call_6_5_5_13(SH2_struct* sh)
{
  SH2extuw(sh,5,5);
}
#define init_6_5_6_13
void FASTCALL call_6_5_6_13(SH2_struct* sh)
{
  SH2extuw(sh,5,6);
}
#define init_6_5_7_13
void FASTCALL call_6_5_7_13(SH2_struct* sh)
{
  SH2extuw(sh,5,7);
}
#define init_6_5_8_13
void FASTCALL call_6_5_8_13(SH2_struct* sh)
{
  SH2extuw(sh,5,8);
}
#define init_6_5_9_13
void FASTCALL call_6_5_9_13(SH2_struct* sh)
{
  SH2extuw(sh,5,9);
}
#define init_6_5_10_13
void FASTCALL call_6_5_10_13(SH2_struct* sh)
{
  SH2extuw(sh,5,10);
}
#define init_6_5_11_13
void FASTCALL call_6_5_11_13(SH2_struct* sh)
{
  SH2extuw(sh,5,11);
}
#define init_6_5_12_13
void FASTCALL call_6_5_12_13(SH2_struct* sh)
{
  SH2extuw(sh,5,12);
}
#define init_6_5_13_13
void FASTCALL call_6_5_13_13(SH2_struct* sh)
{
  SH2extuw(sh,5,13);
}
#define init_6_5_14_13
void FASTCALL call_6_5_14_13(SH2_struct* sh)
{
  SH2extuw(sh,5,14);
}
#define init_6_5_15_13
void FASTCALL call_6_5_15_13(SH2_struct* sh)
{
  SH2extuw(sh,5,15);
}
#define init_6_6_0_13
void FASTCALL call_6_6_0_13(SH2_struct* sh)
{
  SH2extuw(sh,6,0);
}
#define init_6_6_1_13
void FASTCALL call_6_6_1_13(SH2_struct* sh)
{
  SH2extuw(sh,6,1);
}
#define init_6_6_2_13
void FASTCALL call_6_6_2_13(SH2_struct* sh)
{
  SH2extuw(sh,6,2);
}
#define init_6_6_3_13
void FASTCALL call_6_6_3_13(SH2_struct* sh)
{
  SH2extuw(sh,6,3);
}
#define init_6_6_4_13
void FASTCALL call_6_6_4_13(SH2_struct* sh)
{
  SH2extuw(sh,6,4);
}
#define init_6_6_5_13
void FASTCALL call_6_6_5_13(SH2_struct* sh)
{
  SH2extuw(sh,6,5);
}
#define init_6_6_6_13
void FASTCALL call_6_6_6_13(SH2_struct* sh)
{
  SH2extuw(sh,6,6);
}
#define init_6_6_7_13
void FASTCALL call_6_6_7_13(SH2_struct* sh)
{
  SH2extuw(sh,6,7);
}
#define init_6_6_8_13
void FASTCALL call_6_6_8_13(SH2_struct* sh)
{
  SH2extuw(sh,6,8);
}
#define init_6_6_9_13
void FASTCALL call_6_6_9_13(SH2_struct* sh)
{
  SH2extuw(sh,6,9);
}
#define init_6_6_10_13
void FASTCALL call_6_6_10_13(SH2_struct* sh)
{
  SH2extuw(sh,6,10);
}
#define init_6_6_11_13
void FASTCALL call_6_6_11_13(SH2_struct* sh)
{
  SH2extuw(sh,6,11);
}
#define init_6_6_12_13
void FASTCALL call_6_6_12_13(SH2_struct* sh)
{
  SH2extuw(sh,6,12);
}
#define init_6_6_13_13
void FASTCALL call_6_6_13_13(SH2_struct* sh)
{
  SH2extuw(sh,6,13);
}
#define init_6_6_14_13
void FASTCALL call_6_6_14_13(SH2_struct* sh)
{
  SH2extuw(sh,6,14);
}
#define init_6_6_15_13
void FASTCALL call_6_6_15_13(SH2_struct* sh)
{
  SH2extuw(sh,6,15);
}
#define init_6_7_0_13
void FASTCALL call_6_7_0_13(SH2_struct* sh)
{
  SH2extuw(sh,7,0);
}
#define init_6_7_1_13
void FASTCALL call_6_7_1_13(SH2_struct* sh)
{
  SH2extuw(sh,7,1);
}
#define init_6_7_2_13
void FASTCALL call_6_7_2_13(SH2_struct* sh)
{
  SH2extuw(sh,7,2);
}
#define init_6_7_3_13
void FASTCALL call_6_7_3_13(SH2_struct* sh)
{
  SH2extuw(sh,7,3);
}
#define init_6_7_4_13
void FASTCALL call_6_7_4_13(SH2_struct* sh)
{
  SH2extuw(sh,7,4);
}
#define init_6_7_5_13
void FASTCALL call_6_7_5_13(SH2_struct* sh)
{
  SH2extuw(sh,7,5);
}
#define init_6_7_6_13
void FASTCALL call_6_7_6_13(SH2_struct* sh)
{
  SH2extuw(sh,7,6);
}
#define init_6_7_7_13
void FASTCALL call_6_7_7_13(SH2_struct* sh)
{
  SH2extuw(sh,7,7);
}
#define init_6_7_8_13
void FASTCALL call_6_7_8_13(SH2_struct* sh)
{
  SH2extuw(sh,7,8);
}
#define init_6_7_9_13
void FASTCALL call_6_7_9_13(SH2_struct* sh)
{
  SH2extuw(sh,7,9);
}
#define init_6_7_10_13
void FASTCALL call_6_7_10_13(SH2_struct* sh)
{
  SH2extuw(sh,7,10);
}
#define init_6_7_11_13
void FASTCALL call_6_7_11_13(SH2_struct* sh)
{
  SH2extuw(sh,7,11);
}
#define init_6_7_12_13
void FASTCALL call_6_7_12_13(SH2_struct* sh)
{
  SH2extuw(sh,7,12);
}
#define init_6_7_13_13
void FASTCALL call_6_7_13_13(SH2_struct* sh)
{
  SH2extuw(sh,7,13);
}
#define init_6_7_14_13
void FASTCALL call_6_7_14_13(SH2_struct* sh)
{
  SH2extuw(sh,7,14);
}
#define init_6_7_15_13
void FASTCALL call_6_7_15_13(SH2_struct* sh)
{
  SH2extuw(sh,7,15);
}
#define init_6_8_0_13
void FASTCALL call_6_8_0_13(SH2_struct* sh)
{
  SH2extuw(sh,8,0);
}
#define init_6_8_1_13
void FASTCALL call_6_8_1_13(SH2_struct* sh)
{
  SH2extuw(sh,8,1);
}
#define init_6_8_2_13
void FASTCALL call_6_8_2_13(SH2_struct* sh)
{
  SH2extuw(sh,8,2);
}
#define init_6_8_3_13
void FASTCALL call_6_8_3_13(SH2_struct* sh)
{
  SH2extuw(sh,8,3);
}
#define init_6_8_4_13
void FASTCALL call_6_8_4_13(SH2_struct* sh)
{
  SH2extuw(sh,8,4);
}
#define init_6_8_5_13
void FASTCALL call_6_8_5_13(SH2_struct* sh)
{
  SH2extuw(sh,8,5);
}
#define init_6_8_6_13
void FASTCALL call_6_8_6_13(SH2_struct* sh)
{
  SH2extuw(sh,8,6);
}
#define init_6_8_7_13
void FASTCALL call_6_8_7_13(SH2_struct* sh)
{
  SH2extuw(sh,8,7);
}
#define init_6_8_8_13
void FASTCALL call_6_8_8_13(SH2_struct* sh)
{
  SH2extuw(sh,8,8);
}
#define init_6_8_9_13
void FASTCALL call_6_8_9_13(SH2_struct* sh)
{
  SH2extuw(sh,8,9);
}
#define init_6_8_10_13
void FASTCALL call_6_8_10_13(SH2_struct* sh)
{
  SH2extuw(sh,8,10);
}
#define init_6_8_11_13
void FASTCALL call_6_8_11_13(SH2_struct* sh)
{
  SH2extuw(sh,8,11);
}
#define init_6_8_12_13
void FASTCALL call_6_8_12_13(SH2_struct* sh)
{
  SH2extuw(sh,8,12);
}
#define init_6_8_13_13
void FASTCALL call_6_8_13_13(SH2_struct* sh)
{
  SH2extuw(sh,8,13);
}
#define init_6_8_14_13
void FASTCALL call_6_8_14_13(SH2_struct* sh)
{
  SH2extuw(sh,8,14);
}
#define init_6_8_15_13
void FASTCALL call_6_8_15_13(SH2_struct* sh)
{
  SH2extuw(sh,8,15);
}
#define init_6_9_0_13
void FASTCALL call_6_9_0_13(SH2_struct* sh)
{
  SH2extuw(sh,9,0);
}
#define init_6_9_1_13
void FASTCALL call_6_9_1_13(SH2_struct* sh)
{
  SH2extuw(sh,9,1);
}
#define init_6_9_2_13
void FASTCALL call_6_9_2_13(SH2_struct* sh)
{
  SH2extuw(sh,9,2);
}
#define init_6_9_3_13
void FASTCALL call_6_9_3_13(SH2_struct* sh)
{
  SH2extuw(sh,9,3);
}
#define init_6_9_4_13
void FASTCALL call_6_9_4_13(SH2_struct* sh)
{
  SH2extuw(sh,9,4);
}
#define init_6_9_5_13
void FASTCALL call_6_9_5_13(SH2_struct* sh)
{
  SH2extuw(sh,9,5);
}
#define init_6_9_6_13
void FASTCALL call_6_9_6_13(SH2_struct* sh)
{
  SH2extuw(sh,9,6);
}
#define init_6_9_7_13
void FASTCALL call_6_9_7_13(SH2_struct* sh)
{
  SH2extuw(sh,9,7);
}
#define init_6_9_8_13
void FASTCALL call_6_9_8_13(SH2_struct* sh)
{
  SH2extuw(sh,9,8);
}
#define init_6_9_9_13
void FASTCALL call_6_9_9_13(SH2_struct* sh)
{
  SH2extuw(sh,9,9);
}
#define init_6_9_10_13
void FASTCALL call_6_9_10_13(SH2_struct* sh)
{
  SH2extuw(sh,9,10);
}
#define init_6_9_11_13
void FASTCALL call_6_9_11_13(SH2_struct* sh)
{
  SH2extuw(sh,9,11);
}
#define init_6_9_12_13
void FASTCALL call_6_9_12_13(SH2_struct* sh)
{
  SH2extuw(sh,9,12);
}
#define init_6_9_13_13
void FASTCALL call_6_9_13_13(SH2_struct* sh)
{
  SH2extuw(sh,9,13);
}
#define init_6_9_14_13
void FASTCALL call_6_9_14_13(SH2_struct* sh)
{
  SH2extuw(sh,9,14);
}
#define init_6_9_15_13
void FASTCALL call_6_9_15_13(SH2_struct* sh)
{
  SH2extuw(sh,9,15);
}
#define init_6_10_0_13
void FASTCALL call_6_10_0_13(SH2_struct* sh)
{
  SH2extuw(sh,10,0);
}
#define init_6_10_1_13
void FASTCALL call_6_10_1_13(SH2_struct* sh)
{
  SH2extuw(sh,10,1);
}
#define init_6_10_2_13
void FASTCALL call_6_10_2_13(SH2_struct* sh)
{
  SH2extuw(sh,10,2);
}
#define init_6_10_3_13
void FASTCALL call_6_10_3_13(SH2_struct* sh)
{
  SH2extuw(sh,10,3);
}
#define init_6_10_4_13
void FASTCALL call_6_10_4_13(SH2_struct* sh)
{
  SH2extuw(sh,10,4);
}
#define init_6_10_5_13
void FASTCALL call_6_10_5_13(SH2_struct* sh)
{
  SH2extuw(sh,10,5);
}
#define init_6_10_6_13
void FASTCALL call_6_10_6_13(SH2_struct* sh)
{
  SH2extuw(sh,10,6);
}
#define init_6_10_7_13
void FASTCALL call_6_10_7_13(SH2_struct* sh)
{
  SH2extuw(sh,10,7);
}
#define init_6_10_8_13
void FASTCALL call_6_10_8_13(SH2_struct* sh)
{
  SH2extuw(sh,10,8);
}
#define init_6_10_9_13
void FASTCALL call_6_10_9_13(SH2_struct* sh)
{
  SH2extuw(sh,10,9);
}
#define init_6_10_10_13
void FASTCALL call_6_10_10_13(SH2_struct* sh)
{
  SH2extuw(sh,10,10);
}
#define init_6_10_11_13
void FASTCALL call_6_10_11_13(SH2_struct* sh)
{
  SH2extuw(sh,10,11);
}
#define init_6_10_12_13
void FASTCALL call_6_10_12_13(SH2_struct* sh)
{
  SH2extuw(sh,10,12);
}
#define init_6_10_13_13
void FASTCALL call_6_10_13_13(SH2_struct* sh)
{
  SH2extuw(sh,10,13);
}
#define init_6_10_14_13
void FASTCALL call_6_10_14_13(SH2_struct* sh)
{
  SH2extuw(sh,10,14);
}
#define init_6_10_15_13
void FASTCALL call_6_10_15_13(SH2_struct* sh)
{
  SH2extuw(sh,10,15);
}
#define init_6_11_0_13
void FASTCALL call_6_11_0_13(SH2_struct* sh)
{
  SH2extuw(sh,11,0);
}
#define init_6_11_1_13
void FASTCALL call_6_11_1_13(SH2_struct* sh)
{
  SH2extuw(sh,11,1);
}
#define init_6_11_2_13
void FASTCALL call_6_11_2_13(SH2_struct* sh)
{
  SH2extuw(sh,11,2);
}
#define init_6_11_3_13
void FASTCALL call_6_11_3_13(SH2_struct* sh)
{
  SH2extuw(sh,11,3);
}
#define init_6_11_4_13
void FASTCALL call_6_11_4_13(SH2_struct* sh)
{
  SH2extuw(sh,11,4);
}
#define init_6_11_5_13
void FASTCALL call_6_11_5_13(SH2_struct* sh)
{
  SH2extuw(sh,11,5);
}
#define init_6_11_6_13
void FASTCALL call_6_11_6_13(SH2_struct* sh)
{
  SH2extuw(sh,11,6);
}
#define init_6_11_7_13
void FASTCALL call_6_11_7_13(SH2_struct* sh)
{
  SH2extuw(sh,11,7);
}
#define init_6_11_8_13
void FASTCALL call_6_11_8_13(SH2_struct* sh)
{
  SH2extuw(sh,11,8);
}
#define init_6_11_9_13
void FASTCALL call_6_11_9_13(SH2_struct* sh)
{
  SH2extuw(sh,11,9);
}
#define init_6_11_10_13
void FASTCALL call_6_11_10_13(SH2_struct* sh)
{
  SH2extuw(sh,11,10);
}
#define init_6_11_11_13
void FASTCALL call_6_11_11_13(SH2_struct* sh)
{
  SH2extuw(sh,11,11);
}
#define init_6_11_12_13
void FASTCALL call_6_11_12_13(SH2_struct* sh)
{
  SH2extuw(sh,11,12);
}
#define init_6_11_13_13
void FASTCALL call_6_11_13_13(SH2_struct* sh)
{
  SH2extuw(sh,11,13);
}
#define init_6_11_14_13
void FASTCALL call_6_11_14_13(SH2_struct* sh)
{
  SH2extuw(sh,11,14);
}
#define init_6_11_15_13
void FASTCALL call_6_11_15_13(SH2_struct* sh)
{
  SH2extuw(sh,11,15);
}
#define init_6_12_0_13
void FASTCALL call_6_12_0_13(SH2_struct* sh)
{
  SH2extuw(sh,12,0);
}
#define init_6_12_1_13
void FASTCALL call_6_12_1_13(SH2_struct* sh)
{
  SH2extuw(sh,12,1);
}
#define init_6_12_2_13
void FASTCALL call_6_12_2_13(SH2_struct* sh)
{
  SH2extuw(sh,12,2);
}
#define init_6_12_3_13
void FASTCALL call_6_12_3_13(SH2_struct* sh)
{
  SH2extuw(sh,12,3);
}
#define init_6_12_4_13
void FASTCALL call_6_12_4_13(SH2_struct* sh)
{
  SH2extuw(sh,12,4);
}
#define init_6_12_5_13
void FASTCALL call_6_12_5_13(SH2_struct* sh)
{
  SH2extuw(sh,12,5);
}
#define init_6_12_6_13
void FASTCALL call_6_12_6_13(SH2_struct* sh)
{
  SH2extuw(sh,12,6);
}
#define init_6_12_7_13
void FASTCALL call_6_12_7_13(SH2_struct* sh)
{
  SH2extuw(sh,12,7);
}
#define init_6_12_8_13
void FASTCALL call_6_12_8_13(SH2_struct* sh)
{
  SH2extuw(sh,12,8);
}
#define init_6_12_9_13
void FASTCALL call_6_12_9_13(SH2_struct* sh)
{
  SH2extuw(sh,12,9);
}
#define init_6_12_10_13
void FASTCALL call_6_12_10_13(SH2_struct* sh)
{
  SH2extuw(sh,12,10);
}
#define init_6_12_11_13
void FASTCALL call_6_12_11_13(SH2_struct* sh)
{
  SH2extuw(sh,12,11);
}
#define init_6_12_12_13
void FASTCALL call_6_12_12_13(SH2_struct* sh)
{
  SH2extuw(sh,12,12);
}
#define init_6_12_13_13
void FASTCALL call_6_12_13_13(SH2_struct* sh)
{
  SH2extuw(sh,12,13);
}
#define init_6_12_14_13
void FASTCALL call_6_12_14_13(SH2_struct* sh)
{
  SH2extuw(sh,12,14);
}
#define init_6_12_15_13
void FASTCALL call_6_12_15_13(SH2_struct* sh)
{
  SH2extuw(sh,12,15);
}
#define init_6_13_0_13
void FASTCALL call_6_13_0_13(SH2_struct* sh)
{
  SH2extuw(sh,13,0);
}
#define init_6_13_1_13
void FASTCALL call_6_13_1_13(SH2_struct* sh)
{
  SH2extuw(sh,13,1);
}
#define init_6_13_2_13
void FASTCALL call_6_13_2_13(SH2_struct* sh)
{
  SH2extuw(sh,13,2);
}
#define init_6_13_3_13
void FASTCALL call_6_13_3_13(SH2_struct* sh)
{
  SH2extuw(sh,13,3);
}
#define init_6_13_4_13
void FASTCALL call_6_13_4_13(SH2_struct* sh)
{
  SH2extuw(sh,13,4);
}
#define init_6_13_5_13
void FASTCALL call_6_13_5_13(SH2_struct* sh)
{
  SH2extuw(sh,13,5);
}
#define init_6_13_6_13
void FASTCALL call_6_13_6_13(SH2_struct* sh)
{
  SH2extuw(sh,13,6);
}
#define init_6_13_7_13
void FASTCALL call_6_13_7_13(SH2_struct* sh)
{
  SH2extuw(sh,13,7);
}
#define init_6_13_8_13
void FASTCALL call_6_13_8_13(SH2_struct* sh)
{
  SH2extuw(sh,13,8);
}
#define init_6_13_9_13
void FASTCALL call_6_13_9_13(SH2_struct* sh)
{
  SH2extuw(sh,13,9);
}
#define init_6_13_10_13
void FASTCALL call_6_13_10_13(SH2_struct* sh)
{
  SH2extuw(sh,13,10);
}
#define init_6_13_11_13
void FASTCALL call_6_13_11_13(SH2_struct* sh)
{
  SH2extuw(sh,13,11);
}
#define init_6_13_12_13
void FASTCALL call_6_13_12_13(SH2_struct* sh)
{
  SH2extuw(sh,13,12);
}
#define init_6_13_13_13
void FASTCALL call_6_13_13_13(SH2_struct* sh)
{
  SH2extuw(sh,13,13);
}
#define init_6_13_14_13
void FASTCALL call_6_13_14_13(SH2_struct* sh)
{
  SH2extuw(sh,13,14);
}
#define init_6_13_15_13
void FASTCALL call_6_13_15_13(SH2_struct* sh)
{
  SH2extuw(sh,13,15);
}
#define init_6_14_0_13
void FASTCALL call_6_14_0_13(SH2_struct* sh)
{
  SH2extuw(sh,14,0);
}
#define init_6_14_1_13
void FASTCALL call_6_14_1_13(SH2_struct* sh)
{
  SH2extuw(sh,14,1);
}
#define init_6_14_2_13
void FASTCALL call_6_14_2_13(SH2_struct* sh)
{
  SH2extuw(sh,14,2);
}
#define init_6_14_3_13
void FASTCALL call_6_14_3_13(SH2_struct* sh)
{
  SH2extuw(sh,14,3);
}
#define init_6_14_4_13
void FASTCALL call_6_14_4_13(SH2_struct* sh)
{
  SH2extuw(sh,14,4);
}
#define init_6_14_5_13
void FASTCALL call_6_14_5_13(SH2_struct* sh)
{
  SH2extuw(sh,14,5);
}
#define init_6_14_6_13
void FASTCALL call_6_14_6_13(SH2_struct* sh)
{
  SH2extuw(sh,14,6);
}
#define init_6_14_7_13
void FASTCALL call_6_14_7_13(SH2_struct* sh)
{
  SH2extuw(sh,14,7);
}
#define init_6_14_8_13
void FASTCALL call_6_14_8_13(SH2_struct* sh)
{
  SH2extuw(sh,14,8);
}
#define init_6_14_9_13
void FASTCALL call_6_14_9_13(SH2_struct* sh)
{
  SH2extuw(sh,14,9);
}
#define init_6_14_10_13
void FASTCALL call_6_14_10_13(SH2_struct* sh)
{
  SH2extuw(sh,14,10);
}
#define init_6_14_11_13
void FASTCALL call_6_14_11_13(SH2_struct* sh)
{
  SH2extuw(sh,14,11);
}
#define init_6_14_12_13
void FASTCALL call_6_14_12_13(SH2_struct* sh)
{
  SH2extuw(sh,14,12);
}
#define init_6_14_13_13
void FASTCALL call_6_14_13_13(SH2_struct* sh)
{
  SH2extuw(sh,14,13);
}
#define init_6_14_14_13
void FASTCALL call_6_14_14_13(SH2_struct* sh)
{
  SH2extuw(sh,14,14);
}
#define init_6_14_15_13
void FASTCALL call_6_14_15_13(SH2_struct* sh)
{
  SH2extuw(sh,14,15);
}
#define init_6_15_0_13
void FASTCALL call_6_15_0_13(SH2_struct* sh)
{
  SH2extuw(sh,15,0);
}
#define init_6_15_1_13
void FASTCALL call_6_15_1_13(SH2_struct* sh)
{
  SH2extuw(sh,15,1);
}
#define init_6_15_2_13
void FASTCALL call_6_15_2_13(SH2_struct* sh)
{
  SH2extuw(sh,15,2);
}
#define init_6_15_3_13
void FASTCALL call_6_15_3_13(SH2_struct* sh)
{
  SH2extuw(sh,15,3);
}
#define init_6_15_4_13
void FASTCALL call_6_15_4_13(SH2_struct* sh)
{
  SH2extuw(sh,15,4);
}
#define init_6_15_5_13
void FASTCALL call_6_15_5_13(SH2_struct* sh)
{
  SH2extuw(sh,15,5);
}
#define init_6_15_6_13
void FASTCALL call_6_15_6_13(SH2_struct* sh)
{
  SH2extuw(sh,15,6);
}
#define init_6_15_7_13
void FASTCALL call_6_15_7_13(SH2_struct* sh)
{
  SH2extuw(sh,15,7);
}
#define init_6_15_8_13
void FASTCALL call_6_15_8_13(SH2_struct* sh)
{
  SH2extuw(sh,15,8);
}
#define init_6_15_9_13
void FASTCALL call_6_15_9_13(SH2_struct* sh)
{
  SH2extuw(sh,15,9);
}
#define init_6_15_10_13
void FASTCALL call_6_15_10_13(SH2_struct* sh)
{
  SH2extuw(sh,15,10);
}
#define init_6_15_11_13
void FASTCALL call_6_15_11_13(SH2_struct* sh)
{
  SH2extuw(sh,15,11);
}
#define init_6_15_12_13
void FASTCALL call_6_15_12_13(SH2_struct* sh)
{
  SH2extuw(sh,15,12);
}
#define init_6_15_13_13
void FASTCALL call_6_15_13_13(SH2_struct* sh)
{
  SH2extuw(sh,15,13);
}
#define init_6_15_14_13
void FASTCALL call_6_15_14_13(SH2_struct* sh)
{
  SH2extuw(sh,15,14);
}
#define init_6_15_15_13
void FASTCALL call_6_15_15_13(SH2_struct* sh)
{
  SH2extuw(sh,15,15);
}
#define init_6_0_0_12
void FASTCALL call_6_0_0_12(SH2_struct* sh)
{
  SH2extub(sh,0,0);
}
#define init_6_0_1_12
void FASTCALL call_6_0_1_12(SH2_struct* sh)
{
  SH2extub(sh,0,1);
}
#define init_6_0_2_12
void FASTCALL call_6_0_2_12(SH2_struct* sh)
{
  SH2extub(sh,0,2);
}
#define init_6_0_3_12
void FASTCALL call_6_0_3_12(SH2_struct* sh)
{
  SH2extub(sh,0,3);
}
#define init_6_0_4_12
void FASTCALL call_6_0_4_12(SH2_struct* sh)
{
  SH2extub(sh,0,4);
}
#define init_6_0_5_12
void FASTCALL call_6_0_5_12(SH2_struct* sh)
{
  SH2extub(sh,0,5);
}
#define init_6_0_6_12
void FASTCALL call_6_0_6_12(SH2_struct* sh)
{
  SH2extub(sh,0,6);
}
#define init_6_0_7_12
void FASTCALL call_6_0_7_12(SH2_struct* sh)
{
  SH2extub(sh,0,7);
}
#define init_6_0_8_12
void FASTCALL call_6_0_8_12(SH2_struct* sh)
{
  SH2extub(sh,0,8);
}
#define init_6_0_9_12
void FASTCALL call_6_0_9_12(SH2_struct* sh)
{
  SH2extub(sh,0,9);
}
#define init_6_0_10_12
void FASTCALL call_6_0_10_12(SH2_struct* sh)
{
  SH2extub(sh,0,10);
}
#define init_6_0_11_12
void FASTCALL call_6_0_11_12(SH2_struct* sh)
{
  SH2extub(sh,0,11);
}
#define init_6_0_12_12
void FASTCALL call_6_0_12_12(SH2_struct* sh)
{
  SH2extub(sh,0,12);
}
#define init_6_0_13_12
void FASTCALL call_6_0_13_12(SH2_struct* sh)
{
  SH2extub(sh,0,13);
}
#define init_6_0_14_12
void FASTCALL call_6_0_14_12(SH2_struct* sh)
{
  SH2extub(sh,0,14);
}
#define init_6_0_15_12
void FASTCALL call_6_0_15_12(SH2_struct* sh)
{
  SH2extub(sh,0,15);
}
#define init_6_1_0_12
void FASTCALL call_6_1_0_12(SH2_struct* sh)
{
  SH2extub(sh,1,0);
}
#define init_6_1_1_12
void FASTCALL call_6_1_1_12(SH2_struct* sh)
{
  SH2extub(sh,1,1);
}
#define init_6_1_2_12
void FASTCALL call_6_1_2_12(SH2_struct* sh)
{
  SH2extub(sh,1,2);
}
#define init_6_1_3_12
void FASTCALL call_6_1_3_12(SH2_struct* sh)
{
  SH2extub(sh,1,3);
}
#define init_6_1_4_12
void FASTCALL call_6_1_4_12(SH2_struct* sh)
{
  SH2extub(sh,1,4);
}
#define init_6_1_5_12
void FASTCALL call_6_1_5_12(SH2_struct* sh)
{
  SH2extub(sh,1,5);
}
#define init_6_1_6_12
void FASTCALL call_6_1_6_12(SH2_struct* sh)
{
  SH2extub(sh,1,6);
}
#define init_6_1_7_12
void FASTCALL call_6_1_7_12(SH2_struct* sh)
{
  SH2extub(sh,1,7);
}
#define init_6_1_8_12
void FASTCALL call_6_1_8_12(SH2_struct* sh)
{
  SH2extub(sh,1,8);
}
#define init_6_1_9_12
void FASTCALL call_6_1_9_12(SH2_struct* sh)
{
  SH2extub(sh,1,9);
}
#define init_6_1_10_12
void FASTCALL call_6_1_10_12(SH2_struct* sh)
{
  SH2extub(sh,1,10);
}
#define init_6_1_11_12
void FASTCALL call_6_1_11_12(SH2_struct* sh)
{
  SH2extub(sh,1,11);
}
#define init_6_1_12_12
void FASTCALL call_6_1_12_12(SH2_struct* sh)
{
  SH2extub(sh,1,12);
}
#define init_6_1_13_12
void FASTCALL call_6_1_13_12(SH2_struct* sh)
{
  SH2extub(sh,1,13);
}
#define init_6_1_14_12
void FASTCALL call_6_1_14_12(SH2_struct* sh)
{
  SH2extub(sh,1,14);
}
#define init_6_1_15_12
void FASTCALL call_6_1_15_12(SH2_struct* sh)
{
  SH2extub(sh,1,15);
}
#define init_6_2_0_12
void FASTCALL call_6_2_0_12(SH2_struct* sh)
{
  SH2extub(sh,2,0);
}
#define init_6_2_1_12
void FASTCALL call_6_2_1_12(SH2_struct* sh)
{
  SH2extub(sh,2,1);
}
#define init_6_2_2_12
void FASTCALL call_6_2_2_12(SH2_struct* sh)
{
  SH2extub(sh,2,2);
}
#define init_6_2_3_12
void FASTCALL call_6_2_3_12(SH2_struct* sh)
{
  SH2extub(sh,2,3);
}
#define init_6_2_4_12
void FASTCALL call_6_2_4_12(SH2_struct* sh)
{
  SH2extub(sh,2,4);
}
#define init_6_2_5_12
void FASTCALL call_6_2_5_12(SH2_struct* sh)
{
  SH2extub(sh,2,5);
}
#define init_6_2_6_12
void FASTCALL call_6_2_6_12(SH2_struct* sh)
{
  SH2extub(sh,2,6);
}
#define init_6_2_7_12
void FASTCALL call_6_2_7_12(SH2_struct* sh)
{
  SH2extub(sh,2,7);
}
#define init_6_2_8_12
void FASTCALL call_6_2_8_12(SH2_struct* sh)
{
  SH2extub(sh,2,8);
}
#define init_6_2_9_12
void FASTCALL call_6_2_9_12(SH2_struct* sh)
{
  SH2extub(sh,2,9);
}
#define init_6_2_10_12
void FASTCALL call_6_2_10_12(SH2_struct* sh)
{
  SH2extub(sh,2,10);
}
#define init_6_2_11_12
void FASTCALL call_6_2_11_12(SH2_struct* sh)
{
  SH2extub(sh,2,11);
}
#define init_6_2_12_12
void FASTCALL call_6_2_12_12(SH2_struct* sh)
{
  SH2extub(sh,2,12);
}
#define init_6_2_13_12
void FASTCALL call_6_2_13_12(SH2_struct* sh)
{
  SH2extub(sh,2,13);
}
#define init_6_2_14_12
void FASTCALL call_6_2_14_12(SH2_struct* sh)
{
  SH2extub(sh,2,14);
}
#define init_6_2_15_12
void FASTCALL call_6_2_15_12(SH2_struct* sh)
{
  SH2extub(sh,2,15);
}
#define init_6_3_0_12
void FASTCALL call_6_3_0_12(SH2_struct* sh)
{
  SH2extub(sh,3,0);
}
#define init_6_3_1_12
void FASTCALL call_6_3_1_12(SH2_struct* sh)
{
  SH2extub(sh,3,1);
}
#define init_6_3_2_12
void FASTCALL call_6_3_2_12(SH2_struct* sh)
{
  SH2extub(sh,3,2);
}
#define init_6_3_3_12
void FASTCALL call_6_3_3_12(SH2_struct* sh)
{
  SH2extub(sh,3,3);
}
#define init_6_3_4_12
void FASTCALL call_6_3_4_12(SH2_struct* sh)
{
  SH2extub(sh,3,4);
}
#define init_6_3_5_12
void FASTCALL call_6_3_5_12(SH2_struct* sh)
{
  SH2extub(sh,3,5);
}
#define init_6_3_6_12
void FASTCALL call_6_3_6_12(SH2_struct* sh)
{
  SH2extub(sh,3,6);
}
#define init_6_3_7_12
void FASTCALL call_6_3_7_12(SH2_struct* sh)
{
  SH2extub(sh,3,7);
}
#define init_6_3_8_12
void FASTCALL call_6_3_8_12(SH2_struct* sh)
{
  SH2extub(sh,3,8);
}
#define init_6_3_9_12
void FASTCALL call_6_3_9_12(SH2_struct* sh)
{
  SH2extub(sh,3,9);
}
#define init_6_3_10_12
void FASTCALL call_6_3_10_12(SH2_struct* sh)
{
  SH2extub(sh,3,10);
}
#define init_6_3_11_12
void FASTCALL call_6_3_11_12(SH2_struct* sh)
{
  SH2extub(sh,3,11);
}
#define init_6_3_12_12
void FASTCALL call_6_3_12_12(SH2_struct* sh)
{
  SH2extub(sh,3,12);
}
#define init_6_3_13_12
void FASTCALL call_6_3_13_12(SH2_struct* sh)
{
  SH2extub(sh,3,13);
}
#define init_6_3_14_12
void FASTCALL call_6_3_14_12(SH2_struct* sh)
{
  SH2extub(sh,3,14);
}
#define init_6_3_15_12
void FASTCALL call_6_3_15_12(SH2_struct* sh)
{
  SH2extub(sh,3,15);
}
#define init_6_4_0_12
void FASTCALL call_6_4_0_12(SH2_struct* sh)
{
  SH2extub(sh,4,0);
}
#define init_6_4_1_12
void FASTCALL call_6_4_1_12(SH2_struct* sh)
{
  SH2extub(sh,4,1);
}
#define init_6_4_2_12
void FASTCALL call_6_4_2_12(SH2_struct* sh)
{
  SH2extub(sh,4,2);
}
#define init_6_4_3_12
void FASTCALL call_6_4_3_12(SH2_struct* sh)
{
  SH2extub(sh,4,3);
}
#define init_6_4_4_12
void FASTCALL call_6_4_4_12(SH2_struct* sh)
{
  SH2extub(sh,4,4);
}
#define init_6_4_5_12
void FASTCALL call_6_4_5_12(SH2_struct* sh)
{
  SH2extub(sh,4,5);
}
#define init_6_4_6_12
void FASTCALL call_6_4_6_12(SH2_struct* sh)
{
  SH2extub(sh,4,6);
}
#define init_6_4_7_12
void FASTCALL call_6_4_7_12(SH2_struct* sh)
{
  SH2extub(sh,4,7);
}
#define init_6_4_8_12
void FASTCALL call_6_4_8_12(SH2_struct* sh)
{
  SH2extub(sh,4,8);
}
#define init_6_4_9_12
void FASTCALL call_6_4_9_12(SH2_struct* sh)
{
  SH2extub(sh,4,9);
}
#define init_6_4_10_12
void FASTCALL call_6_4_10_12(SH2_struct* sh)
{
  SH2extub(sh,4,10);
}
#define init_6_4_11_12
void FASTCALL call_6_4_11_12(SH2_struct* sh)
{
  SH2extub(sh,4,11);
}
#define init_6_4_12_12
void FASTCALL call_6_4_12_12(SH2_struct* sh)
{
  SH2extub(sh,4,12);
}
#define init_6_4_13_12
void FASTCALL call_6_4_13_12(SH2_struct* sh)
{
  SH2extub(sh,4,13);
}
#define init_6_4_14_12
void FASTCALL call_6_4_14_12(SH2_struct* sh)
{
  SH2extub(sh,4,14);
}
#define init_6_4_15_12
void FASTCALL call_6_4_15_12(SH2_struct* sh)
{
  SH2extub(sh,4,15);
}
#define init_6_5_0_12
void FASTCALL call_6_5_0_12(SH2_struct* sh)
{
  SH2extub(sh,5,0);
}
#define init_6_5_1_12
void FASTCALL call_6_5_1_12(SH2_struct* sh)
{
  SH2extub(sh,5,1);
}
#define init_6_5_2_12
void FASTCALL call_6_5_2_12(SH2_struct* sh)
{
  SH2extub(sh,5,2);
}
#define init_6_5_3_12
void FASTCALL call_6_5_3_12(SH2_struct* sh)
{
  SH2extub(sh,5,3);
}
#define init_6_5_4_12
void FASTCALL call_6_5_4_12(SH2_struct* sh)
{
  SH2extub(sh,5,4);
}
#define init_6_5_5_12
void FASTCALL call_6_5_5_12(SH2_struct* sh)
{
  SH2extub(sh,5,5);
}
#define init_6_5_6_12
void FASTCALL call_6_5_6_12(SH2_struct* sh)
{
  SH2extub(sh,5,6);
}
#define init_6_5_7_12
void FASTCALL call_6_5_7_12(SH2_struct* sh)
{
  SH2extub(sh,5,7);
}
#define init_6_5_8_12
void FASTCALL call_6_5_8_12(SH2_struct* sh)
{
  SH2extub(sh,5,8);
}
#define init_6_5_9_12
void FASTCALL call_6_5_9_12(SH2_struct* sh)
{
  SH2extub(sh,5,9);
}
#define init_6_5_10_12
void FASTCALL call_6_5_10_12(SH2_struct* sh)
{
  SH2extub(sh,5,10);
}
#define init_6_5_11_12
void FASTCALL call_6_5_11_12(SH2_struct* sh)
{
  SH2extub(sh,5,11);
}
#define init_6_5_12_12
void FASTCALL call_6_5_12_12(SH2_struct* sh)
{
  SH2extub(sh,5,12);
}
#define init_6_5_13_12
void FASTCALL call_6_5_13_12(SH2_struct* sh)
{
  SH2extub(sh,5,13);
}
#define init_6_5_14_12
void FASTCALL call_6_5_14_12(SH2_struct* sh)
{
  SH2extub(sh,5,14);
}
#define init_6_5_15_12
void FASTCALL call_6_5_15_12(SH2_struct* sh)
{
  SH2extub(sh,5,15);
}
#define init_6_6_0_12
void FASTCALL call_6_6_0_12(SH2_struct* sh)
{
  SH2extub(sh,6,0);
}
#define init_6_6_1_12
void FASTCALL call_6_6_1_12(SH2_struct* sh)
{
  SH2extub(sh,6,1);
}
#define init_6_6_2_12
void FASTCALL call_6_6_2_12(SH2_struct* sh)
{
  SH2extub(sh,6,2);
}
#define init_6_6_3_12
void FASTCALL call_6_6_3_12(SH2_struct* sh)
{
  SH2extub(sh,6,3);
}
#define init_6_6_4_12
void FASTCALL call_6_6_4_12(SH2_struct* sh)
{
  SH2extub(sh,6,4);
}
#define init_6_6_5_12
void FASTCALL call_6_6_5_12(SH2_struct* sh)
{
  SH2extub(sh,6,5);
}
#define init_6_6_6_12
void FASTCALL call_6_6_6_12(SH2_struct* sh)
{
  SH2extub(sh,6,6);
}
#define init_6_6_7_12
void FASTCALL call_6_6_7_12(SH2_struct* sh)
{
  SH2extub(sh,6,7);
}
#define init_6_6_8_12
void FASTCALL call_6_6_8_12(SH2_struct* sh)
{
  SH2extub(sh,6,8);
}
#define init_6_6_9_12
void FASTCALL call_6_6_9_12(SH2_struct* sh)
{
  SH2extub(sh,6,9);
}
#define init_6_6_10_12
void FASTCALL call_6_6_10_12(SH2_struct* sh)
{
  SH2extub(sh,6,10);
}
#define init_6_6_11_12
void FASTCALL call_6_6_11_12(SH2_struct* sh)
{
  SH2extub(sh,6,11);
}
#define init_6_6_12_12
void FASTCALL call_6_6_12_12(SH2_struct* sh)
{
  SH2extub(sh,6,12);
}
#define init_6_6_13_12
void FASTCALL call_6_6_13_12(SH2_struct* sh)
{
  SH2extub(sh,6,13);
}
#define init_6_6_14_12
void FASTCALL call_6_6_14_12(SH2_struct* sh)
{
  SH2extub(sh,6,14);
}
#define init_6_6_15_12
void FASTCALL call_6_6_15_12(SH2_struct* sh)
{
  SH2extub(sh,6,15);
}
#define init_6_7_0_12
void FASTCALL call_6_7_0_12(SH2_struct* sh)
{
  SH2extub(sh,7,0);
}
#define init_6_7_1_12
void FASTCALL call_6_7_1_12(SH2_struct* sh)
{
  SH2extub(sh,7,1);
}
#define init_6_7_2_12
void FASTCALL call_6_7_2_12(SH2_struct* sh)
{
  SH2extub(sh,7,2);
}
#define init_6_7_3_12
void FASTCALL call_6_7_3_12(SH2_struct* sh)
{
  SH2extub(sh,7,3);
}
#define init_6_7_4_12
void FASTCALL call_6_7_4_12(SH2_struct* sh)
{
  SH2extub(sh,7,4);
}
#define init_6_7_5_12
void FASTCALL call_6_7_5_12(SH2_struct* sh)
{
  SH2extub(sh,7,5);
}
#define init_6_7_6_12
void FASTCALL call_6_7_6_12(SH2_struct* sh)
{
  SH2extub(sh,7,6);
}
#define init_6_7_7_12
void FASTCALL call_6_7_7_12(SH2_struct* sh)
{
  SH2extub(sh,7,7);
}
#define init_6_7_8_12
void FASTCALL call_6_7_8_12(SH2_struct* sh)
{
  SH2extub(sh,7,8);
}
#define init_6_7_9_12
void FASTCALL call_6_7_9_12(SH2_struct* sh)
{
  SH2extub(sh,7,9);
}
#define init_6_7_10_12
void FASTCALL call_6_7_10_12(SH2_struct* sh)
{
  SH2extub(sh,7,10);
}
#define init_6_7_11_12
void FASTCALL call_6_7_11_12(SH2_struct* sh)
{
  SH2extub(sh,7,11);
}
#define init_6_7_12_12
void FASTCALL call_6_7_12_12(SH2_struct* sh)
{
  SH2extub(sh,7,12);
}
#define init_6_7_13_12
void FASTCALL call_6_7_13_12(SH2_struct* sh)
{
  SH2extub(sh,7,13);
}
#define init_6_7_14_12
void FASTCALL call_6_7_14_12(SH2_struct* sh)
{
  SH2extub(sh,7,14);
}
#define init_6_7_15_12
void FASTCALL call_6_7_15_12(SH2_struct* sh)
{
  SH2extub(sh,7,15);
}
#define init_6_8_0_12
void FASTCALL call_6_8_0_12(SH2_struct* sh)
{
  SH2extub(sh,8,0);
}
#define init_6_8_1_12
void FASTCALL call_6_8_1_12(SH2_struct* sh)
{
  SH2extub(sh,8,1);
}
#define init_6_8_2_12
void FASTCALL call_6_8_2_12(SH2_struct* sh)
{
  SH2extub(sh,8,2);
}
#define init_6_8_3_12
void FASTCALL call_6_8_3_12(SH2_struct* sh)
{
  SH2extub(sh,8,3);
}
#define init_6_8_4_12
void FASTCALL call_6_8_4_12(SH2_struct* sh)
{
  SH2extub(sh,8,4);
}
#define init_6_8_5_12
void FASTCALL call_6_8_5_12(SH2_struct* sh)
{
  SH2extub(sh,8,5);
}
#define init_6_8_6_12
void FASTCALL call_6_8_6_12(SH2_struct* sh)
{
  SH2extub(sh,8,6);
}
#define init_6_8_7_12
void FASTCALL call_6_8_7_12(SH2_struct* sh)
{
  SH2extub(sh,8,7);
}
#define init_6_8_8_12
void FASTCALL call_6_8_8_12(SH2_struct* sh)
{
  SH2extub(sh,8,8);
}
#define init_6_8_9_12
void FASTCALL call_6_8_9_12(SH2_struct* sh)
{
  SH2extub(sh,8,9);
}
#define init_6_8_10_12
void FASTCALL call_6_8_10_12(SH2_struct* sh)
{
  SH2extub(sh,8,10);
}
#define init_6_8_11_12
void FASTCALL call_6_8_11_12(SH2_struct* sh)
{
  SH2extub(sh,8,11);
}
#define init_6_8_12_12
void FASTCALL call_6_8_12_12(SH2_struct* sh)
{
  SH2extub(sh,8,12);
}
#define init_6_8_13_12
void FASTCALL call_6_8_13_12(SH2_struct* sh)
{
  SH2extub(sh,8,13);
}
#define init_6_8_14_12
void FASTCALL call_6_8_14_12(SH2_struct* sh)
{
  SH2extub(sh,8,14);
}
#define init_6_8_15_12
void FASTCALL call_6_8_15_12(SH2_struct* sh)
{
  SH2extub(sh,8,15);
}
#define init_6_9_0_12
void FASTCALL call_6_9_0_12(SH2_struct* sh)
{
  SH2extub(sh,9,0);
}
#define init_6_9_1_12
void FASTCALL call_6_9_1_12(SH2_struct* sh)
{
  SH2extub(sh,9,1);
}
#define init_6_9_2_12
void FASTCALL call_6_9_2_12(SH2_struct* sh)
{
  SH2extub(sh,9,2);
}
#define init_6_9_3_12
void FASTCALL call_6_9_3_12(SH2_struct* sh)
{
  SH2extub(sh,9,3);
}
#define init_6_9_4_12
void FASTCALL call_6_9_4_12(SH2_struct* sh)
{
  SH2extub(sh,9,4);
}
#define init_6_9_5_12
void FASTCALL call_6_9_5_12(SH2_struct* sh)
{
  SH2extub(sh,9,5);
}
#define init_6_9_6_12
void FASTCALL call_6_9_6_12(SH2_struct* sh)
{
  SH2extub(sh,9,6);
}
#define init_6_9_7_12
void FASTCALL call_6_9_7_12(SH2_struct* sh)
{
  SH2extub(sh,9,7);
}
#define init_6_9_8_12
void FASTCALL call_6_9_8_12(SH2_struct* sh)
{
  SH2extub(sh,9,8);
}
#define init_6_9_9_12
void FASTCALL call_6_9_9_12(SH2_struct* sh)
{
  SH2extub(sh,9,9);
}
#define init_6_9_10_12
void FASTCALL call_6_9_10_12(SH2_struct* sh)
{
  SH2extub(sh,9,10);
}
#define init_6_9_11_12
void FASTCALL call_6_9_11_12(SH2_struct* sh)
{
  SH2extub(sh,9,11);
}
#define init_6_9_12_12
void FASTCALL call_6_9_12_12(SH2_struct* sh)
{
  SH2extub(sh,9,12);
}
#define init_6_9_13_12
void FASTCALL call_6_9_13_12(SH2_struct* sh)
{
  SH2extub(sh,9,13);
}
#define init_6_9_14_12
void FASTCALL call_6_9_14_12(SH2_struct* sh)
{
  SH2extub(sh,9,14);
}
#define init_6_9_15_12
void FASTCALL call_6_9_15_12(SH2_struct* sh)
{
  SH2extub(sh,9,15);
}
#define init_6_10_0_12
void FASTCALL call_6_10_0_12(SH2_struct* sh)
{
  SH2extub(sh,10,0);
}
#define init_6_10_1_12
void FASTCALL call_6_10_1_12(SH2_struct* sh)
{
  SH2extub(sh,10,1);
}
#define init_6_10_2_12
void FASTCALL call_6_10_2_12(SH2_struct* sh)
{
  SH2extub(sh,10,2);
}
#define init_6_10_3_12
void FASTCALL call_6_10_3_12(SH2_struct* sh)
{
  SH2extub(sh,10,3);
}
#define init_6_10_4_12
void FASTCALL call_6_10_4_12(SH2_struct* sh)
{
  SH2extub(sh,10,4);
}
#define init_6_10_5_12
void FASTCALL call_6_10_5_12(SH2_struct* sh)
{
  SH2extub(sh,10,5);
}
#define init_6_10_6_12
void FASTCALL call_6_10_6_12(SH2_struct* sh)
{
  SH2extub(sh,10,6);
}
#define init_6_10_7_12
void FASTCALL call_6_10_7_12(SH2_struct* sh)
{
  SH2extub(sh,10,7);
}
#define init_6_10_8_12
void FASTCALL call_6_10_8_12(SH2_struct* sh)
{
  SH2extub(sh,10,8);
}
#define init_6_10_9_12
void FASTCALL call_6_10_9_12(SH2_struct* sh)
{
  SH2extub(sh,10,9);
}
#define init_6_10_10_12
void FASTCALL call_6_10_10_12(SH2_struct* sh)
{
  SH2extub(sh,10,10);
}
#define init_6_10_11_12
void FASTCALL call_6_10_11_12(SH2_struct* sh)
{
  SH2extub(sh,10,11);
}
#define init_6_10_12_12
void FASTCALL call_6_10_12_12(SH2_struct* sh)
{
  SH2extub(sh,10,12);
}
#define init_6_10_13_12
void FASTCALL call_6_10_13_12(SH2_struct* sh)
{
  SH2extub(sh,10,13);
}
#define init_6_10_14_12
void FASTCALL call_6_10_14_12(SH2_struct* sh)
{
  SH2extub(sh,10,14);
}
#define init_6_10_15_12
void FASTCALL call_6_10_15_12(SH2_struct* sh)
{
  SH2extub(sh,10,15);
}
#define init_6_11_0_12
void FASTCALL call_6_11_0_12(SH2_struct* sh)
{
  SH2extub(sh,11,0);
}
#define init_6_11_1_12
void FASTCALL call_6_11_1_12(SH2_struct* sh)
{
  SH2extub(sh,11,1);
}
#define init_6_11_2_12
void FASTCALL call_6_11_2_12(SH2_struct* sh)
{
  SH2extub(sh,11,2);
}
#define init_6_11_3_12
void FASTCALL call_6_11_3_12(SH2_struct* sh)
{
  SH2extub(sh,11,3);
}
#define init_6_11_4_12
void FASTCALL call_6_11_4_12(SH2_struct* sh)
{
  SH2extub(sh,11,4);
}
#define init_6_11_5_12
void FASTCALL call_6_11_5_12(SH2_struct* sh)
{
  SH2extub(sh,11,5);
}
#define init_6_11_6_12
void FASTCALL call_6_11_6_12(SH2_struct* sh)
{
  SH2extub(sh,11,6);
}
#define init_6_11_7_12
void FASTCALL call_6_11_7_12(SH2_struct* sh)
{
  SH2extub(sh,11,7);
}
#define init_6_11_8_12
void FASTCALL call_6_11_8_12(SH2_struct* sh)
{
  SH2extub(sh,11,8);
}
#define init_6_11_9_12
void FASTCALL call_6_11_9_12(SH2_struct* sh)
{
  SH2extub(sh,11,9);
}
#define init_6_11_10_12
void FASTCALL call_6_11_10_12(SH2_struct* sh)
{
  SH2extub(sh,11,10);
}
#define init_6_11_11_12
void FASTCALL call_6_11_11_12(SH2_struct* sh)
{
  SH2extub(sh,11,11);
}
#define init_6_11_12_12
void FASTCALL call_6_11_12_12(SH2_struct* sh)
{
  SH2extub(sh,11,12);
}
#define init_6_11_13_12
void FASTCALL call_6_11_13_12(SH2_struct* sh)
{
  SH2extub(sh,11,13);
}
#define init_6_11_14_12
void FASTCALL call_6_11_14_12(SH2_struct* sh)
{
  SH2extub(sh,11,14);
}
#define init_6_11_15_12
void FASTCALL call_6_11_15_12(SH2_struct* sh)
{
  SH2extub(sh,11,15);
}
#define init_6_12_0_12
void FASTCALL call_6_12_0_12(SH2_struct* sh)
{
  SH2extub(sh,12,0);
}
#define init_6_12_1_12
void FASTCALL call_6_12_1_12(SH2_struct* sh)
{
  SH2extub(sh,12,1);
}
#define init_6_12_2_12
void FASTCALL call_6_12_2_12(SH2_struct* sh)
{
  SH2extub(sh,12,2);
}
#define init_6_12_3_12
void FASTCALL call_6_12_3_12(SH2_struct* sh)
{
  SH2extub(sh,12,3);
}
#define init_6_12_4_12
void FASTCALL call_6_12_4_12(SH2_struct* sh)
{
  SH2extub(sh,12,4);
}
#define init_6_12_5_12
void FASTCALL call_6_12_5_12(SH2_struct* sh)
{
  SH2extub(sh,12,5);
}
#define init_6_12_6_12
void FASTCALL call_6_12_6_12(SH2_struct* sh)
{
  SH2extub(sh,12,6);
}
#define init_6_12_7_12
void FASTCALL call_6_12_7_12(SH2_struct* sh)
{
  SH2extub(sh,12,7);
}
#define init_6_12_8_12
void FASTCALL call_6_12_8_12(SH2_struct* sh)
{
  SH2extub(sh,12,8);
}
#define init_6_12_9_12
void FASTCALL call_6_12_9_12(SH2_struct* sh)
{
  SH2extub(sh,12,9);
}
#define init_6_12_10_12
void FASTCALL call_6_12_10_12(SH2_struct* sh)
{
  SH2extub(sh,12,10);
}
#define init_6_12_11_12
void FASTCALL call_6_12_11_12(SH2_struct* sh)
{
  SH2extub(sh,12,11);
}
#define init_6_12_12_12
void FASTCALL call_6_12_12_12(SH2_struct* sh)
{
  SH2extub(sh,12,12);
}
#define init_6_12_13_12
void FASTCALL call_6_12_13_12(SH2_struct* sh)
{
  SH2extub(sh,12,13);
}
#define init_6_12_14_12
void FASTCALL call_6_12_14_12(SH2_struct* sh)
{
  SH2extub(sh,12,14);
}
#define init_6_12_15_12
void FASTCALL call_6_12_15_12(SH2_struct* sh)
{
  SH2extub(sh,12,15);
}
#define init_6_13_0_12
void FASTCALL call_6_13_0_12(SH2_struct* sh)
{
  SH2extub(sh,13,0);
}
#define init_6_13_1_12
void FASTCALL call_6_13_1_12(SH2_struct* sh)
{
  SH2extub(sh,13,1);
}
#define init_6_13_2_12
void FASTCALL call_6_13_2_12(SH2_struct* sh)
{
  SH2extub(sh,13,2);
}
#define init_6_13_3_12
void FASTCALL call_6_13_3_12(SH2_struct* sh)
{
  SH2extub(sh,13,3);
}
#define init_6_13_4_12
void FASTCALL call_6_13_4_12(SH2_struct* sh)
{
  SH2extub(sh,13,4);
}
#define init_6_13_5_12
void FASTCALL call_6_13_5_12(SH2_struct* sh)
{
  SH2extub(sh,13,5);
}
#define init_6_13_6_12
void FASTCALL call_6_13_6_12(SH2_struct* sh)
{
  SH2extub(sh,13,6);
}
#define init_6_13_7_12
void FASTCALL call_6_13_7_12(SH2_struct* sh)
{
  SH2extub(sh,13,7);
}
#define init_6_13_8_12
void FASTCALL call_6_13_8_12(SH2_struct* sh)
{
  SH2extub(sh,13,8);
}
#define init_6_13_9_12
void FASTCALL call_6_13_9_12(SH2_struct* sh)
{
  SH2extub(sh,13,9);
}
#define init_6_13_10_12
void FASTCALL call_6_13_10_12(SH2_struct* sh)
{
  SH2extub(sh,13,10);
}
#define init_6_13_11_12
void FASTCALL call_6_13_11_12(SH2_struct* sh)
{
  SH2extub(sh,13,11);
}
#define init_6_13_12_12
void FASTCALL call_6_13_12_12(SH2_struct* sh)
{
  SH2extub(sh,13,12);
}
#define init_6_13_13_12
void FASTCALL call_6_13_13_12(SH2_struct* sh)
{
  SH2extub(sh,13,13);
}
#define init_6_13_14_12
void FASTCALL call_6_13_14_12(SH2_struct* sh)
{
  SH2extub(sh,13,14);
}
#define init_6_13_15_12
void FASTCALL call_6_13_15_12(SH2_struct* sh)
{
  SH2extub(sh,13,15);
}
#define init_6_14_0_12
void FASTCALL call_6_14_0_12(SH2_struct* sh)
{
  SH2extub(sh,14,0);
}
#define init_6_14_1_12
void FASTCALL call_6_14_1_12(SH2_struct* sh)
{
  SH2extub(sh,14,1);
}
#define init_6_14_2_12
void FASTCALL call_6_14_2_12(SH2_struct* sh)
{
  SH2extub(sh,14,2);
}
#define init_6_14_3_12
void FASTCALL call_6_14_3_12(SH2_struct* sh)
{
  SH2extub(sh,14,3);
}
#define init_6_14_4_12
void FASTCALL call_6_14_4_12(SH2_struct* sh)
{
  SH2extub(sh,14,4);
}
#define init_6_14_5_12
void FASTCALL call_6_14_5_12(SH2_struct* sh)
{
  SH2extub(sh,14,5);
}
#define init_6_14_6_12
void FASTCALL call_6_14_6_12(SH2_struct* sh)
{
  SH2extub(sh,14,6);
}
#define init_6_14_7_12
void FASTCALL call_6_14_7_12(SH2_struct* sh)
{
  SH2extub(sh,14,7);
}
#define init_6_14_8_12
void FASTCALL call_6_14_8_12(SH2_struct* sh)
{
  SH2extub(sh,14,8);
}
#define init_6_14_9_12
void FASTCALL call_6_14_9_12(SH2_struct* sh)
{
  SH2extub(sh,14,9);
}
#define init_6_14_10_12
void FASTCALL call_6_14_10_12(SH2_struct* sh)
{
  SH2extub(sh,14,10);
}
#define init_6_14_11_12
void FASTCALL call_6_14_11_12(SH2_struct* sh)
{
  SH2extub(sh,14,11);
}
#define init_6_14_12_12
void FASTCALL call_6_14_12_12(SH2_struct* sh)
{
  SH2extub(sh,14,12);
}
#define init_6_14_13_12
void FASTCALL call_6_14_13_12(SH2_struct* sh)
{
  SH2extub(sh,14,13);
}
#define init_6_14_14_12
void FASTCALL call_6_14_14_12(SH2_struct* sh)
{
  SH2extub(sh,14,14);
}
#define init_6_14_15_12
void FASTCALL call_6_14_15_12(SH2_struct* sh)
{
  SH2extub(sh,14,15);
}
#define init_6_15_0_12
void FASTCALL call_6_15_0_12(SH2_struct* sh)
{
  SH2extub(sh,15,0);
}
#define init_6_15_1_12
void FASTCALL call_6_15_1_12(SH2_struct* sh)
{
  SH2extub(sh,15,1);
}
#define init_6_15_2_12
void FASTCALL call_6_15_2_12(SH2_struct* sh)
{
  SH2extub(sh,15,2);
}
#define init_6_15_3_12
void FASTCALL call_6_15_3_12(SH2_struct* sh)
{
  SH2extub(sh,15,3);
}
#define init_6_15_4_12
void FASTCALL call_6_15_4_12(SH2_struct* sh)
{
  SH2extub(sh,15,4);
}
#define init_6_15_5_12
void FASTCALL call_6_15_5_12(SH2_struct* sh)
{
  SH2extub(sh,15,5);
}
#define init_6_15_6_12
void FASTCALL call_6_15_6_12(SH2_struct* sh)
{
  SH2extub(sh,15,6);
}
#define init_6_15_7_12
void FASTCALL call_6_15_7_12(SH2_struct* sh)
{
  SH2extub(sh,15,7);
}
#define init_6_15_8_12
void FASTCALL call_6_15_8_12(SH2_struct* sh)
{
  SH2extub(sh,15,8);
}
#define init_6_15_9_12
void FASTCALL call_6_15_9_12(SH2_struct* sh)
{
  SH2extub(sh,15,9);
}
#define init_6_15_10_12
void FASTCALL call_6_15_10_12(SH2_struct* sh)
{
  SH2extub(sh,15,10);
}
#define init_6_15_11_12
void FASTCALL call_6_15_11_12(SH2_struct* sh)
{
  SH2extub(sh,15,11);
}
#define init_6_15_12_12
void FASTCALL call_6_15_12_12(SH2_struct* sh)
{
  SH2extub(sh,15,12);
}
#define init_6_15_13_12
void FASTCALL call_6_15_13_12(SH2_struct* sh)
{
  SH2extub(sh,15,13);
}
#define init_6_15_14_12
void FASTCALL call_6_15_14_12(SH2_struct* sh)
{
  SH2extub(sh,15,14);
}
#define init_6_15_15_12
void FASTCALL call_6_15_15_12(SH2_struct* sh)
{
  SH2extub(sh,15,15);
}
#define init_6_0_0_15
void FASTCALL call_6_0_0_15(SH2_struct* sh)
{
  SH2extsw(sh,0,0);
}
#define init_6_0_1_15
void FASTCALL call_6_0_1_15(SH2_struct* sh)
{
  SH2extsw(sh,0,1);
}
#define init_6_0_2_15
void FASTCALL call_6_0_2_15(SH2_struct* sh)
{
  SH2extsw(sh,0,2);
}
#define init_6_0_3_15
void FASTCALL call_6_0_3_15(SH2_struct* sh)
{
  SH2extsw(sh,0,3);
}
#define init_6_0_4_15
void FASTCALL call_6_0_4_15(SH2_struct* sh)
{
  SH2extsw(sh,0,4);
}
#define init_6_0_5_15
void FASTCALL call_6_0_5_15(SH2_struct* sh)
{
  SH2extsw(sh,0,5);
}
#define init_6_0_6_15
void FASTCALL call_6_0_6_15(SH2_struct* sh)
{
  SH2extsw(sh,0,6);
}
#define init_6_0_7_15
void FASTCALL call_6_0_7_15(SH2_struct* sh)
{
  SH2extsw(sh,0,7);
}
#define init_6_0_8_15
void FASTCALL call_6_0_8_15(SH2_struct* sh)
{
  SH2extsw(sh,0,8);
}
#define init_6_0_9_15
void FASTCALL call_6_0_9_15(SH2_struct* sh)
{
  SH2extsw(sh,0,9);
}
#define init_6_0_10_15
void FASTCALL call_6_0_10_15(SH2_struct* sh)
{
  SH2extsw(sh,0,10);
}
#define init_6_0_11_15
void FASTCALL call_6_0_11_15(SH2_struct* sh)
{
  SH2extsw(sh,0,11);
}
#define init_6_0_12_15
void FASTCALL call_6_0_12_15(SH2_struct* sh)
{
  SH2extsw(sh,0,12);
}
#define init_6_0_13_15
void FASTCALL call_6_0_13_15(SH2_struct* sh)
{
  SH2extsw(sh,0,13);
}
#define init_6_0_14_15
void FASTCALL call_6_0_14_15(SH2_struct* sh)
{
  SH2extsw(sh,0,14);
}
#define init_6_0_15_15
void FASTCALL call_6_0_15_15(SH2_struct* sh)
{
  SH2extsw(sh,0,15);
}
#define init_6_1_0_15
void FASTCALL call_6_1_0_15(SH2_struct* sh)
{
  SH2extsw(sh,1,0);
}
#define init_6_1_1_15
void FASTCALL call_6_1_1_15(SH2_struct* sh)
{
  SH2extsw(sh,1,1);
}
#define init_6_1_2_15
void FASTCALL call_6_1_2_15(SH2_struct* sh)
{
  SH2extsw(sh,1,2);
}
#define init_6_1_3_15
void FASTCALL call_6_1_3_15(SH2_struct* sh)
{
  SH2extsw(sh,1,3);
}
#define init_6_1_4_15
void FASTCALL call_6_1_4_15(SH2_struct* sh)
{
  SH2extsw(sh,1,4);
}
#define init_6_1_5_15
void FASTCALL call_6_1_5_15(SH2_struct* sh)
{
  SH2extsw(sh,1,5);
}
#define init_6_1_6_15
void FASTCALL call_6_1_6_15(SH2_struct* sh)
{
  SH2extsw(sh,1,6);
}
#define init_6_1_7_15
void FASTCALL call_6_1_7_15(SH2_struct* sh)
{
  SH2extsw(sh,1,7);
}
#define init_6_1_8_15
void FASTCALL call_6_1_8_15(SH2_struct* sh)
{
  SH2extsw(sh,1,8);
}
#define init_6_1_9_15
void FASTCALL call_6_1_9_15(SH2_struct* sh)
{
  SH2extsw(sh,1,9);
}
#define init_6_1_10_15
void FASTCALL call_6_1_10_15(SH2_struct* sh)
{
  SH2extsw(sh,1,10);
}
#define init_6_1_11_15
void FASTCALL call_6_1_11_15(SH2_struct* sh)
{
  SH2extsw(sh,1,11);
}
#define init_6_1_12_15
void FASTCALL call_6_1_12_15(SH2_struct* sh)
{
  SH2extsw(sh,1,12);
}
#define init_6_1_13_15
void FASTCALL call_6_1_13_15(SH2_struct* sh)
{
  SH2extsw(sh,1,13);
}
#define init_6_1_14_15
void FASTCALL call_6_1_14_15(SH2_struct* sh)
{
  SH2extsw(sh,1,14);
}
#define init_6_1_15_15
void FASTCALL call_6_1_15_15(SH2_struct* sh)
{
  SH2extsw(sh,1,15);
}
#define init_6_2_0_15
void FASTCALL call_6_2_0_15(SH2_struct* sh)
{
  SH2extsw(sh,2,0);
}
#define init_6_2_1_15
void FASTCALL call_6_2_1_15(SH2_struct* sh)
{
  SH2extsw(sh,2,1);
}
#define init_6_2_2_15
void FASTCALL call_6_2_2_15(SH2_struct* sh)
{
  SH2extsw(sh,2,2);
}
#define init_6_2_3_15
void FASTCALL call_6_2_3_15(SH2_struct* sh)
{
  SH2extsw(sh,2,3);
}
#define init_6_2_4_15
void FASTCALL call_6_2_4_15(SH2_struct* sh)
{
  SH2extsw(sh,2,4);
}
#define init_6_2_5_15
void FASTCALL call_6_2_5_15(SH2_struct* sh)
{
  SH2extsw(sh,2,5);
}
#define init_6_2_6_15
void FASTCALL call_6_2_6_15(SH2_struct* sh)
{
  SH2extsw(sh,2,6);
}
#define init_6_2_7_15
void FASTCALL call_6_2_7_15(SH2_struct* sh)
{
  SH2extsw(sh,2,7);
}
#define init_6_2_8_15
void FASTCALL call_6_2_8_15(SH2_struct* sh)
{
  SH2extsw(sh,2,8);
}
#define init_6_2_9_15
void FASTCALL call_6_2_9_15(SH2_struct* sh)
{
  SH2extsw(sh,2,9);
}
#define init_6_2_10_15
void FASTCALL call_6_2_10_15(SH2_struct* sh)
{
  SH2extsw(sh,2,10);
}
#define init_6_2_11_15
void FASTCALL call_6_2_11_15(SH2_struct* sh)
{
  SH2extsw(sh,2,11);
}
#define init_6_2_12_15
void FASTCALL call_6_2_12_15(SH2_struct* sh)
{
  SH2extsw(sh,2,12);
}
#define init_6_2_13_15
void FASTCALL call_6_2_13_15(SH2_struct* sh)
{
  SH2extsw(sh,2,13);
}
#define init_6_2_14_15
void FASTCALL call_6_2_14_15(SH2_struct* sh)
{
  SH2extsw(sh,2,14);
}
#define init_6_2_15_15
void FASTCALL call_6_2_15_15(SH2_struct* sh)
{
  SH2extsw(sh,2,15);
}
#define init_6_3_0_15
void FASTCALL call_6_3_0_15(SH2_struct* sh)
{
  SH2extsw(sh,3,0);
}
#define init_6_3_1_15
void FASTCALL call_6_3_1_15(SH2_struct* sh)
{
  SH2extsw(sh,3,1);
}
#define init_6_3_2_15
void FASTCALL call_6_3_2_15(SH2_struct* sh)
{
  SH2extsw(sh,3,2);
}
#define init_6_3_3_15
void FASTCALL call_6_3_3_15(SH2_struct* sh)
{
  SH2extsw(sh,3,3);
}
#define init_6_3_4_15
void FASTCALL call_6_3_4_15(SH2_struct* sh)
{
  SH2extsw(sh,3,4);
}
#define init_6_3_5_15
void FASTCALL call_6_3_5_15(SH2_struct* sh)
{
  SH2extsw(sh,3,5);
}
#define init_6_3_6_15
void FASTCALL call_6_3_6_15(SH2_struct* sh)
{
  SH2extsw(sh,3,6);
}
#define init_6_3_7_15
void FASTCALL call_6_3_7_15(SH2_struct* sh)
{
  SH2extsw(sh,3,7);
}
#define init_6_3_8_15
void FASTCALL call_6_3_8_15(SH2_struct* sh)
{
  SH2extsw(sh,3,8);
}
#define init_6_3_9_15
void FASTCALL call_6_3_9_15(SH2_struct* sh)
{
  SH2extsw(sh,3,9);
}
#define init_6_3_10_15
void FASTCALL call_6_3_10_15(SH2_struct* sh)
{
  SH2extsw(sh,3,10);
}
#define init_6_3_11_15
void FASTCALL call_6_3_11_15(SH2_struct* sh)
{
  SH2extsw(sh,3,11);
}
#define init_6_3_12_15
void FASTCALL call_6_3_12_15(SH2_struct* sh)
{
  SH2extsw(sh,3,12);
}
#define init_6_3_13_15
void FASTCALL call_6_3_13_15(SH2_struct* sh)
{
  SH2extsw(sh,3,13);
}
#define init_6_3_14_15
void FASTCALL call_6_3_14_15(SH2_struct* sh)
{
  SH2extsw(sh,3,14);
}
#define init_6_3_15_15
void FASTCALL call_6_3_15_15(SH2_struct* sh)
{
  SH2extsw(sh,3,15);
}
#define init_6_4_0_15
void FASTCALL call_6_4_0_15(SH2_struct* sh)
{
  SH2extsw(sh,4,0);
}
#define init_6_4_1_15
void FASTCALL call_6_4_1_15(SH2_struct* sh)
{
  SH2extsw(sh,4,1);
}
#define init_6_4_2_15
void FASTCALL call_6_4_2_15(SH2_struct* sh)
{
  SH2extsw(sh,4,2);
}
#define init_6_4_3_15
void FASTCALL call_6_4_3_15(SH2_struct* sh)
{
  SH2extsw(sh,4,3);
}
#define init_6_4_4_15
void FASTCALL call_6_4_4_15(SH2_struct* sh)
{
  SH2extsw(sh,4,4);
}
#define init_6_4_5_15
void FASTCALL call_6_4_5_15(SH2_struct* sh)
{
  SH2extsw(sh,4,5);
}
#define init_6_4_6_15
void FASTCALL call_6_4_6_15(SH2_struct* sh)
{
  SH2extsw(sh,4,6);
}
#define init_6_4_7_15
void FASTCALL call_6_4_7_15(SH2_struct* sh)
{
  SH2extsw(sh,4,7);
}
#define init_6_4_8_15
void FASTCALL call_6_4_8_15(SH2_struct* sh)
{
  SH2extsw(sh,4,8);
}
#define init_6_4_9_15
void FASTCALL call_6_4_9_15(SH2_struct* sh)
{
  SH2extsw(sh,4,9);
}
#define init_6_4_10_15
void FASTCALL call_6_4_10_15(SH2_struct* sh)
{
  SH2extsw(sh,4,10);
}
#define init_6_4_11_15
void FASTCALL call_6_4_11_15(SH2_struct* sh)
{
  SH2extsw(sh,4,11);
}
#define init_6_4_12_15
void FASTCALL call_6_4_12_15(SH2_struct* sh)
{
  SH2extsw(sh,4,12);
}
#define init_6_4_13_15
void FASTCALL call_6_4_13_15(SH2_struct* sh)
{
  SH2extsw(sh,4,13);
}
#define init_6_4_14_15
void FASTCALL call_6_4_14_15(SH2_struct* sh)
{
  SH2extsw(sh,4,14);
}
#define init_6_4_15_15
void FASTCALL call_6_4_15_15(SH2_struct* sh)
{
  SH2extsw(sh,4,15);
}
#define init_6_5_0_15
void FASTCALL call_6_5_0_15(SH2_struct* sh)
{
  SH2extsw(sh,5,0);
}
#define init_6_5_1_15
void FASTCALL call_6_5_1_15(SH2_struct* sh)
{
  SH2extsw(sh,5,1);
}
#define init_6_5_2_15
void FASTCALL call_6_5_2_15(SH2_struct* sh)
{
  SH2extsw(sh,5,2);
}
#define init_6_5_3_15
void FASTCALL call_6_5_3_15(SH2_struct* sh)
{
  SH2extsw(sh,5,3);
}
#define init_6_5_4_15
void FASTCALL call_6_5_4_15(SH2_struct* sh)
{
  SH2extsw(sh,5,4);
}
#define init_6_5_5_15
void FASTCALL call_6_5_5_15(SH2_struct* sh)
{
  SH2extsw(sh,5,5);
}
#define init_6_5_6_15
void FASTCALL call_6_5_6_15(SH2_struct* sh)
{
  SH2extsw(sh,5,6);
}
#define init_6_5_7_15
void FASTCALL call_6_5_7_15(SH2_struct* sh)
{
  SH2extsw(sh,5,7);
}
#define init_6_5_8_15
void FASTCALL call_6_5_8_15(SH2_struct* sh)
{
  SH2extsw(sh,5,8);
}
#define init_6_5_9_15
void FASTCALL call_6_5_9_15(SH2_struct* sh)
{
  SH2extsw(sh,5,9);
}
#define init_6_5_10_15
void FASTCALL call_6_5_10_15(SH2_struct* sh)
{
  SH2extsw(sh,5,10);
}
#define init_6_5_11_15
void FASTCALL call_6_5_11_15(SH2_struct* sh)
{
  SH2extsw(sh,5,11);
}
#define init_6_5_12_15
void FASTCALL call_6_5_12_15(SH2_struct* sh)
{
  SH2extsw(sh,5,12);
}
#define init_6_5_13_15
void FASTCALL call_6_5_13_15(SH2_struct* sh)
{
  SH2extsw(sh,5,13);
}
#define init_6_5_14_15
void FASTCALL call_6_5_14_15(SH2_struct* sh)
{
  SH2extsw(sh,5,14);
}
#define init_6_5_15_15
void FASTCALL call_6_5_15_15(SH2_struct* sh)
{
  SH2extsw(sh,5,15);
}
#define init_6_6_0_15
void FASTCALL call_6_6_0_15(SH2_struct* sh)
{
  SH2extsw(sh,6,0);
}
#define init_6_6_1_15
void FASTCALL call_6_6_1_15(SH2_struct* sh)
{
  SH2extsw(sh,6,1);
}
#define init_6_6_2_15
void FASTCALL call_6_6_2_15(SH2_struct* sh)
{
  SH2extsw(sh,6,2);
}
#define init_6_6_3_15
void FASTCALL call_6_6_3_15(SH2_struct* sh)
{
  SH2extsw(sh,6,3);
}
#define init_6_6_4_15
void FASTCALL call_6_6_4_15(SH2_struct* sh)
{
  SH2extsw(sh,6,4);
}
#define init_6_6_5_15
void FASTCALL call_6_6_5_15(SH2_struct* sh)
{
  SH2extsw(sh,6,5);
}
#define init_6_6_6_15
void FASTCALL call_6_6_6_15(SH2_struct* sh)
{
  SH2extsw(sh,6,6);
}
#define init_6_6_7_15
void FASTCALL call_6_6_7_15(SH2_struct* sh)
{
  SH2extsw(sh,6,7);
}
#define init_6_6_8_15
void FASTCALL call_6_6_8_15(SH2_struct* sh)
{
  SH2extsw(sh,6,8);
}
#define init_6_6_9_15
void FASTCALL call_6_6_9_15(SH2_struct* sh)
{
  SH2extsw(sh,6,9);
}
#define init_6_6_10_15
void FASTCALL call_6_6_10_15(SH2_struct* sh)
{
  SH2extsw(sh,6,10);
}
#define init_6_6_11_15
void FASTCALL call_6_6_11_15(SH2_struct* sh)
{
  SH2extsw(sh,6,11);
}
#define init_6_6_12_15
void FASTCALL call_6_6_12_15(SH2_struct* sh)
{
  SH2extsw(sh,6,12);
}
#define init_6_6_13_15
void FASTCALL call_6_6_13_15(SH2_struct* sh)
{
  SH2extsw(sh,6,13);
}
#define init_6_6_14_15
void FASTCALL call_6_6_14_15(SH2_struct* sh)
{
  SH2extsw(sh,6,14);
}
#define init_6_6_15_15
void FASTCALL call_6_6_15_15(SH2_struct* sh)
{
  SH2extsw(sh,6,15);
}
#define init_6_7_0_15
void FASTCALL call_6_7_0_15(SH2_struct* sh)
{
  SH2extsw(sh,7,0);
}
#define init_6_7_1_15
void FASTCALL call_6_7_1_15(SH2_struct* sh)
{
  SH2extsw(sh,7,1);
}
#define init_6_7_2_15
void FASTCALL call_6_7_2_15(SH2_struct* sh)
{
  SH2extsw(sh,7,2);
}
#define init_6_7_3_15
void FASTCALL call_6_7_3_15(SH2_struct* sh)
{
  SH2extsw(sh,7,3);
}
#define init_6_7_4_15
void FASTCALL call_6_7_4_15(SH2_struct* sh)
{
  SH2extsw(sh,7,4);
}
#define init_6_7_5_15
void FASTCALL call_6_7_5_15(SH2_struct* sh)
{
  SH2extsw(sh,7,5);
}
#define init_6_7_6_15
void FASTCALL call_6_7_6_15(SH2_struct* sh)
{
  SH2extsw(sh,7,6);
}
#define init_6_7_7_15
void FASTCALL call_6_7_7_15(SH2_struct* sh)
{
  SH2extsw(sh,7,7);
}
#define init_6_7_8_15
void FASTCALL call_6_7_8_15(SH2_struct* sh)
{
  SH2extsw(sh,7,8);
}
#define init_6_7_9_15
void FASTCALL call_6_7_9_15(SH2_struct* sh)
{
  SH2extsw(sh,7,9);
}
#define init_6_7_10_15
void FASTCALL call_6_7_10_15(SH2_struct* sh)
{
  SH2extsw(sh,7,10);
}
#define init_6_7_11_15
void FASTCALL call_6_7_11_15(SH2_struct* sh)
{
  SH2extsw(sh,7,11);
}
#define init_6_7_12_15
void FASTCALL call_6_7_12_15(SH2_struct* sh)
{
  SH2extsw(sh,7,12);
}
#define init_6_7_13_15
void FASTCALL call_6_7_13_15(SH2_struct* sh)
{
  SH2extsw(sh,7,13);
}
#define init_6_7_14_15
void FASTCALL call_6_7_14_15(SH2_struct* sh)
{
  SH2extsw(sh,7,14);
}
#define init_6_7_15_15
void FASTCALL call_6_7_15_15(SH2_struct* sh)
{
  SH2extsw(sh,7,15);
}
#define init_6_8_0_15
void FASTCALL call_6_8_0_15(SH2_struct* sh)
{
  SH2extsw(sh,8,0);
}
#define init_6_8_1_15
void FASTCALL call_6_8_1_15(SH2_struct* sh)
{
  SH2extsw(sh,8,1);
}
#define init_6_8_2_15
void FASTCALL call_6_8_2_15(SH2_struct* sh)
{
  SH2extsw(sh,8,2);
}
#define init_6_8_3_15
void FASTCALL call_6_8_3_15(SH2_struct* sh)
{
  SH2extsw(sh,8,3);
}
#define init_6_8_4_15
void FASTCALL call_6_8_4_15(SH2_struct* sh)
{
  SH2extsw(sh,8,4);
}
#define init_6_8_5_15
void FASTCALL call_6_8_5_15(SH2_struct* sh)
{
  SH2extsw(sh,8,5);
}
#define init_6_8_6_15
void FASTCALL call_6_8_6_15(SH2_struct* sh)
{
  SH2extsw(sh,8,6);
}
#define init_6_8_7_15
void FASTCALL call_6_8_7_15(SH2_struct* sh)
{
  SH2extsw(sh,8,7);
}
#define init_6_8_8_15
void FASTCALL call_6_8_8_15(SH2_struct* sh)
{
  SH2extsw(sh,8,8);
}
#define init_6_8_9_15
void FASTCALL call_6_8_9_15(SH2_struct* sh)
{
  SH2extsw(sh,8,9);
}
#define init_6_8_10_15
void FASTCALL call_6_8_10_15(SH2_struct* sh)
{
  SH2extsw(sh,8,10);
}
#define init_6_8_11_15
void FASTCALL call_6_8_11_15(SH2_struct* sh)
{
  SH2extsw(sh,8,11);
}
#define init_6_8_12_15
void FASTCALL call_6_8_12_15(SH2_struct* sh)
{
  SH2extsw(sh,8,12);
}
#define init_6_8_13_15
void FASTCALL call_6_8_13_15(SH2_struct* sh)
{
  SH2extsw(sh,8,13);
}
#define init_6_8_14_15
void FASTCALL call_6_8_14_15(SH2_struct* sh)
{
  SH2extsw(sh,8,14);
}
#define init_6_8_15_15
void FASTCALL call_6_8_15_15(SH2_struct* sh)
{
  SH2extsw(sh,8,15);
}
#define init_6_9_0_15
void FASTCALL call_6_9_0_15(SH2_struct* sh)
{
  SH2extsw(sh,9,0);
}
#define init_6_9_1_15
void FASTCALL call_6_9_1_15(SH2_struct* sh)
{
  SH2extsw(sh,9,1);
}
#define init_6_9_2_15
void FASTCALL call_6_9_2_15(SH2_struct* sh)
{
  SH2extsw(sh,9,2);
}
#define init_6_9_3_15
void FASTCALL call_6_9_3_15(SH2_struct* sh)
{
  SH2extsw(sh,9,3);
}
#define init_6_9_4_15
void FASTCALL call_6_9_4_15(SH2_struct* sh)
{
  SH2extsw(sh,9,4);
}
#define init_6_9_5_15
void FASTCALL call_6_9_5_15(SH2_struct* sh)
{
  SH2extsw(sh,9,5);
}
#define init_6_9_6_15
void FASTCALL call_6_9_6_15(SH2_struct* sh)
{
  SH2extsw(sh,9,6);
}
#define init_6_9_7_15
void FASTCALL call_6_9_7_15(SH2_struct* sh)
{
  SH2extsw(sh,9,7);
}
#define init_6_9_8_15
void FASTCALL call_6_9_8_15(SH2_struct* sh)
{
  SH2extsw(sh,9,8);
}
#define init_6_9_9_15
void FASTCALL call_6_9_9_15(SH2_struct* sh)
{
  SH2extsw(sh,9,9);
}
#define init_6_9_10_15
void FASTCALL call_6_9_10_15(SH2_struct* sh)
{
  SH2extsw(sh,9,10);
}
#define init_6_9_11_15
void FASTCALL call_6_9_11_15(SH2_struct* sh)
{
  SH2extsw(sh,9,11);
}
#define init_6_9_12_15
void FASTCALL call_6_9_12_15(SH2_struct* sh)
{
  SH2extsw(sh,9,12);
}
#define init_6_9_13_15
void FASTCALL call_6_9_13_15(SH2_struct* sh)
{
  SH2extsw(sh,9,13);
}
#define init_6_9_14_15
void FASTCALL call_6_9_14_15(SH2_struct* sh)
{
  SH2extsw(sh,9,14);
}
#define init_6_9_15_15
void FASTCALL call_6_9_15_15(SH2_struct* sh)
{
  SH2extsw(sh,9,15);
}
#define init_6_10_0_15
void FASTCALL call_6_10_0_15(SH2_struct* sh)
{
  SH2extsw(sh,10,0);
}
#define init_6_10_1_15
void FASTCALL call_6_10_1_15(SH2_struct* sh)
{
  SH2extsw(sh,10,1);
}
#define init_6_10_2_15
void FASTCALL call_6_10_2_15(SH2_struct* sh)
{
  SH2extsw(sh,10,2);
}
#define init_6_10_3_15
void FASTCALL call_6_10_3_15(SH2_struct* sh)
{
  SH2extsw(sh,10,3);
}
#define init_6_10_4_15
void FASTCALL call_6_10_4_15(SH2_struct* sh)
{
  SH2extsw(sh,10,4);
}
#define init_6_10_5_15
void FASTCALL call_6_10_5_15(SH2_struct* sh)
{
  SH2extsw(sh,10,5);
}
#define init_6_10_6_15
void FASTCALL call_6_10_6_15(SH2_struct* sh)
{
  SH2extsw(sh,10,6);
}
#define init_6_10_7_15
void FASTCALL call_6_10_7_15(SH2_struct* sh)
{
  SH2extsw(sh,10,7);
}
#define init_6_10_8_15
void FASTCALL call_6_10_8_15(SH2_struct* sh)
{
  SH2extsw(sh,10,8);
}
#define init_6_10_9_15
void FASTCALL call_6_10_9_15(SH2_struct* sh)
{
  SH2extsw(sh,10,9);
}
#define init_6_10_10_15
void FASTCALL call_6_10_10_15(SH2_struct* sh)
{
  SH2extsw(sh,10,10);
}
#define init_6_10_11_15
void FASTCALL call_6_10_11_15(SH2_struct* sh)
{
  SH2extsw(sh,10,11);
}
#define init_6_10_12_15
void FASTCALL call_6_10_12_15(SH2_struct* sh)
{
  SH2extsw(sh,10,12);
}
#define init_6_10_13_15
void FASTCALL call_6_10_13_15(SH2_struct* sh)
{
  SH2extsw(sh,10,13);
}
#define init_6_10_14_15
void FASTCALL call_6_10_14_15(SH2_struct* sh)
{
  SH2extsw(sh,10,14);
}
#define init_6_10_15_15
void FASTCALL call_6_10_15_15(SH2_struct* sh)
{
  SH2extsw(sh,10,15);
}
#define init_6_11_0_15
void FASTCALL call_6_11_0_15(SH2_struct* sh)
{
  SH2extsw(sh,11,0);
}
#define init_6_11_1_15
void FASTCALL call_6_11_1_15(SH2_struct* sh)
{
  SH2extsw(sh,11,1);
}
#define init_6_11_2_15
void FASTCALL call_6_11_2_15(SH2_struct* sh)
{
  SH2extsw(sh,11,2);
}
#define init_6_11_3_15
void FASTCALL call_6_11_3_15(SH2_struct* sh)
{
  SH2extsw(sh,11,3);
}
#define init_6_11_4_15
void FASTCALL call_6_11_4_15(SH2_struct* sh)
{
  SH2extsw(sh,11,4);
}
#define init_6_11_5_15
void FASTCALL call_6_11_5_15(SH2_struct* sh)
{
  SH2extsw(sh,11,5);
}
#define init_6_11_6_15
void FASTCALL call_6_11_6_15(SH2_struct* sh)
{
  SH2extsw(sh,11,6);
}
#define init_6_11_7_15
void FASTCALL call_6_11_7_15(SH2_struct* sh)
{
  SH2extsw(sh,11,7);
}
#define init_6_11_8_15
void FASTCALL call_6_11_8_15(SH2_struct* sh)
{
  SH2extsw(sh,11,8);
}
#define init_6_11_9_15
void FASTCALL call_6_11_9_15(SH2_struct* sh)
{
  SH2extsw(sh,11,9);
}
#define init_6_11_10_15
void FASTCALL call_6_11_10_15(SH2_struct* sh)
{
  SH2extsw(sh,11,10);
}
#define init_6_11_11_15
void FASTCALL call_6_11_11_15(SH2_struct* sh)
{
  SH2extsw(sh,11,11);
}
#define init_6_11_12_15
void FASTCALL call_6_11_12_15(SH2_struct* sh)
{
  SH2extsw(sh,11,12);
}
#define init_6_11_13_15
void FASTCALL call_6_11_13_15(SH2_struct* sh)
{
  SH2extsw(sh,11,13);
}
#define init_6_11_14_15
void FASTCALL call_6_11_14_15(SH2_struct* sh)
{
  SH2extsw(sh,11,14);
}
#define init_6_11_15_15
void FASTCALL call_6_11_15_15(SH2_struct* sh)
{
  SH2extsw(sh,11,15);
}
#define init_6_12_0_15
void FASTCALL call_6_12_0_15(SH2_struct* sh)
{
  SH2extsw(sh,12,0);
}
#define init_6_12_1_15
void FASTCALL call_6_12_1_15(SH2_struct* sh)
{
  SH2extsw(sh,12,1);
}
#define init_6_12_2_15
void FASTCALL call_6_12_2_15(SH2_struct* sh)
{
  SH2extsw(sh,12,2);
}
#define init_6_12_3_15
void FASTCALL call_6_12_3_15(SH2_struct* sh)
{
  SH2extsw(sh,12,3);
}
#define init_6_12_4_15
void FASTCALL call_6_12_4_15(SH2_struct* sh)
{
  SH2extsw(sh,12,4);
}
#define init_6_12_5_15
void FASTCALL call_6_12_5_15(SH2_struct* sh)
{
  SH2extsw(sh,12,5);
}
#define init_6_12_6_15
void FASTCALL call_6_12_6_15(SH2_struct* sh)
{
  SH2extsw(sh,12,6);
}
#define init_6_12_7_15
void FASTCALL call_6_12_7_15(SH2_struct* sh)
{
  SH2extsw(sh,12,7);
}
#define init_6_12_8_15
void FASTCALL call_6_12_8_15(SH2_struct* sh)
{
  SH2extsw(sh,12,8);
}
#define init_6_12_9_15
void FASTCALL call_6_12_9_15(SH2_struct* sh)
{
  SH2extsw(sh,12,9);
}
#define init_6_12_10_15
void FASTCALL call_6_12_10_15(SH2_struct* sh)
{
  SH2extsw(sh,12,10);
}
#define init_6_12_11_15
void FASTCALL call_6_12_11_15(SH2_struct* sh)
{
  SH2extsw(sh,12,11);
}
#define init_6_12_12_15
void FASTCALL call_6_12_12_15(SH2_struct* sh)
{
  SH2extsw(sh,12,12);
}
#define init_6_12_13_15
void FASTCALL call_6_12_13_15(SH2_struct* sh)
{
  SH2extsw(sh,12,13);
}
#define init_6_12_14_15
void FASTCALL call_6_12_14_15(SH2_struct* sh)
{
  SH2extsw(sh,12,14);
}
#define init_6_12_15_15
void FASTCALL call_6_12_15_15(SH2_struct* sh)
{
  SH2extsw(sh,12,15);
}
#define init_6_13_0_15
void FASTCALL call_6_13_0_15(SH2_struct* sh)
{
  SH2extsw(sh,13,0);
}
#define init_6_13_1_15
void FASTCALL call_6_13_1_15(SH2_struct* sh)
{
  SH2extsw(sh,13,1);
}
#define init_6_13_2_15
void FASTCALL call_6_13_2_15(SH2_struct* sh)
{
  SH2extsw(sh,13,2);
}
#define init_6_13_3_15
void FASTCALL call_6_13_3_15(SH2_struct* sh)
{
  SH2extsw(sh,13,3);
}
#define init_6_13_4_15
void FASTCALL call_6_13_4_15(SH2_struct* sh)
{
  SH2extsw(sh,13,4);
}
#define init_6_13_5_15
void FASTCALL call_6_13_5_15(SH2_struct* sh)
{
  SH2extsw(sh,13,5);
}
#define init_6_13_6_15
void FASTCALL call_6_13_6_15(SH2_struct* sh)
{
  SH2extsw(sh,13,6);
}
#define init_6_13_7_15
void FASTCALL call_6_13_7_15(SH2_struct* sh)
{
  SH2extsw(sh,13,7);
}
#define init_6_13_8_15
void FASTCALL call_6_13_8_15(SH2_struct* sh)
{
  SH2extsw(sh,13,8);
}
#define init_6_13_9_15
void FASTCALL call_6_13_9_15(SH2_struct* sh)
{
  SH2extsw(sh,13,9);
}
#define init_6_13_10_15
void FASTCALL call_6_13_10_15(SH2_struct* sh)
{
  SH2extsw(sh,13,10);
}
#define init_6_13_11_15
void FASTCALL call_6_13_11_15(SH2_struct* sh)
{
  SH2extsw(sh,13,11);
}
#define init_6_13_12_15
void FASTCALL call_6_13_12_15(SH2_struct* sh)
{
  SH2extsw(sh,13,12);
}
#define init_6_13_13_15
void FASTCALL call_6_13_13_15(SH2_struct* sh)
{
  SH2extsw(sh,13,13);
}
#define init_6_13_14_15
void FASTCALL call_6_13_14_15(SH2_struct* sh)
{
  SH2extsw(sh,13,14);
}
#define init_6_13_15_15
void FASTCALL call_6_13_15_15(SH2_struct* sh)
{
  SH2extsw(sh,13,15);
}
#define init_6_14_0_15
void FASTCALL call_6_14_0_15(SH2_struct* sh)
{
  SH2extsw(sh,14,0);
}
#define init_6_14_1_15
void FASTCALL call_6_14_1_15(SH2_struct* sh)
{
  SH2extsw(sh,14,1);
}
#define init_6_14_2_15
void FASTCALL call_6_14_2_15(SH2_struct* sh)
{
  SH2extsw(sh,14,2);
}
#define init_6_14_3_15
void FASTCALL call_6_14_3_15(SH2_struct* sh)
{
  SH2extsw(sh,14,3);
}
#define init_6_14_4_15
void FASTCALL call_6_14_4_15(SH2_struct* sh)
{
  SH2extsw(sh,14,4);
}
#define init_6_14_5_15
void FASTCALL call_6_14_5_15(SH2_struct* sh)
{
  SH2extsw(sh,14,5);
}
#define init_6_14_6_15
void FASTCALL call_6_14_6_15(SH2_struct* sh)
{
  SH2extsw(sh,14,6);
}
#define init_6_14_7_15
void FASTCALL call_6_14_7_15(SH2_struct* sh)
{
  SH2extsw(sh,14,7);
}
#define init_6_14_8_15
void FASTCALL call_6_14_8_15(SH2_struct* sh)
{
  SH2extsw(sh,14,8);
}
#define init_6_14_9_15
void FASTCALL call_6_14_9_15(SH2_struct* sh)
{
  SH2extsw(sh,14,9);
}
#define init_6_14_10_15
void FASTCALL call_6_14_10_15(SH2_struct* sh)
{
  SH2extsw(sh,14,10);
}
#define init_6_14_11_15
void FASTCALL call_6_14_11_15(SH2_struct* sh)
{
  SH2extsw(sh,14,11);
}
#define init_6_14_12_15
void FASTCALL call_6_14_12_15(SH2_struct* sh)
{
  SH2extsw(sh,14,12);
}
#define init_6_14_13_15
void FASTCALL call_6_14_13_15(SH2_struct* sh)
{
  SH2extsw(sh,14,13);
}
#define init_6_14_14_15
void FASTCALL call_6_14_14_15(SH2_struct* sh)
{
  SH2extsw(sh,14,14);
}
#define init_6_14_15_15
void FASTCALL call_6_14_15_15(SH2_struct* sh)
{
  SH2extsw(sh,14,15);
}
#define init_6_15_0_15
void FASTCALL call_6_15_0_15(SH2_struct* sh)
{
  SH2extsw(sh,15,0);
}
#define init_6_15_1_15
void FASTCALL call_6_15_1_15(SH2_struct* sh)
{
  SH2extsw(sh,15,1);
}
#define init_6_15_2_15
void FASTCALL call_6_15_2_15(SH2_struct* sh)
{
  SH2extsw(sh,15,2);
}
#define init_6_15_3_15
void FASTCALL call_6_15_3_15(SH2_struct* sh)
{
  SH2extsw(sh,15,3);
}
#define init_6_15_4_15
void FASTCALL call_6_15_4_15(SH2_struct* sh)
{
  SH2extsw(sh,15,4);
}
#define init_6_15_5_15
void FASTCALL call_6_15_5_15(SH2_struct* sh)
{
  SH2extsw(sh,15,5);
}
#define init_6_15_6_15
void FASTCALL call_6_15_6_15(SH2_struct* sh)
{
  SH2extsw(sh,15,6);
}
#define init_6_15_7_15
void FASTCALL call_6_15_7_15(SH2_struct* sh)
{
  SH2extsw(sh,15,7);
}
#define init_6_15_8_15
void FASTCALL call_6_15_8_15(SH2_struct* sh)
{
  SH2extsw(sh,15,8);
}
#define init_6_15_9_15
void FASTCALL call_6_15_9_15(SH2_struct* sh)
{
  SH2extsw(sh,15,9);
}
#define init_6_15_10_15
void FASTCALL call_6_15_10_15(SH2_struct* sh)
{
  SH2extsw(sh,15,10);
}
#define init_6_15_11_15
void FASTCALL call_6_15_11_15(SH2_struct* sh)
{
  SH2extsw(sh,15,11);
}
#define init_6_15_12_15
void FASTCALL call_6_15_12_15(SH2_struct* sh)
{
  SH2extsw(sh,15,12);
}
#define init_6_15_13_15
void FASTCALL call_6_15_13_15(SH2_struct* sh)
{
  SH2extsw(sh,15,13);
}
#define init_6_15_14_15
void FASTCALL call_6_15_14_15(SH2_struct* sh)
{
  SH2extsw(sh,15,14);
}
#define init_6_15_15_15
void FASTCALL call_6_15_15_15(SH2_struct* sh)
{
  SH2extsw(sh,15,15);
}
#define init_6_0_0_14
void FASTCALL call_6_0_0_14(SH2_struct* sh)
{
  SH2extsb(sh,0,0);
}
#define init_6_0_1_14
void FASTCALL call_6_0_1_14(SH2_struct* sh)
{
  SH2extsb(sh,0,1);
}
#define init_6_0_2_14
void FASTCALL call_6_0_2_14(SH2_struct* sh)
{
  SH2extsb(sh,0,2);
}
#define init_6_0_3_14
void FASTCALL call_6_0_3_14(SH2_struct* sh)
{
  SH2extsb(sh,0,3);
}
#define init_6_0_4_14
void FASTCALL call_6_0_4_14(SH2_struct* sh)
{
  SH2extsb(sh,0,4);
}
#define init_6_0_5_14
void FASTCALL call_6_0_5_14(SH2_struct* sh)
{
  SH2extsb(sh,0,5);
}
#define init_6_0_6_14
void FASTCALL call_6_0_6_14(SH2_struct* sh)
{
  SH2extsb(sh,0,6);
}
#define init_6_0_7_14
void FASTCALL call_6_0_7_14(SH2_struct* sh)
{
  SH2extsb(sh,0,7);
}
#define init_6_0_8_14
void FASTCALL call_6_0_8_14(SH2_struct* sh)
{
  SH2extsb(sh,0,8);
}
#define init_6_0_9_14
void FASTCALL call_6_0_9_14(SH2_struct* sh)
{
  SH2extsb(sh,0,9);
}
#define init_6_0_10_14
void FASTCALL call_6_0_10_14(SH2_struct* sh)
{
  SH2extsb(sh,0,10);
}
#define init_6_0_11_14
void FASTCALL call_6_0_11_14(SH2_struct* sh)
{
  SH2extsb(sh,0,11);
}
#define init_6_0_12_14
void FASTCALL call_6_0_12_14(SH2_struct* sh)
{
  SH2extsb(sh,0,12);
}
#define init_6_0_13_14
void FASTCALL call_6_0_13_14(SH2_struct* sh)
{
  SH2extsb(sh,0,13);
}
#define init_6_0_14_14
void FASTCALL call_6_0_14_14(SH2_struct* sh)
{
  SH2extsb(sh,0,14);
}
#define init_6_0_15_14
void FASTCALL call_6_0_15_14(SH2_struct* sh)
{
  SH2extsb(sh,0,15);
}
#define init_6_1_0_14
void FASTCALL call_6_1_0_14(SH2_struct* sh)
{
  SH2extsb(sh,1,0);
}
#define init_6_1_1_14
void FASTCALL call_6_1_1_14(SH2_struct* sh)
{
  SH2extsb(sh,1,1);
}
#define init_6_1_2_14
void FASTCALL call_6_1_2_14(SH2_struct* sh)
{
  SH2extsb(sh,1,2);
}
#define init_6_1_3_14
void FASTCALL call_6_1_3_14(SH2_struct* sh)
{
  SH2extsb(sh,1,3);
}
#define init_6_1_4_14
void FASTCALL call_6_1_4_14(SH2_struct* sh)
{
  SH2extsb(sh,1,4);
}
#define init_6_1_5_14
void FASTCALL call_6_1_5_14(SH2_struct* sh)
{
  SH2extsb(sh,1,5);
}
#define init_6_1_6_14
void FASTCALL call_6_1_6_14(SH2_struct* sh)
{
  SH2extsb(sh,1,6);
}
#define init_6_1_7_14
void FASTCALL call_6_1_7_14(SH2_struct* sh)
{
  SH2extsb(sh,1,7);
}
#define init_6_1_8_14
void FASTCALL call_6_1_8_14(SH2_struct* sh)
{
  SH2extsb(sh,1,8);
}
#define init_6_1_9_14
void FASTCALL call_6_1_9_14(SH2_struct* sh)
{
  SH2extsb(sh,1,9);
}
#define init_6_1_10_14
void FASTCALL call_6_1_10_14(SH2_struct* sh)
{
  SH2extsb(sh,1,10);
}
#define init_6_1_11_14
void FASTCALL call_6_1_11_14(SH2_struct* sh)
{
  SH2extsb(sh,1,11);
}
#define init_6_1_12_14
void FASTCALL call_6_1_12_14(SH2_struct* sh)
{
  SH2extsb(sh,1,12);
}
#define init_6_1_13_14
void FASTCALL call_6_1_13_14(SH2_struct* sh)
{
  SH2extsb(sh,1,13);
}
#define init_6_1_14_14
void FASTCALL call_6_1_14_14(SH2_struct* sh)
{
  SH2extsb(sh,1,14);
}
#define init_6_1_15_14
void FASTCALL call_6_1_15_14(SH2_struct* sh)
{
  SH2extsb(sh,1,15);
}
#define init_6_2_0_14
void FASTCALL call_6_2_0_14(SH2_struct* sh)
{
  SH2extsb(sh,2,0);
}
#define init_6_2_1_14
void FASTCALL call_6_2_1_14(SH2_struct* sh)
{
  SH2extsb(sh,2,1);
}
#define init_6_2_2_14
void FASTCALL call_6_2_2_14(SH2_struct* sh)
{
  SH2extsb(sh,2,2);
}
#define init_6_2_3_14
void FASTCALL call_6_2_3_14(SH2_struct* sh)
{
  SH2extsb(sh,2,3);
}
#define init_6_2_4_14
void FASTCALL call_6_2_4_14(SH2_struct* sh)
{
  SH2extsb(sh,2,4);
}
#define init_6_2_5_14
void FASTCALL call_6_2_5_14(SH2_struct* sh)
{
  SH2extsb(sh,2,5);
}
#define init_6_2_6_14
void FASTCALL call_6_2_6_14(SH2_struct* sh)
{
  SH2extsb(sh,2,6);
}
#define init_6_2_7_14
void FASTCALL call_6_2_7_14(SH2_struct* sh)
{
  SH2extsb(sh,2,7);
}
#define init_6_2_8_14
void FASTCALL call_6_2_8_14(SH2_struct* sh)
{
  SH2extsb(sh,2,8);
}
#define init_6_2_9_14
void FASTCALL call_6_2_9_14(SH2_struct* sh)
{
  SH2extsb(sh,2,9);
}
#define init_6_2_10_14
void FASTCALL call_6_2_10_14(SH2_struct* sh)
{
  SH2extsb(sh,2,10);
}
#define init_6_2_11_14
void FASTCALL call_6_2_11_14(SH2_struct* sh)
{
  SH2extsb(sh,2,11);
}
#define init_6_2_12_14
void FASTCALL call_6_2_12_14(SH2_struct* sh)
{
  SH2extsb(sh,2,12);
}
#define init_6_2_13_14
void FASTCALL call_6_2_13_14(SH2_struct* sh)
{
  SH2extsb(sh,2,13);
}
#define init_6_2_14_14
void FASTCALL call_6_2_14_14(SH2_struct* sh)
{
  SH2extsb(sh,2,14);
}
#define init_6_2_15_14
void FASTCALL call_6_2_15_14(SH2_struct* sh)
{
  SH2extsb(sh,2,15);
}
#define init_6_3_0_14
void FASTCALL call_6_3_0_14(SH2_struct* sh)
{
  SH2extsb(sh,3,0);
}
#define init_6_3_1_14
void FASTCALL call_6_3_1_14(SH2_struct* sh)
{
  SH2extsb(sh,3,1);
}
#define init_6_3_2_14
void FASTCALL call_6_3_2_14(SH2_struct* sh)
{
  SH2extsb(sh,3,2);
}
#define init_6_3_3_14
void FASTCALL call_6_3_3_14(SH2_struct* sh)
{
  SH2extsb(sh,3,3);
}
#define init_6_3_4_14
void FASTCALL call_6_3_4_14(SH2_struct* sh)
{
  SH2extsb(sh,3,4);
}
#define init_6_3_5_14
void FASTCALL call_6_3_5_14(SH2_struct* sh)
{
  SH2extsb(sh,3,5);
}
#define init_6_3_6_14
void FASTCALL call_6_3_6_14(SH2_struct* sh)
{
  SH2extsb(sh,3,6);
}
#define init_6_3_7_14
void FASTCALL call_6_3_7_14(SH2_struct* sh)
{
  SH2extsb(sh,3,7);
}
#define init_6_3_8_14
void FASTCALL call_6_3_8_14(SH2_struct* sh)
{
  SH2extsb(sh,3,8);
}
#define init_6_3_9_14
void FASTCALL call_6_3_9_14(SH2_struct* sh)
{
  SH2extsb(sh,3,9);
}
#define init_6_3_10_14
void FASTCALL call_6_3_10_14(SH2_struct* sh)
{
  SH2extsb(sh,3,10);
}
#define init_6_3_11_14
void FASTCALL call_6_3_11_14(SH2_struct* sh)
{
  SH2extsb(sh,3,11);
}
#define init_6_3_12_14
void FASTCALL call_6_3_12_14(SH2_struct* sh)
{
  SH2extsb(sh,3,12);
}
#define init_6_3_13_14
void FASTCALL call_6_3_13_14(SH2_struct* sh)
{
  SH2extsb(sh,3,13);
}
#define init_6_3_14_14
void FASTCALL call_6_3_14_14(SH2_struct* sh)
{
  SH2extsb(sh,3,14);
}
#define init_6_3_15_14
void FASTCALL call_6_3_15_14(SH2_struct* sh)
{
  SH2extsb(sh,3,15);
}
#define init_6_4_0_14
void FASTCALL call_6_4_0_14(SH2_struct* sh)
{
  SH2extsb(sh,4,0);
}
#define init_6_4_1_14
void FASTCALL call_6_4_1_14(SH2_struct* sh)
{
  SH2extsb(sh,4,1);
}
#define init_6_4_2_14
void FASTCALL call_6_4_2_14(SH2_struct* sh)
{
  SH2extsb(sh,4,2);
}
#define init_6_4_3_14
void FASTCALL call_6_4_3_14(SH2_struct* sh)
{
  SH2extsb(sh,4,3);
}
#define init_6_4_4_14
void FASTCALL call_6_4_4_14(SH2_struct* sh)
{
  SH2extsb(sh,4,4);
}
#define init_6_4_5_14
void FASTCALL call_6_4_5_14(SH2_struct* sh)
{
  SH2extsb(sh,4,5);
}
#define init_6_4_6_14
void FASTCALL call_6_4_6_14(SH2_struct* sh)
{
  SH2extsb(sh,4,6);
}
#define init_6_4_7_14
void FASTCALL call_6_4_7_14(SH2_struct* sh)
{
  SH2extsb(sh,4,7);
}
#define init_6_4_8_14
void FASTCALL call_6_4_8_14(SH2_struct* sh)
{
  SH2extsb(sh,4,8);
}
#define init_6_4_9_14
void FASTCALL call_6_4_9_14(SH2_struct* sh)
{
  SH2extsb(sh,4,9);
}
#define init_6_4_10_14
void FASTCALL call_6_4_10_14(SH2_struct* sh)
{
  SH2extsb(sh,4,10);
}
#define init_6_4_11_14
void FASTCALL call_6_4_11_14(SH2_struct* sh)
{
  SH2extsb(sh,4,11);
}
#define init_6_4_12_14
void FASTCALL call_6_4_12_14(SH2_struct* sh)
{
  SH2extsb(sh,4,12);
}
#define init_6_4_13_14
void FASTCALL call_6_4_13_14(SH2_struct* sh)
{
  SH2extsb(sh,4,13);
}
#define init_6_4_14_14
void FASTCALL call_6_4_14_14(SH2_struct* sh)
{
  SH2extsb(sh,4,14);
}
#define init_6_4_15_14
void FASTCALL call_6_4_15_14(SH2_struct* sh)
{
  SH2extsb(sh,4,15);
}
#define init_6_5_0_14
void FASTCALL call_6_5_0_14(SH2_struct* sh)
{
  SH2extsb(sh,5,0);
}
#define init_6_5_1_14
void FASTCALL call_6_5_1_14(SH2_struct* sh)
{
  SH2extsb(sh,5,1);
}
#define init_6_5_2_14
void FASTCALL call_6_5_2_14(SH2_struct* sh)
{
  SH2extsb(sh,5,2);
}
#define init_6_5_3_14
void FASTCALL call_6_5_3_14(SH2_struct* sh)
{
  SH2extsb(sh,5,3);
}
#define init_6_5_4_14
void FASTCALL call_6_5_4_14(SH2_struct* sh)
{
  SH2extsb(sh,5,4);
}
#define init_6_5_5_14
void FASTCALL call_6_5_5_14(SH2_struct* sh)
{
  SH2extsb(sh,5,5);
}
#define init_6_5_6_14
void FASTCALL call_6_5_6_14(SH2_struct* sh)
{
  SH2extsb(sh,5,6);
}
#define init_6_5_7_14
void FASTCALL call_6_5_7_14(SH2_struct* sh)
{
  SH2extsb(sh,5,7);
}
#define init_6_5_8_14
void FASTCALL call_6_5_8_14(SH2_struct* sh)
{
  SH2extsb(sh,5,8);
}
#define init_6_5_9_14
void FASTCALL call_6_5_9_14(SH2_struct* sh)
{
  SH2extsb(sh,5,9);
}
#define init_6_5_10_14
void FASTCALL call_6_5_10_14(SH2_struct* sh)
{
  SH2extsb(sh,5,10);
}
#define init_6_5_11_14
void FASTCALL call_6_5_11_14(SH2_struct* sh)
{
  SH2extsb(sh,5,11);
}
#define init_6_5_12_14
void FASTCALL call_6_5_12_14(SH2_struct* sh)
{
  SH2extsb(sh,5,12);
}
#define init_6_5_13_14
void FASTCALL call_6_5_13_14(SH2_struct* sh)
{
  SH2extsb(sh,5,13);
}
#define init_6_5_14_14
void FASTCALL call_6_5_14_14(SH2_struct* sh)
{
  SH2extsb(sh,5,14);
}
#define init_6_5_15_14
void FASTCALL call_6_5_15_14(SH2_struct* sh)
{
  SH2extsb(sh,5,15);
}
#define init_6_6_0_14
void FASTCALL call_6_6_0_14(SH2_struct* sh)
{
  SH2extsb(sh,6,0);
}
#define init_6_6_1_14
void FASTCALL call_6_6_1_14(SH2_struct* sh)
{
  SH2extsb(sh,6,1);
}
#define init_6_6_2_14
void FASTCALL call_6_6_2_14(SH2_struct* sh)
{
  SH2extsb(sh,6,2);
}
#define init_6_6_3_14
void FASTCALL call_6_6_3_14(SH2_struct* sh)
{
  SH2extsb(sh,6,3);
}
#define init_6_6_4_14
void FASTCALL call_6_6_4_14(SH2_struct* sh)
{
  SH2extsb(sh,6,4);
}
#define init_6_6_5_14
void FASTCALL call_6_6_5_14(SH2_struct* sh)
{
  SH2extsb(sh,6,5);
}
#define init_6_6_6_14
void FASTCALL call_6_6_6_14(SH2_struct* sh)
{
  SH2extsb(sh,6,6);
}
#define init_6_6_7_14
void FASTCALL call_6_6_7_14(SH2_struct* sh)
{
  SH2extsb(sh,6,7);
}
#define init_6_6_8_14
void FASTCALL call_6_6_8_14(SH2_struct* sh)
{
  SH2extsb(sh,6,8);
}
#define init_6_6_9_14
void FASTCALL call_6_6_9_14(SH2_struct* sh)
{
  SH2extsb(sh,6,9);
}
#define init_6_6_10_14
void FASTCALL call_6_6_10_14(SH2_struct* sh)
{
  SH2extsb(sh,6,10);
}
#define init_6_6_11_14
void FASTCALL call_6_6_11_14(SH2_struct* sh)
{
  SH2extsb(sh,6,11);
}
#define init_6_6_12_14
void FASTCALL call_6_6_12_14(SH2_struct* sh)
{
  SH2extsb(sh,6,12);
}
#define init_6_6_13_14
void FASTCALL call_6_6_13_14(SH2_struct* sh)
{
  SH2extsb(sh,6,13);
}
#define init_6_6_14_14
void FASTCALL call_6_6_14_14(SH2_struct* sh)
{
  SH2extsb(sh,6,14);
}
#define init_6_6_15_14
void FASTCALL call_6_6_15_14(SH2_struct* sh)
{
  SH2extsb(sh,6,15);
}
#define init_6_7_0_14
void FASTCALL call_6_7_0_14(SH2_struct* sh)
{
  SH2extsb(sh,7,0);
}
#define init_6_7_1_14
void FASTCALL call_6_7_1_14(SH2_struct* sh)
{
  SH2extsb(sh,7,1);
}
#define init_6_7_2_14
void FASTCALL call_6_7_2_14(SH2_struct* sh)
{
  SH2extsb(sh,7,2);
}
#define init_6_7_3_14
void FASTCALL call_6_7_3_14(SH2_struct* sh)
{
  SH2extsb(sh,7,3);
}
#define init_6_7_4_14
void FASTCALL call_6_7_4_14(SH2_struct* sh)
{
  SH2extsb(sh,7,4);
}
#define init_6_7_5_14
void FASTCALL call_6_7_5_14(SH2_struct* sh)
{
  SH2extsb(sh,7,5);
}
#define init_6_7_6_14
void FASTCALL call_6_7_6_14(SH2_struct* sh)
{
  SH2extsb(sh,7,6);
}
#define init_6_7_7_14
void FASTCALL call_6_7_7_14(SH2_struct* sh)
{
  SH2extsb(sh,7,7);
}
#define init_6_7_8_14
void FASTCALL call_6_7_8_14(SH2_struct* sh)
{
  SH2extsb(sh,7,8);
}
#define init_6_7_9_14
void FASTCALL call_6_7_9_14(SH2_struct* sh)
{
  SH2extsb(sh,7,9);
}
#define init_6_7_10_14
void FASTCALL call_6_7_10_14(SH2_struct* sh)
{
  SH2extsb(sh,7,10);
}
#define init_6_7_11_14
void FASTCALL call_6_7_11_14(SH2_struct* sh)
{
  SH2extsb(sh,7,11);
}
#define init_6_7_12_14
void FASTCALL call_6_7_12_14(SH2_struct* sh)
{
  SH2extsb(sh,7,12);
}
#define init_6_7_13_14
void FASTCALL call_6_7_13_14(SH2_struct* sh)
{
  SH2extsb(sh,7,13);
}
#define init_6_7_14_14
void FASTCALL call_6_7_14_14(SH2_struct* sh)
{
  SH2extsb(sh,7,14);
}
#define init_6_7_15_14
void FASTCALL call_6_7_15_14(SH2_struct* sh)
{
  SH2extsb(sh,7,15);
}
#define init_6_8_0_14
void FASTCALL call_6_8_0_14(SH2_struct* sh)
{
  SH2extsb(sh,8,0);
}
#define init_6_8_1_14
void FASTCALL call_6_8_1_14(SH2_struct* sh)
{
  SH2extsb(sh,8,1);
}
#define init_6_8_2_14
void FASTCALL call_6_8_2_14(SH2_struct* sh)
{
  SH2extsb(sh,8,2);
}
#define init_6_8_3_14
void FASTCALL call_6_8_3_14(SH2_struct* sh)
{
  SH2extsb(sh,8,3);
}
#define init_6_8_4_14
void FASTCALL call_6_8_4_14(SH2_struct* sh)
{
  SH2extsb(sh,8,4);
}
#define init_6_8_5_14
void FASTCALL call_6_8_5_14(SH2_struct* sh)
{
  SH2extsb(sh,8,5);
}
#define init_6_8_6_14
void FASTCALL call_6_8_6_14(SH2_struct* sh)
{
  SH2extsb(sh,8,6);
}
#define init_6_8_7_14
void FASTCALL call_6_8_7_14(SH2_struct* sh)
{
  SH2extsb(sh,8,7);
}
#define init_6_8_8_14
void FASTCALL call_6_8_8_14(SH2_struct* sh)
{
  SH2extsb(sh,8,8);
}
#define init_6_8_9_14
void FASTCALL call_6_8_9_14(SH2_struct* sh)
{
  SH2extsb(sh,8,9);
}
#define init_6_8_10_14
void FASTCALL call_6_8_10_14(SH2_struct* sh)
{
  SH2extsb(sh,8,10);
}
#define init_6_8_11_14
void FASTCALL call_6_8_11_14(SH2_struct* sh)
{
  SH2extsb(sh,8,11);
}
#define init_6_8_12_14
void FASTCALL call_6_8_12_14(SH2_struct* sh)
{
  SH2extsb(sh,8,12);
}
#define init_6_8_13_14
void FASTCALL call_6_8_13_14(SH2_struct* sh)
{
  SH2extsb(sh,8,13);
}
#define init_6_8_14_14
void FASTCALL call_6_8_14_14(SH2_struct* sh)
{
  SH2extsb(sh,8,14);
}
#define init_6_8_15_14
void FASTCALL call_6_8_15_14(SH2_struct* sh)
{
  SH2extsb(sh,8,15);
}
#define init_6_9_0_14
void FASTCALL call_6_9_0_14(SH2_struct* sh)
{
  SH2extsb(sh,9,0);
}
#define init_6_9_1_14
void FASTCALL call_6_9_1_14(SH2_struct* sh)
{
  SH2extsb(sh,9,1);
}
#define init_6_9_2_14
void FASTCALL call_6_9_2_14(SH2_struct* sh)
{
  SH2extsb(sh,9,2);
}
#define init_6_9_3_14
void FASTCALL call_6_9_3_14(SH2_struct* sh)
{
  SH2extsb(sh,9,3);
}
#define init_6_9_4_14
void FASTCALL call_6_9_4_14(SH2_struct* sh)
{
  SH2extsb(sh,9,4);
}
#define init_6_9_5_14
void FASTCALL call_6_9_5_14(SH2_struct* sh)
{
  SH2extsb(sh,9,5);
}
#define init_6_9_6_14
void FASTCALL call_6_9_6_14(SH2_struct* sh)
{
  SH2extsb(sh,9,6);
}
#define init_6_9_7_14
void FASTCALL call_6_9_7_14(SH2_struct* sh)
{
  SH2extsb(sh,9,7);
}
#define init_6_9_8_14
void FASTCALL call_6_9_8_14(SH2_struct* sh)
{
  SH2extsb(sh,9,8);
}
#define init_6_9_9_14
void FASTCALL call_6_9_9_14(SH2_struct* sh)
{
  SH2extsb(sh,9,9);
}
#define init_6_9_10_14
void FASTCALL call_6_9_10_14(SH2_struct* sh)
{
  SH2extsb(sh,9,10);
}
#define init_6_9_11_14
void FASTCALL call_6_9_11_14(SH2_struct* sh)
{
  SH2extsb(sh,9,11);
}
#define init_6_9_12_14
void FASTCALL call_6_9_12_14(SH2_struct* sh)
{
  SH2extsb(sh,9,12);
}
#define init_6_9_13_14
void FASTCALL call_6_9_13_14(SH2_struct* sh)
{
  SH2extsb(sh,9,13);
}
#define init_6_9_14_14
void FASTCALL call_6_9_14_14(SH2_struct* sh)
{
  SH2extsb(sh,9,14);
}
#define init_6_9_15_14
void FASTCALL call_6_9_15_14(SH2_struct* sh)
{
  SH2extsb(sh,9,15);
}
#define init_6_10_0_14
void FASTCALL call_6_10_0_14(SH2_struct* sh)
{
  SH2extsb(sh,10,0);
}
#define init_6_10_1_14
void FASTCALL call_6_10_1_14(SH2_struct* sh)
{
  SH2extsb(sh,10,1);
}
#define init_6_10_2_14
void FASTCALL call_6_10_2_14(SH2_struct* sh)
{
  SH2extsb(sh,10,2);
}
#define init_6_10_3_14
void FASTCALL call_6_10_3_14(SH2_struct* sh)
{
  SH2extsb(sh,10,3);
}
#define init_6_10_4_14
void FASTCALL call_6_10_4_14(SH2_struct* sh)
{
  SH2extsb(sh,10,4);
}
#define init_6_10_5_14
void FASTCALL call_6_10_5_14(SH2_struct* sh)
{
  SH2extsb(sh,10,5);
}
#define init_6_10_6_14
void FASTCALL call_6_10_6_14(SH2_struct* sh)
{
  SH2extsb(sh,10,6);
}
#define init_6_10_7_14
void FASTCALL call_6_10_7_14(SH2_struct* sh)
{
  SH2extsb(sh,10,7);
}
#define init_6_10_8_14
void FASTCALL call_6_10_8_14(SH2_struct* sh)
{
  SH2extsb(sh,10,8);
}
#define init_6_10_9_14
void FASTCALL call_6_10_9_14(SH2_struct* sh)
{
  SH2extsb(sh,10,9);
}
#define init_6_10_10_14
void FASTCALL call_6_10_10_14(SH2_struct* sh)
{
  SH2extsb(sh,10,10);
}
#define init_6_10_11_14
void FASTCALL call_6_10_11_14(SH2_struct* sh)
{
  SH2extsb(sh,10,11);
}
#define init_6_10_12_14
void FASTCALL call_6_10_12_14(SH2_struct* sh)
{
  SH2extsb(sh,10,12);
}
#define init_6_10_13_14
void FASTCALL call_6_10_13_14(SH2_struct* sh)
{
  SH2extsb(sh,10,13);
}
#define init_6_10_14_14
void FASTCALL call_6_10_14_14(SH2_struct* sh)
{
  SH2extsb(sh,10,14);
}
#define init_6_10_15_14
void FASTCALL call_6_10_15_14(SH2_struct* sh)
{
  SH2extsb(sh,10,15);
}
#define init_6_11_0_14
void FASTCALL call_6_11_0_14(SH2_struct* sh)
{
  SH2extsb(sh,11,0);
}
#define init_6_11_1_14
void FASTCALL call_6_11_1_14(SH2_struct* sh)
{
  SH2extsb(sh,11,1);
}
#define init_6_11_2_14
void FASTCALL call_6_11_2_14(SH2_struct* sh)
{
  SH2extsb(sh,11,2);
}
#define init_6_11_3_14
void FASTCALL call_6_11_3_14(SH2_struct* sh)
{
  SH2extsb(sh,11,3);
}
#define init_6_11_4_14
void FASTCALL call_6_11_4_14(SH2_struct* sh)
{
  SH2extsb(sh,11,4);
}
#define init_6_11_5_14
void FASTCALL call_6_11_5_14(SH2_struct* sh)
{
  SH2extsb(sh,11,5);
}
#define init_6_11_6_14
void FASTCALL call_6_11_6_14(SH2_struct* sh)
{
  SH2extsb(sh,11,6);
}
#define init_6_11_7_14
void FASTCALL call_6_11_7_14(SH2_struct* sh)
{
  SH2extsb(sh,11,7);
}
#define init_6_11_8_14
void FASTCALL call_6_11_8_14(SH2_struct* sh)
{
  SH2extsb(sh,11,8);
}
#define init_6_11_9_14
void FASTCALL call_6_11_9_14(SH2_struct* sh)
{
  SH2extsb(sh,11,9);
}
#define init_6_11_10_14
void FASTCALL call_6_11_10_14(SH2_struct* sh)
{
  SH2extsb(sh,11,10);
}
#define init_6_11_11_14
void FASTCALL call_6_11_11_14(SH2_struct* sh)
{
  SH2extsb(sh,11,11);
}
#define init_6_11_12_14
void FASTCALL call_6_11_12_14(SH2_struct* sh)
{
  SH2extsb(sh,11,12);
}
#define init_6_11_13_14
void FASTCALL call_6_11_13_14(SH2_struct* sh)
{
  SH2extsb(sh,11,13);
}
#define init_6_11_14_14
void FASTCALL call_6_11_14_14(SH2_struct* sh)
{
  SH2extsb(sh,11,14);
}
#define init_6_11_15_14
void FASTCALL call_6_11_15_14(SH2_struct* sh)
{
  SH2extsb(sh,11,15);
}
#define init_6_12_0_14
void FASTCALL call_6_12_0_14(SH2_struct* sh)
{
  SH2extsb(sh,12,0);
}
#define init_6_12_1_14
void FASTCALL call_6_12_1_14(SH2_struct* sh)
{
  SH2extsb(sh,12,1);
}
#define init_6_12_2_14
void FASTCALL call_6_12_2_14(SH2_struct* sh)
{
  SH2extsb(sh,12,2);
}
#define init_6_12_3_14
void FASTCALL call_6_12_3_14(SH2_struct* sh)
{
  SH2extsb(sh,12,3);
}
#define init_6_12_4_14
void FASTCALL call_6_12_4_14(SH2_struct* sh)
{
  SH2extsb(sh,12,4);
}
#define init_6_12_5_14
void FASTCALL call_6_12_5_14(SH2_struct* sh)
{
  SH2extsb(sh,12,5);
}
#define init_6_12_6_14
void FASTCALL call_6_12_6_14(SH2_struct* sh)
{
  SH2extsb(sh,12,6);
}
#define init_6_12_7_14
void FASTCALL call_6_12_7_14(SH2_struct* sh)
{
  SH2extsb(sh,12,7);
}
#define init_6_12_8_14
void FASTCALL call_6_12_8_14(SH2_struct* sh)
{
  SH2extsb(sh,12,8);
}
#define init_6_12_9_14
void FASTCALL call_6_12_9_14(SH2_struct* sh)
{
  SH2extsb(sh,12,9);
}
#define init_6_12_10_14
void FASTCALL call_6_12_10_14(SH2_struct* sh)
{
  SH2extsb(sh,12,10);
}
#define init_6_12_11_14
void FASTCALL call_6_12_11_14(SH2_struct* sh)
{
  SH2extsb(sh,12,11);
}
#define init_6_12_12_14
void FASTCALL call_6_12_12_14(SH2_struct* sh)
{
  SH2extsb(sh,12,12);
}
#define init_6_12_13_14
void FASTCALL call_6_12_13_14(SH2_struct* sh)
{
  SH2extsb(sh,12,13);
}
#define init_6_12_14_14
void FASTCALL call_6_12_14_14(SH2_struct* sh)
{
  SH2extsb(sh,12,14);
}
#define init_6_12_15_14
void FASTCALL call_6_12_15_14(SH2_struct* sh)
{
  SH2extsb(sh,12,15);
}
#define init_6_13_0_14
void FASTCALL call_6_13_0_14(SH2_struct* sh)
{
  SH2extsb(sh,13,0);
}
#define init_6_13_1_14
void FASTCALL call_6_13_1_14(SH2_struct* sh)
{
  SH2extsb(sh,13,1);
}
#define init_6_13_2_14
void FASTCALL call_6_13_2_14(SH2_struct* sh)
{
  SH2extsb(sh,13,2);
}
#define init_6_13_3_14
void FASTCALL call_6_13_3_14(SH2_struct* sh)
{
  SH2extsb(sh,13,3);
}
#define init_6_13_4_14
void FASTCALL call_6_13_4_14(SH2_struct* sh)
{
  SH2extsb(sh,13,4);
}
#define init_6_13_5_14
void FASTCALL call_6_13_5_14(SH2_struct* sh)
{
  SH2extsb(sh,13,5);
}
#define init_6_13_6_14
void FASTCALL call_6_13_6_14(SH2_struct* sh)
{
  SH2extsb(sh,13,6);
}
#define init_6_13_7_14
void FASTCALL call_6_13_7_14(SH2_struct* sh)
{
  SH2extsb(sh,13,7);
}
#define init_6_13_8_14
void FASTCALL call_6_13_8_14(SH2_struct* sh)
{
  SH2extsb(sh,13,8);
}
#define init_6_13_9_14
void FASTCALL call_6_13_9_14(SH2_struct* sh)
{
  SH2extsb(sh,13,9);
}
#define init_6_13_10_14
void FASTCALL call_6_13_10_14(SH2_struct* sh)
{
  SH2extsb(sh,13,10);
}
#define init_6_13_11_14
void FASTCALL call_6_13_11_14(SH2_struct* sh)
{
  SH2extsb(sh,13,11);
}
#define init_6_13_12_14
void FASTCALL call_6_13_12_14(SH2_struct* sh)
{
  SH2extsb(sh,13,12);
}
#define init_6_13_13_14
void FASTCALL call_6_13_13_14(SH2_struct* sh)
{
  SH2extsb(sh,13,13);
}
#define init_6_13_14_14
void FASTCALL call_6_13_14_14(SH2_struct* sh)
{
  SH2extsb(sh,13,14);
}
#define init_6_13_15_14
void FASTCALL call_6_13_15_14(SH2_struct* sh)
{
  SH2extsb(sh,13,15);
}
#define init_6_14_0_14
void FASTCALL call_6_14_0_14(SH2_struct* sh)
{
  SH2extsb(sh,14,0);
}
#define init_6_14_1_14
void FASTCALL call_6_14_1_14(SH2_struct* sh)
{
  SH2extsb(sh,14,1);
}
#define init_6_14_2_14
void FASTCALL call_6_14_2_14(SH2_struct* sh)
{
  SH2extsb(sh,14,2);
}
#define init_6_14_3_14
void FASTCALL call_6_14_3_14(SH2_struct* sh)
{
  SH2extsb(sh,14,3);
}
#define init_6_14_4_14
void FASTCALL call_6_14_4_14(SH2_struct* sh)
{
  SH2extsb(sh,14,4);
}
#define init_6_14_5_14
void FASTCALL call_6_14_5_14(SH2_struct* sh)
{
  SH2extsb(sh,14,5);
}
#define init_6_14_6_14
void FASTCALL call_6_14_6_14(SH2_struct* sh)
{
  SH2extsb(sh,14,6);
}
#define init_6_14_7_14
void FASTCALL call_6_14_7_14(SH2_struct* sh)
{
  SH2extsb(sh,14,7);
}
#define init_6_14_8_14
void FASTCALL call_6_14_8_14(SH2_struct* sh)
{
  SH2extsb(sh,14,8);
}
#define init_6_14_9_14
void FASTCALL call_6_14_9_14(SH2_struct* sh)
{
  SH2extsb(sh,14,9);
}
#define init_6_14_10_14
void FASTCALL call_6_14_10_14(SH2_struct* sh)
{
  SH2extsb(sh,14,10);
}
#define init_6_14_11_14
void FASTCALL call_6_14_11_14(SH2_struct* sh)
{
  SH2extsb(sh,14,11);
}
#define init_6_14_12_14
void FASTCALL call_6_14_12_14(SH2_struct* sh)
{
  SH2extsb(sh,14,12);
}
#define init_6_14_13_14
void FASTCALL call_6_14_13_14(SH2_struct* sh)
{
  SH2extsb(sh,14,13);
}
#define init_6_14_14_14
void FASTCALL call_6_14_14_14(SH2_struct* sh)
{
  SH2extsb(sh,14,14);
}
#define init_6_14_15_14
void FASTCALL call_6_14_15_14(SH2_struct* sh)
{
  SH2extsb(sh,14,15);
}
#define init_6_15_0_14
void FASTCALL call_6_15_0_14(SH2_struct* sh)
{
  SH2extsb(sh,15,0);
}
#define init_6_15_1_14
void FASTCALL call_6_15_1_14(SH2_struct* sh)
{
  SH2extsb(sh,15,1);
}
#define init_6_15_2_14
void FASTCALL call_6_15_2_14(SH2_struct* sh)
{
  SH2extsb(sh,15,2);
}
#define init_6_15_3_14
void FASTCALL call_6_15_3_14(SH2_struct* sh)
{
  SH2extsb(sh,15,3);
}
#define init_6_15_4_14
void FASTCALL call_6_15_4_14(SH2_struct* sh)
{
  SH2extsb(sh,15,4);
}
#define init_6_15_5_14
void FASTCALL call_6_15_5_14(SH2_struct* sh)
{
  SH2extsb(sh,15,5);
}
#define init_6_15_6_14
void FASTCALL call_6_15_6_14(SH2_struct* sh)
{
  SH2extsb(sh,15,6);
}
#define init_6_15_7_14
void FASTCALL call_6_15_7_14(SH2_struct* sh)
{
  SH2extsb(sh,15,7);
}
#define init_6_15_8_14
void FASTCALL call_6_15_8_14(SH2_struct* sh)
{
  SH2extsb(sh,15,8);
}
#define init_6_15_9_14
void FASTCALL call_6_15_9_14(SH2_struct* sh)
{
  SH2extsb(sh,15,9);
}
#define init_6_15_10_14
void FASTCALL call_6_15_10_14(SH2_struct* sh)
{
  SH2extsb(sh,15,10);
}
#define init_6_15_11_14
void FASTCALL call_6_15_11_14(SH2_struct* sh)
{
  SH2extsb(sh,15,11);
}
#define init_6_15_12_14
void FASTCALL call_6_15_12_14(SH2_struct* sh)
{
  SH2extsb(sh,15,12);
}
#define init_6_15_13_14
void FASTCALL call_6_15_13_14(SH2_struct* sh)
{
  SH2extsb(sh,15,13);
}
#define init_6_15_14_14
void FASTCALL call_6_15_14_14(SH2_struct* sh)
{
  SH2extsb(sh,15,14);
}
#define init_6_15_15_14
void FASTCALL call_6_15_15_14(SH2_struct* sh)
{
  SH2extsb(sh,15,15);
}
#define init_4_0_1_0
void FASTCALL call_4_0_1_0(SH2_struct* sh)
{
  SH2dt(sh,0);
}
#define init_4_1_1_0
void FASTCALL call_4_1_1_0(SH2_struct* sh)
{
  SH2dt(sh,1);
}
#define init_4_2_1_0
void FASTCALL call_4_2_1_0(SH2_struct* sh)
{
  SH2dt(sh,2);
}
#define init_4_3_1_0
void FASTCALL call_4_3_1_0(SH2_struct* sh)
{
  SH2dt(sh,3);
}
#define init_4_4_1_0
void FASTCALL call_4_4_1_0(SH2_struct* sh)
{
  SH2dt(sh,4);
}
#define init_4_5_1_0
void FASTCALL call_4_5_1_0(SH2_struct* sh)
{
  SH2dt(sh,5);
}
#define init_4_6_1_0
void FASTCALL call_4_6_1_0(SH2_struct* sh)
{
  SH2dt(sh,6);
}
#define init_4_7_1_0
void FASTCALL call_4_7_1_0(SH2_struct* sh)
{
  SH2dt(sh,7);
}
#define init_4_8_1_0
void FASTCALL call_4_8_1_0(SH2_struct* sh)
{
  SH2dt(sh,8);
}
#define init_4_9_1_0
void FASTCALL call_4_9_1_0(SH2_struct* sh)
{
  SH2dt(sh,9);
}
#define init_4_10_1_0
void FASTCALL call_4_10_1_0(SH2_struct* sh)
{
  SH2dt(sh,10);
}
#define init_4_11_1_0
void FASTCALL call_4_11_1_0(SH2_struct* sh)
{
  SH2dt(sh,11);
}
#define init_4_12_1_0
void FASTCALL call_4_12_1_0(SH2_struct* sh)
{
  SH2dt(sh,12);
}
#define init_4_13_1_0
void FASTCALL call_4_13_1_0(SH2_struct* sh)
{
  SH2dt(sh,13);
}
#define init_4_14_1_0
void FASTCALL call_4_14_1_0(SH2_struct* sh)
{
  SH2dt(sh,14);
}
#define init_4_15_1_0
void FASTCALL call_4_15_1_0(SH2_struct* sh)
{
  SH2dt(sh,15);
}
#define init_3_0_0_5
void FASTCALL call_3_0_0_5(SH2_struct* sh)
{
  SH2dmulu(sh,0,0);
}
#define init_3_0_1_5
void FASTCALL call_3_0_1_5(SH2_struct* sh)
{
  SH2dmulu(sh,0,1);
}
#define init_3_0_2_5
void FASTCALL call_3_0_2_5(SH2_struct* sh)
{
  SH2dmulu(sh,0,2);
}
#define init_3_0_3_5
void FASTCALL call_3_0_3_5(SH2_struct* sh)
{
  SH2dmulu(sh,0,3);
}
#define init_3_0_4_5
void FASTCALL call_3_0_4_5(SH2_struct* sh)
{
  SH2dmulu(sh,0,4);
}
#define init_3_0_5_5
void FASTCALL call_3_0_5_5(SH2_struct* sh)
{
  SH2dmulu(sh,0,5);
}
#define init_3_0_6_5
void FASTCALL call_3_0_6_5(SH2_struct* sh)
{
  SH2dmulu(sh,0,6);
}
#define init_3_0_7_5
void FASTCALL call_3_0_7_5(SH2_struct* sh)
{
  SH2dmulu(sh,0,7);
}
#define init_3_0_8_5
void FASTCALL call_3_0_8_5(SH2_struct* sh)
{
  SH2dmulu(sh,0,8);
}
#define init_3_0_9_5
void FASTCALL call_3_0_9_5(SH2_struct* sh)
{
  SH2dmulu(sh,0,9);
}
#define init_3_0_10_5
void FASTCALL call_3_0_10_5(SH2_struct* sh)
{
  SH2dmulu(sh,0,10);
}
#define init_3_0_11_5
void FASTCALL call_3_0_11_5(SH2_struct* sh)
{
  SH2dmulu(sh,0,11);
}
#define init_3_0_12_5
void FASTCALL call_3_0_12_5(SH2_struct* sh)
{
  SH2dmulu(sh,0,12);
}
#define init_3_0_13_5
void FASTCALL call_3_0_13_5(SH2_struct* sh)
{
  SH2dmulu(sh,0,13);
}
#define init_3_0_14_5
void FASTCALL call_3_0_14_5(SH2_struct* sh)
{
  SH2dmulu(sh,0,14);
}
#define init_3_0_15_5
void FASTCALL call_3_0_15_5(SH2_struct* sh)
{
  SH2dmulu(sh,0,15);
}
#define init_3_1_0_5
void FASTCALL call_3_1_0_5(SH2_struct* sh)
{
  SH2dmulu(sh,1,0);
}
#define init_3_1_1_5
void FASTCALL call_3_1_1_5(SH2_struct* sh)
{
  SH2dmulu(sh,1,1);
}
#define init_3_1_2_5
void FASTCALL call_3_1_2_5(SH2_struct* sh)
{
  SH2dmulu(sh,1,2);
}
#define init_3_1_3_5
void FASTCALL call_3_1_3_5(SH2_struct* sh)
{
  SH2dmulu(sh,1,3);
}
#define init_3_1_4_5
void FASTCALL call_3_1_4_5(SH2_struct* sh)
{
  SH2dmulu(sh,1,4);
}
#define init_3_1_5_5
void FASTCALL call_3_1_5_5(SH2_struct* sh)
{
  SH2dmulu(sh,1,5);
}
#define init_3_1_6_5
void FASTCALL call_3_1_6_5(SH2_struct* sh)
{
  SH2dmulu(sh,1,6);
}
#define init_3_1_7_5
void FASTCALL call_3_1_7_5(SH2_struct* sh)
{
  SH2dmulu(sh,1,7);
}
#define init_3_1_8_5
void FASTCALL call_3_1_8_5(SH2_struct* sh)
{
  SH2dmulu(sh,1,8);
}
#define init_3_1_9_5
void FASTCALL call_3_1_9_5(SH2_struct* sh)
{
  SH2dmulu(sh,1,9);
}
#define init_3_1_10_5
void FASTCALL call_3_1_10_5(SH2_struct* sh)
{
  SH2dmulu(sh,1,10);
}
#define init_3_1_11_5
void FASTCALL call_3_1_11_5(SH2_struct* sh)
{
  SH2dmulu(sh,1,11);
}
#define init_3_1_12_5
void FASTCALL call_3_1_12_5(SH2_struct* sh)
{
  SH2dmulu(sh,1,12);
}
#define init_3_1_13_5
void FASTCALL call_3_1_13_5(SH2_struct* sh)
{
  SH2dmulu(sh,1,13);
}
#define init_3_1_14_5
void FASTCALL call_3_1_14_5(SH2_struct* sh)
{
  SH2dmulu(sh,1,14);
}
#define init_3_1_15_5
void FASTCALL call_3_1_15_5(SH2_struct* sh)
{
  SH2dmulu(sh,1,15);
}
#define init_3_2_0_5
void FASTCALL call_3_2_0_5(SH2_struct* sh)
{
  SH2dmulu(sh,2,0);
}
#define init_3_2_1_5
void FASTCALL call_3_2_1_5(SH2_struct* sh)
{
  SH2dmulu(sh,2,1);
}
#define init_3_2_2_5
void FASTCALL call_3_2_2_5(SH2_struct* sh)
{
  SH2dmulu(sh,2,2);
}
#define init_3_2_3_5
void FASTCALL call_3_2_3_5(SH2_struct* sh)
{
  SH2dmulu(sh,2,3);
}
#define init_3_2_4_5
void FASTCALL call_3_2_4_5(SH2_struct* sh)
{
  SH2dmulu(sh,2,4);
}
#define init_3_2_5_5
void FASTCALL call_3_2_5_5(SH2_struct* sh)
{
  SH2dmulu(sh,2,5);
}
#define init_3_2_6_5
void FASTCALL call_3_2_6_5(SH2_struct* sh)
{
  SH2dmulu(sh,2,6);
}
#define init_3_2_7_5
void FASTCALL call_3_2_7_5(SH2_struct* sh)
{
  SH2dmulu(sh,2,7);
}
#define init_3_2_8_5
void FASTCALL call_3_2_8_5(SH2_struct* sh)
{
  SH2dmulu(sh,2,8);
}
#define init_3_2_9_5
void FASTCALL call_3_2_9_5(SH2_struct* sh)
{
  SH2dmulu(sh,2,9);
}
#define init_3_2_10_5
void FASTCALL call_3_2_10_5(SH2_struct* sh)
{
  SH2dmulu(sh,2,10);
}
#define init_3_2_11_5
void FASTCALL call_3_2_11_5(SH2_struct* sh)
{
  SH2dmulu(sh,2,11);
}
#define init_3_2_12_5
void FASTCALL call_3_2_12_5(SH2_struct* sh)
{
  SH2dmulu(sh,2,12);
}
#define init_3_2_13_5
void FASTCALL call_3_2_13_5(SH2_struct* sh)
{
  SH2dmulu(sh,2,13);
}
#define init_3_2_14_5
void FASTCALL call_3_2_14_5(SH2_struct* sh)
{
  SH2dmulu(sh,2,14);
}
#define init_3_2_15_5
void FASTCALL call_3_2_15_5(SH2_struct* sh)
{
  SH2dmulu(sh,2,15);
}
#define init_3_3_0_5
void FASTCALL call_3_3_0_5(SH2_struct* sh)
{
  SH2dmulu(sh,3,0);
}
#define init_3_3_1_5
void FASTCALL call_3_3_1_5(SH2_struct* sh)
{
  SH2dmulu(sh,3,1);
}
#define init_3_3_2_5
void FASTCALL call_3_3_2_5(SH2_struct* sh)
{
  SH2dmulu(sh,3,2);
}
#define init_3_3_3_5
void FASTCALL call_3_3_3_5(SH2_struct* sh)
{
  SH2dmulu(sh,3,3);
}
#define init_3_3_4_5
void FASTCALL call_3_3_4_5(SH2_struct* sh)
{
  SH2dmulu(sh,3,4);
}
#define init_3_3_5_5
void FASTCALL call_3_3_5_5(SH2_struct* sh)
{
  SH2dmulu(sh,3,5);
}
#define init_3_3_6_5
void FASTCALL call_3_3_6_5(SH2_struct* sh)
{
  SH2dmulu(sh,3,6);
}
#define init_3_3_7_5
void FASTCALL call_3_3_7_5(SH2_struct* sh)
{
  SH2dmulu(sh,3,7);
}
#define init_3_3_8_5
void FASTCALL call_3_3_8_5(SH2_struct* sh)
{
  SH2dmulu(sh,3,8);
}
#define init_3_3_9_5
void FASTCALL call_3_3_9_5(SH2_struct* sh)
{
  SH2dmulu(sh,3,9);
}
#define init_3_3_10_5
void FASTCALL call_3_3_10_5(SH2_struct* sh)
{
  SH2dmulu(sh,3,10);
}
#define init_3_3_11_5
void FASTCALL call_3_3_11_5(SH2_struct* sh)
{
  SH2dmulu(sh,3,11);
}
#define init_3_3_12_5
void FASTCALL call_3_3_12_5(SH2_struct* sh)
{
  SH2dmulu(sh,3,12);
}
#define init_3_3_13_5
void FASTCALL call_3_3_13_5(SH2_struct* sh)
{
  SH2dmulu(sh,3,13);
}
#define init_3_3_14_5
void FASTCALL call_3_3_14_5(SH2_struct* sh)
{
  SH2dmulu(sh,3,14);
}
#define init_3_3_15_5
void FASTCALL call_3_3_15_5(SH2_struct* sh)
{
  SH2dmulu(sh,3,15);
}
#define init_3_4_0_5
void FASTCALL call_3_4_0_5(SH2_struct* sh)
{
  SH2dmulu(sh,4,0);
}
#define init_3_4_1_5
void FASTCALL call_3_4_1_5(SH2_struct* sh)
{
  SH2dmulu(sh,4,1);
}
#define init_3_4_2_5
void FASTCALL call_3_4_2_5(SH2_struct* sh)
{
  SH2dmulu(sh,4,2);
}
#define init_3_4_3_5
void FASTCALL call_3_4_3_5(SH2_struct* sh)
{
  SH2dmulu(sh,4,3);
}
#define init_3_4_4_5
void FASTCALL call_3_4_4_5(SH2_struct* sh)
{
  SH2dmulu(sh,4,4);
}
#define init_3_4_5_5
void FASTCALL call_3_4_5_5(SH2_struct* sh)
{
  SH2dmulu(sh,4,5);
}
#define init_3_4_6_5
void FASTCALL call_3_4_6_5(SH2_struct* sh)
{
  SH2dmulu(sh,4,6);
}
#define init_3_4_7_5
void FASTCALL call_3_4_7_5(SH2_struct* sh)
{
  SH2dmulu(sh,4,7);
}
#define init_3_4_8_5
void FASTCALL call_3_4_8_5(SH2_struct* sh)
{
  SH2dmulu(sh,4,8);
}
#define init_3_4_9_5
void FASTCALL call_3_4_9_5(SH2_struct* sh)
{
  SH2dmulu(sh,4,9);
}
#define init_3_4_10_5
void FASTCALL call_3_4_10_5(SH2_struct* sh)
{
  SH2dmulu(sh,4,10);
}
#define init_3_4_11_5
void FASTCALL call_3_4_11_5(SH2_struct* sh)
{
  SH2dmulu(sh,4,11);
}
#define init_3_4_12_5
void FASTCALL call_3_4_12_5(SH2_struct* sh)
{
  SH2dmulu(sh,4,12);
}
#define init_3_4_13_5
void FASTCALL call_3_4_13_5(SH2_struct* sh)
{
  SH2dmulu(sh,4,13);
}
#define init_3_4_14_5
void FASTCALL call_3_4_14_5(SH2_struct* sh)
{
  SH2dmulu(sh,4,14);
}
#define init_3_4_15_5
void FASTCALL call_3_4_15_5(SH2_struct* sh)
{
  SH2dmulu(sh,4,15);
}
#define init_3_5_0_5
void FASTCALL call_3_5_0_5(SH2_struct* sh)
{
  SH2dmulu(sh,5,0);
}
#define init_3_5_1_5
void FASTCALL call_3_5_1_5(SH2_struct* sh)
{
  SH2dmulu(sh,5,1);
}
#define init_3_5_2_5
void FASTCALL call_3_5_2_5(SH2_struct* sh)
{
  SH2dmulu(sh,5,2);
}
#define init_3_5_3_5
void FASTCALL call_3_5_3_5(SH2_struct* sh)
{
  SH2dmulu(sh,5,3);
}
#define init_3_5_4_5
void FASTCALL call_3_5_4_5(SH2_struct* sh)
{
  SH2dmulu(sh,5,4);
}
#define init_3_5_5_5
void FASTCALL call_3_5_5_5(SH2_struct* sh)
{
  SH2dmulu(sh,5,5);
}
#define init_3_5_6_5
void FASTCALL call_3_5_6_5(SH2_struct* sh)
{
  SH2dmulu(sh,5,6);
}
#define init_3_5_7_5
void FASTCALL call_3_5_7_5(SH2_struct* sh)
{
  SH2dmulu(sh,5,7);
}
#define init_3_5_8_5
void FASTCALL call_3_5_8_5(SH2_struct* sh)
{
  SH2dmulu(sh,5,8);
}
#define init_3_5_9_5
void FASTCALL call_3_5_9_5(SH2_struct* sh)
{
  SH2dmulu(sh,5,9);
}
#define init_3_5_10_5
void FASTCALL call_3_5_10_5(SH2_struct* sh)
{
  SH2dmulu(sh,5,10);
}
#define init_3_5_11_5
void FASTCALL call_3_5_11_5(SH2_struct* sh)
{
  SH2dmulu(sh,5,11);
}
#define init_3_5_12_5
void FASTCALL call_3_5_12_5(SH2_struct* sh)
{
  SH2dmulu(sh,5,12);
}
#define init_3_5_13_5
void FASTCALL call_3_5_13_5(SH2_struct* sh)
{
  SH2dmulu(sh,5,13);
}
#define init_3_5_14_5
void FASTCALL call_3_5_14_5(SH2_struct* sh)
{
  SH2dmulu(sh,5,14);
}
#define init_3_5_15_5
void FASTCALL call_3_5_15_5(SH2_struct* sh)
{
  SH2dmulu(sh,5,15);
}
#define init_3_6_0_5
void FASTCALL call_3_6_0_5(SH2_struct* sh)
{
  SH2dmulu(sh,6,0);
}
#define init_3_6_1_5
void FASTCALL call_3_6_1_5(SH2_struct* sh)
{
  SH2dmulu(sh,6,1);
}
#define init_3_6_2_5
void FASTCALL call_3_6_2_5(SH2_struct* sh)
{
  SH2dmulu(sh,6,2);
}
#define init_3_6_3_5
void FASTCALL call_3_6_3_5(SH2_struct* sh)
{
  SH2dmulu(sh,6,3);
}
#define init_3_6_4_5
void FASTCALL call_3_6_4_5(SH2_struct* sh)
{
  SH2dmulu(sh,6,4);
}
#define init_3_6_5_5
void FASTCALL call_3_6_5_5(SH2_struct* sh)
{
  SH2dmulu(sh,6,5);
}
#define init_3_6_6_5
void FASTCALL call_3_6_6_5(SH2_struct* sh)
{
  SH2dmulu(sh,6,6);
}
#define init_3_6_7_5
void FASTCALL call_3_6_7_5(SH2_struct* sh)
{
  SH2dmulu(sh,6,7);
}
#define init_3_6_8_5
void FASTCALL call_3_6_8_5(SH2_struct* sh)
{
  SH2dmulu(sh,6,8);
}
#define init_3_6_9_5
void FASTCALL call_3_6_9_5(SH2_struct* sh)
{
  SH2dmulu(sh,6,9);
}
#define init_3_6_10_5
void FASTCALL call_3_6_10_5(SH2_struct* sh)
{
  SH2dmulu(sh,6,10);
}
#define init_3_6_11_5
void FASTCALL call_3_6_11_5(SH2_struct* sh)
{
  SH2dmulu(sh,6,11);
}
#define init_3_6_12_5
void FASTCALL call_3_6_12_5(SH2_struct* sh)
{
  SH2dmulu(sh,6,12);
}
#define init_3_6_13_5
void FASTCALL call_3_6_13_5(SH2_struct* sh)
{
  SH2dmulu(sh,6,13);
}
#define init_3_6_14_5
void FASTCALL call_3_6_14_5(SH2_struct* sh)
{
  SH2dmulu(sh,6,14);
}
#define init_3_6_15_5
void FASTCALL call_3_6_15_5(SH2_struct* sh)
{
  SH2dmulu(sh,6,15);
}
#define init_3_7_0_5
void FASTCALL call_3_7_0_5(SH2_struct* sh)
{
  SH2dmulu(sh,7,0);
}
#define init_3_7_1_5
void FASTCALL call_3_7_1_5(SH2_struct* sh)
{
  SH2dmulu(sh,7,1);
}
#define init_3_7_2_5
void FASTCALL call_3_7_2_5(SH2_struct* sh)
{
  SH2dmulu(sh,7,2);
}
#define init_3_7_3_5
void FASTCALL call_3_7_3_5(SH2_struct* sh)
{
  SH2dmulu(sh,7,3);
}
#define init_3_7_4_5
void FASTCALL call_3_7_4_5(SH2_struct* sh)
{
  SH2dmulu(sh,7,4);
}
#define init_3_7_5_5
void FASTCALL call_3_7_5_5(SH2_struct* sh)
{
  SH2dmulu(sh,7,5);
}
#define init_3_7_6_5
void FASTCALL call_3_7_6_5(SH2_struct* sh)
{
  SH2dmulu(sh,7,6);
}
#define init_3_7_7_5
void FASTCALL call_3_7_7_5(SH2_struct* sh)
{
  SH2dmulu(sh,7,7);
}
#define init_3_7_8_5
void FASTCALL call_3_7_8_5(SH2_struct* sh)
{
  SH2dmulu(sh,7,8);
}
#define init_3_7_9_5
void FASTCALL call_3_7_9_5(SH2_struct* sh)
{
  SH2dmulu(sh,7,9);
}
#define init_3_7_10_5
void FASTCALL call_3_7_10_5(SH2_struct* sh)
{
  SH2dmulu(sh,7,10);
}
#define init_3_7_11_5
void FASTCALL call_3_7_11_5(SH2_struct* sh)
{
  SH2dmulu(sh,7,11);
}
#define init_3_7_12_5
void FASTCALL call_3_7_12_5(SH2_struct* sh)
{
  SH2dmulu(sh,7,12);
}
#define init_3_7_13_5
void FASTCALL call_3_7_13_5(SH2_struct* sh)
{
  SH2dmulu(sh,7,13);
}
#define init_3_7_14_5
void FASTCALL call_3_7_14_5(SH2_struct* sh)
{
  SH2dmulu(sh,7,14);
}
#define init_3_7_15_5
void FASTCALL call_3_7_15_5(SH2_struct* sh)
{
  SH2dmulu(sh,7,15);
}
#define init_3_8_0_5
void FASTCALL call_3_8_0_5(SH2_struct* sh)
{
  SH2dmulu(sh,8,0);
}
#define init_3_8_1_5
void FASTCALL call_3_8_1_5(SH2_struct* sh)
{
  SH2dmulu(sh,8,1);
}
#define init_3_8_2_5
void FASTCALL call_3_8_2_5(SH2_struct* sh)
{
  SH2dmulu(sh,8,2);
}
#define init_3_8_3_5
void FASTCALL call_3_8_3_5(SH2_struct* sh)
{
  SH2dmulu(sh,8,3);
}
#define init_3_8_4_5
void FASTCALL call_3_8_4_5(SH2_struct* sh)
{
  SH2dmulu(sh,8,4);
}
#define init_3_8_5_5
void FASTCALL call_3_8_5_5(SH2_struct* sh)
{
  SH2dmulu(sh,8,5);
}
#define init_3_8_6_5
void FASTCALL call_3_8_6_5(SH2_struct* sh)
{
  SH2dmulu(sh,8,6);
}
#define init_3_8_7_5
void FASTCALL call_3_8_7_5(SH2_struct* sh)
{
  SH2dmulu(sh,8,7);
}
#define init_3_8_8_5
void FASTCALL call_3_8_8_5(SH2_struct* sh)
{
  SH2dmulu(sh,8,8);
}
#define init_3_8_9_5
void FASTCALL call_3_8_9_5(SH2_struct* sh)
{
  SH2dmulu(sh,8,9);
}
#define init_3_8_10_5
void FASTCALL call_3_8_10_5(SH2_struct* sh)
{
  SH2dmulu(sh,8,10);
}
#define init_3_8_11_5
void FASTCALL call_3_8_11_5(SH2_struct* sh)
{
  SH2dmulu(sh,8,11);
}
#define init_3_8_12_5
void FASTCALL call_3_8_12_5(SH2_struct* sh)
{
  SH2dmulu(sh,8,12);
}
#define init_3_8_13_5
void FASTCALL call_3_8_13_5(SH2_struct* sh)
{
  SH2dmulu(sh,8,13);
}
#define init_3_8_14_5
void FASTCALL call_3_8_14_5(SH2_struct* sh)
{
  SH2dmulu(sh,8,14);
}
#define init_3_8_15_5
void FASTCALL call_3_8_15_5(SH2_struct* sh)
{
  SH2dmulu(sh,8,15);
}
#define init_3_9_0_5
void FASTCALL call_3_9_0_5(SH2_struct* sh)
{
  SH2dmulu(sh,9,0);
}
#define init_3_9_1_5
void FASTCALL call_3_9_1_5(SH2_struct* sh)
{
  SH2dmulu(sh,9,1);
}
#define init_3_9_2_5
void FASTCALL call_3_9_2_5(SH2_struct* sh)
{
  SH2dmulu(sh,9,2);
}
#define init_3_9_3_5
void FASTCALL call_3_9_3_5(SH2_struct* sh)
{
  SH2dmulu(sh,9,3);
}
#define init_3_9_4_5
void FASTCALL call_3_9_4_5(SH2_struct* sh)
{
  SH2dmulu(sh,9,4);
}
#define init_3_9_5_5
void FASTCALL call_3_9_5_5(SH2_struct* sh)
{
  SH2dmulu(sh,9,5);
}
#define init_3_9_6_5
void FASTCALL call_3_9_6_5(SH2_struct* sh)
{
  SH2dmulu(sh,9,6);
}
#define init_3_9_7_5
void FASTCALL call_3_9_7_5(SH2_struct* sh)
{
  SH2dmulu(sh,9,7);
}
#define init_3_9_8_5
void FASTCALL call_3_9_8_5(SH2_struct* sh)
{
  SH2dmulu(sh,9,8);
}
#define init_3_9_9_5
void FASTCALL call_3_9_9_5(SH2_struct* sh)
{
  SH2dmulu(sh,9,9);
}
#define init_3_9_10_5
void FASTCALL call_3_9_10_5(SH2_struct* sh)
{
  SH2dmulu(sh,9,10);
}
#define init_3_9_11_5
void FASTCALL call_3_9_11_5(SH2_struct* sh)
{
  SH2dmulu(sh,9,11);
}
#define init_3_9_12_5
void FASTCALL call_3_9_12_5(SH2_struct* sh)
{
  SH2dmulu(sh,9,12);
}
#define init_3_9_13_5
void FASTCALL call_3_9_13_5(SH2_struct* sh)
{
  SH2dmulu(sh,9,13);
}
#define init_3_9_14_5
void FASTCALL call_3_9_14_5(SH2_struct* sh)
{
  SH2dmulu(sh,9,14);
}
#define init_3_9_15_5
void FASTCALL call_3_9_15_5(SH2_struct* sh)
{
  SH2dmulu(sh,9,15);
}
#define init_3_10_0_5
void FASTCALL call_3_10_0_5(SH2_struct* sh)
{
  SH2dmulu(sh,10,0);
}
#define init_3_10_1_5
void FASTCALL call_3_10_1_5(SH2_struct* sh)
{
  SH2dmulu(sh,10,1);
}
#define init_3_10_2_5
void FASTCALL call_3_10_2_5(SH2_struct* sh)
{
  SH2dmulu(sh,10,2);
}
#define init_3_10_3_5
void FASTCALL call_3_10_3_5(SH2_struct* sh)
{
  SH2dmulu(sh,10,3);
}
#define init_3_10_4_5
void FASTCALL call_3_10_4_5(SH2_struct* sh)
{
  SH2dmulu(sh,10,4);
}
#define init_3_10_5_5
void FASTCALL call_3_10_5_5(SH2_struct* sh)
{
  SH2dmulu(sh,10,5);
}
#define init_3_10_6_5
void FASTCALL call_3_10_6_5(SH2_struct* sh)
{
  SH2dmulu(sh,10,6);
}
#define init_3_10_7_5
void FASTCALL call_3_10_7_5(SH2_struct* sh)
{
  SH2dmulu(sh,10,7);
}
#define init_3_10_8_5
void FASTCALL call_3_10_8_5(SH2_struct* sh)
{
  SH2dmulu(sh,10,8);
}
#define init_3_10_9_5
void FASTCALL call_3_10_9_5(SH2_struct* sh)
{
  SH2dmulu(sh,10,9);
}
#define init_3_10_10_5
void FASTCALL call_3_10_10_5(SH2_struct* sh)
{
  SH2dmulu(sh,10,10);
}
#define init_3_10_11_5
void FASTCALL call_3_10_11_5(SH2_struct* sh)
{
  SH2dmulu(sh,10,11);
}
#define init_3_10_12_5
void FASTCALL call_3_10_12_5(SH2_struct* sh)
{
  SH2dmulu(sh,10,12);
}
#define init_3_10_13_5
void FASTCALL call_3_10_13_5(SH2_struct* sh)
{
  SH2dmulu(sh,10,13);
}
#define init_3_10_14_5
void FASTCALL call_3_10_14_5(SH2_struct* sh)
{
  SH2dmulu(sh,10,14);
}
#define init_3_10_15_5
void FASTCALL call_3_10_15_5(SH2_struct* sh)
{
  SH2dmulu(sh,10,15);
}
#define init_3_11_0_5
void FASTCALL call_3_11_0_5(SH2_struct* sh)
{
  SH2dmulu(sh,11,0);
}
#define init_3_11_1_5
void FASTCALL call_3_11_1_5(SH2_struct* sh)
{
  SH2dmulu(sh,11,1);
}
#define init_3_11_2_5
void FASTCALL call_3_11_2_5(SH2_struct* sh)
{
  SH2dmulu(sh,11,2);
}
#define init_3_11_3_5
void FASTCALL call_3_11_3_5(SH2_struct* sh)
{
  SH2dmulu(sh,11,3);
}
#define init_3_11_4_5
void FASTCALL call_3_11_4_5(SH2_struct* sh)
{
  SH2dmulu(sh,11,4);
}
#define init_3_11_5_5
void FASTCALL call_3_11_5_5(SH2_struct* sh)
{
  SH2dmulu(sh,11,5);
}
#define init_3_11_6_5
void FASTCALL call_3_11_6_5(SH2_struct* sh)
{
  SH2dmulu(sh,11,6);
}
#define init_3_11_7_5
void FASTCALL call_3_11_7_5(SH2_struct* sh)
{
  SH2dmulu(sh,11,7);
}
#define init_3_11_8_5
void FASTCALL call_3_11_8_5(SH2_struct* sh)
{
  SH2dmulu(sh,11,8);
}
#define init_3_11_9_5
void FASTCALL call_3_11_9_5(SH2_struct* sh)
{
  SH2dmulu(sh,11,9);
}
#define init_3_11_10_5
void FASTCALL call_3_11_10_5(SH2_struct* sh)
{
  SH2dmulu(sh,11,10);
}
#define init_3_11_11_5
void FASTCALL call_3_11_11_5(SH2_struct* sh)
{
  SH2dmulu(sh,11,11);
}
#define init_3_11_12_5
void FASTCALL call_3_11_12_5(SH2_struct* sh)
{
  SH2dmulu(sh,11,12);
}
#define init_3_11_13_5
void FASTCALL call_3_11_13_5(SH2_struct* sh)
{
  SH2dmulu(sh,11,13);
}
#define init_3_11_14_5
void FASTCALL call_3_11_14_5(SH2_struct* sh)
{
  SH2dmulu(sh,11,14);
}
#define init_3_11_15_5
void FASTCALL call_3_11_15_5(SH2_struct* sh)
{
  SH2dmulu(sh,11,15);
}
#define init_3_12_0_5
void FASTCALL call_3_12_0_5(SH2_struct* sh)
{
  SH2dmulu(sh,12,0);
}
#define init_3_12_1_5
void FASTCALL call_3_12_1_5(SH2_struct* sh)
{
  SH2dmulu(sh,12,1);
}
#define init_3_12_2_5
void FASTCALL call_3_12_2_5(SH2_struct* sh)
{
  SH2dmulu(sh,12,2);
}
#define init_3_12_3_5
void FASTCALL call_3_12_3_5(SH2_struct* sh)
{
  SH2dmulu(sh,12,3);
}
#define init_3_12_4_5
void FASTCALL call_3_12_4_5(SH2_struct* sh)
{
  SH2dmulu(sh,12,4);
}
#define init_3_12_5_5
void FASTCALL call_3_12_5_5(SH2_struct* sh)
{
  SH2dmulu(sh,12,5);
}
#define init_3_12_6_5
void FASTCALL call_3_12_6_5(SH2_struct* sh)
{
  SH2dmulu(sh,12,6);
}
#define init_3_12_7_5
void FASTCALL call_3_12_7_5(SH2_struct* sh)
{
  SH2dmulu(sh,12,7);
}
#define init_3_12_8_5
void FASTCALL call_3_12_8_5(SH2_struct* sh)
{
  SH2dmulu(sh,12,8);
}
#define init_3_12_9_5
void FASTCALL call_3_12_9_5(SH2_struct* sh)
{
  SH2dmulu(sh,12,9);
}
#define init_3_12_10_5
void FASTCALL call_3_12_10_5(SH2_struct* sh)
{
  SH2dmulu(sh,12,10);
}
#define init_3_12_11_5
void FASTCALL call_3_12_11_5(SH2_struct* sh)
{
  SH2dmulu(sh,12,11);
}
#define init_3_12_12_5
void FASTCALL call_3_12_12_5(SH2_struct* sh)
{
  SH2dmulu(sh,12,12);
}
#define init_3_12_13_5
void FASTCALL call_3_12_13_5(SH2_struct* sh)
{
  SH2dmulu(sh,12,13);
}
#define init_3_12_14_5
void FASTCALL call_3_12_14_5(SH2_struct* sh)
{
  SH2dmulu(sh,12,14);
}
#define init_3_12_15_5
void FASTCALL call_3_12_15_5(SH2_struct* sh)
{
  SH2dmulu(sh,12,15);
}
#define init_3_13_0_5
void FASTCALL call_3_13_0_5(SH2_struct* sh)
{
  SH2dmulu(sh,13,0);
}
#define init_3_13_1_5
void FASTCALL call_3_13_1_5(SH2_struct* sh)
{
  SH2dmulu(sh,13,1);
}
#define init_3_13_2_5
void FASTCALL call_3_13_2_5(SH2_struct* sh)
{
  SH2dmulu(sh,13,2);
}
#define init_3_13_3_5
void FASTCALL call_3_13_3_5(SH2_struct* sh)
{
  SH2dmulu(sh,13,3);
}
#define init_3_13_4_5
void FASTCALL call_3_13_4_5(SH2_struct* sh)
{
  SH2dmulu(sh,13,4);
}
#define init_3_13_5_5
void FASTCALL call_3_13_5_5(SH2_struct* sh)
{
  SH2dmulu(sh,13,5);
}
#define init_3_13_6_5
void FASTCALL call_3_13_6_5(SH2_struct* sh)
{
  SH2dmulu(sh,13,6);
}
#define init_3_13_7_5
void FASTCALL call_3_13_7_5(SH2_struct* sh)
{
  SH2dmulu(sh,13,7);
}
#define init_3_13_8_5
void FASTCALL call_3_13_8_5(SH2_struct* sh)
{
  SH2dmulu(sh,13,8);
}
#define init_3_13_9_5
void FASTCALL call_3_13_9_5(SH2_struct* sh)
{
  SH2dmulu(sh,13,9);
}
#define init_3_13_10_5
void FASTCALL call_3_13_10_5(SH2_struct* sh)
{
  SH2dmulu(sh,13,10);
}
#define init_3_13_11_5
void FASTCALL call_3_13_11_5(SH2_struct* sh)
{
  SH2dmulu(sh,13,11);
}
#define init_3_13_12_5
void FASTCALL call_3_13_12_5(SH2_struct* sh)
{
  SH2dmulu(sh,13,12);
}
#define init_3_13_13_5
void FASTCALL call_3_13_13_5(SH2_struct* sh)
{
  SH2dmulu(sh,13,13);
}
#define init_3_13_14_5
void FASTCALL call_3_13_14_5(SH2_struct* sh)
{
  SH2dmulu(sh,13,14);
}
#define init_3_13_15_5
void FASTCALL call_3_13_15_5(SH2_struct* sh)
{
  SH2dmulu(sh,13,15);
}
#define init_3_14_0_5
void FASTCALL call_3_14_0_5(SH2_struct* sh)
{
  SH2dmulu(sh,14,0);
}
#define init_3_14_1_5
void FASTCALL call_3_14_1_5(SH2_struct* sh)
{
  SH2dmulu(sh,14,1);
}
#define init_3_14_2_5
void FASTCALL call_3_14_2_5(SH2_struct* sh)
{
  SH2dmulu(sh,14,2);
}
#define init_3_14_3_5
void FASTCALL call_3_14_3_5(SH2_struct* sh)
{
  SH2dmulu(sh,14,3);
}
#define init_3_14_4_5
void FASTCALL call_3_14_4_5(SH2_struct* sh)
{
  SH2dmulu(sh,14,4);
}
#define init_3_14_5_5
void FASTCALL call_3_14_5_5(SH2_struct* sh)
{
  SH2dmulu(sh,14,5);
}
#define init_3_14_6_5
void FASTCALL call_3_14_6_5(SH2_struct* sh)
{
  SH2dmulu(sh,14,6);
}
#define init_3_14_7_5
void FASTCALL call_3_14_7_5(SH2_struct* sh)
{
  SH2dmulu(sh,14,7);
}
#define init_3_14_8_5
void FASTCALL call_3_14_8_5(SH2_struct* sh)
{
  SH2dmulu(sh,14,8);
}
#define init_3_14_9_5
void FASTCALL call_3_14_9_5(SH2_struct* sh)
{
  SH2dmulu(sh,14,9);
}
#define init_3_14_10_5
void FASTCALL call_3_14_10_5(SH2_struct* sh)
{
  SH2dmulu(sh,14,10);
}
#define init_3_14_11_5
void FASTCALL call_3_14_11_5(SH2_struct* sh)
{
  SH2dmulu(sh,14,11);
}
#define init_3_14_12_5
void FASTCALL call_3_14_12_5(SH2_struct* sh)
{
  SH2dmulu(sh,14,12);
}
#define init_3_14_13_5
void FASTCALL call_3_14_13_5(SH2_struct* sh)
{
  SH2dmulu(sh,14,13);
}
#define init_3_14_14_5
void FASTCALL call_3_14_14_5(SH2_struct* sh)
{
  SH2dmulu(sh,14,14);
}
#define init_3_14_15_5
void FASTCALL call_3_14_15_5(SH2_struct* sh)
{
  SH2dmulu(sh,14,15);
}
#define init_3_15_0_5
void FASTCALL call_3_15_0_5(SH2_struct* sh)
{
  SH2dmulu(sh,15,0);
}
#define init_3_15_1_5
void FASTCALL call_3_15_1_5(SH2_struct* sh)
{
  SH2dmulu(sh,15,1);
}
#define init_3_15_2_5
void FASTCALL call_3_15_2_5(SH2_struct* sh)
{
  SH2dmulu(sh,15,2);
}
#define init_3_15_3_5
void FASTCALL call_3_15_3_5(SH2_struct* sh)
{
  SH2dmulu(sh,15,3);
}
#define init_3_15_4_5
void FASTCALL call_3_15_4_5(SH2_struct* sh)
{
  SH2dmulu(sh,15,4);
}
#define init_3_15_5_5
void FASTCALL call_3_15_5_5(SH2_struct* sh)
{
  SH2dmulu(sh,15,5);
}
#define init_3_15_6_5
void FASTCALL call_3_15_6_5(SH2_struct* sh)
{
  SH2dmulu(sh,15,6);
}
#define init_3_15_7_5
void FASTCALL call_3_15_7_5(SH2_struct* sh)
{
  SH2dmulu(sh,15,7);
}
#define init_3_15_8_5
void FASTCALL call_3_15_8_5(SH2_struct* sh)
{
  SH2dmulu(sh,15,8);
}
#define init_3_15_9_5
void FASTCALL call_3_15_9_5(SH2_struct* sh)
{
  SH2dmulu(sh,15,9);
}
#define init_3_15_10_5
void FASTCALL call_3_15_10_5(SH2_struct* sh)
{
  SH2dmulu(sh,15,10);
}
#define init_3_15_11_5
void FASTCALL call_3_15_11_5(SH2_struct* sh)
{
  SH2dmulu(sh,15,11);
}
#define init_3_15_12_5
void FASTCALL call_3_15_12_5(SH2_struct* sh)
{
  SH2dmulu(sh,15,12);
}
#define init_3_15_13_5
void FASTCALL call_3_15_13_5(SH2_struct* sh)
{
  SH2dmulu(sh,15,13);
}
#define init_3_15_14_5
void FASTCALL call_3_15_14_5(SH2_struct* sh)
{
  SH2dmulu(sh,15,14);
}
#define init_3_15_15_5
void FASTCALL call_3_15_15_5(SH2_struct* sh)
{
  SH2dmulu(sh,15,15);
}
#define init_3_0_0_13
void FASTCALL call_3_0_0_13(SH2_struct* sh)
{
  SH2dmuls(sh,0,0);
}
#define init_3_0_1_13
void FASTCALL call_3_0_1_13(SH2_struct* sh)
{
  SH2dmuls(sh,0,1);
}
#define init_3_0_2_13
void FASTCALL call_3_0_2_13(SH2_struct* sh)
{
  SH2dmuls(sh,0,2);
}
#define init_3_0_3_13
void FASTCALL call_3_0_3_13(SH2_struct* sh)
{
  SH2dmuls(sh,0,3);
}
#define init_3_0_4_13
void FASTCALL call_3_0_4_13(SH2_struct* sh)
{
  SH2dmuls(sh,0,4);
}
#define init_3_0_5_13
void FASTCALL call_3_0_5_13(SH2_struct* sh)
{
  SH2dmuls(sh,0,5);
}
#define init_3_0_6_13
void FASTCALL call_3_0_6_13(SH2_struct* sh)
{
  SH2dmuls(sh,0,6);
}
#define init_3_0_7_13
void FASTCALL call_3_0_7_13(SH2_struct* sh)
{
  SH2dmuls(sh,0,7);
}
#define init_3_0_8_13
void FASTCALL call_3_0_8_13(SH2_struct* sh)
{
  SH2dmuls(sh,0,8);
}
#define init_3_0_9_13
void FASTCALL call_3_0_9_13(SH2_struct* sh)
{
  SH2dmuls(sh,0,9);
}
#define init_3_0_10_13
void FASTCALL call_3_0_10_13(SH2_struct* sh)
{
  SH2dmuls(sh,0,10);
}
#define init_3_0_11_13
void FASTCALL call_3_0_11_13(SH2_struct* sh)
{
  SH2dmuls(sh,0,11);
}
#define init_3_0_12_13
void FASTCALL call_3_0_12_13(SH2_struct* sh)
{
  SH2dmuls(sh,0,12);
}
#define init_3_0_13_13
void FASTCALL call_3_0_13_13(SH2_struct* sh)
{
  SH2dmuls(sh,0,13);
}
#define init_3_0_14_13
void FASTCALL call_3_0_14_13(SH2_struct* sh)
{
  SH2dmuls(sh,0,14);
}
#define init_3_0_15_13
void FASTCALL call_3_0_15_13(SH2_struct* sh)
{
  SH2dmuls(sh,0,15);
}
#define init_3_1_0_13
void FASTCALL call_3_1_0_13(SH2_struct* sh)
{
  SH2dmuls(sh,1,0);
}
#define init_3_1_1_13
void FASTCALL call_3_1_1_13(SH2_struct* sh)
{
  SH2dmuls(sh,1,1);
}
#define init_3_1_2_13
void FASTCALL call_3_1_2_13(SH2_struct* sh)
{
  SH2dmuls(sh,1,2);
}
#define init_3_1_3_13
void FASTCALL call_3_1_3_13(SH2_struct* sh)
{
  SH2dmuls(sh,1,3);
}
#define init_3_1_4_13
void FASTCALL call_3_1_4_13(SH2_struct* sh)
{
  SH2dmuls(sh,1,4);
}
#define init_3_1_5_13
void FASTCALL call_3_1_5_13(SH2_struct* sh)
{
  SH2dmuls(sh,1,5);
}
#define init_3_1_6_13
void FASTCALL call_3_1_6_13(SH2_struct* sh)
{
  SH2dmuls(sh,1,6);
}
#define init_3_1_7_13
void FASTCALL call_3_1_7_13(SH2_struct* sh)
{
  SH2dmuls(sh,1,7);
}
#define init_3_1_8_13
void FASTCALL call_3_1_8_13(SH2_struct* sh)
{
  SH2dmuls(sh,1,8);
}
#define init_3_1_9_13
void FASTCALL call_3_1_9_13(SH2_struct* sh)
{
  SH2dmuls(sh,1,9);
}
#define init_3_1_10_13
void FASTCALL call_3_1_10_13(SH2_struct* sh)
{
  SH2dmuls(sh,1,10);
}
#define init_3_1_11_13
void FASTCALL call_3_1_11_13(SH2_struct* sh)
{
  SH2dmuls(sh,1,11);
}
#define init_3_1_12_13
void FASTCALL call_3_1_12_13(SH2_struct* sh)
{
  SH2dmuls(sh,1,12);
}
#define init_3_1_13_13
void FASTCALL call_3_1_13_13(SH2_struct* sh)
{
  SH2dmuls(sh,1,13);
}
#define init_3_1_14_13
void FASTCALL call_3_1_14_13(SH2_struct* sh)
{
  SH2dmuls(sh,1,14);
}
#define init_3_1_15_13
void FASTCALL call_3_1_15_13(SH2_struct* sh)
{
  SH2dmuls(sh,1,15);
}
#define init_3_2_0_13
void FASTCALL call_3_2_0_13(SH2_struct* sh)
{
  SH2dmuls(sh,2,0);
}
#define init_3_2_1_13
void FASTCALL call_3_2_1_13(SH2_struct* sh)
{
  SH2dmuls(sh,2,1);
}
#define init_3_2_2_13
void FASTCALL call_3_2_2_13(SH2_struct* sh)
{
  SH2dmuls(sh,2,2);
}
#define init_3_2_3_13
void FASTCALL call_3_2_3_13(SH2_struct* sh)
{
  SH2dmuls(sh,2,3);
}
#define init_3_2_4_13
void FASTCALL call_3_2_4_13(SH2_struct* sh)
{
  SH2dmuls(sh,2,4);
}
#define init_3_2_5_13
void FASTCALL call_3_2_5_13(SH2_struct* sh)
{
  SH2dmuls(sh,2,5);
}
#define init_3_2_6_13
void FASTCALL call_3_2_6_13(SH2_struct* sh)
{
  SH2dmuls(sh,2,6);
}
#define init_3_2_7_13
void FASTCALL call_3_2_7_13(SH2_struct* sh)
{
  SH2dmuls(sh,2,7);
}
#define init_3_2_8_13
void FASTCALL call_3_2_8_13(SH2_struct* sh)
{
  SH2dmuls(sh,2,8);
}
#define init_3_2_9_13
void FASTCALL call_3_2_9_13(SH2_struct* sh)
{
  SH2dmuls(sh,2,9);
}
#define init_3_2_10_13
void FASTCALL call_3_2_10_13(SH2_struct* sh)
{
  SH2dmuls(sh,2,10);
}
#define init_3_2_11_13
void FASTCALL call_3_2_11_13(SH2_struct* sh)
{
  SH2dmuls(sh,2,11);
}
#define init_3_2_12_13
void FASTCALL call_3_2_12_13(SH2_struct* sh)
{
  SH2dmuls(sh,2,12);
}
#define init_3_2_13_13
void FASTCALL call_3_2_13_13(SH2_struct* sh)
{
  SH2dmuls(sh,2,13);
}
#define init_3_2_14_13
void FASTCALL call_3_2_14_13(SH2_struct* sh)
{
  SH2dmuls(sh,2,14);
}
#define init_3_2_15_13
void FASTCALL call_3_2_15_13(SH2_struct* sh)
{
  SH2dmuls(sh,2,15);
}
#define init_3_3_0_13
void FASTCALL call_3_3_0_13(SH2_struct* sh)
{
  SH2dmuls(sh,3,0);
}
#define init_3_3_1_13
void FASTCALL call_3_3_1_13(SH2_struct* sh)
{
  SH2dmuls(sh,3,1);
}
#define init_3_3_2_13
void FASTCALL call_3_3_2_13(SH2_struct* sh)
{
  SH2dmuls(sh,3,2);
}
#define init_3_3_3_13
void FASTCALL call_3_3_3_13(SH2_struct* sh)
{
  SH2dmuls(sh,3,3);
}
#define init_3_3_4_13
void FASTCALL call_3_3_4_13(SH2_struct* sh)
{
  SH2dmuls(sh,3,4);
}
#define init_3_3_5_13
void FASTCALL call_3_3_5_13(SH2_struct* sh)
{
  SH2dmuls(sh,3,5);
}
#define init_3_3_6_13
void FASTCALL call_3_3_6_13(SH2_struct* sh)
{
  SH2dmuls(sh,3,6);
}
#define init_3_3_7_13
void FASTCALL call_3_3_7_13(SH2_struct* sh)
{
  SH2dmuls(sh,3,7);
}
#define init_3_3_8_13
void FASTCALL call_3_3_8_13(SH2_struct* sh)
{
  SH2dmuls(sh,3,8);
}
#define init_3_3_9_13
void FASTCALL call_3_3_9_13(SH2_struct* sh)
{
  SH2dmuls(sh,3,9);
}
#define init_3_3_10_13
void FASTCALL call_3_3_10_13(SH2_struct* sh)
{
  SH2dmuls(sh,3,10);
}
#define init_3_3_11_13
void FASTCALL call_3_3_11_13(SH2_struct* sh)
{
  SH2dmuls(sh,3,11);
}
#define init_3_3_12_13
void FASTCALL call_3_3_12_13(SH2_struct* sh)
{
  SH2dmuls(sh,3,12);
}
#define init_3_3_13_13
void FASTCALL call_3_3_13_13(SH2_struct* sh)
{
  SH2dmuls(sh,3,13);
}
#define init_3_3_14_13
void FASTCALL call_3_3_14_13(SH2_struct* sh)
{
  SH2dmuls(sh,3,14);
}
#define init_3_3_15_13
void FASTCALL call_3_3_15_13(SH2_struct* sh)
{
  SH2dmuls(sh,3,15);
}
#define init_3_4_0_13
void FASTCALL call_3_4_0_13(SH2_struct* sh)
{
  SH2dmuls(sh,4,0);
}
#define init_3_4_1_13
void FASTCALL call_3_4_1_13(SH2_struct* sh)
{
  SH2dmuls(sh,4,1);
}
#define init_3_4_2_13
void FASTCALL call_3_4_2_13(SH2_struct* sh)
{
  SH2dmuls(sh,4,2);
}
#define init_3_4_3_13
void FASTCALL call_3_4_3_13(SH2_struct* sh)
{
  SH2dmuls(sh,4,3);
}
#define init_3_4_4_13
void FASTCALL call_3_4_4_13(SH2_struct* sh)
{
  SH2dmuls(sh,4,4);
}
#define init_3_4_5_13
void FASTCALL call_3_4_5_13(SH2_struct* sh)
{
  SH2dmuls(sh,4,5);
}
#define init_3_4_6_13
void FASTCALL call_3_4_6_13(SH2_struct* sh)
{
  SH2dmuls(sh,4,6);
}
#define init_3_4_7_13
void FASTCALL call_3_4_7_13(SH2_struct* sh)
{
  SH2dmuls(sh,4,7);
}
#define init_3_4_8_13
void FASTCALL call_3_4_8_13(SH2_struct* sh)
{
  SH2dmuls(sh,4,8);
}
#define init_3_4_9_13
void FASTCALL call_3_4_9_13(SH2_struct* sh)
{
  SH2dmuls(sh,4,9);
}
#define init_3_4_10_13
void FASTCALL call_3_4_10_13(SH2_struct* sh)
{
  SH2dmuls(sh,4,10);
}
#define init_3_4_11_13
void FASTCALL call_3_4_11_13(SH2_struct* sh)
{
  SH2dmuls(sh,4,11);
}
#define init_3_4_12_13
void FASTCALL call_3_4_12_13(SH2_struct* sh)
{
  SH2dmuls(sh,4,12);
}
#define init_3_4_13_13
void FASTCALL call_3_4_13_13(SH2_struct* sh)
{
  SH2dmuls(sh,4,13);
}
#define init_3_4_14_13
void FASTCALL call_3_4_14_13(SH2_struct* sh)
{
  SH2dmuls(sh,4,14);
}
#define init_3_4_15_13
void FASTCALL call_3_4_15_13(SH2_struct* sh)
{
  SH2dmuls(sh,4,15);
}
#define init_3_5_0_13
void FASTCALL call_3_5_0_13(SH2_struct* sh)
{
  SH2dmuls(sh,5,0);
}
#define init_3_5_1_13
void FASTCALL call_3_5_1_13(SH2_struct* sh)
{
  SH2dmuls(sh,5,1);
}
#define init_3_5_2_13
void FASTCALL call_3_5_2_13(SH2_struct* sh)
{
  SH2dmuls(sh,5,2);
}
#define init_3_5_3_13
void FASTCALL call_3_5_3_13(SH2_struct* sh)
{
  SH2dmuls(sh,5,3);
}
#define init_3_5_4_13
void FASTCALL call_3_5_4_13(SH2_struct* sh)
{
  SH2dmuls(sh,5,4);
}
#define init_3_5_5_13
void FASTCALL call_3_5_5_13(SH2_struct* sh)
{
  SH2dmuls(sh,5,5);
}
#define init_3_5_6_13
void FASTCALL call_3_5_6_13(SH2_struct* sh)
{
  SH2dmuls(sh,5,6);
}
#define init_3_5_7_13
void FASTCALL call_3_5_7_13(SH2_struct* sh)
{
  SH2dmuls(sh,5,7);
}
#define init_3_5_8_13
void FASTCALL call_3_5_8_13(SH2_struct* sh)
{
  SH2dmuls(sh,5,8);
}
#define init_3_5_9_13
void FASTCALL call_3_5_9_13(SH2_struct* sh)
{
  SH2dmuls(sh,5,9);
}
#define init_3_5_10_13
void FASTCALL call_3_5_10_13(SH2_struct* sh)
{
  SH2dmuls(sh,5,10);
}
#define init_3_5_11_13
void FASTCALL call_3_5_11_13(SH2_struct* sh)
{
  SH2dmuls(sh,5,11);
}
#define init_3_5_12_13
void FASTCALL call_3_5_12_13(SH2_struct* sh)
{
  SH2dmuls(sh,5,12);
}
#define init_3_5_13_13
void FASTCALL call_3_5_13_13(SH2_struct* sh)
{
  SH2dmuls(sh,5,13);
}
#define init_3_5_14_13
void FASTCALL call_3_5_14_13(SH2_struct* sh)
{
  SH2dmuls(sh,5,14);
}
#define init_3_5_15_13
void FASTCALL call_3_5_15_13(SH2_struct* sh)
{
  SH2dmuls(sh,5,15);
}
#define init_3_6_0_13
void FASTCALL call_3_6_0_13(SH2_struct* sh)
{
  SH2dmuls(sh,6,0);
}
#define init_3_6_1_13
void FASTCALL call_3_6_1_13(SH2_struct* sh)
{
  SH2dmuls(sh,6,1);
}
#define init_3_6_2_13
void FASTCALL call_3_6_2_13(SH2_struct* sh)
{
  SH2dmuls(sh,6,2);
}
#define init_3_6_3_13
void FASTCALL call_3_6_3_13(SH2_struct* sh)
{
  SH2dmuls(sh,6,3);
}
#define init_3_6_4_13
void FASTCALL call_3_6_4_13(SH2_struct* sh)
{
  SH2dmuls(sh,6,4);
}
#define init_3_6_5_13
void FASTCALL call_3_6_5_13(SH2_struct* sh)
{
  SH2dmuls(sh,6,5);
}
#define init_3_6_6_13
void FASTCALL call_3_6_6_13(SH2_struct* sh)
{
  SH2dmuls(sh,6,6);
}
#define init_3_6_7_13
void FASTCALL call_3_6_7_13(SH2_struct* sh)
{
  SH2dmuls(sh,6,7);
}
#define init_3_6_8_13
void FASTCALL call_3_6_8_13(SH2_struct* sh)
{
  SH2dmuls(sh,6,8);
}
#define init_3_6_9_13
void FASTCALL call_3_6_9_13(SH2_struct* sh)
{
  SH2dmuls(sh,6,9);
}
#define init_3_6_10_13
void FASTCALL call_3_6_10_13(SH2_struct* sh)
{
  SH2dmuls(sh,6,10);
}
#define init_3_6_11_13
void FASTCALL call_3_6_11_13(SH2_struct* sh)
{
  SH2dmuls(sh,6,11);
}
#define init_3_6_12_13
void FASTCALL call_3_6_12_13(SH2_struct* sh)
{
  SH2dmuls(sh,6,12);
}
#define init_3_6_13_13
void FASTCALL call_3_6_13_13(SH2_struct* sh)
{
  SH2dmuls(sh,6,13);
}
#define init_3_6_14_13
void FASTCALL call_3_6_14_13(SH2_struct* sh)
{
  SH2dmuls(sh,6,14);
}
#define init_3_6_15_13
void FASTCALL call_3_6_15_13(SH2_struct* sh)
{
  SH2dmuls(sh,6,15);
}
#define init_3_7_0_13
void FASTCALL call_3_7_0_13(SH2_struct* sh)
{
  SH2dmuls(sh,7,0);
}
#define init_3_7_1_13
void FASTCALL call_3_7_1_13(SH2_struct* sh)
{
  SH2dmuls(sh,7,1);
}
#define init_3_7_2_13
void FASTCALL call_3_7_2_13(SH2_struct* sh)
{
  SH2dmuls(sh,7,2);
}
#define init_3_7_3_13
void FASTCALL call_3_7_3_13(SH2_struct* sh)
{
  SH2dmuls(sh,7,3);
}
#define init_3_7_4_13
void FASTCALL call_3_7_4_13(SH2_struct* sh)
{
  SH2dmuls(sh,7,4);
}
#define init_3_7_5_13
void FASTCALL call_3_7_5_13(SH2_struct* sh)
{
  SH2dmuls(sh,7,5);
}
#define init_3_7_6_13
void FASTCALL call_3_7_6_13(SH2_struct* sh)
{
  SH2dmuls(sh,7,6);
}
#define init_3_7_7_13
void FASTCALL call_3_7_7_13(SH2_struct* sh)
{
  SH2dmuls(sh,7,7);
}
#define init_3_7_8_13
void FASTCALL call_3_7_8_13(SH2_struct* sh)
{
  SH2dmuls(sh,7,8);
}
#define init_3_7_9_13
void FASTCALL call_3_7_9_13(SH2_struct* sh)
{
  SH2dmuls(sh,7,9);
}
#define init_3_7_10_13
void FASTCALL call_3_7_10_13(SH2_struct* sh)
{
  SH2dmuls(sh,7,10);
}
#define init_3_7_11_13
void FASTCALL call_3_7_11_13(SH2_struct* sh)
{
  SH2dmuls(sh,7,11);
}
#define init_3_7_12_13
void FASTCALL call_3_7_12_13(SH2_struct* sh)
{
  SH2dmuls(sh,7,12);
}
#define init_3_7_13_13
void FASTCALL call_3_7_13_13(SH2_struct* sh)
{
  SH2dmuls(sh,7,13);
}
#define init_3_7_14_13
void FASTCALL call_3_7_14_13(SH2_struct* sh)
{
  SH2dmuls(sh,7,14);
}
#define init_3_7_15_13
void FASTCALL call_3_7_15_13(SH2_struct* sh)
{
  SH2dmuls(sh,7,15);
}
#define init_3_8_0_13
void FASTCALL call_3_8_0_13(SH2_struct* sh)
{
  SH2dmuls(sh,8,0);
}
#define init_3_8_1_13
void FASTCALL call_3_8_1_13(SH2_struct* sh)
{
  SH2dmuls(sh,8,1);
}
#define init_3_8_2_13
void FASTCALL call_3_8_2_13(SH2_struct* sh)
{
  SH2dmuls(sh,8,2);
}
#define init_3_8_3_13
void FASTCALL call_3_8_3_13(SH2_struct* sh)
{
  SH2dmuls(sh,8,3);
}
#define init_3_8_4_13
void FASTCALL call_3_8_4_13(SH2_struct* sh)
{
  SH2dmuls(sh,8,4);
}
#define init_3_8_5_13
void FASTCALL call_3_8_5_13(SH2_struct* sh)
{
  SH2dmuls(sh,8,5);
}
#define init_3_8_6_13
void FASTCALL call_3_8_6_13(SH2_struct* sh)
{
  SH2dmuls(sh,8,6);
}
#define init_3_8_7_13
void FASTCALL call_3_8_7_13(SH2_struct* sh)
{
  SH2dmuls(sh,8,7);
}
#define init_3_8_8_13
void FASTCALL call_3_8_8_13(SH2_struct* sh)
{
  SH2dmuls(sh,8,8);
}
#define init_3_8_9_13
void FASTCALL call_3_8_9_13(SH2_struct* sh)
{
  SH2dmuls(sh,8,9);
}
#define init_3_8_10_13
void FASTCALL call_3_8_10_13(SH2_struct* sh)
{
  SH2dmuls(sh,8,10);
}
#define init_3_8_11_13
void FASTCALL call_3_8_11_13(SH2_struct* sh)
{
  SH2dmuls(sh,8,11);
}
#define init_3_8_12_13
void FASTCALL call_3_8_12_13(SH2_struct* sh)
{
  SH2dmuls(sh,8,12);
}
#define init_3_8_13_13
void FASTCALL call_3_8_13_13(SH2_struct* sh)
{
  SH2dmuls(sh,8,13);
}
#define init_3_8_14_13
void FASTCALL call_3_8_14_13(SH2_struct* sh)
{
  SH2dmuls(sh,8,14);
}
#define init_3_8_15_13
void FASTCALL call_3_8_15_13(SH2_struct* sh)
{
  SH2dmuls(sh,8,15);
}
#define init_3_9_0_13
void FASTCALL call_3_9_0_13(SH2_struct* sh)
{
  SH2dmuls(sh,9,0);
}
#define init_3_9_1_13
void FASTCALL call_3_9_1_13(SH2_struct* sh)
{
  SH2dmuls(sh,9,1);
}
#define init_3_9_2_13
void FASTCALL call_3_9_2_13(SH2_struct* sh)
{
  SH2dmuls(sh,9,2);
}
#define init_3_9_3_13
void FASTCALL call_3_9_3_13(SH2_struct* sh)
{
  SH2dmuls(sh,9,3);
}
#define init_3_9_4_13
void FASTCALL call_3_9_4_13(SH2_struct* sh)
{
  SH2dmuls(sh,9,4);
}
#define init_3_9_5_13
void FASTCALL call_3_9_5_13(SH2_struct* sh)
{
  SH2dmuls(sh,9,5);
}
#define init_3_9_6_13
void FASTCALL call_3_9_6_13(SH2_struct* sh)
{
  SH2dmuls(sh,9,6);
}
#define init_3_9_7_13
void FASTCALL call_3_9_7_13(SH2_struct* sh)
{
  SH2dmuls(sh,9,7);
}
#define init_3_9_8_13
void FASTCALL call_3_9_8_13(SH2_struct* sh)
{
  SH2dmuls(sh,9,8);
}
#define init_3_9_9_13
void FASTCALL call_3_9_9_13(SH2_struct* sh)
{
  SH2dmuls(sh,9,9);
}
#define init_3_9_10_13
void FASTCALL call_3_9_10_13(SH2_struct* sh)
{
  SH2dmuls(sh,9,10);
}
#define init_3_9_11_13
void FASTCALL call_3_9_11_13(SH2_struct* sh)
{
  SH2dmuls(sh,9,11);
}
#define init_3_9_12_13
void FASTCALL call_3_9_12_13(SH2_struct* sh)
{
  SH2dmuls(sh,9,12);
}
#define init_3_9_13_13
void FASTCALL call_3_9_13_13(SH2_struct* sh)
{
  SH2dmuls(sh,9,13);
}
#define init_3_9_14_13
void FASTCALL call_3_9_14_13(SH2_struct* sh)
{
  SH2dmuls(sh,9,14);
}
#define init_3_9_15_13
void FASTCALL call_3_9_15_13(SH2_struct* sh)
{
  SH2dmuls(sh,9,15);
}
#define init_3_10_0_13
void FASTCALL call_3_10_0_13(SH2_struct* sh)
{
  SH2dmuls(sh,10,0);
}
#define init_3_10_1_13
void FASTCALL call_3_10_1_13(SH2_struct* sh)
{
  SH2dmuls(sh,10,1);
}
#define init_3_10_2_13
void FASTCALL call_3_10_2_13(SH2_struct* sh)
{
  SH2dmuls(sh,10,2);
}
#define init_3_10_3_13
void FASTCALL call_3_10_3_13(SH2_struct* sh)
{
  SH2dmuls(sh,10,3);
}
#define init_3_10_4_13
void FASTCALL call_3_10_4_13(SH2_struct* sh)
{
  SH2dmuls(sh,10,4);
}
#define init_3_10_5_13
void FASTCALL call_3_10_5_13(SH2_struct* sh)
{
  SH2dmuls(sh,10,5);
}
#define init_3_10_6_13
void FASTCALL call_3_10_6_13(SH2_struct* sh)
{
  SH2dmuls(sh,10,6);
}
#define init_3_10_7_13
void FASTCALL call_3_10_7_13(SH2_struct* sh)
{
  SH2dmuls(sh,10,7);
}
#define init_3_10_8_13
void FASTCALL call_3_10_8_13(SH2_struct* sh)
{
  SH2dmuls(sh,10,8);
}
#define init_3_10_9_13
void FASTCALL call_3_10_9_13(SH2_struct* sh)
{
  SH2dmuls(sh,10,9);
}
#define init_3_10_10_13
void FASTCALL call_3_10_10_13(SH2_struct* sh)
{
  SH2dmuls(sh,10,10);
}
#define init_3_10_11_13
void FASTCALL call_3_10_11_13(SH2_struct* sh)
{
  SH2dmuls(sh,10,11);
}
#define init_3_10_12_13
void FASTCALL call_3_10_12_13(SH2_struct* sh)
{
  SH2dmuls(sh,10,12);
}
#define init_3_10_13_13
void FASTCALL call_3_10_13_13(SH2_struct* sh)
{
  SH2dmuls(sh,10,13);
}
#define init_3_10_14_13
void FASTCALL call_3_10_14_13(SH2_struct* sh)
{
  SH2dmuls(sh,10,14);
}
#define init_3_10_15_13
void FASTCALL call_3_10_15_13(SH2_struct* sh)
{
  SH2dmuls(sh,10,15);
}
#define init_3_11_0_13
void FASTCALL call_3_11_0_13(SH2_struct* sh)
{
  SH2dmuls(sh,11,0);
}
#define init_3_11_1_13
void FASTCALL call_3_11_1_13(SH2_struct* sh)
{
  SH2dmuls(sh,11,1);
}
#define init_3_11_2_13
void FASTCALL call_3_11_2_13(SH2_struct* sh)
{
  SH2dmuls(sh,11,2);
}
#define init_3_11_3_13
void FASTCALL call_3_11_3_13(SH2_struct* sh)
{
  SH2dmuls(sh,11,3);
}
#define init_3_11_4_13
void FASTCALL call_3_11_4_13(SH2_struct* sh)
{
  SH2dmuls(sh,11,4);
}
#define init_3_11_5_13
void FASTCALL call_3_11_5_13(SH2_struct* sh)
{
  SH2dmuls(sh,11,5);
}
#define init_3_11_6_13
void FASTCALL call_3_11_6_13(SH2_struct* sh)
{
  SH2dmuls(sh,11,6);
}
#define init_3_11_7_13
void FASTCALL call_3_11_7_13(SH2_struct* sh)
{
  SH2dmuls(sh,11,7);
}
#define init_3_11_8_13
void FASTCALL call_3_11_8_13(SH2_struct* sh)
{
  SH2dmuls(sh,11,8);
}
#define init_3_11_9_13
void FASTCALL call_3_11_9_13(SH2_struct* sh)
{
  SH2dmuls(sh,11,9);
}
#define init_3_11_10_13
void FASTCALL call_3_11_10_13(SH2_struct* sh)
{
  SH2dmuls(sh,11,10);
}
#define init_3_11_11_13
void FASTCALL call_3_11_11_13(SH2_struct* sh)
{
  SH2dmuls(sh,11,11);
}
#define init_3_11_12_13
void FASTCALL call_3_11_12_13(SH2_struct* sh)
{
  SH2dmuls(sh,11,12);
}
#define init_3_11_13_13
void FASTCALL call_3_11_13_13(SH2_struct* sh)
{
  SH2dmuls(sh,11,13);
}
#define init_3_11_14_13
void FASTCALL call_3_11_14_13(SH2_struct* sh)
{
  SH2dmuls(sh,11,14);
}
#define init_3_11_15_13
void FASTCALL call_3_11_15_13(SH2_struct* sh)
{
  SH2dmuls(sh,11,15);
}
#define init_3_12_0_13
void FASTCALL call_3_12_0_13(SH2_struct* sh)
{
  SH2dmuls(sh,12,0);
}
#define init_3_12_1_13
void FASTCALL call_3_12_1_13(SH2_struct* sh)
{
  SH2dmuls(sh,12,1);
}
#define init_3_12_2_13
void FASTCALL call_3_12_2_13(SH2_struct* sh)
{
  SH2dmuls(sh,12,2);
}
#define init_3_12_3_13
void FASTCALL call_3_12_3_13(SH2_struct* sh)
{
  SH2dmuls(sh,12,3);
}
#define init_3_12_4_13
void FASTCALL call_3_12_4_13(SH2_struct* sh)
{
  SH2dmuls(sh,12,4);
}
#define init_3_12_5_13
void FASTCALL call_3_12_5_13(SH2_struct* sh)
{
  SH2dmuls(sh,12,5);
}
#define init_3_12_6_13
void FASTCALL call_3_12_6_13(SH2_struct* sh)
{
  SH2dmuls(sh,12,6);
}
#define init_3_12_7_13
void FASTCALL call_3_12_7_13(SH2_struct* sh)
{
  SH2dmuls(sh,12,7);
}
#define init_3_12_8_13
void FASTCALL call_3_12_8_13(SH2_struct* sh)
{
  SH2dmuls(sh,12,8);
}
#define init_3_12_9_13
void FASTCALL call_3_12_9_13(SH2_struct* sh)
{
  SH2dmuls(sh,12,9);
}
#define init_3_12_10_13
void FASTCALL call_3_12_10_13(SH2_struct* sh)
{
  SH2dmuls(sh,12,10);
}
#define init_3_12_11_13
void FASTCALL call_3_12_11_13(SH2_struct* sh)
{
  SH2dmuls(sh,12,11);
}
#define init_3_12_12_13
void FASTCALL call_3_12_12_13(SH2_struct* sh)
{
  SH2dmuls(sh,12,12);
}
#define init_3_12_13_13
void FASTCALL call_3_12_13_13(SH2_struct* sh)
{
  SH2dmuls(sh,12,13);
}
#define init_3_12_14_13
void FASTCALL call_3_12_14_13(SH2_struct* sh)
{
  SH2dmuls(sh,12,14);
}
#define init_3_12_15_13
void FASTCALL call_3_12_15_13(SH2_struct* sh)
{
  SH2dmuls(sh,12,15);
}
#define init_3_13_0_13
void FASTCALL call_3_13_0_13(SH2_struct* sh)
{
  SH2dmuls(sh,13,0);
}
#define init_3_13_1_13
void FASTCALL call_3_13_1_13(SH2_struct* sh)
{
  SH2dmuls(sh,13,1);
}
#define init_3_13_2_13
void FASTCALL call_3_13_2_13(SH2_struct* sh)
{
  SH2dmuls(sh,13,2);
}
#define init_3_13_3_13
void FASTCALL call_3_13_3_13(SH2_struct* sh)
{
  SH2dmuls(sh,13,3);
}
#define init_3_13_4_13
void FASTCALL call_3_13_4_13(SH2_struct* sh)
{
  SH2dmuls(sh,13,4);
}
#define init_3_13_5_13
void FASTCALL call_3_13_5_13(SH2_struct* sh)
{
  SH2dmuls(sh,13,5);
}
#define init_3_13_6_13
void FASTCALL call_3_13_6_13(SH2_struct* sh)
{
  SH2dmuls(sh,13,6);
}
#define init_3_13_7_13
void FASTCALL call_3_13_7_13(SH2_struct* sh)
{
  SH2dmuls(sh,13,7);
}
#define init_3_13_8_13
void FASTCALL call_3_13_8_13(SH2_struct* sh)
{
  SH2dmuls(sh,13,8);
}
#define init_3_13_9_13
void FASTCALL call_3_13_9_13(SH2_struct* sh)
{
  SH2dmuls(sh,13,9);
}
#define init_3_13_10_13
void FASTCALL call_3_13_10_13(SH2_struct* sh)
{
  SH2dmuls(sh,13,10);
}
#define init_3_13_11_13
void FASTCALL call_3_13_11_13(SH2_struct* sh)
{
  SH2dmuls(sh,13,11);
}
#define init_3_13_12_13
void FASTCALL call_3_13_12_13(SH2_struct* sh)
{
  SH2dmuls(sh,13,12);
}
#define init_3_13_13_13
void FASTCALL call_3_13_13_13(SH2_struct* sh)
{
  SH2dmuls(sh,13,13);
}
#define init_3_13_14_13
void FASTCALL call_3_13_14_13(SH2_struct* sh)
{
  SH2dmuls(sh,13,14);
}
#define init_3_13_15_13
void FASTCALL call_3_13_15_13(SH2_struct* sh)
{
  SH2dmuls(sh,13,15);
}
#define init_3_14_0_13
void FASTCALL call_3_14_0_13(SH2_struct* sh)
{
  SH2dmuls(sh,14,0);
}
#define init_3_14_1_13
void FASTCALL call_3_14_1_13(SH2_struct* sh)
{
  SH2dmuls(sh,14,1);
}
#define init_3_14_2_13
void FASTCALL call_3_14_2_13(SH2_struct* sh)
{
  SH2dmuls(sh,14,2);
}
#define init_3_14_3_13
void FASTCALL call_3_14_3_13(SH2_struct* sh)
{
  SH2dmuls(sh,14,3);
}
#define init_3_14_4_13
void FASTCALL call_3_14_4_13(SH2_struct* sh)
{
  SH2dmuls(sh,14,4);
}
#define init_3_14_5_13
void FASTCALL call_3_14_5_13(SH2_struct* sh)
{
  SH2dmuls(sh,14,5);
}
#define init_3_14_6_13
void FASTCALL call_3_14_6_13(SH2_struct* sh)
{
  SH2dmuls(sh,14,6);
}
#define init_3_14_7_13
void FASTCALL call_3_14_7_13(SH2_struct* sh)
{
  SH2dmuls(sh,14,7);
}
#define init_3_14_8_13
void FASTCALL call_3_14_8_13(SH2_struct* sh)
{
  SH2dmuls(sh,14,8);
}
#define init_3_14_9_13
void FASTCALL call_3_14_9_13(SH2_struct* sh)
{
  SH2dmuls(sh,14,9);
}
#define init_3_14_10_13
void FASTCALL call_3_14_10_13(SH2_struct* sh)
{
  SH2dmuls(sh,14,10);
}
#define init_3_14_11_13
void FASTCALL call_3_14_11_13(SH2_struct* sh)
{
  SH2dmuls(sh,14,11);
}
#define init_3_14_12_13
void FASTCALL call_3_14_12_13(SH2_struct* sh)
{
  SH2dmuls(sh,14,12);
}
#define init_3_14_13_13
void FASTCALL call_3_14_13_13(SH2_struct* sh)
{
  SH2dmuls(sh,14,13);
}
#define init_3_14_14_13
void FASTCALL call_3_14_14_13(SH2_struct* sh)
{
  SH2dmuls(sh,14,14);
}
#define init_3_14_15_13
void FASTCALL call_3_14_15_13(SH2_struct* sh)
{
  SH2dmuls(sh,14,15);
}
#define init_3_15_0_13
void FASTCALL call_3_15_0_13(SH2_struct* sh)
{
  SH2dmuls(sh,15,0);
}
#define init_3_15_1_13
void FASTCALL call_3_15_1_13(SH2_struct* sh)
{
  SH2dmuls(sh,15,1);
}
#define init_3_15_2_13
void FASTCALL call_3_15_2_13(SH2_struct* sh)
{
  SH2dmuls(sh,15,2);
}
#define init_3_15_3_13
void FASTCALL call_3_15_3_13(SH2_struct* sh)
{
  SH2dmuls(sh,15,3);
}
#define init_3_15_4_13
void FASTCALL call_3_15_4_13(SH2_struct* sh)
{
  SH2dmuls(sh,15,4);
}
#define init_3_15_5_13
void FASTCALL call_3_15_5_13(SH2_struct* sh)
{
  SH2dmuls(sh,15,5);
}
#define init_3_15_6_13
void FASTCALL call_3_15_6_13(SH2_struct* sh)
{
  SH2dmuls(sh,15,6);
}
#define init_3_15_7_13
void FASTCALL call_3_15_7_13(SH2_struct* sh)
{
  SH2dmuls(sh,15,7);
}
#define init_3_15_8_13
void FASTCALL call_3_15_8_13(SH2_struct* sh)
{
  SH2dmuls(sh,15,8);
}
#define init_3_15_9_13
void FASTCALL call_3_15_9_13(SH2_struct* sh)
{
  SH2dmuls(sh,15,9);
}
#define init_3_15_10_13
void FASTCALL call_3_15_10_13(SH2_struct* sh)
{
  SH2dmuls(sh,15,10);
}
#define init_3_15_11_13
void FASTCALL call_3_15_11_13(SH2_struct* sh)
{
  SH2dmuls(sh,15,11);
}
#define init_3_15_12_13
void FASTCALL call_3_15_12_13(SH2_struct* sh)
{
  SH2dmuls(sh,15,12);
}
#define init_3_15_13_13
void FASTCALL call_3_15_13_13(SH2_struct* sh)
{
  SH2dmuls(sh,15,13);
}
#define init_3_15_14_13
void FASTCALL call_3_15_14_13(SH2_struct* sh)
{
  SH2dmuls(sh,15,14);
}
#define init_3_15_15_13
void FASTCALL call_3_15_15_13(SH2_struct* sh)
{
  SH2dmuls(sh,15,15);
}
#define init_3_0_0_4
void FASTCALL call_3_0_0_4(SH2_struct* sh)
{
  SH2div1(sh,0,0);
}
#define init_3_0_1_4
void FASTCALL call_3_0_1_4(SH2_struct* sh)
{
  SH2div1(sh,0,1);
}
#define init_3_0_2_4
void FASTCALL call_3_0_2_4(SH2_struct* sh)
{
  SH2div1(sh,0,2);
}
#define init_3_0_3_4
void FASTCALL call_3_0_3_4(SH2_struct* sh)
{
  SH2div1(sh,0,3);
}
#define init_3_0_4_4
void FASTCALL call_3_0_4_4(SH2_struct* sh)
{
  SH2div1(sh,0,4);
}
#define init_3_0_5_4
void FASTCALL call_3_0_5_4(SH2_struct* sh)
{
  SH2div1(sh,0,5);
}
#define init_3_0_6_4
void FASTCALL call_3_0_6_4(SH2_struct* sh)
{
  SH2div1(sh,0,6);
}
#define init_3_0_7_4
void FASTCALL call_3_0_7_4(SH2_struct* sh)
{
  SH2div1(sh,0,7);
}
#define init_3_0_8_4
void FASTCALL call_3_0_8_4(SH2_struct* sh)
{
  SH2div1(sh,0,8);
}
#define init_3_0_9_4
void FASTCALL call_3_0_9_4(SH2_struct* sh)
{
  SH2div1(sh,0,9);
}
#define init_3_0_10_4
void FASTCALL call_3_0_10_4(SH2_struct* sh)
{
  SH2div1(sh,0,10);
}
#define init_3_0_11_4
void FASTCALL call_3_0_11_4(SH2_struct* sh)
{
  SH2div1(sh,0,11);
}
#define init_3_0_12_4
void FASTCALL call_3_0_12_4(SH2_struct* sh)
{
  SH2div1(sh,0,12);
}
#define init_3_0_13_4
void FASTCALL call_3_0_13_4(SH2_struct* sh)
{
  SH2div1(sh,0,13);
}
#define init_3_0_14_4
void FASTCALL call_3_0_14_4(SH2_struct* sh)
{
  SH2div1(sh,0,14);
}
#define init_3_0_15_4
void FASTCALL call_3_0_15_4(SH2_struct* sh)
{
  SH2div1(sh,0,15);
}
#define init_3_1_0_4
void FASTCALL call_3_1_0_4(SH2_struct* sh)
{
  SH2div1(sh,1,0);
}
#define init_3_1_1_4
void FASTCALL call_3_1_1_4(SH2_struct* sh)
{
  SH2div1(sh,1,1);
}
#define init_3_1_2_4
void FASTCALL call_3_1_2_4(SH2_struct* sh)
{
  SH2div1(sh,1,2);
}
#define init_3_1_3_4
void FASTCALL call_3_1_3_4(SH2_struct* sh)
{
  SH2div1(sh,1,3);
}
#define init_3_1_4_4
void FASTCALL call_3_1_4_4(SH2_struct* sh)
{
  SH2div1(sh,1,4);
}
#define init_3_1_5_4
void FASTCALL call_3_1_5_4(SH2_struct* sh)
{
  SH2div1(sh,1,5);
}
#define init_3_1_6_4
void FASTCALL call_3_1_6_4(SH2_struct* sh)
{
  SH2div1(sh,1,6);
}
#define init_3_1_7_4
void FASTCALL call_3_1_7_4(SH2_struct* sh)
{
  SH2div1(sh,1,7);
}
#define init_3_1_8_4
void FASTCALL call_3_1_8_4(SH2_struct* sh)
{
  SH2div1(sh,1,8);
}
#define init_3_1_9_4
void FASTCALL call_3_1_9_4(SH2_struct* sh)
{
  SH2div1(sh,1,9);
}
#define init_3_1_10_4
void FASTCALL call_3_1_10_4(SH2_struct* sh)
{
  SH2div1(sh,1,10);
}
#define init_3_1_11_4
void FASTCALL call_3_1_11_4(SH2_struct* sh)
{
  SH2div1(sh,1,11);
}
#define init_3_1_12_4
void FASTCALL call_3_1_12_4(SH2_struct* sh)
{
  SH2div1(sh,1,12);
}
#define init_3_1_13_4
void FASTCALL call_3_1_13_4(SH2_struct* sh)
{
  SH2div1(sh,1,13);
}
#define init_3_1_14_4
void FASTCALL call_3_1_14_4(SH2_struct* sh)
{
  SH2div1(sh,1,14);
}
#define init_3_1_15_4
void FASTCALL call_3_1_15_4(SH2_struct* sh)
{
  SH2div1(sh,1,15);
}
#define init_3_2_0_4
void FASTCALL call_3_2_0_4(SH2_struct* sh)
{
  SH2div1(sh,2,0);
}
#define init_3_2_1_4
void FASTCALL call_3_2_1_4(SH2_struct* sh)
{
  SH2div1(sh,2,1);
}
#define init_3_2_2_4
void FASTCALL call_3_2_2_4(SH2_struct* sh)
{
  SH2div1(sh,2,2);
}
#define init_3_2_3_4
void FASTCALL call_3_2_3_4(SH2_struct* sh)
{
  SH2div1(sh,2,3);
}
#define init_3_2_4_4
void FASTCALL call_3_2_4_4(SH2_struct* sh)
{
  SH2div1(sh,2,4);
}
#define init_3_2_5_4
void FASTCALL call_3_2_5_4(SH2_struct* sh)
{
  SH2div1(sh,2,5);
}
#define init_3_2_6_4
void FASTCALL call_3_2_6_4(SH2_struct* sh)
{
  SH2div1(sh,2,6);
}
#define init_3_2_7_4
void FASTCALL call_3_2_7_4(SH2_struct* sh)
{
  SH2div1(sh,2,7);
}
#define init_3_2_8_4
void FASTCALL call_3_2_8_4(SH2_struct* sh)
{
  SH2div1(sh,2,8);
}
#define init_3_2_9_4
void FASTCALL call_3_2_9_4(SH2_struct* sh)
{
  SH2div1(sh,2,9);
}
#define init_3_2_10_4
void FASTCALL call_3_2_10_4(SH2_struct* sh)
{
  SH2div1(sh,2,10);
}
#define init_3_2_11_4
void FASTCALL call_3_2_11_4(SH2_struct* sh)
{
  SH2div1(sh,2,11);
}
#define init_3_2_12_4
void FASTCALL call_3_2_12_4(SH2_struct* sh)
{
  SH2div1(sh,2,12);
}
#define init_3_2_13_4
void FASTCALL call_3_2_13_4(SH2_struct* sh)
{
  SH2div1(sh,2,13);
}
#define init_3_2_14_4
void FASTCALL call_3_2_14_4(SH2_struct* sh)
{
  SH2div1(sh,2,14);
}
#define init_3_2_15_4
void FASTCALL call_3_2_15_4(SH2_struct* sh)
{
  SH2div1(sh,2,15);
}
#define init_3_3_0_4
void FASTCALL call_3_3_0_4(SH2_struct* sh)
{
  SH2div1(sh,3,0);
}
#define init_3_3_1_4
void FASTCALL call_3_3_1_4(SH2_struct* sh)
{
  SH2div1(sh,3,1);
}
#define init_3_3_2_4
void FASTCALL call_3_3_2_4(SH2_struct* sh)
{
  SH2div1(sh,3,2);
}
#define init_3_3_3_4
void FASTCALL call_3_3_3_4(SH2_struct* sh)
{
  SH2div1(sh,3,3);
}
#define init_3_3_4_4
void FASTCALL call_3_3_4_4(SH2_struct* sh)
{
  SH2div1(sh,3,4);
}
#define init_3_3_5_4
void FASTCALL call_3_3_5_4(SH2_struct* sh)
{
  SH2div1(sh,3,5);
}
#define init_3_3_6_4
void FASTCALL call_3_3_6_4(SH2_struct* sh)
{
  SH2div1(sh,3,6);
}
#define init_3_3_7_4
void FASTCALL call_3_3_7_4(SH2_struct* sh)
{
  SH2div1(sh,3,7);
}
#define init_3_3_8_4
void FASTCALL call_3_3_8_4(SH2_struct* sh)
{
  SH2div1(sh,3,8);
}
#define init_3_3_9_4
void FASTCALL call_3_3_9_4(SH2_struct* sh)
{
  SH2div1(sh,3,9);
}
#define init_3_3_10_4
void FASTCALL call_3_3_10_4(SH2_struct* sh)
{
  SH2div1(sh,3,10);
}
#define init_3_3_11_4
void FASTCALL call_3_3_11_4(SH2_struct* sh)
{
  SH2div1(sh,3,11);
}
#define init_3_3_12_4
void FASTCALL call_3_3_12_4(SH2_struct* sh)
{
  SH2div1(sh,3,12);
}
#define init_3_3_13_4
void FASTCALL call_3_3_13_4(SH2_struct* sh)
{
  SH2div1(sh,3,13);
}
#define init_3_3_14_4
void FASTCALL call_3_3_14_4(SH2_struct* sh)
{
  SH2div1(sh,3,14);
}
#define init_3_3_15_4
void FASTCALL call_3_3_15_4(SH2_struct* sh)
{
  SH2div1(sh,3,15);
}
#define init_3_4_0_4
void FASTCALL call_3_4_0_4(SH2_struct* sh)
{
  SH2div1(sh,4,0);
}
#define init_3_4_1_4
void FASTCALL call_3_4_1_4(SH2_struct* sh)
{
  SH2div1(sh,4,1);
}
#define init_3_4_2_4
void FASTCALL call_3_4_2_4(SH2_struct* sh)
{
  SH2div1(sh,4,2);
}
#define init_3_4_3_4
void FASTCALL call_3_4_3_4(SH2_struct* sh)
{
  SH2div1(sh,4,3);
}
#define init_3_4_4_4
void FASTCALL call_3_4_4_4(SH2_struct* sh)
{
  SH2div1(sh,4,4);
}
#define init_3_4_5_4
void FASTCALL call_3_4_5_4(SH2_struct* sh)
{
  SH2div1(sh,4,5);
}
#define init_3_4_6_4
void FASTCALL call_3_4_6_4(SH2_struct* sh)
{
  SH2div1(sh,4,6);
}
#define init_3_4_7_4
void FASTCALL call_3_4_7_4(SH2_struct* sh)
{
  SH2div1(sh,4,7);
}
#define init_3_4_8_4
void FASTCALL call_3_4_8_4(SH2_struct* sh)
{
  SH2div1(sh,4,8);
}
#define init_3_4_9_4
void FASTCALL call_3_4_9_4(SH2_struct* sh)
{
  SH2div1(sh,4,9);
}
#define init_3_4_10_4
void FASTCALL call_3_4_10_4(SH2_struct* sh)
{
  SH2div1(sh,4,10);
}
#define init_3_4_11_4
void FASTCALL call_3_4_11_4(SH2_struct* sh)
{
  SH2div1(sh,4,11);
}
#define init_3_4_12_4
void FASTCALL call_3_4_12_4(SH2_struct* sh)
{
  SH2div1(sh,4,12);
}
#define init_3_4_13_4
void FASTCALL call_3_4_13_4(SH2_struct* sh)
{
  SH2div1(sh,4,13);
}
#define init_3_4_14_4
void FASTCALL call_3_4_14_4(SH2_struct* sh)
{
  SH2div1(sh,4,14);
}
#define init_3_4_15_4
void FASTCALL call_3_4_15_4(SH2_struct* sh)
{
  SH2div1(sh,4,15);
}
#define init_3_5_0_4
void FASTCALL call_3_5_0_4(SH2_struct* sh)
{
  SH2div1(sh,5,0);
}
#define init_3_5_1_4
void FASTCALL call_3_5_1_4(SH2_struct* sh)
{
  SH2div1(sh,5,1);
}
#define init_3_5_2_4
void FASTCALL call_3_5_2_4(SH2_struct* sh)
{
  SH2div1(sh,5,2);
}
#define init_3_5_3_4
void FASTCALL call_3_5_3_4(SH2_struct* sh)
{
  SH2div1(sh,5,3);
}
#define init_3_5_4_4
void FASTCALL call_3_5_4_4(SH2_struct* sh)
{
  SH2div1(sh,5,4);
}
#define init_3_5_5_4
void FASTCALL call_3_5_5_4(SH2_struct* sh)
{
  SH2div1(sh,5,5);
}
#define init_3_5_6_4
void FASTCALL call_3_5_6_4(SH2_struct* sh)
{
  SH2div1(sh,5,6);
}
#define init_3_5_7_4
void FASTCALL call_3_5_7_4(SH2_struct* sh)
{
  SH2div1(sh,5,7);
}
#define init_3_5_8_4
void FASTCALL call_3_5_8_4(SH2_struct* sh)
{
  SH2div1(sh,5,8);
}
#define init_3_5_9_4
void FASTCALL call_3_5_9_4(SH2_struct* sh)
{
  SH2div1(sh,5,9);
}
#define init_3_5_10_4
void FASTCALL call_3_5_10_4(SH2_struct* sh)
{
  SH2div1(sh,5,10);
}
#define init_3_5_11_4
void FASTCALL call_3_5_11_4(SH2_struct* sh)
{
  SH2div1(sh,5,11);
}
#define init_3_5_12_4
void FASTCALL call_3_5_12_4(SH2_struct* sh)
{
  SH2div1(sh,5,12);
}
#define init_3_5_13_4
void FASTCALL call_3_5_13_4(SH2_struct* sh)
{
  SH2div1(sh,5,13);
}
#define init_3_5_14_4
void FASTCALL call_3_5_14_4(SH2_struct* sh)
{
  SH2div1(sh,5,14);
}
#define init_3_5_15_4
void FASTCALL call_3_5_15_4(SH2_struct* sh)
{
  SH2div1(sh,5,15);
}
#define init_3_6_0_4
void FASTCALL call_3_6_0_4(SH2_struct* sh)
{
  SH2div1(sh,6,0);
}
#define init_3_6_1_4
void FASTCALL call_3_6_1_4(SH2_struct* sh)
{
  SH2div1(sh,6,1);
}
#define init_3_6_2_4
void FASTCALL call_3_6_2_4(SH2_struct* sh)
{
  SH2div1(sh,6,2);
}
#define init_3_6_3_4
void FASTCALL call_3_6_3_4(SH2_struct* sh)
{
  SH2div1(sh,6,3);
}
#define init_3_6_4_4
void FASTCALL call_3_6_4_4(SH2_struct* sh)
{
  SH2div1(sh,6,4);
}
#define init_3_6_5_4
void FASTCALL call_3_6_5_4(SH2_struct* sh)
{
  SH2div1(sh,6,5);
}
#define init_3_6_6_4
void FASTCALL call_3_6_6_4(SH2_struct* sh)
{
  SH2div1(sh,6,6);
}
#define init_3_6_7_4
void FASTCALL call_3_6_7_4(SH2_struct* sh)
{
  SH2div1(sh,6,7);
}
#define init_3_6_8_4
void FASTCALL call_3_6_8_4(SH2_struct* sh)
{
  SH2div1(sh,6,8);
}
#define init_3_6_9_4
void FASTCALL call_3_6_9_4(SH2_struct* sh)
{
  SH2div1(sh,6,9);
}
#define init_3_6_10_4
void FASTCALL call_3_6_10_4(SH2_struct* sh)
{
  SH2div1(sh,6,10);
}
#define init_3_6_11_4
void FASTCALL call_3_6_11_4(SH2_struct* sh)
{
  SH2div1(sh,6,11);
}
#define init_3_6_12_4
void FASTCALL call_3_6_12_4(SH2_struct* sh)
{
  SH2div1(sh,6,12);
}
#define init_3_6_13_4
void FASTCALL call_3_6_13_4(SH2_struct* sh)
{
  SH2div1(sh,6,13);
}
#define init_3_6_14_4
void FASTCALL call_3_6_14_4(SH2_struct* sh)
{
  SH2div1(sh,6,14);
}
#define init_3_6_15_4
void FASTCALL call_3_6_15_4(SH2_struct* sh)
{
  SH2div1(sh,6,15);
}
#define init_3_7_0_4
void FASTCALL call_3_7_0_4(SH2_struct* sh)
{
  SH2div1(sh,7,0);
}
#define init_3_7_1_4
void FASTCALL call_3_7_1_4(SH2_struct* sh)
{
  SH2div1(sh,7,1);
}
#define init_3_7_2_4
void FASTCALL call_3_7_2_4(SH2_struct* sh)
{
  SH2div1(sh,7,2);
}
#define init_3_7_3_4
void FASTCALL call_3_7_3_4(SH2_struct* sh)
{
  SH2div1(sh,7,3);
}
#define init_3_7_4_4
void FASTCALL call_3_7_4_4(SH2_struct* sh)
{
  SH2div1(sh,7,4);
}
#define init_3_7_5_4
void FASTCALL call_3_7_5_4(SH2_struct* sh)
{
  SH2div1(sh,7,5);
}
#define init_3_7_6_4
void FASTCALL call_3_7_6_4(SH2_struct* sh)
{
  SH2div1(sh,7,6);
}
#define init_3_7_7_4
void FASTCALL call_3_7_7_4(SH2_struct* sh)
{
  SH2div1(sh,7,7);
}
#define init_3_7_8_4
void FASTCALL call_3_7_8_4(SH2_struct* sh)
{
  SH2div1(sh,7,8);
}
#define init_3_7_9_4
void FASTCALL call_3_7_9_4(SH2_struct* sh)
{
  SH2div1(sh,7,9);
}
#define init_3_7_10_4
void FASTCALL call_3_7_10_4(SH2_struct* sh)
{
  SH2div1(sh,7,10);
}
#define init_3_7_11_4
void FASTCALL call_3_7_11_4(SH2_struct* sh)
{
  SH2div1(sh,7,11);
}
#define init_3_7_12_4
void FASTCALL call_3_7_12_4(SH2_struct* sh)
{
  SH2div1(sh,7,12);
}
#define init_3_7_13_4
void FASTCALL call_3_7_13_4(SH2_struct* sh)
{
  SH2div1(sh,7,13);
}
#define init_3_7_14_4
void FASTCALL call_3_7_14_4(SH2_struct* sh)
{
  SH2div1(sh,7,14);
}
#define init_3_7_15_4
void FASTCALL call_3_7_15_4(SH2_struct* sh)
{
  SH2div1(sh,7,15);
}
#define init_3_8_0_4
void FASTCALL call_3_8_0_4(SH2_struct* sh)
{
  SH2div1(sh,8,0);
}
#define init_3_8_1_4
void FASTCALL call_3_8_1_4(SH2_struct* sh)
{
  SH2div1(sh,8,1);
}
#define init_3_8_2_4
void FASTCALL call_3_8_2_4(SH2_struct* sh)
{
  SH2div1(sh,8,2);
}
#define init_3_8_3_4
void FASTCALL call_3_8_3_4(SH2_struct* sh)
{
  SH2div1(sh,8,3);
}
#define init_3_8_4_4
void FASTCALL call_3_8_4_4(SH2_struct* sh)
{
  SH2div1(sh,8,4);
}
#define init_3_8_5_4
void FASTCALL call_3_8_5_4(SH2_struct* sh)
{
  SH2div1(sh,8,5);
}
#define init_3_8_6_4
void FASTCALL call_3_8_6_4(SH2_struct* sh)
{
  SH2div1(sh,8,6);
}
#define init_3_8_7_4
void FASTCALL call_3_8_7_4(SH2_struct* sh)
{
  SH2div1(sh,8,7);
}
#define init_3_8_8_4
void FASTCALL call_3_8_8_4(SH2_struct* sh)
{
  SH2div1(sh,8,8);
}
#define init_3_8_9_4
void FASTCALL call_3_8_9_4(SH2_struct* sh)
{
  SH2div1(sh,8,9);
}
#define init_3_8_10_4
void FASTCALL call_3_8_10_4(SH2_struct* sh)
{
  SH2div1(sh,8,10);
}
#define init_3_8_11_4
void FASTCALL call_3_8_11_4(SH2_struct* sh)
{
  SH2div1(sh,8,11);
}
#define init_3_8_12_4
void FASTCALL call_3_8_12_4(SH2_struct* sh)
{
  SH2div1(sh,8,12);
}
#define init_3_8_13_4
void FASTCALL call_3_8_13_4(SH2_struct* sh)
{
  SH2div1(sh,8,13);
}
#define init_3_8_14_4
void FASTCALL call_3_8_14_4(SH2_struct* sh)
{
  SH2div1(sh,8,14);
}
#define init_3_8_15_4
void FASTCALL call_3_8_15_4(SH2_struct* sh)
{
  SH2div1(sh,8,15);
}
#define init_3_9_0_4
void FASTCALL call_3_9_0_4(SH2_struct* sh)
{
  SH2div1(sh,9,0);
}
#define init_3_9_1_4
void FASTCALL call_3_9_1_4(SH2_struct* sh)
{
  SH2div1(sh,9,1);
}
#define init_3_9_2_4
void FASTCALL call_3_9_2_4(SH2_struct* sh)
{
  SH2div1(sh,9,2);
}
#define init_3_9_3_4
void FASTCALL call_3_9_3_4(SH2_struct* sh)
{
  SH2div1(sh,9,3);
}
#define init_3_9_4_4
void FASTCALL call_3_9_4_4(SH2_struct* sh)
{
  SH2div1(sh,9,4);
}
#define init_3_9_5_4
void FASTCALL call_3_9_5_4(SH2_struct* sh)
{
  SH2div1(sh,9,5);
}
#define init_3_9_6_4
void FASTCALL call_3_9_6_4(SH2_struct* sh)
{
  SH2div1(sh,9,6);
}
#define init_3_9_7_4
void FASTCALL call_3_9_7_4(SH2_struct* sh)
{
  SH2div1(sh,9,7);
}
#define init_3_9_8_4
void FASTCALL call_3_9_8_4(SH2_struct* sh)
{
  SH2div1(sh,9,8);
}
#define init_3_9_9_4
void FASTCALL call_3_9_9_4(SH2_struct* sh)
{
  SH2div1(sh,9,9);
}
#define init_3_9_10_4
void FASTCALL call_3_9_10_4(SH2_struct* sh)
{
  SH2div1(sh,9,10);
}
#define init_3_9_11_4
void FASTCALL call_3_9_11_4(SH2_struct* sh)
{
  SH2div1(sh,9,11);
}
#define init_3_9_12_4
void FASTCALL call_3_9_12_4(SH2_struct* sh)
{
  SH2div1(sh,9,12);
}
#define init_3_9_13_4
void FASTCALL call_3_9_13_4(SH2_struct* sh)
{
  SH2div1(sh,9,13);
}
#define init_3_9_14_4
void FASTCALL call_3_9_14_4(SH2_struct* sh)
{
  SH2div1(sh,9,14);
}
#define init_3_9_15_4
void FASTCALL call_3_9_15_4(SH2_struct* sh)
{
  SH2div1(sh,9,15);
}
#define init_3_10_0_4
void FASTCALL call_3_10_0_4(SH2_struct* sh)
{
  SH2div1(sh,10,0);
}
#define init_3_10_1_4
void FASTCALL call_3_10_1_4(SH2_struct* sh)
{
  SH2div1(sh,10,1);
}
#define init_3_10_2_4
void FASTCALL call_3_10_2_4(SH2_struct* sh)
{
  SH2div1(sh,10,2);
}
#define init_3_10_3_4
void FASTCALL call_3_10_3_4(SH2_struct* sh)
{
  SH2div1(sh,10,3);
}
#define init_3_10_4_4
void FASTCALL call_3_10_4_4(SH2_struct* sh)
{
  SH2div1(sh,10,4);
}
#define init_3_10_5_4
void FASTCALL call_3_10_5_4(SH2_struct* sh)
{
  SH2div1(sh,10,5);
}
#define init_3_10_6_4
void FASTCALL call_3_10_6_4(SH2_struct* sh)
{
  SH2div1(sh,10,6);
}
#define init_3_10_7_4
void FASTCALL call_3_10_7_4(SH2_struct* sh)
{
  SH2div1(sh,10,7);
}
#define init_3_10_8_4
void FASTCALL call_3_10_8_4(SH2_struct* sh)
{
  SH2div1(sh,10,8);
}
#define init_3_10_9_4
void FASTCALL call_3_10_9_4(SH2_struct* sh)
{
  SH2div1(sh,10,9);
}
#define init_3_10_10_4
void FASTCALL call_3_10_10_4(SH2_struct* sh)
{
  SH2div1(sh,10,10);
}
#define init_3_10_11_4
void FASTCALL call_3_10_11_4(SH2_struct* sh)
{
  SH2div1(sh,10,11);
}
#define init_3_10_12_4
void FASTCALL call_3_10_12_4(SH2_struct* sh)
{
  SH2div1(sh,10,12);
}
#define init_3_10_13_4
void FASTCALL call_3_10_13_4(SH2_struct* sh)
{
  SH2div1(sh,10,13);
}
#define init_3_10_14_4
void FASTCALL call_3_10_14_4(SH2_struct* sh)
{
  SH2div1(sh,10,14);
}
#define init_3_10_15_4
void FASTCALL call_3_10_15_4(SH2_struct* sh)
{
  SH2div1(sh,10,15);
}
#define init_3_11_0_4
void FASTCALL call_3_11_0_4(SH2_struct* sh)
{
  SH2div1(sh,11,0);
}
#define init_3_11_1_4
void FASTCALL call_3_11_1_4(SH2_struct* sh)
{
  SH2div1(sh,11,1);
}
#define init_3_11_2_4
void FASTCALL call_3_11_2_4(SH2_struct* sh)
{
  SH2div1(sh,11,2);
}
#define init_3_11_3_4
void FASTCALL call_3_11_3_4(SH2_struct* sh)
{
  SH2div1(sh,11,3);
}
#define init_3_11_4_4
void FASTCALL call_3_11_4_4(SH2_struct* sh)
{
  SH2div1(sh,11,4);
}
#define init_3_11_5_4
void FASTCALL call_3_11_5_4(SH2_struct* sh)
{
  SH2div1(sh,11,5);
}
#define init_3_11_6_4
void FASTCALL call_3_11_6_4(SH2_struct* sh)
{
  SH2div1(sh,11,6);
}
#define init_3_11_7_4
void FASTCALL call_3_11_7_4(SH2_struct* sh)
{
  SH2div1(sh,11,7);
}
#define init_3_11_8_4
void FASTCALL call_3_11_8_4(SH2_struct* sh)
{
  SH2div1(sh,11,8);
}
#define init_3_11_9_4
void FASTCALL call_3_11_9_4(SH2_struct* sh)
{
  SH2div1(sh,11,9);
}
#define init_3_11_10_4
void FASTCALL call_3_11_10_4(SH2_struct* sh)
{
  SH2div1(sh,11,10);
}
#define init_3_11_11_4
void FASTCALL call_3_11_11_4(SH2_struct* sh)
{
  SH2div1(sh,11,11);
}
#define init_3_11_12_4
void FASTCALL call_3_11_12_4(SH2_struct* sh)
{
  SH2div1(sh,11,12);
}
#define init_3_11_13_4
void FASTCALL call_3_11_13_4(SH2_struct* sh)
{
  SH2div1(sh,11,13);
}
#define init_3_11_14_4
void FASTCALL call_3_11_14_4(SH2_struct* sh)
{
  SH2div1(sh,11,14);
}
#define init_3_11_15_4
void FASTCALL call_3_11_15_4(SH2_struct* sh)
{
  SH2div1(sh,11,15);
}
#define init_3_12_0_4
void FASTCALL call_3_12_0_4(SH2_struct* sh)
{
  SH2div1(sh,12,0);
}
#define init_3_12_1_4
void FASTCALL call_3_12_1_4(SH2_struct* sh)
{
  SH2div1(sh,12,1);
}
#define init_3_12_2_4
void FASTCALL call_3_12_2_4(SH2_struct* sh)
{
  SH2div1(sh,12,2);
}
#define init_3_12_3_4
void FASTCALL call_3_12_3_4(SH2_struct* sh)
{
  SH2div1(sh,12,3);
}
#define init_3_12_4_4
void FASTCALL call_3_12_4_4(SH2_struct* sh)
{
  SH2div1(sh,12,4);
}
#define init_3_12_5_4
void FASTCALL call_3_12_5_4(SH2_struct* sh)
{
  SH2div1(sh,12,5);
}
#define init_3_12_6_4
void FASTCALL call_3_12_6_4(SH2_struct* sh)
{
  SH2div1(sh,12,6);
}
#define init_3_12_7_4
void FASTCALL call_3_12_7_4(SH2_struct* sh)
{
  SH2div1(sh,12,7);
}
#define init_3_12_8_4
void FASTCALL call_3_12_8_4(SH2_struct* sh)
{
  SH2div1(sh,12,8);
}
#define init_3_12_9_4
void FASTCALL call_3_12_9_4(SH2_struct* sh)
{
  SH2div1(sh,12,9);
}
#define init_3_12_10_4
void FASTCALL call_3_12_10_4(SH2_struct* sh)
{
  SH2div1(sh,12,10);
}
#define init_3_12_11_4
void FASTCALL call_3_12_11_4(SH2_struct* sh)
{
  SH2div1(sh,12,11);
}
#define init_3_12_12_4
void FASTCALL call_3_12_12_4(SH2_struct* sh)
{
  SH2div1(sh,12,12);
}
#define init_3_12_13_4
void FASTCALL call_3_12_13_4(SH2_struct* sh)
{
  SH2div1(sh,12,13);
}
#define init_3_12_14_4
void FASTCALL call_3_12_14_4(SH2_struct* sh)
{
  SH2div1(sh,12,14);
}
#define init_3_12_15_4
void FASTCALL call_3_12_15_4(SH2_struct* sh)
{
  SH2div1(sh,12,15);
}
#define init_3_13_0_4
void FASTCALL call_3_13_0_4(SH2_struct* sh)
{
  SH2div1(sh,13,0);
}
#define init_3_13_1_4
void FASTCALL call_3_13_1_4(SH2_struct* sh)
{
  SH2div1(sh,13,1);
}
#define init_3_13_2_4
void FASTCALL call_3_13_2_4(SH2_struct* sh)
{
  SH2div1(sh,13,2);
}
#define init_3_13_3_4
void FASTCALL call_3_13_3_4(SH2_struct* sh)
{
  SH2div1(sh,13,3);
}
#define init_3_13_4_4
void FASTCALL call_3_13_4_4(SH2_struct* sh)
{
  SH2div1(sh,13,4);
}
#define init_3_13_5_4
void FASTCALL call_3_13_5_4(SH2_struct* sh)
{
  SH2div1(sh,13,5);
}
#define init_3_13_6_4
void FASTCALL call_3_13_6_4(SH2_struct* sh)
{
  SH2div1(sh,13,6);
}
#define init_3_13_7_4
void FASTCALL call_3_13_7_4(SH2_struct* sh)
{
  SH2div1(sh,13,7);
}
#define init_3_13_8_4
void FASTCALL call_3_13_8_4(SH2_struct* sh)
{
  SH2div1(sh,13,8);
}
#define init_3_13_9_4
void FASTCALL call_3_13_9_4(SH2_struct* sh)
{
  SH2div1(sh,13,9);
}
#define init_3_13_10_4
void FASTCALL call_3_13_10_4(SH2_struct* sh)
{
  SH2div1(sh,13,10);
}
#define init_3_13_11_4
void FASTCALL call_3_13_11_4(SH2_struct* sh)
{
  SH2div1(sh,13,11);
}
#define init_3_13_12_4
void FASTCALL call_3_13_12_4(SH2_struct* sh)
{
  SH2div1(sh,13,12);
}
#define init_3_13_13_4
void FASTCALL call_3_13_13_4(SH2_struct* sh)
{
  SH2div1(sh,13,13);
}
#define init_3_13_14_4
void FASTCALL call_3_13_14_4(SH2_struct* sh)
{
  SH2div1(sh,13,14);
}
#define init_3_13_15_4
void FASTCALL call_3_13_15_4(SH2_struct* sh)
{
  SH2div1(sh,13,15);
}
#define init_3_14_0_4
void FASTCALL call_3_14_0_4(SH2_struct* sh)
{
  SH2div1(sh,14,0);
}
#define init_3_14_1_4
void FASTCALL call_3_14_1_4(SH2_struct* sh)
{
  SH2div1(sh,14,1);
}
#define init_3_14_2_4
void FASTCALL call_3_14_2_4(SH2_struct* sh)
{
  SH2div1(sh,14,2);
}
#define init_3_14_3_4
void FASTCALL call_3_14_3_4(SH2_struct* sh)
{
  SH2div1(sh,14,3);
}
#define init_3_14_4_4
void FASTCALL call_3_14_4_4(SH2_struct* sh)
{
  SH2div1(sh,14,4);
}
#define init_3_14_5_4
void FASTCALL call_3_14_5_4(SH2_struct* sh)
{
  SH2div1(sh,14,5);
}
#define init_3_14_6_4
void FASTCALL call_3_14_6_4(SH2_struct* sh)
{
  SH2div1(sh,14,6);
}
#define init_3_14_7_4
void FASTCALL call_3_14_7_4(SH2_struct* sh)
{
  SH2div1(sh,14,7);
}
#define init_3_14_8_4
void FASTCALL call_3_14_8_4(SH2_struct* sh)
{
  SH2div1(sh,14,8);
}
#define init_3_14_9_4
void FASTCALL call_3_14_9_4(SH2_struct* sh)
{
  SH2div1(sh,14,9);
}
#define init_3_14_10_4
void FASTCALL call_3_14_10_4(SH2_struct* sh)
{
  SH2div1(sh,14,10);
}
#define init_3_14_11_4
void FASTCALL call_3_14_11_4(SH2_struct* sh)
{
  SH2div1(sh,14,11);
}
#define init_3_14_12_4
void FASTCALL call_3_14_12_4(SH2_struct* sh)
{
  SH2div1(sh,14,12);
}
#define init_3_14_13_4
void FASTCALL call_3_14_13_4(SH2_struct* sh)
{
  SH2div1(sh,14,13);
}
#define init_3_14_14_4
void FASTCALL call_3_14_14_4(SH2_struct* sh)
{
  SH2div1(sh,14,14);
}
#define init_3_14_15_4
void FASTCALL call_3_14_15_4(SH2_struct* sh)
{
  SH2div1(sh,14,15);
}
#define init_3_15_0_4
void FASTCALL call_3_15_0_4(SH2_struct* sh)
{
  SH2div1(sh,15,0);
}
#define init_3_15_1_4
void FASTCALL call_3_15_1_4(SH2_struct* sh)
{
  SH2div1(sh,15,1);
}
#define init_3_15_2_4
void FASTCALL call_3_15_2_4(SH2_struct* sh)
{
  SH2div1(sh,15,2);
}
#define init_3_15_3_4
void FASTCALL call_3_15_3_4(SH2_struct* sh)
{
  SH2div1(sh,15,3);
}
#define init_3_15_4_4
void FASTCALL call_3_15_4_4(SH2_struct* sh)
{
  SH2div1(sh,15,4);
}
#define init_3_15_5_4
void FASTCALL call_3_15_5_4(SH2_struct* sh)
{
  SH2div1(sh,15,5);
}
#define init_3_15_6_4
void FASTCALL call_3_15_6_4(SH2_struct* sh)
{
  SH2div1(sh,15,6);
}
#define init_3_15_7_4
void FASTCALL call_3_15_7_4(SH2_struct* sh)
{
  SH2div1(sh,15,7);
}
#define init_3_15_8_4
void FASTCALL call_3_15_8_4(SH2_struct* sh)
{
  SH2div1(sh,15,8);
}
#define init_3_15_9_4
void FASTCALL call_3_15_9_4(SH2_struct* sh)
{
  SH2div1(sh,15,9);
}
#define init_3_15_10_4
void FASTCALL call_3_15_10_4(SH2_struct* sh)
{
  SH2div1(sh,15,10);
}
#define init_3_15_11_4
void FASTCALL call_3_15_11_4(SH2_struct* sh)
{
  SH2div1(sh,15,11);
}
#define init_3_15_12_4
void FASTCALL call_3_15_12_4(SH2_struct* sh)
{
  SH2div1(sh,15,12);
}
#define init_3_15_13_4
void FASTCALL call_3_15_13_4(SH2_struct* sh)
{
  SH2div1(sh,15,13);
}
#define init_3_15_14_4
void FASTCALL call_3_15_14_4(SH2_struct* sh)
{
  SH2div1(sh,15,14);
}
#define init_3_15_15_4
void FASTCALL call_3_15_15_4(SH2_struct* sh)
{
  SH2div1(sh,15,15);
}
#define init_2_0_0_7
void FASTCALL call_2_0_0_7(SH2_struct* sh)
{
  SH2div0s(sh,0,0);
}
#define init_2_0_1_7
void FASTCALL call_2_0_1_7(SH2_struct* sh)
{
  SH2div0s(sh,0,1);
}
#define init_2_0_2_7
void FASTCALL call_2_0_2_7(SH2_struct* sh)
{
  SH2div0s(sh,0,2);
}
#define init_2_0_3_7
void FASTCALL call_2_0_3_7(SH2_struct* sh)
{
  SH2div0s(sh,0,3);
}
#define init_2_0_4_7
void FASTCALL call_2_0_4_7(SH2_struct* sh)
{
  SH2div0s(sh,0,4);
}
#define init_2_0_5_7
void FASTCALL call_2_0_5_7(SH2_struct* sh)
{
  SH2div0s(sh,0,5);
}
#define init_2_0_6_7
void FASTCALL call_2_0_6_7(SH2_struct* sh)
{
  SH2div0s(sh,0,6);
}
#define init_2_0_7_7
void FASTCALL call_2_0_7_7(SH2_struct* sh)
{
  SH2div0s(sh,0,7);
}
#define init_2_0_8_7
void FASTCALL call_2_0_8_7(SH2_struct* sh)
{
  SH2div0s(sh,0,8);
}
#define init_2_0_9_7
void FASTCALL call_2_0_9_7(SH2_struct* sh)
{
  SH2div0s(sh,0,9);
}
#define init_2_0_10_7
void FASTCALL call_2_0_10_7(SH2_struct* sh)
{
  SH2div0s(sh,0,10);
}
#define init_2_0_11_7
void FASTCALL call_2_0_11_7(SH2_struct* sh)
{
  SH2div0s(sh,0,11);
}
#define init_2_0_12_7
void FASTCALL call_2_0_12_7(SH2_struct* sh)
{
  SH2div0s(sh,0,12);
}
#define init_2_0_13_7
void FASTCALL call_2_0_13_7(SH2_struct* sh)
{
  SH2div0s(sh,0,13);
}
#define init_2_0_14_7
void FASTCALL call_2_0_14_7(SH2_struct* sh)
{
  SH2div0s(sh,0,14);
}
#define init_2_0_15_7
void FASTCALL call_2_0_15_7(SH2_struct* sh)
{
  SH2div0s(sh,0,15);
}
#define init_2_1_0_7
void FASTCALL call_2_1_0_7(SH2_struct* sh)
{
  SH2div0s(sh,1,0);
}
#define init_2_1_1_7
void FASTCALL call_2_1_1_7(SH2_struct* sh)
{
  SH2div0s(sh,1,1);
}
#define init_2_1_2_7
void FASTCALL call_2_1_2_7(SH2_struct* sh)
{
  SH2div0s(sh,1,2);
}
#define init_2_1_3_7
void FASTCALL call_2_1_3_7(SH2_struct* sh)
{
  SH2div0s(sh,1,3);
}
#define init_2_1_4_7
void FASTCALL call_2_1_4_7(SH2_struct* sh)
{
  SH2div0s(sh,1,4);
}
#define init_2_1_5_7
void FASTCALL call_2_1_5_7(SH2_struct* sh)
{
  SH2div0s(sh,1,5);
}
#define init_2_1_6_7
void FASTCALL call_2_1_6_7(SH2_struct* sh)
{
  SH2div0s(sh,1,6);
}
#define init_2_1_7_7
void FASTCALL call_2_1_7_7(SH2_struct* sh)
{
  SH2div0s(sh,1,7);
}
#define init_2_1_8_7
void FASTCALL call_2_1_8_7(SH2_struct* sh)
{
  SH2div0s(sh,1,8);
}
#define init_2_1_9_7
void FASTCALL call_2_1_9_7(SH2_struct* sh)
{
  SH2div0s(sh,1,9);
}
#define init_2_1_10_7
void FASTCALL call_2_1_10_7(SH2_struct* sh)
{
  SH2div0s(sh,1,10);
}
#define init_2_1_11_7
void FASTCALL call_2_1_11_7(SH2_struct* sh)
{
  SH2div0s(sh,1,11);
}
#define init_2_1_12_7
void FASTCALL call_2_1_12_7(SH2_struct* sh)
{
  SH2div0s(sh,1,12);
}
#define init_2_1_13_7
void FASTCALL call_2_1_13_7(SH2_struct* sh)
{
  SH2div0s(sh,1,13);
}
#define init_2_1_14_7
void FASTCALL call_2_1_14_7(SH2_struct* sh)
{
  SH2div0s(sh,1,14);
}
#define init_2_1_15_7
void FASTCALL call_2_1_15_7(SH2_struct* sh)
{
  SH2div0s(sh,1,15);
}
#define init_2_2_0_7
void FASTCALL call_2_2_0_7(SH2_struct* sh)
{
  SH2div0s(sh,2,0);
}
#define init_2_2_1_7
void FASTCALL call_2_2_1_7(SH2_struct* sh)
{
  SH2div0s(sh,2,1);
}
#define init_2_2_2_7
void FASTCALL call_2_2_2_7(SH2_struct* sh)
{
  SH2div0s(sh,2,2);
}
#define init_2_2_3_7
void FASTCALL call_2_2_3_7(SH2_struct* sh)
{
  SH2div0s(sh,2,3);
}
#define init_2_2_4_7
void FASTCALL call_2_2_4_7(SH2_struct* sh)
{
  SH2div0s(sh,2,4);
}
#define init_2_2_5_7
void FASTCALL call_2_2_5_7(SH2_struct* sh)
{
  SH2div0s(sh,2,5);
}
#define init_2_2_6_7
void FASTCALL call_2_2_6_7(SH2_struct* sh)
{
  SH2div0s(sh,2,6);
}
#define init_2_2_7_7
void FASTCALL call_2_2_7_7(SH2_struct* sh)
{
  SH2div0s(sh,2,7);
}
#define init_2_2_8_7
void FASTCALL call_2_2_8_7(SH2_struct* sh)
{
  SH2div0s(sh,2,8);
}
#define init_2_2_9_7
void FASTCALL call_2_2_9_7(SH2_struct* sh)
{
  SH2div0s(sh,2,9);
}
#define init_2_2_10_7
void FASTCALL call_2_2_10_7(SH2_struct* sh)
{
  SH2div0s(sh,2,10);
}
#define init_2_2_11_7
void FASTCALL call_2_2_11_7(SH2_struct* sh)
{
  SH2div0s(sh,2,11);
}
#define init_2_2_12_7
void FASTCALL call_2_2_12_7(SH2_struct* sh)
{
  SH2div0s(sh,2,12);
}
#define init_2_2_13_7
void FASTCALL call_2_2_13_7(SH2_struct* sh)
{
  SH2div0s(sh,2,13);
}
#define init_2_2_14_7
void FASTCALL call_2_2_14_7(SH2_struct* sh)
{
  SH2div0s(sh,2,14);
}
#define init_2_2_15_7
void FASTCALL call_2_2_15_7(SH2_struct* sh)
{
  SH2div0s(sh,2,15);
}
#define init_2_3_0_7
void FASTCALL call_2_3_0_7(SH2_struct* sh)
{
  SH2div0s(sh,3,0);
}
#define init_2_3_1_7
void FASTCALL call_2_3_1_7(SH2_struct* sh)
{
  SH2div0s(sh,3,1);
}
#define init_2_3_2_7
void FASTCALL call_2_3_2_7(SH2_struct* sh)
{
  SH2div0s(sh,3,2);
}
#define init_2_3_3_7
void FASTCALL call_2_3_3_7(SH2_struct* sh)
{
  SH2div0s(sh,3,3);
}
#define init_2_3_4_7
void FASTCALL call_2_3_4_7(SH2_struct* sh)
{
  SH2div0s(sh,3,4);
}
#define init_2_3_5_7
void FASTCALL call_2_3_5_7(SH2_struct* sh)
{
  SH2div0s(sh,3,5);
}
#define init_2_3_6_7
void FASTCALL call_2_3_6_7(SH2_struct* sh)
{
  SH2div0s(sh,3,6);
}
#define init_2_3_7_7
void FASTCALL call_2_3_7_7(SH2_struct* sh)
{
  SH2div0s(sh,3,7);
}
#define init_2_3_8_7
void FASTCALL call_2_3_8_7(SH2_struct* sh)
{
  SH2div0s(sh,3,8);
}
#define init_2_3_9_7
void FASTCALL call_2_3_9_7(SH2_struct* sh)
{
  SH2div0s(sh,3,9);
}
#define init_2_3_10_7
void FASTCALL call_2_3_10_7(SH2_struct* sh)
{
  SH2div0s(sh,3,10);
}
#define init_2_3_11_7
void FASTCALL call_2_3_11_7(SH2_struct* sh)
{
  SH2div0s(sh,3,11);
}
#define init_2_3_12_7
void FASTCALL call_2_3_12_7(SH2_struct* sh)
{
  SH2div0s(sh,3,12);
}
#define init_2_3_13_7
void FASTCALL call_2_3_13_7(SH2_struct* sh)
{
  SH2div0s(sh,3,13);
}
#define init_2_3_14_7
void FASTCALL call_2_3_14_7(SH2_struct* sh)
{
  SH2div0s(sh,3,14);
}
#define init_2_3_15_7
void FASTCALL call_2_3_15_7(SH2_struct* sh)
{
  SH2div0s(sh,3,15);
}
#define init_2_4_0_7
void FASTCALL call_2_4_0_7(SH2_struct* sh)
{
  SH2div0s(sh,4,0);
}
#define init_2_4_1_7
void FASTCALL call_2_4_1_7(SH2_struct* sh)
{
  SH2div0s(sh,4,1);
}
#define init_2_4_2_7
void FASTCALL call_2_4_2_7(SH2_struct* sh)
{
  SH2div0s(sh,4,2);
}
#define init_2_4_3_7
void FASTCALL call_2_4_3_7(SH2_struct* sh)
{
  SH2div0s(sh,4,3);
}
#define init_2_4_4_7
void FASTCALL call_2_4_4_7(SH2_struct* sh)
{
  SH2div0s(sh,4,4);
}
#define init_2_4_5_7
void FASTCALL call_2_4_5_7(SH2_struct* sh)
{
  SH2div0s(sh,4,5);
}
#define init_2_4_6_7
void FASTCALL call_2_4_6_7(SH2_struct* sh)
{
  SH2div0s(sh,4,6);
}
#define init_2_4_7_7
void FASTCALL call_2_4_7_7(SH2_struct* sh)
{
  SH2div0s(sh,4,7);
}
#define init_2_4_8_7
void FASTCALL call_2_4_8_7(SH2_struct* sh)
{
  SH2div0s(sh,4,8);
}
#define init_2_4_9_7
void FASTCALL call_2_4_9_7(SH2_struct* sh)
{
  SH2div0s(sh,4,9);
}
#define init_2_4_10_7
void FASTCALL call_2_4_10_7(SH2_struct* sh)
{
  SH2div0s(sh,4,10);
}
#define init_2_4_11_7
void FASTCALL call_2_4_11_7(SH2_struct* sh)
{
  SH2div0s(sh,4,11);
}
#define init_2_4_12_7
void FASTCALL call_2_4_12_7(SH2_struct* sh)
{
  SH2div0s(sh,4,12);
}
#define init_2_4_13_7
void FASTCALL call_2_4_13_7(SH2_struct* sh)
{
  SH2div0s(sh,4,13);
}
#define init_2_4_14_7
void FASTCALL call_2_4_14_7(SH2_struct* sh)
{
  SH2div0s(sh,4,14);
}
#define init_2_4_15_7
void FASTCALL call_2_4_15_7(SH2_struct* sh)
{
  SH2div0s(sh,4,15);
}
#define init_2_5_0_7
void FASTCALL call_2_5_0_7(SH2_struct* sh)
{
  SH2div0s(sh,5,0);
}
#define init_2_5_1_7
void FASTCALL call_2_5_1_7(SH2_struct* sh)
{
  SH2div0s(sh,5,1);
}
#define init_2_5_2_7
void FASTCALL call_2_5_2_7(SH2_struct* sh)
{
  SH2div0s(sh,5,2);
}
#define init_2_5_3_7
void FASTCALL call_2_5_3_7(SH2_struct* sh)
{
  SH2div0s(sh,5,3);
}
#define init_2_5_4_7
void FASTCALL call_2_5_4_7(SH2_struct* sh)
{
  SH2div0s(sh,5,4);
}
#define init_2_5_5_7
void FASTCALL call_2_5_5_7(SH2_struct* sh)
{
  SH2div0s(sh,5,5);
}
#define init_2_5_6_7
void FASTCALL call_2_5_6_7(SH2_struct* sh)
{
  SH2div0s(sh,5,6);
}
#define init_2_5_7_7
void FASTCALL call_2_5_7_7(SH2_struct* sh)
{
  SH2div0s(sh,5,7);
}
#define init_2_5_8_7
void FASTCALL call_2_5_8_7(SH2_struct* sh)
{
  SH2div0s(sh,5,8);
}
#define init_2_5_9_7
void FASTCALL call_2_5_9_7(SH2_struct* sh)
{
  SH2div0s(sh,5,9);
}
#define init_2_5_10_7
void FASTCALL call_2_5_10_7(SH2_struct* sh)
{
  SH2div0s(sh,5,10);
}
#define init_2_5_11_7
void FASTCALL call_2_5_11_7(SH2_struct* sh)
{
  SH2div0s(sh,5,11);
}
#define init_2_5_12_7
void FASTCALL call_2_5_12_7(SH2_struct* sh)
{
  SH2div0s(sh,5,12);
}
#define init_2_5_13_7
void FASTCALL call_2_5_13_7(SH2_struct* sh)
{
  SH2div0s(sh,5,13);
}
#define init_2_5_14_7
void FASTCALL call_2_5_14_7(SH2_struct* sh)
{
  SH2div0s(sh,5,14);
}
#define init_2_5_15_7
void FASTCALL call_2_5_15_7(SH2_struct* sh)
{
  SH2div0s(sh,5,15);
}
#define init_2_6_0_7
void FASTCALL call_2_6_0_7(SH2_struct* sh)
{
  SH2div0s(sh,6,0);
}
#define init_2_6_1_7
void FASTCALL call_2_6_1_7(SH2_struct* sh)
{
  SH2div0s(sh,6,1);
}
#define init_2_6_2_7
void FASTCALL call_2_6_2_7(SH2_struct* sh)
{
  SH2div0s(sh,6,2);
}
#define init_2_6_3_7
void FASTCALL call_2_6_3_7(SH2_struct* sh)
{
  SH2div0s(sh,6,3);
}
#define init_2_6_4_7
void FASTCALL call_2_6_4_7(SH2_struct* sh)
{
  SH2div0s(sh,6,4);
}
#define init_2_6_5_7
void FASTCALL call_2_6_5_7(SH2_struct* sh)
{
  SH2div0s(sh,6,5);
}
#define init_2_6_6_7
void FASTCALL call_2_6_6_7(SH2_struct* sh)
{
  SH2div0s(sh,6,6);
}
#define init_2_6_7_7
void FASTCALL call_2_6_7_7(SH2_struct* sh)
{
  SH2div0s(sh,6,7);
}
#define init_2_6_8_7
void FASTCALL call_2_6_8_7(SH2_struct* sh)
{
  SH2div0s(sh,6,8);
}
#define init_2_6_9_7
void FASTCALL call_2_6_9_7(SH2_struct* sh)
{
  SH2div0s(sh,6,9);
}
#define init_2_6_10_7
void FASTCALL call_2_6_10_7(SH2_struct* sh)
{
  SH2div0s(sh,6,10);
}
#define init_2_6_11_7
void FASTCALL call_2_6_11_7(SH2_struct* sh)
{
  SH2div0s(sh,6,11);
}
#define init_2_6_12_7
void FASTCALL call_2_6_12_7(SH2_struct* sh)
{
  SH2div0s(sh,6,12);
}
#define init_2_6_13_7
void FASTCALL call_2_6_13_7(SH2_struct* sh)
{
  SH2div0s(sh,6,13);
}
#define init_2_6_14_7
void FASTCALL call_2_6_14_7(SH2_struct* sh)
{
  SH2div0s(sh,6,14);
}
#define init_2_6_15_7
void FASTCALL call_2_6_15_7(SH2_struct* sh)
{
  SH2div0s(sh,6,15);
}
#define init_2_7_0_7
void FASTCALL call_2_7_0_7(SH2_struct* sh)
{
  SH2div0s(sh,7,0);
}
#define init_2_7_1_7
void FASTCALL call_2_7_1_7(SH2_struct* sh)
{
  SH2div0s(sh,7,1);
}
#define init_2_7_2_7
void FASTCALL call_2_7_2_7(SH2_struct* sh)
{
  SH2div0s(sh,7,2);
}
#define init_2_7_3_7
void FASTCALL call_2_7_3_7(SH2_struct* sh)
{
  SH2div0s(sh,7,3);
}
#define init_2_7_4_7
void FASTCALL call_2_7_4_7(SH2_struct* sh)
{
  SH2div0s(sh,7,4);
}
#define init_2_7_5_7
void FASTCALL call_2_7_5_7(SH2_struct* sh)
{
  SH2div0s(sh,7,5);
}
#define init_2_7_6_7
void FASTCALL call_2_7_6_7(SH2_struct* sh)
{
  SH2div0s(sh,7,6);
}
#define init_2_7_7_7
void FASTCALL call_2_7_7_7(SH2_struct* sh)
{
  SH2div0s(sh,7,7);
}
#define init_2_7_8_7
void FASTCALL call_2_7_8_7(SH2_struct* sh)
{
  SH2div0s(sh,7,8);
}
#define init_2_7_9_7
void FASTCALL call_2_7_9_7(SH2_struct* sh)
{
  SH2div0s(sh,7,9);
}
#define init_2_7_10_7
void FASTCALL call_2_7_10_7(SH2_struct* sh)
{
  SH2div0s(sh,7,10);
}
#define init_2_7_11_7
void FASTCALL call_2_7_11_7(SH2_struct* sh)
{
  SH2div0s(sh,7,11);
}
#define init_2_7_12_7
void FASTCALL call_2_7_12_7(SH2_struct* sh)
{
  SH2div0s(sh,7,12);
}
#define init_2_7_13_7
void FASTCALL call_2_7_13_7(SH2_struct* sh)
{
  SH2div0s(sh,7,13);
}
#define init_2_7_14_7
void FASTCALL call_2_7_14_7(SH2_struct* sh)
{
  SH2div0s(sh,7,14);
}
#define init_2_7_15_7
void FASTCALL call_2_7_15_7(SH2_struct* sh)
{
  SH2div0s(sh,7,15);
}
#define init_2_8_0_7
void FASTCALL call_2_8_0_7(SH2_struct* sh)
{
  SH2div0s(sh,8,0);
}
#define init_2_8_1_7
void FASTCALL call_2_8_1_7(SH2_struct* sh)
{
  SH2div0s(sh,8,1);
}
#define init_2_8_2_7
void FASTCALL call_2_8_2_7(SH2_struct* sh)
{
  SH2div0s(sh,8,2);
}
#define init_2_8_3_7
void FASTCALL call_2_8_3_7(SH2_struct* sh)
{
  SH2div0s(sh,8,3);
}
#define init_2_8_4_7
void FASTCALL call_2_8_4_7(SH2_struct* sh)
{
  SH2div0s(sh,8,4);
}
#define init_2_8_5_7
void FASTCALL call_2_8_5_7(SH2_struct* sh)
{
  SH2div0s(sh,8,5);
}
#define init_2_8_6_7
void FASTCALL call_2_8_6_7(SH2_struct* sh)
{
  SH2div0s(sh,8,6);
}
#define init_2_8_7_7
void FASTCALL call_2_8_7_7(SH2_struct* sh)
{
  SH2div0s(sh,8,7);
}
#define init_2_8_8_7
void FASTCALL call_2_8_8_7(SH2_struct* sh)
{
  SH2div0s(sh,8,8);
}
#define init_2_8_9_7
void FASTCALL call_2_8_9_7(SH2_struct* sh)
{
  SH2div0s(sh,8,9);
}
#define init_2_8_10_7
void FASTCALL call_2_8_10_7(SH2_struct* sh)
{
  SH2div0s(sh,8,10);
}
#define init_2_8_11_7
void FASTCALL call_2_8_11_7(SH2_struct* sh)
{
  SH2div0s(sh,8,11);
}
#define init_2_8_12_7
void FASTCALL call_2_8_12_7(SH2_struct* sh)
{
  SH2div0s(sh,8,12);
}
#define init_2_8_13_7
void FASTCALL call_2_8_13_7(SH2_struct* sh)
{
  SH2div0s(sh,8,13);
}
#define init_2_8_14_7
void FASTCALL call_2_8_14_7(SH2_struct* sh)
{
  SH2div0s(sh,8,14);
}
#define init_2_8_15_7
void FASTCALL call_2_8_15_7(SH2_struct* sh)
{
  SH2div0s(sh,8,15);
}
#define init_2_9_0_7
void FASTCALL call_2_9_0_7(SH2_struct* sh)
{
  SH2div0s(sh,9,0);
}
#define init_2_9_1_7
void FASTCALL call_2_9_1_7(SH2_struct* sh)
{
  SH2div0s(sh,9,1);
}
#define init_2_9_2_7
void FASTCALL call_2_9_2_7(SH2_struct* sh)
{
  SH2div0s(sh,9,2);
}
#define init_2_9_3_7
void FASTCALL call_2_9_3_7(SH2_struct* sh)
{
  SH2div0s(sh,9,3);
}
#define init_2_9_4_7
void FASTCALL call_2_9_4_7(SH2_struct* sh)
{
  SH2div0s(sh,9,4);
}
#define init_2_9_5_7
void FASTCALL call_2_9_5_7(SH2_struct* sh)
{
  SH2div0s(sh,9,5);
}
#define init_2_9_6_7
void FASTCALL call_2_9_6_7(SH2_struct* sh)
{
  SH2div0s(sh,9,6);
}
#define init_2_9_7_7
void FASTCALL call_2_9_7_7(SH2_struct* sh)
{
  SH2div0s(sh,9,7);
}
#define init_2_9_8_7
void FASTCALL call_2_9_8_7(SH2_struct* sh)
{
  SH2div0s(sh,9,8);
}
#define init_2_9_9_7
void FASTCALL call_2_9_9_7(SH2_struct* sh)
{
  SH2div0s(sh,9,9);
}
#define init_2_9_10_7
void FASTCALL call_2_9_10_7(SH2_struct* sh)
{
  SH2div0s(sh,9,10);
}
#define init_2_9_11_7
void FASTCALL call_2_9_11_7(SH2_struct* sh)
{
  SH2div0s(sh,9,11);
}
#define init_2_9_12_7
void FASTCALL call_2_9_12_7(SH2_struct* sh)
{
  SH2div0s(sh,9,12);
}
#define init_2_9_13_7
void FASTCALL call_2_9_13_7(SH2_struct* sh)
{
  SH2div0s(sh,9,13);
}
#define init_2_9_14_7
void FASTCALL call_2_9_14_7(SH2_struct* sh)
{
  SH2div0s(sh,9,14);
}
#define init_2_9_15_7
void FASTCALL call_2_9_15_7(SH2_struct* sh)
{
  SH2div0s(sh,9,15);
}
#define init_2_10_0_7
void FASTCALL call_2_10_0_7(SH2_struct* sh)
{
  SH2div0s(sh,10,0);
}
#define init_2_10_1_7
void FASTCALL call_2_10_1_7(SH2_struct* sh)
{
  SH2div0s(sh,10,1);
}
#define init_2_10_2_7
void FASTCALL call_2_10_2_7(SH2_struct* sh)
{
  SH2div0s(sh,10,2);
}
#define init_2_10_3_7
void FASTCALL call_2_10_3_7(SH2_struct* sh)
{
  SH2div0s(sh,10,3);
}
#define init_2_10_4_7
void FASTCALL call_2_10_4_7(SH2_struct* sh)
{
  SH2div0s(sh,10,4);
}
#define init_2_10_5_7
void FASTCALL call_2_10_5_7(SH2_struct* sh)
{
  SH2div0s(sh,10,5);
}
#define init_2_10_6_7
void FASTCALL call_2_10_6_7(SH2_struct* sh)
{
  SH2div0s(sh,10,6);
}
#define init_2_10_7_7
void FASTCALL call_2_10_7_7(SH2_struct* sh)
{
  SH2div0s(sh,10,7);
}
#define init_2_10_8_7
void FASTCALL call_2_10_8_7(SH2_struct* sh)
{
  SH2div0s(sh,10,8);
}
#define init_2_10_9_7
void FASTCALL call_2_10_9_7(SH2_struct* sh)
{
  SH2div0s(sh,10,9);
}
#define init_2_10_10_7
void FASTCALL call_2_10_10_7(SH2_struct* sh)
{
  SH2div0s(sh,10,10);
}
#define init_2_10_11_7
void FASTCALL call_2_10_11_7(SH2_struct* sh)
{
  SH2div0s(sh,10,11);
}
#define init_2_10_12_7
void FASTCALL call_2_10_12_7(SH2_struct* sh)
{
  SH2div0s(sh,10,12);
}
#define init_2_10_13_7
void FASTCALL call_2_10_13_7(SH2_struct* sh)
{
  SH2div0s(sh,10,13);
}
#define init_2_10_14_7
void FASTCALL call_2_10_14_7(SH2_struct* sh)
{
  SH2div0s(sh,10,14);
}
#define init_2_10_15_7
void FASTCALL call_2_10_15_7(SH2_struct* sh)
{
  SH2div0s(sh,10,15);
}
#define init_2_11_0_7
void FASTCALL call_2_11_0_7(SH2_struct* sh)
{
  SH2div0s(sh,11,0);
}
#define init_2_11_1_7
void FASTCALL call_2_11_1_7(SH2_struct* sh)
{
  SH2div0s(sh,11,1);
}
#define init_2_11_2_7
void FASTCALL call_2_11_2_7(SH2_struct* sh)
{
  SH2div0s(sh,11,2);
}
#define init_2_11_3_7
void FASTCALL call_2_11_3_7(SH2_struct* sh)
{
  SH2div0s(sh,11,3);
}
#define init_2_11_4_7
void FASTCALL call_2_11_4_7(SH2_struct* sh)
{
  SH2div0s(sh,11,4);
}
#define init_2_11_5_7
void FASTCALL call_2_11_5_7(SH2_struct* sh)
{
  SH2div0s(sh,11,5);
}
#define init_2_11_6_7
void FASTCALL call_2_11_6_7(SH2_struct* sh)
{
  SH2div0s(sh,11,6);
}
#define init_2_11_7_7
void FASTCALL call_2_11_7_7(SH2_struct* sh)
{
  SH2div0s(sh,11,7);
}
#define init_2_11_8_7
void FASTCALL call_2_11_8_7(SH2_struct* sh)
{
  SH2div0s(sh,11,8);
}
#define init_2_11_9_7
void FASTCALL call_2_11_9_7(SH2_struct* sh)
{
  SH2div0s(sh,11,9);
}
#define init_2_11_10_7
void FASTCALL call_2_11_10_7(SH2_struct* sh)
{
  SH2div0s(sh,11,10);
}
#define init_2_11_11_7
void FASTCALL call_2_11_11_7(SH2_struct* sh)
{
  SH2div0s(sh,11,11);
}
#define init_2_11_12_7
void FASTCALL call_2_11_12_7(SH2_struct* sh)
{
  SH2div0s(sh,11,12);
}
#define init_2_11_13_7
void FASTCALL call_2_11_13_7(SH2_struct* sh)
{
  SH2div0s(sh,11,13);
}
#define init_2_11_14_7
void FASTCALL call_2_11_14_7(SH2_struct* sh)
{
  SH2div0s(sh,11,14);
}
#define init_2_11_15_7
void FASTCALL call_2_11_15_7(SH2_struct* sh)
{
  SH2div0s(sh,11,15);
}
#define init_2_12_0_7
void FASTCALL call_2_12_0_7(SH2_struct* sh)
{
  SH2div0s(sh,12,0);
}
#define init_2_12_1_7
void FASTCALL call_2_12_1_7(SH2_struct* sh)
{
  SH2div0s(sh,12,1);
}
#define init_2_12_2_7
void FASTCALL call_2_12_2_7(SH2_struct* sh)
{
  SH2div0s(sh,12,2);
}
#define init_2_12_3_7
void FASTCALL call_2_12_3_7(SH2_struct* sh)
{
  SH2div0s(sh,12,3);
}
#define init_2_12_4_7
void FASTCALL call_2_12_4_7(SH2_struct* sh)
{
  SH2div0s(sh,12,4);
}
#define init_2_12_5_7
void FASTCALL call_2_12_5_7(SH2_struct* sh)
{
  SH2div0s(sh,12,5);
}
#define init_2_12_6_7
void FASTCALL call_2_12_6_7(SH2_struct* sh)
{
  SH2div0s(sh,12,6);
}
#define init_2_12_7_7
void FASTCALL call_2_12_7_7(SH2_struct* sh)
{
  SH2div0s(sh,12,7);
}
#define init_2_12_8_7
void FASTCALL call_2_12_8_7(SH2_struct* sh)
{
  SH2div0s(sh,12,8);
}
#define init_2_12_9_7
void FASTCALL call_2_12_9_7(SH2_struct* sh)
{
  SH2div0s(sh,12,9);
}
#define init_2_12_10_7
void FASTCALL call_2_12_10_7(SH2_struct* sh)
{
  SH2div0s(sh,12,10);
}
#define init_2_12_11_7
void FASTCALL call_2_12_11_7(SH2_struct* sh)
{
  SH2div0s(sh,12,11);
}
#define init_2_12_12_7
void FASTCALL call_2_12_12_7(SH2_struct* sh)
{
  SH2div0s(sh,12,12);
}
#define init_2_12_13_7
void FASTCALL call_2_12_13_7(SH2_struct* sh)
{
  SH2div0s(sh,12,13);
}
#define init_2_12_14_7
void FASTCALL call_2_12_14_7(SH2_struct* sh)
{
  SH2div0s(sh,12,14);
}
#define init_2_12_15_7
void FASTCALL call_2_12_15_7(SH2_struct* sh)
{
  SH2div0s(sh,12,15);
}
#define init_2_13_0_7
void FASTCALL call_2_13_0_7(SH2_struct* sh)
{
  SH2div0s(sh,13,0);
}
#define init_2_13_1_7
void FASTCALL call_2_13_1_7(SH2_struct* sh)
{
  SH2div0s(sh,13,1);
}
#define init_2_13_2_7
void FASTCALL call_2_13_2_7(SH2_struct* sh)
{
  SH2div0s(sh,13,2);
}
#define init_2_13_3_7
void FASTCALL call_2_13_3_7(SH2_struct* sh)
{
  SH2div0s(sh,13,3);
}
#define init_2_13_4_7
void FASTCALL call_2_13_4_7(SH2_struct* sh)
{
  SH2div0s(sh,13,4);
}
#define init_2_13_5_7
void FASTCALL call_2_13_5_7(SH2_struct* sh)
{
  SH2div0s(sh,13,5);
}
#define init_2_13_6_7
void FASTCALL call_2_13_6_7(SH2_struct* sh)
{
  SH2div0s(sh,13,6);
}
#define init_2_13_7_7
void FASTCALL call_2_13_7_7(SH2_struct* sh)
{
  SH2div0s(sh,13,7);
}
#define init_2_13_8_7
void FASTCALL call_2_13_8_7(SH2_struct* sh)
{
  SH2div0s(sh,13,8);
}
#define init_2_13_9_7
void FASTCALL call_2_13_9_7(SH2_struct* sh)
{
  SH2div0s(sh,13,9);
}
#define init_2_13_10_7
void FASTCALL call_2_13_10_7(SH2_struct* sh)
{
  SH2div0s(sh,13,10);
}
#define init_2_13_11_7
void FASTCALL call_2_13_11_7(SH2_struct* sh)
{
  SH2div0s(sh,13,11);
}
#define init_2_13_12_7
void FASTCALL call_2_13_12_7(SH2_struct* sh)
{
  SH2div0s(sh,13,12);
}
#define init_2_13_13_7
void FASTCALL call_2_13_13_7(SH2_struct* sh)
{
  SH2div0s(sh,13,13);
}
#define init_2_13_14_7
void FASTCALL call_2_13_14_7(SH2_struct* sh)
{
  SH2div0s(sh,13,14);
}
#define init_2_13_15_7
void FASTCALL call_2_13_15_7(SH2_struct* sh)
{
  SH2div0s(sh,13,15);
}
#define init_2_14_0_7
void FASTCALL call_2_14_0_7(SH2_struct* sh)
{
  SH2div0s(sh,14,0);
}
#define init_2_14_1_7
void FASTCALL call_2_14_1_7(SH2_struct* sh)
{
  SH2div0s(sh,14,1);
}
#define init_2_14_2_7
void FASTCALL call_2_14_2_7(SH2_struct* sh)
{
  SH2div0s(sh,14,2);
}
#define init_2_14_3_7
void FASTCALL call_2_14_3_7(SH2_struct* sh)
{
  SH2div0s(sh,14,3);
}
#define init_2_14_4_7
void FASTCALL call_2_14_4_7(SH2_struct* sh)
{
  SH2div0s(sh,14,4);
}
#define init_2_14_5_7
void FASTCALL call_2_14_5_7(SH2_struct* sh)
{
  SH2div0s(sh,14,5);
}
#define init_2_14_6_7
void FASTCALL call_2_14_6_7(SH2_struct* sh)
{
  SH2div0s(sh,14,6);
}
#define init_2_14_7_7
void FASTCALL call_2_14_7_7(SH2_struct* sh)
{
  SH2div0s(sh,14,7);
}
#define init_2_14_8_7
void FASTCALL call_2_14_8_7(SH2_struct* sh)
{
  SH2div0s(sh,14,8);
}
#define init_2_14_9_7
void FASTCALL call_2_14_9_7(SH2_struct* sh)
{
  SH2div0s(sh,14,9);
}
#define init_2_14_10_7
void FASTCALL call_2_14_10_7(SH2_struct* sh)
{
  SH2div0s(sh,14,10);
}
#define init_2_14_11_7
void FASTCALL call_2_14_11_7(SH2_struct* sh)
{
  SH2div0s(sh,14,11);
}
#define init_2_14_12_7
void FASTCALL call_2_14_12_7(SH2_struct* sh)
{
  SH2div0s(sh,14,12);
}
#define init_2_14_13_7
void FASTCALL call_2_14_13_7(SH2_struct* sh)
{
  SH2div0s(sh,14,13);
}
#define init_2_14_14_7
void FASTCALL call_2_14_14_7(SH2_struct* sh)
{
  SH2div0s(sh,14,14);
}
#define init_2_14_15_7
void FASTCALL call_2_14_15_7(SH2_struct* sh)
{
  SH2div0s(sh,14,15);
}
#define init_2_15_0_7
void FASTCALL call_2_15_0_7(SH2_struct* sh)
{
  SH2div0s(sh,15,0);
}
#define init_2_15_1_7
void FASTCALL call_2_15_1_7(SH2_struct* sh)
{
  SH2div0s(sh,15,1);
}
#define init_2_15_2_7
void FASTCALL call_2_15_2_7(SH2_struct* sh)
{
  SH2div0s(sh,15,2);
}
#define init_2_15_3_7
void FASTCALL call_2_15_3_7(SH2_struct* sh)
{
  SH2div0s(sh,15,3);
}
#define init_2_15_4_7
void FASTCALL call_2_15_4_7(SH2_struct* sh)
{
  SH2div0s(sh,15,4);
}
#define init_2_15_5_7
void FASTCALL call_2_15_5_7(SH2_struct* sh)
{
  SH2div0s(sh,15,5);
}
#define init_2_15_6_7
void FASTCALL call_2_15_6_7(SH2_struct* sh)
{
  SH2div0s(sh,15,6);
}
#define init_2_15_7_7
void FASTCALL call_2_15_7_7(SH2_struct* sh)
{
  SH2div0s(sh,15,7);
}
#define init_2_15_8_7
void FASTCALL call_2_15_8_7(SH2_struct* sh)
{
  SH2div0s(sh,15,8);
}
#define init_2_15_9_7
void FASTCALL call_2_15_9_7(SH2_struct* sh)
{
  SH2div0s(sh,15,9);
}
#define init_2_15_10_7
void FASTCALL call_2_15_10_7(SH2_struct* sh)
{
  SH2div0s(sh,15,10);
}
#define init_2_15_11_7
void FASTCALL call_2_15_11_7(SH2_struct* sh)
{
  SH2div0s(sh,15,11);
}
#define init_2_15_12_7
void FASTCALL call_2_15_12_7(SH2_struct* sh)
{
  SH2div0s(sh,15,12);
}
#define init_2_15_13_7
void FASTCALL call_2_15_13_7(SH2_struct* sh)
{
  SH2div0s(sh,15,13);
}
#define init_2_15_14_7
void FASTCALL call_2_15_14_7(SH2_struct* sh)
{
  SH2div0s(sh,15,14);
}
#define init_2_15_15_7
void FASTCALL call_2_15_15_7(SH2_struct* sh)
{
  SH2div0s(sh,15,15);
}
#define init_2_0_0_12
void FASTCALL call_2_0_0_12(SH2_struct* sh)
{
  SH2cmpstr(sh,0,0);
}
#define init_2_0_1_12
void FASTCALL call_2_0_1_12(SH2_struct* sh)
{
  SH2cmpstr(sh,0,1);
}
#define init_2_0_2_12
void FASTCALL call_2_0_2_12(SH2_struct* sh)
{
  SH2cmpstr(sh,0,2);
}
#define init_2_0_3_12
void FASTCALL call_2_0_3_12(SH2_struct* sh)
{
  SH2cmpstr(sh,0,3);
}
#define init_2_0_4_12
void FASTCALL call_2_0_4_12(SH2_struct* sh)
{
  SH2cmpstr(sh,0,4);
}
#define init_2_0_5_12
void FASTCALL call_2_0_5_12(SH2_struct* sh)
{
  SH2cmpstr(sh,0,5);
}
#define init_2_0_6_12
void FASTCALL call_2_0_6_12(SH2_struct* sh)
{
  SH2cmpstr(sh,0,6);
}
#define init_2_0_7_12
void FASTCALL call_2_0_7_12(SH2_struct* sh)
{
  SH2cmpstr(sh,0,7);
}
#define init_2_0_8_12
void FASTCALL call_2_0_8_12(SH2_struct* sh)
{
  SH2cmpstr(sh,0,8);
}
#define init_2_0_9_12
void FASTCALL call_2_0_9_12(SH2_struct* sh)
{
  SH2cmpstr(sh,0,9);
}
#define init_2_0_10_12
void FASTCALL call_2_0_10_12(SH2_struct* sh)
{
  SH2cmpstr(sh,0,10);
}
#define init_2_0_11_12
void FASTCALL call_2_0_11_12(SH2_struct* sh)
{
  SH2cmpstr(sh,0,11);
}
#define init_2_0_12_12
void FASTCALL call_2_0_12_12(SH2_struct* sh)
{
  SH2cmpstr(sh,0,12);
}
#define init_2_0_13_12
void FASTCALL call_2_0_13_12(SH2_struct* sh)
{
  SH2cmpstr(sh,0,13);
}
#define init_2_0_14_12
void FASTCALL call_2_0_14_12(SH2_struct* sh)
{
  SH2cmpstr(sh,0,14);
}
#define init_2_0_15_12
void FASTCALL call_2_0_15_12(SH2_struct* sh)
{
  SH2cmpstr(sh,0,15);
}
#define init_2_1_0_12
void FASTCALL call_2_1_0_12(SH2_struct* sh)
{
  SH2cmpstr(sh,1,0);
}
#define init_2_1_1_12
void FASTCALL call_2_1_1_12(SH2_struct* sh)
{
  SH2cmpstr(sh,1,1);
}
#define init_2_1_2_12
void FASTCALL call_2_1_2_12(SH2_struct* sh)
{
  SH2cmpstr(sh,1,2);
}
#define init_2_1_3_12
void FASTCALL call_2_1_3_12(SH2_struct* sh)
{
  SH2cmpstr(sh,1,3);
}
#define init_2_1_4_12
void FASTCALL call_2_1_4_12(SH2_struct* sh)
{
  SH2cmpstr(sh,1,4);
}
#define init_2_1_5_12
void FASTCALL call_2_1_5_12(SH2_struct* sh)
{
  SH2cmpstr(sh,1,5);
}
#define init_2_1_6_12
void FASTCALL call_2_1_6_12(SH2_struct* sh)
{
  SH2cmpstr(sh,1,6);
}
#define init_2_1_7_12
void FASTCALL call_2_1_7_12(SH2_struct* sh)
{
  SH2cmpstr(sh,1,7);
}
#define init_2_1_8_12
void FASTCALL call_2_1_8_12(SH2_struct* sh)
{
  SH2cmpstr(sh,1,8);
}
#define init_2_1_9_12
void FASTCALL call_2_1_9_12(SH2_struct* sh)
{
  SH2cmpstr(sh,1,9);
}
#define init_2_1_10_12
void FASTCALL call_2_1_10_12(SH2_struct* sh)
{
  SH2cmpstr(sh,1,10);
}
#define init_2_1_11_12
void FASTCALL call_2_1_11_12(SH2_struct* sh)
{
  SH2cmpstr(sh,1,11);
}
#define init_2_1_12_12
void FASTCALL call_2_1_12_12(SH2_struct* sh)
{
  SH2cmpstr(sh,1,12);
}
#define init_2_1_13_12
void FASTCALL call_2_1_13_12(SH2_struct* sh)
{
  SH2cmpstr(sh,1,13);
}
#define init_2_1_14_12
void FASTCALL call_2_1_14_12(SH2_struct* sh)
{
  SH2cmpstr(sh,1,14);
}
#define init_2_1_15_12
void FASTCALL call_2_1_15_12(SH2_struct* sh)
{
  SH2cmpstr(sh,1,15);
}
#define init_2_2_0_12
void FASTCALL call_2_2_0_12(SH2_struct* sh)
{
  SH2cmpstr(sh,2,0);
}
#define init_2_2_1_12
void FASTCALL call_2_2_1_12(SH2_struct* sh)
{
  SH2cmpstr(sh,2,1);
}
#define init_2_2_2_12
void FASTCALL call_2_2_2_12(SH2_struct* sh)
{
  SH2cmpstr(sh,2,2);
}
#define init_2_2_3_12
void FASTCALL call_2_2_3_12(SH2_struct* sh)
{
  SH2cmpstr(sh,2,3);
}
#define init_2_2_4_12
void FASTCALL call_2_2_4_12(SH2_struct* sh)
{
  SH2cmpstr(sh,2,4);
}
#define init_2_2_5_12
void FASTCALL call_2_2_5_12(SH2_struct* sh)
{
  SH2cmpstr(sh,2,5);
}
#define init_2_2_6_12
void FASTCALL call_2_2_6_12(SH2_struct* sh)
{
  SH2cmpstr(sh,2,6);
}
#define init_2_2_7_12
void FASTCALL call_2_2_7_12(SH2_struct* sh)
{
  SH2cmpstr(sh,2,7);
}
#define init_2_2_8_12
void FASTCALL call_2_2_8_12(SH2_struct* sh)
{
  SH2cmpstr(sh,2,8);
}
#define init_2_2_9_12
void FASTCALL call_2_2_9_12(SH2_struct* sh)
{
  SH2cmpstr(sh,2,9);
}
#define init_2_2_10_12
void FASTCALL call_2_2_10_12(SH2_struct* sh)
{
  SH2cmpstr(sh,2,10);
}
#define init_2_2_11_12
void FASTCALL call_2_2_11_12(SH2_struct* sh)
{
  SH2cmpstr(sh,2,11);
}
#define init_2_2_12_12
void FASTCALL call_2_2_12_12(SH2_struct* sh)
{
  SH2cmpstr(sh,2,12);
}
#define init_2_2_13_12
void FASTCALL call_2_2_13_12(SH2_struct* sh)
{
  SH2cmpstr(sh,2,13);
}
#define init_2_2_14_12
void FASTCALL call_2_2_14_12(SH2_struct* sh)
{
  SH2cmpstr(sh,2,14);
}
#define init_2_2_15_12
void FASTCALL call_2_2_15_12(SH2_struct* sh)
{
  SH2cmpstr(sh,2,15);
}
#define init_2_3_0_12
void FASTCALL call_2_3_0_12(SH2_struct* sh)
{
  SH2cmpstr(sh,3,0);
}
#define init_2_3_1_12
void FASTCALL call_2_3_1_12(SH2_struct* sh)
{
  SH2cmpstr(sh,3,1);
}
#define init_2_3_2_12
void FASTCALL call_2_3_2_12(SH2_struct* sh)
{
  SH2cmpstr(sh,3,2);
}
#define init_2_3_3_12
void FASTCALL call_2_3_3_12(SH2_struct* sh)
{
  SH2cmpstr(sh,3,3);
}
#define init_2_3_4_12
void FASTCALL call_2_3_4_12(SH2_struct* sh)
{
  SH2cmpstr(sh,3,4);
}
#define init_2_3_5_12
void FASTCALL call_2_3_5_12(SH2_struct* sh)
{
  SH2cmpstr(sh,3,5);
}
#define init_2_3_6_12
void FASTCALL call_2_3_6_12(SH2_struct* sh)
{
  SH2cmpstr(sh,3,6);
}
#define init_2_3_7_12
void FASTCALL call_2_3_7_12(SH2_struct* sh)
{
  SH2cmpstr(sh,3,7);
}
#define init_2_3_8_12
void FASTCALL call_2_3_8_12(SH2_struct* sh)
{
  SH2cmpstr(sh,3,8);
}
#define init_2_3_9_12
void FASTCALL call_2_3_9_12(SH2_struct* sh)
{
  SH2cmpstr(sh,3,9);
}
#define init_2_3_10_12
void FASTCALL call_2_3_10_12(SH2_struct* sh)
{
  SH2cmpstr(sh,3,10);
}
#define init_2_3_11_12
void FASTCALL call_2_3_11_12(SH2_struct* sh)
{
  SH2cmpstr(sh,3,11);
}
#define init_2_3_12_12
void FASTCALL call_2_3_12_12(SH2_struct* sh)
{
  SH2cmpstr(sh,3,12);
}
#define init_2_3_13_12
void FASTCALL call_2_3_13_12(SH2_struct* sh)
{
  SH2cmpstr(sh,3,13);
}
#define init_2_3_14_12
void FASTCALL call_2_3_14_12(SH2_struct* sh)
{
  SH2cmpstr(sh,3,14);
}
#define init_2_3_15_12
void FASTCALL call_2_3_15_12(SH2_struct* sh)
{
  SH2cmpstr(sh,3,15);
}
#define init_2_4_0_12
void FASTCALL call_2_4_0_12(SH2_struct* sh)
{
  SH2cmpstr(sh,4,0);
}
#define init_2_4_1_12
void FASTCALL call_2_4_1_12(SH2_struct* sh)
{
  SH2cmpstr(sh,4,1);
}
#define init_2_4_2_12
void FASTCALL call_2_4_2_12(SH2_struct* sh)
{
  SH2cmpstr(sh,4,2);
}
#define init_2_4_3_12
void FASTCALL call_2_4_3_12(SH2_struct* sh)
{
  SH2cmpstr(sh,4,3);
}
#define init_2_4_4_12
void FASTCALL call_2_4_4_12(SH2_struct* sh)
{
  SH2cmpstr(sh,4,4);
}
#define init_2_4_5_12
void FASTCALL call_2_4_5_12(SH2_struct* sh)
{
  SH2cmpstr(sh,4,5);
}
#define init_2_4_6_12
void FASTCALL call_2_4_6_12(SH2_struct* sh)
{
  SH2cmpstr(sh,4,6);
}
#define init_2_4_7_12
void FASTCALL call_2_4_7_12(SH2_struct* sh)
{
  SH2cmpstr(sh,4,7);
}
#define init_2_4_8_12
void FASTCALL call_2_4_8_12(SH2_struct* sh)
{
  SH2cmpstr(sh,4,8);
}
#define init_2_4_9_12
void FASTCALL call_2_4_9_12(SH2_struct* sh)
{
  SH2cmpstr(sh,4,9);
}
#define init_2_4_10_12
void FASTCALL call_2_4_10_12(SH2_struct* sh)
{
  SH2cmpstr(sh,4,10);
}
#define init_2_4_11_12
void FASTCALL call_2_4_11_12(SH2_struct* sh)
{
  SH2cmpstr(sh,4,11);
}
#define init_2_4_12_12
void FASTCALL call_2_4_12_12(SH2_struct* sh)
{
  SH2cmpstr(sh,4,12);
}
#define init_2_4_13_12
void FASTCALL call_2_4_13_12(SH2_struct* sh)
{
  SH2cmpstr(sh,4,13);
}
#define init_2_4_14_12
void FASTCALL call_2_4_14_12(SH2_struct* sh)
{
  SH2cmpstr(sh,4,14);
}
#define init_2_4_15_12
void FASTCALL call_2_4_15_12(SH2_struct* sh)
{
  SH2cmpstr(sh,4,15);
}
#define init_2_5_0_12
void FASTCALL call_2_5_0_12(SH2_struct* sh)
{
  SH2cmpstr(sh,5,0);
}
#define init_2_5_1_12
void FASTCALL call_2_5_1_12(SH2_struct* sh)
{
  SH2cmpstr(sh,5,1);
}
#define init_2_5_2_12
void FASTCALL call_2_5_2_12(SH2_struct* sh)
{
  SH2cmpstr(sh,5,2);
}
#define init_2_5_3_12
void FASTCALL call_2_5_3_12(SH2_struct* sh)
{
  SH2cmpstr(sh,5,3);
}
#define init_2_5_4_12
void FASTCALL call_2_5_4_12(SH2_struct* sh)
{
  SH2cmpstr(sh,5,4);
}
#define init_2_5_5_12
void FASTCALL call_2_5_5_12(SH2_struct* sh)
{
  SH2cmpstr(sh,5,5);
}
#define init_2_5_6_12
void FASTCALL call_2_5_6_12(SH2_struct* sh)
{
  SH2cmpstr(sh,5,6);
}
#define init_2_5_7_12
void FASTCALL call_2_5_7_12(SH2_struct* sh)
{
  SH2cmpstr(sh,5,7);
}
#define init_2_5_8_12
void FASTCALL call_2_5_8_12(SH2_struct* sh)
{
  SH2cmpstr(sh,5,8);
}
#define init_2_5_9_12
void FASTCALL call_2_5_9_12(SH2_struct* sh)
{
  SH2cmpstr(sh,5,9);
}
#define init_2_5_10_12
void FASTCALL call_2_5_10_12(SH2_struct* sh)
{
  SH2cmpstr(sh,5,10);
}
#define init_2_5_11_12
void FASTCALL call_2_5_11_12(SH2_struct* sh)
{
  SH2cmpstr(sh,5,11);
}
#define init_2_5_12_12
void FASTCALL call_2_5_12_12(SH2_struct* sh)
{
  SH2cmpstr(sh,5,12);
}
#define init_2_5_13_12
void FASTCALL call_2_5_13_12(SH2_struct* sh)
{
  SH2cmpstr(sh,5,13);
}
#define init_2_5_14_12
void FASTCALL call_2_5_14_12(SH2_struct* sh)
{
  SH2cmpstr(sh,5,14);
}
#define init_2_5_15_12
void FASTCALL call_2_5_15_12(SH2_struct* sh)
{
  SH2cmpstr(sh,5,15);
}
#define init_2_6_0_12
void FASTCALL call_2_6_0_12(SH2_struct* sh)
{
  SH2cmpstr(sh,6,0);
}
#define init_2_6_1_12
void FASTCALL call_2_6_1_12(SH2_struct* sh)
{
  SH2cmpstr(sh,6,1);
}
#define init_2_6_2_12
void FASTCALL call_2_6_2_12(SH2_struct* sh)
{
  SH2cmpstr(sh,6,2);
}
#define init_2_6_3_12
void FASTCALL call_2_6_3_12(SH2_struct* sh)
{
  SH2cmpstr(sh,6,3);
}
#define init_2_6_4_12
void FASTCALL call_2_6_4_12(SH2_struct* sh)
{
  SH2cmpstr(sh,6,4);
}
#define init_2_6_5_12
void FASTCALL call_2_6_5_12(SH2_struct* sh)
{
  SH2cmpstr(sh,6,5);
}
#define init_2_6_6_12
void FASTCALL call_2_6_6_12(SH2_struct* sh)
{
  SH2cmpstr(sh,6,6);
}
#define init_2_6_7_12
void FASTCALL call_2_6_7_12(SH2_struct* sh)
{
  SH2cmpstr(sh,6,7);
}
#define init_2_6_8_12
void FASTCALL call_2_6_8_12(SH2_struct* sh)
{
  SH2cmpstr(sh,6,8);
}
#define init_2_6_9_12
void FASTCALL call_2_6_9_12(SH2_struct* sh)
{
  SH2cmpstr(sh,6,9);
}
#define init_2_6_10_12
void FASTCALL call_2_6_10_12(SH2_struct* sh)
{
  SH2cmpstr(sh,6,10);
}
#define init_2_6_11_12
void FASTCALL call_2_6_11_12(SH2_struct* sh)
{
  SH2cmpstr(sh,6,11);
}
#define init_2_6_12_12
void FASTCALL call_2_6_12_12(SH2_struct* sh)
{
  SH2cmpstr(sh,6,12);
}
#define init_2_6_13_12
void FASTCALL call_2_6_13_12(SH2_struct* sh)
{
  SH2cmpstr(sh,6,13);
}
#define init_2_6_14_12
void FASTCALL call_2_6_14_12(SH2_struct* sh)
{
  SH2cmpstr(sh,6,14);
}
#define init_2_6_15_12
void FASTCALL call_2_6_15_12(SH2_struct* sh)
{
  SH2cmpstr(sh,6,15);
}
#define init_2_7_0_12
void FASTCALL call_2_7_0_12(SH2_struct* sh)
{
  SH2cmpstr(sh,7,0);
}
#define init_2_7_1_12
void FASTCALL call_2_7_1_12(SH2_struct* sh)
{
  SH2cmpstr(sh,7,1);
}
#define init_2_7_2_12
void FASTCALL call_2_7_2_12(SH2_struct* sh)
{
  SH2cmpstr(sh,7,2);
}
#define init_2_7_3_12
void FASTCALL call_2_7_3_12(SH2_struct* sh)
{
  SH2cmpstr(sh,7,3);
}
#define init_2_7_4_12
void FASTCALL call_2_7_4_12(SH2_struct* sh)
{
  SH2cmpstr(sh,7,4);
}
#define init_2_7_5_12
void FASTCALL call_2_7_5_12(SH2_struct* sh)
{
  SH2cmpstr(sh,7,5);
}
#define init_2_7_6_12
void FASTCALL call_2_7_6_12(SH2_struct* sh)
{
  SH2cmpstr(sh,7,6);
}
#define init_2_7_7_12
void FASTCALL call_2_7_7_12(SH2_struct* sh)
{
  SH2cmpstr(sh,7,7);
}
#define init_2_7_8_12
void FASTCALL call_2_7_8_12(SH2_struct* sh)
{
  SH2cmpstr(sh,7,8);
}
#define init_2_7_9_12
void FASTCALL call_2_7_9_12(SH2_struct* sh)
{
  SH2cmpstr(sh,7,9);
}
#define init_2_7_10_12
void FASTCALL call_2_7_10_12(SH2_struct* sh)
{
  SH2cmpstr(sh,7,10);
}
#define init_2_7_11_12
void FASTCALL call_2_7_11_12(SH2_struct* sh)
{
  SH2cmpstr(sh,7,11);
}
#define init_2_7_12_12
void FASTCALL call_2_7_12_12(SH2_struct* sh)
{
  SH2cmpstr(sh,7,12);
}
#define init_2_7_13_12
void FASTCALL call_2_7_13_12(SH2_struct* sh)
{
  SH2cmpstr(sh,7,13);
}
#define init_2_7_14_12
void FASTCALL call_2_7_14_12(SH2_struct* sh)
{
  SH2cmpstr(sh,7,14);
}
#define init_2_7_15_12
void FASTCALL call_2_7_15_12(SH2_struct* sh)
{
  SH2cmpstr(sh,7,15);
}
#define init_2_8_0_12
void FASTCALL call_2_8_0_12(SH2_struct* sh)
{
  SH2cmpstr(sh,8,0);
}
#define init_2_8_1_12
void FASTCALL call_2_8_1_12(SH2_struct* sh)
{
  SH2cmpstr(sh,8,1);
}
#define init_2_8_2_12
void FASTCALL call_2_8_2_12(SH2_struct* sh)
{
  SH2cmpstr(sh,8,2);
}
#define init_2_8_3_12
void FASTCALL call_2_8_3_12(SH2_struct* sh)
{
  SH2cmpstr(sh,8,3);
}
#define init_2_8_4_12
void FASTCALL call_2_8_4_12(SH2_struct* sh)
{
  SH2cmpstr(sh,8,4);
}
#define init_2_8_5_12
void FASTCALL call_2_8_5_12(SH2_struct* sh)
{
  SH2cmpstr(sh,8,5);
}
#define init_2_8_6_12
void FASTCALL call_2_8_6_12(SH2_struct* sh)
{
  SH2cmpstr(sh,8,6);
}
#define init_2_8_7_12
void FASTCALL call_2_8_7_12(SH2_struct* sh)
{
  SH2cmpstr(sh,8,7);
}
#define init_2_8_8_12
void FASTCALL call_2_8_8_12(SH2_struct* sh)
{
  SH2cmpstr(sh,8,8);
}
#define init_2_8_9_12
void FASTCALL call_2_8_9_12(SH2_struct* sh)
{
  SH2cmpstr(sh,8,9);
}
#define init_2_8_10_12
void FASTCALL call_2_8_10_12(SH2_struct* sh)
{
  SH2cmpstr(sh,8,10);
}
#define init_2_8_11_12
void FASTCALL call_2_8_11_12(SH2_struct* sh)
{
  SH2cmpstr(sh,8,11);
}
#define init_2_8_12_12
void FASTCALL call_2_8_12_12(SH2_struct* sh)
{
  SH2cmpstr(sh,8,12);
}
#define init_2_8_13_12
void FASTCALL call_2_8_13_12(SH2_struct* sh)
{
  SH2cmpstr(sh,8,13);
}
#define init_2_8_14_12
void FASTCALL call_2_8_14_12(SH2_struct* sh)
{
  SH2cmpstr(sh,8,14);
}
#define init_2_8_15_12
void FASTCALL call_2_8_15_12(SH2_struct* sh)
{
  SH2cmpstr(sh,8,15);
}
#define init_2_9_0_12
void FASTCALL call_2_9_0_12(SH2_struct* sh)
{
  SH2cmpstr(sh,9,0);
}
#define init_2_9_1_12
void FASTCALL call_2_9_1_12(SH2_struct* sh)
{
  SH2cmpstr(sh,9,1);
}
#define init_2_9_2_12
void FASTCALL call_2_9_2_12(SH2_struct* sh)
{
  SH2cmpstr(sh,9,2);
}
#define init_2_9_3_12
void FASTCALL call_2_9_3_12(SH2_struct* sh)
{
  SH2cmpstr(sh,9,3);
}
#define init_2_9_4_12
void FASTCALL call_2_9_4_12(SH2_struct* sh)
{
  SH2cmpstr(sh,9,4);
}
#define init_2_9_5_12
void FASTCALL call_2_9_5_12(SH2_struct* sh)
{
  SH2cmpstr(sh,9,5);
}
#define init_2_9_6_12
void FASTCALL call_2_9_6_12(SH2_struct* sh)
{
  SH2cmpstr(sh,9,6);
}
#define init_2_9_7_12
void FASTCALL call_2_9_7_12(SH2_struct* sh)
{
  SH2cmpstr(sh,9,7);
}
#define init_2_9_8_12
void FASTCALL call_2_9_8_12(SH2_struct* sh)
{
  SH2cmpstr(sh,9,8);
}
#define init_2_9_9_12
void FASTCALL call_2_9_9_12(SH2_struct* sh)
{
  SH2cmpstr(sh,9,9);
}
#define init_2_9_10_12
void FASTCALL call_2_9_10_12(SH2_struct* sh)
{
  SH2cmpstr(sh,9,10);
}
#define init_2_9_11_12
void FASTCALL call_2_9_11_12(SH2_struct* sh)
{
  SH2cmpstr(sh,9,11);
}
#define init_2_9_12_12
void FASTCALL call_2_9_12_12(SH2_struct* sh)
{
  SH2cmpstr(sh,9,12);
}
#define init_2_9_13_12
void FASTCALL call_2_9_13_12(SH2_struct* sh)
{
  SH2cmpstr(sh,9,13);
}
#define init_2_9_14_12
void FASTCALL call_2_9_14_12(SH2_struct* sh)
{
  SH2cmpstr(sh,9,14);
}
#define init_2_9_15_12
void FASTCALL call_2_9_15_12(SH2_struct* sh)
{
  SH2cmpstr(sh,9,15);
}
#define init_2_10_0_12
void FASTCALL call_2_10_0_12(SH2_struct* sh)
{
  SH2cmpstr(sh,10,0);
}
#define init_2_10_1_12
void FASTCALL call_2_10_1_12(SH2_struct* sh)
{
  SH2cmpstr(sh,10,1);
}
#define init_2_10_2_12
void FASTCALL call_2_10_2_12(SH2_struct* sh)
{
  SH2cmpstr(sh,10,2);
}
#define init_2_10_3_12
void FASTCALL call_2_10_3_12(SH2_struct* sh)
{
  SH2cmpstr(sh,10,3);
}
#define init_2_10_4_12
void FASTCALL call_2_10_4_12(SH2_struct* sh)
{
  SH2cmpstr(sh,10,4);
}
#define init_2_10_5_12
void FASTCALL call_2_10_5_12(SH2_struct* sh)
{
  SH2cmpstr(sh,10,5);
}
#define init_2_10_6_12
void FASTCALL call_2_10_6_12(SH2_struct* sh)
{
  SH2cmpstr(sh,10,6);
}
#define init_2_10_7_12
void FASTCALL call_2_10_7_12(SH2_struct* sh)
{
  SH2cmpstr(sh,10,7);
}
#define init_2_10_8_12
void FASTCALL call_2_10_8_12(SH2_struct* sh)
{
  SH2cmpstr(sh,10,8);
}
#define init_2_10_9_12
void FASTCALL call_2_10_9_12(SH2_struct* sh)
{
  SH2cmpstr(sh,10,9);
}
#define init_2_10_10_12
void FASTCALL call_2_10_10_12(SH2_struct* sh)
{
  SH2cmpstr(sh,10,10);
}
#define init_2_10_11_12
void FASTCALL call_2_10_11_12(SH2_struct* sh)
{
  SH2cmpstr(sh,10,11);
}
#define init_2_10_12_12
void FASTCALL call_2_10_12_12(SH2_struct* sh)
{
  SH2cmpstr(sh,10,12);
}
#define init_2_10_13_12
void FASTCALL call_2_10_13_12(SH2_struct* sh)
{
  SH2cmpstr(sh,10,13);
}
#define init_2_10_14_12
void FASTCALL call_2_10_14_12(SH2_struct* sh)
{
  SH2cmpstr(sh,10,14);
}
#define init_2_10_15_12
void FASTCALL call_2_10_15_12(SH2_struct* sh)
{
  SH2cmpstr(sh,10,15);
}
#define init_2_11_0_12
void FASTCALL call_2_11_0_12(SH2_struct* sh)
{
  SH2cmpstr(sh,11,0);
}
#define init_2_11_1_12
void FASTCALL call_2_11_1_12(SH2_struct* sh)
{
  SH2cmpstr(sh,11,1);
}
#define init_2_11_2_12
void FASTCALL call_2_11_2_12(SH2_struct* sh)
{
  SH2cmpstr(sh,11,2);
}
#define init_2_11_3_12
void FASTCALL call_2_11_3_12(SH2_struct* sh)
{
  SH2cmpstr(sh,11,3);
}
#define init_2_11_4_12
void FASTCALL call_2_11_4_12(SH2_struct* sh)
{
  SH2cmpstr(sh,11,4);
}
#define init_2_11_5_12
void FASTCALL call_2_11_5_12(SH2_struct* sh)
{
  SH2cmpstr(sh,11,5);
}
#define init_2_11_6_12
void FASTCALL call_2_11_6_12(SH2_struct* sh)
{
  SH2cmpstr(sh,11,6);
}
#define init_2_11_7_12
void FASTCALL call_2_11_7_12(SH2_struct* sh)
{
  SH2cmpstr(sh,11,7);
}
#define init_2_11_8_12
void FASTCALL call_2_11_8_12(SH2_struct* sh)
{
  SH2cmpstr(sh,11,8);
}
#define init_2_11_9_12
void FASTCALL call_2_11_9_12(SH2_struct* sh)
{
  SH2cmpstr(sh,11,9);
}
#define init_2_11_10_12
void FASTCALL call_2_11_10_12(SH2_struct* sh)
{
  SH2cmpstr(sh,11,10);
}
#define init_2_11_11_12
void FASTCALL call_2_11_11_12(SH2_struct* sh)
{
  SH2cmpstr(sh,11,11);
}
#define init_2_11_12_12
void FASTCALL call_2_11_12_12(SH2_struct* sh)
{
  SH2cmpstr(sh,11,12);
}
#define init_2_11_13_12
void FASTCALL call_2_11_13_12(SH2_struct* sh)
{
  SH2cmpstr(sh,11,13);
}
#define init_2_11_14_12
void FASTCALL call_2_11_14_12(SH2_struct* sh)
{
  SH2cmpstr(sh,11,14);
}
#define init_2_11_15_12
void FASTCALL call_2_11_15_12(SH2_struct* sh)
{
  SH2cmpstr(sh,11,15);
}
#define init_2_12_0_12
void FASTCALL call_2_12_0_12(SH2_struct* sh)
{
  SH2cmpstr(sh,12,0);
}
#define init_2_12_1_12
void FASTCALL call_2_12_1_12(SH2_struct* sh)
{
  SH2cmpstr(sh,12,1);
}
#define init_2_12_2_12
void FASTCALL call_2_12_2_12(SH2_struct* sh)
{
  SH2cmpstr(sh,12,2);
}
#define init_2_12_3_12
void FASTCALL call_2_12_3_12(SH2_struct* sh)
{
  SH2cmpstr(sh,12,3);
}
#define init_2_12_4_12
void FASTCALL call_2_12_4_12(SH2_struct* sh)
{
  SH2cmpstr(sh,12,4);
}
#define init_2_12_5_12
void FASTCALL call_2_12_5_12(SH2_struct* sh)
{
  SH2cmpstr(sh,12,5);
}
#define init_2_12_6_12
void FASTCALL call_2_12_6_12(SH2_struct* sh)
{
  SH2cmpstr(sh,12,6);
}
#define init_2_12_7_12
void FASTCALL call_2_12_7_12(SH2_struct* sh)
{
  SH2cmpstr(sh,12,7);
}
#define init_2_12_8_12
void FASTCALL call_2_12_8_12(SH2_struct* sh)
{
  SH2cmpstr(sh,12,8);
}
#define init_2_12_9_12
void FASTCALL call_2_12_9_12(SH2_struct* sh)
{
  SH2cmpstr(sh,12,9);
}
#define init_2_12_10_12
void FASTCALL call_2_12_10_12(SH2_struct* sh)
{
  SH2cmpstr(sh,12,10);
}
#define init_2_12_11_12
void FASTCALL call_2_12_11_12(SH2_struct* sh)
{
  SH2cmpstr(sh,12,11);
}
#define init_2_12_12_12
void FASTCALL call_2_12_12_12(SH2_struct* sh)
{
  SH2cmpstr(sh,12,12);
}
#define init_2_12_13_12
void FASTCALL call_2_12_13_12(SH2_struct* sh)
{
  SH2cmpstr(sh,12,13);
}
#define init_2_12_14_12
void FASTCALL call_2_12_14_12(SH2_struct* sh)
{
  SH2cmpstr(sh,12,14);
}
#define init_2_12_15_12
void FASTCALL call_2_12_15_12(SH2_struct* sh)
{
  SH2cmpstr(sh,12,15);
}
#define init_2_13_0_12
void FASTCALL call_2_13_0_12(SH2_struct* sh)
{
  SH2cmpstr(sh,13,0);
}
#define init_2_13_1_12
void FASTCALL call_2_13_1_12(SH2_struct* sh)
{
  SH2cmpstr(sh,13,1);
}
#define init_2_13_2_12
void FASTCALL call_2_13_2_12(SH2_struct* sh)
{
  SH2cmpstr(sh,13,2);
}
#define init_2_13_3_12
void FASTCALL call_2_13_3_12(SH2_struct* sh)
{
  SH2cmpstr(sh,13,3);
}
#define init_2_13_4_12
void FASTCALL call_2_13_4_12(SH2_struct* sh)
{
  SH2cmpstr(sh,13,4);
}
#define init_2_13_5_12
void FASTCALL call_2_13_5_12(SH2_struct* sh)
{
  SH2cmpstr(sh,13,5);
}
#define init_2_13_6_12
void FASTCALL call_2_13_6_12(SH2_struct* sh)
{
  SH2cmpstr(sh,13,6);
}
#define init_2_13_7_12
void FASTCALL call_2_13_7_12(SH2_struct* sh)
{
  SH2cmpstr(sh,13,7);
}
#define init_2_13_8_12
void FASTCALL call_2_13_8_12(SH2_struct* sh)
{
  SH2cmpstr(sh,13,8);
}
#define init_2_13_9_12
void FASTCALL call_2_13_9_12(SH2_struct* sh)
{
  SH2cmpstr(sh,13,9);
}
#define init_2_13_10_12
void FASTCALL call_2_13_10_12(SH2_struct* sh)
{
  SH2cmpstr(sh,13,10);
}
#define init_2_13_11_12
void FASTCALL call_2_13_11_12(SH2_struct* sh)
{
  SH2cmpstr(sh,13,11);
}
#define init_2_13_12_12
void FASTCALL call_2_13_12_12(SH2_struct* sh)
{
  SH2cmpstr(sh,13,12);
}
#define init_2_13_13_12
void FASTCALL call_2_13_13_12(SH2_struct* sh)
{
  SH2cmpstr(sh,13,13);
}
#define init_2_13_14_12
void FASTCALL call_2_13_14_12(SH2_struct* sh)
{
  SH2cmpstr(sh,13,14);
}
#define init_2_13_15_12
void FASTCALL call_2_13_15_12(SH2_struct* sh)
{
  SH2cmpstr(sh,13,15);
}
#define init_2_14_0_12
void FASTCALL call_2_14_0_12(SH2_struct* sh)
{
  SH2cmpstr(sh,14,0);
}
#define init_2_14_1_12
void FASTCALL call_2_14_1_12(SH2_struct* sh)
{
  SH2cmpstr(sh,14,1);
}
#define init_2_14_2_12
void FASTCALL call_2_14_2_12(SH2_struct* sh)
{
  SH2cmpstr(sh,14,2);
}
#define init_2_14_3_12
void FASTCALL call_2_14_3_12(SH2_struct* sh)
{
  SH2cmpstr(sh,14,3);
}
#define init_2_14_4_12
void FASTCALL call_2_14_4_12(SH2_struct* sh)
{
  SH2cmpstr(sh,14,4);
}
#define init_2_14_5_12
void FASTCALL call_2_14_5_12(SH2_struct* sh)
{
  SH2cmpstr(sh,14,5);
}
#define init_2_14_6_12
void FASTCALL call_2_14_6_12(SH2_struct* sh)
{
  SH2cmpstr(sh,14,6);
}
#define init_2_14_7_12
void FASTCALL call_2_14_7_12(SH2_struct* sh)
{
  SH2cmpstr(sh,14,7);
}
#define init_2_14_8_12
void FASTCALL call_2_14_8_12(SH2_struct* sh)
{
  SH2cmpstr(sh,14,8);
}
#define init_2_14_9_12
void FASTCALL call_2_14_9_12(SH2_struct* sh)
{
  SH2cmpstr(sh,14,9);
}
#define init_2_14_10_12
void FASTCALL call_2_14_10_12(SH2_struct* sh)
{
  SH2cmpstr(sh,14,10);
}
#define init_2_14_11_12
void FASTCALL call_2_14_11_12(SH2_struct* sh)
{
  SH2cmpstr(sh,14,11);
}
#define init_2_14_12_12
void FASTCALL call_2_14_12_12(SH2_struct* sh)
{
  SH2cmpstr(sh,14,12);
}
#define init_2_14_13_12
void FASTCALL call_2_14_13_12(SH2_struct* sh)
{
  SH2cmpstr(sh,14,13);
}
#define init_2_14_14_12
void FASTCALL call_2_14_14_12(SH2_struct* sh)
{
  SH2cmpstr(sh,14,14);
}
#define init_2_14_15_12
void FASTCALL call_2_14_15_12(SH2_struct* sh)
{
  SH2cmpstr(sh,14,15);
}
#define init_2_15_0_12
void FASTCALL call_2_15_0_12(SH2_struct* sh)
{
  SH2cmpstr(sh,15,0);
}
#define init_2_15_1_12
void FASTCALL call_2_15_1_12(SH2_struct* sh)
{
  SH2cmpstr(sh,15,1);
}
#define init_2_15_2_12
void FASTCALL call_2_15_2_12(SH2_struct* sh)
{
  SH2cmpstr(sh,15,2);
}
#define init_2_15_3_12
void FASTCALL call_2_15_3_12(SH2_struct* sh)
{
  SH2cmpstr(sh,15,3);
}
#define init_2_15_4_12
void FASTCALL call_2_15_4_12(SH2_struct* sh)
{
  SH2cmpstr(sh,15,4);
}
#define init_2_15_5_12
void FASTCALL call_2_15_5_12(SH2_struct* sh)
{
  SH2cmpstr(sh,15,5);
}
#define init_2_15_6_12
void FASTCALL call_2_15_6_12(SH2_struct* sh)
{
  SH2cmpstr(sh,15,6);
}
#define init_2_15_7_12
void FASTCALL call_2_15_7_12(SH2_struct* sh)
{
  SH2cmpstr(sh,15,7);
}
#define init_2_15_8_12
void FASTCALL call_2_15_8_12(SH2_struct* sh)
{
  SH2cmpstr(sh,15,8);
}
#define init_2_15_9_12
void FASTCALL call_2_15_9_12(SH2_struct* sh)
{
  SH2cmpstr(sh,15,9);
}
#define init_2_15_10_12
void FASTCALL call_2_15_10_12(SH2_struct* sh)
{
  SH2cmpstr(sh,15,10);
}
#define init_2_15_11_12
void FASTCALL call_2_15_11_12(SH2_struct* sh)
{
  SH2cmpstr(sh,15,11);
}
#define init_2_15_12_12
void FASTCALL call_2_15_12_12(SH2_struct* sh)
{
  SH2cmpstr(sh,15,12);
}
#define init_2_15_13_12
void FASTCALL call_2_15_13_12(SH2_struct* sh)
{
  SH2cmpstr(sh,15,13);
}
#define init_2_15_14_12
void FASTCALL call_2_15_14_12(SH2_struct* sh)
{
  SH2cmpstr(sh,15,14);
}
#define init_2_15_15_12
void FASTCALL call_2_15_15_12(SH2_struct* sh)
{
  SH2cmpstr(sh,15,15);
}
#define init_4_0_1_1
void FASTCALL call_4_0_1_1(SH2_struct* sh)
{
  SH2cmppz(sh,0);
}
#define init_4_1_1_1
void FASTCALL call_4_1_1_1(SH2_struct* sh)
{
  SH2cmppz(sh,1);
}
#define init_4_2_1_1
void FASTCALL call_4_2_1_1(SH2_struct* sh)
{
  SH2cmppz(sh,2);
}
#define init_4_3_1_1
void FASTCALL call_4_3_1_1(SH2_struct* sh)
{
  SH2cmppz(sh,3);
}
#define init_4_4_1_1
void FASTCALL call_4_4_1_1(SH2_struct* sh)
{
  SH2cmppz(sh,4);
}
#define init_4_5_1_1
void FASTCALL call_4_5_1_1(SH2_struct* sh)
{
  SH2cmppz(sh,5);
}
#define init_4_6_1_1
void FASTCALL call_4_6_1_1(SH2_struct* sh)
{
  SH2cmppz(sh,6);
}
#define init_4_7_1_1
void FASTCALL call_4_7_1_1(SH2_struct* sh)
{
  SH2cmppz(sh,7);
}
#define init_4_8_1_1
void FASTCALL call_4_8_1_1(SH2_struct* sh)
{
  SH2cmppz(sh,8);
}
#define init_4_9_1_1
void FASTCALL call_4_9_1_1(SH2_struct* sh)
{
  SH2cmppz(sh,9);
}
#define init_4_10_1_1
void FASTCALL call_4_10_1_1(SH2_struct* sh)
{
  SH2cmppz(sh,10);
}
#define init_4_11_1_1
void FASTCALL call_4_11_1_1(SH2_struct* sh)
{
  SH2cmppz(sh,11);
}
#define init_4_12_1_1
void FASTCALL call_4_12_1_1(SH2_struct* sh)
{
  SH2cmppz(sh,12);
}
#define init_4_13_1_1
void FASTCALL call_4_13_1_1(SH2_struct* sh)
{
  SH2cmppz(sh,13);
}
#define init_4_14_1_1
void FASTCALL call_4_14_1_1(SH2_struct* sh)
{
  SH2cmppz(sh,14);
}
#define init_4_15_1_1
void FASTCALL call_4_15_1_1(SH2_struct* sh)
{
  SH2cmppz(sh,15);
}
#define init_4_0_1_5
void FASTCALL call_4_0_1_5(SH2_struct* sh)
{
  SH2cmppl(sh,0);
}
#define init_4_1_1_5
void FASTCALL call_4_1_1_5(SH2_struct* sh)
{
  SH2cmppl(sh,1);
}
#define init_4_2_1_5
void FASTCALL call_4_2_1_5(SH2_struct* sh)
{
  SH2cmppl(sh,2);
}
#define init_4_3_1_5
void FASTCALL call_4_3_1_5(SH2_struct* sh)
{
  SH2cmppl(sh,3);
}
#define init_4_4_1_5
void FASTCALL call_4_4_1_5(SH2_struct* sh)
{
  SH2cmppl(sh,4);
}
#define init_4_5_1_5
void FASTCALL call_4_5_1_5(SH2_struct* sh)
{
  SH2cmppl(sh,5);
}
#define init_4_6_1_5
void FASTCALL call_4_6_1_5(SH2_struct* sh)
{
  SH2cmppl(sh,6);
}
#define init_4_7_1_5
void FASTCALL call_4_7_1_5(SH2_struct* sh)
{
  SH2cmppl(sh,7);
}
#define init_4_8_1_5
void FASTCALL call_4_8_1_5(SH2_struct* sh)
{
  SH2cmppl(sh,8);
}
#define init_4_9_1_5
void FASTCALL call_4_9_1_5(SH2_struct* sh)
{
  SH2cmppl(sh,9);
}
#define init_4_10_1_5
void FASTCALL call_4_10_1_5(SH2_struct* sh)
{
  SH2cmppl(sh,10);
}
#define init_4_11_1_5
void FASTCALL call_4_11_1_5(SH2_struct* sh)
{
  SH2cmppl(sh,11);
}
#define init_4_12_1_5
void FASTCALL call_4_12_1_5(SH2_struct* sh)
{
  SH2cmppl(sh,12);
}
#define init_4_13_1_5
void FASTCALL call_4_13_1_5(SH2_struct* sh)
{
  SH2cmppl(sh,13);
}
#define init_4_14_1_5
void FASTCALL call_4_14_1_5(SH2_struct* sh)
{
  SH2cmppl(sh,14);
}
#define init_4_15_1_5
void FASTCALL call_4_15_1_5(SH2_struct* sh)
{
  SH2cmppl(sh,15);
}
#define init_8_8_0_0
void FASTCALL call_8_8_0_0(SH2_struct* sh)
{
  SH2cmpim(sh,(0<<4)+0);
}
#define init_8_8_0_1
void FASTCALL call_8_8_0_1(SH2_struct* sh)
{
  SH2cmpim(sh,(0<<4)+1);
}
#define init_8_8_0_2
void FASTCALL call_8_8_0_2(SH2_struct* sh)
{
  SH2cmpim(sh,(0<<4)+2);
}
#define init_8_8_0_3
void FASTCALL call_8_8_0_3(SH2_struct* sh)
{
  SH2cmpim(sh,(0<<4)+3);
}
#define init_8_8_0_4
void FASTCALL call_8_8_0_4(SH2_struct* sh)
{
  SH2cmpim(sh,(0<<4)+4);
}
#define init_8_8_0_5
void FASTCALL call_8_8_0_5(SH2_struct* sh)
{
  SH2cmpim(sh,(0<<4)+5);
}
#define init_8_8_0_6
void FASTCALL call_8_8_0_6(SH2_struct* sh)
{
  SH2cmpim(sh,(0<<4)+6);
}
#define init_8_8_0_7
void FASTCALL call_8_8_0_7(SH2_struct* sh)
{
  SH2cmpim(sh,(0<<4)+7);
}
#define init_8_8_0_8
void FASTCALL call_8_8_0_8(SH2_struct* sh)
{
  SH2cmpim(sh,(0<<4)+8);
}
#define init_8_8_0_9
void FASTCALL call_8_8_0_9(SH2_struct* sh)
{
  SH2cmpim(sh,(0<<4)+9);
}
#define init_8_8_0_10
void FASTCALL call_8_8_0_10(SH2_struct* sh)
{
  SH2cmpim(sh,(0<<4)+10);
}
#define init_8_8_0_11
void FASTCALL call_8_8_0_11(SH2_struct* sh)
{
  SH2cmpim(sh,(0<<4)+11);
}
#define init_8_8_0_12
void FASTCALL call_8_8_0_12(SH2_struct* sh)
{
  SH2cmpim(sh,(0<<4)+12);
}
#define init_8_8_0_13
void FASTCALL call_8_8_0_13(SH2_struct* sh)
{
  SH2cmpim(sh,(0<<4)+13);
}
#define init_8_8_0_14
void FASTCALL call_8_8_0_14(SH2_struct* sh)
{
  SH2cmpim(sh,(0<<4)+14);
}
#define init_8_8_0_15
void FASTCALL call_8_8_0_15(SH2_struct* sh)
{
  SH2cmpim(sh,(0<<4)+15);
}
#define init_8_8_1_0
void FASTCALL call_8_8_1_0(SH2_struct* sh)
{
  SH2cmpim(sh,(1<<4)+0);
}
#define init_8_8_1_1
void FASTCALL call_8_8_1_1(SH2_struct* sh)
{
  SH2cmpim(sh,(1<<4)+1);
}
#define init_8_8_1_2
void FASTCALL call_8_8_1_2(SH2_struct* sh)
{
  SH2cmpim(sh,(1<<4)+2);
}
#define init_8_8_1_3
void FASTCALL call_8_8_1_3(SH2_struct* sh)
{
  SH2cmpim(sh,(1<<4)+3);
}
#define init_8_8_1_4
void FASTCALL call_8_8_1_4(SH2_struct* sh)
{
  SH2cmpim(sh,(1<<4)+4);
}
#define init_8_8_1_5
void FASTCALL call_8_8_1_5(SH2_struct* sh)
{
  SH2cmpim(sh,(1<<4)+5);
}
#define init_8_8_1_6
void FASTCALL call_8_8_1_6(SH2_struct* sh)
{
  SH2cmpim(sh,(1<<4)+6);
}
#define init_8_8_1_7
void FASTCALL call_8_8_1_7(SH2_struct* sh)
{
  SH2cmpim(sh,(1<<4)+7);
}
#define init_8_8_1_8
void FASTCALL call_8_8_1_8(SH2_struct* sh)
{
  SH2cmpim(sh,(1<<4)+8);
}
#define init_8_8_1_9
void FASTCALL call_8_8_1_9(SH2_struct* sh)
{
  SH2cmpim(sh,(1<<4)+9);
}
#define init_8_8_1_10
void FASTCALL call_8_8_1_10(SH2_struct* sh)
{
  SH2cmpim(sh,(1<<4)+10);
}
#define init_8_8_1_11
void FASTCALL call_8_8_1_11(SH2_struct* sh)
{
  SH2cmpim(sh,(1<<4)+11);
}
#define init_8_8_1_12
void FASTCALL call_8_8_1_12(SH2_struct* sh)
{
  SH2cmpim(sh,(1<<4)+12);
}
#define init_8_8_1_13
void FASTCALL call_8_8_1_13(SH2_struct* sh)
{
  SH2cmpim(sh,(1<<4)+13);
}
#define init_8_8_1_14
void FASTCALL call_8_8_1_14(SH2_struct* sh)
{
  SH2cmpim(sh,(1<<4)+14);
}
#define init_8_8_1_15
void FASTCALL call_8_8_1_15(SH2_struct* sh)
{
  SH2cmpim(sh,(1<<4)+15);
}
#define init_8_8_2_0
void FASTCALL call_8_8_2_0(SH2_struct* sh)
{
  SH2cmpim(sh,(2<<4)+0);
}
#define init_8_8_2_1
void FASTCALL call_8_8_2_1(SH2_struct* sh)
{
  SH2cmpim(sh,(2<<4)+1);
}
#define init_8_8_2_2
void FASTCALL call_8_8_2_2(SH2_struct* sh)
{
  SH2cmpim(sh,(2<<4)+2);
}
#define init_8_8_2_3
void FASTCALL call_8_8_2_3(SH2_struct* sh)
{
  SH2cmpim(sh,(2<<4)+3);
}
#define init_8_8_2_4
void FASTCALL call_8_8_2_4(SH2_struct* sh)
{
  SH2cmpim(sh,(2<<4)+4);
}
#define init_8_8_2_5
void FASTCALL call_8_8_2_5(SH2_struct* sh)
{
  SH2cmpim(sh,(2<<4)+5);
}
#define init_8_8_2_6
void FASTCALL call_8_8_2_6(SH2_struct* sh)
{
  SH2cmpim(sh,(2<<4)+6);
}
#define init_8_8_2_7
void FASTCALL call_8_8_2_7(SH2_struct* sh)
{
  SH2cmpim(sh,(2<<4)+7);
}
#define init_8_8_2_8
void FASTCALL call_8_8_2_8(SH2_struct* sh)
{
  SH2cmpim(sh,(2<<4)+8);
}
#define init_8_8_2_9
void FASTCALL call_8_8_2_9(SH2_struct* sh)
{
  SH2cmpim(sh,(2<<4)+9);
}
#define init_8_8_2_10
void FASTCALL call_8_8_2_10(SH2_struct* sh)
{
  SH2cmpim(sh,(2<<4)+10);
}
#define init_8_8_2_11
void FASTCALL call_8_8_2_11(SH2_struct* sh)
{
  SH2cmpim(sh,(2<<4)+11);
}
#define init_8_8_2_12
void FASTCALL call_8_8_2_12(SH2_struct* sh)
{
  SH2cmpim(sh,(2<<4)+12);
}
#define init_8_8_2_13
void FASTCALL call_8_8_2_13(SH2_struct* sh)
{
  SH2cmpim(sh,(2<<4)+13);
}
#define init_8_8_2_14
void FASTCALL call_8_8_2_14(SH2_struct* sh)
{
  SH2cmpim(sh,(2<<4)+14);
}
#define init_8_8_2_15
void FASTCALL call_8_8_2_15(SH2_struct* sh)
{
  SH2cmpim(sh,(2<<4)+15);
}
#define init_8_8_3_0
void FASTCALL call_8_8_3_0(SH2_struct* sh)
{
  SH2cmpim(sh,(3<<4)+0);
}
#define init_8_8_3_1
void FASTCALL call_8_8_3_1(SH2_struct* sh)
{
  SH2cmpim(sh,(3<<4)+1);
}
#define init_8_8_3_2
void FASTCALL call_8_8_3_2(SH2_struct* sh)
{
  SH2cmpim(sh,(3<<4)+2);
}
#define init_8_8_3_3
void FASTCALL call_8_8_3_3(SH2_struct* sh)
{
  SH2cmpim(sh,(3<<4)+3);
}
#define init_8_8_3_4
void FASTCALL call_8_8_3_4(SH2_struct* sh)
{
  SH2cmpim(sh,(3<<4)+4);
}
#define init_8_8_3_5
void FASTCALL call_8_8_3_5(SH2_struct* sh)
{
  SH2cmpim(sh,(3<<4)+5);
}
#define init_8_8_3_6
void FASTCALL call_8_8_3_6(SH2_struct* sh)
{
  SH2cmpim(sh,(3<<4)+6);
}
#define init_8_8_3_7
void FASTCALL call_8_8_3_7(SH2_struct* sh)
{
  SH2cmpim(sh,(3<<4)+7);
}
#define init_8_8_3_8
void FASTCALL call_8_8_3_8(SH2_struct* sh)
{
  SH2cmpim(sh,(3<<4)+8);
}
#define init_8_8_3_9
void FASTCALL call_8_8_3_9(SH2_struct* sh)
{
  SH2cmpim(sh,(3<<4)+9);
}
#define init_8_8_3_10
void FASTCALL call_8_8_3_10(SH2_struct* sh)
{
  SH2cmpim(sh,(3<<4)+10);
}
#define init_8_8_3_11
void FASTCALL call_8_8_3_11(SH2_struct* sh)
{
  SH2cmpim(sh,(3<<4)+11);
}
#define init_8_8_3_12
void FASTCALL call_8_8_3_12(SH2_struct* sh)
{
  SH2cmpim(sh,(3<<4)+12);
}
#define init_8_8_3_13
void FASTCALL call_8_8_3_13(SH2_struct* sh)
{
  SH2cmpim(sh,(3<<4)+13);
}
#define init_8_8_3_14
void FASTCALL call_8_8_3_14(SH2_struct* sh)
{
  SH2cmpim(sh,(3<<4)+14);
}
#define init_8_8_3_15
void FASTCALL call_8_8_3_15(SH2_struct* sh)
{
  SH2cmpim(sh,(3<<4)+15);
}
#define init_8_8_4_0
void FASTCALL call_8_8_4_0(SH2_struct* sh)
{
  SH2cmpim(sh,(4<<4)+0);
}
#define init_8_8_4_1
void FASTCALL call_8_8_4_1(SH2_struct* sh)
{
  SH2cmpim(sh,(4<<4)+1);
}
#define init_8_8_4_2
void FASTCALL call_8_8_4_2(SH2_struct* sh)
{
  SH2cmpim(sh,(4<<4)+2);
}
#define init_8_8_4_3
void FASTCALL call_8_8_4_3(SH2_struct* sh)
{
  SH2cmpim(sh,(4<<4)+3);
}
#define init_8_8_4_4
void FASTCALL call_8_8_4_4(SH2_struct* sh)
{
  SH2cmpim(sh,(4<<4)+4);
}
#define init_8_8_4_5
void FASTCALL call_8_8_4_5(SH2_struct* sh)
{
  SH2cmpim(sh,(4<<4)+5);
}
#define init_8_8_4_6
void FASTCALL call_8_8_4_6(SH2_struct* sh)
{
  SH2cmpim(sh,(4<<4)+6);
}
#define init_8_8_4_7
void FASTCALL call_8_8_4_7(SH2_struct* sh)
{
  SH2cmpim(sh,(4<<4)+7);
}
#define init_8_8_4_8
void FASTCALL call_8_8_4_8(SH2_struct* sh)
{
  SH2cmpim(sh,(4<<4)+8);
}
#define init_8_8_4_9
void FASTCALL call_8_8_4_9(SH2_struct* sh)
{
  SH2cmpim(sh,(4<<4)+9);
}
#define init_8_8_4_10
void FASTCALL call_8_8_4_10(SH2_struct* sh)
{
  SH2cmpim(sh,(4<<4)+10);
}
#define init_8_8_4_11
void FASTCALL call_8_8_4_11(SH2_struct* sh)
{
  SH2cmpim(sh,(4<<4)+11);
}
#define init_8_8_4_12
void FASTCALL call_8_8_4_12(SH2_struct* sh)
{
  SH2cmpim(sh,(4<<4)+12);
}
#define init_8_8_4_13
void FASTCALL call_8_8_4_13(SH2_struct* sh)
{
  SH2cmpim(sh,(4<<4)+13);
}
#define init_8_8_4_14
void FASTCALL call_8_8_4_14(SH2_struct* sh)
{
  SH2cmpim(sh,(4<<4)+14);
}
#define init_8_8_4_15
void FASTCALL call_8_8_4_15(SH2_struct* sh)
{
  SH2cmpim(sh,(4<<4)+15);
}
#define init_8_8_5_0
void FASTCALL call_8_8_5_0(SH2_struct* sh)
{
  SH2cmpim(sh,(5<<4)+0);
}
#define init_8_8_5_1
void FASTCALL call_8_8_5_1(SH2_struct* sh)
{
  SH2cmpim(sh,(5<<4)+1);
}
#define init_8_8_5_2
void FASTCALL call_8_8_5_2(SH2_struct* sh)
{
  SH2cmpim(sh,(5<<4)+2);
}
#define init_8_8_5_3
void FASTCALL call_8_8_5_3(SH2_struct* sh)
{
  SH2cmpim(sh,(5<<4)+3);
}
#define init_8_8_5_4
void FASTCALL call_8_8_5_4(SH2_struct* sh)
{
  SH2cmpim(sh,(5<<4)+4);
}
#define init_8_8_5_5
void FASTCALL call_8_8_5_5(SH2_struct* sh)
{
  SH2cmpim(sh,(5<<4)+5);
}
#define init_8_8_5_6
void FASTCALL call_8_8_5_6(SH2_struct* sh)
{
  SH2cmpim(sh,(5<<4)+6);
}
#define init_8_8_5_7
void FASTCALL call_8_8_5_7(SH2_struct* sh)
{
  SH2cmpim(sh,(5<<4)+7);
}
#define init_8_8_5_8
void FASTCALL call_8_8_5_8(SH2_struct* sh)
{
  SH2cmpim(sh,(5<<4)+8);
}
#define init_8_8_5_9
void FASTCALL call_8_8_5_9(SH2_struct* sh)
{
  SH2cmpim(sh,(5<<4)+9);
}
#define init_8_8_5_10
void FASTCALL call_8_8_5_10(SH2_struct* sh)
{
  SH2cmpim(sh,(5<<4)+10);
}
#define init_8_8_5_11
void FASTCALL call_8_8_5_11(SH2_struct* sh)
{
  SH2cmpim(sh,(5<<4)+11);
}
#define init_8_8_5_12
void FASTCALL call_8_8_5_12(SH2_struct* sh)
{
  SH2cmpim(sh,(5<<4)+12);
}
#define init_8_8_5_13
void FASTCALL call_8_8_5_13(SH2_struct* sh)
{
  SH2cmpim(sh,(5<<4)+13);
}
#define init_8_8_5_14
void FASTCALL call_8_8_5_14(SH2_struct* sh)
{
  SH2cmpim(sh,(5<<4)+14);
}
#define init_8_8_5_15
void FASTCALL call_8_8_5_15(SH2_struct* sh)
{
  SH2cmpim(sh,(5<<4)+15);
}
#define init_8_8_6_0
void FASTCALL call_8_8_6_0(SH2_struct* sh)
{
  SH2cmpim(sh,(6<<4)+0);
}
#define init_8_8_6_1
void FASTCALL call_8_8_6_1(SH2_struct* sh)
{
  SH2cmpim(sh,(6<<4)+1);
}
#define init_8_8_6_2
void FASTCALL call_8_8_6_2(SH2_struct* sh)
{
  SH2cmpim(sh,(6<<4)+2);
}
#define init_8_8_6_3
void FASTCALL call_8_8_6_3(SH2_struct* sh)
{
  SH2cmpim(sh,(6<<4)+3);
}
#define init_8_8_6_4
void FASTCALL call_8_8_6_4(SH2_struct* sh)
{
  SH2cmpim(sh,(6<<4)+4);
}
#define init_8_8_6_5
void FASTCALL call_8_8_6_5(SH2_struct* sh)
{
  SH2cmpim(sh,(6<<4)+5);
}
#define init_8_8_6_6
void FASTCALL call_8_8_6_6(SH2_struct* sh)
{
  SH2cmpim(sh,(6<<4)+6);
}
#define init_8_8_6_7
void FASTCALL call_8_8_6_7(SH2_struct* sh)
{
  SH2cmpim(sh,(6<<4)+7);
}
#define init_8_8_6_8
void FASTCALL call_8_8_6_8(SH2_struct* sh)
{
  SH2cmpim(sh,(6<<4)+8);
}
#define init_8_8_6_9
void FASTCALL call_8_8_6_9(SH2_struct* sh)
{
  SH2cmpim(sh,(6<<4)+9);
}
#define init_8_8_6_10
void FASTCALL call_8_8_6_10(SH2_struct* sh)
{
  SH2cmpim(sh,(6<<4)+10);
}
#define init_8_8_6_11
void FASTCALL call_8_8_6_11(SH2_struct* sh)
{
  SH2cmpim(sh,(6<<4)+11);
}
#define init_8_8_6_12
void FASTCALL call_8_8_6_12(SH2_struct* sh)
{
  SH2cmpim(sh,(6<<4)+12);
}
#define init_8_8_6_13
void FASTCALL call_8_8_6_13(SH2_struct* sh)
{
  SH2cmpim(sh,(6<<4)+13);
}
#define init_8_8_6_14
void FASTCALL call_8_8_6_14(SH2_struct* sh)
{
  SH2cmpim(sh,(6<<4)+14);
}
#define init_8_8_6_15
void FASTCALL call_8_8_6_15(SH2_struct* sh)
{
  SH2cmpim(sh,(6<<4)+15);
}
#define init_8_8_7_0
void FASTCALL call_8_8_7_0(SH2_struct* sh)
{
  SH2cmpim(sh,(7<<4)+0);
}
#define init_8_8_7_1
void FASTCALL call_8_8_7_1(SH2_struct* sh)
{
  SH2cmpim(sh,(7<<4)+1);
}
#define init_8_8_7_2
void FASTCALL call_8_8_7_2(SH2_struct* sh)
{
  SH2cmpim(sh,(7<<4)+2);
}
#define init_8_8_7_3
void FASTCALL call_8_8_7_3(SH2_struct* sh)
{
  SH2cmpim(sh,(7<<4)+3);
}
#define init_8_8_7_4
void FASTCALL call_8_8_7_4(SH2_struct* sh)
{
  SH2cmpim(sh,(7<<4)+4);
}
#define init_8_8_7_5
void FASTCALL call_8_8_7_5(SH2_struct* sh)
{
  SH2cmpim(sh,(7<<4)+5);
}
#define init_8_8_7_6
void FASTCALL call_8_8_7_6(SH2_struct* sh)
{
  SH2cmpim(sh,(7<<4)+6);
}
#define init_8_8_7_7
void FASTCALL call_8_8_7_7(SH2_struct* sh)
{
  SH2cmpim(sh,(7<<4)+7);
}
#define init_8_8_7_8
void FASTCALL call_8_8_7_8(SH2_struct* sh)
{
  SH2cmpim(sh,(7<<4)+8);
}
#define init_8_8_7_9
void FASTCALL call_8_8_7_9(SH2_struct* sh)
{
  SH2cmpim(sh,(7<<4)+9);
}
#define init_8_8_7_10
void FASTCALL call_8_8_7_10(SH2_struct* sh)
{
  SH2cmpim(sh,(7<<4)+10);
}
#define init_8_8_7_11
void FASTCALL call_8_8_7_11(SH2_struct* sh)
{
  SH2cmpim(sh,(7<<4)+11);
}
#define init_8_8_7_12
void FASTCALL call_8_8_7_12(SH2_struct* sh)
{
  SH2cmpim(sh,(7<<4)+12);
}
#define init_8_8_7_13
void FASTCALL call_8_8_7_13(SH2_struct* sh)
{
  SH2cmpim(sh,(7<<4)+13);
}
#define init_8_8_7_14
void FASTCALL call_8_8_7_14(SH2_struct* sh)
{
  SH2cmpim(sh,(7<<4)+14);
}
#define init_8_8_7_15
void FASTCALL call_8_8_7_15(SH2_struct* sh)
{
  SH2cmpim(sh,(7<<4)+15);
}
#define init_8_8_8_0
void FASTCALL call_8_8_8_0(SH2_struct* sh)
{
  SH2cmpim(sh,(8<<4)+0);
}
#define init_8_8_8_1
void FASTCALL call_8_8_8_1(SH2_struct* sh)
{
  SH2cmpim(sh,(8<<4)+1);
}
#define init_8_8_8_2
void FASTCALL call_8_8_8_2(SH2_struct* sh)
{
  SH2cmpim(sh,(8<<4)+2);
}
#define init_8_8_8_3
void FASTCALL call_8_8_8_3(SH2_struct* sh)
{
  SH2cmpim(sh,(8<<4)+3);
}
#define init_8_8_8_4
void FASTCALL call_8_8_8_4(SH2_struct* sh)
{
  SH2cmpim(sh,(8<<4)+4);
}
#define init_8_8_8_5
void FASTCALL call_8_8_8_5(SH2_struct* sh)
{
  SH2cmpim(sh,(8<<4)+5);
}
#define init_8_8_8_6
void FASTCALL call_8_8_8_6(SH2_struct* sh)
{
  SH2cmpim(sh,(8<<4)+6);
}
#define init_8_8_8_7
void FASTCALL call_8_8_8_7(SH2_struct* sh)
{
  SH2cmpim(sh,(8<<4)+7);
}
#define init_8_8_8_8
void FASTCALL call_8_8_8_8(SH2_struct* sh)
{
  SH2cmpim(sh,(8<<4)+8);
}
#define init_8_8_8_9
void FASTCALL call_8_8_8_9(SH2_struct* sh)
{
  SH2cmpim(sh,(8<<4)+9);
}
#define init_8_8_8_10
void FASTCALL call_8_8_8_10(SH2_struct* sh)
{
  SH2cmpim(sh,(8<<4)+10);
}
#define init_8_8_8_11
void FASTCALL call_8_8_8_11(SH2_struct* sh)
{
  SH2cmpim(sh,(8<<4)+11);
}
#define init_8_8_8_12
void FASTCALL call_8_8_8_12(SH2_struct* sh)
{
  SH2cmpim(sh,(8<<4)+12);
}
#define init_8_8_8_13
void FASTCALL call_8_8_8_13(SH2_struct* sh)
{
  SH2cmpim(sh,(8<<4)+13);
}
#define init_8_8_8_14
void FASTCALL call_8_8_8_14(SH2_struct* sh)
{
  SH2cmpim(sh,(8<<4)+14);
}
#define init_8_8_8_15
void FASTCALL call_8_8_8_15(SH2_struct* sh)
{
  SH2cmpim(sh,(8<<4)+15);
}
#define init_8_8_9_0
void FASTCALL call_8_8_9_0(SH2_struct* sh)
{
  SH2cmpim(sh,(9<<4)+0);
}
#define init_8_8_9_1
void FASTCALL call_8_8_9_1(SH2_struct* sh)
{
  SH2cmpim(sh,(9<<4)+1);
}
#define init_8_8_9_2
void FASTCALL call_8_8_9_2(SH2_struct* sh)
{
  SH2cmpim(sh,(9<<4)+2);
}
#define init_8_8_9_3
void FASTCALL call_8_8_9_3(SH2_struct* sh)
{
  SH2cmpim(sh,(9<<4)+3);
}
#define init_8_8_9_4
void FASTCALL call_8_8_9_4(SH2_struct* sh)
{
  SH2cmpim(sh,(9<<4)+4);
}
#define init_8_8_9_5
void FASTCALL call_8_8_9_5(SH2_struct* sh)
{
  SH2cmpim(sh,(9<<4)+5);
}
#define init_8_8_9_6
void FASTCALL call_8_8_9_6(SH2_struct* sh)
{
  SH2cmpim(sh,(9<<4)+6);
}
#define init_8_8_9_7
void FASTCALL call_8_8_9_7(SH2_struct* sh)
{
  SH2cmpim(sh,(9<<4)+7);
}
#define init_8_8_9_8
void FASTCALL call_8_8_9_8(SH2_struct* sh)
{
  SH2cmpim(sh,(9<<4)+8);
}
#define init_8_8_9_9
void FASTCALL call_8_8_9_9(SH2_struct* sh)
{
  SH2cmpim(sh,(9<<4)+9);
}
#define init_8_8_9_10
void FASTCALL call_8_8_9_10(SH2_struct* sh)
{
  SH2cmpim(sh,(9<<4)+10);
}
#define init_8_8_9_11
void FASTCALL call_8_8_9_11(SH2_struct* sh)
{
  SH2cmpim(sh,(9<<4)+11);
}
#define init_8_8_9_12
void FASTCALL call_8_8_9_12(SH2_struct* sh)
{
  SH2cmpim(sh,(9<<4)+12);
}
#define init_8_8_9_13
void FASTCALL call_8_8_9_13(SH2_struct* sh)
{
  SH2cmpim(sh,(9<<4)+13);
}
#define init_8_8_9_14
void FASTCALL call_8_8_9_14(SH2_struct* sh)
{
  SH2cmpim(sh,(9<<4)+14);
}
#define init_8_8_9_15
void FASTCALL call_8_8_9_15(SH2_struct* sh)
{
  SH2cmpim(sh,(9<<4)+15);
}
#define init_8_8_10_0
void FASTCALL call_8_8_10_0(SH2_struct* sh)
{
  SH2cmpim(sh,(10<<4)+0);
}
#define init_8_8_10_1
void FASTCALL call_8_8_10_1(SH2_struct* sh)
{
  SH2cmpim(sh,(10<<4)+1);
}
#define init_8_8_10_2
void FASTCALL call_8_8_10_2(SH2_struct* sh)
{
  SH2cmpim(sh,(10<<4)+2);
}
#define init_8_8_10_3
void FASTCALL call_8_8_10_3(SH2_struct* sh)
{
  SH2cmpim(sh,(10<<4)+3);
}
#define init_8_8_10_4
void FASTCALL call_8_8_10_4(SH2_struct* sh)
{
  SH2cmpim(sh,(10<<4)+4);
}
#define init_8_8_10_5
void FASTCALL call_8_8_10_5(SH2_struct* sh)
{
  SH2cmpim(sh,(10<<4)+5);
}
#define init_8_8_10_6
void FASTCALL call_8_8_10_6(SH2_struct* sh)
{
  SH2cmpim(sh,(10<<4)+6);
}
#define init_8_8_10_7
void FASTCALL call_8_8_10_7(SH2_struct* sh)
{
  SH2cmpim(sh,(10<<4)+7);
}
#define init_8_8_10_8
void FASTCALL call_8_8_10_8(SH2_struct* sh)
{
  SH2cmpim(sh,(10<<4)+8);
}
#define init_8_8_10_9
void FASTCALL call_8_8_10_9(SH2_struct* sh)
{
  SH2cmpim(sh,(10<<4)+9);
}
#define init_8_8_10_10
void FASTCALL call_8_8_10_10(SH2_struct* sh)
{
  SH2cmpim(sh,(10<<4)+10);
}
#define init_8_8_10_11
void FASTCALL call_8_8_10_11(SH2_struct* sh)
{
  SH2cmpim(sh,(10<<4)+11);
}
#define init_8_8_10_12
void FASTCALL call_8_8_10_12(SH2_struct* sh)
{
  SH2cmpim(sh,(10<<4)+12);
}
#define init_8_8_10_13
void FASTCALL call_8_8_10_13(SH2_struct* sh)
{
  SH2cmpim(sh,(10<<4)+13);
}
#define init_8_8_10_14
void FASTCALL call_8_8_10_14(SH2_struct* sh)
{
  SH2cmpim(sh,(10<<4)+14);
}
#define init_8_8_10_15
void FASTCALL call_8_8_10_15(SH2_struct* sh)
{
  SH2cmpim(sh,(10<<4)+15);
}
#define init_8_8_11_0
void FASTCALL call_8_8_11_0(SH2_struct* sh)
{
  SH2cmpim(sh,(11<<4)+0);
}
#define init_8_8_11_1
void FASTCALL call_8_8_11_1(SH2_struct* sh)
{
  SH2cmpim(sh,(11<<4)+1);
}
#define init_8_8_11_2
void FASTCALL call_8_8_11_2(SH2_struct* sh)
{
  SH2cmpim(sh,(11<<4)+2);
}
#define init_8_8_11_3
void FASTCALL call_8_8_11_3(SH2_struct* sh)
{
  SH2cmpim(sh,(11<<4)+3);
}
#define init_8_8_11_4
void FASTCALL call_8_8_11_4(SH2_struct* sh)
{
  SH2cmpim(sh,(11<<4)+4);
}
#define init_8_8_11_5
void FASTCALL call_8_8_11_5(SH2_struct* sh)
{
  SH2cmpim(sh,(11<<4)+5);
}
#define init_8_8_11_6
void FASTCALL call_8_8_11_6(SH2_struct* sh)
{
  SH2cmpim(sh,(11<<4)+6);
}
#define init_8_8_11_7
void FASTCALL call_8_8_11_7(SH2_struct* sh)
{
  SH2cmpim(sh,(11<<4)+7);
}
#define init_8_8_11_8
void FASTCALL call_8_8_11_8(SH2_struct* sh)
{
  SH2cmpim(sh,(11<<4)+8);
}
#define init_8_8_11_9
void FASTCALL call_8_8_11_9(SH2_struct* sh)
{
  SH2cmpim(sh,(11<<4)+9);
}
#define init_8_8_11_10
void FASTCALL call_8_8_11_10(SH2_struct* sh)
{
  SH2cmpim(sh,(11<<4)+10);
}
#define init_8_8_11_11
void FASTCALL call_8_8_11_11(SH2_struct* sh)
{
  SH2cmpim(sh,(11<<4)+11);
}
#define init_8_8_11_12
void FASTCALL call_8_8_11_12(SH2_struct* sh)
{
  SH2cmpim(sh,(11<<4)+12);
}
#define init_8_8_11_13
void FASTCALL call_8_8_11_13(SH2_struct* sh)
{
  SH2cmpim(sh,(11<<4)+13);
}
#define init_8_8_11_14
void FASTCALL call_8_8_11_14(SH2_struct* sh)
{
  SH2cmpim(sh,(11<<4)+14);
}
#define init_8_8_11_15
void FASTCALL call_8_8_11_15(SH2_struct* sh)
{
  SH2cmpim(sh,(11<<4)+15);
}
#define init_8_8_12_0
void FASTCALL call_8_8_12_0(SH2_struct* sh)
{
  SH2cmpim(sh,(12<<4)+0);
}
#define init_8_8_12_1
void FASTCALL call_8_8_12_1(SH2_struct* sh)
{
  SH2cmpim(sh,(12<<4)+1);
}
#define init_8_8_12_2
void FASTCALL call_8_8_12_2(SH2_struct* sh)
{
  SH2cmpim(sh,(12<<4)+2);
}
#define init_8_8_12_3
void FASTCALL call_8_8_12_3(SH2_struct* sh)
{
  SH2cmpim(sh,(12<<4)+3);
}
#define init_8_8_12_4
void FASTCALL call_8_8_12_4(SH2_struct* sh)
{
  SH2cmpim(sh,(12<<4)+4);
}
#define init_8_8_12_5
void FASTCALL call_8_8_12_5(SH2_struct* sh)
{
  SH2cmpim(sh,(12<<4)+5);
}
#define init_8_8_12_6
void FASTCALL call_8_8_12_6(SH2_struct* sh)
{
  SH2cmpim(sh,(12<<4)+6);
}
#define init_8_8_12_7
void FASTCALL call_8_8_12_7(SH2_struct* sh)
{
  SH2cmpim(sh,(12<<4)+7);
}
#define init_8_8_12_8
void FASTCALL call_8_8_12_8(SH2_struct* sh)
{
  SH2cmpim(sh,(12<<4)+8);
}
#define init_8_8_12_9
void FASTCALL call_8_8_12_9(SH2_struct* sh)
{
  SH2cmpim(sh,(12<<4)+9);
}
#define init_8_8_12_10
void FASTCALL call_8_8_12_10(SH2_struct* sh)
{
  SH2cmpim(sh,(12<<4)+10);
}
#define init_8_8_12_11
void FASTCALL call_8_8_12_11(SH2_struct* sh)
{
  SH2cmpim(sh,(12<<4)+11);
}
#define init_8_8_12_12
void FASTCALL call_8_8_12_12(SH2_struct* sh)
{
  SH2cmpim(sh,(12<<4)+12);
}
#define init_8_8_12_13
void FASTCALL call_8_8_12_13(SH2_struct* sh)
{
  SH2cmpim(sh,(12<<4)+13);
}
#define init_8_8_12_14
void FASTCALL call_8_8_12_14(SH2_struct* sh)
{
  SH2cmpim(sh,(12<<4)+14);
}
#define init_8_8_12_15
void FASTCALL call_8_8_12_15(SH2_struct* sh)
{
  SH2cmpim(sh,(12<<4)+15);
}
#define init_8_8_13_0
void FASTCALL call_8_8_13_0(SH2_struct* sh)
{
  SH2cmpim(sh,(13<<4)+0);
}
#define init_8_8_13_1
void FASTCALL call_8_8_13_1(SH2_struct* sh)
{
  SH2cmpim(sh,(13<<4)+1);
}
#define init_8_8_13_2
void FASTCALL call_8_8_13_2(SH2_struct* sh)
{
  SH2cmpim(sh,(13<<4)+2);
}
#define init_8_8_13_3
void FASTCALL call_8_8_13_3(SH2_struct* sh)
{
  SH2cmpim(sh,(13<<4)+3);
}
#define init_8_8_13_4
void FASTCALL call_8_8_13_4(SH2_struct* sh)
{
  SH2cmpim(sh,(13<<4)+4);
}
#define init_8_8_13_5
void FASTCALL call_8_8_13_5(SH2_struct* sh)
{
  SH2cmpim(sh,(13<<4)+5);
}
#define init_8_8_13_6
void FASTCALL call_8_8_13_6(SH2_struct* sh)
{
  SH2cmpim(sh,(13<<4)+6);
}
#define init_8_8_13_7
void FASTCALL call_8_8_13_7(SH2_struct* sh)
{
  SH2cmpim(sh,(13<<4)+7);
}
#define init_8_8_13_8
void FASTCALL call_8_8_13_8(SH2_struct* sh)
{
  SH2cmpim(sh,(13<<4)+8);
}
#define init_8_8_13_9
void FASTCALL call_8_8_13_9(SH2_struct* sh)
{
  SH2cmpim(sh,(13<<4)+9);
}
#define init_8_8_13_10
void FASTCALL call_8_8_13_10(SH2_struct* sh)
{
  SH2cmpim(sh,(13<<4)+10);
}
#define init_8_8_13_11
void FASTCALL call_8_8_13_11(SH2_struct* sh)
{
  SH2cmpim(sh,(13<<4)+11);
}
#define init_8_8_13_12
void FASTCALL call_8_8_13_12(SH2_struct* sh)
{
  SH2cmpim(sh,(13<<4)+12);
}
#define init_8_8_13_13
void FASTCALL call_8_8_13_13(SH2_struct* sh)
{
  SH2cmpim(sh,(13<<4)+13);
}
#define init_8_8_13_14
void FASTCALL call_8_8_13_14(SH2_struct* sh)
{
  SH2cmpim(sh,(13<<4)+14);
}
#define init_8_8_13_15
void FASTCALL call_8_8_13_15(SH2_struct* sh)
{
  SH2cmpim(sh,(13<<4)+15);
}
#define init_8_8_14_0
void FASTCALL call_8_8_14_0(SH2_struct* sh)
{
  SH2cmpim(sh,(14<<4)+0);
}
#define init_8_8_14_1
void FASTCALL call_8_8_14_1(SH2_struct* sh)
{
  SH2cmpim(sh,(14<<4)+1);
}
#define init_8_8_14_2
void FASTCALL call_8_8_14_2(SH2_struct* sh)
{
  SH2cmpim(sh,(14<<4)+2);
}
#define init_8_8_14_3
void FASTCALL call_8_8_14_3(SH2_struct* sh)
{
  SH2cmpim(sh,(14<<4)+3);
}
#define init_8_8_14_4
void FASTCALL call_8_8_14_4(SH2_struct* sh)
{
  SH2cmpim(sh,(14<<4)+4);
}
#define init_8_8_14_5
void FASTCALL call_8_8_14_5(SH2_struct* sh)
{
  SH2cmpim(sh,(14<<4)+5);
}
#define init_8_8_14_6
void FASTCALL call_8_8_14_6(SH2_struct* sh)
{
  SH2cmpim(sh,(14<<4)+6);
}
#define init_8_8_14_7
void FASTCALL call_8_8_14_7(SH2_struct* sh)
{
  SH2cmpim(sh,(14<<4)+7);
}
#define init_8_8_14_8
void FASTCALL call_8_8_14_8(SH2_struct* sh)
{
  SH2cmpim(sh,(14<<4)+8);
}
#define init_8_8_14_9
void FASTCALL call_8_8_14_9(SH2_struct* sh)
{
  SH2cmpim(sh,(14<<4)+9);
}
#define init_8_8_14_10
void FASTCALL call_8_8_14_10(SH2_struct* sh)
{
  SH2cmpim(sh,(14<<4)+10);
}
#define init_8_8_14_11
void FASTCALL call_8_8_14_11(SH2_struct* sh)
{
  SH2cmpim(sh,(14<<4)+11);
}
#define init_8_8_14_12
void FASTCALL call_8_8_14_12(SH2_struct* sh)
{
  SH2cmpim(sh,(14<<4)+12);
}
#define init_8_8_14_13
void FASTCALL call_8_8_14_13(SH2_struct* sh)
{
  SH2cmpim(sh,(14<<4)+13);
}
#define init_8_8_14_14
void FASTCALL call_8_8_14_14(SH2_struct* sh)
{
  SH2cmpim(sh,(14<<4)+14);
}
#define init_8_8_14_15
void FASTCALL call_8_8_14_15(SH2_struct* sh)
{
  SH2cmpim(sh,(14<<4)+15);
}
#define init_8_8_15_0
void FASTCALL call_8_8_15_0(SH2_struct* sh)
{
  SH2cmpim(sh,(15<<4)+0);
}
#define init_8_8_15_1
void FASTCALL call_8_8_15_1(SH2_struct* sh)
{
  SH2cmpim(sh,(15<<4)+1);
}
#define init_8_8_15_2
void FASTCALL call_8_8_15_2(SH2_struct* sh)
{
  SH2cmpim(sh,(15<<4)+2);
}
#define init_8_8_15_3
void FASTCALL call_8_8_15_3(SH2_struct* sh)
{
  SH2cmpim(sh,(15<<4)+3);
}
#define init_8_8_15_4
void FASTCALL call_8_8_15_4(SH2_struct* sh)
{
  SH2cmpim(sh,(15<<4)+4);
}
#define init_8_8_15_5
void FASTCALL call_8_8_15_5(SH2_struct* sh)
{
  SH2cmpim(sh,(15<<4)+5);
}
#define init_8_8_15_6
void FASTCALL call_8_8_15_6(SH2_struct* sh)
{
  SH2cmpim(sh,(15<<4)+6);
}
#define init_8_8_15_7
void FASTCALL call_8_8_15_7(SH2_struct* sh)
{
  SH2cmpim(sh,(15<<4)+7);
}
#define init_8_8_15_8
void FASTCALL call_8_8_15_8(SH2_struct* sh)
{
  SH2cmpim(sh,(15<<4)+8);
}
#define init_8_8_15_9
void FASTCALL call_8_8_15_9(SH2_struct* sh)
{
  SH2cmpim(sh,(15<<4)+9);
}
#define init_8_8_15_10
void FASTCALL call_8_8_15_10(SH2_struct* sh)
{
  SH2cmpim(sh,(15<<4)+10);
}
#define init_8_8_15_11
void FASTCALL call_8_8_15_11(SH2_struct* sh)
{
  SH2cmpim(sh,(15<<4)+11);
}
#define init_8_8_15_12
void FASTCALL call_8_8_15_12(SH2_struct* sh)
{
  SH2cmpim(sh,(15<<4)+12);
}
#define init_8_8_15_13
void FASTCALL call_8_8_15_13(SH2_struct* sh)
{
  SH2cmpim(sh,(15<<4)+13);
}
#define init_8_8_15_14
void FASTCALL call_8_8_15_14(SH2_struct* sh)
{
  SH2cmpim(sh,(15<<4)+14);
}
#define init_8_8_15_15
void FASTCALL call_8_8_15_15(SH2_struct* sh)
{
  SH2cmpim(sh,(15<<4)+15);
}
#define init_3_0_0_2
void FASTCALL call_3_0_0_2(SH2_struct* sh)
{
  SH2cmphs(sh,0,0);
}
#define init_3_0_1_2
void FASTCALL call_3_0_1_2(SH2_struct* sh)
{
  SH2cmphs(sh,0,1);
}
#define init_3_0_2_2
void FASTCALL call_3_0_2_2(SH2_struct* sh)
{
  SH2cmphs(sh,0,2);
}
#define init_3_0_3_2
void FASTCALL call_3_0_3_2(SH2_struct* sh)
{
  SH2cmphs(sh,0,3);
}
#define init_3_0_4_2
void FASTCALL call_3_0_4_2(SH2_struct* sh)
{
  SH2cmphs(sh,0,4);
}
#define init_3_0_5_2
void FASTCALL call_3_0_5_2(SH2_struct* sh)
{
  SH2cmphs(sh,0,5);
}
#define init_3_0_6_2
void FASTCALL call_3_0_6_2(SH2_struct* sh)
{
  SH2cmphs(sh,0,6);
}
#define init_3_0_7_2
void FASTCALL call_3_0_7_2(SH2_struct* sh)
{
  SH2cmphs(sh,0,7);
}
#define init_3_0_8_2
void FASTCALL call_3_0_8_2(SH2_struct* sh)
{
  SH2cmphs(sh,0,8);
}
#define init_3_0_9_2
void FASTCALL call_3_0_9_2(SH2_struct* sh)
{
  SH2cmphs(sh,0,9);
}
#define init_3_0_10_2
void FASTCALL call_3_0_10_2(SH2_struct* sh)
{
  SH2cmphs(sh,0,10);
}
#define init_3_0_11_2
void FASTCALL call_3_0_11_2(SH2_struct* sh)
{
  SH2cmphs(sh,0,11);
}
#define init_3_0_12_2
void FASTCALL call_3_0_12_2(SH2_struct* sh)
{
  SH2cmphs(sh,0,12);
}
#define init_3_0_13_2
void FASTCALL call_3_0_13_2(SH2_struct* sh)
{
  SH2cmphs(sh,0,13);
}
#define init_3_0_14_2
void FASTCALL call_3_0_14_2(SH2_struct* sh)
{
  SH2cmphs(sh,0,14);
}
#define init_3_0_15_2
void FASTCALL call_3_0_15_2(SH2_struct* sh)
{
  SH2cmphs(sh,0,15);
}
#define init_3_1_0_2
void FASTCALL call_3_1_0_2(SH2_struct* sh)
{
  SH2cmphs(sh,1,0);
}
#define init_3_1_1_2
void FASTCALL call_3_1_1_2(SH2_struct* sh)
{
  SH2cmphs(sh,1,1);
}
#define init_3_1_2_2
void FASTCALL call_3_1_2_2(SH2_struct* sh)
{
  SH2cmphs(sh,1,2);
}
#define init_3_1_3_2
void FASTCALL call_3_1_3_2(SH2_struct* sh)
{
  SH2cmphs(sh,1,3);
}
#define init_3_1_4_2
void FASTCALL call_3_1_4_2(SH2_struct* sh)
{
  SH2cmphs(sh,1,4);
}
#define init_3_1_5_2
void FASTCALL call_3_1_5_2(SH2_struct* sh)
{
  SH2cmphs(sh,1,5);
}
#define init_3_1_6_2
void FASTCALL call_3_1_6_2(SH2_struct* sh)
{
  SH2cmphs(sh,1,6);
}
#define init_3_1_7_2
void FASTCALL call_3_1_7_2(SH2_struct* sh)
{
  SH2cmphs(sh,1,7);
}
#define init_3_1_8_2
void FASTCALL call_3_1_8_2(SH2_struct* sh)
{
  SH2cmphs(sh,1,8);
}
#define init_3_1_9_2
void FASTCALL call_3_1_9_2(SH2_struct* sh)
{
  SH2cmphs(sh,1,9);
}
#define init_3_1_10_2
void FASTCALL call_3_1_10_2(SH2_struct* sh)
{
  SH2cmphs(sh,1,10);
}
#define init_3_1_11_2
void FASTCALL call_3_1_11_2(SH2_struct* sh)
{
  SH2cmphs(sh,1,11);
}
#define init_3_1_12_2
void FASTCALL call_3_1_12_2(SH2_struct* sh)
{
  SH2cmphs(sh,1,12);
}
#define init_3_1_13_2
void FASTCALL call_3_1_13_2(SH2_struct* sh)
{
  SH2cmphs(sh,1,13);
}
#define init_3_1_14_2
void FASTCALL call_3_1_14_2(SH2_struct* sh)
{
  SH2cmphs(sh,1,14);
}
#define init_3_1_15_2
void FASTCALL call_3_1_15_2(SH2_struct* sh)
{
  SH2cmphs(sh,1,15);
}
#define init_3_2_0_2
void FASTCALL call_3_2_0_2(SH2_struct* sh)
{
  SH2cmphs(sh,2,0);
}
#define init_3_2_1_2
void FASTCALL call_3_2_1_2(SH2_struct* sh)
{
  SH2cmphs(sh,2,1);
}
#define init_3_2_2_2
void FASTCALL call_3_2_2_2(SH2_struct* sh)
{
  SH2cmphs(sh,2,2);
}
#define init_3_2_3_2
void FASTCALL call_3_2_3_2(SH2_struct* sh)
{
  SH2cmphs(sh,2,3);
}
#define init_3_2_4_2
void FASTCALL call_3_2_4_2(SH2_struct* sh)
{
  SH2cmphs(sh,2,4);
}
#define init_3_2_5_2
void FASTCALL call_3_2_5_2(SH2_struct* sh)
{
  SH2cmphs(sh,2,5);
}
#define init_3_2_6_2
void FASTCALL call_3_2_6_2(SH2_struct* sh)
{
  SH2cmphs(sh,2,6);
}
#define init_3_2_7_2
void FASTCALL call_3_2_7_2(SH2_struct* sh)
{
  SH2cmphs(sh,2,7);
}
#define init_3_2_8_2
void FASTCALL call_3_2_8_2(SH2_struct* sh)
{
  SH2cmphs(sh,2,8);
}
#define init_3_2_9_2
void FASTCALL call_3_2_9_2(SH2_struct* sh)
{
  SH2cmphs(sh,2,9);
}
#define init_3_2_10_2
void FASTCALL call_3_2_10_2(SH2_struct* sh)
{
  SH2cmphs(sh,2,10);
}
#define init_3_2_11_2
void FASTCALL call_3_2_11_2(SH2_struct* sh)
{
  SH2cmphs(sh,2,11);
}
#define init_3_2_12_2
void FASTCALL call_3_2_12_2(SH2_struct* sh)
{
  SH2cmphs(sh,2,12);
}
#define init_3_2_13_2
void FASTCALL call_3_2_13_2(SH2_struct* sh)
{
  SH2cmphs(sh,2,13);
}
#define init_3_2_14_2
void FASTCALL call_3_2_14_2(SH2_struct* sh)
{
  SH2cmphs(sh,2,14);
}
#define init_3_2_15_2
void FASTCALL call_3_2_15_2(SH2_struct* sh)
{
  SH2cmphs(sh,2,15);
}
#define init_3_3_0_2
void FASTCALL call_3_3_0_2(SH2_struct* sh)
{
  SH2cmphs(sh,3,0);
}
#define init_3_3_1_2
void FASTCALL call_3_3_1_2(SH2_struct* sh)
{
  SH2cmphs(sh,3,1);
}
#define init_3_3_2_2
void FASTCALL call_3_3_2_2(SH2_struct* sh)
{
  SH2cmphs(sh,3,2);
}
#define init_3_3_3_2
void FASTCALL call_3_3_3_2(SH2_struct* sh)
{
  SH2cmphs(sh,3,3);
}
#define init_3_3_4_2
void FASTCALL call_3_3_4_2(SH2_struct* sh)
{
  SH2cmphs(sh,3,4);
}
#define init_3_3_5_2
void FASTCALL call_3_3_5_2(SH2_struct* sh)
{
  SH2cmphs(sh,3,5);
}
#define init_3_3_6_2
void FASTCALL call_3_3_6_2(SH2_struct* sh)
{
  SH2cmphs(sh,3,6);
}
#define init_3_3_7_2
void FASTCALL call_3_3_7_2(SH2_struct* sh)
{
  SH2cmphs(sh,3,7);
}
#define init_3_3_8_2
void FASTCALL call_3_3_8_2(SH2_struct* sh)
{
  SH2cmphs(sh,3,8);
}
#define init_3_3_9_2
void FASTCALL call_3_3_9_2(SH2_struct* sh)
{
  SH2cmphs(sh,3,9);
}
#define init_3_3_10_2
void FASTCALL call_3_3_10_2(SH2_struct* sh)
{
  SH2cmphs(sh,3,10);
}
#define init_3_3_11_2
void FASTCALL call_3_3_11_2(SH2_struct* sh)
{
  SH2cmphs(sh,3,11);
}
#define init_3_3_12_2
void FASTCALL call_3_3_12_2(SH2_struct* sh)
{
  SH2cmphs(sh,3,12);
}
#define init_3_3_13_2
void FASTCALL call_3_3_13_2(SH2_struct* sh)
{
  SH2cmphs(sh,3,13);
}
#define init_3_3_14_2
void FASTCALL call_3_3_14_2(SH2_struct* sh)
{
  SH2cmphs(sh,3,14);
}
#define init_3_3_15_2
void FASTCALL call_3_3_15_2(SH2_struct* sh)
{
  SH2cmphs(sh,3,15);
}
#define init_3_4_0_2
void FASTCALL call_3_4_0_2(SH2_struct* sh)
{
  SH2cmphs(sh,4,0);
}
#define init_3_4_1_2
void FASTCALL call_3_4_1_2(SH2_struct* sh)
{
  SH2cmphs(sh,4,1);
}
#define init_3_4_2_2
void FASTCALL call_3_4_2_2(SH2_struct* sh)
{
  SH2cmphs(sh,4,2);
}
#define init_3_4_3_2
void FASTCALL call_3_4_3_2(SH2_struct* sh)
{
  SH2cmphs(sh,4,3);
}
#define init_3_4_4_2
void FASTCALL call_3_4_4_2(SH2_struct* sh)
{
  SH2cmphs(sh,4,4);
}
#define init_3_4_5_2
void FASTCALL call_3_4_5_2(SH2_struct* sh)
{
  SH2cmphs(sh,4,5);
}
#define init_3_4_6_2
void FASTCALL call_3_4_6_2(SH2_struct* sh)
{
  SH2cmphs(sh,4,6);
}
#define init_3_4_7_2
void FASTCALL call_3_4_7_2(SH2_struct* sh)
{
  SH2cmphs(sh,4,7);
}
#define init_3_4_8_2
void FASTCALL call_3_4_8_2(SH2_struct* sh)
{
  SH2cmphs(sh,4,8);
}
#define init_3_4_9_2
void FASTCALL call_3_4_9_2(SH2_struct* sh)
{
  SH2cmphs(sh,4,9);
}
#define init_3_4_10_2
void FASTCALL call_3_4_10_2(SH2_struct* sh)
{
  SH2cmphs(sh,4,10);
}
#define init_3_4_11_2
void FASTCALL call_3_4_11_2(SH2_struct* sh)
{
  SH2cmphs(sh,4,11);
}
#define init_3_4_12_2
void FASTCALL call_3_4_12_2(SH2_struct* sh)
{
  SH2cmphs(sh,4,12);
}
#define init_3_4_13_2
void FASTCALL call_3_4_13_2(SH2_struct* sh)
{
  SH2cmphs(sh,4,13);
}
#define init_3_4_14_2
void FASTCALL call_3_4_14_2(SH2_struct* sh)
{
  SH2cmphs(sh,4,14);
}
#define init_3_4_15_2
void FASTCALL call_3_4_15_2(SH2_struct* sh)
{
  SH2cmphs(sh,4,15);
}
#define init_3_5_0_2
void FASTCALL call_3_5_0_2(SH2_struct* sh)
{
  SH2cmphs(sh,5,0);
}
#define init_3_5_1_2
void FASTCALL call_3_5_1_2(SH2_struct* sh)
{
  SH2cmphs(sh,5,1);
}
#define init_3_5_2_2
void FASTCALL call_3_5_2_2(SH2_struct* sh)
{
  SH2cmphs(sh,5,2);
}
#define init_3_5_3_2
void FASTCALL call_3_5_3_2(SH2_struct* sh)
{
  SH2cmphs(sh,5,3);
}
#define init_3_5_4_2
void FASTCALL call_3_5_4_2(SH2_struct* sh)
{
  SH2cmphs(sh,5,4);
}
#define init_3_5_5_2
void FASTCALL call_3_5_5_2(SH2_struct* sh)
{
  SH2cmphs(sh,5,5);
}
#define init_3_5_6_2
void FASTCALL call_3_5_6_2(SH2_struct* sh)
{
  SH2cmphs(sh,5,6);
}
#define init_3_5_7_2
void FASTCALL call_3_5_7_2(SH2_struct* sh)
{
  SH2cmphs(sh,5,7);
}
#define init_3_5_8_2
void FASTCALL call_3_5_8_2(SH2_struct* sh)
{
  SH2cmphs(sh,5,8);
}
#define init_3_5_9_2
void FASTCALL call_3_5_9_2(SH2_struct* sh)
{
  SH2cmphs(sh,5,9);
}
#define init_3_5_10_2
void FASTCALL call_3_5_10_2(SH2_struct* sh)
{
  SH2cmphs(sh,5,10);
}
#define init_3_5_11_2
void FASTCALL call_3_5_11_2(SH2_struct* sh)
{
  SH2cmphs(sh,5,11);
}
#define init_3_5_12_2
void FASTCALL call_3_5_12_2(SH2_struct* sh)
{
  SH2cmphs(sh,5,12);
}
#define init_3_5_13_2
void FASTCALL call_3_5_13_2(SH2_struct* sh)
{
  SH2cmphs(sh,5,13);
}
#define init_3_5_14_2
void FASTCALL call_3_5_14_2(SH2_struct* sh)
{
  SH2cmphs(sh,5,14);
}
#define init_3_5_15_2
void FASTCALL call_3_5_15_2(SH2_struct* sh)
{
  SH2cmphs(sh,5,15);
}
#define init_3_6_0_2
void FASTCALL call_3_6_0_2(SH2_struct* sh)
{
  SH2cmphs(sh,6,0);
}
#define init_3_6_1_2
void FASTCALL call_3_6_1_2(SH2_struct* sh)
{
  SH2cmphs(sh,6,1);
}
#define init_3_6_2_2
void FASTCALL call_3_6_2_2(SH2_struct* sh)
{
  SH2cmphs(sh,6,2);
}
#define init_3_6_3_2
void FASTCALL call_3_6_3_2(SH2_struct* sh)
{
  SH2cmphs(sh,6,3);
}
#define init_3_6_4_2
void FASTCALL call_3_6_4_2(SH2_struct* sh)
{
  SH2cmphs(sh,6,4);
}
#define init_3_6_5_2
void FASTCALL call_3_6_5_2(SH2_struct* sh)
{
  SH2cmphs(sh,6,5);
}
#define init_3_6_6_2
void FASTCALL call_3_6_6_2(SH2_struct* sh)
{
  SH2cmphs(sh,6,6);
}
#define init_3_6_7_2
void FASTCALL call_3_6_7_2(SH2_struct* sh)
{
  SH2cmphs(sh,6,7);
}
#define init_3_6_8_2
void FASTCALL call_3_6_8_2(SH2_struct* sh)
{
  SH2cmphs(sh,6,8);
}
#define init_3_6_9_2
void FASTCALL call_3_6_9_2(SH2_struct* sh)
{
  SH2cmphs(sh,6,9);
}
#define init_3_6_10_2
void FASTCALL call_3_6_10_2(SH2_struct* sh)
{
  SH2cmphs(sh,6,10);
}
#define init_3_6_11_2
void FASTCALL call_3_6_11_2(SH2_struct* sh)
{
  SH2cmphs(sh,6,11);
}
#define init_3_6_12_2
void FASTCALL call_3_6_12_2(SH2_struct* sh)
{
  SH2cmphs(sh,6,12);
}
#define init_3_6_13_2
void FASTCALL call_3_6_13_2(SH2_struct* sh)
{
  SH2cmphs(sh,6,13);
}
#define init_3_6_14_2
void FASTCALL call_3_6_14_2(SH2_struct* sh)
{
  SH2cmphs(sh,6,14);
}
#define init_3_6_15_2
void FASTCALL call_3_6_15_2(SH2_struct* sh)
{
  SH2cmphs(sh,6,15);
}
#define init_3_7_0_2
void FASTCALL call_3_7_0_2(SH2_struct* sh)
{
  SH2cmphs(sh,7,0);
}
#define init_3_7_1_2
void FASTCALL call_3_7_1_2(SH2_struct* sh)
{
  SH2cmphs(sh,7,1);
}
#define init_3_7_2_2
void FASTCALL call_3_7_2_2(SH2_struct* sh)
{
  SH2cmphs(sh,7,2);
}
#define init_3_7_3_2
void FASTCALL call_3_7_3_2(SH2_struct* sh)
{
  SH2cmphs(sh,7,3);
}
#define init_3_7_4_2
void FASTCALL call_3_7_4_2(SH2_struct* sh)
{
  SH2cmphs(sh,7,4);
}
#define init_3_7_5_2
void FASTCALL call_3_7_5_2(SH2_struct* sh)
{
  SH2cmphs(sh,7,5);
}
#define init_3_7_6_2
void FASTCALL call_3_7_6_2(SH2_struct* sh)
{
  SH2cmphs(sh,7,6);
}
#define init_3_7_7_2
void FASTCALL call_3_7_7_2(SH2_struct* sh)
{
  SH2cmphs(sh,7,7);
}
#define init_3_7_8_2
void FASTCALL call_3_7_8_2(SH2_struct* sh)
{
  SH2cmphs(sh,7,8);
}
#define init_3_7_9_2
void FASTCALL call_3_7_9_2(SH2_struct* sh)
{
  SH2cmphs(sh,7,9);
}
#define init_3_7_10_2
void FASTCALL call_3_7_10_2(SH2_struct* sh)
{
  SH2cmphs(sh,7,10);
}
#define init_3_7_11_2
void FASTCALL call_3_7_11_2(SH2_struct* sh)
{
  SH2cmphs(sh,7,11);
}
#define init_3_7_12_2
void FASTCALL call_3_7_12_2(SH2_struct* sh)
{
  SH2cmphs(sh,7,12);
}
#define init_3_7_13_2
void FASTCALL call_3_7_13_2(SH2_struct* sh)
{
  SH2cmphs(sh,7,13);
}
#define init_3_7_14_2
void FASTCALL call_3_7_14_2(SH2_struct* sh)
{
  SH2cmphs(sh,7,14);
}
#define init_3_7_15_2
void FASTCALL call_3_7_15_2(SH2_struct* sh)
{
  SH2cmphs(sh,7,15);
}
#define init_3_8_0_2
void FASTCALL call_3_8_0_2(SH2_struct* sh)
{
  SH2cmphs(sh,8,0);
}
#define init_3_8_1_2
void FASTCALL call_3_8_1_2(SH2_struct* sh)
{
  SH2cmphs(sh,8,1);
}
#define init_3_8_2_2
void FASTCALL call_3_8_2_2(SH2_struct* sh)
{
  SH2cmphs(sh,8,2);
}
#define init_3_8_3_2
void FASTCALL call_3_8_3_2(SH2_struct* sh)
{
  SH2cmphs(sh,8,3);
}
#define init_3_8_4_2
void FASTCALL call_3_8_4_2(SH2_struct* sh)
{
  SH2cmphs(sh,8,4);
}
#define init_3_8_5_2
void FASTCALL call_3_8_5_2(SH2_struct* sh)
{
  SH2cmphs(sh,8,5);
}
#define init_3_8_6_2
void FASTCALL call_3_8_6_2(SH2_struct* sh)
{
  SH2cmphs(sh,8,6);
}
#define init_3_8_7_2
void FASTCALL call_3_8_7_2(SH2_struct* sh)
{
  SH2cmphs(sh,8,7);
}
#define init_3_8_8_2
void FASTCALL call_3_8_8_2(SH2_struct* sh)
{
  SH2cmphs(sh,8,8);
}
#define init_3_8_9_2
void FASTCALL call_3_8_9_2(SH2_struct* sh)
{
  SH2cmphs(sh,8,9);
}
#define init_3_8_10_2
void FASTCALL call_3_8_10_2(SH2_struct* sh)
{
  SH2cmphs(sh,8,10);
}
#define init_3_8_11_2
void FASTCALL call_3_8_11_2(SH2_struct* sh)
{
  SH2cmphs(sh,8,11);
}
#define init_3_8_12_2
void FASTCALL call_3_8_12_2(SH2_struct* sh)
{
  SH2cmphs(sh,8,12);
}
#define init_3_8_13_2
void FASTCALL call_3_8_13_2(SH2_struct* sh)
{
  SH2cmphs(sh,8,13);
}
#define init_3_8_14_2
void FASTCALL call_3_8_14_2(SH2_struct* sh)
{
  SH2cmphs(sh,8,14);
}
#define init_3_8_15_2
void FASTCALL call_3_8_15_2(SH2_struct* sh)
{
  SH2cmphs(sh,8,15);
}
#define init_3_9_0_2
void FASTCALL call_3_9_0_2(SH2_struct* sh)
{
  SH2cmphs(sh,9,0);
}
#define init_3_9_1_2
void FASTCALL call_3_9_1_2(SH2_struct* sh)
{
  SH2cmphs(sh,9,1);
}
#define init_3_9_2_2
void FASTCALL call_3_9_2_2(SH2_struct* sh)
{
  SH2cmphs(sh,9,2);
}
#define init_3_9_3_2
void FASTCALL call_3_9_3_2(SH2_struct* sh)
{
  SH2cmphs(sh,9,3);
}
#define init_3_9_4_2
void FASTCALL call_3_9_4_2(SH2_struct* sh)
{
  SH2cmphs(sh,9,4);
}
#define init_3_9_5_2
void FASTCALL call_3_9_5_2(SH2_struct* sh)
{
  SH2cmphs(sh,9,5);
}
#define init_3_9_6_2
void FASTCALL call_3_9_6_2(SH2_struct* sh)
{
  SH2cmphs(sh,9,6);
}
#define init_3_9_7_2
void FASTCALL call_3_9_7_2(SH2_struct* sh)
{
  SH2cmphs(sh,9,7);
}
#define init_3_9_8_2
void FASTCALL call_3_9_8_2(SH2_struct* sh)
{
  SH2cmphs(sh,9,8);
}
#define init_3_9_9_2
void FASTCALL call_3_9_9_2(SH2_struct* sh)
{
  SH2cmphs(sh,9,9);
}
#define init_3_9_10_2
void FASTCALL call_3_9_10_2(SH2_struct* sh)
{
  SH2cmphs(sh,9,10);
}
#define init_3_9_11_2
void FASTCALL call_3_9_11_2(SH2_struct* sh)
{
  SH2cmphs(sh,9,11);
}
#define init_3_9_12_2
void FASTCALL call_3_9_12_2(SH2_struct* sh)
{
  SH2cmphs(sh,9,12);
}
#define init_3_9_13_2
void FASTCALL call_3_9_13_2(SH2_struct* sh)
{
  SH2cmphs(sh,9,13);
}
#define init_3_9_14_2
void FASTCALL call_3_9_14_2(SH2_struct* sh)
{
  SH2cmphs(sh,9,14);
}
#define init_3_9_15_2
void FASTCALL call_3_9_15_2(SH2_struct* sh)
{
  SH2cmphs(sh,9,15);
}
#define init_3_10_0_2
void FASTCALL call_3_10_0_2(SH2_struct* sh)
{
  SH2cmphs(sh,10,0);
}
#define init_3_10_1_2
void FASTCALL call_3_10_1_2(SH2_struct* sh)
{
  SH2cmphs(sh,10,1);
}
#define init_3_10_2_2
void FASTCALL call_3_10_2_2(SH2_struct* sh)
{
  SH2cmphs(sh,10,2);
}
#define init_3_10_3_2
void FASTCALL call_3_10_3_2(SH2_struct* sh)
{
  SH2cmphs(sh,10,3);
}
#define init_3_10_4_2
void FASTCALL call_3_10_4_2(SH2_struct* sh)
{
  SH2cmphs(sh,10,4);
}
#define init_3_10_5_2
void FASTCALL call_3_10_5_2(SH2_struct* sh)
{
  SH2cmphs(sh,10,5);
}
#define init_3_10_6_2
void FASTCALL call_3_10_6_2(SH2_struct* sh)
{
  SH2cmphs(sh,10,6);
}
#define init_3_10_7_2
void FASTCALL call_3_10_7_2(SH2_struct* sh)
{
  SH2cmphs(sh,10,7);
}
#define init_3_10_8_2
void FASTCALL call_3_10_8_2(SH2_struct* sh)
{
  SH2cmphs(sh,10,8);
}
#define init_3_10_9_2
void FASTCALL call_3_10_9_2(SH2_struct* sh)
{
  SH2cmphs(sh,10,9);
}
#define init_3_10_10_2
void FASTCALL call_3_10_10_2(SH2_struct* sh)
{
  SH2cmphs(sh,10,10);
}
#define init_3_10_11_2
void FASTCALL call_3_10_11_2(SH2_struct* sh)
{
  SH2cmphs(sh,10,11);
}
#define init_3_10_12_2
void FASTCALL call_3_10_12_2(SH2_struct* sh)
{
  SH2cmphs(sh,10,12);
}
#define init_3_10_13_2
void FASTCALL call_3_10_13_2(SH2_struct* sh)
{
  SH2cmphs(sh,10,13);
}
#define init_3_10_14_2
void FASTCALL call_3_10_14_2(SH2_struct* sh)
{
  SH2cmphs(sh,10,14);
}
#define init_3_10_15_2
void FASTCALL call_3_10_15_2(SH2_struct* sh)
{
  SH2cmphs(sh,10,15);
}
#define init_3_11_0_2
void FASTCALL call_3_11_0_2(SH2_struct* sh)
{
  SH2cmphs(sh,11,0);
}
#define init_3_11_1_2
void FASTCALL call_3_11_1_2(SH2_struct* sh)
{
  SH2cmphs(sh,11,1);
}
#define init_3_11_2_2
void FASTCALL call_3_11_2_2(SH2_struct* sh)
{
  SH2cmphs(sh,11,2);
}
#define init_3_11_3_2
void FASTCALL call_3_11_3_2(SH2_struct* sh)
{
  SH2cmphs(sh,11,3);
}
#define init_3_11_4_2
void FASTCALL call_3_11_4_2(SH2_struct* sh)
{
  SH2cmphs(sh,11,4);
}
#define init_3_11_5_2
void FASTCALL call_3_11_5_2(SH2_struct* sh)
{
  SH2cmphs(sh,11,5);
}
#define init_3_11_6_2
void FASTCALL call_3_11_6_2(SH2_struct* sh)
{
  SH2cmphs(sh,11,6);
}
#define init_3_11_7_2
void FASTCALL call_3_11_7_2(SH2_struct* sh)
{
  SH2cmphs(sh,11,7);
}
#define init_3_11_8_2
void FASTCALL call_3_11_8_2(SH2_struct* sh)
{
  SH2cmphs(sh,11,8);
}
#define init_3_11_9_2
void FASTCALL call_3_11_9_2(SH2_struct* sh)
{
  SH2cmphs(sh,11,9);
}
#define init_3_11_10_2
void FASTCALL call_3_11_10_2(SH2_struct* sh)
{
  SH2cmphs(sh,11,10);
}
#define init_3_11_11_2
void FASTCALL call_3_11_11_2(SH2_struct* sh)
{
  SH2cmphs(sh,11,11);
}
#define init_3_11_12_2
void FASTCALL call_3_11_12_2(SH2_struct* sh)
{
  SH2cmphs(sh,11,12);
}
#define init_3_11_13_2
void FASTCALL call_3_11_13_2(SH2_struct* sh)
{
  SH2cmphs(sh,11,13);
}
#define init_3_11_14_2
void FASTCALL call_3_11_14_2(SH2_struct* sh)
{
  SH2cmphs(sh,11,14);
}
#define init_3_11_15_2
void FASTCALL call_3_11_15_2(SH2_struct* sh)
{
  SH2cmphs(sh,11,15);
}
#define init_3_12_0_2
void FASTCALL call_3_12_0_2(SH2_struct* sh)
{
  SH2cmphs(sh,12,0);
}
#define init_3_12_1_2
void FASTCALL call_3_12_1_2(SH2_struct* sh)
{
  SH2cmphs(sh,12,1);
}
#define init_3_12_2_2
void FASTCALL call_3_12_2_2(SH2_struct* sh)
{
  SH2cmphs(sh,12,2);
}
#define init_3_12_3_2
void FASTCALL call_3_12_3_2(SH2_struct* sh)
{
  SH2cmphs(sh,12,3);
}
#define init_3_12_4_2
void FASTCALL call_3_12_4_2(SH2_struct* sh)
{
  SH2cmphs(sh,12,4);
}
#define init_3_12_5_2
void FASTCALL call_3_12_5_2(SH2_struct* sh)
{
  SH2cmphs(sh,12,5);
}
#define init_3_12_6_2
void FASTCALL call_3_12_6_2(SH2_struct* sh)
{
  SH2cmphs(sh,12,6);
}
#define init_3_12_7_2
void FASTCALL call_3_12_7_2(SH2_struct* sh)
{
  SH2cmphs(sh,12,7);
}
#define init_3_12_8_2
void FASTCALL call_3_12_8_2(SH2_struct* sh)
{
  SH2cmphs(sh,12,8);
}
#define init_3_12_9_2
void FASTCALL call_3_12_9_2(SH2_struct* sh)
{
  SH2cmphs(sh,12,9);
}
#define init_3_12_10_2
void FASTCALL call_3_12_10_2(SH2_struct* sh)
{
  SH2cmphs(sh,12,10);
}
#define init_3_12_11_2
void FASTCALL call_3_12_11_2(SH2_struct* sh)
{
  SH2cmphs(sh,12,11);
}
#define init_3_12_12_2
void FASTCALL call_3_12_12_2(SH2_struct* sh)
{
  SH2cmphs(sh,12,12);
}
#define init_3_12_13_2
void FASTCALL call_3_12_13_2(SH2_struct* sh)
{
  SH2cmphs(sh,12,13);
}
#define init_3_12_14_2
void FASTCALL call_3_12_14_2(SH2_struct* sh)
{
  SH2cmphs(sh,12,14);
}
#define init_3_12_15_2
void FASTCALL call_3_12_15_2(SH2_struct* sh)
{
  SH2cmphs(sh,12,15);
}
#define init_3_13_0_2
void FASTCALL call_3_13_0_2(SH2_struct* sh)
{
  SH2cmphs(sh,13,0);
}
#define init_3_13_1_2
void FASTCALL call_3_13_1_2(SH2_struct* sh)
{
  SH2cmphs(sh,13,1);
}
#define init_3_13_2_2
void FASTCALL call_3_13_2_2(SH2_struct* sh)
{
  SH2cmphs(sh,13,2);
}
#define init_3_13_3_2
void FASTCALL call_3_13_3_2(SH2_struct* sh)
{
  SH2cmphs(sh,13,3);
}
#define init_3_13_4_2
void FASTCALL call_3_13_4_2(SH2_struct* sh)
{
  SH2cmphs(sh,13,4);
}
#define init_3_13_5_2
void FASTCALL call_3_13_5_2(SH2_struct* sh)
{
  SH2cmphs(sh,13,5);
}
#define init_3_13_6_2
void FASTCALL call_3_13_6_2(SH2_struct* sh)
{
  SH2cmphs(sh,13,6);
}
#define init_3_13_7_2
void FASTCALL call_3_13_7_2(SH2_struct* sh)
{
  SH2cmphs(sh,13,7);
}
#define init_3_13_8_2
void FASTCALL call_3_13_8_2(SH2_struct* sh)
{
  SH2cmphs(sh,13,8);
}
#define init_3_13_9_2
void FASTCALL call_3_13_9_2(SH2_struct* sh)
{
  SH2cmphs(sh,13,9);
}
#define init_3_13_10_2
void FASTCALL call_3_13_10_2(SH2_struct* sh)
{
  SH2cmphs(sh,13,10);
}
#define init_3_13_11_2
void FASTCALL call_3_13_11_2(SH2_struct* sh)
{
  SH2cmphs(sh,13,11);
}
#define init_3_13_12_2
void FASTCALL call_3_13_12_2(SH2_struct* sh)
{
  SH2cmphs(sh,13,12);
}
#define init_3_13_13_2
void FASTCALL call_3_13_13_2(SH2_struct* sh)
{
  SH2cmphs(sh,13,13);
}
#define init_3_13_14_2
void FASTCALL call_3_13_14_2(SH2_struct* sh)
{
  SH2cmphs(sh,13,14);
}
#define init_3_13_15_2
void FASTCALL call_3_13_15_2(SH2_struct* sh)
{
  SH2cmphs(sh,13,15);
}
#define init_3_14_0_2
void FASTCALL call_3_14_0_2(SH2_struct* sh)
{
  SH2cmphs(sh,14,0);
}
#define init_3_14_1_2
void FASTCALL call_3_14_1_2(SH2_struct* sh)
{
  SH2cmphs(sh,14,1);
}
#define init_3_14_2_2
void FASTCALL call_3_14_2_2(SH2_struct* sh)
{
  SH2cmphs(sh,14,2);
}
#define init_3_14_3_2
void FASTCALL call_3_14_3_2(SH2_struct* sh)
{
  SH2cmphs(sh,14,3);
}
#define init_3_14_4_2
void FASTCALL call_3_14_4_2(SH2_struct* sh)
{
  SH2cmphs(sh,14,4);
}
#define init_3_14_5_2
void FASTCALL call_3_14_5_2(SH2_struct* sh)
{
  SH2cmphs(sh,14,5);
}
#define init_3_14_6_2
void FASTCALL call_3_14_6_2(SH2_struct* sh)
{
  SH2cmphs(sh,14,6);
}
#define init_3_14_7_2
void FASTCALL call_3_14_7_2(SH2_struct* sh)
{
  SH2cmphs(sh,14,7);
}
#define init_3_14_8_2
void FASTCALL call_3_14_8_2(SH2_struct* sh)
{
  SH2cmphs(sh,14,8);
}
#define init_3_14_9_2
void FASTCALL call_3_14_9_2(SH2_struct* sh)
{
  SH2cmphs(sh,14,9);
}
#define init_3_14_10_2
void FASTCALL call_3_14_10_2(SH2_struct* sh)
{
  SH2cmphs(sh,14,10);
}
#define init_3_14_11_2
void FASTCALL call_3_14_11_2(SH2_struct* sh)
{
  SH2cmphs(sh,14,11);
}
#define init_3_14_12_2
void FASTCALL call_3_14_12_2(SH2_struct* sh)
{
  SH2cmphs(sh,14,12);
}
#define init_3_14_13_2
void FASTCALL call_3_14_13_2(SH2_struct* sh)
{
  SH2cmphs(sh,14,13);
}
#define init_3_14_14_2
void FASTCALL call_3_14_14_2(SH2_struct* sh)
{
  SH2cmphs(sh,14,14);
}
#define init_3_14_15_2
void FASTCALL call_3_14_15_2(SH2_struct* sh)
{
  SH2cmphs(sh,14,15);
}
#define init_3_15_0_2
void FASTCALL call_3_15_0_2(SH2_struct* sh)
{
  SH2cmphs(sh,15,0);
}
#define init_3_15_1_2
void FASTCALL call_3_15_1_2(SH2_struct* sh)
{
  SH2cmphs(sh,15,1);
}
#define init_3_15_2_2
void FASTCALL call_3_15_2_2(SH2_struct* sh)
{
  SH2cmphs(sh,15,2);
}
#define init_3_15_3_2
void FASTCALL call_3_15_3_2(SH2_struct* sh)
{
  SH2cmphs(sh,15,3);
}
#define init_3_15_4_2
void FASTCALL call_3_15_4_2(SH2_struct* sh)
{
  SH2cmphs(sh,15,4);
}
#define init_3_15_5_2
void FASTCALL call_3_15_5_2(SH2_struct* sh)
{
  SH2cmphs(sh,15,5);
}
#define init_3_15_6_2
void FASTCALL call_3_15_6_2(SH2_struct* sh)
{
  SH2cmphs(sh,15,6);
}
#define init_3_15_7_2
void FASTCALL call_3_15_7_2(SH2_struct* sh)
{
  SH2cmphs(sh,15,7);
}
#define init_3_15_8_2
void FASTCALL call_3_15_8_2(SH2_struct* sh)
{
  SH2cmphs(sh,15,8);
}
#define init_3_15_9_2
void FASTCALL call_3_15_9_2(SH2_struct* sh)
{
  SH2cmphs(sh,15,9);
}
#define init_3_15_10_2
void FASTCALL call_3_15_10_2(SH2_struct* sh)
{
  SH2cmphs(sh,15,10);
}
#define init_3_15_11_2
void FASTCALL call_3_15_11_2(SH2_struct* sh)
{
  SH2cmphs(sh,15,11);
}
#define init_3_15_12_2
void FASTCALL call_3_15_12_2(SH2_struct* sh)
{
  SH2cmphs(sh,15,12);
}
#define init_3_15_13_2
void FASTCALL call_3_15_13_2(SH2_struct* sh)
{
  SH2cmphs(sh,15,13);
}
#define init_3_15_14_2
void FASTCALL call_3_15_14_2(SH2_struct* sh)
{
  SH2cmphs(sh,15,14);
}
#define init_3_15_15_2
void FASTCALL call_3_15_15_2(SH2_struct* sh)
{
  SH2cmphs(sh,15,15);
}
#define init_3_0_0_6
void FASTCALL call_3_0_0_6(SH2_struct* sh)
{
  SH2cmphi(sh,0,0);
}
#define init_3_0_1_6
void FASTCALL call_3_0_1_6(SH2_struct* sh)
{
  SH2cmphi(sh,0,1);
}
#define init_3_0_2_6
void FASTCALL call_3_0_2_6(SH2_struct* sh)
{
  SH2cmphi(sh,0,2);
}
#define init_3_0_3_6
void FASTCALL call_3_0_3_6(SH2_struct* sh)
{
  SH2cmphi(sh,0,3);
}
#define init_3_0_4_6
void FASTCALL call_3_0_4_6(SH2_struct* sh)
{
  SH2cmphi(sh,0,4);
}
#define init_3_0_5_6
void FASTCALL call_3_0_5_6(SH2_struct* sh)
{
  SH2cmphi(sh,0,5);
}
#define init_3_0_6_6
void FASTCALL call_3_0_6_6(SH2_struct* sh)
{
  SH2cmphi(sh,0,6);
}
#define init_3_0_7_6
void FASTCALL call_3_0_7_6(SH2_struct* sh)
{
  SH2cmphi(sh,0,7);
}
#define init_3_0_8_6
void FASTCALL call_3_0_8_6(SH2_struct* sh)
{
  SH2cmphi(sh,0,8);
}
#define init_3_0_9_6
void FASTCALL call_3_0_9_6(SH2_struct* sh)
{
  SH2cmphi(sh,0,9);
}
#define init_3_0_10_6
void FASTCALL call_3_0_10_6(SH2_struct* sh)
{
  SH2cmphi(sh,0,10);
}
#define init_3_0_11_6
void FASTCALL call_3_0_11_6(SH2_struct* sh)
{
  SH2cmphi(sh,0,11);
}
#define init_3_0_12_6
void FASTCALL call_3_0_12_6(SH2_struct* sh)
{
  SH2cmphi(sh,0,12);
}
#define init_3_0_13_6
void FASTCALL call_3_0_13_6(SH2_struct* sh)
{
  SH2cmphi(sh,0,13);
}
#define init_3_0_14_6
void FASTCALL call_3_0_14_6(SH2_struct* sh)
{
  SH2cmphi(sh,0,14);
}
#define init_3_0_15_6
void FASTCALL call_3_0_15_6(SH2_struct* sh)
{
  SH2cmphi(sh,0,15);
}
#define init_3_1_0_6
void FASTCALL call_3_1_0_6(SH2_struct* sh)
{
  SH2cmphi(sh,1,0);
}
#define init_3_1_1_6
void FASTCALL call_3_1_1_6(SH2_struct* sh)
{
  SH2cmphi(sh,1,1);
}
#define init_3_1_2_6
void FASTCALL call_3_1_2_6(SH2_struct* sh)
{
  SH2cmphi(sh,1,2);
}
#define init_3_1_3_6
void FASTCALL call_3_1_3_6(SH2_struct* sh)
{
  SH2cmphi(sh,1,3);
}
#define init_3_1_4_6
void FASTCALL call_3_1_4_6(SH2_struct* sh)
{
  SH2cmphi(sh,1,4);
}
#define init_3_1_5_6
void FASTCALL call_3_1_5_6(SH2_struct* sh)
{
  SH2cmphi(sh,1,5);
}
#define init_3_1_6_6
void FASTCALL call_3_1_6_6(SH2_struct* sh)
{
  SH2cmphi(sh,1,6);
}
#define init_3_1_7_6
void FASTCALL call_3_1_7_6(SH2_struct* sh)
{
  SH2cmphi(sh,1,7);
}
#define init_3_1_8_6
void FASTCALL call_3_1_8_6(SH2_struct* sh)
{
  SH2cmphi(sh,1,8);
}
#define init_3_1_9_6
void FASTCALL call_3_1_9_6(SH2_struct* sh)
{
  SH2cmphi(sh,1,9);
}
#define init_3_1_10_6
void FASTCALL call_3_1_10_6(SH2_struct* sh)
{
  SH2cmphi(sh,1,10);
}
#define init_3_1_11_6
void FASTCALL call_3_1_11_6(SH2_struct* sh)
{
  SH2cmphi(sh,1,11);
}
#define init_3_1_12_6
void FASTCALL call_3_1_12_6(SH2_struct* sh)
{
  SH2cmphi(sh,1,12);
}
#define init_3_1_13_6
void FASTCALL call_3_1_13_6(SH2_struct* sh)
{
  SH2cmphi(sh,1,13);
}
#define init_3_1_14_6
void FASTCALL call_3_1_14_6(SH2_struct* sh)
{
  SH2cmphi(sh,1,14);
}
#define init_3_1_15_6
void FASTCALL call_3_1_15_6(SH2_struct* sh)
{
  SH2cmphi(sh,1,15);
}
#define init_3_2_0_6
void FASTCALL call_3_2_0_6(SH2_struct* sh)
{
  SH2cmphi(sh,2,0);
}
#define init_3_2_1_6
void FASTCALL call_3_2_1_6(SH2_struct* sh)
{
  SH2cmphi(sh,2,1);
}
#define init_3_2_2_6
void FASTCALL call_3_2_2_6(SH2_struct* sh)
{
  SH2cmphi(sh,2,2);
}
#define init_3_2_3_6
void FASTCALL call_3_2_3_6(SH2_struct* sh)
{
  SH2cmphi(sh,2,3);
}
#define init_3_2_4_6
void FASTCALL call_3_2_4_6(SH2_struct* sh)
{
  SH2cmphi(sh,2,4);
}
#define init_3_2_5_6
void FASTCALL call_3_2_5_6(SH2_struct* sh)
{
  SH2cmphi(sh,2,5);
}
#define init_3_2_6_6
void FASTCALL call_3_2_6_6(SH2_struct* sh)
{
  SH2cmphi(sh,2,6);
}
#define init_3_2_7_6
void FASTCALL call_3_2_7_6(SH2_struct* sh)
{
  SH2cmphi(sh,2,7);
}
#define init_3_2_8_6
void FASTCALL call_3_2_8_6(SH2_struct* sh)
{
  SH2cmphi(sh,2,8);
}
#define init_3_2_9_6
void FASTCALL call_3_2_9_6(SH2_struct* sh)
{
  SH2cmphi(sh,2,9);
}
#define init_3_2_10_6
void FASTCALL call_3_2_10_6(SH2_struct* sh)
{
  SH2cmphi(sh,2,10);
}
#define init_3_2_11_6
void FASTCALL call_3_2_11_6(SH2_struct* sh)
{
  SH2cmphi(sh,2,11);
}
#define init_3_2_12_6
void FASTCALL call_3_2_12_6(SH2_struct* sh)
{
  SH2cmphi(sh,2,12);
}
#define init_3_2_13_6
void FASTCALL call_3_2_13_6(SH2_struct* sh)
{
  SH2cmphi(sh,2,13);
}
#define init_3_2_14_6
void FASTCALL call_3_2_14_6(SH2_struct* sh)
{
  SH2cmphi(sh,2,14);
}
#define init_3_2_15_6
void FASTCALL call_3_2_15_6(SH2_struct* sh)
{
  SH2cmphi(sh,2,15);
}
#define init_3_3_0_6
void FASTCALL call_3_3_0_6(SH2_struct* sh)
{
  SH2cmphi(sh,3,0);
}
#define init_3_3_1_6
void FASTCALL call_3_3_1_6(SH2_struct* sh)
{
  SH2cmphi(sh,3,1);
}
#define init_3_3_2_6
void FASTCALL call_3_3_2_6(SH2_struct* sh)
{
  SH2cmphi(sh,3,2);
}
#define init_3_3_3_6
void FASTCALL call_3_3_3_6(SH2_struct* sh)
{
  SH2cmphi(sh,3,3);
}
#define init_3_3_4_6
void FASTCALL call_3_3_4_6(SH2_struct* sh)
{
  SH2cmphi(sh,3,4);
}
#define init_3_3_5_6
void FASTCALL call_3_3_5_6(SH2_struct* sh)
{
  SH2cmphi(sh,3,5);
}
#define init_3_3_6_6
void FASTCALL call_3_3_6_6(SH2_struct* sh)
{
  SH2cmphi(sh,3,6);
}
#define init_3_3_7_6
void FASTCALL call_3_3_7_6(SH2_struct* sh)
{
  SH2cmphi(sh,3,7);
}
#define init_3_3_8_6
void FASTCALL call_3_3_8_6(SH2_struct* sh)
{
  SH2cmphi(sh,3,8);
}
#define init_3_3_9_6
void FASTCALL call_3_3_9_6(SH2_struct* sh)
{
  SH2cmphi(sh,3,9);
}
#define init_3_3_10_6
void FASTCALL call_3_3_10_6(SH2_struct* sh)
{
  SH2cmphi(sh,3,10);
}
#define init_3_3_11_6
void FASTCALL call_3_3_11_6(SH2_struct* sh)
{
  SH2cmphi(sh,3,11);
}
#define init_3_3_12_6
void FASTCALL call_3_3_12_6(SH2_struct* sh)
{
  SH2cmphi(sh,3,12);
}
#define init_3_3_13_6
void FASTCALL call_3_3_13_6(SH2_struct* sh)
{
  SH2cmphi(sh,3,13);
}
#define init_3_3_14_6
void FASTCALL call_3_3_14_6(SH2_struct* sh)
{
  SH2cmphi(sh,3,14);
}
#define init_3_3_15_6
void FASTCALL call_3_3_15_6(SH2_struct* sh)
{
  SH2cmphi(sh,3,15);
}
#define init_3_4_0_6
void FASTCALL call_3_4_0_6(SH2_struct* sh)
{
  SH2cmphi(sh,4,0);
}
#define init_3_4_1_6
void FASTCALL call_3_4_1_6(SH2_struct* sh)
{
  SH2cmphi(sh,4,1);
}
#define init_3_4_2_6
void FASTCALL call_3_4_2_6(SH2_struct* sh)
{
  SH2cmphi(sh,4,2);
}
#define init_3_4_3_6
void FASTCALL call_3_4_3_6(SH2_struct* sh)
{
  SH2cmphi(sh,4,3);
}
#define init_3_4_4_6
void FASTCALL call_3_4_4_6(SH2_struct* sh)
{
  SH2cmphi(sh,4,4);
}
#define init_3_4_5_6
void FASTCALL call_3_4_5_6(SH2_struct* sh)
{
  SH2cmphi(sh,4,5);
}
#define init_3_4_6_6
void FASTCALL call_3_4_6_6(SH2_struct* sh)
{
  SH2cmphi(sh,4,6);
}
#define init_3_4_7_6
void FASTCALL call_3_4_7_6(SH2_struct* sh)
{
  SH2cmphi(sh,4,7);
}
#define init_3_4_8_6
void FASTCALL call_3_4_8_6(SH2_struct* sh)
{
  SH2cmphi(sh,4,8);
}
#define init_3_4_9_6
void FASTCALL call_3_4_9_6(SH2_struct* sh)
{
  SH2cmphi(sh,4,9);
}
#define init_3_4_10_6
void FASTCALL call_3_4_10_6(SH2_struct* sh)
{
  SH2cmphi(sh,4,10);
}
#define init_3_4_11_6
void FASTCALL call_3_4_11_6(SH2_struct* sh)
{
  SH2cmphi(sh,4,11);
}
#define init_3_4_12_6
void FASTCALL call_3_4_12_6(SH2_struct* sh)
{
  SH2cmphi(sh,4,12);
}
#define init_3_4_13_6
void FASTCALL call_3_4_13_6(SH2_struct* sh)
{
  SH2cmphi(sh,4,13);
}
#define init_3_4_14_6
void FASTCALL call_3_4_14_6(SH2_struct* sh)
{
  SH2cmphi(sh,4,14);
}
#define init_3_4_15_6
void FASTCALL call_3_4_15_6(SH2_struct* sh)
{
  SH2cmphi(sh,4,15);
}
#define init_3_5_0_6
void FASTCALL call_3_5_0_6(SH2_struct* sh)
{
  SH2cmphi(sh,5,0);
}
#define init_3_5_1_6
void FASTCALL call_3_5_1_6(SH2_struct* sh)
{
  SH2cmphi(sh,5,1);
}
#define init_3_5_2_6
void FASTCALL call_3_5_2_6(SH2_struct* sh)
{
  SH2cmphi(sh,5,2);
}
#define init_3_5_3_6
void FASTCALL call_3_5_3_6(SH2_struct* sh)
{
  SH2cmphi(sh,5,3);
}
#define init_3_5_4_6
void FASTCALL call_3_5_4_6(SH2_struct* sh)
{
  SH2cmphi(sh,5,4);
}
#define init_3_5_5_6
void FASTCALL call_3_5_5_6(SH2_struct* sh)
{
  SH2cmphi(sh,5,5);
}
#define init_3_5_6_6
void FASTCALL call_3_5_6_6(SH2_struct* sh)
{
  SH2cmphi(sh,5,6);
}
#define init_3_5_7_6
void FASTCALL call_3_5_7_6(SH2_struct* sh)
{
  SH2cmphi(sh,5,7);
}
#define init_3_5_8_6
void FASTCALL call_3_5_8_6(SH2_struct* sh)
{
  SH2cmphi(sh,5,8);
}
#define init_3_5_9_6
void FASTCALL call_3_5_9_6(SH2_struct* sh)
{
  SH2cmphi(sh,5,9);
}
#define init_3_5_10_6
void FASTCALL call_3_5_10_6(SH2_struct* sh)
{
  SH2cmphi(sh,5,10);
}
#define init_3_5_11_6
void FASTCALL call_3_5_11_6(SH2_struct* sh)
{
  SH2cmphi(sh,5,11);
}
#define init_3_5_12_6
void FASTCALL call_3_5_12_6(SH2_struct* sh)
{
  SH2cmphi(sh,5,12);
}
#define init_3_5_13_6
void FASTCALL call_3_5_13_6(SH2_struct* sh)
{
  SH2cmphi(sh,5,13);
}
#define init_3_5_14_6
void FASTCALL call_3_5_14_6(SH2_struct* sh)
{
  SH2cmphi(sh,5,14);
}
#define init_3_5_15_6
void FASTCALL call_3_5_15_6(SH2_struct* sh)
{
  SH2cmphi(sh,5,15);
}
#define init_3_6_0_6
void FASTCALL call_3_6_0_6(SH2_struct* sh)
{
  SH2cmphi(sh,6,0);
}
#define init_3_6_1_6
void FASTCALL call_3_6_1_6(SH2_struct* sh)
{
  SH2cmphi(sh,6,1);
}
#define init_3_6_2_6
void FASTCALL call_3_6_2_6(SH2_struct* sh)
{
  SH2cmphi(sh,6,2);
}
#define init_3_6_3_6
void FASTCALL call_3_6_3_6(SH2_struct* sh)
{
  SH2cmphi(sh,6,3);
}
#define init_3_6_4_6
void FASTCALL call_3_6_4_6(SH2_struct* sh)
{
  SH2cmphi(sh,6,4);
}
#define init_3_6_5_6
void FASTCALL call_3_6_5_6(SH2_struct* sh)
{
  SH2cmphi(sh,6,5);
}
#define init_3_6_6_6
void FASTCALL call_3_6_6_6(SH2_struct* sh)
{
  SH2cmphi(sh,6,6);
}
#define init_3_6_7_6
void FASTCALL call_3_6_7_6(SH2_struct* sh)
{
  SH2cmphi(sh,6,7);
}
#define init_3_6_8_6
void FASTCALL call_3_6_8_6(SH2_struct* sh)
{
  SH2cmphi(sh,6,8);
}
#define init_3_6_9_6
void FASTCALL call_3_6_9_6(SH2_struct* sh)
{
  SH2cmphi(sh,6,9);
}
#define init_3_6_10_6
void FASTCALL call_3_6_10_6(SH2_struct* sh)
{
  SH2cmphi(sh,6,10);
}
#define init_3_6_11_6
void FASTCALL call_3_6_11_6(SH2_struct* sh)
{
  SH2cmphi(sh,6,11);
}
#define init_3_6_12_6
void FASTCALL call_3_6_12_6(SH2_struct* sh)
{
  SH2cmphi(sh,6,12);
}
#define init_3_6_13_6
void FASTCALL call_3_6_13_6(SH2_struct* sh)
{
  SH2cmphi(sh,6,13);
}
#define init_3_6_14_6
void FASTCALL call_3_6_14_6(SH2_struct* sh)
{
  SH2cmphi(sh,6,14);
}
#define init_3_6_15_6
void FASTCALL call_3_6_15_6(SH2_struct* sh)
{
  SH2cmphi(sh,6,15);
}
#define init_3_7_0_6
void FASTCALL call_3_7_0_6(SH2_struct* sh)
{
  SH2cmphi(sh,7,0);
}
#define init_3_7_1_6
void FASTCALL call_3_7_1_6(SH2_struct* sh)
{
  SH2cmphi(sh,7,1);
}
#define init_3_7_2_6
void FASTCALL call_3_7_2_6(SH2_struct* sh)
{
  SH2cmphi(sh,7,2);
}
#define init_3_7_3_6
void FASTCALL call_3_7_3_6(SH2_struct* sh)
{
  SH2cmphi(sh,7,3);
}
#define init_3_7_4_6
void FASTCALL call_3_7_4_6(SH2_struct* sh)
{
  SH2cmphi(sh,7,4);
}
#define init_3_7_5_6
void FASTCALL call_3_7_5_6(SH2_struct* sh)
{
  SH2cmphi(sh,7,5);
}
#define init_3_7_6_6
void FASTCALL call_3_7_6_6(SH2_struct* sh)
{
  SH2cmphi(sh,7,6);
}
#define init_3_7_7_6
void FASTCALL call_3_7_7_6(SH2_struct* sh)
{
  SH2cmphi(sh,7,7);
}
#define init_3_7_8_6
void FASTCALL call_3_7_8_6(SH2_struct* sh)
{
  SH2cmphi(sh,7,8);
}
#define init_3_7_9_6
void FASTCALL call_3_7_9_6(SH2_struct* sh)
{
  SH2cmphi(sh,7,9);
}
#define init_3_7_10_6
void FASTCALL call_3_7_10_6(SH2_struct* sh)
{
  SH2cmphi(sh,7,10);
}
#define init_3_7_11_6
void FASTCALL call_3_7_11_6(SH2_struct* sh)
{
  SH2cmphi(sh,7,11);
}
#define init_3_7_12_6
void FASTCALL call_3_7_12_6(SH2_struct* sh)
{
  SH2cmphi(sh,7,12);
}
#define init_3_7_13_6
void FASTCALL call_3_7_13_6(SH2_struct* sh)
{
  SH2cmphi(sh,7,13);
}
#define init_3_7_14_6
void FASTCALL call_3_7_14_6(SH2_struct* sh)
{
  SH2cmphi(sh,7,14);
}
#define init_3_7_15_6
void FASTCALL call_3_7_15_6(SH2_struct* sh)
{
  SH2cmphi(sh,7,15);
}
#define init_3_8_0_6
void FASTCALL call_3_8_0_6(SH2_struct* sh)
{
  SH2cmphi(sh,8,0);
}
#define init_3_8_1_6
void FASTCALL call_3_8_1_6(SH2_struct* sh)
{
  SH2cmphi(sh,8,1);
}
#define init_3_8_2_6
void FASTCALL call_3_8_2_6(SH2_struct* sh)
{
  SH2cmphi(sh,8,2);
}
#define init_3_8_3_6
void FASTCALL call_3_8_3_6(SH2_struct* sh)
{
  SH2cmphi(sh,8,3);
}
#define init_3_8_4_6
void FASTCALL call_3_8_4_6(SH2_struct* sh)
{
  SH2cmphi(sh,8,4);
}
#define init_3_8_5_6
void FASTCALL call_3_8_5_6(SH2_struct* sh)
{
  SH2cmphi(sh,8,5);
}
#define init_3_8_6_6
void FASTCALL call_3_8_6_6(SH2_struct* sh)
{
  SH2cmphi(sh,8,6);
}
#define init_3_8_7_6
void FASTCALL call_3_8_7_6(SH2_struct* sh)
{
  SH2cmphi(sh,8,7);
}
#define init_3_8_8_6
void FASTCALL call_3_8_8_6(SH2_struct* sh)
{
  SH2cmphi(sh,8,8);
}
#define init_3_8_9_6
void FASTCALL call_3_8_9_6(SH2_struct* sh)
{
  SH2cmphi(sh,8,9);
}
#define init_3_8_10_6
void FASTCALL call_3_8_10_6(SH2_struct* sh)
{
  SH2cmphi(sh,8,10);
}
#define init_3_8_11_6
void FASTCALL call_3_8_11_6(SH2_struct* sh)
{
  SH2cmphi(sh,8,11);
}
#define init_3_8_12_6
void FASTCALL call_3_8_12_6(SH2_struct* sh)
{
  SH2cmphi(sh,8,12);
}
#define init_3_8_13_6
void FASTCALL call_3_8_13_6(SH2_struct* sh)
{
  SH2cmphi(sh,8,13);
}
#define init_3_8_14_6
void FASTCALL call_3_8_14_6(SH2_struct* sh)
{
  SH2cmphi(sh,8,14);
}
#define init_3_8_15_6
void FASTCALL call_3_8_15_6(SH2_struct* sh)
{
  SH2cmphi(sh,8,15);
}
#define init_3_9_0_6
void FASTCALL call_3_9_0_6(SH2_struct* sh)
{
  SH2cmphi(sh,9,0);
}
#define init_3_9_1_6
void FASTCALL call_3_9_1_6(SH2_struct* sh)
{
  SH2cmphi(sh,9,1);
}
#define init_3_9_2_6
void FASTCALL call_3_9_2_6(SH2_struct* sh)
{
  SH2cmphi(sh,9,2);
}
#define init_3_9_3_6
void FASTCALL call_3_9_3_6(SH2_struct* sh)
{
  SH2cmphi(sh,9,3);
}
#define init_3_9_4_6
void FASTCALL call_3_9_4_6(SH2_struct* sh)
{
  SH2cmphi(sh,9,4);
}
#define init_3_9_5_6
void FASTCALL call_3_9_5_6(SH2_struct* sh)
{
  SH2cmphi(sh,9,5);
}
#define init_3_9_6_6
void FASTCALL call_3_9_6_6(SH2_struct* sh)
{
  SH2cmphi(sh,9,6);
}
#define init_3_9_7_6
void FASTCALL call_3_9_7_6(SH2_struct* sh)
{
  SH2cmphi(sh,9,7);
}
#define init_3_9_8_6
void FASTCALL call_3_9_8_6(SH2_struct* sh)
{
  SH2cmphi(sh,9,8);
}
#define init_3_9_9_6
void FASTCALL call_3_9_9_6(SH2_struct* sh)
{
  SH2cmphi(sh,9,9);
}
#define init_3_9_10_6
void FASTCALL call_3_9_10_6(SH2_struct* sh)
{
  SH2cmphi(sh,9,10);
}
#define init_3_9_11_6
void FASTCALL call_3_9_11_6(SH2_struct* sh)
{
  SH2cmphi(sh,9,11);
}
#define init_3_9_12_6
void FASTCALL call_3_9_12_6(SH2_struct* sh)
{
  SH2cmphi(sh,9,12);
}
#define init_3_9_13_6
void FASTCALL call_3_9_13_6(SH2_struct* sh)
{
  SH2cmphi(sh,9,13);
}
#define init_3_9_14_6
void FASTCALL call_3_9_14_6(SH2_struct* sh)
{
  SH2cmphi(sh,9,14);
}
#define init_3_9_15_6
void FASTCALL call_3_9_15_6(SH2_struct* sh)
{
  SH2cmphi(sh,9,15);
}
#define init_3_10_0_6
void FASTCALL call_3_10_0_6(SH2_struct* sh)
{
  SH2cmphi(sh,10,0);
}
#define init_3_10_1_6
void FASTCALL call_3_10_1_6(SH2_struct* sh)
{
  SH2cmphi(sh,10,1);
}
#define init_3_10_2_6
void FASTCALL call_3_10_2_6(SH2_struct* sh)
{
  SH2cmphi(sh,10,2);
}
#define init_3_10_3_6
void FASTCALL call_3_10_3_6(SH2_struct* sh)
{
  SH2cmphi(sh,10,3);
}
#define init_3_10_4_6
void FASTCALL call_3_10_4_6(SH2_struct* sh)
{
  SH2cmphi(sh,10,4);
}
#define init_3_10_5_6
void FASTCALL call_3_10_5_6(SH2_struct* sh)
{
  SH2cmphi(sh,10,5);
}
#define init_3_10_6_6
void FASTCALL call_3_10_6_6(SH2_struct* sh)
{
  SH2cmphi(sh,10,6);
}
#define init_3_10_7_6
void FASTCALL call_3_10_7_6(SH2_struct* sh)
{
  SH2cmphi(sh,10,7);
}
#define init_3_10_8_6
void FASTCALL call_3_10_8_6(SH2_struct* sh)
{
  SH2cmphi(sh,10,8);
}
#define init_3_10_9_6
void FASTCALL call_3_10_9_6(SH2_struct* sh)
{
  SH2cmphi(sh,10,9);
}
#define init_3_10_10_6
void FASTCALL call_3_10_10_6(SH2_struct* sh)
{
  SH2cmphi(sh,10,10);
}
#define init_3_10_11_6
void FASTCALL call_3_10_11_6(SH2_struct* sh)
{
  SH2cmphi(sh,10,11);
}
#define init_3_10_12_6
void FASTCALL call_3_10_12_6(SH2_struct* sh)
{
  SH2cmphi(sh,10,12);
}
#define init_3_10_13_6
void FASTCALL call_3_10_13_6(SH2_struct* sh)
{
  SH2cmphi(sh,10,13);
}
#define init_3_10_14_6
void FASTCALL call_3_10_14_6(SH2_struct* sh)
{
  SH2cmphi(sh,10,14);
}
#define init_3_10_15_6
void FASTCALL call_3_10_15_6(SH2_struct* sh)
{
  SH2cmphi(sh,10,15);
}
#define init_3_11_0_6
void FASTCALL call_3_11_0_6(SH2_struct* sh)
{
  SH2cmphi(sh,11,0);
}
#define init_3_11_1_6
void FASTCALL call_3_11_1_6(SH2_struct* sh)
{
  SH2cmphi(sh,11,1);
}
#define init_3_11_2_6
void FASTCALL call_3_11_2_6(SH2_struct* sh)
{
  SH2cmphi(sh,11,2);
}
#define init_3_11_3_6
void FASTCALL call_3_11_3_6(SH2_struct* sh)
{
  SH2cmphi(sh,11,3);
}
#define init_3_11_4_6
void FASTCALL call_3_11_4_6(SH2_struct* sh)
{
  SH2cmphi(sh,11,4);
}
#define init_3_11_5_6
void FASTCALL call_3_11_5_6(SH2_struct* sh)
{
  SH2cmphi(sh,11,5);
}
#define init_3_11_6_6
void FASTCALL call_3_11_6_6(SH2_struct* sh)
{
  SH2cmphi(sh,11,6);
}
#define init_3_11_7_6
void FASTCALL call_3_11_7_6(SH2_struct* sh)
{
  SH2cmphi(sh,11,7);
}
#define init_3_11_8_6
void FASTCALL call_3_11_8_6(SH2_struct* sh)
{
  SH2cmphi(sh,11,8);
}
#define init_3_11_9_6
void FASTCALL call_3_11_9_6(SH2_struct* sh)
{
  SH2cmphi(sh,11,9);
}
#define init_3_11_10_6
void FASTCALL call_3_11_10_6(SH2_struct* sh)
{
  SH2cmphi(sh,11,10);
}
#define init_3_11_11_6
void FASTCALL call_3_11_11_6(SH2_struct* sh)
{
  SH2cmphi(sh,11,11);
}
#define init_3_11_12_6
void FASTCALL call_3_11_12_6(SH2_struct* sh)
{
  SH2cmphi(sh,11,12);
}
#define init_3_11_13_6
void FASTCALL call_3_11_13_6(SH2_struct* sh)
{
  SH2cmphi(sh,11,13);
}
#define init_3_11_14_6
void FASTCALL call_3_11_14_6(SH2_struct* sh)
{
  SH2cmphi(sh,11,14);
}
#define init_3_11_15_6
void FASTCALL call_3_11_15_6(SH2_struct* sh)
{
  SH2cmphi(sh,11,15);
}
#define init_3_12_0_6
void FASTCALL call_3_12_0_6(SH2_struct* sh)
{
  SH2cmphi(sh,12,0);
}
#define init_3_12_1_6
void FASTCALL call_3_12_1_6(SH2_struct* sh)
{
  SH2cmphi(sh,12,1);
}
#define init_3_12_2_6
void FASTCALL call_3_12_2_6(SH2_struct* sh)
{
  SH2cmphi(sh,12,2);
}
#define init_3_12_3_6
void FASTCALL call_3_12_3_6(SH2_struct* sh)
{
  SH2cmphi(sh,12,3);
}
#define init_3_12_4_6
void FASTCALL call_3_12_4_6(SH2_struct* sh)
{
  SH2cmphi(sh,12,4);
}
#define init_3_12_5_6
void FASTCALL call_3_12_5_6(SH2_struct* sh)
{
  SH2cmphi(sh,12,5);
}
#define init_3_12_6_6
void FASTCALL call_3_12_6_6(SH2_struct* sh)
{
  SH2cmphi(sh,12,6);
}
#define init_3_12_7_6
void FASTCALL call_3_12_7_6(SH2_struct* sh)
{
  SH2cmphi(sh,12,7);
}
#define init_3_12_8_6
void FASTCALL call_3_12_8_6(SH2_struct* sh)
{
  SH2cmphi(sh,12,8);
}
#define init_3_12_9_6
void FASTCALL call_3_12_9_6(SH2_struct* sh)
{
  SH2cmphi(sh,12,9);
}
#define init_3_12_10_6
void FASTCALL call_3_12_10_6(SH2_struct* sh)
{
  SH2cmphi(sh,12,10);
}
#define init_3_12_11_6
void FASTCALL call_3_12_11_6(SH2_struct* sh)
{
  SH2cmphi(sh,12,11);
}
#define init_3_12_12_6
void FASTCALL call_3_12_12_6(SH2_struct* sh)
{
  SH2cmphi(sh,12,12);
}
#define init_3_12_13_6
void FASTCALL call_3_12_13_6(SH2_struct* sh)
{
  SH2cmphi(sh,12,13);
}
#define init_3_12_14_6
void FASTCALL call_3_12_14_6(SH2_struct* sh)
{
  SH2cmphi(sh,12,14);
}
#define init_3_12_15_6
void FASTCALL call_3_12_15_6(SH2_struct* sh)
{
  SH2cmphi(sh,12,15);
}
#define init_3_13_0_6
void FASTCALL call_3_13_0_6(SH2_struct* sh)
{
  SH2cmphi(sh,13,0);
}
#define init_3_13_1_6
void FASTCALL call_3_13_1_6(SH2_struct* sh)
{
  SH2cmphi(sh,13,1);
}
#define init_3_13_2_6
void FASTCALL call_3_13_2_6(SH2_struct* sh)
{
  SH2cmphi(sh,13,2);
}
#define init_3_13_3_6
void FASTCALL call_3_13_3_6(SH2_struct* sh)
{
  SH2cmphi(sh,13,3);
}
#define init_3_13_4_6
void FASTCALL call_3_13_4_6(SH2_struct* sh)
{
  SH2cmphi(sh,13,4);
}
#define init_3_13_5_6
void FASTCALL call_3_13_5_6(SH2_struct* sh)
{
  SH2cmphi(sh,13,5);
}
#define init_3_13_6_6
void FASTCALL call_3_13_6_6(SH2_struct* sh)
{
  SH2cmphi(sh,13,6);
}
#define init_3_13_7_6
void FASTCALL call_3_13_7_6(SH2_struct* sh)
{
  SH2cmphi(sh,13,7);
}
#define init_3_13_8_6
void FASTCALL call_3_13_8_6(SH2_struct* sh)
{
  SH2cmphi(sh,13,8);
}
#define init_3_13_9_6
void FASTCALL call_3_13_9_6(SH2_struct* sh)
{
  SH2cmphi(sh,13,9);
}
#define init_3_13_10_6
void FASTCALL call_3_13_10_6(SH2_struct* sh)
{
  SH2cmphi(sh,13,10);
}
#define init_3_13_11_6
void FASTCALL call_3_13_11_6(SH2_struct* sh)
{
  SH2cmphi(sh,13,11);
}
#define init_3_13_12_6
void FASTCALL call_3_13_12_6(SH2_struct* sh)
{
  SH2cmphi(sh,13,12);
}
#define init_3_13_13_6
void FASTCALL call_3_13_13_6(SH2_struct* sh)
{
  SH2cmphi(sh,13,13);
}
#define init_3_13_14_6
void FASTCALL call_3_13_14_6(SH2_struct* sh)
{
  SH2cmphi(sh,13,14);
}
#define init_3_13_15_6
void FASTCALL call_3_13_15_6(SH2_struct* sh)
{
  SH2cmphi(sh,13,15);
}
#define init_3_14_0_6
void FASTCALL call_3_14_0_6(SH2_struct* sh)
{
  SH2cmphi(sh,14,0);
}
#define init_3_14_1_6
void FASTCALL call_3_14_1_6(SH2_struct* sh)
{
  SH2cmphi(sh,14,1);
}
#define init_3_14_2_6
void FASTCALL call_3_14_2_6(SH2_struct* sh)
{
  SH2cmphi(sh,14,2);
}
#define init_3_14_3_6
void FASTCALL call_3_14_3_6(SH2_struct* sh)
{
  SH2cmphi(sh,14,3);
}
#define init_3_14_4_6
void FASTCALL call_3_14_4_6(SH2_struct* sh)
{
  SH2cmphi(sh,14,4);
}
#define init_3_14_5_6
void FASTCALL call_3_14_5_6(SH2_struct* sh)
{
  SH2cmphi(sh,14,5);
}
#define init_3_14_6_6
void FASTCALL call_3_14_6_6(SH2_struct* sh)
{
  SH2cmphi(sh,14,6);
}
#define init_3_14_7_6
void FASTCALL call_3_14_7_6(SH2_struct* sh)
{
  SH2cmphi(sh,14,7);
}
#define init_3_14_8_6
void FASTCALL call_3_14_8_6(SH2_struct* sh)
{
  SH2cmphi(sh,14,8);
}
#define init_3_14_9_6
void FASTCALL call_3_14_9_6(SH2_struct* sh)
{
  SH2cmphi(sh,14,9);
}
#define init_3_14_10_6
void FASTCALL call_3_14_10_6(SH2_struct* sh)
{
  SH2cmphi(sh,14,10);
}
#define init_3_14_11_6
void FASTCALL call_3_14_11_6(SH2_struct* sh)
{
  SH2cmphi(sh,14,11);
}
#define init_3_14_12_6
void FASTCALL call_3_14_12_6(SH2_struct* sh)
{
  SH2cmphi(sh,14,12);
}
#define init_3_14_13_6
void FASTCALL call_3_14_13_6(SH2_struct* sh)
{
  SH2cmphi(sh,14,13);
}
#define init_3_14_14_6
void FASTCALL call_3_14_14_6(SH2_struct* sh)
{
  SH2cmphi(sh,14,14);
}
#define init_3_14_15_6
void FASTCALL call_3_14_15_6(SH2_struct* sh)
{
  SH2cmphi(sh,14,15);
}
#define init_3_15_0_6
void FASTCALL call_3_15_0_6(SH2_struct* sh)
{
  SH2cmphi(sh,15,0);
}
#define init_3_15_1_6
void FASTCALL call_3_15_1_6(SH2_struct* sh)
{
  SH2cmphi(sh,15,1);
}
#define init_3_15_2_6
void FASTCALL call_3_15_2_6(SH2_struct* sh)
{
  SH2cmphi(sh,15,2);
}
#define init_3_15_3_6
void FASTCALL call_3_15_3_6(SH2_struct* sh)
{
  SH2cmphi(sh,15,3);
}
#define init_3_15_4_6
void FASTCALL call_3_15_4_6(SH2_struct* sh)
{
  SH2cmphi(sh,15,4);
}
#define init_3_15_5_6
void FASTCALL call_3_15_5_6(SH2_struct* sh)
{
  SH2cmphi(sh,15,5);
}
#define init_3_15_6_6
void FASTCALL call_3_15_6_6(SH2_struct* sh)
{
  SH2cmphi(sh,15,6);
}
#define init_3_15_7_6
void FASTCALL call_3_15_7_6(SH2_struct* sh)
{
  SH2cmphi(sh,15,7);
}
#define init_3_15_8_6
void FASTCALL call_3_15_8_6(SH2_struct* sh)
{
  SH2cmphi(sh,15,8);
}
#define init_3_15_9_6
void FASTCALL call_3_15_9_6(SH2_struct* sh)
{
  SH2cmphi(sh,15,9);
}
#define init_3_15_10_6
void FASTCALL call_3_15_10_6(SH2_struct* sh)
{
  SH2cmphi(sh,15,10);
}
#define init_3_15_11_6
void FASTCALL call_3_15_11_6(SH2_struct* sh)
{
  SH2cmphi(sh,15,11);
}
#define init_3_15_12_6
void FASTCALL call_3_15_12_6(SH2_struct* sh)
{
  SH2cmphi(sh,15,12);
}
#define init_3_15_13_6
void FASTCALL call_3_15_13_6(SH2_struct* sh)
{
  SH2cmphi(sh,15,13);
}
#define init_3_15_14_6
void FASTCALL call_3_15_14_6(SH2_struct* sh)
{
  SH2cmphi(sh,15,14);
}
#define init_3_15_15_6
void FASTCALL call_3_15_15_6(SH2_struct* sh)
{
  SH2cmphi(sh,15,15);
}
#define init_3_0_0_7
void FASTCALL call_3_0_0_7(SH2_struct* sh)
{
  SH2cmpgt(sh,0,0);
}
#define init_3_0_1_7
void FASTCALL call_3_0_1_7(SH2_struct* sh)
{
  SH2cmpgt(sh,0,1);
}
#define init_3_0_2_7
void FASTCALL call_3_0_2_7(SH2_struct* sh)
{
  SH2cmpgt(sh,0,2);
}
#define init_3_0_3_7
void FASTCALL call_3_0_3_7(SH2_struct* sh)
{
  SH2cmpgt(sh,0,3);
}
#define init_3_0_4_7
void FASTCALL call_3_0_4_7(SH2_struct* sh)
{
  SH2cmpgt(sh,0,4);
}
#define init_3_0_5_7
void FASTCALL call_3_0_5_7(SH2_struct* sh)
{
  SH2cmpgt(sh,0,5);
}
#define init_3_0_6_7
void FASTCALL call_3_0_6_7(SH2_struct* sh)
{
  SH2cmpgt(sh,0,6);
}
#define init_3_0_7_7
void FASTCALL call_3_0_7_7(SH2_struct* sh)
{
  SH2cmpgt(sh,0,7);
}
#define init_3_0_8_7
void FASTCALL call_3_0_8_7(SH2_struct* sh)
{
  SH2cmpgt(sh,0,8);
}
#define init_3_0_9_7
void FASTCALL call_3_0_9_7(SH2_struct* sh)
{
  SH2cmpgt(sh,0,9);
}
#define init_3_0_10_7
void FASTCALL call_3_0_10_7(SH2_struct* sh)
{
  SH2cmpgt(sh,0,10);
}
#define init_3_0_11_7
void FASTCALL call_3_0_11_7(SH2_struct* sh)
{
  SH2cmpgt(sh,0,11);
}
#define init_3_0_12_7
void FASTCALL call_3_0_12_7(SH2_struct* sh)
{
  SH2cmpgt(sh,0,12);
}
#define init_3_0_13_7
void FASTCALL call_3_0_13_7(SH2_struct* sh)
{
  SH2cmpgt(sh,0,13);
}
#define init_3_0_14_7
void FASTCALL call_3_0_14_7(SH2_struct* sh)
{
  SH2cmpgt(sh,0,14);
}
#define init_3_0_15_7
void FASTCALL call_3_0_15_7(SH2_struct* sh)
{
  SH2cmpgt(sh,0,15);
}
#define init_3_1_0_7
void FASTCALL call_3_1_0_7(SH2_struct* sh)
{
  SH2cmpgt(sh,1,0);
}
#define init_3_1_1_7
void FASTCALL call_3_1_1_7(SH2_struct* sh)
{
  SH2cmpgt(sh,1,1);
}
#define init_3_1_2_7
void FASTCALL call_3_1_2_7(SH2_struct* sh)
{
  SH2cmpgt(sh,1,2);
}
#define init_3_1_3_7
void FASTCALL call_3_1_3_7(SH2_struct* sh)
{
  SH2cmpgt(sh,1,3);
}
#define init_3_1_4_7
void FASTCALL call_3_1_4_7(SH2_struct* sh)
{
  SH2cmpgt(sh,1,4);
}
#define init_3_1_5_7
void FASTCALL call_3_1_5_7(SH2_struct* sh)
{
  SH2cmpgt(sh,1,5);
}
#define init_3_1_6_7
void FASTCALL call_3_1_6_7(SH2_struct* sh)
{
  SH2cmpgt(sh,1,6);
}
#define init_3_1_7_7
void FASTCALL call_3_1_7_7(SH2_struct* sh)
{
  SH2cmpgt(sh,1,7);
}
#define init_3_1_8_7
void FASTCALL call_3_1_8_7(SH2_struct* sh)
{
  SH2cmpgt(sh,1,8);
}
#define init_3_1_9_7
void FASTCALL call_3_1_9_7(SH2_struct* sh)
{
  SH2cmpgt(sh,1,9);
}
#define init_3_1_10_7
void FASTCALL call_3_1_10_7(SH2_struct* sh)
{
  SH2cmpgt(sh,1,10);
}
#define init_3_1_11_7
void FASTCALL call_3_1_11_7(SH2_struct* sh)
{
  SH2cmpgt(sh,1,11);
}
#define init_3_1_12_7
void FASTCALL call_3_1_12_7(SH2_struct* sh)
{
  SH2cmpgt(sh,1,12);
}
#define init_3_1_13_7
void FASTCALL call_3_1_13_7(SH2_struct* sh)
{
  SH2cmpgt(sh,1,13);
}
#define init_3_1_14_7
void FASTCALL call_3_1_14_7(SH2_struct* sh)
{
  SH2cmpgt(sh,1,14);
}
#define init_3_1_15_7
void FASTCALL call_3_1_15_7(SH2_struct* sh)
{
  SH2cmpgt(sh,1,15);
}
#define init_3_2_0_7
void FASTCALL call_3_2_0_7(SH2_struct* sh)
{
  SH2cmpgt(sh,2,0);
}
#define init_3_2_1_7
void FASTCALL call_3_2_1_7(SH2_struct* sh)
{
  SH2cmpgt(sh,2,1);
}
#define init_3_2_2_7
void FASTCALL call_3_2_2_7(SH2_struct* sh)
{
  SH2cmpgt(sh,2,2);
}
#define init_3_2_3_7
void FASTCALL call_3_2_3_7(SH2_struct* sh)
{
  SH2cmpgt(sh,2,3);
}
#define init_3_2_4_7
void FASTCALL call_3_2_4_7(SH2_struct* sh)
{
  SH2cmpgt(sh,2,4);
}
#define init_3_2_5_7
void FASTCALL call_3_2_5_7(SH2_struct* sh)
{
  SH2cmpgt(sh,2,5);
}
#define init_3_2_6_7
void FASTCALL call_3_2_6_7(SH2_struct* sh)
{
  SH2cmpgt(sh,2,6);
}
#define init_3_2_7_7
void FASTCALL call_3_2_7_7(SH2_struct* sh)
{
  SH2cmpgt(sh,2,7);
}
#define init_3_2_8_7
void FASTCALL call_3_2_8_7(SH2_struct* sh)
{
  SH2cmpgt(sh,2,8);
}
#define init_3_2_9_7
void FASTCALL call_3_2_9_7(SH2_struct* sh)
{
  SH2cmpgt(sh,2,9);
}
#define init_3_2_10_7
void FASTCALL call_3_2_10_7(SH2_struct* sh)
{
  SH2cmpgt(sh,2,10);
}
#define init_3_2_11_7
void FASTCALL call_3_2_11_7(SH2_struct* sh)
{
  SH2cmpgt(sh,2,11);
}
#define init_3_2_12_7
void FASTCALL call_3_2_12_7(SH2_struct* sh)
{
  SH2cmpgt(sh,2,12);
}
#define init_3_2_13_7
void FASTCALL call_3_2_13_7(SH2_struct* sh)
{
  SH2cmpgt(sh,2,13);
}
#define init_3_2_14_7
void FASTCALL call_3_2_14_7(SH2_struct* sh)
{
  SH2cmpgt(sh,2,14);
}
#define init_3_2_15_7
void FASTCALL call_3_2_15_7(SH2_struct* sh)
{
  SH2cmpgt(sh,2,15);
}
#define init_3_3_0_7
void FASTCALL call_3_3_0_7(SH2_struct* sh)
{
  SH2cmpgt(sh,3,0);
}
#define init_3_3_1_7
void FASTCALL call_3_3_1_7(SH2_struct* sh)
{
  SH2cmpgt(sh,3,1);
}
#define init_3_3_2_7
void FASTCALL call_3_3_2_7(SH2_struct* sh)
{
  SH2cmpgt(sh,3,2);
}
#define init_3_3_3_7
void FASTCALL call_3_3_3_7(SH2_struct* sh)
{
  SH2cmpgt(sh,3,3);
}
#define init_3_3_4_7
void FASTCALL call_3_3_4_7(SH2_struct* sh)
{
  SH2cmpgt(sh,3,4);
}
#define init_3_3_5_7
void FASTCALL call_3_3_5_7(SH2_struct* sh)
{
  SH2cmpgt(sh,3,5);
}
#define init_3_3_6_7
void FASTCALL call_3_3_6_7(SH2_struct* sh)
{
  SH2cmpgt(sh,3,6);
}
#define init_3_3_7_7
void FASTCALL call_3_3_7_7(SH2_struct* sh)
{
  SH2cmpgt(sh,3,7);
}
#define init_3_3_8_7
void FASTCALL call_3_3_8_7(SH2_struct* sh)
{
  SH2cmpgt(sh,3,8);
}
#define init_3_3_9_7
void FASTCALL call_3_3_9_7(SH2_struct* sh)
{
  SH2cmpgt(sh,3,9);
}
#define init_3_3_10_7
void FASTCALL call_3_3_10_7(SH2_struct* sh)
{
  SH2cmpgt(sh,3,10);
}
#define init_3_3_11_7
void FASTCALL call_3_3_11_7(SH2_struct* sh)
{
  SH2cmpgt(sh,3,11);
}
#define init_3_3_12_7
void FASTCALL call_3_3_12_7(SH2_struct* sh)
{
  SH2cmpgt(sh,3,12);
}
#define init_3_3_13_7
void FASTCALL call_3_3_13_7(SH2_struct* sh)
{
  SH2cmpgt(sh,3,13);
}
#define init_3_3_14_7
void FASTCALL call_3_3_14_7(SH2_struct* sh)
{
  SH2cmpgt(sh,3,14);
}
#define init_3_3_15_7
void FASTCALL call_3_3_15_7(SH2_struct* sh)
{
  SH2cmpgt(sh,3,15);
}
#define init_3_4_0_7
void FASTCALL call_3_4_0_7(SH2_struct* sh)
{
  SH2cmpgt(sh,4,0);
}
#define init_3_4_1_7
void FASTCALL call_3_4_1_7(SH2_struct* sh)
{
  SH2cmpgt(sh,4,1);
}
#define init_3_4_2_7
void FASTCALL call_3_4_2_7(SH2_struct* sh)
{
  SH2cmpgt(sh,4,2);
}
#define init_3_4_3_7
void FASTCALL call_3_4_3_7(SH2_struct* sh)
{
  SH2cmpgt(sh,4,3);
}
#define init_3_4_4_7
void FASTCALL call_3_4_4_7(SH2_struct* sh)
{
  SH2cmpgt(sh,4,4);
}
#define init_3_4_5_7
void FASTCALL call_3_4_5_7(SH2_struct* sh)
{
  SH2cmpgt(sh,4,5);
}
#define init_3_4_6_7
void FASTCALL call_3_4_6_7(SH2_struct* sh)
{
  SH2cmpgt(sh,4,6);
}
#define init_3_4_7_7
void FASTCALL call_3_4_7_7(SH2_struct* sh)
{
  SH2cmpgt(sh,4,7);
}
#define init_3_4_8_7
void FASTCALL call_3_4_8_7(SH2_struct* sh)
{
  SH2cmpgt(sh,4,8);
}
#define init_3_4_9_7
void FASTCALL call_3_4_9_7(SH2_struct* sh)
{
  SH2cmpgt(sh,4,9);
}
#define init_3_4_10_7
void FASTCALL call_3_4_10_7(SH2_struct* sh)
{
  SH2cmpgt(sh,4,10);
}
#define init_3_4_11_7
void FASTCALL call_3_4_11_7(SH2_struct* sh)
{
  SH2cmpgt(sh,4,11);
}
#define init_3_4_12_7
void FASTCALL call_3_4_12_7(SH2_struct* sh)
{
  SH2cmpgt(sh,4,12);
}
#define init_3_4_13_7
void FASTCALL call_3_4_13_7(SH2_struct* sh)
{
  SH2cmpgt(sh,4,13);
}
#define init_3_4_14_7
void FASTCALL call_3_4_14_7(SH2_struct* sh)
{
  SH2cmpgt(sh,4,14);
}
#define init_3_4_15_7
void FASTCALL call_3_4_15_7(SH2_struct* sh)
{
  SH2cmpgt(sh,4,15);
}
#define init_3_5_0_7
void FASTCALL call_3_5_0_7(SH2_struct* sh)
{
  SH2cmpgt(sh,5,0);
}
#define init_3_5_1_7
void FASTCALL call_3_5_1_7(SH2_struct* sh)
{
  SH2cmpgt(sh,5,1);
}
#define init_3_5_2_7
void FASTCALL call_3_5_2_7(SH2_struct* sh)
{
  SH2cmpgt(sh,5,2);
}
#define init_3_5_3_7
void FASTCALL call_3_5_3_7(SH2_struct* sh)
{
  SH2cmpgt(sh,5,3);
}
#define init_3_5_4_7
void FASTCALL call_3_5_4_7(SH2_struct* sh)
{
  SH2cmpgt(sh,5,4);
}
#define init_3_5_5_7
void FASTCALL call_3_5_5_7(SH2_struct* sh)
{
  SH2cmpgt(sh,5,5);
}
#define init_3_5_6_7
void FASTCALL call_3_5_6_7(SH2_struct* sh)
{
  SH2cmpgt(sh,5,6);
}
#define init_3_5_7_7
void FASTCALL call_3_5_7_7(SH2_struct* sh)
{
  SH2cmpgt(sh,5,7);
}
#define init_3_5_8_7
void FASTCALL call_3_5_8_7(SH2_struct* sh)
{
  SH2cmpgt(sh,5,8);
}
#define init_3_5_9_7
void FASTCALL call_3_5_9_7(SH2_struct* sh)
{
  SH2cmpgt(sh,5,9);
}
#define init_3_5_10_7
void FASTCALL call_3_5_10_7(SH2_struct* sh)
{
  SH2cmpgt(sh,5,10);
}
#define init_3_5_11_7
void FASTCALL call_3_5_11_7(SH2_struct* sh)
{
  SH2cmpgt(sh,5,11);
}
#define init_3_5_12_7
void FASTCALL call_3_5_12_7(SH2_struct* sh)
{
  SH2cmpgt(sh,5,12);
}
#define init_3_5_13_7
void FASTCALL call_3_5_13_7(SH2_struct* sh)
{
  SH2cmpgt(sh,5,13);
}
#define init_3_5_14_7
void FASTCALL call_3_5_14_7(SH2_struct* sh)
{
  SH2cmpgt(sh,5,14);
}
#define init_3_5_15_7
void FASTCALL call_3_5_15_7(SH2_struct* sh)
{
  SH2cmpgt(sh,5,15);
}
#define init_3_6_0_7
void FASTCALL call_3_6_0_7(SH2_struct* sh)
{
  SH2cmpgt(sh,6,0);
}
#define init_3_6_1_7
void FASTCALL call_3_6_1_7(SH2_struct* sh)
{
  SH2cmpgt(sh,6,1);
}
#define init_3_6_2_7
void FASTCALL call_3_6_2_7(SH2_struct* sh)
{
  SH2cmpgt(sh,6,2);
}
#define init_3_6_3_7
void FASTCALL call_3_6_3_7(SH2_struct* sh)
{
  SH2cmpgt(sh,6,3);
}
#define init_3_6_4_7
void FASTCALL call_3_6_4_7(SH2_struct* sh)
{
  SH2cmpgt(sh,6,4);
}
#define init_3_6_5_7
void FASTCALL call_3_6_5_7(SH2_struct* sh)
{
  SH2cmpgt(sh,6,5);
}
#define init_3_6_6_7
void FASTCALL call_3_6_6_7(SH2_struct* sh)
{
  SH2cmpgt(sh,6,6);
}
#define init_3_6_7_7
void FASTCALL call_3_6_7_7(SH2_struct* sh)
{
  SH2cmpgt(sh,6,7);
}
#define init_3_6_8_7
void FASTCALL call_3_6_8_7(SH2_struct* sh)
{
  SH2cmpgt(sh,6,8);
}
#define init_3_6_9_7
void FASTCALL call_3_6_9_7(SH2_struct* sh)
{
  SH2cmpgt(sh,6,9);
}
#define init_3_6_10_7
void FASTCALL call_3_6_10_7(SH2_struct* sh)
{
  SH2cmpgt(sh,6,10);
}
#define init_3_6_11_7
void FASTCALL call_3_6_11_7(SH2_struct* sh)
{
  SH2cmpgt(sh,6,11);
}
#define init_3_6_12_7
void FASTCALL call_3_6_12_7(SH2_struct* sh)
{
  SH2cmpgt(sh,6,12);
}
#define init_3_6_13_7
void FASTCALL call_3_6_13_7(SH2_struct* sh)
{
  SH2cmpgt(sh,6,13);
}
#define init_3_6_14_7
void FASTCALL call_3_6_14_7(SH2_struct* sh)
{
  SH2cmpgt(sh,6,14);
}
#define init_3_6_15_7
void FASTCALL call_3_6_15_7(SH2_struct* sh)
{
  SH2cmpgt(sh,6,15);
}
#define init_3_7_0_7
void FASTCALL call_3_7_0_7(SH2_struct* sh)
{
  SH2cmpgt(sh,7,0);
}
#define init_3_7_1_7
void FASTCALL call_3_7_1_7(SH2_struct* sh)
{
  SH2cmpgt(sh,7,1);
}
#define init_3_7_2_7
void FASTCALL call_3_7_2_7(SH2_struct* sh)
{
  SH2cmpgt(sh,7,2);
}
#define init_3_7_3_7
void FASTCALL call_3_7_3_7(SH2_struct* sh)
{
  SH2cmpgt(sh,7,3);
}
#define init_3_7_4_7
void FASTCALL call_3_7_4_7(SH2_struct* sh)
{
  SH2cmpgt(sh,7,4);
}
#define init_3_7_5_7
void FASTCALL call_3_7_5_7(SH2_struct* sh)
{
  SH2cmpgt(sh,7,5);
}
#define init_3_7_6_7
void FASTCALL call_3_7_6_7(SH2_struct* sh)
{
  SH2cmpgt(sh,7,6);
}
#define init_3_7_7_7
void FASTCALL call_3_7_7_7(SH2_struct* sh)
{
  SH2cmpgt(sh,7,7);
}
#define init_3_7_8_7
void FASTCALL call_3_7_8_7(SH2_struct* sh)
{
  SH2cmpgt(sh,7,8);
}
#define init_3_7_9_7
void FASTCALL call_3_7_9_7(SH2_struct* sh)
{
  SH2cmpgt(sh,7,9);
}
#define init_3_7_10_7
void FASTCALL call_3_7_10_7(SH2_struct* sh)
{
  SH2cmpgt(sh,7,10);
}
#define init_3_7_11_7
void FASTCALL call_3_7_11_7(SH2_struct* sh)
{
  SH2cmpgt(sh,7,11);
}
#define init_3_7_12_7
void FASTCALL call_3_7_12_7(SH2_struct* sh)
{
  SH2cmpgt(sh,7,12);
}
#define init_3_7_13_7
void FASTCALL call_3_7_13_7(SH2_struct* sh)
{
  SH2cmpgt(sh,7,13);
}
#define init_3_7_14_7
void FASTCALL call_3_7_14_7(SH2_struct* sh)
{
  SH2cmpgt(sh,7,14);
}
#define init_3_7_15_7
void FASTCALL call_3_7_15_7(SH2_struct* sh)
{
  SH2cmpgt(sh,7,15);
}
#define init_3_8_0_7
void FASTCALL call_3_8_0_7(SH2_struct* sh)
{
  SH2cmpgt(sh,8,0);
}
#define init_3_8_1_7
void FASTCALL call_3_8_1_7(SH2_struct* sh)
{
  SH2cmpgt(sh,8,1);
}
#define init_3_8_2_7
void FASTCALL call_3_8_2_7(SH2_struct* sh)
{
  SH2cmpgt(sh,8,2);
}
#define init_3_8_3_7
void FASTCALL call_3_8_3_7(SH2_struct* sh)
{
  SH2cmpgt(sh,8,3);
}
#define init_3_8_4_7
void FASTCALL call_3_8_4_7(SH2_struct* sh)
{
  SH2cmpgt(sh,8,4);
}
#define init_3_8_5_7
void FASTCALL call_3_8_5_7(SH2_struct* sh)
{
  SH2cmpgt(sh,8,5);
}
#define init_3_8_6_7
void FASTCALL call_3_8_6_7(SH2_struct* sh)
{
  SH2cmpgt(sh,8,6);
}
#define init_3_8_7_7
void FASTCALL call_3_8_7_7(SH2_struct* sh)
{
  SH2cmpgt(sh,8,7);
}
#define init_3_8_8_7
void FASTCALL call_3_8_8_7(SH2_struct* sh)
{
  SH2cmpgt(sh,8,8);
}
#define init_3_8_9_7
void FASTCALL call_3_8_9_7(SH2_struct* sh)
{
  SH2cmpgt(sh,8,9);
}
#define init_3_8_10_7
void FASTCALL call_3_8_10_7(SH2_struct* sh)
{
  SH2cmpgt(sh,8,10);
}
#define init_3_8_11_7
void FASTCALL call_3_8_11_7(SH2_struct* sh)
{
  SH2cmpgt(sh,8,11);
}
#define init_3_8_12_7
void FASTCALL call_3_8_12_7(SH2_struct* sh)
{
  SH2cmpgt(sh,8,12);
}
#define init_3_8_13_7
void FASTCALL call_3_8_13_7(SH2_struct* sh)
{
  SH2cmpgt(sh,8,13);
}
#define init_3_8_14_7
void FASTCALL call_3_8_14_7(SH2_struct* sh)
{
  SH2cmpgt(sh,8,14);
}
#define init_3_8_15_7
void FASTCALL call_3_8_15_7(SH2_struct* sh)
{
  SH2cmpgt(sh,8,15);
}
#define init_3_9_0_7
void FASTCALL call_3_9_0_7(SH2_struct* sh)
{
  SH2cmpgt(sh,9,0);
}
#define init_3_9_1_7
void FASTCALL call_3_9_1_7(SH2_struct* sh)
{
  SH2cmpgt(sh,9,1);
}
#define init_3_9_2_7
void FASTCALL call_3_9_2_7(SH2_struct* sh)
{
  SH2cmpgt(sh,9,2);
}
#define init_3_9_3_7
void FASTCALL call_3_9_3_7(SH2_struct* sh)
{
  SH2cmpgt(sh,9,3);
}
#define init_3_9_4_7
void FASTCALL call_3_9_4_7(SH2_struct* sh)
{
  SH2cmpgt(sh,9,4);
}
#define init_3_9_5_7
void FASTCALL call_3_9_5_7(SH2_struct* sh)
{
  SH2cmpgt(sh,9,5);
}
#define init_3_9_6_7
void FASTCALL call_3_9_6_7(SH2_struct* sh)
{
  SH2cmpgt(sh,9,6);
}
#define init_3_9_7_7
void FASTCALL call_3_9_7_7(SH2_struct* sh)
{
  SH2cmpgt(sh,9,7);
}
#define init_3_9_8_7
void FASTCALL call_3_9_8_7(SH2_struct* sh)
{
  SH2cmpgt(sh,9,8);
}
#define init_3_9_9_7
void FASTCALL call_3_9_9_7(SH2_struct* sh)
{
  SH2cmpgt(sh,9,9);
}
#define init_3_9_10_7
void FASTCALL call_3_9_10_7(SH2_struct* sh)
{
  SH2cmpgt(sh,9,10);
}
#define init_3_9_11_7
void FASTCALL call_3_9_11_7(SH2_struct* sh)
{
  SH2cmpgt(sh,9,11);
}
#define init_3_9_12_7
void FASTCALL call_3_9_12_7(SH2_struct* sh)
{
  SH2cmpgt(sh,9,12);
}
#define init_3_9_13_7
void FASTCALL call_3_9_13_7(SH2_struct* sh)
{
  SH2cmpgt(sh,9,13);
}
#define init_3_9_14_7
void FASTCALL call_3_9_14_7(SH2_struct* sh)
{
  SH2cmpgt(sh,9,14);
}
#define init_3_9_15_7
void FASTCALL call_3_9_15_7(SH2_struct* sh)
{
  SH2cmpgt(sh,9,15);
}
#define init_3_10_0_7
void FASTCALL call_3_10_0_7(SH2_struct* sh)
{
  SH2cmpgt(sh,10,0);
}
#define init_3_10_1_7
void FASTCALL call_3_10_1_7(SH2_struct* sh)
{
  SH2cmpgt(sh,10,1);
}
#define init_3_10_2_7
void FASTCALL call_3_10_2_7(SH2_struct* sh)
{
  SH2cmpgt(sh,10,2);
}
#define init_3_10_3_7
void FASTCALL call_3_10_3_7(SH2_struct* sh)
{
  SH2cmpgt(sh,10,3);
}
#define init_3_10_4_7
void FASTCALL call_3_10_4_7(SH2_struct* sh)
{
  SH2cmpgt(sh,10,4);
}
#define init_3_10_5_7
void FASTCALL call_3_10_5_7(SH2_struct* sh)
{
  SH2cmpgt(sh,10,5);
}
#define init_3_10_6_7
void FASTCALL call_3_10_6_7(SH2_struct* sh)
{
  SH2cmpgt(sh,10,6);
}
#define init_3_10_7_7
void FASTCALL call_3_10_7_7(SH2_struct* sh)
{
  SH2cmpgt(sh,10,7);
}
#define init_3_10_8_7
void FASTCALL call_3_10_8_7(SH2_struct* sh)
{
  SH2cmpgt(sh,10,8);
}
#define init_3_10_9_7
void FASTCALL call_3_10_9_7(SH2_struct* sh)
{
  SH2cmpgt(sh,10,9);
}
#define init_3_10_10_7
void FASTCALL call_3_10_10_7(SH2_struct* sh)
{
  SH2cmpgt(sh,10,10);
}
#define init_3_10_11_7
void FASTCALL call_3_10_11_7(SH2_struct* sh)
{
  SH2cmpgt(sh,10,11);
}
#define init_3_10_12_7
void FASTCALL call_3_10_12_7(SH2_struct* sh)
{
  SH2cmpgt(sh,10,12);
}
#define init_3_10_13_7
void FASTCALL call_3_10_13_7(SH2_struct* sh)
{
  SH2cmpgt(sh,10,13);
}
#define init_3_10_14_7
void FASTCALL call_3_10_14_7(SH2_struct* sh)
{
  SH2cmpgt(sh,10,14);
}
#define init_3_10_15_7
void FASTCALL call_3_10_15_7(SH2_struct* sh)
{
  SH2cmpgt(sh,10,15);
}
#define init_3_11_0_7
void FASTCALL call_3_11_0_7(SH2_struct* sh)
{
  SH2cmpgt(sh,11,0);
}
#define init_3_11_1_7
void FASTCALL call_3_11_1_7(SH2_struct* sh)
{
  SH2cmpgt(sh,11,1);
}
#define init_3_11_2_7
void FASTCALL call_3_11_2_7(SH2_struct* sh)
{
  SH2cmpgt(sh,11,2);
}
#define init_3_11_3_7
void FASTCALL call_3_11_3_7(SH2_struct* sh)
{
  SH2cmpgt(sh,11,3);
}
#define init_3_11_4_7
void FASTCALL call_3_11_4_7(SH2_struct* sh)
{
  SH2cmpgt(sh,11,4);
}
#define init_3_11_5_7
void FASTCALL call_3_11_5_7(SH2_struct* sh)
{
  SH2cmpgt(sh,11,5);
}
#define init_3_11_6_7
void FASTCALL call_3_11_6_7(SH2_struct* sh)
{
  SH2cmpgt(sh,11,6);
}
#define init_3_11_7_7
void FASTCALL call_3_11_7_7(SH2_struct* sh)
{
  SH2cmpgt(sh,11,7);
}
#define init_3_11_8_7
void FASTCALL call_3_11_8_7(SH2_struct* sh)
{
  SH2cmpgt(sh,11,8);
}
#define init_3_11_9_7
void FASTCALL call_3_11_9_7(SH2_struct* sh)
{
  SH2cmpgt(sh,11,9);
}
#define init_3_11_10_7
void FASTCALL call_3_11_10_7(SH2_struct* sh)
{
  SH2cmpgt(sh,11,10);
}
#define init_3_11_11_7
void FASTCALL call_3_11_11_7(SH2_struct* sh)
{
  SH2cmpgt(sh,11,11);
}
#define init_3_11_12_7
void FASTCALL call_3_11_12_7(SH2_struct* sh)
{
  SH2cmpgt(sh,11,12);
}
#define init_3_11_13_7
void FASTCALL call_3_11_13_7(SH2_struct* sh)
{
  SH2cmpgt(sh,11,13);
}
#define init_3_11_14_7
void FASTCALL call_3_11_14_7(SH2_struct* sh)
{
  SH2cmpgt(sh,11,14);
}
#define init_3_11_15_7
void FASTCALL call_3_11_15_7(SH2_struct* sh)
{
  SH2cmpgt(sh,11,15);
}
#define init_3_12_0_7
void FASTCALL call_3_12_0_7(SH2_struct* sh)
{
  SH2cmpgt(sh,12,0);
}
#define init_3_12_1_7
void FASTCALL call_3_12_1_7(SH2_struct* sh)
{
  SH2cmpgt(sh,12,1);
}
#define init_3_12_2_7
void FASTCALL call_3_12_2_7(SH2_struct* sh)
{
  SH2cmpgt(sh,12,2);
}
#define init_3_12_3_7
void FASTCALL call_3_12_3_7(SH2_struct* sh)
{
  SH2cmpgt(sh,12,3);
}
#define init_3_12_4_7
void FASTCALL call_3_12_4_7(SH2_struct* sh)
{
  SH2cmpgt(sh,12,4);
}
#define init_3_12_5_7
void FASTCALL call_3_12_5_7(SH2_struct* sh)
{
  SH2cmpgt(sh,12,5);
}
#define init_3_12_6_7
void FASTCALL call_3_12_6_7(SH2_struct* sh)
{
  SH2cmpgt(sh,12,6);
}
#define init_3_12_7_7
void FASTCALL call_3_12_7_7(SH2_struct* sh)
{
  SH2cmpgt(sh,12,7);
}
#define init_3_12_8_7
void FASTCALL call_3_12_8_7(SH2_struct* sh)
{
  SH2cmpgt(sh,12,8);
}
#define init_3_12_9_7
void FASTCALL call_3_12_9_7(SH2_struct* sh)
{
  SH2cmpgt(sh,12,9);
}
#define init_3_12_10_7
void FASTCALL call_3_12_10_7(SH2_struct* sh)
{
  SH2cmpgt(sh,12,10);
}
#define init_3_12_11_7
void FASTCALL call_3_12_11_7(SH2_struct* sh)
{
  SH2cmpgt(sh,12,11);
}
#define init_3_12_12_7
void FASTCALL call_3_12_12_7(SH2_struct* sh)
{
  SH2cmpgt(sh,12,12);
}
#define init_3_12_13_7
void FASTCALL call_3_12_13_7(SH2_struct* sh)
{
  SH2cmpgt(sh,12,13);
}
#define init_3_12_14_7
void FASTCALL call_3_12_14_7(SH2_struct* sh)
{
  SH2cmpgt(sh,12,14);
}
#define init_3_12_15_7
void FASTCALL call_3_12_15_7(SH2_struct* sh)
{
  SH2cmpgt(sh,12,15);
}
#define init_3_13_0_7
void FASTCALL call_3_13_0_7(SH2_struct* sh)
{
  SH2cmpgt(sh,13,0);
}
#define init_3_13_1_7
void FASTCALL call_3_13_1_7(SH2_struct* sh)
{
  SH2cmpgt(sh,13,1);
}
#define init_3_13_2_7
void FASTCALL call_3_13_2_7(SH2_struct* sh)
{
  SH2cmpgt(sh,13,2);
}
#define init_3_13_3_7
void FASTCALL call_3_13_3_7(SH2_struct* sh)
{
  SH2cmpgt(sh,13,3);
}
#define init_3_13_4_7
void FASTCALL call_3_13_4_7(SH2_struct* sh)
{
  SH2cmpgt(sh,13,4);
}
#define init_3_13_5_7
void FASTCALL call_3_13_5_7(SH2_struct* sh)
{
  SH2cmpgt(sh,13,5);
}
#define init_3_13_6_7
void FASTCALL call_3_13_6_7(SH2_struct* sh)
{
  SH2cmpgt(sh,13,6);
}
#define init_3_13_7_7
void FASTCALL call_3_13_7_7(SH2_struct* sh)
{
  SH2cmpgt(sh,13,7);
}
#define init_3_13_8_7
void FASTCALL call_3_13_8_7(SH2_struct* sh)
{
  SH2cmpgt(sh,13,8);
}
#define init_3_13_9_7
void FASTCALL call_3_13_9_7(SH2_struct* sh)
{
  SH2cmpgt(sh,13,9);
}
#define init_3_13_10_7
void FASTCALL call_3_13_10_7(SH2_struct* sh)
{
  SH2cmpgt(sh,13,10);
}
#define init_3_13_11_7
void FASTCALL call_3_13_11_7(SH2_struct* sh)
{
  SH2cmpgt(sh,13,11);
}
#define init_3_13_12_7
void FASTCALL call_3_13_12_7(SH2_struct* sh)
{
  SH2cmpgt(sh,13,12);
}
#define init_3_13_13_7
void FASTCALL call_3_13_13_7(SH2_struct* sh)
{
  SH2cmpgt(sh,13,13);
}
#define init_3_13_14_7
void FASTCALL call_3_13_14_7(SH2_struct* sh)
{
  SH2cmpgt(sh,13,14);
}
#define init_3_13_15_7
void FASTCALL call_3_13_15_7(SH2_struct* sh)
{
  SH2cmpgt(sh,13,15);
}
#define init_3_14_0_7
void FASTCALL call_3_14_0_7(SH2_struct* sh)
{
  SH2cmpgt(sh,14,0);
}
#define init_3_14_1_7
void FASTCALL call_3_14_1_7(SH2_struct* sh)
{
  SH2cmpgt(sh,14,1);
}
#define init_3_14_2_7
void FASTCALL call_3_14_2_7(SH2_struct* sh)
{
  SH2cmpgt(sh,14,2);
}
#define init_3_14_3_7
void FASTCALL call_3_14_3_7(SH2_struct* sh)
{
  SH2cmpgt(sh,14,3);
}
#define init_3_14_4_7
void FASTCALL call_3_14_4_7(SH2_struct* sh)
{
  SH2cmpgt(sh,14,4);
}
#define init_3_14_5_7
void FASTCALL call_3_14_5_7(SH2_struct* sh)
{
  SH2cmpgt(sh,14,5);
}
#define init_3_14_6_7
void FASTCALL call_3_14_6_7(SH2_struct* sh)
{
  SH2cmpgt(sh,14,6);
}
#define init_3_14_7_7
void FASTCALL call_3_14_7_7(SH2_struct* sh)
{
  SH2cmpgt(sh,14,7);
}
#define init_3_14_8_7
void FASTCALL call_3_14_8_7(SH2_struct* sh)
{
  SH2cmpgt(sh,14,8);
}
#define init_3_14_9_7
void FASTCALL call_3_14_9_7(SH2_struct* sh)
{
  SH2cmpgt(sh,14,9);
}
#define init_3_14_10_7
void FASTCALL call_3_14_10_7(SH2_struct* sh)
{
  SH2cmpgt(sh,14,10);
}
#define init_3_14_11_7
void FASTCALL call_3_14_11_7(SH2_struct* sh)
{
  SH2cmpgt(sh,14,11);
}
#define init_3_14_12_7
void FASTCALL call_3_14_12_7(SH2_struct* sh)
{
  SH2cmpgt(sh,14,12);
}
#define init_3_14_13_7
void FASTCALL call_3_14_13_7(SH2_struct* sh)
{
  SH2cmpgt(sh,14,13);
}
#define init_3_14_14_7
void FASTCALL call_3_14_14_7(SH2_struct* sh)
{
  SH2cmpgt(sh,14,14);
}
#define init_3_14_15_7
void FASTCALL call_3_14_15_7(SH2_struct* sh)
{
  SH2cmpgt(sh,14,15);
}
#define init_3_15_0_7
void FASTCALL call_3_15_0_7(SH2_struct* sh)
{
  SH2cmpgt(sh,15,0);
}
#define init_3_15_1_7
void FASTCALL call_3_15_1_7(SH2_struct* sh)
{
  SH2cmpgt(sh,15,1);
}
#define init_3_15_2_7
void FASTCALL call_3_15_2_7(SH2_struct* sh)
{
  SH2cmpgt(sh,15,2);
}
#define init_3_15_3_7
void FASTCALL call_3_15_3_7(SH2_struct* sh)
{
  SH2cmpgt(sh,15,3);
}
#define init_3_15_4_7
void FASTCALL call_3_15_4_7(SH2_struct* sh)
{
  SH2cmpgt(sh,15,4);
}
#define init_3_15_5_7
void FASTCALL call_3_15_5_7(SH2_struct* sh)
{
  SH2cmpgt(sh,15,5);
}
#define init_3_15_6_7
void FASTCALL call_3_15_6_7(SH2_struct* sh)
{
  SH2cmpgt(sh,15,6);
}
#define init_3_15_7_7
void FASTCALL call_3_15_7_7(SH2_struct* sh)
{
  SH2cmpgt(sh,15,7);
}
#define init_3_15_8_7
void FASTCALL call_3_15_8_7(SH2_struct* sh)
{
  SH2cmpgt(sh,15,8);
}
#define init_3_15_9_7
void FASTCALL call_3_15_9_7(SH2_struct* sh)
{
  SH2cmpgt(sh,15,9);
}
#define init_3_15_10_7
void FASTCALL call_3_15_10_7(SH2_struct* sh)
{
  SH2cmpgt(sh,15,10);
}
#define init_3_15_11_7
void FASTCALL call_3_15_11_7(SH2_struct* sh)
{
  SH2cmpgt(sh,15,11);
}
#define init_3_15_12_7
void FASTCALL call_3_15_12_7(SH2_struct* sh)
{
  SH2cmpgt(sh,15,12);
}
#define init_3_15_13_7
void FASTCALL call_3_15_13_7(SH2_struct* sh)
{
  SH2cmpgt(sh,15,13);
}
#define init_3_15_14_7
void FASTCALL call_3_15_14_7(SH2_struct* sh)
{
  SH2cmpgt(sh,15,14);
}
#define init_3_15_15_7
void FASTCALL call_3_15_15_7(SH2_struct* sh)
{
  SH2cmpgt(sh,15,15);
}
#define init_3_0_0_3
void FASTCALL call_3_0_0_3(SH2_struct* sh)
{
  SH2cmpge(sh,0,0);
}
#define init_3_0_1_3
void FASTCALL call_3_0_1_3(SH2_struct* sh)
{
  SH2cmpge(sh,0,1);
}
#define init_3_0_2_3
void FASTCALL call_3_0_2_3(SH2_struct* sh)
{
  SH2cmpge(sh,0,2);
}
#define init_3_0_3_3
void FASTCALL call_3_0_3_3(SH2_struct* sh)
{
  SH2cmpge(sh,0,3);
}
#define init_3_0_4_3
void FASTCALL call_3_0_4_3(SH2_struct* sh)
{
  SH2cmpge(sh,0,4);
}
#define init_3_0_5_3
void FASTCALL call_3_0_5_3(SH2_struct* sh)
{
  SH2cmpge(sh,0,5);
}
#define init_3_0_6_3
void FASTCALL call_3_0_6_3(SH2_struct* sh)
{
  SH2cmpge(sh,0,6);
}
#define init_3_0_7_3
void FASTCALL call_3_0_7_3(SH2_struct* sh)
{
  SH2cmpge(sh,0,7);
}
#define init_3_0_8_3
void FASTCALL call_3_0_8_3(SH2_struct* sh)
{
  SH2cmpge(sh,0,8);
}
#define init_3_0_9_3
void FASTCALL call_3_0_9_3(SH2_struct* sh)
{
  SH2cmpge(sh,0,9);
}
#define init_3_0_10_3
void FASTCALL call_3_0_10_3(SH2_struct* sh)
{
  SH2cmpge(sh,0,10);
}
#define init_3_0_11_3
void FASTCALL call_3_0_11_3(SH2_struct* sh)
{
  SH2cmpge(sh,0,11);
}
#define init_3_0_12_3
void FASTCALL call_3_0_12_3(SH2_struct* sh)
{
  SH2cmpge(sh,0,12);
}
#define init_3_0_13_3
void FASTCALL call_3_0_13_3(SH2_struct* sh)
{
  SH2cmpge(sh,0,13);
}
#define init_3_0_14_3
void FASTCALL call_3_0_14_3(SH2_struct* sh)
{
  SH2cmpge(sh,0,14);
}
#define init_3_0_15_3
void FASTCALL call_3_0_15_3(SH2_struct* sh)
{
  SH2cmpge(sh,0,15);
}
#define init_3_1_0_3
void FASTCALL call_3_1_0_3(SH2_struct* sh)
{
  SH2cmpge(sh,1,0);
}
#define init_3_1_1_3
void FASTCALL call_3_1_1_3(SH2_struct* sh)
{
  SH2cmpge(sh,1,1);
}
#define init_3_1_2_3
void FASTCALL call_3_1_2_3(SH2_struct* sh)
{
  SH2cmpge(sh,1,2);
}
#define init_3_1_3_3
void FASTCALL call_3_1_3_3(SH2_struct* sh)
{
  SH2cmpge(sh,1,3);
}
#define init_3_1_4_3
void FASTCALL call_3_1_4_3(SH2_struct* sh)
{
  SH2cmpge(sh,1,4);
}
#define init_3_1_5_3
void FASTCALL call_3_1_5_3(SH2_struct* sh)
{
  SH2cmpge(sh,1,5);
}
#define init_3_1_6_3
void FASTCALL call_3_1_6_3(SH2_struct* sh)
{
  SH2cmpge(sh,1,6);
}
#define init_3_1_7_3
void FASTCALL call_3_1_7_3(SH2_struct* sh)
{
  SH2cmpge(sh,1,7);
}
#define init_3_1_8_3
void FASTCALL call_3_1_8_3(SH2_struct* sh)
{
  SH2cmpge(sh,1,8);
}
#define init_3_1_9_3
void FASTCALL call_3_1_9_3(SH2_struct* sh)
{
  SH2cmpge(sh,1,9);
}
#define init_3_1_10_3
void FASTCALL call_3_1_10_3(SH2_struct* sh)
{
  SH2cmpge(sh,1,10);
}
#define init_3_1_11_3
void FASTCALL call_3_1_11_3(SH2_struct* sh)
{
  SH2cmpge(sh,1,11);
}
#define init_3_1_12_3
void FASTCALL call_3_1_12_3(SH2_struct* sh)
{
  SH2cmpge(sh,1,12);
}
#define init_3_1_13_3
void FASTCALL call_3_1_13_3(SH2_struct* sh)
{
  SH2cmpge(sh,1,13);
}
#define init_3_1_14_3
void FASTCALL call_3_1_14_3(SH2_struct* sh)
{
  SH2cmpge(sh,1,14);
}
#define init_3_1_15_3
void FASTCALL call_3_1_15_3(SH2_struct* sh)
{
  SH2cmpge(sh,1,15);
}
#define init_3_2_0_3
void FASTCALL call_3_2_0_3(SH2_struct* sh)
{
  SH2cmpge(sh,2,0);
}
#define init_3_2_1_3
void FASTCALL call_3_2_1_3(SH2_struct* sh)
{
  SH2cmpge(sh,2,1);
}
#define init_3_2_2_3
void FASTCALL call_3_2_2_3(SH2_struct* sh)
{
  SH2cmpge(sh,2,2);
}
#define init_3_2_3_3
void FASTCALL call_3_2_3_3(SH2_struct* sh)
{
  SH2cmpge(sh,2,3);
}
#define init_3_2_4_3
void FASTCALL call_3_2_4_3(SH2_struct* sh)
{
  SH2cmpge(sh,2,4);
}
#define init_3_2_5_3
void FASTCALL call_3_2_5_3(SH2_struct* sh)
{
  SH2cmpge(sh,2,5);
}
#define init_3_2_6_3
void FASTCALL call_3_2_6_3(SH2_struct* sh)
{
  SH2cmpge(sh,2,6);
}
#define init_3_2_7_3
void FASTCALL call_3_2_7_3(SH2_struct* sh)
{
  SH2cmpge(sh,2,7);
}
#define init_3_2_8_3
void FASTCALL call_3_2_8_3(SH2_struct* sh)
{
  SH2cmpge(sh,2,8);
}
#define init_3_2_9_3
void FASTCALL call_3_2_9_3(SH2_struct* sh)
{
  SH2cmpge(sh,2,9);
}
#define init_3_2_10_3
void FASTCALL call_3_2_10_3(SH2_struct* sh)
{
  SH2cmpge(sh,2,10);
}
#define init_3_2_11_3
void FASTCALL call_3_2_11_3(SH2_struct* sh)
{
  SH2cmpge(sh,2,11);
}
#define init_3_2_12_3
void FASTCALL call_3_2_12_3(SH2_struct* sh)
{
  SH2cmpge(sh,2,12);
}
#define init_3_2_13_3
void FASTCALL call_3_2_13_3(SH2_struct* sh)
{
  SH2cmpge(sh,2,13);
}
#define init_3_2_14_3
void FASTCALL call_3_2_14_3(SH2_struct* sh)
{
  SH2cmpge(sh,2,14);
}
#define init_3_2_15_3
void FASTCALL call_3_2_15_3(SH2_struct* sh)
{
  SH2cmpge(sh,2,15);
}
#define init_3_3_0_3
void FASTCALL call_3_3_0_3(SH2_struct* sh)
{
  SH2cmpge(sh,3,0);
}
#define init_3_3_1_3
void FASTCALL call_3_3_1_3(SH2_struct* sh)
{
  SH2cmpge(sh,3,1);
}
#define init_3_3_2_3
void FASTCALL call_3_3_2_3(SH2_struct* sh)
{
  SH2cmpge(sh,3,2);
}
#define init_3_3_3_3
void FASTCALL call_3_3_3_3(SH2_struct* sh)
{
  SH2cmpge(sh,3,3);
}
#define init_3_3_4_3
void FASTCALL call_3_3_4_3(SH2_struct* sh)
{
  SH2cmpge(sh,3,4);
}
#define init_3_3_5_3
void FASTCALL call_3_3_5_3(SH2_struct* sh)
{
  SH2cmpge(sh,3,5);
}
#define init_3_3_6_3
void FASTCALL call_3_3_6_3(SH2_struct* sh)
{
  SH2cmpge(sh,3,6);
}
#define init_3_3_7_3
void FASTCALL call_3_3_7_3(SH2_struct* sh)
{
  SH2cmpge(sh,3,7);
}
#define init_3_3_8_3
void FASTCALL call_3_3_8_3(SH2_struct* sh)
{
  SH2cmpge(sh,3,8);
}
#define init_3_3_9_3
void FASTCALL call_3_3_9_3(SH2_struct* sh)
{
  SH2cmpge(sh,3,9);
}
#define init_3_3_10_3
void FASTCALL call_3_3_10_3(SH2_struct* sh)
{
  SH2cmpge(sh,3,10);
}
#define init_3_3_11_3
void FASTCALL call_3_3_11_3(SH2_struct* sh)
{
  SH2cmpge(sh,3,11);
}
#define init_3_3_12_3
void FASTCALL call_3_3_12_3(SH2_struct* sh)
{
  SH2cmpge(sh,3,12);
}
#define init_3_3_13_3
void FASTCALL call_3_3_13_3(SH2_struct* sh)
{
  SH2cmpge(sh,3,13);
}
#define init_3_3_14_3
void FASTCALL call_3_3_14_3(SH2_struct* sh)
{
  SH2cmpge(sh,3,14);
}
#define init_3_3_15_3
void FASTCALL call_3_3_15_3(SH2_struct* sh)
{
  SH2cmpge(sh,3,15);
}
#define init_3_4_0_3
void FASTCALL call_3_4_0_3(SH2_struct* sh)
{
  SH2cmpge(sh,4,0);
}
#define init_3_4_1_3
void FASTCALL call_3_4_1_3(SH2_struct* sh)
{
  SH2cmpge(sh,4,1);
}
#define init_3_4_2_3
void FASTCALL call_3_4_2_3(SH2_struct* sh)
{
  SH2cmpge(sh,4,2);
}
#define init_3_4_3_3
void FASTCALL call_3_4_3_3(SH2_struct* sh)
{
  SH2cmpge(sh,4,3);
}
#define init_3_4_4_3
void FASTCALL call_3_4_4_3(SH2_struct* sh)
{
  SH2cmpge(sh,4,4);
}
#define init_3_4_5_3
void FASTCALL call_3_4_5_3(SH2_struct* sh)
{
  SH2cmpge(sh,4,5);
}
#define init_3_4_6_3
void FASTCALL call_3_4_6_3(SH2_struct* sh)
{
  SH2cmpge(sh,4,6);
}
#define init_3_4_7_3
void FASTCALL call_3_4_7_3(SH2_struct* sh)
{
  SH2cmpge(sh,4,7);
}
#define init_3_4_8_3
void FASTCALL call_3_4_8_3(SH2_struct* sh)
{
  SH2cmpge(sh,4,8);
}
#define init_3_4_9_3
void FASTCALL call_3_4_9_3(SH2_struct* sh)
{
  SH2cmpge(sh,4,9);
}
#define init_3_4_10_3
void FASTCALL call_3_4_10_3(SH2_struct* sh)
{
  SH2cmpge(sh,4,10);
}
#define init_3_4_11_3
void FASTCALL call_3_4_11_3(SH2_struct* sh)
{
  SH2cmpge(sh,4,11);
}
#define init_3_4_12_3
void FASTCALL call_3_4_12_3(SH2_struct* sh)
{
  SH2cmpge(sh,4,12);
}
#define init_3_4_13_3
void FASTCALL call_3_4_13_3(SH2_struct* sh)
{
  SH2cmpge(sh,4,13);
}
#define init_3_4_14_3
void FASTCALL call_3_4_14_3(SH2_struct* sh)
{
  SH2cmpge(sh,4,14);
}
#define init_3_4_15_3
void FASTCALL call_3_4_15_3(SH2_struct* sh)
{
  SH2cmpge(sh,4,15);
}
#define init_3_5_0_3
void FASTCALL call_3_5_0_3(SH2_struct* sh)
{
  SH2cmpge(sh,5,0);
}
#define init_3_5_1_3
void FASTCALL call_3_5_1_3(SH2_struct* sh)
{
  SH2cmpge(sh,5,1);
}
#define init_3_5_2_3
void FASTCALL call_3_5_2_3(SH2_struct* sh)
{
  SH2cmpge(sh,5,2);
}
#define init_3_5_3_3
void FASTCALL call_3_5_3_3(SH2_struct* sh)
{
  SH2cmpge(sh,5,3);
}
#define init_3_5_4_3
void FASTCALL call_3_5_4_3(SH2_struct* sh)
{
  SH2cmpge(sh,5,4);
}
#define init_3_5_5_3
void FASTCALL call_3_5_5_3(SH2_struct* sh)
{
  SH2cmpge(sh,5,5);
}
#define init_3_5_6_3
void FASTCALL call_3_5_6_3(SH2_struct* sh)
{
  SH2cmpge(sh,5,6);
}
#define init_3_5_7_3
void FASTCALL call_3_5_7_3(SH2_struct* sh)
{
  SH2cmpge(sh,5,7);
}
#define init_3_5_8_3
void FASTCALL call_3_5_8_3(SH2_struct* sh)
{
  SH2cmpge(sh,5,8);
}
#define init_3_5_9_3
void FASTCALL call_3_5_9_3(SH2_struct* sh)
{
  SH2cmpge(sh,5,9);
}
#define init_3_5_10_3
void FASTCALL call_3_5_10_3(SH2_struct* sh)
{
  SH2cmpge(sh,5,10);
}
#define init_3_5_11_3
void FASTCALL call_3_5_11_3(SH2_struct* sh)
{
  SH2cmpge(sh,5,11);
}
#define init_3_5_12_3
void FASTCALL call_3_5_12_3(SH2_struct* sh)
{
  SH2cmpge(sh,5,12);
}
#define init_3_5_13_3
void FASTCALL call_3_5_13_3(SH2_struct* sh)
{
  SH2cmpge(sh,5,13);
}
#define init_3_5_14_3
void FASTCALL call_3_5_14_3(SH2_struct* sh)
{
  SH2cmpge(sh,5,14);
}
#define init_3_5_15_3
void FASTCALL call_3_5_15_3(SH2_struct* sh)
{
  SH2cmpge(sh,5,15);
}
#define init_3_6_0_3
void FASTCALL call_3_6_0_3(SH2_struct* sh)
{
  SH2cmpge(sh,6,0);
}
#define init_3_6_1_3
void FASTCALL call_3_6_1_3(SH2_struct* sh)
{
  SH2cmpge(sh,6,1);
}
#define init_3_6_2_3
void FASTCALL call_3_6_2_3(SH2_struct* sh)
{
  SH2cmpge(sh,6,2);
}
#define init_3_6_3_3
void FASTCALL call_3_6_3_3(SH2_struct* sh)
{
  SH2cmpge(sh,6,3);
}
#define init_3_6_4_3
void FASTCALL call_3_6_4_3(SH2_struct* sh)
{
  SH2cmpge(sh,6,4);
}
#define init_3_6_5_3
void FASTCALL call_3_6_5_3(SH2_struct* sh)
{
  SH2cmpge(sh,6,5);
}
#define init_3_6_6_3
void FASTCALL call_3_6_6_3(SH2_struct* sh)
{
  SH2cmpge(sh,6,6);
}
#define init_3_6_7_3
void FASTCALL call_3_6_7_3(SH2_struct* sh)
{
  SH2cmpge(sh,6,7);
}
#define init_3_6_8_3
void FASTCALL call_3_6_8_3(SH2_struct* sh)
{
  SH2cmpge(sh,6,8);
}
#define init_3_6_9_3
void FASTCALL call_3_6_9_3(SH2_struct* sh)
{
  SH2cmpge(sh,6,9);
}
#define init_3_6_10_3
void FASTCALL call_3_6_10_3(SH2_struct* sh)
{
  SH2cmpge(sh,6,10);
}
#define init_3_6_11_3
void FASTCALL call_3_6_11_3(SH2_struct* sh)
{
  SH2cmpge(sh,6,11);
}
#define init_3_6_12_3
void FASTCALL call_3_6_12_3(SH2_struct* sh)
{
  SH2cmpge(sh,6,12);
}
#define init_3_6_13_3
void FASTCALL call_3_6_13_3(SH2_struct* sh)
{
  SH2cmpge(sh,6,13);
}
#define init_3_6_14_3
void FASTCALL call_3_6_14_3(SH2_struct* sh)
{
  SH2cmpge(sh,6,14);
}
#define init_3_6_15_3
void FASTCALL call_3_6_15_3(SH2_struct* sh)
{
  SH2cmpge(sh,6,15);
}
#define init_3_7_0_3
void FASTCALL call_3_7_0_3(SH2_struct* sh)
{
  SH2cmpge(sh,7,0);
}
#define init_3_7_1_3
void FASTCALL call_3_7_1_3(SH2_struct* sh)
{
  SH2cmpge(sh,7,1);
}
#define init_3_7_2_3
void FASTCALL call_3_7_2_3(SH2_struct* sh)
{
  SH2cmpge(sh,7,2);
}
#define init_3_7_3_3
void FASTCALL call_3_7_3_3(SH2_struct* sh)
{
  SH2cmpge(sh,7,3);
}
#define init_3_7_4_3
void FASTCALL call_3_7_4_3(SH2_struct* sh)
{
  SH2cmpge(sh,7,4);
}
#define init_3_7_5_3
void FASTCALL call_3_7_5_3(SH2_struct* sh)
{
  SH2cmpge(sh,7,5);
}
#define init_3_7_6_3
void FASTCALL call_3_7_6_3(SH2_struct* sh)
{
  SH2cmpge(sh,7,6);
}
#define init_3_7_7_3
void FASTCALL call_3_7_7_3(SH2_struct* sh)
{
  SH2cmpge(sh,7,7);
}
#define init_3_7_8_3
void FASTCALL call_3_7_8_3(SH2_struct* sh)
{
  SH2cmpge(sh,7,8);
}
#define init_3_7_9_3
void FASTCALL call_3_7_9_3(SH2_struct* sh)
{
  SH2cmpge(sh,7,9);
}
#define init_3_7_10_3
void FASTCALL call_3_7_10_3(SH2_struct* sh)
{
  SH2cmpge(sh,7,10);
}
#define init_3_7_11_3
void FASTCALL call_3_7_11_3(SH2_struct* sh)
{
  SH2cmpge(sh,7,11);
}
#define init_3_7_12_3
void FASTCALL call_3_7_12_3(SH2_struct* sh)
{
  SH2cmpge(sh,7,12);
}
#define init_3_7_13_3
void FASTCALL call_3_7_13_3(SH2_struct* sh)
{
  SH2cmpge(sh,7,13);
}
#define init_3_7_14_3
void FASTCALL call_3_7_14_3(SH2_struct* sh)
{
  SH2cmpge(sh,7,14);
}
#define init_3_7_15_3
void FASTCALL call_3_7_15_3(SH2_struct* sh)
{
  SH2cmpge(sh,7,15);
}
#define init_3_8_0_3
void FASTCALL call_3_8_0_3(SH2_struct* sh)
{
  SH2cmpge(sh,8,0);
}
#define init_3_8_1_3
void FASTCALL call_3_8_1_3(SH2_struct* sh)
{
  SH2cmpge(sh,8,1);
}
#define init_3_8_2_3
void FASTCALL call_3_8_2_3(SH2_struct* sh)
{
  SH2cmpge(sh,8,2);
}
#define init_3_8_3_3
void FASTCALL call_3_8_3_3(SH2_struct* sh)
{
  SH2cmpge(sh,8,3);
}
#define init_3_8_4_3
void FASTCALL call_3_8_4_3(SH2_struct* sh)
{
  SH2cmpge(sh,8,4);
}
#define init_3_8_5_3
void FASTCALL call_3_8_5_3(SH2_struct* sh)
{
  SH2cmpge(sh,8,5);
}
#define init_3_8_6_3
void FASTCALL call_3_8_6_3(SH2_struct* sh)
{
  SH2cmpge(sh,8,6);
}
#define init_3_8_7_3
void FASTCALL call_3_8_7_3(SH2_struct* sh)
{
  SH2cmpge(sh,8,7);
}
#define init_3_8_8_3
void FASTCALL call_3_8_8_3(SH2_struct* sh)
{
  SH2cmpge(sh,8,8);
}
#define init_3_8_9_3
void FASTCALL call_3_8_9_3(SH2_struct* sh)
{
  SH2cmpge(sh,8,9);
}
#define init_3_8_10_3
void FASTCALL call_3_8_10_3(SH2_struct* sh)
{
  SH2cmpge(sh,8,10);
}
#define init_3_8_11_3
void FASTCALL call_3_8_11_3(SH2_struct* sh)
{
  SH2cmpge(sh,8,11);
}
#define init_3_8_12_3
void FASTCALL call_3_8_12_3(SH2_struct* sh)
{
  SH2cmpge(sh,8,12);
}
#define init_3_8_13_3
void FASTCALL call_3_8_13_3(SH2_struct* sh)
{
  SH2cmpge(sh,8,13);
}
#define init_3_8_14_3
void FASTCALL call_3_8_14_3(SH2_struct* sh)
{
  SH2cmpge(sh,8,14);
}
#define init_3_8_15_3
void FASTCALL call_3_8_15_3(SH2_struct* sh)
{
  SH2cmpge(sh,8,15);
}
#define init_3_9_0_3
void FASTCALL call_3_9_0_3(SH2_struct* sh)
{
  SH2cmpge(sh,9,0);
}
#define init_3_9_1_3
void FASTCALL call_3_9_1_3(SH2_struct* sh)
{
  SH2cmpge(sh,9,1);
}
#define init_3_9_2_3
void FASTCALL call_3_9_2_3(SH2_struct* sh)
{
  SH2cmpge(sh,9,2);
}
#define init_3_9_3_3
void FASTCALL call_3_9_3_3(SH2_struct* sh)
{
  SH2cmpge(sh,9,3);
}
#define init_3_9_4_3
void FASTCALL call_3_9_4_3(SH2_struct* sh)
{
  SH2cmpge(sh,9,4);
}
#define init_3_9_5_3
void FASTCALL call_3_9_5_3(SH2_struct* sh)
{
  SH2cmpge(sh,9,5);
}
#define init_3_9_6_3
void FASTCALL call_3_9_6_3(SH2_struct* sh)
{
  SH2cmpge(sh,9,6);
}
#define init_3_9_7_3
void FASTCALL call_3_9_7_3(SH2_struct* sh)
{
  SH2cmpge(sh,9,7);
}
#define init_3_9_8_3
void FASTCALL call_3_9_8_3(SH2_struct* sh)
{
  SH2cmpge(sh,9,8);
}
#define init_3_9_9_3
void FASTCALL call_3_9_9_3(SH2_struct* sh)
{
  SH2cmpge(sh,9,9);
}
#define init_3_9_10_3
void FASTCALL call_3_9_10_3(SH2_struct* sh)
{
  SH2cmpge(sh,9,10);
}
#define init_3_9_11_3
void FASTCALL call_3_9_11_3(SH2_struct* sh)
{
  SH2cmpge(sh,9,11);
}
#define init_3_9_12_3
void FASTCALL call_3_9_12_3(SH2_struct* sh)
{
  SH2cmpge(sh,9,12);
}
#define init_3_9_13_3
void FASTCALL call_3_9_13_3(SH2_struct* sh)
{
  SH2cmpge(sh,9,13);
}
#define init_3_9_14_3
void FASTCALL call_3_9_14_3(SH2_struct* sh)
{
  SH2cmpge(sh,9,14);
}
#define init_3_9_15_3
void FASTCALL call_3_9_15_3(SH2_struct* sh)
{
  SH2cmpge(sh,9,15);
}
#define init_3_10_0_3
void FASTCALL call_3_10_0_3(SH2_struct* sh)
{
  SH2cmpge(sh,10,0);
}
#define init_3_10_1_3
void FASTCALL call_3_10_1_3(SH2_struct* sh)
{
  SH2cmpge(sh,10,1);
}
#define init_3_10_2_3
void FASTCALL call_3_10_2_3(SH2_struct* sh)
{
  SH2cmpge(sh,10,2);
}
#define init_3_10_3_3
void FASTCALL call_3_10_3_3(SH2_struct* sh)
{
  SH2cmpge(sh,10,3);
}
#define init_3_10_4_3
void FASTCALL call_3_10_4_3(SH2_struct* sh)
{
  SH2cmpge(sh,10,4);
}
#define init_3_10_5_3
void FASTCALL call_3_10_5_3(SH2_struct* sh)
{
  SH2cmpge(sh,10,5);
}
#define init_3_10_6_3
void FASTCALL call_3_10_6_3(SH2_struct* sh)
{
  SH2cmpge(sh,10,6);
}
#define init_3_10_7_3
void FASTCALL call_3_10_7_3(SH2_struct* sh)
{
  SH2cmpge(sh,10,7);
}
#define init_3_10_8_3
void FASTCALL call_3_10_8_3(SH2_struct* sh)
{
  SH2cmpge(sh,10,8);
}
#define init_3_10_9_3
void FASTCALL call_3_10_9_3(SH2_struct* sh)
{
  SH2cmpge(sh,10,9);
}
#define init_3_10_10_3
void FASTCALL call_3_10_10_3(SH2_struct* sh)
{
  SH2cmpge(sh,10,10);
}
#define init_3_10_11_3
void FASTCALL call_3_10_11_3(SH2_struct* sh)
{
  SH2cmpge(sh,10,11);
}
#define init_3_10_12_3
void FASTCALL call_3_10_12_3(SH2_struct* sh)
{
  SH2cmpge(sh,10,12);
}
#define init_3_10_13_3
void FASTCALL call_3_10_13_3(SH2_struct* sh)
{
  SH2cmpge(sh,10,13);
}
#define init_3_10_14_3
void FASTCALL call_3_10_14_3(SH2_struct* sh)
{
  SH2cmpge(sh,10,14);
}
#define init_3_10_15_3
void FASTCALL call_3_10_15_3(SH2_struct* sh)
{
  SH2cmpge(sh,10,15);
}
#define init_3_11_0_3
void FASTCALL call_3_11_0_3(SH2_struct* sh)
{
  SH2cmpge(sh,11,0);
}
#define init_3_11_1_3
void FASTCALL call_3_11_1_3(SH2_struct* sh)
{
  SH2cmpge(sh,11,1);
}
#define init_3_11_2_3
void FASTCALL call_3_11_2_3(SH2_struct* sh)
{
  SH2cmpge(sh,11,2);
}
#define init_3_11_3_3
void FASTCALL call_3_11_3_3(SH2_struct* sh)
{
  SH2cmpge(sh,11,3);
}
#define init_3_11_4_3
void FASTCALL call_3_11_4_3(SH2_struct* sh)
{
  SH2cmpge(sh,11,4);
}
#define init_3_11_5_3
void FASTCALL call_3_11_5_3(SH2_struct* sh)
{
  SH2cmpge(sh,11,5);
}
#define init_3_11_6_3
void FASTCALL call_3_11_6_3(SH2_struct* sh)
{
  SH2cmpge(sh,11,6);
}
#define init_3_11_7_3
void FASTCALL call_3_11_7_3(SH2_struct* sh)
{
  SH2cmpge(sh,11,7);
}
#define init_3_11_8_3
void FASTCALL call_3_11_8_3(SH2_struct* sh)
{
  SH2cmpge(sh,11,8);
}
#define init_3_11_9_3
void FASTCALL call_3_11_9_3(SH2_struct* sh)
{
  SH2cmpge(sh,11,9);
}
#define init_3_11_10_3
void FASTCALL call_3_11_10_3(SH2_struct* sh)
{
  SH2cmpge(sh,11,10);
}
#define init_3_11_11_3
void FASTCALL call_3_11_11_3(SH2_struct* sh)
{
  SH2cmpge(sh,11,11);
}
#define init_3_11_12_3
void FASTCALL call_3_11_12_3(SH2_struct* sh)
{
  SH2cmpge(sh,11,12);
}
#define init_3_11_13_3
void FASTCALL call_3_11_13_3(SH2_struct* sh)
{
  SH2cmpge(sh,11,13);
}
#define init_3_11_14_3
void FASTCALL call_3_11_14_3(SH2_struct* sh)
{
  SH2cmpge(sh,11,14);
}
#define init_3_11_15_3
void FASTCALL call_3_11_15_3(SH2_struct* sh)
{
  SH2cmpge(sh,11,15);
}
#define init_3_12_0_3
void FASTCALL call_3_12_0_3(SH2_struct* sh)
{
  SH2cmpge(sh,12,0);
}
#define init_3_12_1_3
void FASTCALL call_3_12_1_3(SH2_struct* sh)
{
  SH2cmpge(sh,12,1);
}
#define init_3_12_2_3
void FASTCALL call_3_12_2_3(SH2_struct* sh)
{
  SH2cmpge(sh,12,2);
}
#define init_3_12_3_3
void FASTCALL call_3_12_3_3(SH2_struct* sh)
{
  SH2cmpge(sh,12,3);
}
#define init_3_12_4_3
void FASTCALL call_3_12_4_3(SH2_struct* sh)
{
  SH2cmpge(sh,12,4);
}
#define init_3_12_5_3
void FASTCALL call_3_12_5_3(SH2_struct* sh)
{
  SH2cmpge(sh,12,5);
}
#define init_3_12_6_3
void FASTCALL call_3_12_6_3(SH2_struct* sh)
{
  SH2cmpge(sh,12,6);
}
#define init_3_12_7_3
void FASTCALL call_3_12_7_3(SH2_struct* sh)
{
  SH2cmpge(sh,12,7);
}
#define init_3_12_8_3
void FASTCALL call_3_12_8_3(SH2_struct* sh)
{
  SH2cmpge(sh,12,8);
}
#define init_3_12_9_3
void FASTCALL call_3_12_9_3(SH2_struct* sh)
{
  SH2cmpge(sh,12,9);
}
#define init_3_12_10_3
void FASTCALL call_3_12_10_3(SH2_struct* sh)
{
  SH2cmpge(sh,12,10);
}
#define init_3_12_11_3
void FASTCALL call_3_12_11_3(SH2_struct* sh)
{
  SH2cmpge(sh,12,11);
}
#define init_3_12_12_3
void FASTCALL call_3_12_12_3(SH2_struct* sh)
{
  SH2cmpge(sh,12,12);
}
#define init_3_12_13_3
void FASTCALL call_3_12_13_3(SH2_struct* sh)
{
  SH2cmpge(sh,12,13);
}
#define init_3_12_14_3
void FASTCALL call_3_12_14_3(SH2_struct* sh)
{
  SH2cmpge(sh,12,14);
}
#define init_3_12_15_3
void FASTCALL call_3_12_15_3(SH2_struct* sh)
{
  SH2cmpge(sh,12,15);
}
#define init_3_13_0_3
void FASTCALL call_3_13_0_3(SH2_struct* sh)
{
  SH2cmpge(sh,13,0);
}
#define init_3_13_1_3
void FASTCALL call_3_13_1_3(SH2_struct* sh)
{
  SH2cmpge(sh,13,1);
}
#define init_3_13_2_3
void FASTCALL call_3_13_2_3(SH2_struct* sh)
{
  SH2cmpge(sh,13,2);
}
#define init_3_13_3_3
void FASTCALL call_3_13_3_3(SH2_struct* sh)
{
  SH2cmpge(sh,13,3);
}
#define init_3_13_4_3
void FASTCALL call_3_13_4_3(SH2_struct* sh)
{
  SH2cmpge(sh,13,4);
}
#define init_3_13_5_3
void FASTCALL call_3_13_5_3(SH2_struct* sh)
{
  SH2cmpge(sh,13,5);
}
#define init_3_13_6_3
void FASTCALL call_3_13_6_3(SH2_struct* sh)
{
  SH2cmpge(sh,13,6);
}
#define init_3_13_7_3
void FASTCALL call_3_13_7_3(SH2_struct* sh)
{
  SH2cmpge(sh,13,7);
}
#define init_3_13_8_3
void FASTCALL call_3_13_8_3(SH2_struct* sh)
{
  SH2cmpge(sh,13,8);
}
#define init_3_13_9_3
void FASTCALL call_3_13_9_3(SH2_struct* sh)
{
  SH2cmpge(sh,13,9);
}
#define init_3_13_10_3
void FASTCALL call_3_13_10_3(SH2_struct* sh)
{
  SH2cmpge(sh,13,10);
}
#define init_3_13_11_3
void FASTCALL call_3_13_11_3(SH2_struct* sh)
{
  SH2cmpge(sh,13,11);
}
#define init_3_13_12_3
void FASTCALL call_3_13_12_3(SH2_struct* sh)
{
  SH2cmpge(sh,13,12);
}
#define init_3_13_13_3
void FASTCALL call_3_13_13_3(SH2_struct* sh)
{
  SH2cmpge(sh,13,13);
}
#define init_3_13_14_3
void FASTCALL call_3_13_14_3(SH2_struct* sh)
{
  SH2cmpge(sh,13,14);
}
#define init_3_13_15_3
void FASTCALL call_3_13_15_3(SH2_struct* sh)
{
  SH2cmpge(sh,13,15);
}
#define init_3_14_0_3
void FASTCALL call_3_14_0_3(SH2_struct* sh)
{
  SH2cmpge(sh,14,0);
}
#define init_3_14_1_3
void FASTCALL call_3_14_1_3(SH2_struct* sh)
{
  SH2cmpge(sh,14,1);
}
#define init_3_14_2_3
void FASTCALL call_3_14_2_3(SH2_struct* sh)
{
  SH2cmpge(sh,14,2);
}
#define init_3_14_3_3
void FASTCALL call_3_14_3_3(SH2_struct* sh)
{
  SH2cmpge(sh,14,3);
}
#define init_3_14_4_3
void FASTCALL call_3_14_4_3(SH2_struct* sh)
{
  SH2cmpge(sh,14,4);
}
#define init_3_14_5_3
void FASTCALL call_3_14_5_3(SH2_struct* sh)
{
  SH2cmpge(sh,14,5);
}
#define init_3_14_6_3
void FASTCALL call_3_14_6_3(SH2_struct* sh)
{
  SH2cmpge(sh,14,6);
}
#define init_3_14_7_3
void FASTCALL call_3_14_7_3(SH2_struct* sh)
{
  SH2cmpge(sh,14,7);
}
#define init_3_14_8_3
void FASTCALL call_3_14_8_3(SH2_struct* sh)
{
  SH2cmpge(sh,14,8);
}
#define init_3_14_9_3
void FASTCALL call_3_14_9_3(SH2_struct* sh)
{
  SH2cmpge(sh,14,9);
}
#define init_3_14_10_3
void FASTCALL call_3_14_10_3(SH2_struct* sh)
{
  SH2cmpge(sh,14,10);
}
#define init_3_14_11_3
void FASTCALL call_3_14_11_3(SH2_struct* sh)
{
  SH2cmpge(sh,14,11);
}
#define init_3_14_12_3
void FASTCALL call_3_14_12_3(SH2_struct* sh)
{
  SH2cmpge(sh,14,12);
}
#define init_3_14_13_3
void FASTCALL call_3_14_13_3(SH2_struct* sh)
{
  SH2cmpge(sh,14,13);
}
#define init_3_14_14_3
void FASTCALL call_3_14_14_3(SH2_struct* sh)
{
  SH2cmpge(sh,14,14);
}
#define init_3_14_15_3
void FASTCALL call_3_14_15_3(SH2_struct* sh)
{
  SH2cmpge(sh,14,15);
}
#define init_3_15_0_3
void FASTCALL call_3_15_0_3(SH2_struct* sh)
{
  SH2cmpge(sh,15,0);
}
#define init_3_15_1_3
void FASTCALL call_3_15_1_3(SH2_struct* sh)
{
  SH2cmpge(sh,15,1);
}
#define init_3_15_2_3
void FASTCALL call_3_15_2_3(SH2_struct* sh)
{
  SH2cmpge(sh,15,2);
}
#define init_3_15_3_3
void FASTCALL call_3_15_3_3(SH2_struct* sh)
{
  SH2cmpge(sh,15,3);
}
#define init_3_15_4_3
void FASTCALL call_3_15_4_3(SH2_struct* sh)
{
  SH2cmpge(sh,15,4);
}
#define init_3_15_5_3
void FASTCALL call_3_15_5_3(SH2_struct* sh)
{
  SH2cmpge(sh,15,5);
}
#define init_3_15_6_3
void FASTCALL call_3_15_6_3(SH2_struct* sh)
{
  SH2cmpge(sh,15,6);
}
#define init_3_15_7_3
void FASTCALL call_3_15_7_3(SH2_struct* sh)
{
  SH2cmpge(sh,15,7);
}
#define init_3_15_8_3
void FASTCALL call_3_15_8_3(SH2_struct* sh)
{
  SH2cmpge(sh,15,8);
}
#define init_3_15_9_3
void FASTCALL call_3_15_9_3(SH2_struct* sh)
{
  SH2cmpge(sh,15,9);
}
#define init_3_15_10_3
void FASTCALL call_3_15_10_3(SH2_struct* sh)
{
  SH2cmpge(sh,15,10);
}
#define init_3_15_11_3
void FASTCALL call_3_15_11_3(SH2_struct* sh)
{
  SH2cmpge(sh,15,11);
}
#define init_3_15_12_3
void FASTCALL call_3_15_12_3(SH2_struct* sh)
{
  SH2cmpge(sh,15,12);
}
#define init_3_15_13_3
void FASTCALL call_3_15_13_3(SH2_struct* sh)
{
  SH2cmpge(sh,15,13);
}
#define init_3_15_14_3
void FASTCALL call_3_15_14_3(SH2_struct* sh)
{
  SH2cmpge(sh,15,14);
}
#define init_3_15_15_3
void FASTCALL call_3_15_15_3(SH2_struct* sh)
{
  SH2cmpge(sh,15,15);
}
#define init_3_0_0_0
void FASTCALL call_3_0_0_0(SH2_struct* sh)
{
  SH2cmpeq(sh,0,0);
}
#define init_3_0_1_0
void FASTCALL call_3_0_1_0(SH2_struct* sh)
{
  SH2cmpeq(sh,0,1);
}
#define init_3_0_2_0
void FASTCALL call_3_0_2_0(SH2_struct* sh)
{
  SH2cmpeq(sh,0,2);
}
#define init_3_0_3_0
void FASTCALL call_3_0_3_0(SH2_struct* sh)
{
  SH2cmpeq(sh,0,3);
}
#define init_3_0_4_0
void FASTCALL call_3_0_4_0(SH2_struct* sh)
{
  SH2cmpeq(sh,0,4);
}
#define init_3_0_5_0
void FASTCALL call_3_0_5_0(SH2_struct* sh)
{
  SH2cmpeq(sh,0,5);
}
#define init_3_0_6_0
void FASTCALL call_3_0_6_0(SH2_struct* sh)
{
  SH2cmpeq(sh,0,6);
}
#define init_3_0_7_0
void FASTCALL call_3_0_7_0(SH2_struct* sh)
{
  SH2cmpeq(sh,0,7);
}
#define init_3_0_8_0
void FASTCALL call_3_0_8_0(SH2_struct* sh)
{
  SH2cmpeq(sh,0,8);
}
#define init_3_0_9_0
void FASTCALL call_3_0_9_0(SH2_struct* sh)
{
  SH2cmpeq(sh,0,9);
}
#define init_3_0_10_0
void FASTCALL call_3_0_10_0(SH2_struct* sh)
{
  SH2cmpeq(sh,0,10);
}
#define init_3_0_11_0
void FASTCALL call_3_0_11_0(SH2_struct* sh)
{
  SH2cmpeq(sh,0,11);
}
#define init_3_0_12_0
void FASTCALL call_3_0_12_0(SH2_struct* sh)
{
  SH2cmpeq(sh,0,12);
}
#define init_3_0_13_0
void FASTCALL call_3_0_13_0(SH2_struct* sh)
{
  SH2cmpeq(sh,0,13);
}
#define init_3_0_14_0
void FASTCALL call_3_0_14_0(SH2_struct* sh)
{
  SH2cmpeq(sh,0,14);
}
#define init_3_0_15_0
void FASTCALL call_3_0_15_0(SH2_struct* sh)
{
  SH2cmpeq(sh,0,15);
}
#define init_3_1_0_0
void FASTCALL call_3_1_0_0(SH2_struct* sh)
{
  SH2cmpeq(sh,1,0);
}
#define init_3_1_1_0
void FASTCALL call_3_1_1_0(SH2_struct* sh)
{
  SH2cmpeq(sh,1,1);
}
#define init_3_1_2_0
void FASTCALL call_3_1_2_0(SH2_struct* sh)
{
  SH2cmpeq(sh,1,2);
}
#define init_3_1_3_0
void FASTCALL call_3_1_3_0(SH2_struct* sh)
{
  SH2cmpeq(sh,1,3);
}
#define init_3_1_4_0
void FASTCALL call_3_1_4_0(SH2_struct* sh)
{
  SH2cmpeq(sh,1,4);
}
#define init_3_1_5_0
void FASTCALL call_3_1_5_0(SH2_struct* sh)
{
  SH2cmpeq(sh,1,5);
}
#define init_3_1_6_0
void FASTCALL call_3_1_6_0(SH2_struct* sh)
{
  SH2cmpeq(sh,1,6);
}
#define init_3_1_7_0
void FASTCALL call_3_1_7_0(SH2_struct* sh)
{
  SH2cmpeq(sh,1,7);
}
#define init_3_1_8_0
void FASTCALL call_3_1_8_0(SH2_struct* sh)
{
  SH2cmpeq(sh,1,8);
}
#define init_3_1_9_0
void FASTCALL call_3_1_9_0(SH2_struct* sh)
{
  SH2cmpeq(sh,1,9);
}
#define init_3_1_10_0
void FASTCALL call_3_1_10_0(SH2_struct* sh)
{
  SH2cmpeq(sh,1,10);
}
#define init_3_1_11_0
void FASTCALL call_3_1_11_0(SH2_struct* sh)
{
  SH2cmpeq(sh,1,11);
}
#define init_3_1_12_0
void FASTCALL call_3_1_12_0(SH2_struct* sh)
{
  SH2cmpeq(sh,1,12);
}
#define init_3_1_13_0
void FASTCALL call_3_1_13_0(SH2_struct* sh)
{
  SH2cmpeq(sh,1,13);
}
#define init_3_1_14_0
void FASTCALL call_3_1_14_0(SH2_struct* sh)
{
  SH2cmpeq(sh,1,14);
}
#define init_3_1_15_0
void FASTCALL call_3_1_15_0(SH2_struct* sh)
{
  SH2cmpeq(sh,1,15);
}
#define init_3_2_0_0
void FASTCALL call_3_2_0_0(SH2_struct* sh)
{
  SH2cmpeq(sh,2,0);
}
#define init_3_2_1_0
void FASTCALL call_3_2_1_0(SH2_struct* sh)
{
  SH2cmpeq(sh,2,1);
}
#define init_3_2_2_0
void FASTCALL call_3_2_2_0(SH2_struct* sh)
{
  SH2cmpeq(sh,2,2);
}
#define init_3_2_3_0
void FASTCALL call_3_2_3_0(SH2_struct* sh)
{
  SH2cmpeq(sh,2,3);
}
#define init_3_2_4_0
void FASTCALL call_3_2_4_0(SH2_struct* sh)
{
  SH2cmpeq(sh,2,4);
}
#define init_3_2_5_0
void FASTCALL call_3_2_5_0(SH2_struct* sh)
{
  SH2cmpeq(sh,2,5);
}
#define init_3_2_6_0
void FASTCALL call_3_2_6_0(SH2_struct* sh)
{
  SH2cmpeq(sh,2,6);
}
#define init_3_2_7_0
void FASTCALL call_3_2_7_0(SH2_struct* sh)
{
  SH2cmpeq(sh,2,7);
}
#define init_3_2_8_0
void FASTCALL call_3_2_8_0(SH2_struct* sh)
{
  SH2cmpeq(sh,2,8);
}
#define init_3_2_9_0
void FASTCALL call_3_2_9_0(SH2_struct* sh)
{
  SH2cmpeq(sh,2,9);
}
#define init_3_2_10_0
void FASTCALL call_3_2_10_0(SH2_struct* sh)
{
  SH2cmpeq(sh,2,10);
}
#define init_3_2_11_0
void FASTCALL call_3_2_11_0(SH2_struct* sh)
{
  SH2cmpeq(sh,2,11);
}
#define init_3_2_12_0
void FASTCALL call_3_2_12_0(SH2_struct* sh)
{
  SH2cmpeq(sh,2,12);
}
#define init_3_2_13_0
void FASTCALL call_3_2_13_0(SH2_struct* sh)
{
  SH2cmpeq(sh,2,13);
}
#define init_3_2_14_0
void FASTCALL call_3_2_14_0(SH2_struct* sh)
{
  SH2cmpeq(sh,2,14);
}
#define init_3_2_15_0
void FASTCALL call_3_2_15_0(SH2_struct* sh)
{
  SH2cmpeq(sh,2,15);
}
#define init_3_3_0_0
void FASTCALL call_3_3_0_0(SH2_struct* sh)
{
  SH2cmpeq(sh,3,0);
}
#define init_3_3_1_0
void FASTCALL call_3_3_1_0(SH2_struct* sh)
{
  SH2cmpeq(sh,3,1);
}
#define init_3_3_2_0
void FASTCALL call_3_3_2_0(SH2_struct* sh)
{
  SH2cmpeq(sh,3,2);
}
#define init_3_3_3_0
void FASTCALL call_3_3_3_0(SH2_struct* sh)
{
  SH2cmpeq(sh,3,3);
}
#define init_3_3_4_0
void FASTCALL call_3_3_4_0(SH2_struct* sh)
{
  SH2cmpeq(sh,3,4);
}
#define init_3_3_5_0
void FASTCALL call_3_3_5_0(SH2_struct* sh)
{
  SH2cmpeq(sh,3,5);
}
#define init_3_3_6_0
void FASTCALL call_3_3_6_0(SH2_struct* sh)
{
  SH2cmpeq(sh,3,6);
}
#define init_3_3_7_0
void FASTCALL call_3_3_7_0(SH2_struct* sh)
{
  SH2cmpeq(sh,3,7);
}
#define init_3_3_8_0
void FASTCALL call_3_3_8_0(SH2_struct* sh)
{
  SH2cmpeq(sh,3,8);
}
#define init_3_3_9_0
void FASTCALL call_3_3_9_0(SH2_struct* sh)
{
  SH2cmpeq(sh,3,9);
}
#define init_3_3_10_0
void FASTCALL call_3_3_10_0(SH2_struct* sh)
{
  SH2cmpeq(sh,3,10);
}
#define init_3_3_11_0
void FASTCALL call_3_3_11_0(SH2_struct* sh)
{
  SH2cmpeq(sh,3,11);
}
#define init_3_3_12_0
void FASTCALL call_3_3_12_0(SH2_struct* sh)
{
  SH2cmpeq(sh,3,12);
}
#define init_3_3_13_0
void FASTCALL call_3_3_13_0(SH2_struct* sh)
{
  SH2cmpeq(sh,3,13);
}
#define init_3_3_14_0
void FASTCALL call_3_3_14_0(SH2_struct* sh)
{
  SH2cmpeq(sh,3,14);
}
#define init_3_3_15_0
void FASTCALL call_3_3_15_0(SH2_struct* sh)
{
  SH2cmpeq(sh,3,15);
}
#define init_3_4_0_0
void FASTCALL call_3_4_0_0(SH2_struct* sh)
{
  SH2cmpeq(sh,4,0);
}
#define init_3_4_1_0
void FASTCALL call_3_4_1_0(SH2_struct* sh)
{
  SH2cmpeq(sh,4,1);
}
#define init_3_4_2_0
void FASTCALL call_3_4_2_0(SH2_struct* sh)
{
  SH2cmpeq(sh,4,2);
}
#define init_3_4_3_0
void FASTCALL call_3_4_3_0(SH2_struct* sh)
{
  SH2cmpeq(sh,4,3);
}
#define init_3_4_4_0
void FASTCALL call_3_4_4_0(SH2_struct* sh)
{
  SH2cmpeq(sh,4,4);
}
#define init_3_4_5_0
void FASTCALL call_3_4_5_0(SH2_struct* sh)
{
  SH2cmpeq(sh,4,5);
}
#define init_3_4_6_0
void FASTCALL call_3_4_6_0(SH2_struct* sh)
{
  SH2cmpeq(sh,4,6);
}
#define init_3_4_7_0
void FASTCALL call_3_4_7_0(SH2_struct* sh)
{
  SH2cmpeq(sh,4,7);
}
#define init_3_4_8_0
void FASTCALL call_3_4_8_0(SH2_struct* sh)
{
  SH2cmpeq(sh,4,8);
}
#define init_3_4_9_0
void FASTCALL call_3_4_9_0(SH2_struct* sh)
{
  SH2cmpeq(sh,4,9);
}
#define init_3_4_10_0
void FASTCALL call_3_4_10_0(SH2_struct* sh)
{
  SH2cmpeq(sh,4,10);
}
#define init_3_4_11_0
void FASTCALL call_3_4_11_0(SH2_struct* sh)
{
  SH2cmpeq(sh,4,11);
}
#define init_3_4_12_0
void FASTCALL call_3_4_12_0(SH2_struct* sh)
{
  SH2cmpeq(sh,4,12);
}
#define init_3_4_13_0
void FASTCALL call_3_4_13_0(SH2_struct* sh)
{
  SH2cmpeq(sh,4,13);
}
#define init_3_4_14_0
void FASTCALL call_3_4_14_0(SH2_struct* sh)
{
  SH2cmpeq(sh,4,14);
}
#define init_3_4_15_0
void FASTCALL call_3_4_15_0(SH2_struct* sh)
{
  SH2cmpeq(sh,4,15);
}
#define init_3_5_0_0
void FASTCALL call_3_5_0_0(SH2_struct* sh)
{
  SH2cmpeq(sh,5,0);
}
#define init_3_5_1_0
void FASTCALL call_3_5_1_0(SH2_struct* sh)
{
  SH2cmpeq(sh,5,1);
}
#define init_3_5_2_0
void FASTCALL call_3_5_2_0(SH2_struct* sh)
{
  SH2cmpeq(sh,5,2);
}
#define init_3_5_3_0
void FASTCALL call_3_5_3_0(SH2_struct* sh)
{
  SH2cmpeq(sh,5,3);
}
#define init_3_5_4_0
void FASTCALL call_3_5_4_0(SH2_struct* sh)
{
  SH2cmpeq(sh,5,4);
}
#define init_3_5_5_0
void FASTCALL call_3_5_5_0(SH2_struct* sh)
{
  SH2cmpeq(sh,5,5);
}
#define init_3_5_6_0
void FASTCALL call_3_5_6_0(SH2_struct* sh)
{
  SH2cmpeq(sh,5,6);
}
#define init_3_5_7_0
void FASTCALL call_3_5_7_0(SH2_struct* sh)
{
  SH2cmpeq(sh,5,7);
}
#define init_3_5_8_0
void FASTCALL call_3_5_8_0(SH2_struct* sh)
{
  SH2cmpeq(sh,5,8);
}
#define init_3_5_9_0
void FASTCALL call_3_5_9_0(SH2_struct* sh)
{
  SH2cmpeq(sh,5,9);
}
#define init_3_5_10_0
void FASTCALL call_3_5_10_0(SH2_struct* sh)
{
  SH2cmpeq(sh,5,10);
}
#define init_3_5_11_0
void FASTCALL call_3_5_11_0(SH2_struct* sh)
{
  SH2cmpeq(sh,5,11);
}
#define init_3_5_12_0
void FASTCALL call_3_5_12_0(SH2_struct* sh)
{
  SH2cmpeq(sh,5,12);
}
#define init_3_5_13_0
void FASTCALL call_3_5_13_0(SH2_struct* sh)
{
  SH2cmpeq(sh,5,13);
}
#define init_3_5_14_0
void FASTCALL call_3_5_14_0(SH2_struct* sh)
{
  SH2cmpeq(sh,5,14);
}
#define init_3_5_15_0
void FASTCALL call_3_5_15_0(SH2_struct* sh)
{
  SH2cmpeq(sh,5,15);
}
#define init_3_6_0_0
void FASTCALL call_3_6_0_0(SH2_struct* sh)
{
  SH2cmpeq(sh,6,0);
}
#define init_3_6_1_0
void FASTCALL call_3_6_1_0(SH2_struct* sh)
{
  SH2cmpeq(sh,6,1);
}
#define init_3_6_2_0
void FASTCALL call_3_6_2_0(SH2_struct* sh)
{
  SH2cmpeq(sh,6,2);
}
#define init_3_6_3_0
void FASTCALL call_3_6_3_0(SH2_struct* sh)
{
  SH2cmpeq(sh,6,3);
}
#define init_3_6_4_0
void FASTCALL call_3_6_4_0(SH2_struct* sh)
{
  SH2cmpeq(sh,6,4);
}
#define init_3_6_5_0
void FASTCALL call_3_6_5_0(SH2_struct* sh)
{
  SH2cmpeq(sh,6,5);
}
#define init_3_6_6_0
void FASTCALL call_3_6_6_0(SH2_struct* sh)
{
  SH2cmpeq(sh,6,6);
}
#define init_3_6_7_0
void FASTCALL call_3_6_7_0(SH2_struct* sh)
{
  SH2cmpeq(sh,6,7);
}
#define init_3_6_8_0
void FASTCALL call_3_6_8_0(SH2_struct* sh)
{
  SH2cmpeq(sh,6,8);
}
#define init_3_6_9_0
void FASTCALL call_3_6_9_0(SH2_struct* sh)
{
  SH2cmpeq(sh,6,9);
}
#define init_3_6_10_0
void FASTCALL call_3_6_10_0(SH2_struct* sh)
{
  SH2cmpeq(sh,6,10);
}
#define init_3_6_11_0
void FASTCALL call_3_6_11_0(SH2_struct* sh)
{
  SH2cmpeq(sh,6,11);
}
#define init_3_6_12_0
void FASTCALL call_3_6_12_0(SH2_struct* sh)
{
  SH2cmpeq(sh,6,12);
}
#define init_3_6_13_0
void FASTCALL call_3_6_13_0(SH2_struct* sh)
{
  SH2cmpeq(sh,6,13);
}
#define init_3_6_14_0
void FASTCALL call_3_6_14_0(SH2_struct* sh)
{
  SH2cmpeq(sh,6,14);
}
#define init_3_6_15_0
void FASTCALL call_3_6_15_0(SH2_struct* sh)
{
  SH2cmpeq(sh,6,15);
}
#define init_3_7_0_0
void FASTCALL call_3_7_0_0(SH2_struct* sh)
{
  SH2cmpeq(sh,7,0);
}
#define init_3_7_1_0
void FASTCALL call_3_7_1_0(SH2_struct* sh)
{
  SH2cmpeq(sh,7,1);
}
#define init_3_7_2_0
void FASTCALL call_3_7_2_0(SH2_struct* sh)
{
  SH2cmpeq(sh,7,2);
}
#define init_3_7_3_0
void FASTCALL call_3_7_3_0(SH2_struct* sh)
{
  SH2cmpeq(sh,7,3);
}
#define init_3_7_4_0
void FASTCALL call_3_7_4_0(SH2_struct* sh)
{
  SH2cmpeq(sh,7,4);
}
#define init_3_7_5_0
void FASTCALL call_3_7_5_0(SH2_struct* sh)
{
  SH2cmpeq(sh,7,5);
}
#define init_3_7_6_0
void FASTCALL call_3_7_6_0(SH2_struct* sh)
{
  SH2cmpeq(sh,7,6);
}
#define init_3_7_7_0
void FASTCALL call_3_7_7_0(SH2_struct* sh)
{
  SH2cmpeq(sh,7,7);
}
#define init_3_7_8_0
void FASTCALL call_3_7_8_0(SH2_struct* sh)
{
  SH2cmpeq(sh,7,8);
}
#define init_3_7_9_0
void FASTCALL call_3_7_9_0(SH2_struct* sh)
{
  SH2cmpeq(sh,7,9);
}
#define init_3_7_10_0
void FASTCALL call_3_7_10_0(SH2_struct* sh)
{
  SH2cmpeq(sh,7,10);
}
#define init_3_7_11_0
void FASTCALL call_3_7_11_0(SH2_struct* sh)
{
  SH2cmpeq(sh,7,11);
}
#define init_3_7_12_0
void FASTCALL call_3_7_12_0(SH2_struct* sh)
{
  SH2cmpeq(sh,7,12);
}
#define init_3_7_13_0
void FASTCALL call_3_7_13_0(SH2_struct* sh)
{
  SH2cmpeq(sh,7,13);
}
#define init_3_7_14_0
void FASTCALL call_3_7_14_0(SH2_struct* sh)
{
  SH2cmpeq(sh,7,14);
}
#define init_3_7_15_0
void FASTCALL call_3_7_15_0(SH2_struct* sh)
{
  SH2cmpeq(sh,7,15);
}
#define init_3_8_0_0
void FASTCALL call_3_8_0_0(SH2_struct* sh)
{
  SH2cmpeq(sh,8,0);
}
#define init_3_8_1_0
void FASTCALL call_3_8_1_0(SH2_struct* sh)
{
  SH2cmpeq(sh,8,1);
}
#define init_3_8_2_0
void FASTCALL call_3_8_2_0(SH2_struct* sh)
{
  SH2cmpeq(sh,8,2);
}
#define init_3_8_3_0
void FASTCALL call_3_8_3_0(SH2_struct* sh)
{
  SH2cmpeq(sh,8,3);
}
#define init_3_8_4_0
void FASTCALL call_3_8_4_0(SH2_struct* sh)
{
  SH2cmpeq(sh,8,4);
}
#define init_3_8_5_0
void FASTCALL call_3_8_5_0(SH2_struct* sh)
{
  SH2cmpeq(sh,8,5);
}
#define init_3_8_6_0
void FASTCALL call_3_8_6_0(SH2_struct* sh)
{
  SH2cmpeq(sh,8,6);
}
#define init_3_8_7_0
void FASTCALL call_3_8_7_0(SH2_struct* sh)
{
  SH2cmpeq(sh,8,7);
}
#define init_3_8_8_0
void FASTCALL call_3_8_8_0(SH2_struct* sh)
{
  SH2cmpeq(sh,8,8);
}
#define init_3_8_9_0
void FASTCALL call_3_8_9_0(SH2_struct* sh)
{
  SH2cmpeq(sh,8,9);
}
#define init_3_8_10_0
void FASTCALL call_3_8_10_0(SH2_struct* sh)
{
  SH2cmpeq(sh,8,10);
}
#define init_3_8_11_0
void FASTCALL call_3_8_11_0(SH2_struct* sh)
{
  SH2cmpeq(sh,8,11);
}
#define init_3_8_12_0
void FASTCALL call_3_8_12_0(SH2_struct* sh)
{
  SH2cmpeq(sh,8,12);
}
#define init_3_8_13_0
void FASTCALL call_3_8_13_0(SH2_struct* sh)
{
  SH2cmpeq(sh,8,13);
}
#define init_3_8_14_0
void FASTCALL call_3_8_14_0(SH2_struct* sh)
{
  SH2cmpeq(sh,8,14);
}
#define init_3_8_15_0
void FASTCALL call_3_8_15_0(SH2_struct* sh)
{
  SH2cmpeq(sh,8,15);
}
#define init_3_9_0_0
void FASTCALL call_3_9_0_0(SH2_struct* sh)
{
  SH2cmpeq(sh,9,0);
}
#define init_3_9_1_0
void FASTCALL call_3_9_1_0(SH2_struct* sh)
{
  SH2cmpeq(sh,9,1);
}
#define init_3_9_2_0
void FASTCALL call_3_9_2_0(SH2_struct* sh)
{
  SH2cmpeq(sh,9,2);
}
#define init_3_9_3_0
void FASTCALL call_3_9_3_0(SH2_struct* sh)
{
  SH2cmpeq(sh,9,3);
}
#define init_3_9_4_0
void FASTCALL call_3_9_4_0(SH2_struct* sh)
{
  SH2cmpeq(sh,9,4);
}
#define init_3_9_5_0
void FASTCALL call_3_9_5_0(SH2_struct* sh)
{
  SH2cmpeq(sh,9,5);
}
#define init_3_9_6_0
void FASTCALL call_3_9_6_0(SH2_struct* sh)
{
  SH2cmpeq(sh,9,6);
}
#define init_3_9_7_0
void FASTCALL call_3_9_7_0(SH2_struct* sh)
{
  SH2cmpeq(sh,9,7);
}
#define init_3_9_8_0
void FASTCALL call_3_9_8_0(SH2_struct* sh)
{
  SH2cmpeq(sh,9,8);
}
#define init_3_9_9_0
void FASTCALL call_3_9_9_0(SH2_struct* sh)
{
  SH2cmpeq(sh,9,9);
}
#define init_3_9_10_0
void FASTCALL call_3_9_10_0(SH2_struct* sh)
{
  SH2cmpeq(sh,9,10);
}
#define init_3_9_11_0
void FASTCALL call_3_9_11_0(SH2_struct* sh)
{
  SH2cmpeq(sh,9,11);
}
#define init_3_9_12_0
void FASTCALL call_3_9_12_0(SH2_struct* sh)
{
  SH2cmpeq(sh,9,12);
}
#define init_3_9_13_0
void FASTCALL call_3_9_13_0(SH2_struct* sh)
{
  SH2cmpeq(sh,9,13);
}
#define init_3_9_14_0
void FASTCALL call_3_9_14_0(SH2_struct* sh)
{
  SH2cmpeq(sh,9,14);
}
#define init_3_9_15_0
void FASTCALL call_3_9_15_0(SH2_struct* sh)
{
  SH2cmpeq(sh,9,15);
}
#define init_3_10_0_0
void FASTCALL call_3_10_0_0(SH2_struct* sh)
{
  SH2cmpeq(sh,10,0);
}
#define init_3_10_1_0
void FASTCALL call_3_10_1_0(SH2_struct* sh)
{
  SH2cmpeq(sh,10,1);
}
#define init_3_10_2_0
void FASTCALL call_3_10_2_0(SH2_struct* sh)
{
  SH2cmpeq(sh,10,2);
}
#define init_3_10_3_0
void FASTCALL call_3_10_3_0(SH2_struct* sh)
{
  SH2cmpeq(sh,10,3);
}
#define init_3_10_4_0
void FASTCALL call_3_10_4_0(SH2_struct* sh)
{
  SH2cmpeq(sh,10,4);
}
#define init_3_10_5_0
void FASTCALL call_3_10_5_0(SH2_struct* sh)
{
  SH2cmpeq(sh,10,5);
}
#define init_3_10_6_0
void FASTCALL call_3_10_6_0(SH2_struct* sh)
{
  SH2cmpeq(sh,10,6);
}
#define init_3_10_7_0
void FASTCALL call_3_10_7_0(SH2_struct* sh)
{
  SH2cmpeq(sh,10,7);
}
#define init_3_10_8_0
void FASTCALL call_3_10_8_0(SH2_struct* sh)
{
  SH2cmpeq(sh,10,8);
}
#define init_3_10_9_0
void FASTCALL call_3_10_9_0(SH2_struct* sh)
{
  SH2cmpeq(sh,10,9);
}
#define init_3_10_10_0
void FASTCALL call_3_10_10_0(SH2_struct* sh)
{
  SH2cmpeq(sh,10,10);
}
#define init_3_10_11_0
void FASTCALL call_3_10_11_0(SH2_struct* sh)
{
  SH2cmpeq(sh,10,11);
}
#define init_3_10_12_0
void FASTCALL call_3_10_12_0(SH2_struct* sh)
{
  SH2cmpeq(sh,10,12);
}
#define init_3_10_13_0
void FASTCALL call_3_10_13_0(SH2_struct* sh)
{
  SH2cmpeq(sh,10,13);
}
#define init_3_10_14_0
void FASTCALL call_3_10_14_0(SH2_struct* sh)
{
  SH2cmpeq(sh,10,14);
}
#define init_3_10_15_0
void FASTCALL call_3_10_15_0(SH2_struct* sh)
{
  SH2cmpeq(sh,10,15);
}
#define init_3_11_0_0
void FASTCALL call_3_11_0_0(SH2_struct* sh)
{
  SH2cmpeq(sh,11,0);
}
#define init_3_11_1_0
void FASTCALL call_3_11_1_0(SH2_struct* sh)
{
  SH2cmpeq(sh,11,1);
}
#define init_3_11_2_0
void FASTCALL call_3_11_2_0(SH2_struct* sh)
{
  SH2cmpeq(sh,11,2);
}
#define init_3_11_3_0
void FASTCALL call_3_11_3_0(SH2_struct* sh)
{
  SH2cmpeq(sh,11,3);
}
#define init_3_11_4_0
void FASTCALL call_3_11_4_0(SH2_struct* sh)
{
  SH2cmpeq(sh,11,4);
}
#define init_3_11_5_0
void FASTCALL call_3_11_5_0(SH2_struct* sh)
{
  SH2cmpeq(sh,11,5);
}
#define init_3_11_6_0
void FASTCALL call_3_11_6_0(SH2_struct* sh)
{
  SH2cmpeq(sh,11,6);
}
#define init_3_11_7_0
void FASTCALL call_3_11_7_0(SH2_struct* sh)
{
  SH2cmpeq(sh,11,7);
}
#define init_3_11_8_0
void FASTCALL call_3_11_8_0(SH2_struct* sh)
{
  SH2cmpeq(sh,11,8);
}
#define init_3_11_9_0
void FASTCALL call_3_11_9_0(SH2_struct* sh)
{
  SH2cmpeq(sh,11,9);
}
#define init_3_11_10_0
void FASTCALL call_3_11_10_0(SH2_struct* sh)
{
  SH2cmpeq(sh,11,10);
}
#define init_3_11_11_0
void FASTCALL call_3_11_11_0(SH2_struct* sh)
{
  SH2cmpeq(sh,11,11);
}
#define init_3_11_12_0
void FASTCALL call_3_11_12_0(SH2_struct* sh)
{
  SH2cmpeq(sh,11,12);
}
#define init_3_11_13_0
void FASTCALL call_3_11_13_0(SH2_struct* sh)
{
  SH2cmpeq(sh,11,13);
}
#define init_3_11_14_0
void FASTCALL call_3_11_14_0(SH2_struct* sh)
{
  SH2cmpeq(sh,11,14);
}
#define init_3_11_15_0
void FASTCALL call_3_11_15_0(SH2_struct* sh)
{
  SH2cmpeq(sh,11,15);
}
#define init_3_12_0_0
void FASTCALL call_3_12_0_0(SH2_struct* sh)
{
  SH2cmpeq(sh,12,0);
}
#define init_3_12_1_0
void FASTCALL call_3_12_1_0(SH2_struct* sh)
{
  SH2cmpeq(sh,12,1);
}
#define init_3_12_2_0
void FASTCALL call_3_12_2_0(SH2_struct* sh)
{
  SH2cmpeq(sh,12,2);
}
#define init_3_12_3_0
void FASTCALL call_3_12_3_0(SH2_struct* sh)
{
  SH2cmpeq(sh,12,3);
}
#define init_3_12_4_0
void FASTCALL call_3_12_4_0(SH2_struct* sh)
{
  SH2cmpeq(sh,12,4);
}
#define init_3_12_5_0
void FASTCALL call_3_12_5_0(SH2_struct* sh)
{
  SH2cmpeq(sh,12,5);
}
#define init_3_12_6_0
void FASTCALL call_3_12_6_0(SH2_struct* sh)
{
  SH2cmpeq(sh,12,6);
}
#define init_3_12_7_0
void FASTCALL call_3_12_7_0(SH2_struct* sh)
{
  SH2cmpeq(sh,12,7);
}
#define init_3_12_8_0
void FASTCALL call_3_12_8_0(SH2_struct* sh)
{
  SH2cmpeq(sh,12,8);
}
#define init_3_12_9_0
void FASTCALL call_3_12_9_0(SH2_struct* sh)
{
  SH2cmpeq(sh,12,9);
}
#define init_3_12_10_0
void FASTCALL call_3_12_10_0(SH2_struct* sh)
{
  SH2cmpeq(sh,12,10);
}
#define init_3_12_11_0
void FASTCALL call_3_12_11_0(SH2_struct* sh)
{
  SH2cmpeq(sh,12,11);
}
#define init_3_12_12_0
void FASTCALL call_3_12_12_0(SH2_struct* sh)
{
  SH2cmpeq(sh,12,12);
}
#define init_3_12_13_0
void FASTCALL call_3_12_13_0(SH2_struct* sh)
{
  SH2cmpeq(sh,12,13);
}
#define init_3_12_14_0
void FASTCALL call_3_12_14_0(SH2_struct* sh)
{
  SH2cmpeq(sh,12,14);
}
#define init_3_12_15_0
void FASTCALL call_3_12_15_0(SH2_struct* sh)
{
  SH2cmpeq(sh,12,15);
}
#define init_3_13_0_0
void FASTCALL call_3_13_0_0(SH2_struct* sh)
{
  SH2cmpeq(sh,13,0);
}
#define init_3_13_1_0
void FASTCALL call_3_13_1_0(SH2_struct* sh)
{
  SH2cmpeq(sh,13,1);
}
#define init_3_13_2_0
void FASTCALL call_3_13_2_0(SH2_struct* sh)
{
  SH2cmpeq(sh,13,2);
}
#define init_3_13_3_0
void FASTCALL call_3_13_3_0(SH2_struct* sh)
{
  SH2cmpeq(sh,13,3);
}
#define init_3_13_4_0
void FASTCALL call_3_13_4_0(SH2_struct* sh)
{
  SH2cmpeq(sh,13,4);
}
#define init_3_13_5_0
void FASTCALL call_3_13_5_0(SH2_struct* sh)
{
  SH2cmpeq(sh,13,5);
}
#define init_3_13_6_0
void FASTCALL call_3_13_6_0(SH2_struct* sh)
{
  SH2cmpeq(sh,13,6);
}
#define init_3_13_7_0
void FASTCALL call_3_13_7_0(SH2_struct* sh)
{
  SH2cmpeq(sh,13,7);
}
#define init_3_13_8_0
void FASTCALL call_3_13_8_0(SH2_struct* sh)
{
  SH2cmpeq(sh,13,8);
}
#define init_3_13_9_0
void FASTCALL call_3_13_9_0(SH2_struct* sh)
{
  SH2cmpeq(sh,13,9);
}
#define init_3_13_10_0
void FASTCALL call_3_13_10_0(SH2_struct* sh)
{
  SH2cmpeq(sh,13,10);
}
#define init_3_13_11_0
void FASTCALL call_3_13_11_0(SH2_struct* sh)
{
  SH2cmpeq(sh,13,11);
}
#define init_3_13_12_0
void FASTCALL call_3_13_12_0(SH2_struct* sh)
{
  SH2cmpeq(sh,13,12);
}
#define init_3_13_13_0
void FASTCALL call_3_13_13_0(SH2_struct* sh)
{
  SH2cmpeq(sh,13,13);
}
#define init_3_13_14_0
void FASTCALL call_3_13_14_0(SH2_struct* sh)
{
  SH2cmpeq(sh,13,14);
}
#define init_3_13_15_0
void FASTCALL call_3_13_15_0(SH2_struct* sh)
{
  SH2cmpeq(sh,13,15);
}
#define init_3_14_0_0
void FASTCALL call_3_14_0_0(SH2_struct* sh)
{
  SH2cmpeq(sh,14,0);
}
#define init_3_14_1_0
void FASTCALL call_3_14_1_0(SH2_struct* sh)
{
  SH2cmpeq(sh,14,1);
}
#define init_3_14_2_0
void FASTCALL call_3_14_2_0(SH2_struct* sh)
{
  SH2cmpeq(sh,14,2);
}
#define init_3_14_3_0
void FASTCALL call_3_14_3_0(SH2_struct* sh)
{
  SH2cmpeq(sh,14,3);
}
#define init_3_14_4_0
void FASTCALL call_3_14_4_0(SH2_struct* sh)
{
  SH2cmpeq(sh,14,4);
}
#define init_3_14_5_0
void FASTCALL call_3_14_5_0(SH2_struct* sh)
{
  SH2cmpeq(sh,14,5);
}
#define init_3_14_6_0
void FASTCALL call_3_14_6_0(SH2_struct* sh)
{
  SH2cmpeq(sh,14,6);
}
#define init_3_14_7_0
void FASTCALL call_3_14_7_0(SH2_struct* sh)
{
  SH2cmpeq(sh,14,7);
}
#define init_3_14_8_0
void FASTCALL call_3_14_8_0(SH2_struct* sh)
{
  SH2cmpeq(sh,14,8);
}
#define init_3_14_9_0
void FASTCALL call_3_14_9_0(SH2_struct* sh)
{
  SH2cmpeq(sh,14,9);
}
#define init_3_14_10_0
void FASTCALL call_3_14_10_0(SH2_struct* sh)
{
  SH2cmpeq(sh,14,10);
}
#define init_3_14_11_0
void FASTCALL call_3_14_11_0(SH2_struct* sh)
{
  SH2cmpeq(sh,14,11);
}
#define init_3_14_12_0
void FASTCALL call_3_14_12_0(SH2_struct* sh)
{
  SH2cmpeq(sh,14,12);
}
#define init_3_14_13_0
void FASTCALL call_3_14_13_0(SH2_struct* sh)
{
  SH2cmpeq(sh,14,13);
}
#define init_3_14_14_0
void FASTCALL call_3_14_14_0(SH2_struct* sh)
{
  SH2cmpeq(sh,14,14);
}
#define init_3_14_15_0
void FASTCALL call_3_14_15_0(SH2_struct* sh)
{
  SH2cmpeq(sh,14,15);
}
#define init_3_15_0_0
void FASTCALL call_3_15_0_0(SH2_struct* sh)
{
  SH2cmpeq(sh,15,0);
}
#define init_3_15_1_0
void FASTCALL call_3_15_1_0(SH2_struct* sh)
{
  SH2cmpeq(sh,15,1);
}
#define init_3_15_2_0
void FASTCALL call_3_15_2_0(SH2_struct* sh)
{
  SH2cmpeq(sh,15,2);
}
#define init_3_15_3_0
void FASTCALL call_3_15_3_0(SH2_struct* sh)
{
  SH2cmpeq(sh,15,3);
}
#define init_3_15_4_0
void FASTCALL call_3_15_4_0(SH2_struct* sh)
{
  SH2cmpeq(sh,15,4);
}
#define init_3_15_5_0
void FASTCALL call_3_15_5_0(SH2_struct* sh)
{
  SH2cmpeq(sh,15,5);
}
#define init_3_15_6_0
void FASTCALL call_3_15_6_0(SH2_struct* sh)
{
  SH2cmpeq(sh,15,6);
}
#define init_3_15_7_0
void FASTCALL call_3_15_7_0(SH2_struct* sh)
{
  SH2cmpeq(sh,15,7);
}
#define init_3_15_8_0
void FASTCALL call_3_15_8_0(SH2_struct* sh)
{
  SH2cmpeq(sh,15,8);
}
#define init_3_15_9_0
void FASTCALL call_3_15_9_0(SH2_struct* sh)
{
  SH2cmpeq(sh,15,9);
}
#define init_3_15_10_0
void FASTCALL call_3_15_10_0(SH2_struct* sh)
{
  SH2cmpeq(sh,15,10);
}
#define init_3_15_11_0
void FASTCALL call_3_15_11_0(SH2_struct* sh)
{
  SH2cmpeq(sh,15,11);
}
#define init_3_15_12_0
void FASTCALL call_3_15_12_0(SH2_struct* sh)
{
  SH2cmpeq(sh,15,12);
}
#define init_3_15_13_0
void FASTCALL call_3_15_13_0(SH2_struct* sh)
{
  SH2cmpeq(sh,15,13);
}
#define init_3_15_14_0
void FASTCALL call_3_15_14_0(SH2_struct* sh)
{
  SH2cmpeq(sh,15,14);
}
#define init_3_15_15_0
void FASTCALL call_3_15_15_0(SH2_struct* sh)
{
  SH2cmpeq(sh,15,15);
}
#define init_8_13_0_0
void FASTCALL call_8_13_0_0(SH2_struct* sh)
{
  SH2bts(sh,(0<<4)+0);
}
#define init_8_13_0_1
void FASTCALL call_8_13_0_1(SH2_struct* sh)
{
  SH2bts(sh,(0<<4)+1);
}
#define init_8_13_0_2
void FASTCALL call_8_13_0_2(SH2_struct* sh)
{
  SH2bts(sh,(0<<4)+2);
}
#define init_8_13_0_3
void FASTCALL call_8_13_0_3(SH2_struct* sh)
{
  SH2bts(sh,(0<<4)+3);
}
#define init_8_13_0_4
void FASTCALL call_8_13_0_4(SH2_struct* sh)
{
  SH2bts(sh,(0<<4)+4);
}
#define init_8_13_0_5
void FASTCALL call_8_13_0_5(SH2_struct* sh)
{
  SH2bts(sh,(0<<4)+5);
}
#define init_8_13_0_6
void FASTCALL call_8_13_0_6(SH2_struct* sh)
{
  SH2bts(sh,(0<<4)+6);
}
#define init_8_13_0_7
void FASTCALL call_8_13_0_7(SH2_struct* sh)
{
  SH2bts(sh,(0<<4)+7);
}
#define init_8_13_0_8
void FASTCALL call_8_13_0_8(SH2_struct* sh)
{
  SH2bts(sh,(0<<4)+8);
}
#define init_8_13_0_9
void FASTCALL call_8_13_0_9(SH2_struct* sh)
{
  SH2bts(sh,(0<<4)+9);
}
#define init_8_13_0_10
void FASTCALL call_8_13_0_10(SH2_struct* sh)
{
  SH2bts(sh,(0<<4)+10);
}
#define init_8_13_0_11
void FASTCALL call_8_13_0_11(SH2_struct* sh)
{
  SH2bts(sh,(0<<4)+11);
}
#define init_8_13_0_12
void FASTCALL call_8_13_0_12(SH2_struct* sh)
{
  SH2bts(sh,(0<<4)+12);
}
#define init_8_13_0_13
void FASTCALL call_8_13_0_13(SH2_struct* sh)
{
  SH2bts(sh,(0<<4)+13);
}
#define init_8_13_0_14
void FASTCALL call_8_13_0_14(SH2_struct* sh)
{
  SH2bts(sh,(0<<4)+14);
}
#define init_8_13_0_15
void FASTCALL call_8_13_0_15(SH2_struct* sh)
{
  SH2bts(sh,(0<<4)+15);
}
#define init_8_13_1_0
void FASTCALL call_8_13_1_0(SH2_struct* sh)
{
  SH2bts(sh,(1<<4)+0);
}
#define init_8_13_1_1
void FASTCALL call_8_13_1_1(SH2_struct* sh)
{
  SH2bts(sh,(1<<4)+1);
}
#define init_8_13_1_2
void FASTCALL call_8_13_1_2(SH2_struct* sh)
{
  SH2bts(sh,(1<<4)+2);
}
#define init_8_13_1_3
void FASTCALL call_8_13_1_3(SH2_struct* sh)
{
  SH2bts(sh,(1<<4)+3);
}
#define init_8_13_1_4
void FASTCALL call_8_13_1_4(SH2_struct* sh)
{
  SH2bts(sh,(1<<4)+4);
}
#define init_8_13_1_5
void FASTCALL call_8_13_1_5(SH2_struct* sh)
{
  SH2bts(sh,(1<<4)+5);
}
#define init_8_13_1_6
void FASTCALL call_8_13_1_6(SH2_struct* sh)
{
  SH2bts(sh,(1<<4)+6);
}
#define init_8_13_1_7
void FASTCALL call_8_13_1_7(SH2_struct* sh)
{
  SH2bts(sh,(1<<4)+7);
}
#define init_8_13_1_8
void FASTCALL call_8_13_1_8(SH2_struct* sh)
{
  SH2bts(sh,(1<<4)+8);
}
#define init_8_13_1_9
void FASTCALL call_8_13_1_9(SH2_struct* sh)
{
  SH2bts(sh,(1<<4)+9);
}
#define init_8_13_1_10
void FASTCALL call_8_13_1_10(SH2_struct* sh)
{
  SH2bts(sh,(1<<4)+10);
}
#define init_8_13_1_11
void FASTCALL call_8_13_1_11(SH2_struct* sh)
{
  SH2bts(sh,(1<<4)+11);
}
#define init_8_13_1_12
void FASTCALL call_8_13_1_12(SH2_struct* sh)
{
  SH2bts(sh,(1<<4)+12);
}
#define init_8_13_1_13
void FASTCALL call_8_13_1_13(SH2_struct* sh)
{
  SH2bts(sh,(1<<4)+13);
}
#define init_8_13_1_14
void FASTCALL call_8_13_1_14(SH2_struct* sh)
{
  SH2bts(sh,(1<<4)+14);
}
#define init_8_13_1_15
void FASTCALL call_8_13_1_15(SH2_struct* sh)
{
  SH2bts(sh,(1<<4)+15);
}
#define init_8_13_2_0
void FASTCALL call_8_13_2_0(SH2_struct* sh)
{
  SH2bts(sh,(2<<4)+0);
}
#define init_8_13_2_1
void FASTCALL call_8_13_2_1(SH2_struct* sh)
{
  SH2bts(sh,(2<<4)+1);
}
#define init_8_13_2_2
void FASTCALL call_8_13_2_2(SH2_struct* sh)
{
  SH2bts(sh,(2<<4)+2);
}
#define init_8_13_2_3
void FASTCALL call_8_13_2_3(SH2_struct* sh)
{
  SH2bts(sh,(2<<4)+3);
}
#define init_8_13_2_4
void FASTCALL call_8_13_2_4(SH2_struct* sh)
{
  SH2bts(sh,(2<<4)+4);
}
#define init_8_13_2_5
void FASTCALL call_8_13_2_5(SH2_struct* sh)
{
  SH2bts(sh,(2<<4)+5);
}
#define init_8_13_2_6
void FASTCALL call_8_13_2_6(SH2_struct* sh)
{
  SH2bts(sh,(2<<4)+6);
}
#define init_8_13_2_7
void FASTCALL call_8_13_2_7(SH2_struct* sh)
{
  SH2bts(sh,(2<<4)+7);
}
#define init_8_13_2_8
void FASTCALL call_8_13_2_8(SH2_struct* sh)
{
  SH2bts(sh,(2<<4)+8);
}
#define init_8_13_2_9
void FASTCALL call_8_13_2_9(SH2_struct* sh)
{
  SH2bts(sh,(2<<4)+9);
}
#define init_8_13_2_10
void FASTCALL call_8_13_2_10(SH2_struct* sh)
{
  SH2bts(sh,(2<<4)+10);
}
#define init_8_13_2_11
void FASTCALL call_8_13_2_11(SH2_struct* sh)
{
  SH2bts(sh,(2<<4)+11);
}
#define init_8_13_2_12
void FASTCALL call_8_13_2_12(SH2_struct* sh)
{
  SH2bts(sh,(2<<4)+12);
}
#define init_8_13_2_13
void FASTCALL call_8_13_2_13(SH2_struct* sh)
{
  SH2bts(sh,(2<<4)+13);
}
#define init_8_13_2_14
void FASTCALL call_8_13_2_14(SH2_struct* sh)
{
  SH2bts(sh,(2<<4)+14);
}
#define init_8_13_2_15
void FASTCALL call_8_13_2_15(SH2_struct* sh)
{
  SH2bts(sh,(2<<4)+15);
}
#define init_8_13_3_0
void FASTCALL call_8_13_3_0(SH2_struct* sh)
{
  SH2bts(sh,(3<<4)+0);
}
#define init_8_13_3_1
void FASTCALL call_8_13_3_1(SH2_struct* sh)
{
  SH2bts(sh,(3<<4)+1);
}
#define init_8_13_3_2
void FASTCALL call_8_13_3_2(SH2_struct* sh)
{
  SH2bts(sh,(3<<4)+2);
}
#define init_8_13_3_3
void FASTCALL call_8_13_3_3(SH2_struct* sh)
{
  SH2bts(sh,(3<<4)+3);
}
#define init_8_13_3_4
void FASTCALL call_8_13_3_4(SH2_struct* sh)
{
  SH2bts(sh,(3<<4)+4);
}
#define init_8_13_3_5
void FASTCALL call_8_13_3_5(SH2_struct* sh)
{
  SH2bts(sh,(3<<4)+5);
}
#define init_8_13_3_6
void FASTCALL call_8_13_3_6(SH2_struct* sh)
{
  SH2bts(sh,(3<<4)+6);
}
#define init_8_13_3_7
void FASTCALL call_8_13_3_7(SH2_struct* sh)
{
  SH2bts(sh,(3<<4)+7);
}
#define init_8_13_3_8
void FASTCALL call_8_13_3_8(SH2_struct* sh)
{
  SH2bts(sh,(3<<4)+8);
}
#define init_8_13_3_9
void FASTCALL call_8_13_3_9(SH2_struct* sh)
{
  SH2bts(sh,(3<<4)+9);
}
#define init_8_13_3_10
void FASTCALL call_8_13_3_10(SH2_struct* sh)
{
  SH2bts(sh,(3<<4)+10);
}
#define init_8_13_3_11
void FASTCALL call_8_13_3_11(SH2_struct* sh)
{
  SH2bts(sh,(3<<4)+11);
}
#define init_8_13_3_12
void FASTCALL call_8_13_3_12(SH2_struct* sh)
{
  SH2bts(sh,(3<<4)+12);
}
#define init_8_13_3_13
void FASTCALL call_8_13_3_13(SH2_struct* sh)
{
  SH2bts(sh,(3<<4)+13);
}
#define init_8_13_3_14
void FASTCALL call_8_13_3_14(SH2_struct* sh)
{
  SH2bts(sh,(3<<4)+14);
}
#define init_8_13_3_15
void FASTCALL call_8_13_3_15(SH2_struct* sh)
{
  SH2bts(sh,(3<<4)+15);
}
#define init_8_13_4_0
void FASTCALL call_8_13_4_0(SH2_struct* sh)
{
  SH2bts(sh,(4<<4)+0);
}
#define init_8_13_4_1
void FASTCALL call_8_13_4_1(SH2_struct* sh)
{
  SH2bts(sh,(4<<4)+1);
}
#define init_8_13_4_2
void FASTCALL call_8_13_4_2(SH2_struct* sh)
{
  SH2bts(sh,(4<<4)+2);
}
#define init_8_13_4_3
void FASTCALL call_8_13_4_3(SH2_struct* sh)
{
  SH2bts(sh,(4<<4)+3);
}
#define init_8_13_4_4
void FASTCALL call_8_13_4_4(SH2_struct* sh)
{
  SH2bts(sh,(4<<4)+4);
}
#define init_8_13_4_5
void FASTCALL call_8_13_4_5(SH2_struct* sh)
{
  SH2bts(sh,(4<<4)+5);
}
#define init_8_13_4_6
void FASTCALL call_8_13_4_6(SH2_struct* sh)
{
  SH2bts(sh,(4<<4)+6);
}
#define init_8_13_4_7
void FASTCALL call_8_13_4_7(SH2_struct* sh)
{
  SH2bts(sh,(4<<4)+7);
}
#define init_8_13_4_8
void FASTCALL call_8_13_4_8(SH2_struct* sh)
{
  SH2bts(sh,(4<<4)+8);
}
#define init_8_13_4_9
void FASTCALL call_8_13_4_9(SH2_struct* sh)
{
  SH2bts(sh,(4<<4)+9);
}
#define init_8_13_4_10
void FASTCALL call_8_13_4_10(SH2_struct* sh)
{
  SH2bts(sh,(4<<4)+10);
}
#define init_8_13_4_11
void FASTCALL call_8_13_4_11(SH2_struct* sh)
{
  SH2bts(sh,(4<<4)+11);
}
#define init_8_13_4_12
void FASTCALL call_8_13_4_12(SH2_struct* sh)
{
  SH2bts(sh,(4<<4)+12);
}
#define init_8_13_4_13
void FASTCALL call_8_13_4_13(SH2_struct* sh)
{
  SH2bts(sh,(4<<4)+13);
}
#define init_8_13_4_14
void FASTCALL call_8_13_4_14(SH2_struct* sh)
{
  SH2bts(sh,(4<<4)+14);
}
#define init_8_13_4_15
void FASTCALL call_8_13_4_15(SH2_struct* sh)
{
  SH2bts(sh,(4<<4)+15);
}
#define init_8_13_5_0
void FASTCALL call_8_13_5_0(SH2_struct* sh)
{
  SH2bts(sh,(5<<4)+0);
}
#define init_8_13_5_1
void FASTCALL call_8_13_5_1(SH2_struct* sh)
{
  SH2bts(sh,(5<<4)+1);
}
#define init_8_13_5_2
void FASTCALL call_8_13_5_2(SH2_struct* sh)
{
  SH2bts(sh,(5<<4)+2);
}
#define init_8_13_5_3
void FASTCALL call_8_13_5_3(SH2_struct* sh)
{
  SH2bts(sh,(5<<4)+3);
}
#define init_8_13_5_4
void FASTCALL call_8_13_5_4(SH2_struct* sh)
{
  SH2bts(sh,(5<<4)+4);
}
#define init_8_13_5_5
void FASTCALL call_8_13_5_5(SH2_struct* sh)
{
  SH2bts(sh,(5<<4)+5);
}
#define init_8_13_5_6
void FASTCALL call_8_13_5_6(SH2_struct* sh)
{
  SH2bts(sh,(5<<4)+6);
}
#define init_8_13_5_7
void FASTCALL call_8_13_5_7(SH2_struct* sh)
{
  SH2bts(sh,(5<<4)+7);
}
#define init_8_13_5_8
void FASTCALL call_8_13_5_8(SH2_struct* sh)
{
  SH2bts(sh,(5<<4)+8);
}
#define init_8_13_5_9
void FASTCALL call_8_13_5_9(SH2_struct* sh)
{
  SH2bts(sh,(5<<4)+9);
}
#define init_8_13_5_10
void FASTCALL call_8_13_5_10(SH2_struct* sh)
{
  SH2bts(sh,(5<<4)+10);
}
#define init_8_13_5_11
void FASTCALL call_8_13_5_11(SH2_struct* sh)
{
  SH2bts(sh,(5<<4)+11);
}
#define init_8_13_5_12
void FASTCALL call_8_13_5_12(SH2_struct* sh)
{
  SH2bts(sh,(5<<4)+12);
}
#define init_8_13_5_13
void FASTCALL call_8_13_5_13(SH2_struct* sh)
{
  SH2bts(sh,(5<<4)+13);
}
#define init_8_13_5_14
void FASTCALL call_8_13_5_14(SH2_struct* sh)
{
  SH2bts(sh,(5<<4)+14);
}
#define init_8_13_5_15
void FASTCALL call_8_13_5_15(SH2_struct* sh)
{
  SH2bts(sh,(5<<4)+15);
}
#define init_8_13_6_0
void FASTCALL call_8_13_6_0(SH2_struct* sh)
{
  SH2bts(sh,(6<<4)+0);
}
#define init_8_13_6_1
void FASTCALL call_8_13_6_1(SH2_struct* sh)
{
  SH2bts(sh,(6<<4)+1);
}
#define init_8_13_6_2
void FASTCALL call_8_13_6_2(SH2_struct* sh)
{
  SH2bts(sh,(6<<4)+2);
}
#define init_8_13_6_3
void FASTCALL call_8_13_6_3(SH2_struct* sh)
{
  SH2bts(sh,(6<<4)+3);
}
#define init_8_13_6_4
void FASTCALL call_8_13_6_4(SH2_struct* sh)
{
  SH2bts(sh,(6<<4)+4);
}
#define init_8_13_6_5
void FASTCALL call_8_13_6_5(SH2_struct* sh)
{
  SH2bts(sh,(6<<4)+5);
}
#define init_8_13_6_6
void FASTCALL call_8_13_6_6(SH2_struct* sh)
{
  SH2bts(sh,(6<<4)+6);
}
#define init_8_13_6_7
void FASTCALL call_8_13_6_7(SH2_struct* sh)
{
  SH2bts(sh,(6<<4)+7);
}
#define init_8_13_6_8
void FASTCALL call_8_13_6_8(SH2_struct* sh)
{
  SH2bts(sh,(6<<4)+8);
}
#define init_8_13_6_9
void FASTCALL call_8_13_6_9(SH2_struct* sh)
{
  SH2bts(sh,(6<<4)+9);
}
#define init_8_13_6_10
void FASTCALL call_8_13_6_10(SH2_struct* sh)
{
  SH2bts(sh,(6<<4)+10);
}
#define init_8_13_6_11
void FASTCALL call_8_13_6_11(SH2_struct* sh)
{
  SH2bts(sh,(6<<4)+11);
}
#define init_8_13_6_12
void FASTCALL call_8_13_6_12(SH2_struct* sh)
{
  SH2bts(sh,(6<<4)+12);
}
#define init_8_13_6_13
void FASTCALL call_8_13_6_13(SH2_struct* sh)
{
  SH2bts(sh,(6<<4)+13);
}
#define init_8_13_6_14
void FASTCALL call_8_13_6_14(SH2_struct* sh)
{
  SH2bts(sh,(6<<4)+14);
}
#define init_8_13_6_15
void FASTCALL call_8_13_6_15(SH2_struct* sh)
{
  SH2bts(sh,(6<<4)+15);
}
#define init_8_13_7_0
void FASTCALL call_8_13_7_0(SH2_struct* sh)
{
  SH2bts(sh,(7<<4)+0);
}
#define init_8_13_7_1
void FASTCALL call_8_13_7_1(SH2_struct* sh)
{
  SH2bts(sh,(7<<4)+1);
}
#define init_8_13_7_2
void FASTCALL call_8_13_7_2(SH2_struct* sh)
{
  SH2bts(sh,(7<<4)+2);
}
#define init_8_13_7_3
void FASTCALL call_8_13_7_3(SH2_struct* sh)
{
  SH2bts(sh,(7<<4)+3);
}
#define init_8_13_7_4
void FASTCALL call_8_13_7_4(SH2_struct* sh)
{
  SH2bts(sh,(7<<4)+4);
}
#define init_8_13_7_5
void FASTCALL call_8_13_7_5(SH2_struct* sh)
{
  SH2bts(sh,(7<<4)+5);
}
#define init_8_13_7_6
void FASTCALL call_8_13_7_6(SH2_struct* sh)
{
  SH2bts(sh,(7<<4)+6);
}
#define init_8_13_7_7
void FASTCALL call_8_13_7_7(SH2_struct* sh)
{
  SH2bts(sh,(7<<4)+7);
}
#define init_8_13_7_8
void FASTCALL call_8_13_7_8(SH2_struct* sh)
{
  SH2bts(sh,(7<<4)+8);
}
#define init_8_13_7_9
void FASTCALL call_8_13_7_9(SH2_struct* sh)
{
  SH2bts(sh,(7<<4)+9);
}
#define init_8_13_7_10
void FASTCALL call_8_13_7_10(SH2_struct* sh)
{
  SH2bts(sh,(7<<4)+10);
}
#define init_8_13_7_11
void FASTCALL call_8_13_7_11(SH2_struct* sh)
{
  SH2bts(sh,(7<<4)+11);
}
#define init_8_13_7_12
void FASTCALL call_8_13_7_12(SH2_struct* sh)
{
  SH2bts(sh,(7<<4)+12);
}
#define init_8_13_7_13
void FASTCALL call_8_13_7_13(SH2_struct* sh)
{
  SH2bts(sh,(7<<4)+13);
}
#define init_8_13_7_14
void FASTCALL call_8_13_7_14(SH2_struct* sh)
{
  SH2bts(sh,(7<<4)+14);
}
#define init_8_13_7_15
void FASTCALL call_8_13_7_15(SH2_struct* sh)
{
  SH2bts(sh,(7<<4)+15);
}
#define init_8_13_8_0
void FASTCALL call_8_13_8_0(SH2_struct* sh)
{
  SH2bts(sh,(8<<4)+0);
}
#define init_8_13_8_1
void FASTCALL call_8_13_8_1(SH2_struct* sh)
{
  SH2bts(sh,(8<<4)+1);
}
#define init_8_13_8_2
void FASTCALL call_8_13_8_2(SH2_struct* sh)
{
  SH2bts(sh,(8<<4)+2);
}
#define init_8_13_8_3
void FASTCALL call_8_13_8_3(SH2_struct* sh)
{
  SH2bts(sh,(8<<4)+3);
}
#define init_8_13_8_4
void FASTCALL call_8_13_8_4(SH2_struct* sh)
{
  SH2bts(sh,(8<<4)+4);
}
#define init_8_13_8_5
void FASTCALL call_8_13_8_5(SH2_struct* sh)
{
  SH2bts(sh,(8<<4)+5);
}
#define init_8_13_8_6
void FASTCALL call_8_13_8_6(SH2_struct* sh)
{
  SH2bts(sh,(8<<4)+6);
}
#define init_8_13_8_7
void FASTCALL call_8_13_8_7(SH2_struct* sh)
{
  SH2bts(sh,(8<<4)+7);
}
#define init_8_13_8_8
void FASTCALL call_8_13_8_8(SH2_struct* sh)
{
  SH2bts(sh,(8<<4)+8);
}
#define init_8_13_8_9
void FASTCALL call_8_13_8_9(SH2_struct* sh)
{
  SH2bts(sh,(8<<4)+9);
}
#define init_8_13_8_10
void FASTCALL call_8_13_8_10(SH2_struct* sh)
{
  SH2bts(sh,(8<<4)+10);
}
#define init_8_13_8_11
void FASTCALL call_8_13_8_11(SH2_struct* sh)
{
  SH2bts(sh,(8<<4)+11);
}
#define init_8_13_8_12
void FASTCALL call_8_13_8_12(SH2_struct* sh)
{
  SH2bts(sh,(8<<4)+12);
}
#define init_8_13_8_13
void FASTCALL call_8_13_8_13(SH2_struct* sh)
{
  SH2bts(sh,(8<<4)+13);
}
#define init_8_13_8_14
void FASTCALL call_8_13_8_14(SH2_struct* sh)
{
  SH2bts(sh,(8<<4)+14);
}
#define init_8_13_8_15
void FASTCALL call_8_13_8_15(SH2_struct* sh)
{
  SH2bts(sh,(8<<4)+15);
}
#define init_8_13_9_0
void FASTCALL call_8_13_9_0(SH2_struct* sh)
{
  SH2bts(sh,(9<<4)+0);
}
#define init_8_13_9_1
void FASTCALL call_8_13_9_1(SH2_struct* sh)
{
  SH2bts(sh,(9<<4)+1);
}
#define init_8_13_9_2
void FASTCALL call_8_13_9_2(SH2_struct* sh)
{
  SH2bts(sh,(9<<4)+2);
}
#define init_8_13_9_3
void FASTCALL call_8_13_9_3(SH2_struct* sh)
{
  SH2bts(sh,(9<<4)+3);
}
#define init_8_13_9_4
void FASTCALL call_8_13_9_4(SH2_struct* sh)
{
  SH2bts(sh,(9<<4)+4);
}
#define init_8_13_9_5
void FASTCALL call_8_13_9_5(SH2_struct* sh)
{
  SH2bts(sh,(9<<4)+5);
}
#define init_8_13_9_6
void FASTCALL call_8_13_9_6(SH2_struct* sh)
{
  SH2bts(sh,(9<<4)+6);
}
#define init_8_13_9_7
void FASTCALL call_8_13_9_7(SH2_struct* sh)
{
  SH2bts(sh,(9<<4)+7);
}
#define init_8_13_9_8
void FASTCALL call_8_13_9_8(SH2_struct* sh)
{
  SH2bts(sh,(9<<4)+8);
}
#define init_8_13_9_9
void FASTCALL call_8_13_9_9(SH2_struct* sh)
{
  SH2bts(sh,(9<<4)+9);
}
#define init_8_13_9_10
void FASTCALL call_8_13_9_10(SH2_struct* sh)
{
  SH2bts(sh,(9<<4)+10);
}
#define init_8_13_9_11
void FASTCALL call_8_13_9_11(SH2_struct* sh)
{
  SH2bts(sh,(9<<4)+11);
}
#define init_8_13_9_12
void FASTCALL call_8_13_9_12(SH2_struct* sh)
{
  SH2bts(sh,(9<<4)+12);
}
#define init_8_13_9_13
void FASTCALL call_8_13_9_13(SH2_struct* sh)
{
  SH2bts(sh,(9<<4)+13);
}
#define init_8_13_9_14
void FASTCALL call_8_13_9_14(SH2_struct* sh)
{
  SH2bts(sh,(9<<4)+14);
}
#define init_8_13_9_15
void FASTCALL call_8_13_9_15(SH2_struct* sh)
{
  SH2bts(sh,(9<<4)+15);
}
#define init_8_13_10_0
void FASTCALL call_8_13_10_0(SH2_struct* sh)
{
  SH2bts(sh,(10<<4)+0);
}
#define init_8_13_10_1
void FASTCALL call_8_13_10_1(SH2_struct* sh)
{
  SH2bts(sh,(10<<4)+1);
}
#define init_8_13_10_2
void FASTCALL call_8_13_10_2(SH2_struct* sh)
{
  SH2bts(sh,(10<<4)+2);
}
#define init_8_13_10_3
void FASTCALL call_8_13_10_3(SH2_struct* sh)
{
  SH2bts(sh,(10<<4)+3);
}
#define init_8_13_10_4
void FASTCALL call_8_13_10_4(SH2_struct* sh)
{
  SH2bts(sh,(10<<4)+4);
}
#define init_8_13_10_5
void FASTCALL call_8_13_10_5(SH2_struct* sh)
{
  SH2bts(sh,(10<<4)+5);
}
#define init_8_13_10_6
void FASTCALL call_8_13_10_6(SH2_struct* sh)
{
  SH2bts(sh,(10<<4)+6);
}
#define init_8_13_10_7
void FASTCALL call_8_13_10_7(SH2_struct* sh)
{
  SH2bts(sh,(10<<4)+7);
}
#define init_8_13_10_8
void FASTCALL call_8_13_10_8(SH2_struct* sh)
{
  SH2bts(sh,(10<<4)+8);
}
#define init_8_13_10_9
void FASTCALL call_8_13_10_9(SH2_struct* sh)
{
  SH2bts(sh,(10<<4)+9);
}
#define init_8_13_10_10
void FASTCALL call_8_13_10_10(SH2_struct* sh)
{
  SH2bts(sh,(10<<4)+10);
}
#define init_8_13_10_11
void FASTCALL call_8_13_10_11(SH2_struct* sh)
{
  SH2bts(sh,(10<<4)+11);
}
#define init_8_13_10_12
void FASTCALL call_8_13_10_12(SH2_struct* sh)
{
  SH2bts(sh,(10<<4)+12);
}
#define init_8_13_10_13
void FASTCALL call_8_13_10_13(SH2_struct* sh)
{
  SH2bts(sh,(10<<4)+13);
}
#define init_8_13_10_14
void FASTCALL call_8_13_10_14(SH2_struct* sh)
{
  SH2bts(sh,(10<<4)+14);
}
#define init_8_13_10_15
void FASTCALL call_8_13_10_15(SH2_struct* sh)
{
  SH2bts(sh,(10<<4)+15);
}
#define init_8_13_11_0
void FASTCALL call_8_13_11_0(SH2_struct* sh)
{
  SH2bts(sh,(11<<4)+0);
}
#define init_8_13_11_1
void FASTCALL call_8_13_11_1(SH2_struct* sh)
{
  SH2bts(sh,(11<<4)+1);
}
#define init_8_13_11_2
void FASTCALL call_8_13_11_2(SH2_struct* sh)
{
  SH2bts(sh,(11<<4)+2);
}
#define init_8_13_11_3
void FASTCALL call_8_13_11_3(SH2_struct* sh)
{
  SH2bts(sh,(11<<4)+3);
}
#define init_8_13_11_4
void FASTCALL call_8_13_11_4(SH2_struct* sh)
{
  SH2bts(sh,(11<<4)+4);
}
#define init_8_13_11_5
void FASTCALL call_8_13_11_5(SH2_struct* sh)
{
  SH2bts(sh,(11<<4)+5);
}
#define init_8_13_11_6
void FASTCALL call_8_13_11_6(SH2_struct* sh)
{
  SH2bts(sh,(11<<4)+6);
}
#define init_8_13_11_7
void FASTCALL call_8_13_11_7(SH2_struct* sh)
{
  SH2bts(sh,(11<<4)+7);
}
#define init_8_13_11_8
void FASTCALL call_8_13_11_8(SH2_struct* sh)
{
  SH2bts(sh,(11<<4)+8);
}
#define init_8_13_11_9
void FASTCALL call_8_13_11_9(SH2_struct* sh)
{
  SH2bts(sh,(11<<4)+9);
}
#define init_8_13_11_10
void FASTCALL call_8_13_11_10(SH2_struct* sh)
{
  SH2bts(sh,(11<<4)+10);
}
#define init_8_13_11_11
void FASTCALL call_8_13_11_11(SH2_struct* sh)
{
  SH2bts(sh,(11<<4)+11);
}
#define init_8_13_11_12
void FASTCALL call_8_13_11_12(SH2_struct* sh)
{
  SH2bts(sh,(11<<4)+12);
}
#define init_8_13_11_13
void FASTCALL call_8_13_11_13(SH2_struct* sh)
{
  SH2bts(sh,(11<<4)+13);
}
#define init_8_13_11_14
void FASTCALL call_8_13_11_14(SH2_struct* sh)
{
  SH2bts(sh,(11<<4)+14);
}
#define init_8_13_11_15
void FASTCALL call_8_13_11_15(SH2_struct* sh)
{
  SH2bts(sh,(11<<4)+15);
}
#define init_8_13_12_0
void FASTCALL call_8_13_12_0(SH2_struct* sh)
{
  SH2bts(sh,(12<<4)+0);
}
#define init_8_13_12_1
void FASTCALL call_8_13_12_1(SH2_struct* sh)
{
  SH2bts(sh,(12<<4)+1);
}
#define init_8_13_12_2
void FASTCALL call_8_13_12_2(SH2_struct* sh)
{
  SH2bts(sh,(12<<4)+2);
}
#define init_8_13_12_3
void FASTCALL call_8_13_12_3(SH2_struct* sh)
{
  SH2bts(sh,(12<<4)+3);
}
#define init_8_13_12_4
void FASTCALL call_8_13_12_4(SH2_struct* sh)
{
  SH2bts(sh,(12<<4)+4);
}
#define init_8_13_12_5
void FASTCALL call_8_13_12_5(SH2_struct* sh)
{
  SH2bts(sh,(12<<4)+5);
}
#define init_8_13_12_6
void FASTCALL call_8_13_12_6(SH2_struct* sh)
{
  SH2bts(sh,(12<<4)+6);
}
#define init_8_13_12_7
void FASTCALL call_8_13_12_7(SH2_struct* sh)
{
  SH2bts(sh,(12<<4)+7);
}
#define init_8_13_12_8
void FASTCALL call_8_13_12_8(SH2_struct* sh)
{
  SH2bts(sh,(12<<4)+8);
}
#define init_8_13_12_9
void FASTCALL call_8_13_12_9(SH2_struct* sh)
{
  SH2bts(sh,(12<<4)+9);
}
#define init_8_13_12_10
void FASTCALL call_8_13_12_10(SH2_struct* sh)
{
  SH2bts(sh,(12<<4)+10);
}
#define init_8_13_12_11
void FASTCALL call_8_13_12_11(SH2_struct* sh)
{
  SH2bts(sh,(12<<4)+11);
}
#define init_8_13_12_12
void FASTCALL call_8_13_12_12(SH2_struct* sh)
{
  SH2bts(sh,(12<<4)+12);
}
#define init_8_13_12_13
void FASTCALL call_8_13_12_13(SH2_struct* sh)
{
  SH2bts(sh,(12<<4)+13);
}
#define init_8_13_12_14
void FASTCALL call_8_13_12_14(SH2_struct* sh)
{
  SH2bts(sh,(12<<4)+14);
}
#define init_8_13_12_15
void FASTCALL call_8_13_12_15(SH2_struct* sh)
{
  SH2bts(sh,(12<<4)+15);
}
#define init_8_13_13_0
void FASTCALL call_8_13_13_0(SH2_struct* sh)
{
  SH2bts(sh,(13<<4)+0);
}
#define init_8_13_13_1
void FASTCALL call_8_13_13_1(SH2_struct* sh)
{
  SH2bts(sh,(13<<4)+1);
}
#define init_8_13_13_2
void FASTCALL call_8_13_13_2(SH2_struct* sh)
{
  SH2bts(sh,(13<<4)+2);
}
#define init_8_13_13_3
void FASTCALL call_8_13_13_3(SH2_struct* sh)
{
  SH2bts(sh,(13<<4)+3);
}
#define init_8_13_13_4
void FASTCALL call_8_13_13_4(SH2_struct* sh)
{
  SH2bts(sh,(13<<4)+4);
}
#define init_8_13_13_5
void FASTCALL call_8_13_13_5(SH2_struct* sh)
{
  SH2bts(sh,(13<<4)+5);
}
#define init_8_13_13_6
void FASTCALL call_8_13_13_6(SH2_struct* sh)
{
  SH2bts(sh,(13<<4)+6);
}
#define init_8_13_13_7
void FASTCALL call_8_13_13_7(SH2_struct* sh)
{
  SH2bts(sh,(13<<4)+7);
}
#define init_8_13_13_8
void FASTCALL call_8_13_13_8(SH2_struct* sh)
{
  SH2bts(sh,(13<<4)+8);
}
#define init_8_13_13_9
void FASTCALL call_8_13_13_9(SH2_struct* sh)
{
  SH2bts(sh,(13<<4)+9);
}
#define init_8_13_13_10
void FASTCALL call_8_13_13_10(SH2_struct* sh)
{
  SH2bts(sh,(13<<4)+10);
}
#define init_8_13_13_11
void FASTCALL call_8_13_13_11(SH2_struct* sh)
{
  SH2bts(sh,(13<<4)+11);
}
#define init_8_13_13_12
void FASTCALL call_8_13_13_12(SH2_struct* sh)
{
  SH2bts(sh,(13<<4)+12);
}
#define init_8_13_13_13
void FASTCALL call_8_13_13_13(SH2_struct* sh)
{
  SH2bts(sh,(13<<4)+13);
}
#define init_8_13_13_14
void FASTCALL call_8_13_13_14(SH2_struct* sh)
{
  SH2bts(sh,(13<<4)+14);
}
#define init_8_13_13_15
void FASTCALL call_8_13_13_15(SH2_struct* sh)
{
  SH2bts(sh,(13<<4)+15);
}
#define init_8_13_14_0
void FASTCALL call_8_13_14_0(SH2_struct* sh)
{
  SH2bts(sh,(14<<4)+0);
}
#define init_8_13_14_1
void FASTCALL call_8_13_14_1(SH2_struct* sh)
{
  SH2bts(sh,(14<<4)+1);
}
#define init_8_13_14_2
void FASTCALL call_8_13_14_2(SH2_struct* sh)
{
  SH2bts(sh,(14<<4)+2);
}
#define init_8_13_14_3
void FASTCALL call_8_13_14_3(SH2_struct* sh)
{
  SH2bts(sh,(14<<4)+3);
}
#define init_8_13_14_4
void FASTCALL call_8_13_14_4(SH2_struct* sh)
{
  SH2bts(sh,(14<<4)+4);
}
#define init_8_13_14_5
void FASTCALL call_8_13_14_5(SH2_struct* sh)
{
  SH2bts(sh,(14<<4)+5);
}
#define init_8_13_14_6
void FASTCALL call_8_13_14_6(SH2_struct* sh)
{
  SH2bts(sh,(14<<4)+6);
}
#define init_8_13_14_7
void FASTCALL call_8_13_14_7(SH2_struct* sh)
{
  SH2bts(sh,(14<<4)+7);
}
#define init_8_13_14_8
void FASTCALL call_8_13_14_8(SH2_struct* sh)
{
  SH2bts(sh,(14<<4)+8);
}
#define init_8_13_14_9
void FASTCALL call_8_13_14_9(SH2_struct* sh)
{
  SH2bts(sh,(14<<4)+9);
}
#define init_8_13_14_10
void FASTCALL call_8_13_14_10(SH2_struct* sh)
{
  SH2bts(sh,(14<<4)+10);
}
#define init_8_13_14_11
void FASTCALL call_8_13_14_11(SH2_struct* sh)
{
  SH2bts(sh,(14<<4)+11);
}
#define init_8_13_14_12
void FASTCALL call_8_13_14_12(SH2_struct* sh)
{
  SH2bts(sh,(14<<4)+12);
}
#define init_8_13_14_13
void FASTCALL call_8_13_14_13(SH2_struct* sh)
{
  SH2bts(sh,(14<<4)+13);
}
#define init_8_13_14_14
void FASTCALL call_8_13_14_14(SH2_struct* sh)
{
  SH2bts(sh,(14<<4)+14);
}
#define init_8_13_14_15
void FASTCALL call_8_13_14_15(SH2_struct* sh)
{
  SH2bts(sh,(14<<4)+15);
}
#define init_8_13_15_0
void FASTCALL call_8_13_15_0(SH2_struct* sh)
{
  SH2bts(sh,(15<<4)+0);
}
#define init_8_13_15_1
void FASTCALL call_8_13_15_1(SH2_struct* sh)
{
  SH2bts(sh,(15<<4)+1);
}
#define init_8_13_15_2
void FASTCALL call_8_13_15_2(SH2_struct* sh)
{
  SH2bts(sh,(15<<4)+2);
}
#define init_8_13_15_3
void FASTCALL call_8_13_15_3(SH2_struct* sh)
{
  SH2bts(sh,(15<<4)+3);
}
#define init_8_13_15_4
void FASTCALL call_8_13_15_4(SH2_struct* sh)
{
  SH2bts(sh,(15<<4)+4);
}
#define init_8_13_15_5
void FASTCALL call_8_13_15_5(SH2_struct* sh)
{
  SH2bts(sh,(15<<4)+5);
}
#define init_8_13_15_6
void FASTCALL call_8_13_15_6(SH2_struct* sh)
{
  SH2bts(sh,(15<<4)+6);
}
#define init_8_13_15_7
void FASTCALL call_8_13_15_7(SH2_struct* sh)
{
  SH2bts(sh,(15<<4)+7);
}
#define init_8_13_15_8
void FASTCALL call_8_13_15_8(SH2_struct* sh)
{
  SH2bts(sh,(15<<4)+8);
}
#define init_8_13_15_9
void FASTCALL call_8_13_15_9(SH2_struct* sh)
{
  SH2bts(sh,(15<<4)+9);
}
#define init_8_13_15_10
void FASTCALL call_8_13_15_10(SH2_struct* sh)
{
  SH2bts(sh,(15<<4)+10);
}
#define init_8_13_15_11
void FASTCALL call_8_13_15_11(SH2_struct* sh)
{
  SH2bts(sh,(15<<4)+11);
}
#define init_8_13_15_12
void FASTCALL call_8_13_15_12(SH2_struct* sh)
{
  SH2bts(sh,(15<<4)+12);
}
#define init_8_13_15_13
void FASTCALL call_8_13_15_13(SH2_struct* sh)
{
  SH2bts(sh,(15<<4)+13);
}
#define init_8_13_15_14
void FASTCALL call_8_13_15_14(SH2_struct* sh)
{
  SH2bts(sh,(15<<4)+14);
}
#define init_8_13_15_15
void FASTCALL call_8_13_15_15(SH2_struct* sh)
{
  SH2bts(sh,(15<<4)+15);
}
#define init_8_9_0_0
void FASTCALL call_8_9_0_0(SH2_struct* sh)
{
  SH2bt(sh,(0<<4)+0);
}
#define init_8_9_0_1
void FASTCALL call_8_9_0_1(SH2_struct* sh)
{
  SH2bt(sh,(0<<4)+1);
}
#define init_8_9_0_2
void FASTCALL call_8_9_0_2(SH2_struct* sh)
{
  SH2bt(sh,(0<<4)+2);
}
#define init_8_9_0_3
void FASTCALL call_8_9_0_3(SH2_struct* sh)
{
  SH2bt(sh,(0<<4)+3);
}
#define init_8_9_0_4
void FASTCALL call_8_9_0_4(SH2_struct* sh)
{
  SH2bt(sh,(0<<4)+4);
}
#define init_8_9_0_5
void FASTCALL call_8_9_0_5(SH2_struct* sh)
{
  SH2bt(sh,(0<<4)+5);
}
#define init_8_9_0_6
void FASTCALL call_8_9_0_6(SH2_struct* sh)
{
  SH2bt(sh,(0<<4)+6);
}
#define init_8_9_0_7
void FASTCALL call_8_9_0_7(SH2_struct* sh)
{
  SH2bt(sh,(0<<4)+7);
}
#define init_8_9_0_8
void FASTCALL call_8_9_0_8(SH2_struct* sh)
{
  SH2bt(sh,(0<<4)+8);
}
#define init_8_9_0_9
void FASTCALL call_8_9_0_9(SH2_struct* sh)
{
  SH2bt(sh,(0<<4)+9);
}
#define init_8_9_0_10
void FASTCALL call_8_9_0_10(SH2_struct* sh)
{
  SH2bt(sh,(0<<4)+10);
}
#define init_8_9_0_11
void FASTCALL call_8_9_0_11(SH2_struct* sh)
{
  SH2bt(sh,(0<<4)+11);
}
#define init_8_9_0_12
void FASTCALL call_8_9_0_12(SH2_struct* sh)
{
  SH2bt(sh,(0<<4)+12);
}
#define init_8_9_0_13
void FASTCALL call_8_9_0_13(SH2_struct* sh)
{
  SH2bt(sh,(0<<4)+13);
}
#define init_8_9_0_14
void FASTCALL call_8_9_0_14(SH2_struct* sh)
{
  SH2bt(sh,(0<<4)+14);
}
#define init_8_9_0_15
void FASTCALL call_8_9_0_15(SH2_struct* sh)
{
  SH2bt(sh,(0<<4)+15);
}
#define init_8_9_1_0
void FASTCALL call_8_9_1_0(SH2_struct* sh)
{
  SH2bt(sh,(1<<4)+0);
}
#define init_8_9_1_1
void FASTCALL call_8_9_1_1(SH2_struct* sh)
{
  SH2bt(sh,(1<<4)+1);
}
#define init_8_9_1_2
void FASTCALL call_8_9_1_2(SH2_struct* sh)
{
  SH2bt(sh,(1<<4)+2);
}
#define init_8_9_1_3
void FASTCALL call_8_9_1_3(SH2_struct* sh)
{
  SH2bt(sh,(1<<4)+3);
}
#define init_8_9_1_4
void FASTCALL call_8_9_1_4(SH2_struct* sh)
{
  SH2bt(sh,(1<<4)+4);
}
#define init_8_9_1_5
void FASTCALL call_8_9_1_5(SH2_struct* sh)
{
  SH2bt(sh,(1<<4)+5);
}
#define init_8_9_1_6
void FASTCALL call_8_9_1_6(SH2_struct* sh)
{
  SH2bt(sh,(1<<4)+6);
}
#define init_8_9_1_7
void FASTCALL call_8_9_1_7(SH2_struct* sh)
{
  SH2bt(sh,(1<<4)+7);
}
#define init_8_9_1_8
void FASTCALL call_8_9_1_8(SH2_struct* sh)
{
  SH2bt(sh,(1<<4)+8);
}
#define init_8_9_1_9
void FASTCALL call_8_9_1_9(SH2_struct* sh)
{
  SH2bt(sh,(1<<4)+9);
}
#define init_8_9_1_10
void FASTCALL call_8_9_1_10(SH2_struct* sh)
{
  SH2bt(sh,(1<<4)+10);
}
#define init_8_9_1_11
void FASTCALL call_8_9_1_11(SH2_struct* sh)
{
  SH2bt(sh,(1<<4)+11);
}
#define init_8_9_1_12
void FASTCALL call_8_9_1_12(SH2_struct* sh)
{
  SH2bt(sh,(1<<4)+12);
}
#define init_8_9_1_13
void FASTCALL call_8_9_1_13(SH2_struct* sh)
{
  SH2bt(sh,(1<<4)+13);
}
#define init_8_9_1_14
void FASTCALL call_8_9_1_14(SH2_struct* sh)
{
  SH2bt(sh,(1<<4)+14);
}
#define init_8_9_1_15
void FASTCALL call_8_9_1_15(SH2_struct* sh)
{
  SH2bt(sh,(1<<4)+15);
}
#define init_8_9_2_0
void FASTCALL call_8_9_2_0(SH2_struct* sh)
{
  SH2bt(sh,(2<<4)+0);
}
#define init_8_9_2_1
void FASTCALL call_8_9_2_1(SH2_struct* sh)
{
  SH2bt(sh,(2<<4)+1);
}
#define init_8_9_2_2
void FASTCALL call_8_9_2_2(SH2_struct* sh)
{
  SH2bt(sh,(2<<4)+2);
}
#define init_8_9_2_3
void FASTCALL call_8_9_2_3(SH2_struct* sh)
{
  SH2bt(sh,(2<<4)+3);
}
#define init_8_9_2_4
void FASTCALL call_8_9_2_4(SH2_struct* sh)
{
  SH2bt(sh,(2<<4)+4);
}
#define init_8_9_2_5
void FASTCALL call_8_9_2_5(SH2_struct* sh)
{
  SH2bt(sh,(2<<4)+5);
}
#define init_8_9_2_6
void FASTCALL call_8_9_2_6(SH2_struct* sh)
{
  SH2bt(sh,(2<<4)+6);
}
#define init_8_9_2_7
void FASTCALL call_8_9_2_7(SH2_struct* sh)
{
  SH2bt(sh,(2<<4)+7);
}
#define init_8_9_2_8
void FASTCALL call_8_9_2_8(SH2_struct* sh)
{
  SH2bt(sh,(2<<4)+8);
}
#define init_8_9_2_9
void FASTCALL call_8_9_2_9(SH2_struct* sh)
{
  SH2bt(sh,(2<<4)+9);
}
#define init_8_9_2_10
void FASTCALL call_8_9_2_10(SH2_struct* sh)
{
  SH2bt(sh,(2<<4)+10);
}
#define init_8_9_2_11
void FASTCALL call_8_9_2_11(SH2_struct* sh)
{
  SH2bt(sh,(2<<4)+11);
}
#define init_8_9_2_12
void FASTCALL call_8_9_2_12(SH2_struct* sh)
{
  SH2bt(sh,(2<<4)+12);
}
#define init_8_9_2_13
void FASTCALL call_8_9_2_13(SH2_struct* sh)
{
  SH2bt(sh,(2<<4)+13);
}
#define init_8_9_2_14
void FASTCALL call_8_9_2_14(SH2_struct* sh)
{
  SH2bt(sh,(2<<4)+14);
}
#define init_8_9_2_15
void FASTCALL call_8_9_2_15(SH2_struct* sh)
{
  SH2bt(sh,(2<<4)+15);
}
#define init_8_9_3_0
void FASTCALL call_8_9_3_0(SH2_struct* sh)
{
  SH2bt(sh,(3<<4)+0);
}
#define init_8_9_3_1
void FASTCALL call_8_9_3_1(SH2_struct* sh)
{
  SH2bt(sh,(3<<4)+1);
}
#define init_8_9_3_2
void FASTCALL call_8_9_3_2(SH2_struct* sh)
{
  SH2bt(sh,(3<<4)+2);
}
#define init_8_9_3_3
void FASTCALL call_8_9_3_3(SH2_struct* sh)
{
  SH2bt(sh,(3<<4)+3);
}
#define init_8_9_3_4
void FASTCALL call_8_9_3_4(SH2_struct* sh)
{
  SH2bt(sh,(3<<4)+4);
}
#define init_8_9_3_5
void FASTCALL call_8_9_3_5(SH2_struct* sh)
{
  SH2bt(sh,(3<<4)+5);
}
#define init_8_9_3_6
void FASTCALL call_8_9_3_6(SH2_struct* sh)
{
  SH2bt(sh,(3<<4)+6);
}
#define init_8_9_3_7
void FASTCALL call_8_9_3_7(SH2_struct* sh)
{
  SH2bt(sh,(3<<4)+7);
}
#define init_8_9_3_8
void FASTCALL call_8_9_3_8(SH2_struct* sh)
{
  SH2bt(sh,(3<<4)+8);
}
#define init_8_9_3_9
void FASTCALL call_8_9_3_9(SH2_struct* sh)
{
  SH2bt(sh,(3<<4)+9);
}
#define init_8_9_3_10
void FASTCALL call_8_9_3_10(SH2_struct* sh)
{
  SH2bt(sh,(3<<4)+10);
}
#define init_8_9_3_11
void FASTCALL call_8_9_3_11(SH2_struct* sh)
{
  SH2bt(sh,(3<<4)+11);
}
#define init_8_9_3_12
void FASTCALL call_8_9_3_12(SH2_struct* sh)
{
  SH2bt(sh,(3<<4)+12);
}
#define init_8_9_3_13
void FASTCALL call_8_9_3_13(SH2_struct* sh)
{
  SH2bt(sh,(3<<4)+13);
}
#define init_8_9_3_14
void FASTCALL call_8_9_3_14(SH2_struct* sh)
{
  SH2bt(sh,(3<<4)+14);
}
#define init_8_9_3_15
void FASTCALL call_8_9_3_15(SH2_struct* sh)
{
  SH2bt(sh,(3<<4)+15);
}
#define init_8_9_4_0
void FASTCALL call_8_9_4_0(SH2_struct* sh)
{
  SH2bt(sh,(4<<4)+0);
}
#define init_8_9_4_1
void FASTCALL call_8_9_4_1(SH2_struct* sh)
{
  SH2bt(sh,(4<<4)+1);
}
#define init_8_9_4_2
void FASTCALL call_8_9_4_2(SH2_struct* sh)
{
  SH2bt(sh,(4<<4)+2);
}
#define init_8_9_4_3
void FASTCALL call_8_9_4_3(SH2_struct* sh)
{
  SH2bt(sh,(4<<4)+3);
}
#define init_8_9_4_4
void FASTCALL call_8_9_4_4(SH2_struct* sh)
{
  SH2bt(sh,(4<<4)+4);
}
#define init_8_9_4_5
void FASTCALL call_8_9_4_5(SH2_struct* sh)
{
  SH2bt(sh,(4<<4)+5);
}
#define init_8_9_4_6
void FASTCALL call_8_9_4_6(SH2_struct* sh)
{
  SH2bt(sh,(4<<4)+6);
}
#define init_8_9_4_7
void FASTCALL call_8_9_4_7(SH2_struct* sh)
{
  SH2bt(sh,(4<<4)+7);
}
#define init_8_9_4_8
void FASTCALL call_8_9_4_8(SH2_struct* sh)
{
  SH2bt(sh,(4<<4)+8);
}
#define init_8_9_4_9
void FASTCALL call_8_9_4_9(SH2_struct* sh)
{
  SH2bt(sh,(4<<4)+9);
}
#define init_8_9_4_10
void FASTCALL call_8_9_4_10(SH2_struct* sh)
{
  SH2bt(sh,(4<<4)+10);
}
#define init_8_9_4_11
void FASTCALL call_8_9_4_11(SH2_struct* sh)
{
  SH2bt(sh,(4<<4)+11);
}
#define init_8_9_4_12
void FASTCALL call_8_9_4_12(SH2_struct* sh)
{
  SH2bt(sh,(4<<4)+12);
}
#define init_8_9_4_13
void FASTCALL call_8_9_4_13(SH2_struct* sh)
{
  SH2bt(sh,(4<<4)+13);
}
#define init_8_9_4_14
void FASTCALL call_8_9_4_14(SH2_struct* sh)
{
  SH2bt(sh,(4<<4)+14);
}
#define init_8_9_4_15
void FASTCALL call_8_9_4_15(SH2_struct* sh)
{
  SH2bt(sh,(4<<4)+15);
}
#define init_8_9_5_0
void FASTCALL call_8_9_5_0(SH2_struct* sh)
{
  SH2bt(sh,(5<<4)+0);
}
#define init_8_9_5_1
void FASTCALL call_8_9_5_1(SH2_struct* sh)
{
  SH2bt(sh,(5<<4)+1);
}
#define init_8_9_5_2
void FASTCALL call_8_9_5_2(SH2_struct* sh)
{
  SH2bt(sh,(5<<4)+2);
}
#define init_8_9_5_3
void FASTCALL call_8_9_5_3(SH2_struct* sh)
{
  SH2bt(sh,(5<<4)+3);
}
#define init_8_9_5_4
void FASTCALL call_8_9_5_4(SH2_struct* sh)
{
  SH2bt(sh,(5<<4)+4);
}
#define init_8_9_5_5
void FASTCALL call_8_9_5_5(SH2_struct* sh)
{
  SH2bt(sh,(5<<4)+5);
}
#define init_8_9_5_6
void FASTCALL call_8_9_5_6(SH2_struct* sh)
{
  SH2bt(sh,(5<<4)+6);
}
#define init_8_9_5_7
void FASTCALL call_8_9_5_7(SH2_struct* sh)
{
  SH2bt(sh,(5<<4)+7);
}
#define init_8_9_5_8
void FASTCALL call_8_9_5_8(SH2_struct* sh)
{
  SH2bt(sh,(5<<4)+8);
}
#define init_8_9_5_9
void FASTCALL call_8_9_5_9(SH2_struct* sh)
{
  SH2bt(sh,(5<<4)+9);
}
#define init_8_9_5_10
void FASTCALL call_8_9_5_10(SH2_struct* sh)
{
  SH2bt(sh,(5<<4)+10);
}
#define init_8_9_5_11
void FASTCALL call_8_9_5_11(SH2_struct* sh)
{
  SH2bt(sh,(5<<4)+11);
}
#define init_8_9_5_12
void FASTCALL call_8_9_5_12(SH2_struct* sh)
{
  SH2bt(sh,(5<<4)+12);
}
#define init_8_9_5_13
void FASTCALL call_8_9_5_13(SH2_struct* sh)
{
  SH2bt(sh,(5<<4)+13);
}
#define init_8_9_5_14
void FASTCALL call_8_9_5_14(SH2_struct* sh)
{
  SH2bt(sh,(5<<4)+14);
}
#define init_8_9_5_15
void FASTCALL call_8_9_5_15(SH2_struct* sh)
{
  SH2bt(sh,(5<<4)+15);
}
#define init_8_9_6_0
void FASTCALL call_8_9_6_0(SH2_struct* sh)
{
  SH2bt(sh,(6<<4)+0);
}
#define init_8_9_6_1
void FASTCALL call_8_9_6_1(SH2_struct* sh)
{
  SH2bt(sh,(6<<4)+1);
}
#define init_8_9_6_2
void FASTCALL call_8_9_6_2(SH2_struct* sh)
{
  SH2bt(sh,(6<<4)+2);
}
#define init_8_9_6_3
void FASTCALL call_8_9_6_3(SH2_struct* sh)
{
  SH2bt(sh,(6<<4)+3);
}
#define init_8_9_6_4
void FASTCALL call_8_9_6_4(SH2_struct* sh)
{
  SH2bt(sh,(6<<4)+4);
}
#define init_8_9_6_5
void FASTCALL call_8_9_6_5(SH2_struct* sh)
{
  SH2bt(sh,(6<<4)+5);
}
#define init_8_9_6_6
void FASTCALL call_8_9_6_6(SH2_struct* sh)
{
  SH2bt(sh,(6<<4)+6);
}
#define init_8_9_6_7
void FASTCALL call_8_9_6_7(SH2_struct* sh)
{
  SH2bt(sh,(6<<4)+7);
}
#define init_8_9_6_8
void FASTCALL call_8_9_6_8(SH2_struct* sh)
{
  SH2bt(sh,(6<<4)+8);
}
#define init_8_9_6_9
void FASTCALL call_8_9_6_9(SH2_struct* sh)
{
  SH2bt(sh,(6<<4)+9);
}
#define init_8_9_6_10
void FASTCALL call_8_9_6_10(SH2_struct* sh)
{
  SH2bt(sh,(6<<4)+10);
}
#define init_8_9_6_11
void FASTCALL call_8_9_6_11(SH2_struct* sh)
{
  SH2bt(sh,(6<<4)+11);
}
#define init_8_9_6_12
void FASTCALL call_8_9_6_12(SH2_struct* sh)
{
  SH2bt(sh,(6<<4)+12);
}
#define init_8_9_6_13
void FASTCALL call_8_9_6_13(SH2_struct* sh)
{
  SH2bt(sh,(6<<4)+13);
}
#define init_8_9_6_14
void FASTCALL call_8_9_6_14(SH2_struct* sh)
{
  SH2bt(sh,(6<<4)+14);
}
#define init_8_9_6_15
void FASTCALL call_8_9_6_15(SH2_struct* sh)
{
  SH2bt(sh,(6<<4)+15);
}
#define init_8_9_7_0
void FASTCALL call_8_9_7_0(SH2_struct* sh)
{
  SH2bt(sh,(7<<4)+0);
}
#define init_8_9_7_1
void FASTCALL call_8_9_7_1(SH2_struct* sh)
{
  SH2bt(sh,(7<<4)+1);
}
#define init_8_9_7_2
void FASTCALL call_8_9_7_2(SH2_struct* sh)
{
  SH2bt(sh,(7<<4)+2);
}
#define init_8_9_7_3
void FASTCALL call_8_9_7_3(SH2_struct* sh)
{
  SH2bt(sh,(7<<4)+3);
}
#define init_8_9_7_4
void FASTCALL call_8_9_7_4(SH2_struct* sh)
{
  SH2bt(sh,(7<<4)+4);
}
#define init_8_9_7_5
void FASTCALL call_8_9_7_5(SH2_struct* sh)
{
  SH2bt(sh,(7<<4)+5);
}
#define init_8_9_7_6
void FASTCALL call_8_9_7_6(SH2_struct* sh)
{
  SH2bt(sh,(7<<4)+6);
}
#define init_8_9_7_7
void FASTCALL call_8_9_7_7(SH2_struct* sh)
{
  SH2bt(sh,(7<<4)+7);
}
#define init_8_9_7_8
void FASTCALL call_8_9_7_8(SH2_struct* sh)
{
  SH2bt(sh,(7<<4)+8);
}
#define init_8_9_7_9
void FASTCALL call_8_9_7_9(SH2_struct* sh)
{
  SH2bt(sh,(7<<4)+9);
}
#define init_8_9_7_10
void FASTCALL call_8_9_7_10(SH2_struct* sh)
{
  SH2bt(sh,(7<<4)+10);
}
#define init_8_9_7_11
void FASTCALL call_8_9_7_11(SH2_struct* sh)
{
  SH2bt(sh,(7<<4)+11);
}
#define init_8_9_7_12
void FASTCALL call_8_9_7_12(SH2_struct* sh)
{
  SH2bt(sh,(7<<4)+12);
}
#define init_8_9_7_13
void FASTCALL call_8_9_7_13(SH2_struct* sh)
{
  SH2bt(sh,(7<<4)+13);
}
#define init_8_9_7_14
void FASTCALL call_8_9_7_14(SH2_struct* sh)
{
  SH2bt(sh,(7<<4)+14);
}
#define init_8_9_7_15
void FASTCALL call_8_9_7_15(SH2_struct* sh)
{
  SH2bt(sh,(7<<4)+15);
}
#define init_8_9_8_0
void FASTCALL call_8_9_8_0(SH2_struct* sh)
{
  SH2bt(sh,(8<<4)+0);
}
#define init_8_9_8_1
void FASTCALL call_8_9_8_1(SH2_struct* sh)
{
  SH2bt(sh,(8<<4)+1);
}
#define init_8_9_8_2
void FASTCALL call_8_9_8_2(SH2_struct* sh)
{
  SH2bt(sh,(8<<4)+2);
}
#define init_8_9_8_3
void FASTCALL call_8_9_8_3(SH2_struct* sh)
{
  SH2bt(sh,(8<<4)+3);
}
#define init_8_9_8_4
void FASTCALL call_8_9_8_4(SH2_struct* sh)
{
  SH2bt(sh,(8<<4)+4);
}
#define init_8_9_8_5
void FASTCALL call_8_9_8_5(SH2_struct* sh)
{
  SH2bt(sh,(8<<4)+5);
}
#define init_8_9_8_6
void FASTCALL call_8_9_8_6(SH2_struct* sh)
{
  SH2bt(sh,(8<<4)+6);
}
#define init_8_9_8_7
void FASTCALL call_8_9_8_7(SH2_struct* sh)
{
  SH2bt(sh,(8<<4)+7);
}
#define init_8_9_8_8
void FASTCALL call_8_9_8_8(SH2_struct* sh)
{
  SH2bt(sh,(8<<4)+8);
}
#define init_8_9_8_9
void FASTCALL call_8_9_8_9(SH2_struct* sh)
{
  SH2bt(sh,(8<<4)+9);
}
#define init_8_9_8_10
void FASTCALL call_8_9_8_10(SH2_struct* sh)
{
  SH2bt(sh,(8<<4)+10);
}
#define init_8_9_8_11
void FASTCALL call_8_9_8_11(SH2_struct* sh)
{
  SH2bt(sh,(8<<4)+11);
}
#define init_8_9_8_12
void FASTCALL call_8_9_8_12(SH2_struct* sh)
{
  SH2bt(sh,(8<<4)+12);
}
#define init_8_9_8_13
void FASTCALL call_8_9_8_13(SH2_struct* sh)
{
  SH2bt(sh,(8<<4)+13);
}
#define init_8_9_8_14
void FASTCALL call_8_9_8_14(SH2_struct* sh)
{
  SH2bt(sh,(8<<4)+14);
}
#define init_8_9_8_15
void FASTCALL call_8_9_8_15(SH2_struct* sh)
{
  SH2bt(sh,(8<<4)+15);
}
#define init_8_9_9_0
void FASTCALL call_8_9_9_0(SH2_struct* sh)
{
  SH2bt(sh,(9<<4)+0);
}
#define init_8_9_9_1
void FASTCALL call_8_9_9_1(SH2_struct* sh)
{
  SH2bt(sh,(9<<4)+1);
}
#define init_8_9_9_2
void FASTCALL call_8_9_9_2(SH2_struct* sh)
{
  SH2bt(sh,(9<<4)+2);
}
#define init_8_9_9_3
void FASTCALL call_8_9_9_3(SH2_struct* sh)
{
  SH2bt(sh,(9<<4)+3);
}
#define init_8_9_9_4
void FASTCALL call_8_9_9_4(SH2_struct* sh)
{
  SH2bt(sh,(9<<4)+4);
}
#define init_8_9_9_5
void FASTCALL call_8_9_9_5(SH2_struct* sh)
{
  SH2bt(sh,(9<<4)+5);
}
#define init_8_9_9_6
void FASTCALL call_8_9_9_6(SH2_struct* sh)
{
  SH2bt(sh,(9<<4)+6);
}
#define init_8_9_9_7
void FASTCALL call_8_9_9_7(SH2_struct* sh)
{
  SH2bt(sh,(9<<4)+7);
}
#define init_8_9_9_8
void FASTCALL call_8_9_9_8(SH2_struct* sh)
{
  SH2bt(sh,(9<<4)+8);
}
#define init_8_9_9_9
void FASTCALL call_8_9_9_9(SH2_struct* sh)
{
  SH2bt(sh,(9<<4)+9);
}
#define init_8_9_9_10
void FASTCALL call_8_9_9_10(SH2_struct* sh)
{
  SH2bt(sh,(9<<4)+10);
}
#define init_8_9_9_11
void FASTCALL call_8_9_9_11(SH2_struct* sh)
{
  SH2bt(sh,(9<<4)+11);
}
#define init_8_9_9_12
void FASTCALL call_8_9_9_12(SH2_struct* sh)
{
  SH2bt(sh,(9<<4)+12);
}
#define init_8_9_9_13
void FASTCALL call_8_9_9_13(SH2_struct* sh)
{
  SH2bt(sh,(9<<4)+13);
}
#define init_8_9_9_14
void FASTCALL call_8_9_9_14(SH2_struct* sh)
{
  SH2bt(sh,(9<<4)+14);
}
#define init_8_9_9_15
void FASTCALL call_8_9_9_15(SH2_struct* sh)
{
  SH2bt(sh,(9<<4)+15);
}
#define init_8_9_10_0
void FASTCALL call_8_9_10_0(SH2_struct* sh)
{
  SH2bt(sh,(10<<4)+0);
}
#define init_8_9_10_1
void FASTCALL call_8_9_10_1(SH2_struct* sh)
{
  SH2bt(sh,(10<<4)+1);
}
#define init_8_9_10_2
void FASTCALL call_8_9_10_2(SH2_struct* sh)
{
  SH2bt(sh,(10<<4)+2);
}
#define init_8_9_10_3
void FASTCALL call_8_9_10_3(SH2_struct* sh)
{
  SH2bt(sh,(10<<4)+3);
}
#define init_8_9_10_4
void FASTCALL call_8_9_10_4(SH2_struct* sh)
{
  SH2bt(sh,(10<<4)+4);
}
#define init_8_9_10_5
void FASTCALL call_8_9_10_5(SH2_struct* sh)
{
  SH2bt(sh,(10<<4)+5);
}
#define init_8_9_10_6
void FASTCALL call_8_9_10_6(SH2_struct* sh)
{
  SH2bt(sh,(10<<4)+6);
}
#define init_8_9_10_7
void FASTCALL call_8_9_10_7(SH2_struct* sh)
{
  SH2bt(sh,(10<<4)+7);
}
#define init_8_9_10_8
void FASTCALL call_8_9_10_8(SH2_struct* sh)
{
  SH2bt(sh,(10<<4)+8);
}
#define init_8_9_10_9
void FASTCALL call_8_9_10_9(SH2_struct* sh)
{
  SH2bt(sh,(10<<4)+9);
}
#define init_8_9_10_10
void FASTCALL call_8_9_10_10(SH2_struct* sh)
{
  SH2bt(sh,(10<<4)+10);
}
#define init_8_9_10_11
void FASTCALL call_8_9_10_11(SH2_struct* sh)
{
  SH2bt(sh,(10<<4)+11);
}
#define init_8_9_10_12
void FASTCALL call_8_9_10_12(SH2_struct* sh)
{
  SH2bt(sh,(10<<4)+12);
}
#define init_8_9_10_13
void FASTCALL call_8_9_10_13(SH2_struct* sh)
{
  SH2bt(sh,(10<<4)+13);
}
#define init_8_9_10_14
void FASTCALL call_8_9_10_14(SH2_struct* sh)
{
  SH2bt(sh,(10<<4)+14);
}
#define init_8_9_10_15
void FASTCALL call_8_9_10_15(SH2_struct* sh)
{
  SH2bt(sh,(10<<4)+15);
}
#define init_8_9_11_0
void FASTCALL call_8_9_11_0(SH2_struct* sh)
{
  SH2bt(sh,(11<<4)+0);
}
#define init_8_9_11_1
void FASTCALL call_8_9_11_1(SH2_struct* sh)
{
  SH2bt(sh,(11<<4)+1);
}
#define init_8_9_11_2
void FASTCALL call_8_9_11_2(SH2_struct* sh)
{
  SH2bt(sh,(11<<4)+2);
}
#define init_8_9_11_3
void FASTCALL call_8_9_11_3(SH2_struct* sh)
{
  SH2bt(sh,(11<<4)+3);
}
#define init_8_9_11_4
void FASTCALL call_8_9_11_4(SH2_struct* sh)
{
  SH2bt(sh,(11<<4)+4);
}
#define init_8_9_11_5
void FASTCALL call_8_9_11_5(SH2_struct* sh)
{
  SH2bt(sh,(11<<4)+5);
}
#define init_8_9_11_6
void FASTCALL call_8_9_11_6(SH2_struct* sh)
{
  SH2bt(sh,(11<<4)+6);
}
#define init_8_9_11_7
void FASTCALL call_8_9_11_7(SH2_struct* sh)
{
  SH2bt(sh,(11<<4)+7);
}
#define init_8_9_11_8
void FASTCALL call_8_9_11_8(SH2_struct* sh)
{
  SH2bt(sh,(11<<4)+8);
}
#define init_8_9_11_9
void FASTCALL call_8_9_11_9(SH2_struct* sh)
{
  SH2bt(sh,(11<<4)+9);
}
#define init_8_9_11_10
void FASTCALL call_8_9_11_10(SH2_struct* sh)
{
  SH2bt(sh,(11<<4)+10);
}
#define init_8_9_11_11
void FASTCALL call_8_9_11_11(SH2_struct* sh)
{
  SH2bt(sh,(11<<4)+11);
}
#define init_8_9_11_12
void FASTCALL call_8_9_11_12(SH2_struct* sh)
{
  SH2bt(sh,(11<<4)+12);
}
#define init_8_9_11_13
void FASTCALL call_8_9_11_13(SH2_struct* sh)
{
  SH2bt(sh,(11<<4)+13);
}
#define init_8_9_11_14
void FASTCALL call_8_9_11_14(SH2_struct* sh)
{
  SH2bt(sh,(11<<4)+14);
}
#define init_8_9_11_15
void FASTCALL call_8_9_11_15(SH2_struct* sh)
{
  SH2bt(sh,(11<<4)+15);
}
#define init_8_9_12_0
void FASTCALL call_8_9_12_0(SH2_struct* sh)
{
  SH2bt(sh,(12<<4)+0);
}
#define init_8_9_12_1
void FASTCALL call_8_9_12_1(SH2_struct* sh)
{
  SH2bt(sh,(12<<4)+1);
}
#define init_8_9_12_2
void FASTCALL call_8_9_12_2(SH2_struct* sh)
{
  SH2bt(sh,(12<<4)+2);
}
#define init_8_9_12_3
void FASTCALL call_8_9_12_3(SH2_struct* sh)
{
  SH2bt(sh,(12<<4)+3);
}
#define init_8_9_12_4
void FASTCALL call_8_9_12_4(SH2_struct* sh)
{
  SH2bt(sh,(12<<4)+4);
}
#define init_8_9_12_5
void FASTCALL call_8_9_12_5(SH2_struct* sh)
{
  SH2bt(sh,(12<<4)+5);
}
#define init_8_9_12_6
void FASTCALL call_8_9_12_6(SH2_struct* sh)
{
  SH2bt(sh,(12<<4)+6);
}
#define init_8_9_12_7
void FASTCALL call_8_9_12_7(SH2_struct* sh)
{
  SH2bt(sh,(12<<4)+7);
}
#define init_8_9_12_8
void FASTCALL call_8_9_12_8(SH2_struct* sh)
{
  SH2bt(sh,(12<<4)+8);
}
#define init_8_9_12_9
void FASTCALL call_8_9_12_9(SH2_struct* sh)
{
  SH2bt(sh,(12<<4)+9);
}
#define init_8_9_12_10
void FASTCALL call_8_9_12_10(SH2_struct* sh)
{
  SH2bt(sh,(12<<4)+10);
}
#define init_8_9_12_11
void FASTCALL call_8_9_12_11(SH2_struct* sh)
{
  SH2bt(sh,(12<<4)+11);
}
#define init_8_9_12_12
void FASTCALL call_8_9_12_12(SH2_struct* sh)
{
  SH2bt(sh,(12<<4)+12);
}
#define init_8_9_12_13
void FASTCALL call_8_9_12_13(SH2_struct* sh)
{
  SH2bt(sh,(12<<4)+13);
}
#define init_8_9_12_14
void FASTCALL call_8_9_12_14(SH2_struct* sh)
{
  SH2bt(sh,(12<<4)+14);
}
#define init_8_9_12_15
void FASTCALL call_8_9_12_15(SH2_struct* sh)
{
  SH2bt(sh,(12<<4)+15);
}
#define init_8_9_13_0
void FASTCALL call_8_9_13_0(SH2_struct* sh)
{
  SH2bt(sh,(13<<4)+0);
}
#define init_8_9_13_1
void FASTCALL call_8_9_13_1(SH2_struct* sh)
{
  SH2bt(sh,(13<<4)+1);
}
#define init_8_9_13_2
void FASTCALL call_8_9_13_2(SH2_struct* sh)
{
  SH2bt(sh,(13<<4)+2);
}
#define init_8_9_13_3
void FASTCALL call_8_9_13_3(SH2_struct* sh)
{
  SH2bt(sh,(13<<4)+3);
}
#define init_8_9_13_4
void FASTCALL call_8_9_13_4(SH2_struct* sh)
{
  SH2bt(sh,(13<<4)+4);
}
#define init_8_9_13_5
void FASTCALL call_8_9_13_5(SH2_struct* sh)
{
  SH2bt(sh,(13<<4)+5);
}
#define init_8_9_13_6
void FASTCALL call_8_9_13_6(SH2_struct* sh)
{
  SH2bt(sh,(13<<4)+6);
}
#define init_8_9_13_7
void FASTCALL call_8_9_13_7(SH2_struct* sh)
{
  SH2bt(sh,(13<<4)+7);
}
#define init_8_9_13_8
void FASTCALL call_8_9_13_8(SH2_struct* sh)
{
  SH2bt(sh,(13<<4)+8);
}
#define init_8_9_13_9
void FASTCALL call_8_9_13_9(SH2_struct* sh)
{
  SH2bt(sh,(13<<4)+9);
}
#define init_8_9_13_10
void FASTCALL call_8_9_13_10(SH2_struct* sh)
{
  SH2bt(sh,(13<<4)+10);
}
#define init_8_9_13_11
void FASTCALL call_8_9_13_11(SH2_struct* sh)
{
  SH2bt(sh,(13<<4)+11);
}
#define init_8_9_13_12
void FASTCALL call_8_9_13_12(SH2_struct* sh)
{
  SH2bt(sh,(13<<4)+12);
}
#define init_8_9_13_13
void FASTCALL call_8_9_13_13(SH2_struct* sh)
{
  SH2bt(sh,(13<<4)+13);
}
#define init_8_9_13_14
void FASTCALL call_8_9_13_14(SH2_struct* sh)
{
  SH2bt(sh,(13<<4)+14);
}
#define init_8_9_13_15
void FASTCALL call_8_9_13_15(SH2_struct* sh)
{
  SH2bt(sh,(13<<4)+15);
}
#define init_8_9_14_0
void FASTCALL call_8_9_14_0(SH2_struct* sh)
{
  SH2bt(sh,(14<<4)+0);
}
#define init_8_9_14_1
void FASTCALL call_8_9_14_1(SH2_struct* sh)
{
  SH2bt(sh,(14<<4)+1);
}
#define init_8_9_14_2
void FASTCALL call_8_9_14_2(SH2_struct* sh)
{
  SH2bt(sh,(14<<4)+2);
}
#define init_8_9_14_3
void FASTCALL call_8_9_14_3(SH2_struct* sh)
{
  SH2bt(sh,(14<<4)+3);
}
#define init_8_9_14_4
void FASTCALL call_8_9_14_4(SH2_struct* sh)
{
  SH2bt(sh,(14<<4)+4);
}
#define init_8_9_14_5
void FASTCALL call_8_9_14_5(SH2_struct* sh)
{
  SH2bt(sh,(14<<4)+5);
}
#define init_8_9_14_6
void FASTCALL call_8_9_14_6(SH2_struct* sh)
{
  SH2bt(sh,(14<<4)+6);
}
#define init_8_9_14_7
void FASTCALL call_8_9_14_7(SH2_struct* sh)
{
  SH2bt(sh,(14<<4)+7);
}
#define init_8_9_14_8
void FASTCALL call_8_9_14_8(SH2_struct* sh)
{
  SH2bt(sh,(14<<4)+8);
}
#define init_8_9_14_9
void FASTCALL call_8_9_14_9(SH2_struct* sh)
{
  SH2bt(sh,(14<<4)+9);
}
#define init_8_9_14_10
void FASTCALL call_8_9_14_10(SH2_struct* sh)
{
  SH2bt(sh,(14<<4)+10);
}
#define init_8_9_14_11
void FASTCALL call_8_9_14_11(SH2_struct* sh)
{
  SH2bt(sh,(14<<4)+11);
}
#define init_8_9_14_12
void FASTCALL call_8_9_14_12(SH2_struct* sh)
{
  SH2bt(sh,(14<<4)+12);
}
#define init_8_9_14_13
void FASTCALL call_8_9_14_13(SH2_struct* sh)
{
  SH2bt(sh,(14<<4)+13);
}
#define init_8_9_14_14
void FASTCALL call_8_9_14_14(SH2_struct* sh)
{
  SH2bt(sh,(14<<4)+14);
}
#define init_8_9_14_15
void FASTCALL call_8_9_14_15(SH2_struct* sh)
{
  SH2bt(sh,(14<<4)+15);
}
#define init_8_9_15_0
void FASTCALL call_8_9_15_0(SH2_struct* sh)
{
  SH2bt(sh,(15<<4)+0);
}
#define init_8_9_15_1
void FASTCALL call_8_9_15_1(SH2_struct* sh)
{
  SH2bt(sh,(15<<4)+1);
}
#define init_8_9_15_2
void FASTCALL call_8_9_15_2(SH2_struct* sh)
{
  SH2bt(sh,(15<<4)+2);
}
#define init_8_9_15_3
void FASTCALL call_8_9_15_3(SH2_struct* sh)
{
  SH2bt(sh,(15<<4)+3);
}
#define init_8_9_15_4
void FASTCALL call_8_9_15_4(SH2_struct* sh)
{
  SH2bt(sh,(15<<4)+4);
}
#define init_8_9_15_5
void FASTCALL call_8_9_15_5(SH2_struct* sh)
{
  SH2bt(sh,(15<<4)+5);
}
#define init_8_9_15_6
void FASTCALL call_8_9_15_6(SH2_struct* sh)
{
  SH2bt(sh,(15<<4)+6);
}
#define init_8_9_15_7
void FASTCALL call_8_9_15_7(SH2_struct* sh)
{
  SH2bt(sh,(15<<4)+7);
}
#define init_8_9_15_8
void FASTCALL call_8_9_15_8(SH2_struct* sh)
{
  SH2bt(sh,(15<<4)+8);
}
#define init_8_9_15_9
void FASTCALL call_8_9_15_9(SH2_struct* sh)
{
  SH2bt(sh,(15<<4)+9);
}
#define init_8_9_15_10
void FASTCALL call_8_9_15_10(SH2_struct* sh)
{
  SH2bt(sh,(15<<4)+10);
}
#define init_8_9_15_11
void FASTCALL call_8_9_15_11(SH2_struct* sh)
{
  SH2bt(sh,(15<<4)+11);
}
#define init_8_9_15_12
void FASTCALL call_8_9_15_12(SH2_struct* sh)
{
  SH2bt(sh,(15<<4)+12);
}
#define init_8_9_15_13
void FASTCALL call_8_9_15_13(SH2_struct* sh)
{
  SH2bt(sh,(15<<4)+13);
}
#define init_8_9_15_14
void FASTCALL call_8_9_15_14(SH2_struct* sh)
{
  SH2bt(sh,(15<<4)+14);
}
#define init_8_9_15_15
void FASTCALL call_8_9_15_15(SH2_struct* sh)
{
  SH2bt(sh,(15<<4)+15);
}
#define init_11_0_0_0
void FASTCALL call_11_0_0_0(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(0<<4)+0);
}
#define init_11_0_0_1
void FASTCALL call_11_0_0_1(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(0<<4)+1);
}
#define init_11_0_0_2
void FASTCALL call_11_0_0_2(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(0<<4)+2);
}
#define init_11_0_0_3
void FASTCALL call_11_0_0_3(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(0<<4)+3);
}
#define init_11_0_0_4
void FASTCALL call_11_0_0_4(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(0<<4)+4);
}
#define init_11_0_0_5
void FASTCALL call_11_0_0_5(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(0<<4)+5);
}
#define init_11_0_0_6
void FASTCALL call_11_0_0_6(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(0<<4)+6);
}
#define init_11_0_0_7
void FASTCALL call_11_0_0_7(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(0<<4)+7);
}
#define init_11_0_0_8
void FASTCALL call_11_0_0_8(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(0<<4)+8);
}
#define init_11_0_0_9
void FASTCALL call_11_0_0_9(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(0<<4)+9);
}
#define init_11_0_0_10
void FASTCALL call_11_0_0_10(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(0<<4)+10);
}
#define init_11_0_0_11
void FASTCALL call_11_0_0_11(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(0<<4)+11);
}
#define init_11_0_0_12
void FASTCALL call_11_0_0_12(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(0<<4)+12);
}
#define init_11_0_0_13
void FASTCALL call_11_0_0_13(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(0<<4)+13);
}
#define init_11_0_0_14
void FASTCALL call_11_0_0_14(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(0<<4)+14);
}
#define init_11_0_0_15
void FASTCALL call_11_0_0_15(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(0<<4)+15);
}
#define init_11_0_1_0
void FASTCALL call_11_0_1_0(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(1<<4)+0);
}
#define init_11_0_1_1
void FASTCALL call_11_0_1_1(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(1<<4)+1);
}
#define init_11_0_1_2
void FASTCALL call_11_0_1_2(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(1<<4)+2);
}
#define init_11_0_1_3
void FASTCALL call_11_0_1_3(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(1<<4)+3);
}
#define init_11_0_1_4
void FASTCALL call_11_0_1_4(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(1<<4)+4);
}
#define init_11_0_1_5
void FASTCALL call_11_0_1_5(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(1<<4)+5);
}
#define init_11_0_1_6
void FASTCALL call_11_0_1_6(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(1<<4)+6);
}
#define init_11_0_1_7
void FASTCALL call_11_0_1_7(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(1<<4)+7);
}
#define init_11_0_1_8
void FASTCALL call_11_0_1_8(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(1<<4)+8);
}
#define init_11_0_1_9
void FASTCALL call_11_0_1_9(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(1<<4)+9);
}
#define init_11_0_1_10
void FASTCALL call_11_0_1_10(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(1<<4)+10);
}
#define init_11_0_1_11
void FASTCALL call_11_0_1_11(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(1<<4)+11);
}
#define init_11_0_1_12
void FASTCALL call_11_0_1_12(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(1<<4)+12);
}
#define init_11_0_1_13
void FASTCALL call_11_0_1_13(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(1<<4)+13);
}
#define init_11_0_1_14
void FASTCALL call_11_0_1_14(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(1<<4)+14);
}
#define init_11_0_1_15
void FASTCALL call_11_0_1_15(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(1<<4)+15);
}
#define init_11_0_2_0
void FASTCALL call_11_0_2_0(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(2<<4)+0);
}
#define init_11_0_2_1
void FASTCALL call_11_0_2_1(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(2<<4)+1);
}
#define init_11_0_2_2
void FASTCALL call_11_0_2_2(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(2<<4)+2);
}
#define init_11_0_2_3
void FASTCALL call_11_0_2_3(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(2<<4)+3);
}
#define init_11_0_2_4
void FASTCALL call_11_0_2_4(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(2<<4)+4);
}
#define init_11_0_2_5
void FASTCALL call_11_0_2_5(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(2<<4)+5);
}
#define init_11_0_2_6
void FASTCALL call_11_0_2_6(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(2<<4)+6);
}
#define init_11_0_2_7
void FASTCALL call_11_0_2_7(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(2<<4)+7);
}
#define init_11_0_2_8
void FASTCALL call_11_0_2_8(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(2<<4)+8);
}
#define init_11_0_2_9
void FASTCALL call_11_0_2_9(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(2<<4)+9);
}
#define init_11_0_2_10
void FASTCALL call_11_0_2_10(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(2<<4)+10);
}
#define init_11_0_2_11
void FASTCALL call_11_0_2_11(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(2<<4)+11);
}
#define init_11_0_2_12
void FASTCALL call_11_0_2_12(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(2<<4)+12);
}
#define init_11_0_2_13
void FASTCALL call_11_0_2_13(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(2<<4)+13);
}
#define init_11_0_2_14
void FASTCALL call_11_0_2_14(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(2<<4)+14);
}
#define init_11_0_2_15
void FASTCALL call_11_0_2_15(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(2<<4)+15);
}
#define init_11_0_3_0
void FASTCALL call_11_0_3_0(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(3<<4)+0);
}
#define init_11_0_3_1
void FASTCALL call_11_0_3_1(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(3<<4)+1);
}
#define init_11_0_3_2
void FASTCALL call_11_0_3_2(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(3<<4)+2);
}
#define init_11_0_3_3
void FASTCALL call_11_0_3_3(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(3<<4)+3);
}
#define init_11_0_3_4
void FASTCALL call_11_0_3_4(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(3<<4)+4);
}
#define init_11_0_3_5
void FASTCALL call_11_0_3_5(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(3<<4)+5);
}
#define init_11_0_3_6
void FASTCALL call_11_0_3_6(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(3<<4)+6);
}
#define init_11_0_3_7
void FASTCALL call_11_0_3_7(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(3<<4)+7);
}
#define init_11_0_3_8
void FASTCALL call_11_0_3_8(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(3<<4)+8);
}
#define init_11_0_3_9
void FASTCALL call_11_0_3_9(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(3<<4)+9);
}
#define init_11_0_3_10
void FASTCALL call_11_0_3_10(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(3<<4)+10);
}
#define init_11_0_3_11
void FASTCALL call_11_0_3_11(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(3<<4)+11);
}
#define init_11_0_3_12
void FASTCALL call_11_0_3_12(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(3<<4)+12);
}
#define init_11_0_3_13
void FASTCALL call_11_0_3_13(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(3<<4)+13);
}
#define init_11_0_3_14
void FASTCALL call_11_0_3_14(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(3<<4)+14);
}
#define init_11_0_3_15
void FASTCALL call_11_0_3_15(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(3<<4)+15);
}
#define init_11_0_4_0
void FASTCALL call_11_0_4_0(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(4<<4)+0);
}
#define init_11_0_4_1
void FASTCALL call_11_0_4_1(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(4<<4)+1);
}
#define init_11_0_4_2
void FASTCALL call_11_0_4_2(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(4<<4)+2);
}
#define init_11_0_4_3
void FASTCALL call_11_0_4_3(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(4<<4)+3);
}
#define init_11_0_4_4
void FASTCALL call_11_0_4_4(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(4<<4)+4);
}
#define init_11_0_4_5
void FASTCALL call_11_0_4_5(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(4<<4)+5);
}
#define init_11_0_4_6
void FASTCALL call_11_0_4_6(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(4<<4)+6);
}
#define init_11_0_4_7
void FASTCALL call_11_0_4_7(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(4<<4)+7);
}
#define init_11_0_4_8
void FASTCALL call_11_0_4_8(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(4<<4)+8);
}
#define init_11_0_4_9
void FASTCALL call_11_0_4_9(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(4<<4)+9);
}
#define init_11_0_4_10
void FASTCALL call_11_0_4_10(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(4<<4)+10);
}
#define init_11_0_4_11
void FASTCALL call_11_0_4_11(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(4<<4)+11);
}
#define init_11_0_4_12
void FASTCALL call_11_0_4_12(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(4<<4)+12);
}
#define init_11_0_4_13
void FASTCALL call_11_0_4_13(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(4<<4)+13);
}
#define init_11_0_4_14
void FASTCALL call_11_0_4_14(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(4<<4)+14);
}
#define init_11_0_4_15
void FASTCALL call_11_0_4_15(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(4<<4)+15);
}
#define init_11_0_5_0
void FASTCALL call_11_0_5_0(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(5<<4)+0);
}
#define init_11_0_5_1
void FASTCALL call_11_0_5_1(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(5<<4)+1);
}
#define init_11_0_5_2
void FASTCALL call_11_0_5_2(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(5<<4)+2);
}
#define init_11_0_5_3
void FASTCALL call_11_0_5_3(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(5<<4)+3);
}
#define init_11_0_5_4
void FASTCALL call_11_0_5_4(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(5<<4)+4);
}
#define init_11_0_5_5
void FASTCALL call_11_0_5_5(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(5<<4)+5);
}
#define init_11_0_5_6
void FASTCALL call_11_0_5_6(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(5<<4)+6);
}
#define init_11_0_5_7
void FASTCALL call_11_0_5_7(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(5<<4)+7);
}
#define init_11_0_5_8
void FASTCALL call_11_0_5_8(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(5<<4)+8);
}
#define init_11_0_5_9
void FASTCALL call_11_0_5_9(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(5<<4)+9);
}
#define init_11_0_5_10
void FASTCALL call_11_0_5_10(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(5<<4)+10);
}
#define init_11_0_5_11
void FASTCALL call_11_0_5_11(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(5<<4)+11);
}
#define init_11_0_5_12
void FASTCALL call_11_0_5_12(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(5<<4)+12);
}
#define init_11_0_5_13
void FASTCALL call_11_0_5_13(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(5<<4)+13);
}
#define init_11_0_5_14
void FASTCALL call_11_0_5_14(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(5<<4)+14);
}
#define init_11_0_5_15
void FASTCALL call_11_0_5_15(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(5<<4)+15);
}
#define init_11_0_6_0
void FASTCALL call_11_0_6_0(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(6<<4)+0);
}
#define init_11_0_6_1
void FASTCALL call_11_0_6_1(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(6<<4)+1);
}
#define init_11_0_6_2
void FASTCALL call_11_0_6_2(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(6<<4)+2);
}
#define init_11_0_6_3
void FASTCALL call_11_0_6_3(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(6<<4)+3);
}
#define init_11_0_6_4
void FASTCALL call_11_0_6_4(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(6<<4)+4);
}
#define init_11_0_6_5
void FASTCALL call_11_0_6_5(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(6<<4)+5);
}
#define init_11_0_6_6
void FASTCALL call_11_0_6_6(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(6<<4)+6);
}
#define init_11_0_6_7
void FASTCALL call_11_0_6_7(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(6<<4)+7);
}
#define init_11_0_6_8
void FASTCALL call_11_0_6_8(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(6<<4)+8);
}
#define init_11_0_6_9
void FASTCALL call_11_0_6_9(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(6<<4)+9);
}
#define init_11_0_6_10
void FASTCALL call_11_0_6_10(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(6<<4)+10);
}
#define init_11_0_6_11
void FASTCALL call_11_0_6_11(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(6<<4)+11);
}
#define init_11_0_6_12
void FASTCALL call_11_0_6_12(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(6<<4)+12);
}
#define init_11_0_6_13
void FASTCALL call_11_0_6_13(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(6<<4)+13);
}
#define init_11_0_6_14
void FASTCALL call_11_0_6_14(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(6<<4)+14);
}
#define init_11_0_6_15
void FASTCALL call_11_0_6_15(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(6<<4)+15);
}
#define init_11_0_7_0
void FASTCALL call_11_0_7_0(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(7<<4)+0);
}
#define init_11_0_7_1
void FASTCALL call_11_0_7_1(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(7<<4)+1);
}
#define init_11_0_7_2
void FASTCALL call_11_0_7_2(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(7<<4)+2);
}
#define init_11_0_7_3
void FASTCALL call_11_0_7_3(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(7<<4)+3);
}
#define init_11_0_7_4
void FASTCALL call_11_0_7_4(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(7<<4)+4);
}
#define init_11_0_7_5
void FASTCALL call_11_0_7_5(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(7<<4)+5);
}
#define init_11_0_7_6
void FASTCALL call_11_0_7_6(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(7<<4)+6);
}
#define init_11_0_7_7
void FASTCALL call_11_0_7_7(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(7<<4)+7);
}
#define init_11_0_7_8
void FASTCALL call_11_0_7_8(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(7<<4)+8);
}
#define init_11_0_7_9
void FASTCALL call_11_0_7_9(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(7<<4)+9);
}
#define init_11_0_7_10
void FASTCALL call_11_0_7_10(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(7<<4)+10);
}
#define init_11_0_7_11
void FASTCALL call_11_0_7_11(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(7<<4)+11);
}
#define init_11_0_7_12
void FASTCALL call_11_0_7_12(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(7<<4)+12);
}
#define init_11_0_7_13
void FASTCALL call_11_0_7_13(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(7<<4)+13);
}
#define init_11_0_7_14
void FASTCALL call_11_0_7_14(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(7<<4)+14);
}
#define init_11_0_7_15
void FASTCALL call_11_0_7_15(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(7<<4)+15);
}
#define init_11_0_8_0
void FASTCALL call_11_0_8_0(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(8<<4)+0);
}
#define init_11_0_8_1
void FASTCALL call_11_0_8_1(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(8<<4)+1);
}
#define init_11_0_8_2
void FASTCALL call_11_0_8_2(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(8<<4)+2);
}
#define init_11_0_8_3
void FASTCALL call_11_0_8_3(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(8<<4)+3);
}
#define init_11_0_8_4
void FASTCALL call_11_0_8_4(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(8<<4)+4);
}
#define init_11_0_8_5
void FASTCALL call_11_0_8_5(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(8<<4)+5);
}
#define init_11_0_8_6
void FASTCALL call_11_0_8_6(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(8<<4)+6);
}
#define init_11_0_8_7
void FASTCALL call_11_0_8_7(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(8<<4)+7);
}
#define init_11_0_8_8
void FASTCALL call_11_0_8_8(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(8<<4)+8);
}
#define init_11_0_8_9
void FASTCALL call_11_0_8_9(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(8<<4)+9);
}
#define init_11_0_8_10
void FASTCALL call_11_0_8_10(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(8<<4)+10);
}
#define init_11_0_8_11
void FASTCALL call_11_0_8_11(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(8<<4)+11);
}
#define init_11_0_8_12
void FASTCALL call_11_0_8_12(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(8<<4)+12);
}
#define init_11_0_8_13
void FASTCALL call_11_0_8_13(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(8<<4)+13);
}
#define init_11_0_8_14
void FASTCALL call_11_0_8_14(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(8<<4)+14);
}
#define init_11_0_8_15
void FASTCALL call_11_0_8_15(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(8<<4)+15);
}
#define init_11_0_9_0
void FASTCALL call_11_0_9_0(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(9<<4)+0);
}
#define init_11_0_9_1
void FASTCALL call_11_0_9_1(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(9<<4)+1);
}
#define init_11_0_9_2
void FASTCALL call_11_0_9_2(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(9<<4)+2);
}
#define init_11_0_9_3
void FASTCALL call_11_0_9_3(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(9<<4)+3);
}
#define init_11_0_9_4
void FASTCALL call_11_0_9_4(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(9<<4)+4);
}
#define init_11_0_9_5
void FASTCALL call_11_0_9_5(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(9<<4)+5);
}
#define init_11_0_9_6
void FASTCALL call_11_0_9_6(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(9<<4)+6);
}
#define init_11_0_9_7
void FASTCALL call_11_0_9_7(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(9<<4)+7);
}
#define init_11_0_9_8
void FASTCALL call_11_0_9_8(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(9<<4)+8);
}
#define init_11_0_9_9
void FASTCALL call_11_0_9_9(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(9<<4)+9);
}
#define init_11_0_9_10
void FASTCALL call_11_0_9_10(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(9<<4)+10);
}
#define init_11_0_9_11
void FASTCALL call_11_0_9_11(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(9<<4)+11);
}
#define init_11_0_9_12
void FASTCALL call_11_0_9_12(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(9<<4)+12);
}
#define init_11_0_9_13
void FASTCALL call_11_0_9_13(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(9<<4)+13);
}
#define init_11_0_9_14
void FASTCALL call_11_0_9_14(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(9<<4)+14);
}
#define init_11_0_9_15
void FASTCALL call_11_0_9_15(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(9<<4)+15);
}
#define init_11_0_10_0
void FASTCALL call_11_0_10_0(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(10<<4)+0);
}
#define init_11_0_10_1
void FASTCALL call_11_0_10_1(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(10<<4)+1);
}
#define init_11_0_10_2
void FASTCALL call_11_0_10_2(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(10<<4)+2);
}
#define init_11_0_10_3
void FASTCALL call_11_0_10_3(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(10<<4)+3);
}
#define init_11_0_10_4
void FASTCALL call_11_0_10_4(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(10<<4)+4);
}
#define init_11_0_10_5
void FASTCALL call_11_0_10_5(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(10<<4)+5);
}
#define init_11_0_10_6
void FASTCALL call_11_0_10_6(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(10<<4)+6);
}
#define init_11_0_10_7
void FASTCALL call_11_0_10_7(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(10<<4)+7);
}
#define init_11_0_10_8
void FASTCALL call_11_0_10_8(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(10<<4)+8);
}
#define init_11_0_10_9
void FASTCALL call_11_0_10_9(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(10<<4)+9);
}
#define init_11_0_10_10
void FASTCALL call_11_0_10_10(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(10<<4)+10);
}
#define init_11_0_10_11
void FASTCALL call_11_0_10_11(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(10<<4)+11);
}
#define init_11_0_10_12
void FASTCALL call_11_0_10_12(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(10<<4)+12);
}
#define init_11_0_10_13
void FASTCALL call_11_0_10_13(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(10<<4)+13);
}
#define init_11_0_10_14
void FASTCALL call_11_0_10_14(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(10<<4)+14);
}
#define init_11_0_10_15
void FASTCALL call_11_0_10_15(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(10<<4)+15);
}
#define init_11_0_11_0
void FASTCALL call_11_0_11_0(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(11<<4)+0);
}
#define init_11_0_11_1
void FASTCALL call_11_0_11_1(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(11<<4)+1);
}
#define init_11_0_11_2
void FASTCALL call_11_0_11_2(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(11<<4)+2);
}
#define init_11_0_11_3
void FASTCALL call_11_0_11_3(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(11<<4)+3);
}
#define init_11_0_11_4
void FASTCALL call_11_0_11_4(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(11<<4)+4);
}
#define init_11_0_11_5
void FASTCALL call_11_0_11_5(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(11<<4)+5);
}
#define init_11_0_11_6
void FASTCALL call_11_0_11_6(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(11<<4)+6);
}
#define init_11_0_11_7
void FASTCALL call_11_0_11_7(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(11<<4)+7);
}
#define init_11_0_11_8
void FASTCALL call_11_0_11_8(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(11<<4)+8);
}
#define init_11_0_11_9
void FASTCALL call_11_0_11_9(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(11<<4)+9);
}
#define init_11_0_11_10
void FASTCALL call_11_0_11_10(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(11<<4)+10);
}
#define init_11_0_11_11
void FASTCALL call_11_0_11_11(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(11<<4)+11);
}
#define init_11_0_11_12
void FASTCALL call_11_0_11_12(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(11<<4)+12);
}
#define init_11_0_11_13
void FASTCALL call_11_0_11_13(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(11<<4)+13);
}
#define init_11_0_11_14
void FASTCALL call_11_0_11_14(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(11<<4)+14);
}
#define init_11_0_11_15
void FASTCALL call_11_0_11_15(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(11<<4)+15);
}
#define init_11_0_12_0
void FASTCALL call_11_0_12_0(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(12<<4)+0);
}
#define init_11_0_12_1
void FASTCALL call_11_0_12_1(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(12<<4)+1);
}
#define init_11_0_12_2
void FASTCALL call_11_0_12_2(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(12<<4)+2);
}
#define init_11_0_12_3
void FASTCALL call_11_0_12_3(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(12<<4)+3);
}
#define init_11_0_12_4
void FASTCALL call_11_0_12_4(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(12<<4)+4);
}
#define init_11_0_12_5
void FASTCALL call_11_0_12_5(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(12<<4)+5);
}
#define init_11_0_12_6
void FASTCALL call_11_0_12_6(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(12<<4)+6);
}
#define init_11_0_12_7
void FASTCALL call_11_0_12_7(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(12<<4)+7);
}
#define init_11_0_12_8
void FASTCALL call_11_0_12_8(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(12<<4)+8);
}
#define init_11_0_12_9
void FASTCALL call_11_0_12_9(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(12<<4)+9);
}
#define init_11_0_12_10
void FASTCALL call_11_0_12_10(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(12<<4)+10);
}
#define init_11_0_12_11
void FASTCALL call_11_0_12_11(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(12<<4)+11);
}
#define init_11_0_12_12
void FASTCALL call_11_0_12_12(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(12<<4)+12);
}
#define init_11_0_12_13
void FASTCALL call_11_0_12_13(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(12<<4)+13);
}
#define init_11_0_12_14
void FASTCALL call_11_0_12_14(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(12<<4)+14);
}
#define init_11_0_12_15
void FASTCALL call_11_0_12_15(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(12<<4)+15);
}
#define init_11_0_13_0
void FASTCALL call_11_0_13_0(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(13<<4)+0);
}
#define init_11_0_13_1
void FASTCALL call_11_0_13_1(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(13<<4)+1);
}
#define init_11_0_13_2
void FASTCALL call_11_0_13_2(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(13<<4)+2);
}
#define init_11_0_13_3
void FASTCALL call_11_0_13_3(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(13<<4)+3);
}
#define init_11_0_13_4
void FASTCALL call_11_0_13_4(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(13<<4)+4);
}
#define init_11_0_13_5
void FASTCALL call_11_0_13_5(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(13<<4)+5);
}
#define init_11_0_13_6
void FASTCALL call_11_0_13_6(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(13<<4)+6);
}
#define init_11_0_13_7
void FASTCALL call_11_0_13_7(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(13<<4)+7);
}
#define init_11_0_13_8
void FASTCALL call_11_0_13_8(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(13<<4)+8);
}
#define init_11_0_13_9
void FASTCALL call_11_0_13_9(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(13<<4)+9);
}
#define init_11_0_13_10
void FASTCALL call_11_0_13_10(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(13<<4)+10);
}
#define init_11_0_13_11
void FASTCALL call_11_0_13_11(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(13<<4)+11);
}
#define init_11_0_13_12
void FASTCALL call_11_0_13_12(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(13<<4)+12);
}
#define init_11_0_13_13
void FASTCALL call_11_0_13_13(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(13<<4)+13);
}
#define init_11_0_13_14
void FASTCALL call_11_0_13_14(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(13<<4)+14);
}
#define init_11_0_13_15
void FASTCALL call_11_0_13_15(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(13<<4)+15);
}
#define init_11_0_14_0
void FASTCALL call_11_0_14_0(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(14<<4)+0);
}
#define init_11_0_14_1
void FASTCALL call_11_0_14_1(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(14<<4)+1);
}
#define init_11_0_14_2
void FASTCALL call_11_0_14_2(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(14<<4)+2);
}
#define init_11_0_14_3
void FASTCALL call_11_0_14_3(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(14<<4)+3);
}
#define init_11_0_14_4
void FASTCALL call_11_0_14_4(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(14<<4)+4);
}
#define init_11_0_14_5
void FASTCALL call_11_0_14_5(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(14<<4)+5);
}
#define init_11_0_14_6
void FASTCALL call_11_0_14_6(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(14<<4)+6);
}
#define init_11_0_14_7
void FASTCALL call_11_0_14_7(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(14<<4)+7);
}
#define init_11_0_14_8
void FASTCALL call_11_0_14_8(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(14<<4)+8);
}
#define init_11_0_14_9
void FASTCALL call_11_0_14_9(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(14<<4)+9);
}
#define init_11_0_14_10
void FASTCALL call_11_0_14_10(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(14<<4)+10);
}
#define init_11_0_14_11
void FASTCALL call_11_0_14_11(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(14<<4)+11);
}
#define init_11_0_14_12
void FASTCALL call_11_0_14_12(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(14<<4)+12);
}
#define init_11_0_14_13
void FASTCALL call_11_0_14_13(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(14<<4)+13);
}
#define init_11_0_14_14
void FASTCALL call_11_0_14_14(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(14<<4)+14);
}
#define init_11_0_14_15
void FASTCALL call_11_0_14_15(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(14<<4)+15);
}
#define init_11_0_15_0
void FASTCALL call_11_0_15_0(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(15<<4)+0);
}
#define init_11_0_15_1
void FASTCALL call_11_0_15_1(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(15<<4)+1);
}
#define init_11_0_15_2
void FASTCALL call_11_0_15_2(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(15<<4)+2);
}
#define init_11_0_15_3
void FASTCALL call_11_0_15_3(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(15<<4)+3);
}
#define init_11_0_15_4
void FASTCALL call_11_0_15_4(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(15<<4)+4);
}
#define init_11_0_15_5
void FASTCALL call_11_0_15_5(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(15<<4)+5);
}
#define init_11_0_15_6
void FASTCALL call_11_0_15_6(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(15<<4)+6);
}
#define init_11_0_15_7
void FASTCALL call_11_0_15_7(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(15<<4)+7);
}
#define init_11_0_15_8
void FASTCALL call_11_0_15_8(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(15<<4)+8);
}
#define init_11_0_15_9
void FASTCALL call_11_0_15_9(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(15<<4)+9);
}
#define init_11_0_15_10
void FASTCALL call_11_0_15_10(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(15<<4)+10);
}
#define init_11_0_15_11
void FASTCALL call_11_0_15_11(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(15<<4)+11);
}
#define init_11_0_15_12
void FASTCALL call_11_0_15_12(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(15<<4)+12);
}
#define init_11_0_15_13
void FASTCALL call_11_0_15_13(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(15<<4)+13);
}
#define init_11_0_15_14
void FASTCALL call_11_0_15_14(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(15<<4)+14);
}
#define init_11_0_15_15
void FASTCALL call_11_0_15_15(SH2_struct* sh)
{
  SH2bsr(sh, (0<<8)+(15<<4)+15);
}
#define init_11_1_0_0
void FASTCALL call_11_1_0_0(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(0<<4)+0);
}
#define init_11_1_0_1
void FASTCALL call_11_1_0_1(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(0<<4)+1);
}
#define init_11_1_0_2
void FASTCALL call_11_1_0_2(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(0<<4)+2);
}
#define init_11_1_0_3
void FASTCALL call_11_1_0_3(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(0<<4)+3);
}
#define init_11_1_0_4
void FASTCALL call_11_1_0_4(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(0<<4)+4);
}
#define init_11_1_0_5
void FASTCALL call_11_1_0_5(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(0<<4)+5);
}
#define init_11_1_0_6
void FASTCALL call_11_1_0_6(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(0<<4)+6);
}
#define init_11_1_0_7
void FASTCALL call_11_1_0_7(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(0<<4)+7);
}
#define init_11_1_0_8
void FASTCALL call_11_1_0_8(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(0<<4)+8);
}
#define init_11_1_0_9
void FASTCALL call_11_1_0_9(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(0<<4)+9);
}
#define init_11_1_0_10
void FASTCALL call_11_1_0_10(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(0<<4)+10);
}
#define init_11_1_0_11
void FASTCALL call_11_1_0_11(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(0<<4)+11);
}
#define init_11_1_0_12
void FASTCALL call_11_1_0_12(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(0<<4)+12);
}
#define init_11_1_0_13
void FASTCALL call_11_1_0_13(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(0<<4)+13);
}
#define init_11_1_0_14
void FASTCALL call_11_1_0_14(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(0<<4)+14);
}
#define init_11_1_0_15
void FASTCALL call_11_1_0_15(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(0<<4)+15);
}
#define init_11_1_1_0
void FASTCALL call_11_1_1_0(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(1<<4)+0);
}
#define init_11_1_1_1
void FASTCALL call_11_1_1_1(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(1<<4)+1);
}
#define init_11_1_1_2
void FASTCALL call_11_1_1_2(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(1<<4)+2);
}
#define init_11_1_1_3
void FASTCALL call_11_1_1_3(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(1<<4)+3);
}
#define init_11_1_1_4
void FASTCALL call_11_1_1_4(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(1<<4)+4);
}
#define init_11_1_1_5
void FASTCALL call_11_1_1_5(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(1<<4)+5);
}
#define init_11_1_1_6
void FASTCALL call_11_1_1_6(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(1<<4)+6);
}
#define init_11_1_1_7
void FASTCALL call_11_1_1_7(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(1<<4)+7);
}
#define init_11_1_1_8
void FASTCALL call_11_1_1_8(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(1<<4)+8);
}
#define init_11_1_1_9
void FASTCALL call_11_1_1_9(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(1<<4)+9);
}
#define init_11_1_1_10
void FASTCALL call_11_1_1_10(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(1<<4)+10);
}
#define init_11_1_1_11
void FASTCALL call_11_1_1_11(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(1<<4)+11);
}
#define init_11_1_1_12
void FASTCALL call_11_1_1_12(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(1<<4)+12);
}
#define init_11_1_1_13
void FASTCALL call_11_1_1_13(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(1<<4)+13);
}
#define init_11_1_1_14
void FASTCALL call_11_1_1_14(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(1<<4)+14);
}
#define init_11_1_1_15
void FASTCALL call_11_1_1_15(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(1<<4)+15);
}
#define init_11_1_2_0
void FASTCALL call_11_1_2_0(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(2<<4)+0);
}
#define init_11_1_2_1
void FASTCALL call_11_1_2_1(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(2<<4)+1);
}
#define init_11_1_2_2
void FASTCALL call_11_1_2_2(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(2<<4)+2);
}
#define init_11_1_2_3
void FASTCALL call_11_1_2_3(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(2<<4)+3);
}
#define init_11_1_2_4
void FASTCALL call_11_1_2_4(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(2<<4)+4);
}
#define init_11_1_2_5
void FASTCALL call_11_1_2_5(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(2<<4)+5);
}
#define init_11_1_2_6
void FASTCALL call_11_1_2_6(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(2<<4)+6);
}
#define init_11_1_2_7
void FASTCALL call_11_1_2_7(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(2<<4)+7);
}
#define init_11_1_2_8
void FASTCALL call_11_1_2_8(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(2<<4)+8);
}
#define init_11_1_2_9
void FASTCALL call_11_1_2_9(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(2<<4)+9);
}
#define init_11_1_2_10
void FASTCALL call_11_1_2_10(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(2<<4)+10);
}
#define init_11_1_2_11
void FASTCALL call_11_1_2_11(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(2<<4)+11);
}
#define init_11_1_2_12
void FASTCALL call_11_1_2_12(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(2<<4)+12);
}
#define init_11_1_2_13
void FASTCALL call_11_1_2_13(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(2<<4)+13);
}
#define init_11_1_2_14
void FASTCALL call_11_1_2_14(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(2<<4)+14);
}
#define init_11_1_2_15
void FASTCALL call_11_1_2_15(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(2<<4)+15);
}
#define init_11_1_3_0
void FASTCALL call_11_1_3_0(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(3<<4)+0);
}
#define init_11_1_3_1
void FASTCALL call_11_1_3_1(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(3<<4)+1);
}
#define init_11_1_3_2
void FASTCALL call_11_1_3_2(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(3<<4)+2);
}
#define init_11_1_3_3
void FASTCALL call_11_1_3_3(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(3<<4)+3);
}
#define init_11_1_3_4
void FASTCALL call_11_1_3_4(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(3<<4)+4);
}
#define init_11_1_3_5
void FASTCALL call_11_1_3_5(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(3<<4)+5);
}
#define init_11_1_3_6
void FASTCALL call_11_1_3_6(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(3<<4)+6);
}
#define init_11_1_3_7
void FASTCALL call_11_1_3_7(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(3<<4)+7);
}
#define init_11_1_3_8
void FASTCALL call_11_1_3_8(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(3<<4)+8);
}
#define init_11_1_3_9
void FASTCALL call_11_1_3_9(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(3<<4)+9);
}
#define init_11_1_3_10
void FASTCALL call_11_1_3_10(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(3<<4)+10);
}
#define init_11_1_3_11
void FASTCALL call_11_1_3_11(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(3<<4)+11);
}
#define init_11_1_3_12
void FASTCALL call_11_1_3_12(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(3<<4)+12);
}
#define init_11_1_3_13
void FASTCALL call_11_1_3_13(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(3<<4)+13);
}
#define init_11_1_3_14
void FASTCALL call_11_1_3_14(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(3<<4)+14);
}
#define init_11_1_3_15
void FASTCALL call_11_1_3_15(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(3<<4)+15);
}
#define init_11_1_4_0
void FASTCALL call_11_1_4_0(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(4<<4)+0);
}
#define init_11_1_4_1
void FASTCALL call_11_1_4_1(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(4<<4)+1);
}
#define init_11_1_4_2
void FASTCALL call_11_1_4_2(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(4<<4)+2);
}
#define init_11_1_4_3
void FASTCALL call_11_1_4_3(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(4<<4)+3);
}
#define init_11_1_4_4
void FASTCALL call_11_1_4_4(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(4<<4)+4);
}
#define init_11_1_4_5
void FASTCALL call_11_1_4_5(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(4<<4)+5);
}
#define init_11_1_4_6
void FASTCALL call_11_1_4_6(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(4<<4)+6);
}
#define init_11_1_4_7
void FASTCALL call_11_1_4_7(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(4<<4)+7);
}
#define init_11_1_4_8
void FASTCALL call_11_1_4_8(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(4<<4)+8);
}
#define init_11_1_4_9
void FASTCALL call_11_1_4_9(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(4<<4)+9);
}
#define init_11_1_4_10
void FASTCALL call_11_1_4_10(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(4<<4)+10);
}
#define init_11_1_4_11
void FASTCALL call_11_1_4_11(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(4<<4)+11);
}
#define init_11_1_4_12
void FASTCALL call_11_1_4_12(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(4<<4)+12);
}
#define init_11_1_4_13
void FASTCALL call_11_1_4_13(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(4<<4)+13);
}
#define init_11_1_4_14
void FASTCALL call_11_1_4_14(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(4<<4)+14);
}
#define init_11_1_4_15
void FASTCALL call_11_1_4_15(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(4<<4)+15);
}
#define init_11_1_5_0
void FASTCALL call_11_1_5_0(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(5<<4)+0);
}
#define init_11_1_5_1
void FASTCALL call_11_1_5_1(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(5<<4)+1);
}
#define init_11_1_5_2
void FASTCALL call_11_1_5_2(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(5<<4)+2);
}
#define init_11_1_5_3
void FASTCALL call_11_1_5_3(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(5<<4)+3);
}
#define init_11_1_5_4
void FASTCALL call_11_1_5_4(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(5<<4)+4);
}
#define init_11_1_5_5
void FASTCALL call_11_1_5_5(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(5<<4)+5);
}
#define init_11_1_5_6
void FASTCALL call_11_1_5_6(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(5<<4)+6);
}
#define init_11_1_5_7
void FASTCALL call_11_1_5_7(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(5<<4)+7);
}
#define init_11_1_5_8
void FASTCALL call_11_1_5_8(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(5<<4)+8);
}
#define init_11_1_5_9
void FASTCALL call_11_1_5_9(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(5<<4)+9);
}
#define init_11_1_5_10
void FASTCALL call_11_1_5_10(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(5<<4)+10);
}
#define init_11_1_5_11
void FASTCALL call_11_1_5_11(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(5<<4)+11);
}
#define init_11_1_5_12
void FASTCALL call_11_1_5_12(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(5<<4)+12);
}
#define init_11_1_5_13
void FASTCALL call_11_1_5_13(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(5<<4)+13);
}
#define init_11_1_5_14
void FASTCALL call_11_1_5_14(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(5<<4)+14);
}
#define init_11_1_5_15
void FASTCALL call_11_1_5_15(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(5<<4)+15);
}
#define init_11_1_6_0
void FASTCALL call_11_1_6_0(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(6<<4)+0);
}
#define init_11_1_6_1
void FASTCALL call_11_1_6_1(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(6<<4)+1);
}
#define init_11_1_6_2
void FASTCALL call_11_1_6_2(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(6<<4)+2);
}
#define init_11_1_6_3
void FASTCALL call_11_1_6_3(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(6<<4)+3);
}
#define init_11_1_6_4
void FASTCALL call_11_1_6_4(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(6<<4)+4);
}
#define init_11_1_6_5
void FASTCALL call_11_1_6_5(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(6<<4)+5);
}
#define init_11_1_6_6
void FASTCALL call_11_1_6_6(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(6<<4)+6);
}
#define init_11_1_6_7
void FASTCALL call_11_1_6_7(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(6<<4)+7);
}
#define init_11_1_6_8
void FASTCALL call_11_1_6_8(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(6<<4)+8);
}
#define init_11_1_6_9
void FASTCALL call_11_1_6_9(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(6<<4)+9);
}
#define init_11_1_6_10
void FASTCALL call_11_1_6_10(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(6<<4)+10);
}
#define init_11_1_6_11
void FASTCALL call_11_1_6_11(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(6<<4)+11);
}
#define init_11_1_6_12
void FASTCALL call_11_1_6_12(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(6<<4)+12);
}
#define init_11_1_6_13
void FASTCALL call_11_1_6_13(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(6<<4)+13);
}
#define init_11_1_6_14
void FASTCALL call_11_1_6_14(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(6<<4)+14);
}
#define init_11_1_6_15
void FASTCALL call_11_1_6_15(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(6<<4)+15);
}
#define init_11_1_7_0
void FASTCALL call_11_1_7_0(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(7<<4)+0);
}
#define init_11_1_7_1
void FASTCALL call_11_1_7_1(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(7<<4)+1);
}
#define init_11_1_7_2
void FASTCALL call_11_1_7_2(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(7<<4)+2);
}
#define init_11_1_7_3
void FASTCALL call_11_1_7_3(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(7<<4)+3);
}
#define init_11_1_7_4
void FASTCALL call_11_1_7_4(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(7<<4)+4);
}
#define init_11_1_7_5
void FASTCALL call_11_1_7_5(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(7<<4)+5);
}
#define init_11_1_7_6
void FASTCALL call_11_1_7_6(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(7<<4)+6);
}
#define init_11_1_7_7
void FASTCALL call_11_1_7_7(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(7<<4)+7);
}
#define init_11_1_7_8
void FASTCALL call_11_1_7_8(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(7<<4)+8);
}
#define init_11_1_7_9
void FASTCALL call_11_1_7_9(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(7<<4)+9);
}
#define init_11_1_7_10
void FASTCALL call_11_1_7_10(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(7<<4)+10);
}
#define init_11_1_7_11
void FASTCALL call_11_1_7_11(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(7<<4)+11);
}
#define init_11_1_7_12
void FASTCALL call_11_1_7_12(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(7<<4)+12);
}
#define init_11_1_7_13
void FASTCALL call_11_1_7_13(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(7<<4)+13);
}
#define init_11_1_7_14
void FASTCALL call_11_1_7_14(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(7<<4)+14);
}
#define init_11_1_7_15
void FASTCALL call_11_1_7_15(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(7<<4)+15);
}
#define init_11_1_8_0
void FASTCALL call_11_1_8_0(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(8<<4)+0);
}
#define init_11_1_8_1
void FASTCALL call_11_1_8_1(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(8<<4)+1);
}
#define init_11_1_8_2
void FASTCALL call_11_1_8_2(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(8<<4)+2);
}
#define init_11_1_8_3
void FASTCALL call_11_1_8_3(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(8<<4)+3);
}
#define init_11_1_8_4
void FASTCALL call_11_1_8_4(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(8<<4)+4);
}
#define init_11_1_8_5
void FASTCALL call_11_1_8_5(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(8<<4)+5);
}
#define init_11_1_8_6
void FASTCALL call_11_1_8_6(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(8<<4)+6);
}
#define init_11_1_8_7
void FASTCALL call_11_1_8_7(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(8<<4)+7);
}
#define init_11_1_8_8
void FASTCALL call_11_1_8_8(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(8<<4)+8);
}
#define init_11_1_8_9
void FASTCALL call_11_1_8_9(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(8<<4)+9);
}
#define init_11_1_8_10
void FASTCALL call_11_1_8_10(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(8<<4)+10);
}
#define init_11_1_8_11
void FASTCALL call_11_1_8_11(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(8<<4)+11);
}
#define init_11_1_8_12
void FASTCALL call_11_1_8_12(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(8<<4)+12);
}
#define init_11_1_8_13
void FASTCALL call_11_1_8_13(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(8<<4)+13);
}
#define init_11_1_8_14
void FASTCALL call_11_1_8_14(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(8<<4)+14);
}
#define init_11_1_8_15
void FASTCALL call_11_1_8_15(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(8<<4)+15);
}
#define init_11_1_9_0
void FASTCALL call_11_1_9_0(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(9<<4)+0);
}
#define init_11_1_9_1
void FASTCALL call_11_1_9_1(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(9<<4)+1);
}
#define init_11_1_9_2
void FASTCALL call_11_1_9_2(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(9<<4)+2);
}
#define init_11_1_9_3
void FASTCALL call_11_1_9_3(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(9<<4)+3);
}
#define init_11_1_9_4
void FASTCALL call_11_1_9_4(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(9<<4)+4);
}
#define init_11_1_9_5
void FASTCALL call_11_1_9_5(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(9<<4)+5);
}
#define init_11_1_9_6
void FASTCALL call_11_1_9_6(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(9<<4)+6);
}
#define init_11_1_9_7
void FASTCALL call_11_1_9_7(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(9<<4)+7);
}
#define init_11_1_9_8
void FASTCALL call_11_1_9_8(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(9<<4)+8);
}
#define init_11_1_9_9
void FASTCALL call_11_1_9_9(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(9<<4)+9);
}
#define init_11_1_9_10
void FASTCALL call_11_1_9_10(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(9<<4)+10);
}
#define init_11_1_9_11
void FASTCALL call_11_1_9_11(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(9<<4)+11);
}
#define init_11_1_9_12
void FASTCALL call_11_1_9_12(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(9<<4)+12);
}
#define init_11_1_9_13
void FASTCALL call_11_1_9_13(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(9<<4)+13);
}
#define init_11_1_9_14
void FASTCALL call_11_1_9_14(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(9<<4)+14);
}
#define init_11_1_9_15
void FASTCALL call_11_1_9_15(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(9<<4)+15);
}
#define init_11_1_10_0
void FASTCALL call_11_1_10_0(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(10<<4)+0);
}
#define init_11_1_10_1
void FASTCALL call_11_1_10_1(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(10<<4)+1);
}
#define init_11_1_10_2
void FASTCALL call_11_1_10_2(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(10<<4)+2);
}
#define init_11_1_10_3
void FASTCALL call_11_1_10_3(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(10<<4)+3);
}
#define init_11_1_10_4
void FASTCALL call_11_1_10_4(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(10<<4)+4);
}
#define init_11_1_10_5
void FASTCALL call_11_1_10_5(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(10<<4)+5);
}
#define init_11_1_10_6
void FASTCALL call_11_1_10_6(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(10<<4)+6);
}
#define init_11_1_10_7
void FASTCALL call_11_1_10_7(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(10<<4)+7);
}
#define init_11_1_10_8
void FASTCALL call_11_1_10_8(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(10<<4)+8);
}
#define init_11_1_10_9
void FASTCALL call_11_1_10_9(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(10<<4)+9);
}
#define init_11_1_10_10
void FASTCALL call_11_1_10_10(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(10<<4)+10);
}
#define init_11_1_10_11
void FASTCALL call_11_1_10_11(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(10<<4)+11);
}
#define init_11_1_10_12
void FASTCALL call_11_1_10_12(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(10<<4)+12);
}
#define init_11_1_10_13
void FASTCALL call_11_1_10_13(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(10<<4)+13);
}
#define init_11_1_10_14
void FASTCALL call_11_1_10_14(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(10<<4)+14);
}
#define init_11_1_10_15
void FASTCALL call_11_1_10_15(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(10<<4)+15);
}
#define init_11_1_11_0
void FASTCALL call_11_1_11_0(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(11<<4)+0);
}
#define init_11_1_11_1
void FASTCALL call_11_1_11_1(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(11<<4)+1);
}
#define init_11_1_11_2
void FASTCALL call_11_1_11_2(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(11<<4)+2);
}
#define init_11_1_11_3
void FASTCALL call_11_1_11_3(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(11<<4)+3);
}
#define init_11_1_11_4
void FASTCALL call_11_1_11_4(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(11<<4)+4);
}
#define init_11_1_11_5
void FASTCALL call_11_1_11_5(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(11<<4)+5);
}
#define init_11_1_11_6
void FASTCALL call_11_1_11_6(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(11<<4)+6);
}
#define init_11_1_11_7
void FASTCALL call_11_1_11_7(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(11<<4)+7);
}
#define init_11_1_11_8
void FASTCALL call_11_1_11_8(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(11<<4)+8);
}
#define init_11_1_11_9
void FASTCALL call_11_1_11_9(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(11<<4)+9);
}
#define init_11_1_11_10
void FASTCALL call_11_1_11_10(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(11<<4)+10);
}
#define init_11_1_11_11
void FASTCALL call_11_1_11_11(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(11<<4)+11);
}
#define init_11_1_11_12
void FASTCALL call_11_1_11_12(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(11<<4)+12);
}
#define init_11_1_11_13
void FASTCALL call_11_1_11_13(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(11<<4)+13);
}
#define init_11_1_11_14
void FASTCALL call_11_1_11_14(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(11<<4)+14);
}
#define init_11_1_11_15
void FASTCALL call_11_1_11_15(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(11<<4)+15);
}
#define init_11_1_12_0
void FASTCALL call_11_1_12_0(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(12<<4)+0);
}
#define init_11_1_12_1
void FASTCALL call_11_1_12_1(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(12<<4)+1);
}
#define init_11_1_12_2
void FASTCALL call_11_1_12_2(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(12<<4)+2);
}
#define init_11_1_12_3
void FASTCALL call_11_1_12_3(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(12<<4)+3);
}
#define init_11_1_12_4
void FASTCALL call_11_1_12_4(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(12<<4)+4);
}
#define init_11_1_12_5
void FASTCALL call_11_1_12_5(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(12<<4)+5);
}
#define init_11_1_12_6
void FASTCALL call_11_1_12_6(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(12<<4)+6);
}
#define init_11_1_12_7
void FASTCALL call_11_1_12_7(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(12<<4)+7);
}
#define init_11_1_12_8
void FASTCALL call_11_1_12_8(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(12<<4)+8);
}
#define init_11_1_12_9
void FASTCALL call_11_1_12_9(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(12<<4)+9);
}
#define init_11_1_12_10
void FASTCALL call_11_1_12_10(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(12<<4)+10);
}
#define init_11_1_12_11
void FASTCALL call_11_1_12_11(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(12<<4)+11);
}
#define init_11_1_12_12
void FASTCALL call_11_1_12_12(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(12<<4)+12);
}
#define init_11_1_12_13
void FASTCALL call_11_1_12_13(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(12<<4)+13);
}
#define init_11_1_12_14
void FASTCALL call_11_1_12_14(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(12<<4)+14);
}
#define init_11_1_12_15
void FASTCALL call_11_1_12_15(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(12<<4)+15);
}
#define init_11_1_13_0
void FASTCALL call_11_1_13_0(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(13<<4)+0);
}
#define init_11_1_13_1
void FASTCALL call_11_1_13_1(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(13<<4)+1);
}
#define init_11_1_13_2
void FASTCALL call_11_1_13_2(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(13<<4)+2);
}
#define init_11_1_13_3
void FASTCALL call_11_1_13_3(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(13<<4)+3);
}
#define init_11_1_13_4
void FASTCALL call_11_1_13_4(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(13<<4)+4);
}
#define init_11_1_13_5
void FASTCALL call_11_1_13_5(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(13<<4)+5);
}
#define init_11_1_13_6
void FASTCALL call_11_1_13_6(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(13<<4)+6);
}
#define init_11_1_13_7
void FASTCALL call_11_1_13_7(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(13<<4)+7);
}
#define init_11_1_13_8
void FASTCALL call_11_1_13_8(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(13<<4)+8);
}
#define init_11_1_13_9
void FASTCALL call_11_1_13_9(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(13<<4)+9);
}
#define init_11_1_13_10
void FASTCALL call_11_1_13_10(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(13<<4)+10);
}
#define init_11_1_13_11
void FASTCALL call_11_1_13_11(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(13<<4)+11);
}
#define init_11_1_13_12
void FASTCALL call_11_1_13_12(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(13<<4)+12);
}
#define init_11_1_13_13
void FASTCALL call_11_1_13_13(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(13<<4)+13);
}
#define init_11_1_13_14
void FASTCALL call_11_1_13_14(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(13<<4)+14);
}
#define init_11_1_13_15
void FASTCALL call_11_1_13_15(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(13<<4)+15);
}
#define init_11_1_14_0
void FASTCALL call_11_1_14_0(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(14<<4)+0);
}
#define init_11_1_14_1
void FASTCALL call_11_1_14_1(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(14<<4)+1);
}
#define init_11_1_14_2
void FASTCALL call_11_1_14_2(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(14<<4)+2);
}
#define init_11_1_14_3
void FASTCALL call_11_1_14_3(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(14<<4)+3);
}
#define init_11_1_14_4
void FASTCALL call_11_1_14_4(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(14<<4)+4);
}
#define init_11_1_14_5
void FASTCALL call_11_1_14_5(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(14<<4)+5);
}
#define init_11_1_14_6
void FASTCALL call_11_1_14_6(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(14<<4)+6);
}
#define init_11_1_14_7
void FASTCALL call_11_1_14_7(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(14<<4)+7);
}
#define init_11_1_14_8
void FASTCALL call_11_1_14_8(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(14<<4)+8);
}
#define init_11_1_14_9
void FASTCALL call_11_1_14_9(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(14<<4)+9);
}
#define init_11_1_14_10
void FASTCALL call_11_1_14_10(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(14<<4)+10);
}
#define init_11_1_14_11
void FASTCALL call_11_1_14_11(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(14<<4)+11);
}
#define init_11_1_14_12
void FASTCALL call_11_1_14_12(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(14<<4)+12);
}
#define init_11_1_14_13
void FASTCALL call_11_1_14_13(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(14<<4)+13);
}
#define init_11_1_14_14
void FASTCALL call_11_1_14_14(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(14<<4)+14);
}
#define init_11_1_14_15
void FASTCALL call_11_1_14_15(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(14<<4)+15);
}
#define init_11_1_15_0
void FASTCALL call_11_1_15_0(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(15<<4)+0);
}
#define init_11_1_15_1
void FASTCALL call_11_1_15_1(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(15<<4)+1);
}
#define init_11_1_15_2
void FASTCALL call_11_1_15_2(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(15<<4)+2);
}
#define init_11_1_15_3
void FASTCALL call_11_1_15_3(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(15<<4)+3);
}
#define init_11_1_15_4
void FASTCALL call_11_1_15_4(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(15<<4)+4);
}
#define init_11_1_15_5
void FASTCALL call_11_1_15_5(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(15<<4)+5);
}
#define init_11_1_15_6
void FASTCALL call_11_1_15_6(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(15<<4)+6);
}
#define init_11_1_15_7
void FASTCALL call_11_1_15_7(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(15<<4)+7);
}
#define init_11_1_15_8
void FASTCALL call_11_1_15_8(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(15<<4)+8);
}
#define init_11_1_15_9
void FASTCALL call_11_1_15_9(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(15<<4)+9);
}
#define init_11_1_15_10
void FASTCALL call_11_1_15_10(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(15<<4)+10);
}
#define init_11_1_15_11
void FASTCALL call_11_1_15_11(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(15<<4)+11);
}
#define init_11_1_15_12
void FASTCALL call_11_1_15_12(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(15<<4)+12);
}
#define init_11_1_15_13
void FASTCALL call_11_1_15_13(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(15<<4)+13);
}
#define init_11_1_15_14
void FASTCALL call_11_1_15_14(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(15<<4)+14);
}
#define init_11_1_15_15
void FASTCALL call_11_1_15_15(SH2_struct* sh)
{
  SH2bsr(sh, (1<<8)+(15<<4)+15);
}
#define init_11_2_0_0
void FASTCALL call_11_2_0_0(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(0<<4)+0);
}
#define init_11_2_0_1
void FASTCALL call_11_2_0_1(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(0<<4)+1);
}
#define init_11_2_0_2
void FASTCALL call_11_2_0_2(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(0<<4)+2);
}
#define init_11_2_0_3
void FASTCALL call_11_2_0_3(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(0<<4)+3);
}
#define init_11_2_0_4
void FASTCALL call_11_2_0_4(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(0<<4)+4);
}
#define init_11_2_0_5
void FASTCALL call_11_2_0_5(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(0<<4)+5);
}
#define init_11_2_0_6
void FASTCALL call_11_2_0_6(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(0<<4)+6);
}
#define init_11_2_0_7
void FASTCALL call_11_2_0_7(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(0<<4)+7);
}
#define init_11_2_0_8
void FASTCALL call_11_2_0_8(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(0<<4)+8);
}
#define init_11_2_0_9
void FASTCALL call_11_2_0_9(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(0<<4)+9);
}
#define init_11_2_0_10
void FASTCALL call_11_2_0_10(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(0<<4)+10);
}
#define init_11_2_0_11
void FASTCALL call_11_2_0_11(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(0<<4)+11);
}
#define init_11_2_0_12
void FASTCALL call_11_2_0_12(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(0<<4)+12);
}
#define init_11_2_0_13
void FASTCALL call_11_2_0_13(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(0<<4)+13);
}
#define init_11_2_0_14
void FASTCALL call_11_2_0_14(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(0<<4)+14);
}
#define init_11_2_0_15
void FASTCALL call_11_2_0_15(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(0<<4)+15);
}
#define init_11_2_1_0
void FASTCALL call_11_2_1_0(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(1<<4)+0);
}
#define init_11_2_1_1
void FASTCALL call_11_2_1_1(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(1<<4)+1);
}
#define init_11_2_1_2
void FASTCALL call_11_2_1_2(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(1<<4)+2);
}
#define init_11_2_1_3
void FASTCALL call_11_2_1_3(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(1<<4)+3);
}
#define init_11_2_1_4
void FASTCALL call_11_2_1_4(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(1<<4)+4);
}
#define init_11_2_1_5
void FASTCALL call_11_2_1_5(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(1<<4)+5);
}
#define init_11_2_1_6
void FASTCALL call_11_2_1_6(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(1<<4)+6);
}
#define init_11_2_1_7
void FASTCALL call_11_2_1_7(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(1<<4)+7);
}
#define init_11_2_1_8
void FASTCALL call_11_2_1_8(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(1<<4)+8);
}
#define init_11_2_1_9
void FASTCALL call_11_2_1_9(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(1<<4)+9);
}
#define init_11_2_1_10
void FASTCALL call_11_2_1_10(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(1<<4)+10);
}
#define init_11_2_1_11
void FASTCALL call_11_2_1_11(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(1<<4)+11);
}
#define init_11_2_1_12
void FASTCALL call_11_2_1_12(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(1<<4)+12);
}
#define init_11_2_1_13
void FASTCALL call_11_2_1_13(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(1<<4)+13);
}
#define init_11_2_1_14
void FASTCALL call_11_2_1_14(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(1<<4)+14);
}
#define init_11_2_1_15
void FASTCALL call_11_2_1_15(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(1<<4)+15);
}
#define init_11_2_2_0
void FASTCALL call_11_2_2_0(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(2<<4)+0);
}
#define init_11_2_2_1
void FASTCALL call_11_2_2_1(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(2<<4)+1);
}
#define init_11_2_2_2
void FASTCALL call_11_2_2_2(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(2<<4)+2);
}
#define init_11_2_2_3
void FASTCALL call_11_2_2_3(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(2<<4)+3);
}
#define init_11_2_2_4
void FASTCALL call_11_2_2_4(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(2<<4)+4);
}
#define init_11_2_2_5
void FASTCALL call_11_2_2_5(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(2<<4)+5);
}
#define init_11_2_2_6
void FASTCALL call_11_2_2_6(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(2<<4)+6);
}
#define init_11_2_2_7
void FASTCALL call_11_2_2_7(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(2<<4)+7);
}
#define init_11_2_2_8
void FASTCALL call_11_2_2_8(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(2<<4)+8);
}
#define init_11_2_2_9
void FASTCALL call_11_2_2_9(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(2<<4)+9);
}
#define init_11_2_2_10
void FASTCALL call_11_2_2_10(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(2<<4)+10);
}
#define init_11_2_2_11
void FASTCALL call_11_2_2_11(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(2<<4)+11);
}
#define init_11_2_2_12
void FASTCALL call_11_2_2_12(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(2<<4)+12);
}
#define init_11_2_2_13
void FASTCALL call_11_2_2_13(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(2<<4)+13);
}
#define init_11_2_2_14
void FASTCALL call_11_2_2_14(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(2<<4)+14);
}
#define init_11_2_2_15
void FASTCALL call_11_2_2_15(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(2<<4)+15);
}
#define init_11_2_3_0
void FASTCALL call_11_2_3_0(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(3<<4)+0);
}
#define init_11_2_3_1
void FASTCALL call_11_2_3_1(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(3<<4)+1);
}
#define init_11_2_3_2
void FASTCALL call_11_2_3_2(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(3<<4)+2);
}
#define init_11_2_3_3
void FASTCALL call_11_2_3_3(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(3<<4)+3);
}
#define init_11_2_3_4
void FASTCALL call_11_2_3_4(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(3<<4)+4);
}
#define init_11_2_3_5
void FASTCALL call_11_2_3_5(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(3<<4)+5);
}
#define init_11_2_3_6
void FASTCALL call_11_2_3_6(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(3<<4)+6);
}
#define init_11_2_3_7
void FASTCALL call_11_2_3_7(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(3<<4)+7);
}
#define init_11_2_3_8
void FASTCALL call_11_2_3_8(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(3<<4)+8);
}
#define init_11_2_3_9
void FASTCALL call_11_2_3_9(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(3<<4)+9);
}
#define init_11_2_3_10
void FASTCALL call_11_2_3_10(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(3<<4)+10);
}
#define init_11_2_3_11
void FASTCALL call_11_2_3_11(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(3<<4)+11);
}
#define init_11_2_3_12
void FASTCALL call_11_2_3_12(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(3<<4)+12);
}
#define init_11_2_3_13
void FASTCALL call_11_2_3_13(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(3<<4)+13);
}
#define init_11_2_3_14
void FASTCALL call_11_2_3_14(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(3<<4)+14);
}
#define init_11_2_3_15
void FASTCALL call_11_2_3_15(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(3<<4)+15);
}
#define init_11_2_4_0
void FASTCALL call_11_2_4_0(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(4<<4)+0);
}
#define init_11_2_4_1
void FASTCALL call_11_2_4_1(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(4<<4)+1);
}
#define init_11_2_4_2
void FASTCALL call_11_2_4_2(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(4<<4)+2);
}
#define init_11_2_4_3
void FASTCALL call_11_2_4_3(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(4<<4)+3);
}
#define init_11_2_4_4
void FASTCALL call_11_2_4_4(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(4<<4)+4);
}
#define init_11_2_4_5
void FASTCALL call_11_2_4_5(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(4<<4)+5);
}
#define init_11_2_4_6
void FASTCALL call_11_2_4_6(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(4<<4)+6);
}
#define init_11_2_4_7
void FASTCALL call_11_2_4_7(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(4<<4)+7);
}
#define init_11_2_4_8
void FASTCALL call_11_2_4_8(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(4<<4)+8);
}
#define init_11_2_4_9
void FASTCALL call_11_2_4_9(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(4<<4)+9);
}
#define init_11_2_4_10
void FASTCALL call_11_2_4_10(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(4<<4)+10);
}
#define init_11_2_4_11
void FASTCALL call_11_2_4_11(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(4<<4)+11);
}
#define init_11_2_4_12
void FASTCALL call_11_2_4_12(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(4<<4)+12);
}
#define init_11_2_4_13
void FASTCALL call_11_2_4_13(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(4<<4)+13);
}
#define init_11_2_4_14
void FASTCALL call_11_2_4_14(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(4<<4)+14);
}
#define init_11_2_4_15
void FASTCALL call_11_2_4_15(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(4<<4)+15);
}
#define init_11_2_5_0
void FASTCALL call_11_2_5_0(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(5<<4)+0);
}
#define init_11_2_5_1
void FASTCALL call_11_2_5_1(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(5<<4)+1);
}
#define init_11_2_5_2
void FASTCALL call_11_2_5_2(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(5<<4)+2);
}
#define init_11_2_5_3
void FASTCALL call_11_2_5_3(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(5<<4)+3);
}
#define init_11_2_5_4
void FASTCALL call_11_2_5_4(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(5<<4)+4);
}
#define init_11_2_5_5
void FASTCALL call_11_2_5_5(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(5<<4)+5);
}
#define init_11_2_5_6
void FASTCALL call_11_2_5_6(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(5<<4)+6);
}
#define init_11_2_5_7
void FASTCALL call_11_2_5_7(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(5<<4)+7);
}
#define init_11_2_5_8
void FASTCALL call_11_2_5_8(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(5<<4)+8);
}
#define init_11_2_5_9
void FASTCALL call_11_2_5_9(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(5<<4)+9);
}
#define init_11_2_5_10
void FASTCALL call_11_2_5_10(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(5<<4)+10);
}
#define init_11_2_5_11
void FASTCALL call_11_2_5_11(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(5<<4)+11);
}
#define init_11_2_5_12
void FASTCALL call_11_2_5_12(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(5<<4)+12);
}
#define init_11_2_5_13
void FASTCALL call_11_2_5_13(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(5<<4)+13);
}
#define init_11_2_5_14
void FASTCALL call_11_2_5_14(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(5<<4)+14);
}
#define init_11_2_5_15
void FASTCALL call_11_2_5_15(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(5<<4)+15);
}
#define init_11_2_6_0
void FASTCALL call_11_2_6_0(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(6<<4)+0);
}
#define init_11_2_6_1
void FASTCALL call_11_2_6_1(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(6<<4)+1);
}
#define init_11_2_6_2
void FASTCALL call_11_2_6_2(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(6<<4)+2);
}
#define init_11_2_6_3
void FASTCALL call_11_2_6_3(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(6<<4)+3);
}
#define init_11_2_6_4
void FASTCALL call_11_2_6_4(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(6<<4)+4);
}
#define init_11_2_6_5
void FASTCALL call_11_2_6_5(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(6<<4)+5);
}
#define init_11_2_6_6
void FASTCALL call_11_2_6_6(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(6<<4)+6);
}
#define init_11_2_6_7
void FASTCALL call_11_2_6_7(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(6<<4)+7);
}
#define init_11_2_6_8
void FASTCALL call_11_2_6_8(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(6<<4)+8);
}
#define init_11_2_6_9
void FASTCALL call_11_2_6_9(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(6<<4)+9);
}
#define init_11_2_6_10
void FASTCALL call_11_2_6_10(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(6<<4)+10);
}
#define init_11_2_6_11
void FASTCALL call_11_2_6_11(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(6<<4)+11);
}
#define init_11_2_6_12
void FASTCALL call_11_2_6_12(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(6<<4)+12);
}
#define init_11_2_6_13
void FASTCALL call_11_2_6_13(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(6<<4)+13);
}
#define init_11_2_6_14
void FASTCALL call_11_2_6_14(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(6<<4)+14);
}
#define init_11_2_6_15
void FASTCALL call_11_2_6_15(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(6<<4)+15);
}
#define init_11_2_7_0
void FASTCALL call_11_2_7_0(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(7<<4)+0);
}
#define init_11_2_7_1
void FASTCALL call_11_2_7_1(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(7<<4)+1);
}
#define init_11_2_7_2
void FASTCALL call_11_2_7_2(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(7<<4)+2);
}
#define init_11_2_7_3
void FASTCALL call_11_2_7_3(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(7<<4)+3);
}
#define init_11_2_7_4
void FASTCALL call_11_2_7_4(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(7<<4)+4);
}
#define init_11_2_7_5
void FASTCALL call_11_2_7_5(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(7<<4)+5);
}
#define init_11_2_7_6
void FASTCALL call_11_2_7_6(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(7<<4)+6);
}
#define init_11_2_7_7
void FASTCALL call_11_2_7_7(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(7<<4)+7);
}
#define init_11_2_7_8
void FASTCALL call_11_2_7_8(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(7<<4)+8);
}
#define init_11_2_7_9
void FASTCALL call_11_2_7_9(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(7<<4)+9);
}
#define init_11_2_7_10
void FASTCALL call_11_2_7_10(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(7<<4)+10);
}
#define init_11_2_7_11
void FASTCALL call_11_2_7_11(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(7<<4)+11);
}
#define init_11_2_7_12
void FASTCALL call_11_2_7_12(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(7<<4)+12);
}
#define init_11_2_7_13
void FASTCALL call_11_2_7_13(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(7<<4)+13);
}
#define init_11_2_7_14
void FASTCALL call_11_2_7_14(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(7<<4)+14);
}
#define init_11_2_7_15
void FASTCALL call_11_2_7_15(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(7<<4)+15);
}
#define init_11_2_8_0
void FASTCALL call_11_2_8_0(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(8<<4)+0);
}
#define init_11_2_8_1
void FASTCALL call_11_2_8_1(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(8<<4)+1);
}
#define init_11_2_8_2
void FASTCALL call_11_2_8_2(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(8<<4)+2);
}
#define init_11_2_8_3
void FASTCALL call_11_2_8_3(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(8<<4)+3);
}
#define init_11_2_8_4
void FASTCALL call_11_2_8_4(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(8<<4)+4);
}
#define init_11_2_8_5
void FASTCALL call_11_2_8_5(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(8<<4)+5);
}
#define init_11_2_8_6
void FASTCALL call_11_2_8_6(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(8<<4)+6);
}
#define init_11_2_8_7
void FASTCALL call_11_2_8_7(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(8<<4)+7);
}
#define init_11_2_8_8
void FASTCALL call_11_2_8_8(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(8<<4)+8);
}
#define init_11_2_8_9
void FASTCALL call_11_2_8_9(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(8<<4)+9);
}
#define init_11_2_8_10
void FASTCALL call_11_2_8_10(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(8<<4)+10);
}
#define init_11_2_8_11
void FASTCALL call_11_2_8_11(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(8<<4)+11);
}
#define init_11_2_8_12
void FASTCALL call_11_2_8_12(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(8<<4)+12);
}
#define init_11_2_8_13
void FASTCALL call_11_2_8_13(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(8<<4)+13);
}
#define init_11_2_8_14
void FASTCALL call_11_2_8_14(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(8<<4)+14);
}
#define init_11_2_8_15
void FASTCALL call_11_2_8_15(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(8<<4)+15);
}
#define init_11_2_9_0
void FASTCALL call_11_2_9_0(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(9<<4)+0);
}
#define init_11_2_9_1
void FASTCALL call_11_2_9_1(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(9<<4)+1);
}
#define init_11_2_9_2
void FASTCALL call_11_2_9_2(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(9<<4)+2);
}
#define init_11_2_9_3
void FASTCALL call_11_2_9_3(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(9<<4)+3);
}
#define init_11_2_9_4
void FASTCALL call_11_2_9_4(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(9<<4)+4);
}
#define init_11_2_9_5
void FASTCALL call_11_2_9_5(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(9<<4)+5);
}
#define init_11_2_9_6
void FASTCALL call_11_2_9_6(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(9<<4)+6);
}
#define init_11_2_9_7
void FASTCALL call_11_2_9_7(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(9<<4)+7);
}
#define init_11_2_9_8
void FASTCALL call_11_2_9_8(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(9<<4)+8);
}
#define init_11_2_9_9
void FASTCALL call_11_2_9_9(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(9<<4)+9);
}
#define init_11_2_9_10
void FASTCALL call_11_2_9_10(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(9<<4)+10);
}
#define init_11_2_9_11
void FASTCALL call_11_2_9_11(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(9<<4)+11);
}
#define init_11_2_9_12
void FASTCALL call_11_2_9_12(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(9<<4)+12);
}
#define init_11_2_9_13
void FASTCALL call_11_2_9_13(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(9<<4)+13);
}
#define init_11_2_9_14
void FASTCALL call_11_2_9_14(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(9<<4)+14);
}
#define init_11_2_9_15
void FASTCALL call_11_2_9_15(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(9<<4)+15);
}
#define init_11_2_10_0
void FASTCALL call_11_2_10_0(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(10<<4)+0);
}
#define init_11_2_10_1
void FASTCALL call_11_2_10_1(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(10<<4)+1);
}
#define init_11_2_10_2
void FASTCALL call_11_2_10_2(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(10<<4)+2);
}
#define init_11_2_10_3
void FASTCALL call_11_2_10_3(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(10<<4)+3);
}
#define init_11_2_10_4
void FASTCALL call_11_2_10_4(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(10<<4)+4);
}
#define init_11_2_10_5
void FASTCALL call_11_2_10_5(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(10<<4)+5);
}
#define init_11_2_10_6
void FASTCALL call_11_2_10_6(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(10<<4)+6);
}
#define init_11_2_10_7
void FASTCALL call_11_2_10_7(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(10<<4)+7);
}
#define init_11_2_10_8
void FASTCALL call_11_2_10_8(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(10<<4)+8);
}
#define init_11_2_10_9
void FASTCALL call_11_2_10_9(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(10<<4)+9);
}
#define init_11_2_10_10
void FASTCALL call_11_2_10_10(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(10<<4)+10);
}
#define init_11_2_10_11
void FASTCALL call_11_2_10_11(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(10<<4)+11);
}
#define init_11_2_10_12
void FASTCALL call_11_2_10_12(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(10<<4)+12);
}
#define init_11_2_10_13
void FASTCALL call_11_2_10_13(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(10<<4)+13);
}
#define init_11_2_10_14
void FASTCALL call_11_2_10_14(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(10<<4)+14);
}
#define init_11_2_10_15
void FASTCALL call_11_2_10_15(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(10<<4)+15);
}
#define init_11_2_11_0
void FASTCALL call_11_2_11_0(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(11<<4)+0);
}
#define init_11_2_11_1
void FASTCALL call_11_2_11_1(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(11<<4)+1);
}
#define init_11_2_11_2
void FASTCALL call_11_2_11_2(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(11<<4)+2);
}
#define init_11_2_11_3
void FASTCALL call_11_2_11_3(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(11<<4)+3);
}
#define init_11_2_11_4
void FASTCALL call_11_2_11_4(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(11<<4)+4);
}
#define init_11_2_11_5
void FASTCALL call_11_2_11_5(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(11<<4)+5);
}
#define init_11_2_11_6
void FASTCALL call_11_2_11_6(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(11<<4)+6);
}
#define init_11_2_11_7
void FASTCALL call_11_2_11_7(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(11<<4)+7);
}
#define init_11_2_11_8
void FASTCALL call_11_2_11_8(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(11<<4)+8);
}
#define init_11_2_11_9
void FASTCALL call_11_2_11_9(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(11<<4)+9);
}
#define init_11_2_11_10
void FASTCALL call_11_2_11_10(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(11<<4)+10);
}
#define init_11_2_11_11
void FASTCALL call_11_2_11_11(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(11<<4)+11);
}
#define init_11_2_11_12
void FASTCALL call_11_2_11_12(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(11<<4)+12);
}
#define init_11_2_11_13
void FASTCALL call_11_2_11_13(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(11<<4)+13);
}
#define init_11_2_11_14
void FASTCALL call_11_2_11_14(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(11<<4)+14);
}
#define init_11_2_11_15
void FASTCALL call_11_2_11_15(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(11<<4)+15);
}
#define init_11_2_12_0
void FASTCALL call_11_2_12_0(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(12<<4)+0);
}
#define init_11_2_12_1
void FASTCALL call_11_2_12_1(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(12<<4)+1);
}
#define init_11_2_12_2
void FASTCALL call_11_2_12_2(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(12<<4)+2);
}
#define init_11_2_12_3
void FASTCALL call_11_2_12_3(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(12<<4)+3);
}
#define init_11_2_12_4
void FASTCALL call_11_2_12_4(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(12<<4)+4);
}
#define init_11_2_12_5
void FASTCALL call_11_2_12_5(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(12<<4)+5);
}
#define init_11_2_12_6
void FASTCALL call_11_2_12_6(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(12<<4)+6);
}
#define init_11_2_12_7
void FASTCALL call_11_2_12_7(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(12<<4)+7);
}
#define init_11_2_12_8
void FASTCALL call_11_2_12_8(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(12<<4)+8);
}
#define init_11_2_12_9
void FASTCALL call_11_2_12_9(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(12<<4)+9);
}
#define init_11_2_12_10
void FASTCALL call_11_2_12_10(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(12<<4)+10);
}
#define init_11_2_12_11
void FASTCALL call_11_2_12_11(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(12<<4)+11);
}
#define init_11_2_12_12
void FASTCALL call_11_2_12_12(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(12<<4)+12);
}
#define init_11_2_12_13
void FASTCALL call_11_2_12_13(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(12<<4)+13);
}
#define init_11_2_12_14
void FASTCALL call_11_2_12_14(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(12<<4)+14);
}
#define init_11_2_12_15
void FASTCALL call_11_2_12_15(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(12<<4)+15);
}
#define init_11_2_13_0
void FASTCALL call_11_2_13_0(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(13<<4)+0);
}
#define init_11_2_13_1
void FASTCALL call_11_2_13_1(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(13<<4)+1);
}
#define init_11_2_13_2
void FASTCALL call_11_2_13_2(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(13<<4)+2);
}
#define init_11_2_13_3
void FASTCALL call_11_2_13_3(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(13<<4)+3);
}
#define init_11_2_13_4
void FASTCALL call_11_2_13_4(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(13<<4)+4);
}
#define init_11_2_13_5
void FASTCALL call_11_2_13_5(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(13<<4)+5);
}
#define init_11_2_13_6
void FASTCALL call_11_2_13_6(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(13<<4)+6);
}
#define init_11_2_13_7
void FASTCALL call_11_2_13_7(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(13<<4)+7);
}
#define init_11_2_13_8
void FASTCALL call_11_2_13_8(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(13<<4)+8);
}
#define init_11_2_13_9
void FASTCALL call_11_2_13_9(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(13<<4)+9);
}
#define init_11_2_13_10
void FASTCALL call_11_2_13_10(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(13<<4)+10);
}
#define init_11_2_13_11
void FASTCALL call_11_2_13_11(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(13<<4)+11);
}
#define init_11_2_13_12
void FASTCALL call_11_2_13_12(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(13<<4)+12);
}
#define init_11_2_13_13
void FASTCALL call_11_2_13_13(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(13<<4)+13);
}
#define init_11_2_13_14
void FASTCALL call_11_2_13_14(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(13<<4)+14);
}
#define init_11_2_13_15
void FASTCALL call_11_2_13_15(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(13<<4)+15);
}
#define init_11_2_14_0
void FASTCALL call_11_2_14_0(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(14<<4)+0);
}
#define init_11_2_14_1
void FASTCALL call_11_2_14_1(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(14<<4)+1);
}
#define init_11_2_14_2
void FASTCALL call_11_2_14_2(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(14<<4)+2);
}
#define init_11_2_14_3
void FASTCALL call_11_2_14_3(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(14<<4)+3);
}
#define init_11_2_14_4
void FASTCALL call_11_2_14_4(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(14<<4)+4);
}
#define init_11_2_14_5
void FASTCALL call_11_2_14_5(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(14<<4)+5);
}
#define init_11_2_14_6
void FASTCALL call_11_2_14_6(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(14<<4)+6);
}
#define init_11_2_14_7
void FASTCALL call_11_2_14_7(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(14<<4)+7);
}
#define init_11_2_14_8
void FASTCALL call_11_2_14_8(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(14<<4)+8);
}
#define init_11_2_14_9
void FASTCALL call_11_2_14_9(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(14<<4)+9);
}
#define init_11_2_14_10
void FASTCALL call_11_2_14_10(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(14<<4)+10);
}
#define init_11_2_14_11
void FASTCALL call_11_2_14_11(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(14<<4)+11);
}
#define init_11_2_14_12
void FASTCALL call_11_2_14_12(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(14<<4)+12);
}
#define init_11_2_14_13
void FASTCALL call_11_2_14_13(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(14<<4)+13);
}
#define init_11_2_14_14
void FASTCALL call_11_2_14_14(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(14<<4)+14);
}
#define init_11_2_14_15
void FASTCALL call_11_2_14_15(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(14<<4)+15);
}
#define init_11_2_15_0
void FASTCALL call_11_2_15_0(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(15<<4)+0);
}
#define init_11_2_15_1
void FASTCALL call_11_2_15_1(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(15<<4)+1);
}
#define init_11_2_15_2
void FASTCALL call_11_2_15_2(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(15<<4)+2);
}
#define init_11_2_15_3
void FASTCALL call_11_2_15_3(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(15<<4)+3);
}
#define init_11_2_15_4
void FASTCALL call_11_2_15_4(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(15<<4)+4);
}
#define init_11_2_15_5
void FASTCALL call_11_2_15_5(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(15<<4)+5);
}
#define init_11_2_15_6
void FASTCALL call_11_2_15_6(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(15<<4)+6);
}
#define init_11_2_15_7
void FASTCALL call_11_2_15_7(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(15<<4)+7);
}
#define init_11_2_15_8
void FASTCALL call_11_2_15_8(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(15<<4)+8);
}
#define init_11_2_15_9
void FASTCALL call_11_2_15_9(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(15<<4)+9);
}
#define init_11_2_15_10
void FASTCALL call_11_2_15_10(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(15<<4)+10);
}
#define init_11_2_15_11
void FASTCALL call_11_2_15_11(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(15<<4)+11);
}
#define init_11_2_15_12
void FASTCALL call_11_2_15_12(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(15<<4)+12);
}
#define init_11_2_15_13
void FASTCALL call_11_2_15_13(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(15<<4)+13);
}
#define init_11_2_15_14
void FASTCALL call_11_2_15_14(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(15<<4)+14);
}
#define init_11_2_15_15
void FASTCALL call_11_2_15_15(SH2_struct* sh)
{
  SH2bsr(sh, (2<<8)+(15<<4)+15);
}
#define init_11_3_0_0
void FASTCALL call_11_3_0_0(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(0<<4)+0);
}
#define init_11_3_0_1
void FASTCALL call_11_3_0_1(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(0<<4)+1);
}
#define init_11_3_0_2
void FASTCALL call_11_3_0_2(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(0<<4)+2);
}
#define init_11_3_0_3
void FASTCALL call_11_3_0_3(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(0<<4)+3);
}
#define init_11_3_0_4
void FASTCALL call_11_3_0_4(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(0<<4)+4);
}
#define init_11_3_0_5
void FASTCALL call_11_3_0_5(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(0<<4)+5);
}
#define init_11_3_0_6
void FASTCALL call_11_3_0_6(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(0<<4)+6);
}
#define init_11_3_0_7
void FASTCALL call_11_3_0_7(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(0<<4)+7);
}
#define init_11_3_0_8
void FASTCALL call_11_3_0_8(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(0<<4)+8);
}
#define init_11_3_0_9
void FASTCALL call_11_3_0_9(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(0<<4)+9);
}
#define init_11_3_0_10
void FASTCALL call_11_3_0_10(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(0<<4)+10);
}
#define init_11_3_0_11
void FASTCALL call_11_3_0_11(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(0<<4)+11);
}
#define init_11_3_0_12
void FASTCALL call_11_3_0_12(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(0<<4)+12);
}
#define init_11_3_0_13
void FASTCALL call_11_3_0_13(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(0<<4)+13);
}
#define init_11_3_0_14
void FASTCALL call_11_3_0_14(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(0<<4)+14);
}
#define init_11_3_0_15
void FASTCALL call_11_3_0_15(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(0<<4)+15);
}
#define init_11_3_1_0
void FASTCALL call_11_3_1_0(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(1<<4)+0);
}
#define init_11_3_1_1
void FASTCALL call_11_3_1_1(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(1<<4)+1);
}
#define init_11_3_1_2
void FASTCALL call_11_3_1_2(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(1<<4)+2);
}
#define init_11_3_1_3
void FASTCALL call_11_3_1_3(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(1<<4)+3);
}
#define init_11_3_1_4
void FASTCALL call_11_3_1_4(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(1<<4)+4);
}
#define init_11_3_1_5
void FASTCALL call_11_3_1_5(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(1<<4)+5);
}
#define init_11_3_1_6
void FASTCALL call_11_3_1_6(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(1<<4)+6);
}
#define init_11_3_1_7
void FASTCALL call_11_3_1_7(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(1<<4)+7);
}
#define init_11_3_1_8
void FASTCALL call_11_3_1_8(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(1<<4)+8);
}
#define init_11_3_1_9
void FASTCALL call_11_3_1_9(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(1<<4)+9);
}
#define init_11_3_1_10
void FASTCALL call_11_3_1_10(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(1<<4)+10);
}
#define init_11_3_1_11
void FASTCALL call_11_3_1_11(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(1<<4)+11);
}
#define init_11_3_1_12
void FASTCALL call_11_3_1_12(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(1<<4)+12);
}
#define init_11_3_1_13
void FASTCALL call_11_3_1_13(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(1<<4)+13);
}
#define init_11_3_1_14
void FASTCALL call_11_3_1_14(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(1<<4)+14);
}
#define init_11_3_1_15
void FASTCALL call_11_3_1_15(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(1<<4)+15);
}
#define init_11_3_2_0
void FASTCALL call_11_3_2_0(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(2<<4)+0);
}
#define init_11_3_2_1
void FASTCALL call_11_3_2_1(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(2<<4)+1);
}
#define init_11_3_2_2
void FASTCALL call_11_3_2_2(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(2<<4)+2);
}
#define init_11_3_2_3
void FASTCALL call_11_3_2_3(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(2<<4)+3);
}
#define init_11_3_2_4
void FASTCALL call_11_3_2_4(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(2<<4)+4);
}
#define init_11_3_2_5
void FASTCALL call_11_3_2_5(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(2<<4)+5);
}
#define init_11_3_2_6
void FASTCALL call_11_3_2_6(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(2<<4)+6);
}
#define init_11_3_2_7
void FASTCALL call_11_3_2_7(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(2<<4)+7);
}
#define init_11_3_2_8
void FASTCALL call_11_3_2_8(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(2<<4)+8);
}
#define init_11_3_2_9
void FASTCALL call_11_3_2_9(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(2<<4)+9);
}
#define init_11_3_2_10
void FASTCALL call_11_3_2_10(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(2<<4)+10);
}
#define init_11_3_2_11
void FASTCALL call_11_3_2_11(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(2<<4)+11);
}
#define init_11_3_2_12
void FASTCALL call_11_3_2_12(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(2<<4)+12);
}
#define init_11_3_2_13
void FASTCALL call_11_3_2_13(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(2<<4)+13);
}
#define init_11_3_2_14
void FASTCALL call_11_3_2_14(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(2<<4)+14);
}
#define init_11_3_2_15
void FASTCALL call_11_3_2_15(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(2<<4)+15);
}
#define init_11_3_3_0
void FASTCALL call_11_3_3_0(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(3<<4)+0);
}
#define init_11_3_3_1
void FASTCALL call_11_3_3_1(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(3<<4)+1);
}
#define init_11_3_3_2
void FASTCALL call_11_3_3_2(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(3<<4)+2);
}
#define init_11_3_3_3
void FASTCALL call_11_3_3_3(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(3<<4)+3);
}
#define init_11_3_3_4
void FASTCALL call_11_3_3_4(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(3<<4)+4);
}
#define init_11_3_3_5
void FASTCALL call_11_3_3_5(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(3<<4)+5);
}
#define init_11_3_3_6
void FASTCALL call_11_3_3_6(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(3<<4)+6);
}
#define init_11_3_3_7
void FASTCALL call_11_3_3_7(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(3<<4)+7);
}
#define init_11_3_3_8
void FASTCALL call_11_3_3_8(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(3<<4)+8);
}
#define init_11_3_3_9
void FASTCALL call_11_3_3_9(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(3<<4)+9);
}
#define init_11_3_3_10
void FASTCALL call_11_3_3_10(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(3<<4)+10);
}
#define init_11_3_3_11
void FASTCALL call_11_3_3_11(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(3<<4)+11);
}
#define init_11_3_3_12
void FASTCALL call_11_3_3_12(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(3<<4)+12);
}
#define init_11_3_3_13
void FASTCALL call_11_3_3_13(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(3<<4)+13);
}
#define init_11_3_3_14
void FASTCALL call_11_3_3_14(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(3<<4)+14);
}
#define init_11_3_3_15
void FASTCALL call_11_3_3_15(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(3<<4)+15);
}
#define init_11_3_4_0
void FASTCALL call_11_3_4_0(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(4<<4)+0);
}
#define init_11_3_4_1
void FASTCALL call_11_3_4_1(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(4<<4)+1);
}
#define init_11_3_4_2
void FASTCALL call_11_3_4_2(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(4<<4)+2);
}
#define init_11_3_4_3
void FASTCALL call_11_3_4_3(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(4<<4)+3);
}
#define init_11_3_4_4
void FASTCALL call_11_3_4_4(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(4<<4)+4);
}
#define init_11_3_4_5
void FASTCALL call_11_3_4_5(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(4<<4)+5);
}
#define init_11_3_4_6
void FASTCALL call_11_3_4_6(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(4<<4)+6);
}
#define init_11_3_4_7
void FASTCALL call_11_3_4_7(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(4<<4)+7);
}
#define init_11_3_4_8
void FASTCALL call_11_3_4_8(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(4<<4)+8);
}
#define init_11_3_4_9
void FASTCALL call_11_3_4_9(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(4<<4)+9);
}
#define init_11_3_4_10
void FASTCALL call_11_3_4_10(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(4<<4)+10);
}
#define init_11_3_4_11
void FASTCALL call_11_3_4_11(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(4<<4)+11);
}
#define init_11_3_4_12
void FASTCALL call_11_3_4_12(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(4<<4)+12);
}
#define init_11_3_4_13
void FASTCALL call_11_3_4_13(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(4<<4)+13);
}
#define init_11_3_4_14
void FASTCALL call_11_3_4_14(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(4<<4)+14);
}
#define init_11_3_4_15
void FASTCALL call_11_3_4_15(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(4<<4)+15);
}
#define init_11_3_5_0
void FASTCALL call_11_3_5_0(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(5<<4)+0);
}
#define init_11_3_5_1
void FASTCALL call_11_3_5_1(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(5<<4)+1);
}
#define init_11_3_5_2
void FASTCALL call_11_3_5_2(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(5<<4)+2);
}
#define init_11_3_5_3
void FASTCALL call_11_3_5_3(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(5<<4)+3);
}
#define init_11_3_5_4
void FASTCALL call_11_3_5_4(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(5<<4)+4);
}
#define init_11_3_5_5
void FASTCALL call_11_3_5_5(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(5<<4)+5);
}
#define init_11_3_5_6
void FASTCALL call_11_3_5_6(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(5<<4)+6);
}
#define init_11_3_5_7
void FASTCALL call_11_3_5_7(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(5<<4)+7);
}
#define init_11_3_5_8
void FASTCALL call_11_3_5_8(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(5<<4)+8);
}
#define init_11_3_5_9
void FASTCALL call_11_3_5_9(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(5<<4)+9);
}
#define init_11_3_5_10
void FASTCALL call_11_3_5_10(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(5<<4)+10);
}
#define init_11_3_5_11
void FASTCALL call_11_3_5_11(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(5<<4)+11);
}
#define init_11_3_5_12
void FASTCALL call_11_3_5_12(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(5<<4)+12);
}
#define init_11_3_5_13
void FASTCALL call_11_3_5_13(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(5<<4)+13);
}
#define init_11_3_5_14
void FASTCALL call_11_3_5_14(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(5<<4)+14);
}
#define init_11_3_5_15
void FASTCALL call_11_3_5_15(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(5<<4)+15);
}
#define init_11_3_6_0
void FASTCALL call_11_3_6_0(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(6<<4)+0);
}
#define init_11_3_6_1
void FASTCALL call_11_3_6_1(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(6<<4)+1);
}
#define init_11_3_6_2
void FASTCALL call_11_3_6_2(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(6<<4)+2);
}
#define init_11_3_6_3
void FASTCALL call_11_3_6_3(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(6<<4)+3);
}
#define init_11_3_6_4
void FASTCALL call_11_3_6_4(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(6<<4)+4);
}
#define init_11_3_6_5
void FASTCALL call_11_3_6_5(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(6<<4)+5);
}
#define init_11_3_6_6
void FASTCALL call_11_3_6_6(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(6<<4)+6);
}
#define init_11_3_6_7
void FASTCALL call_11_3_6_7(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(6<<4)+7);
}
#define init_11_3_6_8
void FASTCALL call_11_3_6_8(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(6<<4)+8);
}
#define init_11_3_6_9
void FASTCALL call_11_3_6_9(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(6<<4)+9);
}
#define init_11_3_6_10
void FASTCALL call_11_3_6_10(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(6<<4)+10);
}
#define init_11_3_6_11
void FASTCALL call_11_3_6_11(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(6<<4)+11);
}
#define init_11_3_6_12
void FASTCALL call_11_3_6_12(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(6<<4)+12);
}
#define init_11_3_6_13
void FASTCALL call_11_3_6_13(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(6<<4)+13);
}
#define init_11_3_6_14
void FASTCALL call_11_3_6_14(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(6<<4)+14);
}
#define init_11_3_6_15
void FASTCALL call_11_3_6_15(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(6<<4)+15);
}
#define init_11_3_7_0
void FASTCALL call_11_3_7_0(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(7<<4)+0);
}
#define init_11_3_7_1
void FASTCALL call_11_3_7_1(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(7<<4)+1);
}
#define init_11_3_7_2
void FASTCALL call_11_3_7_2(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(7<<4)+2);
}
#define init_11_3_7_3
void FASTCALL call_11_3_7_3(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(7<<4)+3);
}
#define init_11_3_7_4
void FASTCALL call_11_3_7_4(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(7<<4)+4);
}
#define init_11_3_7_5
void FASTCALL call_11_3_7_5(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(7<<4)+5);
}
#define init_11_3_7_6
void FASTCALL call_11_3_7_6(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(7<<4)+6);
}
#define init_11_3_7_7
void FASTCALL call_11_3_7_7(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(7<<4)+7);
}
#define init_11_3_7_8
void FASTCALL call_11_3_7_8(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(7<<4)+8);
}
#define init_11_3_7_9
void FASTCALL call_11_3_7_9(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(7<<4)+9);
}
#define init_11_3_7_10
void FASTCALL call_11_3_7_10(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(7<<4)+10);
}
#define init_11_3_7_11
void FASTCALL call_11_3_7_11(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(7<<4)+11);
}
#define init_11_3_7_12
void FASTCALL call_11_3_7_12(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(7<<4)+12);
}
#define init_11_3_7_13
void FASTCALL call_11_3_7_13(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(7<<4)+13);
}
#define init_11_3_7_14
void FASTCALL call_11_3_7_14(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(7<<4)+14);
}
#define init_11_3_7_15
void FASTCALL call_11_3_7_15(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(7<<4)+15);
}
#define init_11_3_8_0
void FASTCALL call_11_3_8_0(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(8<<4)+0);
}
#define init_11_3_8_1
void FASTCALL call_11_3_8_1(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(8<<4)+1);
}
#define init_11_3_8_2
void FASTCALL call_11_3_8_2(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(8<<4)+2);
}
#define init_11_3_8_3
void FASTCALL call_11_3_8_3(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(8<<4)+3);
}
#define init_11_3_8_4
void FASTCALL call_11_3_8_4(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(8<<4)+4);
}
#define init_11_3_8_5
void FASTCALL call_11_3_8_5(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(8<<4)+5);
}
#define init_11_3_8_6
void FASTCALL call_11_3_8_6(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(8<<4)+6);
}
#define init_11_3_8_7
void FASTCALL call_11_3_8_7(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(8<<4)+7);
}
#define init_11_3_8_8
void FASTCALL call_11_3_8_8(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(8<<4)+8);
}
#define init_11_3_8_9
void FASTCALL call_11_3_8_9(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(8<<4)+9);
}
#define init_11_3_8_10
void FASTCALL call_11_3_8_10(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(8<<4)+10);
}
#define init_11_3_8_11
void FASTCALL call_11_3_8_11(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(8<<4)+11);
}
#define init_11_3_8_12
void FASTCALL call_11_3_8_12(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(8<<4)+12);
}
#define init_11_3_8_13
void FASTCALL call_11_3_8_13(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(8<<4)+13);
}
#define init_11_3_8_14
void FASTCALL call_11_3_8_14(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(8<<4)+14);
}
#define init_11_3_8_15
void FASTCALL call_11_3_8_15(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(8<<4)+15);
}
#define init_11_3_9_0
void FASTCALL call_11_3_9_0(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(9<<4)+0);
}
#define init_11_3_9_1
void FASTCALL call_11_3_9_1(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(9<<4)+1);
}
#define init_11_3_9_2
void FASTCALL call_11_3_9_2(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(9<<4)+2);
}
#define init_11_3_9_3
void FASTCALL call_11_3_9_3(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(9<<4)+3);
}
#define init_11_3_9_4
void FASTCALL call_11_3_9_4(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(9<<4)+4);
}
#define init_11_3_9_5
void FASTCALL call_11_3_9_5(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(9<<4)+5);
}
#define init_11_3_9_6
void FASTCALL call_11_3_9_6(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(9<<4)+6);
}
#define init_11_3_9_7
void FASTCALL call_11_3_9_7(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(9<<4)+7);
}
#define init_11_3_9_8
void FASTCALL call_11_3_9_8(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(9<<4)+8);
}
#define init_11_3_9_9
void FASTCALL call_11_3_9_9(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(9<<4)+9);
}
#define init_11_3_9_10
void FASTCALL call_11_3_9_10(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(9<<4)+10);
}
#define init_11_3_9_11
void FASTCALL call_11_3_9_11(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(9<<4)+11);
}
#define init_11_3_9_12
void FASTCALL call_11_3_9_12(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(9<<4)+12);
}
#define init_11_3_9_13
void FASTCALL call_11_3_9_13(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(9<<4)+13);
}
#define init_11_3_9_14
void FASTCALL call_11_3_9_14(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(9<<4)+14);
}
#define init_11_3_9_15
void FASTCALL call_11_3_9_15(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(9<<4)+15);
}
#define init_11_3_10_0
void FASTCALL call_11_3_10_0(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(10<<4)+0);
}
#define init_11_3_10_1
void FASTCALL call_11_3_10_1(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(10<<4)+1);
}
#define init_11_3_10_2
void FASTCALL call_11_3_10_2(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(10<<4)+2);
}
#define init_11_3_10_3
void FASTCALL call_11_3_10_3(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(10<<4)+3);
}
#define init_11_3_10_4
void FASTCALL call_11_3_10_4(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(10<<4)+4);
}
#define init_11_3_10_5
void FASTCALL call_11_3_10_5(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(10<<4)+5);
}
#define init_11_3_10_6
void FASTCALL call_11_3_10_6(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(10<<4)+6);
}
#define init_11_3_10_7
void FASTCALL call_11_3_10_7(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(10<<4)+7);
}
#define init_11_3_10_8
void FASTCALL call_11_3_10_8(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(10<<4)+8);
}
#define init_11_3_10_9
void FASTCALL call_11_3_10_9(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(10<<4)+9);
}
#define init_11_3_10_10
void FASTCALL call_11_3_10_10(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(10<<4)+10);
}
#define init_11_3_10_11
void FASTCALL call_11_3_10_11(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(10<<4)+11);
}
#define init_11_3_10_12
void FASTCALL call_11_3_10_12(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(10<<4)+12);
}
#define init_11_3_10_13
void FASTCALL call_11_3_10_13(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(10<<4)+13);
}
#define init_11_3_10_14
void FASTCALL call_11_3_10_14(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(10<<4)+14);
}
#define init_11_3_10_15
void FASTCALL call_11_3_10_15(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(10<<4)+15);
}
#define init_11_3_11_0
void FASTCALL call_11_3_11_0(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(11<<4)+0);
}
#define init_11_3_11_1
void FASTCALL call_11_3_11_1(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(11<<4)+1);
}
#define init_11_3_11_2
void FASTCALL call_11_3_11_2(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(11<<4)+2);
}
#define init_11_3_11_3
void FASTCALL call_11_3_11_3(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(11<<4)+3);
}
#define init_11_3_11_4
void FASTCALL call_11_3_11_4(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(11<<4)+4);
}
#define init_11_3_11_5
void FASTCALL call_11_3_11_5(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(11<<4)+5);
}
#define init_11_3_11_6
void FASTCALL call_11_3_11_6(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(11<<4)+6);
}
#define init_11_3_11_7
void FASTCALL call_11_3_11_7(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(11<<4)+7);
}
#define init_11_3_11_8
void FASTCALL call_11_3_11_8(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(11<<4)+8);
}
#define init_11_3_11_9
void FASTCALL call_11_3_11_9(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(11<<4)+9);
}
#define init_11_3_11_10
void FASTCALL call_11_3_11_10(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(11<<4)+10);
}
#define init_11_3_11_11
void FASTCALL call_11_3_11_11(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(11<<4)+11);
}
#define init_11_3_11_12
void FASTCALL call_11_3_11_12(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(11<<4)+12);
}
#define init_11_3_11_13
void FASTCALL call_11_3_11_13(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(11<<4)+13);
}
#define init_11_3_11_14
void FASTCALL call_11_3_11_14(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(11<<4)+14);
}
#define init_11_3_11_15
void FASTCALL call_11_3_11_15(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(11<<4)+15);
}
#define init_11_3_12_0
void FASTCALL call_11_3_12_0(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(12<<4)+0);
}
#define init_11_3_12_1
void FASTCALL call_11_3_12_1(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(12<<4)+1);
}
#define init_11_3_12_2
void FASTCALL call_11_3_12_2(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(12<<4)+2);
}
#define init_11_3_12_3
void FASTCALL call_11_3_12_3(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(12<<4)+3);
}
#define init_11_3_12_4
void FASTCALL call_11_3_12_4(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(12<<4)+4);
}
#define init_11_3_12_5
void FASTCALL call_11_3_12_5(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(12<<4)+5);
}
#define init_11_3_12_6
void FASTCALL call_11_3_12_6(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(12<<4)+6);
}
#define init_11_3_12_7
void FASTCALL call_11_3_12_7(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(12<<4)+7);
}
#define init_11_3_12_8
void FASTCALL call_11_3_12_8(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(12<<4)+8);
}
#define init_11_3_12_9
void FASTCALL call_11_3_12_9(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(12<<4)+9);
}
#define init_11_3_12_10
void FASTCALL call_11_3_12_10(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(12<<4)+10);
}
#define init_11_3_12_11
void FASTCALL call_11_3_12_11(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(12<<4)+11);
}
#define init_11_3_12_12
void FASTCALL call_11_3_12_12(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(12<<4)+12);
}
#define init_11_3_12_13
void FASTCALL call_11_3_12_13(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(12<<4)+13);
}
#define init_11_3_12_14
void FASTCALL call_11_3_12_14(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(12<<4)+14);
}
#define init_11_3_12_15
void FASTCALL call_11_3_12_15(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(12<<4)+15);
}
#define init_11_3_13_0
void FASTCALL call_11_3_13_0(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(13<<4)+0);
}
#define init_11_3_13_1
void FASTCALL call_11_3_13_1(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(13<<4)+1);
}
#define init_11_3_13_2
void FASTCALL call_11_3_13_2(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(13<<4)+2);
}
#define init_11_3_13_3
void FASTCALL call_11_3_13_3(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(13<<4)+3);
}
#define init_11_3_13_4
void FASTCALL call_11_3_13_4(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(13<<4)+4);
}
#define init_11_3_13_5
void FASTCALL call_11_3_13_5(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(13<<4)+5);
}
#define init_11_3_13_6
void FASTCALL call_11_3_13_6(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(13<<4)+6);
}
#define init_11_3_13_7
void FASTCALL call_11_3_13_7(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(13<<4)+7);
}
#define init_11_3_13_8
void FASTCALL call_11_3_13_8(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(13<<4)+8);
}
#define init_11_3_13_9
void FASTCALL call_11_3_13_9(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(13<<4)+9);
}
#define init_11_3_13_10
void FASTCALL call_11_3_13_10(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(13<<4)+10);
}
#define init_11_3_13_11
void FASTCALL call_11_3_13_11(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(13<<4)+11);
}
#define init_11_3_13_12
void FASTCALL call_11_3_13_12(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(13<<4)+12);
}
#define init_11_3_13_13
void FASTCALL call_11_3_13_13(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(13<<4)+13);
}
#define init_11_3_13_14
void FASTCALL call_11_3_13_14(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(13<<4)+14);
}
#define init_11_3_13_15
void FASTCALL call_11_3_13_15(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(13<<4)+15);
}
#define init_11_3_14_0
void FASTCALL call_11_3_14_0(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(14<<4)+0);
}
#define init_11_3_14_1
void FASTCALL call_11_3_14_1(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(14<<4)+1);
}
#define init_11_3_14_2
void FASTCALL call_11_3_14_2(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(14<<4)+2);
}
#define init_11_3_14_3
void FASTCALL call_11_3_14_3(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(14<<4)+3);
}
#define init_11_3_14_4
void FASTCALL call_11_3_14_4(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(14<<4)+4);
}
#define init_11_3_14_5
void FASTCALL call_11_3_14_5(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(14<<4)+5);
}
#define init_11_3_14_6
void FASTCALL call_11_3_14_6(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(14<<4)+6);
}
#define init_11_3_14_7
void FASTCALL call_11_3_14_7(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(14<<4)+7);
}
#define init_11_3_14_8
void FASTCALL call_11_3_14_8(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(14<<4)+8);
}
#define init_11_3_14_9
void FASTCALL call_11_3_14_9(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(14<<4)+9);
}
#define init_11_3_14_10
void FASTCALL call_11_3_14_10(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(14<<4)+10);
}
#define init_11_3_14_11
void FASTCALL call_11_3_14_11(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(14<<4)+11);
}
#define init_11_3_14_12
void FASTCALL call_11_3_14_12(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(14<<4)+12);
}
#define init_11_3_14_13
void FASTCALL call_11_3_14_13(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(14<<4)+13);
}
#define init_11_3_14_14
void FASTCALL call_11_3_14_14(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(14<<4)+14);
}
#define init_11_3_14_15
void FASTCALL call_11_3_14_15(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(14<<4)+15);
}
#define init_11_3_15_0
void FASTCALL call_11_3_15_0(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(15<<4)+0);
}
#define init_11_3_15_1
void FASTCALL call_11_3_15_1(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(15<<4)+1);
}
#define init_11_3_15_2
void FASTCALL call_11_3_15_2(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(15<<4)+2);
}
#define init_11_3_15_3
void FASTCALL call_11_3_15_3(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(15<<4)+3);
}
#define init_11_3_15_4
void FASTCALL call_11_3_15_4(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(15<<4)+4);
}
#define init_11_3_15_5
void FASTCALL call_11_3_15_5(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(15<<4)+5);
}
#define init_11_3_15_6
void FASTCALL call_11_3_15_6(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(15<<4)+6);
}
#define init_11_3_15_7
void FASTCALL call_11_3_15_7(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(15<<4)+7);
}
#define init_11_3_15_8
void FASTCALL call_11_3_15_8(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(15<<4)+8);
}
#define init_11_3_15_9
void FASTCALL call_11_3_15_9(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(15<<4)+9);
}
#define init_11_3_15_10
void FASTCALL call_11_3_15_10(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(15<<4)+10);
}
#define init_11_3_15_11
void FASTCALL call_11_3_15_11(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(15<<4)+11);
}
#define init_11_3_15_12
void FASTCALL call_11_3_15_12(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(15<<4)+12);
}
#define init_11_3_15_13
void FASTCALL call_11_3_15_13(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(15<<4)+13);
}
#define init_11_3_15_14
void FASTCALL call_11_3_15_14(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(15<<4)+14);
}
#define init_11_3_15_15
void FASTCALL call_11_3_15_15(SH2_struct* sh)
{
  SH2bsr(sh, (3<<8)+(15<<4)+15);
}
#define init_11_4_0_0
void FASTCALL call_11_4_0_0(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(0<<4)+0);
}
#define init_11_4_0_1
void FASTCALL call_11_4_0_1(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(0<<4)+1);
}
#define init_11_4_0_2
void FASTCALL call_11_4_0_2(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(0<<4)+2);
}
#define init_11_4_0_3
void FASTCALL call_11_4_0_3(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(0<<4)+3);
}
#define init_11_4_0_4
void FASTCALL call_11_4_0_4(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(0<<4)+4);
}
#define init_11_4_0_5
void FASTCALL call_11_4_0_5(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(0<<4)+5);
}
#define init_11_4_0_6
void FASTCALL call_11_4_0_6(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(0<<4)+6);
}
#define init_11_4_0_7
void FASTCALL call_11_4_0_7(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(0<<4)+7);
}
#define init_11_4_0_8
void FASTCALL call_11_4_0_8(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(0<<4)+8);
}
#define init_11_4_0_9
void FASTCALL call_11_4_0_9(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(0<<4)+9);
}
#define init_11_4_0_10
void FASTCALL call_11_4_0_10(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(0<<4)+10);
}
#define init_11_4_0_11
void FASTCALL call_11_4_0_11(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(0<<4)+11);
}
#define init_11_4_0_12
void FASTCALL call_11_4_0_12(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(0<<4)+12);
}
#define init_11_4_0_13
void FASTCALL call_11_4_0_13(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(0<<4)+13);
}
#define init_11_4_0_14
void FASTCALL call_11_4_0_14(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(0<<4)+14);
}
#define init_11_4_0_15
void FASTCALL call_11_4_0_15(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(0<<4)+15);
}
#define init_11_4_1_0
void FASTCALL call_11_4_1_0(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(1<<4)+0);
}
#define init_11_4_1_1
void FASTCALL call_11_4_1_1(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(1<<4)+1);
}
#define init_11_4_1_2
void FASTCALL call_11_4_1_2(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(1<<4)+2);
}
#define init_11_4_1_3
void FASTCALL call_11_4_1_3(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(1<<4)+3);
}
#define init_11_4_1_4
void FASTCALL call_11_4_1_4(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(1<<4)+4);
}
#define init_11_4_1_5
void FASTCALL call_11_4_1_5(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(1<<4)+5);
}
#define init_11_4_1_6
void FASTCALL call_11_4_1_6(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(1<<4)+6);
}
#define init_11_4_1_7
void FASTCALL call_11_4_1_7(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(1<<4)+7);
}
#define init_11_4_1_8
void FASTCALL call_11_4_1_8(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(1<<4)+8);
}
#define init_11_4_1_9
void FASTCALL call_11_4_1_9(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(1<<4)+9);
}
#define init_11_4_1_10
void FASTCALL call_11_4_1_10(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(1<<4)+10);
}
#define init_11_4_1_11
void FASTCALL call_11_4_1_11(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(1<<4)+11);
}
#define init_11_4_1_12
void FASTCALL call_11_4_1_12(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(1<<4)+12);
}
#define init_11_4_1_13
void FASTCALL call_11_4_1_13(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(1<<4)+13);
}
#define init_11_4_1_14
void FASTCALL call_11_4_1_14(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(1<<4)+14);
}
#define init_11_4_1_15
void FASTCALL call_11_4_1_15(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(1<<4)+15);
}
#define init_11_4_2_0
void FASTCALL call_11_4_2_0(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(2<<4)+0);
}
#define init_11_4_2_1
void FASTCALL call_11_4_2_1(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(2<<4)+1);
}
#define init_11_4_2_2
void FASTCALL call_11_4_2_2(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(2<<4)+2);
}
#define init_11_4_2_3
void FASTCALL call_11_4_2_3(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(2<<4)+3);
}
#define init_11_4_2_4
void FASTCALL call_11_4_2_4(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(2<<4)+4);
}
#define init_11_4_2_5
void FASTCALL call_11_4_2_5(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(2<<4)+5);
}
#define init_11_4_2_6
void FASTCALL call_11_4_2_6(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(2<<4)+6);
}
#define init_11_4_2_7
void FASTCALL call_11_4_2_7(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(2<<4)+7);
}
#define init_11_4_2_8
void FASTCALL call_11_4_2_8(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(2<<4)+8);
}
#define init_11_4_2_9
void FASTCALL call_11_4_2_9(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(2<<4)+9);
}
#define init_11_4_2_10
void FASTCALL call_11_4_2_10(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(2<<4)+10);
}
#define init_11_4_2_11
void FASTCALL call_11_4_2_11(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(2<<4)+11);
}
#define init_11_4_2_12
void FASTCALL call_11_4_2_12(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(2<<4)+12);
}
#define init_11_4_2_13
void FASTCALL call_11_4_2_13(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(2<<4)+13);
}
#define init_11_4_2_14
void FASTCALL call_11_4_2_14(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(2<<4)+14);
}
#define init_11_4_2_15
void FASTCALL call_11_4_2_15(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(2<<4)+15);
}
#define init_11_4_3_0
void FASTCALL call_11_4_3_0(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(3<<4)+0);
}
#define init_11_4_3_1
void FASTCALL call_11_4_3_1(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(3<<4)+1);
}
#define init_11_4_3_2
void FASTCALL call_11_4_3_2(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(3<<4)+2);
}
#define init_11_4_3_3
void FASTCALL call_11_4_3_3(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(3<<4)+3);
}
#define init_11_4_3_4
void FASTCALL call_11_4_3_4(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(3<<4)+4);
}
#define init_11_4_3_5
void FASTCALL call_11_4_3_5(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(3<<4)+5);
}
#define init_11_4_3_6
void FASTCALL call_11_4_3_6(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(3<<4)+6);
}
#define init_11_4_3_7
void FASTCALL call_11_4_3_7(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(3<<4)+7);
}
#define init_11_4_3_8
void FASTCALL call_11_4_3_8(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(3<<4)+8);
}
#define init_11_4_3_9
void FASTCALL call_11_4_3_9(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(3<<4)+9);
}
#define init_11_4_3_10
void FASTCALL call_11_4_3_10(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(3<<4)+10);
}
#define init_11_4_3_11
void FASTCALL call_11_4_3_11(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(3<<4)+11);
}
#define init_11_4_3_12
void FASTCALL call_11_4_3_12(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(3<<4)+12);
}
#define init_11_4_3_13
void FASTCALL call_11_4_3_13(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(3<<4)+13);
}
#define init_11_4_3_14
void FASTCALL call_11_4_3_14(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(3<<4)+14);
}
#define init_11_4_3_15
void FASTCALL call_11_4_3_15(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(3<<4)+15);
}
#define init_11_4_4_0
void FASTCALL call_11_4_4_0(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(4<<4)+0);
}
#define init_11_4_4_1
void FASTCALL call_11_4_4_1(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(4<<4)+1);
}
#define init_11_4_4_2
void FASTCALL call_11_4_4_2(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(4<<4)+2);
}
#define init_11_4_4_3
void FASTCALL call_11_4_4_3(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(4<<4)+3);
}
#define init_11_4_4_4
void FASTCALL call_11_4_4_4(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(4<<4)+4);
}
#define init_11_4_4_5
void FASTCALL call_11_4_4_5(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(4<<4)+5);
}
#define init_11_4_4_6
void FASTCALL call_11_4_4_6(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(4<<4)+6);
}
#define init_11_4_4_7
void FASTCALL call_11_4_4_7(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(4<<4)+7);
}
#define init_11_4_4_8
void FASTCALL call_11_4_4_8(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(4<<4)+8);
}
#define init_11_4_4_9
void FASTCALL call_11_4_4_9(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(4<<4)+9);
}
#define init_11_4_4_10
void FASTCALL call_11_4_4_10(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(4<<4)+10);
}
#define init_11_4_4_11
void FASTCALL call_11_4_4_11(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(4<<4)+11);
}
#define init_11_4_4_12
void FASTCALL call_11_4_4_12(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(4<<4)+12);
}
#define init_11_4_4_13
void FASTCALL call_11_4_4_13(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(4<<4)+13);
}
#define init_11_4_4_14
void FASTCALL call_11_4_4_14(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(4<<4)+14);
}
#define init_11_4_4_15
void FASTCALL call_11_4_4_15(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(4<<4)+15);
}
#define init_11_4_5_0
void FASTCALL call_11_4_5_0(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(5<<4)+0);
}
#define init_11_4_5_1
void FASTCALL call_11_4_5_1(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(5<<4)+1);
}
#define init_11_4_5_2
void FASTCALL call_11_4_5_2(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(5<<4)+2);
}
#define init_11_4_5_3
void FASTCALL call_11_4_5_3(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(5<<4)+3);
}
#define init_11_4_5_4
void FASTCALL call_11_4_5_4(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(5<<4)+4);
}
#define init_11_4_5_5
void FASTCALL call_11_4_5_5(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(5<<4)+5);
}
#define init_11_4_5_6
void FASTCALL call_11_4_5_6(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(5<<4)+6);
}
#define init_11_4_5_7
void FASTCALL call_11_4_5_7(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(5<<4)+7);
}
#define init_11_4_5_8
void FASTCALL call_11_4_5_8(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(5<<4)+8);
}
#define init_11_4_5_9
void FASTCALL call_11_4_5_9(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(5<<4)+9);
}
#define init_11_4_5_10
void FASTCALL call_11_4_5_10(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(5<<4)+10);
}
#define init_11_4_5_11
void FASTCALL call_11_4_5_11(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(5<<4)+11);
}
#define init_11_4_5_12
void FASTCALL call_11_4_5_12(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(5<<4)+12);
}
#define init_11_4_5_13
void FASTCALL call_11_4_5_13(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(5<<4)+13);
}
#define init_11_4_5_14
void FASTCALL call_11_4_5_14(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(5<<4)+14);
}
#define init_11_4_5_15
void FASTCALL call_11_4_5_15(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(5<<4)+15);
}
#define init_11_4_6_0
void FASTCALL call_11_4_6_0(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(6<<4)+0);
}
#define init_11_4_6_1
void FASTCALL call_11_4_6_1(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(6<<4)+1);
}
#define init_11_4_6_2
void FASTCALL call_11_4_6_2(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(6<<4)+2);
}
#define init_11_4_6_3
void FASTCALL call_11_4_6_3(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(6<<4)+3);
}
#define init_11_4_6_4
void FASTCALL call_11_4_6_4(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(6<<4)+4);
}
#define init_11_4_6_5
void FASTCALL call_11_4_6_5(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(6<<4)+5);
}
#define init_11_4_6_6
void FASTCALL call_11_4_6_6(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(6<<4)+6);
}
#define init_11_4_6_7
void FASTCALL call_11_4_6_7(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(6<<4)+7);
}
#define init_11_4_6_8
void FASTCALL call_11_4_6_8(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(6<<4)+8);
}
#define init_11_4_6_9
void FASTCALL call_11_4_6_9(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(6<<4)+9);
}
#define init_11_4_6_10
void FASTCALL call_11_4_6_10(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(6<<4)+10);
}
#define init_11_4_6_11
void FASTCALL call_11_4_6_11(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(6<<4)+11);
}
#define init_11_4_6_12
void FASTCALL call_11_4_6_12(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(6<<4)+12);
}
#define init_11_4_6_13
void FASTCALL call_11_4_6_13(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(6<<4)+13);
}
#define init_11_4_6_14
void FASTCALL call_11_4_6_14(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(6<<4)+14);
}
#define init_11_4_6_15
void FASTCALL call_11_4_6_15(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(6<<4)+15);
}
#define init_11_4_7_0
void FASTCALL call_11_4_7_0(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(7<<4)+0);
}
#define init_11_4_7_1
void FASTCALL call_11_4_7_1(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(7<<4)+1);
}
#define init_11_4_7_2
void FASTCALL call_11_4_7_2(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(7<<4)+2);
}
#define init_11_4_7_3
void FASTCALL call_11_4_7_3(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(7<<4)+3);
}
#define init_11_4_7_4
void FASTCALL call_11_4_7_4(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(7<<4)+4);
}
#define init_11_4_7_5
void FASTCALL call_11_4_7_5(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(7<<4)+5);
}
#define init_11_4_7_6
void FASTCALL call_11_4_7_6(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(7<<4)+6);
}
#define init_11_4_7_7
void FASTCALL call_11_4_7_7(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(7<<4)+7);
}
#define init_11_4_7_8
void FASTCALL call_11_4_7_8(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(7<<4)+8);
}
#define init_11_4_7_9
void FASTCALL call_11_4_7_9(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(7<<4)+9);
}
#define init_11_4_7_10
void FASTCALL call_11_4_7_10(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(7<<4)+10);
}
#define init_11_4_7_11
void FASTCALL call_11_4_7_11(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(7<<4)+11);
}
#define init_11_4_7_12
void FASTCALL call_11_4_7_12(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(7<<4)+12);
}
#define init_11_4_7_13
void FASTCALL call_11_4_7_13(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(7<<4)+13);
}
#define init_11_4_7_14
void FASTCALL call_11_4_7_14(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(7<<4)+14);
}
#define init_11_4_7_15
void FASTCALL call_11_4_7_15(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(7<<4)+15);
}
#define init_11_4_8_0
void FASTCALL call_11_4_8_0(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(8<<4)+0);
}
#define init_11_4_8_1
void FASTCALL call_11_4_8_1(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(8<<4)+1);
}
#define init_11_4_8_2
void FASTCALL call_11_4_8_2(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(8<<4)+2);
}
#define init_11_4_8_3
void FASTCALL call_11_4_8_3(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(8<<4)+3);
}
#define init_11_4_8_4
void FASTCALL call_11_4_8_4(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(8<<4)+4);
}
#define init_11_4_8_5
void FASTCALL call_11_4_8_5(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(8<<4)+5);
}
#define init_11_4_8_6
void FASTCALL call_11_4_8_6(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(8<<4)+6);
}
#define init_11_4_8_7
void FASTCALL call_11_4_8_7(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(8<<4)+7);
}
#define init_11_4_8_8
void FASTCALL call_11_4_8_8(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(8<<4)+8);
}
#define init_11_4_8_9
void FASTCALL call_11_4_8_9(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(8<<4)+9);
}
#define init_11_4_8_10
void FASTCALL call_11_4_8_10(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(8<<4)+10);
}
#define init_11_4_8_11
void FASTCALL call_11_4_8_11(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(8<<4)+11);
}
#define init_11_4_8_12
void FASTCALL call_11_4_8_12(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(8<<4)+12);
}
#define init_11_4_8_13
void FASTCALL call_11_4_8_13(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(8<<4)+13);
}
#define init_11_4_8_14
void FASTCALL call_11_4_8_14(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(8<<4)+14);
}
#define init_11_4_8_15
void FASTCALL call_11_4_8_15(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(8<<4)+15);
}
#define init_11_4_9_0
void FASTCALL call_11_4_9_0(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(9<<4)+0);
}
#define init_11_4_9_1
void FASTCALL call_11_4_9_1(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(9<<4)+1);
}
#define init_11_4_9_2
void FASTCALL call_11_4_9_2(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(9<<4)+2);
}
#define init_11_4_9_3
void FASTCALL call_11_4_9_3(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(9<<4)+3);
}
#define init_11_4_9_4
void FASTCALL call_11_4_9_4(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(9<<4)+4);
}
#define init_11_4_9_5
void FASTCALL call_11_4_9_5(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(9<<4)+5);
}
#define init_11_4_9_6
void FASTCALL call_11_4_9_6(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(9<<4)+6);
}
#define init_11_4_9_7
void FASTCALL call_11_4_9_7(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(9<<4)+7);
}
#define init_11_4_9_8
void FASTCALL call_11_4_9_8(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(9<<4)+8);
}
#define init_11_4_9_9
void FASTCALL call_11_4_9_9(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(9<<4)+9);
}
#define init_11_4_9_10
void FASTCALL call_11_4_9_10(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(9<<4)+10);
}
#define init_11_4_9_11
void FASTCALL call_11_4_9_11(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(9<<4)+11);
}
#define init_11_4_9_12
void FASTCALL call_11_4_9_12(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(9<<4)+12);
}
#define init_11_4_9_13
void FASTCALL call_11_4_9_13(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(9<<4)+13);
}
#define init_11_4_9_14
void FASTCALL call_11_4_9_14(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(9<<4)+14);
}
#define init_11_4_9_15
void FASTCALL call_11_4_9_15(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(9<<4)+15);
}
#define init_11_4_10_0
void FASTCALL call_11_4_10_0(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(10<<4)+0);
}
#define init_11_4_10_1
void FASTCALL call_11_4_10_1(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(10<<4)+1);
}
#define init_11_4_10_2
void FASTCALL call_11_4_10_2(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(10<<4)+2);
}
#define init_11_4_10_3
void FASTCALL call_11_4_10_3(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(10<<4)+3);
}
#define init_11_4_10_4
void FASTCALL call_11_4_10_4(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(10<<4)+4);
}
#define init_11_4_10_5
void FASTCALL call_11_4_10_5(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(10<<4)+5);
}
#define init_11_4_10_6
void FASTCALL call_11_4_10_6(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(10<<4)+6);
}
#define init_11_4_10_7
void FASTCALL call_11_4_10_7(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(10<<4)+7);
}
#define init_11_4_10_8
void FASTCALL call_11_4_10_8(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(10<<4)+8);
}
#define init_11_4_10_9
void FASTCALL call_11_4_10_9(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(10<<4)+9);
}
#define init_11_4_10_10
void FASTCALL call_11_4_10_10(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(10<<4)+10);
}
#define init_11_4_10_11
void FASTCALL call_11_4_10_11(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(10<<4)+11);
}
#define init_11_4_10_12
void FASTCALL call_11_4_10_12(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(10<<4)+12);
}
#define init_11_4_10_13
void FASTCALL call_11_4_10_13(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(10<<4)+13);
}
#define init_11_4_10_14
void FASTCALL call_11_4_10_14(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(10<<4)+14);
}
#define init_11_4_10_15
void FASTCALL call_11_4_10_15(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(10<<4)+15);
}
#define init_11_4_11_0
void FASTCALL call_11_4_11_0(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(11<<4)+0);
}
#define init_11_4_11_1
void FASTCALL call_11_4_11_1(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(11<<4)+1);
}
#define init_11_4_11_2
void FASTCALL call_11_4_11_2(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(11<<4)+2);
}
#define init_11_4_11_3
void FASTCALL call_11_4_11_3(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(11<<4)+3);
}
#define init_11_4_11_4
void FASTCALL call_11_4_11_4(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(11<<4)+4);
}
#define init_11_4_11_5
void FASTCALL call_11_4_11_5(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(11<<4)+5);
}
#define init_11_4_11_6
void FASTCALL call_11_4_11_6(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(11<<4)+6);
}
#define init_11_4_11_7
void FASTCALL call_11_4_11_7(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(11<<4)+7);
}
#define init_11_4_11_8
void FASTCALL call_11_4_11_8(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(11<<4)+8);
}
#define init_11_4_11_9
void FASTCALL call_11_4_11_9(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(11<<4)+9);
}
#define init_11_4_11_10
void FASTCALL call_11_4_11_10(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(11<<4)+10);
}
#define init_11_4_11_11
void FASTCALL call_11_4_11_11(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(11<<4)+11);
}
#define init_11_4_11_12
void FASTCALL call_11_4_11_12(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(11<<4)+12);
}
#define init_11_4_11_13
void FASTCALL call_11_4_11_13(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(11<<4)+13);
}
#define init_11_4_11_14
void FASTCALL call_11_4_11_14(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(11<<4)+14);
}
#define init_11_4_11_15
void FASTCALL call_11_4_11_15(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(11<<4)+15);
}
#define init_11_4_12_0
void FASTCALL call_11_4_12_0(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(12<<4)+0);
}
#define init_11_4_12_1
void FASTCALL call_11_4_12_1(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(12<<4)+1);
}
#define init_11_4_12_2
void FASTCALL call_11_4_12_2(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(12<<4)+2);
}
#define init_11_4_12_3
void FASTCALL call_11_4_12_3(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(12<<4)+3);
}
#define init_11_4_12_4
void FASTCALL call_11_4_12_4(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(12<<4)+4);
}
#define init_11_4_12_5
void FASTCALL call_11_4_12_5(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(12<<4)+5);
}
#define init_11_4_12_6
void FASTCALL call_11_4_12_6(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(12<<4)+6);
}
#define init_11_4_12_7
void FASTCALL call_11_4_12_7(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(12<<4)+7);
}
#define init_11_4_12_8
void FASTCALL call_11_4_12_8(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(12<<4)+8);
}
#define init_11_4_12_9
void FASTCALL call_11_4_12_9(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(12<<4)+9);
}
#define init_11_4_12_10
void FASTCALL call_11_4_12_10(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(12<<4)+10);
}
#define init_11_4_12_11
void FASTCALL call_11_4_12_11(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(12<<4)+11);
}
#define init_11_4_12_12
void FASTCALL call_11_4_12_12(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(12<<4)+12);
}
#define init_11_4_12_13
void FASTCALL call_11_4_12_13(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(12<<4)+13);
}
#define init_11_4_12_14
void FASTCALL call_11_4_12_14(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(12<<4)+14);
}
#define init_11_4_12_15
void FASTCALL call_11_4_12_15(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(12<<4)+15);
}
#define init_11_4_13_0
void FASTCALL call_11_4_13_0(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(13<<4)+0);
}
#define init_11_4_13_1
void FASTCALL call_11_4_13_1(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(13<<4)+1);
}
#define init_11_4_13_2
void FASTCALL call_11_4_13_2(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(13<<4)+2);
}
#define init_11_4_13_3
void FASTCALL call_11_4_13_3(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(13<<4)+3);
}
#define init_11_4_13_4
void FASTCALL call_11_4_13_4(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(13<<4)+4);
}
#define init_11_4_13_5
void FASTCALL call_11_4_13_5(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(13<<4)+5);
}
#define init_11_4_13_6
void FASTCALL call_11_4_13_6(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(13<<4)+6);
}
#define init_11_4_13_7
void FASTCALL call_11_4_13_7(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(13<<4)+7);
}
#define init_11_4_13_8
void FASTCALL call_11_4_13_8(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(13<<4)+8);
}
#define init_11_4_13_9
void FASTCALL call_11_4_13_9(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(13<<4)+9);
}
#define init_11_4_13_10
void FASTCALL call_11_4_13_10(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(13<<4)+10);
}
#define init_11_4_13_11
void FASTCALL call_11_4_13_11(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(13<<4)+11);
}
#define init_11_4_13_12
void FASTCALL call_11_4_13_12(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(13<<4)+12);
}
#define init_11_4_13_13
void FASTCALL call_11_4_13_13(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(13<<4)+13);
}
#define init_11_4_13_14
void FASTCALL call_11_4_13_14(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(13<<4)+14);
}
#define init_11_4_13_15
void FASTCALL call_11_4_13_15(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(13<<4)+15);
}
#define init_11_4_14_0
void FASTCALL call_11_4_14_0(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(14<<4)+0);
}
#define init_11_4_14_1
void FASTCALL call_11_4_14_1(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(14<<4)+1);
}
#define init_11_4_14_2
void FASTCALL call_11_4_14_2(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(14<<4)+2);
}
#define init_11_4_14_3
void FASTCALL call_11_4_14_3(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(14<<4)+3);
}
#define init_11_4_14_4
void FASTCALL call_11_4_14_4(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(14<<4)+4);
}
#define init_11_4_14_5
void FASTCALL call_11_4_14_5(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(14<<4)+5);
}
#define init_11_4_14_6
void FASTCALL call_11_4_14_6(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(14<<4)+6);
}
#define init_11_4_14_7
void FASTCALL call_11_4_14_7(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(14<<4)+7);
}
#define init_11_4_14_8
void FASTCALL call_11_4_14_8(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(14<<4)+8);
}
#define init_11_4_14_9
void FASTCALL call_11_4_14_9(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(14<<4)+9);
}
#define init_11_4_14_10
void FASTCALL call_11_4_14_10(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(14<<4)+10);
}
#define init_11_4_14_11
void FASTCALL call_11_4_14_11(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(14<<4)+11);
}
#define init_11_4_14_12
void FASTCALL call_11_4_14_12(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(14<<4)+12);
}
#define init_11_4_14_13
void FASTCALL call_11_4_14_13(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(14<<4)+13);
}
#define init_11_4_14_14
void FASTCALL call_11_4_14_14(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(14<<4)+14);
}
#define init_11_4_14_15
void FASTCALL call_11_4_14_15(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(14<<4)+15);
}
#define init_11_4_15_0
void FASTCALL call_11_4_15_0(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(15<<4)+0);
}
#define init_11_4_15_1
void FASTCALL call_11_4_15_1(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(15<<4)+1);
}
#define init_11_4_15_2
void FASTCALL call_11_4_15_2(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(15<<4)+2);
}
#define init_11_4_15_3
void FASTCALL call_11_4_15_3(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(15<<4)+3);
}
#define init_11_4_15_4
void FASTCALL call_11_4_15_4(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(15<<4)+4);
}
#define init_11_4_15_5
void FASTCALL call_11_4_15_5(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(15<<4)+5);
}
#define init_11_4_15_6
void FASTCALL call_11_4_15_6(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(15<<4)+6);
}
#define init_11_4_15_7
void FASTCALL call_11_4_15_7(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(15<<4)+7);
}
#define init_11_4_15_8
void FASTCALL call_11_4_15_8(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(15<<4)+8);
}
#define init_11_4_15_9
void FASTCALL call_11_4_15_9(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(15<<4)+9);
}
#define init_11_4_15_10
void FASTCALL call_11_4_15_10(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(15<<4)+10);
}
#define init_11_4_15_11
void FASTCALL call_11_4_15_11(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(15<<4)+11);
}
#define init_11_4_15_12
void FASTCALL call_11_4_15_12(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(15<<4)+12);
}
#define init_11_4_15_13
void FASTCALL call_11_4_15_13(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(15<<4)+13);
}
#define init_11_4_15_14
void FASTCALL call_11_4_15_14(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(15<<4)+14);
}
#define init_11_4_15_15
void FASTCALL call_11_4_15_15(SH2_struct* sh)
{
  SH2bsr(sh, (4<<8)+(15<<4)+15);
}
#define init_11_5_0_0
void FASTCALL call_11_5_0_0(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(0<<4)+0);
}
#define init_11_5_0_1
void FASTCALL call_11_5_0_1(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(0<<4)+1);
}
#define init_11_5_0_2
void FASTCALL call_11_5_0_2(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(0<<4)+2);
}
#define init_11_5_0_3
void FASTCALL call_11_5_0_3(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(0<<4)+3);
}
#define init_11_5_0_4
void FASTCALL call_11_5_0_4(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(0<<4)+4);
}
#define init_11_5_0_5
void FASTCALL call_11_5_0_5(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(0<<4)+5);
}
#define init_11_5_0_6
void FASTCALL call_11_5_0_6(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(0<<4)+6);
}
#define init_11_5_0_7
void FASTCALL call_11_5_0_7(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(0<<4)+7);
}
#define init_11_5_0_8
void FASTCALL call_11_5_0_8(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(0<<4)+8);
}
#define init_11_5_0_9
void FASTCALL call_11_5_0_9(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(0<<4)+9);
}
#define init_11_5_0_10
void FASTCALL call_11_5_0_10(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(0<<4)+10);
}
#define init_11_5_0_11
void FASTCALL call_11_5_0_11(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(0<<4)+11);
}
#define init_11_5_0_12
void FASTCALL call_11_5_0_12(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(0<<4)+12);
}
#define init_11_5_0_13
void FASTCALL call_11_5_0_13(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(0<<4)+13);
}
#define init_11_5_0_14
void FASTCALL call_11_5_0_14(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(0<<4)+14);
}
#define init_11_5_0_15
void FASTCALL call_11_5_0_15(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(0<<4)+15);
}
#define init_11_5_1_0
void FASTCALL call_11_5_1_0(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(1<<4)+0);
}
#define init_11_5_1_1
void FASTCALL call_11_5_1_1(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(1<<4)+1);
}
#define init_11_5_1_2
void FASTCALL call_11_5_1_2(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(1<<4)+2);
}
#define init_11_5_1_3
void FASTCALL call_11_5_1_3(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(1<<4)+3);
}
#define init_11_5_1_4
void FASTCALL call_11_5_1_4(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(1<<4)+4);
}
#define init_11_5_1_5
void FASTCALL call_11_5_1_5(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(1<<4)+5);
}
#define init_11_5_1_6
void FASTCALL call_11_5_1_6(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(1<<4)+6);
}
#define init_11_5_1_7
void FASTCALL call_11_5_1_7(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(1<<4)+7);
}
#define init_11_5_1_8
void FASTCALL call_11_5_1_8(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(1<<4)+8);
}
#define init_11_5_1_9
void FASTCALL call_11_5_1_9(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(1<<4)+9);
}
#define init_11_5_1_10
void FASTCALL call_11_5_1_10(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(1<<4)+10);
}
#define init_11_5_1_11
void FASTCALL call_11_5_1_11(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(1<<4)+11);
}
#define init_11_5_1_12
void FASTCALL call_11_5_1_12(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(1<<4)+12);
}
#define init_11_5_1_13
void FASTCALL call_11_5_1_13(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(1<<4)+13);
}
#define init_11_5_1_14
void FASTCALL call_11_5_1_14(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(1<<4)+14);
}
#define init_11_5_1_15
void FASTCALL call_11_5_1_15(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(1<<4)+15);
}
#define init_11_5_2_0
void FASTCALL call_11_5_2_0(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(2<<4)+0);
}
#define init_11_5_2_1
void FASTCALL call_11_5_2_1(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(2<<4)+1);
}
#define init_11_5_2_2
void FASTCALL call_11_5_2_2(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(2<<4)+2);
}
#define init_11_5_2_3
void FASTCALL call_11_5_2_3(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(2<<4)+3);
}
#define init_11_5_2_4
void FASTCALL call_11_5_2_4(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(2<<4)+4);
}
#define init_11_5_2_5
void FASTCALL call_11_5_2_5(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(2<<4)+5);
}
#define init_11_5_2_6
void FASTCALL call_11_5_2_6(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(2<<4)+6);
}
#define init_11_5_2_7
void FASTCALL call_11_5_2_7(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(2<<4)+7);
}
#define init_11_5_2_8
void FASTCALL call_11_5_2_8(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(2<<4)+8);
}
#define init_11_5_2_9
void FASTCALL call_11_5_2_9(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(2<<4)+9);
}
#define init_11_5_2_10
void FASTCALL call_11_5_2_10(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(2<<4)+10);
}
#define init_11_5_2_11
void FASTCALL call_11_5_2_11(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(2<<4)+11);
}
#define init_11_5_2_12
void FASTCALL call_11_5_2_12(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(2<<4)+12);
}
#define init_11_5_2_13
void FASTCALL call_11_5_2_13(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(2<<4)+13);
}
#define init_11_5_2_14
void FASTCALL call_11_5_2_14(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(2<<4)+14);
}
#define init_11_5_2_15
void FASTCALL call_11_5_2_15(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(2<<4)+15);
}
#define init_11_5_3_0
void FASTCALL call_11_5_3_0(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(3<<4)+0);
}
#define init_11_5_3_1
void FASTCALL call_11_5_3_1(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(3<<4)+1);
}
#define init_11_5_3_2
void FASTCALL call_11_5_3_2(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(3<<4)+2);
}
#define init_11_5_3_3
void FASTCALL call_11_5_3_3(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(3<<4)+3);
}
#define init_11_5_3_4
void FASTCALL call_11_5_3_4(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(3<<4)+4);
}
#define init_11_5_3_5
void FASTCALL call_11_5_3_5(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(3<<4)+5);
}
#define init_11_5_3_6
void FASTCALL call_11_5_3_6(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(3<<4)+6);
}
#define init_11_5_3_7
void FASTCALL call_11_5_3_7(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(3<<4)+7);
}
#define init_11_5_3_8
void FASTCALL call_11_5_3_8(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(3<<4)+8);
}
#define init_11_5_3_9
void FASTCALL call_11_5_3_9(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(3<<4)+9);
}
#define init_11_5_3_10
void FASTCALL call_11_5_3_10(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(3<<4)+10);
}
#define init_11_5_3_11
void FASTCALL call_11_5_3_11(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(3<<4)+11);
}
#define init_11_5_3_12
void FASTCALL call_11_5_3_12(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(3<<4)+12);
}
#define init_11_5_3_13
void FASTCALL call_11_5_3_13(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(3<<4)+13);
}
#define init_11_5_3_14
void FASTCALL call_11_5_3_14(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(3<<4)+14);
}
#define init_11_5_3_15
void FASTCALL call_11_5_3_15(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(3<<4)+15);
}
#define init_11_5_4_0
void FASTCALL call_11_5_4_0(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(4<<4)+0);
}
#define init_11_5_4_1
void FASTCALL call_11_5_4_1(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(4<<4)+1);
}
#define init_11_5_4_2
void FASTCALL call_11_5_4_2(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(4<<4)+2);
}
#define init_11_5_4_3
void FASTCALL call_11_5_4_3(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(4<<4)+3);
}
#define init_11_5_4_4
void FASTCALL call_11_5_4_4(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(4<<4)+4);
}
#define init_11_5_4_5
void FASTCALL call_11_5_4_5(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(4<<4)+5);
}
#define init_11_5_4_6
void FASTCALL call_11_5_4_6(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(4<<4)+6);
}
#define init_11_5_4_7
void FASTCALL call_11_5_4_7(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(4<<4)+7);
}
#define init_11_5_4_8
void FASTCALL call_11_5_4_8(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(4<<4)+8);
}
#define init_11_5_4_9
void FASTCALL call_11_5_4_9(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(4<<4)+9);
}
#define init_11_5_4_10
void FASTCALL call_11_5_4_10(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(4<<4)+10);
}
#define init_11_5_4_11
void FASTCALL call_11_5_4_11(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(4<<4)+11);
}
#define init_11_5_4_12
void FASTCALL call_11_5_4_12(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(4<<4)+12);
}
#define init_11_5_4_13
void FASTCALL call_11_5_4_13(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(4<<4)+13);
}
#define init_11_5_4_14
void FASTCALL call_11_5_4_14(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(4<<4)+14);
}
#define init_11_5_4_15
void FASTCALL call_11_5_4_15(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(4<<4)+15);
}
#define init_11_5_5_0
void FASTCALL call_11_5_5_0(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(5<<4)+0);
}
#define init_11_5_5_1
void FASTCALL call_11_5_5_1(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(5<<4)+1);
}
#define init_11_5_5_2
void FASTCALL call_11_5_5_2(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(5<<4)+2);
}
#define init_11_5_5_3
void FASTCALL call_11_5_5_3(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(5<<4)+3);
}
#define init_11_5_5_4
void FASTCALL call_11_5_5_4(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(5<<4)+4);
}
#define init_11_5_5_5
void FASTCALL call_11_5_5_5(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(5<<4)+5);
}
#define init_11_5_5_6
void FASTCALL call_11_5_5_6(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(5<<4)+6);
}
#define init_11_5_5_7
void FASTCALL call_11_5_5_7(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(5<<4)+7);
}
#define init_11_5_5_8
void FASTCALL call_11_5_5_8(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(5<<4)+8);
}
#define init_11_5_5_9
void FASTCALL call_11_5_5_9(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(5<<4)+9);
}
#define init_11_5_5_10
void FASTCALL call_11_5_5_10(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(5<<4)+10);
}
#define init_11_5_5_11
void FASTCALL call_11_5_5_11(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(5<<4)+11);
}
#define init_11_5_5_12
void FASTCALL call_11_5_5_12(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(5<<4)+12);
}
#define init_11_5_5_13
void FASTCALL call_11_5_5_13(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(5<<4)+13);
}
#define init_11_5_5_14
void FASTCALL call_11_5_5_14(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(5<<4)+14);
}
#define init_11_5_5_15
void FASTCALL call_11_5_5_15(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(5<<4)+15);
}
#define init_11_5_6_0
void FASTCALL call_11_5_6_0(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(6<<4)+0);
}
#define init_11_5_6_1
void FASTCALL call_11_5_6_1(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(6<<4)+1);
}
#define init_11_5_6_2
void FASTCALL call_11_5_6_2(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(6<<4)+2);
}
#define init_11_5_6_3
void FASTCALL call_11_5_6_3(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(6<<4)+3);
}
#define init_11_5_6_4
void FASTCALL call_11_5_6_4(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(6<<4)+4);
}
#define init_11_5_6_5
void FASTCALL call_11_5_6_5(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(6<<4)+5);
}
#define init_11_5_6_6
void FASTCALL call_11_5_6_6(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(6<<4)+6);
}
#define init_11_5_6_7
void FASTCALL call_11_5_6_7(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(6<<4)+7);
}
#define init_11_5_6_8
void FASTCALL call_11_5_6_8(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(6<<4)+8);
}
#define init_11_5_6_9
void FASTCALL call_11_5_6_9(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(6<<4)+9);
}
#define init_11_5_6_10
void FASTCALL call_11_5_6_10(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(6<<4)+10);
}
#define init_11_5_6_11
void FASTCALL call_11_5_6_11(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(6<<4)+11);
}
#define init_11_5_6_12
void FASTCALL call_11_5_6_12(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(6<<4)+12);
}
#define init_11_5_6_13
void FASTCALL call_11_5_6_13(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(6<<4)+13);
}
#define init_11_5_6_14
void FASTCALL call_11_5_6_14(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(6<<4)+14);
}
#define init_11_5_6_15
void FASTCALL call_11_5_6_15(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(6<<4)+15);
}
#define init_11_5_7_0
void FASTCALL call_11_5_7_0(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(7<<4)+0);
}
#define init_11_5_7_1
void FASTCALL call_11_5_7_1(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(7<<4)+1);
}
#define init_11_5_7_2
void FASTCALL call_11_5_7_2(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(7<<4)+2);
}
#define init_11_5_7_3
void FASTCALL call_11_5_7_3(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(7<<4)+3);
}
#define init_11_5_7_4
void FASTCALL call_11_5_7_4(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(7<<4)+4);
}
#define init_11_5_7_5
void FASTCALL call_11_5_7_5(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(7<<4)+5);
}
#define init_11_5_7_6
void FASTCALL call_11_5_7_6(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(7<<4)+6);
}
#define init_11_5_7_7
void FASTCALL call_11_5_7_7(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(7<<4)+7);
}
#define init_11_5_7_8
void FASTCALL call_11_5_7_8(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(7<<4)+8);
}
#define init_11_5_7_9
void FASTCALL call_11_5_7_9(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(7<<4)+9);
}
#define init_11_5_7_10
void FASTCALL call_11_5_7_10(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(7<<4)+10);
}
#define init_11_5_7_11
void FASTCALL call_11_5_7_11(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(7<<4)+11);
}
#define init_11_5_7_12
void FASTCALL call_11_5_7_12(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(7<<4)+12);
}
#define init_11_5_7_13
void FASTCALL call_11_5_7_13(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(7<<4)+13);
}
#define init_11_5_7_14
void FASTCALL call_11_5_7_14(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(7<<4)+14);
}
#define init_11_5_7_15
void FASTCALL call_11_5_7_15(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(7<<4)+15);
}
#define init_11_5_8_0
void FASTCALL call_11_5_8_0(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(8<<4)+0);
}
#define init_11_5_8_1
void FASTCALL call_11_5_8_1(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(8<<4)+1);
}
#define init_11_5_8_2
void FASTCALL call_11_5_8_2(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(8<<4)+2);
}
#define init_11_5_8_3
void FASTCALL call_11_5_8_3(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(8<<4)+3);
}
#define init_11_5_8_4
void FASTCALL call_11_5_8_4(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(8<<4)+4);
}
#define init_11_5_8_5
void FASTCALL call_11_5_8_5(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(8<<4)+5);
}
#define init_11_5_8_6
void FASTCALL call_11_5_8_6(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(8<<4)+6);
}
#define init_11_5_8_7
void FASTCALL call_11_5_8_7(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(8<<4)+7);
}
#define init_11_5_8_8
void FASTCALL call_11_5_8_8(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(8<<4)+8);
}
#define init_11_5_8_9
void FASTCALL call_11_5_8_9(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(8<<4)+9);
}
#define init_11_5_8_10
void FASTCALL call_11_5_8_10(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(8<<4)+10);
}
#define init_11_5_8_11
void FASTCALL call_11_5_8_11(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(8<<4)+11);
}
#define init_11_5_8_12
void FASTCALL call_11_5_8_12(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(8<<4)+12);
}
#define init_11_5_8_13
void FASTCALL call_11_5_8_13(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(8<<4)+13);
}
#define init_11_5_8_14
void FASTCALL call_11_5_8_14(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(8<<4)+14);
}
#define init_11_5_8_15
void FASTCALL call_11_5_8_15(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(8<<4)+15);
}
#define init_11_5_9_0
void FASTCALL call_11_5_9_0(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(9<<4)+0);
}
#define init_11_5_9_1
void FASTCALL call_11_5_9_1(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(9<<4)+1);
}
#define init_11_5_9_2
void FASTCALL call_11_5_9_2(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(9<<4)+2);
}
#define init_11_5_9_3
void FASTCALL call_11_5_9_3(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(9<<4)+3);
}
#define init_11_5_9_4
void FASTCALL call_11_5_9_4(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(9<<4)+4);
}
#define init_11_5_9_5
void FASTCALL call_11_5_9_5(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(9<<4)+5);
}
#define init_11_5_9_6
void FASTCALL call_11_5_9_6(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(9<<4)+6);
}
#define init_11_5_9_7
void FASTCALL call_11_5_9_7(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(9<<4)+7);
}
#define init_11_5_9_8
void FASTCALL call_11_5_9_8(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(9<<4)+8);
}
#define init_11_5_9_9
void FASTCALL call_11_5_9_9(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(9<<4)+9);
}
#define init_11_5_9_10
void FASTCALL call_11_5_9_10(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(9<<4)+10);
}
#define init_11_5_9_11
void FASTCALL call_11_5_9_11(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(9<<4)+11);
}
#define init_11_5_9_12
void FASTCALL call_11_5_9_12(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(9<<4)+12);
}
#define init_11_5_9_13
void FASTCALL call_11_5_9_13(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(9<<4)+13);
}
#define init_11_5_9_14
void FASTCALL call_11_5_9_14(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(9<<4)+14);
}
#define init_11_5_9_15
void FASTCALL call_11_5_9_15(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(9<<4)+15);
}
#define init_11_5_10_0
void FASTCALL call_11_5_10_0(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(10<<4)+0);
}
#define init_11_5_10_1
void FASTCALL call_11_5_10_1(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(10<<4)+1);
}
#define init_11_5_10_2
void FASTCALL call_11_5_10_2(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(10<<4)+2);
}
#define init_11_5_10_3
void FASTCALL call_11_5_10_3(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(10<<4)+3);
}
#define init_11_5_10_4
void FASTCALL call_11_5_10_4(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(10<<4)+4);
}
#define init_11_5_10_5
void FASTCALL call_11_5_10_5(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(10<<4)+5);
}
#define init_11_5_10_6
void FASTCALL call_11_5_10_6(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(10<<4)+6);
}
#define init_11_5_10_7
void FASTCALL call_11_5_10_7(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(10<<4)+7);
}
#define init_11_5_10_8
void FASTCALL call_11_5_10_8(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(10<<4)+8);
}
#define init_11_5_10_9
void FASTCALL call_11_5_10_9(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(10<<4)+9);
}
#define init_11_5_10_10
void FASTCALL call_11_5_10_10(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(10<<4)+10);
}
#define init_11_5_10_11
void FASTCALL call_11_5_10_11(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(10<<4)+11);
}
#define init_11_5_10_12
void FASTCALL call_11_5_10_12(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(10<<4)+12);
}
#define init_11_5_10_13
void FASTCALL call_11_5_10_13(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(10<<4)+13);
}
#define init_11_5_10_14
void FASTCALL call_11_5_10_14(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(10<<4)+14);
}
#define init_11_5_10_15
void FASTCALL call_11_5_10_15(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(10<<4)+15);
}
#define init_11_5_11_0
void FASTCALL call_11_5_11_0(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(11<<4)+0);
}
#define init_11_5_11_1
void FASTCALL call_11_5_11_1(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(11<<4)+1);
}
#define init_11_5_11_2
void FASTCALL call_11_5_11_2(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(11<<4)+2);
}
#define init_11_5_11_3
void FASTCALL call_11_5_11_3(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(11<<4)+3);
}
#define init_11_5_11_4
void FASTCALL call_11_5_11_4(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(11<<4)+4);
}
#define init_11_5_11_5
void FASTCALL call_11_5_11_5(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(11<<4)+5);
}
#define init_11_5_11_6
void FASTCALL call_11_5_11_6(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(11<<4)+6);
}
#define init_11_5_11_7
void FASTCALL call_11_5_11_7(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(11<<4)+7);
}
#define init_11_5_11_8
void FASTCALL call_11_5_11_8(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(11<<4)+8);
}
#define init_11_5_11_9
void FASTCALL call_11_5_11_9(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(11<<4)+9);
}
#define init_11_5_11_10
void FASTCALL call_11_5_11_10(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(11<<4)+10);
}
#define init_11_5_11_11
void FASTCALL call_11_5_11_11(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(11<<4)+11);
}
#define init_11_5_11_12
void FASTCALL call_11_5_11_12(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(11<<4)+12);
}
#define init_11_5_11_13
void FASTCALL call_11_5_11_13(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(11<<4)+13);
}
#define init_11_5_11_14
void FASTCALL call_11_5_11_14(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(11<<4)+14);
}
#define init_11_5_11_15
void FASTCALL call_11_5_11_15(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(11<<4)+15);
}
#define init_11_5_12_0
void FASTCALL call_11_5_12_0(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(12<<4)+0);
}
#define init_11_5_12_1
void FASTCALL call_11_5_12_1(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(12<<4)+1);
}
#define init_11_5_12_2
void FASTCALL call_11_5_12_2(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(12<<4)+2);
}
#define init_11_5_12_3
void FASTCALL call_11_5_12_3(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(12<<4)+3);
}
#define init_11_5_12_4
void FASTCALL call_11_5_12_4(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(12<<4)+4);
}
#define init_11_5_12_5
void FASTCALL call_11_5_12_5(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(12<<4)+5);
}
#define init_11_5_12_6
void FASTCALL call_11_5_12_6(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(12<<4)+6);
}
#define init_11_5_12_7
void FASTCALL call_11_5_12_7(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(12<<4)+7);
}
#define init_11_5_12_8
void FASTCALL call_11_5_12_8(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(12<<4)+8);
}
#define init_11_5_12_9
void FASTCALL call_11_5_12_9(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(12<<4)+9);
}
#define init_11_5_12_10
void FASTCALL call_11_5_12_10(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(12<<4)+10);
}
#define init_11_5_12_11
void FASTCALL call_11_5_12_11(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(12<<4)+11);
}
#define init_11_5_12_12
void FASTCALL call_11_5_12_12(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(12<<4)+12);
}
#define init_11_5_12_13
void FASTCALL call_11_5_12_13(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(12<<4)+13);
}
#define init_11_5_12_14
void FASTCALL call_11_5_12_14(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(12<<4)+14);
}
#define init_11_5_12_15
void FASTCALL call_11_5_12_15(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(12<<4)+15);
}
#define init_11_5_13_0
void FASTCALL call_11_5_13_0(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(13<<4)+0);
}
#define init_11_5_13_1
void FASTCALL call_11_5_13_1(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(13<<4)+1);
}
#define init_11_5_13_2
void FASTCALL call_11_5_13_2(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(13<<4)+2);
}
#define init_11_5_13_3
void FASTCALL call_11_5_13_3(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(13<<4)+3);
}
#define init_11_5_13_4
void FASTCALL call_11_5_13_4(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(13<<4)+4);
}
#define init_11_5_13_5
void FASTCALL call_11_5_13_5(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(13<<4)+5);
}
#define init_11_5_13_6
void FASTCALL call_11_5_13_6(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(13<<4)+6);
}
#define init_11_5_13_7
void FASTCALL call_11_5_13_7(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(13<<4)+7);
}
#define init_11_5_13_8
void FASTCALL call_11_5_13_8(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(13<<4)+8);
}
#define init_11_5_13_9
void FASTCALL call_11_5_13_9(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(13<<4)+9);
}
#define init_11_5_13_10
void FASTCALL call_11_5_13_10(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(13<<4)+10);
}
#define init_11_5_13_11
void FASTCALL call_11_5_13_11(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(13<<4)+11);
}
#define init_11_5_13_12
void FASTCALL call_11_5_13_12(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(13<<4)+12);
}
#define init_11_5_13_13
void FASTCALL call_11_5_13_13(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(13<<4)+13);
}
#define init_11_5_13_14
void FASTCALL call_11_5_13_14(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(13<<4)+14);
}
#define init_11_5_13_15
void FASTCALL call_11_5_13_15(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(13<<4)+15);
}
#define init_11_5_14_0
void FASTCALL call_11_5_14_0(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(14<<4)+0);
}
#define init_11_5_14_1
void FASTCALL call_11_5_14_1(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(14<<4)+1);
}
#define init_11_5_14_2
void FASTCALL call_11_5_14_2(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(14<<4)+2);
}
#define init_11_5_14_3
void FASTCALL call_11_5_14_3(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(14<<4)+3);
}
#define init_11_5_14_4
void FASTCALL call_11_5_14_4(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(14<<4)+4);
}
#define init_11_5_14_5
void FASTCALL call_11_5_14_5(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(14<<4)+5);
}
#define init_11_5_14_6
void FASTCALL call_11_5_14_6(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(14<<4)+6);
}
#define init_11_5_14_7
void FASTCALL call_11_5_14_7(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(14<<4)+7);
}
#define init_11_5_14_8
void FASTCALL call_11_5_14_8(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(14<<4)+8);
}
#define init_11_5_14_9
void FASTCALL call_11_5_14_9(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(14<<4)+9);
}
#define init_11_5_14_10
void FASTCALL call_11_5_14_10(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(14<<4)+10);
}
#define init_11_5_14_11
void FASTCALL call_11_5_14_11(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(14<<4)+11);
}
#define init_11_5_14_12
void FASTCALL call_11_5_14_12(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(14<<4)+12);
}
#define init_11_5_14_13
void FASTCALL call_11_5_14_13(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(14<<4)+13);
}
#define init_11_5_14_14
void FASTCALL call_11_5_14_14(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(14<<4)+14);
}
#define init_11_5_14_15
void FASTCALL call_11_5_14_15(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(14<<4)+15);
}
#define init_11_5_15_0
void FASTCALL call_11_5_15_0(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(15<<4)+0);
}
#define init_11_5_15_1
void FASTCALL call_11_5_15_1(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(15<<4)+1);
}
#define init_11_5_15_2
void FASTCALL call_11_5_15_2(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(15<<4)+2);
}
#define init_11_5_15_3
void FASTCALL call_11_5_15_3(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(15<<4)+3);
}
#define init_11_5_15_4
void FASTCALL call_11_5_15_4(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(15<<4)+4);
}
#define init_11_5_15_5
void FASTCALL call_11_5_15_5(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(15<<4)+5);
}
#define init_11_5_15_6
void FASTCALL call_11_5_15_6(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(15<<4)+6);
}
#define init_11_5_15_7
void FASTCALL call_11_5_15_7(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(15<<4)+7);
}
#define init_11_5_15_8
void FASTCALL call_11_5_15_8(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(15<<4)+8);
}
#define init_11_5_15_9
void FASTCALL call_11_5_15_9(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(15<<4)+9);
}
#define init_11_5_15_10
void FASTCALL call_11_5_15_10(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(15<<4)+10);
}
#define init_11_5_15_11
void FASTCALL call_11_5_15_11(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(15<<4)+11);
}
#define init_11_5_15_12
void FASTCALL call_11_5_15_12(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(15<<4)+12);
}
#define init_11_5_15_13
void FASTCALL call_11_5_15_13(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(15<<4)+13);
}
#define init_11_5_15_14
void FASTCALL call_11_5_15_14(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(15<<4)+14);
}
#define init_11_5_15_15
void FASTCALL call_11_5_15_15(SH2_struct* sh)
{
  SH2bsr(sh, (5<<8)+(15<<4)+15);
}
#define init_11_6_0_0
void FASTCALL call_11_6_0_0(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(0<<4)+0);
}
#define init_11_6_0_1
void FASTCALL call_11_6_0_1(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(0<<4)+1);
}
#define init_11_6_0_2
void FASTCALL call_11_6_0_2(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(0<<4)+2);
}
#define init_11_6_0_3
void FASTCALL call_11_6_0_3(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(0<<4)+3);
}
#define init_11_6_0_4
void FASTCALL call_11_6_0_4(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(0<<4)+4);
}
#define init_11_6_0_5
void FASTCALL call_11_6_0_5(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(0<<4)+5);
}
#define init_11_6_0_6
void FASTCALL call_11_6_0_6(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(0<<4)+6);
}
#define init_11_6_0_7
void FASTCALL call_11_6_0_7(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(0<<4)+7);
}
#define init_11_6_0_8
void FASTCALL call_11_6_0_8(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(0<<4)+8);
}
#define init_11_6_0_9
void FASTCALL call_11_6_0_9(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(0<<4)+9);
}
#define init_11_6_0_10
void FASTCALL call_11_6_0_10(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(0<<4)+10);
}
#define init_11_6_0_11
void FASTCALL call_11_6_0_11(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(0<<4)+11);
}
#define init_11_6_0_12
void FASTCALL call_11_6_0_12(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(0<<4)+12);
}
#define init_11_6_0_13
void FASTCALL call_11_6_0_13(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(0<<4)+13);
}
#define init_11_6_0_14
void FASTCALL call_11_6_0_14(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(0<<4)+14);
}
#define init_11_6_0_15
void FASTCALL call_11_6_0_15(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(0<<4)+15);
}
#define init_11_6_1_0
void FASTCALL call_11_6_1_0(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(1<<4)+0);
}
#define init_11_6_1_1
void FASTCALL call_11_6_1_1(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(1<<4)+1);
}
#define init_11_6_1_2
void FASTCALL call_11_6_1_2(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(1<<4)+2);
}
#define init_11_6_1_3
void FASTCALL call_11_6_1_3(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(1<<4)+3);
}
#define init_11_6_1_4
void FASTCALL call_11_6_1_4(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(1<<4)+4);
}
#define init_11_6_1_5
void FASTCALL call_11_6_1_5(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(1<<4)+5);
}
#define init_11_6_1_6
void FASTCALL call_11_6_1_6(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(1<<4)+6);
}
#define init_11_6_1_7
void FASTCALL call_11_6_1_7(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(1<<4)+7);
}
#define init_11_6_1_8
void FASTCALL call_11_6_1_8(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(1<<4)+8);
}
#define init_11_6_1_9
void FASTCALL call_11_6_1_9(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(1<<4)+9);
}
#define init_11_6_1_10
void FASTCALL call_11_6_1_10(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(1<<4)+10);
}
#define init_11_6_1_11
void FASTCALL call_11_6_1_11(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(1<<4)+11);
}
#define init_11_6_1_12
void FASTCALL call_11_6_1_12(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(1<<4)+12);
}
#define init_11_6_1_13
void FASTCALL call_11_6_1_13(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(1<<4)+13);
}
#define init_11_6_1_14
void FASTCALL call_11_6_1_14(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(1<<4)+14);
}
#define init_11_6_1_15
void FASTCALL call_11_6_1_15(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(1<<4)+15);
}
#define init_11_6_2_0
void FASTCALL call_11_6_2_0(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(2<<4)+0);
}
#define init_11_6_2_1
void FASTCALL call_11_6_2_1(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(2<<4)+1);
}
#define init_11_6_2_2
void FASTCALL call_11_6_2_2(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(2<<4)+2);
}
#define init_11_6_2_3
void FASTCALL call_11_6_2_3(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(2<<4)+3);
}
#define init_11_6_2_4
void FASTCALL call_11_6_2_4(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(2<<4)+4);
}
#define init_11_6_2_5
void FASTCALL call_11_6_2_5(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(2<<4)+5);
}
#define init_11_6_2_6
void FASTCALL call_11_6_2_6(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(2<<4)+6);
}
#define init_11_6_2_7
void FASTCALL call_11_6_2_7(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(2<<4)+7);
}
#define init_11_6_2_8
void FASTCALL call_11_6_2_8(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(2<<4)+8);
}
#define init_11_6_2_9
void FASTCALL call_11_6_2_9(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(2<<4)+9);
}
#define init_11_6_2_10
void FASTCALL call_11_6_2_10(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(2<<4)+10);
}
#define init_11_6_2_11
void FASTCALL call_11_6_2_11(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(2<<4)+11);
}
#define init_11_6_2_12
void FASTCALL call_11_6_2_12(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(2<<4)+12);
}
#define init_11_6_2_13
void FASTCALL call_11_6_2_13(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(2<<4)+13);
}
#define init_11_6_2_14
void FASTCALL call_11_6_2_14(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(2<<4)+14);
}
#define init_11_6_2_15
void FASTCALL call_11_6_2_15(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(2<<4)+15);
}
#define init_11_6_3_0
void FASTCALL call_11_6_3_0(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(3<<4)+0);
}
#define init_11_6_3_1
void FASTCALL call_11_6_3_1(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(3<<4)+1);
}
#define init_11_6_3_2
void FASTCALL call_11_6_3_2(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(3<<4)+2);
}
#define init_11_6_3_3
void FASTCALL call_11_6_3_3(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(3<<4)+3);
}
#define init_11_6_3_4
void FASTCALL call_11_6_3_4(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(3<<4)+4);
}
#define init_11_6_3_5
void FASTCALL call_11_6_3_5(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(3<<4)+5);
}
#define init_11_6_3_6
void FASTCALL call_11_6_3_6(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(3<<4)+6);
}
#define init_11_6_3_7
void FASTCALL call_11_6_3_7(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(3<<4)+7);
}
#define init_11_6_3_8
void FASTCALL call_11_6_3_8(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(3<<4)+8);
}
#define init_11_6_3_9
void FASTCALL call_11_6_3_9(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(3<<4)+9);
}
#define init_11_6_3_10
void FASTCALL call_11_6_3_10(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(3<<4)+10);
}
#define init_11_6_3_11
void FASTCALL call_11_6_3_11(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(3<<4)+11);
}
#define init_11_6_3_12
void FASTCALL call_11_6_3_12(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(3<<4)+12);
}
#define init_11_6_3_13
void FASTCALL call_11_6_3_13(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(3<<4)+13);
}
#define init_11_6_3_14
void FASTCALL call_11_6_3_14(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(3<<4)+14);
}
#define init_11_6_3_15
void FASTCALL call_11_6_3_15(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(3<<4)+15);
}
#define init_11_6_4_0
void FASTCALL call_11_6_4_0(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(4<<4)+0);
}
#define init_11_6_4_1
void FASTCALL call_11_6_4_1(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(4<<4)+1);
}
#define init_11_6_4_2
void FASTCALL call_11_6_4_2(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(4<<4)+2);
}
#define init_11_6_4_3
void FASTCALL call_11_6_4_3(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(4<<4)+3);
}
#define init_11_6_4_4
void FASTCALL call_11_6_4_4(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(4<<4)+4);
}
#define init_11_6_4_5
void FASTCALL call_11_6_4_5(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(4<<4)+5);
}
#define init_11_6_4_6
void FASTCALL call_11_6_4_6(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(4<<4)+6);
}
#define init_11_6_4_7
void FASTCALL call_11_6_4_7(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(4<<4)+7);
}
#define init_11_6_4_8
void FASTCALL call_11_6_4_8(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(4<<4)+8);
}
#define init_11_6_4_9
void FASTCALL call_11_6_4_9(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(4<<4)+9);
}
#define init_11_6_4_10
void FASTCALL call_11_6_4_10(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(4<<4)+10);
}
#define init_11_6_4_11
void FASTCALL call_11_6_4_11(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(4<<4)+11);
}
#define init_11_6_4_12
void FASTCALL call_11_6_4_12(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(4<<4)+12);
}
#define init_11_6_4_13
void FASTCALL call_11_6_4_13(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(4<<4)+13);
}
#define init_11_6_4_14
void FASTCALL call_11_6_4_14(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(4<<4)+14);
}
#define init_11_6_4_15
void FASTCALL call_11_6_4_15(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(4<<4)+15);
}
#define init_11_6_5_0
void FASTCALL call_11_6_5_0(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(5<<4)+0);
}
#define init_11_6_5_1
void FASTCALL call_11_6_5_1(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(5<<4)+1);
}
#define init_11_6_5_2
void FASTCALL call_11_6_5_2(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(5<<4)+2);
}
#define init_11_6_5_3
void FASTCALL call_11_6_5_3(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(5<<4)+3);
}
#define init_11_6_5_4
void FASTCALL call_11_6_5_4(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(5<<4)+4);
}
#define init_11_6_5_5
void FASTCALL call_11_6_5_5(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(5<<4)+5);
}
#define init_11_6_5_6
void FASTCALL call_11_6_5_6(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(5<<4)+6);
}
#define init_11_6_5_7
void FASTCALL call_11_6_5_7(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(5<<4)+7);
}
#define init_11_6_5_8
void FASTCALL call_11_6_5_8(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(5<<4)+8);
}
#define init_11_6_5_9
void FASTCALL call_11_6_5_9(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(5<<4)+9);
}
#define init_11_6_5_10
void FASTCALL call_11_6_5_10(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(5<<4)+10);
}
#define init_11_6_5_11
void FASTCALL call_11_6_5_11(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(5<<4)+11);
}
#define init_11_6_5_12
void FASTCALL call_11_6_5_12(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(5<<4)+12);
}
#define init_11_6_5_13
void FASTCALL call_11_6_5_13(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(5<<4)+13);
}
#define init_11_6_5_14
void FASTCALL call_11_6_5_14(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(5<<4)+14);
}
#define init_11_6_5_15
void FASTCALL call_11_6_5_15(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(5<<4)+15);
}
#define init_11_6_6_0
void FASTCALL call_11_6_6_0(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(6<<4)+0);
}
#define init_11_6_6_1
void FASTCALL call_11_6_6_1(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(6<<4)+1);
}
#define init_11_6_6_2
void FASTCALL call_11_6_6_2(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(6<<4)+2);
}
#define init_11_6_6_3
void FASTCALL call_11_6_6_3(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(6<<4)+3);
}
#define init_11_6_6_4
void FASTCALL call_11_6_6_4(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(6<<4)+4);
}
#define init_11_6_6_5
void FASTCALL call_11_6_6_5(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(6<<4)+5);
}
#define init_11_6_6_6
void FASTCALL call_11_6_6_6(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(6<<4)+6);
}
#define init_11_6_6_7
void FASTCALL call_11_6_6_7(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(6<<4)+7);
}
#define init_11_6_6_8
void FASTCALL call_11_6_6_8(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(6<<4)+8);
}
#define init_11_6_6_9
void FASTCALL call_11_6_6_9(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(6<<4)+9);
}
#define init_11_6_6_10
void FASTCALL call_11_6_6_10(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(6<<4)+10);
}
#define init_11_6_6_11
void FASTCALL call_11_6_6_11(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(6<<4)+11);
}
#define init_11_6_6_12
void FASTCALL call_11_6_6_12(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(6<<4)+12);
}
#define init_11_6_6_13
void FASTCALL call_11_6_6_13(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(6<<4)+13);
}
#define init_11_6_6_14
void FASTCALL call_11_6_6_14(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(6<<4)+14);
}
#define init_11_6_6_15
void FASTCALL call_11_6_6_15(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(6<<4)+15);
}
#define init_11_6_7_0
void FASTCALL call_11_6_7_0(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(7<<4)+0);
}
#define init_11_6_7_1
void FASTCALL call_11_6_7_1(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(7<<4)+1);
}
#define init_11_6_7_2
void FASTCALL call_11_6_7_2(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(7<<4)+2);
}
#define init_11_6_7_3
void FASTCALL call_11_6_7_3(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(7<<4)+3);
}
#define init_11_6_7_4
void FASTCALL call_11_6_7_4(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(7<<4)+4);
}
#define init_11_6_7_5
void FASTCALL call_11_6_7_5(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(7<<4)+5);
}
#define init_11_6_7_6
void FASTCALL call_11_6_7_6(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(7<<4)+6);
}
#define init_11_6_7_7
void FASTCALL call_11_6_7_7(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(7<<4)+7);
}
#define init_11_6_7_8
void FASTCALL call_11_6_7_8(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(7<<4)+8);
}
#define init_11_6_7_9
void FASTCALL call_11_6_7_9(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(7<<4)+9);
}
#define init_11_6_7_10
void FASTCALL call_11_6_7_10(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(7<<4)+10);
}
#define init_11_6_7_11
void FASTCALL call_11_6_7_11(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(7<<4)+11);
}
#define init_11_6_7_12
void FASTCALL call_11_6_7_12(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(7<<4)+12);
}
#define init_11_6_7_13
void FASTCALL call_11_6_7_13(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(7<<4)+13);
}
#define init_11_6_7_14
void FASTCALL call_11_6_7_14(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(7<<4)+14);
}
#define init_11_6_7_15
void FASTCALL call_11_6_7_15(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(7<<4)+15);
}
#define init_11_6_8_0
void FASTCALL call_11_6_8_0(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(8<<4)+0);
}
#define init_11_6_8_1
void FASTCALL call_11_6_8_1(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(8<<4)+1);
}
#define init_11_6_8_2
void FASTCALL call_11_6_8_2(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(8<<4)+2);
}
#define init_11_6_8_3
void FASTCALL call_11_6_8_3(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(8<<4)+3);
}
#define init_11_6_8_4
void FASTCALL call_11_6_8_4(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(8<<4)+4);
}
#define init_11_6_8_5
void FASTCALL call_11_6_8_5(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(8<<4)+5);
}
#define init_11_6_8_6
void FASTCALL call_11_6_8_6(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(8<<4)+6);
}
#define init_11_6_8_7
void FASTCALL call_11_6_8_7(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(8<<4)+7);
}
#define init_11_6_8_8
void FASTCALL call_11_6_8_8(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(8<<4)+8);
}
#define init_11_6_8_9
void FASTCALL call_11_6_8_9(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(8<<4)+9);
}
#define init_11_6_8_10
void FASTCALL call_11_6_8_10(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(8<<4)+10);
}
#define init_11_6_8_11
void FASTCALL call_11_6_8_11(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(8<<4)+11);
}
#define init_11_6_8_12
void FASTCALL call_11_6_8_12(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(8<<4)+12);
}
#define init_11_6_8_13
void FASTCALL call_11_6_8_13(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(8<<4)+13);
}
#define init_11_6_8_14
void FASTCALL call_11_6_8_14(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(8<<4)+14);
}
#define init_11_6_8_15
void FASTCALL call_11_6_8_15(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(8<<4)+15);
}
#define init_11_6_9_0
void FASTCALL call_11_6_9_0(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(9<<4)+0);
}
#define init_11_6_9_1
void FASTCALL call_11_6_9_1(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(9<<4)+1);
}
#define init_11_6_9_2
void FASTCALL call_11_6_9_2(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(9<<4)+2);
}
#define init_11_6_9_3
void FASTCALL call_11_6_9_3(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(9<<4)+3);
}
#define init_11_6_9_4
void FASTCALL call_11_6_9_4(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(9<<4)+4);
}
#define init_11_6_9_5
void FASTCALL call_11_6_9_5(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(9<<4)+5);
}
#define init_11_6_9_6
void FASTCALL call_11_6_9_6(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(9<<4)+6);
}
#define init_11_6_9_7
void FASTCALL call_11_6_9_7(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(9<<4)+7);
}
#define init_11_6_9_8
void FASTCALL call_11_6_9_8(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(9<<4)+8);
}
#define init_11_6_9_9
void FASTCALL call_11_6_9_9(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(9<<4)+9);
}
#define init_11_6_9_10
void FASTCALL call_11_6_9_10(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(9<<4)+10);
}
#define init_11_6_9_11
void FASTCALL call_11_6_9_11(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(9<<4)+11);
}
#define init_11_6_9_12
void FASTCALL call_11_6_9_12(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(9<<4)+12);
}
#define init_11_6_9_13
void FASTCALL call_11_6_9_13(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(9<<4)+13);
}
#define init_11_6_9_14
void FASTCALL call_11_6_9_14(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(9<<4)+14);
}
#define init_11_6_9_15
void FASTCALL call_11_6_9_15(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(9<<4)+15);
}
#define init_11_6_10_0
void FASTCALL call_11_6_10_0(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(10<<4)+0);
}
#define init_11_6_10_1
void FASTCALL call_11_6_10_1(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(10<<4)+1);
}
#define init_11_6_10_2
void FASTCALL call_11_6_10_2(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(10<<4)+2);
}
#define init_11_6_10_3
void FASTCALL call_11_6_10_3(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(10<<4)+3);
}
#define init_11_6_10_4
void FASTCALL call_11_6_10_4(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(10<<4)+4);
}
#define init_11_6_10_5
void FASTCALL call_11_6_10_5(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(10<<4)+5);
}
#define init_11_6_10_6
void FASTCALL call_11_6_10_6(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(10<<4)+6);
}
#define init_11_6_10_7
void FASTCALL call_11_6_10_7(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(10<<4)+7);
}
#define init_11_6_10_8
void FASTCALL call_11_6_10_8(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(10<<4)+8);
}
#define init_11_6_10_9
void FASTCALL call_11_6_10_9(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(10<<4)+9);
}
#define init_11_6_10_10
void FASTCALL call_11_6_10_10(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(10<<4)+10);
}
#define init_11_6_10_11
void FASTCALL call_11_6_10_11(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(10<<4)+11);
}
#define init_11_6_10_12
void FASTCALL call_11_6_10_12(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(10<<4)+12);
}
#define init_11_6_10_13
void FASTCALL call_11_6_10_13(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(10<<4)+13);
}
#define init_11_6_10_14
void FASTCALL call_11_6_10_14(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(10<<4)+14);
}
#define init_11_6_10_15
void FASTCALL call_11_6_10_15(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(10<<4)+15);
}
#define init_11_6_11_0
void FASTCALL call_11_6_11_0(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(11<<4)+0);
}
#define init_11_6_11_1
void FASTCALL call_11_6_11_1(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(11<<4)+1);
}
#define init_11_6_11_2
void FASTCALL call_11_6_11_2(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(11<<4)+2);
}
#define init_11_6_11_3
void FASTCALL call_11_6_11_3(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(11<<4)+3);
}
#define init_11_6_11_4
void FASTCALL call_11_6_11_4(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(11<<4)+4);
}
#define init_11_6_11_5
void FASTCALL call_11_6_11_5(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(11<<4)+5);
}
#define init_11_6_11_6
void FASTCALL call_11_6_11_6(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(11<<4)+6);
}
#define init_11_6_11_7
void FASTCALL call_11_6_11_7(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(11<<4)+7);
}
#define init_11_6_11_8
void FASTCALL call_11_6_11_8(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(11<<4)+8);
}
#define init_11_6_11_9
void FASTCALL call_11_6_11_9(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(11<<4)+9);
}
#define init_11_6_11_10
void FASTCALL call_11_6_11_10(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(11<<4)+10);
}
#define init_11_6_11_11
void FASTCALL call_11_6_11_11(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(11<<4)+11);
}
#define init_11_6_11_12
void FASTCALL call_11_6_11_12(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(11<<4)+12);
}
#define init_11_6_11_13
void FASTCALL call_11_6_11_13(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(11<<4)+13);
}
#define init_11_6_11_14
void FASTCALL call_11_6_11_14(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(11<<4)+14);
}
#define init_11_6_11_15
void FASTCALL call_11_6_11_15(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(11<<4)+15);
}
#define init_11_6_12_0
void FASTCALL call_11_6_12_0(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(12<<4)+0);
}
#define init_11_6_12_1
void FASTCALL call_11_6_12_1(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(12<<4)+1);
}
#define init_11_6_12_2
void FASTCALL call_11_6_12_2(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(12<<4)+2);
}
#define init_11_6_12_3
void FASTCALL call_11_6_12_3(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(12<<4)+3);
}
#define init_11_6_12_4
void FASTCALL call_11_6_12_4(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(12<<4)+4);
}
#define init_11_6_12_5
void FASTCALL call_11_6_12_5(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(12<<4)+5);
}
#define init_11_6_12_6
void FASTCALL call_11_6_12_6(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(12<<4)+6);
}
#define init_11_6_12_7
void FASTCALL call_11_6_12_7(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(12<<4)+7);
}
#define init_11_6_12_8
void FASTCALL call_11_6_12_8(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(12<<4)+8);
}
#define init_11_6_12_9
void FASTCALL call_11_6_12_9(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(12<<4)+9);
}
#define init_11_6_12_10
void FASTCALL call_11_6_12_10(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(12<<4)+10);
}
#define init_11_6_12_11
void FASTCALL call_11_6_12_11(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(12<<4)+11);
}
#define init_11_6_12_12
void FASTCALL call_11_6_12_12(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(12<<4)+12);
}
#define init_11_6_12_13
void FASTCALL call_11_6_12_13(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(12<<4)+13);
}
#define init_11_6_12_14
void FASTCALL call_11_6_12_14(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(12<<4)+14);
}
#define init_11_6_12_15
void FASTCALL call_11_6_12_15(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(12<<4)+15);
}
#define init_11_6_13_0
void FASTCALL call_11_6_13_0(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(13<<4)+0);
}
#define init_11_6_13_1
void FASTCALL call_11_6_13_1(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(13<<4)+1);
}
#define init_11_6_13_2
void FASTCALL call_11_6_13_2(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(13<<4)+2);
}
#define init_11_6_13_3
void FASTCALL call_11_6_13_3(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(13<<4)+3);
}
#define init_11_6_13_4
void FASTCALL call_11_6_13_4(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(13<<4)+4);
}
#define init_11_6_13_5
void FASTCALL call_11_6_13_5(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(13<<4)+5);
}
#define init_11_6_13_6
void FASTCALL call_11_6_13_6(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(13<<4)+6);
}
#define init_11_6_13_7
void FASTCALL call_11_6_13_7(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(13<<4)+7);
}
#define init_11_6_13_8
void FASTCALL call_11_6_13_8(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(13<<4)+8);
}
#define init_11_6_13_9
void FASTCALL call_11_6_13_9(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(13<<4)+9);
}
#define init_11_6_13_10
void FASTCALL call_11_6_13_10(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(13<<4)+10);
}
#define init_11_6_13_11
void FASTCALL call_11_6_13_11(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(13<<4)+11);
}
#define init_11_6_13_12
void FASTCALL call_11_6_13_12(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(13<<4)+12);
}
#define init_11_6_13_13
void FASTCALL call_11_6_13_13(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(13<<4)+13);
}
#define init_11_6_13_14
void FASTCALL call_11_6_13_14(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(13<<4)+14);
}
#define init_11_6_13_15
void FASTCALL call_11_6_13_15(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(13<<4)+15);
}
#define init_11_6_14_0
void FASTCALL call_11_6_14_0(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(14<<4)+0);
}
#define init_11_6_14_1
void FASTCALL call_11_6_14_1(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(14<<4)+1);
}
#define init_11_6_14_2
void FASTCALL call_11_6_14_2(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(14<<4)+2);
}
#define init_11_6_14_3
void FASTCALL call_11_6_14_3(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(14<<4)+3);
}
#define init_11_6_14_4
void FASTCALL call_11_6_14_4(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(14<<4)+4);
}
#define init_11_6_14_5
void FASTCALL call_11_6_14_5(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(14<<4)+5);
}
#define init_11_6_14_6
void FASTCALL call_11_6_14_6(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(14<<4)+6);
}
#define init_11_6_14_7
void FASTCALL call_11_6_14_7(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(14<<4)+7);
}
#define init_11_6_14_8
void FASTCALL call_11_6_14_8(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(14<<4)+8);
}
#define init_11_6_14_9
void FASTCALL call_11_6_14_9(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(14<<4)+9);
}
#define init_11_6_14_10
void FASTCALL call_11_6_14_10(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(14<<4)+10);
}
#define init_11_6_14_11
void FASTCALL call_11_6_14_11(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(14<<4)+11);
}
#define init_11_6_14_12
void FASTCALL call_11_6_14_12(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(14<<4)+12);
}
#define init_11_6_14_13
void FASTCALL call_11_6_14_13(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(14<<4)+13);
}
#define init_11_6_14_14
void FASTCALL call_11_6_14_14(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(14<<4)+14);
}
#define init_11_6_14_15
void FASTCALL call_11_6_14_15(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(14<<4)+15);
}
#define init_11_6_15_0
void FASTCALL call_11_6_15_0(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(15<<4)+0);
}
#define init_11_6_15_1
void FASTCALL call_11_6_15_1(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(15<<4)+1);
}
#define init_11_6_15_2
void FASTCALL call_11_6_15_2(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(15<<4)+2);
}
#define init_11_6_15_3
void FASTCALL call_11_6_15_3(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(15<<4)+3);
}
#define init_11_6_15_4
void FASTCALL call_11_6_15_4(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(15<<4)+4);
}
#define init_11_6_15_5
void FASTCALL call_11_6_15_5(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(15<<4)+5);
}
#define init_11_6_15_6
void FASTCALL call_11_6_15_6(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(15<<4)+6);
}
#define init_11_6_15_7
void FASTCALL call_11_6_15_7(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(15<<4)+7);
}
#define init_11_6_15_8
void FASTCALL call_11_6_15_8(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(15<<4)+8);
}
#define init_11_6_15_9
void FASTCALL call_11_6_15_9(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(15<<4)+9);
}
#define init_11_6_15_10
void FASTCALL call_11_6_15_10(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(15<<4)+10);
}
#define init_11_6_15_11
void FASTCALL call_11_6_15_11(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(15<<4)+11);
}
#define init_11_6_15_12
void FASTCALL call_11_6_15_12(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(15<<4)+12);
}
#define init_11_6_15_13
void FASTCALL call_11_6_15_13(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(15<<4)+13);
}
#define init_11_6_15_14
void FASTCALL call_11_6_15_14(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(15<<4)+14);
}
#define init_11_6_15_15
void FASTCALL call_11_6_15_15(SH2_struct* sh)
{
  SH2bsr(sh, (6<<8)+(15<<4)+15);
}
#define init_11_7_0_0
void FASTCALL call_11_7_0_0(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(0<<4)+0);
}
#define init_11_7_0_1
void FASTCALL call_11_7_0_1(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(0<<4)+1);
}
#define init_11_7_0_2
void FASTCALL call_11_7_0_2(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(0<<4)+2);
}
#define init_11_7_0_3
void FASTCALL call_11_7_0_3(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(0<<4)+3);
}
#define init_11_7_0_4
void FASTCALL call_11_7_0_4(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(0<<4)+4);
}
#define init_11_7_0_5
void FASTCALL call_11_7_0_5(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(0<<4)+5);
}
#define init_11_7_0_6
void FASTCALL call_11_7_0_6(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(0<<4)+6);
}
#define init_11_7_0_7
void FASTCALL call_11_7_0_7(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(0<<4)+7);
}
#define init_11_7_0_8
void FASTCALL call_11_7_0_8(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(0<<4)+8);
}
#define init_11_7_0_9
void FASTCALL call_11_7_0_9(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(0<<4)+9);
}
#define init_11_7_0_10
void FASTCALL call_11_7_0_10(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(0<<4)+10);
}
#define init_11_7_0_11
void FASTCALL call_11_7_0_11(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(0<<4)+11);
}
#define init_11_7_0_12
void FASTCALL call_11_7_0_12(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(0<<4)+12);
}
#define init_11_7_0_13
void FASTCALL call_11_7_0_13(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(0<<4)+13);
}
#define init_11_7_0_14
void FASTCALL call_11_7_0_14(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(0<<4)+14);
}
#define init_11_7_0_15
void FASTCALL call_11_7_0_15(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(0<<4)+15);
}
#define init_11_7_1_0
void FASTCALL call_11_7_1_0(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(1<<4)+0);
}
#define init_11_7_1_1
void FASTCALL call_11_7_1_1(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(1<<4)+1);
}
#define init_11_7_1_2
void FASTCALL call_11_7_1_2(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(1<<4)+2);
}
#define init_11_7_1_3
void FASTCALL call_11_7_1_3(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(1<<4)+3);
}
#define init_11_7_1_4
void FASTCALL call_11_7_1_4(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(1<<4)+4);
}
#define init_11_7_1_5
void FASTCALL call_11_7_1_5(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(1<<4)+5);
}
#define init_11_7_1_6
void FASTCALL call_11_7_1_6(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(1<<4)+6);
}
#define init_11_7_1_7
void FASTCALL call_11_7_1_7(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(1<<4)+7);
}
#define init_11_7_1_8
void FASTCALL call_11_7_1_8(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(1<<4)+8);
}
#define init_11_7_1_9
void FASTCALL call_11_7_1_9(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(1<<4)+9);
}
#define init_11_7_1_10
void FASTCALL call_11_7_1_10(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(1<<4)+10);
}
#define init_11_7_1_11
void FASTCALL call_11_7_1_11(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(1<<4)+11);
}
#define init_11_7_1_12
void FASTCALL call_11_7_1_12(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(1<<4)+12);
}
#define init_11_7_1_13
void FASTCALL call_11_7_1_13(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(1<<4)+13);
}
#define init_11_7_1_14
void FASTCALL call_11_7_1_14(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(1<<4)+14);
}
#define init_11_7_1_15
void FASTCALL call_11_7_1_15(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(1<<4)+15);
}
#define init_11_7_2_0
void FASTCALL call_11_7_2_0(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(2<<4)+0);
}
#define init_11_7_2_1
void FASTCALL call_11_7_2_1(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(2<<4)+1);
}
#define init_11_7_2_2
void FASTCALL call_11_7_2_2(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(2<<4)+2);
}
#define init_11_7_2_3
void FASTCALL call_11_7_2_3(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(2<<4)+3);
}
#define init_11_7_2_4
void FASTCALL call_11_7_2_4(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(2<<4)+4);
}
#define init_11_7_2_5
void FASTCALL call_11_7_2_5(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(2<<4)+5);
}
#define init_11_7_2_6
void FASTCALL call_11_7_2_6(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(2<<4)+6);
}
#define init_11_7_2_7
void FASTCALL call_11_7_2_7(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(2<<4)+7);
}
#define init_11_7_2_8
void FASTCALL call_11_7_2_8(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(2<<4)+8);
}
#define init_11_7_2_9
void FASTCALL call_11_7_2_9(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(2<<4)+9);
}
#define init_11_7_2_10
void FASTCALL call_11_7_2_10(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(2<<4)+10);
}
#define init_11_7_2_11
void FASTCALL call_11_7_2_11(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(2<<4)+11);
}
#define init_11_7_2_12
void FASTCALL call_11_7_2_12(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(2<<4)+12);
}
#define init_11_7_2_13
void FASTCALL call_11_7_2_13(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(2<<4)+13);
}
#define init_11_7_2_14
void FASTCALL call_11_7_2_14(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(2<<4)+14);
}
#define init_11_7_2_15
void FASTCALL call_11_7_2_15(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(2<<4)+15);
}
#define init_11_7_3_0
void FASTCALL call_11_7_3_0(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(3<<4)+0);
}
#define init_11_7_3_1
void FASTCALL call_11_7_3_1(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(3<<4)+1);
}
#define init_11_7_3_2
void FASTCALL call_11_7_3_2(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(3<<4)+2);
}
#define init_11_7_3_3
void FASTCALL call_11_7_3_3(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(3<<4)+3);
}
#define init_11_7_3_4
void FASTCALL call_11_7_3_4(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(3<<4)+4);
}
#define init_11_7_3_5
void FASTCALL call_11_7_3_5(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(3<<4)+5);
}
#define init_11_7_3_6
void FASTCALL call_11_7_3_6(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(3<<4)+6);
}
#define init_11_7_3_7
void FASTCALL call_11_7_3_7(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(3<<4)+7);
}
#define init_11_7_3_8
void FASTCALL call_11_7_3_8(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(3<<4)+8);
}
#define init_11_7_3_9
void FASTCALL call_11_7_3_9(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(3<<4)+9);
}
#define init_11_7_3_10
void FASTCALL call_11_7_3_10(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(3<<4)+10);
}
#define init_11_7_3_11
void FASTCALL call_11_7_3_11(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(3<<4)+11);
}
#define init_11_7_3_12
void FASTCALL call_11_7_3_12(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(3<<4)+12);
}
#define init_11_7_3_13
void FASTCALL call_11_7_3_13(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(3<<4)+13);
}
#define init_11_7_3_14
void FASTCALL call_11_7_3_14(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(3<<4)+14);
}
#define init_11_7_3_15
void FASTCALL call_11_7_3_15(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(3<<4)+15);
}
#define init_11_7_4_0
void FASTCALL call_11_7_4_0(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(4<<4)+0);
}
#define init_11_7_4_1
void FASTCALL call_11_7_4_1(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(4<<4)+1);
}
#define init_11_7_4_2
void FASTCALL call_11_7_4_2(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(4<<4)+2);
}
#define init_11_7_4_3
void FASTCALL call_11_7_4_3(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(4<<4)+3);
}
#define init_11_7_4_4
void FASTCALL call_11_7_4_4(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(4<<4)+4);
}
#define init_11_7_4_5
void FASTCALL call_11_7_4_5(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(4<<4)+5);
}
#define init_11_7_4_6
void FASTCALL call_11_7_4_6(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(4<<4)+6);
}
#define init_11_7_4_7
void FASTCALL call_11_7_4_7(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(4<<4)+7);
}
#define init_11_7_4_8
void FASTCALL call_11_7_4_8(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(4<<4)+8);
}
#define init_11_7_4_9
void FASTCALL call_11_7_4_9(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(4<<4)+9);
}
#define init_11_7_4_10
void FASTCALL call_11_7_4_10(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(4<<4)+10);
}
#define init_11_7_4_11
void FASTCALL call_11_7_4_11(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(4<<4)+11);
}
#define init_11_7_4_12
void FASTCALL call_11_7_4_12(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(4<<4)+12);
}
#define init_11_7_4_13
void FASTCALL call_11_7_4_13(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(4<<4)+13);
}
#define init_11_7_4_14
void FASTCALL call_11_7_4_14(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(4<<4)+14);
}
#define init_11_7_4_15
void FASTCALL call_11_7_4_15(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(4<<4)+15);
}
#define init_11_7_5_0
void FASTCALL call_11_7_5_0(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(5<<4)+0);
}
#define init_11_7_5_1
void FASTCALL call_11_7_5_1(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(5<<4)+1);
}
#define init_11_7_5_2
void FASTCALL call_11_7_5_2(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(5<<4)+2);
}
#define init_11_7_5_3
void FASTCALL call_11_7_5_3(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(5<<4)+3);
}
#define init_11_7_5_4
void FASTCALL call_11_7_5_4(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(5<<4)+4);
}
#define init_11_7_5_5
void FASTCALL call_11_7_5_5(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(5<<4)+5);
}
#define init_11_7_5_6
void FASTCALL call_11_7_5_6(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(5<<4)+6);
}
#define init_11_7_5_7
void FASTCALL call_11_7_5_7(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(5<<4)+7);
}
#define init_11_7_5_8
void FASTCALL call_11_7_5_8(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(5<<4)+8);
}
#define init_11_7_5_9
void FASTCALL call_11_7_5_9(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(5<<4)+9);
}
#define init_11_7_5_10
void FASTCALL call_11_7_5_10(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(5<<4)+10);
}
#define init_11_7_5_11
void FASTCALL call_11_7_5_11(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(5<<4)+11);
}
#define init_11_7_5_12
void FASTCALL call_11_7_5_12(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(5<<4)+12);
}
#define init_11_7_5_13
void FASTCALL call_11_7_5_13(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(5<<4)+13);
}
#define init_11_7_5_14
void FASTCALL call_11_7_5_14(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(5<<4)+14);
}
#define init_11_7_5_15
void FASTCALL call_11_7_5_15(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(5<<4)+15);
}
#define init_11_7_6_0
void FASTCALL call_11_7_6_0(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(6<<4)+0);
}
#define init_11_7_6_1
void FASTCALL call_11_7_6_1(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(6<<4)+1);
}
#define init_11_7_6_2
void FASTCALL call_11_7_6_2(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(6<<4)+2);
}
#define init_11_7_6_3
void FASTCALL call_11_7_6_3(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(6<<4)+3);
}
#define init_11_7_6_4
void FASTCALL call_11_7_6_4(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(6<<4)+4);
}
#define init_11_7_6_5
void FASTCALL call_11_7_6_5(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(6<<4)+5);
}
#define init_11_7_6_6
void FASTCALL call_11_7_6_6(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(6<<4)+6);
}
#define init_11_7_6_7
void FASTCALL call_11_7_6_7(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(6<<4)+7);
}
#define init_11_7_6_8
void FASTCALL call_11_7_6_8(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(6<<4)+8);
}
#define init_11_7_6_9
void FASTCALL call_11_7_6_9(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(6<<4)+9);
}
#define init_11_7_6_10
void FASTCALL call_11_7_6_10(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(6<<4)+10);
}
#define init_11_7_6_11
void FASTCALL call_11_7_6_11(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(6<<4)+11);
}
#define init_11_7_6_12
void FASTCALL call_11_7_6_12(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(6<<4)+12);
}
#define init_11_7_6_13
void FASTCALL call_11_7_6_13(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(6<<4)+13);
}
#define init_11_7_6_14
void FASTCALL call_11_7_6_14(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(6<<4)+14);
}
#define init_11_7_6_15
void FASTCALL call_11_7_6_15(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(6<<4)+15);
}
#define init_11_7_7_0
void FASTCALL call_11_7_7_0(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(7<<4)+0);
}
#define init_11_7_7_1
void FASTCALL call_11_7_7_1(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(7<<4)+1);
}
#define init_11_7_7_2
void FASTCALL call_11_7_7_2(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(7<<4)+2);
}
#define init_11_7_7_3
void FASTCALL call_11_7_7_3(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(7<<4)+3);
}
#define init_11_7_7_4
void FASTCALL call_11_7_7_4(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(7<<4)+4);
}
#define init_11_7_7_5
void FASTCALL call_11_7_7_5(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(7<<4)+5);
}
#define init_11_7_7_6
void FASTCALL call_11_7_7_6(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(7<<4)+6);
}
#define init_11_7_7_7
void FASTCALL call_11_7_7_7(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(7<<4)+7);
}
#define init_11_7_7_8
void FASTCALL call_11_7_7_8(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(7<<4)+8);
}
#define init_11_7_7_9
void FASTCALL call_11_7_7_9(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(7<<4)+9);
}
#define init_11_7_7_10
void FASTCALL call_11_7_7_10(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(7<<4)+10);
}
#define init_11_7_7_11
void FASTCALL call_11_7_7_11(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(7<<4)+11);
}
#define init_11_7_7_12
void FASTCALL call_11_7_7_12(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(7<<4)+12);
}
#define init_11_7_7_13
void FASTCALL call_11_7_7_13(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(7<<4)+13);
}
#define init_11_7_7_14
void FASTCALL call_11_7_7_14(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(7<<4)+14);
}
#define init_11_7_7_15
void FASTCALL call_11_7_7_15(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(7<<4)+15);
}
#define init_11_7_8_0
void FASTCALL call_11_7_8_0(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(8<<4)+0);
}
#define init_11_7_8_1
void FASTCALL call_11_7_8_1(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(8<<4)+1);
}
#define init_11_7_8_2
void FASTCALL call_11_7_8_2(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(8<<4)+2);
}
#define init_11_7_8_3
void FASTCALL call_11_7_8_3(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(8<<4)+3);
}
#define init_11_7_8_4
void FASTCALL call_11_7_8_4(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(8<<4)+4);
}
#define init_11_7_8_5
void FASTCALL call_11_7_8_5(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(8<<4)+5);
}
#define init_11_7_8_6
void FASTCALL call_11_7_8_6(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(8<<4)+6);
}
#define init_11_7_8_7
void FASTCALL call_11_7_8_7(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(8<<4)+7);
}
#define init_11_7_8_8
void FASTCALL call_11_7_8_8(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(8<<4)+8);
}
#define init_11_7_8_9
void FASTCALL call_11_7_8_9(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(8<<4)+9);
}
#define init_11_7_8_10
void FASTCALL call_11_7_8_10(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(8<<4)+10);
}
#define init_11_7_8_11
void FASTCALL call_11_7_8_11(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(8<<4)+11);
}
#define init_11_7_8_12
void FASTCALL call_11_7_8_12(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(8<<4)+12);
}
#define init_11_7_8_13
void FASTCALL call_11_7_8_13(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(8<<4)+13);
}
#define init_11_7_8_14
void FASTCALL call_11_7_8_14(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(8<<4)+14);
}
#define init_11_7_8_15
void FASTCALL call_11_7_8_15(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(8<<4)+15);
}
#define init_11_7_9_0
void FASTCALL call_11_7_9_0(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(9<<4)+0);
}
#define init_11_7_9_1
void FASTCALL call_11_7_9_1(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(9<<4)+1);
}
#define init_11_7_9_2
void FASTCALL call_11_7_9_2(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(9<<4)+2);
}
#define init_11_7_9_3
void FASTCALL call_11_7_9_3(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(9<<4)+3);
}
#define init_11_7_9_4
void FASTCALL call_11_7_9_4(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(9<<4)+4);
}
#define init_11_7_9_5
void FASTCALL call_11_7_9_5(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(9<<4)+5);
}
#define init_11_7_9_6
void FASTCALL call_11_7_9_6(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(9<<4)+6);
}
#define init_11_7_9_7
void FASTCALL call_11_7_9_7(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(9<<4)+7);
}
#define init_11_7_9_8
void FASTCALL call_11_7_9_8(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(9<<4)+8);
}
#define init_11_7_9_9
void FASTCALL call_11_7_9_9(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(9<<4)+9);
}
#define init_11_7_9_10
void FASTCALL call_11_7_9_10(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(9<<4)+10);
}
#define init_11_7_9_11
void FASTCALL call_11_7_9_11(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(9<<4)+11);
}
#define init_11_7_9_12
void FASTCALL call_11_7_9_12(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(9<<4)+12);
}
#define init_11_7_9_13
void FASTCALL call_11_7_9_13(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(9<<4)+13);
}
#define init_11_7_9_14
void FASTCALL call_11_7_9_14(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(9<<4)+14);
}
#define init_11_7_9_15
void FASTCALL call_11_7_9_15(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(9<<4)+15);
}
#define init_11_7_10_0
void FASTCALL call_11_7_10_0(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(10<<4)+0);
}
#define init_11_7_10_1
void FASTCALL call_11_7_10_1(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(10<<4)+1);
}
#define init_11_7_10_2
void FASTCALL call_11_7_10_2(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(10<<4)+2);
}
#define init_11_7_10_3
void FASTCALL call_11_7_10_3(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(10<<4)+3);
}
#define init_11_7_10_4
void FASTCALL call_11_7_10_4(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(10<<4)+4);
}
#define init_11_7_10_5
void FASTCALL call_11_7_10_5(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(10<<4)+5);
}
#define init_11_7_10_6
void FASTCALL call_11_7_10_6(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(10<<4)+6);
}
#define init_11_7_10_7
void FASTCALL call_11_7_10_7(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(10<<4)+7);
}
#define init_11_7_10_8
void FASTCALL call_11_7_10_8(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(10<<4)+8);
}
#define init_11_7_10_9
void FASTCALL call_11_7_10_9(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(10<<4)+9);
}
#define init_11_7_10_10
void FASTCALL call_11_7_10_10(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(10<<4)+10);
}
#define init_11_7_10_11
void FASTCALL call_11_7_10_11(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(10<<4)+11);
}
#define init_11_7_10_12
void FASTCALL call_11_7_10_12(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(10<<4)+12);
}
#define init_11_7_10_13
void FASTCALL call_11_7_10_13(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(10<<4)+13);
}
#define init_11_7_10_14
void FASTCALL call_11_7_10_14(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(10<<4)+14);
}
#define init_11_7_10_15
void FASTCALL call_11_7_10_15(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(10<<4)+15);
}
#define init_11_7_11_0
void FASTCALL call_11_7_11_0(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(11<<4)+0);
}
#define init_11_7_11_1
void FASTCALL call_11_7_11_1(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(11<<4)+1);
}
#define init_11_7_11_2
void FASTCALL call_11_7_11_2(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(11<<4)+2);
}
#define init_11_7_11_3
void FASTCALL call_11_7_11_3(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(11<<4)+3);
}
#define init_11_7_11_4
void FASTCALL call_11_7_11_4(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(11<<4)+4);
}
#define init_11_7_11_5
void FASTCALL call_11_7_11_5(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(11<<4)+5);
}
#define init_11_7_11_6
void FASTCALL call_11_7_11_6(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(11<<4)+6);
}
#define init_11_7_11_7
void FASTCALL call_11_7_11_7(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(11<<4)+7);
}
#define init_11_7_11_8
void FASTCALL call_11_7_11_8(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(11<<4)+8);
}
#define init_11_7_11_9
void FASTCALL call_11_7_11_9(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(11<<4)+9);
}
#define init_11_7_11_10
void FASTCALL call_11_7_11_10(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(11<<4)+10);
}
#define init_11_7_11_11
void FASTCALL call_11_7_11_11(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(11<<4)+11);
}
#define init_11_7_11_12
void FASTCALL call_11_7_11_12(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(11<<4)+12);
}
#define init_11_7_11_13
void FASTCALL call_11_7_11_13(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(11<<4)+13);
}
#define init_11_7_11_14
void FASTCALL call_11_7_11_14(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(11<<4)+14);
}
#define init_11_7_11_15
void FASTCALL call_11_7_11_15(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(11<<4)+15);
}
#define init_11_7_12_0
void FASTCALL call_11_7_12_0(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(12<<4)+0);
}
#define init_11_7_12_1
void FASTCALL call_11_7_12_1(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(12<<4)+1);
}
#define init_11_7_12_2
void FASTCALL call_11_7_12_2(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(12<<4)+2);
}
#define init_11_7_12_3
void FASTCALL call_11_7_12_3(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(12<<4)+3);
}
#define init_11_7_12_4
void FASTCALL call_11_7_12_4(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(12<<4)+4);
}
#define init_11_7_12_5
void FASTCALL call_11_7_12_5(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(12<<4)+5);
}
#define init_11_7_12_6
void FASTCALL call_11_7_12_6(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(12<<4)+6);
}
#define init_11_7_12_7
void FASTCALL call_11_7_12_7(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(12<<4)+7);
}
#define init_11_7_12_8
void FASTCALL call_11_7_12_8(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(12<<4)+8);
}
#define init_11_7_12_9
void FASTCALL call_11_7_12_9(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(12<<4)+9);
}
#define init_11_7_12_10
void FASTCALL call_11_7_12_10(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(12<<4)+10);
}
#define init_11_7_12_11
void FASTCALL call_11_7_12_11(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(12<<4)+11);
}
#define init_11_7_12_12
void FASTCALL call_11_7_12_12(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(12<<4)+12);
}
#define init_11_7_12_13
void FASTCALL call_11_7_12_13(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(12<<4)+13);
}
#define init_11_7_12_14
void FASTCALL call_11_7_12_14(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(12<<4)+14);
}
#define init_11_7_12_15
void FASTCALL call_11_7_12_15(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(12<<4)+15);
}
#define init_11_7_13_0
void FASTCALL call_11_7_13_0(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(13<<4)+0);
}
#define init_11_7_13_1
void FASTCALL call_11_7_13_1(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(13<<4)+1);
}
#define init_11_7_13_2
void FASTCALL call_11_7_13_2(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(13<<4)+2);
}
#define init_11_7_13_3
void FASTCALL call_11_7_13_3(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(13<<4)+3);
}
#define init_11_7_13_4
void FASTCALL call_11_7_13_4(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(13<<4)+4);
}
#define init_11_7_13_5
void FASTCALL call_11_7_13_5(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(13<<4)+5);
}
#define init_11_7_13_6
void FASTCALL call_11_7_13_6(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(13<<4)+6);
}
#define init_11_7_13_7
void FASTCALL call_11_7_13_7(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(13<<4)+7);
}
#define init_11_7_13_8
void FASTCALL call_11_7_13_8(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(13<<4)+8);
}
#define init_11_7_13_9
void FASTCALL call_11_7_13_9(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(13<<4)+9);
}
#define init_11_7_13_10
void FASTCALL call_11_7_13_10(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(13<<4)+10);
}
#define init_11_7_13_11
void FASTCALL call_11_7_13_11(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(13<<4)+11);
}
#define init_11_7_13_12
void FASTCALL call_11_7_13_12(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(13<<4)+12);
}
#define init_11_7_13_13
void FASTCALL call_11_7_13_13(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(13<<4)+13);
}
#define init_11_7_13_14
void FASTCALL call_11_7_13_14(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(13<<4)+14);
}
#define init_11_7_13_15
void FASTCALL call_11_7_13_15(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(13<<4)+15);
}
#define init_11_7_14_0
void FASTCALL call_11_7_14_0(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(14<<4)+0);
}
#define init_11_7_14_1
void FASTCALL call_11_7_14_1(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(14<<4)+1);
}
#define init_11_7_14_2
void FASTCALL call_11_7_14_2(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(14<<4)+2);
}
#define init_11_7_14_3
void FASTCALL call_11_7_14_3(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(14<<4)+3);
}
#define init_11_7_14_4
void FASTCALL call_11_7_14_4(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(14<<4)+4);
}
#define init_11_7_14_5
void FASTCALL call_11_7_14_5(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(14<<4)+5);
}
#define init_11_7_14_6
void FASTCALL call_11_7_14_6(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(14<<4)+6);
}
#define init_11_7_14_7
void FASTCALL call_11_7_14_7(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(14<<4)+7);
}
#define init_11_7_14_8
void FASTCALL call_11_7_14_8(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(14<<4)+8);
}
#define init_11_7_14_9
void FASTCALL call_11_7_14_9(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(14<<4)+9);
}
#define init_11_7_14_10
void FASTCALL call_11_7_14_10(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(14<<4)+10);
}
#define init_11_7_14_11
void FASTCALL call_11_7_14_11(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(14<<4)+11);
}
#define init_11_7_14_12
void FASTCALL call_11_7_14_12(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(14<<4)+12);
}
#define init_11_7_14_13
void FASTCALL call_11_7_14_13(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(14<<4)+13);
}
#define init_11_7_14_14
void FASTCALL call_11_7_14_14(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(14<<4)+14);
}
#define init_11_7_14_15
void FASTCALL call_11_7_14_15(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(14<<4)+15);
}
#define init_11_7_15_0
void FASTCALL call_11_7_15_0(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(15<<4)+0);
}
#define init_11_7_15_1
void FASTCALL call_11_7_15_1(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(15<<4)+1);
}
#define init_11_7_15_2
void FASTCALL call_11_7_15_2(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(15<<4)+2);
}
#define init_11_7_15_3
void FASTCALL call_11_7_15_3(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(15<<4)+3);
}
#define init_11_7_15_4
void FASTCALL call_11_7_15_4(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(15<<4)+4);
}
#define init_11_7_15_5
void FASTCALL call_11_7_15_5(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(15<<4)+5);
}
#define init_11_7_15_6
void FASTCALL call_11_7_15_6(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(15<<4)+6);
}
#define init_11_7_15_7
void FASTCALL call_11_7_15_7(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(15<<4)+7);
}
#define init_11_7_15_8
void FASTCALL call_11_7_15_8(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(15<<4)+8);
}
#define init_11_7_15_9
void FASTCALL call_11_7_15_9(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(15<<4)+9);
}
#define init_11_7_15_10
void FASTCALL call_11_7_15_10(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(15<<4)+10);
}
#define init_11_7_15_11
void FASTCALL call_11_7_15_11(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(15<<4)+11);
}
#define init_11_7_15_12
void FASTCALL call_11_7_15_12(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(15<<4)+12);
}
#define init_11_7_15_13
void FASTCALL call_11_7_15_13(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(15<<4)+13);
}
#define init_11_7_15_14
void FASTCALL call_11_7_15_14(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(15<<4)+14);
}
#define init_11_7_15_15
void FASTCALL call_11_7_15_15(SH2_struct* sh)
{
  SH2bsr(sh, (7<<8)+(15<<4)+15);
}
#define init_11_8_0_0
void FASTCALL call_11_8_0_0(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(0<<4)+0);
}
#define init_11_8_0_1
void FASTCALL call_11_8_0_1(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(0<<4)+1);
}
#define init_11_8_0_2
void FASTCALL call_11_8_0_2(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(0<<4)+2);
}
#define init_11_8_0_3
void FASTCALL call_11_8_0_3(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(0<<4)+3);
}
#define init_11_8_0_4
void FASTCALL call_11_8_0_4(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(0<<4)+4);
}
#define init_11_8_0_5
void FASTCALL call_11_8_0_5(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(0<<4)+5);
}
#define init_11_8_0_6
void FASTCALL call_11_8_0_6(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(0<<4)+6);
}
#define init_11_8_0_7
void FASTCALL call_11_8_0_7(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(0<<4)+7);
}
#define init_11_8_0_8
void FASTCALL call_11_8_0_8(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(0<<4)+8);
}
#define init_11_8_0_9
void FASTCALL call_11_8_0_9(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(0<<4)+9);
}
#define init_11_8_0_10
void FASTCALL call_11_8_0_10(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(0<<4)+10);
}
#define init_11_8_0_11
void FASTCALL call_11_8_0_11(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(0<<4)+11);
}
#define init_11_8_0_12
void FASTCALL call_11_8_0_12(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(0<<4)+12);
}
#define init_11_8_0_13
void FASTCALL call_11_8_0_13(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(0<<4)+13);
}
#define init_11_8_0_14
void FASTCALL call_11_8_0_14(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(0<<4)+14);
}
#define init_11_8_0_15
void FASTCALL call_11_8_0_15(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(0<<4)+15);
}
#define init_11_8_1_0
void FASTCALL call_11_8_1_0(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(1<<4)+0);
}
#define init_11_8_1_1
void FASTCALL call_11_8_1_1(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(1<<4)+1);
}
#define init_11_8_1_2
void FASTCALL call_11_8_1_2(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(1<<4)+2);
}
#define init_11_8_1_3
void FASTCALL call_11_8_1_3(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(1<<4)+3);
}
#define init_11_8_1_4
void FASTCALL call_11_8_1_4(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(1<<4)+4);
}
#define init_11_8_1_5
void FASTCALL call_11_8_1_5(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(1<<4)+5);
}
#define init_11_8_1_6
void FASTCALL call_11_8_1_6(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(1<<4)+6);
}
#define init_11_8_1_7
void FASTCALL call_11_8_1_7(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(1<<4)+7);
}
#define init_11_8_1_8
void FASTCALL call_11_8_1_8(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(1<<4)+8);
}
#define init_11_8_1_9
void FASTCALL call_11_8_1_9(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(1<<4)+9);
}
#define init_11_8_1_10
void FASTCALL call_11_8_1_10(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(1<<4)+10);
}
#define init_11_8_1_11
void FASTCALL call_11_8_1_11(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(1<<4)+11);
}
#define init_11_8_1_12
void FASTCALL call_11_8_1_12(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(1<<4)+12);
}
#define init_11_8_1_13
void FASTCALL call_11_8_1_13(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(1<<4)+13);
}
#define init_11_8_1_14
void FASTCALL call_11_8_1_14(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(1<<4)+14);
}
#define init_11_8_1_15
void FASTCALL call_11_8_1_15(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(1<<4)+15);
}
#define init_11_8_2_0
void FASTCALL call_11_8_2_0(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(2<<4)+0);
}
#define init_11_8_2_1
void FASTCALL call_11_8_2_1(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(2<<4)+1);
}
#define init_11_8_2_2
void FASTCALL call_11_8_2_2(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(2<<4)+2);
}
#define init_11_8_2_3
void FASTCALL call_11_8_2_3(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(2<<4)+3);
}
#define init_11_8_2_4
void FASTCALL call_11_8_2_4(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(2<<4)+4);
}
#define init_11_8_2_5
void FASTCALL call_11_8_2_5(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(2<<4)+5);
}
#define init_11_8_2_6
void FASTCALL call_11_8_2_6(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(2<<4)+6);
}
#define init_11_8_2_7
void FASTCALL call_11_8_2_7(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(2<<4)+7);
}
#define init_11_8_2_8
void FASTCALL call_11_8_2_8(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(2<<4)+8);
}
#define init_11_8_2_9
void FASTCALL call_11_8_2_9(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(2<<4)+9);
}
#define init_11_8_2_10
void FASTCALL call_11_8_2_10(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(2<<4)+10);
}
#define init_11_8_2_11
void FASTCALL call_11_8_2_11(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(2<<4)+11);
}
#define init_11_8_2_12
void FASTCALL call_11_8_2_12(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(2<<4)+12);
}
#define init_11_8_2_13
void FASTCALL call_11_8_2_13(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(2<<4)+13);
}
#define init_11_8_2_14
void FASTCALL call_11_8_2_14(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(2<<4)+14);
}
#define init_11_8_2_15
void FASTCALL call_11_8_2_15(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(2<<4)+15);
}
#define init_11_8_3_0
void FASTCALL call_11_8_3_0(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(3<<4)+0);
}
#define init_11_8_3_1
void FASTCALL call_11_8_3_1(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(3<<4)+1);
}
#define init_11_8_3_2
void FASTCALL call_11_8_3_2(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(3<<4)+2);
}
#define init_11_8_3_3
void FASTCALL call_11_8_3_3(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(3<<4)+3);
}
#define init_11_8_3_4
void FASTCALL call_11_8_3_4(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(3<<4)+4);
}
#define init_11_8_3_5
void FASTCALL call_11_8_3_5(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(3<<4)+5);
}
#define init_11_8_3_6
void FASTCALL call_11_8_3_6(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(3<<4)+6);
}
#define init_11_8_3_7
void FASTCALL call_11_8_3_7(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(3<<4)+7);
}
#define init_11_8_3_8
void FASTCALL call_11_8_3_8(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(3<<4)+8);
}
#define init_11_8_3_9
void FASTCALL call_11_8_3_9(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(3<<4)+9);
}
#define init_11_8_3_10
void FASTCALL call_11_8_3_10(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(3<<4)+10);
}
#define init_11_8_3_11
void FASTCALL call_11_8_3_11(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(3<<4)+11);
}
#define init_11_8_3_12
void FASTCALL call_11_8_3_12(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(3<<4)+12);
}
#define init_11_8_3_13
void FASTCALL call_11_8_3_13(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(3<<4)+13);
}
#define init_11_8_3_14
void FASTCALL call_11_8_3_14(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(3<<4)+14);
}
#define init_11_8_3_15
void FASTCALL call_11_8_3_15(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(3<<4)+15);
}
#define init_11_8_4_0
void FASTCALL call_11_8_4_0(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(4<<4)+0);
}
#define init_11_8_4_1
void FASTCALL call_11_8_4_1(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(4<<4)+1);
}
#define init_11_8_4_2
void FASTCALL call_11_8_4_2(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(4<<4)+2);
}
#define init_11_8_4_3
void FASTCALL call_11_8_4_3(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(4<<4)+3);
}
#define init_11_8_4_4
void FASTCALL call_11_8_4_4(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(4<<4)+4);
}
#define init_11_8_4_5
void FASTCALL call_11_8_4_5(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(4<<4)+5);
}
#define init_11_8_4_6
void FASTCALL call_11_8_4_6(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(4<<4)+6);
}
#define init_11_8_4_7
void FASTCALL call_11_8_4_7(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(4<<4)+7);
}
#define init_11_8_4_8
void FASTCALL call_11_8_4_8(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(4<<4)+8);
}
#define init_11_8_4_9
void FASTCALL call_11_8_4_9(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(4<<4)+9);
}
#define init_11_8_4_10
void FASTCALL call_11_8_4_10(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(4<<4)+10);
}
#define init_11_8_4_11
void FASTCALL call_11_8_4_11(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(4<<4)+11);
}
#define init_11_8_4_12
void FASTCALL call_11_8_4_12(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(4<<4)+12);
}
#define init_11_8_4_13
void FASTCALL call_11_8_4_13(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(4<<4)+13);
}
#define init_11_8_4_14
void FASTCALL call_11_8_4_14(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(4<<4)+14);
}
#define init_11_8_4_15
void FASTCALL call_11_8_4_15(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(4<<4)+15);
}
#define init_11_8_5_0
void FASTCALL call_11_8_5_0(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(5<<4)+0);
}
#define init_11_8_5_1
void FASTCALL call_11_8_5_1(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(5<<4)+1);
}
#define init_11_8_5_2
void FASTCALL call_11_8_5_2(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(5<<4)+2);
}
#define init_11_8_5_3
void FASTCALL call_11_8_5_3(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(5<<4)+3);
}
#define init_11_8_5_4
void FASTCALL call_11_8_5_4(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(5<<4)+4);
}
#define init_11_8_5_5
void FASTCALL call_11_8_5_5(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(5<<4)+5);
}
#define init_11_8_5_6
void FASTCALL call_11_8_5_6(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(5<<4)+6);
}
#define init_11_8_5_7
void FASTCALL call_11_8_5_7(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(5<<4)+7);
}
#define init_11_8_5_8
void FASTCALL call_11_8_5_8(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(5<<4)+8);
}
#define init_11_8_5_9
void FASTCALL call_11_8_5_9(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(5<<4)+9);
}
#define init_11_8_5_10
void FASTCALL call_11_8_5_10(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(5<<4)+10);
}
#define init_11_8_5_11
void FASTCALL call_11_8_5_11(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(5<<4)+11);
}
#define init_11_8_5_12
void FASTCALL call_11_8_5_12(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(5<<4)+12);
}
#define init_11_8_5_13
void FASTCALL call_11_8_5_13(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(5<<4)+13);
}
#define init_11_8_5_14
void FASTCALL call_11_8_5_14(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(5<<4)+14);
}
#define init_11_8_5_15
void FASTCALL call_11_8_5_15(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(5<<4)+15);
}
#define init_11_8_6_0
void FASTCALL call_11_8_6_0(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(6<<4)+0);
}
#define init_11_8_6_1
void FASTCALL call_11_8_6_1(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(6<<4)+1);
}
#define init_11_8_6_2
void FASTCALL call_11_8_6_2(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(6<<4)+2);
}
#define init_11_8_6_3
void FASTCALL call_11_8_6_3(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(6<<4)+3);
}
#define init_11_8_6_4
void FASTCALL call_11_8_6_4(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(6<<4)+4);
}
#define init_11_8_6_5
void FASTCALL call_11_8_6_5(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(6<<4)+5);
}
#define init_11_8_6_6
void FASTCALL call_11_8_6_6(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(6<<4)+6);
}
#define init_11_8_6_7
void FASTCALL call_11_8_6_7(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(6<<4)+7);
}
#define init_11_8_6_8
void FASTCALL call_11_8_6_8(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(6<<4)+8);
}
#define init_11_8_6_9
void FASTCALL call_11_8_6_9(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(6<<4)+9);
}
#define init_11_8_6_10
void FASTCALL call_11_8_6_10(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(6<<4)+10);
}
#define init_11_8_6_11
void FASTCALL call_11_8_6_11(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(6<<4)+11);
}
#define init_11_8_6_12
void FASTCALL call_11_8_6_12(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(6<<4)+12);
}
#define init_11_8_6_13
void FASTCALL call_11_8_6_13(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(6<<4)+13);
}
#define init_11_8_6_14
void FASTCALL call_11_8_6_14(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(6<<4)+14);
}
#define init_11_8_6_15
void FASTCALL call_11_8_6_15(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(6<<4)+15);
}
#define init_11_8_7_0
void FASTCALL call_11_8_7_0(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(7<<4)+0);
}
#define init_11_8_7_1
void FASTCALL call_11_8_7_1(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(7<<4)+1);
}
#define init_11_8_7_2
void FASTCALL call_11_8_7_2(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(7<<4)+2);
}
#define init_11_8_7_3
void FASTCALL call_11_8_7_3(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(7<<4)+3);
}
#define init_11_8_7_4
void FASTCALL call_11_8_7_4(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(7<<4)+4);
}
#define init_11_8_7_5
void FASTCALL call_11_8_7_5(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(7<<4)+5);
}
#define init_11_8_7_6
void FASTCALL call_11_8_7_6(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(7<<4)+6);
}
#define init_11_8_7_7
void FASTCALL call_11_8_7_7(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(7<<4)+7);
}
#define init_11_8_7_8
void FASTCALL call_11_8_7_8(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(7<<4)+8);
}
#define init_11_8_7_9
void FASTCALL call_11_8_7_9(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(7<<4)+9);
}
#define init_11_8_7_10
void FASTCALL call_11_8_7_10(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(7<<4)+10);
}
#define init_11_8_7_11
void FASTCALL call_11_8_7_11(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(7<<4)+11);
}
#define init_11_8_7_12
void FASTCALL call_11_8_7_12(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(7<<4)+12);
}
#define init_11_8_7_13
void FASTCALL call_11_8_7_13(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(7<<4)+13);
}
#define init_11_8_7_14
void FASTCALL call_11_8_7_14(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(7<<4)+14);
}
#define init_11_8_7_15
void FASTCALL call_11_8_7_15(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(7<<4)+15);
}
#define init_11_8_8_0
void FASTCALL call_11_8_8_0(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(8<<4)+0);
}
#define init_11_8_8_1
void FASTCALL call_11_8_8_1(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(8<<4)+1);
}
#define init_11_8_8_2
void FASTCALL call_11_8_8_2(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(8<<4)+2);
}
#define init_11_8_8_3
void FASTCALL call_11_8_8_3(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(8<<4)+3);
}
#define init_11_8_8_4
void FASTCALL call_11_8_8_4(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(8<<4)+4);
}
#define init_11_8_8_5
void FASTCALL call_11_8_8_5(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(8<<4)+5);
}
#define init_11_8_8_6
void FASTCALL call_11_8_8_6(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(8<<4)+6);
}
#define init_11_8_8_7
void FASTCALL call_11_8_8_7(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(8<<4)+7);
}
#define init_11_8_8_8
void FASTCALL call_11_8_8_8(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(8<<4)+8);
}
#define init_11_8_8_9
void FASTCALL call_11_8_8_9(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(8<<4)+9);
}
#define init_11_8_8_10
void FASTCALL call_11_8_8_10(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(8<<4)+10);
}
#define init_11_8_8_11
void FASTCALL call_11_8_8_11(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(8<<4)+11);
}
#define init_11_8_8_12
void FASTCALL call_11_8_8_12(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(8<<4)+12);
}
#define init_11_8_8_13
void FASTCALL call_11_8_8_13(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(8<<4)+13);
}
#define init_11_8_8_14
void FASTCALL call_11_8_8_14(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(8<<4)+14);
}
#define init_11_8_8_15
void FASTCALL call_11_8_8_15(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(8<<4)+15);
}
#define init_11_8_9_0
void FASTCALL call_11_8_9_0(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(9<<4)+0);
}
#define init_11_8_9_1
void FASTCALL call_11_8_9_1(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(9<<4)+1);
}
#define init_11_8_9_2
void FASTCALL call_11_8_9_2(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(9<<4)+2);
}
#define init_11_8_9_3
void FASTCALL call_11_8_9_3(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(9<<4)+3);
}
#define init_11_8_9_4
void FASTCALL call_11_8_9_4(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(9<<4)+4);
}
#define init_11_8_9_5
void FASTCALL call_11_8_9_5(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(9<<4)+5);
}
#define init_11_8_9_6
void FASTCALL call_11_8_9_6(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(9<<4)+6);
}
#define init_11_8_9_7
void FASTCALL call_11_8_9_7(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(9<<4)+7);
}
#define init_11_8_9_8
void FASTCALL call_11_8_9_8(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(9<<4)+8);
}
#define init_11_8_9_9
void FASTCALL call_11_8_9_9(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(9<<4)+9);
}
#define init_11_8_9_10
void FASTCALL call_11_8_9_10(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(9<<4)+10);
}
#define init_11_8_9_11
void FASTCALL call_11_8_9_11(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(9<<4)+11);
}
#define init_11_8_9_12
void FASTCALL call_11_8_9_12(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(9<<4)+12);
}
#define init_11_8_9_13
void FASTCALL call_11_8_9_13(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(9<<4)+13);
}
#define init_11_8_9_14
void FASTCALL call_11_8_9_14(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(9<<4)+14);
}
#define init_11_8_9_15
void FASTCALL call_11_8_9_15(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(9<<4)+15);
}
#define init_11_8_10_0
void FASTCALL call_11_8_10_0(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(10<<4)+0);
}
#define init_11_8_10_1
void FASTCALL call_11_8_10_1(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(10<<4)+1);
}
#define init_11_8_10_2
void FASTCALL call_11_8_10_2(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(10<<4)+2);
}
#define init_11_8_10_3
void FASTCALL call_11_8_10_3(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(10<<4)+3);
}
#define init_11_8_10_4
void FASTCALL call_11_8_10_4(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(10<<4)+4);
}
#define init_11_8_10_5
void FASTCALL call_11_8_10_5(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(10<<4)+5);
}
#define init_11_8_10_6
void FASTCALL call_11_8_10_6(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(10<<4)+6);
}
#define init_11_8_10_7
void FASTCALL call_11_8_10_7(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(10<<4)+7);
}
#define init_11_8_10_8
void FASTCALL call_11_8_10_8(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(10<<4)+8);
}
#define init_11_8_10_9
void FASTCALL call_11_8_10_9(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(10<<4)+9);
}
#define init_11_8_10_10
void FASTCALL call_11_8_10_10(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(10<<4)+10);
}
#define init_11_8_10_11
void FASTCALL call_11_8_10_11(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(10<<4)+11);
}
#define init_11_8_10_12
void FASTCALL call_11_8_10_12(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(10<<4)+12);
}
#define init_11_8_10_13
void FASTCALL call_11_8_10_13(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(10<<4)+13);
}
#define init_11_8_10_14
void FASTCALL call_11_8_10_14(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(10<<4)+14);
}
#define init_11_8_10_15
void FASTCALL call_11_8_10_15(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(10<<4)+15);
}
#define init_11_8_11_0
void FASTCALL call_11_8_11_0(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(11<<4)+0);
}
#define init_11_8_11_1
void FASTCALL call_11_8_11_1(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(11<<4)+1);
}
#define init_11_8_11_2
void FASTCALL call_11_8_11_2(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(11<<4)+2);
}
#define init_11_8_11_3
void FASTCALL call_11_8_11_3(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(11<<4)+3);
}
#define init_11_8_11_4
void FASTCALL call_11_8_11_4(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(11<<4)+4);
}
#define init_11_8_11_5
void FASTCALL call_11_8_11_5(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(11<<4)+5);
}
#define init_11_8_11_6
void FASTCALL call_11_8_11_6(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(11<<4)+6);
}
#define init_11_8_11_7
void FASTCALL call_11_8_11_7(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(11<<4)+7);
}
#define init_11_8_11_8
void FASTCALL call_11_8_11_8(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(11<<4)+8);
}
#define init_11_8_11_9
void FASTCALL call_11_8_11_9(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(11<<4)+9);
}
#define init_11_8_11_10
void FASTCALL call_11_8_11_10(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(11<<4)+10);
}
#define init_11_8_11_11
void FASTCALL call_11_8_11_11(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(11<<4)+11);
}
#define init_11_8_11_12
void FASTCALL call_11_8_11_12(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(11<<4)+12);
}
#define init_11_8_11_13
void FASTCALL call_11_8_11_13(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(11<<4)+13);
}
#define init_11_8_11_14
void FASTCALL call_11_8_11_14(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(11<<4)+14);
}
#define init_11_8_11_15
void FASTCALL call_11_8_11_15(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(11<<4)+15);
}
#define init_11_8_12_0
void FASTCALL call_11_8_12_0(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(12<<4)+0);
}
#define init_11_8_12_1
void FASTCALL call_11_8_12_1(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(12<<4)+1);
}
#define init_11_8_12_2
void FASTCALL call_11_8_12_2(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(12<<4)+2);
}
#define init_11_8_12_3
void FASTCALL call_11_8_12_3(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(12<<4)+3);
}
#define init_11_8_12_4
void FASTCALL call_11_8_12_4(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(12<<4)+4);
}
#define init_11_8_12_5
void FASTCALL call_11_8_12_5(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(12<<4)+5);
}
#define init_11_8_12_6
void FASTCALL call_11_8_12_6(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(12<<4)+6);
}
#define init_11_8_12_7
void FASTCALL call_11_8_12_7(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(12<<4)+7);
}
#define init_11_8_12_8
void FASTCALL call_11_8_12_8(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(12<<4)+8);
}
#define init_11_8_12_9
void FASTCALL call_11_8_12_9(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(12<<4)+9);
}
#define init_11_8_12_10
void FASTCALL call_11_8_12_10(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(12<<4)+10);
}
#define init_11_8_12_11
void FASTCALL call_11_8_12_11(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(12<<4)+11);
}
#define init_11_8_12_12
void FASTCALL call_11_8_12_12(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(12<<4)+12);
}
#define init_11_8_12_13
void FASTCALL call_11_8_12_13(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(12<<4)+13);
}
#define init_11_8_12_14
void FASTCALL call_11_8_12_14(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(12<<4)+14);
}
#define init_11_8_12_15
void FASTCALL call_11_8_12_15(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(12<<4)+15);
}
#define init_11_8_13_0
void FASTCALL call_11_8_13_0(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(13<<4)+0);
}
#define init_11_8_13_1
void FASTCALL call_11_8_13_1(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(13<<4)+1);
}
#define init_11_8_13_2
void FASTCALL call_11_8_13_2(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(13<<4)+2);
}
#define init_11_8_13_3
void FASTCALL call_11_8_13_3(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(13<<4)+3);
}
#define init_11_8_13_4
void FASTCALL call_11_8_13_4(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(13<<4)+4);
}
#define init_11_8_13_5
void FASTCALL call_11_8_13_5(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(13<<4)+5);
}
#define init_11_8_13_6
void FASTCALL call_11_8_13_6(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(13<<4)+6);
}
#define init_11_8_13_7
void FASTCALL call_11_8_13_7(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(13<<4)+7);
}
#define init_11_8_13_8
void FASTCALL call_11_8_13_8(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(13<<4)+8);
}
#define init_11_8_13_9
void FASTCALL call_11_8_13_9(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(13<<4)+9);
}
#define init_11_8_13_10
void FASTCALL call_11_8_13_10(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(13<<4)+10);
}
#define init_11_8_13_11
void FASTCALL call_11_8_13_11(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(13<<4)+11);
}
#define init_11_8_13_12
void FASTCALL call_11_8_13_12(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(13<<4)+12);
}
#define init_11_8_13_13
void FASTCALL call_11_8_13_13(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(13<<4)+13);
}
#define init_11_8_13_14
void FASTCALL call_11_8_13_14(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(13<<4)+14);
}
#define init_11_8_13_15
void FASTCALL call_11_8_13_15(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(13<<4)+15);
}
#define init_11_8_14_0
void FASTCALL call_11_8_14_0(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(14<<4)+0);
}
#define init_11_8_14_1
void FASTCALL call_11_8_14_1(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(14<<4)+1);
}
#define init_11_8_14_2
void FASTCALL call_11_8_14_2(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(14<<4)+2);
}
#define init_11_8_14_3
void FASTCALL call_11_8_14_3(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(14<<4)+3);
}
#define init_11_8_14_4
void FASTCALL call_11_8_14_4(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(14<<4)+4);
}
#define init_11_8_14_5
void FASTCALL call_11_8_14_5(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(14<<4)+5);
}
#define init_11_8_14_6
void FASTCALL call_11_8_14_6(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(14<<4)+6);
}
#define init_11_8_14_7
void FASTCALL call_11_8_14_7(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(14<<4)+7);
}
#define init_11_8_14_8
void FASTCALL call_11_8_14_8(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(14<<4)+8);
}
#define init_11_8_14_9
void FASTCALL call_11_8_14_9(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(14<<4)+9);
}
#define init_11_8_14_10
void FASTCALL call_11_8_14_10(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(14<<4)+10);
}
#define init_11_8_14_11
void FASTCALL call_11_8_14_11(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(14<<4)+11);
}
#define init_11_8_14_12
void FASTCALL call_11_8_14_12(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(14<<4)+12);
}
#define init_11_8_14_13
void FASTCALL call_11_8_14_13(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(14<<4)+13);
}
#define init_11_8_14_14
void FASTCALL call_11_8_14_14(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(14<<4)+14);
}
#define init_11_8_14_15
void FASTCALL call_11_8_14_15(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(14<<4)+15);
}
#define init_11_8_15_0
void FASTCALL call_11_8_15_0(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(15<<4)+0);
}
#define init_11_8_15_1
void FASTCALL call_11_8_15_1(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(15<<4)+1);
}
#define init_11_8_15_2
void FASTCALL call_11_8_15_2(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(15<<4)+2);
}
#define init_11_8_15_3
void FASTCALL call_11_8_15_3(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(15<<4)+3);
}
#define init_11_8_15_4
void FASTCALL call_11_8_15_4(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(15<<4)+4);
}
#define init_11_8_15_5
void FASTCALL call_11_8_15_5(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(15<<4)+5);
}
#define init_11_8_15_6
void FASTCALL call_11_8_15_6(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(15<<4)+6);
}
#define init_11_8_15_7
void FASTCALL call_11_8_15_7(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(15<<4)+7);
}
#define init_11_8_15_8
void FASTCALL call_11_8_15_8(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(15<<4)+8);
}
#define init_11_8_15_9
void FASTCALL call_11_8_15_9(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(15<<4)+9);
}
#define init_11_8_15_10
void FASTCALL call_11_8_15_10(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(15<<4)+10);
}
#define init_11_8_15_11
void FASTCALL call_11_8_15_11(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(15<<4)+11);
}
#define init_11_8_15_12
void FASTCALL call_11_8_15_12(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(15<<4)+12);
}
#define init_11_8_15_13
void FASTCALL call_11_8_15_13(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(15<<4)+13);
}
#define init_11_8_15_14
void FASTCALL call_11_8_15_14(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(15<<4)+14);
}
#define init_11_8_15_15
void FASTCALL call_11_8_15_15(SH2_struct* sh)
{
  SH2bsr(sh, (8<<8)+(15<<4)+15);
}
#define init_11_9_0_0
void FASTCALL call_11_9_0_0(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(0<<4)+0);
}
#define init_11_9_0_1
void FASTCALL call_11_9_0_1(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(0<<4)+1);
}
#define init_11_9_0_2
void FASTCALL call_11_9_0_2(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(0<<4)+2);
}
#define init_11_9_0_3
void FASTCALL call_11_9_0_3(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(0<<4)+3);
}
#define init_11_9_0_4
void FASTCALL call_11_9_0_4(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(0<<4)+4);
}
#define init_11_9_0_5
void FASTCALL call_11_9_0_5(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(0<<4)+5);
}
#define init_11_9_0_6
void FASTCALL call_11_9_0_6(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(0<<4)+6);
}
#define init_11_9_0_7
void FASTCALL call_11_9_0_7(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(0<<4)+7);
}
#define init_11_9_0_8
void FASTCALL call_11_9_0_8(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(0<<4)+8);
}
#define init_11_9_0_9
void FASTCALL call_11_9_0_9(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(0<<4)+9);
}
#define init_11_9_0_10
void FASTCALL call_11_9_0_10(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(0<<4)+10);
}
#define init_11_9_0_11
void FASTCALL call_11_9_0_11(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(0<<4)+11);
}
#define init_11_9_0_12
void FASTCALL call_11_9_0_12(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(0<<4)+12);
}
#define init_11_9_0_13
void FASTCALL call_11_9_0_13(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(0<<4)+13);
}
#define init_11_9_0_14
void FASTCALL call_11_9_0_14(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(0<<4)+14);
}
#define init_11_9_0_15
void FASTCALL call_11_9_0_15(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(0<<4)+15);
}
#define init_11_9_1_0
void FASTCALL call_11_9_1_0(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(1<<4)+0);
}
#define init_11_9_1_1
void FASTCALL call_11_9_1_1(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(1<<4)+1);
}
#define init_11_9_1_2
void FASTCALL call_11_9_1_2(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(1<<4)+2);
}
#define init_11_9_1_3
void FASTCALL call_11_9_1_3(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(1<<4)+3);
}
#define init_11_9_1_4
void FASTCALL call_11_9_1_4(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(1<<4)+4);
}
#define init_11_9_1_5
void FASTCALL call_11_9_1_5(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(1<<4)+5);
}
#define init_11_9_1_6
void FASTCALL call_11_9_1_6(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(1<<4)+6);
}
#define init_11_9_1_7
void FASTCALL call_11_9_1_7(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(1<<4)+7);
}
#define init_11_9_1_8
void FASTCALL call_11_9_1_8(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(1<<4)+8);
}
#define init_11_9_1_9
void FASTCALL call_11_9_1_9(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(1<<4)+9);
}
#define init_11_9_1_10
void FASTCALL call_11_9_1_10(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(1<<4)+10);
}
#define init_11_9_1_11
void FASTCALL call_11_9_1_11(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(1<<4)+11);
}
#define init_11_9_1_12
void FASTCALL call_11_9_1_12(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(1<<4)+12);
}
#define init_11_9_1_13
void FASTCALL call_11_9_1_13(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(1<<4)+13);
}
#define init_11_9_1_14
void FASTCALL call_11_9_1_14(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(1<<4)+14);
}
#define init_11_9_1_15
void FASTCALL call_11_9_1_15(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(1<<4)+15);
}
#define init_11_9_2_0
void FASTCALL call_11_9_2_0(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(2<<4)+0);
}
#define init_11_9_2_1
void FASTCALL call_11_9_2_1(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(2<<4)+1);
}
#define init_11_9_2_2
void FASTCALL call_11_9_2_2(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(2<<4)+2);
}
#define init_11_9_2_3
void FASTCALL call_11_9_2_3(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(2<<4)+3);
}
#define init_11_9_2_4
void FASTCALL call_11_9_2_4(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(2<<4)+4);
}
#define init_11_9_2_5
void FASTCALL call_11_9_2_5(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(2<<4)+5);
}
#define init_11_9_2_6
void FASTCALL call_11_9_2_6(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(2<<4)+6);
}
#define init_11_9_2_7
void FASTCALL call_11_9_2_7(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(2<<4)+7);
}
#define init_11_9_2_8
void FASTCALL call_11_9_2_8(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(2<<4)+8);
}
#define init_11_9_2_9
void FASTCALL call_11_9_2_9(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(2<<4)+9);
}
#define init_11_9_2_10
void FASTCALL call_11_9_2_10(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(2<<4)+10);
}
#define init_11_9_2_11
void FASTCALL call_11_9_2_11(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(2<<4)+11);
}
#define init_11_9_2_12
void FASTCALL call_11_9_2_12(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(2<<4)+12);
}
#define init_11_9_2_13
void FASTCALL call_11_9_2_13(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(2<<4)+13);
}
#define init_11_9_2_14
void FASTCALL call_11_9_2_14(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(2<<4)+14);
}
#define init_11_9_2_15
void FASTCALL call_11_9_2_15(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(2<<4)+15);
}
#define init_11_9_3_0
void FASTCALL call_11_9_3_0(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(3<<4)+0);
}
#define init_11_9_3_1
void FASTCALL call_11_9_3_1(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(3<<4)+1);
}
#define init_11_9_3_2
void FASTCALL call_11_9_3_2(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(3<<4)+2);
}
#define init_11_9_3_3
void FASTCALL call_11_9_3_3(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(3<<4)+3);
}
#define init_11_9_3_4
void FASTCALL call_11_9_3_4(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(3<<4)+4);
}
#define init_11_9_3_5
void FASTCALL call_11_9_3_5(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(3<<4)+5);
}
#define init_11_9_3_6
void FASTCALL call_11_9_3_6(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(3<<4)+6);
}
#define init_11_9_3_7
void FASTCALL call_11_9_3_7(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(3<<4)+7);
}
#define init_11_9_3_8
void FASTCALL call_11_9_3_8(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(3<<4)+8);
}
#define init_11_9_3_9
void FASTCALL call_11_9_3_9(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(3<<4)+9);
}
#define init_11_9_3_10
void FASTCALL call_11_9_3_10(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(3<<4)+10);
}
#define init_11_9_3_11
void FASTCALL call_11_9_3_11(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(3<<4)+11);
}
#define init_11_9_3_12
void FASTCALL call_11_9_3_12(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(3<<4)+12);
}
#define init_11_9_3_13
void FASTCALL call_11_9_3_13(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(3<<4)+13);
}
#define init_11_9_3_14
void FASTCALL call_11_9_3_14(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(3<<4)+14);
}
#define init_11_9_3_15
void FASTCALL call_11_9_3_15(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(3<<4)+15);
}
#define init_11_9_4_0
void FASTCALL call_11_9_4_0(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(4<<4)+0);
}
#define init_11_9_4_1
void FASTCALL call_11_9_4_1(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(4<<4)+1);
}
#define init_11_9_4_2
void FASTCALL call_11_9_4_2(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(4<<4)+2);
}
#define init_11_9_4_3
void FASTCALL call_11_9_4_3(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(4<<4)+3);
}
#define init_11_9_4_4
void FASTCALL call_11_9_4_4(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(4<<4)+4);
}
#define init_11_9_4_5
void FASTCALL call_11_9_4_5(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(4<<4)+5);
}
#define init_11_9_4_6
void FASTCALL call_11_9_4_6(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(4<<4)+6);
}
#define init_11_9_4_7
void FASTCALL call_11_9_4_7(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(4<<4)+7);
}
#define init_11_9_4_8
void FASTCALL call_11_9_4_8(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(4<<4)+8);
}
#define init_11_9_4_9
void FASTCALL call_11_9_4_9(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(4<<4)+9);
}
#define init_11_9_4_10
void FASTCALL call_11_9_4_10(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(4<<4)+10);
}
#define init_11_9_4_11
void FASTCALL call_11_9_4_11(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(4<<4)+11);
}
#define init_11_9_4_12
void FASTCALL call_11_9_4_12(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(4<<4)+12);
}
#define init_11_9_4_13
void FASTCALL call_11_9_4_13(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(4<<4)+13);
}
#define init_11_9_4_14
void FASTCALL call_11_9_4_14(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(4<<4)+14);
}
#define init_11_9_4_15
void FASTCALL call_11_9_4_15(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(4<<4)+15);
}
#define init_11_9_5_0
void FASTCALL call_11_9_5_0(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(5<<4)+0);
}
#define init_11_9_5_1
void FASTCALL call_11_9_5_1(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(5<<4)+1);
}
#define init_11_9_5_2
void FASTCALL call_11_9_5_2(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(5<<4)+2);
}
#define init_11_9_5_3
void FASTCALL call_11_9_5_3(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(5<<4)+3);
}
#define init_11_9_5_4
void FASTCALL call_11_9_5_4(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(5<<4)+4);
}
#define init_11_9_5_5
void FASTCALL call_11_9_5_5(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(5<<4)+5);
}
#define init_11_9_5_6
void FASTCALL call_11_9_5_6(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(5<<4)+6);
}
#define init_11_9_5_7
void FASTCALL call_11_9_5_7(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(5<<4)+7);
}
#define init_11_9_5_8
void FASTCALL call_11_9_5_8(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(5<<4)+8);
}
#define init_11_9_5_9
void FASTCALL call_11_9_5_9(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(5<<4)+9);
}
#define init_11_9_5_10
void FASTCALL call_11_9_5_10(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(5<<4)+10);
}
#define init_11_9_5_11
void FASTCALL call_11_9_5_11(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(5<<4)+11);
}
#define init_11_9_5_12
void FASTCALL call_11_9_5_12(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(5<<4)+12);
}
#define init_11_9_5_13
void FASTCALL call_11_9_5_13(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(5<<4)+13);
}
#define init_11_9_5_14
void FASTCALL call_11_9_5_14(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(5<<4)+14);
}
#define init_11_9_5_15
void FASTCALL call_11_9_5_15(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(5<<4)+15);
}
#define init_11_9_6_0
void FASTCALL call_11_9_6_0(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(6<<4)+0);
}
#define init_11_9_6_1
void FASTCALL call_11_9_6_1(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(6<<4)+1);
}
#define init_11_9_6_2
void FASTCALL call_11_9_6_2(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(6<<4)+2);
}
#define init_11_9_6_3
void FASTCALL call_11_9_6_3(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(6<<4)+3);
}
#define init_11_9_6_4
void FASTCALL call_11_9_6_4(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(6<<4)+4);
}
#define init_11_9_6_5
void FASTCALL call_11_9_6_5(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(6<<4)+5);
}
#define init_11_9_6_6
void FASTCALL call_11_9_6_6(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(6<<4)+6);
}
#define init_11_9_6_7
void FASTCALL call_11_9_6_7(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(6<<4)+7);
}
#define init_11_9_6_8
void FASTCALL call_11_9_6_8(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(6<<4)+8);
}
#define init_11_9_6_9
void FASTCALL call_11_9_6_9(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(6<<4)+9);
}
#define init_11_9_6_10
void FASTCALL call_11_9_6_10(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(6<<4)+10);
}
#define init_11_9_6_11
void FASTCALL call_11_9_6_11(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(6<<4)+11);
}
#define init_11_9_6_12
void FASTCALL call_11_9_6_12(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(6<<4)+12);
}
#define init_11_9_6_13
void FASTCALL call_11_9_6_13(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(6<<4)+13);
}
#define init_11_9_6_14
void FASTCALL call_11_9_6_14(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(6<<4)+14);
}
#define init_11_9_6_15
void FASTCALL call_11_9_6_15(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(6<<4)+15);
}
#define init_11_9_7_0
void FASTCALL call_11_9_7_0(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(7<<4)+0);
}
#define init_11_9_7_1
void FASTCALL call_11_9_7_1(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(7<<4)+1);
}
#define init_11_9_7_2
void FASTCALL call_11_9_7_2(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(7<<4)+2);
}
#define init_11_9_7_3
void FASTCALL call_11_9_7_3(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(7<<4)+3);
}
#define init_11_9_7_4
void FASTCALL call_11_9_7_4(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(7<<4)+4);
}
#define init_11_9_7_5
void FASTCALL call_11_9_7_5(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(7<<4)+5);
}
#define init_11_9_7_6
void FASTCALL call_11_9_7_6(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(7<<4)+6);
}
#define init_11_9_7_7
void FASTCALL call_11_9_7_7(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(7<<4)+7);
}
#define init_11_9_7_8
void FASTCALL call_11_9_7_8(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(7<<4)+8);
}
#define init_11_9_7_9
void FASTCALL call_11_9_7_9(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(7<<4)+9);
}
#define init_11_9_7_10
void FASTCALL call_11_9_7_10(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(7<<4)+10);
}
#define init_11_9_7_11
void FASTCALL call_11_9_7_11(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(7<<4)+11);
}
#define init_11_9_7_12
void FASTCALL call_11_9_7_12(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(7<<4)+12);
}
#define init_11_9_7_13
void FASTCALL call_11_9_7_13(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(7<<4)+13);
}
#define init_11_9_7_14
void FASTCALL call_11_9_7_14(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(7<<4)+14);
}
#define init_11_9_7_15
void FASTCALL call_11_9_7_15(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(7<<4)+15);
}
#define init_11_9_8_0
void FASTCALL call_11_9_8_0(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(8<<4)+0);
}
#define init_11_9_8_1
void FASTCALL call_11_9_8_1(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(8<<4)+1);
}
#define init_11_9_8_2
void FASTCALL call_11_9_8_2(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(8<<4)+2);
}
#define init_11_9_8_3
void FASTCALL call_11_9_8_3(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(8<<4)+3);
}
#define init_11_9_8_4
void FASTCALL call_11_9_8_4(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(8<<4)+4);
}
#define init_11_9_8_5
void FASTCALL call_11_9_8_5(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(8<<4)+5);
}
#define init_11_9_8_6
void FASTCALL call_11_9_8_6(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(8<<4)+6);
}
#define init_11_9_8_7
void FASTCALL call_11_9_8_7(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(8<<4)+7);
}
#define init_11_9_8_8
void FASTCALL call_11_9_8_8(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(8<<4)+8);
}
#define init_11_9_8_9
void FASTCALL call_11_9_8_9(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(8<<4)+9);
}
#define init_11_9_8_10
void FASTCALL call_11_9_8_10(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(8<<4)+10);
}
#define init_11_9_8_11
void FASTCALL call_11_9_8_11(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(8<<4)+11);
}
#define init_11_9_8_12
void FASTCALL call_11_9_8_12(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(8<<4)+12);
}
#define init_11_9_8_13
void FASTCALL call_11_9_8_13(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(8<<4)+13);
}
#define init_11_9_8_14
void FASTCALL call_11_9_8_14(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(8<<4)+14);
}
#define init_11_9_8_15
void FASTCALL call_11_9_8_15(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(8<<4)+15);
}
#define init_11_9_9_0
void FASTCALL call_11_9_9_0(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(9<<4)+0);
}
#define init_11_9_9_1
void FASTCALL call_11_9_9_1(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(9<<4)+1);
}
#define init_11_9_9_2
void FASTCALL call_11_9_9_2(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(9<<4)+2);
}
#define init_11_9_9_3
void FASTCALL call_11_9_9_3(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(9<<4)+3);
}
#define init_11_9_9_4
void FASTCALL call_11_9_9_4(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(9<<4)+4);
}
#define init_11_9_9_5
void FASTCALL call_11_9_9_5(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(9<<4)+5);
}
#define init_11_9_9_6
void FASTCALL call_11_9_9_6(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(9<<4)+6);
}
#define init_11_9_9_7
void FASTCALL call_11_9_9_7(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(9<<4)+7);
}
#define init_11_9_9_8
void FASTCALL call_11_9_9_8(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(9<<4)+8);
}
#define init_11_9_9_9
void FASTCALL call_11_9_9_9(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(9<<4)+9);
}
#define init_11_9_9_10
void FASTCALL call_11_9_9_10(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(9<<4)+10);
}
#define init_11_9_9_11
void FASTCALL call_11_9_9_11(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(9<<4)+11);
}
#define init_11_9_9_12
void FASTCALL call_11_9_9_12(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(9<<4)+12);
}
#define init_11_9_9_13
void FASTCALL call_11_9_9_13(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(9<<4)+13);
}
#define init_11_9_9_14
void FASTCALL call_11_9_9_14(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(9<<4)+14);
}
#define init_11_9_9_15
void FASTCALL call_11_9_9_15(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(9<<4)+15);
}
#define init_11_9_10_0
void FASTCALL call_11_9_10_0(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(10<<4)+0);
}
#define init_11_9_10_1
void FASTCALL call_11_9_10_1(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(10<<4)+1);
}
#define init_11_9_10_2
void FASTCALL call_11_9_10_2(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(10<<4)+2);
}
#define init_11_9_10_3
void FASTCALL call_11_9_10_3(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(10<<4)+3);
}
#define init_11_9_10_4
void FASTCALL call_11_9_10_4(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(10<<4)+4);
}
#define init_11_9_10_5
void FASTCALL call_11_9_10_5(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(10<<4)+5);
}
#define init_11_9_10_6
void FASTCALL call_11_9_10_6(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(10<<4)+6);
}
#define init_11_9_10_7
void FASTCALL call_11_9_10_7(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(10<<4)+7);
}
#define init_11_9_10_8
void FASTCALL call_11_9_10_8(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(10<<4)+8);
}
#define init_11_9_10_9
void FASTCALL call_11_9_10_9(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(10<<4)+9);
}
#define init_11_9_10_10
void FASTCALL call_11_9_10_10(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(10<<4)+10);
}
#define init_11_9_10_11
void FASTCALL call_11_9_10_11(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(10<<4)+11);
}
#define init_11_9_10_12
void FASTCALL call_11_9_10_12(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(10<<4)+12);
}
#define init_11_9_10_13
void FASTCALL call_11_9_10_13(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(10<<4)+13);
}
#define init_11_9_10_14
void FASTCALL call_11_9_10_14(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(10<<4)+14);
}
#define init_11_9_10_15
void FASTCALL call_11_9_10_15(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(10<<4)+15);
}
#define init_11_9_11_0
void FASTCALL call_11_9_11_0(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(11<<4)+0);
}
#define init_11_9_11_1
void FASTCALL call_11_9_11_1(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(11<<4)+1);
}
#define init_11_9_11_2
void FASTCALL call_11_9_11_2(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(11<<4)+2);
}
#define init_11_9_11_3
void FASTCALL call_11_9_11_3(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(11<<4)+3);
}
#define init_11_9_11_4
void FASTCALL call_11_9_11_4(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(11<<4)+4);
}
#define init_11_9_11_5
void FASTCALL call_11_9_11_5(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(11<<4)+5);
}
#define init_11_9_11_6
void FASTCALL call_11_9_11_6(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(11<<4)+6);
}
#define init_11_9_11_7
void FASTCALL call_11_9_11_7(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(11<<4)+7);
}
#define init_11_9_11_8
void FASTCALL call_11_9_11_8(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(11<<4)+8);
}
#define init_11_9_11_9
void FASTCALL call_11_9_11_9(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(11<<4)+9);
}
#define init_11_9_11_10
void FASTCALL call_11_9_11_10(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(11<<4)+10);
}
#define init_11_9_11_11
void FASTCALL call_11_9_11_11(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(11<<4)+11);
}
#define init_11_9_11_12
void FASTCALL call_11_9_11_12(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(11<<4)+12);
}
#define init_11_9_11_13
void FASTCALL call_11_9_11_13(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(11<<4)+13);
}
#define init_11_9_11_14
void FASTCALL call_11_9_11_14(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(11<<4)+14);
}
#define init_11_9_11_15
void FASTCALL call_11_9_11_15(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(11<<4)+15);
}
#define init_11_9_12_0
void FASTCALL call_11_9_12_0(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(12<<4)+0);
}
#define init_11_9_12_1
void FASTCALL call_11_9_12_1(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(12<<4)+1);
}
#define init_11_9_12_2
void FASTCALL call_11_9_12_2(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(12<<4)+2);
}
#define init_11_9_12_3
void FASTCALL call_11_9_12_3(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(12<<4)+3);
}
#define init_11_9_12_4
void FASTCALL call_11_9_12_4(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(12<<4)+4);
}
#define init_11_9_12_5
void FASTCALL call_11_9_12_5(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(12<<4)+5);
}
#define init_11_9_12_6
void FASTCALL call_11_9_12_6(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(12<<4)+6);
}
#define init_11_9_12_7
void FASTCALL call_11_9_12_7(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(12<<4)+7);
}
#define init_11_9_12_8
void FASTCALL call_11_9_12_8(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(12<<4)+8);
}
#define init_11_9_12_9
void FASTCALL call_11_9_12_9(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(12<<4)+9);
}
#define init_11_9_12_10
void FASTCALL call_11_9_12_10(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(12<<4)+10);
}
#define init_11_9_12_11
void FASTCALL call_11_9_12_11(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(12<<4)+11);
}
#define init_11_9_12_12
void FASTCALL call_11_9_12_12(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(12<<4)+12);
}
#define init_11_9_12_13
void FASTCALL call_11_9_12_13(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(12<<4)+13);
}
#define init_11_9_12_14
void FASTCALL call_11_9_12_14(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(12<<4)+14);
}
#define init_11_9_12_15
void FASTCALL call_11_9_12_15(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(12<<4)+15);
}
#define init_11_9_13_0
void FASTCALL call_11_9_13_0(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(13<<4)+0);
}
#define init_11_9_13_1
void FASTCALL call_11_9_13_1(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(13<<4)+1);
}
#define init_11_9_13_2
void FASTCALL call_11_9_13_2(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(13<<4)+2);
}
#define init_11_9_13_3
void FASTCALL call_11_9_13_3(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(13<<4)+3);
}
#define init_11_9_13_4
void FASTCALL call_11_9_13_4(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(13<<4)+4);
}
#define init_11_9_13_5
void FASTCALL call_11_9_13_5(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(13<<4)+5);
}
#define init_11_9_13_6
void FASTCALL call_11_9_13_6(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(13<<4)+6);
}
#define init_11_9_13_7
void FASTCALL call_11_9_13_7(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(13<<4)+7);
}
#define init_11_9_13_8
void FASTCALL call_11_9_13_8(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(13<<4)+8);
}
#define init_11_9_13_9
void FASTCALL call_11_9_13_9(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(13<<4)+9);
}
#define init_11_9_13_10
void FASTCALL call_11_9_13_10(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(13<<4)+10);
}
#define init_11_9_13_11
void FASTCALL call_11_9_13_11(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(13<<4)+11);
}
#define init_11_9_13_12
void FASTCALL call_11_9_13_12(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(13<<4)+12);
}
#define init_11_9_13_13
void FASTCALL call_11_9_13_13(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(13<<4)+13);
}
#define init_11_9_13_14
void FASTCALL call_11_9_13_14(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(13<<4)+14);
}
#define init_11_9_13_15
void FASTCALL call_11_9_13_15(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(13<<4)+15);
}
#define init_11_9_14_0
void FASTCALL call_11_9_14_0(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(14<<4)+0);
}
#define init_11_9_14_1
void FASTCALL call_11_9_14_1(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(14<<4)+1);
}
#define init_11_9_14_2
void FASTCALL call_11_9_14_2(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(14<<4)+2);
}
#define init_11_9_14_3
void FASTCALL call_11_9_14_3(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(14<<4)+3);
}
#define init_11_9_14_4
void FASTCALL call_11_9_14_4(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(14<<4)+4);
}
#define init_11_9_14_5
void FASTCALL call_11_9_14_5(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(14<<4)+5);
}
#define init_11_9_14_6
void FASTCALL call_11_9_14_6(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(14<<4)+6);
}
#define init_11_9_14_7
void FASTCALL call_11_9_14_7(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(14<<4)+7);
}
#define init_11_9_14_8
void FASTCALL call_11_9_14_8(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(14<<4)+8);
}
#define init_11_9_14_9
void FASTCALL call_11_9_14_9(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(14<<4)+9);
}
#define init_11_9_14_10
void FASTCALL call_11_9_14_10(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(14<<4)+10);
}
#define init_11_9_14_11
void FASTCALL call_11_9_14_11(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(14<<4)+11);
}
#define init_11_9_14_12
void FASTCALL call_11_9_14_12(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(14<<4)+12);
}
#define init_11_9_14_13
void FASTCALL call_11_9_14_13(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(14<<4)+13);
}
#define init_11_9_14_14
void FASTCALL call_11_9_14_14(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(14<<4)+14);
}
#define init_11_9_14_15
void FASTCALL call_11_9_14_15(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(14<<4)+15);
}
#define init_11_9_15_0
void FASTCALL call_11_9_15_0(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(15<<4)+0);
}
#define init_11_9_15_1
void FASTCALL call_11_9_15_1(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(15<<4)+1);
}
#define init_11_9_15_2
void FASTCALL call_11_9_15_2(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(15<<4)+2);
}
#define init_11_9_15_3
void FASTCALL call_11_9_15_3(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(15<<4)+3);
}
#define init_11_9_15_4
void FASTCALL call_11_9_15_4(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(15<<4)+4);
}
#define init_11_9_15_5
void FASTCALL call_11_9_15_5(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(15<<4)+5);
}
#define init_11_9_15_6
void FASTCALL call_11_9_15_6(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(15<<4)+6);
}
#define init_11_9_15_7
void FASTCALL call_11_9_15_7(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(15<<4)+7);
}
#define init_11_9_15_8
void FASTCALL call_11_9_15_8(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(15<<4)+8);
}
#define init_11_9_15_9
void FASTCALL call_11_9_15_9(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(15<<4)+9);
}
#define init_11_9_15_10
void FASTCALL call_11_9_15_10(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(15<<4)+10);
}
#define init_11_9_15_11
void FASTCALL call_11_9_15_11(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(15<<4)+11);
}
#define init_11_9_15_12
void FASTCALL call_11_9_15_12(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(15<<4)+12);
}
#define init_11_9_15_13
void FASTCALL call_11_9_15_13(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(15<<4)+13);
}
#define init_11_9_15_14
void FASTCALL call_11_9_15_14(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(15<<4)+14);
}
#define init_11_9_15_15
void FASTCALL call_11_9_15_15(SH2_struct* sh)
{
  SH2bsr(sh, (9<<8)+(15<<4)+15);
}
#define init_11_10_0_0
void FASTCALL call_11_10_0_0(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(0<<4)+0);
}
#define init_11_10_0_1
void FASTCALL call_11_10_0_1(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(0<<4)+1);
}
#define init_11_10_0_2
void FASTCALL call_11_10_0_2(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(0<<4)+2);
}
#define init_11_10_0_3
void FASTCALL call_11_10_0_3(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(0<<4)+3);
}
#define init_11_10_0_4
void FASTCALL call_11_10_0_4(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(0<<4)+4);
}
#define init_11_10_0_5
void FASTCALL call_11_10_0_5(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(0<<4)+5);
}
#define init_11_10_0_6
void FASTCALL call_11_10_0_6(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(0<<4)+6);
}
#define init_11_10_0_7
void FASTCALL call_11_10_0_7(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(0<<4)+7);
}
#define init_11_10_0_8
void FASTCALL call_11_10_0_8(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(0<<4)+8);
}
#define init_11_10_0_9
void FASTCALL call_11_10_0_9(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(0<<4)+9);
}
#define init_11_10_0_10
void FASTCALL call_11_10_0_10(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(0<<4)+10);
}
#define init_11_10_0_11
void FASTCALL call_11_10_0_11(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(0<<4)+11);
}
#define init_11_10_0_12
void FASTCALL call_11_10_0_12(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(0<<4)+12);
}
#define init_11_10_0_13
void FASTCALL call_11_10_0_13(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(0<<4)+13);
}
#define init_11_10_0_14
void FASTCALL call_11_10_0_14(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(0<<4)+14);
}
#define init_11_10_0_15
void FASTCALL call_11_10_0_15(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(0<<4)+15);
}
#define init_11_10_1_0
void FASTCALL call_11_10_1_0(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(1<<4)+0);
}
#define init_11_10_1_1
void FASTCALL call_11_10_1_1(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(1<<4)+1);
}
#define init_11_10_1_2
void FASTCALL call_11_10_1_2(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(1<<4)+2);
}
#define init_11_10_1_3
void FASTCALL call_11_10_1_3(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(1<<4)+3);
}
#define init_11_10_1_4
void FASTCALL call_11_10_1_4(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(1<<4)+4);
}
#define init_11_10_1_5
void FASTCALL call_11_10_1_5(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(1<<4)+5);
}
#define init_11_10_1_6
void FASTCALL call_11_10_1_6(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(1<<4)+6);
}
#define init_11_10_1_7
void FASTCALL call_11_10_1_7(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(1<<4)+7);
}
#define init_11_10_1_8
void FASTCALL call_11_10_1_8(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(1<<4)+8);
}
#define init_11_10_1_9
void FASTCALL call_11_10_1_9(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(1<<4)+9);
}
#define init_11_10_1_10
void FASTCALL call_11_10_1_10(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(1<<4)+10);
}
#define init_11_10_1_11
void FASTCALL call_11_10_1_11(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(1<<4)+11);
}
#define init_11_10_1_12
void FASTCALL call_11_10_1_12(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(1<<4)+12);
}
#define init_11_10_1_13
void FASTCALL call_11_10_1_13(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(1<<4)+13);
}
#define init_11_10_1_14
void FASTCALL call_11_10_1_14(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(1<<4)+14);
}
#define init_11_10_1_15
void FASTCALL call_11_10_1_15(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(1<<4)+15);
}
#define init_11_10_2_0
void FASTCALL call_11_10_2_0(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(2<<4)+0);
}
#define init_11_10_2_1
void FASTCALL call_11_10_2_1(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(2<<4)+1);
}
#define init_11_10_2_2
void FASTCALL call_11_10_2_2(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(2<<4)+2);
}
#define init_11_10_2_3
void FASTCALL call_11_10_2_3(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(2<<4)+3);
}
#define init_11_10_2_4
void FASTCALL call_11_10_2_4(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(2<<4)+4);
}
#define init_11_10_2_5
void FASTCALL call_11_10_2_5(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(2<<4)+5);
}
#define init_11_10_2_6
void FASTCALL call_11_10_2_6(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(2<<4)+6);
}
#define init_11_10_2_7
void FASTCALL call_11_10_2_7(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(2<<4)+7);
}
#define init_11_10_2_8
void FASTCALL call_11_10_2_8(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(2<<4)+8);
}
#define init_11_10_2_9
void FASTCALL call_11_10_2_9(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(2<<4)+9);
}
#define init_11_10_2_10
void FASTCALL call_11_10_2_10(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(2<<4)+10);
}
#define init_11_10_2_11
void FASTCALL call_11_10_2_11(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(2<<4)+11);
}
#define init_11_10_2_12
void FASTCALL call_11_10_2_12(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(2<<4)+12);
}
#define init_11_10_2_13
void FASTCALL call_11_10_2_13(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(2<<4)+13);
}
#define init_11_10_2_14
void FASTCALL call_11_10_2_14(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(2<<4)+14);
}
#define init_11_10_2_15
void FASTCALL call_11_10_2_15(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(2<<4)+15);
}
#define init_11_10_3_0
void FASTCALL call_11_10_3_0(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(3<<4)+0);
}
#define init_11_10_3_1
void FASTCALL call_11_10_3_1(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(3<<4)+1);
}
#define init_11_10_3_2
void FASTCALL call_11_10_3_2(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(3<<4)+2);
}
#define init_11_10_3_3
void FASTCALL call_11_10_3_3(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(3<<4)+3);
}
#define init_11_10_3_4
void FASTCALL call_11_10_3_4(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(3<<4)+4);
}
#define init_11_10_3_5
void FASTCALL call_11_10_3_5(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(3<<4)+5);
}
#define init_11_10_3_6
void FASTCALL call_11_10_3_6(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(3<<4)+6);
}
#define init_11_10_3_7
void FASTCALL call_11_10_3_7(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(3<<4)+7);
}
#define init_11_10_3_8
void FASTCALL call_11_10_3_8(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(3<<4)+8);
}
#define init_11_10_3_9
void FASTCALL call_11_10_3_9(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(3<<4)+9);
}
#define init_11_10_3_10
void FASTCALL call_11_10_3_10(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(3<<4)+10);
}
#define init_11_10_3_11
void FASTCALL call_11_10_3_11(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(3<<4)+11);
}
#define init_11_10_3_12
void FASTCALL call_11_10_3_12(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(3<<4)+12);
}
#define init_11_10_3_13
void FASTCALL call_11_10_3_13(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(3<<4)+13);
}
#define init_11_10_3_14
void FASTCALL call_11_10_3_14(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(3<<4)+14);
}
#define init_11_10_3_15
void FASTCALL call_11_10_3_15(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(3<<4)+15);
}
#define init_11_10_4_0
void FASTCALL call_11_10_4_0(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(4<<4)+0);
}
#define init_11_10_4_1
void FASTCALL call_11_10_4_1(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(4<<4)+1);
}
#define init_11_10_4_2
void FASTCALL call_11_10_4_2(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(4<<4)+2);
}
#define init_11_10_4_3
void FASTCALL call_11_10_4_3(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(4<<4)+3);
}
#define init_11_10_4_4
void FASTCALL call_11_10_4_4(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(4<<4)+4);
}
#define init_11_10_4_5
void FASTCALL call_11_10_4_5(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(4<<4)+5);
}
#define init_11_10_4_6
void FASTCALL call_11_10_4_6(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(4<<4)+6);
}
#define init_11_10_4_7
void FASTCALL call_11_10_4_7(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(4<<4)+7);
}
#define init_11_10_4_8
void FASTCALL call_11_10_4_8(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(4<<4)+8);
}
#define init_11_10_4_9
void FASTCALL call_11_10_4_9(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(4<<4)+9);
}
#define init_11_10_4_10
void FASTCALL call_11_10_4_10(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(4<<4)+10);
}
#define init_11_10_4_11
void FASTCALL call_11_10_4_11(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(4<<4)+11);
}
#define init_11_10_4_12
void FASTCALL call_11_10_4_12(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(4<<4)+12);
}
#define init_11_10_4_13
void FASTCALL call_11_10_4_13(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(4<<4)+13);
}
#define init_11_10_4_14
void FASTCALL call_11_10_4_14(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(4<<4)+14);
}
#define init_11_10_4_15
void FASTCALL call_11_10_4_15(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(4<<4)+15);
}
#define init_11_10_5_0
void FASTCALL call_11_10_5_0(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(5<<4)+0);
}
#define init_11_10_5_1
void FASTCALL call_11_10_5_1(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(5<<4)+1);
}
#define init_11_10_5_2
void FASTCALL call_11_10_5_2(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(5<<4)+2);
}
#define init_11_10_5_3
void FASTCALL call_11_10_5_3(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(5<<4)+3);
}
#define init_11_10_5_4
void FASTCALL call_11_10_5_4(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(5<<4)+4);
}
#define init_11_10_5_5
void FASTCALL call_11_10_5_5(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(5<<4)+5);
}
#define init_11_10_5_6
void FASTCALL call_11_10_5_6(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(5<<4)+6);
}
#define init_11_10_5_7
void FASTCALL call_11_10_5_7(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(5<<4)+7);
}
#define init_11_10_5_8
void FASTCALL call_11_10_5_8(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(5<<4)+8);
}
#define init_11_10_5_9
void FASTCALL call_11_10_5_9(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(5<<4)+9);
}
#define init_11_10_5_10
void FASTCALL call_11_10_5_10(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(5<<4)+10);
}
#define init_11_10_5_11
void FASTCALL call_11_10_5_11(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(5<<4)+11);
}
#define init_11_10_5_12
void FASTCALL call_11_10_5_12(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(5<<4)+12);
}
#define init_11_10_5_13
void FASTCALL call_11_10_5_13(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(5<<4)+13);
}
#define init_11_10_5_14
void FASTCALL call_11_10_5_14(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(5<<4)+14);
}
#define init_11_10_5_15
void FASTCALL call_11_10_5_15(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(5<<4)+15);
}
#define init_11_10_6_0
void FASTCALL call_11_10_6_0(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(6<<4)+0);
}
#define init_11_10_6_1
void FASTCALL call_11_10_6_1(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(6<<4)+1);
}
#define init_11_10_6_2
void FASTCALL call_11_10_6_2(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(6<<4)+2);
}
#define init_11_10_6_3
void FASTCALL call_11_10_6_3(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(6<<4)+3);
}
#define init_11_10_6_4
void FASTCALL call_11_10_6_4(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(6<<4)+4);
}
#define init_11_10_6_5
void FASTCALL call_11_10_6_5(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(6<<4)+5);
}
#define init_11_10_6_6
void FASTCALL call_11_10_6_6(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(6<<4)+6);
}
#define init_11_10_6_7
void FASTCALL call_11_10_6_7(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(6<<4)+7);
}
#define init_11_10_6_8
void FASTCALL call_11_10_6_8(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(6<<4)+8);
}
#define init_11_10_6_9
void FASTCALL call_11_10_6_9(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(6<<4)+9);
}
#define init_11_10_6_10
void FASTCALL call_11_10_6_10(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(6<<4)+10);
}
#define init_11_10_6_11
void FASTCALL call_11_10_6_11(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(6<<4)+11);
}
#define init_11_10_6_12
void FASTCALL call_11_10_6_12(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(6<<4)+12);
}
#define init_11_10_6_13
void FASTCALL call_11_10_6_13(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(6<<4)+13);
}
#define init_11_10_6_14
void FASTCALL call_11_10_6_14(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(6<<4)+14);
}
#define init_11_10_6_15
void FASTCALL call_11_10_6_15(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(6<<4)+15);
}
#define init_11_10_7_0
void FASTCALL call_11_10_7_0(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(7<<4)+0);
}
#define init_11_10_7_1
void FASTCALL call_11_10_7_1(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(7<<4)+1);
}
#define init_11_10_7_2
void FASTCALL call_11_10_7_2(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(7<<4)+2);
}
#define init_11_10_7_3
void FASTCALL call_11_10_7_3(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(7<<4)+3);
}
#define init_11_10_7_4
void FASTCALL call_11_10_7_4(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(7<<4)+4);
}
#define init_11_10_7_5
void FASTCALL call_11_10_7_5(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(7<<4)+5);
}
#define init_11_10_7_6
void FASTCALL call_11_10_7_6(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(7<<4)+6);
}
#define init_11_10_7_7
void FASTCALL call_11_10_7_7(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(7<<4)+7);
}
#define init_11_10_7_8
void FASTCALL call_11_10_7_8(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(7<<4)+8);
}
#define init_11_10_7_9
void FASTCALL call_11_10_7_9(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(7<<4)+9);
}
#define init_11_10_7_10
void FASTCALL call_11_10_7_10(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(7<<4)+10);
}
#define init_11_10_7_11
void FASTCALL call_11_10_7_11(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(7<<4)+11);
}
#define init_11_10_7_12
void FASTCALL call_11_10_7_12(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(7<<4)+12);
}
#define init_11_10_7_13
void FASTCALL call_11_10_7_13(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(7<<4)+13);
}
#define init_11_10_7_14
void FASTCALL call_11_10_7_14(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(7<<4)+14);
}
#define init_11_10_7_15
void FASTCALL call_11_10_7_15(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(7<<4)+15);
}
#define init_11_10_8_0
void FASTCALL call_11_10_8_0(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(8<<4)+0);
}
#define init_11_10_8_1
void FASTCALL call_11_10_8_1(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(8<<4)+1);
}
#define init_11_10_8_2
void FASTCALL call_11_10_8_2(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(8<<4)+2);
}
#define init_11_10_8_3
void FASTCALL call_11_10_8_3(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(8<<4)+3);
}
#define init_11_10_8_4
void FASTCALL call_11_10_8_4(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(8<<4)+4);
}
#define init_11_10_8_5
void FASTCALL call_11_10_8_5(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(8<<4)+5);
}
#define init_11_10_8_6
void FASTCALL call_11_10_8_6(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(8<<4)+6);
}
#define init_11_10_8_7
void FASTCALL call_11_10_8_7(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(8<<4)+7);
}
#define init_11_10_8_8
void FASTCALL call_11_10_8_8(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(8<<4)+8);
}
#define init_11_10_8_9
void FASTCALL call_11_10_8_9(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(8<<4)+9);
}
#define init_11_10_8_10
void FASTCALL call_11_10_8_10(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(8<<4)+10);
}
#define init_11_10_8_11
void FASTCALL call_11_10_8_11(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(8<<4)+11);
}
#define init_11_10_8_12
void FASTCALL call_11_10_8_12(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(8<<4)+12);
}
#define init_11_10_8_13
void FASTCALL call_11_10_8_13(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(8<<4)+13);
}
#define init_11_10_8_14
void FASTCALL call_11_10_8_14(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(8<<4)+14);
}
#define init_11_10_8_15
void FASTCALL call_11_10_8_15(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(8<<4)+15);
}
#define init_11_10_9_0
void FASTCALL call_11_10_9_0(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(9<<4)+0);
}
#define init_11_10_9_1
void FASTCALL call_11_10_9_1(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(9<<4)+1);
}
#define init_11_10_9_2
void FASTCALL call_11_10_9_2(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(9<<4)+2);
}
#define init_11_10_9_3
void FASTCALL call_11_10_9_3(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(9<<4)+3);
}
#define init_11_10_9_4
void FASTCALL call_11_10_9_4(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(9<<4)+4);
}
#define init_11_10_9_5
void FASTCALL call_11_10_9_5(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(9<<4)+5);
}
#define init_11_10_9_6
void FASTCALL call_11_10_9_6(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(9<<4)+6);
}
#define init_11_10_9_7
void FASTCALL call_11_10_9_7(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(9<<4)+7);
}
#define init_11_10_9_8
void FASTCALL call_11_10_9_8(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(9<<4)+8);
}
#define init_11_10_9_9
void FASTCALL call_11_10_9_9(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(9<<4)+9);
}
#define init_11_10_9_10
void FASTCALL call_11_10_9_10(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(9<<4)+10);
}
#define init_11_10_9_11
void FASTCALL call_11_10_9_11(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(9<<4)+11);
}
#define init_11_10_9_12
void FASTCALL call_11_10_9_12(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(9<<4)+12);
}
#define init_11_10_9_13
void FASTCALL call_11_10_9_13(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(9<<4)+13);
}
#define init_11_10_9_14
void FASTCALL call_11_10_9_14(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(9<<4)+14);
}
#define init_11_10_9_15
void FASTCALL call_11_10_9_15(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(9<<4)+15);
}
#define init_11_10_10_0
void FASTCALL call_11_10_10_0(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(10<<4)+0);
}
#define init_11_10_10_1
void FASTCALL call_11_10_10_1(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(10<<4)+1);
}
#define init_11_10_10_2
void FASTCALL call_11_10_10_2(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(10<<4)+2);
}
#define init_11_10_10_3
void FASTCALL call_11_10_10_3(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(10<<4)+3);
}
#define init_11_10_10_4
void FASTCALL call_11_10_10_4(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(10<<4)+4);
}
#define init_11_10_10_5
void FASTCALL call_11_10_10_5(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(10<<4)+5);
}
#define init_11_10_10_6
void FASTCALL call_11_10_10_6(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(10<<4)+6);
}
#define init_11_10_10_7
void FASTCALL call_11_10_10_7(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(10<<4)+7);
}
#define init_11_10_10_8
void FASTCALL call_11_10_10_8(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(10<<4)+8);
}
#define init_11_10_10_9
void FASTCALL call_11_10_10_9(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(10<<4)+9);
}
#define init_11_10_10_10
void FASTCALL call_11_10_10_10(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(10<<4)+10);
}
#define init_11_10_10_11
void FASTCALL call_11_10_10_11(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(10<<4)+11);
}
#define init_11_10_10_12
void FASTCALL call_11_10_10_12(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(10<<4)+12);
}
#define init_11_10_10_13
void FASTCALL call_11_10_10_13(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(10<<4)+13);
}
#define init_11_10_10_14
void FASTCALL call_11_10_10_14(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(10<<4)+14);
}
#define init_11_10_10_15
void FASTCALL call_11_10_10_15(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(10<<4)+15);
}
#define init_11_10_11_0
void FASTCALL call_11_10_11_0(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(11<<4)+0);
}
#define init_11_10_11_1
void FASTCALL call_11_10_11_1(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(11<<4)+1);
}
#define init_11_10_11_2
void FASTCALL call_11_10_11_2(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(11<<4)+2);
}
#define init_11_10_11_3
void FASTCALL call_11_10_11_3(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(11<<4)+3);
}
#define init_11_10_11_4
void FASTCALL call_11_10_11_4(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(11<<4)+4);
}
#define init_11_10_11_5
void FASTCALL call_11_10_11_5(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(11<<4)+5);
}
#define init_11_10_11_6
void FASTCALL call_11_10_11_6(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(11<<4)+6);
}
#define init_11_10_11_7
void FASTCALL call_11_10_11_7(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(11<<4)+7);
}
#define init_11_10_11_8
void FASTCALL call_11_10_11_8(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(11<<4)+8);
}
#define init_11_10_11_9
void FASTCALL call_11_10_11_9(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(11<<4)+9);
}
#define init_11_10_11_10
void FASTCALL call_11_10_11_10(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(11<<4)+10);
}
#define init_11_10_11_11
void FASTCALL call_11_10_11_11(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(11<<4)+11);
}
#define init_11_10_11_12
void FASTCALL call_11_10_11_12(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(11<<4)+12);
}
#define init_11_10_11_13
void FASTCALL call_11_10_11_13(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(11<<4)+13);
}
#define init_11_10_11_14
void FASTCALL call_11_10_11_14(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(11<<4)+14);
}
#define init_11_10_11_15
void FASTCALL call_11_10_11_15(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(11<<4)+15);
}
#define init_11_10_12_0
void FASTCALL call_11_10_12_0(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(12<<4)+0);
}
#define init_11_10_12_1
void FASTCALL call_11_10_12_1(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(12<<4)+1);
}
#define init_11_10_12_2
void FASTCALL call_11_10_12_2(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(12<<4)+2);
}
#define init_11_10_12_3
void FASTCALL call_11_10_12_3(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(12<<4)+3);
}
#define init_11_10_12_4
void FASTCALL call_11_10_12_4(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(12<<4)+4);
}
#define init_11_10_12_5
void FASTCALL call_11_10_12_5(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(12<<4)+5);
}
#define init_11_10_12_6
void FASTCALL call_11_10_12_6(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(12<<4)+6);
}
#define init_11_10_12_7
void FASTCALL call_11_10_12_7(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(12<<4)+7);
}
#define init_11_10_12_8
void FASTCALL call_11_10_12_8(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(12<<4)+8);
}
#define init_11_10_12_9
void FASTCALL call_11_10_12_9(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(12<<4)+9);
}
#define init_11_10_12_10
void FASTCALL call_11_10_12_10(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(12<<4)+10);
}
#define init_11_10_12_11
void FASTCALL call_11_10_12_11(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(12<<4)+11);
}
#define init_11_10_12_12
void FASTCALL call_11_10_12_12(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(12<<4)+12);
}
#define init_11_10_12_13
void FASTCALL call_11_10_12_13(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(12<<4)+13);
}
#define init_11_10_12_14
void FASTCALL call_11_10_12_14(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(12<<4)+14);
}
#define init_11_10_12_15
void FASTCALL call_11_10_12_15(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(12<<4)+15);
}
#define init_11_10_13_0
void FASTCALL call_11_10_13_0(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(13<<4)+0);
}
#define init_11_10_13_1
void FASTCALL call_11_10_13_1(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(13<<4)+1);
}
#define init_11_10_13_2
void FASTCALL call_11_10_13_2(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(13<<4)+2);
}
#define init_11_10_13_3
void FASTCALL call_11_10_13_3(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(13<<4)+3);
}
#define init_11_10_13_4
void FASTCALL call_11_10_13_4(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(13<<4)+4);
}
#define init_11_10_13_5
void FASTCALL call_11_10_13_5(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(13<<4)+5);
}
#define init_11_10_13_6
void FASTCALL call_11_10_13_6(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(13<<4)+6);
}
#define init_11_10_13_7
void FASTCALL call_11_10_13_7(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(13<<4)+7);
}
#define init_11_10_13_8
void FASTCALL call_11_10_13_8(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(13<<4)+8);
}
#define init_11_10_13_9
void FASTCALL call_11_10_13_9(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(13<<4)+9);
}
#define init_11_10_13_10
void FASTCALL call_11_10_13_10(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(13<<4)+10);
}
#define init_11_10_13_11
void FASTCALL call_11_10_13_11(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(13<<4)+11);
}
#define init_11_10_13_12
void FASTCALL call_11_10_13_12(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(13<<4)+12);
}
#define init_11_10_13_13
void FASTCALL call_11_10_13_13(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(13<<4)+13);
}
#define init_11_10_13_14
void FASTCALL call_11_10_13_14(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(13<<4)+14);
}
#define init_11_10_13_15
void FASTCALL call_11_10_13_15(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(13<<4)+15);
}
#define init_11_10_14_0
void FASTCALL call_11_10_14_0(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(14<<4)+0);
}
#define init_11_10_14_1
void FASTCALL call_11_10_14_1(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(14<<4)+1);
}
#define init_11_10_14_2
void FASTCALL call_11_10_14_2(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(14<<4)+2);
}
#define init_11_10_14_3
void FASTCALL call_11_10_14_3(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(14<<4)+3);
}
#define init_11_10_14_4
void FASTCALL call_11_10_14_4(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(14<<4)+4);
}
#define init_11_10_14_5
void FASTCALL call_11_10_14_5(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(14<<4)+5);
}
#define init_11_10_14_6
void FASTCALL call_11_10_14_6(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(14<<4)+6);
}
#define init_11_10_14_7
void FASTCALL call_11_10_14_7(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(14<<4)+7);
}
#define init_11_10_14_8
void FASTCALL call_11_10_14_8(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(14<<4)+8);
}
#define init_11_10_14_9
void FASTCALL call_11_10_14_9(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(14<<4)+9);
}
#define init_11_10_14_10
void FASTCALL call_11_10_14_10(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(14<<4)+10);
}
#define init_11_10_14_11
void FASTCALL call_11_10_14_11(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(14<<4)+11);
}
#define init_11_10_14_12
void FASTCALL call_11_10_14_12(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(14<<4)+12);
}
#define init_11_10_14_13
void FASTCALL call_11_10_14_13(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(14<<4)+13);
}
#define init_11_10_14_14
void FASTCALL call_11_10_14_14(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(14<<4)+14);
}
#define init_11_10_14_15
void FASTCALL call_11_10_14_15(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(14<<4)+15);
}
#define init_11_10_15_0
void FASTCALL call_11_10_15_0(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(15<<4)+0);
}
#define init_11_10_15_1
void FASTCALL call_11_10_15_1(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(15<<4)+1);
}
#define init_11_10_15_2
void FASTCALL call_11_10_15_2(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(15<<4)+2);
}
#define init_11_10_15_3
void FASTCALL call_11_10_15_3(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(15<<4)+3);
}
#define init_11_10_15_4
void FASTCALL call_11_10_15_4(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(15<<4)+4);
}
#define init_11_10_15_5
void FASTCALL call_11_10_15_5(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(15<<4)+5);
}
#define init_11_10_15_6
void FASTCALL call_11_10_15_6(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(15<<4)+6);
}
#define init_11_10_15_7
void FASTCALL call_11_10_15_7(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(15<<4)+7);
}
#define init_11_10_15_8
void FASTCALL call_11_10_15_8(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(15<<4)+8);
}
#define init_11_10_15_9
void FASTCALL call_11_10_15_9(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(15<<4)+9);
}
#define init_11_10_15_10
void FASTCALL call_11_10_15_10(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(15<<4)+10);
}
#define init_11_10_15_11
void FASTCALL call_11_10_15_11(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(15<<4)+11);
}
#define init_11_10_15_12
void FASTCALL call_11_10_15_12(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(15<<4)+12);
}
#define init_11_10_15_13
void FASTCALL call_11_10_15_13(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(15<<4)+13);
}
#define init_11_10_15_14
void FASTCALL call_11_10_15_14(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(15<<4)+14);
}
#define init_11_10_15_15
void FASTCALL call_11_10_15_15(SH2_struct* sh)
{
  SH2bsr(sh, (10<<8)+(15<<4)+15);
}
#define init_11_11_0_0
void FASTCALL call_11_11_0_0(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(0<<4)+0);
}
#define init_11_11_0_1
void FASTCALL call_11_11_0_1(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(0<<4)+1);
}
#define init_11_11_0_2
void FASTCALL call_11_11_0_2(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(0<<4)+2);
}
#define init_11_11_0_3
void FASTCALL call_11_11_0_3(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(0<<4)+3);
}
#define init_11_11_0_4
void FASTCALL call_11_11_0_4(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(0<<4)+4);
}
#define init_11_11_0_5
void FASTCALL call_11_11_0_5(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(0<<4)+5);
}
#define init_11_11_0_6
void FASTCALL call_11_11_0_6(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(0<<4)+6);
}
#define init_11_11_0_7
void FASTCALL call_11_11_0_7(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(0<<4)+7);
}
#define init_11_11_0_8
void FASTCALL call_11_11_0_8(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(0<<4)+8);
}
#define init_11_11_0_9
void FASTCALL call_11_11_0_9(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(0<<4)+9);
}
#define init_11_11_0_10
void FASTCALL call_11_11_0_10(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(0<<4)+10);
}
#define init_11_11_0_11
void FASTCALL call_11_11_0_11(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(0<<4)+11);
}
#define init_11_11_0_12
void FASTCALL call_11_11_0_12(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(0<<4)+12);
}
#define init_11_11_0_13
void FASTCALL call_11_11_0_13(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(0<<4)+13);
}
#define init_11_11_0_14
void FASTCALL call_11_11_0_14(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(0<<4)+14);
}
#define init_11_11_0_15
void FASTCALL call_11_11_0_15(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(0<<4)+15);
}
#define init_11_11_1_0
void FASTCALL call_11_11_1_0(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(1<<4)+0);
}
#define init_11_11_1_1
void FASTCALL call_11_11_1_1(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(1<<4)+1);
}
#define init_11_11_1_2
void FASTCALL call_11_11_1_2(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(1<<4)+2);
}
#define init_11_11_1_3
void FASTCALL call_11_11_1_3(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(1<<4)+3);
}
#define init_11_11_1_4
void FASTCALL call_11_11_1_4(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(1<<4)+4);
}
#define init_11_11_1_5
void FASTCALL call_11_11_1_5(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(1<<4)+5);
}
#define init_11_11_1_6
void FASTCALL call_11_11_1_6(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(1<<4)+6);
}
#define init_11_11_1_7
void FASTCALL call_11_11_1_7(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(1<<4)+7);
}
#define init_11_11_1_8
void FASTCALL call_11_11_1_8(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(1<<4)+8);
}
#define init_11_11_1_9
void FASTCALL call_11_11_1_9(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(1<<4)+9);
}
#define init_11_11_1_10
void FASTCALL call_11_11_1_10(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(1<<4)+10);
}
#define init_11_11_1_11
void FASTCALL call_11_11_1_11(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(1<<4)+11);
}
#define init_11_11_1_12
void FASTCALL call_11_11_1_12(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(1<<4)+12);
}
#define init_11_11_1_13
void FASTCALL call_11_11_1_13(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(1<<4)+13);
}
#define init_11_11_1_14
void FASTCALL call_11_11_1_14(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(1<<4)+14);
}
#define init_11_11_1_15
void FASTCALL call_11_11_1_15(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(1<<4)+15);
}
#define init_11_11_2_0
void FASTCALL call_11_11_2_0(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(2<<4)+0);
}
#define init_11_11_2_1
void FASTCALL call_11_11_2_1(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(2<<4)+1);
}
#define init_11_11_2_2
void FASTCALL call_11_11_2_2(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(2<<4)+2);
}
#define init_11_11_2_3
void FASTCALL call_11_11_2_3(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(2<<4)+3);
}
#define init_11_11_2_4
void FASTCALL call_11_11_2_4(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(2<<4)+4);
}
#define init_11_11_2_5
void FASTCALL call_11_11_2_5(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(2<<4)+5);
}
#define init_11_11_2_6
void FASTCALL call_11_11_2_6(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(2<<4)+6);
}
#define init_11_11_2_7
void FASTCALL call_11_11_2_7(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(2<<4)+7);
}
#define init_11_11_2_8
void FASTCALL call_11_11_2_8(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(2<<4)+8);
}
#define init_11_11_2_9
void FASTCALL call_11_11_2_9(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(2<<4)+9);
}
#define init_11_11_2_10
void FASTCALL call_11_11_2_10(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(2<<4)+10);
}
#define init_11_11_2_11
void FASTCALL call_11_11_2_11(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(2<<4)+11);
}
#define init_11_11_2_12
void FASTCALL call_11_11_2_12(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(2<<4)+12);
}
#define init_11_11_2_13
void FASTCALL call_11_11_2_13(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(2<<4)+13);
}
#define init_11_11_2_14
void FASTCALL call_11_11_2_14(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(2<<4)+14);
}
#define init_11_11_2_15
void FASTCALL call_11_11_2_15(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(2<<4)+15);
}
#define init_11_11_3_0
void FASTCALL call_11_11_3_0(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(3<<4)+0);
}
#define init_11_11_3_1
void FASTCALL call_11_11_3_1(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(3<<4)+1);
}
#define init_11_11_3_2
void FASTCALL call_11_11_3_2(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(3<<4)+2);
}
#define init_11_11_3_3
void FASTCALL call_11_11_3_3(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(3<<4)+3);
}
#define init_11_11_3_4
void FASTCALL call_11_11_3_4(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(3<<4)+4);
}
#define init_11_11_3_5
void FASTCALL call_11_11_3_5(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(3<<4)+5);
}
#define init_11_11_3_6
void FASTCALL call_11_11_3_6(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(3<<4)+6);
}
#define init_11_11_3_7
void FASTCALL call_11_11_3_7(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(3<<4)+7);
}
#define init_11_11_3_8
void FASTCALL call_11_11_3_8(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(3<<4)+8);
}
#define init_11_11_3_9
void FASTCALL call_11_11_3_9(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(3<<4)+9);
}
#define init_11_11_3_10
void FASTCALL call_11_11_3_10(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(3<<4)+10);
}
#define init_11_11_3_11
void FASTCALL call_11_11_3_11(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(3<<4)+11);
}
#define init_11_11_3_12
void FASTCALL call_11_11_3_12(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(3<<4)+12);
}
#define init_11_11_3_13
void FASTCALL call_11_11_3_13(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(3<<4)+13);
}
#define init_11_11_3_14
void FASTCALL call_11_11_3_14(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(3<<4)+14);
}
#define init_11_11_3_15
void FASTCALL call_11_11_3_15(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(3<<4)+15);
}
#define init_11_11_4_0
void FASTCALL call_11_11_4_0(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(4<<4)+0);
}
#define init_11_11_4_1
void FASTCALL call_11_11_4_1(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(4<<4)+1);
}
#define init_11_11_4_2
void FASTCALL call_11_11_4_2(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(4<<4)+2);
}
#define init_11_11_4_3
void FASTCALL call_11_11_4_3(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(4<<4)+3);
}
#define init_11_11_4_4
void FASTCALL call_11_11_4_4(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(4<<4)+4);
}
#define init_11_11_4_5
void FASTCALL call_11_11_4_5(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(4<<4)+5);
}
#define init_11_11_4_6
void FASTCALL call_11_11_4_6(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(4<<4)+6);
}
#define init_11_11_4_7
void FASTCALL call_11_11_4_7(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(4<<4)+7);
}
#define init_11_11_4_8
void FASTCALL call_11_11_4_8(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(4<<4)+8);
}
#define init_11_11_4_9
void FASTCALL call_11_11_4_9(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(4<<4)+9);
}
#define init_11_11_4_10
void FASTCALL call_11_11_4_10(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(4<<4)+10);
}
#define init_11_11_4_11
void FASTCALL call_11_11_4_11(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(4<<4)+11);
}
#define init_11_11_4_12
void FASTCALL call_11_11_4_12(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(4<<4)+12);
}
#define init_11_11_4_13
void FASTCALL call_11_11_4_13(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(4<<4)+13);
}
#define init_11_11_4_14
void FASTCALL call_11_11_4_14(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(4<<4)+14);
}
#define init_11_11_4_15
void FASTCALL call_11_11_4_15(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(4<<4)+15);
}
#define init_11_11_5_0
void FASTCALL call_11_11_5_0(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(5<<4)+0);
}
#define init_11_11_5_1
void FASTCALL call_11_11_5_1(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(5<<4)+1);
}
#define init_11_11_5_2
void FASTCALL call_11_11_5_2(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(5<<4)+2);
}
#define init_11_11_5_3
void FASTCALL call_11_11_5_3(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(5<<4)+3);
}
#define init_11_11_5_4
void FASTCALL call_11_11_5_4(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(5<<4)+4);
}
#define init_11_11_5_5
void FASTCALL call_11_11_5_5(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(5<<4)+5);
}
#define init_11_11_5_6
void FASTCALL call_11_11_5_6(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(5<<4)+6);
}
#define init_11_11_5_7
void FASTCALL call_11_11_5_7(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(5<<4)+7);
}
#define init_11_11_5_8
void FASTCALL call_11_11_5_8(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(5<<4)+8);
}
#define init_11_11_5_9
void FASTCALL call_11_11_5_9(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(5<<4)+9);
}
#define init_11_11_5_10
void FASTCALL call_11_11_5_10(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(5<<4)+10);
}
#define init_11_11_5_11
void FASTCALL call_11_11_5_11(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(5<<4)+11);
}
#define init_11_11_5_12
void FASTCALL call_11_11_5_12(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(5<<4)+12);
}
#define init_11_11_5_13
void FASTCALL call_11_11_5_13(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(5<<4)+13);
}
#define init_11_11_5_14
void FASTCALL call_11_11_5_14(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(5<<4)+14);
}
#define init_11_11_5_15
void FASTCALL call_11_11_5_15(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(5<<4)+15);
}
#define init_11_11_6_0
void FASTCALL call_11_11_6_0(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(6<<4)+0);
}
#define init_11_11_6_1
void FASTCALL call_11_11_6_1(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(6<<4)+1);
}
#define init_11_11_6_2
void FASTCALL call_11_11_6_2(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(6<<4)+2);
}
#define init_11_11_6_3
void FASTCALL call_11_11_6_3(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(6<<4)+3);
}
#define init_11_11_6_4
void FASTCALL call_11_11_6_4(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(6<<4)+4);
}
#define init_11_11_6_5
void FASTCALL call_11_11_6_5(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(6<<4)+5);
}
#define init_11_11_6_6
void FASTCALL call_11_11_6_6(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(6<<4)+6);
}
#define init_11_11_6_7
void FASTCALL call_11_11_6_7(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(6<<4)+7);
}
#define init_11_11_6_8
void FASTCALL call_11_11_6_8(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(6<<4)+8);
}
#define init_11_11_6_9
void FASTCALL call_11_11_6_9(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(6<<4)+9);
}
#define init_11_11_6_10
void FASTCALL call_11_11_6_10(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(6<<4)+10);
}
#define init_11_11_6_11
void FASTCALL call_11_11_6_11(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(6<<4)+11);
}
#define init_11_11_6_12
void FASTCALL call_11_11_6_12(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(6<<4)+12);
}
#define init_11_11_6_13
void FASTCALL call_11_11_6_13(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(6<<4)+13);
}
#define init_11_11_6_14
void FASTCALL call_11_11_6_14(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(6<<4)+14);
}
#define init_11_11_6_15
void FASTCALL call_11_11_6_15(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(6<<4)+15);
}
#define init_11_11_7_0
void FASTCALL call_11_11_7_0(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(7<<4)+0);
}
#define init_11_11_7_1
void FASTCALL call_11_11_7_1(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(7<<4)+1);
}
#define init_11_11_7_2
void FASTCALL call_11_11_7_2(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(7<<4)+2);
}
#define init_11_11_7_3
void FASTCALL call_11_11_7_3(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(7<<4)+3);
}
#define init_11_11_7_4
void FASTCALL call_11_11_7_4(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(7<<4)+4);
}
#define init_11_11_7_5
void FASTCALL call_11_11_7_5(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(7<<4)+5);
}
#define init_11_11_7_6
void FASTCALL call_11_11_7_6(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(7<<4)+6);
}
#define init_11_11_7_7
void FASTCALL call_11_11_7_7(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(7<<4)+7);
}
#define init_11_11_7_8
void FASTCALL call_11_11_7_8(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(7<<4)+8);
}
#define init_11_11_7_9
void FASTCALL call_11_11_7_9(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(7<<4)+9);
}
#define init_11_11_7_10
void FASTCALL call_11_11_7_10(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(7<<4)+10);
}
#define init_11_11_7_11
void FASTCALL call_11_11_7_11(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(7<<4)+11);
}
#define init_11_11_7_12
void FASTCALL call_11_11_7_12(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(7<<4)+12);
}
#define init_11_11_7_13
void FASTCALL call_11_11_7_13(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(7<<4)+13);
}
#define init_11_11_7_14
void FASTCALL call_11_11_7_14(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(7<<4)+14);
}
#define init_11_11_7_15
void FASTCALL call_11_11_7_15(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(7<<4)+15);
}
#define init_11_11_8_0
void FASTCALL call_11_11_8_0(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(8<<4)+0);
}
#define init_11_11_8_1
void FASTCALL call_11_11_8_1(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(8<<4)+1);
}
#define init_11_11_8_2
void FASTCALL call_11_11_8_2(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(8<<4)+2);
}
#define init_11_11_8_3
void FASTCALL call_11_11_8_3(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(8<<4)+3);
}
#define init_11_11_8_4
void FASTCALL call_11_11_8_4(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(8<<4)+4);
}
#define init_11_11_8_5
void FASTCALL call_11_11_8_5(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(8<<4)+5);
}
#define init_11_11_8_6
void FASTCALL call_11_11_8_6(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(8<<4)+6);
}
#define init_11_11_8_7
void FASTCALL call_11_11_8_7(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(8<<4)+7);
}
#define init_11_11_8_8
void FASTCALL call_11_11_8_8(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(8<<4)+8);
}
#define init_11_11_8_9
void FASTCALL call_11_11_8_9(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(8<<4)+9);
}
#define init_11_11_8_10
void FASTCALL call_11_11_8_10(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(8<<4)+10);
}
#define init_11_11_8_11
void FASTCALL call_11_11_8_11(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(8<<4)+11);
}
#define init_11_11_8_12
void FASTCALL call_11_11_8_12(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(8<<4)+12);
}
#define init_11_11_8_13
void FASTCALL call_11_11_8_13(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(8<<4)+13);
}
#define init_11_11_8_14
void FASTCALL call_11_11_8_14(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(8<<4)+14);
}
#define init_11_11_8_15
void FASTCALL call_11_11_8_15(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(8<<4)+15);
}
#define init_11_11_9_0
void FASTCALL call_11_11_9_0(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(9<<4)+0);
}
#define init_11_11_9_1
void FASTCALL call_11_11_9_1(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(9<<4)+1);
}
#define init_11_11_9_2
void FASTCALL call_11_11_9_2(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(9<<4)+2);
}
#define init_11_11_9_3
void FASTCALL call_11_11_9_3(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(9<<4)+3);
}
#define init_11_11_9_4
void FASTCALL call_11_11_9_4(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(9<<4)+4);
}
#define init_11_11_9_5
void FASTCALL call_11_11_9_5(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(9<<4)+5);
}
#define init_11_11_9_6
void FASTCALL call_11_11_9_6(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(9<<4)+6);
}
#define init_11_11_9_7
void FASTCALL call_11_11_9_7(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(9<<4)+7);
}
#define init_11_11_9_8
void FASTCALL call_11_11_9_8(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(9<<4)+8);
}
#define init_11_11_9_9
void FASTCALL call_11_11_9_9(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(9<<4)+9);
}
#define init_11_11_9_10
void FASTCALL call_11_11_9_10(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(9<<4)+10);
}
#define init_11_11_9_11
void FASTCALL call_11_11_9_11(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(9<<4)+11);
}
#define init_11_11_9_12
void FASTCALL call_11_11_9_12(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(9<<4)+12);
}
#define init_11_11_9_13
void FASTCALL call_11_11_9_13(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(9<<4)+13);
}
#define init_11_11_9_14
void FASTCALL call_11_11_9_14(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(9<<4)+14);
}
#define init_11_11_9_15
void FASTCALL call_11_11_9_15(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(9<<4)+15);
}
#define init_11_11_10_0
void FASTCALL call_11_11_10_0(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(10<<4)+0);
}
#define init_11_11_10_1
void FASTCALL call_11_11_10_1(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(10<<4)+1);
}
#define init_11_11_10_2
void FASTCALL call_11_11_10_2(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(10<<4)+2);
}
#define init_11_11_10_3
void FASTCALL call_11_11_10_3(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(10<<4)+3);
}
#define init_11_11_10_4
void FASTCALL call_11_11_10_4(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(10<<4)+4);
}
#define init_11_11_10_5
void FASTCALL call_11_11_10_5(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(10<<4)+5);
}
#define init_11_11_10_6
void FASTCALL call_11_11_10_6(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(10<<4)+6);
}
#define init_11_11_10_7
void FASTCALL call_11_11_10_7(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(10<<4)+7);
}
#define init_11_11_10_8
void FASTCALL call_11_11_10_8(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(10<<4)+8);
}
#define init_11_11_10_9
void FASTCALL call_11_11_10_9(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(10<<4)+9);
}
#define init_11_11_10_10
void FASTCALL call_11_11_10_10(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(10<<4)+10);
}
#define init_11_11_10_11
void FASTCALL call_11_11_10_11(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(10<<4)+11);
}
#define init_11_11_10_12
void FASTCALL call_11_11_10_12(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(10<<4)+12);
}
#define init_11_11_10_13
void FASTCALL call_11_11_10_13(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(10<<4)+13);
}
#define init_11_11_10_14
void FASTCALL call_11_11_10_14(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(10<<4)+14);
}
#define init_11_11_10_15
void FASTCALL call_11_11_10_15(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(10<<4)+15);
}
#define init_11_11_11_0
void FASTCALL call_11_11_11_0(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(11<<4)+0);
}
#define init_11_11_11_1
void FASTCALL call_11_11_11_1(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(11<<4)+1);
}
#define init_11_11_11_2
void FASTCALL call_11_11_11_2(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(11<<4)+2);
}
#define init_11_11_11_3
void FASTCALL call_11_11_11_3(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(11<<4)+3);
}
#define init_11_11_11_4
void FASTCALL call_11_11_11_4(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(11<<4)+4);
}
#define init_11_11_11_5
void FASTCALL call_11_11_11_5(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(11<<4)+5);
}
#define init_11_11_11_6
void FASTCALL call_11_11_11_6(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(11<<4)+6);
}
#define init_11_11_11_7
void FASTCALL call_11_11_11_7(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(11<<4)+7);
}
#define init_11_11_11_8
void FASTCALL call_11_11_11_8(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(11<<4)+8);
}
#define init_11_11_11_9
void FASTCALL call_11_11_11_9(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(11<<4)+9);
}
#define init_11_11_11_10
void FASTCALL call_11_11_11_10(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(11<<4)+10);
}
#define init_11_11_11_11
void FASTCALL call_11_11_11_11(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(11<<4)+11);
}
#define init_11_11_11_12
void FASTCALL call_11_11_11_12(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(11<<4)+12);
}
#define init_11_11_11_13
void FASTCALL call_11_11_11_13(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(11<<4)+13);
}
#define init_11_11_11_14
void FASTCALL call_11_11_11_14(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(11<<4)+14);
}
#define init_11_11_11_15
void FASTCALL call_11_11_11_15(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(11<<4)+15);
}
#define init_11_11_12_0
void FASTCALL call_11_11_12_0(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(12<<4)+0);
}
#define init_11_11_12_1
void FASTCALL call_11_11_12_1(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(12<<4)+1);
}
#define init_11_11_12_2
void FASTCALL call_11_11_12_2(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(12<<4)+2);
}
#define init_11_11_12_3
void FASTCALL call_11_11_12_3(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(12<<4)+3);
}
#define init_11_11_12_4
void FASTCALL call_11_11_12_4(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(12<<4)+4);
}
#define init_11_11_12_5
void FASTCALL call_11_11_12_5(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(12<<4)+5);
}
#define init_11_11_12_6
void FASTCALL call_11_11_12_6(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(12<<4)+6);
}
#define init_11_11_12_7
void FASTCALL call_11_11_12_7(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(12<<4)+7);
}
#define init_11_11_12_8
void FASTCALL call_11_11_12_8(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(12<<4)+8);
}
#define init_11_11_12_9
void FASTCALL call_11_11_12_9(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(12<<4)+9);
}
#define init_11_11_12_10
void FASTCALL call_11_11_12_10(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(12<<4)+10);
}
#define init_11_11_12_11
void FASTCALL call_11_11_12_11(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(12<<4)+11);
}
#define init_11_11_12_12
void FASTCALL call_11_11_12_12(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(12<<4)+12);
}
#define init_11_11_12_13
void FASTCALL call_11_11_12_13(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(12<<4)+13);
}
#define init_11_11_12_14
void FASTCALL call_11_11_12_14(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(12<<4)+14);
}
#define init_11_11_12_15
void FASTCALL call_11_11_12_15(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(12<<4)+15);
}
#define init_11_11_13_0
void FASTCALL call_11_11_13_0(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(13<<4)+0);
}
#define init_11_11_13_1
void FASTCALL call_11_11_13_1(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(13<<4)+1);
}
#define init_11_11_13_2
void FASTCALL call_11_11_13_2(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(13<<4)+2);
}
#define init_11_11_13_3
void FASTCALL call_11_11_13_3(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(13<<4)+3);
}
#define init_11_11_13_4
void FASTCALL call_11_11_13_4(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(13<<4)+4);
}
#define init_11_11_13_5
void FASTCALL call_11_11_13_5(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(13<<4)+5);
}
#define init_11_11_13_6
void FASTCALL call_11_11_13_6(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(13<<4)+6);
}
#define init_11_11_13_7
void FASTCALL call_11_11_13_7(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(13<<4)+7);
}
#define init_11_11_13_8
void FASTCALL call_11_11_13_8(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(13<<4)+8);
}
#define init_11_11_13_9
void FASTCALL call_11_11_13_9(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(13<<4)+9);
}
#define init_11_11_13_10
void FASTCALL call_11_11_13_10(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(13<<4)+10);
}
#define init_11_11_13_11
void FASTCALL call_11_11_13_11(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(13<<4)+11);
}
#define init_11_11_13_12
void FASTCALL call_11_11_13_12(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(13<<4)+12);
}
#define init_11_11_13_13
void FASTCALL call_11_11_13_13(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(13<<4)+13);
}
#define init_11_11_13_14
void FASTCALL call_11_11_13_14(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(13<<4)+14);
}
#define init_11_11_13_15
void FASTCALL call_11_11_13_15(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(13<<4)+15);
}
#define init_11_11_14_0
void FASTCALL call_11_11_14_0(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(14<<4)+0);
}
#define init_11_11_14_1
void FASTCALL call_11_11_14_1(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(14<<4)+1);
}
#define init_11_11_14_2
void FASTCALL call_11_11_14_2(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(14<<4)+2);
}
#define init_11_11_14_3
void FASTCALL call_11_11_14_3(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(14<<4)+3);
}
#define init_11_11_14_4
void FASTCALL call_11_11_14_4(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(14<<4)+4);
}
#define init_11_11_14_5
void FASTCALL call_11_11_14_5(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(14<<4)+5);
}
#define init_11_11_14_6
void FASTCALL call_11_11_14_6(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(14<<4)+6);
}
#define init_11_11_14_7
void FASTCALL call_11_11_14_7(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(14<<4)+7);
}
#define init_11_11_14_8
void FASTCALL call_11_11_14_8(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(14<<4)+8);
}
#define init_11_11_14_9
void FASTCALL call_11_11_14_9(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(14<<4)+9);
}
#define init_11_11_14_10
void FASTCALL call_11_11_14_10(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(14<<4)+10);
}
#define init_11_11_14_11
void FASTCALL call_11_11_14_11(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(14<<4)+11);
}
#define init_11_11_14_12
void FASTCALL call_11_11_14_12(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(14<<4)+12);
}
#define init_11_11_14_13
void FASTCALL call_11_11_14_13(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(14<<4)+13);
}
#define init_11_11_14_14
void FASTCALL call_11_11_14_14(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(14<<4)+14);
}
#define init_11_11_14_15
void FASTCALL call_11_11_14_15(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(14<<4)+15);
}
#define init_11_11_15_0
void FASTCALL call_11_11_15_0(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(15<<4)+0);
}
#define init_11_11_15_1
void FASTCALL call_11_11_15_1(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(15<<4)+1);
}
#define init_11_11_15_2
void FASTCALL call_11_11_15_2(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(15<<4)+2);
}
#define init_11_11_15_3
void FASTCALL call_11_11_15_3(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(15<<4)+3);
}
#define init_11_11_15_4
void FASTCALL call_11_11_15_4(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(15<<4)+4);
}
#define init_11_11_15_5
void FASTCALL call_11_11_15_5(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(15<<4)+5);
}
#define init_11_11_15_6
void FASTCALL call_11_11_15_6(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(15<<4)+6);
}
#define init_11_11_15_7
void FASTCALL call_11_11_15_7(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(15<<4)+7);
}
#define init_11_11_15_8
void FASTCALL call_11_11_15_8(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(15<<4)+8);
}
#define init_11_11_15_9
void FASTCALL call_11_11_15_9(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(15<<4)+9);
}
#define init_11_11_15_10
void FASTCALL call_11_11_15_10(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(15<<4)+10);
}
#define init_11_11_15_11
void FASTCALL call_11_11_15_11(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(15<<4)+11);
}
#define init_11_11_15_12
void FASTCALL call_11_11_15_12(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(15<<4)+12);
}
#define init_11_11_15_13
void FASTCALL call_11_11_15_13(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(15<<4)+13);
}
#define init_11_11_15_14
void FASTCALL call_11_11_15_14(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(15<<4)+14);
}
#define init_11_11_15_15
void FASTCALL call_11_11_15_15(SH2_struct* sh)
{
  SH2bsr(sh, (11<<8)+(15<<4)+15);
}
#define init_11_12_0_0
void FASTCALL call_11_12_0_0(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(0<<4)+0);
}
#define init_11_12_0_1
void FASTCALL call_11_12_0_1(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(0<<4)+1);
}
#define init_11_12_0_2
void FASTCALL call_11_12_0_2(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(0<<4)+2);
}
#define init_11_12_0_3
void FASTCALL call_11_12_0_3(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(0<<4)+3);
}
#define init_11_12_0_4
void FASTCALL call_11_12_0_4(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(0<<4)+4);
}
#define init_11_12_0_5
void FASTCALL call_11_12_0_5(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(0<<4)+5);
}
#define init_11_12_0_6
void FASTCALL call_11_12_0_6(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(0<<4)+6);
}
#define init_11_12_0_7
void FASTCALL call_11_12_0_7(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(0<<4)+7);
}
#define init_11_12_0_8
void FASTCALL call_11_12_0_8(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(0<<4)+8);
}
#define init_11_12_0_9
void FASTCALL call_11_12_0_9(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(0<<4)+9);
}
#define init_11_12_0_10
void FASTCALL call_11_12_0_10(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(0<<4)+10);
}
#define init_11_12_0_11
void FASTCALL call_11_12_0_11(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(0<<4)+11);
}
#define init_11_12_0_12
void FASTCALL call_11_12_0_12(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(0<<4)+12);
}
#define init_11_12_0_13
void FASTCALL call_11_12_0_13(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(0<<4)+13);
}
#define init_11_12_0_14
void FASTCALL call_11_12_0_14(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(0<<4)+14);
}
#define init_11_12_0_15
void FASTCALL call_11_12_0_15(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(0<<4)+15);
}
#define init_11_12_1_0
void FASTCALL call_11_12_1_0(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(1<<4)+0);
}
#define init_11_12_1_1
void FASTCALL call_11_12_1_1(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(1<<4)+1);
}
#define init_11_12_1_2
void FASTCALL call_11_12_1_2(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(1<<4)+2);
}
#define init_11_12_1_3
void FASTCALL call_11_12_1_3(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(1<<4)+3);
}
#define init_11_12_1_4
void FASTCALL call_11_12_1_4(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(1<<4)+4);
}
#define init_11_12_1_5
void FASTCALL call_11_12_1_5(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(1<<4)+5);
}
#define init_11_12_1_6
void FASTCALL call_11_12_1_6(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(1<<4)+6);
}
#define init_11_12_1_7
void FASTCALL call_11_12_1_7(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(1<<4)+7);
}
#define init_11_12_1_8
void FASTCALL call_11_12_1_8(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(1<<4)+8);
}
#define init_11_12_1_9
void FASTCALL call_11_12_1_9(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(1<<4)+9);
}
#define init_11_12_1_10
void FASTCALL call_11_12_1_10(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(1<<4)+10);
}
#define init_11_12_1_11
void FASTCALL call_11_12_1_11(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(1<<4)+11);
}
#define init_11_12_1_12
void FASTCALL call_11_12_1_12(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(1<<4)+12);
}
#define init_11_12_1_13
void FASTCALL call_11_12_1_13(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(1<<4)+13);
}
#define init_11_12_1_14
void FASTCALL call_11_12_1_14(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(1<<4)+14);
}
#define init_11_12_1_15
void FASTCALL call_11_12_1_15(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(1<<4)+15);
}
#define init_11_12_2_0
void FASTCALL call_11_12_2_0(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(2<<4)+0);
}
#define init_11_12_2_1
void FASTCALL call_11_12_2_1(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(2<<4)+1);
}
#define init_11_12_2_2
void FASTCALL call_11_12_2_2(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(2<<4)+2);
}
#define init_11_12_2_3
void FASTCALL call_11_12_2_3(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(2<<4)+3);
}
#define init_11_12_2_4
void FASTCALL call_11_12_2_4(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(2<<4)+4);
}
#define init_11_12_2_5
void FASTCALL call_11_12_2_5(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(2<<4)+5);
}
#define init_11_12_2_6
void FASTCALL call_11_12_2_6(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(2<<4)+6);
}
#define init_11_12_2_7
void FASTCALL call_11_12_2_7(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(2<<4)+7);
}
#define init_11_12_2_8
void FASTCALL call_11_12_2_8(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(2<<4)+8);
}
#define init_11_12_2_9
void FASTCALL call_11_12_2_9(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(2<<4)+9);
}
#define init_11_12_2_10
void FASTCALL call_11_12_2_10(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(2<<4)+10);
}
#define init_11_12_2_11
void FASTCALL call_11_12_2_11(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(2<<4)+11);
}
#define init_11_12_2_12
void FASTCALL call_11_12_2_12(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(2<<4)+12);
}
#define init_11_12_2_13
void FASTCALL call_11_12_2_13(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(2<<4)+13);
}
#define init_11_12_2_14
void FASTCALL call_11_12_2_14(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(2<<4)+14);
}
#define init_11_12_2_15
void FASTCALL call_11_12_2_15(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(2<<4)+15);
}
#define init_11_12_3_0
void FASTCALL call_11_12_3_0(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(3<<4)+0);
}
#define init_11_12_3_1
void FASTCALL call_11_12_3_1(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(3<<4)+1);
}
#define init_11_12_3_2
void FASTCALL call_11_12_3_2(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(3<<4)+2);
}
#define init_11_12_3_3
void FASTCALL call_11_12_3_3(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(3<<4)+3);
}
#define init_11_12_3_4
void FASTCALL call_11_12_3_4(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(3<<4)+4);
}
#define init_11_12_3_5
void FASTCALL call_11_12_3_5(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(3<<4)+5);
}
#define init_11_12_3_6
void FASTCALL call_11_12_3_6(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(3<<4)+6);
}
#define init_11_12_3_7
void FASTCALL call_11_12_3_7(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(3<<4)+7);
}
#define init_11_12_3_8
void FASTCALL call_11_12_3_8(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(3<<4)+8);
}
#define init_11_12_3_9
void FASTCALL call_11_12_3_9(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(3<<4)+9);
}
#define init_11_12_3_10
void FASTCALL call_11_12_3_10(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(3<<4)+10);
}
#define init_11_12_3_11
void FASTCALL call_11_12_3_11(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(3<<4)+11);
}
#define init_11_12_3_12
void FASTCALL call_11_12_3_12(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(3<<4)+12);
}
#define init_11_12_3_13
void FASTCALL call_11_12_3_13(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(3<<4)+13);
}
#define init_11_12_3_14
void FASTCALL call_11_12_3_14(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(3<<4)+14);
}
#define init_11_12_3_15
void FASTCALL call_11_12_3_15(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(3<<4)+15);
}
#define init_11_12_4_0
void FASTCALL call_11_12_4_0(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(4<<4)+0);
}
#define init_11_12_4_1
void FASTCALL call_11_12_4_1(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(4<<4)+1);
}
#define init_11_12_4_2
void FASTCALL call_11_12_4_2(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(4<<4)+2);
}
#define init_11_12_4_3
void FASTCALL call_11_12_4_3(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(4<<4)+3);
}
#define init_11_12_4_4
void FASTCALL call_11_12_4_4(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(4<<4)+4);
}
#define init_11_12_4_5
void FASTCALL call_11_12_4_5(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(4<<4)+5);
}
#define init_11_12_4_6
void FASTCALL call_11_12_4_6(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(4<<4)+6);
}
#define init_11_12_4_7
void FASTCALL call_11_12_4_7(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(4<<4)+7);
}
#define init_11_12_4_8
void FASTCALL call_11_12_4_8(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(4<<4)+8);
}
#define init_11_12_4_9
void FASTCALL call_11_12_4_9(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(4<<4)+9);
}
#define init_11_12_4_10
void FASTCALL call_11_12_4_10(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(4<<4)+10);
}
#define init_11_12_4_11
void FASTCALL call_11_12_4_11(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(4<<4)+11);
}
#define init_11_12_4_12
void FASTCALL call_11_12_4_12(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(4<<4)+12);
}
#define init_11_12_4_13
void FASTCALL call_11_12_4_13(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(4<<4)+13);
}
#define init_11_12_4_14
void FASTCALL call_11_12_4_14(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(4<<4)+14);
}
#define init_11_12_4_15
void FASTCALL call_11_12_4_15(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(4<<4)+15);
}
#define init_11_12_5_0
void FASTCALL call_11_12_5_0(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(5<<4)+0);
}
#define init_11_12_5_1
void FASTCALL call_11_12_5_1(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(5<<4)+1);
}
#define init_11_12_5_2
void FASTCALL call_11_12_5_2(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(5<<4)+2);
}
#define init_11_12_5_3
void FASTCALL call_11_12_5_3(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(5<<4)+3);
}
#define init_11_12_5_4
void FASTCALL call_11_12_5_4(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(5<<4)+4);
}
#define init_11_12_5_5
void FASTCALL call_11_12_5_5(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(5<<4)+5);
}
#define init_11_12_5_6
void FASTCALL call_11_12_5_6(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(5<<4)+6);
}
#define init_11_12_5_7
void FASTCALL call_11_12_5_7(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(5<<4)+7);
}
#define init_11_12_5_8
void FASTCALL call_11_12_5_8(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(5<<4)+8);
}
#define init_11_12_5_9
void FASTCALL call_11_12_5_9(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(5<<4)+9);
}
#define init_11_12_5_10
void FASTCALL call_11_12_5_10(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(5<<4)+10);
}
#define init_11_12_5_11
void FASTCALL call_11_12_5_11(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(5<<4)+11);
}
#define init_11_12_5_12
void FASTCALL call_11_12_5_12(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(5<<4)+12);
}
#define init_11_12_5_13
void FASTCALL call_11_12_5_13(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(5<<4)+13);
}
#define init_11_12_5_14
void FASTCALL call_11_12_5_14(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(5<<4)+14);
}
#define init_11_12_5_15
void FASTCALL call_11_12_5_15(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(5<<4)+15);
}
#define init_11_12_6_0
void FASTCALL call_11_12_6_0(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(6<<4)+0);
}
#define init_11_12_6_1
void FASTCALL call_11_12_6_1(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(6<<4)+1);
}
#define init_11_12_6_2
void FASTCALL call_11_12_6_2(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(6<<4)+2);
}
#define init_11_12_6_3
void FASTCALL call_11_12_6_3(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(6<<4)+3);
}
#define init_11_12_6_4
void FASTCALL call_11_12_6_4(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(6<<4)+4);
}
#define init_11_12_6_5
void FASTCALL call_11_12_6_5(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(6<<4)+5);
}
#define init_11_12_6_6
void FASTCALL call_11_12_6_6(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(6<<4)+6);
}
#define init_11_12_6_7
void FASTCALL call_11_12_6_7(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(6<<4)+7);
}
#define init_11_12_6_8
void FASTCALL call_11_12_6_8(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(6<<4)+8);
}
#define init_11_12_6_9
void FASTCALL call_11_12_6_9(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(6<<4)+9);
}
#define init_11_12_6_10
void FASTCALL call_11_12_6_10(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(6<<4)+10);
}
#define init_11_12_6_11
void FASTCALL call_11_12_6_11(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(6<<4)+11);
}
#define init_11_12_6_12
void FASTCALL call_11_12_6_12(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(6<<4)+12);
}
#define init_11_12_6_13
void FASTCALL call_11_12_6_13(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(6<<4)+13);
}
#define init_11_12_6_14
void FASTCALL call_11_12_6_14(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(6<<4)+14);
}
#define init_11_12_6_15
void FASTCALL call_11_12_6_15(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(6<<4)+15);
}
#define init_11_12_7_0
void FASTCALL call_11_12_7_0(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(7<<4)+0);
}
#define init_11_12_7_1
void FASTCALL call_11_12_7_1(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(7<<4)+1);
}
#define init_11_12_7_2
void FASTCALL call_11_12_7_2(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(7<<4)+2);
}
#define init_11_12_7_3
void FASTCALL call_11_12_7_3(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(7<<4)+3);
}
#define init_11_12_7_4
void FASTCALL call_11_12_7_4(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(7<<4)+4);
}
#define init_11_12_7_5
void FASTCALL call_11_12_7_5(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(7<<4)+5);
}
#define init_11_12_7_6
void FASTCALL call_11_12_7_6(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(7<<4)+6);
}
#define init_11_12_7_7
void FASTCALL call_11_12_7_7(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(7<<4)+7);
}
#define init_11_12_7_8
void FASTCALL call_11_12_7_8(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(7<<4)+8);
}
#define init_11_12_7_9
void FASTCALL call_11_12_7_9(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(7<<4)+9);
}
#define init_11_12_7_10
void FASTCALL call_11_12_7_10(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(7<<4)+10);
}
#define init_11_12_7_11
void FASTCALL call_11_12_7_11(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(7<<4)+11);
}
#define init_11_12_7_12
void FASTCALL call_11_12_7_12(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(7<<4)+12);
}
#define init_11_12_7_13
void FASTCALL call_11_12_7_13(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(7<<4)+13);
}
#define init_11_12_7_14
void FASTCALL call_11_12_7_14(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(7<<4)+14);
}
#define init_11_12_7_15
void FASTCALL call_11_12_7_15(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(7<<4)+15);
}
#define init_11_12_8_0
void FASTCALL call_11_12_8_0(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(8<<4)+0);
}
#define init_11_12_8_1
void FASTCALL call_11_12_8_1(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(8<<4)+1);
}
#define init_11_12_8_2
void FASTCALL call_11_12_8_2(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(8<<4)+2);
}
#define init_11_12_8_3
void FASTCALL call_11_12_8_3(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(8<<4)+3);
}
#define init_11_12_8_4
void FASTCALL call_11_12_8_4(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(8<<4)+4);
}
#define init_11_12_8_5
void FASTCALL call_11_12_8_5(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(8<<4)+5);
}
#define init_11_12_8_6
void FASTCALL call_11_12_8_6(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(8<<4)+6);
}
#define init_11_12_8_7
void FASTCALL call_11_12_8_7(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(8<<4)+7);
}
#define init_11_12_8_8
void FASTCALL call_11_12_8_8(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(8<<4)+8);
}
#define init_11_12_8_9
void FASTCALL call_11_12_8_9(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(8<<4)+9);
}
#define init_11_12_8_10
void FASTCALL call_11_12_8_10(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(8<<4)+10);
}
#define init_11_12_8_11
void FASTCALL call_11_12_8_11(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(8<<4)+11);
}
#define init_11_12_8_12
void FASTCALL call_11_12_8_12(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(8<<4)+12);
}
#define init_11_12_8_13
void FASTCALL call_11_12_8_13(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(8<<4)+13);
}
#define init_11_12_8_14
void FASTCALL call_11_12_8_14(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(8<<4)+14);
}
#define init_11_12_8_15
void FASTCALL call_11_12_8_15(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(8<<4)+15);
}
#define init_11_12_9_0
void FASTCALL call_11_12_9_0(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(9<<4)+0);
}
#define init_11_12_9_1
void FASTCALL call_11_12_9_1(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(9<<4)+1);
}
#define init_11_12_9_2
void FASTCALL call_11_12_9_2(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(9<<4)+2);
}
#define init_11_12_9_3
void FASTCALL call_11_12_9_3(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(9<<4)+3);
}
#define init_11_12_9_4
void FASTCALL call_11_12_9_4(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(9<<4)+4);
}
#define init_11_12_9_5
void FASTCALL call_11_12_9_5(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(9<<4)+5);
}
#define init_11_12_9_6
void FASTCALL call_11_12_9_6(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(9<<4)+6);
}
#define init_11_12_9_7
void FASTCALL call_11_12_9_7(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(9<<4)+7);
}
#define init_11_12_9_8
void FASTCALL call_11_12_9_8(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(9<<4)+8);
}
#define init_11_12_9_9
void FASTCALL call_11_12_9_9(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(9<<4)+9);
}
#define init_11_12_9_10
void FASTCALL call_11_12_9_10(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(9<<4)+10);
}
#define init_11_12_9_11
void FASTCALL call_11_12_9_11(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(9<<4)+11);
}
#define init_11_12_9_12
void FASTCALL call_11_12_9_12(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(9<<4)+12);
}
#define init_11_12_9_13
void FASTCALL call_11_12_9_13(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(9<<4)+13);
}
#define init_11_12_9_14
void FASTCALL call_11_12_9_14(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(9<<4)+14);
}
#define init_11_12_9_15
void FASTCALL call_11_12_9_15(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(9<<4)+15);
}
#define init_11_12_10_0
void FASTCALL call_11_12_10_0(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(10<<4)+0);
}
#define init_11_12_10_1
void FASTCALL call_11_12_10_1(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(10<<4)+1);
}
#define init_11_12_10_2
void FASTCALL call_11_12_10_2(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(10<<4)+2);
}
#define init_11_12_10_3
void FASTCALL call_11_12_10_3(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(10<<4)+3);
}
#define init_11_12_10_4
void FASTCALL call_11_12_10_4(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(10<<4)+4);
}
#define init_11_12_10_5
void FASTCALL call_11_12_10_5(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(10<<4)+5);
}
#define init_11_12_10_6
void FASTCALL call_11_12_10_6(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(10<<4)+6);
}
#define init_11_12_10_7
void FASTCALL call_11_12_10_7(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(10<<4)+7);
}
#define init_11_12_10_8
void FASTCALL call_11_12_10_8(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(10<<4)+8);
}
#define init_11_12_10_9
void FASTCALL call_11_12_10_9(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(10<<4)+9);
}
#define init_11_12_10_10
void FASTCALL call_11_12_10_10(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(10<<4)+10);
}
#define init_11_12_10_11
void FASTCALL call_11_12_10_11(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(10<<4)+11);
}
#define init_11_12_10_12
void FASTCALL call_11_12_10_12(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(10<<4)+12);
}
#define init_11_12_10_13
void FASTCALL call_11_12_10_13(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(10<<4)+13);
}
#define init_11_12_10_14
void FASTCALL call_11_12_10_14(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(10<<4)+14);
}
#define init_11_12_10_15
void FASTCALL call_11_12_10_15(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(10<<4)+15);
}
#define init_11_12_11_0
void FASTCALL call_11_12_11_0(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(11<<4)+0);
}
#define init_11_12_11_1
void FASTCALL call_11_12_11_1(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(11<<4)+1);
}
#define init_11_12_11_2
void FASTCALL call_11_12_11_2(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(11<<4)+2);
}
#define init_11_12_11_3
void FASTCALL call_11_12_11_3(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(11<<4)+3);
}
#define init_11_12_11_4
void FASTCALL call_11_12_11_4(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(11<<4)+4);
}
#define init_11_12_11_5
void FASTCALL call_11_12_11_5(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(11<<4)+5);
}
#define init_11_12_11_6
void FASTCALL call_11_12_11_6(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(11<<4)+6);
}
#define init_11_12_11_7
void FASTCALL call_11_12_11_7(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(11<<4)+7);
}
#define init_11_12_11_8
void FASTCALL call_11_12_11_8(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(11<<4)+8);
}
#define init_11_12_11_9
void FASTCALL call_11_12_11_9(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(11<<4)+9);
}
#define init_11_12_11_10
void FASTCALL call_11_12_11_10(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(11<<4)+10);
}
#define init_11_12_11_11
void FASTCALL call_11_12_11_11(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(11<<4)+11);
}
#define init_11_12_11_12
void FASTCALL call_11_12_11_12(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(11<<4)+12);
}
#define init_11_12_11_13
void FASTCALL call_11_12_11_13(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(11<<4)+13);
}
#define init_11_12_11_14
void FASTCALL call_11_12_11_14(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(11<<4)+14);
}
#define init_11_12_11_15
void FASTCALL call_11_12_11_15(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(11<<4)+15);
}
#define init_11_12_12_0
void FASTCALL call_11_12_12_0(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(12<<4)+0);
}
#define init_11_12_12_1
void FASTCALL call_11_12_12_1(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(12<<4)+1);
}
#define init_11_12_12_2
void FASTCALL call_11_12_12_2(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(12<<4)+2);
}
#define init_11_12_12_3
void FASTCALL call_11_12_12_3(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(12<<4)+3);
}
#define init_11_12_12_4
void FASTCALL call_11_12_12_4(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(12<<4)+4);
}
#define init_11_12_12_5
void FASTCALL call_11_12_12_5(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(12<<4)+5);
}
#define init_11_12_12_6
void FASTCALL call_11_12_12_6(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(12<<4)+6);
}
#define init_11_12_12_7
void FASTCALL call_11_12_12_7(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(12<<4)+7);
}
#define init_11_12_12_8
void FASTCALL call_11_12_12_8(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(12<<4)+8);
}
#define init_11_12_12_9
void FASTCALL call_11_12_12_9(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(12<<4)+9);
}
#define init_11_12_12_10
void FASTCALL call_11_12_12_10(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(12<<4)+10);
}
#define init_11_12_12_11
void FASTCALL call_11_12_12_11(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(12<<4)+11);
}
#define init_11_12_12_12
void FASTCALL call_11_12_12_12(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(12<<4)+12);
}
#define init_11_12_12_13
void FASTCALL call_11_12_12_13(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(12<<4)+13);
}
#define init_11_12_12_14
void FASTCALL call_11_12_12_14(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(12<<4)+14);
}
#define init_11_12_12_15
void FASTCALL call_11_12_12_15(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(12<<4)+15);
}
#define init_11_12_13_0
void FASTCALL call_11_12_13_0(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(13<<4)+0);
}
#define init_11_12_13_1
void FASTCALL call_11_12_13_1(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(13<<4)+1);
}
#define init_11_12_13_2
void FASTCALL call_11_12_13_2(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(13<<4)+2);
}
#define init_11_12_13_3
void FASTCALL call_11_12_13_3(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(13<<4)+3);
}
#define init_11_12_13_4
void FASTCALL call_11_12_13_4(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(13<<4)+4);
}
#define init_11_12_13_5
void FASTCALL call_11_12_13_5(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(13<<4)+5);
}
#define init_11_12_13_6
void FASTCALL call_11_12_13_6(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(13<<4)+6);
}
#define init_11_12_13_7
void FASTCALL call_11_12_13_7(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(13<<4)+7);
}
#define init_11_12_13_8
void FASTCALL call_11_12_13_8(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(13<<4)+8);
}
#define init_11_12_13_9
void FASTCALL call_11_12_13_9(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(13<<4)+9);
}
#define init_11_12_13_10
void FASTCALL call_11_12_13_10(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(13<<4)+10);
}
#define init_11_12_13_11
void FASTCALL call_11_12_13_11(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(13<<4)+11);
}
#define init_11_12_13_12
void FASTCALL call_11_12_13_12(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(13<<4)+12);
}
#define init_11_12_13_13
void FASTCALL call_11_12_13_13(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(13<<4)+13);
}
#define init_11_12_13_14
void FASTCALL call_11_12_13_14(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(13<<4)+14);
}
#define init_11_12_13_15
void FASTCALL call_11_12_13_15(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(13<<4)+15);
}
#define init_11_12_14_0
void FASTCALL call_11_12_14_0(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(14<<4)+0);
}
#define init_11_12_14_1
void FASTCALL call_11_12_14_1(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(14<<4)+1);
}
#define init_11_12_14_2
void FASTCALL call_11_12_14_2(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(14<<4)+2);
}
#define init_11_12_14_3
void FASTCALL call_11_12_14_3(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(14<<4)+3);
}
#define init_11_12_14_4
void FASTCALL call_11_12_14_4(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(14<<4)+4);
}
#define init_11_12_14_5
void FASTCALL call_11_12_14_5(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(14<<4)+5);
}
#define init_11_12_14_6
void FASTCALL call_11_12_14_6(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(14<<4)+6);
}
#define init_11_12_14_7
void FASTCALL call_11_12_14_7(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(14<<4)+7);
}
#define init_11_12_14_8
void FASTCALL call_11_12_14_8(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(14<<4)+8);
}
#define init_11_12_14_9
void FASTCALL call_11_12_14_9(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(14<<4)+9);
}
#define init_11_12_14_10
void FASTCALL call_11_12_14_10(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(14<<4)+10);
}
#define init_11_12_14_11
void FASTCALL call_11_12_14_11(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(14<<4)+11);
}
#define init_11_12_14_12
void FASTCALL call_11_12_14_12(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(14<<4)+12);
}
#define init_11_12_14_13
void FASTCALL call_11_12_14_13(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(14<<4)+13);
}
#define init_11_12_14_14
void FASTCALL call_11_12_14_14(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(14<<4)+14);
}
#define init_11_12_14_15
void FASTCALL call_11_12_14_15(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(14<<4)+15);
}
#define init_11_12_15_0
void FASTCALL call_11_12_15_0(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(15<<4)+0);
}
#define init_11_12_15_1
void FASTCALL call_11_12_15_1(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(15<<4)+1);
}
#define init_11_12_15_2
void FASTCALL call_11_12_15_2(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(15<<4)+2);
}
#define init_11_12_15_3
void FASTCALL call_11_12_15_3(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(15<<4)+3);
}
#define init_11_12_15_4
void FASTCALL call_11_12_15_4(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(15<<4)+4);
}
#define init_11_12_15_5
void FASTCALL call_11_12_15_5(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(15<<4)+5);
}
#define init_11_12_15_6
void FASTCALL call_11_12_15_6(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(15<<4)+6);
}
#define init_11_12_15_7
void FASTCALL call_11_12_15_7(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(15<<4)+7);
}
#define init_11_12_15_8
void FASTCALL call_11_12_15_8(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(15<<4)+8);
}
#define init_11_12_15_9
void FASTCALL call_11_12_15_9(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(15<<4)+9);
}
#define init_11_12_15_10
void FASTCALL call_11_12_15_10(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(15<<4)+10);
}
#define init_11_12_15_11
void FASTCALL call_11_12_15_11(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(15<<4)+11);
}
#define init_11_12_15_12
void FASTCALL call_11_12_15_12(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(15<<4)+12);
}
#define init_11_12_15_13
void FASTCALL call_11_12_15_13(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(15<<4)+13);
}
#define init_11_12_15_14
void FASTCALL call_11_12_15_14(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(15<<4)+14);
}
#define init_11_12_15_15
void FASTCALL call_11_12_15_15(SH2_struct* sh)
{
  SH2bsr(sh, (12<<8)+(15<<4)+15);
}
#define init_11_13_0_0
void FASTCALL call_11_13_0_0(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(0<<4)+0);
}
#define init_11_13_0_1
void FASTCALL call_11_13_0_1(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(0<<4)+1);
}
#define init_11_13_0_2
void FASTCALL call_11_13_0_2(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(0<<4)+2);
}
#define init_11_13_0_3
void FASTCALL call_11_13_0_3(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(0<<4)+3);
}
#define init_11_13_0_4
void FASTCALL call_11_13_0_4(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(0<<4)+4);
}
#define init_11_13_0_5
void FASTCALL call_11_13_0_5(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(0<<4)+5);
}
#define init_11_13_0_6
void FASTCALL call_11_13_0_6(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(0<<4)+6);
}
#define init_11_13_0_7
void FASTCALL call_11_13_0_7(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(0<<4)+7);
}
#define init_11_13_0_8
void FASTCALL call_11_13_0_8(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(0<<4)+8);
}
#define init_11_13_0_9
void FASTCALL call_11_13_0_9(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(0<<4)+9);
}
#define init_11_13_0_10
void FASTCALL call_11_13_0_10(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(0<<4)+10);
}
#define init_11_13_0_11
void FASTCALL call_11_13_0_11(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(0<<4)+11);
}
#define init_11_13_0_12
void FASTCALL call_11_13_0_12(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(0<<4)+12);
}
#define init_11_13_0_13
void FASTCALL call_11_13_0_13(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(0<<4)+13);
}
#define init_11_13_0_14
void FASTCALL call_11_13_0_14(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(0<<4)+14);
}
#define init_11_13_0_15
void FASTCALL call_11_13_0_15(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(0<<4)+15);
}
#define init_11_13_1_0
void FASTCALL call_11_13_1_0(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(1<<4)+0);
}
#define init_11_13_1_1
void FASTCALL call_11_13_1_1(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(1<<4)+1);
}
#define init_11_13_1_2
void FASTCALL call_11_13_1_2(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(1<<4)+2);
}
#define init_11_13_1_3
void FASTCALL call_11_13_1_3(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(1<<4)+3);
}
#define init_11_13_1_4
void FASTCALL call_11_13_1_4(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(1<<4)+4);
}
#define init_11_13_1_5
void FASTCALL call_11_13_1_5(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(1<<4)+5);
}
#define init_11_13_1_6
void FASTCALL call_11_13_1_6(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(1<<4)+6);
}
#define init_11_13_1_7
void FASTCALL call_11_13_1_7(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(1<<4)+7);
}
#define init_11_13_1_8
void FASTCALL call_11_13_1_8(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(1<<4)+8);
}
#define init_11_13_1_9
void FASTCALL call_11_13_1_9(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(1<<4)+9);
}
#define init_11_13_1_10
void FASTCALL call_11_13_1_10(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(1<<4)+10);
}
#define init_11_13_1_11
void FASTCALL call_11_13_1_11(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(1<<4)+11);
}
#define init_11_13_1_12
void FASTCALL call_11_13_1_12(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(1<<4)+12);
}
#define init_11_13_1_13
void FASTCALL call_11_13_1_13(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(1<<4)+13);
}
#define init_11_13_1_14
void FASTCALL call_11_13_1_14(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(1<<4)+14);
}
#define init_11_13_1_15
void FASTCALL call_11_13_1_15(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(1<<4)+15);
}
#define init_11_13_2_0
void FASTCALL call_11_13_2_0(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(2<<4)+0);
}
#define init_11_13_2_1
void FASTCALL call_11_13_2_1(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(2<<4)+1);
}
#define init_11_13_2_2
void FASTCALL call_11_13_2_2(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(2<<4)+2);
}
#define init_11_13_2_3
void FASTCALL call_11_13_2_3(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(2<<4)+3);
}
#define init_11_13_2_4
void FASTCALL call_11_13_2_4(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(2<<4)+4);
}
#define init_11_13_2_5
void FASTCALL call_11_13_2_5(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(2<<4)+5);
}
#define init_11_13_2_6
void FASTCALL call_11_13_2_6(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(2<<4)+6);
}
#define init_11_13_2_7
void FASTCALL call_11_13_2_7(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(2<<4)+7);
}
#define init_11_13_2_8
void FASTCALL call_11_13_2_8(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(2<<4)+8);
}
#define init_11_13_2_9
void FASTCALL call_11_13_2_9(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(2<<4)+9);
}
#define init_11_13_2_10
void FASTCALL call_11_13_2_10(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(2<<4)+10);
}
#define init_11_13_2_11
void FASTCALL call_11_13_2_11(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(2<<4)+11);
}
#define init_11_13_2_12
void FASTCALL call_11_13_2_12(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(2<<4)+12);
}
#define init_11_13_2_13
void FASTCALL call_11_13_2_13(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(2<<4)+13);
}
#define init_11_13_2_14
void FASTCALL call_11_13_2_14(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(2<<4)+14);
}
#define init_11_13_2_15
void FASTCALL call_11_13_2_15(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(2<<4)+15);
}
#define init_11_13_3_0
void FASTCALL call_11_13_3_0(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(3<<4)+0);
}
#define init_11_13_3_1
void FASTCALL call_11_13_3_1(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(3<<4)+1);
}
#define init_11_13_3_2
void FASTCALL call_11_13_3_2(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(3<<4)+2);
}
#define init_11_13_3_3
void FASTCALL call_11_13_3_3(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(3<<4)+3);
}
#define init_11_13_3_4
void FASTCALL call_11_13_3_4(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(3<<4)+4);
}
#define init_11_13_3_5
void FASTCALL call_11_13_3_5(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(3<<4)+5);
}
#define init_11_13_3_6
void FASTCALL call_11_13_3_6(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(3<<4)+6);
}
#define init_11_13_3_7
void FASTCALL call_11_13_3_7(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(3<<4)+7);
}
#define init_11_13_3_8
void FASTCALL call_11_13_3_8(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(3<<4)+8);
}
#define init_11_13_3_9
void FASTCALL call_11_13_3_9(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(3<<4)+9);
}
#define init_11_13_3_10
void FASTCALL call_11_13_3_10(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(3<<4)+10);
}
#define init_11_13_3_11
void FASTCALL call_11_13_3_11(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(3<<4)+11);
}
#define init_11_13_3_12
void FASTCALL call_11_13_3_12(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(3<<4)+12);
}
#define init_11_13_3_13
void FASTCALL call_11_13_3_13(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(3<<4)+13);
}
#define init_11_13_3_14
void FASTCALL call_11_13_3_14(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(3<<4)+14);
}
#define init_11_13_3_15
void FASTCALL call_11_13_3_15(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(3<<4)+15);
}
#define init_11_13_4_0
void FASTCALL call_11_13_4_0(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(4<<4)+0);
}
#define init_11_13_4_1
void FASTCALL call_11_13_4_1(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(4<<4)+1);
}
#define init_11_13_4_2
void FASTCALL call_11_13_4_2(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(4<<4)+2);
}
#define init_11_13_4_3
void FASTCALL call_11_13_4_3(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(4<<4)+3);
}
#define init_11_13_4_4
void FASTCALL call_11_13_4_4(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(4<<4)+4);
}
#define init_11_13_4_5
void FASTCALL call_11_13_4_5(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(4<<4)+5);
}
#define init_11_13_4_6
void FASTCALL call_11_13_4_6(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(4<<4)+6);
}
#define init_11_13_4_7
void FASTCALL call_11_13_4_7(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(4<<4)+7);
}
#define init_11_13_4_8
void FASTCALL call_11_13_4_8(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(4<<4)+8);
}
#define init_11_13_4_9
void FASTCALL call_11_13_4_9(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(4<<4)+9);
}
#define init_11_13_4_10
void FASTCALL call_11_13_4_10(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(4<<4)+10);
}
#define init_11_13_4_11
void FASTCALL call_11_13_4_11(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(4<<4)+11);
}
#define init_11_13_4_12
void FASTCALL call_11_13_4_12(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(4<<4)+12);
}
#define init_11_13_4_13
void FASTCALL call_11_13_4_13(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(4<<4)+13);
}
#define init_11_13_4_14
void FASTCALL call_11_13_4_14(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(4<<4)+14);
}
#define init_11_13_4_15
void FASTCALL call_11_13_4_15(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(4<<4)+15);
}
#define init_11_13_5_0
void FASTCALL call_11_13_5_0(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(5<<4)+0);
}
#define init_11_13_5_1
void FASTCALL call_11_13_5_1(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(5<<4)+1);
}
#define init_11_13_5_2
void FASTCALL call_11_13_5_2(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(5<<4)+2);
}
#define init_11_13_5_3
void FASTCALL call_11_13_5_3(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(5<<4)+3);
}
#define init_11_13_5_4
void FASTCALL call_11_13_5_4(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(5<<4)+4);
}
#define init_11_13_5_5
void FASTCALL call_11_13_5_5(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(5<<4)+5);
}
#define init_11_13_5_6
void FASTCALL call_11_13_5_6(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(5<<4)+6);
}
#define init_11_13_5_7
void FASTCALL call_11_13_5_7(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(5<<4)+7);
}
#define init_11_13_5_8
void FASTCALL call_11_13_5_8(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(5<<4)+8);
}
#define init_11_13_5_9
void FASTCALL call_11_13_5_9(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(5<<4)+9);
}
#define init_11_13_5_10
void FASTCALL call_11_13_5_10(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(5<<4)+10);
}
#define init_11_13_5_11
void FASTCALL call_11_13_5_11(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(5<<4)+11);
}
#define init_11_13_5_12
void FASTCALL call_11_13_5_12(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(5<<4)+12);
}
#define init_11_13_5_13
void FASTCALL call_11_13_5_13(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(5<<4)+13);
}
#define init_11_13_5_14
void FASTCALL call_11_13_5_14(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(5<<4)+14);
}
#define init_11_13_5_15
void FASTCALL call_11_13_5_15(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(5<<4)+15);
}
#define init_11_13_6_0
void FASTCALL call_11_13_6_0(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(6<<4)+0);
}
#define init_11_13_6_1
void FASTCALL call_11_13_6_1(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(6<<4)+1);
}
#define init_11_13_6_2
void FASTCALL call_11_13_6_2(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(6<<4)+2);
}
#define init_11_13_6_3
void FASTCALL call_11_13_6_3(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(6<<4)+3);
}
#define init_11_13_6_4
void FASTCALL call_11_13_6_4(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(6<<4)+4);
}
#define init_11_13_6_5
void FASTCALL call_11_13_6_5(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(6<<4)+5);
}
#define init_11_13_6_6
void FASTCALL call_11_13_6_6(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(6<<4)+6);
}
#define init_11_13_6_7
void FASTCALL call_11_13_6_7(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(6<<4)+7);
}
#define init_11_13_6_8
void FASTCALL call_11_13_6_8(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(6<<4)+8);
}
#define init_11_13_6_9
void FASTCALL call_11_13_6_9(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(6<<4)+9);
}
#define init_11_13_6_10
void FASTCALL call_11_13_6_10(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(6<<4)+10);
}
#define init_11_13_6_11
void FASTCALL call_11_13_6_11(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(6<<4)+11);
}
#define init_11_13_6_12
void FASTCALL call_11_13_6_12(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(6<<4)+12);
}
#define init_11_13_6_13
void FASTCALL call_11_13_6_13(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(6<<4)+13);
}
#define init_11_13_6_14
void FASTCALL call_11_13_6_14(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(6<<4)+14);
}
#define init_11_13_6_15
void FASTCALL call_11_13_6_15(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(6<<4)+15);
}
#define init_11_13_7_0
void FASTCALL call_11_13_7_0(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(7<<4)+0);
}
#define init_11_13_7_1
void FASTCALL call_11_13_7_1(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(7<<4)+1);
}
#define init_11_13_7_2
void FASTCALL call_11_13_7_2(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(7<<4)+2);
}
#define init_11_13_7_3
void FASTCALL call_11_13_7_3(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(7<<4)+3);
}
#define init_11_13_7_4
void FASTCALL call_11_13_7_4(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(7<<4)+4);
}
#define init_11_13_7_5
void FASTCALL call_11_13_7_5(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(7<<4)+5);
}
#define init_11_13_7_6
void FASTCALL call_11_13_7_6(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(7<<4)+6);
}
#define init_11_13_7_7
void FASTCALL call_11_13_7_7(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(7<<4)+7);
}
#define init_11_13_7_8
void FASTCALL call_11_13_7_8(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(7<<4)+8);
}
#define init_11_13_7_9
void FASTCALL call_11_13_7_9(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(7<<4)+9);
}
#define init_11_13_7_10
void FASTCALL call_11_13_7_10(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(7<<4)+10);
}
#define init_11_13_7_11
void FASTCALL call_11_13_7_11(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(7<<4)+11);
}
#define init_11_13_7_12
void FASTCALL call_11_13_7_12(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(7<<4)+12);
}
#define init_11_13_7_13
void FASTCALL call_11_13_7_13(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(7<<4)+13);
}
#define init_11_13_7_14
void FASTCALL call_11_13_7_14(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(7<<4)+14);
}
#define init_11_13_7_15
void FASTCALL call_11_13_7_15(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(7<<4)+15);
}
#define init_11_13_8_0
void FASTCALL call_11_13_8_0(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(8<<4)+0);
}
#define init_11_13_8_1
void FASTCALL call_11_13_8_1(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(8<<4)+1);
}
#define init_11_13_8_2
void FASTCALL call_11_13_8_2(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(8<<4)+2);
}
#define init_11_13_8_3
void FASTCALL call_11_13_8_3(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(8<<4)+3);
}
#define init_11_13_8_4
void FASTCALL call_11_13_8_4(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(8<<4)+4);
}
#define init_11_13_8_5
void FASTCALL call_11_13_8_5(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(8<<4)+5);
}
#define init_11_13_8_6
void FASTCALL call_11_13_8_6(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(8<<4)+6);
}
#define init_11_13_8_7
void FASTCALL call_11_13_8_7(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(8<<4)+7);
}
#define init_11_13_8_8
void FASTCALL call_11_13_8_8(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(8<<4)+8);
}
#define init_11_13_8_9
void FASTCALL call_11_13_8_9(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(8<<4)+9);
}
#define init_11_13_8_10
void FASTCALL call_11_13_8_10(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(8<<4)+10);
}
#define init_11_13_8_11
void FASTCALL call_11_13_8_11(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(8<<4)+11);
}
#define init_11_13_8_12
void FASTCALL call_11_13_8_12(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(8<<4)+12);
}
#define init_11_13_8_13
void FASTCALL call_11_13_8_13(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(8<<4)+13);
}
#define init_11_13_8_14
void FASTCALL call_11_13_8_14(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(8<<4)+14);
}
#define init_11_13_8_15
void FASTCALL call_11_13_8_15(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(8<<4)+15);
}
#define init_11_13_9_0
void FASTCALL call_11_13_9_0(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(9<<4)+0);
}
#define init_11_13_9_1
void FASTCALL call_11_13_9_1(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(9<<4)+1);
}
#define init_11_13_9_2
void FASTCALL call_11_13_9_2(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(9<<4)+2);
}
#define init_11_13_9_3
void FASTCALL call_11_13_9_3(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(9<<4)+3);
}
#define init_11_13_9_4
void FASTCALL call_11_13_9_4(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(9<<4)+4);
}
#define init_11_13_9_5
void FASTCALL call_11_13_9_5(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(9<<4)+5);
}
#define init_11_13_9_6
void FASTCALL call_11_13_9_6(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(9<<4)+6);
}
#define init_11_13_9_7
void FASTCALL call_11_13_9_7(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(9<<4)+7);
}
#define init_11_13_9_8
void FASTCALL call_11_13_9_8(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(9<<4)+8);
}
#define init_11_13_9_9
void FASTCALL call_11_13_9_9(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(9<<4)+9);
}
#define init_11_13_9_10
void FASTCALL call_11_13_9_10(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(9<<4)+10);
}
#define init_11_13_9_11
void FASTCALL call_11_13_9_11(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(9<<4)+11);
}
#define init_11_13_9_12
void FASTCALL call_11_13_9_12(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(9<<4)+12);
}
#define init_11_13_9_13
void FASTCALL call_11_13_9_13(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(9<<4)+13);
}
#define init_11_13_9_14
void FASTCALL call_11_13_9_14(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(9<<4)+14);
}
#define init_11_13_9_15
void FASTCALL call_11_13_9_15(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(9<<4)+15);
}
#define init_11_13_10_0
void FASTCALL call_11_13_10_0(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(10<<4)+0);
}
#define init_11_13_10_1
void FASTCALL call_11_13_10_1(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(10<<4)+1);
}
#define init_11_13_10_2
void FASTCALL call_11_13_10_2(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(10<<4)+2);
}
#define init_11_13_10_3
void FASTCALL call_11_13_10_3(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(10<<4)+3);
}
#define init_11_13_10_4
void FASTCALL call_11_13_10_4(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(10<<4)+4);
}
#define init_11_13_10_5
void FASTCALL call_11_13_10_5(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(10<<4)+5);
}
#define init_11_13_10_6
void FASTCALL call_11_13_10_6(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(10<<4)+6);
}
#define init_11_13_10_7
void FASTCALL call_11_13_10_7(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(10<<4)+7);
}
#define init_11_13_10_8
void FASTCALL call_11_13_10_8(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(10<<4)+8);
}
#define init_11_13_10_9
void FASTCALL call_11_13_10_9(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(10<<4)+9);
}
#define init_11_13_10_10
void FASTCALL call_11_13_10_10(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(10<<4)+10);
}
#define init_11_13_10_11
void FASTCALL call_11_13_10_11(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(10<<4)+11);
}
#define init_11_13_10_12
void FASTCALL call_11_13_10_12(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(10<<4)+12);
}
#define init_11_13_10_13
void FASTCALL call_11_13_10_13(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(10<<4)+13);
}
#define init_11_13_10_14
void FASTCALL call_11_13_10_14(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(10<<4)+14);
}
#define init_11_13_10_15
void FASTCALL call_11_13_10_15(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(10<<4)+15);
}
#define init_11_13_11_0
void FASTCALL call_11_13_11_0(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(11<<4)+0);
}
#define init_11_13_11_1
void FASTCALL call_11_13_11_1(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(11<<4)+1);
}
#define init_11_13_11_2
void FASTCALL call_11_13_11_2(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(11<<4)+2);
}
#define init_11_13_11_3
void FASTCALL call_11_13_11_3(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(11<<4)+3);
}
#define init_11_13_11_4
void FASTCALL call_11_13_11_4(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(11<<4)+4);
}
#define init_11_13_11_5
void FASTCALL call_11_13_11_5(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(11<<4)+5);
}
#define init_11_13_11_6
void FASTCALL call_11_13_11_6(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(11<<4)+6);
}
#define init_11_13_11_7
void FASTCALL call_11_13_11_7(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(11<<4)+7);
}
#define init_11_13_11_8
void FASTCALL call_11_13_11_8(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(11<<4)+8);
}
#define init_11_13_11_9
void FASTCALL call_11_13_11_9(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(11<<4)+9);
}
#define init_11_13_11_10
void FASTCALL call_11_13_11_10(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(11<<4)+10);
}
#define init_11_13_11_11
void FASTCALL call_11_13_11_11(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(11<<4)+11);
}
#define init_11_13_11_12
void FASTCALL call_11_13_11_12(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(11<<4)+12);
}
#define init_11_13_11_13
void FASTCALL call_11_13_11_13(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(11<<4)+13);
}
#define init_11_13_11_14
void FASTCALL call_11_13_11_14(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(11<<4)+14);
}
#define init_11_13_11_15
void FASTCALL call_11_13_11_15(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(11<<4)+15);
}
#define init_11_13_12_0
void FASTCALL call_11_13_12_0(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(12<<4)+0);
}
#define init_11_13_12_1
void FASTCALL call_11_13_12_1(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(12<<4)+1);
}
#define init_11_13_12_2
void FASTCALL call_11_13_12_2(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(12<<4)+2);
}
#define init_11_13_12_3
void FASTCALL call_11_13_12_3(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(12<<4)+3);
}
#define init_11_13_12_4
void FASTCALL call_11_13_12_4(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(12<<4)+4);
}
#define init_11_13_12_5
void FASTCALL call_11_13_12_5(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(12<<4)+5);
}
#define init_11_13_12_6
void FASTCALL call_11_13_12_6(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(12<<4)+6);
}
#define init_11_13_12_7
void FASTCALL call_11_13_12_7(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(12<<4)+7);
}
#define init_11_13_12_8
void FASTCALL call_11_13_12_8(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(12<<4)+8);
}
#define init_11_13_12_9
void FASTCALL call_11_13_12_9(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(12<<4)+9);
}
#define init_11_13_12_10
void FASTCALL call_11_13_12_10(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(12<<4)+10);
}
#define init_11_13_12_11
void FASTCALL call_11_13_12_11(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(12<<4)+11);
}
#define init_11_13_12_12
void FASTCALL call_11_13_12_12(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(12<<4)+12);
}
#define init_11_13_12_13
void FASTCALL call_11_13_12_13(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(12<<4)+13);
}
#define init_11_13_12_14
void FASTCALL call_11_13_12_14(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(12<<4)+14);
}
#define init_11_13_12_15
void FASTCALL call_11_13_12_15(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(12<<4)+15);
}
#define init_11_13_13_0
void FASTCALL call_11_13_13_0(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(13<<4)+0);
}
#define init_11_13_13_1
void FASTCALL call_11_13_13_1(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(13<<4)+1);
}
#define init_11_13_13_2
void FASTCALL call_11_13_13_2(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(13<<4)+2);
}
#define init_11_13_13_3
void FASTCALL call_11_13_13_3(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(13<<4)+3);
}
#define init_11_13_13_4
void FASTCALL call_11_13_13_4(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(13<<4)+4);
}
#define init_11_13_13_5
void FASTCALL call_11_13_13_5(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(13<<4)+5);
}
#define init_11_13_13_6
void FASTCALL call_11_13_13_6(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(13<<4)+6);
}
#define init_11_13_13_7
void FASTCALL call_11_13_13_7(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(13<<4)+7);
}
#define init_11_13_13_8
void FASTCALL call_11_13_13_8(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(13<<4)+8);
}
#define init_11_13_13_9
void FASTCALL call_11_13_13_9(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(13<<4)+9);
}
#define init_11_13_13_10
void FASTCALL call_11_13_13_10(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(13<<4)+10);
}
#define init_11_13_13_11
void FASTCALL call_11_13_13_11(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(13<<4)+11);
}
#define init_11_13_13_12
void FASTCALL call_11_13_13_12(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(13<<4)+12);
}
#define init_11_13_13_13
void FASTCALL call_11_13_13_13(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(13<<4)+13);
}
#define init_11_13_13_14
void FASTCALL call_11_13_13_14(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(13<<4)+14);
}
#define init_11_13_13_15
void FASTCALL call_11_13_13_15(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(13<<4)+15);
}
#define init_11_13_14_0
void FASTCALL call_11_13_14_0(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(14<<4)+0);
}
#define init_11_13_14_1
void FASTCALL call_11_13_14_1(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(14<<4)+1);
}
#define init_11_13_14_2
void FASTCALL call_11_13_14_2(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(14<<4)+2);
}
#define init_11_13_14_3
void FASTCALL call_11_13_14_3(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(14<<4)+3);
}
#define init_11_13_14_4
void FASTCALL call_11_13_14_4(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(14<<4)+4);
}
#define init_11_13_14_5
void FASTCALL call_11_13_14_5(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(14<<4)+5);
}
#define init_11_13_14_6
void FASTCALL call_11_13_14_6(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(14<<4)+6);
}
#define init_11_13_14_7
void FASTCALL call_11_13_14_7(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(14<<4)+7);
}
#define init_11_13_14_8
void FASTCALL call_11_13_14_8(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(14<<4)+8);
}
#define init_11_13_14_9
void FASTCALL call_11_13_14_9(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(14<<4)+9);
}
#define init_11_13_14_10
void FASTCALL call_11_13_14_10(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(14<<4)+10);
}
#define init_11_13_14_11
void FASTCALL call_11_13_14_11(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(14<<4)+11);
}
#define init_11_13_14_12
void FASTCALL call_11_13_14_12(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(14<<4)+12);
}
#define init_11_13_14_13
void FASTCALL call_11_13_14_13(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(14<<4)+13);
}
#define init_11_13_14_14
void FASTCALL call_11_13_14_14(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(14<<4)+14);
}
#define init_11_13_14_15
void FASTCALL call_11_13_14_15(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(14<<4)+15);
}
#define init_11_13_15_0
void FASTCALL call_11_13_15_0(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(15<<4)+0);
}
#define init_11_13_15_1
void FASTCALL call_11_13_15_1(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(15<<4)+1);
}
#define init_11_13_15_2
void FASTCALL call_11_13_15_2(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(15<<4)+2);
}
#define init_11_13_15_3
void FASTCALL call_11_13_15_3(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(15<<4)+3);
}
#define init_11_13_15_4
void FASTCALL call_11_13_15_4(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(15<<4)+4);
}
#define init_11_13_15_5
void FASTCALL call_11_13_15_5(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(15<<4)+5);
}
#define init_11_13_15_6
void FASTCALL call_11_13_15_6(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(15<<4)+6);
}
#define init_11_13_15_7
void FASTCALL call_11_13_15_7(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(15<<4)+7);
}
#define init_11_13_15_8
void FASTCALL call_11_13_15_8(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(15<<4)+8);
}
#define init_11_13_15_9
void FASTCALL call_11_13_15_9(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(15<<4)+9);
}
#define init_11_13_15_10
void FASTCALL call_11_13_15_10(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(15<<4)+10);
}
#define init_11_13_15_11
void FASTCALL call_11_13_15_11(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(15<<4)+11);
}
#define init_11_13_15_12
void FASTCALL call_11_13_15_12(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(15<<4)+12);
}
#define init_11_13_15_13
void FASTCALL call_11_13_15_13(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(15<<4)+13);
}
#define init_11_13_15_14
void FASTCALL call_11_13_15_14(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(15<<4)+14);
}
#define init_11_13_15_15
void FASTCALL call_11_13_15_15(SH2_struct* sh)
{
  SH2bsr(sh, (13<<8)+(15<<4)+15);
}
#define init_11_14_0_0
void FASTCALL call_11_14_0_0(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(0<<4)+0);
}
#define init_11_14_0_1
void FASTCALL call_11_14_0_1(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(0<<4)+1);
}
#define init_11_14_0_2
void FASTCALL call_11_14_0_2(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(0<<4)+2);
}
#define init_11_14_0_3
void FASTCALL call_11_14_0_3(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(0<<4)+3);
}
#define init_11_14_0_4
void FASTCALL call_11_14_0_4(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(0<<4)+4);
}
#define init_11_14_0_5
void FASTCALL call_11_14_0_5(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(0<<4)+5);
}
#define init_11_14_0_6
void FASTCALL call_11_14_0_6(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(0<<4)+6);
}
#define init_11_14_0_7
void FASTCALL call_11_14_0_7(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(0<<4)+7);
}
#define init_11_14_0_8
void FASTCALL call_11_14_0_8(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(0<<4)+8);
}
#define init_11_14_0_9
void FASTCALL call_11_14_0_9(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(0<<4)+9);
}
#define init_11_14_0_10
void FASTCALL call_11_14_0_10(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(0<<4)+10);
}
#define init_11_14_0_11
void FASTCALL call_11_14_0_11(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(0<<4)+11);
}
#define init_11_14_0_12
void FASTCALL call_11_14_0_12(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(0<<4)+12);
}
#define init_11_14_0_13
void FASTCALL call_11_14_0_13(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(0<<4)+13);
}
#define init_11_14_0_14
void FASTCALL call_11_14_0_14(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(0<<4)+14);
}
#define init_11_14_0_15
void FASTCALL call_11_14_0_15(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(0<<4)+15);
}
#define init_11_14_1_0
void FASTCALL call_11_14_1_0(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(1<<4)+0);
}
#define init_11_14_1_1
void FASTCALL call_11_14_1_1(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(1<<4)+1);
}
#define init_11_14_1_2
void FASTCALL call_11_14_1_2(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(1<<4)+2);
}
#define init_11_14_1_3
void FASTCALL call_11_14_1_3(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(1<<4)+3);
}
#define init_11_14_1_4
void FASTCALL call_11_14_1_4(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(1<<4)+4);
}
#define init_11_14_1_5
void FASTCALL call_11_14_1_5(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(1<<4)+5);
}
#define init_11_14_1_6
void FASTCALL call_11_14_1_6(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(1<<4)+6);
}
#define init_11_14_1_7
void FASTCALL call_11_14_1_7(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(1<<4)+7);
}
#define init_11_14_1_8
void FASTCALL call_11_14_1_8(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(1<<4)+8);
}
#define init_11_14_1_9
void FASTCALL call_11_14_1_9(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(1<<4)+9);
}
#define init_11_14_1_10
void FASTCALL call_11_14_1_10(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(1<<4)+10);
}
#define init_11_14_1_11
void FASTCALL call_11_14_1_11(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(1<<4)+11);
}
#define init_11_14_1_12
void FASTCALL call_11_14_1_12(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(1<<4)+12);
}
#define init_11_14_1_13
void FASTCALL call_11_14_1_13(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(1<<4)+13);
}
#define init_11_14_1_14
void FASTCALL call_11_14_1_14(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(1<<4)+14);
}
#define init_11_14_1_15
void FASTCALL call_11_14_1_15(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(1<<4)+15);
}
#define init_11_14_2_0
void FASTCALL call_11_14_2_0(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(2<<4)+0);
}
#define init_11_14_2_1
void FASTCALL call_11_14_2_1(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(2<<4)+1);
}
#define init_11_14_2_2
void FASTCALL call_11_14_2_2(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(2<<4)+2);
}
#define init_11_14_2_3
void FASTCALL call_11_14_2_3(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(2<<4)+3);
}
#define init_11_14_2_4
void FASTCALL call_11_14_2_4(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(2<<4)+4);
}
#define init_11_14_2_5
void FASTCALL call_11_14_2_5(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(2<<4)+5);
}
#define init_11_14_2_6
void FASTCALL call_11_14_2_6(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(2<<4)+6);
}
#define init_11_14_2_7
void FASTCALL call_11_14_2_7(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(2<<4)+7);
}
#define init_11_14_2_8
void FASTCALL call_11_14_2_8(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(2<<4)+8);
}
#define init_11_14_2_9
void FASTCALL call_11_14_2_9(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(2<<4)+9);
}
#define init_11_14_2_10
void FASTCALL call_11_14_2_10(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(2<<4)+10);
}
#define init_11_14_2_11
void FASTCALL call_11_14_2_11(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(2<<4)+11);
}
#define init_11_14_2_12
void FASTCALL call_11_14_2_12(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(2<<4)+12);
}
#define init_11_14_2_13
void FASTCALL call_11_14_2_13(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(2<<4)+13);
}
#define init_11_14_2_14
void FASTCALL call_11_14_2_14(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(2<<4)+14);
}
#define init_11_14_2_15
void FASTCALL call_11_14_2_15(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(2<<4)+15);
}
#define init_11_14_3_0
void FASTCALL call_11_14_3_0(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(3<<4)+0);
}
#define init_11_14_3_1
void FASTCALL call_11_14_3_1(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(3<<4)+1);
}
#define init_11_14_3_2
void FASTCALL call_11_14_3_2(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(3<<4)+2);
}
#define init_11_14_3_3
void FASTCALL call_11_14_3_3(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(3<<4)+3);
}
#define init_11_14_3_4
void FASTCALL call_11_14_3_4(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(3<<4)+4);
}
#define init_11_14_3_5
void FASTCALL call_11_14_3_5(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(3<<4)+5);
}
#define init_11_14_3_6
void FASTCALL call_11_14_3_6(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(3<<4)+6);
}
#define init_11_14_3_7
void FASTCALL call_11_14_3_7(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(3<<4)+7);
}
#define init_11_14_3_8
void FASTCALL call_11_14_3_8(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(3<<4)+8);
}
#define init_11_14_3_9
void FASTCALL call_11_14_3_9(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(3<<4)+9);
}
#define init_11_14_3_10
void FASTCALL call_11_14_3_10(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(3<<4)+10);
}
#define init_11_14_3_11
void FASTCALL call_11_14_3_11(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(3<<4)+11);
}
#define init_11_14_3_12
void FASTCALL call_11_14_3_12(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(3<<4)+12);
}
#define init_11_14_3_13
void FASTCALL call_11_14_3_13(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(3<<4)+13);
}
#define init_11_14_3_14
void FASTCALL call_11_14_3_14(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(3<<4)+14);
}
#define init_11_14_3_15
void FASTCALL call_11_14_3_15(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(3<<4)+15);
}
#define init_11_14_4_0
void FASTCALL call_11_14_4_0(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(4<<4)+0);
}
#define init_11_14_4_1
void FASTCALL call_11_14_4_1(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(4<<4)+1);
}
#define init_11_14_4_2
void FASTCALL call_11_14_4_2(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(4<<4)+2);
}
#define init_11_14_4_3
void FASTCALL call_11_14_4_3(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(4<<4)+3);
}
#define init_11_14_4_4
void FASTCALL call_11_14_4_4(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(4<<4)+4);
}
#define init_11_14_4_5
void FASTCALL call_11_14_4_5(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(4<<4)+5);
}
#define init_11_14_4_6
void FASTCALL call_11_14_4_6(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(4<<4)+6);
}
#define init_11_14_4_7
void FASTCALL call_11_14_4_7(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(4<<4)+7);
}
#define init_11_14_4_8
void FASTCALL call_11_14_4_8(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(4<<4)+8);
}
#define init_11_14_4_9
void FASTCALL call_11_14_4_9(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(4<<4)+9);
}
#define init_11_14_4_10
void FASTCALL call_11_14_4_10(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(4<<4)+10);
}
#define init_11_14_4_11
void FASTCALL call_11_14_4_11(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(4<<4)+11);
}
#define init_11_14_4_12
void FASTCALL call_11_14_4_12(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(4<<4)+12);
}
#define init_11_14_4_13
void FASTCALL call_11_14_4_13(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(4<<4)+13);
}
#define init_11_14_4_14
void FASTCALL call_11_14_4_14(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(4<<4)+14);
}
#define init_11_14_4_15
void FASTCALL call_11_14_4_15(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(4<<4)+15);
}
#define init_11_14_5_0
void FASTCALL call_11_14_5_0(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(5<<4)+0);
}
#define init_11_14_5_1
void FASTCALL call_11_14_5_1(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(5<<4)+1);
}
#define init_11_14_5_2
void FASTCALL call_11_14_5_2(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(5<<4)+2);
}
#define init_11_14_5_3
void FASTCALL call_11_14_5_3(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(5<<4)+3);
}
#define init_11_14_5_4
void FASTCALL call_11_14_5_4(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(5<<4)+4);
}
#define init_11_14_5_5
void FASTCALL call_11_14_5_5(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(5<<4)+5);
}
#define init_11_14_5_6
void FASTCALL call_11_14_5_6(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(5<<4)+6);
}
#define init_11_14_5_7
void FASTCALL call_11_14_5_7(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(5<<4)+7);
}
#define init_11_14_5_8
void FASTCALL call_11_14_5_8(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(5<<4)+8);
}
#define init_11_14_5_9
void FASTCALL call_11_14_5_9(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(5<<4)+9);
}
#define init_11_14_5_10
void FASTCALL call_11_14_5_10(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(5<<4)+10);
}
#define init_11_14_5_11
void FASTCALL call_11_14_5_11(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(5<<4)+11);
}
#define init_11_14_5_12
void FASTCALL call_11_14_5_12(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(5<<4)+12);
}
#define init_11_14_5_13
void FASTCALL call_11_14_5_13(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(5<<4)+13);
}
#define init_11_14_5_14
void FASTCALL call_11_14_5_14(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(5<<4)+14);
}
#define init_11_14_5_15
void FASTCALL call_11_14_5_15(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(5<<4)+15);
}
#define init_11_14_6_0
void FASTCALL call_11_14_6_0(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(6<<4)+0);
}
#define init_11_14_6_1
void FASTCALL call_11_14_6_1(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(6<<4)+1);
}
#define init_11_14_6_2
void FASTCALL call_11_14_6_2(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(6<<4)+2);
}
#define init_11_14_6_3
void FASTCALL call_11_14_6_3(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(6<<4)+3);
}
#define init_11_14_6_4
void FASTCALL call_11_14_6_4(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(6<<4)+4);
}
#define init_11_14_6_5
void FASTCALL call_11_14_6_5(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(6<<4)+5);
}
#define init_11_14_6_6
void FASTCALL call_11_14_6_6(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(6<<4)+6);
}
#define init_11_14_6_7
void FASTCALL call_11_14_6_7(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(6<<4)+7);
}
#define init_11_14_6_8
void FASTCALL call_11_14_6_8(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(6<<4)+8);
}
#define init_11_14_6_9
void FASTCALL call_11_14_6_9(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(6<<4)+9);
}
#define init_11_14_6_10
void FASTCALL call_11_14_6_10(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(6<<4)+10);
}
#define init_11_14_6_11
void FASTCALL call_11_14_6_11(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(6<<4)+11);
}
#define init_11_14_6_12
void FASTCALL call_11_14_6_12(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(6<<4)+12);
}
#define init_11_14_6_13
void FASTCALL call_11_14_6_13(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(6<<4)+13);
}
#define init_11_14_6_14
void FASTCALL call_11_14_6_14(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(6<<4)+14);
}
#define init_11_14_6_15
void FASTCALL call_11_14_6_15(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(6<<4)+15);
}
#define init_11_14_7_0
void FASTCALL call_11_14_7_0(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(7<<4)+0);
}
#define init_11_14_7_1
void FASTCALL call_11_14_7_1(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(7<<4)+1);
}
#define init_11_14_7_2
void FASTCALL call_11_14_7_2(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(7<<4)+2);
}
#define init_11_14_7_3
void FASTCALL call_11_14_7_3(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(7<<4)+3);
}
#define init_11_14_7_4
void FASTCALL call_11_14_7_4(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(7<<4)+4);
}
#define init_11_14_7_5
void FASTCALL call_11_14_7_5(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(7<<4)+5);
}
#define init_11_14_7_6
void FASTCALL call_11_14_7_6(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(7<<4)+6);
}
#define init_11_14_7_7
void FASTCALL call_11_14_7_7(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(7<<4)+7);
}
#define init_11_14_7_8
void FASTCALL call_11_14_7_8(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(7<<4)+8);
}
#define init_11_14_7_9
void FASTCALL call_11_14_7_9(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(7<<4)+9);
}
#define init_11_14_7_10
void FASTCALL call_11_14_7_10(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(7<<4)+10);
}
#define init_11_14_7_11
void FASTCALL call_11_14_7_11(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(7<<4)+11);
}
#define init_11_14_7_12
void FASTCALL call_11_14_7_12(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(7<<4)+12);
}
#define init_11_14_7_13
void FASTCALL call_11_14_7_13(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(7<<4)+13);
}
#define init_11_14_7_14
void FASTCALL call_11_14_7_14(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(7<<4)+14);
}
#define init_11_14_7_15
void FASTCALL call_11_14_7_15(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(7<<4)+15);
}
#define init_11_14_8_0
void FASTCALL call_11_14_8_0(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(8<<4)+0);
}
#define init_11_14_8_1
void FASTCALL call_11_14_8_1(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(8<<4)+1);
}
#define init_11_14_8_2
void FASTCALL call_11_14_8_2(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(8<<4)+2);
}
#define init_11_14_8_3
void FASTCALL call_11_14_8_3(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(8<<4)+3);
}
#define init_11_14_8_4
void FASTCALL call_11_14_8_4(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(8<<4)+4);
}
#define init_11_14_8_5
void FASTCALL call_11_14_8_5(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(8<<4)+5);
}
#define init_11_14_8_6
void FASTCALL call_11_14_8_6(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(8<<4)+6);
}
#define init_11_14_8_7
void FASTCALL call_11_14_8_7(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(8<<4)+7);
}
#define init_11_14_8_8
void FASTCALL call_11_14_8_8(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(8<<4)+8);
}
#define init_11_14_8_9
void FASTCALL call_11_14_8_9(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(8<<4)+9);
}
#define init_11_14_8_10
void FASTCALL call_11_14_8_10(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(8<<4)+10);
}
#define init_11_14_8_11
void FASTCALL call_11_14_8_11(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(8<<4)+11);
}
#define init_11_14_8_12
void FASTCALL call_11_14_8_12(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(8<<4)+12);
}
#define init_11_14_8_13
void FASTCALL call_11_14_8_13(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(8<<4)+13);
}
#define init_11_14_8_14
void FASTCALL call_11_14_8_14(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(8<<4)+14);
}
#define init_11_14_8_15
void FASTCALL call_11_14_8_15(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(8<<4)+15);
}
#define init_11_14_9_0
void FASTCALL call_11_14_9_0(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(9<<4)+0);
}
#define init_11_14_9_1
void FASTCALL call_11_14_9_1(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(9<<4)+1);
}
#define init_11_14_9_2
void FASTCALL call_11_14_9_2(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(9<<4)+2);
}
#define init_11_14_9_3
void FASTCALL call_11_14_9_3(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(9<<4)+3);
}
#define init_11_14_9_4
void FASTCALL call_11_14_9_4(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(9<<4)+4);
}
#define init_11_14_9_5
void FASTCALL call_11_14_9_5(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(9<<4)+5);
}
#define init_11_14_9_6
void FASTCALL call_11_14_9_6(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(9<<4)+6);
}
#define init_11_14_9_7
void FASTCALL call_11_14_9_7(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(9<<4)+7);
}
#define init_11_14_9_8
void FASTCALL call_11_14_9_8(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(9<<4)+8);
}
#define init_11_14_9_9
void FASTCALL call_11_14_9_9(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(9<<4)+9);
}
#define init_11_14_9_10
void FASTCALL call_11_14_9_10(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(9<<4)+10);
}
#define init_11_14_9_11
void FASTCALL call_11_14_9_11(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(9<<4)+11);
}
#define init_11_14_9_12
void FASTCALL call_11_14_9_12(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(9<<4)+12);
}
#define init_11_14_9_13
void FASTCALL call_11_14_9_13(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(9<<4)+13);
}
#define init_11_14_9_14
void FASTCALL call_11_14_9_14(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(9<<4)+14);
}
#define init_11_14_9_15
void FASTCALL call_11_14_9_15(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(9<<4)+15);
}
#define init_11_14_10_0
void FASTCALL call_11_14_10_0(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(10<<4)+0);
}
#define init_11_14_10_1
void FASTCALL call_11_14_10_1(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(10<<4)+1);
}
#define init_11_14_10_2
void FASTCALL call_11_14_10_2(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(10<<4)+2);
}
#define init_11_14_10_3
void FASTCALL call_11_14_10_3(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(10<<4)+3);
}
#define init_11_14_10_4
void FASTCALL call_11_14_10_4(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(10<<4)+4);
}
#define init_11_14_10_5
void FASTCALL call_11_14_10_5(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(10<<4)+5);
}
#define init_11_14_10_6
void FASTCALL call_11_14_10_6(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(10<<4)+6);
}
#define init_11_14_10_7
void FASTCALL call_11_14_10_7(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(10<<4)+7);
}
#define init_11_14_10_8
void FASTCALL call_11_14_10_8(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(10<<4)+8);
}
#define init_11_14_10_9
void FASTCALL call_11_14_10_9(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(10<<4)+9);
}
#define init_11_14_10_10
void FASTCALL call_11_14_10_10(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(10<<4)+10);
}
#define init_11_14_10_11
void FASTCALL call_11_14_10_11(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(10<<4)+11);
}
#define init_11_14_10_12
void FASTCALL call_11_14_10_12(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(10<<4)+12);
}
#define init_11_14_10_13
void FASTCALL call_11_14_10_13(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(10<<4)+13);
}
#define init_11_14_10_14
void FASTCALL call_11_14_10_14(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(10<<4)+14);
}
#define init_11_14_10_15
void FASTCALL call_11_14_10_15(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(10<<4)+15);
}
#define init_11_14_11_0
void FASTCALL call_11_14_11_0(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(11<<4)+0);
}
#define init_11_14_11_1
void FASTCALL call_11_14_11_1(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(11<<4)+1);
}
#define init_11_14_11_2
void FASTCALL call_11_14_11_2(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(11<<4)+2);
}
#define init_11_14_11_3
void FASTCALL call_11_14_11_3(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(11<<4)+3);
}
#define init_11_14_11_4
void FASTCALL call_11_14_11_4(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(11<<4)+4);
}
#define init_11_14_11_5
void FASTCALL call_11_14_11_5(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(11<<4)+5);
}
#define init_11_14_11_6
void FASTCALL call_11_14_11_6(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(11<<4)+6);
}
#define init_11_14_11_7
void FASTCALL call_11_14_11_7(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(11<<4)+7);
}
#define init_11_14_11_8
void FASTCALL call_11_14_11_8(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(11<<4)+8);
}
#define init_11_14_11_9
void FASTCALL call_11_14_11_9(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(11<<4)+9);
}
#define init_11_14_11_10
void FASTCALL call_11_14_11_10(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(11<<4)+10);
}
#define init_11_14_11_11
void FASTCALL call_11_14_11_11(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(11<<4)+11);
}
#define init_11_14_11_12
void FASTCALL call_11_14_11_12(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(11<<4)+12);
}
#define init_11_14_11_13
void FASTCALL call_11_14_11_13(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(11<<4)+13);
}
#define init_11_14_11_14
void FASTCALL call_11_14_11_14(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(11<<4)+14);
}
#define init_11_14_11_15
void FASTCALL call_11_14_11_15(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(11<<4)+15);
}
#define init_11_14_12_0
void FASTCALL call_11_14_12_0(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(12<<4)+0);
}
#define init_11_14_12_1
void FASTCALL call_11_14_12_1(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(12<<4)+1);
}
#define init_11_14_12_2
void FASTCALL call_11_14_12_2(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(12<<4)+2);
}
#define init_11_14_12_3
void FASTCALL call_11_14_12_3(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(12<<4)+3);
}
#define init_11_14_12_4
void FASTCALL call_11_14_12_4(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(12<<4)+4);
}
#define init_11_14_12_5
void FASTCALL call_11_14_12_5(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(12<<4)+5);
}
#define init_11_14_12_6
void FASTCALL call_11_14_12_6(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(12<<4)+6);
}
#define init_11_14_12_7
void FASTCALL call_11_14_12_7(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(12<<4)+7);
}
#define init_11_14_12_8
void FASTCALL call_11_14_12_8(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(12<<4)+8);
}
#define init_11_14_12_9
void FASTCALL call_11_14_12_9(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(12<<4)+9);
}
#define init_11_14_12_10
void FASTCALL call_11_14_12_10(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(12<<4)+10);
}
#define init_11_14_12_11
void FASTCALL call_11_14_12_11(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(12<<4)+11);
}
#define init_11_14_12_12
void FASTCALL call_11_14_12_12(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(12<<4)+12);
}
#define init_11_14_12_13
void FASTCALL call_11_14_12_13(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(12<<4)+13);
}
#define init_11_14_12_14
void FASTCALL call_11_14_12_14(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(12<<4)+14);
}
#define init_11_14_12_15
void FASTCALL call_11_14_12_15(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(12<<4)+15);
}
#define init_11_14_13_0
void FASTCALL call_11_14_13_0(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(13<<4)+0);
}
#define init_11_14_13_1
void FASTCALL call_11_14_13_1(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(13<<4)+1);
}
#define init_11_14_13_2
void FASTCALL call_11_14_13_2(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(13<<4)+2);
}
#define init_11_14_13_3
void FASTCALL call_11_14_13_3(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(13<<4)+3);
}
#define init_11_14_13_4
void FASTCALL call_11_14_13_4(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(13<<4)+4);
}
#define init_11_14_13_5
void FASTCALL call_11_14_13_5(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(13<<4)+5);
}
#define init_11_14_13_6
void FASTCALL call_11_14_13_6(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(13<<4)+6);
}
#define init_11_14_13_7
void FASTCALL call_11_14_13_7(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(13<<4)+7);
}
#define init_11_14_13_8
void FASTCALL call_11_14_13_8(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(13<<4)+8);
}
#define init_11_14_13_9
void FASTCALL call_11_14_13_9(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(13<<4)+9);
}
#define init_11_14_13_10
void FASTCALL call_11_14_13_10(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(13<<4)+10);
}
#define init_11_14_13_11
void FASTCALL call_11_14_13_11(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(13<<4)+11);
}
#define init_11_14_13_12
void FASTCALL call_11_14_13_12(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(13<<4)+12);
}
#define init_11_14_13_13
void FASTCALL call_11_14_13_13(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(13<<4)+13);
}
#define init_11_14_13_14
void FASTCALL call_11_14_13_14(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(13<<4)+14);
}
#define init_11_14_13_15
void FASTCALL call_11_14_13_15(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(13<<4)+15);
}
#define init_11_14_14_0
void FASTCALL call_11_14_14_0(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(14<<4)+0);
}
#define init_11_14_14_1
void FASTCALL call_11_14_14_1(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(14<<4)+1);
}
#define init_11_14_14_2
void FASTCALL call_11_14_14_2(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(14<<4)+2);
}
#define init_11_14_14_3
void FASTCALL call_11_14_14_3(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(14<<4)+3);
}
#define init_11_14_14_4
void FASTCALL call_11_14_14_4(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(14<<4)+4);
}
#define init_11_14_14_5
void FASTCALL call_11_14_14_5(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(14<<4)+5);
}
#define init_11_14_14_6
void FASTCALL call_11_14_14_6(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(14<<4)+6);
}
#define init_11_14_14_7
void FASTCALL call_11_14_14_7(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(14<<4)+7);
}
#define init_11_14_14_8
void FASTCALL call_11_14_14_8(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(14<<4)+8);
}
#define init_11_14_14_9
void FASTCALL call_11_14_14_9(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(14<<4)+9);
}
#define init_11_14_14_10
void FASTCALL call_11_14_14_10(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(14<<4)+10);
}
#define init_11_14_14_11
void FASTCALL call_11_14_14_11(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(14<<4)+11);
}
#define init_11_14_14_12
void FASTCALL call_11_14_14_12(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(14<<4)+12);
}
#define init_11_14_14_13
void FASTCALL call_11_14_14_13(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(14<<4)+13);
}
#define init_11_14_14_14
void FASTCALL call_11_14_14_14(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(14<<4)+14);
}
#define init_11_14_14_15
void FASTCALL call_11_14_14_15(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(14<<4)+15);
}
#define init_11_14_15_0
void FASTCALL call_11_14_15_0(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(15<<4)+0);
}
#define init_11_14_15_1
void FASTCALL call_11_14_15_1(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(15<<4)+1);
}
#define init_11_14_15_2
void FASTCALL call_11_14_15_2(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(15<<4)+2);
}
#define init_11_14_15_3
void FASTCALL call_11_14_15_3(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(15<<4)+3);
}
#define init_11_14_15_4
void FASTCALL call_11_14_15_4(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(15<<4)+4);
}
#define init_11_14_15_5
void FASTCALL call_11_14_15_5(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(15<<4)+5);
}
#define init_11_14_15_6
void FASTCALL call_11_14_15_6(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(15<<4)+6);
}
#define init_11_14_15_7
void FASTCALL call_11_14_15_7(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(15<<4)+7);
}
#define init_11_14_15_8
void FASTCALL call_11_14_15_8(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(15<<4)+8);
}
#define init_11_14_15_9
void FASTCALL call_11_14_15_9(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(15<<4)+9);
}
#define init_11_14_15_10
void FASTCALL call_11_14_15_10(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(15<<4)+10);
}
#define init_11_14_15_11
void FASTCALL call_11_14_15_11(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(15<<4)+11);
}
#define init_11_14_15_12
void FASTCALL call_11_14_15_12(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(15<<4)+12);
}
#define init_11_14_15_13
void FASTCALL call_11_14_15_13(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(15<<4)+13);
}
#define init_11_14_15_14
void FASTCALL call_11_14_15_14(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(15<<4)+14);
}
#define init_11_14_15_15
void FASTCALL call_11_14_15_15(SH2_struct* sh)
{
  SH2bsr(sh, (14<<8)+(15<<4)+15);
}
#define init_11_15_0_0
void FASTCALL call_11_15_0_0(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(0<<4)+0);
}
#define init_11_15_0_1
void FASTCALL call_11_15_0_1(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(0<<4)+1);
}
#define init_11_15_0_2
void FASTCALL call_11_15_0_2(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(0<<4)+2);
}
#define init_11_15_0_3
void FASTCALL call_11_15_0_3(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(0<<4)+3);
}
#define init_11_15_0_4
void FASTCALL call_11_15_0_4(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(0<<4)+4);
}
#define init_11_15_0_5
void FASTCALL call_11_15_0_5(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(0<<4)+5);
}
#define init_11_15_0_6
void FASTCALL call_11_15_0_6(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(0<<4)+6);
}
#define init_11_15_0_7
void FASTCALL call_11_15_0_7(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(0<<4)+7);
}
#define init_11_15_0_8
void FASTCALL call_11_15_0_8(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(0<<4)+8);
}
#define init_11_15_0_9
void FASTCALL call_11_15_0_9(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(0<<4)+9);
}
#define init_11_15_0_10
void FASTCALL call_11_15_0_10(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(0<<4)+10);
}
#define init_11_15_0_11
void FASTCALL call_11_15_0_11(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(0<<4)+11);
}
#define init_11_15_0_12
void FASTCALL call_11_15_0_12(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(0<<4)+12);
}
#define init_11_15_0_13
void FASTCALL call_11_15_0_13(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(0<<4)+13);
}
#define init_11_15_0_14
void FASTCALL call_11_15_0_14(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(0<<4)+14);
}
#define init_11_15_0_15
void FASTCALL call_11_15_0_15(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(0<<4)+15);
}
#define init_11_15_1_0
void FASTCALL call_11_15_1_0(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(1<<4)+0);
}
#define init_11_15_1_1
void FASTCALL call_11_15_1_1(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(1<<4)+1);
}
#define init_11_15_1_2
void FASTCALL call_11_15_1_2(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(1<<4)+2);
}
#define init_11_15_1_3
void FASTCALL call_11_15_1_3(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(1<<4)+3);
}
#define init_11_15_1_4
void FASTCALL call_11_15_1_4(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(1<<4)+4);
}
#define init_11_15_1_5
void FASTCALL call_11_15_1_5(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(1<<4)+5);
}
#define init_11_15_1_6
void FASTCALL call_11_15_1_6(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(1<<4)+6);
}
#define init_11_15_1_7
void FASTCALL call_11_15_1_7(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(1<<4)+7);
}
#define init_11_15_1_8
void FASTCALL call_11_15_1_8(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(1<<4)+8);
}
#define init_11_15_1_9
void FASTCALL call_11_15_1_9(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(1<<4)+9);
}
#define init_11_15_1_10
void FASTCALL call_11_15_1_10(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(1<<4)+10);
}
#define init_11_15_1_11
void FASTCALL call_11_15_1_11(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(1<<4)+11);
}
#define init_11_15_1_12
void FASTCALL call_11_15_1_12(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(1<<4)+12);
}
#define init_11_15_1_13
void FASTCALL call_11_15_1_13(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(1<<4)+13);
}
#define init_11_15_1_14
void FASTCALL call_11_15_1_14(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(1<<4)+14);
}
#define init_11_15_1_15
void FASTCALL call_11_15_1_15(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(1<<4)+15);
}
#define init_11_15_2_0
void FASTCALL call_11_15_2_0(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(2<<4)+0);
}
#define init_11_15_2_1
void FASTCALL call_11_15_2_1(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(2<<4)+1);
}
#define init_11_15_2_2
void FASTCALL call_11_15_2_2(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(2<<4)+2);
}
#define init_11_15_2_3
void FASTCALL call_11_15_2_3(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(2<<4)+3);
}
#define init_11_15_2_4
void FASTCALL call_11_15_2_4(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(2<<4)+4);
}
#define init_11_15_2_5
void FASTCALL call_11_15_2_5(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(2<<4)+5);
}
#define init_11_15_2_6
void FASTCALL call_11_15_2_6(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(2<<4)+6);
}
#define init_11_15_2_7
void FASTCALL call_11_15_2_7(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(2<<4)+7);
}
#define init_11_15_2_8
void FASTCALL call_11_15_2_8(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(2<<4)+8);
}
#define init_11_15_2_9
void FASTCALL call_11_15_2_9(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(2<<4)+9);
}
#define init_11_15_2_10
void FASTCALL call_11_15_2_10(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(2<<4)+10);
}
#define init_11_15_2_11
void FASTCALL call_11_15_2_11(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(2<<4)+11);
}
#define init_11_15_2_12
void FASTCALL call_11_15_2_12(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(2<<4)+12);
}
#define init_11_15_2_13
void FASTCALL call_11_15_2_13(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(2<<4)+13);
}
#define init_11_15_2_14
void FASTCALL call_11_15_2_14(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(2<<4)+14);
}
#define init_11_15_2_15
void FASTCALL call_11_15_2_15(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(2<<4)+15);
}
#define init_11_15_3_0
void FASTCALL call_11_15_3_0(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(3<<4)+0);
}
#define init_11_15_3_1
void FASTCALL call_11_15_3_1(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(3<<4)+1);
}
#define init_11_15_3_2
void FASTCALL call_11_15_3_2(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(3<<4)+2);
}
#define init_11_15_3_3
void FASTCALL call_11_15_3_3(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(3<<4)+3);
}
#define init_11_15_3_4
void FASTCALL call_11_15_3_4(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(3<<4)+4);
}
#define init_11_15_3_5
void FASTCALL call_11_15_3_5(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(3<<4)+5);
}
#define init_11_15_3_6
void FASTCALL call_11_15_3_6(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(3<<4)+6);
}
#define init_11_15_3_7
void FASTCALL call_11_15_3_7(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(3<<4)+7);
}
#define init_11_15_3_8
void FASTCALL call_11_15_3_8(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(3<<4)+8);
}
#define init_11_15_3_9
void FASTCALL call_11_15_3_9(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(3<<4)+9);
}
#define init_11_15_3_10
void FASTCALL call_11_15_3_10(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(3<<4)+10);
}
#define init_11_15_3_11
void FASTCALL call_11_15_3_11(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(3<<4)+11);
}
#define init_11_15_3_12
void FASTCALL call_11_15_3_12(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(3<<4)+12);
}
#define init_11_15_3_13
void FASTCALL call_11_15_3_13(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(3<<4)+13);
}
#define init_11_15_3_14
void FASTCALL call_11_15_3_14(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(3<<4)+14);
}
#define init_11_15_3_15
void FASTCALL call_11_15_3_15(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(3<<4)+15);
}
#define init_11_15_4_0
void FASTCALL call_11_15_4_0(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(4<<4)+0);
}
#define init_11_15_4_1
void FASTCALL call_11_15_4_1(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(4<<4)+1);
}
#define init_11_15_4_2
void FASTCALL call_11_15_4_2(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(4<<4)+2);
}
#define init_11_15_4_3
void FASTCALL call_11_15_4_3(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(4<<4)+3);
}
#define init_11_15_4_4
void FASTCALL call_11_15_4_4(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(4<<4)+4);
}
#define init_11_15_4_5
void FASTCALL call_11_15_4_5(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(4<<4)+5);
}
#define init_11_15_4_6
void FASTCALL call_11_15_4_6(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(4<<4)+6);
}
#define init_11_15_4_7
void FASTCALL call_11_15_4_7(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(4<<4)+7);
}
#define init_11_15_4_8
void FASTCALL call_11_15_4_8(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(4<<4)+8);
}
#define init_11_15_4_9
void FASTCALL call_11_15_4_9(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(4<<4)+9);
}
#define init_11_15_4_10
void FASTCALL call_11_15_4_10(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(4<<4)+10);
}
#define init_11_15_4_11
void FASTCALL call_11_15_4_11(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(4<<4)+11);
}
#define init_11_15_4_12
void FASTCALL call_11_15_4_12(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(4<<4)+12);
}
#define init_11_15_4_13
void FASTCALL call_11_15_4_13(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(4<<4)+13);
}
#define init_11_15_4_14
void FASTCALL call_11_15_4_14(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(4<<4)+14);
}
#define init_11_15_4_15
void FASTCALL call_11_15_4_15(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(4<<4)+15);
}
#define init_11_15_5_0
void FASTCALL call_11_15_5_0(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(5<<4)+0);
}
#define init_11_15_5_1
void FASTCALL call_11_15_5_1(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(5<<4)+1);
}
#define init_11_15_5_2
void FASTCALL call_11_15_5_2(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(5<<4)+2);
}
#define init_11_15_5_3
void FASTCALL call_11_15_5_3(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(5<<4)+3);
}
#define init_11_15_5_4
void FASTCALL call_11_15_5_4(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(5<<4)+4);
}
#define init_11_15_5_5
void FASTCALL call_11_15_5_5(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(5<<4)+5);
}
#define init_11_15_5_6
void FASTCALL call_11_15_5_6(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(5<<4)+6);
}
#define init_11_15_5_7
void FASTCALL call_11_15_5_7(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(5<<4)+7);
}
#define init_11_15_5_8
void FASTCALL call_11_15_5_8(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(5<<4)+8);
}
#define init_11_15_5_9
void FASTCALL call_11_15_5_9(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(5<<4)+9);
}
#define init_11_15_5_10
void FASTCALL call_11_15_5_10(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(5<<4)+10);
}
#define init_11_15_5_11
void FASTCALL call_11_15_5_11(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(5<<4)+11);
}
#define init_11_15_5_12
void FASTCALL call_11_15_5_12(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(5<<4)+12);
}
#define init_11_15_5_13
void FASTCALL call_11_15_5_13(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(5<<4)+13);
}
#define init_11_15_5_14
void FASTCALL call_11_15_5_14(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(5<<4)+14);
}
#define init_11_15_5_15
void FASTCALL call_11_15_5_15(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(5<<4)+15);
}
#define init_11_15_6_0
void FASTCALL call_11_15_6_0(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(6<<4)+0);
}
#define init_11_15_6_1
void FASTCALL call_11_15_6_1(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(6<<4)+1);
}
#define init_11_15_6_2
void FASTCALL call_11_15_6_2(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(6<<4)+2);
}
#define init_11_15_6_3
void FASTCALL call_11_15_6_3(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(6<<4)+3);
}
#define init_11_15_6_4
void FASTCALL call_11_15_6_4(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(6<<4)+4);
}
#define init_11_15_6_5
void FASTCALL call_11_15_6_5(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(6<<4)+5);
}
#define init_11_15_6_6
void FASTCALL call_11_15_6_6(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(6<<4)+6);
}
#define init_11_15_6_7
void FASTCALL call_11_15_6_7(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(6<<4)+7);
}
#define init_11_15_6_8
void FASTCALL call_11_15_6_8(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(6<<4)+8);
}
#define init_11_15_6_9
void FASTCALL call_11_15_6_9(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(6<<4)+9);
}
#define init_11_15_6_10
void FASTCALL call_11_15_6_10(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(6<<4)+10);
}
#define init_11_15_6_11
void FASTCALL call_11_15_6_11(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(6<<4)+11);
}
#define init_11_15_6_12
void FASTCALL call_11_15_6_12(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(6<<4)+12);
}
#define init_11_15_6_13
void FASTCALL call_11_15_6_13(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(6<<4)+13);
}
#define init_11_15_6_14
void FASTCALL call_11_15_6_14(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(6<<4)+14);
}
#define init_11_15_6_15
void FASTCALL call_11_15_6_15(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(6<<4)+15);
}
#define init_11_15_7_0
void FASTCALL call_11_15_7_0(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(7<<4)+0);
}
#define init_11_15_7_1
void FASTCALL call_11_15_7_1(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(7<<4)+1);
}
#define init_11_15_7_2
void FASTCALL call_11_15_7_2(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(7<<4)+2);
}
#define init_11_15_7_3
void FASTCALL call_11_15_7_3(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(7<<4)+3);
}
#define init_11_15_7_4
void FASTCALL call_11_15_7_4(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(7<<4)+4);
}
#define init_11_15_7_5
void FASTCALL call_11_15_7_5(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(7<<4)+5);
}
#define init_11_15_7_6
void FASTCALL call_11_15_7_6(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(7<<4)+6);
}
#define init_11_15_7_7
void FASTCALL call_11_15_7_7(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(7<<4)+7);
}
#define init_11_15_7_8
void FASTCALL call_11_15_7_8(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(7<<4)+8);
}
#define init_11_15_7_9
void FASTCALL call_11_15_7_9(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(7<<4)+9);
}
#define init_11_15_7_10
void FASTCALL call_11_15_7_10(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(7<<4)+10);
}
#define init_11_15_7_11
void FASTCALL call_11_15_7_11(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(7<<4)+11);
}
#define init_11_15_7_12
void FASTCALL call_11_15_7_12(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(7<<4)+12);
}
#define init_11_15_7_13
void FASTCALL call_11_15_7_13(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(7<<4)+13);
}
#define init_11_15_7_14
void FASTCALL call_11_15_7_14(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(7<<4)+14);
}
#define init_11_15_7_15
void FASTCALL call_11_15_7_15(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(7<<4)+15);
}
#define init_11_15_8_0
void FASTCALL call_11_15_8_0(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(8<<4)+0);
}
#define init_11_15_8_1
void FASTCALL call_11_15_8_1(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(8<<4)+1);
}
#define init_11_15_8_2
void FASTCALL call_11_15_8_2(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(8<<4)+2);
}
#define init_11_15_8_3
void FASTCALL call_11_15_8_3(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(8<<4)+3);
}
#define init_11_15_8_4
void FASTCALL call_11_15_8_4(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(8<<4)+4);
}
#define init_11_15_8_5
void FASTCALL call_11_15_8_5(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(8<<4)+5);
}
#define init_11_15_8_6
void FASTCALL call_11_15_8_6(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(8<<4)+6);
}
#define init_11_15_8_7
void FASTCALL call_11_15_8_7(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(8<<4)+7);
}
#define init_11_15_8_8
void FASTCALL call_11_15_8_8(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(8<<4)+8);
}
#define init_11_15_8_9
void FASTCALL call_11_15_8_9(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(8<<4)+9);
}
#define init_11_15_8_10
void FASTCALL call_11_15_8_10(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(8<<4)+10);
}
#define init_11_15_8_11
void FASTCALL call_11_15_8_11(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(8<<4)+11);
}
#define init_11_15_8_12
void FASTCALL call_11_15_8_12(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(8<<4)+12);
}
#define init_11_15_8_13
void FASTCALL call_11_15_8_13(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(8<<4)+13);
}
#define init_11_15_8_14
void FASTCALL call_11_15_8_14(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(8<<4)+14);
}
#define init_11_15_8_15
void FASTCALL call_11_15_8_15(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(8<<4)+15);
}
#define init_11_15_9_0
void FASTCALL call_11_15_9_0(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(9<<4)+0);
}
#define init_11_15_9_1
void FASTCALL call_11_15_9_1(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(9<<4)+1);
}
#define init_11_15_9_2
void FASTCALL call_11_15_9_2(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(9<<4)+2);
}
#define init_11_15_9_3
void FASTCALL call_11_15_9_3(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(9<<4)+3);
}
#define init_11_15_9_4
void FASTCALL call_11_15_9_4(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(9<<4)+4);
}
#define init_11_15_9_5
void FASTCALL call_11_15_9_5(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(9<<4)+5);
}
#define init_11_15_9_6
void FASTCALL call_11_15_9_6(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(9<<4)+6);
}
#define init_11_15_9_7
void FASTCALL call_11_15_9_7(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(9<<4)+7);
}
#define init_11_15_9_8
void FASTCALL call_11_15_9_8(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(9<<4)+8);
}
#define init_11_15_9_9
void FASTCALL call_11_15_9_9(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(9<<4)+9);
}
#define init_11_15_9_10
void FASTCALL call_11_15_9_10(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(9<<4)+10);
}
#define init_11_15_9_11
void FASTCALL call_11_15_9_11(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(9<<4)+11);
}
#define init_11_15_9_12
void FASTCALL call_11_15_9_12(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(9<<4)+12);
}
#define init_11_15_9_13
void FASTCALL call_11_15_9_13(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(9<<4)+13);
}
#define init_11_15_9_14
void FASTCALL call_11_15_9_14(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(9<<4)+14);
}
#define init_11_15_9_15
void FASTCALL call_11_15_9_15(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(9<<4)+15);
}
#define init_11_15_10_0
void FASTCALL call_11_15_10_0(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(10<<4)+0);
}
#define init_11_15_10_1
void FASTCALL call_11_15_10_1(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(10<<4)+1);
}
#define init_11_15_10_2
void FASTCALL call_11_15_10_2(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(10<<4)+2);
}
#define init_11_15_10_3
void FASTCALL call_11_15_10_3(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(10<<4)+3);
}
#define init_11_15_10_4
void FASTCALL call_11_15_10_4(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(10<<4)+4);
}
#define init_11_15_10_5
void FASTCALL call_11_15_10_5(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(10<<4)+5);
}
#define init_11_15_10_6
void FASTCALL call_11_15_10_6(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(10<<4)+6);
}
#define init_11_15_10_7
void FASTCALL call_11_15_10_7(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(10<<4)+7);
}
#define init_11_15_10_8
void FASTCALL call_11_15_10_8(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(10<<4)+8);
}
#define init_11_15_10_9
void FASTCALL call_11_15_10_9(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(10<<4)+9);
}
#define init_11_15_10_10
void FASTCALL call_11_15_10_10(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(10<<4)+10);
}
#define init_11_15_10_11
void FASTCALL call_11_15_10_11(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(10<<4)+11);
}
#define init_11_15_10_12
void FASTCALL call_11_15_10_12(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(10<<4)+12);
}
#define init_11_15_10_13
void FASTCALL call_11_15_10_13(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(10<<4)+13);
}
#define init_11_15_10_14
void FASTCALL call_11_15_10_14(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(10<<4)+14);
}
#define init_11_15_10_15
void FASTCALL call_11_15_10_15(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(10<<4)+15);
}
#define init_11_15_11_0
void FASTCALL call_11_15_11_0(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(11<<4)+0);
}
#define init_11_15_11_1
void FASTCALL call_11_15_11_1(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(11<<4)+1);
}
#define init_11_15_11_2
void FASTCALL call_11_15_11_2(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(11<<4)+2);
}
#define init_11_15_11_3
void FASTCALL call_11_15_11_3(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(11<<4)+3);
}
#define init_11_15_11_4
void FASTCALL call_11_15_11_4(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(11<<4)+4);
}
#define init_11_15_11_5
void FASTCALL call_11_15_11_5(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(11<<4)+5);
}
#define init_11_15_11_6
void FASTCALL call_11_15_11_6(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(11<<4)+6);
}
#define init_11_15_11_7
void FASTCALL call_11_15_11_7(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(11<<4)+7);
}
#define init_11_15_11_8
void FASTCALL call_11_15_11_8(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(11<<4)+8);
}
#define init_11_15_11_9
void FASTCALL call_11_15_11_9(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(11<<4)+9);
}
#define init_11_15_11_10
void FASTCALL call_11_15_11_10(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(11<<4)+10);
}
#define init_11_15_11_11
void FASTCALL call_11_15_11_11(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(11<<4)+11);
}
#define init_11_15_11_12
void FASTCALL call_11_15_11_12(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(11<<4)+12);
}
#define init_11_15_11_13
void FASTCALL call_11_15_11_13(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(11<<4)+13);
}
#define init_11_15_11_14
void FASTCALL call_11_15_11_14(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(11<<4)+14);
}
#define init_11_15_11_15
void FASTCALL call_11_15_11_15(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(11<<4)+15);
}
#define init_11_15_12_0
void FASTCALL call_11_15_12_0(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(12<<4)+0);
}
#define init_11_15_12_1
void FASTCALL call_11_15_12_1(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(12<<4)+1);
}
#define init_11_15_12_2
void FASTCALL call_11_15_12_2(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(12<<4)+2);
}
#define init_11_15_12_3
void FASTCALL call_11_15_12_3(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(12<<4)+3);
}
#define init_11_15_12_4
void FASTCALL call_11_15_12_4(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(12<<4)+4);
}
#define init_11_15_12_5
void FASTCALL call_11_15_12_5(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(12<<4)+5);
}
#define init_11_15_12_6
void FASTCALL call_11_15_12_6(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(12<<4)+6);
}
#define init_11_15_12_7
void FASTCALL call_11_15_12_7(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(12<<4)+7);
}
#define init_11_15_12_8
void FASTCALL call_11_15_12_8(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(12<<4)+8);
}
#define init_11_15_12_9
void FASTCALL call_11_15_12_9(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(12<<4)+9);
}
#define init_11_15_12_10
void FASTCALL call_11_15_12_10(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(12<<4)+10);
}
#define init_11_15_12_11
void FASTCALL call_11_15_12_11(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(12<<4)+11);
}
#define init_11_15_12_12
void FASTCALL call_11_15_12_12(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(12<<4)+12);
}
#define init_11_15_12_13
void FASTCALL call_11_15_12_13(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(12<<4)+13);
}
#define init_11_15_12_14
void FASTCALL call_11_15_12_14(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(12<<4)+14);
}
#define init_11_15_12_15
void FASTCALL call_11_15_12_15(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(12<<4)+15);
}
#define init_11_15_13_0
void FASTCALL call_11_15_13_0(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(13<<4)+0);
}
#define init_11_15_13_1
void FASTCALL call_11_15_13_1(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(13<<4)+1);
}
#define init_11_15_13_2
void FASTCALL call_11_15_13_2(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(13<<4)+2);
}
#define init_11_15_13_3
void FASTCALL call_11_15_13_3(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(13<<4)+3);
}
#define init_11_15_13_4
void FASTCALL call_11_15_13_4(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(13<<4)+4);
}
#define init_11_15_13_5
void FASTCALL call_11_15_13_5(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(13<<4)+5);
}
#define init_11_15_13_6
void FASTCALL call_11_15_13_6(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(13<<4)+6);
}
#define init_11_15_13_7
void FASTCALL call_11_15_13_7(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(13<<4)+7);
}
#define init_11_15_13_8
void FASTCALL call_11_15_13_8(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(13<<4)+8);
}
#define init_11_15_13_9
void FASTCALL call_11_15_13_9(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(13<<4)+9);
}
#define init_11_15_13_10
void FASTCALL call_11_15_13_10(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(13<<4)+10);
}
#define init_11_15_13_11
void FASTCALL call_11_15_13_11(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(13<<4)+11);
}
#define init_11_15_13_12
void FASTCALL call_11_15_13_12(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(13<<4)+12);
}
#define init_11_15_13_13
void FASTCALL call_11_15_13_13(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(13<<4)+13);
}
#define init_11_15_13_14
void FASTCALL call_11_15_13_14(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(13<<4)+14);
}
#define init_11_15_13_15
void FASTCALL call_11_15_13_15(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(13<<4)+15);
}
#define init_11_15_14_0
void FASTCALL call_11_15_14_0(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(14<<4)+0);
}
#define init_11_15_14_1
void FASTCALL call_11_15_14_1(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(14<<4)+1);
}
#define init_11_15_14_2
void FASTCALL call_11_15_14_2(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(14<<4)+2);
}
#define init_11_15_14_3
void FASTCALL call_11_15_14_3(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(14<<4)+3);
}
#define init_11_15_14_4
void FASTCALL call_11_15_14_4(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(14<<4)+4);
}
#define init_11_15_14_5
void FASTCALL call_11_15_14_5(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(14<<4)+5);
}
#define init_11_15_14_6
void FASTCALL call_11_15_14_6(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(14<<4)+6);
}
#define init_11_15_14_7
void FASTCALL call_11_15_14_7(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(14<<4)+7);
}
#define init_11_15_14_8
void FASTCALL call_11_15_14_8(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(14<<4)+8);
}
#define init_11_15_14_9
void FASTCALL call_11_15_14_9(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(14<<4)+9);
}
#define init_11_15_14_10
void FASTCALL call_11_15_14_10(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(14<<4)+10);
}
#define init_11_15_14_11
void FASTCALL call_11_15_14_11(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(14<<4)+11);
}
#define init_11_15_14_12
void FASTCALL call_11_15_14_12(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(14<<4)+12);
}
#define init_11_15_14_13
void FASTCALL call_11_15_14_13(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(14<<4)+13);
}
#define init_11_15_14_14
void FASTCALL call_11_15_14_14(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(14<<4)+14);
}
#define init_11_15_14_15
void FASTCALL call_11_15_14_15(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(14<<4)+15);
}
#define init_11_15_15_0
void FASTCALL call_11_15_15_0(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(15<<4)+0);
}
#define init_11_15_15_1
void FASTCALL call_11_15_15_1(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(15<<4)+1);
}
#define init_11_15_15_2
void FASTCALL call_11_15_15_2(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(15<<4)+2);
}
#define init_11_15_15_3
void FASTCALL call_11_15_15_3(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(15<<4)+3);
}
#define init_11_15_15_4
void FASTCALL call_11_15_15_4(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(15<<4)+4);
}
#define init_11_15_15_5
void FASTCALL call_11_15_15_5(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(15<<4)+5);
}
#define init_11_15_15_6
void FASTCALL call_11_15_15_6(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(15<<4)+6);
}
#define init_11_15_15_7
void FASTCALL call_11_15_15_7(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(15<<4)+7);
}
#define init_11_15_15_8
void FASTCALL call_11_15_15_8(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(15<<4)+8);
}
#define init_11_15_15_9
void FASTCALL call_11_15_15_9(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(15<<4)+9);
}
#define init_11_15_15_10
void FASTCALL call_11_15_15_10(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(15<<4)+10);
}
#define init_11_15_15_11
void FASTCALL call_11_15_15_11(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(15<<4)+11);
}
#define init_11_15_15_12
void FASTCALL call_11_15_15_12(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(15<<4)+12);
}
#define init_11_15_15_13
void FASTCALL call_11_15_15_13(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(15<<4)+13);
}
#define init_11_15_15_14
void FASTCALL call_11_15_15_14(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(15<<4)+14);
}
#define init_11_15_15_15
void FASTCALL call_11_15_15_15(SH2_struct* sh)
{
  SH2bsr(sh, (15<<8)+(15<<4)+15);
}
#define init_10_0_0_0
void FASTCALL call_10_0_0_0(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(0<<4)+0);
}
#define init_10_0_0_1
void FASTCALL call_10_0_0_1(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(0<<4)+1);
}
#define init_10_0_0_2
void FASTCALL call_10_0_0_2(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(0<<4)+2);
}
#define init_10_0_0_3
void FASTCALL call_10_0_0_3(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(0<<4)+3);
}
#define init_10_0_0_4
void FASTCALL call_10_0_0_4(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(0<<4)+4);
}
#define init_10_0_0_5
void FASTCALL call_10_0_0_5(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(0<<4)+5);
}
#define init_10_0_0_6
void FASTCALL call_10_0_0_6(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(0<<4)+6);
}
#define init_10_0_0_7
void FASTCALL call_10_0_0_7(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(0<<4)+7);
}
#define init_10_0_0_8
void FASTCALL call_10_0_0_8(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(0<<4)+8);
}
#define init_10_0_0_9
void FASTCALL call_10_0_0_9(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(0<<4)+9);
}
#define init_10_0_0_10
void FASTCALL call_10_0_0_10(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(0<<4)+10);
}
#define init_10_0_0_11
void FASTCALL call_10_0_0_11(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(0<<4)+11);
}
#define init_10_0_0_12
void FASTCALL call_10_0_0_12(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(0<<4)+12);
}
#define init_10_0_0_13
void FASTCALL call_10_0_0_13(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(0<<4)+13);
}
#define init_10_0_0_14
void FASTCALL call_10_0_0_14(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(0<<4)+14);
}
#define init_10_0_0_15
void FASTCALL call_10_0_0_15(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(0<<4)+15);
}
#define init_10_0_1_0
void FASTCALL call_10_0_1_0(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(1<<4)+0);
}
#define init_10_0_1_1
void FASTCALL call_10_0_1_1(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(1<<4)+1);
}
#define init_10_0_1_2
void FASTCALL call_10_0_1_2(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(1<<4)+2);
}
#define init_10_0_1_3
void FASTCALL call_10_0_1_3(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(1<<4)+3);
}
#define init_10_0_1_4
void FASTCALL call_10_0_1_4(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(1<<4)+4);
}
#define init_10_0_1_5
void FASTCALL call_10_0_1_5(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(1<<4)+5);
}
#define init_10_0_1_6
void FASTCALL call_10_0_1_6(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(1<<4)+6);
}
#define init_10_0_1_7
void FASTCALL call_10_0_1_7(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(1<<4)+7);
}
#define init_10_0_1_8
void FASTCALL call_10_0_1_8(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(1<<4)+8);
}
#define init_10_0_1_9
void FASTCALL call_10_0_1_9(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(1<<4)+9);
}
#define init_10_0_1_10
void FASTCALL call_10_0_1_10(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(1<<4)+10);
}
#define init_10_0_1_11
void FASTCALL call_10_0_1_11(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(1<<4)+11);
}
#define init_10_0_1_12
void FASTCALL call_10_0_1_12(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(1<<4)+12);
}
#define init_10_0_1_13
void FASTCALL call_10_0_1_13(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(1<<4)+13);
}
#define init_10_0_1_14
void FASTCALL call_10_0_1_14(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(1<<4)+14);
}
#define init_10_0_1_15
void FASTCALL call_10_0_1_15(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(1<<4)+15);
}
#define init_10_0_2_0
void FASTCALL call_10_0_2_0(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(2<<4)+0);
}
#define init_10_0_2_1
void FASTCALL call_10_0_2_1(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(2<<4)+1);
}
#define init_10_0_2_2
void FASTCALL call_10_0_2_2(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(2<<4)+2);
}
#define init_10_0_2_3
void FASTCALL call_10_0_2_3(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(2<<4)+3);
}
#define init_10_0_2_4
void FASTCALL call_10_0_2_4(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(2<<4)+4);
}
#define init_10_0_2_5
void FASTCALL call_10_0_2_5(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(2<<4)+5);
}
#define init_10_0_2_6
void FASTCALL call_10_0_2_6(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(2<<4)+6);
}
#define init_10_0_2_7
void FASTCALL call_10_0_2_7(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(2<<4)+7);
}
#define init_10_0_2_8
void FASTCALL call_10_0_2_8(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(2<<4)+8);
}
#define init_10_0_2_9
void FASTCALL call_10_0_2_9(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(2<<4)+9);
}
#define init_10_0_2_10
void FASTCALL call_10_0_2_10(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(2<<4)+10);
}
#define init_10_0_2_11
void FASTCALL call_10_0_2_11(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(2<<4)+11);
}
#define init_10_0_2_12
void FASTCALL call_10_0_2_12(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(2<<4)+12);
}
#define init_10_0_2_13
void FASTCALL call_10_0_2_13(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(2<<4)+13);
}
#define init_10_0_2_14
void FASTCALL call_10_0_2_14(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(2<<4)+14);
}
#define init_10_0_2_15
void FASTCALL call_10_0_2_15(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(2<<4)+15);
}
#define init_10_0_3_0
void FASTCALL call_10_0_3_0(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(3<<4)+0);
}
#define init_10_0_3_1
void FASTCALL call_10_0_3_1(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(3<<4)+1);
}
#define init_10_0_3_2
void FASTCALL call_10_0_3_2(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(3<<4)+2);
}
#define init_10_0_3_3
void FASTCALL call_10_0_3_3(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(3<<4)+3);
}
#define init_10_0_3_4
void FASTCALL call_10_0_3_4(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(3<<4)+4);
}
#define init_10_0_3_5
void FASTCALL call_10_0_3_5(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(3<<4)+5);
}
#define init_10_0_3_6
void FASTCALL call_10_0_3_6(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(3<<4)+6);
}
#define init_10_0_3_7
void FASTCALL call_10_0_3_7(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(3<<4)+7);
}
#define init_10_0_3_8
void FASTCALL call_10_0_3_8(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(3<<4)+8);
}
#define init_10_0_3_9
void FASTCALL call_10_0_3_9(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(3<<4)+9);
}
#define init_10_0_3_10
void FASTCALL call_10_0_3_10(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(3<<4)+10);
}
#define init_10_0_3_11
void FASTCALL call_10_0_3_11(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(3<<4)+11);
}
#define init_10_0_3_12
void FASTCALL call_10_0_3_12(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(3<<4)+12);
}
#define init_10_0_3_13
void FASTCALL call_10_0_3_13(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(3<<4)+13);
}
#define init_10_0_3_14
void FASTCALL call_10_0_3_14(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(3<<4)+14);
}
#define init_10_0_3_15
void FASTCALL call_10_0_3_15(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(3<<4)+15);
}
#define init_10_0_4_0
void FASTCALL call_10_0_4_0(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(4<<4)+0);
}
#define init_10_0_4_1
void FASTCALL call_10_0_4_1(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(4<<4)+1);
}
#define init_10_0_4_2
void FASTCALL call_10_0_4_2(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(4<<4)+2);
}
#define init_10_0_4_3
void FASTCALL call_10_0_4_3(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(4<<4)+3);
}
#define init_10_0_4_4
void FASTCALL call_10_0_4_4(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(4<<4)+4);
}
#define init_10_0_4_5
void FASTCALL call_10_0_4_5(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(4<<4)+5);
}
#define init_10_0_4_6
void FASTCALL call_10_0_4_6(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(4<<4)+6);
}
#define init_10_0_4_7
void FASTCALL call_10_0_4_7(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(4<<4)+7);
}
#define init_10_0_4_8
void FASTCALL call_10_0_4_8(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(4<<4)+8);
}
#define init_10_0_4_9
void FASTCALL call_10_0_4_9(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(4<<4)+9);
}
#define init_10_0_4_10
void FASTCALL call_10_0_4_10(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(4<<4)+10);
}
#define init_10_0_4_11
void FASTCALL call_10_0_4_11(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(4<<4)+11);
}
#define init_10_0_4_12
void FASTCALL call_10_0_4_12(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(4<<4)+12);
}
#define init_10_0_4_13
void FASTCALL call_10_0_4_13(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(4<<4)+13);
}
#define init_10_0_4_14
void FASTCALL call_10_0_4_14(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(4<<4)+14);
}
#define init_10_0_4_15
void FASTCALL call_10_0_4_15(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(4<<4)+15);
}
#define init_10_0_5_0
void FASTCALL call_10_0_5_0(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(5<<4)+0);
}
#define init_10_0_5_1
void FASTCALL call_10_0_5_1(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(5<<4)+1);
}
#define init_10_0_5_2
void FASTCALL call_10_0_5_2(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(5<<4)+2);
}
#define init_10_0_5_3
void FASTCALL call_10_0_5_3(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(5<<4)+3);
}
#define init_10_0_5_4
void FASTCALL call_10_0_5_4(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(5<<4)+4);
}
#define init_10_0_5_5
void FASTCALL call_10_0_5_5(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(5<<4)+5);
}
#define init_10_0_5_6
void FASTCALL call_10_0_5_6(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(5<<4)+6);
}
#define init_10_0_5_7
void FASTCALL call_10_0_5_7(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(5<<4)+7);
}
#define init_10_0_5_8
void FASTCALL call_10_0_5_8(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(5<<4)+8);
}
#define init_10_0_5_9
void FASTCALL call_10_0_5_9(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(5<<4)+9);
}
#define init_10_0_5_10
void FASTCALL call_10_0_5_10(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(5<<4)+10);
}
#define init_10_0_5_11
void FASTCALL call_10_0_5_11(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(5<<4)+11);
}
#define init_10_0_5_12
void FASTCALL call_10_0_5_12(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(5<<4)+12);
}
#define init_10_0_5_13
void FASTCALL call_10_0_5_13(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(5<<4)+13);
}
#define init_10_0_5_14
void FASTCALL call_10_0_5_14(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(5<<4)+14);
}
#define init_10_0_5_15
void FASTCALL call_10_0_5_15(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(5<<4)+15);
}
#define init_10_0_6_0
void FASTCALL call_10_0_6_0(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(6<<4)+0);
}
#define init_10_0_6_1
void FASTCALL call_10_0_6_1(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(6<<4)+1);
}
#define init_10_0_6_2
void FASTCALL call_10_0_6_2(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(6<<4)+2);
}
#define init_10_0_6_3
void FASTCALL call_10_0_6_3(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(6<<4)+3);
}
#define init_10_0_6_4
void FASTCALL call_10_0_6_4(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(6<<4)+4);
}
#define init_10_0_6_5
void FASTCALL call_10_0_6_5(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(6<<4)+5);
}
#define init_10_0_6_6
void FASTCALL call_10_0_6_6(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(6<<4)+6);
}
#define init_10_0_6_7
void FASTCALL call_10_0_6_7(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(6<<4)+7);
}
#define init_10_0_6_8
void FASTCALL call_10_0_6_8(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(6<<4)+8);
}
#define init_10_0_6_9
void FASTCALL call_10_0_6_9(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(6<<4)+9);
}
#define init_10_0_6_10
void FASTCALL call_10_0_6_10(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(6<<4)+10);
}
#define init_10_0_6_11
void FASTCALL call_10_0_6_11(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(6<<4)+11);
}
#define init_10_0_6_12
void FASTCALL call_10_0_6_12(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(6<<4)+12);
}
#define init_10_0_6_13
void FASTCALL call_10_0_6_13(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(6<<4)+13);
}
#define init_10_0_6_14
void FASTCALL call_10_0_6_14(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(6<<4)+14);
}
#define init_10_0_6_15
void FASTCALL call_10_0_6_15(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(6<<4)+15);
}
#define init_10_0_7_0
void FASTCALL call_10_0_7_0(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(7<<4)+0);
}
#define init_10_0_7_1
void FASTCALL call_10_0_7_1(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(7<<4)+1);
}
#define init_10_0_7_2
void FASTCALL call_10_0_7_2(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(7<<4)+2);
}
#define init_10_0_7_3
void FASTCALL call_10_0_7_3(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(7<<4)+3);
}
#define init_10_0_7_4
void FASTCALL call_10_0_7_4(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(7<<4)+4);
}
#define init_10_0_7_5
void FASTCALL call_10_0_7_5(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(7<<4)+5);
}
#define init_10_0_7_6
void FASTCALL call_10_0_7_6(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(7<<4)+6);
}
#define init_10_0_7_7
void FASTCALL call_10_0_7_7(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(7<<4)+7);
}
#define init_10_0_7_8
void FASTCALL call_10_0_7_8(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(7<<4)+8);
}
#define init_10_0_7_9
void FASTCALL call_10_0_7_9(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(7<<4)+9);
}
#define init_10_0_7_10
void FASTCALL call_10_0_7_10(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(7<<4)+10);
}
#define init_10_0_7_11
void FASTCALL call_10_0_7_11(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(7<<4)+11);
}
#define init_10_0_7_12
void FASTCALL call_10_0_7_12(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(7<<4)+12);
}
#define init_10_0_7_13
void FASTCALL call_10_0_7_13(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(7<<4)+13);
}
#define init_10_0_7_14
void FASTCALL call_10_0_7_14(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(7<<4)+14);
}
#define init_10_0_7_15
void FASTCALL call_10_0_7_15(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(7<<4)+15);
}
#define init_10_0_8_0
void FASTCALL call_10_0_8_0(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(8<<4)+0);
}
#define init_10_0_8_1
void FASTCALL call_10_0_8_1(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(8<<4)+1);
}
#define init_10_0_8_2
void FASTCALL call_10_0_8_2(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(8<<4)+2);
}
#define init_10_0_8_3
void FASTCALL call_10_0_8_3(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(8<<4)+3);
}
#define init_10_0_8_4
void FASTCALL call_10_0_8_4(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(8<<4)+4);
}
#define init_10_0_8_5
void FASTCALL call_10_0_8_5(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(8<<4)+5);
}
#define init_10_0_8_6
void FASTCALL call_10_0_8_6(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(8<<4)+6);
}
#define init_10_0_8_7
void FASTCALL call_10_0_8_7(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(8<<4)+7);
}
#define init_10_0_8_8
void FASTCALL call_10_0_8_8(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(8<<4)+8);
}
#define init_10_0_8_9
void FASTCALL call_10_0_8_9(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(8<<4)+9);
}
#define init_10_0_8_10
void FASTCALL call_10_0_8_10(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(8<<4)+10);
}
#define init_10_0_8_11
void FASTCALL call_10_0_8_11(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(8<<4)+11);
}
#define init_10_0_8_12
void FASTCALL call_10_0_8_12(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(8<<4)+12);
}
#define init_10_0_8_13
void FASTCALL call_10_0_8_13(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(8<<4)+13);
}
#define init_10_0_8_14
void FASTCALL call_10_0_8_14(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(8<<4)+14);
}
#define init_10_0_8_15
void FASTCALL call_10_0_8_15(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(8<<4)+15);
}
#define init_10_0_9_0
void FASTCALL call_10_0_9_0(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(9<<4)+0);
}
#define init_10_0_9_1
void FASTCALL call_10_0_9_1(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(9<<4)+1);
}
#define init_10_0_9_2
void FASTCALL call_10_0_9_2(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(9<<4)+2);
}
#define init_10_0_9_3
void FASTCALL call_10_0_9_3(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(9<<4)+3);
}
#define init_10_0_9_4
void FASTCALL call_10_0_9_4(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(9<<4)+4);
}
#define init_10_0_9_5
void FASTCALL call_10_0_9_5(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(9<<4)+5);
}
#define init_10_0_9_6
void FASTCALL call_10_0_9_6(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(9<<4)+6);
}
#define init_10_0_9_7
void FASTCALL call_10_0_9_7(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(9<<4)+7);
}
#define init_10_0_9_8
void FASTCALL call_10_0_9_8(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(9<<4)+8);
}
#define init_10_0_9_9
void FASTCALL call_10_0_9_9(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(9<<4)+9);
}
#define init_10_0_9_10
void FASTCALL call_10_0_9_10(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(9<<4)+10);
}
#define init_10_0_9_11
void FASTCALL call_10_0_9_11(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(9<<4)+11);
}
#define init_10_0_9_12
void FASTCALL call_10_0_9_12(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(9<<4)+12);
}
#define init_10_0_9_13
void FASTCALL call_10_0_9_13(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(9<<4)+13);
}
#define init_10_0_9_14
void FASTCALL call_10_0_9_14(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(9<<4)+14);
}
#define init_10_0_9_15
void FASTCALL call_10_0_9_15(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(9<<4)+15);
}
#define init_10_0_10_0
void FASTCALL call_10_0_10_0(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(10<<4)+0);
}
#define init_10_0_10_1
void FASTCALL call_10_0_10_1(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(10<<4)+1);
}
#define init_10_0_10_2
void FASTCALL call_10_0_10_2(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(10<<4)+2);
}
#define init_10_0_10_3
void FASTCALL call_10_0_10_3(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(10<<4)+3);
}
#define init_10_0_10_4
void FASTCALL call_10_0_10_4(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(10<<4)+4);
}
#define init_10_0_10_5
void FASTCALL call_10_0_10_5(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(10<<4)+5);
}
#define init_10_0_10_6
void FASTCALL call_10_0_10_6(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(10<<4)+6);
}
#define init_10_0_10_7
void FASTCALL call_10_0_10_7(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(10<<4)+7);
}
#define init_10_0_10_8
void FASTCALL call_10_0_10_8(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(10<<4)+8);
}
#define init_10_0_10_9
void FASTCALL call_10_0_10_9(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(10<<4)+9);
}
#define init_10_0_10_10
void FASTCALL call_10_0_10_10(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(10<<4)+10);
}
#define init_10_0_10_11
void FASTCALL call_10_0_10_11(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(10<<4)+11);
}
#define init_10_0_10_12
void FASTCALL call_10_0_10_12(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(10<<4)+12);
}
#define init_10_0_10_13
void FASTCALL call_10_0_10_13(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(10<<4)+13);
}
#define init_10_0_10_14
void FASTCALL call_10_0_10_14(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(10<<4)+14);
}
#define init_10_0_10_15
void FASTCALL call_10_0_10_15(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(10<<4)+15);
}
#define init_10_0_11_0
void FASTCALL call_10_0_11_0(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(11<<4)+0);
}
#define init_10_0_11_1
void FASTCALL call_10_0_11_1(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(11<<4)+1);
}
#define init_10_0_11_2
void FASTCALL call_10_0_11_2(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(11<<4)+2);
}
#define init_10_0_11_3
void FASTCALL call_10_0_11_3(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(11<<4)+3);
}
#define init_10_0_11_4
void FASTCALL call_10_0_11_4(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(11<<4)+4);
}
#define init_10_0_11_5
void FASTCALL call_10_0_11_5(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(11<<4)+5);
}
#define init_10_0_11_6
void FASTCALL call_10_0_11_6(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(11<<4)+6);
}
#define init_10_0_11_7
void FASTCALL call_10_0_11_7(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(11<<4)+7);
}
#define init_10_0_11_8
void FASTCALL call_10_0_11_8(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(11<<4)+8);
}
#define init_10_0_11_9
void FASTCALL call_10_0_11_9(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(11<<4)+9);
}
#define init_10_0_11_10
void FASTCALL call_10_0_11_10(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(11<<4)+10);
}
#define init_10_0_11_11
void FASTCALL call_10_0_11_11(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(11<<4)+11);
}
#define init_10_0_11_12
void FASTCALL call_10_0_11_12(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(11<<4)+12);
}
#define init_10_0_11_13
void FASTCALL call_10_0_11_13(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(11<<4)+13);
}
#define init_10_0_11_14
void FASTCALL call_10_0_11_14(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(11<<4)+14);
}
#define init_10_0_11_15
void FASTCALL call_10_0_11_15(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(11<<4)+15);
}
#define init_10_0_12_0
void FASTCALL call_10_0_12_0(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(12<<4)+0);
}
#define init_10_0_12_1
void FASTCALL call_10_0_12_1(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(12<<4)+1);
}
#define init_10_0_12_2
void FASTCALL call_10_0_12_2(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(12<<4)+2);
}
#define init_10_0_12_3
void FASTCALL call_10_0_12_3(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(12<<4)+3);
}
#define init_10_0_12_4
void FASTCALL call_10_0_12_4(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(12<<4)+4);
}
#define init_10_0_12_5
void FASTCALL call_10_0_12_5(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(12<<4)+5);
}
#define init_10_0_12_6
void FASTCALL call_10_0_12_6(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(12<<4)+6);
}
#define init_10_0_12_7
void FASTCALL call_10_0_12_7(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(12<<4)+7);
}
#define init_10_0_12_8
void FASTCALL call_10_0_12_8(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(12<<4)+8);
}
#define init_10_0_12_9
void FASTCALL call_10_0_12_9(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(12<<4)+9);
}
#define init_10_0_12_10
void FASTCALL call_10_0_12_10(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(12<<4)+10);
}
#define init_10_0_12_11
void FASTCALL call_10_0_12_11(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(12<<4)+11);
}
#define init_10_0_12_12
void FASTCALL call_10_0_12_12(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(12<<4)+12);
}
#define init_10_0_12_13
void FASTCALL call_10_0_12_13(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(12<<4)+13);
}
#define init_10_0_12_14
void FASTCALL call_10_0_12_14(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(12<<4)+14);
}
#define init_10_0_12_15
void FASTCALL call_10_0_12_15(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(12<<4)+15);
}
#define init_10_0_13_0
void FASTCALL call_10_0_13_0(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(13<<4)+0);
}
#define init_10_0_13_1
void FASTCALL call_10_0_13_1(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(13<<4)+1);
}
#define init_10_0_13_2
void FASTCALL call_10_0_13_2(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(13<<4)+2);
}
#define init_10_0_13_3
void FASTCALL call_10_0_13_3(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(13<<4)+3);
}
#define init_10_0_13_4
void FASTCALL call_10_0_13_4(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(13<<4)+4);
}
#define init_10_0_13_5
void FASTCALL call_10_0_13_5(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(13<<4)+5);
}
#define init_10_0_13_6
void FASTCALL call_10_0_13_6(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(13<<4)+6);
}
#define init_10_0_13_7
void FASTCALL call_10_0_13_7(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(13<<4)+7);
}
#define init_10_0_13_8
void FASTCALL call_10_0_13_8(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(13<<4)+8);
}
#define init_10_0_13_9
void FASTCALL call_10_0_13_9(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(13<<4)+9);
}
#define init_10_0_13_10
void FASTCALL call_10_0_13_10(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(13<<4)+10);
}
#define init_10_0_13_11
void FASTCALL call_10_0_13_11(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(13<<4)+11);
}
#define init_10_0_13_12
void FASTCALL call_10_0_13_12(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(13<<4)+12);
}
#define init_10_0_13_13
void FASTCALL call_10_0_13_13(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(13<<4)+13);
}
#define init_10_0_13_14
void FASTCALL call_10_0_13_14(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(13<<4)+14);
}
#define init_10_0_13_15
void FASTCALL call_10_0_13_15(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(13<<4)+15);
}
#define init_10_0_14_0
void FASTCALL call_10_0_14_0(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(14<<4)+0);
}
#define init_10_0_14_1
void FASTCALL call_10_0_14_1(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(14<<4)+1);
}
#define init_10_0_14_2
void FASTCALL call_10_0_14_2(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(14<<4)+2);
}
#define init_10_0_14_3
void FASTCALL call_10_0_14_3(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(14<<4)+3);
}
#define init_10_0_14_4
void FASTCALL call_10_0_14_4(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(14<<4)+4);
}
#define init_10_0_14_5
void FASTCALL call_10_0_14_5(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(14<<4)+5);
}
#define init_10_0_14_6
void FASTCALL call_10_0_14_6(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(14<<4)+6);
}
#define init_10_0_14_7
void FASTCALL call_10_0_14_7(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(14<<4)+7);
}
#define init_10_0_14_8
void FASTCALL call_10_0_14_8(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(14<<4)+8);
}
#define init_10_0_14_9
void FASTCALL call_10_0_14_9(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(14<<4)+9);
}
#define init_10_0_14_10
void FASTCALL call_10_0_14_10(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(14<<4)+10);
}
#define init_10_0_14_11
void FASTCALL call_10_0_14_11(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(14<<4)+11);
}
#define init_10_0_14_12
void FASTCALL call_10_0_14_12(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(14<<4)+12);
}
#define init_10_0_14_13
void FASTCALL call_10_0_14_13(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(14<<4)+13);
}
#define init_10_0_14_14
void FASTCALL call_10_0_14_14(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(14<<4)+14);
}
#define init_10_0_14_15
void FASTCALL call_10_0_14_15(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(14<<4)+15);
}
#define init_10_0_15_0
void FASTCALL call_10_0_15_0(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(15<<4)+0);
}
#define init_10_0_15_1
void FASTCALL call_10_0_15_1(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(15<<4)+1);
}
#define init_10_0_15_2
void FASTCALL call_10_0_15_2(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(15<<4)+2);
}
#define init_10_0_15_3
void FASTCALL call_10_0_15_3(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(15<<4)+3);
}
#define init_10_0_15_4
void FASTCALL call_10_0_15_4(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(15<<4)+4);
}
#define init_10_0_15_5
void FASTCALL call_10_0_15_5(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(15<<4)+5);
}
#define init_10_0_15_6
void FASTCALL call_10_0_15_6(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(15<<4)+6);
}
#define init_10_0_15_7
void FASTCALL call_10_0_15_7(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(15<<4)+7);
}
#define init_10_0_15_8
void FASTCALL call_10_0_15_8(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(15<<4)+8);
}
#define init_10_0_15_9
void FASTCALL call_10_0_15_9(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(15<<4)+9);
}
#define init_10_0_15_10
void FASTCALL call_10_0_15_10(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(15<<4)+10);
}
#define init_10_0_15_11
void FASTCALL call_10_0_15_11(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(15<<4)+11);
}
#define init_10_0_15_12
void FASTCALL call_10_0_15_12(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(15<<4)+12);
}
#define init_10_0_15_13
void FASTCALL call_10_0_15_13(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(15<<4)+13);
}
#define init_10_0_15_14
void FASTCALL call_10_0_15_14(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(15<<4)+14);
}
#define init_10_0_15_15
void FASTCALL call_10_0_15_15(SH2_struct* sh)
{
  SH2bra(sh, (0<<8)+(15<<4)+15);
}
#define init_10_1_0_0
void FASTCALL call_10_1_0_0(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(0<<4)+0);
}
#define init_10_1_0_1
void FASTCALL call_10_1_0_1(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(0<<4)+1);
}
#define init_10_1_0_2
void FASTCALL call_10_1_0_2(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(0<<4)+2);
}
#define init_10_1_0_3
void FASTCALL call_10_1_0_3(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(0<<4)+3);
}
#define init_10_1_0_4
void FASTCALL call_10_1_0_4(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(0<<4)+4);
}
#define init_10_1_0_5
void FASTCALL call_10_1_0_5(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(0<<4)+5);
}
#define init_10_1_0_6
void FASTCALL call_10_1_0_6(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(0<<4)+6);
}
#define init_10_1_0_7
void FASTCALL call_10_1_0_7(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(0<<4)+7);
}
#define init_10_1_0_8
void FASTCALL call_10_1_0_8(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(0<<4)+8);
}
#define init_10_1_0_9
void FASTCALL call_10_1_0_9(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(0<<4)+9);
}
#define init_10_1_0_10
void FASTCALL call_10_1_0_10(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(0<<4)+10);
}
#define init_10_1_0_11
void FASTCALL call_10_1_0_11(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(0<<4)+11);
}
#define init_10_1_0_12
void FASTCALL call_10_1_0_12(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(0<<4)+12);
}
#define init_10_1_0_13
void FASTCALL call_10_1_0_13(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(0<<4)+13);
}
#define init_10_1_0_14
void FASTCALL call_10_1_0_14(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(0<<4)+14);
}
#define init_10_1_0_15
void FASTCALL call_10_1_0_15(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(0<<4)+15);
}
#define init_10_1_1_0
void FASTCALL call_10_1_1_0(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(1<<4)+0);
}
#define init_10_1_1_1
void FASTCALL call_10_1_1_1(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(1<<4)+1);
}
#define init_10_1_1_2
void FASTCALL call_10_1_1_2(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(1<<4)+2);
}
#define init_10_1_1_3
void FASTCALL call_10_1_1_3(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(1<<4)+3);
}
#define init_10_1_1_4
void FASTCALL call_10_1_1_4(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(1<<4)+4);
}
#define init_10_1_1_5
void FASTCALL call_10_1_1_5(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(1<<4)+5);
}
#define init_10_1_1_6
void FASTCALL call_10_1_1_6(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(1<<4)+6);
}
#define init_10_1_1_7
void FASTCALL call_10_1_1_7(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(1<<4)+7);
}
#define init_10_1_1_8
void FASTCALL call_10_1_1_8(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(1<<4)+8);
}
#define init_10_1_1_9
void FASTCALL call_10_1_1_9(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(1<<4)+9);
}
#define init_10_1_1_10
void FASTCALL call_10_1_1_10(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(1<<4)+10);
}
#define init_10_1_1_11
void FASTCALL call_10_1_1_11(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(1<<4)+11);
}
#define init_10_1_1_12
void FASTCALL call_10_1_1_12(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(1<<4)+12);
}
#define init_10_1_1_13
void FASTCALL call_10_1_1_13(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(1<<4)+13);
}
#define init_10_1_1_14
void FASTCALL call_10_1_1_14(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(1<<4)+14);
}
#define init_10_1_1_15
void FASTCALL call_10_1_1_15(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(1<<4)+15);
}
#define init_10_1_2_0
void FASTCALL call_10_1_2_0(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(2<<4)+0);
}
#define init_10_1_2_1
void FASTCALL call_10_1_2_1(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(2<<4)+1);
}
#define init_10_1_2_2
void FASTCALL call_10_1_2_2(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(2<<4)+2);
}
#define init_10_1_2_3
void FASTCALL call_10_1_2_3(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(2<<4)+3);
}
#define init_10_1_2_4
void FASTCALL call_10_1_2_4(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(2<<4)+4);
}
#define init_10_1_2_5
void FASTCALL call_10_1_2_5(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(2<<4)+5);
}
#define init_10_1_2_6
void FASTCALL call_10_1_2_6(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(2<<4)+6);
}
#define init_10_1_2_7
void FASTCALL call_10_1_2_7(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(2<<4)+7);
}
#define init_10_1_2_8
void FASTCALL call_10_1_2_8(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(2<<4)+8);
}
#define init_10_1_2_9
void FASTCALL call_10_1_2_9(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(2<<4)+9);
}
#define init_10_1_2_10
void FASTCALL call_10_1_2_10(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(2<<4)+10);
}
#define init_10_1_2_11
void FASTCALL call_10_1_2_11(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(2<<4)+11);
}
#define init_10_1_2_12
void FASTCALL call_10_1_2_12(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(2<<4)+12);
}
#define init_10_1_2_13
void FASTCALL call_10_1_2_13(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(2<<4)+13);
}
#define init_10_1_2_14
void FASTCALL call_10_1_2_14(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(2<<4)+14);
}
#define init_10_1_2_15
void FASTCALL call_10_1_2_15(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(2<<4)+15);
}
#define init_10_1_3_0
void FASTCALL call_10_1_3_0(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(3<<4)+0);
}
#define init_10_1_3_1
void FASTCALL call_10_1_3_1(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(3<<4)+1);
}
#define init_10_1_3_2
void FASTCALL call_10_1_3_2(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(3<<4)+2);
}
#define init_10_1_3_3
void FASTCALL call_10_1_3_3(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(3<<4)+3);
}
#define init_10_1_3_4
void FASTCALL call_10_1_3_4(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(3<<4)+4);
}
#define init_10_1_3_5
void FASTCALL call_10_1_3_5(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(3<<4)+5);
}
#define init_10_1_3_6
void FASTCALL call_10_1_3_6(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(3<<4)+6);
}
#define init_10_1_3_7
void FASTCALL call_10_1_3_7(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(3<<4)+7);
}
#define init_10_1_3_8
void FASTCALL call_10_1_3_8(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(3<<4)+8);
}
#define init_10_1_3_9
void FASTCALL call_10_1_3_9(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(3<<4)+9);
}
#define init_10_1_3_10
void FASTCALL call_10_1_3_10(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(3<<4)+10);
}
#define init_10_1_3_11
void FASTCALL call_10_1_3_11(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(3<<4)+11);
}
#define init_10_1_3_12
void FASTCALL call_10_1_3_12(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(3<<4)+12);
}
#define init_10_1_3_13
void FASTCALL call_10_1_3_13(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(3<<4)+13);
}
#define init_10_1_3_14
void FASTCALL call_10_1_3_14(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(3<<4)+14);
}
#define init_10_1_3_15
void FASTCALL call_10_1_3_15(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(3<<4)+15);
}
#define init_10_1_4_0
void FASTCALL call_10_1_4_0(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(4<<4)+0);
}
#define init_10_1_4_1
void FASTCALL call_10_1_4_1(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(4<<4)+1);
}
#define init_10_1_4_2
void FASTCALL call_10_1_4_2(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(4<<4)+2);
}
#define init_10_1_4_3
void FASTCALL call_10_1_4_3(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(4<<4)+3);
}
#define init_10_1_4_4
void FASTCALL call_10_1_4_4(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(4<<4)+4);
}
#define init_10_1_4_5
void FASTCALL call_10_1_4_5(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(4<<4)+5);
}
#define init_10_1_4_6
void FASTCALL call_10_1_4_6(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(4<<4)+6);
}
#define init_10_1_4_7
void FASTCALL call_10_1_4_7(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(4<<4)+7);
}
#define init_10_1_4_8
void FASTCALL call_10_1_4_8(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(4<<4)+8);
}
#define init_10_1_4_9
void FASTCALL call_10_1_4_9(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(4<<4)+9);
}
#define init_10_1_4_10
void FASTCALL call_10_1_4_10(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(4<<4)+10);
}
#define init_10_1_4_11
void FASTCALL call_10_1_4_11(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(4<<4)+11);
}
#define init_10_1_4_12
void FASTCALL call_10_1_4_12(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(4<<4)+12);
}
#define init_10_1_4_13
void FASTCALL call_10_1_4_13(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(4<<4)+13);
}
#define init_10_1_4_14
void FASTCALL call_10_1_4_14(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(4<<4)+14);
}
#define init_10_1_4_15
void FASTCALL call_10_1_4_15(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(4<<4)+15);
}
#define init_10_1_5_0
void FASTCALL call_10_1_5_0(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(5<<4)+0);
}
#define init_10_1_5_1
void FASTCALL call_10_1_5_1(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(5<<4)+1);
}
#define init_10_1_5_2
void FASTCALL call_10_1_5_2(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(5<<4)+2);
}
#define init_10_1_5_3
void FASTCALL call_10_1_5_3(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(5<<4)+3);
}
#define init_10_1_5_4
void FASTCALL call_10_1_5_4(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(5<<4)+4);
}
#define init_10_1_5_5
void FASTCALL call_10_1_5_5(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(5<<4)+5);
}
#define init_10_1_5_6
void FASTCALL call_10_1_5_6(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(5<<4)+6);
}
#define init_10_1_5_7
void FASTCALL call_10_1_5_7(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(5<<4)+7);
}
#define init_10_1_5_8
void FASTCALL call_10_1_5_8(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(5<<4)+8);
}
#define init_10_1_5_9
void FASTCALL call_10_1_5_9(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(5<<4)+9);
}
#define init_10_1_5_10
void FASTCALL call_10_1_5_10(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(5<<4)+10);
}
#define init_10_1_5_11
void FASTCALL call_10_1_5_11(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(5<<4)+11);
}
#define init_10_1_5_12
void FASTCALL call_10_1_5_12(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(5<<4)+12);
}
#define init_10_1_5_13
void FASTCALL call_10_1_5_13(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(5<<4)+13);
}
#define init_10_1_5_14
void FASTCALL call_10_1_5_14(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(5<<4)+14);
}
#define init_10_1_5_15
void FASTCALL call_10_1_5_15(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(5<<4)+15);
}
#define init_10_1_6_0
void FASTCALL call_10_1_6_0(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(6<<4)+0);
}
#define init_10_1_6_1
void FASTCALL call_10_1_6_1(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(6<<4)+1);
}
#define init_10_1_6_2
void FASTCALL call_10_1_6_2(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(6<<4)+2);
}
#define init_10_1_6_3
void FASTCALL call_10_1_6_3(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(6<<4)+3);
}
#define init_10_1_6_4
void FASTCALL call_10_1_6_4(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(6<<4)+4);
}
#define init_10_1_6_5
void FASTCALL call_10_1_6_5(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(6<<4)+5);
}
#define init_10_1_6_6
void FASTCALL call_10_1_6_6(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(6<<4)+6);
}
#define init_10_1_6_7
void FASTCALL call_10_1_6_7(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(6<<4)+7);
}
#define init_10_1_6_8
void FASTCALL call_10_1_6_8(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(6<<4)+8);
}
#define init_10_1_6_9
void FASTCALL call_10_1_6_9(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(6<<4)+9);
}
#define init_10_1_6_10
void FASTCALL call_10_1_6_10(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(6<<4)+10);
}
#define init_10_1_6_11
void FASTCALL call_10_1_6_11(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(6<<4)+11);
}
#define init_10_1_6_12
void FASTCALL call_10_1_6_12(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(6<<4)+12);
}
#define init_10_1_6_13
void FASTCALL call_10_1_6_13(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(6<<4)+13);
}
#define init_10_1_6_14
void FASTCALL call_10_1_6_14(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(6<<4)+14);
}
#define init_10_1_6_15
void FASTCALL call_10_1_6_15(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(6<<4)+15);
}
#define init_10_1_7_0
void FASTCALL call_10_1_7_0(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(7<<4)+0);
}
#define init_10_1_7_1
void FASTCALL call_10_1_7_1(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(7<<4)+1);
}
#define init_10_1_7_2
void FASTCALL call_10_1_7_2(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(7<<4)+2);
}
#define init_10_1_7_3
void FASTCALL call_10_1_7_3(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(7<<4)+3);
}
#define init_10_1_7_4
void FASTCALL call_10_1_7_4(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(7<<4)+4);
}
#define init_10_1_7_5
void FASTCALL call_10_1_7_5(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(7<<4)+5);
}
#define init_10_1_7_6
void FASTCALL call_10_1_7_6(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(7<<4)+6);
}
#define init_10_1_7_7
void FASTCALL call_10_1_7_7(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(7<<4)+7);
}
#define init_10_1_7_8
void FASTCALL call_10_1_7_8(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(7<<4)+8);
}
#define init_10_1_7_9
void FASTCALL call_10_1_7_9(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(7<<4)+9);
}
#define init_10_1_7_10
void FASTCALL call_10_1_7_10(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(7<<4)+10);
}
#define init_10_1_7_11
void FASTCALL call_10_1_7_11(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(7<<4)+11);
}
#define init_10_1_7_12
void FASTCALL call_10_1_7_12(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(7<<4)+12);
}
#define init_10_1_7_13
void FASTCALL call_10_1_7_13(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(7<<4)+13);
}
#define init_10_1_7_14
void FASTCALL call_10_1_7_14(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(7<<4)+14);
}
#define init_10_1_7_15
void FASTCALL call_10_1_7_15(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(7<<4)+15);
}
#define init_10_1_8_0
void FASTCALL call_10_1_8_0(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(8<<4)+0);
}
#define init_10_1_8_1
void FASTCALL call_10_1_8_1(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(8<<4)+1);
}
#define init_10_1_8_2
void FASTCALL call_10_1_8_2(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(8<<4)+2);
}
#define init_10_1_8_3
void FASTCALL call_10_1_8_3(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(8<<4)+3);
}
#define init_10_1_8_4
void FASTCALL call_10_1_8_4(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(8<<4)+4);
}
#define init_10_1_8_5
void FASTCALL call_10_1_8_5(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(8<<4)+5);
}
#define init_10_1_8_6
void FASTCALL call_10_1_8_6(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(8<<4)+6);
}
#define init_10_1_8_7
void FASTCALL call_10_1_8_7(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(8<<4)+7);
}
#define init_10_1_8_8
void FASTCALL call_10_1_8_8(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(8<<4)+8);
}
#define init_10_1_8_9
void FASTCALL call_10_1_8_9(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(8<<4)+9);
}
#define init_10_1_8_10
void FASTCALL call_10_1_8_10(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(8<<4)+10);
}
#define init_10_1_8_11
void FASTCALL call_10_1_8_11(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(8<<4)+11);
}
#define init_10_1_8_12
void FASTCALL call_10_1_8_12(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(8<<4)+12);
}
#define init_10_1_8_13
void FASTCALL call_10_1_8_13(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(8<<4)+13);
}
#define init_10_1_8_14
void FASTCALL call_10_1_8_14(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(8<<4)+14);
}
#define init_10_1_8_15
void FASTCALL call_10_1_8_15(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(8<<4)+15);
}
#define init_10_1_9_0
void FASTCALL call_10_1_9_0(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(9<<4)+0);
}
#define init_10_1_9_1
void FASTCALL call_10_1_9_1(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(9<<4)+1);
}
#define init_10_1_9_2
void FASTCALL call_10_1_9_2(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(9<<4)+2);
}
#define init_10_1_9_3
void FASTCALL call_10_1_9_3(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(9<<4)+3);
}
#define init_10_1_9_4
void FASTCALL call_10_1_9_4(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(9<<4)+4);
}
#define init_10_1_9_5
void FASTCALL call_10_1_9_5(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(9<<4)+5);
}
#define init_10_1_9_6
void FASTCALL call_10_1_9_6(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(9<<4)+6);
}
#define init_10_1_9_7
void FASTCALL call_10_1_9_7(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(9<<4)+7);
}
#define init_10_1_9_8
void FASTCALL call_10_1_9_8(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(9<<4)+8);
}
#define init_10_1_9_9
void FASTCALL call_10_1_9_9(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(9<<4)+9);
}
#define init_10_1_9_10
void FASTCALL call_10_1_9_10(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(9<<4)+10);
}
#define init_10_1_9_11
void FASTCALL call_10_1_9_11(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(9<<4)+11);
}
#define init_10_1_9_12
void FASTCALL call_10_1_9_12(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(9<<4)+12);
}
#define init_10_1_9_13
void FASTCALL call_10_1_9_13(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(9<<4)+13);
}
#define init_10_1_9_14
void FASTCALL call_10_1_9_14(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(9<<4)+14);
}
#define init_10_1_9_15
void FASTCALL call_10_1_9_15(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(9<<4)+15);
}
#define init_10_1_10_0
void FASTCALL call_10_1_10_0(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(10<<4)+0);
}
#define init_10_1_10_1
void FASTCALL call_10_1_10_1(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(10<<4)+1);
}
#define init_10_1_10_2
void FASTCALL call_10_1_10_2(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(10<<4)+2);
}
#define init_10_1_10_3
void FASTCALL call_10_1_10_3(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(10<<4)+3);
}
#define init_10_1_10_4
void FASTCALL call_10_1_10_4(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(10<<4)+4);
}
#define init_10_1_10_5
void FASTCALL call_10_1_10_5(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(10<<4)+5);
}
#define init_10_1_10_6
void FASTCALL call_10_1_10_6(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(10<<4)+6);
}
#define init_10_1_10_7
void FASTCALL call_10_1_10_7(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(10<<4)+7);
}
#define init_10_1_10_8
void FASTCALL call_10_1_10_8(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(10<<4)+8);
}
#define init_10_1_10_9
void FASTCALL call_10_1_10_9(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(10<<4)+9);
}
#define init_10_1_10_10
void FASTCALL call_10_1_10_10(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(10<<4)+10);
}
#define init_10_1_10_11
void FASTCALL call_10_1_10_11(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(10<<4)+11);
}
#define init_10_1_10_12
void FASTCALL call_10_1_10_12(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(10<<4)+12);
}
#define init_10_1_10_13
void FASTCALL call_10_1_10_13(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(10<<4)+13);
}
#define init_10_1_10_14
void FASTCALL call_10_1_10_14(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(10<<4)+14);
}
#define init_10_1_10_15
void FASTCALL call_10_1_10_15(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(10<<4)+15);
}
#define init_10_1_11_0
void FASTCALL call_10_1_11_0(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(11<<4)+0);
}
#define init_10_1_11_1
void FASTCALL call_10_1_11_1(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(11<<4)+1);
}
#define init_10_1_11_2
void FASTCALL call_10_1_11_2(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(11<<4)+2);
}
#define init_10_1_11_3
void FASTCALL call_10_1_11_3(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(11<<4)+3);
}
#define init_10_1_11_4
void FASTCALL call_10_1_11_4(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(11<<4)+4);
}
#define init_10_1_11_5
void FASTCALL call_10_1_11_5(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(11<<4)+5);
}
#define init_10_1_11_6
void FASTCALL call_10_1_11_6(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(11<<4)+6);
}
#define init_10_1_11_7
void FASTCALL call_10_1_11_7(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(11<<4)+7);
}
#define init_10_1_11_8
void FASTCALL call_10_1_11_8(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(11<<4)+8);
}
#define init_10_1_11_9
void FASTCALL call_10_1_11_9(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(11<<4)+9);
}
#define init_10_1_11_10
void FASTCALL call_10_1_11_10(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(11<<4)+10);
}
#define init_10_1_11_11
void FASTCALL call_10_1_11_11(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(11<<4)+11);
}
#define init_10_1_11_12
void FASTCALL call_10_1_11_12(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(11<<4)+12);
}
#define init_10_1_11_13
void FASTCALL call_10_1_11_13(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(11<<4)+13);
}
#define init_10_1_11_14
void FASTCALL call_10_1_11_14(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(11<<4)+14);
}
#define init_10_1_11_15
void FASTCALL call_10_1_11_15(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(11<<4)+15);
}
#define init_10_1_12_0
void FASTCALL call_10_1_12_0(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(12<<4)+0);
}
#define init_10_1_12_1
void FASTCALL call_10_1_12_1(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(12<<4)+1);
}
#define init_10_1_12_2
void FASTCALL call_10_1_12_2(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(12<<4)+2);
}
#define init_10_1_12_3
void FASTCALL call_10_1_12_3(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(12<<4)+3);
}
#define init_10_1_12_4
void FASTCALL call_10_1_12_4(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(12<<4)+4);
}
#define init_10_1_12_5
void FASTCALL call_10_1_12_5(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(12<<4)+5);
}
#define init_10_1_12_6
void FASTCALL call_10_1_12_6(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(12<<4)+6);
}
#define init_10_1_12_7
void FASTCALL call_10_1_12_7(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(12<<4)+7);
}
#define init_10_1_12_8
void FASTCALL call_10_1_12_8(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(12<<4)+8);
}
#define init_10_1_12_9
void FASTCALL call_10_1_12_9(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(12<<4)+9);
}
#define init_10_1_12_10
void FASTCALL call_10_1_12_10(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(12<<4)+10);
}
#define init_10_1_12_11
void FASTCALL call_10_1_12_11(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(12<<4)+11);
}
#define init_10_1_12_12
void FASTCALL call_10_1_12_12(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(12<<4)+12);
}
#define init_10_1_12_13
void FASTCALL call_10_1_12_13(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(12<<4)+13);
}
#define init_10_1_12_14
void FASTCALL call_10_1_12_14(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(12<<4)+14);
}
#define init_10_1_12_15
void FASTCALL call_10_1_12_15(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(12<<4)+15);
}
#define init_10_1_13_0
void FASTCALL call_10_1_13_0(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(13<<4)+0);
}
#define init_10_1_13_1
void FASTCALL call_10_1_13_1(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(13<<4)+1);
}
#define init_10_1_13_2
void FASTCALL call_10_1_13_2(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(13<<4)+2);
}
#define init_10_1_13_3
void FASTCALL call_10_1_13_3(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(13<<4)+3);
}
#define init_10_1_13_4
void FASTCALL call_10_1_13_4(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(13<<4)+4);
}
#define init_10_1_13_5
void FASTCALL call_10_1_13_5(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(13<<4)+5);
}
#define init_10_1_13_6
void FASTCALL call_10_1_13_6(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(13<<4)+6);
}
#define init_10_1_13_7
void FASTCALL call_10_1_13_7(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(13<<4)+7);
}
#define init_10_1_13_8
void FASTCALL call_10_1_13_8(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(13<<4)+8);
}
#define init_10_1_13_9
void FASTCALL call_10_1_13_9(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(13<<4)+9);
}
#define init_10_1_13_10
void FASTCALL call_10_1_13_10(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(13<<4)+10);
}
#define init_10_1_13_11
void FASTCALL call_10_1_13_11(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(13<<4)+11);
}
#define init_10_1_13_12
void FASTCALL call_10_1_13_12(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(13<<4)+12);
}
#define init_10_1_13_13
void FASTCALL call_10_1_13_13(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(13<<4)+13);
}
#define init_10_1_13_14
void FASTCALL call_10_1_13_14(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(13<<4)+14);
}
#define init_10_1_13_15
void FASTCALL call_10_1_13_15(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(13<<4)+15);
}
#define init_10_1_14_0
void FASTCALL call_10_1_14_0(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(14<<4)+0);
}
#define init_10_1_14_1
void FASTCALL call_10_1_14_1(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(14<<4)+1);
}
#define init_10_1_14_2
void FASTCALL call_10_1_14_2(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(14<<4)+2);
}
#define init_10_1_14_3
void FASTCALL call_10_1_14_3(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(14<<4)+3);
}
#define init_10_1_14_4
void FASTCALL call_10_1_14_4(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(14<<4)+4);
}
#define init_10_1_14_5
void FASTCALL call_10_1_14_5(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(14<<4)+5);
}
#define init_10_1_14_6
void FASTCALL call_10_1_14_6(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(14<<4)+6);
}
#define init_10_1_14_7
void FASTCALL call_10_1_14_7(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(14<<4)+7);
}
#define init_10_1_14_8
void FASTCALL call_10_1_14_8(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(14<<4)+8);
}
#define init_10_1_14_9
void FASTCALL call_10_1_14_9(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(14<<4)+9);
}
#define init_10_1_14_10
void FASTCALL call_10_1_14_10(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(14<<4)+10);
}
#define init_10_1_14_11
void FASTCALL call_10_1_14_11(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(14<<4)+11);
}
#define init_10_1_14_12
void FASTCALL call_10_1_14_12(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(14<<4)+12);
}
#define init_10_1_14_13
void FASTCALL call_10_1_14_13(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(14<<4)+13);
}
#define init_10_1_14_14
void FASTCALL call_10_1_14_14(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(14<<4)+14);
}
#define init_10_1_14_15
void FASTCALL call_10_1_14_15(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(14<<4)+15);
}
#define init_10_1_15_0
void FASTCALL call_10_1_15_0(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(15<<4)+0);
}
#define init_10_1_15_1
void FASTCALL call_10_1_15_1(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(15<<4)+1);
}
#define init_10_1_15_2
void FASTCALL call_10_1_15_2(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(15<<4)+2);
}
#define init_10_1_15_3
void FASTCALL call_10_1_15_3(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(15<<4)+3);
}
#define init_10_1_15_4
void FASTCALL call_10_1_15_4(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(15<<4)+4);
}
#define init_10_1_15_5
void FASTCALL call_10_1_15_5(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(15<<4)+5);
}
#define init_10_1_15_6
void FASTCALL call_10_1_15_6(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(15<<4)+6);
}
#define init_10_1_15_7
void FASTCALL call_10_1_15_7(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(15<<4)+7);
}
#define init_10_1_15_8
void FASTCALL call_10_1_15_8(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(15<<4)+8);
}
#define init_10_1_15_9
void FASTCALL call_10_1_15_9(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(15<<4)+9);
}
#define init_10_1_15_10
void FASTCALL call_10_1_15_10(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(15<<4)+10);
}
#define init_10_1_15_11
void FASTCALL call_10_1_15_11(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(15<<4)+11);
}
#define init_10_1_15_12
void FASTCALL call_10_1_15_12(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(15<<4)+12);
}
#define init_10_1_15_13
void FASTCALL call_10_1_15_13(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(15<<4)+13);
}
#define init_10_1_15_14
void FASTCALL call_10_1_15_14(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(15<<4)+14);
}
#define init_10_1_15_15
void FASTCALL call_10_1_15_15(SH2_struct* sh)
{
  SH2bra(sh, (1<<8)+(15<<4)+15);
}
#define init_10_2_0_0
void FASTCALL call_10_2_0_0(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(0<<4)+0);
}
#define init_10_2_0_1
void FASTCALL call_10_2_0_1(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(0<<4)+1);
}
#define init_10_2_0_2
void FASTCALL call_10_2_0_2(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(0<<4)+2);
}
#define init_10_2_0_3
void FASTCALL call_10_2_0_3(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(0<<4)+3);
}
#define init_10_2_0_4
void FASTCALL call_10_2_0_4(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(0<<4)+4);
}
#define init_10_2_0_5
void FASTCALL call_10_2_0_5(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(0<<4)+5);
}
#define init_10_2_0_6
void FASTCALL call_10_2_0_6(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(0<<4)+6);
}
#define init_10_2_0_7
void FASTCALL call_10_2_0_7(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(0<<4)+7);
}
#define init_10_2_0_8
void FASTCALL call_10_2_0_8(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(0<<4)+8);
}
#define init_10_2_0_9
void FASTCALL call_10_2_0_9(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(0<<4)+9);
}
#define init_10_2_0_10
void FASTCALL call_10_2_0_10(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(0<<4)+10);
}
#define init_10_2_0_11
void FASTCALL call_10_2_0_11(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(0<<4)+11);
}
#define init_10_2_0_12
void FASTCALL call_10_2_0_12(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(0<<4)+12);
}
#define init_10_2_0_13
void FASTCALL call_10_2_0_13(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(0<<4)+13);
}
#define init_10_2_0_14
void FASTCALL call_10_2_0_14(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(0<<4)+14);
}
#define init_10_2_0_15
void FASTCALL call_10_2_0_15(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(0<<4)+15);
}
#define init_10_2_1_0
void FASTCALL call_10_2_1_0(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(1<<4)+0);
}
#define init_10_2_1_1
void FASTCALL call_10_2_1_1(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(1<<4)+1);
}
#define init_10_2_1_2
void FASTCALL call_10_2_1_2(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(1<<4)+2);
}
#define init_10_2_1_3
void FASTCALL call_10_2_1_3(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(1<<4)+3);
}
#define init_10_2_1_4
void FASTCALL call_10_2_1_4(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(1<<4)+4);
}
#define init_10_2_1_5
void FASTCALL call_10_2_1_5(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(1<<4)+5);
}
#define init_10_2_1_6
void FASTCALL call_10_2_1_6(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(1<<4)+6);
}
#define init_10_2_1_7
void FASTCALL call_10_2_1_7(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(1<<4)+7);
}
#define init_10_2_1_8
void FASTCALL call_10_2_1_8(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(1<<4)+8);
}
#define init_10_2_1_9
void FASTCALL call_10_2_1_9(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(1<<4)+9);
}
#define init_10_2_1_10
void FASTCALL call_10_2_1_10(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(1<<4)+10);
}
#define init_10_2_1_11
void FASTCALL call_10_2_1_11(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(1<<4)+11);
}
#define init_10_2_1_12
void FASTCALL call_10_2_1_12(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(1<<4)+12);
}
#define init_10_2_1_13
void FASTCALL call_10_2_1_13(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(1<<4)+13);
}
#define init_10_2_1_14
void FASTCALL call_10_2_1_14(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(1<<4)+14);
}
#define init_10_2_1_15
void FASTCALL call_10_2_1_15(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(1<<4)+15);
}
#define init_10_2_2_0
void FASTCALL call_10_2_2_0(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(2<<4)+0);
}
#define init_10_2_2_1
void FASTCALL call_10_2_2_1(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(2<<4)+1);
}
#define init_10_2_2_2
void FASTCALL call_10_2_2_2(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(2<<4)+2);
}
#define init_10_2_2_3
void FASTCALL call_10_2_2_3(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(2<<4)+3);
}
#define init_10_2_2_4
void FASTCALL call_10_2_2_4(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(2<<4)+4);
}
#define init_10_2_2_5
void FASTCALL call_10_2_2_5(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(2<<4)+5);
}
#define init_10_2_2_6
void FASTCALL call_10_2_2_6(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(2<<4)+6);
}
#define init_10_2_2_7
void FASTCALL call_10_2_2_7(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(2<<4)+7);
}
#define init_10_2_2_8
void FASTCALL call_10_2_2_8(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(2<<4)+8);
}
#define init_10_2_2_9
void FASTCALL call_10_2_2_9(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(2<<4)+9);
}
#define init_10_2_2_10
void FASTCALL call_10_2_2_10(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(2<<4)+10);
}
#define init_10_2_2_11
void FASTCALL call_10_2_2_11(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(2<<4)+11);
}
#define init_10_2_2_12
void FASTCALL call_10_2_2_12(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(2<<4)+12);
}
#define init_10_2_2_13
void FASTCALL call_10_2_2_13(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(2<<4)+13);
}
#define init_10_2_2_14
void FASTCALL call_10_2_2_14(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(2<<4)+14);
}
#define init_10_2_2_15
void FASTCALL call_10_2_2_15(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(2<<4)+15);
}
#define init_10_2_3_0
void FASTCALL call_10_2_3_0(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(3<<4)+0);
}
#define init_10_2_3_1
void FASTCALL call_10_2_3_1(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(3<<4)+1);
}
#define init_10_2_3_2
void FASTCALL call_10_2_3_2(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(3<<4)+2);
}
#define init_10_2_3_3
void FASTCALL call_10_2_3_3(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(3<<4)+3);
}
#define init_10_2_3_4
void FASTCALL call_10_2_3_4(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(3<<4)+4);
}
#define init_10_2_3_5
void FASTCALL call_10_2_3_5(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(3<<4)+5);
}
#define init_10_2_3_6
void FASTCALL call_10_2_3_6(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(3<<4)+6);
}
#define init_10_2_3_7
void FASTCALL call_10_2_3_7(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(3<<4)+7);
}
#define init_10_2_3_8
void FASTCALL call_10_2_3_8(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(3<<4)+8);
}
#define init_10_2_3_9
void FASTCALL call_10_2_3_9(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(3<<4)+9);
}
#define init_10_2_3_10
void FASTCALL call_10_2_3_10(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(3<<4)+10);
}
#define init_10_2_3_11
void FASTCALL call_10_2_3_11(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(3<<4)+11);
}
#define init_10_2_3_12
void FASTCALL call_10_2_3_12(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(3<<4)+12);
}
#define init_10_2_3_13
void FASTCALL call_10_2_3_13(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(3<<4)+13);
}
#define init_10_2_3_14
void FASTCALL call_10_2_3_14(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(3<<4)+14);
}
#define init_10_2_3_15
void FASTCALL call_10_2_3_15(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(3<<4)+15);
}
#define init_10_2_4_0
void FASTCALL call_10_2_4_0(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(4<<4)+0);
}
#define init_10_2_4_1
void FASTCALL call_10_2_4_1(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(4<<4)+1);
}
#define init_10_2_4_2
void FASTCALL call_10_2_4_2(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(4<<4)+2);
}
#define init_10_2_4_3
void FASTCALL call_10_2_4_3(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(4<<4)+3);
}
#define init_10_2_4_4
void FASTCALL call_10_2_4_4(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(4<<4)+4);
}
#define init_10_2_4_5
void FASTCALL call_10_2_4_5(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(4<<4)+5);
}
#define init_10_2_4_6
void FASTCALL call_10_2_4_6(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(4<<4)+6);
}
#define init_10_2_4_7
void FASTCALL call_10_2_4_7(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(4<<4)+7);
}
#define init_10_2_4_8
void FASTCALL call_10_2_4_8(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(4<<4)+8);
}
#define init_10_2_4_9
void FASTCALL call_10_2_4_9(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(4<<4)+9);
}
#define init_10_2_4_10
void FASTCALL call_10_2_4_10(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(4<<4)+10);
}
#define init_10_2_4_11
void FASTCALL call_10_2_4_11(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(4<<4)+11);
}
#define init_10_2_4_12
void FASTCALL call_10_2_4_12(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(4<<4)+12);
}
#define init_10_2_4_13
void FASTCALL call_10_2_4_13(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(4<<4)+13);
}
#define init_10_2_4_14
void FASTCALL call_10_2_4_14(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(4<<4)+14);
}
#define init_10_2_4_15
void FASTCALL call_10_2_4_15(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(4<<4)+15);
}
#define init_10_2_5_0
void FASTCALL call_10_2_5_0(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(5<<4)+0);
}
#define init_10_2_5_1
void FASTCALL call_10_2_5_1(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(5<<4)+1);
}
#define init_10_2_5_2
void FASTCALL call_10_2_5_2(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(5<<4)+2);
}
#define init_10_2_5_3
void FASTCALL call_10_2_5_3(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(5<<4)+3);
}
#define init_10_2_5_4
void FASTCALL call_10_2_5_4(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(5<<4)+4);
}
#define init_10_2_5_5
void FASTCALL call_10_2_5_5(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(5<<4)+5);
}
#define init_10_2_5_6
void FASTCALL call_10_2_5_6(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(5<<4)+6);
}
#define init_10_2_5_7
void FASTCALL call_10_2_5_7(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(5<<4)+7);
}
#define init_10_2_5_8
void FASTCALL call_10_2_5_8(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(5<<4)+8);
}
#define init_10_2_5_9
void FASTCALL call_10_2_5_9(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(5<<4)+9);
}
#define init_10_2_5_10
void FASTCALL call_10_2_5_10(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(5<<4)+10);
}
#define init_10_2_5_11
void FASTCALL call_10_2_5_11(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(5<<4)+11);
}
#define init_10_2_5_12
void FASTCALL call_10_2_5_12(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(5<<4)+12);
}
#define init_10_2_5_13
void FASTCALL call_10_2_5_13(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(5<<4)+13);
}
#define init_10_2_5_14
void FASTCALL call_10_2_5_14(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(5<<4)+14);
}
#define init_10_2_5_15
void FASTCALL call_10_2_5_15(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(5<<4)+15);
}
#define init_10_2_6_0
void FASTCALL call_10_2_6_0(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(6<<4)+0);
}
#define init_10_2_6_1
void FASTCALL call_10_2_6_1(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(6<<4)+1);
}
#define init_10_2_6_2
void FASTCALL call_10_2_6_2(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(6<<4)+2);
}
#define init_10_2_6_3
void FASTCALL call_10_2_6_3(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(6<<4)+3);
}
#define init_10_2_6_4
void FASTCALL call_10_2_6_4(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(6<<4)+4);
}
#define init_10_2_6_5
void FASTCALL call_10_2_6_5(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(6<<4)+5);
}
#define init_10_2_6_6
void FASTCALL call_10_2_6_6(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(6<<4)+6);
}
#define init_10_2_6_7
void FASTCALL call_10_2_6_7(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(6<<4)+7);
}
#define init_10_2_6_8
void FASTCALL call_10_2_6_8(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(6<<4)+8);
}
#define init_10_2_6_9
void FASTCALL call_10_2_6_9(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(6<<4)+9);
}
#define init_10_2_6_10
void FASTCALL call_10_2_6_10(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(6<<4)+10);
}
#define init_10_2_6_11
void FASTCALL call_10_2_6_11(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(6<<4)+11);
}
#define init_10_2_6_12
void FASTCALL call_10_2_6_12(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(6<<4)+12);
}
#define init_10_2_6_13
void FASTCALL call_10_2_6_13(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(6<<4)+13);
}
#define init_10_2_6_14
void FASTCALL call_10_2_6_14(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(6<<4)+14);
}
#define init_10_2_6_15
void FASTCALL call_10_2_6_15(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(6<<4)+15);
}
#define init_10_2_7_0
void FASTCALL call_10_2_7_0(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(7<<4)+0);
}
#define init_10_2_7_1
void FASTCALL call_10_2_7_1(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(7<<4)+1);
}
#define init_10_2_7_2
void FASTCALL call_10_2_7_2(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(7<<4)+2);
}
#define init_10_2_7_3
void FASTCALL call_10_2_7_3(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(7<<4)+3);
}
#define init_10_2_7_4
void FASTCALL call_10_2_7_4(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(7<<4)+4);
}
#define init_10_2_7_5
void FASTCALL call_10_2_7_5(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(7<<4)+5);
}
#define init_10_2_7_6
void FASTCALL call_10_2_7_6(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(7<<4)+6);
}
#define init_10_2_7_7
void FASTCALL call_10_2_7_7(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(7<<4)+7);
}
#define init_10_2_7_8
void FASTCALL call_10_2_7_8(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(7<<4)+8);
}
#define init_10_2_7_9
void FASTCALL call_10_2_7_9(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(7<<4)+9);
}
#define init_10_2_7_10
void FASTCALL call_10_2_7_10(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(7<<4)+10);
}
#define init_10_2_7_11
void FASTCALL call_10_2_7_11(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(7<<4)+11);
}
#define init_10_2_7_12
void FASTCALL call_10_2_7_12(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(7<<4)+12);
}
#define init_10_2_7_13
void FASTCALL call_10_2_7_13(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(7<<4)+13);
}
#define init_10_2_7_14
void FASTCALL call_10_2_7_14(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(7<<4)+14);
}
#define init_10_2_7_15
void FASTCALL call_10_2_7_15(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(7<<4)+15);
}
#define init_10_2_8_0
void FASTCALL call_10_2_8_0(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(8<<4)+0);
}
#define init_10_2_8_1
void FASTCALL call_10_2_8_1(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(8<<4)+1);
}
#define init_10_2_8_2
void FASTCALL call_10_2_8_2(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(8<<4)+2);
}
#define init_10_2_8_3
void FASTCALL call_10_2_8_3(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(8<<4)+3);
}
#define init_10_2_8_4
void FASTCALL call_10_2_8_4(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(8<<4)+4);
}
#define init_10_2_8_5
void FASTCALL call_10_2_8_5(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(8<<4)+5);
}
#define init_10_2_8_6
void FASTCALL call_10_2_8_6(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(8<<4)+6);
}
#define init_10_2_8_7
void FASTCALL call_10_2_8_7(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(8<<4)+7);
}
#define init_10_2_8_8
void FASTCALL call_10_2_8_8(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(8<<4)+8);
}
#define init_10_2_8_9
void FASTCALL call_10_2_8_9(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(8<<4)+9);
}
#define init_10_2_8_10
void FASTCALL call_10_2_8_10(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(8<<4)+10);
}
#define init_10_2_8_11
void FASTCALL call_10_2_8_11(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(8<<4)+11);
}
#define init_10_2_8_12
void FASTCALL call_10_2_8_12(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(8<<4)+12);
}
#define init_10_2_8_13
void FASTCALL call_10_2_8_13(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(8<<4)+13);
}
#define init_10_2_8_14
void FASTCALL call_10_2_8_14(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(8<<4)+14);
}
#define init_10_2_8_15
void FASTCALL call_10_2_8_15(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(8<<4)+15);
}
#define init_10_2_9_0
void FASTCALL call_10_2_9_0(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(9<<4)+0);
}
#define init_10_2_9_1
void FASTCALL call_10_2_9_1(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(9<<4)+1);
}
#define init_10_2_9_2
void FASTCALL call_10_2_9_2(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(9<<4)+2);
}
#define init_10_2_9_3
void FASTCALL call_10_2_9_3(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(9<<4)+3);
}
#define init_10_2_9_4
void FASTCALL call_10_2_9_4(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(9<<4)+4);
}
#define init_10_2_9_5
void FASTCALL call_10_2_9_5(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(9<<4)+5);
}
#define init_10_2_9_6
void FASTCALL call_10_2_9_6(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(9<<4)+6);
}
#define init_10_2_9_7
void FASTCALL call_10_2_9_7(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(9<<4)+7);
}
#define init_10_2_9_8
void FASTCALL call_10_2_9_8(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(9<<4)+8);
}
#define init_10_2_9_9
void FASTCALL call_10_2_9_9(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(9<<4)+9);
}
#define init_10_2_9_10
void FASTCALL call_10_2_9_10(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(9<<4)+10);
}
#define init_10_2_9_11
void FASTCALL call_10_2_9_11(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(9<<4)+11);
}
#define init_10_2_9_12
void FASTCALL call_10_2_9_12(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(9<<4)+12);
}
#define init_10_2_9_13
void FASTCALL call_10_2_9_13(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(9<<4)+13);
}
#define init_10_2_9_14
void FASTCALL call_10_2_9_14(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(9<<4)+14);
}
#define init_10_2_9_15
void FASTCALL call_10_2_9_15(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(9<<4)+15);
}
#define init_10_2_10_0
void FASTCALL call_10_2_10_0(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(10<<4)+0);
}
#define init_10_2_10_1
void FASTCALL call_10_2_10_1(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(10<<4)+1);
}
#define init_10_2_10_2
void FASTCALL call_10_2_10_2(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(10<<4)+2);
}
#define init_10_2_10_3
void FASTCALL call_10_2_10_3(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(10<<4)+3);
}
#define init_10_2_10_4
void FASTCALL call_10_2_10_4(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(10<<4)+4);
}
#define init_10_2_10_5
void FASTCALL call_10_2_10_5(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(10<<4)+5);
}
#define init_10_2_10_6
void FASTCALL call_10_2_10_6(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(10<<4)+6);
}
#define init_10_2_10_7
void FASTCALL call_10_2_10_7(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(10<<4)+7);
}
#define init_10_2_10_8
void FASTCALL call_10_2_10_8(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(10<<4)+8);
}
#define init_10_2_10_9
void FASTCALL call_10_2_10_9(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(10<<4)+9);
}
#define init_10_2_10_10
void FASTCALL call_10_2_10_10(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(10<<4)+10);
}
#define init_10_2_10_11
void FASTCALL call_10_2_10_11(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(10<<4)+11);
}
#define init_10_2_10_12
void FASTCALL call_10_2_10_12(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(10<<4)+12);
}
#define init_10_2_10_13
void FASTCALL call_10_2_10_13(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(10<<4)+13);
}
#define init_10_2_10_14
void FASTCALL call_10_2_10_14(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(10<<4)+14);
}
#define init_10_2_10_15
void FASTCALL call_10_2_10_15(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(10<<4)+15);
}
#define init_10_2_11_0
void FASTCALL call_10_2_11_0(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(11<<4)+0);
}
#define init_10_2_11_1
void FASTCALL call_10_2_11_1(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(11<<4)+1);
}
#define init_10_2_11_2
void FASTCALL call_10_2_11_2(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(11<<4)+2);
}
#define init_10_2_11_3
void FASTCALL call_10_2_11_3(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(11<<4)+3);
}
#define init_10_2_11_4
void FASTCALL call_10_2_11_4(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(11<<4)+4);
}
#define init_10_2_11_5
void FASTCALL call_10_2_11_5(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(11<<4)+5);
}
#define init_10_2_11_6
void FASTCALL call_10_2_11_6(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(11<<4)+6);
}
#define init_10_2_11_7
void FASTCALL call_10_2_11_7(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(11<<4)+7);
}
#define init_10_2_11_8
void FASTCALL call_10_2_11_8(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(11<<4)+8);
}
#define init_10_2_11_9
void FASTCALL call_10_2_11_9(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(11<<4)+9);
}
#define init_10_2_11_10
void FASTCALL call_10_2_11_10(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(11<<4)+10);
}
#define init_10_2_11_11
void FASTCALL call_10_2_11_11(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(11<<4)+11);
}
#define init_10_2_11_12
void FASTCALL call_10_2_11_12(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(11<<4)+12);
}
#define init_10_2_11_13
void FASTCALL call_10_2_11_13(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(11<<4)+13);
}
#define init_10_2_11_14
void FASTCALL call_10_2_11_14(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(11<<4)+14);
}
#define init_10_2_11_15
void FASTCALL call_10_2_11_15(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(11<<4)+15);
}
#define init_10_2_12_0
void FASTCALL call_10_2_12_0(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(12<<4)+0);
}
#define init_10_2_12_1
void FASTCALL call_10_2_12_1(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(12<<4)+1);
}
#define init_10_2_12_2
void FASTCALL call_10_2_12_2(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(12<<4)+2);
}
#define init_10_2_12_3
void FASTCALL call_10_2_12_3(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(12<<4)+3);
}
#define init_10_2_12_4
void FASTCALL call_10_2_12_4(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(12<<4)+4);
}
#define init_10_2_12_5
void FASTCALL call_10_2_12_5(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(12<<4)+5);
}
#define init_10_2_12_6
void FASTCALL call_10_2_12_6(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(12<<4)+6);
}
#define init_10_2_12_7
void FASTCALL call_10_2_12_7(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(12<<4)+7);
}
#define init_10_2_12_8
void FASTCALL call_10_2_12_8(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(12<<4)+8);
}
#define init_10_2_12_9
void FASTCALL call_10_2_12_9(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(12<<4)+9);
}
#define init_10_2_12_10
void FASTCALL call_10_2_12_10(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(12<<4)+10);
}
#define init_10_2_12_11
void FASTCALL call_10_2_12_11(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(12<<4)+11);
}
#define init_10_2_12_12
void FASTCALL call_10_2_12_12(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(12<<4)+12);
}
#define init_10_2_12_13
void FASTCALL call_10_2_12_13(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(12<<4)+13);
}
#define init_10_2_12_14
void FASTCALL call_10_2_12_14(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(12<<4)+14);
}
#define init_10_2_12_15
void FASTCALL call_10_2_12_15(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(12<<4)+15);
}
#define init_10_2_13_0
void FASTCALL call_10_2_13_0(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(13<<4)+0);
}
#define init_10_2_13_1
void FASTCALL call_10_2_13_1(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(13<<4)+1);
}
#define init_10_2_13_2
void FASTCALL call_10_2_13_2(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(13<<4)+2);
}
#define init_10_2_13_3
void FASTCALL call_10_2_13_3(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(13<<4)+3);
}
#define init_10_2_13_4
void FASTCALL call_10_2_13_4(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(13<<4)+4);
}
#define init_10_2_13_5
void FASTCALL call_10_2_13_5(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(13<<4)+5);
}
#define init_10_2_13_6
void FASTCALL call_10_2_13_6(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(13<<4)+6);
}
#define init_10_2_13_7
void FASTCALL call_10_2_13_7(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(13<<4)+7);
}
#define init_10_2_13_8
void FASTCALL call_10_2_13_8(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(13<<4)+8);
}
#define init_10_2_13_9
void FASTCALL call_10_2_13_9(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(13<<4)+9);
}
#define init_10_2_13_10
void FASTCALL call_10_2_13_10(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(13<<4)+10);
}
#define init_10_2_13_11
void FASTCALL call_10_2_13_11(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(13<<4)+11);
}
#define init_10_2_13_12
void FASTCALL call_10_2_13_12(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(13<<4)+12);
}
#define init_10_2_13_13
void FASTCALL call_10_2_13_13(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(13<<4)+13);
}
#define init_10_2_13_14
void FASTCALL call_10_2_13_14(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(13<<4)+14);
}
#define init_10_2_13_15
void FASTCALL call_10_2_13_15(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(13<<4)+15);
}
#define init_10_2_14_0
void FASTCALL call_10_2_14_0(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(14<<4)+0);
}
#define init_10_2_14_1
void FASTCALL call_10_2_14_1(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(14<<4)+1);
}
#define init_10_2_14_2
void FASTCALL call_10_2_14_2(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(14<<4)+2);
}
#define init_10_2_14_3
void FASTCALL call_10_2_14_3(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(14<<4)+3);
}
#define init_10_2_14_4
void FASTCALL call_10_2_14_4(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(14<<4)+4);
}
#define init_10_2_14_5
void FASTCALL call_10_2_14_5(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(14<<4)+5);
}
#define init_10_2_14_6
void FASTCALL call_10_2_14_6(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(14<<4)+6);
}
#define init_10_2_14_7
void FASTCALL call_10_2_14_7(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(14<<4)+7);
}
#define init_10_2_14_8
void FASTCALL call_10_2_14_8(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(14<<4)+8);
}
#define init_10_2_14_9
void FASTCALL call_10_2_14_9(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(14<<4)+9);
}
#define init_10_2_14_10
void FASTCALL call_10_2_14_10(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(14<<4)+10);
}
#define init_10_2_14_11
void FASTCALL call_10_2_14_11(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(14<<4)+11);
}
#define init_10_2_14_12
void FASTCALL call_10_2_14_12(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(14<<4)+12);
}
#define init_10_2_14_13
void FASTCALL call_10_2_14_13(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(14<<4)+13);
}
#define init_10_2_14_14
void FASTCALL call_10_2_14_14(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(14<<4)+14);
}
#define init_10_2_14_15
void FASTCALL call_10_2_14_15(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(14<<4)+15);
}
#define init_10_2_15_0
void FASTCALL call_10_2_15_0(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(15<<4)+0);
}
#define init_10_2_15_1
void FASTCALL call_10_2_15_1(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(15<<4)+1);
}
#define init_10_2_15_2
void FASTCALL call_10_2_15_2(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(15<<4)+2);
}
#define init_10_2_15_3
void FASTCALL call_10_2_15_3(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(15<<4)+3);
}
#define init_10_2_15_4
void FASTCALL call_10_2_15_4(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(15<<4)+4);
}
#define init_10_2_15_5
void FASTCALL call_10_2_15_5(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(15<<4)+5);
}
#define init_10_2_15_6
void FASTCALL call_10_2_15_6(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(15<<4)+6);
}
#define init_10_2_15_7
void FASTCALL call_10_2_15_7(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(15<<4)+7);
}
#define init_10_2_15_8
void FASTCALL call_10_2_15_8(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(15<<4)+8);
}
#define init_10_2_15_9
void FASTCALL call_10_2_15_9(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(15<<4)+9);
}
#define init_10_2_15_10
void FASTCALL call_10_2_15_10(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(15<<4)+10);
}
#define init_10_2_15_11
void FASTCALL call_10_2_15_11(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(15<<4)+11);
}
#define init_10_2_15_12
void FASTCALL call_10_2_15_12(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(15<<4)+12);
}
#define init_10_2_15_13
void FASTCALL call_10_2_15_13(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(15<<4)+13);
}
#define init_10_2_15_14
void FASTCALL call_10_2_15_14(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(15<<4)+14);
}
#define init_10_2_15_15
void FASTCALL call_10_2_15_15(SH2_struct* sh)
{
  SH2bra(sh, (2<<8)+(15<<4)+15);
}
#define init_10_3_0_0
void FASTCALL call_10_3_0_0(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(0<<4)+0);
}
#define init_10_3_0_1
void FASTCALL call_10_3_0_1(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(0<<4)+1);
}
#define init_10_3_0_2
void FASTCALL call_10_3_0_2(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(0<<4)+2);
}
#define init_10_3_0_3
void FASTCALL call_10_3_0_3(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(0<<4)+3);
}
#define init_10_3_0_4
void FASTCALL call_10_3_0_4(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(0<<4)+4);
}
#define init_10_3_0_5
void FASTCALL call_10_3_0_5(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(0<<4)+5);
}
#define init_10_3_0_6
void FASTCALL call_10_3_0_6(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(0<<4)+6);
}
#define init_10_3_0_7
void FASTCALL call_10_3_0_7(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(0<<4)+7);
}
#define init_10_3_0_8
void FASTCALL call_10_3_0_8(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(0<<4)+8);
}
#define init_10_3_0_9
void FASTCALL call_10_3_0_9(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(0<<4)+9);
}
#define init_10_3_0_10
void FASTCALL call_10_3_0_10(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(0<<4)+10);
}
#define init_10_3_0_11
void FASTCALL call_10_3_0_11(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(0<<4)+11);
}
#define init_10_3_0_12
void FASTCALL call_10_3_0_12(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(0<<4)+12);
}
#define init_10_3_0_13
void FASTCALL call_10_3_0_13(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(0<<4)+13);
}
#define init_10_3_0_14
void FASTCALL call_10_3_0_14(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(0<<4)+14);
}
#define init_10_3_0_15
void FASTCALL call_10_3_0_15(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(0<<4)+15);
}
#define init_10_3_1_0
void FASTCALL call_10_3_1_0(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(1<<4)+0);
}
#define init_10_3_1_1
void FASTCALL call_10_3_1_1(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(1<<4)+1);
}
#define init_10_3_1_2
void FASTCALL call_10_3_1_2(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(1<<4)+2);
}
#define init_10_3_1_3
void FASTCALL call_10_3_1_3(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(1<<4)+3);
}
#define init_10_3_1_4
void FASTCALL call_10_3_1_4(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(1<<4)+4);
}
#define init_10_3_1_5
void FASTCALL call_10_3_1_5(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(1<<4)+5);
}
#define init_10_3_1_6
void FASTCALL call_10_3_1_6(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(1<<4)+6);
}
#define init_10_3_1_7
void FASTCALL call_10_3_1_7(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(1<<4)+7);
}
#define init_10_3_1_8
void FASTCALL call_10_3_1_8(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(1<<4)+8);
}
#define init_10_3_1_9
void FASTCALL call_10_3_1_9(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(1<<4)+9);
}
#define init_10_3_1_10
void FASTCALL call_10_3_1_10(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(1<<4)+10);
}
#define init_10_3_1_11
void FASTCALL call_10_3_1_11(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(1<<4)+11);
}
#define init_10_3_1_12
void FASTCALL call_10_3_1_12(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(1<<4)+12);
}
#define init_10_3_1_13
void FASTCALL call_10_3_1_13(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(1<<4)+13);
}
#define init_10_3_1_14
void FASTCALL call_10_3_1_14(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(1<<4)+14);
}
#define init_10_3_1_15
void FASTCALL call_10_3_1_15(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(1<<4)+15);
}
#define init_10_3_2_0
void FASTCALL call_10_3_2_0(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(2<<4)+0);
}
#define init_10_3_2_1
void FASTCALL call_10_3_2_1(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(2<<4)+1);
}
#define init_10_3_2_2
void FASTCALL call_10_3_2_2(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(2<<4)+2);
}
#define init_10_3_2_3
void FASTCALL call_10_3_2_3(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(2<<4)+3);
}
#define init_10_3_2_4
void FASTCALL call_10_3_2_4(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(2<<4)+4);
}
#define init_10_3_2_5
void FASTCALL call_10_3_2_5(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(2<<4)+5);
}
#define init_10_3_2_6
void FASTCALL call_10_3_2_6(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(2<<4)+6);
}
#define init_10_3_2_7
void FASTCALL call_10_3_2_7(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(2<<4)+7);
}
#define init_10_3_2_8
void FASTCALL call_10_3_2_8(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(2<<4)+8);
}
#define init_10_3_2_9
void FASTCALL call_10_3_2_9(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(2<<4)+9);
}
#define init_10_3_2_10
void FASTCALL call_10_3_2_10(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(2<<4)+10);
}
#define init_10_3_2_11
void FASTCALL call_10_3_2_11(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(2<<4)+11);
}
#define init_10_3_2_12
void FASTCALL call_10_3_2_12(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(2<<4)+12);
}
#define init_10_3_2_13
void FASTCALL call_10_3_2_13(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(2<<4)+13);
}
#define init_10_3_2_14
void FASTCALL call_10_3_2_14(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(2<<4)+14);
}
#define init_10_3_2_15
void FASTCALL call_10_3_2_15(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(2<<4)+15);
}
#define init_10_3_3_0
void FASTCALL call_10_3_3_0(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(3<<4)+0);
}
#define init_10_3_3_1
void FASTCALL call_10_3_3_1(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(3<<4)+1);
}
#define init_10_3_3_2
void FASTCALL call_10_3_3_2(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(3<<4)+2);
}
#define init_10_3_3_3
void FASTCALL call_10_3_3_3(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(3<<4)+3);
}
#define init_10_3_3_4
void FASTCALL call_10_3_3_4(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(3<<4)+4);
}
#define init_10_3_3_5
void FASTCALL call_10_3_3_5(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(3<<4)+5);
}
#define init_10_3_3_6
void FASTCALL call_10_3_3_6(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(3<<4)+6);
}
#define init_10_3_3_7
void FASTCALL call_10_3_3_7(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(3<<4)+7);
}
#define init_10_3_3_8
void FASTCALL call_10_3_3_8(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(3<<4)+8);
}
#define init_10_3_3_9
void FASTCALL call_10_3_3_9(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(3<<4)+9);
}
#define init_10_3_3_10
void FASTCALL call_10_3_3_10(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(3<<4)+10);
}
#define init_10_3_3_11
void FASTCALL call_10_3_3_11(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(3<<4)+11);
}
#define init_10_3_3_12
void FASTCALL call_10_3_3_12(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(3<<4)+12);
}
#define init_10_3_3_13
void FASTCALL call_10_3_3_13(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(3<<4)+13);
}
#define init_10_3_3_14
void FASTCALL call_10_3_3_14(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(3<<4)+14);
}
#define init_10_3_3_15
void FASTCALL call_10_3_3_15(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(3<<4)+15);
}
#define init_10_3_4_0
void FASTCALL call_10_3_4_0(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(4<<4)+0);
}
#define init_10_3_4_1
void FASTCALL call_10_3_4_1(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(4<<4)+1);
}
#define init_10_3_4_2
void FASTCALL call_10_3_4_2(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(4<<4)+2);
}
#define init_10_3_4_3
void FASTCALL call_10_3_4_3(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(4<<4)+3);
}
#define init_10_3_4_4
void FASTCALL call_10_3_4_4(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(4<<4)+4);
}
#define init_10_3_4_5
void FASTCALL call_10_3_4_5(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(4<<4)+5);
}
#define init_10_3_4_6
void FASTCALL call_10_3_4_6(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(4<<4)+6);
}
#define init_10_3_4_7
void FASTCALL call_10_3_4_7(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(4<<4)+7);
}
#define init_10_3_4_8
void FASTCALL call_10_3_4_8(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(4<<4)+8);
}
#define init_10_3_4_9
void FASTCALL call_10_3_4_9(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(4<<4)+9);
}
#define init_10_3_4_10
void FASTCALL call_10_3_4_10(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(4<<4)+10);
}
#define init_10_3_4_11
void FASTCALL call_10_3_4_11(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(4<<4)+11);
}
#define init_10_3_4_12
void FASTCALL call_10_3_4_12(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(4<<4)+12);
}
#define init_10_3_4_13
void FASTCALL call_10_3_4_13(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(4<<4)+13);
}
#define init_10_3_4_14
void FASTCALL call_10_3_4_14(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(4<<4)+14);
}
#define init_10_3_4_15
void FASTCALL call_10_3_4_15(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(4<<4)+15);
}
#define init_10_3_5_0
void FASTCALL call_10_3_5_0(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(5<<4)+0);
}
#define init_10_3_5_1
void FASTCALL call_10_3_5_1(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(5<<4)+1);
}
#define init_10_3_5_2
void FASTCALL call_10_3_5_2(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(5<<4)+2);
}
#define init_10_3_5_3
void FASTCALL call_10_3_5_3(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(5<<4)+3);
}
#define init_10_3_5_4
void FASTCALL call_10_3_5_4(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(5<<4)+4);
}
#define init_10_3_5_5
void FASTCALL call_10_3_5_5(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(5<<4)+5);
}
#define init_10_3_5_6
void FASTCALL call_10_3_5_6(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(5<<4)+6);
}
#define init_10_3_5_7
void FASTCALL call_10_3_5_7(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(5<<4)+7);
}
#define init_10_3_5_8
void FASTCALL call_10_3_5_8(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(5<<4)+8);
}
#define init_10_3_5_9
void FASTCALL call_10_3_5_9(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(5<<4)+9);
}
#define init_10_3_5_10
void FASTCALL call_10_3_5_10(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(5<<4)+10);
}
#define init_10_3_5_11
void FASTCALL call_10_3_5_11(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(5<<4)+11);
}
#define init_10_3_5_12
void FASTCALL call_10_3_5_12(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(5<<4)+12);
}
#define init_10_3_5_13
void FASTCALL call_10_3_5_13(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(5<<4)+13);
}
#define init_10_3_5_14
void FASTCALL call_10_3_5_14(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(5<<4)+14);
}
#define init_10_3_5_15
void FASTCALL call_10_3_5_15(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(5<<4)+15);
}
#define init_10_3_6_0
void FASTCALL call_10_3_6_0(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(6<<4)+0);
}
#define init_10_3_6_1
void FASTCALL call_10_3_6_1(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(6<<4)+1);
}
#define init_10_3_6_2
void FASTCALL call_10_3_6_2(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(6<<4)+2);
}
#define init_10_3_6_3
void FASTCALL call_10_3_6_3(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(6<<4)+3);
}
#define init_10_3_6_4
void FASTCALL call_10_3_6_4(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(6<<4)+4);
}
#define init_10_3_6_5
void FASTCALL call_10_3_6_5(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(6<<4)+5);
}
#define init_10_3_6_6
void FASTCALL call_10_3_6_6(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(6<<4)+6);
}
#define init_10_3_6_7
void FASTCALL call_10_3_6_7(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(6<<4)+7);
}
#define init_10_3_6_8
void FASTCALL call_10_3_6_8(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(6<<4)+8);
}
#define init_10_3_6_9
void FASTCALL call_10_3_6_9(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(6<<4)+9);
}
#define init_10_3_6_10
void FASTCALL call_10_3_6_10(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(6<<4)+10);
}
#define init_10_3_6_11
void FASTCALL call_10_3_6_11(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(6<<4)+11);
}
#define init_10_3_6_12
void FASTCALL call_10_3_6_12(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(6<<4)+12);
}
#define init_10_3_6_13
void FASTCALL call_10_3_6_13(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(6<<4)+13);
}
#define init_10_3_6_14
void FASTCALL call_10_3_6_14(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(6<<4)+14);
}
#define init_10_3_6_15
void FASTCALL call_10_3_6_15(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(6<<4)+15);
}
#define init_10_3_7_0
void FASTCALL call_10_3_7_0(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(7<<4)+0);
}
#define init_10_3_7_1
void FASTCALL call_10_3_7_1(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(7<<4)+1);
}
#define init_10_3_7_2
void FASTCALL call_10_3_7_2(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(7<<4)+2);
}
#define init_10_3_7_3
void FASTCALL call_10_3_7_3(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(7<<4)+3);
}
#define init_10_3_7_4
void FASTCALL call_10_3_7_4(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(7<<4)+4);
}
#define init_10_3_7_5
void FASTCALL call_10_3_7_5(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(7<<4)+5);
}
#define init_10_3_7_6
void FASTCALL call_10_3_7_6(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(7<<4)+6);
}
#define init_10_3_7_7
void FASTCALL call_10_3_7_7(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(7<<4)+7);
}
#define init_10_3_7_8
void FASTCALL call_10_3_7_8(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(7<<4)+8);
}
#define init_10_3_7_9
void FASTCALL call_10_3_7_9(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(7<<4)+9);
}
#define init_10_3_7_10
void FASTCALL call_10_3_7_10(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(7<<4)+10);
}
#define init_10_3_7_11
void FASTCALL call_10_3_7_11(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(7<<4)+11);
}
#define init_10_3_7_12
void FASTCALL call_10_3_7_12(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(7<<4)+12);
}
#define init_10_3_7_13
void FASTCALL call_10_3_7_13(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(7<<4)+13);
}
#define init_10_3_7_14
void FASTCALL call_10_3_7_14(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(7<<4)+14);
}
#define init_10_3_7_15
void FASTCALL call_10_3_7_15(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(7<<4)+15);
}
#define init_10_3_8_0
void FASTCALL call_10_3_8_0(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(8<<4)+0);
}
#define init_10_3_8_1
void FASTCALL call_10_3_8_1(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(8<<4)+1);
}
#define init_10_3_8_2
void FASTCALL call_10_3_8_2(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(8<<4)+2);
}
#define init_10_3_8_3
void FASTCALL call_10_3_8_3(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(8<<4)+3);
}
#define init_10_3_8_4
void FASTCALL call_10_3_8_4(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(8<<4)+4);
}
#define init_10_3_8_5
void FASTCALL call_10_3_8_5(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(8<<4)+5);
}
#define init_10_3_8_6
void FASTCALL call_10_3_8_6(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(8<<4)+6);
}
#define init_10_3_8_7
void FASTCALL call_10_3_8_7(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(8<<4)+7);
}
#define init_10_3_8_8
void FASTCALL call_10_3_8_8(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(8<<4)+8);
}
#define init_10_3_8_9
void FASTCALL call_10_3_8_9(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(8<<4)+9);
}
#define init_10_3_8_10
void FASTCALL call_10_3_8_10(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(8<<4)+10);
}
#define init_10_3_8_11
void FASTCALL call_10_3_8_11(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(8<<4)+11);
}
#define init_10_3_8_12
void FASTCALL call_10_3_8_12(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(8<<4)+12);
}
#define init_10_3_8_13
void FASTCALL call_10_3_8_13(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(8<<4)+13);
}
#define init_10_3_8_14
void FASTCALL call_10_3_8_14(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(8<<4)+14);
}
#define init_10_3_8_15
void FASTCALL call_10_3_8_15(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(8<<4)+15);
}
#define init_10_3_9_0
void FASTCALL call_10_3_9_0(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(9<<4)+0);
}
#define init_10_3_9_1
void FASTCALL call_10_3_9_1(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(9<<4)+1);
}
#define init_10_3_9_2
void FASTCALL call_10_3_9_2(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(9<<4)+2);
}
#define init_10_3_9_3
void FASTCALL call_10_3_9_3(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(9<<4)+3);
}
#define init_10_3_9_4
void FASTCALL call_10_3_9_4(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(9<<4)+4);
}
#define init_10_3_9_5
void FASTCALL call_10_3_9_5(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(9<<4)+5);
}
#define init_10_3_9_6
void FASTCALL call_10_3_9_6(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(9<<4)+6);
}
#define init_10_3_9_7
void FASTCALL call_10_3_9_7(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(9<<4)+7);
}
#define init_10_3_9_8
void FASTCALL call_10_3_9_8(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(9<<4)+8);
}
#define init_10_3_9_9
void FASTCALL call_10_3_9_9(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(9<<4)+9);
}
#define init_10_3_9_10
void FASTCALL call_10_3_9_10(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(9<<4)+10);
}
#define init_10_3_9_11
void FASTCALL call_10_3_9_11(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(9<<4)+11);
}
#define init_10_3_9_12
void FASTCALL call_10_3_9_12(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(9<<4)+12);
}
#define init_10_3_9_13
void FASTCALL call_10_3_9_13(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(9<<4)+13);
}
#define init_10_3_9_14
void FASTCALL call_10_3_9_14(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(9<<4)+14);
}
#define init_10_3_9_15
void FASTCALL call_10_3_9_15(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(9<<4)+15);
}
#define init_10_3_10_0
void FASTCALL call_10_3_10_0(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(10<<4)+0);
}
#define init_10_3_10_1
void FASTCALL call_10_3_10_1(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(10<<4)+1);
}
#define init_10_3_10_2
void FASTCALL call_10_3_10_2(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(10<<4)+2);
}
#define init_10_3_10_3
void FASTCALL call_10_3_10_3(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(10<<4)+3);
}
#define init_10_3_10_4
void FASTCALL call_10_3_10_4(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(10<<4)+4);
}
#define init_10_3_10_5
void FASTCALL call_10_3_10_5(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(10<<4)+5);
}
#define init_10_3_10_6
void FASTCALL call_10_3_10_6(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(10<<4)+6);
}
#define init_10_3_10_7
void FASTCALL call_10_3_10_7(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(10<<4)+7);
}
#define init_10_3_10_8
void FASTCALL call_10_3_10_8(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(10<<4)+8);
}
#define init_10_3_10_9
void FASTCALL call_10_3_10_9(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(10<<4)+9);
}
#define init_10_3_10_10
void FASTCALL call_10_3_10_10(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(10<<4)+10);
}
#define init_10_3_10_11
void FASTCALL call_10_3_10_11(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(10<<4)+11);
}
#define init_10_3_10_12
void FASTCALL call_10_3_10_12(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(10<<4)+12);
}
#define init_10_3_10_13
void FASTCALL call_10_3_10_13(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(10<<4)+13);
}
#define init_10_3_10_14
void FASTCALL call_10_3_10_14(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(10<<4)+14);
}
#define init_10_3_10_15
void FASTCALL call_10_3_10_15(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(10<<4)+15);
}
#define init_10_3_11_0
void FASTCALL call_10_3_11_0(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(11<<4)+0);
}
#define init_10_3_11_1
void FASTCALL call_10_3_11_1(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(11<<4)+1);
}
#define init_10_3_11_2
void FASTCALL call_10_3_11_2(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(11<<4)+2);
}
#define init_10_3_11_3
void FASTCALL call_10_3_11_3(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(11<<4)+3);
}
#define init_10_3_11_4
void FASTCALL call_10_3_11_4(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(11<<4)+4);
}
#define init_10_3_11_5
void FASTCALL call_10_3_11_5(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(11<<4)+5);
}
#define init_10_3_11_6
void FASTCALL call_10_3_11_6(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(11<<4)+6);
}
#define init_10_3_11_7
void FASTCALL call_10_3_11_7(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(11<<4)+7);
}
#define init_10_3_11_8
void FASTCALL call_10_3_11_8(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(11<<4)+8);
}
#define init_10_3_11_9
void FASTCALL call_10_3_11_9(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(11<<4)+9);
}
#define init_10_3_11_10
void FASTCALL call_10_3_11_10(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(11<<4)+10);
}
#define init_10_3_11_11
void FASTCALL call_10_3_11_11(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(11<<4)+11);
}
#define init_10_3_11_12
void FASTCALL call_10_3_11_12(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(11<<4)+12);
}
#define init_10_3_11_13
void FASTCALL call_10_3_11_13(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(11<<4)+13);
}
#define init_10_3_11_14
void FASTCALL call_10_3_11_14(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(11<<4)+14);
}
#define init_10_3_11_15
void FASTCALL call_10_3_11_15(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(11<<4)+15);
}
#define init_10_3_12_0
void FASTCALL call_10_3_12_0(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(12<<4)+0);
}
#define init_10_3_12_1
void FASTCALL call_10_3_12_1(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(12<<4)+1);
}
#define init_10_3_12_2
void FASTCALL call_10_3_12_2(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(12<<4)+2);
}
#define init_10_3_12_3
void FASTCALL call_10_3_12_3(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(12<<4)+3);
}
#define init_10_3_12_4
void FASTCALL call_10_3_12_4(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(12<<4)+4);
}
#define init_10_3_12_5
void FASTCALL call_10_3_12_5(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(12<<4)+5);
}
#define init_10_3_12_6
void FASTCALL call_10_3_12_6(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(12<<4)+6);
}
#define init_10_3_12_7
void FASTCALL call_10_3_12_7(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(12<<4)+7);
}
#define init_10_3_12_8
void FASTCALL call_10_3_12_8(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(12<<4)+8);
}
#define init_10_3_12_9
void FASTCALL call_10_3_12_9(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(12<<4)+9);
}
#define init_10_3_12_10
void FASTCALL call_10_3_12_10(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(12<<4)+10);
}
#define init_10_3_12_11
void FASTCALL call_10_3_12_11(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(12<<4)+11);
}
#define init_10_3_12_12
void FASTCALL call_10_3_12_12(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(12<<4)+12);
}
#define init_10_3_12_13
void FASTCALL call_10_3_12_13(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(12<<4)+13);
}
#define init_10_3_12_14
void FASTCALL call_10_3_12_14(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(12<<4)+14);
}
#define init_10_3_12_15
void FASTCALL call_10_3_12_15(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(12<<4)+15);
}
#define init_10_3_13_0
void FASTCALL call_10_3_13_0(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(13<<4)+0);
}
#define init_10_3_13_1
void FASTCALL call_10_3_13_1(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(13<<4)+1);
}
#define init_10_3_13_2
void FASTCALL call_10_3_13_2(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(13<<4)+2);
}
#define init_10_3_13_3
void FASTCALL call_10_3_13_3(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(13<<4)+3);
}
#define init_10_3_13_4
void FASTCALL call_10_3_13_4(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(13<<4)+4);
}
#define init_10_3_13_5
void FASTCALL call_10_3_13_5(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(13<<4)+5);
}
#define init_10_3_13_6
void FASTCALL call_10_3_13_6(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(13<<4)+6);
}
#define init_10_3_13_7
void FASTCALL call_10_3_13_7(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(13<<4)+7);
}
#define init_10_3_13_8
void FASTCALL call_10_3_13_8(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(13<<4)+8);
}
#define init_10_3_13_9
void FASTCALL call_10_3_13_9(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(13<<4)+9);
}
#define init_10_3_13_10
void FASTCALL call_10_3_13_10(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(13<<4)+10);
}
#define init_10_3_13_11
void FASTCALL call_10_3_13_11(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(13<<4)+11);
}
#define init_10_3_13_12
void FASTCALL call_10_3_13_12(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(13<<4)+12);
}
#define init_10_3_13_13
void FASTCALL call_10_3_13_13(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(13<<4)+13);
}
#define init_10_3_13_14
void FASTCALL call_10_3_13_14(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(13<<4)+14);
}
#define init_10_3_13_15
void FASTCALL call_10_3_13_15(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(13<<4)+15);
}
#define init_10_3_14_0
void FASTCALL call_10_3_14_0(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(14<<4)+0);
}
#define init_10_3_14_1
void FASTCALL call_10_3_14_1(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(14<<4)+1);
}
#define init_10_3_14_2
void FASTCALL call_10_3_14_2(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(14<<4)+2);
}
#define init_10_3_14_3
void FASTCALL call_10_3_14_3(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(14<<4)+3);
}
#define init_10_3_14_4
void FASTCALL call_10_3_14_4(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(14<<4)+4);
}
#define init_10_3_14_5
void FASTCALL call_10_3_14_5(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(14<<4)+5);
}
#define init_10_3_14_6
void FASTCALL call_10_3_14_6(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(14<<4)+6);
}
#define init_10_3_14_7
void FASTCALL call_10_3_14_7(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(14<<4)+7);
}
#define init_10_3_14_8
void FASTCALL call_10_3_14_8(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(14<<4)+8);
}
#define init_10_3_14_9
void FASTCALL call_10_3_14_9(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(14<<4)+9);
}
#define init_10_3_14_10
void FASTCALL call_10_3_14_10(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(14<<4)+10);
}
#define init_10_3_14_11
void FASTCALL call_10_3_14_11(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(14<<4)+11);
}
#define init_10_3_14_12
void FASTCALL call_10_3_14_12(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(14<<4)+12);
}
#define init_10_3_14_13
void FASTCALL call_10_3_14_13(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(14<<4)+13);
}
#define init_10_3_14_14
void FASTCALL call_10_3_14_14(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(14<<4)+14);
}
#define init_10_3_14_15
void FASTCALL call_10_3_14_15(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(14<<4)+15);
}
#define init_10_3_15_0
void FASTCALL call_10_3_15_0(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(15<<4)+0);
}
#define init_10_3_15_1
void FASTCALL call_10_3_15_1(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(15<<4)+1);
}
#define init_10_3_15_2
void FASTCALL call_10_3_15_2(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(15<<4)+2);
}
#define init_10_3_15_3
void FASTCALL call_10_3_15_3(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(15<<4)+3);
}
#define init_10_3_15_4
void FASTCALL call_10_3_15_4(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(15<<4)+4);
}
#define init_10_3_15_5
void FASTCALL call_10_3_15_5(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(15<<4)+5);
}
#define init_10_3_15_6
void FASTCALL call_10_3_15_6(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(15<<4)+6);
}
#define init_10_3_15_7
void FASTCALL call_10_3_15_7(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(15<<4)+7);
}
#define init_10_3_15_8
void FASTCALL call_10_3_15_8(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(15<<4)+8);
}
#define init_10_3_15_9
void FASTCALL call_10_3_15_9(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(15<<4)+9);
}
#define init_10_3_15_10
void FASTCALL call_10_3_15_10(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(15<<4)+10);
}
#define init_10_3_15_11
void FASTCALL call_10_3_15_11(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(15<<4)+11);
}
#define init_10_3_15_12
void FASTCALL call_10_3_15_12(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(15<<4)+12);
}
#define init_10_3_15_13
void FASTCALL call_10_3_15_13(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(15<<4)+13);
}
#define init_10_3_15_14
void FASTCALL call_10_3_15_14(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(15<<4)+14);
}
#define init_10_3_15_15
void FASTCALL call_10_3_15_15(SH2_struct* sh)
{
  SH2bra(sh, (3<<8)+(15<<4)+15);
}
#define init_10_4_0_0
void FASTCALL call_10_4_0_0(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(0<<4)+0);
}
#define init_10_4_0_1
void FASTCALL call_10_4_0_1(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(0<<4)+1);
}
#define init_10_4_0_2
void FASTCALL call_10_4_0_2(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(0<<4)+2);
}
#define init_10_4_0_3
void FASTCALL call_10_4_0_3(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(0<<4)+3);
}
#define init_10_4_0_4
void FASTCALL call_10_4_0_4(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(0<<4)+4);
}
#define init_10_4_0_5
void FASTCALL call_10_4_0_5(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(0<<4)+5);
}
#define init_10_4_0_6
void FASTCALL call_10_4_0_6(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(0<<4)+6);
}
#define init_10_4_0_7
void FASTCALL call_10_4_0_7(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(0<<4)+7);
}
#define init_10_4_0_8
void FASTCALL call_10_4_0_8(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(0<<4)+8);
}
#define init_10_4_0_9
void FASTCALL call_10_4_0_9(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(0<<4)+9);
}
#define init_10_4_0_10
void FASTCALL call_10_4_0_10(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(0<<4)+10);
}
#define init_10_4_0_11
void FASTCALL call_10_4_0_11(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(0<<4)+11);
}
#define init_10_4_0_12
void FASTCALL call_10_4_0_12(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(0<<4)+12);
}
#define init_10_4_0_13
void FASTCALL call_10_4_0_13(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(0<<4)+13);
}
#define init_10_4_0_14
void FASTCALL call_10_4_0_14(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(0<<4)+14);
}
#define init_10_4_0_15
void FASTCALL call_10_4_0_15(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(0<<4)+15);
}
#define init_10_4_1_0
void FASTCALL call_10_4_1_0(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(1<<4)+0);
}
#define init_10_4_1_1
void FASTCALL call_10_4_1_1(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(1<<4)+1);
}
#define init_10_4_1_2
void FASTCALL call_10_4_1_2(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(1<<4)+2);
}
#define init_10_4_1_3
void FASTCALL call_10_4_1_3(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(1<<4)+3);
}
#define init_10_4_1_4
void FASTCALL call_10_4_1_4(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(1<<4)+4);
}
#define init_10_4_1_5
void FASTCALL call_10_4_1_5(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(1<<4)+5);
}
#define init_10_4_1_6
void FASTCALL call_10_4_1_6(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(1<<4)+6);
}
#define init_10_4_1_7
void FASTCALL call_10_4_1_7(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(1<<4)+7);
}
#define init_10_4_1_8
void FASTCALL call_10_4_1_8(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(1<<4)+8);
}
#define init_10_4_1_9
void FASTCALL call_10_4_1_9(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(1<<4)+9);
}
#define init_10_4_1_10
void FASTCALL call_10_4_1_10(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(1<<4)+10);
}
#define init_10_4_1_11
void FASTCALL call_10_4_1_11(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(1<<4)+11);
}
#define init_10_4_1_12
void FASTCALL call_10_4_1_12(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(1<<4)+12);
}
#define init_10_4_1_13
void FASTCALL call_10_4_1_13(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(1<<4)+13);
}
#define init_10_4_1_14
void FASTCALL call_10_4_1_14(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(1<<4)+14);
}
#define init_10_4_1_15
void FASTCALL call_10_4_1_15(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(1<<4)+15);
}
#define init_10_4_2_0
void FASTCALL call_10_4_2_0(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(2<<4)+0);
}
#define init_10_4_2_1
void FASTCALL call_10_4_2_1(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(2<<4)+1);
}
#define init_10_4_2_2
void FASTCALL call_10_4_2_2(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(2<<4)+2);
}
#define init_10_4_2_3
void FASTCALL call_10_4_2_3(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(2<<4)+3);
}
#define init_10_4_2_4
void FASTCALL call_10_4_2_4(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(2<<4)+4);
}
#define init_10_4_2_5
void FASTCALL call_10_4_2_5(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(2<<4)+5);
}
#define init_10_4_2_6
void FASTCALL call_10_4_2_6(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(2<<4)+6);
}
#define init_10_4_2_7
void FASTCALL call_10_4_2_7(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(2<<4)+7);
}
#define init_10_4_2_8
void FASTCALL call_10_4_2_8(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(2<<4)+8);
}
#define init_10_4_2_9
void FASTCALL call_10_4_2_9(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(2<<4)+9);
}
#define init_10_4_2_10
void FASTCALL call_10_4_2_10(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(2<<4)+10);
}
#define init_10_4_2_11
void FASTCALL call_10_4_2_11(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(2<<4)+11);
}
#define init_10_4_2_12
void FASTCALL call_10_4_2_12(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(2<<4)+12);
}
#define init_10_4_2_13
void FASTCALL call_10_4_2_13(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(2<<4)+13);
}
#define init_10_4_2_14
void FASTCALL call_10_4_2_14(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(2<<4)+14);
}
#define init_10_4_2_15
void FASTCALL call_10_4_2_15(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(2<<4)+15);
}
#define init_10_4_3_0
void FASTCALL call_10_4_3_0(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(3<<4)+0);
}
#define init_10_4_3_1
void FASTCALL call_10_4_3_1(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(3<<4)+1);
}
#define init_10_4_3_2
void FASTCALL call_10_4_3_2(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(3<<4)+2);
}
#define init_10_4_3_3
void FASTCALL call_10_4_3_3(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(3<<4)+3);
}
#define init_10_4_3_4
void FASTCALL call_10_4_3_4(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(3<<4)+4);
}
#define init_10_4_3_5
void FASTCALL call_10_4_3_5(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(3<<4)+5);
}
#define init_10_4_3_6
void FASTCALL call_10_4_3_6(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(3<<4)+6);
}
#define init_10_4_3_7
void FASTCALL call_10_4_3_7(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(3<<4)+7);
}
#define init_10_4_3_8
void FASTCALL call_10_4_3_8(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(3<<4)+8);
}
#define init_10_4_3_9
void FASTCALL call_10_4_3_9(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(3<<4)+9);
}
#define init_10_4_3_10
void FASTCALL call_10_4_3_10(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(3<<4)+10);
}
#define init_10_4_3_11
void FASTCALL call_10_4_3_11(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(3<<4)+11);
}
#define init_10_4_3_12
void FASTCALL call_10_4_3_12(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(3<<4)+12);
}
#define init_10_4_3_13
void FASTCALL call_10_4_3_13(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(3<<4)+13);
}
#define init_10_4_3_14
void FASTCALL call_10_4_3_14(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(3<<4)+14);
}
#define init_10_4_3_15
void FASTCALL call_10_4_3_15(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(3<<4)+15);
}
#define init_10_4_4_0
void FASTCALL call_10_4_4_0(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(4<<4)+0);
}
#define init_10_4_4_1
void FASTCALL call_10_4_4_1(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(4<<4)+1);
}
#define init_10_4_4_2
void FASTCALL call_10_4_4_2(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(4<<4)+2);
}
#define init_10_4_4_3
void FASTCALL call_10_4_4_3(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(4<<4)+3);
}
#define init_10_4_4_4
void FASTCALL call_10_4_4_4(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(4<<4)+4);
}
#define init_10_4_4_5
void FASTCALL call_10_4_4_5(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(4<<4)+5);
}
#define init_10_4_4_6
void FASTCALL call_10_4_4_6(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(4<<4)+6);
}
#define init_10_4_4_7
void FASTCALL call_10_4_4_7(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(4<<4)+7);
}
#define init_10_4_4_8
void FASTCALL call_10_4_4_8(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(4<<4)+8);
}
#define init_10_4_4_9
void FASTCALL call_10_4_4_9(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(4<<4)+9);
}
#define init_10_4_4_10
void FASTCALL call_10_4_4_10(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(4<<4)+10);
}
#define init_10_4_4_11
void FASTCALL call_10_4_4_11(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(4<<4)+11);
}
#define init_10_4_4_12
void FASTCALL call_10_4_4_12(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(4<<4)+12);
}
#define init_10_4_4_13
void FASTCALL call_10_4_4_13(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(4<<4)+13);
}
#define init_10_4_4_14
void FASTCALL call_10_4_4_14(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(4<<4)+14);
}
#define init_10_4_4_15
void FASTCALL call_10_4_4_15(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(4<<4)+15);
}
#define init_10_4_5_0
void FASTCALL call_10_4_5_0(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(5<<4)+0);
}
#define init_10_4_5_1
void FASTCALL call_10_4_5_1(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(5<<4)+1);
}
#define init_10_4_5_2
void FASTCALL call_10_4_5_2(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(5<<4)+2);
}
#define init_10_4_5_3
void FASTCALL call_10_4_5_3(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(5<<4)+3);
}
#define init_10_4_5_4
void FASTCALL call_10_4_5_4(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(5<<4)+4);
}
#define init_10_4_5_5
void FASTCALL call_10_4_5_5(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(5<<4)+5);
}
#define init_10_4_5_6
void FASTCALL call_10_4_5_6(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(5<<4)+6);
}
#define init_10_4_5_7
void FASTCALL call_10_4_5_7(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(5<<4)+7);
}
#define init_10_4_5_8
void FASTCALL call_10_4_5_8(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(5<<4)+8);
}
#define init_10_4_5_9
void FASTCALL call_10_4_5_9(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(5<<4)+9);
}
#define init_10_4_5_10
void FASTCALL call_10_4_5_10(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(5<<4)+10);
}
#define init_10_4_5_11
void FASTCALL call_10_4_5_11(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(5<<4)+11);
}
#define init_10_4_5_12
void FASTCALL call_10_4_5_12(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(5<<4)+12);
}
#define init_10_4_5_13
void FASTCALL call_10_4_5_13(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(5<<4)+13);
}
#define init_10_4_5_14
void FASTCALL call_10_4_5_14(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(5<<4)+14);
}
#define init_10_4_5_15
void FASTCALL call_10_4_5_15(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(5<<4)+15);
}
#define init_10_4_6_0
void FASTCALL call_10_4_6_0(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(6<<4)+0);
}
#define init_10_4_6_1
void FASTCALL call_10_4_6_1(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(6<<4)+1);
}
#define init_10_4_6_2
void FASTCALL call_10_4_6_2(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(6<<4)+2);
}
#define init_10_4_6_3
void FASTCALL call_10_4_6_3(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(6<<4)+3);
}
#define init_10_4_6_4
void FASTCALL call_10_4_6_4(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(6<<4)+4);
}
#define init_10_4_6_5
void FASTCALL call_10_4_6_5(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(6<<4)+5);
}
#define init_10_4_6_6
void FASTCALL call_10_4_6_6(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(6<<4)+6);
}
#define init_10_4_6_7
void FASTCALL call_10_4_6_7(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(6<<4)+7);
}
#define init_10_4_6_8
void FASTCALL call_10_4_6_8(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(6<<4)+8);
}
#define init_10_4_6_9
void FASTCALL call_10_4_6_9(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(6<<4)+9);
}
#define init_10_4_6_10
void FASTCALL call_10_4_6_10(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(6<<4)+10);
}
#define init_10_4_6_11
void FASTCALL call_10_4_6_11(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(6<<4)+11);
}
#define init_10_4_6_12
void FASTCALL call_10_4_6_12(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(6<<4)+12);
}
#define init_10_4_6_13
void FASTCALL call_10_4_6_13(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(6<<4)+13);
}
#define init_10_4_6_14
void FASTCALL call_10_4_6_14(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(6<<4)+14);
}
#define init_10_4_6_15
void FASTCALL call_10_4_6_15(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(6<<4)+15);
}
#define init_10_4_7_0
void FASTCALL call_10_4_7_0(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(7<<4)+0);
}
#define init_10_4_7_1
void FASTCALL call_10_4_7_1(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(7<<4)+1);
}
#define init_10_4_7_2
void FASTCALL call_10_4_7_2(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(7<<4)+2);
}
#define init_10_4_7_3
void FASTCALL call_10_4_7_3(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(7<<4)+3);
}
#define init_10_4_7_4
void FASTCALL call_10_4_7_4(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(7<<4)+4);
}
#define init_10_4_7_5
void FASTCALL call_10_4_7_5(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(7<<4)+5);
}
#define init_10_4_7_6
void FASTCALL call_10_4_7_6(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(7<<4)+6);
}
#define init_10_4_7_7
void FASTCALL call_10_4_7_7(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(7<<4)+7);
}
#define init_10_4_7_8
void FASTCALL call_10_4_7_8(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(7<<4)+8);
}
#define init_10_4_7_9
void FASTCALL call_10_4_7_9(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(7<<4)+9);
}
#define init_10_4_7_10
void FASTCALL call_10_4_7_10(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(7<<4)+10);
}
#define init_10_4_7_11
void FASTCALL call_10_4_7_11(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(7<<4)+11);
}
#define init_10_4_7_12
void FASTCALL call_10_4_7_12(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(7<<4)+12);
}
#define init_10_4_7_13
void FASTCALL call_10_4_7_13(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(7<<4)+13);
}
#define init_10_4_7_14
void FASTCALL call_10_4_7_14(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(7<<4)+14);
}
#define init_10_4_7_15
void FASTCALL call_10_4_7_15(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(7<<4)+15);
}
#define init_10_4_8_0
void FASTCALL call_10_4_8_0(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(8<<4)+0);
}
#define init_10_4_8_1
void FASTCALL call_10_4_8_1(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(8<<4)+1);
}
#define init_10_4_8_2
void FASTCALL call_10_4_8_2(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(8<<4)+2);
}
#define init_10_4_8_3
void FASTCALL call_10_4_8_3(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(8<<4)+3);
}
#define init_10_4_8_4
void FASTCALL call_10_4_8_4(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(8<<4)+4);
}
#define init_10_4_8_5
void FASTCALL call_10_4_8_5(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(8<<4)+5);
}
#define init_10_4_8_6
void FASTCALL call_10_4_8_6(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(8<<4)+6);
}
#define init_10_4_8_7
void FASTCALL call_10_4_8_7(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(8<<4)+7);
}
#define init_10_4_8_8
void FASTCALL call_10_4_8_8(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(8<<4)+8);
}
#define init_10_4_8_9
void FASTCALL call_10_4_8_9(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(8<<4)+9);
}
#define init_10_4_8_10
void FASTCALL call_10_4_8_10(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(8<<4)+10);
}
#define init_10_4_8_11
void FASTCALL call_10_4_8_11(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(8<<4)+11);
}
#define init_10_4_8_12
void FASTCALL call_10_4_8_12(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(8<<4)+12);
}
#define init_10_4_8_13
void FASTCALL call_10_4_8_13(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(8<<4)+13);
}
#define init_10_4_8_14
void FASTCALL call_10_4_8_14(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(8<<4)+14);
}
#define init_10_4_8_15
void FASTCALL call_10_4_8_15(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(8<<4)+15);
}
#define init_10_4_9_0
void FASTCALL call_10_4_9_0(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(9<<4)+0);
}
#define init_10_4_9_1
void FASTCALL call_10_4_9_1(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(9<<4)+1);
}
#define init_10_4_9_2
void FASTCALL call_10_4_9_2(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(9<<4)+2);
}
#define init_10_4_9_3
void FASTCALL call_10_4_9_3(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(9<<4)+3);
}
#define init_10_4_9_4
void FASTCALL call_10_4_9_4(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(9<<4)+4);
}
#define init_10_4_9_5
void FASTCALL call_10_4_9_5(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(9<<4)+5);
}
#define init_10_4_9_6
void FASTCALL call_10_4_9_6(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(9<<4)+6);
}
#define init_10_4_9_7
void FASTCALL call_10_4_9_7(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(9<<4)+7);
}
#define init_10_4_9_8
void FASTCALL call_10_4_9_8(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(9<<4)+8);
}
#define init_10_4_9_9
void FASTCALL call_10_4_9_9(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(9<<4)+9);
}
#define init_10_4_9_10
void FASTCALL call_10_4_9_10(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(9<<4)+10);
}
#define init_10_4_9_11
void FASTCALL call_10_4_9_11(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(9<<4)+11);
}
#define init_10_4_9_12
void FASTCALL call_10_4_9_12(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(9<<4)+12);
}
#define init_10_4_9_13
void FASTCALL call_10_4_9_13(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(9<<4)+13);
}
#define init_10_4_9_14
void FASTCALL call_10_4_9_14(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(9<<4)+14);
}
#define init_10_4_9_15
void FASTCALL call_10_4_9_15(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(9<<4)+15);
}
#define init_10_4_10_0
void FASTCALL call_10_4_10_0(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(10<<4)+0);
}
#define init_10_4_10_1
void FASTCALL call_10_4_10_1(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(10<<4)+1);
}
#define init_10_4_10_2
void FASTCALL call_10_4_10_2(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(10<<4)+2);
}
#define init_10_4_10_3
void FASTCALL call_10_4_10_3(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(10<<4)+3);
}
#define init_10_4_10_4
void FASTCALL call_10_4_10_4(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(10<<4)+4);
}
#define init_10_4_10_5
void FASTCALL call_10_4_10_5(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(10<<4)+5);
}
#define init_10_4_10_6
void FASTCALL call_10_4_10_6(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(10<<4)+6);
}
#define init_10_4_10_7
void FASTCALL call_10_4_10_7(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(10<<4)+7);
}
#define init_10_4_10_8
void FASTCALL call_10_4_10_8(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(10<<4)+8);
}
#define init_10_4_10_9
void FASTCALL call_10_4_10_9(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(10<<4)+9);
}
#define init_10_4_10_10
void FASTCALL call_10_4_10_10(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(10<<4)+10);
}
#define init_10_4_10_11
void FASTCALL call_10_4_10_11(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(10<<4)+11);
}
#define init_10_4_10_12
void FASTCALL call_10_4_10_12(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(10<<4)+12);
}
#define init_10_4_10_13
void FASTCALL call_10_4_10_13(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(10<<4)+13);
}
#define init_10_4_10_14
void FASTCALL call_10_4_10_14(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(10<<4)+14);
}
#define init_10_4_10_15
void FASTCALL call_10_4_10_15(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(10<<4)+15);
}
#define init_10_4_11_0
void FASTCALL call_10_4_11_0(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(11<<4)+0);
}
#define init_10_4_11_1
void FASTCALL call_10_4_11_1(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(11<<4)+1);
}
#define init_10_4_11_2
void FASTCALL call_10_4_11_2(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(11<<4)+2);
}
#define init_10_4_11_3
void FASTCALL call_10_4_11_3(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(11<<4)+3);
}
#define init_10_4_11_4
void FASTCALL call_10_4_11_4(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(11<<4)+4);
}
#define init_10_4_11_5
void FASTCALL call_10_4_11_5(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(11<<4)+5);
}
#define init_10_4_11_6
void FASTCALL call_10_4_11_6(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(11<<4)+6);
}
#define init_10_4_11_7
void FASTCALL call_10_4_11_7(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(11<<4)+7);
}
#define init_10_4_11_8
void FASTCALL call_10_4_11_8(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(11<<4)+8);
}
#define init_10_4_11_9
void FASTCALL call_10_4_11_9(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(11<<4)+9);
}
#define init_10_4_11_10
void FASTCALL call_10_4_11_10(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(11<<4)+10);
}
#define init_10_4_11_11
void FASTCALL call_10_4_11_11(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(11<<4)+11);
}
#define init_10_4_11_12
void FASTCALL call_10_4_11_12(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(11<<4)+12);
}
#define init_10_4_11_13
void FASTCALL call_10_4_11_13(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(11<<4)+13);
}
#define init_10_4_11_14
void FASTCALL call_10_4_11_14(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(11<<4)+14);
}
#define init_10_4_11_15
void FASTCALL call_10_4_11_15(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(11<<4)+15);
}
#define init_10_4_12_0
void FASTCALL call_10_4_12_0(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(12<<4)+0);
}
#define init_10_4_12_1
void FASTCALL call_10_4_12_1(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(12<<4)+1);
}
#define init_10_4_12_2
void FASTCALL call_10_4_12_2(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(12<<4)+2);
}
#define init_10_4_12_3
void FASTCALL call_10_4_12_3(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(12<<4)+3);
}
#define init_10_4_12_4
void FASTCALL call_10_4_12_4(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(12<<4)+4);
}
#define init_10_4_12_5
void FASTCALL call_10_4_12_5(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(12<<4)+5);
}
#define init_10_4_12_6
void FASTCALL call_10_4_12_6(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(12<<4)+6);
}
#define init_10_4_12_7
void FASTCALL call_10_4_12_7(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(12<<4)+7);
}
#define init_10_4_12_8
void FASTCALL call_10_4_12_8(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(12<<4)+8);
}
#define init_10_4_12_9
void FASTCALL call_10_4_12_9(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(12<<4)+9);
}
#define init_10_4_12_10
void FASTCALL call_10_4_12_10(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(12<<4)+10);
}
#define init_10_4_12_11
void FASTCALL call_10_4_12_11(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(12<<4)+11);
}
#define init_10_4_12_12
void FASTCALL call_10_4_12_12(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(12<<4)+12);
}
#define init_10_4_12_13
void FASTCALL call_10_4_12_13(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(12<<4)+13);
}
#define init_10_4_12_14
void FASTCALL call_10_4_12_14(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(12<<4)+14);
}
#define init_10_4_12_15
void FASTCALL call_10_4_12_15(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(12<<4)+15);
}
#define init_10_4_13_0
void FASTCALL call_10_4_13_0(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(13<<4)+0);
}
#define init_10_4_13_1
void FASTCALL call_10_4_13_1(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(13<<4)+1);
}
#define init_10_4_13_2
void FASTCALL call_10_4_13_2(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(13<<4)+2);
}
#define init_10_4_13_3
void FASTCALL call_10_4_13_3(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(13<<4)+3);
}
#define init_10_4_13_4
void FASTCALL call_10_4_13_4(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(13<<4)+4);
}
#define init_10_4_13_5
void FASTCALL call_10_4_13_5(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(13<<4)+5);
}
#define init_10_4_13_6
void FASTCALL call_10_4_13_6(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(13<<4)+6);
}
#define init_10_4_13_7
void FASTCALL call_10_4_13_7(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(13<<4)+7);
}
#define init_10_4_13_8
void FASTCALL call_10_4_13_8(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(13<<4)+8);
}
#define init_10_4_13_9
void FASTCALL call_10_4_13_9(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(13<<4)+9);
}
#define init_10_4_13_10
void FASTCALL call_10_4_13_10(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(13<<4)+10);
}
#define init_10_4_13_11
void FASTCALL call_10_4_13_11(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(13<<4)+11);
}
#define init_10_4_13_12
void FASTCALL call_10_4_13_12(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(13<<4)+12);
}
#define init_10_4_13_13
void FASTCALL call_10_4_13_13(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(13<<4)+13);
}
#define init_10_4_13_14
void FASTCALL call_10_4_13_14(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(13<<4)+14);
}
#define init_10_4_13_15
void FASTCALL call_10_4_13_15(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(13<<4)+15);
}
#define init_10_4_14_0
void FASTCALL call_10_4_14_0(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(14<<4)+0);
}
#define init_10_4_14_1
void FASTCALL call_10_4_14_1(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(14<<4)+1);
}
#define init_10_4_14_2
void FASTCALL call_10_4_14_2(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(14<<4)+2);
}
#define init_10_4_14_3
void FASTCALL call_10_4_14_3(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(14<<4)+3);
}
#define init_10_4_14_4
void FASTCALL call_10_4_14_4(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(14<<4)+4);
}
#define init_10_4_14_5
void FASTCALL call_10_4_14_5(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(14<<4)+5);
}
#define init_10_4_14_6
void FASTCALL call_10_4_14_6(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(14<<4)+6);
}
#define init_10_4_14_7
void FASTCALL call_10_4_14_7(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(14<<4)+7);
}
#define init_10_4_14_8
void FASTCALL call_10_4_14_8(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(14<<4)+8);
}
#define init_10_4_14_9
void FASTCALL call_10_4_14_9(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(14<<4)+9);
}
#define init_10_4_14_10
void FASTCALL call_10_4_14_10(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(14<<4)+10);
}
#define init_10_4_14_11
void FASTCALL call_10_4_14_11(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(14<<4)+11);
}
#define init_10_4_14_12
void FASTCALL call_10_4_14_12(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(14<<4)+12);
}
#define init_10_4_14_13
void FASTCALL call_10_4_14_13(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(14<<4)+13);
}
#define init_10_4_14_14
void FASTCALL call_10_4_14_14(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(14<<4)+14);
}
#define init_10_4_14_15
void FASTCALL call_10_4_14_15(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(14<<4)+15);
}
#define init_10_4_15_0
void FASTCALL call_10_4_15_0(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(15<<4)+0);
}
#define init_10_4_15_1
void FASTCALL call_10_4_15_1(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(15<<4)+1);
}
#define init_10_4_15_2
void FASTCALL call_10_4_15_2(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(15<<4)+2);
}
#define init_10_4_15_3
void FASTCALL call_10_4_15_3(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(15<<4)+3);
}
#define init_10_4_15_4
void FASTCALL call_10_4_15_4(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(15<<4)+4);
}
#define init_10_4_15_5
void FASTCALL call_10_4_15_5(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(15<<4)+5);
}
#define init_10_4_15_6
void FASTCALL call_10_4_15_6(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(15<<4)+6);
}
#define init_10_4_15_7
void FASTCALL call_10_4_15_7(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(15<<4)+7);
}
#define init_10_4_15_8
void FASTCALL call_10_4_15_8(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(15<<4)+8);
}
#define init_10_4_15_9
void FASTCALL call_10_4_15_9(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(15<<4)+9);
}
#define init_10_4_15_10
void FASTCALL call_10_4_15_10(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(15<<4)+10);
}
#define init_10_4_15_11
void FASTCALL call_10_4_15_11(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(15<<4)+11);
}
#define init_10_4_15_12
void FASTCALL call_10_4_15_12(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(15<<4)+12);
}
#define init_10_4_15_13
void FASTCALL call_10_4_15_13(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(15<<4)+13);
}
#define init_10_4_15_14
void FASTCALL call_10_4_15_14(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(15<<4)+14);
}
#define init_10_4_15_15
void FASTCALL call_10_4_15_15(SH2_struct* sh)
{
  SH2bra(sh, (4<<8)+(15<<4)+15);
}
#define init_10_5_0_0
void FASTCALL call_10_5_0_0(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(0<<4)+0);
}
#define init_10_5_0_1
void FASTCALL call_10_5_0_1(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(0<<4)+1);
}
#define init_10_5_0_2
void FASTCALL call_10_5_0_2(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(0<<4)+2);
}
#define init_10_5_0_3
void FASTCALL call_10_5_0_3(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(0<<4)+3);
}
#define init_10_5_0_4
void FASTCALL call_10_5_0_4(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(0<<4)+4);
}
#define init_10_5_0_5
void FASTCALL call_10_5_0_5(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(0<<4)+5);
}
#define init_10_5_0_6
void FASTCALL call_10_5_0_6(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(0<<4)+6);
}
#define init_10_5_0_7
void FASTCALL call_10_5_0_7(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(0<<4)+7);
}
#define init_10_5_0_8
void FASTCALL call_10_5_0_8(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(0<<4)+8);
}
#define init_10_5_0_9
void FASTCALL call_10_5_0_9(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(0<<4)+9);
}
#define init_10_5_0_10
void FASTCALL call_10_5_0_10(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(0<<4)+10);
}
#define init_10_5_0_11
void FASTCALL call_10_5_0_11(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(0<<4)+11);
}
#define init_10_5_0_12
void FASTCALL call_10_5_0_12(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(0<<4)+12);
}
#define init_10_5_0_13
void FASTCALL call_10_5_0_13(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(0<<4)+13);
}
#define init_10_5_0_14
void FASTCALL call_10_5_0_14(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(0<<4)+14);
}
#define init_10_5_0_15
void FASTCALL call_10_5_0_15(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(0<<4)+15);
}
#define init_10_5_1_0
void FASTCALL call_10_5_1_0(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(1<<4)+0);
}
#define init_10_5_1_1
void FASTCALL call_10_5_1_1(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(1<<4)+1);
}
#define init_10_5_1_2
void FASTCALL call_10_5_1_2(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(1<<4)+2);
}
#define init_10_5_1_3
void FASTCALL call_10_5_1_3(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(1<<4)+3);
}
#define init_10_5_1_4
void FASTCALL call_10_5_1_4(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(1<<4)+4);
}
#define init_10_5_1_5
void FASTCALL call_10_5_1_5(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(1<<4)+5);
}
#define init_10_5_1_6
void FASTCALL call_10_5_1_6(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(1<<4)+6);
}
#define init_10_5_1_7
void FASTCALL call_10_5_1_7(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(1<<4)+7);
}
#define init_10_5_1_8
void FASTCALL call_10_5_1_8(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(1<<4)+8);
}
#define init_10_5_1_9
void FASTCALL call_10_5_1_9(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(1<<4)+9);
}
#define init_10_5_1_10
void FASTCALL call_10_5_1_10(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(1<<4)+10);
}
#define init_10_5_1_11
void FASTCALL call_10_5_1_11(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(1<<4)+11);
}
#define init_10_5_1_12
void FASTCALL call_10_5_1_12(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(1<<4)+12);
}
#define init_10_5_1_13
void FASTCALL call_10_5_1_13(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(1<<4)+13);
}
#define init_10_5_1_14
void FASTCALL call_10_5_1_14(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(1<<4)+14);
}
#define init_10_5_1_15
void FASTCALL call_10_5_1_15(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(1<<4)+15);
}
#define init_10_5_2_0
void FASTCALL call_10_5_2_0(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(2<<4)+0);
}
#define init_10_5_2_1
void FASTCALL call_10_5_2_1(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(2<<4)+1);
}
#define init_10_5_2_2
void FASTCALL call_10_5_2_2(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(2<<4)+2);
}
#define init_10_5_2_3
void FASTCALL call_10_5_2_3(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(2<<4)+3);
}
#define init_10_5_2_4
void FASTCALL call_10_5_2_4(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(2<<4)+4);
}
#define init_10_5_2_5
void FASTCALL call_10_5_2_5(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(2<<4)+5);
}
#define init_10_5_2_6
void FASTCALL call_10_5_2_6(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(2<<4)+6);
}
#define init_10_5_2_7
void FASTCALL call_10_5_2_7(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(2<<4)+7);
}
#define init_10_5_2_8
void FASTCALL call_10_5_2_8(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(2<<4)+8);
}
#define init_10_5_2_9
void FASTCALL call_10_5_2_9(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(2<<4)+9);
}
#define init_10_5_2_10
void FASTCALL call_10_5_2_10(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(2<<4)+10);
}
#define init_10_5_2_11
void FASTCALL call_10_5_2_11(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(2<<4)+11);
}
#define init_10_5_2_12
void FASTCALL call_10_5_2_12(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(2<<4)+12);
}
#define init_10_5_2_13
void FASTCALL call_10_5_2_13(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(2<<4)+13);
}
#define init_10_5_2_14
void FASTCALL call_10_5_2_14(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(2<<4)+14);
}
#define init_10_5_2_15
void FASTCALL call_10_5_2_15(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(2<<4)+15);
}
#define init_10_5_3_0
void FASTCALL call_10_5_3_0(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(3<<4)+0);
}
#define init_10_5_3_1
void FASTCALL call_10_5_3_1(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(3<<4)+1);
}
#define init_10_5_3_2
void FASTCALL call_10_5_3_2(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(3<<4)+2);
}
#define init_10_5_3_3
void FASTCALL call_10_5_3_3(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(3<<4)+3);
}
#define init_10_5_3_4
void FASTCALL call_10_5_3_4(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(3<<4)+4);
}
#define init_10_5_3_5
void FASTCALL call_10_5_3_5(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(3<<4)+5);
}
#define init_10_5_3_6
void FASTCALL call_10_5_3_6(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(3<<4)+6);
}
#define init_10_5_3_7
void FASTCALL call_10_5_3_7(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(3<<4)+7);
}
#define init_10_5_3_8
void FASTCALL call_10_5_3_8(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(3<<4)+8);
}
#define init_10_5_3_9
void FASTCALL call_10_5_3_9(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(3<<4)+9);
}
#define init_10_5_3_10
void FASTCALL call_10_5_3_10(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(3<<4)+10);
}
#define init_10_5_3_11
void FASTCALL call_10_5_3_11(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(3<<4)+11);
}
#define init_10_5_3_12
void FASTCALL call_10_5_3_12(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(3<<4)+12);
}
#define init_10_5_3_13
void FASTCALL call_10_5_3_13(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(3<<4)+13);
}
#define init_10_5_3_14
void FASTCALL call_10_5_3_14(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(3<<4)+14);
}
#define init_10_5_3_15
void FASTCALL call_10_5_3_15(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(3<<4)+15);
}
#define init_10_5_4_0
void FASTCALL call_10_5_4_0(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(4<<4)+0);
}
#define init_10_5_4_1
void FASTCALL call_10_5_4_1(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(4<<4)+1);
}
#define init_10_5_4_2
void FASTCALL call_10_5_4_2(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(4<<4)+2);
}
#define init_10_5_4_3
void FASTCALL call_10_5_4_3(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(4<<4)+3);
}
#define init_10_5_4_4
void FASTCALL call_10_5_4_4(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(4<<4)+4);
}
#define init_10_5_4_5
void FASTCALL call_10_5_4_5(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(4<<4)+5);
}
#define init_10_5_4_6
void FASTCALL call_10_5_4_6(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(4<<4)+6);
}
#define init_10_5_4_7
void FASTCALL call_10_5_4_7(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(4<<4)+7);
}
#define init_10_5_4_8
void FASTCALL call_10_5_4_8(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(4<<4)+8);
}
#define init_10_5_4_9
void FASTCALL call_10_5_4_9(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(4<<4)+9);
}
#define init_10_5_4_10
void FASTCALL call_10_5_4_10(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(4<<4)+10);
}
#define init_10_5_4_11
void FASTCALL call_10_5_4_11(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(4<<4)+11);
}
#define init_10_5_4_12
void FASTCALL call_10_5_4_12(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(4<<4)+12);
}
#define init_10_5_4_13
void FASTCALL call_10_5_4_13(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(4<<4)+13);
}
#define init_10_5_4_14
void FASTCALL call_10_5_4_14(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(4<<4)+14);
}
#define init_10_5_4_15
void FASTCALL call_10_5_4_15(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(4<<4)+15);
}
#define init_10_5_5_0
void FASTCALL call_10_5_5_0(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(5<<4)+0);
}
#define init_10_5_5_1
void FASTCALL call_10_5_5_1(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(5<<4)+1);
}
#define init_10_5_5_2
void FASTCALL call_10_5_5_2(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(5<<4)+2);
}
#define init_10_5_5_3
void FASTCALL call_10_5_5_3(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(5<<4)+3);
}
#define init_10_5_5_4
void FASTCALL call_10_5_5_4(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(5<<4)+4);
}
#define init_10_5_5_5
void FASTCALL call_10_5_5_5(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(5<<4)+5);
}
#define init_10_5_5_6
void FASTCALL call_10_5_5_6(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(5<<4)+6);
}
#define init_10_5_5_7
void FASTCALL call_10_5_5_7(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(5<<4)+7);
}
#define init_10_5_5_8
void FASTCALL call_10_5_5_8(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(5<<4)+8);
}
#define init_10_5_5_9
void FASTCALL call_10_5_5_9(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(5<<4)+9);
}
#define init_10_5_5_10
void FASTCALL call_10_5_5_10(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(5<<4)+10);
}
#define init_10_5_5_11
void FASTCALL call_10_5_5_11(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(5<<4)+11);
}
#define init_10_5_5_12
void FASTCALL call_10_5_5_12(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(5<<4)+12);
}
#define init_10_5_5_13
void FASTCALL call_10_5_5_13(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(5<<4)+13);
}
#define init_10_5_5_14
void FASTCALL call_10_5_5_14(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(5<<4)+14);
}
#define init_10_5_5_15
void FASTCALL call_10_5_5_15(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(5<<4)+15);
}
#define init_10_5_6_0
void FASTCALL call_10_5_6_0(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(6<<4)+0);
}
#define init_10_5_6_1
void FASTCALL call_10_5_6_1(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(6<<4)+1);
}
#define init_10_5_6_2
void FASTCALL call_10_5_6_2(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(6<<4)+2);
}
#define init_10_5_6_3
void FASTCALL call_10_5_6_3(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(6<<4)+3);
}
#define init_10_5_6_4
void FASTCALL call_10_5_6_4(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(6<<4)+4);
}
#define init_10_5_6_5
void FASTCALL call_10_5_6_5(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(6<<4)+5);
}
#define init_10_5_6_6
void FASTCALL call_10_5_6_6(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(6<<4)+6);
}
#define init_10_5_6_7
void FASTCALL call_10_5_6_7(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(6<<4)+7);
}
#define init_10_5_6_8
void FASTCALL call_10_5_6_8(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(6<<4)+8);
}
#define init_10_5_6_9
void FASTCALL call_10_5_6_9(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(6<<4)+9);
}
#define init_10_5_6_10
void FASTCALL call_10_5_6_10(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(6<<4)+10);
}
#define init_10_5_6_11
void FASTCALL call_10_5_6_11(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(6<<4)+11);
}
#define init_10_5_6_12
void FASTCALL call_10_5_6_12(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(6<<4)+12);
}
#define init_10_5_6_13
void FASTCALL call_10_5_6_13(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(6<<4)+13);
}
#define init_10_5_6_14
void FASTCALL call_10_5_6_14(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(6<<4)+14);
}
#define init_10_5_6_15
void FASTCALL call_10_5_6_15(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(6<<4)+15);
}
#define init_10_5_7_0
void FASTCALL call_10_5_7_0(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(7<<4)+0);
}
#define init_10_5_7_1
void FASTCALL call_10_5_7_1(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(7<<4)+1);
}
#define init_10_5_7_2
void FASTCALL call_10_5_7_2(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(7<<4)+2);
}
#define init_10_5_7_3
void FASTCALL call_10_5_7_3(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(7<<4)+3);
}
#define init_10_5_7_4
void FASTCALL call_10_5_7_4(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(7<<4)+4);
}
#define init_10_5_7_5
void FASTCALL call_10_5_7_5(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(7<<4)+5);
}
#define init_10_5_7_6
void FASTCALL call_10_5_7_6(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(7<<4)+6);
}
#define init_10_5_7_7
void FASTCALL call_10_5_7_7(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(7<<4)+7);
}
#define init_10_5_7_8
void FASTCALL call_10_5_7_8(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(7<<4)+8);
}
#define init_10_5_7_9
void FASTCALL call_10_5_7_9(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(7<<4)+9);
}
#define init_10_5_7_10
void FASTCALL call_10_5_7_10(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(7<<4)+10);
}
#define init_10_5_7_11
void FASTCALL call_10_5_7_11(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(7<<4)+11);
}
#define init_10_5_7_12
void FASTCALL call_10_5_7_12(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(7<<4)+12);
}
#define init_10_5_7_13
void FASTCALL call_10_5_7_13(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(7<<4)+13);
}
#define init_10_5_7_14
void FASTCALL call_10_5_7_14(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(7<<4)+14);
}
#define init_10_5_7_15
void FASTCALL call_10_5_7_15(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(7<<4)+15);
}
#define init_10_5_8_0
void FASTCALL call_10_5_8_0(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(8<<4)+0);
}
#define init_10_5_8_1
void FASTCALL call_10_5_8_1(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(8<<4)+1);
}
#define init_10_5_8_2
void FASTCALL call_10_5_8_2(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(8<<4)+2);
}
#define init_10_5_8_3
void FASTCALL call_10_5_8_3(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(8<<4)+3);
}
#define init_10_5_8_4
void FASTCALL call_10_5_8_4(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(8<<4)+4);
}
#define init_10_5_8_5
void FASTCALL call_10_5_8_5(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(8<<4)+5);
}
#define init_10_5_8_6
void FASTCALL call_10_5_8_6(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(8<<4)+6);
}
#define init_10_5_8_7
void FASTCALL call_10_5_8_7(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(8<<4)+7);
}
#define init_10_5_8_8
void FASTCALL call_10_5_8_8(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(8<<4)+8);
}
#define init_10_5_8_9
void FASTCALL call_10_5_8_9(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(8<<4)+9);
}
#define init_10_5_8_10
void FASTCALL call_10_5_8_10(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(8<<4)+10);
}
#define init_10_5_8_11
void FASTCALL call_10_5_8_11(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(8<<4)+11);
}
#define init_10_5_8_12
void FASTCALL call_10_5_8_12(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(8<<4)+12);
}
#define init_10_5_8_13
void FASTCALL call_10_5_8_13(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(8<<4)+13);
}
#define init_10_5_8_14
void FASTCALL call_10_5_8_14(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(8<<4)+14);
}
#define init_10_5_8_15
void FASTCALL call_10_5_8_15(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(8<<4)+15);
}
#define init_10_5_9_0
void FASTCALL call_10_5_9_0(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(9<<4)+0);
}
#define init_10_5_9_1
void FASTCALL call_10_5_9_1(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(9<<4)+1);
}
#define init_10_5_9_2
void FASTCALL call_10_5_9_2(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(9<<4)+2);
}
#define init_10_5_9_3
void FASTCALL call_10_5_9_3(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(9<<4)+3);
}
#define init_10_5_9_4
void FASTCALL call_10_5_9_4(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(9<<4)+4);
}
#define init_10_5_9_5
void FASTCALL call_10_5_9_5(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(9<<4)+5);
}
#define init_10_5_9_6
void FASTCALL call_10_5_9_6(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(9<<4)+6);
}
#define init_10_5_9_7
void FASTCALL call_10_5_9_7(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(9<<4)+7);
}
#define init_10_5_9_8
void FASTCALL call_10_5_9_8(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(9<<4)+8);
}
#define init_10_5_9_9
void FASTCALL call_10_5_9_9(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(9<<4)+9);
}
#define init_10_5_9_10
void FASTCALL call_10_5_9_10(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(9<<4)+10);
}
#define init_10_5_9_11
void FASTCALL call_10_5_9_11(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(9<<4)+11);
}
#define init_10_5_9_12
void FASTCALL call_10_5_9_12(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(9<<4)+12);
}
#define init_10_5_9_13
void FASTCALL call_10_5_9_13(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(9<<4)+13);
}
#define init_10_5_9_14
void FASTCALL call_10_5_9_14(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(9<<4)+14);
}
#define init_10_5_9_15
void FASTCALL call_10_5_9_15(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(9<<4)+15);
}
#define init_10_5_10_0
void FASTCALL call_10_5_10_0(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(10<<4)+0);
}
#define init_10_5_10_1
void FASTCALL call_10_5_10_1(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(10<<4)+1);
}
#define init_10_5_10_2
void FASTCALL call_10_5_10_2(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(10<<4)+2);
}
#define init_10_5_10_3
void FASTCALL call_10_5_10_3(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(10<<4)+3);
}
#define init_10_5_10_4
void FASTCALL call_10_5_10_4(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(10<<4)+4);
}
#define init_10_5_10_5
void FASTCALL call_10_5_10_5(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(10<<4)+5);
}
#define init_10_5_10_6
void FASTCALL call_10_5_10_6(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(10<<4)+6);
}
#define init_10_5_10_7
void FASTCALL call_10_5_10_7(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(10<<4)+7);
}
#define init_10_5_10_8
void FASTCALL call_10_5_10_8(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(10<<4)+8);
}
#define init_10_5_10_9
void FASTCALL call_10_5_10_9(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(10<<4)+9);
}
#define init_10_5_10_10
void FASTCALL call_10_5_10_10(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(10<<4)+10);
}
#define init_10_5_10_11
void FASTCALL call_10_5_10_11(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(10<<4)+11);
}
#define init_10_5_10_12
void FASTCALL call_10_5_10_12(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(10<<4)+12);
}
#define init_10_5_10_13
void FASTCALL call_10_5_10_13(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(10<<4)+13);
}
#define init_10_5_10_14
void FASTCALL call_10_5_10_14(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(10<<4)+14);
}
#define init_10_5_10_15
void FASTCALL call_10_5_10_15(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(10<<4)+15);
}
#define init_10_5_11_0
void FASTCALL call_10_5_11_0(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(11<<4)+0);
}
#define init_10_5_11_1
void FASTCALL call_10_5_11_1(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(11<<4)+1);
}
#define init_10_5_11_2
void FASTCALL call_10_5_11_2(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(11<<4)+2);
}
#define init_10_5_11_3
void FASTCALL call_10_5_11_3(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(11<<4)+3);
}
#define init_10_5_11_4
void FASTCALL call_10_5_11_4(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(11<<4)+4);
}
#define init_10_5_11_5
void FASTCALL call_10_5_11_5(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(11<<4)+5);
}
#define init_10_5_11_6
void FASTCALL call_10_5_11_6(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(11<<4)+6);
}
#define init_10_5_11_7
void FASTCALL call_10_5_11_7(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(11<<4)+7);
}
#define init_10_5_11_8
void FASTCALL call_10_5_11_8(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(11<<4)+8);
}
#define init_10_5_11_9
void FASTCALL call_10_5_11_9(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(11<<4)+9);
}
#define init_10_5_11_10
void FASTCALL call_10_5_11_10(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(11<<4)+10);
}
#define init_10_5_11_11
void FASTCALL call_10_5_11_11(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(11<<4)+11);
}
#define init_10_5_11_12
void FASTCALL call_10_5_11_12(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(11<<4)+12);
}
#define init_10_5_11_13
void FASTCALL call_10_5_11_13(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(11<<4)+13);
}
#define init_10_5_11_14
void FASTCALL call_10_5_11_14(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(11<<4)+14);
}
#define init_10_5_11_15
void FASTCALL call_10_5_11_15(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(11<<4)+15);
}
#define init_10_5_12_0
void FASTCALL call_10_5_12_0(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(12<<4)+0);
}
#define init_10_5_12_1
void FASTCALL call_10_5_12_1(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(12<<4)+1);
}
#define init_10_5_12_2
void FASTCALL call_10_5_12_2(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(12<<4)+2);
}
#define init_10_5_12_3
void FASTCALL call_10_5_12_3(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(12<<4)+3);
}
#define init_10_5_12_4
void FASTCALL call_10_5_12_4(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(12<<4)+4);
}
#define init_10_5_12_5
void FASTCALL call_10_5_12_5(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(12<<4)+5);
}
#define init_10_5_12_6
void FASTCALL call_10_5_12_6(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(12<<4)+6);
}
#define init_10_5_12_7
void FASTCALL call_10_5_12_7(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(12<<4)+7);
}
#define init_10_5_12_8
void FASTCALL call_10_5_12_8(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(12<<4)+8);
}
#define init_10_5_12_9
void FASTCALL call_10_5_12_9(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(12<<4)+9);
}
#define init_10_5_12_10
void FASTCALL call_10_5_12_10(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(12<<4)+10);
}
#define init_10_5_12_11
void FASTCALL call_10_5_12_11(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(12<<4)+11);
}
#define init_10_5_12_12
void FASTCALL call_10_5_12_12(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(12<<4)+12);
}
#define init_10_5_12_13
void FASTCALL call_10_5_12_13(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(12<<4)+13);
}
#define init_10_5_12_14
void FASTCALL call_10_5_12_14(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(12<<4)+14);
}
#define init_10_5_12_15
void FASTCALL call_10_5_12_15(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(12<<4)+15);
}
#define init_10_5_13_0
void FASTCALL call_10_5_13_0(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(13<<4)+0);
}
#define init_10_5_13_1
void FASTCALL call_10_5_13_1(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(13<<4)+1);
}
#define init_10_5_13_2
void FASTCALL call_10_5_13_2(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(13<<4)+2);
}
#define init_10_5_13_3
void FASTCALL call_10_5_13_3(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(13<<4)+3);
}
#define init_10_5_13_4
void FASTCALL call_10_5_13_4(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(13<<4)+4);
}
#define init_10_5_13_5
void FASTCALL call_10_5_13_5(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(13<<4)+5);
}
#define init_10_5_13_6
void FASTCALL call_10_5_13_6(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(13<<4)+6);
}
#define init_10_5_13_7
void FASTCALL call_10_5_13_7(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(13<<4)+7);
}
#define init_10_5_13_8
void FASTCALL call_10_5_13_8(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(13<<4)+8);
}
#define init_10_5_13_9
void FASTCALL call_10_5_13_9(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(13<<4)+9);
}
#define init_10_5_13_10
void FASTCALL call_10_5_13_10(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(13<<4)+10);
}
#define init_10_5_13_11
void FASTCALL call_10_5_13_11(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(13<<4)+11);
}
#define init_10_5_13_12
void FASTCALL call_10_5_13_12(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(13<<4)+12);
}
#define init_10_5_13_13
void FASTCALL call_10_5_13_13(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(13<<4)+13);
}
#define init_10_5_13_14
void FASTCALL call_10_5_13_14(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(13<<4)+14);
}
#define init_10_5_13_15
void FASTCALL call_10_5_13_15(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(13<<4)+15);
}
#define init_10_5_14_0
void FASTCALL call_10_5_14_0(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(14<<4)+0);
}
#define init_10_5_14_1
void FASTCALL call_10_5_14_1(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(14<<4)+1);
}
#define init_10_5_14_2
void FASTCALL call_10_5_14_2(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(14<<4)+2);
}
#define init_10_5_14_3
void FASTCALL call_10_5_14_3(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(14<<4)+3);
}
#define init_10_5_14_4
void FASTCALL call_10_5_14_4(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(14<<4)+4);
}
#define init_10_5_14_5
void FASTCALL call_10_5_14_5(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(14<<4)+5);
}
#define init_10_5_14_6
void FASTCALL call_10_5_14_6(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(14<<4)+6);
}
#define init_10_5_14_7
void FASTCALL call_10_5_14_7(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(14<<4)+7);
}
#define init_10_5_14_8
void FASTCALL call_10_5_14_8(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(14<<4)+8);
}
#define init_10_5_14_9
void FASTCALL call_10_5_14_9(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(14<<4)+9);
}
#define init_10_5_14_10
void FASTCALL call_10_5_14_10(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(14<<4)+10);
}
#define init_10_5_14_11
void FASTCALL call_10_5_14_11(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(14<<4)+11);
}
#define init_10_5_14_12
void FASTCALL call_10_5_14_12(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(14<<4)+12);
}
#define init_10_5_14_13
void FASTCALL call_10_5_14_13(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(14<<4)+13);
}
#define init_10_5_14_14
void FASTCALL call_10_5_14_14(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(14<<4)+14);
}
#define init_10_5_14_15
void FASTCALL call_10_5_14_15(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(14<<4)+15);
}
#define init_10_5_15_0
void FASTCALL call_10_5_15_0(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(15<<4)+0);
}
#define init_10_5_15_1
void FASTCALL call_10_5_15_1(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(15<<4)+1);
}
#define init_10_5_15_2
void FASTCALL call_10_5_15_2(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(15<<4)+2);
}
#define init_10_5_15_3
void FASTCALL call_10_5_15_3(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(15<<4)+3);
}
#define init_10_5_15_4
void FASTCALL call_10_5_15_4(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(15<<4)+4);
}
#define init_10_5_15_5
void FASTCALL call_10_5_15_5(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(15<<4)+5);
}
#define init_10_5_15_6
void FASTCALL call_10_5_15_6(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(15<<4)+6);
}
#define init_10_5_15_7
void FASTCALL call_10_5_15_7(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(15<<4)+7);
}
#define init_10_5_15_8
void FASTCALL call_10_5_15_8(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(15<<4)+8);
}
#define init_10_5_15_9
void FASTCALL call_10_5_15_9(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(15<<4)+9);
}
#define init_10_5_15_10
void FASTCALL call_10_5_15_10(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(15<<4)+10);
}
#define init_10_5_15_11
void FASTCALL call_10_5_15_11(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(15<<4)+11);
}
#define init_10_5_15_12
void FASTCALL call_10_5_15_12(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(15<<4)+12);
}
#define init_10_5_15_13
void FASTCALL call_10_5_15_13(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(15<<4)+13);
}
#define init_10_5_15_14
void FASTCALL call_10_5_15_14(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(15<<4)+14);
}
#define init_10_5_15_15
void FASTCALL call_10_5_15_15(SH2_struct* sh)
{
  SH2bra(sh, (5<<8)+(15<<4)+15);
}
#define init_10_6_0_0
void FASTCALL call_10_6_0_0(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(0<<4)+0);
}
#define init_10_6_0_1
void FASTCALL call_10_6_0_1(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(0<<4)+1);
}
#define init_10_6_0_2
void FASTCALL call_10_6_0_2(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(0<<4)+2);
}
#define init_10_6_0_3
void FASTCALL call_10_6_0_3(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(0<<4)+3);
}
#define init_10_6_0_4
void FASTCALL call_10_6_0_4(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(0<<4)+4);
}
#define init_10_6_0_5
void FASTCALL call_10_6_0_5(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(0<<4)+5);
}
#define init_10_6_0_6
void FASTCALL call_10_6_0_6(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(0<<4)+6);
}
#define init_10_6_0_7
void FASTCALL call_10_6_0_7(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(0<<4)+7);
}
#define init_10_6_0_8
void FASTCALL call_10_6_0_8(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(0<<4)+8);
}
#define init_10_6_0_9
void FASTCALL call_10_6_0_9(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(0<<4)+9);
}
#define init_10_6_0_10
void FASTCALL call_10_6_0_10(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(0<<4)+10);
}
#define init_10_6_0_11
void FASTCALL call_10_6_0_11(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(0<<4)+11);
}
#define init_10_6_0_12
void FASTCALL call_10_6_0_12(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(0<<4)+12);
}
#define init_10_6_0_13
void FASTCALL call_10_6_0_13(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(0<<4)+13);
}
#define init_10_6_0_14
void FASTCALL call_10_6_0_14(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(0<<4)+14);
}
#define init_10_6_0_15
void FASTCALL call_10_6_0_15(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(0<<4)+15);
}
#define init_10_6_1_0
void FASTCALL call_10_6_1_0(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(1<<4)+0);
}
#define init_10_6_1_1
void FASTCALL call_10_6_1_1(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(1<<4)+1);
}
#define init_10_6_1_2
void FASTCALL call_10_6_1_2(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(1<<4)+2);
}
#define init_10_6_1_3
void FASTCALL call_10_6_1_3(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(1<<4)+3);
}
#define init_10_6_1_4
void FASTCALL call_10_6_1_4(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(1<<4)+4);
}
#define init_10_6_1_5
void FASTCALL call_10_6_1_5(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(1<<4)+5);
}
#define init_10_6_1_6
void FASTCALL call_10_6_1_6(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(1<<4)+6);
}
#define init_10_6_1_7
void FASTCALL call_10_6_1_7(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(1<<4)+7);
}
#define init_10_6_1_8
void FASTCALL call_10_6_1_8(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(1<<4)+8);
}
#define init_10_6_1_9
void FASTCALL call_10_6_1_9(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(1<<4)+9);
}
#define init_10_6_1_10
void FASTCALL call_10_6_1_10(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(1<<4)+10);
}
#define init_10_6_1_11
void FASTCALL call_10_6_1_11(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(1<<4)+11);
}
#define init_10_6_1_12
void FASTCALL call_10_6_1_12(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(1<<4)+12);
}
#define init_10_6_1_13
void FASTCALL call_10_6_1_13(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(1<<4)+13);
}
#define init_10_6_1_14
void FASTCALL call_10_6_1_14(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(1<<4)+14);
}
#define init_10_6_1_15
void FASTCALL call_10_6_1_15(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(1<<4)+15);
}
#define init_10_6_2_0
void FASTCALL call_10_6_2_0(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(2<<4)+0);
}
#define init_10_6_2_1
void FASTCALL call_10_6_2_1(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(2<<4)+1);
}
#define init_10_6_2_2
void FASTCALL call_10_6_2_2(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(2<<4)+2);
}
#define init_10_6_2_3
void FASTCALL call_10_6_2_3(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(2<<4)+3);
}
#define init_10_6_2_4
void FASTCALL call_10_6_2_4(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(2<<4)+4);
}
#define init_10_6_2_5
void FASTCALL call_10_6_2_5(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(2<<4)+5);
}
#define init_10_6_2_6
void FASTCALL call_10_6_2_6(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(2<<4)+6);
}
#define init_10_6_2_7
void FASTCALL call_10_6_2_7(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(2<<4)+7);
}
#define init_10_6_2_8
void FASTCALL call_10_6_2_8(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(2<<4)+8);
}
#define init_10_6_2_9
void FASTCALL call_10_6_2_9(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(2<<4)+9);
}
#define init_10_6_2_10
void FASTCALL call_10_6_2_10(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(2<<4)+10);
}
#define init_10_6_2_11
void FASTCALL call_10_6_2_11(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(2<<4)+11);
}
#define init_10_6_2_12
void FASTCALL call_10_6_2_12(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(2<<4)+12);
}
#define init_10_6_2_13
void FASTCALL call_10_6_2_13(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(2<<4)+13);
}
#define init_10_6_2_14
void FASTCALL call_10_6_2_14(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(2<<4)+14);
}
#define init_10_6_2_15
void FASTCALL call_10_6_2_15(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(2<<4)+15);
}
#define init_10_6_3_0
void FASTCALL call_10_6_3_0(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(3<<4)+0);
}
#define init_10_6_3_1
void FASTCALL call_10_6_3_1(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(3<<4)+1);
}
#define init_10_6_3_2
void FASTCALL call_10_6_3_2(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(3<<4)+2);
}
#define init_10_6_3_3
void FASTCALL call_10_6_3_3(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(3<<4)+3);
}
#define init_10_6_3_4
void FASTCALL call_10_6_3_4(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(3<<4)+4);
}
#define init_10_6_3_5
void FASTCALL call_10_6_3_5(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(3<<4)+5);
}
#define init_10_6_3_6
void FASTCALL call_10_6_3_6(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(3<<4)+6);
}
#define init_10_6_3_7
void FASTCALL call_10_6_3_7(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(3<<4)+7);
}
#define init_10_6_3_8
void FASTCALL call_10_6_3_8(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(3<<4)+8);
}
#define init_10_6_3_9
void FASTCALL call_10_6_3_9(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(3<<4)+9);
}
#define init_10_6_3_10
void FASTCALL call_10_6_3_10(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(3<<4)+10);
}
#define init_10_6_3_11
void FASTCALL call_10_6_3_11(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(3<<4)+11);
}
#define init_10_6_3_12
void FASTCALL call_10_6_3_12(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(3<<4)+12);
}
#define init_10_6_3_13
void FASTCALL call_10_6_3_13(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(3<<4)+13);
}
#define init_10_6_3_14
void FASTCALL call_10_6_3_14(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(3<<4)+14);
}
#define init_10_6_3_15
void FASTCALL call_10_6_3_15(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(3<<4)+15);
}
#define init_10_6_4_0
void FASTCALL call_10_6_4_0(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(4<<4)+0);
}
#define init_10_6_4_1
void FASTCALL call_10_6_4_1(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(4<<4)+1);
}
#define init_10_6_4_2
void FASTCALL call_10_6_4_2(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(4<<4)+2);
}
#define init_10_6_4_3
void FASTCALL call_10_6_4_3(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(4<<4)+3);
}
#define init_10_6_4_4
void FASTCALL call_10_6_4_4(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(4<<4)+4);
}
#define init_10_6_4_5
void FASTCALL call_10_6_4_5(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(4<<4)+5);
}
#define init_10_6_4_6
void FASTCALL call_10_6_4_6(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(4<<4)+6);
}
#define init_10_6_4_7
void FASTCALL call_10_6_4_7(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(4<<4)+7);
}
#define init_10_6_4_8
void FASTCALL call_10_6_4_8(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(4<<4)+8);
}
#define init_10_6_4_9
void FASTCALL call_10_6_4_9(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(4<<4)+9);
}
#define init_10_6_4_10
void FASTCALL call_10_6_4_10(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(4<<4)+10);
}
#define init_10_6_4_11
void FASTCALL call_10_6_4_11(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(4<<4)+11);
}
#define init_10_6_4_12
void FASTCALL call_10_6_4_12(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(4<<4)+12);
}
#define init_10_6_4_13
void FASTCALL call_10_6_4_13(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(4<<4)+13);
}
#define init_10_6_4_14
void FASTCALL call_10_6_4_14(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(4<<4)+14);
}
#define init_10_6_4_15
void FASTCALL call_10_6_4_15(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(4<<4)+15);
}
#define init_10_6_5_0
void FASTCALL call_10_6_5_0(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(5<<4)+0);
}
#define init_10_6_5_1
void FASTCALL call_10_6_5_1(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(5<<4)+1);
}
#define init_10_6_5_2
void FASTCALL call_10_6_5_2(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(5<<4)+2);
}
#define init_10_6_5_3
void FASTCALL call_10_6_5_3(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(5<<4)+3);
}
#define init_10_6_5_4
void FASTCALL call_10_6_5_4(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(5<<4)+4);
}
#define init_10_6_5_5
void FASTCALL call_10_6_5_5(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(5<<4)+5);
}
#define init_10_6_5_6
void FASTCALL call_10_6_5_6(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(5<<4)+6);
}
#define init_10_6_5_7
void FASTCALL call_10_6_5_7(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(5<<4)+7);
}
#define init_10_6_5_8
void FASTCALL call_10_6_5_8(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(5<<4)+8);
}
#define init_10_6_5_9
void FASTCALL call_10_6_5_9(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(5<<4)+9);
}
#define init_10_6_5_10
void FASTCALL call_10_6_5_10(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(5<<4)+10);
}
#define init_10_6_5_11
void FASTCALL call_10_6_5_11(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(5<<4)+11);
}
#define init_10_6_5_12
void FASTCALL call_10_6_5_12(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(5<<4)+12);
}
#define init_10_6_5_13
void FASTCALL call_10_6_5_13(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(5<<4)+13);
}
#define init_10_6_5_14
void FASTCALL call_10_6_5_14(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(5<<4)+14);
}
#define init_10_6_5_15
void FASTCALL call_10_6_5_15(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(5<<4)+15);
}
#define init_10_6_6_0
void FASTCALL call_10_6_6_0(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(6<<4)+0);
}
#define init_10_6_6_1
void FASTCALL call_10_6_6_1(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(6<<4)+1);
}
#define init_10_6_6_2
void FASTCALL call_10_6_6_2(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(6<<4)+2);
}
#define init_10_6_6_3
void FASTCALL call_10_6_6_3(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(6<<4)+3);
}
#define init_10_6_6_4
void FASTCALL call_10_6_6_4(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(6<<4)+4);
}
#define init_10_6_6_5
void FASTCALL call_10_6_6_5(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(6<<4)+5);
}
#define init_10_6_6_6
void FASTCALL call_10_6_6_6(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(6<<4)+6);
}
#define init_10_6_6_7
void FASTCALL call_10_6_6_7(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(6<<4)+7);
}
#define init_10_6_6_8
void FASTCALL call_10_6_6_8(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(6<<4)+8);
}
#define init_10_6_6_9
void FASTCALL call_10_6_6_9(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(6<<4)+9);
}
#define init_10_6_6_10
void FASTCALL call_10_6_6_10(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(6<<4)+10);
}
#define init_10_6_6_11
void FASTCALL call_10_6_6_11(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(6<<4)+11);
}
#define init_10_6_6_12
void FASTCALL call_10_6_6_12(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(6<<4)+12);
}
#define init_10_6_6_13
void FASTCALL call_10_6_6_13(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(6<<4)+13);
}
#define init_10_6_6_14
void FASTCALL call_10_6_6_14(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(6<<4)+14);
}
#define init_10_6_6_15
void FASTCALL call_10_6_6_15(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(6<<4)+15);
}
#define init_10_6_7_0
void FASTCALL call_10_6_7_0(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(7<<4)+0);
}
#define init_10_6_7_1
void FASTCALL call_10_6_7_1(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(7<<4)+1);
}
#define init_10_6_7_2
void FASTCALL call_10_6_7_2(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(7<<4)+2);
}
#define init_10_6_7_3
void FASTCALL call_10_6_7_3(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(7<<4)+3);
}
#define init_10_6_7_4
void FASTCALL call_10_6_7_4(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(7<<4)+4);
}
#define init_10_6_7_5
void FASTCALL call_10_6_7_5(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(7<<4)+5);
}
#define init_10_6_7_6
void FASTCALL call_10_6_7_6(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(7<<4)+6);
}
#define init_10_6_7_7
void FASTCALL call_10_6_7_7(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(7<<4)+7);
}
#define init_10_6_7_8
void FASTCALL call_10_6_7_8(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(7<<4)+8);
}
#define init_10_6_7_9
void FASTCALL call_10_6_7_9(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(7<<4)+9);
}
#define init_10_6_7_10
void FASTCALL call_10_6_7_10(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(7<<4)+10);
}
#define init_10_6_7_11
void FASTCALL call_10_6_7_11(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(7<<4)+11);
}
#define init_10_6_7_12
void FASTCALL call_10_6_7_12(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(7<<4)+12);
}
#define init_10_6_7_13
void FASTCALL call_10_6_7_13(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(7<<4)+13);
}
#define init_10_6_7_14
void FASTCALL call_10_6_7_14(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(7<<4)+14);
}
#define init_10_6_7_15
void FASTCALL call_10_6_7_15(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(7<<4)+15);
}
#define init_10_6_8_0
void FASTCALL call_10_6_8_0(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(8<<4)+0);
}
#define init_10_6_8_1
void FASTCALL call_10_6_8_1(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(8<<4)+1);
}
#define init_10_6_8_2
void FASTCALL call_10_6_8_2(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(8<<4)+2);
}
#define init_10_6_8_3
void FASTCALL call_10_6_8_3(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(8<<4)+3);
}
#define init_10_6_8_4
void FASTCALL call_10_6_8_4(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(8<<4)+4);
}
#define init_10_6_8_5
void FASTCALL call_10_6_8_5(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(8<<4)+5);
}
#define init_10_6_8_6
void FASTCALL call_10_6_8_6(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(8<<4)+6);
}
#define init_10_6_8_7
void FASTCALL call_10_6_8_7(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(8<<4)+7);
}
#define init_10_6_8_8
void FASTCALL call_10_6_8_8(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(8<<4)+8);
}
#define init_10_6_8_9
void FASTCALL call_10_6_8_9(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(8<<4)+9);
}
#define init_10_6_8_10
void FASTCALL call_10_6_8_10(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(8<<4)+10);
}
#define init_10_6_8_11
void FASTCALL call_10_6_8_11(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(8<<4)+11);
}
#define init_10_6_8_12
void FASTCALL call_10_6_8_12(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(8<<4)+12);
}
#define init_10_6_8_13
void FASTCALL call_10_6_8_13(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(8<<4)+13);
}
#define init_10_6_8_14
void FASTCALL call_10_6_8_14(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(8<<4)+14);
}
#define init_10_6_8_15
void FASTCALL call_10_6_8_15(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(8<<4)+15);
}
#define init_10_6_9_0
void FASTCALL call_10_6_9_0(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(9<<4)+0);
}
#define init_10_6_9_1
void FASTCALL call_10_6_9_1(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(9<<4)+1);
}
#define init_10_6_9_2
void FASTCALL call_10_6_9_2(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(9<<4)+2);
}
#define init_10_6_9_3
void FASTCALL call_10_6_9_3(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(9<<4)+3);
}
#define init_10_6_9_4
void FASTCALL call_10_6_9_4(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(9<<4)+4);
}
#define init_10_6_9_5
void FASTCALL call_10_6_9_5(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(9<<4)+5);
}
#define init_10_6_9_6
void FASTCALL call_10_6_9_6(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(9<<4)+6);
}
#define init_10_6_9_7
void FASTCALL call_10_6_9_7(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(9<<4)+7);
}
#define init_10_6_9_8
void FASTCALL call_10_6_9_8(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(9<<4)+8);
}
#define init_10_6_9_9
void FASTCALL call_10_6_9_9(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(9<<4)+9);
}
#define init_10_6_9_10
void FASTCALL call_10_6_9_10(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(9<<4)+10);
}
#define init_10_6_9_11
void FASTCALL call_10_6_9_11(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(9<<4)+11);
}
#define init_10_6_9_12
void FASTCALL call_10_6_9_12(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(9<<4)+12);
}
#define init_10_6_9_13
void FASTCALL call_10_6_9_13(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(9<<4)+13);
}
#define init_10_6_9_14
void FASTCALL call_10_6_9_14(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(9<<4)+14);
}
#define init_10_6_9_15
void FASTCALL call_10_6_9_15(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(9<<4)+15);
}
#define init_10_6_10_0
void FASTCALL call_10_6_10_0(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(10<<4)+0);
}
#define init_10_6_10_1
void FASTCALL call_10_6_10_1(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(10<<4)+1);
}
#define init_10_6_10_2
void FASTCALL call_10_6_10_2(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(10<<4)+2);
}
#define init_10_6_10_3
void FASTCALL call_10_6_10_3(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(10<<4)+3);
}
#define init_10_6_10_4
void FASTCALL call_10_6_10_4(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(10<<4)+4);
}
#define init_10_6_10_5
void FASTCALL call_10_6_10_5(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(10<<4)+5);
}
#define init_10_6_10_6
void FASTCALL call_10_6_10_6(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(10<<4)+6);
}
#define init_10_6_10_7
void FASTCALL call_10_6_10_7(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(10<<4)+7);
}
#define init_10_6_10_8
void FASTCALL call_10_6_10_8(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(10<<4)+8);
}
#define init_10_6_10_9
void FASTCALL call_10_6_10_9(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(10<<4)+9);
}
#define init_10_6_10_10
void FASTCALL call_10_6_10_10(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(10<<4)+10);
}
#define init_10_6_10_11
void FASTCALL call_10_6_10_11(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(10<<4)+11);
}
#define init_10_6_10_12
void FASTCALL call_10_6_10_12(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(10<<4)+12);
}
#define init_10_6_10_13
void FASTCALL call_10_6_10_13(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(10<<4)+13);
}
#define init_10_6_10_14
void FASTCALL call_10_6_10_14(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(10<<4)+14);
}
#define init_10_6_10_15
void FASTCALL call_10_6_10_15(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(10<<4)+15);
}
#define init_10_6_11_0
void FASTCALL call_10_6_11_0(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(11<<4)+0);
}
#define init_10_6_11_1
void FASTCALL call_10_6_11_1(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(11<<4)+1);
}
#define init_10_6_11_2
void FASTCALL call_10_6_11_2(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(11<<4)+2);
}
#define init_10_6_11_3
void FASTCALL call_10_6_11_3(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(11<<4)+3);
}
#define init_10_6_11_4
void FASTCALL call_10_6_11_4(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(11<<4)+4);
}
#define init_10_6_11_5
void FASTCALL call_10_6_11_5(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(11<<4)+5);
}
#define init_10_6_11_6
void FASTCALL call_10_6_11_6(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(11<<4)+6);
}
#define init_10_6_11_7
void FASTCALL call_10_6_11_7(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(11<<4)+7);
}
#define init_10_6_11_8
void FASTCALL call_10_6_11_8(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(11<<4)+8);
}
#define init_10_6_11_9
void FASTCALL call_10_6_11_9(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(11<<4)+9);
}
#define init_10_6_11_10
void FASTCALL call_10_6_11_10(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(11<<4)+10);
}
#define init_10_6_11_11
void FASTCALL call_10_6_11_11(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(11<<4)+11);
}
#define init_10_6_11_12
void FASTCALL call_10_6_11_12(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(11<<4)+12);
}
#define init_10_6_11_13
void FASTCALL call_10_6_11_13(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(11<<4)+13);
}
#define init_10_6_11_14
void FASTCALL call_10_6_11_14(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(11<<4)+14);
}
#define init_10_6_11_15
void FASTCALL call_10_6_11_15(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(11<<4)+15);
}
#define init_10_6_12_0
void FASTCALL call_10_6_12_0(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(12<<4)+0);
}
#define init_10_6_12_1
void FASTCALL call_10_6_12_1(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(12<<4)+1);
}
#define init_10_6_12_2
void FASTCALL call_10_6_12_2(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(12<<4)+2);
}
#define init_10_6_12_3
void FASTCALL call_10_6_12_3(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(12<<4)+3);
}
#define init_10_6_12_4
void FASTCALL call_10_6_12_4(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(12<<4)+4);
}
#define init_10_6_12_5
void FASTCALL call_10_6_12_5(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(12<<4)+5);
}
#define init_10_6_12_6
void FASTCALL call_10_6_12_6(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(12<<4)+6);
}
#define init_10_6_12_7
void FASTCALL call_10_6_12_7(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(12<<4)+7);
}
#define init_10_6_12_8
void FASTCALL call_10_6_12_8(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(12<<4)+8);
}
#define init_10_6_12_9
void FASTCALL call_10_6_12_9(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(12<<4)+9);
}
#define init_10_6_12_10
void FASTCALL call_10_6_12_10(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(12<<4)+10);
}
#define init_10_6_12_11
void FASTCALL call_10_6_12_11(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(12<<4)+11);
}
#define init_10_6_12_12
void FASTCALL call_10_6_12_12(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(12<<4)+12);
}
#define init_10_6_12_13
void FASTCALL call_10_6_12_13(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(12<<4)+13);
}
#define init_10_6_12_14
void FASTCALL call_10_6_12_14(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(12<<4)+14);
}
#define init_10_6_12_15
void FASTCALL call_10_6_12_15(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(12<<4)+15);
}
#define init_10_6_13_0
void FASTCALL call_10_6_13_0(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(13<<4)+0);
}
#define init_10_6_13_1
void FASTCALL call_10_6_13_1(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(13<<4)+1);
}
#define init_10_6_13_2
void FASTCALL call_10_6_13_2(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(13<<4)+2);
}
#define init_10_6_13_3
void FASTCALL call_10_6_13_3(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(13<<4)+3);
}
#define init_10_6_13_4
void FASTCALL call_10_6_13_4(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(13<<4)+4);
}
#define init_10_6_13_5
void FASTCALL call_10_6_13_5(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(13<<4)+5);
}
#define init_10_6_13_6
void FASTCALL call_10_6_13_6(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(13<<4)+6);
}
#define init_10_6_13_7
void FASTCALL call_10_6_13_7(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(13<<4)+7);
}
#define init_10_6_13_8
void FASTCALL call_10_6_13_8(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(13<<4)+8);
}
#define init_10_6_13_9
void FASTCALL call_10_6_13_9(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(13<<4)+9);
}
#define init_10_6_13_10
void FASTCALL call_10_6_13_10(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(13<<4)+10);
}
#define init_10_6_13_11
void FASTCALL call_10_6_13_11(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(13<<4)+11);
}
#define init_10_6_13_12
void FASTCALL call_10_6_13_12(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(13<<4)+12);
}
#define init_10_6_13_13
void FASTCALL call_10_6_13_13(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(13<<4)+13);
}
#define init_10_6_13_14
void FASTCALL call_10_6_13_14(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(13<<4)+14);
}
#define init_10_6_13_15
void FASTCALL call_10_6_13_15(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(13<<4)+15);
}
#define init_10_6_14_0
void FASTCALL call_10_6_14_0(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(14<<4)+0);
}
#define init_10_6_14_1
void FASTCALL call_10_6_14_1(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(14<<4)+1);
}
#define init_10_6_14_2
void FASTCALL call_10_6_14_2(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(14<<4)+2);
}
#define init_10_6_14_3
void FASTCALL call_10_6_14_3(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(14<<4)+3);
}
#define init_10_6_14_4
void FASTCALL call_10_6_14_4(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(14<<4)+4);
}
#define init_10_6_14_5
void FASTCALL call_10_6_14_5(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(14<<4)+5);
}
#define init_10_6_14_6
void FASTCALL call_10_6_14_6(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(14<<4)+6);
}
#define init_10_6_14_7
void FASTCALL call_10_6_14_7(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(14<<4)+7);
}
#define init_10_6_14_8
void FASTCALL call_10_6_14_8(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(14<<4)+8);
}
#define init_10_6_14_9
void FASTCALL call_10_6_14_9(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(14<<4)+9);
}
#define init_10_6_14_10
void FASTCALL call_10_6_14_10(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(14<<4)+10);
}
#define init_10_6_14_11
void FASTCALL call_10_6_14_11(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(14<<4)+11);
}
#define init_10_6_14_12
void FASTCALL call_10_6_14_12(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(14<<4)+12);
}
#define init_10_6_14_13
void FASTCALL call_10_6_14_13(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(14<<4)+13);
}
#define init_10_6_14_14
void FASTCALL call_10_6_14_14(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(14<<4)+14);
}
#define init_10_6_14_15
void FASTCALL call_10_6_14_15(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(14<<4)+15);
}
#define init_10_6_15_0
void FASTCALL call_10_6_15_0(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(15<<4)+0);
}
#define init_10_6_15_1
void FASTCALL call_10_6_15_1(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(15<<4)+1);
}
#define init_10_6_15_2
void FASTCALL call_10_6_15_2(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(15<<4)+2);
}
#define init_10_6_15_3
void FASTCALL call_10_6_15_3(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(15<<4)+3);
}
#define init_10_6_15_4
void FASTCALL call_10_6_15_4(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(15<<4)+4);
}
#define init_10_6_15_5
void FASTCALL call_10_6_15_5(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(15<<4)+5);
}
#define init_10_6_15_6
void FASTCALL call_10_6_15_6(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(15<<4)+6);
}
#define init_10_6_15_7
void FASTCALL call_10_6_15_7(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(15<<4)+7);
}
#define init_10_6_15_8
void FASTCALL call_10_6_15_8(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(15<<4)+8);
}
#define init_10_6_15_9
void FASTCALL call_10_6_15_9(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(15<<4)+9);
}
#define init_10_6_15_10
void FASTCALL call_10_6_15_10(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(15<<4)+10);
}
#define init_10_6_15_11
void FASTCALL call_10_6_15_11(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(15<<4)+11);
}
#define init_10_6_15_12
void FASTCALL call_10_6_15_12(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(15<<4)+12);
}
#define init_10_6_15_13
void FASTCALL call_10_6_15_13(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(15<<4)+13);
}
#define init_10_6_15_14
void FASTCALL call_10_6_15_14(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(15<<4)+14);
}
#define init_10_6_15_15
void FASTCALL call_10_6_15_15(SH2_struct* sh)
{
  SH2bra(sh, (6<<8)+(15<<4)+15);
}
#define init_10_7_0_0
void FASTCALL call_10_7_0_0(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(0<<4)+0);
}
#define init_10_7_0_1
void FASTCALL call_10_7_0_1(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(0<<4)+1);
}
#define init_10_7_0_2
void FASTCALL call_10_7_0_2(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(0<<4)+2);
}
#define init_10_7_0_3
void FASTCALL call_10_7_0_3(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(0<<4)+3);
}
#define init_10_7_0_4
void FASTCALL call_10_7_0_4(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(0<<4)+4);
}
#define init_10_7_0_5
void FASTCALL call_10_7_0_5(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(0<<4)+5);
}
#define init_10_7_0_6
void FASTCALL call_10_7_0_6(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(0<<4)+6);
}
#define init_10_7_0_7
void FASTCALL call_10_7_0_7(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(0<<4)+7);
}
#define init_10_7_0_8
void FASTCALL call_10_7_0_8(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(0<<4)+8);
}
#define init_10_7_0_9
void FASTCALL call_10_7_0_9(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(0<<4)+9);
}
#define init_10_7_0_10
void FASTCALL call_10_7_0_10(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(0<<4)+10);
}
#define init_10_7_0_11
void FASTCALL call_10_7_0_11(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(0<<4)+11);
}
#define init_10_7_0_12
void FASTCALL call_10_7_0_12(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(0<<4)+12);
}
#define init_10_7_0_13
void FASTCALL call_10_7_0_13(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(0<<4)+13);
}
#define init_10_7_0_14
void FASTCALL call_10_7_0_14(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(0<<4)+14);
}
#define init_10_7_0_15
void FASTCALL call_10_7_0_15(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(0<<4)+15);
}
#define init_10_7_1_0
void FASTCALL call_10_7_1_0(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(1<<4)+0);
}
#define init_10_7_1_1
void FASTCALL call_10_7_1_1(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(1<<4)+1);
}
#define init_10_7_1_2
void FASTCALL call_10_7_1_2(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(1<<4)+2);
}
#define init_10_7_1_3
void FASTCALL call_10_7_1_3(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(1<<4)+3);
}
#define init_10_7_1_4
void FASTCALL call_10_7_1_4(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(1<<4)+4);
}
#define init_10_7_1_5
void FASTCALL call_10_7_1_5(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(1<<4)+5);
}
#define init_10_7_1_6
void FASTCALL call_10_7_1_6(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(1<<4)+6);
}
#define init_10_7_1_7
void FASTCALL call_10_7_1_7(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(1<<4)+7);
}
#define init_10_7_1_8
void FASTCALL call_10_7_1_8(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(1<<4)+8);
}
#define init_10_7_1_9
void FASTCALL call_10_7_1_9(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(1<<4)+9);
}
#define init_10_7_1_10
void FASTCALL call_10_7_1_10(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(1<<4)+10);
}
#define init_10_7_1_11
void FASTCALL call_10_7_1_11(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(1<<4)+11);
}
#define init_10_7_1_12
void FASTCALL call_10_7_1_12(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(1<<4)+12);
}
#define init_10_7_1_13
void FASTCALL call_10_7_1_13(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(1<<4)+13);
}
#define init_10_7_1_14
void FASTCALL call_10_7_1_14(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(1<<4)+14);
}
#define init_10_7_1_15
void FASTCALL call_10_7_1_15(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(1<<4)+15);
}
#define init_10_7_2_0
void FASTCALL call_10_7_2_0(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(2<<4)+0);
}
#define init_10_7_2_1
void FASTCALL call_10_7_2_1(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(2<<4)+1);
}
#define init_10_7_2_2
void FASTCALL call_10_7_2_2(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(2<<4)+2);
}
#define init_10_7_2_3
void FASTCALL call_10_7_2_3(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(2<<4)+3);
}
#define init_10_7_2_4
void FASTCALL call_10_7_2_4(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(2<<4)+4);
}
#define init_10_7_2_5
void FASTCALL call_10_7_2_5(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(2<<4)+5);
}
#define init_10_7_2_6
void FASTCALL call_10_7_2_6(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(2<<4)+6);
}
#define init_10_7_2_7
void FASTCALL call_10_7_2_7(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(2<<4)+7);
}
#define init_10_7_2_8
void FASTCALL call_10_7_2_8(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(2<<4)+8);
}
#define init_10_7_2_9
void FASTCALL call_10_7_2_9(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(2<<4)+9);
}
#define init_10_7_2_10
void FASTCALL call_10_7_2_10(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(2<<4)+10);
}
#define init_10_7_2_11
void FASTCALL call_10_7_2_11(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(2<<4)+11);
}
#define init_10_7_2_12
void FASTCALL call_10_7_2_12(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(2<<4)+12);
}
#define init_10_7_2_13
void FASTCALL call_10_7_2_13(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(2<<4)+13);
}
#define init_10_7_2_14
void FASTCALL call_10_7_2_14(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(2<<4)+14);
}
#define init_10_7_2_15
void FASTCALL call_10_7_2_15(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(2<<4)+15);
}
#define init_10_7_3_0
void FASTCALL call_10_7_3_0(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(3<<4)+0);
}
#define init_10_7_3_1
void FASTCALL call_10_7_3_1(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(3<<4)+1);
}
#define init_10_7_3_2
void FASTCALL call_10_7_3_2(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(3<<4)+2);
}
#define init_10_7_3_3
void FASTCALL call_10_7_3_3(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(3<<4)+3);
}
#define init_10_7_3_4
void FASTCALL call_10_7_3_4(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(3<<4)+4);
}
#define init_10_7_3_5
void FASTCALL call_10_7_3_5(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(3<<4)+5);
}
#define init_10_7_3_6
void FASTCALL call_10_7_3_6(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(3<<4)+6);
}
#define init_10_7_3_7
void FASTCALL call_10_7_3_7(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(3<<4)+7);
}
#define init_10_7_3_8
void FASTCALL call_10_7_3_8(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(3<<4)+8);
}
#define init_10_7_3_9
void FASTCALL call_10_7_3_9(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(3<<4)+9);
}
#define init_10_7_3_10
void FASTCALL call_10_7_3_10(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(3<<4)+10);
}
#define init_10_7_3_11
void FASTCALL call_10_7_3_11(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(3<<4)+11);
}
#define init_10_7_3_12
void FASTCALL call_10_7_3_12(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(3<<4)+12);
}
#define init_10_7_3_13
void FASTCALL call_10_7_3_13(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(3<<4)+13);
}
#define init_10_7_3_14
void FASTCALL call_10_7_3_14(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(3<<4)+14);
}
#define init_10_7_3_15
void FASTCALL call_10_7_3_15(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(3<<4)+15);
}
#define init_10_7_4_0
void FASTCALL call_10_7_4_0(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(4<<4)+0);
}
#define init_10_7_4_1
void FASTCALL call_10_7_4_1(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(4<<4)+1);
}
#define init_10_7_4_2
void FASTCALL call_10_7_4_2(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(4<<4)+2);
}
#define init_10_7_4_3
void FASTCALL call_10_7_4_3(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(4<<4)+3);
}
#define init_10_7_4_4
void FASTCALL call_10_7_4_4(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(4<<4)+4);
}
#define init_10_7_4_5
void FASTCALL call_10_7_4_5(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(4<<4)+5);
}
#define init_10_7_4_6
void FASTCALL call_10_7_4_6(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(4<<4)+6);
}
#define init_10_7_4_7
void FASTCALL call_10_7_4_7(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(4<<4)+7);
}
#define init_10_7_4_8
void FASTCALL call_10_7_4_8(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(4<<4)+8);
}
#define init_10_7_4_9
void FASTCALL call_10_7_4_9(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(4<<4)+9);
}
#define init_10_7_4_10
void FASTCALL call_10_7_4_10(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(4<<4)+10);
}
#define init_10_7_4_11
void FASTCALL call_10_7_4_11(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(4<<4)+11);
}
#define init_10_7_4_12
void FASTCALL call_10_7_4_12(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(4<<4)+12);
}
#define init_10_7_4_13
void FASTCALL call_10_7_4_13(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(4<<4)+13);
}
#define init_10_7_4_14
void FASTCALL call_10_7_4_14(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(4<<4)+14);
}
#define init_10_7_4_15
void FASTCALL call_10_7_4_15(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(4<<4)+15);
}
#define init_10_7_5_0
void FASTCALL call_10_7_5_0(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(5<<4)+0);
}
#define init_10_7_5_1
void FASTCALL call_10_7_5_1(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(5<<4)+1);
}
#define init_10_7_5_2
void FASTCALL call_10_7_5_2(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(5<<4)+2);
}
#define init_10_7_5_3
void FASTCALL call_10_7_5_3(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(5<<4)+3);
}
#define init_10_7_5_4
void FASTCALL call_10_7_5_4(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(5<<4)+4);
}
#define init_10_7_5_5
void FASTCALL call_10_7_5_5(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(5<<4)+5);
}
#define init_10_7_5_6
void FASTCALL call_10_7_5_6(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(5<<4)+6);
}
#define init_10_7_5_7
void FASTCALL call_10_7_5_7(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(5<<4)+7);
}
#define init_10_7_5_8
void FASTCALL call_10_7_5_8(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(5<<4)+8);
}
#define init_10_7_5_9
void FASTCALL call_10_7_5_9(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(5<<4)+9);
}
#define init_10_7_5_10
void FASTCALL call_10_7_5_10(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(5<<4)+10);
}
#define init_10_7_5_11
void FASTCALL call_10_7_5_11(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(5<<4)+11);
}
#define init_10_7_5_12
void FASTCALL call_10_7_5_12(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(5<<4)+12);
}
#define init_10_7_5_13
void FASTCALL call_10_7_5_13(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(5<<4)+13);
}
#define init_10_7_5_14
void FASTCALL call_10_7_5_14(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(5<<4)+14);
}
#define init_10_7_5_15
void FASTCALL call_10_7_5_15(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(5<<4)+15);
}
#define init_10_7_6_0
void FASTCALL call_10_7_6_0(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(6<<4)+0);
}
#define init_10_7_6_1
void FASTCALL call_10_7_6_1(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(6<<4)+1);
}
#define init_10_7_6_2
void FASTCALL call_10_7_6_2(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(6<<4)+2);
}
#define init_10_7_6_3
void FASTCALL call_10_7_6_3(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(6<<4)+3);
}
#define init_10_7_6_4
void FASTCALL call_10_7_6_4(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(6<<4)+4);
}
#define init_10_7_6_5
void FASTCALL call_10_7_6_5(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(6<<4)+5);
}
#define init_10_7_6_6
void FASTCALL call_10_7_6_6(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(6<<4)+6);
}
#define init_10_7_6_7
void FASTCALL call_10_7_6_7(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(6<<4)+7);
}
#define init_10_7_6_8
void FASTCALL call_10_7_6_8(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(6<<4)+8);
}
#define init_10_7_6_9
void FASTCALL call_10_7_6_9(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(6<<4)+9);
}
#define init_10_7_6_10
void FASTCALL call_10_7_6_10(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(6<<4)+10);
}
#define init_10_7_6_11
void FASTCALL call_10_7_6_11(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(6<<4)+11);
}
#define init_10_7_6_12
void FASTCALL call_10_7_6_12(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(6<<4)+12);
}
#define init_10_7_6_13
void FASTCALL call_10_7_6_13(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(6<<4)+13);
}
#define init_10_7_6_14
void FASTCALL call_10_7_6_14(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(6<<4)+14);
}
#define init_10_7_6_15
void FASTCALL call_10_7_6_15(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(6<<4)+15);
}
#define init_10_7_7_0
void FASTCALL call_10_7_7_0(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(7<<4)+0);
}
#define init_10_7_7_1
void FASTCALL call_10_7_7_1(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(7<<4)+1);
}
#define init_10_7_7_2
void FASTCALL call_10_7_7_2(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(7<<4)+2);
}
#define init_10_7_7_3
void FASTCALL call_10_7_7_3(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(7<<4)+3);
}
#define init_10_7_7_4
void FASTCALL call_10_7_7_4(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(7<<4)+4);
}
#define init_10_7_7_5
void FASTCALL call_10_7_7_5(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(7<<4)+5);
}
#define init_10_7_7_6
void FASTCALL call_10_7_7_6(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(7<<4)+6);
}
#define init_10_7_7_7
void FASTCALL call_10_7_7_7(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(7<<4)+7);
}
#define init_10_7_7_8
void FASTCALL call_10_7_7_8(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(7<<4)+8);
}
#define init_10_7_7_9
void FASTCALL call_10_7_7_9(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(7<<4)+9);
}
#define init_10_7_7_10
void FASTCALL call_10_7_7_10(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(7<<4)+10);
}
#define init_10_7_7_11
void FASTCALL call_10_7_7_11(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(7<<4)+11);
}
#define init_10_7_7_12
void FASTCALL call_10_7_7_12(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(7<<4)+12);
}
#define init_10_7_7_13
void FASTCALL call_10_7_7_13(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(7<<4)+13);
}
#define init_10_7_7_14
void FASTCALL call_10_7_7_14(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(7<<4)+14);
}
#define init_10_7_7_15
void FASTCALL call_10_7_7_15(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(7<<4)+15);
}
#define init_10_7_8_0
void FASTCALL call_10_7_8_0(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(8<<4)+0);
}
#define init_10_7_8_1
void FASTCALL call_10_7_8_1(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(8<<4)+1);
}
#define init_10_7_8_2
void FASTCALL call_10_7_8_2(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(8<<4)+2);
}
#define init_10_7_8_3
void FASTCALL call_10_7_8_3(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(8<<4)+3);
}
#define init_10_7_8_4
void FASTCALL call_10_7_8_4(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(8<<4)+4);
}
#define init_10_7_8_5
void FASTCALL call_10_7_8_5(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(8<<4)+5);
}
#define init_10_7_8_6
void FASTCALL call_10_7_8_6(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(8<<4)+6);
}
#define init_10_7_8_7
void FASTCALL call_10_7_8_7(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(8<<4)+7);
}
#define init_10_7_8_8
void FASTCALL call_10_7_8_8(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(8<<4)+8);
}
#define init_10_7_8_9
void FASTCALL call_10_7_8_9(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(8<<4)+9);
}
#define init_10_7_8_10
void FASTCALL call_10_7_8_10(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(8<<4)+10);
}
#define init_10_7_8_11
void FASTCALL call_10_7_8_11(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(8<<4)+11);
}
#define init_10_7_8_12
void FASTCALL call_10_7_8_12(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(8<<4)+12);
}
#define init_10_7_8_13
void FASTCALL call_10_7_8_13(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(8<<4)+13);
}
#define init_10_7_8_14
void FASTCALL call_10_7_8_14(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(8<<4)+14);
}
#define init_10_7_8_15
void FASTCALL call_10_7_8_15(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(8<<4)+15);
}
#define init_10_7_9_0
void FASTCALL call_10_7_9_0(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(9<<4)+0);
}
#define init_10_7_9_1
void FASTCALL call_10_7_9_1(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(9<<4)+1);
}
#define init_10_7_9_2
void FASTCALL call_10_7_9_2(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(9<<4)+2);
}
#define init_10_7_9_3
void FASTCALL call_10_7_9_3(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(9<<4)+3);
}
#define init_10_7_9_4
void FASTCALL call_10_7_9_4(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(9<<4)+4);
}
#define init_10_7_9_5
void FASTCALL call_10_7_9_5(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(9<<4)+5);
}
#define init_10_7_9_6
void FASTCALL call_10_7_9_6(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(9<<4)+6);
}
#define init_10_7_9_7
void FASTCALL call_10_7_9_7(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(9<<4)+7);
}
#define init_10_7_9_8
void FASTCALL call_10_7_9_8(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(9<<4)+8);
}
#define init_10_7_9_9
void FASTCALL call_10_7_9_9(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(9<<4)+9);
}
#define init_10_7_9_10
void FASTCALL call_10_7_9_10(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(9<<4)+10);
}
#define init_10_7_9_11
void FASTCALL call_10_7_9_11(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(9<<4)+11);
}
#define init_10_7_9_12
void FASTCALL call_10_7_9_12(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(9<<4)+12);
}
#define init_10_7_9_13
void FASTCALL call_10_7_9_13(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(9<<4)+13);
}
#define init_10_7_9_14
void FASTCALL call_10_7_9_14(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(9<<4)+14);
}
#define init_10_7_9_15
void FASTCALL call_10_7_9_15(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(9<<4)+15);
}
#define init_10_7_10_0
void FASTCALL call_10_7_10_0(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(10<<4)+0);
}
#define init_10_7_10_1
void FASTCALL call_10_7_10_1(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(10<<4)+1);
}
#define init_10_7_10_2
void FASTCALL call_10_7_10_2(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(10<<4)+2);
}
#define init_10_7_10_3
void FASTCALL call_10_7_10_3(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(10<<4)+3);
}
#define init_10_7_10_4
void FASTCALL call_10_7_10_4(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(10<<4)+4);
}
#define init_10_7_10_5
void FASTCALL call_10_7_10_5(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(10<<4)+5);
}
#define init_10_7_10_6
void FASTCALL call_10_7_10_6(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(10<<4)+6);
}
#define init_10_7_10_7
void FASTCALL call_10_7_10_7(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(10<<4)+7);
}
#define init_10_7_10_8
void FASTCALL call_10_7_10_8(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(10<<4)+8);
}
#define init_10_7_10_9
void FASTCALL call_10_7_10_9(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(10<<4)+9);
}
#define init_10_7_10_10
void FASTCALL call_10_7_10_10(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(10<<4)+10);
}
#define init_10_7_10_11
void FASTCALL call_10_7_10_11(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(10<<4)+11);
}
#define init_10_7_10_12
void FASTCALL call_10_7_10_12(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(10<<4)+12);
}
#define init_10_7_10_13
void FASTCALL call_10_7_10_13(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(10<<4)+13);
}
#define init_10_7_10_14
void FASTCALL call_10_7_10_14(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(10<<4)+14);
}
#define init_10_7_10_15
void FASTCALL call_10_7_10_15(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(10<<4)+15);
}
#define init_10_7_11_0
void FASTCALL call_10_7_11_0(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(11<<4)+0);
}
#define init_10_7_11_1
void FASTCALL call_10_7_11_1(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(11<<4)+1);
}
#define init_10_7_11_2
void FASTCALL call_10_7_11_2(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(11<<4)+2);
}
#define init_10_7_11_3
void FASTCALL call_10_7_11_3(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(11<<4)+3);
}
#define init_10_7_11_4
void FASTCALL call_10_7_11_4(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(11<<4)+4);
}
#define init_10_7_11_5
void FASTCALL call_10_7_11_5(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(11<<4)+5);
}
#define init_10_7_11_6
void FASTCALL call_10_7_11_6(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(11<<4)+6);
}
#define init_10_7_11_7
void FASTCALL call_10_7_11_7(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(11<<4)+7);
}
#define init_10_7_11_8
void FASTCALL call_10_7_11_8(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(11<<4)+8);
}
#define init_10_7_11_9
void FASTCALL call_10_7_11_9(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(11<<4)+9);
}
#define init_10_7_11_10
void FASTCALL call_10_7_11_10(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(11<<4)+10);
}
#define init_10_7_11_11
void FASTCALL call_10_7_11_11(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(11<<4)+11);
}
#define init_10_7_11_12
void FASTCALL call_10_7_11_12(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(11<<4)+12);
}
#define init_10_7_11_13
void FASTCALL call_10_7_11_13(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(11<<4)+13);
}
#define init_10_7_11_14
void FASTCALL call_10_7_11_14(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(11<<4)+14);
}
#define init_10_7_11_15
void FASTCALL call_10_7_11_15(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(11<<4)+15);
}
#define init_10_7_12_0
void FASTCALL call_10_7_12_0(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(12<<4)+0);
}
#define init_10_7_12_1
void FASTCALL call_10_7_12_1(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(12<<4)+1);
}
#define init_10_7_12_2
void FASTCALL call_10_7_12_2(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(12<<4)+2);
}
#define init_10_7_12_3
void FASTCALL call_10_7_12_3(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(12<<4)+3);
}
#define init_10_7_12_4
void FASTCALL call_10_7_12_4(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(12<<4)+4);
}
#define init_10_7_12_5
void FASTCALL call_10_7_12_5(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(12<<4)+5);
}
#define init_10_7_12_6
void FASTCALL call_10_7_12_6(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(12<<4)+6);
}
#define init_10_7_12_7
void FASTCALL call_10_7_12_7(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(12<<4)+7);
}
#define init_10_7_12_8
void FASTCALL call_10_7_12_8(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(12<<4)+8);
}
#define init_10_7_12_9
void FASTCALL call_10_7_12_9(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(12<<4)+9);
}
#define init_10_7_12_10
void FASTCALL call_10_7_12_10(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(12<<4)+10);
}
#define init_10_7_12_11
void FASTCALL call_10_7_12_11(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(12<<4)+11);
}
#define init_10_7_12_12
void FASTCALL call_10_7_12_12(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(12<<4)+12);
}
#define init_10_7_12_13
void FASTCALL call_10_7_12_13(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(12<<4)+13);
}
#define init_10_7_12_14
void FASTCALL call_10_7_12_14(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(12<<4)+14);
}
#define init_10_7_12_15
void FASTCALL call_10_7_12_15(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(12<<4)+15);
}
#define init_10_7_13_0
void FASTCALL call_10_7_13_0(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(13<<4)+0);
}
#define init_10_7_13_1
void FASTCALL call_10_7_13_1(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(13<<4)+1);
}
#define init_10_7_13_2
void FASTCALL call_10_7_13_2(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(13<<4)+2);
}
#define init_10_7_13_3
void FASTCALL call_10_7_13_3(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(13<<4)+3);
}
#define init_10_7_13_4
void FASTCALL call_10_7_13_4(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(13<<4)+4);
}
#define init_10_7_13_5
void FASTCALL call_10_7_13_5(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(13<<4)+5);
}
#define init_10_7_13_6
void FASTCALL call_10_7_13_6(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(13<<4)+6);
}
#define init_10_7_13_7
void FASTCALL call_10_7_13_7(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(13<<4)+7);
}
#define init_10_7_13_8
void FASTCALL call_10_7_13_8(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(13<<4)+8);
}
#define init_10_7_13_9
void FASTCALL call_10_7_13_9(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(13<<4)+9);
}
#define init_10_7_13_10
void FASTCALL call_10_7_13_10(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(13<<4)+10);
}
#define init_10_7_13_11
void FASTCALL call_10_7_13_11(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(13<<4)+11);
}
#define init_10_7_13_12
void FASTCALL call_10_7_13_12(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(13<<4)+12);
}
#define init_10_7_13_13
void FASTCALL call_10_7_13_13(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(13<<4)+13);
}
#define init_10_7_13_14
void FASTCALL call_10_7_13_14(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(13<<4)+14);
}
#define init_10_7_13_15
void FASTCALL call_10_7_13_15(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(13<<4)+15);
}
#define init_10_7_14_0
void FASTCALL call_10_7_14_0(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(14<<4)+0);
}
#define init_10_7_14_1
void FASTCALL call_10_7_14_1(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(14<<4)+1);
}
#define init_10_7_14_2
void FASTCALL call_10_7_14_2(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(14<<4)+2);
}
#define init_10_7_14_3
void FASTCALL call_10_7_14_3(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(14<<4)+3);
}
#define init_10_7_14_4
void FASTCALL call_10_7_14_4(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(14<<4)+4);
}
#define init_10_7_14_5
void FASTCALL call_10_7_14_5(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(14<<4)+5);
}
#define init_10_7_14_6
void FASTCALL call_10_7_14_6(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(14<<4)+6);
}
#define init_10_7_14_7
void FASTCALL call_10_7_14_7(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(14<<4)+7);
}
#define init_10_7_14_8
void FASTCALL call_10_7_14_8(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(14<<4)+8);
}
#define init_10_7_14_9
void FASTCALL call_10_7_14_9(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(14<<4)+9);
}
#define init_10_7_14_10
void FASTCALL call_10_7_14_10(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(14<<4)+10);
}
#define init_10_7_14_11
void FASTCALL call_10_7_14_11(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(14<<4)+11);
}
#define init_10_7_14_12
void FASTCALL call_10_7_14_12(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(14<<4)+12);
}
#define init_10_7_14_13
void FASTCALL call_10_7_14_13(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(14<<4)+13);
}
#define init_10_7_14_14
void FASTCALL call_10_7_14_14(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(14<<4)+14);
}
#define init_10_7_14_15
void FASTCALL call_10_7_14_15(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(14<<4)+15);
}
#define init_10_7_15_0
void FASTCALL call_10_7_15_0(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(15<<4)+0);
}
#define init_10_7_15_1
void FASTCALL call_10_7_15_1(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(15<<4)+1);
}
#define init_10_7_15_2
void FASTCALL call_10_7_15_2(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(15<<4)+2);
}
#define init_10_7_15_3
void FASTCALL call_10_7_15_3(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(15<<4)+3);
}
#define init_10_7_15_4
void FASTCALL call_10_7_15_4(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(15<<4)+4);
}
#define init_10_7_15_5
void FASTCALL call_10_7_15_5(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(15<<4)+5);
}
#define init_10_7_15_6
void FASTCALL call_10_7_15_6(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(15<<4)+6);
}
#define init_10_7_15_7
void FASTCALL call_10_7_15_7(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(15<<4)+7);
}
#define init_10_7_15_8
void FASTCALL call_10_7_15_8(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(15<<4)+8);
}
#define init_10_7_15_9
void FASTCALL call_10_7_15_9(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(15<<4)+9);
}
#define init_10_7_15_10
void FASTCALL call_10_7_15_10(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(15<<4)+10);
}
#define init_10_7_15_11
void FASTCALL call_10_7_15_11(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(15<<4)+11);
}
#define init_10_7_15_12
void FASTCALL call_10_7_15_12(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(15<<4)+12);
}
#define init_10_7_15_13
void FASTCALL call_10_7_15_13(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(15<<4)+13);
}
#define init_10_7_15_14
void FASTCALL call_10_7_15_14(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(15<<4)+14);
}
#define init_10_7_15_15
void FASTCALL call_10_7_15_15(SH2_struct* sh)
{
  SH2bra(sh, (7<<8)+(15<<4)+15);
}
#define init_10_8_0_0
void FASTCALL call_10_8_0_0(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(0<<4)+0);
}
#define init_10_8_0_1
void FASTCALL call_10_8_0_1(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(0<<4)+1);
}
#define init_10_8_0_2
void FASTCALL call_10_8_0_2(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(0<<4)+2);
}
#define init_10_8_0_3
void FASTCALL call_10_8_0_3(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(0<<4)+3);
}
#define init_10_8_0_4
void FASTCALL call_10_8_0_4(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(0<<4)+4);
}
#define init_10_8_0_5
void FASTCALL call_10_8_0_5(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(0<<4)+5);
}
#define init_10_8_0_6
void FASTCALL call_10_8_0_6(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(0<<4)+6);
}
#define init_10_8_0_7
void FASTCALL call_10_8_0_7(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(0<<4)+7);
}
#define init_10_8_0_8
void FASTCALL call_10_8_0_8(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(0<<4)+8);
}
#define init_10_8_0_9
void FASTCALL call_10_8_0_9(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(0<<4)+9);
}
#define init_10_8_0_10
void FASTCALL call_10_8_0_10(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(0<<4)+10);
}
#define init_10_8_0_11
void FASTCALL call_10_8_0_11(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(0<<4)+11);
}
#define init_10_8_0_12
void FASTCALL call_10_8_0_12(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(0<<4)+12);
}
#define init_10_8_0_13
void FASTCALL call_10_8_0_13(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(0<<4)+13);
}
#define init_10_8_0_14
void FASTCALL call_10_8_0_14(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(0<<4)+14);
}
#define init_10_8_0_15
void FASTCALL call_10_8_0_15(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(0<<4)+15);
}
#define init_10_8_1_0
void FASTCALL call_10_8_1_0(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(1<<4)+0);
}
#define init_10_8_1_1
void FASTCALL call_10_8_1_1(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(1<<4)+1);
}
#define init_10_8_1_2
void FASTCALL call_10_8_1_2(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(1<<4)+2);
}
#define init_10_8_1_3
void FASTCALL call_10_8_1_3(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(1<<4)+3);
}
#define init_10_8_1_4
void FASTCALL call_10_8_1_4(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(1<<4)+4);
}
#define init_10_8_1_5
void FASTCALL call_10_8_1_5(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(1<<4)+5);
}
#define init_10_8_1_6
void FASTCALL call_10_8_1_6(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(1<<4)+6);
}
#define init_10_8_1_7
void FASTCALL call_10_8_1_7(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(1<<4)+7);
}
#define init_10_8_1_8
void FASTCALL call_10_8_1_8(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(1<<4)+8);
}
#define init_10_8_1_9
void FASTCALL call_10_8_1_9(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(1<<4)+9);
}
#define init_10_8_1_10
void FASTCALL call_10_8_1_10(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(1<<4)+10);
}
#define init_10_8_1_11
void FASTCALL call_10_8_1_11(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(1<<4)+11);
}
#define init_10_8_1_12
void FASTCALL call_10_8_1_12(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(1<<4)+12);
}
#define init_10_8_1_13
void FASTCALL call_10_8_1_13(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(1<<4)+13);
}
#define init_10_8_1_14
void FASTCALL call_10_8_1_14(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(1<<4)+14);
}
#define init_10_8_1_15
void FASTCALL call_10_8_1_15(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(1<<4)+15);
}
#define init_10_8_2_0
void FASTCALL call_10_8_2_0(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(2<<4)+0);
}
#define init_10_8_2_1
void FASTCALL call_10_8_2_1(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(2<<4)+1);
}
#define init_10_8_2_2
void FASTCALL call_10_8_2_2(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(2<<4)+2);
}
#define init_10_8_2_3
void FASTCALL call_10_8_2_3(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(2<<4)+3);
}
#define init_10_8_2_4
void FASTCALL call_10_8_2_4(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(2<<4)+4);
}
#define init_10_8_2_5
void FASTCALL call_10_8_2_5(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(2<<4)+5);
}
#define init_10_8_2_6
void FASTCALL call_10_8_2_6(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(2<<4)+6);
}
#define init_10_8_2_7
void FASTCALL call_10_8_2_7(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(2<<4)+7);
}
#define init_10_8_2_8
void FASTCALL call_10_8_2_8(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(2<<4)+8);
}
#define init_10_8_2_9
void FASTCALL call_10_8_2_9(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(2<<4)+9);
}
#define init_10_8_2_10
void FASTCALL call_10_8_2_10(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(2<<4)+10);
}
#define init_10_8_2_11
void FASTCALL call_10_8_2_11(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(2<<4)+11);
}
#define init_10_8_2_12
void FASTCALL call_10_8_2_12(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(2<<4)+12);
}
#define init_10_8_2_13
void FASTCALL call_10_8_2_13(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(2<<4)+13);
}
#define init_10_8_2_14
void FASTCALL call_10_8_2_14(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(2<<4)+14);
}
#define init_10_8_2_15
void FASTCALL call_10_8_2_15(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(2<<4)+15);
}
#define init_10_8_3_0
void FASTCALL call_10_8_3_0(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(3<<4)+0);
}
#define init_10_8_3_1
void FASTCALL call_10_8_3_1(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(3<<4)+1);
}
#define init_10_8_3_2
void FASTCALL call_10_8_3_2(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(3<<4)+2);
}
#define init_10_8_3_3
void FASTCALL call_10_8_3_3(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(3<<4)+3);
}
#define init_10_8_3_4
void FASTCALL call_10_8_3_4(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(3<<4)+4);
}
#define init_10_8_3_5
void FASTCALL call_10_8_3_5(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(3<<4)+5);
}
#define init_10_8_3_6
void FASTCALL call_10_8_3_6(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(3<<4)+6);
}
#define init_10_8_3_7
void FASTCALL call_10_8_3_7(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(3<<4)+7);
}
#define init_10_8_3_8
void FASTCALL call_10_8_3_8(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(3<<4)+8);
}
#define init_10_8_3_9
void FASTCALL call_10_8_3_9(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(3<<4)+9);
}
#define init_10_8_3_10
void FASTCALL call_10_8_3_10(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(3<<4)+10);
}
#define init_10_8_3_11
void FASTCALL call_10_8_3_11(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(3<<4)+11);
}
#define init_10_8_3_12
void FASTCALL call_10_8_3_12(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(3<<4)+12);
}
#define init_10_8_3_13
void FASTCALL call_10_8_3_13(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(3<<4)+13);
}
#define init_10_8_3_14
void FASTCALL call_10_8_3_14(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(3<<4)+14);
}
#define init_10_8_3_15
void FASTCALL call_10_8_3_15(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(3<<4)+15);
}
#define init_10_8_4_0
void FASTCALL call_10_8_4_0(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(4<<4)+0);
}
#define init_10_8_4_1
void FASTCALL call_10_8_4_1(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(4<<4)+1);
}
#define init_10_8_4_2
void FASTCALL call_10_8_4_2(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(4<<4)+2);
}
#define init_10_8_4_3
void FASTCALL call_10_8_4_3(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(4<<4)+3);
}
#define init_10_8_4_4
void FASTCALL call_10_8_4_4(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(4<<4)+4);
}
#define init_10_8_4_5
void FASTCALL call_10_8_4_5(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(4<<4)+5);
}
#define init_10_8_4_6
void FASTCALL call_10_8_4_6(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(4<<4)+6);
}
#define init_10_8_4_7
void FASTCALL call_10_8_4_7(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(4<<4)+7);
}
#define init_10_8_4_8
void FASTCALL call_10_8_4_8(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(4<<4)+8);
}
#define init_10_8_4_9
void FASTCALL call_10_8_4_9(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(4<<4)+9);
}
#define init_10_8_4_10
void FASTCALL call_10_8_4_10(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(4<<4)+10);
}
#define init_10_8_4_11
void FASTCALL call_10_8_4_11(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(4<<4)+11);
}
#define init_10_8_4_12
void FASTCALL call_10_8_4_12(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(4<<4)+12);
}
#define init_10_8_4_13
void FASTCALL call_10_8_4_13(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(4<<4)+13);
}
#define init_10_8_4_14
void FASTCALL call_10_8_4_14(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(4<<4)+14);
}
#define init_10_8_4_15
void FASTCALL call_10_8_4_15(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(4<<4)+15);
}
#define init_10_8_5_0
void FASTCALL call_10_8_5_0(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(5<<4)+0);
}
#define init_10_8_5_1
void FASTCALL call_10_8_5_1(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(5<<4)+1);
}
#define init_10_8_5_2
void FASTCALL call_10_8_5_2(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(5<<4)+2);
}
#define init_10_8_5_3
void FASTCALL call_10_8_5_3(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(5<<4)+3);
}
#define init_10_8_5_4
void FASTCALL call_10_8_5_4(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(5<<4)+4);
}
#define init_10_8_5_5
void FASTCALL call_10_8_5_5(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(5<<4)+5);
}
#define init_10_8_5_6
void FASTCALL call_10_8_5_6(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(5<<4)+6);
}
#define init_10_8_5_7
void FASTCALL call_10_8_5_7(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(5<<4)+7);
}
#define init_10_8_5_8
void FASTCALL call_10_8_5_8(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(5<<4)+8);
}
#define init_10_8_5_9
void FASTCALL call_10_8_5_9(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(5<<4)+9);
}
#define init_10_8_5_10
void FASTCALL call_10_8_5_10(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(5<<4)+10);
}
#define init_10_8_5_11
void FASTCALL call_10_8_5_11(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(5<<4)+11);
}
#define init_10_8_5_12
void FASTCALL call_10_8_5_12(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(5<<4)+12);
}
#define init_10_8_5_13
void FASTCALL call_10_8_5_13(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(5<<4)+13);
}
#define init_10_8_5_14
void FASTCALL call_10_8_5_14(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(5<<4)+14);
}
#define init_10_8_5_15
void FASTCALL call_10_8_5_15(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(5<<4)+15);
}
#define init_10_8_6_0
void FASTCALL call_10_8_6_0(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(6<<4)+0);
}
#define init_10_8_6_1
void FASTCALL call_10_8_6_1(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(6<<4)+1);
}
#define init_10_8_6_2
void FASTCALL call_10_8_6_2(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(6<<4)+2);
}
#define init_10_8_6_3
void FASTCALL call_10_8_6_3(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(6<<4)+3);
}
#define init_10_8_6_4
void FASTCALL call_10_8_6_4(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(6<<4)+4);
}
#define init_10_8_6_5
void FASTCALL call_10_8_6_5(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(6<<4)+5);
}
#define init_10_8_6_6
void FASTCALL call_10_8_6_6(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(6<<4)+6);
}
#define init_10_8_6_7
void FASTCALL call_10_8_6_7(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(6<<4)+7);
}
#define init_10_8_6_8
void FASTCALL call_10_8_6_8(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(6<<4)+8);
}
#define init_10_8_6_9
void FASTCALL call_10_8_6_9(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(6<<4)+9);
}
#define init_10_8_6_10
void FASTCALL call_10_8_6_10(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(6<<4)+10);
}
#define init_10_8_6_11
void FASTCALL call_10_8_6_11(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(6<<4)+11);
}
#define init_10_8_6_12
void FASTCALL call_10_8_6_12(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(6<<4)+12);
}
#define init_10_8_6_13
void FASTCALL call_10_8_6_13(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(6<<4)+13);
}
#define init_10_8_6_14
void FASTCALL call_10_8_6_14(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(6<<4)+14);
}
#define init_10_8_6_15
void FASTCALL call_10_8_6_15(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(6<<4)+15);
}
#define init_10_8_7_0
void FASTCALL call_10_8_7_0(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(7<<4)+0);
}
#define init_10_8_7_1
void FASTCALL call_10_8_7_1(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(7<<4)+1);
}
#define init_10_8_7_2
void FASTCALL call_10_8_7_2(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(7<<4)+2);
}
#define init_10_8_7_3
void FASTCALL call_10_8_7_3(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(7<<4)+3);
}
#define init_10_8_7_4
void FASTCALL call_10_8_7_4(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(7<<4)+4);
}
#define init_10_8_7_5
void FASTCALL call_10_8_7_5(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(7<<4)+5);
}
#define init_10_8_7_6
void FASTCALL call_10_8_7_6(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(7<<4)+6);
}
#define init_10_8_7_7
void FASTCALL call_10_8_7_7(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(7<<4)+7);
}
#define init_10_8_7_8
void FASTCALL call_10_8_7_8(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(7<<4)+8);
}
#define init_10_8_7_9
void FASTCALL call_10_8_7_9(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(7<<4)+9);
}
#define init_10_8_7_10
void FASTCALL call_10_8_7_10(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(7<<4)+10);
}
#define init_10_8_7_11
void FASTCALL call_10_8_7_11(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(7<<4)+11);
}
#define init_10_8_7_12
void FASTCALL call_10_8_7_12(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(7<<4)+12);
}
#define init_10_8_7_13
void FASTCALL call_10_8_7_13(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(7<<4)+13);
}
#define init_10_8_7_14
void FASTCALL call_10_8_7_14(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(7<<4)+14);
}
#define init_10_8_7_15
void FASTCALL call_10_8_7_15(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(7<<4)+15);
}
#define init_10_8_8_0
void FASTCALL call_10_8_8_0(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(8<<4)+0);
}
#define init_10_8_8_1
void FASTCALL call_10_8_8_1(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(8<<4)+1);
}
#define init_10_8_8_2
void FASTCALL call_10_8_8_2(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(8<<4)+2);
}
#define init_10_8_8_3
void FASTCALL call_10_8_8_3(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(8<<4)+3);
}
#define init_10_8_8_4
void FASTCALL call_10_8_8_4(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(8<<4)+4);
}
#define init_10_8_8_5
void FASTCALL call_10_8_8_5(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(8<<4)+5);
}
#define init_10_8_8_6
void FASTCALL call_10_8_8_6(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(8<<4)+6);
}
#define init_10_8_8_7
void FASTCALL call_10_8_8_7(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(8<<4)+7);
}
#define init_10_8_8_8
void FASTCALL call_10_8_8_8(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(8<<4)+8);
}
#define init_10_8_8_9
void FASTCALL call_10_8_8_9(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(8<<4)+9);
}
#define init_10_8_8_10
void FASTCALL call_10_8_8_10(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(8<<4)+10);
}
#define init_10_8_8_11
void FASTCALL call_10_8_8_11(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(8<<4)+11);
}
#define init_10_8_8_12
void FASTCALL call_10_8_8_12(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(8<<4)+12);
}
#define init_10_8_8_13
void FASTCALL call_10_8_8_13(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(8<<4)+13);
}
#define init_10_8_8_14
void FASTCALL call_10_8_8_14(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(8<<4)+14);
}
#define init_10_8_8_15
void FASTCALL call_10_8_8_15(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(8<<4)+15);
}
#define init_10_8_9_0
void FASTCALL call_10_8_9_0(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(9<<4)+0);
}
#define init_10_8_9_1
void FASTCALL call_10_8_9_1(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(9<<4)+1);
}
#define init_10_8_9_2
void FASTCALL call_10_8_9_2(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(9<<4)+2);
}
#define init_10_8_9_3
void FASTCALL call_10_8_9_3(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(9<<4)+3);
}
#define init_10_8_9_4
void FASTCALL call_10_8_9_4(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(9<<4)+4);
}
#define init_10_8_9_5
void FASTCALL call_10_8_9_5(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(9<<4)+5);
}
#define init_10_8_9_6
void FASTCALL call_10_8_9_6(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(9<<4)+6);
}
#define init_10_8_9_7
void FASTCALL call_10_8_9_7(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(9<<4)+7);
}
#define init_10_8_9_8
void FASTCALL call_10_8_9_8(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(9<<4)+8);
}
#define init_10_8_9_9
void FASTCALL call_10_8_9_9(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(9<<4)+9);
}
#define init_10_8_9_10
void FASTCALL call_10_8_9_10(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(9<<4)+10);
}
#define init_10_8_9_11
void FASTCALL call_10_8_9_11(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(9<<4)+11);
}
#define init_10_8_9_12
void FASTCALL call_10_8_9_12(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(9<<4)+12);
}
#define init_10_8_9_13
void FASTCALL call_10_8_9_13(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(9<<4)+13);
}
#define init_10_8_9_14
void FASTCALL call_10_8_9_14(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(9<<4)+14);
}
#define init_10_8_9_15
void FASTCALL call_10_8_9_15(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(9<<4)+15);
}
#define init_10_8_10_0
void FASTCALL call_10_8_10_0(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(10<<4)+0);
}
#define init_10_8_10_1
void FASTCALL call_10_8_10_1(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(10<<4)+1);
}
#define init_10_8_10_2
void FASTCALL call_10_8_10_2(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(10<<4)+2);
}
#define init_10_8_10_3
void FASTCALL call_10_8_10_3(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(10<<4)+3);
}
#define init_10_8_10_4
void FASTCALL call_10_8_10_4(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(10<<4)+4);
}
#define init_10_8_10_5
void FASTCALL call_10_8_10_5(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(10<<4)+5);
}
#define init_10_8_10_6
void FASTCALL call_10_8_10_6(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(10<<4)+6);
}
#define init_10_8_10_7
void FASTCALL call_10_8_10_7(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(10<<4)+7);
}
#define init_10_8_10_8
void FASTCALL call_10_8_10_8(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(10<<4)+8);
}
#define init_10_8_10_9
void FASTCALL call_10_8_10_9(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(10<<4)+9);
}
#define init_10_8_10_10
void FASTCALL call_10_8_10_10(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(10<<4)+10);
}
#define init_10_8_10_11
void FASTCALL call_10_8_10_11(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(10<<4)+11);
}
#define init_10_8_10_12
void FASTCALL call_10_8_10_12(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(10<<4)+12);
}
#define init_10_8_10_13
void FASTCALL call_10_8_10_13(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(10<<4)+13);
}
#define init_10_8_10_14
void FASTCALL call_10_8_10_14(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(10<<4)+14);
}
#define init_10_8_10_15
void FASTCALL call_10_8_10_15(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(10<<4)+15);
}
#define init_10_8_11_0
void FASTCALL call_10_8_11_0(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(11<<4)+0);
}
#define init_10_8_11_1
void FASTCALL call_10_8_11_1(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(11<<4)+1);
}
#define init_10_8_11_2
void FASTCALL call_10_8_11_2(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(11<<4)+2);
}
#define init_10_8_11_3
void FASTCALL call_10_8_11_3(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(11<<4)+3);
}
#define init_10_8_11_4
void FASTCALL call_10_8_11_4(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(11<<4)+4);
}
#define init_10_8_11_5
void FASTCALL call_10_8_11_5(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(11<<4)+5);
}
#define init_10_8_11_6
void FASTCALL call_10_8_11_6(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(11<<4)+6);
}
#define init_10_8_11_7
void FASTCALL call_10_8_11_7(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(11<<4)+7);
}
#define init_10_8_11_8
void FASTCALL call_10_8_11_8(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(11<<4)+8);
}
#define init_10_8_11_9
void FASTCALL call_10_8_11_9(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(11<<4)+9);
}
#define init_10_8_11_10
void FASTCALL call_10_8_11_10(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(11<<4)+10);
}
#define init_10_8_11_11
void FASTCALL call_10_8_11_11(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(11<<4)+11);
}
#define init_10_8_11_12
void FASTCALL call_10_8_11_12(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(11<<4)+12);
}
#define init_10_8_11_13
void FASTCALL call_10_8_11_13(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(11<<4)+13);
}
#define init_10_8_11_14
void FASTCALL call_10_8_11_14(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(11<<4)+14);
}
#define init_10_8_11_15
void FASTCALL call_10_8_11_15(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(11<<4)+15);
}
#define init_10_8_12_0
void FASTCALL call_10_8_12_0(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(12<<4)+0);
}
#define init_10_8_12_1
void FASTCALL call_10_8_12_1(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(12<<4)+1);
}
#define init_10_8_12_2
void FASTCALL call_10_8_12_2(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(12<<4)+2);
}
#define init_10_8_12_3
void FASTCALL call_10_8_12_3(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(12<<4)+3);
}
#define init_10_8_12_4
void FASTCALL call_10_8_12_4(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(12<<4)+4);
}
#define init_10_8_12_5
void FASTCALL call_10_8_12_5(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(12<<4)+5);
}
#define init_10_8_12_6
void FASTCALL call_10_8_12_6(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(12<<4)+6);
}
#define init_10_8_12_7
void FASTCALL call_10_8_12_7(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(12<<4)+7);
}
#define init_10_8_12_8
void FASTCALL call_10_8_12_8(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(12<<4)+8);
}
#define init_10_8_12_9
void FASTCALL call_10_8_12_9(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(12<<4)+9);
}
#define init_10_8_12_10
void FASTCALL call_10_8_12_10(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(12<<4)+10);
}
#define init_10_8_12_11
void FASTCALL call_10_8_12_11(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(12<<4)+11);
}
#define init_10_8_12_12
void FASTCALL call_10_8_12_12(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(12<<4)+12);
}
#define init_10_8_12_13
void FASTCALL call_10_8_12_13(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(12<<4)+13);
}
#define init_10_8_12_14
void FASTCALL call_10_8_12_14(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(12<<4)+14);
}
#define init_10_8_12_15
void FASTCALL call_10_8_12_15(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(12<<4)+15);
}
#define init_10_8_13_0
void FASTCALL call_10_8_13_0(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(13<<4)+0);
}
#define init_10_8_13_1
void FASTCALL call_10_8_13_1(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(13<<4)+1);
}
#define init_10_8_13_2
void FASTCALL call_10_8_13_2(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(13<<4)+2);
}
#define init_10_8_13_3
void FASTCALL call_10_8_13_3(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(13<<4)+3);
}
#define init_10_8_13_4
void FASTCALL call_10_8_13_4(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(13<<4)+4);
}
#define init_10_8_13_5
void FASTCALL call_10_8_13_5(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(13<<4)+5);
}
#define init_10_8_13_6
void FASTCALL call_10_8_13_6(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(13<<4)+6);
}
#define init_10_8_13_7
void FASTCALL call_10_8_13_7(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(13<<4)+7);
}
#define init_10_8_13_8
void FASTCALL call_10_8_13_8(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(13<<4)+8);
}
#define init_10_8_13_9
void FASTCALL call_10_8_13_9(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(13<<4)+9);
}
#define init_10_8_13_10
void FASTCALL call_10_8_13_10(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(13<<4)+10);
}
#define init_10_8_13_11
void FASTCALL call_10_8_13_11(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(13<<4)+11);
}
#define init_10_8_13_12
void FASTCALL call_10_8_13_12(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(13<<4)+12);
}
#define init_10_8_13_13
void FASTCALL call_10_8_13_13(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(13<<4)+13);
}
#define init_10_8_13_14
void FASTCALL call_10_8_13_14(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(13<<4)+14);
}
#define init_10_8_13_15
void FASTCALL call_10_8_13_15(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(13<<4)+15);
}
#define init_10_8_14_0
void FASTCALL call_10_8_14_0(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(14<<4)+0);
}
#define init_10_8_14_1
void FASTCALL call_10_8_14_1(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(14<<4)+1);
}
#define init_10_8_14_2
void FASTCALL call_10_8_14_2(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(14<<4)+2);
}
#define init_10_8_14_3
void FASTCALL call_10_8_14_3(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(14<<4)+3);
}
#define init_10_8_14_4
void FASTCALL call_10_8_14_4(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(14<<4)+4);
}
#define init_10_8_14_5
void FASTCALL call_10_8_14_5(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(14<<4)+5);
}
#define init_10_8_14_6
void FASTCALL call_10_8_14_6(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(14<<4)+6);
}
#define init_10_8_14_7
void FASTCALL call_10_8_14_7(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(14<<4)+7);
}
#define init_10_8_14_8
void FASTCALL call_10_8_14_8(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(14<<4)+8);
}
#define init_10_8_14_9
void FASTCALL call_10_8_14_9(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(14<<4)+9);
}
#define init_10_8_14_10
void FASTCALL call_10_8_14_10(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(14<<4)+10);
}
#define init_10_8_14_11
void FASTCALL call_10_8_14_11(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(14<<4)+11);
}
#define init_10_8_14_12
void FASTCALL call_10_8_14_12(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(14<<4)+12);
}
#define init_10_8_14_13
void FASTCALL call_10_8_14_13(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(14<<4)+13);
}
#define init_10_8_14_14
void FASTCALL call_10_8_14_14(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(14<<4)+14);
}
#define init_10_8_14_15
void FASTCALL call_10_8_14_15(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(14<<4)+15);
}
#define init_10_8_15_0
void FASTCALL call_10_8_15_0(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(15<<4)+0);
}
#define init_10_8_15_1
void FASTCALL call_10_8_15_1(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(15<<4)+1);
}
#define init_10_8_15_2
void FASTCALL call_10_8_15_2(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(15<<4)+2);
}
#define init_10_8_15_3
void FASTCALL call_10_8_15_3(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(15<<4)+3);
}
#define init_10_8_15_4
void FASTCALL call_10_8_15_4(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(15<<4)+4);
}
#define init_10_8_15_5
void FASTCALL call_10_8_15_5(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(15<<4)+5);
}
#define init_10_8_15_6
void FASTCALL call_10_8_15_6(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(15<<4)+6);
}
#define init_10_8_15_7
void FASTCALL call_10_8_15_7(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(15<<4)+7);
}
#define init_10_8_15_8
void FASTCALL call_10_8_15_8(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(15<<4)+8);
}
#define init_10_8_15_9
void FASTCALL call_10_8_15_9(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(15<<4)+9);
}
#define init_10_8_15_10
void FASTCALL call_10_8_15_10(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(15<<4)+10);
}
#define init_10_8_15_11
void FASTCALL call_10_8_15_11(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(15<<4)+11);
}
#define init_10_8_15_12
void FASTCALL call_10_8_15_12(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(15<<4)+12);
}
#define init_10_8_15_13
void FASTCALL call_10_8_15_13(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(15<<4)+13);
}
#define init_10_8_15_14
void FASTCALL call_10_8_15_14(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(15<<4)+14);
}
#define init_10_8_15_15
void FASTCALL call_10_8_15_15(SH2_struct* sh)
{
  SH2bra(sh, (8<<8)+(15<<4)+15);
}
#define init_10_9_0_0
void FASTCALL call_10_9_0_0(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(0<<4)+0);
}
#define init_10_9_0_1
void FASTCALL call_10_9_0_1(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(0<<4)+1);
}
#define init_10_9_0_2
void FASTCALL call_10_9_0_2(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(0<<4)+2);
}
#define init_10_9_0_3
void FASTCALL call_10_9_0_3(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(0<<4)+3);
}
#define init_10_9_0_4
void FASTCALL call_10_9_0_4(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(0<<4)+4);
}
#define init_10_9_0_5
void FASTCALL call_10_9_0_5(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(0<<4)+5);
}
#define init_10_9_0_6
void FASTCALL call_10_9_0_6(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(0<<4)+6);
}
#define init_10_9_0_7
void FASTCALL call_10_9_0_7(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(0<<4)+7);
}
#define init_10_9_0_8
void FASTCALL call_10_9_0_8(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(0<<4)+8);
}
#define init_10_9_0_9
void FASTCALL call_10_9_0_9(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(0<<4)+9);
}
#define init_10_9_0_10
void FASTCALL call_10_9_0_10(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(0<<4)+10);
}
#define init_10_9_0_11
void FASTCALL call_10_9_0_11(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(0<<4)+11);
}
#define init_10_9_0_12
void FASTCALL call_10_9_0_12(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(0<<4)+12);
}
#define init_10_9_0_13
void FASTCALL call_10_9_0_13(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(0<<4)+13);
}
#define init_10_9_0_14
void FASTCALL call_10_9_0_14(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(0<<4)+14);
}
#define init_10_9_0_15
void FASTCALL call_10_9_0_15(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(0<<4)+15);
}
#define init_10_9_1_0
void FASTCALL call_10_9_1_0(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(1<<4)+0);
}
#define init_10_9_1_1
void FASTCALL call_10_9_1_1(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(1<<4)+1);
}
#define init_10_9_1_2
void FASTCALL call_10_9_1_2(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(1<<4)+2);
}
#define init_10_9_1_3
void FASTCALL call_10_9_1_3(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(1<<4)+3);
}
#define init_10_9_1_4
void FASTCALL call_10_9_1_4(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(1<<4)+4);
}
#define init_10_9_1_5
void FASTCALL call_10_9_1_5(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(1<<4)+5);
}
#define init_10_9_1_6
void FASTCALL call_10_9_1_6(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(1<<4)+6);
}
#define init_10_9_1_7
void FASTCALL call_10_9_1_7(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(1<<4)+7);
}
#define init_10_9_1_8
void FASTCALL call_10_9_1_8(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(1<<4)+8);
}
#define init_10_9_1_9
void FASTCALL call_10_9_1_9(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(1<<4)+9);
}
#define init_10_9_1_10
void FASTCALL call_10_9_1_10(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(1<<4)+10);
}
#define init_10_9_1_11
void FASTCALL call_10_9_1_11(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(1<<4)+11);
}
#define init_10_9_1_12
void FASTCALL call_10_9_1_12(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(1<<4)+12);
}
#define init_10_9_1_13
void FASTCALL call_10_9_1_13(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(1<<4)+13);
}
#define init_10_9_1_14
void FASTCALL call_10_9_1_14(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(1<<4)+14);
}
#define init_10_9_1_15
void FASTCALL call_10_9_1_15(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(1<<4)+15);
}
#define init_10_9_2_0
void FASTCALL call_10_9_2_0(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(2<<4)+0);
}
#define init_10_9_2_1
void FASTCALL call_10_9_2_1(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(2<<4)+1);
}
#define init_10_9_2_2
void FASTCALL call_10_9_2_2(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(2<<4)+2);
}
#define init_10_9_2_3
void FASTCALL call_10_9_2_3(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(2<<4)+3);
}
#define init_10_9_2_4
void FASTCALL call_10_9_2_4(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(2<<4)+4);
}
#define init_10_9_2_5
void FASTCALL call_10_9_2_5(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(2<<4)+5);
}
#define init_10_9_2_6
void FASTCALL call_10_9_2_6(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(2<<4)+6);
}
#define init_10_9_2_7
void FASTCALL call_10_9_2_7(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(2<<4)+7);
}
#define init_10_9_2_8
void FASTCALL call_10_9_2_8(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(2<<4)+8);
}
#define init_10_9_2_9
void FASTCALL call_10_9_2_9(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(2<<4)+9);
}
#define init_10_9_2_10
void FASTCALL call_10_9_2_10(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(2<<4)+10);
}
#define init_10_9_2_11
void FASTCALL call_10_9_2_11(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(2<<4)+11);
}
#define init_10_9_2_12
void FASTCALL call_10_9_2_12(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(2<<4)+12);
}
#define init_10_9_2_13
void FASTCALL call_10_9_2_13(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(2<<4)+13);
}
#define init_10_9_2_14
void FASTCALL call_10_9_2_14(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(2<<4)+14);
}
#define init_10_9_2_15
void FASTCALL call_10_9_2_15(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(2<<4)+15);
}
#define init_10_9_3_0
void FASTCALL call_10_9_3_0(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(3<<4)+0);
}
#define init_10_9_3_1
void FASTCALL call_10_9_3_1(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(3<<4)+1);
}
#define init_10_9_3_2
void FASTCALL call_10_9_3_2(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(3<<4)+2);
}
#define init_10_9_3_3
void FASTCALL call_10_9_3_3(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(3<<4)+3);
}
#define init_10_9_3_4
void FASTCALL call_10_9_3_4(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(3<<4)+4);
}
#define init_10_9_3_5
void FASTCALL call_10_9_3_5(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(3<<4)+5);
}
#define init_10_9_3_6
void FASTCALL call_10_9_3_6(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(3<<4)+6);
}
#define init_10_9_3_7
void FASTCALL call_10_9_3_7(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(3<<4)+7);
}
#define init_10_9_3_8
void FASTCALL call_10_9_3_8(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(3<<4)+8);
}
#define init_10_9_3_9
void FASTCALL call_10_9_3_9(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(3<<4)+9);
}
#define init_10_9_3_10
void FASTCALL call_10_9_3_10(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(3<<4)+10);
}
#define init_10_9_3_11
void FASTCALL call_10_9_3_11(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(3<<4)+11);
}
#define init_10_9_3_12
void FASTCALL call_10_9_3_12(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(3<<4)+12);
}
#define init_10_9_3_13
void FASTCALL call_10_9_3_13(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(3<<4)+13);
}
#define init_10_9_3_14
void FASTCALL call_10_9_3_14(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(3<<4)+14);
}
#define init_10_9_3_15
void FASTCALL call_10_9_3_15(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(3<<4)+15);
}
#define init_10_9_4_0
void FASTCALL call_10_9_4_0(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(4<<4)+0);
}
#define init_10_9_4_1
void FASTCALL call_10_9_4_1(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(4<<4)+1);
}
#define init_10_9_4_2
void FASTCALL call_10_9_4_2(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(4<<4)+2);
}
#define init_10_9_4_3
void FASTCALL call_10_9_4_3(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(4<<4)+3);
}
#define init_10_9_4_4
void FASTCALL call_10_9_4_4(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(4<<4)+4);
}
#define init_10_9_4_5
void FASTCALL call_10_9_4_5(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(4<<4)+5);
}
#define init_10_9_4_6
void FASTCALL call_10_9_4_6(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(4<<4)+6);
}
#define init_10_9_4_7
void FASTCALL call_10_9_4_7(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(4<<4)+7);
}
#define init_10_9_4_8
void FASTCALL call_10_9_4_8(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(4<<4)+8);
}
#define init_10_9_4_9
void FASTCALL call_10_9_4_9(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(4<<4)+9);
}
#define init_10_9_4_10
void FASTCALL call_10_9_4_10(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(4<<4)+10);
}
#define init_10_9_4_11
void FASTCALL call_10_9_4_11(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(4<<4)+11);
}
#define init_10_9_4_12
void FASTCALL call_10_9_4_12(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(4<<4)+12);
}
#define init_10_9_4_13
void FASTCALL call_10_9_4_13(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(4<<4)+13);
}
#define init_10_9_4_14
void FASTCALL call_10_9_4_14(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(4<<4)+14);
}
#define init_10_9_4_15
void FASTCALL call_10_9_4_15(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(4<<4)+15);
}
#define init_10_9_5_0
void FASTCALL call_10_9_5_0(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(5<<4)+0);
}
#define init_10_9_5_1
void FASTCALL call_10_9_5_1(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(5<<4)+1);
}
#define init_10_9_5_2
void FASTCALL call_10_9_5_2(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(5<<4)+2);
}
#define init_10_9_5_3
void FASTCALL call_10_9_5_3(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(5<<4)+3);
}
#define init_10_9_5_4
void FASTCALL call_10_9_5_4(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(5<<4)+4);
}
#define init_10_9_5_5
void FASTCALL call_10_9_5_5(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(5<<4)+5);
}
#define init_10_9_5_6
void FASTCALL call_10_9_5_6(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(5<<4)+6);
}
#define init_10_9_5_7
void FASTCALL call_10_9_5_7(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(5<<4)+7);
}
#define init_10_9_5_8
void FASTCALL call_10_9_5_8(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(5<<4)+8);
}
#define init_10_9_5_9
void FASTCALL call_10_9_5_9(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(5<<4)+9);
}
#define init_10_9_5_10
void FASTCALL call_10_9_5_10(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(5<<4)+10);
}
#define init_10_9_5_11
void FASTCALL call_10_9_5_11(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(5<<4)+11);
}
#define init_10_9_5_12
void FASTCALL call_10_9_5_12(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(5<<4)+12);
}
#define init_10_9_5_13
void FASTCALL call_10_9_5_13(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(5<<4)+13);
}
#define init_10_9_5_14
void FASTCALL call_10_9_5_14(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(5<<4)+14);
}
#define init_10_9_5_15
void FASTCALL call_10_9_5_15(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(5<<4)+15);
}
#define init_10_9_6_0
void FASTCALL call_10_9_6_0(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(6<<4)+0);
}
#define init_10_9_6_1
void FASTCALL call_10_9_6_1(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(6<<4)+1);
}
#define init_10_9_6_2
void FASTCALL call_10_9_6_2(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(6<<4)+2);
}
#define init_10_9_6_3
void FASTCALL call_10_9_6_3(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(6<<4)+3);
}
#define init_10_9_6_4
void FASTCALL call_10_9_6_4(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(6<<4)+4);
}
#define init_10_9_6_5
void FASTCALL call_10_9_6_5(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(6<<4)+5);
}
#define init_10_9_6_6
void FASTCALL call_10_9_6_6(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(6<<4)+6);
}
#define init_10_9_6_7
void FASTCALL call_10_9_6_7(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(6<<4)+7);
}
#define init_10_9_6_8
void FASTCALL call_10_9_6_8(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(6<<4)+8);
}
#define init_10_9_6_9
void FASTCALL call_10_9_6_9(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(6<<4)+9);
}
#define init_10_9_6_10
void FASTCALL call_10_9_6_10(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(6<<4)+10);
}
#define init_10_9_6_11
void FASTCALL call_10_9_6_11(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(6<<4)+11);
}
#define init_10_9_6_12
void FASTCALL call_10_9_6_12(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(6<<4)+12);
}
#define init_10_9_6_13
void FASTCALL call_10_9_6_13(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(6<<4)+13);
}
#define init_10_9_6_14
void FASTCALL call_10_9_6_14(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(6<<4)+14);
}
#define init_10_9_6_15
void FASTCALL call_10_9_6_15(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(6<<4)+15);
}
#define init_10_9_7_0
void FASTCALL call_10_9_7_0(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(7<<4)+0);
}
#define init_10_9_7_1
void FASTCALL call_10_9_7_1(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(7<<4)+1);
}
#define init_10_9_7_2
void FASTCALL call_10_9_7_2(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(7<<4)+2);
}
#define init_10_9_7_3
void FASTCALL call_10_9_7_3(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(7<<4)+3);
}
#define init_10_9_7_4
void FASTCALL call_10_9_7_4(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(7<<4)+4);
}
#define init_10_9_7_5
void FASTCALL call_10_9_7_5(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(7<<4)+5);
}
#define init_10_9_7_6
void FASTCALL call_10_9_7_6(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(7<<4)+6);
}
#define init_10_9_7_7
void FASTCALL call_10_9_7_7(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(7<<4)+7);
}
#define init_10_9_7_8
void FASTCALL call_10_9_7_8(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(7<<4)+8);
}
#define init_10_9_7_9
void FASTCALL call_10_9_7_9(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(7<<4)+9);
}
#define init_10_9_7_10
void FASTCALL call_10_9_7_10(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(7<<4)+10);
}
#define init_10_9_7_11
void FASTCALL call_10_9_7_11(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(7<<4)+11);
}
#define init_10_9_7_12
void FASTCALL call_10_9_7_12(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(7<<4)+12);
}
#define init_10_9_7_13
void FASTCALL call_10_9_7_13(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(7<<4)+13);
}
#define init_10_9_7_14
void FASTCALL call_10_9_7_14(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(7<<4)+14);
}
#define init_10_9_7_15
void FASTCALL call_10_9_7_15(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(7<<4)+15);
}
#define init_10_9_8_0
void FASTCALL call_10_9_8_0(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(8<<4)+0);
}
#define init_10_9_8_1
void FASTCALL call_10_9_8_1(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(8<<4)+1);
}
#define init_10_9_8_2
void FASTCALL call_10_9_8_2(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(8<<4)+2);
}
#define init_10_9_8_3
void FASTCALL call_10_9_8_3(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(8<<4)+3);
}
#define init_10_9_8_4
void FASTCALL call_10_9_8_4(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(8<<4)+4);
}
#define init_10_9_8_5
void FASTCALL call_10_9_8_5(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(8<<4)+5);
}
#define init_10_9_8_6
void FASTCALL call_10_9_8_6(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(8<<4)+6);
}
#define init_10_9_8_7
void FASTCALL call_10_9_8_7(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(8<<4)+7);
}
#define init_10_9_8_8
void FASTCALL call_10_9_8_8(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(8<<4)+8);
}
#define init_10_9_8_9
void FASTCALL call_10_9_8_9(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(8<<4)+9);
}
#define init_10_9_8_10
void FASTCALL call_10_9_8_10(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(8<<4)+10);
}
#define init_10_9_8_11
void FASTCALL call_10_9_8_11(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(8<<4)+11);
}
#define init_10_9_8_12
void FASTCALL call_10_9_8_12(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(8<<4)+12);
}
#define init_10_9_8_13
void FASTCALL call_10_9_8_13(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(8<<4)+13);
}
#define init_10_9_8_14
void FASTCALL call_10_9_8_14(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(8<<4)+14);
}
#define init_10_9_8_15
void FASTCALL call_10_9_8_15(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(8<<4)+15);
}
#define init_10_9_9_0
void FASTCALL call_10_9_9_0(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(9<<4)+0);
}
#define init_10_9_9_1
void FASTCALL call_10_9_9_1(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(9<<4)+1);
}
#define init_10_9_9_2
void FASTCALL call_10_9_9_2(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(9<<4)+2);
}
#define init_10_9_9_3
void FASTCALL call_10_9_9_3(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(9<<4)+3);
}
#define init_10_9_9_4
void FASTCALL call_10_9_9_4(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(9<<4)+4);
}
#define init_10_9_9_5
void FASTCALL call_10_9_9_5(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(9<<4)+5);
}
#define init_10_9_9_6
void FASTCALL call_10_9_9_6(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(9<<4)+6);
}
#define init_10_9_9_7
void FASTCALL call_10_9_9_7(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(9<<4)+7);
}
#define init_10_9_9_8
void FASTCALL call_10_9_9_8(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(9<<4)+8);
}
#define init_10_9_9_9
void FASTCALL call_10_9_9_9(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(9<<4)+9);
}
#define init_10_9_9_10
void FASTCALL call_10_9_9_10(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(9<<4)+10);
}
#define init_10_9_9_11
void FASTCALL call_10_9_9_11(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(9<<4)+11);
}
#define init_10_9_9_12
void FASTCALL call_10_9_9_12(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(9<<4)+12);
}
#define init_10_9_9_13
void FASTCALL call_10_9_9_13(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(9<<4)+13);
}
#define init_10_9_9_14
void FASTCALL call_10_9_9_14(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(9<<4)+14);
}
#define init_10_9_9_15
void FASTCALL call_10_9_9_15(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(9<<4)+15);
}
#define init_10_9_10_0
void FASTCALL call_10_9_10_0(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(10<<4)+0);
}
#define init_10_9_10_1
void FASTCALL call_10_9_10_1(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(10<<4)+1);
}
#define init_10_9_10_2
void FASTCALL call_10_9_10_2(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(10<<4)+2);
}
#define init_10_9_10_3
void FASTCALL call_10_9_10_3(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(10<<4)+3);
}
#define init_10_9_10_4
void FASTCALL call_10_9_10_4(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(10<<4)+4);
}
#define init_10_9_10_5
void FASTCALL call_10_9_10_5(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(10<<4)+5);
}
#define init_10_9_10_6
void FASTCALL call_10_9_10_6(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(10<<4)+6);
}
#define init_10_9_10_7
void FASTCALL call_10_9_10_7(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(10<<4)+7);
}
#define init_10_9_10_8
void FASTCALL call_10_9_10_8(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(10<<4)+8);
}
#define init_10_9_10_9
void FASTCALL call_10_9_10_9(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(10<<4)+9);
}
#define init_10_9_10_10
void FASTCALL call_10_9_10_10(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(10<<4)+10);
}
#define init_10_9_10_11
void FASTCALL call_10_9_10_11(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(10<<4)+11);
}
#define init_10_9_10_12
void FASTCALL call_10_9_10_12(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(10<<4)+12);
}
#define init_10_9_10_13
void FASTCALL call_10_9_10_13(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(10<<4)+13);
}
#define init_10_9_10_14
void FASTCALL call_10_9_10_14(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(10<<4)+14);
}
#define init_10_9_10_15
void FASTCALL call_10_9_10_15(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(10<<4)+15);
}
#define init_10_9_11_0
void FASTCALL call_10_9_11_0(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(11<<4)+0);
}
#define init_10_9_11_1
void FASTCALL call_10_9_11_1(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(11<<4)+1);
}
#define init_10_9_11_2
void FASTCALL call_10_9_11_2(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(11<<4)+2);
}
#define init_10_9_11_3
void FASTCALL call_10_9_11_3(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(11<<4)+3);
}
#define init_10_9_11_4
void FASTCALL call_10_9_11_4(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(11<<4)+4);
}
#define init_10_9_11_5
void FASTCALL call_10_9_11_5(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(11<<4)+5);
}
#define init_10_9_11_6
void FASTCALL call_10_9_11_6(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(11<<4)+6);
}
#define init_10_9_11_7
void FASTCALL call_10_9_11_7(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(11<<4)+7);
}
#define init_10_9_11_8
void FASTCALL call_10_9_11_8(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(11<<4)+8);
}
#define init_10_9_11_9
void FASTCALL call_10_9_11_9(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(11<<4)+9);
}
#define init_10_9_11_10
void FASTCALL call_10_9_11_10(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(11<<4)+10);
}
#define init_10_9_11_11
void FASTCALL call_10_9_11_11(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(11<<4)+11);
}
#define init_10_9_11_12
void FASTCALL call_10_9_11_12(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(11<<4)+12);
}
#define init_10_9_11_13
void FASTCALL call_10_9_11_13(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(11<<4)+13);
}
#define init_10_9_11_14
void FASTCALL call_10_9_11_14(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(11<<4)+14);
}
#define init_10_9_11_15
void FASTCALL call_10_9_11_15(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(11<<4)+15);
}
#define init_10_9_12_0
void FASTCALL call_10_9_12_0(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(12<<4)+0);
}
#define init_10_9_12_1
void FASTCALL call_10_9_12_1(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(12<<4)+1);
}
#define init_10_9_12_2
void FASTCALL call_10_9_12_2(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(12<<4)+2);
}
#define init_10_9_12_3
void FASTCALL call_10_9_12_3(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(12<<4)+3);
}
#define init_10_9_12_4
void FASTCALL call_10_9_12_4(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(12<<4)+4);
}
#define init_10_9_12_5
void FASTCALL call_10_9_12_5(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(12<<4)+5);
}
#define init_10_9_12_6
void FASTCALL call_10_9_12_6(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(12<<4)+6);
}
#define init_10_9_12_7
void FASTCALL call_10_9_12_7(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(12<<4)+7);
}
#define init_10_9_12_8
void FASTCALL call_10_9_12_8(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(12<<4)+8);
}
#define init_10_9_12_9
void FASTCALL call_10_9_12_9(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(12<<4)+9);
}
#define init_10_9_12_10
void FASTCALL call_10_9_12_10(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(12<<4)+10);
}
#define init_10_9_12_11
void FASTCALL call_10_9_12_11(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(12<<4)+11);
}
#define init_10_9_12_12
void FASTCALL call_10_9_12_12(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(12<<4)+12);
}
#define init_10_9_12_13
void FASTCALL call_10_9_12_13(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(12<<4)+13);
}
#define init_10_9_12_14
void FASTCALL call_10_9_12_14(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(12<<4)+14);
}
#define init_10_9_12_15
void FASTCALL call_10_9_12_15(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(12<<4)+15);
}
#define init_10_9_13_0
void FASTCALL call_10_9_13_0(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(13<<4)+0);
}
#define init_10_9_13_1
void FASTCALL call_10_9_13_1(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(13<<4)+1);
}
#define init_10_9_13_2
void FASTCALL call_10_9_13_2(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(13<<4)+2);
}
#define init_10_9_13_3
void FASTCALL call_10_9_13_3(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(13<<4)+3);
}
#define init_10_9_13_4
void FASTCALL call_10_9_13_4(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(13<<4)+4);
}
#define init_10_9_13_5
void FASTCALL call_10_9_13_5(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(13<<4)+5);
}
#define init_10_9_13_6
void FASTCALL call_10_9_13_6(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(13<<4)+6);
}
#define init_10_9_13_7
void FASTCALL call_10_9_13_7(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(13<<4)+7);
}
#define init_10_9_13_8
void FASTCALL call_10_9_13_8(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(13<<4)+8);
}
#define init_10_9_13_9
void FASTCALL call_10_9_13_9(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(13<<4)+9);
}
#define init_10_9_13_10
void FASTCALL call_10_9_13_10(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(13<<4)+10);
}
#define init_10_9_13_11
void FASTCALL call_10_9_13_11(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(13<<4)+11);
}
#define init_10_9_13_12
void FASTCALL call_10_9_13_12(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(13<<4)+12);
}
#define init_10_9_13_13
void FASTCALL call_10_9_13_13(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(13<<4)+13);
}
#define init_10_9_13_14
void FASTCALL call_10_9_13_14(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(13<<4)+14);
}
#define init_10_9_13_15
void FASTCALL call_10_9_13_15(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(13<<4)+15);
}
#define init_10_9_14_0
void FASTCALL call_10_9_14_0(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(14<<4)+0);
}
#define init_10_9_14_1
void FASTCALL call_10_9_14_1(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(14<<4)+1);
}
#define init_10_9_14_2
void FASTCALL call_10_9_14_2(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(14<<4)+2);
}
#define init_10_9_14_3
void FASTCALL call_10_9_14_3(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(14<<4)+3);
}
#define init_10_9_14_4
void FASTCALL call_10_9_14_4(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(14<<4)+4);
}
#define init_10_9_14_5
void FASTCALL call_10_9_14_5(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(14<<4)+5);
}
#define init_10_9_14_6
void FASTCALL call_10_9_14_6(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(14<<4)+6);
}
#define init_10_9_14_7
void FASTCALL call_10_9_14_7(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(14<<4)+7);
}
#define init_10_9_14_8
void FASTCALL call_10_9_14_8(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(14<<4)+8);
}
#define init_10_9_14_9
void FASTCALL call_10_9_14_9(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(14<<4)+9);
}
#define init_10_9_14_10
void FASTCALL call_10_9_14_10(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(14<<4)+10);
}
#define init_10_9_14_11
void FASTCALL call_10_9_14_11(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(14<<4)+11);
}
#define init_10_9_14_12
void FASTCALL call_10_9_14_12(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(14<<4)+12);
}
#define init_10_9_14_13
void FASTCALL call_10_9_14_13(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(14<<4)+13);
}
#define init_10_9_14_14
void FASTCALL call_10_9_14_14(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(14<<4)+14);
}
#define init_10_9_14_15
void FASTCALL call_10_9_14_15(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(14<<4)+15);
}
#define init_10_9_15_0
void FASTCALL call_10_9_15_0(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(15<<4)+0);
}
#define init_10_9_15_1
void FASTCALL call_10_9_15_1(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(15<<4)+1);
}
#define init_10_9_15_2
void FASTCALL call_10_9_15_2(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(15<<4)+2);
}
#define init_10_9_15_3
void FASTCALL call_10_9_15_3(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(15<<4)+3);
}
#define init_10_9_15_4
void FASTCALL call_10_9_15_4(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(15<<4)+4);
}
#define init_10_9_15_5
void FASTCALL call_10_9_15_5(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(15<<4)+5);
}
#define init_10_9_15_6
void FASTCALL call_10_9_15_6(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(15<<4)+6);
}
#define init_10_9_15_7
void FASTCALL call_10_9_15_7(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(15<<4)+7);
}
#define init_10_9_15_8
void FASTCALL call_10_9_15_8(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(15<<4)+8);
}
#define init_10_9_15_9
void FASTCALL call_10_9_15_9(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(15<<4)+9);
}
#define init_10_9_15_10
void FASTCALL call_10_9_15_10(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(15<<4)+10);
}
#define init_10_9_15_11
void FASTCALL call_10_9_15_11(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(15<<4)+11);
}
#define init_10_9_15_12
void FASTCALL call_10_9_15_12(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(15<<4)+12);
}
#define init_10_9_15_13
void FASTCALL call_10_9_15_13(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(15<<4)+13);
}
#define init_10_9_15_14
void FASTCALL call_10_9_15_14(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(15<<4)+14);
}
#define init_10_9_15_15
void FASTCALL call_10_9_15_15(SH2_struct* sh)
{
  SH2bra(sh, (9<<8)+(15<<4)+15);
}
#define init_10_10_0_0
void FASTCALL call_10_10_0_0(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(0<<4)+0);
}
#define init_10_10_0_1
void FASTCALL call_10_10_0_1(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(0<<4)+1);
}
#define init_10_10_0_2
void FASTCALL call_10_10_0_2(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(0<<4)+2);
}
#define init_10_10_0_3
void FASTCALL call_10_10_0_3(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(0<<4)+3);
}
#define init_10_10_0_4
void FASTCALL call_10_10_0_4(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(0<<4)+4);
}
#define init_10_10_0_5
void FASTCALL call_10_10_0_5(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(0<<4)+5);
}
#define init_10_10_0_6
void FASTCALL call_10_10_0_6(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(0<<4)+6);
}
#define init_10_10_0_7
void FASTCALL call_10_10_0_7(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(0<<4)+7);
}
#define init_10_10_0_8
void FASTCALL call_10_10_0_8(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(0<<4)+8);
}
#define init_10_10_0_9
void FASTCALL call_10_10_0_9(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(0<<4)+9);
}
#define init_10_10_0_10
void FASTCALL call_10_10_0_10(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(0<<4)+10);
}
#define init_10_10_0_11
void FASTCALL call_10_10_0_11(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(0<<4)+11);
}
#define init_10_10_0_12
void FASTCALL call_10_10_0_12(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(0<<4)+12);
}
#define init_10_10_0_13
void FASTCALL call_10_10_0_13(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(0<<4)+13);
}
#define init_10_10_0_14
void FASTCALL call_10_10_0_14(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(0<<4)+14);
}
#define init_10_10_0_15
void FASTCALL call_10_10_0_15(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(0<<4)+15);
}
#define init_10_10_1_0
void FASTCALL call_10_10_1_0(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(1<<4)+0);
}
#define init_10_10_1_1
void FASTCALL call_10_10_1_1(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(1<<4)+1);
}
#define init_10_10_1_2
void FASTCALL call_10_10_1_2(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(1<<4)+2);
}
#define init_10_10_1_3
void FASTCALL call_10_10_1_3(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(1<<4)+3);
}
#define init_10_10_1_4
void FASTCALL call_10_10_1_4(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(1<<4)+4);
}
#define init_10_10_1_5
void FASTCALL call_10_10_1_5(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(1<<4)+5);
}
#define init_10_10_1_6
void FASTCALL call_10_10_1_6(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(1<<4)+6);
}
#define init_10_10_1_7
void FASTCALL call_10_10_1_7(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(1<<4)+7);
}
#define init_10_10_1_8
void FASTCALL call_10_10_1_8(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(1<<4)+8);
}
#define init_10_10_1_9
void FASTCALL call_10_10_1_9(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(1<<4)+9);
}
#define init_10_10_1_10
void FASTCALL call_10_10_1_10(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(1<<4)+10);
}
#define init_10_10_1_11
void FASTCALL call_10_10_1_11(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(1<<4)+11);
}
#define init_10_10_1_12
void FASTCALL call_10_10_1_12(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(1<<4)+12);
}
#define init_10_10_1_13
void FASTCALL call_10_10_1_13(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(1<<4)+13);
}
#define init_10_10_1_14
void FASTCALL call_10_10_1_14(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(1<<4)+14);
}
#define init_10_10_1_15
void FASTCALL call_10_10_1_15(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(1<<4)+15);
}
#define init_10_10_2_0
void FASTCALL call_10_10_2_0(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(2<<4)+0);
}
#define init_10_10_2_1
void FASTCALL call_10_10_2_1(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(2<<4)+1);
}
#define init_10_10_2_2
void FASTCALL call_10_10_2_2(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(2<<4)+2);
}
#define init_10_10_2_3
void FASTCALL call_10_10_2_3(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(2<<4)+3);
}
#define init_10_10_2_4
void FASTCALL call_10_10_2_4(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(2<<4)+4);
}
#define init_10_10_2_5
void FASTCALL call_10_10_2_5(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(2<<4)+5);
}
#define init_10_10_2_6
void FASTCALL call_10_10_2_6(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(2<<4)+6);
}
#define init_10_10_2_7
void FASTCALL call_10_10_2_7(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(2<<4)+7);
}
#define init_10_10_2_8
void FASTCALL call_10_10_2_8(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(2<<4)+8);
}
#define init_10_10_2_9
void FASTCALL call_10_10_2_9(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(2<<4)+9);
}
#define init_10_10_2_10
void FASTCALL call_10_10_2_10(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(2<<4)+10);
}
#define init_10_10_2_11
void FASTCALL call_10_10_2_11(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(2<<4)+11);
}
#define init_10_10_2_12
void FASTCALL call_10_10_2_12(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(2<<4)+12);
}
#define init_10_10_2_13
void FASTCALL call_10_10_2_13(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(2<<4)+13);
}
#define init_10_10_2_14
void FASTCALL call_10_10_2_14(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(2<<4)+14);
}
#define init_10_10_2_15
void FASTCALL call_10_10_2_15(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(2<<4)+15);
}
#define init_10_10_3_0
void FASTCALL call_10_10_3_0(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(3<<4)+0);
}
#define init_10_10_3_1
void FASTCALL call_10_10_3_1(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(3<<4)+1);
}
#define init_10_10_3_2
void FASTCALL call_10_10_3_2(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(3<<4)+2);
}
#define init_10_10_3_3
void FASTCALL call_10_10_3_3(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(3<<4)+3);
}
#define init_10_10_3_4
void FASTCALL call_10_10_3_4(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(3<<4)+4);
}
#define init_10_10_3_5
void FASTCALL call_10_10_3_5(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(3<<4)+5);
}
#define init_10_10_3_6
void FASTCALL call_10_10_3_6(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(3<<4)+6);
}
#define init_10_10_3_7
void FASTCALL call_10_10_3_7(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(3<<4)+7);
}
#define init_10_10_3_8
void FASTCALL call_10_10_3_8(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(3<<4)+8);
}
#define init_10_10_3_9
void FASTCALL call_10_10_3_9(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(3<<4)+9);
}
#define init_10_10_3_10
void FASTCALL call_10_10_3_10(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(3<<4)+10);
}
#define init_10_10_3_11
void FASTCALL call_10_10_3_11(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(3<<4)+11);
}
#define init_10_10_3_12
void FASTCALL call_10_10_3_12(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(3<<4)+12);
}
#define init_10_10_3_13
void FASTCALL call_10_10_3_13(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(3<<4)+13);
}
#define init_10_10_3_14
void FASTCALL call_10_10_3_14(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(3<<4)+14);
}
#define init_10_10_3_15
void FASTCALL call_10_10_3_15(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(3<<4)+15);
}
#define init_10_10_4_0
void FASTCALL call_10_10_4_0(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(4<<4)+0);
}
#define init_10_10_4_1
void FASTCALL call_10_10_4_1(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(4<<4)+1);
}
#define init_10_10_4_2
void FASTCALL call_10_10_4_2(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(4<<4)+2);
}
#define init_10_10_4_3
void FASTCALL call_10_10_4_3(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(4<<4)+3);
}
#define init_10_10_4_4
void FASTCALL call_10_10_4_4(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(4<<4)+4);
}
#define init_10_10_4_5
void FASTCALL call_10_10_4_5(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(4<<4)+5);
}
#define init_10_10_4_6
void FASTCALL call_10_10_4_6(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(4<<4)+6);
}
#define init_10_10_4_7
void FASTCALL call_10_10_4_7(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(4<<4)+7);
}
#define init_10_10_4_8
void FASTCALL call_10_10_4_8(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(4<<4)+8);
}
#define init_10_10_4_9
void FASTCALL call_10_10_4_9(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(4<<4)+9);
}
#define init_10_10_4_10
void FASTCALL call_10_10_4_10(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(4<<4)+10);
}
#define init_10_10_4_11
void FASTCALL call_10_10_4_11(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(4<<4)+11);
}
#define init_10_10_4_12
void FASTCALL call_10_10_4_12(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(4<<4)+12);
}
#define init_10_10_4_13
void FASTCALL call_10_10_4_13(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(4<<4)+13);
}
#define init_10_10_4_14
void FASTCALL call_10_10_4_14(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(4<<4)+14);
}
#define init_10_10_4_15
void FASTCALL call_10_10_4_15(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(4<<4)+15);
}
#define init_10_10_5_0
void FASTCALL call_10_10_5_0(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(5<<4)+0);
}
#define init_10_10_5_1
void FASTCALL call_10_10_5_1(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(5<<4)+1);
}
#define init_10_10_5_2
void FASTCALL call_10_10_5_2(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(5<<4)+2);
}
#define init_10_10_5_3
void FASTCALL call_10_10_5_3(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(5<<4)+3);
}
#define init_10_10_5_4
void FASTCALL call_10_10_5_4(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(5<<4)+4);
}
#define init_10_10_5_5
void FASTCALL call_10_10_5_5(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(5<<4)+5);
}
#define init_10_10_5_6
void FASTCALL call_10_10_5_6(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(5<<4)+6);
}
#define init_10_10_5_7
void FASTCALL call_10_10_5_7(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(5<<4)+7);
}
#define init_10_10_5_8
void FASTCALL call_10_10_5_8(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(5<<4)+8);
}
#define init_10_10_5_9
void FASTCALL call_10_10_5_9(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(5<<4)+9);
}
#define init_10_10_5_10
void FASTCALL call_10_10_5_10(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(5<<4)+10);
}
#define init_10_10_5_11
void FASTCALL call_10_10_5_11(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(5<<4)+11);
}
#define init_10_10_5_12
void FASTCALL call_10_10_5_12(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(5<<4)+12);
}
#define init_10_10_5_13
void FASTCALL call_10_10_5_13(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(5<<4)+13);
}
#define init_10_10_5_14
void FASTCALL call_10_10_5_14(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(5<<4)+14);
}
#define init_10_10_5_15
void FASTCALL call_10_10_5_15(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(5<<4)+15);
}
#define init_10_10_6_0
void FASTCALL call_10_10_6_0(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(6<<4)+0);
}
#define init_10_10_6_1
void FASTCALL call_10_10_6_1(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(6<<4)+1);
}
#define init_10_10_6_2
void FASTCALL call_10_10_6_2(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(6<<4)+2);
}
#define init_10_10_6_3
void FASTCALL call_10_10_6_3(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(6<<4)+3);
}
#define init_10_10_6_4
void FASTCALL call_10_10_6_4(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(6<<4)+4);
}
#define init_10_10_6_5
void FASTCALL call_10_10_6_5(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(6<<4)+5);
}
#define init_10_10_6_6
void FASTCALL call_10_10_6_6(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(6<<4)+6);
}
#define init_10_10_6_7
void FASTCALL call_10_10_6_7(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(6<<4)+7);
}
#define init_10_10_6_8
void FASTCALL call_10_10_6_8(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(6<<4)+8);
}
#define init_10_10_6_9
void FASTCALL call_10_10_6_9(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(6<<4)+9);
}
#define init_10_10_6_10
void FASTCALL call_10_10_6_10(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(6<<4)+10);
}
#define init_10_10_6_11
void FASTCALL call_10_10_6_11(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(6<<4)+11);
}
#define init_10_10_6_12
void FASTCALL call_10_10_6_12(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(6<<4)+12);
}
#define init_10_10_6_13
void FASTCALL call_10_10_6_13(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(6<<4)+13);
}
#define init_10_10_6_14
void FASTCALL call_10_10_6_14(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(6<<4)+14);
}
#define init_10_10_6_15
void FASTCALL call_10_10_6_15(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(6<<4)+15);
}
#define init_10_10_7_0
void FASTCALL call_10_10_7_0(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(7<<4)+0);
}
#define init_10_10_7_1
void FASTCALL call_10_10_7_1(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(7<<4)+1);
}
#define init_10_10_7_2
void FASTCALL call_10_10_7_2(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(7<<4)+2);
}
#define init_10_10_7_3
void FASTCALL call_10_10_7_3(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(7<<4)+3);
}
#define init_10_10_7_4
void FASTCALL call_10_10_7_4(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(7<<4)+4);
}
#define init_10_10_7_5
void FASTCALL call_10_10_7_5(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(7<<4)+5);
}
#define init_10_10_7_6
void FASTCALL call_10_10_7_6(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(7<<4)+6);
}
#define init_10_10_7_7
void FASTCALL call_10_10_7_7(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(7<<4)+7);
}
#define init_10_10_7_8
void FASTCALL call_10_10_7_8(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(7<<4)+8);
}
#define init_10_10_7_9
void FASTCALL call_10_10_7_9(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(7<<4)+9);
}
#define init_10_10_7_10
void FASTCALL call_10_10_7_10(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(7<<4)+10);
}
#define init_10_10_7_11
void FASTCALL call_10_10_7_11(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(7<<4)+11);
}
#define init_10_10_7_12
void FASTCALL call_10_10_7_12(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(7<<4)+12);
}
#define init_10_10_7_13
void FASTCALL call_10_10_7_13(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(7<<4)+13);
}
#define init_10_10_7_14
void FASTCALL call_10_10_7_14(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(7<<4)+14);
}
#define init_10_10_7_15
void FASTCALL call_10_10_7_15(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(7<<4)+15);
}
#define init_10_10_8_0
void FASTCALL call_10_10_8_0(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(8<<4)+0);
}
#define init_10_10_8_1
void FASTCALL call_10_10_8_1(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(8<<4)+1);
}
#define init_10_10_8_2
void FASTCALL call_10_10_8_2(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(8<<4)+2);
}
#define init_10_10_8_3
void FASTCALL call_10_10_8_3(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(8<<4)+3);
}
#define init_10_10_8_4
void FASTCALL call_10_10_8_4(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(8<<4)+4);
}
#define init_10_10_8_5
void FASTCALL call_10_10_8_5(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(8<<4)+5);
}
#define init_10_10_8_6
void FASTCALL call_10_10_8_6(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(8<<4)+6);
}
#define init_10_10_8_7
void FASTCALL call_10_10_8_7(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(8<<4)+7);
}
#define init_10_10_8_8
void FASTCALL call_10_10_8_8(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(8<<4)+8);
}
#define init_10_10_8_9
void FASTCALL call_10_10_8_9(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(8<<4)+9);
}
#define init_10_10_8_10
void FASTCALL call_10_10_8_10(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(8<<4)+10);
}
#define init_10_10_8_11
void FASTCALL call_10_10_8_11(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(8<<4)+11);
}
#define init_10_10_8_12
void FASTCALL call_10_10_8_12(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(8<<4)+12);
}
#define init_10_10_8_13
void FASTCALL call_10_10_8_13(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(8<<4)+13);
}
#define init_10_10_8_14
void FASTCALL call_10_10_8_14(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(8<<4)+14);
}
#define init_10_10_8_15
void FASTCALL call_10_10_8_15(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(8<<4)+15);
}
#define init_10_10_9_0
void FASTCALL call_10_10_9_0(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(9<<4)+0);
}
#define init_10_10_9_1
void FASTCALL call_10_10_9_1(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(9<<4)+1);
}
#define init_10_10_9_2
void FASTCALL call_10_10_9_2(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(9<<4)+2);
}
#define init_10_10_9_3
void FASTCALL call_10_10_9_3(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(9<<4)+3);
}
#define init_10_10_9_4
void FASTCALL call_10_10_9_4(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(9<<4)+4);
}
#define init_10_10_9_5
void FASTCALL call_10_10_9_5(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(9<<4)+5);
}
#define init_10_10_9_6
void FASTCALL call_10_10_9_6(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(9<<4)+6);
}
#define init_10_10_9_7
void FASTCALL call_10_10_9_7(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(9<<4)+7);
}
#define init_10_10_9_8
void FASTCALL call_10_10_9_8(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(9<<4)+8);
}
#define init_10_10_9_9
void FASTCALL call_10_10_9_9(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(9<<4)+9);
}
#define init_10_10_9_10
void FASTCALL call_10_10_9_10(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(9<<4)+10);
}
#define init_10_10_9_11
void FASTCALL call_10_10_9_11(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(9<<4)+11);
}
#define init_10_10_9_12
void FASTCALL call_10_10_9_12(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(9<<4)+12);
}
#define init_10_10_9_13
void FASTCALL call_10_10_9_13(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(9<<4)+13);
}
#define init_10_10_9_14
void FASTCALL call_10_10_9_14(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(9<<4)+14);
}
#define init_10_10_9_15
void FASTCALL call_10_10_9_15(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(9<<4)+15);
}
#define init_10_10_10_0
void FASTCALL call_10_10_10_0(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(10<<4)+0);
}
#define init_10_10_10_1
void FASTCALL call_10_10_10_1(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(10<<4)+1);
}
#define init_10_10_10_2
void FASTCALL call_10_10_10_2(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(10<<4)+2);
}
#define init_10_10_10_3
void FASTCALL call_10_10_10_3(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(10<<4)+3);
}
#define init_10_10_10_4
void FASTCALL call_10_10_10_4(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(10<<4)+4);
}
#define init_10_10_10_5
void FASTCALL call_10_10_10_5(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(10<<4)+5);
}
#define init_10_10_10_6
void FASTCALL call_10_10_10_6(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(10<<4)+6);
}
#define init_10_10_10_7
void FASTCALL call_10_10_10_7(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(10<<4)+7);
}
#define init_10_10_10_8
void FASTCALL call_10_10_10_8(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(10<<4)+8);
}
#define init_10_10_10_9
void FASTCALL call_10_10_10_9(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(10<<4)+9);
}
#define init_10_10_10_10
void FASTCALL call_10_10_10_10(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(10<<4)+10);
}
#define init_10_10_10_11
void FASTCALL call_10_10_10_11(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(10<<4)+11);
}
#define init_10_10_10_12
void FASTCALL call_10_10_10_12(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(10<<4)+12);
}
#define init_10_10_10_13
void FASTCALL call_10_10_10_13(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(10<<4)+13);
}
#define init_10_10_10_14
void FASTCALL call_10_10_10_14(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(10<<4)+14);
}
#define init_10_10_10_15
void FASTCALL call_10_10_10_15(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(10<<4)+15);
}
#define init_10_10_11_0
void FASTCALL call_10_10_11_0(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(11<<4)+0);
}
#define init_10_10_11_1
void FASTCALL call_10_10_11_1(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(11<<4)+1);
}
#define init_10_10_11_2
void FASTCALL call_10_10_11_2(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(11<<4)+2);
}
#define init_10_10_11_3
void FASTCALL call_10_10_11_3(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(11<<4)+3);
}
#define init_10_10_11_4
void FASTCALL call_10_10_11_4(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(11<<4)+4);
}
#define init_10_10_11_5
void FASTCALL call_10_10_11_5(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(11<<4)+5);
}
#define init_10_10_11_6
void FASTCALL call_10_10_11_6(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(11<<4)+6);
}
#define init_10_10_11_7
void FASTCALL call_10_10_11_7(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(11<<4)+7);
}
#define init_10_10_11_8
void FASTCALL call_10_10_11_8(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(11<<4)+8);
}
#define init_10_10_11_9
void FASTCALL call_10_10_11_9(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(11<<4)+9);
}
#define init_10_10_11_10
void FASTCALL call_10_10_11_10(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(11<<4)+10);
}
#define init_10_10_11_11
void FASTCALL call_10_10_11_11(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(11<<4)+11);
}
#define init_10_10_11_12
void FASTCALL call_10_10_11_12(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(11<<4)+12);
}
#define init_10_10_11_13
void FASTCALL call_10_10_11_13(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(11<<4)+13);
}
#define init_10_10_11_14
void FASTCALL call_10_10_11_14(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(11<<4)+14);
}
#define init_10_10_11_15
void FASTCALL call_10_10_11_15(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(11<<4)+15);
}
#define init_10_10_12_0
void FASTCALL call_10_10_12_0(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(12<<4)+0);
}
#define init_10_10_12_1
void FASTCALL call_10_10_12_1(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(12<<4)+1);
}
#define init_10_10_12_2
void FASTCALL call_10_10_12_2(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(12<<4)+2);
}
#define init_10_10_12_3
void FASTCALL call_10_10_12_3(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(12<<4)+3);
}
#define init_10_10_12_4
void FASTCALL call_10_10_12_4(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(12<<4)+4);
}
#define init_10_10_12_5
void FASTCALL call_10_10_12_5(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(12<<4)+5);
}
#define init_10_10_12_6
void FASTCALL call_10_10_12_6(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(12<<4)+6);
}
#define init_10_10_12_7
void FASTCALL call_10_10_12_7(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(12<<4)+7);
}
#define init_10_10_12_8
void FASTCALL call_10_10_12_8(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(12<<4)+8);
}
#define init_10_10_12_9
void FASTCALL call_10_10_12_9(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(12<<4)+9);
}
#define init_10_10_12_10
void FASTCALL call_10_10_12_10(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(12<<4)+10);
}
#define init_10_10_12_11
void FASTCALL call_10_10_12_11(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(12<<4)+11);
}
#define init_10_10_12_12
void FASTCALL call_10_10_12_12(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(12<<4)+12);
}
#define init_10_10_12_13
void FASTCALL call_10_10_12_13(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(12<<4)+13);
}
#define init_10_10_12_14
void FASTCALL call_10_10_12_14(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(12<<4)+14);
}
#define init_10_10_12_15
void FASTCALL call_10_10_12_15(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(12<<4)+15);
}
#define init_10_10_13_0
void FASTCALL call_10_10_13_0(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(13<<4)+0);
}
#define init_10_10_13_1
void FASTCALL call_10_10_13_1(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(13<<4)+1);
}
#define init_10_10_13_2
void FASTCALL call_10_10_13_2(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(13<<4)+2);
}
#define init_10_10_13_3
void FASTCALL call_10_10_13_3(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(13<<4)+3);
}
#define init_10_10_13_4
void FASTCALL call_10_10_13_4(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(13<<4)+4);
}
#define init_10_10_13_5
void FASTCALL call_10_10_13_5(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(13<<4)+5);
}
#define init_10_10_13_6
void FASTCALL call_10_10_13_6(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(13<<4)+6);
}
#define init_10_10_13_7
void FASTCALL call_10_10_13_7(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(13<<4)+7);
}
#define init_10_10_13_8
void FASTCALL call_10_10_13_8(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(13<<4)+8);
}
#define init_10_10_13_9
void FASTCALL call_10_10_13_9(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(13<<4)+9);
}
#define init_10_10_13_10
void FASTCALL call_10_10_13_10(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(13<<4)+10);
}
#define init_10_10_13_11
void FASTCALL call_10_10_13_11(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(13<<4)+11);
}
#define init_10_10_13_12
void FASTCALL call_10_10_13_12(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(13<<4)+12);
}
#define init_10_10_13_13
void FASTCALL call_10_10_13_13(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(13<<4)+13);
}
#define init_10_10_13_14
void FASTCALL call_10_10_13_14(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(13<<4)+14);
}
#define init_10_10_13_15
void FASTCALL call_10_10_13_15(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(13<<4)+15);
}
#define init_10_10_14_0
void FASTCALL call_10_10_14_0(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(14<<4)+0);
}
#define init_10_10_14_1
void FASTCALL call_10_10_14_1(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(14<<4)+1);
}
#define init_10_10_14_2
void FASTCALL call_10_10_14_2(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(14<<4)+2);
}
#define init_10_10_14_3
void FASTCALL call_10_10_14_3(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(14<<4)+3);
}
#define init_10_10_14_4
void FASTCALL call_10_10_14_4(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(14<<4)+4);
}
#define init_10_10_14_5
void FASTCALL call_10_10_14_5(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(14<<4)+5);
}
#define init_10_10_14_6
void FASTCALL call_10_10_14_6(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(14<<4)+6);
}
#define init_10_10_14_7
void FASTCALL call_10_10_14_7(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(14<<4)+7);
}
#define init_10_10_14_8
void FASTCALL call_10_10_14_8(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(14<<4)+8);
}
#define init_10_10_14_9
void FASTCALL call_10_10_14_9(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(14<<4)+9);
}
#define init_10_10_14_10
void FASTCALL call_10_10_14_10(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(14<<4)+10);
}
#define init_10_10_14_11
void FASTCALL call_10_10_14_11(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(14<<4)+11);
}
#define init_10_10_14_12
void FASTCALL call_10_10_14_12(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(14<<4)+12);
}
#define init_10_10_14_13
void FASTCALL call_10_10_14_13(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(14<<4)+13);
}
#define init_10_10_14_14
void FASTCALL call_10_10_14_14(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(14<<4)+14);
}
#define init_10_10_14_15
void FASTCALL call_10_10_14_15(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(14<<4)+15);
}
#define init_10_10_15_0
void FASTCALL call_10_10_15_0(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(15<<4)+0);
}
#define init_10_10_15_1
void FASTCALL call_10_10_15_1(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(15<<4)+1);
}
#define init_10_10_15_2
void FASTCALL call_10_10_15_2(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(15<<4)+2);
}
#define init_10_10_15_3
void FASTCALL call_10_10_15_3(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(15<<4)+3);
}
#define init_10_10_15_4
void FASTCALL call_10_10_15_4(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(15<<4)+4);
}
#define init_10_10_15_5
void FASTCALL call_10_10_15_5(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(15<<4)+5);
}
#define init_10_10_15_6
void FASTCALL call_10_10_15_6(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(15<<4)+6);
}
#define init_10_10_15_7
void FASTCALL call_10_10_15_7(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(15<<4)+7);
}
#define init_10_10_15_8
void FASTCALL call_10_10_15_8(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(15<<4)+8);
}
#define init_10_10_15_9
void FASTCALL call_10_10_15_9(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(15<<4)+9);
}
#define init_10_10_15_10
void FASTCALL call_10_10_15_10(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(15<<4)+10);
}
#define init_10_10_15_11
void FASTCALL call_10_10_15_11(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(15<<4)+11);
}
#define init_10_10_15_12
void FASTCALL call_10_10_15_12(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(15<<4)+12);
}
#define init_10_10_15_13
void FASTCALL call_10_10_15_13(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(15<<4)+13);
}
#define init_10_10_15_14
void FASTCALL call_10_10_15_14(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(15<<4)+14);
}
#define init_10_10_15_15
void FASTCALL call_10_10_15_15(SH2_struct* sh)
{
  SH2bra(sh, (10<<8)+(15<<4)+15);
}
#define init_10_11_0_0
void FASTCALL call_10_11_0_0(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(0<<4)+0);
}
#define init_10_11_0_1
void FASTCALL call_10_11_0_1(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(0<<4)+1);
}
#define init_10_11_0_2
void FASTCALL call_10_11_0_2(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(0<<4)+2);
}
#define init_10_11_0_3
void FASTCALL call_10_11_0_3(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(0<<4)+3);
}
#define init_10_11_0_4
void FASTCALL call_10_11_0_4(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(0<<4)+4);
}
#define init_10_11_0_5
void FASTCALL call_10_11_0_5(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(0<<4)+5);
}
#define init_10_11_0_6
void FASTCALL call_10_11_0_6(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(0<<4)+6);
}
#define init_10_11_0_7
void FASTCALL call_10_11_0_7(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(0<<4)+7);
}
#define init_10_11_0_8
void FASTCALL call_10_11_0_8(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(0<<4)+8);
}
#define init_10_11_0_9
void FASTCALL call_10_11_0_9(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(0<<4)+9);
}
#define init_10_11_0_10
void FASTCALL call_10_11_0_10(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(0<<4)+10);
}
#define init_10_11_0_11
void FASTCALL call_10_11_0_11(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(0<<4)+11);
}
#define init_10_11_0_12
void FASTCALL call_10_11_0_12(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(0<<4)+12);
}
#define init_10_11_0_13
void FASTCALL call_10_11_0_13(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(0<<4)+13);
}
#define init_10_11_0_14
void FASTCALL call_10_11_0_14(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(0<<4)+14);
}
#define init_10_11_0_15
void FASTCALL call_10_11_0_15(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(0<<4)+15);
}
#define init_10_11_1_0
void FASTCALL call_10_11_1_0(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(1<<4)+0);
}
#define init_10_11_1_1
void FASTCALL call_10_11_1_1(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(1<<4)+1);
}
#define init_10_11_1_2
void FASTCALL call_10_11_1_2(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(1<<4)+2);
}
#define init_10_11_1_3
void FASTCALL call_10_11_1_3(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(1<<4)+3);
}
#define init_10_11_1_4
void FASTCALL call_10_11_1_4(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(1<<4)+4);
}
#define init_10_11_1_5
void FASTCALL call_10_11_1_5(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(1<<4)+5);
}
#define init_10_11_1_6
void FASTCALL call_10_11_1_6(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(1<<4)+6);
}
#define init_10_11_1_7
void FASTCALL call_10_11_1_7(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(1<<4)+7);
}
#define init_10_11_1_8
void FASTCALL call_10_11_1_8(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(1<<4)+8);
}
#define init_10_11_1_9
void FASTCALL call_10_11_1_9(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(1<<4)+9);
}
#define init_10_11_1_10
void FASTCALL call_10_11_1_10(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(1<<4)+10);
}
#define init_10_11_1_11
void FASTCALL call_10_11_1_11(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(1<<4)+11);
}
#define init_10_11_1_12
void FASTCALL call_10_11_1_12(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(1<<4)+12);
}
#define init_10_11_1_13
void FASTCALL call_10_11_1_13(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(1<<4)+13);
}
#define init_10_11_1_14
void FASTCALL call_10_11_1_14(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(1<<4)+14);
}
#define init_10_11_1_15
void FASTCALL call_10_11_1_15(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(1<<4)+15);
}
#define init_10_11_2_0
void FASTCALL call_10_11_2_0(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(2<<4)+0);
}
#define init_10_11_2_1
void FASTCALL call_10_11_2_1(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(2<<4)+1);
}
#define init_10_11_2_2
void FASTCALL call_10_11_2_2(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(2<<4)+2);
}
#define init_10_11_2_3
void FASTCALL call_10_11_2_3(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(2<<4)+3);
}
#define init_10_11_2_4
void FASTCALL call_10_11_2_4(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(2<<4)+4);
}
#define init_10_11_2_5
void FASTCALL call_10_11_2_5(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(2<<4)+5);
}
#define init_10_11_2_6
void FASTCALL call_10_11_2_6(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(2<<4)+6);
}
#define init_10_11_2_7
void FASTCALL call_10_11_2_7(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(2<<4)+7);
}
#define init_10_11_2_8
void FASTCALL call_10_11_2_8(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(2<<4)+8);
}
#define init_10_11_2_9
void FASTCALL call_10_11_2_9(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(2<<4)+9);
}
#define init_10_11_2_10
void FASTCALL call_10_11_2_10(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(2<<4)+10);
}
#define init_10_11_2_11
void FASTCALL call_10_11_2_11(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(2<<4)+11);
}
#define init_10_11_2_12
void FASTCALL call_10_11_2_12(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(2<<4)+12);
}
#define init_10_11_2_13
void FASTCALL call_10_11_2_13(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(2<<4)+13);
}
#define init_10_11_2_14
void FASTCALL call_10_11_2_14(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(2<<4)+14);
}
#define init_10_11_2_15
void FASTCALL call_10_11_2_15(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(2<<4)+15);
}
#define init_10_11_3_0
void FASTCALL call_10_11_3_0(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(3<<4)+0);
}
#define init_10_11_3_1
void FASTCALL call_10_11_3_1(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(3<<4)+1);
}
#define init_10_11_3_2
void FASTCALL call_10_11_3_2(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(3<<4)+2);
}
#define init_10_11_3_3
void FASTCALL call_10_11_3_3(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(3<<4)+3);
}
#define init_10_11_3_4
void FASTCALL call_10_11_3_4(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(3<<4)+4);
}
#define init_10_11_3_5
void FASTCALL call_10_11_3_5(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(3<<4)+5);
}
#define init_10_11_3_6
void FASTCALL call_10_11_3_6(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(3<<4)+6);
}
#define init_10_11_3_7
void FASTCALL call_10_11_3_7(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(3<<4)+7);
}
#define init_10_11_3_8
void FASTCALL call_10_11_3_8(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(3<<4)+8);
}
#define init_10_11_3_9
void FASTCALL call_10_11_3_9(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(3<<4)+9);
}
#define init_10_11_3_10
void FASTCALL call_10_11_3_10(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(3<<4)+10);
}
#define init_10_11_3_11
void FASTCALL call_10_11_3_11(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(3<<4)+11);
}
#define init_10_11_3_12
void FASTCALL call_10_11_3_12(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(3<<4)+12);
}
#define init_10_11_3_13
void FASTCALL call_10_11_3_13(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(3<<4)+13);
}
#define init_10_11_3_14
void FASTCALL call_10_11_3_14(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(3<<4)+14);
}
#define init_10_11_3_15
void FASTCALL call_10_11_3_15(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(3<<4)+15);
}
#define init_10_11_4_0
void FASTCALL call_10_11_4_0(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(4<<4)+0);
}
#define init_10_11_4_1
void FASTCALL call_10_11_4_1(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(4<<4)+1);
}
#define init_10_11_4_2
void FASTCALL call_10_11_4_2(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(4<<4)+2);
}
#define init_10_11_4_3
void FASTCALL call_10_11_4_3(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(4<<4)+3);
}
#define init_10_11_4_4
void FASTCALL call_10_11_4_4(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(4<<4)+4);
}
#define init_10_11_4_5
void FASTCALL call_10_11_4_5(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(4<<4)+5);
}
#define init_10_11_4_6
void FASTCALL call_10_11_4_6(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(4<<4)+6);
}
#define init_10_11_4_7
void FASTCALL call_10_11_4_7(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(4<<4)+7);
}
#define init_10_11_4_8
void FASTCALL call_10_11_4_8(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(4<<4)+8);
}
#define init_10_11_4_9
void FASTCALL call_10_11_4_9(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(4<<4)+9);
}
#define init_10_11_4_10
void FASTCALL call_10_11_4_10(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(4<<4)+10);
}
#define init_10_11_4_11
void FASTCALL call_10_11_4_11(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(4<<4)+11);
}
#define init_10_11_4_12
void FASTCALL call_10_11_4_12(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(4<<4)+12);
}
#define init_10_11_4_13
void FASTCALL call_10_11_4_13(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(4<<4)+13);
}
#define init_10_11_4_14
void FASTCALL call_10_11_4_14(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(4<<4)+14);
}
#define init_10_11_4_15
void FASTCALL call_10_11_4_15(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(4<<4)+15);
}
#define init_10_11_5_0
void FASTCALL call_10_11_5_0(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(5<<4)+0);
}
#define init_10_11_5_1
void FASTCALL call_10_11_5_1(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(5<<4)+1);
}
#define init_10_11_5_2
void FASTCALL call_10_11_5_2(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(5<<4)+2);
}
#define init_10_11_5_3
void FASTCALL call_10_11_5_3(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(5<<4)+3);
}
#define init_10_11_5_4
void FASTCALL call_10_11_5_4(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(5<<4)+4);
}
#define init_10_11_5_5
void FASTCALL call_10_11_5_5(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(5<<4)+5);
}
#define init_10_11_5_6
void FASTCALL call_10_11_5_6(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(5<<4)+6);
}
#define init_10_11_5_7
void FASTCALL call_10_11_5_7(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(5<<4)+7);
}
#define init_10_11_5_8
void FASTCALL call_10_11_5_8(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(5<<4)+8);
}
#define init_10_11_5_9
void FASTCALL call_10_11_5_9(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(5<<4)+9);
}
#define init_10_11_5_10
void FASTCALL call_10_11_5_10(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(5<<4)+10);
}
#define init_10_11_5_11
void FASTCALL call_10_11_5_11(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(5<<4)+11);
}
#define init_10_11_5_12
void FASTCALL call_10_11_5_12(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(5<<4)+12);
}
#define init_10_11_5_13
void FASTCALL call_10_11_5_13(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(5<<4)+13);
}
#define init_10_11_5_14
void FASTCALL call_10_11_5_14(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(5<<4)+14);
}
#define init_10_11_5_15
void FASTCALL call_10_11_5_15(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(5<<4)+15);
}
#define init_10_11_6_0
void FASTCALL call_10_11_6_0(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(6<<4)+0);
}
#define init_10_11_6_1
void FASTCALL call_10_11_6_1(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(6<<4)+1);
}
#define init_10_11_6_2
void FASTCALL call_10_11_6_2(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(6<<4)+2);
}
#define init_10_11_6_3
void FASTCALL call_10_11_6_3(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(6<<4)+3);
}
#define init_10_11_6_4
void FASTCALL call_10_11_6_4(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(6<<4)+4);
}
#define init_10_11_6_5
void FASTCALL call_10_11_6_5(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(6<<4)+5);
}
#define init_10_11_6_6
void FASTCALL call_10_11_6_6(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(6<<4)+6);
}
#define init_10_11_6_7
void FASTCALL call_10_11_6_7(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(6<<4)+7);
}
#define init_10_11_6_8
void FASTCALL call_10_11_6_8(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(6<<4)+8);
}
#define init_10_11_6_9
void FASTCALL call_10_11_6_9(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(6<<4)+9);
}
#define init_10_11_6_10
void FASTCALL call_10_11_6_10(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(6<<4)+10);
}
#define init_10_11_6_11
void FASTCALL call_10_11_6_11(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(6<<4)+11);
}
#define init_10_11_6_12
void FASTCALL call_10_11_6_12(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(6<<4)+12);
}
#define init_10_11_6_13
void FASTCALL call_10_11_6_13(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(6<<4)+13);
}
#define init_10_11_6_14
void FASTCALL call_10_11_6_14(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(6<<4)+14);
}
#define init_10_11_6_15
void FASTCALL call_10_11_6_15(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(6<<4)+15);
}
#define init_10_11_7_0
void FASTCALL call_10_11_7_0(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(7<<4)+0);
}
#define init_10_11_7_1
void FASTCALL call_10_11_7_1(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(7<<4)+1);
}
#define init_10_11_7_2
void FASTCALL call_10_11_7_2(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(7<<4)+2);
}
#define init_10_11_7_3
void FASTCALL call_10_11_7_3(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(7<<4)+3);
}
#define init_10_11_7_4
void FASTCALL call_10_11_7_4(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(7<<4)+4);
}
#define init_10_11_7_5
void FASTCALL call_10_11_7_5(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(7<<4)+5);
}
#define init_10_11_7_6
void FASTCALL call_10_11_7_6(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(7<<4)+6);
}
#define init_10_11_7_7
void FASTCALL call_10_11_7_7(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(7<<4)+7);
}
#define init_10_11_7_8
void FASTCALL call_10_11_7_8(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(7<<4)+8);
}
#define init_10_11_7_9
void FASTCALL call_10_11_7_9(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(7<<4)+9);
}
#define init_10_11_7_10
void FASTCALL call_10_11_7_10(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(7<<4)+10);
}
#define init_10_11_7_11
void FASTCALL call_10_11_7_11(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(7<<4)+11);
}
#define init_10_11_7_12
void FASTCALL call_10_11_7_12(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(7<<4)+12);
}
#define init_10_11_7_13
void FASTCALL call_10_11_7_13(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(7<<4)+13);
}
#define init_10_11_7_14
void FASTCALL call_10_11_7_14(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(7<<4)+14);
}
#define init_10_11_7_15
void FASTCALL call_10_11_7_15(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(7<<4)+15);
}
#define init_10_11_8_0
void FASTCALL call_10_11_8_0(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(8<<4)+0);
}
#define init_10_11_8_1
void FASTCALL call_10_11_8_1(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(8<<4)+1);
}
#define init_10_11_8_2
void FASTCALL call_10_11_8_2(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(8<<4)+2);
}
#define init_10_11_8_3
void FASTCALL call_10_11_8_3(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(8<<4)+3);
}
#define init_10_11_8_4
void FASTCALL call_10_11_8_4(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(8<<4)+4);
}
#define init_10_11_8_5
void FASTCALL call_10_11_8_5(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(8<<4)+5);
}
#define init_10_11_8_6
void FASTCALL call_10_11_8_6(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(8<<4)+6);
}
#define init_10_11_8_7
void FASTCALL call_10_11_8_7(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(8<<4)+7);
}
#define init_10_11_8_8
void FASTCALL call_10_11_8_8(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(8<<4)+8);
}
#define init_10_11_8_9
void FASTCALL call_10_11_8_9(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(8<<4)+9);
}
#define init_10_11_8_10
void FASTCALL call_10_11_8_10(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(8<<4)+10);
}
#define init_10_11_8_11
void FASTCALL call_10_11_8_11(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(8<<4)+11);
}
#define init_10_11_8_12
void FASTCALL call_10_11_8_12(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(8<<4)+12);
}
#define init_10_11_8_13
void FASTCALL call_10_11_8_13(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(8<<4)+13);
}
#define init_10_11_8_14
void FASTCALL call_10_11_8_14(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(8<<4)+14);
}
#define init_10_11_8_15
void FASTCALL call_10_11_8_15(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(8<<4)+15);
}
#define init_10_11_9_0
void FASTCALL call_10_11_9_0(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(9<<4)+0);
}
#define init_10_11_9_1
void FASTCALL call_10_11_9_1(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(9<<4)+1);
}
#define init_10_11_9_2
void FASTCALL call_10_11_9_2(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(9<<4)+2);
}
#define init_10_11_9_3
void FASTCALL call_10_11_9_3(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(9<<4)+3);
}
#define init_10_11_9_4
void FASTCALL call_10_11_9_4(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(9<<4)+4);
}
#define init_10_11_9_5
void FASTCALL call_10_11_9_5(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(9<<4)+5);
}
#define init_10_11_9_6
void FASTCALL call_10_11_9_6(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(9<<4)+6);
}
#define init_10_11_9_7
void FASTCALL call_10_11_9_7(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(9<<4)+7);
}
#define init_10_11_9_8
void FASTCALL call_10_11_9_8(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(9<<4)+8);
}
#define init_10_11_9_9
void FASTCALL call_10_11_9_9(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(9<<4)+9);
}
#define init_10_11_9_10
void FASTCALL call_10_11_9_10(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(9<<4)+10);
}
#define init_10_11_9_11
void FASTCALL call_10_11_9_11(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(9<<4)+11);
}
#define init_10_11_9_12
void FASTCALL call_10_11_9_12(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(9<<4)+12);
}
#define init_10_11_9_13
void FASTCALL call_10_11_9_13(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(9<<4)+13);
}
#define init_10_11_9_14
void FASTCALL call_10_11_9_14(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(9<<4)+14);
}
#define init_10_11_9_15
void FASTCALL call_10_11_9_15(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(9<<4)+15);
}
#define init_10_11_10_0
void FASTCALL call_10_11_10_0(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(10<<4)+0);
}
#define init_10_11_10_1
void FASTCALL call_10_11_10_1(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(10<<4)+1);
}
#define init_10_11_10_2
void FASTCALL call_10_11_10_2(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(10<<4)+2);
}
#define init_10_11_10_3
void FASTCALL call_10_11_10_3(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(10<<4)+3);
}
#define init_10_11_10_4
void FASTCALL call_10_11_10_4(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(10<<4)+4);
}
#define init_10_11_10_5
void FASTCALL call_10_11_10_5(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(10<<4)+5);
}
#define init_10_11_10_6
void FASTCALL call_10_11_10_6(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(10<<4)+6);
}
#define init_10_11_10_7
void FASTCALL call_10_11_10_7(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(10<<4)+7);
}
#define init_10_11_10_8
void FASTCALL call_10_11_10_8(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(10<<4)+8);
}
#define init_10_11_10_9
void FASTCALL call_10_11_10_9(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(10<<4)+9);
}
#define init_10_11_10_10
void FASTCALL call_10_11_10_10(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(10<<4)+10);
}
#define init_10_11_10_11
void FASTCALL call_10_11_10_11(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(10<<4)+11);
}
#define init_10_11_10_12
void FASTCALL call_10_11_10_12(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(10<<4)+12);
}
#define init_10_11_10_13
void FASTCALL call_10_11_10_13(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(10<<4)+13);
}
#define init_10_11_10_14
void FASTCALL call_10_11_10_14(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(10<<4)+14);
}
#define init_10_11_10_15
void FASTCALL call_10_11_10_15(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(10<<4)+15);
}
#define init_10_11_11_0
void FASTCALL call_10_11_11_0(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(11<<4)+0);
}
#define init_10_11_11_1
void FASTCALL call_10_11_11_1(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(11<<4)+1);
}
#define init_10_11_11_2
void FASTCALL call_10_11_11_2(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(11<<4)+2);
}
#define init_10_11_11_3
void FASTCALL call_10_11_11_3(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(11<<4)+3);
}
#define init_10_11_11_4
void FASTCALL call_10_11_11_4(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(11<<4)+4);
}
#define init_10_11_11_5
void FASTCALL call_10_11_11_5(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(11<<4)+5);
}
#define init_10_11_11_6
void FASTCALL call_10_11_11_6(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(11<<4)+6);
}
#define init_10_11_11_7
void FASTCALL call_10_11_11_7(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(11<<4)+7);
}
#define init_10_11_11_8
void FASTCALL call_10_11_11_8(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(11<<4)+8);
}
#define init_10_11_11_9
void FASTCALL call_10_11_11_9(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(11<<4)+9);
}
#define init_10_11_11_10
void FASTCALL call_10_11_11_10(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(11<<4)+10);
}
#define init_10_11_11_11
void FASTCALL call_10_11_11_11(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(11<<4)+11);
}
#define init_10_11_11_12
void FASTCALL call_10_11_11_12(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(11<<4)+12);
}
#define init_10_11_11_13
void FASTCALL call_10_11_11_13(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(11<<4)+13);
}
#define init_10_11_11_14
void FASTCALL call_10_11_11_14(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(11<<4)+14);
}
#define init_10_11_11_15
void FASTCALL call_10_11_11_15(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(11<<4)+15);
}
#define init_10_11_12_0
void FASTCALL call_10_11_12_0(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(12<<4)+0);
}
#define init_10_11_12_1
void FASTCALL call_10_11_12_1(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(12<<4)+1);
}
#define init_10_11_12_2
void FASTCALL call_10_11_12_2(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(12<<4)+2);
}
#define init_10_11_12_3
void FASTCALL call_10_11_12_3(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(12<<4)+3);
}
#define init_10_11_12_4
void FASTCALL call_10_11_12_4(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(12<<4)+4);
}
#define init_10_11_12_5
void FASTCALL call_10_11_12_5(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(12<<4)+5);
}
#define init_10_11_12_6
void FASTCALL call_10_11_12_6(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(12<<4)+6);
}
#define init_10_11_12_7
void FASTCALL call_10_11_12_7(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(12<<4)+7);
}
#define init_10_11_12_8
void FASTCALL call_10_11_12_8(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(12<<4)+8);
}
#define init_10_11_12_9
void FASTCALL call_10_11_12_9(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(12<<4)+9);
}
#define init_10_11_12_10
void FASTCALL call_10_11_12_10(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(12<<4)+10);
}
#define init_10_11_12_11
void FASTCALL call_10_11_12_11(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(12<<4)+11);
}
#define init_10_11_12_12
void FASTCALL call_10_11_12_12(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(12<<4)+12);
}
#define init_10_11_12_13
void FASTCALL call_10_11_12_13(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(12<<4)+13);
}
#define init_10_11_12_14
void FASTCALL call_10_11_12_14(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(12<<4)+14);
}
#define init_10_11_12_15
void FASTCALL call_10_11_12_15(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(12<<4)+15);
}
#define init_10_11_13_0
void FASTCALL call_10_11_13_0(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(13<<4)+0);
}
#define init_10_11_13_1
void FASTCALL call_10_11_13_1(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(13<<4)+1);
}
#define init_10_11_13_2
void FASTCALL call_10_11_13_2(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(13<<4)+2);
}
#define init_10_11_13_3
void FASTCALL call_10_11_13_3(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(13<<4)+3);
}
#define init_10_11_13_4
void FASTCALL call_10_11_13_4(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(13<<4)+4);
}
#define init_10_11_13_5
void FASTCALL call_10_11_13_5(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(13<<4)+5);
}
#define init_10_11_13_6
void FASTCALL call_10_11_13_6(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(13<<4)+6);
}
#define init_10_11_13_7
void FASTCALL call_10_11_13_7(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(13<<4)+7);
}
#define init_10_11_13_8
void FASTCALL call_10_11_13_8(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(13<<4)+8);
}
#define init_10_11_13_9
void FASTCALL call_10_11_13_9(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(13<<4)+9);
}
#define init_10_11_13_10
void FASTCALL call_10_11_13_10(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(13<<4)+10);
}
#define init_10_11_13_11
void FASTCALL call_10_11_13_11(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(13<<4)+11);
}
#define init_10_11_13_12
void FASTCALL call_10_11_13_12(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(13<<4)+12);
}
#define init_10_11_13_13
void FASTCALL call_10_11_13_13(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(13<<4)+13);
}
#define init_10_11_13_14
void FASTCALL call_10_11_13_14(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(13<<4)+14);
}
#define init_10_11_13_15
void FASTCALL call_10_11_13_15(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(13<<4)+15);
}
#define init_10_11_14_0
void FASTCALL call_10_11_14_0(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(14<<4)+0);
}
#define init_10_11_14_1
void FASTCALL call_10_11_14_1(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(14<<4)+1);
}
#define init_10_11_14_2
void FASTCALL call_10_11_14_2(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(14<<4)+2);
}
#define init_10_11_14_3
void FASTCALL call_10_11_14_3(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(14<<4)+3);
}
#define init_10_11_14_4
void FASTCALL call_10_11_14_4(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(14<<4)+4);
}
#define init_10_11_14_5
void FASTCALL call_10_11_14_5(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(14<<4)+5);
}
#define init_10_11_14_6
void FASTCALL call_10_11_14_6(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(14<<4)+6);
}
#define init_10_11_14_7
void FASTCALL call_10_11_14_7(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(14<<4)+7);
}
#define init_10_11_14_8
void FASTCALL call_10_11_14_8(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(14<<4)+8);
}
#define init_10_11_14_9
void FASTCALL call_10_11_14_9(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(14<<4)+9);
}
#define init_10_11_14_10
void FASTCALL call_10_11_14_10(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(14<<4)+10);
}
#define init_10_11_14_11
void FASTCALL call_10_11_14_11(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(14<<4)+11);
}
#define init_10_11_14_12
void FASTCALL call_10_11_14_12(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(14<<4)+12);
}
#define init_10_11_14_13
void FASTCALL call_10_11_14_13(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(14<<4)+13);
}
#define init_10_11_14_14
void FASTCALL call_10_11_14_14(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(14<<4)+14);
}
#define init_10_11_14_15
void FASTCALL call_10_11_14_15(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(14<<4)+15);
}
#define init_10_11_15_0
void FASTCALL call_10_11_15_0(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(15<<4)+0);
}
#define init_10_11_15_1
void FASTCALL call_10_11_15_1(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(15<<4)+1);
}
#define init_10_11_15_2
void FASTCALL call_10_11_15_2(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(15<<4)+2);
}
#define init_10_11_15_3
void FASTCALL call_10_11_15_3(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(15<<4)+3);
}
#define init_10_11_15_4
void FASTCALL call_10_11_15_4(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(15<<4)+4);
}
#define init_10_11_15_5
void FASTCALL call_10_11_15_5(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(15<<4)+5);
}
#define init_10_11_15_6
void FASTCALL call_10_11_15_6(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(15<<4)+6);
}
#define init_10_11_15_7
void FASTCALL call_10_11_15_7(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(15<<4)+7);
}
#define init_10_11_15_8
void FASTCALL call_10_11_15_8(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(15<<4)+8);
}
#define init_10_11_15_9
void FASTCALL call_10_11_15_9(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(15<<4)+9);
}
#define init_10_11_15_10
void FASTCALL call_10_11_15_10(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(15<<4)+10);
}
#define init_10_11_15_11
void FASTCALL call_10_11_15_11(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(15<<4)+11);
}
#define init_10_11_15_12
void FASTCALL call_10_11_15_12(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(15<<4)+12);
}
#define init_10_11_15_13
void FASTCALL call_10_11_15_13(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(15<<4)+13);
}
#define init_10_11_15_14
void FASTCALL call_10_11_15_14(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(15<<4)+14);
}
#define init_10_11_15_15
void FASTCALL call_10_11_15_15(SH2_struct* sh)
{
  SH2bra(sh, (11<<8)+(15<<4)+15);
}
#define init_10_12_0_0
void FASTCALL call_10_12_0_0(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(0<<4)+0);
}
#define init_10_12_0_1
void FASTCALL call_10_12_0_1(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(0<<4)+1);
}
#define init_10_12_0_2
void FASTCALL call_10_12_0_2(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(0<<4)+2);
}
#define init_10_12_0_3
void FASTCALL call_10_12_0_3(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(0<<4)+3);
}
#define init_10_12_0_4
void FASTCALL call_10_12_0_4(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(0<<4)+4);
}
#define init_10_12_0_5
void FASTCALL call_10_12_0_5(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(0<<4)+5);
}
#define init_10_12_0_6
void FASTCALL call_10_12_0_6(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(0<<4)+6);
}
#define init_10_12_0_7
void FASTCALL call_10_12_0_7(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(0<<4)+7);
}
#define init_10_12_0_8
void FASTCALL call_10_12_0_8(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(0<<4)+8);
}
#define init_10_12_0_9
void FASTCALL call_10_12_0_9(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(0<<4)+9);
}
#define init_10_12_0_10
void FASTCALL call_10_12_0_10(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(0<<4)+10);
}
#define init_10_12_0_11
void FASTCALL call_10_12_0_11(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(0<<4)+11);
}
#define init_10_12_0_12
void FASTCALL call_10_12_0_12(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(0<<4)+12);
}
#define init_10_12_0_13
void FASTCALL call_10_12_0_13(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(0<<4)+13);
}
#define init_10_12_0_14
void FASTCALL call_10_12_0_14(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(0<<4)+14);
}
#define init_10_12_0_15
void FASTCALL call_10_12_0_15(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(0<<4)+15);
}
#define init_10_12_1_0
void FASTCALL call_10_12_1_0(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(1<<4)+0);
}
#define init_10_12_1_1
void FASTCALL call_10_12_1_1(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(1<<4)+1);
}
#define init_10_12_1_2
void FASTCALL call_10_12_1_2(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(1<<4)+2);
}
#define init_10_12_1_3
void FASTCALL call_10_12_1_3(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(1<<4)+3);
}
#define init_10_12_1_4
void FASTCALL call_10_12_1_4(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(1<<4)+4);
}
#define init_10_12_1_5
void FASTCALL call_10_12_1_5(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(1<<4)+5);
}
#define init_10_12_1_6
void FASTCALL call_10_12_1_6(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(1<<4)+6);
}
#define init_10_12_1_7
void FASTCALL call_10_12_1_7(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(1<<4)+7);
}
#define init_10_12_1_8
void FASTCALL call_10_12_1_8(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(1<<4)+8);
}
#define init_10_12_1_9
void FASTCALL call_10_12_1_9(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(1<<4)+9);
}
#define init_10_12_1_10
void FASTCALL call_10_12_1_10(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(1<<4)+10);
}
#define init_10_12_1_11
void FASTCALL call_10_12_1_11(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(1<<4)+11);
}
#define init_10_12_1_12
void FASTCALL call_10_12_1_12(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(1<<4)+12);
}
#define init_10_12_1_13
void FASTCALL call_10_12_1_13(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(1<<4)+13);
}
#define init_10_12_1_14
void FASTCALL call_10_12_1_14(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(1<<4)+14);
}
#define init_10_12_1_15
void FASTCALL call_10_12_1_15(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(1<<4)+15);
}
#define init_10_12_2_0
void FASTCALL call_10_12_2_0(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(2<<4)+0);
}
#define init_10_12_2_1
void FASTCALL call_10_12_2_1(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(2<<4)+1);
}
#define init_10_12_2_2
void FASTCALL call_10_12_2_2(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(2<<4)+2);
}
#define init_10_12_2_3
void FASTCALL call_10_12_2_3(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(2<<4)+3);
}
#define init_10_12_2_4
void FASTCALL call_10_12_2_4(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(2<<4)+4);
}
#define init_10_12_2_5
void FASTCALL call_10_12_2_5(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(2<<4)+5);
}
#define init_10_12_2_6
void FASTCALL call_10_12_2_6(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(2<<4)+6);
}
#define init_10_12_2_7
void FASTCALL call_10_12_2_7(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(2<<4)+7);
}
#define init_10_12_2_8
void FASTCALL call_10_12_2_8(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(2<<4)+8);
}
#define init_10_12_2_9
void FASTCALL call_10_12_2_9(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(2<<4)+9);
}
#define init_10_12_2_10
void FASTCALL call_10_12_2_10(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(2<<4)+10);
}
#define init_10_12_2_11
void FASTCALL call_10_12_2_11(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(2<<4)+11);
}
#define init_10_12_2_12
void FASTCALL call_10_12_2_12(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(2<<4)+12);
}
#define init_10_12_2_13
void FASTCALL call_10_12_2_13(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(2<<4)+13);
}
#define init_10_12_2_14
void FASTCALL call_10_12_2_14(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(2<<4)+14);
}
#define init_10_12_2_15
void FASTCALL call_10_12_2_15(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(2<<4)+15);
}
#define init_10_12_3_0
void FASTCALL call_10_12_3_0(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(3<<4)+0);
}
#define init_10_12_3_1
void FASTCALL call_10_12_3_1(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(3<<4)+1);
}
#define init_10_12_3_2
void FASTCALL call_10_12_3_2(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(3<<4)+2);
}
#define init_10_12_3_3
void FASTCALL call_10_12_3_3(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(3<<4)+3);
}
#define init_10_12_3_4
void FASTCALL call_10_12_3_4(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(3<<4)+4);
}
#define init_10_12_3_5
void FASTCALL call_10_12_3_5(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(3<<4)+5);
}
#define init_10_12_3_6
void FASTCALL call_10_12_3_6(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(3<<4)+6);
}
#define init_10_12_3_7
void FASTCALL call_10_12_3_7(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(3<<4)+7);
}
#define init_10_12_3_8
void FASTCALL call_10_12_3_8(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(3<<4)+8);
}
#define init_10_12_3_9
void FASTCALL call_10_12_3_9(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(3<<4)+9);
}
#define init_10_12_3_10
void FASTCALL call_10_12_3_10(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(3<<4)+10);
}
#define init_10_12_3_11
void FASTCALL call_10_12_3_11(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(3<<4)+11);
}
#define init_10_12_3_12
void FASTCALL call_10_12_3_12(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(3<<4)+12);
}
#define init_10_12_3_13
void FASTCALL call_10_12_3_13(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(3<<4)+13);
}
#define init_10_12_3_14
void FASTCALL call_10_12_3_14(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(3<<4)+14);
}
#define init_10_12_3_15
void FASTCALL call_10_12_3_15(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(3<<4)+15);
}
#define init_10_12_4_0
void FASTCALL call_10_12_4_0(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(4<<4)+0);
}
#define init_10_12_4_1
void FASTCALL call_10_12_4_1(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(4<<4)+1);
}
#define init_10_12_4_2
void FASTCALL call_10_12_4_2(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(4<<4)+2);
}
#define init_10_12_4_3
void FASTCALL call_10_12_4_3(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(4<<4)+3);
}
#define init_10_12_4_4
void FASTCALL call_10_12_4_4(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(4<<4)+4);
}
#define init_10_12_4_5
void FASTCALL call_10_12_4_5(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(4<<4)+5);
}
#define init_10_12_4_6
void FASTCALL call_10_12_4_6(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(4<<4)+6);
}
#define init_10_12_4_7
void FASTCALL call_10_12_4_7(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(4<<4)+7);
}
#define init_10_12_4_8
void FASTCALL call_10_12_4_8(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(4<<4)+8);
}
#define init_10_12_4_9
void FASTCALL call_10_12_4_9(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(4<<4)+9);
}
#define init_10_12_4_10
void FASTCALL call_10_12_4_10(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(4<<4)+10);
}
#define init_10_12_4_11
void FASTCALL call_10_12_4_11(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(4<<4)+11);
}
#define init_10_12_4_12
void FASTCALL call_10_12_4_12(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(4<<4)+12);
}
#define init_10_12_4_13
void FASTCALL call_10_12_4_13(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(4<<4)+13);
}
#define init_10_12_4_14
void FASTCALL call_10_12_4_14(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(4<<4)+14);
}
#define init_10_12_4_15
void FASTCALL call_10_12_4_15(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(4<<4)+15);
}
#define init_10_12_5_0
void FASTCALL call_10_12_5_0(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(5<<4)+0);
}
#define init_10_12_5_1
void FASTCALL call_10_12_5_1(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(5<<4)+1);
}
#define init_10_12_5_2
void FASTCALL call_10_12_5_2(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(5<<4)+2);
}
#define init_10_12_5_3
void FASTCALL call_10_12_5_3(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(5<<4)+3);
}
#define init_10_12_5_4
void FASTCALL call_10_12_5_4(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(5<<4)+4);
}
#define init_10_12_5_5
void FASTCALL call_10_12_5_5(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(5<<4)+5);
}
#define init_10_12_5_6
void FASTCALL call_10_12_5_6(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(5<<4)+6);
}
#define init_10_12_5_7
void FASTCALL call_10_12_5_7(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(5<<4)+7);
}
#define init_10_12_5_8
void FASTCALL call_10_12_5_8(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(5<<4)+8);
}
#define init_10_12_5_9
void FASTCALL call_10_12_5_9(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(5<<4)+9);
}
#define init_10_12_5_10
void FASTCALL call_10_12_5_10(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(5<<4)+10);
}
#define init_10_12_5_11
void FASTCALL call_10_12_5_11(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(5<<4)+11);
}
#define init_10_12_5_12
void FASTCALL call_10_12_5_12(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(5<<4)+12);
}
#define init_10_12_5_13
void FASTCALL call_10_12_5_13(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(5<<4)+13);
}
#define init_10_12_5_14
void FASTCALL call_10_12_5_14(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(5<<4)+14);
}
#define init_10_12_5_15
void FASTCALL call_10_12_5_15(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(5<<4)+15);
}
#define init_10_12_6_0
void FASTCALL call_10_12_6_0(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(6<<4)+0);
}
#define init_10_12_6_1
void FASTCALL call_10_12_6_1(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(6<<4)+1);
}
#define init_10_12_6_2
void FASTCALL call_10_12_6_2(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(6<<4)+2);
}
#define init_10_12_6_3
void FASTCALL call_10_12_6_3(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(6<<4)+3);
}
#define init_10_12_6_4
void FASTCALL call_10_12_6_4(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(6<<4)+4);
}
#define init_10_12_6_5
void FASTCALL call_10_12_6_5(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(6<<4)+5);
}
#define init_10_12_6_6
void FASTCALL call_10_12_6_6(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(6<<4)+6);
}
#define init_10_12_6_7
void FASTCALL call_10_12_6_7(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(6<<4)+7);
}
#define init_10_12_6_8
void FASTCALL call_10_12_6_8(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(6<<4)+8);
}
#define init_10_12_6_9
void FASTCALL call_10_12_6_9(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(6<<4)+9);
}
#define init_10_12_6_10
void FASTCALL call_10_12_6_10(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(6<<4)+10);
}
#define init_10_12_6_11
void FASTCALL call_10_12_6_11(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(6<<4)+11);
}
#define init_10_12_6_12
void FASTCALL call_10_12_6_12(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(6<<4)+12);
}
#define init_10_12_6_13
void FASTCALL call_10_12_6_13(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(6<<4)+13);
}
#define init_10_12_6_14
void FASTCALL call_10_12_6_14(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(6<<4)+14);
}
#define init_10_12_6_15
void FASTCALL call_10_12_6_15(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(6<<4)+15);
}
#define init_10_12_7_0
void FASTCALL call_10_12_7_0(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(7<<4)+0);
}
#define init_10_12_7_1
void FASTCALL call_10_12_7_1(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(7<<4)+1);
}
#define init_10_12_7_2
void FASTCALL call_10_12_7_2(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(7<<4)+2);
}
#define init_10_12_7_3
void FASTCALL call_10_12_7_3(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(7<<4)+3);
}
#define init_10_12_7_4
void FASTCALL call_10_12_7_4(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(7<<4)+4);
}
#define init_10_12_7_5
void FASTCALL call_10_12_7_5(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(7<<4)+5);
}
#define init_10_12_7_6
void FASTCALL call_10_12_7_6(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(7<<4)+6);
}
#define init_10_12_7_7
void FASTCALL call_10_12_7_7(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(7<<4)+7);
}
#define init_10_12_7_8
void FASTCALL call_10_12_7_8(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(7<<4)+8);
}
#define init_10_12_7_9
void FASTCALL call_10_12_7_9(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(7<<4)+9);
}
#define init_10_12_7_10
void FASTCALL call_10_12_7_10(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(7<<4)+10);
}
#define init_10_12_7_11
void FASTCALL call_10_12_7_11(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(7<<4)+11);
}
#define init_10_12_7_12
void FASTCALL call_10_12_7_12(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(7<<4)+12);
}
#define init_10_12_7_13
void FASTCALL call_10_12_7_13(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(7<<4)+13);
}
#define init_10_12_7_14
void FASTCALL call_10_12_7_14(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(7<<4)+14);
}
#define init_10_12_7_15
void FASTCALL call_10_12_7_15(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(7<<4)+15);
}
#define init_10_12_8_0
void FASTCALL call_10_12_8_0(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(8<<4)+0);
}
#define init_10_12_8_1
void FASTCALL call_10_12_8_1(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(8<<4)+1);
}
#define init_10_12_8_2
void FASTCALL call_10_12_8_2(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(8<<4)+2);
}
#define init_10_12_8_3
void FASTCALL call_10_12_8_3(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(8<<4)+3);
}
#define init_10_12_8_4
void FASTCALL call_10_12_8_4(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(8<<4)+4);
}
#define init_10_12_8_5
void FASTCALL call_10_12_8_5(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(8<<4)+5);
}
#define init_10_12_8_6
void FASTCALL call_10_12_8_6(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(8<<4)+6);
}
#define init_10_12_8_7
void FASTCALL call_10_12_8_7(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(8<<4)+7);
}
#define init_10_12_8_8
void FASTCALL call_10_12_8_8(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(8<<4)+8);
}
#define init_10_12_8_9
void FASTCALL call_10_12_8_9(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(8<<4)+9);
}
#define init_10_12_8_10
void FASTCALL call_10_12_8_10(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(8<<4)+10);
}
#define init_10_12_8_11
void FASTCALL call_10_12_8_11(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(8<<4)+11);
}
#define init_10_12_8_12
void FASTCALL call_10_12_8_12(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(8<<4)+12);
}
#define init_10_12_8_13
void FASTCALL call_10_12_8_13(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(8<<4)+13);
}
#define init_10_12_8_14
void FASTCALL call_10_12_8_14(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(8<<4)+14);
}
#define init_10_12_8_15
void FASTCALL call_10_12_8_15(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(8<<4)+15);
}
#define init_10_12_9_0
void FASTCALL call_10_12_9_0(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(9<<4)+0);
}
#define init_10_12_9_1
void FASTCALL call_10_12_9_1(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(9<<4)+1);
}
#define init_10_12_9_2
void FASTCALL call_10_12_9_2(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(9<<4)+2);
}
#define init_10_12_9_3
void FASTCALL call_10_12_9_3(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(9<<4)+3);
}
#define init_10_12_9_4
void FASTCALL call_10_12_9_4(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(9<<4)+4);
}
#define init_10_12_9_5
void FASTCALL call_10_12_9_5(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(9<<4)+5);
}
#define init_10_12_9_6
void FASTCALL call_10_12_9_6(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(9<<4)+6);
}
#define init_10_12_9_7
void FASTCALL call_10_12_9_7(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(9<<4)+7);
}
#define init_10_12_9_8
void FASTCALL call_10_12_9_8(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(9<<4)+8);
}
#define init_10_12_9_9
void FASTCALL call_10_12_9_9(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(9<<4)+9);
}
#define init_10_12_9_10
void FASTCALL call_10_12_9_10(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(9<<4)+10);
}
#define init_10_12_9_11
void FASTCALL call_10_12_9_11(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(9<<4)+11);
}
#define init_10_12_9_12
void FASTCALL call_10_12_9_12(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(9<<4)+12);
}
#define init_10_12_9_13
void FASTCALL call_10_12_9_13(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(9<<4)+13);
}
#define init_10_12_9_14
void FASTCALL call_10_12_9_14(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(9<<4)+14);
}
#define init_10_12_9_15
void FASTCALL call_10_12_9_15(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(9<<4)+15);
}
#define init_10_12_10_0
void FASTCALL call_10_12_10_0(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(10<<4)+0);
}
#define init_10_12_10_1
void FASTCALL call_10_12_10_1(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(10<<4)+1);
}
#define init_10_12_10_2
void FASTCALL call_10_12_10_2(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(10<<4)+2);
}
#define init_10_12_10_3
void FASTCALL call_10_12_10_3(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(10<<4)+3);
}
#define init_10_12_10_4
void FASTCALL call_10_12_10_4(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(10<<4)+4);
}
#define init_10_12_10_5
void FASTCALL call_10_12_10_5(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(10<<4)+5);
}
#define init_10_12_10_6
void FASTCALL call_10_12_10_6(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(10<<4)+6);
}
#define init_10_12_10_7
void FASTCALL call_10_12_10_7(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(10<<4)+7);
}
#define init_10_12_10_8
void FASTCALL call_10_12_10_8(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(10<<4)+8);
}
#define init_10_12_10_9
void FASTCALL call_10_12_10_9(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(10<<4)+9);
}
#define init_10_12_10_10
void FASTCALL call_10_12_10_10(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(10<<4)+10);
}
#define init_10_12_10_11
void FASTCALL call_10_12_10_11(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(10<<4)+11);
}
#define init_10_12_10_12
void FASTCALL call_10_12_10_12(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(10<<4)+12);
}
#define init_10_12_10_13
void FASTCALL call_10_12_10_13(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(10<<4)+13);
}
#define init_10_12_10_14
void FASTCALL call_10_12_10_14(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(10<<4)+14);
}
#define init_10_12_10_15
void FASTCALL call_10_12_10_15(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(10<<4)+15);
}
#define init_10_12_11_0
void FASTCALL call_10_12_11_0(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(11<<4)+0);
}
#define init_10_12_11_1
void FASTCALL call_10_12_11_1(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(11<<4)+1);
}
#define init_10_12_11_2
void FASTCALL call_10_12_11_2(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(11<<4)+2);
}
#define init_10_12_11_3
void FASTCALL call_10_12_11_3(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(11<<4)+3);
}
#define init_10_12_11_4
void FASTCALL call_10_12_11_4(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(11<<4)+4);
}
#define init_10_12_11_5
void FASTCALL call_10_12_11_5(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(11<<4)+5);
}
#define init_10_12_11_6
void FASTCALL call_10_12_11_6(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(11<<4)+6);
}
#define init_10_12_11_7
void FASTCALL call_10_12_11_7(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(11<<4)+7);
}
#define init_10_12_11_8
void FASTCALL call_10_12_11_8(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(11<<4)+8);
}
#define init_10_12_11_9
void FASTCALL call_10_12_11_9(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(11<<4)+9);
}
#define init_10_12_11_10
void FASTCALL call_10_12_11_10(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(11<<4)+10);
}
#define init_10_12_11_11
void FASTCALL call_10_12_11_11(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(11<<4)+11);
}
#define init_10_12_11_12
void FASTCALL call_10_12_11_12(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(11<<4)+12);
}
#define init_10_12_11_13
void FASTCALL call_10_12_11_13(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(11<<4)+13);
}
#define init_10_12_11_14
void FASTCALL call_10_12_11_14(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(11<<4)+14);
}
#define init_10_12_11_15
void FASTCALL call_10_12_11_15(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(11<<4)+15);
}
#define init_10_12_12_0
void FASTCALL call_10_12_12_0(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(12<<4)+0);
}
#define init_10_12_12_1
void FASTCALL call_10_12_12_1(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(12<<4)+1);
}
#define init_10_12_12_2
void FASTCALL call_10_12_12_2(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(12<<4)+2);
}
#define init_10_12_12_3
void FASTCALL call_10_12_12_3(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(12<<4)+3);
}
#define init_10_12_12_4
void FASTCALL call_10_12_12_4(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(12<<4)+4);
}
#define init_10_12_12_5
void FASTCALL call_10_12_12_5(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(12<<4)+5);
}
#define init_10_12_12_6
void FASTCALL call_10_12_12_6(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(12<<4)+6);
}
#define init_10_12_12_7
void FASTCALL call_10_12_12_7(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(12<<4)+7);
}
#define init_10_12_12_8
void FASTCALL call_10_12_12_8(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(12<<4)+8);
}
#define init_10_12_12_9
void FASTCALL call_10_12_12_9(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(12<<4)+9);
}
#define init_10_12_12_10
void FASTCALL call_10_12_12_10(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(12<<4)+10);
}
#define init_10_12_12_11
void FASTCALL call_10_12_12_11(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(12<<4)+11);
}
#define init_10_12_12_12
void FASTCALL call_10_12_12_12(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(12<<4)+12);
}
#define init_10_12_12_13
void FASTCALL call_10_12_12_13(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(12<<4)+13);
}
#define init_10_12_12_14
void FASTCALL call_10_12_12_14(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(12<<4)+14);
}
#define init_10_12_12_15
void FASTCALL call_10_12_12_15(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(12<<4)+15);
}
#define init_10_12_13_0
void FASTCALL call_10_12_13_0(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(13<<4)+0);
}
#define init_10_12_13_1
void FASTCALL call_10_12_13_1(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(13<<4)+1);
}
#define init_10_12_13_2
void FASTCALL call_10_12_13_2(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(13<<4)+2);
}
#define init_10_12_13_3
void FASTCALL call_10_12_13_3(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(13<<4)+3);
}
#define init_10_12_13_4
void FASTCALL call_10_12_13_4(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(13<<4)+4);
}
#define init_10_12_13_5
void FASTCALL call_10_12_13_5(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(13<<4)+5);
}
#define init_10_12_13_6
void FASTCALL call_10_12_13_6(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(13<<4)+6);
}
#define init_10_12_13_7
void FASTCALL call_10_12_13_7(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(13<<4)+7);
}
#define init_10_12_13_8
void FASTCALL call_10_12_13_8(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(13<<4)+8);
}
#define init_10_12_13_9
void FASTCALL call_10_12_13_9(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(13<<4)+9);
}
#define init_10_12_13_10
void FASTCALL call_10_12_13_10(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(13<<4)+10);
}
#define init_10_12_13_11
void FASTCALL call_10_12_13_11(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(13<<4)+11);
}
#define init_10_12_13_12
void FASTCALL call_10_12_13_12(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(13<<4)+12);
}
#define init_10_12_13_13
void FASTCALL call_10_12_13_13(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(13<<4)+13);
}
#define init_10_12_13_14
void FASTCALL call_10_12_13_14(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(13<<4)+14);
}
#define init_10_12_13_15
void FASTCALL call_10_12_13_15(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(13<<4)+15);
}
#define init_10_12_14_0
void FASTCALL call_10_12_14_0(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(14<<4)+0);
}
#define init_10_12_14_1
void FASTCALL call_10_12_14_1(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(14<<4)+1);
}
#define init_10_12_14_2
void FASTCALL call_10_12_14_2(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(14<<4)+2);
}
#define init_10_12_14_3
void FASTCALL call_10_12_14_3(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(14<<4)+3);
}
#define init_10_12_14_4
void FASTCALL call_10_12_14_4(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(14<<4)+4);
}
#define init_10_12_14_5
void FASTCALL call_10_12_14_5(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(14<<4)+5);
}
#define init_10_12_14_6
void FASTCALL call_10_12_14_6(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(14<<4)+6);
}
#define init_10_12_14_7
void FASTCALL call_10_12_14_7(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(14<<4)+7);
}
#define init_10_12_14_8
void FASTCALL call_10_12_14_8(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(14<<4)+8);
}
#define init_10_12_14_9
void FASTCALL call_10_12_14_9(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(14<<4)+9);
}
#define init_10_12_14_10
void FASTCALL call_10_12_14_10(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(14<<4)+10);
}
#define init_10_12_14_11
void FASTCALL call_10_12_14_11(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(14<<4)+11);
}
#define init_10_12_14_12
void FASTCALL call_10_12_14_12(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(14<<4)+12);
}
#define init_10_12_14_13
void FASTCALL call_10_12_14_13(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(14<<4)+13);
}
#define init_10_12_14_14
void FASTCALL call_10_12_14_14(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(14<<4)+14);
}
#define init_10_12_14_15
void FASTCALL call_10_12_14_15(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(14<<4)+15);
}
#define init_10_12_15_0
void FASTCALL call_10_12_15_0(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(15<<4)+0);
}
#define init_10_12_15_1
void FASTCALL call_10_12_15_1(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(15<<4)+1);
}
#define init_10_12_15_2
void FASTCALL call_10_12_15_2(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(15<<4)+2);
}
#define init_10_12_15_3
void FASTCALL call_10_12_15_3(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(15<<4)+3);
}
#define init_10_12_15_4
void FASTCALL call_10_12_15_4(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(15<<4)+4);
}
#define init_10_12_15_5
void FASTCALL call_10_12_15_5(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(15<<4)+5);
}
#define init_10_12_15_6
void FASTCALL call_10_12_15_6(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(15<<4)+6);
}
#define init_10_12_15_7
void FASTCALL call_10_12_15_7(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(15<<4)+7);
}
#define init_10_12_15_8
void FASTCALL call_10_12_15_8(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(15<<4)+8);
}
#define init_10_12_15_9
void FASTCALL call_10_12_15_9(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(15<<4)+9);
}
#define init_10_12_15_10
void FASTCALL call_10_12_15_10(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(15<<4)+10);
}
#define init_10_12_15_11
void FASTCALL call_10_12_15_11(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(15<<4)+11);
}
#define init_10_12_15_12
void FASTCALL call_10_12_15_12(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(15<<4)+12);
}
#define init_10_12_15_13
void FASTCALL call_10_12_15_13(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(15<<4)+13);
}
#define init_10_12_15_14
void FASTCALL call_10_12_15_14(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(15<<4)+14);
}
#define init_10_12_15_15
void FASTCALL call_10_12_15_15(SH2_struct* sh)
{
  SH2bra(sh, (12<<8)+(15<<4)+15);
}
#define init_10_13_0_0
void FASTCALL call_10_13_0_0(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(0<<4)+0);
}
#define init_10_13_0_1
void FASTCALL call_10_13_0_1(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(0<<4)+1);
}
#define init_10_13_0_2
void FASTCALL call_10_13_0_2(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(0<<4)+2);
}
#define init_10_13_0_3
void FASTCALL call_10_13_0_3(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(0<<4)+3);
}
#define init_10_13_0_4
void FASTCALL call_10_13_0_4(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(0<<4)+4);
}
#define init_10_13_0_5
void FASTCALL call_10_13_0_5(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(0<<4)+5);
}
#define init_10_13_0_6
void FASTCALL call_10_13_0_6(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(0<<4)+6);
}
#define init_10_13_0_7
void FASTCALL call_10_13_0_7(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(0<<4)+7);
}
#define init_10_13_0_8
void FASTCALL call_10_13_0_8(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(0<<4)+8);
}
#define init_10_13_0_9
void FASTCALL call_10_13_0_9(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(0<<4)+9);
}
#define init_10_13_0_10
void FASTCALL call_10_13_0_10(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(0<<4)+10);
}
#define init_10_13_0_11
void FASTCALL call_10_13_0_11(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(0<<4)+11);
}
#define init_10_13_0_12
void FASTCALL call_10_13_0_12(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(0<<4)+12);
}
#define init_10_13_0_13
void FASTCALL call_10_13_0_13(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(0<<4)+13);
}
#define init_10_13_0_14
void FASTCALL call_10_13_0_14(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(0<<4)+14);
}
#define init_10_13_0_15
void FASTCALL call_10_13_0_15(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(0<<4)+15);
}
#define init_10_13_1_0
void FASTCALL call_10_13_1_0(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(1<<4)+0);
}
#define init_10_13_1_1
void FASTCALL call_10_13_1_1(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(1<<4)+1);
}
#define init_10_13_1_2
void FASTCALL call_10_13_1_2(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(1<<4)+2);
}
#define init_10_13_1_3
void FASTCALL call_10_13_1_3(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(1<<4)+3);
}
#define init_10_13_1_4
void FASTCALL call_10_13_1_4(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(1<<4)+4);
}
#define init_10_13_1_5
void FASTCALL call_10_13_1_5(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(1<<4)+5);
}
#define init_10_13_1_6
void FASTCALL call_10_13_1_6(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(1<<4)+6);
}
#define init_10_13_1_7
void FASTCALL call_10_13_1_7(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(1<<4)+7);
}
#define init_10_13_1_8
void FASTCALL call_10_13_1_8(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(1<<4)+8);
}
#define init_10_13_1_9
void FASTCALL call_10_13_1_9(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(1<<4)+9);
}
#define init_10_13_1_10
void FASTCALL call_10_13_1_10(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(1<<4)+10);
}
#define init_10_13_1_11
void FASTCALL call_10_13_1_11(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(1<<4)+11);
}
#define init_10_13_1_12
void FASTCALL call_10_13_1_12(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(1<<4)+12);
}
#define init_10_13_1_13
void FASTCALL call_10_13_1_13(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(1<<4)+13);
}
#define init_10_13_1_14
void FASTCALL call_10_13_1_14(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(1<<4)+14);
}
#define init_10_13_1_15
void FASTCALL call_10_13_1_15(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(1<<4)+15);
}
#define init_10_13_2_0
void FASTCALL call_10_13_2_0(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(2<<4)+0);
}
#define init_10_13_2_1
void FASTCALL call_10_13_2_1(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(2<<4)+1);
}
#define init_10_13_2_2
void FASTCALL call_10_13_2_2(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(2<<4)+2);
}
#define init_10_13_2_3
void FASTCALL call_10_13_2_3(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(2<<4)+3);
}
#define init_10_13_2_4
void FASTCALL call_10_13_2_4(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(2<<4)+4);
}
#define init_10_13_2_5
void FASTCALL call_10_13_2_5(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(2<<4)+5);
}
#define init_10_13_2_6
void FASTCALL call_10_13_2_6(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(2<<4)+6);
}
#define init_10_13_2_7
void FASTCALL call_10_13_2_7(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(2<<4)+7);
}
#define init_10_13_2_8
void FASTCALL call_10_13_2_8(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(2<<4)+8);
}
#define init_10_13_2_9
void FASTCALL call_10_13_2_9(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(2<<4)+9);
}
#define init_10_13_2_10
void FASTCALL call_10_13_2_10(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(2<<4)+10);
}
#define init_10_13_2_11
void FASTCALL call_10_13_2_11(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(2<<4)+11);
}
#define init_10_13_2_12
void FASTCALL call_10_13_2_12(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(2<<4)+12);
}
#define init_10_13_2_13
void FASTCALL call_10_13_2_13(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(2<<4)+13);
}
#define init_10_13_2_14
void FASTCALL call_10_13_2_14(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(2<<4)+14);
}
#define init_10_13_2_15
void FASTCALL call_10_13_2_15(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(2<<4)+15);
}
#define init_10_13_3_0
void FASTCALL call_10_13_3_0(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(3<<4)+0);
}
#define init_10_13_3_1
void FASTCALL call_10_13_3_1(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(3<<4)+1);
}
#define init_10_13_3_2
void FASTCALL call_10_13_3_2(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(3<<4)+2);
}
#define init_10_13_3_3
void FASTCALL call_10_13_3_3(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(3<<4)+3);
}
#define init_10_13_3_4
void FASTCALL call_10_13_3_4(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(3<<4)+4);
}
#define init_10_13_3_5
void FASTCALL call_10_13_3_5(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(3<<4)+5);
}
#define init_10_13_3_6
void FASTCALL call_10_13_3_6(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(3<<4)+6);
}
#define init_10_13_3_7
void FASTCALL call_10_13_3_7(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(3<<4)+7);
}
#define init_10_13_3_8
void FASTCALL call_10_13_3_8(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(3<<4)+8);
}
#define init_10_13_3_9
void FASTCALL call_10_13_3_9(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(3<<4)+9);
}
#define init_10_13_3_10
void FASTCALL call_10_13_3_10(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(3<<4)+10);
}
#define init_10_13_3_11
void FASTCALL call_10_13_3_11(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(3<<4)+11);
}
#define init_10_13_3_12
void FASTCALL call_10_13_3_12(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(3<<4)+12);
}
#define init_10_13_3_13
void FASTCALL call_10_13_3_13(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(3<<4)+13);
}
#define init_10_13_3_14
void FASTCALL call_10_13_3_14(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(3<<4)+14);
}
#define init_10_13_3_15
void FASTCALL call_10_13_3_15(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(3<<4)+15);
}
#define init_10_13_4_0
void FASTCALL call_10_13_4_0(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(4<<4)+0);
}
#define init_10_13_4_1
void FASTCALL call_10_13_4_1(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(4<<4)+1);
}
#define init_10_13_4_2
void FASTCALL call_10_13_4_2(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(4<<4)+2);
}
#define init_10_13_4_3
void FASTCALL call_10_13_4_3(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(4<<4)+3);
}
#define init_10_13_4_4
void FASTCALL call_10_13_4_4(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(4<<4)+4);
}
#define init_10_13_4_5
void FASTCALL call_10_13_4_5(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(4<<4)+5);
}
#define init_10_13_4_6
void FASTCALL call_10_13_4_6(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(4<<4)+6);
}
#define init_10_13_4_7
void FASTCALL call_10_13_4_7(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(4<<4)+7);
}
#define init_10_13_4_8
void FASTCALL call_10_13_4_8(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(4<<4)+8);
}
#define init_10_13_4_9
void FASTCALL call_10_13_4_9(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(4<<4)+9);
}
#define init_10_13_4_10
void FASTCALL call_10_13_4_10(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(4<<4)+10);
}
#define init_10_13_4_11
void FASTCALL call_10_13_4_11(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(4<<4)+11);
}
#define init_10_13_4_12
void FASTCALL call_10_13_4_12(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(4<<4)+12);
}
#define init_10_13_4_13
void FASTCALL call_10_13_4_13(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(4<<4)+13);
}
#define init_10_13_4_14
void FASTCALL call_10_13_4_14(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(4<<4)+14);
}
#define init_10_13_4_15
void FASTCALL call_10_13_4_15(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(4<<4)+15);
}
#define init_10_13_5_0
void FASTCALL call_10_13_5_0(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(5<<4)+0);
}
#define init_10_13_5_1
void FASTCALL call_10_13_5_1(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(5<<4)+1);
}
#define init_10_13_5_2
void FASTCALL call_10_13_5_2(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(5<<4)+2);
}
#define init_10_13_5_3
void FASTCALL call_10_13_5_3(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(5<<4)+3);
}
#define init_10_13_5_4
void FASTCALL call_10_13_5_4(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(5<<4)+4);
}
#define init_10_13_5_5
void FASTCALL call_10_13_5_5(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(5<<4)+5);
}
#define init_10_13_5_6
void FASTCALL call_10_13_5_6(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(5<<4)+6);
}
#define init_10_13_5_7
void FASTCALL call_10_13_5_7(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(5<<4)+7);
}
#define init_10_13_5_8
void FASTCALL call_10_13_5_8(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(5<<4)+8);
}
#define init_10_13_5_9
void FASTCALL call_10_13_5_9(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(5<<4)+9);
}
#define init_10_13_5_10
void FASTCALL call_10_13_5_10(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(5<<4)+10);
}
#define init_10_13_5_11
void FASTCALL call_10_13_5_11(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(5<<4)+11);
}
#define init_10_13_5_12
void FASTCALL call_10_13_5_12(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(5<<4)+12);
}
#define init_10_13_5_13
void FASTCALL call_10_13_5_13(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(5<<4)+13);
}
#define init_10_13_5_14
void FASTCALL call_10_13_5_14(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(5<<4)+14);
}
#define init_10_13_5_15
void FASTCALL call_10_13_5_15(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(5<<4)+15);
}
#define init_10_13_6_0
void FASTCALL call_10_13_6_0(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(6<<4)+0);
}
#define init_10_13_6_1
void FASTCALL call_10_13_6_1(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(6<<4)+1);
}
#define init_10_13_6_2
void FASTCALL call_10_13_6_2(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(6<<4)+2);
}
#define init_10_13_6_3
void FASTCALL call_10_13_6_3(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(6<<4)+3);
}
#define init_10_13_6_4
void FASTCALL call_10_13_6_4(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(6<<4)+4);
}
#define init_10_13_6_5
void FASTCALL call_10_13_6_5(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(6<<4)+5);
}
#define init_10_13_6_6
void FASTCALL call_10_13_6_6(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(6<<4)+6);
}
#define init_10_13_6_7
void FASTCALL call_10_13_6_7(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(6<<4)+7);
}
#define init_10_13_6_8
void FASTCALL call_10_13_6_8(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(6<<4)+8);
}
#define init_10_13_6_9
void FASTCALL call_10_13_6_9(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(6<<4)+9);
}
#define init_10_13_6_10
void FASTCALL call_10_13_6_10(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(6<<4)+10);
}
#define init_10_13_6_11
void FASTCALL call_10_13_6_11(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(6<<4)+11);
}
#define init_10_13_6_12
void FASTCALL call_10_13_6_12(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(6<<4)+12);
}
#define init_10_13_6_13
void FASTCALL call_10_13_6_13(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(6<<4)+13);
}
#define init_10_13_6_14
void FASTCALL call_10_13_6_14(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(6<<4)+14);
}
#define init_10_13_6_15
void FASTCALL call_10_13_6_15(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(6<<4)+15);
}
#define init_10_13_7_0
void FASTCALL call_10_13_7_0(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(7<<4)+0);
}
#define init_10_13_7_1
void FASTCALL call_10_13_7_1(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(7<<4)+1);
}
#define init_10_13_7_2
void FASTCALL call_10_13_7_2(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(7<<4)+2);
}
#define init_10_13_7_3
void FASTCALL call_10_13_7_3(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(7<<4)+3);
}
#define init_10_13_7_4
void FASTCALL call_10_13_7_4(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(7<<4)+4);
}
#define init_10_13_7_5
void FASTCALL call_10_13_7_5(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(7<<4)+5);
}
#define init_10_13_7_6
void FASTCALL call_10_13_7_6(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(7<<4)+6);
}
#define init_10_13_7_7
void FASTCALL call_10_13_7_7(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(7<<4)+7);
}
#define init_10_13_7_8
void FASTCALL call_10_13_7_8(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(7<<4)+8);
}
#define init_10_13_7_9
void FASTCALL call_10_13_7_9(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(7<<4)+9);
}
#define init_10_13_7_10
void FASTCALL call_10_13_7_10(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(7<<4)+10);
}
#define init_10_13_7_11
void FASTCALL call_10_13_7_11(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(7<<4)+11);
}
#define init_10_13_7_12
void FASTCALL call_10_13_7_12(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(7<<4)+12);
}
#define init_10_13_7_13
void FASTCALL call_10_13_7_13(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(7<<4)+13);
}
#define init_10_13_7_14
void FASTCALL call_10_13_7_14(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(7<<4)+14);
}
#define init_10_13_7_15
void FASTCALL call_10_13_7_15(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(7<<4)+15);
}
#define init_10_13_8_0
void FASTCALL call_10_13_8_0(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(8<<4)+0);
}
#define init_10_13_8_1
void FASTCALL call_10_13_8_1(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(8<<4)+1);
}
#define init_10_13_8_2
void FASTCALL call_10_13_8_2(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(8<<4)+2);
}
#define init_10_13_8_3
void FASTCALL call_10_13_8_3(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(8<<4)+3);
}
#define init_10_13_8_4
void FASTCALL call_10_13_8_4(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(8<<4)+4);
}
#define init_10_13_8_5
void FASTCALL call_10_13_8_5(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(8<<4)+5);
}
#define init_10_13_8_6
void FASTCALL call_10_13_8_6(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(8<<4)+6);
}
#define init_10_13_8_7
void FASTCALL call_10_13_8_7(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(8<<4)+7);
}
#define init_10_13_8_8
void FASTCALL call_10_13_8_8(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(8<<4)+8);
}
#define init_10_13_8_9
void FASTCALL call_10_13_8_9(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(8<<4)+9);
}
#define init_10_13_8_10
void FASTCALL call_10_13_8_10(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(8<<4)+10);
}
#define init_10_13_8_11
void FASTCALL call_10_13_8_11(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(8<<4)+11);
}
#define init_10_13_8_12
void FASTCALL call_10_13_8_12(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(8<<4)+12);
}
#define init_10_13_8_13
void FASTCALL call_10_13_8_13(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(8<<4)+13);
}
#define init_10_13_8_14
void FASTCALL call_10_13_8_14(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(8<<4)+14);
}
#define init_10_13_8_15
void FASTCALL call_10_13_8_15(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(8<<4)+15);
}
#define init_10_13_9_0
void FASTCALL call_10_13_9_0(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(9<<4)+0);
}
#define init_10_13_9_1
void FASTCALL call_10_13_9_1(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(9<<4)+1);
}
#define init_10_13_9_2
void FASTCALL call_10_13_9_2(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(9<<4)+2);
}
#define init_10_13_9_3
void FASTCALL call_10_13_9_3(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(9<<4)+3);
}
#define init_10_13_9_4
void FASTCALL call_10_13_9_4(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(9<<4)+4);
}
#define init_10_13_9_5
void FASTCALL call_10_13_9_5(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(9<<4)+5);
}
#define init_10_13_9_6
void FASTCALL call_10_13_9_6(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(9<<4)+6);
}
#define init_10_13_9_7
void FASTCALL call_10_13_9_7(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(9<<4)+7);
}
#define init_10_13_9_8
void FASTCALL call_10_13_9_8(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(9<<4)+8);
}
#define init_10_13_9_9
void FASTCALL call_10_13_9_9(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(9<<4)+9);
}
#define init_10_13_9_10
void FASTCALL call_10_13_9_10(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(9<<4)+10);
}
#define init_10_13_9_11
void FASTCALL call_10_13_9_11(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(9<<4)+11);
}
#define init_10_13_9_12
void FASTCALL call_10_13_9_12(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(9<<4)+12);
}
#define init_10_13_9_13
void FASTCALL call_10_13_9_13(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(9<<4)+13);
}
#define init_10_13_9_14
void FASTCALL call_10_13_9_14(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(9<<4)+14);
}
#define init_10_13_9_15
void FASTCALL call_10_13_9_15(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(9<<4)+15);
}
#define init_10_13_10_0
void FASTCALL call_10_13_10_0(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(10<<4)+0);
}
#define init_10_13_10_1
void FASTCALL call_10_13_10_1(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(10<<4)+1);
}
#define init_10_13_10_2
void FASTCALL call_10_13_10_2(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(10<<4)+2);
}
#define init_10_13_10_3
void FASTCALL call_10_13_10_3(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(10<<4)+3);
}
#define init_10_13_10_4
void FASTCALL call_10_13_10_4(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(10<<4)+4);
}
#define init_10_13_10_5
void FASTCALL call_10_13_10_5(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(10<<4)+5);
}
#define init_10_13_10_6
void FASTCALL call_10_13_10_6(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(10<<4)+6);
}
#define init_10_13_10_7
void FASTCALL call_10_13_10_7(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(10<<4)+7);
}
#define init_10_13_10_8
void FASTCALL call_10_13_10_8(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(10<<4)+8);
}
#define init_10_13_10_9
void FASTCALL call_10_13_10_9(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(10<<4)+9);
}
#define init_10_13_10_10
void FASTCALL call_10_13_10_10(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(10<<4)+10);
}
#define init_10_13_10_11
void FASTCALL call_10_13_10_11(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(10<<4)+11);
}
#define init_10_13_10_12
void FASTCALL call_10_13_10_12(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(10<<4)+12);
}
#define init_10_13_10_13
void FASTCALL call_10_13_10_13(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(10<<4)+13);
}
#define init_10_13_10_14
void FASTCALL call_10_13_10_14(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(10<<4)+14);
}
#define init_10_13_10_15
void FASTCALL call_10_13_10_15(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(10<<4)+15);
}
#define init_10_13_11_0
void FASTCALL call_10_13_11_0(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(11<<4)+0);
}
#define init_10_13_11_1
void FASTCALL call_10_13_11_1(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(11<<4)+1);
}
#define init_10_13_11_2
void FASTCALL call_10_13_11_2(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(11<<4)+2);
}
#define init_10_13_11_3
void FASTCALL call_10_13_11_3(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(11<<4)+3);
}
#define init_10_13_11_4
void FASTCALL call_10_13_11_4(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(11<<4)+4);
}
#define init_10_13_11_5
void FASTCALL call_10_13_11_5(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(11<<4)+5);
}
#define init_10_13_11_6
void FASTCALL call_10_13_11_6(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(11<<4)+6);
}
#define init_10_13_11_7
void FASTCALL call_10_13_11_7(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(11<<4)+7);
}
#define init_10_13_11_8
void FASTCALL call_10_13_11_8(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(11<<4)+8);
}
#define init_10_13_11_9
void FASTCALL call_10_13_11_9(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(11<<4)+9);
}
#define init_10_13_11_10
void FASTCALL call_10_13_11_10(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(11<<4)+10);
}
#define init_10_13_11_11
void FASTCALL call_10_13_11_11(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(11<<4)+11);
}
#define init_10_13_11_12
void FASTCALL call_10_13_11_12(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(11<<4)+12);
}
#define init_10_13_11_13
void FASTCALL call_10_13_11_13(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(11<<4)+13);
}
#define init_10_13_11_14
void FASTCALL call_10_13_11_14(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(11<<4)+14);
}
#define init_10_13_11_15
void FASTCALL call_10_13_11_15(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(11<<4)+15);
}
#define init_10_13_12_0
void FASTCALL call_10_13_12_0(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(12<<4)+0);
}
#define init_10_13_12_1
void FASTCALL call_10_13_12_1(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(12<<4)+1);
}
#define init_10_13_12_2
void FASTCALL call_10_13_12_2(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(12<<4)+2);
}
#define init_10_13_12_3
void FASTCALL call_10_13_12_3(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(12<<4)+3);
}
#define init_10_13_12_4
void FASTCALL call_10_13_12_4(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(12<<4)+4);
}
#define init_10_13_12_5
void FASTCALL call_10_13_12_5(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(12<<4)+5);
}
#define init_10_13_12_6
void FASTCALL call_10_13_12_6(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(12<<4)+6);
}
#define init_10_13_12_7
void FASTCALL call_10_13_12_7(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(12<<4)+7);
}
#define init_10_13_12_8
void FASTCALL call_10_13_12_8(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(12<<4)+8);
}
#define init_10_13_12_9
void FASTCALL call_10_13_12_9(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(12<<4)+9);
}
#define init_10_13_12_10
void FASTCALL call_10_13_12_10(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(12<<4)+10);
}
#define init_10_13_12_11
void FASTCALL call_10_13_12_11(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(12<<4)+11);
}
#define init_10_13_12_12
void FASTCALL call_10_13_12_12(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(12<<4)+12);
}
#define init_10_13_12_13
void FASTCALL call_10_13_12_13(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(12<<4)+13);
}
#define init_10_13_12_14
void FASTCALL call_10_13_12_14(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(12<<4)+14);
}
#define init_10_13_12_15
void FASTCALL call_10_13_12_15(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(12<<4)+15);
}
#define init_10_13_13_0
void FASTCALL call_10_13_13_0(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(13<<4)+0);
}
#define init_10_13_13_1
void FASTCALL call_10_13_13_1(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(13<<4)+1);
}
#define init_10_13_13_2
void FASTCALL call_10_13_13_2(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(13<<4)+2);
}
#define init_10_13_13_3
void FASTCALL call_10_13_13_3(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(13<<4)+3);
}
#define init_10_13_13_4
void FASTCALL call_10_13_13_4(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(13<<4)+4);
}
#define init_10_13_13_5
void FASTCALL call_10_13_13_5(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(13<<4)+5);
}
#define init_10_13_13_6
void FASTCALL call_10_13_13_6(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(13<<4)+6);
}
#define init_10_13_13_7
void FASTCALL call_10_13_13_7(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(13<<4)+7);
}
#define init_10_13_13_8
void FASTCALL call_10_13_13_8(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(13<<4)+8);
}
#define init_10_13_13_9
void FASTCALL call_10_13_13_9(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(13<<4)+9);
}
#define init_10_13_13_10
void FASTCALL call_10_13_13_10(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(13<<4)+10);
}
#define init_10_13_13_11
void FASTCALL call_10_13_13_11(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(13<<4)+11);
}
#define init_10_13_13_12
void FASTCALL call_10_13_13_12(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(13<<4)+12);
}
#define init_10_13_13_13
void FASTCALL call_10_13_13_13(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(13<<4)+13);
}
#define init_10_13_13_14
void FASTCALL call_10_13_13_14(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(13<<4)+14);
}
#define init_10_13_13_15
void FASTCALL call_10_13_13_15(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(13<<4)+15);
}
#define init_10_13_14_0
void FASTCALL call_10_13_14_0(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(14<<4)+0);
}
#define init_10_13_14_1
void FASTCALL call_10_13_14_1(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(14<<4)+1);
}
#define init_10_13_14_2
void FASTCALL call_10_13_14_2(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(14<<4)+2);
}
#define init_10_13_14_3
void FASTCALL call_10_13_14_3(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(14<<4)+3);
}
#define init_10_13_14_4
void FASTCALL call_10_13_14_4(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(14<<4)+4);
}
#define init_10_13_14_5
void FASTCALL call_10_13_14_5(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(14<<4)+5);
}
#define init_10_13_14_6
void FASTCALL call_10_13_14_6(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(14<<4)+6);
}
#define init_10_13_14_7
void FASTCALL call_10_13_14_7(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(14<<4)+7);
}
#define init_10_13_14_8
void FASTCALL call_10_13_14_8(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(14<<4)+8);
}
#define init_10_13_14_9
void FASTCALL call_10_13_14_9(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(14<<4)+9);
}
#define init_10_13_14_10
void FASTCALL call_10_13_14_10(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(14<<4)+10);
}
#define init_10_13_14_11
void FASTCALL call_10_13_14_11(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(14<<4)+11);
}
#define init_10_13_14_12
void FASTCALL call_10_13_14_12(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(14<<4)+12);
}
#define init_10_13_14_13
void FASTCALL call_10_13_14_13(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(14<<4)+13);
}
#define init_10_13_14_14
void FASTCALL call_10_13_14_14(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(14<<4)+14);
}
#define init_10_13_14_15
void FASTCALL call_10_13_14_15(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(14<<4)+15);
}
#define init_10_13_15_0
void FASTCALL call_10_13_15_0(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(15<<4)+0);
}
#define init_10_13_15_1
void FASTCALL call_10_13_15_1(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(15<<4)+1);
}
#define init_10_13_15_2
void FASTCALL call_10_13_15_2(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(15<<4)+2);
}
#define init_10_13_15_3
void FASTCALL call_10_13_15_3(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(15<<4)+3);
}
#define init_10_13_15_4
void FASTCALL call_10_13_15_4(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(15<<4)+4);
}
#define init_10_13_15_5
void FASTCALL call_10_13_15_5(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(15<<4)+5);
}
#define init_10_13_15_6
void FASTCALL call_10_13_15_6(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(15<<4)+6);
}
#define init_10_13_15_7
void FASTCALL call_10_13_15_7(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(15<<4)+7);
}
#define init_10_13_15_8
void FASTCALL call_10_13_15_8(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(15<<4)+8);
}
#define init_10_13_15_9
void FASTCALL call_10_13_15_9(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(15<<4)+9);
}
#define init_10_13_15_10
void FASTCALL call_10_13_15_10(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(15<<4)+10);
}
#define init_10_13_15_11
void FASTCALL call_10_13_15_11(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(15<<4)+11);
}
#define init_10_13_15_12
void FASTCALL call_10_13_15_12(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(15<<4)+12);
}
#define init_10_13_15_13
void FASTCALL call_10_13_15_13(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(15<<4)+13);
}
#define init_10_13_15_14
void FASTCALL call_10_13_15_14(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(15<<4)+14);
}
#define init_10_13_15_15
void FASTCALL call_10_13_15_15(SH2_struct* sh)
{
  SH2bra(sh, (13<<8)+(15<<4)+15);
}
#define init_10_14_0_0
void FASTCALL call_10_14_0_0(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(0<<4)+0);
}
#define init_10_14_0_1
void FASTCALL call_10_14_0_1(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(0<<4)+1);
}
#define init_10_14_0_2
void FASTCALL call_10_14_0_2(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(0<<4)+2);
}
#define init_10_14_0_3
void FASTCALL call_10_14_0_3(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(0<<4)+3);
}
#define init_10_14_0_4
void FASTCALL call_10_14_0_4(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(0<<4)+4);
}
#define init_10_14_0_5
void FASTCALL call_10_14_0_5(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(0<<4)+5);
}
#define init_10_14_0_6
void FASTCALL call_10_14_0_6(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(0<<4)+6);
}
#define init_10_14_0_7
void FASTCALL call_10_14_0_7(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(0<<4)+7);
}
#define init_10_14_0_8
void FASTCALL call_10_14_0_8(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(0<<4)+8);
}
#define init_10_14_0_9
void FASTCALL call_10_14_0_9(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(0<<4)+9);
}
#define init_10_14_0_10
void FASTCALL call_10_14_0_10(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(0<<4)+10);
}
#define init_10_14_0_11
void FASTCALL call_10_14_0_11(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(0<<4)+11);
}
#define init_10_14_0_12
void FASTCALL call_10_14_0_12(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(0<<4)+12);
}
#define init_10_14_0_13
void FASTCALL call_10_14_0_13(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(0<<4)+13);
}
#define init_10_14_0_14
void FASTCALL call_10_14_0_14(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(0<<4)+14);
}
#define init_10_14_0_15
void FASTCALL call_10_14_0_15(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(0<<4)+15);
}
#define init_10_14_1_0
void FASTCALL call_10_14_1_0(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(1<<4)+0);
}
#define init_10_14_1_1
void FASTCALL call_10_14_1_1(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(1<<4)+1);
}
#define init_10_14_1_2
void FASTCALL call_10_14_1_2(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(1<<4)+2);
}
#define init_10_14_1_3
void FASTCALL call_10_14_1_3(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(1<<4)+3);
}
#define init_10_14_1_4
void FASTCALL call_10_14_1_4(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(1<<4)+4);
}
#define init_10_14_1_5
void FASTCALL call_10_14_1_5(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(1<<4)+5);
}
#define init_10_14_1_6
void FASTCALL call_10_14_1_6(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(1<<4)+6);
}
#define init_10_14_1_7
void FASTCALL call_10_14_1_7(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(1<<4)+7);
}
#define init_10_14_1_8
void FASTCALL call_10_14_1_8(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(1<<4)+8);
}
#define init_10_14_1_9
void FASTCALL call_10_14_1_9(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(1<<4)+9);
}
#define init_10_14_1_10
void FASTCALL call_10_14_1_10(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(1<<4)+10);
}
#define init_10_14_1_11
void FASTCALL call_10_14_1_11(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(1<<4)+11);
}
#define init_10_14_1_12
void FASTCALL call_10_14_1_12(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(1<<4)+12);
}
#define init_10_14_1_13
void FASTCALL call_10_14_1_13(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(1<<4)+13);
}
#define init_10_14_1_14
void FASTCALL call_10_14_1_14(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(1<<4)+14);
}
#define init_10_14_1_15
void FASTCALL call_10_14_1_15(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(1<<4)+15);
}
#define init_10_14_2_0
void FASTCALL call_10_14_2_0(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(2<<4)+0);
}
#define init_10_14_2_1
void FASTCALL call_10_14_2_1(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(2<<4)+1);
}
#define init_10_14_2_2
void FASTCALL call_10_14_2_2(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(2<<4)+2);
}
#define init_10_14_2_3
void FASTCALL call_10_14_2_3(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(2<<4)+3);
}
#define init_10_14_2_4
void FASTCALL call_10_14_2_4(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(2<<4)+4);
}
#define init_10_14_2_5
void FASTCALL call_10_14_2_5(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(2<<4)+5);
}
#define init_10_14_2_6
void FASTCALL call_10_14_2_6(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(2<<4)+6);
}
#define init_10_14_2_7
void FASTCALL call_10_14_2_7(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(2<<4)+7);
}
#define init_10_14_2_8
void FASTCALL call_10_14_2_8(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(2<<4)+8);
}
#define init_10_14_2_9
void FASTCALL call_10_14_2_9(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(2<<4)+9);
}
#define init_10_14_2_10
void FASTCALL call_10_14_2_10(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(2<<4)+10);
}
#define init_10_14_2_11
void FASTCALL call_10_14_2_11(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(2<<4)+11);
}
#define init_10_14_2_12
void FASTCALL call_10_14_2_12(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(2<<4)+12);
}
#define init_10_14_2_13
void FASTCALL call_10_14_2_13(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(2<<4)+13);
}
#define init_10_14_2_14
void FASTCALL call_10_14_2_14(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(2<<4)+14);
}
#define init_10_14_2_15
void FASTCALL call_10_14_2_15(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(2<<4)+15);
}
#define init_10_14_3_0
void FASTCALL call_10_14_3_0(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(3<<4)+0);
}
#define init_10_14_3_1
void FASTCALL call_10_14_3_1(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(3<<4)+1);
}
#define init_10_14_3_2
void FASTCALL call_10_14_3_2(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(3<<4)+2);
}
#define init_10_14_3_3
void FASTCALL call_10_14_3_3(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(3<<4)+3);
}
#define init_10_14_3_4
void FASTCALL call_10_14_3_4(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(3<<4)+4);
}
#define init_10_14_3_5
void FASTCALL call_10_14_3_5(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(3<<4)+5);
}
#define init_10_14_3_6
void FASTCALL call_10_14_3_6(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(3<<4)+6);
}
#define init_10_14_3_7
void FASTCALL call_10_14_3_7(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(3<<4)+7);
}
#define init_10_14_3_8
void FASTCALL call_10_14_3_8(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(3<<4)+8);
}
#define init_10_14_3_9
void FASTCALL call_10_14_3_9(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(3<<4)+9);
}
#define init_10_14_3_10
void FASTCALL call_10_14_3_10(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(3<<4)+10);
}
#define init_10_14_3_11
void FASTCALL call_10_14_3_11(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(3<<4)+11);
}
#define init_10_14_3_12
void FASTCALL call_10_14_3_12(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(3<<4)+12);
}
#define init_10_14_3_13
void FASTCALL call_10_14_3_13(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(3<<4)+13);
}
#define init_10_14_3_14
void FASTCALL call_10_14_3_14(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(3<<4)+14);
}
#define init_10_14_3_15
void FASTCALL call_10_14_3_15(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(3<<4)+15);
}
#define init_10_14_4_0
void FASTCALL call_10_14_4_0(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(4<<4)+0);
}
#define init_10_14_4_1
void FASTCALL call_10_14_4_1(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(4<<4)+1);
}
#define init_10_14_4_2
void FASTCALL call_10_14_4_2(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(4<<4)+2);
}
#define init_10_14_4_3
void FASTCALL call_10_14_4_3(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(4<<4)+3);
}
#define init_10_14_4_4
void FASTCALL call_10_14_4_4(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(4<<4)+4);
}
#define init_10_14_4_5
void FASTCALL call_10_14_4_5(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(4<<4)+5);
}
#define init_10_14_4_6
void FASTCALL call_10_14_4_6(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(4<<4)+6);
}
#define init_10_14_4_7
void FASTCALL call_10_14_4_7(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(4<<4)+7);
}
#define init_10_14_4_8
void FASTCALL call_10_14_4_8(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(4<<4)+8);
}
#define init_10_14_4_9
void FASTCALL call_10_14_4_9(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(4<<4)+9);
}
#define init_10_14_4_10
void FASTCALL call_10_14_4_10(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(4<<4)+10);
}
#define init_10_14_4_11
void FASTCALL call_10_14_4_11(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(4<<4)+11);
}
#define init_10_14_4_12
void FASTCALL call_10_14_4_12(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(4<<4)+12);
}
#define init_10_14_4_13
void FASTCALL call_10_14_4_13(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(4<<4)+13);
}
#define init_10_14_4_14
void FASTCALL call_10_14_4_14(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(4<<4)+14);
}
#define init_10_14_4_15
void FASTCALL call_10_14_4_15(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(4<<4)+15);
}
#define init_10_14_5_0
void FASTCALL call_10_14_5_0(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(5<<4)+0);
}
#define init_10_14_5_1
void FASTCALL call_10_14_5_1(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(5<<4)+1);
}
#define init_10_14_5_2
void FASTCALL call_10_14_5_2(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(5<<4)+2);
}
#define init_10_14_5_3
void FASTCALL call_10_14_5_3(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(5<<4)+3);
}
#define init_10_14_5_4
void FASTCALL call_10_14_5_4(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(5<<4)+4);
}
#define init_10_14_5_5
void FASTCALL call_10_14_5_5(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(5<<4)+5);
}
#define init_10_14_5_6
void FASTCALL call_10_14_5_6(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(5<<4)+6);
}
#define init_10_14_5_7
void FASTCALL call_10_14_5_7(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(5<<4)+7);
}
#define init_10_14_5_8
void FASTCALL call_10_14_5_8(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(5<<4)+8);
}
#define init_10_14_5_9
void FASTCALL call_10_14_5_9(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(5<<4)+9);
}
#define init_10_14_5_10
void FASTCALL call_10_14_5_10(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(5<<4)+10);
}
#define init_10_14_5_11
void FASTCALL call_10_14_5_11(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(5<<4)+11);
}
#define init_10_14_5_12
void FASTCALL call_10_14_5_12(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(5<<4)+12);
}
#define init_10_14_5_13
void FASTCALL call_10_14_5_13(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(5<<4)+13);
}
#define init_10_14_5_14
void FASTCALL call_10_14_5_14(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(5<<4)+14);
}
#define init_10_14_5_15
void FASTCALL call_10_14_5_15(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(5<<4)+15);
}
#define init_10_14_6_0
void FASTCALL call_10_14_6_0(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(6<<4)+0);
}
#define init_10_14_6_1
void FASTCALL call_10_14_6_1(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(6<<4)+1);
}
#define init_10_14_6_2
void FASTCALL call_10_14_6_2(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(6<<4)+2);
}
#define init_10_14_6_3
void FASTCALL call_10_14_6_3(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(6<<4)+3);
}
#define init_10_14_6_4
void FASTCALL call_10_14_6_4(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(6<<4)+4);
}
#define init_10_14_6_5
void FASTCALL call_10_14_6_5(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(6<<4)+5);
}
#define init_10_14_6_6
void FASTCALL call_10_14_6_6(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(6<<4)+6);
}
#define init_10_14_6_7
void FASTCALL call_10_14_6_7(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(6<<4)+7);
}
#define init_10_14_6_8
void FASTCALL call_10_14_6_8(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(6<<4)+8);
}
#define init_10_14_6_9
void FASTCALL call_10_14_6_9(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(6<<4)+9);
}
#define init_10_14_6_10
void FASTCALL call_10_14_6_10(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(6<<4)+10);
}
#define init_10_14_6_11
void FASTCALL call_10_14_6_11(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(6<<4)+11);
}
#define init_10_14_6_12
void FASTCALL call_10_14_6_12(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(6<<4)+12);
}
#define init_10_14_6_13
void FASTCALL call_10_14_6_13(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(6<<4)+13);
}
#define init_10_14_6_14
void FASTCALL call_10_14_6_14(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(6<<4)+14);
}
#define init_10_14_6_15
void FASTCALL call_10_14_6_15(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(6<<4)+15);
}
#define init_10_14_7_0
void FASTCALL call_10_14_7_0(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(7<<4)+0);
}
#define init_10_14_7_1
void FASTCALL call_10_14_7_1(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(7<<4)+1);
}
#define init_10_14_7_2
void FASTCALL call_10_14_7_2(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(7<<4)+2);
}
#define init_10_14_7_3
void FASTCALL call_10_14_7_3(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(7<<4)+3);
}
#define init_10_14_7_4
void FASTCALL call_10_14_7_4(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(7<<4)+4);
}
#define init_10_14_7_5
void FASTCALL call_10_14_7_5(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(7<<4)+5);
}
#define init_10_14_7_6
void FASTCALL call_10_14_7_6(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(7<<4)+6);
}
#define init_10_14_7_7
void FASTCALL call_10_14_7_7(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(7<<4)+7);
}
#define init_10_14_7_8
void FASTCALL call_10_14_7_8(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(7<<4)+8);
}
#define init_10_14_7_9
void FASTCALL call_10_14_7_9(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(7<<4)+9);
}
#define init_10_14_7_10
void FASTCALL call_10_14_7_10(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(7<<4)+10);
}
#define init_10_14_7_11
void FASTCALL call_10_14_7_11(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(7<<4)+11);
}
#define init_10_14_7_12
void FASTCALL call_10_14_7_12(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(7<<4)+12);
}
#define init_10_14_7_13
void FASTCALL call_10_14_7_13(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(7<<4)+13);
}
#define init_10_14_7_14
void FASTCALL call_10_14_7_14(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(7<<4)+14);
}
#define init_10_14_7_15
void FASTCALL call_10_14_7_15(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(7<<4)+15);
}
#define init_10_14_8_0
void FASTCALL call_10_14_8_0(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(8<<4)+0);
}
#define init_10_14_8_1
void FASTCALL call_10_14_8_1(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(8<<4)+1);
}
#define init_10_14_8_2
void FASTCALL call_10_14_8_2(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(8<<4)+2);
}
#define init_10_14_8_3
void FASTCALL call_10_14_8_3(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(8<<4)+3);
}
#define init_10_14_8_4
void FASTCALL call_10_14_8_4(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(8<<4)+4);
}
#define init_10_14_8_5
void FASTCALL call_10_14_8_5(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(8<<4)+5);
}
#define init_10_14_8_6
void FASTCALL call_10_14_8_6(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(8<<4)+6);
}
#define init_10_14_8_7
void FASTCALL call_10_14_8_7(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(8<<4)+7);
}
#define init_10_14_8_8
void FASTCALL call_10_14_8_8(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(8<<4)+8);
}
#define init_10_14_8_9
void FASTCALL call_10_14_8_9(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(8<<4)+9);
}
#define init_10_14_8_10
void FASTCALL call_10_14_8_10(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(8<<4)+10);
}
#define init_10_14_8_11
void FASTCALL call_10_14_8_11(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(8<<4)+11);
}
#define init_10_14_8_12
void FASTCALL call_10_14_8_12(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(8<<4)+12);
}
#define init_10_14_8_13
void FASTCALL call_10_14_8_13(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(8<<4)+13);
}
#define init_10_14_8_14
void FASTCALL call_10_14_8_14(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(8<<4)+14);
}
#define init_10_14_8_15
void FASTCALL call_10_14_8_15(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(8<<4)+15);
}
#define init_10_14_9_0
void FASTCALL call_10_14_9_0(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(9<<4)+0);
}
#define init_10_14_9_1
void FASTCALL call_10_14_9_1(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(9<<4)+1);
}
#define init_10_14_9_2
void FASTCALL call_10_14_9_2(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(9<<4)+2);
}
#define init_10_14_9_3
void FASTCALL call_10_14_9_3(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(9<<4)+3);
}
#define init_10_14_9_4
void FASTCALL call_10_14_9_4(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(9<<4)+4);
}
#define init_10_14_9_5
void FASTCALL call_10_14_9_5(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(9<<4)+5);
}
#define init_10_14_9_6
void FASTCALL call_10_14_9_6(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(9<<4)+6);
}
#define init_10_14_9_7
void FASTCALL call_10_14_9_7(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(9<<4)+7);
}
#define init_10_14_9_8
void FASTCALL call_10_14_9_8(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(9<<4)+8);
}
#define init_10_14_9_9
void FASTCALL call_10_14_9_9(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(9<<4)+9);
}
#define init_10_14_9_10
void FASTCALL call_10_14_9_10(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(9<<4)+10);
}
#define init_10_14_9_11
void FASTCALL call_10_14_9_11(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(9<<4)+11);
}
#define init_10_14_9_12
void FASTCALL call_10_14_9_12(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(9<<4)+12);
}
#define init_10_14_9_13
void FASTCALL call_10_14_9_13(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(9<<4)+13);
}
#define init_10_14_9_14
void FASTCALL call_10_14_9_14(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(9<<4)+14);
}
#define init_10_14_9_15
void FASTCALL call_10_14_9_15(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(9<<4)+15);
}
#define init_10_14_10_0
void FASTCALL call_10_14_10_0(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(10<<4)+0);
}
#define init_10_14_10_1
void FASTCALL call_10_14_10_1(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(10<<4)+1);
}
#define init_10_14_10_2
void FASTCALL call_10_14_10_2(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(10<<4)+2);
}
#define init_10_14_10_3
void FASTCALL call_10_14_10_3(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(10<<4)+3);
}
#define init_10_14_10_4
void FASTCALL call_10_14_10_4(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(10<<4)+4);
}
#define init_10_14_10_5
void FASTCALL call_10_14_10_5(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(10<<4)+5);
}
#define init_10_14_10_6
void FASTCALL call_10_14_10_6(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(10<<4)+6);
}
#define init_10_14_10_7
void FASTCALL call_10_14_10_7(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(10<<4)+7);
}
#define init_10_14_10_8
void FASTCALL call_10_14_10_8(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(10<<4)+8);
}
#define init_10_14_10_9
void FASTCALL call_10_14_10_9(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(10<<4)+9);
}
#define init_10_14_10_10
void FASTCALL call_10_14_10_10(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(10<<4)+10);
}
#define init_10_14_10_11
void FASTCALL call_10_14_10_11(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(10<<4)+11);
}
#define init_10_14_10_12
void FASTCALL call_10_14_10_12(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(10<<4)+12);
}
#define init_10_14_10_13
void FASTCALL call_10_14_10_13(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(10<<4)+13);
}
#define init_10_14_10_14
void FASTCALL call_10_14_10_14(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(10<<4)+14);
}
#define init_10_14_10_15
void FASTCALL call_10_14_10_15(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(10<<4)+15);
}
#define init_10_14_11_0
void FASTCALL call_10_14_11_0(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(11<<4)+0);
}
#define init_10_14_11_1
void FASTCALL call_10_14_11_1(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(11<<4)+1);
}
#define init_10_14_11_2
void FASTCALL call_10_14_11_2(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(11<<4)+2);
}
#define init_10_14_11_3
void FASTCALL call_10_14_11_3(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(11<<4)+3);
}
#define init_10_14_11_4
void FASTCALL call_10_14_11_4(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(11<<4)+4);
}
#define init_10_14_11_5
void FASTCALL call_10_14_11_5(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(11<<4)+5);
}
#define init_10_14_11_6
void FASTCALL call_10_14_11_6(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(11<<4)+6);
}
#define init_10_14_11_7
void FASTCALL call_10_14_11_7(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(11<<4)+7);
}
#define init_10_14_11_8
void FASTCALL call_10_14_11_8(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(11<<4)+8);
}
#define init_10_14_11_9
void FASTCALL call_10_14_11_9(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(11<<4)+9);
}
#define init_10_14_11_10
void FASTCALL call_10_14_11_10(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(11<<4)+10);
}
#define init_10_14_11_11
void FASTCALL call_10_14_11_11(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(11<<4)+11);
}
#define init_10_14_11_12
void FASTCALL call_10_14_11_12(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(11<<4)+12);
}
#define init_10_14_11_13
void FASTCALL call_10_14_11_13(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(11<<4)+13);
}
#define init_10_14_11_14
void FASTCALL call_10_14_11_14(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(11<<4)+14);
}
#define init_10_14_11_15
void FASTCALL call_10_14_11_15(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(11<<4)+15);
}
#define init_10_14_12_0
void FASTCALL call_10_14_12_0(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(12<<4)+0);
}
#define init_10_14_12_1
void FASTCALL call_10_14_12_1(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(12<<4)+1);
}
#define init_10_14_12_2
void FASTCALL call_10_14_12_2(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(12<<4)+2);
}
#define init_10_14_12_3
void FASTCALL call_10_14_12_3(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(12<<4)+3);
}
#define init_10_14_12_4
void FASTCALL call_10_14_12_4(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(12<<4)+4);
}
#define init_10_14_12_5
void FASTCALL call_10_14_12_5(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(12<<4)+5);
}
#define init_10_14_12_6
void FASTCALL call_10_14_12_6(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(12<<4)+6);
}
#define init_10_14_12_7
void FASTCALL call_10_14_12_7(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(12<<4)+7);
}
#define init_10_14_12_8
void FASTCALL call_10_14_12_8(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(12<<4)+8);
}
#define init_10_14_12_9
void FASTCALL call_10_14_12_9(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(12<<4)+9);
}
#define init_10_14_12_10
void FASTCALL call_10_14_12_10(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(12<<4)+10);
}
#define init_10_14_12_11
void FASTCALL call_10_14_12_11(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(12<<4)+11);
}
#define init_10_14_12_12
void FASTCALL call_10_14_12_12(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(12<<4)+12);
}
#define init_10_14_12_13
void FASTCALL call_10_14_12_13(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(12<<4)+13);
}
#define init_10_14_12_14
void FASTCALL call_10_14_12_14(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(12<<4)+14);
}
#define init_10_14_12_15
void FASTCALL call_10_14_12_15(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(12<<4)+15);
}
#define init_10_14_13_0
void FASTCALL call_10_14_13_0(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(13<<4)+0);
}
#define init_10_14_13_1
void FASTCALL call_10_14_13_1(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(13<<4)+1);
}
#define init_10_14_13_2
void FASTCALL call_10_14_13_2(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(13<<4)+2);
}
#define init_10_14_13_3
void FASTCALL call_10_14_13_3(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(13<<4)+3);
}
#define init_10_14_13_4
void FASTCALL call_10_14_13_4(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(13<<4)+4);
}
#define init_10_14_13_5
void FASTCALL call_10_14_13_5(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(13<<4)+5);
}
#define init_10_14_13_6
void FASTCALL call_10_14_13_6(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(13<<4)+6);
}
#define init_10_14_13_7
void FASTCALL call_10_14_13_7(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(13<<4)+7);
}
#define init_10_14_13_8
void FASTCALL call_10_14_13_8(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(13<<4)+8);
}
#define init_10_14_13_9
void FASTCALL call_10_14_13_9(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(13<<4)+9);
}
#define init_10_14_13_10
void FASTCALL call_10_14_13_10(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(13<<4)+10);
}
#define init_10_14_13_11
void FASTCALL call_10_14_13_11(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(13<<4)+11);
}
#define init_10_14_13_12
void FASTCALL call_10_14_13_12(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(13<<4)+12);
}
#define init_10_14_13_13
void FASTCALL call_10_14_13_13(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(13<<4)+13);
}
#define init_10_14_13_14
void FASTCALL call_10_14_13_14(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(13<<4)+14);
}
#define init_10_14_13_15
void FASTCALL call_10_14_13_15(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(13<<4)+15);
}
#define init_10_14_14_0
void FASTCALL call_10_14_14_0(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(14<<4)+0);
}
#define init_10_14_14_1
void FASTCALL call_10_14_14_1(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(14<<4)+1);
}
#define init_10_14_14_2
void FASTCALL call_10_14_14_2(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(14<<4)+2);
}
#define init_10_14_14_3
void FASTCALL call_10_14_14_3(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(14<<4)+3);
}
#define init_10_14_14_4
void FASTCALL call_10_14_14_4(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(14<<4)+4);
}
#define init_10_14_14_5
void FASTCALL call_10_14_14_5(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(14<<4)+5);
}
#define init_10_14_14_6
void FASTCALL call_10_14_14_6(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(14<<4)+6);
}
#define init_10_14_14_7
void FASTCALL call_10_14_14_7(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(14<<4)+7);
}
#define init_10_14_14_8
void FASTCALL call_10_14_14_8(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(14<<4)+8);
}
#define init_10_14_14_9
void FASTCALL call_10_14_14_9(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(14<<4)+9);
}
#define init_10_14_14_10
void FASTCALL call_10_14_14_10(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(14<<4)+10);
}
#define init_10_14_14_11
void FASTCALL call_10_14_14_11(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(14<<4)+11);
}
#define init_10_14_14_12
void FASTCALL call_10_14_14_12(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(14<<4)+12);
}
#define init_10_14_14_13
void FASTCALL call_10_14_14_13(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(14<<4)+13);
}
#define init_10_14_14_14
void FASTCALL call_10_14_14_14(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(14<<4)+14);
}
#define init_10_14_14_15
void FASTCALL call_10_14_14_15(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(14<<4)+15);
}
#define init_10_14_15_0
void FASTCALL call_10_14_15_0(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(15<<4)+0);
}
#define init_10_14_15_1
void FASTCALL call_10_14_15_1(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(15<<4)+1);
}
#define init_10_14_15_2
void FASTCALL call_10_14_15_2(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(15<<4)+2);
}
#define init_10_14_15_3
void FASTCALL call_10_14_15_3(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(15<<4)+3);
}
#define init_10_14_15_4
void FASTCALL call_10_14_15_4(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(15<<4)+4);
}
#define init_10_14_15_5
void FASTCALL call_10_14_15_5(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(15<<4)+5);
}
#define init_10_14_15_6
void FASTCALL call_10_14_15_6(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(15<<4)+6);
}
#define init_10_14_15_7
void FASTCALL call_10_14_15_7(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(15<<4)+7);
}
#define init_10_14_15_8
void FASTCALL call_10_14_15_8(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(15<<4)+8);
}
#define init_10_14_15_9
void FASTCALL call_10_14_15_9(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(15<<4)+9);
}
#define init_10_14_15_10
void FASTCALL call_10_14_15_10(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(15<<4)+10);
}
#define init_10_14_15_11
void FASTCALL call_10_14_15_11(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(15<<4)+11);
}
#define init_10_14_15_12
void FASTCALL call_10_14_15_12(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(15<<4)+12);
}
#define init_10_14_15_13
void FASTCALL call_10_14_15_13(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(15<<4)+13);
}
#define init_10_14_15_14
void FASTCALL call_10_14_15_14(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(15<<4)+14);
}
#define init_10_14_15_15
void FASTCALL call_10_14_15_15(SH2_struct* sh)
{
  SH2bra(sh, (14<<8)+(15<<4)+15);
}
#define init_10_15_0_0
void FASTCALL call_10_15_0_0(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(0<<4)+0);
}
#define init_10_15_0_1
void FASTCALL call_10_15_0_1(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(0<<4)+1);
}
#define init_10_15_0_2
void FASTCALL call_10_15_0_2(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(0<<4)+2);
}
#define init_10_15_0_3
void FASTCALL call_10_15_0_3(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(0<<4)+3);
}
#define init_10_15_0_4
void FASTCALL call_10_15_0_4(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(0<<4)+4);
}
#define init_10_15_0_5
void FASTCALL call_10_15_0_5(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(0<<4)+5);
}
#define init_10_15_0_6
void FASTCALL call_10_15_0_6(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(0<<4)+6);
}
#define init_10_15_0_7
void FASTCALL call_10_15_0_7(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(0<<4)+7);
}
#define init_10_15_0_8
void FASTCALL call_10_15_0_8(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(0<<4)+8);
}
#define init_10_15_0_9
void FASTCALL call_10_15_0_9(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(0<<4)+9);
}
#define init_10_15_0_10
void FASTCALL call_10_15_0_10(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(0<<4)+10);
}
#define init_10_15_0_11
void FASTCALL call_10_15_0_11(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(0<<4)+11);
}
#define init_10_15_0_12
void FASTCALL call_10_15_0_12(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(0<<4)+12);
}
#define init_10_15_0_13
void FASTCALL call_10_15_0_13(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(0<<4)+13);
}
#define init_10_15_0_14
void FASTCALL call_10_15_0_14(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(0<<4)+14);
}
#define init_10_15_0_15
void FASTCALL call_10_15_0_15(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(0<<4)+15);
}
#define init_10_15_1_0
void FASTCALL call_10_15_1_0(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(1<<4)+0);
}
#define init_10_15_1_1
void FASTCALL call_10_15_1_1(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(1<<4)+1);
}
#define init_10_15_1_2
void FASTCALL call_10_15_1_2(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(1<<4)+2);
}
#define init_10_15_1_3
void FASTCALL call_10_15_1_3(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(1<<4)+3);
}
#define init_10_15_1_4
void FASTCALL call_10_15_1_4(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(1<<4)+4);
}
#define init_10_15_1_5
void FASTCALL call_10_15_1_5(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(1<<4)+5);
}
#define init_10_15_1_6
void FASTCALL call_10_15_1_6(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(1<<4)+6);
}
#define init_10_15_1_7
void FASTCALL call_10_15_1_7(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(1<<4)+7);
}
#define init_10_15_1_8
void FASTCALL call_10_15_1_8(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(1<<4)+8);
}
#define init_10_15_1_9
void FASTCALL call_10_15_1_9(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(1<<4)+9);
}
#define init_10_15_1_10
void FASTCALL call_10_15_1_10(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(1<<4)+10);
}
#define init_10_15_1_11
void FASTCALL call_10_15_1_11(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(1<<4)+11);
}
#define init_10_15_1_12
void FASTCALL call_10_15_1_12(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(1<<4)+12);
}
#define init_10_15_1_13
void FASTCALL call_10_15_1_13(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(1<<4)+13);
}
#define init_10_15_1_14
void FASTCALL call_10_15_1_14(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(1<<4)+14);
}
#define init_10_15_1_15
void FASTCALL call_10_15_1_15(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(1<<4)+15);
}
#define init_10_15_2_0
void FASTCALL call_10_15_2_0(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(2<<4)+0);
}
#define init_10_15_2_1
void FASTCALL call_10_15_2_1(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(2<<4)+1);
}
#define init_10_15_2_2
void FASTCALL call_10_15_2_2(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(2<<4)+2);
}
#define init_10_15_2_3
void FASTCALL call_10_15_2_3(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(2<<4)+3);
}
#define init_10_15_2_4
void FASTCALL call_10_15_2_4(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(2<<4)+4);
}
#define init_10_15_2_5
void FASTCALL call_10_15_2_5(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(2<<4)+5);
}
#define init_10_15_2_6
void FASTCALL call_10_15_2_6(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(2<<4)+6);
}
#define init_10_15_2_7
void FASTCALL call_10_15_2_7(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(2<<4)+7);
}
#define init_10_15_2_8
void FASTCALL call_10_15_2_8(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(2<<4)+8);
}
#define init_10_15_2_9
void FASTCALL call_10_15_2_9(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(2<<4)+9);
}
#define init_10_15_2_10
void FASTCALL call_10_15_2_10(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(2<<4)+10);
}
#define init_10_15_2_11
void FASTCALL call_10_15_2_11(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(2<<4)+11);
}
#define init_10_15_2_12
void FASTCALL call_10_15_2_12(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(2<<4)+12);
}
#define init_10_15_2_13
void FASTCALL call_10_15_2_13(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(2<<4)+13);
}
#define init_10_15_2_14
void FASTCALL call_10_15_2_14(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(2<<4)+14);
}
#define init_10_15_2_15
void FASTCALL call_10_15_2_15(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(2<<4)+15);
}
#define init_10_15_3_0
void FASTCALL call_10_15_3_0(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(3<<4)+0);
}
#define init_10_15_3_1
void FASTCALL call_10_15_3_1(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(3<<4)+1);
}
#define init_10_15_3_2
void FASTCALL call_10_15_3_2(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(3<<4)+2);
}
#define init_10_15_3_3
void FASTCALL call_10_15_3_3(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(3<<4)+3);
}
#define init_10_15_3_4
void FASTCALL call_10_15_3_4(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(3<<4)+4);
}
#define init_10_15_3_5
void FASTCALL call_10_15_3_5(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(3<<4)+5);
}
#define init_10_15_3_6
void FASTCALL call_10_15_3_6(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(3<<4)+6);
}
#define init_10_15_3_7
void FASTCALL call_10_15_3_7(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(3<<4)+7);
}
#define init_10_15_3_8
void FASTCALL call_10_15_3_8(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(3<<4)+8);
}
#define init_10_15_3_9
void FASTCALL call_10_15_3_9(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(3<<4)+9);
}
#define init_10_15_3_10
void FASTCALL call_10_15_3_10(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(3<<4)+10);
}
#define init_10_15_3_11
void FASTCALL call_10_15_3_11(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(3<<4)+11);
}
#define init_10_15_3_12
void FASTCALL call_10_15_3_12(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(3<<4)+12);
}
#define init_10_15_3_13
void FASTCALL call_10_15_3_13(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(3<<4)+13);
}
#define init_10_15_3_14
void FASTCALL call_10_15_3_14(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(3<<4)+14);
}
#define init_10_15_3_15
void FASTCALL call_10_15_3_15(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(3<<4)+15);
}
#define init_10_15_4_0
void FASTCALL call_10_15_4_0(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(4<<4)+0);
}
#define init_10_15_4_1
void FASTCALL call_10_15_4_1(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(4<<4)+1);
}
#define init_10_15_4_2
void FASTCALL call_10_15_4_2(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(4<<4)+2);
}
#define init_10_15_4_3
void FASTCALL call_10_15_4_3(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(4<<4)+3);
}
#define init_10_15_4_4
void FASTCALL call_10_15_4_4(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(4<<4)+4);
}
#define init_10_15_4_5
void FASTCALL call_10_15_4_5(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(4<<4)+5);
}
#define init_10_15_4_6
void FASTCALL call_10_15_4_6(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(4<<4)+6);
}
#define init_10_15_4_7
void FASTCALL call_10_15_4_7(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(4<<4)+7);
}
#define init_10_15_4_8
void FASTCALL call_10_15_4_8(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(4<<4)+8);
}
#define init_10_15_4_9
void FASTCALL call_10_15_4_9(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(4<<4)+9);
}
#define init_10_15_4_10
void FASTCALL call_10_15_4_10(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(4<<4)+10);
}
#define init_10_15_4_11
void FASTCALL call_10_15_4_11(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(4<<4)+11);
}
#define init_10_15_4_12
void FASTCALL call_10_15_4_12(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(4<<4)+12);
}
#define init_10_15_4_13
void FASTCALL call_10_15_4_13(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(4<<4)+13);
}
#define init_10_15_4_14
void FASTCALL call_10_15_4_14(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(4<<4)+14);
}
#define init_10_15_4_15
void FASTCALL call_10_15_4_15(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(4<<4)+15);
}
#define init_10_15_5_0
void FASTCALL call_10_15_5_0(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(5<<4)+0);
}
#define init_10_15_5_1
void FASTCALL call_10_15_5_1(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(5<<4)+1);
}
#define init_10_15_5_2
void FASTCALL call_10_15_5_2(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(5<<4)+2);
}
#define init_10_15_5_3
void FASTCALL call_10_15_5_3(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(5<<4)+3);
}
#define init_10_15_5_4
void FASTCALL call_10_15_5_4(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(5<<4)+4);
}
#define init_10_15_5_5
void FASTCALL call_10_15_5_5(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(5<<4)+5);
}
#define init_10_15_5_6
void FASTCALL call_10_15_5_6(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(5<<4)+6);
}
#define init_10_15_5_7
void FASTCALL call_10_15_5_7(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(5<<4)+7);
}
#define init_10_15_5_8
void FASTCALL call_10_15_5_8(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(5<<4)+8);
}
#define init_10_15_5_9
void FASTCALL call_10_15_5_9(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(5<<4)+9);
}
#define init_10_15_5_10
void FASTCALL call_10_15_5_10(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(5<<4)+10);
}
#define init_10_15_5_11
void FASTCALL call_10_15_5_11(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(5<<4)+11);
}
#define init_10_15_5_12
void FASTCALL call_10_15_5_12(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(5<<4)+12);
}
#define init_10_15_5_13
void FASTCALL call_10_15_5_13(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(5<<4)+13);
}
#define init_10_15_5_14
void FASTCALL call_10_15_5_14(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(5<<4)+14);
}
#define init_10_15_5_15
void FASTCALL call_10_15_5_15(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(5<<4)+15);
}
#define init_10_15_6_0
void FASTCALL call_10_15_6_0(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(6<<4)+0);
}
#define init_10_15_6_1
void FASTCALL call_10_15_6_1(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(6<<4)+1);
}
#define init_10_15_6_2
void FASTCALL call_10_15_6_2(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(6<<4)+2);
}
#define init_10_15_6_3
void FASTCALL call_10_15_6_3(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(6<<4)+3);
}
#define init_10_15_6_4
void FASTCALL call_10_15_6_4(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(6<<4)+4);
}
#define init_10_15_6_5
void FASTCALL call_10_15_6_5(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(6<<4)+5);
}
#define init_10_15_6_6
void FASTCALL call_10_15_6_6(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(6<<4)+6);
}
#define init_10_15_6_7
void FASTCALL call_10_15_6_7(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(6<<4)+7);
}
#define init_10_15_6_8
void FASTCALL call_10_15_6_8(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(6<<4)+8);
}
#define init_10_15_6_9
void FASTCALL call_10_15_6_9(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(6<<4)+9);
}
#define init_10_15_6_10
void FASTCALL call_10_15_6_10(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(6<<4)+10);
}
#define init_10_15_6_11
void FASTCALL call_10_15_6_11(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(6<<4)+11);
}
#define init_10_15_6_12
void FASTCALL call_10_15_6_12(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(6<<4)+12);
}
#define init_10_15_6_13
void FASTCALL call_10_15_6_13(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(6<<4)+13);
}
#define init_10_15_6_14
void FASTCALL call_10_15_6_14(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(6<<4)+14);
}
#define init_10_15_6_15
void FASTCALL call_10_15_6_15(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(6<<4)+15);
}
#define init_10_15_7_0
void FASTCALL call_10_15_7_0(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(7<<4)+0);
}
#define init_10_15_7_1
void FASTCALL call_10_15_7_1(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(7<<4)+1);
}
#define init_10_15_7_2
void FASTCALL call_10_15_7_2(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(7<<4)+2);
}
#define init_10_15_7_3
void FASTCALL call_10_15_7_3(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(7<<4)+3);
}
#define init_10_15_7_4
void FASTCALL call_10_15_7_4(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(7<<4)+4);
}
#define init_10_15_7_5
void FASTCALL call_10_15_7_5(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(7<<4)+5);
}
#define init_10_15_7_6
void FASTCALL call_10_15_7_6(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(7<<4)+6);
}
#define init_10_15_7_7
void FASTCALL call_10_15_7_7(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(7<<4)+7);
}
#define init_10_15_7_8
void FASTCALL call_10_15_7_8(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(7<<4)+8);
}
#define init_10_15_7_9
void FASTCALL call_10_15_7_9(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(7<<4)+9);
}
#define init_10_15_7_10
void FASTCALL call_10_15_7_10(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(7<<4)+10);
}
#define init_10_15_7_11
void FASTCALL call_10_15_7_11(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(7<<4)+11);
}
#define init_10_15_7_12
void FASTCALL call_10_15_7_12(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(7<<4)+12);
}
#define init_10_15_7_13
void FASTCALL call_10_15_7_13(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(7<<4)+13);
}
#define init_10_15_7_14
void FASTCALL call_10_15_7_14(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(7<<4)+14);
}
#define init_10_15_7_15
void FASTCALL call_10_15_7_15(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(7<<4)+15);
}
#define init_10_15_8_0
void FASTCALL call_10_15_8_0(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(8<<4)+0);
}
#define init_10_15_8_1
void FASTCALL call_10_15_8_1(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(8<<4)+1);
}
#define init_10_15_8_2
void FASTCALL call_10_15_8_2(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(8<<4)+2);
}
#define init_10_15_8_3
void FASTCALL call_10_15_8_3(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(8<<4)+3);
}
#define init_10_15_8_4
void FASTCALL call_10_15_8_4(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(8<<4)+4);
}
#define init_10_15_8_5
void FASTCALL call_10_15_8_5(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(8<<4)+5);
}
#define init_10_15_8_6
void FASTCALL call_10_15_8_6(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(8<<4)+6);
}
#define init_10_15_8_7
void FASTCALL call_10_15_8_7(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(8<<4)+7);
}
#define init_10_15_8_8
void FASTCALL call_10_15_8_8(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(8<<4)+8);
}
#define init_10_15_8_9
void FASTCALL call_10_15_8_9(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(8<<4)+9);
}
#define init_10_15_8_10
void FASTCALL call_10_15_8_10(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(8<<4)+10);
}
#define init_10_15_8_11
void FASTCALL call_10_15_8_11(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(8<<4)+11);
}
#define init_10_15_8_12
void FASTCALL call_10_15_8_12(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(8<<4)+12);
}
#define init_10_15_8_13
void FASTCALL call_10_15_8_13(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(8<<4)+13);
}
#define init_10_15_8_14
void FASTCALL call_10_15_8_14(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(8<<4)+14);
}
#define init_10_15_8_15
void FASTCALL call_10_15_8_15(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(8<<4)+15);
}
#define init_10_15_9_0
void FASTCALL call_10_15_9_0(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(9<<4)+0);
}
#define init_10_15_9_1
void FASTCALL call_10_15_9_1(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(9<<4)+1);
}
#define init_10_15_9_2
void FASTCALL call_10_15_9_2(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(9<<4)+2);
}
#define init_10_15_9_3
void FASTCALL call_10_15_9_3(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(9<<4)+3);
}
#define init_10_15_9_4
void FASTCALL call_10_15_9_4(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(9<<4)+4);
}
#define init_10_15_9_5
void FASTCALL call_10_15_9_5(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(9<<4)+5);
}
#define init_10_15_9_6
void FASTCALL call_10_15_9_6(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(9<<4)+6);
}
#define init_10_15_9_7
void FASTCALL call_10_15_9_7(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(9<<4)+7);
}
#define init_10_15_9_8
void FASTCALL call_10_15_9_8(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(9<<4)+8);
}
#define init_10_15_9_9
void FASTCALL call_10_15_9_9(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(9<<4)+9);
}
#define init_10_15_9_10
void FASTCALL call_10_15_9_10(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(9<<4)+10);
}
#define init_10_15_9_11
void FASTCALL call_10_15_9_11(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(9<<4)+11);
}
#define init_10_15_9_12
void FASTCALL call_10_15_9_12(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(9<<4)+12);
}
#define init_10_15_9_13
void FASTCALL call_10_15_9_13(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(9<<4)+13);
}
#define init_10_15_9_14
void FASTCALL call_10_15_9_14(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(9<<4)+14);
}
#define init_10_15_9_15
void FASTCALL call_10_15_9_15(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(9<<4)+15);
}
#define init_10_15_10_0
void FASTCALL call_10_15_10_0(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(10<<4)+0);
}
#define init_10_15_10_1
void FASTCALL call_10_15_10_1(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(10<<4)+1);
}
#define init_10_15_10_2
void FASTCALL call_10_15_10_2(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(10<<4)+2);
}
#define init_10_15_10_3
void FASTCALL call_10_15_10_3(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(10<<4)+3);
}
#define init_10_15_10_4
void FASTCALL call_10_15_10_4(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(10<<4)+4);
}
#define init_10_15_10_5
void FASTCALL call_10_15_10_5(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(10<<4)+5);
}
#define init_10_15_10_6
void FASTCALL call_10_15_10_6(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(10<<4)+6);
}
#define init_10_15_10_7
void FASTCALL call_10_15_10_7(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(10<<4)+7);
}
#define init_10_15_10_8
void FASTCALL call_10_15_10_8(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(10<<4)+8);
}
#define init_10_15_10_9
void FASTCALL call_10_15_10_9(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(10<<4)+9);
}
#define init_10_15_10_10
void FASTCALL call_10_15_10_10(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(10<<4)+10);
}
#define init_10_15_10_11
void FASTCALL call_10_15_10_11(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(10<<4)+11);
}
#define init_10_15_10_12
void FASTCALL call_10_15_10_12(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(10<<4)+12);
}
#define init_10_15_10_13
void FASTCALL call_10_15_10_13(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(10<<4)+13);
}
#define init_10_15_10_14
void FASTCALL call_10_15_10_14(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(10<<4)+14);
}
#define init_10_15_10_15
void FASTCALL call_10_15_10_15(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(10<<4)+15);
}
#define init_10_15_11_0
void FASTCALL call_10_15_11_0(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(11<<4)+0);
}
#define init_10_15_11_1
void FASTCALL call_10_15_11_1(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(11<<4)+1);
}
#define init_10_15_11_2
void FASTCALL call_10_15_11_2(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(11<<4)+2);
}
#define init_10_15_11_3
void FASTCALL call_10_15_11_3(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(11<<4)+3);
}
#define init_10_15_11_4
void FASTCALL call_10_15_11_4(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(11<<4)+4);
}
#define init_10_15_11_5
void FASTCALL call_10_15_11_5(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(11<<4)+5);
}
#define init_10_15_11_6
void FASTCALL call_10_15_11_6(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(11<<4)+6);
}
#define init_10_15_11_7
void FASTCALL call_10_15_11_7(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(11<<4)+7);
}
#define init_10_15_11_8
void FASTCALL call_10_15_11_8(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(11<<4)+8);
}
#define init_10_15_11_9
void FASTCALL call_10_15_11_9(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(11<<4)+9);
}
#define init_10_15_11_10
void FASTCALL call_10_15_11_10(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(11<<4)+10);
}
#define init_10_15_11_11
void FASTCALL call_10_15_11_11(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(11<<4)+11);
}
#define init_10_15_11_12
void FASTCALL call_10_15_11_12(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(11<<4)+12);
}
#define init_10_15_11_13
void FASTCALL call_10_15_11_13(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(11<<4)+13);
}
#define init_10_15_11_14
void FASTCALL call_10_15_11_14(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(11<<4)+14);
}
#define init_10_15_11_15
void FASTCALL call_10_15_11_15(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(11<<4)+15);
}
#define init_10_15_12_0
void FASTCALL call_10_15_12_0(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(12<<4)+0);
}
#define init_10_15_12_1
void FASTCALL call_10_15_12_1(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(12<<4)+1);
}
#define init_10_15_12_2
void FASTCALL call_10_15_12_2(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(12<<4)+2);
}
#define init_10_15_12_3
void FASTCALL call_10_15_12_3(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(12<<4)+3);
}
#define init_10_15_12_4
void FASTCALL call_10_15_12_4(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(12<<4)+4);
}
#define init_10_15_12_5
void FASTCALL call_10_15_12_5(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(12<<4)+5);
}
#define init_10_15_12_6
void FASTCALL call_10_15_12_6(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(12<<4)+6);
}
#define init_10_15_12_7
void FASTCALL call_10_15_12_7(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(12<<4)+7);
}
#define init_10_15_12_8
void FASTCALL call_10_15_12_8(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(12<<4)+8);
}
#define init_10_15_12_9
void FASTCALL call_10_15_12_9(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(12<<4)+9);
}
#define init_10_15_12_10
void FASTCALL call_10_15_12_10(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(12<<4)+10);
}
#define init_10_15_12_11
void FASTCALL call_10_15_12_11(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(12<<4)+11);
}
#define init_10_15_12_12
void FASTCALL call_10_15_12_12(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(12<<4)+12);
}
#define init_10_15_12_13
void FASTCALL call_10_15_12_13(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(12<<4)+13);
}
#define init_10_15_12_14
void FASTCALL call_10_15_12_14(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(12<<4)+14);
}
#define init_10_15_12_15
void FASTCALL call_10_15_12_15(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(12<<4)+15);
}
#define init_10_15_13_0
void FASTCALL call_10_15_13_0(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(13<<4)+0);
}
#define init_10_15_13_1
void FASTCALL call_10_15_13_1(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(13<<4)+1);
}
#define init_10_15_13_2
void FASTCALL call_10_15_13_2(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(13<<4)+2);
}
#define init_10_15_13_3
void FASTCALL call_10_15_13_3(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(13<<4)+3);
}
#define init_10_15_13_4
void FASTCALL call_10_15_13_4(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(13<<4)+4);
}
#define init_10_15_13_5
void FASTCALL call_10_15_13_5(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(13<<4)+5);
}
#define init_10_15_13_6
void FASTCALL call_10_15_13_6(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(13<<4)+6);
}
#define init_10_15_13_7
void FASTCALL call_10_15_13_7(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(13<<4)+7);
}
#define init_10_15_13_8
void FASTCALL call_10_15_13_8(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(13<<4)+8);
}
#define init_10_15_13_9
void FASTCALL call_10_15_13_9(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(13<<4)+9);
}
#define init_10_15_13_10
void FASTCALL call_10_15_13_10(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(13<<4)+10);
}
#define init_10_15_13_11
void FASTCALL call_10_15_13_11(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(13<<4)+11);
}
#define init_10_15_13_12
void FASTCALL call_10_15_13_12(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(13<<4)+12);
}
#define init_10_15_13_13
void FASTCALL call_10_15_13_13(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(13<<4)+13);
}
#define init_10_15_13_14
void FASTCALL call_10_15_13_14(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(13<<4)+14);
}
#define init_10_15_13_15
void FASTCALL call_10_15_13_15(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(13<<4)+15);
}
#define init_10_15_14_0
void FASTCALL call_10_15_14_0(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(14<<4)+0);
}
#define init_10_15_14_1
void FASTCALL call_10_15_14_1(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(14<<4)+1);
}
#define init_10_15_14_2
void FASTCALL call_10_15_14_2(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(14<<4)+2);
}
#define init_10_15_14_3
void FASTCALL call_10_15_14_3(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(14<<4)+3);
}
#define init_10_15_14_4
void FASTCALL call_10_15_14_4(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(14<<4)+4);
}
#define init_10_15_14_5
void FASTCALL call_10_15_14_5(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(14<<4)+5);
}
#define init_10_15_14_6
void FASTCALL call_10_15_14_6(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(14<<4)+6);
}
#define init_10_15_14_7
void FASTCALL call_10_15_14_7(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(14<<4)+7);
}
#define init_10_15_14_8
void FASTCALL call_10_15_14_8(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(14<<4)+8);
}
#define init_10_15_14_9
void FASTCALL call_10_15_14_9(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(14<<4)+9);
}
#define init_10_15_14_10
void FASTCALL call_10_15_14_10(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(14<<4)+10);
}
#define init_10_15_14_11
void FASTCALL call_10_15_14_11(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(14<<4)+11);
}
#define init_10_15_14_12
void FASTCALL call_10_15_14_12(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(14<<4)+12);
}
#define init_10_15_14_13
void FASTCALL call_10_15_14_13(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(14<<4)+13);
}
#define init_10_15_14_14
void FASTCALL call_10_15_14_14(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(14<<4)+14);
}
#define init_10_15_14_15
void FASTCALL call_10_15_14_15(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(14<<4)+15);
}
#define init_10_15_15_0
void FASTCALL call_10_15_15_0(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(15<<4)+0);
}
#define init_10_15_15_1
void FASTCALL call_10_15_15_1(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(15<<4)+1);
}
#define init_10_15_15_2
void FASTCALL call_10_15_15_2(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(15<<4)+2);
}
#define init_10_15_15_3
void FASTCALL call_10_15_15_3(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(15<<4)+3);
}
#define init_10_15_15_4
void FASTCALL call_10_15_15_4(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(15<<4)+4);
}
#define init_10_15_15_5
void FASTCALL call_10_15_15_5(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(15<<4)+5);
}
#define init_10_15_15_6
void FASTCALL call_10_15_15_6(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(15<<4)+6);
}
#define init_10_15_15_7
void FASTCALL call_10_15_15_7(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(15<<4)+7);
}
#define init_10_15_15_8
void FASTCALL call_10_15_15_8(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(15<<4)+8);
}
#define init_10_15_15_9
void FASTCALL call_10_15_15_9(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(15<<4)+9);
}
#define init_10_15_15_10
void FASTCALL call_10_15_15_10(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(15<<4)+10);
}
#define init_10_15_15_11
void FASTCALL call_10_15_15_11(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(15<<4)+11);
}
#define init_10_15_15_12
void FASTCALL call_10_15_15_12(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(15<<4)+12);
}
#define init_10_15_15_13
void FASTCALL call_10_15_15_13(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(15<<4)+13);
}
#define init_10_15_15_14
void FASTCALL call_10_15_15_14(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(15<<4)+14);
}
#define init_10_15_15_15
void FASTCALL call_10_15_15_15(SH2_struct* sh)
{
  SH2bra(sh, (15<<8)+(15<<4)+15);
}
#define init_8_15_0_0
void FASTCALL call_8_15_0_0(SH2_struct* sh)
{
  SH2bfs(sh,(0<<4)+0);
}
#define init_8_15_0_1
void FASTCALL call_8_15_0_1(SH2_struct* sh)
{
  SH2bfs(sh,(0<<4)+1);
}
#define init_8_15_0_2
void FASTCALL call_8_15_0_2(SH2_struct* sh)
{
  SH2bfs(sh,(0<<4)+2);
}
#define init_8_15_0_3
void FASTCALL call_8_15_0_3(SH2_struct* sh)
{
  SH2bfs(sh,(0<<4)+3);
}
#define init_8_15_0_4
void FASTCALL call_8_15_0_4(SH2_struct* sh)
{
  SH2bfs(sh,(0<<4)+4);
}
#define init_8_15_0_5
void FASTCALL call_8_15_0_5(SH2_struct* sh)
{
  SH2bfs(sh,(0<<4)+5);
}
#define init_8_15_0_6
void FASTCALL call_8_15_0_6(SH2_struct* sh)
{
  SH2bfs(sh,(0<<4)+6);
}
#define init_8_15_0_7
void FASTCALL call_8_15_0_7(SH2_struct* sh)
{
  SH2bfs(sh,(0<<4)+7);
}
#define init_8_15_0_8
void FASTCALL call_8_15_0_8(SH2_struct* sh)
{
  SH2bfs(sh,(0<<4)+8);
}
#define init_8_15_0_9
void FASTCALL call_8_15_0_9(SH2_struct* sh)
{
  SH2bfs(sh,(0<<4)+9);
}
#define init_8_15_0_10
void FASTCALL call_8_15_0_10(SH2_struct* sh)
{
  SH2bfs(sh,(0<<4)+10);
}
#define init_8_15_0_11
void FASTCALL call_8_15_0_11(SH2_struct* sh)
{
  SH2bfs(sh,(0<<4)+11);
}
#define init_8_15_0_12
void FASTCALL call_8_15_0_12(SH2_struct* sh)
{
  SH2bfs(sh,(0<<4)+12);
}
#define init_8_15_0_13
void FASTCALL call_8_15_0_13(SH2_struct* sh)
{
  SH2bfs(sh,(0<<4)+13);
}
#define init_8_15_0_14
void FASTCALL call_8_15_0_14(SH2_struct* sh)
{
  SH2bfs(sh,(0<<4)+14);
}
#define init_8_15_0_15
void FASTCALL call_8_15_0_15(SH2_struct* sh)
{
  SH2bfs(sh,(0<<4)+15);
}
#define init_8_15_1_0
void FASTCALL call_8_15_1_0(SH2_struct* sh)
{
  SH2bfs(sh,(1<<4)+0);
}
#define init_8_15_1_1
void FASTCALL call_8_15_1_1(SH2_struct* sh)
{
  SH2bfs(sh,(1<<4)+1);
}
#define init_8_15_1_2
void FASTCALL call_8_15_1_2(SH2_struct* sh)
{
  SH2bfs(sh,(1<<4)+2);
}
#define init_8_15_1_3
void FASTCALL call_8_15_1_3(SH2_struct* sh)
{
  SH2bfs(sh,(1<<4)+3);
}
#define init_8_15_1_4
void FASTCALL call_8_15_1_4(SH2_struct* sh)
{
  SH2bfs(sh,(1<<4)+4);
}
#define init_8_15_1_5
void FASTCALL call_8_15_1_5(SH2_struct* sh)
{
  SH2bfs(sh,(1<<4)+5);
}
#define init_8_15_1_6
void FASTCALL call_8_15_1_6(SH2_struct* sh)
{
  SH2bfs(sh,(1<<4)+6);
}
#define init_8_15_1_7
void FASTCALL call_8_15_1_7(SH2_struct* sh)
{
  SH2bfs(sh,(1<<4)+7);
}
#define init_8_15_1_8
void FASTCALL call_8_15_1_8(SH2_struct* sh)
{
  SH2bfs(sh,(1<<4)+8);
}
#define init_8_15_1_9
void FASTCALL call_8_15_1_9(SH2_struct* sh)
{
  SH2bfs(sh,(1<<4)+9);
}
#define init_8_15_1_10
void FASTCALL call_8_15_1_10(SH2_struct* sh)
{
  SH2bfs(sh,(1<<4)+10);
}
#define init_8_15_1_11
void FASTCALL call_8_15_1_11(SH2_struct* sh)
{
  SH2bfs(sh,(1<<4)+11);
}
#define init_8_15_1_12
void FASTCALL call_8_15_1_12(SH2_struct* sh)
{
  SH2bfs(sh,(1<<4)+12);
}
#define init_8_15_1_13
void FASTCALL call_8_15_1_13(SH2_struct* sh)
{
  SH2bfs(sh,(1<<4)+13);
}
#define init_8_15_1_14
void FASTCALL call_8_15_1_14(SH2_struct* sh)
{
  SH2bfs(sh,(1<<4)+14);
}
#define init_8_15_1_15
void FASTCALL call_8_15_1_15(SH2_struct* sh)
{
  SH2bfs(sh,(1<<4)+15);
}
#define init_8_15_2_0
void FASTCALL call_8_15_2_0(SH2_struct* sh)
{
  SH2bfs(sh,(2<<4)+0);
}
#define init_8_15_2_1
void FASTCALL call_8_15_2_1(SH2_struct* sh)
{
  SH2bfs(sh,(2<<4)+1);
}
#define init_8_15_2_2
void FASTCALL call_8_15_2_2(SH2_struct* sh)
{
  SH2bfs(sh,(2<<4)+2);
}
#define init_8_15_2_3
void FASTCALL call_8_15_2_3(SH2_struct* sh)
{
  SH2bfs(sh,(2<<4)+3);
}
#define init_8_15_2_4
void FASTCALL call_8_15_2_4(SH2_struct* sh)
{
  SH2bfs(sh,(2<<4)+4);
}
#define init_8_15_2_5
void FASTCALL call_8_15_2_5(SH2_struct* sh)
{
  SH2bfs(sh,(2<<4)+5);
}
#define init_8_15_2_6
void FASTCALL call_8_15_2_6(SH2_struct* sh)
{
  SH2bfs(sh,(2<<4)+6);
}
#define init_8_15_2_7
void FASTCALL call_8_15_2_7(SH2_struct* sh)
{
  SH2bfs(sh,(2<<4)+7);
}
#define init_8_15_2_8
void FASTCALL call_8_15_2_8(SH2_struct* sh)
{
  SH2bfs(sh,(2<<4)+8);
}
#define init_8_15_2_9
void FASTCALL call_8_15_2_9(SH2_struct* sh)
{
  SH2bfs(sh,(2<<4)+9);
}
#define init_8_15_2_10
void FASTCALL call_8_15_2_10(SH2_struct* sh)
{
  SH2bfs(sh,(2<<4)+10);
}
#define init_8_15_2_11
void FASTCALL call_8_15_2_11(SH2_struct* sh)
{
  SH2bfs(sh,(2<<4)+11);
}
#define init_8_15_2_12
void FASTCALL call_8_15_2_12(SH2_struct* sh)
{
  SH2bfs(sh,(2<<4)+12);
}
#define init_8_15_2_13
void FASTCALL call_8_15_2_13(SH2_struct* sh)
{
  SH2bfs(sh,(2<<4)+13);
}
#define init_8_15_2_14
void FASTCALL call_8_15_2_14(SH2_struct* sh)
{
  SH2bfs(sh,(2<<4)+14);
}
#define init_8_15_2_15
void FASTCALL call_8_15_2_15(SH2_struct* sh)
{
  SH2bfs(sh,(2<<4)+15);
}
#define init_8_15_3_0
void FASTCALL call_8_15_3_0(SH2_struct* sh)
{
  SH2bfs(sh,(3<<4)+0);
}
#define init_8_15_3_1
void FASTCALL call_8_15_3_1(SH2_struct* sh)
{
  SH2bfs(sh,(3<<4)+1);
}
#define init_8_15_3_2
void FASTCALL call_8_15_3_2(SH2_struct* sh)
{
  SH2bfs(sh,(3<<4)+2);
}
#define init_8_15_3_3
void FASTCALL call_8_15_3_3(SH2_struct* sh)
{
  SH2bfs(sh,(3<<4)+3);
}
#define init_8_15_3_4
void FASTCALL call_8_15_3_4(SH2_struct* sh)
{
  SH2bfs(sh,(3<<4)+4);
}
#define init_8_15_3_5
void FASTCALL call_8_15_3_5(SH2_struct* sh)
{
  SH2bfs(sh,(3<<4)+5);
}
#define init_8_15_3_6
void FASTCALL call_8_15_3_6(SH2_struct* sh)
{
  SH2bfs(sh,(3<<4)+6);
}
#define init_8_15_3_7
void FASTCALL call_8_15_3_7(SH2_struct* sh)
{
  SH2bfs(sh,(3<<4)+7);
}
#define init_8_15_3_8
void FASTCALL call_8_15_3_8(SH2_struct* sh)
{
  SH2bfs(sh,(3<<4)+8);
}
#define init_8_15_3_9
void FASTCALL call_8_15_3_9(SH2_struct* sh)
{
  SH2bfs(sh,(3<<4)+9);
}
#define init_8_15_3_10
void FASTCALL call_8_15_3_10(SH2_struct* sh)
{
  SH2bfs(sh,(3<<4)+10);
}
#define init_8_15_3_11
void FASTCALL call_8_15_3_11(SH2_struct* sh)
{
  SH2bfs(sh,(3<<4)+11);
}
#define init_8_15_3_12
void FASTCALL call_8_15_3_12(SH2_struct* sh)
{
  SH2bfs(sh,(3<<4)+12);
}
#define init_8_15_3_13
void FASTCALL call_8_15_3_13(SH2_struct* sh)
{
  SH2bfs(sh,(3<<4)+13);
}
#define init_8_15_3_14
void FASTCALL call_8_15_3_14(SH2_struct* sh)
{
  SH2bfs(sh,(3<<4)+14);
}
#define init_8_15_3_15
void FASTCALL call_8_15_3_15(SH2_struct* sh)
{
  SH2bfs(sh,(3<<4)+15);
}
#define init_8_15_4_0
void FASTCALL call_8_15_4_0(SH2_struct* sh)
{
  SH2bfs(sh,(4<<4)+0);
}
#define init_8_15_4_1
void FASTCALL call_8_15_4_1(SH2_struct* sh)
{
  SH2bfs(sh,(4<<4)+1);
}
#define init_8_15_4_2
void FASTCALL call_8_15_4_2(SH2_struct* sh)
{
  SH2bfs(sh,(4<<4)+2);
}
#define init_8_15_4_3
void FASTCALL call_8_15_4_3(SH2_struct* sh)
{
  SH2bfs(sh,(4<<4)+3);
}
#define init_8_15_4_4
void FASTCALL call_8_15_4_4(SH2_struct* sh)
{
  SH2bfs(sh,(4<<4)+4);
}
#define init_8_15_4_5
void FASTCALL call_8_15_4_5(SH2_struct* sh)
{
  SH2bfs(sh,(4<<4)+5);
}
#define init_8_15_4_6
void FASTCALL call_8_15_4_6(SH2_struct* sh)
{
  SH2bfs(sh,(4<<4)+6);
}
#define init_8_15_4_7
void FASTCALL call_8_15_4_7(SH2_struct* sh)
{
  SH2bfs(sh,(4<<4)+7);
}
#define init_8_15_4_8
void FASTCALL call_8_15_4_8(SH2_struct* sh)
{
  SH2bfs(sh,(4<<4)+8);
}
#define init_8_15_4_9
void FASTCALL call_8_15_4_9(SH2_struct* sh)
{
  SH2bfs(sh,(4<<4)+9);
}
#define init_8_15_4_10
void FASTCALL call_8_15_4_10(SH2_struct* sh)
{
  SH2bfs(sh,(4<<4)+10);
}
#define init_8_15_4_11
void FASTCALL call_8_15_4_11(SH2_struct* sh)
{
  SH2bfs(sh,(4<<4)+11);
}
#define init_8_15_4_12
void FASTCALL call_8_15_4_12(SH2_struct* sh)
{
  SH2bfs(sh,(4<<4)+12);
}
#define init_8_15_4_13
void FASTCALL call_8_15_4_13(SH2_struct* sh)
{
  SH2bfs(sh,(4<<4)+13);
}
#define init_8_15_4_14
void FASTCALL call_8_15_4_14(SH2_struct* sh)
{
  SH2bfs(sh,(4<<4)+14);
}
#define init_8_15_4_15
void FASTCALL call_8_15_4_15(SH2_struct* sh)
{
  SH2bfs(sh,(4<<4)+15);
}
#define init_8_15_5_0
void FASTCALL call_8_15_5_0(SH2_struct* sh)
{
  SH2bfs(sh,(5<<4)+0);
}
#define init_8_15_5_1
void FASTCALL call_8_15_5_1(SH2_struct* sh)
{
  SH2bfs(sh,(5<<4)+1);
}
#define init_8_15_5_2
void FASTCALL call_8_15_5_2(SH2_struct* sh)
{
  SH2bfs(sh,(5<<4)+2);
}
#define init_8_15_5_3
void FASTCALL call_8_15_5_3(SH2_struct* sh)
{
  SH2bfs(sh,(5<<4)+3);
}
#define init_8_15_5_4
void FASTCALL call_8_15_5_4(SH2_struct* sh)
{
  SH2bfs(sh,(5<<4)+4);
}
#define init_8_15_5_5
void FASTCALL call_8_15_5_5(SH2_struct* sh)
{
  SH2bfs(sh,(5<<4)+5);
}
#define init_8_15_5_6
void FASTCALL call_8_15_5_6(SH2_struct* sh)
{
  SH2bfs(sh,(5<<4)+6);
}
#define init_8_15_5_7
void FASTCALL call_8_15_5_7(SH2_struct* sh)
{
  SH2bfs(sh,(5<<4)+7);
}
#define init_8_15_5_8
void FASTCALL call_8_15_5_8(SH2_struct* sh)
{
  SH2bfs(sh,(5<<4)+8);
}
#define init_8_15_5_9
void FASTCALL call_8_15_5_9(SH2_struct* sh)
{
  SH2bfs(sh,(5<<4)+9);
}
#define init_8_15_5_10
void FASTCALL call_8_15_5_10(SH2_struct* sh)
{
  SH2bfs(sh,(5<<4)+10);
}
#define init_8_15_5_11
void FASTCALL call_8_15_5_11(SH2_struct* sh)
{
  SH2bfs(sh,(5<<4)+11);
}
#define init_8_15_5_12
void FASTCALL call_8_15_5_12(SH2_struct* sh)
{
  SH2bfs(sh,(5<<4)+12);
}
#define init_8_15_5_13
void FASTCALL call_8_15_5_13(SH2_struct* sh)
{
  SH2bfs(sh,(5<<4)+13);
}
#define init_8_15_5_14
void FASTCALL call_8_15_5_14(SH2_struct* sh)
{
  SH2bfs(sh,(5<<4)+14);
}
#define init_8_15_5_15
void FASTCALL call_8_15_5_15(SH2_struct* sh)
{
  SH2bfs(sh,(5<<4)+15);
}
#define init_8_15_6_0
void FASTCALL call_8_15_6_0(SH2_struct* sh)
{
  SH2bfs(sh,(6<<4)+0);
}
#define init_8_15_6_1
void FASTCALL call_8_15_6_1(SH2_struct* sh)
{
  SH2bfs(sh,(6<<4)+1);
}
#define init_8_15_6_2
void FASTCALL call_8_15_6_2(SH2_struct* sh)
{
  SH2bfs(sh,(6<<4)+2);
}
#define init_8_15_6_3
void FASTCALL call_8_15_6_3(SH2_struct* sh)
{
  SH2bfs(sh,(6<<4)+3);
}
#define init_8_15_6_4
void FASTCALL call_8_15_6_4(SH2_struct* sh)
{
  SH2bfs(sh,(6<<4)+4);
}
#define init_8_15_6_5
void FASTCALL call_8_15_6_5(SH2_struct* sh)
{
  SH2bfs(sh,(6<<4)+5);
}
#define init_8_15_6_6
void FASTCALL call_8_15_6_6(SH2_struct* sh)
{
  SH2bfs(sh,(6<<4)+6);
}
#define init_8_15_6_7
void FASTCALL call_8_15_6_7(SH2_struct* sh)
{
  SH2bfs(sh,(6<<4)+7);
}
#define init_8_15_6_8
void FASTCALL call_8_15_6_8(SH2_struct* sh)
{
  SH2bfs(sh,(6<<4)+8);
}
#define init_8_15_6_9
void FASTCALL call_8_15_6_9(SH2_struct* sh)
{
  SH2bfs(sh,(6<<4)+9);
}
#define init_8_15_6_10
void FASTCALL call_8_15_6_10(SH2_struct* sh)
{
  SH2bfs(sh,(6<<4)+10);
}
#define init_8_15_6_11
void FASTCALL call_8_15_6_11(SH2_struct* sh)
{
  SH2bfs(sh,(6<<4)+11);
}
#define init_8_15_6_12
void FASTCALL call_8_15_6_12(SH2_struct* sh)
{
  SH2bfs(sh,(6<<4)+12);
}
#define init_8_15_6_13
void FASTCALL call_8_15_6_13(SH2_struct* sh)
{
  SH2bfs(sh,(6<<4)+13);
}
#define init_8_15_6_14
void FASTCALL call_8_15_6_14(SH2_struct* sh)
{
  SH2bfs(sh,(6<<4)+14);
}
#define init_8_15_6_15
void FASTCALL call_8_15_6_15(SH2_struct* sh)
{
  SH2bfs(sh,(6<<4)+15);
}
#define init_8_15_7_0
void FASTCALL call_8_15_7_0(SH2_struct* sh)
{
  SH2bfs(sh,(7<<4)+0);
}
#define init_8_15_7_1
void FASTCALL call_8_15_7_1(SH2_struct* sh)
{
  SH2bfs(sh,(7<<4)+1);
}
#define init_8_15_7_2
void FASTCALL call_8_15_7_2(SH2_struct* sh)
{
  SH2bfs(sh,(7<<4)+2);
}
#define init_8_15_7_3
void FASTCALL call_8_15_7_3(SH2_struct* sh)
{
  SH2bfs(sh,(7<<4)+3);
}
#define init_8_15_7_4
void FASTCALL call_8_15_7_4(SH2_struct* sh)
{
  SH2bfs(sh,(7<<4)+4);
}
#define init_8_15_7_5
void FASTCALL call_8_15_7_5(SH2_struct* sh)
{
  SH2bfs(sh,(7<<4)+5);
}
#define init_8_15_7_6
void FASTCALL call_8_15_7_6(SH2_struct* sh)
{
  SH2bfs(sh,(7<<4)+6);
}
#define init_8_15_7_7
void FASTCALL call_8_15_7_7(SH2_struct* sh)
{
  SH2bfs(sh,(7<<4)+7);
}
#define init_8_15_7_8
void FASTCALL call_8_15_7_8(SH2_struct* sh)
{
  SH2bfs(sh,(7<<4)+8);
}
#define init_8_15_7_9
void FASTCALL call_8_15_7_9(SH2_struct* sh)
{
  SH2bfs(sh,(7<<4)+9);
}
#define init_8_15_7_10
void FASTCALL call_8_15_7_10(SH2_struct* sh)
{
  SH2bfs(sh,(7<<4)+10);
}
#define init_8_15_7_11
void FASTCALL call_8_15_7_11(SH2_struct* sh)
{
  SH2bfs(sh,(7<<4)+11);
}
#define init_8_15_7_12
void FASTCALL call_8_15_7_12(SH2_struct* sh)
{
  SH2bfs(sh,(7<<4)+12);
}
#define init_8_15_7_13
void FASTCALL call_8_15_7_13(SH2_struct* sh)
{
  SH2bfs(sh,(7<<4)+13);
}
#define init_8_15_7_14
void FASTCALL call_8_15_7_14(SH2_struct* sh)
{
  SH2bfs(sh,(7<<4)+14);
}
#define init_8_15_7_15
void FASTCALL call_8_15_7_15(SH2_struct* sh)
{
  SH2bfs(sh,(7<<4)+15);
}
#define init_8_15_8_0
void FASTCALL call_8_15_8_0(SH2_struct* sh)
{
  SH2bfs(sh,(8<<4)+0);
}
#define init_8_15_8_1
void FASTCALL call_8_15_8_1(SH2_struct* sh)
{
  SH2bfs(sh,(8<<4)+1);
}
#define init_8_15_8_2
void FASTCALL call_8_15_8_2(SH2_struct* sh)
{
  SH2bfs(sh,(8<<4)+2);
}
#define init_8_15_8_3
void FASTCALL call_8_15_8_3(SH2_struct* sh)
{
  SH2bfs(sh,(8<<4)+3);
}
#define init_8_15_8_4
void FASTCALL call_8_15_8_4(SH2_struct* sh)
{
  SH2bfs(sh,(8<<4)+4);
}
#define init_8_15_8_5
void FASTCALL call_8_15_8_5(SH2_struct* sh)
{
  SH2bfs(sh,(8<<4)+5);
}
#define init_8_15_8_6
void FASTCALL call_8_15_8_6(SH2_struct* sh)
{
  SH2bfs(sh,(8<<4)+6);
}
#define init_8_15_8_7
void FASTCALL call_8_15_8_7(SH2_struct* sh)
{
  SH2bfs(sh,(8<<4)+7);
}
#define init_8_15_8_8
void FASTCALL call_8_15_8_8(SH2_struct* sh)
{
  SH2bfs(sh,(8<<4)+8);
}
#define init_8_15_8_9
void FASTCALL call_8_15_8_9(SH2_struct* sh)
{
  SH2bfs(sh,(8<<4)+9);
}
#define init_8_15_8_10
void FASTCALL call_8_15_8_10(SH2_struct* sh)
{
  SH2bfs(sh,(8<<4)+10);
}
#define init_8_15_8_11
void FASTCALL call_8_15_8_11(SH2_struct* sh)
{
  SH2bfs(sh,(8<<4)+11);
}
#define init_8_15_8_12
void FASTCALL call_8_15_8_12(SH2_struct* sh)
{
  SH2bfs(sh,(8<<4)+12);
}
#define init_8_15_8_13
void FASTCALL call_8_15_8_13(SH2_struct* sh)
{
  SH2bfs(sh,(8<<4)+13);
}
#define init_8_15_8_14
void FASTCALL call_8_15_8_14(SH2_struct* sh)
{
  SH2bfs(sh,(8<<4)+14);
}
#define init_8_15_8_15
void FASTCALL call_8_15_8_15(SH2_struct* sh)
{
  SH2bfs(sh,(8<<4)+15);
}
#define init_8_15_9_0
void FASTCALL call_8_15_9_0(SH2_struct* sh)
{
  SH2bfs(sh,(9<<4)+0);
}
#define init_8_15_9_1
void FASTCALL call_8_15_9_1(SH2_struct* sh)
{
  SH2bfs(sh,(9<<4)+1);
}
#define init_8_15_9_2
void FASTCALL call_8_15_9_2(SH2_struct* sh)
{
  SH2bfs(sh,(9<<4)+2);
}
#define init_8_15_9_3
void FASTCALL call_8_15_9_3(SH2_struct* sh)
{
  SH2bfs(sh,(9<<4)+3);
}
#define init_8_15_9_4
void FASTCALL call_8_15_9_4(SH2_struct* sh)
{
  SH2bfs(sh,(9<<4)+4);
}
#define init_8_15_9_5
void FASTCALL call_8_15_9_5(SH2_struct* sh)
{
  SH2bfs(sh,(9<<4)+5);
}
#define init_8_15_9_6
void FASTCALL call_8_15_9_6(SH2_struct* sh)
{
  SH2bfs(sh,(9<<4)+6);
}
#define init_8_15_9_7
void FASTCALL call_8_15_9_7(SH2_struct* sh)
{
  SH2bfs(sh,(9<<4)+7);
}
#define init_8_15_9_8
void FASTCALL call_8_15_9_8(SH2_struct* sh)
{
  SH2bfs(sh,(9<<4)+8);
}
#define init_8_15_9_9
void FASTCALL call_8_15_9_9(SH2_struct* sh)
{
  SH2bfs(sh,(9<<4)+9);
}
#define init_8_15_9_10
void FASTCALL call_8_15_9_10(SH2_struct* sh)
{
  SH2bfs(sh,(9<<4)+10);
}
#define init_8_15_9_11
void FASTCALL call_8_15_9_11(SH2_struct* sh)
{
  SH2bfs(sh,(9<<4)+11);
}
#define init_8_15_9_12
void FASTCALL call_8_15_9_12(SH2_struct* sh)
{
  SH2bfs(sh,(9<<4)+12);
}
#define init_8_15_9_13
void FASTCALL call_8_15_9_13(SH2_struct* sh)
{
  SH2bfs(sh,(9<<4)+13);
}
#define init_8_15_9_14
void FASTCALL call_8_15_9_14(SH2_struct* sh)
{
  SH2bfs(sh,(9<<4)+14);
}
#define init_8_15_9_15
void FASTCALL call_8_15_9_15(SH2_struct* sh)
{
  SH2bfs(sh,(9<<4)+15);
}
#define init_8_15_10_0
void FASTCALL call_8_15_10_0(SH2_struct* sh)
{
  SH2bfs(sh,(10<<4)+0);
}
#define init_8_15_10_1
void FASTCALL call_8_15_10_1(SH2_struct* sh)
{
  SH2bfs(sh,(10<<4)+1);
}
#define init_8_15_10_2
void FASTCALL call_8_15_10_2(SH2_struct* sh)
{
  SH2bfs(sh,(10<<4)+2);
}
#define init_8_15_10_3
void FASTCALL call_8_15_10_3(SH2_struct* sh)
{
  SH2bfs(sh,(10<<4)+3);
}
#define init_8_15_10_4
void FASTCALL call_8_15_10_4(SH2_struct* sh)
{
  SH2bfs(sh,(10<<4)+4);
}
#define init_8_15_10_5
void FASTCALL call_8_15_10_5(SH2_struct* sh)
{
  SH2bfs(sh,(10<<4)+5);
}
#define init_8_15_10_6
void FASTCALL call_8_15_10_6(SH2_struct* sh)
{
  SH2bfs(sh,(10<<4)+6);
}
#define init_8_15_10_7
void FASTCALL call_8_15_10_7(SH2_struct* sh)
{
  SH2bfs(sh,(10<<4)+7);
}
#define init_8_15_10_8
void FASTCALL call_8_15_10_8(SH2_struct* sh)
{
  SH2bfs(sh,(10<<4)+8);
}
#define init_8_15_10_9
void FASTCALL call_8_15_10_9(SH2_struct* sh)
{
  SH2bfs(sh,(10<<4)+9);
}
#define init_8_15_10_10
void FASTCALL call_8_15_10_10(SH2_struct* sh)
{
  SH2bfs(sh,(10<<4)+10);
}
#define init_8_15_10_11
void FASTCALL call_8_15_10_11(SH2_struct* sh)
{
  SH2bfs(sh,(10<<4)+11);
}
#define init_8_15_10_12
void FASTCALL call_8_15_10_12(SH2_struct* sh)
{
  SH2bfs(sh,(10<<4)+12);
}
#define init_8_15_10_13
void FASTCALL call_8_15_10_13(SH2_struct* sh)
{
  SH2bfs(sh,(10<<4)+13);
}
#define init_8_15_10_14
void FASTCALL call_8_15_10_14(SH2_struct* sh)
{
  SH2bfs(sh,(10<<4)+14);
}
#define init_8_15_10_15
void FASTCALL call_8_15_10_15(SH2_struct* sh)
{
  SH2bfs(sh,(10<<4)+15);
}
#define init_8_15_11_0
void FASTCALL call_8_15_11_0(SH2_struct* sh)
{
  SH2bfs(sh,(11<<4)+0);
}
#define init_8_15_11_1
void FASTCALL call_8_15_11_1(SH2_struct* sh)
{
  SH2bfs(sh,(11<<4)+1);
}
#define init_8_15_11_2
void FASTCALL call_8_15_11_2(SH2_struct* sh)
{
  SH2bfs(sh,(11<<4)+2);
}
#define init_8_15_11_3
void FASTCALL call_8_15_11_3(SH2_struct* sh)
{
  SH2bfs(sh,(11<<4)+3);
}
#define init_8_15_11_4
void FASTCALL call_8_15_11_4(SH2_struct* sh)
{
  SH2bfs(sh,(11<<4)+4);
}
#define init_8_15_11_5
void FASTCALL call_8_15_11_5(SH2_struct* sh)
{
  SH2bfs(sh,(11<<4)+5);
}
#define init_8_15_11_6
void FASTCALL call_8_15_11_6(SH2_struct* sh)
{
  SH2bfs(sh,(11<<4)+6);
}
#define init_8_15_11_7
void FASTCALL call_8_15_11_7(SH2_struct* sh)
{
  SH2bfs(sh,(11<<4)+7);
}
#define init_8_15_11_8
void FASTCALL call_8_15_11_8(SH2_struct* sh)
{
  SH2bfs(sh,(11<<4)+8);
}
#define init_8_15_11_9
void FASTCALL call_8_15_11_9(SH2_struct* sh)
{
  SH2bfs(sh,(11<<4)+9);
}
#define init_8_15_11_10
void FASTCALL call_8_15_11_10(SH2_struct* sh)
{
  SH2bfs(sh,(11<<4)+10);
}
#define init_8_15_11_11
void FASTCALL call_8_15_11_11(SH2_struct* sh)
{
  SH2bfs(sh,(11<<4)+11);
}
#define init_8_15_11_12
void FASTCALL call_8_15_11_12(SH2_struct* sh)
{
  SH2bfs(sh,(11<<4)+12);
}
#define init_8_15_11_13
void FASTCALL call_8_15_11_13(SH2_struct* sh)
{
  SH2bfs(sh,(11<<4)+13);
}
#define init_8_15_11_14
void FASTCALL call_8_15_11_14(SH2_struct* sh)
{
  SH2bfs(sh,(11<<4)+14);
}
#define init_8_15_11_15
void FASTCALL call_8_15_11_15(SH2_struct* sh)
{
  SH2bfs(sh,(11<<4)+15);
}
#define init_8_15_12_0
void FASTCALL call_8_15_12_0(SH2_struct* sh)
{
  SH2bfs(sh,(12<<4)+0);
}
#define init_8_15_12_1
void FASTCALL call_8_15_12_1(SH2_struct* sh)
{
  SH2bfs(sh,(12<<4)+1);
}
#define init_8_15_12_2
void FASTCALL call_8_15_12_2(SH2_struct* sh)
{
  SH2bfs(sh,(12<<4)+2);
}
#define init_8_15_12_3
void FASTCALL call_8_15_12_3(SH2_struct* sh)
{
  SH2bfs(sh,(12<<4)+3);
}
#define init_8_15_12_4
void FASTCALL call_8_15_12_4(SH2_struct* sh)
{
  SH2bfs(sh,(12<<4)+4);
}
#define init_8_15_12_5
void FASTCALL call_8_15_12_5(SH2_struct* sh)
{
  SH2bfs(sh,(12<<4)+5);
}
#define init_8_15_12_6
void FASTCALL call_8_15_12_6(SH2_struct* sh)
{
  SH2bfs(sh,(12<<4)+6);
}
#define init_8_15_12_7
void FASTCALL call_8_15_12_7(SH2_struct* sh)
{
  SH2bfs(sh,(12<<4)+7);
}
#define init_8_15_12_8
void FASTCALL call_8_15_12_8(SH2_struct* sh)
{
  SH2bfs(sh,(12<<4)+8);
}
#define init_8_15_12_9
void FASTCALL call_8_15_12_9(SH2_struct* sh)
{
  SH2bfs(sh,(12<<4)+9);
}
#define init_8_15_12_10
void FASTCALL call_8_15_12_10(SH2_struct* sh)
{
  SH2bfs(sh,(12<<4)+10);
}
#define init_8_15_12_11
void FASTCALL call_8_15_12_11(SH2_struct* sh)
{
  SH2bfs(sh,(12<<4)+11);
}
#define init_8_15_12_12
void FASTCALL call_8_15_12_12(SH2_struct* sh)
{
  SH2bfs(sh,(12<<4)+12);
}
#define init_8_15_12_13
void FASTCALL call_8_15_12_13(SH2_struct* sh)
{
  SH2bfs(sh,(12<<4)+13);
}
#define init_8_15_12_14
void FASTCALL call_8_15_12_14(SH2_struct* sh)
{
  SH2bfs(sh,(12<<4)+14);
}
#define init_8_15_12_15
void FASTCALL call_8_15_12_15(SH2_struct* sh)
{
  SH2bfs(sh,(12<<4)+15);
}
#define init_8_15_13_0
void FASTCALL call_8_15_13_0(SH2_struct* sh)
{
  SH2bfs(sh,(13<<4)+0);
}
#define init_8_15_13_1
void FASTCALL call_8_15_13_1(SH2_struct* sh)
{
  SH2bfs(sh,(13<<4)+1);
}
#define init_8_15_13_2
void FASTCALL call_8_15_13_2(SH2_struct* sh)
{
  SH2bfs(sh,(13<<4)+2);
}
#define init_8_15_13_3
void FASTCALL call_8_15_13_3(SH2_struct* sh)
{
  SH2bfs(sh,(13<<4)+3);
}
#define init_8_15_13_4
void FASTCALL call_8_15_13_4(SH2_struct* sh)
{
  SH2bfs(sh,(13<<4)+4);
}
#define init_8_15_13_5
void FASTCALL call_8_15_13_5(SH2_struct* sh)
{
  SH2bfs(sh,(13<<4)+5);
}
#define init_8_15_13_6
void FASTCALL call_8_15_13_6(SH2_struct* sh)
{
  SH2bfs(sh,(13<<4)+6);
}
#define init_8_15_13_7
void FASTCALL call_8_15_13_7(SH2_struct* sh)
{
  SH2bfs(sh,(13<<4)+7);
}
#define init_8_15_13_8
void FASTCALL call_8_15_13_8(SH2_struct* sh)
{
  SH2bfs(sh,(13<<4)+8);
}
#define init_8_15_13_9
void FASTCALL call_8_15_13_9(SH2_struct* sh)
{
  SH2bfs(sh,(13<<4)+9);
}
#define init_8_15_13_10
void FASTCALL call_8_15_13_10(SH2_struct* sh)
{
  SH2bfs(sh,(13<<4)+10);
}
#define init_8_15_13_11
void FASTCALL call_8_15_13_11(SH2_struct* sh)
{
  SH2bfs(sh,(13<<4)+11);
}
#define init_8_15_13_12
void FASTCALL call_8_15_13_12(SH2_struct* sh)
{
  SH2bfs(sh,(13<<4)+12);
}
#define init_8_15_13_13
void FASTCALL call_8_15_13_13(SH2_struct* sh)
{
  SH2bfs(sh,(13<<4)+13);
}
#define init_8_15_13_14
void FASTCALL call_8_15_13_14(SH2_struct* sh)
{
  SH2bfs(sh,(13<<4)+14);
}
#define init_8_15_13_15
void FASTCALL call_8_15_13_15(SH2_struct* sh)
{
  SH2bfs(sh,(13<<4)+15);
}
#define init_8_15_14_0
void FASTCALL call_8_15_14_0(SH2_struct* sh)
{
  SH2bfs(sh,(14<<4)+0);
}
#define init_8_15_14_1
void FASTCALL call_8_15_14_1(SH2_struct* sh)
{
  SH2bfs(sh,(14<<4)+1);
}
#define init_8_15_14_2
void FASTCALL call_8_15_14_2(SH2_struct* sh)
{
  SH2bfs(sh,(14<<4)+2);
}
#define init_8_15_14_3
void FASTCALL call_8_15_14_3(SH2_struct* sh)
{
  SH2bfs(sh,(14<<4)+3);
}
#define init_8_15_14_4
void FASTCALL call_8_15_14_4(SH2_struct* sh)
{
  SH2bfs(sh,(14<<4)+4);
}
#define init_8_15_14_5
void FASTCALL call_8_15_14_5(SH2_struct* sh)
{
  SH2bfs(sh,(14<<4)+5);
}
#define init_8_15_14_6
void FASTCALL call_8_15_14_6(SH2_struct* sh)
{
  SH2bfs(sh,(14<<4)+6);
}
#define init_8_15_14_7
void FASTCALL call_8_15_14_7(SH2_struct* sh)
{
  SH2bfs(sh,(14<<4)+7);
}
#define init_8_15_14_8
void FASTCALL call_8_15_14_8(SH2_struct* sh)
{
  SH2bfs(sh,(14<<4)+8);
}
#define init_8_15_14_9
void FASTCALL call_8_15_14_9(SH2_struct* sh)
{
  SH2bfs(sh,(14<<4)+9);
}
#define init_8_15_14_10
void FASTCALL call_8_15_14_10(SH2_struct* sh)
{
  SH2bfs(sh,(14<<4)+10);
}
#define init_8_15_14_11
void FASTCALL call_8_15_14_11(SH2_struct* sh)
{
  SH2bfs(sh,(14<<4)+11);
}
#define init_8_15_14_12
void FASTCALL call_8_15_14_12(SH2_struct* sh)
{
  SH2bfs(sh,(14<<4)+12);
}
#define init_8_15_14_13
void FASTCALL call_8_15_14_13(SH2_struct* sh)
{
  SH2bfs(sh,(14<<4)+13);
}
#define init_8_15_14_14
void FASTCALL call_8_15_14_14(SH2_struct* sh)
{
  SH2bfs(sh,(14<<4)+14);
}
#define init_8_15_14_15
void FASTCALL call_8_15_14_15(SH2_struct* sh)
{
  SH2bfs(sh,(14<<4)+15);
}
#define init_8_15_15_0
void FASTCALL call_8_15_15_0(SH2_struct* sh)
{
  SH2bfs(sh,(15<<4)+0);
}
#define init_8_15_15_1
void FASTCALL call_8_15_15_1(SH2_struct* sh)
{
  SH2bfs(sh,(15<<4)+1);
}
#define init_8_15_15_2
void FASTCALL call_8_15_15_2(SH2_struct* sh)
{
  SH2bfs(sh,(15<<4)+2);
}
#define init_8_15_15_3
void FASTCALL call_8_15_15_3(SH2_struct* sh)
{
  SH2bfs(sh,(15<<4)+3);
}
#define init_8_15_15_4
void FASTCALL call_8_15_15_4(SH2_struct* sh)
{
  SH2bfs(sh,(15<<4)+4);
}
#define init_8_15_15_5
void FASTCALL call_8_15_15_5(SH2_struct* sh)
{
  SH2bfs(sh,(15<<4)+5);
}
#define init_8_15_15_6
void FASTCALL call_8_15_15_6(SH2_struct* sh)
{
  SH2bfs(sh,(15<<4)+6);
}
#define init_8_15_15_7
void FASTCALL call_8_15_15_7(SH2_struct* sh)
{
  SH2bfs(sh,(15<<4)+7);
}
#define init_8_15_15_8
void FASTCALL call_8_15_15_8(SH2_struct* sh)
{
  SH2bfs(sh,(15<<4)+8);
}
#define init_8_15_15_9
void FASTCALL call_8_15_15_9(SH2_struct* sh)
{
  SH2bfs(sh,(15<<4)+9);
}
#define init_8_15_15_10
void FASTCALL call_8_15_15_10(SH2_struct* sh)
{
  SH2bfs(sh,(15<<4)+10);
}
#define init_8_15_15_11
void FASTCALL call_8_15_15_11(SH2_struct* sh)
{
  SH2bfs(sh,(15<<4)+11);
}
#define init_8_15_15_12
void FASTCALL call_8_15_15_12(SH2_struct* sh)
{
  SH2bfs(sh,(15<<4)+12);
}
#define init_8_15_15_13
void FASTCALL call_8_15_15_13(SH2_struct* sh)
{
  SH2bfs(sh,(15<<4)+13);
}
#define init_8_15_15_14
void FASTCALL call_8_15_15_14(SH2_struct* sh)
{
  SH2bfs(sh,(15<<4)+14);
}
#define init_8_15_15_15
void FASTCALL call_8_15_15_15(SH2_struct* sh)
{
  SH2bfs(sh,(15<<4)+15);
}
#define init_8_11_0_0
void FASTCALL call_8_11_0_0(SH2_struct* sh)
{
  SH2bf(sh,(0<<4)+0);
}
#define init_8_11_0_1
void FASTCALL call_8_11_0_1(SH2_struct* sh)
{
  SH2bf(sh,(0<<4)+1);
}
#define init_8_11_0_2
void FASTCALL call_8_11_0_2(SH2_struct* sh)
{
  SH2bf(sh,(0<<4)+2);
}
#define init_8_11_0_3
void FASTCALL call_8_11_0_3(SH2_struct* sh)
{
  SH2bf(sh,(0<<4)+3);
}
#define init_8_11_0_4
void FASTCALL call_8_11_0_4(SH2_struct* sh)
{
  SH2bf(sh,(0<<4)+4);
}
#define init_8_11_0_5
void FASTCALL call_8_11_0_5(SH2_struct* sh)
{
  SH2bf(sh,(0<<4)+5);
}
#define init_8_11_0_6
void FASTCALL call_8_11_0_6(SH2_struct* sh)
{
  SH2bf(sh,(0<<4)+6);
}
#define init_8_11_0_7
void FASTCALL call_8_11_0_7(SH2_struct* sh)
{
  SH2bf(sh,(0<<4)+7);
}
#define init_8_11_0_8
void FASTCALL call_8_11_0_8(SH2_struct* sh)
{
  SH2bf(sh,(0<<4)+8);
}
#define init_8_11_0_9
void FASTCALL call_8_11_0_9(SH2_struct* sh)
{
  SH2bf(sh,(0<<4)+9);
}
#define init_8_11_0_10
void FASTCALL call_8_11_0_10(SH2_struct* sh)
{
  SH2bf(sh,(0<<4)+10);
}
#define init_8_11_0_11
void FASTCALL call_8_11_0_11(SH2_struct* sh)
{
  SH2bf(sh,(0<<4)+11);
}
#define init_8_11_0_12
void FASTCALL call_8_11_0_12(SH2_struct* sh)
{
  SH2bf(sh,(0<<4)+12);
}
#define init_8_11_0_13
void FASTCALL call_8_11_0_13(SH2_struct* sh)
{
  SH2bf(sh,(0<<4)+13);
}
#define init_8_11_0_14
void FASTCALL call_8_11_0_14(SH2_struct* sh)
{
  SH2bf(sh,(0<<4)+14);
}
#define init_8_11_0_15
void FASTCALL call_8_11_0_15(SH2_struct* sh)
{
  SH2bf(sh,(0<<4)+15);
}
#define init_8_11_1_0
void FASTCALL call_8_11_1_0(SH2_struct* sh)
{
  SH2bf(sh,(1<<4)+0);
}
#define init_8_11_1_1
void FASTCALL call_8_11_1_1(SH2_struct* sh)
{
  SH2bf(sh,(1<<4)+1);
}
#define init_8_11_1_2
void FASTCALL call_8_11_1_2(SH2_struct* sh)
{
  SH2bf(sh,(1<<4)+2);
}
#define init_8_11_1_3
void FASTCALL call_8_11_1_3(SH2_struct* sh)
{
  SH2bf(sh,(1<<4)+3);
}
#define init_8_11_1_4
void FASTCALL call_8_11_1_4(SH2_struct* sh)
{
  SH2bf(sh,(1<<4)+4);
}
#define init_8_11_1_5
void FASTCALL call_8_11_1_5(SH2_struct* sh)
{
  SH2bf(sh,(1<<4)+5);
}
#define init_8_11_1_6
void FASTCALL call_8_11_1_6(SH2_struct* sh)
{
  SH2bf(sh,(1<<4)+6);
}
#define init_8_11_1_7
void FASTCALL call_8_11_1_7(SH2_struct* sh)
{
  SH2bf(sh,(1<<4)+7);
}
#define init_8_11_1_8
void FASTCALL call_8_11_1_8(SH2_struct* sh)
{
  SH2bf(sh,(1<<4)+8);
}
#define init_8_11_1_9
void FASTCALL call_8_11_1_9(SH2_struct* sh)
{
  SH2bf(sh,(1<<4)+9);
}
#define init_8_11_1_10
void FASTCALL call_8_11_1_10(SH2_struct* sh)
{
  SH2bf(sh,(1<<4)+10);
}
#define init_8_11_1_11
void FASTCALL call_8_11_1_11(SH2_struct* sh)
{
  SH2bf(sh,(1<<4)+11);
}
#define init_8_11_1_12
void FASTCALL call_8_11_1_12(SH2_struct* sh)
{
  SH2bf(sh,(1<<4)+12);
}
#define init_8_11_1_13
void FASTCALL call_8_11_1_13(SH2_struct* sh)
{
  SH2bf(sh,(1<<4)+13);
}
#define init_8_11_1_14
void FASTCALL call_8_11_1_14(SH2_struct* sh)
{
  SH2bf(sh,(1<<4)+14);
}
#define init_8_11_1_15
void FASTCALL call_8_11_1_15(SH2_struct* sh)
{
  SH2bf(sh,(1<<4)+15);
}
#define init_8_11_2_0
void FASTCALL call_8_11_2_0(SH2_struct* sh)
{
  SH2bf(sh,(2<<4)+0);
}
#define init_8_11_2_1
void FASTCALL call_8_11_2_1(SH2_struct* sh)
{
  SH2bf(sh,(2<<4)+1);
}
#define init_8_11_2_2
void FASTCALL call_8_11_2_2(SH2_struct* sh)
{
  SH2bf(sh,(2<<4)+2);
}
#define init_8_11_2_3
void FASTCALL call_8_11_2_3(SH2_struct* sh)
{
  SH2bf(sh,(2<<4)+3);
}
#define init_8_11_2_4
void FASTCALL call_8_11_2_4(SH2_struct* sh)
{
  SH2bf(sh,(2<<4)+4);
}
#define init_8_11_2_5
void FASTCALL call_8_11_2_5(SH2_struct* sh)
{
  SH2bf(sh,(2<<4)+5);
}
#define init_8_11_2_6
void FASTCALL call_8_11_2_6(SH2_struct* sh)
{
  SH2bf(sh,(2<<4)+6);
}
#define init_8_11_2_7
void FASTCALL call_8_11_2_7(SH2_struct* sh)
{
  SH2bf(sh,(2<<4)+7);
}
#define init_8_11_2_8
void FASTCALL call_8_11_2_8(SH2_struct* sh)
{
  SH2bf(sh,(2<<4)+8);
}
#define init_8_11_2_9
void FASTCALL call_8_11_2_9(SH2_struct* sh)
{
  SH2bf(sh,(2<<4)+9);
}
#define init_8_11_2_10
void FASTCALL call_8_11_2_10(SH2_struct* sh)
{
  SH2bf(sh,(2<<4)+10);
}
#define init_8_11_2_11
void FASTCALL call_8_11_2_11(SH2_struct* sh)
{
  SH2bf(sh,(2<<4)+11);
}
#define init_8_11_2_12
void FASTCALL call_8_11_2_12(SH2_struct* sh)
{
  SH2bf(sh,(2<<4)+12);
}
#define init_8_11_2_13
void FASTCALL call_8_11_2_13(SH2_struct* sh)
{
  SH2bf(sh,(2<<4)+13);
}
#define init_8_11_2_14
void FASTCALL call_8_11_2_14(SH2_struct* sh)
{
  SH2bf(sh,(2<<4)+14);
}
#define init_8_11_2_15
void FASTCALL call_8_11_2_15(SH2_struct* sh)
{
  SH2bf(sh,(2<<4)+15);
}
#define init_8_11_3_0
void FASTCALL call_8_11_3_0(SH2_struct* sh)
{
  SH2bf(sh,(3<<4)+0);
}
#define init_8_11_3_1
void FASTCALL call_8_11_3_1(SH2_struct* sh)
{
  SH2bf(sh,(3<<4)+1);
}
#define init_8_11_3_2
void FASTCALL call_8_11_3_2(SH2_struct* sh)
{
  SH2bf(sh,(3<<4)+2);
}
#define init_8_11_3_3
void FASTCALL call_8_11_3_3(SH2_struct* sh)
{
  SH2bf(sh,(3<<4)+3);
}
#define init_8_11_3_4
void FASTCALL call_8_11_3_4(SH2_struct* sh)
{
  SH2bf(sh,(3<<4)+4);
}
#define init_8_11_3_5
void FASTCALL call_8_11_3_5(SH2_struct* sh)
{
  SH2bf(sh,(3<<4)+5);
}
#define init_8_11_3_6
void FASTCALL call_8_11_3_6(SH2_struct* sh)
{
  SH2bf(sh,(3<<4)+6);
}
#define init_8_11_3_7
void FASTCALL call_8_11_3_7(SH2_struct* sh)
{
  SH2bf(sh,(3<<4)+7);
}
#define init_8_11_3_8
void FASTCALL call_8_11_3_8(SH2_struct* sh)
{
  SH2bf(sh,(3<<4)+8);
}
#define init_8_11_3_9
void FASTCALL call_8_11_3_9(SH2_struct* sh)
{
  SH2bf(sh,(3<<4)+9);
}
#define init_8_11_3_10
void FASTCALL call_8_11_3_10(SH2_struct* sh)
{
  SH2bf(sh,(3<<4)+10);
}
#define init_8_11_3_11
void FASTCALL call_8_11_3_11(SH2_struct* sh)
{
  SH2bf(sh,(3<<4)+11);
}
#define init_8_11_3_12
void FASTCALL call_8_11_3_12(SH2_struct* sh)
{
  SH2bf(sh,(3<<4)+12);
}
#define init_8_11_3_13
void FASTCALL call_8_11_3_13(SH2_struct* sh)
{
  SH2bf(sh,(3<<4)+13);
}
#define init_8_11_3_14
void FASTCALL call_8_11_3_14(SH2_struct* sh)
{
  SH2bf(sh,(3<<4)+14);
}
#define init_8_11_3_15
void FASTCALL call_8_11_3_15(SH2_struct* sh)
{
  SH2bf(sh,(3<<4)+15);
}
#define init_8_11_4_0
void FASTCALL call_8_11_4_0(SH2_struct* sh)
{
  SH2bf(sh,(4<<4)+0);
}
#define init_8_11_4_1
void FASTCALL call_8_11_4_1(SH2_struct* sh)
{
  SH2bf(sh,(4<<4)+1);
}
#define init_8_11_4_2
void FASTCALL call_8_11_4_2(SH2_struct* sh)
{
  SH2bf(sh,(4<<4)+2);
}
#define init_8_11_4_3
void FASTCALL call_8_11_4_3(SH2_struct* sh)
{
  SH2bf(sh,(4<<4)+3);
}
#define init_8_11_4_4
void FASTCALL call_8_11_4_4(SH2_struct* sh)
{
  SH2bf(sh,(4<<4)+4);
}
#define init_8_11_4_5
void FASTCALL call_8_11_4_5(SH2_struct* sh)
{
  SH2bf(sh,(4<<4)+5);
}
#define init_8_11_4_6
void FASTCALL call_8_11_4_6(SH2_struct* sh)
{
  SH2bf(sh,(4<<4)+6);
}
#define init_8_11_4_7
void FASTCALL call_8_11_4_7(SH2_struct* sh)
{
  SH2bf(sh,(4<<4)+7);
}
#define init_8_11_4_8
void FASTCALL call_8_11_4_8(SH2_struct* sh)
{
  SH2bf(sh,(4<<4)+8);
}
#define init_8_11_4_9
void FASTCALL call_8_11_4_9(SH2_struct* sh)
{
  SH2bf(sh,(4<<4)+9);
}
#define init_8_11_4_10
void FASTCALL call_8_11_4_10(SH2_struct* sh)
{
  SH2bf(sh,(4<<4)+10);
}
#define init_8_11_4_11
void FASTCALL call_8_11_4_11(SH2_struct* sh)
{
  SH2bf(sh,(4<<4)+11);
}
#define init_8_11_4_12
void FASTCALL call_8_11_4_12(SH2_struct* sh)
{
  SH2bf(sh,(4<<4)+12);
}
#define init_8_11_4_13
void FASTCALL call_8_11_4_13(SH2_struct* sh)
{
  SH2bf(sh,(4<<4)+13);
}
#define init_8_11_4_14
void FASTCALL call_8_11_4_14(SH2_struct* sh)
{
  SH2bf(sh,(4<<4)+14);
}
#define init_8_11_4_15
void FASTCALL call_8_11_4_15(SH2_struct* sh)
{
  SH2bf(sh,(4<<4)+15);
}
#define init_8_11_5_0
void FASTCALL call_8_11_5_0(SH2_struct* sh)
{
  SH2bf(sh,(5<<4)+0);
}
#define init_8_11_5_1
void FASTCALL call_8_11_5_1(SH2_struct* sh)
{
  SH2bf(sh,(5<<4)+1);
}
#define init_8_11_5_2
void FASTCALL call_8_11_5_2(SH2_struct* sh)
{
  SH2bf(sh,(5<<4)+2);
}
#define init_8_11_5_3
void FASTCALL call_8_11_5_3(SH2_struct* sh)
{
  SH2bf(sh,(5<<4)+3);
}
#define init_8_11_5_4
void FASTCALL call_8_11_5_4(SH2_struct* sh)
{
  SH2bf(sh,(5<<4)+4);
}
#define init_8_11_5_5
void FASTCALL call_8_11_5_5(SH2_struct* sh)
{
  SH2bf(sh,(5<<4)+5);
}
#define init_8_11_5_6
void FASTCALL call_8_11_5_6(SH2_struct* sh)
{
  SH2bf(sh,(5<<4)+6);
}
#define init_8_11_5_7
void FASTCALL call_8_11_5_7(SH2_struct* sh)
{
  SH2bf(sh,(5<<4)+7);
}
#define init_8_11_5_8
void FASTCALL call_8_11_5_8(SH2_struct* sh)
{
  SH2bf(sh,(5<<4)+8);
}
#define init_8_11_5_9
void FASTCALL call_8_11_5_9(SH2_struct* sh)
{
  SH2bf(sh,(5<<4)+9);
}
#define init_8_11_5_10
void FASTCALL call_8_11_5_10(SH2_struct* sh)
{
  SH2bf(sh,(5<<4)+10);
}
#define init_8_11_5_11
void FASTCALL call_8_11_5_11(SH2_struct* sh)
{
  SH2bf(sh,(5<<4)+11);
}
#define init_8_11_5_12
void FASTCALL call_8_11_5_12(SH2_struct* sh)
{
  SH2bf(sh,(5<<4)+12);
}
#define init_8_11_5_13
void FASTCALL call_8_11_5_13(SH2_struct* sh)
{
  SH2bf(sh,(5<<4)+13);
}
#define init_8_11_5_14
void FASTCALL call_8_11_5_14(SH2_struct* sh)
{
  SH2bf(sh,(5<<4)+14);
}
#define init_8_11_5_15
void FASTCALL call_8_11_5_15(SH2_struct* sh)
{
  SH2bf(sh,(5<<4)+15);
}
#define init_8_11_6_0
void FASTCALL call_8_11_6_0(SH2_struct* sh)
{
  SH2bf(sh,(6<<4)+0);
}
#define init_8_11_6_1
void FASTCALL call_8_11_6_1(SH2_struct* sh)
{
  SH2bf(sh,(6<<4)+1);
}
#define init_8_11_6_2
void FASTCALL call_8_11_6_2(SH2_struct* sh)
{
  SH2bf(sh,(6<<4)+2);
}
#define init_8_11_6_3
void FASTCALL call_8_11_6_3(SH2_struct* sh)
{
  SH2bf(sh,(6<<4)+3);
}
#define init_8_11_6_4
void FASTCALL call_8_11_6_4(SH2_struct* sh)
{
  SH2bf(sh,(6<<4)+4);
}
#define init_8_11_6_5
void FASTCALL call_8_11_6_5(SH2_struct* sh)
{
  SH2bf(sh,(6<<4)+5);
}
#define init_8_11_6_6
void FASTCALL call_8_11_6_6(SH2_struct* sh)
{
  SH2bf(sh,(6<<4)+6);
}
#define init_8_11_6_7
void FASTCALL call_8_11_6_7(SH2_struct* sh)
{
  SH2bf(sh,(6<<4)+7);
}
#define init_8_11_6_8
void FASTCALL call_8_11_6_8(SH2_struct* sh)
{
  SH2bf(sh,(6<<4)+8);
}
#define init_8_11_6_9
void FASTCALL call_8_11_6_9(SH2_struct* sh)
{
  SH2bf(sh,(6<<4)+9);
}
#define init_8_11_6_10
void FASTCALL call_8_11_6_10(SH2_struct* sh)
{
  SH2bf(sh,(6<<4)+10);
}
#define init_8_11_6_11
void FASTCALL call_8_11_6_11(SH2_struct* sh)
{
  SH2bf(sh,(6<<4)+11);
}
#define init_8_11_6_12
void FASTCALL call_8_11_6_12(SH2_struct* sh)
{
  SH2bf(sh,(6<<4)+12);
}
#define init_8_11_6_13
void FASTCALL call_8_11_6_13(SH2_struct* sh)
{
  SH2bf(sh,(6<<4)+13);
}
#define init_8_11_6_14
void FASTCALL call_8_11_6_14(SH2_struct* sh)
{
  SH2bf(sh,(6<<4)+14);
}
#define init_8_11_6_15
void FASTCALL call_8_11_6_15(SH2_struct* sh)
{
  SH2bf(sh,(6<<4)+15);
}
#define init_8_11_7_0
void FASTCALL call_8_11_7_0(SH2_struct* sh)
{
  SH2bf(sh,(7<<4)+0);
}
#define init_8_11_7_1
void FASTCALL call_8_11_7_1(SH2_struct* sh)
{
  SH2bf(sh,(7<<4)+1);
}
#define init_8_11_7_2
void FASTCALL call_8_11_7_2(SH2_struct* sh)
{
  SH2bf(sh,(7<<4)+2);
}
#define init_8_11_7_3
void FASTCALL call_8_11_7_3(SH2_struct* sh)
{
  SH2bf(sh,(7<<4)+3);
}
#define init_8_11_7_4
void FASTCALL call_8_11_7_4(SH2_struct* sh)
{
  SH2bf(sh,(7<<4)+4);
}
#define init_8_11_7_5
void FASTCALL call_8_11_7_5(SH2_struct* sh)
{
  SH2bf(sh,(7<<4)+5);
}
#define init_8_11_7_6
void FASTCALL call_8_11_7_6(SH2_struct* sh)
{
  SH2bf(sh,(7<<4)+6);
}
#define init_8_11_7_7
void FASTCALL call_8_11_7_7(SH2_struct* sh)
{
  SH2bf(sh,(7<<4)+7);
}
#define init_8_11_7_8
void FASTCALL call_8_11_7_8(SH2_struct* sh)
{
  SH2bf(sh,(7<<4)+8);
}
#define init_8_11_7_9
void FASTCALL call_8_11_7_9(SH2_struct* sh)
{
  SH2bf(sh,(7<<4)+9);
}
#define init_8_11_7_10
void FASTCALL call_8_11_7_10(SH2_struct* sh)
{
  SH2bf(sh,(7<<4)+10);
}
#define init_8_11_7_11
void FASTCALL call_8_11_7_11(SH2_struct* sh)
{
  SH2bf(sh,(7<<4)+11);
}
#define init_8_11_7_12
void FASTCALL call_8_11_7_12(SH2_struct* sh)
{
  SH2bf(sh,(7<<4)+12);
}
#define init_8_11_7_13
void FASTCALL call_8_11_7_13(SH2_struct* sh)
{
  SH2bf(sh,(7<<4)+13);
}
#define init_8_11_7_14
void FASTCALL call_8_11_7_14(SH2_struct* sh)
{
  SH2bf(sh,(7<<4)+14);
}
#define init_8_11_7_15
void FASTCALL call_8_11_7_15(SH2_struct* sh)
{
  SH2bf(sh,(7<<4)+15);
}
#define init_8_11_8_0
void FASTCALL call_8_11_8_0(SH2_struct* sh)
{
  SH2bf(sh,(8<<4)+0);
}
#define init_8_11_8_1
void FASTCALL call_8_11_8_1(SH2_struct* sh)
{
  SH2bf(sh,(8<<4)+1);
}
#define init_8_11_8_2
void FASTCALL call_8_11_8_2(SH2_struct* sh)
{
  SH2bf(sh,(8<<4)+2);
}
#define init_8_11_8_3
void FASTCALL call_8_11_8_3(SH2_struct* sh)
{
  SH2bf(sh,(8<<4)+3);
}
#define init_8_11_8_4
void FASTCALL call_8_11_8_4(SH2_struct* sh)
{
  SH2bf(sh,(8<<4)+4);
}
#define init_8_11_8_5
void FASTCALL call_8_11_8_5(SH2_struct* sh)
{
  SH2bf(sh,(8<<4)+5);
}
#define init_8_11_8_6
void FASTCALL call_8_11_8_6(SH2_struct* sh)
{
  SH2bf(sh,(8<<4)+6);
}
#define init_8_11_8_7
void FASTCALL call_8_11_8_7(SH2_struct* sh)
{
  SH2bf(sh,(8<<4)+7);
}
#define init_8_11_8_8
void FASTCALL call_8_11_8_8(SH2_struct* sh)
{
  SH2bf(sh,(8<<4)+8);
}
#define init_8_11_8_9
void FASTCALL call_8_11_8_9(SH2_struct* sh)
{
  SH2bf(sh,(8<<4)+9);
}
#define init_8_11_8_10
void FASTCALL call_8_11_8_10(SH2_struct* sh)
{
  SH2bf(sh,(8<<4)+10);
}
#define init_8_11_8_11
void FASTCALL call_8_11_8_11(SH2_struct* sh)
{
  SH2bf(sh,(8<<4)+11);
}
#define init_8_11_8_12
void FASTCALL call_8_11_8_12(SH2_struct* sh)
{
  SH2bf(sh,(8<<4)+12);
}
#define init_8_11_8_13
void FASTCALL call_8_11_8_13(SH2_struct* sh)
{
  SH2bf(sh,(8<<4)+13);
}
#define init_8_11_8_14
void FASTCALL call_8_11_8_14(SH2_struct* sh)
{
  SH2bf(sh,(8<<4)+14);
}
#define init_8_11_8_15
void FASTCALL call_8_11_8_15(SH2_struct* sh)
{
  SH2bf(sh,(8<<4)+15);
}
#define init_8_11_9_0
void FASTCALL call_8_11_9_0(SH2_struct* sh)
{
  SH2bf(sh,(9<<4)+0);
}
#define init_8_11_9_1
void FASTCALL call_8_11_9_1(SH2_struct* sh)
{
  SH2bf(sh,(9<<4)+1);
}
#define init_8_11_9_2
void FASTCALL call_8_11_9_2(SH2_struct* sh)
{
  SH2bf(sh,(9<<4)+2);
}
#define init_8_11_9_3
void FASTCALL call_8_11_9_3(SH2_struct* sh)
{
  SH2bf(sh,(9<<4)+3);
}
#define init_8_11_9_4
void FASTCALL call_8_11_9_4(SH2_struct* sh)
{
  SH2bf(sh,(9<<4)+4);
}
#define init_8_11_9_5
void FASTCALL call_8_11_9_5(SH2_struct* sh)
{
  SH2bf(sh,(9<<4)+5);
}
#define init_8_11_9_6
void FASTCALL call_8_11_9_6(SH2_struct* sh)
{
  SH2bf(sh,(9<<4)+6);
}
#define init_8_11_9_7
void FASTCALL call_8_11_9_7(SH2_struct* sh)
{
  SH2bf(sh,(9<<4)+7);
}
#define init_8_11_9_8
void FASTCALL call_8_11_9_8(SH2_struct* sh)
{
  SH2bf(sh,(9<<4)+8);
}
#define init_8_11_9_9
void FASTCALL call_8_11_9_9(SH2_struct* sh)
{
  SH2bf(sh,(9<<4)+9);
}
#define init_8_11_9_10
void FASTCALL call_8_11_9_10(SH2_struct* sh)
{
  SH2bf(sh,(9<<4)+10);
}
#define init_8_11_9_11
void FASTCALL call_8_11_9_11(SH2_struct* sh)
{
  SH2bf(sh,(9<<4)+11);
}
#define init_8_11_9_12
void FASTCALL call_8_11_9_12(SH2_struct* sh)
{
  SH2bf(sh,(9<<4)+12);
}
#define init_8_11_9_13
void FASTCALL call_8_11_9_13(SH2_struct* sh)
{
  SH2bf(sh,(9<<4)+13);
}
#define init_8_11_9_14
void FASTCALL call_8_11_9_14(SH2_struct* sh)
{
  SH2bf(sh,(9<<4)+14);
}
#define init_8_11_9_15
void FASTCALL call_8_11_9_15(SH2_struct* sh)
{
  SH2bf(sh,(9<<4)+15);
}
#define init_8_11_10_0
void FASTCALL call_8_11_10_0(SH2_struct* sh)
{
  SH2bf(sh,(10<<4)+0);
}
#define init_8_11_10_1
void FASTCALL call_8_11_10_1(SH2_struct* sh)
{
  SH2bf(sh,(10<<4)+1);
}
#define init_8_11_10_2
void FASTCALL call_8_11_10_2(SH2_struct* sh)
{
  SH2bf(sh,(10<<4)+2);
}
#define init_8_11_10_3
void FASTCALL call_8_11_10_3(SH2_struct* sh)
{
  SH2bf(sh,(10<<4)+3);
}
#define init_8_11_10_4
void FASTCALL call_8_11_10_4(SH2_struct* sh)
{
  SH2bf(sh,(10<<4)+4);
}
#define init_8_11_10_5
void FASTCALL call_8_11_10_5(SH2_struct* sh)
{
  SH2bf(sh,(10<<4)+5);
}
#define init_8_11_10_6
void FASTCALL call_8_11_10_6(SH2_struct* sh)
{
  SH2bf(sh,(10<<4)+6);
}
#define init_8_11_10_7
void FASTCALL call_8_11_10_7(SH2_struct* sh)
{
  SH2bf(sh,(10<<4)+7);
}
#define init_8_11_10_8
void FASTCALL call_8_11_10_8(SH2_struct* sh)
{
  SH2bf(sh,(10<<4)+8);
}
#define init_8_11_10_9
void FASTCALL call_8_11_10_9(SH2_struct* sh)
{
  SH2bf(sh,(10<<4)+9);
}
#define init_8_11_10_10
void FASTCALL call_8_11_10_10(SH2_struct* sh)
{
  SH2bf(sh,(10<<4)+10);
}
#define init_8_11_10_11
void FASTCALL call_8_11_10_11(SH2_struct* sh)
{
  SH2bf(sh,(10<<4)+11);
}
#define init_8_11_10_12
void FASTCALL call_8_11_10_12(SH2_struct* sh)
{
  SH2bf(sh,(10<<4)+12);
}
#define init_8_11_10_13
void FASTCALL call_8_11_10_13(SH2_struct* sh)
{
  SH2bf(sh,(10<<4)+13);
}
#define init_8_11_10_14
void FASTCALL call_8_11_10_14(SH2_struct* sh)
{
  SH2bf(sh,(10<<4)+14);
}
#define init_8_11_10_15
void FASTCALL call_8_11_10_15(SH2_struct* sh)
{
  SH2bf(sh,(10<<4)+15);
}
#define init_8_11_11_0
void FASTCALL call_8_11_11_0(SH2_struct* sh)
{
  SH2bf(sh,(11<<4)+0);
}
#define init_8_11_11_1
void FASTCALL call_8_11_11_1(SH2_struct* sh)
{
  SH2bf(sh,(11<<4)+1);
}
#define init_8_11_11_2
void FASTCALL call_8_11_11_2(SH2_struct* sh)
{
  SH2bf(sh,(11<<4)+2);
}
#define init_8_11_11_3
void FASTCALL call_8_11_11_3(SH2_struct* sh)
{
  SH2bf(sh,(11<<4)+3);
}
#define init_8_11_11_4
void FASTCALL call_8_11_11_4(SH2_struct* sh)
{
  SH2bf(sh,(11<<4)+4);
}
#define init_8_11_11_5
void FASTCALL call_8_11_11_5(SH2_struct* sh)
{
  SH2bf(sh,(11<<4)+5);
}
#define init_8_11_11_6
void FASTCALL call_8_11_11_6(SH2_struct* sh)
{
  SH2bf(sh,(11<<4)+6);
}
#define init_8_11_11_7
void FASTCALL call_8_11_11_7(SH2_struct* sh)
{
  SH2bf(sh,(11<<4)+7);
}
#define init_8_11_11_8
void FASTCALL call_8_11_11_8(SH2_struct* sh)
{
  SH2bf(sh,(11<<4)+8);
}
#define init_8_11_11_9
void FASTCALL call_8_11_11_9(SH2_struct* sh)
{
  SH2bf(sh,(11<<4)+9);
}
#define init_8_11_11_10
void FASTCALL call_8_11_11_10(SH2_struct* sh)
{
  SH2bf(sh,(11<<4)+10);
}
#define init_8_11_11_11
void FASTCALL call_8_11_11_11(SH2_struct* sh)
{
  SH2bf(sh,(11<<4)+11);
}
#define init_8_11_11_12
void FASTCALL call_8_11_11_12(SH2_struct* sh)
{
  SH2bf(sh,(11<<4)+12);
}
#define init_8_11_11_13
void FASTCALL call_8_11_11_13(SH2_struct* sh)
{
  SH2bf(sh,(11<<4)+13);
}
#define init_8_11_11_14
void FASTCALL call_8_11_11_14(SH2_struct* sh)
{
  SH2bf(sh,(11<<4)+14);
}
#define init_8_11_11_15
void FASTCALL call_8_11_11_15(SH2_struct* sh)
{
  SH2bf(sh,(11<<4)+15);
}
#define init_8_11_12_0
void FASTCALL call_8_11_12_0(SH2_struct* sh)
{
  SH2bf(sh,(12<<4)+0);
}
#define init_8_11_12_1
void FASTCALL call_8_11_12_1(SH2_struct* sh)
{
  SH2bf(sh,(12<<4)+1);
}
#define init_8_11_12_2
void FASTCALL call_8_11_12_2(SH2_struct* sh)
{
  SH2bf(sh,(12<<4)+2);
}
#define init_8_11_12_3
void FASTCALL call_8_11_12_3(SH2_struct* sh)
{
  SH2bf(sh,(12<<4)+3);
}
#define init_8_11_12_4
void FASTCALL call_8_11_12_4(SH2_struct* sh)
{
  SH2bf(sh,(12<<4)+4);
}
#define init_8_11_12_5
void FASTCALL call_8_11_12_5(SH2_struct* sh)
{
  SH2bf(sh,(12<<4)+5);
}
#define init_8_11_12_6
void FASTCALL call_8_11_12_6(SH2_struct* sh)
{
  SH2bf(sh,(12<<4)+6);
}
#define init_8_11_12_7
void FASTCALL call_8_11_12_7(SH2_struct* sh)
{
  SH2bf(sh,(12<<4)+7);
}
#define init_8_11_12_8
void FASTCALL call_8_11_12_8(SH2_struct* sh)
{
  SH2bf(sh,(12<<4)+8);
}
#define init_8_11_12_9
void FASTCALL call_8_11_12_9(SH2_struct* sh)
{
  SH2bf(sh,(12<<4)+9);
}
#define init_8_11_12_10
void FASTCALL call_8_11_12_10(SH2_struct* sh)
{
  SH2bf(sh,(12<<4)+10);
}
#define init_8_11_12_11
void FASTCALL call_8_11_12_11(SH2_struct* sh)
{
  SH2bf(sh,(12<<4)+11);
}
#define init_8_11_12_12
void FASTCALL call_8_11_12_12(SH2_struct* sh)
{
  SH2bf(sh,(12<<4)+12);
}
#define init_8_11_12_13
void FASTCALL call_8_11_12_13(SH2_struct* sh)
{
  SH2bf(sh,(12<<4)+13);
}
#define init_8_11_12_14
void FASTCALL call_8_11_12_14(SH2_struct* sh)
{
  SH2bf(sh,(12<<4)+14);
}
#define init_8_11_12_15
void FASTCALL call_8_11_12_15(SH2_struct* sh)
{
  SH2bf(sh,(12<<4)+15);
}
#define init_8_11_13_0
void FASTCALL call_8_11_13_0(SH2_struct* sh)
{
  SH2bf(sh,(13<<4)+0);
}
#define init_8_11_13_1
void FASTCALL call_8_11_13_1(SH2_struct* sh)
{
  SH2bf(sh,(13<<4)+1);
}
#define init_8_11_13_2
void FASTCALL call_8_11_13_2(SH2_struct* sh)
{
  SH2bf(sh,(13<<4)+2);
}
#define init_8_11_13_3
void FASTCALL call_8_11_13_3(SH2_struct* sh)
{
  SH2bf(sh,(13<<4)+3);
}
#define init_8_11_13_4
void FASTCALL call_8_11_13_4(SH2_struct* sh)
{
  SH2bf(sh,(13<<4)+4);
}
#define init_8_11_13_5
void FASTCALL call_8_11_13_5(SH2_struct* sh)
{
  SH2bf(sh,(13<<4)+5);
}
#define init_8_11_13_6
void FASTCALL call_8_11_13_6(SH2_struct* sh)
{
  SH2bf(sh,(13<<4)+6);
}
#define init_8_11_13_7
void FASTCALL call_8_11_13_7(SH2_struct* sh)
{
  SH2bf(sh,(13<<4)+7);
}
#define init_8_11_13_8
void FASTCALL call_8_11_13_8(SH2_struct* sh)
{
  SH2bf(sh,(13<<4)+8);
}
#define init_8_11_13_9
void FASTCALL call_8_11_13_9(SH2_struct* sh)
{
  SH2bf(sh,(13<<4)+9);
}
#define init_8_11_13_10
void FASTCALL call_8_11_13_10(SH2_struct* sh)
{
  SH2bf(sh,(13<<4)+10);
}
#define init_8_11_13_11
void FASTCALL call_8_11_13_11(SH2_struct* sh)
{
  SH2bf(sh,(13<<4)+11);
}
#define init_8_11_13_12
void FASTCALL call_8_11_13_12(SH2_struct* sh)
{
  SH2bf(sh,(13<<4)+12);
}
#define init_8_11_13_13
void FASTCALL call_8_11_13_13(SH2_struct* sh)
{
  SH2bf(sh,(13<<4)+13);
}
#define init_8_11_13_14
void FASTCALL call_8_11_13_14(SH2_struct* sh)
{
  SH2bf(sh,(13<<4)+14);
}
#define init_8_11_13_15
void FASTCALL call_8_11_13_15(SH2_struct* sh)
{
  SH2bf(sh,(13<<4)+15);
}
#define init_8_11_14_0
void FASTCALL call_8_11_14_0(SH2_struct* sh)
{
  SH2bf(sh,(14<<4)+0);
}
#define init_8_11_14_1
void FASTCALL call_8_11_14_1(SH2_struct* sh)
{
  SH2bf(sh,(14<<4)+1);
}
#define init_8_11_14_2
void FASTCALL call_8_11_14_2(SH2_struct* sh)
{
  SH2bf(sh,(14<<4)+2);
}
#define init_8_11_14_3
void FASTCALL call_8_11_14_3(SH2_struct* sh)
{
  SH2bf(sh,(14<<4)+3);
}
#define init_8_11_14_4
void FASTCALL call_8_11_14_4(SH2_struct* sh)
{
  SH2bf(sh,(14<<4)+4);
}
#define init_8_11_14_5
void FASTCALL call_8_11_14_5(SH2_struct* sh)
{
  SH2bf(sh,(14<<4)+5);
}
#define init_8_11_14_6
void FASTCALL call_8_11_14_6(SH2_struct* sh)
{
  SH2bf(sh,(14<<4)+6);
}
#define init_8_11_14_7
void FASTCALL call_8_11_14_7(SH2_struct* sh)
{
  SH2bf(sh,(14<<4)+7);
}
#define init_8_11_14_8
void FASTCALL call_8_11_14_8(SH2_struct* sh)
{
  SH2bf(sh,(14<<4)+8);
}
#define init_8_11_14_9
void FASTCALL call_8_11_14_9(SH2_struct* sh)
{
  SH2bf(sh,(14<<4)+9);
}
#define init_8_11_14_10
void FASTCALL call_8_11_14_10(SH2_struct* sh)
{
  SH2bf(sh,(14<<4)+10);
}
#define init_8_11_14_11
void FASTCALL call_8_11_14_11(SH2_struct* sh)
{
  SH2bf(sh,(14<<4)+11);
}
#define init_8_11_14_12
void FASTCALL call_8_11_14_12(SH2_struct* sh)
{
  SH2bf(sh,(14<<4)+12);
}
#define init_8_11_14_13
void FASTCALL call_8_11_14_13(SH2_struct* sh)
{
  SH2bf(sh,(14<<4)+13);
}
#define init_8_11_14_14
void FASTCALL call_8_11_14_14(SH2_struct* sh)
{
  SH2bf(sh,(14<<4)+14);
}
#define init_8_11_14_15
void FASTCALL call_8_11_14_15(SH2_struct* sh)
{
  SH2bf(sh,(14<<4)+15);
}
#define init_8_11_15_0
void FASTCALL call_8_11_15_0(SH2_struct* sh)
{
  SH2bf(sh,(15<<4)+0);
}
#define init_8_11_15_1
void FASTCALL call_8_11_15_1(SH2_struct* sh)
{
  SH2bf(sh,(15<<4)+1);
}
#define init_8_11_15_2
void FASTCALL call_8_11_15_2(SH2_struct* sh)
{
  SH2bf(sh,(15<<4)+2);
}
#define init_8_11_15_3
void FASTCALL call_8_11_15_3(SH2_struct* sh)
{
  SH2bf(sh,(15<<4)+3);
}
#define init_8_11_15_4
void FASTCALL call_8_11_15_4(SH2_struct* sh)
{
  SH2bf(sh,(15<<4)+4);
}
#define init_8_11_15_5
void FASTCALL call_8_11_15_5(SH2_struct* sh)
{
  SH2bf(sh,(15<<4)+5);
}
#define init_8_11_15_6
void FASTCALL call_8_11_15_6(SH2_struct* sh)
{
  SH2bf(sh,(15<<4)+6);
}
#define init_8_11_15_7
void FASTCALL call_8_11_15_7(SH2_struct* sh)
{
  SH2bf(sh,(15<<4)+7);
}
#define init_8_11_15_8
void FASTCALL call_8_11_15_8(SH2_struct* sh)
{
  SH2bf(sh,(15<<4)+8);
}
#define init_8_11_15_9
void FASTCALL call_8_11_15_9(SH2_struct* sh)
{
  SH2bf(sh,(15<<4)+9);
}
#define init_8_11_15_10
void FASTCALL call_8_11_15_10(SH2_struct* sh)
{
  SH2bf(sh,(15<<4)+10);
}
#define init_8_11_15_11
void FASTCALL call_8_11_15_11(SH2_struct* sh)
{
  SH2bf(sh,(15<<4)+11);
}
#define init_8_11_15_12
void FASTCALL call_8_11_15_12(SH2_struct* sh)
{
  SH2bf(sh,(15<<4)+12);
}
#define init_8_11_15_13
void FASTCALL call_8_11_15_13(SH2_struct* sh)
{
  SH2bf(sh,(15<<4)+13);
}
#define init_8_11_15_14
void FASTCALL call_8_11_15_14(SH2_struct* sh)
{
  SH2bf(sh,(15<<4)+14);
}
#define init_8_11_15_15
void FASTCALL call_8_11_15_15(SH2_struct* sh)
{
  SH2bf(sh,(15<<4)+15);
}
#define init_12_13_0_0
void FASTCALL call_12_13_0_0(SH2_struct* sh)
{
  SH2andm(sh,(0<<4)+0);
}
#define init_12_13_0_1
void FASTCALL call_12_13_0_1(SH2_struct* sh)
{
  SH2andm(sh,(0<<4)+1);
}
#define init_12_13_0_2
void FASTCALL call_12_13_0_2(SH2_struct* sh)
{
  SH2andm(sh,(0<<4)+2);
}
#define init_12_13_0_3
void FASTCALL call_12_13_0_3(SH2_struct* sh)
{
  SH2andm(sh,(0<<4)+3);
}
#define init_12_13_0_4
void FASTCALL call_12_13_0_4(SH2_struct* sh)
{
  SH2andm(sh,(0<<4)+4);
}
#define init_12_13_0_5
void FASTCALL call_12_13_0_5(SH2_struct* sh)
{
  SH2andm(sh,(0<<4)+5);
}
#define init_12_13_0_6
void FASTCALL call_12_13_0_6(SH2_struct* sh)
{
  SH2andm(sh,(0<<4)+6);
}
#define init_12_13_0_7
void FASTCALL call_12_13_0_7(SH2_struct* sh)
{
  SH2andm(sh,(0<<4)+7);
}
#define init_12_13_0_8
void FASTCALL call_12_13_0_8(SH2_struct* sh)
{
  SH2andm(sh,(0<<4)+8);
}
#define init_12_13_0_9
void FASTCALL call_12_13_0_9(SH2_struct* sh)
{
  SH2andm(sh,(0<<4)+9);
}
#define init_12_13_0_10
void FASTCALL call_12_13_0_10(SH2_struct* sh)
{
  SH2andm(sh,(0<<4)+10);
}
#define init_12_13_0_11
void FASTCALL call_12_13_0_11(SH2_struct* sh)
{
  SH2andm(sh,(0<<4)+11);
}
#define init_12_13_0_12
void FASTCALL call_12_13_0_12(SH2_struct* sh)
{
  SH2andm(sh,(0<<4)+12);
}
#define init_12_13_0_13
void FASTCALL call_12_13_0_13(SH2_struct* sh)
{
  SH2andm(sh,(0<<4)+13);
}
#define init_12_13_0_14
void FASTCALL call_12_13_0_14(SH2_struct* sh)
{
  SH2andm(sh,(0<<4)+14);
}
#define init_12_13_0_15
void FASTCALL call_12_13_0_15(SH2_struct* sh)
{
  SH2andm(sh,(0<<4)+15);
}
#define init_12_13_1_0
void FASTCALL call_12_13_1_0(SH2_struct* sh)
{
  SH2andm(sh,(1<<4)+0);
}
#define init_12_13_1_1
void FASTCALL call_12_13_1_1(SH2_struct* sh)
{
  SH2andm(sh,(1<<4)+1);
}
#define init_12_13_1_2
void FASTCALL call_12_13_1_2(SH2_struct* sh)
{
  SH2andm(sh,(1<<4)+2);
}
#define init_12_13_1_3
void FASTCALL call_12_13_1_3(SH2_struct* sh)
{
  SH2andm(sh,(1<<4)+3);
}
#define init_12_13_1_4
void FASTCALL call_12_13_1_4(SH2_struct* sh)
{
  SH2andm(sh,(1<<4)+4);
}
#define init_12_13_1_5
void FASTCALL call_12_13_1_5(SH2_struct* sh)
{
  SH2andm(sh,(1<<4)+5);
}
#define init_12_13_1_6
void FASTCALL call_12_13_1_6(SH2_struct* sh)
{
  SH2andm(sh,(1<<4)+6);
}
#define init_12_13_1_7
void FASTCALL call_12_13_1_7(SH2_struct* sh)
{
  SH2andm(sh,(1<<4)+7);
}
#define init_12_13_1_8
void FASTCALL call_12_13_1_8(SH2_struct* sh)
{
  SH2andm(sh,(1<<4)+8);
}
#define init_12_13_1_9
void FASTCALL call_12_13_1_9(SH2_struct* sh)
{
  SH2andm(sh,(1<<4)+9);
}
#define init_12_13_1_10
void FASTCALL call_12_13_1_10(SH2_struct* sh)
{
  SH2andm(sh,(1<<4)+10);
}
#define init_12_13_1_11
void FASTCALL call_12_13_1_11(SH2_struct* sh)
{
  SH2andm(sh,(1<<4)+11);
}
#define init_12_13_1_12
void FASTCALL call_12_13_1_12(SH2_struct* sh)
{
  SH2andm(sh,(1<<4)+12);
}
#define init_12_13_1_13
void FASTCALL call_12_13_1_13(SH2_struct* sh)
{
  SH2andm(sh,(1<<4)+13);
}
#define init_12_13_1_14
void FASTCALL call_12_13_1_14(SH2_struct* sh)
{
  SH2andm(sh,(1<<4)+14);
}
#define init_12_13_1_15
void FASTCALL call_12_13_1_15(SH2_struct* sh)
{
  SH2andm(sh,(1<<4)+15);
}
#define init_12_13_2_0
void FASTCALL call_12_13_2_0(SH2_struct* sh)
{
  SH2andm(sh,(2<<4)+0);
}
#define init_12_13_2_1
void FASTCALL call_12_13_2_1(SH2_struct* sh)
{
  SH2andm(sh,(2<<4)+1);
}
#define init_12_13_2_2
void FASTCALL call_12_13_2_2(SH2_struct* sh)
{
  SH2andm(sh,(2<<4)+2);
}
#define init_12_13_2_3
void FASTCALL call_12_13_2_3(SH2_struct* sh)
{
  SH2andm(sh,(2<<4)+3);
}
#define init_12_13_2_4
void FASTCALL call_12_13_2_4(SH2_struct* sh)
{
  SH2andm(sh,(2<<4)+4);
}
#define init_12_13_2_5
void FASTCALL call_12_13_2_5(SH2_struct* sh)
{
  SH2andm(sh,(2<<4)+5);
}
#define init_12_13_2_6
void FASTCALL call_12_13_2_6(SH2_struct* sh)
{
  SH2andm(sh,(2<<4)+6);
}
#define init_12_13_2_7
void FASTCALL call_12_13_2_7(SH2_struct* sh)
{
  SH2andm(sh,(2<<4)+7);
}
#define init_12_13_2_8
void FASTCALL call_12_13_2_8(SH2_struct* sh)
{
  SH2andm(sh,(2<<4)+8);
}
#define init_12_13_2_9
void FASTCALL call_12_13_2_9(SH2_struct* sh)
{
  SH2andm(sh,(2<<4)+9);
}
#define init_12_13_2_10
void FASTCALL call_12_13_2_10(SH2_struct* sh)
{
  SH2andm(sh,(2<<4)+10);
}
#define init_12_13_2_11
void FASTCALL call_12_13_2_11(SH2_struct* sh)
{
  SH2andm(sh,(2<<4)+11);
}
#define init_12_13_2_12
void FASTCALL call_12_13_2_12(SH2_struct* sh)
{
  SH2andm(sh,(2<<4)+12);
}
#define init_12_13_2_13
void FASTCALL call_12_13_2_13(SH2_struct* sh)
{
  SH2andm(sh,(2<<4)+13);
}
#define init_12_13_2_14
void FASTCALL call_12_13_2_14(SH2_struct* sh)
{
  SH2andm(sh,(2<<4)+14);
}
#define init_12_13_2_15
void FASTCALL call_12_13_2_15(SH2_struct* sh)
{
  SH2andm(sh,(2<<4)+15);
}
#define init_12_13_3_0
void FASTCALL call_12_13_3_0(SH2_struct* sh)
{
  SH2andm(sh,(3<<4)+0);
}
#define init_12_13_3_1
void FASTCALL call_12_13_3_1(SH2_struct* sh)
{
  SH2andm(sh,(3<<4)+1);
}
#define init_12_13_3_2
void FASTCALL call_12_13_3_2(SH2_struct* sh)
{
  SH2andm(sh,(3<<4)+2);
}
#define init_12_13_3_3
void FASTCALL call_12_13_3_3(SH2_struct* sh)
{
  SH2andm(sh,(3<<4)+3);
}
#define init_12_13_3_4
void FASTCALL call_12_13_3_4(SH2_struct* sh)
{
  SH2andm(sh,(3<<4)+4);
}
#define init_12_13_3_5
void FASTCALL call_12_13_3_5(SH2_struct* sh)
{
  SH2andm(sh,(3<<4)+5);
}
#define init_12_13_3_6
void FASTCALL call_12_13_3_6(SH2_struct* sh)
{
  SH2andm(sh,(3<<4)+6);
}
#define init_12_13_3_7
void FASTCALL call_12_13_3_7(SH2_struct* sh)
{
  SH2andm(sh,(3<<4)+7);
}
#define init_12_13_3_8
void FASTCALL call_12_13_3_8(SH2_struct* sh)
{
  SH2andm(sh,(3<<4)+8);
}
#define init_12_13_3_9
void FASTCALL call_12_13_3_9(SH2_struct* sh)
{
  SH2andm(sh,(3<<4)+9);
}
#define init_12_13_3_10
void FASTCALL call_12_13_3_10(SH2_struct* sh)
{
  SH2andm(sh,(3<<4)+10);
}
#define init_12_13_3_11
void FASTCALL call_12_13_3_11(SH2_struct* sh)
{
  SH2andm(sh,(3<<4)+11);
}
#define init_12_13_3_12
void FASTCALL call_12_13_3_12(SH2_struct* sh)
{
  SH2andm(sh,(3<<4)+12);
}
#define init_12_13_3_13
void FASTCALL call_12_13_3_13(SH2_struct* sh)
{
  SH2andm(sh,(3<<4)+13);
}
#define init_12_13_3_14
void FASTCALL call_12_13_3_14(SH2_struct* sh)
{
  SH2andm(sh,(3<<4)+14);
}
#define init_12_13_3_15
void FASTCALL call_12_13_3_15(SH2_struct* sh)
{
  SH2andm(sh,(3<<4)+15);
}
#define init_12_13_4_0
void FASTCALL call_12_13_4_0(SH2_struct* sh)
{
  SH2andm(sh,(4<<4)+0);
}
#define init_12_13_4_1
void FASTCALL call_12_13_4_1(SH2_struct* sh)
{
  SH2andm(sh,(4<<4)+1);
}
#define init_12_13_4_2
void FASTCALL call_12_13_4_2(SH2_struct* sh)
{
  SH2andm(sh,(4<<4)+2);
}
#define init_12_13_4_3
void FASTCALL call_12_13_4_3(SH2_struct* sh)
{
  SH2andm(sh,(4<<4)+3);
}
#define init_12_13_4_4
void FASTCALL call_12_13_4_4(SH2_struct* sh)
{
  SH2andm(sh,(4<<4)+4);
}
#define init_12_13_4_5
void FASTCALL call_12_13_4_5(SH2_struct* sh)
{
  SH2andm(sh,(4<<4)+5);
}
#define init_12_13_4_6
void FASTCALL call_12_13_4_6(SH2_struct* sh)
{
  SH2andm(sh,(4<<4)+6);
}
#define init_12_13_4_7
void FASTCALL call_12_13_4_7(SH2_struct* sh)
{
  SH2andm(sh,(4<<4)+7);
}
#define init_12_13_4_8
void FASTCALL call_12_13_4_8(SH2_struct* sh)
{
  SH2andm(sh,(4<<4)+8);
}
#define init_12_13_4_9
void FASTCALL call_12_13_4_9(SH2_struct* sh)
{
  SH2andm(sh,(4<<4)+9);
}
#define init_12_13_4_10
void FASTCALL call_12_13_4_10(SH2_struct* sh)
{
  SH2andm(sh,(4<<4)+10);
}
#define init_12_13_4_11
void FASTCALL call_12_13_4_11(SH2_struct* sh)
{
  SH2andm(sh,(4<<4)+11);
}
#define init_12_13_4_12
void FASTCALL call_12_13_4_12(SH2_struct* sh)
{
  SH2andm(sh,(4<<4)+12);
}
#define init_12_13_4_13
void FASTCALL call_12_13_4_13(SH2_struct* sh)
{
  SH2andm(sh,(4<<4)+13);
}
#define init_12_13_4_14
void FASTCALL call_12_13_4_14(SH2_struct* sh)
{
  SH2andm(sh,(4<<4)+14);
}
#define init_12_13_4_15
void FASTCALL call_12_13_4_15(SH2_struct* sh)
{
  SH2andm(sh,(4<<4)+15);
}
#define init_12_13_5_0
void FASTCALL call_12_13_5_0(SH2_struct* sh)
{
  SH2andm(sh,(5<<4)+0);
}
#define init_12_13_5_1
void FASTCALL call_12_13_5_1(SH2_struct* sh)
{
  SH2andm(sh,(5<<4)+1);
}
#define init_12_13_5_2
void FASTCALL call_12_13_5_2(SH2_struct* sh)
{
  SH2andm(sh,(5<<4)+2);
}
#define init_12_13_5_3
void FASTCALL call_12_13_5_3(SH2_struct* sh)
{
  SH2andm(sh,(5<<4)+3);
}
#define init_12_13_5_4
void FASTCALL call_12_13_5_4(SH2_struct* sh)
{
  SH2andm(sh,(5<<4)+4);
}
#define init_12_13_5_5
void FASTCALL call_12_13_5_5(SH2_struct* sh)
{
  SH2andm(sh,(5<<4)+5);
}
#define init_12_13_5_6
void FASTCALL call_12_13_5_6(SH2_struct* sh)
{
  SH2andm(sh,(5<<4)+6);
}
#define init_12_13_5_7
void FASTCALL call_12_13_5_7(SH2_struct* sh)
{
  SH2andm(sh,(5<<4)+7);
}
#define init_12_13_5_8
void FASTCALL call_12_13_5_8(SH2_struct* sh)
{
  SH2andm(sh,(5<<4)+8);
}
#define init_12_13_5_9
void FASTCALL call_12_13_5_9(SH2_struct* sh)
{
  SH2andm(sh,(5<<4)+9);
}
#define init_12_13_5_10
void FASTCALL call_12_13_5_10(SH2_struct* sh)
{
  SH2andm(sh,(5<<4)+10);
}
#define init_12_13_5_11
void FASTCALL call_12_13_5_11(SH2_struct* sh)
{
  SH2andm(sh,(5<<4)+11);
}
#define init_12_13_5_12
void FASTCALL call_12_13_5_12(SH2_struct* sh)
{
  SH2andm(sh,(5<<4)+12);
}
#define init_12_13_5_13
void FASTCALL call_12_13_5_13(SH2_struct* sh)
{
  SH2andm(sh,(5<<4)+13);
}
#define init_12_13_5_14
void FASTCALL call_12_13_5_14(SH2_struct* sh)
{
  SH2andm(sh,(5<<4)+14);
}
#define init_12_13_5_15
void FASTCALL call_12_13_5_15(SH2_struct* sh)
{
  SH2andm(sh,(5<<4)+15);
}
#define init_12_13_6_0
void FASTCALL call_12_13_6_0(SH2_struct* sh)
{
  SH2andm(sh,(6<<4)+0);
}
#define init_12_13_6_1
void FASTCALL call_12_13_6_1(SH2_struct* sh)
{
  SH2andm(sh,(6<<4)+1);
}
#define init_12_13_6_2
void FASTCALL call_12_13_6_2(SH2_struct* sh)
{
  SH2andm(sh,(6<<4)+2);
}
#define init_12_13_6_3
void FASTCALL call_12_13_6_3(SH2_struct* sh)
{
  SH2andm(sh,(6<<4)+3);
}
#define init_12_13_6_4
void FASTCALL call_12_13_6_4(SH2_struct* sh)
{
  SH2andm(sh,(6<<4)+4);
}
#define init_12_13_6_5
void FASTCALL call_12_13_6_5(SH2_struct* sh)
{
  SH2andm(sh,(6<<4)+5);
}
#define init_12_13_6_6
void FASTCALL call_12_13_6_6(SH2_struct* sh)
{
  SH2andm(sh,(6<<4)+6);
}
#define init_12_13_6_7
void FASTCALL call_12_13_6_7(SH2_struct* sh)
{
  SH2andm(sh,(6<<4)+7);
}
#define init_12_13_6_8
void FASTCALL call_12_13_6_8(SH2_struct* sh)
{
  SH2andm(sh,(6<<4)+8);
}
#define init_12_13_6_9
void FASTCALL call_12_13_6_9(SH2_struct* sh)
{
  SH2andm(sh,(6<<4)+9);
}
#define init_12_13_6_10
void FASTCALL call_12_13_6_10(SH2_struct* sh)
{
  SH2andm(sh,(6<<4)+10);
}
#define init_12_13_6_11
void FASTCALL call_12_13_6_11(SH2_struct* sh)
{
  SH2andm(sh,(6<<4)+11);
}
#define init_12_13_6_12
void FASTCALL call_12_13_6_12(SH2_struct* sh)
{
  SH2andm(sh,(6<<4)+12);
}
#define init_12_13_6_13
void FASTCALL call_12_13_6_13(SH2_struct* sh)
{
  SH2andm(sh,(6<<4)+13);
}
#define init_12_13_6_14
void FASTCALL call_12_13_6_14(SH2_struct* sh)
{
  SH2andm(sh,(6<<4)+14);
}
#define init_12_13_6_15
void FASTCALL call_12_13_6_15(SH2_struct* sh)
{
  SH2andm(sh,(6<<4)+15);
}
#define init_12_13_7_0
void FASTCALL call_12_13_7_0(SH2_struct* sh)
{
  SH2andm(sh,(7<<4)+0);
}
#define init_12_13_7_1
void FASTCALL call_12_13_7_1(SH2_struct* sh)
{
  SH2andm(sh,(7<<4)+1);
}
#define init_12_13_7_2
void FASTCALL call_12_13_7_2(SH2_struct* sh)
{
  SH2andm(sh,(7<<4)+2);
}
#define init_12_13_7_3
void FASTCALL call_12_13_7_3(SH2_struct* sh)
{
  SH2andm(sh,(7<<4)+3);
}
#define init_12_13_7_4
void FASTCALL call_12_13_7_4(SH2_struct* sh)
{
  SH2andm(sh,(7<<4)+4);
}
#define init_12_13_7_5
void FASTCALL call_12_13_7_5(SH2_struct* sh)
{
  SH2andm(sh,(7<<4)+5);
}
#define init_12_13_7_6
void FASTCALL call_12_13_7_6(SH2_struct* sh)
{
  SH2andm(sh,(7<<4)+6);
}
#define init_12_13_7_7
void FASTCALL call_12_13_7_7(SH2_struct* sh)
{
  SH2andm(sh,(7<<4)+7);
}
#define init_12_13_7_8
void FASTCALL call_12_13_7_8(SH2_struct* sh)
{
  SH2andm(sh,(7<<4)+8);
}
#define init_12_13_7_9
void FASTCALL call_12_13_7_9(SH2_struct* sh)
{
  SH2andm(sh,(7<<4)+9);
}
#define init_12_13_7_10
void FASTCALL call_12_13_7_10(SH2_struct* sh)
{
  SH2andm(sh,(7<<4)+10);
}
#define init_12_13_7_11
void FASTCALL call_12_13_7_11(SH2_struct* sh)
{
  SH2andm(sh,(7<<4)+11);
}
#define init_12_13_7_12
void FASTCALL call_12_13_7_12(SH2_struct* sh)
{
  SH2andm(sh,(7<<4)+12);
}
#define init_12_13_7_13
void FASTCALL call_12_13_7_13(SH2_struct* sh)
{
  SH2andm(sh,(7<<4)+13);
}
#define init_12_13_7_14
void FASTCALL call_12_13_7_14(SH2_struct* sh)
{
  SH2andm(sh,(7<<4)+14);
}
#define init_12_13_7_15
void FASTCALL call_12_13_7_15(SH2_struct* sh)
{
  SH2andm(sh,(7<<4)+15);
}
#define init_12_13_8_0
void FASTCALL call_12_13_8_0(SH2_struct* sh)
{
  SH2andm(sh,(8<<4)+0);
}
#define init_12_13_8_1
void FASTCALL call_12_13_8_1(SH2_struct* sh)
{
  SH2andm(sh,(8<<4)+1);
}
#define init_12_13_8_2
void FASTCALL call_12_13_8_2(SH2_struct* sh)
{
  SH2andm(sh,(8<<4)+2);
}
#define init_12_13_8_3
void FASTCALL call_12_13_8_3(SH2_struct* sh)
{
  SH2andm(sh,(8<<4)+3);
}
#define init_12_13_8_4
void FASTCALL call_12_13_8_4(SH2_struct* sh)
{
  SH2andm(sh,(8<<4)+4);
}
#define init_12_13_8_5
void FASTCALL call_12_13_8_5(SH2_struct* sh)
{
  SH2andm(sh,(8<<4)+5);
}
#define init_12_13_8_6
void FASTCALL call_12_13_8_6(SH2_struct* sh)
{
  SH2andm(sh,(8<<4)+6);
}
#define init_12_13_8_7
void FASTCALL call_12_13_8_7(SH2_struct* sh)
{
  SH2andm(sh,(8<<4)+7);
}
#define init_12_13_8_8
void FASTCALL call_12_13_8_8(SH2_struct* sh)
{
  SH2andm(sh,(8<<4)+8);
}
#define init_12_13_8_9
void FASTCALL call_12_13_8_9(SH2_struct* sh)
{
  SH2andm(sh,(8<<4)+9);
}
#define init_12_13_8_10
void FASTCALL call_12_13_8_10(SH2_struct* sh)
{
  SH2andm(sh,(8<<4)+10);
}
#define init_12_13_8_11
void FASTCALL call_12_13_8_11(SH2_struct* sh)
{
  SH2andm(sh,(8<<4)+11);
}
#define init_12_13_8_12
void FASTCALL call_12_13_8_12(SH2_struct* sh)
{
  SH2andm(sh,(8<<4)+12);
}
#define init_12_13_8_13
void FASTCALL call_12_13_8_13(SH2_struct* sh)
{
  SH2andm(sh,(8<<4)+13);
}
#define init_12_13_8_14
void FASTCALL call_12_13_8_14(SH2_struct* sh)
{
  SH2andm(sh,(8<<4)+14);
}
#define init_12_13_8_15
void FASTCALL call_12_13_8_15(SH2_struct* sh)
{
  SH2andm(sh,(8<<4)+15);
}
#define init_12_13_9_0
void FASTCALL call_12_13_9_0(SH2_struct* sh)
{
  SH2andm(sh,(9<<4)+0);
}
#define init_12_13_9_1
void FASTCALL call_12_13_9_1(SH2_struct* sh)
{
  SH2andm(sh,(9<<4)+1);
}
#define init_12_13_9_2
void FASTCALL call_12_13_9_2(SH2_struct* sh)
{
  SH2andm(sh,(9<<4)+2);
}
#define init_12_13_9_3
void FASTCALL call_12_13_9_3(SH2_struct* sh)
{
  SH2andm(sh,(9<<4)+3);
}
#define init_12_13_9_4
void FASTCALL call_12_13_9_4(SH2_struct* sh)
{
  SH2andm(sh,(9<<4)+4);
}
#define init_12_13_9_5
void FASTCALL call_12_13_9_5(SH2_struct* sh)
{
  SH2andm(sh,(9<<4)+5);
}
#define init_12_13_9_6
void FASTCALL call_12_13_9_6(SH2_struct* sh)
{
  SH2andm(sh,(9<<4)+6);
}
#define init_12_13_9_7
void FASTCALL call_12_13_9_7(SH2_struct* sh)
{
  SH2andm(sh,(9<<4)+7);
}
#define init_12_13_9_8
void FASTCALL call_12_13_9_8(SH2_struct* sh)
{
  SH2andm(sh,(9<<4)+8);
}
#define init_12_13_9_9
void FASTCALL call_12_13_9_9(SH2_struct* sh)
{
  SH2andm(sh,(9<<4)+9);
}
#define init_12_13_9_10
void FASTCALL call_12_13_9_10(SH2_struct* sh)
{
  SH2andm(sh,(9<<4)+10);
}
#define init_12_13_9_11
void FASTCALL call_12_13_9_11(SH2_struct* sh)
{
  SH2andm(sh,(9<<4)+11);
}
#define init_12_13_9_12
void FASTCALL call_12_13_9_12(SH2_struct* sh)
{
  SH2andm(sh,(9<<4)+12);
}
#define init_12_13_9_13
void FASTCALL call_12_13_9_13(SH2_struct* sh)
{
  SH2andm(sh,(9<<4)+13);
}
#define init_12_13_9_14
void FASTCALL call_12_13_9_14(SH2_struct* sh)
{
  SH2andm(sh,(9<<4)+14);
}
#define init_12_13_9_15
void FASTCALL call_12_13_9_15(SH2_struct* sh)
{
  SH2andm(sh,(9<<4)+15);
}
#define init_12_13_10_0
void FASTCALL call_12_13_10_0(SH2_struct* sh)
{
  SH2andm(sh,(10<<4)+0);
}
#define init_12_13_10_1
void FASTCALL call_12_13_10_1(SH2_struct* sh)
{
  SH2andm(sh,(10<<4)+1);
}
#define init_12_13_10_2
void FASTCALL call_12_13_10_2(SH2_struct* sh)
{
  SH2andm(sh,(10<<4)+2);
}
#define init_12_13_10_3
void FASTCALL call_12_13_10_3(SH2_struct* sh)
{
  SH2andm(sh,(10<<4)+3);
}
#define init_12_13_10_4
void FASTCALL call_12_13_10_4(SH2_struct* sh)
{
  SH2andm(sh,(10<<4)+4);
}
#define init_12_13_10_5
void FASTCALL call_12_13_10_5(SH2_struct* sh)
{
  SH2andm(sh,(10<<4)+5);
}
#define init_12_13_10_6
void FASTCALL call_12_13_10_6(SH2_struct* sh)
{
  SH2andm(sh,(10<<4)+6);
}
#define init_12_13_10_7
void FASTCALL call_12_13_10_7(SH2_struct* sh)
{
  SH2andm(sh,(10<<4)+7);
}
#define init_12_13_10_8
void FASTCALL call_12_13_10_8(SH2_struct* sh)
{
  SH2andm(sh,(10<<4)+8);
}
#define init_12_13_10_9
void FASTCALL call_12_13_10_9(SH2_struct* sh)
{
  SH2andm(sh,(10<<4)+9);
}
#define init_12_13_10_10
void FASTCALL call_12_13_10_10(SH2_struct* sh)
{
  SH2andm(sh,(10<<4)+10);
}
#define init_12_13_10_11
void FASTCALL call_12_13_10_11(SH2_struct* sh)
{
  SH2andm(sh,(10<<4)+11);
}
#define init_12_13_10_12
void FASTCALL call_12_13_10_12(SH2_struct* sh)
{
  SH2andm(sh,(10<<4)+12);
}
#define init_12_13_10_13
void FASTCALL call_12_13_10_13(SH2_struct* sh)
{
  SH2andm(sh,(10<<4)+13);
}
#define init_12_13_10_14
void FASTCALL call_12_13_10_14(SH2_struct* sh)
{
  SH2andm(sh,(10<<4)+14);
}
#define init_12_13_10_15
void FASTCALL call_12_13_10_15(SH2_struct* sh)
{
  SH2andm(sh,(10<<4)+15);
}
#define init_12_13_11_0
void FASTCALL call_12_13_11_0(SH2_struct* sh)
{
  SH2andm(sh,(11<<4)+0);
}
#define init_12_13_11_1
void FASTCALL call_12_13_11_1(SH2_struct* sh)
{
  SH2andm(sh,(11<<4)+1);
}
#define init_12_13_11_2
void FASTCALL call_12_13_11_2(SH2_struct* sh)
{
  SH2andm(sh,(11<<4)+2);
}
#define init_12_13_11_3
void FASTCALL call_12_13_11_3(SH2_struct* sh)
{
  SH2andm(sh,(11<<4)+3);
}
#define init_12_13_11_4
void FASTCALL call_12_13_11_4(SH2_struct* sh)
{
  SH2andm(sh,(11<<4)+4);
}
#define init_12_13_11_5
void FASTCALL call_12_13_11_5(SH2_struct* sh)
{
  SH2andm(sh,(11<<4)+5);
}
#define init_12_13_11_6
void FASTCALL call_12_13_11_6(SH2_struct* sh)
{
  SH2andm(sh,(11<<4)+6);
}
#define init_12_13_11_7
void FASTCALL call_12_13_11_7(SH2_struct* sh)
{
  SH2andm(sh,(11<<4)+7);
}
#define init_12_13_11_8
void FASTCALL call_12_13_11_8(SH2_struct* sh)
{
  SH2andm(sh,(11<<4)+8);
}
#define init_12_13_11_9
void FASTCALL call_12_13_11_9(SH2_struct* sh)
{
  SH2andm(sh,(11<<4)+9);
}
#define init_12_13_11_10
void FASTCALL call_12_13_11_10(SH2_struct* sh)
{
  SH2andm(sh,(11<<4)+10);
}
#define init_12_13_11_11
void FASTCALL call_12_13_11_11(SH2_struct* sh)
{
  SH2andm(sh,(11<<4)+11);
}
#define init_12_13_11_12
void FASTCALL call_12_13_11_12(SH2_struct* sh)
{
  SH2andm(sh,(11<<4)+12);
}
#define init_12_13_11_13
void FASTCALL call_12_13_11_13(SH2_struct* sh)
{
  SH2andm(sh,(11<<4)+13);
}
#define init_12_13_11_14
void FASTCALL call_12_13_11_14(SH2_struct* sh)
{
  SH2andm(sh,(11<<4)+14);
}
#define init_12_13_11_15
void FASTCALL call_12_13_11_15(SH2_struct* sh)
{
  SH2andm(sh,(11<<4)+15);
}
#define init_12_13_12_0
void FASTCALL call_12_13_12_0(SH2_struct* sh)
{
  SH2andm(sh,(12<<4)+0);
}
#define init_12_13_12_1
void FASTCALL call_12_13_12_1(SH2_struct* sh)
{
  SH2andm(sh,(12<<4)+1);
}
#define init_12_13_12_2
void FASTCALL call_12_13_12_2(SH2_struct* sh)
{
  SH2andm(sh,(12<<4)+2);
}
#define init_12_13_12_3
void FASTCALL call_12_13_12_3(SH2_struct* sh)
{
  SH2andm(sh,(12<<4)+3);
}
#define init_12_13_12_4
void FASTCALL call_12_13_12_4(SH2_struct* sh)
{
  SH2andm(sh,(12<<4)+4);
}
#define init_12_13_12_5
void FASTCALL call_12_13_12_5(SH2_struct* sh)
{
  SH2andm(sh,(12<<4)+5);
}
#define init_12_13_12_6
void FASTCALL call_12_13_12_6(SH2_struct* sh)
{
  SH2andm(sh,(12<<4)+6);
}
#define init_12_13_12_7
void FASTCALL call_12_13_12_7(SH2_struct* sh)
{
  SH2andm(sh,(12<<4)+7);
}
#define init_12_13_12_8
void FASTCALL call_12_13_12_8(SH2_struct* sh)
{
  SH2andm(sh,(12<<4)+8);
}
#define init_12_13_12_9
void FASTCALL call_12_13_12_9(SH2_struct* sh)
{
  SH2andm(sh,(12<<4)+9);
}
#define init_12_13_12_10
void FASTCALL call_12_13_12_10(SH2_struct* sh)
{
  SH2andm(sh,(12<<4)+10);
}
#define init_12_13_12_11
void FASTCALL call_12_13_12_11(SH2_struct* sh)
{
  SH2andm(sh,(12<<4)+11);
}
#define init_12_13_12_12
void FASTCALL call_12_13_12_12(SH2_struct* sh)
{
  SH2andm(sh,(12<<4)+12);
}
#define init_12_13_12_13
void FASTCALL call_12_13_12_13(SH2_struct* sh)
{
  SH2andm(sh,(12<<4)+13);
}
#define init_12_13_12_14
void FASTCALL call_12_13_12_14(SH2_struct* sh)
{
  SH2andm(sh,(12<<4)+14);
}
#define init_12_13_12_15
void FASTCALL call_12_13_12_15(SH2_struct* sh)
{
  SH2andm(sh,(12<<4)+15);
}
#define init_12_13_13_0
void FASTCALL call_12_13_13_0(SH2_struct* sh)
{
  SH2andm(sh,(13<<4)+0);
}
#define init_12_13_13_1
void FASTCALL call_12_13_13_1(SH2_struct* sh)
{
  SH2andm(sh,(13<<4)+1);
}
#define init_12_13_13_2
void FASTCALL call_12_13_13_2(SH2_struct* sh)
{
  SH2andm(sh,(13<<4)+2);
}
#define init_12_13_13_3
void FASTCALL call_12_13_13_3(SH2_struct* sh)
{
  SH2andm(sh,(13<<4)+3);
}
#define init_12_13_13_4
void FASTCALL call_12_13_13_4(SH2_struct* sh)
{
  SH2andm(sh,(13<<4)+4);
}
#define init_12_13_13_5
void FASTCALL call_12_13_13_5(SH2_struct* sh)
{
  SH2andm(sh,(13<<4)+5);
}
#define init_12_13_13_6
void FASTCALL call_12_13_13_6(SH2_struct* sh)
{
  SH2andm(sh,(13<<4)+6);
}
#define init_12_13_13_7
void FASTCALL call_12_13_13_7(SH2_struct* sh)
{
  SH2andm(sh,(13<<4)+7);
}
#define init_12_13_13_8
void FASTCALL call_12_13_13_8(SH2_struct* sh)
{
  SH2andm(sh,(13<<4)+8);
}
#define init_12_13_13_9
void FASTCALL call_12_13_13_9(SH2_struct* sh)
{
  SH2andm(sh,(13<<4)+9);
}
#define init_12_13_13_10
void FASTCALL call_12_13_13_10(SH2_struct* sh)
{
  SH2andm(sh,(13<<4)+10);
}
#define init_12_13_13_11
void FASTCALL call_12_13_13_11(SH2_struct* sh)
{
  SH2andm(sh,(13<<4)+11);
}
#define init_12_13_13_12
void FASTCALL call_12_13_13_12(SH2_struct* sh)
{
  SH2andm(sh,(13<<4)+12);
}
#define init_12_13_13_13
void FASTCALL call_12_13_13_13(SH2_struct* sh)
{
  SH2andm(sh,(13<<4)+13);
}
#define init_12_13_13_14
void FASTCALL call_12_13_13_14(SH2_struct* sh)
{
  SH2andm(sh,(13<<4)+14);
}
#define init_12_13_13_15
void FASTCALL call_12_13_13_15(SH2_struct* sh)
{
  SH2andm(sh,(13<<4)+15);
}
#define init_12_13_14_0
void FASTCALL call_12_13_14_0(SH2_struct* sh)
{
  SH2andm(sh,(14<<4)+0);
}
#define init_12_13_14_1
void FASTCALL call_12_13_14_1(SH2_struct* sh)
{
  SH2andm(sh,(14<<4)+1);
}
#define init_12_13_14_2
void FASTCALL call_12_13_14_2(SH2_struct* sh)
{
  SH2andm(sh,(14<<4)+2);
}
#define init_12_13_14_3
void FASTCALL call_12_13_14_3(SH2_struct* sh)
{
  SH2andm(sh,(14<<4)+3);
}
#define init_12_13_14_4
void FASTCALL call_12_13_14_4(SH2_struct* sh)
{
  SH2andm(sh,(14<<4)+4);
}
#define init_12_13_14_5
void FASTCALL call_12_13_14_5(SH2_struct* sh)
{
  SH2andm(sh,(14<<4)+5);
}
#define init_12_13_14_6
void FASTCALL call_12_13_14_6(SH2_struct* sh)
{
  SH2andm(sh,(14<<4)+6);
}
#define init_12_13_14_7
void FASTCALL call_12_13_14_7(SH2_struct* sh)
{
  SH2andm(sh,(14<<4)+7);
}
#define init_12_13_14_8
void FASTCALL call_12_13_14_8(SH2_struct* sh)
{
  SH2andm(sh,(14<<4)+8);
}
#define init_12_13_14_9
void FASTCALL call_12_13_14_9(SH2_struct* sh)
{
  SH2andm(sh,(14<<4)+9);
}
#define init_12_13_14_10
void FASTCALL call_12_13_14_10(SH2_struct* sh)
{
  SH2andm(sh,(14<<4)+10);
}
#define init_12_13_14_11
void FASTCALL call_12_13_14_11(SH2_struct* sh)
{
  SH2andm(sh,(14<<4)+11);
}
#define init_12_13_14_12
void FASTCALL call_12_13_14_12(SH2_struct* sh)
{
  SH2andm(sh,(14<<4)+12);
}
#define init_12_13_14_13
void FASTCALL call_12_13_14_13(SH2_struct* sh)
{
  SH2andm(sh,(14<<4)+13);
}
#define init_12_13_14_14
void FASTCALL call_12_13_14_14(SH2_struct* sh)
{
  SH2andm(sh,(14<<4)+14);
}
#define init_12_13_14_15
void FASTCALL call_12_13_14_15(SH2_struct* sh)
{
  SH2andm(sh,(14<<4)+15);
}
#define init_12_13_15_0
void FASTCALL call_12_13_15_0(SH2_struct* sh)
{
  SH2andm(sh,(15<<4)+0);
}
#define init_12_13_15_1
void FASTCALL call_12_13_15_1(SH2_struct* sh)
{
  SH2andm(sh,(15<<4)+1);
}
#define init_12_13_15_2
void FASTCALL call_12_13_15_2(SH2_struct* sh)
{
  SH2andm(sh,(15<<4)+2);
}
#define init_12_13_15_3
void FASTCALL call_12_13_15_3(SH2_struct* sh)
{
  SH2andm(sh,(15<<4)+3);
}
#define init_12_13_15_4
void FASTCALL call_12_13_15_4(SH2_struct* sh)
{
  SH2andm(sh,(15<<4)+4);
}
#define init_12_13_15_5
void FASTCALL call_12_13_15_5(SH2_struct* sh)
{
  SH2andm(sh,(15<<4)+5);
}
#define init_12_13_15_6
void FASTCALL call_12_13_15_6(SH2_struct* sh)
{
  SH2andm(sh,(15<<4)+6);
}
#define init_12_13_15_7
void FASTCALL call_12_13_15_7(SH2_struct* sh)
{
  SH2andm(sh,(15<<4)+7);
}
#define init_12_13_15_8
void FASTCALL call_12_13_15_8(SH2_struct* sh)
{
  SH2andm(sh,(15<<4)+8);
}
#define init_12_13_15_9
void FASTCALL call_12_13_15_9(SH2_struct* sh)
{
  SH2andm(sh,(15<<4)+9);
}
#define init_12_13_15_10
void FASTCALL call_12_13_15_10(SH2_struct* sh)
{
  SH2andm(sh,(15<<4)+10);
}
#define init_12_13_15_11
void FASTCALL call_12_13_15_11(SH2_struct* sh)
{
  SH2andm(sh,(15<<4)+11);
}
#define init_12_13_15_12
void FASTCALL call_12_13_15_12(SH2_struct* sh)
{
  SH2andm(sh,(15<<4)+12);
}
#define init_12_13_15_13
void FASTCALL call_12_13_15_13(SH2_struct* sh)
{
  SH2andm(sh,(15<<4)+13);
}
#define init_12_13_15_14
void FASTCALL call_12_13_15_14(SH2_struct* sh)
{
  SH2andm(sh,(15<<4)+14);
}
#define init_12_13_15_15
void FASTCALL call_12_13_15_15(SH2_struct* sh)
{
  SH2andm(sh,(15<<4)+15);
}
#define init_12_9_0_0
void FASTCALL call_12_9_0_0(SH2_struct* sh)
{
  SH2andi(sh,(0<<4)+0);
}
#define init_12_9_0_1
void FASTCALL call_12_9_0_1(SH2_struct* sh)
{
  SH2andi(sh,(0<<4)+1);
}
#define init_12_9_0_2
void FASTCALL call_12_9_0_2(SH2_struct* sh)
{
  SH2andi(sh,(0<<4)+2);
}
#define init_12_9_0_3
void FASTCALL call_12_9_0_3(SH2_struct* sh)
{
  SH2andi(sh,(0<<4)+3);
}
#define init_12_9_0_4
void FASTCALL call_12_9_0_4(SH2_struct* sh)
{
  SH2andi(sh,(0<<4)+4);
}
#define init_12_9_0_5
void FASTCALL call_12_9_0_5(SH2_struct* sh)
{
  SH2andi(sh,(0<<4)+5);
}
#define init_12_9_0_6
void FASTCALL call_12_9_0_6(SH2_struct* sh)
{
  SH2andi(sh,(0<<4)+6);
}
#define init_12_9_0_7
void FASTCALL call_12_9_0_7(SH2_struct* sh)
{
  SH2andi(sh,(0<<4)+7);
}
#define init_12_9_0_8
void FASTCALL call_12_9_0_8(SH2_struct* sh)
{
  SH2andi(sh,(0<<4)+8);
}
#define init_12_9_0_9
void FASTCALL call_12_9_0_9(SH2_struct* sh)
{
  SH2andi(sh,(0<<4)+9);
}
#define init_12_9_0_10
void FASTCALL call_12_9_0_10(SH2_struct* sh)
{
  SH2andi(sh,(0<<4)+10);
}
#define init_12_9_0_11
void FASTCALL call_12_9_0_11(SH2_struct* sh)
{
  SH2andi(sh,(0<<4)+11);
}
#define init_12_9_0_12
void FASTCALL call_12_9_0_12(SH2_struct* sh)
{
  SH2andi(sh,(0<<4)+12);
}
#define init_12_9_0_13
void FASTCALL call_12_9_0_13(SH2_struct* sh)
{
  SH2andi(sh,(0<<4)+13);
}
#define init_12_9_0_14
void FASTCALL call_12_9_0_14(SH2_struct* sh)
{
  SH2andi(sh,(0<<4)+14);
}
#define init_12_9_0_15
void FASTCALL call_12_9_0_15(SH2_struct* sh)
{
  SH2andi(sh,(0<<4)+15);
}
#define init_12_9_1_0
void FASTCALL call_12_9_1_0(SH2_struct* sh)
{
  SH2andi(sh,(1<<4)+0);
}
#define init_12_9_1_1
void FASTCALL call_12_9_1_1(SH2_struct* sh)
{
  SH2andi(sh,(1<<4)+1);
}
#define init_12_9_1_2
void FASTCALL call_12_9_1_2(SH2_struct* sh)
{
  SH2andi(sh,(1<<4)+2);
}
#define init_12_9_1_3
void FASTCALL call_12_9_1_3(SH2_struct* sh)
{
  SH2andi(sh,(1<<4)+3);
}
#define init_12_9_1_4
void FASTCALL call_12_9_1_4(SH2_struct* sh)
{
  SH2andi(sh,(1<<4)+4);
}
#define init_12_9_1_5
void FASTCALL call_12_9_1_5(SH2_struct* sh)
{
  SH2andi(sh,(1<<4)+5);
}
#define init_12_9_1_6
void FASTCALL call_12_9_1_6(SH2_struct* sh)
{
  SH2andi(sh,(1<<4)+6);
}
#define init_12_9_1_7
void FASTCALL call_12_9_1_7(SH2_struct* sh)
{
  SH2andi(sh,(1<<4)+7);
}
#define init_12_9_1_8
void FASTCALL call_12_9_1_8(SH2_struct* sh)
{
  SH2andi(sh,(1<<4)+8);
}
#define init_12_9_1_9
void FASTCALL call_12_9_1_9(SH2_struct* sh)
{
  SH2andi(sh,(1<<4)+9);
}
#define init_12_9_1_10
void FASTCALL call_12_9_1_10(SH2_struct* sh)
{
  SH2andi(sh,(1<<4)+10);
}
#define init_12_9_1_11
void FASTCALL call_12_9_1_11(SH2_struct* sh)
{
  SH2andi(sh,(1<<4)+11);
}
#define init_12_9_1_12
void FASTCALL call_12_9_1_12(SH2_struct* sh)
{
  SH2andi(sh,(1<<4)+12);
}
#define init_12_9_1_13
void FASTCALL call_12_9_1_13(SH2_struct* sh)
{
  SH2andi(sh,(1<<4)+13);
}
#define init_12_9_1_14
void FASTCALL call_12_9_1_14(SH2_struct* sh)
{
  SH2andi(sh,(1<<4)+14);
}
#define init_12_9_1_15
void FASTCALL call_12_9_1_15(SH2_struct* sh)
{
  SH2andi(sh,(1<<4)+15);
}
#define init_12_9_2_0
void FASTCALL call_12_9_2_0(SH2_struct* sh)
{
  SH2andi(sh,(2<<4)+0);
}
#define init_12_9_2_1
void FASTCALL call_12_9_2_1(SH2_struct* sh)
{
  SH2andi(sh,(2<<4)+1);
}
#define init_12_9_2_2
void FASTCALL call_12_9_2_2(SH2_struct* sh)
{
  SH2andi(sh,(2<<4)+2);
}
#define init_12_9_2_3
void FASTCALL call_12_9_2_3(SH2_struct* sh)
{
  SH2andi(sh,(2<<4)+3);
}
#define init_12_9_2_4
void FASTCALL call_12_9_2_4(SH2_struct* sh)
{
  SH2andi(sh,(2<<4)+4);
}
#define init_12_9_2_5
void FASTCALL call_12_9_2_5(SH2_struct* sh)
{
  SH2andi(sh,(2<<4)+5);
}
#define init_12_9_2_6
void FASTCALL call_12_9_2_6(SH2_struct* sh)
{
  SH2andi(sh,(2<<4)+6);
}
#define init_12_9_2_7
void FASTCALL call_12_9_2_7(SH2_struct* sh)
{
  SH2andi(sh,(2<<4)+7);
}
#define init_12_9_2_8
void FASTCALL call_12_9_2_8(SH2_struct* sh)
{
  SH2andi(sh,(2<<4)+8);
}
#define init_12_9_2_9
void FASTCALL call_12_9_2_9(SH2_struct* sh)
{
  SH2andi(sh,(2<<4)+9);
}
#define init_12_9_2_10
void FASTCALL call_12_9_2_10(SH2_struct* sh)
{
  SH2andi(sh,(2<<4)+10);
}
#define init_12_9_2_11
void FASTCALL call_12_9_2_11(SH2_struct* sh)
{
  SH2andi(sh,(2<<4)+11);
}
#define init_12_9_2_12
void FASTCALL call_12_9_2_12(SH2_struct* sh)
{
  SH2andi(sh,(2<<4)+12);
}
#define init_12_9_2_13
void FASTCALL call_12_9_2_13(SH2_struct* sh)
{
  SH2andi(sh,(2<<4)+13);
}
#define init_12_9_2_14
void FASTCALL call_12_9_2_14(SH2_struct* sh)
{
  SH2andi(sh,(2<<4)+14);
}
#define init_12_9_2_15
void FASTCALL call_12_9_2_15(SH2_struct* sh)
{
  SH2andi(sh,(2<<4)+15);
}
#define init_12_9_3_0
void FASTCALL call_12_9_3_0(SH2_struct* sh)
{
  SH2andi(sh,(3<<4)+0);
}
#define init_12_9_3_1
void FASTCALL call_12_9_3_1(SH2_struct* sh)
{
  SH2andi(sh,(3<<4)+1);
}
#define init_12_9_3_2
void FASTCALL call_12_9_3_2(SH2_struct* sh)
{
  SH2andi(sh,(3<<4)+2);
}
#define init_12_9_3_3
void FASTCALL call_12_9_3_3(SH2_struct* sh)
{
  SH2andi(sh,(3<<4)+3);
}
#define init_12_9_3_4
void FASTCALL call_12_9_3_4(SH2_struct* sh)
{
  SH2andi(sh,(3<<4)+4);
}
#define init_12_9_3_5
void FASTCALL call_12_9_3_5(SH2_struct* sh)
{
  SH2andi(sh,(3<<4)+5);
}
#define init_12_9_3_6
void FASTCALL call_12_9_3_6(SH2_struct* sh)
{
  SH2andi(sh,(3<<4)+6);
}
#define init_12_9_3_7
void FASTCALL call_12_9_3_7(SH2_struct* sh)
{
  SH2andi(sh,(3<<4)+7);
}
#define init_12_9_3_8
void FASTCALL call_12_9_3_8(SH2_struct* sh)
{
  SH2andi(sh,(3<<4)+8);
}
#define init_12_9_3_9
void FASTCALL call_12_9_3_9(SH2_struct* sh)
{
  SH2andi(sh,(3<<4)+9);
}
#define init_12_9_3_10
void FASTCALL call_12_9_3_10(SH2_struct* sh)
{
  SH2andi(sh,(3<<4)+10);
}
#define init_12_9_3_11
void FASTCALL call_12_9_3_11(SH2_struct* sh)
{
  SH2andi(sh,(3<<4)+11);
}
#define init_12_9_3_12
void FASTCALL call_12_9_3_12(SH2_struct* sh)
{
  SH2andi(sh,(3<<4)+12);
}
#define init_12_9_3_13
void FASTCALL call_12_9_3_13(SH2_struct* sh)
{
  SH2andi(sh,(3<<4)+13);
}
#define init_12_9_3_14
void FASTCALL call_12_9_3_14(SH2_struct* sh)
{
  SH2andi(sh,(3<<4)+14);
}
#define init_12_9_3_15
void FASTCALL call_12_9_3_15(SH2_struct* sh)
{
  SH2andi(sh,(3<<4)+15);
}
#define init_12_9_4_0
void FASTCALL call_12_9_4_0(SH2_struct* sh)
{
  SH2andi(sh,(4<<4)+0);
}
#define init_12_9_4_1
void FASTCALL call_12_9_4_1(SH2_struct* sh)
{
  SH2andi(sh,(4<<4)+1);
}
#define init_12_9_4_2
void FASTCALL call_12_9_4_2(SH2_struct* sh)
{
  SH2andi(sh,(4<<4)+2);
}
#define init_12_9_4_3
void FASTCALL call_12_9_4_3(SH2_struct* sh)
{
  SH2andi(sh,(4<<4)+3);
}
#define init_12_9_4_4
void FASTCALL call_12_9_4_4(SH2_struct* sh)
{
  SH2andi(sh,(4<<4)+4);
}
#define init_12_9_4_5
void FASTCALL call_12_9_4_5(SH2_struct* sh)
{
  SH2andi(sh,(4<<4)+5);
}
#define init_12_9_4_6
void FASTCALL call_12_9_4_6(SH2_struct* sh)
{
  SH2andi(sh,(4<<4)+6);
}
#define init_12_9_4_7
void FASTCALL call_12_9_4_7(SH2_struct* sh)
{
  SH2andi(sh,(4<<4)+7);
}
#define init_12_9_4_8
void FASTCALL call_12_9_4_8(SH2_struct* sh)
{
  SH2andi(sh,(4<<4)+8);
}
#define init_12_9_4_9
void FASTCALL call_12_9_4_9(SH2_struct* sh)
{
  SH2andi(sh,(4<<4)+9);
}
#define init_12_9_4_10
void FASTCALL call_12_9_4_10(SH2_struct* sh)
{
  SH2andi(sh,(4<<4)+10);
}
#define init_12_9_4_11
void FASTCALL call_12_9_4_11(SH2_struct* sh)
{
  SH2andi(sh,(4<<4)+11);
}
#define init_12_9_4_12
void FASTCALL call_12_9_4_12(SH2_struct* sh)
{
  SH2andi(sh,(4<<4)+12);
}
#define init_12_9_4_13
void FASTCALL call_12_9_4_13(SH2_struct* sh)
{
  SH2andi(sh,(4<<4)+13);
}
#define init_12_9_4_14
void FASTCALL call_12_9_4_14(SH2_struct* sh)
{
  SH2andi(sh,(4<<4)+14);
}
#define init_12_9_4_15
void FASTCALL call_12_9_4_15(SH2_struct* sh)
{
  SH2andi(sh,(4<<4)+15);
}
#define init_12_9_5_0
void FASTCALL call_12_9_5_0(SH2_struct* sh)
{
  SH2andi(sh,(5<<4)+0);
}
#define init_12_9_5_1
void FASTCALL call_12_9_5_1(SH2_struct* sh)
{
  SH2andi(sh,(5<<4)+1);
}
#define init_12_9_5_2
void FASTCALL call_12_9_5_2(SH2_struct* sh)
{
  SH2andi(sh,(5<<4)+2);
}
#define init_12_9_5_3
void FASTCALL call_12_9_5_3(SH2_struct* sh)
{
  SH2andi(sh,(5<<4)+3);
}
#define init_12_9_5_4
void FASTCALL call_12_9_5_4(SH2_struct* sh)
{
  SH2andi(sh,(5<<4)+4);
}
#define init_12_9_5_5
void FASTCALL call_12_9_5_5(SH2_struct* sh)
{
  SH2andi(sh,(5<<4)+5);
}
#define init_12_9_5_6
void FASTCALL call_12_9_5_6(SH2_struct* sh)
{
  SH2andi(sh,(5<<4)+6);
}
#define init_12_9_5_7
void FASTCALL call_12_9_5_7(SH2_struct* sh)
{
  SH2andi(sh,(5<<4)+7);
}
#define init_12_9_5_8
void FASTCALL call_12_9_5_8(SH2_struct* sh)
{
  SH2andi(sh,(5<<4)+8);
}
#define init_12_9_5_9
void FASTCALL call_12_9_5_9(SH2_struct* sh)
{
  SH2andi(sh,(5<<4)+9);
}
#define init_12_9_5_10
void FASTCALL call_12_9_5_10(SH2_struct* sh)
{
  SH2andi(sh,(5<<4)+10);
}
#define init_12_9_5_11
void FASTCALL call_12_9_5_11(SH2_struct* sh)
{
  SH2andi(sh,(5<<4)+11);
}
#define init_12_9_5_12
void FASTCALL call_12_9_5_12(SH2_struct* sh)
{
  SH2andi(sh,(5<<4)+12);
}
#define init_12_9_5_13
void FASTCALL call_12_9_5_13(SH2_struct* sh)
{
  SH2andi(sh,(5<<4)+13);
}
#define init_12_9_5_14
void FASTCALL call_12_9_5_14(SH2_struct* sh)
{
  SH2andi(sh,(5<<4)+14);
}
#define init_12_9_5_15
void FASTCALL call_12_9_5_15(SH2_struct* sh)
{
  SH2andi(sh,(5<<4)+15);
}
#define init_12_9_6_0
void FASTCALL call_12_9_6_0(SH2_struct* sh)
{
  SH2andi(sh,(6<<4)+0);
}
#define init_12_9_6_1
void FASTCALL call_12_9_6_1(SH2_struct* sh)
{
  SH2andi(sh,(6<<4)+1);
}
#define init_12_9_6_2
void FASTCALL call_12_9_6_2(SH2_struct* sh)
{
  SH2andi(sh,(6<<4)+2);
}
#define init_12_9_6_3
void FASTCALL call_12_9_6_3(SH2_struct* sh)
{
  SH2andi(sh,(6<<4)+3);
}
#define init_12_9_6_4
void FASTCALL call_12_9_6_4(SH2_struct* sh)
{
  SH2andi(sh,(6<<4)+4);
}
#define init_12_9_6_5
void FASTCALL call_12_9_6_5(SH2_struct* sh)
{
  SH2andi(sh,(6<<4)+5);
}
#define init_12_9_6_6
void FASTCALL call_12_9_6_6(SH2_struct* sh)
{
  SH2andi(sh,(6<<4)+6);
}
#define init_12_9_6_7
void FASTCALL call_12_9_6_7(SH2_struct* sh)
{
  SH2andi(sh,(6<<4)+7);
}
#define init_12_9_6_8
void FASTCALL call_12_9_6_8(SH2_struct* sh)
{
  SH2andi(sh,(6<<4)+8);
}
#define init_12_9_6_9
void FASTCALL call_12_9_6_9(SH2_struct* sh)
{
  SH2andi(sh,(6<<4)+9);
}
#define init_12_9_6_10
void FASTCALL call_12_9_6_10(SH2_struct* sh)
{
  SH2andi(sh,(6<<4)+10);
}
#define init_12_9_6_11
void FASTCALL call_12_9_6_11(SH2_struct* sh)
{
  SH2andi(sh,(6<<4)+11);
}
#define init_12_9_6_12
void FASTCALL call_12_9_6_12(SH2_struct* sh)
{
  SH2andi(sh,(6<<4)+12);
}
#define init_12_9_6_13
void FASTCALL call_12_9_6_13(SH2_struct* sh)
{
  SH2andi(sh,(6<<4)+13);
}
#define init_12_9_6_14
void FASTCALL call_12_9_6_14(SH2_struct* sh)
{
  SH2andi(sh,(6<<4)+14);
}
#define init_12_9_6_15
void FASTCALL call_12_9_6_15(SH2_struct* sh)
{
  SH2andi(sh,(6<<4)+15);
}
#define init_12_9_7_0
void FASTCALL call_12_9_7_0(SH2_struct* sh)
{
  SH2andi(sh,(7<<4)+0);
}
#define init_12_9_7_1
void FASTCALL call_12_9_7_1(SH2_struct* sh)
{
  SH2andi(sh,(7<<4)+1);
}
#define init_12_9_7_2
void FASTCALL call_12_9_7_2(SH2_struct* sh)
{
  SH2andi(sh,(7<<4)+2);
}
#define init_12_9_7_3
void FASTCALL call_12_9_7_3(SH2_struct* sh)
{
  SH2andi(sh,(7<<4)+3);
}
#define init_12_9_7_4
void FASTCALL call_12_9_7_4(SH2_struct* sh)
{
  SH2andi(sh,(7<<4)+4);
}
#define init_12_9_7_5
void FASTCALL call_12_9_7_5(SH2_struct* sh)
{
  SH2andi(sh,(7<<4)+5);
}
#define init_12_9_7_6
void FASTCALL call_12_9_7_6(SH2_struct* sh)
{
  SH2andi(sh,(7<<4)+6);
}
#define init_12_9_7_7
void FASTCALL call_12_9_7_7(SH2_struct* sh)
{
  SH2andi(sh,(7<<4)+7);
}
#define init_12_9_7_8
void FASTCALL call_12_9_7_8(SH2_struct* sh)
{
  SH2andi(sh,(7<<4)+8);
}
#define init_12_9_7_9
void FASTCALL call_12_9_7_9(SH2_struct* sh)
{
  SH2andi(sh,(7<<4)+9);
}
#define init_12_9_7_10
void FASTCALL call_12_9_7_10(SH2_struct* sh)
{
  SH2andi(sh,(7<<4)+10);
}
#define init_12_9_7_11
void FASTCALL call_12_9_7_11(SH2_struct* sh)
{
  SH2andi(sh,(7<<4)+11);
}
#define init_12_9_7_12
void FASTCALL call_12_9_7_12(SH2_struct* sh)
{
  SH2andi(sh,(7<<4)+12);
}
#define init_12_9_7_13
void FASTCALL call_12_9_7_13(SH2_struct* sh)
{
  SH2andi(sh,(7<<4)+13);
}
#define init_12_9_7_14
void FASTCALL call_12_9_7_14(SH2_struct* sh)
{
  SH2andi(sh,(7<<4)+14);
}
#define init_12_9_7_15
void FASTCALL call_12_9_7_15(SH2_struct* sh)
{
  SH2andi(sh,(7<<4)+15);
}
#define init_12_9_8_0
void FASTCALL call_12_9_8_0(SH2_struct* sh)
{
  SH2andi(sh,(8<<4)+0);
}
#define init_12_9_8_1
void FASTCALL call_12_9_8_1(SH2_struct* sh)
{
  SH2andi(sh,(8<<4)+1);
}
#define init_12_9_8_2
void FASTCALL call_12_9_8_2(SH2_struct* sh)
{
  SH2andi(sh,(8<<4)+2);
}
#define init_12_9_8_3
void FASTCALL call_12_9_8_3(SH2_struct* sh)
{
  SH2andi(sh,(8<<4)+3);
}
#define init_12_9_8_4
void FASTCALL call_12_9_8_4(SH2_struct* sh)
{
  SH2andi(sh,(8<<4)+4);
}
#define init_12_9_8_5
void FASTCALL call_12_9_8_5(SH2_struct* sh)
{
  SH2andi(sh,(8<<4)+5);
}
#define init_12_9_8_6
void FASTCALL call_12_9_8_6(SH2_struct* sh)
{
  SH2andi(sh,(8<<4)+6);
}
#define init_12_9_8_7
void FASTCALL call_12_9_8_7(SH2_struct* sh)
{
  SH2andi(sh,(8<<4)+7);
}
#define init_12_9_8_8
void FASTCALL call_12_9_8_8(SH2_struct* sh)
{
  SH2andi(sh,(8<<4)+8);
}
#define init_12_9_8_9
void FASTCALL call_12_9_8_9(SH2_struct* sh)
{
  SH2andi(sh,(8<<4)+9);
}
#define init_12_9_8_10
void FASTCALL call_12_9_8_10(SH2_struct* sh)
{
  SH2andi(sh,(8<<4)+10);
}
#define init_12_9_8_11
void FASTCALL call_12_9_8_11(SH2_struct* sh)
{
  SH2andi(sh,(8<<4)+11);
}
#define init_12_9_8_12
void FASTCALL call_12_9_8_12(SH2_struct* sh)
{
  SH2andi(sh,(8<<4)+12);
}
#define init_12_9_8_13
void FASTCALL call_12_9_8_13(SH2_struct* sh)
{
  SH2andi(sh,(8<<4)+13);
}
#define init_12_9_8_14
void FASTCALL call_12_9_8_14(SH2_struct* sh)
{
  SH2andi(sh,(8<<4)+14);
}
#define init_12_9_8_15
void FASTCALL call_12_9_8_15(SH2_struct* sh)
{
  SH2andi(sh,(8<<4)+15);
}
#define init_12_9_9_0
void FASTCALL call_12_9_9_0(SH2_struct* sh)
{
  SH2andi(sh,(9<<4)+0);
}
#define init_12_9_9_1
void FASTCALL call_12_9_9_1(SH2_struct* sh)
{
  SH2andi(sh,(9<<4)+1);
}
#define init_12_9_9_2
void FASTCALL call_12_9_9_2(SH2_struct* sh)
{
  SH2andi(sh,(9<<4)+2);
}
#define init_12_9_9_3
void FASTCALL call_12_9_9_3(SH2_struct* sh)
{
  SH2andi(sh,(9<<4)+3);
}
#define init_12_9_9_4
void FASTCALL call_12_9_9_4(SH2_struct* sh)
{
  SH2andi(sh,(9<<4)+4);
}
#define init_12_9_9_5
void FASTCALL call_12_9_9_5(SH2_struct* sh)
{
  SH2andi(sh,(9<<4)+5);
}
#define init_12_9_9_6
void FASTCALL call_12_9_9_6(SH2_struct* sh)
{
  SH2andi(sh,(9<<4)+6);
}
#define init_12_9_9_7
void FASTCALL call_12_9_9_7(SH2_struct* sh)
{
  SH2andi(sh,(9<<4)+7);
}
#define init_12_9_9_8
void FASTCALL call_12_9_9_8(SH2_struct* sh)
{
  SH2andi(sh,(9<<4)+8);
}
#define init_12_9_9_9
void FASTCALL call_12_9_9_9(SH2_struct* sh)
{
  SH2andi(sh,(9<<4)+9);
}
#define init_12_9_9_10
void FASTCALL call_12_9_9_10(SH2_struct* sh)
{
  SH2andi(sh,(9<<4)+10);
}
#define init_12_9_9_11
void FASTCALL call_12_9_9_11(SH2_struct* sh)
{
  SH2andi(sh,(9<<4)+11);
}
#define init_12_9_9_12
void FASTCALL call_12_9_9_12(SH2_struct* sh)
{
  SH2andi(sh,(9<<4)+12);
}
#define init_12_9_9_13
void FASTCALL call_12_9_9_13(SH2_struct* sh)
{
  SH2andi(sh,(9<<4)+13);
}
#define init_12_9_9_14
void FASTCALL call_12_9_9_14(SH2_struct* sh)
{
  SH2andi(sh,(9<<4)+14);
}
#define init_12_9_9_15
void FASTCALL call_12_9_9_15(SH2_struct* sh)
{
  SH2andi(sh,(9<<4)+15);
}
#define init_12_9_10_0
void FASTCALL call_12_9_10_0(SH2_struct* sh)
{
  SH2andi(sh,(10<<4)+0);
}
#define init_12_9_10_1
void FASTCALL call_12_9_10_1(SH2_struct* sh)
{
  SH2andi(sh,(10<<4)+1);
}
#define init_12_9_10_2
void FASTCALL call_12_9_10_2(SH2_struct* sh)
{
  SH2andi(sh,(10<<4)+2);
}
#define init_12_9_10_3
void FASTCALL call_12_9_10_3(SH2_struct* sh)
{
  SH2andi(sh,(10<<4)+3);
}
#define init_12_9_10_4
void FASTCALL call_12_9_10_4(SH2_struct* sh)
{
  SH2andi(sh,(10<<4)+4);
}
#define init_12_9_10_5
void FASTCALL call_12_9_10_5(SH2_struct* sh)
{
  SH2andi(sh,(10<<4)+5);
}
#define init_12_9_10_6
void FASTCALL call_12_9_10_6(SH2_struct* sh)
{
  SH2andi(sh,(10<<4)+6);
}
#define init_12_9_10_7
void FASTCALL call_12_9_10_7(SH2_struct* sh)
{
  SH2andi(sh,(10<<4)+7);
}
#define init_12_9_10_8
void FASTCALL call_12_9_10_8(SH2_struct* sh)
{
  SH2andi(sh,(10<<4)+8);
}
#define init_12_9_10_9
void FASTCALL call_12_9_10_9(SH2_struct* sh)
{
  SH2andi(sh,(10<<4)+9);
}
#define init_12_9_10_10
void FASTCALL call_12_9_10_10(SH2_struct* sh)
{
  SH2andi(sh,(10<<4)+10);
}
#define init_12_9_10_11
void FASTCALL call_12_9_10_11(SH2_struct* sh)
{
  SH2andi(sh,(10<<4)+11);
}
#define init_12_9_10_12
void FASTCALL call_12_9_10_12(SH2_struct* sh)
{
  SH2andi(sh,(10<<4)+12);
}
#define init_12_9_10_13
void FASTCALL call_12_9_10_13(SH2_struct* sh)
{
  SH2andi(sh,(10<<4)+13);
}
#define init_12_9_10_14
void FASTCALL call_12_9_10_14(SH2_struct* sh)
{
  SH2andi(sh,(10<<4)+14);
}
#define init_12_9_10_15
void FASTCALL call_12_9_10_15(SH2_struct* sh)
{
  SH2andi(sh,(10<<4)+15);
}
#define init_12_9_11_0
void FASTCALL call_12_9_11_0(SH2_struct* sh)
{
  SH2andi(sh,(11<<4)+0);
}
#define init_12_9_11_1
void FASTCALL call_12_9_11_1(SH2_struct* sh)
{
  SH2andi(sh,(11<<4)+1);
}
#define init_12_9_11_2
void FASTCALL call_12_9_11_2(SH2_struct* sh)
{
  SH2andi(sh,(11<<4)+2);
}
#define init_12_9_11_3
void FASTCALL call_12_9_11_3(SH2_struct* sh)
{
  SH2andi(sh,(11<<4)+3);
}
#define init_12_9_11_4
void FASTCALL call_12_9_11_4(SH2_struct* sh)
{
  SH2andi(sh,(11<<4)+4);
}
#define init_12_9_11_5
void FASTCALL call_12_9_11_5(SH2_struct* sh)
{
  SH2andi(sh,(11<<4)+5);
}
#define init_12_9_11_6
void FASTCALL call_12_9_11_6(SH2_struct* sh)
{
  SH2andi(sh,(11<<4)+6);
}
#define init_12_9_11_7
void FASTCALL call_12_9_11_7(SH2_struct* sh)
{
  SH2andi(sh,(11<<4)+7);
}
#define init_12_9_11_8
void FASTCALL call_12_9_11_8(SH2_struct* sh)
{
  SH2andi(sh,(11<<4)+8);
}
#define init_12_9_11_9
void FASTCALL call_12_9_11_9(SH2_struct* sh)
{
  SH2andi(sh,(11<<4)+9);
}
#define init_12_9_11_10
void FASTCALL call_12_9_11_10(SH2_struct* sh)
{
  SH2andi(sh,(11<<4)+10);
}
#define init_12_9_11_11
void FASTCALL call_12_9_11_11(SH2_struct* sh)
{
  SH2andi(sh,(11<<4)+11);
}
#define init_12_9_11_12
void FASTCALL call_12_9_11_12(SH2_struct* sh)
{
  SH2andi(sh,(11<<4)+12);
}
#define init_12_9_11_13
void FASTCALL call_12_9_11_13(SH2_struct* sh)
{
  SH2andi(sh,(11<<4)+13);
}
#define init_12_9_11_14
void FASTCALL call_12_9_11_14(SH2_struct* sh)
{
  SH2andi(sh,(11<<4)+14);
}
#define init_12_9_11_15
void FASTCALL call_12_9_11_15(SH2_struct* sh)
{
  SH2andi(sh,(11<<4)+15);
}
#define init_12_9_12_0
void FASTCALL call_12_9_12_0(SH2_struct* sh)
{
  SH2andi(sh,(12<<4)+0);
}
#define init_12_9_12_1
void FASTCALL call_12_9_12_1(SH2_struct* sh)
{
  SH2andi(sh,(12<<4)+1);
}
#define init_12_9_12_2
void FASTCALL call_12_9_12_2(SH2_struct* sh)
{
  SH2andi(sh,(12<<4)+2);
}
#define init_12_9_12_3
void FASTCALL call_12_9_12_3(SH2_struct* sh)
{
  SH2andi(sh,(12<<4)+3);
}
#define init_12_9_12_4
void FASTCALL call_12_9_12_4(SH2_struct* sh)
{
  SH2andi(sh,(12<<4)+4);
}
#define init_12_9_12_5
void FASTCALL call_12_9_12_5(SH2_struct* sh)
{
  SH2andi(sh,(12<<4)+5);
}
#define init_12_9_12_6
void FASTCALL call_12_9_12_6(SH2_struct* sh)
{
  SH2andi(sh,(12<<4)+6);
}
#define init_12_9_12_7
void FASTCALL call_12_9_12_7(SH2_struct* sh)
{
  SH2andi(sh,(12<<4)+7);
}
#define init_12_9_12_8
void FASTCALL call_12_9_12_8(SH2_struct* sh)
{
  SH2andi(sh,(12<<4)+8);
}
#define init_12_9_12_9
void FASTCALL call_12_9_12_9(SH2_struct* sh)
{
  SH2andi(sh,(12<<4)+9);
}
#define init_12_9_12_10
void FASTCALL call_12_9_12_10(SH2_struct* sh)
{
  SH2andi(sh,(12<<4)+10);
}
#define init_12_9_12_11
void FASTCALL call_12_9_12_11(SH2_struct* sh)
{
  SH2andi(sh,(12<<4)+11);
}
#define init_12_9_12_12
void FASTCALL call_12_9_12_12(SH2_struct* sh)
{
  SH2andi(sh,(12<<4)+12);
}
#define init_12_9_12_13
void FASTCALL call_12_9_12_13(SH2_struct* sh)
{
  SH2andi(sh,(12<<4)+13);
}
#define init_12_9_12_14
void FASTCALL call_12_9_12_14(SH2_struct* sh)
{
  SH2andi(sh,(12<<4)+14);
}
#define init_12_9_12_15
void FASTCALL call_12_9_12_15(SH2_struct* sh)
{
  SH2andi(sh,(12<<4)+15);
}
#define init_12_9_13_0
void FASTCALL call_12_9_13_0(SH2_struct* sh)
{
  SH2andi(sh,(13<<4)+0);
}
#define init_12_9_13_1
void FASTCALL call_12_9_13_1(SH2_struct* sh)
{
  SH2andi(sh,(13<<4)+1);
}
#define init_12_9_13_2
void FASTCALL call_12_9_13_2(SH2_struct* sh)
{
  SH2andi(sh,(13<<4)+2);
}
#define init_12_9_13_3
void FASTCALL call_12_9_13_3(SH2_struct* sh)
{
  SH2andi(sh,(13<<4)+3);
}
#define init_12_9_13_4
void FASTCALL call_12_9_13_4(SH2_struct* sh)
{
  SH2andi(sh,(13<<4)+4);
}
#define init_12_9_13_5
void FASTCALL call_12_9_13_5(SH2_struct* sh)
{
  SH2andi(sh,(13<<4)+5);
}
#define init_12_9_13_6
void FASTCALL call_12_9_13_6(SH2_struct* sh)
{
  SH2andi(sh,(13<<4)+6);
}
#define init_12_9_13_7
void FASTCALL call_12_9_13_7(SH2_struct* sh)
{
  SH2andi(sh,(13<<4)+7);
}
#define init_12_9_13_8
void FASTCALL call_12_9_13_8(SH2_struct* sh)
{
  SH2andi(sh,(13<<4)+8);
}
#define init_12_9_13_9
void FASTCALL call_12_9_13_9(SH2_struct* sh)
{
  SH2andi(sh,(13<<4)+9);
}
#define init_12_9_13_10
void FASTCALL call_12_9_13_10(SH2_struct* sh)
{
  SH2andi(sh,(13<<4)+10);
}
#define init_12_9_13_11
void FASTCALL call_12_9_13_11(SH2_struct* sh)
{
  SH2andi(sh,(13<<4)+11);
}
#define init_12_9_13_12
void FASTCALL call_12_9_13_12(SH2_struct* sh)
{
  SH2andi(sh,(13<<4)+12);
}
#define init_12_9_13_13
void FASTCALL call_12_9_13_13(SH2_struct* sh)
{
  SH2andi(sh,(13<<4)+13);
}
#define init_12_9_13_14
void FASTCALL call_12_9_13_14(SH2_struct* sh)
{
  SH2andi(sh,(13<<4)+14);
}
#define init_12_9_13_15
void FASTCALL call_12_9_13_15(SH2_struct* sh)
{
  SH2andi(sh,(13<<4)+15);
}
#define init_12_9_14_0
void FASTCALL call_12_9_14_0(SH2_struct* sh)
{
  SH2andi(sh,(14<<4)+0);
}
#define init_12_9_14_1
void FASTCALL call_12_9_14_1(SH2_struct* sh)
{
  SH2andi(sh,(14<<4)+1);
}
#define init_12_9_14_2
void FASTCALL call_12_9_14_2(SH2_struct* sh)
{
  SH2andi(sh,(14<<4)+2);
}
#define init_12_9_14_3
void FASTCALL call_12_9_14_3(SH2_struct* sh)
{
  SH2andi(sh,(14<<4)+3);
}
#define init_12_9_14_4
void FASTCALL call_12_9_14_4(SH2_struct* sh)
{
  SH2andi(sh,(14<<4)+4);
}
#define init_12_9_14_5
void FASTCALL call_12_9_14_5(SH2_struct* sh)
{
  SH2andi(sh,(14<<4)+5);
}
#define init_12_9_14_6
void FASTCALL call_12_9_14_6(SH2_struct* sh)
{
  SH2andi(sh,(14<<4)+6);
}
#define init_12_9_14_7
void FASTCALL call_12_9_14_7(SH2_struct* sh)
{
  SH2andi(sh,(14<<4)+7);
}
#define init_12_9_14_8
void FASTCALL call_12_9_14_8(SH2_struct* sh)
{
  SH2andi(sh,(14<<4)+8);
}
#define init_12_9_14_9
void FASTCALL call_12_9_14_9(SH2_struct* sh)
{
  SH2andi(sh,(14<<4)+9);
}
#define init_12_9_14_10
void FASTCALL call_12_9_14_10(SH2_struct* sh)
{
  SH2andi(sh,(14<<4)+10);
}
#define init_12_9_14_11
void FASTCALL call_12_9_14_11(SH2_struct* sh)
{
  SH2andi(sh,(14<<4)+11);
}
#define init_12_9_14_12
void FASTCALL call_12_9_14_12(SH2_struct* sh)
{
  SH2andi(sh,(14<<4)+12);
}
#define init_12_9_14_13
void FASTCALL call_12_9_14_13(SH2_struct* sh)
{
  SH2andi(sh,(14<<4)+13);
}
#define init_12_9_14_14
void FASTCALL call_12_9_14_14(SH2_struct* sh)
{
  SH2andi(sh,(14<<4)+14);
}
#define init_12_9_14_15
void FASTCALL call_12_9_14_15(SH2_struct* sh)
{
  SH2andi(sh,(14<<4)+15);
}
#define init_12_9_15_0
void FASTCALL call_12_9_15_0(SH2_struct* sh)
{
  SH2andi(sh,(15<<4)+0);
}
#define init_12_9_15_1
void FASTCALL call_12_9_15_1(SH2_struct* sh)
{
  SH2andi(sh,(15<<4)+1);
}
#define init_12_9_15_2
void FASTCALL call_12_9_15_2(SH2_struct* sh)
{
  SH2andi(sh,(15<<4)+2);
}
#define init_12_9_15_3
void FASTCALL call_12_9_15_3(SH2_struct* sh)
{
  SH2andi(sh,(15<<4)+3);
}
#define init_12_9_15_4
void FASTCALL call_12_9_15_4(SH2_struct* sh)
{
  SH2andi(sh,(15<<4)+4);
}
#define init_12_9_15_5
void FASTCALL call_12_9_15_5(SH2_struct* sh)
{
  SH2andi(sh,(15<<4)+5);
}
#define init_12_9_15_6
void FASTCALL call_12_9_15_6(SH2_struct* sh)
{
  SH2andi(sh,(15<<4)+6);
}
#define init_12_9_15_7
void FASTCALL call_12_9_15_7(SH2_struct* sh)
{
  SH2andi(sh,(15<<4)+7);
}
#define init_12_9_15_8
void FASTCALL call_12_9_15_8(SH2_struct* sh)
{
  SH2andi(sh,(15<<4)+8);
}
#define init_12_9_15_9
void FASTCALL call_12_9_15_9(SH2_struct* sh)
{
  SH2andi(sh,(15<<4)+9);
}
#define init_12_9_15_10
void FASTCALL call_12_9_15_10(SH2_struct* sh)
{
  SH2andi(sh,(15<<4)+10);
}
#define init_12_9_15_11
void FASTCALL call_12_9_15_11(SH2_struct* sh)
{
  SH2andi(sh,(15<<4)+11);
}
#define init_12_9_15_12
void FASTCALL call_12_9_15_12(SH2_struct* sh)
{
  SH2andi(sh,(15<<4)+12);
}
#define init_12_9_15_13
void FASTCALL call_12_9_15_13(SH2_struct* sh)
{
  SH2andi(sh,(15<<4)+13);
}
#define init_12_9_15_14
void FASTCALL call_12_9_15_14(SH2_struct* sh)
{
  SH2andi(sh,(15<<4)+14);
}
#define init_12_9_15_15
void FASTCALL call_12_9_15_15(SH2_struct* sh)
{
  SH2andi(sh,(15<<4)+15);
}
#define init_2_0_0_9
void FASTCALL call_2_0_0_9(SH2_struct* sh)
{
  SH2and(sh,0,0);
}
#define init_2_0_1_9
void FASTCALL call_2_0_1_9(SH2_struct* sh)
{
  SH2and(sh,0,1);
}
#define init_2_0_2_9
void FASTCALL call_2_0_2_9(SH2_struct* sh)
{
  SH2and(sh,0,2);
}
#define init_2_0_3_9
void FASTCALL call_2_0_3_9(SH2_struct* sh)
{
  SH2and(sh,0,3);
}
#define init_2_0_4_9
void FASTCALL call_2_0_4_9(SH2_struct* sh)
{
  SH2and(sh,0,4);
}
#define init_2_0_5_9
void FASTCALL call_2_0_5_9(SH2_struct* sh)
{
  SH2and(sh,0,5);
}
#define init_2_0_6_9
void FASTCALL call_2_0_6_9(SH2_struct* sh)
{
  SH2and(sh,0,6);
}
#define init_2_0_7_9
void FASTCALL call_2_0_7_9(SH2_struct* sh)
{
  SH2and(sh,0,7);
}
#define init_2_0_8_9
void FASTCALL call_2_0_8_9(SH2_struct* sh)
{
  SH2and(sh,0,8);
}
#define init_2_0_9_9
void FASTCALL call_2_0_9_9(SH2_struct* sh)
{
  SH2and(sh,0,9);
}
#define init_2_0_10_9
void FASTCALL call_2_0_10_9(SH2_struct* sh)
{
  SH2and(sh,0,10);
}
#define init_2_0_11_9
void FASTCALL call_2_0_11_9(SH2_struct* sh)
{
  SH2and(sh,0,11);
}
#define init_2_0_12_9
void FASTCALL call_2_0_12_9(SH2_struct* sh)
{
  SH2and(sh,0,12);
}
#define init_2_0_13_9
void FASTCALL call_2_0_13_9(SH2_struct* sh)
{
  SH2and(sh,0,13);
}
#define init_2_0_14_9
void FASTCALL call_2_0_14_9(SH2_struct* sh)
{
  SH2and(sh,0,14);
}
#define init_2_0_15_9
void FASTCALL call_2_0_15_9(SH2_struct* sh)
{
  SH2and(sh,0,15);
}
#define init_2_1_0_9
void FASTCALL call_2_1_0_9(SH2_struct* sh)
{
  SH2and(sh,1,0);
}
#define init_2_1_1_9
void FASTCALL call_2_1_1_9(SH2_struct* sh)
{
  SH2and(sh,1,1);
}
#define init_2_1_2_9
void FASTCALL call_2_1_2_9(SH2_struct* sh)
{
  SH2and(sh,1,2);
}
#define init_2_1_3_9
void FASTCALL call_2_1_3_9(SH2_struct* sh)
{
  SH2and(sh,1,3);
}
#define init_2_1_4_9
void FASTCALL call_2_1_4_9(SH2_struct* sh)
{
  SH2and(sh,1,4);
}
#define init_2_1_5_9
void FASTCALL call_2_1_5_9(SH2_struct* sh)
{
  SH2and(sh,1,5);
}
#define init_2_1_6_9
void FASTCALL call_2_1_6_9(SH2_struct* sh)
{
  SH2and(sh,1,6);
}
#define init_2_1_7_9
void FASTCALL call_2_1_7_9(SH2_struct* sh)
{
  SH2and(sh,1,7);
}
#define init_2_1_8_9
void FASTCALL call_2_1_8_9(SH2_struct* sh)
{
  SH2and(sh,1,8);
}
#define init_2_1_9_9
void FASTCALL call_2_1_9_9(SH2_struct* sh)
{
  SH2and(sh,1,9);
}
#define init_2_1_10_9
void FASTCALL call_2_1_10_9(SH2_struct* sh)
{
  SH2and(sh,1,10);
}
#define init_2_1_11_9
void FASTCALL call_2_1_11_9(SH2_struct* sh)
{
  SH2and(sh,1,11);
}
#define init_2_1_12_9
void FASTCALL call_2_1_12_9(SH2_struct* sh)
{
  SH2and(sh,1,12);
}
#define init_2_1_13_9
void FASTCALL call_2_1_13_9(SH2_struct* sh)
{
  SH2and(sh,1,13);
}
#define init_2_1_14_9
void FASTCALL call_2_1_14_9(SH2_struct* sh)
{
  SH2and(sh,1,14);
}
#define init_2_1_15_9
void FASTCALL call_2_1_15_9(SH2_struct* sh)
{
  SH2and(sh,1,15);
}
#define init_2_2_0_9
void FASTCALL call_2_2_0_9(SH2_struct* sh)
{
  SH2and(sh,2,0);
}
#define init_2_2_1_9
void FASTCALL call_2_2_1_9(SH2_struct* sh)
{
  SH2and(sh,2,1);
}
#define init_2_2_2_9
void FASTCALL call_2_2_2_9(SH2_struct* sh)
{
  SH2and(sh,2,2);
}
#define init_2_2_3_9
void FASTCALL call_2_2_3_9(SH2_struct* sh)
{
  SH2and(sh,2,3);
}
#define init_2_2_4_9
void FASTCALL call_2_2_4_9(SH2_struct* sh)
{
  SH2and(sh,2,4);
}
#define init_2_2_5_9
void FASTCALL call_2_2_5_9(SH2_struct* sh)
{
  SH2and(sh,2,5);
}
#define init_2_2_6_9
void FASTCALL call_2_2_6_9(SH2_struct* sh)
{
  SH2and(sh,2,6);
}
#define init_2_2_7_9
void FASTCALL call_2_2_7_9(SH2_struct* sh)
{
  SH2and(sh,2,7);
}
#define init_2_2_8_9
void FASTCALL call_2_2_8_9(SH2_struct* sh)
{
  SH2and(sh,2,8);
}
#define init_2_2_9_9
void FASTCALL call_2_2_9_9(SH2_struct* sh)
{
  SH2and(sh,2,9);
}
#define init_2_2_10_9
void FASTCALL call_2_2_10_9(SH2_struct* sh)
{
  SH2and(sh,2,10);
}
#define init_2_2_11_9
void FASTCALL call_2_2_11_9(SH2_struct* sh)
{
  SH2and(sh,2,11);
}
#define init_2_2_12_9
void FASTCALL call_2_2_12_9(SH2_struct* sh)
{
  SH2and(sh,2,12);
}
#define init_2_2_13_9
void FASTCALL call_2_2_13_9(SH2_struct* sh)
{
  SH2and(sh,2,13);
}
#define init_2_2_14_9
void FASTCALL call_2_2_14_9(SH2_struct* sh)
{
  SH2and(sh,2,14);
}
#define init_2_2_15_9
void FASTCALL call_2_2_15_9(SH2_struct* sh)
{
  SH2and(sh,2,15);
}
#define init_2_3_0_9
void FASTCALL call_2_3_0_9(SH2_struct* sh)
{
  SH2and(sh,3,0);
}
#define init_2_3_1_9
void FASTCALL call_2_3_1_9(SH2_struct* sh)
{
  SH2and(sh,3,1);
}
#define init_2_3_2_9
void FASTCALL call_2_3_2_9(SH2_struct* sh)
{
  SH2and(sh,3,2);
}
#define init_2_3_3_9
void FASTCALL call_2_3_3_9(SH2_struct* sh)
{
  SH2and(sh,3,3);
}
#define init_2_3_4_9
void FASTCALL call_2_3_4_9(SH2_struct* sh)
{
  SH2and(sh,3,4);
}
#define init_2_3_5_9
void FASTCALL call_2_3_5_9(SH2_struct* sh)
{
  SH2and(sh,3,5);
}
#define init_2_3_6_9
void FASTCALL call_2_3_6_9(SH2_struct* sh)
{
  SH2and(sh,3,6);
}
#define init_2_3_7_9
void FASTCALL call_2_3_7_9(SH2_struct* sh)
{
  SH2and(sh,3,7);
}
#define init_2_3_8_9
void FASTCALL call_2_3_8_9(SH2_struct* sh)
{
  SH2and(sh,3,8);
}
#define init_2_3_9_9
void FASTCALL call_2_3_9_9(SH2_struct* sh)
{
  SH2and(sh,3,9);
}
#define init_2_3_10_9
void FASTCALL call_2_3_10_9(SH2_struct* sh)
{
  SH2and(sh,3,10);
}
#define init_2_3_11_9
void FASTCALL call_2_3_11_9(SH2_struct* sh)
{
  SH2and(sh,3,11);
}
#define init_2_3_12_9
void FASTCALL call_2_3_12_9(SH2_struct* sh)
{
  SH2and(sh,3,12);
}
#define init_2_3_13_9
void FASTCALL call_2_3_13_9(SH2_struct* sh)
{
  SH2and(sh,3,13);
}
#define init_2_3_14_9
void FASTCALL call_2_3_14_9(SH2_struct* sh)
{
  SH2and(sh,3,14);
}
#define init_2_3_15_9
void FASTCALL call_2_3_15_9(SH2_struct* sh)
{
  SH2and(sh,3,15);
}
#define init_2_4_0_9
void FASTCALL call_2_4_0_9(SH2_struct* sh)
{
  SH2and(sh,4,0);
}
#define init_2_4_1_9
void FASTCALL call_2_4_1_9(SH2_struct* sh)
{
  SH2and(sh,4,1);
}
#define init_2_4_2_9
void FASTCALL call_2_4_2_9(SH2_struct* sh)
{
  SH2and(sh,4,2);
}
#define init_2_4_3_9
void FASTCALL call_2_4_3_9(SH2_struct* sh)
{
  SH2and(sh,4,3);
}
#define init_2_4_4_9
void FASTCALL call_2_4_4_9(SH2_struct* sh)
{
  SH2and(sh,4,4);
}
#define init_2_4_5_9
void FASTCALL call_2_4_5_9(SH2_struct* sh)
{
  SH2and(sh,4,5);
}
#define init_2_4_6_9
void FASTCALL call_2_4_6_9(SH2_struct* sh)
{
  SH2and(sh,4,6);
}
#define init_2_4_7_9
void FASTCALL call_2_4_7_9(SH2_struct* sh)
{
  SH2and(sh,4,7);
}
#define init_2_4_8_9
void FASTCALL call_2_4_8_9(SH2_struct* sh)
{
  SH2and(sh,4,8);
}
#define init_2_4_9_9
void FASTCALL call_2_4_9_9(SH2_struct* sh)
{
  SH2and(sh,4,9);
}
#define init_2_4_10_9
void FASTCALL call_2_4_10_9(SH2_struct* sh)
{
  SH2and(sh,4,10);
}
#define init_2_4_11_9
void FASTCALL call_2_4_11_9(SH2_struct* sh)
{
  SH2and(sh,4,11);
}
#define init_2_4_12_9
void FASTCALL call_2_4_12_9(SH2_struct* sh)
{
  SH2and(sh,4,12);
}
#define init_2_4_13_9
void FASTCALL call_2_4_13_9(SH2_struct* sh)
{
  SH2and(sh,4,13);
}
#define init_2_4_14_9
void FASTCALL call_2_4_14_9(SH2_struct* sh)
{
  SH2and(sh,4,14);
}
#define init_2_4_15_9
void FASTCALL call_2_4_15_9(SH2_struct* sh)
{
  SH2and(sh,4,15);
}
#define init_2_5_0_9
void FASTCALL call_2_5_0_9(SH2_struct* sh)
{
  SH2and(sh,5,0);
}
#define init_2_5_1_9
void FASTCALL call_2_5_1_9(SH2_struct* sh)
{
  SH2and(sh,5,1);
}
#define init_2_5_2_9
void FASTCALL call_2_5_2_9(SH2_struct* sh)
{
  SH2and(sh,5,2);
}
#define init_2_5_3_9
void FASTCALL call_2_5_3_9(SH2_struct* sh)
{
  SH2and(sh,5,3);
}
#define init_2_5_4_9
void FASTCALL call_2_5_4_9(SH2_struct* sh)
{
  SH2and(sh,5,4);
}
#define init_2_5_5_9
void FASTCALL call_2_5_5_9(SH2_struct* sh)
{
  SH2and(sh,5,5);
}
#define init_2_5_6_9
void FASTCALL call_2_5_6_9(SH2_struct* sh)
{
  SH2and(sh,5,6);
}
#define init_2_5_7_9
void FASTCALL call_2_5_7_9(SH2_struct* sh)
{
  SH2and(sh,5,7);
}
#define init_2_5_8_9
void FASTCALL call_2_5_8_9(SH2_struct* sh)
{
  SH2and(sh,5,8);
}
#define init_2_5_9_9
void FASTCALL call_2_5_9_9(SH2_struct* sh)
{
  SH2and(sh,5,9);
}
#define init_2_5_10_9
void FASTCALL call_2_5_10_9(SH2_struct* sh)
{
  SH2and(sh,5,10);
}
#define init_2_5_11_9
void FASTCALL call_2_5_11_9(SH2_struct* sh)
{
  SH2and(sh,5,11);
}
#define init_2_5_12_9
void FASTCALL call_2_5_12_9(SH2_struct* sh)
{
  SH2and(sh,5,12);
}
#define init_2_5_13_9
void FASTCALL call_2_5_13_9(SH2_struct* sh)
{
  SH2and(sh,5,13);
}
#define init_2_5_14_9
void FASTCALL call_2_5_14_9(SH2_struct* sh)
{
  SH2and(sh,5,14);
}
#define init_2_5_15_9
void FASTCALL call_2_5_15_9(SH2_struct* sh)
{
  SH2and(sh,5,15);
}
#define init_2_6_0_9
void FASTCALL call_2_6_0_9(SH2_struct* sh)
{
  SH2and(sh,6,0);
}
#define init_2_6_1_9
void FASTCALL call_2_6_1_9(SH2_struct* sh)
{
  SH2and(sh,6,1);
}
#define init_2_6_2_9
void FASTCALL call_2_6_2_9(SH2_struct* sh)
{
  SH2and(sh,6,2);
}
#define init_2_6_3_9
void FASTCALL call_2_6_3_9(SH2_struct* sh)
{
  SH2and(sh,6,3);
}
#define init_2_6_4_9
void FASTCALL call_2_6_4_9(SH2_struct* sh)
{
  SH2and(sh,6,4);
}
#define init_2_6_5_9
void FASTCALL call_2_6_5_9(SH2_struct* sh)
{
  SH2and(sh,6,5);
}
#define init_2_6_6_9
void FASTCALL call_2_6_6_9(SH2_struct* sh)
{
  SH2and(sh,6,6);
}
#define init_2_6_7_9
void FASTCALL call_2_6_7_9(SH2_struct* sh)
{
  SH2and(sh,6,7);
}
#define init_2_6_8_9
void FASTCALL call_2_6_8_9(SH2_struct* sh)
{
  SH2and(sh,6,8);
}
#define init_2_6_9_9
void FASTCALL call_2_6_9_9(SH2_struct* sh)
{
  SH2and(sh,6,9);
}
#define init_2_6_10_9
void FASTCALL call_2_6_10_9(SH2_struct* sh)
{
  SH2and(sh,6,10);
}
#define init_2_6_11_9
void FASTCALL call_2_6_11_9(SH2_struct* sh)
{
  SH2and(sh,6,11);
}
#define init_2_6_12_9
void FASTCALL call_2_6_12_9(SH2_struct* sh)
{
  SH2and(sh,6,12);
}
#define init_2_6_13_9
void FASTCALL call_2_6_13_9(SH2_struct* sh)
{
  SH2and(sh,6,13);
}
#define init_2_6_14_9
void FASTCALL call_2_6_14_9(SH2_struct* sh)
{
  SH2and(sh,6,14);
}
#define init_2_6_15_9
void FASTCALL call_2_6_15_9(SH2_struct* sh)
{
  SH2and(sh,6,15);
}
#define init_2_7_0_9
void FASTCALL call_2_7_0_9(SH2_struct* sh)
{
  SH2and(sh,7,0);
}
#define init_2_7_1_9
void FASTCALL call_2_7_1_9(SH2_struct* sh)
{
  SH2and(sh,7,1);
}
#define init_2_7_2_9
void FASTCALL call_2_7_2_9(SH2_struct* sh)
{
  SH2and(sh,7,2);
}
#define init_2_7_3_9
void FASTCALL call_2_7_3_9(SH2_struct* sh)
{
  SH2and(sh,7,3);
}
#define init_2_7_4_9
void FASTCALL call_2_7_4_9(SH2_struct* sh)
{
  SH2and(sh,7,4);
}
#define init_2_7_5_9
void FASTCALL call_2_7_5_9(SH2_struct* sh)
{
  SH2and(sh,7,5);
}
#define init_2_7_6_9
void FASTCALL call_2_7_6_9(SH2_struct* sh)
{
  SH2and(sh,7,6);
}
#define init_2_7_7_9
void FASTCALL call_2_7_7_9(SH2_struct* sh)
{
  SH2and(sh,7,7);
}
#define init_2_7_8_9
void FASTCALL call_2_7_8_9(SH2_struct* sh)
{
  SH2and(sh,7,8);
}
#define init_2_7_9_9
void FASTCALL call_2_7_9_9(SH2_struct* sh)
{
  SH2and(sh,7,9);
}
#define init_2_7_10_9
void FASTCALL call_2_7_10_9(SH2_struct* sh)
{
  SH2and(sh,7,10);
}
#define init_2_7_11_9
void FASTCALL call_2_7_11_9(SH2_struct* sh)
{
  SH2and(sh,7,11);
}
#define init_2_7_12_9
void FASTCALL call_2_7_12_9(SH2_struct* sh)
{
  SH2and(sh,7,12);
}
#define init_2_7_13_9
void FASTCALL call_2_7_13_9(SH2_struct* sh)
{
  SH2and(sh,7,13);
}
#define init_2_7_14_9
void FASTCALL call_2_7_14_9(SH2_struct* sh)
{
  SH2and(sh,7,14);
}
#define init_2_7_15_9
void FASTCALL call_2_7_15_9(SH2_struct* sh)
{
  SH2and(sh,7,15);
}
#define init_2_8_0_9
void FASTCALL call_2_8_0_9(SH2_struct* sh)
{
  SH2and(sh,8,0);
}
#define init_2_8_1_9
void FASTCALL call_2_8_1_9(SH2_struct* sh)
{
  SH2and(sh,8,1);
}
#define init_2_8_2_9
void FASTCALL call_2_8_2_9(SH2_struct* sh)
{
  SH2and(sh,8,2);
}
#define init_2_8_3_9
void FASTCALL call_2_8_3_9(SH2_struct* sh)
{
  SH2and(sh,8,3);
}
#define init_2_8_4_9
void FASTCALL call_2_8_4_9(SH2_struct* sh)
{
  SH2and(sh,8,4);
}
#define init_2_8_5_9
void FASTCALL call_2_8_5_9(SH2_struct* sh)
{
  SH2and(sh,8,5);
}
#define init_2_8_6_9
void FASTCALL call_2_8_6_9(SH2_struct* sh)
{
  SH2and(sh,8,6);
}
#define init_2_8_7_9
void FASTCALL call_2_8_7_9(SH2_struct* sh)
{
  SH2and(sh,8,7);
}
#define init_2_8_8_9
void FASTCALL call_2_8_8_9(SH2_struct* sh)
{
  SH2and(sh,8,8);
}
#define init_2_8_9_9
void FASTCALL call_2_8_9_9(SH2_struct* sh)
{
  SH2and(sh,8,9);
}
#define init_2_8_10_9
void FASTCALL call_2_8_10_9(SH2_struct* sh)
{
  SH2and(sh,8,10);
}
#define init_2_8_11_9
void FASTCALL call_2_8_11_9(SH2_struct* sh)
{
  SH2and(sh,8,11);
}
#define init_2_8_12_9
void FASTCALL call_2_8_12_9(SH2_struct* sh)
{
  SH2and(sh,8,12);
}
#define init_2_8_13_9
void FASTCALL call_2_8_13_9(SH2_struct* sh)
{
  SH2and(sh,8,13);
}
#define init_2_8_14_9
void FASTCALL call_2_8_14_9(SH2_struct* sh)
{
  SH2and(sh,8,14);
}
#define init_2_8_15_9
void FASTCALL call_2_8_15_9(SH2_struct* sh)
{
  SH2and(sh,8,15);
}
#define init_2_9_0_9
void FASTCALL call_2_9_0_9(SH2_struct* sh)
{
  SH2and(sh,9,0);
}
#define init_2_9_1_9
void FASTCALL call_2_9_1_9(SH2_struct* sh)
{
  SH2and(sh,9,1);
}
#define init_2_9_2_9
void FASTCALL call_2_9_2_9(SH2_struct* sh)
{
  SH2and(sh,9,2);
}
#define init_2_9_3_9
void FASTCALL call_2_9_3_9(SH2_struct* sh)
{
  SH2and(sh,9,3);
}
#define init_2_9_4_9
void FASTCALL call_2_9_4_9(SH2_struct* sh)
{
  SH2and(sh,9,4);
}
#define init_2_9_5_9
void FASTCALL call_2_9_5_9(SH2_struct* sh)
{
  SH2and(sh,9,5);
}
#define init_2_9_6_9
void FASTCALL call_2_9_6_9(SH2_struct* sh)
{
  SH2and(sh,9,6);
}
#define init_2_9_7_9
void FASTCALL call_2_9_7_9(SH2_struct* sh)
{
  SH2and(sh,9,7);
}
#define init_2_9_8_9
void FASTCALL call_2_9_8_9(SH2_struct* sh)
{
  SH2and(sh,9,8);
}
#define init_2_9_9_9
void FASTCALL call_2_9_9_9(SH2_struct* sh)
{
  SH2and(sh,9,9);
}
#define init_2_9_10_9
void FASTCALL call_2_9_10_9(SH2_struct* sh)
{
  SH2and(sh,9,10);
}
#define init_2_9_11_9
void FASTCALL call_2_9_11_9(SH2_struct* sh)
{
  SH2and(sh,9,11);
}
#define init_2_9_12_9
void FASTCALL call_2_9_12_9(SH2_struct* sh)
{
  SH2and(sh,9,12);
}
#define init_2_9_13_9
void FASTCALL call_2_9_13_9(SH2_struct* sh)
{
  SH2and(sh,9,13);
}
#define init_2_9_14_9
void FASTCALL call_2_9_14_9(SH2_struct* sh)
{
  SH2and(sh,9,14);
}
#define init_2_9_15_9
void FASTCALL call_2_9_15_9(SH2_struct* sh)
{
  SH2and(sh,9,15);
}
#define init_2_10_0_9
void FASTCALL call_2_10_0_9(SH2_struct* sh)
{
  SH2and(sh,10,0);
}
#define init_2_10_1_9
void FASTCALL call_2_10_1_9(SH2_struct* sh)
{
  SH2and(sh,10,1);
}
#define init_2_10_2_9
void FASTCALL call_2_10_2_9(SH2_struct* sh)
{
  SH2and(sh,10,2);
}
#define init_2_10_3_9
void FASTCALL call_2_10_3_9(SH2_struct* sh)
{
  SH2and(sh,10,3);
}
#define init_2_10_4_9
void FASTCALL call_2_10_4_9(SH2_struct* sh)
{
  SH2and(sh,10,4);
}
#define init_2_10_5_9
void FASTCALL call_2_10_5_9(SH2_struct* sh)
{
  SH2and(sh,10,5);
}
#define init_2_10_6_9
void FASTCALL call_2_10_6_9(SH2_struct* sh)
{
  SH2and(sh,10,6);
}
#define init_2_10_7_9
void FASTCALL call_2_10_7_9(SH2_struct* sh)
{
  SH2and(sh,10,7);
}
#define init_2_10_8_9
void FASTCALL call_2_10_8_9(SH2_struct* sh)
{
  SH2and(sh,10,8);
}
#define init_2_10_9_9
void FASTCALL call_2_10_9_9(SH2_struct* sh)
{
  SH2and(sh,10,9);
}
#define init_2_10_10_9
void FASTCALL call_2_10_10_9(SH2_struct* sh)
{
  SH2and(sh,10,10);
}
#define init_2_10_11_9
void FASTCALL call_2_10_11_9(SH2_struct* sh)
{
  SH2and(sh,10,11);
}
#define init_2_10_12_9
void FASTCALL call_2_10_12_9(SH2_struct* sh)
{
  SH2and(sh,10,12);
}
#define init_2_10_13_9
void FASTCALL call_2_10_13_9(SH2_struct* sh)
{
  SH2and(sh,10,13);
}
#define init_2_10_14_9
void FASTCALL call_2_10_14_9(SH2_struct* sh)
{
  SH2and(sh,10,14);
}
#define init_2_10_15_9
void FASTCALL call_2_10_15_9(SH2_struct* sh)
{
  SH2and(sh,10,15);
}
#define init_2_11_0_9
void FASTCALL call_2_11_0_9(SH2_struct* sh)
{
  SH2and(sh,11,0);
}
#define init_2_11_1_9
void FASTCALL call_2_11_1_9(SH2_struct* sh)
{
  SH2and(sh,11,1);
}
#define init_2_11_2_9
void FASTCALL call_2_11_2_9(SH2_struct* sh)
{
  SH2and(sh,11,2);
}
#define init_2_11_3_9
void FASTCALL call_2_11_3_9(SH2_struct* sh)
{
  SH2and(sh,11,3);
}
#define init_2_11_4_9
void FASTCALL call_2_11_4_9(SH2_struct* sh)
{
  SH2and(sh,11,4);
}
#define init_2_11_5_9
void FASTCALL call_2_11_5_9(SH2_struct* sh)
{
  SH2and(sh,11,5);
}
#define init_2_11_6_9
void FASTCALL call_2_11_6_9(SH2_struct* sh)
{
  SH2and(sh,11,6);
}
#define init_2_11_7_9
void FASTCALL call_2_11_7_9(SH2_struct* sh)
{
  SH2and(sh,11,7);
}
#define init_2_11_8_9
void FASTCALL call_2_11_8_9(SH2_struct* sh)
{
  SH2and(sh,11,8);
}
#define init_2_11_9_9
void FASTCALL call_2_11_9_9(SH2_struct* sh)
{
  SH2and(sh,11,9);
}
#define init_2_11_10_9
void FASTCALL call_2_11_10_9(SH2_struct* sh)
{
  SH2and(sh,11,10);
}
#define init_2_11_11_9
void FASTCALL call_2_11_11_9(SH2_struct* sh)
{
  SH2and(sh,11,11);
}
#define init_2_11_12_9
void FASTCALL call_2_11_12_9(SH2_struct* sh)
{
  SH2and(sh,11,12);
}
#define init_2_11_13_9
void FASTCALL call_2_11_13_9(SH2_struct* sh)
{
  SH2and(sh,11,13);
}
#define init_2_11_14_9
void FASTCALL call_2_11_14_9(SH2_struct* sh)
{
  SH2and(sh,11,14);
}
#define init_2_11_15_9
void FASTCALL call_2_11_15_9(SH2_struct* sh)
{
  SH2and(sh,11,15);
}
#define init_2_12_0_9
void FASTCALL call_2_12_0_9(SH2_struct* sh)
{
  SH2and(sh,12,0);
}
#define init_2_12_1_9
void FASTCALL call_2_12_1_9(SH2_struct* sh)
{
  SH2and(sh,12,1);
}
#define init_2_12_2_9
void FASTCALL call_2_12_2_9(SH2_struct* sh)
{
  SH2and(sh,12,2);
}
#define init_2_12_3_9
void FASTCALL call_2_12_3_9(SH2_struct* sh)
{
  SH2and(sh,12,3);
}
#define init_2_12_4_9
void FASTCALL call_2_12_4_9(SH2_struct* sh)
{
  SH2and(sh,12,4);
}
#define init_2_12_5_9
void FASTCALL call_2_12_5_9(SH2_struct* sh)
{
  SH2and(sh,12,5);
}
#define init_2_12_6_9
void FASTCALL call_2_12_6_9(SH2_struct* sh)
{
  SH2and(sh,12,6);
}
#define init_2_12_7_9
void FASTCALL call_2_12_7_9(SH2_struct* sh)
{
  SH2and(sh,12,7);
}
#define init_2_12_8_9
void FASTCALL call_2_12_8_9(SH2_struct* sh)
{
  SH2and(sh,12,8);
}
#define init_2_12_9_9
void FASTCALL call_2_12_9_9(SH2_struct* sh)
{
  SH2and(sh,12,9);
}
#define init_2_12_10_9
void FASTCALL call_2_12_10_9(SH2_struct* sh)
{
  SH2and(sh,12,10);
}
#define init_2_12_11_9
void FASTCALL call_2_12_11_9(SH2_struct* sh)
{
  SH2and(sh,12,11);
}
#define init_2_12_12_9
void FASTCALL call_2_12_12_9(SH2_struct* sh)
{
  SH2and(sh,12,12);
}
#define init_2_12_13_9
void FASTCALL call_2_12_13_9(SH2_struct* sh)
{
  SH2and(sh,12,13);
}
#define init_2_12_14_9
void FASTCALL call_2_12_14_9(SH2_struct* sh)
{
  SH2and(sh,12,14);
}
#define init_2_12_15_9
void FASTCALL call_2_12_15_9(SH2_struct* sh)
{
  SH2and(sh,12,15);
}
#define init_2_13_0_9
void FASTCALL call_2_13_0_9(SH2_struct* sh)
{
  SH2and(sh,13,0);
}
#define init_2_13_1_9
void FASTCALL call_2_13_1_9(SH2_struct* sh)
{
  SH2and(sh,13,1);
}
#define init_2_13_2_9
void FASTCALL call_2_13_2_9(SH2_struct* sh)
{
  SH2and(sh,13,2);
}
#define init_2_13_3_9
void FASTCALL call_2_13_3_9(SH2_struct* sh)
{
  SH2and(sh,13,3);
}
#define init_2_13_4_9
void FASTCALL call_2_13_4_9(SH2_struct* sh)
{
  SH2and(sh,13,4);
}
#define init_2_13_5_9
void FASTCALL call_2_13_5_9(SH2_struct* sh)
{
  SH2and(sh,13,5);
}
#define init_2_13_6_9
void FASTCALL call_2_13_6_9(SH2_struct* sh)
{
  SH2and(sh,13,6);
}
#define init_2_13_7_9
void FASTCALL call_2_13_7_9(SH2_struct* sh)
{
  SH2and(sh,13,7);
}
#define init_2_13_8_9
void FASTCALL call_2_13_8_9(SH2_struct* sh)
{
  SH2and(sh,13,8);
}
#define init_2_13_9_9
void FASTCALL call_2_13_9_9(SH2_struct* sh)
{
  SH2and(sh,13,9);
}
#define init_2_13_10_9
void FASTCALL call_2_13_10_9(SH2_struct* sh)
{
  SH2and(sh,13,10);
}
#define init_2_13_11_9
void FASTCALL call_2_13_11_9(SH2_struct* sh)
{
  SH2and(sh,13,11);
}
#define init_2_13_12_9
void FASTCALL call_2_13_12_9(SH2_struct* sh)
{
  SH2and(sh,13,12);
}
#define init_2_13_13_9
void FASTCALL call_2_13_13_9(SH2_struct* sh)
{
  SH2and(sh,13,13);
}
#define init_2_13_14_9
void FASTCALL call_2_13_14_9(SH2_struct* sh)
{
  SH2and(sh,13,14);
}
#define init_2_13_15_9
void FASTCALL call_2_13_15_9(SH2_struct* sh)
{
  SH2and(sh,13,15);
}
#define init_2_14_0_9
void FASTCALL call_2_14_0_9(SH2_struct* sh)
{
  SH2and(sh,14,0);
}
#define init_2_14_1_9
void FASTCALL call_2_14_1_9(SH2_struct* sh)
{
  SH2and(sh,14,1);
}
#define init_2_14_2_9
void FASTCALL call_2_14_2_9(SH2_struct* sh)
{
  SH2and(sh,14,2);
}
#define init_2_14_3_9
void FASTCALL call_2_14_3_9(SH2_struct* sh)
{
  SH2and(sh,14,3);
}
#define init_2_14_4_9
void FASTCALL call_2_14_4_9(SH2_struct* sh)
{
  SH2and(sh,14,4);
}
#define init_2_14_5_9
void FASTCALL call_2_14_5_9(SH2_struct* sh)
{
  SH2and(sh,14,5);
}
#define init_2_14_6_9
void FASTCALL call_2_14_6_9(SH2_struct* sh)
{
  SH2and(sh,14,6);
}
#define init_2_14_7_9
void FASTCALL call_2_14_7_9(SH2_struct* sh)
{
  SH2and(sh,14,7);
}
#define init_2_14_8_9
void FASTCALL call_2_14_8_9(SH2_struct* sh)
{
  SH2and(sh,14,8);
}
#define init_2_14_9_9
void FASTCALL call_2_14_9_9(SH2_struct* sh)
{
  SH2and(sh,14,9);
}
#define init_2_14_10_9
void FASTCALL call_2_14_10_9(SH2_struct* sh)
{
  SH2and(sh,14,10);
}
#define init_2_14_11_9
void FASTCALL call_2_14_11_9(SH2_struct* sh)
{
  SH2and(sh,14,11);
}
#define init_2_14_12_9
void FASTCALL call_2_14_12_9(SH2_struct* sh)
{
  SH2and(sh,14,12);
}
#define init_2_14_13_9
void FASTCALL call_2_14_13_9(SH2_struct* sh)
{
  SH2and(sh,14,13);
}
#define init_2_14_14_9
void FASTCALL call_2_14_14_9(SH2_struct* sh)
{
  SH2and(sh,14,14);
}
#define init_2_14_15_9
void FASTCALL call_2_14_15_9(SH2_struct* sh)
{
  SH2and(sh,14,15);
}
#define init_2_15_0_9
void FASTCALL call_2_15_0_9(SH2_struct* sh)
{
  SH2and(sh,15,0);
}
#define init_2_15_1_9
void FASTCALL call_2_15_1_9(SH2_struct* sh)
{
  SH2and(sh,15,1);
}
#define init_2_15_2_9
void FASTCALL call_2_15_2_9(SH2_struct* sh)
{
  SH2and(sh,15,2);
}
#define init_2_15_3_9
void FASTCALL call_2_15_3_9(SH2_struct* sh)
{
  SH2and(sh,15,3);
}
#define init_2_15_4_9
void FASTCALL call_2_15_4_9(SH2_struct* sh)
{
  SH2and(sh,15,4);
}
#define init_2_15_5_9
void FASTCALL call_2_15_5_9(SH2_struct* sh)
{
  SH2and(sh,15,5);
}
#define init_2_15_6_9
void FASTCALL call_2_15_6_9(SH2_struct* sh)
{
  SH2and(sh,15,6);
}
#define init_2_15_7_9
void FASTCALL call_2_15_7_9(SH2_struct* sh)
{
  SH2and(sh,15,7);
}
#define init_2_15_8_9
void FASTCALL call_2_15_8_9(SH2_struct* sh)
{
  SH2and(sh,15,8);
}
#define init_2_15_9_9
void FASTCALL call_2_15_9_9(SH2_struct* sh)
{
  SH2and(sh,15,9);
}
#define init_2_15_10_9
void FASTCALL call_2_15_10_9(SH2_struct* sh)
{
  SH2and(sh,15,10);
}
#define init_2_15_11_9
void FASTCALL call_2_15_11_9(SH2_struct* sh)
{
  SH2and(sh,15,11);
}
#define init_2_15_12_9
void FASTCALL call_2_15_12_9(SH2_struct* sh)
{
  SH2and(sh,15,12);
}
#define init_2_15_13_9
void FASTCALL call_2_15_13_9(SH2_struct* sh)
{
  SH2and(sh,15,13);
}
#define init_2_15_14_9
void FASTCALL call_2_15_14_9(SH2_struct* sh)
{
  SH2and(sh,15,14);
}
#define init_2_15_15_9
void FASTCALL call_2_15_15_9(SH2_struct* sh)
{
  SH2and(sh,15,15);
}
#define init_3_0_0_15
void FASTCALL call_3_0_0_15(SH2_struct* sh)
{
  SH2addv(sh,0,0);
}
#define init_3_0_1_15
void FASTCALL call_3_0_1_15(SH2_struct* sh)
{
  SH2addv(sh,0,1);
}
#define init_3_0_2_15
void FASTCALL call_3_0_2_15(SH2_struct* sh)
{
  SH2addv(sh,0,2);
}
#define init_3_0_3_15
void FASTCALL call_3_0_3_15(SH2_struct* sh)
{
  SH2addv(sh,0,3);
}
#define init_3_0_4_15
void FASTCALL call_3_0_4_15(SH2_struct* sh)
{
  SH2addv(sh,0,4);
}
#define init_3_0_5_15
void FASTCALL call_3_0_5_15(SH2_struct* sh)
{
  SH2addv(sh,0,5);
}
#define init_3_0_6_15
void FASTCALL call_3_0_6_15(SH2_struct* sh)
{
  SH2addv(sh,0,6);
}
#define init_3_0_7_15
void FASTCALL call_3_0_7_15(SH2_struct* sh)
{
  SH2addv(sh,0,7);
}
#define init_3_0_8_15
void FASTCALL call_3_0_8_15(SH2_struct* sh)
{
  SH2addv(sh,0,8);
}
#define init_3_0_9_15
void FASTCALL call_3_0_9_15(SH2_struct* sh)
{
  SH2addv(sh,0,9);
}
#define init_3_0_10_15
void FASTCALL call_3_0_10_15(SH2_struct* sh)
{
  SH2addv(sh,0,10);
}
#define init_3_0_11_15
void FASTCALL call_3_0_11_15(SH2_struct* sh)
{
  SH2addv(sh,0,11);
}
#define init_3_0_12_15
void FASTCALL call_3_0_12_15(SH2_struct* sh)
{
  SH2addv(sh,0,12);
}
#define init_3_0_13_15
void FASTCALL call_3_0_13_15(SH2_struct* sh)
{
  SH2addv(sh,0,13);
}
#define init_3_0_14_15
void FASTCALL call_3_0_14_15(SH2_struct* sh)
{
  SH2addv(sh,0,14);
}
#define init_3_0_15_15
void FASTCALL call_3_0_15_15(SH2_struct* sh)
{
  SH2addv(sh,0,15);
}
#define init_3_1_0_15
void FASTCALL call_3_1_0_15(SH2_struct* sh)
{
  SH2addv(sh,1,0);
}
#define init_3_1_1_15
void FASTCALL call_3_1_1_15(SH2_struct* sh)
{
  SH2addv(sh,1,1);
}
#define init_3_1_2_15
void FASTCALL call_3_1_2_15(SH2_struct* sh)
{
  SH2addv(sh,1,2);
}
#define init_3_1_3_15
void FASTCALL call_3_1_3_15(SH2_struct* sh)
{
  SH2addv(sh,1,3);
}
#define init_3_1_4_15
void FASTCALL call_3_1_4_15(SH2_struct* sh)
{
  SH2addv(sh,1,4);
}
#define init_3_1_5_15
void FASTCALL call_3_1_5_15(SH2_struct* sh)
{
  SH2addv(sh,1,5);
}
#define init_3_1_6_15
void FASTCALL call_3_1_6_15(SH2_struct* sh)
{
  SH2addv(sh,1,6);
}
#define init_3_1_7_15
void FASTCALL call_3_1_7_15(SH2_struct* sh)
{
  SH2addv(sh,1,7);
}
#define init_3_1_8_15
void FASTCALL call_3_1_8_15(SH2_struct* sh)
{
  SH2addv(sh,1,8);
}
#define init_3_1_9_15
void FASTCALL call_3_1_9_15(SH2_struct* sh)
{
  SH2addv(sh,1,9);
}
#define init_3_1_10_15
void FASTCALL call_3_1_10_15(SH2_struct* sh)
{
  SH2addv(sh,1,10);
}
#define init_3_1_11_15
void FASTCALL call_3_1_11_15(SH2_struct* sh)
{
  SH2addv(sh,1,11);
}
#define init_3_1_12_15
void FASTCALL call_3_1_12_15(SH2_struct* sh)
{
  SH2addv(sh,1,12);
}
#define init_3_1_13_15
void FASTCALL call_3_1_13_15(SH2_struct* sh)
{
  SH2addv(sh,1,13);
}
#define init_3_1_14_15
void FASTCALL call_3_1_14_15(SH2_struct* sh)
{
  SH2addv(sh,1,14);
}
#define init_3_1_15_15
void FASTCALL call_3_1_15_15(SH2_struct* sh)
{
  SH2addv(sh,1,15);
}
#define init_3_2_0_15
void FASTCALL call_3_2_0_15(SH2_struct* sh)
{
  SH2addv(sh,2,0);
}
#define init_3_2_1_15
void FASTCALL call_3_2_1_15(SH2_struct* sh)
{
  SH2addv(sh,2,1);
}
#define init_3_2_2_15
void FASTCALL call_3_2_2_15(SH2_struct* sh)
{
  SH2addv(sh,2,2);
}
#define init_3_2_3_15
void FASTCALL call_3_2_3_15(SH2_struct* sh)
{
  SH2addv(sh,2,3);
}
#define init_3_2_4_15
void FASTCALL call_3_2_4_15(SH2_struct* sh)
{
  SH2addv(sh,2,4);
}
#define init_3_2_5_15
void FASTCALL call_3_2_5_15(SH2_struct* sh)
{
  SH2addv(sh,2,5);
}
#define init_3_2_6_15
void FASTCALL call_3_2_6_15(SH2_struct* sh)
{
  SH2addv(sh,2,6);
}
#define init_3_2_7_15
void FASTCALL call_3_2_7_15(SH2_struct* sh)
{
  SH2addv(sh,2,7);
}
#define init_3_2_8_15
void FASTCALL call_3_2_8_15(SH2_struct* sh)
{
  SH2addv(sh,2,8);
}
#define init_3_2_9_15
void FASTCALL call_3_2_9_15(SH2_struct* sh)
{
  SH2addv(sh,2,9);
}
#define init_3_2_10_15
void FASTCALL call_3_2_10_15(SH2_struct* sh)
{
  SH2addv(sh,2,10);
}
#define init_3_2_11_15
void FASTCALL call_3_2_11_15(SH2_struct* sh)
{
  SH2addv(sh,2,11);
}
#define init_3_2_12_15
void FASTCALL call_3_2_12_15(SH2_struct* sh)
{
  SH2addv(sh,2,12);
}
#define init_3_2_13_15
void FASTCALL call_3_2_13_15(SH2_struct* sh)
{
  SH2addv(sh,2,13);
}
#define init_3_2_14_15
void FASTCALL call_3_2_14_15(SH2_struct* sh)
{
  SH2addv(sh,2,14);
}
#define init_3_2_15_15
void FASTCALL call_3_2_15_15(SH2_struct* sh)
{
  SH2addv(sh,2,15);
}
#define init_3_3_0_15
void FASTCALL call_3_3_0_15(SH2_struct* sh)
{
  SH2addv(sh,3,0);
}
#define init_3_3_1_15
void FASTCALL call_3_3_1_15(SH2_struct* sh)
{
  SH2addv(sh,3,1);
}
#define init_3_3_2_15
void FASTCALL call_3_3_2_15(SH2_struct* sh)
{
  SH2addv(sh,3,2);
}
#define init_3_3_3_15
void FASTCALL call_3_3_3_15(SH2_struct* sh)
{
  SH2addv(sh,3,3);
}
#define init_3_3_4_15
void FASTCALL call_3_3_4_15(SH2_struct* sh)
{
  SH2addv(sh,3,4);
}
#define init_3_3_5_15
void FASTCALL call_3_3_5_15(SH2_struct* sh)
{
  SH2addv(sh,3,5);
}
#define init_3_3_6_15
void FASTCALL call_3_3_6_15(SH2_struct* sh)
{
  SH2addv(sh,3,6);
}
#define init_3_3_7_15
void FASTCALL call_3_3_7_15(SH2_struct* sh)
{
  SH2addv(sh,3,7);
}
#define init_3_3_8_15
void FASTCALL call_3_3_8_15(SH2_struct* sh)
{
  SH2addv(sh,3,8);
}
#define init_3_3_9_15
void FASTCALL call_3_3_9_15(SH2_struct* sh)
{
  SH2addv(sh,3,9);
}
#define init_3_3_10_15
void FASTCALL call_3_3_10_15(SH2_struct* sh)
{
  SH2addv(sh,3,10);
}
#define init_3_3_11_15
void FASTCALL call_3_3_11_15(SH2_struct* sh)
{
  SH2addv(sh,3,11);
}
#define init_3_3_12_15
void FASTCALL call_3_3_12_15(SH2_struct* sh)
{
  SH2addv(sh,3,12);
}
#define init_3_3_13_15
void FASTCALL call_3_3_13_15(SH2_struct* sh)
{
  SH2addv(sh,3,13);
}
#define init_3_3_14_15
void FASTCALL call_3_3_14_15(SH2_struct* sh)
{
  SH2addv(sh,3,14);
}
#define init_3_3_15_15
void FASTCALL call_3_3_15_15(SH2_struct* sh)
{
  SH2addv(sh,3,15);
}
#define init_3_4_0_15
void FASTCALL call_3_4_0_15(SH2_struct* sh)
{
  SH2addv(sh,4,0);
}
#define init_3_4_1_15
void FASTCALL call_3_4_1_15(SH2_struct* sh)
{
  SH2addv(sh,4,1);
}
#define init_3_4_2_15
void FASTCALL call_3_4_2_15(SH2_struct* sh)
{
  SH2addv(sh,4,2);
}
#define init_3_4_3_15
void FASTCALL call_3_4_3_15(SH2_struct* sh)
{
  SH2addv(sh,4,3);
}
#define init_3_4_4_15
void FASTCALL call_3_4_4_15(SH2_struct* sh)
{
  SH2addv(sh,4,4);
}
#define init_3_4_5_15
void FASTCALL call_3_4_5_15(SH2_struct* sh)
{
  SH2addv(sh,4,5);
}
#define init_3_4_6_15
void FASTCALL call_3_4_6_15(SH2_struct* sh)
{
  SH2addv(sh,4,6);
}
#define init_3_4_7_15
void FASTCALL call_3_4_7_15(SH2_struct* sh)
{
  SH2addv(sh,4,7);
}
#define init_3_4_8_15
void FASTCALL call_3_4_8_15(SH2_struct* sh)
{
  SH2addv(sh,4,8);
}
#define init_3_4_9_15
void FASTCALL call_3_4_9_15(SH2_struct* sh)
{
  SH2addv(sh,4,9);
}
#define init_3_4_10_15
void FASTCALL call_3_4_10_15(SH2_struct* sh)
{
  SH2addv(sh,4,10);
}
#define init_3_4_11_15
void FASTCALL call_3_4_11_15(SH2_struct* sh)
{
  SH2addv(sh,4,11);
}
#define init_3_4_12_15
void FASTCALL call_3_4_12_15(SH2_struct* sh)
{
  SH2addv(sh,4,12);
}
#define init_3_4_13_15
void FASTCALL call_3_4_13_15(SH2_struct* sh)
{
  SH2addv(sh,4,13);
}
#define init_3_4_14_15
void FASTCALL call_3_4_14_15(SH2_struct* sh)
{
  SH2addv(sh,4,14);
}
#define init_3_4_15_15
void FASTCALL call_3_4_15_15(SH2_struct* sh)
{
  SH2addv(sh,4,15);
}
#define init_3_5_0_15
void FASTCALL call_3_5_0_15(SH2_struct* sh)
{
  SH2addv(sh,5,0);
}
#define init_3_5_1_15
void FASTCALL call_3_5_1_15(SH2_struct* sh)
{
  SH2addv(sh,5,1);
}
#define init_3_5_2_15
void FASTCALL call_3_5_2_15(SH2_struct* sh)
{
  SH2addv(sh,5,2);
}
#define init_3_5_3_15
void FASTCALL call_3_5_3_15(SH2_struct* sh)
{
  SH2addv(sh,5,3);
}
#define init_3_5_4_15
void FASTCALL call_3_5_4_15(SH2_struct* sh)
{
  SH2addv(sh,5,4);
}
#define init_3_5_5_15
void FASTCALL call_3_5_5_15(SH2_struct* sh)
{
  SH2addv(sh,5,5);
}
#define init_3_5_6_15
void FASTCALL call_3_5_6_15(SH2_struct* sh)
{
  SH2addv(sh,5,6);
}
#define init_3_5_7_15
void FASTCALL call_3_5_7_15(SH2_struct* sh)
{
  SH2addv(sh,5,7);
}
#define init_3_5_8_15
void FASTCALL call_3_5_8_15(SH2_struct* sh)
{
  SH2addv(sh,5,8);
}
#define init_3_5_9_15
void FASTCALL call_3_5_9_15(SH2_struct* sh)
{
  SH2addv(sh,5,9);
}
#define init_3_5_10_15
void FASTCALL call_3_5_10_15(SH2_struct* sh)
{
  SH2addv(sh,5,10);
}
#define init_3_5_11_15
void FASTCALL call_3_5_11_15(SH2_struct* sh)
{
  SH2addv(sh,5,11);
}
#define init_3_5_12_15
void FASTCALL call_3_5_12_15(SH2_struct* sh)
{
  SH2addv(sh,5,12);
}
#define init_3_5_13_15
void FASTCALL call_3_5_13_15(SH2_struct* sh)
{
  SH2addv(sh,5,13);
}
#define init_3_5_14_15
void FASTCALL call_3_5_14_15(SH2_struct* sh)
{
  SH2addv(sh,5,14);
}
#define init_3_5_15_15
void FASTCALL call_3_5_15_15(SH2_struct* sh)
{
  SH2addv(sh,5,15);
}
#define init_3_6_0_15
void FASTCALL call_3_6_0_15(SH2_struct* sh)
{
  SH2addv(sh,6,0);
}
#define init_3_6_1_15
void FASTCALL call_3_6_1_15(SH2_struct* sh)
{
  SH2addv(sh,6,1);
}
#define init_3_6_2_15
void FASTCALL call_3_6_2_15(SH2_struct* sh)
{
  SH2addv(sh,6,2);
}
#define init_3_6_3_15
void FASTCALL call_3_6_3_15(SH2_struct* sh)
{
  SH2addv(sh,6,3);
}
#define init_3_6_4_15
void FASTCALL call_3_6_4_15(SH2_struct* sh)
{
  SH2addv(sh,6,4);
}
#define init_3_6_5_15
void FASTCALL call_3_6_5_15(SH2_struct* sh)
{
  SH2addv(sh,6,5);
}
#define init_3_6_6_15
void FASTCALL call_3_6_6_15(SH2_struct* sh)
{
  SH2addv(sh,6,6);
}
#define init_3_6_7_15
void FASTCALL call_3_6_7_15(SH2_struct* sh)
{
  SH2addv(sh,6,7);
}
#define init_3_6_8_15
void FASTCALL call_3_6_8_15(SH2_struct* sh)
{
  SH2addv(sh,6,8);
}
#define init_3_6_9_15
void FASTCALL call_3_6_9_15(SH2_struct* sh)
{
  SH2addv(sh,6,9);
}
#define init_3_6_10_15
void FASTCALL call_3_6_10_15(SH2_struct* sh)
{
  SH2addv(sh,6,10);
}
#define init_3_6_11_15
void FASTCALL call_3_6_11_15(SH2_struct* sh)
{
  SH2addv(sh,6,11);
}
#define init_3_6_12_15
void FASTCALL call_3_6_12_15(SH2_struct* sh)
{
  SH2addv(sh,6,12);
}
#define init_3_6_13_15
void FASTCALL call_3_6_13_15(SH2_struct* sh)
{
  SH2addv(sh,6,13);
}
#define init_3_6_14_15
void FASTCALL call_3_6_14_15(SH2_struct* sh)
{
  SH2addv(sh,6,14);
}
#define init_3_6_15_15
void FASTCALL call_3_6_15_15(SH2_struct* sh)
{
  SH2addv(sh,6,15);
}
#define init_3_7_0_15
void FASTCALL call_3_7_0_15(SH2_struct* sh)
{
  SH2addv(sh,7,0);
}
#define init_3_7_1_15
void FASTCALL call_3_7_1_15(SH2_struct* sh)
{
  SH2addv(sh,7,1);
}
#define init_3_7_2_15
void FASTCALL call_3_7_2_15(SH2_struct* sh)
{
  SH2addv(sh,7,2);
}
#define init_3_7_3_15
void FASTCALL call_3_7_3_15(SH2_struct* sh)
{
  SH2addv(sh,7,3);
}
#define init_3_7_4_15
void FASTCALL call_3_7_4_15(SH2_struct* sh)
{
  SH2addv(sh,7,4);
}
#define init_3_7_5_15
void FASTCALL call_3_7_5_15(SH2_struct* sh)
{
  SH2addv(sh,7,5);
}
#define init_3_7_6_15
void FASTCALL call_3_7_6_15(SH2_struct* sh)
{
  SH2addv(sh,7,6);
}
#define init_3_7_7_15
void FASTCALL call_3_7_7_15(SH2_struct* sh)
{
  SH2addv(sh,7,7);
}
#define init_3_7_8_15
void FASTCALL call_3_7_8_15(SH2_struct* sh)
{
  SH2addv(sh,7,8);
}
#define init_3_7_9_15
void FASTCALL call_3_7_9_15(SH2_struct* sh)
{
  SH2addv(sh,7,9);
}
#define init_3_7_10_15
void FASTCALL call_3_7_10_15(SH2_struct* sh)
{
  SH2addv(sh,7,10);
}
#define init_3_7_11_15
void FASTCALL call_3_7_11_15(SH2_struct* sh)
{
  SH2addv(sh,7,11);
}
#define init_3_7_12_15
void FASTCALL call_3_7_12_15(SH2_struct* sh)
{
  SH2addv(sh,7,12);
}
#define init_3_7_13_15
void FASTCALL call_3_7_13_15(SH2_struct* sh)
{
  SH2addv(sh,7,13);
}
#define init_3_7_14_15
void FASTCALL call_3_7_14_15(SH2_struct* sh)
{
  SH2addv(sh,7,14);
}
#define init_3_7_15_15
void FASTCALL call_3_7_15_15(SH2_struct* sh)
{
  SH2addv(sh,7,15);
}
#define init_3_8_0_15
void FASTCALL call_3_8_0_15(SH2_struct* sh)
{
  SH2addv(sh,8,0);
}
#define init_3_8_1_15
void FASTCALL call_3_8_1_15(SH2_struct* sh)
{
  SH2addv(sh,8,1);
}
#define init_3_8_2_15
void FASTCALL call_3_8_2_15(SH2_struct* sh)
{
  SH2addv(sh,8,2);
}
#define init_3_8_3_15
void FASTCALL call_3_8_3_15(SH2_struct* sh)
{
  SH2addv(sh,8,3);
}
#define init_3_8_4_15
void FASTCALL call_3_8_4_15(SH2_struct* sh)
{
  SH2addv(sh,8,4);
}
#define init_3_8_5_15
void FASTCALL call_3_8_5_15(SH2_struct* sh)
{
  SH2addv(sh,8,5);
}
#define init_3_8_6_15
void FASTCALL call_3_8_6_15(SH2_struct* sh)
{
  SH2addv(sh,8,6);
}
#define init_3_8_7_15
void FASTCALL call_3_8_7_15(SH2_struct* sh)
{
  SH2addv(sh,8,7);
}
#define init_3_8_8_15
void FASTCALL call_3_8_8_15(SH2_struct* sh)
{
  SH2addv(sh,8,8);
}
#define init_3_8_9_15
void FASTCALL call_3_8_9_15(SH2_struct* sh)
{
  SH2addv(sh,8,9);
}
#define init_3_8_10_15
void FASTCALL call_3_8_10_15(SH2_struct* sh)
{
  SH2addv(sh,8,10);
}
#define init_3_8_11_15
void FASTCALL call_3_8_11_15(SH2_struct* sh)
{
  SH2addv(sh,8,11);
}
#define init_3_8_12_15
void FASTCALL call_3_8_12_15(SH2_struct* sh)
{
  SH2addv(sh,8,12);
}
#define init_3_8_13_15
void FASTCALL call_3_8_13_15(SH2_struct* sh)
{
  SH2addv(sh,8,13);
}
#define init_3_8_14_15
void FASTCALL call_3_8_14_15(SH2_struct* sh)
{
  SH2addv(sh,8,14);
}
#define init_3_8_15_15
void FASTCALL call_3_8_15_15(SH2_struct* sh)
{
  SH2addv(sh,8,15);
}
#define init_3_9_0_15
void FASTCALL call_3_9_0_15(SH2_struct* sh)
{
  SH2addv(sh,9,0);
}
#define init_3_9_1_15
void FASTCALL call_3_9_1_15(SH2_struct* sh)
{
  SH2addv(sh,9,1);
}
#define init_3_9_2_15
void FASTCALL call_3_9_2_15(SH2_struct* sh)
{
  SH2addv(sh,9,2);
}
#define init_3_9_3_15
void FASTCALL call_3_9_3_15(SH2_struct* sh)
{
  SH2addv(sh,9,3);
}
#define init_3_9_4_15
void FASTCALL call_3_9_4_15(SH2_struct* sh)
{
  SH2addv(sh,9,4);
}
#define init_3_9_5_15
void FASTCALL call_3_9_5_15(SH2_struct* sh)
{
  SH2addv(sh,9,5);
}
#define init_3_9_6_15
void FASTCALL call_3_9_6_15(SH2_struct* sh)
{
  SH2addv(sh,9,6);
}
#define init_3_9_7_15
void FASTCALL call_3_9_7_15(SH2_struct* sh)
{
  SH2addv(sh,9,7);
}
#define init_3_9_8_15
void FASTCALL call_3_9_8_15(SH2_struct* sh)
{
  SH2addv(sh,9,8);
}
#define init_3_9_9_15
void FASTCALL call_3_9_9_15(SH2_struct* sh)
{
  SH2addv(sh,9,9);
}
#define init_3_9_10_15
void FASTCALL call_3_9_10_15(SH2_struct* sh)
{
  SH2addv(sh,9,10);
}
#define init_3_9_11_15
void FASTCALL call_3_9_11_15(SH2_struct* sh)
{
  SH2addv(sh,9,11);
}
#define init_3_9_12_15
void FASTCALL call_3_9_12_15(SH2_struct* sh)
{
  SH2addv(sh,9,12);
}
#define init_3_9_13_15
void FASTCALL call_3_9_13_15(SH2_struct* sh)
{
  SH2addv(sh,9,13);
}
#define init_3_9_14_15
void FASTCALL call_3_9_14_15(SH2_struct* sh)
{
  SH2addv(sh,9,14);
}
#define init_3_9_15_15
void FASTCALL call_3_9_15_15(SH2_struct* sh)
{
  SH2addv(sh,9,15);
}
#define init_3_10_0_15
void FASTCALL call_3_10_0_15(SH2_struct* sh)
{
  SH2addv(sh,10,0);
}
#define init_3_10_1_15
void FASTCALL call_3_10_1_15(SH2_struct* sh)
{
  SH2addv(sh,10,1);
}
#define init_3_10_2_15
void FASTCALL call_3_10_2_15(SH2_struct* sh)
{
  SH2addv(sh,10,2);
}
#define init_3_10_3_15
void FASTCALL call_3_10_3_15(SH2_struct* sh)
{
  SH2addv(sh,10,3);
}
#define init_3_10_4_15
void FASTCALL call_3_10_4_15(SH2_struct* sh)
{
  SH2addv(sh,10,4);
}
#define init_3_10_5_15
void FASTCALL call_3_10_5_15(SH2_struct* sh)
{
  SH2addv(sh,10,5);
}
#define init_3_10_6_15
void FASTCALL call_3_10_6_15(SH2_struct* sh)
{
  SH2addv(sh,10,6);
}
#define init_3_10_7_15
void FASTCALL call_3_10_7_15(SH2_struct* sh)
{
  SH2addv(sh,10,7);
}
#define init_3_10_8_15
void FASTCALL call_3_10_8_15(SH2_struct* sh)
{
  SH2addv(sh,10,8);
}
#define init_3_10_9_15
void FASTCALL call_3_10_9_15(SH2_struct* sh)
{
  SH2addv(sh,10,9);
}
#define init_3_10_10_15
void FASTCALL call_3_10_10_15(SH2_struct* sh)
{
  SH2addv(sh,10,10);
}
#define init_3_10_11_15
void FASTCALL call_3_10_11_15(SH2_struct* sh)
{
  SH2addv(sh,10,11);
}
#define init_3_10_12_15
void FASTCALL call_3_10_12_15(SH2_struct* sh)
{
  SH2addv(sh,10,12);
}
#define init_3_10_13_15
void FASTCALL call_3_10_13_15(SH2_struct* sh)
{
  SH2addv(sh,10,13);
}
#define init_3_10_14_15
void FASTCALL call_3_10_14_15(SH2_struct* sh)
{
  SH2addv(sh,10,14);
}
#define init_3_10_15_15
void FASTCALL call_3_10_15_15(SH2_struct* sh)
{
  SH2addv(sh,10,15);
}
#define init_3_11_0_15
void FASTCALL call_3_11_0_15(SH2_struct* sh)
{
  SH2addv(sh,11,0);
}
#define init_3_11_1_15
void FASTCALL call_3_11_1_15(SH2_struct* sh)
{
  SH2addv(sh,11,1);
}
#define init_3_11_2_15
void FASTCALL call_3_11_2_15(SH2_struct* sh)
{
  SH2addv(sh,11,2);
}
#define init_3_11_3_15
void FASTCALL call_3_11_3_15(SH2_struct* sh)
{
  SH2addv(sh,11,3);
}
#define init_3_11_4_15
void FASTCALL call_3_11_4_15(SH2_struct* sh)
{
  SH2addv(sh,11,4);
}
#define init_3_11_5_15
void FASTCALL call_3_11_5_15(SH2_struct* sh)
{
  SH2addv(sh,11,5);
}
#define init_3_11_6_15
void FASTCALL call_3_11_6_15(SH2_struct* sh)
{
  SH2addv(sh,11,6);
}
#define init_3_11_7_15
void FASTCALL call_3_11_7_15(SH2_struct* sh)
{
  SH2addv(sh,11,7);
}
#define init_3_11_8_15
void FASTCALL call_3_11_8_15(SH2_struct* sh)
{
  SH2addv(sh,11,8);
}
#define init_3_11_9_15
void FASTCALL call_3_11_9_15(SH2_struct* sh)
{
  SH2addv(sh,11,9);
}
#define init_3_11_10_15
void FASTCALL call_3_11_10_15(SH2_struct* sh)
{
  SH2addv(sh,11,10);
}
#define init_3_11_11_15
void FASTCALL call_3_11_11_15(SH2_struct* sh)
{
  SH2addv(sh,11,11);
}
#define init_3_11_12_15
void FASTCALL call_3_11_12_15(SH2_struct* sh)
{
  SH2addv(sh,11,12);
}
#define init_3_11_13_15
void FASTCALL call_3_11_13_15(SH2_struct* sh)
{
  SH2addv(sh,11,13);
}
#define init_3_11_14_15
void FASTCALL call_3_11_14_15(SH2_struct* sh)
{
  SH2addv(sh,11,14);
}
#define init_3_11_15_15
void FASTCALL call_3_11_15_15(SH2_struct* sh)
{
  SH2addv(sh,11,15);
}
#define init_3_12_0_15
void FASTCALL call_3_12_0_15(SH2_struct* sh)
{
  SH2addv(sh,12,0);
}
#define init_3_12_1_15
void FASTCALL call_3_12_1_15(SH2_struct* sh)
{
  SH2addv(sh,12,1);
}
#define init_3_12_2_15
void FASTCALL call_3_12_2_15(SH2_struct* sh)
{
  SH2addv(sh,12,2);
}
#define init_3_12_3_15
void FASTCALL call_3_12_3_15(SH2_struct* sh)
{
  SH2addv(sh,12,3);
}
#define init_3_12_4_15
void FASTCALL call_3_12_4_15(SH2_struct* sh)
{
  SH2addv(sh,12,4);
}
#define init_3_12_5_15
void FASTCALL call_3_12_5_15(SH2_struct* sh)
{
  SH2addv(sh,12,5);
}
#define init_3_12_6_15
void FASTCALL call_3_12_6_15(SH2_struct* sh)
{
  SH2addv(sh,12,6);
}
#define init_3_12_7_15
void FASTCALL call_3_12_7_15(SH2_struct* sh)
{
  SH2addv(sh,12,7);
}
#define init_3_12_8_15
void FASTCALL call_3_12_8_15(SH2_struct* sh)
{
  SH2addv(sh,12,8);
}
#define init_3_12_9_15
void FASTCALL call_3_12_9_15(SH2_struct* sh)
{
  SH2addv(sh,12,9);
}
#define init_3_12_10_15
void FASTCALL call_3_12_10_15(SH2_struct* sh)
{
  SH2addv(sh,12,10);
}
#define init_3_12_11_15
void FASTCALL call_3_12_11_15(SH2_struct* sh)
{
  SH2addv(sh,12,11);
}
#define init_3_12_12_15
void FASTCALL call_3_12_12_15(SH2_struct* sh)
{
  SH2addv(sh,12,12);
}
#define init_3_12_13_15
void FASTCALL call_3_12_13_15(SH2_struct* sh)
{
  SH2addv(sh,12,13);
}
#define init_3_12_14_15
void FASTCALL call_3_12_14_15(SH2_struct* sh)
{
  SH2addv(sh,12,14);
}
#define init_3_12_15_15
void FASTCALL call_3_12_15_15(SH2_struct* sh)
{
  SH2addv(sh,12,15);
}
#define init_3_13_0_15
void FASTCALL call_3_13_0_15(SH2_struct* sh)
{
  SH2addv(sh,13,0);
}
#define init_3_13_1_15
void FASTCALL call_3_13_1_15(SH2_struct* sh)
{
  SH2addv(sh,13,1);
}
#define init_3_13_2_15
void FASTCALL call_3_13_2_15(SH2_struct* sh)
{
  SH2addv(sh,13,2);
}
#define init_3_13_3_15
void FASTCALL call_3_13_3_15(SH2_struct* sh)
{
  SH2addv(sh,13,3);
}
#define init_3_13_4_15
void FASTCALL call_3_13_4_15(SH2_struct* sh)
{
  SH2addv(sh,13,4);
}
#define init_3_13_5_15
void FASTCALL call_3_13_5_15(SH2_struct* sh)
{
  SH2addv(sh,13,5);
}
#define init_3_13_6_15
void FASTCALL call_3_13_6_15(SH2_struct* sh)
{
  SH2addv(sh,13,6);
}
#define init_3_13_7_15
void FASTCALL call_3_13_7_15(SH2_struct* sh)
{
  SH2addv(sh,13,7);
}
#define init_3_13_8_15
void FASTCALL call_3_13_8_15(SH2_struct* sh)
{
  SH2addv(sh,13,8);
}
#define init_3_13_9_15
void FASTCALL call_3_13_9_15(SH2_struct* sh)
{
  SH2addv(sh,13,9);
}
#define init_3_13_10_15
void FASTCALL call_3_13_10_15(SH2_struct* sh)
{
  SH2addv(sh,13,10);
}
#define init_3_13_11_15
void FASTCALL call_3_13_11_15(SH2_struct* sh)
{
  SH2addv(sh,13,11);
}
#define init_3_13_12_15
void FASTCALL call_3_13_12_15(SH2_struct* sh)
{
  SH2addv(sh,13,12);
}
#define init_3_13_13_15
void FASTCALL call_3_13_13_15(SH2_struct* sh)
{
  SH2addv(sh,13,13);
}
#define init_3_13_14_15
void FASTCALL call_3_13_14_15(SH2_struct* sh)
{
  SH2addv(sh,13,14);
}
#define init_3_13_15_15
void FASTCALL call_3_13_15_15(SH2_struct* sh)
{
  SH2addv(sh,13,15);
}
#define init_3_14_0_15
void FASTCALL call_3_14_0_15(SH2_struct* sh)
{
  SH2addv(sh,14,0);
}
#define init_3_14_1_15
void FASTCALL call_3_14_1_15(SH2_struct* sh)
{
  SH2addv(sh,14,1);
}
#define init_3_14_2_15
void FASTCALL call_3_14_2_15(SH2_struct* sh)
{
  SH2addv(sh,14,2);
}
#define init_3_14_3_15
void FASTCALL call_3_14_3_15(SH2_struct* sh)
{
  SH2addv(sh,14,3);
}
#define init_3_14_4_15
void FASTCALL call_3_14_4_15(SH2_struct* sh)
{
  SH2addv(sh,14,4);
}
#define init_3_14_5_15
void FASTCALL call_3_14_5_15(SH2_struct* sh)
{
  SH2addv(sh,14,5);
}
#define init_3_14_6_15
void FASTCALL call_3_14_6_15(SH2_struct* sh)
{
  SH2addv(sh,14,6);
}
#define init_3_14_7_15
void FASTCALL call_3_14_7_15(SH2_struct* sh)
{
  SH2addv(sh,14,7);
}
#define init_3_14_8_15
void FASTCALL call_3_14_8_15(SH2_struct* sh)
{
  SH2addv(sh,14,8);
}
#define init_3_14_9_15
void FASTCALL call_3_14_9_15(SH2_struct* sh)
{
  SH2addv(sh,14,9);
}
#define init_3_14_10_15
void FASTCALL call_3_14_10_15(SH2_struct* sh)
{
  SH2addv(sh,14,10);
}
#define init_3_14_11_15
void FASTCALL call_3_14_11_15(SH2_struct* sh)
{
  SH2addv(sh,14,11);
}
#define init_3_14_12_15
void FASTCALL call_3_14_12_15(SH2_struct* sh)
{
  SH2addv(sh,14,12);
}
#define init_3_14_13_15
void FASTCALL call_3_14_13_15(SH2_struct* sh)
{
  SH2addv(sh,14,13);
}
#define init_3_14_14_15
void FASTCALL call_3_14_14_15(SH2_struct* sh)
{
  SH2addv(sh,14,14);
}
#define init_3_14_15_15
void FASTCALL call_3_14_15_15(SH2_struct* sh)
{
  SH2addv(sh,14,15);
}
#define init_3_15_0_15
void FASTCALL call_3_15_0_15(SH2_struct* sh)
{
  SH2addv(sh,15,0);
}
#define init_3_15_1_15
void FASTCALL call_3_15_1_15(SH2_struct* sh)
{
  SH2addv(sh,15,1);
}
#define init_3_15_2_15
void FASTCALL call_3_15_2_15(SH2_struct* sh)
{
  SH2addv(sh,15,2);
}
#define init_3_15_3_15
void FASTCALL call_3_15_3_15(SH2_struct* sh)
{
  SH2addv(sh,15,3);
}
#define init_3_15_4_15
void FASTCALL call_3_15_4_15(SH2_struct* sh)
{
  SH2addv(sh,15,4);
}
#define init_3_15_5_15
void FASTCALL call_3_15_5_15(SH2_struct* sh)
{
  SH2addv(sh,15,5);
}
#define init_3_15_6_15
void FASTCALL call_3_15_6_15(SH2_struct* sh)
{
  SH2addv(sh,15,6);
}
#define init_3_15_7_15
void FASTCALL call_3_15_7_15(SH2_struct* sh)
{
  SH2addv(sh,15,7);
}
#define init_3_15_8_15
void FASTCALL call_3_15_8_15(SH2_struct* sh)
{
  SH2addv(sh,15,8);
}
#define init_3_15_9_15
void FASTCALL call_3_15_9_15(SH2_struct* sh)
{
  SH2addv(sh,15,9);
}
#define init_3_15_10_15
void FASTCALL call_3_15_10_15(SH2_struct* sh)
{
  SH2addv(sh,15,10);
}
#define init_3_15_11_15
void FASTCALL call_3_15_11_15(SH2_struct* sh)
{
  SH2addv(sh,15,11);
}
#define init_3_15_12_15
void FASTCALL call_3_15_12_15(SH2_struct* sh)
{
  SH2addv(sh,15,12);
}
#define init_3_15_13_15
void FASTCALL call_3_15_13_15(SH2_struct* sh)
{
  SH2addv(sh,15,13);
}
#define init_3_15_14_15
void FASTCALL call_3_15_14_15(SH2_struct* sh)
{
  SH2addv(sh,15,14);
}
#define init_3_15_15_15
void FASTCALL call_3_15_15_15(SH2_struct* sh)
{
  SH2addv(sh,15,15);
}
#define init_3_0_0_14
void FASTCALL call_3_0_0_14(SH2_struct* sh)
{
  SH2addc(sh,0,0);
}
#define init_3_0_1_14
void FASTCALL call_3_0_1_14(SH2_struct* sh)
{
  SH2addc(sh,0,1);
}
#define init_3_0_2_14
void FASTCALL call_3_0_2_14(SH2_struct* sh)
{
  SH2addc(sh,0,2);
}
#define init_3_0_3_14
void FASTCALL call_3_0_3_14(SH2_struct* sh)
{
  SH2addc(sh,0,3);
}
#define init_3_0_4_14
void FASTCALL call_3_0_4_14(SH2_struct* sh)
{
  SH2addc(sh,0,4);
}
#define init_3_0_5_14
void FASTCALL call_3_0_5_14(SH2_struct* sh)
{
  SH2addc(sh,0,5);
}
#define init_3_0_6_14
void FASTCALL call_3_0_6_14(SH2_struct* sh)
{
  SH2addc(sh,0,6);
}
#define init_3_0_7_14
void FASTCALL call_3_0_7_14(SH2_struct* sh)
{
  SH2addc(sh,0,7);
}
#define init_3_0_8_14
void FASTCALL call_3_0_8_14(SH2_struct* sh)
{
  SH2addc(sh,0,8);
}
#define init_3_0_9_14
void FASTCALL call_3_0_9_14(SH2_struct* sh)
{
  SH2addc(sh,0,9);
}
#define init_3_0_10_14
void FASTCALL call_3_0_10_14(SH2_struct* sh)
{
  SH2addc(sh,0,10);
}
#define init_3_0_11_14
void FASTCALL call_3_0_11_14(SH2_struct* sh)
{
  SH2addc(sh,0,11);
}
#define init_3_0_12_14
void FASTCALL call_3_0_12_14(SH2_struct* sh)
{
  SH2addc(sh,0,12);
}
#define init_3_0_13_14
void FASTCALL call_3_0_13_14(SH2_struct* sh)
{
  SH2addc(sh,0,13);
}
#define init_3_0_14_14
void FASTCALL call_3_0_14_14(SH2_struct* sh)
{
  SH2addc(sh,0,14);
}
#define init_3_0_15_14
void FASTCALL call_3_0_15_14(SH2_struct* sh)
{
  SH2addc(sh,0,15);
}
#define init_3_1_0_14
void FASTCALL call_3_1_0_14(SH2_struct* sh)
{
  SH2addc(sh,1,0);
}
#define init_3_1_1_14
void FASTCALL call_3_1_1_14(SH2_struct* sh)
{
  SH2addc(sh,1,1);
}
#define init_3_1_2_14
void FASTCALL call_3_1_2_14(SH2_struct* sh)
{
  SH2addc(sh,1,2);
}
#define init_3_1_3_14
void FASTCALL call_3_1_3_14(SH2_struct* sh)
{
  SH2addc(sh,1,3);
}
#define init_3_1_4_14
void FASTCALL call_3_1_4_14(SH2_struct* sh)
{
  SH2addc(sh,1,4);
}
#define init_3_1_5_14
void FASTCALL call_3_1_5_14(SH2_struct* sh)
{
  SH2addc(sh,1,5);
}
#define init_3_1_6_14
void FASTCALL call_3_1_6_14(SH2_struct* sh)
{
  SH2addc(sh,1,6);
}
#define init_3_1_7_14
void FASTCALL call_3_1_7_14(SH2_struct* sh)
{
  SH2addc(sh,1,7);
}
#define init_3_1_8_14
void FASTCALL call_3_1_8_14(SH2_struct* sh)
{
  SH2addc(sh,1,8);
}
#define init_3_1_9_14
void FASTCALL call_3_1_9_14(SH2_struct* sh)
{
  SH2addc(sh,1,9);
}
#define init_3_1_10_14
void FASTCALL call_3_1_10_14(SH2_struct* sh)
{
  SH2addc(sh,1,10);
}
#define init_3_1_11_14
void FASTCALL call_3_1_11_14(SH2_struct* sh)
{
  SH2addc(sh,1,11);
}
#define init_3_1_12_14
void FASTCALL call_3_1_12_14(SH2_struct* sh)
{
  SH2addc(sh,1,12);
}
#define init_3_1_13_14
void FASTCALL call_3_1_13_14(SH2_struct* sh)
{
  SH2addc(sh,1,13);
}
#define init_3_1_14_14
void FASTCALL call_3_1_14_14(SH2_struct* sh)
{
  SH2addc(sh,1,14);
}
#define init_3_1_15_14
void FASTCALL call_3_1_15_14(SH2_struct* sh)
{
  SH2addc(sh,1,15);
}
#define init_3_2_0_14
void FASTCALL call_3_2_0_14(SH2_struct* sh)
{
  SH2addc(sh,2,0);
}
#define init_3_2_1_14
void FASTCALL call_3_2_1_14(SH2_struct* sh)
{
  SH2addc(sh,2,1);
}
#define init_3_2_2_14
void FASTCALL call_3_2_2_14(SH2_struct* sh)
{
  SH2addc(sh,2,2);
}
#define init_3_2_3_14
void FASTCALL call_3_2_3_14(SH2_struct* sh)
{
  SH2addc(sh,2,3);
}
#define init_3_2_4_14
void FASTCALL call_3_2_4_14(SH2_struct* sh)
{
  SH2addc(sh,2,4);
}
#define init_3_2_5_14
void FASTCALL call_3_2_5_14(SH2_struct* sh)
{
  SH2addc(sh,2,5);
}
#define init_3_2_6_14
void FASTCALL call_3_2_6_14(SH2_struct* sh)
{
  SH2addc(sh,2,6);
}
#define init_3_2_7_14
void FASTCALL call_3_2_7_14(SH2_struct* sh)
{
  SH2addc(sh,2,7);
}
#define init_3_2_8_14
void FASTCALL call_3_2_8_14(SH2_struct* sh)
{
  SH2addc(sh,2,8);
}
#define init_3_2_9_14
void FASTCALL call_3_2_9_14(SH2_struct* sh)
{
  SH2addc(sh,2,9);
}
#define init_3_2_10_14
void FASTCALL call_3_2_10_14(SH2_struct* sh)
{
  SH2addc(sh,2,10);
}
#define init_3_2_11_14
void FASTCALL call_3_2_11_14(SH2_struct* sh)
{
  SH2addc(sh,2,11);
}
#define init_3_2_12_14
void FASTCALL call_3_2_12_14(SH2_struct* sh)
{
  SH2addc(sh,2,12);
}
#define init_3_2_13_14
void FASTCALL call_3_2_13_14(SH2_struct* sh)
{
  SH2addc(sh,2,13);
}
#define init_3_2_14_14
void FASTCALL call_3_2_14_14(SH2_struct* sh)
{
  SH2addc(sh,2,14);
}
#define init_3_2_15_14
void FASTCALL call_3_2_15_14(SH2_struct* sh)
{
  SH2addc(sh,2,15);
}
#define init_3_3_0_14
void FASTCALL call_3_3_0_14(SH2_struct* sh)
{
  SH2addc(sh,3,0);
}
#define init_3_3_1_14
void FASTCALL call_3_3_1_14(SH2_struct* sh)
{
  SH2addc(sh,3,1);
}
#define init_3_3_2_14
void FASTCALL call_3_3_2_14(SH2_struct* sh)
{
  SH2addc(sh,3,2);
}
#define init_3_3_3_14
void FASTCALL call_3_3_3_14(SH2_struct* sh)
{
  SH2addc(sh,3,3);
}
#define init_3_3_4_14
void FASTCALL call_3_3_4_14(SH2_struct* sh)
{
  SH2addc(sh,3,4);
}
#define init_3_3_5_14
void FASTCALL call_3_3_5_14(SH2_struct* sh)
{
  SH2addc(sh,3,5);
}
#define init_3_3_6_14
void FASTCALL call_3_3_6_14(SH2_struct* sh)
{
  SH2addc(sh,3,6);
}
#define init_3_3_7_14
void FASTCALL call_3_3_7_14(SH2_struct* sh)
{
  SH2addc(sh,3,7);
}
#define init_3_3_8_14
void FASTCALL call_3_3_8_14(SH2_struct* sh)
{
  SH2addc(sh,3,8);
}
#define init_3_3_9_14
void FASTCALL call_3_3_9_14(SH2_struct* sh)
{
  SH2addc(sh,3,9);
}
#define init_3_3_10_14
void FASTCALL call_3_3_10_14(SH2_struct* sh)
{
  SH2addc(sh,3,10);
}
#define init_3_3_11_14
void FASTCALL call_3_3_11_14(SH2_struct* sh)
{
  SH2addc(sh,3,11);
}
#define init_3_3_12_14
void FASTCALL call_3_3_12_14(SH2_struct* sh)
{
  SH2addc(sh,3,12);
}
#define init_3_3_13_14
void FASTCALL call_3_3_13_14(SH2_struct* sh)
{
  SH2addc(sh,3,13);
}
#define init_3_3_14_14
void FASTCALL call_3_3_14_14(SH2_struct* sh)
{
  SH2addc(sh,3,14);
}
#define init_3_3_15_14
void FASTCALL call_3_3_15_14(SH2_struct* sh)
{
  SH2addc(sh,3,15);
}
#define init_3_4_0_14
void FASTCALL call_3_4_0_14(SH2_struct* sh)
{
  SH2addc(sh,4,0);
}
#define init_3_4_1_14
void FASTCALL call_3_4_1_14(SH2_struct* sh)
{
  SH2addc(sh,4,1);
}
#define init_3_4_2_14
void FASTCALL call_3_4_2_14(SH2_struct* sh)
{
  SH2addc(sh,4,2);
}
#define init_3_4_3_14
void FASTCALL call_3_4_3_14(SH2_struct* sh)
{
  SH2addc(sh,4,3);
}
#define init_3_4_4_14
void FASTCALL call_3_4_4_14(SH2_struct* sh)
{
  SH2addc(sh,4,4);
}
#define init_3_4_5_14
void FASTCALL call_3_4_5_14(SH2_struct* sh)
{
  SH2addc(sh,4,5);
}
#define init_3_4_6_14
void FASTCALL call_3_4_6_14(SH2_struct* sh)
{
  SH2addc(sh,4,6);
}
#define init_3_4_7_14
void FASTCALL call_3_4_7_14(SH2_struct* sh)
{
  SH2addc(sh,4,7);
}
#define init_3_4_8_14
void FASTCALL call_3_4_8_14(SH2_struct* sh)
{
  SH2addc(sh,4,8);
}
#define init_3_4_9_14
void FASTCALL call_3_4_9_14(SH2_struct* sh)
{
  SH2addc(sh,4,9);
}
#define init_3_4_10_14
void FASTCALL call_3_4_10_14(SH2_struct* sh)
{
  SH2addc(sh,4,10);
}
#define init_3_4_11_14
void FASTCALL call_3_4_11_14(SH2_struct* sh)
{
  SH2addc(sh,4,11);
}
#define init_3_4_12_14
void FASTCALL call_3_4_12_14(SH2_struct* sh)
{
  SH2addc(sh,4,12);
}
#define init_3_4_13_14
void FASTCALL call_3_4_13_14(SH2_struct* sh)
{
  SH2addc(sh,4,13);
}
#define init_3_4_14_14
void FASTCALL call_3_4_14_14(SH2_struct* sh)
{
  SH2addc(sh,4,14);
}
#define init_3_4_15_14
void FASTCALL call_3_4_15_14(SH2_struct* sh)
{
  SH2addc(sh,4,15);
}
#define init_3_5_0_14
void FASTCALL call_3_5_0_14(SH2_struct* sh)
{
  SH2addc(sh,5,0);
}
#define init_3_5_1_14
void FASTCALL call_3_5_1_14(SH2_struct* sh)
{
  SH2addc(sh,5,1);
}
#define init_3_5_2_14
void FASTCALL call_3_5_2_14(SH2_struct* sh)
{
  SH2addc(sh,5,2);
}
#define init_3_5_3_14
void FASTCALL call_3_5_3_14(SH2_struct* sh)
{
  SH2addc(sh,5,3);
}
#define init_3_5_4_14
void FASTCALL call_3_5_4_14(SH2_struct* sh)
{
  SH2addc(sh,5,4);
}
#define init_3_5_5_14
void FASTCALL call_3_5_5_14(SH2_struct* sh)
{
  SH2addc(sh,5,5);
}
#define init_3_5_6_14
void FASTCALL call_3_5_6_14(SH2_struct* sh)
{
  SH2addc(sh,5,6);
}
#define init_3_5_7_14
void FASTCALL call_3_5_7_14(SH2_struct* sh)
{
  SH2addc(sh,5,7);
}
#define init_3_5_8_14
void FASTCALL call_3_5_8_14(SH2_struct* sh)
{
  SH2addc(sh,5,8);
}
#define init_3_5_9_14
void FASTCALL call_3_5_9_14(SH2_struct* sh)
{
  SH2addc(sh,5,9);
}
#define init_3_5_10_14
void FASTCALL call_3_5_10_14(SH2_struct* sh)
{
  SH2addc(sh,5,10);
}
#define init_3_5_11_14
void FASTCALL call_3_5_11_14(SH2_struct* sh)
{
  SH2addc(sh,5,11);
}
#define init_3_5_12_14
void FASTCALL call_3_5_12_14(SH2_struct* sh)
{
  SH2addc(sh,5,12);
}
#define init_3_5_13_14
void FASTCALL call_3_5_13_14(SH2_struct* sh)
{
  SH2addc(sh,5,13);
}
#define init_3_5_14_14
void FASTCALL call_3_5_14_14(SH2_struct* sh)
{
  SH2addc(sh,5,14);
}
#define init_3_5_15_14
void FASTCALL call_3_5_15_14(SH2_struct* sh)
{
  SH2addc(sh,5,15);
}
#define init_3_6_0_14
void FASTCALL call_3_6_0_14(SH2_struct* sh)
{
  SH2addc(sh,6,0);
}
#define init_3_6_1_14
void FASTCALL call_3_6_1_14(SH2_struct* sh)
{
  SH2addc(sh,6,1);
}
#define init_3_6_2_14
void FASTCALL call_3_6_2_14(SH2_struct* sh)
{
  SH2addc(sh,6,2);
}
#define init_3_6_3_14
void FASTCALL call_3_6_3_14(SH2_struct* sh)
{
  SH2addc(sh,6,3);
}
#define init_3_6_4_14
void FASTCALL call_3_6_4_14(SH2_struct* sh)
{
  SH2addc(sh,6,4);
}
#define init_3_6_5_14
void FASTCALL call_3_6_5_14(SH2_struct* sh)
{
  SH2addc(sh,6,5);
}
#define init_3_6_6_14
void FASTCALL call_3_6_6_14(SH2_struct* sh)
{
  SH2addc(sh,6,6);
}
#define init_3_6_7_14
void FASTCALL call_3_6_7_14(SH2_struct* sh)
{
  SH2addc(sh,6,7);
}
#define init_3_6_8_14
void FASTCALL call_3_6_8_14(SH2_struct* sh)
{
  SH2addc(sh,6,8);
}
#define init_3_6_9_14
void FASTCALL call_3_6_9_14(SH2_struct* sh)
{
  SH2addc(sh,6,9);
}
#define init_3_6_10_14
void FASTCALL call_3_6_10_14(SH2_struct* sh)
{
  SH2addc(sh,6,10);
}
#define init_3_6_11_14
void FASTCALL call_3_6_11_14(SH2_struct* sh)
{
  SH2addc(sh,6,11);
}
#define init_3_6_12_14
void FASTCALL call_3_6_12_14(SH2_struct* sh)
{
  SH2addc(sh,6,12);
}
#define init_3_6_13_14
void FASTCALL call_3_6_13_14(SH2_struct* sh)
{
  SH2addc(sh,6,13);
}
#define init_3_6_14_14
void FASTCALL call_3_6_14_14(SH2_struct* sh)
{
  SH2addc(sh,6,14);
}
#define init_3_6_15_14
void FASTCALL call_3_6_15_14(SH2_struct* sh)
{
  SH2addc(sh,6,15);
}
#define init_3_7_0_14
void FASTCALL call_3_7_0_14(SH2_struct* sh)
{
  SH2addc(sh,7,0);
}
#define init_3_7_1_14
void FASTCALL call_3_7_1_14(SH2_struct* sh)
{
  SH2addc(sh,7,1);
}
#define init_3_7_2_14
void FASTCALL call_3_7_2_14(SH2_struct* sh)
{
  SH2addc(sh,7,2);
}
#define init_3_7_3_14
void FASTCALL call_3_7_3_14(SH2_struct* sh)
{
  SH2addc(sh,7,3);
}
#define init_3_7_4_14
void FASTCALL call_3_7_4_14(SH2_struct* sh)
{
  SH2addc(sh,7,4);
}
#define init_3_7_5_14
void FASTCALL call_3_7_5_14(SH2_struct* sh)
{
  SH2addc(sh,7,5);
}
#define init_3_7_6_14
void FASTCALL call_3_7_6_14(SH2_struct* sh)
{
  SH2addc(sh,7,6);
}
#define init_3_7_7_14
void FASTCALL call_3_7_7_14(SH2_struct* sh)
{
  SH2addc(sh,7,7);
}
#define init_3_7_8_14
void FASTCALL call_3_7_8_14(SH2_struct* sh)
{
  SH2addc(sh,7,8);
}
#define init_3_7_9_14
void FASTCALL call_3_7_9_14(SH2_struct* sh)
{
  SH2addc(sh,7,9);
}
#define init_3_7_10_14
void FASTCALL call_3_7_10_14(SH2_struct* sh)
{
  SH2addc(sh,7,10);
}
#define init_3_7_11_14
void FASTCALL call_3_7_11_14(SH2_struct* sh)
{
  SH2addc(sh,7,11);
}
#define init_3_7_12_14
void FASTCALL call_3_7_12_14(SH2_struct* sh)
{
  SH2addc(sh,7,12);
}
#define init_3_7_13_14
void FASTCALL call_3_7_13_14(SH2_struct* sh)
{
  SH2addc(sh,7,13);
}
#define init_3_7_14_14
void FASTCALL call_3_7_14_14(SH2_struct* sh)
{
  SH2addc(sh,7,14);
}
#define init_3_7_15_14
void FASTCALL call_3_7_15_14(SH2_struct* sh)
{
  SH2addc(sh,7,15);
}
#define init_3_8_0_14
void FASTCALL call_3_8_0_14(SH2_struct* sh)
{
  SH2addc(sh,8,0);
}
#define init_3_8_1_14
void FASTCALL call_3_8_1_14(SH2_struct* sh)
{
  SH2addc(sh,8,1);
}
#define init_3_8_2_14
void FASTCALL call_3_8_2_14(SH2_struct* sh)
{
  SH2addc(sh,8,2);
}
#define init_3_8_3_14
void FASTCALL call_3_8_3_14(SH2_struct* sh)
{
  SH2addc(sh,8,3);
}
#define init_3_8_4_14
void FASTCALL call_3_8_4_14(SH2_struct* sh)
{
  SH2addc(sh,8,4);
}
#define init_3_8_5_14
void FASTCALL call_3_8_5_14(SH2_struct* sh)
{
  SH2addc(sh,8,5);
}
#define init_3_8_6_14
void FASTCALL call_3_8_6_14(SH2_struct* sh)
{
  SH2addc(sh,8,6);
}
#define init_3_8_7_14
void FASTCALL call_3_8_7_14(SH2_struct* sh)
{
  SH2addc(sh,8,7);
}
#define init_3_8_8_14
void FASTCALL call_3_8_8_14(SH2_struct* sh)
{
  SH2addc(sh,8,8);
}
#define init_3_8_9_14
void FASTCALL call_3_8_9_14(SH2_struct* sh)
{
  SH2addc(sh,8,9);
}
#define init_3_8_10_14
void FASTCALL call_3_8_10_14(SH2_struct* sh)
{
  SH2addc(sh,8,10);
}
#define init_3_8_11_14
void FASTCALL call_3_8_11_14(SH2_struct* sh)
{
  SH2addc(sh,8,11);
}
#define init_3_8_12_14
void FASTCALL call_3_8_12_14(SH2_struct* sh)
{
  SH2addc(sh,8,12);
}
#define init_3_8_13_14
void FASTCALL call_3_8_13_14(SH2_struct* sh)
{
  SH2addc(sh,8,13);
}
#define init_3_8_14_14
void FASTCALL call_3_8_14_14(SH2_struct* sh)
{
  SH2addc(sh,8,14);
}
#define init_3_8_15_14
void FASTCALL call_3_8_15_14(SH2_struct* sh)
{
  SH2addc(sh,8,15);
}
#define init_3_9_0_14
void FASTCALL call_3_9_0_14(SH2_struct* sh)
{
  SH2addc(sh,9,0);
}
#define init_3_9_1_14
void FASTCALL call_3_9_1_14(SH2_struct* sh)
{
  SH2addc(sh,9,1);
}
#define init_3_9_2_14
void FASTCALL call_3_9_2_14(SH2_struct* sh)
{
  SH2addc(sh,9,2);
}
#define init_3_9_3_14
void FASTCALL call_3_9_3_14(SH2_struct* sh)
{
  SH2addc(sh,9,3);
}
#define init_3_9_4_14
void FASTCALL call_3_9_4_14(SH2_struct* sh)
{
  SH2addc(sh,9,4);
}
#define init_3_9_5_14
void FASTCALL call_3_9_5_14(SH2_struct* sh)
{
  SH2addc(sh,9,5);
}
#define init_3_9_6_14
void FASTCALL call_3_9_6_14(SH2_struct* sh)
{
  SH2addc(sh,9,6);
}
#define init_3_9_7_14
void FASTCALL call_3_9_7_14(SH2_struct* sh)
{
  SH2addc(sh,9,7);
}
#define init_3_9_8_14
void FASTCALL call_3_9_8_14(SH2_struct* sh)
{
  SH2addc(sh,9,8);
}
#define init_3_9_9_14
void FASTCALL call_3_9_9_14(SH2_struct* sh)
{
  SH2addc(sh,9,9);
}
#define init_3_9_10_14
void FASTCALL call_3_9_10_14(SH2_struct* sh)
{
  SH2addc(sh,9,10);
}
#define init_3_9_11_14
void FASTCALL call_3_9_11_14(SH2_struct* sh)
{
  SH2addc(sh,9,11);
}
#define init_3_9_12_14
void FASTCALL call_3_9_12_14(SH2_struct* sh)
{
  SH2addc(sh,9,12);
}
#define init_3_9_13_14
void FASTCALL call_3_9_13_14(SH2_struct* sh)
{
  SH2addc(sh,9,13);
}
#define init_3_9_14_14
void FASTCALL call_3_9_14_14(SH2_struct* sh)
{
  SH2addc(sh,9,14);
}
#define init_3_9_15_14
void FASTCALL call_3_9_15_14(SH2_struct* sh)
{
  SH2addc(sh,9,15);
}
#define init_3_10_0_14
void FASTCALL call_3_10_0_14(SH2_struct* sh)
{
  SH2addc(sh,10,0);
}
#define init_3_10_1_14
void FASTCALL call_3_10_1_14(SH2_struct* sh)
{
  SH2addc(sh,10,1);
}
#define init_3_10_2_14
void FASTCALL call_3_10_2_14(SH2_struct* sh)
{
  SH2addc(sh,10,2);
}
#define init_3_10_3_14
void FASTCALL call_3_10_3_14(SH2_struct* sh)
{
  SH2addc(sh,10,3);
}
#define init_3_10_4_14
void FASTCALL call_3_10_4_14(SH2_struct* sh)
{
  SH2addc(sh,10,4);
}
#define init_3_10_5_14
void FASTCALL call_3_10_5_14(SH2_struct* sh)
{
  SH2addc(sh,10,5);
}
#define init_3_10_6_14
void FASTCALL call_3_10_6_14(SH2_struct* sh)
{
  SH2addc(sh,10,6);
}
#define init_3_10_7_14
void FASTCALL call_3_10_7_14(SH2_struct* sh)
{
  SH2addc(sh,10,7);
}
#define init_3_10_8_14
void FASTCALL call_3_10_8_14(SH2_struct* sh)
{
  SH2addc(sh,10,8);
}
#define init_3_10_9_14
void FASTCALL call_3_10_9_14(SH2_struct* sh)
{
  SH2addc(sh,10,9);
}
#define init_3_10_10_14
void FASTCALL call_3_10_10_14(SH2_struct* sh)
{
  SH2addc(sh,10,10);
}
#define init_3_10_11_14
void FASTCALL call_3_10_11_14(SH2_struct* sh)
{
  SH2addc(sh,10,11);
}
#define init_3_10_12_14
void FASTCALL call_3_10_12_14(SH2_struct* sh)
{
  SH2addc(sh,10,12);
}
#define init_3_10_13_14
void FASTCALL call_3_10_13_14(SH2_struct* sh)
{
  SH2addc(sh,10,13);
}
#define init_3_10_14_14
void FASTCALL call_3_10_14_14(SH2_struct* sh)
{
  SH2addc(sh,10,14);
}
#define init_3_10_15_14
void FASTCALL call_3_10_15_14(SH2_struct* sh)
{
  SH2addc(sh,10,15);
}
#define init_3_11_0_14
void FASTCALL call_3_11_0_14(SH2_struct* sh)
{
  SH2addc(sh,11,0);
}
#define init_3_11_1_14
void FASTCALL call_3_11_1_14(SH2_struct* sh)
{
  SH2addc(sh,11,1);
}
#define init_3_11_2_14
void FASTCALL call_3_11_2_14(SH2_struct* sh)
{
  SH2addc(sh,11,2);
}
#define init_3_11_3_14
void FASTCALL call_3_11_3_14(SH2_struct* sh)
{
  SH2addc(sh,11,3);
}
#define init_3_11_4_14
void FASTCALL call_3_11_4_14(SH2_struct* sh)
{
  SH2addc(sh,11,4);
}
#define init_3_11_5_14
void FASTCALL call_3_11_5_14(SH2_struct* sh)
{
  SH2addc(sh,11,5);
}
#define init_3_11_6_14
void FASTCALL call_3_11_6_14(SH2_struct* sh)
{
  SH2addc(sh,11,6);
}
#define init_3_11_7_14
void FASTCALL call_3_11_7_14(SH2_struct* sh)
{
  SH2addc(sh,11,7);
}
#define init_3_11_8_14
void FASTCALL call_3_11_8_14(SH2_struct* sh)
{
  SH2addc(sh,11,8);
}
#define init_3_11_9_14
void FASTCALL call_3_11_9_14(SH2_struct* sh)
{
  SH2addc(sh,11,9);
}
#define init_3_11_10_14
void FASTCALL call_3_11_10_14(SH2_struct* sh)
{
  SH2addc(sh,11,10);
}
#define init_3_11_11_14
void FASTCALL call_3_11_11_14(SH2_struct* sh)
{
  SH2addc(sh,11,11);
}
#define init_3_11_12_14
void FASTCALL call_3_11_12_14(SH2_struct* sh)
{
  SH2addc(sh,11,12);
}
#define init_3_11_13_14
void FASTCALL call_3_11_13_14(SH2_struct* sh)
{
  SH2addc(sh,11,13);
}
#define init_3_11_14_14
void FASTCALL call_3_11_14_14(SH2_struct* sh)
{
  SH2addc(sh,11,14);
}
#define init_3_11_15_14
void FASTCALL call_3_11_15_14(SH2_struct* sh)
{
  SH2addc(sh,11,15);
}
#define init_3_12_0_14
void FASTCALL call_3_12_0_14(SH2_struct* sh)
{
  SH2addc(sh,12,0);
}
#define init_3_12_1_14
void FASTCALL call_3_12_1_14(SH2_struct* sh)
{
  SH2addc(sh,12,1);
}
#define init_3_12_2_14
void FASTCALL call_3_12_2_14(SH2_struct* sh)
{
  SH2addc(sh,12,2);
}
#define init_3_12_3_14
void FASTCALL call_3_12_3_14(SH2_struct* sh)
{
  SH2addc(sh,12,3);
}
#define init_3_12_4_14
void FASTCALL call_3_12_4_14(SH2_struct* sh)
{
  SH2addc(sh,12,4);
}
#define init_3_12_5_14
void FASTCALL call_3_12_5_14(SH2_struct* sh)
{
  SH2addc(sh,12,5);
}
#define init_3_12_6_14
void FASTCALL call_3_12_6_14(SH2_struct* sh)
{
  SH2addc(sh,12,6);
}
#define init_3_12_7_14
void FASTCALL call_3_12_7_14(SH2_struct* sh)
{
  SH2addc(sh,12,7);
}
#define init_3_12_8_14
void FASTCALL call_3_12_8_14(SH2_struct* sh)
{
  SH2addc(sh,12,8);
}
#define init_3_12_9_14
void FASTCALL call_3_12_9_14(SH2_struct* sh)
{
  SH2addc(sh,12,9);
}
#define init_3_12_10_14
void FASTCALL call_3_12_10_14(SH2_struct* sh)
{
  SH2addc(sh,12,10);
}
#define init_3_12_11_14
void FASTCALL call_3_12_11_14(SH2_struct* sh)
{
  SH2addc(sh,12,11);
}
#define init_3_12_12_14
void FASTCALL call_3_12_12_14(SH2_struct* sh)
{
  SH2addc(sh,12,12);
}
#define init_3_12_13_14
void FASTCALL call_3_12_13_14(SH2_struct* sh)
{
  SH2addc(sh,12,13);
}
#define init_3_12_14_14
void FASTCALL call_3_12_14_14(SH2_struct* sh)
{
  SH2addc(sh,12,14);
}
#define init_3_12_15_14
void FASTCALL call_3_12_15_14(SH2_struct* sh)
{
  SH2addc(sh,12,15);
}
#define init_3_13_0_14
void FASTCALL call_3_13_0_14(SH2_struct* sh)
{
  SH2addc(sh,13,0);
}
#define init_3_13_1_14
void FASTCALL call_3_13_1_14(SH2_struct* sh)
{
  SH2addc(sh,13,1);
}
#define init_3_13_2_14
void FASTCALL call_3_13_2_14(SH2_struct* sh)
{
  SH2addc(sh,13,2);
}
#define init_3_13_3_14
void FASTCALL call_3_13_3_14(SH2_struct* sh)
{
  SH2addc(sh,13,3);
}
#define init_3_13_4_14
void FASTCALL call_3_13_4_14(SH2_struct* sh)
{
  SH2addc(sh,13,4);
}
#define init_3_13_5_14
void FASTCALL call_3_13_5_14(SH2_struct* sh)
{
  SH2addc(sh,13,5);
}
#define init_3_13_6_14
void FASTCALL call_3_13_6_14(SH2_struct* sh)
{
  SH2addc(sh,13,6);
}
#define init_3_13_7_14
void FASTCALL call_3_13_7_14(SH2_struct* sh)
{
  SH2addc(sh,13,7);
}
#define init_3_13_8_14
void FASTCALL call_3_13_8_14(SH2_struct* sh)
{
  SH2addc(sh,13,8);
}
#define init_3_13_9_14
void FASTCALL call_3_13_9_14(SH2_struct* sh)
{
  SH2addc(sh,13,9);
}
#define init_3_13_10_14
void FASTCALL call_3_13_10_14(SH2_struct* sh)
{
  SH2addc(sh,13,10);
}
#define init_3_13_11_14
void FASTCALL call_3_13_11_14(SH2_struct* sh)
{
  SH2addc(sh,13,11);
}
#define init_3_13_12_14
void FASTCALL call_3_13_12_14(SH2_struct* sh)
{
  SH2addc(sh,13,12);
}
#define init_3_13_13_14
void FASTCALL call_3_13_13_14(SH2_struct* sh)
{
  SH2addc(sh,13,13);
}
#define init_3_13_14_14
void FASTCALL call_3_13_14_14(SH2_struct* sh)
{
  SH2addc(sh,13,14);
}
#define init_3_13_15_14
void FASTCALL call_3_13_15_14(SH2_struct* sh)
{
  SH2addc(sh,13,15);
}
#define init_3_14_0_14
void FASTCALL call_3_14_0_14(SH2_struct* sh)
{
  SH2addc(sh,14,0);
}
#define init_3_14_1_14
void FASTCALL call_3_14_1_14(SH2_struct* sh)
{
  SH2addc(sh,14,1);
}
#define init_3_14_2_14
void FASTCALL call_3_14_2_14(SH2_struct* sh)
{
  SH2addc(sh,14,2);
}
#define init_3_14_3_14
void FASTCALL call_3_14_3_14(SH2_struct* sh)
{
  SH2addc(sh,14,3);
}
#define init_3_14_4_14
void FASTCALL call_3_14_4_14(SH2_struct* sh)
{
  SH2addc(sh,14,4);
}
#define init_3_14_5_14
void FASTCALL call_3_14_5_14(SH2_struct* sh)
{
  SH2addc(sh,14,5);
}
#define init_3_14_6_14
void FASTCALL call_3_14_6_14(SH2_struct* sh)
{
  SH2addc(sh,14,6);
}
#define init_3_14_7_14
void FASTCALL call_3_14_7_14(SH2_struct* sh)
{
  SH2addc(sh,14,7);
}
#define init_3_14_8_14
void FASTCALL call_3_14_8_14(SH2_struct* sh)
{
  SH2addc(sh,14,8);
}
#define init_3_14_9_14
void FASTCALL call_3_14_9_14(SH2_struct* sh)
{
  SH2addc(sh,14,9);
}
#define init_3_14_10_14
void FASTCALL call_3_14_10_14(SH2_struct* sh)
{
  SH2addc(sh,14,10);
}
#define init_3_14_11_14
void FASTCALL call_3_14_11_14(SH2_struct* sh)
{
  SH2addc(sh,14,11);
}
#define init_3_14_12_14
void FASTCALL call_3_14_12_14(SH2_struct* sh)
{
  SH2addc(sh,14,12);
}
#define init_3_14_13_14
void FASTCALL call_3_14_13_14(SH2_struct* sh)
{
  SH2addc(sh,14,13);
}
#define init_3_14_14_14
void FASTCALL call_3_14_14_14(SH2_struct* sh)
{
  SH2addc(sh,14,14);
}
#define init_3_14_15_14
void FASTCALL call_3_14_15_14(SH2_struct* sh)
{
  SH2addc(sh,14,15);
}
#define init_3_15_0_14
void FASTCALL call_3_15_0_14(SH2_struct* sh)
{
  SH2addc(sh,15,0);
}
#define init_3_15_1_14
void FASTCALL call_3_15_1_14(SH2_struct* sh)
{
  SH2addc(sh,15,1);
}
#define init_3_15_2_14
void FASTCALL call_3_15_2_14(SH2_struct* sh)
{
  SH2addc(sh,15,2);
}
#define init_3_15_3_14
void FASTCALL call_3_15_3_14(SH2_struct* sh)
{
  SH2addc(sh,15,3);
}
#define init_3_15_4_14
void FASTCALL call_3_15_4_14(SH2_struct* sh)
{
  SH2addc(sh,15,4);
}
#define init_3_15_5_14
void FASTCALL call_3_15_5_14(SH2_struct* sh)
{
  SH2addc(sh,15,5);
}
#define init_3_15_6_14
void FASTCALL call_3_15_6_14(SH2_struct* sh)
{
  SH2addc(sh,15,6);
}
#define init_3_15_7_14
void FASTCALL call_3_15_7_14(SH2_struct* sh)
{
  SH2addc(sh,15,7);
}
#define init_3_15_8_14
void FASTCALL call_3_15_8_14(SH2_struct* sh)
{
  SH2addc(sh,15,8);
}
#define init_3_15_9_14
void FASTCALL call_3_15_9_14(SH2_struct* sh)
{
  SH2addc(sh,15,9);
}
#define init_3_15_10_14
void FASTCALL call_3_15_10_14(SH2_struct* sh)
{
  SH2addc(sh,15,10);
}
#define init_3_15_11_14
void FASTCALL call_3_15_11_14(SH2_struct* sh)
{
  SH2addc(sh,15,11);
}
#define init_3_15_12_14
void FASTCALL call_3_15_12_14(SH2_struct* sh)
{
  SH2addc(sh,15,12);
}
#define init_3_15_13_14
void FASTCALL call_3_15_13_14(SH2_struct* sh)
{
  SH2addc(sh,15,13);
}
#define init_3_15_14_14
void FASTCALL call_3_15_14_14(SH2_struct* sh)
{
  SH2addc(sh,15,14);
}
#define init_3_15_15_14
void FASTCALL call_3_15_15_14(SH2_struct* sh)
{
  SH2addc(sh,15,15);
}
#define init_7_0_0_0
void FASTCALL call_7_0_0_0(SH2_struct* sh)
{
  SH2addi(sh, 0, (0<<4)+0);
}
#define init_7_0_0_1
void FASTCALL call_7_0_0_1(SH2_struct* sh)
{
  SH2addi(sh, 0, (0<<4)+1);
}
#define init_7_0_0_2
void FASTCALL call_7_0_0_2(SH2_struct* sh)
{
  SH2addi(sh, 0, (0<<4)+2);
}
#define init_7_0_0_3
void FASTCALL call_7_0_0_3(SH2_struct* sh)
{
  SH2addi(sh, 0, (0<<4)+3);
}
#define init_7_0_0_4
void FASTCALL call_7_0_0_4(SH2_struct* sh)
{
  SH2addi(sh, 0, (0<<4)+4);
}
#define init_7_0_0_5
void FASTCALL call_7_0_0_5(SH2_struct* sh)
{
  SH2addi(sh, 0, (0<<4)+5);
}
#define init_7_0_0_6
void FASTCALL call_7_0_0_6(SH2_struct* sh)
{
  SH2addi(sh, 0, (0<<4)+6);
}
#define init_7_0_0_7
void FASTCALL call_7_0_0_7(SH2_struct* sh)
{
  SH2addi(sh, 0, (0<<4)+7);
}
#define init_7_0_0_8
void FASTCALL call_7_0_0_8(SH2_struct* sh)
{
  SH2addi(sh, 0, (0<<4)+8);
}
#define init_7_0_0_9
void FASTCALL call_7_0_0_9(SH2_struct* sh)
{
  SH2addi(sh, 0, (0<<4)+9);
}
#define init_7_0_0_10
void FASTCALL call_7_0_0_10(SH2_struct* sh)
{
  SH2addi(sh, 0, (0<<4)+10);
}
#define init_7_0_0_11
void FASTCALL call_7_0_0_11(SH2_struct* sh)
{
  SH2addi(sh, 0, (0<<4)+11);
}
#define init_7_0_0_12
void FASTCALL call_7_0_0_12(SH2_struct* sh)
{
  SH2addi(sh, 0, (0<<4)+12);
}
#define init_7_0_0_13
void FASTCALL call_7_0_0_13(SH2_struct* sh)
{
  SH2addi(sh, 0, (0<<4)+13);
}
#define init_7_0_0_14
void FASTCALL call_7_0_0_14(SH2_struct* sh)
{
  SH2addi(sh, 0, (0<<4)+14);
}
#define init_7_0_0_15
void FASTCALL call_7_0_0_15(SH2_struct* sh)
{
  SH2addi(sh, 0, (0<<4)+15);
}
#define init_7_0_1_0
void FASTCALL call_7_0_1_0(SH2_struct* sh)
{
  SH2addi(sh, 0, (1<<4)+0);
}
#define init_7_0_1_1
void FASTCALL call_7_0_1_1(SH2_struct* sh)
{
  SH2addi(sh, 0, (1<<4)+1);
}
#define init_7_0_1_2
void FASTCALL call_7_0_1_2(SH2_struct* sh)
{
  SH2addi(sh, 0, (1<<4)+2);
}
#define init_7_0_1_3
void FASTCALL call_7_0_1_3(SH2_struct* sh)
{
  SH2addi(sh, 0, (1<<4)+3);
}
#define init_7_0_1_4
void FASTCALL call_7_0_1_4(SH2_struct* sh)
{
  SH2addi(sh, 0, (1<<4)+4);
}
#define init_7_0_1_5
void FASTCALL call_7_0_1_5(SH2_struct* sh)
{
  SH2addi(sh, 0, (1<<4)+5);
}
#define init_7_0_1_6
void FASTCALL call_7_0_1_6(SH2_struct* sh)
{
  SH2addi(sh, 0, (1<<4)+6);
}
#define init_7_0_1_7
void FASTCALL call_7_0_1_7(SH2_struct* sh)
{
  SH2addi(sh, 0, (1<<4)+7);
}
#define init_7_0_1_8
void FASTCALL call_7_0_1_8(SH2_struct* sh)
{
  SH2addi(sh, 0, (1<<4)+8);
}
#define init_7_0_1_9
void FASTCALL call_7_0_1_9(SH2_struct* sh)
{
  SH2addi(sh, 0, (1<<4)+9);
}
#define init_7_0_1_10
void FASTCALL call_7_0_1_10(SH2_struct* sh)
{
  SH2addi(sh, 0, (1<<4)+10);
}
#define init_7_0_1_11
void FASTCALL call_7_0_1_11(SH2_struct* sh)
{
  SH2addi(sh, 0, (1<<4)+11);
}
#define init_7_0_1_12
void FASTCALL call_7_0_1_12(SH2_struct* sh)
{
  SH2addi(sh, 0, (1<<4)+12);
}
#define init_7_0_1_13
void FASTCALL call_7_0_1_13(SH2_struct* sh)
{
  SH2addi(sh, 0, (1<<4)+13);
}
#define init_7_0_1_14
void FASTCALL call_7_0_1_14(SH2_struct* sh)
{
  SH2addi(sh, 0, (1<<4)+14);
}
#define init_7_0_1_15
void FASTCALL call_7_0_1_15(SH2_struct* sh)
{
  SH2addi(sh, 0, (1<<4)+15);
}
#define init_7_0_2_0
void FASTCALL call_7_0_2_0(SH2_struct* sh)
{
  SH2addi(sh, 0, (2<<4)+0);
}
#define init_7_0_2_1
void FASTCALL call_7_0_2_1(SH2_struct* sh)
{
  SH2addi(sh, 0, (2<<4)+1);
}
#define init_7_0_2_2
void FASTCALL call_7_0_2_2(SH2_struct* sh)
{
  SH2addi(sh, 0, (2<<4)+2);
}
#define init_7_0_2_3
void FASTCALL call_7_0_2_3(SH2_struct* sh)
{
  SH2addi(sh, 0, (2<<4)+3);
}
#define init_7_0_2_4
void FASTCALL call_7_0_2_4(SH2_struct* sh)
{
  SH2addi(sh, 0, (2<<4)+4);
}
#define init_7_0_2_5
void FASTCALL call_7_0_2_5(SH2_struct* sh)
{
  SH2addi(sh, 0, (2<<4)+5);
}
#define init_7_0_2_6
void FASTCALL call_7_0_2_6(SH2_struct* sh)
{
  SH2addi(sh, 0, (2<<4)+6);
}
#define init_7_0_2_7
void FASTCALL call_7_0_2_7(SH2_struct* sh)
{
  SH2addi(sh, 0, (2<<4)+7);
}
#define init_7_0_2_8
void FASTCALL call_7_0_2_8(SH2_struct* sh)
{
  SH2addi(sh, 0, (2<<4)+8);
}
#define init_7_0_2_9
void FASTCALL call_7_0_2_9(SH2_struct* sh)
{
  SH2addi(sh, 0, (2<<4)+9);
}
#define init_7_0_2_10
void FASTCALL call_7_0_2_10(SH2_struct* sh)
{
  SH2addi(sh, 0, (2<<4)+10);
}
#define init_7_0_2_11
void FASTCALL call_7_0_2_11(SH2_struct* sh)
{
  SH2addi(sh, 0, (2<<4)+11);
}
#define init_7_0_2_12
void FASTCALL call_7_0_2_12(SH2_struct* sh)
{
  SH2addi(sh, 0, (2<<4)+12);
}
#define init_7_0_2_13
void FASTCALL call_7_0_2_13(SH2_struct* sh)
{
  SH2addi(sh, 0, (2<<4)+13);
}
#define init_7_0_2_14
void FASTCALL call_7_0_2_14(SH2_struct* sh)
{
  SH2addi(sh, 0, (2<<4)+14);
}
#define init_7_0_2_15
void FASTCALL call_7_0_2_15(SH2_struct* sh)
{
  SH2addi(sh, 0, (2<<4)+15);
}
#define init_7_0_3_0
void FASTCALL call_7_0_3_0(SH2_struct* sh)
{
  SH2addi(sh, 0, (3<<4)+0);
}
#define init_7_0_3_1
void FASTCALL call_7_0_3_1(SH2_struct* sh)
{
  SH2addi(sh, 0, (3<<4)+1);
}
#define init_7_0_3_2
void FASTCALL call_7_0_3_2(SH2_struct* sh)
{
  SH2addi(sh, 0, (3<<4)+2);
}
#define init_7_0_3_3
void FASTCALL call_7_0_3_3(SH2_struct* sh)
{
  SH2addi(sh, 0, (3<<4)+3);
}
#define init_7_0_3_4
void FASTCALL call_7_0_3_4(SH2_struct* sh)
{
  SH2addi(sh, 0, (3<<4)+4);
}
#define init_7_0_3_5
void FASTCALL call_7_0_3_5(SH2_struct* sh)
{
  SH2addi(sh, 0, (3<<4)+5);
}
#define init_7_0_3_6
void FASTCALL call_7_0_3_6(SH2_struct* sh)
{
  SH2addi(sh, 0, (3<<4)+6);
}
#define init_7_0_3_7
void FASTCALL call_7_0_3_7(SH2_struct* sh)
{
  SH2addi(sh, 0, (3<<4)+7);
}
#define init_7_0_3_8
void FASTCALL call_7_0_3_8(SH2_struct* sh)
{
  SH2addi(sh, 0, (3<<4)+8);
}
#define init_7_0_3_9
void FASTCALL call_7_0_3_9(SH2_struct* sh)
{
  SH2addi(sh, 0, (3<<4)+9);
}
#define init_7_0_3_10
void FASTCALL call_7_0_3_10(SH2_struct* sh)
{
  SH2addi(sh, 0, (3<<4)+10);
}
#define init_7_0_3_11
void FASTCALL call_7_0_3_11(SH2_struct* sh)
{
  SH2addi(sh, 0, (3<<4)+11);
}
#define init_7_0_3_12
void FASTCALL call_7_0_3_12(SH2_struct* sh)
{
  SH2addi(sh, 0, (3<<4)+12);
}
#define init_7_0_3_13
void FASTCALL call_7_0_3_13(SH2_struct* sh)
{
  SH2addi(sh, 0, (3<<4)+13);
}
#define init_7_0_3_14
void FASTCALL call_7_0_3_14(SH2_struct* sh)
{
  SH2addi(sh, 0, (3<<4)+14);
}
#define init_7_0_3_15
void FASTCALL call_7_0_3_15(SH2_struct* sh)
{
  SH2addi(sh, 0, (3<<4)+15);
}
#define init_7_0_4_0
void FASTCALL call_7_0_4_0(SH2_struct* sh)
{
  SH2addi(sh, 0, (4<<4)+0);
}
#define init_7_0_4_1
void FASTCALL call_7_0_4_1(SH2_struct* sh)
{
  SH2addi(sh, 0, (4<<4)+1);
}
#define init_7_0_4_2
void FASTCALL call_7_0_4_2(SH2_struct* sh)
{
  SH2addi(sh, 0, (4<<4)+2);
}
#define init_7_0_4_3
void FASTCALL call_7_0_4_3(SH2_struct* sh)
{
  SH2addi(sh, 0, (4<<4)+3);
}
#define init_7_0_4_4
void FASTCALL call_7_0_4_4(SH2_struct* sh)
{
  SH2addi(sh, 0, (4<<4)+4);
}
#define init_7_0_4_5
void FASTCALL call_7_0_4_5(SH2_struct* sh)
{
  SH2addi(sh, 0, (4<<4)+5);
}
#define init_7_0_4_6
void FASTCALL call_7_0_4_6(SH2_struct* sh)
{
  SH2addi(sh, 0, (4<<4)+6);
}
#define init_7_0_4_7
void FASTCALL call_7_0_4_7(SH2_struct* sh)
{
  SH2addi(sh, 0, (4<<4)+7);
}
#define init_7_0_4_8
void FASTCALL call_7_0_4_8(SH2_struct* sh)
{
  SH2addi(sh, 0, (4<<4)+8);
}
#define init_7_0_4_9
void FASTCALL call_7_0_4_9(SH2_struct* sh)
{
  SH2addi(sh, 0, (4<<4)+9);
}
#define init_7_0_4_10
void FASTCALL call_7_0_4_10(SH2_struct* sh)
{
  SH2addi(sh, 0, (4<<4)+10);
}
#define init_7_0_4_11
void FASTCALL call_7_0_4_11(SH2_struct* sh)
{
  SH2addi(sh, 0, (4<<4)+11);
}
#define init_7_0_4_12
void FASTCALL call_7_0_4_12(SH2_struct* sh)
{
  SH2addi(sh, 0, (4<<4)+12);
}
#define init_7_0_4_13
void FASTCALL call_7_0_4_13(SH2_struct* sh)
{
  SH2addi(sh, 0, (4<<4)+13);
}
#define init_7_0_4_14
void FASTCALL call_7_0_4_14(SH2_struct* sh)
{
  SH2addi(sh, 0, (4<<4)+14);
}
#define init_7_0_4_15
void FASTCALL call_7_0_4_15(SH2_struct* sh)
{
  SH2addi(sh, 0, (4<<4)+15);
}
#define init_7_0_5_0
void FASTCALL call_7_0_5_0(SH2_struct* sh)
{
  SH2addi(sh, 0, (5<<4)+0);
}
#define init_7_0_5_1
void FASTCALL call_7_0_5_1(SH2_struct* sh)
{
  SH2addi(sh, 0, (5<<4)+1);
}
#define init_7_0_5_2
void FASTCALL call_7_0_5_2(SH2_struct* sh)
{
  SH2addi(sh, 0, (5<<4)+2);
}
#define init_7_0_5_3
void FASTCALL call_7_0_5_3(SH2_struct* sh)
{
  SH2addi(sh, 0, (5<<4)+3);
}
#define init_7_0_5_4
void FASTCALL call_7_0_5_4(SH2_struct* sh)
{
  SH2addi(sh, 0, (5<<4)+4);
}
#define init_7_0_5_5
void FASTCALL call_7_0_5_5(SH2_struct* sh)
{
  SH2addi(sh, 0, (5<<4)+5);
}
#define init_7_0_5_6
void FASTCALL call_7_0_5_6(SH2_struct* sh)
{
  SH2addi(sh, 0, (5<<4)+6);
}
#define init_7_0_5_7
void FASTCALL call_7_0_5_7(SH2_struct* sh)
{
  SH2addi(sh, 0, (5<<4)+7);
}
#define init_7_0_5_8
void FASTCALL call_7_0_5_8(SH2_struct* sh)
{
  SH2addi(sh, 0, (5<<4)+8);
}
#define init_7_0_5_9
void FASTCALL call_7_0_5_9(SH2_struct* sh)
{
  SH2addi(sh, 0, (5<<4)+9);
}
#define init_7_0_5_10
void FASTCALL call_7_0_5_10(SH2_struct* sh)
{
  SH2addi(sh, 0, (5<<4)+10);
}
#define init_7_0_5_11
void FASTCALL call_7_0_5_11(SH2_struct* sh)
{
  SH2addi(sh, 0, (5<<4)+11);
}
#define init_7_0_5_12
void FASTCALL call_7_0_5_12(SH2_struct* sh)
{
  SH2addi(sh, 0, (5<<4)+12);
}
#define init_7_0_5_13
void FASTCALL call_7_0_5_13(SH2_struct* sh)
{
  SH2addi(sh, 0, (5<<4)+13);
}
#define init_7_0_5_14
void FASTCALL call_7_0_5_14(SH2_struct* sh)
{
  SH2addi(sh, 0, (5<<4)+14);
}
#define init_7_0_5_15
void FASTCALL call_7_0_5_15(SH2_struct* sh)
{
  SH2addi(sh, 0, (5<<4)+15);
}
#define init_7_0_6_0
void FASTCALL call_7_0_6_0(SH2_struct* sh)
{
  SH2addi(sh, 0, (6<<4)+0);
}
#define init_7_0_6_1
void FASTCALL call_7_0_6_1(SH2_struct* sh)
{
  SH2addi(sh, 0, (6<<4)+1);
}
#define init_7_0_6_2
void FASTCALL call_7_0_6_2(SH2_struct* sh)
{
  SH2addi(sh, 0, (6<<4)+2);
}
#define init_7_0_6_3
void FASTCALL call_7_0_6_3(SH2_struct* sh)
{
  SH2addi(sh, 0, (6<<4)+3);
}
#define init_7_0_6_4
void FASTCALL call_7_0_6_4(SH2_struct* sh)
{
  SH2addi(sh, 0, (6<<4)+4);
}
#define init_7_0_6_5
void FASTCALL call_7_0_6_5(SH2_struct* sh)
{
  SH2addi(sh, 0, (6<<4)+5);
}
#define init_7_0_6_6
void FASTCALL call_7_0_6_6(SH2_struct* sh)
{
  SH2addi(sh, 0, (6<<4)+6);
}
#define init_7_0_6_7
void FASTCALL call_7_0_6_7(SH2_struct* sh)
{
  SH2addi(sh, 0, (6<<4)+7);
}
#define init_7_0_6_8
void FASTCALL call_7_0_6_8(SH2_struct* sh)
{
  SH2addi(sh, 0, (6<<4)+8);
}
#define init_7_0_6_9
void FASTCALL call_7_0_6_9(SH2_struct* sh)
{
  SH2addi(sh, 0, (6<<4)+9);
}
#define init_7_0_6_10
void FASTCALL call_7_0_6_10(SH2_struct* sh)
{
  SH2addi(sh, 0, (6<<4)+10);
}
#define init_7_0_6_11
void FASTCALL call_7_0_6_11(SH2_struct* sh)
{
  SH2addi(sh, 0, (6<<4)+11);
}
#define init_7_0_6_12
void FASTCALL call_7_0_6_12(SH2_struct* sh)
{
  SH2addi(sh, 0, (6<<4)+12);
}
#define init_7_0_6_13
void FASTCALL call_7_0_6_13(SH2_struct* sh)
{
  SH2addi(sh, 0, (6<<4)+13);
}
#define init_7_0_6_14
void FASTCALL call_7_0_6_14(SH2_struct* sh)
{
  SH2addi(sh, 0, (6<<4)+14);
}
#define init_7_0_6_15
void FASTCALL call_7_0_6_15(SH2_struct* sh)
{
  SH2addi(sh, 0, (6<<4)+15);
}
#define init_7_0_7_0
void FASTCALL call_7_0_7_0(SH2_struct* sh)
{
  SH2addi(sh, 0, (7<<4)+0);
}
#define init_7_0_7_1
void FASTCALL call_7_0_7_1(SH2_struct* sh)
{
  SH2addi(sh, 0, (7<<4)+1);
}
#define init_7_0_7_2
void FASTCALL call_7_0_7_2(SH2_struct* sh)
{
  SH2addi(sh, 0, (7<<4)+2);
}
#define init_7_0_7_3
void FASTCALL call_7_0_7_3(SH2_struct* sh)
{
  SH2addi(sh, 0, (7<<4)+3);
}
#define init_7_0_7_4
void FASTCALL call_7_0_7_4(SH2_struct* sh)
{
  SH2addi(sh, 0, (7<<4)+4);
}
#define init_7_0_7_5
void FASTCALL call_7_0_7_5(SH2_struct* sh)
{
  SH2addi(sh, 0, (7<<4)+5);
}
#define init_7_0_7_6
void FASTCALL call_7_0_7_6(SH2_struct* sh)
{
  SH2addi(sh, 0, (7<<4)+6);
}
#define init_7_0_7_7
void FASTCALL call_7_0_7_7(SH2_struct* sh)
{
  SH2addi(sh, 0, (7<<4)+7);
}
#define init_7_0_7_8
void FASTCALL call_7_0_7_8(SH2_struct* sh)
{
  SH2addi(sh, 0, (7<<4)+8);
}
#define init_7_0_7_9
void FASTCALL call_7_0_7_9(SH2_struct* sh)
{
  SH2addi(sh, 0, (7<<4)+9);
}
#define init_7_0_7_10
void FASTCALL call_7_0_7_10(SH2_struct* sh)
{
  SH2addi(sh, 0, (7<<4)+10);
}
#define init_7_0_7_11
void FASTCALL call_7_0_7_11(SH2_struct* sh)
{
  SH2addi(sh, 0, (7<<4)+11);
}
#define init_7_0_7_12
void FASTCALL call_7_0_7_12(SH2_struct* sh)
{
  SH2addi(sh, 0, (7<<4)+12);
}
#define init_7_0_7_13
void FASTCALL call_7_0_7_13(SH2_struct* sh)
{
  SH2addi(sh, 0, (7<<4)+13);
}
#define init_7_0_7_14
void FASTCALL call_7_0_7_14(SH2_struct* sh)
{
  SH2addi(sh, 0, (7<<4)+14);
}
#define init_7_0_7_15
void FASTCALL call_7_0_7_15(SH2_struct* sh)
{
  SH2addi(sh, 0, (7<<4)+15);
}
#define init_7_0_8_0
void FASTCALL call_7_0_8_0(SH2_struct* sh)
{
  SH2addi(sh, 0, (8<<4)+0);
}
#define init_7_0_8_1
void FASTCALL call_7_0_8_1(SH2_struct* sh)
{
  SH2addi(sh, 0, (8<<4)+1);
}
#define init_7_0_8_2
void FASTCALL call_7_0_8_2(SH2_struct* sh)
{
  SH2addi(sh, 0, (8<<4)+2);
}
#define init_7_0_8_3
void FASTCALL call_7_0_8_3(SH2_struct* sh)
{
  SH2addi(sh, 0, (8<<4)+3);
}
#define init_7_0_8_4
void FASTCALL call_7_0_8_4(SH2_struct* sh)
{
  SH2addi(sh, 0, (8<<4)+4);
}
#define init_7_0_8_5
void FASTCALL call_7_0_8_5(SH2_struct* sh)
{
  SH2addi(sh, 0, (8<<4)+5);
}
#define init_7_0_8_6
void FASTCALL call_7_0_8_6(SH2_struct* sh)
{
  SH2addi(sh, 0, (8<<4)+6);
}
#define init_7_0_8_7
void FASTCALL call_7_0_8_7(SH2_struct* sh)
{
  SH2addi(sh, 0, (8<<4)+7);
}
#define init_7_0_8_8
void FASTCALL call_7_0_8_8(SH2_struct* sh)
{
  SH2addi(sh, 0, (8<<4)+8);
}
#define init_7_0_8_9
void FASTCALL call_7_0_8_9(SH2_struct* sh)
{
  SH2addi(sh, 0, (8<<4)+9);
}
#define init_7_0_8_10
void FASTCALL call_7_0_8_10(SH2_struct* sh)
{
  SH2addi(sh, 0, (8<<4)+10);
}
#define init_7_0_8_11
void FASTCALL call_7_0_8_11(SH2_struct* sh)
{
  SH2addi(sh, 0, (8<<4)+11);
}
#define init_7_0_8_12
void FASTCALL call_7_0_8_12(SH2_struct* sh)
{
  SH2addi(sh, 0, (8<<4)+12);
}
#define init_7_0_8_13
void FASTCALL call_7_0_8_13(SH2_struct* sh)
{
  SH2addi(sh, 0, (8<<4)+13);
}
#define init_7_0_8_14
void FASTCALL call_7_0_8_14(SH2_struct* sh)
{
  SH2addi(sh, 0, (8<<4)+14);
}
#define init_7_0_8_15
void FASTCALL call_7_0_8_15(SH2_struct* sh)
{
  SH2addi(sh, 0, (8<<4)+15);
}
#define init_7_0_9_0
void FASTCALL call_7_0_9_0(SH2_struct* sh)
{
  SH2addi(sh, 0, (9<<4)+0);
}
#define init_7_0_9_1
void FASTCALL call_7_0_9_1(SH2_struct* sh)
{
  SH2addi(sh, 0, (9<<4)+1);
}
#define init_7_0_9_2
void FASTCALL call_7_0_9_2(SH2_struct* sh)
{
  SH2addi(sh, 0, (9<<4)+2);
}
#define init_7_0_9_3
void FASTCALL call_7_0_9_3(SH2_struct* sh)
{
  SH2addi(sh, 0, (9<<4)+3);
}
#define init_7_0_9_4
void FASTCALL call_7_0_9_4(SH2_struct* sh)
{
  SH2addi(sh, 0, (9<<4)+4);
}
#define init_7_0_9_5
void FASTCALL call_7_0_9_5(SH2_struct* sh)
{
  SH2addi(sh, 0, (9<<4)+5);
}
#define init_7_0_9_6
void FASTCALL call_7_0_9_6(SH2_struct* sh)
{
  SH2addi(sh, 0, (9<<4)+6);
}
#define init_7_0_9_7
void FASTCALL call_7_0_9_7(SH2_struct* sh)
{
  SH2addi(sh, 0, (9<<4)+7);
}
#define init_7_0_9_8
void FASTCALL call_7_0_9_8(SH2_struct* sh)
{
  SH2addi(sh, 0, (9<<4)+8);
}
#define init_7_0_9_9
void FASTCALL call_7_0_9_9(SH2_struct* sh)
{
  SH2addi(sh, 0, (9<<4)+9);
}
#define init_7_0_9_10
void FASTCALL call_7_0_9_10(SH2_struct* sh)
{
  SH2addi(sh, 0, (9<<4)+10);
}
#define init_7_0_9_11
void FASTCALL call_7_0_9_11(SH2_struct* sh)
{
  SH2addi(sh, 0, (9<<4)+11);
}
#define init_7_0_9_12
void FASTCALL call_7_0_9_12(SH2_struct* sh)
{
  SH2addi(sh, 0, (9<<4)+12);
}
#define init_7_0_9_13
void FASTCALL call_7_0_9_13(SH2_struct* sh)
{
  SH2addi(sh, 0, (9<<4)+13);
}
#define init_7_0_9_14
void FASTCALL call_7_0_9_14(SH2_struct* sh)
{
  SH2addi(sh, 0, (9<<4)+14);
}
#define init_7_0_9_15
void FASTCALL call_7_0_9_15(SH2_struct* sh)
{
  SH2addi(sh, 0, (9<<4)+15);
}
#define init_7_0_10_0
void FASTCALL call_7_0_10_0(SH2_struct* sh)
{
  SH2addi(sh, 0, (10<<4)+0);
}
#define init_7_0_10_1
void FASTCALL call_7_0_10_1(SH2_struct* sh)
{
  SH2addi(sh, 0, (10<<4)+1);
}
#define init_7_0_10_2
void FASTCALL call_7_0_10_2(SH2_struct* sh)
{
  SH2addi(sh, 0, (10<<4)+2);
}
#define init_7_0_10_3
void FASTCALL call_7_0_10_3(SH2_struct* sh)
{
  SH2addi(sh, 0, (10<<4)+3);
}
#define init_7_0_10_4
void FASTCALL call_7_0_10_4(SH2_struct* sh)
{
  SH2addi(sh, 0, (10<<4)+4);
}
#define init_7_0_10_5
void FASTCALL call_7_0_10_5(SH2_struct* sh)
{
  SH2addi(sh, 0, (10<<4)+5);
}
#define init_7_0_10_6
void FASTCALL call_7_0_10_6(SH2_struct* sh)
{
  SH2addi(sh, 0, (10<<4)+6);
}
#define init_7_0_10_7
void FASTCALL call_7_0_10_7(SH2_struct* sh)
{
  SH2addi(sh, 0, (10<<4)+7);
}
#define init_7_0_10_8
void FASTCALL call_7_0_10_8(SH2_struct* sh)
{
  SH2addi(sh, 0, (10<<4)+8);
}
#define init_7_0_10_9
void FASTCALL call_7_0_10_9(SH2_struct* sh)
{
  SH2addi(sh, 0, (10<<4)+9);
}
#define init_7_0_10_10
void FASTCALL call_7_0_10_10(SH2_struct* sh)
{
  SH2addi(sh, 0, (10<<4)+10);
}
#define init_7_0_10_11
void FASTCALL call_7_0_10_11(SH2_struct* sh)
{
  SH2addi(sh, 0, (10<<4)+11);
}
#define init_7_0_10_12
void FASTCALL call_7_0_10_12(SH2_struct* sh)
{
  SH2addi(sh, 0, (10<<4)+12);
}
#define init_7_0_10_13
void FASTCALL call_7_0_10_13(SH2_struct* sh)
{
  SH2addi(sh, 0, (10<<4)+13);
}
#define init_7_0_10_14
void FASTCALL call_7_0_10_14(SH2_struct* sh)
{
  SH2addi(sh, 0, (10<<4)+14);
}
#define init_7_0_10_15
void FASTCALL call_7_0_10_15(SH2_struct* sh)
{
  SH2addi(sh, 0, (10<<4)+15);
}
#define init_7_0_11_0
void FASTCALL call_7_0_11_0(SH2_struct* sh)
{
  SH2addi(sh, 0, (11<<4)+0);
}
#define init_7_0_11_1
void FASTCALL call_7_0_11_1(SH2_struct* sh)
{
  SH2addi(sh, 0, (11<<4)+1);
}
#define init_7_0_11_2
void FASTCALL call_7_0_11_2(SH2_struct* sh)
{
  SH2addi(sh, 0, (11<<4)+2);
}
#define init_7_0_11_3
void FASTCALL call_7_0_11_3(SH2_struct* sh)
{
  SH2addi(sh, 0, (11<<4)+3);
}
#define init_7_0_11_4
void FASTCALL call_7_0_11_4(SH2_struct* sh)
{
  SH2addi(sh, 0, (11<<4)+4);
}
#define init_7_0_11_5
void FASTCALL call_7_0_11_5(SH2_struct* sh)
{
  SH2addi(sh, 0, (11<<4)+5);
}
#define init_7_0_11_6
void FASTCALL call_7_0_11_6(SH2_struct* sh)
{
  SH2addi(sh, 0, (11<<4)+6);
}
#define init_7_0_11_7
void FASTCALL call_7_0_11_7(SH2_struct* sh)
{
  SH2addi(sh, 0, (11<<4)+7);
}
#define init_7_0_11_8
void FASTCALL call_7_0_11_8(SH2_struct* sh)
{
  SH2addi(sh, 0, (11<<4)+8);
}
#define init_7_0_11_9
void FASTCALL call_7_0_11_9(SH2_struct* sh)
{
  SH2addi(sh, 0, (11<<4)+9);
}
#define init_7_0_11_10
void FASTCALL call_7_0_11_10(SH2_struct* sh)
{
  SH2addi(sh, 0, (11<<4)+10);
}
#define init_7_0_11_11
void FASTCALL call_7_0_11_11(SH2_struct* sh)
{
  SH2addi(sh, 0, (11<<4)+11);
}
#define init_7_0_11_12
void FASTCALL call_7_0_11_12(SH2_struct* sh)
{
  SH2addi(sh, 0, (11<<4)+12);
}
#define init_7_0_11_13
void FASTCALL call_7_0_11_13(SH2_struct* sh)
{
  SH2addi(sh, 0, (11<<4)+13);
}
#define init_7_0_11_14
void FASTCALL call_7_0_11_14(SH2_struct* sh)
{
  SH2addi(sh, 0, (11<<4)+14);
}
#define init_7_0_11_15
void FASTCALL call_7_0_11_15(SH2_struct* sh)
{
  SH2addi(sh, 0, (11<<4)+15);
}
#define init_7_0_12_0
void FASTCALL call_7_0_12_0(SH2_struct* sh)
{
  SH2addi(sh, 0, (12<<4)+0);
}
#define init_7_0_12_1
void FASTCALL call_7_0_12_1(SH2_struct* sh)
{
  SH2addi(sh, 0, (12<<4)+1);
}
#define init_7_0_12_2
void FASTCALL call_7_0_12_2(SH2_struct* sh)
{
  SH2addi(sh, 0, (12<<4)+2);
}
#define init_7_0_12_3
void FASTCALL call_7_0_12_3(SH2_struct* sh)
{
  SH2addi(sh, 0, (12<<4)+3);
}
#define init_7_0_12_4
void FASTCALL call_7_0_12_4(SH2_struct* sh)
{
  SH2addi(sh, 0, (12<<4)+4);
}
#define init_7_0_12_5
void FASTCALL call_7_0_12_5(SH2_struct* sh)
{
  SH2addi(sh, 0, (12<<4)+5);
}
#define init_7_0_12_6
void FASTCALL call_7_0_12_6(SH2_struct* sh)
{
  SH2addi(sh, 0, (12<<4)+6);
}
#define init_7_0_12_7
void FASTCALL call_7_0_12_7(SH2_struct* sh)
{
  SH2addi(sh, 0, (12<<4)+7);
}
#define init_7_0_12_8
void FASTCALL call_7_0_12_8(SH2_struct* sh)
{
  SH2addi(sh, 0, (12<<4)+8);
}
#define init_7_0_12_9
void FASTCALL call_7_0_12_9(SH2_struct* sh)
{
  SH2addi(sh, 0, (12<<4)+9);
}
#define init_7_0_12_10
void FASTCALL call_7_0_12_10(SH2_struct* sh)
{
  SH2addi(sh, 0, (12<<4)+10);
}
#define init_7_0_12_11
void FASTCALL call_7_0_12_11(SH2_struct* sh)
{
  SH2addi(sh, 0, (12<<4)+11);
}
#define init_7_0_12_12
void FASTCALL call_7_0_12_12(SH2_struct* sh)
{
  SH2addi(sh, 0, (12<<4)+12);
}
#define init_7_0_12_13
void FASTCALL call_7_0_12_13(SH2_struct* sh)
{
  SH2addi(sh, 0, (12<<4)+13);
}
#define init_7_0_12_14
void FASTCALL call_7_0_12_14(SH2_struct* sh)
{
  SH2addi(sh, 0, (12<<4)+14);
}
#define init_7_0_12_15
void FASTCALL call_7_0_12_15(SH2_struct* sh)
{
  SH2addi(sh, 0, (12<<4)+15);
}
#define init_7_0_13_0
void FASTCALL call_7_0_13_0(SH2_struct* sh)
{
  SH2addi(sh, 0, (13<<4)+0);
}
#define init_7_0_13_1
void FASTCALL call_7_0_13_1(SH2_struct* sh)
{
  SH2addi(sh, 0, (13<<4)+1);
}
#define init_7_0_13_2
void FASTCALL call_7_0_13_2(SH2_struct* sh)
{
  SH2addi(sh, 0, (13<<4)+2);
}
#define init_7_0_13_3
void FASTCALL call_7_0_13_3(SH2_struct* sh)
{
  SH2addi(sh, 0, (13<<4)+3);
}
#define init_7_0_13_4
void FASTCALL call_7_0_13_4(SH2_struct* sh)
{
  SH2addi(sh, 0, (13<<4)+4);
}
#define init_7_0_13_5
void FASTCALL call_7_0_13_5(SH2_struct* sh)
{
  SH2addi(sh, 0, (13<<4)+5);
}
#define init_7_0_13_6
void FASTCALL call_7_0_13_6(SH2_struct* sh)
{
  SH2addi(sh, 0, (13<<4)+6);
}
#define init_7_0_13_7
void FASTCALL call_7_0_13_7(SH2_struct* sh)
{
  SH2addi(sh, 0, (13<<4)+7);
}
#define init_7_0_13_8
void FASTCALL call_7_0_13_8(SH2_struct* sh)
{
  SH2addi(sh, 0, (13<<4)+8);
}
#define init_7_0_13_9
void FASTCALL call_7_0_13_9(SH2_struct* sh)
{
  SH2addi(sh, 0, (13<<4)+9);
}
#define init_7_0_13_10
void FASTCALL call_7_0_13_10(SH2_struct* sh)
{
  SH2addi(sh, 0, (13<<4)+10);
}
#define init_7_0_13_11
void FASTCALL call_7_0_13_11(SH2_struct* sh)
{
  SH2addi(sh, 0, (13<<4)+11);
}
#define init_7_0_13_12
void FASTCALL call_7_0_13_12(SH2_struct* sh)
{
  SH2addi(sh, 0, (13<<4)+12);
}
#define init_7_0_13_13
void FASTCALL call_7_0_13_13(SH2_struct* sh)
{
  SH2addi(sh, 0, (13<<4)+13);
}
#define init_7_0_13_14
void FASTCALL call_7_0_13_14(SH2_struct* sh)
{
  SH2addi(sh, 0, (13<<4)+14);
}
#define init_7_0_13_15
void FASTCALL call_7_0_13_15(SH2_struct* sh)
{
  SH2addi(sh, 0, (13<<4)+15);
}
#define init_7_0_14_0
void FASTCALL call_7_0_14_0(SH2_struct* sh)
{
  SH2addi(sh, 0, (14<<4)+0);
}
#define init_7_0_14_1
void FASTCALL call_7_0_14_1(SH2_struct* sh)
{
  SH2addi(sh, 0, (14<<4)+1);
}
#define init_7_0_14_2
void FASTCALL call_7_0_14_2(SH2_struct* sh)
{
  SH2addi(sh, 0, (14<<4)+2);
}
#define init_7_0_14_3
void FASTCALL call_7_0_14_3(SH2_struct* sh)
{
  SH2addi(sh, 0, (14<<4)+3);
}
#define init_7_0_14_4
void FASTCALL call_7_0_14_4(SH2_struct* sh)
{
  SH2addi(sh, 0, (14<<4)+4);
}
#define init_7_0_14_5
void FASTCALL call_7_0_14_5(SH2_struct* sh)
{
  SH2addi(sh, 0, (14<<4)+5);
}
#define init_7_0_14_6
void FASTCALL call_7_0_14_6(SH2_struct* sh)
{
  SH2addi(sh, 0, (14<<4)+6);
}
#define init_7_0_14_7
void FASTCALL call_7_0_14_7(SH2_struct* sh)
{
  SH2addi(sh, 0, (14<<4)+7);
}
#define init_7_0_14_8
void FASTCALL call_7_0_14_8(SH2_struct* sh)
{
  SH2addi(sh, 0, (14<<4)+8);
}
#define init_7_0_14_9
void FASTCALL call_7_0_14_9(SH2_struct* sh)
{
  SH2addi(sh, 0, (14<<4)+9);
}
#define init_7_0_14_10
void FASTCALL call_7_0_14_10(SH2_struct* sh)
{
  SH2addi(sh, 0, (14<<4)+10);
}
#define init_7_0_14_11
void FASTCALL call_7_0_14_11(SH2_struct* sh)
{
  SH2addi(sh, 0, (14<<4)+11);
}
#define init_7_0_14_12
void FASTCALL call_7_0_14_12(SH2_struct* sh)
{
  SH2addi(sh, 0, (14<<4)+12);
}
#define init_7_0_14_13
void FASTCALL call_7_0_14_13(SH2_struct* sh)
{
  SH2addi(sh, 0, (14<<4)+13);
}
#define init_7_0_14_14
void FASTCALL call_7_0_14_14(SH2_struct* sh)
{
  SH2addi(sh, 0, (14<<4)+14);
}
#define init_7_0_14_15
void FASTCALL call_7_0_14_15(SH2_struct* sh)
{
  SH2addi(sh, 0, (14<<4)+15);
}
#define init_7_0_15_0
void FASTCALL call_7_0_15_0(SH2_struct* sh)
{
  SH2addi(sh, 0, (15<<4)+0);
}
#define init_7_0_15_1
void FASTCALL call_7_0_15_1(SH2_struct* sh)
{
  SH2addi(sh, 0, (15<<4)+1);
}
#define init_7_0_15_2
void FASTCALL call_7_0_15_2(SH2_struct* sh)
{
  SH2addi(sh, 0, (15<<4)+2);
}
#define init_7_0_15_3
void FASTCALL call_7_0_15_3(SH2_struct* sh)
{
  SH2addi(sh, 0, (15<<4)+3);
}
#define init_7_0_15_4
void FASTCALL call_7_0_15_4(SH2_struct* sh)
{
  SH2addi(sh, 0, (15<<4)+4);
}
#define init_7_0_15_5
void FASTCALL call_7_0_15_5(SH2_struct* sh)
{
  SH2addi(sh, 0, (15<<4)+5);
}
#define init_7_0_15_6
void FASTCALL call_7_0_15_6(SH2_struct* sh)
{
  SH2addi(sh, 0, (15<<4)+6);
}
#define init_7_0_15_7
void FASTCALL call_7_0_15_7(SH2_struct* sh)
{
  SH2addi(sh, 0, (15<<4)+7);
}
#define init_7_0_15_8
void FASTCALL call_7_0_15_8(SH2_struct* sh)
{
  SH2addi(sh, 0, (15<<4)+8);
}
#define init_7_0_15_9
void FASTCALL call_7_0_15_9(SH2_struct* sh)
{
  SH2addi(sh, 0, (15<<4)+9);
}
#define init_7_0_15_10
void FASTCALL call_7_0_15_10(SH2_struct* sh)
{
  SH2addi(sh, 0, (15<<4)+10);
}
#define init_7_0_15_11
void FASTCALL call_7_0_15_11(SH2_struct* sh)
{
  SH2addi(sh, 0, (15<<4)+11);
}
#define init_7_0_15_12
void FASTCALL call_7_0_15_12(SH2_struct* sh)
{
  SH2addi(sh, 0, (15<<4)+12);
}
#define init_7_0_15_13
void FASTCALL call_7_0_15_13(SH2_struct* sh)
{
  SH2addi(sh, 0, (15<<4)+13);
}
#define init_7_0_15_14
void FASTCALL call_7_0_15_14(SH2_struct* sh)
{
  SH2addi(sh, 0, (15<<4)+14);
}
#define init_7_0_15_15
void FASTCALL call_7_0_15_15(SH2_struct* sh)
{
  SH2addi(sh, 0, (15<<4)+15);
}
#define init_7_1_0_0
void FASTCALL call_7_1_0_0(SH2_struct* sh)
{
  SH2addi(sh, 1, (0<<4)+0);
}
#define init_7_1_0_1
void FASTCALL call_7_1_0_1(SH2_struct* sh)
{
  SH2addi(sh, 1, (0<<4)+1);
}
#define init_7_1_0_2
void FASTCALL call_7_1_0_2(SH2_struct* sh)
{
  SH2addi(sh, 1, (0<<4)+2);
}
#define init_7_1_0_3
void FASTCALL call_7_1_0_3(SH2_struct* sh)
{
  SH2addi(sh, 1, (0<<4)+3);
}
#define init_7_1_0_4
void FASTCALL call_7_1_0_4(SH2_struct* sh)
{
  SH2addi(sh, 1, (0<<4)+4);
}
#define init_7_1_0_5
void FASTCALL call_7_1_0_5(SH2_struct* sh)
{
  SH2addi(sh, 1, (0<<4)+5);
}
#define init_7_1_0_6
void FASTCALL call_7_1_0_6(SH2_struct* sh)
{
  SH2addi(sh, 1, (0<<4)+6);
}
#define init_7_1_0_7
void FASTCALL call_7_1_0_7(SH2_struct* sh)
{
  SH2addi(sh, 1, (0<<4)+7);
}
#define init_7_1_0_8
void FASTCALL call_7_1_0_8(SH2_struct* sh)
{
  SH2addi(sh, 1, (0<<4)+8);
}
#define init_7_1_0_9
void FASTCALL call_7_1_0_9(SH2_struct* sh)
{
  SH2addi(sh, 1, (0<<4)+9);
}
#define init_7_1_0_10
void FASTCALL call_7_1_0_10(SH2_struct* sh)
{
  SH2addi(sh, 1, (0<<4)+10);
}
#define init_7_1_0_11
void FASTCALL call_7_1_0_11(SH2_struct* sh)
{
  SH2addi(sh, 1, (0<<4)+11);
}
#define init_7_1_0_12
void FASTCALL call_7_1_0_12(SH2_struct* sh)
{
  SH2addi(sh, 1, (0<<4)+12);
}
#define init_7_1_0_13
void FASTCALL call_7_1_0_13(SH2_struct* sh)
{
  SH2addi(sh, 1, (0<<4)+13);
}
#define init_7_1_0_14
void FASTCALL call_7_1_0_14(SH2_struct* sh)
{
  SH2addi(sh, 1, (0<<4)+14);
}
#define init_7_1_0_15
void FASTCALL call_7_1_0_15(SH2_struct* sh)
{
  SH2addi(sh, 1, (0<<4)+15);
}
#define init_7_1_1_0
void FASTCALL call_7_1_1_0(SH2_struct* sh)
{
  SH2addi(sh, 1, (1<<4)+0);
}
#define init_7_1_1_1
void FASTCALL call_7_1_1_1(SH2_struct* sh)
{
  SH2addi(sh, 1, (1<<4)+1);
}
#define init_7_1_1_2
void FASTCALL call_7_1_1_2(SH2_struct* sh)
{
  SH2addi(sh, 1, (1<<4)+2);
}
#define init_7_1_1_3
void FASTCALL call_7_1_1_3(SH2_struct* sh)
{
  SH2addi(sh, 1, (1<<4)+3);
}
#define init_7_1_1_4
void FASTCALL call_7_1_1_4(SH2_struct* sh)
{
  SH2addi(sh, 1, (1<<4)+4);
}
#define init_7_1_1_5
void FASTCALL call_7_1_1_5(SH2_struct* sh)
{
  SH2addi(sh, 1, (1<<4)+5);
}
#define init_7_1_1_6
void FASTCALL call_7_1_1_6(SH2_struct* sh)
{
  SH2addi(sh, 1, (1<<4)+6);
}
#define init_7_1_1_7
void FASTCALL call_7_1_1_7(SH2_struct* sh)
{
  SH2addi(sh, 1, (1<<4)+7);
}
#define init_7_1_1_8
void FASTCALL call_7_1_1_8(SH2_struct* sh)
{
  SH2addi(sh, 1, (1<<4)+8);
}
#define init_7_1_1_9
void FASTCALL call_7_1_1_9(SH2_struct* sh)
{
  SH2addi(sh, 1, (1<<4)+9);
}
#define init_7_1_1_10
void FASTCALL call_7_1_1_10(SH2_struct* sh)
{
  SH2addi(sh, 1, (1<<4)+10);
}
#define init_7_1_1_11
void FASTCALL call_7_1_1_11(SH2_struct* sh)
{
  SH2addi(sh, 1, (1<<4)+11);
}
#define init_7_1_1_12
void FASTCALL call_7_1_1_12(SH2_struct* sh)
{
  SH2addi(sh, 1, (1<<4)+12);
}
#define init_7_1_1_13
void FASTCALL call_7_1_1_13(SH2_struct* sh)
{
  SH2addi(sh, 1, (1<<4)+13);
}
#define init_7_1_1_14
void FASTCALL call_7_1_1_14(SH2_struct* sh)
{
  SH2addi(sh, 1, (1<<4)+14);
}
#define init_7_1_1_15
void FASTCALL call_7_1_1_15(SH2_struct* sh)
{
  SH2addi(sh, 1, (1<<4)+15);
}
#define init_7_1_2_0
void FASTCALL call_7_1_2_0(SH2_struct* sh)
{
  SH2addi(sh, 1, (2<<4)+0);
}
#define init_7_1_2_1
void FASTCALL call_7_1_2_1(SH2_struct* sh)
{
  SH2addi(sh, 1, (2<<4)+1);
}
#define init_7_1_2_2
void FASTCALL call_7_1_2_2(SH2_struct* sh)
{
  SH2addi(sh, 1, (2<<4)+2);
}
#define init_7_1_2_3
void FASTCALL call_7_1_2_3(SH2_struct* sh)
{
  SH2addi(sh, 1, (2<<4)+3);
}
#define init_7_1_2_4
void FASTCALL call_7_1_2_4(SH2_struct* sh)
{
  SH2addi(sh, 1, (2<<4)+4);
}
#define init_7_1_2_5
void FASTCALL call_7_1_2_5(SH2_struct* sh)
{
  SH2addi(sh, 1, (2<<4)+5);
}
#define init_7_1_2_6
void FASTCALL call_7_1_2_6(SH2_struct* sh)
{
  SH2addi(sh, 1, (2<<4)+6);
}
#define init_7_1_2_7
void FASTCALL call_7_1_2_7(SH2_struct* sh)
{
  SH2addi(sh, 1, (2<<4)+7);
}
#define init_7_1_2_8
void FASTCALL call_7_1_2_8(SH2_struct* sh)
{
  SH2addi(sh, 1, (2<<4)+8);
}
#define init_7_1_2_9
void FASTCALL call_7_1_2_9(SH2_struct* sh)
{
  SH2addi(sh, 1, (2<<4)+9);
}
#define init_7_1_2_10
void FASTCALL call_7_1_2_10(SH2_struct* sh)
{
  SH2addi(sh, 1, (2<<4)+10);
}
#define init_7_1_2_11
void FASTCALL call_7_1_2_11(SH2_struct* sh)
{
  SH2addi(sh, 1, (2<<4)+11);
}
#define init_7_1_2_12
void FASTCALL call_7_1_2_12(SH2_struct* sh)
{
  SH2addi(sh, 1, (2<<4)+12);
}
#define init_7_1_2_13
void FASTCALL call_7_1_2_13(SH2_struct* sh)
{
  SH2addi(sh, 1, (2<<4)+13);
}
#define init_7_1_2_14
void FASTCALL call_7_1_2_14(SH2_struct* sh)
{
  SH2addi(sh, 1, (2<<4)+14);
}
#define init_7_1_2_15
void FASTCALL call_7_1_2_15(SH2_struct* sh)
{
  SH2addi(sh, 1, (2<<4)+15);
}
#define init_7_1_3_0
void FASTCALL call_7_1_3_0(SH2_struct* sh)
{
  SH2addi(sh, 1, (3<<4)+0);
}
#define init_7_1_3_1
void FASTCALL call_7_1_3_1(SH2_struct* sh)
{
  SH2addi(sh, 1, (3<<4)+1);
}
#define init_7_1_3_2
void FASTCALL call_7_1_3_2(SH2_struct* sh)
{
  SH2addi(sh, 1, (3<<4)+2);
}
#define init_7_1_3_3
void FASTCALL call_7_1_3_3(SH2_struct* sh)
{
  SH2addi(sh, 1, (3<<4)+3);
}
#define init_7_1_3_4
void FASTCALL call_7_1_3_4(SH2_struct* sh)
{
  SH2addi(sh, 1, (3<<4)+4);
}
#define init_7_1_3_5
void FASTCALL call_7_1_3_5(SH2_struct* sh)
{
  SH2addi(sh, 1, (3<<4)+5);
}
#define init_7_1_3_6
void FASTCALL call_7_1_3_6(SH2_struct* sh)
{
  SH2addi(sh, 1, (3<<4)+6);
}
#define init_7_1_3_7
void FASTCALL call_7_1_3_7(SH2_struct* sh)
{
  SH2addi(sh, 1, (3<<4)+7);
}
#define init_7_1_3_8
void FASTCALL call_7_1_3_8(SH2_struct* sh)
{
  SH2addi(sh, 1, (3<<4)+8);
}
#define init_7_1_3_9
void FASTCALL call_7_1_3_9(SH2_struct* sh)
{
  SH2addi(sh, 1, (3<<4)+9);
}
#define init_7_1_3_10
void FASTCALL call_7_1_3_10(SH2_struct* sh)
{
  SH2addi(sh, 1, (3<<4)+10);
}
#define init_7_1_3_11
void FASTCALL call_7_1_3_11(SH2_struct* sh)
{
  SH2addi(sh, 1, (3<<4)+11);
}
#define init_7_1_3_12
void FASTCALL call_7_1_3_12(SH2_struct* sh)
{
  SH2addi(sh, 1, (3<<4)+12);
}
#define init_7_1_3_13
void FASTCALL call_7_1_3_13(SH2_struct* sh)
{
  SH2addi(sh, 1, (3<<4)+13);
}
#define init_7_1_3_14
void FASTCALL call_7_1_3_14(SH2_struct* sh)
{
  SH2addi(sh, 1, (3<<4)+14);
}
#define init_7_1_3_15
void FASTCALL call_7_1_3_15(SH2_struct* sh)
{
  SH2addi(sh, 1, (3<<4)+15);
}
#define init_7_1_4_0
void FASTCALL call_7_1_4_0(SH2_struct* sh)
{
  SH2addi(sh, 1, (4<<4)+0);
}
#define init_7_1_4_1
void FASTCALL call_7_1_4_1(SH2_struct* sh)
{
  SH2addi(sh, 1, (4<<4)+1);
}
#define init_7_1_4_2
void FASTCALL call_7_1_4_2(SH2_struct* sh)
{
  SH2addi(sh, 1, (4<<4)+2);
}
#define init_7_1_4_3
void FASTCALL call_7_1_4_3(SH2_struct* sh)
{
  SH2addi(sh, 1, (4<<4)+3);
}
#define init_7_1_4_4
void FASTCALL call_7_1_4_4(SH2_struct* sh)
{
  SH2addi(sh, 1, (4<<4)+4);
}
#define init_7_1_4_5
void FASTCALL call_7_1_4_5(SH2_struct* sh)
{
  SH2addi(sh, 1, (4<<4)+5);
}
#define init_7_1_4_6
void FASTCALL call_7_1_4_6(SH2_struct* sh)
{
  SH2addi(sh, 1, (4<<4)+6);
}
#define init_7_1_4_7
void FASTCALL call_7_1_4_7(SH2_struct* sh)
{
  SH2addi(sh, 1, (4<<4)+7);
}
#define init_7_1_4_8
void FASTCALL call_7_1_4_8(SH2_struct* sh)
{
  SH2addi(sh, 1, (4<<4)+8);
}
#define init_7_1_4_9
void FASTCALL call_7_1_4_9(SH2_struct* sh)
{
  SH2addi(sh, 1, (4<<4)+9);
}
#define init_7_1_4_10
void FASTCALL call_7_1_4_10(SH2_struct* sh)
{
  SH2addi(sh, 1, (4<<4)+10);
}
#define init_7_1_4_11
void FASTCALL call_7_1_4_11(SH2_struct* sh)
{
  SH2addi(sh, 1, (4<<4)+11);
}
#define init_7_1_4_12
void FASTCALL call_7_1_4_12(SH2_struct* sh)
{
  SH2addi(sh, 1, (4<<4)+12);
}
#define init_7_1_4_13
void FASTCALL call_7_1_4_13(SH2_struct* sh)
{
  SH2addi(sh, 1, (4<<4)+13);
}
#define init_7_1_4_14
void FASTCALL call_7_1_4_14(SH2_struct* sh)
{
  SH2addi(sh, 1, (4<<4)+14);
}
#define init_7_1_4_15
void FASTCALL call_7_1_4_15(SH2_struct* sh)
{
  SH2addi(sh, 1, (4<<4)+15);
}
#define init_7_1_5_0
void FASTCALL call_7_1_5_0(SH2_struct* sh)
{
  SH2addi(sh, 1, (5<<4)+0);
}
#define init_7_1_5_1
void FASTCALL call_7_1_5_1(SH2_struct* sh)
{
  SH2addi(sh, 1, (5<<4)+1);
}
#define init_7_1_5_2
void FASTCALL call_7_1_5_2(SH2_struct* sh)
{
  SH2addi(sh, 1, (5<<4)+2);
}
#define init_7_1_5_3
void FASTCALL call_7_1_5_3(SH2_struct* sh)
{
  SH2addi(sh, 1, (5<<4)+3);
}
#define init_7_1_5_4
void FASTCALL call_7_1_5_4(SH2_struct* sh)
{
  SH2addi(sh, 1, (5<<4)+4);
}
#define init_7_1_5_5
void FASTCALL call_7_1_5_5(SH2_struct* sh)
{
  SH2addi(sh, 1, (5<<4)+5);
}
#define init_7_1_5_6
void FASTCALL call_7_1_5_6(SH2_struct* sh)
{
  SH2addi(sh, 1, (5<<4)+6);
}
#define init_7_1_5_7
void FASTCALL call_7_1_5_7(SH2_struct* sh)
{
  SH2addi(sh, 1, (5<<4)+7);
}
#define init_7_1_5_8
void FASTCALL call_7_1_5_8(SH2_struct* sh)
{
  SH2addi(sh, 1, (5<<4)+8);
}
#define init_7_1_5_9
void FASTCALL call_7_1_5_9(SH2_struct* sh)
{
  SH2addi(sh, 1, (5<<4)+9);
}
#define init_7_1_5_10
void FASTCALL call_7_1_5_10(SH2_struct* sh)
{
  SH2addi(sh, 1, (5<<4)+10);
}
#define init_7_1_5_11
void FASTCALL call_7_1_5_11(SH2_struct* sh)
{
  SH2addi(sh, 1, (5<<4)+11);
}
#define init_7_1_5_12
void FASTCALL call_7_1_5_12(SH2_struct* sh)
{
  SH2addi(sh, 1, (5<<4)+12);
}
#define init_7_1_5_13
void FASTCALL call_7_1_5_13(SH2_struct* sh)
{
  SH2addi(sh, 1, (5<<4)+13);
}
#define init_7_1_5_14
void FASTCALL call_7_1_5_14(SH2_struct* sh)
{
  SH2addi(sh, 1, (5<<4)+14);
}
#define init_7_1_5_15
void FASTCALL call_7_1_5_15(SH2_struct* sh)
{
  SH2addi(sh, 1, (5<<4)+15);
}
#define init_7_1_6_0
void FASTCALL call_7_1_6_0(SH2_struct* sh)
{
  SH2addi(sh, 1, (6<<4)+0);
}
#define init_7_1_6_1
void FASTCALL call_7_1_6_1(SH2_struct* sh)
{
  SH2addi(sh, 1, (6<<4)+1);
}
#define init_7_1_6_2
void FASTCALL call_7_1_6_2(SH2_struct* sh)
{
  SH2addi(sh, 1, (6<<4)+2);
}
#define init_7_1_6_3
void FASTCALL call_7_1_6_3(SH2_struct* sh)
{
  SH2addi(sh, 1, (6<<4)+3);
}
#define init_7_1_6_4
void FASTCALL call_7_1_6_4(SH2_struct* sh)
{
  SH2addi(sh, 1, (6<<4)+4);
}
#define init_7_1_6_5
void FASTCALL call_7_1_6_5(SH2_struct* sh)
{
  SH2addi(sh, 1, (6<<4)+5);
}
#define init_7_1_6_6
void FASTCALL call_7_1_6_6(SH2_struct* sh)
{
  SH2addi(sh, 1, (6<<4)+6);
}
#define init_7_1_6_7
void FASTCALL call_7_1_6_7(SH2_struct* sh)
{
  SH2addi(sh, 1, (6<<4)+7);
}
#define init_7_1_6_8
void FASTCALL call_7_1_6_8(SH2_struct* sh)
{
  SH2addi(sh, 1, (6<<4)+8);
}
#define init_7_1_6_9
void FASTCALL call_7_1_6_9(SH2_struct* sh)
{
  SH2addi(sh, 1, (6<<4)+9);
}
#define init_7_1_6_10
void FASTCALL call_7_1_6_10(SH2_struct* sh)
{
  SH2addi(sh, 1, (6<<4)+10);
}
#define init_7_1_6_11
void FASTCALL call_7_1_6_11(SH2_struct* sh)
{
  SH2addi(sh, 1, (6<<4)+11);
}
#define init_7_1_6_12
void FASTCALL call_7_1_6_12(SH2_struct* sh)
{
  SH2addi(sh, 1, (6<<4)+12);
}
#define init_7_1_6_13
void FASTCALL call_7_1_6_13(SH2_struct* sh)
{
  SH2addi(sh, 1, (6<<4)+13);
}
#define init_7_1_6_14
void FASTCALL call_7_1_6_14(SH2_struct* sh)
{
  SH2addi(sh, 1, (6<<4)+14);
}
#define init_7_1_6_15
void FASTCALL call_7_1_6_15(SH2_struct* sh)
{
  SH2addi(sh, 1, (6<<4)+15);
}
#define init_7_1_7_0
void FASTCALL call_7_1_7_0(SH2_struct* sh)
{
  SH2addi(sh, 1, (7<<4)+0);
}
#define init_7_1_7_1
void FASTCALL call_7_1_7_1(SH2_struct* sh)
{
  SH2addi(sh, 1, (7<<4)+1);
}
#define init_7_1_7_2
void FASTCALL call_7_1_7_2(SH2_struct* sh)
{
  SH2addi(sh, 1, (7<<4)+2);
}
#define init_7_1_7_3
void FASTCALL call_7_1_7_3(SH2_struct* sh)
{
  SH2addi(sh, 1, (7<<4)+3);
}
#define init_7_1_7_4
void FASTCALL call_7_1_7_4(SH2_struct* sh)
{
  SH2addi(sh, 1, (7<<4)+4);
}
#define init_7_1_7_5
void FASTCALL call_7_1_7_5(SH2_struct* sh)
{
  SH2addi(sh, 1, (7<<4)+5);
}
#define init_7_1_7_6
void FASTCALL call_7_1_7_6(SH2_struct* sh)
{
  SH2addi(sh, 1, (7<<4)+6);
}
#define init_7_1_7_7
void FASTCALL call_7_1_7_7(SH2_struct* sh)
{
  SH2addi(sh, 1, (7<<4)+7);
}
#define init_7_1_7_8
void FASTCALL call_7_1_7_8(SH2_struct* sh)
{
  SH2addi(sh, 1, (7<<4)+8);
}
#define init_7_1_7_9
void FASTCALL call_7_1_7_9(SH2_struct* sh)
{
  SH2addi(sh, 1, (7<<4)+9);
}
#define init_7_1_7_10
void FASTCALL call_7_1_7_10(SH2_struct* sh)
{
  SH2addi(sh, 1, (7<<4)+10);
}
#define init_7_1_7_11
void FASTCALL call_7_1_7_11(SH2_struct* sh)
{
  SH2addi(sh, 1, (7<<4)+11);
}
#define init_7_1_7_12
void FASTCALL call_7_1_7_12(SH2_struct* sh)
{
  SH2addi(sh, 1, (7<<4)+12);
}
#define init_7_1_7_13
void FASTCALL call_7_1_7_13(SH2_struct* sh)
{
  SH2addi(sh, 1, (7<<4)+13);
}
#define init_7_1_7_14
void FASTCALL call_7_1_7_14(SH2_struct* sh)
{
  SH2addi(sh, 1, (7<<4)+14);
}
#define init_7_1_7_15
void FASTCALL call_7_1_7_15(SH2_struct* sh)
{
  SH2addi(sh, 1, (7<<4)+15);
}
#define init_7_1_8_0
void FASTCALL call_7_1_8_0(SH2_struct* sh)
{
  SH2addi(sh, 1, (8<<4)+0);
}
#define init_7_1_8_1
void FASTCALL call_7_1_8_1(SH2_struct* sh)
{
  SH2addi(sh, 1, (8<<4)+1);
}
#define init_7_1_8_2
void FASTCALL call_7_1_8_2(SH2_struct* sh)
{
  SH2addi(sh, 1, (8<<4)+2);
}
#define init_7_1_8_3
void FASTCALL call_7_1_8_3(SH2_struct* sh)
{
  SH2addi(sh, 1, (8<<4)+3);
}
#define init_7_1_8_4
void FASTCALL call_7_1_8_4(SH2_struct* sh)
{
  SH2addi(sh, 1, (8<<4)+4);
}
#define init_7_1_8_5
void FASTCALL call_7_1_8_5(SH2_struct* sh)
{
  SH2addi(sh, 1, (8<<4)+5);
}
#define init_7_1_8_6
void FASTCALL call_7_1_8_6(SH2_struct* sh)
{
  SH2addi(sh, 1, (8<<4)+6);
}
#define init_7_1_8_7
void FASTCALL call_7_1_8_7(SH2_struct* sh)
{
  SH2addi(sh, 1, (8<<4)+7);
}
#define init_7_1_8_8
void FASTCALL call_7_1_8_8(SH2_struct* sh)
{
  SH2addi(sh, 1, (8<<4)+8);
}
#define init_7_1_8_9
void FASTCALL call_7_1_8_9(SH2_struct* sh)
{
  SH2addi(sh, 1, (8<<4)+9);
}
#define init_7_1_8_10
void FASTCALL call_7_1_8_10(SH2_struct* sh)
{
  SH2addi(sh, 1, (8<<4)+10);
}
#define init_7_1_8_11
void FASTCALL call_7_1_8_11(SH2_struct* sh)
{
  SH2addi(sh, 1, (8<<4)+11);
}
#define init_7_1_8_12
void FASTCALL call_7_1_8_12(SH2_struct* sh)
{
  SH2addi(sh, 1, (8<<4)+12);
}
#define init_7_1_8_13
void FASTCALL call_7_1_8_13(SH2_struct* sh)
{
  SH2addi(sh, 1, (8<<4)+13);
}
#define init_7_1_8_14
void FASTCALL call_7_1_8_14(SH2_struct* sh)
{
  SH2addi(sh, 1, (8<<4)+14);
}
#define init_7_1_8_15
void FASTCALL call_7_1_8_15(SH2_struct* sh)
{
  SH2addi(sh, 1, (8<<4)+15);
}
#define init_7_1_9_0
void FASTCALL call_7_1_9_0(SH2_struct* sh)
{
  SH2addi(sh, 1, (9<<4)+0);
}
#define init_7_1_9_1
void FASTCALL call_7_1_9_1(SH2_struct* sh)
{
  SH2addi(sh, 1, (9<<4)+1);
}
#define init_7_1_9_2
void FASTCALL call_7_1_9_2(SH2_struct* sh)
{
  SH2addi(sh, 1, (9<<4)+2);
}
#define init_7_1_9_3
void FASTCALL call_7_1_9_3(SH2_struct* sh)
{
  SH2addi(sh, 1, (9<<4)+3);
}
#define init_7_1_9_4
void FASTCALL call_7_1_9_4(SH2_struct* sh)
{
  SH2addi(sh, 1, (9<<4)+4);
}
#define init_7_1_9_5
void FASTCALL call_7_1_9_5(SH2_struct* sh)
{
  SH2addi(sh, 1, (9<<4)+5);
}
#define init_7_1_9_6
void FASTCALL call_7_1_9_6(SH2_struct* sh)
{
  SH2addi(sh, 1, (9<<4)+6);
}
#define init_7_1_9_7
void FASTCALL call_7_1_9_7(SH2_struct* sh)
{
  SH2addi(sh, 1, (9<<4)+7);
}
#define init_7_1_9_8
void FASTCALL call_7_1_9_8(SH2_struct* sh)
{
  SH2addi(sh, 1, (9<<4)+8);
}
#define init_7_1_9_9
void FASTCALL call_7_1_9_9(SH2_struct* sh)
{
  SH2addi(sh, 1, (9<<4)+9);
}
#define init_7_1_9_10
void FASTCALL call_7_1_9_10(SH2_struct* sh)
{
  SH2addi(sh, 1, (9<<4)+10);
}
#define init_7_1_9_11
void FASTCALL call_7_1_9_11(SH2_struct* sh)
{
  SH2addi(sh, 1, (9<<4)+11);
}
#define init_7_1_9_12
void FASTCALL call_7_1_9_12(SH2_struct* sh)
{
  SH2addi(sh, 1, (9<<4)+12);
}
#define init_7_1_9_13
void FASTCALL call_7_1_9_13(SH2_struct* sh)
{
  SH2addi(sh, 1, (9<<4)+13);
}
#define init_7_1_9_14
void FASTCALL call_7_1_9_14(SH2_struct* sh)
{
  SH2addi(sh, 1, (9<<4)+14);
}
#define init_7_1_9_15
void FASTCALL call_7_1_9_15(SH2_struct* sh)
{
  SH2addi(sh, 1, (9<<4)+15);
}
#define init_7_1_10_0
void FASTCALL call_7_1_10_0(SH2_struct* sh)
{
  SH2addi(sh, 1, (10<<4)+0);
}
#define init_7_1_10_1
void FASTCALL call_7_1_10_1(SH2_struct* sh)
{
  SH2addi(sh, 1, (10<<4)+1);
}
#define init_7_1_10_2
void FASTCALL call_7_1_10_2(SH2_struct* sh)
{
  SH2addi(sh, 1, (10<<4)+2);
}
#define init_7_1_10_3
void FASTCALL call_7_1_10_3(SH2_struct* sh)
{
  SH2addi(sh, 1, (10<<4)+3);
}
#define init_7_1_10_4
void FASTCALL call_7_1_10_4(SH2_struct* sh)
{
  SH2addi(sh, 1, (10<<4)+4);
}
#define init_7_1_10_5
void FASTCALL call_7_1_10_5(SH2_struct* sh)
{
  SH2addi(sh, 1, (10<<4)+5);
}
#define init_7_1_10_6
void FASTCALL call_7_1_10_6(SH2_struct* sh)
{
  SH2addi(sh, 1, (10<<4)+6);
}
#define init_7_1_10_7
void FASTCALL call_7_1_10_7(SH2_struct* sh)
{
  SH2addi(sh, 1, (10<<4)+7);
}
#define init_7_1_10_8
void FASTCALL call_7_1_10_8(SH2_struct* sh)
{
  SH2addi(sh, 1, (10<<4)+8);
}
#define init_7_1_10_9
void FASTCALL call_7_1_10_9(SH2_struct* sh)
{
  SH2addi(sh, 1, (10<<4)+9);
}
#define init_7_1_10_10
void FASTCALL call_7_1_10_10(SH2_struct* sh)
{
  SH2addi(sh, 1, (10<<4)+10);
}
#define init_7_1_10_11
void FASTCALL call_7_1_10_11(SH2_struct* sh)
{
  SH2addi(sh, 1, (10<<4)+11);
}
#define init_7_1_10_12
void FASTCALL call_7_1_10_12(SH2_struct* sh)
{
  SH2addi(sh, 1, (10<<4)+12);
}
#define init_7_1_10_13
void FASTCALL call_7_1_10_13(SH2_struct* sh)
{
  SH2addi(sh, 1, (10<<4)+13);
}
#define init_7_1_10_14
void FASTCALL call_7_1_10_14(SH2_struct* sh)
{
  SH2addi(sh, 1, (10<<4)+14);
}
#define init_7_1_10_15
void FASTCALL call_7_1_10_15(SH2_struct* sh)
{
  SH2addi(sh, 1, (10<<4)+15);
}
#define init_7_1_11_0
void FASTCALL call_7_1_11_0(SH2_struct* sh)
{
  SH2addi(sh, 1, (11<<4)+0);
}
#define init_7_1_11_1
void FASTCALL call_7_1_11_1(SH2_struct* sh)
{
  SH2addi(sh, 1, (11<<4)+1);
}
#define init_7_1_11_2
void FASTCALL call_7_1_11_2(SH2_struct* sh)
{
  SH2addi(sh, 1, (11<<4)+2);
}
#define init_7_1_11_3
void FASTCALL call_7_1_11_3(SH2_struct* sh)
{
  SH2addi(sh, 1, (11<<4)+3);
}
#define init_7_1_11_4
void FASTCALL call_7_1_11_4(SH2_struct* sh)
{
  SH2addi(sh, 1, (11<<4)+4);
}
#define init_7_1_11_5
void FASTCALL call_7_1_11_5(SH2_struct* sh)
{
  SH2addi(sh, 1, (11<<4)+5);
}
#define init_7_1_11_6
void FASTCALL call_7_1_11_6(SH2_struct* sh)
{
  SH2addi(sh, 1, (11<<4)+6);
}
#define init_7_1_11_7
void FASTCALL call_7_1_11_7(SH2_struct* sh)
{
  SH2addi(sh, 1, (11<<4)+7);
}
#define init_7_1_11_8
void FASTCALL call_7_1_11_8(SH2_struct* sh)
{
  SH2addi(sh, 1, (11<<4)+8);
}
#define init_7_1_11_9
void FASTCALL call_7_1_11_9(SH2_struct* sh)
{
  SH2addi(sh, 1, (11<<4)+9);
}
#define init_7_1_11_10
void FASTCALL call_7_1_11_10(SH2_struct* sh)
{
  SH2addi(sh, 1, (11<<4)+10);
}
#define init_7_1_11_11
void FASTCALL call_7_1_11_11(SH2_struct* sh)
{
  SH2addi(sh, 1, (11<<4)+11);
}
#define init_7_1_11_12
void FASTCALL call_7_1_11_12(SH2_struct* sh)
{
  SH2addi(sh, 1, (11<<4)+12);
}
#define init_7_1_11_13
void FASTCALL call_7_1_11_13(SH2_struct* sh)
{
  SH2addi(sh, 1, (11<<4)+13);
}
#define init_7_1_11_14
void FASTCALL call_7_1_11_14(SH2_struct* sh)
{
  SH2addi(sh, 1, (11<<4)+14);
}
#define init_7_1_11_15
void FASTCALL call_7_1_11_15(SH2_struct* sh)
{
  SH2addi(sh, 1, (11<<4)+15);
}
#define init_7_1_12_0
void FASTCALL call_7_1_12_0(SH2_struct* sh)
{
  SH2addi(sh, 1, (12<<4)+0);
}
#define init_7_1_12_1
void FASTCALL call_7_1_12_1(SH2_struct* sh)
{
  SH2addi(sh, 1, (12<<4)+1);
}
#define init_7_1_12_2
void FASTCALL call_7_1_12_2(SH2_struct* sh)
{
  SH2addi(sh, 1, (12<<4)+2);
}
#define init_7_1_12_3
void FASTCALL call_7_1_12_3(SH2_struct* sh)
{
  SH2addi(sh, 1, (12<<4)+3);
}
#define init_7_1_12_4
void FASTCALL call_7_1_12_4(SH2_struct* sh)
{
  SH2addi(sh, 1, (12<<4)+4);
}
#define init_7_1_12_5
void FASTCALL call_7_1_12_5(SH2_struct* sh)
{
  SH2addi(sh, 1, (12<<4)+5);
}
#define init_7_1_12_6
void FASTCALL call_7_1_12_6(SH2_struct* sh)
{
  SH2addi(sh, 1, (12<<4)+6);
}
#define init_7_1_12_7
void FASTCALL call_7_1_12_7(SH2_struct* sh)
{
  SH2addi(sh, 1, (12<<4)+7);
}
#define init_7_1_12_8
void FASTCALL call_7_1_12_8(SH2_struct* sh)
{
  SH2addi(sh, 1, (12<<4)+8);
}
#define init_7_1_12_9
void FASTCALL call_7_1_12_9(SH2_struct* sh)
{
  SH2addi(sh, 1, (12<<4)+9);
}
#define init_7_1_12_10
void FASTCALL call_7_1_12_10(SH2_struct* sh)
{
  SH2addi(sh, 1, (12<<4)+10);
}
#define init_7_1_12_11
void FASTCALL call_7_1_12_11(SH2_struct* sh)
{
  SH2addi(sh, 1, (12<<4)+11);
}
#define init_7_1_12_12
void FASTCALL call_7_1_12_12(SH2_struct* sh)
{
  SH2addi(sh, 1, (12<<4)+12);
}
#define init_7_1_12_13
void FASTCALL call_7_1_12_13(SH2_struct* sh)
{
  SH2addi(sh, 1, (12<<4)+13);
}
#define init_7_1_12_14
void FASTCALL call_7_1_12_14(SH2_struct* sh)
{
  SH2addi(sh, 1, (12<<4)+14);
}
#define init_7_1_12_15
void FASTCALL call_7_1_12_15(SH2_struct* sh)
{
  SH2addi(sh, 1, (12<<4)+15);
}
#define init_7_1_13_0
void FASTCALL call_7_1_13_0(SH2_struct* sh)
{
  SH2addi(sh, 1, (13<<4)+0);
}
#define init_7_1_13_1
void FASTCALL call_7_1_13_1(SH2_struct* sh)
{
  SH2addi(sh, 1, (13<<4)+1);
}
#define init_7_1_13_2
void FASTCALL call_7_1_13_2(SH2_struct* sh)
{
  SH2addi(sh, 1, (13<<4)+2);
}
#define init_7_1_13_3
void FASTCALL call_7_1_13_3(SH2_struct* sh)
{
  SH2addi(sh, 1, (13<<4)+3);
}
#define init_7_1_13_4
void FASTCALL call_7_1_13_4(SH2_struct* sh)
{
  SH2addi(sh, 1, (13<<4)+4);
}
#define init_7_1_13_5
void FASTCALL call_7_1_13_5(SH2_struct* sh)
{
  SH2addi(sh, 1, (13<<4)+5);
}
#define init_7_1_13_6
void FASTCALL call_7_1_13_6(SH2_struct* sh)
{
  SH2addi(sh, 1, (13<<4)+6);
}
#define init_7_1_13_7
void FASTCALL call_7_1_13_7(SH2_struct* sh)
{
  SH2addi(sh, 1, (13<<4)+7);
}
#define init_7_1_13_8
void FASTCALL call_7_1_13_8(SH2_struct* sh)
{
  SH2addi(sh, 1, (13<<4)+8);
}
#define init_7_1_13_9
void FASTCALL call_7_1_13_9(SH2_struct* sh)
{
  SH2addi(sh, 1, (13<<4)+9);
}
#define init_7_1_13_10
void FASTCALL call_7_1_13_10(SH2_struct* sh)
{
  SH2addi(sh, 1, (13<<4)+10);
}
#define init_7_1_13_11
void FASTCALL call_7_1_13_11(SH2_struct* sh)
{
  SH2addi(sh, 1, (13<<4)+11);
}
#define init_7_1_13_12
void FASTCALL call_7_1_13_12(SH2_struct* sh)
{
  SH2addi(sh, 1, (13<<4)+12);
}
#define init_7_1_13_13
void FASTCALL call_7_1_13_13(SH2_struct* sh)
{
  SH2addi(sh, 1, (13<<4)+13);
}
#define init_7_1_13_14
void FASTCALL call_7_1_13_14(SH2_struct* sh)
{
  SH2addi(sh, 1, (13<<4)+14);
}
#define init_7_1_13_15
void FASTCALL call_7_1_13_15(SH2_struct* sh)
{
  SH2addi(sh, 1, (13<<4)+15);
}
#define init_7_1_14_0
void FASTCALL call_7_1_14_0(SH2_struct* sh)
{
  SH2addi(sh, 1, (14<<4)+0);
}
#define init_7_1_14_1
void FASTCALL call_7_1_14_1(SH2_struct* sh)
{
  SH2addi(sh, 1, (14<<4)+1);
}
#define init_7_1_14_2
void FASTCALL call_7_1_14_2(SH2_struct* sh)
{
  SH2addi(sh, 1, (14<<4)+2);
}
#define init_7_1_14_3
void FASTCALL call_7_1_14_3(SH2_struct* sh)
{
  SH2addi(sh, 1, (14<<4)+3);
}
#define init_7_1_14_4
void FASTCALL call_7_1_14_4(SH2_struct* sh)
{
  SH2addi(sh, 1, (14<<4)+4);
}
#define init_7_1_14_5
void FASTCALL call_7_1_14_5(SH2_struct* sh)
{
  SH2addi(sh, 1, (14<<4)+5);
}
#define init_7_1_14_6
void FASTCALL call_7_1_14_6(SH2_struct* sh)
{
  SH2addi(sh, 1, (14<<4)+6);
}
#define init_7_1_14_7
void FASTCALL call_7_1_14_7(SH2_struct* sh)
{
  SH2addi(sh, 1, (14<<4)+7);
}
#define init_7_1_14_8
void FASTCALL call_7_1_14_8(SH2_struct* sh)
{
  SH2addi(sh, 1, (14<<4)+8);
}
#define init_7_1_14_9
void FASTCALL call_7_1_14_9(SH2_struct* sh)
{
  SH2addi(sh, 1, (14<<4)+9);
}
#define init_7_1_14_10
void FASTCALL call_7_1_14_10(SH2_struct* sh)
{
  SH2addi(sh, 1, (14<<4)+10);
}
#define init_7_1_14_11
void FASTCALL call_7_1_14_11(SH2_struct* sh)
{
  SH2addi(sh, 1, (14<<4)+11);
}
#define init_7_1_14_12
void FASTCALL call_7_1_14_12(SH2_struct* sh)
{
  SH2addi(sh, 1, (14<<4)+12);
}
#define init_7_1_14_13
void FASTCALL call_7_1_14_13(SH2_struct* sh)
{
  SH2addi(sh, 1, (14<<4)+13);
}
#define init_7_1_14_14
void FASTCALL call_7_1_14_14(SH2_struct* sh)
{
  SH2addi(sh, 1, (14<<4)+14);
}
#define init_7_1_14_15
void FASTCALL call_7_1_14_15(SH2_struct* sh)
{
  SH2addi(sh, 1, (14<<4)+15);
}
#define init_7_1_15_0
void FASTCALL call_7_1_15_0(SH2_struct* sh)
{
  SH2addi(sh, 1, (15<<4)+0);
}
#define init_7_1_15_1
void FASTCALL call_7_1_15_1(SH2_struct* sh)
{
  SH2addi(sh, 1, (15<<4)+1);
}
#define init_7_1_15_2
void FASTCALL call_7_1_15_2(SH2_struct* sh)
{
  SH2addi(sh, 1, (15<<4)+2);
}
#define init_7_1_15_3
void FASTCALL call_7_1_15_3(SH2_struct* sh)
{
  SH2addi(sh, 1, (15<<4)+3);
}
#define init_7_1_15_4
void FASTCALL call_7_1_15_4(SH2_struct* sh)
{
  SH2addi(sh, 1, (15<<4)+4);
}
#define init_7_1_15_5
void FASTCALL call_7_1_15_5(SH2_struct* sh)
{
  SH2addi(sh, 1, (15<<4)+5);
}
#define init_7_1_15_6
void FASTCALL call_7_1_15_6(SH2_struct* sh)
{
  SH2addi(sh, 1, (15<<4)+6);
}
#define init_7_1_15_7
void FASTCALL call_7_1_15_7(SH2_struct* sh)
{
  SH2addi(sh, 1, (15<<4)+7);
}
#define init_7_1_15_8
void FASTCALL call_7_1_15_8(SH2_struct* sh)
{
  SH2addi(sh, 1, (15<<4)+8);
}
#define init_7_1_15_9
void FASTCALL call_7_1_15_9(SH2_struct* sh)
{
  SH2addi(sh, 1, (15<<4)+9);
}
#define init_7_1_15_10
void FASTCALL call_7_1_15_10(SH2_struct* sh)
{
  SH2addi(sh, 1, (15<<4)+10);
}
#define init_7_1_15_11
void FASTCALL call_7_1_15_11(SH2_struct* sh)
{
  SH2addi(sh, 1, (15<<4)+11);
}
#define init_7_1_15_12
void FASTCALL call_7_1_15_12(SH2_struct* sh)
{
  SH2addi(sh, 1, (15<<4)+12);
}
#define init_7_1_15_13
void FASTCALL call_7_1_15_13(SH2_struct* sh)
{
  SH2addi(sh, 1, (15<<4)+13);
}
#define init_7_1_15_14
void FASTCALL call_7_1_15_14(SH2_struct* sh)
{
  SH2addi(sh, 1, (15<<4)+14);
}
#define init_7_1_15_15
void FASTCALL call_7_1_15_15(SH2_struct* sh)
{
  SH2addi(sh, 1, (15<<4)+15);
}
#define init_7_2_0_0
void FASTCALL call_7_2_0_0(SH2_struct* sh)
{
  SH2addi(sh, 2, (0<<4)+0);
}
#define init_7_2_0_1
void FASTCALL call_7_2_0_1(SH2_struct* sh)
{
  SH2addi(sh, 2, (0<<4)+1);
}
#define init_7_2_0_2
void FASTCALL call_7_2_0_2(SH2_struct* sh)
{
  SH2addi(sh, 2, (0<<4)+2);
}
#define init_7_2_0_3
void FASTCALL call_7_2_0_3(SH2_struct* sh)
{
  SH2addi(sh, 2, (0<<4)+3);
}
#define init_7_2_0_4
void FASTCALL call_7_2_0_4(SH2_struct* sh)
{
  SH2addi(sh, 2, (0<<4)+4);
}
#define init_7_2_0_5
void FASTCALL call_7_2_0_5(SH2_struct* sh)
{
  SH2addi(sh, 2, (0<<4)+5);
}
#define init_7_2_0_6
void FASTCALL call_7_2_0_6(SH2_struct* sh)
{
  SH2addi(sh, 2, (0<<4)+6);
}
#define init_7_2_0_7
void FASTCALL call_7_2_0_7(SH2_struct* sh)
{
  SH2addi(sh, 2, (0<<4)+7);
}
#define init_7_2_0_8
void FASTCALL call_7_2_0_8(SH2_struct* sh)
{
  SH2addi(sh, 2, (0<<4)+8);
}
#define init_7_2_0_9
void FASTCALL call_7_2_0_9(SH2_struct* sh)
{
  SH2addi(sh, 2, (0<<4)+9);
}
#define init_7_2_0_10
void FASTCALL call_7_2_0_10(SH2_struct* sh)
{
  SH2addi(sh, 2, (0<<4)+10);
}
#define init_7_2_0_11
void FASTCALL call_7_2_0_11(SH2_struct* sh)
{
  SH2addi(sh, 2, (0<<4)+11);
}
#define init_7_2_0_12
void FASTCALL call_7_2_0_12(SH2_struct* sh)
{
  SH2addi(sh, 2, (0<<4)+12);
}
#define init_7_2_0_13
void FASTCALL call_7_2_0_13(SH2_struct* sh)
{
  SH2addi(sh, 2, (0<<4)+13);
}
#define init_7_2_0_14
void FASTCALL call_7_2_0_14(SH2_struct* sh)
{
  SH2addi(sh, 2, (0<<4)+14);
}
#define init_7_2_0_15
void FASTCALL call_7_2_0_15(SH2_struct* sh)
{
  SH2addi(sh, 2, (0<<4)+15);
}
#define init_7_2_1_0
void FASTCALL call_7_2_1_0(SH2_struct* sh)
{
  SH2addi(sh, 2, (1<<4)+0);
}
#define init_7_2_1_1
void FASTCALL call_7_2_1_1(SH2_struct* sh)
{
  SH2addi(sh, 2, (1<<4)+1);
}
#define init_7_2_1_2
void FASTCALL call_7_2_1_2(SH2_struct* sh)
{
  SH2addi(sh, 2, (1<<4)+2);
}
#define init_7_2_1_3
void FASTCALL call_7_2_1_3(SH2_struct* sh)
{
  SH2addi(sh, 2, (1<<4)+3);
}
#define init_7_2_1_4
void FASTCALL call_7_2_1_4(SH2_struct* sh)
{
  SH2addi(sh, 2, (1<<4)+4);
}
#define init_7_2_1_5
void FASTCALL call_7_2_1_5(SH2_struct* sh)
{
  SH2addi(sh, 2, (1<<4)+5);
}
#define init_7_2_1_6
void FASTCALL call_7_2_1_6(SH2_struct* sh)
{
  SH2addi(sh, 2, (1<<4)+6);
}
#define init_7_2_1_7
void FASTCALL call_7_2_1_7(SH2_struct* sh)
{
  SH2addi(sh, 2, (1<<4)+7);
}
#define init_7_2_1_8
void FASTCALL call_7_2_1_8(SH2_struct* sh)
{
  SH2addi(sh, 2, (1<<4)+8);
}
#define init_7_2_1_9
void FASTCALL call_7_2_1_9(SH2_struct* sh)
{
  SH2addi(sh, 2, (1<<4)+9);
}
#define init_7_2_1_10
void FASTCALL call_7_2_1_10(SH2_struct* sh)
{
  SH2addi(sh, 2, (1<<4)+10);
}
#define init_7_2_1_11
void FASTCALL call_7_2_1_11(SH2_struct* sh)
{
  SH2addi(sh, 2, (1<<4)+11);
}
#define init_7_2_1_12
void FASTCALL call_7_2_1_12(SH2_struct* sh)
{
  SH2addi(sh, 2, (1<<4)+12);
}
#define init_7_2_1_13
void FASTCALL call_7_2_1_13(SH2_struct* sh)
{
  SH2addi(sh, 2, (1<<4)+13);
}
#define init_7_2_1_14
void FASTCALL call_7_2_1_14(SH2_struct* sh)
{
  SH2addi(sh, 2, (1<<4)+14);
}
#define init_7_2_1_15
void FASTCALL call_7_2_1_15(SH2_struct* sh)
{
  SH2addi(sh, 2, (1<<4)+15);
}
#define init_7_2_2_0
void FASTCALL call_7_2_2_0(SH2_struct* sh)
{
  SH2addi(sh, 2, (2<<4)+0);
}
#define init_7_2_2_1
void FASTCALL call_7_2_2_1(SH2_struct* sh)
{
  SH2addi(sh, 2, (2<<4)+1);
}
#define init_7_2_2_2
void FASTCALL call_7_2_2_2(SH2_struct* sh)
{
  SH2addi(sh, 2, (2<<4)+2);
}
#define init_7_2_2_3
void FASTCALL call_7_2_2_3(SH2_struct* sh)
{
  SH2addi(sh, 2, (2<<4)+3);
}
#define init_7_2_2_4
void FASTCALL call_7_2_2_4(SH2_struct* sh)
{
  SH2addi(sh, 2, (2<<4)+4);
}
#define init_7_2_2_5
void FASTCALL call_7_2_2_5(SH2_struct* sh)
{
  SH2addi(sh, 2, (2<<4)+5);
}
#define init_7_2_2_6
void FASTCALL call_7_2_2_6(SH2_struct* sh)
{
  SH2addi(sh, 2, (2<<4)+6);
}
#define init_7_2_2_7
void FASTCALL call_7_2_2_7(SH2_struct* sh)
{
  SH2addi(sh, 2, (2<<4)+7);
}
#define init_7_2_2_8
void FASTCALL call_7_2_2_8(SH2_struct* sh)
{
  SH2addi(sh, 2, (2<<4)+8);
}
#define init_7_2_2_9
void FASTCALL call_7_2_2_9(SH2_struct* sh)
{
  SH2addi(sh, 2, (2<<4)+9);
}
#define init_7_2_2_10
void FASTCALL call_7_2_2_10(SH2_struct* sh)
{
  SH2addi(sh, 2, (2<<4)+10);
}
#define init_7_2_2_11
void FASTCALL call_7_2_2_11(SH2_struct* sh)
{
  SH2addi(sh, 2, (2<<4)+11);
}
#define init_7_2_2_12
void FASTCALL call_7_2_2_12(SH2_struct* sh)
{
  SH2addi(sh, 2, (2<<4)+12);
}
#define init_7_2_2_13
void FASTCALL call_7_2_2_13(SH2_struct* sh)
{
  SH2addi(sh, 2, (2<<4)+13);
}
#define init_7_2_2_14
void FASTCALL call_7_2_2_14(SH2_struct* sh)
{
  SH2addi(sh, 2, (2<<4)+14);
}
#define init_7_2_2_15
void FASTCALL call_7_2_2_15(SH2_struct* sh)
{
  SH2addi(sh, 2, (2<<4)+15);
}
#define init_7_2_3_0
void FASTCALL call_7_2_3_0(SH2_struct* sh)
{
  SH2addi(sh, 2, (3<<4)+0);
}
#define init_7_2_3_1
void FASTCALL call_7_2_3_1(SH2_struct* sh)
{
  SH2addi(sh, 2, (3<<4)+1);
}
#define init_7_2_3_2
void FASTCALL call_7_2_3_2(SH2_struct* sh)
{
  SH2addi(sh, 2, (3<<4)+2);
}
#define init_7_2_3_3
void FASTCALL call_7_2_3_3(SH2_struct* sh)
{
  SH2addi(sh, 2, (3<<4)+3);
}
#define init_7_2_3_4
void FASTCALL call_7_2_3_4(SH2_struct* sh)
{
  SH2addi(sh, 2, (3<<4)+4);
}
#define init_7_2_3_5
void FASTCALL call_7_2_3_5(SH2_struct* sh)
{
  SH2addi(sh, 2, (3<<4)+5);
}
#define init_7_2_3_6
void FASTCALL call_7_2_3_6(SH2_struct* sh)
{
  SH2addi(sh, 2, (3<<4)+6);
}
#define init_7_2_3_7
void FASTCALL call_7_2_3_7(SH2_struct* sh)
{
  SH2addi(sh, 2, (3<<4)+7);
}
#define init_7_2_3_8
void FASTCALL call_7_2_3_8(SH2_struct* sh)
{
  SH2addi(sh, 2, (3<<4)+8);
}
#define init_7_2_3_9
void FASTCALL call_7_2_3_9(SH2_struct* sh)
{
  SH2addi(sh, 2, (3<<4)+9);
}
#define init_7_2_3_10
void FASTCALL call_7_2_3_10(SH2_struct* sh)
{
  SH2addi(sh, 2, (3<<4)+10);
}
#define init_7_2_3_11
void FASTCALL call_7_2_3_11(SH2_struct* sh)
{
  SH2addi(sh, 2, (3<<4)+11);
}
#define init_7_2_3_12
void FASTCALL call_7_2_3_12(SH2_struct* sh)
{
  SH2addi(sh, 2, (3<<4)+12);
}
#define init_7_2_3_13
void FASTCALL call_7_2_3_13(SH2_struct* sh)
{
  SH2addi(sh, 2, (3<<4)+13);
}
#define init_7_2_3_14
void FASTCALL call_7_2_3_14(SH2_struct* sh)
{
  SH2addi(sh, 2, (3<<4)+14);
}
#define init_7_2_3_15
void FASTCALL call_7_2_3_15(SH2_struct* sh)
{
  SH2addi(sh, 2, (3<<4)+15);
}
#define init_7_2_4_0
void FASTCALL call_7_2_4_0(SH2_struct* sh)
{
  SH2addi(sh, 2, (4<<4)+0);
}
#define init_7_2_4_1
void FASTCALL call_7_2_4_1(SH2_struct* sh)
{
  SH2addi(sh, 2, (4<<4)+1);
}
#define init_7_2_4_2
void FASTCALL call_7_2_4_2(SH2_struct* sh)
{
  SH2addi(sh, 2, (4<<4)+2);
}
#define init_7_2_4_3
void FASTCALL call_7_2_4_3(SH2_struct* sh)
{
  SH2addi(sh, 2, (4<<4)+3);
}
#define init_7_2_4_4
void FASTCALL call_7_2_4_4(SH2_struct* sh)
{
  SH2addi(sh, 2, (4<<4)+4);
}
#define init_7_2_4_5
void FASTCALL call_7_2_4_5(SH2_struct* sh)
{
  SH2addi(sh, 2, (4<<4)+5);
}
#define init_7_2_4_6
void FASTCALL call_7_2_4_6(SH2_struct* sh)
{
  SH2addi(sh, 2, (4<<4)+6);
}
#define init_7_2_4_7
void FASTCALL call_7_2_4_7(SH2_struct* sh)
{
  SH2addi(sh, 2, (4<<4)+7);
}
#define init_7_2_4_8
void FASTCALL call_7_2_4_8(SH2_struct* sh)
{
  SH2addi(sh, 2, (4<<4)+8);
}
#define init_7_2_4_9
void FASTCALL call_7_2_4_9(SH2_struct* sh)
{
  SH2addi(sh, 2, (4<<4)+9);
}
#define init_7_2_4_10
void FASTCALL call_7_2_4_10(SH2_struct* sh)
{
  SH2addi(sh, 2, (4<<4)+10);
}
#define init_7_2_4_11
void FASTCALL call_7_2_4_11(SH2_struct* sh)
{
  SH2addi(sh, 2, (4<<4)+11);
}
#define init_7_2_4_12
void FASTCALL call_7_2_4_12(SH2_struct* sh)
{
  SH2addi(sh, 2, (4<<4)+12);
}
#define init_7_2_4_13
void FASTCALL call_7_2_4_13(SH2_struct* sh)
{
  SH2addi(sh, 2, (4<<4)+13);
}
#define init_7_2_4_14
void FASTCALL call_7_2_4_14(SH2_struct* sh)
{
  SH2addi(sh, 2, (4<<4)+14);
}
#define init_7_2_4_15
void FASTCALL call_7_2_4_15(SH2_struct* sh)
{
  SH2addi(sh, 2, (4<<4)+15);
}
#define init_7_2_5_0
void FASTCALL call_7_2_5_0(SH2_struct* sh)
{
  SH2addi(sh, 2, (5<<4)+0);
}
#define init_7_2_5_1
void FASTCALL call_7_2_5_1(SH2_struct* sh)
{
  SH2addi(sh, 2, (5<<4)+1);
}
#define init_7_2_5_2
void FASTCALL call_7_2_5_2(SH2_struct* sh)
{
  SH2addi(sh, 2, (5<<4)+2);
}
#define init_7_2_5_3
void FASTCALL call_7_2_5_3(SH2_struct* sh)
{
  SH2addi(sh, 2, (5<<4)+3);
}
#define init_7_2_5_4
void FASTCALL call_7_2_5_4(SH2_struct* sh)
{
  SH2addi(sh, 2, (5<<4)+4);
}
#define init_7_2_5_5
void FASTCALL call_7_2_5_5(SH2_struct* sh)
{
  SH2addi(sh, 2, (5<<4)+5);
}
#define init_7_2_5_6
void FASTCALL call_7_2_5_6(SH2_struct* sh)
{
  SH2addi(sh, 2, (5<<4)+6);
}
#define init_7_2_5_7
void FASTCALL call_7_2_5_7(SH2_struct* sh)
{
  SH2addi(sh, 2, (5<<4)+7);
}
#define init_7_2_5_8
void FASTCALL call_7_2_5_8(SH2_struct* sh)
{
  SH2addi(sh, 2, (5<<4)+8);
}
#define init_7_2_5_9
void FASTCALL call_7_2_5_9(SH2_struct* sh)
{
  SH2addi(sh, 2, (5<<4)+9);
}
#define init_7_2_5_10
void FASTCALL call_7_2_5_10(SH2_struct* sh)
{
  SH2addi(sh, 2, (5<<4)+10);
}
#define init_7_2_5_11
void FASTCALL call_7_2_5_11(SH2_struct* sh)
{
  SH2addi(sh, 2, (5<<4)+11);
}
#define init_7_2_5_12
void FASTCALL call_7_2_5_12(SH2_struct* sh)
{
  SH2addi(sh, 2, (5<<4)+12);
}
#define init_7_2_5_13
void FASTCALL call_7_2_5_13(SH2_struct* sh)
{
  SH2addi(sh, 2, (5<<4)+13);
}
#define init_7_2_5_14
void FASTCALL call_7_2_5_14(SH2_struct* sh)
{
  SH2addi(sh, 2, (5<<4)+14);
}
#define init_7_2_5_15
void FASTCALL call_7_2_5_15(SH2_struct* sh)
{
  SH2addi(sh, 2, (5<<4)+15);
}
#define init_7_2_6_0
void FASTCALL call_7_2_6_0(SH2_struct* sh)
{
  SH2addi(sh, 2, (6<<4)+0);
}
#define init_7_2_6_1
void FASTCALL call_7_2_6_1(SH2_struct* sh)
{
  SH2addi(sh, 2, (6<<4)+1);
}
#define init_7_2_6_2
void FASTCALL call_7_2_6_2(SH2_struct* sh)
{
  SH2addi(sh, 2, (6<<4)+2);
}
#define init_7_2_6_3
void FASTCALL call_7_2_6_3(SH2_struct* sh)
{
  SH2addi(sh, 2, (6<<4)+3);
}
#define init_7_2_6_4
void FASTCALL call_7_2_6_4(SH2_struct* sh)
{
  SH2addi(sh, 2, (6<<4)+4);
}
#define init_7_2_6_5
void FASTCALL call_7_2_6_5(SH2_struct* sh)
{
  SH2addi(sh, 2, (6<<4)+5);
}
#define init_7_2_6_6
void FASTCALL call_7_2_6_6(SH2_struct* sh)
{
  SH2addi(sh, 2, (6<<4)+6);
}
#define init_7_2_6_7
void FASTCALL call_7_2_6_7(SH2_struct* sh)
{
  SH2addi(sh, 2, (6<<4)+7);
}
#define init_7_2_6_8
void FASTCALL call_7_2_6_8(SH2_struct* sh)
{
  SH2addi(sh, 2, (6<<4)+8);
}
#define init_7_2_6_9
void FASTCALL call_7_2_6_9(SH2_struct* sh)
{
  SH2addi(sh, 2, (6<<4)+9);
}
#define init_7_2_6_10
void FASTCALL call_7_2_6_10(SH2_struct* sh)
{
  SH2addi(sh, 2, (6<<4)+10);
}
#define init_7_2_6_11
void FASTCALL call_7_2_6_11(SH2_struct* sh)
{
  SH2addi(sh, 2, (6<<4)+11);
}
#define init_7_2_6_12
void FASTCALL call_7_2_6_12(SH2_struct* sh)
{
  SH2addi(sh, 2, (6<<4)+12);
}
#define init_7_2_6_13
void FASTCALL call_7_2_6_13(SH2_struct* sh)
{
  SH2addi(sh, 2, (6<<4)+13);
}
#define init_7_2_6_14
void FASTCALL call_7_2_6_14(SH2_struct* sh)
{
  SH2addi(sh, 2, (6<<4)+14);
}
#define init_7_2_6_15
void FASTCALL call_7_2_6_15(SH2_struct* sh)
{
  SH2addi(sh, 2, (6<<4)+15);
}
#define init_7_2_7_0
void FASTCALL call_7_2_7_0(SH2_struct* sh)
{
  SH2addi(sh, 2, (7<<4)+0);
}
#define init_7_2_7_1
void FASTCALL call_7_2_7_1(SH2_struct* sh)
{
  SH2addi(sh, 2, (7<<4)+1);
}
#define init_7_2_7_2
void FASTCALL call_7_2_7_2(SH2_struct* sh)
{
  SH2addi(sh, 2, (7<<4)+2);
}
#define init_7_2_7_3
void FASTCALL call_7_2_7_3(SH2_struct* sh)
{
  SH2addi(sh, 2, (7<<4)+3);
}
#define init_7_2_7_4
void FASTCALL call_7_2_7_4(SH2_struct* sh)
{
  SH2addi(sh, 2, (7<<4)+4);
}
#define init_7_2_7_5
void FASTCALL call_7_2_7_5(SH2_struct* sh)
{
  SH2addi(sh, 2, (7<<4)+5);
}
#define init_7_2_7_6
void FASTCALL call_7_2_7_6(SH2_struct* sh)
{
  SH2addi(sh, 2, (7<<4)+6);
}
#define init_7_2_7_7
void FASTCALL call_7_2_7_7(SH2_struct* sh)
{
  SH2addi(sh, 2, (7<<4)+7);
}
#define init_7_2_7_8
void FASTCALL call_7_2_7_8(SH2_struct* sh)
{
  SH2addi(sh, 2, (7<<4)+8);
}
#define init_7_2_7_9
void FASTCALL call_7_2_7_9(SH2_struct* sh)
{
  SH2addi(sh, 2, (7<<4)+9);
}
#define init_7_2_7_10
void FASTCALL call_7_2_7_10(SH2_struct* sh)
{
  SH2addi(sh, 2, (7<<4)+10);
}
#define init_7_2_7_11
void FASTCALL call_7_2_7_11(SH2_struct* sh)
{
  SH2addi(sh, 2, (7<<4)+11);
}
#define init_7_2_7_12
void FASTCALL call_7_2_7_12(SH2_struct* sh)
{
  SH2addi(sh, 2, (7<<4)+12);
}
#define init_7_2_7_13
void FASTCALL call_7_2_7_13(SH2_struct* sh)
{
  SH2addi(sh, 2, (7<<4)+13);
}
#define init_7_2_7_14
void FASTCALL call_7_2_7_14(SH2_struct* sh)
{
  SH2addi(sh, 2, (7<<4)+14);
}
#define init_7_2_7_15
void FASTCALL call_7_2_7_15(SH2_struct* sh)
{
  SH2addi(sh, 2, (7<<4)+15);
}
#define init_7_2_8_0
void FASTCALL call_7_2_8_0(SH2_struct* sh)
{
  SH2addi(sh, 2, (8<<4)+0);
}
#define init_7_2_8_1
void FASTCALL call_7_2_8_1(SH2_struct* sh)
{
  SH2addi(sh, 2, (8<<4)+1);
}
#define init_7_2_8_2
void FASTCALL call_7_2_8_2(SH2_struct* sh)
{
  SH2addi(sh, 2, (8<<4)+2);
}
#define init_7_2_8_3
void FASTCALL call_7_2_8_3(SH2_struct* sh)
{
  SH2addi(sh, 2, (8<<4)+3);
}
#define init_7_2_8_4
void FASTCALL call_7_2_8_4(SH2_struct* sh)
{
  SH2addi(sh, 2, (8<<4)+4);
}
#define init_7_2_8_5
void FASTCALL call_7_2_8_5(SH2_struct* sh)
{
  SH2addi(sh, 2, (8<<4)+5);
}
#define init_7_2_8_6
void FASTCALL call_7_2_8_6(SH2_struct* sh)
{
  SH2addi(sh, 2, (8<<4)+6);
}
#define init_7_2_8_7
void FASTCALL call_7_2_8_7(SH2_struct* sh)
{
  SH2addi(sh, 2, (8<<4)+7);
}
#define init_7_2_8_8
void FASTCALL call_7_2_8_8(SH2_struct* sh)
{
  SH2addi(sh, 2, (8<<4)+8);
}
#define init_7_2_8_9
void FASTCALL call_7_2_8_9(SH2_struct* sh)
{
  SH2addi(sh, 2, (8<<4)+9);
}
#define init_7_2_8_10
void FASTCALL call_7_2_8_10(SH2_struct* sh)
{
  SH2addi(sh, 2, (8<<4)+10);
}
#define init_7_2_8_11
void FASTCALL call_7_2_8_11(SH2_struct* sh)
{
  SH2addi(sh, 2, (8<<4)+11);
}
#define init_7_2_8_12
void FASTCALL call_7_2_8_12(SH2_struct* sh)
{
  SH2addi(sh, 2, (8<<4)+12);
}
#define init_7_2_8_13
void FASTCALL call_7_2_8_13(SH2_struct* sh)
{
  SH2addi(sh, 2, (8<<4)+13);
}
#define init_7_2_8_14
void FASTCALL call_7_2_8_14(SH2_struct* sh)
{
  SH2addi(sh, 2, (8<<4)+14);
}
#define init_7_2_8_15
void FASTCALL call_7_2_8_15(SH2_struct* sh)
{
  SH2addi(sh, 2, (8<<4)+15);
}
#define init_7_2_9_0
void FASTCALL call_7_2_9_0(SH2_struct* sh)
{
  SH2addi(sh, 2, (9<<4)+0);
}
#define init_7_2_9_1
void FASTCALL call_7_2_9_1(SH2_struct* sh)
{
  SH2addi(sh, 2, (9<<4)+1);
}
#define init_7_2_9_2
void FASTCALL call_7_2_9_2(SH2_struct* sh)
{
  SH2addi(sh, 2, (9<<4)+2);
}
#define init_7_2_9_3
void FASTCALL call_7_2_9_3(SH2_struct* sh)
{
  SH2addi(sh, 2, (9<<4)+3);
}
#define init_7_2_9_4
void FASTCALL call_7_2_9_4(SH2_struct* sh)
{
  SH2addi(sh, 2, (9<<4)+4);
}
#define init_7_2_9_5
void FASTCALL call_7_2_9_5(SH2_struct* sh)
{
  SH2addi(sh, 2, (9<<4)+5);
}
#define init_7_2_9_6
void FASTCALL call_7_2_9_6(SH2_struct* sh)
{
  SH2addi(sh, 2, (9<<4)+6);
}
#define init_7_2_9_7
void FASTCALL call_7_2_9_7(SH2_struct* sh)
{
  SH2addi(sh, 2, (9<<4)+7);
}
#define init_7_2_9_8
void FASTCALL call_7_2_9_8(SH2_struct* sh)
{
  SH2addi(sh, 2, (9<<4)+8);
}
#define init_7_2_9_9
void FASTCALL call_7_2_9_9(SH2_struct* sh)
{
  SH2addi(sh, 2, (9<<4)+9);
}
#define init_7_2_9_10
void FASTCALL call_7_2_9_10(SH2_struct* sh)
{
  SH2addi(sh, 2, (9<<4)+10);
}
#define init_7_2_9_11
void FASTCALL call_7_2_9_11(SH2_struct* sh)
{
  SH2addi(sh, 2, (9<<4)+11);
}
#define init_7_2_9_12
void FASTCALL call_7_2_9_12(SH2_struct* sh)
{
  SH2addi(sh, 2, (9<<4)+12);
}
#define init_7_2_9_13
void FASTCALL call_7_2_9_13(SH2_struct* sh)
{
  SH2addi(sh, 2, (9<<4)+13);
}
#define init_7_2_9_14
void FASTCALL call_7_2_9_14(SH2_struct* sh)
{
  SH2addi(sh, 2, (9<<4)+14);
}
#define init_7_2_9_15
void FASTCALL call_7_2_9_15(SH2_struct* sh)
{
  SH2addi(sh, 2, (9<<4)+15);
}
#define init_7_2_10_0
void FASTCALL call_7_2_10_0(SH2_struct* sh)
{
  SH2addi(sh, 2, (10<<4)+0);
}
#define init_7_2_10_1
void FASTCALL call_7_2_10_1(SH2_struct* sh)
{
  SH2addi(sh, 2, (10<<4)+1);
}
#define init_7_2_10_2
void FASTCALL call_7_2_10_2(SH2_struct* sh)
{
  SH2addi(sh, 2, (10<<4)+2);
}
#define init_7_2_10_3
void FASTCALL call_7_2_10_3(SH2_struct* sh)
{
  SH2addi(sh, 2, (10<<4)+3);
}
#define init_7_2_10_4
void FASTCALL call_7_2_10_4(SH2_struct* sh)
{
  SH2addi(sh, 2, (10<<4)+4);
}
#define init_7_2_10_5
void FASTCALL call_7_2_10_5(SH2_struct* sh)
{
  SH2addi(sh, 2, (10<<4)+5);
}
#define init_7_2_10_6
void FASTCALL call_7_2_10_6(SH2_struct* sh)
{
  SH2addi(sh, 2, (10<<4)+6);
}
#define init_7_2_10_7
void FASTCALL call_7_2_10_7(SH2_struct* sh)
{
  SH2addi(sh, 2, (10<<4)+7);
}
#define init_7_2_10_8
void FASTCALL call_7_2_10_8(SH2_struct* sh)
{
  SH2addi(sh, 2, (10<<4)+8);
}
#define init_7_2_10_9
void FASTCALL call_7_2_10_9(SH2_struct* sh)
{
  SH2addi(sh, 2, (10<<4)+9);
}
#define init_7_2_10_10
void FASTCALL call_7_2_10_10(SH2_struct* sh)
{
  SH2addi(sh, 2, (10<<4)+10);
}
#define init_7_2_10_11
void FASTCALL call_7_2_10_11(SH2_struct* sh)
{
  SH2addi(sh, 2, (10<<4)+11);
}
#define init_7_2_10_12
void FASTCALL call_7_2_10_12(SH2_struct* sh)
{
  SH2addi(sh, 2, (10<<4)+12);
}
#define init_7_2_10_13
void FASTCALL call_7_2_10_13(SH2_struct* sh)
{
  SH2addi(sh, 2, (10<<4)+13);
}
#define init_7_2_10_14
void FASTCALL call_7_2_10_14(SH2_struct* sh)
{
  SH2addi(sh, 2, (10<<4)+14);
}
#define init_7_2_10_15
void FASTCALL call_7_2_10_15(SH2_struct* sh)
{
  SH2addi(sh, 2, (10<<4)+15);
}
#define init_7_2_11_0
void FASTCALL call_7_2_11_0(SH2_struct* sh)
{
  SH2addi(sh, 2, (11<<4)+0);
}
#define init_7_2_11_1
void FASTCALL call_7_2_11_1(SH2_struct* sh)
{
  SH2addi(sh, 2, (11<<4)+1);
}
#define init_7_2_11_2
void FASTCALL call_7_2_11_2(SH2_struct* sh)
{
  SH2addi(sh, 2, (11<<4)+2);
}
#define init_7_2_11_3
void FASTCALL call_7_2_11_3(SH2_struct* sh)
{
  SH2addi(sh, 2, (11<<4)+3);
}
#define init_7_2_11_4
void FASTCALL call_7_2_11_4(SH2_struct* sh)
{
  SH2addi(sh, 2, (11<<4)+4);
}
#define init_7_2_11_5
void FASTCALL call_7_2_11_5(SH2_struct* sh)
{
  SH2addi(sh, 2, (11<<4)+5);
}
#define init_7_2_11_6
void FASTCALL call_7_2_11_6(SH2_struct* sh)
{
  SH2addi(sh, 2, (11<<4)+6);
}
#define init_7_2_11_7
void FASTCALL call_7_2_11_7(SH2_struct* sh)
{
  SH2addi(sh, 2, (11<<4)+7);
}
#define init_7_2_11_8
void FASTCALL call_7_2_11_8(SH2_struct* sh)
{
  SH2addi(sh, 2, (11<<4)+8);
}
#define init_7_2_11_9
void FASTCALL call_7_2_11_9(SH2_struct* sh)
{
  SH2addi(sh, 2, (11<<4)+9);
}
#define init_7_2_11_10
void FASTCALL call_7_2_11_10(SH2_struct* sh)
{
  SH2addi(sh, 2, (11<<4)+10);
}
#define init_7_2_11_11
void FASTCALL call_7_2_11_11(SH2_struct* sh)
{
  SH2addi(sh, 2, (11<<4)+11);
}
#define init_7_2_11_12
void FASTCALL call_7_2_11_12(SH2_struct* sh)
{
  SH2addi(sh, 2, (11<<4)+12);
}
#define init_7_2_11_13
void FASTCALL call_7_2_11_13(SH2_struct* sh)
{
  SH2addi(sh, 2, (11<<4)+13);
}
#define init_7_2_11_14
void FASTCALL call_7_2_11_14(SH2_struct* sh)
{
  SH2addi(sh, 2, (11<<4)+14);
}
#define init_7_2_11_15
void FASTCALL call_7_2_11_15(SH2_struct* sh)
{
  SH2addi(sh, 2, (11<<4)+15);
}
#define init_7_2_12_0
void FASTCALL call_7_2_12_0(SH2_struct* sh)
{
  SH2addi(sh, 2, (12<<4)+0);
}
#define init_7_2_12_1
void FASTCALL call_7_2_12_1(SH2_struct* sh)
{
  SH2addi(sh, 2, (12<<4)+1);
}
#define init_7_2_12_2
void FASTCALL call_7_2_12_2(SH2_struct* sh)
{
  SH2addi(sh, 2, (12<<4)+2);
}
#define init_7_2_12_3
void FASTCALL call_7_2_12_3(SH2_struct* sh)
{
  SH2addi(sh, 2, (12<<4)+3);
}
#define init_7_2_12_4
void FASTCALL call_7_2_12_4(SH2_struct* sh)
{
  SH2addi(sh, 2, (12<<4)+4);
}
#define init_7_2_12_5
void FASTCALL call_7_2_12_5(SH2_struct* sh)
{
  SH2addi(sh, 2, (12<<4)+5);
}
#define init_7_2_12_6
void FASTCALL call_7_2_12_6(SH2_struct* sh)
{
  SH2addi(sh, 2, (12<<4)+6);
}
#define init_7_2_12_7
void FASTCALL call_7_2_12_7(SH2_struct* sh)
{
  SH2addi(sh, 2, (12<<4)+7);
}
#define init_7_2_12_8
void FASTCALL call_7_2_12_8(SH2_struct* sh)
{
  SH2addi(sh, 2, (12<<4)+8);
}
#define init_7_2_12_9
void FASTCALL call_7_2_12_9(SH2_struct* sh)
{
  SH2addi(sh, 2, (12<<4)+9);
}
#define init_7_2_12_10
void FASTCALL call_7_2_12_10(SH2_struct* sh)
{
  SH2addi(sh, 2, (12<<4)+10);
}
#define init_7_2_12_11
void FASTCALL call_7_2_12_11(SH2_struct* sh)
{
  SH2addi(sh, 2, (12<<4)+11);
}
#define init_7_2_12_12
void FASTCALL call_7_2_12_12(SH2_struct* sh)
{
  SH2addi(sh, 2, (12<<4)+12);
}
#define init_7_2_12_13
void FASTCALL call_7_2_12_13(SH2_struct* sh)
{
  SH2addi(sh, 2, (12<<4)+13);
}
#define init_7_2_12_14
void FASTCALL call_7_2_12_14(SH2_struct* sh)
{
  SH2addi(sh, 2, (12<<4)+14);
}
#define init_7_2_12_15
void FASTCALL call_7_2_12_15(SH2_struct* sh)
{
  SH2addi(sh, 2, (12<<4)+15);
}
#define init_7_2_13_0
void FASTCALL call_7_2_13_0(SH2_struct* sh)
{
  SH2addi(sh, 2, (13<<4)+0);
}
#define init_7_2_13_1
void FASTCALL call_7_2_13_1(SH2_struct* sh)
{
  SH2addi(sh, 2, (13<<4)+1);
}
#define init_7_2_13_2
void FASTCALL call_7_2_13_2(SH2_struct* sh)
{
  SH2addi(sh, 2, (13<<4)+2);
}
#define init_7_2_13_3
void FASTCALL call_7_2_13_3(SH2_struct* sh)
{
  SH2addi(sh, 2, (13<<4)+3);
}
#define init_7_2_13_4
void FASTCALL call_7_2_13_4(SH2_struct* sh)
{
  SH2addi(sh, 2, (13<<4)+4);
}
#define init_7_2_13_5
void FASTCALL call_7_2_13_5(SH2_struct* sh)
{
  SH2addi(sh, 2, (13<<4)+5);
}
#define init_7_2_13_6
void FASTCALL call_7_2_13_6(SH2_struct* sh)
{
  SH2addi(sh, 2, (13<<4)+6);
}
#define init_7_2_13_7
void FASTCALL call_7_2_13_7(SH2_struct* sh)
{
  SH2addi(sh, 2, (13<<4)+7);
}
#define init_7_2_13_8
void FASTCALL call_7_2_13_8(SH2_struct* sh)
{
  SH2addi(sh, 2, (13<<4)+8);
}
#define init_7_2_13_9
void FASTCALL call_7_2_13_9(SH2_struct* sh)
{
  SH2addi(sh, 2, (13<<4)+9);
}
#define init_7_2_13_10
void FASTCALL call_7_2_13_10(SH2_struct* sh)
{
  SH2addi(sh, 2, (13<<4)+10);
}
#define init_7_2_13_11
void FASTCALL call_7_2_13_11(SH2_struct* sh)
{
  SH2addi(sh, 2, (13<<4)+11);
}
#define init_7_2_13_12
void FASTCALL call_7_2_13_12(SH2_struct* sh)
{
  SH2addi(sh, 2, (13<<4)+12);
}
#define init_7_2_13_13
void FASTCALL call_7_2_13_13(SH2_struct* sh)
{
  SH2addi(sh, 2, (13<<4)+13);
}
#define init_7_2_13_14
void FASTCALL call_7_2_13_14(SH2_struct* sh)
{
  SH2addi(sh, 2, (13<<4)+14);
}
#define init_7_2_13_15
void FASTCALL call_7_2_13_15(SH2_struct* sh)
{
  SH2addi(sh, 2, (13<<4)+15);
}
#define init_7_2_14_0
void FASTCALL call_7_2_14_0(SH2_struct* sh)
{
  SH2addi(sh, 2, (14<<4)+0);
}
#define init_7_2_14_1
void FASTCALL call_7_2_14_1(SH2_struct* sh)
{
  SH2addi(sh, 2, (14<<4)+1);
}
#define init_7_2_14_2
void FASTCALL call_7_2_14_2(SH2_struct* sh)
{
  SH2addi(sh, 2, (14<<4)+2);
}
#define init_7_2_14_3
void FASTCALL call_7_2_14_3(SH2_struct* sh)
{
  SH2addi(sh, 2, (14<<4)+3);
}
#define init_7_2_14_4
void FASTCALL call_7_2_14_4(SH2_struct* sh)
{
  SH2addi(sh, 2, (14<<4)+4);
}
#define init_7_2_14_5
void FASTCALL call_7_2_14_5(SH2_struct* sh)
{
  SH2addi(sh, 2, (14<<4)+5);
}
#define init_7_2_14_6
void FASTCALL call_7_2_14_6(SH2_struct* sh)
{
  SH2addi(sh, 2, (14<<4)+6);
}
#define init_7_2_14_7
void FASTCALL call_7_2_14_7(SH2_struct* sh)
{
  SH2addi(sh, 2, (14<<4)+7);
}
#define init_7_2_14_8
void FASTCALL call_7_2_14_8(SH2_struct* sh)
{
  SH2addi(sh, 2, (14<<4)+8);
}
#define init_7_2_14_9
void FASTCALL call_7_2_14_9(SH2_struct* sh)
{
  SH2addi(sh, 2, (14<<4)+9);
}
#define init_7_2_14_10
void FASTCALL call_7_2_14_10(SH2_struct* sh)
{
  SH2addi(sh, 2, (14<<4)+10);
}
#define init_7_2_14_11
void FASTCALL call_7_2_14_11(SH2_struct* sh)
{
  SH2addi(sh, 2, (14<<4)+11);
}
#define init_7_2_14_12
void FASTCALL call_7_2_14_12(SH2_struct* sh)
{
  SH2addi(sh, 2, (14<<4)+12);
}
#define init_7_2_14_13
void FASTCALL call_7_2_14_13(SH2_struct* sh)
{
  SH2addi(sh, 2, (14<<4)+13);
}
#define init_7_2_14_14
void FASTCALL call_7_2_14_14(SH2_struct* sh)
{
  SH2addi(sh, 2, (14<<4)+14);
}
#define init_7_2_14_15
void FASTCALL call_7_2_14_15(SH2_struct* sh)
{
  SH2addi(sh, 2, (14<<4)+15);
}
#define init_7_2_15_0
void FASTCALL call_7_2_15_0(SH2_struct* sh)
{
  SH2addi(sh, 2, (15<<4)+0);
}
#define init_7_2_15_1
void FASTCALL call_7_2_15_1(SH2_struct* sh)
{
  SH2addi(sh, 2, (15<<4)+1);
}
#define init_7_2_15_2
void FASTCALL call_7_2_15_2(SH2_struct* sh)
{
  SH2addi(sh, 2, (15<<4)+2);
}
#define init_7_2_15_3
void FASTCALL call_7_2_15_3(SH2_struct* sh)
{
  SH2addi(sh, 2, (15<<4)+3);
}
#define init_7_2_15_4
void FASTCALL call_7_2_15_4(SH2_struct* sh)
{
  SH2addi(sh, 2, (15<<4)+4);
}
#define init_7_2_15_5
void FASTCALL call_7_2_15_5(SH2_struct* sh)
{
  SH2addi(sh, 2, (15<<4)+5);
}
#define init_7_2_15_6
void FASTCALL call_7_2_15_6(SH2_struct* sh)
{
  SH2addi(sh, 2, (15<<4)+6);
}
#define init_7_2_15_7
void FASTCALL call_7_2_15_7(SH2_struct* sh)
{
  SH2addi(sh, 2, (15<<4)+7);
}
#define init_7_2_15_8
void FASTCALL call_7_2_15_8(SH2_struct* sh)
{
  SH2addi(sh, 2, (15<<4)+8);
}
#define init_7_2_15_9
void FASTCALL call_7_2_15_9(SH2_struct* sh)
{
  SH2addi(sh, 2, (15<<4)+9);
}
#define init_7_2_15_10
void FASTCALL call_7_2_15_10(SH2_struct* sh)
{
  SH2addi(sh, 2, (15<<4)+10);
}
#define init_7_2_15_11
void FASTCALL call_7_2_15_11(SH2_struct* sh)
{
  SH2addi(sh, 2, (15<<4)+11);
}
#define init_7_2_15_12
void FASTCALL call_7_2_15_12(SH2_struct* sh)
{
  SH2addi(sh, 2, (15<<4)+12);
}
#define init_7_2_15_13
void FASTCALL call_7_2_15_13(SH2_struct* sh)
{
  SH2addi(sh, 2, (15<<4)+13);
}
#define init_7_2_15_14
void FASTCALL call_7_2_15_14(SH2_struct* sh)
{
  SH2addi(sh, 2, (15<<4)+14);
}
#define init_7_2_15_15
void FASTCALL call_7_2_15_15(SH2_struct* sh)
{
  SH2addi(sh, 2, (15<<4)+15);
}
#define init_7_3_0_0
void FASTCALL call_7_3_0_0(SH2_struct* sh)
{
  SH2addi(sh, 3, (0<<4)+0);
}
#define init_7_3_0_1
void FASTCALL call_7_3_0_1(SH2_struct* sh)
{
  SH2addi(sh, 3, (0<<4)+1);
}
#define init_7_3_0_2
void FASTCALL call_7_3_0_2(SH2_struct* sh)
{
  SH2addi(sh, 3, (0<<4)+2);
}
#define init_7_3_0_3
void FASTCALL call_7_3_0_3(SH2_struct* sh)
{
  SH2addi(sh, 3, (0<<4)+3);
}
#define init_7_3_0_4
void FASTCALL call_7_3_0_4(SH2_struct* sh)
{
  SH2addi(sh, 3, (0<<4)+4);
}
#define init_7_3_0_5
void FASTCALL call_7_3_0_5(SH2_struct* sh)
{
  SH2addi(sh, 3, (0<<4)+5);
}
#define init_7_3_0_6
void FASTCALL call_7_3_0_6(SH2_struct* sh)
{
  SH2addi(sh, 3, (0<<4)+6);
}
#define init_7_3_0_7
void FASTCALL call_7_3_0_7(SH2_struct* sh)
{
  SH2addi(sh, 3, (0<<4)+7);
}
#define init_7_3_0_8
void FASTCALL call_7_3_0_8(SH2_struct* sh)
{
  SH2addi(sh, 3, (0<<4)+8);
}
#define init_7_3_0_9
void FASTCALL call_7_3_0_9(SH2_struct* sh)
{
  SH2addi(sh, 3, (0<<4)+9);
}
#define init_7_3_0_10
void FASTCALL call_7_3_0_10(SH2_struct* sh)
{
  SH2addi(sh, 3, (0<<4)+10);
}
#define init_7_3_0_11
void FASTCALL call_7_3_0_11(SH2_struct* sh)
{
  SH2addi(sh, 3, (0<<4)+11);
}
#define init_7_3_0_12
void FASTCALL call_7_3_0_12(SH2_struct* sh)
{
  SH2addi(sh, 3, (0<<4)+12);
}
#define init_7_3_0_13
void FASTCALL call_7_3_0_13(SH2_struct* sh)
{
  SH2addi(sh, 3, (0<<4)+13);
}
#define init_7_3_0_14
void FASTCALL call_7_3_0_14(SH2_struct* sh)
{
  SH2addi(sh, 3, (0<<4)+14);
}
#define init_7_3_0_15
void FASTCALL call_7_3_0_15(SH2_struct* sh)
{
  SH2addi(sh, 3, (0<<4)+15);
}
#define init_7_3_1_0
void FASTCALL call_7_3_1_0(SH2_struct* sh)
{
  SH2addi(sh, 3, (1<<4)+0);
}
#define init_7_3_1_1
void FASTCALL call_7_3_1_1(SH2_struct* sh)
{
  SH2addi(sh, 3, (1<<4)+1);
}
#define init_7_3_1_2
void FASTCALL call_7_3_1_2(SH2_struct* sh)
{
  SH2addi(sh, 3, (1<<4)+2);
}
#define init_7_3_1_3
void FASTCALL call_7_3_1_3(SH2_struct* sh)
{
  SH2addi(sh, 3, (1<<4)+3);
}
#define init_7_3_1_4
void FASTCALL call_7_3_1_4(SH2_struct* sh)
{
  SH2addi(sh, 3, (1<<4)+4);
}
#define init_7_3_1_5
void FASTCALL call_7_3_1_5(SH2_struct* sh)
{
  SH2addi(sh, 3, (1<<4)+5);
}
#define init_7_3_1_6
void FASTCALL call_7_3_1_6(SH2_struct* sh)
{
  SH2addi(sh, 3, (1<<4)+6);
}
#define init_7_3_1_7
void FASTCALL call_7_3_1_7(SH2_struct* sh)
{
  SH2addi(sh, 3, (1<<4)+7);
}
#define init_7_3_1_8
void FASTCALL call_7_3_1_8(SH2_struct* sh)
{
  SH2addi(sh, 3, (1<<4)+8);
}
#define init_7_3_1_9
void FASTCALL call_7_3_1_9(SH2_struct* sh)
{
  SH2addi(sh, 3, (1<<4)+9);
}
#define init_7_3_1_10
void FASTCALL call_7_3_1_10(SH2_struct* sh)
{
  SH2addi(sh, 3, (1<<4)+10);
}
#define init_7_3_1_11
void FASTCALL call_7_3_1_11(SH2_struct* sh)
{
  SH2addi(sh, 3, (1<<4)+11);
}
#define init_7_3_1_12
void FASTCALL call_7_3_1_12(SH2_struct* sh)
{
  SH2addi(sh, 3, (1<<4)+12);
}
#define init_7_3_1_13
void FASTCALL call_7_3_1_13(SH2_struct* sh)
{
  SH2addi(sh, 3, (1<<4)+13);
}
#define init_7_3_1_14
void FASTCALL call_7_3_1_14(SH2_struct* sh)
{
  SH2addi(sh, 3, (1<<4)+14);
}
#define init_7_3_1_15
void FASTCALL call_7_3_1_15(SH2_struct* sh)
{
  SH2addi(sh, 3, (1<<4)+15);
}
#define init_7_3_2_0
void FASTCALL call_7_3_2_0(SH2_struct* sh)
{
  SH2addi(sh, 3, (2<<4)+0);
}
#define init_7_3_2_1
void FASTCALL call_7_3_2_1(SH2_struct* sh)
{
  SH2addi(sh, 3, (2<<4)+1);
}
#define init_7_3_2_2
void FASTCALL call_7_3_2_2(SH2_struct* sh)
{
  SH2addi(sh, 3, (2<<4)+2);
}
#define init_7_3_2_3
void FASTCALL call_7_3_2_3(SH2_struct* sh)
{
  SH2addi(sh, 3, (2<<4)+3);
}
#define init_7_3_2_4
void FASTCALL call_7_3_2_4(SH2_struct* sh)
{
  SH2addi(sh, 3, (2<<4)+4);
}
#define init_7_3_2_5
void FASTCALL call_7_3_2_5(SH2_struct* sh)
{
  SH2addi(sh, 3, (2<<4)+5);
}
#define init_7_3_2_6
void FASTCALL call_7_3_2_6(SH2_struct* sh)
{
  SH2addi(sh, 3, (2<<4)+6);
}
#define init_7_3_2_7
void FASTCALL call_7_3_2_7(SH2_struct* sh)
{
  SH2addi(sh, 3, (2<<4)+7);
}
#define init_7_3_2_8
void FASTCALL call_7_3_2_8(SH2_struct* sh)
{
  SH2addi(sh, 3, (2<<4)+8);
}
#define init_7_3_2_9
void FASTCALL call_7_3_2_9(SH2_struct* sh)
{
  SH2addi(sh, 3, (2<<4)+9);
}
#define init_7_3_2_10
void FASTCALL call_7_3_2_10(SH2_struct* sh)
{
  SH2addi(sh, 3, (2<<4)+10);
}
#define init_7_3_2_11
void FASTCALL call_7_3_2_11(SH2_struct* sh)
{
  SH2addi(sh, 3, (2<<4)+11);
}
#define init_7_3_2_12
void FASTCALL call_7_3_2_12(SH2_struct* sh)
{
  SH2addi(sh, 3, (2<<4)+12);
}
#define init_7_3_2_13
void FASTCALL call_7_3_2_13(SH2_struct* sh)
{
  SH2addi(sh, 3, (2<<4)+13);
}
#define init_7_3_2_14
void FASTCALL call_7_3_2_14(SH2_struct* sh)
{
  SH2addi(sh, 3, (2<<4)+14);
}
#define init_7_3_2_15
void FASTCALL call_7_3_2_15(SH2_struct* sh)
{
  SH2addi(sh, 3, (2<<4)+15);
}
#define init_7_3_3_0
void FASTCALL call_7_3_3_0(SH2_struct* sh)
{
  SH2addi(sh, 3, (3<<4)+0);
}
#define init_7_3_3_1
void FASTCALL call_7_3_3_1(SH2_struct* sh)
{
  SH2addi(sh, 3, (3<<4)+1);
}
#define init_7_3_3_2
void FASTCALL call_7_3_3_2(SH2_struct* sh)
{
  SH2addi(sh, 3, (3<<4)+2);
}
#define init_7_3_3_3
void FASTCALL call_7_3_3_3(SH2_struct* sh)
{
  SH2addi(sh, 3, (3<<4)+3);
}
#define init_7_3_3_4
void FASTCALL call_7_3_3_4(SH2_struct* sh)
{
  SH2addi(sh, 3, (3<<4)+4);
}
#define init_7_3_3_5
void FASTCALL call_7_3_3_5(SH2_struct* sh)
{
  SH2addi(sh, 3, (3<<4)+5);
}
#define init_7_3_3_6
void FASTCALL call_7_3_3_6(SH2_struct* sh)
{
  SH2addi(sh, 3, (3<<4)+6);
}
#define init_7_3_3_7
void FASTCALL call_7_3_3_7(SH2_struct* sh)
{
  SH2addi(sh, 3, (3<<4)+7);
}
#define init_7_3_3_8
void FASTCALL call_7_3_3_8(SH2_struct* sh)
{
  SH2addi(sh, 3, (3<<4)+8);
}
#define init_7_3_3_9
void FASTCALL call_7_3_3_9(SH2_struct* sh)
{
  SH2addi(sh, 3, (3<<4)+9);
}
#define init_7_3_3_10
void FASTCALL call_7_3_3_10(SH2_struct* sh)
{
  SH2addi(sh, 3, (3<<4)+10);
}
#define init_7_3_3_11
void FASTCALL call_7_3_3_11(SH2_struct* sh)
{
  SH2addi(sh, 3, (3<<4)+11);
}
#define init_7_3_3_12
void FASTCALL call_7_3_3_12(SH2_struct* sh)
{
  SH2addi(sh, 3, (3<<4)+12);
}
#define init_7_3_3_13
void FASTCALL call_7_3_3_13(SH2_struct* sh)
{
  SH2addi(sh, 3, (3<<4)+13);
}
#define init_7_3_3_14
void FASTCALL call_7_3_3_14(SH2_struct* sh)
{
  SH2addi(sh, 3, (3<<4)+14);
}
#define init_7_3_3_15
void FASTCALL call_7_3_3_15(SH2_struct* sh)
{
  SH2addi(sh, 3, (3<<4)+15);
}
#define init_7_3_4_0
void FASTCALL call_7_3_4_0(SH2_struct* sh)
{
  SH2addi(sh, 3, (4<<4)+0);
}
#define init_7_3_4_1
void FASTCALL call_7_3_4_1(SH2_struct* sh)
{
  SH2addi(sh, 3, (4<<4)+1);
}
#define init_7_3_4_2
void FASTCALL call_7_3_4_2(SH2_struct* sh)
{
  SH2addi(sh, 3, (4<<4)+2);
}
#define init_7_3_4_3
void FASTCALL call_7_3_4_3(SH2_struct* sh)
{
  SH2addi(sh, 3, (4<<4)+3);
}
#define init_7_3_4_4
void FASTCALL call_7_3_4_4(SH2_struct* sh)
{
  SH2addi(sh, 3, (4<<4)+4);
}
#define init_7_3_4_5
void FASTCALL call_7_3_4_5(SH2_struct* sh)
{
  SH2addi(sh, 3, (4<<4)+5);
}
#define init_7_3_4_6
void FASTCALL call_7_3_4_6(SH2_struct* sh)
{
  SH2addi(sh, 3, (4<<4)+6);
}
#define init_7_3_4_7
void FASTCALL call_7_3_4_7(SH2_struct* sh)
{
  SH2addi(sh, 3, (4<<4)+7);
}
#define init_7_3_4_8
void FASTCALL call_7_3_4_8(SH2_struct* sh)
{
  SH2addi(sh, 3, (4<<4)+8);
}
#define init_7_3_4_9
void FASTCALL call_7_3_4_9(SH2_struct* sh)
{
  SH2addi(sh, 3, (4<<4)+9);
}
#define init_7_3_4_10
void FASTCALL call_7_3_4_10(SH2_struct* sh)
{
  SH2addi(sh, 3, (4<<4)+10);
}
#define init_7_3_4_11
void FASTCALL call_7_3_4_11(SH2_struct* sh)
{
  SH2addi(sh, 3, (4<<4)+11);
}
#define init_7_3_4_12
void FASTCALL call_7_3_4_12(SH2_struct* sh)
{
  SH2addi(sh, 3, (4<<4)+12);
}
#define init_7_3_4_13
void FASTCALL call_7_3_4_13(SH2_struct* sh)
{
  SH2addi(sh, 3, (4<<4)+13);
}
#define init_7_3_4_14
void FASTCALL call_7_3_4_14(SH2_struct* sh)
{
  SH2addi(sh, 3, (4<<4)+14);
}
#define init_7_3_4_15
void FASTCALL call_7_3_4_15(SH2_struct* sh)
{
  SH2addi(sh, 3, (4<<4)+15);
}
#define init_7_3_5_0
void FASTCALL call_7_3_5_0(SH2_struct* sh)
{
  SH2addi(sh, 3, (5<<4)+0);
}
#define init_7_3_5_1
void FASTCALL call_7_3_5_1(SH2_struct* sh)
{
  SH2addi(sh, 3, (5<<4)+1);
}
#define init_7_3_5_2
void FASTCALL call_7_3_5_2(SH2_struct* sh)
{
  SH2addi(sh, 3, (5<<4)+2);
}
#define init_7_3_5_3
void FASTCALL call_7_3_5_3(SH2_struct* sh)
{
  SH2addi(sh, 3, (5<<4)+3);
}
#define init_7_3_5_4
void FASTCALL call_7_3_5_4(SH2_struct* sh)
{
  SH2addi(sh, 3, (5<<4)+4);
}
#define init_7_3_5_5
void FASTCALL call_7_3_5_5(SH2_struct* sh)
{
  SH2addi(sh, 3, (5<<4)+5);
}
#define init_7_3_5_6
void FASTCALL call_7_3_5_6(SH2_struct* sh)
{
  SH2addi(sh, 3, (5<<4)+6);
}
#define init_7_3_5_7
void FASTCALL call_7_3_5_7(SH2_struct* sh)
{
  SH2addi(sh, 3, (5<<4)+7);
}
#define init_7_3_5_8
void FASTCALL call_7_3_5_8(SH2_struct* sh)
{
  SH2addi(sh, 3, (5<<4)+8);
}
#define init_7_3_5_9
void FASTCALL call_7_3_5_9(SH2_struct* sh)
{
  SH2addi(sh, 3, (5<<4)+9);
}
#define init_7_3_5_10
void FASTCALL call_7_3_5_10(SH2_struct* sh)
{
  SH2addi(sh, 3, (5<<4)+10);
}
#define init_7_3_5_11
void FASTCALL call_7_3_5_11(SH2_struct* sh)
{
  SH2addi(sh, 3, (5<<4)+11);
}
#define init_7_3_5_12
void FASTCALL call_7_3_5_12(SH2_struct* sh)
{
  SH2addi(sh, 3, (5<<4)+12);
}
#define init_7_3_5_13
void FASTCALL call_7_3_5_13(SH2_struct* sh)
{
  SH2addi(sh, 3, (5<<4)+13);
}
#define init_7_3_5_14
void FASTCALL call_7_3_5_14(SH2_struct* sh)
{
  SH2addi(sh, 3, (5<<4)+14);
}
#define init_7_3_5_15
void FASTCALL call_7_3_5_15(SH2_struct* sh)
{
  SH2addi(sh, 3, (5<<4)+15);
}
#define init_7_3_6_0
void FASTCALL call_7_3_6_0(SH2_struct* sh)
{
  SH2addi(sh, 3, (6<<4)+0);
}
#define init_7_3_6_1
void FASTCALL call_7_3_6_1(SH2_struct* sh)
{
  SH2addi(sh, 3, (6<<4)+1);
}
#define init_7_3_6_2
void FASTCALL call_7_3_6_2(SH2_struct* sh)
{
  SH2addi(sh, 3, (6<<4)+2);
}
#define init_7_3_6_3
void FASTCALL call_7_3_6_3(SH2_struct* sh)
{
  SH2addi(sh, 3, (6<<4)+3);
}
#define init_7_3_6_4
void FASTCALL call_7_3_6_4(SH2_struct* sh)
{
  SH2addi(sh, 3, (6<<4)+4);
}
#define init_7_3_6_5
void FASTCALL call_7_3_6_5(SH2_struct* sh)
{
  SH2addi(sh, 3, (6<<4)+5);
}
#define init_7_3_6_6
void FASTCALL call_7_3_6_6(SH2_struct* sh)
{
  SH2addi(sh, 3, (6<<4)+6);
}
#define init_7_3_6_7
void FASTCALL call_7_3_6_7(SH2_struct* sh)
{
  SH2addi(sh, 3, (6<<4)+7);
}
#define init_7_3_6_8
void FASTCALL call_7_3_6_8(SH2_struct* sh)
{
  SH2addi(sh, 3, (6<<4)+8);
}
#define init_7_3_6_9
void FASTCALL call_7_3_6_9(SH2_struct* sh)
{
  SH2addi(sh, 3, (6<<4)+9);
}
#define init_7_3_6_10
void FASTCALL call_7_3_6_10(SH2_struct* sh)
{
  SH2addi(sh, 3, (6<<4)+10);
}
#define init_7_3_6_11
void FASTCALL call_7_3_6_11(SH2_struct* sh)
{
  SH2addi(sh, 3, (6<<4)+11);
}
#define init_7_3_6_12
void FASTCALL call_7_3_6_12(SH2_struct* sh)
{
  SH2addi(sh, 3, (6<<4)+12);
}
#define init_7_3_6_13
void FASTCALL call_7_3_6_13(SH2_struct* sh)
{
  SH2addi(sh, 3, (6<<4)+13);
}
#define init_7_3_6_14
void FASTCALL call_7_3_6_14(SH2_struct* sh)
{
  SH2addi(sh, 3, (6<<4)+14);
}
#define init_7_3_6_15
void FASTCALL call_7_3_6_15(SH2_struct* sh)
{
  SH2addi(sh, 3, (6<<4)+15);
}
#define init_7_3_7_0
void FASTCALL call_7_3_7_0(SH2_struct* sh)
{
  SH2addi(sh, 3, (7<<4)+0);
}
#define init_7_3_7_1
void FASTCALL call_7_3_7_1(SH2_struct* sh)
{
  SH2addi(sh, 3, (7<<4)+1);
}
#define init_7_3_7_2
void FASTCALL call_7_3_7_2(SH2_struct* sh)
{
  SH2addi(sh, 3, (7<<4)+2);
}
#define init_7_3_7_3
void FASTCALL call_7_3_7_3(SH2_struct* sh)
{
  SH2addi(sh, 3, (7<<4)+3);
}
#define init_7_3_7_4
void FASTCALL call_7_3_7_4(SH2_struct* sh)
{
  SH2addi(sh, 3, (7<<4)+4);
}
#define init_7_3_7_5
void FASTCALL call_7_3_7_5(SH2_struct* sh)
{
  SH2addi(sh, 3, (7<<4)+5);
}
#define init_7_3_7_6
void FASTCALL call_7_3_7_6(SH2_struct* sh)
{
  SH2addi(sh, 3, (7<<4)+6);
}
#define init_7_3_7_7
void FASTCALL call_7_3_7_7(SH2_struct* sh)
{
  SH2addi(sh, 3, (7<<4)+7);
}
#define init_7_3_7_8
void FASTCALL call_7_3_7_8(SH2_struct* sh)
{
  SH2addi(sh, 3, (7<<4)+8);
}
#define init_7_3_7_9
void FASTCALL call_7_3_7_9(SH2_struct* sh)
{
  SH2addi(sh, 3, (7<<4)+9);
}
#define init_7_3_7_10
void FASTCALL call_7_3_7_10(SH2_struct* sh)
{
  SH2addi(sh, 3, (7<<4)+10);
}
#define init_7_3_7_11
void FASTCALL call_7_3_7_11(SH2_struct* sh)
{
  SH2addi(sh, 3, (7<<4)+11);
}
#define init_7_3_7_12
void FASTCALL call_7_3_7_12(SH2_struct* sh)
{
  SH2addi(sh, 3, (7<<4)+12);
}
#define init_7_3_7_13
void FASTCALL call_7_3_7_13(SH2_struct* sh)
{
  SH2addi(sh, 3, (7<<4)+13);
}
#define init_7_3_7_14
void FASTCALL call_7_3_7_14(SH2_struct* sh)
{
  SH2addi(sh, 3, (7<<4)+14);
}
#define init_7_3_7_15
void FASTCALL call_7_3_7_15(SH2_struct* sh)
{
  SH2addi(sh, 3, (7<<4)+15);
}
#define init_7_3_8_0
void FASTCALL call_7_3_8_0(SH2_struct* sh)
{
  SH2addi(sh, 3, (8<<4)+0);
}
#define init_7_3_8_1
void FASTCALL call_7_3_8_1(SH2_struct* sh)
{
  SH2addi(sh, 3, (8<<4)+1);
}
#define init_7_3_8_2
void FASTCALL call_7_3_8_2(SH2_struct* sh)
{
  SH2addi(sh, 3, (8<<4)+2);
}
#define init_7_3_8_3
void FASTCALL call_7_3_8_3(SH2_struct* sh)
{
  SH2addi(sh, 3, (8<<4)+3);
}
#define init_7_3_8_4
void FASTCALL call_7_3_8_4(SH2_struct* sh)
{
  SH2addi(sh, 3, (8<<4)+4);
}
#define init_7_3_8_5
void FASTCALL call_7_3_8_5(SH2_struct* sh)
{
  SH2addi(sh, 3, (8<<4)+5);
}
#define init_7_3_8_6
void FASTCALL call_7_3_8_6(SH2_struct* sh)
{
  SH2addi(sh, 3, (8<<4)+6);
}
#define init_7_3_8_7
void FASTCALL call_7_3_8_7(SH2_struct* sh)
{
  SH2addi(sh, 3, (8<<4)+7);
}
#define init_7_3_8_8
void FASTCALL call_7_3_8_8(SH2_struct* sh)
{
  SH2addi(sh, 3, (8<<4)+8);
}
#define init_7_3_8_9
void FASTCALL call_7_3_8_9(SH2_struct* sh)
{
  SH2addi(sh, 3, (8<<4)+9);
}
#define init_7_3_8_10
void FASTCALL call_7_3_8_10(SH2_struct* sh)
{
  SH2addi(sh, 3, (8<<4)+10);
}
#define init_7_3_8_11
void FASTCALL call_7_3_8_11(SH2_struct* sh)
{
  SH2addi(sh, 3, (8<<4)+11);
}
#define init_7_3_8_12
void FASTCALL call_7_3_8_12(SH2_struct* sh)
{
  SH2addi(sh, 3, (8<<4)+12);
}
#define init_7_3_8_13
void FASTCALL call_7_3_8_13(SH2_struct* sh)
{
  SH2addi(sh, 3, (8<<4)+13);
}
#define init_7_3_8_14
void FASTCALL call_7_3_8_14(SH2_struct* sh)
{
  SH2addi(sh, 3, (8<<4)+14);
}
#define init_7_3_8_15
void FASTCALL call_7_3_8_15(SH2_struct* sh)
{
  SH2addi(sh, 3, (8<<4)+15);
}
#define init_7_3_9_0
void FASTCALL call_7_3_9_0(SH2_struct* sh)
{
  SH2addi(sh, 3, (9<<4)+0);
}
#define init_7_3_9_1
void FASTCALL call_7_3_9_1(SH2_struct* sh)
{
  SH2addi(sh, 3, (9<<4)+1);
}
#define init_7_3_9_2
void FASTCALL call_7_3_9_2(SH2_struct* sh)
{
  SH2addi(sh, 3, (9<<4)+2);
}
#define init_7_3_9_3
void FASTCALL call_7_3_9_3(SH2_struct* sh)
{
  SH2addi(sh, 3, (9<<4)+3);
}
#define init_7_3_9_4
void FASTCALL call_7_3_9_4(SH2_struct* sh)
{
  SH2addi(sh, 3, (9<<4)+4);
}
#define init_7_3_9_5
void FASTCALL call_7_3_9_5(SH2_struct* sh)
{
  SH2addi(sh, 3, (9<<4)+5);
}
#define init_7_3_9_6
void FASTCALL call_7_3_9_6(SH2_struct* sh)
{
  SH2addi(sh, 3, (9<<4)+6);
}
#define init_7_3_9_7
void FASTCALL call_7_3_9_7(SH2_struct* sh)
{
  SH2addi(sh, 3, (9<<4)+7);
}
#define init_7_3_9_8
void FASTCALL call_7_3_9_8(SH2_struct* sh)
{
  SH2addi(sh, 3, (9<<4)+8);
}
#define init_7_3_9_9
void FASTCALL call_7_3_9_9(SH2_struct* sh)
{
  SH2addi(sh, 3, (9<<4)+9);
}
#define init_7_3_9_10
void FASTCALL call_7_3_9_10(SH2_struct* sh)
{
  SH2addi(sh, 3, (9<<4)+10);
}
#define init_7_3_9_11
void FASTCALL call_7_3_9_11(SH2_struct* sh)
{
  SH2addi(sh, 3, (9<<4)+11);
}
#define init_7_3_9_12
void FASTCALL call_7_3_9_12(SH2_struct* sh)
{
  SH2addi(sh, 3, (9<<4)+12);
}
#define init_7_3_9_13
void FASTCALL call_7_3_9_13(SH2_struct* sh)
{
  SH2addi(sh, 3, (9<<4)+13);
}
#define init_7_3_9_14
void FASTCALL call_7_3_9_14(SH2_struct* sh)
{
  SH2addi(sh, 3, (9<<4)+14);
}
#define init_7_3_9_15
void FASTCALL call_7_3_9_15(SH2_struct* sh)
{
  SH2addi(sh, 3, (9<<4)+15);
}
#define init_7_3_10_0
void FASTCALL call_7_3_10_0(SH2_struct* sh)
{
  SH2addi(sh, 3, (10<<4)+0);
}
#define init_7_3_10_1
void FASTCALL call_7_3_10_1(SH2_struct* sh)
{
  SH2addi(sh, 3, (10<<4)+1);
}
#define init_7_3_10_2
void FASTCALL call_7_3_10_2(SH2_struct* sh)
{
  SH2addi(sh, 3, (10<<4)+2);
}
#define init_7_3_10_3
void FASTCALL call_7_3_10_3(SH2_struct* sh)
{
  SH2addi(sh, 3, (10<<4)+3);
}
#define init_7_3_10_4
void FASTCALL call_7_3_10_4(SH2_struct* sh)
{
  SH2addi(sh, 3, (10<<4)+4);
}
#define init_7_3_10_5
void FASTCALL call_7_3_10_5(SH2_struct* sh)
{
  SH2addi(sh, 3, (10<<4)+5);
}
#define init_7_3_10_6
void FASTCALL call_7_3_10_6(SH2_struct* sh)
{
  SH2addi(sh, 3, (10<<4)+6);
}
#define init_7_3_10_7
void FASTCALL call_7_3_10_7(SH2_struct* sh)
{
  SH2addi(sh, 3, (10<<4)+7);
}
#define init_7_3_10_8
void FASTCALL call_7_3_10_8(SH2_struct* sh)
{
  SH2addi(sh, 3, (10<<4)+8);
}
#define init_7_3_10_9
void FASTCALL call_7_3_10_9(SH2_struct* sh)
{
  SH2addi(sh, 3, (10<<4)+9);
}
#define init_7_3_10_10
void FASTCALL call_7_3_10_10(SH2_struct* sh)
{
  SH2addi(sh, 3, (10<<4)+10);
}
#define init_7_3_10_11
void FASTCALL call_7_3_10_11(SH2_struct* sh)
{
  SH2addi(sh, 3, (10<<4)+11);
}
#define init_7_3_10_12
void FASTCALL call_7_3_10_12(SH2_struct* sh)
{
  SH2addi(sh, 3, (10<<4)+12);
}
#define init_7_3_10_13
void FASTCALL call_7_3_10_13(SH2_struct* sh)
{
  SH2addi(sh, 3, (10<<4)+13);
}
#define init_7_3_10_14
void FASTCALL call_7_3_10_14(SH2_struct* sh)
{
  SH2addi(sh, 3, (10<<4)+14);
}
#define init_7_3_10_15
void FASTCALL call_7_3_10_15(SH2_struct* sh)
{
  SH2addi(sh, 3, (10<<4)+15);
}
#define init_7_3_11_0
void FASTCALL call_7_3_11_0(SH2_struct* sh)
{
  SH2addi(sh, 3, (11<<4)+0);
}
#define init_7_3_11_1
void FASTCALL call_7_3_11_1(SH2_struct* sh)
{
  SH2addi(sh, 3, (11<<4)+1);
}
#define init_7_3_11_2
void FASTCALL call_7_3_11_2(SH2_struct* sh)
{
  SH2addi(sh, 3, (11<<4)+2);
}
#define init_7_3_11_3
void FASTCALL call_7_3_11_3(SH2_struct* sh)
{
  SH2addi(sh, 3, (11<<4)+3);
}
#define init_7_3_11_4
void FASTCALL call_7_3_11_4(SH2_struct* sh)
{
  SH2addi(sh, 3, (11<<4)+4);
}
#define init_7_3_11_5
void FASTCALL call_7_3_11_5(SH2_struct* sh)
{
  SH2addi(sh, 3, (11<<4)+5);
}
#define init_7_3_11_6
void FASTCALL call_7_3_11_6(SH2_struct* sh)
{
  SH2addi(sh, 3, (11<<4)+6);
}
#define init_7_3_11_7
void FASTCALL call_7_3_11_7(SH2_struct* sh)
{
  SH2addi(sh, 3, (11<<4)+7);
}
#define init_7_3_11_8
void FASTCALL call_7_3_11_8(SH2_struct* sh)
{
  SH2addi(sh, 3, (11<<4)+8);
}
#define init_7_3_11_9
void FASTCALL call_7_3_11_9(SH2_struct* sh)
{
  SH2addi(sh, 3, (11<<4)+9);
}
#define init_7_3_11_10
void FASTCALL call_7_3_11_10(SH2_struct* sh)
{
  SH2addi(sh, 3, (11<<4)+10);
}
#define init_7_3_11_11
void FASTCALL call_7_3_11_11(SH2_struct* sh)
{
  SH2addi(sh, 3, (11<<4)+11);
}
#define init_7_3_11_12
void FASTCALL call_7_3_11_12(SH2_struct* sh)
{
  SH2addi(sh, 3, (11<<4)+12);
}
#define init_7_3_11_13
void FASTCALL call_7_3_11_13(SH2_struct* sh)
{
  SH2addi(sh, 3, (11<<4)+13);
}
#define init_7_3_11_14
void FASTCALL call_7_3_11_14(SH2_struct* sh)
{
  SH2addi(sh, 3, (11<<4)+14);
}
#define init_7_3_11_15
void FASTCALL call_7_3_11_15(SH2_struct* sh)
{
  SH2addi(sh, 3, (11<<4)+15);
}
#define init_7_3_12_0
void FASTCALL call_7_3_12_0(SH2_struct* sh)
{
  SH2addi(sh, 3, (12<<4)+0);
}
#define init_7_3_12_1
void FASTCALL call_7_3_12_1(SH2_struct* sh)
{
  SH2addi(sh, 3, (12<<4)+1);
}
#define init_7_3_12_2
void FASTCALL call_7_3_12_2(SH2_struct* sh)
{
  SH2addi(sh, 3, (12<<4)+2);
}
#define init_7_3_12_3
void FASTCALL call_7_3_12_3(SH2_struct* sh)
{
  SH2addi(sh, 3, (12<<4)+3);
}
#define init_7_3_12_4
void FASTCALL call_7_3_12_4(SH2_struct* sh)
{
  SH2addi(sh, 3, (12<<4)+4);
}
#define init_7_3_12_5
void FASTCALL call_7_3_12_5(SH2_struct* sh)
{
  SH2addi(sh, 3, (12<<4)+5);
}
#define init_7_3_12_6
void FASTCALL call_7_3_12_6(SH2_struct* sh)
{
  SH2addi(sh, 3, (12<<4)+6);
}
#define init_7_3_12_7
void FASTCALL call_7_3_12_7(SH2_struct* sh)
{
  SH2addi(sh, 3, (12<<4)+7);
}
#define init_7_3_12_8
void FASTCALL call_7_3_12_8(SH2_struct* sh)
{
  SH2addi(sh, 3, (12<<4)+8);
}
#define init_7_3_12_9
void FASTCALL call_7_3_12_9(SH2_struct* sh)
{
  SH2addi(sh, 3, (12<<4)+9);
}
#define init_7_3_12_10
void FASTCALL call_7_3_12_10(SH2_struct* sh)
{
  SH2addi(sh, 3, (12<<4)+10);
}
#define init_7_3_12_11
void FASTCALL call_7_3_12_11(SH2_struct* sh)
{
  SH2addi(sh, 3, (12<<4)+11);
}
#define init_7_3_12_12
void FASTCALL call_7_3_12_12(SH2_struct* sh)
{
  SH2addi(sh, 3, (12<<4)+12);
}
#define init_7_3_12_13
void FASTCALL call_7_3_12_13(SH2_struct* sh)
{
  SH2addi(sh, 3, (12<<4)+13);
}
#define init_7_3_12_14
void FASTCALL call_7_3_12_14(SH2_struct* sh)
{
  SH2addi(sh, 3, (12<<4)+14);
}
#define init_7_3_12_15
void FASTCALL call_7_3_12_15(SH2_struct* sh)
{
  SH2addi(sh, 3, (12<<4)+15);
}
#define init_7_3_13_0
void FASTCALL call_7_3_13_0(SH2_struct* sh)
{
  SH2addi(sh, 3, (13<<4)+0);
}
#define init_7_3_13_1
void FASTCALL call_7_3_13_1(SH2_struct* sh)
{
  SH2addi(sh, 3, (13<<4)+1);
}
#define init_7_3_13_2
void FASTCALL call_7_3_13_2(SH2_struct* sh)
{
  SH2addi(sh, 3, (13<<4)+2);
}
#define init_7_3_13_3
void FASTCALL call_7_3_13_3(SH2_struct* sh)
{
  SH2addi(sh, 3, (13<<4)+3);
}
#define init_7_3_13_4
void FASTCALL call_7_3_13_4(SH2_struct* sh)
{
  SH2addi(sh, 3, (13<<4)+4);
}
#define init_7_3_13_5
void FASTCALL call_7_3_13_5(SH2_struct* sh)
{
  SH2addi(sh, 3, (13<<4)+5);
}
#define init_7_3_13_6
void FASTCALL call_7_3_13_6(SH2_struct* sh)
{
  SH2addi(sh, 3, (13<<4)+6);
}
#define init_7_3_13_7
void FASTCALL call_7_3_13_7(SH2_struct* sh)
{
  SH2addi(sh, 3, (13<<4)+7);
}
#define init_7_3_13_8
void FASTCALL call_7_3_13_8(SH2_struct* sh)
{
  SH2addi(sh, 3, (13<<4)+8);
}
#define init_7_3_13_9
void FASTCALL call_7_3_13_9(SH2_struct* sh)
{
  SH2addi(sh, 3, (13<<4)+9);
}
#define init_7_3_13_10
void FASTCALL call_7_3_13_10(SH2_struct* sh)
{
  SH2addi(sh, 3, (13<<4)+10);
}
#define init_7_3_13_11
void FASTCALL call_7_3_13_11(SH2_struct* sh)
{
  SH2addi(sh, 3, (13<<4)+11);
}
#define init_7_3_13_12
void FASTCALL call_7_3_13_12(SH2_struct* sh)
{
  SH2addi(sh, 3, (13<<4)+12);
}
#define init_7_3_13_13
void FASTCALL call_7_3_13_13(SH2_struct* sh)
{
  SH2addi(sh, 3, (13<<4)+13);
}
#define init_7_3_13_14
void FASTCALL call_7_3_13_14(SH2_struct* sh)
{
  SH2addi(sh, 3, (13<<4)+14);
}
#define init_7_3_13_15
void FASTCALL call_7_3_13_15(SH2_struct* sh)
{
  SH2addi(sh, 3, (13<<4)+15);
}
#define init_7_3_14_0
void FASTCALL call_7_3_14_0(SH2_struct* sh)
{
  SH2addi(sh, 3, (14<<4)+0);
}
#define init_7_3_14_1
void FASTCALL call_7_3_14_1(SH2_struct* sh)
{
  SH2addi(sh, 3, (14<<4)+1);
}
#define init_7_3_14_2
void FASTCALL call_7_3_14_2(SH2_struct* sh)
{
  SH2addi(sh, 3, (14<<4)+2);
}
#define init_7_3_14_3
void FASTCALL call_7_3_14_3(SH2_struct* sh)
{
  SH2addi(sh, 3, (14<<4)+3);
}
#define init_7_3_14_4
void FASTCALL call_7_3_14_4(SH2_struct* sh)
{
  SH2addi(sh, 3, (14<<4)+4);
}
#define init_7_3_14_5
void FASTCALL call_7_3_14_5(SH2_struct* sh)
{
  SH2addi(sh, 3, (14<<4)+5);
}
#define init_7_3_14_6
void FASTCALL call_7_3_14_6(SH2_struct* sh)
{
  SH2addi(sh, 3, (14<<4)+6);
}
#define init_7_3_14_7
void FASTCALL call_7_3_14_7(SH2_struct* sh)
{
  SH2addi(sh, 3, (14<<4)+7);
}
#define init_7_3_14_8
void FASTCALL call_7_3_14_8(SH2_struct* sh)
{
  SH2addi(sh, 3, (14<<4)+8);
}
#define init_7_3_14_9
void FASTCALL call_7_3_14_9(SH2_struct* sh)
{
  SH2addi(sh, 3, (14<<4)+9);
}
#define init_7_3_14_10
void FASTCALL call_7_3_14_10(SH2_struct* sh)
{
  SH2addi(sh, 3, (14<<4)+10);
}
#define init_7_3_14_11
void FASTCALL call_7_3_14_11(SH2_struct* sh)
{
  SH2addi(sh, 3, (14<<4)+11);
}
#define init_7_3_14_12
void FASTCALL call_7_3_14_12(SH2_struct* sh)
{
  SH2addi(sh, 3, (14<<4)+12);
}
#define init_7_3_14_13
void FASTCALL call_7_3_14_13(SH2_struct* sh)
{
  SH2addi(sh, 3, (14<<4)+13);
}
#define init_7_3_14_14
void FASTCALL call_7_3_14_14(SH2_struct* sh)
{
  SH2addi(sh, 3, (14<<4)+14);
}
#define init_7_3_14_15
void FASTCALL call_7_3_14_15(SH2_struct* sh)
{
  SH2addi(sh, 3, (14<<4)+15);
}
#define init_7_3_15_0
void FASTCALL call_7_3_15_0(SH2_struct* sh)
{
  SH2addi(sh, 3, (15<<4)+0);
}
#define init_7_3_15_1
void FASTCALL call_7_3_15_1(SH2_struct* sh)
{
  SH2addi(sh, 3, (15<<4)+1);
}
#define init_7_3_15_2
void FASTCALL call_7_3_15_2(SH2_struct* sh)
{
  SH2addi(sh, 3, (15<<4)+2);
}
#define init_7_3_15_3
void FASTCALL call_7_3_15_3(SH2_struct* sh)
{
  SH2addi(sh, 3, (15<<4)+3);
}
#define init_7_3_15_4
void FASTCALL call_7_3_15_4(SH2_struct* sh)
{
  SH2addi(sh, 3, (15<<4)+4);
}
#define init_7_3_15_5
void FASTCALL call_7_3_15_5(SH2_struct* sh)
{
  SH2addi(sh, 3, (15<<4)+5);
}
#define init_7_3_15_6
void FASTCALL call_7_3_15_6(SH2_struct* sh)
{
  SH2addi(sh, 3, (15<<4)+6);
}
#define init_7_3_15_7
void FASTCALL call_7_3_15_7(SH2_struct* sh)
{
  SH2addi(sh, 3, (15<<4)+7);
}
#define init_7_3_15_8
void FASTCALL call_7_3_15_8(SH2_struct* sh)
{
  SH2addi(sh, 3, (15<<4)+8);
}
#define init_7_3_15_9
void FASTCALL call_7_3_15_9(SH2_struct* sh)
{
  SH2addi(sh, 3, (15<<4)+9);
}
#define init_7_3_15_10
void FASTCALL call_7_3_15_10(SH2_struct* sh)
{
  SH2addi(sh, 3, (15<<4)+10);
}
#define init_7_3_15_11
void FASTCALL call_7_3_15_11(SH2_struct* sh)
{
  SH2addi(sh, 3, (15<<4)+11);
}
#define init_7_3_15_12
void FASTCALL call_7_3_15_12(SH2_struct* sh)
{
  SH2addi(sh, 3, (15<<4)+12);
}
#define init_7_3_15_13
void FASTCALL call_7_3_15_13(SH2_struct* sh)
{
  SH2addi(sh, 3, (15<<4)+13);
}
#define init_7_3_15_14
void FASTCALL call_7_3_15_14(SH2_struct* sh)
{
  SH2addi(sh, 3, (15<<4)+14);
}
#define init_7_3_15_15
void FASTCALL call_7_3_15_15(SH2_struct* sh)
{
  SH2addi(sh, 3, (15<<4)+15);
}
#define init_7_4_0_0
void FASTCALL call_7_4_0_0(SH2_struct* sh)
{
  SH2addi(sh, 4, (0<<4)+0);
}
#define init_7_4_0_1
void FASTCALL call_7_4_0_1(SH2_struct* sh)
{
  SH2addi(sh, 4, (0<<4)+1);
}
#define init_7_4_0_2
void FASTCALL call_7_4_0_2(SH2_struct* sh)
{
  SH2addi(sh, 4, (0<<4)+2);
}
#define init_7_4_0_3
void FASTCALL call_7_4_0_3(SH2_struct* sh)
{
  SH2addi(sh, 4, (0<<4)+3);
}
#define init_7_4_0_4
void FASTCALL call_7_4_0_4(SH2_struct* sh)
{
  SH2addi(sh, 4, (0<<4)+4);
}
#define init_7_4_0_5
void FASTCALL call_7_4_0_5(SH2_struct* sh)
{
  SH2addi(sh, 4, (0<<4)+5);
}
#define init_7_4_0_6
void FASTCALL call_7_4_0_6(SH2_struct* sh)
{
  SH2addi(sh, 4, (0<<4)+6);
}
#define init_7_4_0_7
void FASTCALL call_7_4_0_7(SH2_struct* sh)
{
  SH2addi(sh, 4, (0<<4)+7);
}
#define init_7_4_0_8
void FASTCALL call_7_4_0_8(SH2_struct* sh)
{
  SH2addi(sh, 4, (0<<4)+8);
}
#define init_7_4_0_9
void FASTCALL call_7_4_0_9(SH2_struct* sh)
{
  SH2addi(sh, 4, (0<<4)+9);
}
#define init_7_4_0_10
void FASTCALL call_7_4_0_10(SH2_struct* sh)
{
  SH2addi(sh, 4, (0<<4)+10);
}
#define init_7_4_0_11
void FASTCALL call_7_4_0_11(SH2_struct* sh)
{
  SH2addi(sh, 4, (0<<4)+11);
}
#define init_7_4_0_12
void FASTCALL call_7_4_0_12(SH2_struct* sh)
{
  SH2addi(sh, 4, (0<<4)+12);
}
#define init_7_4_0_13
void FASTCALL call_7_4_0_13(SH2_struct* sh)
{
  SH2addi(sh, 4, (0<<4)+13);
}
#define init_7_4_0_14
void FASTCALL call_7_4_0_14(SH2_struct* sh)
{
  SH2addi(sh, 4, (0<<4)+14);
}
#define init_7_4_0_15
void FASTCALL call_7_4_0_15(SH2_struct* sh)
{
  SH2addi(sh, 4, (0<<4)+15);
}
#define init_7_4_1_0
void FASTCALL call_7_4_1_0(SH2_struct* sh)
{
  SH2addi(sh, 4, (1<<4)+0);
}
#define init_7_4_1_1
void FASTCALL call_7_4_1_1(SH2_struct* sh)
{
  SH2addi(sh, 4, (1<<4)+1);
}
#define init_7_4_1_2
void FASTCALL call_7_4_1_2(SH2_struct* sh)
{
  SH2addi(sh, 4, (1<<4)+2);
}
#define init_7_4_1_3
void FASTCALL call_7_4_1_3(SH2_struct* sh)
{
  SH2addi(sh, 4, (1<<4)+3);
}
#define init_7_4_1_4
void FASTCALL call_7_4_1_4(SH2_struct* sh)
{
  SH2addi(sh, 4, (1<<4)+4);
}
#define init_7_4_1_5
void FASTCALL call_7_4_1_5(SH2_struct* sh)
{
  SH2addi(sh, 4, (1<<4)+5);
}
#define init_7_4_1_6
void FASTCALL call_7_4_1_6(SH2_struct* sh)
{
  SH2addi(sh, 4, (1<<4)+6);
}
#define init_7_4_1_7
void FASTCALL call_7_4_1_7(SH2_struct* sh)
{
  SH2addi(sh, 4, (1<<4)+7);
}
#define init_7_4_1_8
void FASTCALL call_7_4_1_8(SH2_struct* sh)
{
  SH2addi(sh, 4, (1<<4)+8);
}
#define init_7_4_1_9
void FASTCALL call_7_4_1_9(SH2_struct* sh)
{
  SH2addi(sh, 4, (1<<4)+9);
}
#define init_7_4_1_10
void FASTCALL call_7_4_1_10(SH2_struct* sh)
{
  SH2addi(sh, 4, (1<<4)+10);
}
#define init_7_4_1_11
void FASTCALL call_7_4_1_11(SH2_struct* sh)
{
  SH2addi(sh, 4, (1<<4)+11);
}
#define init_7_4_1_12
void FASTCALL call_7_4_1_12(SH2_struct* sh)
{
  SH2addi(sh, 4, (1<<4)+12);
}
#define init_7_4_1_13
void FASTCALL call_7_4_1_13(SH2_struct* sh)
{
  SH2addi(sh, 4, (1<<4)+13);
}
#define init_7_4_1_14
void FASTCALL call_7_4_1_14(SH2_struct* sh)
{
  SH2addi(sh, 4, (1<<4)+14);
}
#define init_7_4_1_15
void FASTCALL call_7_4_1_15(SH2_struct* sh)
{
  SH2addi(sh, 4, (1<<4)+15);
}
#define init_7_4_2_0
void FASTCALL call_7_4_2_0(SH2_struct* sh)
{
  SH2addi(sh, 4, (2<<4)+0);
}
#define init_7_4_2_1
void FASTCALL call_7_4_2_1(SH2_struct* sh)
{
  SH2addi(sh, 4, (2<<4)+1);
}
#define init_7_4_2_2
void FASTCALL call_7_4_2_2(SH2_struct* sh)
{
  SH2addi(sh, 4, (2<<4)+2);
}
#define init_7_4_2_3
void FASTCALL call_7_4_2_3(SH2_struct* sh)
{
  SH2addi(sh, 4, (2<<4)+3);
}
#define init_7_4_2_4
void FASTCALL call_7_4_2_4(SH2_struct* sh)
{
  SH2addi(sh, 4, (2<<4)+4);
}
#define init_7_4_2_5
void FASTCALL call_7_4_2_5(SH2_struct* sh)
{
  SH2addi(sh, 4, (2<<4)+5);
}
#define init_7_4_2_6
void FASTCALL call_7_4_2_6(SH2_struct* sh)
{
  SH2addi(sh, 4, (2<<4)+6);
}
#define init_7_4_2_7
void FASTCALL call_7_4_2_7(SH2_struct* sh)
{
  SH2addi(sh, 4, (2<<4)+7);
}
#define init_7_4_2_8
void FASTCALL call_7_4_2_8(SH2_struct* sh)
{
  SH2addi(sh, 4, (2<<4)+8);
}
#define init_7_4_2_9
void FASTCALL call_7_4_2_9(SH2_struct* sh)
{
  SH2addi(sh, 4, (2<<4)+9);
}
#define init_7_4_2_10
void FASTCALL call_7_4_2_10(SH2_struct* sh)
{
  SH2addi(sh, 4, (2<<4)+10);
}
#define init_7_4_2_11
void FASTCALL call_7_4_2_11(SH2_struct* sh)
{
  SH2addi(sh, 4, (2<<4)+11);
}
#define init_7_4_2_12
void FASTCALL call_7_4_2_12(SH2_struct* sh)
{
  SH2addi(sh, 4, (2<<4)+12);
}
#define init_7_4_2_13
void FASTCALL call_7_4_2_13(SH2_struct* sh)
{
  SH2addi(sh, 4, (2<<4)+13);
}
#define init_7_4_2_14
void FASTCALL call_7_4_2_14(SH2_struct* sh)
{
  SH2addi(sh, 4, (2<<4)+14);
}
#define init_7_4_2_15
void FASTCALL call_7_4_2_15(SH2_struct* sh)
{
  SH2addi(sh, 4, (2<<4)+15);
}
#define init_7_4_3_0
void FASTCALL call_7_4_3_0(SH2_struct* sh)
{
  SH2addi(sh, 4, (3<<4)+0);
}
#define init_7_4_3_1
void FASTCALL call_7_4_3_1(SH2_struct* sh)
{
  SH2addi(sh, 4, (3<<4)+1);
}
#define init_7_4_3_2
void FASTCALL call_7_4_3_2(SH2_struct* sh)
{
  SH2addi(sh, 4, (3<<4)+2);
}
#define init_7_4_3_3
void FASTCALL call_7_4_3_3(SH2_struct* sh)
{
  SH2addi(sh, 4, (3<<4)+3);
}
#define init_7_4_3_4
void FASTCALL call_7_4_3_4(SH2_struct* sh)
{
  SH2addi(sh, 4, (3<<4)+4);
}
#define init_7_4_3_5
void FASTCALL call_7_4_3_5(SH2_struct* sh)
{
  SH2addi(sh, 4, (3<<4)+5);
}
#define init_7_4_3_6
void FASTCALL call_7_4_3_6(SH2_struct* sh)
{
  SH2addi(sh, 4, (3<<4)+6);
}
#define init_7_4_3_7
void FASTCALL call_7_4_3_7(SH2_struct* sh)
{
  SH2addi(sh, 4, (3<<4)+7);
}
#define init_7_4_3_8
void FASTCALL call_7_4_3_8(SH2_struct* sh)
{
  SH2addi(sh, 4, (3<<4)+8);
}
#define init_7_4_3_9
void FASTCALL call_7_4_3_9(SH2_struct* sh)
{
  SH2addi(sh, 4, (3<<4)+9);
}
#define init_7_4_3_10
void FASTCALL call_7_4_3_10(SH2_struct* sh)
{
  SH2addi(sh, 4, (3<<4)+10);
}
#define init_7_4_3_11
void FASTCALL call_7_4_3_11(SH2_struct* sh)
{
  SH2addi(sh, 4, (3<<4)+11);
}
#define init_7_4_3_12
void FASTCALL call_7_4_3_12(SH2_struct* sh)
{
  SH2addi(sh, 4, (3<<4)+12);
}
#define init_7_4_3_13
void FASTCALL call_7_4_3_13(SH2_struct* sh)
{
  SH2addi(sh, 4, (3<<4)+13);
}
#define init_7_4_3_14
void FASTCALL call_7_4_3_14(SH2_struct* sh)
{
  SH2addi(sh, 4, (3<<4)+14);
}
#define init_7_4_3_15
void FASTCALL call_7_4_3_15(SH2_struct* sh)
{
  SH2addi(sh, 4, (3<<4)+15);
}
#define init_7_4_4_0
void FASTCALL call_7_4_4_0(SH2_struct* sh)
{
  SH2addi(sh, 4, (4<<4)+0);
}
#define init_7_4_4_1
void FASTCALL call_7_4_4_1(SH2_struct* sh)
{
  SH2addi(sh, 4, (4<<4)+1);
}
#define init_7_4_4_2
void FASTCALL call_7_4_4_2(SH2_struct* sh)
{
  SH2addi(sh, 4, (4<<4)+2);
}
#define init_7_4_4_3
void FASTCALL call_7_4_4_3(SH2_struct* sh)
{
  SH2addi(sh, 4, (4<<4)+3);
}
#define init_7_4_4_4
void FASTCALL call_7_4_4_4(SH2_struct* sh)
{
  SH2addi(sh, 4, (4<<4)+4);
}
#define init_7_4_4_5
void FASTCALL call_7_4_4_5(SH2_struct* sh)
{
  SH2addi(sh, 4, (4<<4)+5);
}
#define init_7_4_4_6
void FASTCALL call_7_4_4_6(SH2_struct* sh)
{
  SH2addi(sh, 4, (4<<4)+6);
}
#define init_7_4_4_7
void FASTCALL call_7_4_4_7(SH2_struct* sh)
{
  SH2addi(sh, 4, (4<<4)+7);
}
#define init_7_4_4_8
void FASTCALL call_7_4_4_8(SH2_struct* sh)
{
  SH2addi(sh, 4, (4<<4)+8);
}
#define init_7_4_4_9
void FASTCALL call_7_4_4_9(SH2_struct* sh)
{
  SH2addi(sh, 4, (4<<4)+9);
}
#define init_7_4_4_10
void FASTCALL call_7_4_4_10(SH2_struct* sh)
{
  SH2addi(sh, 4, (4<<4)+10);
}
#define init_7_4_4_11
void FASTCALL call_7_4_4_11(SH2_struct* sh)
{
  SH2addi(sh, 4, (4<<4)+11);
}
#define init_7_4_4_12
void FASTCALL call_7_4_4_12(SH2_struct* sh)
{
  SH2addi(sh, 4, (4<<4)+12);
}
#define init_7_4_4_13
void FASTCALL call_7_4_4_13(SH2_struct* sh)
{
  SH2addi(sh, 4, (4<<4)+13);
}
#define init_7_4_4_14
void FASTCALL call_7_4_4_14(SH2_struct* sh)
{
  SH2addi(sh, 4, (4<<4)+14);
}
#define init_7_4_4_15
void FASTCALL call_7_4_4_15(SH2_struct* sh)
{
  SH2addi(sh, 4, (4<<4)+15);
}
#define init_7_4_5_0
void FASTCALL call_7_4_5_0(SH2_struct* sh)
{
  SH2addi(sh, 4, (5<<4)+0);
}
#define init_7_4_5_1
void FASTCALL call_7_4_5_1(SH2_struct* sh)
{
  SH2addi(sh, 4, (5<<4)+1);
}
#define init_7_4_5_2
void FASTCALL call_7_4_5_2(SH2_struct* sh)
{
  SH2addi(sh, 4, (5<<4)+2);
}
#define init_7_4_5_3
void FASTCALL call_7_4_5_3(SH2_struct* sh)
{
  SH2addi(sh, 4, (5<<4)+3);
}
#define init_7_4_5_4
void FASTCALL call_7_4_5_4(SH2_struct* sh)
{
  SH2addi(sh, 4, (5<<4)+4);
}
#define init_7_4_5_5
void FASTCALL call_7_4_5_5(SH2_struct* sh)
{
  SH2addi(sh, 4, (5<<4)+5);
}
#define init_7_4_5_6
void FASTCALL call_7_4_5_6(SH2_struct* sh)
{
  SH2addi(sh, 4, (5<<4)+6);
}
#define init_7_4_5_7
void FASTCALL call_7_4_5_7(SH2_struct* sh)
{
  SH2addi(sh, 4, (5<<4)+7);
}
#define init_7_4_5_8
void FASTCALL call_7_4_5_8(SH2_struct* sh)
{
  SH2addi(sh, 4, (5<<4)+8);
}
#define init_7_4_5_9
void FASTCALL call_7_4_5_9(SH2_struct* sh)
{
  SH2addi(sh, 4, (5<<4)+9);
}
#define init_7_4_5_10
void FASTCALL call_7_4_5_10(SH2_struct* sh)
{
  SH2addi(sh, 4, (5<<4)+10);
}
#define init_7_4_5_11
void FASTCALL call_7_4_5_11(SH2_struct* sh)
{
  SH2addi(sh, 4, (5<<4)+11);
}
#define init_7_4_5_12
void FASTCALL call_7_4_5_12(SH2_struct* sh)
{
  SH2addi(sh, 4, (5<<4)+12);
}
#define init_7_4_5_13
void FASTCALL call_7_4_5_13(SH2_struct* sh)
{
  SH2addi(sh, 4, (5<<4)+13);
}
#define init_7_4_5_14
void FASTCALL call_7_4_5_14(SH2_struct* sh)
{
  SH2addi(sh, 4, (5<<4)+14);
}
#define init_7_4_5_15
void FASTCALL call_7_4_5_15(SH2_struct* sh)
{
  SH2addi(sh, 4, (5<<4)+15);
}
#define init_7_4_6_0
void FASTCALL call_7_4_6_0(SH2_struct* sh)
{
  SH2addi(sh, 4, (6<<4)+0);
}
#define init_7_4_6_1
void FASTCALL call_7_4_6_1(SH2_struct* sh)
{
  SH2addi(sh, 4, (6<<4)+1);
}
#define init_7_4_6_2
void FASTCALL call_7_4_6_2(SH2_struct* sh)
{
  SH2addi(sh, 4, (6<<4)+2);
}
#define init_7_4_6_3
void FASTCALL call_7_4_6_3(SH2_struct* sh)
{
  SH2addi(sh, 4, (6<<4)+3);
}
#define init_7_4_6_4
void FASTCALL call_7_4_6_4(SH2_struct* sh)
{
  SH2addi(sh, 4, (6<<4)+4);
}
#define init_7_4_6_5
void FASTCALL call_7_4_6_5(SH2_struct* sh)
{
  SH2addi(sh, 4, (6<<4)+5);
}
#define init_7_4_6_6
void FASTCALL call_7_4_6_6(SH2_struct* sh)
{
  SH2addi(sh, 4, (6<<4)+6);
}
#define init_7_4_6_7
void FASTCALL call_7_4_6_7(SH2_struct* sh)
{
  SH2addi(sh, 4, (6<<4)+7);
}
#define init_7_4_6_8
void FASTCALL call_7_4_6_8(SH2_struct* sh)
{
  SH2addi(sh, 4, (6<<4)+8);
}
#define init_7_4_6_9
void FASTCALL call_7_4_6_9(SH2_struct* sh)
{
  SH2addi(sh, 4, (6<<4)+9);
}
#define init_7_4_6_10
void FASTCALL call_7_4_6_10(SH2_struct* sh)
{
  SH2addi(sh, 4, (6<<4)+10);
}
#define init_7_4_6_11
void FASTCALL call_7_4_6_11(SH2_struct* sh)
{
  SH2addi(sh, 4, (6<<4)+11);
}
#define init_7_4_6_12
void FASTCALL call_7_4_6_12(SH2_struct* sh)
{
  SH2addi(sh, 4, (6<<4)+12);
}
#define init_7_4_6_13
void FASTCALL call_7_4_6_13(SH2_struct* sh)
{
  SH2addi(sh, 4, (6<<4)+13);
}
#define init_7_4_6_14
void FASTCALL call_7_4_6_14(SH2_struct* sh)
{
  SH2addi(sh, 4, (6<<4)+14);
}
#define init_7_4_6_15
void FASTCALL call_7_4_6_15(SH2_struct* sh)
{
  SH2addi(sh, 4, (6<<4)+15);
}
#define init_7_4_7_0
void FASTCALL call_7_4_7_0(SH2_struct* sh)
{
  SH2addi(sh, 4, (7<<4)+0);
}
#define init_7_4_7_1
void FASTCALL call_7_4_7_1(SH2_struct* sh)
{
  SH2addi(sh, 4, (7<<4)+1);
}
#define init_7_4_7_2
void FASTCALL call_7_4_7_2(SH2_struct* sh)
{
  SH2addi(sh, 4, (7<<4)+2);
}
#define init_7_4_7_3
void FASTCALL call_7_4_7_3(SH2_struct* sh)
{
  SH2addi(sh, 4, (7<<4)+3);
}
#define init_7_4_7_4
void FASTCALL call_7_4_7_4(SH2_struct* sh)
{
  SH2addi(sh, 4, (7<<4)+4);
}
#define init_7_4_7_5
void FASTCALL call_7_4_7_5(SH2_struct* sh)
{
  SH2addi(sh, 4, (7<<4)+5);
}
#define init_7_4_7_6
void FASTCALL call_7_4_7_6(SH2_struct* sh)
{
  SH2addi(sh, 4, (7<<4)+6);
}
#define init_7_4_7_7
void FASTCALL call_7_4_7_7(SH2_struct* sh)
{
  SH2addi(sh, 4, (7<<4)+7);
}
#define init_7_4_7_8
void FASTCALL call_7_4_7_8(SH2_struct* sh)
{
  SH2addi(sh, 4, (7<<4)+8);
}
#define init_7_4_7_9
void FASTCALL call_7_4_7_9(SH2_struct* sh)
{
  SH2addi(sh, 4, (7<<4)+9);
}
#define init_7_4_7_10
void FASTCALL call_7_4_7_10(SH2_struct* sh)
{
  SH2addi(sh, 4, (7<<4)+10);
}
#define init_7_4_7_11
void FASTCALL call_7_4_7_11(SH2_struct* sh)
{
  SH2addi(sh, 4, (7<<4)+11);
}
#define init_7_4_7_12
void FASTCALL call_7_4_7_12(SH2_struct* sh)
{
  SH2addi(sh, 4, (7<<4)+12);
}
#define init_7_4_7_13
void FASTCALL call_7_4_7_13(SH2_struct* sh)
{
  SH2addi(sh, 4, (7<<4)+13);
}
#define init_7_4_7_14
void FASTCALL call_7_4_7_14(SH2_struct* sh)
{
  SH2addi(sh, 4, (7<<4)+14);
}
#define init_7_4_7_15
void FASTCALL call_7_4_7_15(SH2_struct* sh)
{
  SH2addi(sh, 4, (7<<4)+15);
}
#define init_7_4_8_0
void FASTCALL call_7_4_8_0(SH2_struct* sh)
{
  SH2addi(sh, 4, (8<<4)+0);
}
#define init_7_4_8_1
void FASTCALL call_7_4_8_1(SH2_struct* sh)
{
  SH2addi(sh, 4, (8<<4)+1);
}
#define init_7_4_8_2
void FASTCALL call_7_4_8_2(SH2_struct* sh)
{
  SH2addi(sh, 4, (8<<4)+2);
}
#define init_7_4_8_3
void FASTCALL call_7_4_8_3(SH2_struct* sh)
{
  SH2addi(sh, 4, (8<<4)+3);
}
#define init_7_4_8_4
void FASTCALL call_7_4_8_4(SH2_struct* sh)
{
  SH2addi(sh, 4, (8<<4)+4);
}
#define init_7_4_8_5
void FASTCALL call_7_4_8_5(SH2_struct* sh)
{
  SH2addi(sh, 4, (8<<4)+5);
}
#define init_7_4_8_6
void FASTCALL call_7_4_8_6(SH2_struct* sh)
{
  SH2addi(sh, 4, (8<<4)+6);
}
#define init_7_4_8_7
void FASTCALL call_7_4_8_7(SH2_struct* sh)
{
  SH2addi(sh, 4, (8<<4)+7);
}
#define init_7_4_8_8
void FASTCALL call_7_4_8_8(SH2_struct* sh)
{
  SH2addi(sh, 4, (8<<4)+8);
}
#define init_7_4_8_9
void FASTCALL call_7_4_8_9(SH2_struct* sh)
{
  SH2addi(sh, 4, (8<<4)+9);
}
#define init_7_4_8_10
void FASTCALL call_7_4_8_10(SH2_struct* sh)
{
  SH2addi(sh, 4, (8<<4)+10);
}
#define init_7_4_8_11
void FASTCALL call_7_4_8_11(SH2_struct* sh)
{
  SH2addi(sh, 4, (8<<4)+11);
}
#define init_7_4_8_12
void FASTCALL call_7_4_8_12(SH2_struct* sh)
{
  SH2addi(sh, 4, (8<<4)+12);
}
#define init_7_4_8_13
void FASTCALL call_7_4_8_13(SH2_struct* sh)
{
  SH2addi(sh, 4, (8<<4)+13);
}
#define init_7_4_8_14
void FASTCALL call_7_4_8_14(SH2_struct* sh)
{
  SH2addi(sh, 4, (8<<4)+14);
}
#define init_7_4_8_15
void FASTCALL call_7_4_8_15(SH2_struct* sh)
{
  SH2addi(sh, 4, (8<<4)+15);
}
#define init_7_4_9_0
void FASTCALL call_7_4_9_0(SH2_struct* sh)
{
  SH2addi(sh, 4, (9<<4)+0);
}
#define init_7_4_9_1
void FASTCALL call_7_4_9_1(SH2_struct* sh)
{
  SH2addi(sh, 4, (9<<4)+1);
}
#define init_7_4_9_2
void FASTCALL call_7_4_9_2(SH2_struct* sh)
{
  SH2addi(sh, 4, (9<<4)+2);
}
#define init_7_4_9_3
void FASTCALL call_7_4_9_3(SH2_struct* sh)
{
  SH2addi(sh, 4, (9<<4)+3);
}
#define init_7_4_9_4
void FASTCALL call_7_4_9_4(SH2_struct* sh)
{
  SH2addi(sh, 4, (9<<4)+4);
}
#define init_7_4_9_5
void FASTCALL call_7_4_9_5(SH2_struct* sh)
{
  SH2addi(sh, 4, (9<<4)+5);
}
#define init_7_4_9_6
void FASTCALL call_7_4_9_6(SH2_struct* sh)
{
  SH2addi(sh, 4, (9<<4)+6);
}
#define init_7_4_9_7
void FASTCALL call_7_4_9_7(SH2_struct* sh)
{
  SH2addi(sh, 4, (9<<4)+7);
}
#define init_7_4_9_8
void FASTCALL call_7_4_9_8(SH2_struct* sh)
{
  SH2addi(sh, 4, (9<<4)+8);
}
#define init_7_4_9_9
void FASTCALL call_7_4_9_9(SH2_struct* sh)
{
  SH2addi(sh, 4, (9<<4)+9);
}
#define init_7_4_9_10
void FASTCALL call_7_4_9_10(SH2_struct* sh)
{
  SH2addi(sh, 4, (9<<4)+10);
}
#define init_7_4_9_11
void FASTCALL call_7_4_9_11(SH2_struct* sh)
{
  SH2addi(sh, 4, (9<<4)+11);
}
#define init_7_4_9_12
void FASTCALL call_7_4_9_12(SH2_struct* sh)
{
  SH2addi(sh, 4, (9<<4)+12);
}
#define init_7_4_9_13
void FASTCALL call_7_4_9_13(SH2_struct* sh)
{
  SH2addi(sh, 4, (9<<4)+13);
}
#define init_7_4_9_14
void FASTCALL call_7_4_9_14(SH2_struct* sh)
{
  SH2addi(sh, 4, (9<<4)+14);
}
#define init_7_4_9_15
void FASTCALL call_7_4_9_15(SH2_struct* sh)
{
  SH2addi(sh, 4, (9<<4)+15);
}
#define init_7_4_10_0
void FASTCALL call_7_4_10_0(SH2_struct* sh)
{
  SH2addi(sh, 4, (10<<4)+0);
}
#define init_7_4_10_1
void FASTCALL call_7_4_10_1(SH2_struct* sh)
{
  SH2addi(sh, 4, (10<<4)+1);
}
#define init_7_4_10_2
void FASTCALL call_7_4_10_2(SH2_struct* sh)
{
  SH2addi(sh, 4, (10<<4)+2);
}
#define init_7_4_10_3
void FASTCALL call_7_4_10_3(SH2_struct* sh)
{
  SH2addi(sh, 4, (10<<4)+3);
}
#define init_7_4_10_4
void FASTCALL call_7_4_10_4(SH2_struct* sh)
{
  SH2addi(sh, 4, (10<<4)+4);
}
#define init_7_4_10_5
void FASTCALL call_7_4_10_5(SH2_struct* sh)
{
  SH2addi(sh, 4, (10<<4)+5);
}
#define init_7_4_10_6
void FASTCALL call_7_4_10_6(SH2_struct* sh)
{
  SH2addi(sh, 4, (10<<4)+6);
}
#define init_7_4_10_7
void FASTCALL call_7_4_10_7(SH2_struct* sh)
{
  SH2addi(sh, 4, (10<<4)+7);
}
#define init_7_4_10_8
void FASTCALL call_7_4_10_8(SH2_struct* sh)
{
  SH2addi(sh, 4, (10<<4)+8);
}
#define init_7_4_10_9
void FASTCALL call_7_4_10_9(SH2_struct* sh)
{
  SH2addi(sh, 4, (10<<4)+9);
}
#define init_7_4_10_10
void FASTCALL call_7_4_10_10(SH2_struct* sh)
{
  SH2addi(sh, 4, (10<<4)+10);
}
#define init_7_4_10_11
void FASTCALL call_7_4_10_11(SH2_struct* sh)
{
  SH2addi(sh, 4, (10<<4)+11);
}
#define init_7_4_10_12
void FASTCALL call_7_4_10_12(SH2_struct* sh)
{
  SH2addi(sh, 4, (10<<4)+12);
}
#define init_7_4_10_13
void FASTCALL call_7_4_10_13(SH2_struct* sh)
{
  SH2addi(sh, 4, (10<<4)+13);
}
#define init_7_4_10_14
void FASTCALL call_7_4_10_14(SH2_struct* sh)
{
  SH2addi(sh, 4, (10<<4)+14);
}
#define init_7_4_10_15
void FASTCALL call_7_4_10_15(SH2_struct* sh)
{
  SH2addi(sh, 4, (10<<4)+15);
}
#define init_7_4_11_0
void FASTCALL call_7_4_11_0(SH2_struct* sh)
{
  SH2addi(sh, 4, (11<<4)+0);
}
#define init_7_4_11_1
void FASTCALL call_7_4_11_1(SH2_struct* sh)
{
  SH2addi(sh, 4, (11<<4)+1);
}
#define init_7_4_11_2
void FASTCALL call_7_4_11_2(SH2_struct* sh)
{
  SH2addi(sh, 4, (11<<4)+2);
}
#define init_7_4_11_3
void FASTCALL call_7_4_11_3(SH2_struct* sh)
{
  SH2addi(sh, 4, (11<<4)+3);
}
#define init_7_4_11_4
void FASTCALL call_7_4_11_4(SH2_struct* sh)
{
  SH2addi(sh, 4, (11<<4)+4);
}
#define init_7_4_11_5
void FASTCALL call_7_4_11_5(SH2_struct* sh)
{
  SH2addi(sh, 4, (11<<4)+5);
}
#define init_7_4_11_6
void FASTCALL call_7_4_11_6(SH2_struct* sh)
{
  SH2addi(sh, 4, (11<<4)+6);
}
#define init_7_4_11_7
void FASTCALL call_7_4_11_7(SH2_struct* sh)
{
  SH2addi(sh, 4, (11<<4)+7);
}
#define init_7_4_11_8
void FASTCALL call_7_4_11_8(SH2_struct* sh)
{
  SH2addi(sh, 4, (11<<4)+8);
}
#define init_7_4_11_9
void FASTCALL call_7_4_11_9(SH2_struct* sh)
{
  SH2addi(sh, 4, (11<<4)+9);
}
#define init_7_4_11_10
void FASTCALL call_7_4_11_10(SH2_struct* sh)
{
  SH2addi(sh, 4, (11<<4)+10);
}
#define init_7_4_11_11
void FASTCALL call_7_4_11_11(SH2_struct* sh)
{
  SH2addi(sh, 4, (11<<4)+11);
}
#define init_7_4_11_12
void FASTCALL call_7_4_11_12(SH2_struct* sh)
{
  SH2addi(sh, 4, (11<<4)+12);
}
#define init_7_4_11_13
void FASTCALL call_7_4_11_13(SH2_struct* sh)
{
  SH2addi(sh, 4, (11<<4)+13);
}
#define init_7_4_11_14
void FASTCALL call_7_4_11_14(SH2_struct* sh)
{
  SH2addi(sh, 4, (11<<4)+14);
}
#define init_7_4_11_15
void FASTCALL call_7_4_11_15(SH2_struct* sh)
{
  SH2addi(sh, 4, (11<<4)+15);
}
#define init_7_4_12_0
void FASTCALL call_7_4_12_0(SH2_struct* sh)
{
  SH2addi(sh, 4, (12<<4)+0);
}
#define init_7_4_12_1
void FASTCALL call_7_4_12_1(SH2_struct* sh)
{
  SH2addi(sh, 4, (12<<4)+1);
}
#define init_7_4_12_2
void FASTCALL call_7_4_12_2(SH2_struct* sh)
{
  SH2addi(sh, 4, (12<<4)+2);
}
#define init_7_4_12_3
void FASTCALL call_7_4_12_3(SH2_struct* sh)
{
  SH2addi(sh, 4, (12<<4)+3);
}
#define init_7_4_12_4
void FASTCALL call_7_4_12_4(SH2_struct* sh)
{
  SH2addi(sh, 4, (12<<4)+4);
}
#define init_7_4_12_5
void FASTCALL call_7_4_12_5(SH2_struct* sh)
{
  SH2addi(sh, 4, (12<<4)+5);
}
#define init_7_4_12_6
void FASTCALL call_7_4_12_6(SH2_struct* sh)
{
  SH2addi(sh, 4, (12<<4)+6);
}
#define init_7_4_12_7
void FASTCALL call_7_4_12_7(SH2_struct* sh)
{
  SH2addi(sh, 4, (12<<4)+7);
}
#define init_7_4_12_8
void FASTCALL call_7_4_12_8(SH2_struct* sh)
{
  SH2addi(sh, 4, (12<<4)+8);
}
#define init_7_4_12_9
void FASTCALL call_7_4_12_9(SH2_struct* sh)
{
  SH2addi(sh, 4, (12<<4)+9);
}
#define init_7_4_12_10
void FASTCALL call_7_4_12_10(SH2_struct* sh)
{
  SH2addi(sh, 4, (12<<4)+10);
}
#define init_7_4_12_11
void FASTCALL call_7_4_12_11(SH2_struct* sh)
{
  SH2addi(sh, 4, (12<<4)+11);
}
#define init_7_4_12_12
void FASTCALL call_7_4_12_12(SH2_struct* sh)
{
  SH2addi(sh, 4, (12<<4)+12);
}
#define init_7_4_12_13
void FASTCALL call_7_4_12_13(SH2_struct* sh)
{
  SH2addi(sh, 4, (12<<4)+13);
}
#define init_7_4_12_14
void FASTCALL call_7_4_12_14(SH2_struct* sh)
{
  SH2addi(sh, 4, (12<<4)+14);
}
#define init_7_4_12_15
void FASTCALL call_7_4_12_15(SH2_struct* sh)
{
  SH2addi(sh, 4, (12<<4)+15);
}
#define init_7_4_13_0
void FASTCALL call_7_4_13_0(SH2_struct* sh)
{
  SH2addi(sh, 4, (13<<4)+0);
}
#define init_7_4_13_1
void FASTCALL call_7_4_13_1(SH2_struct* sh)
{
  SH2addi(sh, 4, (13<<4)+1);
}
#define init_7_4_13_2
void FASTCALL call_7_4_13_2(SH2_struct* sh)
{
  SH2addi(sh, 4, (13<<4)+2);
}
#define init_7_4_13_3
void FASTCALL call_7_4_13_3(SH2_struct* sh)
{
  SH2addi(sh, 4, (13<<4)+3);
}
#define init_7_4_13_4
void FASTCALL call_7_4_13_4(SH2_struct* sh)
{
  SH2addi(sh, 4, (13<<4)+4);
}
#define init_7_4_13_5
void FASTCALL call_7_4_13_5(SH2_struct* sh)
{
  SH2addi(sh, 4, (13<<4)+5);
}
#define init_7_4_13_6
void FASTCALL call_7_4_13_6(SH2_struct* sh)
{
  SH2addi(sh, 4, (13<<4)+6);
}
#define init_7_4_13_7
void FASTCALL call_7_4_13_7(SH2_struct* sh)
{
  SH2addi(sh, 4, (13<<4)+7);
}
#define init_7_4_13_8
void FASTCALL call_7_4_13_8(SH2_struct* sh)
{
  SH2addi(sh, 4, (13<<4)+8);
}
#define init_7_4_13_9
void FASTCALL call_7_4_13_9(SH2_struct* sh)
{
  SH2addi(sh, 4, (13<<4)+9);
}
#define init_7_4_13_10
void FASTCALL call_7_4_13_10(SH2_struct* sh)
{
  SH2addi(sh, 4, (13<<4)+10);
}
#define init_7_4_13_11
void FASTCALL call_7_4_13_11(SH2_struct* sh)
{
  SH2addi(sh, 4, (13<<4)+11);
}
#define init_7_4_13_12
void FASTCALL call_7_4_13_12(SH2_struct* sh)
{
  SH2addi(sh, 4, (13<<4)+12);
}
#define init_7_4_13_13
void FASTCALL call_7_4_13_13(SH2_struct* sh)
{
  SH2addi(sh, 4, (13<<4)+13);
}
#define init_7_4_13_14
void FASTCALL call_7_4_13_14(SH2_struct* sh)
{
  SH2addi(sh, 4, (13<<4)+14);
}
#define init_7_4_13_15
void FASTCALL call_7_4_13_15(SH2_struct* sh)
{
  SH2addi(sh, 4, (13<<4)+15);
}
#define init_7_4_14_0
void FASTCALL call_7_4_14_0(SH2_struct* sh)
{
  SH2addi(sh, 4, (14<<4)+0);
}
#define init_7_4_14_1
void FASTCALL call_7_4_14_1(SH2_struct* sh)
{
  SH2addi(sh, 4, (14<<4)+1);
}
#define init_7_4_14_2
void FASTCALL call_7_4_14_2(SH2_struct* sh)
{
  SH2addi(sh, 4, (14<<4)+2);
}
#define init_7_4_14_3
void FASTCALL call_7_4_14_3(SH2_struct* sh)
{
  SH2addi(sh, 4, (14<<4)+3);
}
#define init_7_4_14_4
void FASTCALL call_7_4_14_4(SH2_struct* sh)
{
  SH2addi(sh, 4, (14<<4)+4);
}
#define init_7_4_14_5
void FASTCALL call_7_4_14_5(SH2_struct* sh)
{
  SH2addi(sh, 4, (14<<4)+5);
}
#define init_7_4_14_6
void FASTCALL call_7_4_14_6(SH2_struct* sh)
{
  SH2addi(sh, 4, (14<<4)+6);
}
#define init_7_4_14_7
void FASTCALL call_7_4_14_7(SH2_struct* sh)
{
  SH2addi(sh, 4, (14<<4)+7);
}
#define init_7_4_14_8
void FASTCALL call_7_4_14_8(SH2_struct* sh)
{
  SH2addi(sh, 4, (14<<4)+8);
}
#define init_7_4_14_9
void FASTCALL call_7_4_14_9(SH2_struct* sh)
{
  SH2addi(sh, 4, (14<<4)+9);
}
#define init_7_4_14_10
void FASTCALL call_7_4_14_10(SH2_struct* sh)
{
  SH2addi(sh, 4, (14<<4)+10);
}
#define init_7_4_14_11
void FASTCALL call_7_4_14_11(SH2_struct* sh)
{
  SH2addi(sh, 4, (14<<4)+11);
}
#define init_7_4_14_12
void FASTCALL call_7_4_14_12(SH2_struct* sh)
{
  SH2addi(sh, 4, (14<<4)+12);
}
#define init_7_4_14_13
void FASTCALL call_7_4_14_13(SH2_struct* sh)
{
  SH2addi(sh, 4, (14<<4)+13);
}
#define init_7_4_14_14
void FASTCALL call_7_4_14_14(SH2_struct* sh)
{
  SH2addi(sh, 4, (14<<4)+14);
}
#define init_7_4_14_15
void FASTCALL call_7_4_14_15(SH2_struct* sh)
{
  SH2addi(sh, 4, (14<<4)+15);
}
#define init_7_4_15_0
void FASTCALL call_7_4_15_0(SH2_struct* sh)
{
  SH2addi(sh, 4, (15<<4)+0);
}
#define init_7_4_15_1
void FASTCALL call_7_4_15_1(SH2_struct* sh)
{
  SH2addi(sh, 4, (15<<4)+1);
}
#define init_7_4_15_2
void FASTCALL call_7_4_15_2(SH2_struct* sh)
{
  SH2addi(sh, 4, (15<<4)+2);
}
#define init_7_4_15_3
void FASTCALL call_7_4_15_3(SH2_struct* sh)
{
  SH2addi(sh, 4, (15<<4)+3);
}
#define init_7_4_15_4
void FASTCALL call_7_4_15_4(SH2_struct* sh)
{
  SH2addi(sh, 4, (15<<4)+4);
}
#define init_7_4_15_5
void FASTCALL call_7_4_15_5(SH2_struct* sh)
{
  SH2addi(sh, 4, (15<<4)+5);
}
#define init_7_4_15_6
void FASTCALL call_7_4_15_6(SH2_struct* sh)
{
  SH2addi(sh, 4, (15<<4)+6);
}
#define init_7_4_15_7
void FASTCALL call_7_4_15_7(SH2_struct* sh)
{
  SH2addi(sh, 4, (15<<4)+7);
}
#define init_7_4_15_8
void FASTCALL call_7_4_15_8(SH2_struct* sh)
{
  SH2addi(sh, 4, (15<<4)+8);
}
#define init_7_4_15_9
void FASTCALL call_7_4_15_9(SH2_struct* sh)
{
  SH2addi(sh, 4, (15<<4)+9);
}
#define init_7_4_15_10
void FASTCALL call_7_4_15_10(SH2_struct* sh)
{
  SH2addi(sh, 4, (15<<4)+10);
}
#define init_7_4_15_11
void FASTCALL call_7_4_15_11(SH2_struct* sh)
{
  SH2addi(sh, 4, (15<<4)+11);
}
#define init_7_4_15_12
void FASTCALL call_7_4_15_12(SH2_struct* sh)
{
  SH2addi(sh, 4, (15<<4)+12);
}
#define init_7_4_15_13
void FASTCALL call_7_4_15_13(SH2_struct* sh)
{
  SH2addi(sh, 4, (15<<4)+13);
}
#define init_7_4_15_14
void FASTCALL call_7_4_15_14(SH2_struct* sh)
{
  SH2addi(sh, 4, (15<<4)+14);
}
#define init_7_4_15_15
void FASTCALL call_7_4_15_15(SH2_struct* sh)
{
  SH2addi(sh, 4, (15<<4)+15);
}
#define init_7_5_0_0
void FASTCALL call_7_5_0_0(SH2_struct* sh)
{
  SH2addi(sh, 5, (0<<4)+0);
}
#define init_7_5_0_1
void FASTCALL call_7_5_0_1(SH2_struct* sh)
{
  SH2addi(sh, 5, (0<<4)+1);
}
#define init_7_5_0_2
void FASTCALL call_7_5_0_2(SH2_struct* sh)
{
  SH2addi(sh, 5, (0<<4)+2);
}
#define init_7_5_0_3
void FASTCALL call_7_5_0_3(SH2_struct* sh)
{
  SH2addi(sh, 5, (0<<4)+3);
}
#define init_7_5_0_4
void FASTCALL call_7_5_0_4(SH2_struct* sh)
{
  SH2addi(sh, 5, (0<<4)+4);
}
#define init_7_5_0_5
void FASTCALL call_7_5_0_5(SH2_struct* sh)
{
  SH2addi(sh, 5, (0<<4)+5);
}
#define init_7_5_0_6
void FASTCALL call_7_5_0_6(SH2_struct* sh)
{
  SH2addi(sh, 5, (0<<4)+6);
}
#define init_7_5_0_7
void FASTCALL call_7_5_0_7(SH2_struct* sh)
{
  SH2addi(sh, 5, (0<<4)+7);
}
#define init_7_5_0_8
void FASTCALL call_7_5_0_8(SH2_struct* sh)
{
  SH2addi(sh, 5, (0<<4)+8);
}
#define init_7_5_0_9
void FASTCALL call_7_5_0_9(SH2_struct* sh)
{
  SH2addi(sh, 5, (0<<4)+9);
}
#define init_7_5_0_10
void FASTCALL call_7_5_0_10(SH2_struct* sh)
{
  SH2addi(sh, 5, (0<<4)+10);
}
#define init_7_5_0_11
void FASTCALL call_7_5_0_11(SH2_struct* sh)
{
  SH2addi(sh, 5, (0<<4)+11);
}
#define init_7_5_0_12
void FASTCALL call_7_5_0_12(SH2_struct* sh)
{
  SH2addi(sh, 5, (0<<4)+12);
}
#define init_7_5_0_13
void FASTCALL call_7_5_0_13(SH2_struct* sh)
{
  SH2addi(sh, 5, (0<<4)+13);
}
#define init_7_5_0_14
void FASTCALL call_7_5_0_14(SH2_struct* sh)
{
  SH2addi(sh, 5, (0<<4)+14);
}
#define init_7_5_0_15
void FASTCALL call_7_5_0_15(SH2_struct* sh)
{
  SH2addi(sh, 5, (0<<4)+15);
}
#define init_7_5_1_0
void FASTCALL call_7_5_1_0(SH2_struct* sh)
{
  SH2addi(sh, 5, (1<<4)+0);
}
#define init_7_5_1_1
void FASTCALL call_7_5_1_1(SH2_struct* sh)
{
  SH2addi(sh, 5, (1<<4)+1);
}
#define init_7_5_1_2
void FASTCALL call_7_5_1_2(SH2_struct* sh)
{
  SH2addi(sh, 5, (1<<4)+2);
}
#define init_7_5_1_3
void FASTCALL call_7_5_1_3(SH2_struct* sh)
{
  SH2addi(sh, 5, (1<<4)+3);
}
#define init_7_5_1_4
void FASTCALL call_7_5_1_4(SH2_struct* sh)
{
  SH2addi(sh, 5, (1<<4)+4);
}
#define init_7_5_1_5
void FASTCALL call_7_5_1_5(SH2_struct* sh)
{
  SH2addi(sh, 5, (1<<4)+5);
}
#define init_7_5_1_6
void FASTCALL call_7_5_1_6(SH2_struct* sh)
{
  SH2addi(sh, 5, (1<<4)+6);
}
#define init_7_5_1_7
void FASTCALL call_7_5_1_7(SH2_struct* sh)
{
  SH2addi(sh, 5, (1<<4)+7);
}
#define init_7_5_1_8
void FASTCALL call_7_5_1_8(SH2_struct* sh)
{
  SH2addi(sh, 5, (1<<4)+8);
}
#define init_7_5_1_9
void FASTCALL call_7_5_1_9(SH2_struct* sh)
{
  SH2addi(sh, 5, (1<<4)+9);
}
#define init_7_5_1_10
void FASTCALL call_7_5_1_10(SH2_struct* sh)
{
  SH2addi(sh, 5, (1<<4)+10);
}
#define init_7_5_1_11
void FASTCALL call_7_5_1_11(SH2_struct* sh)
{
  SH2addi(sh, 5, (1<<4)+11);
}
#define init_7_5_1_12
void FASTCALL call_7_5_1_12(SH2_struct* sh)
{
  SH2addi(sh, 5, (1<<4)+12);
}
#define init_7_5_1_13
void FASTCALL call_7_5_1_13(SH2_struct* sh)
{
  SH2addi(sh, 5, (1<<4)+13);
}
#define init_7_5_1_14
void FASTCALL call_7_5_1_14(SH2_struct* sh)
{
  SH2addi(sh, 5, (1<<4)+14);
}
#define init_7_5_1_15
void FASTCALL call_7_5_1_15(SH2_struct* sh)
{
  SH2addi(sh, 5, (1<<4)+15);
}
#define init_7_5_2_0
void FASTCALL call_7_5_2_0(SH2_struct* sh)
{
  SH2addi(sh, 5, (2<<4)+0);
}
#define init_7_5_2_1
void FASTCALL call_7_5_2_1(SH2_struct* sh)
{
  SH2addi(sh, 5, (2<<4)+1);
}
#define init_7_5_2_2
void FASTCALL call_7_5_2_2(SH2_struct* sh)
{
  SH2addi(sh, 5, (2<<4)+2);
}
#define init_7_5_2_3
void FASTCALL call_7_5_2_3(SH2_struct* sh)
{
  SH2addi(sh, 5, (2<<4)+3);
}
#define init_7_5_2_4
void FASTCALL call_7_5_2_4(SH2_struct* sh)
{
  SH2addi(sh, 5, (2<<4)+4);
}
#define init_7_5_2_5
void FASTCALL call_7_5_2_5(SH2_struct* sh)
{
  SH2addi(sh, 5, (2<<4)+5);
}
#define init_7_5_2_6
void FASTCALL call_7_5_2_6(SH2_struct* sh)
{
  SH2addi(sh, 5, (2<<4)+6);
}
#define init_7_5_2_7
void FASTCALL call_7_5_2_7(SH2_struct* sh)
{
  SH2addi(sh, 5, (2<<4)+7);
}
#define init_7_5_2_8
void FASTCALL call_7_5_2_8(SH2_struct* sh)
{
  SH2addi(sh, 5, (2<<4)+8);
}
#define init_7_5_2_9
void FASTCALL call_7_5_2_9(SH2_struct* sh)
{
  SH2addi(sh, 5, (2<<4)+9);
}
#define init_7_5_2_10
void FASTCALL call_7_5_2_10(SH2_struct* sh)
{
  SH2addi(sh, 5, (2<<4)+10);
}
#define init_7_5_2_11
void FASTCALL call_7_5_2_11(SH2_struct* sh)
{
  SH2addi(sh, 5, (2<<4)+11);
}
#define init_7_5_2_12
void FASTCALL call_7_5_2_12(SH2_struct* sh)
{
  SH2addi(sh, 5, (2<<4)+12);
}
#define init_7_5_2_13
void FASTCALL call_7_5_2_13(SH2_struct* sh)
{
  SH2addi(sh, 5, (2<<4)+13);
}
#define init_7_5_2_14
void FASTCALL call_7_5_2_14(SH2_struct* sh)
{
  SH2addi(sh, 5, (2<<4)+14);
}
#define init_7_5_2_15
void FASTCALL call_7_5_2_15(SH2_struct* sh)
{
  SH2addi(sh, 5, (2<<4)+15);
}
#define init_7_5_3_0
void FASTCALL call_7_5_3_0(SH2_struct* sh)
{
  SH2addi(sh, 5, (3<<4)+0);
}
#define init_7_5_3_1
void FASTCALL call_7_5_3_1(SH2_struct* sh)
{
  SH2addi(sh, 5, (3<<4)+1);
}
#define init_7_5_3_2
void FASTCALL call_7_5_3_2(SH2_struct* sh)
{
  SH2addi(sh, 5, (3<<4)+2);
}
#define init_7_5_3_3
void FASTCALL call_7_5_3_3(SH2_struct* sh)
{
  SH2addi(sh, 5, (3<<4)+3);
}
#define init_7_5_3_4
void FASTCALL call_7_5_3_4(SH2_struct* sh)
{
  SH2addi(sh, 5, (3<<4)+4);
}
#define init_7_5_3_5
void FASTCALL call_7_5_3_5(SH2_struct* sh)
{
  SH2addi(sh, 5, (3<<4)+5);
}
#define init_7_5_3_6
void FASTCALL call_7_5_3_6(SH2_struct* sh)
{
  SH2addi(sh, 5, (3<<4)+6);
}
#define init_7_5_3_7
void FASTCALL call_7_5_3_7(SH2_struct* sh)
{
  SH2addi(sh, 5, (3<<4)+7);
}
#define init_7_5_3_8
void FASTCALL call_7_5_3_8(SH2_struct* sh)
{
  SH2addi(sh, 5, (3<<4)+8);
}
#define init_7_5_3_9
void FASTCALL call_7_5_3_9(SH2_struct* sh)
{
  SH2addi(sh, 5, (3<<4)+9);
}
#define init_7_5_3_10
void FASTCALL call_7_5_3_10(SH2_struct* sh)
{
  SH2addi(sh, 5, (3<<4)+10);
}
#define init_7_5_3_11
void FASTCALL call_7_5_3_11(SH2_struct* sh)
{
  SH2addi(sh, 5, (3<<4)+11);
}
#define init_7_5_3_12
void FASTCALL call_7_5_3_12(SH2_struct* sh)
{
  SH2addi(sh, 5, (3<<4)+12);
}
#define init_7_5_3_13
void FASTCALL call_7_5_3_13(SH2_struct* sh)
{
  SH2addi(sh, 5, (3<<4)+13);
}
#define init_7_5_3_14
void FASTCALL call_7_5_3_14(SH2_struct* sh)
{
  SH2addi(sh, 5, (3<<4)+14);
}
#define init_7_5_3_15
void FASTCALL call_7_5_3_15(SH2_struct* sh)
{
  SH2addi(sh, 5, (3<<4)+15);
}
#define init_7_5_4_0
void FASTCALL call_7_5_4_0(SH2_struct* sh)
{
  SH2addi(sh, 5, (4<<4)+0);
}
#define init_7_5_4_1
void FASTCALL call_7_5_4_1(SH2_struct* sh)
{
  SH2addi(sh, 5, (4<<4)+1);
}
#define init_7_5_4_2
void FASTCALL call_7_5_4_2(SH2_struct* sh)
{
  SH2addi(sh, 5, (4<<4)+2);
}
#define init_7_5_4_3
void FASTCALL call_7_5_4_3(SH2_struct* sh)
{
  SH2addi(sh, 5, (4<<4)+3);
}
#define init_7_5_4_4
void FASTCALL call_7_5_4_4(SH2_struct* sh)
{
  SH2addi(sh, 5, (4<<4)+4);
}
#define init_7_5_4_5
void FASTCALL call_7_5_4_5(SH2_struct* sh)
{
  SH2addi(sh, 5, (4<<4)+5);
}
#define init_7_5_4_6
void FASTCALL call_7_5_4_6(SH2_struct* sh)
{
  SH2addi(sh, 5, (4<<4)+6);
}
#define init_7_5_4_7
void FASTCALL call_7_5_4_7(SH2_struct* sh)
{
  SH2addi(sh, 5, (4<<4)+7);
}
#define init_7_5_4_8
void FASTCALL call_7_5_4_8(SH2_struct* sh)
{
  SH2addi(sh, 5, (4<<4)+8);
}
#define init_7_5_4_9
void FASTCALL call_7_5_4_9(SH2_struct* sh)
{
  SH2addi(sh, 5, (4<<4)+9);
}
#define init_7_5_4_10
void FASTCALL call_7_5_4_10(SH2_struct* sh)
{
  SH2addi(sh, 5, (4<<4)+10);
}
#define init_7_5_4_11
void FASTCALL call_7_5_4_11(SH2_struct* sh)
{
  SH2addi(sh, 5, (4<<4)+11);
}
#define init_7_5_4_12
void FASTCALL call_7_5_4_12(SH2_struct* sh)
{
  SH2addi(sh, 5, (4<<4)+12);
}
#define init_7_5_4_13
void FASTCALL call_7_5_4_13(SH2_struct* sh)
{
  SH2addi(sh, 5, (4<<4)+13);
}
#define init_7_5_4_14
void FASTCALL call_7_5_4_14(SH2_struct* sh)
{
  SH2addi(sh, 5, (4<<4)+14);
}
#define init_7_5_4_15
void FASTCALL call_7_5_4_15(SH2_struct* sh)
{
  SH2addi(sh, 5, (4<<4)+15);
}
#define init_7_5_5_0
void FASTCALL call_7_5_5_0(SH2_struct* sh)
{
  SH2addi(sh, 5, (5<<4)+0);
}
#define init_7_5_5_1
void FASTCALL call_7_5_5_1(SH2_struct* sh)
{
  SH2addi(sh, 5, (5<<4)+1);
}
#define init_7_5_5_2
void FASTCALL call_7_5_5_2(SH2_struct* sh)
{
  SH2addi(sh, 5, (5<<4)+2);
}
#define init_7_5_5_3
void FASTCALL call_7_5_5_3(SH2_struct* sh)
{
  SH2addi(sh, 5, (5<<4)+3);
}
#define init_7_5_5_4
void FASTCALL call_7_5_5_4(SH2_struct* sh)
{
  SH2addi(sh, 5, (5<<4)+4);
}
#define init_7_5_5_5
void FASTCALL call_7_5_5_5(SH2_struct* sh)
{
  SH2addi(sh, 5, (5<<4)+5);
}
#define init_7_5_5_6
void FASTCALL call_7_5_5_6(SH2_struct* sh)
{
  SH2addi(sh, 5, (5<<4)+6);
}
#define init_7_5_5_7
void FASTCALL call_7_5_5_7(SH2_struct* sh)
{
  SH2addi(sh, 5, (5<<4)+7);
}
#define init_7_5_5_8
void FASTCALL call_7_5_5_8(SH2_struct* sh)
{
  SH2addi(sh, 5, (5<<4)+8);
}
#define init_7_5_5_9
void FASTCALL call_7_5_5_9(SH2_struct* sh)
{
  SH2addi(sh, 5, (5<<4)+9);
}
#define init_7_5_5_10
void FASTCALL call_7_5_5_10(SH2_struct* sh)
{
  SH2addi(sh, 5, (5<<4)+10);
}
#define init_7_5_5_11
void FASTCALL call_7_5_5_11(SH2_struct* sh)
{
  SH2addi(sh, 5, (5<<4)+11);
}
#define init_7_5_5_12
void FASTCALL call_7_5_5_12(SH2_struct* sh)
{
  SH2addi(sh, 5, (5<<4)+12);
}
#define init_7_5_5_13
void FASTCALL call_7_5_5_13(SH2_struct* sh)
{
  SH2addi(sh, 5, (5<<4)+13);
}
#define init_7_5_5_14
void FASTCALL call_7_5_5_14(SH2_struct* sh)
{
  SH2addi(sh, 5, (5<<4)+14);
}
#define init_7_5_5_15
void FASTCALL call_7_5_5_15(SH2_struct* sh)
{
  SH2addi(sh, 5, (5<<4)+15);
}
#define init_7_5_6_0
void FASTCALL call_7_5_6_0(SH2_struct* sh)
{
  SH2addi(sh, 5, (6<<4)+0);
}
#define init_7_5_6_1
void FASTCALL call_7_5_6_1(SH2_struct* sh)
{
  SH2addi(sh, 5, (6<<4)+1);
}
#define init_7_5_6_2
void FASTCALL call_7_5_6_2(SH2_struct* sh)
{
  SH2addi(sh, 5, (6<<4)+2);
}
#define init_7_5_6_3
void FASTCALL call_7_5_6_3(SH2_struct* sh)
{
  SH2addi(sh, 5, (6<<4)+3);
}
#define init_7_5_6_4
void FASTCALL call_7_5_6_4(SH2_struct* sh)
{
  SH2addi(sh, 5, (6<<4)+4);
}
#define init_7_5_6_5
void FASTCALL call_7_5_6_5(SH2_struct* sh)
{
  SH2addi(sh, 5, (6<<4)+5);
}
#define init_7_5_6_6
void FASTCALL call_7_5_6_6(SH2_struct* sh)
{
  SH2addi(sh, 5, (6<<4)+6);
}
#define init_7_5_6_7
void FASTCALL call_7_5_6_7(SH2_struct* sh)
{
  SH2addi(sh, 5, (6<<4)+7);
}
#define init_7_5_6_8
void FASTCALL call_7_5_6_8(SH2_struct* sh)
{
  SH2addi(sh, 5, (6<<4)+8);
}
#define init_7_5_6_9
void FASTCALL call_7_5_6_9(SH2_struct* sh)
{
  SH2addi(sh, 5, (6<<4)+9);
}
#define init_7_5_6_10
void FASTCALL call_7_5_6_10(SH2_struct* sh)
{
  SH2addi(sh, 5, (6<<4)+10);
}
#define init_7_5_6_11
void FASTCALL call_7_5_6_11(SH2_struct* sh)
{
  SH2addi(sh, 5, (6<<4)+11);
}
#define init_7_5_6_12
void FASTCALL call_7_5_6_12(SH2_struct* sh)
{
  SH2addi(sh, 5, (6<<4)+12);
}
#define init_7_5_6_13
void FASTCALL call_7_5_6_13(SH2_struct* sh)
{
  SH2addi(sh, 5, (6<<4)+13);
}
#define init_7_5_6_14
void FASTCALL call_7_5_6_14(SH2_struct* sh)
{
  SH2addi(sh, 5, (6<<4)+14);
}
#define init_7_5_6_15
void FASTCALL call_7_5_6_15(SH2_struct* sh)
{
  SH2addi(sh, 5, (6<<4)+15);
}
#define init_7_5_7_0
void FASTCALL call_7_5_7_0(SH2_struct* sh)
{
  SH2addi(sh, 5, (7<<4)+0);
}
#define init_7_5_7_1
void FASTCALL call_7_5_7_1(SH2_struct* sh)
{
  SH2addi(sh, 5, (7<<4)+1);
}
#define init_7_5_7_2
void FASTCALL call_7_5_7_2(SH2_struct* sh)
{
  SH2addi(sh, 5, (7<<4)+2);
}
#define init_7_5_7_3
void FASTCALL call_7_5_7_3(SH2_struct* sh)
{
  SH2addi(sh, 5, (7<<4)+3);
}
#define init_7_5_7_4
void FASTCALL call_7_5_7_4(SH2_struct* sh)
{
  SH2addi(sh, 5, (7<<4)+4);
}
#define init_7_5_7_5
void FASTCALL call_7_5_7_5(SH2_struct* sh)
{
  SH2addi(sh, 5, (7<<4)+5);
}
#define init_7_5_7_6
void FASTCALL call_7_5_7_6(SH2_struct* sh)
{
  SH2addi(sh, 5, (7<<4)+6);
}
#define init_7_5_7_7
void FASTCALL call_7_5_7_7(SH2_struct* sh)
{
  SH2addi(sh, 5, (7<<4)+7);
}
#define init_7_5_7_8
void FASTCALL call_7_5_7_8(SH2_struct* sh)
{
  SH2addi(sh, 5, (7<<4)+8);
}
#define init_7_5_7_9
void FASTCALL call_7_5_7_9(SH2_struct* sh)
{
  SH2addi(sh, 5, (7<<4)+9);
}
#define init_7_5_7_10
void FASTCALL call_7_5_7_10(SH2_struct* sh)
{
  SH2addi(sh, 5, (7<<4)+10);
}
#define init_7_5_7_11
void FASTCALL call_7_5_7_11(SH2_struct* sh)
{
  SH2addi(sh, 5, (7<<4)+11);
}
#define init_7_5_7_12
void FASTCALL call_7_5_7_12(SH2_struct* sh)
{
  SH2addi(sh, 5, (7<<4)+12);
}
#define init_7_5_7_13
void FASTCALL call_7_5_7_13(SH2_struct* sh)
{
  SH2addi(sh, 5, (7<<4)+13);
}
#define init_7_5_7_14
void FASTCALL call_7_5_7_14(SH2_struct* sh)
{
  SH2addi(sh, 5, (7<<4)+14);
}
#define init_7_5_7_15
void FASTCALL call_7_5_7_15(SH2_struct* sh)
{
  SH2addi(sh, 5, (7<<4)+15);
}
#define init_7_5_8_0
void FASTCALL call_7_5_8_0(SH2_struct* sh)
{
  SH2addi(sh, 5, (8<<4)+0);
}
#define init_7_5_8_1
void FASTCALL call_7_5_8_1(SH2_struct* sh)
{
  SH2addi(sh, 5, (8<<4)+1);
}
#define init_7_5_8_2
void FASTCALL call_7_5_8_2(SH2_struct* sh)
{
  SH2addi(sh, 5, (8<<4)+2);
}
#define init_7_5_8_3
void FASTCALL call_7_5_8_3(SH2_struct* sh)
{
  SH2addi(sh, 5, (8<<4)+3);
}
#define init_7_5_8_4
void FASTCALL call_7_5_8_4(SH2_struct* sh)
{
  SH2addi(sh, 5, (8<<4)+4);
}
#define init_7_5_8_5
void FASTCALL call_7_5_8_5(SH2_struct* sh)
{
  SH2addi(sh, 5, (8<<4)+5);
}
#define init_7_5_8_6
void FASTCALL call_7_5_8_6(SH2_struct* sh)
{
  SH2addi(sh, 5, (8<<4)+6);
}
#define init_7_5_8_7
void FASTCALL call_7_5_8_7(SH2_struct* sh)
{
  SH2addi(sh, 5, (8<<4)+7);
}
#define init_7_5_8_8
void FASTCALL call_7_5_8_8(SH2_struct* sh)
{
  SH2addi(sh, 5, (8<<4)+8);
}
#define init_7_5_8_9
void FASTCALL call_7_5_8_9(SH2_struct* sh)
{
  SH2addi(sh, 5, (8<<4)+9);
}
#define init_7_5_8_10
void FASTCALL call_7_5_8_10(SH2_struct* sh)
{
  SH2addi(sh, 5, (8<<4)+10);
}
#define init_7_5_8_11
void FASTCALL call_7_5_8_11(SH2_struct* sh)
{
  SH2addi(sh, 5, (8<<4)+11);
}
#define init_7_5_8_12
void FASTCALL call_7_5_8_12(SH2_struct* sh)
{
  SH2addi(sh, 5, (8<<4)+12);
}
#define init_7_5_8_13
void FASTCALL call_7_5_8_13(SH2_struct* sh)
{
  SH2addi(sh, 5, (8<<4)+13);
}
#define init_7_5_8_14
void FASTCALL call_7_5_8_14(SH2_struct* sh)
{
  SH2addi(sh, 5, (8<<4)+14);
}
#define init_7_5_8_15
void FASTCALL call_7_5_8_15(SH2_struct* sh)
{
  SH2addi(sh, 5, (8<<4)+15);
}
#define init_7_5_9_0
void FASTCALL call_7_5_9_0(SH2_struct* sh)
{
  SH2addi(sh, 5, (9<<4)+0);
}
#define init_7_5_9_1
void FASTCALL call_7_5_9_1(SH2_struct* sh)
{
  SH2addi(sh, 5, (9<<4)+1);
}
#define init_7_5_9_2
void FASTCALL call_7_5_9_2(SH2_struct* sh)
{
  SH2addi(sh, 5, (9<<4)+2);
}
#define init_7_5_9_3
void FASTCALL call_7_5_9_3(SH2_struct* sh)
{
  SH2addi(sh, 5, (9<<4)+3);
}
#define init_7_5_9_4
void FASTCALL call_7_5_9_4(SH2_struct* sh)
{
  SH2addi(sh, 5, (9<<4)+4);
}
#define init_7_5_9_5
void FASTCALL call_7_5_9_5(SH2_struct* sh)
{
  SH2addi(sh, 5, (9<<4)+5);
}
#define init_7_5_9_6
void FASTCALL call_7_5_9_6(SH2_struct* sh)
{
  SH2addi(sh, 5, (9<<4)+6);
}
#define init_7_5_9_7
void FASTCALL call_7_5_9_7(SH2_struct* sh)
{
  SH2addi(sh, 5, (9<<4)+7);
}
#define init_7_5_9_8
void FASTCALL call_7_5_9_8(SH2_struct* sh)
{
  SH2addi(sh, 5, (9<<4)+8);
}
#define init_7_5_9_9
void FASTCALL call_7_5_9_9(SH2_struct* sh)
{
  SH2addi(sh, 5, (9<<4)+9);
}
#define init_7_5_9_10
void FASTCALL call_7_5_9_10(SH2_struct* sh)
{
  SH2addi(sh, 5, (9<<4)+10);
}
#define init_7_5_9_11
void FASTCALL call_7_5_9_11(SH2_struct* sh)
{
  SH2addi(sh, 5, (9<<4)+11);
}
#define init_7_5_9_12
void FASTCALL call_7_5_9_12(SH2_struct* sh)
{
  SH2addi(sh, 5, (9<<4)+12);
}
#define init_7_5_9_13
void FASTCALL call_7_5_9_13(SH2_struct* sh)
{
  SH2addi(sh, 5, (9<<4)+13);
}
#define init_7_5_9_14
void FASTCALL call_7_5_9_14(SH2_struct* sh)
{
  SH2addi(sh, 5, (9<<4)+14);
}
#define init_7_5_9_15
void FASTCALL call_7_5_9_15(SH2_struct* sh)
{
  SH2addi(sh, 5, (9<<4)+15);
}
#define init_7_5_10_0
void FASTCALL call_7_5_10_0(SH2_struct* sh)
{
  SH2addi(sh, 5, (10<<4)+0);
}
#define init_7_5_10_1
void FASTCALL call_7_5_10_1(SH2_struct* sh)
{
  SH2addi(sh, 5, (10<<4)+1);
}
#define init_7_5_10_2
void FASTCALL call_7_5_10_2(SH2_struct* sh)
{
  SH2addi(sh, 5, (10<<4)+2);
}
#define init_7_5_10_3
void FASTCALL call_7_5_10_3(SH2_struct* sh)
{
  SH2addi(sh, 5, (10<<4)+3);
}
#define init_7_5_10_4
void FASTCALL call_7_5_10_4(SH2_struct* sh)
{
  SH2addi(sh, 5, (10<<4)+4);
}
#define init_7_5_10_5
void FASTCALL call_7_5_10_5(SH2_struct* sh)
{
  SH2addi(sh, 5, (10<<4)+5);
}
#define init_7_5_10_6
void FASTCALL call_7_5_10_6(SH2_struct* sh)
{
  SH2addi(sh, 5, (10<<4)+6);
}
#define init_7_5_10_7
void FASTCALL call_7_5_10_7(SH2_struct* sh)
{
  SH2addi(sh, 5, (10<<4)+7);
}
#define init_7_5_10_8
void FASTCALL call_7_5_10_8(SH2_struct* sh)
{
  SH2addi(sh, 5, (10<<4)+8);
}
#define init_7_5_10_9
void FASTCALL call_7_5_10_9(SH2_struct* sh)
{
  SH2addi(sh, 5, (10<<4)+9);
}
#define init_7_5_10_10
void FASTCALL call_7_5_10_10(SH2_struct* sh)
{
  SH2addi(sh, 5, (10<<4)+10);
}
#define init_7_5_10_11
void FASTCALL call_7_5_10_11(SH2_struct* sh)
{
  SH2addi(sh, 5, (10<<4)+11);
}
#define init_7_5_10_12
void FASTCALL call_7_5_10_12(SH2_struct* sh)
{
  SH2addi(sh, 5, (10<<4)+12);
}
#define init_7_5_10_13
void FASTCALL call_7_5_10_13(SH2_struct* sh)
{
  SH2addi(sh, 5, (10<<4)+13);
}
#define init_7_5_10_14
void FASTCALL call_7_5_10_14(SH2_struct* sh)
{
  SH2addi(sh, 5, (10<<4)+14);
}
#define init_7_5_10_15
void FASTCALL call_7_5_10_15(SH2_struct* sh)
{
  SH2addi(sh, 5, (10<<4)+15);
}
#define init_7_5_11_0
void FASTCALL call_7_5_11_0(SH2_struct* sh)
{
  SH2addi(sh, 5, (11<<4)+0);
}
#define init_7_5_11_1
void FASTCALL call_7_5_11_1(SH2_struct* sh)
{
  SH2addi(sh, 5, (11<<4)+1);
}
#define init_7_5_11_2
void FASTCALL call_7_5_11_2(SH2_struct* sh)
{
  SH2addi(sh, 5, (11<<4)+2);
}
#define init_7_5_11_3
void FASTCALL call_7_5_11_3(SH2_struct* sh)
{
  SH2addi(sh, 5, (11<<4)+3);
}
#define init_7_5_11_4
void FASTCALL call_7_5_11_4(SH2_struct* sh)
{
  SH2addi(sh, 5, (11<<4)+4);
}
#define init_7_5_11_5
void FASTCALL call_7_5_11_5(SH2_struct* sh)
{
  SH2addi(sh, 5, (11<<4)+5);
}
#define init_7_5_11_6
void FASTCALL call_7_5_11_6(SH2_struct* sh)
{
  SH2addi(sh, 5, (11<<4)+6);
}
#define init_7_5_11_7
void FASTCALL call_7_5_11_7(SH2_struct* sh)
{
  SH2addi(sh, 5, (11<<4)+7);
}
#define init_7_5_11_8
void FASTCALL call_7_5_11_8(SH2_struct* sh)
{
  SH2addi(sh, 5, (11<<4)+8);
}
#define init_7_5_11_9
void FASTCALL call_7_5_11_9(SH2_struct* sh)
{
  SH2addi(sh, 5, (11<<4)+9);
}
#define init_7_5_11_10
void FASTCALL call_7_5_11_10(SH2_struct* sh)
{
  SH2addi(sh, 5, (11<<4)+10);
}
#define init_7_5_11_11
void FASTCALL call_7_5_11_11(SH2_struct* sh)
{
  SH2addi(sh, 5, (11<<4)+11);
}
#define init_7_5_11_12
void FASTCALL call_7_5_11_12(SH2_struct* sh)
{
  SH2addi(sh, 5, (11<<4)+12);
}
#define init_7_5_11_13
void FASTCALL call_7_5_11_13(SH2_struct* sh)
{
  SH2addi(sh, 5, (11<<4)+13);
}
#define init_7_5_11_14
void FASTCALL call_7_5_11_14(SH2_struct* sh)
{
  SH2addi(sh, 5, (11<<4)+14);
}
#define init_7_5_11_15
void FASTCALL call_7_5_11_15(SH2_struct* sh)
{
  SH2addi(sh, 5, (11<<4)+15);
}
#define init_7_5_12_0
void FASTCALL call_7_5_12_0(SH2_struct* sh)
{
  SH2addi(sh, 5, (12<<4)+0);
}
#define init_7_5_12_1
void FASTCALL call_7_5_12_1(SH2_struct* sh)
{
  SH2addi(sh, 5, (12<<4)+1);
}
#define init_7_5_12_2
void FASTCALL call_7_5_12_2(SH2_struct* sh)
{
  SH2addi(sh, 5, (12<<4)+2);
}
#define init_7_5_12_3
void FASTCALL call_7_5_12_3(SH2_struct* sh)
{
  SH2addi(sh, 5, (12<<4)+3);
}
#define init_7_5_12_4
void FASTCALL call_7_5_12_4(SH2_struct* sh)
{
  SH2addi(sh, 5, (12<<4)+4);
}
#define init_7_5_12_5
void FASTCALL call_7_5_12_5(SH2_struct* sh)
{
  SH2addi(sh, 5, (12<<4)+5);
}
#define init_7_5_12_6
void FASTCALL call_7_5_12_6(SH2_struct* sh)
{
  SH2addi(sh, 5, (12<<4)+6);
}
#define init_7_5_12_7
void FASTCALL call_7_5_12_7(SH2_struct* sh)
{
  SH2addi(sh, 5, (12<<4)+7);
}
#define init_7_5_12_8
void FASTCALL call_7_5_12_8(SH2_struct* sh)
{
  SH2addi(sh, 5, (12<<4)+8);
}
#define init_7_5_12_9
void FASTCALL call_7_5_12_9(SH2_struct* sh)
{
  SH2addi(sh, 5, (12<<4)+9);
}
#define init_7_5_12_10
void FASTCALL call_7_5_12_10(SH2_struct* sh)
{
  SH2addi(sh, 5, (12<<4)+10);
}
#define init_7_5_12_11
void FASTCALL call_7_5_12_11(SH2_struct* sh)
{
  SH2addi(sh, 5, (12<<4)+11);
}
#define init_7_5_12_12
void FASTCALL call_7_5_12_12(SH2_struct* sh)
{
  SH2addi(sh, 5, (12<<4)+12);
}
#define init_7_5_12_13
void FASTCALL call_7_5_12_13(SH2_struct* sh)
{
  SH2addi(sh, 5, (12<<4)+13);
}
#define init_7_5_12_14
void FASTCALL call_7_5_12_14(SH2_struct* sh)
{
  SH2addi(sh, 5, (12<<4)+14);
}
#define init_7_5_12_15
void FASTCALL call_7_5_12_15(SH2_struct* sh)
{
  SH2addi(sh, 5, (12<<4)+15);
}
#define init_7_5_13_0
void FASTCALL call_7_5_13_0(SH2_struct* sh)
{
  SH2addi(sh, 5, (13<<4)+0);
}
#define init_7_5_13_1
void FASTCALL call_7_5_13_1(SH2_struct* sh)
{
  SH2addi(sh, 5, (13<<4)+1);
}
#define init_7_5_13_2
void FASTCALL call_7_5_13_2(SH2_struct* sh)
{
  SH2addi(sh, 5, (13<<4)+2);
}
#define init_7_5_13_3
void FASTCALL call_7_5_13_3(SH2_struct* sh)
{
  SH2addi(sh, 5, (13<<4)+3);
}
#define init_7_5_13_4
void FASTCALL call_7_5_13_4(SH2_struct* sh)
{
  SH2addi(sh, 5, (13<<4)+4);
}
#define init_7_5_13_5
void FASTCALL call_7_5_13_5(SH2_struct* sh)
{
  SH2addi(sh, 5, (13<<4)+5);
}
#define init_7_5_13_6
void FASTCALL call_7_5_13_6(SH2_struct* sh)
{
  SH2addi(sh, 5, (13<<4)+6);
}
#define init_7_5_13_7
void FASTCALL call_7_5_13_7(SH2_struct* sh)
{
  SH2addi(sh, 5, (13<<4)+7);
}
#define init_7_5_13_8
void FASTCALL call_7_5_13_8(SH2_struct* sh)
{
  SH2addi(sh, 5, (13<<4)+8);
}
#define init_7_5_13_9
void FASTCALL call_7_5_13_9(SH2_struct* sh)
{
  SH2addi(sh, 5, (13<<4)+9);
}
#define init_7_5_13_10
void FASTCALL call_7_5_13_10(SH2_struct* sh)
{
  SH2addi(sh, 5, (13<<4)+10);
}
#define init_7_5_13_11
void FASTCALL call_7_5_13_11(SH2_struct* sh)
{
  SH2addi(sh, 5, (13<<4)+11);
}
#define init_7_5_13_12
void FASTCALL call_7_5_13_12(SH2_struct* sh)
{
  SH2addi(sh, 5, (13<<4)+12);
}
#define init_7_5_13_13
void FASTCALL call_7_5_13_13(SH2_struct* sh)
{
  SH2addi(sh, 5, (13<<4)+13);
}
#define init_7_5_13_14
void FASTCALL call_7_5_13_14(SH2_struct* sh)
{
  SH2addi(sh, 5, (13<<4)+14);
}
#define init_7_5_13_15
void FASTCALL call_7_5_13_15(SH2_struct* sh)
{
  SH2addi(sh, 5, (13<<4)+15);
}
#define init_7_5_14_0
void FASTCALL call_7_5_14_0(SH2_struct* sh)
{
  SH2addi(sh, 5, (14<<4)+0);
}
#define init_7_5_14_1
void FASTCALL call_7_5_14_1(SH2_struct* sh)
{
  SH2addi(sh, 5, (14<<4)+1);
}
#define init_7_5_14_2
void FASTCALL call_7_5_14_2(SH2_struct* sh)
{
  SH2addi(sh, 5, (14<<4)+2);
}
#define init_7_5_14_3
void FASTCALL call_7_5_14_3(SH2_struct* sh)
{
  SH2addi(sh, 5, (14<<4)+3);
}
#define init_7_5_14_4
void FASTCALL call_7_5_14_4(SH2_struct* sh)
{
  SH2addi(sh, 5, (14<<4)+4);
}
#define init_7_5_14_5
void FASTCALL call_7_5_14_5(SH2_struct* sh)
{
  SH2addi(sh, 5, (14<<4)+5);
}
#define init_7_5_14_6
void FASTCALL call_7_5_14_6(SH2_struct* sh)
{
  SH2addi(sh, 5, (14<<4)+6);
}
#define init_7_5_14_7
void FASTCALL call_7_5_14_7(SH2_struct* sh)
{
  SH2addi(sh, 5, (14<<4)+7);
}
#define init_7_5_14_8
void FASTCALL call_7_5_14_8(SH2_struct* sh)
{
  SH2addi(sh, 5, (14<<4)+8);
}
#define init_7_5_14_9
void FASTCALL call_7_5_14_9(SH2_struct* sh)
{
  SH2addi(sh, 5, (14<<4)+9);
}
#define init_7_5_14_10
void FASTCALL call_7_5_14_10(SH2_struct* sh)
{
  SH2addi(sh, 5, (14<<4)+10);
}
#define init_7_5_14_11
void FASTCALL call_7_5_14_11(SH2_struct* sh)
{
  SH2addi(sh, 5, (14<<4)+11);
}
#define init_7_5_14_12
void FASTCALL call_7_5_14_12(SH2_struct* sh)
{
  SH2addi(sh, 5, (14<<4)+12);
}
#define init_7_5_14_13
void FASTCALL call_7_5_14_13(SH2_struct* sh)
{
  SH2addi(sh, 5, (14<<4)+13);
}
#define init_7_5_14_14
void FASTCALL call_7_5_14_14(SH2_struct* sh)
{
  SH2addi(sh, 5, (14<<4)+14);
}
#define init_7_5_14_15
void FASTCALL call_7_5_14_15(SH2_struct* sh)
{
  SH2addi(sh, 5, (14<<4)+15);
}
#define init_7_5_15_0
void FASTCALL call_7_5_15_0(SH2_struct* sh)
{
  SH2addi(sh, 5, (15<<4)+0);
}
#define init_7_5_15_1
void FASTCALL call_7_5_15_1(SH2_struct* sh)
{
  SH2addi(sh, 5, (15<<4)+1);
}
#define init_7_5_15_2
void FASTCALL call_7_5_15_2(SH2_struct* sh)
{
  SH2addi(sh, 5, (15<<4)+2);
}
#define init_7_5_15_3
void FASTCALL call_7_5_15_3(SH2_struct* sh)
{
  SH2addi(sh, 5, (15<<4)+3);
}
#define init_7_5_15_4
void FASTCALL call_7_5_15_4(SH2_struct* sh)
{
  SH2addi(sh, 5, (15<<4)+4);
}
#define init_7_5_15_5
void FASTCALL call_7_5_15_5(SH2_struct* sh)
{
  SH2addi(sh, 5, (15<<4)+5);
}
#define init_7_5_15_6
void FASTCALL call_7_5_15_6(SH2_struct* sh)
{
  SH2addi(sh, 5, (15<<4)+6);
}
#define init_7_5_15_7
void FASTCALL call_7_5_15_7(SH2_struct* sh)
{
  SH2addi(sh, 5, (15<<4)+7);
}
#define init_7_5_15_8
void FASTCALL call_7_5_15_8(SH2_struct* sh)
{
  SH2addi(sh, 5, (15<<4)+8);
}
#define init_7_5_15_9
void FASTCALL call_7_5_15_9(SH2_struct* sh)
{
  SH2addi(sh, 5, (15<<4)+9);
}
#define init_7_5_15_10
void FASTCALL call_7_5_15_10(SH2_struct* sh)
{
  SH2addi(sh, 5, (15<<4)+10);
}
#define init_7_5_15_11
void FASTCALL call_7_5_15_11(SH2_struct* sh)
{
  SH2addi(sh, 5, (15<<4)+11);
}
#define init_7_5_15_12
void FASTCALL call_7_5_15_12(SH2_struct* sh)
{
  SH2addi(sh, 5, (15<<4)+12);
}
#define init_7_5_15_13
void FASTCALL call_7_5_15_13(SH2_struct* sh)
{
  SH2addi(sh, 5, (15<<4)+13);
}
#define init_7_5_15_14
void FASTCALL call_7_5_15_14(SH2_struct* sh)
{
  SH2addi(sh, 5, (15<<4)+14);
}
#define init_7_5_15_15
void FASTCALL call_7_5_15_15(SH2_struct* sh)
{
  SH2addi(sh, 5, (15<<4)+15);
}
#define init_7_6_0_0
void FASTCALL call_7_6_0_0(SH2_struct* sh)
{
  SH2addi(sh, 6, (0<<4)+0);
}
#define init_7_6_0_1
void FASTCALL call_7_6_0_1(SH2_struct* sh)
{
  SH2addi(sh, 6, (0<<4)+1);
}
#define init_7_6_0_2
void FASTCALL call_7_6_0_2(SH2_struct* sh)
{
  SH2addi(sh, 6, (0<<4)+2);
}
#define init_7_6_0_3
void FASTCALL call_7_6_0_3(SH2_struct* sh)
{
  SH2addi(sh, 6, (0<<4)+3);
}
#define init_7_6_0_4
void FASTCALL call_7_6_0_4(SH2_struct* sh)
{
  SH2addi(sh, 6, (0<<4)+4);
}
#define init_7_6_0_5
void FASTCALL call_7_6_0_5(SH2_struct* sh)
{
  SH2addi(sh, 6, (0<<4)+5);
}
#define init_7_6_0_6
void FASTCALL call_7_6_0_6(SH2_struct* sh)
{
  SH2addi(sh, 6, (0<<4)+6);
}
#define init_7_6_0_7
void FASTCALL call_7_6_0_7(SH2_struct* sh)
{
  SH2addi(sh, 6, (0<<4)+7);
}
#define init_7_6_0_8
void FASTCALL call_7_6_0_8(SH2_struct* sh)
{
  SH2addi(sh, 6, (0<<4)+8);
}
#define init_7_6_0_9
void FASTCALL call_7_6_0_9(SH2_struct* sh)
{
  SH2addi(sh, 6, (0<<4)+9);
}
#define init_7_6_0_10
void FASTCALL call_7_6_0_10(SH2_struct* sh)
{
  SH2addi(sh, 6, (0<<4)+10);
}
#define init_7_6_0_11
void FASTCALL call_7_6_0_11(SH2_struct* sh)
{
  SH2addi(sh, 6, (0<<4)+11);
}
#define init_7_6_0_12
void FASTCALL call_7_6_0_12(SH2_struct* sh)
{
  SH2addi(sh, 6, (0<<4)+12);
}
#define init_7_6_0_13
void FASTCALL call_7_6_0_13(SH2_struct* sh)
{
  SH2addi(sh, 6, (0<<4)+13);
}
#define init_7_6_0_14
void FASTCALL call_7_6_0_14(SH2_struct* sh)
{
  SH2addi(sh, 6, (0<<4)+14);
}
#define init_7_6_0_15
void FASTCALL call_7_6_0_15(SH2_struct* sh)
{
  SH2addi(sh, 6, (0<<4)+15);
}
#define init_7_6_1_0
void FASTCALL call_7_6_1_0(SH2_struct* sh)
{
  SH2addi(sh, 6, (1<<4)+0);
}
#define init_7_6_1_1
void FASTCALL call_7_6_1_1(SH2_struct* sh)
{
  SH2addi(sh, 6, (1<<4)+1);
}
#define init_7_6_1_2
void FASTCALL call_7_6_1_2(SH2_struct* sh)
{
  SH2addi(sh, 6, (1<<4)+2);
}
#define init_7_6_1_3
void FASTCALL call_7_6_1_3(SH2_struct* sh)
{
  SH2addi(sh, 6, (1<<4)+3);
}
#define init_7_6_1_4
void FASTCALL call_7_6_1_4(SH2_struct* sh)
{
  SH2addi(sh, 6, (1<<4)+4);
}
#define init_7_6_1_5
void FASTCALL call_7_6_1_5(SH2_struct* sh)
{
  SH2addi(sh, 6, (1<<4)+5);
}
#define init_7_6_1_6
void FASTCALL call_7_6_1_6(SH2_struct* sh)
{
  SH2addi(sh, 6, (1<<4)+6);
}
#define init_7_6_1_7
void FASTCALL call_7_6_1_7(SH2_struct* sh)
{
  SH2addi(sh, 6, (1<<4)+7);
}
#define init_7_6_1_8
void FASTCALL call_7_6_1_8(SH2_struct* sh)
{
  SH2addi(sh, 6, (1<<4)+8);
}
#define init_7_6_1_9
void FASTCALL call_7_6_1_9(SH2_struct* sh)
{
  SH2addi(sh, 6, (1<<4)+9);
}
#define init_7_6_1_10
void FASTCALL call_7_6_1_10(SH2_struct* sh)
{
  SH2addi(sh, 6, (1<<4)+10);
}
#define init_7_6_1_11
void FASTCALL call_7_6_1_11(SH2_struct* sh)
{
  SH2addi(sh, 6, (1<<4)+11);
}
#define init_7_6_1_12
void FASTCALL call_7_6_1_12(SH2_struct* sh)
{
  SH2addi(sh, 6, (1<<4)+12);
}
#define init_7_6_1_13
void FASTCALL call_7_6_1_13(SH2_struct* sh)
{
  SH2addi(sh, 6, (1<<4)+13);
}
#define init_7_6_1_14
void FASTCALL call_7_6_1_14(SH2_struct* sh)
{
  SH2addi(sh, 6, (1<<4)+14);
}
#define init_7_6_1_15
void FASTCALL call_7_6_1_15(SH2_struct* sh)
{
  SH2addi(sh, 6, (1<<4)+15);
}
#define init_7_6_2_0
void FASTCALL call_7_6_2_0(SH2_struct* sh)
{
  SH2addi(sh, 6, (2<<4)+0);
}
#define init_7_6_2_1
void FASTCALL call_7_6_2_1(SH2_struct* sh)
{
  SH2addi(sh, 6, (2<<4)+1);
}
#define init_7_6_2_2
void FASTCALL call_7_6_2_2(SH2_struct* sh)
{
  SH2addi(sh, 6, (2<<4)+2);
}
#define init_7_6_2_3
void FASTCALL call_7_6_2_3(SH2_struct* sh)
{
  SH2addi(sh, 6, (2<<4)+3);
}
#define init_7_6_2_4
void FASTCALL call_7_6_2_4(SH2_struct* sh)
{
  SH2addi(sh, 6, (2<<4)+4);
}
#define init_7_6_2_5
void FASTCALL call_7_6_2_5(SH2_struct* sh)
{
  SH2addi(sh, 6, (2<<4)+5);
}
#define init_7_6_2_6
void FASTCALL call_7_6_2_6(SH2_struct* sh)
{
  SH2addi(sh, 6, (2<<4)+6);
}
#define init_7_6_2_7
void FASTCALL call_7_6_2_7(SH2_struct* sh)
{
  SH2addi(sh, 6, (2<<4)+7);
}
#define init_7_6_2_8
void FASTCALL call_7_6_2_8(SH2_struct* sh)
{
  SH2addi(sh, 6, (2<<4)+8);
}
#define init_7_6_2_9
void FASTCALL call_7_6_2_9(SH2_struct* sh)
{
  SH2addi(sh, 6, (2<<4)+9);
}
#define init_7_6_2_10
void FASTCALL call_7_6_2_10(SH2_struct* sh)
{
  SH2addi(sh, 6, (2<<4)+10);
}
#define init_7_6_2_11
void FASTCALL call_7_6_2_11(SH2_struct* sh)
{
  SH2addi(sh, 6, (2<<4)+11);
}
#define init_7_6_2_12
void FASTCALL call_7_6_2_12(SH2_struct* sh)
{
  SH2addi(sh, 6, (2<<4)+12);
}
#define init_7_6_2_13
void FASTCALL call_7_6_2_13(SH2_struct* sh)
{
  SH2addi(sh, 6, (2<<4)+13);
}
#define init_7_6_2_14
void FASTCALL call_7_6_2_14(SH2_struct* sh)
{
  SH2addi(sh, 6, (2<<4)+14);
}
#define init_7_6_2_15
void FASTCALL call_7_6_2_15(SH2_struct* sh)
{
  SH2addi(sh, 6, (2<<4)+15);
}
#define init_7_6_3_0
void FASTCALL call_7_6_3_0(SH2_struct* sh)
{
  SH2addi(sh, 6, (3<<4)+0);
}
#define init_7_6_3_1
void FASTCALL call_7_6_3_1(SH2_struct* sh)
{
  SH2addi(sh, 6, (3<<4)+1);
}
#define init_7_6_3_2
void FASTCALL call_7_6_3_2(SH2_struct* sh)
{
  SH2addi(sh, 6, (3<<4)+2);
}
#define init_7_6_3_3
void FASTCALL call_7_6_3_3(SH2_struct* sh)
{
  SH2addi(sh, 6, (3<<4)+3);
}
#define init_7_6_3_4
void FASTCALL call_7_6_3_4(SH2_struct* sh)
{
  SH2addi(sh, 6, (3<<4)+4);
}
#define init_7_6_3_5
void FASTCALL call_7_6_3_5(SH2_struct* sh)
{
  SH2addi(sh, 6, (3<<4)+5);
}
#define init_7_6_3_6
void FASTCALL call_7_6_3_6(SH2_struct* sh)
{
  SH2addi(sh, 6, (3<<4)+6);
}
#define init_7_6_3_7
void FASTCALL call_7_6_3_7(SH2_struct* sh)
{
  SH2addi(sh, 6, (3<<4)+7);
}
#define init_7_6_3_8
void FASTCALL call_7_6_3_8(SH2_struct* sh)
{
  SH2addi(sh, 6, (3<<4)+8);
}
#define init_7_6_3_9
void FASTCALL call_7_6_3_9(SH2_struct* sh)
{
  SH2addi(sh, 6, (3<<4)+9);
}
#define init_7_6_3_10
void FASTCALL call_7_6_3_10(SH2_struct* sh)
{
  SH2addi(sh, 6, (3<<4)+10);
}
#define init_7_6_3_11
void FASTCALL call_7_6_3_11(SH2_struct* sh)
{
  SH2addi(sh, 6, (3<<4)+11);
}
#define init_7_6_3_12
void FASTCALL call_7_6_3_12(SH2_struct* sh)
{
  SH2addi(sh, 6, (3<<4)+12);
}
#define init_7_6_3_13
void FASTCALL call_7_6_3_13(SH2_struct* sh)
{
  SH2addi(sh, 6, (3<<4)+13);
}
#define init_7_6_3_14
void FASTCALL call_7_6_3_14(SH2_struct* sh)
{
  SH2addi(sh, 6, (3<<4)+14);
}
#define init_7_6_3_15
void FASTCALL call_7_6_3_15(SH2_struct* sh)
{
  SH2addi(sh, 6, (3<<4)+15);
}
#define init_7_6_4_0
void FASTCALL call_7_6_4_0(SH2_struct* sh)
{
  SH2addi(sh, 6, (4<<4)+0);
}
#define init_7_6_4_1
void FASTCALL call_7_6_4_1(SH2_struct* sh)
{
  SH2addi(sh, 6, (4<<4)+1);
}
#define init_7_6_4_2
void FASTCALL call_7_6_4_2(SH2_struct* sh)
{
  SH2addi(sh, 6, (4<<4)+2);
}
#define init_7_6_4_3
void FASTCALL call_7_6_4_3(SH2_struct* sh)
{
  SH2addi(sh, 6, (4<<4)+3);
}
#define init_7_6_4_4
void FASTCALL call_7_6_4_4(SH2_struct* sh)
{
  SH2addi(sh, 6, (4<<4)+4);
}
#define init_7_6_4_5
void FASTCALL call_7_6_4_5(SH2_struct* sh)
{
  SH2addi(sh, 6, (4<<4)+5);
}
#define init_7_6_4_6
void FASTCALL call_7_6_4_6(SH2_struct* sh)
{
  SH2addi(sh, 6, (4<<4)+6);
}
#define init_7_6_4_7
void FASTCALL call_7_6_4_7(SH2_struct* sh)
{
  SH2addi(sh, 6, (4<<4)+7);
}
#define init_7_6_4_8
void FASTCALL call_7_6_4_8(SH2_struct* sh)
{
  SH2addi(sh, 6, (4<<4)+8);
}
#define init_7_6_4_9
void FASTCALL call_7_6_4_9(SH2_struct* sh)
{
  SH2addi(sh, 6, (4<<4)+9);
}
#define init_7_6_4_10
void FASTCALL call_7_6_4_10(SH2_struct* sh)
{
  SH2addi(sh, 6, (4<<4)+10);
}
#define init_7_6_4_11
void FASTCALL call_7_6_4_11(SH2_struct* sh)
{
  SH2addi(sh, 6, (4<<4)+11);
}
#define init_7_6_4_12
void FASTCALL call_7_6_4_12(SH2_struct* sh)
{
  SH2addi(sh, 6, (4<<4)+12);
}
#define init_7_6_4_13
void FASTCALL call_7_6_4_13(SH2_struct* sh)
{
  SH2addi(sh, 6, (4<<4)+13);
}
#define init_7_6_4_14
void FASTCALL call_7_6_4_14(SH2_struct* sh)
{
  SH2addi(sh, 6, (4<<4)+14);
}
#define init_7_6_4_15
void FASTCALL call_7_6_4_15(SH2_struct* sh)
{
  SH2addi(sh, 6, (4<<4)+15);
}
#define init_7_6_5_0
void FASTCALL call_7_6_5_0(SH2_struct* sh)
{
  SH2addi(sh, 6, (5<<4)+0);
}
#define init_7_6_5_1
void FASTCALL call_7_6_5_1(SH2_struct* sh)
{
  SH2addi(sh, 6, (5<<4)+1);
}
#define init_7_6_5_2
void FASTCALL call_7_6_5_2(SH2_struct* sh)
{
  SH2addi(sh, 6, (5<<4)+2);
}
#define init_7_6_5_3
void FASTCALL call_7_6_5_3(SH2_struct* sh)
{
  SH2addi(sh, 6, (5<<4)+3);
}
#define init_7_6_5_4
void FASTCALL call_7_6_5_4(SH2_struct* sh)
{
  SH2addi(sh, 6, (5<<4)+4);
}
#define init_7_6_5_5
void FASTCALL call_7_6_5_5(SH2_struct* sh)
{
  SH2addi(sh, 6, (5<<4)+5);
}
#define init_7_6_5_6
void FASTCALL call_7_6_5_6(SH2_struct* sh)
{
  SH2addi(sh, 6, (5<<4)+6);
}
#define init_7_6_5_7
void FASTCALL call_7_6_5_7(SH2_struct* sh)
{
  SH2addi(sh, 6, (5<<4)+7);
}
#define init_7_6_5_8
void FASTCALL call_7_6_5_8(SH2_struct* sh)
{
  SH2addi(sh, 6, (5<<4)+8);
}
#define init_7_6_5_9
void FASTCALL call_7_6_5_9(SH2_struct* sh)
{
  SH2addi(sh, 6, (5<<4)+9);
}
#define init_7_6_5_10
void FASTCALL call_7_6_5_10(SH2_struct* sh)
{
  SH2addi(sh, 6, (5<<4)+10);
}
#define init_7_6_5_11
void FASTCALL call_7_6_5_11(SH2_struct* sh)
{
  SH2addi(sh, 6, (5<<4)+11);
}
#define init_7_6_5_12
void FASTCALL call_7_6_5_12(SH2_struct* sh)
{
  SH2addi(sh, 6, (5<<4)+12);
}
#define init_7_6_5_13
void FASTCALL call_7_6_5_13(SH2_struct* sh)
{
  SH2addi(sh, 6, (5<<4)+13);
}
#define init_7_6_5_14
void FASTCALL call_7_6_5_14(SH2_struct* sh)
{
  SH2addi(sh, 6, (5<<4)+14);
}
#define init_7_6_5_15
void FASTCALL call_7_6_5_15(SH2_struct* sh)
{
  SH2addi(sh, 6, (5<<4)+15);
}
#define init_7_6_6_0
void FASTCALL call_7_6_6_0(SH2_struct* sh)
{
  SH2addi(sh, 6, (6<<4)+0);
}
#define init_7_6_6_1
void FASTCALL call_7_6_6_1(SH2_struct* sh)
{
  SH2addi(sh, 6, (6<<4)+1);
}
#define init_7_6_6_2
void FASTCALL call_7_6_6_2(SH2_struct* sh)
{
  SH2addi(sh, 6, (6<<4)+2);
}
#define init_7_6_6_3
void FASTCALL call_7_6_6_3(SH2_struct* sh)
{
  SH2addi(sh, 6, (6<<4)+3);
}
#define init_7_6_6_4
void FASTCALL call_7_6_6_4(SH2_struct* sh)
{
  SH2addi(sh, 6, (6<<4)+4);
}
#define init_7_6_6_5
void FASTCALL call_7_6_6_5(SH2_struct* sh)
{
  SH2addi(sh, 6, (6<<4)+5);
}
#define init_7_6_6_6
void FASTCALL call_7_6_6_6(SH2_struct* sh)
{
  SH2addi(sh, 6, (6<<4)+6);
}
#define init_7_6_6_7
void FASTCALL call_7_6_6_7(SH2_struct* sh)
{
  SH2addi(sh, 6, (6<<4)+7);
}
#define init_7_6_6_8
void FASTCALL call_7_6_6_8(SH2_struct* sh)
{
  SH2addi(sh, 6, (6<<4)+8);
}
#define init_7_6_6_9
void FASTCALL call_7_6_6_9(SH2_struct* sh)
{
  SH2addi(sh, 6, (6<<4)+9);
}
#define init_7_6_6_10
void FASTCALL call_7_6_6_10(SH2_struct* sh)
{
  SH2addi(sh, 6, (6<<4)+10);
}
#define init_7_6_6_11
void FASTCALL call_7_6_6_11(SH2_struct* sh)
{
  SH2addi(sh, 6, (6<<4)+11);
}
#define init_7_6_6_12
void FASTCALL call_7_6_6_12(SH2_struct* sh)
{
  SH2addi(sh, 6, (6<<4)+12);
}
#define init_7_6_6_13
void FASTCALL call_7_6_6_13(SH2_struct* sh)
{
  SH2addi(sh, 6, (6<<4)+13);
}
#define init_7_6_6_14
void FASTCALL call_7_6_6_14(SH2_struct* sh)
{
  SH2addi(sh, 6, (6<<4)+14);
}
#define init_7_6_6_15
void FASTCALL call_7_6_6_15(SH2_struct* sh)
{
  SH2addi(sh, 6, (6<<4)+15);
}
#define init_7_6_7_0
void FASTCALL call_7_6_7_0(SH2_struct* sh)
{
  SH2addi(sh, 6, (7<<4)+0);
}
#define init_7_6_7_1
void FASTCALL call_7_6_7_1(SH2_struct* sh)
{
  SH2addi(sh, 6, (7<<4)+1);
}
#define init_7_6_7_2
void FASTCALL call_7_6_7_2(SH2_struct* sh)
{
  SH2addi(sh, 6, (7<<4)+2);
}
#define init_7_6_7_3
void FASTCALL call_7_6_7_3(SH2_struct* sh)
{
  SH2addi(sh, 6, (7<<4)+3);
}
#define init_7_6_7_4
void FASTCALL call_7_6_7_4(SH2_struct* sh)
{
  SH2addi(sh, 6, (7<<4)+4);
}
#define init_7_6_7_5
void FASTCALL call_7_6_7_5(SH2_struct* sh)
{
  SH2addi(sh, 6, (7<<4)+5);
}
#define init_7_6_7_6
void FASTCALL call_7_6_7_6(SH2_struct* sh)
{
  SH2addi(sh, 6, (7<<4)+6);
}
#define init_7_6_7_7
void FASTCALL call_7_6_7_7(SH2_struct* sh)
{
  SH2addi(sh, 6, (7<<4)+7);
}
#define init_7_6_7_8
void FASTCALL call_7_6_7_8(SH2_struct* sh)
{
  SH2addi(sh, 6, (7<<4)+8);
}
#define init_7_6_7_9
void FASTCALL call_7_6_7_9(SH2_struct* sh)
{
  SH2addi(sh, 6, (7<<4)+9);
}
#define init_7_6_7_10
void FASTCALL call_7_6_7_10(SH2_struct* sh)
{
  SH2addi(sh, 6, (7<<4)+10);
}
#define init_7_6_7_11
void FASTCALL call_7_6_7_11(SH2_struct* sh)
{
  SH2addi(sh, 6, (7<<4)+11);
}
#define init_7_6_7_12
void FASTCALL call_7_6_7_12(SH2_struct* sh)
{
  SH2addi(sh, 6, (7<<4)+12);
}
#define init_7_6_7_13
void FASTCALL call_7_6_7_13(SH2_struct* sh)
{
  SH2addi(sh, 6, (7<<4)+13);
}
#define init_7_6_7_14
void FASTCALL call_7_6_7_14(SH2_struct* sh)
{
  SH2addi(sh, 6, (7<<4)+14);
}
#define init_7_6_7_15
void FASTCALL call_7_6_7_15(SH2_struct* sh)
{
  SH2addi(sh, 6, (7<<4)+15);
}
#define init_7_6_8_0
void FASTCALL call_7_6_8_0(SH2_struct* sh)
{
  SH2addi(sh, 6, (8<<4)+0);
}
#define init_7_6_8_1
void FASTCALL call_7_6_8_1(SH2_struct* sh)
{
  SH2addi(sh, 6, (8<<4)+1);
}
#define init_7_6_8_2
void FASTCALL call_7_6_8_2(SH2_struct* sh)
{
  SH2addi(sh, 6, (8<<4)+2);
}
#define init_7_6_8_3
void FASTCALL call_7_6_8_3(SH2_struct* sh)
{
  SH2addi(sh, 6, (8<<4)+3);
}
#define init_7_6_8_4
void FASTCALL call_7_6_8_4(SH2_struct* sh)
{
  SH2addi(sh, 6, (8<<4)+4);
}
#define init_7_6_8_5
void FASTCALL call_7_6_8_5(SH2_struct* sh)
{
  SH2addi(sh, 6, (8<<4)+5);
}
#define init_7_6_8_6
void FASTCALL call_7_6_8_6(SH2_struct* sh)
{
  SH2addi(sh, 6, (8<<4)+6);
}
#define init_7_6_8_7
void FASTCALL call_7_6_8_7(SH2_struct* sh)
{
  SH2addi(sh, 6, (8<<4)+7);
}
#define init_7_6_8_8
void FASTCALL call_7_6_8_8(SH2_struct* sh)
{
  SH2addi(sh, 6, (8<<4)+8);
}
#define init_7_6_8_9
void FASTCALL call_7_6_8_9(SH2_struct* sh)
{
  SH2addi(sh, 6, (8<<4)+9);
}
#define init_7_6_8_10
void FASTCALL call_7_6_8_10(SH2_struct* sh)
{
  SH2addi(sh, 6, (8<<4)+10);
}
#define init_7_6_8_11
void FASTCALL call_7_6_8_11(SH2_struct* sh)
{
  SH2addi(sh, 6, (8<<4)+11);
}
#define init_7_6_8_12
void FASTCALL call_7_6_8_12(SH2_struct* sh)
{
  SH2addi(sh, 6, (8<<4)+12);
}
#define init_7_6_8_13
void FASTCALL call_7_6_8_13(SH2_struct* sh)
{
  SH2addi(sh, 6, (8<<4)+13);
}
#define init_7_6_8_14
void FASTCALL call_7_6_8_14(SH2_struct* sh)
{
  SH2addi(sh, 6, (8<<4)+14);
}
#define init_7_6_8_15
void FASTCALL call_7_6_8_15(SH2_struct* sh)
{
  SH2addi(sh, 6, (8<<4)+15);
}
#define init_7_6_9_0
void FASTCALL call_7_6_9_0(SH2_struct* sh)
{
  SH2addi(sh, 6, (9<<4)+0);
}
#define init_7_6_9_1
void FASTCALL call_7_6_9_1(SH2_struct* sh)
{
  SH2addi(sh, 6, (9<<4)+1);
}
#define init_7_6_9_2
void FASTCALL call_7_6_9_2(SH2_struct* sh)
{
  SH2addi(sh, 6, (9<<4)+2);
}
#define init_7_6_9_3
void FASTCALL call_7_6_9_3(SH2_struct* sh)
{
  SH2addi(sh, 6, (9<<4)+3);
}
#define init_7_6_9_4
void FASTCALL call_7_6_9_4(SH2_struct* sh)
{
  SH2addi(sh, 6, (9<<4)+4);
}
#define init_7_6_9_5
void FASTCALL call_7_6_9_5(SH2_struct* sh)
{
  SH2addi(sh, 6, (9<<4)+5);
}
#define init_7_6_9_6
void FASTCALL call_7_6_9_6(SH2_struct* sh)
{
  SH2addi(sh, 6, (9<<4)+6);
}
#define init_7_6_9_7
void FASTCALL call_7_6_9_7(SH2_struct* sh)
{
  SH2addi(sh, 6, (9<<4)+7);
}
#define init_7_6_9_8
void FASTCALL call_7_6_9_8(SH2_struct* sh)
{
  SH2addi(sh, 6, (9<<4)+8);
}
#define init_7_6_9_9
void FASTCALL call_7_6_9_9(SH2_struct* sh)
{
  SH2addi(sh, 6, (9<<4)+9);
}
#define init_7_6_9_10
void FASTCALL call_7_6_9_10(SH2_struct* sh)
{
  SH2addi(sh, 6, (9<<4)+10);
}
#define init_7_6_9_11
void FASTCALL call_7_6_9_11(SH2_struct* sh)
{
  SH2addi(sh, 6, (9<<4)+11);
}
#define init_7_6_9_12
void FASTCALL call_7_6_9_12(SH2_struct* sh)
{
  SH2addi(sh, 6, (9<<4)+12);
}
#define init_7_6_9_13
void FASTCALL call_7_6_9_13(SH2_struct* sh)
{
  SH2addi(sh, 6, (9<<4)+13);
}
#define init_7_6_9_14
void FASTCALL call_7_6_9_14(SH2_struct* sh)
{
  SH2addi(sh, 6, (9<<4)+14);
}
#define init_7_6_9_15
void FASTCALL call_7_6_9_15(SH2_struct* sh)
{
  SH2addi(sh, 6, (9<<4)+15);
}
#define init_7_6_10_0
void FASTCALL call_7_6_10_0(SH2_struct* sh)
{
  SH2addi(sh, 6, (10<<4)+0);
}
#define init_7_6_10_1
void FASTCALL call_7_6_10_1(SH2_struct* sh)
{
  SH2addi(sh, 6, (10<<4)+1);
}
#define init_7_6_10_2
void FASTCALL call_7_6_10_2(SH2_struct* sh)
{
  SH2addi(sh, 6, (10<<4)+2);
}
#define init_7_6_10_3
void FASTCALL call_7_6_10_3(SH2_struct* sh)
{
  SH2addi(sh, 6, (10<<4)+3);
}
#define init_7_6_10_4
void FASTCALL call_7_6_10_4(SH2_struct* sh)
{
  SH2addi(sh, 6, (10<<4)+4);
}
#define init_7_6_10_5
void FASTCALL call_7_6_10_5(SH2_struct* sh)
{
  SH2addi(sh, 6, (10<<4)+5);
}
#define init_7_6_10_6
void FASTCALL call_7_6_10_6(SH2_struct* sh)
{
  SH2addi(sh, 6, (10<<4)+6);
}
#define init_7_6_10_7
void FASTCALL call_7_6_10_7(SH2_struct* sh)
{
  SH2addi(sh, 6, (10<<4)+7);
}
#define init_7_6_10_8
void FASTCALL call_7_6_10_8(SH2_struct* sh)
{
  SH2addi(sh, 6, (10<<4)+8);
}
#define init_7_6_10_9
void FASTCALL call_7_6_10_9(SH2_struct* sh)
{
  SH2addi(sh, 6, (10<<4)+9);
}
#define init_7_6_10_10
void FASTCALL call_7_6_10_10(SH2_struct* sh)
{
  SH2addi(sh, 6, (10<<4)+10);
}
#define init_7_6_10_11
void FASTCALL call_7_6_10_11(SH2_struct* sh)
{
  SH2addi(sh, 6, (10<<4)+11);
}
#define init_7_6_10_12
void FASTCALL call_7_6_10_12(SH2_struct* sh)
{
  SH2addi(sh, 6, (10<<4)+12);
}
#define init_7_6_10_13
void FASTCALL call_7_6_10_13(SH2_struct* sh)
{
  SH2addi(sh, 6, (10<<4)+13);
}
#define init_7_6_10_14
void FASTCALL call_7_6_10_14(SH2_struct* sh)
{
  SH2addi(sh, 6, (10<<4)+14);
}
#define init_7_6_10_15
void FASTCALL call_7_6_10_15(SH2_struct* sh)
{
  SH2addi(sh, 6, (10<<4)+15);
}
#define init_7_6_11_0
void FASTCALL call_7_6_11_0(SH2_struct* sh)
{
  SH2addi(sh, 6, (11<<4)+0);
}
#define init_7_6_11_1
void FASTCALL call_7_6_11_1(SH2_struct* sh)
{
  SH2addi(sh, 6, (11<<4)+1);
}
#define init_7_6_11_2
void FASTCALL call_7_6_11_2(SH2_struct* sh)
{
  SH2addi(sh, 6, (11<<4)+2);
}
#define init_7_6_11_3
void FASTCALL call_7_6_11_3(SH2_struct* sh)
{
  SH2addi(sh, 6, (11<<4)+3);
}
#define init_7_6_11_4
void FASTCALL call_7_6_11_4(SH2_struct* sh)
{
  SH2addi(sh, 6, (11<<4)+4);
}
#define init_7_6_11_5
void FASTCALL call_7_6_11_5(SH2_struct* sh)
{
  SH2addi(sh, 6, (11<<4)+5);
}
#define init_7_6_11_6
void FASTCALL call_7_6_11_6(SH2_struct* sh)
{
  SH2addi(sh, 6, (11<<4)+6);
}
#define init_7_6_11_7
void FASTCALL call_7_6_11_7(SH2_struct* sh)
{
  SH2addi(sh, 6, (11<<4)+7);
}
#define init_7_6_11_8
void FASTCALL call_7_6_11_8(SH2_struct* sh)
{
  SH2addi(sh, 6, (11<<4)+8);
}
#define init_7_6_11_9
void FASTCALL call_7_6_11_9(SH2_struct* sh)
{
  SH2addi(sh, 6, (11<<4)+9);
}
#define init_7_6_11_10
void FASTCALL call_7_6_11_10(SH2_struct* sh)
{
  SH2addi(sh, 6, (11<<4)+10);
}
#define init_7_6_11_11
void FASTCALL call_7_6_11_11(SH2_struct* sh)
{
  SH2addi(sh, 6, (11<<4)+11);
}
#define init_7_6_11_12
void FASTCALL call_7_6_11_12(SH2_struct* sh)
{
  SH2addi(sh, 6, (11<<4)+12);
}
#define init_7_6_11_13
void FASTCALL call_7_6_11_13(SH2_struct* sh)
{
  SH2addi(sh, 6, (11<<4)+13);
}
#define init_7_6_11_14
void FASTCALL call_7_6_11_14(SH2_struct* sh)
{
  SH2addi(sh, 6, (11<<4)+14);
}
#define init_7_6_11_15
void FASTCALL call_7_6_11_15(SH2_struct* sh)
{
  SH2addi(sh, 6, (11<<4)+15);
}
#define init_7_6_12_0
void FASTCALL call_7_6_12_0(SH2_struct* sh)
{
  SH2addi(sh, 6, (12<<4)+0);
}
#define init_7_6_12_1
void FASTCALL call_7_6_12_1(SH2_struct* sh)
{
  SH2addi(sh, 6, (12<<4)+1);
}
#define init_7_6_12_2
void FASTCALL call_7_6_12_2(SH2_struct* sh)
{
  SH2addi(sh, 6, (12<<4)+2);
}
#define init_7_6_12_3
void FASTCALL call_7_6_12_3(SH2_struct* sh)
{
  SH2addi(sh, 6, (12<<4)+3);
}
#define init_7_6_12_4
void FASTCALL call_7_6_12_4(SH2_struct* sh)
{
  SH2addi(sh, 6, (12<<4)+4);
}
#define init_7_6_12_5
void FASTCALL call_7_6_12_5(SH2_struct* sh)
{
  SH2addi(sh, 6, (12<<4)+5);
}
#define init_7_6_12_6
void FASTCALL call_7_6_12_6(SH2_struct* sh)
{
  SH2addi(sh, 6, (12<<4)+6);
}
#define init_7_6_12_7
void FASTCALL call_7_6_12_7(SH2_struct* sh)
{
  SH2addi(sh, 6, (12<<4)+7);
}
#define init_7_6_12_8
void FASTCALL call_7_6_12_8(SH2_struct* sh)
{
  SH2addi(sh, 6, (12<<4)+8);
}
#define init_7_6_12_9
void FASTCALL call_7_6_12_9(SH2_struct* sh)
{
  SH2addi(sh, 6, (12<<4)+9);
}
#define init_7_6_12_10
void FASTCALL call_7_6_12_10(SH2_struct* sh)
{
  SH2addi(sh, 6, (12<<4)+10);
}
#define init_7_6_12_11
void FASTCALL call_7_6_12_11(SH2_struct* sh)
{
  SH2addi(sh, 6, (12<<4)+11);
}
#define init_7_6_12_12
void FASTCALL call_7_6_12_12(SH2_struct* sh)
{
  SH2addi(sh, 6, (12<<4)+12);
}
#define init_7_6_12_13
void FASTCALL call_7_6_12_13(SH2_struct* sh)
{
  SH2addi(sh, 6, (12<<4)+13);
}
#define init_7_6_12_14
void FASTCALL call_7_6_12_14(SH2_struct* sh)
{
  SH2addi(sh, 6, (12<<4)+14);
}
#define init_7_6_12_15
void FASTCALL call_7_6_12_15(SH2_struct* sh)
{
  SH2addi(sh, 6, (12<<4)+15);
}
#define init_7_6_13_0
void FASTCALL call_7_6_13_0(SH2_struct* sh)
{
  SH2addi(sh, 6, (13<<4)+0);
}
#define init_7_6_13_1
void FASTCALL call_7_6_13_1(SH2_struct* sh)
{
  SH2addi(sh, 6, (13<<4)+1);
}
#define init_7_6_13_2
void FASTCALL call_7_6_13_2(SH2_struct* sh)
{
  SH2addi(sh, 6, (13<<4)+2);
}
#define init_7_6_13_3
void FASTCALL call_7_6_13_3(SH2_struct* sh)
{
  SH2addi(sh, 6, (13<<4)+3);
}
#define init_7_6_13_4
void FASTCALL call_7_6_13_4(SH2_struct* sh)
{
  SH2addi(sh, 6, (13<<4)+4);
}
#define init_7_6_13_5
void FASTCALL call_7_6_13_5(SH2_struct* sh)
{
  SH2addi(sh, 6, (13<<4)+5);
}
#define init_7_6_13_6
void FASTCALL call_7_6_13_6(SH2_struct* sh)
{
  SH2addi(sh, 6, (13<<4)+6);
}
#define init_7_6_13_7
void FASTCALL call_7_6_13_7(SH2_struct* sh)
{
  SH2addi(sh, 6, (13<<4)+7);
}
#define init_7_6_13_8
void FASTCALL call_7_6_13_8(SH2_struct* sh)
{
  SH2addi(sh, 6, (13<<4)+8);
}
#define init_7_6_13_9
void FASTCALL call_7_6_13_9(SH2_struct* sh)
{
  SH2addi(sh, 6, (13<<4)+9);
}
#define init_7_6_13_10
void FASTCALL call_7_6_13_10(SH2_struct* sh)
{
  SH2addi(sh, 6, (13<<4)+10);
}
#define init_7_6_13_11
void FASTCALL call_7_6_13_11(SH2_struct* sh)
{
  SH2addi(sh, 6, (13<<4)+11);
}
#define init_7_6_13_12
void FASTCALL call_7_6_13_12(SH2_struct* sh)
{
  SH2addi(sh, 6, (13<<4)+12);
}
#define init_7_6_13_13
void FASTCALL call_7_6_13_13(SH2_struct* sh)
{
  SH2addi(sh, 6, (13<<4)+13);
}
#define init_7_6_13_14
void FASTCALL call_7_6_13_14(SH2_struct* sh)
{
  SH2addi(sh, 6, (13<<4)+14);
}
#define init_7_6_13_15
void FASTCALL call_7_6_13_15(SH2_struct* sh)
{
  SH2addi(sh, 6, (13<<4)+15);
}
#define init_7_6_14_0
void FASTCALL call_7_6_14_0(SH2_struct* sh)
{
  SH2addi(sh, 6, (14<<4)+0);
}
#define init_7_6_14_1
void FASTCALL call_7_6_14_1(SH2_struct* sh)
{
  SH2addi(sh, 6, (14<<4)+1);
}
#define init_7_6_14_2
void FASTCALL call_7_6_14_2(SH2_struct* sh)
{
  SH2addi(sh, 6, (14<<4)+2);
}
#define init_7_6_14_3
void FASTCALL call_7_6_14_3(SH2_struct* sh)
{
  SH2addi(sh, 6, (14<<4)+3);
}
#define init_7_6_14_4
void FASTCALL call_7_6_14_4(SH2_struct* sh)
{
  SH2addi(sh, 6, (14<<4)+4);
}
#define init_7_6_14_5
void FASTCALL call_7_6_14_5(SH2_struct* sh)
{
  SH2addi(sh, 6, (14<<4)+5);
}
#define init_7_6_14_6
void FASTCALL call_7_6_14_6(SH2_struct* sh)
{
  SH2addi(sh, 6, (14<<4)+6);
}
#define init_7_6_14_7
void FASTCALL call_7_6_14_7(SH2_struct* sh)
{
  SH2addi(sh, 6, (14<<4)+7);
}
#define init_7_6_14_8
void FASTCALL call_7_6_14_8(SH2_struct* sh)
{
  SH2addi(sh, 6, (14<<4)+8);
}
#define init_7_6_14_9
void FASTCALL call_7_6_14_9(SH2_struct* sh)
{
  SH2addi(sh, 6, (14<<4)+9);
}
#define init_7_6_14_10
void FASTCALL call_7_6_14_10(SH2_struct* sh)
{
  SH2addi(sh, 6, (14<<4)+10);
}
#define init_7_6_14_11
void FASTCALL call_7_6_14_11(SH2_struct* sh)
{
  SH2addi(sh, 6, (14<<4)+11);
}
#define init_7_6_14_12
void FASTCALL call_7_6_14_12(SH2_struct* sh)
{
  SH2addi(sh, 6, (14<<4)+12);
}
#define init_7_6_14_13
void FASTCALL call_7_6_14_13(SH2_struct* sh)
{
  SH2addi(sh, 6, (14<<4)+13);
}
#define init_7_6_14_14
void FASTCALL call_7_6_14_14(SH2_struct* sh)
{
  SH2addi(sh, 6, (14<<4)+14);
}
#define init_7_6_14_15
void FASTCALL call_7_6_14_15(SH2_struct* sh)
{
  SH2addi(sh, 6, (14<<4)+15);
}
#define init_7_6_15_0
void FASTCALL call_7_6_15_0(SH2_struct* sh)
{
  SH2addi(sh, 6, (15<<4)+0);
}
#define init_7_6_15_1
void FASTCALL call_7_6_15_1(SH2_struct* sh)
{
  SH2addi(sh, 6, (15<<4)+1);
}
#define init_7_6_15_2
void FASTCALL call_7_6_15_2(SH2_struct* sh)
{
  SH2addi(sh, 6, (15<<4)+2);
}
#define init_7_6_15_3
void FASTCALL call_7_6_15_3(SH2_struct* sh)
{
  SH2addi(sh, 6, (15<<4)+3);
}
#define init_7_6_15_4
void FASTCALL call_7_6_15_4(SH2_struct* sh)
{
  SH2addi(sh, 6, (15<<4)+4);
}
#define init_7_6_15_5
void FASTCALL call_7_6_15_5(SH2_struct* sh)
{
  SH2addi(sh, 6, (15<<4)+5);
}
#define init_7_6_15_6
void FASTCALL call_7_6_15_6(SH2_struct* sh)
{
  SH2addi(sh, 6, (15<<4)+6);
}
#define init_7_6_15_7
void FASTCALL call_7_6_15_7(SH2_struct* sh)
{
  SH2addi(sh, 6, (15<<4)+7);
}
#define init_7_6_15_8
void FASTCALL call_7_6_15_8(SH2_struct* sh)
{
  SH2addi(sh, 6, (15<<4)+8);
}
#define init_7_6_15_9
void FASTCALL call_7_6_15_9(SH2_struct* sh)
{
  SH2addi(sh, 6, (15<<4)+9);
}
#define init_7_6_15_10
void FASTCALL call_7_6_15_10(SH2_struct* sh)
{
  SH2addi(sh, 6, (15<<4)+10);
}
#define init_7_6_15_11
void FASTCALL call_7_6_15_11(SH2_struct* sh)
{
  SH2addi(sh, 6, (15<<4)+11);
}
#define init_7_6_15_12
void FASTCALL call_7_6_15_12(SH2_struct* sh)
{
  SH2addi(sh, 6, (15<<4)+12);
}
#define init_7_6_15_13
void FASTCALL call_7_6_15_13(SH2_struct* sh)
{
  SH2addi(sh, 6, (15<<4)+13);
}
#define init_7_6_15_14
void FASTCALL call_7_6_15_14(SH2_struct* sh)
{
  SH2addi(sh, 6, (15<<4)+14);
}
#define init_7_6_15_15
void FASTCALL call_7_6_15_15(SH2_struct* sh)
{
  SH2addi(sh, 6, (15<<4)+15);
}
#define init_7_7_0_0
void FASTCALL call_7_7_0_0(SH2_struct* sh)
{
  SH2addi(sh, 7, (0<<4)+0);
}
#define init_7_7_0_1
void FASTCALL call_7_7_0_1(SH2_struct* sh)
{
  SH2addi(sh, 7, (0<<4)+1);
}
#define init_7_7_0_2
void FASTCALL call_7_7_0_2(SH2_struct* sh)
{
  SH2addi(sh, 7, (0<<4)+2);
}
#define init_7_7_0_3
void FASTCALL call_7_7_0_3(SH2_struct* sh)
{
  SH2addi(sh, 7, (0<<4)+3);
}
#define init_7_7_0_4
void FASTCALL call_7_7_0_4(SH2_struct* sh)
{
  SH2addi(sh, 7, (0<<4)+4);
}
#define init_7_7_0_5
void FASTCALL call_7_7_0_5(SH2_struct* sh)
{
  SH2addi(sh, 7, (0<<4)+5);
}
#define init_7_7_0_6
void FASTCALL call_7_7_0_6(SH2_struct* sh)
{
  SH2addi(sh, 7, (0<<4)+6);
}
#define init_7_7_0_7
void FASTCALL call_7_7_0_7(SH2_struct* sh)
{
  SH2addi(sh, 7, (0<<4)+7);
}
#define init_7_7_0_8
void FASTCALL call_7_7_0_8(SH2_struct* sh)
{
  SH2addi(sh, 7, (0<<4)+8);
}
#define init_7_7_0_9
void FASTCALL call_7_7_0_9(SH2_struct* sh)
{
  SH2addi(sh, 7, (0<<4)+9);
}
#define init_7_7_0_10
void FASTCALL call_7_7_0_10(SH2_struct* sh)
{
  SH2addi(sh, 7, (0<<4)+10);
}
#define init_7_7_0_11
void FASTCALL call_7_7_0_11(SH2_struct* sh)
{
  SH2addi(sh, 7, (0<<4)+11);
}
#define init_7_7_0_12
void FASTCALL call_7_7_0_12(SH2_struct* sh)
{
  SH2addi(sh, 7, (0<<4)+12);
}
#define init_7_7_0_13
void FASTCALL call_7_7_0_13(SH2_struct* sh)
{
  SH2addi(sh, 7, (0<<4)+13);
}
#define init_7_7_0_14
void FASTCALL call_7_7_0_14(SH2_struct* sh)
{
  SH2addi(sh, 7, (0<<4)+14);
}
#define init_7_7_0_15
void FASTCALL call_7_7_0_15(SH2_struct* sh)
{
  SH2addi(sh, 7, (0<<4)+15);
}
#define init_7_7_1_0
void FASTCALL call_7_7_1_0(SH2_struct* sh)
{
  SH2addi(sh, 7, (1<<4)+0);
}
#define init_7_7_1_1
void FASTCALL call_7_7_1_1(SH2_struct* sh)
{
  SH2addi(sh, 7, (1<<4)+1);
}
#define init_7_7_1_2
void FASTCALL call_7_7_1_2(SH2_struct* sh)
{
  SH2addi(sh, 7, (1<<4)+2);
}
#define init_7_7_1_3
void FASTCALL call_7_7_1_3(SH2_struct* sh)
{
  SH2addi(sh, 7, (1<<4)+3);
}
#define init_7_7_1_4
void FASTCALL call_7_7_1_4(SH2_struct* sh)
{
  SH2addi(sh, 7, (1<<4)+4);
}
#define init_7_7_1_5
void FASTCALL call_7_7_1_5(SH2_struct* sh)
{
  SH2addi(sh, 7, (1<<4)+5);
}
#define init_7_7_1_6
void FASTCALL call_7_7_1_6(SH2_struct* sh)
{
  SH2addi(sh, 7, (1<<4)+6);
}
#define init_7_7_1_7
void FASTCALL call_7_7_1_7(SH2_struct* sh)
{
  SH2addi(sh, 7, (1<<4)+7);
}
#define init_7_7_1_8
void FASTCALL call_7_7_1_8(SH2_struct* sh)
{
  SH2addi(sh, 7, (1<<4)+8);
}
#define init_7_7_1_9
void FASTCALL call_7_7_1_9(SH2_struct* sh)
{
  SH2addi(sh, 7, (1<<4)+9);
}
#define init_7_7_1_10
void FASTCALL call_7_7_1_10(SH2_struct* sh)
{
  SH2addi(sh, 7, (1<<4)+10);
}
#define init_7_7_1_11
void FASTCALL call_7_7_1_11(SH2_struct* sh)
{
  SH2addi(sh, 7, (1<<4)+11);
}
#define init_7_7_1_12
void FASTCALL call_7_7_1_12(SH2_struct* sh)
{
  SH2addi(sh, 7, (1<<4)+12);
}
#define init_7_7_1_13
void FASTCALL call_7_7_1_13(SH2_struct* sh)
{
  SH2addi(sh, 7, (1<<4)+13);
}
#define init_7_7_1_14
void FASTCALL call_7_7_1_14(SH2_struct* sh)
{
  SH2addi(sh, 7, (1<<4)+14);
}
#define init_7_7_1_15
void FASTCALL call_7_7_1_15(SH2_struct* sh)
{
  SH2addi(sh, 7, (1<<4)+15);
}
#define init_7_7_2_0
void FASTCALL call_7_7_2_0(SH2_struct* sh)
{
  SH2addi(sh, 7, (2<<4)+0);
}
#define init_7_7_2_1
void FASTCALL call_7_7_2_1(SH2_struct* sh)
{
  SH2addi(sh, 7, (2<<4)+1);
}
#define init_7_7_2_2
void FASTCALL call_7_7_2_2(SH2_struct* sh)
{
  SH2addi(sh, 7, (2<<4)+2);
}
#define init_7_7_2_3
void FASTCALL call_7_7_2_3(SH2_struct* sh)
{
  SH2addi(sh, 7, (2<<4)+3);
}
#define init_7_7_2_4
void FASTCALL call_7_7_2_4(SH2_struct* sh)
{
  SH2addi(sh, 7, (2<<4)+4);
}
#define init_7_7_2_5
void FASTCALL call_7_7_2_5(SH2_struct* sh)
{
  SH2addi(sh, 7, (2<<4)+5);
}
#define init_7_7_2_6
void FASTCALL call_7_7_2_6(SH2_struct* sh)
{
  SH2addi(sh, 7, (2<<4)+6);
}
#define init_7_7_2_7
void FASTCALL call_7_7_2_7(SH2_struct* sh)
{
  SH2addi(sh, 7, (2<<4)+7);
}
#define init_7_7_2_8
void FASTCALL call_7_7_2_8(SH2_struct* sh)
{
  SH2addi(sh, 7, (2<<4)+8);
}
#define init_7_7_2_9
void FASTCALL call_7_7_2_9(SH2_struct* sh)
{
  SH2addi(sh, 7, (2<<4)+9);
}
#define init_7_7_2_10
void FASTCALL call_7_7_2_10(SH2_struct* sh)
{
  SH2addi(sh, 7, (2<<4)+10);
}
#define init_7_7_2_11
void FASTCALL call_7_7_2_11(SH2_struct* sh)
{
  SH2addi(sh, 7, (2<<4)+11);
}
#define init_7_7_2_12
void FASTCALL call_7_7_2_12(SH2_struct* sh)
{
  SH2addi(sh, 7, (2<<4)+12);
}
#define init_7_7_2_13
void FASTCALL call_7_7_2_13(SH2_struct* sh)
{
  SH2addi(sh, 7, (2<<4)+13);
}
#define init_7_7_2_14
void FASTCALL call_7_7_2_14(SH2_struct* sh)
{
  SH2addi(sh, 7, (2<<4)+14);
}
#define init_7_7_2_15
void FASTCALL call_7_7_2_15(SH2_struct* sh)
{
  SH2addi(sh, 7, (2<<4)+15);
}
#define init_7_7_3_0
void FASTCALL call_7_7_3_0(SH2_struct* sh)
{
  SH2addi(sh, 7, (3<<4)+0);
}
#define init_7_7_3_1
void FASTCALL call_7_7_3_1(SH2_struct* sh)
{
  SH2addi(sh, 7, (3<<4)+1);
}
#define init_7_7_3_2
void FASTCALL call_7_7_3_2(SH2_struct* sh)
{
  SH2addi(sh, 7, (3<<4)+2);
}
#define init_7_7_3_3
void FASTCALL call_7_7_3_3(SH2_struct* sh)
{
  SH2addi(sh, 7, (3<<4)+3);
}
#define init_7_7_3_4
void FASTCALL call_7_7_3_4(SH2_struct* sh)
{
  SH2addi(sh, 7, (3<<4)+4);
}
#define init_7_7_3_5
void FASTCALL call_7_7_3_5(SH2_struct* sh)
{
  SH2addi(sh, 7, (3<<4)+5);
}
#define init_7_7_3_6
void FASTCALL call_7_7_3_6(SH2_struct* sh)
{
  SH2addi(sh, 7, (3<<4)+6);
}
#define init_7_7_3_7
void FASTCALL call_7_7_3_7(SH2_struct* sh)
{
  SH2addi(sh, 7, (3<<4)+7);
}
#define init_7_7_3_8
void FASTCALL call_7_7_3_8(SH2_struct* sh)
{
  SH2addi(sh, 7, (3<<4)+8);
}
#define init_7_7_3_9
void FASTCALL call_7_7_3_9(SH2_struct* sh)
{
  SH2addi(sh, 7, (3<<4)+9);
}
#define init_7_7_3_10
void FASTCALL call_7_7_3_10(SH2_struct* sh)
{
  SH2addi(sh, 7, (3<<4)+10);
}
#define init_7_7_3_11
void FASTCALL call_7_7_3_11(SH2_struct* sh)
{
  SH2addi(sh, 7, (3<<4)+11);
}
#define init_7_7_3_12
void FASTCALL call_7_7_3_12(SH2_struct* sh)
{
  SH2addi(sh, 7, (3<<4)+12);
}
#define init_7_7_3_13
void FASTCALL call_7_7_3_13(SH2_struct* sh)
{
  SH2addi(sh, 7, (3<<4)+13);
}
#define init_7_7_3_14
void FASTCALL call_7_7_3_14(SH2_struct* sh)
{
  SH2addi(sh, 7, (3<<4)+14);
}
#define init_7_7_3_15
void FASTCALL call_7_7_3_15(SH2_struct* sh)
{
  SH2addi(sh, 7, (3<<4)+15);
}
#define init_7_7_4_0
void FASTCALL call_7_7_4_0(SH2_struct* sh)
{
  SH2addi(sh, 7, (4<<4)+0);
}
#define init_7_7_4_1
void FASTCALL call_7_7_4_1(SH2_struct* sh)
{
  SH2addi(sh, 7, (4<<4)+1);
}
#define init_7_7_4_2
void FASTCALL call_7_7_4_2(SH2_struct* sh)
{
  SH2addi(sh, 7, (4<<4)+2);
}
#define init_7_7_4_3
void FASTCALL call_7_7_4_3(SH2_struct* sh)
{
  SH2addi(sh, 7, (4<<4)+3);
}
#define init_7_7_4_4
void FASTCALL call_7_7_4_4(SH2_struct* sh)
{
  SH2addi(sh, 7, (4<<4)+4);
}
#define init_7_7_4_5
void FASTCALL call_7_7_4_5(SH2_struct* sh)
{
  SH2addi(sh, 7, (4<<4)+5);
}
#define init_7_7_4_6
void FASTCALL call_7_7_4_6(SH2_struct* sh)
{
  SH2addi(sh, 7, (4<<4)+6);
}
#define init_7_7_4_7
void FASTCALL call_7_7_4_7(SH2_struct* sh)
{
  SH2addi(sh, 7, (4<<4)+7);
}
#define init_7_7_4_8
void FASTCALL call_7_7_4_8(SH2_struct* sh)
{
  SH2addi(sh, 7, (4<<4)+8);
}
#define init_7_7_4_9
void FASTCALL call_7_7_4_9(SH2_struct* sh)
{
  SH2addi(sh, 7, (4<<4)+9);
}
#define init_7_7_4_10
void FASTCALL call_7_7_4_10(SH2_struct* sh)
{
  SH2addi(sh, 7, (4<<4)+10);
}
#define init_7_7_4_11
void FASTCALL call_7_7_4_11(SH2_struct* sh)
{
  SH2addi(sh, 7, (4<<4)+11);
}
#define init_7_7_4_12
void FASTCALL call_7_7_4_12(SH2_struct* sh)
{
  SH2addi(sh, 7, (4<<4)+12);
}
#define init_7_7_4_13
void FASTCALL call_7_7_4_13(SH2_struct* sh)
{
  SH2addi(sh, 7, (4<<4)+13);
}
#define init_7_7_4_14
void FASTCALL call_7_7_4_14(SH2_struct* sh)
{
  SH2addi(sh, 7, (4<<4)+14);
}
#define init_7_7_4_15
void FASTCALL call_7_7_4_15(SH2_struct* sh)
{
  SH2addi(sh, 7, (4<<4)+15);
}
#define init_7_7_5_0
void FASTCALL call_7_7_5_0(SH2_struct* sh)
{
  SH2addi(sh, 7, (5<<4)+0);
}
#define init_7_7_5_1
void FASTCALL call_7_7_5_1(SH2_struct* sh)
{
  SH2addi(sh, 7, (5<<4)+1);
}
#define init_7_7_5_2
void FASTCALL call_7_7_5_2(SH2_struct* sh)
{
  SH2addi(sh, 7, (5<<4)+2);
}
#define init_7_7_5_3
void FASTCALL call_7_7_5_3(SH2_struct* sh)
{
  SH2addi(sh, 7, (5<<4)+3);
}
#define init_7_7_5_4
void FASTCALL call_7_7_5_4(SH2_struct* sh)
{
  SH2addi(sh, 7, (5<<4)+4);
}
#define init_7_7_5_5
void FASTCALL call_7_7_5_5(SH2_struct* sh)
{
  SH2addi(sh, 7, (5<<4)+5);
}
#define init_7_7_5_6
void FASTCALL call_7_7_5_6(SH2_struct* sh)
{
  SH2addi(sh, 7, (5<<4)+6);
}
#define init_7_7_5_7
void FASTCALL call_7_7_5_7(SH2_struct* sh)
{
  SH2addi(sh, 7, (5<<4)+7);
}
#define init_7_7_5_8
void FASTCALL call_7_7_5_8(SH2_struct* sh)
{
  SH2addi(sh, 7, (5<<4)+8);
}
#define init_7_7_5_9
void FASTCALL call_7_7_5_9(SH2_struct* sh)
{
  SH2addi(sh, 7, (5<<4)+9);
}
#define init_7_7_5_10
void FASTCALL call_7_7_5_10(SH2_struct* sh)
{
  SH2addi(sh, 7, (5<<4)+10);
}
#define init_7_7_5_11
void FASTCALL call_7_7_5_11(SH2_struct* sh)
{
  SH2addi(sh, 7, (5<<4)+11);
}
#define init_7_7_5_12
void FASTCALL call_7_7_5_12(SH2_struct* sh)
{
  SH2addi(sh, 7, (5<<4)+12);
}
#define init_7_7_5_13
void FASTCALL call_7_7_5_13(SH2_struct* sh)
{
  SH2addi(sh, 7, (5<<4)+13);
}
#define init_7_7_5_14
void FASTCALL call_7_7_5_14(SH2_struct* sh)
{
  SH2addi(sh, 7, (5<<4)+14);
}
#define init_7_7_5_15
void FASTCALL call_7_7_5_15(SH2_struct* sh)
{
  SH2addi(sh, 7, (5<<4)+15);
}
#define init_7_7_6_0
void FASTCALL call_7_7_6_0(SH2_struct* sh)
{
  SH2addi(sh, 7, (6<<4)+0);
}
#define init_7_7_6_1
void FASTCALL call_7_7_6_1(SH2_struct* sh)
{
  SH2addi(sh, 7, (6<<4)+1);
}
#define init_7_7_6_2
void FASTCALL call_7_7_6_2(SH2_struct* sh)
{
  SH2addi(sh, 7, (6<<4)+2);
}
#define init_7_7_6_3
void FASTCALL call_7_7_6_3(SH2_struct* sh)
{
  SH2addi(sh, 7, (6<<4)+3);
}
#define init_7_7_6_4
void FASTCALL call_7_7_6_4(SH2_struct* sh)
{
  SH2addi(sh, 7, (6<<4)+4);
}
#define init_7_7_6_5
void FASTCALL call_7_7_6_5(SH2_struct* sh)
{
  SH2addi(sh, 7, (6<<4)+5);
}
#define init_7_7_6_6
void FASTCALL call_7_7_6_6(SH2_struct* sh)
{
  SH2addi(sh, 7, (6<<4)+6);
}
#define init_7_7_6_7
void FASTCALL call_7_7_6_7(SH2_struct* sh)
{
  SH2addi(sh, 7, (6<<4)+7);
}
#define init_7_7_6_8
void FASTCALL call_7_7_6_8(SH2_struct* sh)
{
  SH2addi(sh, 7, (6<<4)+8);
}
#define init_7_7_6_9
void FASTCALL call_7_7_6_9(SH2_struct* sh)
{
  SH2addi(sh, 7, (6<<4)+9);
}
#define init_7_7_6_10
void FASTCALL call_7_7_6_10(SH2_struct* sh)
{
  SH2addi(sh, 7, (6<<4)+10);
}
#define init_7_7_6_11
void FASTCALL call_7_7_6_11(SH2_struct* sh)
{
  SH2addi(sh, 7, (6<<4)+11);
}
#define init_7_7_6_12
void FASTCALL call_7_7_6_12(SH2_struct* sh)
{
  SH2addi(sh, 7, (6<<4)+12);
}
#define init_7_7_6_13
void FASTCALL call_7_7_6_13(SH2_struct* sh)
{
  SH2addi(sh, 7, (6<<4)+13);
}
#define init_7_7_6_14
void FASTCALL call_7_7_6_14(SH2_struct* sh)
{
  SH2addi(sh, 7, (6<<4)+14);
}
#define init_7_7_6_15
void FASTCALL call_7_7_6_15(SH2_struct* sh)
{
  SH2addi(sh, 7, (6<<4)+15);
}
#define init_7_7_7_0
void FASTCALL call_7_7_7_0(SH2_struct* sh)
{
  SH2addi(sh, 7, (7<<4)+0);
}
#define init_7_7_7_1
void FASTCALL call_7_7_7_1(SH2_struct* sh)
{
  SH2addi(sh, 7, (7<<4)+1);
}
#define init_7_7_7_2
void FASTCALL call_7_7_7_2(SH2_struct* sh)
{
  SH2addi(sh, 7, (7<<4)+2);
}
#define init_7_7_7_3
void FASTCALL call_7_7_7_3(SH2_struct* sh)
{
  SH2addi(sh, 7, (7<<4)+3);
}
#define init_7_7_7_4
void FASTCALL call_7_7_7_4(SH2_struct* sh)
{
  SH2addi(sh, 7, (7<<4)+4);
}
#define init_7_7_7_5
void FASTCALL call_7_7_7_5(SH2_struct* sh)
{
  SH2addi(sh, 7, (7<<4)+5);
}
#define init_7_7_7_6
void FASTCALL call_7_7_7_6(SH2_struct* sh)
{
  SH2addi(sh, 7, (7<<4)+6);
}
#define init_7_7_7_7
void FASTCALL call_7_7_7_7(SH2_struct* sh)
{
  SH2addi(sh, 7, (7<<4)+7);
}
#define init_7_7_7_8
void FASTCALL call_7_7_7_8(SH2_struct* sh)
{
  SH2addi(sh, 7, (7<<4)+8);
}
#define init_7_7_7_9
void FASTCALL call_7_7_7_9(SH2_struct* sh)
{
  SH2addi(sh, 7, (7<<4)+9);
}
#define init_7_7_7_10
void FASTCALL call_7_7_7_10(SH2_struct* sh)
{
  SH2addi(sh, 7, (7<<4)+10);
}
#define init_7_7_7_11
void FASTCALL call_7_7_7_11(SH2_struct* sh)
{
  SH2addi(sh, 7, (7<<4)+11);
}
#define init_7_7_7_12
void FASTCALL call_7_7_7_12(SH2_struct* sh)
{
  SH2addi(sh, 7, (7<<4)+12);
}
#define init_7_7_7_13
void FASTCALL call_7_7_7_13(SH2_struct* sh)
{
  SH2addi(sh, 7, (7<<4)+13);
}
#define init_7_7_7_14
void FASTCALL call_7_7_7_14(SH2_struct* sh)
{
  SH2addi(sh, 7, (7<<4)+14);
}
#define init_7_7_7_15
void FASTCALL call_7_7_7_15(SH2_struct* sh)
{
  SH2addi(sh, 7, (7<<4)+15);
}
#define init_7_7_8_0
void FASTCALL call_7_7_8_0(SH2_struct* sh)
{
  SH2addi(sh, 7, (8<<4)+0);
}
#define init_7_7_8_1
void FASTCALL call_7_7_8_1(SH2_struct* sh)
{
  SH2addi(sh, 7, (8<<4)+1);
}
#define init_7_7_8_2
void FASTCALL call_7_7_8_2(SH2_struct* sh)
{
  SH2addi(sh, 7, (8<<4)+2);
}
#define init_7_7_8_3
void FASTCALL call_7_7_8_3(SH2_struct* sh)
{
  SH2addi(sh, 7, (8<<4)+3);
}
#define init_7_7_8_4
void FASTCALL call_7_7_8_4(SH2_struct* sh)
{
  SH2addi(sh, 7, (8<<4)+4);
}
#define init_7_7_8_5
void FASTCALL call_7_7_8_5(SH2_struct* sh)
{
  SH2addi(sh, 7, (8<<4)+5);
}
#define init_7_7_8_6
void FASTCALL call_7_7_8_6(SH2_struct* sh)
{
  SH2addi(sh, 7, (8<<4)+6);
}
#define init_7_7_8_7
void FASTCALL call_7_7_8_7(SH2_struct* sh)
{
  SH2addi(sh, 7, (8<<4)+7);
}
#define init_7_7_8_8
void FASTCALL call_7_7_8_8(SH2_struct* sh)
{
  SH2addi(sh, 7, (8<<4)+8);
}
#define init_7_7_8_9
void FASTCALL call_7_7_8_9(SH2_struct* sh)
{
  SH2addi(sh, 7, (8<<4)+9);
}
#define init_7_7_8_10
void FASTCALL call_7_7_8_10(SH2_struct* sh)
{
  SH2addi(sh, 7, (8<<4)+10);
}
#define init_7_7_8_11
void FASTCALL call_7_7_8_11(SH2_struct* sh)
{
  SH2addi(sh, 7, (8<<4)+11);
}
#define init_7_7_8_12
void FASTCALL call_7_7_8_12(SH2_struct* sh)
{
  SH2addi(sh, 7, (8<<4)+12);
}
#define init_7_7_8_13
void FASTCALL call_7_7_8_13(SH2_struct* sh)
{
  SH2addi(sh, 7, (8<<4)+13);
}
#define init_7_7_8_14
void FASTCALL call_7_7_8_14(SH2_struct* sh)
{
  SH2addi(sh, 7, (8<<4)+14);
}
#define init_7_7_8_15
void FASTCALL call_7_7_8_15(SH2_struct* sh)
{
  SH2addi(sh, 7, (8<<4)+15);
}
#define init_7_7_9_0
void FASTCALL call_7_7_9_0(SH2_struct* sh)
{
  SH2addi(sh, 7, (9<<4)+0);
}
#define init_7_7_9_1
void FASTCALL call_7_7_9_1(SH2_struct* sh)
{
  SH2addi(sh, 7, (9<<4)+1);
}
#define init_7_7_9_2
void FASTCALL call_7_7_9_2(SH2_struct* sh)
{
  SH2addi(sh, 7, (9<<4)+2);
}
#define init_7_7_9_3
void FASTCALL call_7_7_9_3(SH2_struct* sh)
{
  SH2addi(sh, 7, (9<<4)+3);
}
#define init_7_7_9_4
void FASTCALL call_7_7_9_4(SH2_struct* sh)
{
  SH2addi(sh, 7, (9<<4)+4);
}
#define init_7_7_9_5
void FASTCALL call_7_7_9_5(SH2_struct* sh)
{
  SH2addi(sh, 7, (9<<4)+5);
}
#define init_7_7_9_6
void FASTCALL call_7_7_9_6(SH2_struct* sh)
{
  SH2addi(sh, 7, (9<<4)+6);
}
#define init_7_7_9_7
void FASTCALL call_7_7_9_7(SH2_struct* sh)
{
  SH2addi(sh, 7, (9<<4)+7);
}
#define init_7_7_9_8
void FASTCALL call_7_7_9_8(SH2_struct* sh)
{
  SH2addi(sh, 7, (9<<4)+8);
}
#define init_7_7_9_9
void FASTCALL call_7_7_9_9(SH2_struct* sh)
{
  SH2addi(sh, 7, (9<<4)+9);
}
#define init_7_7_9_10
void FASTCALL call_7_7_9_10(SH2_struct* sh)
{
  SH2addi(sh, 7, (9<<4)+10);
}
#define init_7_7_9_11
void FASTCALL call_7_7_9_11(SH2_struct* sh)
{
  SH2addi(sh, 7, (9<<4)+11);
}
#define init_7_7_9_12
void FASTCALL call_7_7_9_12(SH2_struct* sh)
{
  SH2addi(sh, 7, (9<<4)+12);
}
#define init_7_7_9_13
void FASTCALL call_7_7_9_13(SH2_struct* sh)
{
  SH2addi(sh, 7, (9<<4)+13);
}
#define init_7_7_9_14
void FASTCALL call_7_7_9_14(SH2_struct* sh)
{
  SH2addi(sh, 7, (9<<4)+14);
}
#define init_7_7_9_15
void FASTCALL call_7_7_9_15(SH2_struct* sh)
{
  SH2addi(sh, 7, (9<<4)+15);
}
#define init_7_7_10_0
void FASTCALL call_7_7_10_0(SH2_struct* sh)
{
  SH2addi(sh, 7, (10<<4)+0);
}
#define init_7_7_10_1
void FASTCALL call_7_7_10_1(SH2_struct* sh)
{
  SH2addi(sh, 7, (10<<4)+1);
}
#define init_7_7_10_2
void FASTCALL call_7_7_10_2(SH2_struct* sh)
{
  SH2addi(sh, 7, (10<<4)+2);
}
#define init_7_7_10_3
void FASTCALL call_7_7_10_3(SH2_struct* sh)
{
  SH2addi(sh, 7, (10<<4)+3);
}
#define init_7_7_10_4
void FASTCALL call_7_7_10_4(SH2_struct* sh)
{
  SH2addi(sh, 7, (10<<4)+4);
}
#define init_7_7_10_5
void FASTCALL call_7_7_10_5(SH2_struct* sh)
{
  SH2addi(sh, 7, (10<<4)+5);
}
#define init_7_7_10_6
void FASTCALL call_7_7_10_6(SH2_struct* sh)
{
  SH2addi(sh, 7, (10<<4)+6);
}
#define init_7_7_10_7
void FASTCALL call_7_7_10_7(SH2_struct* sh)
{
  SH2addi(sh, 7, (10<<4)+7);
}
#define init_7_7_10_8
void FASTCALL call_7_7_10_8(SH2_struct* sh)
{
  SH2addi(sh, 7, (10<<4)+8);
}
#define init_7_7_10_9
void FASTCALL call_7_7_10_9(SH2_struct* sh)
{
  SH2addi(sh, 7, (10<<4)+9);
}
#define init_7_7_10_10
void FASTCALL call_7_7_10_10(SH2_struct* sh)
{
  SH2addi(sh, 7, (10<<4)+10);
}
#define init_7_7_10_11
void FASTCALL call_7_7_10_11(SH2_struct* sh)
{
  SH2addi(sh, 7, (10<<4)+11);
}
#define init_7_7_10_12
void FASTCALL call_7_7_10_12(SH2_struct* sh)
{
  SH2addi(sh, 7, (10<<4)+12);
}
#define init_7_7_10_13
void FASTCALL call_7_7_10_13(SH2_struct* sh)
{
  SH2addi(sh, 7, (10<<4)+13);
}
#define init_7_7_10_14
void FASTCALL call_7_7_10_14(SH2_struct* sh)
{
  SH2addi(sh, 7, (10<<4)+14);
}
#define init_7_7_10_15
void FASTCALL call_7_7_10_15(SH2_struct* sh)
{
  SH2addi(sh, 7, (10<<4)+15);
}
#define init_7_7_11_0
void FASTCALL call_7_7_11_0(SH2_struct* sh)
{
  SH2addi(sh, 7, (11<<4)+0);
}
#define init_7_7_11_1
void FASTCALL call_7_7_11_1(SH2_struct* sh)
{
  SH2addi(sh, 7, (11<<4)+1);
}
#define init_7_7_11_2
void FASTCALL call_7_7_11_2(SH2_struct* sh)
{
  SH2addi(sh, 7, (11<<4)+2);
}
#define init_7_7_11_3
void FASTCALL call_7_7_11_3(SH2_struct* sh)
{
  SH2addi(sh, 7, (11<<4)+3);
}
#define init_7_7_11_4
void FASTCALL call_7_7_11_4(SH2_struct* sh)
{
  SH2addi(sh, 7, (11<<4)+4);
}
#define init_7_7_11_5
void FASTCALL call_7_7_11_5(SH2_struct* sh)
{
  SH2addi(sh, 7, (11<<4)+5);
}
#define init_7_7_11_6
void FASTCALL call_7_7_11_6(SH2_struct* sh)
{
  SH2addi(sh, 7, (11<<4)+6);
}
#define init_7_7_11_7
void FASTCALL call_7_7_11_7(SH2_struct* sh)
{
  SH2addi(sh, 7, (11<<4)+7);
}
#define init_7_7_11_8
void FASTCALL call_7_7_11_8(SH2_struct* sh)
{
  SH2addi(sh, 7, (11<<4)+8);
}
#define init_7_7_11_9
void FASTCALL call_7_7_11_9(SH2_struct* sh)
{
  SH2addi(sh, 7, (11<<4)+9);
}
#define init_7_7_11_10
void FASTCALL call_7_7_11_10(SH2_struct* sh)
{
  SH2addi(sh, 7, (11<<4)+10);
}
#define init_7_7_11_11
void FASTCALL call_7_7_11_11(SH2_struct* sh)
{
  SH2addi(sh, 7, (11<<4)+11);
}
#define init_7_7_11_12
void FASTCALL call_7_7_11_12(SH2_struct* sh)
{
  SH2addi(sh, 7, (11<<4)+12);
}
#define init_7_7_11_13
void FASTCALL call_7_7_11_13(SH2_struct* sh)
{
  SH2addi(sh, 7, (11<<4)+13);
}
#define init_7_7_11_14
void FASTCALL call_7_7_11_14(SH2_struct* sh)
{
  SH2addi(sh, 7, (11<<4)+14);
}
#define init_7_7_11_15
void FASTCALL call_7_7_11_15(SH2_struct* sh)
{
  SH2addi(sh, 7, (11<<4)+15);
}
#define init_7_7_12_0
void FASTCALL call_7_7_12_0(SH2_struct* sh)
{
  SH2addi(sh, 7, (12<<4)+0);
}
#define init_7_7_12_1
void FASTCALL call_7_7_12_1(SH2_struct* sh)
{
  SH2addi(sh, 7, (12<<4)+1);
}
#define init_7_7_12_2
void FASTCALL call_7_7_12_2(SH2_struct* sh)
{
  SH2addi(sh, 7, (12<<4)+2);
}
#define init_7_7_12_3
void FASTCALL call_7_7_12_3(SH2_struct* sh)
{
  SH2addi(sh, 7, (12<<4)+3);
}
#define init_7_7_12_4
void FASTCALL call_7_7_12_4(SH2_struct* sh)
{
  SH2addi(sh, 7, (12<<4)+4);
}
#define init_7_7_12_5
void FASTCALL call_7_7_12_5(SH2_struct* sh)
{
  SH2addi(sh, 7, (12<<4)+5);
}
#define init_7_7_12_6
void FASTCALL call_7_7_12_6(SH2_struct* sh)
{
  SH2addi(sh, 7, (12<<4)+6);
}
#define init_7_7_12_7
void FASTCALL call_7_7_12_7(SH2_struct* sh)
{
  SH2addi(sh, 7, (12<<4)+7);
}
#define init_7_7_12_8
void FASTCALL call_7_7_12_8(SH2_struct* sh)
{
  SH2addi(sh, 7, (12<<4)+8);
}
#define init_7_7_12_9
void FASTCALL call_7_7_12_9(SH2_struct* sh)
{
  SH2addi(sh, 7, (12<<4)+9);
}
#define init_7_7_12_10
void FASTCALL call_7_7_12_10(SH2_struct* sh)
{
  SH2addi(sh, 7, (12<<4)+10);
}
#define init_7_7_12_11
void FASTCALL call_7_7_12_11(SH2_struct* sh)
{
  SH2addi(sh, 7, (12<<4)+11);
}
#define init_7_7_12_12
void FASTCALL call_7_7_12_12(SH2_struct* sh)
{
  SH2addi(sh, 7, (12<<4)+12);
}
#define init_7_7_12_13
void FASTCALL call_7_7_12_13(SH2_struct* sh)
{
  SH2addi(sh, 7, (12<<4)+13);
}
#define init_7_7_12_14
void FASTCALL call_7_7_12_14(SH2_struct* sh)
{
  SH2addi(sh, 7, (12<<4)+14);
}
#define init_7_7_12_15
void FASTCALL call_7_7_12_15(SH2_struct* sh)
{
  SH2addi(sh, 7, (12<<4)+15);
}
#define init_7_7_13_0
void FASTCALL call_7_7_13_0(SH2_struct* sh)
{
  SH2addi(sh, 7, (13<<4)+0);
}
#define init_7_7_13_1
void FASTCALL call_7_7_13_1(SH2_struct* sh)
{
  SH2addi(sh, 7, (13<<4)+1);
}
#define init_7_7_13_2
void FASTCALL call_7_7_13_2(SH2_struct* sh)
{
  SH2addi(sh, 7, (13<<4)+2);
}
#define init_7_7_13_3
void FASTCALL call_7_7_13_3(SH2_struct* sh)
{
  SH2addi(sh, 7, (13<<4)+3);
}
#define init_7_7_13_4
void FASTCALL call_7_7_13_4(SH2_struct* sh)
{
  SH2addi(sh, 7, (13<<4)+4);
}
#define init_7_7_13_5
void FASTCALL call_7_7_13_5(SH2_struct* sh)
{
  SH2addi(sh, 7, (13<<4)+5);
}
#define init_7_7_13_6
void FASTCALL call_7_7_13_6(SH2_struct* sh)
{
  SH2addi(sh, 7, (13<<4)+6);
}
#define init_7_7_13_7
void FASTCALL call_7_7_13_7(SH2_struct* sh)
{
  SH2addi(sh, 7, (13<<4)+7);
}
#define init_7_7_13_8
void FASTCALL call_7_7_13_8(SH2_struct* sh)
{
  SH2addi(sh, 7, (13<<4)+8);
}
#define init_7_7_13_9
void FASTCALL call_7_7_13_9(SH2_struct* sh)
{
  SH2addi(sh, 7, (13<<4)+9);
}
#define init_7_7_13_10
void FASTCALL call_7_7_13_10(SH2_struct* sh)
{
  SH2addi(sh, 7, (13<<4)+10);
}
#define init_7_7_13_11
void FASTCALL call_7_7_13_11(SH2_struct* sh)
{
  SH2addi(sh, 7, (13<<4)+11);
}
#define init_7_7_13_12
void FASTCALL call_7_7_13_12(SH2_struct* sh)
{
  SH2addi(sh, 7, (13<<4)+12);
}
#define init_7_7_13_13
void FASTCALL call_7_7_13_13(SH2_struct* sh)
{
  SH2addi(sh, 7, (13<<4)+13);
}
#define init_7_7_13_14
void FASTCALL call_7_7_13_14(SH2_struct* sh)
{
  SH2addi(sh, 7, (13<<4)+14);
}
#define init_7_7_13_15
void FASTCALL call_7_7_13_15(SH2_struct* sh)
{
  SH2addi(sh, 7, (13<<4)+15);
}
#define init_7_7_14_0
void FASTCALL call_7_7_14_0(SH2_struct* sh)
{
  SH2addi(sh, 7, (14<<4)+0);
}
#define init_7_7_14_1
void FASTCALL call_7_7_14_1(SH2_struct* sh)
{
  SH2addi(sh, 7, (14<<4)+1);
}
#define init_7_7_14_2
void FASTCALL call_7_7_14_2(SH2_struct* sh)
{
  SH2addi(sh, 7, (14<<4)+2);
}
#define init_7_7_14_3
void FASTCALL call_7_7_14_3(SH2_struct* sh)
{
  SH2addi(sh, 7, (14<<4)+3);
}
#define init_7_7_14_4
void FASTCALL call_7_7_14_4(SH2_struct* sh)
{
  SH2addi(sh, 7, (14<<4)+4);
}
#define init_7_7_14_5
void FASTCALL call_7_7_14_5(SH2_struct* sh)
{
  SH2addi(sh, 7, (14<<4)+5);
}
#define init_7_7_14_6
void FASTCALL call_7_7_14_6(SH2_struct* sh)
{
  SH2addi(sh, 7, (14<<4)+6);
}
#define init_7_7_14_7
void FASTCALL call_7_7_14_7(SH2_struct* sh)
{
  SH2addi(sh, 7, (14<<4)+7);
}
#define init_7_7_14_8
void FASTCALL call_7_7_14_8(SH2_struct* sh)
{
  SH2addi(sh, 7, (14<<4)+8);
}
#define init_7_7_14_9
void FASTCALL call_7_7_14_9(SH2_struct* sh)
{
  SH2addi(sh, 7, (14<<4)+9);
}
#define init_7_7_14_10
void FASTCALL call_7_7_14_10(SH2_struct* sh)
{
  SH2addi(sh, 7, (14<<4)+10);
}
#define init_7_7_14_11
void FASTCALL call_7_7_14_11(SH2_struct* sh)
{
  SH2addi(sh, 7, (14<<4)+11);
}
#define init_7_7_14_12
void FASTCALL call_7_7_14_12(SH2_struct* sh)
{
  SH2addi(sh, 7, (14<<4)+12);
}
#define init_7_7_14_13
void FASTCALL call_7_7_14_13(SH2_struct* sh)
{
  SH2addi(sh, 7, (14<<4)+13);
}
#define init_7_7_14_14
void FASTCALL call_7_7_14_14(SH2_struct* sh)
{
  SH2addi(sh, 7, (14<<4)+14);
}
#define init_7_7_14_15
void FASTCALL call_7_7_14_15(SH2_struct* sh)
{
  SH2addi(sh, 7, (14<<4)+15);
}
#define init_7_7_15_0
void FASTCALL call_7_7_15_0(SH2_struct* sh)
{
  SH2addi(sh, 7, (15<<4)+0);
}
#define init_7_7_15_1
void FASTCALL call_7_7_15_1(SH2_struct* sh)
{
  SH2addi(sh, 7, (15<<4)+1);
}
#define init_7_7_15_2
void FASTCALL call_7_7_15_2(SH2_struct* sh)
{
  SH2addi(sh, 7, (15<<4)+2);
}
#define init_7_7_15_3
void FASTCALL call_7_7_15_3(SH2_struct* sh)
{
  SH2addi(sh, 7, (15<<4)+3);
}
#define init_7_7_15_4
void FASTCALL call_7_7_15_4(SH2_struct* sh)
{
  SH2addi(sh, 7, (15<<4)+4);
}
#define init_7_7_15_5
void FASTCALL call_7_7_15_5(SH2_struct* sh)
{
  SH2addi(sh, 7, (15<<4)+5);
}
#define init_7_7_15_6
void FASTCALL call_7_7_15_6(SH2_struct* sh)
{
  SH2addi(sh, 7, (15<<4)+6);
}
#define init_7_7_15_7
void FASTCALL call_7_7_15_7(SH2_struct* sh)
{
  SH2addi(sh, 7, (15<<4)+7);
}
#define init_7_7_15_8
void FASTCALL call_7_7_15_8(SH2_struct* sh)
{
  SH2addi(sh, 7, (15<<4)+8);
}
#define init_7_7_15_9
void FASTCALL call_7_7_15_9(SH2_struct* sh)
{
  SH2addi(sh, 7, (15<<4)+9);
}
#define init_7_7_15_10
void FASTCALL call_7_7_15_10(SH2_struct* sh)
{
  SH2addi(sh, 7, (15<<4)+10);
}
#define init_7_7_15_11
void FASTCALL call_7_7_15_11(SH2_struct* sh)
{
  SH2addi(sh, 7, (15<<4)+11);
}
#define init_7_7_15_12
void FASTCALL call_7_7_15_12(SH2_struct* sh)
{
  SH2addi(sh, 7, (15<<4)+12);
}
#define init_7_7_15_13
void FASTCALL call_7_7_15_13(SH2_struct* sh)
{
  SH2addi(sh, 7, (15<<4)+13);
}
#define init_7_7_15_14
void FASTCALL call_7_7_15_14(SH2_struct* sh)
{
  SH2addi(sh, 7, (15<<4)+14);
}
#define init_7_7_15_15
void FASTCALL call_7_7_15_15(SH2_struct* sh)
{
  SH2addi(sh, 7, (15<<4)+15);
}
#define init_7_8_0_0
void FASTCALL call_7_8_0_0(SH2_struct* sh)
{
  SH2addi(sh, 8, (0<<4)+0);
}
#define init_7_8_0_1
void FASTCALL call_7_8_0_1(SH2_struct* sh)
{
  SH2addi(sh, 8, (0<<4)+1);
}
#define init_7_8_0_2
void FASTCALL call_7_8_0_2(SH2_struct* sh)
{
  SH2addi(sh, 8, (0<<4)+2);
}
#define init_7_8_0_3
void FASTCALL call_7_8_0_3(SH2_struct* sh)
{
  SH2addi(sh, 8, (0<<4)+3);
}
#define init_7_8_0_4
void FASTCALL call_7_8_0_4(SH2_struct* sh)
{
  SH2addi(sh, 8, (0<<4)+4);
}
#define init_7_8_0_5
void FASTCALL call_7_8_0_5(SH2_struct* sh)
{
  SH2addi(sh, 8, (0<<4)+5);
}
#define init_7_8_0_6
void FASTCALL call_7_8_0_6(SH2_struct* sh)
{
  SH2addi(sh, 8, (0<<4)+6);
}
#define init_7_8_0_7
void FASTCALL call_7_8_0_7(SH2_struct* sh)
{
  SH2addi(sh, 8, (0<<4)+7);
}
#define init_7_8_0_8
void FASTCALL call_7_8_0_8(SH2_struct* sh)
{
  SH2addi(sh, 8, (0<<4)+8);
}
#define init_7_8_0_9
void FASTCALL call_7_8_0_9(SH2_struct* sh)
{
  SH2addi(sh, 8, (0<<4)+9);
}
#define init_7_8_0_10
void FASTCALL call_7_8_0_10(SH2_struct* sh)
{
  SH2addi(sh, 8, (0<<4)+10);
}
#define init_7_8_0_11
void FASTCALL call_7_8_0_11(SH2_struct* sh)
{
  SH2addi(sh, 8, (0<<4)+11);
}
#define init_7_8_0_12
void FASTCALL call_7_8_0_12(SH2_struct* sh)
{
  SH2addi(sh, 8, (0<<4)+12);
}
#define init_7_8_0_13
void FASTCALL call_7_8_0_13(SH2_struct* sh)
{
  SH2addi(sh, 8, (0<<4)+13);
}
#define init_7_8_0_14
void FASTCALL call_7_8_0_14(SH2_struct* sh)
{
  SH2addi(sh, 8, (0<<4)+14);
}
#define init_7_8_0_15
void FASTCALL call_7_8_0_15(SH2_struct* sh)
{
  SH2addi(sh, 8, (0<<4)+15);
}
#define init_7_8_1_0
void FASTCALL call_7_8_1_0(SH2_struct* sh)
{
  SH2addi(sh, 8, (1<<4)+0);
}
#define init_7_8_1_1
void FASTCALL call_7_8_1_1(SH2_struct* sh)
{
  SH2addi(sh, 8, (1<<4)+1);
}
#define init_7_8_1_2
void FASTCALL call_7_8_1_2(SH2_struct* sh)
{
  SH2addi(sh, 8, (1<<4)+2);
}
#define init_7_8_1_3
void FASTCALL call_7_8_1_3(SH2_struct* sh)
{
  SH2addi(sh, 8, (1<<4)+3);
}
#define init_7_8_1_4
void FASTCALL call_7_8_1_4(SH2_struct* sh)
{
  SH2addi(sh, 8, (1<<4)+4);
}
#define init_7_8_1_5
void FASTCALL call_7_8_1_5(SH2_struct* sh)
{
  SH2addi(sh, 8, (1<<4)+5);
}
#define init_7_8_1_6
void FASTCALL call_7_8_1_6(SH2_struct* sh)
{
  SH2addi(sh, 8, (1<<4)+6);
}
#define init_7_8_1_7
void FASTCALL call_7_8_1_7(SH2_struct* sh)
{
  SH2addi(sh, 8, (1<<4)+7);
}
#define init_7_8_1_8
void FASTCALL call_7_8_1_8(SH2_struct* sh)
{
  SH2addi(sh, 8, (1<<4)+8);
}
#define init_7_8_1_9
void FASTCALL call_7_8_1_9(SH2_struct* sh)
{
  SH2addi(sh, 8, (1<<4)+9);
}
#define init_7_8_1_10
void FASTCALL call_7_8_1_10(SH2_struct* sh)
{
  SH2addi(sh, 8, (1<<4)+10);
}
#define init_7_8_1_11
void FASTCALL call_7_8_1_11(SH2_struct* sh)
{
  SH2addi(sh, 8, (1<<4)+11);
}
#define init_7_8_1_12
void FASTCALL call_7_8_1_12(SH2_struct* sh)
{
  SH2addi(sh, 8, (1<<4)+12);
}
#define init_7_8_1_13
void FASTCALL call_7_8_1_13(SH2_struct* sh)
{
  SH2addi(sh, 8, (1<<4)+13);
}
#define init_7_8_1_14
void FASTCALL call_7_8_1_14(SH2_struct* sh)
{
  SH2addi(sh, 8, (1<<4)+14);
}
#define init_7_8_1_15
void FASTCALL call_7_8_1_15(SH2_struct* sh)
{
  SH2addi(sh, 8, (1<<4)+15);
}
#define init_7_8_2_0
void FASTCALL call_7_8_2_0(SH2_struct* sh)
{
  SH2addi(sh, 8, (2<<4)+0);
}
#define init_7_8_2_1
void FASTCALL call_7_8_2_1(SH2_struct* sh)
{
  SH2addi(sh, 8, (2<<4)+1);
}
#define init_7_8_2_2
void FASTCALL call_7_8_2_2(SH2_struct* sh)
{
  SH2addi(sh, 8, (2<<4)+2);
}
#define init_7_8_2_3
void FASTCALL call_7_8_2_3(SH2_struct* sh)
{
  SH2addi(sh, 8, (2<<4)+3);
}
#define init_7_8_2_4
void FASTCALL call_7_8_2_4(SH2_struct* sh)
{
  SH2addi(sh, 8, (2<<4)+4);
}
#define init_7_8_2_5
void FASTCALL call_7_8_2_5(SH2_struct* sh)
{
  SH2addi(sh, 8, (2<<4)+5);
}
#define init_7_8_2_6
void FASTCALL call_7_8_2_6(SH2_struct* sh)
{
  SH2addi(sh, 8, (2<<4)+6);
}
#define init_7_8_2_7
void FASTCALL call_7_8_2_7(SH2_struct* sh)
{
  SH2addi(sh, 8, (2<<4)+7);
}
#define init_7_8_2_8
void FASTCALL call_7_8_2_8(SH2_struct* sh)
{
  SH2addi(sh, 8, (2<<4)+8);
}
#define init_7_8_2_9
void FASTCALL call_7_8_2_9(SH2_struct* sh)
{
  SH2addi(sh, 8, (2<<4)+9);
}
#define init_7_8_2_10
void FASTCALL call_7_8_2_10(SH2_struct* sh)
{
  SH2addi(sh, 8, (2<<4)+10);
}
#define init_7_8_2_11
void FASTCALL call_7_8_2_11(SH2_struct* sh)
{
  SH2addi(sh, 8, (2<<4)+11);
}
#define init_7_8_2_12
void FASTCALL call_7_8_2_12(SH2_struct* sh)
{
  SH2addi(sh, 8, (2<<4)+12);
}
#define init_7_8_2_13
void FASTCALL call_7_8_2_13(SH2_struct* sh)
{
  SH2addi(sh, 8, (2<<4)+13);
}
#define init_7_8_2_14
void FASTCALL call_7_8_2_14(SH2_struct* sh)
{
  SH2addi(sh, 8, (2<<4)+14);
}
#define init_7_8_2_15
void FASTCALL call_7_8_2_15(SH2_struct* sh)
{
  SH2addi(sh, 8, (2<<4)+15);
}
#define init_7_8_3_0
void FASTCALL call_7_8_3_0(SH2_struct* sh)
{
  SH2addi(sh, 8, (3<<4)+0);
}
#define init_7_8_3_1
void FASTCALL call_7_8_3_1(SH2_struct* sh)
{
  SH2addi(sh, 8, (3<<4)+1);
}
#define init_7_8_3_2
void FASTCALL call_7_8_3_2(SH2_struct* sh)
{
  SH2addi(sh, 8, (3<<4)+2);
}
#define init_7_8_3_3
void FASTCALL call_7_8_3_3(SH2_struct* sh)
{
  SH2addi(sh, 8, (3<<4)+3);
}
#define init_7_8_3_4
void FASTCALL call_7_8_3_4(SH2_struct* sh)
{
  SH2addi(sh, 8, (3<<4)+4);
}
#define init_7_8_3_5
void FASTCALL call_7_8_3_5(SH2_struct* sh)
{
  SH2addi(sh, 8, (3<<4)+5);
}
#define init_7_8_3_6
void FASTCALL call_7_8_3_6(SH2_struct* sh)
{
  SH2addi(sh, 8, (3<<4)+6);
}
#define init_7_8_3_7
void FASTCALL call_7_8_3_7(SH2_struct* sh)
{
  SH2addi(sh, 8, (3<<4)+7);
}
#define init_7_8_3_8
void FASTCALL call_7_8_3_8(SH2_struct* sh)
{
  SH2addi(sh, 8, (3<<4)+8);
}
#define init_7_8_3_9
void FASTCALL call_7_8_3_9(SH2_struct* sh)
{
  SH2addi(sh, 8, (3<<4)+9);
}
#define init_7_8_3_10
void FASTCALL call_7_8_3_10(SH2_struct* sh)
{
  SH2addi(sh, 8, (3<<4)+10);
}
#define init_7_8_3_11
void FASTCALL call_7_8_3_11(SH2_struct* sh)
{
  SH2addi(sh, 8, (3<<4)+11);
}
#define init_7_8_3_12
void FASTCALL call_7_8_3_12(SH2_struct* sh)
{
  SH2addi(sh, 8, (3<<4)+12);
}
#define init_7_8_3_13
void FASTCALL call_7_8_3_13(SH2_struct* sh)
{
  SH2addi(sh, 8, (3<<4)+13);
}
#define init_7_8_3_14
void FASTCALL call_7_8_3_14(SH2_struct* sh)
{
  SH2addi(sh, 8, (3<<4)+14);
}
#define init_7_8_3_15
void FASTCALL call_7_8_3_15(SH2_struct* sh)
{
  SH2addi(sh, 8, (3<<4)+15);
}
#define init_7_8_4_0
void FASTCALL call_7_8_4_0(SH2_struct* sh)
{
  SH2addi(sh, 8, (4<<4)+0);
}
#define init_7_8_4_1
void FASTCALL call_7_8_4_1(SH2_struct* sh)
{
  SH2addi(sh, 8, (4<<4)+1);
}
#define init_7_8_4_2
void FASTCALL call_7_8_4_2(SH2_struct* sh)
{
  SH2addi(sh, 8, (4<<4)+2);
}
#define init_7_8_4_3
void FASTCALL call_7_8_4_3(SH2_struct* sh)
{
  SH2addi(sh, 8, (4<<4)+3);
}
#define init_7_8_4_4
void FASTCALL call_7_8_4_4(SH2_struct* sh)
{
  SH2addi(sh, 8, (4<<4)+4);
}
#define init_7_8_4_5
void FASTCALL call_7_8_4_5(SH2_struct* sh)
{
  SH2addi(sh, 8, (4<<4)+5);
}
#define init_7_8_4_6
void FASTCALL call_7_8_4_6(SH2_struct* sh)
{
  SH2addi(sh, 8, (4<<4)+6);
}
#define init_7_8_4_7
void FASTCALL call_7_8_4_7(SH2_struct* sh)
{
  SH2addi(sh, 8, (4<<4)+7);
}
#define init_7_8_4_8
void FASTCALL call_7_8_4_8(SH2_struct* sh)
{
  SH2addi(sh, 8, (4<<4)+8);
}
#define init_7_8_4_9
void FASTCALL call_7_8_4_9(SH2_struct* sh)
{
  SH2addi(sh, 8, (4<<4)+9);
}
#define init_7_8_4_10
void FASTCALL call_7_8_4_10(SH2_struct* sh)
{
  SH2addi(sh, 8, (4<<4)+10);
}
#define init_7_8_4_11
void FASTCALL call_7_8_4_11(SH2_struct* sh)
{
  SH2addi(sh, 8, (4<<4)+11);
}
#define init_7_8_4_12
void FASTCALL call_7_8_4_12(SH2_struct* sh)
{
  SH2addi(sh, 8, (4<<4)+12);
}
#define init_7_8_4_13
void FASTCALL call_7_8_4_13(SH2_struct* sh)
{
  SH2addi(sh, 8, (4<<4)+13);
}
#define init_7_8_4_14
void FASTCALL call_7_8_4_14(SH2_struct* sh)
{
  SH2addi(sh, 8, (4<<4)+14);
}
#define init_7_8_4_15
void FASTCALL call_7_8_4_15(SH2_struct* sh)
{
  SH2addi(sh, 8, (4<<4)+15);
}
#define init_7_8_5_0
void FASTCALL call_7_8_5_0(SH2_struct* sh)
{
  SH2addi(sh, 8, (5<<4)+0);
}
#define init_7_8_5_1
void FASTCALL call_7_8_5_1(SH2_struct* sh)
{
  SH2addi(sh, 8, (5<<4)+1);
}
#define init_7_8_5_2
void FASTCALL call_7_8_5_2(SH2_struct* sh)
{
  SH2addi(sh, 8, (5<<4)+2);
}
#define init_7_8_5_3
void FASTCALL call_7_8_5_3(SH2_struct* sh)
{
  SH2addi(sh, 8, (5<<4)+3);
}
#define init_7_8_5_4
void FASTCALL call_7_8_5_4(SH2_struct* sh)
{
  SH2addi(sh, 8, (5<<4)+4);
}
#define init_7_8_5_5
void FASTCALL call_7_8_5_5(SH2_struct* sh)
{
  SH2addi(sh, 8, (5<<4)+5);
}
#define init_7_8_5_6
void FASTCALL call_7_8_5_6(SH2_struct* sh)
{
  SH2addi(sh, 8, (5<<4)+6);
}
#define init_7_8_5_7
void FASTCALL call_7_8_5_7(SH2_struct* sh)
{
  SH2addi(sh, 8, (5<<4)+7);
}
#define init_7_8_5_8
void FASTCALL call_7_8_5_8(SH2_struct* sh)
{
  SH2addi(sh, 8, (5<<4)+8);
}
#define init_7_8_5_9
void FASTCALL call_7_8_5_9(SH2_struct* sh)
{
  SH2addi(sh, 8, (5<<4)+9);
}
#define init_7_8_5_10
void FASTCALL call_7_8_5_10(SH2_struct* sh)
{
  SH2addi(sh, 8, (5<<4)+10);
}
#define init_7_8_5_11
void FASTCALL call_7_8_5_11(SH2_struct* sh)
{
  SH2addi(sh, 8, (5<<4)+11);
}
#define init_7_8_5_12
void FASTCALL call_7_8_5_12(SH2_struct* sh)
{
  SH2addi(sh, 8, (5<<4)+12);
}
#define init_7_8_5_13
void FASTCALL call_7_8_5_13(SH2_struct* sh)
{
  SH2addi(sh, 8, (5<<4)+13);
}
#define init_7_8_5_14
void FASTCALL call_7_8_5_14(SH2_struct* sh)
{
  SH2addi(sh, 8, (5<<4)+14);
}
#define init_7_8_5_15
void FASTCALL call_7_8_5_15(SH2_struct* sh)
{
  SH2addi(sh, 8, (5<<4)+15);
}
#define init_7_8_6_0
void FASTCALL call_7_8_6_0(SH2_struct* sh)
{
  SH2addi(sh, 8, (6<<4)+0);
}
#define init_7_8_6_1
void FASTCALL call_7_8_6_1(SH2_struct* sh)
{
  SH2addi(sh, 8, (6<<4)+1);
}
#define init_7_8_6_2
void FASTCALL call_7_8_6_2(SH2_struct* sh)
{
  SH2addi(sh, 8, (6<<4)+2);
}
#define init_7_8_6_3
void FASTCALL call_7_8_6_3(SH2_struct* sh)
{
  SH2addi(sh, 8, (6<<4)+3);
}
#define init_7_8_6_4
void FASTCALL call_7_8_6_4(SH2_struct* sh)
{
  SH2addi(sh, 8, (6<<4)+4);
}
#define init_7_8_6_5
void FASTCALL call_7_8_6_5(SH2_struct* sh)
{
  SH2addi(sh, 8, (6<<4)+5);
}
#define init_7_8_6_6
void FASTCALL call_7_8_6_6(SH2_struct* sh)
{
  SH2addi(sh, 8, (6<<4)+6);
}
#define init_7_8_6_7
void FASTCALL call_7_8_6_7(SH2_struct* sh)
{
  SH2addi(sh, 8, (6<<4)+7);
}
#define init_7_8_6_8
void FASTCALL call_7_8_6_8(SH2_struct* sh)
{
  SH2addi(sh, 8, (6<<4)+8);
}
#define init_7_8_6_9
void FASTCALL call_7_8_6_9(SH2_struct* sh)
{
  SH2addi(sh, 8, (6<<4)+9);
}
#define init_7_8_6_10
void FASTCALL call_7_8_6_10(SH2_struct* sh)
{
  SH2addi(sh, 8, (6<<4)+10);
}
#define init_7_8_6_11
void FASTCALL call_7_8_6_11(SH2_struct* sh)
{
  SH2addi(sh, 8, (6<<4)+11);
}
#define init_7_8_6_12
void FASTCALL call_7_8_6_12(SH2_struct* sh)
{
  SH2addi(sh, 8, (6<<4)+12);
}
#define init_7_8_6_13
void FASTCALL call_7_8_6_13(SH2_struct* sh)
{
  SH2addi(sh, 8, (6<<4)+13);
}
#define init_7_8_6_14
void FASTCALL call_7_8_6_14(SH2_struct* sh)
{
  SH2addi(sh, 8, (6<<4)+14);
}
#define init_7_8_6_15
void FASTCALL call_7_8_6_15(SH2_struct* sh)
{
  SH2addi(sh, 8, (6<<4)+15);
}
#define init_7_8_7_0
void FASTCALL call_7_8_7_0(SH2_struct* sh)
{
  SH2addi(sh, 8, (7<<4)+0);
}
#define init_7_8_7_1
void FASTCALL call_7_8_7_1(SH2_struct* sh)
{
  SH2addi(sh, 8, (7<<4)+1);
}
#define init_7_8_7_2
void FASTCALL call_7_8_7_2(SH2_struct* sh)
{
  SH2addi(sh, 8, (7<<4)+2);
}
#define init_7_8_7_3
void FASTCALL call_7_8_7_3(SH2_struct* sh)
{
  SH2addi(sh, 8, (7<<4)+3);
}
#define init_7_8_7_4
void FASTCALL call_7_8_7_4(SH2_struct* sh)
{
  SH2addi(sh, 8, (7<<4)+4);
}
#define init_7_8_7_5
void FASTCALL call_7_8_7_5(SH2_struct* sh)
{
  SH2addi(sh, 8, (7<<4)+5);
}
#define init_7_8_7_6
void FASTCALL call_7_8_7_6(SH2_struct* sh)
{
  SH2addi(sh, 8, (7<<4)+6);
}
#define init_7_8_7_7
void FASTCALL call_7_8_7_7(SH2_struct* sh)
{
  SH2addi(sh, 8, (7<<4)+7);
}
#define init_7_8_7_8
void FASTCALL call_7_8_7_8(SH2_struct* sh)
{
  SH2addi(sh, 8, (7<<4)+8);
}
#define init_7_8_7_9
void FASTCALL call_7_8_7_9(SH2_struct* sh)
{
  SH2addi(sh, 8, (7<<4)+9);
}
#define init_7_8_7_10
void FASTCALL call_7_8_7_10(SH2_struct* sh)
{
  SH2addi(sh, 8, (7<<4)+10);
}
#define init_7_8_7_11
void FASTCALL call_7_8_7_11(SH2_struct* sh)
{
  SH2addi(sh, 8, (7<<4)+11);
}
#define init_7_8_7_12
void FASTCALL call_7_8_7_12(SH2_struct* sh)
{
  SH2addi(sh, 8, (7<<4)+12);
}
#define init_7_8_7_13
void FASTCALL call_7_8_7_13(SH2_struct* sh)
{
  SH2addi(sh, 8, (7<<4)+13);
}
#define init_7_8_7_14
void FASTCALL call_7_8_7_14(SH2_struct* sh)
{
  SH2addi(sh, 8, (7<<4)+14);
}
#define init_7_8_7_15
void FASTCALL call_7_8_7_15(SH2_struct* sh)
{
  SH2addi(sh, 8, (7<<4)+15);
}
#define init_7_8_8_0
void FASTCALL call_7_8_8_0(SH2_struct* sh)
{
  SH2addi(sh, 8, (8<<4)+0);
}
#define init_7_8_8_1
void FASTCALL call_7_8_8_1(SH2_struct* sh)
{
  SH2addi(sh, 8, (8<<4)+1);
}
#define init_7_8_8_2
void FASTCALL call_7_8_8_2(SH2_struct* sh)
{
  SH2addi(sh, 8, (8<<4)+2);
}
#define init_7_8_8_3
void FASTCALL call_7_8_8_3(SH2_struct* sh)
{
  SH2addi(sh, 8, (8<<4)+3);
}
#define init_7_8_8_4
void FASTCALL call_7_8_8_4(SH2_struct* sh)
{
  SH2addi(sh, 8, (8<<4)+4);
}
#define init_7_8_8_5
void FASTCALL call_7_8_8_5(SH2_struct* sh)
{
  SH2addi(sh, 8, (8<<4)+5);
}
#define init_7_8_8_6
void FASTCALL call_7_8_8_6(SH2_struct* sh)
{
  SH2addi(sh, 8, (8<<4)+6);
}
#define init_7_8_8_7
void FASTCALL call_7_8_8_7(SH2_struct* sh)
{
  SH2addi(sh, 8, (8<<4)+7);
}
#define init_7_8_8_8
void FASTCALL call_7_8_8_8(SH2_struct* sh)
{
  SH2addi(sh, 8, (8<<4)+8);
}
#define init_7_8_8_9
void FASTCALL call_7_8_8_9(SH2_struct* sh)
{
  SH2addi(sh, 8, (8<<4)+9);
}
#define init_7_8_8_10
void FASTCALL call_7_8_8_10(SH2_struct* sh)
{
  SH2addi(sh, 8, (8<<4)+10);
}
#define init_7_8_8_11
void FASTCALL call_7_8_8_11(SH2_struct* sh)
{
  SH2addi(sh, 8, (8<<4)+11);
}
#define init_7_8_8_12
void FASTCALL call_7_8_8_12(SH2_struct* sh)
{
  SH2addi(sh, 8, (8<<4)+12);
}
#define init_7_8_8_13
void FASTCALL call_7_8_8_13(SH2_struct* sh)
{
  SH2addi(sh, 8, (8<<4)+13);
}
#define init_7_8_8_14
void FASTCALL call_7_8_8_14(SH2_struct* sh)
{
  SH2addi(sh, 8, (8<<4)+14);
}
#define init_7_8_8_15
void FASTCALL call_7_8_8_15(SH2_struct* sh)
{
  SH2addi(sh, 8, (8<<4)+15);
}
#define init_7_8_9_0
void FASTCALL call_7_8_9_0(SH2_struct* sh)
{
  SH2addi(sh, 8, (9<<4)+0);
}
#define init_7_8_9_1
void FASTCALL call_7_8_9_1(SH2_struct* sh)
{
  SH2addi(sh, 8, (9<<4)+1);
}
#define init_7_8_9_2
void FASTCALL call_7_8_9_2(SH2_struct* sh)
{
  SH2addi(sh, 8, (9<<4)+2);
}
#define init_7_8_9_3
void FASTCALL call_7_8_9_3(SH2_struct* sh)
{
  SH2addi(sh, 8, (9<<4)+3);
}
#define init_7_8_9_4
void FASTCALL call_7_8_9_4(SH2_struct* sh)
{
  SH2addi(sh, 8, (9<<4)+4);
}
#define init_7_8_9_5
void FASTCALL call_7_8_9_5(SH2_struct* sh)
{
  SH2addi(sh, 8, (9<<4)+5);
}
#define init_7_8_9_6
void FASTCALL call_7_8_9_6(SH2_struct* sh)
{
  SH2addi(sh, 8, (9<<4)+6);
}
#define init_7_8_9_7
void FASTCALL call_7_8_9_7(SH2_struct* sh)
{
  SH2addi(sh, 8, (9<<4)+7);
}
#define init_7_8_9_8
void FASTCALL call_7_8_9_8(SH2_struct* sh)
{
  SH2addi(sh, 8, (9<<4)+8);
}
#define init_7_8_9_9
void FASTCALL call_7_8_9_9(SH2_struct* sh)
{
  SH2addi(sh, 8, (9<<4)+9);
}
#define init_7_8_9_10
void FASTCALL call_7_8_9_10(SH2_struct* sh)
{
  SH2addi(sh, 8, (9<<4)+10);
}
#define init_7_8_9_11
void FASTCALL call_7_8_9_11(SH2_struct* sh)
{
  SH2addi(sh, 8, (9<<4)+11);
}
#define init_7_8_9_12
void FASTCALL call_7_8_9_12(SH2_struct* sh)
{
  SH2addi(sh, 8, (9<<4)+12);
}
#define init_7_8_9_13
void FASTCALL call_7_8_9_13(SH2_struct* sh)
{
  SH2addi(sh, 8, (9<<4)+13);
}
#define init_7_8_9_14
void FASTCALL call_7_8_9_14(SH2_struct* sh)
{
  SH2addi(sh, 8, (9<<4)+14);
}
#define init_7_8_9_15
void FASTCALL call_7_8_9_15(SH2_struct* sh)
{
  SH2addi(sh, 8, (9<<4)+15);
}
#define init_7_8_10_0
void FASTCALL call_7_8_10_0(SH2_struct* sh)
{
  SH2addi(sh, 8, (10<<4)+0);
}
#define init_7_8_10_1
void FASTCALL call_7_8_10_1(SH2_struct* sh)
{
  SH2addi(sh, 8, (10<<4)+1);
}
#define init_7_8_10_2
void FASTCALL call_7_8_10_2(SH2_struct* sh)
{
  SH2addi(sh, 8, (10<<4)+2);
}
#define init_7_8_10_3
void FASTCALL call_7_8_10_3(SH2_struct* sh)
{
  SH2addi(sh, 8, (10<<4)+3);
}
#define init_7_8_10_4
void FASTCALL call_7_8_10_4(SH2_struct* sh)
{
  SH2addi(sh, 8, (10<<4)+4);
}
#define init_7_8_10_5
void FASTCALL call_7_8_10_5(SH2_struct* sh)
{
  SH2addi(sh, 8, (10<<4)+5);
}
#define init_7_8_10_6
void FASTCALL call_7_8_10_6(SH2_struct* sh)
{
  SH2addi(sh, 8, (10<<4)+6);
}
#define init_7_8_10_7
void FASTCALL call_7_8_10_7(SH2_struct* sh)
{
  SH2addi(sh, 8, (10<<4)+7);
}
#define init_7_8_10_8
void FASTCALL call_7_8_10_8(SH2_struct* sh)
{
  SH2addi(sh, 8, (10<<4)+8);
}
#define init_7_8_10_9
void FASTCALL call_7_8_10_9(SH2_struct* sh)
{
  SH2addi(sh, 8, (10<<4)+9);
}
#define init_7_8_10_10
void FASTCALL call_7_8_10_10(SH2_struct* sh)
{
  SH2addi(sh, 8, (10<<4)+10);
}
#define init_7_8_10_11
void FASTCALL call_7_8_10_11(SH2_struct* sh)
{
  SH2addi(sh, 8, (10<<4)+11);
}
#define init_7_8_10_12
void FASTCALL call_7_8_10_12(SH2_struct* sh)
{
  SH2addi(sh, 8, (10<<4)+12);
}
#define init_7_8_10_13
void FASTCALL call_7_8_10_13(SH2_struct* sh)
{
  SH2addi(sh, 8, (10<<4)+13);
}
#define init_7_8_10_14
void FASTCALL call_7_8_10_14(SH2_struct* sh)
{
  SH2addi(sh, 8, (10<<4)+14);
}
#define init_7_8_10_15
void FASTCALL call_7_8_10_15(SH2_struct* sh)
{
  SH2addi(sh, 8, (10<<4)+15);
}
#define init_7_8_11_0
void FASTCALL call_7_8_11_0(SH2_struct* sh)
{
  SH2addi(sh, 8, (11<<4)+0);
}
#define init_7_8_11_1
void FASTCALL call_7_8_11_1(SH2_struct* sh)
{
  SH2addi(sh, 8, (11<<4)+1);
}
#define init_7_8_11_2
void FASTCALL call_7_8_11_2(SH2_struct* sh)
{
  SH2addi(sh, 8, (11<<4)+2);
}
#define init_7_8_11_3
void FASTCALL call_7_8_11_3(SH2_struct* sh)
{
  SH2addi(sh, 8, (11<<4)+3);
}
#define init_7_8_11_4
void FASTCALL call_7_8_11_4(SH2_struct* sh)
{
  SH2addi(sh, 8, (11<<4)+4);
}
#define init_7_8_11_5
void FASTCALL call_7_8_11_5(SH2_struct* sh)
{
  SH2addi(sh, 8, (11<<4)+5);
}
#define init_7_8_11_6
void FASTCALL call_7_8_11_6(SH2_struct* sh)
{
  SH2addi(sh, 8, (11<<4)+6);
}
#define init_7_8_11_7
void FASTCALL call_7_8_11_7(SH2_struct* sh)
{
  SH2addi(sh, 8, (11<<4)+7);
}
#define init_7_8_11_8
void FASTCALL call_7_8_11_8(SH2_struct* sh)
{
  SH2addi(sh, 8, (11<<4)+8);
}
#define init_7_8_11_9
void FASTCALL call_7_8_11_9(SH2_struct* sh)
{
  SH2addi(sh, 8, (11<<4)+9);
}
#define init_7_8_11_10
void FASTCALL call_7_8_11_10(SH2_struct* sh)
{
  SH2addi(sh, 8, (11<<4)+10);
}
#define init_7_8_11_11
void FASTCALL call_7_8_11_11(SH2_struct* sh)
{
  SH2addi(sh, 8, (11<<4)+11);
}
#define init_7_8_11_12
void FASTCALL call_7_8_11_12(SH2_struct* sh)
{
  SH2addi(sh, 8, (11<<4)+12);
}
#define init_7_8_11_13
void FASTCALL call_7_8_11_13(SH2_struct* sh)
{
  SH2addi(sh, 8, (11<<4)+13);
}
#define init_7_8_11_14
void FASTCALL call_7_8_11_14(SH2_struct* sh)
{
  SH2addi(sh, 8, (11<<4)+14);
}
#define init_7_8_11_15
void FASTCALL call_7_8_11_15(SH2_struct* sh)
{
  SH2addi(sh, 8, (11<<4)+15);
}
#define init_7_8_12_0
void FASTCALL call_7_8_12_0(SH2_struct* sh)
{
  SH2addi(sh, 8, (12<<4)+0);
}
#define init_7_8_12_1
void FASTCALL call_7_8_12_1(SH2_struct* sh)
{
  SH2addi(sh, 8, (12<<4)+1);
}
#define init_7_8_12_2
void FASTCALL call_7_8_12_2(SH2_struct* sh)
{
  SH2addi(sh, 8, (12<<4)+2);
}
#define init_7_8_12_3
void FASTCALL call_7_8_12_3(SH2_struct* sh)
{
  SH2addi(sh, 8, (12<<4)+3);
}
#define init_7_8_12_4
void FASTCALL call_7_8_12_4(SH2_struct* sh)
{
  SH2addi(sh, 8, (12<<4)+4);
}
#define init_7_8_12_5
void FASTCALL call_7_8_12_5(SH2_struct* sh)
{
  SH2addi(sh, 8, (12<<4)+5);
}
#define init_7_8_12_6
void FASTCALL call_7_8_12_6(SH2_struct* sh)
{
  SH2addi(sh, 8, (12<<4)+6);
}
#define init_7_8_12_7
void FASTCALL call_7_8_12_7(SH2_struct* sh)
{
  SH2addi(sh, 8, (12<<4)+7);
}
#define init_7_8_12_8
void FASTCALL call_7_8_12_8(SH2_struct* sh)
{
  SH2addi(sh, 8, (12<<4)+8);
}
#define init_7_8_12_9
void FASTCALL call_7_8_12_9(SH2_struct* sh)
{
  SH2addi(sh, 8, (12<<4)+9);
}
#define init_7_8_12_10
void FASTCALL call_7_8_12_10(SH2_struct* sh)
{
  SH2addi(sh, 8, (12<<4)+10);
}
#define init_7_8_12_11
void FASTCALL call_7_8_12_11(SH2_struct* sh)
{
  SH2addi(sh, 8, (12<<4)+11);
}
#define init_7_8_12_12
void FASTCALL call_7_8_12_12(SH2_struct* sh)
{
  SH2addi(sh, 8, (12<<4)+12);
}
#define init_7_8_12_13
void FASTCALL call_7_8_12_13(SH2_struct* sh)
{
  SH2addi(sh, 8, (12<<4)+13);
}
#define init_7_8_12_14
void FASTCALL call_7_8_12_14(SH2_struct* sh)
{
  SH2addi(sh, 8, (12<<4)+14);
}
#define init_7_8_12_15
void FASTCALL call_7_8_12_15(SH2_struct* sh)
{
  SH2addi(sh, 8, (12<<4)+15);
}
#define init_7_8_13_0
void FASTCALL call_7_8_13_0(SH2_struct* sh)
{
  SH2addi(sh, 8, (13<<4)+0);
}
#define init_7_8_13_1
void FASTCALL call_7_8_13_1(SH2_struct* sh)
{
  SH2addi(sh, 8, (13<<4)+1);
}
#define init_7_8_13_2
void FASTCALL call_7_8_13_2(SH2_struct* sh)
{
  SH2addi(sh, 8, (13<<4)+2);
}
#define init_7_8_13_3
void FASTCALL call_7_8_13_3(SH2_struct* sh)
{
  SH2addi(sh, 8, (13<<4)+3);
}
#define init_7_8_13_4
void FASTCALL call_7_8_13_4(SH2_struct* sh)
{
  SH2addi(sh, 8, (13<<4)+4);
}
#define init_7_8_13_5
void FASTCALL call_7_8_13_5(SH2_struct* sh)
{
  SH2addi(sh, 8, (13<<4)+5);
}
#define init_7_8_13_6
void FASTCALL call_7_8_13_6(SH2_struct* sh)
{
  SH2addi(sh, 8, (13<<4)+6);
}
#define init_7_8_13_7
void FASTCALL call_7_8_13_7(SH2_struct* sh)
{
  SH2addi(sh, 8, (13<<4)+7);
}
#define init_7_8_13_8
void FASTCALL call_7_8_13_8(SH2_struct* sh)
{
  SH2addi(sh, 8, (13<<4)+8);
}
#define init_7_8_13_9
void FASTCALL call_7_8_13_9(SH2_struct* sh)
{
  SH2addi(sh, 8, (13<<4)+9);
}
#define init_7_8_13_10
void FASTCALL call_7_8_13_10(SH2_struct* sh)
{
  SH2addi(sh, 8, (13<<4)+10);
}
#define init_7_8_13_11
void FASTCALL call_7_8_13_11(SH2_struct* sh)
{
  SH2addi(sh, 8, (13<<4)+11);
}
#define init_7_8_13_12
void FASTCALL call_7_8_13_12(SH2_struct* sh)
{
  SH2addi(sh, 8, (13<<4)+12);
}
#define init_7_8_13_13
void FASTCALL call_7_8_13_13(SH2_struct* sh)
{
  SH2addi(sh, 8, (13<<4)+13);
}
#define init_7_8_13_14
void FASTCALL call_7_8_13_14(SH2_struct* sh)
{
  SH2addi(sh, 8, (13<<4)+14);
}
#define init_7_8_13_15
void FASTCALL call_7_8_13_15(SH2_struct* sh)
{
  SH2addi(sh, 8, (13<<4)+15);
}
#define init_7_8_14_0
void FASTCALL call_7_8_14_0(SH2_struct* sh)
{
  SH2addi(sh, 8, (14<<4)+0);
}
#define init_7_8_14_1
void FASTCALL call_7_8_14_1(SH2_struct* sh)
{
  SH2addi(sh, 8, (14<<4)+1);
}
#define init_7_8_14_2
void FASTCALL call_7_8_14_2(SH2_struct* sh)
{
  SH2addi(sh, 8, (14<<4)+2);
}
#define init_7_8_14_3
void FASTCALL call_7_8_14_3(SH2_struct* sh)
{
  SH2addi(sh, 8, (14<<4)+3);
}
#define init_7_8_14_4
void FASTCALL call_7_8_14_4(SH2_struct* sh)
{
  SH2addi(sh, 8, (14<<4)+4);
}
#define init_7_8_14_5
void FASTCALL call_7_8_14_5(SH2_struct* sh)
{
  SH2addi(sh, 8, (14<<4)+5);
}
#define init_7_8_14_6
void FASTCALL call_7_8_14_6(SH2_struct* sh)
{
  SH2addi(sh, 8, (14<<4)+6);
}
#define init_7_8_14_7
void FASTCALL call_7_8_14_7(SH2_struct* sh)
{
  SH2addi(sh, 8, (14<<4)+7);
}
#define init_7_8_14_8
void FASTCALL call_7_8_14_8(SH2_struct* sh)
{
  SH2addi(sh, 8, (14<<4)+8);
}
#define init_7_8_14_9
void FASTCALL call_7_8_14_9(SH2_struct* sh)
{
  SH2addi(sh, 8, (14<<4)+9);
}
#define init_7_8_14_10
void FASTCALL call_7_8_14_10(SH2_struct* sh)
{
  SH2addi(sh, 8, (14<<4)+10);
}
#define init_7_8_14_11
void FASTCALL call_7_8_14_11(SH2_struct* sh)
{
  SH2addi(sh, 8, (14<<4)+11);
}
#define init_7_8_14_12
void FASTCALL call_7_8_14_12(SH2_struct* sh)
{
  SH2addi(sh, 8, (14<<4)+12);
}
#define init_7_8_14_13
void FASTCALL call_7_8_14_13(SH2_struct* sh)
{
  SH2addi(sh, 8, (14<<4)+13);
}
#define init_7_8_14_14
void FASTCALL call_7_8_14_14(SH2_struct* sh)
{
  SH2addi(sh, 8, (14<<4)+14);
}
#define init_7_8_14_15
void FASTCALL call_7_8_14_15(SH2_struct* sh)
{
  SH2addi(sh, 8, (14<<4)+15);
}
#define init_7_8_15_0
void FASTCALL call_7_8_15_0(SH2_struct* sh)
{
  SH2addi(sh, 8, (15<<4)+0);
}
#define init_7_8_15_1
void FASTCALL call_7_8_15_1(SH2_struct* sh)
{
  SH2addi(sh, 8, (15<<4)+1);
}
#define init_7_8_15_2
void FASTCALL call_7_8_15_2(SH2_struct* sh)
{
  SH2addi(sh, 8, (15<<4)+2);
}
#define init_7_8_15_3
void FASTCALL call_7_8_15_3(SH2_struct* sh)
{
  SH2addi(sh, 8, (15<<4)+3);
}
#define init_7_8_15_4
void FASTCALL call_7_8_15_4(SH2_struct* sh)
{
  SH2addi(sh, 8, (15<<4)+4);
}
#define init_7_8_15_5
void FASTCALL call_7_8_15_5(SH2_struct* sh)
{
  SH2addi(sh, 8, (15<<4)+5);
}
#define init_7_8_15_6
void FASTCALL call_7_8_15_6(SH2_struct* sh)
{
  SH2addi(sh, 8, (15<<4)+6);
}
#define init_7_8_15_7
void FASTCALL call_7_8_15_7(SH2_struct* sh)
{
  SH2addi(sh, 8, (15<<4)+7);
}
#define init_7_8_15_8
void FASTCALL call_7_8_15_8(SH2_struct* sh)
{
  SH2addi(sh, 8, (15<<4)+8);
}
#define init_7_8_15_9
void FASTCALL call_7_8_15_9(SH2_struct* sh)
{
  SH2addi(sh, 8, (15<<4)+9);
}
#define init_7_8_15_10
void FASTCALL call_7_8_15_10(SH2_struct* sh)
{
  SH2addi(sh, 8, (15<<4)+10);
}
#define init_7_8_15_11
void FASTCALL call_7_8_15_11(SH2_struct* sh)
{
  SH2addi(sh, 8, (15<<4)+11);
}
#define init_7_8_15_12
void FASTCALL call_7_8_15_12(SH2_struct* sh)
{
  SH2addi(sh, 8, (15<<4)+12);
}
#define init_7_8_15_13
void FASTCALL call_7_8_15_13(SH2_struct* sh)
{
  SH2addi(sh, 8, (15<<4)+13);
}
#define init_7_8_15_14
void FASTCALL call_7_8_15_14(SH2_struct* sh)
{
  SH2addi(sh, 8, (15<<4)+14);
}
#define init_7_8_15_15
void FASTCALL call_7_8_15_15(SH2_struct* sh)
{
  SH2addi(sh, 8, (15<<4)+15);
}
#define init_7_9_0_0
void FASTCALL call_7_9_0_0(SH2_struct* sh)
{
  SH2addi(sh, 9, (0<<4)+0);
}
#define init_7_9_0_1
void FASTCALL call_7_9_0_1(SH2_struct* sh)
{
  SH2addi(sh, 9, (0<<4)+1);
}
#define init_7_9_0_2
void FASTCALL call_7_9_0_2(SH2_struct* sh)
{
  SH2addi(sh, 9, (0<<4)+2);
}
#define init_7_9_0_3
void FASTCALL call_7_9_0_3(SH2_struct* sh)
{
  SH2addi(sh, 9, (0<<4)+3);
}
#define init_7_9_0_4
void FASTCALL call_7_9_0_4(SH2_struct* sh)
{
  SH2addi(sh, 9, (0<<4)+4);
}
#define init_7_9_0_5
void FASTCALL call_7_9_0_5(SH2_struct* sh)
{
  SH2addi(sh, 9, (0<<4)+5);
}
#define init_7_9_0_6
void FASTCALL call_7_9_0_6(SH2_struct* sh)
{
  SH2addi(sh, 9, (0<<4)+6);
}
#define init_7_9_0_7
void FASTCALL call_7_9_0_7(SH2_struct* sh)
{
  SH2addi(sh, 9, (0<<4)+7);
}
#define init_7_9_0_8
void FASTCALL call_7_9_0_8(SH2_struct* sh)
{
  SH2addi(sh, 9, (0<<4)+8);
}
#define init_7_9_0_9
void FASTCALL call_7_9_0_9(SH2_struct* sh)
{
  SH2addi(sh, 9, (0<<4)+9);
}
#define init_7_9_0_10
void FASTCALL call_7_9_0_10(SH2_struct* sh)
{
  SH2addi(sh, 9, (0<<4)+10);
}
#define init_7_9_0_11
void FASTCALL call_7_9_0_11(SH2_struct* sh)
{
  SH2addi(sh, 9, (0<<4)+11);
}
#define init_7_9_0_12
void FASTCALL call_7_9_0_12(SH2_struct* sh)
{
  SH2addi(sh, 9, (0<<4)+12);
}
#define init_7_9_0_13
void FASTCALL call_7_9_0_13(SH2_struct* sh)
{
  SH2addi(sh, 9, (0<<4)+13);
}
#define init_7_9_0_14
void FASTCALL call_7_9_0_14(SH2_struct* sh)
{
  SH2addi(sh, 9, (0<<4)+14);
}
#define init_7_9_0_15
void FASTCALL call_7_9_0_15(SH2_struct* sh)
{
  SH2addi(sh, 9, (0<<4)+15);
}
#define init_7_9_1_0
void FASTCALL call_7_9_1_0(SH2_struct* sh)
{
  SH2addi(sh, 9, (1<<4)+0);
}
#define init_7_9_1_1
void FASTCALL call_7_9_1_1(SH2_struct* sh)
{
  SH2addi(sh, 9, (1<<4)+1);
}
#define init_7_9_1_2
void FASTCALL call_7_9_1_2(SH2_struct* sh)
{
  SH2addi(sh, 9, (1<<4)+2);
}
#define init_7_9_1_3
void FASTCALL call_7_9_1_3(SH2_struct* sh)
{
  SH2addi(sh, 9, (1<<4)+3);
}
#define init_7_9_1_4
void FASTCALL call_7_9_1_4(SH2_struct* sh)
{
  SH2addi(sh, 9, (1<<4)+4);
}
#define init_7_9_1_5
void FASTCALL call_7_9_1_5(SH2_struct* sh)
{
  SH2addi(sh, 9, (1<<4)+5);
}
#define init_7_9_1_6
void FASTCALL call_7_9_1_6(SH2_struct* sh)
{
  SH2addi(sh, 9, (1<<4)+6);
}
#define init_7_9_1_7
void FASTCALL call_7_9_1_7(SH2_struct* sh)
{
  SH2addi(sh, 9, (1<<4)+7);
}
#define init_7_9_1_8
void FASTCALL call_7_9_1_8(SH2_struct* sh)
{
  SH2addi(sh, 9, (1<<4)+8);
}
#define init_7_9_1_9
void FASTCALL call_7_9_1_9(SH2_struct* sh)
{
  SH2addi(sh, 9, (1<<4)+9);
}
#define init_7_9_1_10
void FASTCALL call_7_9_1_10(SH2_struct* sh)
{
  SH2addi(sh, 9, (1<<4)+10);
}
#define init_7_9_1_11
void FASTCALL call_7_9_1_11(SH2_struct* sh)
{
  SH2addi(sh, 9, (1<<4)+11);
}
#define init_7_9_1_12
void FASTCALL call_7_9_1_12(SH2_struct* sh)
{
  SH2addi(sh, 9, (1<<4)+12);
}
#define init_7_9_1_13
void FASTCALL call_7_9_1_13(SH2_struct* sh)
{
  SH2addi(sh, 9, (1<<4)+13);
}
#define init_7_9_1_14
void FASTCALL call_7_9_1_14(SH2_struct* sh)
{
  SH2addi(sh, 9, (1<<4)+14);
}
#define init_7_9_1_15
void FASTCALL call_7_9_1_15(SH2_struct* sh)
{
  SH2addi(sh, 9, (1<<4)+15);
}
#define init_7_9_2_0
void FASTCALL call_7_9_2_0(SH2_struct* sh)
{
  SH2addi(sh, 9, (2<<4)+0);
}
#define init_7_9_2_1
void FASTCALL call_7_9_2_1(SH2_struct* sh)
{
  SH2addi(sh, 9, (2<<4)+1);
}
#define init_7_9_2_2
void FASTCALL call_7_9_2_2(SH2_struct* sh)
{
  SH2addi(sh, 9, (2<<4)+2);
}
#define init_7_9_2_3
void FASTCALL call_7_9_2_3(SH2_struct* sh)
{
  SH2addi(sh, 9, (2<<4)+3);
}
#define init_7_9_2_4
void FASTCALL call_7_9_2_4(SH2_struct* sh)
{
  SH2addi(sh, 9, (2<<4)+4);
}
#define init_7_9_2_5
void FASTCALL call_7_9_2_5(SH2_struct* sh)
{
  SH2addi(sh, 9, (2<<4)+5);
}
#define init_7_9_2_6
void FASTCALL call_7_9_2_6(SH2_struct* sh)
{
  SH2addi(sh, 9, (2<<4)+6);
}
#define init_7_9_2_7
void FASTCALL call_7_9_2_7(SH2_struct* sh)
{
  SH2addi(sh, 9, (2<<4)+7);
}
#define init_7_9_2_8
void FASTCALL call_7_9_2_8(SH2_struct* sh)
{
  SH2addi(sh, 9, (2<<4)+8);
}
#define init_7_9_2_9
void FASTCALL call_7_9_2_9(SH2_struct* sh)
{
  SH2addi(sh, 9, (2<<4)+9);
}
#define init_7_9_2_10
void FASTCALL call_7_9_2_10(SH2_struct* sh)
{
  SH2addi(sh, 9, (2<<4)+10);
}
#define init_7_9_2_11
void FASTCALL call_7_9_2_11(SH2_struct* sh)
{
  SH2addi(sh, 9, (2<<4)+11);
}
#define init_7_9_2_12
void FASTCALL call_7_9_2_12(SH2_struct* sh)
{
  SH2addi(sh, 9, (2<<4)+12);
}
#define init_7_9_2_13
void FASTCALL call_7_9_2_13(SH2_struct* sh)
{
  SH2addi(sh, 9, (2<<4)+13);
}
#define init_7_9_2_14
void FASTCALL call_7_9_2_14(SH2_struct* sh)
{
  SH2addi(sh, 9, (2<<4)+14);
}
#define init_7_9_2_15
void FASTCALL call_7_9_2_15(SH2_struct* sh)
{
  SH2addi(sh, 9, (2<<4)+15);
}
#define init_7_9_3_0
void FASTCALL call_7_9_3_0(SH2_struct* sh)
{
  SH2addi(sh, 9, (3<<4)+0);
}
#define init_7_9_3_1
void FASTCALL call_7_9_3_1(SH2_struct* sh)
{
  SH2addi(sh, 9, (3<<4)+1);
}
#define init_7_9_3_2
void FASTCALL call_7_9_3_2(SH2_struct* sh)
{
  SH2addi(sh, 9, (3<<4)+2);
}
#define init_7_9_3_3
void FASTCALL call_7_9_3_3(SH2_struct* sh)
{
  SH2addi(sh, 9, (3<<4)+3);
}
#define init_7_9_3_4
void FASTCALL call_7_9_3_4(SH2_struct* sh)
{
  SH2addi(sh, 9, (3<<4)+4);
}
#define init_7_9_3_5
void FASTCALL call_7_9_3_5(SH2_struct* sh)
{
  SH2addi(sh, 9, (3<<4)+5);
}
#define init_7_9_3_6
void FASTCALL call_7_9_3_6(SH2_struct* sh)
{
  SH2addi(sh, 9, (3<<4)+6);
}
#define init_7_9_3_7
void FASTCALL call_7_9_3_7(SH2_struct* sh)
{
  SH2addi(sh, 9, (3<<4)+7);
}
#define init_7_9_3_8
void FASTCALL call_7_9_3_8(SH2_struct* sh)
{
  SH2addi(sh, 9, (3<<4)+8);
}
#define init_7_9_3_9
void FASTCALL call_7_9_3_9(SH2_struct* sh)
{
  SH2addi(sh, 9, (3<<4)+9);
}
#define init_7_9_3_10
void FASTCALL call_7_9_3_10(SH2_struct* sh)
{
  SH2addi(sh, 9, (3<<4)+10);
}
#define init_7_9_3_11
void FASTCALL call_7_9_3_11(SH2_struct* sh)
{
  SH2addi(sh, 9, (3<<4)+11);
}
#define init_7_9_3_12
void FASTCALL call_7_9_3_12(SH2_struct* sh)
{
  SH2addi(sh, 9, (3<<4)+12);
}
#define init_7_9_3_13
void FASTCALL call_7_9_3_13(SH2_struct* sh)
{
  SH2addi(sh, 9, (3<<4)+13);
}
#define init_7_9_3_14
void FASTCALL call_7_9_3_14(SH2_struct* sh)
{
  SH2addi(sh, 9, (3<<4)+14);
}
#define init_7_9_3_15
void FASTCALL call_7_9_3_15(SH2_struct* sh)
{
  SH2addi(sh, 9, (3<<4)+15);
}
#define init_7_9_4_0
void FASTCALL call_7_9_4_0(SH2_struct* sh)
{
  SH2addi(sh, 9, (4<<4)+0);
}
#define init_7_9_4_1
void FASTCALL call_7_9_4_1(SH2_struct* sh)
{
  SH2addi(sh, 9, (4<<4)+1);
}
#define init_7_9_4_2
void FASTCALL call_7_9_4_2(SH2_struct* sh)
{
  SH2addi(sh, 9, (4<<4)+2);
}
#define init_7_9_4_3
void FASTCALL call_7_9_4_3(SH2_struct* sh)
{
  SH2addi(sh, 9, (4<<4)+3);
}
#define init_7_9_4_4
void FASTCALL call_7_9_4_4(SH2_struct* sh)
{
  SH2addi(sh, 9, (4<<4)+4);
}
#define init_7_9_4_5
void FASTCALL call_7_9_4_5(SH2_struct* sh)
{
  SH2addi(sh, 9, (4<<4)+5);
}
#define init_7_9_4_6
void FASTCALL call_7_9_4_6(SH2_struct* sh)
{
  SH2addi(sh, 9, (4<<4)+6);
}
#define init_7_9_4_7
void FASTCALL call_7_9_4_7(SH2_struct* sh)
{
  SH2addi(sh, 9, (4<<4)+7);
}
#define init_7_9_4_8
void FASTCALL call_7_9_4_8(SH2_struct* sh)
{
  SH2addi(sh, 9, (4<<4)+8);
}
#define init_7_9_4_9
void FASTCALL call_7_9_4_9(SH2_struct* sh)
{
  SH2addi(sh, 9, (4<<4)+9);
}
#define init_7_9_4_10
void FASTCALL call_7_9_4_10(SH2_struct* sh)
{
  SH2addi(sh, 9, (4<<4)+10);
}
#define init_7_9_4_11
void FASTCALL call_7_9_4_11(SH2_struct* sh)
{
  SH2addi(sh, 9, (4<<4)+11);
}
#define init_7_9_4_12
void FASTCALL call_7_9_4_12(SH2_struct* sh)
{
  SH2addi(sh, 9, (4<<4)+12);
}
#define init_7_9_4_13
void FASTCALL call_7_9_4_13(SH2_struct* sh)
{
  SH2addi(sh, 9, (4<<4)+13);
}
#define init_7_9_4_14
void FASTCALL call_7_9_4_14(SH2_struct* sh)
{
  SH2addi(sh, 9, (4<<4)+14);
}
#define init_7_9_4_15
void FASTCALL call_7_9_4_15(SH2_struct* sh)
{
  SH2addi(sh, 9, (4<<4)+15);
}
#define init_7_9_5_0
void FASTCALL call_7_9_5_0(SH2_struct* sh)
{
  SH2addi(sh, 9, (5<<4)+0);
}
#define init_7_9_5_1
void FASTCALL call_7_9_5_1(SH2_struct* sh)
{
  SH2addi(sh, 9, (5<<4)+1);
}
#define init_7_9_5_2
void FASTCALL call_7_9_5_2(SH2_struct* sh)
{
  SH2addi(sh, 9, (5<<4)+2);
}
#define init_7_9_5_3
void FASTCALL call_7_9_5_3(SH2_struct* sh)
{
  SH2addi(sh, 9, (5<<4)+3);
}
#define init_7_9_5_4
void FASTCALL call_7_9_5_4(SH2_struct* sh)
{
  SH2addi(sh, 9, (5<<4)+4);
}
#define init_7_9_5_5
void FASTCALL call_7_9_5_5(SH2_struct* sh)
{
  SH2addi(sh, 9, (5<<4)+5);
}
#define init_7_9_5_6
void FASTCALL call_7_9_5_6(SH2_struct* sh)
{
  SH2addi(sh, 9, (5<<4)+6);
}
#define init_7_9_5_7
void FASTCALL call_7_9_5_7(SH2_struct* sh)
{
  SH2addi(sh, 9, (5<<4)+7);
}
#define init_7_9_5_8
void FASTCALL call_7_9_5_8(SH2_struct* sh)
{
  SH2addi(sh, 9, (5<<4)+8);
}
#define init_7_9_5_9
void FASTCALL call_7_9_5_9(SH2_struct* sh)
{
  SH2addi(sh, 9, (5<<4)+9);
}
#define init_7_9_5_10
void FASTCALL call_7_9_5_10(SH2_struct* sh)
{
  SH2addi(sh, 9, (5<<4)+10);
}
#define init_7_9_5_11
void FASTCALL call_7_9_5_11(SH2_struct* sh)
{
  SH2addi(sh, 9, (5<<4)+11);
}
#define init_7_9_5_12
void FASTCALL call_7_9_5_12(SH2_struct* sh)
{
  SH2addi(sh, 9, (5<<4)+12);
}
#define init_7_9_5_13
void FASTCALL call_7_9_5_13(SH2_struct* sh)
{
  SH2addi(sh, 9, (5<<4)+13);
}
#define init_7_9_5_14
void FASTCALL call_7_9_5_14(SH2_struct* sh)
{
  SH2addi(sh, 9, (5<<4)+14);
}
#define init_7_9_5_15
void FASTCALL call_7_9_5_15(SH2_struct* sh)
{
  SH2addi(sh, 9, (5<<4)+15);
}
#define init_7_9_6_0
void FASTCALL call_7_9_6_0(SH2_struct* sh)
{
  SH2addi(sh, 9, (6<<4)+0);
}
#define init_7_9_6_1
void FASTCALL call_7_9_6_1(SH2_struct* sh)
{
  SH2addi(sh, 9, (6<<4)+1);
}
#define init_7_9_6_2
void FASTCALL call_7_9_6_2(SH2_struct* sh)
{
  SH2addi(sh, 9, (6<<4)+2);
}
#define init_7_9_6_3
void FASTCALL call_7_9_6_3(SH2_struct* sh)
{
  SH2addi(sh, 9, (6<<4)+3);
}
#define init_7_9_6_4
void FASTCALL call_7_9_6_4(SH2_struct* sh)
{
  SH2addi(sh, 9, (6<<4)+4);
}
#define init_7_9_6_5
void FASTCALL call_7_9_6_5(SH2_struct* sh)
{
  SH2addi(sh, 9, (6<<4)+5);
}
#define init_7_9_6_6
void FASTCALL call_7_9_6_6(SH2_struct* sh)
{
  SH2addi(sh, 9, (6<<4)+6);
}
#define init_7_9_6_7
void FASTCALL call_7_9_6_7(SH2_struct* sh)
{
  SH2addi(sh, 9, (6<<4)+7);
}
#define init_7_9_6_8
void FASTCALL call_7_9_6_8(SH2_struct* sh)
{
  SH2addi(sh, 9, (6<<4)+8);
}
#define init_7_9_6_9
void FASTCALL call_7_9_6_9(SH2_struct* sh)
{
  SH2addi(sh, 9, (6<<4)+9);
}
#define init_7_9_6_10
void FASTCALL call_7_9_6_10(SH2_struct* sh)
{
  SH2addi(sh, 9, (6<<4)+10);
}
#define init_7_9_6_11
void FASTCALL call_7_9_6_11(SH2_struct* sh)
{
  SH2addi(sh, 9, (6<<4)+11);
}
#define init_7_9_6_12
void FASTCALL call_7_9_6_12(SH2_struct* sh)
{
  SH2addi(sh, 9, (6<<4)+12);
}
#define init_7_9_6_13
void FASTCALL call_7_9_6_13(SH2_struct* sh)
{
  SH2addi(sh, 9, (6<<4)+13);
}
#define init_7_9_6_14
void FASTCALL call_7_9_6_14(SH2_struct* sh)
{
  SH2addi(sh, 9, (6<<4)+14);
}
#define init_7_9_6_15
void FASTCALL call_7_9_6_15(SH2_struct* sh)
{
  SH2addi(sh, 9, (6<<4)+15);
}
#define init_7_9_7_0
void FASTCALL call_7_9_7_0(SH2_struct* sh)
{
  SH2addi(sh, 9, (7<<4)+0);
}
#define init_7_9_7_1
void FASTCALL call_7_9_7_1(SH2_struct* sh)
{
  SH2addi(sh, 9, (7<<4)+1);
}
#define init_7_9_7_2
void FASTCALL call_7_9_7_2(SH2_struct* sh)
{
  SH2addi(sh, 9, (7<<4)+2);
}
#define init_7_9_7_3
void FASTCALL call_7_9_7_3(SH2_struct* sh)
{
  SH2addi(sh, 9, (7<<4)+3);
}
#define init_7_9_7_4
void FASTCALL call_7_9_7_4(SH2_struct* sh)
{
  SH2addi(sh, 9, (7<<4)+4);
}
#define init_7_9_7_5
void FASTCALL call_7_9_7_5(SH2_struct* sh)
{
  SH2addi(sh, 9, (7<<4)+5);
}
#define init_7_9_7_6
void FASTCALL call_7_9_7_6(SH2_struct* sh)
{
  SH2addi(sh, 9, (7<<4)+6);
}
#define init_7_9_7_7
void FASTCALL call_7_9_7_7(SH2_struct* sh)
{
  SH2addi(sh, 9, (7<<4)+7);
}
#define init_7_9_7_8
void FASTCALL call_7_9_7_8(SH2_struct* sh)
{
  SH2addi(sh, 9, (7<<4)+8);
}
#define init_7_9_7_9
void FASTCALL call_7_9_7_9(SH2_struct* sh)
{
  SH2addi(sh, 9, (7<<4)+9);
}
#define init_7_9_7_10
void FASTCALL call_7_9_7_10(SH2_struct* sh)
{
  SH2addi(sh, 9, (7<<4)+10);
}
#define init_7_9_7_11
void FASTCALL call_7_9_7_11(SH2_struct* sh)
{
  SH2addi(sh, 9, (7<<4)+11);
}
#define init_7_9_7_12
void FASTCALL call_7_9_7_12(SH2_struct* sh)
{
  SH2addi(sh, 9, (7<<4)+12);
}
#define init_7_9_7_13
void FASTCALL call_7_9_7_13(SH2_struct* sh)
{
  SH2addi(sh, 9, (7<<4)+13);
}
#define init_7_9_7_14
void FASTCALL call_7_9_7_14(SH2_struct* sh)
{
  SH2addi(sh, 9, (7<<4)+14);
}
#define init_7_9_7_15
void FASTCALL call_7_9_7_15(SH2_struct* sh)
{
  SH2addi(sh, 9, (7<<4)+15);
}
#define init_7_9_8_0
void FASTCALL call_7_9_8_0(SH2_struct* sh)
{
  SH2addi(sh, 9, (8<<4)+0);
}
#define init_7_9_8_1
void FASTCALL call_7_9_8_1(SH2_struct* sh)
{
  SH2addi(sh, 9, (8<<4)+1);
}
#define init_7_9_8_2
void FASTCALL call_7_9_8_2(SH2_struct* sh)
{
  SH2addi(sh, 9, (8<<4)+2);
}
#define init_7_9_8_3
void FASTCALL call_7_9_8_3(SH2_struct* sh)
{
  SH2addi(sh, 9, (8<<4)+3);
}
#define init_7_9_8_4
void FASTCALL call_7_9_8_4(SH2_struct* sh)
{
  SH2addi(sh, 9, (8<<4)+4);
}
#define init_7_9_8_5
void FASTCALL call_7_9_8_5(SH2_struct* sh)
{
  SH2addi(sh, 9, (8<<4)+5);
}
#define init_7_9_8_6
void FASTCALL call_7_9_8_6(SH2_struct* sh)
{
  SH2addi(sh, 9, (8<<4)+6);
}
#define init_7_9_8_7
void FASTCALL call_7_9_8_7(SH2_struct* sh)
{
  SH2addi(sh, 9, (8<<4)+7);
}
#define init_7_9_8_8
void FASTCALL call_7_9_8_8(SH2_struct* sh)
{
  SH2addi(sh, 9, (8<<4)+8);
}
#define init_7_9_8_9
void FASTCALL call_7_9_8_9(SH2_struct* sh)
{
  SH2addi(sh, 9, (8<<4)+9);
}
#define init_7_9_8_10
void FASTCALL call_7_9_8_10(SH2_struct* sh)
{
  SH2addi(sh, 9, (8<<4)+10);
}
#define init_7_9_8_11
void FASTCALL call_7_9_8_11(SH2_struct* sh)
{
  SH2addi(sh, 9, (8<<4)+11);
}
#define init_7_9_8_12
void FASTCALL call_7_9_8_12(SH2_struct* sh)
{
  SH2addi(sh, 9, (8<<4)+12);
}
#define init_7_9_8_13
void FASTCALL call_7_9_8_13(SH2_struct* sh)
{
  SH2addi(sh, 9, (8<<4)+13);
}
#define init_7_9_8_14
void FASTCALL call_7_9_8_14(SH2_struct* sh)
{
  SH2addi(sh, 9, (8<<4)+14);
}
#define init_7_9_8_15
void FASTCALL call_7_9_8_15(SH2_struct* sh)
{
  SH2addi(sh, 9, (8<<4)+15);
}
#define init_7_9_9_0
void FASTCALL call_7_9_9_0(SH2_struct* sh)
{
  SH2addi(sh, 9, (9<<4)+0);
}
#define init_7_9_9_1
void FASTCALL call_7_9_9_1(SH2_struct* sh)
{
  SH2addi(sh, 9, (9<<4)+1);
}
#define init_7_9_9_2
void FASTCALL call_7_9_9_2(SH2_struct* sh)
{
  SH2addi(sh, 9, (9<<4)+2);
}
#define init_7_9_9_3
void FASTCALL call_7_9_9_3(SH2_struct* sh)
{
  SH2addi(sh, 9, (9<<4)+3);
}
#define init_7_9_9_4
void FASTCALL call_7_9_9_4(SH2_struct* sh)
{
  SH2addi(sh, 9, (9<<4)+4);
}
#define init_7_9_9_5
void FASTCALL call_7_9_9_5(SH2_struct* sh)
{
  SH2addi(sh, 9, (9<<4)+5);
}
#define init_7_9_9_6
void FASTCALL call_7_9_9_6(SH2_struct* sh)
{
  SH2addi(sh, 9, (9<<4)+6);
}
#define init_7_9_9_7
void FASTCALL call_7_9_9_7(SH2_struct* sh)
{
  SH2addi(sh, 9, (9<<4)+7);
}
#define init_7_9_9_8
void FASTCALL call_7_9_9_8(SH2_struct* sh)
{
  SH2addi(sh, 9, (9<<4)+8);
}
#define init_7_9_9_9
void FASTCALL call_7_9_9_9(SH2_struct* sh)
{
  SH2addi(sh, 9, (9<<4)+9);
}
#define init_7_9_9_10
void FASTCALL call_7_9_9_10(SH2_struct* sh)
{
  SH2addi(sh, 9, (9<<4)+10);
}
#define init_7_9_9_11
void FASTCALL call_7_9_9_11(SH2_struct* sh)
{
  SH2addi(sh, 9, (9<<4)+11);
}
#define init_7_9_9_12
void FASTCALL call_7_9_9_12(SH2_struct* sh)
{
  SH2addi(sh, 9, (9<<4)+12);
}
#define init_7_9_9_13
void FASTCALL call_7_9_9_13(SH2_struct* sh)
{
  SH2addi(sh, 9, (9<<4)+13);
}
#define init_7_9_9_14
void FASTCALL call_7_9_9_14(SH2_struct* sh)
{
  SH2addi(sh, 9, (9<<4)+14);
}
#define init_7_9_9_15
void FASTCALL call_7_9_9_15(SH2_struct* sh)
{
  SH2addi(sh, 9, (9<<4)+15);
}
#define init_7_9_10_0
void FASTCALL call_7_9_10_0(SH2_struct* sh)
{
  SH2addi(sh, 9, (10<<4)+0);
}
#define init_7_9_10_1
void FASTCALL call_7_9_10_1(SH2_struct* sh)
{
  SH2addi(sh, 9, (10<<4)+1);
}
#define init_7_9_10_2
void FASTCALL call_7_9_10_2(SH2_struct* sh)
{
  SH2addi(sh, 9, (10<<4)+2);
}
#define init_7_9_10_3
void FASTCALL call_7_9_10_3(SH2_struct* sh)
{
  SH2addi(sh, 9, (10<<4)+3);
}
#define init_7_9_10_4
void FASTCALL call_7_9_10_4(SH2_struct* sh)
{
  SH2addi(sh, 9, (10<<4)+4);
}
#define init_7_9_10_5
void FASTCALL call_7_9_10_5(SH2_struct* sh)
{
  SH2addi(sh, 9, (10<<4)+5);
}
#define init_7_9_10_6
void FASTCALL call_7_9_10_6(SH2_struct* sh)
{
  SH2addi(sh, 9, (10<<4)+6);
}
#define init_7_9_10_7
void FASTCALL call_7_9_10_7(SH2_struct* sh)
{
  SH2addi(sh, 9, (10<<4)+7);
}
#define init_7_9_10_8
void FASTCALL call_7_9_10_8(SH2_struct* sh)
{
  SH2addi(sh, 9, (10<<4)+8);
}
#define init_7_9_10_9
void FASTCALL call_7_9_10_9(SH2_struct* sh)
{
  SH2addi(sh, 9, (10<<4)+9);
}
#define init_7_9_10_10
void FASTCALL call_7_9_10_10(SH2_struct* sh)
{
  SH2addi(sh, 9, (10<<4)+10);
}
#define init_7_9_10_11
void FASTCALL call_7_9_10_11(SH2_struct* sh)
{
  SH2addi(sh, 9, (10<<4)+11);
}
#define init_7_9_10_12
void FASTCALL call_7_9_10_12(SH2_struct* sh)
{
  SH2addi(sh, 9, (10<<4)+12);
}
#define init_7_9_10_13
void FASTCALL call_7_9_10_13(SH2_struct* sh)
{
  SH2addi(sh, 9, (10<<4)+13);
}
#define init_7_9_10_14
void FASTCALL call_7_9_10_14(SH2_struct* sh)
{
  SH2addi(sh, 9, (10<<4)+14);
}
#define init_7_9_10_15
void FASTCALL call_7_9_10_15(SH2_struct* sh)
{
  SH2addi(sh, 9, (10<<4)+15);
}
#define init_7_9_11_0
void FASTCALL call_7_9_11_0(SH2_struct* sh)
{
  SH2addi(sh, 9, (11<<4)+0);
}
#define init_7_9_11_1
void FASTCALL call_7_9_11_1(SH2_struct* sh)
{
  SH2addi(sh, 9, (11<<4)+1);
}
#define init_7_9_11_2
void FASTCALL call_7_9_11_2(SH2_struct* sh)
{
  SH2addi(sh, 9, (11<<4)+2);
}
#define init_7_9_11_3
void FASTCALL call_7_9_11_3(SH2_struct* sh)
{
  SH2addi(sh, 9, (11<<4)+3);
}
#define init_7_9_11_4
void FASTCALL call_7_9_11_4(SH2_struct* sh)
{
  SH2addi(sh, 9, (11<<4)+4);
}
#define init_7_9_11_5
void FASTCALL call_7_9_11_5(SH2_struct* sh)
{
  SH2addi(sh, 9, (11<<4)+5);
}
#define init_7_9_11_6
void FASTCALL call_7_9_11_6(SH2_struct* sh)
{
  SH2addi(sh, 9, (11<<4)+6);
}
#define init_7_9_11_7
void FASTCALL call_7_9_11_7(SH2_struct* sh)
{
  SH2addi(sh, 9, (11<<4)+7);
}
#define init_7_9_11_8
void FASTCALL call_7_9_11_8(SH2_struct* sh)
{
  SH2addi(sh, 9, (11<<4)+8);
}
#define init_7_9_11_9
void FASTCALL call_7_9_11_9(SH2_struct* sh)
{
  SH2addi(sh, 9, (11<<4)+9);
}
#define init_7_9_11_10
void FASTCALL call_7_9_11_10(SH2_struct* sh)
{
  SH2addi(sh, 9, (11<<4)+10);
}
#define init_7_9_11_11
void FASTCALL call_7_9_11_11(SH2_struct* sh)
{
  SH2addi(sh, 9, (11<<4)+11);
}
#define init_7_9_11_12
void FASTCALL call_7_9_11_12(SH2_struct* sh)
{
  SH2addi(sh, 9, (11<<4)+12);
}
#define init_7_9_11_13
void FASTCALL call_7_9_11_13(SH2_struct* sh)
{
  SH2addi(sh, 9, (11<<4)+13);
}
#define init_7_9_11_14
void FASTCALL call_7_9_11_14(SH2_struct* sh)
{
  SH2addi(sh, 9, (11<<4)+14);
}
#define init_7_9_11_15
void FASTCALL call_7_9_11_15(SH2_struct* sh)
{
  SH2addi(sh, 9, (11<<4)+15);
}
#define init_7_9_12_0
void FASTCALL call_7_9_12_0(SH2_struct* sh)
{
  SH2addi(sh, 9, (12<<4)+0);
}
#define init_7_9_12_1
void FASTCALL call_7_9_12_1(SH2_struct* sh)
{
  SH2addi(sh, 9, (12<<4)+1);
}
#define init_7_9_12_2
void FASTCALL call_7_9_12_2(SH2_struct* sh)
{
  SH2addi(sh, 9, (12<<4)+2);
}
#define init_7_9_12_3
void FASTCALL call_7_9_12_3(SH2_struct* sh)
{
  SH2addi(sh, 9, (12<<4)+3);
}
#define init_7_9_12_4
void FASTCALL call_7_9_12_4(SH2_struct* sh)
{
  SH2addi(sh, 9, (12<<4)+4);
}
#define init_7_9_12_5
void FASTCALL call_7_9_12_5(SH2_struct* sh)
{
  SH2addi(sh, 9, (12<<4)+5);
}
#define init_7_9_12_6
void FASTCALL call_7_9_12_6(SH2_struct* sh)
{
  SH2addi(sh, 9, (12<<4)+6);
}
#define init_7_9_12_7
void FASTCALL call_7_9_12_7(SH2_struct* sh)
{
  SH2addi(sh, 9, (12<<4)+7);
}
#define init_7_9_12_8
void FASTCALL call_7_9_12_8(SH2_struct* sh)
{
  SH2addi(sh, 9, (12<<4)+8);
}
#define init_7_9_12_9
void FASTCALL call_7_9_12_9(SH2_struct* sh)
{
  SH2addi(sh, 9, (12<<4)+9);
}
#define init_7_9_12_10
void FASTCALL call_7_9_12_10(SH2_struct* sh)
{
  SH2addi(sh, 9, (12<<4)+10);
}
#define init_7_9_12_11
void FASTCALL call_7_9_12_11(SH2_struct* sh)
{
  SH2addi(sh, 9, (12<<4)+11);
}
#define init_7_9_12_12
void FASTCALL call_7_9_12_12(SH2_struct* sh)
{
  SH2addi(sh, 9, (12<<4)+12);
}
#define init_7_9_12_13
void FASTCALL call_7_9_12_13(SH2_struct* sh)
{
  SH2addi(sh, 9, (12<<4)+13);
}
#define init_7_9_12_14
void FASTCALL call_7_9_12_14(SH2_struct* sh)
{
  SH2addi(sh, 9, (12<<4)+14);
}
#define init_7_9_12_15
void FASTCALL call_7_9_12_15(SH2_struct* sh)
{
  SH2addi(sh, 9, (12<<4)+15);
}
#define init_7_9_13_0
void FASTCALL call_7_9_13_0(SH2_struct* sh)
{
  SH2addi(sh, 9, (13<<4)+0);
}
#define init_7_9_13_1
void FASTCALL call_7_9_13_1(SH2_struct* sh)
{
  SH2addi(sh, 9, (13<<4)+1);
}
#define init_7_9_13_2
void FASTCALL call_7_9_13_2(SH2_struct* sh)
{
  SH2addi(sh, 9, (13<<4)+2);
}
#define init_7_9_13_3
void FASTCALL call_7_9_13_3(SH2_struct* sh)
{
  SH2addi(sh, 9, (13<<4)+3);
}
#define init_7_9_13_4
void FASTCALL call_7_9_13_4(SH2_struct* sh)
{
  SH2addi(sh, 9, (13<<4)+4);
}
#define init_7_9_13_5
void FASTCALL call_7_9_13_5(SH2_struct* sh)
{
  SH2addi(sh, 9, (13<<4)+5);
}
#define init_7_9_13_6
void FASTCALL call_7_9_13_6(SH2_struct* sh)
{
  SH2addi(sh, 9, (13<<4)+6);
}
#define init_7_9_13_7
void FASTCALL call_7_9_13_7(SH2_struct* sh)
{
  SH2addi(sh, 9, (13<<4)+7);
}
#define init_7_9_13_8
void FASTCALL call_7_9_13_8(SH2_struct* sh)
{
  SH2addi(sh, 9, (13<<4)+8);
}
#define init_7_9_13_9
void FASTCALL call_7_9_13_9(SH2_struct* sh)
{
  SH2addi(sh, 9, (13<<4)+9);
}
#define init_7_9_13_10
void FASTCALL call_7_9_13_10(SH2_struct* sh)
{
  SH2addi(sh, 9, (13<<4)+10);
}
#define init_7_9_13_11
void FASTCALL call_7_9_13_11(SH2_struct* sh)
{
  SH2addi(sh, 9, (13<<4)+11);
}
#define init_7_9_13_12
void FASTCALL call_7_9_13_12(SH2_struct* sh)
{
  SH2addi(sh, 9, (13<<4)+12);
}
#define init_7_9_13_13
void FASTCALL call_7_9_13_13(SH2_struct* sh)
{
  SH2addi(sh, 9, (13<<4)+13);
}
#define init_7_9_13_14
void FASTCALL call_7_9_13_14(SH2_struct* sh)
{
  SH2addi(sh, 9, (13<<4)+14);
}
#define init_7_9_13_15
void FASTCALL call_7_9_13_15(SH2_struct* sh)
{
  SH2addi(sh, 9, (13<<4)+15);
}
#define init_7_9_14_0
void FASTCALL call_7_9_14_0(SH2_struct* sh)
{
  SH2addi(sh, 9, (14<<4)+0);
}
#define init_7_9_14_1
void FASTCALL call_7_9_14_1(SH2_struct* sh)
{
  SH2addi(sh, 9, (14<<4)+1);
}
#define init_7_9_14_2
void FASTCALL call_7_9_14_2(SH2_struct* sh)
{
  SH2addi(sh, 9, (14<<4)+2);
}
#define init_7_9_14_3
void FASTCALL call_7_9_14_3(SH2_struct* sh)
{
  SH2addi(sh, 9, (14<<4)+3);
}
#define init_7_9_14_4
void FASTCALL call_7_9_14_4(SH2_struct* sh)
{
  SH2addi(sh, 9, (14<<4)+4);
}
#define init_7_9_14_5
void FASTCALL call_7_9_14_5(SH2_struct* sh)
{
  SH2addi(sh, 9, (14<<4)+5);
}
#define init_7_9_14_6
void FASTCALL call_7_9_14_6(SH2_struct* sh)
{
  SH2addi(sh, 9, (14<<4)+6);
}
#define init_7_9_14_7
void FASTCALL call_7_9_14_7(SH2_struct* sh)
{
  SH2addi(sh, 9, (14<<4)+7);
}
#define init_7_9_14_8
void FASTCALL call_7_9_14_8(SH2_struct* sh)
{
  SH2addi(sh, 9, (14<<4)+8);
}
#define init_7_9_14_9
void FASTCALL call_7_9_14_9(SH2_struct* sh)
{
  SH2addi(sh, 9, (14<<4)+9);
}
#define init_7_9_14_10
void FASTCALL call_7_9_14_10(SH2_struct* sh)
{
  SH2addi(sh, 9, (14<<4)+10);
}
#define init_7_9_14_11
void FASTCALL call_7_9_14_11(SH2_struct* sh)
{
  SH2addi(sh, 9, (14<<4)+11);
}
#define init_7_9_14_12
void FASTCALL call_7_9_14_12(SH2_struct* sh)
{
  SH2addi(sh, 9, (14<<4)+12);
}
#define init_7_9_14_13
void FASTCALL call_7_9_14_13(SH2_struct* sh)
{
  SH2addi(sh, 9, (14<<4)+13);
}
#define init_7_9_14_14
void FASTCALL call_7_9_14_14(SH2_struct* sh)
{
  SH2addi(sh, 9, (14<<4)+14);
}
#define init_7_9_14_15
void FASTCALL call_7_9_14_15(SH2_struct* sh)
{
  SH2addi(sh, 9, (14<<4)+15);
}
#define init_7_9_15_0
void FASTCALL call_7_9_15_0(SH2_struct* sh)
{
  SH2addi(sh, 9, (15<<4)+0);
}
#define init_7_9_15_1
void FASTCALL call_7_9_15_1(SH2_struct* sh)
{
  SH2addi(sh, 9, (15<<4)+1);
}
#define init_7_9_15_2
void FASTCALL call_7_9_15_2(SH2_struct* sh)
{
  SH2addi(sh, 9, (15<<4)+2);
}
#define init_7_9_15_3
void FASTCALL call_7_9_15_3(SH2_struct* sh)
{
  SH2addi(sh, 9, (15<<4)+3);
}
#define init_7_9_15_4
void FASTCALL call_7_9_15_4(SH2_struct* sh)
{
  SH2addi(sh, 9, (15<<4)+4);
}
#define init_7_9_15_5
void FASTCALL call_7_9_15_5(SH2_struct* sh)
{
  SH2addi(sh, 9, (15<<4)+5);
}
#define init_7_9_15_6
void FASTCALL call_7_9_15_6(SH2_struct* sh)
{
  SH2addi(sh, 9, (15<<4)+6);
}
#define init_7_9_15_7
void FASTCALL call_7_9_15_7(SH2_struct* sh)
{
  SH2addi(sh, 9, (15<<4)+7);
}
#define init_7_9_15_8
void FASTCALL call_7_9_15_8(SH2_struct* sh)
{
  SH2addi(sh, 9, (15<<4)+8);
}
#define init_7_9_15_9
void FASTCALL call_7_9_15_9(SH2_struct* sh)
{
  SH2addi(sh, 9, (15<<4)+9);
}
#define init_7_9_15_10
void FASTCALL call_7_9_15_10(SH2_struct* sh)
{
  SH2addi(sh, 9, (15<<4)+10);
}
#define init_7_9_15_11
void FASTCALL call_7_9_15_11(SH2_struct* sh)
{
  SH2addi(sh, 9, (15<<4)+11);
}
#define init_7_9_15_12
void FASTCALL call_7_9_15_12(SH2_struct* sh)
{
  SH2addi(sh, 9, (15<<4)+12);
}
#define init_7_9_15_13
void FASTCALL call_7_9_15_13(SH2_struct* sh)
{
  SH2addi(sh, 9, (15<<4)+13);
}
#define init_7_9_15_14
void FASTCALL call_7_9_15_14(SH2_struct* sh)
{
  SH2addi(sh, 9, (15<<4)+14);
}
#define init_7_9_15_15
void FASTCALL call_7_9_15_15(SH2_struct* sh)
{
  SH2addi(sh, 9, (15<<4)+15);
}
#define init_7_10_0_0
void FASTCALL call_7_10_0_0(SH2_struct* sh)
{
  SH2addi(sh, 10, (0<<4)+0);
}
#define init_7_10_0_1
void FASTCALL call_7_10_0_1(SH2_struct* sh)
{
  SH2addi(sh, 10, (0<<4)+1);
}
#define init_7_10_0_2
void FASTCALL call_7_10_0_2(SH2_struct* sh)
{
  SH2addi(sh, 10, (0<<4)+2);
}
#define init_7_10_0_3
void FASTCALL call_7_10_0_3(SH2_struct* sh)
{
  SH2addi(sh, 10, (0<<4)+3);
}
#define init_7_10_0_4
void FASTCALL call_7_10_0_4(SH2_struct* sh)
{
  SH2addi(sh, 10, (0<<4)+4);
}
#define init_7_10_0_5
void FASTCALL call_7_10_0_5(SH2_struct* sh)
{
  SH2addi(sh, 10, (0<<4)+5);
}
#define init_7_10_0_6
void FASTCALL call_7_10_0_6(SH2_struct* sh)
{
  SH2addi(sh, 10, (0<<4)+6);
}
#define init_7_10_0_7
void FASTCALL call_7_10_0_7(SH2_struct* sh)
{
  SH2addi(sh, 10, (0<<4)+7);
}
#define init_7_10_0_8
void FASTCALL call_7_10_0_8(SH2_struct* sh)
{
  SH2addi(sh, 10, (0<<4)+8);
}
#define init_7_10_0_9
void FASTCALL call_7_10_0_9(SH2_struct* sh)
{
  SH2addi(sh, 10, (0<<4)+9);
}
#define init_7_10_0_10
void FASTCALL call_7_10_0_10(SH2_struct* sh)
{
  SH2addi(sh, 10, (0<<4)+10);
}
#define init_7_10_0_11
void FASTCALL call_7_10_0_11(SH2_struct* sh)
{
  SH2addi(sh, 10, (0<<4)+11);
}
#define init_7_10_0_12
void FASTCALL call_7_10_0_12(SH2_struct* sh)
{
  SH2addi(sh, 10, (0<<4)+12);
}
#define init_7_10_0_13
void FASTCALL call_7_10_0_13(SH2_struct* sh)
{
  SH2addi(sh, 10, (0<<4)+13);
}
#define init_7_10_0_14
void FASTCALL call_7_10_0_14(SH2_struct* sh)
{
  SH2addi(sh, 10, (0<<4)+14);
}
#define init_7_10_0_15
void FASTCALL call_7_10_0_15(SH2_struct* sh)
{
  SH2addi(sh, 10, (0<<4)+15);
}
#define init_7_10_1_0
void FASTCALL call_7_10_1_0(SH2_struct* sh)
{
  SH2addi(sh, 10, (1<<4)+0);
}
#define init_7_10_1_1
void FASTCALL call_7_10_1_1(SH2_struct* sh)
{
  SH2addi(sh, 10, (1<<4)+1);
}
#define init_7_10_1_2
void FASTCALL call_7_10_1_2(SH2_struct* sh)
{
  SH2addi(sh, 10, (1<<4)+2);
}
#define init_7_10_1_3
void FASTCALL call_7_10_1_3(SH2_struct* sh)
{
  SH2addi(sh, 10, (1<<4)+3);
}
#define init_7_10_1_4
void FASTCALL call_7_10_1_4(SH2_struct* sh)
{
  SH2addi(sh, 10, (1<<4)+4);
}
#define init_7_10_1_5
void FASTCALL call_7_10_1_5(SH2_struct* sh)
{
  SH2addi(sh, 10, (1<<4)+5);
}
#define init_7_10_1_6
void FASTCALL call_7_10_1_6(SH2_struct* sh)
{
  SH2addi(sh, 10, (1<<4)+6);
}
#define init_7_10_1_7
void FASTCALL call_7_10_1_7(SH2_struct* sh)
{
  SH2addi(sh, 10, (1<<4)+7);
}
#define init_7_10_1_8
void FASTCALL call_7_10_1_8(SH2_struct* sh)
{
  SH2addi(sh, 10, (1<<4)+8);
}
#define init_7_10_1_9
void FASTCALL call_7_10_1_9(SH2_struct* sh)
{
  SH2addi(sh, 10, (1<<4)+9);
}
#define init_7_10_1_10
void FASTCALL call_7_10_1_10(SH2_struct* sh)
{
  SH2addi(sh, 10, (1<<4)+10);
}
#define init_7_10_1_11
void FASTCALL call_7_10_1_11(SH2_struct* sh)
{
  SH2addi(sh, 10, (1<<4)+11);
}
#define init_7_10_1_12
void FASTCALL call_7_10_1_12(SH2_struct* sh)
{
  SH2addi(sh, 10, (1<<4)+12);
}
#define init_7_10_1_13
void FASTCALL call_7_10_1_13(SH2_struct* sh)
{
  SH2addi(sh, 10, (1<<4)+13);
}
#define init_7_10_1_14
void FASTCALL call_7_10_1_14(SH2_struct* sh)
{
  SH2addi(sh, 10, (1<<4)+14);
}
#define init_7_10_1_15
void FASTCALL call_7_10_1_15(SH2_struct* sh)
{
  SH2addi(sh, 10, (1<<4)+15);
}
#define init_7_10_2_0
void FASTCALL call_7_10_2_0(SH2_struct* sh)
{
  SH2addi(sh, 10, (2<<4)+0);
}
#define init_7_10_2_1
void FASTCALL call_7_10_2_1(SH2_struct* sh)
{
  SH2addi(sh, 10, (2<<4)+1);
}
#define init_7_10_2_2
void FASTCALL call_7_10_2_2(SH2_struct* sh)
{
  SH2addi(sh, 10, (2<<4)+2);
}
#define init_7_10_2_3
void FASTCALL call_7_10_2_3(SH2_struct* sh)
{
  SH2addi(sh, 10, (2<<4)+3);
}
#define init_7_10_2_4
void FASTCALL call_7_10_2_4(SH2_struct* sh)
{
  SH2addi(sh, 10, (2<<4)+4);
}
#define init_7_10_2_5
void FASTCALL call_7_10_2_5(SH2_struct* sh)
{
  SH2addi(sh, 10, (2<<4)+5);
}
#define init_7_10_2_6
void FASTCALL call_7_10_2_6(SH2_struct* sh)
{
  SH2addi(sh, 10, (2<<4)+6);
}
#define init_7_10_2_7
void FASTCALL call_7_10_2_7(SH2_struct* sh)
{
  SH2addi(sh, 10, (2<<4)+7);
}
#define init_7_10_2_8
void FASTCALL call_7_10_2_8(SH2_struct* sh)
{
  SH2addi(sh, 10, (2<<4)+8);
}
#define init_7_10_2_9
void FASTCALL call_7_10_2_9(SH2_struct* sh)
{
  SH2addi(sh, 10, (2<<4)+9);
}
#define init_7_10_2_10
void FASTCALL call_7_10_2_10(SH2_struct* sh)
{
  SH2addi(sh, 10, (2<<4)+10);
}
#define init_7_10_2_11
void FASTCALL call_7_10_2_11(SH2_struct* sh)
{
  SH2addi(sh, 10, (2<<4)+11);
}
#define init_7_10_2_12
void FASTCALL call_7_10_2_12(SH2_struct* sh)
{
  SH2addi(sh, 10, (2<<4)+12);
}
#define init_7_10_2_13
void FASTCALL call_7_10_2_13(SH2_struct* sh)
{
  SH2addi(sh, 10, (2<<4)+13);
}
#define init_7_10_2_14
void FASTCALL call_7_10_2_14(SH2_struct* sh)
{
  SH2addi(sh, 10, (2<<4)+14);
}
#define init_7_10_2_15
void FASTCALL call_7_10_2_15(SH2_struct* sh)
{
  SH2addi(sh, 10, (2<<4)+15);
}
#define init_7_10_3_0
void FASTCALL call_7_10_3_0(SH2_struct* sh)
{
  SH2addi(sh, 10, (3<<4)+0);
}
#define init_7_10_3_1
void FASTCALL call_7_10_3_1(SH2_struct* sh)
{
  SH2addi(sh, 10, (3<<4)+1);
}
#define init_7_10_3_2
void FASTCALL call_7_10_3_2(SH2_struct* sh)
{
  SH2addi(sh, 10, (3<<4)+2);
}
#define init_7_10_3_3
void FASTCALL call_7_10_3_3(SH2_struct* sh)
{
  SH2addi(sh, 10, (3<<4)+3);
}
#define init_7_10_3_4
void FASTCALL call_7_10_3_4(SH2_struct* sh)
{
  SH2addi(sh, 10, (3<<4)+4);
}
#define init_7_10_3_5
void FASTCALL call_7_10_3_5(SH2_struct* sh)
{
  SH2addi(sh, 10, (3<<4)+5);
}
#define init_7_10_3_6
void FASTCALL call_7_10_3_6(SH2_struct* sh)
{
  SH2addi(sh, 10, (3<<4)+6);
}
#define init_7_10_3_7
void FASTCALL call_7_10_3_7(SH2_struct* sh)
{
  SH2addi(sh, 10, (3<<4)+7);
}
#define init_7_10_3_8
void FASTCALL call_7_10_3_8(SH2_struct* sh)
{
  SH2addi(sh, 10, (3<<4)+8);
}
#define init_7_10_3_9
void FASTCALL call_7_10_3_9(SH2_struct* sh)
{
  SH2addi(sh, 10, (3<<4)+9);
}
#define init_7_10_3_10
void FASTCALL call_7_10_3_10(SH2_struct* sh)
{
  SH2addi(sh, 10, (3<<4)+10);
}
#define init_7_10_3_11
void FASTCALL call_7_10_3_11(SH2_struct* sh)
{
  SH2addi(sh, 10, (3<<4)+11);
}
#define init_7_10_3_12
void FASTCALL call_7_10_3_12(SH2_struct* sh)
{
  SH2addi(sh, 10, (3<<4)+12);
}
#define init_7_10_3_13
void FASTCALL call_7_10_3_13(SH2_struct* sh)
{
  SH2addi(sh, 10, (3<<4)+13);
}
#define init_7_10_3_14
void FASTCALL call_7_10_3_14(SH2_struct* sh)
{
  SH2addi(sh, 10, (3<<4)+14);
}
#define init_7_10_3_15
void FASTCALL call_7_10_3_15(SH2_struct* sh)
{
  SH2addi(sh, 10, (3<<4)+15);
}
#define init_7_10_4_0
void FASTCALL call_7_10_4_0(SH2_struct* sh)
{
  SH2addi(sh, 10, (4<<4)+0);
}
#define init_7_10_4_1
void FASTCALL call_7_10_4_1(SH2_struct* sh)
{
  SH2addi(sh, 10, (4<<4)+1);
}
#define init_7_10_4_2
void FASTCALL call_7_10_4_2(SH2_struct* sh)
{
  SH2addi(sh, 10, (4<<4)+2);
}
#define init_7_10_4_3
void FASTCALL call_7_10_4_3(SH2_struct* sh)
{
  SH2addi(sh, 10, (4<<4)+3);
}
#define init_7_10_4_4
void FASTCALL call_7_10_4_4(SH2_struct* sh)
{
  SH2addi(sh, 10, (4<<4)+4);
}
#define init_7_10_4_5
void FASTCALL call_7_10_4_5(SH2_struct* sh)
{
  SH2addi(sh, 10, (4<<4)+5);
}
#define init_7_10_4_6
void FASTCALL call_7_10_4_6(SH2_struct* sh)
{
  SH2addi(sh, 10, (4<<4)+6);
}
#define init_7_10_4_7
void FASTCALL call_7_10_4_7(SH2_struct* sh)
{
  SH2addi(sh, 10, (4<<4)+7);
}
#define init_7_10_4_8
void FASTCALL call_7_10_4_8(SH2_struct* sh)
{
  SH2addi(sh, 10, (4<<4)+8);
}
#define init_7_10_4_9
void FASTCALL call_7_10_4_9(SH2_struct* sh)
{
  SH2addi(sh, 10, (4<<4)+9);
}
#define init_7_10_4_10
void FASTCALL call_7_10_4_10(SH2_struct* sh)
{
  SH2addi(sh, 10, (4<<4)+10);
}
#define init_7_10_4_11
void FASTCALL call_7_10_4_11(SH2_struct* sh)
{
  SH2addi(sh, 10, (4<<4)+11);
}
#define init_7_10_4_12
void FASTCALL call_7_10_4_12(SH2_struct* sh)
{
  SH2addi(sh, 10, (4<<4)+12);
}
#define init_7_10_4_13
void FASTCALL call_7_10_4_13(SH2_struct* sh)
{
  SH2addi(sh, 10, (4<<4)+13);
}
#define init_7_10_4_14
void FASTCALL call_7_10_4_14(SH2_struct* sh)
{
  SH2addi(sh, 10, (4<<4)+14);
}
#define init_7_10_4_15
void FASTCALL call_7_10_4_15(SH2_struct* sh)
{
  SH2addi(sh, 10, (4<<4)+15);
}
#define init_7_10_5_0
void FASTCALL call_7_10_5_0(SH2_struct* sh)
{
  SH2addi(sh, 10, (5<<4)+0);
}
#define init_7_10_5_1
void FASTCALL call_7_10_5_1(SH2_struct* sh)
{
  SH2addi(sh, 10, (5<<4)+1);
}
#define init_7_10_5_2
void FASTCALL call_7_10_5_2(SH2_struct* sh)
{
  SH2addi(sh, 10, (5<<4)+2);
}
#define init_7_10_5_3
void FASTCALL call_7_10_5_3(SH2_struct* sh)
{
  SH2addi(sh, 10, (5<<4)+3);
}
#define init_7_10_5_4
void FASTCALL call_7_10_5_4(SH2_struct* sh)
{
  SH2addi(sh, 10, (5<<4)+4);
}
#define init_7_10_5_5
void FASTCALL call_7_10_5_5(SH2_struct* sh)
{
  SH2addi(sh, 10, (5<<4)+5);
}
#define init_7_10_5_6
void FASTCALL call_7_10_5_6(SH2_struct* sh)
{
  SH2addi(sh, 10, (5<<4)+6);
}
#define init_7_10_5_7
void FASTCALL call_7_10_5_7(SH2_struct* sh)
{
  SH2addi(sh, 10, (5<<4)+7);
}
#define init_7_10_5_8
void FASTCALL call_7_10_5_8(SH2_struct* sh)
{
  SH2addi(sh, 10, (5<<4)+8);
}
#define init_7_10_5_9
void FASTCALL call_7_10_5_9(SH2_struct* sh)
{
  SH2addi(sh, 10, (5<<4)+9);
}
#define init_7_10_5_10
void FASTCALL call_7_10_5_10(SH2_struct* sh)
{
  SH2addi(sh, 10, (5<<4)+10);
}
#define init_7_10_5_11
void FASTCALL call_7_10_5_11(SH2_struct* sh)
{
  SH2addi(sh, 10, (5<<4)+11);
}
#define init_7_10_5_12
void FASTCALL call_7_10_5_12(SH2_struct* sh)
{
  SH2addi(sh, 10, (5<<4)+12);
}
#define init_7_10_5_13
void FASTCALL call_7_10_5_13(SH2_struct* sh)
{
  SH2addi(sh, 10, (5<<4)+13);
}
#define init_7_10_5_14
void FASTCALL call_7_10_5_14(SH2_struct* sh)
{
  SH2addi(sh, 10, (5<<4)+14);
}
#define init_7_10_5_15
void FASTCALL call_7_10_5_15(SH2_struct* sh)
{
  SH2addi(sh, 10, (5<<4)+15);
}
#define init_7_10_6_0
void FASTCALL call_7_10_6_0(SH2_struct* sh)
{
  SH2addi(sh, 10, (6<<4)+0);
}
#define init_7_10_6_1
void FASTCALL call_7_10_6_1(SH2_struct* sh)
{
  SH2addi(sh, 10, (6<<4)+1);
}
#define init_7_10_6_2
void FASTCALL call_7_10_6_2(SH2_struct* sh)
{
  SH2addi(sh, 10, (6<<4)+2);
}
#define init_7_10_6_3
void FASTCALL call_7_10_6_3(SH2_struct* sh)
{
  SH2addi(sh, 10, (6<<4)+3);
}
#define init_7_10_6_4
void FASTCALL call_7_10_6_4(SH2_struct* sh)
{
  SH2addi(sh, 10, (6<<4)+4);
}
#define init_7_10_6_5
void FASTCALL call_7_10_6_5(SH2_struct* sh)
{
  SH2addi(sh, 10, (6<<4)+5);
}
#define init_7_10_6_6
void FASTCALL call_7_10_6_6(SH2_struct* sh)
{
  SH2addi(sh, 10, (6<<4)+6);
}
#define init_7_10_6_7
void FASTCALL call_7_10_6_7(SH2_struct* sh)
{
  SH2addi(sh, 10, (6<<4)+7);
}
#define init_7_10_6_8
void FASTCALL call_7_10_6_8(SH2_struct* sh)
{
  SH2addi(sh, 10, (6<<4)+8);
}
#define init_7_10_6_9
void FASTCALL call_7_10_6_9(SH2_struct* sh)
{
  SH2addi(sh, 10, (6<<4)+9);
}
#define init_7_10_6_10
void FASTCALL call_7_10_6_10(SH2_struct* sh)
{
  SH2addi(sh, 10, (6<<4)+10);
}
#define init_7_10_6_11
void FASTCALL call_7_10_6_11(SH2_struct* sh)
{
  SH2addi(sh, 10, (6<<4)+11);
}
#define init_7_10_6_12
void FASTCALL call_7_10_6_12(SH2_struct* sh)
{
  SH2addi(sh, 10, (6<<4)+12);
}
#define init_7_10_6_13
void FASTCALL call_7_10_6_13(SH2_struct* sh)
{
  SH2addi(sh, 10, (6<<4)+13);
}
#define init_7_10_6_14
void FASTCALL call_7_10_6_14(SH2_struct* sh)
{
  SH2addi(sh, 10, (6<<4)+14);
}
#define init_7_10_6_15
void FASTCALL call_7_10_6_15(SH2_struct* sh)
{
  SH2addi(sh, 10, (6<<4)+15);
}
#define init_7_10_7_0
void FASTCALL call_7_10_7_0(SH2_struct* sh)
{
  SH2addi(sh, 10, (7<<4)+0);
}
#define init_7_10_7_1
void FASTCALL call_7_10_7_1(SH2_struct* sh)
{
  SH2addi(sh, 10, (7<<4)+1);
}
#define init_7_10_7_2
void FASTCALL call_7_10_7_2(SH2_struct* sh)
{
  SH2addi(sh, 10, (7<<4)+2);
}
#define init_7_10_7_3
void FASTCALL call_7_10_7_3(SH2_struct* sh)
{
  SH2addi(sh, 10, (7<<4)+3);
}
#define init_7_10_7_4
void FASTCALL call_7_10_7_4(SH2_struct* sh)
{
  SH2addi(sh, 10, (7<<4)+4);
}
#define init_7_10_7_5
void FASTCALL call_7_10_7_5(SH2_struct* sh)
{
  SH2addi(sh, 10, (7<<4)+5);
}
#define init_7_10_7_6
void FASTCALL call_7_10_7_6(SH2_struct* sh)
{
  SH2addi(sh, 10, (7<<4)+6);
}
#define init_7_10_7_7
void FASTCALL call_7_10_7_7(SH2_struct* sh)
{
  SH2addi(sh, 10, (7<<4)+7);
}
#define init_7_10_7_8
void FASTCALL call_7_10_7_8(SH2_struct* sh)
{
  SH2addi(sh, 10, (7<<4)+8);
}
#define init_7_10_7_9
void FASTCALL call_7_10_7_9(SH2_struct* sh)
{
  SH2addi(sh, 10, (7<<4)+9);
}
#define init_7_10_7_10
void FASTCALL call_7_10_7_10(SH2_struct* sh)
{
  SH2addi(sh, 10, (7<<4)+10);
}
#define init_7_10_7_11
void FASTCALL call_7_10_7_11(SH2_struct* sh)
{
  SH2addi(sh, 10, (7<<4)+11);
}
#define init_7_10_7_12
void FASTCALL call_7_10_7_12(SH2_struct* sh)
{
  SH2addi(sh, 10, (7<<4)+12);
}
#define init_7_10_7_13
void FASTCALL call_7_10_7_13(SH2_struct* sh)
{
  SH2addi(sh, 10, (7<<4)+13);
}
#define init_7_10_7_14
void FASTCALL call_7_10_7_14(SH2_struct* sh)
{
  SH2addi(sh, 10, (7<<4)+14);
}
#define init_7_10_7_15
void FASTCALL call_7_10_7_15(SH2_struct* sh)
{
  SH2addi(sh, 10, (7<<4)+15);
}
#define init_7_10_8_0
void FASTCALL call_7_10_8_0(SH2_struct* sh)
{
  SH2addi(sh, 10, (8<<4)+0);
}
#define init_7_10_8_1
void FASTCALL call_7_10_8_1(SH2_struct* sh)
{
  SH2addi(sh, 10, (8<<4)+1);
}
#define init_7_10_8_2
void FASTCALL call_7_10_8_2(SH2_struct* sh)
{
  SH2addi(sh, 10, (8<<4)+2);
}
#define init_7_10_8_3
void FASTCALL call_7_10_8_3(SH2_struct* sh)
{
  SH2addi(sh, 10, (8<<4)+3);
}
#define init_7_10_8_4
void FASTCALL call_7_10_8_4(SH2_struct* sh)
{
  SH2addi(sh, 10, (8<<4)+4);
}
#define init_7_10_8_5
void FASTCALL call_7_10_8_5(SH2_struct* sh)
{
  SH2addi(sh, 10, (8<<4)+5);
}
#define init_7_10_8_6
void FASTCALL call_7_10_8_6(SH2_struct* sh)
{
  SH2addi(sh, 10, (8<<4)+6);
}
#define init_7_10_8_7
void FASTCALL call_7_10_8_7(SH2_struct* sh)
{
  SH2addi(sh, 10, (8<<4)+7);
}
#define init_7_10_8_8
void FASTCALL call_7_10_8_8(SH2_struct* sh)
{
  SH2addi(sh, 10, (8<<4)+8);
}
#define init_7_10_8_9
void FASTCALL call_7_10_8_9(SH2_struct* sh)
{
  SH2addi(sh, 10, (8<<4)+9);
}
#define init_7_10_8_10
void FASTCALL call_7_10_8_10(SH2_struct* sh)
{
  SH2addi(sh, 10, (8<<4)+10);
}
#define init_7_10_8_11
void FASTCALL call_7_10_8_11(SH2_struct* sh)
{
  SH2addi(sh, 10, (8<<4)+11);
}
#define init_7_10_8_12
void FASTCALL call_7_10_8_12(SH2_struct* sh)
{
  SH2addi(sh, 10, (8<<4)+12);
}
#define init_7_10_8_13
void FASTCALL call_7_10_8_13(SH2_struct* sh)
{
  SH2addi(sh, 10, (8<<4)+13);
}
#define init_7_10_8_14
void FASTCALL call_7_10_8_14(SH2_struct* sh)
{
  SH2addi(sh, 10, (8<<4)+14);
}
#define init_7_10_8_15
void FASTCALL call_7_10_8_15(SH2_struct* sh)
{
  SH2addi(sh, 10, (8<<4)+15);
}
#define init_7_10_9_0
void FASTCALL call_7_10_9_0(SH2_struct* sh)
{
  SH2addi(sh, 10, (9<<4)+0);
}
#define init_7_10_9_1
void FASTCALL call_7_10_9_1(SH2_struct* sh)
{
  SH2addi(sh, 10, (9<<4)+1);
}
#define init_7_10_9_2
void FASTCALL call_7_10_9_2(SH2_struct* sh)
{
  SH2addi(sh, 10, (9<<4)+2);
}
#define init_7_10_9_3
void FASTCALL call_7_10_9_3(SH2_struct* sh)
{
  SH2addi(sh, 10, (9<<4)+3);
}
#define init_7_10_9_4
void FASTCALL call_7_10_9_4(SH2_struct* sh)
{
  SH2addi(sh, 10, (9<<4)+4);
}
#define init_7_10_9_5
void FASTCALL call_7_10_9_5(SH2_struct* sh)
{
  SH2addi(sh, 10, (9<<4)+5);
}
#define init_7_10_9_6
void FASTCALL call_7_10_9_6(SH2_struct* sh)
{
  SH2addi(sh, 10, (9<<4)+6);
}
#define init_7_10_9_7
void FASTCALL call_7_10_9_7(SH2_struct* sh)
{
  SH2addi(sh, 10, (9<<4)+7);
}
#define init_7_10_9_8
void FASTCALL call_7_10_9_8(SH2_struct* sh)
{
  SH2addi(sh, 10, (9<<4)+8);
}
#define init_7_10_9_9
void FASTCALL call_7_10_9_9(SH2_struct* sh)
{
  SH2addi(sh, 10, (9<<4)+9);
}
#define init_7_10_9_10
void FASTCALL call_7_10_9_10(SH2_struct* sh)
{
  SH2addi(sh, 10, (9<<4)+10);
}
#define init_7_10_9_11
void FASTCALL call_7_10_9_11(SH2_struct* sh)
{
  SH2addi(sh, 10, (9<<4)+11);
}
#define init_7_10_9_12
void FASTCALL call_7_10_9_12(SH2_struct* sh)
{
  SH2addi(sh, 10, (9<<4)+12);
}
#define init_7_10_9_13
void FASTCALL call_7_10_9_13(SH2_struct* sh)
{
  SH2addi(sh, 10, (9<<4)+13);
}
#define init_7_10_9_14
void FASTCALL call_7_10_9_14(SH2_struct* sh)
{
  SH2addi(sh, 10, (9<<4)+14);
}
#define init_7_10_9_15
void FASTCALL call_7_10_9_15(SH2_struct* sh)
{
  SH2addi(sh, 10, (9<<4)+15);
}
#define init_7_10_10_0
void FASTCALL call_7_10_10_0(SH2_struct* sh)
{
  SH2addi(sh, 10, (10<<4)+0);
}
#define init_7_10_10_1
void FASTCALL call_7_10_10_1(SH2_struct* sh)
{
  SH2addi(sh, 10, (10<<4)+1);
}
#define init_7_10_10_2
void FASTCALL call_7_10_10_2(SH2_struct* sh)
{
  SH2addi(sh, 10, (10<<4)+2);
}
#define init_7_10_10_3
void FASTCALL call_7_10_10_3(SH2_struct* sh)
{
  SH2addi(sh, 10, (10<<4)+3);
}
#define init_7_10_10_4
void FASTCALL call_7_10_10_4(SH2_struct* sh)
{
  SH2addi(sh, 10, (10<<4)+4);
}
#define init_7_10_10_5
void FASTCALL call_7_10_10_5(SH2_struct* sh)
{
  SH2addi(sh, 10, (10<<4)+5);
}
#define init_7_10_10_6
void FASTCALL call_7_10_10_6(SH2_struct* sh)
{
  SH2addi(sh, 10, (10<<4)+6);
}
#define init_7_10_10_7
void FASTCALL call_7_10_10_7(SH2_struct* sh)
{
  SH2addi(sh, 10, (10<<4)+7);
}
#define init_7_10_10_8
void FASTCALL call_7_10_10_8(SH2_struct* sh)
{
  SH2addi(sh, 10, (10<<4)+8);
}
#define init_7_10_10_9
void FASTCALL call_7_10_10_9(SH2_struct* sh)
{
  SH2addi(sh, 10, (10<<4)+9);
}
#define init_7_10_10_10
void FASTCALL call_7_10_10_10(SH2_struct* sh)
{
  SH2addi(sh, 10, (10<<4)+10);
}
#define init_7_10_10_11
void FASTCALL call_7_10_10_11(SH2_struct* sh)
{
  SH2addi(sh, 10, (10<<4)+11);
}
#define init_7_10_10_12
void FASTCALL call_7_10_10_12(SH2_struct* sh)
{
  SH2addi(sh, 10, (10<<4)+12);
}
#define init_7_10_10_13
void FASTCALL call_7_10_10_13(SH2_struct* sh)
{
  SH2addi(sh, 10, (10<<4)+13);
}
#define init_7_10_10_14
void FASTCALL call_7_10_10_14(SH2_struct* sh)
{
  SH2addi(sh, 10, (10<<4)+14);
}
#define init_7_10_10_15
void FASTCALL call_7_10_10_15(SH2_struct* sh)
{
  SH2addi(sh, 10, (10<<4)+15);
}
#define init_7_10_11_0
void FASTCALL call_7_10_11_0(SH2_struct* sh)
{
  SH2addi(sh, 10, (11<<4)+0);
}
#define init_7_10_11_1
void FASTCALL call_7_10_11_1(SH2_struct* sh)
{
  SH2addi(sh, 10, (11<<4)+1);
}
#define init_7_10_11_2
void FASTCALL call_7_10_11_2(SH2_struct* sh)
{
  SH2addi(sh, 10, (11<<4)+2);
}
#define init_7_10_11_3
void FASTCALL call_7_10_11_3(SH2_struct* sh)
{
  SH2addi(sh, 10, (11<<4)+3);
}
#define init_7_10_11_4
void FASTCALL call_7_10_11_4(SH2_struct* sh)
{
  SH2addi(sh, 10, (11<<4)+4);
}
#define init_7_10_11_5
void FASTCALL call_7_10_11_5(SH2_struct* sh)
{
  SH2addi(sh, 10, (11<<4)+5);
}
#define init_7_10_11_6
void FASTCALL call_7_10_11_6(SH2_struct* sh)
{
  SH2addi(sh, 10, (11<<4)+6);
}
#define init_7_10_11_7
void FASTCALL call_7_10_11_7(SH2_struct* sh)
{
  SH2addi(sh, 10, (11<<4)+7);
}
#define init_7_10_11_8
void FASTCALL call_7_10_11_8(SH2_struct* sh)
{
  SH2addi(sh, 10, (11<<4)+8);
}
#define init_7_10_11_9
void FASTCALL call_7_10_11_9(SH2_struct* sh)
{
  SH2addi(sh, 10, (11<<4)+9);
}
#define init_7_10_11_10
void FASTCALL call_7_10_11_10(SH2_struct* sh)
{
  SH2addi(sh, 10, (11<<4)+10);
}
#define init_7_10_11_11
void FASTCALL call_7_10_11_11(SH2_struct* sh)
{
  SH2addi(sh, 10, (11<<4)+11);
}
#define init_7_10_11_12
void FASTCALL call_7_10_11_12(SH2_struct* sh)
{
  SH2addi(sh, 10, (11<<4)+12);
}
#define init_7_10_11_13
void FASTCALL call_7_10_11_13(SH2_struct* sh)
{
  SH2addi(sh, 10, (11<<4)+13);
}
#define init_7_10_11_14
void FASTCALL call_7_10_11_14(SH2_struct* sh)
{
  SH2addi(sh, 10, (11<<4)+14);
}
#define init_7_10_11_15
void FASTCALL call_7_10_11_15(SH2_struct* sh)
{
  SH2addi(sh, 10, (11<<4)+15);
}
#define init_7_10_12_0
void FASTCALL call_7_10_12_0(SH2_struct* sh)
{
  SH2addi(sh, 10, (12<<4)+0);
}
#define init_7_10_12_1
void FASTCALL call_7_10_12_1(SH2_struct* sh)
{
  SH2addi(sh, 10, (12<<4)+1);
}
#define init_7_10_12_2
void FASTCALL call_7_10_12_2(SH2_struct* sh)
{
  SH2addi(sh, 10, (12<<4)+2);
}
#define init_7_10_12_3
void FASTCALL call_7_10_12_3(SH2_struct* sh)
{
  SH2addi(sh, 10, (12<<4)+3);
}
#define init_7_10_12_4
void FASTCALL call_7_10_12_4(SH2_struct* sh)
{
  SH2addi(sh, 10, (12<<4)+4);
}
#define init_7_10_12_5
void FASTCALL call_7_10_12_5(SH2_struct* sh)
{
  SH2addi(sh, 10, (12<<4)+5);
}
#define init_7_10_12_6
void FASTCALL call_7_10_12_6(SH2_struct* sh)
{
  SH2addi(sh, 10, (12<<4)+6);
}
#define init_7_10_12_7
void FASTCALL call_7_10_12_7(SH2_struct* sh)
{
  SH2addi(sh, 10, (12<<4)+7);
}
#define init_7_10_12_8
void FASTCALL call_7_10_12_8(SH2_struct* sh)
{
  SH2addi(sh, 10, (12<<4)+8);
}
#define init_7_10_12_9
void FASTCALL call_7_10_12_9(SH2_struct* sh)
{
  SH2addi(sh, 10, (12<<4)+9);
}
#define init_7_10_12_10
void FASTCALL call_7_10_12_10(SH2_struct* sh)
{
  SH2addi(sh, 10, (12<<4)+10);
}
#define init_7_10_12_11
void FASTCALL call_7_10_12_11(SH2_struct* sh)
{
  SH2addi(sh, 10, (12<<4)+11);
}
#define init_7_10_12_12
void FASTCALL call_7_10_12_12(SH2_struct* sh)
{
  SH2addi(sh, 10, (12<<4)+12);
}
#define init_7_10_12_13
void FASTCALL call_7_10_12_13(SH2_struct* sh)
{
  SH2addi(sh, 10, (12<<4)+13);
}
#define init_7_10_12_14
void FASTCALL call_7_10_12_14(SH2_struct* sh)
{
  SH2addi(sh, 10, (12<<4)+14);
}
#define init_7_10_12_15
void FASTCALL call_7_10_12_15(SH2_struct* sh)
{
  SH2addi(sh, 10, (12<<4)+15);
}
#define init_7_10_13_0
void FASTCALL call_7_10_13_0(SH2_struct* sh)
{
  SH2addi(sh, 10, (13<<4)+0);
}
#define init_7_10_13_1
void FASTCALL call_7_10_13_1(SH2_struct* sh)
{
  SH2addi(sh, 10, (13<<4)+1);
}
#define init_7_10_13_2
void FASTCALL call_7_10_13_2(SH2_struct* sh)
{
  SH2addi(sh, 10, (13<<4)+2);
}
#define init_7_10_13_3
void FASTCALL call_7_10_13_3(SH2_struct* sh)
{
  SH2addi(sh, 10, (13<<4)+3);
}
#define init_7_10_13_4
void FASTCALL call_7_10_13_4(SH2_struct* sh)
{
  SH2addi(sh, 10, (13<<4)+4);
}
#define init_7_10_13_5
void FASTCALL call_7_10_13_5(SH2_struct* sh)
{
  SH2addi(sh, 10, (13<<4)+5);
}
#define init_7_10_13_6
void FASTCALL call_7_10_13_6(SH2_struct* sh)
{
  SH2addi(sh, 10, (13<<4)+6);
}
#define init_7_10_13_7
void FASTCALL call_7_10_13_7(SH2_struct* sh)
{
  SH2addi(sh, 10, (13<<4)+7);
}
#define init_7_10_13_8
void FASTCALL call_7_10_13_8(SH2_struct* sh)
{
  SH2addi(sh, 10, (13<<4)+8);
}
#define init_7_10_13_9
void FASTCALL call_7_10_13_9(SH2_struct* sh)
{
  SH2addi(sh, 10, (13<<4)+9);
}
#define init_7_10_13_10
void FASTCALL call_7_10_13_10(SH2_struct* sh)
{
  SH2addi(sh, 10, (13<<4)+10);
}
#define init_7_10_13_11
void FASTCALL call_7_10_13_11(SH2_struct* sh)
{
  SH2addi(sh, 10, (13<<4)+11);
}
#define init_7_10_13_12
void FASTCALL call_7_10_13_12(SH2_struct* sh)
{
  SH2addi(sh, 10, (13<<4)+12);
}
#define init_7_10_13_13
void FASTCALL call_7_10_13_13(SH2_struct* sh)
{
  SH2addi(sh, 10, (13<<4)+13);
}
#define init_7_10_13_14
void FASTCALL call_7_10_13_14(SH2_struct* sh)
{
  SH2addi(sh, 10, (13<<4)+14);
}
#define init_7_10_13_15
void FASTCALL call_7_10_13_15(SH2_struct* sh)
{
  SH2addi(sh, 10, (13<<4)+15);
}
#define init_7_10_14_0
void FASTCALL call_7_10_14_0(SH2_struct* sh)
{
  SH2addi(sh, 10, (14<<4)+0);
}
#define init_7_10_14_1
void FASTCALL call_7_10_14_1(SH2_struct* sh)
{
  SH2addi(sh, 10, (14<<4)+1);
}
#define init_7_10_14_2
void FASTCALL call_7_10_14_2(SH2_struct* sh)
{
  SH2addi(sh, 10, (14<<4)+2);
}
#define init_7_10_14_3
void FASTCALL call_7_10_14_3(SH2_struct* sh)
{
  SH2addi(sh, 10, (14<<4)+3);
}
#define init_7_10_14_4
void FASTCALL call_7_10_14_4(SH2_struct* sh)
{
  SH2addi(sh, 10, (14<<4)+4);
}
#define init_7_10_14_5
void FASTCALL call_7_10_14_5(SH2_struct* sh)
{
  SH2addi(sh, 10, (14<<4)+5);
}
#define init_7_10_14_6
void FASTCALL call_7_10_14_6(SH2_struct* sh)
{
  SH2addi(sh, 10, (14<<4)+6);
}
#define init_7_10_14_7
void FASTCALL call_7_10_14_7(SH2_struct* sh)
{
  SH2addi(sh, 10, (14<<4)+7);
}
#define init_7_10_14_8
void FASTCALL call_7_10_14_8(SH2_struct* sh)
{
  SH2addi(sh, 10, (14<<4)+8);
}
#define init_7_10_14_9
void FASTCALL call_7_10_14_9(SH2_struct* sh)
{
  SH2addi(sh, 10, (14<<4)+9);
}
#define init_7_10_14_10
void FASTCALL call_7_10_14_10(SH2_struct* sh)
{
  SH2addi(sh, 10, (14<<4)+10);
}
#define init_7_10_14_11
void FASTCALL call_7_10_14_11(SH2_struct* sh)
{
  SH2addi(sh, 10, (14<<4)+11);
}
#define init_7_10_14_12
void FASTCALL call_7_10_14_12(SH2_struct* sh)
{
  SH2addi(sh, 10, (14<<4)+12);
}
#define init_7_10_14_13
void FASTCALL call_7_10_14_13(SH2_struct* sh)
{
  SH2addi(sh, 10, (14<<4)+13);
}
#define init_7_10_14_14
void FASTCALL call_7_10_14_14(SH2_struct* sh)
{
  SH2addi(sh, 10, (14<<4)+14);
}
#define init_7_10_14_15
void FASTCALL call_7_10_14_15(SH2_struct* sh)
{
  SH2addi(sh, 10, (14<<4)+15);
}
#define init_7_10_15_0
void FASTCALL call_7_10_15_0(SH2_struct* sh)
{
  SH2addi(sh, 10, (15<<4)+0);
}
#define init_7_10_15_1
void FASTCALL call_7_10_15_1(SH2_struct* sh)
{
  SH2addi(sh, 10, (15<<4)+1);
}
#define init_7_10_15_2
void FASTCALL call_7_10_15_2(SH2_struct* sh)
{
  SH2addi(sh, 10, (15<<4)+2);
}
#define init_7_10_15_3
void FASTCALL call_7_10_15_3(SH2_struct* sh)
{
  SH2addi(sh, 10, (15<<4)+3);
}
#define init_7_10_15_4
void FASTCALL call_7_10_15_4(SH2_struct* sh)
{
  SH2addi(sh, 10, (15<<4)+4);
}
#define init_7_10_15_5
void FASTCALL call_7_10_15_5(SH2_struct* sh)
{
  SH2addi(sh, 10, (15<<4)+5);
}
#define init_7_10_15_6
void FASTCALL call_7_10_15_6(SH2_struct* sh)
{
  SH2addi(sh, 10, (15<<4)+6);
}
#define init_7_10_15_7
void FASTCALL call_7_10_15_7(SH2_struct* sh)
{
  SH2addi(sh, 10, (15<<4)+7);
}
#define init_7_10_15_8
void FASTCALL call_7_10_15_8(SH2_struct* sh)
{
  SH2addi(sh, 10, (15<<4)+8);
}
#define init_7_10_15_9
void FASTCALL call_7_10_15_9(SH2_struct* sh)
{
  SH2addi(sh, 10, (15<<4)+9);
}
#define init_7_10_15_10
void FASTCALL call_7_10_15_10(SH2_struct* sh)
{
  SH2addi(sh, 10, (15<<4)+10);
}
#define init_7_10_15_11
void FASTCALL call_7_10_15_11(SH2_struct* sh)
{
  SH2addi(sh, 10, (15<<4)+11);
}
#define init_7_10_15_12
void FASTCALL call_7_10_15_12(SH2_struct* sh)
{
  SH2addi(sh, 10, (15<<4)+12);
}
#define init_7_10_15_13
void FASTCALL call_7_10_15_13(SH2_struct* sh)
{
  SH2addi(sh, 10, (15<<4)+13);
}
#define init_7_10_15_14
void FASTCALL call_7_10_15_14(SH2_struct* sh)
{
  SH2addi(sh, 10, (15<<4)+14);
}
#define init_7_10_15_15
void FASTCALL call_7_10_15_15(SH2_struct* sh)
{
  SH2addi(sh, 10, (15<<4)+15);
}
#define init_7_11_0_0
void FASTCALL call_7_11_0_0(SH2_struct* sh)
{
  SH2addi(sh, 11, (0<<4)+0);
}
#define init_7_11_0_1
void FASTCALL call_7_11_0_1(SH2_struct* sh)
{
  SH2addi(sh, 11, (0<<4)+1);
}
#define init_7_11_0_2
void FASTCALL call_7_11_0_2(SH2_struct* sh)
{
  SH2addi(sh, 11, (0<<4)+2);
}
#define init_7_11_0_3
void FASTCALL call_7_11_0_3(SH2_struct* sh)
{
  SH2addi(sh, 11, (0<<4)+3);
}
#define init_7_11_0_4
void FASTCALL call_7_11_0_4(SH2_struct* sh)
{
  SH2addi(sh, 11, (0<<4)+4);
}
#define init_7_11_0_5
void FASTCALL call_7_11_0_5(SH2_struct* sh)
{
  SH2addi(sh, 11, (0<<4)+5);
}
#define init_7_11_0_6
void FASTCALL call_7_11_0_6(SH2_struct* sh)
{
  SH2addi(sh, 11, (0<<4)+6);
}
#define init_7_11_0_7
void FASTCALL call_7_11_0_7(SH2_struct* sh)
{
  SH2addi(sh, 11, (0<<4)+7);
}
#define init_7_11_0_8
void FASTCALL call_7_11_0_8(SH2_struct* sh)
{
  SH2addi(sh, 11, (0<<4)+8);
}
#define init_7_11_0_9
void FASTCALL call_7_11_0_9(SH2_struct* sh)
{
  SH2addi(sh, 11, (0<<4)+9);
}
#define init_7_11_0_10
void FASTCALL call_7_11_0_10(SH2_struct* sh)
{
  SH2addi(sh, 11, (0<<4)+10);
}
#define init_7_11_0_11
void FASTCALL call_7_11_0_11(SH2_struct* sh)
{
  SH2addi(sh, 11, (0<<4)+11);
}
#define init_7_11_0_12
void FASTCALL call_7_11_0_12(SH2_struct* sh)
{
  SH2addi(sh, 11, (0<<4)+12);
}
#define init_7_11_0_13
void FASTCALL call_7_11_0_13(SH2_struct* sh)
{
  SH2addi(sh, 11, (0<<4)+13);
}
#define init_7_11_0_14
void FASTCALL call_7_11_0_14(SH2_struct* sh)
{
  SH2addi(sh, 11, (0<<4)+14);
}
#define init_7_11_0_15
void FASTCALL call_7_11_0_15(SH2_struct* sh)
{
  SH2addi(sh, 11, (0<<4)+15);
}
#define init_7_11_1_0
void FASTCALL call_7_11_1_0(SH2_struct* sh)
{
  SH2addi(sh, 11, (1<<4)+0);
}
#define init_7_11_1_1
void FASTCALL call_7_11_1_1(SH2_struct* sh)
{
  SH2addi(sh, 11, (1<<4)+1);
}
#define init_7_11_1_2
void FASTCALL call_7_11_1_2(SH2_struct* sh)
{
  SH2addi(sh, 11, (1<<4)+2);
}
#define init_7_11_1_3
void FASTCALL call_7_11_1_3(SH2_struct* sh)
{
  SH2addi(sh, 11, (1<<4)+3);
}
#define init_7_11_1_4
void FASTCALL call_7_11_1_4(SH2_struct* sh)
{
  SH2addi(sh, 11, (1<<4)+4);
}
#define init_7_11_1_5
void FASTCALL call_7_11_1_5(SH2_struct* sh)
{
  SH2addi(sh, 11, (1<<4)+5);
}
#define init_7_11_1_6
void FASTCALL call_7_11_1_6(SH2_struct* sh)
{
  SH2addi(sh, 11, (1<<4)+6);
}
#define init_7_11_1_7
void FASTCALL call_7_11_1_7(SH2_struct* sh)
{
  SH2addi(sh, 11, (1<<4)+7);
}
#define init_7_11_1_8
void FASTCALL call_7_11_1_8(SH2_struct* sh)
{
  SH2addi(sh, 11, (1<<4)+8);
}
#define init_7_11_1_9
void FASTCALL call_7_11_1_9(SH2_struct* sh)
{
  SH2addi(sh, 11, (1<<4)+9);
}
#define init_7_11_1_10
void FASTCALL call_7_11_1_10(SH2_struct* sh)
{
  SH2addi(sh, 11, (1<<4)+10);
}
#define init_7_11_1_11
void FASTCALL call_7_11_1_11(SH2_struct* sh)
{
  SH2addi(sh, 11, (1<<4)+11);
}
#define init_7_11_1_12
void FASTCALL call_7_11_1_12(SH2_struct* sh)
{
  SH2addi(sh, 11, (1<<4)+12);
}
#define init_7_11_1_13
void FASTCALL call_7_11_1_13(SH2_struct* sh)
{
  SH2addi(sh, 11, (1<<4)+13);
}
#define init_7_11_1_14
void FASTCALL call_7_11_1_14(SH2_struct* sh)
{
  SH2addi(sh, 11, (1<<4)+14);
}
#define init_7_11_1_15
void FASTCALL call_7_11_1_15(SH2_struct* sh)
{
  SH2addi(sh, 11, (1<<4)+15);
}
#define init_7_11_2_0
void FASTCALL call_7_11_2_0(SH2_struct* sh)
{
  SH2addi(sh, 11, (2<<4)+0);
}
#define init_7_11_2_1
void FASTCALL call_7_11_2_1(SH2_struct* sh)
{
  SH2addi(sh, 11, (2<<4)+1);
}
#define init_7_11_2_2
void FASTCALL call_7_11_2_2(SH2_struct* sh)
{
  SH2addi(sh, 11, (2<<4)+2);
}
#define init_7_11_2_3
void FASTCALL call_7_11_2_3(SH2_struct* sh)
{
  SH2addi(sh, 11, (2<<4)+3);
}
#define init_7_11_2_4
void FASTCALL call_7_11_2_4(SH2_struct* sh)
{
  SH2addi(sh, 11, (2<<4)+4);
}
#define init_7_11_2_5
void FASTCALL call_7_11_2_5(SH2_struct* sh)
{
  SH2addi(sh, 11, (2<<4)+5);
}
#define init_7_11_2_6
void FASTCALL call_7_11_2_6(SH2_struct* sh)
{
  SH2addi(sh, 11, (2<<4)+6);
}
#define init_7_11_2_7
void FASTCALL call_7_11_2_7(SH2_struct* sh)
{
  SH2addi(sh, 11, (2<<4)+7);
}
#define init_7_11_2_8
void FASTCALL call_7_11_2_8(SH2_struct* sh)
{
  SH2addi(sh, 11, (2<<4)+8);
}
#define init_7_11_2_9
void FASTCALL call_7_11_2_9(SH2_struct* sh)
{
  SH2addi(sh, 11, (2<<4)+9);
}
#define init_7_11_2_10
void FASTCALL call_7_11_2_10(SH2_struct* sh)
{
  SH2addi(sh, 11, (2<<4)+10);
}
#define init_7_11_2_11
void FASTCALL call_7_11_2_11(SH2_struct* sh)
{
  SH2addi(sh, 11, (2<<4)+11);
}
#define init_7_11_2_12
void FASTCALL call_7_11_2_12(SH2_struct* sh)
{
  SH2addi(sh, 11, (2<<4)+12);
}
#define init_7_11_2_13
void FASTCALL call_7_11_2_13(SH2_struct* sh)
{
  SH2addi(sh, 11, (2<<4)+13);
}
#define init_7_11_2_14
void FASTCALL call_7_11_2_14(SH2_struct* sh)
{
  SH2addi(sh, 11, (2<<4)+14);
}
#define init_7_11_2_15
void FASTCALL call_7_11_2_15(SH2_struct* sh)
{
  SH2addi(sh, 11, (2<<4)+15);
}
#define init_7_11_3_0
void FASTCALL call_7_11_3_0(SH2_struct* sh)
{
  SH2addi(sh, 11, (3<<4)+0);
}
#define init_7_11_3_1
void FASTCALL call_7_11_3_1(SH2_struct* sh)
{
  SH2addi(sh, 11, (3<<4)+1);
}
#define init_7_11_3_2
void FASTCALL call_7_11_3_2(SH2_struct* sh)
{
  SH2addi(sh, 11, (3<<4)+2);
}
#define init_7_11_3_3
void FASTCALL call_7_11_3_3(SH2_struct* sh)
{
  SH2addi(sh, 11, (3<<4)+3);
}
#define init_7_11_3_4
void FASTCALL call_7_11_3_4(SH2_struct* sh)
{
  SH2addi(sh, 11, (3<<4)+4);
}
#define init_7_11_3_5
void FASTCALL call_7_11_3_5(SH2_struct* sh)
{
  SH2addi(sh, 11, (3<<4)+5);
}
#define init_7_11_3_6
void FASTCALL call_7_11_3_6(SH2_struct* sh)
{
  SH2addi(sh, 11, (3<<4)+6);
}
#define init_7_11_3_7
void FASTCALL call_7_11_3_7(SH2_struct* sh)
{
  SH2addi(sh, 11, (3<<4)+7);
}
#define init_7_11_3_8
void FASTCALL call_7_11_3_8(SH2_struct* sh)
{
  SH2addi(sh, 11, (3<<4)+8);
}
#define init_7_11_3_9
void FASTCALL call_7_11_3_9(SH2_struct* sh)
{
  SH2addi(sh, 11, (3<<4)+9);
}
#define init_7_11_3_10
void FASTCALL call_7_11_3_10(SH2_struct* sh)
{
  SH2addi(sh, 11, (3<<4)+10);
}
#define init_7_11_3_11
void FASTCALL call_7_11_3_11(SH2_struct* sh)
{
  SH2addi(sh, 11, (3<<4)+11);
}
#define init_7_11_3_12
void FASTCALL call_7_11_3_12(SH2_struct* sh)
{
  SH2addi(sh, 11, (3<<4)+12);
}
#define init_7_11_3_13
void FASTCALL call_7_11_3_13(SH2_struct* sh)
{
  SH2addi(sh, 11, (3<<4)+13);
}
#define init_7_11_3_14
void FASTCALL call_7_11_3_14(SH2_struct* sh)
{
  SH2addi(sh, 11, (3<<4)+14);
}
#define init_7_11_3_15
void FASTCALL call_7_11_3_15(SH2_struct* sh)
{
  SH2addi(sh, 11, (3<<4)+15);
}
#define init_7_11_4_0
void FASTCALL call_7_11_4_0(SH2_struct* sh)
{
  SH2addi(sh, 11, (4<<4)+0);
}
#define init_7_11_4_1
void FASTCALL call_7_11_4_1(SH2_struct* sh)
{
  SH2addi(sh, 11, (4<<4)+1);
}
#define init_7_11_4_2
void FASTCALL call_7_11_4_2(SH2_struct* sh)
{
  SH2addi(sh, 11, (4<<4)+2);
}
#define init_7_11_4_3
void FASTCALL call_7_11_4_3(SH2_struct* sh)
{
  SH2addi(sh, 11, (4<<4)+3);
}
#define init_7_11_4_4
void FASTCALL call_7_11_4_4(SH2_struct* sh)
{
  SH2addi(sh, 11, (4<<4)+4);
}
#define init_7_11_4_5
void FASTCALL call_7_11_4_5(SH2_struct* sh)
{
  SH2addi(sh, 11, (4<<4)+5);
}
#define init_7_11_4_6
void FASTCALL call_7_11_4_6(SH2_struct* sh)
{
  SH2addi(sh, 11, (4<<4)+6);
}
#define init_7_11_4_7
void FASTCALL call_7_11_4_7(SH2_struct* sh)
{
  SH2addi(sh, 11, (4<<4)+7);
}
#define init_7_11_4_8
void FASTCALL call_7_11_4_8(SH2_struct* sh)
{
  SH2addi(sh, 11, (4<<4)+8);
}
#define init_7_11_4_9
void FASTCALL call_7_11_4_9(SH2_struct* sh)
{
  SH2addi(sh, 11, (4<<4)+9);
}
#define init_7_11_4_10
void FASTCALL call_7_11_4_10(SH2_struct* sh)
{
  SH2addi(sh, 11, (4<<4)+10);
}
#define init_7_11_4_11
void FASTCALL call_7_11_4_11(SH2_struct* sh)
{
  SH2addi(sh, 11, (4<<4)+11);
}
#define init_7_11_4_12
void FASTCALL call_7_11_4_12(SH2_struct* sh)
{
  SH2addi(sh, 11, (4<<4)+12);
}
#define init_7_11_4_13
void FASTCALL call_7_11_4_13(SH2_struct* sh)
{
  SH2addi(sh, 11, (4<<4)+13);
}
#define init_7_11_4_14
void FASTCALL call_7_11_4_14(SH2_struct* sh)
{
  SH2addi(sh, 11, (4<<4)+14);
}
#define init_7_11_4_15
void FASTCALL call_7_11_4_15(SH2_struct* sh)
{
  SH2addi(sh, 11, (4<<4)+15);
}
#define init_7_11_5_0
void FASTCALL call_7_11_5_0(SH2_struct* sh)
{
  SH2addi(sh, 11, (5<<4)+0);
}
#define init_7_11_5_1
void FASTCALL call_7_11_5_1(SH2_struct* sh)
{
  SH2addi(sh, 11, (5<<4)+1);
}
#define init_7_11_5_2
void FASTCALL call_7_11_5_2(SH2_struct* sh)
{
  SH2addi(sh, 11, (5<<4)+2);
}
#define init_7_11_5_3
void FASTCALL call_7_11_5_3(SH2_struct* sh)
{
  SH2addi(sh, 11, (5<<4)+3);
}
#define init_7_11_5_4
void FASTCALL call_7_11_5_4(SH2_struct* sh)
{
  SH2addi(sh, 11, (5<<4)+4);
}
#define init_7_11_5_5
void FASTCALL call_7_11_5_5(SH2_struct* sh)
{
  SH2addi(sh, 11, (5<<4)+5);
}
#define init_7_11_5_6
void FASTCALL call_7_11_5_6(SH2_struct* sh)
{
  SH2addi(sh, 11, (5<<4)+6);
}
#define init_7_11_5_7
void FASTCALL call_7_11_5_7(SH2_struct* sh)
{
  SH2addi(sh, 11, (5<<4)+7);
}
#define init_7_11_5_8
void FASTCALL call_7_11_5_8(SH2_struct* sh)
{
  SH2addi(sh, 11, (5<<4)+8);
}
#define init_7_11_5_9
void FASTCALL call_7_11_5_9(SH2_struct* sh)
{
  SH2addi(sh, 11, (5<<4)+9);
}
#define init_7_11_5_10
void FASTCALL call_7_11_5_10(SH2_struct* sh)
{
  SH2addi(sh, 11, (5<<4)+10);
}
#define init_7_11_5_11
void FASTCALL call_7_11_5_11(SH2_struct* sh)
{
  SH2addi(sh, 11, (5<<4)+11);
}
#define init_7_11_5_12
void FASTCALL call_7_11_5_12(SH2_struct* sh)
{
  SH2addi(sh, 11, (5<<4)+12);
}
#define init_7_11_5_13
void FASTCALL call_7_11_5_13(SH2_struct* sh)
{
  SH2addi(sh, 11, (5<<4)+13);
}
#define init_7_11_5_14
void FASTCALL call_7_11_5_14(SH2_struct* sh)
{
  SH2addi(sh, 11, (5<<4)+14);
}
#define init_7_11_5_15
void FASTCALL call_7_11_5_15(SH2_struct* sh)
{
  SH2addi(sh, 11, (5<<4)+15);
}
#define init_7_11_6_0
void FASTCALL call_7_11_6_0(SH2_struct* sh)
{
  SH2addi(sh, 11, (6<<4)+0);
}
#define init_7_11_6_1
void FASTCALL call_7_11_6_1(SH2_struct* sh)
{
  SH2addi(sh, 11, (6<<4)+1);
}
#define init_7_11_6_2
void FASTCALL call_7_11_6_2(SH2_struct* sh)
{
  SH2addi(sh, 11, (6<<4)+2);
}
#define init_7_11_6_3
void FASTCALL call_7_11_6_3(SH2_struct* sh)
{
  SH2addi(sh, 11, (6<<4)+3);
}
#define init_7_11_6_4
void FASTCALL call_7_11_6_4(SH2_struct* sh)
{
  SH2addi(sh, 11, (6<<4)+4);
}
#define init_7_11_6_5
void FASTCALL call_7_11_6_5(SH2_struct* sh)
{
  SH2addi(sh, 11, (6<<4)+5);
}
#define init_7_11_6_6
void FASTCALL call_7_11_6_6(SH2_struct* sh)
{
  SH2addi(sh, 11, (6<<4)+6);
}
#define init_7_11_6_7
void FASTCALL call_7_11_6_7(SH2_struct* sh)
{
  SH2addi(sh, 11, (6<<4)+7);
}
#define init_7_11_6_8
void FASTCALL call_7_11_6_8(SH2_struct* sh)
{
  SH2addi(sh, 11, (6<<4)+8);
}
#define init_7_11_6_9
void FASTCALL call_7_11_6_9(SH2_struct* sh)
{
  SH2addi(sh, 11, (6<<4)+9);
}
#define init_7_11_6_10
void FASTCALL call_7_11_6_10(SH2_struct* sh)
{
  SH2addi(sh, 11, (6<<4)+10);
}
#define init_7_11_6_11
void FASTCALL call_7_11_6_11(SH2_struct* sh)
{
  SH2addi(sh, 11, (6<<4)+11);
}
#define init_7_11_6_12
void FASTCALL call_7_11_6_12(SH2_struct* sh)
{
  SH2addi(sh, 11, (6<<4)+12);
}
#define init_7_11_6_13
void FASTCALL call_7_11_6_13(SH2_struct* sh)
{
  SH2addi(sh, 11, (6<<4)+13);
}
#define init_7_11_6_14
void FASTCALL call_7_11_6_14(SH2_struct* sh)
{
  SH2addi(sh, 11, (6<<4)+14);
}
#define init_7_11_6_15
void FASTCALL call_7_11_6_15(SH2_struct* sh)
{
  SH2addi(sh, 11, (6<<4)+15);
}
#define init_7_11_7_0
void FASTCALL call_7_11_7_0(SH2_struct* sh)
{
  SH2addi(sh, 11, (7<<4)+0);
}
#define init_7_11_7_1
void FASTCALL call_7_11_7_1(SH2_struct* sh)
{
  SH2addi(sh, 11, (7<<4)+1);
}
#define init_7_11_7_2
void FASTCALL call_7_11_7_2(SH2_struct* sh)
{
  SH2addi(sh, 11, (7<<4)+2);
}
#define init_7_11_7_3
void FASTCALL call_7_11_7_3(SH2_struct* sh)
{
  SH2addi(sh, 11, (7<<4)+3);
}
#define init_7_11_7_4
void FASTCALL call_7_11_7_4(SH2_struct* sh)
{
  SH2addi(sh, 11, (7<<4)+4);
}
#define init_7_11_7_5
void FASTCALL call_7_11_7_5(SH2_struct* sh)
{
  SH2addi(sh, 11, (7<<4)+5);
}
#define init_7_11_7_6
void FASTCALL call_7_11_7_6(SH2_struct* sh)
{
  SH2addi(sh, 11, (7<<4)+6);
}
#define init_7_11_7_7
void FASTCALL call_7_11_7_7(SH2_struct* sh)
{
  SH2addi(sh, 11, (7<<4)+7);
}
#define init_7_11_7_8
void FASTCALL call_7_11_7_8(SH2_struct* sh)
{
  SH2addi(sh, 11, (7<<4)+8);
}
#define init_7_11_7_9
void FASTCALL call_7_11_7_9(SH2_struct* sh)
{
  SH2addi(sh, 11, (7<<4)+9);
}
#define init_7_11_7_10
void FASTCALL call_7_11_7_10(SH2_struct* sh)
{
  SH2addi(sh, 11, (7<<4)+10);
}
#define init_7_11_7_11
void FASTCALL call_7_11_7_11(SH2_struct* sh)
{
  SH2addi(sh, 11, (7<<4)+11);
}
#define init_7_11_7_12
void FASTCALL call_7_11_7_12(SH2_struct* sh)
{
  SH2addi(sh, 11, (7<<4)+12);
}
#define init_7_11_7_13
void FASTCALL call_7_11_7_13(SH2_struct* sh)
{
  SH2addi(sh, 11, (7<<4)+13);
}
#define init_7_11_7_14
void FASTCALL call_7_11_7_14(SH2_struct* sh)
{
  SH2addi(sh, 11, (7<<4)+14);
}
#define init_7_11_7_15
void FASTCALL call_7_11_7_15(SH2_struct* sh)
{
  SH2addi(sh, 11, (7<<4)+15);
}
#define init_7_11_8_0
void FASTCALL call_7_11_8_0(SH2_struct* sh)
{
  SH2addi(sh, 11, (8<<4)+0);
}
#define init_7_11_8_1
void FASTCALL call_7_11_8_1(SH2_struct* sh)
{
  SH2addi(sh, 11, (8<<4)+1);
}
#define init_7_11_8_2
void FASTCALL call_7_11_8_2(SH2_struct* sh)
{
  SH2addi(sh, 11, (8<<4)+2);
}
#define init_7_11_8_3
void FASTCALL call_7_11_8_3(SH2_struct* sh)
{
  SH2addi(sh, 11, (8<<4)+3);
}
#define init_7_11_8_4
void FASTCALL call_7_11_8_4(SH2_struct* sh)
{
  SH2addi(sh, 11, (8<<4)+4);
}
#define init_7_11_8_5
void FASTCALL call_7_11_8_5(SH2_struct* sh)
{
  SH2addi(sh, 11, (8<<4)+5);
}
#define init_7_11_8_6
void FASTCALL call_7_11_8_6(SH2_struct* sh)
{
  SH2addi(sh, 11, (8<<4)+6);
}
#define init_7_11_8_7
void FASTCALL call_7_11_8_7(SH2_struct* sh)
{
  SH2addi(sh, 11, (8<<4)+7);
}
#define init_7_11_8_8
void FASTCALL call_7_11_8_8(SH2_struct* sh)
{
  SH2addi(sh, 11, (8<<4)+8);
}
#define init_7_11_8_9
void FASTCALL call_7_11_8_9(SH2_struct* sh)
{
  SH2addi(sh, 11, (8<<4)+9);
}
#define init_7_11_8_10
void FASTCALL call_7_11_8_10(SH2_struct* sh)
{
  SH2addi(sh, 11, (8<<4)+10);
}
#define init_7_11_8_11
void FASTCALL call_7_11_8_11(SH2_struct* sh)
{
  SH2addi(sh, 11, (8<<4)+11);
}
#define init_7_11_8_12
void FASTCALL call_7_11_8_12(SH2_struct* sh)
{
  SH2addi(sh, 11, (8<<4)+12);
}
#define init_7_11_8_13
void FASTCALL call_7_11_8_13(SH2_struct* sh)
{
  SH2addi(sh, 11, (8<<4)+13);
}
#define init_7_11_8_14
void FASTCALL call_7_11_8_14(SH2_struct* sh)
{
  SH2addi(sh, 11, (8<<4)+14);
}
#define init_7_11_8_15
void FASTCALL call_7_11_8_15(SH2_struct* sh)
{
  SH2addi(sh, 11, (8<<4)+15);
}
#define init_7_11_9_0
void FASTCALL call_7_11_9_0(SH2_struct* sh)
{
  SH2addi(sh, 11, (9<<4)+0);
}
#define init_7_11_9_1
void FASTCALL call_7_11_9_1(SH2_struct* sh)
{
  SH2addi(sh, 11, (9<<4)+1);
}
#define init_7_11_9_2
void FASTCALL call_7_11_9_2(SH2_struct* sh)
{
  SH2addi(sh, 11, (9<<4)+2);
}
#define init_7_11_9_3
void FASTCALL call_7_11_9_3(SH2_struct* sh)
{
  SH2addi(sh, 11, (9<<4)+3);
}
#define init_7_11_9_4
void FASTCALL call_7_11_9_4(SH2_struct* sh)
{
  SH2addi(sh, 11, (9<<4)+4);
}
#define init_7_11_9_5
void FASTCALL call_7_11_9_5(SH2_struct* sh)
{
  SH2addi(sh, 11, (9<<4)+5);
}
#define init_7_11_9_6
void FASTCALL call_7_11_9_6(SH2_struct* sh)
{
  SH2addi(sh, 11, (9<<4)+6);
}
#define init_7_11_9_7
void FASTCALL call_7_11_9_7(SH2_struct* sh)
{
  SH2addi(sh, 11, (9<<4)+7);
}
#define init_7_11_9_8
void FASTCALL call_7_11_9_8(SH2_struct* sh)
{
  SH2addi(sh, 11, (9<<4)+8);
}
#define init_7_11_9_9
void FASTCALL call_7_11_9_9(SH2_struct* sh)
{
  SH2addi(sh, 11, (9<<4)+9);
}
#define init_7_11_9_10
void FASTCALL call_7_11_9_10(SH2_struct* sh)
{
  SH2addi(sh, 11, (9<<4)+10);
}
#define init_7_11_9_11
void FASTCALL call_7_11_9_11(SH2_struct* sh)
{
  SH2addi(sh, 11, (9<<4)+11);
}
#define init_7_11_9_12
void FASTCALL call_7_11_9_12(SH2_struct* sh)
{
  SH2addi(sh, 11, (9<<4)+12);
}
#define init_7_11_9_13
void FASTCALL call_7_11_9_13(SH2_struct* sh)
{
  SH2addi(sh, 11, (9<<4)+13);
}
#define init_7_11_9_14
void FASTCALL call_7_11_9_14(SH2_struct* sh)
{
  SH2addi(sh, 11, (9<<4)+14);
}
#define init_7_11_9_15
void FASTCALL call_7_11_9_15(SH2_struct* sh)
{
  SH2addi(sh, 11, (9<<4)+15);
}
#define init_7_11_10_0
void FASTCALL call_7_11_10_0(SH2_struct* sh)
{
  SH2addi(sh, 11, (10<<4)+0);
}
#define init_7_11_10_1
void FASTCALL call_7_11_10_1(SH2_struct* sh)
{
  SH2addi(sh, 11, (10<<4)+1);
}
#define init_7_11_10_2
void FASTCALL call_7_11_10_2(SH2_struct* sh)
{
  SH2addi(sh, 11, (10<<4)+2);
}
#define init_7_11_10_3
void FASTCALL call_7_11_10_3(SH2_struct* sh)
{
  SH2addi(sh, 11, (10<<4)+3);
}
#define init_7_11_10_4
void FASTCALL call_7_11_10_4(SH2_struct* sh)
{
  SH2addi(sh, 11, (10<<4)+4);
}
#define init_7_11_10_5
void FASTCALL call_7_11_10_5(SH2_struct* sh)
{
  SH2addi(sh, 11, (10<<4)+5);
}
#define init_7_11_10_6
void FASTCALL call_7_11_10_6(SH2_struct* sh)
{
  SH2addi(sh, 11, (10<<4)+6);
}
#define init_7_11_10_7
void FASTCALL call_7_11_10_7(SH2_struct* sh)
{
  SH2addi(sh, 11, (10<<4)+7);
}
#define init_7_11_10_8
void FASTCALL call_7_11_10_8(SH2_struct* sh)
{
  SH2addi(sh, 11, (10<<4)+8);
}
#define init_7_11_10_9
void FASTCALL call_7_11_10_9(SH2_struct* sh)
{
  SH2addi(sh, 11, (10<<4)+9);
}
#define init_7_11_10_10
void FASTCALL call_7_11_10_10(SH2_struct* sh)
{
  SH2addi(sh, 11, (10<<4)+10);
}
#define init_7_11_10_11
void FASTCALL call_7_11_10_11(SH2_struct* sh)
{
  SH2addi(sh, 11, (10<<4)+11);
}
#define init_7_11_10_12
void FASTCALL call_7_11_10_12(SH2_struct* sh)
{
  SH2addi(sh, 11, (10<<4)+12);
}
#define init_7_11_10_13
void FASTCALL call_7_11_10_13(SH2_struct* sh)
{
  SH2addi(sh, 11, (10<<4)+13);
}
#define init_7_11_10_14
void FASTCALL call_7_11_10_14(SH2_struct* sh)
{
  SH2addi(sh, 11, (10<<4)+14);
}
#define init_7_11_10_15
void FASTCALL call_7_11_10_15(SH2_struct* sh)
{
  SH2addi(sh, 11, (10<<4)+15);
}
#define init_7_11_11_0
void FASTCALL call_7_11_11_0(SH2_struct* sh)
{
  SH2addi(sh, 11, (11<<4)+0);
}
#define init_7_11_11_1
void FASTCALL call_7_11_11_1(SH2_struct* sh)
{
  SH2addi(sh, 11, (11<<4)+1);
}
#define init_7_11_11_2
void FASTCALL call_7_11_11_2(SH2_struct* sh)
{
  SH2addi(sh, 11, (11<<4)+2);
}
#define init_7_11_11_3
void FASTCALL call_7_11_11_3(SH2_struct* sh)
{
  SH2addi(sh, 11, (11<<4)+3);
}
#define init_7_11_11_4
void FASTCALL call_7_11_11_4(SH2_struct* sh)
{
  SH2addi(sh, 11, (11<<4)+4);
}
#define init_7_11_11_5
void FASTCALL call_7_11_11_5(SH2_struct* sh)
{
  SH2addi(sh, 11, (11<<4)+5);
}
#define init_7_11_11_6
void FASTCALL call_7_11_11_6(SH2_struct* sh)
{
  SH2addi(sh, 11, (11<<4)+6);
}
#define init_7_11_11_7
void FASTCALL call_7_11_11_7(SH2_struct* sh)
{
  SH2addi(sh, 11, (11<<4)+7);
}
#define init_7_11_11_8
void FASTCALL call_7_11_11_8(SH2_struct* sh)
{
  SH2addi(sh, 11, (11<<4)+8);
}
#define init_7_11_11_9
void FASTCALL call_7_11_11_9(SH2_struct* sh)
{
  SH2addi(sh, 11, (11<<4)+9);
}
#define init_7_11_11_10
void FASTCALL call_7_11_11_10(SH2_struct* sh)
{
  SH2addi(sh, 11, (11<<4)+10);
}
#define init_7_11_11_11
void FASTCALL call_7_11_11_11(SH2_struct* sh)
{
  SH2addi(sh, 11, (11<<4)+11);
}
#define init_7_11_11_12
void FASTCALL call_7_11_11_12(SH2_struct* sh)
{
  SH2addi(sh, 11, (11<<4)+12);
}
#define init_7_11_11_13
void FASTCALL call_7_11_11_13(SH2_struct* sh)
{
  SH2addi(sh, 11, (11<<4)+13);
}
#define init_7_11_11_14
void FASTCALL call_7_11_11_14(SH2_struct* sh)
{
  SH2addi(sh, 11, (11<<4)+14);
}
#define init_7_11_11_15
void FASTCALL call_7_11_11_15(SH2_struct* sh)
{
  SH2addi(sh, 11, (11<<4)+15);
}
#define init_7_11_12_0
void FASTCALL call_7_11_12_0(SH2_struct* sh)
{
  SH2addi(sh, 11, (12<<4)+0);
}
#define init_7_11_12_1
void FASTCALL call_7_11_12_1(SH2_struct* sh)
{
  SH2addi(sh, 11, (12<<4)+1);
}
#define init_7_11_12_2
void FASTCALL call_7_11_12_2(SH2_struct* sh)
{
  SH2addi(sh, 11, (12<<4)+2);
}
#define init_7_11_12_3
void FASTCALL call_7_11_12_3(SH2_struct* sh)
{
  SH2addi(sh, 11, (12<<4)+3);
}
#define init_7_11_12_4
void FASTCALL call_7_11_12_4(SH2_struct* sh)
{
  SH2addi(sh, 11, (12<<4)+4);
}
#define init_7_11_12_5
void FASTCALL call_7_11_12_5(SH2_struct* sh)
{
  SH2addi(sh, 11, (12<<4)+5);
}
#define init_7_11_12_6
void FASTCALL call_7_11_12_6(SH2_struct* sh)
{
  SH2addi(sh, 11, (12<<4)+6);
}
#define init_7_11_12_7
void FASTCALL call_7_11_12_7(SH2_struct* sh)
{
  SH2addi(sh, 11, (12<<4)+7);
}
#define init_7_11_12_8
void FASTCALL call_7_11_12_8(SH2_struct* sh)
{
  SH2addi(sh, 11, (12<<4)+8);
}
#define init_7_11_12_9
void FASTCALL call_7_11_12_9(SH2_struct* sh)
{
  SH2addi(sh, 11, (12<<4)+9);
}
#define init_7_11_12_10
void FASTCALL call_7_11_12_10(SH2_struct* sh)
{
  SH2addi(sh, 11, (12<<4)+10);
}
#define init_7_11_12_11
void FASTCALL call_7_11_12_11(SH2_struct* sh)
{
  SH2addi(sh, 11, (12<<4)+11);
}
#define init_7_11_12_12
void FASTCALL call_7_11_12_12(SH2_struct* sh)
{
  SH2addi(sh, 11, (12<<4)+12);
}
#define init_7_11_12_13
void FASTCALL call_7_11_12_13(SH2_struct* sh)
{
  SH2addi(sh, 11, (12<<4)+13);
}
#define init_7_11_12_14
void FASTCALL call_7_11_12_14(SH2_struct* sh)
{
  SH2addi(sh, 11, (12<<4)+14);
}
#define init_7_11_12_15
void FASTCALL call_7_11_12_15(SH2_struct* sh)
{
  SH2addi(sh, 11, (12<<4)+15);
}
#define init_7_11_13_0
void FASTCALL call_7_11_13_0(SH2_struct* sh)
{
  SH2addi(sh, 11, (13<<4)+0);
}
#define init_7_11_13_1
void FASTCALL call_7_11_13_1(SH2_struct* sh)
{
  SH2addi(sh, 11, (13<<4)+1);
}
#define init_7_11_13_2
void FASTCALL call_7_11_13_2(SH2_struct* sh)
{
  SH2addi(sh, 11, (13<<4)+2);
}
#define init_7_11_13_3
void FASTCALL call_7_11_13_3(SH2_struct* sh)
{
  SH2addi(sh, 11, (13<<4)+3);
}
#define init_7_11_13_4
void FASTCALL call_7_11_13_4(SH2_struct* sh)
{
  SH2addi(sh, 11, (13<<4)+4);
}
#define init_7_11_13_5
void FASTCALL call_7_11_13_5(SH2_struct* sh)
{
  SH2addi(sh, 11, (13<<4)+5);
}
#define init_7_11_13_6
void FASTCALL call_7_11_13_6(SH2_struct* sh)
{
  SH2addi(sh, 11, (13<<4)+6);
}
#define init_7_11_13_7
void FASTCALL call_7_11_13_7(SH2_struct* sh)
{
  SH2addi(sh, 11, (13<<4)+7);
}
#define init_7_11_13_8
void FASTCALL call_7_11_13_8(SH2_struct* sh)
{
  SH2addi(sh, 11, (13<<4)+8);
}
#define init_7_11_13_9
void FASTCALL call_7_11_13_9(SH2_struct* sh)
{
  SH2addi(sh, 11, (13<<4)+9);
}
#define init_7_11_13_10
void FASTCALL call_7_11_13_10(SH2_struct* sh)
{
  SH2addi(sh, 11, (13<<4)+10);
}
#define init_7_11_13_11
void FASTCALL call_7_11_13_11(SH2_struct* sh)
{
  SH2addi(sh, 11, (13<<4)+11);
}
#define init_7_11_13_12
void FASTCALL call_7_11_13_12(SH2_struct* sh)
{
  SH2addi(sh, 11, (13<<4)+12);
}
#define init_7_11_13_13
void FASTCALL call_7_11_13_13(SH2_struct* sh)
{
  SH2addi(sh, 11, (13<<4)+13);
}
#define init_7_11_13_14
void FASTCALL call_7_11_13_14(SH2_struct* sh)
{
  SH2addi(sh, 11, (13<<4)+14);
}
#define init_7_11_13_15
void FASTCALL call_7_11_13_15(SH2_struct* sh)
{
  SH2addi(sh, 11, (13<<4)+15);
}
#define init_7_11_14_0
void FASTCALL call_7_11_14_0(SH2_struct* sh)
{
  SH2addi(sh, 11, (14<<4)+0);
}
#define init_7_11_14_1
void FASTCALL call_7_11_14_1(SH2_struct* sh)
{
  SH2addi(sh, 11, (14<<4)+1);
}
#define init_7_11_14_2
void FASTCALL call_7_11_14_2(SH2_struct* sh)
{
  SH2addi(sh, 11, (14<<4)+2);
}
#define init_7_11_14_3
void FASTCALL call_7_11_14_3(SH2_struct* sh)
{
  SH2addi(sh, 11, (14<<4)+3);
}
#define init_7_11_14_4
void FASTCALL call_7_11_14_4(SH2_struct* sh)
{
  SH2addi(sh, 11, (14<<4)+4);
}
#define init_7_11_14_5
void FASTCALL call_7_11_14_5(SH2_struct* sh)
{
  SH2addi(sh, 11, (14<<4)+5);
}
#define init_7_11_14_6
void FASTCALL call_7_11_14_6(SH2_struct* sh)
{
  SH2addi(sh, 11, (14<<4)+6);
}
#define init_7_11_14_7
void FASTCALL call_7_11_14_7(SH2_struct* sh)
{
  SH2addi(sh, 11, (14<<4)+7);
}
#define init_7_11_14_8
void FASTCALL call_7_11_14_8(SH2_struct* sh)
{
  SH2addi(sh, 11, (14<<4)+8);
}
#define init_7_11_14_9
void FASTCALL call_7_11_14_9(SH2_struct* sh)
{
  SH2addi(sh, 11, (14<<4)+9);
}
#define init_7_11_14_10
void FASTCALL call_7_11_14_10(SH2_struct* sh)
{
  SH2addi(sh, 11, (14<<4)+10);
}
#define init_7_11_14_11
void FASTCALL call_7_11_14_11(SH2_struct* sh)
{
  SH2addi(sh, 11, (14<<4)+11);
}
#define init_7_11_14_12
void FASTCALL call_7_11_14_12(SH2_struct* sh)
{
  SH2addi(sh, 11, (14<<4)+12);
}
#define init_7_11_14_13
void FASTCALL call_7_11_14_13(SH2_struct* sh)
{
  SH2addi(sh, 11, (14<<4)+13);
}
#define init_7_11_14_14
void FASTCALL call_7_11_14_14(SH2_struct* sh)
{
  SH2addi(sh, 11, (14<<4)+14);
}
#define init_7_11_14_15
void FASTCALL call_7_11_14_15(SH2_struct* sh)
{
  SH2addi(sh, 11, (14<<4)+15);
}
#define init_7_11_15_0
void FASTCALL call_7_11_15_0(SH2_struct* sh)
{
  SH2addi(sh, 11, (15<<4)+0);
}
#define init_7_11_15_1
void FASTCALL call_7_11_15_1(SH2_struct* sh)
{
  SH2addi(sh, 11, (15<<4)+1);
}
#define init_7_11_15_2
void FASTCALL call_7_11_15_2(SH2_struct* sh)
{
  SH2addi(sh, 11, (15<<4)+2);
}
#define init_7_11_15_3
void FASTCALL call_7_11_15_3(SH2_struct* sh)
{
  SH2addi(sh, 11, (15<<4)+3);
}
#define init_7_11_15_4
void FASTCALL call_7_11_15_4(SH2_struct* sh)
{
  SH2addi(sh, 11, (15<<4)+4);
}
#define init_7_11_15_5
void FASTCALL call_7_11_15_5(SH2_struct* sh)
{
  SH2addi(sh, 11, (15<<4)+5);
}
#define init_7_11_15_6
void FASTCALL call_7_11_15_6(SH2_struct* sh)
{
  SH2addi(sh, 11, (15<<4)+6);
}
#define init_7_11_15_7
void FASTCALL call_7_11_15_7(SH2_struct* sh)
{
  SH2addi(sh, 11, (15<<4)+7);
}
#define init_7_11_15_8
void FASTCALL call_7_11_15_8(SH2_struct* sh)
{
  SH2addi(sh, 11, (15<<4)+8);
}
#define init_7_11_15_9
void FASTCALL call_7_11_15_9(SH2_struct* sh)
{
  SH2addi(sh, 11, (15<<4)+9);
}
#define init_7_11_15_10
void FASTCALL call_7_11_15_10(SH2_struct* sh)
{
  SH2addi(sh, 11, (15<<4)+10);
}
#define init_7_11_15_11
void FASTCALL call_7_11_15_11(SH2_struct* sh)
{
  SH2addi(sh, 11, (15<<4)+11);
}
#define init_7_11_15_12
void FASTCALL call_7_11_15_12(SH2_struct* sh)
{
  SH2addi(sh, 11, (15<<4)+12);
}
#define init_7_11_15_13
void FASTCALL call_7_11_15_13(SH2_struct* sh)
{
  SH2addi(sh, 11, (15<<4)+13);
}
#define init_7_11_15_14
void FASTCALL call_7_11_15_14(SH2_struct* sh)
{
  SH2addi(sh, 11, (15<<4)+14);
}
#define init_7_11_15_15
void FASTCALL call_7_11_15_15(SH2_struct* sh)
{
  SH2addi(sh, 11, (15<<4)+15);
}
#define init_7_12_0_0
void FASTCALL call_7_12_0_0(SH2_struct* sh)
{
  SH2addi(sh, 12, (0<<4)+0);
}
#define init_7_12_0_1
void FASTCALL call_7_12_0_1(SH2_struct* sh)
{
  SH2addi(sh, 12, (0<<4)+1);
}
#define init_7_12_0_2
void FASTCALL call_7_12_0_2(SH2_struct* sh)
{
  SH2addi(sh, 12, (0<<4)+2);
}
#define init_7_12_0_3
void FASTCALL call_7_12_0_3(SH2_struct* sh)
{
  SH2addi(sh, 12, (0<<4)+3);
}
#define init_7_12_0_4
void FASTCALL call_7_12_0_4(SH2_struct* sh)
{
  SH2addi(sh, 12, (0<<4)+4);
}
#define init_7_12_0_5
void FASTCALL call_7_12_0_5(SH2_struct* sh)
{
  SH2addi(sh, 12, (0<<4)+5);
}
#define init_7_12_0_6
void FASTCALL call_7_12_0_6(SH2_struct* sh)
{
  SH2addi(sh, 12, (0<<4)+6);
}
#define init_7_12_0_7
void FASTCALL call_7_12_0_7(SH2_struct* sh)
{
  SH2addi(sh, 12, (0<<4)+7);
}
#define init_7_12_0_8
void FASTCALL call_7_12_0_8(SH2_struct* sh)
{
  SH2addi(sh, 12, (0<<4)+8);
}
#define init_7_12_0_9
void FASTCALL call_7_12_0_9(SH2_struct* sh)
{
  SH2addi(sh, 12, (0<<4)+9);
}
#define init_7_12_0_10
void FASTCALL call_7_12_0_10(SH2_struct* sh)
{
  SH2addi(sh, 12, (0<<4)+10);
}
#define init_7_12_0_11
void FASTCALL call_7_12_0_11(SH2_struct* sh)
{
  SH2addi(sh, 12, (0<<4)+11);
}
#define init_7_12_0_12
void FASTCALL call_7_12_0_12(SH2_struct* sh)
{
  SH2addi(sh, 12, (0<<4)+12);
}
#define init_7_12_0_13
void FASTCALL call_7_12_0_13(SH2_struct* sh)
{
  SH2addi(sh, 12, (0<<4)+13);
}
#define init_7_12_0_14
void FASTCALL call_7_12_0_14(SH2_struct* sh)
{
  SH2addi(sh, 12, (0<<4)+14);
}
#define init_7_12_0_15
void FASTCALL call_7_12_0_15(SH2_struct* sh)
{
  SH2addi(sh, 12, (0<<4)+15);
}
#define init_7_12_1_0
void FASTCALL call_7_12_1_0(SH2_struct* sh)
{
  SH2addi(sh, 12, (1<<4)+0);
}
#define init_7_12_1_1
void FASTCALL call_7_12_1_1(SH2_struct* sh)
{
  SH2addi(sh, 12, (1<<4)+1);
}
#define init_7_12_1_2
void FASTCALL call_7_12_1_2(SH2_struct* sh)
{
  SH2addi(sh, 12, (1<<4)+2);
}
#define init_7_12_1_3
void FASTCALL call_7_12_1_3(SH2_struct* sh)
{
  SH2addi(sh, 12, (1<<4)+3);
}
#define init_7_12_1_4
void FASTCALL call_7_12_1_4(SH2_struct* sh)
{
  SH2addi(sh, 12, (1<<4)+4);
}
#define init_7_12_1_5
void FASTCALL call_7_12_1_5(SH2_struct* sh)
{
  SH2addi(sh, 12, (1<<4)+5);
}
#define init_7_12_1_6
void FASTCALL call_7_12_1_6(SH2_struct* sh)
{
  SH2addi(sh, 12, (1<<4)+6);
}
#define init_7_12_1_7
void FASTCALL call_7_12_1_7(SH2_struct* sh)
{
  SH2addi(sh, 12, (1<<4)+7);
}
#define init_7_12_1_8
void FASTCALL call_7_12_1_8(SH2_struct* sh)
{
  SH2addi(sh, 12, (1<<4)+8);
}
#define init_7_12_1_9
void FASTCALL call_7_12_1_9(SH2_struct* sh)
{
  SH2addi(sh, 12, (1<<4)+9);
}
#define init_7_12_1_10
void FASTCALL call_7_12_1_10(SH2_struct* sh)
{
  SH2addi(sh, 12, (1<<4)+10);
}
#define init_7_12_1_11
void FASTCALL call_7_12_1_11(SH2_struct* sh)
{
  SH2addi(sh, 12, (1<<4)+11);
}
#define init_7_12_1_12
void FASTCALL call_7_12_1_12(SH2_struct* sh)
{
  SH2addi(sh, 12, (1<<4)+12);
}
#define init_7_12_1_13
void FASTCALL call_7_12_1_13(SH2_struct* sh)
{
  SH2addi(sh, 12, (1<<4)+13);
}
#define init_7_12_1_14
void FASTCALL call_7_12_1_14(SH2_struct* sh)
{
  SH2addi(sh, 12, (1<<4)+14);
}
#define init_7_12_1_15
void FASTCALL call_7_12_1_15(SH2_struct* sh)
{
  SH2addi(sh, 12, (1<<4)+15);
}
#define init_7_12_2_0
void FASTCALL call_7_12_2_0(SH2_struct* sh)
{
  SH2addi(sh, 12, (2<<4)+0);
}
#define init_7_12_2_1
void FASTCALL call_7_12_2_1(SH2_struct* sh)
{
  SH2addi(sh, 12, (2<<4)+1);
}
#define init_7_12_2_2
void FASTCALL call_7_12_2_2(SH2_struct* sh)
{
  SH2addi(sh, 12, (2<<4)+2);
}
#define init_7_12_2_3
void FASTCALL call_7_12_2_3(SH2_struct* sh)
{
  SH2addi(sh, 12, (2<<4)+3);
}
#define init_7_12_2_4
void FASTCALL call_7_12_2_4(SH2_struct* sh)
{
  SH2addi(sh, 12, (2<<4)+4);
}
#define init_7_12_2_5
void FASTCALL call_7_12_2_5(SH2_struct* sh)
{
  SH2addi(sh, 12, (2<<4)+5);
}
#define init_7_12_2_6
void FASTCALL call_7_12_2_6(SH2_struct* sh)
{
  SH2addi(sh, 12, (2<<4)+6);
}
#define init_7_12_2_7
void FASTCALL call_7_12_2_7(SH2_struct* sh)
{
  SH2addi(sh, 12, (2<<4)+7);
}
#define init_7_12_2_8
void FASTCALL call_7_12_2_8(SH2_struct* sh)
{
  SH2addi(sh, 12, (2<<4)+8);
}
#define init_7_12_2_9
void FASTCALL call_7_12_2_9(SH2_struct* sh)
{
  SH2addi(sh, 12, (2<<4)+9);
}
#define init_7_12_2_10
void FASTCALL call_7_12_2_10(SH2_struct* sh)
{
  SH2addi(sh, 12, (2<<4)+10);
}
#define init_7_12_2_11
void FASTCALL call_7_12_2_11(SH2_struct* sh)
{
  SH2addi(sh, 12, (2<<4)+11);
}
#define init_7_12_2_12
void FASTCALL call_7_12_2_12(SH2_struct* sh)
{
  SH2addi(sh, 12, (2<<4)+12);
}
#define init_7_12_2_13
void FASTCALL call_7_12_2_13(SH2_struct* sh)
{
  SH2addi(sh, 12, (2<<4)+13);
}
#define init_7_12_2_14
void FASTCALL call_7_12_2_14(SH2_struct* sh)
{
  SH2addi(sh, 12, (2<<4)+14);
}
#define init_7_12_2_15
void FASTCALL call_7_12_2_15(SH2_struct* sh)
{
  SH2addi(sh, 12, (2<<4)+15);
}
#define init_7_12_3_0
void FASTCALL call_7_12_3_0(SH2_struct* sh)
{
  SH2addi(sh, 12, (3<<4)+0);
}
#define init_7_12_3_1
void FASTCALL call_7_12_3_1(SH2_struct* sh)
{
  SH2addi(sh, 12, (3<<4)+1);
}
#define init_7_12_3_2
void FASTCALL call_7_12_3_2(SH2_struct* sh)
{
  SH2addi(sh, 12, (3<<4)+2);
}
#define init_7_12_3_3
void FASTCALL call_7_12_3_3(SH2_struct* sh)
{
  SH2addi(sh, 12, (3<<4)+3);
}
#define init_7_12_3_4
void FASTCALL call_7_12_3_4(SH2_struct* sh)
{
  SH2addi(sh, 12, (3<<4)+4);
}
#define init_7_12_3_5
void FASTCALL call_7_12_3_5(SH2_struct* sh)
{
  SH2addi(sh, 12, (3<<4)+5);
}
#define init_7_12_3_6
void FASTCALL call_7_12_3_6(SH2_struct* sh)
{
  SH2addi(sh, 12, (3<<4)+6);
}
#define init_7_12_3_7
void FASTCALL call_7_12_3_7(SH2_struct* sh)
{
  SH2addi(sh, 12, (3<<4)+7);
}
#define init_7_12_3_8
void FASTCALL call_7_12_3_8(SH2_struct* sh)
{
  SH2addi(sh, 12, (3<<4)+8);
}
#define init_7_12_3_9
void FASTCALL call_7_12_3_9(SH2_struct* sh)
{
  SH2addi(sh, 12, (3<<4)+9);
}
#define init_7_12_3_10
void FASTCALL call_7_12_3_10(SH2_struct* sh)
{
  SH2addi(sh, 12, (3<<4)+10);
}
#define init_7_12_3_11
void FASTCALL call_7_12_3_11(SH2_struct* sh)
{
  SH2addi(sh, 12, (3<<4)+11);
}
#define init_7_12_3_12
void FASTCALL call_7_12_3_12(SH2_struct* sh)
{
  SH2addi(sh, 12, (3<<4)+12);
}
#define init_7_12_3_13
void FASTCALL call_7_12_3_13(SH2_struct* sh)
{
  SH2addi(sh, 12, (3<<4)+13);
}
#define init_7_12_3_14
void FASTCALL call_7_12_3_14(SH2_struct* sh)
{
  SH2addi(sh, 12, (3<<4)+14);
}
#define init_7_12_3_15
void FASTCALL call_7_12_3_15(SH2_struct* sh)
{
  SH2addi(sh, 12, (3<<4)+15);
}
#define init_7_12_4_0
void FASTCALL call_7_12_4_0(SH2_struct* sh)
{
  SH2addi(sh, 12, (4<<4)+0);
}
#define init_7_12_4_1
void FASTCALL call_7_12_4_1(SH2_struct* sh)
{
  SH2addi(sh, 12, (4<<4)+1);
}
#define init_7_12_4_2
void FASTCALL call_7_12_4_2(SH2_struct* sh)
{
  SH2addi(sh, 12, (4<<4)+2);
}
#define init_7_12_4_3
void FASTCALL call_7_12_4_3(SH2_struct* sh)
{
  SH2addi(sh, 12, (4<<4)+3);
}
#define init_7_12_4_4
void FASTCALL call_7_12_4_4(SH2_struct* sh)
{
  SH2addi(sh, 12, (4<<4)+4);
}
#define init_7_12_4_5
void FASTCALL call_7_12_4_5(SH2_struct* sh)
{
  SH2addi(sh, 12, (4<<4)+5);
}
#define init_7_12_4_6
void FASTCALL call_7_12_4_6(SH2_struct* sh)
{
  SH2addi(sh, 12, (4<<4)+6);
}
#define init_7_12_4_7
void FASTCALL call_7_12_4_7(SH2_struct* sh)
{
  SH2addi(sh, 12, (4<<4)+7);
}
#define init_7_12_4_8
void FASTCALL call_7_12_4_8(SH2_struct* sh)
{
  SH2addi(sh, 12, (4<<4)+8);
}
#define init_7_12_4_9
void FASTCALL call_7_12_4_9(SH2_struct* sh)
{
  SH2addi(sh, 12, (4<<4)+9);
}
#define init_7_12_4_10
void FASTCALL call_7_12_4_10(SH2_struct* sh)
{
  SH2addi(sh, 12, (4<<4)+10);
}
#define init_7_12_4_11
void FASTCALL call_7_12_4_11(SH2_struct* sh)
{
  SH2addi(sh, 12, (4<<4)+11);
}
#define init_7_12_4_12
void FASTCALL call_7_12_4_12(SH2_struct* sh)
{
  SH2addi(sh, 12, (4<<4)+12);
}
#define init_7_12_4_13
void FASTCALL call_7_12_4_13(SH2_struct* sh)
{
  SH2addi(sh, 12, (4<<4)+13);
}
#define init_7_12_4_14
void FASTCALL call_7_12_4_14(SH2_struct* sh)
{
  SH2addi(sh, 12, (4<<4)+14);
}
#define init_7_12_4_15
void FASTCALL call_7_12_4_15(SH2_struct* sh)
{
  SH2addi(sh, 12, (4<<4)+15);
}
#define init_7_12_5_0
void FASTCALL call_7_12_5_0(SH2_struct* sh)
{
  SH2addi(sh, 12, (5<<4)+0);
}
#define init_7_12_5_1
void FASTCALL call_7_12_5_1(SH2_struct* sh)
{
  SH2addi(sh, 12, (5<<4)+1);
}
#define init_7_12_5_2
void FASTCALL call_7_12_5_2(SH2_struct* sh)
{
  SH2addi(sh, 12, (5<<4)+2);
}
#define init_7_12_5_3
void FASTCALL call_7_12_5_3(SH2_struct* sh)
{
  SH2addi(sh, 12, (5<<4)+3);
}
#define init_7_12_5_4
void FASTCALL call_7_12_5_4(SH2_struct* sh)
{
  SH2addi(sh, 12, (5<<4)+4);
}
#define init_7_12_5_5
void FASTCALL call_7_12_5_5(SH2_struct* sh)
{
  SH2addi(sh, 12, (5<<4)+5);
}
#define init_7_12_5_6
void FASTCALL call_7_12_5_6(SH2_struct* sh)
{
  SH2addi(sh, 12, (5<<4)+6);
}
#define init_7_12_5_7
void FASTCALL call_7_12_5_7(SH2_struct* sh)
{
  SH2addi(sh, 12, (5<<4)+7);
}
#define init_7_12_5_8
void FASTCALL call_7_12_5_8(SH2_struct* sh)
{
  SH2addi(sh, 12, (5<<4)+8);
}
#define init_7_12_5_9
void FASTCALL call_7_12_5_9(SH2_struct* sh)
{
  SH2addi(sh, 12, (5<<4)+9);
}
#define init_7_12_5_10
void FASTCALL call_7_12_5_10(SH2_struct* sh)
{
  SH2addi(sh, 12, (5<<4)+10);
}
#define init_7_12_5_11
void FASTCALL call_7_12_5_11(SH2_struct* sh)
{
  SH2addi(sh, 12, (5<<4)+11);
}
#define init_7_12_5_12
void FASTCALL call_7_12_5_12(SH2_struct* sh)
{
  SH2addi(sh, 12, (5<<4)+12);
}
#define init_7_12_5_13
void FASTCALL call_7_12_5_13(SH2_struct* sh)
{
  SH2addi(sh, 12, (5<<4)+13);
}
#define init_7_12_5_14
void FASTCALL call_7_12_5_14(SH2_struct* sh)
{
  SH2addi(sh, 12, (5<<4)+14);
}
#define init_7_12_5_15
void FASTCALL call_7_12_5_15(SH2_struct* sh)
{
  SH2addi(sh, 12, (5<<4)+15);
}
#define init_7_12_6_0
void FASTCALL call_7_12_6_0(SH2_struct* sh)
{
  SH2addi(sh, 12, (6<<4)+0);
}
#define init_7_12_6_1
void FASTCALL call_7_12_6_1(SH2_struct* sh)
{
  SH2addi(sh, 12, (6<<4)+1);
}
#define init_7_12_6_2
void FASTCALL call_7_12_6_2(SH2_struct* sh)
{
  SH2addi(sh, 12, (6<<4)+2);
}
#define init_7_12_6_3
void FASTCALL call_7_12_6_3(SH2_struct* sh)
{
  SH2addi(sh, 12, (6<<4)+3);
}
#define init_7_12_6_4
void FASTCALL call_7_12_6_4(SH2_struct* sh)
{
  SH2addi(sh, 12, (6<<4)+4);
}
#define init_7_12_6_5
void FASTCALL call_7_12_6_5(SH2_struct* sh)
{
  SH2addi(sh, 12, (6<<4)+5);
}
#define init_7_12_6_6
void FASTCALL call_7_12_6_6(SH2_struct* sh)
{
  SH2addi(sh, 12, (6<<4)+6);
}
#define init_7_12_6_7
void FASTCALL call_7_12_6_7(SH2_struct* sh)
{
  SH2addi(sh, 12, (6<<4)+7);
}
#define init_7_12_6_8
void FASTCALL call_7_12_6_8(SH2_struct* sh)
{
  SH2addi(sh, 12, (6<<4)+8);
}
#define init_7_12_6_9
void FASTCALL call_7_12_6_9(SH2_struct* sh)
{
  SH2addi(sh, 12, (6<<4)+9);
}
#define init_7_12_6_10
void FASTCALL call_7_12_6_10(SH2_struct* sh)
{
  SH2addi(sh, 12, (6<<4)+10);
}
#define init_7_12_6_11
void FASTCALL call_7_12_6_11(SH2_struct* sh)
{
  SH2addi(sh, 12, (6<<4)+11);
}
#define init_7_12_6_12
void FASTCALL call_7_12_6_12(SH2_struct* sh)
{
  SH2addi(sh, 12, (6<<4)+12);
}
#define init_7_12_6_13
void FASTCALL call_7_12_6_13(SH2_struct* sh)
{
  SH2addi(sh, 12, (6<<4)+13);
}
#define init_7_12_6_14
void FASTCALL call_7_12_6_14(SH2_struct* sh)
{
  SH2addi(sh, 12, (6<<4)+14);
}
#define init_7_12_6_15
void FASTCALL call_7_12_6_15(SH2_struct* sh)
{
  SH2addi(sh, 12, (6<<4)+15);
}
#define init_7_12_7_0
void FASTCALL call_7_12_7_0(SH2_struct* sh)
{
  SH2addi(sh, 12, (7<<4)+0);
}
#define init_7_12_7_1
void FASTCALL call_7_12_7_1(SH2_struct* sh)
{
  SH2addi(sh, 12, (7<<4)+1);
}
#define init_7_12_7_2
void FASTCALL call_7_12_7_2(SH2_struct* sh)
{
  SH2addi(sh, 12, (7<<4)+2);
}
#define init_7_12_7_3
void FASTCALL call_7_12_7_3(SH2_struct* sh)
{
  SH2addi(sh, 12, (7<<4)+3);
}
#define init_7_12_7_4
void FASTCALL call_7_12_7_4(SH2_struct* sh)
{
  SH2addi(sh, 12, (7<<4)+4);
}
#define init_7_12_7_5
void FASTCALL call_7_12_7_5(SH2_struct* sh)
{
  SH2addi(sh, 12, (7<<4)+5);
}
#define init_7_12_7_6
void FASTCALL call_7_12_7_6(SH2_struct* sh)
{
  SH2addi(sh, 12, (7<<4)+6);
}
#define init_7_12_7_7
void FASTCALL call_7_12_7_7(SH2_struct* sh)
{
  SH2addi(sh, 12, (7<<4)+7);
}
#define init_7_12_7_8
void FASTCALL call_7_12_7_8(SH2_struct* sh)
{
  SH2addi(sh, 12, (7<<4)+8);
}
#define init_7_12_7_9
void FASTCALL call_7_12_7_9(SH2_struct* sh)
{
  SH2addi(sh, 12, (7<<4)+9);
}
#define init_7_12_7_10
void FASTCALL call_7_12_7_10(SH2_struct* sh)
{
  SH2addi(sh, 12, (7<<4)+10);
}
#define init_7_12_7_11
void FASTCALL call_7_12_7_11(SH2_struct* sh)
{
  SH2addi(sh, 12, (7<<4)+11);
}
#define init_7_12_7_12
void FASTCALL call_7_12_7_12(SH2_struct* sh)
{
  SH2addi(sh, 12, (7<<4)+12);
}
#define init_7_12_7_13
void FASTCALL call_7_12_7_13(SH2_struct* sh)
{
  SH2addi(sh, 12, (7<<4)+13);
}
#define init_7_12_7_14
void FASTCALL call_7_12_7_14(SH2_struct* sh)
{
  SH2addi(sh, 12, (7<<4)+14);
}
#define init_7_12_7_15
void FASTCALL call_7_12_7_15(SH2_struct* sh)
{
  SH2addi(sh, 12, (7<<4)+15);
}
#define init_7_12_8_0
void FASTCALL call_7_12_8_0(SH2_struct* sh)
{
  SH2addi(sh, 12, (8<<4)+0);
}
#define init_7_12_8_1
void FASTCALL call_7_12_8_1(SH2_struct* sh)
{
  SH2addi(sh, 12, (8<<4)+1);
}
#define init_7_12_8_2
void FASTCALL call_7_12_8_2(SH2_struct* sh)
{
  SH2addi(sh, 12, (8<<4)+2);
}
#define init_7_12_8_3
void FASTCALL call_7_12_8_3(SH2_struct* sh)
{
  SH2addi(sh, 12, (8<<4)+3);
}
#define init_7_12_8_4
void FASTCALL call_7_12_8_4(SH2_struct* sh)
{
  SH2addi(sh, 12, (8<<4)+4);
}
#define init_7_12_8_5
void FASTCALL call_7_12_8_5(SH2_struct* sh)
{
  SH2addi(sh, 12, (8<<4)+5);
}
#define init_7_12_8_6
void FASTCALL call_7_12_8_6(SH2_struct* sh)
{
  SH2addi(sh, 12, (8<<4)+6);
}
#define init_7_12_8_7
void FASTCALL call_7_12_8_7(SH2_struct* sh)
{
  SH2addi(sh, 12, (8<<4)+7);
}
#define init_7_12_8_8
void FASTCALL call_7_12_8_8(SH2_struct* sh)
{
  SH2addi(sh, 12, (8<<4)+8);
}
#define init_7_12_8_9
void FASTCALL call_7_12_8_9(SH2_struct* sh)
{
  SH2addi(sh, 12, (8<<4)+9);
}
#define init_7_12_8_10
void FASTCALL call_7_12_8_10(SH2_struct* sh)
{
  SH2addi(sh, 12, (8<<4)+10);
}
#define init_7_12_8_11
void FASTCALL call_7_12_8_11(SH2_struct* sh)
{
  SH2addi(sh, 12, (8<<4)+11);
}
#define init_7_12_8_12
void FASTCALL call_7_12_8_12(SH2_struct* sh)
{
  SH2addi(sh, 12, (8<<4)+12);
}
#define init_7_12_8_13
void FASTCALL call_7_12_8_13(SH2_struct* sh)
{
  SH2addi(sh, 12, (8<<4)+13);
}
#define init_7_12_8_14
void FASTCALL call_7_12_8_14(SH2_struct* sh)
{
  SH2addi(sh, 12, (8<<4)+14);
}
#define init_7_12_8_15
void FASTCALL call_7_12_8_15(SH2_struct* sh)
{
  SH2addi(sh, 12, (8<<4)+15);
}
#define init_7_12_9_0
void FASTCALL call_7_12_9_0(SH2_struct* sh)
{
  SH2addi(sh, 12, (9<<4)+0);
}
#define init_7_12_9_1
void FASTCALL call_7_12_9_1(SH2_struct* sh)
{
  SH2addi(sh, 12, (9<<4)+1);
}
#define init_7_12_9_2
void FASTCALL call_7_12_9_2(SH2_struct* sh)
{
  SH2addi(sh, 12, (9<<4)+2);
}
#define init_7_12_9_3
void FASTCALL call_7_12_9_3(SH2_struct* sh)
{
  SH2addi(sh, 12, (9<<4)+3);
}
#define init_7_12_9_4
void FASTCALL call_7_12_9_4(SH2_struct* sh)
{
  SH2addi(sh, 12, (9<<4)+4);
}
#define init_7_12_9_5
void FASTCALL call_7_12_9_5(SH2_struct* sh)
{
  SH2addi(sh, 12, (9<<4)+5);
}
#define init_7_12_9_6
void FASTCALL call_7_12_9_6(SH2_struct* sh)
{
  SH2addi(sh, 12, (9<<4)+6);
}
#define init_7_12_9_7
void FASTCALL call_7_12_9_7(SH2_struct* sh)
{
  SH2addi(sh, 12, (9<<4)+7);
}
#define init_7_12_9_8
void FASTCALL call_7_12_9_8(SH2_struct* sh)
{
  SH2addi(sh, 12, (9<<4)+8);
}
#define init_7_12_9_9
void FASTCALL call_7_12_9_9(SH2_struct* sh)
{
  SH2addi(sh, 12, (9<<4)+9);
}
#define init_7_12_9_10
void FASTCALL call_7_12_9_10(SH2_struct* sh)
{
  SH2addi(sh, 12, (9<<4)+10);
}
#define init_7_12_9_11
void FASTCALL call_7_12_9_11(SH2_struct* sh)
{
  SH2addi(sh, 12, (9<<4)+11);
}
#define init_7_12_9_12
void FASTCALL call_7_12_9_12(SH2_struct* sh)
{
  SH2addi(sh, 12, (9<<4)+12);
}
#define init_7_12_9_13
void FASTCALL call_7_12_9_13(SH2_struct* sh)
{
  SH2addi(sh, 12, (9<<4)+13);
}
#define init_7_12_9_14
void FASTCALL call_7_12_9_14(SH2_struct* sh)
{
  SH2addi(sh, 12, (9<<4)+14);
}
#define init_7_12_9_15
void FASTCALL call_7_12_9_15(SH2_struct* sh)
{
  SH2addi(sh, 12, (9<<4)+15);
}
#define init_7_12_10_0
void FASTCALL call_7_12_10_0(SH2_struct* sh)
{
  SH2addi(sh, 12, (10<<4)+0);
}
#define init_7_12_10_1
void FASTCALL call_7_12_10_1(SH2_struct* sh)
{
  SH2addi(sh, 12, (10<<4)+1);
}
#define init_7_12_10_2
void FASTCALL call_7_12_10_2(SH2_struct* sh)
{
  SH2addi(sh, 12, (10<<4)+2);
}
#define init_7_12_10_3
void FASTCALL call_7_12_10_3(SH2_struct* sh)
{
  SH2addi(sh, 12, (10<<4)+3);
}
#define init_7_12_10_4
void FASTCALL call_7_12_10_4(SH2_struct* sh)
{
  SH2addi(sh, 12, (10<<4)+4);
}
#define init_7_12_10_5
void FASTCALL call_7_12_10_5(SH2_struct* sh)
{
  SH2addi(sh, 12, (10<<4)+5);
}
#define init_7_12_10_6
void FASTCALL call_7_12_10_6(SH2_struct* sh)
{
  SH2addi(sh, 12, (10<<4)+6);
}
#define init_7_12_10_7
void FASTCALL call_7_12_10_7(SH2_struct* sh)
{
  SH2addi(sh, 12, (10<<4)+7);
}
#define init_7_12_10_8
void FASTCALL call_7_12_10_8(SH2_struct* sh)
{
  SH2addi(sh, 12, (10<<4)+8);
}
#define init_7_12_10_9
void FASTCALL call_7_12_10_9(SH2_struct* sh)
{
  SH2addi(sh, 12, (10<<4)+9);
}
#define init_7_12_10_10
void FASTCALL call_7_12_10_10(SH2_struct* sh)
{
  SH2addi(sh, 12, (10<<4)+10);
}
#define init_7_12_10_11
void FASTCALL call_7_12_10_11(SH2_struct* sh)
{
  SH2addi(sh, 12, (10<<4)+11);
}
#define init_7_12_10_12
void FASTCALL call_7_12_10_12(SH2_struct* sh)
{
  SH2addi(sh, 12, (10<<4)+12);
}
#define init_7_12_10_13
void FASTCALL call_7_12_10_13(SH2_struct* sh)
{
  SH2addi(sh, 12, (10<<4)+13);
}
#define init_7_12_10_14
void FASTCALL call_7_12_10_14(SH2_struct* sh)
{
  SH2addi(sh, 12, (10<<4)+14);
}
#define init_7_12_10_15
void FASTCALL call_7_12_10_15(SH2_struct* sh)
{
  SH2addi(sh, 12, (10<<4)+15);
}
#define init_7_12_11_0
void FASTCALL call_7_12_11_0(SH2_struct* sh)
{
  SH2addi(sh, 12, (11<<4)+0);
}
#define init_7_12_11_1
void FASTCALL call_7_12_11_1(SH2_struct* sh)
{
  SH2addi(sh, 12, (11<<4)+1);
}
#define init_7_12_11_2
void FASTCALL call_7_12_11_2(SH2_struct* sh)
{
  SH2addi(sh, 12, (11<<4)+2);
}
#define init_7_12_11_3
void FASTCALL call_7_12_11_3(SH2_struct* sh)
{
  SH2addi(sh, 12, (11<<4)+3);
}
#define init_7_12_11_4
void FASTCALL call_7_12_11_4(SH2_struct* sh)
{
  SH2addi(sh, 12, (11<<4)+4);
}
#define init_7_12_11_5
void FASTCALL call_7_12_11_5(SH2_struct* sh)
{
  SH2addi(sh, 12, (11<<4)+5);
}
#define init_7_12_11_6
void FASTCALL call_7_12_11_6(SH2_struct* sh)
{
  SH2addi(sh, 12, (11<<4)+6);
}
#define init_7_12_11_7
void FASTCALL call_7_12_11_7(SH2_struct* sh)
{
  SH2addi(sh, 12, (11<<4)+7);
}
#define init_7_12_11_8
void FASTCALL call_7_12_11_8(SH2_struct* sh)
{
  SH2addi(sh, 12, (11<<4)+8);
}
#define init_7_12_11_9
void FASTCALL call_7_12_11_9(SH2_struct* sh)
{
  SH2addi(sh, 12, (11<<4)+9);
}
#define init_7_12_11_10
void FASTCALL call_7_12_11_10(SH2_struct* sh)
{
  SH2addi(sh, 12, (11<<4)+10);
}
#define init_7_12_11_11
void FASTCALL call_7_12_11_11(SH2_struct* sh)
{
  SH2addi(sh, 12, (11<<4)+11);
}
#define init_7_12_11_12
void FASTCALL call_7_12_11_12(SH2_struct* sh)
{
  SH2addi(sh, 12, (11<<4)+12);
}
#define init_7_12_11_13
void FASTCALL call_7_12_11_13(SH2_struct* sh)
{
  SH2addi(sh, 12, (11<<4)+13);
}
#define init_7_12_11_14
void FASTCALL call_7_12_11_14(SH2_struct* sh)
{
  SH2addi(sh, 12, (11<<4)+14);
}
#define init_7_12_11_15
void FASTCALL call_7_12_11_15(SH2_struct* sh)
{
  SH2addi(sh, 12, (11<<4)+15);
}
#define init_7_12_12_0
void FASTCALL call_7_12_12_0(SH2_struct* sh)
{
  SH2addi(sh, 12, (12<<4)+0);
}
#define init_7_12_12_1
void FASTCALL call_7_12_12_1(SH2_struct* sh)
{
  SH2addi(sh, 12, (12<<4)+1);
}
#define init_7_12_12_2
void FASTCALL call_7_12_12_2(SH2_struct* sh)
{
  SH2addi(sh, 12, (12<<4)+2);
}
#define init_7_12_12_3
void FASTCALL call_7_12_12_3(SH2_struct* sh)
{
  SH2addi(sh, 12, (12<<4)+3);
}
#define init_7_12_12_4
void FASTCALL call_7_12_12_4(SH2_struct* sh)
{
  SH2addi(sh, 12, (12<<4)+4);
}
#define init_7_12_12_5
void FASTCALL call_7_12_12_5(SH2_struct* sh)
{
  SH2addi(sh, 12, (12<<4)+5);
}
#define init_7_12_12_6
void FASTCALL call_7_12_12_6(SH2_struct* sh)
{
  SH2addi(sh, 12, (12<<4)+6);
}
#define init_7_12_12_7
void FASTCALL call_7_12_12_7(SH2_struct* sh)
{
  SH2addi(sh, 12, (12<<4)+7);
}
#define init_7_12_12_8
void FASTCALL call_7_12_12_8(SH2_struct* sh)
{
  SH2addi(sh, 12, (12<<4)+8);
}
#define init_7_12_12_9
void FASTCALL call_7_12_12_9(SH2_struct* sh)
{
  SH2addi(sh, 12, (12<<4)+9);
}
#define init_7_12_12_10
void FASTCALL call_7_12_12_10(SH2_struct* sh)
{
  SH2addi(sh, 12, (12<<4)+10);
}
#define init_7_12_12_11
void FASTCALL call_7_12_12_11(SH2_struct* sh)
{
  SH2addi(sh, 12, (12<<4)+11);
}
#define init_7_12_12_12
void FASTCALL call_7_12_12_12(SH2_struct* sh)
{
  SH2addi(sh, 12, (12<<4)+12);
}
#define init_7_12_12_13
void FASTCALL call_7_12_12_13(SH2_struct* sh)
{
  SH2addi(sh, 12, (12<<4)+13);
}
#define init_7_12_12_14
void FASTCALL call_7_12_12_14(SH2_struct* sh)
{
  SH2addi(sh, 12, (12<<4)+14);
}
#define init_7_12_12_15
void FASTCALL call_7_12_12_15(SH2_struct* sh)
{
  SH2addi(sh, 12, (12<<4)+15);
}
#define init_7_12_13_0
void FASTCALL call_7_12_13_0(SH2_struct* sh)
{
  SH2addi(sh, 12, (13<<4)+0);
}
#define init_7_12_13_1
void FASTCALL call_7_12_13_1(SH2_struct* sh)
{
  SH2addi(sh, 12, (13<<4)+1);
}
#define init_7_12_13_2
void FASTCALL call_7_12_13_2(SH2_struct* sh)
{
  SH2addi(sh, 12, (13<<4)+2);
}
#define init_7_12_13_3
void FASTCALL call_7_12_13_3(SH2_struct* sh)
{
  SH2addi(sh, 12, (13<<4)+3);
}
#define init_7_12_13_4
void FASTCALL call_7_12_13_4(SH2_struct* sh)
{
  SH2addi(sh, 12, (13<<4)+4);
}
#define init_7_12_13_5
void FASTCALL call_7_12_13_5(SH2_struct* sh)
{
  SH2addi(sh, 12, (13<<4)+5);
}
#define init_7_12_13_6
void FASTCALL call_7_12_13_6(SH2_struct* sh)
{
  SH2addi(sh, 12, (13<<4)+6);
}
#define init_7_12_13_7
void FASTCALL call_7_12_13_7(SH2_struct* sh)
{
  SH2addi(sh, 12, (13<<4)+7);
}
#define init_7_12_13_8
void FASTCALL call_7_12_13_8(SH2_struct* sh)
{
  SH2addi(sh, 12, (13<<4)+8);
}
#define init_7_12_13_9
void FASTCALL call_7_12_13_9(SH2_struct* sh)
{
  SH2addi(sh, 12, (13<<4)+9);
}
#define init_7_12_13_10
void FASTCALL call_7_12_13_10(SH2_struct* sh)
{
  SH2addi(sh, 12, (13<<4)+10);
}
#define init_7_12_13_11
void FASTCALL call_7_12_13_11(SH2_struct* sh)
{
  SH2addi(sh, 12, (13<<4)+11);
}
#define init_7_12_13_12
void FASTCALL call_7_12_13_12(SH2_struct* sh)
{
  SH2addi(sh, 12, (13<<4)+12);
}
#define init_7_12_13_13
void FASTCALL call_7_12_13_13(SH2_struct* sh)
{
  SH2addi(sh, 12, (13<<4)+13);
}
#define init_7_12_13_14
void FASTCALL call_7_12_13_14(SH2_struct* sh)
{
  SH2addi(sh, 12, (13<<4)+14);
}
#define init_7_12_13_15
void FASTCALL call_7_12_13_15(SH2_struct* sh)
{
  SH2addi(sh, 12, (13<<4)+15);
}
#define init_7_12_14_0
void FASTCALL call_7_12_14_0(SH2_struct* sh)
{
  SH2addi(sh, 12, (14<<4)+0);
}
#define init_7_12_14_1
void FASTCALL call_7_12_14_1(SH2_struct* sh)
{
  SH2addi(sh, 12, (14<<4)+1);
}
#define init_7_12_14_2
void FASTCALL call_7_12_14_2(SH2_struct* sh)
{
  SH2addi(sh, 12, (14<<4)+2);
}
#define init_7_12_14_3
void FASTCALL call_7_12_14_3(SH2_struct* sh)
{
  SH2addi(sh, 12, (14<<4)+3);
}
#define init_7_12_14_4
void FASTCALL call_7_12_14_4(SH2_struct* sh)
{
  SH2addi(sh, 12, (14<<4)+4);
}
#define init_7_12_14_5
void FASTCALL call_7_12_14_5(SH2_struct* sh)
{
  SH2addi(sh, 12, (14<<4)+5);
}
#define init_7_12_14_6
void FASTCALL call_7_12_14_6(SH2_struct* sh)
{
  SH2addi(sh, 12, (14<<4)+6);
}
#define init_7_12_14_7
void FASTCALL call_7_12_14_7(SH2_struct* sh)
{
  SH2addi(sh, 12, (14<<4)+7);
}
#define init_7_12_14_8
void FASTCALL call_7_12_14_8(SH2_struct* sh)
{
  SH2addi(sh, 12, (14<<4)+8);
}
#define init_7_12_14_9
void FASTCALL call_7_12_14_9(SH2_struct* sh)
{
  SH2addi(sh, 12, (14<<4)+9);
}
#define init_7_12_14_10
void FASTCALL call_7_12_14_10(SH2_struct* sh)
{
  SH2addi(sh, 12, (14<<4)+10);
}
#define init_7_12_14_11
void FASTCALL call_7_12_14_11(SH2_struct* sh)
{
  SH2addi(sh, 12, (14<<4)+11);
}
#define init_7_12_14_12
void FASTCALL call_7_12_14_12(SH2_struct* sh)
{
  SH2addi(sh, 12, (14<<4)+12);
}
#define init_7_12_14_13
void FASTCALL call_7_12_14_13(SH2_struct* sh)
{
  SH2addi(sh, 12, (14<<4)+13);
}
#define init_7_12_14_14
void FASTCALL call_7_12_14_14(SH2_struct* sh)
{
  SH2addi(sh, 12, (14<<4)+14);
}
#define init_7_12_14_15
void FASTCALL call_7_12_14_15(SH2_struct* sh)
{
  SH2addi(sh, 12, (14<<4)+15);
}
#define init_7_12_15_0
void FASTCALL call_7_12_15_0(SH2_struct* sh)
{
  SH2addi(sh, 12, (15<<4)+0);
}
#define init_7_12_15_1
void FASTCALL call_7_12_15_1(SH2_struct* sh)
{
  SH2addi(sh, 12, (15<<4)+1);
}
#define init_7_12_15_2
void FASTCALL call_7_12_15_2(SH2_struct* sh)
{
  SH2addi(sh, 12, (15<<4)+2);
}
#define init_7_12_15_3
void FASTCALL call_7_12_15_3(SH2_struct* sh)
{
  SH2addi(sh, 12, (15<<4)+3);
}
#define init_7_12_15_4
void FASTCALL call_7_12_15_4(SH2_struct* sh)
{
  SH2addi(sh, 12, (15<<4)+4);
}
#define init_7_12_15_5
void FASTCALL call_7_12_15_5(SH2_struct* sh)
{
  SH2addi(sh, 12, (15<<4)+5);
}
#define init_7_12_15_6
void FASTCALL call_7_12_15_6(SH2_struct* sh)
{
  SH2addi(sh, 12, (15<<4)+6);
}
#define init_7_12_15_7
void FASTCALL call_7_12_15_7(SH2_struct* sh)
{
  SH2addi(sh, 12, (15<<4)+7);
}
#define init_7_12_15_8
void FASTCALL call_7_12_15_8(SH2_struct* sh)
{
  SH2addi(sh, 12, (15<<4)+8);
}
#define init_7_12_15_9
void FASTCALL call_7_12_15_9(SH2_struct* sh)
{
  SH2addi(sh, 12, (15<<4)+9);
}
#define init_7_12_15_10
void FASTCALL call_7_12_15_10(SH2_struct* sh)
{
  SH2addi(sh, 12, (15<<4)+10);
}
#define init_7_12_15_11
void FASTCALL call_7_12_15_11(SH2_struct* sh)
{
  SH2addi(sh, 12, (15<<4)+11);
}
#define init_7_12_15_12
void FASTCALL call_7_12_15_12(SH2_struct* sh)
{
  SH2addi(sh, 12, (15<<4)+12);
}
#define init_7_12_15_13
void FASTCALL call_7_12_15_13(SH2_struct* sh)
{
  SH2addi(sh, 12, (15<<4)+13);
}
#define init_7_12_15_14
void FASTCALL call_7_12_15_14(SH2_struct* sh)
{
  SH2addi(sh, 12, (15<<4)+14);
}
#define init_7_12_15_15
void FASTCALL call_7_12_15_15(SH2_struct* sh)
{
  SH2addi(sh, 12, (15<<4)+15);
}
#define init_7_13_0_0
void FASTCALL call_7_13_0_0(SH2_struct* sh)
{
  SH2addi(sh, 13, (0<<4)+0);
}
#define init_7_13_0_1
void FASTCALL call_7_13_0_1(SH2_struct* sh)
{
  SH2addi(sh, 13, (0<<4)+1);
}
#define init_7_13_0_2
void FASTCALL call_7_13_0_2(SH2_struct* sh)
{
  SH2addi(sh, 13, (0<<4)+2);
}
#define init_7_13_0_3
void FASTCALL call_7_13_0_3(SH2_struct* sh)
{
  SH2addi(sh, 13, (0<<4)+3);
}
#define init_7_13_0_4
void FASTCALL call_7_13_0_4(SH2_struct* sh)
{
  SH2addi(sh, 13, (0<<4)+4);
}
#define init_7_13_0_5
void FASTCALL call_7_13_0_5(SH2_struct* sh)
{
  SH2addi(sh, 13, (0<<4)+5);
}
#define init_7_13_0_6
void FASTCALL call_7_13_0_6(SH2_struct* sh)
{
  SH2addi(sh, 13, (0<<4)+6);
}
#define init_7_13_0_7
void FASTCALL call_7_13_0_7(SH2_struct* sh)
{
  SH2addi(sh, 13, (0<<4)+7);
}
#define init_7_13_0_8
void FASTCALL call_7_13_0_8(SH2_struct* sh)
{
  SH2addi(sh, 13, (0<<4)+8);
}
#define init_7_13_0_9
void FASTCALL call_7_13_0_9(SH2_struct* sh)
{
  SH2addi(sh, 13, (0<<4)+9);
}
#define init_7_13_0_10
void FASTCALL call_7_13_0_10(SH2_struct* sh)
{
  SH2addi(sh, 13, (0<<4)+10);
}
#define init_7_13_0_11
void FASTCALL call_7_13_0_11(SH2_struct* sh)
{
  SH2addi(sh, 13, (0<<4)+11);
}
#define init_7_13_0_12
void FASTCALL call_7_13_0_12(SH2_struct* sh)
{
  SH2addi(sh, 13, (0<<4)+12);
}
#define init_7_13_0_13
void FASTCALL call_7_13_0_13(SH2_struct* sh)
{
  SH2addi(sh, 13, (0<<4)+13);
}
#define init_7_13_0_14
void FASTCALL call_7_13_0_14(SH2_struct* sh)
{
  SH2addi(sh, 13, (0<<4)+14);
}
#define init_7_13_0_15
void FASTCALL call_7_13_0_15(SH2_struct* sh)
{
  SH2addi(sh, 13, (0<<4)+15);
}
#define init_7_13_1_0
void FASTCALL call_7_13_1_0(SH2_struct* sh)
{
  SH2addi(sh, 13, (1<<4)+0);
}
#define init_7_13_1_1
void FASTCALL call_7_13_1_1(SH2_struct* sh)
{
  SH2addi(sh, 13, (1<<4)+1);
}
#define init_7_13_1_2
void FASTCALL call_7_13_1_2(SH2_struct* sh)
{
  SH2addi(sh, 13, (1<<4)+2);
}
#define init_7_13_1_3
void FASTCALL call_7_13_1_3(SH2_struct* sh)
{
  SH2addi(sh, 13, (1<<4)+3);
}
#define init_7_13_1_4
void FASTCALL call_7_13_1_4(SH2_struct* sh)
{
  SH2addi(sh, 13, (1<<4)+4);
}
#define init_7_13_1_5
void FASTCALL call_7_13_1_5(SH2_struct* sh)
{
  SH2addi(sh, 13, (1<<4)+5);
}
#define init_7_13_1_6
void FASTCALL call_7_13_1_6(SH2_struct* sh)
{
  SH2addi(sh, 13, (1<<4)+6);
}
#define init_7_13_1_7
void FASTCALL call_7_13_1_7(SH2_struct* sh)
{
  SH2addi(sh, 13, (1<<4)+7);
}
#define init_7_13_1_8
void FASTCALL call_7_13_1_8(SH2_struct* sh)
{
  SH2addi(sh, 13, (1<<4)+8);
}
#define init_7_13_1_9
void FASTCALL call_7_13_1_9(SH2_struct* sh)
{
  SH2addi(sh, 13, (1<<4)+9);
}
#define init_7_13_1_10
void FASTCALL call_7_13_1_10(SH2_struct* sh)
{
  SH2addi(sh, 13, (1<<4)+10);
}
#define init_7_13_1_11
void FASTCALL call_7_13_1_11(SH2_struct* sh)
{
  SH2addi(sh, 13, (1<<4)+11);
}
#define init_7_13_1_12
void FASTCALL call_7_13_1_12(SH2_struct* sh)
{
  SH2addi(sh, 13, (1<<4)+12);
}
#define init_7_13_1_13
void FASTCALL call_7_13_1_13(SH2_struct* sh)
{
  SH2addi(sh, 13, (1<<4)+13);
}
#define init_7_13_1_14
void FASTCALL call_7_13_1_14(SH2_struct* sh)
{
  SH2addi(sh, 13, (1<<4)+14);
}
#define init_7_13_1_15
void FASTCALL call_7_13_1_15(SH2_struct* sh)
{
  SH2addi(sh, 13, (1<<4)+15);
}
#define init_7_13_2_0
void FASTCALL call_7_13_2_0(SH2_struct* sh)
{
  SH2addi(sh, 13, (2<<4)+0);
}
#define init_7_13_2_1
void FASTCALL call_7_13_2_1(SH2_struct* sh)
{
  SH2addi(sh, 13, (2<<4)+1);
}
#define init_7_13_2_2
void FASTCALL call_7_13_2_2(SH2_struct* sh)
{
  SH2addi(sh, 13, (2<<4)+2);
}
#define init_7_13_2_3
void FASTCALL call_7_13_2_3(SH2_struct* sh)
{
  SH2addi(sh, 13, (2<<4)+3);
}
#define init_7_13_2_4
void FASTCALL call_7_13_2_4(SH2_struct* sh)
{
  SH2addi(sh, 13, (2<<4)+4);
}
#define init_7_13_2_5
void FASTCALL call_7_13_2_5(SH2_struct* sh)
{
  SH2addi(sh, 13, (2<<4)+5);
}
#define init_7_13_2_6
void FASTCALL call_7_13_2_6(SH2_struct* sh)
{
  SH2addi(sh, 13, (2<<4)+6);
}
#define init_7_13_2_7
void FASTCALL call_7_13_2_7(SH2_struct* sh)
{
  SH2addi(sh, 13, (2<<4)+7);
}
#define init_7_13_2_8
void FASTCALL call_7_13_2_8(SH2_struct* sh)
{
  SH2addi(sh, 13, (2<<4)+8);
}
#define init_7_13_2_9
void FASTCALL call_7_13_2_9(SH2_struct* sh)
{
  SH2addi(sh, 13, (2<<4)+9);
}
#define init_7_13_2_10
void FASTCALL call_7_13_2_10(SH2_struct* sh)
{
  SH2addi(sh, 13, (2<<4)+10);
}
#define init_7_13_2_11
void FASTCALL call_7_13_2_11(SH2_struct* sh)
{
  SH2addi(sh, 13, (2<<4)+11);
}
#define init_7_13_2_12
void FASTCALL call_7_13_2_12(SH2_struct* sh)
{
  SH2addi(sh, 13, (2<<4)+12);
}
#define init_7_13_2_13
void FASTCALL call_7_13_2_13(SH2_struct* sh)
{
  SH2addi(sh, 13, (2<<4)+13);
}
#define init_7_13_2_14
void FASTCALL call_7_13_2_14(SH2_struct* sh)
{
  SH2addi(sh, 13, (2<<4)+14);
}
#define init_7_13_2_15
void FASTCALL call_7_13_2_15(SH2_struct* sh)
{
  SH2addi(sh, 13, (2<<4)+15);
}
#define init_7_13_3_0
void FASTCALL call_7_13_3_0(SH2_struct* sh)
{
  SH2addi(sh, 13, (3<<4)+0);
}
#define init_7_13_3_1
void FASTCALL call_7_13_3_1(SH2_struct* sh)
{
  SH2addi(sh, 13, (3<<4)+1);
}
#define init_7_13_3_2
void FASTCALL call_7_13_3_2(SH2_struct* sh)
{
  SH2addi(sh, 13, (3<<4)+2);
}
#define init_7_13_3_3
void FASTCALL call_7_13_3_3(SH2_struct* sh)
{
  SH2addi(sh, 13, (3<<4)+3);
}
#define init_7_13_3_4
void FASTCALL call_7_13_3_4(SH2_struct* sh)
{
  SH2addi(sh, 13, (3<<4)+4);
}
#define init_7_13_3_5
void FASTCALL call_7_13_3_5(SH2_struct* sh)
{
  SH2addi(sh, 13, (3<<4)+5);
}
#define init_7_13_3_6
void FASTCALL call_7_13_3_6(SH2_struct* sh)
{
  SH2addi(sh, 13, (3<<4)+6);
}
#define init_7_13_3_7
void FASTCALL call_7_13_3_7(SH2_struct* sh)
{
  SH2addi(sh, 13, (3<<4)+7);
}
#define init_7_13_3_8
void FASTCALL call_7_13_3_8(SH2_struct* sh)
{
  SH2addi(sh, 13, (3<<4)+8);
}
#define init_7_13_3_9
void FASTCALL call_7_13_3_9(SH2_struct* sh)
{
  SH2addi(sh, 13, (3<<4)+9);
}
#define init_7_13_3_10
void FASTCALL call_7_13_3_10(SH2_struct* sh)
{
  SH2addi(sh, 13, (3<<4)+10);
}
#define init_7_13_3_11
void FASTCALL call_7_13_3_11(SH2_struct* sh)
{
  SH2addi(sh, 13, (3<<4)+11);
}
#define init_7_13_3_12
void FASTCALL call_7_13_3_12(SH2_struct* sh)
{
  SH2addi(sh, 13, (3<<4)+12);
}
#define init_7_13_3_13
void FASTCALL call_7_13_3_13(SH2_struct* sh)
{
  SH2addi(sh, 13, (3<<4)+13);
}
#define init_7_13_3_14
void FASTCALL call_7_13_3_14(SH2_struct* sh)
{
  SH2addi(sh, 13, (3<<4)+14);
}
#define init_7_13_3_15
void FASTCALL call_7_13_3_15(SH2_struct* sh)
{
  SH2addi(sh, 13, (3<<4)+15);
}
#define init_7_13_4_0
void FASTCALL call_7_13_4_0(SH2_struct* sh)
{
  SH2addi(sh, 13, (4<<4)+0);
}
#define init_7_13_4_1
void FASTCALL call_7_13_4_1(SH2_struct* sh)
{
  SH2addi(sh, 13, (4<<4)+1);
}
#define init_7_13_4_2
void FASTCALL call_7_13_4_2(SH2_struct* sh)
{
  SH2addi(sh, 13, (4<<4)+2);
}
#define init_7_13_4_3
void FASTCALL call_7_13_4_3(SH2_struct* sh)
{
  SH2addi(sh, 13, (4<<4)+3);
}
#define init_7_13_4_4
void FASTCALL call_7_13_4_4(SH2_struct* sh)
{
  SH2addi(sh, 13, (4<<4)+4);
}
#define init_7_13_4_5
void FASTCALL call_7_13_4_5(SH2_struct* sh)
{
  SH2addi(sh, 13, (4<<4)+5);
}
#define init_7_13_4_6
void FASTCALL call_7_13_4_6(SH2_struct* sh)
{
  SH2addi(sh, 13, (4<<4)+6);
}
#define init_7_13_4_7
void FASTCALL call_7_13_4_7(SH2_struct* sh)
{
  SH2addi(sh, 13, (4<<4)+7);
}
#define init_7_13_4_8
void FASTCALL call_7_13_4_8(SH2_struct* sh)
{
  SH2addi(sh, 13, (4<<4)+8);
}
#define init_7_13_4_9
void FASTCALL call_7_13_4_9(SH2_struct* sh)
{
  SH2addi(sh, 13, (4<<4)+9);
}
#define init_7_13_4_10
void FASTCALL call_7_13_4_10(SH2_struct* sh)
{
  SH2addi(sh, 13, (4<<4)+10);
}
#define init_7_13_4_11
void FASTCALL call_7_13_4_11(SH2_struct* sh)
{
  SH2addi(sh, 13, (4<<4)+11);
}
#define init_7_13_4_12
void FASTCALL call_7_13_4_12(SH2_struct* sh)
{
  SH2addi(sh, 13, (4<<4)+12);
}
#define init_7_13_4_13
void FASTCALL call_7_13_4_13(SH2_struct* sh)
{
  SH2addi(sh, 13, (4<<4)+13);
}
#define init_7_13_4_14
void FASTCALL call_7_13_4_14(SH2_struct* sh)
{
  SH2addi(sh, 13, (4<<4)+14);
}
#define init_7_13_4_15
void FASTCALL call_7_13_4_15(SH2_struct* sh)
{
  SH2addi(sh, 13, (4<<4)+15);
}
#define init_7_13_5_0
void FASTCALL call_7_13_5_0(SH2_struct* sh)
{
  SH2addi(sh, 13, (5<<4)+0);
}
#define init_7_13_5_1
void FASTCALL call_7_13_5_1(SH2_struct* sh)
{
  SH2addi(sh, 13, (5<<4)+1);
}
#define init_7_13_5_2
void FASTCALL call_7_13_5_2(SH2_struct* sh)
{
  SH2addi(sh, 13, (5<<4)+2);
}
#define init_7_13_5_3
void FASTCALL call_7_13_5_3(SH2_struct* sh)
{
  SH2addi(sh, 13, (5<<4)+3);
}
#define init_7_13_5_4
void FASTCALL call_7_13_5_4(SH2_struct* sh)
{
  SH2addi(sh, 13, (5<<4)+4);
}
#define init_7_13_5_5
void FASTCALL call_7_13_5_5(SH2_struct* sh)
{
  SH2addi(sh, 13, (5<<4)+5);
}
#define init_7_13_5_6
void FASTCALL call_7_13_5_6(SH2_struct* sh)
{
  SH2addi(sh, 13, (5<<4)+6);
}
#define init_7_13_5_7
void FASTCALL call_7_13_5_7(SH2_struct* sh)
{
  SH2addi(sh, 13, (5<<4)+7);
}
#define init_7_13_5_8
void FASTCALL call_7_13_5_8(SH2_struct* sh)
{
  SH2addi(sh, 13, (5<<4)+8);
}
#define init_7_13_5_9
void FASTCALL call_7_13_5_9(SH2_struct* sh)
{
  SH2addi(sh, 13, (5<<4)+9);
}
#define init_7_13_5_10
void FASTCALL call_7_13_5_10(SH2_struct* sh)
{
  SH2addi(sh, 13, (5<<4)+10);
}
#define init_7_13_5_11
void FASTCALL call_7_13_5_11(SH2_struct* sh)
{
  SH2addi(sh, 13, (5<<4)+11);
}
#define init_7_13_5_12
void FASTCALL call_7_13_5_12(SH2_struct* sh)
{
  SH2addi(sh, 13, (5<<4)+12);
}
#define init_7_13_5_13
void FASTCALL call_7_13_5_13(SH2_struct* sh)
{
  SH2addi(sh, 13, (5<<4)+13);
}
#define init_7_13_5_14
void FASTCALL call_7_13_5_14(SH2_struct* sh)
{
  SH2addi(sh, 13, (5<<4)+14);
}
#define init_7_13_5_15
void FASTCALL call_7_13_5_15(SH2_struct* sh)
{
  SH2addi(sh, 13, (5<<4)+15);
}
#define init_7_13_6_0
void FASTCALL call_7_13_6_0(SH2_struct* sh)
{
  SH2addi(sh, 13, (6<<4)+0);
}
#define init_7_13_6_1
void FASTCALL call_7_13_6_1(SH2_struct* sh)
{
  SH2addi(sh, 13, (6<<4)+1);
}
#define init_7_13_6_2
void FASTCALL call_7_13_6_2(SH2_struct* sh)
{
  SH2addi(sh, 13, (6<<4)+2);
}
#define init_7_13_6_3
void FASTCALL call_7_13_6_3(SH2_struct* sh)
{
  SH2addi(sh, 13, (6<<4)+3);
}
#define init_7_13_6_4
void FASTCALL call_7_13_6_4(SH2_struct* sh)
{
  SH2addi(sh, 13, (6<<4)+4);
}
#define init_7_13_6_5
void FASTCALL call_7_13_6_5(SH2_struct* sh)
{
  SH2addi(sh, 13, (6<<4)+5);
}
#define init_7_13_6_6
void FASTCALL call_7_13_6_6(SH2_struct* sh)
{
  SH2addi(sh, 13, (6<<4)+6);
}
#define init_7_13_6_7
void FASTCALL call_7_13_6_7(SH2_struct* sh)
{
  SH2addi(sh, 13, (6<<4)+7);
}
#define init_7_13_6_8
void FASTCALL call_7_13_6_8(SH2_struct* sh)
{
  SH2addi(sh, 13, (6<<4)+8);
}
#define init_7_13_6_9
void FASTCALL call_7_13_6_9(SH2_struct* sh)
{
  SH2addi(sh, 13, (6<<4)+9);
}
#define init_7_13_6_10
void FASTCALL call_7_13_6_10(SH2_struct* sh)
{
  SH2addi(sh, 13, (6<<4)+10);
}
#define init_7_13_6_11
void FASTCALL call_7_13_6_11(SH2_struct* sh)
{
  SH2addi(sh, 13, (6<<4)+11);
}
#define init_7_13_6_12
void FASTCALL call_7_13_6_12(SH2_struct* sh)
{
  SH2addi(sh, 13, (6<<4)+12);
}
#define init_7_13_6_13
void FASTCALL call_7_13_6_13(SH2_struct* sh)
{
  SH2addi(sh, 13, (6<<4)+13);
}
#define init_7_13_6_14
void FASTCALL call_7_13_6_14(SH2_struct* sh)
{
  SH2addi(sh, 13, (6<<4)+14);
}
#define init_7_13_6_15
void FASTCALL call_7_13_6_15(SH2_struct* sh)
{
  SH2addi(sh, 13, (6<<4)+15);
}
#define init_7_13_7_0
void FASTCALL call_7_13_7_0(SH2_struct* sh)
{
  SH2addi(sh, 13, (7<<4)+0);
}
#define init_7_13_7_1
void FASTCALL call_7_13_7_1(SH2_struct* sh)
{
  SH2addi(sh, 13, (7<<4)+1);
}
#define init_7_13_7_2
void FASTCALL call_7_13_7_2(SH2_struct* sh)
{
  SH2addi(sh, 13, (7<<4)+2);
}
#define init_7_13_7_3
void FASTCALL call_7_13_7_3(SH2_struct* sh)
{
  SH2addi(sh, 13, (7<<4)+3);
}
#define init_7_13_7_4
void FASTCALL call_7_13_7_4(SH2_struct* sh)
{
  SH2addi(sh, 13, (7<<4)+4);
}
#define init_7_13_7_5
void FASTCALL call_7_13_7_5(SH2_struct* sh)
{
  SH2addi(sh, 13, (7<<4)+5);
}
#define init_7_13_7_6
void FASTCALL call_7_13_7_6(SH2_struct* sh)
{
  SH2addi(sh, 13, (7<<4)+6);
}
#define init_7_13_7_7
void FASTCALL call_7_13_7_7(SH2_struct* sh)
{
  SH2addi(sh, 13, (7<<4)+7);
}
#define init_7_13_7_8
void FASTCALL call_7_13_7_8(SH2_struct* sh)
{
  SH2addi(sh, 13, (7<<4)+8);
}
#define init_7_13_7_9
void FASTCALL call_7_13_7_9(SH2_struct* sh)
{
  SH2addi(sh, 13, (7<<4)+9);
}
#define init_7_13_7_10
void FASTCALL call_7_13_7_10(SH2_struct* sh)
{
  SH2addi(sh, 13, (7<<4)+10);
}
#define init_7_13_7_11
void FASTCALL call_7_13_7_11(SH2_struct* sh)
{
  SH2addi(sh, 13, (7<<4)+11);
}
#define init_7_13_7_12
void FASTCALL call_7_13_7_12(SH2_struct* sh)
{
  SH2addi(sh, 13, (7<<4)+12);
}
#define init_7_13_7_13
void FASTCALL call_7_13_7_13(SH2_struct* sh)
{
  SH2addi(sh, 13, (7<<4)+13);
}
#define init_7_13_7_14
void FASTCALL call_7_13_7_14(SH2_struct* sh)
{
  SH2addi(sh, 13, (7<<4)+14);
}
#define init_7_13_7_15
void FASTCALL call_7_13_7_15(SH2_struct* sh)
{
  SH2addi(sh, 13, (7<<4)+15);
}
#define init_7_13_8_0
void FASTCALL call_7_13_8_0(SH2_struct* sh)
{
  SH2addi(sh, 13, (8<<4)+0);
}
#define init_7_13_8_1
void FASTCALL call_7_13_8_1(SH2_struct* sh)
{
  SH2addi(sh, 13, (8<<4)+1);
}
#define init_7_13_8_2
void FASTCALL call_7_13_8_2(SH2_struct* sh)
{
  SH2addi(sh, 13, (8<<4)+2);
}
#define init_7_13_8_3
void FASTCALL call_7_13_8_3(SH2_struct* sh)
{
  SH2addi(sh, 13, (8<<4)+3);
}
#define init_7_13_8_4
void FASTCALL call_7_13_8_4(SH2_struct* sh)
{
  SH2addi(sh, 13, (8<<4)+4);
}
#define init_7_13_8_5
void FASTCALL call_7_13_8_5(SH2_struct* sh)
{
  SH2addi(sh, 13, (8<<4)+5);
}
#define init_7_13_8_6
void FASTCALL call_7_13_8_6(SH2_struct* sh)
{
  SH2addi(sh, 13, (8<<4)+6);
}
#define init_7_13_8_7
void FASTCALL call_7_13_8_7(SH2_struct* sh)
{
  SH2addi(sh, 13, (8<<4)+7);
}
#define init_7_13_8_8
void FASTCALL call_7_13_8_8(SH2_struct* sh)
{
  SH2addi(sh, 13, (8<<4)+8);
}
#define init_7_13_8_9
void FASTCALL call_7_13_8_9(SH2_struct* sh)
{
  SH2addi(sh, 13, (8<<4)+9);
}
#define init_7_13_8_10
void FASTCALL call_7_13_8_10(SH2_struct* sh)
{
  SH2addi(sh, 13, (8<<4)+10);
}
#define init_7_13_8_11
void FASTCALL call_7_13_8_11(SH2_struct* sh)
{
  SH2addi(sh, 13, (8<<4)+11);
}
#define init_7_13_8_12
void FASTCALL call_7_13_8_12(SH2_struct* sh)
{
  SH2addi(sh, 13, (8<<4)+12);
}
#define init_7_13_8_13
void FASTCALL call_7_13_8_13(SH2_struct* sh)
{
  SH2addi(sh, 13, (8<<4)+13);
}
#define init_7_13_8_14
void FASTCALL call_7_13_8_14(SH2_struct* sh)
{
  SH2addi(sh, 13, (8<<4)+14);
}
#define init_7_13_8_15
void FASTCALL call_7_13_8_15(SH2_struct* sh)
{
  SH2addi(sh, 13, (8<<4)+15);
}
#define init_7_13_9_0
void FASTCALL call_7_13_9_0(SH2_struct* sh)
{
  SH2addi(sh, 13, (9<<4)+0);
}
#define init_7_13_9_1
void FASTCALL call_7_13_9_1(SH2_struct* sh)
{
  SH2addi(sh, 13, (9<<4)+1);
}
#define init_7_13_9_2
void FASTCALL call_7_13_9_2(SH2_struct* sh)
{
  SH2addi(sh, 13, (9<<4)+2);
}
#define init_7_13_9_3
void FASTCALL call_7_13_9_3(SH2_struct* sh)
{
  SH2addi(sh, 13, (9<<4)+3);
}
#define init_7_13_9_4
void FASTCALL call_7_13_9_4(SH2_struct* sh)
{
  SH2addi(sh, 13, (9<<4)+4);
}
#define init_7_13_9_5
void FASTCALL call_7_13_9_5(SH2_struct* sh)
{
  SH2addi(sh, 13, (9<<4)+5);
}
#define init_7_13_9_6
void FASTCALL call_7_13_9_6(SH2_struct* sh)
{
  SH2addi(sh, 13, (9<<4)+6);
}
#define init_7_13_9_7
void FASTCALL call_7_13_9_7(SH2_struct* sh)
{
  SH2addi(sh, 13, (9<<4)+7);
}
#define init_7_13_9_8
void FASTCALL call_7_13_9_8(SH2_struct* sh)
{
  SH2addi(sh, 13, (9<<4)+8);
}
#define init_7_13_9_9
void FASTCALL call_7_13_9_9(SH2_struct* sh)
{
  SH2addi(sh, 13, (9<<4)+9);
}
#define init_7_13_9_10
void FASTCALL call_7_13_9_10(SH2_struct* sh)
{
  SH2addi(sh, 13, (9<<4)+10);
}
#define init_7_13_9_11
void FASTCALL call_7_13_9_11(SH2_struct* sh)
{
  SH2addi(sh, 13, (9<<4)+11);
}
#define init_7_13_9_12
void FASTCALL call_7_13_9_12(SH2_struct* sh)
{
  SH2addi(sh, 13, (9<<4)+12);
}
#define init_7_13_9_13
void FASTCALL call_7_13_9_13(SH2_struct* sh)
{
  SH2addi(sh, 13, (9<<4)+13);
}
#define init_7_13_9_14
void FASTCALL call_7_13_9_14(SH2_struct* sh)
{
  SH2addi(sh, 13, (9<<4)+14);
}
#define init_7_13_9_15
void FASTCALL call_7_13_9_15(SH2_struct* sh)
{
  SH2addi(sh, 13, (9<<4)+15);
}
#define init_7_13_10_0
void FASTCALL call_7_13_10_0(SH2_struct* sh)
{
  SH2addi(sh, 13, (10<<4)+0);
}
#define init_7_13_10_1
void FASTCALL call_7_13_10_1(SH2_struct* sh)
{
  SH2addi(sh, 13, (10<<4)+1);
}
#define init_7_13_10_2
void FASTCALL call_7_13_10_2(SH2_struct* sh)
{
  SH2addi(sh, 13, (10<<4)+2);
}
#define init_7_13_10_3
void FASTCALL call_7_13_10_3(SH2_struct* sh)
{
  SH2addi(sh, 13, (10<<4)+3);
}
#define init_7_13_10_4
void FASTCALL call_7_13_10_4(SH2_struct* sh)
{
  SH2addi(sh, 13, (10<<4)+4);
}
#define init_7_13_10_5
void FASTCALL call_7_13_10_5(SH2_struct* sh)
{
  SH2addi(sh, 13, (10<<4)+5);
}
#define init_7_13_10_6
void FASTCALL call_7_13_10_6(SH2_struct* sh)
{
  SH2addi(sh, 13, (10<<4)+6);
}
#define init_7_13_10_7
void FASTCALL call_7_13_10_7(SH2_struct* sh)
{
  SH2addi(sh, 13, (10<<4)+7);
}
#define init_7_13_10_8
void FASTCALL call_7_13_10_8(SH2_struct* sh)
{
  SH2addi(sh, 13, (10<<4)+8);
}
#define init_7_13_10_9
void FASTCALL call_7_13_10_9(SH2_struct* sh)
{
  SH2addi(sh, 13, (10<<4)+9);
}
#define init_7_13_10_10
void FASTCALL call_7_13_10_10(SH2_struct* sh)
{
  SH2addi(sh, 13, (10<<4)+10);
}
#define init_7_13_10_11
void FASTCALL call_7_13_10_11(SH2_struct* sh)
{
  SH2addi(sh, 13, (10<<4)+11);
}
#define init_7_13_10_12
void FASTCALL call_7_13_10_12(SH2_struct* sh)
{
  SH2addi(sh, 13, (10<<4)+12);
}
#define init_7_13_10_13
void FASTCALL call_7_13_10_13(SH2_struct* sh)
{
  SH2addi(sh, 13, (10<<4)+13);
}
#define init_7_13_10_14
void FASTCALL call_7_13_10_14(SH2_struct* sh)
{
  SH2addi(sh, 13, (10<<4)+14);
}
#define init_7_13_10_15
void FASTCALL call_7_13_10_15(SH2_struct* sh)
{
  SH2addi(sh, 13, (10<<4)+15);
}
#define init_7_13_11_0
void FASTCALL call_7_13_11_0(SH2_struct* sh)
{
  SH2addi(sh, 13, (11<<4)+0);
}
#define init_7_13_11_1
void FASTCALL call_7_13_11_1(SH2_struct* sh)
{
  SH2addi(sh, 13, (11<<4)+1);
}
#define init_7_13_11_2
void FASTCALL call_7_13_11_2(SH2_struct* sh)
{
  SH2addi(sh, 13, (11<<4)+2);
}
#define init_7_13_11_3
void FASTCALL call_7_13_11_3(SH2_struct* sh)
{
  SH2addi(sh, 13, (11<<4)+3);
}
#define init_7_13_11_4
void FASTCALL call_7_13_11_4(SH2_struct* sh)
{
  SH2addi(sh, 13, (11<<4)+4);
}
#define init_7_13_11_5
void FASTCALL call_7_13_11_5(SH2_struct* sh)
{
  SH2addi(sh, 13, (11<<4)+5);
}
#define init_7_13_11_6
void FASTCALL call_7_13_11_6(SH2_struct* sh)
{
  SH2addi(sh, 13, (11<<4)+6);
}
#define init_7_13_11_7
void FASTCALL call_7_13_11_7(SH2_struct* sh)
{
  SH2addi(sh, 13, (11<<4)+7);
}
#define init_7_13_11_8
void FASTCALL call_7_13_11_8(SH2_struct* sh)
{
  SH2addi(sh, 13, (11<<4)+8);
}
#define init_7_13_11_9
void FASTCALL call_7_13_11_9(SH2_struct* sh)
{
  SH2addi(sh, 13, (11<<4)+9);
}
#define init_7_13_11_10
void FASTCALL call_7_13_11_10(SH2_struct* sh)
{
  SH2addi(sh, 13, (11<<4)+10);
}
#define init_7_13_11_11
void FASTCALL call_7_13_11_11(SH2_struct* sh)
{
  SH2addi(sh, 13, (11<<4)+11);
}
#define init_7_13_11_12
void FASTCALL call_7_13_11_12(SH2_struct* sh)
{
  SH2addi(sh, 13, (11<<4)+12);
}
#define init_7_13_11_13
void FASTCALL call_7_13_11_13(SH2_struct* sh)
{
  SH2addi(sh, 13, (11<<4)+13);
}
#define init_7_13_11_14
void FASTCALL call_7_13_11_14(SH2_struct* sh)
{
  SH2addi(sh, 13, (11<<4)+14);
}
#define init_7_13_11_15
void FASTCALL call_7_13_11_15(SH2_struct* sh)
{
  SH2addi(sh, 13, (11<<4)+15);
}
#define init_7_13_12_0
void FASTCALL call_7_13_12_0(SH2_struct* sh)
{
  SH2addi(sh, 13, (12<<4)+0);
}
#define init_7_13_12_1
void FASTCALL call_7_13_12_1(SH2_struct* sh)
{
  SH2addi(sh, 13, (12<<4)+1);
}
#define init_7_13_12_2
void FASTCALL call_7_13_12_2(SH2_struct* sh)
{
  SH2addi(sh, 13, (12<<4)+2);
}
#define init_7_13_12_3
void FASTCALL call_7_13_12_3(SH2_struct* sh)
{
  SH2addi(sh, 13, (12<<4)+3);
}
#define init_7_13_12_4
void FASTCALL call_7_13_12_4(SH2_struct* sh)
{
  SH2addi(sh, 13, (12<<4)+4);
}
#define init_7_13_12_5
void FASTCALL call_7_13_12_5(SH2_struct* sh)
{
  SH2addi(sh, 13, (12<<4)+5);
}
#define init_7_13_12_6
void FASTCALL call_7_13_12_6(SH2_struct* sh)
{
  SH2addi(sh, 13, (12<<4)+6);
}
#define init_7_13_12_7
void FASTCALL call_7_13_12_7(SH2_struct* sh)
{
  SH2addi(sh, 13, (12<<4)+7);
}
#define init_7_13_12_8
void FASTCALL call_7_13_12_8(SH2_struct* sh)
{
  SH2addi(sh, 13, (12<<4)+8);
}
#define init_7_13_12_9
void FASTCALL call_7_13_12_9(SH2_struct* sh)
{
  SH2addi(sh, 13, (12<<4)+9);
}
#define init_7_13_12_10
void FASTCALL call_7_13_12_10(SH2_struct* sh)
{
  SH2addi(sh, 13, (12<<4)+10);
}
#define init_7_13_12_11
void FASTCALL call_7_13_12_11(SH2_struct* sh)
{
  SH2addi(sh, 13, (12<<4)+11);
}
#define init_7_13_12_12
void FASTCALL call_7_13_12_12(SH2_struct* sh)
{
  SH2addi(sh, 13, (12<<4)+12);
}
#define init_7_13_12_13
void FASTCALL call_7_13_12_13(SH2_struct* sh)
{
  SH2addi(sh, 13, (12<<4)+13);
}
#define init_7_13_12_14
void FASTCALL call_7_13_12_14(SH2_struct* sh)
{
  SH2addi(sh, 13, (12<<4)+14);
}
#define init_7_13_12_15
void FASTCALL call_7_13_12_15(SH2_struct* sh)
{
  SH2addi(sh, 13, (12<<4)+15);
}
#define init_7_13_13_0
void FASTCALL call_7_13_13_0(SH2_struct* sh)
{
  SH2addi(sh, 13, (13<<4)+0);
}
#define init_7_13_13_1
void FASTCALL call_7_13_13_1(SH2_struct* sh)
{
  SH2addi(sh, 13, (13<<4)+1);
}
#define init_7_13_13_2
void FASTCALL call_7_13_13_2(SH2_struct* sh)
{
  SH2addi(sh, 13, (13<<4)+2);
}
#define init_7_13_13_3
void FASTCALL call_7_13_13_3(SH2_struct* sh)
{
  SH2addi(sh, 13, (13<<4)+3);
}
#define init_7_13_13_4
void FASTCALL call_7_13_13_4(SH2_struct* sh)
{
  SH2addi(sh, 13, (13<<4)+4);
}
#define init_7_13_13_5
void FASTCALL call_7_13_13_5(SH2_struct* sh)
{
  SH2addi(sh, 13, (13<<4)+5);
}
#define init_7_13_13_6
void FASTCALL call_7_13_13_6(SH2_struct* sh)
{
  SH2addi(sh, 13, (13<<4)+6);
}
#define init_7_13_13_7
void FASTCALL call_7_13_13_7(SH2_struct* sh)
{
  SH2addi(sh, 13, (13<<4)+7);
}
#define init_7_13_13_8
void FASTCALL call_7_13_13_8(SH2_struct* sh)
{
  SH2addi(sh, 13, (13<<4)+8);
}
#define init_7_13_13_9
void FASTCALL call_7_13_13_9(SH2_struct* sh)
{
  SH2addi(sh, 13, (13<<4)+9);
}
#define init_7_13_13_10
void FASTCALL call_7_13_13_10(SH2_struct* sh)
{
  SH2addi(sh, 13, (13<<4)+10);
}
#define init_7_13_13_11
void FASTCALL call_7_13_13_11(SH2_struct* sh)
{
  SH2addi(sh, 13, (13<<4)+11);
}
#define init_7_13_13_12
void FASTCALL call_7_13_13_12(SH2_struct* sh)
{
  SH2addi(sh, 13, (13<<4)+12);
}
#define init_7_13_13_13
void FASTCALL call_7_13_13_13(SH2_struct* sh)
{
  SH2addi(sh, 13, (13<<4)+13);
}
#define init_7_13_13_14
void FASTCALL call_7_13_13_14(SH2_struct* sh)
{
  SH2addi(sh, 13, (13<<4)+14);
}
#define init_7_13_13_15
void FASTCALL call_7_13_13_15(SH2_struct* sh)
{
  SH2addi(sh, 13, (13<<4)+15);
}
#define init_7_13_14_0
void FASTCALL call_7_13_14_0(SH2_struct* sh)
{
  SH2addi(sh, 13, (14<<4)+0);
}
#define init_7_13_14_1
void FASTCALL call_7_13_14_1(SH2_struct* sh)
{
  SH2addi(sh, 13, (14<<4)+1);
}
#define init_7_13_14_2
void FASTCALL call_7_13_14_2(SH2_struct* sh)
{
  SH2addi(sh, 13, (14<<4)+2);
}
#define init_7_13_14_3
void FASTCALL call_7_13_14_3(SH2_struct* sh)
{
  SH2addi(sh, 13, (14<<4)+3);
}
#define init_7_13_14_4
void FASTCALL call_7_13_14_4(SH2_struct* sh)
{
  SH2addi(sh, 13, (14<<4)+4);
}
#define init_7_13_14_5
void FASTCALL call_7_13_14_5(SH2_struct* sh)
{
  SH2addi(sh, 13, (14<<4)+5);
}
#define init_7_13_14_6
void FASTCALL call_7_13_14_6(SH2_struct* sh)
{
  SH2addi(sh, 13, (14<<4)+6);
}
#define init_7_13_14_7
void FASTCALL call_7_13_14_7(SH2_struct* sh)
{
  SH2addi(sh, 13, (14<<4)+7);
}
#define init_7_13_14_8
void FASTCALL call_7_13_14_8(SH2_struct* sh)
{
  SH2addi(sh, 13, (14<<4)+8);
}
#define init_7_13_14_9
void FASTCALL call_7_13_14_9(SH2_struct* sh)
{
  SH2addi(sh, 13, (14<<4)+9);
}
#define init_7_13_14_10
void FASTCALL call_7_13_14_10(SH2_struct* sh)
{
  SH2addi(sh, 13, (14<<4)+10);
}
#define init_7_13_14_11
void FASTCALL call_7_13_14_11(SH2_struct* sh)
{
  SH2addi(sh, 13, (14<<4)+11);
}
#define init_7_13_14_12
void FASTCALL call_7_13_14_12(SH2_struct* sh)
{
  SH2addi(sh, 13, (14<<4)+12);
}
#define init_7_13_14_13
void FASTCALL call_7_13_14_13(SH2_struct* sh)
{
  SH2addi(sh, 13, (14<<4)+13);
}
#define init_7_13_14_14
void FASTCALL call_7_13_14_14(SH2_struct* sh)
{
  SH2addi(sh, 13, (14<<4)+14);
}
#define init_7_13_14_15
void FASTCALL call_7_13_14_15(SH2_struct* sh)
{
  SH2addi(sh, 13, (14<<4)+15);
}
#define init_7_13_15_0
void FASTCALL call_7_13_15_0(SH2_struct* sh)
{
  SH2addi(sh, 13, (15<<4)+0);
}
#define init_7_13_15_1
void FASTCALL call_7_13_15_1(SH2_struct* sh)
{
  SH2addi(sh, 13, (15<<4)+1);
}
#define init_7_13_15_2
void FASTCALL call_7_13_15_2(SH2_struct* sh)
{
  SH2addi(sh, 13, (15<<4)+2);
}
#define init_7_13_15_3
void FASTCALL call_7_13_15_3(SH2_struct* sh)
{
  SH2addi(sh, 13, (15<<4)+3);
}
#define init_7_13_15_4
void FASTCALL call_7_13_15_4(SH2_struct* sh)
{
  SH2addi(sh, 13, (15<<4)+4);
}
#define init_7_13_15_5
void FASTCALL call_7_13_15_5(SH2_struct* sh)
{
  SH2addi(sh, 13, (15<<4)+5);
}
#define init_7_13_15_6
void FASTCALL call_7_13_15_6(SH2_struct* sh)
{
  SH2addi(sh, 13, (15<<4)+6);
}
#define init_7_13_15_7
void FASTCALL call_7_13_15_7(SH2_struct* sh)
{
  SH2addi(sh, 13, (15<<4)+7);
}
#define init_7_13_15_8
void FASTCALL call_7_13_15_8(SH2_struct* sh)
{
  SH2addi(sh, 13, (15<<4)+8);
}
#define init_7_13_15_9
void FASTCALL call_7_13_15_9(SH2_struct* sh)
{
  SH2addi(sh, 13, (15<<4)+9);
}
#define init_7_13_15_10
void FASTCALL call_7_13_15_10(SH2_struct* sh)
{
  SH2addi(sh, 13, (15<<4)+10);
}
#define init_7_13_15_11
void FASTCALL call_7_13_15_11(SH2_struct* sh)
{
  SH2addi(sh, 13, (15<<4)+11);
}
#define init_7_13_15_12
void FASTCALL call_7_13_15_12(SH2_struct* sh)
{
  SH2addi(sh, 13, (15<<4)+12);
}
#define init_7_13_15_13
void FASTCALL call_7_13_15_13(SH2_struct* sh)
{
  SH2addi(sh, 13, (15<<4)+13);
}
#define init_7_13_15_14
void FASTCALL call_7_13_15_14(SH2_struct* sh)
{
  SH2addi(sh, 13, (15<<4)+14);
}
#define init_7_13_15_15
void FASTCALL call_7_13_15_15(SH2_struct* sh)
{
  SH2addi(sh, 13, (15<<4)+15);
}
#define init_7_14_0_0
void FASTCALL call_7_14_0_0(SH2_struct* sh)
{
  SH2addi(sh, 14, (0<<4)+0);
}
#define init_7_14_0_1
void FASTCALL call_7_14_0_1(SH2_struct* sh)
{
  SH2addi(sh, 14, (0<<4)+1);
}
#define init_7_14_0_2
void FASTCALL call_7_14_0_2(SH2_struct* sh)
{
  SH2addi(sh, 14, (0<<4)+2);
}
#define init_7_14_0_3
void FASTCALL call_7_14_0_3(SH2_struct* sh)
{
  SH2addi(sh, 14, (0<<4)+3);
}
#define init_7_14_0_4
void FASTCALL call_7_14_0_4(SH2_struct* sh)
{
  SH2addi(sh, 14, (0<<4)+4);
}
#define init_7_14_0_5
void FASTCALL call_7_14_0_5(SH2_struct* sh)
{
  SH2addi(sh, 14, (0<<4)+5);
}
#define init_7_14_0_6
void FASTCALL call_7_14_0_6(SH2_struct* sh)
{
  SH2addi(sh, 14, (0<<4)+6);
}
#define init_7_14_0_7
void FASTCALL call_7_14_0_7(SH2_struct* sh)
{
  SH2addi(sh, 14, (0<<4)+7);
}
#define init_7_14_0_8
void FASTCALL call_7_14_0_8(SH2_struct* sh)
{
  SH2addi(sh, 14, (0<<4)+8);
}
#define init_7_14_0_9
void FASTCALL call_7_14_0_9(SH2_struct* sh)
{
  SH2addi(sh, 14, (0<<4)+9);
}
#define init_7_14_0_10
void FASTCALL call_7_14_0_10(SH2_struct* sh)
{
  SH2addi(sh, 14, (0<<4)+10);
}
#define init_7_14_0_11
void FASTCALL call_7_14_0_11(SH2_struct* sh)
{
  SH2addi(sh, 14, (0<<4)+11);
}
#define init_7_14_0_12
void FASTCALL call_7_14_0_12(SH2_struct* sh)
{
  SH2addi(sh, 14, (0<<4)+12);
}
#define init_7_14_0_13
void FASTCALL call_7_14_0_13(SH2_struct* sh)
{
  SH2addi(sh, 14, (0<<4)+13);
}
#define init_7_14_0_14
void FASTCALL call_7_14_0_14(SH2_struct* sh)
{
  SH2addi(sh, 14, (0<<4)+14);
}
#define init_7_14_0_15
void FASTCALL call_7_14_0_15(SH2_struct* sh)
{
  SH2addi(sh, 14, (0<<4)+15);
}
#define init_7_14_1_0
void FASTCALL call_7_14_1_0(SH2_struct* sh)
{
  SH2addi(sh, 14, (1<<4)+0);
}
#define init_7_14_1_1
void FASTCALL call_7_14_1_1(SH2_struct* sh)
{
  SH2addi(sh, 14, (1<<4)+1);
}
#define init_7_14_1_2
void FASTCALL call_7_14_1_2(SH2_struct* sh)
{
  SH2addi(sh, 14, (1<<4)+2);
}
#define init_7_14_1_3
void FASTCALL call_7_14_1_3(SH2_struct* sh)
{
  SH2addi(sh, 14, (1<<4)+3);
}
#define init_7_14_1_4
void FASTCALL call_7_14_1_4(SH2_struct* sh)
{
  SH2addi(sh, 14, (1<<4)+4);
}
#define init_7_14_1_5
void FASTCALL call_7_14_1_5(SH2_struct* sh)
{
  SH2addi(sh, 14, (1<<4)+5);
}
#define init_7_14_1_6
void FASTCALL call_7_14_1_6(SH2_struct* sh)
{
  SH2addi(sh, 14, (1<<4)+6);
}
#define init_7_14_1_7
void FASTCALL call_7_14_1_7(SH2_struct* sh)
{
  SH2addi(sh, 14, (1<<4)+7);
}
#define init_7_14_1_8
void FASTCALL call_7_14_1_8(SH2_struct* sh)
{
  SH2addi(sh, 14, (1<<4)+8);
}
#define init_7_14_1_9
void FASTCALL call_7_14_1_9(SH2_struct* sh)
{
  SH2addi(sh, 14, (1<<4)+9);
}
#define init_7_14_1_10
void FASTCALL call_7_14_1_10(SH2_struct* sh)
{
  SH2addi(sh, 14, (1<<4)+10);
}
#define init_7_14_1_11
void FASTCALL call_7_14_1_11(SH2_struct* sh)
{
  SH2addi(sh, 14, (1<<4)+11);
}
#define init_7_14_1_12
void FASTCALL call_7_14_1_12(SH2_struct* sh)
{
  SH2addi(sh, 14, (1<<4)+12);
}
#define init_7_14_1_13
void FASTCALL call_7_14_1_13(SH2_struct* sh)
{
  SH2addi(sh, 14, (1<<4)+13);
}
#define init_7_14_1_14
void FASTCALL call_7_14_1_14(SH2_struct* sh)
{
  SH2addi(sh, 14, (1<<4)+14);
}
#define init_7_14_1_15
void FASTCALL call_7_14_1_15(SH2_struct* sh)
{
  SH2addi(sh, 14, (1<<4)+15);
}
#define init_7_14_2_0
void FASTCALL call_7_14_2_0(SH2_struct* sh)
{
  SH2addi(sh, 14, (2<<4)+0);
}
#define init_7_14_2_1
void FASTCALL call_7_14_2_1(SH2_struct* sh)
{
  SH2addi(sh, 14, (2<<4)+1);
}
#define init_7_14_2_2
void FASTCALL call_7_14_2_2(SH2_struct* sh)
{
  SH2addi(sh, 14, (2<<4)+2);
}
#define init_7_14_2_3
void FASTCALL call_7_14_2_3(SH2_struct* sh)
{
  SH2addi(sh, 14, (2<<4)+3);
}
#define init_7_14_2_4
void FASTCALL call_7_14_2_4(SH2_struct* sh)
{
  SH2addi(sh, 14, (2<<4)+4);
}
#define init_7_14_2_5
void FASTCALL call_7_14_2_5(SH2_struct* sh)
{
  SH2addi(sh, 14, (2<<4)+5);
}
#define init_7_14_2_6
void FASTCALL call_7_14_2_6(SH2_struct* sh)
{
  SH2addi(sh, 14, (2<<4)+6);
}
#define init_7_14_2_7
void FASTCALL call_7_14_2_7(SH2_struct* sh)
{
  SH2addi(sh, 14, (2<<4)+7);
}
#define init_7_14_2_8
void FASTCALL call_7_14_2_8(SH2_struct* sh)
{
  SH2addi(sh, 14, (2<<4)+8);
}
#define init_7_14_2_9
void FASTCALL call_7_14_2_9(SH2_struct* sh)
{
  SH2addi(sh, 14, (2<<4)+9);
}
#define init_7_14_2_10
void FASTCALL call_7_14_2_10(SH2_struct* sh)
{
  SH2addi(sh, 14, (2<<4)+10);
}
#define init_7_14_2_11
void FASTCALL call_7_14_2_11(SH2_struct* sh)
{
  SH2addi(sh, 14, (2<<4)+11);
}
#define init_7_14_2_12
void FASTCALL call_7_14_2_12(SH2_struct* sh)
{
  SH2addi(sh, 14, (2<<4)+12);
}
#define init_7_14_2_13
void FASTCALL call_7_14_2_13(SH2_struct* sh)
{
  SH2addi(sh, 14, (2<<4)+13);
}
#define init_7_14_2_14
void FASTCALL call_7_14_2_14(SH2_struct* sh)
{
  SH2addi(sh, 14, (2<<4)+14);
}
#define init_7_14_2_15
void FASTCALL call_7_14_2_15(SH2_struct* sh)
{
  SH2addi(sh, 14, (2<<4)+15);
}
#define init_7_14_3_0
void FASTCALL call_7_14_3_0(SH2_struct* sh)
{
  SH2addi(sh, 14, (3<<4)+0);
}
#define init_7_14_3_1
void FASTCALL call_7_14_3_1(SH2_struct* sh)
{
  SH2addi(sh, 14, (3<<4)+1);
}
#define init_7_14_3_2
void FASTCALL call_7_14_3_2(SH2_struct* sh)
{
  SH2addi(sh, 14, (3<<4)+2);
}
#define init_7_14_3_3
void FASTCALL call_7_14_3_3(SH2_struct* sh)
{
  SH2addi(sh, 14, (3<<4)+3);
}
#define init_7_14_3_4
void FASTCALL call_7_14_3_4(SH2_struct* sh)
{
  SH2addi(sh, 14, (3<<4)+4);
}
#define init_7_14_3_5
void FASTCALL call_7_14_3_5(SH2_struct* sh)
{
  SH2addi(sh, 14, (3<<4)+5);
}
#define init_7_14_3_6
void FASTCALL call_7_14_3_6(SH2_struct* sh)
{
  SH2addi(sh, 14, (3<<4)+6);
}
#define init_7_14_3_7
void FASTCALL call_7_14_3_7(SH2_struct* sh)
{
  SH2addi(sh, 14, (3<<4)+7);
}
#define init_7_14_3_8
void FASTCALL call_7_14_3_8(SH2_struct* sh)
{
  SH2addi(sh, 14, (3<<4)+8);
}
#define init_7_14_3_9
void FASTCALL call_7_14_3_9(SH2_struct* sh)
{
  SH2addi(sh, 14, (3<<4)+9);
}
#define init_7_14_3_10
void FASTCALL call_7_14_3_10(SH2_struct* sh)
{
  SH2addi(sh, 14, (3<<4)+10);
}
#define init_7_14_3_11
void FASTCALL call_7_14_3_11(SH2_struct* sh)
{
  SH2addi(sh, 14, (3<<4)+11);
}
#define init_7_14_3_12
void FASTCALL call_7_14_3_12(SH2_struct* sh)
{
  SH2addi(sh, 14, (3<<4)+12);
}
#define init_7_14_3_13
void FASTCALL call_7_14_3_13(SH2_struct* sh)
{
  SH2addi(sh, 14, (3<<4)+13);
}
#define init_7_14_3_14
void FASTCALL call_7_14_3_14(SH2_struct* sh)
{
  SH2addi(sh, 14, (3<<4)+14);
}
#define init_7_14_3_15
void FASTCALL call_7_14_3_15(SH2_struct* sh)
{
  SH2addi(sh, 14, (3<<4)+15);
}
#define init_7_14_4_0
void FASTCALL call_7_14_4_0(SH2_struct* sh)
{
  SH2addi(sh, 14, (4<<4)+0);
}
#define init_7_14_4_1
void FASTCALL call_7_14_4_1(SH2_struct* sh)
{
  SH2addi(sh, 14, (4<<4)+1);
}
#define init_7_14_4_2
void FASTCALL call_7_14_4_2(SH2_struct* sh)
{
  SH2addi(sh, 14, (4<<4)+2);
}
#define init_7_14_4_3
void FASTCALL call_7_14_4_3(SH2_struct* sh)
{
  SH2addi(sh, 14, (4<<4)+3);
}
#define init_7_14_4_4
void FASTCALL call_7_14_4_4(SH2_struct* sh)
{
  SH2addi(sh, 14, (4<<4)+4);
}
#define init_7_14_4_5
void FASTCALL call_7_14_4_5(SH2_struct* sh)
{
  SH2addi(sh, 14, (4<<4)+5);
}
#define init_7_14_4_6
void FASTCALL call_7_14_4_6(SH2_struct* sh)
{
  SH2addi(sh, 14, (4<<4)+6);
}
#define init_7_14_4_7
void FASTCALL call_7_14_4_7(SH2_struct* sh)
{
  SH2addi(sh, 14, (4<<4)+7);
}
#define init_7_14_4_8
void FASTCALL call_7_14_4_8(SH2_struct* sh)
{
  SH2addi(sh, 14, (4<<4)+8);
}
#define init_7_14_4_9
void FASTCALL call_7_14_4_9(SH2_struct* sh)
{
  SH2addi(sh, 14, (4<<4)+9);
}
#define init_7_14_4_10
void FASTCALL call_7_14_4_10(SH2_struct* sh)
{
  SH2addi(sh, 14, (4<<4)+10);
}
#define init_7_14_4_11
void FASTCALL call_7_14_4_11(SH2_struct* sh)
{
  SH2addi(sh, 14, (4<<4)+11);
}
#define init_7_14_4_12
void FASTCALL call_7_14_4_12(SH2_struct* sh)
{
  SH2addi(sh, 14, (4<<4)+12);
}
#define init_7_14_4_13
void FASTCALL call_7_14_4_13(SH2_struct* sh)
{
  SH2addi(sh, 14, (4<<4)+13);
}
#define init_7_14_4_14
void FASTCALL call_7_14_4_14(SH2_struct* sh)
{
  SH2addi(sh, 14, (4<<4)+14);
}
#define init_7_14_4_15
void FASTCALL call_7_14_4_15(SH2_struct* sh)
{
  SH2addi(sh, 14, (4<<4)+15);
}
#define init_7_14_5_0
void FASTCALL call_7_14_5_0(SH2_struct* sh)
{
  SH2addi(sh, 14, (5<<4)+0);
}
#define init_7_14_5_1
void FASTCALL call_7_14_5_1(SH2_struct* sh)
{
  SH2addi(sh, 14, (5<<4)+1);
}
#define init_7_14_5_2
void FASTCALL call_7_14_5_2(SH2_struct* sh)
{
  SH2addi(sh, 14, (5<<4)+2);
}
#define init_7_14_5_3
void FASTCALL call_7_14_5_3(SH2_struct* sh)
{
  SH2addi(sh, 14, (5<<4)+3);
}
#define init_7_14_5_4
void FASTCALL call_7_14_5_4(SH2_struct* sh)
{
  SH2addi(sh, 14, (5<<4)+4);
}
#define init_7_14_5_5
void FASTCALL call_7_14_5_5(SH2_struct* sh)
{
  SH2addi(sh, 14, (5<<4)+5);
}
#define init_7_14_5_6
void FASTCALL call_7_14_5_6(SH2_struct* sh)
{
  SH2addi(sh, 14, (5<<4)+6);
}
#define init_7_14_5_7
void FASTCALL call_7_14_5_7(SH2_struct* sh)
{
  SH2addi(sh, 14, (5<<4)+7);
}
#define init_7_14_5_8
void FASTCALL call_7_14_5_8(SH2_struct* sh)
{
  SH2addi(sh, 14, (5<<4)+8);
}
#define init_7_14_5_9
void FASTCALL call_7_14_5_9(SH2_struct* sh)
{
  SH2addi(sh, 14, (5<<4)+9);
}
#define init_7_14_5_10
void FASTCALL call_7_14_5_10(SH2_struct* sh)
{
  SH2addi(sh, 14, (5<<4)+10);
}
#define init_7_14_5_11
void FASTCALL call_7_14_5_11(SH2_struct* sh)
{
  SH2addi(sh, 14, (5<<4)+11);
}
#define init_7_14_5_12
void FASTCALL call_7_14_5_12(SH2_struct* sh)
{
  SH2addi(sh, 14, (5<<4)+12);
}
#define init_7_14_5_13
void FASTCALL call_7_14_5_13(SH2_struct* sh)
{
  SH2addi(sh, 14, (5<<4)+13);
}
#define init_7_14_5_14
void FASTCALL call_7_14_5_14(SH2_struct* sh)
{
  SH2addi(sh, 14, (5<<4)+14);
}
#define init_7_14_5_15
void FASTCALL call_7_14_5_15(SH2_struct* sh)
{
  SH2addi(sh, 14, (5<<4)+15);
}
#define init_7_14_6_0
void FASTCALL call_7_14_6_0(SH2_struct* sh)
{
  SH2addi(sh, 14, (6<<4)+0);
}
#define init_7_14_6_1
void FASTCALL call_7_14_6_1(SH2_struct* sh)
{
  SH2addi(sh, 14, (6<<4)+1);
}
#define init_7_14_6_2
void FASTCALL call_7_14_6_2(SH2_struct* sh)
{
  SH2addi(sh, 14, (6<<4)+2);
}
#define init_7_14_6_3
void FASTCALL call_7_14_6_3(SH2_struct* sh)
{
  SH2addi(sh, 14, (6<<4)+3);
}
#define init_7_14_6_4
void FASTCALL call_7_14_6_4(SH2_struct* sh)
{
  SH2addi(sh, 14, (6<<4)+4);
}
#define init_7_14_6_5
void FASTCALL call_7_14_6_5(SH2_struct* sh)
{
  SH2addi(sh, 14, (6<<4)+5);
}
#define init_7_14_6_6
void FASTCALL call_7_14_6_6(SH2_struct* sh)
{
  SH2addi(sh, 14, (6<<4)+6);
}
#define init_7_14_6_7
void FASTCALL call_7_14_6_7(SH2_struct* sh)
{
  SH2addi(sh, 14, (6<<4)+7);
}
#define init_7_14_6_8
void FASTCALL call_7_14_6_8(SH2_struct* sh)
{
  SH2addi(sh, 14, (6<<4)+8);
}
#define init_7_14_6_9
void FASTCALL call_7_14_6_9(SH2_struct* sh)
{
  SH2addi(sh, 14, (6<<4)+9);
}
#define init_7_14_6_10
void FASTCALL call_7_14_6_10(SH2_struct* sh)
{
  SH2addi(sh, 14, (6<<4)+10);
}
#define init_7_14_6_11
void FASTCALL call_7_14_6_11(SH2_struct* sh)
{
  SH2addi(sh, 14, (6<<4)+11);
}
#define init_7_14_6_12
void FASTCALL call_7_14_6_12(SH2_struct* sh)
{
  SH2addi(sh, 14, (6<<4)+12);
}
#define init_7_14_6_13
void FASTCALL call_7_14_6_13(SH2_struct* sh)
{
  SH2addi(sh, 14, (6<<4)+13);
}
#define init_7_14_6_14
void FASTCALL call_7_14_6_14(SH2_struct* sh)
{
  SH2addi(sh, 14, (6<<4)+14);
}
#define init_7_14_6_15
void FASTCALL call_7_14_6_15(SH2_struct* sh)
{
  SH2addi(sh, 14, (6<<4)+15);
}
#define init_7_14_7_0
void FASTCALL call_7_14_7_0(SH2_struct* sh)
{
  SH2addi(sh, 14, (7<<4)+0);
}
#define init_7_14_7_1
void FASTCALL call_7_14_7_1(SH2_struct* sh)
{
  SH2addi(sh, 14, (7<<4)+1);
}
#define init_7_14_7_2
void FASTCALL call_7_14_7_2(SH2_struct* sh)
{
  SH2addi(sh, 14, (7<<4)+2);
}
#define init_7_14_7_3
void FASTCALL call_7_14_7_3(SH2_struct* sh)
{
  SH2addi(sh, 14, (7<<4)+3);
}
#define init_7_14_7_4
void FASTCALL call_7_14_7_4(SH2_struct* sh)
{
  SH2addi(sh, 14, (7<<4)+4);
}
#define init_7_14_7_5
void FASTCALL call_7_14_7_5(SH2_struct* sh)
{
  SH2addi(sh, 14, (7<<4)+5);
}
#define init_7_14_7_6
void FASTCALL call_7_14_7_6(SH2_struct* sh)
{
  SH2addi(sh, 14, (7<<4)+6);
}
#define init_7_14_7_7
void FASTCALL call_7_14_7_7(SH2_struct* sh)
{
  SH2addi(sh, 14, (7<<4)+7);
}
#define init_7_14_7_8
void FASTCALL call_7_14_7_8(SH2_struct* sh)
{
  SH2addi(sh, 14, (7<<4)+8);
}
#define init_7_14_7_9
void FASTCALL call_7_14_7_9(SH2_struct* sh)
{
  SH2addi(sh, 14, (7<<4)+9);
}
#define init_7_14_7_10
void FASTCALL call_7_14_7_10(SH2_struct* sh)
{
  SH2addi(sh, 14, (7<<4)+10);
}
#define init_7_14_7_11
void FASTCALL call_7_14_7_11(SH2_struct* sh)
{
  SH2addi(sh, 14, (7<<4)+11);
}
#define init_7_14_7_12
void FASTCALL call_7_14_7_12(SH2_struct* sh)
{
  SH2addi(sh, 14, (7<<4)+12);
}
#define init_7_14_7_13
void FASTCALL call_7_14_7_13(SH2_struct* sh)
{
  SH2addi(sh, 14, (7<<4)+13);
}
#define init_7_14_7_14
void FASTCALL call_7_14_7_14(SH2_struct* sh)
{
  SH2addi(sh, 14, (7<<4)+14);
}
#define init_7_14_7_15
void FASTCALL call_7_14_7_15(SH2_struct* sh)
{
  SH2addi(sh, 14, (7<<4)+15);
}
#define init_7_14_8_0
void FASTCALL call_7_14_8_0(SH2_struct* sh)
{
  SH2addi(sh, 14, (8<<4)+0);
}
#define init_7_14_8_1
void FASTCALL call_7_14_8_1(SH2_struct* sh)
{
  SH2addi(sh, 14, (8<<4)+1);
}
#define init_7_14_8_2
void FASTCALL call_7_14_8_2(SH2_struct* sh)
{
  SH2addi(sh, 14, (8<<4)+2);
}
#define init_7_14_8_3
void FASTCALL call_7_14_8_3(SH2_struct* sh)
{
  SH2addi(sh, 14, (8<<4)+3);
}
#define init_7_14_8_4
void FASTCALL call_7_14_8_4(SH2_struct* sh)
{
  SH2addi(sh, 14, (8<<4)+4);
}
#define init_7_14_8_5
void FASTCALL call_7_14_8_5(SH2_struct* sh)
{
  SH2addi(sh, 14, (8<<4)+5);
}
#define init_7_14_8_6
void FASTCALL call_7_14_8_6(SH2_struct* sh)
{
  SH2addi(sh, 14, (8<<4)+6);
}
#define init_7_14_8_7
void FASTCALL call_7_14_8_7(SH2_struct* sh)
{
  SH2addi(sh, 14, (8<<4)+7);
}
#define init_7_14_8_8
void FASTCALL call_7_14_8_8(SH2_struct* sh)
{
  SH2addi(sh, 14, (8<<4)+8);
}
#define init_7_14_8_9
void FASTCALL call_7_14_8_9(SH2_struct* sh)
{
  SH2addi(sh, 14, (8<<4)+9);
}
#define init_7_14_8_10
void FASTCALL call_7_14_8_10(SH2_struct* sh)
{
  SH2addi(sh, 14, (8<<4)+10);
}
#define init_7_14_8_11
void FASTCALL call_7_14_8_11(SH2_struct* sh)
{
  SH2addi(sh, 14, (8<<4)+11);
}
#define init_7_14_8_12
void FASTCALL call_7_14_8_12(SH2_struct* sh)
{
  SH2addi(sh, 14, (8<<4)+12);
}
#define init_7_14_8_13
void FASTCALL call_7_14_8_13(SH2_struct* sh)
{
  SH2addi(sh, 14, (8<<4)+13);
}
#define init_7_14_8_14
void FASTCALL call_7_14_8_14(SH2_struct* sh)
{
  SH2addi(sh, 14, (8<<4)+14);
}
#define init_7_14_8_15
void FASTCALL call_7_14_8_15(SH2_struct* sh)
{
  SH2addi(sh, 14, (8<<4)+15);
}
#define init_7_14_9_0
void FASTCALL call_7_14_9_0(SH2_struct* sh)
{
  SH2addi(sh, 14, (9<<4)+0);
}
#define init_7_14_9_1
void FASTCALL call_7_14_9_1(SH2_struct* sh)
{
  SH2addi(sh, 14, (9<<4)+1);
}
#define init_7_14_9_2
void FASTCALL call_7_14_9_2(SH2_struct* sh)
{
  SH2addi(sh, 14, (9<<4)+2);
}
#define init_7_14_9_3
void FASTCALL call_7_14_9_3(SH2_struct* sh)
{
  SH2addi(sh, 14, (9<<4)+3);
}
#define init_7_14_9_4
void FASTCALL call_7_14_9_4(SH2_struct* sh)
{
  SH2addi(sh, 14, (9<<4)+4);
}
#define init_7_14_9_5
void FASTCALL call_7_14_9_5(SH2_struct* sh)
{
  SH2addi(sh, 14, (9<<4)+5);
}
#define init_7_14_9_6
void FASTCALL call_7_14_9_6(SH2_struct* sh)
{
  SH2addi(sh, 14, (9<<4)+6);
}
#define init_7_14_9_7
void FASTCALL call_7_14_9_7(SH2_struct* sh)
{
  SH2addi(sh, 14, (9<<4)+7);
}
#define init_7_14_9_8
void FASTCALL call_7_14_9_8(SH2_struct* sh)
{
  SH2addi(sh, 14, (9<<4)+8);
}
#define init_7_14_9_9
void FASTCALL call_7_14_9_9(SH2_struct* sh)
{
  SH2addi(sh, 14, (9<<4)+9);
}
#define init_7_14_9_10
void FASTCALL call_7_14_9_10(SH2_struct* sh)
{
  SH2addi(sh, 14, (9<<4)+10);
}
#define init_7_14_9_11
void FASTCALL call_7_14_9_11(SH2_struct* sh)
{
  SH2addi(sh, 14, (9<<4)+11);
}
#define init_7_14_9_12
void FASTCALL call_7_14_9_12(SH2_struct* sh)
{
  SH2addi(sh, 14, (9<<4)+12);
}
#define init_7_14_9_13
void FASTCALL call_7_14_9_13(SH2_struct* sh)
{
  SH2addi(sh, 14, (9<<4)+13);
}
#define init_7_14_9_14
void FASTCALL call_7_14_9_14(SH2_struct* sh)
{
  SH2addi(sh, 14, (9<<4)+14);
}
#define init_7_14_9_15
void FASTCALL call_7_14_9_15(SH2_struct* sh)
{
  SH2addi(sh, 14, (9<<4)+15);
}
#define init_7_14_10_0
void FASTCALL call_7_14_10_0(SH2_struct* sh)
{
  SH2addi(sh, 14, (10<<4)+0);
}
#define init_7_14_10_1
void FASTCALL call_7_14_10_1(SH2_struct* sh)
{
  SH2addi(sh, 14, (10<<4)+1);
}
#define init_7_14_10_2
void FASTCALL call_7_14_10_2(SH2_struct* sh)
{
  SH2addi(sh, 14, (10<<4)+2);
}
#define init_7_14_10_3
void FASTCALL call_7_14_10_3(SH2_struct* sh)
{
  SH2addi(sh, 14, (10<<4)+3);
}
#define init_7_14_10_4
void FASTCALL call_7_14_10_4(SH2_struct* sh)
{
  SH2addi(sh, 14, (10<<4)+4);
}
#define init_7_14_10_5
void FASTCALL call_7_14_10_5(SH2_struct* sh)
{
  SH2addi(sh, 14, (10<<4)+5);
}
#define init_7_14_10_6
void FASTCALL call_7_14_10_6(SH2_struct* sh)
{
  SH2addi(sh, 14, (10<<4)+6);
}
#define init_7_14_10_7
void FASTCALL call_7_14_10_7(SH2_struct* sh)
{
  SH2addi(sh, 14, (10<<4)+7);
}
#define init_7_14_10_8
void FASTCALL call_7_14_10_8(SH2_struct* sh)
{
  SH2addi(sh, 14, (10<<4)+8);
}
#define init_7_14_10_9
void FASTCALL call_7_14_10_9(SH2_struct* sh)
{
  SH2addi(sh, 14, (10<<4)+9);
}
#define init_7_14_10_10
void FASTCALL call_7_14_10_10(SH2_struct* sh)
{
  SH2addi(sh, 14, (10<<4)+10);
}
#define init_7_14_10_11
void FASTCALL call_7_14_10_11(SH2_struct* sh)
{
  SH2addi(sh, 14, (10<<4)+11);
}
#define init_7_14_10_12
void FASTCALL call_7_14_10_12(SH2_struct* sh)
{
  SH2addi(sh, 14, (10<<4)+12);
}
#define init_7_14_10_13
void FASTCALL call_7_14_10_13(SH2_struct* sh)
{
  SH2addi(sh, 14, (10<<4)+13);
}
#define init_7_14_10_14
void FASTCALL call_7_14_10_14(SH2_struct* sh)
{
  SH2addi(sh, 14, (10<<4)+14);
}
#define init_7_14_10_15
void FASTCALL call_7_14_10_15(SH2_struct* sh)
{
  SH2addi(sh, 14, (10<<4)+15);
}
#define init_7_14_11_0
void FASTCALL call_7_14_11_0(SH2_struct* sh)
{
  SH2addi(sh, 14, (11<<4)+0);
}
#define init_7_14_11_1
void FASTCALL call_7_14_11_1(SH2_struct* sh)
{
  SH2addi(sh, 14, (11<<4)+1);
}
#define init_7_14_11_2
void FASTCALL call_7_14_11_2(SH2_struct* sh)
{
  SH2addi(sh, 14, (11<<4)+2);
}
#define init_7_14_11_3
void FASTCALL call_7_14_11_3(SH2_struct* sh)
{
  SH2addi(sh, 14, (11<<4)+3);
}
#define init_7_14_11_4
void FASTCALL call_7_14_11_4(SH2_struct* sh)
{
  SH2addi(sh, 14, (11<<4)+4);
}
#define init_7_14_11_5
void FASTCALL call_7_14_11_5(SH2_struct* sh)
{
  SH2addi(sh, 14, (11<<4)+5);
}
#define init_7_14_11_6
void FASTCALL call_7_14_11_6(SH2_struct* sh)
{
  SH2addi(sh, 14, (11<<4)+6);
}
#define init_7_14_11_7
void FASTCALL call_7_14_11_7(SH2_struct* sh)
{
  SH2addi(sh, 14, (11<<4)+7);
}
#define init_7_14_11_8
void FASTCALL call_7_14_11_8(SH2_struct* sh)
{
  SH2addi(sh, 14, (11<<4)+8);
}
#define init_7_14_11_9
void FASTCALL call_7_14_11_9(SH2_struct* sh)
{
  SH2addi(sh, 14, (11<<4)+9);
}
#define init_7_14_11_10
void FASTCALL call_7_14_11_10(SH2_struct* sh)
{
  SH2addi(sh, 14, (11<<4)+10);
}
#define init_7_14_11_11
void FASTCALL call_7_14_11_11(SH2_struct* sh)
{
  SH2addi(sh, 14, (11<<4)+11);
}
#define init_7_14_11_12
void FASTCALL call_7_14_11_12(SH2_struct* sh)
{
  SH2addi(sh, 14, (11<<4)+12);
}
#define init_7_14_11_13
void FASTCALL call_7_14_11_13(SH2_struct* sh)
{
  SH2addi(sh, 14, (11<<4)+13);
}
#define init_7_14_11_14
void FASTCALL call_7_14_11_14(SH2_struct* sh)
{
  SH2addi(sh, 14, (11<<4)+14);
}
#define init_7_14_11_15
void FASTCALL call_7_14_11_15(SH2_struct* sh)
{
  SH2addi(sh, 14, (11<<4)+15);
}
#define init_7_14_12_0
void FASTCALL call_7_14_12_0(SH2_struct* sh)
{
  SH2addi(sh, 14, (12<<4)+0);
}
#define init_7_14_12_1
void FASTCALL call_7_14_12_1(SH2_struct* sh)
{
  SH2addi(sh, 14, (12<<4)+1);
}
#define init_7_14_12_2
void FASTCALL call_7_14_12_2(SH2_struct* sh)
{
  SH2addi(sh, 14, (12<<4)+2);
}
#define init_7_14_12_3
void FASTCALL call_7_14_12_3(SH2_struct* sh)
{
  SH2addi(sh, 14, (12<<4)+3);
}
#define init_7_14_12_4
void FASTCALL call_7_14_12_4(SH2_struct* sh)
{
  SH2addi(sh, 14, (12<<4)+4);
}
#define init_7_14_12_5
void FASTCALL call_7_14_12_5(SH2_struct* sh)
{
  SH2addi(sh, 14, (12<<4)+5);
}
#define init_7_14_12_6
void FASTCALL call_7_14_12_6(SH2_struct* sh)
{
  SH2addi(sh, 14, (12<<4)+6);
}
#define init_7_14_12_7
void FASTCALL call_7_14_12_7(SH2_struct* sh)
{
  SH2addi(sh, 14, (12<<4)+7);
}
#define init_7_14_12_8
void FASTCALL call_7_14_12_8(SH2_struct* sh)
{
  SH2addi(sh, 14, (12<<4)+8);
}
#define init_7_14_12_9
void FASTCALL call_7_14_12_9(SH2_struct* sh)
{
  SH2addi(sh, 14, (12<<4)+9);
}
#define init_7_14_12_10
void FASTCALL call_7_14_12_10(SH2_struct* sh)
{
  SH2addi(sh, 14, (12<<4)+10);
}
#define init_7_14_12_11
void FASTCALL call_7_14_12_11(SH2_struct* sh)
{
  SH2addi(sh, 14, (12<<4)+11);
}
#define init_7_14_12_12
void FASTCALL call_7_14_12_12(SH2_struct* sh)
{
  SH2addi(sh, 14, (12<<4)+12);
}
#define init_7_14_12_13
void FASTCALL call_7_14_12_13(SH2_struct* sh)
{
  SH2addi(sh, 14, (12<<4)+13);
}
#define init_7_14_12_14
void FASTCALL call_7_14_12_14(SH2_struct* sh)
{
  SH2addi(sh, 14, (12<<4)+14);
}
#define init_7_14_12_15
void FASTCALL call_7_14_12_15(SH2_struct* sh)
{
  SH2addi(sh, 14, (12<<4)+15);
}
#define init_7_14_13_0
void FASTCALL call_7_14_13_0(SH2_struct* sh)
{
  SH2addi(sh, 14, (13<<4)+0);
}
#define init_7_14_13_1
void FASTCALL call_7_14_13_1(SH2_struct* sh)
{
  SH2addi(sh, 14, (13<<4)+1);
}
#define init_7_14_13_2
void FASTCALL call_7_14_13_2(SH2_struct* sh)
{
  SH2addi(sh, 14, (13<<4)+2);
}
#define init_7_14_13_3
void FASTCALL call_7_14_13_3(SH2_struct* sh)
{
  SH2addi(sh, 14, (13<<4)+3);
}
#define init_7_14_13_4
void FASTCALL call_7_14_13_4(SH2_struct* sh)
{
  SH2addi(sh, 14, (13<<4)+4);
}
#define init_7_14_13_5
void FASTCALL call_7_14_13_5(SH2_struct* sh)
{
  SH2addi(sh, 14, (13<<4)+5);
}
#define init_7_14_13_6
void FASTCALL call_7_14_13_6(SH2_struct* sh)
{
  SH2addi(sh, 14, (13<<4)+6);
}
#define init_7_14_13_7
void FASTCALL call_7_14_13_7(SH2_struct* sh)
{
  SH2addi(sh, 14, (13<<4)+7);
}
#define init_7_14_13_8
void FASTCALL call_7_14_13_8(SH2_struct* sh)
{
  SH2addi(sh, 14, (13<<4)+8);
}
#define init_7_14_13_9
void FASTCALL call_7_14_13_9(SH2_struct* sh)
{
  SH2addi(sh, 14, (13<<4)+9);
}
#define init_7_14_13_10
void FASTCALL call_7_14_13_10(SH2_struct* sh)
{
  SH2addi(sh, 14, (13<<4)+10);
}
#define init_7_14_13_11
void FASTCALL call_7_14_13_11(SH2_struct* sh)
{
  SH2addi(sh, 14, (13<<4)+11);
}
#define init_7_14_13_12
void FASTCALL call_7_14_13_12(SH2_struct* sh)
{
  SH2addi(sh, 14, (13<<4)+12);
}
#define init_7_14_13_13
void FASTCALL call_7_14_13_13(SH2_struct* sh)
{
  SH2addi(sh, 14, (13<<4)+13);
}
#define init_7_14_13_14
void FASTCALL call_7_14_13_14(SH2_struct* sh)
{
  SH2addi(sh, 14, (13<<4)+14);
}
#define init_7_14_13_15
void FASTCALL call_7_14_13_15(SH2_struct* sh)
{
  SH2addi(sh, 14, (13<<4)+15);
}
#define init_7_14_14_0
void FASTCALL call_7_14_14_0(SH2_struct* sh)
{
  SH2addi(sh, 14, (14<<4)+0);
}
#define init_7_14_14_1
void FASTCALL call_7_14_14_1(SH2_struct* sh)
{
  SH2addi(sh, 14, (14<<4)+1);
}
#define init_7_14_14_2
void FASTCALL call_7_14_14_2(SH2_struct* sh)
{
  SH2addi(sh, 14, (14<<4)+2);
}
#define init_7_14_14_3
void FASTCALL call_7_14_14_3(SH2_struct* sh)
{
  SH2addi(sh, 14, (14<<4)+3);
}
#define init_7_14_14_4
void FASTCALL call_7_14_14_4(SH2_struct* sh)
{
  SH2addi(sh, 14, (14<<4)+4);
}
#define init_7_14_14_5
void FASTCALL call_7_14_14_5(SH2_struct* sh)
{
  SH2addi(sh, 14, (14<<4)+5);
}
#define init_7_14_14_6
void FASTCALL call_7_14_14_6(SH2_struct* sh)
{
  SH2addi(sh, 14, (14<<4)+6);
}
#define init_7_14_14_7
void FASTCALL call_7_14_14_7(SH2_struct* sh)
{
  SH2addi(sh, 14, (14<<4)+7);
}
#define init_7_14_14_8
void FASTCALL call_7_14_14_8(SH2_struct* sh)
{
  SH2addi(sh, 14, (14<<4)+8);
}
#define init_7_14_14_9
void FASTCALL call_7_14_14_9(SH2_struct* sh)
{
  SH2addi(sh, 14, (14<<4)+9);
}
#define init_7_14_14_10
void FASTCALL call_7_14_14_10(SH2_struct* sh)
{
  SH2addi(sh, 14, (14<<4)+10);
}
#define init_7_14_14_11
void FASTCALL call_7_14_14_11(SH2_struct* sh)
{
  SH2addi(sh, 14, (14<<4)+11);
}
#define init_7_14_14_12
void FASTCALL call_7_14_14_12(SH2_struct* sh)
{
  SH2addi(sh, 14, (14<<4)+12);
}
#define init_7_14_14_13
void FASTCALL call_7_14_14_13(SH2_struct* sh)
{
  SH2addi(sh, 14, (14<<4)+13);
}
#define init_7_14_14_14
void FASTCALL call_7_14_14_14(SH2_struct* sh)
{
  SH2addi(sh, 14, (14<<4)+14);
}
#define init_7_14_14_15
void FASTCALL call_7_14_14_15(SH2_struct* sh)
{
  SH2addi(sh, 14, (14<<4)+15);
}
#define init_7_14_15_0
void FASTCALL call_7_14_15_0(SH2_struct* sh)
{
  SH2addi(sh, 14, (15<<4)+0);
}
#define init_7_14_15_1
void FASTCALL call_7_14_15_1(SH2_struct* sh)
{
  SH2addi(sh, 14, (15<<4)+1);
}
#define init_7_14_15_2
void FASTCALL call_7_14_15_2(SH2_struct* sh)
{
  SH2addi(sh, 14, (15<<4)+2);
}
#define init_7_14_15_3
void FASTCALL call_7_14_15_3(SH2_struct* sh)
{
  SH2addi(sh, 14, (15<<4)+3);
}
#define init_7_14_15_4
void FASTCALL call_7_14_15_4(SH2_struct* sh)
{
  SH2addi(sh, 14, (15<<4)+4);
}
#define init_7_14_15_5
void FASTCALL call_7_14_15_5(SH2_struct* sh)
{
  SH2addi(sh, 14, (15<<4)+5);
}
#define init_7_14_15_6
void FASTCALL call_7_14_15_6(SH2_struct* sh)
{
  SH2addi(sh, 14, (15<<4)+6);
}
#define init_7_14_15_7
void FASTCALL call_7_14_15_7(SH2_struct* sh)
{
  SH2addi(sh, 14, (15<<4)+7);
}
#define init_7_14_15_8
void FASTCALL call_7_14_15_8(SH2_struct* sh)
{
  SH2addi(sh, 14, (15<<4)+8);
}
#define init_7_14_15_9
void FASTCALL call_7_14_15_9(SH2_struct* sh)
{
  SH2addi(sh, 14, (15<<4)+9);
}
#define init_7_14_15_10
void FASTCALL call_7_14_15_10(SH2_struct* sh)
{
  SH2addi(sh, 14, (15<<4)+10);
}
#define init_7_14_15_11
void FASTCALL call_7_14_15_11(SH2_struct* sh)
{
  SH2addi(sh, 14, (15<<4)+11);
}
#define init_7_14_15_12
void FASTCALL call_7_14_15_12(SH2_struct* sh)
{
  SH2addi(sh, 14, (15<<4)+12);
}
#define init_7_14_15_13
void FASTCALL call_7_14_15_13(SH2_struct* sh)
{
  SH2addi(sh, 14, (15<<4)+13);
}
#define init_7_14_15_14
void FASTCALL call_7_14_15_14(SH2_struct* sh)
{
  SH2addi(sh, 14, (15<<4)+14);
}
#define init_7_14_15_15
void FASTCALL call_7_14_15_15(SH2_struct* sh)
{
  SH2addi(sh, 14, (15<<4)+15);
}
#define init_7_15_0_0
void FASTCALL call_7_15_0_0(SH2_struct* sh)
{
  SH2addi(sh, 15, (0<<4)+0);
}
#define init_7_15_0_1
void FASTCALL call_7_15_0_1(SH2_struct* sh)
{
  SH2addi(sh, 15, (0<<4)+1);
}
#define init_7_15_0_2
void FASTCALL call_7_15_0_2(SH2_struct* sh)
{
  SH2addi(sh, 15, (0<<4)+2);
}
#define init_7_15_0_3
void FASTCALL call_7_15_0_3(SH2_struct* sh)
{
  SH2addi(sh, 15, (0<<4)+3);
}
#define init_7_15_0_4
void FASTCALL call_7_15_0_4(SH2_struct* sh)
{
  SH2addi(sh, 15, (0<<4)+4);
}
#define init_7_15_0_5
void FASTCALL call_7_15_0_5(SH2_struct* sh)
{
  SH2addi(sh, 15, (0<<4)+5);
}
#define init_7_15_0_6
void FASTCALL call_7_15_0_6(SH2_struct* sh)
{
  SH2addi(sh, 15, (0<<4)+6);
}
#define init_7_15_0_7
void FASTCALL call_7_15_0_7(SH2_struct* sh)
{
  SH2addi(sh, 15, (0<<4)+7);
}
#define init_7_15_0_8
void FASTCALL call_7_15_0_8(SH2_struct* sh)
{
  SH2addi(sh, 15, (0<<4)+8);
}
#define init_7_15_0_9
void FASTCALL call_7_15_0_9(SH2_struct* sh)
{
  SH2addi(sh, 15, (0<<4)+9);
}
#define init_7_15_0_10
void FASTCALL call_7_15_0_10(SH2_struct* sh)
{
  SH2addi(sh, 15, (0<<4)+10);
}
#define init_7_15_0_11
void FASTCALL call_7_15_0_11(SH2_struct* sh)
{
  SH2addi(sh, 15, (0<<4)+11);
}
#define init_7_15_0_12
void FASTCALL call_7_15_0_12(SH2_struct* sh)
{
  SH2addi(sh, 15, (0<<4)+12);
}
#define init_7_15_0_13
void FASTCALL call_7_15_0_13(SH2_struct* sh)
{
  SH2addi(sh, 15, (0<<4)+13);
}
#define init_7_15_0_14
void FASTCALL call_7_15_0_14(SH2_struct* sh)
{
  SH2addi(sh, 15, (0<<4)+14);
}
#define init_7_15_0_15
void FASTCALL call_7_15_0_15(SH2_struct* sh)
{
  SH2addi(sh, 15, (0<<4)+15);
}
#define init_7_15_1_0
void FASTCALL call_7_15_1_0(SH2_struct* sh)
{
  SH2addi(sh, 15, (1<<4)+0);
}
#define init_7_15_1_1
void FASTCALL call_7_15_1_1(SH2_struct* sh)
{
  SH2addi(sh, 15, (1<<4)+1);
}
#define init_7_15_1_2
void FASTCALL call_7_15_1_2(SH2_struct* sh)
{
  SH2addi(sh, 15, (1<<4)+2);
}
#define init_7_15_1_3
void FASTCALL call_7_15_1_3(SH2_struct* sh)
{
  SH2addi(sh, 15, (1<<4)+3);
}
#define init_7_15_1_4
void FASTCALL call_7_15_1_4(SH2_struct* sh)
{
  SH2addi(sh, 15, (1<<4)+4);
}
#define init_7_15_1_5
void FASTCALL call_7_15_1_5(SH2_struct* sh)
{
  SH2addi(sh, 15, (1<<4)+5);
}
#define init_7_15_1_6
void FASTCALL call_7_15_1_6(SH2_struct* sh)
{
  SH2addi(sh, 15, (1<<4)+6);
}
#define init_7_15_1_7
void FASTCALL call_7_15_1_7(SH2_struct* sh)
{
  SH2addi(sh, 15, (1<<4)+7);
}
#define init_7_15_1_8
void FASTCALL call_7_15_1_8(SH2_struct* sh)
{
  SH2addi(sh, 15, (1<<4)+8);
}
#define init_7_15_1_9
void FASTCALL call_7_15_1_9(SH2_struct* sh)
{
  SH2addi(sh, 15, (1<<4)+9);
}
#define init_7_15_1_10
void FASTCALL call_7_15_1_10(SH2_struct* sh)
{
  SH2addi(sh, 15, (1<<4)+10);
}
#define init_7_15_1_11
void FASTCALL call_7_15_1_11(SH2_struct* sh)
{
  SH2addi(sh, 15, (1<<4)+11);
}
#define init_7_15_1_12
void FASTCALL call_7_15_1_12(SH2_struct* sh)
{
  SH2addi(sh, 15, (1<<4)+12);
}
#define init_7_15_1_13
void FASTCALL call_7_15_1_13(SH2_struct* sh)
{
  SH2addi(sh, 15, (1<<4)+13);
}
#define init_7_15_1_14
void FASTCALL call_7_15_1_14(SH2_struct* sh)
{
  SH2addi(sh, 15, (1<<4)+14);
}
#define init_7_15_1_15
void FASTCALL call_7_15_1_15(SH2_struct* sh)
{
  SH2addi(sh, 15, (1<<4)+15);
}
#define init_7_15_2_0
void FASTCALL call_7_15_2_0(SH2_struct* sh)
{
  SH2addi(sh, 15, (2<<4)+0);
}
#define init_7_15_2_1
void FASTCALL call_7_15_2_1(SH2_struct* sh)
{
  SH2addi(sh, 15, (2<<4)+1);
}
#define init_7_15_2_2
void FASTCALL call_7_15_2_2(SH2_struct* sh)
{
  SH2addi(sh, 15, (2<<4)+2);
}
#define init_7_15_2_3
void FASTCALL call_7_15_2_3(SH2_struct* sh)
{
  SH2addi(sh, 15, (2<<4)+3);
}
#define init_7_15_2_4
void FASTCALL call_7_15_2_4(SH2_struct* sh)
{
  SH2addi(sh, 15, (2<<4)+4);
}
#define init_7_15_2_5
void FASTCALL call_7_15_2_5(SH2_struct* sh)
{
  SH2addi(sh, 15, (2<<4)+5);
}
#define init_7_15_2_6
void FASTCALL call_7_15_2_6(SH2_struct* sh)
{
  SH2addi(sh, 15, (2<<4)+6);
}
#define init_7_15_2_7
void FASTCALL call_7_15_2_7(SH2_struct* sh)
{
  SH2addi(sh, 15, (2<<4)+7);
}
#define init_7_15_2_8
void FASTCALL call_7_15_2_8(SH2_struct* sh)
{
  SH2addi(sh, 15, (2<<4)+8);
}
#define init_7_15_2_9
void FASTCALL call_7_15_2_9(SH2_struct* sh)
{
  SH2addi(sh, 15, (2<<4)+9);
}
#define init_7_15_2_10
void FASTCALL call_7_15_2_10(SH2_struct* sh)
{
  SH2addi(sh, 15, (2<<4)+10);
}
#define init_7_15_2_11
void FASTCALL call_7_15_2_11(SH2_struct* sh)
{
  SH2addi(sh, 15, (2<<4)+11);
}
#define init_7_15_2_12
void FASTCALL call_7_15_2_12(SH2_struct* sh)
{
  SH2addi(sh, 15, (2<<4)+12);
}
#define init_7_15_2_13
void FASTCALL call_7_15_2_13(SH2_struct* sh)
{
  SH2addi(sh, 15, (2<<4)+13);
}
#define init_7_15_2_14
void FASTCALL call_7_15_2_14(SH2_struct* sh)
{
  SH2addi(sh, 15, (2<<4)+14);
}
#define init_7_15_2_15
void FASTCALL call_7_15_2_15(SH2_struct* sh)
{
  SH2addi(sh, 15, (2<<4)+15);
}
#define init_7_15_3_0
void FASTCALL call_7_15_3_0(SH2_struct* sh)
{
  SH2addi(sh, 15, (3<<4)+0);
}
#define init_7_15_3_1
void FASTCALL call_7_15_3_1(SH2_struct* sh)
{
  SH2addi(sh, 15, (3<<4)+1);
}
#define init_7_15_3_2
void FASTCALL call_7_15_3_2(SH2_struct* sh)
{
  SH2addi(sh, 15, (3<<4)+2);
}
#define init_7_15_3_3
void FASTCALL call_7_15_3_3(SH2_struct* sh)
{
  SH2addi(sh, 15, (3<<4)+3);
}
#define init_7_15_3_4
void FASTCALL call_7_15_3_4(SH2_struct* sh)
{
  SH2addi(sh, 15, (3<<4)+4);
}
#define init_7_15_3_5
void FASTCALL call_7_15_3_5(SH2_struct* sh)
{
  SH2addi(sh, 15, (3<<4)+5);
}
#define init_7_15_3_6
void FASTCALL call_7_15_3_6(SH2_struct* sh)
{
  SH2addi(sh, 15, (3<<4)+6);
}
#define init_7_15_3_7
void FASTCALL call_7_15_3_7(SH2_struct* sh)
{
  SH2addi(sh, 15, (3<<4)+7);
}
#define init_7_15_3_8
void FASTCALL call_7_15_3_8(SH2_struct* sh)
{
  SH2addi(sh, 15, (3<<4)+8);
}
#define init_7_15_3_9
void FASTCALL call_7_15_3_9(SH2_struct* sh)
{
  SH2addi(sh, 15, (3<<4)+9);
}
#define init_7_15_3_10
void FASTCALL call_7_15_3_10(SH2_struct* sh)
{
  SH2addi(sh, 15, (3<<4)+10);
}
#define init_7_15_3_11
void FASTCALL call_7_15_3_11(SH2_struct* sh)
{
  SH2addi(sh, 15, (3<<4)+11);
}
#define init_7_15_3_12
void FASTCALL call_7_15_3_12(SH2_struct* sh)
{
  SH2addi(sh, 15, (3<<4)+12);
}
#define init_7_15_3_13
void FASTCALL call_7_15_3_13(SH2_struct* sh)
{
  SH2addi(sh, 15, (3<<4)+13);
}
#define init_7_15_3_14
void FASTCALL call_7_15_3_14(SH2_struct* sh)
{
  SH2addi(sh, 15, (3<<4)+14);
}
#define init_7_15_3_15
void FASTCALL call_7_15_3_15(SH2_struct* sh)
{
  SH2addi(sh, 15, (3<<4)+15);
}
#define init_7_15_4_0
void FASTCALL call_7_15_4_0(SH2_struct* sh)
{
  SH2addi(sh, 15, (4<<4)+0);
}
#define init_7_15_4_1
void FASTCALL call_7_15_4_1(SH2_struct* sh)
{
  SH2addi(sh, 15, (4<<4)+1);
}
#define init_7_15_4_2
void FASTCALL call_7_15_4_2(SH2_struct* sh)
{
  SH2addi(sh, 15, (4<<4)+2);
}
#define init_7_15_4_3
void FASTCALL call_7_15_4_3(SH2_struct* sh)
{
  SH2addi(sh, 15, (4<<4)+3);
}
#define init_7_15_4_4
void FASTCALL call_7_15_4_4(SH2_struct* sh)
{
  SH2addi(sh, 15, (4<<4)+4);
}
#define init_7_15_4_5
void FASTCALL call_7_15_4_5(SH2_struct* sh)
{
  SH2addi(sh, 15, (4<<4)+5);
}
#define init_7_15_4_6
void FASTCALL call_7_15_4_6(SH2_struct* sh)
{
  SH2addi(sh, 15, (4<<4)+6);
}
#define init_7_15_4_7
void FASTCALL call_7_15_4_7(SH2_struct* sh)
{
  SH2addi(sh, 15, (4<<4)+7);
}
#define init_7_15_4_8
void FASTCALL call_7_15_4_8(SH2_struct* sh)
{
  SH2addi(sh, 15, (4<<4)+8);
}
#define init_7_15_4_9
void FASTCALL call_7_15_4_9(SH2_struct* sh)
{
  SH2addi(sh, 15, (4<<4)+9);
}
#define init_7_15_4_10
void FASTCALL call_7_15_4_10(SH2_struct* sh)
{
  SH2addi(sh, 15, (4<<4)+10);
}
#define init_7_15_4_11
void FASTCALL call_7_15_4_11(SH2_struct* sh)
{
  SH2addi(sh, 15, (4<<4)+11);
}
#define init_7_15_4_12
void FASTCALL call_7_15_4_12(SH2_struct* sh)
{
  SH2addi(sh, 15, (4<<4)+12);
}
#define init_7_15_4_13
void FASTCALL call_7_15_4_13(SH2_struct* sh)
{
  SH2addi(sh, 15, (4<<4)+13);
}
#define init_7_15_4_14
void FASTCALL call_7_15_4_14(SH2_struct* sh)
{
  SH2addi(sh, 15, (4<<4)+14);
}
#define init_7_15_4_15
void FASTCALL call_7_15_4_15(SH2_struct* sh)
{
  SH2addi(sh, 15, (4<<4)+15);
}
#define init_7_15_5_0
void FASTCALL call_7_15_5_0(SH2_struct* sh)
{
  SH2addi(sh, 15, (5<<4)+0);
}
#define init_7_15_5_1
void FASTCALL call_7_15_5_1(SH2_struct* sh)
{
  SH2addi(sh, 15, (5<<4)+1);
}
#define init_7_15_5_2
void FASTCALL call_7_15_5_2(SH2_struct* sh)
{
  SH2addi(sh, 15, (5<<4)+2);
}
#define init_7_15_5_3
void FASTCALL call_7_15_5_3(SH2_struct* sh)
{
  SH2addi(sh, 15, (5<<4)+3);
}
#define init_7_15_5_4
void FASTCALL call_7_15_5_4(SH2_struct* sh)
{
  SH2addi(sh, 15, (5<<4)+4);
}
#define init_7_15_5_5
void FASTCALL call_7_15_5_5(SH2_struct* sh)
{
  SH2addi(sh, 15, (5<<4)+5);
}
#define init_7_15_5_6
void FASTCALL call_7_15_5_6(SH2_struct* sh)
{
  SH2addi(sh, 15, (5<<4)+6);
}
#define init_7_15_5_7
void FASTCALL call_7_15_5_7(SH2_struct* sh)
{
  SH2addi(sh, 15, (5<<4)+7);
}
#define init_7_15_5_8
void FASTCALL call_7_15_5_8(SH2_struct* sh)
{
  SH2addi(sh, 15, (5<<4)+8);
}
#define init_7_15_5_9
void FASTCALL call_7_15_5_9(SH2_struct* sh)
{
  SH2addi(sh, 15, (5<<4)+9);
}
#define init_7_15_5_10
void FASTCALL call_7_15_5_10(SH2_struct* sh)
{
  SH2addi(sh, 15, (5<<4)+10);
}
#define init_7_15_5_11
void FASTCALL call_7_15_5_11(SH2_struct* sh)
{
  SH2addi(sh, 15, (5<<4)+11);
}
#define init_7_15_5_12
void FASTCALL call_7_15_5_12(SH2_struct* sh)
{
  SH2addi(sh, 15, (5<<4)+12);
}
#define init_7_15_5_13
void FASTCALL call_7_15_5_13(SH2_struct* sh)
{
  SH2addi(sh, 15, (5<<4)+13);
}
#define init_7_15_5_14
void FASTCALL call_7_15_5_14(SH2_struct* sh)
{
  SH2addi(sh, 15, (5<<4)+14);
}
#define init_7_15_5_15
void FASTCALL call_7_15_5_15(SH2_struct* sh)
{
  SH2addi(sh, 15, (5<<4)+15);
}
#define init_7_15_6_0
void FASTCALL call_7_15_6_0(SH2_struct* sh)
{
  SH2addi(sh, 15, (6<<4)+0);
}
#define init_7_15_6_1
void FASTCALL call_7_15_6_1(SH2_struct* sh)
{
  SH2addi(sh, 15, (6<<4)+1);
}
#define init_7_15_6_2
void FASTCALL call_7_15_6_2(SH2_struct* sh)
{
  SH2addi(sh, 15, (6<<4)+2);
}
#define init_7_15_6_3
void FASTCALL call_7_15_6_3(SH2_struct* sh)
{
  SH2addi(sh, 15, (6<<4)+3);
}
#define init_7_15_6_4
void FASTCALL call_7_15_6_4(SH2_struct* sh)
{
  SH2addi(sh, 15, (6<<4)+4);
}
#define init_7_15_6_5
void FASTCALL call_7_15_6_5(SH2_struct* sh)
{
  SH2addi(sh, 15, (6<<4)+5);
}
#define init_7_15_6_6
void FASTCALL call_7_15_6_6(SH2_struct* sh)
{
  SH2addi(sh, 15, (6<<4)+6);
}
#define init_7_15_6_7
void FASTCALL call_7_15_6_7(SH2_struct* sh)
{
  SH2addi(sh, 15, (6<<4)+7);
}
#define init_7_15_6_8
void FASTCALL call_7_15_6_8(SH2_struct* sh)
{
  SH2addi(sh, 15, (6<<4)+8);
}
#define init_7_15_6_9
void FASTCALL call_7_15_6_9(SH2_struct* sh)
{
  SH2addi(sh, 15, (6<<4)+9);
}
#define init_7_15_6_10
void FASTCALL call_7_15_6_10(SH2_struct* sh)
{
  SH2addi(sh, 15, (6<<4)+10);
}
#define init_7_15_6_11
void FASTCALL call_7_15_6_11(SH2_struct* sh)
{
  SH2addi(sh, 15, (6<<4)+11);
}
#define init_7_15_6_12
void FASTCALL call_7_15_6_12(SH2_struct* sh)
{
  SH2addi(sh, 15, (6<<4)+12);
}
#define init_7_15_6_13
void FASTCALL call_7_15_6_13(SH2_struct* sh)
{
  SH2addi(sh, 15, (6<<4)+13);
}
#define init_7_15_6_14
void FASTCALL call_7_15_6_14(SH2_struct* sh)
{
  SH2addi(sh, 15, (6<<4)+14);
}
#define init_7_15_6_15
void FASTCALL call_7_15_6_15(SH2_struct* sh)
{
  SH2addi(sh, 15, (6<<4)+15);
}
#define init_7_15_7_0
void FASTCALL call_7_15_7_0(SH2_struct* sh)
{
  SH2addi(sh, 15, (7<<4)+0);
}
#define init_7_15_7_1
void FASTCALL call_7_15_7_1(SH2_struct* sh)
{
  SH2addi(sh, 15, (7<<4)+1);
}
#define init_7_15_7_2
void FASTCALL call_7_15_7_2(SH2_struct* sh)
{
  SH2addi(sh, 15, (7<<4)+2);
}
#define init_7_15_7_3
void FASTCALL call_7_15_7_3(SH2_struct* sh)
{
  SH2addi(sh, 15, (7<<4)+3);
}
#define init_7_15_7_4
void FASTCALL call_7_15_7_4(SH2_struct* sh)
{
  SH2addi(sh, 15, (7<<4)+4);
}
#define init_7_15_7_5
void FASTCALL call_7_15_7_5(SH2_struct* sh)
{
  SH2addi(sh, 15, (7<<4)+5);
}
#define init_7_15_7_6
void FASTCALL call_7_15_7_6(SH2_struct* sh)
{
  SH2addi(sh, 15, (7<<4)+6);
}
#define init_7_15_7_7
void FASTCALL call_7_15_7_7(SH2_struct* sh)
{
  SH2addi(sh, 15, (7<<4)+7);
}
#define init_7_15_7_8
void FASTCALL call_7_15_7_8(SH2_struct* sh)
{
  SH2addi(sh, 15, (7<<4)+8);
}
#define init_7_15_7_9
void FASTCALL call_7_15_7_9(SH2_struct* sh)
{
  SH2addi(sh, 15, (7<<4)+9);
}
#define init_7_15_7_10
void FASTCALL call_7_15_7_10(SH2_struct* sh)
{
  SH2addi(sh, 15, (7<<4)+10);
}
#define init_7_15_7_11
void FASTCALL call_7_15_7_11(SH2_struct* sh)
{
  SH2addi(sh, 15, (7<<4)+11);
}
#define init_7_15_7_12
void FASTCALL call_7_15_7_12(SH2_struct* sh)
{
  SH2addi(sh, 15, (7<<4)+12);
}
#define init_7_15_7_13
void FASTCALL call_7_15_7_13(SH2_struct* sh)
{
  SH2addi(sh, 15, (7<<4)+13);
}
#define init_7_15_7_14
void FASTCALL call_7_15_7_14(SH2_struct* sh)
{
  SH2addi(sh, 15, (7<<4)+14);
}
#define init_7_15_7_15
void FASTCALL call_7_15_7_15(SH2_struct* sh)
{
  SH2addi(sh, 15, (7<<4)+15);
}
#define init_7_15_8_0
void FASTCALL call_7_15_8_0(SH2_struct* sh)
{
  SH2addi(sh, 15, (8<<4)+0);
}
#define init_7_15_8_1
void FASTCALL call_7_15_8_1(SH2_struct* sh)
{
  SH2addi(sh, 15, (8<<4)+1);
}
#define init_7_15_8_2
void FASTCALL call_7_15_8_2(SH2_struct* sh)
{
  SH2addi(sh, 15, (8<<4)+2);
}
#define init_7_15_8_3
void FASTCALL call_7_15_8_3(SH2_struct* sh)
{
  SH2addi(sh, 15, (8<<4)+3);
}
#define init_7_15_8_4
void FASTCALL call_7_15_8_4(SH2_struct* sh)
{
  SH2addi(sh, 15, (8<<4)+4);
}
#define init_7_15_8_5
void FASTCALL call_7_15_8_5(SH2_struct* sh)
{
  SH2addi(sh, 15, (8<<4)+5);
}
#define init_7_15_8_6
void FASTCALL call_7_15_8_6(SH2_struct* sh)
{
  SH2addi(sh, 15, (8<<4)+6);
}
#define init_7_15_8_7
void FASTCALL call_7_15_8_7(SH2_struct* sh)
{
  SH2addi(sh, 15, (8<<4)+7);
}
#define init_7_15_8_8
void FASTCALL call_7_15_8_8(SH2_struct* sh)
{
  SH2addi(sh, 15, (8<<4)+8);
}
#define init_7_15_8_9
void FASTCALL call_7_15_8_9(SH2_struct* sh)
{
  SH2addi(sh, 15, (8<<4)+9);
}
#define init_7_15_8_10
void FASTCALL call_7_15_8_10(SH2_struct* sh)
{
  SH2addi(sh, 15, (8<<4)+10);
}
#define init_7_15_8_11
void FASTCALL call_7_15_8_11(SH2_struct* sh)
{
  SH2addi(sh, 15, (8<<4)+11);
}
#define init_7_15_8_12
void FASTCALL call_7_15_8_12(SH2_struct* sh)
{
  SH2addi(sh, 15, (8<<4)+12);
}
#define init_7_15_8_13
void FASTCALL call_7_15_8_13(SH2_struct* sh)
{
  SH2addi(sh, 15, (8<<4)+13);
}
#define init_7_15_8_14
void FASTCALL call_7_15_8_14(SH2_struct* sh)
{
  SH2addi(sh, 15, (8<<4)+14);
}
#define init_7_15_8_15
void FASTCALL call_7_15_8_15(SH2_struct* sh)
{
  SH2addi(sh, 15, (8<<4)+15);
}
#define init_7_15_9_0
void FASTCALL call_7_15_9_0(SH2_struct* sh)
{
  SH2addi(sh, 15, (9<<4)+0);
}
#define init_7_15_9_1
void FASTCALL call_7_15_9_1(SH2_struct* sh)
{
  SH2addi(sh, 15, (9<<4)+1);
}
#define init_7_15_9_2
void FASTCALL call_7_15_9_2(SH2_struct* sh)
{
  SH2addi(sh, 15, (9<<4)+2);
}
#define init_7_15_9_3
void FASTCALL call_7_15_9_3(SH2_struct* sh)
{
  SH2addi(sh, 15, (9<<4)+3);
}
#define init_7_15_9_4
void FASTCALL call_7_15_9_4(SH2_struct* sh)
{
  SH2addi(sh, 15, (9<<4)+4);
}
#define init_7_15_9_5
void FASTCALL call_7_15_9_5(SH2_struct* sh)
{
  SH2addi(sh, 15, (9<<4)+5);
}
#define init_7_15_9_6
void FASTCALL call_7_15_9_6(SH2_struct* sh)
{
  SH2addi(sh, 15, (9<<4)+6);
}
#define init_7_15_9_7
void FASTCALL call_7_15_9_7(SH2_struct* sh)
{
  SH2addi(sh, 15, (9<<4)+7);
}
#define init_7_15_9_8
void FASTCALL call_7_15_9_8(SH2_struct* sh)
{
  SH2addi(sh, 15, (9<<4)+8);
}
#define init_7_15_9_9
void FASTCALL call_7_15_9_9(SH2_struct* sh)
{
  SH2addi(sh, 15, (9<<4)+9);
}
#define init_7_15_9_10
void FASTCALL call_7_15_9_10(SH2_struct* sh)
{
  SH2addi(sh, 15, (9<<4)+10);
}
#define init_7_15_9_11
void FASTCALL call_7_15_9_11(SH2_struct* sh)
{
  SH2addi(sh, 15, (9<<4)+11);
}
#define init_7_15_9_12
void FASTCALL call_7_15_9_12(SH2_struct* sh)
{
  SH2addi(sh, 15, (9<<4)+12);
}
#define init_7_15_9_13
void FASTCALL call_7_15_9_13(SH2_struct* sh)
{
  SH2addi(sh, 15, (9<<4)+13);
}
#define init_7_15_9_14
void FASTCALL call_7_15_9_14(SH2_struct* sh)
{
  SH2addi(sh, 15, (9<<4)+14);
}
#define init_7_15_9_15
void FASTCALL call_7_15_9_15(SH2_struct* sh)
{
  SH2addi(sh, 15, (9<<4)+15);
}
#define init_7_15_10_0
void FASTCALL call_7_15_10_0(SH2_struct* sh)
{
  SH2addi(sh, 15, (10<<4)+0);
}
#define init_7_15_10_1
void FASTCALL call_7_15_10_1(SH2_struct* sh)
{
  SH2addi(sh, 15, (10<<4)+1);
}
#define init_7_15_10_2
void FASTCALL call_7_15_10_2(SH2_struct* sh)
{
  SH2addi(sh, 15, (10<<4)+2);
}
#define init_7_15_10_3
void FASTCALL call_7_15_10_3(SH2_struct* sh)
{
  SH2addi(sh, 15, (10<<4)+3);
}
#define init_7_15_10_4
void FASTCALL call_7_15_10_4(SH2_struct* sh)
{
  SH2addi(sh, 15, (10<<4)+4);
}
#define init_7_15_10_5
void FASTCALL call_7_15_10_5(SH2_struct* sh)
{
  SH2addi(sh, 15, (10<<4)+5);
}
#define init_7_15_10_6
void FASTCALL call_7_15_10_6(SH2_struct* sh)
{
  SH2addi(sh, 15, (10<<4)+6);
}
#define init_7_15_10_7
void FASTCALL call_7_15_10_7(SH2_struct* sh)
{
  SH2addi(sh, 15, (10<<4)+7);
}
#define init_7_15_10_8
void FASTCALL call_7_15_10_8(SH2_struct* sh)
{
  SH2addi(sh, 15, (10<<4)+8);
}
#define init_7_15_10_9
void FASTCALL call_7_15_10_9(SH2_struct* sh)
{
  SH2addi(sh, 15, (10<<4)+9);
}
#define init_7_15_10_10
void FASTCALL call_7_15_10_10(SH2_struct* sh)
{
  SH2addi(sh, 15, (10<<4)+10);
}
#define init_7_15_10_11
void FASTCALL call_7_15_10_11(SH2_struct* sh)
{
  SH2addi(sh, 15, (10<<4)+11);
}
#define init_7_15_10_12
void FASTCALL call_7_15_10_12(SH2_struct* sh)
{
  SH2addi(sh, 15, (10<<4)+12);
}
#define init_7_15_10_13
void FASTCALL call_7_15_10_13(SH2_struct* sh)
{
  SH2addi(sh, 15, (10<<4)+13);
}
#define init_7_15_10_14
void FASTCALL call_7_15_10_14(SH2_struct* sh)
{
  SH2addi(sh, 15, (10<<4)+14);
}
#define init_7_15_10_15
void FASTCALL call_7_15_10_15(SH2_struct* sh)
{
  SH2addi(sh, 15, (10<<4)+15);
}
#define init_7_15_11_0
void FASTCALL call_7_15_11_0(SH2_struct* sh)
{
  SH2addi(sh, 15, (11<<4)+0);
}
#define init_7_15_11_1
void FASTCALL call_7_15_11_1(SH2_struct* sh)
{
  SH2addi(sh, 15, (11<<4)+1);
}
#define init_7_15_11_2
void FASTCALL call_7_15_11_2(SH2_struct* sh)
{
  SH2addi(sh, 15, (11<<4)+2);
}
#define init_7_15_11_3
void FASTCALL call_7_15_11_3(SH2_struct* sh)
{
  SH2addi(sh, 15, (11<<4)+3);
}
#define init_7_15_11_4
void FASTCALL call_7_15_11_4(SH2_struct* sh)
{
  SH2addi(sh, 15, (11<<4)+4);
}
#define init_7_15_11_5
void FASTCALL call_7_15_11_5(SH2_struct* sh)
{
  SH2addi(sh, 15, (11<<4)+5);
}
#define init_7_15_11_6
void FASTCALL call_7_15_11_6(SH2_struct* sh)
{
  SH2addi(sh, 15, (11<<4)+6);
}
#define init_7_15_11_7
void FASTCALL call_7_15_11_7(SH2_struct* sh)
{
  SH2addi(sh, 15, (11<<4)+7);
}
#define init_7_15_11_8
void FASTCALL call_7_15_11_8(SH2_struct* sh)
{
  SH2addi(sh, 15, (11<<4)+8);
}
#define init_7_15_11_9
void FASTCALL call_7_15_11_9(SH2_struct* sh)
{
  SH2addi(sh, 15, (11<<4)+9);
}
#define init_7_15_11_10
void FASTCALL call_7_15_11_10(SH2_struct* sh)
{
  SH2addi(sh, 15, (11<<4)+10);
}
#define init_7_15_11_11
void FASTCALL call_7_15_11_11(SH2_struct* sh)
{
  SH2addi(sh, 15, (11<<4)+11);
}
#define init_7_15_11_12
void FASTCALL call_7_15_11_12(SH2_struct* sh)
{
  SH2addi(sh, 15, (11<<4)+12);
}
#define init_7_15_11_13
void FASTCALL call_7_15_11_13(SH2_struct* sh)
{
  SH2addi(sh, 15, (11<<4)+13);
}
#define init_7_15_11_14
void FASTCALL call_7_15_11_14(SH2_struct* sh)
{
  SH2addi(sh, 15, (11<<4)+14);
}
#define init_7_15_11_15
void FASTCALL call_7_15_11_15(SH2_struct* sh)
{
  SH2addi(sh, 15, (11<<4)+15);
}
#define init_7_15_12_0
void FASTCALL call_7_15_12_0(SH2_struct* sh)
{
  SH2addi(sh, 15, (12<<4)+0);
}
#define init_7_15_12_1
void FASTCALL call_7_15_12_1(SH2_struct* sh)
{
  SH2addi(sh, 15, (12<<4)+1);
}
#define init_7_15_12_2
void FASTCALL call_7_15_12_2(SH2_struct* sh)
{
  SH2addi(sh, 15, (12<<4)+2);
}
#define init_7_15_12_3
void FASTCALL call_7_15_12_3(SH2_struct* sh)
{
  SH2addi(sh, 15, (12<<4)+3);
}
#define init_7_15_12_4
void FASTCALL call_7_15_12_4(SH2_struct* sh)
{
  SH2addi(sh, 15, (12<<4)+4);
}
#define init_7_15_12_5
void FASTCALL call_7_15_12_5(SH2_struct* sh)
{
  SH2addi(sh, 15, (12<<4)+5);
}
#define init_7_15_12_6
void FASTCALL call_7_15_12_6(SH2_struct* sh)
{
  SH2addi(sh, 15, (12<<4)+6);
}
#define init_7_15_12_7
void FASTCALL call_7_15_12_7(SH2_struct* sh)
{
  SH2addi(sh, 15, (12<<4)+7);
}
#define init_7_15_12_8
void FASTCALL call_7_15_12_8(SH2_struct* sh)
{
  SH2addi(sh, 15, (12<<4)+8);
}
#define init_7_15_12_9
void FASTCALL call_7_15_12_9(SH2_struct* sh)
{
  SH2addi(sh, 15, (12<<4)+9);
}
#define init_7_15_12_10
void FASTCALL call_7_15_12_10(SH2_struct* sh)
{
  SH2addi(sh, 15, (12<<4)+10);
}
#define init_7_15_12_11
void FASTCALL call_7_15_12_11(SH2_struct* sh)
{
  SH2addi(sh, 15, (12<<4)+11);
}
#define init_7_15_12_12
void FASTCALL call_7_15_12_12(SH2_struct* sh)
{
  SH2addi(sh, 15, (12<<4)+12);
}
#define init_7_15_12_13
void FASTCALL call_7_15_12_13(SH2_struct* sh)
{
  SH2addi(sh, 15, (12<<4)+13);
}
#define init_7_15_12_14
void FASTCALL call_7_15_12_14(SH2_struct* sh)
{
  SH2addi(sh, 15, (12<<4)+14);
}
#define init_7_15_12_15
void FASTCALL call_7_15_12_15(SH2_struct* sh)
{
  SH2addi(sh, 15, (12<<4)+15);
}
#define init_7_15_13_0
void FASTCALL call_7_15_13_0(SH2_struct* sh)
{
  SH2addi(sh, 15, (13<<4)+0);
}
#define init_7_15_13_1
void FASTCALL call_7_15_13_1(SH2_struct* sh)
{
  SH2addi(sh, 15, (13<<4)+1);
}
#define init_7_15_13_2
void FASTCALL call_7_15_13_2(SH2_struct* sh)
{
  SH2addi(sh, 15, (13<<4)+2);
}
#define init_7_15_13_3
void FASTCALL call_7_15_13_3(SH2_struct* sh)
{
  SH2addi(sh, 15, (13<<4)+3);
}
#define init_7_15_13_4
void FASTCALL call_7_15_13_4(SH2_struct* sh)
{
  SH2addi(sh, 15, (13<<4)+4);
}
#define init_7_15_13_5
void FASTCALL call_7_15_13_5(SH2_struct* sh)
{
  SH2addi(sh, 15, (13<<4)+5);
}
#define init_7_15_13_6
void FASTCALL call_7_15_13_6(SH2_struct* sh)
{
  SH2addi(sh, 15, (13<<4)+6);
}
#define init_7_15_13_7
void FASTCALL call_7_15_13_7(SH2_struct* sh)
{
  SH2addi(sh, 15, (13<<4)+7);
}
#define init_7_15_13_8
void FASTCALL call_7_15_13_8(SH2_struct* sh)
{
  SH2addi(sh, 15, (13<<4)+8);
}
#define init_7_15_13_9
void FASTCALL call_7_15_13_9(SH2_struct* sh)
{
  SH2addi(sh, 15, (13<<4)+9);
}
#define init_7_15_13_10
void FASTCALL call_7_15_13_10(SH2_struct* sh)
{
  SH2addi(sh, 15, (13<<4)+10);
}
#define init_7_15_13_11
void FASTCALL call_7_15_13_11(SH2_struct* sh)
{
  SH2addi(sh, 15, (13<<4)+11);
}
#define init_7_15_13_12
void FASTCALL call_7_15_13_12(SH2_struct* sh)
{
  SH2addi(sh, 15, (13<<4)+12);
}
#define init_7_15_13_13
void FASTCALL call_7_15_13_13(SH2_struct* sh)
{
  SH2addi(sh, 15, (13<<4)+13);
}
#define init_7_15_13_14
void FASTCALL call_7_15_13_14(SH2_struct* sh)
{
  SH2addi(sh, 15, (13<<4)+14);
}
#define init_7_15_13_15
void FASTCALL call_7_15_13_15(SH2_struct* sh)
{
  SH2addi(sh, 15, (13<<4)+15);
}
#define init_7_15_14_0
void FASTCALL call_7_15_14_0(SH2_struct* sh)
{
  SH2addi(sh, 15, (14<<4)+0);
}
#define init_7_15_14_1
void FASTCALL call_7_15_14_1(SH2_struct* sh)
{
  SH2addi(sh, 15, (14<<4)+1);
}
#define init_7_15_14_2
void FASTCALL call_7_15_14_2(SH2_struct* sh)
{
  SH2addi(sh, 15, (14<<4)+2);
}
#define init_7_15_14_3
void FASTCALL call_7_15_14_3(SH2_struct* sh)
{
  SH2addi(sh, 15, (14<<4)+3);
}
#define init_7_15_14_4
void FASTCALL call_7_15_14_4(SH2_struct* sh)
{
  SH2addi(sh, 15, (14<<4)+4);
}
#define init_7_15_14_5
void FASTCALL call_7_15_14_5(SH2_struct* sh)
{
  SH2addi(sh, 15, (14<<4)+5);
}
#define init_7_15_14_6
void FASTCALL call_7_15_14_6(SH2_struct* sh)
{
  SH2addi(sh, 15, (14<<4)+6);
}
#define init_7_15_14_7
void FASTCALL call_7_15_14_7(SH2_struct* sh)
{
  SH2addi(sh, 15, (14<<4)+7);
}
#define init_7_15_14_8
void FASTCALL call_7_15_14_8(SH2_struct* sh)
{
  SH2addi(sh, 15, (14<<4)+8);
}
#define init_7_15_14_9
void FASTCALL call_7_15_14_9(SH2_struct* sh)
{
  SH2addi(sh, 15, (14<<4)+9);
}
#define init_7_15_14_10
void FASTCALL call_7_15_14_10(SH2_struct* sh)
{
  SH2addi(sh, 15, (14<<4)+10);
}
#define init_7_15_14_11
void FASTCALL call_7_15_14_11(SH2_struct* sh)
{
  SH2addi(sh, 15, (14<<4)+11);
}
#define init_7_15_14_12
void FASTCALL call_7_15_14_12(SH2_struct* sh)
{
  SH2addi(sh, 15, (14<<4)+12);
}
#define init_7_15_14_13
void FASTCALL call_7_15_14_13(SH2_struct* sh)
{
  SH2addi(sh, 15, (14<<4)+13);
}
#define init_7_15_14_14
void FASTCALL call_7_15_14_14(SH2_struct* sh)
{
  SH2addi(sh, 15, (14<<4)+14);
}
#define init_7_15_14_15
void FASTCALL call_7_15_14_15(SH2_struct* sh)
{
  SH2addi(sh, 15, (14<<4)+15);
}
#define init_7_15_15_0
void FASTCALL call_7_15_15_0(SH2_struct* sh)
{
  SH2addi(sh, 15, (15<<4)+0);
}
#define init_7_15_15_1
void FASTCALL call_7_15_15_1(SH2_struct* sh)
{
  SH2addi(sh, 15, (15<<4)+1);
}
#define init_7_15_15_2
void FASTCALL call_7_15_15_2(SH2_struct* sh)
{
  SH2addi(sh, 15, (15<<4)+2);
}
#define init_7_15_15_3
void FASTCALL call_7_15_15_3(SH2_struct* sh)
{
  SH2addi(sh, 15, (15<<4)+3);
}
#define init_7_15_15_4
void FASTCALL call_7_15_15_4(SH2_struct* sh)
{
  SH2addi(sh, 15, (15<<4)+4);
}
#define init_7_15_15_5
void FASTCALL call_7_15_15_5(SH2_struct* sh)
{
  SH2addi(sh, 15, (15<<4)+5);
}
#define init_7_15_15_6
void FASTCALL call_7_15_15_6(SH2_struct* sh)
{
  SH2addi(sh, 15, (15<<4)+6);
}
#define init_7_15_15_7
void FASTCALL call_7_15_15_7(SH2_struct* sh)
{
  SH2addi(sh, 15, (15<<4)+7);
}
#define init_7_15_15_8
void FASTCALL call_7_15_15_8(SH2_struct* sh)
{
  SH2addi(sh, 15, (15<<4)+8);
}
#define init_7_15_15_9
void FASTCALL call_7_15_15_9(SH2_struct* sh)
{
  SH2addi(sh, 15, (15<<4)+9);
}
#define init_7_15_15_10
void FASTCALL call_7_15_15_10(SH2_struct* sh)
{
  SH2addi(sh, 15, (15<<4)+10);
}
#define init_7_15_15_11
void FASTCALL call_7_15_15_11(SH2_struct* sh)
{
  SH2addi(sh, 15, (15<<4)+11);
}
#define init_7_15_15_12
void FASTCALL call_7_15_15_12(SH2_struct* sh)
{
  SH2addi(sh, 15, (15<<4)+12);
}
#define init_7_15_15_13
void FASTCALL call_7_15_15_13(SH2_struct* sh)
{
  SH2addi(sh, 15, (15<<4)+13);
}
#define init_7_15_15_14
void FASTCALL call_7_15_15_14(SH2_struct* sh)
{
  SH2addi(sh, 15, (15<<4)+14);
}
#define init_7_15_15_15
void FASTCALL call_7_15_15_15(SH2_struct* sh)
{
  SH2addi(sh, 15, (15<<4)+15);
}
#define init_3_0_0_12
void FASTCALL call_3_0_0_12(SH2_struct* sh)
{
  SH2add(sh,0,0);
}
#define init_3_0_1_12
void FASTCALL call_3_0_1_12(SH2_struct* sh)
{
  SH2add(sh,0,1);
}
#define init_3_0_2_12
void FASTCALL call_3_0_2_12(SH2_struct* sh)
{
  SH2add(sh,0,2);
}
#define init_3_0_3_12
void FASTCALL call_3_0_3_12(SH2_struct* sh)
{
  SH2add(sh,0,3);
}
#define init_3_0_4_12
void FASTCALL call_3_0_4_12(SH2_struct* sh)
{
  SH2add(sh,0,4);
}
#define init_3_0_5_12
void FASTCALL call_3_0_5_12(SH2_struct* sh)
{
  SH2add(sh,0,5);
}
#define init_3_0_6_12
void FASTCALL call_3_0_6_12(SH2_struct* sh)
{
  SH2add(sh,0,6);
}
#define init_3_0_7_12
void FASTCALL call_3_0_7_12(SH2_struct* sh)
{
  SH2add(sh,0,7);
}
#define init_3_0_8_12
void FASTCALL call_3_0_8_12(SH2_struct* sh)
{
  SH2add(sh,0,8);
}
#define init_3_0_9_12
void FASTCALL call_3_0_9_12(SH2_struct* sh)
{
  SH2add(sh,0,9);
}
#define init_3_0_10_12
void FASTCALL call_3_0_10_12(SH2_struct* sh)
{
  SH2add(sh,0,10);
}
#define init_3_0_11_12
void FASTCALL call_3_0_11_12(SH2_struct* sh)
{
  SH2add(sh,0,11);
}
#define init_3_0_12_12
void FASTCALL call_3_0_12_12(SH2_struct* sh)
{
  SH2add(sh,0,12);
}
#define init_3_0_13_12
void FASTCALL call_3_0_13_12(SH2_struct* sh)
{
  SH2add(sh,0,13);
}
#define init_3_0_14_12
void FASTCALL call_3_0_14_12(SH2_struct* sh)
{
  SH2add(sh,0,14);
}
#define init_3_0_15_12
void FASTCALL call_3_0_15_12(SH2_struct* sh)
{
  SH2add(sh,0,15);
}
#define init_3_1_0_12
void FASTCALL call_3_1_0_12(SH2_struct* sh)
{
  SH2add(sh,1,0);
}
#define init_3_1_1_12
void FASTCALL call_3_1_1_12(SH2_struct* sh)
{
  SH2add(sh,1,1);
}
#define init_3_1_2_12
void FASTCALL call_3_1_2_12(SH2_struct* sh)
{
  SH2add(sh,1,2);
}
#define init_3_1_3_12
void FASTCALL call_3_1_3_12(SH2_struct* sh)
{
  SH2add(sh,1,3);
}
#define init_3_1_4_12
void FASTCALL call_3_1_4_12(SH2_struct* sh)
{
  SH2add(sh,1,4);
}
#define init_3_1_5_12
void FASTCALL call_3_1_5_12(SH2_struct* sh)
{
  SH2add(sh,1,5);
}
#define init_3_1_6_12
void FASTCALL call_3_1_6_12(SH2_struct* sh)
{
  SH2add(sh,1,6);
}
#define init_3_1_7_12
void FASTCALL call_3_1_7_12(SH2_struct* sh)
{
  SH2add(sh,1,7);
}
#define init_3_1_8_12
void FASTCALL call_3_1_8_12(SH2_struct* sh)
{
  SH2add(sh,1,8);
}
#define init_3_1_9_12
void FASTCALL call_3_1_9_12(SH2_struct* sh)
{
  SH2add(sh,1,9);
}
#define init_3_1_10_12
void FASTCALL call_3_1_10_12(SH2_struct* sh)
{
  SH2add(sh,1,10);
}
#define init_3_1_11_12
void FASTCALL call_3_1_11_12(SH2_struct* sh)
{
  SH2add(sh,1,11);
}
#define init_3_1_12_12
void FASTCALL call_3_1_12_12(SH2_struct* sh)
{
  SH2add(sh,1,12);
}
#define init_3_1_13_12
void FASTCALL call_3_1_13_12(SH2_struct* sh)
{
  SH2add(sh,1,13);
}
#define init_3_1_14_12
void FASTCALL call_3_1_14_12(SH2_struct* sh)
{
  SH2add(sh,1,14);
}
#define init_3_1_15_12
void FASTCALL call_3_1_15_12(SH2_struct* sh)
{
  SH2add(sh,1,15);
}
#define init_3_2_0_12
void FASTCALL call_3_2_0_12(SH2_struct* sh)
{
  SH2add(sh,2,0);
}
#define init_3_2_1_12
void FASTCALL call_3_2_1_12(SH2_struct* sh)
{
  SH2add(sh,2,1);
}
#define init_3_2_2_12
void FASTCALL call_3_2_2_12(SH2_struct* sh)
{
  SH2add(sh,2,2);
}
#define init_3_2_3_12
void FASTCALL call_3_2_3_12(SH2_struct* sh)
{
  SH2add(sh,2,3);
}
#define init_3_2_4_12
void FASTCALL call_3_2_4_12(SH2_struct* sh)
{
  SH2add(sh,2,4);
}
#define init_3_2_5_12
void FASTCALL call_3_2_5_12(SH2_struct* sh)
{
  SH2add(sh,2,5);
}
#define init_3_2_6_12
void FASTCALL call_3_2_6_12(SH2_struct* sh)
{
  SH2add(sh,2,6);
}
#define init_3_2_7_12
void FASTCALL call_3_2_7_12(SH2_struct* sh)
{
  SH2add(sh,2,7);
}
#define init_3_2_8_12
void FASTCALL call_3_2_8_12(SH2_struct* sh)
{
  SH2add(sh,2,8);
}
#define init_3_2_9_12
void FASTCALL call_3_2_9_12(SH2_struct* sh)
{
  SH2add(sh,2,9);
}
#define init_3_2_10_12
void FASTCALL call_3_2_10_12(SH2_struct* sh)
{
  SH2add(sh,2,10);
}
#define init_3_2_11_12
void FASTCALL call_3_2_11_12(SH2_struct* sh)
{
  SH2add(sh,2,11);
}
#define init_3_2_12_12
void FASTCALL call_3_2_12_12(SH2_struct* sh)
{
  SH2add(sh,2,12);
}
#define init_3_2_13_12
void FASTCALL call_3_2_13_12(SH2_struct* sh)
{
  SH2add(sh,2,13);
}
#define init_3_2_14_12
void FASTCALL call_3_2_14_12(SH2_struct* sh)
{
  SH2add(sh,2,14);
}
#define init_3_2_15_12
void FASTCALL call_3_2_15_12(SH2_struct* sh)
{
  SH2add(sh,2,15);
}
#define init_3_3_0_12
void FASTCALL call_3_3_0_12(SH2_struct* sh)
{
  SH2add(sh,3,0);
}
#define init_3_3_1_12
void FASTCALL call_3_3_1_12(SH2_struct* sh)
{
  SH2add(sh,3,1);
}
#define init_3_3_2_12
void FASTCALL call_3_3_2_12(SH2_struct* sh)
{
  SH2add(sh,3,2);
}
#define init_3_3_3_12
void FASTCALL call_3_3_3_12(SH2_struct* sh)
{
  SH2add(sh,3,3);
}
#define init_3_3_4_12
void FASTCALL call_3_3_4_12(SH2_struct* sh)
{
  SH2add(sh,3,4);
}
#define init_3_3_5_12
void FASTCALL call_3_3_5_12(SH2_struct* sh)
{
  SH2add(sh,3,5);
}
#define init_3_3_6_12
void FASTCALL call_3_3_6_12(SH2_struct* sh)
{
  SH2add(sh,3,6);
}
#define init_3_3_7_12
void FASTCALL call_3_3_7_12(SH2_struct* sh)
{
  SH2add(sh,3,7);
}
#define init_3_3_8_12
void FASTCALL call_3_3_8_12(SH2_struct* sh)
{
  SH2add(sh,3,8);
}
#define init_3_3_9_12
void FASTCALL call_3_3_9_12(SH2_struct* sh)
{
  SH2add(sh,3,9);
}
#define init_3_3_10_12
void FASTCALL call_3_3_10_12(SH2_struct* sh)
{
  SH2add(sh,3,10);
}
#define init_3_3_11_12
void FASTCALL call_3_3_11_12(SH2_struct* sh)
{
  SH2add(sh,3,11);
}
#define init_3_3_12_12
void FASTCALL call_3_3_12_12(SH2_struct* sh)
{
  SH2add(sh,3,12);
}
#define init_3_3_13_12
void FASTCALL call_3_3_13_12(SH2_struct* sh)
{
  SH2add(sh,3,13);
}
#define init_3_3_14_12
void FASTCALL call_3_3_14_12(SH2_struct* sh)
{
  SH2add(sh,3,14);
}
#define init_3_3_15_12
void FASTCALL call_3_3_15_12(SH2_struct* sh)
{
  SH2add(sh,3,15);
}
#define init_3_4_0_12
void FASTCALL call_3_4_0_12(SH2_struct* sh)
{
  SH2add(sh,4,0);
}
#define init_3_4_1_12
void FASTCALL call_3_4_1_12(SH2_struct* sh)
{
  SH2add(sh,4,1);
}
#define init_3_4_2_12
void FASTCALL call_3_4_2_12(SH2_struct* sh)
{
  SH2add(sh,4,2);
}
#define init_3_4_3_12
void FASTCALL call_3_4_3_12(SH2_struct* sh)
{
  SH2add(sh,4,3);
}
#define init_3_4_4_12
void FASTCALL call_3_4_4_12(SH2_struct* sh)
{
  SH2add(sh,4,4);
}
#define init_3_4_5_12
void FASTCALL call_3_4_5_12(SH2_struct* sh)
{
  SH2add(sh,4,5);
}
#define init_3_4_6_12
void FASTCALL call_3_4_6_12(SH2_struct* sh)
{
  SH2add(sh,4,6);
}
#define init_3_4_7_12
void FASTCALL call_3_4_7_12(SH2_struct* sh)
{
  SH2add(sh,4,7);
}
#define init_3_4_8_12
void FASTCALL call_3_4_8_12(SH2_struct* sh)
{
  SH2add(sh,4,8);
}
#define init_3_4_9_12
void FASTCALL call_3_4_9_12(SH2_struct* sh)
{
  SH2add(sh,4,9);
}
#define init_3_4_10_12
void FASTCALL call_3_4_10_12(SH2_struct* sh)
{
  SH2add(sh,4,10);
}
#define init_3_4_11_12
void FASTCALL call_3_4_11_12(SH2_struct* sh)
{
  SH2add(sh,4,11);
}
#define init_3_4_12_12
void FASTCALL call_3_4_12_12(SH2_struct* sh)
{
  SH2add(sh,4,12);
}
#define init_3_4_13_12
void FASTCALL call_3_4_13_12(SH2_struct* sh)
{
  SH2add(sh,4,13);
}
#define init_3_4_14_12
void FASTCALL call_3_4_14_12(SH2_struct* sh)
{
  SH2add(sh,4,14);
}
#define init_3_4_15_12
void FASTCALL call_3_4_15_12(SH2_struct* sh)
{
  SH2add(sh,4,15);
}
#define init_3_5_0_12
void FASTCALL call_3_5_0_12(SH2_struct* sh)
{
  SH2add(sh,5,0);
}
#define init_3_5_1_12
void FASTCALL call_3_5_1_12(SH2_struct* sh)
{
  SH2add(sh,5,1);
}
#define init_3_5_2_12
void FASTCALL call_3_5_2_12(SH2_struct* sh)
{
  SH2add(sh,5,2);
}
#define init_3_5_3_12
void FASTCALL call_3_5_3_12(SH2_struct* sh)
{
  SH2add(sh,5,3);
}
#define init_3_5_4_12
void FASTCALL call_3_5_4_12(SH2_struct* sh)
{
  SH2add(sh,5,4);
}
#define init_3_5_5_12
void FASTCALL call_3_5_5_12(SH2_struct* sh)
{
  SH2add(sh,5,5);
}
#define init_3_5_6_12
void FASTCALL call_3_5_6_12(SH2_struct* sh)
{
  SH2add(sh,5,6);
}
#define init_3_5_7_12
void FASTCALL call_3_5_7_12(SH2_struct* sh)
{
  SH2add(sh,5,7);
}
#define init_3_5_8_12
void FASTCALL call_3_5_8_12(SH2_struct* sh)
{
  SH2add(sh,5,8);
}
#define init_3_5_9_12
void FASTCALL call_3_5_9_12(SH2_struct* sh)
{
  SH2add(sh,5,9);
}
#define init_3_5_10_12
void FASTCALL call_3_5_10_12(SH2_struct* sh)
{
  SH2add(sh,5,10);
}
#define init_3_5_11_12
void FASTCALL call_3_5_11_12(SH2_struct* sh)
{
  SH2add(sh,5,11);
}
#define init_3_5_12_12
void FASTCALL call_3_5_12_12(SH2_struct* sh)
{
  SH2add(sh,5,12);
}
#define init_3_5_13_12
void FASTCALL call_3_5_13_12(SH2_struct* sh)
{
  SH2add(sh,5,13);
}
#define init_3_5_14_12
void FASTCALL call_3_5_14_12(SH2_struct* sh)
{
  SH2add(sh,5,14);
}
#define init_3_5_15_12
void FASTCALL call_3_5_15_12(SH2_struct* sh)
{
  SH2add(sh,5,15);
}
#define init_3_6_0_12
void FASTCALL call_3_6_0_12(SH2_struct* sh)
{
  SH2add(sh,6,0);
}
#define init_3_6_1_12
void FASTCALL call_3_6_1_12(SH2_struct* sh)
{
  SH2add(sh,6,1);
}
#define init_3_6_2_12
void FASTCALL call_3_6_2_12(SH2_struct* sh)
{
  SH2add(sh,6,2);
}
#define init_3_6_3_12
void FASTCALL call_3_6_3_12(SH2_struct* sh)
{
  SH2add(sh,6,3);
}
#define init_3_6_4_12
void FASTCALL call_3_6_4_12(SH2_struct* sh)
{
  SH2add(sh,6,4);
}
#define init_3_6_5_12
void FASTCALL call_3_6_5_12(SH2_struct* sh)
{
  SH2add(sh,6,5);
}
#define init_3_6_6_12
void FASTCALL call_3_6_6_12(SH2_struct* sh)
{
  SH2add(sh,6,6);
}
#define init_3_6_7_12
void FASTCALL call_3_6_7_12(SH2_struct* sh)
{
  SH2add(sh,6,7);
}
#define init_3_6_8_12
void FASTCALL call_3_6_8_12(SH2_struct* sh)
{
  SH2add(sh,6,8);
}
#define init_3_6_9_12
void FASTCALL call_3_6_9_12(SH2_struct* sh)
{
  SH2add(sh,6,9);
}
#define init_3_6_10_12
void FASTCALL call_3_6_10_12(SH2_struct* sh)
{
  SH2add(sh,6,10);
}
#define init_3_6_11_12
void FASTCALL call_3_6_11_12(SH2_struct* sh)
{
  SH2add(sh,6,11);
}
#define init_3_6_12_12
void FASTCALL call_3_6_12_12(SH2_struct* sh)
{
  SH2add(sh,6,12);
}
#define init_3_6_13_12
void FASTCALL call_3_6_13_12(SH2_struct* sh)
{
  SH2add(sh,6,13);
}
#define init_3_6_14_12
void FASTCALL call_3_6_14_12(SH2_struct* sh)
{
  SH2add(sh,6,14);
}
#define init_3_6_15_12
void FASTCALL call_3_6_15_12(SH2_struct* sh)
{
  SH2add(sh,6,15);
}
#define init_3_7_0_12
void FASTCALL call_3_7_0_12(SH2_struct* sh)
{
  SH2add(sh,7,0);
}
#define init_3_7_1_12
void FASTCALL call_3_7_1_12(SH2_struct* sh)
{
  SH2add(sh,7,1);
}
#define init_3_7_2_12
void FASTCALL call_3_7_2_12(SH2_struct* sh)
{
  SH2add(sh,7,2);
}
#define init_3_7_3_12
void FASTCALL call_3_7_3_12(SH2_struct* sh)
{
  SH2add(sh,7,3);
}
#define init_3_7_4_12
void FASTCALL call_3_7_4_12(SH2_struct* sh)
{
  SH2add(sh,7,4);
}
#define init_3_7_5_12
void FASTCALL call_3_7_5_12(SH2_struct* sh)
{
  SH2add(sh,7,5);
}
#define init_3_7_6_12
void FASTCALL call_3_7_6_12(SH2_struct* sh)
{
  SH2add(sh,7,6);
}
#define init_3_7_7_12
void FASTCALL call_3_7_7_12(SH2_struct* sh)
{
  SH2add(sh,7,7);
}
#define init_3_7_8_12
void FASTCALL call_3_7_8_12(SH2_struct* sh)
{
  SH2add(sh,7,8);
}
#define init_3_7_9_12
void FASTCALL call_3_7_9_12(SH2_struct* sh)
{
  SH2add(sh,7,9);
}
#define init_3_7_10_12
void FASTCALL call_3_7_10_12(SH2_struct* sh)
{
  SH2add(sh,7,10);
}
#define init_3_7_11_12
void FASTCALL call_3_7_11_12(SH2_struct* sh)
{
  SH2add(sh,7,11);
}
#define init_3_7_12_12
void FASTCALL call_3_7_12_12(SH2_struct* sh)
{
  SH2add(sh,7,12);
}
#define init_3_7_13_12
void FASTCALL call_3_7_13_12(SH2_struct* sh)
{
  SH2add(sh,7,13);
}
#define init_3_7_14_12
void FASTCALL call_3_7_14_12(SH2_struct* sh)
{
  SH2add(sh,7,14);
}
#define init_3_7_15_12
void FASTCALL call_3_7_15_12(SH2_struct* sh)
{
  SH2add(sh,7,15);
}
#define init_3_8_0_12
void FASTCALL call_3_8_0_12(SH2_struct* sh)
{
  SH2add(sh,8,0);
}
#define init_3_8_1_12
void FASTCALL call_3_8_1_12(SH2_struct* sh)
{
  SH2add(sh,8,1);
}
#define init_3_8_2_12
void FASTCALL call_3_8_2_12(SH2_struct* sh)
{
  SH2add(sh,8,2);
}
#define init_3_8_3_12
void FASTCALL call_3_8_3_12(SH2_struct* sh)
{
  SH2add(sh,8,3);
}
#define init_3_8_4_12
void FASTCALL call_3_8_4_12(SH2_struct* sh)
{
  SH2add(sh,8,4);
}
#define init_3_8_5_12
void FASTCALL call_3_8_5_12(SH2_struct* sh)
{
  SH2add(sh,8,5);
}
#define init_3_8_6_12
void FASTCALL call_3_8_6_12(SH2_struct* sh)
{
  SH2add(sh,8,6);
}
#define init_3_8_7_12
void FASTCALL call_3_8_7_12(SH2_struct* sh)
{
  SH2add(sh,8,7);
}
#define init_3_8_8_12
void FASTCALL call_3_8_8_12(SH2_struct* sh)
{
  SH2add(sh,8,8);
}
#define init_3_8_9_12
void FASTCALL call_3_8_9_12(SH2_struct* sh)
{
  SH2add(sh,8,9);
}
#define init_3_8_10_12
void FASTCALL call_3_8_10_12(SH2_struct* sh)
{
  SH2add(sh,8,10);
}
#define init_3_8_11_12
void FASTCALL call_3_8_11_12(SH2_struct* sh)
{
  SH2add(sh,8,11);
}
#define init_3_8_12_12
void FASTCALL call_3_8_12_12(SH2_struct* sh)
{
  SH2add(sh,8,12);
}
#define init_3_8_13_12
void FASTCALL call_3_8_13_12(SH2_struct* sh)
{
  SH2add(sh,8,13);
}
#define init_3_8_14_12
void FASTCALL call_3_8_14_12(SH2_struct* sh)
{
  SH2add(sh,8,14);
}
#define init_3_8_15_12
void FASTCALL call_3_8_15_12(SH2_struct* sh)
{
  SH2add(sh,8,15);
}
#define init_3_9_0_12
void FASTCALL call_3_9_0_12(SH2_struct* sh)
{
  SH2add(sh,9,0);
}
#define init_3_9_1_12
void FASTCALL call_3_9_1_12(SH2_struct* sh)
{
  SH2add(sh,9,1);
}
#define init_3_9_2_12
void FASTCALL call_3_9_2_12(SH2_struct* sh)
{
  SH2add(sh,9,2);
}
#define init_3_9_3_12
void FASTCALL call_3_9_3_12(SH2_struct* sh)
{
  SH2add(sh,9,3);
}
#define init_3_9_4_12
void FASTCALL call_3_9_4_12(SH2_struct* sh)
{
  SH2add(sh,9,4);
}
#define init_3_9_5_12
void FASTCALL call_3_9_5_12(SH2_struct* sh)
{
  SH2add(sh,9,5);
}
#define init_3_9_6_12
void FASTCALL call_3_9_6_12(SH2_struct* sh)
{
  SH2add(sh,9,6);
}
#define init_3_9_7_12
void FASTCALL call_3_9_7_12(SH2_struct* sh)
{
  SH2add(sh,9,7);
}
#define init_3_9_8_12
void FASTCALL call_3_9_8_12(SH2_struct* sh)
{
  SH2add(sh,9,8);
}
#define init_3_9_9_12
void FASTCALL call_3_9_9_12(SH2_struct* sh)
{
  SH2add(sh,9,9);
}
#define init_3_9_10_12
void FASTCALL call_3_9_10_12(SH2_struct* sh)
{
  SH2add(sh,9,10);
}
#define init_3_9_11_12
void FASTCALL call_3_9_11_12(SH2_struct* sh)
{
  SH2add(sh,9,11);
}
#define init_3_9_12_12
void FASTCALL call_3_9_12_12(SH2_struct* sh)
{
  SH2add(sh,9,12);
}
#define init_3_9_13_12
void FASTCALL call_3_9_13_12(SH2_struct* sh)
{
  SH2add(sh,9,13);
}
#define init_3_9_14_12
void FASTCALL call_3_9_14_12(SH2_struct* sh)
{
  SH2add(sh,9,14);
}
#define init_3_9_15_12
void FASTCALL call_3_9_15_12(SH2_struct* sh)
{
  SH2add(sh,9,15);
}
#define init_3_10_0_12
void FASTCALL call_3_10_0_12(SH2_struct* sh)
{
  SH2add(sh,10,0);
}
#define init_3_10_1_12
void FASTCALL call_3_10_1_12(SH2_struct* sh)
{
  SH2add(sh,10,1);
}
#define init_3_10_2_12
void FASTCALL call_3_10_2_12(SH2_struct* sh)
{
  SH2add(sh,10,2);
}
#define init_3_10_3_12
void FASTCALL call_3_10_3_12(SH2_struct* sh)
{
  SH2add(sh,10,3);
}
#define init_3_10_4_12
void FASTCALL call_3_10_4_12(SH2_struct* sh)
{
  SH2add(sh,10,4);
}
#define init_3_10_5_12
void FASTCALL call_3_10_5_12(SH2_struct* sh)
{
  SH2add(sh,10,5);
}
#define init_3_10_6_12
void FASTCALL call_3_10_6_12(SH2_struct* sh)
{
  SH2add(sh,10,6);
}
#define init_3_10_7_12
void FASTCALL call_3_10_7_12(SH2_struct* sh)
{
  SH2add(sh,10,7);
}
#define init_3_10_8_12
void FASTCALL call_3_10_8_12(SH2_struct* sh)
{
  SH2add(sh,10,8);
}
#define init_3_10_9_12
void FASTCALL call_3_10_9_12(SH2_struct* sh)
{
  SH2add(sh,10,9);
}
#define init_3_10_10_12
void FASTCALL call_3_10_10_12(SH2_struct* sh)
{
  SH2add(sh,10,10);
}
#define init_3_10_11_12
void FASTCALL call_3_10_11_12(SH2_struct* sh)
{
  SH2add(sh,10,11);
}
#define init_3_10_12_12
void FASTCALL call_3_10_12_12(SH2_struct* sh)
{
  SH2add(sh,10,12);
}
#define init_3_10_13_12
void FASTCALL call_3_10_13_12(SH2_struct* sh)
{
  SH2add(sh,10,13);
}
#define init_3_10_14_12
void FASTCALL call_3_10_14_12(SH2_struct* sh)
{
  SH2add(sh,10,14);
}
#define init_3_10_15_12
void FASTCALL call_3_10_15_12(SH2_struct* sh)
{
  SH2add(sh,10,15);
}
#define init_3_11_0_12
void FASTCALL call_3_11_0_12(SH2_struct* sh)
{
  SH2add(sh,11,0);
}
#define init_3_11_1_12
void FASTCALL call_3_11_1_12(SH2_struct* sh)
{
  SH2add(sh,11,1);
}
#define init_3_11_2_12
void FASTCALL call_3_11_2_12(SH2_struct* sh)
{
  SH2add(sh,11,2);
}
#define init_3_11_3_12
void FASTCALL call_3_11_3_12(SH2_struct* sh)
{
  SH2add(sh,11,3);
}
#define init_3_11_4_12
void FASTCALL call_3_11_4_12(SH2_struct* sh)
{
  SH2add(sh,11,4);
}
#define init_3_11_5_12
void FASTCALL call_3_11_5_12(SH2_struct* sh)
{
  SH2add(sh,11,5);
}
#define init_3_11_6_12
void FASTCALL call_3_11_6_12(SH2_struct* sh)
{
  SH2add(sh,11,6);
}
#define init_3_11_7_12
void FASTCALL call_3_11_7_12(SH2_struct* sh)
{
  SH2add(sh,11,7);
}
#define init_3_11_8_12
void FASTCALL call_3_11_8_12(SH2_struct* sh)
{
  SH2add(sh,11,8);
}
#define init_3_11_9_12
void FASTCALL call_3_11_9_12(SH2_struct* sh)
{
  SH2add(sh,11,9);
}
#define init_3_11_10_12
void FASTCALL call_3_11_10_12(SH2_struct* sh)
{
  SH2add(sh,11,10);
}
#define init_3_11_11_12
void FASTCALL call_3_11_11_12(SH2_struct* sh)
{
  SH2add(sh,11,11);
}
#define init_3_11_12_12
void FASTCALL call_3_11_12_12(SH2_struct* sh)
{
  SH2add(sh,11,12);
}
#define init_3_11_13_12
void FASTCALL call_3_11_13_12(SH2_struct* sh)
{
  SH2add(sh,11,13);
}
#define init_3_11_14_12
void FASTCALL call_3_11_14_12(SH2_struct* sh)
{
  SH2add(sh,11,14);
}
#define init_3_11_15_12
void FASTCALL call_3_11_15_12(SH2_struct* sh)
{
  SH2add(sh,11,15);
}
#define init_3_12_0_12
void FASTCALL call_3_12_0_12(SH2_struct* sh)
{
  SH2add(sh,12,0);
}
#define init_3_12_1_12
void FASTCALL call_3_12_1_12(SH2_struct* sh)
{
  SH2add(sh,12,1);
}
#define init_3_12_2_12
void FASTCALL call_3_12_2_12(SH2_struct* sh)
{
  SH2add(sh,12,2);
}
#define init_3_12_3_12
void FASTCALL call_3_12_3_12(SH2_struct* sh)
{
  SH2add(sh,12,3);
}
#define init_3_12_4_12
void FASTCALL call_3_12_4_12(SH2_struct* sh)
{
  SH2add(sh,12,4);
}
#define init_3_12_5_12
void FASTCALL call_3_12_5_12(SH2_struct* sh)
{
  SH2add(sh,12,5);
}
#define init_3_12_6_12
void FASTCALL call_3_12_6_12(SH2_struct* sh)
{
  SH2add(sh,12,6);
}
#define init_3_12_7_12
void FASTCALL call_3_12_7_12(SH2_struct* sh)
{
  SH2add(sh,12,7);
}
#define init_3_12_8_12
void FASTCALL call_3_12_8_12(SH2_struct* sh)
{
  SH2add(sh,12,8);
}
#define init_3_12_9_12
void FASTCALL call_3_12_9_12(SH2_struct* sh)
{
  SH2add(sh,12,9);
}
#define init_3_12_10_12
void FASTCALL call_3_12_10_12(SH2_struct* sh)
{
  SH2add(sh,12,10);
}
#define init_3_12_11_12
void FASTCALL call_3_12_11_12(SH2_struct* sh)
{
  SH2add(sh,12,11);
}
#define init_3_12_12_12
void FASTCALL call_3_12_12_12(SH2_struct* sh)
{
  SH2add(sh,12,12);
}
#define init_3_12_13_12
void FASTCALL call_3_12_13_12(SH2_struct* sh)
{
  SH2add(sh,12,13);
}
#define init_3_12_14_12
void FASTCALL call_3_12_14_12(SH2_struct* sh)
{
  SH2add(sh,12,14);
}
#define init_3_12_15_12
void FASTCALL call_3_12_15_12(SH2_struct* sh)
{
  SH2add(sh,12,15);
}
#define init_3_13_0_12
void FASTCALL call_3_13_0_12(SH2_struct* sh)
{
  SH2add(sh,13,0);
}
#define init_3_13_1_12
void FASTCALL call_3_13_1_12(SH2_struct* sh)
{
  SH2add(sh,13,1);
}
#define init_3_13_2_12
void FASTCALL call_3_13_2_12(SH2_struct* sh)
{
  SH2add(sh,13,2);
}
#define init_3_13_3_12
void FASTCALL call_3_13_3_12(SH2_struct* sh)
{
  SH2add(sh,13,3);
}
#define init_3_13_4_12
void FASTCALL call_3_13_4_12(SH2_struct* sh)
{
  SH2add(sh,13,4);
}
#define init_3_13_5_12
void FASTCALL call_3_13_5_12(SH2_struct* sh)
{
  SH2add(sh,13,5);
}
#define init_3_13_6_12
void FASTCALL call_3_13_6_12(SH2_struct* sh)
{
  SH2add(sh,13,6);
}
#define init_3_13_7_12
void FASTCALL call_3_13_7_12(SH2_struct* sh)
{
  SH2add(sh,13,7);
}
#define init_3_13_8_12
void FASTCALL call_3_13_8_12(SH2_struct* sh)
{
  SH2add(sh,13,8);
}
#define init_3_13_9_12
void FASTCALL call_3_13_9_12(SH2_struct* sh)
{
  SH2add(sh,13,9);
}
#define init_3_13_10_12
void FASTCALL call_3_13_10_12(SH2_struct* sh)
{
  SH2add(sh,13,10);
}
#define init_3_13_11_12
void FASTCALL call_3_13_11_12(SH2_struct* sh)
{
  SH2add(sh,13,11);
}
#define init_3_13_12_12
void FASTCALL call_3_13_12_12(SH2_struct* sh)
{
  SH2add(sh,13,12);
}
#define init_3_13_13_12
void FASTCALL call_3_13_13_12(SH2_struct* sh)
{
  SH2add(sh,13,13);
}
#define init_3_13_14_12
void FASTCALL call_3_13_14_12(SH2_struct* sh)
{
  SH2add(sh,13,14);
}
#define init_3_13_15_12
void FASTCALL call_3_13_15_12(SH2_struct* sh)
{
  SH2add(sh,13,15);
}
#define init_3_14_0_12
void FASTCALL call_3_14_0_12(SH2_struct* sh)
{
  SH2add(sh,14,0);
}
#define init_3_14_1_12
void FASTCALL call_3_14_1_12(SH2_struct* sh)
{
  SH2add(sh,14,1);
}
#define init_3_14_2_12
void FASTCALL call_3_14_2_12(SH2_struct* sh)
{
  SH2add(sh,14,2);
}
#define init_3_14_3_12
void FASTCALL call_3_14_3_12(SH2_struct* sh)
{
  SH2add(sh,14,3);
}
#define init_3_14_4_12
void FASTCALL call_3_14_4_12(SH2_struct* sh)
{
  SH2add(sh,14,4);
}
#define init_3_14_5_12
void FASTCALL call_3_14_5_12(SH2_struct* sh)
{
  SH2add(sh,14,5);
}
#define init_3_14_6_12
void FASTCALL call_3_14_6_12(SH2_struct* sh)
{
  SH2add(sh,14,6);
}
#define init_3_14_7_12
void FASTCALL call_3_14_7_12(SH2_struct* sh)
{
  SH2add(sh,14,7);
}
#define init_3_14_8_12
void FASTCALL call_3_14_8_12(SH2_struct* sh)
{
  SH2add(sh,14,8);
}
#define init_3_14_9_12
void FASTCALL call_3_14_9_12(SH2_struct* sh)
{
  SH2add(sh,14,9);
}
#define init_3_14_10_12
void FASTCALL call_3_14_10_12(SH2_struct* sh)
{
  SH2add(sh,14,10);
}
#define init_3_14_11_12
void FASTCALL call_3_14_11_12(SH2_struct* sh)
{
  SH2add(sh,14,11);
}
#define init_3_14_12_12
void FASTCALL call_3_14_12_12(SH2_struct* sh)
{
  SH2add(sh,14,12);
}
#define init_3_14_13_12
void FASTCALL call_3_14_13_12(SH2_struct* sh)
{
  SH2add(sh,14,13);
}
#define init_3_14_14_12
void FASTCALL call_3_14_14_12(SH2_struct* sh)
{
  SH2add(sh,14,14);
}
#define init_3_14_15_12
void FASTCALL call_3_14_15_12(SH2_struct* sh)
{
  SH2add(sh,14,15);
}
#define init_3_15_0_12
void FASTCALL call_3_15_0_12(SH2_struct* sh)
{
  SH2add(sh,15,0);
}
#define init_3_15_1_12
void FASTCALL call_3_15_1_12(SH2_struct* sh)
{
  SH2add(sh,15,1);
}
#define init_3_15_2_12
void FASTCALL call_3_15_2_12(SH2_struct* sh)
{
  SH2add(sh,15,2);
}
#define init_3_15_3_12
void FASTCALL call_3_15_3_12(SH2_struct* sh)
{
  SH2add(sh,15,3);
}
#define init_3_15_4_12
void FASTCALL call_3_15_4_12(SH2_struct* sh)
{
  SH2add(sh,15,4);
}
#define init_3_15_5_12
void FASTCALL call_3_15_5_12(SH2_struct* sh)
{
  SH2add(sh,15,5);
}
#define init_3_15_6_12
void FASTCALL call_3_15_6_12(SH2_struct* sh)
{
  SH2add(sh,15,6);
}
#define init_3_15_7_12
void FASTCALL call_3_15_7_12(SH2_struct* sh)
{
  SH2add(sh,15,7);
}
#define init_3_15_8_12
void FASTCALL call_3_15_8_12(SH2_struct* sh)
{
  SH2add(sh,15,8);
}
#define init_3_15_9_12
void FASTCALL call_3_15_9_12(SH2_struct* sh)
{
  SH2add(sh,15,9);
}
#define init_3_15_10_12
void FASTCALL call_3_15_10_12(SH2_struct* sh)
{
  SH2add(sh,15,10);
}
#define init_3_15_11_12
void FASTCALL call_3_15_11_12(SH2_struct* sh)
{
  SH2add(sh,15,11);
}
#define init_3_15_12_12
void FASTCALL call_3_15_12_12(SH2_struct* sh)
{
  SH2add(sh,15,12);
}
#define init_3_15_13_12
void FASTCALL call_3_15_13_12(SH2_struct* sh)
{
  SH2add(sh,15,13);
}
#define init_3_15_14_12
void FASTCALL call_3_15_14_12(SH2_struct* sh)
{
  SH2add(sh,15,14);
}
#define init_3_15_15_12
void FASTCALL call_3_15_15_12(SH2_struct* sh)
{
  SH2add(sh,15,15);
}
#ifndef init_0_0_0_0
#define init_0_0_0_0
void FASTCALL call_0_0_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_0_1
#define init_0_0_0_1
void FASTCALL call_0_0_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_0_2
#define init_0_0_0_2
void FASTCALL call_0_0_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_0_3
#define init_0_0_0_3
void FASTCALL call_0_0_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_0_4
#define init_0_0_0_4
void FASTCALL call_0_0_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_0_5
#define init_0_0_0_5
void FASTCALL call_0_0_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_0_6
#define init_0_0_0_6
void FASTCALL call_0_0_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_0_7
#define init_0_0_0_7
void FASTCALL call_0_0_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_0_8
#define init_0_0_0_8
void FASTCALL call_0_0_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_0_9
#define init_0_0_0_9
void FASTCALL call_0_0_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_0_10
#define init_0_0_0_10
void FASTCALL call_0_0_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_0_11
#define init_0_0_0_11
void FASTCALL call_0_0_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_0_12
#define init_0_0_0_12
void FASTCALL call_0_0_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_0_13
#define init_0_0_0_13
void FASTCALL call_0_0_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_0_14
#define init_0_0_0_14
void FASTCALL call_0_0_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_0_15
#define init_0_0_0_15
void FASTCALL call_0_0_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_1_0
#define init_0_0_1_0
void FASTCALL call_0_0_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_1_1
#define init_0_0_1_1
void FASTCALL call_0_0_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_1_2
#define init_0_0_1_2
void FASTCALL call_0_0_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_1_3
#define init_0_0_1_3
void FASTCALL call_0_0_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_1_4
#define init_0_0_1_4
void FASTCALL call_0_0_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_1_5
#define init_0_0_1_5
void FASTCALL call_0_0_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_1_6
#define init_0_0_1_6
void FASTCALL call_0_0_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_1_7
#define init_0_0_1_7
void FASTCALL call_0_0_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_1_8
#define init_0_0_1_8
void FASTCALL call_0_0_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_1_9
#define init_0_0_1_9
void FASTCALL call_0_0_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_1_10
#define init_0_0_1_10
void FASTCALL call_0_0_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_1_11
#define init_0_0_1_11
void FASTCALL call_0_0_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_1_12
#define init_0_0_1_12
void FASTCALL call_0_0_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_1_13
#define init_0_0_1_13
void FASTCALL call_0_0_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_1_14
#define init_0_0_1_14
void FASTCALL call_0_0_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_1_15
#define init_0_0_1_15
void FASTCALL call_0_0_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_2_0
#define init_0_0_2_0
void FASTCALL call_0_0_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_2_1
#define init_0_0_2_1
void FASTCALL call_0_0_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_2_2
#define init_0_0_2_2
void FASTCALL call_0_0_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_2_3
#define init_0_0_2_3
void FASTCALL call_0_0_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_2_4
#define init_0_0_2_4
void FASTCALL call_0_0_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_2_5
#define init_0_0_2_5
void FASTCALL call_0_0_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_2_6
#define init_0_0_2_6
void FASTCALL call_0_0_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_2_7
#define init_0_0_2_7
void FASTCALL call_0_0_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_2_8
#define init_0_0_2_8
void FASTCALL call_0_0_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_2_9
#define init_0_0_2_9
void FASTCALL call_0_0_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_2_10
#define init_0_0_2_10
void FASTCALL call_0_0_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_2_11
#define init_0_0_2_11
void FASTCALL call_0_0_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_2_12
#define init_0_0_2_12
void FASTCALL call_0_0_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_2_13
#define init_0_0_2_13
void FASTCALL call_0_0_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_2_14
#define init_0_0_2_14
void FASTCALL call_0_0_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_2_15
#define init_0_0_2_15
void FASTCALL call_0_0_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_3_0
#define init_0_0_3_0
void FASTCALL call_0_0_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_3_1
#define init_0_0_3_1
void FASTCALL call_0_0_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_3_2
#define init_0_0_3_2
void FASTCALL call_0_0_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_3_3
#define init_0_0_3_3
void FASTCALL call_0_0_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_3_4
#define init_0_0_3_4
void FASTCALL call_0_0_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_3_5
#define init_0_0_3_5
void FASTCALL call_0_0_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_3_6
#define init_0_0_3_6
void FASTCALL call_0_0_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_3_7
#define init_0_0_3_7
void FASTCALL call_0_0_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_3_8
#define init_0_0_3_8
void FASTCALL call_0_0_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_3_9
#define init_0_0_3_9
void FASTCALL call_0_0_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_3_10
#define init_0_0_3_10
void FASTCALL call_0_0_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_3_11
#define init_0_0_3_11
void FASTCALL call_0_0_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_3_12
#define init_0_0_3_12
void FASTCALL call_0_0_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_3_13
#define init_0_0_3_13
void FASTCALL call_0_0_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_3_14
#define init_0_0_3_14
void FASTCALL call_0_0_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_3_15
#define init_0_0_3_15
void FASTCALL call_0_0_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_4_0
#define init_0_0_4_0
void FASTCALL call_0_0_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_4_1
#define init_0_0_4_1
void FASTCALL call_0_0_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_4_2
#define init_0_0_4_2
void FASTCALL call_0_0_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_4_3
#define init_0_0_4_3
void FASTCALL call_0_0_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_4_4
#define init_0_0_4_4
void FASTCALL call_0_0_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_4_5
#define init_0_0_4_5
void FASTCALL call_0_0_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_4_6
#define init_0_0_4_6
void FASTCALL call_0_0_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_4_7
#define init_0_0_4_7
void FASTCALL call_0_0_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_4_8
#define init_0_0_4_8
void FASTCALL call_0_0_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_4_9
#define init_0_0_4_9
void FASTCALL call_0_0_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_4_10
#define init_0_0_4_10
void FASTCALL call_0_0_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_4_11
#define init_0_0_4_11
void FASTCALL call_0_0_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_4_12
#define init_0_0_4_12
void FASTCALL call_0_0_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_4_13
#define init_0_0_4_13
void FASTCALL call_0_0_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_4_14
#define init_0_0_4_14
void FASTCALL call_0_0_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_4_15
#define init_0_0_4_15
void FASTCALL call_0_0_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_5_0
#define init_0_0_5_0
void FASTCALL call_0_0_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_5_1
#define init_0_0_5_1
void FASTCALL call_0_0_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_5_2
#define init_0_0_5_2
void FASTCALL call_0_0_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_5_3
#define init_0_0_5_3
void FASTCALL call_0_0_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_5_4
#define init_0_0_5_4
void FASTCALL call_0_0_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_5_5
#define init_0_0_5_5
void FASTCALL call_0_0_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_5_6
#define init_0_0_5_6
void FASTCALL call_0_0_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_5_7
#define init_0_0_5_7
void FASTCALL call_0_0_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_5_8
#define init_0_0_5_8
void FASTCALL call_0_0_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_5_9
#define init_0_0_5_9
void FASTCALL call_0_0_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_5_10
#define init_0_0_5_10
void FASTCALL call_0_0_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_5_11
#define init_0_0_5_11
void FASTCALL call_0_0_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_5_12
#define init_0_0_5_12
void FASTCALL call_0_0_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_5_13
#define init_0_0_5_13
void FASTCALL call_0_0_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_5_14
#define init_0_0_5_14
void FASTCALL call_0_0_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_5_15
#define init_0_0_5_15
void FASTCALL call_0_0_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_6_0
#define init_0_0_6_0
void FASTCALL call_0_0_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_6_1
#define init_0_0_6_1
void FASTCALL call_0_0_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_6_2
#define init_0_0_6_2
void FASTCALL call_0_0_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_6_3
#define init_0_0_6_3
void FASTCALL call_0_0_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_6_4
#define init_0_0_6_4
void FASTCALL call_0_0_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_6_5
#define init_0_0_6_5
void FASTCALL call_0_0_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_6_6
#define init_0_0_6_6
void FASTCALL call_0_0_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_6_7
#define init_0_0_6_7
void FASTCALL call_0_0_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_6_8
#define init_0_0_6_8
void FASTCALL call_0_0_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_6_9
#define init_0_0_6_9
void FASTCALL call_0_0_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_6_10
#define init_0_0_6_10
void FASTCALL call_0_0_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_6_11
#define init_0_0_6_11
void FASTCALL call_0_0_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_6_12
#define init_0_0_6_12
void FASTCALL call_0_0_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_6_13
#define init_0_0_6_13
void FASTCALL call_0_0_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_6_14
#define init_0_0_6_14
void FASTCALL call_0_0_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_6_15
#define init_0_0_6_15
void FASTCALL call_0_0_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_7_0
#define init_0_0_7_0
void FASTCALL call_0_0_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_7_1
#define init_0_0_7_1
void FASTCALL call_0_0_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_7_2
#define init_0_0_7_2
void FASTCALL call_0_0_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_7_3
#define init_0_0_7_3
void FASTCALL call_0_0_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_7_4
#define init_0_0_7_4
void FASTCALL call_0_0_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_7_5
#define init_0_0_7_5
void FASTCALL call_0_0_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_7_6
#define init_0_0_7_6
void FASTCALL call_0_0_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_7_7
#define init_0_0_7_7
void FASTCALL call_0_0_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_7_8
#define init_0_0_7_8
void FASTCALL call_0_0_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_7_9
#define init_0_0_7_9
void FASTCALL call_0_0_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_7_10
#define init_0_0_7_10
void FASTCALL call_0_0_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_7_11
#define init_0_0_7_11
void FASTCALL call_0_0_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_7_12
#define init_0_0_7_12
void FASTCALL call_0_0_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_7_13
#define init_0_0_7_13
void FASTCALL call_0_0_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_7_14
#define init_0_0_7_14
void FASTCALL call_0_0_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_7_15
#define init_0_0_7_15
void FASTCALL call_0_0_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_8_0
#define init_0_0_8_0
void FASTCALL call_0_0_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_8_1
#define init_0_0_8_1
void FASTCALL call_0_0_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_8_2
#define init_0_0_8_2
void FASTCALL call_0_0_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_8_3
#define init_0_0_8_3
void FASTCALL call_0_0_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_8_4
#define init_0_0_8_4
void FASTCALL call_0_0_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_8_5
#define init_0_0_8_5
void FASTCALL call_0_0_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_8_6
#define init_0_0_8_6
void FASTCALL call_0_0_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_8_7
#define init_0_0_8_7
void FASTCALL call_0_0_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_8_8
#define init_0_0_8_8
void FASTCALL call_0_0_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_8_9
#define init_0_0_8_9
void FASTCALL call_0_0_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_8_10
#define init_0_0_8_10
void FASTCALL call_0_0_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_8_11
#define init_0_0_8_11
void FASTCALL call_0_0_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_8_12
#define init_0_0_8_12
void FASTCALL call_0_0_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_8_13
#define init_0_0_8_13
void FASTCALL call_0_0_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_8_14
#define init_0_0_8_14
void FASTCALL call_0_0_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_8_15
#define init_0_0_8_15
void FASTCALL call_0_0_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_9_0
#define init_0_0_9_0
void FASTCALL call_0_0_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_9_1
#define init_0_0_9_1
void FASTCALL call_0_0_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_9_2
#define init_0_0_9_2
void FASTCALL call_0_0_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_9_3
#define init_0_0_9_3
void FASTCALL call_0_0_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_9_4
#define init_0_0_9_4
void FASTCALL call_0_0_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_9_5
#define init_0_0_9_5
void FASTCALL call_0_0_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_9_6
#define init_0_0_9_6
void FASTCALL call_0_0_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_9_7
#define init_0_0_9_7
void FASTCALL call_0_0_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_9_8
#define init_0_0_9_8
void FASTCALL call_0_0_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_9_9
#define init_0_0_9_9
void FASTCALL call_0_0_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_9_10
#define init_0_0_9_10
void FASTCALL call_0_0_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_9_11
#define init_0_0_9_11
void FASTCALL call_0_0_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_9_12
#define init_0_0_9_12
void FASTCALL call_0_0_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_9_13
#define init_0_0_9_13
void FASTCALL call_0_0_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_9_14
#define init_0_0_9_14
void FASTCALL call_0_0_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_9_15
#define init_0_0_9_15
void FASTCALL call_0_0_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_10_0
#define init_0_0_10_0
void FASTCALL call_0_0_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_10_1
#define init_0_0_10_1
void FASTCALL call_0_0_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_10_2
#define init_0_0_10_2
void FASTCALL call_0_0_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_10_3
#define init_0_0_10_3
void FASTCALL call_0_0_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_10_4
#define init_0_0_10_4
void FASTCALL call_0_0_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_10_5
#define init_0_0_10_5
void FASTCALL call_0_0_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_10_6
#define init_0_0_10_6
void FASTCALL call_0_0_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_10_7
#define init_0_0_10_7
void FASTCALL call_0_0_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_10_8
#define init_0_0_10_8
void FASTCALL call_0_0_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_10_9
#define init_0_0_10_9
void FASTCALL call_0_0_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_10_10
#define init_0_0_10_10
void FASTCALL call_0_0_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_10_11
#define init_0_0_10_11
void FASTCALL call_0_0_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_10_12
#define init_0_0_10_12
void FASTCALL call_0_0_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_10_13
#define init_0_0_10_13
void FASTCALL call_0_0_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_10_14
#define init_0_0_10_14
void FASTCALL call_0_0_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_10_15
#define init_0_0_10_15
void FASTCALL call_0_0_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_11_0
#define init_0_0_11_0
void FASTCALL call_0_0_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_11_1
#define init_0_0_11_1
void FASTCALL call_0_0_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_11_2
#define init_0_0_11_2
void FASTCALL call_0_0_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_11_3
#define init_0_0_11_3
void FASTCALL call_0_0_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_11_4
#define init_0_0_11_4
void FASTCALL call_0_0_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_11_5
#define init_0_0_11_5
void FASTCALL call_0_0_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_11_6
#define init_0_0_11_6
void FASTCALL call_0_0_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_11_7
#define init_0_0_11_7
void FASTCALL call_0_0_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_11_8
#define init_0_0_11_8
void FASTCALL call_0_0_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_11_9
#define init_0_0_11_9
void FASTCALL call_0_0_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_11_10
#define init_0_0_11_10
void FASTCALL call_0_0_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_11_11
#define init_0_0_11_11
void FASTCALL call_0_0_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_11_12
#define init_0_0_11_12
void FASTCALL call_0_0_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_11_13
#define init_0_0_11_13
void FASTCALL call_0_0_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_11_14
#define init_0_0_11_14
void FASTCALL call_0_0_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_11_15
#define init_0_0_11_15
void FASTCALL call_0_0_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_12_0
#define init_0_0_12_0
void FASTCALL call_0_0_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_12_1
#define init_0_0_12_1
void FASTCALL call_0_0_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_12_2
#define init_0_0_12_2
void FASTCALL call_0_0_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_12_3
#define init_0_0_12_3
void FASTCALL call_0_0_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_12_4
#define init_0_0_12_4
void FASTCALL call_0_0_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_12_5
#define init_0_0_12_5
void FASTCALL call_0_0_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_12_6
#define init_0_0_12_6
void FASTCALL call_0_0_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_12_7
#define init_0_0_12_7
void FASTCALL call_0_0_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_12_8
#define init_0_0_12_8
void FASTCALL call_0_0_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_12_9
#define init_0_0_12_9
void FASTCALL call_0_0_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_12_10
#define init_0_0_12_10
void FASTCALL call_0_0_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_12_11
#define init_0_0_12_11
void FASTCALL call_0_0_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_12_12
#define init_0_0_12_12
void FASTCALL call_0_0_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_12_13
#define init_0_0_12_13
void FASTCALL call_0_0_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_12_14
#define init_0_0_12_14
void FASTCALL call_0_0_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_12_15
#define init_0_0_12_15
void FASTCALL call_0_0_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_13_0
#define init_0_0_13_0
void FASTCALL call_0_0_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_13_1
#define init_0_0_13_1
void FASTCALL call_0_0_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_13_2
#define init_0_0_13_2
void FASTCALL call_0_0_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_13_3
#define init_0_0_13_3
void FASTCALL call_0_0_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_13_4
#define init_0_0_13_4
void FASTCALL call_0_0_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_13_5
#define init_0_0_13_5
void FASTCALL call_0_0_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_13_6
#define init_0_0_13_6
void FASTCALL call_0_0_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_13_7
#define init_0_0_13_7
void FASTCALL call_0_0_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_13_8
#define init_0_0_13_8
void FASTCALL call_0_0_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_13_9
#define init_0_0_13_9
void FASTCALL call_0_0_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_13_10
#define init_0_0_13_10
void FASTCALL call_0_0_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_13_11
#define init_0_0_13_11
void FASTCALL call_0_0_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_13_12
#define init_0_0_13_12
void FASTCALL call_0_0_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_13_13
#define init_0_0_13_13
void FASTCALL call_0_0_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_13_14
#define init_0_0_13_14
void FASTCALL call_0_0_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_13_15
#define init_0_0_13_15
void FASTCALL call_0_0_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_14_0
#define init_0_0_14_0
void FASTCALL call_0_0_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_14_1
#define init_0_0_14_1
void FASTCALL call_0_0_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_14_2
#define init_0_0_14_2
void FASTCALL call_0_0_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_14_3
#define init_0_0_14_3
void FASTCALL call_0_0_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_14_4
#define init_0_0_14_4
void FASTCALL call_0_0_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_14_5
#define init_0_0_14_5
void FASTCALL call_0_0_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_14_6
#define init_0_0_14_6
void FASTCALL call_0_0_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_14_7
#define init_0_0_14_7
void FASTCALL call_0_0_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_14_8
#define init_0_0_14_8
void FASTCALL call_0_0_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_14_9
#define init_0_0_14_9
void FASTCALL call_0_0_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_14_10
#define init_0_0_14_10
void FASTCALL call_0_0_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_14_11
#define init_0_0_14_11
void FASTCALL call_0_0_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_14_12
#define init_0_0_14_12
void FASTCALL call_0_0_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_14_13
#define init_0_0_14_13
void FASTCALL call_0_0_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_14_14
#define init_0_0_14_14
void FASTCALL call_0_0_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_14_15
#define init_0_0_14_15
void FASTCALL call_0_0_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_15_0
#define init_0_0_15_0
void FASTCALL call_0_0_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_15_1
#define init_0_0_15_1
void FASTCALL call_0_0_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_15_2
#define init_0_0_15_2
void FASTCALL call_0_0_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_15_3
#define init_0_0_15_3
void FASTCALL call_0_0_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_15_4
#define init_0_0_15_4
void FASTCALL call_0_0_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_15_5
#define init_0_0_15_5
void FASTCALL call_0_0_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_15_6
#define init_0_0_15_6
void FASTCALL call_0_0_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_15_7
#define init_0_0_15_7
void FASTCALL call_0_0_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_15_8
#define init_0_0_15_8
void FASTCALL call_0_0_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_15_9
#define init_0_0_15_9
void FASTCALL call_0_0_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_15_10
#define init_0_0_15_10
void FASTCALL call_0_0_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_15_11
#define init_0_0_15_11
void FASTCALL call_0_0_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_15_12
#define init_0_0_15_12
void FASTCALL call_0_0_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_15_13
#define init_0_0_15_13
void FASTCALL call_0_0_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_15_14
#define init_0_0_15_14
void FASTCALL call_0_0_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_0_15_15
#define init_0_0_15_15
void FASTCALL call_0_0_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_0_0
#define init_0_1_0_0
void FASTCALL call_0_1_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_0_1
#define init_0_1_0_1
void FASTCALL call_0_1_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_0_2
#define init_0_1_0_2
void FASTCALL call_0_1_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_0_3
#define init_0_1_0_3
void FASTCALL call_0_1_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_0_4
#define init_0_1_0_4
void FASTCALL call_0_1_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_0_5
#define init_0_1_0_5
void FASTCALL call_0_1_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_0_6
#define init_0_1_0_6
void FASTCALL call_0_1_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_0_7
#define init_0_1_0_7
void FASTCALL call_0_1_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_0_8
#define init_0_1_0_8
void FASTCALL call_0_1_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_0_9
#define init_0_1_0_9
void FASTCALL call_0_1_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_0_10
#define init_0_1_0_10
void FASTCALL call_0_1_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_0_11
#define init_0_1_0_11
void FASTCALL call_0_1_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_0_12
#define init_0_1_0_12
void FASTCALL call_0_1_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_0_13
#define init_0_1_0_13
void FASTCALL call_0_1_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_0_14
#define init_0_1_0_14
void FASTCALL call_0_1_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_0_15
#define init_0_1_0_15
void FASTCALL call_0_1_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_1_0
#define init_0_1_1_0
void FASTCALL call_0_1_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_1_1
#define init_0_1_1_1
void FASTCALL call_0_1_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_1_2
#define init_0_1_1_2
void FASTCALL call_0_1_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_1_3
#define init_0_1_1_3
void FASTCALL call_0_1_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_1_4
#define init_0_1_1_4
void FASTCALL call_0_1_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_1_5
#define init_0_1_1_5
void FASTCALL call_0_1_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_1_6
#define init_0_1_1_6
void FASTCALL call_0_1_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_1_7
#define init_0_1_1_7
void FASTCALL call_0_1_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_1_8
#define init_0_1_1_8
void FASTCALL call_0_1_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_1_9
#define init_0_1_1_9
void FASTCALL call_0_1_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_1_10
#define init_0_1_1_10
void FASTCALL call_0_1_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_1_11
#define init_0_1_1_11
void FASTCALL call_0_1_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_1_12
#define init_0_1_1_12
void FASTCALL call_0_1_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_1_13
#define init_0_1_1_13
void FASTCALL call_0_1_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_1_14
#define init_0_1_1_14
void FASTCALL call_0_1_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_1_15
#define init_0_1_1_15
void FASTCALL call_0_1_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_2_0
#define init_0_1_2_0
void FASTCALL call_0_1_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_2_1
#define init_0_1_2_1
void FASTCALL call_0_1_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_2_2
#define init_0_1_2_2
void FASTCALL call_0_1_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_2_3
#define init_0_1_2_3
void FASTCALL call_0_1_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_2_4
#define init_0_1_2_4
void FASTCALL call_0_1_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_2_5
#define init_0_1_2_5
void FASTCALL call_0_1_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_2_6
#define init_0_1_2_6
void FASTCALL call_0_1_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_2_7
#define init_0_1_2_7
void FASTCALL call_0_1_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_2_8
#define init_0_1_2_8
void FASTCALL call_0_1_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_2_9
#define init_0_1_2_9
void FASTCALL call_0_1_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_2_10
#define init_0_1_2_10
void FASTCALL call_0_1_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_2_11
#define init_0_1_2_11
void FASTCALL call_0_1_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_2_12
#define init_0_1_2_12
void FASTCALL call_0_1_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_2_13
#define init_0_1_2_13
void FASTCALL call_0_1_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_2_14
#define init_0_1_2_14
void FASTCALL call_0_1_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_2_15
#define init_0_1_2_15
void FASTCALL call_0_1_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_3_0
#define init_0_1_3_0
void FASTCALL call_0_1_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_3_1
#define init_0_1_3_1
void FASTCALL call_0_1_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_3_2
#define init_0_1_3_2
void FASTCALL call_0_1_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_3_3
#define init_0_1_3_3
void FASTCALL call_0_1_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_3_4
#define init_0_1_3_4
void FASTCALL call_0_1_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_3_5
#define init_0_1_3_5
void FASTCALL call_0_1_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_3_6
#define init_0_1_3_6
void FASTCALL call_0_1_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_3_7
#define init_0_1_3_7
void FASTCALL call_0_1_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_3_8
#define init_0_1_3_8
void FASTCALL call_0_1_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_3_9
#define init_0_1_3_9
void FASTCALL call_0_1_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_3_10
#define init_0_1_3_10
void FASTCALL call_0_1_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_3_11
#define init_0_1_3_11
void FASTCALL call_0_1_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_3_12
#define init_0_1_3_12
void FASTCALL call_0_1_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_3_13
#define init_0_1_3_13
void FASTCALL call_0_1_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_3_14
#define init_0_1_3_14
void FASTCALL call_0_1_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_3_15
#define init_0_1_3_15
void FASTCALL call_0_1_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_4_0
#define init_0_1_4_0
void FASTCALL call_0_1_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_4_1
#define init_0_1_4_1
void FASTCALL call_0_1_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_4_2
#define init_0_1_4_2
void FASTCALL call_0_1_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_4_3
#define init_0_1_4_3
void FASTCALL call_0_1_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_4_4
#define init_0_1_4_4
void FASTCALL call_0_1_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_4_5
#define init_0_1_4_5
void FASTCALL call_0_1_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_4_6
#define init_0_1_4_6
void FASTCALL call_0_1_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_4_7
#define init_0_1_4_7
void FASTCALL call_0_1_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_4_8
#define init_0_1_4_8
void FASTCALL call_0_1_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_4_9
#define init_0_1_4_9
void FASTCALL call_0_1_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_4_10
#define init_0_1_4_10
void FASTCALL call_0_1_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_4_11
#define init_0_1_4_11
void FASTCALL call_0_1_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_4_12
#define init_0_1_4_12
void FASTCALL call_0_1_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_4_13
#define init_0_1_4_13
void FASTCALL call_0_1_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_4_14
#define init_0_1_4_14
void FASTCALL call_0_1_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_4_15
#define init_0_1_4_15
void FASTCALL call_0_1_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_5_0
#define init_0_1_5_0
void FASTCALL call_0_1_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_5_1
#define init_0_1_5_1
void FASTCALL call_0_1_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_5_2
#define init_0_1_5_2
void FASTCALL call_0_1_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_5_3
#define init_0_1_5_3
void FASTCALL call_0_1_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_5_4
#define init_0_1_5_4
void FASTCALL call_0_1_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_5_5
#define init_0_1_5_5
void FASTCALL call_0_1_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_5_6
#define init_0_1_5_6
void FASTCALL call_0_1_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_5_7
#define init_0_1_5_7
void FASTCALL call_0_1_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_5_8
#define init_0_1_5_8
void FASTCALL call_0_1_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_5_9
#define init_0_1_5_9
void FASTCALL call_0_1_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_5_10
#define init_0_1_5_10
void FASTCALL call_0_1_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_5_11
#define init_0_1_5_11
void FASTCALL call_0_1_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_5_12
#define init_0_1_5_12
void FASTCALL call_0_1_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_5_13
#define init_0_1_5_13
void FASTCALL call_0_1_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_5_14
#define init_0_1_5_14
void FASTCALL call_0_1_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_5_15
#define init_0_1_5_15
void FASTCALL call_0_1_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_6_0
#define init_0_1_6_0
void FASTCALL call_0_1_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_6_1
#define init_0_1_6_1
void FASTCALL call_0_1_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_6_2
#define init_0_1_6_2
void FASTCALL call_0_1_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_6_3
#define init_0_1_6_3
void FASTCALL call_0_1_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_6_4
#define init_0_1_6_4
void FASTCALL call_0_1_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_6_5
#define init_0_1_6_5
void FASTCALL call_0_1_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_6_6
#define init_0_1_6_6
void FASTCALL call_0_1_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_6_7
#define init_0_1_6_7
void FASTCALL call_0_1_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_6_8
#define init_0_1_6_8
void FASTCALL call_0_1_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_6_9
#define init_0_1_6_9
void FASTCALL call_0_1_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_6_10
#define init_0_1_6_10
void FASTCALL call_0_1_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_6_11
#define init_0_1_6_11
void FASTCALL call_0_1_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_6_12
#define init_0_1_6_12
void FASTCALL call_0_1_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_6_13
#define init_0_1_6_13
void FASTCALL call_0_1_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_6_14
#define init_0_1_6_14
void FASTCALL call_0_1_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_6_15
#define init_0_1_6_15
void FASTCALL call_0_1_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_7_0
#define init_0_1_7_0
void FASTCALL call_0_1_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_7_1
#define init_0_1_7_1
void FASTCALL call_0_1_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_7_2
#define init_0_1_7_2
void FASTCALL call_0_1_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_7_3
#define init_0_1_7_3
void FASTCALL call_0_1_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_7_4
#define init_0_1_7_4
void FASTCALL call_0_1_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_7_5
#define init_0_1_7_5
void FASTCALL call_0_1_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_7_6
#define init_0_1_7_6
void FASTCALL call_0_1_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_7_7
#define init_0_1_7_7
void FASTCALL call_0_1_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_7_8
#define init_0_1_7_8
void FASTCALL call_0_1_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_7_9
#define init_0_1_7_9
void FASTCALL call_0_1_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_7_10
#define init_0_1_7_10
void FASTCALL call_0_1_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_7_11
#define init_0_1_7_11
void FASTCALL call_0_1_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_7_12
#define init_0_1_7_12
void FASTCALL call_0_1_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_7_13
#define init_0_1_7_13
void FASTCALL call_0_1_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_7_14
#define init_0_1_7_14
void FASTCALL call_0_1_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_7_15
#define init_0_1_7_15
void FASTCALL call_0_1_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_8_0
#define init_0_1_8_0
void FASTCALL call_0_1_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_8_1
#define init_0_1_8_1
void FASTCALL call_0_1_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_8_2
#define init_0_1_8_2
void FASTCALL call_0_1_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_8_3
#define init_0_1_8_3
void FASTCALL call_0_1_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_8_4
#define init_0_1_8_4
void FASTCALL call_0_1_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_8_5
#define init_0_1_8_5
void FASTCALL call_0_1_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_8_6
#define init_0_1_8_6
void FASTCALL call_0_1_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_8_7
#define init_0_1_8_7
void FASTCALL call_0_1_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_8_8
#define init_0_1_8_8
void FASTCALL call_0_1_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_8_9
#define init_0_1_8_9
void FASTCALL call_0_1_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_8_10
#define init_0_1_8_10
void FASTCALL call_0_1_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_8_11
#define init_0_1_8_11
void FASTCALL call_0_1_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_8_12
#define init_0_1_8_12
void FASTCALL call_0_1_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_8_13
#define init_0_1_8_13
void FASTCALL call_0_1_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_8_14
#define init_0_1_8_14
void FASTCALL call_0_1_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_8_15
#define init_0_1_8_15
void FASTCALL call_0_1_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_9_0
#define init_0_1_9_0
void FASTCALL call_0_1_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_9_1
#define init_0_1_9_1
void FASTCALL call_0_1_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_9_2
#define init_0_1_9_2
void FASTCALL call_0_1_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_9_3
#define init_0_1_9_3
void FASTCALL call_0_1_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_9_4
#define init_0_1_9_4
void FASTCALL call_0_1_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_9_5
#define init_0_1_9_5
void FASTCALL call_0_1_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_9_6
#define init_0_1_9_6
void FASTCALL call_0_1_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_9_7
#define init_0_1_9_7
void FASTCALL call_0_1_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_9_8
#define init_0_1_9_8
void FASTCALL call_0_1_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_9_9
#define init_0_1_9_9
void FASTCALL call_0_1_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_9_10
#define init_0_1_9_10
void FASTCALL call_0_1_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_9_11
#define init_0_1_9_11
void FASTCALL call_0_1_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_9_12
#define init_0_1_9_12
void FASTCALL call_0_1_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_9_13
#define init_0_1_9_13
void FASTCALL call_0_1_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_9_14
#define init_0_1_9_14
void FASTCALL call_0_1_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_9_15
#define init_0_1_9_15
void FASTCALL call_0_1_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_10_0
#define init_0_1_10_0
void FASTCALL call_0_1_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_10_1
#define init_0_1_10_1
void FASTCALL call_0_1_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_10_2
#define init_0_1_10_2
void FASTCALL call_0_1_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_10_3
#define init_0_1_10_3
void FASTCALL call_0_1_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_10_4
#define init_0_1_10_4
void FASTCALL call_0_1_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_10_5
#define init_0_1_10_5
void FASTCALL call_0_1_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_10_6
#define init_0_1_10_6
void FASTCALL call_0_1_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_10_7
#define init_0_1_10_7
void FASTCALL call_0_1_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_10_8
#define init_0_1_10_8
void FASTCALL call_0_1_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_10_9
#define init_0_1_10_9
void FASTCALL call_0_1_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_10_10
#define init_0_1_10_10
void FASTCALL call_0_1_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_10_11
#define init_0_1_10_11
void FASTCALL call_0_1_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_10_12
#define init_0_1_10_12
void FASTCALL call_0_1_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_10_13
#define init_0_1_10_13
void FASTCALL call_0_1_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_10_14
#define init_0_1_10_14
void FASTCALL call_0_1_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_10_15
#define init_0_1_10_15
void FASTCALL call_0_1_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_11_0
#define init_0_1_11_0
void FASTCALL call_0_1_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_11_1
#define init_0_1_11_1
void FASTCALL call_0_1_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_11_2
#define init_0_1_11_2
void FASTCALL call_0_1_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_11_3
#define init_0_1_11_3
void FASTCALL call_0_1_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_11_4
#define init_0_1_11_4
void FASTCALL call_0_1_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_11_5
#define init_0_1_11_5
void FASTCALL call_0_1_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_11_6
#define init_0_1_11_6
void FASTCALL call_0_1_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_11_7
#define init_0_1_11_7
void FASTCALL call_0_1_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_11_8
#define init_0_1_11_8
void FASTCALL call_0_1_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_11_9
#define init_0_1_11_9
void FASTCALL call_0_1_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_11_10
#define init_0_1_11_10
void FASTCALL call_0_1_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_11_11
#define init_0_1_11_11
void FASTCALL call_0_1_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_11_12
#define init_0_1_11_12
void FASTCALL call_0_1_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_11_13
#define init_0_1_11_13
void FASTCALL call_0_1_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_11_14
#define init_0_1_11_14
void FASTCALL call_0_1_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_11_15
#define init_0_1_11_15
void FASTCALL call_0_1_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_12_0
#define init_0_1_12_0
void FASTCALL call_0_1_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_12_1
#define init_0_1_12_1
void FASTCALL call_0_1_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_12_2
#define init_0_1_12_2
void FASTCALL call_0_1_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_12_3
#define init_0_1_12_3
void FASTCALL call_0_1_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_12_4
#define init_0_1_12_4
void FASTCALL call_0_1_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_12_5
#define init_0_1_12_5
void FASTCALL call_0_1_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_12_6
#define init_0_1_12_6
void FASTCALL call_0_1_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_12_7
#define init_0_1_12_7
void FASTCALL call_0_1_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_12_8
#define init_0_1_12_8
void FASTCALL call_0_1_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_12_9
#define init_0_1_12_9
void FASTCALL call_0_1_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_12_10
#define init_0_1_12_10
void FASTCALL call_0_1_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_12_11
#define init_0_1_12_11
void FASTCALL call_0_1_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_12_12
#define init_0_1_12_12
void FASTCALL call_0_1_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_12_13
#define init_0_1_12_13
void FASTCALL call_0_1_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_12_14
#define init_0_1_12_14
void FASTCALL call_0_1_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_12_15
#define init_0_1_12_15
void FASTCALL call_0_1_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_13_0
#define init_0_1_13_0
void FASTCALL call_0_1_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_13_1
#define init_0_1_13_1
void FASTCALL call_0_1_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_13_2
#define init_0_1_13_2
void FASTCALL call_0_1_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_13_3
#define init_0_1_13_3
void FASTCALL call_0_1_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_13_4
#define init_0_1_13_4
void FASTCALL call_0_1_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_13_5
#define init_0_1_13_5
void FASTCALL call_0_1_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_13_6
#define init_0_1_13_6
void FASTCALL call_0_1_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_13_7
#define init_0_1_13_7
void FASTCALL call_0_1_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_13_8
#define init_0_1_13_8
void FASTCALL call_0_1_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_13_9
#define init_0_1_13_9
void FASTCALL call_0_1_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_13_10
#define init_0_1_13_10
void FASTCALL call_0_1_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_13_11
#define init_0_1_13_11
void FASTCALL call_0_1_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_13_12
#define init_0_1_13_12
void FASTCALL call_0_1_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_13_13
#define init_0_1_13_13
void FASTCALL call_0_1_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_13_14
#define init_0_1_13_14
void FASTCALL call_0_1_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_13_15
#define init_0_1_13_15
void FASTCALL call_0_1_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_14_0
#define init_0_1_14_0
void FASTCALL call_0_1_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_14_1
#define init_0_1_14_1
void FASTCALL call_0_1_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_14_2
#define init_0_1_14_2
void FASTCALL call_0_1_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_14_3
#define init_0_1_14_3
void FASTCALL call_0_1_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_14_4
#define init_0_1_14_4
void FASTCALL call_0_1_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_14_5
#define init_0_1_14_5
void FASTCALL call_0_1_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_14_6
#define init_0_1_14_6
void FASTCALL call_0_1_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_14_7
#define init_0_1_14_7
void FASTCALL call_0_1_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_14_8
#define init_0_1_14_8
void FASTCALL call_0_1_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_14_9
#define init_0_1_14_9
void FASTCALL call_0_1_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_14_10
#define init_0_1_14_10
void FASTCALL call_0_1_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_14_11
#define init_0_1_14_11
void FASTCALL call_0_1_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_14_12
#define init_0_1_14_12
void FASTCALL call_0_1_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_14_13
#define init_0_1_14_13
void FASTCALL call_0_1_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_14_14
#define init_0_1_14_14
void FASTCALL call_0_1_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_14_15
#define init_0_1_14_15
void FASTCALL call_0_1_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_15_0
#define init_0_1_15_0
void FASTCALL call_0_1_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_15_1
#define init_0_1_15_1
void FASTCALL call_0_1_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_15_2
#define init_0_1_15_2
void FASTCALL call_0_1_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_15_3
#define init_0_1_15_3
void FASTCALL call_0_1_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_15_4
#define init_0_1_15_4
void FASTCALL call_0_1_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_15_5
#define init_0_1_15_5
void FASTCALL call_0_1_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_15_6
#define init_0_1_15_6
void FASTCALL call_0_1_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_15_7
#define init_0_1_15_7
void FASTCALL call_0_1_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_15_8
#define init_0_1_15_8
void FASTCALL call_0_1_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_15_9
#define init_0_1_15_9
void FASTCALL call_0_1_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_15_10
#define init_0_1_15_10
void FASTCALL call_0_1_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_15_11
#define init_0_1_15_11
void FASTCALL call_0_1_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_15_12
#define init_0_1_15_12
void FASTCALL call_0_1_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_15_13
#define init_0_1_15_13
void FASTCALL call_0_1_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_15_14
#define init_0_1_15_14
void FASTCALL call_0_1_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_1_15_15
#define init_0_1_15_15
void FASTCALL call_0_1_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_0_0
#define init_0_2_0_0
void FASTCALL call_0_2_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_0_1
#define init_0_2_0_1
void FASTCALL call_0_2_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_0_2
#define init_0_2_0_2
void FASTCALL call_0_2_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_0_3
#define init_0_2_0_3
void FASTCALL call_0_2_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_0_4
#define init_0_2_0_4
void FASTCALL call_0_2_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_0_5
#define init_0_2_0_5
void FASTCALL call_0_2_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_0_6
#define init_0_2_0_6
void FASTCALL call_0_2_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_0_7
#define init_0_2_0_7
void FASTCALL call_0_2_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_0_8
#define init_0_2_0_8
void FASTCALL call_0_2_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_0_9
#define init_0_2_0_9
void FASTCALL call_0_2_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_0_10
#define init_0_2_0_10
void FASTCALL call_0_2_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_0_11
#define init_0_2_0_11
void FASTCALL call_0_2_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_0_12
#define init_0_2_0_12
void FASTCALL call_0_2_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_0_13
#define init_0_2_0_13
void FASTCALL call_0_2_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_0_14
#define init_0_2_0_14
void FASTCALL call_0_2_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_0_15
#define init_0_2_0_15
void FASTCALL call_0_2_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_1_0
#define init_0_2_1_0
void FASTCALL call_0_2_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_1_1
#define init_0_2_1_1
void FASTCALL call_0_2_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_1_2
#define init_0_2_1_2
void FASTCALL call_0_2_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_1_3
#define init_0_2_1_3
void FASTCALL call_0_2_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_1_4
#define init_0_2_1_4
void FASTCALL call_0_2_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_1_5
#define init_0_2_1_5
void FASTCALL call_0_2_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_1_6
#define init_0_2_1_6
void FASTCALL call_0_2_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_1_7
#define init_0_2_1_7
void FASTCALL call_0_2_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_1_8
#define init_0_2_1_8
void FASTCALL call_0_2_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_1_9
#define init_0_2_1_9
void FASTCALL call_0_2_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_1_10
#define init_0_2_1_10
void FASTCALL call_0_2_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_1_11
#define init_0_2_1_11
void FASTCALL call_0_2_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_1_12
#define init_0_2_1_12
void FASTCALL call_0_2_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_1_13
#define init_0_2_1_13
void FASTCALL call_0_2_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_1_14
#define init_0_2_1_14
void FASTCALL call_0_2_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_1_15
#define init_0_2_1_15
void FASTCALL call_0_2_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_2_0
#define init_0_2_2_0
void FASTCALL call_0_2_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_2_1
#define init_0_2_2_1
void FASTCALL call_0_2_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_2_2
#define init_0_2_2_2
void FASTCALL call_0_2_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_2_3
#define init_0_2_2_3
void FASTCALL call_0_2_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_2_4
#define init_0_2_2_4
void FASTCALL call_0_2_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_2_5
#define init_0_2_2_5
void FASTCALL call_0_2_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_2_6
#define init_0_2_2_6
void FASTCALL call_0_2_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_2_7
#define init_0_2_2_7
void FASTCALL call_0_2_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_2_8
#define init_0_2_2_8
void FASTCALL call_0_2_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_2_9
#define init_0_2_2_9
void FASTCALL call_0_2_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_2_10
#define init_0_2_2_10
void FASTCALL call_0_2_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_2_11
#define init_0_2_2_11
void FASTCALL call_0_2_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_2_12
#define init_0_2_2_12
void FASTCALL call_0_2_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_2_13
#define init_0_2_2_13
void FASTCALL call_0_2_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_2_14
#define init_0_2_2_14
void FASTCALL call_0_2_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_2_15
#define init_0_2_2_15
void FASTCALL call_0_2_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_3_0
#define init_0_2_3_0
void FASTCALL call_0_2_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_3_1
#define init_0_2_3_1
void FASTCALL call_0_2_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_3_2
#define init_0_2_3_2
void FASTCALL call_0_2_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_3_3
#define init_0_2_3_3
void FASTCALL call_0_2_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_3_4
#define init_0_2_3_4
void FASTCALL call_0_2_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_3_5
#define init_0_2_3_5
void FASTCALL call_0_2_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_3_6
#define init_0_2_3_6
void FASTCALL call_0_2_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_3_7
#define init_0_2_3_7
void FASTCALL call_0_2_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_3_8
#define init_0_2_3_8
void FASTCALL call_0_2_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_3_9
#define init_0_2_3_9
void FASTCALL call_0_2_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_3_10
#define init_0_2_3_10
void FASTCALL call_0_2_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_3_11
#define init_0_2_3_11
void FASTCALL call_0_2_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_3_12
#define init_0_2_3_12
void FASTCALL call_0_2_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_3_13
#define init_0_2_3_13
void FASTCALL call_0_2_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_3_14
#define init_0_2_3_14
void FASTCALL call_0_2_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_3_15
#define init_0_2_3_15
void FASTCALL call_0_2_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_4_0
#define init_0_2_4_0
void FASTCALL call_0_2_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_4_1
#define init_0_2_4_1
void FASTCALL call_0_2_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_4_2
#define init_0_2_4_2
void FASTCALL call_0_2_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_4_3
#define init_0_2_4_3
void FASTCALL call_0_2_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_4_4
#define init_0_2_4_4
void FASTCALL call_0_2_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_4_5
#define init_0_2_4_5
void FASTCALL call_0_2_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_4_6
#define init_0_2_4_6
void FASTCALL call_0_2_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_4_7
#define init_0_2_4_7
void FASTCALL call_0_2_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_4_8
#define init_0_2_4_8
void FASTCALL call_0_2_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_4_9
#define init_0_2_4_9
void FASTCALL call_0_2_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_4_10
#define init_0_2_4_10
void FASTCALL call_0_2_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_4_11
#define init_0_2_4_11
void FASTCALL call_0_2_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_4_12
#define init_0_2_4_12
void FASTCALL call_0_2_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_4_13
#define init_0_2_4_13
void FASTCALL call_0_2_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_4_14
#define init_0_2_4_14
void FASTCALL call_0_2_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_4_15
#define init_0_2_4_15
void FASTCALL call_0_2_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_5_0
#define init_0_2_5_0
void FASTCALL call_0_2_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_5_1
#define init_0_2_5_1
void FASTCALL call_0_2_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_5_2
#define init_0_2_5_2
void FASTCALL call_0_2_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_5_3
#define init_0_2_5_3
void FASTCALL call_0_2_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_5_4
#define init_0_2_5_4
void FASTCALL call_0_2_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_5_5
#define init_0_2_5_5
void FASTCALL call_0_2_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_5_6
#define init_0_2_5_6
void FASTCALL call_0_2_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_5_7
#define init_0_2_5_7
void FASTCALL call_0_2_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_5_8
#define init_0_2_5_8
void FASTCALL call_0_2_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_5_9
#define init_0_2_5_9
void FASTCALL call_0_2_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_5_10
#define init_0_2_5_10
void FASTCALL call_0_2_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_5_11
#define init_0_2_5_11
void FASTCALL call_0_2_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_5_12
#define init_0_2_5_12
void FASTCALL call_0_2_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_5_13
#define init_0_2_5_13
void FASTCALL call_0_2_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_5_14
#define init_0_2_5_14
void FASTCALL call_0_2_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_5_15
#define init_0_2_5_15
void FASTCALL call_0_2_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_6_0
#define init_0_2_6_0
void FASTCALL call_0_2_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_6_1
#define init_0_2_6_1
void FASTCALL call_0_2_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_6_2
#define init_0_2_6_2
void FASTCALL call_0_2_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_6_3
#define init_0_2_6_3
void FASTCALL call_0_2_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_6_4
#define init_0_2_6_4
void FASTCALL call_0_2_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_6_5
#define init_0_2_6_5
void FASTCALL call_0_2_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_6_6
#define init_0_2_6_6
void FASTCALL call_0_2_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_6_7
#define init_0_2_6_7
void FASTCALL call_0_2_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_6_8
#define init_0_2_6_8
void FASTCALL call_0_2_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_6_9
#define init_0_2_6_9
void FASTCALL call_0_2_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_6_10
#define init_0_2_6_10
void FASTCALL call_0_2_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_6_11
#define init_0_2_6_11
void FASTCALL call_0_2_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_6_12
#define init_0_2_6_12
void FASTCALL call_0_2_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_6_13
#define init_0_2_6_13
void FASTCALL call_0_2_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_6_14
#define init_0_2_6_14
void FASTCALL call_0_2_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_6_15
#define init_0_2_6_15
void FASTCALL call_0_2_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_7_0
#define init_0_2_7_0
void FASTCALL call_0_2_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_7_1
#define init_0_2_7_1
void FASTCALL call_0_2_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_7_2
#define init_0_2_7_2
void FASTCALL call_0_2_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_7_3
#define init_0_2_7_3
void FASTCALL call_0_2_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_7_4
#define init_0_2_7_4
void FASTCALL call_0_2_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_7_5
#define init_0_2_7_5
void FASTCALL call_0_2_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_7_6
#define init_0_2_7_6
void FASTCALL call_0_2_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_7_7
#define init_0_2_7_7
void FASTCALL call_0_2_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_7_8
#define init_0_2_7_8
void FASTCALL call_0_2_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_7_9
#define init_0_2_7_9
void FASTCALL call_0_2_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_7_10
#define init_0_2_7_10
void FASTCALL call_0_2_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_7_11
#define init_0_2_7_11
void FASTCALL call_0_2_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_7_12
#define init_0_2_7_12
void FASTCALL call_0_2_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_7_13
#define init_0_2_7_13
void FASTCALL call_0_2_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_7_14
#define init_0_2_7_14
void FASTCALL call_0_2_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_7_15
#define init_0_2_7_15
void FASTCALL call_0_2_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_8_0
#define init_0_2_8_0
void FASTCALL call_0_2_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_8_1
#define init_0_2_8_1
void FASTCALL call_0_2_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_8_2
#define init_0_2_8_2
void FASTCALL call_0_2_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_8_3
#define init_0_2_8_3
void FASTCALL call_0_2_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_8_4
#define init_0_2_8_4
void FASTCALL call_0_2_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_8_5
#define init_0_2_8_5
void FASTCALL call_0_2_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_8_6
#define init_0_2_8_6
void FASTCALL call_0_2_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_8_7
#define init_0_2_8_7
void FASTCALL call_0_2_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_8_8
#define init_0_2_8_8
void FASTCALL call_0_2_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_8_9
#define init_0_2_8_9
void FASTCALL call_0_2_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_8_10
#define init_0_2_8_10
void FASTCALL call_0_2_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_8_11
#define init_0_2_8_11
void FASTCALL call_0_2_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_8_12
#define init_0_2_8_12
void FASTCALL call_0_2_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_8_13
#define init_0_2_8_13
void FASTCALL call_0_2_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_8_14
#define init_0_2_8_14
void FASTCALL call_0_2_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_8_15
#define init_0_2_8_15
void FASTCALL call_0_2_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_9_0
#define init_0_2_9_0
void FASTCALL call_0_2_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_9_1
#define init_0_2_9_1
void FASTCALL call_0_2_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_9_2
#define init_0_2_9_2
void FASTCALL call_0_2_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_9_3
#define init_0_2_9_3
void FASTCALL call_0_2_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_9_4
#define init_0_2_9_4
void FASTCALL call_0_2_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_9_5
#define init_0_2_9_5
void FASTCALL call_0_2_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_9_6
#define init_0_2_9_6
void FASTCALL call_0_2_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_9_7
#define init_0_2_9_7
void FASTCALL call_0_2_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_9_8
#define init_0_2_9_8
void FASTCALL call_0_2_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_9_9
#define init_0_2_9_9
void FASTCALL call_0_2_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_9_10
#define init_0_2_9_10
void FASTCALL call_0_2_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_9_11
#define init_0_2_9_11
void FASTCALL call_0_2_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_9_12
#define init_0_2_9_12
void FASTCALL call_0_2_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_9_13
#define init_0_2_9_13
void FASTCALL call_0_2_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_9_14
#define init_0_2_9_14
void FASTCALL call_0_2_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_9_15
#define init_0_2_9_15
void FASTCALL call_0_2_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_10_0
#define init_0_2_10_0
void FASTCALL call_0_2_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_10_1
#define init_0_2_10_1
void FASTCALL call_0_2_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_10_2
#define init_0_2_10_2
void FASTCALL call_0_2_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_10_3
#define init_0_2_10_3
void FASTCALL call_0_2_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_10_4
#define init_0_2_10_4
void FASTCALL call_0_2_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_10_5
#define init_0_2_10_5
void FASTCALL call_0_2_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_10_6
#define init_0_2_10_6
void FASTCALL call_0_2_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_10_7
#define init_0_2_10_7
void FASTCALL call_0_2_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_10_8
#define init_0_2_10_8
void FASTCALL call_0_2_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_10_9
#define init_0_2_10_9
void FASTCALL call_0_2_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_10_10
#define init_0_2_10_10
void FASTCALL call_0_2_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_10_11
#define init_0_2_10_11
void FASTCALL call_0_2_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_10_12
#define init_0_2_10_12
void FASTCALL call_0_2_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_10_13
#define init_0_2_10_13
void FASTCALL call_0_2_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_10_14
#define init_0_2_10_14
void FASTCALL call_0_2_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_10_15
#define init_0_2_10_15
void FASTCALL call_0_2_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_11_0
#define init_0_2_11_0
void FASTCALL call_0_2_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_11_1
#define init_0_2_11_1
void FASTCALL call_0_2_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_11_2
#define init_0_2_11_2
void FASTCALL call_0_2_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_11_3
#define init_0_2_11_3
void FASTCALL call_0_2_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_11_4
#define init_0_2_11_4
void FASTCALL call_0_2_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_11_5
#define init_0_2_11_5
void FASTCALL call_0_2_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_11_6
#define init_0_2_11_6
void FASTCALL call_0_2_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_11_7
#define init_0_2_11_7
void FASTCALL call_0_2_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_11_8
#define init_0_2_11_8
void FASTCALL call_0_2_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_11_9
#define init_0_2_11_9
void FASTCALL call_0_2_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_11_10
#define init_0_2_11_10
void FASTCALL call_0_2_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_11_11
#define init_0_2_11_11
void FASTCALL call_0_2_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_11_12
#define init_0_2_11_12
void FASTCALL call_0_2_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_11_13
#define init_0_2_11_13
void FASTCALL call_0_2_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_11_14
#define init_0_2_11_14
void FASTCALL call_0_2_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_11_15
#define init_0_2_11_15
void FASTCALL call_0_2_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_12_0
#define init_0_2_12_0
void FASTCALL call_0_2_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_12_1
#define init_0_2_12_1
void FASTCALL call_0_2_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_12_2
#define init_0_2_12_2
void FASTCALL call_0_2_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_12_3
#define init_0_2_12_3
void FASTCALL call_0_2_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_12_4
#define init_0_2_12_4
void FASTCALL call_0_2_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_12_5
#define init_0_2_12_5
void FASTCALL call_0_2_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_12_6
#define init_0_2_12_6
void FASTCALL call_0_2_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_12_7
#define init_0_2_12_7
void FASTCALL call_0_2_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_12_8
#define init_0_2_12_8
void FASTCALL call_0_2_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_12_9
#define init_0_2_12_9
void FASTCALL call_0_2_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_12_10
#define init_0_2_12_10
void FASTCALL call_0_2_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_12_11
#define init_0_2_12_11
void FASTCALL call_0_2_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_12_12
#define init_0_2_12_12
void FASTCALL call_0_2_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_12_13
#define init_0_2_12_13
void FASTCALL call_0_2_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_12_14
#define init_0_2_12_14
void FASTCALL call_0_2_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_12_15
#define init_0_2_12_15
void FASTCALL call_0_2_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_13_0
#define init_0_2_13_0
void FASTCALL call_0_2_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_13_1
#define init_0_2_13_1
void FASTCALL call_0_2_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_13_2
#define init_0_2_13_2
void FASTCALL call_0_2_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_13_3
#define init_0_2_13_3
void FASTCALL call_0_2_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_13_4
#define init_0_2_13_4
void FASTCALL call_0_2_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_13_5
#define init_0_2_13_5
void FASTCALL call_0_2_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_13_6
#define init_0_2_13_6
void FASTCALL call_0_2_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_13_7
#define init_0_2_13_7
void FASTCALL call_0_2_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_13_8
#define init_0_2_13_8
void FASTCALL call_0_2_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_13_9
#define init_0_2_13_9
void FASTCALL call_0_2_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_13_10
#define init_0_2_13_10
void FASTCALL call_0_2_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_13_11
#define init_0_2_13_11
void FASTCALL call_0_2_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_13_12
#define init_0_2_13_12
void FASTCALL call_0_2_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_13_13
#define init_0_2_13_13
void FASTCALL call_0_2_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_13_14
#define init_0_2_13_14
void FASTCALL call_0_2_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_13_15
#define init_0_2_13_15
void FASTCALL call_0_2_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_14_0
#define init_0_2_14_0
void FASTCALL call_0_2_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_14_1
#define init_0_2_14_1
void FASTCALL call_0_2_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_14_2
#define init_0_2_14_2
void FASTCALL call_0_2_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_14_3
#define init_0_2_14_3
void FASTCALL call_0_2_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_14_4
#define init_0_2_14_4
void FASTCALL call_0_2_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_14_5
#define init_0_2_14_5
void FASTCALL call_0_2_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_14_6
#define init_0_2_14_6
void FASTCALL call_0_2_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_14_7
#define init_0_2_14_7
void FASTCALL call_0_2_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_14_8
#define init_0_2_14_8
void FASTCALL call_0_2_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_14_9
#define init_0_2_14_9
void FASTCALL call_0_2_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_14_10
#define init_0_2_14_10
void FASTCALL call_0_2_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_14_11
#define init_0_2_14_11
void FASTCALL call_0_2_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_14_12
#define init_0_2_14_12
void FASTCALL call_0_2_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_14_13
#define init_0_2_14_13
void FASTCALL call_0_2_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_14_14
#define init_0_2_14_14
void FASTCALL call_0_2_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_14_15
#define init_0_2_14_15
void FASTCALL call_0_2_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_15_0
#define init_0_2_15_0
void FASTCALL call_0_2_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_15_1
#define init_0_2_15_1
void FASTCALL call_0_2_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_15_2
#define init_0_2_15_2
void FASTCALL call_0_2_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_15_3
#define init_0_2_15_3
void FASTCALL call_0_2_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_15_4
#define init_0_2_15_4
void FASTCALL call_0_2_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_15_5
#define init_0_2_15_5
void FASTCALL call_0_2_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_15_6
#define init_0_2_15_6
void FASTCALL call_0_2_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_15_7
#define init_0_2_15_7
void FASTCALL call_0_2_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_15_8
#define init_0_2_15_8
void FASTCALL call_0_2_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_15_9
#define init_0_2_15_9
void FASTCALL call_0_2_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_15_10
#define init_0_2_15_10
void FASTCALL call_0_2_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_15_11
#define init_0_2_15_11
void FASTCALL call_0_2_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_15_12
#define init_0_2_15_12
void FASTCALL call_0_2_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_15_13
#define init_0_2_15_13
void FASTCALL call_0_2_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_15_14
#define init_0_2_15_14
void FASTCALL call_0_2_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_2_15_15
#define init_0_2_15_15
void FASTCALL call_0_2_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_0_0
#define init_0_3_0_0
void FASTCALL call_0_3_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_0_1
#define init_0_3_0_1
void FASTCALL call_0_3_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_0_2
#define init_0_3_0_2
void FASTCALL call_0_3_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_0_3
#define init_0_3_0_3
void FASTCALL call_0_3_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_0_4
#define init_0_3_0_4
void FASTCALL call_0_3_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_0_5
#define init_0_3_0_5
void FASTCALL call_0_3_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_0_6
#define init_0_3_0_6
void FASTCALL call_0_3_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_0_7
#define init_0_3_0_7
void FASTCALL call_0_3_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_0_8
#define init_0_3_0_8
void FASTCALL call_0_3_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_0_9
#define init_0_3_0_9
void FASTCALL call_0_3_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_0_10
#define init_0_3_0_10
void FASTCALL call_0_3_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_0_11
#define init_0_3_0_11
void FASTCALL call_0_3_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_0_12
#define init_0_3_0_12
void FASTCALL call_0_3_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_0_13
#define init_0_3_0_13
void FASTCALL call_0_3_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_0_14
#define init_0_3_0_14
void FASTCALL call_0_3_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_0_15
#define init_0_3_0_15
void FASTCALL call_0_3_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_1_0
#define init_0_3_1_0
void FASTCALL call_0_3_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_1_1
#define init_0_3_1_1
void FASTCALL call_0_3_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_1_2
#define init_0_3_1_2
void FASTCALL call_0_3_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_1_3
#define init_0_3_1_3
void FASTCALL call_0_3_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_1_4
#define init_0_3_1_4
void FASTCALL call_0_3_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_1_5
#define init_0_3_1_5
void FASTCALL call_0_3_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_1_6
#define init_0_3_1_6
void FASTCALL call_0_3_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_1_7
#define init_0_3_1_7
void FASTCALL call_0_3_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_1_8
#define init_0_3_1_8
void FASTCALL call_0_3_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_1_9
#define init_0_3_1_9
void FASTCALL call_0_3_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_1_10
#define init_0_3_1_10
void FASTCALL call_0_3_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_1_11
#define init_0_3_1_11
void FASTCALL call_0_3_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_1_12
#define init_0_3_1_12
void FASTCALL call_0_3_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_1_13
#define init_0_3_1_13
void FASTCALL call_0_3_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_1_14
#define init_0_3_1_14
void FASTCALL call_0_3_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_1_15
#define init_0_3_1_15
void FASTCALL call_0_3_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_2_0
#define init_0_3_2_0
void FASTCALL call_0_3_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_2_1
#define init_0_3_2_1
void FASTCALL call_0_3_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_2_2
#define init_0_3_2_2
void FASTCALL call_0_3_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_2_3
#define init_0_3_2_3
void FASTCALL call_0_3_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_2_4
#define init_0_3_2_4
void FASTCALL call_0_3_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_2_5
#define init_0_3_2_5
void FASTCALL call_0_3_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_2_6
#define init_0_3_2_6
void FASTCALL call_0_3_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_2_7
#define init_0_3_2_7
void FASTCALL call_0_3_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_2_8
#define init_0_3_2_8
void FASTCALL call_0_3_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_2_9
#define init_0_3_2_9
void FASTCALL call_0_3_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_2_10
#define init_0_3_2_10
void FASTCALL call_0_3_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_2_11
#define init_0_3_2_11
void FASTCALL call_0_3_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_2_12
#define init_0_3_2_12
void FASTCALL call_0_3_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_2_13
#define init_0_3_2_13
void FASTCALL call_0_3_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_2_14
#define init_0_3_2_14
void FASTCALL call_0_3_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_2_15
#define init_0_3_2_15
void FASTCALL call_0_3_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_3_0
#define init_0_3_3_0
void FASTCALL call_0_3_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_3_1
#define init_0_3_3_1
void FASTCALL call_0_3_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_3_2
#define init_0_3_3_2
void FASTCALL call_0_3_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_3_3
#define init_0_3_3_3
void FASTCALL call_0_3_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_3_4
#define init_0_3_3_4
void FASTCALL call_0_3_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_3_5
#define init_0_3_3_5
void FASTCALL call_0_3_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_3_6
#define init_0_3_3_6
void FASTCALL call_0_3_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_3_7
#define init_0_3_3_7
void FASTCALL call_0_3_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_3_8
#define init_0_3_3_8
void FASTCALL call_0_3_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_3_9
#define init_0_3_3_9
void FASTCALL call_0_3_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_3_10
#define init_0_3_3_10
void FASTCALL call_0_3_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_3_11
#define init_0_3_3_11
void FASTCALL call_0_3_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_3_12
#define init_0_3_3_12
void FASTCALL call_0_3_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_3_13
#define init_0_3_3_13
void FASTCALL call_0_3_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_3_14
#define init_0_3_3_14
void FASTCALL call_0_3_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_3_15
#define init_0_3_3_15
void FASTCALL call_0_3_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_4_0
#define init_0_3_4_0
void FASTCALL call_0_3_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_4_1
#define init_0_3_4_1
void FASTCALL call_0_3_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_4_2
#define init_0_3_4_2
void FASTCALL call_0_3_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_4_3
#define init_0_3_4_3
void FASTCALL call_0_3_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_4_4
#define init_0_3_4_4
void FASTCALL call_0_3_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_4_5
#define init_0_3_4_5
void FASTCALL call_0_3_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_4_6
#define init_0_3_4_6
void FASTCALL call_0_3_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_4_7
#define init_0_3_4_7
void FASTCALL call_0_3_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_4_8
#define init_0_3_4_8
void FASTCALL call_0_3_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_4_9
#define init_0_3_4_9
void FASTCALL call_0_3_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_4_10
#define init_0_3_4_10
void FASTCALL call_0_3_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_4_11
#define init_0_3_4_11
void FASTCALL call_0_3_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_4_12
#define init_0_3_4_12
void FASTCALL call_0_3_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_4_13
#define init_0_3_4_13
void FASTCALL call_0_3_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_4_14
#define init_0_3_4_14
void FASTCALL call_0_3_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_4_15
#define init_0_3_4_15
void FASTCALL call_0_3_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_5_0
#define init_0_3_5_0
void FASTCALL call_0_3_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_5_1
#define init_0_3_5_1
void FASTCALL call_0_3_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_5_2
#define init_0_3_5_2
void FASTCALL call_0_3_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_5_3
#define init_0_3_5_3
void FASTCALL call_0_3_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_5_4
#define init_0_3_5_4
void FASTCALL call_0_3_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_5_5
#define init_0_3_5_5
void FASTCALL call_0_3_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_5_6
#define init_0_3_5_6
void FASTCALL call_0_3_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_5_7
#define init_0_3_5_7
void FASTCALL call_0_3_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_5_8
#define init_0_3_5_8
void FASTCALL call_0_3_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_5_9
#define init_0_3_5_9
void FASTCALL call_0_3_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_5_10
#define init_0_3_5_10
void FASTCALL call_0_3_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_5_11
#define init_0_3_5_11
void FASTCALL call_0_3_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_5_12
#define init_0_3_5_12
void FASTCALL call_0_3_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_5_13
#define init_0_3_5_13
void FASTCALL call_0_3_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_5_14
#define init_0_3_5_14
void FASTCALL call_0_3_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_5_15
#define init_0_3_5_15
void FASTCALL call_0_3_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_6_0
#define init_0_3_6_0
void FASTCALL call_0_3_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_6_1
#define init_0_3_6_1
void FASTCALL call_0_3_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_6_2
#define init_0_3_6_2
void FASTCALL call_0_3_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_6_3
#define init_0_3_6_3
void FASTCALL call_0_3_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_6_4
#define init_0_3_6_4
void FASTCALL call_0_3_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_6_5
#define init_0_3_6_5
void FASTCALL call_0_3_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_6_6
#define init_0_3_6_6
void FASTCALL call_0_3_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_6_7
#define init_0_3_6_7
void FASTCALL call_0_3_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_6_8
#define init_0_3_6_8
void FASTCALL call_0_3_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_6_9
#define init_0_3_6_9
void FASTCALL call_0_3_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_6_10
#define init_0_3_6_10
void FASTCALL call_0_3_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_6_11
#define init_0_3_6_11
void FASTCALL call_0_3_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_6_12
#define init_0_3_6_12
void FASTCALL call_0_3_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_6_13
#define init_0_3_6_13
void FASTCALL call_0_3_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_6_14
#define init_0_3_6_14
void FASTCALL call_0_3_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_6_15
#define init_0_3_6_15
void FASTCALL call_0_3_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_7_0
#define init_0_3_7_0
void FASTCALL call_0_3_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_7_1
#define init_0_3_7_1
void FASTCALL call_0_3_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_7_2
#define init_0_3_7_2
void FASTCALL call_0_3_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_7_3
#define init_0_3_7_3
void FASTCALL call_0_3_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_7_4
#define init_0_3_7_4
void FASTCALL call_0_3_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_7_5
#define init_0_3_7_5
void FASTCALL call_0_3_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_7_6
#define init_0_3_7_6
void FASTCALL call_0_3_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_7_7
#define init_0_3_7_7
void FASTCALL call_0_3_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_7_8
#define init_0_3_7_8
void FASTCALL call_0_3_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_7_9
#define init_0_3_7_9
void FASTCALL call_0_3_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_7_10
#define init_0_3_7_10
void FASTCALL call_0_3_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_7_11
#define init_0_3_7_11
void FASTCALL call_0_3_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_7_12
#define init_0_3_7_12
void FASTCALL call_0_3_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_7_13
#define init_0_3_7_13
void FASTCALL call_0_3_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_7_14
#define init_0_3_7_14
void FASTCALL call_0_3_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_7_15
#define init_0_3_7_15
void FASTCALL call_0_3_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_8_0
#define init_0_3_8_0
void FASTCALL call_0_3_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_8_1
#define init_0_3_8_1
void FASTCALL call_0_3_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_8_2
#define init_0_3_8_2
void FASTCALL call_0_3_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_8_3
#define init_0_3_8_3
void FASTCALL call_0_3_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_8_4
#define init_0_3_8_4
void FASTCALL call_0_3_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_8_5
#define init_0_3_8_5
void FASTCALL call_0_3_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_8_6
#define init_0_3_8_6
void FASTCALL call_0_3_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_8_7
#define init_0_3_8_7
void FASTCALL call_0_3_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_8_8
#define init_0_3_8_8
void FASTCALL call_0_3_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_8_9
#define init_0_3_8_9
void FASTCALL call_0_3_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_8_10
#define init_0_3_8_10
void FASTCALL call_0_3_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_8_11
#define init_0_3_8_11
void FASTCALL call_0_3_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_8_12
#define init_0_3_8_12
void FASTCALL call_0_3_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_8_13
#define init_0_3_8_13
void FASTCALL call_0_3_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_8_14
#define init_0_3_8_14
void FASTCALL call_0_3_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_8_15
#define init_0_3_8_15
void FASTCALL call_0_3_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_9_0
#define init_0_3_9_0
void FASTCALL call_0_3_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_9_1
#define init_0_3_9_1
void FASTCALL call_0_3_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_9_2
#define init_0_3_9_2
void FASTCALL call_0_3_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_9_3
#define init_0_3_9_3
void FASTCALL call_0_3_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_9_4
#define init_0_3_9_4
void FASTCALL call_0_3_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_9_5
#define init_0_3_9_5
void FASTCALL call_0_3_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_9_6
#define init_0_3_9_6
void FASTCALL call_0_3_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_9_7
#define init_0_3_9_7
void FASTCALL call_0_3_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_9_8
#define init_0_3_9_8
void FASTCALL call_0_3_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_9_9
#define init_0_3_9_9
void FASTCALL call_0_3_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_9_10
#define init_0_3_9_10
void FASTCALL call_0_3_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_9_11
#define init_0_3_9_11
void FASTCALL call_0_3_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_9_12
#define init_0_3_9_12
void FASTCALL call_0_3_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_9_13
#define init_0_3_9_13
void FASTCALL call_0_3_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_9_14
#define init_0_3_9_14
void FASTCALL call_0_3_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_9_15
#define init_0_3_9_15
void FASTCALL call_0_3_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_10_0
#define init_0_3_10_0
void FASTCALL call_0_3_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_10_1
#define init_0_3_10_1
void FASTCALL call_0_3_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_10_2
#define init_0_3_10_2
void FASTCALL call_0_3_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_10_3
#define init_0_3_10_3
void FASTCALL call_0_3_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_10_4
#define init_0_3_10_4
void FASTCALL call_0_3_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_10_5
#define init_0_3_10_5
void FASTCALL call_0_3_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_10_6
#define init_0_3_10_6
void FASTCALL call_0_3_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_10_7
#define init_0_3_10_7
void FASTCALL call_0_3_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_10_8
#define init_0_3_10_8
void FASTCALL call_0_3_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_10_9
#define init_0_3_10_9
void FASTCALL call_0_3_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_10_10
#define init_0_3_10_10
void FASTCALL call_0_3_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_10_11
#define init_0_3_10_11
void FASTCALL call_0_3_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_10_12
#define init_0_3_10_12
void FASTCALL call_0_3_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_10_13
#define init_0_3_10_13
void FASTCALL call_0_3_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_10_14
#define init_0_3_10_14
void FASTCALL call_0_3_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_10_15
#define init_0_3_10_15
void FASTCALL call_0_3_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_11_0
#define init_0_3_11_0
void FASTCALL call_0_3_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_11_1
#define init_0_3_11_1
void FASTCALL call_0_3_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_11_2
#define init_0_3_11_2
void FASTCALL call_0_3_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_11_3
#define init_0_3_11_3
void FASTCALL call_0_3_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_11_4
#define init_0_3_11_4
void FASTCALL call_0_3_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_11_5
#define init_0_3_11_5
void FASTCALL call_0_3_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_11_6
#define init_0_3_11_6
void FASTCALL call_0_3_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_11_7
#define init_0_3_11_7
void FASTCALL call_0_3_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_11_8
#define init_0_3_11_8
void FASTCALL call_0_3_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_11_9
#define init_0_3_11_9
void FASTCALL call_0_3_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_11_10
#define init_0_3_11_10
void FASTCALL call_0_3_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_11_11
#define init_0_3_11_11
void FASTCALL call_0_3_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_11_12
#define init_0_3_11_12
void FASTCALL call_0_3_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_11_13
#define init_0_3_11_13
void FASTCALL call_0_3_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_11_14
#define init_0_3_11_14
void FASTCALL call_0_3_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_11_15
#define init_0_3_11_15
void FASTCALL call_0_3_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_12_0
#define init_0_3_12_0
void FASTCALL call_0_3_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_12_1
#define init_0_3_12_1
void FASTCALL call_0_3_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_12_2
#define init_0_3_12_2
void FASTCALL call_0_3_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_12_3
#define init_0_3_12_3
void FASTCALL call_0_3_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_12_4
#define init_0_3_12_4
void FASTCALL call_0_3_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_12_5
#define init_0_3_12_5
void FASTCALL call_0_3_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_12_6
#define init_0_3_12_6
void FASTCALL call_0_3_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_12_7
#define init_0_3_12_7
void FASTCALL call_0_3_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_12_8
#define init_0_3_12_8
void FASTCALL call_0_3_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_12_9
#define init_0_3_12_9
void FASTCALL call_0_3_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_12_10
#define init_0_3_12_10
void FASTCALL call_0_3_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_12_11
#define init_0_3_12_11
void FASTCALL call_0_3_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_12_12
#define init_0_3_12_12
void FASTCALL call_0_3_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_12_13
#define init_0_3_12_13
void FASTCALL call_0_3_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_12_14
#define init_0_3_12_14
void FASTCALL call_0_3_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_12_15
#define init_0_3_12_15
void FASTCALL call_0_3_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_13_0
#define init_0_3_13_0
void FASTCALL call_0_3_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_13_1
#define init_0_3_13_1
void FASTCALL call_0_3_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_13_2
#define init_0_3_13_2
void FASTCALL call_0_3_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_13_3
#define init_0_3_13_3
void FASTCALL call_0_3_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_13_4
#define init_0_3_13_4
void FASTCALL call_0_3_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_13_5
#define init_0_3_13_5
void FASTCALL call_0_3_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_13_6
#define init_0_3_13_6
void FASTCALL call_0_3_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_13_7
#define init_0_3_13_7
void FASTCALL call_0_3_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_13_8
#define init_0_3_13_8
void FASTCALL call_0_3_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_13_9
#define init_0_3_13_9
void FASTCALL call_0_3_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_13_10
#define init_0_3_13_10
void FASTCALL call_0_3_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_13_11
#define init_0_3_13_11
void FASTCALL call_0_3_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_13_12
#define init_0_3_13_12
void FASTCALL call_0_3_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_13_13
#define init_0_3_13_13
void FASTCALL call_0_3_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_13_14
#define init_0_3_13_14
void FASTCALL call_0_3_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_13_15
#define init_0_3_13_15
void FASTCALL call_0_3_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_14_0
#define init_0_3_14_0
void FASTCALL call_0_3_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_14_1
#define init_0_3_14_1
void FASTCALL call_0_3_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_14_2
#define init_0_3_14_2
void FASTCALL call_0_3_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_14_3
#define init_0_3_14_3
void FASTCALL call_0_3_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_14_4
#define init_0_3_14_4
void FASTCALL call_0_3_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_14_5
#define init_0_3_14_5
void FASTCALL call_0_3_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_14_6
#define init_0_3_14_6
void FASTCALL call_0_3_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_14_7
#define init_0_3_14_7
void FASTCALL call_0_3_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_14_8
#define init_0_3_14_8
void FASTCALL call_0_3_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_14_9
#define init_0_3_14_9
void FASTCALL call_0_3_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_14_10
#define init_0_3_14_10
void FASTCALL call_0_3_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_14_11
#define init_0_3_14_11
void FASTCALL call_0_3_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_14_12
#define init_0_3_14_12
void FASTCALL call_0_3_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_14_13
#define init_0_3_14_13
void FASTCALL call_0_3_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_14_14
#define init_0_3_14_14
void FASTCALL call_0_3_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_14_15
#define init_0_3_14_15
void FASTCALL call_0_3_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_15_0
#define init_0_3_15_0
void FASTCALL call_0_3_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_15_1
#define init_0_3_15_1
void FASTCALL call_0_3_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_15_2
#define init_0_3_15_2
void FASTCALL call_0_3_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_15_3
#define init_0_3_15_3
void FASTCALL call_0_3_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_15_4
#define init_0_3_15_4
void FASTCALL call_0_3_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_15_5
#define init_0_3_15_5
void FASTCALL call_0_3_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_15_6
#define init_0_3_15_6
void FASTCALL call_0_3_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_15_7
#define init_0_3_15_7
void FASTCALL call_0_3_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_15_8
#define init_0_3_15_8
void FASTCALL call_0_3_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_15_9
#define init_0_3_15_9
void FASTCALL call_0_3_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_15_10
#define init_0_3_15_10
void FASTCALL call_0_3_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_15_11
#define init_0_3_15_11
void FASTCALL call_0_3_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_15_12
#define init_0_3_15_12
void FASTCALL call_0_3_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_15_13
#define init_0_3_15_13
void FASTCALL call_0_3_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_15_14
#define init_0_3_15_14
void FASTCALL call_0_3_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_3_15_15
#define init_0_3_15_15
void FASTCALL call_0_3_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_0_0
#define init_0_4_0_0
void FASTCALL call_0_4_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_0_1
#define init_0_4_0_1
void FASTCALL call_0_4_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_0_2
#define init_0_4_0_2
void FASTCALL call_0_4_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_0_3
#define init_0_4_0_3
void FASTCALL call_0_4_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_0_4
#define init_0_4_0_4
void FASTCALL call_0_4_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_0_5
#define init_0_4_0_5
void FASTCALL call_0_4_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_0_6
#define init_0_4_0_6
void FASTCALL call_0_4_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_0_7
#define init_0_4_0_7
void FASTCALL call_0_4_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_0_8
#define init_0_4_0_8
void FASTCALL call_0_4_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_0_9
#define init_0_4_0_9
void FASTCALL call_0_4_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_0_10
#define init_0_4_0_10
void FASTCALL call_0_4_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_0_11
#define init_0_4_0_11
void FASTCALL call_0_4_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_0_12
#define init_0_4_0_12
void FASTCALL call_0_4_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_0_13
#define init_0_4_0_13
void FASTCALL call_0_4_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_0_14
#define init_0_4_0_14
void FASTCALL call_0_4_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_0_15
#define init_0_4_0_15
void FASTCALL call_0_4_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_1_0
#define init_0_4_1_0
void FASTCALL call_0_4_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_1_1
#define init_0_4_1_1
void FASTCALL call_0_4_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_1_2
#define init_0_4_1_2
void FASTCALL call_0_4_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_1_3
#define init_0_4_1_3
void FASTCALL call_0_4_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_1_4
#define init_0_4_1_4
void FASTCALL call_0_4_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_1_5
#define init_0_4_1_5
void FASTCALL call_0_4_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_1_6
#define init_0_4_1_6
void FASTCALL call_0_4_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_1_7
#define init_0_4_1_7
void FASTCALL call_0_4_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_1_8
#define init_0_4_1_8
void FASTCALL call_0_4_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_1_9
#define init_0_4_1_9
void FASTCALL call_0_4_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_1_10
#define init_0_4_1_10
void FASTCALL call_0_4_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_1_11
#define init_0_4_1_11
void FASTCALL call_0_4_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_1_12
#define init_0_4_1_12
void FASTCALL call_0_4_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_1_13
#define init_0_4_1_13
void FASTCALL call_0_4_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_1_14
#define init_0_4_1_14
void FASTCALL call_0_4_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_1_15
#define init_0_4_1_15
void FASTCALL call_0_4_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_2_0
#define init_0_4_2_0
void FASTCALL call_0_4_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_2_1
#define init_0_4_2_1
void FASTCALL call_0_4_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_2_2
#define init_0_4_2_2
void FASTCALL call_0_4_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_2_3
#define init_0_4_2_3
void FASTCALL call_0_4_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_2_4
#define init_0_4_2_4
void FASTCALL call_0_4_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_2_5
#define init_0_4_2_5
void FASTCALL call_0_4_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_2_6
#define init_0_4_2_6
void FASTCALL call_0_4_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_2_7
#define init_0_4_2_7
void FASTCALL call_0_4_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_2_8
#define init_0_4_2_8
void FASTCALL call_0_4_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_2_9
#define init_0_4_2_9
void FASTCALL call_0_4_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_2_10
#define init_0_4_2_10
void FASTCALL call_0_4_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_2_11
#define init_0_4_2_11
void FASTCALL call_0_4_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_2_12
#define init_0_4_2_12
void FASTCALL call_0_4_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_2_13
#define init_0_4_2_13
void FASTCALL call_0_4_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_2_14
#define init_0_4_2_14
void FASTCALL call_0_4_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_2_15
#define init_0_4_2_15
void FASTCALL call_0_4_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_3_0
#define init_0_4_3_0
void FASTCALL call_0_4_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_3_1
#define init_0_4_3_1
void FASTCALL call_0_4_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_3_2
#define init_0_4_3_2
void FASTCALL call_0_4_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_3_3
#define init_0_4_3_3
void FASTCALL call_0_4_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_3_4
#define init_0_4_3_4
void FASTCALL call_0_4_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_3_5
#define init_0_4_3_5
void FASTCALL call_0_4_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_3_6
#define init_0_4_3_6
void FASTCALL call_0_4_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_3_7
#define init_0_4_3_7
void FASTCALL call_0_4_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_3_8
#define init_0_4_3_8
void FASTCALL call_0_4_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_3_9
#define init_0_4_3_9
void FASTCALL call_0_4_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_3_10
#define init_0_4_3_10
void FASTCALL call_0_4_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_3_11
#define init_0_4_3_11
void FASTCALL call_0_4_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_3_12
#define init_0_4_3_12
void FASTCALL call_0_4_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_3_13
#define init_0_4_3_13
void FASTCALL call_0_4_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_3_14
#define init_0_4_3_14
void FASTCALL call_0_4_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_3_15
#define init_0_4_3_15
void FASTCALL call_0_4_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_4_0
#define init_0_4_4_0
void FASTCALL call_0_4_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_4_1
#define init_0_4_4_1
void FASTCALL call_0_4_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_4_2
#define init_0_4_4_2
void FASTCALL call_0_4_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_4_3
#define init_0_4_4_3
void FASTCALL call_0_4_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_4_4
#define init_0_4_4_4
void FASTCALL call_0_4_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_4_5
#define init_0_4_4_5
void FASTCALL call_0_4_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_4_6
#define init_0_4_4_6
void FASTCALL call_0_4_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_4_7
#define init_0_4_4_7
void FASTCALL call_0_4_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_4_8
#define init_0_4_4_8
void FASTCALL call_0_4_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_4_9
#define init_0_4_4_9
void FASTCALL call_0_4_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_4_10
#define init_0_4_4_10
void FASTCALL call_0_4_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_4_11
#define init_0_4_4_11
void FASTCALL call_0_4_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_4_12
#define init_0_4_4_12
void FASTCALL call_0_4_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_4_13
#define init_0_4_4_13
void FASTCALL call_0_4_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_4_14
#define init_0_4_4_14
void FASTCALL call_0_4_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_4_15
#define init_0_4_4_15
void FASTCALL call_0_4_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_5_0
#define init_0_4_5_0
void FASTCALL call_0_4_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_5_1
#define init_0_4_5_1
void FASTCALL call_0_4_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_5_2
#define init_0_4_5_2
void FASTCALL call_0_4_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_5_3
#define init_0_4_5_3
void FASTCALL call_0_4_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_5_4
#define init_0_4_5_4
void FASTCALL call_0_4_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_5_5
#define init_0_4_5_5
void FASTCALL call_0_4_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_5_6
#define init_0_4_5_6
void FASTCALL call_0_4_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_5_7
#define init_0_4_5_7
void FASTCALL call_0_4_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_5_8
#define init_0_4_5_8
void FASTCALL call_0_4_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_5_9
#define init_0_4_5_9
void FASTCALL call_0_4_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_5_10
#define init_0_4_5_10
void FASTCALL call_0_4_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_5_11
#define init_0_4_5_11
void FASTCALL call_0_4_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_5_12
#define init_0_4_5_12
void FASTCALL call_0_4_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_5_13
#define init_0_4_5_13
void FASTCALL call_0_4_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_5_14
#define init_0_4_5_14
void FASTCALL call_0_4_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_5_15
#define init_0_4_5_15
void FASTCALL call_0_4_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_6_0
#define init_0_4_6_0
void FASTCALL call_0_4_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_6_1
#define init_0_4_6_1
void FASTCALL call_0_4_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_6_2
#define init_0_4_6_2
void FASTCALL call_0_4_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_6_3
#define init_0_4_6_3
void FASTCALL call_0_4_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_6_4
#define init_0_4_6_4
void FASTCALL call_0_4_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_6_5
#define init_0_4_6_5
void FASTCALL call_0_4_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_6_6
#define init_0_4_6_6
void FASTCALL call_0_4_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_6_7
#define init_0_4_6_7
void FASTCALL call_0_4_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_6_8
#define init_0_4_6_8
void FASTCALL call_0_4_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_6_9
#define init_0_4_6_9
void FASTCALL call_0_4_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_6_10
#define init_0_4_6_10
void FASTCALL call_0_4_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_6_11
#define init_0_4_6_11
void FASTCALL call_0_4_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_6_12
#define init_0_4_6_12
void FASTCALL call_0_4_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_6_13
#define init_0_4_6_13
void FASTCALL call_0_4_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_6_14
#define init_0_4_6_14
void FASTCALL call_0_4_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_6_15
#define init_0_4_6_15
void FASTCALL call_0_4_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_7_0
#define init_0_4_7_0
void FASTCALL call_0_4_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_7_1
#define init_0_4_7_1
void FASTCALL call_0_4_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_7_2
#define init_0_4_7_2
void FASTCALL call_0_4_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_7_3
#define init_0_4_7_3
void FASTCALL call_0_4_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_7_4
#define init_0_4_7_4
void FASTCALL call_0_4_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_7_5
#define init_0_4_7_5
void FASTCALL call_0_4_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_7_6
#define init_0_4_7_6
void FASTCALL call_0_4_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_7_7
#define init_0_4_7_7
void FASTCALL call_0_4_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_7_8
#define init_0_4_7_8
void FASTCALL call_0_4_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_7_9
#define init_0_4_7_9
void FASTCALL call_0_4_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_7_10
#define init_0_4_7_10
void FASTCALL call_0_4_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_7_11
#define init_0_4_7_11
void FASTCALL call_0_4_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_7_12
#define init_0_4_7_12
void FASTCALL call_0_4_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_7_13
#define init_0_4_7_13
void FASTCALL call_0_4_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_7_14
#define init_0_4_7_14
void FASTCALL call_0_4_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_7_15
#define init_0_4_7_15
void FASTCALL call_0_4_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_8_0
#define init_0_4_8_0
void FASTCALL call_0_4_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_8_1
#define init_0_4_8_1
void FASTCALL call_0_4_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_8_2
#define init_0_4_8_2
void FASTCALL call_0_4_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_8_3
#define init_0_4_8_3
void FASTCALL call_0_4_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_8_4
#define init_0_4_8_4
void FASTCALL call_0_4_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_8_5
#define init_0_4_8_5
void FASTCALL call_0_4_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_8_6
#define init_0_4_8_6
void FASTCALL call_0_4_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_8_7
#define init_0_4_8_7
void FASTCALL call_0_4_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_8_8
#define init_0_4_8_8
void FASTCALL call_0_4_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_8_9
#define init_0_4_8_9
void FASTCALL call_0_4_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_8_10
#define init_0_4_8_10
void FASTCALL call_0_4_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_8_11
#define init_0_4_8_11
void FASTCALL call_0_4_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_8_12
#define init_0_4_8_12
void FASTCALL call_0_4_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_8_13
#define init_0_4_8_13
void FASTCALL call_0_4_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_8_14
#define init_0_4_8_14
void FASTCALL call_0_4_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_8_15
#define init_0_4_8_15
void FASTCALL call_0_4_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_9_0
#define init_0_4_9_0
void FASTCALL call_0_4_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_9_1
#define init_0_4_9_1
void FASTCALL call_0_4_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_9_2
#define init_0_4_9_2
void FASTCALL call_0_4_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_9_3
#define init_0_4_9_3
void FASTCALL call_0_4_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_9_4
#define init_0_4_9_4
void FASTCALL call_0_4_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_9_5
#define init_0_4_9_5
void FASTCALL call_0_4_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_9_6
#define init_0_4_9_6
void FASTCALL call_0_4_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_9_7
#define init_0_4_9_7
void FASTCALL call_0_4_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_9_8
#define init_0_4_9_8
void FASTCALL call_0_4_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_9_9
#define init_0_4_9_9
void FASTCALL call_0_4_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_9_10
#define init_0_4_9_10
void FASTCALL call_0_4_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_9_11
#define init_0_4_9_11
void FASTCALL call_0_4_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_9_12
#define init_0_4_9_12
void FASTCALL call_0_4_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_9_13
#define init_0_4_9_13
void FASTCALL call_0_4_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_9_14
#define init_0_4_9_14
void FASTCALL call_0_4_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_9_15
#define init_0_4_9_15
void FASTCALL call_0_4_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_10_0
#define init_0_4_10_0
void FASTCALL call_0_4_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_10_1
#define init_0_4_10_1
void FASTCALL call_0_4_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_10_2
#define init_0_4_10_2
void FASTCALL call_0_4_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_10_3
#define init_0_4_10_3
void FASTCALL call_0_4_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_10_4
#define init_0_4_10_4
void FASTCALL call_0_4_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_10_5
#define init_0_4_10_5
void FASTCALL call_0_4_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_10_6
#define init_0_4_10_6
void FASTCALL call_0_4_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_10_7
#define init_0_4_10_7
void FASTCALL call_0_4_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_10_8
#define init_0_4_10_8
void FASTCALL call_0_4_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_10_9
#define init_0_4_10_9
void FASTCALL call_0_4_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_10_10
#define init_0_4_10_10
void FASTCALL call_0_4_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_10_11
#define init_0_4_10_11
void FASTCALL call_0_4_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_10_12
#define init_0_4_10_12
void FASTCALL call_0_4_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_10_13
#define init_0_4_10_13
void FASTCALL call_0_4_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_10_14
#define init_0_4_10_14
void FASTCALL call_0_4_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_10_15
#define init_0_4_10_15
void FASTCALL call_0_4_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_11_0
#define init_0_4_11_0
void FASTCALL call_0_4_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_11_1
#define init_0_4_11_1
void FASTCALL call_0_4_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_11_2
#define init_0_4_11_2
void FASTCALL call_0_4_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_11_3
#define init_0_4_11_3
void FASTCALL call_0_4_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_11_4
#define init_0_4_11_4
void FASTCALL call_0_4_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_11_5
#define init_0_4_11_5
void FASTCALL call_0_4_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_11_6
#define init_0_4_11_6
void FASTCALL call_0_4_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_11_7
#define init_0_4_11_7
void FASTCALL call_0_4_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_11_8
#define init_0_4_11_8
void FASTCALL call_0_4_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_11_9
#define init_0_4_11_9
void FASTCALL call_0_4_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_11_10
#define init_0_4_11_10
void FASTCALL call_0_4_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_11_11
#define init_0_4_11_11
void FASTCALL call_0_4_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_11_12
#define init_0_4_11_12
void FASTCALL call_0_4_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_11_13
#define init_0_4_11_13
void FASTCALL call_0_4_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_11_14
#define init_0_4_11_14
void FASTCALL call_0_4_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_11_15
#define init_0_4_11_15
void FASTCALL call_0_4_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_12_0
#define init_0_4_12_0
void FASTCALL call_0_4_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_12_1
#define init_0_4_12_1
void FASTCALL call_0_4_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_12_2
#define init_0_4_12_2
void FASTCALL call_0_4_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_12_3
#define init_0_4_12_3
void FASTCALL call_0_4_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_12_4
#define init_0_4_12_4
void FASTCALL call_0_4_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_12_5
#define init_0_4_12_5
void FASTCALL call_0_4_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_12_6
#define init_0_4_12_6
void FASTCALL call_0_4_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_12_7
#define init_0_4_12_7
void FASTCALL call_0_4_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_12_8
#define init_0_4_12_8
void FASTCALL call_0_4_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_12_9
#define init_0_4_12_9
void FASTCALL call_0_4_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_12_10
#define init_0_4_12_10
void FASTCALL call_0_4_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_12_11
#define init_0_4_12_11
void FASTCALL call_0_4_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_12_12
#define init_0_4_12_12
void FASTCALL call_0_4_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_12_13
#define init_0_4_12_13
void FASTCALL call_0_4_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_12_14
#define init_0_4_12_14
void FASTCALL call_0_4_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_12_15
#define init_0_4_12_15
void FASTCALL call_0_4_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_13_0
#define init_0_4_13_0
void FASTCALL call_0_4_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_13_1
#define init_0_4_13_1
void FASTCALL call_0_4_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_13_2
#define init_0_4_13_2
void FASTCALL call_0_4_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_13_3
#define init_0_4_13_3
void FASTCALL call_0_4_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_13_4
#define init_0_4_13_4
void FASTCALL call_0_4_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_13_5
#define init_0_4_13_5
void FASTCALL call_0_4_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_13_6
#define init_0_4_13_6
void FASTCALL call_0_4_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_13_7
#define init_0_4_13_7
void FASTCALL call_0_4_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_13_8
#define init_0_4_13_8
void FASTCALL call_0_4_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_13_9
#define init_0_4_13_9
void FASTCALL call_0_4_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_13_10
#define init_0_4_13_10
void FASTCALL call_0_4_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_13_11
#define init_0_4_13_11
void FASTCALL call_0_4_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_13_12
#define init_0_4_13_12
void FASTCALL call_0_4_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_13_13
#define init_0_4_13_13
void FASTCALL call_0_4_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_13_14
#define init_0_4_13_14
void FASTCALL call_0_4_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_13_15
#define init_0_4_13_15
void FASTCALL call_0_4_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_14_0
#define init_0_4_14_0
void FASTCALL call_0_4_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_14_1
#define init_0_4_14_1
void FASTCALL call_0_4_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_14_2
#define init_0_4_14_2
void FASTCALL call_0_4_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_14_3
#define init_0_4_14_3
void FASTCALL call_0_4_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_14_4
#define init_0_4_14_4
void FASTCALL call_0_4_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_14_5
#define init_0_4_14_5
void FASTCALL call_0_4_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_14_6
#define init_0_4_14_6
void FASTCALL call_0_4_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_14_7
#define init_0_4_14_7
void FASTCALL call_0_4_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_14_8
#define init_0_4_14_8
void FASTCALL call_0_4_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_14_9
#define init_0_4_14_9
void FASTCALL call_0_4_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_14_10
#define init_0_4_14_10
void FASTCALL call_0_4_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_14_11
#define init_0_4_14_11
void FASTCALL call_0_4_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_14_12
#define init_0_4_14_12
void FASTCALL call_0_4_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_14_13
#define init_0_4_14_13
void FASTCALL call_0_4_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_14_14
#define init_0_4_14_14
void FASTCALL call_0_4_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_14_15
#define init_0_4_14_15
void FASTCALL call_0_4_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_15_0
#define init_0_4_15_0
void FASTCALL call_0_4_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_15_1
#define init_0_4_15_1
void FASTCALL call_0_4_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_15_2
#define init_0_4_15_2
void FASTCALL call_0_4_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_15_3
#define init_0_4_15_3
void FASTCALL call_0_4_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_15_4
#define init_0_4_15_4
void FASTCALL call_0_4_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_15_5
#define init_0_4_15_5
void FASTCALL call_0_4_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_15_6
#define init_0_4_15_6
void FASTCALL call_0_4_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_15_7
#define init_0_4_15_7
void FASTCALL call_0_4_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_15_8
#define init_0_4_15_8
void FASTCALL call_0_4_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_15_9
#define init_0_4_15_9
void FASTCALL call_0_4_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_15_10
#define init_0_4_15_10
void FASTCALL call_0_4_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_15_11
#define init_0_4_15_11
void FASTCALL call_0_4_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_15_12
#define init_0_4_15_12
void FASTCALL call_0_4_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_15_13
#define init_0_4_15_13
void FASTCALL call_0_4_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_15_14
#define init_0_4_15_14
void FASTCALL call_0_4_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_4_15_15
#define init_0_4_15_15
void FASTCALL call_0_4_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_0_0
#define init_0_5_0_0
void FASTCALL call_0_5_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_0_1
#define init_0_5_0_1
void FASTCALL call_0_5_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_0_2
#define init_0_5_0_2
void FASTCALL call_0_5_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_0_3
#define init_0_5_0_3
void FASTCALL call_0_5_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_0_4
#define init_0_5_0_4
void FASTCALL call_0_5_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_0_5
#define init_0_5_0_5
void FASTCALL call_0_5_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_0_6
#define init_0_5_0_6
void FASTCALL call_0_5_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_0_7
#define init_0_5_0_7
void FASTCALL call_0_5_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_0_8
#define init_0_5_0_8
void FASTCALL call_0_5_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_0_9
#define init_0_5_0_9
void FASTCALL call_0_5_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_0_10
#define init_0_5_0_10
void FASTCALL call_0_5_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_0_11
#define init_0_5_0_11
void FASTCALL call_0_5_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_0_12
#define init_0_5_0_12
void FASTCALL call_0_5_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_0_13
#define init_0_5_0_13
void FASTCALL call_0_5_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_0_14
#define init_0_5_0_14
void FASTCALL call_0_5_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_0_15
#define init_0_5_0_15
void FASTCALL call_0_5_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_1_0
#define init_0_5_1_0
void FASTCALL call_0_5_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_1_1
#define init_0_5_1_1
void FASTCALL call_0_5_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_1_2
#define init_0_5_1_2
void FASTCALL call_0_5_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_1_3
#define init_0_5_1_3
void FASTCALL call_0_5_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_1_4
#define init_0_5_1_4
void FASTCALL call_0_5_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_1_5
#define init_0_5_1_5
void FASTCALL call_0_5_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_1_6
#define init_0_5_1_6
void FASTCALL call_0_5_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_1_7
#define init_0_5_1_7
void FASTCALL call_0_5_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_1_8
#define init_0_5_1_8
void FASTCALL call_0_5_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_1_9
#define init_0_5_1_9
void FASTCALL call_0_5_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_1_10
#define init_0_5_1_10
void FASTCALL call_0_5_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_1_11
#define init_0_5_1_11
void FASTCALL call_0_5_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_1_12
#define init_0_5_1_12
void FASTCALL call_0_5_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_1_13
#define init_0_5_1_13
void FASTCALL call_0_5_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_1_14
#define init_0_5_1_14
void FASTCALL call_0_5_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_1_15
#define init_0_5_1_15
void FASTCALL call_0_5_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_2_0
#define init_0_5_2_0
void FASTCALL call_0_5_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_2_1
#define init_0_5_2_1
void FASTCALL call_0_5_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_2_2
#define init_0_5_2_2
void FASTCALL call_0_5_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_2_3
#define init_0_5_2_3
void FASTCALL call_0_5_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_2_4
#define init_0_5_2_4
void FASTCALL call_0_5_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_2_5
#define init_0_5_2_5
void FASTCALL call_0_5_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_2_6
#define init_0_5_2_6
void FASTCALL call_0_5_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_2_7
#define init_0_5_2_7
void FASTCALL call_0_5_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_2_8
#define init_0_5_2_8
void FASTCALL call_0_5_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_2_9
#define init_0_5_2_9
void FASTCALL call_0_5_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_2_10
#define init_0_5_2_10
void FASTCALL call_0_5_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_2_11
#define init_0_5_2_11
void FASTCALL call_0_5_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_2_12
#define init_0_5_2_12
void FASTCALL call_0_5_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_2_13
#define init_0_5_2_13
void FASTCALL call_0_5_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_2_14
#define init_0_5_2_14
void FASTCALL call_0_5_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_2_15
#define init_0_5_2_15
void FASTCALL call_0_5_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_3_0
#define init_0_5_3_0
void FASTCALL call_0_5_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_3_1
#define init_0_5_3_1
void FASTCALL call_0_5_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_3_2
#define init_0_5_3_2
void FASTCALL call_0_5_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_3_3
#define init_0_5_3_3
void FASTCALL call_0_5_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_3_4
#define init_0_5_3_4
void FASTCALL call_0_5_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_3_5
#define init_0_5_3_5
void FASTCALL call_0_5_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_3_6
#define init_0_5_3_6
void FASTCALL call_0_5_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_3_7
#define init_0_5_3_7
void FASTCALL call_0_5_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_3_8
#define init_0_5_3_8
void FASTCALL call_0_5_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_3_9
#define init_0_5_3_9
void FASTCALL call_0_5_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_3_10
#define init_0_5_3_10
void FASTCALL call_0_5_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_3_11
#define init_0_5_3_11
void FASTCALL call_0_5_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_3_12
#define init_0_5_3_12
void FASTCALL call_0_5_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_3_13
#define init_0_5_3_13
void FASTCALL call_0_5_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_3_14
#define init_0_5_3_14
void FASTCALL call_0_5_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_3_15
#define init_0_5_3_15
void FASTCALL call_0_5_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_4_0
#define init_0_5_4_0
void FASTCALL call_0_5_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_4_1
#define init_0_5_4_1
void FASTCALL call_0_5_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_4_2
#define init_0_5_4_2
void FASTCALL call_0_5_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_4_3
#define init_0_5_4_3
void FASTCALL call_0_5_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_4_4
#define init_0_5_4_4
void FASTCALL call_0_5_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_4_5
#define init_0_5_4_5
void FASTCALL call_0_5_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_4_6
#define init_0_5_4_6
void FASTCALL call_0_5_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_4_7
#define init_0_5_4_7
void FASTCALL call_0_5_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_4_8
#define init_0_5_4_8
void FASTCALL call_0_5_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_4_9
#define init_0_5_4_9
void FASTCALL call_0_5_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_4_10
#define init_0_5_4_10
void FASTCALL call_0_5_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_4_11
#define init_0_5_4_11
void FASTCALL call_0_5_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_4_12
#define init_0_5_4_12
void FASTCALL call_0_5_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_4_13
#define init_0_5_4_13
void FASTCALL call_0_5_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_4_14
#define init_0_5_4_14
void FASTCALL call_0_5_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_4_15
#define init_0_5_4_15
void FASTCALL call_0_5_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_5_0
#define init_0_5_5_0
void FASTCALL call_0_5_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_5_1
#define init_0_5_5_1
void FASTCALL call_0_5_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_5_2
#define init_0_5_5_2
void FASTCALL call_0_5_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_5_3
#define init_0_5_5_3
void FASTCALL call_0_5_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_5_4
#define init_0_5_5_4
void FASTCALL call_0_5_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_5_5
#define init_0_5_5_5
void FASTCALL call_0_5_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_5_6
#define init_0_5_5_6
void FASTCALL call_0_5_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_5_7
#define init_0_5_5_7
void FASTCALL call_0_5_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_5_8
#define init_0_5_5_8
void FASTCALL call_0_5_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_5_9
#define init_0_5_5_9
void FASTCALL call_0_5_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_5_10
#define init_0_5_5_10
void FASTCALL call_0_5_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_5_11
#define init_0_5_5_11
void FASTCALL call_0_5_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_5_12
#define init_0_5_5_12
void FASTCALL call_0_5_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_5_13
#define init_0_5_5_13
void FASTCALL call_0_5_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_5_14
#define init_0_5_5_14
void FASTCALL call_0_5_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_5_15
#define init_0_5_5_15
void FASTCALL call_0_5_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_6_0
#define init_0_5_6_0
void FASTCALL call_0_5_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_6_1
#define init_0_5_6_1
void FASTCALL call_0_5_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_6_2
#define init_0_5_6_2
void FASTCALL call_0_5_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_6_3
#define init_0_5_6_3
void FASTCALL call_0_5_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_6_4
#define init_0_5_6_4
void FASTCALL call_0_5_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_6_5
#define init_0_5_6_5
void FASTCALL call_0_5_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_6_6
#define init_0_5_6_6
void FASTCALL call_0_5_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_6_7
#define init_0_5_6_7
void FASTCALL call_0_5_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_6_8
#define init_0_5_6_8
void FASTCALL call_0_5_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_6_9
#define init_0_5_6_9
void FASTCALL call_0_5_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_6_10
#define init_0_5_6_10
void FASTCALL call_0_5_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_6_11
#define init_0_5_6_11
void FASTCALL call_0_5_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_6_12
#define init_0_5_6_12
void FASTCALL call_0_5_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_6_13
#define init_0_5_6_13
void FASTCALL call_0_5_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_6_14
#define init_0_5_6_14
void FASTCALL call_0_5_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_6_15
#define init_0_5_6_15
void FASTCALL call_0_5_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_7_0
#define init_0_5_7_0
void FASTCALL call_0_5_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_7_1
#define init_0_5_7_1
void FASTCALL call_0_5_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_7_2
#define init_0_5_7_2
void FASTCALL call_0_5_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_7_3
#define init_0_5_7_3
void FASTCALL call_0_5_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_7_4
#define init_0_5_7_4
void FASTCALL call_0_5_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_7_5
#define init_0_5_7_5
void FASTCALL call_0_5_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_7_6
#define init_0_5_7_6
void FASTCALL call_0_5_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_7_7
#define init_0_5_7_7
void FASTCALL call_0_5_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_7_8
#define init_0_5_7_8
void FASTCALL call_0_5_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_7_9
#define init_0_5_7_9
void FASTCALL call_0_5_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_7_10
#define init_0_5_7_10
void FASTCALL call_0_5_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_7_11
#define init_0_5_7_11
void FASTCALL call_0_5_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_7_12
#define init_0_5_7_12
void FASTCALL call_0_5_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_7_13
#define init_0_5_7_13
void FASTCALL call_0_5_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_7_14
#define init_0_5_7_14
void FASTCALL call_0_5_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_7_15
#define init_0_5_7_15
void FASTCALL call_0_5_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_8_0
#define init_0_5_8_0
void FASTCALL call_0_5_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_8_1
#define init_0_5_8_1
void FASTCALL call_0_5_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_8_2
#define init_0_5_8_2
void FASTCALL call_0_5_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_8_3
#define init_0_5_8_3
void FASTCALL call_0_5_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_8_4
#define init_0_5_8_4
void FASTCALL call_0_5_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_8_5
#define init_0_5_8_5
void FASTCALL call_0_5_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_8_6
#define init_0_5_8_6
void FASTCALL call_0_5_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_8_7
#define init_0_5_8_7
void FASTCALL call_0_5_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_8_8
#define init_0_5_8_8
void FASTCALL call_0_5_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_8_9
#define init_0_5_8_9
void FASTCALL call_0_5_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_8_10
#define init_0_5_8_10
void FASTCALL call_0_5_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_8_11
#define init_0_5_8_11
void FASTCALL call_0_5_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_8_12
#define init_0_5_8_12
void FASTCALL call_0_5_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_8_13
#define init_0_5_8_13
void FASTCALL call_0_5_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_8_14
#define init_0_5_8_14
void FASTCALL call_0_5_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_8_15
#define init_0_5_8_15
void FASTCALL call_0_5_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_9_0
#define init_0_5_9_0
void FASTCALL call_0_5_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_9_1
#define init_0_5_9_1
void FASTCALL call_0_5_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_9_2
#define init_0_5_9_2
void FASTCALL call_0_5_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_9_3
#define init_0_5_9_3
void FASTCALL call_0_5_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_9_4
#define init_0_5_9_4
void FASTCALL call_0_5_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_9_5
#define init_0_5_9_5
void FASTCALL call_0_5_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_9_6
#define init_0_5_9_6
void FASTCALL call_0_5_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_9_7
#define init_0_5_9_7
void FASTCALL call_0_5_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_9_8
#define init_0_5_9_8
void FASTCALL call_0_5_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_9_9
#define init_0_5_9_9
void FASTCALL call_0_5_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_9_10
#define init_0_5_9_10
void FASTCALL call_0_5_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_9_11
#define init_0_5_9_11
void FASTCALL call_0_5_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_9_12
#define init_0_5_9_12
void FASTCALL call_0_5_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_9_13
#define init_0_5_9_13
void FASTCALL call_0_5_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_9_14
#define init_0_5_9_14
void FASTCALL call_0_5_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_9_15
#define init_0_5_9_15
void FASTCALL call_0_5_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_10_0
#define init_0_5_10_0
void FASTCALL call_0_5_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_10_1
#define init_0_5_10_1
void FASTCALL call_0_5_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_10_2
#define init_0_5_10_2
void FASTCALL call_0_5_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_10_3
#define init_0_5_10_3
void FASTCALL call_0_5_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_10_4
#define init_0_5_10_4
void FASTCALL call_0_5_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_10_5
#define init_0_5_10_5
void FASTCALL call_0_5_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_10_6
#define init_0_5_10_6
void FASTCALL call_0_5_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_10_7
#define init_0_5_10_7
void FASTCALL call_0_5_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_10_8
#define init_0_5_10_8
void FASTCALL call_0_5_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_10_9
#define init_0_5_10_9
void FASTCALL call_0_5_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_10_10
#define init_0_5_10_10
void FASTCALL call_0_5_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_10_11
#define init_0_5_10_11
void FASTCALL call_0_5_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_10_12
#define init_0_5_10_12
void FASTCALL call_0_5_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_10_13
#define init_0_5_10_13
void FASTCALL call_0_5_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_10_14
#define init_0_5_10_14
void FASTCALL call_0_5_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_10_15
#define init_0_5_10_15
void FASTCALL call_0_5_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_11_0
#define init_0_5_11_0
void FASTCALL call_0_5_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_11_1
#define init_0_5_11_1
void FASTCALL call_0_5_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_11_2
#define init_0_5_11_2
void FASTCALL call_0_5_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_11_3
#define init_0_5_11_3
void FASTCALL call_0_5_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_11_4
#define init_0_5_11_4
void FASTCALL call_0_5_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_11_5
#define init_0_5_11_5
void FASTCALL call_0_5_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_11_6
#define init_0_5_11_6
void FASTCALL call_0_5_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_11_7
#define init_0_5_11_7
void FASTCALL call_0_5_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_11_8
#define init_0_5_11_8
void FASTCALL call_0_5_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_11_9
#define init_0_5_11_9
void FASTCALL call_0_5_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_11_10
#define init_0_5_11_10
void FASTCALL call_0_5_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_11_11
#define init_0_5_11_11
void FASTCALL call_0_5_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_11_12
#define init_0_5_11_12
void FASTCALL call_0_5_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_11_13
#define init_0_5_11_13
void FASTCALL call_0_5_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_11_14
#define init_0_5_11_14
void FASTCALL call_0_5_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_11_15
#define init_0_5_11_15
void FASTCALL call_0_5_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_12_0
#define init_0_5_12_0
void FASTCALL call_0_5_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_12_1
#define init_0_5_12_1
void FASTCALL call_0_5_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_12_2
#define init_0_5_12_2
void FASTCALL call_0_5_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_12_3
#define init_0_5_12_3
void FASTCALL call_0_5_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_12_4
#define init_0_5_12_4
void FASTCALL call_0_5_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_12_5
#define init_0_5_12_5
void FASTCALL call_0_5_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_12_6
#define init_0_5_12_6
void FASTCALL call_0_5_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_12_7
#define init_0_5_12_7
void FASTCALL call_0_5_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_12_8
#define init_0_5_12_8
void FASTCALL call_0_5_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_12_9
#define init_0_5_12_9
void FASTCALL call_0_5_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_12_10
#define init_0_5_12_10
void FASTCALL call_0_5_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_12_11
#define init_0_5_12_11
void FASTCALL call_0_5_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_12_12
#define init_0_5_12_12
void FASTCALL call_0_5_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_12_13
#define init_0_5_12_13
void FASTCALL call_0_5_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_12_14
#define init_0_5_12_14
void FASTCALL call_0_5_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_12_15
#define init_0_5_12_15
void FASTCALL call_0_5_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_13_0
#define init_0_5_13_0
void FASTCALL call_0_5_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_13_1
#define init_0_5_13_1
void FASTCALL call_0_5_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_13_2
#define init_0_5_13_2
void FASTCALL call_0_5_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_13_3
#define init_0_5_13_3
void FASTCALL call_0_5_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_13_4
#define init_0_5_13_4
void FASTCALL call_0_5_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_13_5
#define init_0_5_13_5
void FASTCALL call_0_5_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_13_6
#define init_0_5_13_6
void FASTCALL call_0_5_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_13_7
#define init_0_5_13_7
void FASTCALL call_0_5_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_13_8
#define init_0_5_13_8
void FASTCALL call_0_5_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_13_9
#define init_0_5_13_9
void FASTCALL call_0_5_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_13_10
#define init_0_5_13_10
void FASTCALL call_0_5_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_13_11
#define init_0_5_13_11
void FASTCALL call_0_5_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_13_12
#define init_0_5_13_12
void FASTCALL call_0_5_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_13_13
#define init_0_5_13_13
void FASTCALL call_0_5_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_13_14
#define init_0_5_13_14
void FASTCALL call_0_5_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_13_15
#define init_0_5_13_15
void FASTCALL call_0_5_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_14_0
#define init_0_5_14_0
void FASTCALL call_0_5_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_14_1
#define init_0_5_14_1
void FASTCALL call_0_5_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_14_2
#define init_0_5_14_2
void FASTCALL call_0_5_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_14_3
#define init_0_5_14_3
void FASTCALL call_0_5_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_14_4
#define init_0_5_14_4
void FASTCALL call_0_5_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_14_5
#define init_0_5_14_5
void FASTCALL call_0_5_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_14_6
#define init_0_5_14_6
void FASTCALL call_0_5_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_14_7
#define init_0_5_14_7
void FASTCALL call_0_5_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_14_8
#define init_0_5_14_8
void FASTCALL call_0_5_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_14_9
#define init_0_5_14_9
void FASTCALL call_0_5_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_14_10
#define init_0_5_14_10
void FASTCALL call_0_5_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_14_11
#define init_0_5_14_11
void FASTCALL call_0_5_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_14_12
#define init_0_5_14_12
void FASTCALL call_0_5_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_14_13
#define init_0_5_14_13
void FASTCALL call_0_5_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_14_14
#define init_0_5_14_14
void FASTCALL call_0_5_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_14_15
#define init_0_5_14_15
void FASTCALL call_0_5_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_15_0
#define init_0_5_15_0
void FASTCALL call_0_5_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_15_1
#define init_0_5_15_1
void FASTCALL call_0_5_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_15_2
#define init_0_5_15_2
void FASTCALL call_0_5_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_15_3
#define init_0_5_15_3
void FASTCALL call_0_5_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_15_4
#define init_0_5_15_4
void FASTCALL call_0_5_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_15_5
#define init_0_5_15_5
void FASTCALL call_0_5_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_15_6
#define init_0_5_15_6
void FASTCALL call_0_5_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_15_7
#define init_0_5_15_7
void FASTCALL call_0_5_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_15_8
#define init_0_5_15_8
void FASTCALL call_0_5_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_15_9
#define init_0_5_15_9
void FASTCALL call_0_5_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_15_10
#define init_0_5_15_10
void FASTCALL call_0_5_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_15_11
#define init_0_5_15_11
void FASTCALL call_0_5_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_15_12
#define init_0_5_15_12
void FASTCALL call_0_5_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_15_13
#define init_0_5_15_13
void FASTCALL call_0_5_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_15_14
#define init_0_5_15_14
void FASTCALL call_0_5_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_5_15_15
#define init_0_5_15_15
void FASTCALL call_0_5_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_0_0
#define init_0_6_0_0
void FASTCALL call_0_6_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_0_1
#define init_0_6_0_1
void FASTCALL call_0_6_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_0_2
#define init_0_6_0_2
void FASTCALL call_0_6_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_0_3
#define init_0_6_0_3
void FASTCALL call_0_6_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_0_4
#define init_0_6_0_4
void FASTCALL call_0_6_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_0_5
#define init_0_6_0_5
void FASTCALL call_0_6_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_0_6
#define init_0_6_0_6
void FASTCALL call_0_6_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_0_7
#define init_0_6_0_7
void FASTCALL call_0_6_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_0_8
#define init_0_6_0_8
void FASTCALL call_0_6_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_0_9
#define init_0_6_0_9
void FASTCALL call_0_6_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_0_10
#define init_0_6_0_10
void FASTCALL call_0_6_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_0_11
#define init_0_6_0_11
void FASTCALL call_0_6_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_0_12
#define init_0_6_0_12
void FASTCALL call_0_6_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_0_13
#define init_0_6_0_13
void FASTCALL call_0_6_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_0_14
#define init_0_6_0_14
void FASTCALL call_0_6_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_0_15
#define init_0_6_0_15
void FASTCALL call_0_6_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_1_0
#define init_0_6_1_0
void FASTCALL call_0_6_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_1_1
#define init_0_6_1_1
void FASTCALL call_0_6_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_1_2
#define init_0_6_1_2
void FASTCALL call_0_6_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_1_3
#define init_0_6_1_3
void FASTCALL call_0_6_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_1_4
#define init_0_6_1_4
void FASTCALL call_0_6_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_1_5
#define init_0_6_1_5
void FASTCALL call_0_6_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_1_6
#define init_0_6_1_6
void FASTCALL call_0_6_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_1_7
#define init_0_6_1_7
void FASTCALL call_0_6_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_1_8
#define init_0_6_1_8
void FASTCALL call_0_6_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_1_9
#define init_0_6_1_9
void FASTCALL call_0_6_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_1_10
#define init_0_6_1_10
void FASTCALL call_0_6_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_1_11
#define init_0_6_1_11
void FASTCALL call_0_6_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_1_12
#define init_0_6_1_12
void FASTCALL call_0_6_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_1_13
#define init_0_6_1_13
void FASTCALL call_0_6_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_1_14
#define init_0_6_1_14
void FASTCALL call_0_6_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_1_15
#define init_0_6_1_15
void FASTCALL call_0_6_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_2_0
#define init_0_6_2_0
void FASTCALL call_0_6_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_2_1
#define init_0_6_2_1
void FASTCALL call_0_6_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_2_2
#define init_0_6_2_2
void FASTCALL call_0_6_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_2_3
#define init_0_6_2_3
void FASTCALL call_0_6_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_2_4
#define init_0_6_2_4
void FASTCALL call_0_6_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_2_5
#define init_0_6_2_5
void FASTCALL call_0_6_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_2_6
#define init_0_6_2_6
void FASTCALL call_0_6_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_2_7
#define init_0_6_2_7
void FASTCALL call_0_6_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_2_8
#define init_0_6_2_8
void FASTCALL call_0_6_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_2_9
#define init_0_6_2_9
void FASTCALL call_0_6_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_2_10
#define init_0_6_2_10
void FASTCALL call_0_6_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_2_11
#define init_0_6_2_11
void FASTCALL call_0_6_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_2_12
#define init_0_6_2_12
void FASTCALL call_0_6_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_2_13
#define init_0_6_2_13
void FASTCALL call_0_6_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_2_14
#define init_0_6_2_14
void FASTCALL call_0_6_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_2_15
#define init_0_6_2_15
void FASTCALL call_0_6_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_3_0
#define init_0_6_3_0
void FASTCALL call_0_6_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_3_1
#define init_0_6_3_1
void FASTCALL call_0_6_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_3_2
#define init_0_6_3_2
void FASTCALL call_0_6_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_3_3
#define init_0_6_3_3
void FASTCALL call_0_6_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_3_4
#define init_0_6_3_4
void FASTCALL call_0_6_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_3_5
#define init_0_6_3_5
void FASTCALL call_0_6_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_3_6
#define init_0_6_3_6
void FASTCALL call_0_6_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_3_7
#define init_0_6_3_7
void FASTCALL call_0_6_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_3_8
#define init_0_6_3_8
void FASTCALL call_0_6_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_3_9
#define init_0_6_3_9
void FASTCALL call_0_6_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_3_10
#define init_0_6_3_10
void FASTCALL call_0_6_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_3_11
#define init_0_6_3_11
void FASTCALL call_0_6_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_3_12
#define init_0_6_3_12
void FASTCALL call_0_6_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_3_13
#define init_0_6_3_13
void FASTCALL call_0_6_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_3_14
#define init_0_6_3_14
void FASTCALL call_0_6_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_3_15
#define init_0_6_3_15
void FASTCALL call_0_6_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_4_0
#define init_0_6_4_0
void FASTCALL call_0_6_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_4_1
#define init_0_6_4_1
void FASTCALL call_0_6_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_4_2
#define init_0_6_4_2
void FASTCALL call_0_6_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_4_3
#define init_0_6_4_3
void FASTCALL call_0_6_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_4_4
#define init_0_6_4_4
void FASTCALL call_0_6_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_4_5
#define init_0_6_4_5
void FASTCALL call_0_6_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_4_6
#define init_0_6_4_6
void FASTCALL call_0_6_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_4_7
#define init_0_6_4_7
void FASTCALL call_0_6_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_4_8
#define init_0_6_4_8
void FASTCALL call_0_6_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_4_9
#define init_0_6_4_9
void FASTCALL call_0_6_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_4_10
#define init_0_6_4_10
void FASTCALL call_0_6_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_4_11
#define init_0_6_4_11
void FASTCALL call_0_6_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_4_12
#define init_0_6_4_12
void FASTCALL call_0_6_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_4_13
#define init_0_6_4_13
void FASTCALL call_0_6_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_4_14
#define init_0_6_4_14
void FASTCALL call_0_6_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_4_15
#define init_0_6_4_15
void FASTCALL call_0_6_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_5_0
#define init_0_6_5_0
void FASTCALL call_0_6_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_5_1
#define init_0_6_5_1
void FASTCALL call_0_6_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_5_2
#define init_0_6_5_2
void FASTCALL call_0_6_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_5_3
#define init_0_6_5_3
void FASTCALL call_0_6_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_5_4
#define init_0_6_5_4
void FASTCALL call_0_6_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_5_5
#define init_0_6_5_5
void FASTCALL call_0_6_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_5_6
#define init_0_6_5_6
void FASTCALL call_0_6_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_5_7
#define init_0_6_5_7
void FASTCALL call_0_6_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_5_8
#define init_0_6_5_8
void FASTCALL call_0_6_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_5_9
#define init_0_6_5_9
void FASTCALL call_0_6_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_5_10
#define init_0_6_5_10
void FASTCALL call_0_6_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_5_11
#define init_0_6_5_11
void FASTCALL call_0_6_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_5_12
#define init_0_6_5_12
void FASTCALL call_0_6_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_5_13
#define init_0_6_5_13
void FASTCALL call_0_6_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_5_14
#define init_0_6_5_14
void FASTCALL call_0_6_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_5_15
#define init_0_6_5_15
void FASTCALL call_0_6_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_6_0
#define init_0_6_6_0
void FASTCALL call_0_6_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_6_1
#define init_0_6_6_1
void FASTCALL call_0_6_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_6_2
#define init_0_6_6_2
void FASTCALL call_0_6_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_6_3
#define init_0_6_6_3
void FASTCALL call_0_6_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_6_4
#define init_0_6_6_4
void FASTCALL call_0_6_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_6_5
#define init_0_6_6_5
void FASTCALL call_0_6_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_6_6
#define init_0_6_6_6
void FASTCALL call_0_6_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_6_7
#define init_0_6_6_7
void FASTCALL call_0_6_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_6_8
#define init_0_6_6_8
void FASTCALL call_0_6_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_6_9
#define init_0_6_6_9
void FASTCALL call_0_6_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_6_10
#define init_0_6_6_10
void FASTCALL call_0_6_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_6_11
#define init_0_6_6_11
void FASTCALL call_0_6_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_6_12
#define init_0_6_6_12
void FASTCALL call_0_6_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_6_13
#define init_0_6_6_13
void FASTCALL call_0_6_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_6_14
#define init_0_6_6_14
void FASTCALL call_0_6_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_6_15
#define init_0_6_6_15
void FASTCALL call_0_6_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_7_0
#define init_0_6_7_0
void FASTCALL call_0_6_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_7_1
#define init_0_6_7_1
void FASTCALL call_0_6_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_7_2
#define init_0_6_7_2
void FASTCALL call_0_6_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_7_3
#define init_0_6_7_3
void FASTCALL call_0_6_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_7_4
#define init_0_6_7_4
void FASTCALL call_0_6_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_7_5
#define init_0_6_7_5
void FASTCALL call_0_6_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_7_6
#define init_0_6_7_6
void FASTCALL call_0_6_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_7_7
#define init_0_6_7_7
void FASTCALL call_0_6_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_7_8
#define init_0_6_7_8
void FASTCALL call_0_6_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_7_9
#define init_0_6_7_9
void FASTCALL call_0_6_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_7_10
#define init_0_6_7_10
void FASTCALL call_0_6_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_7_11
#define init_0_6_7_11
void FASTCALL call_0_6_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_7_12
#define init_0_6_7_12
void FASTCALL call_0_6_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_7_13
#define init_0_6_7_13
void FASTCALL call_0_6_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_7_14
#define init_0_6_7_14
void FASTCALL call_0_6_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_7_15
#define init_0_6_7_15
void FASTCALL call_0_6_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_8_0
#define init_0_6_8_0
void FASTCALL call_0_6_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_8_1
#define init_0_6_8_1
void FASTCALL call_0_6_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_8_2
#define init_0_6_8_2
void FASTCALL call_0_6_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_8_3
#define init_0_6_8_3
void FASTCALL call_0_6_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_8_4
#define init_0_6_8_4
void FASTCALL call_0_6_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_8_5
#define init_0_6_8_5
void FASTCALL call_0_6_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_8_6
#define init_0_6_8_6
void FASTCALL call_0_6_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_8_7
#define init_0_6_8_7
void FASTCALL call_0_6_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_8_8
#define init_0_6_8_8
void FASTCALL call_0_6_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_8_9
#define init_0_6_8_9
void FASTCALL call_0_6_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_8_10
#define init_0_6_8_10
void FASTCALL call_0_6_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_8_11
#define init_0_6_8_11
void FASTCALL call_0_6_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_8_12
#define init_0_6_8_12
void FASTCALL call_0_6_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_8_13
#define init_0_6_8_13
void FASTCALL call_0_6_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_8_14
#define init_0_6_8_14
void FASTCALL call_0_6_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_8_15
#define init_0_6_8_15
void FASTCALL call_0_6_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_9_0
#define init_0_6_9_0
void FASTCALL call_0_6_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_9_1
#define init_0_6_9_1
void FASTCALL call_0_6_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_9_2
#define init_0_6_9_2
void FASTCALL call_0_6_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_9_3
#define init_0_6_9_3
void FASTCALL call_0_6_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_9_4
#define init_0_6_9_4
void FASTCALL call_0_6_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_9_5
#define init_0_6_9_5
void FASTCALL call_0_6_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_9_6
#define init_0_6_9_6
void FASTCALL call_0_6_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_9_7
#define init_0_6_9_7
void FASTCALL call_0_6_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_9_8
#define init_0_6_9_8
void FASTCALL call_0_6_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_9_9
#define init_0_6_9_9
void FASTCALL call_0_6_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_9_10
#define init_0_6_9_10
void FASTCALL call_0_6_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_9_11
#define init_0_6_9_11
void FASTCALL call_0_6_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_9_12
#define init_0_6_9_12
void FASTCALL call_0_6_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_9_13
#define init_0_6_9_13
void FASTCALL call_0_6_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_9_14
#define init_0_6_9_14
void FASTCALL call_0_6_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_9_15
#define init_0_6_9_15
void FASTCALL call_0_6_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_10_0
#define init_0_6_10_0
void FASTCALL call_0_6_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_10_1
#define init_0_6_10_1
void FASTCALL call_0_6_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_10_2
#define init_0_6_10_2
void FASTCALL call_0_6_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_10_3
#define init_0_6_10_3
void FASTCALL call_0_6_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_10_4
#define init_0_6_10_4
void FASTCALL call_0_6_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_10_5
#define init_0_6_10_5
void FASTCALL call_0_6_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_10_6
#define init_0_6_10_6
void FASTCALL call_0_6_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_10_7
#define init_0_6_10_7
void FASTCALL call_0_6_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_10_8
#define init_0_6_10_8
void FASTCALL call_0_6_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_10_9
#define init_0_6_10_9
void FASTCALL call_0_6_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_10_10
#define init_0_6_10_10
void FASTCALL call_0_6_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_10_11
#define init_0_6_10_11
void FASTCALL call_0_6_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_10_12
#define init_0_6_10_12
void FASTCALL call_0_6_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_10_13
#define init_0_6_10_13
void FASTCALL call_0_6_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_10_14
#define init_0_6_10_14
void FASTCALL call_0_6_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_10_15
#define init_0_6_10_15
void FASTCALL call_0_6_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_11_0
#define init_0_6_11_0
void FASTCALL call_0_6_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_11_1
#define init_0_6_11_1
void FASTCALL call_0_6_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_11_2
#define init_0_6_11_2
void FASTCALL call_0_6_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_11_3
#define init_0_6_11_3
void FASTCALL call_0_6_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_11_4
#define init_0_6_11_4
void FASTCALL call_0_6_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_11_5
#define init_0_6_11_5
void FASTCALL call_0_6_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_11_6
#define init_0_6_11_6
void FASTCALL call_0_6_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_11_7
#define init_0_6_11_7
void FASTCALL call_0_6_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_11_8
#define init_0_6_11_8
void FASTCALL call_0_6_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_11_9
#define init_0_6_11_9
void FASTCALL call_0_6_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_11_10
#define init_0_6_11_10
void FASTCALL call_0_6_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_11_11
#define init_0_6_11_11
void FASTCALL call_0_6_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_11_12
#define init_0_6_11_12
void FASTCALL call_0_6_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_11_13
#define init_0_6_11_13
void FASTCALL call_0_6_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_11_14
#define init_0_6_11_14
void FASTCALL call_0_6_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_11_15
#define init_0_6_11_15
void FASTCALL call_0_6_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_12_0
#define init_0_6_12_0
void FASTCALL call_0_6_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_12_1
#define init_0_6_12_1
void FASTCALL call_0_6_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_12_2
#define init_0_6_12_2
void FASTCALL call_0_6_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_12_3
#define init_0_6_12_3
void FASTCALL call_0_6_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_12_4
#define init_0_6_12_4
void FASTCALL call_0_6_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_12_5
#define init_0_6_12_5
void FASTCALL call_0_6_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_12_6
#define init_0_6_12_6
void FASTCALL call_0_6_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_12_7
#define init_0_6_12_7
void FASTCALL call_0_6_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_12_8
#define init_0_6_12_8
void FASTCALL call_0_6_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_12_9
#define init_0_6_12_9
void FASTCALL call_0_6_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_12_10
#define init_0_6_12_10
void FASTCALL call_0_6_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_12_11
#define init_0_6_12_11
void FASTCALL call_0_6_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_12_12
#define init_0_6_12_12
void FASTCALL call_0_6_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_12_13
#define init_0_6_12_13
void FASTCALL call_0_6_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_12_14
#define init_0_6_12_14
void FASTCALL call_0_6_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_12_15
#define init_0_6_12_15
void FASTCALL call_0_6_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_13_0
#define init_0_6_13_0
void FASTCALL call_0_6_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_13_1
#define init_0_6_13_1
void FASTCALL call_0_6_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_13_2
#define init_0_6_13_2
void FASTCALL call_0_6_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_13_3
#define init_0_6_13_3
void FASTCALL call_0_6_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_13_4
#define init_0_6_13_4
void FASTCALL call_0_6_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_13_5
#define init_0_6_13_5
void FASTCALL call_0_6_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_13_6
#define init_0_6_13_6
void FASTCALL call_0_6_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_13_7
#define init_0_6_13_7
void FASTCALL call_0_6_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_13_8
#define init_0_6_13_8
void FASTCALL call_0_6_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_13_9
#define init_0_6_13_9
void FASTCALL call_0_6_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_13_10
#define init_0_6_13_10
void FASTCALL call_0_6_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_13_11
#define init_0_6_13_11
void FASTCALL call_0_6_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_13_12
#define init_0_6_13_12
void FASTCALL call_0_6_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_13_13
#define init_0_6_13_13
void FASTCALL call_0_6_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_13_14
#define init_0_6_13_14
void FASTCALL call_0_6_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_13_15
#define init_0_6_13_15
void FASTCALL call_0_6_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_14_0
#define init_0_6_14_0
void FASTCALL call_0_6_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_14_1
#define init_0_6_14_1
void FASTCALL call_0_6_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_14_2
#define init_0_6_14_2
void FASTCALL call_0_6_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_14_3
#define init_0_6_14_3
void FASTCALL call_0_6_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_14_4
#define init_0_6_14_4
void FASTCALL call_0_6_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_14_5
#define init_0_6_14_5
void FASTCALL call_0_6_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_14_6
#define init_0_6_14_6
void FASTCALL call_0_6_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_14_7
#define init_0_6_14_7
void FASTCALL call_0_6_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_14_8
#define init_0_6_14_8
void FASTCALL call_0_6_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_14_9
#define init_0_6_14_9
void FASTCALL call_0_6_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_14_10
#define init_0_6_14_10
void FASTCALL call_0_6_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_14_11
#define init_0_6_14_11
void FASTCALL call_0_6_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_14_12
#define init_0_6_14_12
void FASTCALL call_0_6_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_14_13
#define init_0_6_14_13
void FASTCALL call_0_6_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_14_14
#define init_0_6_14_14
void FASTCALL call_0_6_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_14_15
#define init_0_6_14_15
void FASTCALL call_0_6_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_15_0
#define init_0_6_15_0
void FASTCALL call_0_6_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_15_1
#define init_0_6_15_1
void FASTCALL call_0_6_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_15_2
#define init_0_6_15_2
void FASTCALL call_0_6_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_15_3
#define init_0_6_15_3
void FASTCALL call_0_6_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_15_4
#define init_0_6_15_4
void FASTCALL call_0_6_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_15_5
#define init_0_6_15_5
void FASTCALL call_0_6_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_15_6
#define init_0_6_15_6
void FASTCALL call_0_6_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_15_7
#define init_0_6_15_7
void FASTCALL call_0_6_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_15_8
#define init_0_6_15_8
void FASTCALL call_0_6_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_15_9
#define init_0_6_15_9
void FASTCALL call_0_6_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_15_10
#define init_0_6_15_10
void FASTCALL call_0_6_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_15_11
#define init_0_6_15_11
void FASTCALL call_0_6_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_15_12
#define init_0_6_15_12
void FASTCALL call_0_6_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_15_13
#define init_0_6_15_13
void FASTCALL call_0_6_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_15_14
#define init_0_6_15_14
void FASTCALL call_0_6_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_6_15_15
#define init_0_6_15_15
void FASTCALL call_0_6_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_0_0
#define init_0_7_0_0
void FASTCALL call_0_7_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_0_1
#define init_0_7_0_1
void FASTCALL call_0_7_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_0_2
#define init_0_7_0_2
void FASTCALL call_0_7_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_0_3
#define init_0_7_0_3
void FASTCALL call_0_7_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_0_4
#define init_0_7_0_4
void FASTCALL call_0_7_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_0_5
#define init_0_7_0_5
void FASTCALL call_0_7_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_0_6
#define init_0_7_0_6
void FASTCALL call_0_7_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_0_7
#define init_0_7_0_7
void FASTCALL call_0_7_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_0_8
#define init_0_7_0_8
void FASTCALL call_0_7_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_0_9
#define init_0_7_0_9
void FASTCALL call_0_7_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_0_10
#define init_0_7_0_10
void FASTCALL call_0_7_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_0_11
#define init_0_7_0_11
void FASTCALL call_0_7_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_0_12
#define init_0_7_0_12
void FASTCALL call_0_7_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_0_13
#define init_0_7_0_13
void FASTCALL call_0_7_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_0_14
#define init_0_7_0_14
void FASTCALL call_0_7_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_0_15
#define init_0_7_0_15
void FASTCALL call_0_7_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_1_0
#define init_0_7_1_0
void FASTCALL call_0_7_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_1_1
#define init_0_7_1_1
void FASTCALL call_0_7_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_1_2
#define init_0_7_1_2
void FASTCALL call_0_7_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_1_3
#define init_0_7_1_3
void FASTCALL call_0_7_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_1_4
#define init_0_7_1_4
void FASTCALL call_0_7_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_1_5
#define init_0_7_1_5
void FASTCALL call_0_7_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_1_6
#define init_0_7_1_6
void FASTCALL call_0_7_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_1_7
#define init_0_7_1_7
void FASTCALL call_0_7_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_1_8
#define init_0_7_1_8
void FASTCALL call_0_7_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_1_9
#define init_0_7_1_9
void FASTCALL call_0_7_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_1_10
#define init_0_7_1_10
void FASTCALL call_0_7_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_1_11
#define init_0_7_1_11
void FASTCALL call_0_7_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_1_12
#define init_0_7_1_12
void FASTCALL call_0_7_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_1_13
#define init_0_7_1_13
void FASTCALL call_0_7_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_1_14
#define init_0_7_1_14
void FASTCALL call_0_7_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_1_15
#define init_0_7_1_15
void FASTCALL call_0_7_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_2_0
#define init_0_7_2_0
void FASTCALL call_0_7_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_2_1
#define init_0_7_2_1
void FASTCALL call_0_7_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_2_2
#define init_0_7_2_2
void FASTCALL call_0_7_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_2_3
#define init_0_7_2_3
void FASTCALL call_0_7_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_2_4
#define init_0_7_2_4
void FASTCALL call_0_7_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_2_5
#define init_0_7_2_5
void FASTCALL call_0_7_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_2_6
#define init_0_7_2_6
void FASTCALL call_0_7_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_2_7
#define init_0_7_2_7
void FASTCALL call_0_7_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_2_8
#define init_0_7_2_8
void FASTCALL call_0_7_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_2_9
#define init_0_7_2_9
void FASTCALL call_0_7_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_2_10
#define init_0_7_2_10
void FASTCALL call_0_7_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_2_11
#define init_0_7_2_11
void FASTCALL call_0_7_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_2_12
#define init_0_7_2_12
void FASTCALL call_0_7_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_2_13
#define init_0_7_2_13
void FASTCALL call_0_7_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_2_14
#define init_0_7_2_14
void FASTCALL call_0_7_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_2_15
#define init_0_7_2_15
void FASTCALL call_0_7_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_3_0
#define init_0_7_3_0
void FASTCALL call_0_7_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_3_1
#define init_0_7_3_1
void FASTCALL call_0_7_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_3_2
#define init_0_7_3_2
void FASTCALL call_0_7_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_3_3
#define init_0_7_3_3
void FASTCALL call_0_7_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_3_4
#define init_0_7_3_4
void FASTCALL call_0_7_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_3_5
#define init_0_7_3_5
void FASTCALL call_0_7_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_3_6
#define init_0_7_3_6
void FASTCALL call_0_7_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_3_7
#define init_0_7_3_7
void FASTCALL call_0_7_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_3_8
#define init_0_7_3_8
void FASTCALL call_0_7_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_3_9
#define init_0_7_3_9
void FASTCALL call_0_7_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_3_10
#define init_0_7_3_10
void FASTCALL call_0_7_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_3_11
#define init_0_7_3_11
void FASTCALL call_0_7_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_3_12
#define init_0_7_3_12
void FASTCALL call_0_7_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_3_13
#define init_0_7_3_13
void FASTCALL call_0_7_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_3_14
#define init_0_7_3_14
void FASTCALL call_0_7_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_3_15
#define init_0_7_3_15
void FASTCALL call_0_7_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_4_0
#define init_0_7_4_0
void FASTCALL call_0_7_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_4_1
#define init_0_7_4_1
void FASTCALL call_0_7_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_4_2
#define init_0_7_4_2
void FASTCALL call_0_7_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_4_3
#define init_0_7_4_3
void FASTCALL call_0_7_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_4_4
#define init_0_7_4_4
void FASTCALL call_0_7_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_4_5
#define init_0_7_4_5
void FASTCALL call_0_7_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_4_6
#define init_0_7_4_6
void FASTCALL call_0_7_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_4_7
#define init_0_7_4_7
void FASTCALL call_0_7_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_4_8
#define init_0_7_4_8
void FASTCALL call_0_7_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_4_9
#define init_0_7_4_9
void FASTCALL call_0_7_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_4_10
#define init_0_7_4_10
void FASTCALL call_0_7_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_4_11
#define init_0_7_4_11
void FASTCALL call_0_7_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_4_12
#define init_0_7_4_12
void FASTCALL call_0_7_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_4_13
#define init_0_7_4_13
void FASTCALL call_0_7_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_4_14
#define init_0_7_4_14
void FASTCALL call_0_7_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_4_15
#define init_0_7_4_15
void FASTCALL call_0_7_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_5_0
#define init_0_7_5_0
void FASTCALL call_0_7_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_5_1
#define init_0_7_5_1
void FASTCALL call_0_7_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_5_2
#define init_0_7_5_2
void FASTCALL call_0_7_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_5_3
#define init_0_7_5_3
void FASTCALL call_0_7_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_5_4
#define init_0_7_5_4
void FASTCALL call_0_7_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_5_5
#define init_0_7_5_5
void FASTCALL call_0_7_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_5_6
#define init_0_7_5_6
void FASTCALL call_0_7_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_5_7
#define init_0_7_5_7
void FASTCALL call_0_7_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_5_8
#define init_0_7_5_8
void FASTCALL call_0_7_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_5_9
#define init_0_7_5_9
void FASTCALL call_0_7_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_5_10
#define init_0_7_5_10
void FASTCALL call_0_7_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_5_11
#define init_0_7_5_11
void FASTCALL call_0_7_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_5_12
#define init_0_7_5_12
void FASTCALL call_0_7_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_5_13
#define init_0_7_5_13
void FASTCALL call_0_7_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_5_14
#define init_0_7_5_14
void FASTCALL call_0_7_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_5_15
#define init_0_7_5_15
void FASTCALL call_0_7_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_6_0
#define init_0_7_6_0
void FASTCALL call_0_7_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_6_1
#define init_0_7_6_1
void FASTCALL call_0_7_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_6_2
#define init_0_7_6_2
void FASTCALL call_0_7_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_6_3
#define init_0_7_6_3
void FASTCALL call_0_7_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_6_4
#define init_0_7_6_4
void FASTCALL call_0_7_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_6_5
#define init_0_7_6_5
void FASTCALL call_0_7_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_6_6
#define init_0_7_6_6
void FASTCALL call_0_7_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_6_7
#define init_0_7_6_7
void FASTCALL call_0_7_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_6_8
#define init_0_7_6_8
void FASTCALL call_0_7_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_6_9
#define init_0_7_6_9
void FASTCALL call_0_7_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_6_10
#define init_0_7_6_10
void FASTCALL call_0_7_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_6_11
#define init_0_7_6_11
void FASTCALL call_0_7_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_6_12
#define init_0_7_6_12
void FASTCALL call_0_7_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_6_13
#define init_0_7_6_13
void FASTCALL call_0_7_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_6_14
#define init_0_7_6_14
void FASTCALL call_0_7_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_6_15
#define init_0_7_6_15
void FASTCALL call_0_7_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_7_0
#define init_0_7_7_0
void FASTCALL call_0_7_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_7_1
#define init_0_7_7_1
void FASTCALL call_0_7_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_7_2
#define init_0_7_7_2
void FASTCALL call_0_7_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_7_3
#define init_0_7_7_3
void FASTCALL call_0_7_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_7_4
#define init_0_7_7_4
void FASTCALL call_0_7_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_7_5
#define init_0_7_7_5
void FASTCALL call_0_7_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_7_6
#define init_0_7_7_6
void FASTCALL call_0_7_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_7_7
#define init_0_7_7_7
void FASTCALL call_0_7_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_7_8
#define init_0_7_7_8
void FASTCALL call_0_7_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_7_9
#define init_0_7_7_9
void FASTCALL call_0_7_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_7_10
#define init_0_7_7_10
void FASTCALL call_0_7_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_7_11
#define init_0_7_7_11
void FASTCALL call_0_7_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_7_12
#define init_0_7_7_12
void FASTCALL call_0_7_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_7_13
#define init_0_7_7_13
void FASTCALL call_0_7_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_7_14
#define init_0_7_7_14
void FASTCALL call_0_7_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_7_15
#define init_0_7_7_15
void FASTCALL call_0_7_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_8_0
#define init_0_7_8_0
void FASTCALL call_0_7_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_8_1
#define init_0_7_8_1
void FASTCALL call_0_7_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_8_2
#define init_0_7_8_2
void FASTCALL call_0_7_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_8_3
#define init_0_7_8_3
void FASTCALL call_0_7_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_8_4
#define init_0_7_8_4
void FASTCALL call_0_7_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_8_5
#define init_0_7_8_5
void FASTCALL call_0_7_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_8_6
#define init_0_7_8_6
void FASTCALL call_0_7_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_8_7
#define init_0_7_8_7
void FASTCALL call_0_7_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_8_8
#define init_0_7_8_8
void FASTCALL call_0_7_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_8_9
#define init_0_7_8_9
void FASTCALL call_0_7_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_8_10
#define init_0_7_8_10
void FASTCALL call_0_7_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_8_11
#define init_0_7_8_11
void FASTCALL call_0_7_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_8_12
#define init_0_7_8_12
void FASTCALL call_0_7_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_8_13
#define init_0_7_8_13
void FASTCALL call_0_7_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_8_14
#define init_0_7_8_14
void FASTCALL call_0_7_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_8_15
#define init_0_7_8_15
void FASTCALL call_0_7_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_9_0
#define init_0_7_9_0
void FASTCALL call_0_7_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_9_1
#define init_0_7_9_1
void FASTCALL call_0_7_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_9_2
#define init_0_7_9_2
void FASTCALL call_0_7_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_9_3
#define init_0_7_9_3
void FASTCALL call_0_7_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_9_4
#define init_0_7_9_4
void FASTCALL call_0_7_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_9_5
#define init_0_7_9_5
void FASTCALL call_0_7_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_9_6
#define init_0_7_9_6
void FASTCALL call_0_7_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_9_7
#define init_0_7_9_7
void FASTCALL call_0_7_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_9_8
#define init_0_7_9_8
void FASTCALL call_0_7_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_9_9
#define init_0_7_9_9
void FASTCALL call_0_7_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_9_10
#define init_0_7_9_10
void FASTCALL call_0_7_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_9_11
#define init_0_7_9_11
void FASTCALL call_0_7_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_9_12
#define init_0_7_9_12
void FASTCALL call_0_7_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_9_13
#define init_0_7_9_13
void FASTCALL call_0_7_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_9_14
#define init_0_7_9_14
void FASTCALL call_0_7_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_9_15
#define init_0_7_9_15
void FASTCALL call_0_7_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_10_0
#define init_0_7_10_0
void FASTCALL call_0_7_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_10_1
#define init_0_7_10_1
void FASTCALL call_0_7_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_10_2
#define init_0_7_10_2
void FASTCALL call_0_7_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_10_3
#define init_0_7_10_3
void FASTCALL call_0_7_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_10_4
#define init_0_7_10_4
void FASTCALL call_0_7_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_10_5
#define init_0_7_10_5
void FASTCALL call_0_7_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_10_6
#define init_0_7_10_6
void FASTCALL call_0_7_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_10_7
#define init_0_7_10_7
void FASTCALL call_0_7_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_10_8
#define init_0_7_10_8
void FASTCALL call_0_7_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_10_9
#define init_0_7_10_9
void FASTCALL call_0_7_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_10_10
#define init_0_7_10_10
void FASTCALL call_0_7_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_10_11
#define init_0_7_10_11
void FASTCALL call_0_7_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_10_12
#define init_0_7_10_12
void FASTCALL call_0_7_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_10_13
#define init_0_7_10_13
void FASTCALL call_0_7_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_10_14
#define init_0_7_10_14
void FASTCALL call_0_7_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_10_15
#define init_0_7_10_15
void FASTCALL call_0_7_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_11_0
#define init_0_7_11_0
void FASTCALL call_0_7_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_11_1
#define init_0_7_11_1
void FASTCALL call_0_7_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_11_2
#define init_0_7_11_2
void FASTCALL call_0_7_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_11_3
#define init_0_7_11_3
void FASTCALL call_0_7_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_11_4
#define init_0_7_11_4
void FASTCALL call_0_7_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_11_5
#define init_0_7_11_5
void FASTCALL call_0_7_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_11_6
#define init_0_7_11_6
void FASTCALL call_0_7_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_11_7
#define init_0_7_11_7
void FASTCALL call_0_7_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_11_8
#define init_0_7_11_8
void FASTCALL call_0_7_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_11_9
#define init_0_7_11_9
void FASTCALL call_0_7_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_11_10
#define init_0_7_11_10
void FASTCALL call_0_7_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_11_11
#define init_0_7_11_11
void FASTCALL call_0_7_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_11_12
#define init_0_7_11_12
void FASTCALL call_0_7_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_11_13
#define init_0_7_11_13
void FASTCALL call_0_7_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_11_14
#define init_0_7_11_14
void FASTCALL call_0_7_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_11_15
#define init_0_7_11_15
void FASTCALL call_0_7_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_12_0
#define init_0_7_12_0
void FASTCALL call_0_7_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_12_1
#define init_0_7_12_1
void FASTCALL call_0_7_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_12_2
#define init_0_7_12_2
void FASTCALL call_0_7_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_12_3
#define init_0_7_12_3
void FASTCALL call_0_7_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_12_4
#define init_0_7_12_4
void FASTCALL call_0_7_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_12_5
#define init_0_7_12_5
void FASTCALL call_0_7_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_12_6
#define init_0_7_12_6
void FASTCALL call_0_7_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_12_7
#define init_0_7_12_7
void FASTCALL call_0_7_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_12_8
#define init_0_7_12_8
void FASTCALL call_0_7_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_12_9
#define init_0_7_12_9
void FASTCALL call_0_7_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_12_10
#define init_0_7_12_10
void FASTCALL call_0_7_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_12_11
#define init_0_7_12_11
void FASTCALL call_0_7_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_12_12
#define init_0_7_12_12
void FASTCALL call_0_7_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_12_13
#define init_0_7_12_13
void FASTCALL call_0_7_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_12_14
#define init_0_7_12_14
void FASTCALL call_0_7_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_12_15
#define init_0_7_12_15
void FASTCALL call_0_7_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_13_0
#define init_0_7_13_0
void FASTCALL call_0_7_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_13_1
#define init_0_7_13_1
void FASTCALL call_0_7_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_13_2
#define init_0_7_13_2
void FASTCALL call_0_7_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_13_3
#define init_0_7_13_3
void FASTCALL call_0_7_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_13_4
#define init_0_7_13_4
void FASTCALL call_0_7_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_13_5
#define init_0_7_13_5
void FASTCALL call_0_7_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_13_6
#define init_0_7_13_6
void FASTCALL call_0_7_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_13_7
#define init_0_7_13_7
void FASTCALL call_0_7_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_13_8
#define init_0_7_13_8
void FASTCALL call_0_7_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_13_9
#define init_0_7_13_9
void FASTCALL call_0_7_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_13_10
#define init_0_7_13_10
void FASTCALL call_0_7_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_13_11
#define init_0_7_13_11
void FASTCALL call_0_7_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_13_12
#define init_0_7_13_12
void FASTCALL call_0_7_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_13_13
#define init_0_7_13_13
void FASTCALL call_0_7_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_13_14
#define init_0_7_13_14
void FASTCALL call_0_7_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_13_15
#define init_0_7_13_15
void FASTCALL call_0_7_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_14_0
#define init_0_7_14_0
void FASTCALL call_0_7_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_14_1
#define init_0_7_14_1
void FASTCALL call_0_7_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_14_2
#define init_0_7_14_2
void FASTCALL call_0_7_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_14_3
#define init_0_7_14_3
void FASTCALL call_0_7_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_14_4
#define init_0_7_14_4
void FASTCALL call_0_7_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_14_5
#define init_0_7_14_5
void FASTCALL call_0_7_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_14_6
#define init_0_7_14_6
void FASTCALL call_0_7_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_14_7
#define init_0_7_14_7
void FASTCALL call_0_7_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_14_8
#define init_0_7_14_8
void FASTCALL call_0_7_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_14_9
#define init_0_7_14_9
void FASTCALL call_0_7_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_14_10
#define init_0_7_14_10
void FASTCALL call_0_7_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_14_11
#define init_0_7_14_11
void FASTCALL call_0_7_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_14_12
#define init_0_7_14_12
void FASTCALL call_0_7_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_14_13
#define init_0_7_14_13
void FASTCALL call_0_7_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_14_14
#define init_0_7_14_14
void FASTCALL call_0_7_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_14_15
#define init_0_7_14_15
void FASTCALL call_0_7_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_15_0
#define init_0_7_15_0
void FASTCALL call_0_7_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_15_1
#define init_0_7_15_1
void FASTCALL call_0_7_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_15_2
#define init_0_7_15_2
void FASTCALL call_0_7_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_15_3
#define init_0_7_15_3
void FASTCALL call_0_7_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_15_4
#define init_0_7_15_4
void FASTCALL call_0_7_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_15_5
#define init_0_7_15_5
void FASTCALL call_0_7_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_15_6
#define init_0_7_15_6
void FASTCALL call_0_7_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_15_7
#define init_0_7_15_7
void FASTCALL call_0_7_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_15_8
#define init_0_7_15_8
void FASTCALL call_0_7_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_15_9
#define init_0_7_15_9
void FASTCALL call_0_7_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_15_10
#define init_0_7_15_10
void FASTCALL call_0_7_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_15_11
#define init_0_7_15_11
void FASTCALL call_0_7_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_15_12
#define init_0_7_15_12
void FASTCALL call_0_7_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_15_13
#define init_0_7_15_13
void FASTCALL call_0_7_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_15_14
#define init_0_7_15_14
void FASTCALL call_0_7_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_7_15_15
#define init_0_7_15_15
void FASTCALL call_0_7_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_0_0
#define init_0_8_0_0
void FASTCALL call_0_8_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_0_1
#define init_0_8_0_1
void FASTCALL call_0_8_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_0_2
#define init_0_8_0_2
void FASTCALL call_0_8_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_0_3
#define init_0_8_0_3
void FASTCALL call_0_8_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_0_4
#define init_0_8_0_4
void FASTCALL call_0_8_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_0_5
#define init_0_8_0_5
void FASTCALL call_0_8_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_0_6
#define init_0_8_0_6
void FASTCALL call_0_8_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_0_7
#define init_0_8_0_7
void FASTCALL call_0_8_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_0_8
#define init_0_8_0_8
void FASTCALL call_0_8_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_0_9
#define init_0_8_0_9
void FASTCALL call_0_8_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_0_10
#define init_0_8_0_10
void FASTCALL call_0_8_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_0_11
#define init_0_8_0_11
void FASTCALL call_0_8_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_0_12
#define init_0_8_0_12
void FASTCALL call_0_8_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_0_13
#define init_0_8_0_13
void FASTCALL call_0_8_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_0_14
#define init_0_8_0_14
void FASTCALL call_0_8_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_0_15
#define init_0_8_0_15
void FASTCALL call_0_8_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_1_0
#define init_0_8_1_0
void FASTCALL call_0_8_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_1_1
#define init_0_8_1_1
void FASTCALL call_0_8_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_1_2
#define init_0_8_1_2
void FASTCALL call_0_8_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_1_3
#define init_0_8_1_3
void FASTCALL call_0_8_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_1_4
#define init_0_8_1_4
void FASTCALL call_0_8_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_1_5
#define init_0_8_1_5
void FASTCALL call_0_8_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_1_6
#define init_0_8_1_6
void FASTCALL call_0_8_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_1_7
#define init_0_8_1_7
void FASTCALL call_0_8_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_1_8
#define init_0_8_1_8
void FASTCALL call_0_8_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_1_9
#define init_0_8_1_9
void FASTCALL call_0_8_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_1_10
#define init_0_8_1_10
void FASTCALL call_0_8_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_1_11
#define init_0_8_1_11
void FASTCALL call_0_8_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_1_12
#define init_0_8_1_12
void FASTCALL call_0_8_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_1_13
#define init_0_8_1_13
void FASTCALL call_0_8_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_1_14
#define init_0_8_1_14
void FASTCALL call_0_8_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_1_15
#define init_0_8_1_15
void FASTCALL call_0_8_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_2_0
#define init_0_8_2_0
void FASTCALL call_0_8_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_2_1
#define init_0_8_2_1
void FASTCALL call_0_8_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_2_2
#define init_0_8_2_2
void FASTCALL call_0_8_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_2_3
#define init_0_8_2_3
void FASTCALL call_0_8_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_2_4
#define init_0_8_2_4
void FASTCALL call_0_8_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_2_5
#define init_0_8_2_5
void FASTCALL call_0_8_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_2_6
#define init_0_8_2_6
void FASTCALL call_0_8_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_2_7
#define init_0_8_2_7
void FASTCALL call_0_8_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_2_8
#define init_0_8_2_8
void FASTCALL call_0_8_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_2_9
#define init_0_8_2_9
void FASTCALL call_0_8_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_2_10
#define init_0_8_2_10
void FASTCALL call_0_8_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_2_11
#define init_0_8_2_11
void FASTCALL call_0_8_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_2_12
#define init_0_8_2_12
void FASTCALL call_0_8_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_2_13
#define init_0_8_2_13
void FASTCALL call_0_8_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_2_14
#define init_0_8_2_14
void FASTCALL call_0_8_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_2_15
#define init_0_8_2_15
void FASTCALL call_0_8_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_3_0
#define init_0_8_3_0
void FASTCALL call_0_8_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_3_1
#define init_0_8_3_1
void FASTCALL call_0_8_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_3_2
#define init_0_8_3_2
void FASTCALL call_0_8_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_3_3
#define init_0_8_3_3
void FASTCALL call_0_8_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_3_4
#define init_0_8_3_4
void FASTCALL call_0_8_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_3_5
#define init_0_8_3_5
void FASTCALL call_0_8_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_3_6
#define init_0_8_3_6
void FASTCALL call_0_8_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_3_7
#define init_0_8_3_7
void FASTCALL call_0_8_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_3_8
#define init_0_8_3_8
void FASTCALL call_0_8_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_3_9
#define init_0_8_3_9
void FASTCALL call_0_8_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_3_10
#define init_0_8_3_10
void FASTCALL call_0_8_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_3_11
#define init_0_8_3_11
void FASTCALL call_0_8_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_3_12
#define init_0_8_3_12
void FASTCALL call_0_8_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_3_13
#define init_0_8_3_13
void FASTCALL call_0_8_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_3_14
#define init_0_8_3_14
void FASTCALL call_0_8_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_3_15
#define init_0_8_3_15
void FASTCALL call_0_8_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_4_0
#define init_0_8_4_0
void FASTCALL call_0_8_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_4_1
#define init_0_8_4_1
void FASTCALL call_0_8_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_4_2
#define init_0_8_4_2
void FASTCALL call_0_8_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_4_3
#define init_0_8_4_3
void FASTCALL call_0_8_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_4_4
#define init_0_8_4_4
void FASTCALL call_0_8_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_4_5
#define init_0_8_4_5
void FASTCALL call_0_8_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_4_6
#define init_0_8_4_6
void FASTCALL call_0_8_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_4_7
#define init_0_8_4_7
void FASTCALL call_0_8_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_4_8
#define init_0_8_4_8
void FASTCALL call_0_8_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_4_9
#define init_0_8_4_9
void FASTCALL call_0_8_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_4_10
#define init_0_8_4_10
void FASTCALL call_0_8_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_4_11
#define init_0_8_4_11
void FASTCALL call_0_8_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_4_12
#define init_0_8_4_12
void FASTCALL call_0_8_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_4_13
#define init_0_8_4_13
void FASTCALL call_0_8_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_4_14
#define init_0_8_4_14
void FASTCALL call_0_8_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_4_15
#define init_0_8_4_15
void FASTCALL call_0_8_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_5_0
#define init_0_8_5_0
void FASTCALL call_0_8_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_5_1
#define init_0_8_5_1
void FASTCALL call_0_8_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_5_2
#define init_0_8_5_2
void FASTCALL call_0_8_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_5_3
#define init_0_8_5_3
void FASTCALL call_0_8_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_5_4
#define init_0_8_5_4
void FASTCALL call_0_8_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_5_5
#define init_0_8_5_5
void FASTCALL call_0_8_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_5_6
#define init_0_8_5_6
void FASTCALL call_0_8_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_5_7
#define init_0_8_5_7
void FASTCALL call_0_8_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_5_8
#define init_0_8_5_8
void FASTCALL call_0_8_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_5_9
#define init_0_8_5_9
void FASTCALL call_0_8_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_5_10
#define init_0_8_5_10
void FASTCALL call_0_8_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_5_11
#define init_0_8_5_11
void FASTCALL call_0_8_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_5_12
#define init_0_8_5_12
void FASTCALL call_0_8_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_5_13
#define init_0_8_5_13
void FASTCALL call_0_8_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_5_14
#define init_0_8_5_14
void FASTCALL call_0_8_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_5_15
#define init_0_8_5_15
void FASTCALL call_0_8_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_6_0
#define init_0_8_6_0
void FASTCALL call_0_8_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_6_1
#define init_0_8_6_1
void FASTCALL call_0_8_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_6_2
#define init_0_8_6_2
void FASTCALL call_0_8_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_6_3
#define init_0_8_6_3
void FASTCALL call_0_8_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_6_4
#define init_0_8_6_4
void FASTCALL call_0_8_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_6_5
#define init_0_8_6_5
void FASTCALL call_0_8_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_6_6
#define init_0_8_6_6
void FASTCALL call_0_8_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_6_7
#define init_0_8_6_7
void FASTCALL call_0_8_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_6_8
#define init_0_8_6_8
void FASTCALL call_0_8_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_6_9
#define init_0_8_6_9
void FASTCALL call_0_8_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_6_10
#define init_0_8_6_10
void FASTCALL call_0_8_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_6_11
#define init_0_8_6_11
void FASTCALL call_0_8_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_6_12
#define init_0_8_6_12
void FASTCALL call_0_8_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_6_13
#define init_0_8_6_13
void FASTCALL call_0_8_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_6_14
#define init_0_8_6_14
void FASTCALL call_0_8_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_6_15
#define init_0_8_6_15
void FASTCALL call_0_8_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_7_0
#define init_0_8_7_0
void FASTCALL call_0_8_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_7_1
#define init_0_8_7_1
void FASTCALL call_0_8_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_7_2
#define init_0_8_7_2
void FASTCALL call_0_8_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_7_3
#define init_0_8_7_3
void FASTCALL call_0_8_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_7_4
#define init_0_8_7_4
void FASTCALL call_0_8_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_7_5
#define init_0_8_7_5
void FASTCALL call_0_8_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_7_6
#define init_0_8_7_6
void FASTCALL call_0_8_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_7_7
#define init_0_8_7_7
void FASTCALL call_0_8_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_7_8
#define init_0_8_7_8
void FASTCALL call_0_8_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_7_9
#define init_0_8_7_9
void FASTCALL call_0_8_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_7_10
#define init_0_8_7_10
void FASTCALL call_0_8_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_7_11
#define init_0_8_7_11
void FASTCALL call_0_8_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_7_12
#define init_0_8_7_12
void FASTCALL call_0_8_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_7_13
#define init_0_8_7_13
void FASTCALL call_0_8_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_7_14
#define init_0_8_7_14
void FASTCALL call_0_8_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_7_15
#define init_0_8_7_15
void FASTCALL call_0_8_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_8_0
#define init_0_8_8_0
void FASTCALL call_0_8_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_8_1
#define init_0_8_8_1
void FASTCALL call_0_8_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_8_2
#define init_0_8_8_2
void FASTCALL call_0_8_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_8_3
#define init_0_8_8_3
void FASTCALL call_0_8_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_8_4
#define init_0_8_8_4
void FASTCALL call_0_8_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_8_5
#define init_0_8_8_5
void FASTCALL call_0_8_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_8_6
#define init_0_8_8_6
void FASTCALL call_0_8_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_8_7
#define init_0_8_8_7
void FASTCALL call_0_8_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_8_8
#define init_0_8_8_8
void FASTCALL call_0_8_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_8_9
#define init_0_8_8_9
void FASTCALL call_0_8_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_8_10
#define init_0_8_8_10
void FASTCALL call_0_8_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_8_11
#define init_0_8_8_11
void FASTCALL call_0_8_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_8_12
#define init_0_8_8_12
void FASTCALL call_0_8_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_8_13
#define init_0_8_8_13
void FASTCALL call_0_8_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_8_14
#define init_0_8_8_14
void FASTCALL call_0_8_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_8_15
#define init_0_8_8_15
void FASTCALL call_0_8_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_9_0
#define init_0_8_9_0
void FASTCALL call_0_8_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_9_1
#define init_0_8_9_1
void FASTCALL call_0_8_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_9_2
#define init_0_8_9_2
void FASTCALL call_0_8_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_9_3
#define init_0_8_9_3
void FASTCALL call_0_8_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_9_4
#define init_0_8_9_4
void FASTCALL call_0_8_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_9_5
#define init_0_8_9_5
void FASTCALL call_0_8_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_9_6
#define init_0_8_9_6
void FASTCALL call_0_8_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_9_7
#define init_0_8_9_7
void FASTCALL call_0_8_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_9_8
#define init_0_8_9_8
void FASTCALL call_0_8_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_9_9
#define init_0_8_9_9
void FASTCALL call_0_8_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_9_10
#define init_0_8_9_10
void FASTCALL call_0_8_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_9_11
#define init_0_8_9_11
void FASTCALL call_0_8_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_9_12
#define init_0_8_9_12
void FASTCALL call_0_8_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_9_13
#define init_0_8_9_13
void FASTCALL call_0_8_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_9_14
#define init_0_8_9_14
void FASTCALL call_0_8_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_9_15
#define init_0_8_9_15
void FASTCALL call_0_8_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_10_0
#define init_0_8_10_0
void FASTCALL call_0_8_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_10_1
#define init_0_8_10_1
void FASTCALL call_0_8_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_10_2
#define init_0_8_10_2
void FASTCALL call_0_8_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_10_3
#define init_0_8_10_3
void FASTCALL call_0_8_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_10_4
#define init_0_8_10_4
void FASTCALL call_0_8_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_10_5
#define init_0_8_10_5
void FASTCALL call_0_8_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_10_6
#define init_0_8_10_6
void FASTCALL call_0_8_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_10_7
#define init_0_8_10_7
void FASTCALL call_0_8_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_10_8
#define init_0_8_10_8
void FASTCALL call_0_8_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_10_9
#define init_0_8_10_9
void FASTCALL call_0_8_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_10_10
#define init_0_8_10_10
void FASTCALL call_0_8_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_10_11
#define init_0_8_10_11
void FASTCALL call_0_8_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_10_12
#define init_0_8_10_12
void FASTCALL call_0_8_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_10_13
#define init_0_8_10_13
void FASTCALL call_0_8_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_10_14
#define init_0_8_10_14
void FASTCALL call_0_8_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_10_15
#define init_0_8_10_15
void FASTCALL call_0_8_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_11_0
#define init_0_8_11_0
void FASTCALL call_0_8_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_11_1
#define init_0_8_11_1
void FASTCALL call_0_8_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_11_2
#define init_0_8_11_2
void FASTCALL call_0_8_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_11_3
#define init_0_8_11_3
void FASTCALL call_0_8_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_11_4
#define init_0_8_11_4
void FASTCALL call_0_8_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_11_5
#define init_0_8_11_5
void FASTCALL call_0_8_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_11_6
#define init_0_8_11_6
void FASTCALL call_0_8_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_11_7
#define init_0_8_11_7
void FASTCALL call_0_8_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_11_8
#define init_0_8_11_8
void FASTCALL call_0_8_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_11_9
#define init_0_8_11_9
void FASTCALL call_0_8_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_11_10
#define init_0_8_11_10
void FASTCALL call_0_8_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_11_11
#define init_0_8_11_11
void FASTCALL call_0_8_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_11_12
#define init_0_8_11_12
void FASTCALL call_0_8_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_11_13
#define init_0_8_11_13
void FASTCALL call_0_8_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_11_14
#define init_0_8_11_14
void FASTCALL call_0_8_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_11_15
#define init_0_8_11_15
void FASTCALL call_0_8_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_12_0
#define init_0_8_12_0
void FASTCALL call_0_8_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_12_1
#define init_0_8_12_1
void FASTCALL call_0_8_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_12_2
#define init_0_8_12_2
void FASTCALL call_0_8_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_12_3
#define init_0_8_12_3
void FASTCALL call_0_8_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_12_4
#define init_0_8_12_4
void FASTCALL call_0_8_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_12_5
#define init_0_8_12_5
void FASTCALL call_0_8_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_12_6
#define init_0_8_12_6
void FASTCALL call_0_8_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_12_7
#define init_0_8_12_7
void FASTCALL call_0_8_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_12_8
#define init_0_8_12_8
void FASTCALL call_0_8_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_12_9
#define init_0_8_12_9
void FASTCALL call_0_8_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_12_10
#define init_0_8_12_10
void FASTCALL call_0_8_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_12_11
#define init_0_8_12_11
void FASTCALL call_0_8_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_12_12
#define init_0_8_12_12
void FASTCALL call_0_8_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_12_13
#define init_0_8_12_13
void FASTCALL call_0_8_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_12_14
#define init_0_8_12_14
void FASTCALL call_0_8_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_12_15
#define init_0_8_12_15
void FASTCALL call_0_8_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_13_0
#define init_0_8_13_0
void FASTCALL call_0_8_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_13_1
#define init_0_8_13_1
void FASTCALL call_0_8_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_13_2
#define init_0_8_13_2
void FASTCALL call_0_8_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_13_3
#define init_0_8_13_3
void FASTCALL call_0_8_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_13_4
#define init_0_8_13_4
void FASTCALL call_0_8_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_13_5
#define init_0_8_13_5
void FASTCALL call_0_8_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_13_6
#define init_0_8_13_6
void FASTCALL call_0_8_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_13_7
#define init_0_8_13_7
void FASTCALL call_0_8_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_13_8
#define init_0_8_13_8
void FASTCALL call_0_8_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_13_9
#define init_0_8_13_9
void FASTCALL call_0_8_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_13_10
#define init_0_8_13_10
void FASTCALL call_0_8_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_13_11
#define init_0_8_13_11
void FASTCALL call_0_8_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_13_12
#define init_0_8_13_12
void FASTCALL call_0_8_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_13_13
#define init_0_8_13_13
void FASTCALL call_0_8_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_13_14
#define init_0_8_13_14
void FASTCALL call_0_8_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_13_15
#define init_0_8_13_15
void FASTCALL call_0_8_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_14_0
#define init_0_8_14_0
void FASTCALL call_0_8_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_14_1
#define init_0_8_14_1
void FASTCALL call_0_8_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_14_2
#define init_0_8_14_2
void FASTCALL call_0_8_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_14_3
#define init_0_8_14_3
void FASTCALL call_0_8_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_14_4
#define init_0_8_14_4
void FASTCALL call_0_8_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_14_5
#define init_0_8_14_5
void FASTCALL call_0_8_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_14_6
#define init_0_8_14_6
void FASTCALL call_0_8_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_14_7
#define init_0_8_14_7
void FASTCALL call_0_8_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_14_8
#define init_0_8_14_8
void FASTCALL call_0_8_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_14_9
#define init_0_8_14_9
void FASTCALL call_0_8_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_14_10
#define init_0_8_14_10
void FASTCALL call_0_8_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_14_11
#define init_0_8_14_11
void FASTCALL call_0_8_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_14_12
#define init_0_8_14_12
void FASTCALL call_0_8_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_14_13
#define init_0_8_14_13
void FASTCALL call_0_8_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_14_14
#define init_0_8_14_14
void FASTCALL call_0_8_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_14_15
#define init_0_8_14_15
void FASTCALL call_0_8_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_15_0
#define init_0_8_15_0
void FASTCALL call_0_8_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_15_1
#define init_0_8_15_1
void FASTCALL call_0_8_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_15_2
#define init_0_8_15_2
void FASTCALL call_0_8_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_15_3
#define init_0_8_15_3
void FASTCALL call_0_8_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_15_4
#define init_0_8_15_4
void FASTCALL call_0_8_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_15_5
#define init_0_8_15_5
void FASTCALL call_0_8_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_15_6
#define init_0_8_15_6
void FASTCALL call_0_8_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_15_7
#define init_0_8_15_7
void FASTCALL call_0_8_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_15_8
#define init_0_8_15_8
void FASTCALL call_0_8_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_15_9
#define init_0_8_15_9
void FASTCALL call_0_8_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_15_10
#define init_0_8_15_10
void FASTCALL call_0_8_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_15_11
#define init_0_8_15_11
void FASTCALL call_0_8_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_15_12
#define init_0_8_15_12
void FASTCALL call_0_8_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_15_13
#define init_0_8_15_13
void FASTCALL call_0_8_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_15_14
#define init_0_8_15_14
void FASTCALL call_0_8_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_8_15_15
#define init_0_8_15_15
void FASTCALL call_0_8_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_0_0
#define init_0_9_0_0
void FASTCALL call_0_9_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_0_1
#define init_0_9_0_1
void FASTCALL call_0_9_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_0_2
#define init_0_9_0_2
void FASTCALL call_0_9_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_0_3
#define init_0_9_0_3
void FASTCALL call_0_9_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_0_4
#define init_0_9_0_4
void FASTCALL call_0_9_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_0_5
#define init_0_9_0_5
void FASTCALL call_0_9_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_0_6
#define init_0_9_0_6
void FASTCALL call_0_9_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_0_7
#define init_0_9_0_7
void FASTCALL call_0_9_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_0_8
#define init_0_9_0_8
void FASTCALL call_0_9_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_0_9
#define init_0_9_0_9
void FASTCALL call_0_9_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_0_10
#define init_0_9_0_10
void FASTCALL call_0_9_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_0_11
#define init_0_9_0_11
void FASTCALL call_0_9_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_0_12
#define init_0_9_0_12
void FASTCALL call_0_9_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_0_13
#define init_0_9_0_13
void FASTCALL call_0_9_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_0_14
#define init_0_9_0_14
void FASTCALL call_0_9_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_0_15
#define init_0_9_0_15
void FASTCALL call_0_9_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_1_0
#define init_0_9_1_0
void FASTCALL call_0_9_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_1_1
#define init_0_9_1_1
void FASTCALL call_0_9_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_1_2
#define init_0_9_1_2
void FASTCALL call_0_9_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_1_3
#define init_0_9_1_3
void FASTCALL call_0_9_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_1_4
#define init_0_9_1_4
void FASTCALL call_0_9_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_1_5
#define init_0_9_1_5
void FASTCALL call_0_9_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_1_6
#define init_0_9_1_6
void FASTCALL call_0_9_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_1_7
#define init_0_9_1_7
void FASTCALL call_0_9_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_1_8
#define init_0_9_1_8
void FASTCALL call_0_9_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_1_9
#define init_0_9_1_9
void FASTCALL call_0_9_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_1_10
#define init_0_9_1_10
void FASTCALL call_0_9_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_1_11
#define init_0_9_1_11
void FASTCALL call_0_9_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_1_12
#define init_0_9_1_12
void FASTCALL call_0_9_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_1_13
#define init_0_9_1_13
void FASTCALL call_0_9_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_1_14
#define init_0_9_1_14
void FASTCALL call_0_9_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_1_15
#define init_0_9_1_15
void FASTCALL call_0_9_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_2_0
#define init_0_9_2_0
void FASTCALL call_0_9_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_2_1
#define init_0_9_2_1
void FASTCALL call_0_9_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_2_2
#define init_0_9_2_2
void FASTCALL call_0_9_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_2_3
#define init_0_9_2_3
void FASTCALL call_0_9_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_2_4
#define init_0_9_2_4
void FASTCALL call_0_9_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_2_5
#define init_0_9_2_5
void FASTCALL call_0_9_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_2_6
#define init_0_9_2_6
void FASTCALL call_0_9_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_2_7
#define init_0_9_2_7
void FASTCALL call_0_9_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_2_8
#define init_0_9_2_8
void FASTCALL call_0_9_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_2_9
#define init_0_9_2_9
void FASTCALL call_0_9_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_2_10
#define init_0_9_2_10
void FASTCALL call_0_9_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_2_11
#define init_0_9_2_11
void FASTCALL call_0_9_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_2_12
#define init_0_9_2_12
void FASTCALL call_0_9_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_2_13
#define init_0_9_2_13
void FASTCALL call_0_9_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_2_14
#define init_0_9_2_14
void FASTCALL call_0_9_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_2_15
#define init_0_9_2_15
void FASTCALL call_0_9_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_3_0
#define init_0_9_3_0
void FASTCALL call_0_9_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_3_1
#define init_0_9_3_1
void FASTCALL call_0_9_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_3_2
#define init_0_9_3_2
void FASTCALL call_0_9_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_3_3
#define init_0_9_3_3
void FASTCALL call_0_9_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_3_4
#define init_0_9_3_4
void FASTCALL call_0_9_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_3_5
#define init_0_9_3_5
void FASTCALL call_0_9_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_3_6
#define init_0_9_3_6
void FASTCALL call_0_9_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_3_7
#define init_0_9_3_7
void FASTCALL call_0_9_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_3_8
#define init_0_9_3_8
void FASTCALL call_0_9_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_3_9
#define init_0_9_3_9
void FASTCALL call_0_9_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_3_10
#define init_0_9_3_10
void FASTCALL call_0_9_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_3_11
#define init_0_9_3_11
void FASTCALL call_0_9_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_3_12
#define init_0_9_3_12
void FASTCALL call_0_9_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_3_13
#define init_0_9_3_13
void FASTCALL call_0_9_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_3_14
#define init_0_9_3_14
void FASTCALL call_0_9_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_3_15
#define init_0_9_3_15
void FASTCALL call_0_9_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_4_0
#define init_0_9_4_0
void FASTCALL call_0_9_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_4_1
#define init_0_9_4_1
void FASTCALL call_0_9_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_4_2
#define init_0_9_4_2
void FASTCALL call_0_9_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_4_3
#define init_0_9_4_3
void FASTCALL call_0_9_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_4_4
#define init_0_9_4_4
void FASTCALL call_0_9_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_4_5
#define init_0_9_4_5
void FASTCALL call_0_9_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_4_6
#define init_0_9_4_6
void FASTCALL call_0_9_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_4_7
#define init_0_9_4_7
void FASTCALL call_0_9_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_4_8
#define init_0_9_4_8
void FASTCALL call_0_9_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_4_9
#define init_0_9_4_9
void FASTCALL call_0_9_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_4_10
#define init_0_9_4_10
void FASTCALL call_0_9_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_4_11
#define init_0_9_4_11
void FASTCALL call_0_9_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_4_12
#define init_0_9_4_12
void FASTCALL call_0_9_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_4_13
#define init_0_9_4_13
void FASTCALL call_0_9_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_4_14
#define init_0_9_4_14
void FASTCALL call_0_9_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_4_15
#define init_0_9_4_15
void FASTCALL call_0_9_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_5_0
#define init_0_9_5_0
void FASTCALL call_0_9_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_5_1
#define init_0_9_5_1
void FASTCALL call_0_9_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_5_2
#define init_0_9_5_2
void FASTCALL call_0_9_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_5_3
#define init_0_9_5_3
void FASTCALL call_0_9_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_5_4
#define init_0_9_5_4
void FASTCALL call_0_9_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_5_5
#define init_0_9_5_5
void FASTCALL call_0_9_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_5_6
#define init_0_9_5_6
void FASTCALL call_0_9_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_5_7
#define init_0_9_5_7
void FASTCALL call_0_9_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_5_8
#define init_0_9_5_8
void FASTCALL call_0_9_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_5_9
#define init_0_9_5_9
void FASTCALL call_0_9_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_5_10
#define init_0_9_5_10
void FASTCALL call_0_9_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_5_11
#define init_0_9_5_11
void FASTCALL call_0_9_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_5_12
#define init_0_9_5_12
void FASTCALL call_0_9_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_5_13
#define init_0_9_5_13
void FASTCALL call_0_9_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_5_14
#define init_0_9_5_14
void FASTCALL call_0_9_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_5_15
#define init_0_9_5_15
void FASTCALL call_0_9_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_6_0
#define init_0_9_6_0
void FASTCALL call_0_9_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_6_1
#define init_0_9_6_1
void FASTCALL call_0_9_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_6_2
#define init_0_9_6_2
void FASTCALL call_0_9_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_6_3
#define init_0_9_6_3
void FASTCALL call_0_9_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_6_4
#define init_0_9_6_4
void FASTCALL call_0_9_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_6_5
#define init_0_9_6_5
void FASTCALL call_0_9_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_6_6
#define init_0_9_6_6
void FASTCALL call_0_9_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_6_7
#define init_0_9_6_7
void FASTCALL call_0_9_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_6_8
#define init_0_9_6_8
void FASTCALL call_0_9_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_6_9
#define init_0_9_6_9
void FASTCALL call_0_9_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_6_10
#define init_0_9_6_10
void FASTCALL call_0_9_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_6_11
#define init_0_9_6_11
void FASTCALL call_0_9_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_6_12
#define init_0_9_6_12
void FASTCALL call_0_9_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_6_13
#define init_0_9_6_13
void FASTCALL call_0_9_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_6_14
#define init_0_9_6_14
void FASTCALL call_0_9_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_6_15
#define init_0_9_6_15
void FASTCALL call_0_9_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_7_0
#define init_0_9_7_0
void FASTCALL call_0_9_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_7_1
#define init_0_9_7_1
void FASTCALL call_0_9_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_7_2
#define init_0_9_7_2
void FASTCALL call_0_9_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_7_3
#define init_0_9_7_3
void FASTCALL call_0_9_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_7_4
#define init_0_9_7_4
void FASTCALL call_0_9_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_7_5
#define init_0_9_7_5
void FASTCALL call_0_9_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_7_6
#define init_0_9_7_6
void FASTCALL call_0_9_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_7_7
#define init_0_9_7_7
void FASTCALL call_0_9_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_7_8
#define init_0_9_7_8
void FASTCALL call_0_9_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_7_9
#define init_0_9_7_9
void FASTCALL call_0_9_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_7_10
#define init_0_9_7_10
void FASTCALL call_0_9_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_7_11
#define init_0_9_7_11
void FASTCALL call_0_9_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_7_12
#define init_0_9_7_12
void FASTCALL call_0_9_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_7_13
#define init_0_9_7_13
void FASTCALL call_0_9_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_7_14
#define init_0_9_7_14
void FASTCALL call_0_9_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_7_15
#define init_0_9_7_15
void FASTCALL call_0_9_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_8_0
#define init_0_9_8_0
void FASTCALL call_0_9_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_8_1
#define init_0_9_8_1
void FASTCALL call_0_9_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_8_2
#define init_0_9_8_2
void FASTCALL call_0_9_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_8_3
#define init_0_9_8_3
void FASTCALL call_0_9_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_8_4
#define init_0_9_8_4
void FASTCALL call_0_9_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_8_5
#define init_0_9_8_5
void FASTCALL call_0_9_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_8_6
#define init_0_9_8_6
void FASTCALL call_0_9_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_8_7
#define init_0_9_8_7
void FASTCALL call_0_9_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_8_8
#define init_0_9_8_8
void FASTCALL call_0_9_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_8_9
#define init_0_9_8_9
void FASTCALL call_0_9_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_8_10
#define init_0_9_8_10
void FASTCALL call_0_9_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_8_11
#define init_0_9_8_11
void FASTCALL call_0_9_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_8_12
#define init_0_9_8_12
void FASTCALL call_0_9_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_8_13
#define init_0_9_8_13
void FASTCALL call_0_9_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_8_14
#define init_0_9_8_14
void FASTCALL call_0_9_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_8_15
#define init_0_9_8_15
void FASTCALL call_0_9_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_9_0
#define init_0_9_9_0
void FASTCALL call_0_9_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_9_1
#define init_0_9_9_1
void FASTCALL call_0_9_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_9_2
#define init_0_9_9_2
void FASTCALL call_0_9_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_9_3
#define init_0_9_9_3
void FASTCALL call_0_9_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_9_4
#define init_0_9_9_4
void FASTCALL call_0_9_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_9_5
#define init_0_9_9_5
void FASTCALL call_0_9_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_9_6
#define init_0_9_9_6
void FASTCALL call_0_9_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_9_7
#define init_0_9_9_7
void FASTCALL call_0_9_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_9_8
#define init_0_9_9_8
void FASTCALL call_0_9_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_9_9
#define init_0_9_9_9
void FASTCALL call_0_9_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_9_10
#define init_0_9_9_10
void FASTCALL call_0_9_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_9_11
#define init_0_9_9_11
void FASTCALL call_0_9_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_9_12
#define init_0_9_9_12
void FASTCALL call_0_9_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_9_13
#define init_0_9_9_13
void FASTCALL call_0_9_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_9_14
#define init_0_9_9_14
void FASTCALL call_0_9_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_9_15
#define init_0_9_9_15
void FASTCALL call_0_9_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_10_0
#define init_0_9_10_0
void FASTCALL call_0_9_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_10_1
#define init_0_9_10_1
void FASTCALL call_0_9_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_10_2
#define init_0_9_10_2
void FASTCALL call_0_9_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_10_3
#define init_0_9_10_3
void FASTCALL call_0_9_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_10_4
#define init_0_9_10_4
void FASTCALL call_0_9_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_10_5
#define init_0_9_10_5
void FASTCALL call_0_9_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_10_6
#define init_0_9_10_6
void FASTCALL call_0_9_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_10_7
#define init_0_9_10_7
void FASTCALL call_0_9_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_10_8
#define init_0_9_10_8
void FASTCALL call_0_9_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_10_9
#define init_0_9_10_9
void FASTCALL call_0_9_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_10_10
#define init_0_9_10_10
void FASTCALL call_0_9_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_10_11
#define init_0_9_10_11
void FASTCALL call_0_9_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_10_12
#define init_0_9_10_12
void FASTCALL call_0_9_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_10_13
#define init_0_9_10_13
void FASTCALL call_0_9_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_10_14
#define init_0_9_10_14
void FASTCALL call_0_9_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_10_15
#define init_0_9_10_15
void FASTCALL call_0_9_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_11_0
#define init_0_9_11_0
void FASTCALL call_0_9_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_11_1
#define init_0_9_11_1
void FASTCALL call_0_9_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_11_2
#define init_0_9_11_2
void FASTCALL call_0_9_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_11_3
#define init_0_9_11_3
void FASTCALL call_0_9_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_11_4
#define init_0_9_11_4
void FASTCALL call_0_9_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_11_5
#define init_0_9_11_5
void FASTCALL call_0_9_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_11_6
#define init_0_9_11_6
void FASTCALL call_0_9_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_11_7
#define init_0_9_11_7
void FASTCALL call_0_9_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_11_8
#define init_0_9_11_8
void FASTCALL call_0_9_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_11_9
#define init_0_9_11_9
void FASTCALL call_0_9_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_11_10
#define init_0_9_11_10
void FASTCALL call_0_9_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_11_11
#define init_0_9_11_11
void FASTCALL call_0_9_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_11_12
#define init_0_9_11_12
void FASTCALL call_0_9_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_11_13
#define init_0_9_11_13
void FASTCALL call_0_9_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_11_14
#define init_0_9_11_14
void FASTCALL call_0_9_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_11_15
#define init_0_9_11_15
void FASTCALL call_0_9_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_12_0
#define init_0_9_12_0
void FASTCALL call_0_9_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_12_1
#define init_0_9_12_1
void FASTCALL call_0_9_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_12_2
#define init_0_9_12_2
void FASTCALL call_0_9_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_12_3
#define init_0_9_12_3
void FASTCALL call_0_9_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_12_4
#define init_0_9_12_4
void FASTCALL call_0_9_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_12_5
#define init_0_9_12_5
void FASTCALL call_0_9_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_12_6
#define init_0_9_12_6
void FASTCALL call_0_9_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_12_7
#define init_0_9_12_7
void FASTCALL call_0_9_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_12_8
#define init_0_9_12_8
void FASTCALL call_0_9_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_12_9
#define init_0_9_12_9
void FASTCALL call_0_9_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_12_10
#define init_0_9_12_10
void FASTCALL call_0_9_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_12_11
#define init_0_9_12_11
void FASTCALL call_0_9_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_12_12
#define init_0_9_12_12
void FASTCALL call_0_9_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_12_13
#define init_0_9_12_13
void FASTCALL call_0_9_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_12_14
#define init_0_9_12_14
void FASTCALL call_0_9_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_12_15
#define init_0_9_12_15
void FASTCALL call_0_9_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_13_0
#define init_0_9_13_0
void FASTCALL call_0_9_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_13_1
#define init_0_9_13_1
void FASTCALL call_0_9_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_13_2
#define init_0_9_13_2
void FASTCALL call_0_9_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_13_3
#define init_0_9_13_3
void FASTCALL call_0_9_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_13_4
#define init_0_9_13_4
void FASTCALL call_0_9_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_13_5
#define init_0_9_13_5
void FASTCALL call_0_9_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_13_6
#define init_0_9_13_6
void FASTCALL call_0_9_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_13_7
#define init_0_9_13_7
void FASTCALL call_0_9_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_13_8
#define init_0_9_13_8
void FASTCALL call_0_9_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_13_9
#define init_0_9_13_9
void FASTCALL call_0_9_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_13_10
#define init_0_9_13_10
void FASTCALL call_0_9_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_13_11
#define init_0_9_13_11
void FASTCALL call_0_9_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_13_12
#define init_0_9_13_12
void FASTCALL call_0_9_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_13_13
#define init_0_9_13_13
void FASTCALL call_0_9_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_13_14
#define init_0_9_13_14
void FASTCALL call_0_9_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_13_15
#define init_0_9_13_15
void FASTCALL call_0_9_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_14_0
#define init_0_9_14_0
void FASTCALL call_0_9_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_14_1
#define init_0_9_14_1
void FASTCALL call_0_9_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_14_2
#define init_0_9_14_2
void FASTCALL call_0_9_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_14_3
#define init_0_9_14_3
void FASTCALL call_0_9_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_14_4
#define init_0_9_14_4
void FASTCALL call_0_9_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_14_5
#define init_0_9_14_5
void FASTCALL call_0_9_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_14_6
#define init_0_9_14_6
void FASTCALL call_0_9_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_14_7
#define init_0_9_14_7
void FASTCALL call_0_9_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_14_8
#define init_0_9_14_8
void FASTCALL call_0_9_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_14_9
#define init_0_9_14_9
void FASTCALL call_0_9_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_14_10
#define init_0_9_14_10
void FASTCALL call_0_9_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_14_11
#define init_0_9_14_11
void FASTCALL call_0_9_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_14_12
#define init_0_9_14_12
void FASTCALL call_0_9_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_14_13
#define init_0_9_14_13
void FASTCALL call_0_9_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_14_14
#define init_0_9_14_14
void FASTCALL call_0_9_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_14_15
#define init_0_9_14_15
void FASTCALL call_0_9_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_15_0
#define init_0_9_15_0
void FASTCALL call_0_9_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_15_1
#define init_0_9_15_1
void FASTCALL call_0_9_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_15_2
#define init_0_9_15_2
void FASTCALL call_0_9_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_15_3
#define init_0_9_15_3
void FASTCALL call_0_9_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_15_4
#define init_0_9_15_4
void FASTCALL call_0_9_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_15_5
#define init_0_9_15_5
void FASTCALL call_0_9_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_15_6
#define init_0_9_15_6
void FASTCALL call_0_9_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_15_7
#define init_0_9_15_7
void FASTCALL call_0_9_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_15_8
#define init_0_9_15_8
void FASTCALL call_0_9_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_15_9
#define init_0_9_15_9
void FASTCALL call_0_9_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_15_10
#define init_0_9_15_10
void FASTCALL call_0_9_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_15_11
#define init_0_9_15_11
void FASTCALL call_0_9_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_15_12
#define init_0_9_15_12
void FASTCALL call_0_9_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_15_13
#define init_0_9_15_13
void FASTCALL call_0_9_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_15_14
#define init_0_9_15_14
void FASTCALL call_0_9_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_9_15_15
#define init_0_9_15_15
void FASTCALL call_0_9_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_0_0
#define init_0_10_0_0
void FASTCALL call_0_10_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_0_1
#define init_0_10_0_1
void FASTCALL call_0_10_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_0_2
#define init_0_10_0_2
void FASTCALL call_0_10_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_0_3
#define init_0_10_0_3
void FASTCALL call_0_10_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_0_4
#define init_0_10_0_4
void FASTCALL call_0_10_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_0_5
#define init_0_10_0_5
void FASTCALL call_0_10_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_0_6
#define init_0_10_0_6
void FASTCALL call_0_10_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_0_7
#define init_0_10_0_7
void FASTCALL call_0_10_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_0_8
#define init_0_10_0_8
void FASTCALL call_0_10_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_0_9
#define init_0_10_0_9
void FASTCALL call_0_10_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_0_10
#define init_0_10_0_10
void FASTCALL call_0_10_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_0_11
#define init_0_10_0_11
void FASTCALL call_0_10_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_0_12
#define init_0_10_0_12
void FASTCALL call_0_10_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_0_13
#define init_0_10_0_13
void FASTCALL call_0_10_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_0_14
#define init_0_10_0_14
void FASTCALL call_0_10_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_0_15
#define init_0_10_0_15
void FASTCALL call_0_10_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_1_0
#define init_0_10_1_0
void FASTCALL call_0_10_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_1_1
#define init_0_10_1_1
void FASTCALL call_0_10_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_1_2
#define init_0_10_1_2
void FASTCALL call_0_10_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_1_3
#define init_0_10_1_3
void FASTCALL call_0_10_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_1_4
#define init_0_10_1_4
void FASTCALL call_0_10_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_1_5
#define init_0_10_1_5
void FASTCALL call_0_10_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_1_6
#define init_0_10_1_6
void FASTCALL call_0_10_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_1_7
#define init_0_10_1_7
void FASTCALL call_0_10_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_1_8
#define init_0_10_1_8
void FASTCALL call_0_10_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_1_9
#define init_0_10_1_9
void FASTCALL call_0_10_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_1_10
#define init_0_10_1_10
void FASTCALL call_0_10_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_1_11
#define init_0_10_1_11
void FASTCALL call_0_10_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_1_12
#define init_0_10_1_12
void FASTCALL call_0_10_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_1_13
#define init_0_10_1_13
void FASTCALL call_0_10_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_1_14
#define init_0_10_1_14
void FASTCALL call_0_10_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_1_15
#define init_0_10_1_15
void FASTCALL call_0_10_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_2_0
#define init_0_10_2_0
void FASTCALL call_0_10_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_2_1
#define init_0_10_2_1
void FASTCALL call_0_10_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_2_2
#define init_0_10_2_2
void FASTCALL call_0_10_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_2_3
#define init_0_10_2_3
void FASTCALL call_0_10_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_2_4
#define init_0_10_2_4
void FASTCALL call_0_10_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_2_5
#define init_0_10_2_5
void FASTCALL call_0_10_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_2_6
#define init_0_10_2_6
void FASTCALL call_0_10_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_2_7
#define init_0_10_2_7
void FASTCALL call_0_10_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_2_8
#define init_0_10_2_8
void FASTCALL call_0_10_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_2_9
#define init_0_10_2_9
void FASTCALL call_0_10_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_2_10
#define init_0_10_2_10
void FASTCALL call_0_10_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_2_11
#define init_0_10_2_11
void FASTCALL call_0_10_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_2_12
#define init_0_10_2_12
void FASTCALL call_0_10_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_2_13
#define init_0_10_2_13
void FASTCALL call_0_10_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_2_14
#define init_0_10_2_14
void FASTCALL call_0_10_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_2_15
#define init_0_10_2_15
void FASTCALL call_0_10_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_3_0
#define init_0_10_3_0
void FASTCALL call_0_10_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_3_1
#define init_0_10_3_1
void FASTCALL call_0_10_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_3_2
#define init_0_10_3_2
void FASTCALL call_0_10_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_3_3
#define init_0_10_3_3
void FASTCALL call_0_10_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_3_4
#define init_0_10_3_4
void FASTCALL call_0_10_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_3_5
#define init_0_10_3_5
void FASTCALL call_0_10_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_3_6
#define init_0_10_3_6
void FASTCALL call_0_10_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_3_7
#define init_0_10_3_7
void FASTCALL call_0_10_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_3_8
#define init_0_10_3_8
void FASTCALL call_0_10_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_3_9
#define init_0_10_3_9
void FASTCALL call_0_10_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_3_10
#define init_0_10_3_10
void FASTCALL call_0_10_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_3_11
#define init_0_10_3_11
void FASTCALL call_0_10_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_3_12
#define init_0_10_3_12
void FASTCALL call_0_10_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_3_13
#define init_0_10_3_13
void FASTCALL call_0_10_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_3_14
#define init_0_10_3_14
void FASTCALL call_0_10_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_3_15
#define init_0_10_3_15
void FASTCALL call_0_10_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_4_0
#define init_0_10_4_0
void FASTCALL call_0_10_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_4_1
#define init_0_10_4_1
void FASTCALL call_0_10_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_4_2
#define init_0_10_4_2
void FASTCALL call_0_10_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_4_3
#define init_0_10_4_3
void FASTCALL call_0_10_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_4_4
#define init_0_10_4_4
void FASTCALL call_0_10_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_4_5
#define init_0_10_4_5
void FASTCALL call_0_10_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_4_6
#define init_0_10_4_6
void FASTCALL call_0_10_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_4_7
#define init_0_10_4_7
void FASTCALL call_0_10_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_4_8
#define init_0_10_4_8
void FASTCALL call_0_10_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_4_9
#define init_0_10_4_9
void FASTCALL call_0_10_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_4_10
#define init_0_10_4_10
void FASTCALL call_0_10_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_4_11
#define init_0_10_4_11
void FASTCALL call_0_10_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_4_12
#define init_0_10_4_12
void FASTCALL call_0_10_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_4_13
#define init_0_10_4_13
void FASTCALL call_0_10_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_4_14
#define init_0_10_4_14
void FASTCALL call_0_10_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_4_15
#define init_0_10_4_15
void FASTCALL call_0_10_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_5_0
#define init_0_10_5_0
void FASTCALL call_0_10_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_5_1
#define init_0_10_5_1
void FASTCALL call_0_10_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_5_2
#define init_0_10_5_2
void FASTCALL call_0_10_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_5_3
#define init_0_10_5_3
void FASTCALL call_0_10_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_5_4
#define init_0_10_5_4
void FASTCALL call_0_10_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_5_5
#define init_0_10_5_5
void FASTCALL call_0_10_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_5_6
#define init_0_10_5_6
void FASTCALL call_0_10_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_5_7
#define init_0_10_5_7
void FASTCALL call_0_10_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_5_8
#define init_0_10_5_8
void FASTCALL call_0_10_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_5_9
#define init_0_10_5_9
void FASTCALL call_0_10_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_5_10
#define init_0_10_5_10
void FASTCALL call_0_10_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_5_11
#define init_0_10_5_11
void FASTCALL call_0_10_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_5_12
#define init_0_10_5_12
void FASTCALL call_0_10_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_5_13
#define init_0_10_5_13
void FASTCALL call_0_10_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_5_14
#define init_0_10_5_14
void FASTCALL call_0_10_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_5_15
#define init_0_10_5_15
void FASTCALL call_0_10_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_6_0
#define init_0_10_6_0
void FASTCALL call_0_10_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_6_1
#define init_0_10_6_1
void FASTCALL call_0_10_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_6_2
#define init_0_10_6_2
void FASTCALL call_0_10_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_6_3
#define init_0_10_6_3
void FASTCALL call_0_10_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_6_4
#define init_0_10_6_4
void FASTCALL call_0_10_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_6_5
#define init_0_10_6_5
void FASTCALL call_0_10_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_6_6
#define init_0_10_6_6
void FASTCALL call_0_10_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_6_7
#define init_0_10_6_7
void FASTCALL call_0_10_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_6_8
#define init_0_10_6_8
void FASTCALL call_0_10_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_6_9
#define init_0_10_6_9
void FASTCALL call_0_10_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_6_10
#define init_0_10_6_10
void FASTCALL call_0_10_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_6_11
#define init_0_10_6_11
void FASTCALL call_0_10_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_6_12
#define init_0_10_6_12
void FASTCALL call_0_10_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_6_13
#define init_0_10_6_13
void FASTCALL call_0_10_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_6_14
#define init_0_10_6_14
void FASTCALL call_0_10_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_6_15
#define init_0_10_6_15
void FASTCALL call_0_10_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_7_0
#define init_0_10_7_0
void FASTCALL call_0_10_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_7_1
#define init_0_10_7_1
void FASTCALL call_0_10_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_7_2
#define init_0_10_7_2
void FASTCALL call_0_10_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_7_3
#define init_0_10_7_3
void FASTCALL call_0_10_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_7_4
#define init_0_10_7_4
void FASTCALL call_0_10_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_7_5
#define init_0_10_7_5
void FASTCALL call_0_10_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_7_6
#define init_0_10_7_6
void FASTCALL call_0_10_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_7_7
#define init_0_10_7_7
void FASTCALL call_0_10_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_7_8
#define init_0_10_7_8
void FASTCALL call_0_10_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_7_9
#define init_0_10_7_9
void FASTCALL call_0_10_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_7_10
#define init_0_10_7_10
void FASTCALL call_0_10_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_7_11
#define init_0_10_7_11
void FASTCALL call_0_10_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_7_12
#define init_0_10_7_12
void FASTCALL call_0_10_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_7_13
#define init_0_10_7_13
void FASTCALL call_0_10_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_7_14
#define init_0_10_7_14
void FASTCALL call_0_10_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_7_15
#define init_0_10_7_15
void FASTCALL call_0_10_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_8_0
#define init_0_10_8_0
void FASTCALL call_0_10_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_8_1
#define init_0_10_8_1
void FASTCALL call_0_10_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_8_2
#define init_0_10_8_2
void FASTCALL call_0_10_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_8_3
#define init_0_10_8_3
void FASTCALL call_0_10_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_8_4
#define init_0_10_8_4
void FASTCALL call_0_10_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_8_5
#define init_0_10_8_5
void FASTCALL call_0_10_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_8_6
#define init_0_10_8_6
void FASTCALL call_0_10_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_8_7
#define init_0_10_8_7
void FASTCALL call_0_10_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_8_8
#define init_0_10_8_8
void FASTCALL call_0_10_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_8_9
#define init_0_10_8_9
void FASTCALL call_0_10_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_8_10
#define init_0_10_8_10
void FASTCALL call_0_10_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_8_11
#define init_0_10_8_11
void FASTCALL call_0_10_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_8_12
#define init_0_10_8_12
void FASTCALL call_0_10_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_8_13
#define init_0_10_8_13
void FASTCALL call_0_10_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_8_14
#define init_0_10_8_14
void FASTCALL call_0_10_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_8_15
#define init_0_10_8_15
void FASTCALL call_0_10_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_9_0
#define init_0_10_9_0
void FASTCALL call_0_10_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_9_1
#define init_0_10_9_1
void FASTCALL call_0_10_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_9_2
#define init_0_10_9_2
void FASTCALL call_0_10_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_9_3
#define init_0_10_9_3
void FASTCALL call_0_10_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_9_4
#define init_0_10_9_4
void FASTCALL call_0_10_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_9_5
#define init_0_10_9_5
void FASTCALL call_0_10_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_9_6
#define init_0_10_9_6
void FASTCALL call_0_10_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_9_7
#define init_0_10_9_7
void FASTCALL call_0_10_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_9_8
#define init_0_10_9_8
void FASTCALL call_0_10_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_9_9
#define init_0_10_9_9
void FASTCALL call_0_10_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_9_10
#define init_0_10_9_10
void FASTCALL call_0_10_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_9_11
#define init_0_10_9_11
void FASTCALL call_0_10_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_9_12
#define init_0_10_9_12
void FASTCALL call_0_10_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_9_13
#define init_0_10_9_13
void FASTCALL call_0_10_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_9_14
#define init_0_10_9_14
void FASTCALL call_0_10_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_9_15
#define init_0_10_9_15
void FASTCALL call_0_10_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_10_0
#define init_0_10_10_0
void FASTCALL call_0_10_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_10_1
#define init_0_10_10_1
void FASTCALL call_0_10_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_10_2
#define init_0_10_10_2
void FASTCALL call_0_10_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_10_3
#define init_0_10_10_3
void FASTCALL call_0_10_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_10_4
#define init_0_10_10_4
void FASTCALL call_0_10_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_10_5
#define init_0_10_10_5
void FASTCALL call_0_10_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_10_6
#define init_0_10_10_6
void FASTCALL call_0_10_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_10_7
#define init_0_10_10_7
void FASTCALL call_0_10_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_10_8
#define init_0_10_10_8
void FASTCALL call_0_10_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_10_9
#define init_0_10_10_9
void FASTCALL call_0_10_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_10_10
#define init_0_10_10_10
void FASTCALL call_0_10_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_10_11
#define init_0_10_10_11
void FASTCALL call_0_10_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_10_12
#define init_0_10_10_12
void FASTCALL call_0_10_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_10_13
#define init_0_10_10_13
void FASTCALL call_0_10_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_10_14
#define init_0_10_10_14
void FASTCALL call_0_10_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_10_15
#define init_0_10_10_15
void FASTCALL call_0_10_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_11_0
#define init_0_10_11_0
void FASTCALL call_0_10_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_11_1
#define init_0_10_11_1
void FASTCALL call_0_10_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_11_2
#define init_0_10_11_2
void FASTCALL call_0_10_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_11_3
#define init_0_10_11_3
void FASTCALL call_0_10_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_11_4
#define init_0_10_11_4
void FASTCALL call_0_10_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_11_5
#define init_0_10_11_5
void FASTCALL call_0_10_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_11_6
#define init_0_10_11_6
void FASTCALL call_0_10_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_11_7
#define init_0_10_11_7
void FASTCALL call_0_10_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_11_8
#define init_0_10_11_8
void FASTCALL call_0_10_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_11_9
#define init_0_10_11_9
void FASTCALL call_0_10_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_11_10
#define init_0_10_11_10
void FASTCALL call_0_10_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_11_11
#define init_0_10_11_11
void FASTCALL call_0_10_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_11_12
#define init_0_10_11_12
void FASTCALL call_0_10_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_11_13
#define init_0_10_11_13
void FASTCALL call_0_10_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_11_14
#define init_0_10_11_14
void FASTCALL call_0_10_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_11_15
#define init_0_10_11_15
void FASTCALL call_0_10_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_12_0
#define init_0_10_12_0
void FASTCALL call_0_10_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_12_1
#define init_0_10_12_1
void FASTCALL call_0_10_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_12_2
#define init_0_10_12_2
void FASTCALL call_0_10_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_12_3
#define init_0_10_12_3
void FASTCALL call_0_10_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_12_4
#define init_0_10_12_4
void FASTCALL call_0_10_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_12_5
#define init_0_10_12_5
void FASTCALL call_0_10_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_12_6
#define init_0_10_12_6
void FASTCALL call_0_10_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_12_7
#define init_0_10_12_7
void FASTCALL call_0_10_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_12_8
#define init_0_10_12_8
void FASTCALL call_0_10_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_12_9
#define init_0_10_12_9
void FASTCALL call_0_10_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_12_10
#define init_0_10_12_10
void FASTCALL call_0_10_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_12_11
#define init_0_10_12_11
void FASTCALL call_0_10_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_12_12
#define init_0_10_12_12
void FASTCALL call_0_10_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_12_13
#define init_0_10_12_13
void FASTCALL call_0_10_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_12_14
#define init_0_10_12_14
void FASTCALL call_0_10_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_12_15
#define init_0_10_12_15
void FASTCALL call_0_10_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_13_0
#define init_0_10_13_0
void FASTCALL call_0_10_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_13_1
#define init_0_10_13_1
void FASTCALL call_0_10_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_13_2
#define init_0_10_13_2
void FASTCALL call_0_10_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_13_3
#define init_0_10_13_3
void FASTCALL call_0_10_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_13_4
#define init_0_10_13_4
void FASTCALL call_0_10_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_13_5
#define init_0_10_13_5
void FASTCALL call_0_10_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_13_6
#define init_0_10_13_6
void FASTCALL call_0_10_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_13_7
#define init_0_10_13_7
void FASTCALL call_0_10_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_13_8
#define init_0_10_13_8
void FASTCALL call_0_10_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_13_9
#define init_0_10_13_9
void FASTCALL call_0_10_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_13_10
#define init_0_10_13_10
void FASTCALL call_0_10_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_13_11
#define init_0_10_13_11
void FASTCALL call_0_10_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_13_12
#define init_0_10_13_12
void FASTCALL call_0_10_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_13_13
#define init_0_10_13_13
void FASTCALL call_0_10_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_13_14
#define init_0_10_13_14
void FASTCALL call_0_10_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_13_15
#define init_0_10_13_15
void FASTCALL call_0_10_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_14_0
#define init_0_10_14_0
void FASTCALL call_0_10_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_14_1
#define init_0_10_14_1
void FASTCALL call_0_10_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_14_2
#define init_0_10_14_2
void FASTCALL call_0_10_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_14_3
#define init_0_10_14_3
void FASTCALL call_0_10_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_14_4
#define init_0_10_14_4
void FASTCALL call_0_10_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_14_5
#define init_0_10_14_5
void FASTCALL call_0_10_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_14_6
#define init_0_10_14_6
void FASTCALL call_0_10_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_14_7
#define init_0_10_14_7
void FASTCALL call_0_10_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_14_8
#define init_0_10_14_8
void FASTCALL call_0_10_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_14_9
#define init_0_10_14_9
void FASTCALL call_0_10_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_14_10
#define init_0_10_14_10
void FASTCALL call_0_10_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_14_11
#define init_0_10_14_11
void FASTCALL call_0_10_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_14_12
#define init_0_10_14_12
void FASTCALL call_0_10_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_14_13
#define init_0_10_14_13
void FASTCALL call_0_10_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_14_14
#define init_0_10_14_14
void FASTCALL call_0_10_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_14_15
#define init_0_10_14_15
void FASTCALL call_0_10_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_15_0
#define init_0_10_15_0
void FASTCALL call_0_10_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_15_1
#define init_0_10_15_1
void FASTCALL call_0_10_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_15_2
#define init_0_10_15_2
void FASTCALL call_0_10_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_15_3
#define init_0_10_15_3
void FASTCALL call_0_10_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_15_4
#define init_0_10_15_4
void FASTCALL call_0_10_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_15_5
#define init_0_10_15_5
void FASTCALL call_0_10_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_15_6
#define init_0_10_15_6
void FASTCALL call_0_10_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_15_7
#define init_0_10_15_7
void FASTCALL call_0_10_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_15_8
#define init_0_10_15_8
void FASTCALL call_0_10_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_15_9
#define init_0_10_15_9
void FASTCALL call_0_10_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_15_10
#define init_0_10_15_10
void FASTCALL call_0_10_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_15_11
#define init_0_10_15_11
void FASTCALL call_0_10_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_15_12
#define init_0_10_15_12
void FASTCALL call_0_10_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_15_13
#define init_0_10_15_13
void FASTCALL call_0_10_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_15_14
#define init_0_10_15_14
void FASTCALL call_0_10_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_10_15_15
#define init_0_10_15_15
void FASTCALL call_0_10_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_0_0
#define init_0_11_0_0
void FASTCALL call_0_11_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_0_1
#define init_0_11_0_1
void FASTCALL call_0_11_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_0_2
#define init_0_11_0_2
void FASTCALL call_0_11_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_0_3
#define init_0_11_0_3
void FASTCALL call_0_11_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_0_4
#define init_0_11_0_4
void FASTCALL call_0_11_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_0_5
#define init_0_11_0_5
void FASTCALL call_0_11_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_0_6
#define init_0_11_0_6
void FASTCALL call_0_11_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_0_7
#define init_0_11_0_7
void FASTCALL call_0_11_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_0_8
#define init_0_11_0_8
void FASTCALL call_0_11_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_0_9
#define init_0_11_0_9
void FASTCALL call_0_11_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_0_10
#define init_0_11_0_10
void FASTCALL call_0_11_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_0_11
#define init_0_11_0_11
void FASTCALL call_0_11_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_0_12
#define init_0_11_0_12
void FASTCALL call_0_11_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_0_13
#define init_0_11_0_13
void FASTCALL call_0_11_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_0_14
#define init_0_11_0_14
void FASTCALL call_0_11_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_0_15
#define init_0_11_0_15
void FASTCALL call_0_11_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_1_0
#define init_0_11_1_0
void FASTCALL call_0_11_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_1_1
#define init_0_11_1_1
void FASTCALL call_0_11_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_1_2
#define init_0_11_1_2
void FASTCALL call_0_11_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_1_3
#define init_0_11_1_3
void FASTCALL call_0_11_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_1_4
#define init_0_11_1_4
void FASTCALL call_0_11_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_1_5
#define init_0_11_1_5
void FASTCALL call_0_11_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_1_6
#define init_0_11_1_6
void FASTCALL call_0_11_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_1_7
#define init_0_11_1_7
void FASTCALL call_0_11_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_1_8
#define init_0_11_1_8
void FASTCALL call_0_11_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_1_9
#define init_0_11_1_9
void FASTCALL call_0_11_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_1_10
#define init_0_11_1_10
void FASTCALL call_0_11_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_1_11
#define init_0_11_1_11
void FASTCALL call_0_11_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_1_12
#define init_0_11_1_12
void FASTCALL call_0_11_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_1_13
#define init_0_11_1_13
void FASTCALL call_0_11_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_1_14
#define init_0_11_1_14
void FASTCALL call_0_11_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_1_15
#define init_0_11_1_15
void FASTCALL call_0_11_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_2_0
#define init_0_11_2_0
void FASTCALL call_0_11_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_2_1
#define init_0_11_2_1
void FASTCALL call_0_11_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_2_2
#define init_0_11_2_2
void FASTCALL call_0_11_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_2_3
#define init_0_11_2_3
void FASTCALL call_0_11_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_2_4
#define init_0_11_2_4
void FASTCALL call_0_11_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_2_5
#define init_0_11_2_5
void FASTCALL call_0_11_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_2_6
#define init_0_11_2_6
void FASTCALL call_0_11_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_2_7
#define init_0_11_2_7
void FASTCALL call_0_11_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_2_8
#define init_0_11_2_8
void FASTCALL call_0_11_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_2_9
#define init_0_11_2_9
void FASTCALL call_0_11_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_2_10
#define init_0_11_2_10
void FASTCALL call_0_11_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_2_11
#define init_0_11_2_11
void FASTCALL call_0_11_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_2_12
#define init_0_11_2_12
void FASTCALL call_0_11_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_2_13
#define init_0_11_2_13
void FASTCALL call_0_11_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_2_14
#define init_0_11_2_14
void FASTCALL call_0_11_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_2_15
#define init_0_11_2_15
void FASTCALL call_0_11_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_3_0
#define init_0_11_3_0
void FASTCALL call_0_11_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_3_1
#define init_0_11_3_1
void FASTCALL call_0_11_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_3_2
#define init_0_11_3_2
void FASTCALL call_0_11_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_3_3
#define init_0_11_3_3
void FASTCALL call_0_11_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_3_4
#define init_0_11_3_4
void FASTCALL call_0_11_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_3_5
#define init_0_11_3_5
void FASTCALL call_0_11_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_3_6
#define init_0_11_3_6
void FASTCALL call_0_11_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_3_7
#define init_0_11_3_7
void FASTCALL call_0_11_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_3_8
#define init_0_11_3_8
void FASTCALL call_0_11_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_3_9
#define init_0_11_3_9
void FASTCALL call_0_11_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_3_10
#define init_0_11_3_10
void FASTCALL call_0_11_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_3_11
#define init_0_11_3_11
void FASTCALL call_0_11_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_3_12
#define init_0_11_3_12
void FASTCALL call_0_11_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_3_13
#define init_0_11_3_13
void FASTCALL call_0_11_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_3_14
#define init_0_11_3_14
void FASTCALL call_0_11_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_3_15
#define init_0_11_3_15
void FASTCALL call_0_11_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_4_0
#define init_0_11_4_0
void FASTCALL call_0_11_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_4_1
#define init_0_11_4_1
void FASTCALL call_0_11_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_4_2
#define init_0_11_4_2
void FASTCALL call_0_11_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_4_3
#define init_0_11_4_3
void FASTCALL call_0_11_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_4_4
#define init_0_11_4_4
void FASTCALL call_0_11_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_4_5
#define init_0_11_4_5
void FASTCALL call_0_11_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_4_6
#define init_0_11_4_6
void FASTCALL call_0_11_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_4_7
#define init_0_11_4_7
void FASTCALL call_0_11_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_4_8
#define init_0_11_4_8
void FASTCALL call_0_11_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_4_9
#define init_0_11_4_9
void FASTCALL call_0_11_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_4_10
#define init_0_11_4_10
void FASTCALL call_0_11_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_4_11
#define init_0_11_4_11
void FASTCALL call_0_11_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_4_12
#define init_0_11_4_12
void FASTCALL call_0_11_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_4_13
#define init_0_11_4_13
void FASTCALL call_0_11_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_4_14
#define init_0_11_4_14
void FASTCALL call_0_11_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_4_15
#define init_0_11_4_15
void FASTCALL call_0_11_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_5_0
#define init_0_11_5_0
void FASTCALL call_0_11_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_5_1
#define init_0_11_5_1
void FASTCALL call_0_11_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_5_2
#define init_0_11_5_2
void FASTCALL call_0_11_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_5_3
#define init_0_11_5_3
void FASTCALL call_0_11_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_5_4
#define init_0_11_5_4
void FASTCALL call_0_11_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_5_5
#define init_0_11_5_5
void FASTCALL call_0_11_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_5_6
#define init_0_11_5_6
void FASTCALL call_0_11_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_5_7
#define init_0_11_5_7
void FASTCALL call_0_11_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_5_8
#define init_0_11_5_8
void FASTCALL call_0_11_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_5_9
#define init_0_11_5_9
void FASTCALL call_0_11_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_5_10
#define init_0_11_5_10
void FASTCALL call_0_11_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_5_11
#define init_0_11_5_11
void FASTCALL call_0_11_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_5_12
#define init_0_11_5_12
void FASTCALL call_0_11_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_5_13
#define init_0_11_5_13
void FASTCALL call_0_11_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_5_14
#define init_0_11_5_14
void FASTCALL call_0_11_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_5_15
#define init_0_11_5_15
void FASTCALL call_0_11_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_6_0
#define init_0_11_6_0
void FASTCALL call_0_11_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_6_1
#define init_0_11_6_1
void FASTCALL call_0_11_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_6_2
#define init_0_11_6_2
void FASTCALL call_0_11_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_6_3
#define init_0_11_6_3
void FASTCALL call_0_11_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_6_4
#define init_0_11_6_4
void FASTCALL call_0_11_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_6_5
#define init_0_11_6_5
void FASTCALL call_0_11_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_6_6
#define init_0_11_6_6
void FASTCALL call_0_11_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_6_7
#define init_0_11_6_7
void FASTCALL call_0_11_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_6_8
#define init_0_11_6_8
void FASTCALL call_0_11_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_6_9
#define init_0_11_6_9
void FASTCALL call_0_11_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_6_10
#define init_0_11_6_10
void FASTCALL call_0_11_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_6_11
#define init_0_11_6_11
void FASTCALL call_0_11_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_6_12
#define init_0_11_6_12
void FASTCALL call_0_11_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_6_13
#define init_0_11_6_13
void FASTCALL call_0_11_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_6_14
#define init_0_11_6_14
void FASTCALL call_0_11_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_6_15
#define init_0_11_6_15
void FASTCALL call_0_11_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_7_0
#define init_0_11_7_0
void FASTCALL call_0_11_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_7_1
#define init_0_11_7_1
void FASTCALL call_0_11_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_7_2
#define init_0_11_7_2
void FASTCALL call_0_11_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_7_3
#define init_0_11_7_3
void FASTCALL call_0_11_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_7_4
#define init_0_11_7_4
void FASTCALL call_0_11_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_7_5
#define init_0_11_7_5
void FASTCALL call_0_11_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_7_6
#define init_0_11_7_6
void FASTCALL call_0_11_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_7_7
#define init_0_11_7_7
void FASTCALL call_0_11_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_7_8
#define init_0_11_7_8
void FASTCALL call_0_11_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_7_9
#define init_0_11_7_9
void FASTCALL call_0_11_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_7_10
#define init_0_11_7_10
void FASTCALL call_0_11_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_7_11
#define init_0_11_7_11
void FASTCALL call_0_11_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_7_12
#define init_0_11_7_12
void FASTCALL call_0_11_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_7_13
#define init_0_11_7_13
void FASTCALL call_0_11_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_7_14
#define init_0_11_7_14
void FASTCALL call_0_11_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_7_15
#define init_0_11_7_15
void FASTCALL call_0_11_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_8_0
#define init_0_11_8_0
void FASTCALL call_0_11_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_8_1
#define init_0_11_8_1
void FASTCALL call_0_11_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_8_2
#define init_0_11_8_2
void FASTCALL call_0_11_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_8_3
#define init_0_11_8_3
void FASTCALL call_0_11_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_8_4
#define init_0_11_8_4
void FASTCALL call_0_11_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_8_5
#define init_0_11_8_5
void FASTCALL call_0_11_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_8_6
#define init_0_11_8_6
void FASTCALL call_0_11_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_8_7
#define init_0_11_8_7
void FASTCALL call_0_11_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_8_8
#define init_0_11_8_8
void FASTCALL call_0_11_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_8_9
#define init_0_11_8_9
void FASTCALL call_0_11_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_8_10
#define init_0_11_8_10
void FASTCALL call_0_11_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_8_11
#define init_0_11_8_11
void FASTCALL call_0_11_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_8_12
#define init_0_11_8_12
void FASTCALL call_0_11_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_8_13
#define init_0_11_8_13
void FASTCALL call_0_11_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_8_14
#define init_0_11_8_14
void FASTCALL call_0_11_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_8_15
#define init_0_11_8_15
void FASTCALL call_0_11_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_9_0
#define init_0_11_9_0
void FASTCALL call_0_11_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_9_1
#define init_0_11_9_1
void FASTCALL call_0_11_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_9_2
#define init_0_11_9_2
void FASTCALL call_0_11_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_9_3
#define init_0_11_9_3
void FASTCALL call_0_11_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_9_4
#define init_0_11_9_4
void FASTCALL call_0_11_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_9_5
#define init_0_11_9_5
void FASTCALL call_0_11_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_9_6
#define init_0_11_9_6
void FASTCALL call_0_11_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_9_7
#define init_0_11_9_7
void FASTCALL call_0_11_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_9_8
#define init_0_11_9_8
void FASTCALL call_0_11_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_9_9
#define init_0_11_9_9
void FASTCALL call_0_11_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_9_10
#define init_0_11_9_10
void FASTCALL call_0_11_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_9_11
#define init_0_11_9_11
void FASTCALL call_0_11_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_9_12
#define init_0_11_9_12
void FASTCALL call_0_11_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_9_13
#define init_0_11_9_13
void FASTCALL call_0_11_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_9_14
#define init_0_11_9_14
void FASTCALL call_0_11_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_9_15
#define init_0_11_9_15
void FASTCALL call_0_11_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_10_0
#define init_0_11_10_0
void FASTCALL call_0_11_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_10_1
#define init_0_11_10_1
void FASTCALL call_0_11_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_10_2
#define init_0_11_10_2
void FASTCALL call_0_11_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_10_3
#define init_0_11_10_3
void FASTCALL call_0_11_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_10_4
#define init_0_11_10_4
void FASTCALL call_0_11_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_10_5
#define init_0_11_10_5
void FASTCALL call_0_11_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_10_6
#define init_0_11_10_6
void FASTCALL call_0_11_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_10_7
#define init_0_11_10_7
void FASTCALL call_0_11_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_10_8
#define init_0_11_10_8
void FASTCALL call_0_11_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_10_9
#define init_0_11_10_9
void FASTCALL call_0_11_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_10_10
#define init_0_11_10_10
void FASTCALL call_0_11_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_10_11
#define init_0_11_10_11
void FASTCALL call_0_11_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_10_12
#define init_0_11_10_12
void FASTCALL call_0_11_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_10_13
#define init_0_11_10_13
void FASTCALL call_0_11_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_10_14
#define init_0_11_10_14
void FASTCALL call_0_11_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_10_15
#define init_0_11_10_15
void FASTCALL call_0_11_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_11_0
#define init_0_11_11_0
void FASTCALL call_0_11_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_11_1
#define init_0_11_11_1
void FASTCALL call_0_11_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_11_2
#define init_0_11_11_2
void FASTCALL call_0_11_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_11_3
#define init_0_11_11_3
void FASTCALL call_0_11_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_11_4
#define init_0_11_11_4
void FASTCALL call_0_11_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_11_5
#define init_0_11_11_5
void FASTCALL call_0_11_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_11_6
#define init_0_11_11_6
void FASTCALL call_0_11_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_11_7
#define init_0_11_11_7
void FASTCALL call_0_11_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_11_8
#define init_0_11_11_8
void FASTCALL call_0_11_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_11_9
#define init_0_11_11_9
void FASTCALL call_0_11_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_11_10
#define init_0_11_11_10
void FASTCALL call_0_11_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_11_11
#define init_0_11_11_11
void FASTCALL call_0_11_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_11_12
#define init_0_11_11_12
void FASTCALL call_0_11_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_11_13
#define init_0_11_11_13
void FASTCALL call_0_11_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_11_14
#define init_0_11_11_14
void FASTCALL call_0_11_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_11_15
#define init_0_11_11_15
void FASTCALL call_0_11_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_12_0
#define init_0_11_12_0
void FASTCALL call_0_11_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_12_1
#define init_0_11_12_1
void FASTCALL call_0_11_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_12_2
#define init_0_11_12_2
void FASTCALL call_0_11_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_12_3
#define init_0_11_12_3
void FASTCALL call_0_11_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_12_4
#define init_0_11_12_4
void FASTCALL call_0_11_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_12_5
#define init_0_11_12_5
void FASTCALL call_0_11_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_12_6
#define init_0_11_12_6
void FASTCALL call_0_11_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_12_7
#define init_0_11_12_7
void FASTCALL call_0_11_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_12_8
#define init_0_11_12_8
void FASTCALL call_0_11_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_12_9
#define init_0_11_12_9
void FASTCALL call_0_11_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_12_10
#define init_0_11_12_10
void FASTCALL call_0_11_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_12_11
#define init_0_11_12_11
void FASTCALL call_0_11_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_12_12
#define init_0_11_12_12
void FASTCALL call_0_11_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_12_13
#define init_0_11_12_13
void FASTCALL call_0_11_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_12_14
#define init_0_11_12_14
void FASTCALL call_0_11_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_12_15
#define init_0_11_12_15
void FASTCALL call_0_11_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_13_0
#define init_0_11_13_0
void FASTCALL call_0_11_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_13_1
#define init_0_11_13_1
void FASTCALL call_0_11_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_13_2
#define init_0_11_13_2
void FASTCALL call_0_11_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_13_3
#define init_0_11_13_3
void FASTCALL call_0_11_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_13_4
#define init_0_11_13_4
void FASTCALL call_0_11_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_13_5
#define init_0_11_13_5
void FASTCALL call_0_11_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_13_6
#define init_0_11_13_6
void FASTCALL call_0_11_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_13_7
#define init_0_11_13_7
void FASTCALL call_0_11_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_13_8
#define init_0_11_13_8
void FASTCALL call_0_11_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_13_9
#define init_0_11_13_9
void FASTCALL call_0_11_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_13_10
#define init_0_11_13_10
void FASTCALL call_0_11_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_13_11
#define init_0_11_13_11
void FASTCALL call_0_11_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_13_12
#define init_0_11_13_12
void FASTCALL call_0_11_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_13_13
#define init_0_11_13_13
void FASTCALL call_0_11_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_13_14
#define init_0_11_13_14
void FASTCALL call_0_11_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_13_15
#define init_0_11_13_15
void FASTCALL call_0_11_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_14_0
#define init_0_11_14_0
void FASTCALL call_0_11_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_14_1
#define init_0_11_14_1
void FASTCALL call_0_11_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_14_2
#define init_0_11_14_2
void FASTCALL call_0_11_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_14_3
#define init_0_11_14_3
void FASTCALL call_0_11_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_14_4
#define init_0_11_14_4
void FASTCALL call_0_11_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_14_5
#define init_0_11_14_5
void FASTCALL call_0_11_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_14_6
#define init_0_11_14_6
void FASTCALL call_0_11_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_14_7
#define init_0_11_14_7
void FASTCALL call_0_11_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_14_8
#define init_0_11_14_8
void FASTCALL call_0_11_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_14_9
#define init_0_11_14_9
void FASTCALL call_0_11_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_14_10
#define init_0_11_14_10
void FASTCALL call_0_11_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_14_11
#define init_0_11_14_11
void FASTCALL call_0_11_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_14_12
#define init_0_11_14_12
void FASTCALL call_0_11_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_14_13
#define init_0_11_14_13
void FASTCALL call_0_11_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_14_14
#define init_0_11_14_14
void FASTCALL call_0_11_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_14_15
#define init_0_11_14_15
void FASTCALL call_0_11_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_15_0
#define init_0_11_15_0
void FASTCALL call_0_11_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_15_1
#define init_0_11_15_1
void FASTCALL call_0_11_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_15_2
#define init_0_11_15_2
void FASTCALL call_0_11_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_15_3
#define init_0_11_15_3
void FASTCALL call_0_11_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_15_4
#define init_0_11_15_4
void FASTCALL call_0_11_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_15_5
#define init_0_11_15_5
void FASTCALL call_0_11_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_15_6
#define init_0_11_15_6
void FASTCALL call_0_11_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_15_7
#define init_0_11_15_7
void FASTCALL call_0_11_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_15_8
#define init_0_11_15_8
void FASTCALL call_0_11_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_15_9
#define init_0_11_15_9
void FASTCALL call_0_11_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_15_10
#define init_0_11_15_10
void FASTCALL call_0_11_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_15_11
#define init_0_11_15_11
void FASTCALL call_0_11_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_15_12
#define init_0_11_15_12
void FASTCALL call_0_11_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_15_13
#define init_0_11_15_13
void FASTCALL call_0_11_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_15_14
#define init_0_11_15_14
void FASTCALL call_0_11_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_11_15_15
#define init_0_11_15_15
void FASTCALL call_0_11_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_0_0
#define init_0_12_0_0
void FASTCALL call_0_12_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_0_1
#define init_0_12_0_1
void FASTCALL call_0_12_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_0_2
#define init_0_12_0_2
void FASTCALL call_0_12_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_0_3
#define init_0_12_0_3
void FASTCALL call_0_12_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_0_4
#define init_0_12_0_4
void FASTCALL call_0_12_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_0_5
#define init_0_12_0_5
void FASTCALL call_0_12_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_0_6
#define init_0_12_0_6
void FASTCALL call_0_12_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_0_7
#define init_0_12_0_7
void FASTCALL call_0_12_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_0_8
#define init_0_12_0_8
void FASTCALL call_0_12_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_0_9
#define init_0_12_0_9
void FASTCALL call_0_12_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_0_10
#define init_0_12_0_10
void FASTCALL call_0_12_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_0_11
#define init_0_12_0_11
void FASTCALL call_0_12_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_0_12
#define init_0_12_0_12
void FASTCALL call_0_12_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_0_13
#define init_0_12_0_13
void FASTCALL call_0_12_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_0_14
#define init_0_12_0_14
void FASTCALL call_0_12_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_0_15
#define init_0_12_0_15
void FASTCALL call_0_12_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_1_0
#define init_0_12_1_0
void FASTCALL call_0_12_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_1_1
#define init_0_12_1_1
void FASTCALL call_0_12_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_1_2
#define init_0_12_1_2
void FASTCALL call_0_12_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_1_3
#define init_0_12_1_3
void FASTCALL call_0_12_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_1_4
#define init_0_12_1_4
void FASTCALL call_0_12_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_1_5
#define init_0_12_1_5
void FASTCALL call_0_12_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_1_6
#define init_0_12_1_6
void FASTCALL call_0_12_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_1_7
#define init_0_12_1_7
void FASTCALL call_0_12_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_1_8
#define init_0_12_1_8
void FASTCALL call_0_12_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_1_9
#define init_0_12_1_9
void FASTCALL call_0_12_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_1_10
#define init_0_12_1_10
void FASTCALL call_0_12_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_1_11
#define init_0_12_1_11
void FASTCALL call_0_12_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_1_12
#define init_0_12_1_12
void FASTCALL call_0_12_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_1_13
#define init_0_12_1_13
void FASTCALL call_0_12_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_1_14
#define init_0_12_1_14
void FASTCALL call_0_12_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_1_15
#define init_0_12_1_15
void FASTCALL call_0_12_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_2_0
#define init_0_12_2_0
void FASTCALL call_0_12_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_2_1
#define init_0_12_2_1
void FASTCALL call_0_12_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_2_2
#define init_0_12_2_2
void FASTCALL call_0_12_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_2_3
#define init_0_12_2_3
void FASTCALL call_0_12_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_2_4
#define init_0_12_2_4
void FASTCALL call_0_12_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_2_5
#define init_0_12_2_5
void FASTCALL call_0_12_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_2_6
#define init_0_12_2_6
void FASTCALL call_0_12_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_2_7
#define init_0_12_2_7
void FASTCALL call_0_12_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_2_8
#define init_0_12_2_8
void FASTCALL call_0_12_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_2_9
#define init_0_12_2_9
void FASTCALL call_0_12_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_2_10
#define init_0_12_2_10
void FASTCALL call_0_12_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_2_11
#define init_0_12_2_11
void FASTCALL call_0_12_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_2_12
#define init_0_12_2_12
void FASTCALL call_0_12_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_2_13
#define init_0_12_2_13
void FASTCALL call_0_12_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_2_14
#define init_0_12_2_14
void FASTCALL call_0_12_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_2_15
#define init_0_12_2_15
void FASTCALL call_0_12_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_3_0
#define init_0_12_3_0
void FASTCALL call_0_12_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_3_1
#define init_0_12_3_1
void FASTCALL call_0_12_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_3_2
#define init_0_12_3_2
void FASTCALL call_0_12_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_3_3
#define init_0_12_3_3
void FASTCALL call_0_12_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_3_4
#define init_0_12_3_4
void FASTCALL call_0_12_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_3_5
#define init_0_12_3_5
void FASTCALL call_0_12_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_3_6
#define init_0_12_3_6
void FASTCALL call_0_12_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_3_7
#define init_0_12_3_7
void FASTCALL call_0_12_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_3_8
#define init_0_12_3_8
void FASTCALL call_0_12_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_3_9
#define init_0_12_3_9
void FASTCALL call_0_12_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_3_10
#define init_0_12_3_10
void FASTCALL call_0_12_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_3_11
#define init_0_12_3_11
void FASTCALL call_0_12_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_3_12
#define init_0_12_3_12
void FASTCALL call_0_12_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_3_13
#define init_0_12_3_13
void FASTCALL call_0_12_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_3_14
#define init_0_12_3_14
void FASTCALL call_0_12_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_3_15
#define init_0_12_3_15
void FASTCALL call_0_12_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_4_0
#define init_0_12_4_0
void FASTCALL call_0_12_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_4_1
#define init_0_12_4_1
void FASTCALL call_0_12_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_4_2
#define init_0_12_4_2
void FASTCALL call_0_12_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_4_3
#define init_0_12_4_3
void FASTCALL call_0_12_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_4_4
#define init_0_12_4_4
void FASTCALL call_0_12_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_4_5
#define init_0_12_4_5
void FASTCALL call_0_12_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_4_6
#define init_0_12_4_6
void FASTCALL call_0_12_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_4_7
#define init_0_12_4_7
void FASTCALL call_0_12_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_4_8
#define init_0_12_4_8
void FASTCALL call_0_12_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_4_9
#define init_0_12_4_9
void FASTCALL call_0_12_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_4_10
#define init_0_12_4_10
void FASTCALL call_0_12_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_4_11
#define init_0_12_4_11
void FASTCALL call_0_12_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_4_12
#define init_0_12_4_12
void FASTCALL call_0_12_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_4_13
#define init_0_12_4_13
void FASTCALL call_0_12_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_4_14
#define init_0_12_4_14
void FASTCALL call_0_12_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_4_15
#define init_0_12_4_15
void FASTCALL call_0_12_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_5_0
#define init_0_12_5_0
void FASTCALL call_0_12_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_5_1
#define init_0_12_5_1
void FASTCALL call_0_12_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_5_2
#define init_0_12_5_2
void FASTCALL call_0_12_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_5_3
#define init_0_12_5_3
void FASTCALL call_0_12_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_5_4
#define init_0_12_5_4
void FASTCALL call_0_12_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_5_5
#define init_0_12_5_5
void FASTCALL call_0_12_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_5_6
#define init_0_12_5_6
void FASTCALL call_0_12_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_5_7
#define init_0_12_5_7
void FASTCALL call_0_12_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_5_8
#define init_0_12_5_8
void FASTCALL call_0_12_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_5_9
#define init_0_12_5_9
void FASTCALL call_0_12_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_5_10
#define init_0_12_5_10
void FASTCALL call_0_12_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_5_11
#define init_0_12_5_11
void FASTCALL call_0_12_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_5_12
#define init_0_12_5_12
void FASTCALL call_0_12_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_5_13
#define init_0_12_5_13
void FASTCALL call_0_12_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_5_14
#define init_0_12_5_14
void FASTCALL call_0_12_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_5_15
#define init_0_12_5_15
void FASTCALL call_0_12_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_6_0
#define init_0_12_6_0
void FASTCALL call_0_12_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_6_1
#define init_0_12_6_1
void FASTCALL call_0_12_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_6_2
#define init_0_12_6_2
void FASTCALL call_0_12_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_6_3
#define init_0_12_6_3
void FASTCALL call_0_12_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_6_4
#define init_0_12_6_4
void FASTCALL call_0_12_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_6_5
#define init_0_12_6_5
void FASTCALL call_0_12_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_6_6
#define init_0_12_6_6
void FASTCALL call_0_12_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_6_7
#define init_0_12_6_7
void FASTCALL call_0_12_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_6_8
#define init_0_12_6_8
void FASTCALL call_0_12_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_6_9
#define init_0_12_6_9
void FASTCALL call_0_12_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_6_10
#define init_0_12_6_10
void FASTCALL call_0_12_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_6_11
#define init_0_12_6_11
void FASTCALL call_0_12_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_6_12
#define init_0_12_6_12
void FASTCALL call_0_12_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_6_13
#define init_0_12_6_13
void FASTCALL call_0_12_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_6_14
#define init_0_12_6_14
void FASTCALL call_0_12_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_6_15
#define init_0_12_6_15
void FASTCALL call_0_12_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_7_0
#define init_0_12_7_0
void FASTCALL call_0_12_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_7_1
#define init_0_12_7_1
void FASTCALL call_0_12_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_7_2
#define init_0_12_7_2
void FASTCALL call_0_12_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_7_3
#define init_0_12_7_3
void FASTCALL call_0_12_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_7_4
#define init_0_12_7_4
void FASTCALL call_0_12_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_7_5
#define init_0_12_7_5
void FASTCALL call_0_12_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_7_6
#define init_0_12_7_6
void FASTCALL call_0_12_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_7_7
#define init_0_12_7_7
void FASTCALL call_0_12_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_7_8
#define init_0_12_7_8
void FASTCALL call_0_12_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_7_9
#define init_0_12_7_9
void FASTCALL call_0_12_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_7_10
#define init_0_12_7_10
void FASTCALL call_0_12_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_7_11
#define init_0_12_7_11
void FASTCALL call_0_12_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_7_12
#define init_0_12_7_12
void FASTCALL call_0_12_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_7_13
#define init_0_12_7_13
void FASTCALL call_0_12_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_7_14
#define init_0_12_7_14
void FASTCALL call_0_12_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_7_15
#define init_0_12_7_15
void FASTCALL call_0_12_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_8_0
#define init_0_12_8_0
void FASTCALL call_0_12_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_8_1
#define init_0_12_8_1
void FASTCALL call_0_12_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_8_2
#define init_0_12_8_2
void FASTCALL call_0_12_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_8_3
#define init_0_12_8_3
void FASTCALL call_0_12_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_8_4
#define init_0_12_8_4
void FASTCALL call_0_12_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_8_5
#define init_0_12_8_5
void FASTCALL call_0_12_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_8_6
#define init_0_12_8_6
void FASTCALL call_0_12_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_8_7
#define init_0_12_8_7
void FASTCALL call_0_12_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_8_8
#define init_0_12_8_8
void FASTCALL call_0_12_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_8_9
#define init_0_12_8_9
void FASTCALL call_0_12_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_8_10
#define init_0_12_8_10
void FASTCALL call_0_12_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_8_11
#define init_0_12_8_11
void FASTCALL call_0_12_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_8_12
#define init_0_12_8_12
void FASTCALL call_0_12_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_8_13
#define init_0_12_8_13
void FASTCALL call_0_12_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_8_14
#define init_0_12_8_14
void FASTCALL call_0_12_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_8_15
#define init_0_12_8_15
void FASTCALL call_0_12_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_9_0
#define init_0_12_9_0
void FASTCALL call_0_12_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_9_1
#define init_0_12_9_1
void FASTCALL call_0_12_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_9_2
#define init_0_12_9_2
void FASTCALL call_0_12_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_9_3
#define init_0_12_9_3
void FASTCALL call_0_12_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_9_4
#define init_0_12_9_4
void FASTCALL call_0_12_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_9_5
#define init_0_12_9_5
void FASTCALL call_0_12_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_9_6
#define init_0_12_9_6
void FASTCALL call_0_12_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_9_7
#define init_0_12_9_7
void FASTCALL call_0_12_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_9_8
#define init_0_12_9_8
void FASTCALL call_0_12_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_9_9
#define init_0_12_9_9
void FASTCALL call_0_12_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_9_10
#define init_0_12_9_10
void FASTCALL call_0_12_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_9_11
#define init_0_12_9_11
void FASTCALL call_0_12_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_9_12
#define init_0_12_9_12
void FASTCALL call_0_12_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_9_13
#define init_0_12_9_13
void FASTCALL call_0_12_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_9_14
#define init_0_12_9_14
void FASTCALL call_0_12_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_9_15
#define init_0_12_9_15
void FASTCALL call_0_12_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_10_0
#define init_0_12_10_0
void FASTCALL call_0_12_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_10_1
#define init_0_12_10_1
void FASTCALL call_0_12_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_10_2
#define init_0_12_10_2
void FASTCALL call_0_12_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_10_3
#define init_0_12_10_3
void FASTCALL call_0_12_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_10_4
#define init_0_12_10_4
void FASTCALL call_0_12_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_10_5
#define init_0_12_10_5
void FASTCALL call_0_12_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_10_6
#define init_0_12_10_6
void FASTCALL call_0_12_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_10_7
#define init_0_12_10_7
void FASTCALL call_0_12_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_10_8
#define init_0_12_10_8
void FASTCALL call_0_12_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_10_9
#define init_0_12_10_9
void FASTCALL call_0_12_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_10_10
#define init_0_12_10_10
void FASTCALL call_0_12_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_10_11
#define init_0_12_10_11
void FASTCALL call_0_12_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_10_12
#define init_0_12_10_12
void FASTCALL call_0_12_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_10_13
#define init_0_12_10_13
void FASTCALL call_0_12_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_10_14
#define init_0_12_10_14
void FASTCALL call_0_12_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_10_15
#define init_0_12_10_15
void FASTCALL call_0_12_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_11_0
#define init_0_12_11_0
void FASTCALL call_0_12_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_11_1
#define init_0_12_11_1
void FASTCALL call_0_12_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_11_2
#define init_0_12_11_2
void FASTCALL call_0_12_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_11_3
#define init_0_12_11_3
void FASTCALL call_0_12_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_11_4
#define init_0_12_11_4
void FASTCALL call_0_12_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_11_5
#define init_0_12_11_5
void FASTCALL call_0_12_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_11_6
#define init_0_12_11_6
void FASTCALL call_0_12_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_11_7
#define init_0_12_11_7
void FASTCALL call_0_12_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_11_8
#define init_0_12_11_8
void FASTCALL call_0_12_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_11_9
#define init_0_12_11_9
void FASTCALL call_0_12_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_11_10
#define init_0_12_11_10
void FASTCALL call_0_12_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_11_11
#define init_0_12_11_11
void FASTCALL call_0_12_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_11_12
#define init_0_12_11_12
void FASTCALL call_0_12_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_11_13
#define init_0_12_11_13
void FASTCALL call_0_12_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_11_14
#define init_0_12_11_14
void FASTCALL call_0_12_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_11_15
#define init_0_12_11_15
void FASTCALL call_0_12_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_12_0
#define init_0_12_12_0
void FASTCALL call_0_12_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_12_1
#define init_0_12_12_1
void FASTCALL call_0_12_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_12_2
#define init_0_12_12_2
void FASTCALL call_0_12_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_12_3
#define init_0_12_12_3
void FASTCALL call_0_12_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_12_4
#define init_0_12_12_4
void FASTCALL call_0_12_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_12_5
#define init_0_12_12_5
void FASTCALL call_0_12_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_12_6
#define init_0_12_12_6
void FASTCALL call_0_12_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_12_7
#define init_0_12_12_7
void FASTCALL call_0_12_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_12_8
#define init_0_12_12_8
void FASTCALL call_0_12_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_12_9
#define init_0_12_12_9
void FASTCALL call_0_12_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_12_10
#define init_0_12_12_10
void FASTCALL call_0_12_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_12_11
#define init_0_12_12_11
void FASTCALL call_0_12_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_12_12
#define init_0_12_12_12
void FASTCALL call_0_12_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_12_13
#define init_0_12_12_13
void FASTCALL call_0_12_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_12_14
#define init_0_12_12_14
void FASTCALL call_0_12_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_12_15
#define init_0_12_12_15
void FASTCALL call_0_12_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_13_0
#define init_0_12_13_0
void FASTCALL call_0_12_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_13_1
#define init_0_12_13_1
void FASTCALL call_0_12_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_13_2
#define init_0_12_13_2
void FASTCALL call_0_12_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_13_3
#define init_0_12_13_3
void FASTCALL call_0_12_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_13_4
#define init_0_12_13_4
void FASTCALL call_0_12_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_13_5
#define init_0_12_13_5
void FASTCALL call_0_12_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_13_6
#define init_0_12_13_6
void FASTCALL call_0_12_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_13_7
#define init_0_12_13_7
void FASTCALL call_0_12_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_13_8
#define init_0_12_13_8
void FASTCALL call_0_12_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_13_9
#define init_0_12_13_9
void FASTCALL call_0_12_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_13_10
#define init_0_12_13_10
void FASTCALL call_0_12_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_13_11
#define init_0_12_13_11
void FASTCALL call_0_12_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_13_12
#define init_0_12_13_12
void FASTCALL call_0_12_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_13_13
#define init_0_12_13_13
void FASTCALL call_0_12_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_13_14
#define init_0_12_13_14
void FASTCALL call_0_12_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_13_15
#define init_0_12_13_15
void FASTCALL call_0_12_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_14_0
#define init_0_12_14_0
void FASTCALL call_0_12_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_14_1
#define init_0_12_14_1
void FASTCALL call_0_12_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_14_2
#define init_0_12_14_2
void FASTCALL call_0_12_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_14_3
#define init_0_12_14_3
void FASTCALL call_0_12_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_14_4
#define init_0_12_14_4
void FASTCALL call_0_12_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_14_5
#define init_0_12_14_5
void FASTCALL call_0_12_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_14_6
#define init_0_12_14_6
void FASTCALL call_0_12_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_14_7
#define init_0_12_14_7
void FASTCALL call_0_12_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_14_8
#define init_0_12_14_8
void FASTCALL call_0_12_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_14_9
#define init_0_12_14_9
void FASTCALL call_0_12_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_14_10
#define init_0_12_14_10
void FASTCALL call_0_12_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_14_11
#define init_0_12_14_11
void FASTCALL call_0_12_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_14_12
#define init_0_12_14_12
void FASTCALL call_0_12_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_14_13
#define init_0_12_14_13
void FASTCALL call_0_12_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_14_14
#define init_0_12_14_14
void FASTCALL call_0_12_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_14_15
#define init_0_12_14_15
void FASTCALL call_0_12_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_15_0
#define init_0_12_15_0
void FASTCALL call_0_12_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_15_1
#define init_0_12_15_1
void FASTCALL call_0_12_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_15_2
#define init_0_12_15_2
void FASTCALL call_0_12_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_15_3
#define init_0_12_15_3
void FASTCALL call_0_12_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_15_4
#define init_0_12_15_4
void FASTCALL call_0_12_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_15_5
#define init_0_12_15_5
void FASTCALL call_0_12_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_15_6
#define init_0_12_15_6
void FASTCALL call_0_12_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_15_7
#define init_0_12_15_7
void FASTCALL call_0_12_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_15_8
#define init_0_12_15_8
void FASTCALL call_0_12_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_15_9
#define init_0_12_15_9
void FASTCALL call_0_12_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_15_10
#define init_0_12_15_10
void FASTCALL call_0_12_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_15_11
#define init_0_12_15_11
void FASTCALL call_0_12_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_15_12
#define init_0_12_15_12
void FASTCALL call_0_12_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_15_13
#define init_0_12_15_13
void FASTCALL call_0_12_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_15_14
#define init_0_12_15_14
void FASTCALL call_0_12_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_12_15_15
#define init_0_12_15_15
void FASTCALL call_0_12_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_0_0
#define init_0_13_0_0
void FASTCALL call_0_13_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_0_1
#define init_0_13_0_1
void FASTCALL call_0_13_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_0_2
#define init_0_13_0_2
void FASTCALL call_0_13_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_0_3
#define init_0_13_0_3
void FASTCALL call_0_13_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_0_4
#define init_0_13_0_4
void FASTCALL call_0_13_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_0_5
#define init_0_13_0_5
void FASTCALL call_0_13_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_0_6
#define init_0_13_0_6
void FASTCALL call_0_13_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_0_7
#define init_0_13_0_7
void FASTCALL call_0_13_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_0_8
#define init_0_13_0_8
void FASTCALL call_0_13_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_0_9
#define init_0_13_0_9
void FASTCALL call_0_13_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_0_10
#define init_0_13_0_10
void FASTCALL call_0_13_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_0_11
#define init_0_13_0_11
void FASTCALL call_0_13_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_0_12
#define init_0_13_0_12
void FASTCALL call_0_13_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_0_13
#define init_0_13_0_13
void FASTCALL call_0_13_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_0_14
#define init_0_13_0_14
void FASTCALL call_0_13_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_0_15
#define init_0_13_0_15
void FASTCALL call_0_13_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_1_0
#define init_0_13_1_0
void FASTCALL call_0_13_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_1_1
#define init_0_13_1_1
void FASTCALL call_0_13_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_1_2
#define init_0_13_1_2
void FASTCALL call_0_13_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_1_3
#define init_0_13_1_3
void FASTCALL call_0_13_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_1_4
#define init_0_13_1_4
void FASTCALL call_0_13_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_1_5
#define init_0_13_1_5
void FASTCALL call_0_13_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_1_6
#define init_0_13_1_6
void FASTCALL call_0_13_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_1_7
#define init_0_13_1_7
void FASTCALL call_0_13_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_1_8
#define init_0_13_1_8
void FASTCALL call_0_13_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_1_9
#define init_0_13_1_9
void FASTCALL call_0_13_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_1_10
#define init_0_13_1_10
void FASTCALL call_0_13_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_1_11
#define init_0_13_1_11
void FASTCALL call_0_13_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_1_12
#define init_0_13_1_12
void FASTCALL call_0_13_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_1_13
#define init_0_13_1_13
void FASTCALL call_0_13_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_1_14
#define init_0_13_1_14
void FASTCALL call_0_13_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_1_15
#define init_0_13_1_15
void FASTCALL call_0_13_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_2_0
#define init_0_13_2_0
void FASTCALL call_0_13_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_2_1
#define init_0_13_2_1
void FASTCALL call_0_13_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_2_2
#define init_0_13_2_2
void FASTCALL call_0_13_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_2_3
#define init_0_13_2_3
void FASTCALL call_0_13_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_2_4
#define init_0_13_2_4
void FASTCALL call_0_13_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_2_5
#define init_0_13_2_5
void FASTCALL call_0_13_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_2_6
#define init_0_13_2_6
void FASTCALL call_0_13_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_2_7
#define init_0_13_2_7
void FASTCALL call_0_13_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_2_8
#define init_0_13_2_8
void FASTCALL call_0_13_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_2_9
#define init_0_13_2_9
void FASTCALL call_0_13_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_2_10
#define init_0_13_2_10
void FASTCALL call_0_13_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_2_11
#define init_0_13_2_11
void FASTCALL call_0_13_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_2_12
#define init_0_13_2_12
void FASTCALL call_0_13_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_2_13
#define init_0_13_2_13
void FASTCALL call_0_13_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_2_14
#define init_0_13_2_14
void FASTCALL call_0_13_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_2_15
#define init_0_13_2_15
void FASTCALL call_0_13_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_3_0
#define init_0_13_3_0
void FASTCALL call_0_13_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_3_1
#define init_0_13_3_1
void FASTCALL call_0_13_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_3_2
#define init_0_13_3_2
void FASTCALL call_0_13_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_3_3
#define init_0_13_3_3
void FASTCALL call_0_13_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_3_4
#define init_0_13_3_4
void FASTCALL call_0_13_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_3_5
#define init_0_13_3_5
void FASTCALL call_0_13_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_3_6
#define init_0_13_3_6
void FASTCALL call_0_13_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_3_7
#define init_0_13_3_7
void FASTCALL call_0_13_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_3_8
#define init_0_13_3_8
void FASTCALL call_0_13_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_3_9
#define init_0_13_3_9
void FASTCALL call_0_13_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_3_10
#define init_0_13_3_10
void FASTCALL call_0_13_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_3_11
#define init_0_13_3_11
void FASTCALL call_0_13_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_3_12
#define init_0_13_3_12
void FASTCALL call_0_13_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_3_13
#define init_0_13_3_13
void FASTCALL call_0_13_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_3_14
#define init_0_13_3_14
void FASTCALL call_0_13_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_3_15
#define init_0_13_3_15
void FASTCALL call_0_13_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_4_0
#define init_0_13_4_0
void FASTCALL call_0_13_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_4_1
#define init_0_13_4_1
void FASTCALL call_0_13_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_4_2
#define init_0_13_4_2
void FASTCALL call_0_13_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_4_3
#define init_0_13_4_3
void FASTCALL call_0_13_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_4_4
#define init_0_13_4_4
void FASTCALL call_0_13_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_4_5
#define init_0_13_4_5
void FASTCALL call_0_13_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_4_6
#define init_0_13_4_6
void FASTCALL call_0_13_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_4_7
#define init_0_13_4_7
void FASTCALL call_0_13_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_4_8
#define init_0_13_4_8
void FASTCALL call_0_13_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_4_9
#define init_0_13_4_9
void FASTCALL call_0_13_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_4_10
#define init_0_13_4_10
void FASTCALL call_0_13_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_4_11
#define init_0_13_4_11
void FASTCALL call_0_13_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_4_12
#define init_0_13_4_12
void FASTCALL call_0_13_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_4_13
#define init_0_13_4_13
void FASTCALL call_0_13_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_4_14
#define init_0_13_4_14
void FASTCALL call_0_13_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_4_15
#define init_0_13_4_15
void FASTCALL call_0_13_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_5_0
#define init_0_13_5_0
void FASTCALL call_0_13_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_5_1
#define init_0_13_5_1
void FASTCALL call_0_13_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_5_2
#define init_0_13_5_2
void FASTCALL call_0_13_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_5_3
#define init_0_13_5_3
void FASTCALL call_0_13_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_5_4
#define init_0_13_5_4
void FASTCALL call_0_13_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_5_5
#define init_0_13_5_5
void FASTCALL call_0_13_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_5_6
#define init_0_13_5_6
void FASTCALL call_0_13_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_5_7
#define init_0_13_5_7
void FASTCALL call_0_13_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_5_8
#define init_0_13_5_8
void FASTCALL call_0_13_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_5_9
#define init_0_13_5_9
void FASTCALL call_0_13_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_5_10
#define init_0_13_5_10
void FASTCALL call_0_13_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_5_11
#define init_0_13_5_11
void FASTCALL call_0_13_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_5_12
#define init_0_13_5_12
void FASTCALL call_0_13_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_5_13
#define init_0_13_5_13
void FASTCALL call_0_13_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_5_14
#define init_0_13_5_14
void FASTCALL call_0_13_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_5_15
#define init_0_13_5_15
void FASTCALL call_0_13_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_6_0
#define init_0_13_6_0
void FASTCALL call_0_13_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_6_1
#define init_0_13_6_1
void FASTCALL call_0_13_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_6_2
#define init_0_13_6_2
void FASTCALL call_0_13_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_6_3
#define init_0_13_6_3
void FASTCALL call_0_13_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_6_4
#define init_0_13_6_4
void FASTCALL call_0_13_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_6_5
#define init_0_13_6_5
void FASTCALL call_0_13_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_6_6
#define init_0_13_6_6
void FASTCALL call_0_13_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_6_7
#define init_0_13_6_7
void FASTCALL call_0_13_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_6_8
#define init_0_13_6_8
void FASTCALL call_0_13_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_6_9
#define init_0_13_6_9
void FASTCALL call_0_13_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_6_10
#define init_0_13_6_10
void FASTCALL call_0_13_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_6_11
#define init_0_13_6_11
void FASTCALL call_0_13_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_6_12
#define init_0_13_6_12
void FASTCALL call_0_13_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_6_13
#define init_0_13_6_13
void FASTCALL call_0_13_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_6_14
#define init_0_13_6_14
void FASTCALL call_0_13_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_6_15
#define init_0_13_6_15
void FASTCALL call_0_13_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_7_0
#define init_0_13_7_0
void FASTCALL call_0_13_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_7_1
#define init_0_13_7_1
void FASTCALL call_0_13_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_7_2
#define init_0_13_7_2
void FASTCALL call_0_13_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_7_3
#define init_0_13_7_3
void FASTCALL call_0_13_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_7_4
#define init_0_13_7_4
void FASTCALL call_0_13_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_7_5
#define init_0_13_7_5
void FASTCALL call_0_13_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_7_6
#define init_0_13_7_6
void FASTCALL call_0_13_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_7_7
#define init_0_13_7_7
void FASTCALL call_0_13_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_7_8
#define init_0_13_7_8
void FASTCALL call_0_13_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_7_9
#define init_0_13_7_9
void FASTCALL call_0_13_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_7_10
#define init_0_13_7_10
void FASTCALL call_0_13_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_7_11
#define init_0_13_7_11
void FASTCALL call_0_13_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_7_12
#define init_0_13_7_12
void FASTCALL call_0_13_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_7_13
#define init_0_13_7_13
void FASTCALL call_0_13_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_7_14
#define init_0_13_7_14
void FASTCALL call_0_13_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_7_15
#define init_0_13_7_15
void FASTCALL call_0_13_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_8_0
#define init_0_13_8_0
void FASTCALL call_0_13_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_8_1
#define init_0_13_8_1
void FASTCALL call_0_13_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_8_2
#define init_0_13_8_2
void FASTCALL call_0_13_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_8_3
#define init_0_13_8_3
void FASTCALL call_0_13_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_8_4
#define init_0_13_8_4
void FASTCALL call_0_13_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_8_5
#define init_0_13_8_5
void FASTCALL call_0_13_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_8_6
#define init_0_13_8_6
void FASTCALL call_0_13_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_8_7
#define init_0_13_8_7
void FASTCALL call_0_13_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_8_8
#define init_0_13_8_8
void FASTCALL call_0_13_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_8_9
#define init_0_13_8_9
void FASTCALL call_0_13_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_8_10
#define init_0_13_8_10
void FASTCALL call_0_13_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_8_11
#define init_0_13_8_11
void FASTCALL call_0_13_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_8_12
#define init_0_13_8_12
void FASTCALL call_0_13_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_8_13
#define init_0_13_8_13
void FASTCALL call_0_13_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_8_14
#define init_0_13_8_14
void FASTCALL call_0_13_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_8_15
#define init_0_13_8_15
void FASTCALL call_0_13_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_9_0
#define init_0_13_9_0
void FASTCALL call_0_13_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_9_1
#define init_0_13_9_1
void FASTCALL call_0_13_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_9_2
#define init_0_13_9_2
void FASTCALL call_0_13_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_9_3
#define init_0_13_9_3
void FASTCALL call_0_13_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_9_4
#define init_0_13_9_4
void FASTCALL call_0_13_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_9_5
#define init_0_13_9_5
void FASTCALL call_0_13_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_9_6
#define init_0_13_9_6
void FASTCALL call_0_13_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_9_7
#define init_0_13_9_7
void FASTCALL call_0_13_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_9_8
#define init_0_13_9_8
void FASTCALL call_0_13_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_9_9
#define init_0_13_9_9
void FASTCALL call_0_13_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_9_10
#define init_0_13_9_10
void FASTCALL call_0_13_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_9_11
#define init_0_13_9_11
void FASTCALL call_0_13_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_9_12
#define init_0_13_9_12
void FASTCALL call_0_13_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_9_13
#define init_0_13_9_13
void FASTCALL call_0_13_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_9_14
#define init_0_13_9_14
void FASTCALL call_0_13_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_9_15
#define init_0_13_9_15
void FASTCALL call_0_13_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_10_0
#define init_0_13_10_0
void FASTCALL call_0_13_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_10_1
#define init_0_13_10_1
void FASTCALL call_0_13_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_10_2
#define init_0_13_10_2
void FASTCALL call_0_13_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_10_3
#define init_0_13_10_3
void FASTCALL call_0_13_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_10_4
#define init_0_13_10_4
void FASTCALL call_0_13_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_10_5
#define init_0_13_10_5
void FASTCALL call_0_13_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_10_6
#define init_0_13_10_6
void FASTCALL call_0_13_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_10_7
#define init_0_13_10_7
void FASTCALL call_0_13_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_10_8
#define init_0_13_10_8
void FASTCALL call_0_13_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_10_9
#define init_0_13_10_9
void FASTCALL call_0_13_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_10_10
#define init_0_13_10_10
void FASTCALL call_0_13_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_10_11
#define init_0_13_10_11
void FASTCALL call_0_13_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_10_12
#define init_0_13_10_12
void FASTCALL call_0_13_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_10_13
#define init_0_13_10_13
void FASTCALL call_0_13_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_10_14
#define init_0_13_10_14
void FASTCALL call_0_13_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_10_15
#define init_0_13_10_15
void FASTCALL call_0_13_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_11_0
#define init_0_13_11_0
void FASTCALL call_0_13_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_11_1
#define init_0_13_11_1
void FASTCALL call_0_13_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_11_2
#define init_0_13_11_2
void FASTCALL call_0_13_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_11_3
#define init_0_13_11_3
void FASTCALL call_0_13_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_11_4
#define init_0_13_11_4
void FASTCALL call_0_13_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_11_5
#define init_0_13_11_5
void FASTCALL call_0_13_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_11_6
#define init_0_13_11_6
void FASTCALL call_0_13_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_11_7
#define init_0_13_11_7
void FASTCALL call_0_13_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_11_8
#define init_0_13_11_8
void FASTCALL call_0_13_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_11_9
#define init_0_13_11_9
void FASTCALL call_0_13_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_11_10
#define init_0_13_11_10
void FASTCALL call_0_13_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_11_11
#define init_0_13_11_11
void FASTCALL call_0_13_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_11_12
#define init_0_13_11_12
void FASTCALL call_0_13_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_11_13
#define init_0_13_11_13
void FASTCALL call_0_13_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_11_14
#define init_0_13_11_14
void FASTCALL call_0_13_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_11_15
#define init_0_13_11_15
void FASTCALL call_0_13_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_12_0
#define init_0_13_12_0
void FASTCALL call_0_13_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_12_1
#define init_0_13_12_1
void FASTCALL call_0_13_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_12_2
#define init_0_13_12_2
void FASTCALL call_0_13_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_12_3
#define init_0_13_12_3
void FASTCALL call_0_13_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_12_4
#define init_0_13_12_4
void FASTCALL call_0_13_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_12_5
#define init_0_13_12_5
void FASTCALL call_0_13_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_12_6
#define init_0_13_12_6
void FASTCALL call_0_13_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_12_7
#define init_0_13_12_7
void FASTCALL call_0_13_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_12_8
#define init_0_13_12_8
void FASTCALL call_0_13_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_12_9
#define init_0_13_12_9
void FASTCALL call_0_13_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_12_10
#define init_0_13_12_10
void FASTCALL call_0_13_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_12_11
#define init_0_13_12_11
void FASTCALL call_0_13_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_12_12
#define init_0_13_12_12
void FASTCALL call_0_13_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_12_13
#define init_0_13_12_13
void FASTCALL call_0_13_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_12_14
#define init_0_13_12_14
void FASTCALL call_0_13_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_12_15
#define init_0_13_12_15
void FASTCALL call_0_13_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_13_0
#define init_0_13_13_0
void FASTCALL call_0_13_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_13_1
#define init_0_13_13_1
void FASTCALL call_0_13_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_13_2
#define init_0_13_13_2
void FASTCALL call_0_13_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_13_3
#define init_0_13_13_3
void FASTCALL call_0_13_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_13_4
#define init_0_13_13_4
void FASTCALL call_0_13_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_13_5
#define init_0_13_13_5
void FASTCALL call_0_13_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_13_6
#define init_0_13_13_6
void FASTCALL call_0_13_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_13_7
#define init_0_13_13_7
void FASTCALL call_0_13_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_13_8
#define init_0_13_13_8
void FASTCALL call_0_13_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_13_9
#define init_0_13_13_9
void FASTCALL call_0_13_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_13_10
#define init_0_13_13_10
void FASTCALL call_0_13_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_13_11
#define init_0_13_13_11
void FASTCALL call_0_13_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_13_12
#define init_0_13_13_12
void FASTCALL call_0_13_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_13_13
#define init_0_13_13_13
void FASTCALL call_0_13_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_13_14
#define init_0_13_13_14
void FASTCALL call_0_13_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_13_15
#define init_0_13_13_15
void FASTCALL call_0_13_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_14_0
#define init_0_13_14_0
void FASTCALL call_0_13_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_14_1
#define init_0_13_14_1
void FASTCALL call_0_13_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_14_2
#define init_0_13_14_2
void FASTCALL call_0_13_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_14_3
#define init_0_13_14_3
void FASTCALL call_0_13_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_14_4
#define init_0_13_14_4
void FASTCALL call_0_13_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_14_5
#define init_0_13_14_5
void FASTCALL call_0_13_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_14_6
#define init_0_13_14_6
void FASTCALL call_0_13_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_14_7
#define init_0_13_14_7
void FASTCALL call_0_13_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_14_8
#define init_0_13_14_8
void FASTCALL call_0_13_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_14_9
#define init_0_13_14_9
void FASTCALL call_0_13_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_14_10
#define init_0_13_14_10
void FASTCALL call_0_13_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_14_11
#define init_0_13_14_11
void FASTCALL call_0_13_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_14_12
#define init_0_13_14_12
void FASTCALL call_0_13_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_14_13
#define init_0_13_14_13
void FASTCALL call_0_13_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_14_14
#define init_0_13_14_14
void FASTCALL call_0_13_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_14_15
#define init_0_13_14_15
void FASTCALL call_0_13_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_15_0
#define init_0_13_15_0
void FASTCALL call_0_13_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_15_1
#define init_0_13_15_1
void FASTCALL call_0_13_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_15_2
#define init_0_13_15_2
void FASTCALL call_0_13_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_15_3
#define init_0_13_15_3
void FASTCALL call_0_13_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_15_4
#define init_0_13_15_4
void FASTCALL call_0_13_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_15_5
#define init_0_13_15_5
void FASTCALL call_0_13_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_15_6
#define init_0_13_15_6
void FASTCALL call_0_13_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_15_7
#define init_0_13_15_7
void FASTCALL call_0_13_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_15_8
#define init_0_13_15_8
void FASTCALL call_0_13_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_15_9
#define init_0_13_15_9
void FASTCALL call_0_13_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_15_10
#define init_0_13_15_10
void FASTCALL call_0_13_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_15_11
#define init_0_13_15_11
void FASTCALL call_0_13_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_15_12
#define init_0_13_15_12
void FASTCALL call_0_13_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_15_13
#define init_0_13_15_13
void FASTCALL call_0_13_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_15_14
#define init_0_13_15_14
void FASTCALL call_0_13_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_13_15_15
#define init_0_13_15_15
void FASTCALL call_0_13_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_0_0
#define init_0_14_0_0
void FASTCALL call_0_14_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_0_1
#define init_0_14_0_1
void FASTCALL call_0_14_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_0_2
#define init_0_14_0_2
void FASTCALL call_0_14_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_0_3
#define init_0_14_0_3
void FASTCALL call_0_14_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_0_4
#define init_0_14_0_4
void FASTCALL call_0_14_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_0_5
#define init_0_14_0_5
void FASTCALL call_0_14_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_0_6
#define init_0_14_0_6
void FASTCALL call_0_14_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_0_7
#define init_0_14_0_7
void FASTCALL call_0_14_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_0_8
#define init_0_14_0_8
void FASTCALL call_0_14_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_0_9
#define init_0_14_0_9
void FASTCALL call_0_14_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_0_10
#define init_0_14_0_10
void FASTCALL call_0_14_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_0_11
#define init_0_14_0_11
void FASTCALL call_0_14_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_0_12
#define init_0_14_0_12
void FASTCALL call_0_14_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_0_13
#define init_0_14_0_13
void FASTCALL call_0_14_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_0_14
#define init_0_14_0_14
void FASTCALL call_0_14_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_0_15
#define init_0_14_0_15
void FASTCALL call_0_14_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_1_0
#define init_0_14_1_0
void FASTCALL call_0_14_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_1_1
#define init_0_14_1_1
void FASTCALL call_0_14_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_1_2
#define init_0_14_1_2
void FASTCALL call_0_14_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_1_3
#define init_0_14_1_3
void FASTCALL call_0_14_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_1_4
#define init_0_14_1_4
void FASTCALL call_0_14_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_1_5
#define init_0_14_1_5
void FASTCALL call_0_14_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_1_6
#define init_0_14_1_6
void FASTCALL call_0_14_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_1_7
#define init_0_14_1_7
void FASTCALL call_0_14_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_1_8
#define init_0_14_1_8
void FASTCALL call_0_14_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_1_9
#define init_0_14_1_9
void FASTCALL call_0_14_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_1_10
#define init_0_14_1_10
void FASTCALL call_0_14_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_1_11
#define init_0_14_1_11
void FASTCALL call_0_14_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_1_12
#define init_0_14_1_12
void FASTCALL call_0_14_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_1_13
#define init_0_14_1_13
void FASTCALL call_0_14_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_1_14
#define init_0_14_1_14
void FASTCALL call_0_14_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_1_15
#define init_0_14_1_15
void FASTCALL call_0_14_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_2_0
#define init_0_14_2_0
void FASTCALL call_0_14_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_2_1
#define init_0_14_2_1
void FASTCALL call_0_14_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_2_2
#define init_0_14_2_2
void FASTCALL call_0_14_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_2_3
#define init_0_14_2_3
void FASTCALL call_0_14_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_2_4
#define init_0_14_2_4
void FASTCALL call_0_14_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_2_5
#define init_0_14_2_5
void FASTCALL call_0_14_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_2_6
#define init_0_14_2_6
void FASTCALL call_0_14_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_2_7
#define init_0_14_2_7
void FASTCALL call_0_14_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_2_8
#define init_0_14_2_8
void FASTCALL call_0_14_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_2_9
#define init_0_14_2_9
void FASTCALL call_0_14_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_2_10
#define init_0_14_2_10
void FASTCALL call_0_14_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_2_11
#define init_0_14_2_11
void FASTCALL call_0_14_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_2_12
#define init_0_14_2_12
void FASTCALL call_0_14_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_2_13
#define init_0_14_2_13
void FASTCALL call_0_14_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_2_14
#define init_0_14_2_14
void FASTCALL call_0_14_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_2_15
#define init_0_14_2_15
void FASTCALL call_0_14_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_3_0
#define init_0_14_3_0
void FASTCALL call_0_14_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_3_1
#define init_0_14_3_1
void FASTCALL call_0_14_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_3_2
#define init_0_14_3_2
void FASTCALL call_0_14_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_3_3
#define init_0_14_3_3
void FASTCALL call_0_14_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_3_4
#define init_0_14_3_4
void FASTCALL call_0_14_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_3_5
#define init_0_14_3_5
void FASTCALL call_0_14_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_3_6
#define init_0_14_3_6
void FASTCALL call_0_14_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_3_7
#define init_0_14_3_7
void FASTCALL call_0_14_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_3_8
#define init_0_14_3_8
void FASTCALL call_0_14_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_3_9
#define init_0_14_3_9
void FASTCALL call_0_14_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_3_10
#define init_0_14_3_10
void FASTCALL call_0_14_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_3_11
#define init_0_14_3_11
void FASTCALL call_0_14_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_3_12
#define init_0_14_3_12
void FASTCALL call_0_14_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_3_13
#define init_0_14_3_13
void FASTCALL call_0_14_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_3_14
#define init_0_14_3_14
void FASTCALL call_0_14_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_3_15
#define init_0_14_3_15
void FASTCALL call_0_14_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_4_0
#define init_0_14_4_0
void FASTCALL call_0_14_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_4_1
#define init_0_14_4_1
void FASTCALL call_0_14_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_4_2
#define init_0_14_4_2
void FASTCALL call_0_14_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_4_3
#define init_0_14_4_3
void FASTCALL call_0_14_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_4_4
#define init_0_14_4_4
void FASTCALL call_0_14_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_4_5
#define init_0_14_4_5
void FASTCALL call_0_14_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_4_6
#define init_0_14_4_6
void FASTCALL call_0_14_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_4_7
#define init_0_14_4_7
void FASTCALL call_0_14_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_4_8
#define init_0_14_4_8
void FASTCALL call_0_14_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_4_9
#define init_0_14_4_9
void FASTCALL call_0_14_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_4_10
#define init_0_14_4_10
void FASTCALL call_0_14_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_4_11
#define init_0_14_4_11
void FASTCALL call_0_14_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_4_12
#define init_0_14_4_12
void FASTCALL call_0_14_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_4_13
#define init_0_14_4_13
void FASTCALL call_0_14_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_4_14
#define init_0_14_4_14
void FASTCALL call_0_14_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_4_15
#define init_0_14_4_15
void FASTCALL call_0_14_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_5_0
#define init_0_14_5_0
void FASTCALL call_0_14_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_5_1
#define init_0_14_5_1
void FASTCALL call_0_14_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_5_2
#define init_0_14_5_2
void FASTCALL call_0_14_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_5_3
#define init_0_14_5_3
void FASTCALL call_0_14_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_5_4
#define init_0_14_5_4
void FASTCALL call_0_14_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_5_5
#define init_0_14_5_5
void FASTCALL call_0_14_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_5_6
#define init_0_14_5_6
void FASTCALL call_0_14_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_5_7
#define init_0_14_5_7
void FASTCALL call_0_14_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_5_8
#define init_0_14_5_8
void FASTCALL call_0_14_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_5_9
#define init_0_14_5_9
void FASTCALL call_0_14_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_5_10
#define init_0_14_5_10
void FASTCALL call_0_14_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_5_11
#define init_0_14_5_11
void FASTCALL call_0_14_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_5_12
#define init_0_14_5_12
void FASTCALL call_0_14_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_5_13
#define init_0_14_5_13
void FASTCALL call_0_14_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_5_14
#define init_0_14_5_14
void FASTCALL call_0_14_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_5_15
#define init_0_14_5_15
void FASTCALL call_0_14_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_6_0
#define init_0_14_6_0
void FASTCALL call_0_14_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_6_1
#define init_0_14_6_1
void FASTCALL call_0_14_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_6_2
#define init_0_14_6_2
void FASTCALL call_0_14_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_6_3
#define init_0_14_6_3
void FASTCALL call_0_14_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_6_4
#define init_0_14_6_4
void FASTCALL call_0_14_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_6_5
#define init_0_14_6_5
void FASTCALL call_0_14_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_6_6
#define init_0_14_6_6
void FASTCALL call_0_14_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_6_7
#define init_0_14_6_7
void FASTCALL call_0_14_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_6_8
#define init_0_14_6_8
void FASTCALL call_0_14_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_6_9
#define init_0_14_6_9
void FASTCALL call_0_14_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_6_10
#define init_0_14_6_10
void FASTCALL call_0_14_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_6_11
#define init_0_14_6_11
void FASTCALL call_0_14_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_6_12
#define init_0_14_6_12
void FASTCALL call_0_14_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_6_13
#define init_0_14_6_13
void FASTCALL call_0_14_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_6_14
#define init_0_14_6_14
void FASTCALL call_0_14_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_6_15
#define init_0_14_6_15
void FASTCALL call_0_14_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_7_0
#define init_0_14_7_0
void FASTCALL call_0_14_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_7_1
#define init_0_14_7_1
void FASTCALL call_0_14_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_7_2
#define init_0_14_7_2
void FASTCALL call_0_14_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_7_3
#define init_0_14_7_3
void FASTCALL call_0_14_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_7_4
#define init_0_14_7_4
void FASTCALL call_0_14_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_7_5
#define init_0_14_7_5
void FASTCALL call_0_14_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_7_6
#define init_0_14_7_6
void FASTCALL call_0_14_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_7_7
#define init_0_14_7_7
void FASTCALL call_0_14_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_7_8
#define init_0_14_7_8
void FASTCALL call_0_14_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_7_9
#define init_0_14_7_9
void FASTCALL call_0_14_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_7_10
#define init_0_14_7_10
void FASTCALL call_0_14_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_7_11
#define init_0_14_7_11
void FASTCALL call_0_14_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_7_12
#define init_0_14_7_12
void FASTCALL call_0_14_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_7_13
#define init_0_14_7_13
void FASTCALL call_0_14_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_7_14
#define init_0_14_7_14
void FASTCALL call_0_14_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_7_15
#define init_0_14_7_15
void FASTCALL call_0_14_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_8_0
#define init_0_14_8_0
void FASTCALL call_0_14_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_8_1
#define init_0_14_8_1
void FASTCALL call_0_14_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_8_2
#define init_0_14_8_2
void FASTCALL call_0_14_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_8_3
#define init_0_14_8_3
void FASTCALL call_0_14_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_8_4
#define init_0_14_8_4
void FASTCALL call_0_14_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_8_5
#define init_0_14_8_5
void FASTCALL call_0_14_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_8_6
#define init_0_14_8_6
void FASTCALL call_0_14_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_8_7
#define init_0_14_8_7
void FASTCALL call_0_14_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_8_8
#define init_0_14_8_8
void FASTCALL call_0_14_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_8_9
#define init_0_14_8_9
void FASTCALL call_0_14_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_8_10
#define init_0_14_8_10
void FASTCALL call_0_14_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_8_11
#define init_0_14_8_11
void FASTCALL call_0_14_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_8_12
#define init_0_14_8_12
void FASTCALL call_0_14_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_8_13
#define init_0_14_8_13
void FASTCALL call_0_14_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_8_14
#define init_0_14_8_14
void FASTCALL call_0_14_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_8_15
#define init_0_14_8_15
void FASTCALL call_0_14_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_9_0
#define init_0_14_9_0
void FASTCALL call_0_14_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_9_1
#define init_0_14_9_1
void FASTCALL call_0_14_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_9_2
#define init_0_14_9_2
void FASTCALL call_0_14_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_9_3
#define init_0_14_9_3
void FASTCALL call_0_14_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_9_4
#define init_0_14_9_4
void FASTCALL call_0_14_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_9_5
#define init_0_14_9_5
void FASTCALL call_0_14_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_9_6
#define init_0_14_9_6
void FASTCALL call_0_14_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_9_7
#define init_0_14_9_7
void FASTCALL call_0_14_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_9_8
#define init_0_14_9_8
void FASTCALL call_0_14_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_9_9
#define init_0_14_9_9
void FASTCALL call_0_14_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_9_10
#define init_0_14_9_10
void FASTCALL call_0_14_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_9_11
#define init_0_14_9_11
void FASTCALL call_0_14_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_9_12
#define init_0_14_9_12
void FASTCALL call_0_14_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_9_13
#define init_0_14_9_13
void FASTCALL call_0_14_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_9_14
#define init_0_14_9_14
void FASTCALL call_0_14_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_9_15
#define init_0_14_9_15
void FASTCALL call_0_14_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_10_0
#define init_0_14_10_0
void FASTCALL call_0_14_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_10_1
#define init_0_14_10_1
void FASTCALL call_0_14_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_10_2
#define init_0_14_10_2
void FASTCALL call_0_14_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_10_3
#define init_0_14_10_3
void FASTCALL call_0_14_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_10_4
#define init_0_14_10_4
void FASTCALL call_0_14_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_10_5
#define init_0_14_10_5
void FASTCALL call_0_14_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_10_6
#define init_0_14_10_6
void FASTCALL call_0_14_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_10_7
#define init_0_14_10_7
void FASTCALL call_0_14_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_10_8
#define init_0_14_10_8
void FASTCALL call_0_14_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_10_9
#define init_0_14_10_9
void FASTCALL call_0_14_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_10_10
#define init_0_14_10_10
void FASTCALL call_0_14_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_10_11
#define init_0_14_10_11
void FASTCALL call_0_14_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_10_12
#define init_0_14_10_12
void FASTCALL call_0_14_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_10_13
#define init_0_14_10_13
void FASTCALL call_0_14_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_10_14
#define init_0_14_10_14
void FASTCALL call_0_14_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_10_15
#define init_0_14_10_15
void FASTCALL call_0_14_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_11_0
#define init_0_14_11_0
void FASTCALL call_0_14_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_11_1
#define init_0_14_11_1
void FASTCALL call_0_14_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_11_2
#define init_0_14_11_2
void FASTCALL call_0_14_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_11_3
#define init_0_14_11_3
void FASTCALL call_0_14_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_11_4
#define init_0_14_11_4
void FASTCALL call_0_14_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_11_5
#define init_0_14_11_5
void FASTCALL call_0_14_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_11_6
#define init_0_14_11_6
void FASTCALL call_0_14_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_11_7
#define init_0_14_11_7
void FASTCALL call_0_14_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_11_8
#define init_0_14_11_8
void FASTCALL call_0_14_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_11_9
#define init_0_14_11_9
void FASTCALL call_0_14_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_11_10
#define init_0_14_11_10
void FASTCALL call_0_14_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_11_11
#define init_0_14_11_11
void FASTCALL call_0_14_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_11_12
#define init_0_14_11_12
void FASTCALL call_0_14_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_11_13
#define init_0_14_11_13
void FASTCALL call_0_14_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_11_14
#define init_0_14_11_14
void FASTCALL call_0_14_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_11_15
#define init_0_14_11_15
void FASTCALL call_0_14_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_12_0
#define init_0_14_12_0
void FASTCALL call_0_14_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_12_1
#define init_0_14_12_1
void FASTCALL call_0_14_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_12_2
#define init_0_14_12_2
void FASTCALL call_0_14_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_12_3
#define init_0_14_12_3
void FASTCALL call_0_14_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_12_4
#define init_0_14_12_4
void FASTCALL call_0_14_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_12_5
#define init_0_14_12_5
void FASTCALL call_0_14_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_12_6
#define init_0_14_12_6
void FASTCALL call_0_14_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_12_7
#define init_0_14_12_7
void FASTCALL call_0_14_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_12_8
#define init_0_14_12_8
void FASTCALL call_0_14_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_12_9
#define init_0_14_12_9
void FASTCALL call_0_14_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_12_10
#define init_0_14_12_10
void FASTCALL call_0_14_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_12_11
#define init_0_14_12_11
void FASTCALL call_0_14_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_12_12
#define init_0_14_12_12
void FASTCALL call_0_14_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_12_13
#define init_0_14_12_13
void FASTCALL call_0_14_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_12_14
#define init_0_14_12_14
void FASTCALL call_0_14_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_12_15
#define init_0_14_12_15
void FASTCALL call_0_14_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_13_0
#define init_0_14_13_0
void FASTCALL call_0_14_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_13_1
#define init_0_14_13_1
void FASTCALL call_0_14_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_13_2
#define init_0_14_13_2
void FASTCALL call_0_14_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_13_3
#define init_0_14_13_3
void FASTCALL call_0_14_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_13_4
#define init_0_14_13_4
void FASTCALL call_0_14_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_13_5
#define init_0_14_13_5
void FASTCALL call_0_14_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_13_6
#define init_0_14_13_6
void FASTCALL call_0_14_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_13_7
#define init_0_14_13_7
void FASTCALL call_0_14_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_13_8
#define init_0_14_13_8
void FASTCALL call_0_14_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_13_9
#define init_0_14_13_9
void FASTCALL call_0_14_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_13_10
#define init_0_14_13_10
void FASTCALL call_0_14_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_13_11
#define init_0_14_13_11
void FASTCALL call_0_14_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_13_12
#define init_0_14_13_12
void FASTCALL call_0_14_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_13_13
#define init_0_14_13_13
void FASTCALL call_0_14_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_13_14
#define init_0_14_13_14
void FASTCALL call_0_14_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_13_15
#define init_0_14_13_15
void FASTCALL call_0_14_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_14_0
#define init_0_14_14_0
void FASTCALL call_0_14_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_14_1
#define init_0_14_14_1
void FASTCALL call_0_14_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_14_2
#define init_0_14_14_2
void FASTCALL call_0_14_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_14_3
#define init_0_14_14_3
void FASTCALL call_0_14_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_14_4
#define init_0_14_14_4
void FASTCALL call_0_14_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_14_5
#define init_0_14_14_5
void FASTCALL call_0_14_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_14_6
#define init_0_14_14_6
void FASTCALL call_0_14_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_14_7
#define init_0_14_14_7
void FASTCALL call_0_14_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_14_8
#define init_0_14_14_8
void FASTCALL call_0_14_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_14_9
#define init_0_14_14_9
void FASTCALL call_0_14_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_14_10
#define init_0_14_14_10
void FASTCALL call_0_14_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_14_11
#define init_0_14_14_11
void FASTCALL call_0_14_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_14_12
#define init_0_14_14_12
void FASTCALL call_0_14_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_14_13
#define init_0_14_14_13
void FASTCALL call_0_14_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_14_14
#define init_0_14_14_14
void FASTCALL call_0_14_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_14_15
#define init_0_14_14_15
void FASTCALL call_0_14_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_15_0
#define init_0_14_15_0
void FASTCALL call_0_14_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_15_1
#define init_0_14_15_1
void FASTCALL call_0_14_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_15_2
#define init_0_14_15_2
void FASTCALL call_0_14_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_15_3
#define init_0_14_15_3
void FASTCALL call_0_14_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_15_4
#define init_0_14_15_4
void FASTCALL call_0_14_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_15_5
#define init_0_14_15_5
void FASTCALL call_0_14_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_15_6
#define init_0_14_15_6
void FASTCALL call_0_14_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_15_7
#define init_0_14_15_7
void FASTCALL call_0_14_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_15_8
#define init_0_14_15_8
void FASTCALL call_0_14_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_15_9
#define init_0_14_15_9
void FASTCALL call_0_14_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_15_10
#define init_0_14_15_10
void FASTCALL call_0_14_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_15_11
#define init_0_14_15_11
void FASTCALL call_0_14_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_15_12
#define init_0_14_15_12
void FASTCALL call_0_14_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_15_13
#define init_0_14_15_13
void FASTCALL call_0_14_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_15_14
#define init_0_14_15_14
void FASTCALL call_0_14_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_14_15_15
#define init_0_14_15_15
void FASTCALL call_0_14_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_0_0
#define init_0_15_0_0
void FASTCALL call_0_15_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_0_1
#define init_0_15_0_1
void FASTCALL call_0_15_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_0_2
#define init_0_15_0_2
void FASTCALL call_0_15_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_0_3
#define init_0_15_0_3
void FASTCALL call_0_15_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_0_4
#define init_0_15_0_4
void FASTCALL call_0_15_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_0_5
#define init_0_15_0_5
void FASTCALL call_0_15_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_0_6
#define init_0_15_0_6
void FASTCALL call_0_15_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_0_7
#define init_0_15_0_7
void FASTCALL call_0_15_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_0_8
#define init_0_15_0_8
void FASTCALL call_0_15_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_0_9
#define init_0_15_0_9
void FASTCALL call_0_15_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_0_10
#define init_0_15_0_10
void FASTCALL call_0_15_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_0_11
#define init_0_15_0_11
void FASTCALL call_0_15_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_0_12
#define init_0_15_0_12
void FASTCALL call_0_15_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_0_13
#define init_0_15_0_13
void FASTCALL call_0_15_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_0_14
#define init_0_15_0_14
void FASTCALL call_0_15_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_0_15
#define init_0_15_0_15
void FASTCALL call_0_15_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_1_0
#define init_0_15_1_0
void FASTCALL call_0_15_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_1_1
#define init_0_15_1_1
void FASTCALL call_0_15_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_1_2
#define init_0_15_1_2
void FASTCALL call_0_15_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_1_3
#define init_0_15_1_3
void FASTCALL call_0_15_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_1_4
#define init_0_15_1_4
void FASTCALL call_0_15_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_1_5
#define init_0_15_1_5
void FASTCALL call_0_15_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_1_6
#define init_0_15_1_6
void FASTCALL call_0_15_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_1_7
#define init_0_15_1_7
void FASTCALL call_0_15_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_1_8
#define init_0_15_1_8
void FASTCALL call_0_15_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_1_9
#define init_0_15_1_9
void FASTCALL call_0_15_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_1_10
#define init_0_15_1_10
void FASTCALL call_0_15_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_1_11
#define init_0_15_1_11
void FASTCALL call_0_15_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_1_12
#define init_0_15_1_12
void FASTCALL call_0_15_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_1_13
#define init_0_15_1_13
void FASTCALL call_0_15_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_1_14
#define init_0_15_1_14
void FASTCALL call_0_15_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_1_15
#define init_0_15_1_15
void FASTCALL call_0_15_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_2_0
#define init_0_15_2_0
void FASTCALL call_0_15_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_2_1
#define init_0_15_2_1
void FASTCALL call_0_15_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_2_2
#define init_0_15_2_2
void FASTCALL call_0_15_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_2_3
#define init_0_15_2_3
void FASTCALL call_0_15_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_2_4
#define init_0_15_2_4
void FASTCALL call_0_15_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_2_5
#define init_0_15_2_5
void FASTCALL call_0_15_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_2_6
#define init_0_15_2_6
void FASTCALL call_0_15_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_2_7
#define init_0_15_2_7
void FASTCALL call_0_15_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_2_8
#define init_0_15_2_8
void FASTCALL call_0_15_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_2_9
#define init_0_15_2_9
void FASTCALL call_0_15_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_2_10
#define init_0_15_2_10
void FASTCALL call_0_15_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_2_11
#define init_0_15_2_11
void FASTCALL call_0_15_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_2_12
#define init_0_15_2_12
void FASTCALL call_0_15_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_2_13
#define init_0_15_2_13
void FASTCALL call_0_15_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_2_14
#define init_0_15_2_14
void FASTCALL call_0_15_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_2_15
#define init_0_15_2_15
void FASTCALL call_0_15_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_3_0
#define init_0_15_3_0
void FASTCALL call_0_15_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_3_1
#define init_0_15_3_1
void FASTCALL call_0_15_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_3_2
#define init_0_15_3_2
void FASTCALL call_0_15_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_3_3
#define init_0_15_3_3
void FASTCALL call_0_15_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_3_4
#define init_0_15_3_4
void FASTCALL call_0_15_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_3_5
#define init_0_15_3_5
void FASTCALL call_0_15_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_3_6
#define init_0_15_3_6
void FASTCALL call_0_15_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_3_7
#define init_0_15_3_7
void FASTCALL call_0_15_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_3_8
#define init_0_15_3_8
void FASTCALL call_0_15_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_3_9
#define init_0_15_3_9
void FASTCALL call_0_15_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_3_10
#define init_0_15_3_10
void FASTCALL call_0_15_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_3_11
#define init_0_15_3_11
void FASTCALL call_0_15_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_3_12
#define init_0_15_3_12
void FASTCALL call_0_15_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_3_13
#define init_0_15_3_13
void FASTCALL call_0_15_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_3_14
#define init_0_15_3_14
void FASTCALL call_0_15_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_3_15
#define init_0_15_3_15
void FASTCALL call_0_15_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_4_0
#define init_0_15_4_0
void FASTCALL call_0_15_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_4_1
#define init_0_15_4_1
void FASTCALL call_0_15_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_4_2
#define init_0_15_4_2
void FASTCALL call_0_15_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_4_3
#define init_0_15_4_3
void FASTCALL call_0_15_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_4_4
#define init_0_15_4_4
void FASTCALL call_0_15_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_4_5
#define init_0_15_4_5
void FASTCALL call_0_15_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_4_6
#define init_0_15_4_6
void FASTCALL call_0_15_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_4_7
#define init_0_15_4_7
void FASTCALL call_0_15_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_4_8
#define init_0_15_4_8
void FASTCALL call_0_15_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_4_9
#define init_0_15_4_9
void FASTCALL call_0_15_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_4_10
#define init_0_15_4_10
void FASTCALL call_0_15_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_4_11
#define init_0_15_4_11
void FASTCALL call_0_15_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_4_12
#define init_0_15_4_12
void FASTCALL call_0_15_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_4_13
#define init_0_15_4_13
void FASTCALL call_0_15_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_4_14
#define init_0_15_4_14
void FASTCALL call_0_15_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_4_15
#define init_0_15_4_15
void FASTCALL call_0_15_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_5_0
#define init_0_15_5_0
void FASTCALL call_0_15_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_5_1
#define init_0_15_5_1
void FASTCALL call_0_15_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_5_2
#define init_0_15_5_2
void FASTCALL call_0_15_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_5_3
#define init_0_15_5_3
void FASTCALL call_0_15_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_5_4
#define init_0_15_5_4
void FASTCALL call_0_15_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_5_5
#define init_0_15_5_5
void FASTCALL call_0_15_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_5_6
#define init_0_15_5_6
void FASTCALL call_0_15_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_5_7
#define init_0_15_5_7
void FASTCALL call_0_15_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_5_8
#define init_0_15_5_8
void FASTCALL call_0_15_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_5_9
#define init_0_15_5_9
void FASTCALL call_0_15_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_5_10
#define init_0_15_5_10
void FASTCALL call_0_15_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_5_11
#define init_0_15_5_11
void FASTCALL call_0_15_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_5_12
#define init_0_15_5_12
void FASTCALL call_0_15_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_5_13
#define init_0_15_5_13
void FASTCALL call_0_15_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_5_14
#define init_0_15_5_14
void FASTCALL call_0_15_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_5_15
#define init_0_15_5_15
void FASTCALL call_0_15_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_6_0
#define init_0_15_6_0
void FASTCALL call_0_15_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_6_1
#define init_0_15_6_1
void FASTCALL call_0_15_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_6_2
#define init_0_15_6_2
void FASTCALL call_0_15_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_6_3
#define init_0_15_6_3
void FASTCALL call_0_15_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_6_4
#define init_0_15_6_4
void FASTCALL call_0_15_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_6_5
#define init_0_15_6_5
void FASTCALL call_0_15_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_6_6
#define init_0_15_6_6
void FASTCALL call_0_15_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_6_7
#define init_0_15_6_7
void FASTCALL call_0_15_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_6_8
#define init_0_15_6_8
void FASTCALL call_0_15_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_6_9
#define init_0_15_6_9
void FASTCALL call_0_15_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_6_10
#define init_0_15_6_10
void FASTCALL call_0_15_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_6_11
#define init_0_15_6_11
void FASTCALL call_0_15_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_6_12
#define init_0_15_6_12
void FASTCALL call_0_15_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_6_13
#define init_0_15_6_13
void FASTCALL call_0_15_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_6_14
#define init_0_15_6_14
void FASTCALL call_0_15_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_6_15
#define init_0_15_6_15
void FASTCALL call_0_15_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_7_0
#define init_0_15_7_0
void FASTCALL call_0_15_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_7_1
#define init_0_15_7_1
void FASTCALL call_0_15_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_7_2
#define init_0_15_7_2
void FASTCALL call_0_15_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_7_3
#define init_0_15_7_3
void FASTCALL call_0_15_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_7_4
#define init_0_15_7_4
void FASTCALL call_0_15_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_7_5
#define init_0_15_7_5
void FASTCALL call_0_15_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_7_6
#define init_0_15_7_6
void FASTCALL call_0_15_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_7_7
#define init_0_15_7_7
void FASTCALL call_0_15_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_7_8
#define init_0_15_7_8
void FASTCALL call_0_15_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_7_9
#define init_0_15_7_9
void FASTCALL call_0_15_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_7_10
#define init_0_15_7_10
void FASTCALL call_0_15_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_7_11
#define init_0_15_7_11
void FASTCALL call_0_15_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_7_12
#define init_0_15_7_12
void FASTCALL call_0_15_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_7_13
#define init_0_15_7_13
void FASTCALL call_0_15_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_7_14
#define init_0_15_7_14
void FASTCALL call_0_15_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_7_15
#define init_0_15_7_15
void FASTCALL call_0_15_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_8_0
#define init_0_15_8_0
void FASTCALL call_0_15_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_8_1
#define init_0_15_8_1
void FASTCALL call_0_15_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_8_2
#define init_0_15_8_2
void FASTCALL call_0_15_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_8_3
#define init_0_15_8_3
void FASTCALL call_0_15_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_8_4
#define init_0_15_8_4
void FASTCALL call_0_15_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_8_5
#define init_0_15_8_5
void FASTCALL call_0_15_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_8_6
#define init_0_15_8_6
void FASTCALL call_0_15_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_8_7
#define init_0_15_8_7
void FASTCALL call_0_15_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_8_8
#define init_0_15_8_8
void FASTCALL call_0_15_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_8_9
#define init_0_15_8_9
void FASTCALL call_0_15_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_8_10
#define init_0_15_8_10
void FASTCALL call_0_15_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_8_11
#define init_0_15_8_11
void FASTCALL call_0_15_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_8_12
#define init_0_15_8_12
void FASTCALL call_0_15_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_8_13
#define init_0_15_8_13
void FASTCALL call_0_15_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_8_14
#define init_0_15_8_14
void FASTCALL call_0_15_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_8_15
#define init_0_15_8_15
void FASTCALL call_0_15_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_9_0
#define init_0_15_9_0
void FASTCALL call_0_15_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_9_1
#define init_0_15_9_1
void FASTCALL call_0_15_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_9_2
#define init_0_15_9_2
void FASTCALL call_0_15_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_9_3
#define init_0_15_9_3
void FASTCALL call_0_15_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_9_4
#define init_0_15_9_4
void FASTCALL call_0_15_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_9_5
#define init_0_15_9_5
void FASTCALL call_0_15_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_9_6
#define init_0_15_9_6
void FASTCALL call_0_15_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_9_7
#define init_0_15_9_7
void FASTCALL call_0_15_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_9_8
#define init_0_15_9_8
void FASTCALL call_0_15_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_9_9
#define init_0_15_9_9
void FASTCALL call_0_15_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_9_10
#define init_0_15_9_10
void FASTCALL call_0_15_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_9_11
#define init_0_15_9_11
void FASTCALL call_0_15_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_9_12
#define init_0_15_9_12
void FASTCALL call_0_15_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_9_13
#define init_0_15_9_13
void FASTCALL call_0_15_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_9_14
#define init_0_15_9_14
void FASTCALL call_0_15_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_9_15
#define init_0_15_9_15
void FASTCALL call_0_15_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_10_0
#define init_0_15_10_0
void FASTCALL call_0_15_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_10_1
#define init_0_15_10_1
void FASTCALL call_0_15_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_10_2
#define init_0_15_10_2
void FASTCALL call_0_15_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_10_3
#define init_0_15_10_3
void FASTCALL call_0_15_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_10_4
#define init_0_15_10_4
void FASTCALL call_0_15_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_10_5
#define init_0_15_10_5
void FASTCALL call_0_15_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_10_6
#define init_0_15_10_6
void FASTCALL call_0_15_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_10_7
#define init_0_15_10_7
void FASTCALL call_0_15_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_10_8
#define init_0_15_10_8
void FASTCALL call_0_15_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_10_9
#define init_0_15_10_9
void FASTCALL call_0_15_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_10_10
#define init_0_15_10_10
void FASTCALL call_0_15_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_10_11
#define init_0_15_10_11
void FASTCALL call_0_15_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_10_12
#define init_0_15_10_12
void FASTCALL call_0_15_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_10_13
#define init_0_15_10_13
void FASTCALL call_0_15_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_10_14
#define init_0_15_10_14
void FASTCALL call_0_15_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_10_15
#define init_0_15_10_15
void FASTCALL call_0_15_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_11_0
#define init_0_15_11_0
void FASTCALL call_0_15_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_11_1
#define init_0_15_11_1
void FASTCALL call_0_15_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_11_2
#define init_0_15_11_2
void FASTCALL call_0_15_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_11_3
#define init_0_15_11_3
void FASTCALL call_0_15_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_11_4
#define init_0_15_11_4
void FASTCALL call_0_15_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_11_5
#define init_0_15_11_5
void FASTCALL call_0_15_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_11_6
#define init_0_15_11_6
void FASTCALL call_0_15_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_11_7
#define init_0_15_11_7
void FASTCALL call_0_15_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_11_8
#define init_0_15_11_8
void FASTCALL call_0_15_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_11_9
#define init_0_15_11_9
void FASTCALL call_0_15_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_11_10
#define init_0_15_11_10
void FASTCALL call_0_15_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_11_11
#define init_0_15_11_11
void FASTCALL call_0_15_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_11_12
#define init_0_15_11_12
void FASTCALL call_0_15_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_11_13
#define init_0_15_11_13
void FASTCALL call_0_15_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_11_14
#define init_0_15_11_14
void FASTCALL call_0_15_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_11_15
#define init_0_15_11_15
void FASTCALL call_0_15_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_12_0
#define init_0_15_12_0
void FASTCALL call_0_15_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_12_1
#define init_0_15_12_1
void FASTCALL call_0_15_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_12_2
#define init_0_15_12_2
void FASTCALL call_0_15_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_12_3
#define init_0_15_12_3
void FASTCALL call_0_15_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_12_4
#define init_0_15_12_4
void FASTCALL call_0_15_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_12_5
#define init_0_15_12_5
void FASTCALL call_0_15_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_12_6
#define init_0_15_12_6
void FASTCALL call_0_15_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_12_7
#define init_0_15_12_7
void FASTCALL call_0_15_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_12_8
#define init_0_15_12_8
void FASTCALL call_0_15_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_12_9
#define init_0_15_12_9
void FASTCALL call_0_15_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_12_10
#define init_0_15_12_10
void FASTCALL call_0_15_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_12_11
#define init_0_15_12_11
void FASTCALL call_0_15_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_12_12
#define init_0_15_12_12
void FASTCALL call_0_15_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_12_13
#define init_0_15_12_13
void FASTCALL call_0_15_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_12_14
#define init_0_15_12_14
void FASTCALL call_0_15_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_12_15
#define init_0_15_12_15
void FASTCALL call_0_15_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_13_0
#define init_0_15_13_0
void FASTCALL call_0_15_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_13_1
#define init_0_15_13_1
void FASTCALL call_0_15_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_13_2
#define init_0_15_13_2
void FASTCALL call_0_15_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_13_3
#define init_0_15_13_3
void FASTCALL call_0_15_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_13_4
#define init_0_15_13_4
void FASTCALL call_0_15_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_13_5
#define init_0_15_13_5
void FASTCALL call_0_15_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_13_6
#define init_0_15_13_6
void FASTCALL call_0_15_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_13_7
#define init_0_15_13_7
void FASTCALL call_0_15_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_13_8
#define init_0_15_13_8
void FASTCALL call_0_15_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_13_9
#define init_0_15_13_9
void FASTCALL call_0_15_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_13_10
#define init_0_15_13_10
void FASTCALL call_0_15_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_13_11
#define init_0_15_13_11
void FASTCALL call_0_15_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_13_12
#define init_0_15_13_12
void FASTCALL call_0_15_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_13_13
#define init_0_15_13_13
void FASTCALL call_0_15_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_13_14
#define init_0_15_13_14
void FASTCALL call_0_15_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_13_15
#define init_0_15_13_15
void FASTCALL call_0_15_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_14_0
#define init_0_15_14_0
void FASTCALL call_0_15_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_14_1
#define init_0_15_14_1
void FASTCALL call_0_15_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_14_2
#define init_0_15_14_2
void FASTCALL call_0_15_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_14_3
#define init_0_15_14_3
void FASTCALL call_0_15_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_14_4
#define init_0_15_14_4
void FASTCALL call_0_15_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_14_5
#define init_0_15_14_5
void FASTCALL call_0_15_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_14_6
#define init_0_15_14_6
void FASTCALL call_0_15_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_14_7
#define init_0_15_14_7
void FASTCALL call_0_15_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_14_8
#define init_0_15_14_8
void FASTCALL call_0_15_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_14_9
#define init_0_15_14_9
void FASTCALL call_0_15_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_14_10
#define init_0_15_14_10
void FASTCALL call_0_15_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_14_11
#define init_0_15_14_11
void FASTCALL call_0_15_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_14_12
#define init_0_15_14_12
void FASTCALL call_0_15_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_14_13
#define init_0_15_14_13
void FASTCALL call_0_15_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_14_14
#define init_0_15_14_14
void FASTCALL call_0_15_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_14_15
#define init_0_15_14_15
void FASTCALL call_0_15_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_15_0
#define init_0_15_15_0
void FASTCALL call_0_15_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_15_1
#define init_0_15_15_1
void FASTCALL call_0_15_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_15_2
#define init_0_15_15_2
void FASTCALL call_0_15_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_15_3
#define init_0_15_15_3
void FASTCALL call_0_15_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_15_4
#define init_0_15_15_4
void FASTCALL call_0_15_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_15_5
#define init_0_15_15_5
void FASTCALL call_0_15_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_15_6
#define init_0_15_15_6
void FASTCALL call_0_15_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_15_7
#define init_0_15_15_7
void FASTCALL call_0_15_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_15_8
#define init_0_15_15_8
void FASTCALL call_0_15_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_15_9
#define init_0_15_15_9
void FASTCALL call_0_15_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_15_10
#define init_0_15_15_10
void FASTCALL call_0_15_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_15_11
#define init_0_15_15_11
void FASTCALL call_0_15_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_15_12
#define init_0_15_15_12
void FASTCALL call_0_15_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_15_13
#define init_0_15_15_13
void FASTCALL call_0_15_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_15_14
#define init_0_15_15_14
void FASTCALL call_0_15_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_0_15_15_15
#define init_0_15_15_15
void FASTCALL call_0_15_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_0_0
#define init_1_0_0_0
void FASTCALL call_1_0_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_0_1
#define init_1_0_0_1
void FASTCALL call_1_0_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_0_2
#define init_1_0_0_2
void FASTCALL call_1_0_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_0_3
#define init_1_0_0_3
void FASTCALL call_1_0_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_0_4
#define init_1_0_0_4
void FASTCALL call_1_0_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_0_5
#define init_1_0_0_5
void FASTCALL call_1_0_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_0_6
#define init_1_0_0_6
void FASTCALL call_1_0_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_0_7
#define init_1_0_0_7
void FASTCALL call_1_0_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_0_8
#define init_1_0_0_8
void FASTCALL call_1_0_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_0_9
#define init_1_0_0_9
void FASTCALL call_1_0_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_0_10
#define init_1_0_0_10
void FASTCALL call_1_0_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_0_11
#define init_1_0_0_11
void FASTCALL call_1_0_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_0_12
#define init_1_0_0_12
void FASTCALL call_1_0_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_0_13
#define init_1_0_0_13
void FASTCALL call_1_0_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_0_14
#define init_1_0_0_14
void FASTCALL call_1_0_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_0_15
#define init_1_0_0_15
void FASTCALL call_1_0_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_1_0
#define init_1_0_1_0
void FASTCALL call_1_0_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_1_1
#define init_1_0_1_1
void FASTCALL call_1_0_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_1_2
#define init_1_0_1_2
void FASTCALL call_1_0_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_1_3
#define init_1_0_1_3
void FASTCALL call_1_0_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_1_4
#define init_1_0_1_4
void FASTCALL call_1_0_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_1_5
#define init_1_0_1_5
void FASTCALL call_1_0_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_1_6
#define init_1_0_1_6
void FASTCALL call_1_0_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_1_7
#define init_1_0_1_7
void FASTCALL call_1_0_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_1_8
#define init_1_0_1_8
void FASTCALL call_1_0_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_1_9
#define init_1_0_1_9
void FASTCALL call_1_0_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_1_10
#define init_1_0_1_10
void FASTCALL call_1_0_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_1_11
#define init_1_0_1_11
void FASTCALL call_1_0_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_1_12
#define init_1_0_1_12
void FASTCALL call_1_0_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_1_13
#define init_1_0_1_13
void FASTCALL call_1_0_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_1_14
#define init_1_0_1_14
void FASTCALL call_1_0_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_1_15
#define init_1_0_1_15
void FASTCALL call_1_0_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_2_0
#define init_1_0_2_0
void FASTCALL call_1_0_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_2_1
#define init_1_0_2_1
void FASTCALL call_1_0_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_2_2
#define init_1_0_2_2
void FASTCALL call_1_0_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_2_3
#define init_1_0_2_3
void FASTCALL call_1_0_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_2_4
#define init_1_0_2_4
void FASTCALL call_1_0_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_2_5
#define init_1_0_2_5
void FASTCALL call_1_0_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_2_6
#define init_1_0_2_6
void FASTCALL call_1_0_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_2_7
#define init_1_0_2_7
void FASTCALL call_1_0_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_2_8
#define init_1_0_2_8
void FASTCALL call_1_0_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_2_9
#define init_1_0_2_9
void FASTCALL call_1_0_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_2_10
#define init_1_0_2_10
void FASTCALL call_1_0_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_2_11
#define init_1_0_2_11
void FASTCALL call_1_0_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_2_12
#define init_1_0_2_12
void FASTCALL call_1_0_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_2_13
#define init_1_0_2_13
void FASTCALL call_1_0_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_2_14
#define init_1_0_2_14
void FASTCALL call_1_0_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_2_15
#define init_1_0_2_15
void FASTCALL call_1_0_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_3_0
#define init_1_0_3_0
void FASTCALL call_1_0_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_3_1
#define init_1_0_3_1
void FASTCALL call_1_0_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_3_2
#define init_1_0_3_2
void FASTCALL call_1_0_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_3_3
#define init_1_0_3_3
void FASTCALL call_1_0_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_3_4
#define init_1_0_3_4
void FASTCALL call_1_0_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_3_5
#define init_1_0_3_5
void FASTCALL call_1_0_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_3_6
#define init_1_0_3_6
void FASTCALL call_1_0_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_3_7
#define init_1_0_3_7
void FASTCALL call_1_0_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_3_8
#define init_1_0_3_8
void FASTCALL call_1_0_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_3_9
#define init_1_0_3_9
void FASTCALL call_1_0_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_3_10
#define init_1_0_3_10
void FASTCALL call_1_0_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_3_11
#define init_1_0_3_11
void FASTCALL call_1_0_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_3_12
#define init_1_0_3_12
void FASTCALL call_1_0_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_3_13
#define init_1_0_3_13
void FASTCALL call_1_0_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_3_14
#define init_1_0_3_14
void FASTCALL call_1_0_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_3_15
#define init_1_0_3_15
void FASTCALL call_1_0_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_4_0
#define init_1_0_4_0
void FASTCALL call_1_0_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_4_1
#define init_1_0_4_1
void FASTCALL call_1_0_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_4_2
#define init_1_0_4_2
void FASTCALL call_1_0_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_4_3
#define init_1_0_4_3
void FASTCALL call_1_0_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_4_4
#define init_1_0_4_4
void FASTCALL call_1_0_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_4_5
#define init_1_0_4_5
void FASTCALL call_1_0_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_4_6
#define init_1_0_4_6
void FASTCALL call_1_0_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_4_7
#define init_1_0_4_7
void FASTCALL call_1_0_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_4_8
#define init_1_0_4_8
void FASTCALL call_1_0_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_4_9
#define init_1_0_4_9
void FASTCALL call_1_0_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_4_10
#define init_1_0_4_10
void FASTCALL call_1_0_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_4_11
#define init_1_0_4_11
void FASTCALL call_1_0_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_4_12
#define init_1_0_4_12
void FASTCALL call_1_0_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_4_13
#define init_1_0_4_13
void FASTCALL call_1_0_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_4_14
#define init_1_0_4_14
void FASTCALL call_1_0_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_4_15
#define init_1_0_4_15
void FASTCALL call_1_0_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_5_0
#define init_1_0_5_0
void FASTCALL call_1_0_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_5_1
#define init_1_0_5_1
void FASTCALL call_1_0_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_5_2
#define init_1_0_5_2
void FASTCALL call_1_0_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_5_3
#define init_1_0_5_3
void FASTCALL call_1_0_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_5_4
#define init_1_0_5_4
void FASTCALL call_1_0_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_5_5
#define init_1_0_5_5
void FASTCALL call_1_0_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_5_6
#define init_1_0_5_6
void FASTCALL call_1_0_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_5_7
#define init_1_0_5_7
void FASTCALL call_1_0_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_5_8
#define init_1_0_5_8
void FASTCALL call_1_0_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_5_9
#define init_1_0_5_9
void FASTCALL call_1_0_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_5_10
#define init_1_0_5_10
void FASTCALL call_1_0_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_5_11
#define init_1_0_5_11
void FASTCALL call_1_0_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_5_12
#define init_1_0_5_12
void FASTCALL call_1_0_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_5_13
#define init_1_0_5_13
void FASTCALL call_1_0_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_5_14
#define init_1_0_5_14
void FASTCALL call_1_0_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_5_15
#define init_1_0_5_15
void FASTCALL call_1_0_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_6_0
#define init_1_0_6_0
void FASTCALL call_1_0_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_6_1
#define init_1_0_6_1
void FASTCALL call_1_0_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_6_2
#define init_1_0_6_2
void FASTCALL call_1_0_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_6_3
#define init_1_0_6_3
void FASTCALL call_1_0_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_6_4
#define init_1_0_6_4
void FASTCALL call_1_0_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_6_5
#define init_1_0_6_5
void FASTCALL call_1_0_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_6_6
#define init_1_0_6_6
void FASTCALL call_1_0_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_6_7
#define init_1_0_6_7
void FASTCALL call_1_0_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_6_8
#define init_1_0_6_8
void FASTCALL call_1_0_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_6_9
#define init_1_0_6_9
void FASTCALL call_1_0_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_6_10
#define init_1_0_6_10
void FASTCALL call_1_0_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_6_11
#define init_1_0_6_11
void FASTCALL call_1_0_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_6_12
#define init_1_0_6_12
void FASTCALL call_1_0_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_6_13
#define init_1_0_6_13
void FASTCALL call_1_0_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_6_14
#define init_1_0_6_14
void FASTCALL call_1_0_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_6_15
#define init_1_0_6_15
void FASTCALL call_1_0_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_7_0
#define init_1_0_7_0
void FASTCALL call_1_0_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_7_1
#define init_1_0_7_1
void FASTCALL call_1_0_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_7_2
#define init_1_0_7_2
void FASTCALL call_1_0_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_7_3
#define init_1_0_7_3
void FASTCALL call_1_0_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_7_4
#define init_1_0_7_4
void FASTCALL call_1_0_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_7_5
#define init_1_0_7_5
void FASTCALL call_1_0_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_7_6
#define init_1_0_7_6
void FASTCALL call_1_0_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_7_7
#define init_1_0_7_7
void FASTCALL call_1_0_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_7_8
#define init_1_0_7_8
void FASTCALL call_1_0_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_7_9
#define init_1_0_7_9
void FASTCALL call_1_0_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_7_10
#define init_1_0_7_10
void FASTCALL call_1_0_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_7_11
#define init_1_0_7_11
void FASTCALL call_1_0_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_7_12
#define init_1_0_7_12
void FASTCALL call_1_0_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_7_13
#define init_1_0_7_13
void FASTCALL call_1_0_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_7_14
#define init_1_0_7_14
void FASTCALL call_1_0_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_7_15
#define init_1_0_7_15
void FASTCALL call_1_0_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_8_0
#define init_1_0_8_0
void FASTCALL call_1_0_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_8_1
#define init_1_0_8_1
void FASTCALL call_1_0_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_8_2
#define init_1_0_8_2
void FASTCALL call_1_0_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_8_3
#define init_1_0_8_3
void FASTCALL call_1_0_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_8_4
#define init_1_0_8_4
void FASTCALL call_1_0_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_8_5
#define init_1_0_8_5
void FASTCALL call_1_0_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_8_6
#define init_1_0_8_6
void FASTCALL call_1_0_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_8_7
#define init_1_0_8_7
void FASTCALL call_1_0_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_8_8
#define init_1_0_8_8
void FASTCALL call_1_0_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_8_9
#define init_1_0_8_9
void FASTCALL call_1_0_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_8_10
#define init_1_0_8_10
void FASTCALL call_1_0_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_8_11
#define init_1_0_8_11
void FASTCALL call_1_0_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_8_12
#define init_1_0_8_12
void FASTCALL call_1_0_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_8_13
#define init_1_0_8_13
void FASTCALL call_1_0_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_8_14
#define init_1_0_8_14
void FASTCALL call_1_0_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_8_15
#define init_1_0_8_15
void FASTCALL call_1_0_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_9_0
#define init_1_0_9_0
void FASTCALL call_1_0_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_9_1
#define init_1_0_9_1
void FASTCALL call_1_0_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_9_2
#define init_1_0_9_2
void FASTCALL call_1_0_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_9_3
#define init_1_0_9_3
void FASTCALL call_1_0_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_9_4
#define init_1_0_9_4
void FASTCALL call_1_0_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_9_5
#define init_1_0_9_5
void FASTCALL call_1_0_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_9_6
#define init_1_0_9_6
void FASTCALL call_1_0_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_9_7
#define init_1_0_9_7
void FASTCALL call_1_0_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_9_8
#define init_1_0_9_8
void FASTCALL call_1_0_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_9_9
#define init_1_0_9_9
void FASTCALL call_1_0_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_9_10
#define init_1_0_9_10
void FASTCALL call_1_0_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_9_11
#define init_1_0_9_11
void FASTCALL call_1_0_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_9_12
#define init_1_0_9_12
void FASTCALL call_1_0_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_9_13
#define init_1_0_9_13
void FASTCALL call_1_0_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_9_14
#define init_1_0_9_14
void FASTCALL call_1_0_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_9_15
#define init_1_0_9_15
void FASTCALL call_1_0_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_10_0
#define init_1_0_10_0
void FASTCALL call_1_0_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_10_1
#define init_1_0_10_1
void FASTCALL call_1_0_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_10_2
#define init_1_0_10_2
void FASTCALL call_1_0_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_10_3
#define init_1_0_10_3
void FASTCALL call_1_0_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_10_4
#define init_1_0_10_4
void FASTCALL call_1_0_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_10_5
#define init_1_0_10_5
void FASTCALL call_1_0_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_10_6
#define init_1_0_10_6
void FASTCALL call_1_0_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_10_7
#define init_1_0_10_7
void FASTCALL call_1_0_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_10_8
#define init_1_0_10_8
void FASTCALL call_1_0_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_10_9
#define init_1_0_10_9
void FASTCALL call_1_0_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_10_10
#define init_1_0_10_10
void FASTCALL call_1_0_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_10_11
#define init_1_0_10_11
void FASTCALL call_1_0_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_10_12
#define init_1_0_10_12
void FASTCALL call_1_0_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_10_13
#define init_1_0_10_13
void FASTCALL call_1_0_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_10_14
#define init_1_0_10_14
void FASTCALL call_1_0_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_10_15
#define init_1_0_10_15
void FASTCALL call_1_0_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_11_0
#define init_1_0_11_0
void FASTCALL call_1_0_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_11_1
#define init_1_0_11_1
void FASTCALL call_1_0_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_11_2
#define init_1_0_11_2
void FASTCALL call_1_0_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_11_3
#define init_1_0_11_3
void FASTCALL call_1_0_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_11_4
#define init_1_0_11_4
void FASTCALL call_1_0_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_11_5
#define init_1_0_11_5
void FASTCALL call_1_0_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_11_6
#define init_1_0_11_6
void FASTCALL call_1_0_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_11_7
#define init_1_0_11_7
void FASTCALL call_1_0_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_11_8
#define init_1_0_11_8
void FASTCALL call_1_0_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_11_9
#define init_1_0_11_9
void FASTCALL call_1_0_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_11_10
#define init_1_0_11_10
void FASTCALL call_1_0_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_11_11
#define init_1_0_11_11
void FASTCALL call_1_0_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_11_12
#define init_1_0_11_12
void FASTCALL call_1_0_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_11_13
#define init_1_0_11_13
void FASTCALL call_1_0_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_11_14
#define init_1_0_11_14
void FASTCALL call_1_0_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_11_15
#define init_1_0_11_15
void FASTCALL call_1_0_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_12_0
#define init_1_0_12_0
void FASTCALL call_1_0_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_12_1
#define init_1_0_12_1
void FASTCALL call_1_0_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_12_2
#define init_1_0_12_2
void FASTCALL call_1_0_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_12_3
#define init_1_0_12_3
void FASTCALL call_1_0_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_12_4
#define init_1_0_12_4
void FASTCALL call_1_0_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_12_5
#define init_1_0_12_5
void FASTCALL call_1_0_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_12_6
#define init_1_0_12_6
void FASTCALL call_1_0_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_12_7
#define init_1_0_12_7
void FASTCALL call_1_0_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_12_8
#define init_1_0_12_8
void FASTCALL call_1_0_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_12_9
#define init_1_0_12_9
void FASTCALL call_1_0_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_12_10
#define init_1_0_12_10
void FASTCALL call_1_0_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_12_11
#define init_1_0_12_11
void FASTCALL call_1_0_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_12_12
#define init_1_0_12_12
void FASTCALL call_1_0_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_12_13
#define init_1_0_12_13
void FASTCALL call_1_0_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_12_14
#define init_1_0_12_14
void FASTCALL call_1_0_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_12_15
#define init_1_0_12_15
void FASTCALL call_1_0_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_13_0
#define init_1_0_13_0
void FASTCALL call_1_0_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_13_1
#define init_1_0_13_1
void FASTCALL call_1_0_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_13_2
#define init_1_0_13_2
void FASTCALL call_1_0_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_13_3
#define init_1_0_13_3
void FASTCALL call_1_0_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_13_4
#define init_1_0_13_4
void FASTCALL call_1_0_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_13_5
#define init_1_0_13_5
void FASTCALL call_1_0_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_13_6
#define init_1_0_13_6
void FASTCALL call_1_0_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_13_7
#define init_1_0_13_7
void FASTCALL call_1_0_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_13_8
#define init_1_0_13_8
void FASTCALL call_1_0_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_13_9
#define init_1_0_13_9
void FASTCALL call_1_0_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_13_10
#define init_1_0_13_10
void FASTCALL call_1_0_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_13_11
#define init_1_0_13_11
void FASTCALL call_1_0_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_13_12
#define init_1_0_13_12
void FASTCALL call_1_0_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_13_13
#define init_1_0_13_13
void FASTCALL call_1_0_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_13_14
#define init_1_0_13_14
void FASTCALL call_1_0_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_13_15
#define init_1_0_13_15
void FASTCALL call_1_0_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_14_0
#define init_1_0_14_0
void FASTCALL call_1_0_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_14_1
#define init_1_0_14_1
void FASTCALL call_1_0_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_14_2
#define init_1_0_14_2
void FASTCALL call_1_0_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_14_3
#define init_1_0_14_3
void FASTCALL call_1_0_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_14_4
#define init_1_0_14_4
void FASTCALL call_1_0_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_14_5
#define init_1_0_14_5
void FASTCALL call_1_0_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_14_6
#define init_1_0_14_6
void FASTCALL call_1_0_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_14_7
#define init_1_0_14_7
void FASTCALL call_1_0_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_14_8
#define init_1_0_14_8
void FASTCALL call_1_0_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_14_9
#define init_1_0_14_9
void FASTCALL call_1_0_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_14_10
#define init_1_0_14_10
void FASTCALL call_1_0_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_14_11
#define init_1_0_14_11
void FASTCALL call_1_0_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_14_12
#define init_1_0_14_12
void FASTCALL call_1_0_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_14_13
#define init_1_0_14_13
void FASTCALL call_1_0_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_14_14
#define init_1_0_14_14
void FASTCALL call_1_0_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_14_15
#define init_1_0_14_15
void FASTCALL call_1_0_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_15_0
#define init_1_0_15_0
void FASTCALL call_1_0_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_15_1
#define init_1_0_15_1
void FASTCALL call_1_0_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_15_2
#define init_1_0_15_2
void FASTCALL call_1_0_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_15_3
#define init_1_0_15_3
void FASTCALL call_1_0_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_15_4
#define init_1_0_15_4
void FASTCALL call_1_0_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_15_5
#define init_1_0_15_5
void FASTCALL call_1_0_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_15_6
#define init_1_0_15_6
void FASTCALL call_1_0_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_15_7
#define init_1_0_15_7
void FASTCALL call_1_0_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_15_8
#define init_1_0_15_8
void FASTCALL call_1_0_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_15_9
#define init_1_0_15_9
void FASTCALL call_1_0_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_15_10
#define init_1_0_15_10
void FASTCALL call_1_0_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_15_11
#define init_1_0_15_11
void FASTCALL call_1_0_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_15_12
#define init_1_0_15_12
void FASTCALL call_1_0_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_15_13
#define init_1_0_15_13
void FASTCALL call_1_0_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_15_14
#define init_1_0_15_14
void FASTCALL call_1_0_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_0_15_15
#define init_1_0_15_15
void FASTCALL call_1_0_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_0_0
#define init_1_1_0_0
void FASTCALL call_1_1_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_0_1
#define init_1_1_0_1
void FASTCALL call_1_1_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_0_2
#define init_1_1_0_2
void FASTCALL call_1_1_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_0_3
#define init_1_1_0_3
void FASTCALL call_1_1_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_0_4
#define init_1_1_0_4
void FASTCALL call_1_1_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_0_5
#define init_1_1_0_5
void FASTCALL call_1_1_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_0_6
#define init_1_1_0_6
void FASTCALL call_1_1_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_0_7
#define init_1_1_0_7
void FASTCALL call_1_1_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_0_8
#define init_1_1_0_8
void FASTCALL call_1_1_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_0_9
#define init_1_1_0_9
void FASTCALL call_1_1_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_0_10
#define init_1_1_0_10
void FASTCALL call_1_1_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_0_11
#define init_1_1_0_11
void FASTCALL call_1_1_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_0_12
#define init_1_1_0_12
void FASTCALL call_1_1_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_0_13
#define init_1_1_0_13
void FASTCALL call_1_1_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_0_14
#define init_1_1_0_14
void FASTCALL call_1_1_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_0_15
#define init_1_1_0_15
void FASTCALL call_1_1_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_1_0
#define init_1_1_1_0
void FASTCALL call_1_1_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_1_1
#define init_1_1_1_1
void FASTCALL call_1_1_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_1_2
#define init_1_1_1_2
void FASTCALL call_1_1_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_1_3
#define init_1_1_1_3
void FASTCALL call_1_1_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_1_4
#define init_1_1_1_4
void FASTCALL call_1_1_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_1_5
#define init_1_1_1_5
void FASTCALL call_1_1_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_1_6
#define init_1_1_1_6
void FASTCALL call_1_1_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_1_7
#define init_1_1_1_7
void FASTCALL call_1_1_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_1_8
#define init_1_1_1_8
void FASTCALL call_1_1_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_1_9
#define init_1_1_1_9
void FASTCALL call_1_1_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_1_10
#define init_1_1_1_10
void FASTCALL call_1_1_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_1_11
#define init_1_1_1_11
void FASTCALL call_1_1_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_1_12
#define init_1_1_1_12
void FASTCALL call_1_1_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_1_13
#define init_1_1_1_13
void FASTCALL call_1_1_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_1_14
#define init_1_1_1_14
void FASTCALL call_1_1_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_1_15
#define init_1_1_1_15
void FASTCALL call_1_1_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_2_0
#define init_1_1_2_0
void FASTCALL call_1_1_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_2_1
#define init_1_1_2_1
void FASTCALL call_1_1_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_2_2
#define init_1_1_2_2
void FASTCALL call_1_1_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_2_3
#define init_1_1_2_3
void FASTCALL call_1_1_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_2_4
#define init_1_1_2_4
void FASTCALL call_1_1_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_2_5
#define init_1_1_2_5
void FASTCALL call_1_1_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_2_6
#define init_1_1_2_6
void FASTCALL call_1_1_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_2_7
#define init_1_1_2_7
void FASTCALL call_1_1_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_2_8
#define init_1_1_2_8
void FASTCALL call_1_1_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_2_9
#define init_1_1_2_9
void FASTCALL call_1_1_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_2_10
#define init_1_1_2_10
void FASTCALL call_1_1_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_2_11
#define init_1_1_2_11
void FASTCALL call_1_1_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_2_12
#define init_1_1_2_12
void FASTCALL call_1_1_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_2_13
#define init_1_1_2_13
void FASTCALL call_1_1_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_2_14
#define init_1_1_2_14
void FASTCALL call_1_1_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_2_15
#define init_1_1_2_15
void FASTCALL call_1_1_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_3_0
#define init_1_1_3_0
void FASTCALL call_1_1_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_3_1
#define init_1_1_3_1
void FASTCALL call_1_1_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_3_2
#define init_1_1_3_2
void FASTCALL call_1_1_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_3_3
#define init_1_1_3_3
void FASTCALL call_1_1_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_3_4
#define init_1_1_3_4
void FASTCALL call_1_1_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_3_5
#define init_1_1_3_5
void FASTCALL call_1_1_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_3_6
#define init_1_1_3_6
void FASTCALL call_1_1_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_3_7
#define init_1_1_3_7
void FASTCALL call_1_1_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_3_8
#define init_1_1_3_8
void FASTCALL call_1_1_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_3_9
#define init_1_1_3_9
void FASTCALL call_1_1_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_3_10
#define init_1_1_3_10
void FASTCALL call_1_1_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_3_11
#define init_1_1_3_11
void FASTCALL call_1_1_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_3_12
#define init_1_1_3_12
void FASTCALL call_1_1_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_3_13
#define init_1_1_3_13
void FASTCALL call_1_1_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_3_14
#define init_1_1_3_14
void FASTCALL call_1_1_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_3_15
#define init_1_1_3_15
void FASTCALL call_1_1_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_4_0
#define init_1_1_4_0
void FASTCALL call_1_1_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_4_1
#define init_1_1_4_1
void FASTCALL call_1_1_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_4_2
#define init_1_1_4_2
void FASTCALL call_1_1_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_4_3
#define init_1_1_4_3
void FASTCALL call_1_1_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_4_4
#define init_1_1_4_4
void FASTCALL call_1_1_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_4_5
#define init_1_1_4_5
void FASTCALL call_1_1_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_4_6
#define init_1_1_4_6
void FASTCALL call_1_1_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_4_7
#define init_1_1_4_7
void FASTCALL call_1_1_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_4_8
#define init_1_1_4_8
void FASTCALL call_1_1_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_4_9
#define init_1_1_4_9
void FASTCALL call_1_1_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_4_10
#define init_1_1_4_10
void FASTCALL call_1_1_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_4_11
#define init_1_1_4_11
void FASTCALL call_1_1_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_4_12
#define init_1_1_4_12
void FASTCALL call_1_1_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_4_13
#define init_1_1_4_13
void FASTCALL call_1_1_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_4_14
#define init_1_1_4_14
void FASTCALL call_1_1_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_4_15
#define init_1_1_4_15
void FASTCALL call_1_1_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_5_0
#define init_1_1_5_0
void FASTCALL call_1_1_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_5_1
#define init_1_1_5_1
void FASTCALL call_1_1_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_5_2
#define init_1_1_5_2
void FASTCALL call_1_1_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_5_3
#define init_1_1_5_3
void FASTCALL call_1_1_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_5_4
#define init_1_1_5_4
void FASTCALL call_1_1_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_5_5
#define init_1_1_5_5
void FASTCALL call_1_1_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_5_6
#define init_1_1_5_6
void FASTCALL call_1_1_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_5_7
#define init_1_1_5_7
void FASTCALL call_1_1_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_5_8
#define init_1_1_5_8
void FASTCALL call_1_1_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_5_9
#define init_1_1_5_9
void FASTCALL call_1_1_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_5_10
#define init_1_1_5_10
void FASTCALL call_1_1_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_5_11
#define init_1_1_5_11
void FASTCALL call_1_1_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_5_12
#define init_1_1_5_12
void FASTCALL call_1_1_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_5_13
#define init_1_1_5_13
void FASTCALL call_1_1_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_5_14
#define init_1_1_5_14
void FASTCALL call_1_1_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_5_15
#define init_1_1_5_15
void FASTCALL call_1_1_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_6_0
#define init_1_1_6_0
void FASTCALL call_1_1_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_6_1
#define init_1_1_6_1
void FASTCALL call_1_1_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_6_2
#define init_1_1_6_2
void FASTCALL call_1_1_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_6_3
#define init_1_1_6_3
void FASTCALL call_1_1_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_6_4
#define init_1_1_6_4
void FASTCALL call_1_1_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_6_5
#define init_1_1_6_5
void FASTCALL call_1_1_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_6_6
#define init_1_1_6_6
void FASTCALL call_1_1_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_6_7
#define init_1_1_6_7
void FASTCALL call_1_1_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_6_8
#define init_1_1_6_8
void FASTCALL call_1_1_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_6_9
#define init_1_1_6_9
void FASTCALL call_1_1_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_6_10
#define init_1_1_6_10
void FASTCALL call_1_1_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_6_11
#define init_1_1_6_11
void FASTCALL call_1_1_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_6_12
#define init_1_1_6_12
void FASTCALL call_1_1_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_6_13
#define init_1_1_6_13
void FASTCALL call_1_1_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_6_14
#define init_1_1_6_14
void FASTCALL call_1_1_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_6_15
#define init_1_1_6_15
void FASTCALL call_1_1_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_7_0
#define init_1_1_7_0
void FASTCALL call_1_1_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_7_1
#define init_1_1_7_1
void FASTCALL call_1_1_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_7_2
#define init_1_1_7_2
void FASTCALL call_1_1_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_7_3
#define init_1_1_7_3
void FASTCALL call_1_1_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_7_4
#define init_1_1_7_4
void FASTCALL call_1_1_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_7_5
#define init_1_1_7_5
void FASTCALL call_1_1_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_7_6
#define init_1_1_7_6
void FASTCALL call_1_1_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_7_7
#define init_1_1_7_7
void FASTCALL call_1_1_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_7_8
#define init_1_1_7_8
void FASTCALL call_1_1_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_7_9
#define init_1_1_7_9
void FASTCALL call_1_1_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_7_10
#define init_1_1_7_10
void FASTCALL call_1_1_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_7_11
#define init_1_1_7_11
void FASTCALL call_1_1_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_7_12
#define init_1_1_7_12
void FASTCALL call_1_1_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_7_13
#define init_1_1_7_13
void FASTCALL call_1_1_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_7_14
#define init_1_1_7_14
void FASTCALL call_1_1_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_7_15
#define init_1_1_7_15
void FASTCALL call_1_1_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_8_0
#define init_1_1_8_0
void FASTCALL call_1_1_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_8_1
#define init_1_1_8_1
void FASTCALL call_1_1_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_8_2
#define init_1_1_8_2
void FASTCALL call_1_1_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_8_3
#define init_1_1_8_3
void FASTCALL call_1_1_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_8_4
#define init_1_1_8_4
void FASTCALL call_1_1_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_8_5
#define init_1_1_8_5
void FASTCALL call_1_1_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_8_6
#define init_1_1_8_6
void FASTCALL call_1_1_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_8_7
#define init_1_1_8_7
void FASTCALL call_1_1_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_8_8
#define init_1_1_8_8
void FASTCALL call_1_1_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_8_9
#define init_1_1_8_9
void FASTCALL call_1_1_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_8_10
#define init_1_1_8_10
void FASTCALL call_1_1_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_8_11
#define init_1_1_8_11
void FASTCALL call_1_1_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_8_12
#define init_1_1_8_12
void FASTCALL call_1_1_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_8_13
#define init_1_1_8_13
void FASTCALL call_1_1_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_8_14
#define init_1_1_8_14
void FASTCALL call_1_1_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_8_15
#define init_1_1_8_15
void FASTCALL call_1_1_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_9_0
#define init_1_1_9_0
void FASTCALL call_1_1_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_9_1
#define init_1_1_9_1
void FASTCALL call_1_1_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_9_2
#define init_1_1_9_2
void FASTCALL call_1_1_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_9_3
#define init_1_1_9_3
void FASTCALL call_1_1_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_9_4
#define init_1_1_9_4
void FASTCALL call_1_1_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_9_5
#define init_1_1_9_5
void FASTCALL call_1_1_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_9_6
#define init_1_1_9_6
void FASTCALL call_1_1_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_9_7
#define init_1_1_9_7
void FASTCALL call_1_1_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_9_8
#define init_1_1_9_8
void FASTCALL call_1_1_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_9_9
#define init_1_1_9_9
void FASTCALL call_1_1_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_9_10
#define init_1_1_9_10
void FASTCALL call_1_1_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_9_11
#define init_1_1_9_11
void FASTCALL call_1_1_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_9_12
#define init_1_1_9_12
void FASTCALL call_1_1_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_9_13
#define init_1_1_9_13
void FASTCALL call_1_1_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_9_14
#define init_1_1_9_14
void FASTCALL call_1_1_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_9_15
#define init_1_1_9_15
void FASTCALL call_1_1_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_10_0
#define init_1_1_10_0
void FASTCALL call_1_1_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_10_1
#define init_1_1_10_1
void FASTCALL call_1_1_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_10_2
#define init_1_1_10_2
void FASTCALL call_1_1_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_10_3
#define init_1_1_10_3
void FASTCALL call_1_1_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_10_4
#define init_1_1_10_4
void FASTCALL call_1_1_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_10_5
#define init_1_1_10_5
void FASTCALL call_1_1_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_10_6
#define init_1_1_10_6
void FASTCALL call_1_1_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_10_7
#define init_1_1_10_7
void FASTCALL call_1_1_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_10_8
#define init_1_1_10_8
void FASTCALL call_1_1_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_10_9
#define init_1_1_10_9
void FASTCALL call_1_1_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_10_10
#define init_1_1_10_10
void FASTCALL call_1_1_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_10_11
#define init_1_1_10_11
void FASTCALL call_1_1_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_10_12
#define init_1_1_10_12
void FASTCALL call_1_1_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_10_13
#define init_1_1_10_13
void FASTCALL call_1_1_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_10_14
#define init_1_1_10_14
void FASTCALL call_1_1_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_10_15
#define init_1_1_10_15
void FASTCALL call_1_1_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_11_0
#define init_1_1_11_0
void FASTCALL call_1_1_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_11_1
#define init_1_1_11_1
void FASTCALL call_1_1_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_11_2
#define init_1_1_11_2
void FASTCALL call_1_1_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_11_3
#define init_1_1_11_3
void FASTCALL call_1_1_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_11_4
#define init_1_1_11_4
void FASTCALL call_1_1_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_11_5
#define init_1_1_11_5
void FASTCALL call_1_1_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_11_6
#define init_1_1_11_6
void FASTCALL call_1_1_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_11_7
#define init_1_1_11_7
void FASTCALL call_1_1_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_11_8
#define init_1_1_11_8
void FASTCALL call_1_1_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_11_9
#define init_1_1_11_9
void FASTCALL call_1_1_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_11_10
#define init_1_1_11_10
void FASTCALL call_1_1_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_11_11
#define init_1_1_11_11
void FASTCALL call_1_1_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_11_12
#define init_1_1_11_12
void FASTCALL call_1_1_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_11_13
#define init_1_1_11_13
void FASTCALL call_1_1_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_11_14
#define init_1_1_11_14
void FASTCALL call_1_1_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_11_15
#define init_1_1_11_15
void FASTCALL call_1_1_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_12_0
#define init_1_1_12_0
void FASTCALL call_1_1_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_12_1
#define init_1_1_12_1
void FASTCALL call_1_1_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_12_2
#define init_1_1_12_2
void FASTCALL call_1_1_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_12_3
#define init_1_1_12_3
void FASTCALL call_1_1_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_12_4
#define init_1_1_12_4
void FASTCALL call_1_1_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_12_5
#define init_1_1_12_5
void FASTCALL call_1_1_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_12_6
#define init_1_1_12_6
void FASTCALL call_1_1_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_12_7
#define init_1_1_12_7
void FASTCALL call_1_1_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_12_8
#define init_1_1_12_8
void FASTCALL call_1_1_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_12_9
#define init_1_1_12_9
void FASTCALL call_1_1_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_12_10
#define init_1_1_12_10
void FASTCALL call_1_1_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_12_11
#define init_1_1_12_11
void FASTCALL call_1_1_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_12_12
#define init_1_1_12_12
void FASTCALL call_1_1_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_12_13
#define init_1_1_12_13
void FASTCALL call_1_1_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_12_14
#define init_1_1_12_14
void FASTCALL call_1_1_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_12_15
#define init_1_1_12_15
void FASTCALL call_1_1_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_13_0
#define init_1_1_13_0
void FASTCALL call_1_1_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_13_1
#define init_1_1_13_1
void FASTCALL call_1_1_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_13_2
#define init_1_1_13_2
void FASTCALL call_1_1_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_13_3
#define init_1_1_13_3
void FASTCALL call_1_1_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_13_4
#define init_1_1_13_4
void FASTCALL call_1_1_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_13_5
#define init_1_1_13_5
void FASTCALL call_1_1_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_13_6
#define init_1_1_13_6
void FASTCALL call_1_1_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_13_7
#define init_1_1_13_7
void FASTCALL call_1_1_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_13_8
#define init_1_1_13_8
void FASTCALL call_1_1_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_13_9
#define init_1_1_13_9
void FASTCALL call_1_1_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_13_10
#define init_1_1_13_10
void FASTCALL call_1_1_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_13_11
#define init_1_1_13_11
void FASTCALL call_1_1_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_13_12
#define init_1_1_13_12
void FASTCALL call_1_1_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_13_13
#define init_1_1_13_13
void FASTCALL call_1_1_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_13_14
#define init_1_1_13_14
void FASTCALL call_1_1_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_13_15
#define init_1_1_13_15
void FASTCALL call_1_1_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_14_0
#define init_1_1_14_0
void FASTCALL call_1_1_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_14_1
#define init_1_1_14_1
void FASTCALL call_1_1_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_14_2
#define init_1_1_14_2
void FASTCALL call_1_1_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_14_3
#define init_1_1_14_3
void FASTCALL call_1_1_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_14_4
#define init_1_1_14_4
void FASTCALL call_1_1_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_14_5
#define init_1_1_14_5
void FASTCALL call_1_1_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_14_6
#define init_1_1_14_6
void FASTCALL call_1_1_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_14_7
#define init_1_1_14_7
void FASTCALL call_1_1_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_14_8
#define init_1_1_14_8
void FASTCALL call_1_1_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_14_9
#define init_1_1_14_9
void FASTCALL call_1_1_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_14_10
#define init_1_1_14_10
void FASTCALL call_1_1_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_14_11
#define init_1_1_14_11
void FASTCALL call_1_1_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_14_12
#define init_1_1_14_12
void FASTCALL call_1_1_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_14_13
#define init_1_1_14_13
void FASTCALL call_1_1_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_14_14
#define init_1_1_14_14
void FASTCALL call_1_1_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_14_15
#define init_1_1_14_15
void FASTCALL call_1_1_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_15_0
#define init_1_1_15_0
void FASTCALL call_1_1_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_15_1
#define init_1_1_15_1
void FASTCALL call_1_1_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_15_2
#define init_1_1_15_2
void FASTCALL call_1_1_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_15_3
#define init_1_1_15_3
void FASTCALL call_1_1_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_15_4
#define init_1_1_15_4
void FASTCALL call_1_1_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_15_5
#define init_1_1_15_5
void FASTCALL call_1_1_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_15_6
#define init_1_1_15_6
void FASTCALL call_1_1_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_15_7
#define init_1_1_15_7
void FASTCALL call_1_1_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_15_8
#define init_1_1_15_8
void FASTCALL call_1_1_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_15_9
#define init_1_1_15_9
void FASTCALL call_1_1_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_15_10
#define init_1_1_15_10
void FASTCALL call_1_1_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_15_11
#define init_1_1_15_11
void FASTCALL call_1_1_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_15_12
#define init_1_1_15_12
void FASTCALL call_1_1_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_15_13
#define init_1_1_15_13
void FASTCALL call_1_1_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_15_14
#define init_1_1_15_14
void FASTCALL call_1_1_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_1_15_15
#define init_1_1_15_15
void FASTCALL call_1_1_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_0_0
#define init_1_2_0_0
void FASTCALL call_1_2_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_0_1
#define init_1_2_0_1
void FASTCALL call_1_2_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_0_2
#define init_1_2_0_2
void FASTCALL call_1_2_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_0_3
#define init_1_2_0_3
void FASTCALL call_1_2_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_0_4
#define init_1_2_0_4
void FASTCALL call_1_2_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_0_5
#define init_1_2_0_5
void FASTCALL call_1_2_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_0_6
#define init_1_2_0_6
void FASTCALL call_1_2_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_0_7
#define init_1_2_0_7
void FASTCALL call_1_2_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_0_8
#define init_1_2_0_8
void FASTCALL call_1_2_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_0_9
#define init_1_2_0_9
void FASTCALL call_1_2_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_0_10
#define init_1_2_0_10
void FASTCALL call_1_2_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_0_11
#define init_1_2_0_11
void FASTCALL call_1_2_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_0_12
#define init_1_2_0_12
void FASTCALL call_1_2_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_0_13
#define init_1_2_0_13
void FASTCALL call_1_2_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_0_14
#define init_1_2_0_14
void FASTCALL call_1_2_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_0_15
#define init_1_2_0_15
void FASTCALL call_1_2_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_1_0
#define init_1_2_1_0
void FASTCALL call_1_2_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_1_1
#define init_1_2_1_1
void FASTCALL call_1_2_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_1_2
#define init_1_2_1_2
void FASTCALL call_1_2_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_1_3
#define init_1_2_1_3
void FASTCALL call_1_2_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_1_4
#define init_1_2_1_4
void FASTCALL call_1_2_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_1_5
#define init_1_2_1_5
void FASTCALL call_1_2_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_1_6
#define init_1_2_1_6
void FASTCALL call_1_2_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_1_7
#define init_1_2_1_7
void FASTCALL call_1_2_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_1_8
#define init_1_2_1_8
void FASTCALL call_1_2_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_1_9
#define init_1_2_1_9
void FASTCALL call_1_2_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_1_10
#define init_1_2_1_10
void FASTCALL call_1_2_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_1_11
#define init_1_2_1_11
void FASTCALL call_1_2_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_1_12
#define init_1_2_1_12
void FASTCALL call_1_2_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_1_13
#define init_1_2_1_13
void FASTCALL call_1_2_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_1_14
#define init_1_2_1_14
void FASTCALL call_1_2_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_1_15
#define init_1_2_1_15
void FASTCALL call_1_2_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_2_0
#define init_1_2_2_0
void FASTCALL call_1_2_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_2_1
#define init_1_2_2_1
void FASTCALL call_1_2_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_2_2
#define init_1_2_2_2
void FASTCALL call_1_2_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_2_3
#define init_1_2_2_3
void FASTCALL call_1_2_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_2_4
#define init_1_2_2_4
void FASTCALL call_1_2_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_2_5
#define init_1_2_2_5
void FASTCALL call_1_2_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_2_6
#define init_1_2_2_6
void FASTCALL call_1_2_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_2_7
#define init_1_2_2_7
void FASTCALL call_1_2_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_2_8
#define init_1_2_2_8
void FASTCALL call_1_2_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_2_9
#define init_1_2_2_9
void FASTCALL call_1_2_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_2_10
#define init_1_2_2_10
void FASTCALL call_1_2_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_2_11
#define init_1_2_2_11
void FASTCALL call_1_2_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_2_12
#define init_1_2_2_12
void FASTCALL call_1_2_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_2_13
#define init_1_2_2_13
void FASTCALL call_1_2_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_2_14
#define init_1_2_2_14
void FASTCALL call_1_2_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_2_15
#define init_1_2_2_15
void FASTCALL call_1_2_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_3_0
#define init_1_2_3_0
void FASTCALL call_1_2_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_3_1
#define init_1_2_3_1
void FASTCALL call_1_2_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_3_2
#define init_1_2_3_2
void FASTCALL call_1_2_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_3_3
#define init_1_2_3_3
void FASTCALL call_1_2_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_3_4
#define init_1_2_3_4
void FASTCALL call_1_2_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_3_5
#define init_1_2_3_5
void FASTCALL call_1_2_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_3_6
#define init_1_2_3_6
void FASTCALL call_1_2_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_3_7
#define init_1_2_3_7
void FASTCALL call_1_2_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_3_8
#define init_1_2_3_8
void FASTCALL call_1_2_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_3_9
#define init_1_2_3_9
void FASTCALL call_1_2_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_3_10
#define init_1_2_3_10
void FASTCALL call_1_2_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_3_11
#define init_1_2_3_11
void FASTCALL call_1_2_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_3_12
#define init_1_2_3_12
void FASTCALL call_1_2_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_3_13
#define init_1_2_3_13
void FASTCALL call_1_2_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_3_14
#define init_1_2_3_14
void FASTCALL call_1_2_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_3_15
#define init_1_2_3_15
void FASTCALL call_1_2_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_4_0
#define init_1_2_4_0
void FASTCALL call_1_2_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_4_1
#define init_1_2_4_1
void FASTCALL call_1_2_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_4_2
#define init_1_2_4_2
void FASTCALL call_1_2_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_4_3
#define init_1_2_4_3
void FASTCALL call_1_2_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_4_4
#define init_1_2_4_4
void FASTCALL call_1_2_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_4_5
#define init_1_2_4_5
void FASTCALL call_1_2_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_4_6
#define init_1_2_4_6
void FASTCALL call_1_2_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_4_7
#define init_1_2_4_7
void FASTCALL call_1_2_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_4_8
#define init_1_2_4_8
void FASTCALL call_1_2_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_4_9
#define init_1_2_4_9
void FASTCALL call_1_2_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_4_10
#define init_1_2_4_10
void FASTCALL call_1_2_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_4_11
#define init_1_2_4_11
void FASTCALL call_1_2_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_4_12
#define init_1_2_4_12
void FASTCALL call_1_2_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_4_13
#define init_1_2_4_13
void FASTCALL call_1_2_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_4_14
#define init_1_2_4_14
void FASTCALL call_1_2_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_4_15
#define init_1_2_4_15
void FASTCALL call_1_2_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_5_0
#define init_1_2_5_0
void FASTCALL call_1_2_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_5_1
#define init_1_2_5_1
void FASTCALL call_1_2_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_5_2
#define init_1_2_5_2
void FASTCALL call_1_2_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_5_3
#define init_1_2_5_3
void FASTCALL call_1_2_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_5_4
#define init_1_2_5_4
void FASTCALL call_1_2_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_5_5
#define init_1_2_5_5
void FASTCALL call_1_2_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_5_6
#define init_1_2_5_6
void FASTCALL call_1_2_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_5_7
#define init_1_2_5_7
void FASTCALL call_1_2_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_5_8
#define init_1_2_5_8
void FASTCALL call_1_2_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_5_9
#define init_1_2_5_9
void FASTCALL call_1_2_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_5_10
#define init_1_2_5_10
void FASTCALL call_1_2_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_5_11
#define init_1_2_5_11
void FASTCALL call_1_2_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_5_12
#define init_1_2_5_12
void FASTCALL call_1_2_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_5_13
#define init_1_2_5_13
void FASTCALL call_1_2_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_5_14
#define init_1_2_5_14
void FASTCALL call_1_2_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_5_15
#define init_1_2_5_15
void FASTCALL call_1_2_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_6_0
#define init_1_2_6_0
void FASTCALL call_1_2_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_6_1
#define init_1_2_6_1
void FASTCALL call_1_2_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_6_2
#define init_1_2_6_2
void FASTCALL call_1_2_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_6_3
#define init_1_2_6_3
void FASTCALL call_1_2_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_6_4
#define init_1_2_6_4
void FASTCALL call_1_2_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_6_5
#define init_1_2_6_5
void FASTCALL call_1_2_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_6_6
#define init_1_2_6_6
void FASTCALL call_1_2_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_6_7
#define init_1_2_6_7
void FASTCALL call_1_2_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_6_8
#define init_1_2_6_8
void FASTCALL call_1_2_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_6_9
#define init_1_2_6_9
void FASTCALL call_1_2_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_6_10
#define init_1_2_6_10
void FASTCALL call_1_2_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_6_11
#define init_1_2_6_11
void FASTCALL call_1_2_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_6_12
#define init_1_2_6_12
void FASTCALL call_1_2_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_6_13
#define init_1_2_6_13
void FASTCALL call_1_2_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_6_14
#define init_1_2_6_14
void FASTCALL call_1_2_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_6_15
#define init_1_2_6_15
void FASTCALL call_1_2_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_7_0
#define init_1_2_7_0
void FASTCALL call_1_2_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_7_1
#define init_1_2_7_1
void FASTCALL call_1_2_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_7_2
#define init_1_2_7_2
void FASTCALL call_1_2_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_7_3
#define init_1_2_7_3
void FASTCALL call_1_2_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_7_4
#define init_1_2_7_4
void FASTCALL call_1_2_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_7_5
#define init_1_2_7_5
void FASTCALL call_1_2_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_7_6
#define init_1_2_7_6
void FASTCALL call_1_2_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_7_7
#define init_1_2_7_7
void FASTCALL call_1_2_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_7_8
#define init_1_2_7_8
void FASTCALL call_1_2_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_7_9
#define init_1_2_7_9
void FASTCALL call_1_2_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_7_10
#define init_1_2_7_10
void FASTCALL call_1_2_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_7_11
#define init_1_2_7_11
void FASTCALL call_1_2_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_7_12
#define init_1_2_7_12
void FASTCALL call_1_2_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_7_13
#define init_1_2_7_13
void FASTCALL call_1_2_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_7_14
#define init_1_2_7_14
void FASTCALL call_1_2_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_7_15
#define init_1_2_7_15
void FASTCALL call_1_2_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_8_0
#define init_1_2_8_0
void FASTCALL call_1_2_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_8_1
#define init_1_2_8_1
void FASTCALL call_1_2_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_8_2
#define init_1_2_8_2
void FASTCALL call_1_2_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_8_3
#define init_1_2_8_3
void FASTCALL call_1_2_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_8_4
#define init_1_2_8_4
void FASTCALL call_1_2_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_8_5
#define init_1_2_8_5
void FASTCALL call_1_2_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_8_6
#define init_1_2_8_6
void FASTCALL call_1_2_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_8_7
#define init_1_2_8_7
void FASTCALL call_1_2_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_8_8
#define init_1_2_8_8
void FASTCALL call_1_2_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_8_9
#define init_1_2_8_9
void FASTCALL call_1_2_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_8_10
#define init_1_2_8_10
void FASTCALL call_1_2_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_8_11
#define init_1_2_8_11
void FASTCALL call_1_2_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_8_12
#define init_1_2_8_12
void FASTCALL call_1_2_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_8_13
#define init_1_2_8_13
void FASTCALL call_1_2_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_8_14
#define init_1_2_8_14
void FASTCALL call_1_2_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_8_15
#define init_1_2_8_15
void FASTCALL call_1_2_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_9_0
#define init_1_2_9_0
void FASTCALL call_1_2_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_9_1
#define init_1_2_9_1
void FASTCALL call_1_2_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_9_2
#define init_1_2_9_2
void FASTCALL call_1_2_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_9_3
#define init_1_2_9_3
void FASTCALL call_1_2_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_9_4
#define init_1_2_9_4
void FASTCALL call_1_2_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_9_5
#define init_1_2_9_5
void FASTCALL call_1_2_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_9_6
#define init_1_2_9_6
void FASTCALL call_1_2_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_9_7
#define init_1_2_9_7
void FASTCALL call_1_2_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_9_8
#define init_1_2_9_8
void FASTCALL call_1_2_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_9_9
#define init_1_2_9_9
void FASTCALL call_1_2_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_9_10
#define init_1_2_9_10
void FASTCALL call_1_2_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_9_11
#define init_1_2_9_11
void FASTCALL call_1_2_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_9_12
#define init_1_2_9_12
void FASTCALL call_1_2_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_9_13
#define init_1_2_9_13
void FASTCALL call_1_2_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_9_14
#define init_1_2_9_14
void FASTCALL call_1_2_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_9_15
#define init_1_2_9_15
void FASTCALL call_1_2_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_10_0
#define init_1_2_10_0
void FASTCALL call_1_2_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_10_1
#define init_1_2_10_1
void FASTCALL call_1_2_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_10_2
#define init_1_2_10_2
void FASTCALL call_1_2_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_10_3
#define init_1_2_10_3
void FASTCALL call_1_2_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_10_4
#define init_1_2_10_4
void FASTCALL call_1_2_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_10_5
#define init_1_2_10_5
void FASTCALL call_1_2_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_10_6
#define init_1_2_10_6
void FASTCALL call_1_2_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_10_7
#define init_1_2_10_7
void FASTCALL call_1_2_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_10_8
#define init_1_2_10_8
void FASTCALL call_1_2_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_10_9
#define init_1_2_10_9
void FASTCALL call_1_2_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_10_10
#define init_1_2_10_10
void FASTCALL call_1_2_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_10_11
#define init_1_2_10_11
void FASTCALL call_1_2_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_10_12
#define init_1_2_10_12
void FASTCALL call_1_2_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_10_13
#define init_1_2_10_13
void FASTCALL call_1_2_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_10_14
#define init_1_2_10_14
void FASTCALL call_1_2_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_10_15
#define init_1_2_10_15
void FASTCALL call_1_2_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_11_0
#define init_1_2_11_0
void FASTCALL call_1_2_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_11_1
#define init_1_2_11_1
void FASTCALL call_1_2_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_11_2
#define init_1_2_11_2
void FASTCALL call_1_2_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_11_3
#define init_1_2_11_3
void FASTCALL call_1_2_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_11_4
#define init_1_2_11_4
void FASTCALL call_1_2_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_11_5
#define init_1_2_11_5
void FASTCALL call_1_2_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_11_6
#define init_1_2_11_6
void FASTCALL call_1_2_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_11_7
#define init_1_2_11_7
void FASTCALL call_1_2_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_11_8
#define init_1_2_11_8
void FASTCALL call_1_2_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_11_9
#define init_1_2_11_9
void FASTCALL call_1_2_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_11_10
#define init_1_2_11_10
void FASTCALL call_1_2_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_11_11
#define init_1_2_11_11
void FASTCALL call_1_2_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_11_12
#define init_1_2_11_12
void FASTCALL call_1_2_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_11_13
#define init_1_2_11_13
void FASTCALL call_1_2_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_11_14
#define init_1_2_11_14
void FASTCALL call_1_2_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_11_15
#define init_1_2_11_15
void FASTCALL call_1_2_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_12_0
#define init_1_2_12_0
void FASTCALL call_1_2_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_12_1
#define init_1_2_12_1
void FASTCALL call_1_2_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_12_2
#define init_1_2_12_2
void FASTCALL call_1_2_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_12_3
#define init_1_2_12_3
void FASTCALL call_1_2_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_12_4
#define init_1_2_12_4
void FASTCALL call_1_2_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_12_5
#define init_1_2_12_5
void FASTCALL call_1_2_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_12_6
#define init_1_2_12_6
void FASTCALL call_1_2_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_12_7
#define init_1_2_12_7
void FASTCALL call_1_2_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_12_8
#define init_1_2_12_8
void FASTCALL call_1_2_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_12_9
#define init_1_2_12_9
void FASTCALL call_1_2_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_12_10
#define init_1_2_12_10
void FASTCALL call_1_2_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_12_11
#define init_1_2_12_11
void FASTCALL call_1_2_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_12_12
#define init_1_2_12_12
void FASTCALL call_1_2_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_12_13
#define init_1_2_12_13
void FASTCALL call_1_2_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_12_14
#define init_1_2_12_14
void FASTCALL call_1_2_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_12_15
#define init_1_2_12_15
void FASTCALL call_1_2_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_13_0
#define init_1_2_13_0
void FASTCALL call_1_2_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_13_1
#define init_1_2_13_1
void FASTCALL call_1_2_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_13_2
#define init_1_2_13_2
void FASTCALL call_1_2_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_13_3
#define init_1_2_13_3
void FASTCALL call_1_2_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_13_4
#define init_1_2_13_4
void FASTCALL call_1_2_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_13_5
#define init_1_2_13_5
void FASTCALL call_1_2_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_13_6
#define init_1_2_13_6
void FASTCALL call_1_2_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_13_7
#define init_1_2_13_7
void FASTCALL call_1_2_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_13_8
#define init_1_2_13_8
void FASTCALL call_1_2_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_13_9
#define init_1_2_13_9
void FASTCALL call_1_2_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_13_10
#define init_1_2_13_10
void FASTCALL call_1_2_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_13_11
#define init_1_2_13_11
void FASTCALL call_1_2_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_13_12
#define init_1_2_13_12
void FASTCALL call_1_2_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_13_13
#define init_1_2_13_13
void FASTCALL call_1_2_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_13_14
#define init_1_2_13_14
void FASTCALL call_1_2_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_13_15
#define init_1_2_13_15
void FASTCALL call_1_2_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_14_0
#define init_1_2_14_0
void FASTCALL call_1_2_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_14_1
#define init_1_2_14_1
void FASTCALL call_1_2_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_14_2
#define init_1_2_14_2
void FASTCALL call_1_2_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_14_3
#define init_1_2_14_3
void FASTCALL call_1_2_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_14_4
#define init_1_2_14_4
void FASTCALL call_1_2_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_14_5
#define init_1_2_14_5
void FASTCALL call_1_2_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_14_6
#define init_1_2_14_6
void FASTCALL call_1_2_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_14_7
#define init_1_2_14_7
void FASTCALL call_1_2_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_14_8
#define init_1_2_14_8
void FASTCALL call_1_2_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_14_9
#define init_1_2_14_9
void FASTCALL call_1_2_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_14_10
#define init_1_2_14_10
void FASTCALL call_1_2_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_14_11
#define init_1_2_14_11
void FASTCALL call_1_2_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_14_12
#define init_1_2_14_12
void FASTCALL call_1_2_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_14_13
#define init_1_2_14_13
void FASTCALL call_1_2_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_14_14
#define init_1_2_14_14
void FASTCALL call_1_2_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_14_15
#define init_1_2_14_15
void FASTCALL call_1_2_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_15_0
#define init_1_2_15_0
void FASTCALL call_1_2_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_15_1
#define init_1_2_15_1
void FASTCALL call_1_2_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_15_2
#define init_1_2_15_2
void FASTCALL call_1_2_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_15_3
#define init_1_2_15_3
void FASTCALL call_1_2_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_15_4
#define init_1_2_15_4
void FASTCALL call_1_2_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_15_5
#define init_1_2_15_5
void FASTCALL call_1_2_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_15_6
#define init_1_2_15_6
void FASTCALL call_1_2_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_15_7
#define init_1_2_15_7
void FASTCALL call_1_2_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_15_8
#define init_1_2_15_8
void FASTCALL call_1_2_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_15_9
#define init_1_2_15_9
void FASTCALL call_1_2_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_15_10
#define init_1_2_15_10
void FASTCALL call_1_2_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_15_11
#define init_1_2_15_11
void FASTCALL call_1_2_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_15_12
#define init_1_2_15_12
void FASTCALL call_1_2_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_15_13
#define init_1_2_15_13
void FASTCALL call_1_2_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_15_14
#define init_1_2_15_14
void FASTCALL call_1_2_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_2_15_15
#define init_1_2_15_15
void FASTCALL call_1_2_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_0_0
#define init_1_3_0_0
void FASTCALL call_1_3_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_0_1
#define init_1_3_0_1
void FASTCALL call_1_3_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_0_2
#define init_1_3_0_2
void FASTCALL call_1_3_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_0_3
#define init_1_3_0_3
void FASTCALL call_1_3_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_0_4
#define init_1_3_0_4
void FASTCALL call_1_3_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_0_5
#define init_1_3_0_5
void FASTCALL call_1_3_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_0_6
#define init_1_3_0_6
void FASTCALL call_1_3_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_0_7
#define init_1_3_0_7
void FASTCALL call_1_3_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_0_8
#define init_1_3_0_8
void FASTCALL call_1_3_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_0_9
#define init_1_3_0_9
void FASTCALL call_1_3_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_0_10
#define init_1_3_0_10
void FASTCALL call_1_3_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_0_11
#define init_1_3_0_11
void FASTCALL call_1_3_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_0_12
#define init_1_3_0_12
void FASTCALL call_1_3_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_0_13
#define init_1_3_0_13
void FASTCALL call_1_3_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_0_14
#define init_1_3_0_14
void FASTCALL call_1_3_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_0_15
#define init_1_3_0_15
void FASTCALL call_1_3_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_1_0
#define init_1_3_1_0
void FASTCALL call_1_3_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_1_1
#define init_1_3_1_1
void FASTCALL call_1_3_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_1_2
#define init_1_3_1_2
void FASTCALL call_1_3_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_1_3
#define init_1_3_1_3
void FASTCALL call_1_3_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_1_4
#define init_1_3_1_4
void FASTCALL call_1_3_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_1_5
#define init_1_3_1_5
void FASTCALL call_1_3_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_1_6
#define init_1_3_1_6
void FASTCALL call_1_3_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_1_7
#define init_1_3_1_7
void FASTCALL call_1_3_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_1_8
#define init_1_3_1_8
void FASTCALL call_1_3_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_1_9
#define init_1_3_1_9
void FASTCALL call_1_3_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_1_10
#define init_1_3_1_10
void FASTCALL call_1_3_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_1_11
#define init_1_3_1_11
void FASTCALL call_1_3_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_1_12
#define init_1_3_1_12
void FASTCALL call_1_3_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_1_13
#define init_1_3_1_13
void FASTCALL call_1_3_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_1_14
#define init_1_3_1_14
void FASTCALL call_1_3_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_1_15
#define init_1_3_1_15
void FASTCALL call_1_3_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_2_0
#define init_1_3_2_0
void FASTCALL call_1_3_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_2_1
#define init_1_3_2_1
void FASTCALL call_1_3_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_2_2
#define init_1_3_2_2
void FASTCALL call_1_3_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_2_3
#define init_1_3_2_3
void FASTCALL call_1_3_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_2_4
#define init_1_3_2_4
void FASTCALL call_1_3_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_2_5
#define init_1_3_2_5
void FASTCALL call_1_3_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_2_6
#define init_1_3_2_6
void FASTCALL call_1_3_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_2_7
#define init_1_3_2_7
void FASTCALL call_1_3_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_2_8
#define init_1_3_2_8
void FASTCALL call_1_3_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_2_9
#define init_1_3_2_9
void FASTCALL call_1_3_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_2_10
#define init_1_3_2_10
void FASTCALL call_1_3_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_2_11
#define init_1_3_2_11
void FASTCALL call_1_3_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_2_12
#define init_1_3_2_12
void FASTCALL call_1_3_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_2_13
#define init_1_3_2_13
void FASTCALL call_1_3_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_2_14
#define init_1_3_2_14
void FASTCALL call_1_3_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_2_15
#define init_1_3_2_15
void FASTCALL call_1_3_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_3_0
#define init_1_3_3_0
void FASTCALL call_1_3_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_3_1
#define init_1_3_3_1
void FASTCALL call_1_3_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_3_2
#define init_1_3_3_2
void FASTCALL call_1_3_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_3_3
#define init_1_3_3_3
void FASTCALL call_1_3_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_3_4
#define init_1_3_3_4
void FASTCALL call_1_3_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_3_5
#define init_1_3_3_5
void FASTCALL call_1_3_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_3_6
#define init_1_3_3_6
void FASTCALL call_1_3_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_3_7
#define init_1_3_3_7
void FASTCALL call_1_3_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_3_8
#define init_1_3_3_8
void FASTCALL call_1_3_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_3_9
#define init_1_3_3_9
void FASTCALL call_1_3_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_3_10
#define init_1_3_3_10
void FASTCALL call_1_3_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_3_11
#define init_1_3_3_11
void FASTCALL call_1_3_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_3_12
#define init_1_3_3_12
void FASTCALL call_1_3_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_3_13
#define init_1_3_3_13
void FASTCALL call_1_3_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_3_14
#define init_1_3_3_14
void FASTCALL call_1_3_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_3_15
#define init_1_3_3_15
void FASTCALL call_1_3_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_4_0
#define init_1_3_4_0
void FASTCALL call_1_3_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_4_1
#define init_1_3_4_1
void FASTCALL call_1_3_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_4_2
#define init_1_3_4_2
void FASTCALL call_1_3_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_4_3
#define init_1_3_4_3
void FASTCALL call_1_3_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_4_4
#define init_1_3_4_4
void FASTCALL call_1_3_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_4_5
#define init_1_3_4_5
void FASTCALL call_1_3_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_4_6
#define init_1_3_4_6
void FASTCALL call_1_3_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_4_7
#define init_1_3_4_7
void FASTCALL call_1_3_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_4_8
#define init_1_3_4_8
void FASTCALL call_1_3_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_4_9
#define init_1_3_4_9
void FASTCALL call_1_3_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_4_10
#define init_1_3_4_10
void FASTCALL call_1_3_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_4_11
#define init_1_3_4_11
void FASTCALL call_1_3_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_4_12
#define init_1_3_4_12
void FASTCALL call_1_3_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_4_13
#define init_1_3_4_13
void FASTCALL call_1_3_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_4_14
#define init_1_3_4_14
void FASTCALL call_1_3_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_4_15
#define init_1_3_4_15
void FASTCALL call_1_3_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_5_0
#define init_1_3_5_0
void FASTCALL call_1_3_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_5_1
#define init_1_3_5_1
void FASTCALL call_1_3_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_5_2
#define init_1_3_5_2
void FASTCALL call_1_3_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_5_3
#define init_1_3_5_3
void FASTCALL call_1_3_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_5_4
#define init_1_3_5_4
void FASTCALL call_1_3_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_5_5
#define init_1_3_5_5
void FASTCALL call_1_3_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_5_6
#define init_1_3_5_6
void FASTCALL call_1_3_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_5_7
#define init_1_3_5_7
void FASTCALL call_1_3_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_5_8
#define init_1_3_5_8
void FASTCALL call_1_3_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_5_9
#define init_1_3_5_9
void FASTCALL call_1_3_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_5_10
#define init_1_3_5_10
void FASTCALL call_1_3_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_5_11
#define init_1_3_5_11
void FASTCALL call_1_3_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_5_12
#define init_1_3_5_12
void FASTCALL call_1_3_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_5_13
#define init_1_3_5_13
void FASTCALL call_1_3_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_5_14
#define init_1_3_5_14
void FASTCALL call_1_3_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_5_15
#define init_1_3_5_15
void FASTCALL call_1_3_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_6_0
#define init_1_3_6_0
void FASTCALL call_1_3_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_6_1
#define init_1_3_6_1
void FASTCALL call_1_3_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_6_2
#define init_1_3_6_2
void FASTCALL call_1_3_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_6_3
#define init_1_3_6_3
void FASTCALL call_1_3_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_6_4
#define init_1_3_6_4
void FASTCALL call_1_3_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_6_5
#define init_1_3_6_5
void FASTCALL call_1_3_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_6_6
#define init_1_3_6_6
void FASTCALL call_1_3_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_6_7
#define init_1_3_6_7
void FASTCALL call_1_3_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_6_8
#define init_1_3_6_8
void FASTCALL call_1_3_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_6_9
#define init_1_3_6_9
void FASTCALL call_1_3_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_6_10
#define init_1_3_6_10
void FASTCALL call_1_3_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_6_11
#define init_1_3_6_11
void FASTCALL call_1_3_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_6_12
#define init_1_3_6_12
void FASTCALL call_1_3_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_6_13
#define init_1_3_6_13
void FASTCALL call_1_3_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_6_14
#define init_1_3_6_14
void FASTCALL call_1_3_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_6_15
#define init_1_3_6_15
void FASTCALL call_1_3_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_7_0
#define init_1_3_7_0
void FASTCALL call_1_3_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_7_1
#define init_1_3_7_1
void FASTCALL call_1_3_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_7_2
#define init_1_3_7_2
void FASTCALL call_1_3_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_7_3
#define init_1_3_7_3
void FASTCALL call_1_3_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_7_4
#define init_1_3_7_4
void FASTCALL call_1_3_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_7_5
#define init_1_3_7_5
void FASTCALL call_1_3_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_7_6
#define init_1_3_7_6
void FASTCALL call_1_3_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_7_7
#define init_1_3_7_7
void FASTCALL call_1_3_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_7_8
#define init_1_3_7_8
void FASTCALL call_1_3_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_7_9
#define init_1_3_7_9
void FASTCALL call_1_3_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_7_10
#define init_1_3_7_10
void FASTCALL call_1_3_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_7_11
#define init_1_3_7_11
void FASTCALL call_1_3_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_7_12
#define init_1_3_7_12
void FASTCALL call_1_3_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_7_13
#define init_1_3_7_13
void FASTCALL call_1_3_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_7_14
#define init_1_3_7_14
void FASTCALL call_1_3_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_7_15
#define init_1_3_7_15
void FASTCALL call_1_3_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_8_0
#define init_1_3_8_0
void FASTCALL call_1_3_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_8_1
#define init_1_3_8_1
void FASTCALL call_1_3_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_8_2
#define init_1_3_8_2
void FASTCALL call_1_3_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_8_3
#define init_1_3_8_3
void FASTCALL call_1_3_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_8_4
#define init_1_3_8_4
void FASTCALL call_1_3_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_8_5
#define init_1_3_8_5
void FASTCALL call_1_3_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_8_6
#define init_1_3_8_6
void FASTCALL call_1_3_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_8_7
#define init_1_3_8_7
void FASTCALL call_1_3_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_8_8
#define init_1_3_8_8
void FASTCALL call_1_3_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_8_9
#define init_1_3_8_9
void FASTCALL call_1_3_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_8_10
#define init_1_3_8_10
void FASTCALL call_1_3_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_8_11
#define init_1_3_8_11
void FASTCALL call_1_3_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_8_12
#define init_1_3_8_12
void FASTCALL call_1_3_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_8_13
#define init_1_3_8_13
void FASTCALL call_1_3_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_8_14
#define init_1_3_8_14
void FASTCALL call_1_3_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_8_15
#define init_1_3_8_15
void FASTCALL call_1_3_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_9_0
#define init_1_3_9_0
void FASTCALL call_1_3_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_9_1
#define init_1_3_9_1
void FASTCALL call_1_3_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_9_2
#define init_1_3_9_2
void FASTCALL call_1_3_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_9_3
#define init_1_3_9_3
void FASTCALL call_1_3_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_9_4
#define init_1_3_9_4
void FASTCALL call_1_3_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_9_5
#define init_1_3_9_5
void FASTCALL call_1_3_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_9_6
#define init_1_3_9_6
void FASTCALL call_1_3_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_9_7
#define init_1_3_9_7
void FASTCALL call_1_3_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_9_8
#define init_1_3_9_8
void FASTCALL call_1_3_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_9_9
#define init_1_3_9_9
void FASTCALL call_1_3_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_9_10
#define init_1_3_9_10
void FASTCALL call_1_3_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_9_11
#define init_1_3_9_11
void FASTCALL call_1_3_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_9_12
#define init_1_3_9_12
void FASTCALL call_1_3_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_9_13
#define init_1_3_9_13
void FASTCALL call_1_3_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_9_14
#define init_1_3_9_14
void FASTCALL call_1_3_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_9_15
#define init_1_3_9_15
void FASTCALL call_1_3_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_10_0
#define init_1_3_10_0
void FASTCALL call_1_3_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_10_1
#define init_1_3_10_1
void FASTCALL call_1_3_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_10_2
#define init_1_3_10_2
void FASTCALL call_1_3_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_10_3
#define init_1_3_10_3
void FASTCALL call_1_3_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_10_4
#define init_1_3_10_4
void FASTCALL call_1_3_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_10_5
#define init_1_3_10_5
void FASTCALL call_1_3_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_10_6
#define init_1_3_10_6
void FASTCALL call_1_3_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_10_7
#define init_1_3_10_7
void FASTCALL call_1_3_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_10_8
#define init_1_3_10_8
void FASTCALL call_1_3_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_10_9
#define init_1_3_10_9
void FASTCALL call_1_3_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_10_10
#define init_1_3_10_10
void FASTCALL call_1_3_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_10_11
#define init_1_3_10_11
void FASTCALL call_1_3_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_10_12
#define init_1_3_10_12
void FASTCALL call_1_3_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_10_13
#define init_1_3_10_13
void FASTCALL call_1_3_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_10_14
#define init_1_3_10_14
void FASTCALL call_1_3_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_10_15
#define init_1_3_10_15
void FASTCALL call_1_3_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_11_0
#define init_1_3_11_0
void FASTCALL call_1_3_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_11_1
#define init_1_3_11_1
void FASTCALL call_1_3_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_11_2
#define init_1_3_11_2
void FASTCALL call_1_3_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_11_3
#define init_1_3_11_3
void FASTCALL call_1_3_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_11_4
#define init_1_3_11_4
void FASTCALL call_1_3_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_11_5
#define init_1_3_11_5
void FASTCALL call_1_3_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_11_6
#define init_1_3_11_6
void FASTCALL call_1_3_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_11_7
#define init_1_3_11_7
void FASTCALL call_1_3_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_11_8
#define init_1_3_11_8
void FASTCALL call_1_3_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_11_9
#define init_1_3_11_9
void FASTCALL call_1_3_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_11_10
#define init_1_3_11_10
void FASTCALL call_1_3_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_11_11
#define init_1_3_11_11
void FASTCALL call_1_3_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_11_12
#define init_1_3_11_12
void FASTCALL call_1_3_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_11_13
#define init_1_3_11_13
void FASTCALL call_1_3_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_11_14
#define init_1_3_11_14
void FASTCALL call_1_3_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_11_15
#define init_1_3_11_15
void FASTCALL call_1_3_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_12_0
#define init_1_3_12_0
void FASTCALL call_1_3_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_12_1
#define init_1_3_12_1
void FASTCALL call_1_3_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_12_2
#define init_1_3_12_2
void FASTCALL call_1_3_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_12_3
#define init_1_3_12_3
void FASTCALL call_1_3_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_12_4
#define init_1_3_12_4
void FASTCALL call_1_3_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_12_5
#define init_1_3_12_5
void FASTCALL call_1_3_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_12_6
#define init_1_3_12_6
void FASTCALL call_1_3_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_12_7
#define init_1_3_12_7
void FASTCALL call_1_3_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_12_8
#define init_1_3_12_8
void FASTCALL call_1_3_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_12_9
#define init_1_3_12_9
void FASTCALL call_1_3_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_12_10
#define init_1_3_12_10
void FASTCALL call_1_3_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_12_11
#define init_1_3_12_11
void FASTCALL call_1_3_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_12_12
#define init_1_3_12_12
void FASTCALL call_1_3_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_12_13
#define init_1_3_12_13
void FASTCALL call_1_3_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_12_14
#define init_1_3_12_14
void FASTCALL call_1_3_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_12_15
#define init_1_3_12_15
void FASTCALL call_1_3_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_13_0
#define init_1_3_13_0
void FASTCALL call_1_3_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_13_1
#define init_1_3_13_1
void FASTCALL call_1_3_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_13_2
#define init_1_3_13_2
void FASTCALL call_1_3_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_13_3
#define init_1_3_13_3
void FASTCALL call_1_3_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_13_4
#define init_1_3_13_4
void FASTCALL call_1_3_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_13_5
#define init_1_3_13_5
void FASTCALL call_1_3_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_13_6
#define init_1_3_13_6
void FASTCALL call_1_3_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_13_7
#define init_1_3_13_7
void FASTCALL call_1_3_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_13_8
#define init_1_3_13_8
void FASTCALL call_1_3_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_13_9
#define init_1_3_13_9
void FASTCALL call_1_3_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_13_10
#define init_1_3_13_10
void FASTCALL call_1_3_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_13_11
#define init_1_3_13_11
void FASTCALL call_1_3_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_13_12
#define init_1_3_13_12
void FASTCALL call_1_3_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_13_13
#define init_1_3_13_13
void FASTCALL call_1_3_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_13_14
#define init_1_3_13_14
void FASTCALL call_1_3_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_13_15
#define init_1_3_13_15
void FASTCALL call_1_3_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_14_0
#define init_1_3_14_0
void FASTCALL call_1_3_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_14_1
#define init_1_3_14_1
void FASTCALL call_1_3_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_14_2
#define init_1_3_14_2
void FASTCALL call_1_3_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_14_3
#define init_1_3_14_3
void FASTCALL call_1_3_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_14_4
#define init_1_3_14_4
void FASTCALL call_1_3_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_14_5
#define init_1_3_14_5
void FASTCALL call_1_3_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_14_6
#define init_1_3_14_6
void FASTCALL call_1_3_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_14_7
#define init_1_3_14_7
void FASTCALL call_1_3_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_14_8
#define init_1_3_14_8
void FASTCALL call_1_3_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_14_9
#define init_1_3_14_9
void FASTCALL call_1_3_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_14_10
#define init_1_3_14_10
void FASTCALL call_1_3_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_14_11
#define init_1_3_14_11
void FASTCALL call_1_3_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_14_12
#define init_1_3_14_12
void FASTCALL call_1_3_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_14_13
#define init_1_3_14_13
void FASTCALL call_1_3_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_14_14
#define init_1_3_14_14
void FASTCALL call_1_3_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_14_15
#define init_1_3_14_15
void FASTCALL call_1_3_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_15_0
#define init_1_3_15_0
void FASTCALL call_1_3_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_15_1
#define init_1_3_15_1
void FASTCALL call_1_3_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_15_2
#define init_1_3_15_2
void FASTCALL call_1_3_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_15_3
#define init_1_3_15_3
void FASTCALL call_1_3_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_15_4
#define init_1_3_15_4
void FASTCALL call_1_3_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_15_5
#define init_1_3_15_5
void FASTCALL call_1_3_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_15_6
#define init_1_3_15_6
void FASTCALL call_1_3_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_15_7
#define init_1_3_15_7
void FASTCALL call_1_3_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_15_8
#define init_1_3_15_8
void FASTCALL call_1_3_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_15_9
#define init_1_3_15_9
void FASTCALL call_1_3_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_15_10
#define init_1_3_15_10
void FASTCALL call_1_3_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_15_11
#define init_1_3_15_11
void FASTCALL call_1_3_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_15_12
#define init_1_3_15_12
void FASTCALL call_1_3_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_15_13
#define init_1_3_15_13
void FASTCALL call_1_3_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_15_14
#define init_1_3_15_14
void FASTCALL call_1_3_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_3_15_15
#define init_1_3_15_15
void FASTCALL call_1_3_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_0_0
#define init_1_4_0_0
void FASTCALL call_1_4_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_0_1
#define init_1_4_0_1
void FASTCALL call_1_4_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_0_2
#define init_1_4_0_2
void FASTCALL call_1_4_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_0_3
#define init_1_4_0_3
void FASTCALL call_1_4_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_0_4
#define init_1_4_0_4
void FASTCALL call_1_4_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_0_5
#define init_1_4_0_5
void FASTCALL call_1_4_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_0_6
#define init_1_4_0_6
void FASTCALL call_1_4_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_0_7
#define init_1_4_0_7
void FASTCALL call_1_4_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_0_8
#define init_1_4_0_8
void FASTCALL call_1_4_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_0_9
#define init_1_4_0_9
void FASTCALL call_1_4_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_0_10
#define init_1_4_0_10
void FASTCALL call_1_4_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_0_11
#define init_1_4_0_11
void FASTCALL call_1_4_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_0_12
#define init_1_4_0_12
void FASTCALL call_1_4_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_0_13
#define init_1_4_0_13
void FASTCALL call_1_4_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_0_14
#define init_1_4_0_14
void FASTCALL call_1_4_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_0_15
#define init_1_4_0_15
void FASTCALL call_1_4_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_1_0
#define init_1_4_1_0
void FASTCALL call_1_4_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_1_1
#define init_1_4_1_1
void FASTCALL call_1_4_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_1_2
#define init_1_4_1_2
void FASTCALL call_1_4_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_1_3
#define init_1_4_1_3
void FASTCALL call_1_4_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_1_4
#define init_1_4_1_4
void FASTCALL call_1_4_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_1_5
#define init_1_4_1_5
void FASTCALL call_1_4_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_1_6
#define init_1_4_1_6
void FASTCALL call_1_4_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_1_7
#define init_1_4_1_7
void FASTCALL call_1_4_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_1_8
#define init_1_4_1_8
void FASTCALL call_1_4_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_1_9
#define init_1_4_1_9
void FASTCALL call_1_4_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_1_10
#define init_1_4_1_10
void FASTCALL call_1_4_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_1_11
#define init_1_4_1_11
void FASTCALL call_1_4_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_1_12
#define init_1_4_1_12
void FASTCALL call_1_4_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_1_13
#define init_1_4_1_13
void FASTCALL call_1_4_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_1_14
#define init_1_4_1_14
void FASTCALL call_1_4_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_1_15
#define init_1_4_1_15
void FASTCALL call_1_4_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_2_0
#define init_1_4_2_0
void FASTCALL call_1_4_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_2_1
#define init_1_4_2_1
void FASTCALL call_1_4_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_2_2
#define init_1_4_2_2
void FASTCALL call_1_4_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_2_3
#define init_1_4_2_3
void FASTCALL call_1_4_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_2_4
#define init_1_4_2_4
void FASTCALL call_1_4_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_2_5
#define init_1_4_2_5
void FASTCALL call_1_4_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_2_6
#define init_1_4_2_6
void FASTCALL call_1_4_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_2_7
#define init_1_4_2_7
void FASTCALL call_1_4_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_2_8
#define init_1_4_2_8
void FASTCALL call_1_4_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_2_9
#define init_1_4_2_9
void FASTCALL call_1_4_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_2_10
#define init_1_4_2_10
void FASTCALL call_1_4_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_2_11
#define init_1_4_2_11
void FASTCALL call_1_4_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_2_12
#define init_1_4_2_12
void FASTCALL call_1_4_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_2_13
#define init_1_4_2_13
void FASTCALL call_1_4_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_2_14
#define init_1_4_2_14
void FASTCALL call_1_4_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_2_15
#define init_1_4_2_15
void FASTCALL call_1_4_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_3_0
#define init_1_4_3_0
void FASTCALL call_1_4_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_3_1
#define init_1_4_3_1
void FASTCALL call_1_4_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_3_2
#define init_1_4_3_2
void FASTCALL call_1_4_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_3_3
#define init_1_4_3_3
void FASTCALL call_1_4_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_3_4
#define init_1_4_3_4
void FASTCALL call_1_4_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_3_5
#define init_1_4_3_5
void FASTCALL call_1_4_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_3_6
#define init_1_4_3_6
void FASTCALL call_1_4_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_3_7
#define init_1_4_3_7
void FASTCALL call_1_4_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_3_8
#define init_1_4_3_8
void FASTCALL call_1_4_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_3_9
#define init_1_4_3_9
void FASTCALL call_1_4_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_3_10
#define init_1_4_3_10
void FASTCALL call_1_4_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_3_11
#define init_1_4_3_11
void FASTCALL call_1_4_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_3_12
#define init_1_4_3_12
void FASTCALL call_1_4_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_3_13
#define init_1_4_3_13
void FASTCALL call_1_4_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_3_14
#define init_1_4_3_14
void FASTCALL call_1_4_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_3_15
#define init_1_4_3_15
void FASTCALL call_1_4_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_4_0
#define init_1_4_4_0
void FASTCALL call_1_4_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_4_1
#define init_1_4_4_1
void FASTCALL call_1_4_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_4_2
#define init_1_4_4_2
void FASTCALL call_1_4_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_4_3
#define init_1_4_4_3
void FASTCALL call_1_4_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_4_4
#define init_1_4_4_4
void FASTCALL call_1_4_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_4_5
#define init_1_4_4_5
void FASTCALL call_1_4_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_4_6
#define init_1_4_4_6
void FASTCALL call_1_4_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_4_7
#define init_1_4_4_7
void FASTCALL call_1_4_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_4_8
#define init_1_4_4_8
void FASTCALL call_1_4_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_4_9
#define init_1_4_4_9
void FASTCALL call_1_4_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_4_10
#define init_1_4_4_10
void FASTCALL call_1_4_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_4_11
#define init_1_4_4_11
void FASTCALL call_1_4_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_4_12
#define init_1_4_4_12
void FASTCALL call_1_4_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_4_13
#define init_1_4_4_13
void FASTCALL call_1_4_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_4_14
#define init_1_4_4_14
void FASTCALL call_1_4_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_4_15
#define init_1_4_4_15
void FASTCALL call_1_4_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_5_0
#define init_1_4_5_0
void FASTCALL call_1_4_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_5_1
#define init_1_4_5_1
void FASTCALL call_1_4_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_5_2
#define init_1_4_5_2
void FASTCALL call_1_4_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_5_3
#define init_1_4_5_3
void FASTCALL call_1_4_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_5_4
#define init_1_4_5_4
void FASTCALL call_1_4_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_5_5
#define init_1_4_5_5
void FASTCALL call_1_4_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_5_6
#define init_1_4_5_6
void FASTCALL call_1_4_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_5_7
#define init_1_4_5_7
void FASTCALL call_1_4_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_5_8
#define init_1_4_5_8
void FASTCALL call_1_4_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_5_9
#define init_1_4_5_9
void FASTCALL call_1_4_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_5_10
#define init_1_4_5_10
void FASTCALL call_1_4_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_5_11
#define init_1_4_5_11
void FASTCALL call_1_4_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_5_12
#define init_1_4_5_12
void FASTCALL call_1_4_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_5_13
#define init_1_4_5_13
void FASTCALL call_1_4_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_5_14
#define init_1_4_5_14
void FASTCALL call_1_4_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_5_15
#define init_1_4_5_15
void FASTCALL call_1_4_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_6_0
#define init_1_4_6_0
void FASTCALL call_1_4_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_6_1
#define init_1_4_6_1
void FASTCALL call_1_4_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_6_2
#define init_1_4_6_2
void FASTCALL call_1_4_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_6_3
#define init_1_4_6_3
void FASTCALL call_1_4_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_6_4
#define init_1_4_6_4
void FASTCALL call_1_4_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_6_5
#define init_1_4_6_5
void FASTCALL call_1_4_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_6_6
#define init_1_4_6_6
void FASTCALL call_1_4_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_6_7
#define init_1_4_6_7
void FASTCALL call_1_4_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_6_8
#define init_1_4_6_8
void FASTCALL call_1_4_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_6_9
#define init_1_4_6_9
void FASTCALL call_1_4_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_6_10
#define init_1_4_6_10
void FASTCALL call_1_4_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_6_11
#define init_1_4_6_11
void FASTCALL call_1_4_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_6_12
#define init_1_4_6_12
void FASTCALL call_1_4_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_6_13
#define init_1_4_6_13
void FASTCALL call_1_4_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_6_14
#define init_1_4_6_14
void FASTCALL call_1_4_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_6_15
#define init_1_4_6_15
void FASTCALL call_1_4_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_7_0
#define init_1_4_7_0
void FASTCALL call_1_4_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_7_1
#define init_1_4_7_1
void FASTCALL call_1_4_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_7_2
#define init_1_4_7_2
void FASTCALL call_1_4_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_7_3
#define init_1_4_7_3
void FASTCALL call_1_4_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_7_4
#define init_1_4_7_4
void FASTCALL call_1_4_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_7_5
#define init_1_4_7_5
void FASTCALL call_1_4_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_7_6
#define init_1_4_7_6
void FASTCALL call_1_4_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_7_7
#define init_1_4_7_7
void FASTCALL call_1_4_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_7_8
#define init_1_4_7_8
void FASTCALL call_1_4_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_7_9
#define init_1_4_7_9
void FASTCALL call_1_4_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_7_10
#define init_1_4_7_10
void FASTCALL call_1_4_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_7_11
#define init_1_4_7_11
void FASTCALL call_1_4_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_7_12
#define init_1_4_7_12
void FASTCALL call_1_4_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_7_13
#define init_1_4_7_13
void FASTCALL call_1_4_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_7_14
#define init_1_4_7_14
void FASTCALL call_1_4_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_7_15
#define init_1_4_7_15
void FASTCALL call_1_4_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_8_0
#define init_1_4_8_0
void FASTCALL call_1_4_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_8_1
#define init_1_4_8_1
void FASTCALL call_1_4_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_8_2
#define init_1_4_8_2
void FASTCALL call_1_4_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_8_3
#define init_1_4_8_3
void FASTCALL call_1_4_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_8_4
#define init_1_4_8_4
void FASTCALL call_1_4_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_8_5
#define init_1_4_8_5
void FASTCALL call_1_4_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_8_6
#define init_1_4_8_6
void FASTCALL call_1_4_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_8_7
#define init_1_4_8_7
void FASTCALL call_1_4_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_8_8
#define init_1_4_8_8
void FASTCALL call_1_4_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_8_9
#define init_1_4_8_9
void FASTCALL call_1_4_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_8_10
#define init_1_4_8_10
void FASTCALL call_1_4_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_8_11
#define init_1_4_8_11
void FASTCALL call_1_4_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_8_12
#define init_1_4_8_12
void FASTCALL call_1_4_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_8_13
#define init_1_4_8_13
void FASTCALL call_1_4_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_8_14
#define init_1_4_8_14
void FASTCALL call_1_4_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_8_15
#define init_1_4_8_15
void FASTCALL call_1_4_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_9_0
#define init_1_4_9_0
void FASTCALL call_1_4_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_9_1
#define init_1_4_9_1
void FASTCALL call_1_4_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_9_2
#define init_1_4_9_2
void FASTCALL call_1_4_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_9_3
#define init_1_4_9_3
void FASTCALL call_1_4_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_9_4
#define init_1_4_9_4
void FASTCALL call_1_4_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_9_5
#define init_1_4_9_5
void FASTCALL call_1_4_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_9_6
#define init_1_4_9_6
void FASTCALL call_1_4_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_9_7
#define init_1_4_9_7
void FASTCALL call_1_4_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_9_8
#define init_1_4_9_8
void FASTCALL call_1_4_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_9_9
#define init_1_4_9_9
void FASTCALL call_1_4_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_9_10
#define init_1_4_9_10
void FASTCALL call_1_4_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_9_11
#define init_1_4_9_11
void FASTCALL call_1_4_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_9_12
#define init_1_4_9_12
void FASTCALL call_1_4_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_9_13
#define init_1_4_9_13
void FASTCALL call_1_4_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_9_14
#define init_1_4_9_14
void FASTCALL call_1_4_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_9_15
#define init_1_4_9_15
void FASTCALL call_1_4_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_10_0
#define init_1_4_10_0
void FASTCALL call_1_4_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_10_1
#define init_1_4_10_1
void FASTCALL call_1_4_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_10_2
#define init_1_4_10_2
void FASTCALL call_1_4_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_10_3
#define init_1_4_10_3
void FASTCALL call_1_4_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_10_4
#define init_1_4_10_4
void FASTCALL call_1_4_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_10_5
#define init_1_4_10_5
void FASTCALL call_1_4_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_10_6
#define init_1_4_10_6
void FASTCALL call_1_4_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_10_7
#define init_1_4_10_7
void FASTCALL call_1_4_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_10_8
#define init_1_4_10_8
void FASTCALL call_1_4_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_10_9
#define init_1_4_10_9
void FASTCALL call_1_4_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_10_10
#define init_1_4_10_10
void FASTCALL call_1_4_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_10_11
#define init_1_4_10_11
void FASTCALL call_1_4_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_10_12
#define init_1_4_10_12
void FASTCALL call_1_4_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_10_13
#define init_1_4_10_13
void FASTCALL call_1_4_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_10_14
#define init_1_4_10_14
void FASTCALL call_1_4_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_10_15
#define init_1_4_10_15
void FASTCALL call_1_4_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_11_0
#define init_1_4_11_0
void FASTCALL call_1_4_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_11_1
#define init_1_4_11_1
void FASTCALL call_1_4_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_11_2
#define init_1_4_11_2
void FASTCALL call_1_4_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_11_3
#define init_1_4_11_3
void FASTCALL call_1_4_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_11_4
#define init_1_4_11_4
void FASTCALL call_1_4_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_11_5
#define init_1_4_11_5
void FASTCALL call_1_4_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_11_6
#define init_1_4_11_6
void FASTCALL call_1_4_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_11_7
#define init_1_4_11_7
void FASTCALL call_1_4_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_11_8
#define init_1_4_11_8
void FASTCALL call_1_4_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_11_9
#define init_1_4_11_9
void FASTCALL call_1_4_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_11_10
#define init_1_4_11_10
void FASTCALL call_1_4_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_11_11
#define init_1_4_11_11
void FASTCALL call_1_4_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_11_12
#define init_1_4_11_12
void FASTCALL call_1_4_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_11_13
#define init_1_4_11_13
void FASTCALL call_1_4_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_11_14
#define init_1_4_11_14
void FASTCALL call_1_4_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_11_15
#define init_1_4_11_15
void FASTCALL call_1_4_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_12_0
#define init_1_4_12_0
void FASTCALL call_1_4_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_12_1
#define init_1_4_12_1
void FASTCALL call_1_4_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_12_2
#define init_1_4_12_2
void FASTCALL call_1_4_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_12_3
#define init_1_4_12_3
void FASTCALL call_1_4_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_12_4
#define init_1_4_12_4
void FASTCALL call_1_4_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_12_5
#define init_1_4_12_5
void FASTCALL call_1_4_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_12_6
#define init_1_4_12_6
void FASTCALL call_1_4_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_12_7
#define init_1_4_12_7
void FASTCALL call_1_4_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_12_8
#define init_1_4_12_8
void FASTCALL call_1_4_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_12_9
#define init_1_4_12_9
void FASTCALL call_1_4_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_12_10
#define init_1_4_12_10
void FASTCALL call_1_4_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_12_11
#define init_1_4_12_11
void FASTCALL call_1_4_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_12_12
#define init_1_4_12_12
void FASTCALL call_1_4_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_12_13
#define init_1_4_12_13
void FASTCALL call_1_4_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_12_14
#define init_1_4_12_14
void FASTCALL call_1_4_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_12_15
#define init_1_4_12_15
void FASTCALL call_1_4_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_13_0
#define init_1_4_13_0
void FASTCALL call_1_4_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_13_1
#define init_1_4_13_1
void FASTCALL call_1_4_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_13_2
#define init_1_4_13_2
void FASTCALL call_1_4_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_13_3
#define init_1_4_13_3
void FASTCALL call_1_4_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_13_4
#define init_1_4_13_4
void FASTCALL call_1_4_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_13_5
#define init_1_4_13_5
void FASTCALL call_1_4_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_13_6
#define init_1_4_13_6
void FASTCALL call_1_4_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_13_7
#define init_1_4_13_7
void FASTCALL call_1_4_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_13_8
#define init_1_4_13_8
void FASTCALL call_1_4_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_13_9
#define init_1_4_13_9
void FASTCALL call_1_4_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_13_10
#define init_1_4_13_10
void FASTCALL call_1_4_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_13_11
#define init_1_4_13_11
void FASTCALL call_1_4_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_13_12
#define init_1_4_13_12
void FASTCALL call_1_4_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_13_13
#define init_1_4_13_13
void FASTCALL call_1_4_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_13_14
#define init_1_4_13_14
void FASTCALL call_1_4_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_13_15
#define init_1_4_13_15
void FASTCALL call_1_4_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_14_0
#define init_1_4_14_0
void FASTCALL call_1_4_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_14_1
#define init_1_4_14_1
void FASTCALL call_1_4_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_14_2
#define init_1_4_14_2
void FASTCALL call_1_4_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_14_3
#define init_1_4_14_3
void FASTCALL call_1_4_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_14_4
#define init_1_4_14_4
void FASTCALL call_1_4_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_14_5
#define init_1_4_14_5
void FASTCALL call_1_4_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_14_6
#define init_1_4_14_6
void FASTCALL call_1_4_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_14_7
#define init_1_4_14_7
void FASTCALL call_1_4_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_14_8
#define init_1_4_14_8
void FASTCALL call_1_4_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_14_9
#define init_1_4_14_9
void FASTCALL call_1_4_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_14_10
#define init_1_4_14_10
void FASTCALL call_1_4_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_14_11
#define init_1_4_14_11
void FASTCALL call_1_4_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_14_12
#define init_1_4_14_12
void FASTCALL call_1_4_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_14_13
#define init_1_4_14_13
void FASTCALL call_1_4_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_14_14
#define init_1_4_14_14
void FASTCALL call_1_4_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_14_15
#define init_1_4_14_15
void FASTCALL call_1_4_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_15_0
#define init_1_4_15_0
void FASTCALL call_1_4_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_15_1
#define init_1_4_15_1
void FASTCALL call_1_4_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_15_2
#define init_1_4_15_2
void FASTCALL call_1_4_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_15_3
#define init_1_4_15_3
void FASTCALL call_1_4_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_15_4
#define init_1_4_15_4
void FASTCALL call_1_4_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_15_5
#define init_1_4_15_5
void FASTCALL call_1_4_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_15_6
#define init_1_4_15_6
void FASTCALL call_1_4_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_15_7
#define init_1_4_15_7
void FASTCALL call_1_4_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_15_8
#define init_1_4_15_8
void FASTCALL call_1_4_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_15_9
#define init_1_4_15_9
void FASTCALL call_1_4_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_15_10
#define init_1_4_15_10
void FASTCALL call_1_4_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_15_11
#define init_1_4_15_11
void FASTCALL call_1_4_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_15_12
#define init_1_4_15_12
void FASTCALL call_1_4_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_15_13
#define init_1_4_15_13
void FASTCALL call_1_4_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_15_14
#define init_1_4_15_14
void FASTCALL call_1_4_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_4_15_15
#define init_1_4_15_15
void FASTCALL call_1_4_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_0_0
#define init_1_5_0_0
void FASTCALL call_1_5_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_0_1
#define init_1_5_0_1
void FASTCALL call_1_5_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_0_2
#define init_1_5_0_2
void FASTCALL call_1_5_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_0_3
#define init_1_5_0_3
void FASTCALL call_1_5_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_0_4
#define init_1_5_0_4
void FASTCALL call_1_5_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_0_5
#define init_1_5_0_5
void FASTCALL call_1_5_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_0_6
#define init_1_5_0_6
void FASTCALL call_1_5_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_0_7
#define init_1_5_0_7
void FASTCALL call_1_5_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_0_8
#define init_1_5_0_8
void FASTCALL call_1_5_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_0_9
#define init_1_5_0_9
void FASTCALL call_1_5_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_0_10
#define init_1_5_0_10
void FASTCALL call_1_5_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_0_11
#define init_1_5_0_11
void FASTCALL call_1_5_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_0_12
#define init_1_5_0_12
void FASTCALL call_1_5_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_0_13
#define init_1_5_0_13
void FASTCALL call_1_5_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_0_14
#define init_1_5_0_14
void FASTCALL call_1_5_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_0_15
#define init_1_5_0_15
void FASTCALL call_1_5_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_1_0
#define init_1_5_1_0
void FASTCALL call_1_5_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_1_1
#define init_1_5_1_1
void FASTCALL call_1_5_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_1_2
#define init_1_5_1_2
void FASTCALL call_1_5_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_1_3
#define init_1_5_1_3
void FASTCALL call_1_5_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_1_4
#define init_1_5_1_4
void FASTCALL call_1_5_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_1_5
#define init_1_5_1_5
void FASTCALL call_1_5_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_1_6
#define init_1_5_1_6
void FASTCALL call_1_5_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_1_7
#define init_1_5_1_7
void FASTCALL call_1_5_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_1_8
#define init_1_5_1_8
void FASTCALL call_1_5_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_1_9
#define init_1_5_1_9
void FASTCALL call_1_5_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_1_10
#define init_1_5_1_10
void FASTCALL call_1_5_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_1_11
#define init_1_5_1_11
void FASTCALL call_1_5_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_1_12
#define init_1_5_1_12
void FASTCALL call_1_5_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_1_13
#define init_1_5_1_13
void FASTCALL call_1_5_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_1_14
#define init_1_5_1_14
void FASTCALL call_1_5_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_1_15
#define init_1_5_1_15
void FASTCALL call_1_5_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_2_0
#define init_1_5_2_0
void FASTCALL call_1_5_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_2_1
#define init_1_5_2_1
void FASTCALL call_1_5_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_2_2
#define init_1_5_2_2
void FASTCALL call_1_5_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_2_3
#define init_1_5_2_3
void FASTCALL call_1_5_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_2_4
#define init_1_5_2_4
void FASTCALL call_1_5_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_2_5
#define init_1_5_2_5
void FASTCALL call_1_5_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_2_6
#define init_1_5_2_6
void FASTCALL call_1_5_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_2_7
#define init_1_5_2_7
void FASTCALL call_1_5_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_2_8
#define init_1_5_2_8
void FASTCALL call_1_5_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_2_9
#define init_1_5_2_9
void FASTCALL call_1_5_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_2_10
#define init_1_5_2_10
void FASTCALL call_1_5_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_2_11
#define init_1_5_2_11
void FASTCALL call_1_5_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_2_12
#define init_1_5_2_12
void FASTCALL call_1_5_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_2_13
#define init_1_5_2_13
void FASTCALL call_1_5_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_2_14
#define init_1_5_2_14
void FASTCALL call_1_5_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_2_15
#define init_1_5_2_15
void FASTCALL call_1_5_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_3_0
#define init_1_5_3_0
void FASTCALL call_1_5_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_3_1
#define init_1_5_3_1
void FASTCALL call_1_5_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_3_2
#define init_1_5_3_2
void FASTCALL call_1_5_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_3_3
#define init_1_5_3_3
void FASTCALL call_1_5_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_3_4
#define init_1_5_3_4
void FASTCALL call_1_5_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_3_5
#define init_1_5_3_5
void FASTCALL call_1_5_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_3_6
#define init_1_5_3_6
void FASTCALL call_1_5_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_3_7
#define init_1_5_3_7
void FASTCALL call_1_5_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_3_8
#define init_1_5_3_8
void FASTCALL call_1_5_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_3_9
#define init_1_5_3_9
void FASTCALL call_1_5_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_3_10
#define init_1_5_3_10
void FASTCALL call_1_5_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_3_11
#define init_1_5_3_11
void FASTCALL call_1_5_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_3_12
#define init_1_5_3_12
void FASTCALL call_1_5_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_3_13
#define init_1_5_3_13
void FASTCALL call_1_5_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_3_14
#define init_1_5_3_14
void FASTCALL call_1_5_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_3_15
#define init_1_5_3_15
void FASTCALL call_1_5_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_4_0
#define init_1_5_4_0
void FASTCALL call_1_5_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_4_1
#define init_1_5_4_1
void FASTCALL call_1_5_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_4_2
#define init_1_5_4_2
void FASTCALL call_1_5_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_4_3
#define init_1_5_4_3
void FASTCALL call_1_5_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_4_4
#define init_1_5_4_4
void FASTCALL call_1_5_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_4_5
#define init_1_5_4_5
void FASTCALL call_1_5_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_4_6
#define init_1_5_4_6
void FASTCALL call_1_5_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_4_7
#define init_1_5_4_7
void FASTCALL call_1_5_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_4_8
#define init_1_5_4_8
void FASTCALL call_1_5_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_4_9
#define init_1_5_4_9
void FASTCALL call_1_5_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_4_10
#define init_1_5_4_10
void FASTCALL call_1_5_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_4_11
#define init_1_5_4_11
void FASTCALL call_1_5_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_4_12
#define init_1_5_4_12
void FASTCALL call_1_5_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_4_13
#define init_1_5_4_13
void FASTCALL call_1_5_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_4_14
#define init_1_5_4_14
void FASTCALL call_1_5_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_4_15
#define init_1_5_4_15
void FASTCALL call_1_5_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_5_0
#define init_1_5_5_0
void FASTCALL call_1_5_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_5_1
#define init_1_5_5_1
void FASTCALL call_1_5_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_5_2
#define init_1_5_5_2
void FASTCALL call_1_5_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_5_3
#define init_1_5_5_3
void FASTCALL call_1_5_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_5_4
#define init_1_5_5_4
void FASTCALL call_1_5_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_5_5
#define init_1_5_5_5
void FASTCALL call_1_5_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_5_6
#define init_1_5_5_6
void FASTCALL call_1_5_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_5_7
#define init_1_5_5_7
void FASTCALL call_1_5_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_5_8
#define init_1_5_5_8
void FASTCALL call_1_5_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_5_9
#define init_1_5_5_9
void FASTCALL call_1_5_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_5_10
#define init_1_5_5_10
void FASTCALL call_1_5_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_5_11
#define init_1_5_5_11
void FASTCALL call_1_5_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_5_12
#define init_1_5_5_12
void FASTCALL call_1_5_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_5_13
#define init_1_5_5_13
void FASTCALL call_1_5_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_5_14
#define init_1_5_5_14
void FASTCALL call_1_5_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_5_15
#define init_1_5_5_15
void FASTCALL call_1_5_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_6_0
#define init_1_5_6_0
void FASTCALL call_1_5_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_6_1
#define init_1_5_6_1
void FASTCALL call_1_5_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_6_2
#define init_1_5_6_2
void FASTCALL call_1_5_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_6_3
#define init_1_5_6_3
void FASTCALL call_1_5_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_6_4
#define init_1_5_6_4
void FASTCALL call_1_5_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_6_5
#define init_1_5_6_5
void FASTCALL call_1_5_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_6_6
#define init_1_5_6_6
void FASTCALL call_1_5_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_6_7
#define init_1_5_6_7
void FASTCALL call_1_5_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_6_8
#define init_1_5_6_8
void FASTCALL call_1_5_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_6_9
#define init_1_5_6_9
void FASTCALL call_1_5_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_6_10
#define init_1_5_6_10
void FASTCALL call_1_5_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_6_11
#define init_1_5_6_11
void FASTCALL call_1_5_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_6_12
#define init_1_5_6_12
void FASTCALL call_1_5_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_6_13
#define init_1_5_6_13
void FASTCALL call_1_5_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_6_14
#define init_1_5_6_14
void FASTCALL call_1_5_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_6_15
#define init_1_5_6_15
void FASTCALL call_1_5_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_7_0
#define init_1_5_7_0
void FASTCALL call_1_5_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_7_1
#define init_1_5_7_1
void FASTCALL call_1_5_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_7_2
#define init_1_5_7_2
void FASTCALL call_1_5_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_7_3
#define init_1_5_7_3
void FASTCALL call_1_5_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_7_4
#define init_1_5_7_4
void FASTCALL call_1_5_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_7_5
#define init_1_5_7_5
void FASTCALL call_1_5_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_7_6
#define init_1_5_7_6
void FASTCALL call_1_5_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_7_7
#define init_1_5_7_7
void FASTCALL call_1_5_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_7_8
#define init_1_5_7_8
void FASTCALL call_1_5_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_7_9
#define init_1_5_7_9
void FASTCALL call_1_5_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_7_10
#define init_1_5_7_10
void FASTCALL call_1_5_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_7_11
#define init_1_5_7_11
void FASTCALL call_1_5_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_7_12
#define init_1_5_7_12
void FASTCALL call_1_5_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_7_13
#define init_1_5_7_13
void FASTCALL call_1_5_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_7_14
#define init_1_5_7_14
void FASTCALL call_1_5_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_7_15
#define init_1_5_7_15
void FASTCALL call_1_5_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_8_0
#define init_1_5_8_0
void FASTCALL call_1_5_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_8_1
#define init_1_5_8_1
void FASTCALL call_1_5_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_8_2
#define init_1_5_8_2
void FASTCALL call_1_5_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_8_3
#define init_1_5_8_3
void FASTCALL call_1_5_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_8_4
#define init_1_5_8_4
void FASTCALL call_1_5_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_8_5
#define init_1_5_8_5
void FASTCALL call_1_5_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_8_6
#define init_1_5_8_6
void FASTCALL call_1_5_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_8_7
#define init_1_5_8_7
void FASTCALL call_1_5_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_8_8
#define init_1_5_8_8
void FASTCALL call_1_5_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_8_9
#define init_1_5_8_9
void FASTCALL call_1_5_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_8_10
#define init_1_5_8_10
void FASTCALL call_1_5_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_8_11
#define init_1_5_8_11
void FASTCALL call_1_5_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_8_12
#define init_1_5_8_12
void FASTCALL call_1_5_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_8_13
#define init_1_5_8_13
void FASTCALL call_1_5_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_8_14
#define init_1_5_8_14
void FASTCALL call_1_5_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_8_15
#define init_1_5_8_15
void FASTCALL call_1_5_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_9_0
#define init_1_5_9_0
void FASTCALL call_1_5_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_9_1
#define init_1_5_9_1
void FASTCALL call_1_5_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_9_2
#define init_1_5_9_2
void FASTCALL call_1_5_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_9_3
#define init_1_5_9_3
void FASTCALL call_1_5_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_9_4
#define init_1_5_9_4
void FASTCALL call_1_5_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_9_5
#define init_1_5_9_5
void FASTCALL call_1_5_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_9_6
#define init_1_5_9_6
void FASTCALL call_1_5_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_9_7
#define init_1_5_9_7
void FASTCALL call_1_5_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_9_8
#define init_1_5_9_8
void FASTCALL call_1_5_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_9_9
#define init_1_5_9_9
void FASTCALL call_1_5_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_9_10
#define init_1_5_9_10
void FASTCALL call_1_5_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_9_11
#define init_1_5_9_11
void FASTCALL call_1_5_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_9_12
#define init_1_5_9_12
void FASTCALL call_1_5_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_9_13
#define init_1_5_9_13
void FASTCALL call_1_5_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_9_14
#define init_1_5_9_14
void FASTCALL call_1_5_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_9_15
#define init_1_5_9_15
void FASTCALL call_1_5_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_10_0
#define init_1_5_10_0
void FASTCALL call_1_5_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_10_1
#define init_1_5_10_1
void FASTCALL call_1_5_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_10_2
#define init_1_5_10_2
void FASTCALL call_1_5_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_10_3
#define init_1_5_10_3
void FASTCALL call_1_5_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_10_4
#define init_1_5_10_4
void FASTCALL call_1_5_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_10_5
#define init_1_5_10_5
void FASTCALL call_1_5_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_10_6
#define init_1_5_10_6
void FASTCALL call_1_5_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_10_7
#define init_1_5_10_7
void FASTCALL call_1_5_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_10_8
#define init_1_5_10_8
void FASTCALL call_1_5_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_10_9
#define init_1_5_10_9
void FASTCALL call_1_5_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_10_10
#define init_1_5_10_10
void FASTCALL call_1_5_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_10_11
#define init_1_5_10_11
void FASTCALL call_1_5_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_10_12
#define init_1_5_10_12
void FASTCALL call_1_5_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_10_13
#define init_1_5_10_13
void FASTCALL call_1_5_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_10_14
#define init_1_5_10_14
void FASTCALL call_1_5_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_10_15
#define init_1_5_10_15
void FASTCALL call_1_5_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_11_0
#define init_1_5_11_0
void FASTCALL call_1_5_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_11_1
#define init_1_5_11_1
void FASTCALL call_1_5_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_11_2
#define init_1_5_11_2
void FASTCALL call_1_5_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_11_3
#define init_1_5_11_3
void FASTCALL call_1_5_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_11_4
#define init_1_5_11_4
void FASTCALL call_1_5_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_11_5
#define init_1_5_11_5
void FASTCALL call_1_5_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_11_6
#define init_1_5_11_6
void FASTCALL call_1_5_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_11_7
#define init_1_5_11_7
void FASTCALL call_1_5_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_11_8
#define init_1_5_11_8
void FASTCALL call_1_5_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_11_9
#define init_1_5_11_9
void FASTCALL call_1_5_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_11_10
#define init_1_5_11_10
void FASTCALL call_1_5_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_11_11
#define init_1_5_11_11
void FASTCALL call_1_5_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_11_12
#define init_1_5_11_12
void FASTCALL call_1_5_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_11_13
#define init_1_5_11_13
void FASTCALL call_1_5_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_11_14
#define init_1_5_11_14
void FASTCALL call_1_5_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_11_15
#define init_1_5_11_15
void FASTCALL call_1_5_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_12_0
#define init_1_5_12_0
void FASTCALL call_1_5_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_12_1
#define init_1_5_12_1
void FASTCALL call_1_5_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_12_2
#define init_1_5_12_2
void FASTCALL call_1_5_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_12_3
#define init_1_5_12_3
void FASTCALL call_1_5_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_12_4
#define init_1_5_12_4
void FASTCALL call_1_5_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_12_5
#define init_1_5_12_5
void FASTCALL call_1_5_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_12_6
#define init_1_5_12_6
void FASTCALL call_1_5_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_12_7
#define init_1_5_12_7
void FASTCALL call_1_5_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_12_8
#define init_1_5_12_8
void FASTCALL call_1_5_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_12_9
#define init_1_5_12_9
void FASTCALL call_1_5_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_12_10
#define init_1_5_12_10
void FASTCALL call_1_5_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_12_11
#define init_1_5_12_11
void FASTCALL call_1_5_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_12_12
#define init_1_5_12_12
void FASTCALL call_1_5_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_12_13
#define init_1_5_12_13
void FASTCALL call_1_5_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_12_14
#define init_1_5_12_14
void FASTCALL call_1_5_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_12_15
#define init_1_5_12_15
void FASTCALL call_1_5_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_13_0
#define init_1_5_13_0
void FASTCALL call_1_5_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_13_1
#define init_1_5_13_1
void FASTCALL call_1_5_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_13_2
#define init_1_5_13_2
void FASTCALL call_1_5_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_13_3
#define init_1_5_13_3
void FASTCALL call_1_5_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_13_4
#define init_1_5_13_4
void FASTCALL call_1_5_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_13_5
#define init_1_5_13_5
void FASTCALL call_1_5_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_13_6
#define init_1_5_13_6
void FASTCALL call_1_5_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_13_7
#define init_1_5_13_7
void FASTCALL call_1_5_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_13_8
#define init_1_5_13_8
void FASTCALL call_1_5_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_13_9
#define init_1_5_13_9
void FASTCALL call_1_5_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_13_10
#define init_1_5_13_10
void FASTCALL call_1_5_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_13_11
#define init_1_5_13_11
void FASTCALL call_1_5_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_13_12
#define init_1_5_13_12
void FASTCALL call_1_5_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_13_13
#define init_1_5_13_13
void FASTCALL call_1_5_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_13_14
#define init_1_5_13_14
void FASTCALL call_1_5_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_13_15
#define init_1_5_13_15
void FASTCALL call_1_5_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_14_0
#define init_1_5_14_0
void FASTCALL call_1_5_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_14_1
#define init_1_5_14_1
void FASTCALL call_1_5_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_14_2
#define init_1_5_14_2
void FASTCALL call_1_5_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_14_3
#define init_1_5_14_3
void FASTCALL call_1_5_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_14_4
#define init_1_5_14_4
void FASTCALL call_1_5_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_14_5
#define init_1_5_14_5
void FASTCALL call_1_5_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_14_6
#define init_1_5_14_6
void FASTCALL call_1_5_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_14_7
#define init_1_5_14_7
void FASTCALL call_1_5_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_14_8
#define init_1_5_14_8
void FASTCALL call_1_5_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_14_9
#define init_1_5_14_9
void FASTCALL call_1_5_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_14_10
#define init_1_5_14_10
void FASTCALL call_1_5_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_14_11
#define init_1_5_14_11
void FASTCALL call_1_5_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_14_12
#define init_1_5_14_12
void FASTCALL call_1_5_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_14_13
#define init_1_5_14_13
void FASTCALL call_1_5_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_14_14
#define init_1_5_14_14
void FASTCALL call_1_5_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_14_15
#define init_1_5_14_15
void FASTCALL call_1_5_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_15_0
#define init_1_5_15_0
void FASTCALL call_1_5_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_15_1
#define init_1_5_15_1
void FASTCALL call_1_5_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_15_2
#define init_1_5_15_2
void FASTCALL call_1_5_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_15_3
#define init_1_5_15_3
void FASTCALL call_1_5_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_15_4
#define init_1_5_15_4
void FASTCALL call_1_5_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_15_5
#define init_1_5_15_5
void FASTCALL call_1_5_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_15_6
#define init_1_5_15_6
void FASTCALL call_1_5_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_15_7
#define init_1_5_15_7
void FASTCALL call_1_5_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_15_8
#define init_1_5_15_8
void FASTCALL call_1_5_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_15_9
#define init_1_5_15_9
void FASTCALL call_1_5_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_15_10
#define init_1_5_15_10
void FASTCALL call_1_5_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_15_11
#define init_1_5_15_11
void FASTCALL call_1_5_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_15_12
#define init_1_5_15_12
void FASTCALL call_1_5_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_15_13
#define init_1_5_15_13
void FASTCALL call_1_5_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_15_14
#define init_1_5_15_14
void FASTCALL call_1_5_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_5_15_15
#define init_1_5_15_15
void FASTCALL call_1_5_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_0_0
#define init_1_6_0_0
void FASTCALL call_1_6_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_0_1
#define init_1_6_0_1
void FASTCALL call_1_6_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_0_2
#define init_1_6_0_2
void FASTCALL call_1_6_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_0_3
#define init_1_6_0_3
void FASTCALL call_1_6_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_0_4
#define init_1_6_0_4
void FASTCALL call_1_6_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_0_5
#define init_1_6_0_5
void FASTCALL call_1_6_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_0_6
#define init_1_6_0_6
void FASTCALL call_1_6_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_0_7
#define init_1_6_0_7
void FASTCALL call_1_6_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_0_8
#define init_1_6_0_8
void FASTCALL call_1_6_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_0_9
#define init_1_6_0_9
void FASTCALL call_1_6_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_0_10
#define init_1_6_0_10
void FASTCALL call_1_6_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_0_11
#define init_1_6_0_11
void FASTCALL call_1_6_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_0_12
#define init_1_6_0_12
void FASTCALL call_1_6_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_0_13
#define init_1_6_0_13
void FASTCALL call_1_6_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_0_14
#define init_1_6_0_14
void FASTCALL call_1_6_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_0_15
#define init_1_6_0_15
void FASTCALL call_1_6_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_1_0
#define init_1_6_1_0
void FASTCALL call_1_6_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_1_1
#define init_1_6_1_1
void FASTCALL call_1_6_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_1_2
#define init_1_6_1_2
void FASTCALL call_1_6_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_1_3
#define init_1_6_1_3
void FASTCALL call_1_6_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_1_4
#define init_1_6_1_4
void FASTCALL call_1_6_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_1_5
#define init_1_6_1_5
void FASTCALL call_1_6_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_1_6
#define init_1_6_1_6
void FASTCALL call_1_6_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_1_7
#define init_1_6_1_7
void FASTCALL call_1_6_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_1_8
#define init_1_6_1_8
void FASTCALL call_1_6_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_1_9
#define init_1_6_1_9
void FASTCALL call_1_6_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_1_10
#define init_1_6_1_10
void FASTCALL call_1_6_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_1_11
#define init_1_6_1_11
void FASTCALL call_1_6_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_1_12
#define init_1_6_1_12
void FASTCALL call_1_6_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_1_13
#define init_1_6_1_13
void FASTCALL call_1_6_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_1_14
#define init_1_6_1_14
void FASTCALL call_1_6_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_1_15
#define init_1_6_1_15
void FASTCALL call_1_6_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_2_0
#define init_1_6_2_0
void FASTCALL call_1_6_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_2_1
#define init_1_6_2_1
void FASTCALL call_1_6_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_2_2
#define init_1_6_2_2
void FASTCALL call_1_6_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_2_3
#define init_1_6_2_3
void FASTCALL call_1_6_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_2_4
#define init_1_6_2_4
void FASTCALL call_1_6_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_2_5
#define init_1_6_2_5
void FASTCALL call_1_6_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_2_6
#define init_1_6_2_6
void FASTCALL call_1_6_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_2_7
#define init_1_6_2_7
void FASTCALL call_1_6_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_2_8
#define init_1_6_2_8
void FASTCALL call_1_6_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_2_9
#define init_1_6_2_9
void FASTCALL call_1_6_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_2_10
#define init_1_6_2_10
void FASTCALL call_1_6_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_2_11
#define init_1_6_2_11
void FASTCALL call_1_6_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_2_12
#define init_1_6_2_12
void FASTCALL call_1_6_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_2_13
#define init_1_6_2_13
void FASTCALL call_1_6_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_2_14
#define init_1_6_2_14
void FASTCALL call_1_6_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_2_15
#define init_1_6_2_15
void FASTCALL call_1_6_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_3_0
#define init_1_6_3_0
void FASTCALL call_1_6_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_3_1
#define init_1_6_3_1
void FASTCALL call_1_6_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_3_2
#define init_1_6_3_2
void FASTCALL call_1_6_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_3_3
#define init_1_6_3_3
void FASTCALL call_1_6_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_3_4
#define init_1_6_3_4
void FASTCALL call_1_6_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_3_5
#define init_1_6_3_5
void FASTCALL call_1_6_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_3_6
#define init_1_6_3_6
void FASTCALL call_1_6_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_3_7
#define init_1_6_3_7
void FASTCALL call_1_6_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_3_8
#define init_1_6_3_8
void FASTCALL call_1_6_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_3_9
#define init_1_6_3_9
void FASTCALL call_1_6_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_3_10
#define init_1_6_3_10
void FASTCALL call_1_6_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_3_11
#define init_1_6_3_11
void FASTCALL call_1_6_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_3_12
#define init_1_6_3_12
void FASTCALL call_1_6_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_3_13
#define init_1_6_3_13
void FASTCALL call_1_6_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_3_14
#define init_1_6_3_14
void FASTCALL call_1_6_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_3_15
#define init_1_6_3_15
void FASTCALL call_1_6_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_4_0
#define init_1_6_4_0
void FASTCALL call_1_6_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_4_1
#define init_1_6_4_1
void FASTCALL call_1_6_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_4_2
#define init_1_6_4_2
void FASTCALL call_1_6_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_4_3
#define init_1_6_4_3
void FASTCALL call_1_6_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_4_4
#define init_1_6_4_4
void FASTCALL call_1_6_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_4_5
#define init_1_6_4_5
void FASTCALL call_1_6_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_4_6
#define init_1_6_4_6
void FASTCALL call_1_6_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_4_7
#define init_1_6_4_7
void FASTCALL call_1_6_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_4_8
#define init_1_6_4_8
void FASTCALL call_1_6_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_4_9
#define init_1_6_4_9
void FASTCALL call_1_6_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_4_10
#define init_1_6_4_10
void FASTCALL call_1_6_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_4_11
#define init_1_6_4_11
void FASTCALL call_1_6_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_4_12
#define init_1_6_4_12
void FASTCALL call_1_6_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_4_13
#define init_1_6_4_13
void FASTCALL call_1_6_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_4_14
#define init_1_6_4_14
void FASTCALL call_1_6_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_4_15
#define init_1_6_4_15
void FASTCALL call_1_6_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_5_0
#define init_1_6_5_0
void FASTCALL call_1_6_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_5_1
#define init_1_6_5_1
void FASTCALL call_1_6_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_5_2
#define init_1_6_5_2
void FASTCALL call_1_6_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_5_3
#define init_1_6_5_3
void FASTCALL call_1_6_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_5_4
#define init_1_6_5_4
void FASTCALL call_1_6_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_5_5
#define init_1_6_5_5
void FASTCALL call_1_6_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_5_6
#define init_1_6_5_6
void FASTCALL call_1_6_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_5_7
#define init_1_6_5_7
void FASTCALL call_1_6_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_5_8
#define init_1_6_5_8
void FASTCALL call_1_6_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_5_9
#define init_1_6_5_9
void FASTCALL call_1_6_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_5_10
#define init_1_6_5_10
void FASTCALL call_1_6_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_5_11
#define init_1_6_5_11
void FASTCALL call_1_6_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_5_12
#define init_1_6_5_12
void FASTCALL call_1_6_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_5_13
#define init_1_6_5_13
void FASTCALL call_1_6_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_5_14
#define init_1_6_5_14
void FASTCALL call_1_6_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_5_15
#define init_1_6_5_15
void FASTCALL call_1_6_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_6_0
#define init_1_6_6_0
void FASTCALL call_1_6_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_6_1
#define init_1_6_6_1
void FASTCALL call_1_6_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_6_2
#define init_1_6_6_2
void FASTCALL call_1_6_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_6_3
#define init_1_6_6_3
void FASTCALL call_1_6_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_6_4
#define init_1_6_6_4
void FASTCALL call_1_6_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_6_5
#define init_1_6_6_5
void FASTCALL call_1_6_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_6_6
#define init_1_6_6_6
void FASTCALL call_1_6_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_6_7
#define init_1_6_6_7
void FASTCALL call_1_6_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_6_8
#define init_1_6_6_8
void FASTCALL call_1_6_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_6_9
#define init_1_6_6_9
void FASTCALL call_1_6_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_6_10
#define init_1_6_6_10
void FASTCALL call_1_6_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_6_11
#define init_1_6_6_11
void FASTCALL call_1_6_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_6_12
#define init_1_6_6_12
void FASTCALL call_1_6_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_6_13
#define init_1_6_6_13
void FASTCALL call_1_6_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_6_14
#define init_1_6_6_14
void FASTCALL call_1_6_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_6_15
#define init_1_6_6_15
void FASTCALL call_1_6_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_7_0
#define init_1_6_7_0
void FASTCALL call_1_6_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_7_1
#define init_1_6_7_1
void FASTCALL call_1_6_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_7_2
#define init_1_6_7_2
void FASTCALL call_1_6_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_7_3
#define init_1_6_7_3
void FASTCALL call_1_6_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_7_4
#define init_1_6_7_4
void FASTCALL call_1_6_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_7_5
#define init_1_6_7_5
void FASTCALL call_1_6_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_7_6
#define init_1_6_7_6
void FASTCALL call_1_6_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_7_7
#define init_1_6_7_7
void FASTCALL call_1_6_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_7_8
#define init_1_6_7_8
void FASTCALL call_1_6_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_7_9
#define init_1_6_7_9
void FASTCALL call_1_6_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_7_10
#define init_1_6_7_10
void FASTCALL call_1_6_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_7_11
#define init_1_6_7_11
void FASTCALL call_1_6_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_7_12
#define init_1_6_7_12
void FASTCALL call_1_6_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_7_13
#define init_1_6_7_13
void FASTCALL call_1_6_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_7_14
#define init_1_6_7_14
void FASTCALL call_1_6_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_7_15
#define init_1_6_7_15
void FASTCALL call_1_6_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_8_0
#define init_1_6_8_0
void FASTCALL call_1_6_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_8_1
#define init_1_6_8_1
void FASTCALL call_1_6_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_8_2
#define init_1_6_8_2
void FASTCALL call_1_6_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_8_3
#define init_1_6_8_3
void FASTCALL call_1_6_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_8_4
#define init_1_6_8_4
void FASTCALL call_1_6_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_8_5
#define init_1_6_8_5
void FASTCALL call_1_6_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_8_6
#define init_1_6_8_6
void FASTCALL call_1_6_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_8_7
#define init_1_6_8_7
void FASTCALL call_1_6_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_8_8
#define init_1_6_8_8
void FASTCALL call_1_6_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_8_9
#define init_1_6_8_9
void FASTCALL call_1_6_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_8_10
#define init_1_6_8_10
void FASTCALL call_1_6_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_8_11
#define init_1_6_8_11
void FASTCALL call_1_6_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_8_12
#define init_1_6_8_12
void FASTCALL call_1_6_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_8_13
#define init_1_6_8_13
void FASTCALL call_1_6_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_8_14
#define init_1_6_8_14
void FASTCALL call_1_6_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_8_15
#define init_1_6_8_15
void FASTCALL call_1_6_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_9_0
#define init_1_6_9_0
void FASTCALL call_1_6_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_9_1
#define init_1_6_9_1
void FASTCALL call_1_6_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_9_2
#define init_1_6_9_2
void FASTCALL call_1_6_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_9_3
#define init_1_6_9_3
void FASTCALL call_1_6_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_9_4
#define init_1_6_9_4
void FASTCALL call_1_6_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_9_5
#define init_1_6_9_5
void FASTCALL call_1_6_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_9_6
#define init_1_6_9_6
void FASTCALL call_1_6_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_9_7
#define init_1_6_9_7
void FASTCALL call_1_6_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_9_8
#define init_1_6_9_8
void FASTCALL call_1_6_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_9_9
#define init_1_6_9_9
void FASTCALL call_1_6_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_9_10
#define init_1_6_9_10
void FASTCALL call_1_6_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_9_11
#define init_1_6_9_11
void FASTCALL call_1_6_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_9_12
#define init_1_6_9_12
void FASTCALL call_1_6_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_9_13
#define init_1_6_9_13
void FASTCALL call_1_6_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_9_14
#define init_1_6_9_14
void FASTCALL call_1_6_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_9_15
#define init_1_6_9_15
void FASTCALL call_1_6_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_10_0
#define init_1_6_10_0
void FASTCALL call_1_6_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_10_1
#define init_1_6_10_1
void FASTCALL call_1_6_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_10_2
#define init_1_6_10_2
void FASTCALL call_1_6_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_10_3
#define init_1_6_10_3
void FASTCALL call_1_6_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_10_4
#define init_1_6_10_4
void FASTCALL call_1_6_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_10_5
#define init_1_6_10_5
void FASTCALL call_1_6_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_10_6
#define init_1_6_10_6
void FASTCALL call_1_6_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_10_7
#define init_1_6_10_7
void FASTCALL call_1_6_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_10_8
#define init_1_6_10_8
void FASTCALL call_1_6_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_10_9
#define init_1_6_10_9
void FASTCALL call_1_6_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_10_10
#define init_1_6_10_10
void FASTCALL call_1_6_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_10_11
#define init_1_6_10_11
void FASTCALL call_1_6_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_10_12
#define init_1_6_10_12
void FASTCALL call_1_6_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_10_13
#define init_1_6_10_13
void FASTCALL call_1_6_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_10_14
#define init_1_6_10_14
void FASTCALL call_1_6_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_10_15
#define init_1_6_10_15
void FASTCALL call_1_6_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_11_0
#define init_1_6_11_0
void FASTCALL call_1_6_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_11_1
#define init_1_6_11_1
void FASTCALL call_1_6_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_11_2
#define init_1_6_11_2
void FASTCALL call_1_6_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_11_3
#define init_1_6_11_3
void FASTCALL call_1_6_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_11_4
#define init_1_6_11_4
void FASTCALL call_1_6_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_11_5
#define init_1_6_11_5
void FASTCALL call_1_6_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_11_6
#define init_1_6_11_6
void FASTCALL call_1_6_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_11_7
#define init_1_6_11_7
void FASTCALL call_1_6_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_11_8
#define init_1_6_11_8
void FASTCALL call_1_6_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_11_9
#define init_1_6_11_9
void FASTCALL call_1_6_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_11_10
#define init_1_6_11_10
void FASTCALL call_1_6_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_11_11
#define init_1_6_11_11
void FASTCALL call_1_6_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_11_12
#define init_1_6_11_12
void FASTCALL call_1_6_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_11_13
#define init_1_6_11_13
void FASTCALL call_1_6_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_11_14
#define init_1_6_11_14
void FASTCALL call_1_6_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_11_15
#define init_1_6_11_15
void FASTCALL call_1_6_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_12_0
#define init_1_6_12_0
void FASTCALL call_1_6_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_12_1
#define init_1_6_12_1
void FASTCALL call_1_6_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_12_2
#define init_1_6_12_2
void FASTCALL call_1_6_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_12_3
#define init_1_6_12_3
void FASTCALL call_1_6_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_12_4
#define init_1_6_12_4
void FASTCALL call_1_6_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_12_5
#define init_1_6_12_5
void FASTCALL call_1_6_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_12_6
#define init_1_6_12_6
void FASTCALL call_1_6_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_12_7
#define init_1_6_12_7
void FASTCALL call_1_6_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_12_8
#define init_1_6_12_8
void FASTCALL call_1_6_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_12_9
#define init_1_6_12_9
void FASTCALL call_1_6_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_12_10
#define init_1_6_12_10
void FASTCALL call_1_6_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_12_11
#define init_1_6_12_11
void FASTCALL call_1_6_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_12_12
#define init_1_6_12_12
void FASTCALL call_1_6_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_12_13
#define init_1_6_12_13
void FASTCALL call_1_6_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_12_14
#define init_1_6_12_14
void FASTCALL call_1_6_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_12_15
#define init_1_6_12_15
void FASTCALL call_1_6_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_13_0
#define init_1_6_13_0
void FASTCALL call_1_6_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_13_1
#define init_1_6_13_1
void FASTCALL call_1_6_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_13_2
#define init_1_6_13_2
void FASTCALL call_1_6_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_13_3
#define init_1_6_13_3
void FASTCALL call_1_6_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_13_4
#define init_1_6_13_4
void FASTCALL call_1_6_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_13_5
#define init_1_6_13_5
void FASTCALL call_1_6_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_13_6
#define init_1_6_13_6
void FASTCALL call_1_6_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_13_7
#define init_1_6_13_7
void FASTCALL call_1_6_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_13_8
#define init_1_6_13_8
void FASTCALL call_1_6_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_13_9
#define init_1_6_13_9
void FASTCALL call_1_6_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_13_10
#define init_1_6_13_10
void FASTCALL call_1_6_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_13_11
#define init_1_6_13_11
void FASTCALL call_1_6_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_13_12
#define init_1_6_13_12
void FASTCALL call_1_6_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_13_13
#define init_1_6_13_13
void FASTCALL call_1_6_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_13_14
#define init_1_6_13_14
void FASTCALL call_1_6_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_13_15
#define init_1_6_13_15
void FASTCALL call_1_6_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_14_0
#define init_1_6_14_0
void FASTCALL call_1_6_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_14_1
#define init_1_6_14_1
void FASTCALL call_1_6_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_14_2
#define init_1_6_14_2
void FASTCALL call_1_6_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_14_3
#define init_1_6_14_3
void FASTCALL call_1_6_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_14_4
#define init_1_6_14_4
void FASTCALL call_1_6_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_14_5
#define init_1_6_14_5
void FASTCALL call_1_6_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_14_6
#define init_1_6_14_6
void FASTCALL call_1_6_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_14_7
#define init_1_6_14_7
void FASTCALL call_1_6_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_14_8
#define init_1_6_14_8
void FASTCALL call_1_6_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_14_9
#define init_1_6_14_9
void FASTCALL call_1_6_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_14_10
#define init_1_6_14_10
void FASTCALL call_1_6_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_14_11
#define init_1_6_14_11
void FASTCALL call_1_6_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_14_12
#define init_1_6_14_12
void FASTCALL call_1_6_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_14_13
#define init_1_6_14_13
void FASTCALL call_1_6_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_14_14
#define init_1_6_14_14
void FASTCALL call_1_6_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_14_15
#define init_1_6_14_15
void FASTCALL call_1_6_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_15_0
#define init_1_6_15_0
void FASTCALL call_1_6_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_15_1
#define init_1_6_15_1
void FASTCALL call_1_6_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_15_2
#define init_1_6_15_2
void FASTCALL call_1_6_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_15_3
#define init_1_6_15_3
void FASTCALL call_1_6_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_15_4
#define init_1_6_15_4
void FASTCALL call_1_6_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_15_5
#define init_1_6_15_5
void FASTCALL call_1_6_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_15_6
#define init_1_6_15_6
void FASTCALL call_1_6_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_15_7
#define init_1_6_15_7
void FASTCALL call_1_6_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_15_8
#define init_1_6_15_8
void FASTCALL call_1_6_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_15_9
#define init_1_6_15_9
void FASTCALL call_1_6_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_15_10
#define init_1_6_15_10
void FASTCALL call_1_6_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_15_11
#define init_1_6_15_11
void FASTCALL call_1_6_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_15_12
#define init_1_6_15_12
void FASTCALL call_1_6_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_15_13
#define init_1_6_15_13
void FASTCALL call_1_6_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_15_14
#define init_1_6_15_14
void FASTCALL call_1_6_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_6_15_15
#define init_1_6_15_15
void FASTCALL call_1_6_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_0_0
#define init_1_7_0_0
void FASTCALL call_1_7_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_0_1
#define init_1_7_0_1
void FASTCALL call_1_7_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_0_2
#define init_1_7_0_2
void FASTCALL call_1_7_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_0_3
#define init_1_7_0_3
void FASTCALL call_1_7_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_0_4
#define init_1_7_0_4
void FASTCALL call_1_7_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_0_5
#define init_1_7_0_5
void FASTCALL call_1_7_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_0_6
#define init_1_7_0_6
void FASTCALL call_1_7_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_0_7
#define init_1_7_0_7
void FASTCALL call_1_7_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_0_8
#define init_1_7_0_8
void FASTCALL call_1_7_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_0_9
#define init_1_7_0_9
void FASTCALL call_1_7_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_0_10
#define init_1_7_0_10
void FASTCALL call_1_7_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_0_11
#define init_1_7_0_11
void FASTCALL call_1_7_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_0_12
#define init_1_7_0_12
void FASTCALL call_1_7_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_0_13
#define init_1_7_0_13
void FASTCALL call_1_7_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_0_14
#define init_1_7_0_14
void FASTCALL call_1_7_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_0_15
#define init_1_7_0_15
void FASTCALL call_1_7_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_1_0
#define init_1_7_1_0
void FASTCALL call_1_7_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_1_1
#define init_1_7_1_1
void FASTCALL call_1_7_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_1_2
#define init_1_7_1_2
void FASTCALL call_1_7_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_1_3
#define init_1_7_1_3
void FASTCALL call_1_7_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_1_4
#define init_1_7_1_4
void FASTCALL call_1_7_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_1_5
#define init_1_7_1_5
void FASTCALL call_1_7_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_1_6
#define init_1_7_1_6
void FASTCALL call_1_7_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_1_7
#define init_1_7_1_7
void FASTCALL call_1_7_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_1_8
#define init_1_7_1_8
void FASTCALL call_1_7_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_1_9
#define init_1_7_1_9
void FASTCALL call_1_7_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_1_10
#define init_1_7_1_10
void FASTCALL call_1_7_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_1_11
#define init_1_7_1_11
void FASTCALL call_1_7_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_1_12
#define init_1_7_1_12
void FASTCALL call_1_7_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_1_13
#define init_1_7_1_13
void FASTCALL call_1_7_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_1_14
#define init_1_7_1_14
void FASTCALL call_1_7_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_1_15
#define init_1_7_1_15
void FASTCALL call_1_7_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_2_0
#define init_1_7_2_0
void FASTCALL call_1_7_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_2_1
#define init_1_7_2_1
void FASTCALL call_1_7_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_2_2
#define init_1_7_2_2
void FASTCALL call_1_7_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_2_3
#define init_1_7_2_3
void FASTCALL call_1_7_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_2_4
#define init_1_7_2_4
void FASTCALL call_1_7_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_2_5
#define init_1_7_2_5
void FASTCALL call_1_7_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_2_6
#define init_1_7_2_6
void FASTCALL call_1_7_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_2_7
#define init_1_7_2_7
void FASTCALL call_1_7_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_2_8
#define init_1_7_2_8
void FASTCALL call_1_7_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_2_9
#define init_1_7_2_9
void FASTCALL call_1_7_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_2_10
#define init_1_7_2_10
void FASTCALL call_1_7_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_2_11
#define init_1_7_2_11
void FASTCALL call_1_7_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_2_12
#define init_1_7_2_12
void FASTCALL call_1_7_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_2_13
#define init_1_7_2_13
void FASTCALL call_1_7_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_2_14
#define init_1_7_2_14
void FASTCALL call_1_7_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_2_15
#define init_1_7_2_15
void FASTCALL call_1_7_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_3_0
#define init_1_7_3_0
void FASTCALL call_1_7_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_3_1
#define init_1_7_3_1
void FASTCALL call_1_7_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_3_2
#define init_1_7_3_2
void FASTCALL call_1_7_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_3_3
#define init_1_7_3_3
void FASTCALL call_1_7_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_3_4
#define init_1_7_3_4
void FASTCALL call_1_7_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_3_5
#define init_1_7_3_5
void FASTCALL call_1_7_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_3_6
#define init_1_7_3_6
void FASTCALL call_1_7_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_3_7
#define init_1_7_3_7
void FASTCALL call_1_7_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_3_8
#define init_1_7_3_8
void FASTCALL call_1_7_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_3_9
#define init_1_7_3_9
void FASTCALL call_1_7_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_3_10
#define init_1_7_3_10
void FASTCALL call_1_7_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_3_11
#define init_1_7_3_11
void FASTCALL call_1_7_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_3_12
#define init_1_7_3_12
void FASTCALL call_1_7_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_3_13
#define init_1_7_3_13
void FASTCALL call_1_7_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_3_14
#define init_1_7_3_14
void FASTCALL call_1_7_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_3_15
#define init_1_7_3_15
void FASTCALL call_1_7_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_4_0
#define init_1_7_4_0
void FASTCALL call_1_7_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_4_1
#define init_1_7_4_1
void FASTCALL call_1_7_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_4_2
#define init_1_7_4_2
void FASTCALL call_1_7_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_4_3
#define init_1_7_4_3
void FASTCALL call_1_7_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_4_4
#define init_1_7_4_4
void FASTCALL call_1_7_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_4_5
#define init_1_7_4_5
void FASTCALL call_1_7_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_4_6
#define init_1_7_4_6
void FASTCALL call_1_7_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_4_7
#define init_1_7_4_7
void FASTCALL call_1_7_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_4_8
#define init_1_7_4_8
void FASTCALL call_1_7_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_4_9
#define init_1_7_4_9
void FASTCALL call_1_7_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_4_10
#define init_1_7_4_10
void FASTCALL call_1_7_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_4_11
#define init_1_7_4_11
void FASTCALL call_1_7_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_4_12
#define init_1_7_4_12
void FASTCALL call_1_7_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_4_13
#define init_1_7_4_13
void FASTCALL call_1_7_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_4_14
#define init_1_7_4_14
void FASTCALL call_1_7_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_4_15
#define init_1_7_4_15
void FASTCALL call_1_7_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_5_0
#define init_1_7_5_0
void FASTCALL call_1_7_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_5_1
#define init_1_7_5_1
void FASTCALL call_1_7_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_5_2
#define init_1_7_5_2
void FASTCALL call_1_7_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_5_3
#define init_1_7_5_3
void FASTCALL call_1_7_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_5_4
#define init_1_7_5_4
void FASTCALL call_1_7_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_5_5
#define init_1_7_5_5
void FASTCALL call_1_7_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_5_6
#define init_1_7_5_6
void FASTCALL call_1_7_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_5_7
#define init_1_7_5_7
void FASTCALL call_1_7_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_5_8
#define init_1_7_5_8
void FASTCALL call_1_7_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_5_9
#define init_1_7_5_9
void FASTCALL call_1_7_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_5_10
#define init_1_7_5_10
void FASTCALL call_1_7_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_5_11
#define init_1_7_5_11
void FASTCALL call_1_7_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_5_12
#define init_1_7_5_12
void FASTCALL call_1_7_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_5_13
#define init_1_7_5_13
void FASTCALL call_1_7_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_5_14
#define init_1_7_5_14
void FASTCALL call_1_7_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_5_15
#define init_1_7_5_15
void FASTCALL call_1_7_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_6_0
#define init_1_7_6_0
void FASTCALL call_1_7_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_6_1
#define init_1_7_6_1
void FASTCALL call_1_7_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_6_2
#define init_1_7_6_2
void FASTCALL call_1_7_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_6_3
#define init_1_7_6_3
void FASTCALL call_1_7_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_6_4
#define init_1_7_6_4
void FASTCALL call_1_7_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_6_5
#define init_1_7_6_5
void FASTCALL call_1_7_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_6_6
#define init_1_7_6_6
void FASTCALL call_1_7_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_6_7
#define init_1_7_6_7
void FASTCALL call_1_7_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_6_8
#define init_1_7_6_8
void FASTCALL call_1_7_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_6_9
#define init_1_7_6_9
void FASTCALL call_1_7_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_6_10
#define init_1_7_6_10
void FASTCALL call_1_7_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_6_11
#define init_1_7_6_11
void FASTCALL call_1_7_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_6_12
#define init_1_7_6_12
void FASTCALL call_1_7_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_6_13
#define init_1_7_6_13
void FASTCALL call_1_7_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_6_14
#define init_1_7_6_14
void FASTCALL call_1_7_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_6_15
#define init_1_7_6_15
void FASTCALL call_1_7_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_7_0
#define init_1_7_7_0
void FASTCALL call_1_7_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_7_1
#define init_1_7_7_1
void FASTCALL call_1_7_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_7_2
#define init_1_7_7_2
void FASTCALL call_1_7_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_7_3
#define init_1_7_7_3
void FASTCALL call_1_7_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_7_4
#define init_1_7_7_4
void FASTCALL call_1_7_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_7_5
#define init_1_7_7_5
void FASTCALL call_1_7_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_7_6
#define init_1_7_7_6
void FASTCALL call_1_7_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_7_7
#define init_1_7_7_7
void FASTCALL call_1_7_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_7_8
#define init_1_7_7_8
void FASTCALL call_1_7_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_7_9
#define init_1_7_7_9
void FASTCALL call_1_7_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_7_10
#define init_1_7_7_10
void FASTCALL call_1_7_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_7_11
#define init_1_7_7_11
void FASTCALL call_1_7_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_7_12
#define init_1_7_7_12
void FASTCALL call_1_7_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_7_13
#define init_1_7_7_13
void FASTCALL call_1_7_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_7_14
#define init_1_7_7_14
void FASTCALL call_1_7_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_7_15
#define init_1_7_7_15
void FASTCALL call_1_7_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_8_0
#define init_1_7_8_0
void FASTCALL call_1_7_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_8_1
#define init_1_7_8_1
void FASTCALL call_1_7_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_8_2
#define init_1_7_8_2
void FASTCALL call_1_7_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_8_3
#define init_1_7_8_3
void FASTCALL call_1_7_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_8_4
#define init_1_7_8_4
void FASTCALL call_1_7_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_8_5
#define init_1_7_8_5
void FASTCALL call_1_7_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_8_6
#define init_1_7_8_6
void FASTCALL call_1_7_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_8_7
#define init_1_7_8_7
void FASTCALL call_1_7_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_8_8
#define init_1_7_8_8
void FASTCALL call_1_7_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_8_9
#define init_1_7_8_9
void FASTCALL call_1_7_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_8_10
#define init_1_7_8_10
void FASTCALL call_1_7_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_8_11
#define init_1_7_8_11
void FASTCALL call_1_7_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_8_12
#define init_1_7_8_12
void FASTCALL call_1_7_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_8_13
#define init_1_7_8_13
void FASTCALL call_1_7_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_8_14
#define init_1_7_8_14
void FASTCALL call_1_7_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_8_15
#define init_1_7_8_15
void FASTCALL call_1_7_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_9_0
#define init_1_7_9_0
void FASTCALL call_1_7_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_9_1
#define init_1_7_9_1
void FASTCALL call_1_7_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_9_2
#define init_1_7_9_2
void FASTCALL call_1_7_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_9_3
#define init_1_7_9_3
void FASTCALL call_1_7_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_9_4
#define init_1_7_9_4
void FASTCALL call_1_7_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_9_5
#define init_1_7_9_5
void FASTCALL call_1_7_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_9_6
#define init_1_7_9_6
void FASTCALL call_1_7_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_9_7
#define init_1_7_9_7
void FASTCALL call_1_7_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_9_8
#define init_1_7_9_8
void FASTCALL call_1_7_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_9_9
#define init_1_7_9_9
void FASTCALL call_1_7_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_9_10
#define init_1_7_9_10
void FASTCALL call_1_7_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_9_11
#define init_1_7_9_11
void FASTCALL call_1_7_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_9_12
#define init_1_7_9_12
void FASTCALL call_1_7_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_9_13
#define init_1_7_9_13
void FASTCALL call_1_7_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_9_14
#define init_1_7_9_14
void FASTCALL call_1_7_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_9_15
#define init_1_7_9_15
void FASTCALL call_1_7_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_10_0
#define init_1_7_10_0
void FASTCALL call_1_7_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_10_1
#define init_1_7_10_1
void FASTCALL call_1_7_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_10_2
#define init_1_7_10_2
void FASTCALL call_1_7_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_10_3
#define init_1_7_10_3
void FASTCALL call_1_7_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_10_4
#define init_1_7_10_4
void FASTCALL call_1_7_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_10_5
#define init_1_7_10_5
void FASTCALL call_1_7_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_10_6
#define init_1_7_10_6
void FASTCALL call_1_7_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_10_7
#define init_1_7_10_7
void FASTCALL call_1_7_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_10_8
#define init_1_7_10_8
void FASTCALL call_1_7_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_10_9
#define init_1_7_10_9
void FASTCALL call_1_7_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_10_10
#define init_1_7_10_10
void FASTCALL call_1_7_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_10_11
#define init_1_7_10_11
void FASTCALL call_1_7_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_10_12
#define init_1_7_10_12
void FASTCALL call_1_7_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_10_13
#define init_1_7_10_13
void FASTCALL call_1_7_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_10_14
#define init_1_7_10_14
void FASTCALL call_1_7_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_10_15
#define init_1_7_10_15
void FASTCALL call_1_7_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_11_0
#define init_1_7_11_0
void FASTCALL call_1_7_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_11_1
#define init_1_7_11_1
void FASTCALL call_1_7_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_11_2
#define init_1_7_11_2
void FASTCALL call_1_7_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_11_3
#define init_1_7_11_3
void FASTCALL call_1_7_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_11_4
#define init_1_7_11_4
void FASTCALL call_1_7_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_11_5
#define init_1_7_11_5
void FASTCALL call_1_7_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_11_6
#define init_1_7_11_6
void FASTCALL call_1_7_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_11_7
#define init_1_7_11_7
void FASTCALL call_1_7_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_11_8
#define init_1_7_11_8
void FASTCALL call_1_7_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_11_9
#define init_1_7_11_9
void FASTCALL call_1_7_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_11_10
#define init_1_7_11_10
void FASTCALL call_1_7_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_11_11
#define init_1_7_11_11
void FASTCALL call_1_7_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_11_12
#define init_1_7_11_12
void FASTCALL call_1_7_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_11_13
#define init_1_7_11_13
void FASTCALL call_1_7_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_11_14
#define init_1_7_11_14
void FASTCALL call_1_7_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_11_15
#define init_1_7_11_15
void FASTCALL call_1_7_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_12_0
#define init_1_7_12_0
void FASTCALL call_1_7_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_12_1
#define init_1_7_12_1
void FASTCALL call_1_7_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_12_2
#define init_1_7_12_2
void FASTCALL call_1_7_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_12_3
#define init_1_7_12_3
void FASTCALL call_1_7_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_12_4
#define init_1_7_12_4
void FASTCALL call_1_7_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_12_5
#define init_1_7_12_5
void FASTCALL call_1_7_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_12_6
#define init_1_7_12_6
void FASTCALL call_1_7_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_12_7
#define init_1_7_12_7
void FASTCALL call_1_7_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_12_8
#define init_1_7_12_8
void FASTCALL call_1_7_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_12_9
#define init_1_7_12_9
void FASTCALL call_1_7_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_12_10
#define init_1_7_12_10
void FASTCALL call_1_7_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_12_11
#define init_1_7_12_11
void FASTCALL call_1_7_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_12_12
#define init_1_7_12_12
void FASTCALL call_1_7_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_12_13
#define init_1_7_12_13
void FASTCALL call_1_7_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_12_14
#define init_1_7_12_14
void FASTCALL call_1_7_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_12_15
#define init_1_7_12_15
void FASTCALL call_1_7_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_13_0
#define init_1_7_13_0
void FASTCALL call_1_7_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_13_1
#define init_1_7_13_1
void FASTCALL call_1_7_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_13_2
#define init_1_7_13_2
void FASTCALL call_1_7_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_13_3
#define init_1_7_13_3
void FASTCALL call_1_7_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_13_4
#define init_1_7_13_4
void FASTCALL call_1_7_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_13_5
#define init_1_7_13_5
void FASTCALL call_1_7_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_13_6
#define init_1_7_13_6
void FASTCALL call_1_7_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_13_7
#define init_1_7_13_7
void FASTCALL call_1_7_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_13_8
#define init_1_7_13_8
void FASTCALL call_1_7_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_13_9
#define init_1_7_13_9
void FASTCALL call_1_7_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_13_10
#define init_1_7_13_10
void FASTCALL call_1_7_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_13_11
#define init_1_7_13_11
void FASTCALL call_1_7_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_13_12
#define init_1_7_13_12
void FASTCALL call_1_7_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_13_13
#define init_1_7_13_13
void FASTCALL call_1_7_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_13_14
#define init_1_7_13_14
void FASTCALL call_1_7_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_13_15
#define init_1_7_13_15
void FASTCALL call_1_7_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_14_0
#define init_1_7_14_0
void FASTCALL call_1_7_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_14_1
#define init_1_7_14_1
void FASTCALL call_1_7_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_14_2
#define init_1_7_14_2
void FASTCALL call_1_7_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_14_3
#define init_1_7_14_3
void FASTCALL call_1_7_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_14_4
#define init_1_7_14_4
void FASTCALL call_1_7_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_14_5
#define init_1_7_14_5
void FASTCALL call_1_7_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_14_6
#define init_1_7_14_6
void FASTCALL call_1_7_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_14_7
#define init_1_7_14_7
void FASTCALL call_1_7_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_14_8
#define init_1_7_14_8
void FASTCALL call_1_7_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_14_9
#define init_1_7_14_9
void FASTCALL call_1_7_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_14_10
#define init_1_7_14_10
void FASTCALL call_1_7_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_14_11
#define init_1_7_14_11
void FASTCALL call_1_7_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_14_12
#define init_1_7_14_12
void FASTCALL call_1_7_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_14_13
#define init_1_7_14_13
void FASTCALL call_1_7_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_14_14
#define init_1_7_14_14
void FASTCALL call_1_7_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_14_15
#define init_1_7_14_15
void FASTCALL call_1_7_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_15_0
#define init_1_7_15_0
void FASTCALL call_1_7_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_15_1
#define init_1_7_15_1
void FASTCALL call_1_7_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_15_2
#define init_1_7_15_2
void FASTCALL call_1_7_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_15_3
#define init_1_7_15_3
void FASTCALL call_1_7_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_15_4
#define init_1_7_15_4
void FASTCALL call_1_7_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_15_5
#define init_1_7_15_5
void FASTCALL call_1_7_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_15_6
#define init_1_7_15_6
void FASTCALL call_1_7_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_15_7
#define init_1_7_15_7
void FASTCALL call_1_7_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_15_8
#define init_1_7_15_8
void FASTCALL call_1_7_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_15_9
#define init_1_7_15_9
void FASTCALL call_1_7_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_15_10
#define init_1_7_15_10
void FASTCALL call_1_7_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_15_11
#define init_1_7_15_11
void FASTCALL call_1_7_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_15_12
#define init_1_7_15_12
void FASTCALL call_1_7_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_15_13
#define init_1_7_15_13
void FASTCALL call_1_7_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_15_14
#define init_1_7_15_14
void FASTCALL call_1_7_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_7_15_15
#define init_1_7_15_15
void FASTCALL call_1_7_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_0_0
#define init_1_8_0_0
void FASTCALL call_1_8_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_0_1
#define init_1_8_0_1
void FASTCALL call_1_8_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_0_2
#define init_1_8_0_2
void FASTCALL call_1_8_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_0_3
#define init_1_8_0_3
void FASTCALL call_1_8_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_0_4
#define init_1_8_0_4
void FASTCALL call_1_8_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_0_5
#define init_1_8_0_5
void FASTCALL call_1_8_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_0_6
#define init_1_8_0_6
void FASTCALL call_1_8_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_0_7
#define init_1_8_0_7
void FASTCALL call_1_8_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_0_8
#define init_1_8_0_8
void FASTCALL call_1_8_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_0_9
#define init_1_8_0_9
void FASTCALL call_1_8_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_0_10
#define init_1_8_0_10
void FASTCALL call_1_8_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_0_11
#define init_1_8_0_11
void FASTCALL call_1_8_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_0_12
#define init_1_8_0_12
void FASTCALL call_1_8_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_0_13
#define init_1_8_0_13
void FASTCALL call_1_8_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_0_14
#define init_1_8_0_14
void FASTCALL call_1_8_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_0_15
#define init_1_8_0_15
void FASTCALL call_1_8_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_1_0
#define init_1_8_1_0
void FASTCALL call_1_8_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_1_1
#define init_1_8_1_1
void FASTCALL call_1_8_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_1_2
#define init_1_8_1_2
void FASTCALL call_1_8_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_1_3
#define init_1_8_1_3
void FASTCALL call_1_8_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_1_4
#define init_1_8_1_4
void FASTCALL call_1_8_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_1_5
#define init_1_8_1_5
void FASTCALL call_1_8_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_1_6
#define init_1_8_1_6
void FASTCALL call_1_8_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_1_7
#define init_1_8_1_7
void FASTCALL call_1_8_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_1_8
#define init_1_8_1_8
void FASTCALL call_1_8_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_1_9
#define init_1_8_1_9
void FASTCALL call_1_8_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_1_10
#define init_1_8_1_10
void FASTCALL call_1_8_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_1_11
#define init_1_8_1_11
void FASTCALL call_1_8_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_1_12
#define init_1_8_1_12
void FASTCALL call_1_8_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_1_13
#define init_1_8_1_13
void FASTCALL call_1_8_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_1_14
#define init_1_8_1_14
void FASTCALL call_1_8_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_1_15
#define init_1_8_1_15
void FASTCALL call_1_8_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_2_0
#define init_1_8_2_0
void FASTCALL call_1_8_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_2_1
#define init_1_8_2_1
void FASTCALL call_1_8_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_2_2
#define init_1_8_2_2
void FASTCALL call_1_8_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_2_3
#define init_1_8_2_3
void FASTCALL call_1_8_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_2_4
#define init_1_8_2_4
void FASTCALL call_1_8_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_2_5
#define init_1_8_2_5
void FASTCALL call_1_8_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_2_6
#define init_1_8_2_6
void FASTCALL call_1_8_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_2_7
#define init_1_8_2_7
void FASTCALL call_1_8_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_2_8
#define init_1_8_2_8
void FASTCALL call_1_8_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_2_9
#define init_1_8_2_9
void FASTCALL call_1_8_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_2_10
#define init_1_8_2_10
void FASTCALL call_1_8_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_2_11
#define init_1_8_2_11
void FASTCALL call_1_8_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_2_12
#define init_1_8_2_12
void FASTCALL call_1_8_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_2_13
#define init_1_8_2_13
void FASTCALL call_1_8_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_2_14
#define init_1_8_2_14
void FASTCALL call_1_8_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_2_15
#define init_1_8_2_15
void FASTCALL call_1_8_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_3_0
#define init_1_8_3_0
void FASTCALL call_1_8_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_3_1
#define init_1_8_3_1
void FASTCALL call_1_8_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_3_2
#define init_1_8_3_2
void FASTCALL call_1_8_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_3_3
#define init_1_8_3_3
void FASTCALL call_1_8_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_3_4
#define init_1_8_3_4
void FASTCALL call_1_8_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_3_5
#define init_1_8_3_5
void FASTCALL call_1_8_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_3_6
#define init_1_8_3_6
void FASTCALL call_1_8_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_3_7
#define init_1_8_3_7
void FASTCALL call_1_8_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_3_8
#define init_1_8_3_8
void FASTCALL call_1_8_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_3_9
#define init_1_8_3_9
void FASTCALL call_1_8_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_3_10
#define init_1_8_3_10
void FASTCALL call_1_8_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_3_11
#define init_1_8_3_11
void FASTCALL call_1_8_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_3_12
#define init_1_8_3_12
void FASTCALL call_1_8_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_3_13
#define init_1_8_3_13
void FASTCALL call_1_8_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_3_14
#define init_1_8_3_14
void FASTCALL call_1_8_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_3_15
#define init_1_8_3_15
void FASTCALL call_1_8_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_4_0
#define init_1_8_4_0
void FASTCALL call_1_8_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_4_1
#define init_1_8_4_1
void FASTCALL call_1_8_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_4_2
#define init_1_8_4_2
void FASTCALL call_1_8_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_4_3
#define init_1_8_4_3
void FASTCALL call_1_8_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_4_4
#define init_1_8_4_4
void FASTCALL call_1_8_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_4_5
#define init_1_8_4_5
void FASTCALL call_1_8_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_4_6
#define init_1_8_4_6
void FASTCALL call_1_8_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_4_7
#define init_1_8_4_7
void FASTCALL call_1_8_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_4_8
#define init_1_8_4_8
void FASTCALL call_1_8_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_4_9
#define init_1_8_4_9
void FASTCALL call_1_8_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_4_10
#define init_1_8_4_10
void FASTCALL call_1_8_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_4_11
#define init_1_8_4_11
void FASTCALL call_1_8_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_4_12
#define init_1_8_4_12
void FASTCALL call_1_8_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_4_13
#define init_1_8_4_13
void FASTCALL call_1_8_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_4_14
#define init_1_8_4_14
void FASTCALL call_1_8_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_4_15
#define init_1_8_4_15
void FASTCALL call_1_8_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_5_0
#define init_1_8_5_0
void FASTCALL call_1_8_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_5_1
#define init_1_8_5_1
void FASTCALL call_1_8_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_5_2
#define init_1_8_5_2
void FASTCALL call_1_8_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_5_3
#define init_1_8_5_3
void FASTCALL call_1_8_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_5_4
#define init_1_8_5_4
void FASTCALL call_1_8_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_5_5
#define init_1_8_5_5
void FASTCALL call_1_8_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_5_6
#define init_1_8_5_6
void FASTCALL call_1_8_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_5_7
#define init_1_8_5_7
void FASTCALL call_1_8_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_5_8
#define init_1_8_5_8
void FASTCALL call_1_8_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_5_9
#define init_1_8_5_9
void FASTCALL call_1_8_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_5_10
#define init_1_8_5_10
void FASTCALL call_1_8_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_5_11
#define init_1_8_5_11
void FASTCALL call_1_8_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_5_12
#define init_1_8_5_12
void FASTCALL call_1_8_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_5_13
#define init_1_8_5_13
void FASTCALL call_1_8_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_5_14
#define init_1_8_5_14
void FASTCALL call_1_8_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_5_15
#define init_1_8_5_15
void FASTCALL call_1_8_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_6_0
#define init_1_8_6_0
void FASTCALL call_1_8_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_6_1
#define init_1_8_6_1
void FASTCALL call_1_8_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_6_2
#define init_1_8_6_2
void FASTCALL call_1_8_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_6_3
#define init_1_8_6_3
void FASTCALL call_1_8_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_6_4
#define init_1_8_6_4
void FASTCALL call_1_8_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_6_5
#define init_1_8_6_5
void FASTCALL call_1_8_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_6_6
#define init_1_8_6_6
void FASTCALL call_1_8_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_6_7
#define init_1_8_6_7
void FASTCALL call_1_8_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_6_8
#define init_1_8_6_8
void FASTCALL call_1_8_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_6_9
#define init_1_8_6_9
void FASTCALL call_1_8_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_6_10
#define init_1_8_6_10
void FASTCALL call_1_8_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_6_11
#define init_1_8_6_11
void FASTCALL call_1_8_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_6_12
#define init_1_8_6_12
void FASTCALL call_1_8_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_6_13
#define init_1_8_6_13
void FASTCALL call_1_8_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_6_14
#define init_1_8_6_14
void FASTCALL call_1_8_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_6_15
#define init_1_8_6_15
void FASTCALL call_1_8_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_7_0
#define init_1_8_7_0
void FASTCALL call_1_8_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_7_1
#define init_1_8_7_1
void FASTCALL call_1_8_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_7_2
#define init_1_8_7_2
void FASTCALL call_1_8_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_7_3
#define init_1_8_7_3
void FASTCALL call_1_8_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_7_4
#define init_1_8_7_4
void FASTCALL call_1_8_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_7_5
#define init_1_8_7_5
void FASTCALL call_1_8_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_7_6
#define init_1_8_7_6
void FASTCALL call_1_8_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_7_7
#define init_1_8_7_7
void FASTCALL call_1_8_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_7_8
#define init_1_8_7_8
void FASTCALL call_1_8_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_7_9
#define init_1_8_7_9
void FASTCALL call_1_8_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_7_10
#define init_1_8_7_10
void FASTCALL call_1_8_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_7_11
#define init_1_8_7_11
void FASTCALL call_1_8_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_7_12
#define init_1_8_7_12
void FASTCALL call_1_8_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_7_13
#define init_1_8_7_13
void FASTCALL call_1_8_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_7_14
#define init_1_8_7_14
void FASTCALL call_1_8_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_7_15
#define init_1_8_7_15
void FASTCALL call_1_8_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_8_0
#define init_1_8_8_0
void FASTCALL call_1_8_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_8_1
#define init_1_8_8_1
void FASTCALL call_1_8_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_8_2
#define init_1_8_8_2
void FASTCALL call_1_8_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_8_3
#define init_1_8_8_3
void FASTCALL call_1_8_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_8_4
#define init_1_8_8_4
void FASTCALL call_1_8_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_8_5
#define init_1_8_8_5
void FASTCALL call_1_8_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_8_6
#define init_1_8_8_6
void FASTCALL call_1_8_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_8_7
#define init_1_8_8_7
void FASTCALL call_1_8_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_8_8
#define init_1_8_8_8
void FASTCALL call_1_8_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_8_9
#define init_1_8_8_9
void FASTCALL call_1_8_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_8_10
#define init_1_8_8_10
void FASTCALL call_1_8_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_8_11
#define init_1_8_8_11
void FASTCALL call_1_8_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_8_12
#define init_1_8_8_12
void FASTCALL call_1_8_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_8_13
#define init_1_8_8_13
void FASTCALL call_1_8_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_8_14
#define init_1_8_8_14
void FASTCALL call_1_8_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_8_15
#define init_1_8_8_15
void FASTCALL call_1_8_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_9_0
#define init_1_8_9_0
void FASTCALL call_1_8_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_9_1
#define init_1_8_9_1
void FASTCALL call_1_8_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_9_2
#define init_1_8_9_2
void FASTCALL call_1_8_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_9_3
#define init_1_8_9_3
void FASTCALL call_1_8_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_9_4
#define init_1_8_9_4
void FASTCALL call_1_8_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_9_5
#define init_1_8_9_5
void FASTCALL call_1_8_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_9_6
#define init_1_8_9_6
void FASTCALL call_1_8_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_9_7
#define init_1_8_9_7
void FASTCALL call_1_8_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_9_8
#define init_1_8_9_8
void FASTCALL call_1_8_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_9_9
#define init_1_8_9_9
void FASTCALL call_1_8_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_9_10
#define init_1_8_9_10
void FASTCALL call_1_8_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_9_11
#define init_1_8_9_11
void FASTCALL call_1_8_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_9_12
#define init_1_8_9_12
void FASTCALL call_1_8_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_9_13
#define init_1_8_9_13
void FASTCALL call_1_8_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_9_14
#define init_1_8_9_14
void FASTCALL call_1_8_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_9_15
#define init_1_8_9_15
void FASTCALL call_1_8_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_10_0
#define init_1_8_10_0
void FASTCALL call_1_8_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_10_1
#define init_1_8_10_1
void FASTCALL call_1_8_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_10_2
#define init_1_8_10_2
void FASTCALL call_1_8_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_10_3
#define init_1_8_10_3
void FASTCALL call_1_8_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_10_4
#define init_1_8_10_4
void FASTCALL call_1_8_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_10_5
#define init_1_8_10_5
void FASTCALL call_1_8_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_10_6
#define init_1_8_10_6
void FASTCALL call_1_8_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_10_7
#define init_1_8_10_7
void FASTCALL call_1_8_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_10_8
#define init_1_8_10_8
void FASTCALL call_1_8_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_10_9
#define init_1_8_10_9
void FASTCALL call_1_8_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_10_10
#define init_1_8_10_10
void FASTCALL call_1_8_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_10_11
#define init_1_8_10_11
void FASTCALL call_1_8_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_10_12
#define init_1_8_10_12
void FASTCALL call_1_8_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_10_13
#define init_1_8_10_13
void FASTCALL call_1_8_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_10_14
#define init_1_8_10_14
void FASTCALL call_1_8_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_10_15
#define init_1_8_10_15
void FASTCALL call_1_8_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_11_0
#define init_1_8_11_0
void FASTCALL call_1_8_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_11_1
#define init_1_8_11_1
void FASTCALL call_1_8_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_11_2
#define init_1_8_11_2
void FASTCALL call_1_8_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_11_3
#define init_1_8_11_3
void FASTCALL call_1_8_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_11_4
#define init_1_8_11_4
void FASTCALL call_1_8_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_11_5
#define init_1_8_11_5
void FASTCALL call_1_8_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_11_6
#define init_1_8_11_6
void FASTCALL call_1_8_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_11_7
#define init_1_8_11_7
void FASTCALL call_1_8_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_11_8
#define init_1_8_11_8
void FASTCALL call_1_8_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_11_9
#define init_1_8_11_9
void FASTCALL call_1_8_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_11_10
#define init_1_8_11_10
void FASTCALL call_1_8_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_11_11
#define init_1_8_11_11
void FASTCALL call_1_8_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_11_12
#define init_1_8_11_12
void FASTCALL call_1_8_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_11_13
#define init_1_8_11_13
void FASTCALL call_1_8_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_11_14
#define init_1_8_11_14
void FASTCALL call_1_8_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_11_15
#define init_1_8_11_15
void FASTCALL call_1_8_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_12_0
#define init_1_8_12_0
void FASTCALL call_1_8_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_12_1
#define init_1_8_12_1
void FASTCALL call_1_8_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_12_2
#define init_1_8_12_2
void FASTCALL call_1_8_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_12_3
#define init_1_8_12_3
void FASTCALL call_1_8_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_12_4
#define init_1_8_12_4
void FASTCALL call_1_8_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_12_5
#define init_1_8_12_5
void FASTCALL call_1_8_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_12_6
#define init_1_8_12_6
void FASTCALL call_1_8_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_12_7
#define init_1_8_12_7
void FASTCALL call_1_8_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_12_8
#define init_1_8_12_8
void FASTCALL call_1_8_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_12_9
#define init_1_8_12_9
void FASTCALL call_1_8_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_12_10
#define init_1_8_12_10
void FASTCALL call_1_8_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_12_11
#define init_1_8_12_11
void FASTCALL call_1_8_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_12_12
#define init_1_8_12_12
void FASTCALL call_1_8_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_12_13
#define init_1_8_12_13
void FASTCALL call_1_8_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_12_14
#define init_1_8_12_14
void FASTCALL call_1_8_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_12_15
#define init_1_8_12_15
void FASTCALL call_1_8_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_13_0
#define init_1_8_13_0
void FASTCALL call_1_8_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_13_1
#define init_1_8_13_1
void FASTCALL call_1_8_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_13_2
#define init_1_8_13_2
void FASTCALL call_1_8_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_13_3
#define init_1_8_13_3
void FASTCALL call_1_8_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_13_4
#define init_1_8_13_4
void FASTCALL call_1_8_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_13_5
#define init_1_8_13_5
void FASTCALL call_1_8_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_13_6
#define init_1_8_13_6
void FASTCALL call_1_8_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_13_7
#define init_1_8_13_7
void FASTCALL call_1_8_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_13_8
#define init_1_8_13_8
void FASTCALL call_1_8_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_13_9
#define init_1_8_13_9
void FASTCALL call_1_8_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_13_10
#define init_1_8_13_10
void FASTCALL call_1_8_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_13_11
#define init_1_8_13_11
void FASTCALL call_1_8_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_13_12
#define init_1_8_13_12
void FASTCALL call_1_8_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_13_13
#define init_1_8_13_13
void FASTCALL call_1_8_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_13_14
#define init_1_8_13_14
void FASTCALL call_1_8_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_13_15
#define init_1_8_13_15
void FASTCALL call_1_8_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_14_0
#define init_1_8_14_0
void FASTCALL call_1_8_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_14_1
#define init_1_8_14_1
void FASTCALL call_1_8_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_14_2
#define init_1_8_14_2
void FASTCALL call_1_8_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_14_3
#define init_1_8_14_3
void FASTCALL call_1_8_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_14_4
#define init_1_8_14_4
void FASTCALL call_1_8_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_14_5
#define init_1_8_14_5
void FASTCALL call_1_8_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_14_6
#define init_1_8_14_6
void FASTCALL call_1_8_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_14_7
#define init_1_8_14_7
void FASTCALL call_1_8_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_14_8
#define init_1_8_14_8
void FASTCALL call_1_8_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_14_9
#define init_1_8_14_9
void FASTCALL call_1_8_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_14_10
#define init_1_8_14_10
void FASTCALL call_1_8_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_14_11
#define init_1_8_14_11
void FASTCALL call_1_8_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_14_12
#define init_1_8_14_12
void FASTCALL call_1_8_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_14_13
#define init_1_8_14_13
void FASTCALL call_1_8_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_14_14
#define init_1_8_14_14
void FASTCALL call_1_8_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_14_15
#define init_1_8_14_15
void FASTCALL call_1_8_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_15_0
#define init_1_8_15_0
void FASTCALL call_1_8_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_15_1
#define init_1_8_15_1
void FASTCALL call_1_8_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_15_2
#define init_1_8_15_2
void FASTCALL call_1_8_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_15_3
#define init_1_8_15_3
void FASTCALL call_1_8_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_15_4
#define init_1_8_15_4
void FASTCALL call_1_8_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_15_5
#define init_1_8_15_5
void FASTCALL call_1_8_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_15_6
#define init_1_8_15_6
void FASTCALL call_1_8_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_15_7
#define init_1_8_15_7
void FASTCALL call_1_8_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_15_8
#define init_1_8_15_8
void FASTCALL call_1_8_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_15_9
#define init_1_8_15_9
void FASTCALL call_1_8_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_15_10
#define init_1_8_15_10
void FASTCALL call_1_8_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_15_11
#define init_1_8_15_11
void FASTCALL call_1_8_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_15_12
#define init_1_8_15_12
void FASTCALL call_1_8_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_15_13
#define init_1_8_15_13
void FASTCALL call_1_8_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_15_14
#define init_1_8_15_14
void FASTCALL call_1_8_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_8_15_15
#define init_1_8_15_15
void FASTCALL call_1_8_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_0_0
#define init_1_9_0_0
void FASTCALL call_1_9_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_0_1
#define init_1_9_0_1
void FASTCALL call_1_9_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_0_2
#define init_1_9_0_2
void FASTCALL call_1_9_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_0_3
#define init_1_9_0_3
void FASTCALL call_1_9_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_0_4
#define init_1_9_0_4
void FASTCALL call_1_9_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_0_5
#define init_1_9_0_5
void FASTCALL call_1_9_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_0_6
#define init_1_9_0_6
void FASTCALL call_1_9_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_0_7
#define init_1_9_0_7
void FASTCALL call_1_9_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_0_8
#define init_1_9_0_8
void FASTCALL call_1_9_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_0_9
#define init_1_9_0_9
void FASTCALL call_1_9_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_0_10
#define init_1_9_0_10
void FASTCALL call_1_9_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_0_11
#define init_1_9_0_11
void FASTCALL call_1_9_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_0_12
#define init_1_9_0_12
void FASTCALL call_1_9_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_0_13
#define init_1_9_0_13
void FASTCALL call_1_9_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_0_14
#define init_1_9_0_14
void FASTCALL call_1_9_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_0_15
#define init_1_9_0_15
void FASTCALL call_1_9_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_1_0
#define init_1_9_1_0
void FASTCALL call_1_9_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_1_1
#define init_1_9_1_1
void FASTCALL call_1_9_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_1_2
#define init_1_9_1_2
void FASTCALL call_1_9_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_1_3
#define init_1_9_1_3
void FASTCALL call_1_9_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_1_4
#define init_1_9_1_4
void FASTCALL call_1_9_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_1_5
#define init_1_9_1_5
void FASTCALL call_1_9_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_1_6
#define init_1_9_1_6
void FASTCALL call_1_9_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_1_7
#define init_1_9_1_7
void FASTCALL call_1_9_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_1_8
#define init_1_9_1_8
void FASTCALL call_1_9_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_1_9
#define init_1_9_1_9
void FASTCALL call_1_9_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_1_10
#define init_1_9_1_10
void FASTCALL call_1_9_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_1_11
#define init_1_9_1_11
void FASTCALL call_1_9_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_1_12
#define init_1_9_1_12
void FASTCALL call_1_9_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_1_13
#define init_1_9_1_13
void FASTCALL call_1_9_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_1_14
#define init_1_9_1_14
void FASTCALL call_1_9_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_1_15
#define init_1_9_1_15
void FASTCALL call_1_9_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_2_0
#define init_1_9_2_0
void FASTCALL call_1_9_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_2_1
#define init_1_9_2_1
void FASTCALL call_1_9_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_2_2
#define init_1_9_2_2
void FASTCALL call_1_9_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_2_3
#define init_1_9_2_3
void FASTCALL call_1_9_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_2_4
#define init_1_9_2_4
void FASTCALL call_1_9_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_2_5
#define init_1_9_2_5
void FASTCALL call_1_9_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_2_6
#define init_1_9_2_6
void FASTCALL call_1_9_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_2_7
#define init_1_9_2_7
void FASTCALL call_1_9_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_2_8
#define init_1_9_2_8
void FASTCALL call_1_9_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_2_9
#define init_1_9_2_9
void FASTCALL call_1_9_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_2_10
#define init_1_9_2_10
void FASTCALL call_1_9_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_2_11
#define init_1_9_2_11
void FASTCALL call_1_9_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_2_12
#define init_1_9_2_12
void FASTCALL call_1_9_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_2_13
#define init_1_9_2_13
void FASTCALL call_1_9_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_2_14
#define init_1_9_2_14
void FASTCALL call_1_9_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_2_15
#define init_1_9_2_15
void FASTCALL call_1_9_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_3_0
#define init_1_9_3_0
void FASTCALL call_1_9_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_3_1
#define init_1_9_3_1
void FASTCALL call_1_9_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_3_2
#define init_1_9_3_2
void FASTCALL call_1_9_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_3_3
#define init_1_9_3_3
void FASTCALL call_1_9_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_3_4
#define init_1_9_3_4
void FASTCALL call_1_9_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_3_5
#define init_1_9_3_5
void FASTCALL call_1_9_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_3_6
#define init_1_9_3_6
void FASTCALL call_1_9_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_3_7
#define init_1_9_3_7
void FASTCALL call_1_9_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_3_8
#define init_1_9_3_8
void FASTCALL call_1_9_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_3_9
#define init_1_9_3_9
void FASTCALL call_1_9_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_3_10
#define init_1_9_3_10
void FASTCALL call_1_9_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_3_11
#define init_1_9_3_11
void FASTCALL call_1_9_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_3_12
#define init_1_9_3_12
void FASTCALL call_1_9_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_3_13
#define init_1_9_3_13
void FASTCALL call_1_9_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_3_14
#define init_1_9_3_14
void FASTCALL call_1_9_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_3_15
#define init_1_9_3_15
void FASTCALL call_1_9_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_4_0
#define init_1_9_4_0
void FASTCALL call_1_9_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_4_1
#define init_1_9_4_1
void FASTCALL call_1_9_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_4_2
#define init_1_9_4_2
void FASTCALL call_1_9_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_4_3
#define init_1_9_4_3
void FASTCALL call_1_9_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_4_4
#define init_1_9_4_4
void FASTCALL call_1_9_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_4_5
#define init_1_9_4_5
void FASTCALL call_1_9_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_4_6
#define init_1_9_4_6
void FASTCALL call_1_9_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_4_7
#define init_1_9_4_7
void FASTCALL call_1_9_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_4_8
#define init_1_9_4_8
void FASTCALL call_1_9_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_4_9
#define init_1_9_4_9
void FASTCALL call_1_9_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_4_10
#define init_1_9_4_10
void FASTCALL call_1_9_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_4_11
#define init_1_9_4_11
void FASTCALL call_1_9_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_4_12
#define init_1_9_4_12
void FASTCALL call_1_9_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_4_13
#define init_1_9_4_13
void FASTCALL call_1_9_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_4_14
#define init_1_9_4_14
void FASTCALL call_1_9_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_4_15
#define init_1_9_4_15
void FASTCALL call_1_9_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_5_0
#define init_1_9_5_0
void FASTCALL call_1_9_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_5_1
#define init_1_9_5_1
void FASTCALL call_1_9_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_5_2
#define init_1_9_5_2
void FASTCALL call_1_9_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_5_3
#define init_1_9_5_3
void FASTCALL call_1_9_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_5_4
#define init_1_9_5_4
void FASTCALL call_1_9_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_5_5
#define init_1_9_5_5
void FASTCALL call_1_9_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_5_6
#define init_1_9_5_6
void FASTCALL call_1_9_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_5_7
#define init_1_9_5_7
void FASTCALL call_1_9_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_5_8
#define init_1_9_5_8
void FASTCALL call_1_9_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_5_9
#define init_1_9_5_9
void FASTCALL call_1_9_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_5_10
#define init_1_9_5_10
void FASTCALL call_1_9_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_5_11
#define init_1_9_5_11
void FASTCALL call_1_9_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_5_12
#define init_1_9_5_12
void FASTCALL call_1_9_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_5_13
#define init_1_9_5_13
void FASTCALL call_1_9_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_5_14
#define init_1_9_5_14
void FASTCALL call_1_9_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_5_15
#define init_1_9_5_15
void FASTCALL call_1_9_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_6_0
#define init_1_9_6_0
void FASTCALL call_1_9_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_6_1
#define init_1_9_6_1
void FASTCALL call_1_9_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_6_2
#define init_1_9_6_2
void FASTCALL call_1_9_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_6_3
#define init_1_9_6_3
void FASTCALL call_1_9_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_6_4
#define init_1_9_6_4
void FASTCALL call_1_9_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_6_5
#define init_1_9_6_5
void FASTCALL call_1_9_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_6_6
#define init_1_9_6_6
void FASTCALL call_1_9_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_6_7
#define init_1_9_6_7
void FASTCALL call_1_9_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_6_8
#define init_1_9_6_8
void FASTCALL call_1_9_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_6_9
#define init_1_9_6_9
void FASTCALL call_1_9_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_6_10
#define init_1_9_6_10
void FASTCALL call_1_9_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_6_11
#define init_1_9_6_11
void FASTCALL call_1_9_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_6_12
#define init_1_9_6_12
void FASTCALL call_1_9_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_6_13
#define init_1_9_6_13
void FASTCALL call_1_9_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_6_14
#define init_1_9_6_14
void FASTCALL call_1_9_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_6_15
#define init_1_9_6_15
void FASTCALL call_1_9_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_7_0
#define init_1_9_7_0
void FASTCALL call_1_9_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_7_1
#define init_1_9_7_1
void FASTCALL call_1_9_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_7_2
#define init_1_9_7_2
void FASTCALL call_1_9_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_7_3
#define init_1_9_7_3
void FASTCALL call_1_9_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_7_4
#define init_1_9_7_4
void FASTCALL call_1_9_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_7_5
#define init_1_9_7_5
void FASTCALL call_1_9_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_7_6
#define init_1_9_7_6
void FASTCALL call_1_9_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_7_7
#define init_1_9_7_7
void FASTCALL call_1_9_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_7_8
#define init_1_9_7_8
void FASTCALL call_1_9_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_7_9
#define init_1_9_7_9
void FASTCALL call_1_9_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_7_10
#define init_1_9_7_10
void FASTCALL call_1_9_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_7_11
#define init_1_9_7_11
void FASTCALL call_1_9_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_7_12
#define init_1_9_7_12
void FASTCALL call_1_9_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_7_13
#define init_1_9_7_13
void FASTCALL call_1_9_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_7_14
#define init_1_9_7_14
void FASTCALL call_1_9_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_7_15
#define init_1_9_7_15
void FASTCALL call_1_9_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_8_0
#define init_1_9_8_0
void FASTCALL call_1_9_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_8_1
#define init_1_9_8_1
void FASTCALL call_1_9_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_8_2
#define init_1_9_8_2
void FASTCALL call_1_9_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_8_3
#define init_1_9_8_3
void FASTCALL call_1_9_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_8_4
#define init_1_9_8_4
void FASTCALL call_1_9_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_8_5
#define init_1_9_8_5
void FASTCALL call_1_9_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_8_6
#define init_1_9_8_6
void FASTCALL call_1_9_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_8_7
#define init_1_9_8_7
void FASTCALL call_1_9_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_8_8
#define init_1_9_8_8
void FASTCALL call_1_9_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_8_9
#define init_1_9_8_9
void FASTCALL call_1_9_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_8_10
#define init_1_9_8_10
void FASTCALL call_1_9_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_8_11
#define init_1_9_8_11
void FASTCALL call_1_9_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_8_12
#define init_1_9_8_12
void FASTCALL call_1_9_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_8_13
#define init_1_9_8_13
void FASTCALL call_1_9_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_8_14
#define init_1_9_8_14
void FASTCALL call_1_9_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_8_15
#define init_1_9_8_15
void FASTCALL call_1_9_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_9_0
#define init_1_9_9_0
void FASTCALL call_1_9_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_9_1
#define init_1_9_9_1
void FASTCALL call_1_9_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_9_2
#define init_1_9_9_2
void FASTCALL call_1_9_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_9_3
#define init_1_9_9_3
void FASTCALL call_1_9_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_9_4
#define init_1_9_9_4
void FASTCALL call_1_9_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_9_5
#define init_1_9_9_5
void FASTCALL call_1_9_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_9_6
#define init_1_9_9_6
void FASTCALL call_1_9_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_9_7
#define init_1_9_9_7
void FASTCALL call_1_9_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_9_8
#define init_1_9_9_8
void FASTCALL call_1_9_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_9_9
#define init_1_9_9_9
void FASTCALL call_1_9_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_9_10
#define init_1_9_9_10
void FASTCALL call_1_9_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_9_11
#define init_1_9_9_11
void FASTCALL call_1_9_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_9_12
#define init_1_9_9_12
void FASTCALL call_1_9_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_9_13
#define init_1_9_9_13
void FASTCALL call_1_9_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_9_14
#define init_1_9_9_14
void FASTCALL call_1_9_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_9_15
#define init_1_9_9_15
void FASTCALL call_1_9_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_10_0
#define init_1_9_10_0
void FASTCALL call_1_9_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_10_1
#define init_1_9_10_1
void FASTCALL call_1_9_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_10_2
#define init_1_9_10_2
void FASTCALL call_1_9_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_10_3
#define init_1_9_10_3
void FASTCALL call_1_9_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_10_4
#define init_1_9_10_4
void FASTCALL call_1_9_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_10_5
#define init_1_9_10_5
void FASTCALL call_1_9_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_10_6
#define init_1_9_10_6
void FASTCALL call_1_9_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_10_7
#define init_1_9_10_7
void FASTCALL call_1_9_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_10_8
#define init_1_9_10_8
void FASTCALL call_1_9_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_10_9
#define init_1_9_10_9
void FASTCALL call_1_9_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_10_10
#define init_1_9_10_10
void FASTCALL call_1_9_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_10_11
#define init_1_9_10_11
void FASTCALL call_1_9_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_10_12
#define init_1_9_10_12
void FASTCALL call_1_9_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_10_13
#define init_1_9_10_13
void FASTCALL call_1_9_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_10_14
#define init_1_9_10_14
void FASTCALL call_1_9_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_10_15
#define init_1_9_10_15
void FASTCALL call_1_9_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_11_0
#define init_1_9_11_0
void FASTCALL call_1_9_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_11_1
#define init_1_9_11_1
void FASTCALL call_1_9_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_11_2
#define init_1_9_11_2
void FASTCALL call_1_9_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_11_3
#define init_1_9_11_3
void FASTCALL call_1_9_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_11_4
#define init_1_9_11_4
void FASTCALL call_1_9_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_11_5
#define init_1_9_11_5
void FASTCALL call_1_9_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_11_6
#define init_1_9_11_6
void FASTCALL call_1_9_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_11_7
#define init_1_9_11_7
void FASTCALL call_1_9_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_11_8
#define init_1_9_11_8
void FASTCALL call_1_9_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_11_9
#define init_1_9_11_9
void FASTCALL call_1_9_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_11_10
#define init_1_9_11_10
void FASTCALL call_1_9_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_11_11
#define init_1_9_11_11
void FASTCALL call_1_9_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_11_12
#define init_1_9_11_12
void FASTCALL call_1_9_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_11_13
#define init_1_9_11_13
void FASTCALL call_1_9_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_11_14
#define init_1_9_11_14
void FASTCALL call_1_9_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_11_15
#define init_1_9_11_15
void FASTCALL call_1_9_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_12_0
#define init_1_9_12_0
void FASTCALL call_1_9_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_12_1
#define init_1_9_12_1
void FASTCALL call_1_9_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_12_2
#define init_1_9_12_2
void FASTCALL call_1_9_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_12_3
#define init_1_9_12_3
void FASTCALL call_1_9_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_12_4
#define init_1_9_12_4
void FASTCALL call_1_9_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_12_5
#define init_1_9_12_5
void FASTCALL call_1_9_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_12_6
#define init_1_9_12_6
void FASTCALL call_1_9_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_12_7
#define init_1_9_12_7
void FASTCALL call_1_9_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_12_8
#define init_1_9_12_8
void FASTCALL call_1_9_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_12_9
#define init_1_9_12_9
void FASTCALL call_1_9_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_12_10
#define init_1_9_12_10
void FASTCALL call_1_9_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_12_11
#define init_1_9_12_11
void FASTCALL call_1_9_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_12_12
#define init_1_9_12_12
void FASTCALL call_1_9_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_12_13
#define init_1_9_12_13
void FASTCALL call_1_9_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_12_14
#define init_1_9_12_14
void FASTCALL call_1_9_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_12_15
#define init_1_9_12_15
void FASTCALL call_1_9_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_13_0
#define init_1_9_13_0
void FASTCALL call_1_9_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_13_1
#define init_1_9_13_1
void FASTCALL call_1_9_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_13_2
#define init_1_9_13_2
void FASTCALL call_1_9_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_13_3
#define init_1_9_13_3
void FASTCALL call_1_9_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_13_4
#define init_1_9_13_4
void FASTCALL call_1_9_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_13_5
#define init_1_9_13_5
void FASTCALL call_1_9_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_13_6
#define init_1_9_13_6
void FASTCALL call_1_9_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_13_7
#define init_1_9_13_7
void FASTCALL call_1_9_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_13_8
#define init_1_9_13_8
void FASTCALL call_1_9_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_13_9
#define init_1_9_13_9
void FASTCALL call_1_9_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_13_10
#define init_1_9_13_10
void FASTCALL call_1_9_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_13_11
#define init_1_9_13_11
void FASTCALL call_1_9_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_13_12
#define init_1_9_13_12
void FASTCALL call_1_9_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_13_13
#define init_1_9_13_13
void FASTCALL call_1_9_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_13_14
#define init_1_9_13_14
void FASTCALL call_1_9_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_13_15
#define init_1_9_13_15
void FASTCALL call_1_9_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_14_0
#define init_1_9_14_0
void FASTCALL call_1_9_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_14_1
#define init_1_9_14_1
void FASTCALL call_1_9_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_14_2
#define init_1_9_14_2
void FASTCALL call_1_9_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_14_3
#define init_1_9_14_3
void FASTCALL call_1_9_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_14_4
#define init_1_9_14_4
void FASTCALL call_1_9_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_14_5
#define init_1_9_14_5
void FASTCALL call_1_9_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_14_6
#define init_1_9_14_6
void FASTCALL call_1_9_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_14_7
#define init_1_9_14_7
void FASTCALL call_1_9_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_14_8
#define init_1_9_14_8
void FASTCALL call_1_9_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_14_9
#define init_1_9_14_9
void FASTCALL call_1_9_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_14_10
#define init_1_9_14_10
void FASTCALL call_1_9_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_14_11
#define init_1_9_14_11
void FASTCALL call_1_9_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_14_12
#define init_1_9_14_12
void FASTCALL call_1_9_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_14_13
#define init_1_9_14_13
void FASTCALL call_1_9_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_14_14
#define init_1_9_14_14
void FASTCALL call_1_9_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_14_15
#define init_1_9_14_15
void FASTCALL call_1_9_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_15_0
#define init_1_9_15_0
void FASTCALL call_1_9_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_15_1
#define init_1_9_15_1
void FASTCALL call_1_9_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_15_2
#define init_1_9_15_2
void FASTCALL call_1_9_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_15_3
#define init_1_9_15_3
void FASTCALL call_1_9_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_15_4
#define init_1_9_15_4
void FASTCALL call_1_9_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_15_5
#define init_1_9_15_5
void FASTCALL call_1_9_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_15_6
#define init_1_9_15_6
void FASTCALL call_1_9_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_15_7
#define init_1_9_15_7
void FASTCALL call_1_9_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_15_8
#define init_1_9_15_8
void FASTCALL call_1_9_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_15_9
#define init_1_9_15_9
void FASTCALL call_1_9_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_15_10
#define init_1_9_15_10
void FASTCALL call_1_9_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_15_11
#define init_1_9_15_11
void FASTCALL call_1_9_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_15_12
#define init_1_9_15_12
void FASTCALL call_1_9_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_15_13
#define init_1_9_15_13
void FASTCALL call_1_9_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_15_14
#define init_1_9_15_14
void FASTCALL call_1_9_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_9_15_15
#define init_1_9_15_15
void FASTCALL call_1_9_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_0_0
#define init_1_10_0_0
void FASTCALL call_1_10_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_0_1
#define init_1_10_0_1
void FASTCALL call_1_10_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_0_2
#define init_1_10_0_2
void FASTCALL call_1_10_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_0_3
#define init_1_10_0_3
void FASTCALL call_1_10_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_0_4
#define init_1_10_0_4
void FASTCALL call_1_10_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_0_5
#define init_1_10_0_5
void FASTCALL call_1_10_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_0_6
#define init_1_10_0_6
void FASTCALL call_1_10_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_0_7
#define init_1_10_0_7
void FASTCALL call_1_10_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_0_8
#define init_1_10_0_8
void FASTCALL call_1_10_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_0_9
#define init_1_10_0_9
void FASTCALL call_1_10_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_0_10
#define init_1_10_0_10
void FASTCALL call_1_10_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_0_11
#define init_1_10_0_11
void FASTCALL call_1_10_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_0_12
#define init_1_10_0_12
void FASTCALL call_1_10_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_0_13
#define init_1_10_0_13
void FASTCALL call_1_10_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_0_14
#define init_1_10_0_14
void FASTCALL call_1_10_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_0_15
#define init_1_10_0_15
void FASTCALL call_1_10_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_1_0
#define init_1_10_1_0
void FASTCALL call_1_10_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_1_1
#define init_1_10_1_1
void FASTCALL call_1_10_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_1_2
#define init_1_10_1_2
void FASTCALL call_1_10_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_1_3
#define init_1_10_1_3
void FASTCALL call_1_10_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_1_4
#define init_1_10_1_4
void FASTCALL call_1_10_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_1_5
#define init_1_10_1_5
void FASTCALL call_1_10_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_1_6
#define init_1_10_1_6
void FASTCALL call_1_10_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_1_7
#define init_1_10_1_7
void FASTCALL call_1_10_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_1_8
#define init_1_10_1_8
void FASTCALL call_1_10_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_1_9
#define init_1_10_1_9
void FASTCALL call_1_10_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_1_10
#define init_1_10_1_10
void FASTCALL call_1_10_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_1_11
#define init_1_10_1_11
void FASTCALL call_1_10_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_1_12
#define init_1_10_1_12
void FASTCALL call_1_10_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_1_13
#define init_1_10_1_13
void FASTCALL call_1_10_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_1_14
#define init_1_10_1_14
void FASTCALL call_1_10_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_1_15
#define init_1_10_1_15
void FASTCALL call_1_10_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_2_0
#define init_1_10_2_0
void FASTCALL call_1_10_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_2_1
#define init_1_10_2_1
void FASTCALL call_1_10_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_2_2
#define init_1_10_2_2
void FASTCALL call_1_10_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_2_3
#define init_1_10_2_3
void FASTCALL call_1_10_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_2_4
#define init_1_10_2_4
void FASTCALL call_1_10_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_2_5
#define init_1_10_2_5
void FASTCALL call_1_10_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_2_6
#define init_1_10_2_6
void FASTCALL call_1_10_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_2_7
#define init_1_10_2_7
void FASTCALL call_1_10_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_2_8
#define init_1_10_2_8
void FASTCALL call_1_10_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_2_9
#define init_1_10_2_9
void FASTCALL call_1_10_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_2_10
#define init_1_10_2_10
void FASTCALL call_1_10_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_2_11
#define init_1_10_2_11
void FASTCALL call_1_10_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_2_12
#define init_1_10_2_12
void FASTCALL call_1_10_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_2_13
#define init_1_10_2_13
void FASTCALL call_1_10_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_2_14
#define init_1_10_2_14
void FASTCALL call_1_10_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_2_15
#define init_1_10_2_15
void FASTCALL call_1_10_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_3_0
#define init_1_10_3_0
void FASTCALL call_1_10_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_3_1
#define init_1_10_3_1
void FASTCALL call_1_10_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_3_2
#define init_1_10_3_2
void FASTCALL call_1_10_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_3_3
#define init_1_10_3_3
void FASTCALL call_1_10_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_3_4
#define init_1_10_3_4
void FASTCALL call_1_10_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_3_5
#define init_1_10_3_5
void FASTCALL call_1_10_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_3_6
#define init_1_10_3_6
void FASTCALL call_1_10_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_3_7
#define init_1_10_3_7
void FASTCALL call_1_10_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_3_8
#define init_1_10_3_8
void FASTCALL call_1_10_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_3_9
#define init_1_10_3_9
void FASTCALL call_1_10_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_3_10
#define init_1_10_3_10
void FASTCALL call_1_10_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_3_11
#define init_1_10_3_11
void FASTCALL call_1_10_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_3_12
#define init_1_10_3_12
void FASTCALL call_1_10_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_3_13
#define init_1_10_3_13
void FASTCALL call_1_10_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_3_14
#define init_1_10_3_14
void FASTCALL call_1_10_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_3_15
#define init_1_10_3_15
void FASTCALL call_1_10_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_4_0
#define init_1_10_4_0
void FASTCALL call_1_10_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_4_1
#define init_1_10_4_1
void FASTCALL call_1_10_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_4_2
#define init_1_10_4_2
void FASTCALL call_1_10_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_4_3
#define init_1_10_4_3
void FASTCALL call_1_10_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_4_4
#define init_1_10_4_4
void FASTCALL call_1_10_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_4_5
#define init_1_10_4_5
void FASTCALL call_1_10_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_4_6
#define init_1_10_4_6
void FASTCALL call_1_10_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_4_7
#define init_1_10_4_7
void FASTCALL call_1_10_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_4_8
#define init_1_10_4_8
void FASTCALL call_1_10_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_4_9
#define init_1_10_4_9
void FASTCALL call_1_10_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_4_10
#define init_1_10_4_10
void FASTCALL call_1_10_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_4_11
#define init_1_10_4_11
void FASTCALL call_1_10_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_4_12
#define init_1_10_4_12
void FASTCALL call_1_10_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_4_13
#define init_1_10_4_13
void FASTCALL call_1_10_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_4_14
#define init_1_10_4_14
void FASTCALL call_1_10_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_4_15
#define init_1_10_4_15
void FASTCALL call_1_10_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_5_0
#define init_1_10_5_0
void FASTCALL call_1_10_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_5_1
#define init_1_10_5_1
void FASTCALL call_1_10_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_5_2
#define init_1_10_5_2
void FASTCALL call_1_10_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_5_3
#define init_1_10_5_3
void FASTCALL call_1_10_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_5_4
#define init_1_10_5_4
void FASTCALL call_1_10_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_5_5
#define init_1_10_5_5
void FASTCALL call_1_10_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_5_6
#define init_1_10_5_6
void FASTCALL call_1_10_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_5_7
#define init_1_10_5_7
void FASTCALL call_1_10_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_5_8
#define init_1_10_5_8
void FASTCALL call_1_10_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_5_9
#define init_1_10_5_9
void FASTCALL call_1_10_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_5_10
#define init_1_10_5_10
void FASTCALL call_1_10_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_5_11
#define init_1_10_5_11
void FASTCALL call_1_10_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_5_12
#define init_1_10_5_12
void FASTCALL call_1_10_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_5_13
#define init_1_10_5_13
void FASTCALL call_1_10_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_5_14
#define init_1_10_5_14
void FASTCALL call_1_10_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_5_15
#define init_1_10_5_15
void FASTCALL call_1_10_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_6_0
#define init_1_10_6_0
void FASTCALL call_1_10_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_6_1
#define init_1_10_6_1
void FASTCALL call_1_10_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_6_2
#define init_1_10_6_2
void FASTCALL call_1_10_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_6_3
#define init_1_10_6_3
void FASTCALL call_1_10_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_6_4
#define init_1_10_6_4
void FASTCALL call_1_10_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_6_5
#define init_1_10_6_5
void FASTCALL call_1_10_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_6_6
#define init_1_10_6_6
void FASTCALL call_1_10_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_6_7
#define init_1_10_6_7
void FASTCALL call_1_10_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_6_8
#define init_1_10_6_8
void FASTCALL call_1_10_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_6_9
#define init_1_10_6_9
void FASTCALL call_1_10_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_6_10
#define init_1_10_6_10
void FASTCALL call_1_10_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_6_11
#define init_1_10_6_11
void FASTCALL call_1_10_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_6_12
#define init_1_10_6_12
void FASTCALL call_1_10_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_6_13
#define init_1_10_6_13
void FASTCALL call_1_10_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_6_14
#define init_1_10_6_14
void FASTCALL call_1_10_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_6_15
#define init_1_10_6_15
void FASTCALL call_1_10_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_7_0
#define init_1_10_7_0
void FASTCALL call_1_10_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_7_1
#define init_1_10_7_1
void FASTCALL call_1_10_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_7_2
#define init_1_10_7_2
void FASTCALL call_1_10_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_7_3
#define init_1_10_7_3
void FASTCALL call_1_10_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_7_4
#define init_1_10_7_4
void FASTCALL call_1_10_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_7_5
#define init_1_10_7_5
void FASTCALL call_1_10_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_7_6
#define init_1_10_7_6
void FASTCALL call_1_10_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_7_7
#define init_1_10_7_7
void FASTCALL call_1_10_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_7_8
#define init_1_10_7_8
void FASTCALL call_1_10_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_7_9
#define init_1_10_7_9
void FASTCALL call_1_10_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_7_10
#define init_1_10_7_10
void FASTCALL call_1_10_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_7_11
#define init_1_10_7_11
void FASTCALL call_1_10_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_7_12
#define init_1_10_7_12
void FASTCALL call_1_10_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_7_13
#define init_1_10_7_13
void FASTCALL call_1_10_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_7_14
#define init_1_10_7_14
void FASTCALL call_1_10_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_7_15
#define init_1_10_7_15
void FASTCALL call_1_10_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_8_0
#define init_1_10_8_0
void FASTCALL call_1_10_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_8_1
#define init_1_10_8_1
void FASTCALL call_1_10_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_8_2
#define init_1_10_8_2
void FASTCALL call_1_10_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_8_3
#define init_1_10_8_3
void FASTCALL call_1_10_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_8_4
#define init_1_10_8_4
void FASTCALL call_1_10_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_8_5
#define init_1_10_8_5
void FASTCALL call_1_10_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_8_6
#define init_1_10_8_6
void FASTCALL call_1_10_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_8_7
#define init_1_10_8_7
void FASTCALL call_1_10_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_8_8
#define init_1_10_8_8
void FASTCALL call_1_10_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_8_9
#define init_1_10_8_9
void FASTCALL call_1_10_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_8_10
#define init_1_10_8_10
void FASTCALL call_1_10_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_8_11
#define init_1_10_8_11
void FASTCALL call_1_10_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_8_12
#define init_1_10_8_12
void FASTCALL call_1_10_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_8_13
#define init_1_10_8_13
void FASTCALL call_1_10_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_8_14
#define init_1_10_8_14
void FASTCALL call_1_10_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_8_15
#define init_1_10_8_15
void FASTCALL call_1_10_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_9_0
#define init_1_10_9_0
void FASTCALL call_1_10_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_9_1
#define init_1_10_9_1
void FASTCALL call_1_10_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_9_2
#define init_1_10_9_2
void FASTCALL call_1_10_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_9_3
#define init_1_10_9_3
void FASTCALL call_1_10_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_9_4
#define init_1_10_9_4
void FASTCALL call_1_10_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_9_5
#define init_1_10_9_5
void FASTCALL call_1_10_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_9_6
#define init_1_10_9_6
void FASTCALL call_1_10_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_9_7
#define init_1_10_9_7
void FASTCALL call_1_10_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_9_8
#define init_1_10_9_8
void FASTCALL call_1_10_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_9_9
#define init_1_10_9_9
void FASTCALL call_1_10_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_9_10
#define init_1_10_9_10
void FASTCALL call_1_10_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_9_11
#define init_1_10_9_11
void FASTCALL call_1_10_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_9_12
#define init_1_10_9_12
void FASTCALL call_1_10_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_9_13
#define init_1_10_9_13
void FASTCALL call_1_10_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_9_14
#define init_1_10_9_14
void FASTCALL call_1_10_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_9_15
#define init_1_10_9_15
void FASTCALL call_1_10_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_10_0
#define init_1_10_10_0
void FASTCALL call_1_10_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_10_1
#define init_1_10_10_1
void FASTCALL call_1_10_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_10_2
#define init_1_10_10_2
void FASTCALL call_1_10_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_10_3
#define init_1_10_10_3
void FASTCALL call_1_10_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_10_4
#define init_1_10_10_4
void FASTCALL call_1_10_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_10_5
#define init_1_10_10_5
void FASTCALL call_1_10_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_10_6
#define init_1_10_10_6
void FASTCALL call_1_10_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_10_7
#define init_1_10_10_7
void FASTCALL call_1_10_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_10_8
#define init_1_10_10_8
void FASTCALL call_1_10_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_10_9
#define init_1_10_10_9
void FASTCALL call_1_10_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_10_10
#define init_1_10_10_10
void FASTCALL call_1_10_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_10_11
#define init_1_10_10_11
void FASTCALL call_1_10_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_10_12
#define init_1_10_10_12
void FASTCALL call_1_10_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_10_13
#define init_1_10_10_13
void FASTCALL call_1_10_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_10_14
#define init_1_10_10_14
void FASTCALL call_1_10_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_10_15
#define init_1_10_10_15
void FASTCALL call_1_10_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_11_0
#define init_1_10_11_0
void FASTCALL call_1_10_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_11_1
#define init_1_10_11_1
void FASTCALL call_1_10_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_11_2
#define init_1_10_11_2
void FASTCALL call_1_10_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_11_3
#define init_1_10_11_3
void FASTCALL call_1_10_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_11_4
#define init_1_10_11_4
void FASTCALL call_1_10_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_11_5
#define init_1_10_11_5
void FASTCALL call_1_10_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_11_6
#define init_1_10_11_6
void FASTCALL call_1_10_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_11_7
#define init_1_10_11_7
void FASTCALL call_1_10_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_11_8
#define init_1_10_11_8
void FASTCALL call_1_10_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_11_9
#define init_1_10_11_9
void FASTCALL call_1_10_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_11_10
#define init_1_10_11_10
void FASTCALL call_1_10_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_11_11
#define init_1_10_11_11
void FASTCALL call_1_10_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_11_12
#define init_1_10_11_12
void FASTCALL call_1_10_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_11_13
#define init_1_10_11_13
void FASTCALL call_1_10_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_11_14
#define init_1_10_11_14
void FASTCALL call_1_10_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_11_15
#define init_1_10_11_15
void FASTCALL call_1_10_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_12_0
#define init_1_10_12_0
void FASTCALL call_1_10_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_12_1
#define init_1_10_12_1
void FASTCALL call_1_10_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_12_2
#define init_1_10_12_2
void FASTCALL call_1_10_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_12_3
#define init_1_10_12_3
void FASTCALL call_1_10_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_12_4
#define init_1_10_12_4
void FASTCALL call_1_10_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_12_5
#define init_1_10_12_5
void FASTCALL call_1_10_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_12_6
#define init_1_10_12_6
void FASTCALL call_1_10_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_12_7
#define init_1_10_12_7
void FASTCALL call_1_10_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_12_8
#define init_1_10_12_8
void FASTCALL call_1_10_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_12_9
#define init_1_10_12_9
void FASTCALL call_1_10_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_12_10
#define init_1_10_12_10
void FASTCALL call_1_10_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_12_11
#define init_1_10_12_11
void FASTCALL call_1_10_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_12_12
#define init_1_10_12_12
void FASTCALL call_1_10_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_12_13
#define init_1_10_12_13
void FASTCALL call_1_10_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_12_14
#define init_1_10_12_14
void FASTCALL call_1_10_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_12_15
#define init_1_10_12_15
void FASTCALL call_1_10_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_13_0
#define init_1_10_13_0
void FASTCALL call_1_10_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_13_1
#define init_1_10_13_1
void FASTCALL call_1_10_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_13_2
#define init_1_10_13_2
void FASTCALL call_1_10_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_13_3
#define init_1_10_13_3
void FASTCALL call_1_10_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_13_4
#define init_1_10_13_4
void FASTCALL call_1_10_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_13_5
#define init_1_10_13_5
void FASTCALL call_1_10_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_13_6
#define init_1_10_13_6
void FASTCALL call_1_10_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_13_7
#define init_1_10_13_7
void FASTCALL call_1_10_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_13_8
#define init_1_10_13_8
void FASTCALL call_1_10_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_13_9
#define init_1_10_13_9
void FASTCALL call_1_10_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_13_10
#define init_1_10_13_10
void FASTCALL call_1_10_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_13_11
#define init_1_10_13_11
void FASTCALL call_1_10_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_13_12
#define init_1_10_13_12
void FASTCALL call_1_10_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_13_13
#define init_1_10_13_13
void FASTCALL call_1_10_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_13_14
#define init_1_10_13_14
void FASTCALL call_1_10_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_13_15
#define init_1_10_13_15
void FASTCALL call_1_10_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_14_0
#define init_1_10_14_0
void FASTCALL call_1_10_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_14_1
#define init_1_10_14_1
void FASTCALL call_1_10_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_14_2
#define init_1_10_14_2
void FASTCALL call_1_10_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_14_3
#define init_1_10_14_3
void FASTCALL call_1_10_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_14_4
#define init_1_10_14_4
void FASTCALL call_1_10_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_14_5
#define init_1_10_14_5
void FASTCALL call_1_10_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_14_6
#define init_1_10_14_6
void FASTCALL call_1_10_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_14_7
#define init_1_10_14_7
void FASTCALL call_1_10_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_14_8
#define init_1_10_14_8
void FASTCALL call_1_10_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_14_9
#define init_1_10_14_9
void FASTCALL call_1_10_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_14_10
#define init_1_10_14_10
void FASTCALL call_1_10_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_14_11
#define init_1_10_14_11
void FASTCALL call_1_10_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_14_12
#define init_1_10_14_12
void FASTCALL call_1_10_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_14_13
#define init_1_10_14_13
void FASTCALL call_1_10_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_14_14
#define init_1_10_14_14
void FASTCALL call_1_10_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_14_15
#define init_1_10_14_15
void FASTCALL call_1_10_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_15_0
#define init_1_10_15_0
void FASTCALL call_1_10_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_15_1
#define init_1_10_15_1
void FASTCALL call_1_10_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_15_2
#define init_1_10_15_2
void FASTCALL call_1_10_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_15_3
#define init_1_10_15_3
void FASTCALL call_1_10_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_15_4
#define init_1_10_15_4
void FASTCALL call_1_10_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_15_5
#define init_1_10_15_5
void FASTCALL call_1_10_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_15_6
#define init_1_10_15_6
void FASTCALL call_1_10_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_15_7
#define init_1_10_15_7
void FASTCALL call_1_10_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_15_8
#define init_1_10_15_8
void FASTCALL call_1_10_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_15_9
#define init_1_10_15_9
void FASTCALL call_1_10_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_15_10
#define init_1_10_15_10
void FASTCALL call_1_10_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_15_11
#define init_1_10_15_11
void FASTCALL call_1_10_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_15_12
#define init_1_10_15_12
void FASTCALL call_1_10_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_15_13
#define init_1_10_15_13
void FASTCALL call_1_10_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_15_14
#define init_1_10_15_14
void FASTCALL call_1_10_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_10_15_15
#define init_1_10_15_15
void FASTCALL call_1_10_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_0_0
#define init_1_11_0_0
void FASTCALL call_1_11_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_0_1
#define init_1_11_0_1
void FASTCALL call_1_11_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_0_2
#define init_1_11_0_2
void FASTCALL call_1_11_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_0_3
#define init_1_11_0_3
void FASTCALL call_1_11_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_0_4
#define init_1_11_0_4
void FASTCALL call_1_11_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_0_5
#define init_1_11_0_5
void FASTCALL call_1_11_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_0_6
#define init_1_11_0_6
void FASTCALL call_1_11_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_0_7
#define init_1_11_0_7
void FASTCALL call_1_11_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_0_8
#define init_1_11_0_8
void FASTCALL call_1_11_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_0_9
#define init_1_11_0_9
void FASTCALL call_1_11_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_0_10
#define init_1_11_0_10
void FASTCALL call_1_11_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_0_11
#define init_1_11_0_11
void FASTCALL call_1_11_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_0_12
#define init_1_11_0_12
void FASTCALL call_1_11_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_0_13
#define init_1_11_0_13
void FASTCALL call_1_11_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_0_14
#define init_1_11_0_14
void FASTCALL call_1_11_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_0_15
#define init_1_11_0_15
void FASTCALL call_1_11_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_1_0
#define init_1_11_1_0
void FASTCALL call_1_11_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_1_1
#define init_1_11_1_1
void FASTCALL call_1_11_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_1_2
#define init_1_11_1_2
void FASTCALL call_1_11_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_1_3
#define init_1_11_1_3
void FASTCALL call_1_11_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_1_4
#define init_1_11_1_4
void FASTCALL call_1_11_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_1_5
#define init_1_11_1_5
void FASTCALL call_1_11_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_1_6
#define init_1_11_1_6
void FASTCALL call_1_11_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_1_7
#define init_1_11_1_7
void FASTCALL call_1_11_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_1_8
#define init_1_11_1_8
void FASTCALL call_1_11_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_1_9
#define init_1_11_1_9
void FASTCALL call_1_11_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_1_10
#define init_1_11_1_10
void FASTCALL call_1_11_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_1_11
#define init_1_11_1_11
void FASTCALL call_1_11_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_1_12
#define init_1_11_1_12
void FASTCALL call_1_11_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_1_13
#define init_1_11_1_13
void FASTCALL call_1_11_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_1_14
#define init_1_11_1_14
void FASTCALL call_1_11_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_1_15
#define init_1_11_1_15
void FASTCALL call_1_11_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_2_0
#define init_1_11_2_0
void FASTCALL call_1_11_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_2_1
#define init_1_11_2_1
void FASTCALL call_1_11_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_2_2
#define init_1_11_2_2
void FASTCALL call_1_11_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_2_3
#define init_1_11_2_3
void FASTCALL call_1_11_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_2_4
#define init_1_11_2_4
void FASTCALL call_1_11_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_2_5
#define init_1_11_2_5
void FASTCALL call_1_11_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_2_6
#define init_1_11_2_6
void FASTCALL call_1_11_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_2_7
#define init_1_11_2_7
void FASTCALL call_1_11_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_2_8
#define init_1_11_2_8
void FASTCALL call_1_11_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_2_9
#define init_1_11_2_9
void FASTCALL call_1_11_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_2_10
#define init_1_11_2_10
void FASTCALL call_1_11_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_2_11
#define init_1_11_2_11
void FASTCALL call_1_11_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_2_12
#define init_1_11_2_12
void FASTCALL call_1_11_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_2_13
#define init_1_11_2_13
void FASTCALL call_1_11_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_2_14
#define init_1_11_2_14
void FASTCALL call_1_11_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_2_15
#define init_1_11_2_15
void FASTCALL call_1_11_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_3_0
#define init_1_11_3_0
void FASTCALL call_1_11_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_3_1
#define init_1_11_3_1
void FASTCALL call_1_11_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_3_2
#define init_1_11_3_2
void FASTCALL call_1_11_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_3_3
#define init_1_11_3_3
void FASTCALL call_1_11_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_3_4
#define init_1_11_3_4
void FASTCALL call_1_11_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_3_5
#define init_1_11_3_5
void FASTCALL call_1_11_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_3_6
#define init_1_11_3_6
void FASTCALL call_1_11_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_3_7
#define init_1_11_3_7
void FASTCALL call_1_11_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_3_8
#define init_1_11_3_8
void FASTCALL call_1_11_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_3_9
#define init_1_11_3_9
void FASTCALL call_1_11_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_3_10
#define init_1_11_3_10
void FASTCALL call_1_11_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_3_11
#define init_1_11_3_11
void FASTCALL call_1_11_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_3_12
#define init_1_11_3_12
void FASTCALL call_1_11_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_3_13
#define init_1_11_3_13
void FASTCALL call_1_11_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_3_14
#define init_1_11_3_14
void FASTCALL call_1_11_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_3_15
#define init_1_11_3_15
void FASTCALL call_1_11_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_4_0
#define init_1_11_4_0
void FASTCALL call_1_11_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_4_1
#define init_1_11_4_1
void FASTCALL call_1_11_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_4_2
#define init_1_11_4_2
void FASTCALL call_1_11_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_4_3
#define init_1_11_4_3
void FASTCALL call_1_11_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_4_4
#define init_1_11_4_4
void FASTCALL call_1_11_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_4_5
#define init_1_11_4_5
void FASTCALL call_1_11_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_4_6
#define init_1_11_4_6
void FASTCALL call_1_11_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_4_7
#define init_1_11_4_7
void FASTCALL call_1_11_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_4_8
#define init_1_11_4_8
void FASTCALL call_1_11_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_4_9
#define init_1_11_4_9
void FASTCALL call_1_11_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_4_10
#define init_1_11_4_10
void FASTCALL call_1_11_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_4_11
#define init_1_11_4_11
void FASTCALL call_1_11_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_4_12
#define init_1_11_4_12
void FASTCALL call_1_11_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_4_13
#define init_1_11_4_13
void FASTCALL call_1_11_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_4_14
#define init_1_11_4_14
void FASTCALL call_1_11_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_4_15
#define init_1_11_4_15
void FASTCALL call_1_11_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_5_0
#define init_1_11_5_0
void FASTCALL call_1_11_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_5_1
#define init_1_11_5_1
void FASTCALL call_1_11_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_5_2
#define init_1_11_5_2
void FASTCALL call_1_11_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_5_3
#define init_1_11_5_3
void FASTCALL call_1_11_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_5_4
#define init_1_11_5_4
void FASTCALL call_1_11_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_5_5
#define init_1_11_5_5
void FASTCALL call_1_11_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_5_6
#define init_1_11_5_6
void FASTCALL call_1_11_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_5_7
#define init_1_11_5_7
void FASTCALL call_1_11_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_5_8
#define init_1_11_5_8
void FASTCALL call_1_11_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_5_9
#define init_1_11_5_9
void FASTCALL call_1_11_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_5_10
#define init_1_11_5_10
void FASTCALL call_1_11_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_5_11
#define init_1_11_5_11
void FASTCALL call_1_11_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_5_12
#define init_1_11_5_12
void FASTCALL call_1_11_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_5_13
#define init_1_11_5_13
void FASTCALL call_1_11_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_5_14
#define init_1_11_5_14
void FASTCALL call_1_11_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_5_15
#define init_1_11_5_15
void FASTCALL call_1_11_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_6_0
#define init_1_11_6_0
void FASTCALL call_1_11_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_6_1
#define init_1_11_6_1
void FASTCALL call_1_11_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_6_2
#define init_1_11_6_2
void FASTCALL call_1_11_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_6_3
#define init_1_11_6_3
void FASTCALL call_1_11_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_6_4
#define init_1_11_6_4
void FASTCALL call_1_11_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_6_5
#define init_1_11_6_5
void FASTCALL call_1_11_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_6_6
#define init_1_11_6_6
void FASTCALL call_1_11_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_6_7
#define init_1_11_6_7
void FASTCALL call_1_11_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_6_8
#define init_1_11_6_8
void FASTCALL call_1_11_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_6_9
#define init_1_11_6_9
void FASTCALL call_1_11_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_6_10
#define init_1_11_6_10
void FASTCALL call_1_11_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_6_11
#define init_1_11_6_11
void FASTCALL call_1_11_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_6_12
#define init_1_11_6_12
void FASTCALL call_1_11_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_6_13
#define init_1_11_6_13
void FASTCALL call_1_11_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_6_14
#define init_1_11_6_14
void FASTCALL call_1_11_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_6_15
#define init_1_11_6_15
void FASTCALL call_1_11_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_7_0
#define init_1_11_7_0
void FASTCALL call_1_11_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_7_1
#define init_1_11_7_1
void FASTCALL call_1_11_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_7_2
#define init_1_11_7_2
void FASTCALL call_1_11_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_7_3
#define init_1_11_7_3
void FASTCALL call_1_11_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_7_4
#define init_1_11_7_4
void FASTCALL call_1_11_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_7_5
#define init_1_11_7_5
void FASTCALL call_1_11_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_7_6
#define init_1_11_7_6
void FASTCALL call_1_11_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_7_7
#define init_1_11_7_7
void FASTCALL call_1_11_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_7_8
#define init_1_11_7_8
void FASTCALL call_1_11_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_7_9
#define init_1_11_7_9
void FASTCALL call_1_11_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_7_10
#define init_1_11_7_10
void FASTCALL call_1_11_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_7_11
#define init_1_11_7_11
void FASTCALL call_1_11_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_7_12
#define init_1_11_7_12
void FASTCALL call_1_11_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_7_13
#define init_1_11_7_13
void FASTCALL call_1_11_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_7_14
#define init_1_11_7_14
void FASTCALL call_1_11_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_7_15
#define init_1_11_7_15
void FASTCALL call_1_11_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_8_0
#define init_1_11_8_0
void FASTCALL call_1_11_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_8_1
#define init_1_11_8_1
void FASTCALL call_1_11_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_8_2
#define init_1_11_8_2
void FASTCALL call_1_11_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_8_3
#define init_1_11_8_3
void FASTCALL call_1_11_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_8_4
#define init_1_11_8_4
void FASTCALL call_1_11_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_8_5
#define init_1_11_8_5
void FASTCALL call_1_11_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_8_6
#define init_1_11_8_6
void FASTCALL call_1_11_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_8_7
#define init_1_11_8_7
void FASTCALL call_1_11_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_8_8
#define init_1_11_8_8
void FASTCALL call_1_11_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_8_9
#define init_1_11_8_9
void FASTCALL call_1_11_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_8_10
#define init_1_11_8_10
void FASTCALL call_1_11_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_8_11
#define init_1_11_8_11
void FASTCALL call_1_11_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_8_12
#define init_1_11_8_12
void FASTCALL call_1_11_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_8_13
#define init_1_11_8_13
void FASTCALL call_1_11_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_8_14
#define init_1_11_8_14
void FASTCALL call_1_11_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_8_15
#define init_1_11_8_15
void FASTCALL call_1_11_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_9_0
#define init_1_11_9_0
void FASTCALL call_1_11_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_9_1
#define init_1_11_9_1
void FASTCALL call_1_11_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_9_2
#define init_1_11_9_2
void FASTCALL call_1_11_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_9_3
#define init_1_11_9_3
void FASTCALL call_1_11_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_9_4
#define init_1_11_9_4
void FASTCALL call_1_11_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_9_5
#define init_1_11_9_5
void FASTCALL call_1_11_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_9_6
#define init_1_11_9_6
void FASTCALL call_1_11_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_9_7
#define init_1_11_9_7
void FASTCALL call_1_11_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_9_8
#define init_1_11_9_8
void FASTCALL call_1_11_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_9_9
#define init_1_11_9_9
void FASTCALL call_1_11_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_9_10
#define init_1_11_9_10
void FASTCALL call_1_11_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_9_11
#define init_1_11_9_11
void FASTCALL call_1_11_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_9_12
#define init_1_11_9_12
void FASTCALL call_1_11_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_9_13
#define init_1_11_9_13
void FASTCALL call_1_11_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_9_14
#define init_1_11_9_14
void FASTCALL call_1_11_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_9_15
#define init_1_11_9_15
void FASTCALL call_1_11_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_10_0
#define init_1_11_10_0
void FASTCALL call_1_11_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_10_1
#define init_1_11_10_1
void FASTCALL call_1_11_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_10_2
#define init_1_11_10_2
void FASTCALL call_1_11_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_10_3
#define init_1_11_10_3
void FASTCALL call_1_11_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_10_4
#define init_1_11_10_4
void FASTCALL call_1_11_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_10_5
#define init_1_11_10_5
void FASTCALL call_1_11_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_10_6
#define init_1_11_10_6
void FASTCALL call_1_11_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_10_7
#define init_1_11_10_7
void FASTCALL call_1_11_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_10_8
#define init_1_11_10_8
void FASTCALL call_1_11_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_10_9
#define init_1_11_10_9
void FASTCALL call_1_11_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_10_10
#define init_1_11_10_10
void FASTCALL call_1_11_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_10_11
#define init_1_11_10_11
void FASTCALL call_1_11_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_10_12
#define init_1_11_10_12
void FASTCALL call_1_11_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_10_13
#define init_1_11_10_13
void FASTCALL call_1_11_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_10_14
#define init_1_11_10_14
void FASTCALL call_1_11_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_10_15
#define init_1_11_10_15
void FASTCALL call_1_11_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_11_0
#define init_1_11_11_0
void FASTCALL call_1_11_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_11_1
#define init_1_11_11_1
void FASTCALL call_1_11_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_11_2
#define init_1_11_11_2
void FASTCALL call_1_11_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_11_3
#define init_1_11_11_3
void FASTCALL call_1_11_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_11_4
#define init_1_11_11_4
void FASTCALL call_1_11_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_11_5
#define init_1_11_11_5
void FASTCALL call_1_11_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_11_6
#define init_1_11_11_6
void FASTCALL call_1_11_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_11_7
#define init_1_11_11_7
void FASTCALL call_1_11_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_11_8
#define init_1_11_11_8
void FASTCALL call_1_11_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_11_9
#define init_1_11_11_9
void FASTCALL call_1_11_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_11_10
#define init_1_11_11_10
void FASTCALL call_1_11_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_11_11
#define init_1_11_11_11
void FASTCALL call_1_11_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_11_12
#define init_1_11_11_12
void FASTCALL call_1_11_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_11_13
#define init_1_11_11_13
void FASTCALL call_1_11_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_11_14
#define init_1_11_11_14
void FASTCALL call_1_11_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_11_15
#define init_1_11_11_15
void FASTCALL call_1_11_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_12_0
#define init_1_11_12_0
void FASTCALL call_1_11_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_12_1
#define init_1_11_12_1
void FASTCALL call_1_11_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_12_2
#define init_1_11_12_2
void FASTCALL call_1_11_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_12_3
#define init_1_11_12_3
void FASTCALL call_1_11_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_12_4
#define init_1_11_12_4
void FASTCALL call_1_11_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_12_5
#define init_1_11_12_5
void FASTCALL call_1_11_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_12_6
#define init_1_11_12_6
void FASTCALL call_1_11_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_12_7
#define init_1_11_12_7
void FASTCALL call_1_11_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_12_8
#define init_1_11_12_8
void FASTCALL call_1_11_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_12_9
#define init_1_11_12_9
void FASTCALL call_1_11_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_12_10
#define init_1_11_12_10
void FASTCALL call_1_11_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_12_11
#define init_1_11_12_11
void FASTCALL call_1_11_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_12_12
#define init_1_11_12_12
void FASTCALL call_1_11_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_12_13
#define init_1_11_12_13
void FASTCALL call_1_11_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_12_14
#define init_1_11_12_14
void FASTCALL call_1_11_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_12_15
#define init_1_11_12_15
void FASTCALL call_1_11_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_13_0
#define init_1_11_13_0
void FASTCALL call_1_11_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_13_1
#define init_1_11_13_1
void FASTCALL call_1_11_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_13_2
#define init_1_11_13_2
void FASTCALL call_1_11_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_13_3
#define init_1_11_13_3
void FASTCALL call_1_11_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_13_4
#define init_1_11_13_4
void FASTCALL call_1_11_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_13_5
#define init_1_11_13_5
void FASTCALL call_1_11_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_13_6
#define init_1_11_13_6
void FASTCALL call_1_11_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_13_7
#define init_1_11_13_7
void FASTCALL call_1_11_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_13_8
#define init_1_11_13_8
void FASTCALL call_1_11_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_13_9
#define init_1_11_13_9
void FASTCALL call_1_11_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_13_10
#define init_1_11_13_10
void FASTCALL call_1_11_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_13_11
#define init_1_11_13_11
void FASTCALL call_1_11_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_13_12
#define init_1_11_13_12
void FASTCALL call_1_11_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_13_13
#define init_1_11_13_13
void FASTCALL call_1_11_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_13_14
#define init_1_11_13_14
void FASTCALL call_1_11_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_13_15
#define init_1_11_13_15
void FASTCALL call_1_11_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_14_0
#define init_1_11_14_0
void FASTCALL call_1_11_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_14_1
#define init_1_11_14_1
void FASTCALL call_1_11_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_14_2
#define init_1_11_14_2
void FASTCALL call_1_11_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_14_3
#define init_1_11_14_3
void FASTCALL call_1_11_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_14_4
#define init_1_11_14_4
void FASTCALL call_1_11_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_14_5
#define init_1_11_14_5
void FASTCALL call_1_11_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_14_6
#define init_1_11_14_6
void FASTCALL call_1_11_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_14_7
#define init_1_11_14_7
void FASTCALL call_1_11_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_14_8
#define init_1_11_14_8
void FASTCALL call_1_11_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_14_9
#define init_1_11_14_9
void FASTCALL call_1_11_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_14_10
#define init_1_11_14_10
void FASTCALL call_1_11_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_14_11
#define init_1_11_14_11
void FASTCALL call_1_11_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_14_12
#define init_1_11_14_12
void FASTCALL call_1_11_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_14_13
#define init_1_11_14_13
void FASTCALL call_1_11_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_14_14
#define init_1_11_14_14
void FASTCALL call_1_11_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_14_15
#define init_1_11_14_15
void FASTCALL call_1_11_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_15_0
#define init_1_11_15_0
void FASTCALL call_1_11_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_15_1
#define init_1_11_15_1
void FASTCALL call_1_11_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_15_2
#define init_1_11_15_2
void FASTCALL call_1_11_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_15_3
#define init_1_11_15_3
void FASTCALL call_1_11_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_15_4
#define init_1_11_15_4
void FASTCALL call_1_11_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_15_5
#define init_1_11_15_5
void FASTCALL call_1_11_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_15_6
#define init_1_11_15_6
void FASTCALL call_1_11_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_15_7
#define init_1_11_15_7
void FASTCALL call_1_11_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_15_8
#define init_1_11_15_8
void FASTCALL call_1_11_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_15_9
#define init_1_11_15_9
void FASTCALL call_1_11_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_15_10
#define init_1_11_15_10
void FASTCALL call_1_11_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_15_11
#define init_1_11_15_11
void FASTCALL call_1_11_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_15_12
#define init_1_11_15_12
void FASTCALL call_1_11_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_15_13
#define init_1_11_15_13
void FASTCALL call_1_11_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_15_14
#define init_1_11_15_14
void FASTCALL call_1_11_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_11_15_15
#define init_1_11_15_15
void FASTCALL call_1_11_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_0_0
#define init_1_12_0_0
void FASTCALL call_1_12_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_0_1
#define init_1_12_0_1
void FASTCALL call_1_12_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_0_2
#define init_1_12_0_2
void FASTCALL call_1_12_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_0_3
#define init_1_12_0_3
void FASTCALL call_1_12_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_0_4
#define init_1_12_0_4
void FASTCALL call_1_12_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_0_5
#define init_1_12_0_5
void FASTCALL call_1_12_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_0_6
#define init_1_12_0_6
void FASTCALL call_1_12_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_0_7
#define init_1_12_0_7
void FASTCALL call_1_12_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_0_8
#define init_1_12_0_8
void FASTCALL call_1_12_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_0_9
#define init_1_12_0_9
void FASTCALL call_1_12_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_0_10
#define init_1_12_0_10
void FASTCALL call_1_12_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_0_11
#define init_1_12_0_11
void FASTCALL call_1_12_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_0_12
#define init_1_12_0_12
void FASTCALL call_1_12_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_0_13
#define init_1_12_0_13
void FASTCALL call_1_12_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_0_14
#define init_1_12_0_14
void FASTCALL call_1_12_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_0_15
#define init_1_12_0_15
void FASTCALL call_1_12_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_1_0
#define init_1_12_1_0
void FASTCALL call_1_12_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_1_1
#define init_1_12_1_1
void FASTCALL call_1_12_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_1_2
#define init_1_12_1_2
void FASTCALL call_1_12_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_1_3
#define init_1_12_1_3
void FASTCALL call_1_12_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_1_4
#define init_1_12_1_4
void FASTCALL call_1_12_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_1_5
#define init_1_12_1_5
void FASTCALL call_1_12_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_1_6
#define init_1_12_1_6
void FASTCALL call_1_12_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_1_7
#define init_1_12_1_7
void FASTCALL call_1_12_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_1_8
#define init_1_12_1_8
void FASTCALL call_1_12_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_1_9
#define init_1_12_1_9
void FASTCALL call_1_12_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_1_10
#define init_1_12_1_10
void FASTCALL call_1_12_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_1_11
#define init_1_12_1_11
void FASTCALL call_1_12_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_1_12
#define init_1_12_1_12
void FASTCALL call_1_12_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_1_13
#define init_1_12_1_13
void FASTCALL call_1_12_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_1_14
#define init_1_12_1_14
void FASTCALL call_1_12_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_1_15
#define init_1_12_1_15
void FASTCALL call_1_12_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_2_0
#define init_1_12_2_0
void FASTCALL call_1_12_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_2_1
#define init_1_12_2_1
void FASTCALL call_1_12_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_2_2
#define init_1_12_2_2
void FASTCALL call_1_12_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_2_3
#define init_1_12_2_3
void FASTCALL call_1_12_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_2_4
#define init_1_12_2_4
void FASTCALL call_1_12_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_2_5
#define init_1_12_2_5
void FASTCALL call_1_12_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_2_6
#define init_1_12_2_6
void FASTCALL call_1_12_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_2_7
#define init_1_12_2_7
void FASTCALL call_1_12_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_2_8
#define init_1_12_2_8
void FASTCALL call_1_12_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_2_9
#define init_1_12_2_9
void FASTCALL call_1_12_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_2_10
#define init_1_12_2_10
void FASTCALL call_1_12_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_2_11
#define init_1_12_2_11
void FASTCALL call_1_12_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_2_12
#define init_1_12_2_12
void FASTCALL call_1_12_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_2_13
#define init_1_12_2_13
void FASTCALL call_1_12_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_2_14
#define init_1_12_2_14
void FASTCALL call_1_12_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_2_15
#define init_1_12_2_15
void FASTCALL call_1_12_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_3_0
#define init_1_12_3_0
void FASTCALL call_1_12_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_3_1
#define init_1_12_3_1
void FASTCALL call_1_12_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_3_2
#define init_1_12_3_2
void FASTCALL call_1_12_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_3_3
#define init_1_12_3_3
void FASTCALL call_1_12_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_3_4
#define init_1_12_3_4
void FASTCALL call_1_12_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_3_5
#define init_1_12_3_5
void FASTCALL call_1_12_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_3_6
#define init_1_12_3_6
void FASTCALL call_1_12_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_3_7
#define init_1_12_3_7
void FASTCALL call_1_12_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_3_8
#define init_1_12_3_8
void FASTCALL call_1_12_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_3_9
#define init_1_12_3_9
void FASTCALL call_1_12_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_3_10
#define init_1_12_3_10
void FASTCALL call_1_12_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_3_11
#define init_1_12_3_11
void FASTCALL call_1_12_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_3_12
#define init_1_12_3_12
void FASTCALL call_1_12_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_3_13
#define init_1_12_3_13
void FASTCALL call_1_12_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_3_14
#define init_1_12_3_14
void FASTCALL call_1_12_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_3_15
#define init_1_12_3_15
void FASTCALL call_1_12_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_4_0
#define init_1_12_4_0
void FASTCALL call_1_12_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_4_1
#define init_1_12_4_1
void FASTCALL call_1_12_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_4_2
#define init_1_12_4_2
void FASTCALL call_1_12_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_4_3
#define init_1_12_4_3
void FASTCALL call_1_12_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_4_4
#define init_1_12_4_4
void FASTCALL call_1_12_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_4_5
#define init_1_12_4_5
void FASTCALL call_1_12_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_4_6
#define init_1_12_4_6
void FASTCALL call_1_12_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_4_7
#define init_1_12_4_7
void FASTCALL call_1_12_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_4_8
#define init_1_12_4_8
void FASTCALL call_1_12_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_4_9
#define init_1_12_4_9
void FASTCALL call_1_12_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_4_10
#define init_1_12_4_10
void FASTCALL call_1_12_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_4_11
#define init_1_12_4_11
void FASTCALL call_1_12_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_4_12
#define init_1_12_4_12
void FASTCALL call_1_12_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_4_13
#define init_1_12_4_13
void FASTCALL call_1_12_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_4_14
#define init_1_12_4_14
void FASTCALL call_1_12_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_4_15
#define init_1_12_4_15
void FASTCALL call_1_12_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_5_0
#define init_1_12_5_0
void FASTCALL call_1_12_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_5_1
#define init_1_12_5_1
void FASTCALL call_1_12_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_5_2
#define init_1_12_5_2
void FASTCALL call_1_12_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_5_3
#define init_1_12_5_3
void FASTCALL call_1_12_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_5_4
#define init_1_12_5_4
void FASTCALL call_1_12_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_5_5
#define init_1_12_5_5
void FASTCALL call_1_12_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_5_6
#define init_1_12_5_6
void FASTCALL call_1_12_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_5_7
#define init_1_12_5_7
void FASTCALL call_1_12_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_5_8
#define init_1_12_5_8
void FASTCALL call_1_12_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_5_9
#define init_1_12_5_9
void FASTCALL call_1_12_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_5_10
#define init_1_12_5_10
void FASTCALL call_1_12_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_5_11
#define init_1_12_5_11
void FASTCALL call_1_12_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_5_12
#define init_1_12_5_12
void FASTCALL call_1_12_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_5_13
#define init_1_12_5_13
void FASTCALL call_1_12_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_5_14
#define init_1_12_5_14
void FASTCALL call_1_12_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_5_15
#define init_1_12_5_15
void FASTCALL call_1_12_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_6_0
#define init_1_12_6_0
void FASTCALL call_1_12_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_6_1
#define init_1_12_6_1
void FASTCALL call_1_12_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_6_2
#define init_1_12_6_2
void FASTCALL call_1_12_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_6_3
#define init_1_12_6_3
void FASTCALL call_1_12_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_6_4
#define init_1_12_6_4
void FASTCALL call_1_12_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_6_5
#define init_1_12_6_5
void FASTCALL call_1_12_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_6_6
#define init_1_12_6_6
void FASTCALL call_1_12_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_6_7
#define init_1_12_6_7
void FASTCALL call_1_12_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_6_8
#define init_1_12_6_8
void FASTCALL call_1_12_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_6_9
#define init_1_12_6_9
void FASTCALL call_1_12_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_6_10
#define init_1_12_6_10
void FASTCALL call_1_12_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_6_11
#define init_1_12_6_11
void FASTCALL call_1_12_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_6_12
#define init_1_12_6_12
void FASTCALL call_1_12_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_6_13
#define init_1_12_6_13
void FASTCALL call_1_12_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_6_14
#define init_1_12_6_14
void FASTCALL call_1_12_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_6_15
#define init_1_12_6_15
void FASTCALL call_1_12_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_7_0
#define init_1_12_7_0
void FASTCALL call_1_12_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_7_1
#define init_1_12_7_1
void FASTCALL call_1_12_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_7_2
#define init_1_12_7_2
void FASTCALL call_1_12_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_7_3
#define init_1_12_7_3
void FASTCALL call_1_12_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_7_4
#define init_1_12_7_4
void FASTCALL call_1_12_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_7_5
#define init_1_12_7_5
void FASTCALL call_1_12_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_7_6
#define init_1_12_7_6
void FASTCALL call_1_12_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_7_7
#define init_1_12_7_7
void FASTCALL call_1_12_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_7_8
#define init_1_12_7_8
void FASTCALL call_1_12_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_7_9
#define init_1_12_7_9
void FASTCALL call_1_12_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_7_10
#define init_1_12_7_10
void FASTCALL call_1_12_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_7_11
#define init_1_12_7_11
void FASTCALL call_1_12_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_7_12
#define init_1_12_7_12
void FASTCALL call_1_12_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_7_13
#define init_1_12_7_13
void FASTCALL call_1_12_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_7_14
#define init_1_12_7_14
void FASTCALL call_1_12_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_7_15
#define init_1_12_7_15
void FASTCALL call_1_12_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_8_0
#define init_1_12_8_0
void FASTCALL call_1_12_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_8_1
#define init_1_12_8_1
void FASTCALL call_1_12_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_8_2
#define init_1_12_8_2
void FASTCALL call_1_12_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_8_3
#define init_1_12_8_3
void FASTCALL call_1_12_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_8_4
#define init_1_12_8_4
void FASTCALL call_1_12_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_8_5
#define init_1_12_8_5
void FASTCALL call_1_12_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_8_6
#define init_1_12_8_6
void FASTCALL call_1_12_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_8_7
#define init_1_12_8_7
void FASTCALL call_1_12_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_8_8
#define init_1_12_8_8
void FASTCALL call_1_12_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_8_9
#define init_1_12_8_9
void FASTCALL call_1_12_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_8_10
#define init_1_12_8_10
void FASTCALL call_1_12_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_8_11
#define init_1_12_8_11
void FASTCALL call_1_12_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_8_12
#define init_1_12_8_12
void FASTCALL call_1_12_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_8_13
#define init_1_12_8_13
void FASTCALL call_1_12_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_8_14
#define init_1_12_8_14
void FASTCALL call_1_12_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_8_15
#define init_1_12_8_15
void FASTCALL call_1_12_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_9_0
#define init_1_12_9_0
void FASTCALL call_1_12_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_9_1
#define init_1_12_9_1
void FASTCALL call_1_12_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_9_2
#define init_1_12_9_2
void FASTCALL call_1_12_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_9_3
#define init_1_12_9_3
void FASTCALL call_1_12_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_9_4
#define init_1_12_9_4
void FASTCALL call_1_12_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_9_5
#define init_1_12_9_5
void FASTCALL call_1_12_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_9_6
#define init_1_12_9_6
void FASTCALL call_1_12_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_9_7
#define init_1_12_9_7
void FASTCALL call_1_12_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_9_8
#define init_1_12_9_8
void FASTCALL call_1_12_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_9_9
#define init_1_12_9_9
void FASTCALL call_1_12_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_9_10
#define init_1_12_9_10
void FASTCALL call_1_12_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_9_11
#define init_1_12_9_11
void FASTCALL call_1_12_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_9_12
#define init_1_12_9_12
void FASTCALL call_1_12_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_9_13
#define init_1_12_9_13
void FASTCALL call_1_12_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_9_14
#define init_1_12_9_14
void FASTCALL call_1_12_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_9_15
#define init_1_12_9_15
void FASTCALL call_1_12_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_10_0
#define init_1_12_10_0
void FASTCALL call_1_12_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_10_1
#define init_1_12_10_1
void FASTCALL call_1_12_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_10_2
#define init_1_12_10_2
void FASTCALL call_1_12_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_10_3
#define init_1_12_10_3
void FASTCALL call_1_12_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_10_4
#define init_1_12_10_4
void FASTCALL call_1_12_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_10_5
#define init_1_12_10_5
void FASTCALL call_1_12_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_10_6
#define init_1_12_10_6
void FASTCALL call_1_12_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_10_7
#define init_1_12_10_7
void FASTCALL call_1_12_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_10_8
#define init_1_12_10_8
void FASTCALL call_1_12_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_10_9
#define init_1_12_10_9
void FASTCALL call_1_12_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_10_10
#define init_1_12_10_10
void FASTCALL call_1_12_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_10_11
#define init_1_12_10_11
void FASTCALL call_1_12_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_10_12
#define init_1_12_10_12
void FASTCALL call_1_12_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_10_13
#define init_1_12_10_13
void FASTCALL call_1_12_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_10_14
#define init_1_12_10_14
void FASTCALL call_1_12_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_10_15
#define init_1_12_10_15
void FASTCALL call_1_12_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_11_0
#define init_1_12_11_0
void FASTCALL call_1_12_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_11_1
#define init_1_12_11_1
void FASTCALL call_1_12_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_11_2
#define init_1_12_11_2
void FASTCALL call_1_12_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_11_3
#define init_1_12_11_3
void FASTCALL call_1_12_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_11_4
#define init_1_12_11_4
void FASTCALL call_1_12_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_11_5
#define init_1_12_11_5
void FASTCALL call_1_12_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_11_6
#define init_1_12_11_6
void FASTCALL call_1_12_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_11_7
#define init_1_12_11_7
void FASTCALL call_1_12_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_11_8
#define init_1_12_11_8
void FASTCALL call_1_12_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_11_9
#define init_1_12_11_9
void FASTCALL call_1_12_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_11_10
#define init_1_12_11_10
void FASTCALL call_1_12_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_11_11
#define init_1_12_11_11
void FASTCALL call_1_12_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_11_12
#define init_1_12_11_12
void FASTCALL call_1_12_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_11_13
#define init_1_12_11_13
void FASTCALL call_1_12_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_11_14
#define init_1_12_11_14
void FASTCALL call_1_12_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_11_15
#define init_1_12_11_15
void FASTCALL call_1_12_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_12_0
#define init_1_12_12_0
void FASTCALL call_1_12_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_12_1
#define init_1_12_12_1
void FASTCALL call_1_12_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_12_2
#define init_1_12_12_2
void FASTCALL call_1_12_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_12_3
#define init_1_12_12_3
void FASTCALL call_1_12_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_12_4
#define init_1_12_12_4
void FASTCALL call_1_12_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_12_5
#define init_1_12_12_5
void FASTCALL call_1_12_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_12_6
#define init_1_12_12_6
void FASTCALL call_1_12_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_12_7
#define init_1_12_12_7
void FASTCALL call_1_12_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_12_8
#define init_1_12_12_8
void FASTCALL call_1_12_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_12_9
#define init_1_12_12_9
void FASTCALL call_1_12_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_12_10
#define init_1_12_12_10
void FASTCALL call_1_12_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_12_11
#define init_1_12_12_11
void FASTCALL call_1_12_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_12_12
#define init_1_12_12_12
void FASTCALL call_1_12_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_12_13
#define init_1_12_12_13
void FASTCALL call_1_12_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_12_14
#define init_1_12_12_14
void FASTCALL call_1_12_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_12_15
#define init_1_12_12_15
void FASTCALL call_1_12_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_13_0
#define init_1_12_13_0
void FASTCALL call_1_12_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_13_1
#define init_1_12_13_1
void FASTCALL call_1_12_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_13_2
#define init_1_12_13_2
void FASTCALL call_1_12_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_13_3
#define init_1_12_13_3
void FASTCALL call_1_12_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_13_4
#define init_1_12_13_4
void FASTCALL call_1_12_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_13_5
#define init_1_12_13_5
void FASTCALL call_1_12_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_13_6
#define init_1_12_13_6
void FASTCALL call_1_12_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_13_7
#define init_1_12_13_7
void FASTCALL call_1_12_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_13_8
#define init_1_12_13_8
void FASTCALL call_1_12_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_13_9
#define init_1_12_13_9
void FASTCALL call_1_12_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_13_10
#define init_1_12_13_10
void FASTCALL call_1_12_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_13_11
#define init_1_12_13_11
void FASTCALL call_1_12_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_13_12
#define init_1_12_13_12
void FASTCALL call_1_12_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_13_13
#define init_1_12_13_13
void FASTCALL call_1_12_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_13_14
#define init_1_12_13_14
void FASTCALL call_1_12_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_13_15
#define init_1_12_13_15
void FASTCALL call_1_12_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_14_0
#define init_1_12_14_0
void FASTCALL call_1_12_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_14_1
#define init_1_12_14_1
void FASTCALL call_1_12_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_14_2
#define init_1_12_14_2
void FASTCALL call_1_12_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_14_3
#define init_1_12_14_3
void FASTCALL call_1_12_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_14_4
#define init_1_12_14_4
void FASTCALL call_1_12_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_14_5
#define init_1_12_14_5
void FASTCALL call_1_12_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_14_6
#define init_1_12_14_6
void FASTCALL call_1_12_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_14_7
#define init_1_12_14_7
void FASTCALL call_1_12_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_14_8
#define init_1_12_14_8
void FASTCALL call_1_12_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_14_9
#define init_1_12_14_9
void FASTCALL call_1_12_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_14_10
#define init_1_12_14_10
void FASTCALL call_1_12_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_14_11
#define init_1_12_14_11
void FASTCALL call_1_12_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_14_12
#define init_1_12_14_12
void FASTCALL call_1_12_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_14_13
#define init_1_12_14_13
void FASTCALL call_1_12_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_14_14
#define init_1_12_14_14
void FASTCALL call_1_12_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_14_15
#define init_1_12_14_15
void FASTCALL call_1_12_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_15_0
#define init_1_12_15_0
void FASTCALL call_1_12_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_15_1
#define init_1_12_15_1
void FASTCALL call_1_12_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_15_2
#define init_1_12_15_2
void FASTCALL call_1_12_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_15_3
#define init_1_12_15_3
void FASTCALL call_1_12_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_15_4
#define init_1_12_15_4
void FASTCALL call_1_12_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_15_5
#define init_1_12_15_5
void FASTCALL call_1_12_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_15_6
#define init_1_12_15_6
void FASTCALL call_1_12_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_15_7
#define init_1_12_15_7
void FASTCALL call_1_12_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_15_8
#define init_1_12_15_8
void FASTCALL call_1_12_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_15_9
#define init_1_12_15_9
void FASTCALL call_1_12_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_15_10
#define init_1_12_15_10
void FASTCALL call_1_12_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_15_11
#define init_1_12_15_11
void FASTCALL call_1_12_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_15_12
#define init_1_12_15_12
void FASTCALL call_1_12_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_15_13
#define init_1_12_15_13
void FASTCALL call_1_12_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_15_14
#define init_1_12_15_14
void FASTCALL call_1_12_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_12_15_15
#define init_1_12_15_15
void FASTCALL call_1_12_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_0_0
#define init_1_13_0_0
void FASTCALL call_1_13_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_0_1
#define init_1_13_0_1
void FASTCALL call_1_13_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_0_2
#define init_1_13_0_2
void FASTCALL call_1_13_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_0_3
#define init_1_13_0_3
void FASTCALL call_1_13_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_0_4
#define init_1_13_0_4
void FASTCALL call_1_13_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_0_5
#define init_1_13_0_5
void FASTCALL call_1_13_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_0_6
#define init_1_13_0_6
void FASTCALL call_1_13_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_0_7
#define init_1_13_0_7
void FASTCALL call_1_13_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_0_8
#define init_1_13_0_8
void FASTCALL call_1_13_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_0_9
#define init_1_13_0_9
void FASTCALL call_1_13_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_0_10
#define init_1_13_0_10
void FASTCALL call_1_13_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_0_11
#define init_1_13_0_11
void FASTCALL call_1_13_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_0_12
#define init_1_13_0_12
void FASTCALL call_1_13_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_0_13
#define init_1_13_0_13
void FASTCALL call_1_13_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_0_14
#define init_1_13_0_14
void FASTCALL call_1_13_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_0_15
#define init_1_13_0_15
void FASTCALL call_1_13_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_1_0
#define init_1_13_1_0
void FASTCALL call_1_13_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_1_1
#define init_1_13_1_1
void FASTCALL call_1_13_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_1_2
#define init_1_13_1_2
void FASTCALL call_1_13_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_1_3
#define init_1_13_1_3
void FASTCALL call_1_13_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_1_4
#define init_1_13_1_4
void FASTCALL call_1_13_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_1_5
#define init_1_13_1_5
void FASTCALL call_1_13_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_1_6
#define init_1_13_1_6
void FASTCALL call_1_13_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_1_7
#define init_1_13_1_7
void FASTCALL call_1_13_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_1_8
#define init_1_13_1_8
void FASTCALL call_1_13_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_1_9
#define init_1_13_1_9
void FASTCALL call_1_13_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_1_10
#define init_1_13_1_10
void FASTCALL call_1_13_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_1_11
#define init_1_13_1_11
void FASTCALL call_1_13_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_1_12
#define init_1_13_1_12
void FASTCALL call_1_13_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_1_13
#define init_1_13_1_13
void FASTCALL call_1_13_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_1_14
#define init_1_13_1_14
void FASTCALL call_1_13_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_1_15
#define init_1_13_1_15
void FASTCALL call_1_13_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_2_0
#define init_1_13_2_0
void FASTCALL call_1_13_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_2_1
#define init_1_13_2_1
void FASTCALL call_1_13_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_2_2
#define init_1_13_2_2
void FASTCALL call_1_13_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_2_3
#define init_1_13_2_3
void FASTCALL call_1_13_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_2_4
#define init_1_13_2_4
void FASTCALL call_1_13_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_2_5
#define init_1_13_2_5
void FASTCALL call_1_13_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_2_6
#define init_1_13_2_6
void FASTCALL call_1_13_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_2_7
#define init_1_13_2_7
void FASTCALL call_1_13_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_2_8
#define init_1_13_2_8
void FASTCALL call_1_13_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_2_9
#define init_1_13_2_9
void FASTCALL call_1_13_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_2_10
#define init_1_13_2_10
void FASTCALL call_1_13_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_2_11
#define init_1_13_2_11
void FASTCALL call_1_13_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_2_12
#define init_1_13_2_12
void FASTCALL call_1_13_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_2_13
#define init_1_13_2_13
void FASTCALL call_1_13_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_2_14
#define init_1_13_2_14
void FASTCALL call_1_13_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_2_15
#define init_1_13_2_15
void FASTCALL call_1_13_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_3_0
#define init_1_13_3_0
void FASTCALL call_1_13_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_3_1
#define init_1_13_3_1
void FASTCALL call_1_13_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_3_2
#define init_1_13_3_2
void FASTCALL call_1_13_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_3_3
#define init_1_13_3_3
void FASTCALL call_1_13_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_3_4
#define init_1_13_3_4
void FASTCALL call_1_13_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_3_5
#define init_1_13_3_5
void FASTCALL call_1_13_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_3_6
#define init_1_13_3_6
void FASTCALL call_1_13_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_3_7
#define init_1_13_3_7
void FASTCALL call_1_13_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_3_8
#define init_1_13_3_8
void FASTCALL call_1_13_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_3_9
#define init_1_13_3_9
void FASTCALL call_1_13_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_3_10
#define init_1_13_3_10
void FASTCALL call_1_13_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_3_11
#define init_1_13_3_11
void FASTCALL call_1_13_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_3_12
#define init_1_13_3_12
void FASTCALL call_1_13_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_3_13
#define init_1_13_3_13
void FASTCALL call_1_13_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_3_14
#define init_1_13_3_14
void FASTCALL call_1_13_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_3_15
#define init_1_13_3_15
void FASTCALL call_1_13_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_4_0
#define init_1_13_4_0
void FASTCALL call_1_13_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_4_1
#define init_1_13_4_1
void FASTCALL call_1_13_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_4_2
#define init_1_13_4_2
void FASTCALL call_1_13_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_4_3
#define init_1_13_4_3
void FASTCALL call_1_13_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_4_4
#define init_1_13_4_4
void FASTCALL call_1_13_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_4_5
#define init_1_13_4_5
void FASTCALL call_1_13_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_4_6
#define init_1_13_4_6
void FASTCALL call_1_13_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_4_7
#define init_1_13_4_7
void FASTCALL call_1_13_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_4_8
#define init_1_13_4_8
void FASTCALL call_1_13_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_4_9
#define init_1_13_4_9
void FASTCALL call_1_13_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_4_10
#define init_1_13_4_10
void FASTCALL call_1_13_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_4_11
#define init_1_13_4_11
void FASTCALL call_1_13_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_4_12
#define init_1_13_4_12
void FASTCALL call_1_13_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_4_13
#define init_1_13_4_13
void FASTCALL call_1_13_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_4_14
#define init_1_13_4_14
void FASTCALL call_1_13_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_4_15
#define init_1_13_4_15
void FASTCALL call_1_13_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_5_0
#define init_1_13_5_0
void FASTCALL call_1_13_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_5_1
#define init_1_13_5_1
void FASTCALL call_1_13_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_5_2
#define init_1_13_5_2
void FASTCALL call_1_13_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_5_3
#define init_1_13_5_3
void FASTCALL call_1_13_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_5_4
#define init_1_13_5_4
void FASTCALL call_1_13_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_5_5
#define init_1_13_5_5
void FASTCALL call_1_13_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_5_6
#define init_1_13_5_6
void FASTCALL call_1_13_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_5_7
#define init_1_13_5_7
void FASTCALL call_1_13_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_5_8
#define init_1_13_5_8
void FASTCALL call_1_13_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_5_9
#define init_1_13_5_9
void FASTCALL call_1_13_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_5_10
#define init_1_13_5_10
void FASTCALL call_1_13_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_5_11
#define init_1_13_5_11
void FASTCALL call_1_13_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_5_12
#define init_1_13_5_12
void FASTCALL call_1_13_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_5_13
#define init_1_13_5_13
void FASTCALL call_1_13_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_5_14
#define init_1_13_5_14
void FASTCALL call_1_13_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_5_15
#define init_1_13_5_15
void FASTCALL call_1_13_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_6_0
#define init_1_13_6_0
void FASTCALL call_1_13_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_6_1
#define init_1_13_6_1
void FASTCALL call_1_13_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_6_2
#define init_1_13_6_2
void FASTCALL call_1_13_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_6_3
#define init_1_13_6_3
void FASTCALL call_1_13_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_6_4
#define init_1_13_6_4
void FASTCALL call_1_13_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_6_5
#define init_1_13_6_5
void FASTCALL call_1_13_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_6_6
#define init_1_13_6_6
void FASTCALL call_1_13_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_6_7
#define init_1_13_6_7
void FASTCALL call_1_13_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_6_8
#define init_1_13_6_8
void FASTCALL call_1_13_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_6_9
#define init_1_13_6_9
void FASTCALL call_1_13_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_6_10
#define init_1_13_6_10
void FASTCALL call_1_13_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_6_11
#define init_1_13_6_11
void FASTCALL call_1_13_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_6_12
#define init_1_13_6_12
void FASTCALL call_1_13_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_6_13
#define init_1_13_6_13
void FASTCALL call_1_13_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_6_14
#define init_1_13_6_14
void FASTCALL call_1_13_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_6_15
#define init_1_13_6_15
void FASTCALL call_1_13_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_7_0
#define init_1_13_7_0
void FASTCALL call_1_13_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_7_1
#define init_1_13_7_1
void FASTCALL call_1_13_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_7_2
#define init_1_13_7_2
void FASTCALL call_1_13_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_7_3
#define init_1_13_7_3
void FASTCALL call_1_13_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_7_4
#define init_1_13_7_4
void FASTCALL call_1_13_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_7_5
#define init_1_13_7_5
void FASTCALL call_1_13_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_7_6
#define init_1_13_7_6
void FASTCALL call_1_13_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_7_7
#define init_1_13_7_7
void FASTCALL call_1_13_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_7_8
#define init_1_13_7_8
void FASTCALL call_1_13_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_7_9
#define init_1_13_7_9
void FASTCALL call_1_13_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_7_10
#define init_1_13_7_10
void FASTCALL call_1_13_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_7_11
#define init_1_13_7_11
void FASTCALL call_1_13_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_7_12
#define init_1_13_7_12
void FASTCALL call_1_13_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_7_13
#define init_1_13_7_13
void FASTCALL call_1_13_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_7_14
#define init_1_13_7_14
void FASTCALL call_1_13_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_7_15
#define init_1_13_7_15
void FASTCALL call_1_13_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_8_0
#define init_1_13_8_0
void FASTCALL call_1_13_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_8_1
#define init_1_13_8_1
void FASTCALL call_1_13_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_8_2
#define init_1_13_8_2
void FASTCALL call_1_13_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_8_3
#define init_1_13_8_3
void FASTCALL call_1_13_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_8_4
#define init_1_13_8_4
void FASTCALL call_1_13_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_8_5
#define init_1_13_8_5
void FASTCALL call_1_13_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_8_6
#define init_1_13_8_6
void FASTCALL call_1_13_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_8_7
#define init_1_13_8_7
void FASTCALL call_1_13_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_8_8
#define init_1_13_8_8
void FASTCALL call_1_13_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_8_9
#define init_1_13_8_9
void FASTCALL call_1_13_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_8_10
#define init_1_13_8_10
void FASTCALL call_1_13_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_8_11
#define init_1_13_8_11
void FASTCALL call_1_13_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_8_12
#define init_1_13_8_12
void FASTCALL call_1_13_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_8_13
#define init_1_13_8_13
void FASTCALL call_1_13_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_8_14
#define init_1_13_8_14
void FASTCALL call_1_13_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_8_15
#define init_1_13_8_15
void FASTCALL call_1_13_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_9_0
#define init_1_13_9_0
void FASTCALL call_1_13_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_9_1
#define init_1_13_9_1
void FASTCALL call_1_13_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_9_2
#define init_1_13_9_2
void FASTCALL call_1_13_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_9_3
#define init_1_13_9_3
void FASTCALL call_1_13_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_9_4
#define init_1_13_9_4
void FASTCALL call_1_13_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_9_5
#define init_1_13_9_5
void FASTCALL call_1_13_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_9_6
#define init_1_13_9_6
void FASTCALL call_1_13_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_9_7
#define init_1_13_9_7
void FASTCALL call_1_13_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_9_8
#define init_1_13_9_8
void FASTCALL call_1_13_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_9_9
#define init_1_13_9_9
void FASTCALL call_1_13_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_9_10
#define init_1_13_9_10
void FASTCALL call_1_13_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_9_11
#define init_1_13_9_11
void FASTCALL call_1_13_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_9_12
#define init_1_13_9_12
void FASTCALL call_1_13_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_9_13
#define init_1_13_9_13
void FASTCALL call_1_13_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_9_14
#define init_1_13_9_14
void FASTCALL call_1_13_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_9_15
#define init_1_13_9_15
void FASTCALL call_1_13_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_10_0
#define init_1_13_10_0
void FASTCALL call_1_13_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_10_1
#define init_1_13_10_1
void FASTCALL call_1_13_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_10_2
#define init_1_13_10_2
void FASTCALL call_1_13_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_10_3
#define init_1_13_10_3
void FASTCALL call_1_13_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_10_4
#define init_1_13_10_4
void FASTCALL call_1_13_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_10_5
#define init_1_13_10_5
void FASTCALL call_1_13_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_10_6
#define init_1_13_10_6
void FASTCALL call_1_13_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_10_7
#define init_1_13_10_7
void FASTCALL call_1_13_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_10_8
#define init_1_13_10_8
void FASTCALL call_1_13_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_10_9
#define init_1_13_10_9
void FASTCALL call_1_13_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_10_10
#define init_1_13_10_10
void FASTCALL call_1_13_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_10_11
#define init_1_13_10_11
void FASTCALL call_1_13_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_10_12
#define init_1_13_10_12
void FASTCALL call_1_13_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_10_13
#define init_1_13_10_13
void FASTCALL call_1_13_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_10_14
#define init_1_13_10_14
void FASTCALL call_1_13_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_10_15
#define init_1_13_10_15
void FASTCALL call_1_13_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_11_0
#define init_1_13_11_0
void FASTCALL call_1_13_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_11_1
#define init_1_13_11_1
void FASTCALL call_1_13_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_11_2
#define init_1_13_11_2
void FASTCALL call_1_13_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_11_3
#define init_1_13_11_3
void FASTCALL call_1_13_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_11_4
#define init_1_13_11_4
void FASTCALL call_1_13_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_11_5
#define init_1_13_11_5
void FASTCALL call_1_13_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_11_6
#define init_1_13_11_6
void FASTCALL call_1_13_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_11_7
#define init_1_13_11_7
void FASTCALL call_1_13_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_11_8
#define init_1_13_11_8
void FASTCALL call_1_13_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_11_9
#define init_1_13_11_9
void FASTCALL call_1_13_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_11_10
#define init_1_13_11_10
void FASTCALL call_1_13_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_11_11
#define init_1_13_11_11
void FASTCALL call_1_13_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_11_12
#define init_1_13_11_12
void FASTCALL call_1_13_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_11_13
#define init_1_13_11_13
void FASTCALL call_1_13_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_11_14
#define init_1_13_11_14
void FASTCALL call_1_13_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_11_15
#define init_1_13_11_15
void FASTCALL call_1_13_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_12_0
#define init_1_13_12_0
void FASTCALL call_1_13_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_12_1
#define init_1_13_12_1
void FASTCALL call_1_13_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_12_2
#define init_1_13_12_2
void FASTCALL call_1_13_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_12_3
#define init_1_13_12_3
void FASTCALL call_1_13_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_12_4
#define init_1_13_12_4
void FASTCALL call_1_13_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_12_5
#define init_1_13_12_5
void FASTCALL call_1_13_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_12_6
#define init_1_13_12_6
void FASTCALL call_1_13_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_12_7
#define init_1_13_12_7
void FASTCALL call_1_13_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_12_8
#define init_1_13_12_8
void FASTCALL call_1_13_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_12_9
#define init_1_13_12_9
void FASTCALL call_1_13_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_12_10
#define init_1_13_12_10
void FASTCALL call_1_13_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_12_11
#define init_1_13_12_11
void FASTCALL call_1_13_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_12_12
#define init_1_13_12_12
void FASTCALL call_1_13_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_12_13
#define init_1_13_12_13
void FASTCALL call_1_13_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_12_14
#define init_1_13_12_14
void FASTCALL call_1_13_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_12_15
#define init_1_13_12_15
void FASTCALL call_1_13_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_13_0
#define init_1_13_13_0
void FASTCALL call_1_13_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_13_1
#define init_1_13_13_1
void FASTCALL call_1_13_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_13_2
#define init_1_13_13_2
void FASTCALL call_1_13_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_13_3
#define init_1_13_13_3
void FASTCALL call_1_13_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_13_4
#define init_1_13_13_4
void FASTCALL call_1_13_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_13_5
#define init_1_13_13_5
void FASTCALL call_1_13_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_13_6
#define init_1_13_13_6
void FASTCALL call_1_13_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_13_7
#define init_1_13_13_7
void FASTCALL call_1_13_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_13_8
#define init_1_13_13_8
void FASTCALL call_1_13_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_13_9
#define init_1_13_13_9
void FASTCALL call_1_13_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_13_10
#define init_1_13_13_10
void FASTCALL call_1_13_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_13_11
#define init_1_13_13_11
void FASTCALL call_1_13_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_13_12
#define init_1_13_13_12
void FASTCALL call_1_13_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_13_13
#define init_1_13_13_13
void FASTCALL call_1_13_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_13_14
#define init_1_13_13_14
void FASTCALL call_1_13_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_13_15
#define init_1_13_13_15
void FASTCALL call_1_13_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_14_0
#define init_1_13_14_0
void FASTCALL call_1_13_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_14_1
#define init_1_13_14_1
void FASTCALL call_1_13_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_14_2
#define init_1_13_14_2
void FASTCALL call_1_13_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_14_3
#define init_1_13_14_3
void FASTCALL call_1_13_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_14_4
#define init_1_13_14_4
void FASTCALL call_1_13_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_14_5
#define init_1_13_14_5
void FASTCALL call_1_13_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_14_6
#define init_1_13_14_6
void FASTCALL call_1_13_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_14_7
#define init_1_13_14_7
void FASTCALL call_1_13_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_14_8
#define init_1_13_14_8
void FASTCALL call_1_13_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_14_9
#define init_1_13_14_9
void FASTCALL call_1_13_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_14_10
#define init_1_13_14_10
void FASTCALL call_1_13_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_14_11
#define init_1_13_14_11
void FASTCALL call_1_13_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_14_12
#define init_1_13_14_12
void FASTCALL call_1_13_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_14_13
#define init_1_13_14_13
void FASTCALL call_1_13_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_14_14
#define init_1_13_14_14
void FASTCALL call_1_13_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_14_15
#define init_1_13_14_15
void FASTCALL call_1_13_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_15_0
#define init_1_13_15_0
void FASTCALL call_1_13_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_15_1
#define init_1_13_15_1
void FASTCALL call_1_13_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_15_2
#define init_1_13_15_2
void FASTCALL call_1_13_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_15_3
#define init_1_13_15_3
void FASTCALL call_1_13_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_15_4
#define init_1_13_15_4
void FASTCALL call_1_13_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_15_5
#define init_1_13_15_5
void FASTCALL call_1_13_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_15_6
#define init_1_13_15_6
void FASTCALL call_1_13_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_15_7
#define init_1_13_15_7
void FASTCALL call_1_13_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_15_8
#define init_1_13_15_8
void FASTCALL call_1_13_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_15_9
#define init_1_13_15_9
void FASTCALL call_1_13_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_15_10
#define init_1_13_15_10
void FASTCALL call_1_13_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_15_11
#define init_1_13_15_11
void FASTCALL call_1_13_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_15_12
#define init_1_13_15_12
void FASTCALL call_1_13_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_15_13
#define init_1_13_15_13
void FASTCALL call_1_13_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_15_14
#define init_1_13_15_14
void FASTCALL call_1_13_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_13_15_15
#define init_1_13_15_15
void FASTCALL call_1_13_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_0_0
#define init_1_14_0_0
void FASTCALL call_1_14_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_0_1
#define init_1_14_0_1
void FASTCALL call_1_14_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_0_2
#define init_1_14_0_2
void FASTCALL call_1_14_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_0_3
#define init_1_14_0_3
void FASTCALL call_1_14_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_0_4
#define init_1_14_0_4
void FASTCALL call_1_14_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_0_5
#define init_1_14_0_5
void FASTCALL call_1_14_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_0_6
#define init_1_14_0_6
void FASTCALL call_1_14_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_0_7
#define init_1_14_0_7
void FASTCALL call_1_14_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_0_8
#define init_1_14_0_8
void FASTCALL call_1_14_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_0_9
#define init_1_14_0_9
void FASTCALL call_1_14_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_0_10
#define init_1_14_0_10
void FASTCALL call_1_14_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_0_11
#define init_1_14_0_11
void FASTCALL call_1_14_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_0_12
#define init_1_14_0_12
void FASTCALL call_1_14_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_0_13
#define init_1_14_0_13
void FASTCALL call_1_14_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_0_14
#define init_1_14_0_14
void FASTCALL call_1_14_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_0_15
#define init_1_14_0_15
void FASTCALL call_1_14_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_1_0
#define init_1_14_1_0
void FASTCALL call_1_14_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_1_1
#define init_1_14_1_1
void FASTCALL call_1_14_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_1_2
#define init_1_14_1_2
void FASTCALL call_1_14_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_1_3
#define init_1_14_1_3
void FASTCALL call_1_14_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_1_4
#define init_1_14_1_4
void FASTCALL call_1_14_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_1_5
#define init_1_14_1_5
void FASTCALL call_1_14_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_1_6
#define init_1_14_1_6
void FASTCALL call_1_14_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_1_7
#define init_1_14_1_7
void FASTCALL call_1_14_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_1_8
#define init_1_14_1_8
void FASTCALL call_1_14_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_1_9
#define init_1_14_1_9
void FASTCALL call_1_14_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_1_10
#define init_1_14_1_10
void FASTCALL call_1_14_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_1_11
#define init_1_14_1_11
void FASTCALL call_1_14_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_1_12
#define init_1_14_1_12
void FASTCALL call_1_14_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_1_13
#define init_1_14_1_13
void FASTCALL call_1_14_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_1_14
#define init_1_14_1_14
void FASTCALL call_1_14_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_1_15
#define init_1_14_1_15
void FASTCALL call_1_14_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_2_0
#define init_1_14_2_0
void FASTCALL call_1_14_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_2_1
#define init_1_14_2_1
void FASTCALL call_1_14_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_2_2
#define init_1_14_2_2
void FASTCALL call_1_14_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_2_3
#define init_1_14_2_3
void FASTCALL call_1_14_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_2_4
#define init_1_14_2_4
void FASTCALL call_1_14_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_2_5
#define init_1_14_2_5
void FASTCALL call_1_14_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_2_6
#define init_1_14_2_6
void FASTCALL call_1_14_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_2_7
#define init_1_14_2_7
void FASTCALL call_1_14_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_2_8
#define init_1_14_2_8
void FASTCALL call_1_14_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_2_9
#define init_1_14_2_9
void FASTCALL call_1_14_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_2_10
#define init_1_14_2_10
void FASTCALL call_1_14_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_2_11
#define init_1_14_2_11
void FASTCALL call_1_14_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_2_12
#define init_1_14_2_12
void FASTCALL call_1_14_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_2_13
#define init_1_14_2_13
void FASTCALL call_1_14_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_2_14
#define init_1_14_2_14
void FASTCALL call_1_14_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_2_15
#define init_1_14_2_15
void FASTCALL call_1_14_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_3_0
#define init_1_14_3_0
void FASTCALL call_1_14_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_3_1
#define init_1_14_3_1
void FASTCALL call_1_14_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_3_2
#define init_1_14_3_2
void FASTCALL call_1_14_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_3_3
#define init_1_14_3_3
void FASTCALL call_1_14_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_3_4
#define init_1_14_3_4
void FASTCALL call_1_14_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_3_5
#define init_1_14_3_5
void FASTCALL call_1_14_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_3_6
#define init_1_14_3_6
void FASTCALL call_1_14_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_3_7
#define init_1_14_3_7
void FASTCALL call_1_14_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_3_8
#define init_1_14_3_8
void FASTCALL call_1_14_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_3_9
#define init_1_14_3_9
void FASTCALL call_1_14_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_3_10
#define init_1_14_3_10
void FASTCALL call_1_14_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_3_11
#define init_1_14_3_11
void FASTCALL call_1_14_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_3_12
#define init_1_14_3_12
void FASTCALL call_1_14_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_3_13
#define init_1_14_3_13
void FASTCALL call_1_14_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_3_14
#define init_1_14_3_14
void FASTCALL call_1_14_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_3_15
#define init_1_14_3_15
void FASTCALL call_1_14_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_4_0
#define init_1_14_4_0
void FASTCALL call_1_14_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_4_1
#define init_1_14_4_1
void FASTCALL call_1_14_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_4_2
#define init_1_14_4_2
void FASTCALL call_1_14_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_4_3
#define init_1_14_4_3
void FASTCALL call_1_14_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_4_4
#define init_1_14_4_4
void FASTCALL call_1_14_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_4_5
#define init_1_14_4_5
void FASTCALL call_1_14_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_4_6
#define init_1_14_4_6
void FASTCALL call_1_14_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_4_7
#define init_1_14_4_7
void FASTCALL call_1_14_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_4_8
#define init_1_14_4_8
void FASTCALL call_1_14_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_4_9
#define init_1_14_4_9
void FASTCALL call_1_14_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_4_10
#define init_1_14_4_10
void FASTCALL call_1_14_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_4_11
#define init_1_14_4_11
void FASTCALL call_1_14_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_4_12
#define init_1_14_4_12
void FASTCALL call_1_14_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_4_13
#define init_1_14_4_13
void FASTCALL call_1_14_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_4_14
#define init_1_14_4_14
void FASTCALL call_1_14_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_4_15
#define init_1_14_4_15
void FASTCALL call_1_14_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_5_0
#define init_1_14_5_0
void FASTCALL call_1_14_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_5_1
#define init_1_14_5_1
void FASTCALL call_1_14_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_5_2
#define init_1_14_5_2
void FASTCALL call_1_14_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_5_3
#define init_1_14_5_3
void FASTCALL call_1_14_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_5_4
#define init_1_14_5_4
void FASTCALL call_1_14_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_5_5
#define init_1_14_5_5
void FASTCALL call_1_14_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_5_6
#define init_1_14_5_6
void FASTCALL call_1_14_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_5_7
#define init_1_14_5_7
void FASTCALL call_1_14_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_5_8
#define init_1_14_5_8
void FASTCALL call_1_14_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_5_9
#define init_1_14_5_9
void FASTCALL call_1_14_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_5_10
#define init_1_14_5_10
void FASTCALL call_1_14_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_5_11
#define init_1_14_5_11
void FASTCALL call_1_14_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_5_12
#define init_1_14_5_12
void FASTCALL call_1_14_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_5_13
#define init_1_14_5_13
void FASTCALL call_1_14_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_5_14
#define init_1_14_5_14
void FASTCALL call_1_14_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_5_15
#define init_1_14_5_15
void FASTCALL call_1_14_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_6_0
#define init_1_14_6_0
void FASTCALL call_1_14_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_6_1
#define init_1_14_6_1
void FASTCALL call_1_14_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_6_2
#define init_1_14_6_2
void FASTCALL call_1_14_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_6_3
#define init_1_14_6_3
void FASTCALL call_1_14_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_6_4
#define init_1_14_6_4
void FASTCALL call_1_14_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_6_5
#define init_1_14_6_5
void FASTCALL call_1_14_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_6_6
#define init_1_14_6_6
void FASTCALL call_1_14_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_6_7
#define init_1_14_6_7
void FASTCALL call_1_14_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_6_8
#define init_1_14_6_8
void FASTCALL call_1_14_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_6_9
#define init_1_14_6_9
void FASTCALL call_1_14_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_6_10
#define init_1_14_6_10
void FASTCALL call_1_14_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_6_11
#define init_1_14_6_11
void FASTCALL call_1_14_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_6_12
#define init_1_14_6_12
void FASTCALL call_1_14_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_6_13
#define init_1_14_6_13
void FASTCALL call_1_14_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_6_14
#define init_1_14_6_14
void FASTCALL call_1_14_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_6_15
#define init_1_14_6_15
void FASTCALL call_1_14_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_7_0
#define init_1_14_7_0
void FASTCALL call_1_14_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_7_1
#define init_1_14_7_1
void FASTCALL call_1_14_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_7_2
#define init_1_14_7_2
void FASTCALL call_1_14_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_7_3
#define init_1_14_7_3
void FASTCALL call_1_14_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_7_4
#define init_1_14_7_4
void FASTCALL call_1_14_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_7_5
#define init_1_14_7_5
void FASTCALL call_1_14_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_7_6
#define init_1_14_7_6
void FASTCALL call_1_14_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_7_7
#define init_1_14_7_7
void FASTCALL call_1_14_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_7_8
#define init_1_14_7_8
void FASTCALL call_1_14_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_7_9
#define init_1_14_7_9
void FASTCALL call_1_14_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_7_10
#define init_1_14_7_10
void FASTCALL call_1_14_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_7_11
#define init_1_14_7_11
void FASTCALL call_1_14_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_7_12
#define init_1_14_7_12
void FASTCALL call_1_14_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_7_13
#define init_1_14_7_13
void FASTCALL call_1_14_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_7_14
#define init_1_14_7_14
void FASTCALL call_1_14_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_7_15
#define init_1_14_7_15
void FASTCALL call_1_14_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_8_0
#define init_1_14_8_0
void FASTCALL call_1_14_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_8_1
#define init_1_14_8_1
void FASTCALL call_1_14_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_8_2
#define init_1_14_8_2
void FASTCALL call_1_14_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_8_3
#define init_1_14_8_3
void FASTCALL call_1_14_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_8_4
#define init_1_14_8_4
void FASTCALL call_1_14_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_8_5
#define init_1_14_8_5
void FASTCALL call_1_14_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_8_6
#define init_1_14_8_6
void FASTCALL call_1_14_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_8_7
#define init_1_14_8_7
void FASTCALL call_1_14_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_8_8
#define init_1_14_8_8
void FASTCALL call_1_14_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_8_9
#define init_1_14_8_9
void FASTCALL call_1_14_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_8_10
#define init_1_14_8_10
void FASTCALL call_1_14_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_8_11
#define init_1_14_8_11
void FASTCALL call_1_14_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_8_12
#define init_1_14_8_12
void FASTCALL call_1_14_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_8_13
#define init_1_14_8_13
void FASTCALL call_1_14_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_8_14
#define init_1_14_8_14
void FASTCALL call_1_14_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_8_15
#define init_1_14_8_15
void FASTCALL call_1_14_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_9_0
#define init_1_14_9_0
void FASTCALL call_1_14_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_9_1
#define init_1_14_9_1
void FASTCALL call_1_14_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_9_2
#define init_1_14_9_2
void FASTCALL call_1_14_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_9_3
#define init_1_14_9_3
void FASTCALL call_1_14_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_9_4
#define init_1_14_9_4
void FASTCALL call_1_14_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_9_5
#define init_1_14_9_5
void FASTCALL call_1_14_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_9_6
#define init_1_14_9_6
void FASTCALL call_1_14_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_9_7
#define init_1_14_9_7
void FASTCALL call_1_14_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_9_8
#define init_1_14_9_8
void FASTCALL call_1_14_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_9_9
#define init_1_14_9_9
void FASTCALL call_1_14_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_9_10
#define init_1_14_9_10
void FASTCALL call_1_14_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_9_11
#define init_1_14_9_11
void FASTCALL call_1_14_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_9_12
#define init_1_14_9_12
void FASTCALL call_1_14_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_9_13
#define init_1_14_9_13
void FASTCALL call_1_14_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_9_14
#define init_1_14_9_14
void FASTCALL call_1_14_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_9_15
#define init_1_14_9_15
void FASTCALL call_1_14_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_10_0
#define init_1_14_10_0
void FASTCALL call_1_14_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_10_1
#define init_1_14_10_1
void FASTCALL call_1_14_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_10_2
#define init_1_14_10_2
void FASTCALL call_1_14_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_10_3
#define init_1_14_10_3
void FASTCALL call_1_14_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_10_4
#define init_1_14_10_4
void FASTCALL call_1_14_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_10_5
#define init_1_14_10_5
void FASTCALL call_1_14_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_10_6
#define init_1_14_10_6
void FASTCALL call_1_14_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_10_7
#define init_1_14_10_7
void FASTCALL call_1_14_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_10_8
#define init_1_14_10_8
void FASTCALL call_1_14_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_10_9
#define init_1_14_10_9
void FASTCALL call_1_14_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_10_10
#define init_1_14_10_10
void FASTCALL call_1_14_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_10_11
#define init_1_14_10_11
void FASTCALL call_1_14_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_10_12
#define init_1_14_10_12
void FASTCALL call_1_14_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_10_13
#define init_1_14_10_13
void FASTCALL call_1_14_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_10_14
#define init_1_14_10_14
void FASTCALL call_1_14_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_10_15
#define init_1_14_10_15
void FASTCALL call_1_14_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_11_0
#define init_1_14_11_0
void FASTCALL call_1_14_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_11_1
#define init_1_14_11_1
void FASTCALL call_1_14_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_11_2
#define init_1_14_11_2
void FASTCALL call_1_14_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_11_3
#define init_1_14_11_3
void FASTCALL call_1_14_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_11_4
#define init_1_14_11_4
void FASTCALL call_1_14_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_11_5
#define init_1_14_11_5
void FASTCALL call_1_14_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_11_6
#define init_1_14_11_6
void FASTCALL call_1_14_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_11_7
#define init_1_14_11_7
void FASTCALL call_1_14_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_11_8
#define init_1_14_11_8
void FASTCALL call_1_14_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_11_9
#define init_1_14_11_9
void FASTCALL call_1_14_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_11_10
#define init_1_14_11_10
void FASTCALL call_1_14_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_11_11
#define init_1_14_11_11
void FASTCALL call_1_14_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_11_12
#define init_1_14_11_12
void FASTCALL call_1_14_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_11_13
#define init_1_14_11_13
void FASTCALL call_1_14_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_11_14
#define init_1_14_11_14
void FASTCALL call_1_14_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_11_15
#define init_1_14_11_15
void FASTCALL call_1_14_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_12_0
#define init_1_14_12_0
void FASTCALL call_1_14_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_12_1
#define init_1_14_12_1
void FASTCALL call_1_14_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_12_2
#define init_1_14_12_2
void FASTCALL call_1_14_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_12_3
#define init_1_14_12_3
void FASTCALL call_1_14_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_12_4
#define init_1_14_12_4
void FASTCALL call_1_14_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_12_5
#define init_1_14_12_5
void FASTCALL call_1_14_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_12_6
#define init_1_14_12_6
void FASTCALL call_1_14_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_12_7
#define init_1_14_12_7
void FASTCALL call_1_14_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_12_8
#define init_1_14_12_8
void FASTCALL call_1_14_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_12_9
#define init_1_14_12_9
void FASTCALL call_1_14_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_12_10
#define init_1_14_12_10
void FASTCALL call_1_14_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_12_11
#define init_1_14_12_11
void FASTCALL call_1_14_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_12_12
#define init_1_14_12_12
void FASTCALL call_1_14_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_12_13
#define init_1_14_12_13
void FASTCALL call_1_14_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_12_14
#define init_1_14_12_14
void FASTCALL call_1_14_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_12_15
#define init_1_14_12_15
void FASTCALL call_1_14_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_13_0
#define init_1_14_13_0
void FASTCALL call_1_14_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_13_1
#define init_1_14_13_1
void FASTCALL call_1_14_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_13_2
#define init_1_14_13_2
void FASTCALL call_1_14_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_13_3
#define init_1_14_13_3
void FASTCALL call_1_14_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_13_4
#define init_1_14_13_4
void FASTCALL call_1_14_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_13_5
#define init_1_14_13_5
void FASTCALL call_1_14_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_13_6
#define init_1_14_13_6
void FASTCALL call_1_14_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_13_7
#define init_1_14_13_7
void FASTCALL call_1_14_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_13_8
#define init_1_14_13_8
void FASTCALL call_1_14_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_13_9
#define init_1_14_13_9
void FASTCALL call_1_14_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_13_10
#define init_1_14_13_10
void FASTCALL call_1_14_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_13_11
#define init_1_14_13_11
void FASTCALL call_1_14_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_13_12
#define init_1_14_13_12
void FASTCALL call_1_14_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_13_13
#define init_1_14_13_13
void FASTCALL call_1_14_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_13_14
#define init_1_14_13_14
void FASTCALL call_1_14_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_13_15
#define init_1_14_13_15
void FASTCALL call_1_14_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_14_0
#define init_1_14_14_0
void FASTCALL call_1_14_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_14_1
#define init_1_14_14_1
void FASTCALL call_1_14_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_14_2
#define init_1_14_14_2
void FASTCALL call_1_14_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_14_3
#define init_1_14_14_3
void FASTCALL call_1_14_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_14_4
#define init_1_14_14_4
void FASTCALL call_1_14_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_14_5
#define init_1_14_14_5
void FASTCALL call_1_14_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_14_6
#define init_1_14_14_6
void FASTCALL call_1_14_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_14_7
#define init_1_14_14_7
void FASTCALL call_1_14_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_14_8
#define init_1_14_14_8
void FASTCALL call_1_14_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_14_9
#define init_1_14_14_9
void FASTCALL call_1_14_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_14_10
#define init_1_14_14_10
void FASTCALL call_1_14_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_14_11
#define init_1_14_14_11
void FASTCALL call_1_14_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_14_12
#define init_1_14_14_12
void FASTCALL call_1_14_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_14_13
#define init_1_14_14_13
void FASTCALL call_1_14_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_14_14
#define init_1_14_14_14
void FASTCALL call_1_14_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_14_15
#define init_1_14_14_15
void FASTCALL call_1_14_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_15_0
#define init_1_14_15_0
void FASTCALL call_1_14_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_15_1
#define init_1_14_15_1
void FASTCALL call_1_14_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_15_2
#define init_1_14_15_2
void FASTCALL call_1_14_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_15_3
#define init_1_14_15_3
void FASTCALL call_1_14_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_15_4
#define init_1_14_15_4
void FASTCALL call_1_14_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_15_5
#define init_1_14_15_5
void FASTCALL call_1_14_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_15_6
#define init_1_14_15_6
void FASTCALL call_1_14_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_15_7
#define init_1_14_15_7
void FASTCALL call_1_14_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_15_8
#define init_1_14_15_8
void FASTCALL call_1_14_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_15_9
#define init_1_14_15_9
void FASTCALL call_1_14_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_15_10
#define init_1_14_15_10
void FASTCALL call_1_14_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_15_11
#define init_1_14_15_11
void FASTCALL call_1_14_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_15_12
#define init_1_14_15_12
void FASTCALL call_1_14_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_15_13
#define init_1_14_15_13
void FASTCALL call_1_14_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_15_14
#define init_1_14_15_14
void FASTCALL call_1_14_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_14_15_15
#define init_1_14_15_15
void FASTCALL call_1_14_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_0_0
#define init_1_15_0_0
void FASTCALL call_1_15_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_0_1
#define init_1_15_0_1
void FASTCALL call_1_15_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_0_2
#define init_1_15_0_2
void FASTCALL call_1_15_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_0_3
#define init_1_15_0_3
void FASTCALL call_1_15_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_0_4
#define init_1_15_0_4
void FASTCALL call_1_15_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_0_5
#define init_1_15_0_5
void FASTCALL call_1_15_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_0_6
#define init_1_15_0_6
void FASTCALL call_1_15_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_0_7
#define init_1_15_0_7
void FASTCALL call_1_15_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_0_8
#define init_1_15_0_8
void FASTCALL call_1_15_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_0_9
#define init_1_15_0_9
void FASTCALL call_1_15_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_0_10
#define init_1_15_0_10
void FASTCALL call_1_15_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_0_11
#define init_1_15_0_11
void FASTCALL call_1_15_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_0_12
#define init_1_15_0_12
void FASTCALL call_1_15_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_0_13
#define init_1_15_0_13
void FASTCALL call_1_15_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_0_14
#define init_1_15_0_14
void FASTCALL call_1_15_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_0_15
#define init_1_15_0_15
void FASTCALL call_1_15_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_1_0
#define init_1_15_1_0
void FASTCALL call_1_15_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_1_1
#define init_1_15_1_1
void FASTCALL call_1_15_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_1_2
#define init_1_15_1_2
void FASTCALL call_1_15_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_1_3
#define init_1_15_1_3
void FASTCALL call_1_15_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_1_4
#define init_1_15_1_4
void FASTCALL call_1_15_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_1_5
#define init_1_15_1_5
void FASTCALL call_1_15_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_1_6
#define init_1_15_1_6
void FASTCALL call_1_15_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_1_7
#define init_1_15_1_7
void FASTCALL call_1_15_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_1_8
#define init_1_15_1_8
void FASTCALL call_1_15_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_1_9
#define init_1_15_1_9
void FASTCALL call_1_15_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_1_10
#define init_1_15_1_10
void FASTCALL call_1_15_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_1_11
#define init_1_15_1_11
void FASTCALL call_1_15_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_1_12
#define init_1_15_1_12
void FASTCALL call_1_15_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_1_13
#define init_1_15_1_13
void FASTCALL call_1_15_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_1_14
#define init_1_15_1_14
void FASTCALL call_1_15_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_1_15
#define init_1_15_1_15
void FASTCALL call_1_15_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_2_0
#define init_1_15_2_0
void FASTCALL call_1_15_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_2_1
#define init_1_15_2_1
void FASTCALL call_1_15_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_2_2
#define init_1_15_2_2
void FASTCALL call_1_15_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_2_3
#define init_1_15_2_3
void FASTCALL call_1_15_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_2_4
#define init_1_15_2_4
void FASTCALL call_1_15_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_2_5
#define init_1_15_2_5
void FASTCALL call_1_15_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_2_6
#define init_1_15_2_6
void FASTCALL call_1_15_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_2_7
#define init_1_15_2_7
void FASTCALL call_1_15_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_2_8
#define init_1_15_2_8
void FASTCALL call_1_15_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_2_9
#define init_1_15_2_9
void FASTCALL call_1_15_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_2_10
#define init_1_15_2_10
void FASTCALL call_1_15_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_2_11
#define init_1_15_2_11
void FASTCALL call_1_15_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_2_12
#define init_1_15_2_12
void FASTCALL call_1_15_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_2_13
#define init_1_15_2_13
void FASTCALL call_1_15_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_2_14
#define init_1_15_2_14
void FASTCALL call_1_15_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_2_15
#define init_1_15_2_15
void FASTCALL call_1_15_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_3_0
#define init_1_15_3_0
void FASTCALL call_1_15_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_3_1
#define init_1_15_3_1
void FASTCALL call_1_15_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_3_2
#define init_1_15_3_2
void FASTCALL call_1_15_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_3_3
#define init_1_15_3_3
void FASTCALL call_1_15_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_3_4
#define init_1_15_3_4
void FASTCALL call_1_15_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_3_5
#define init_1_15_3_5
void FASTCALL call_1_15_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_3_6
#define init_1_15_3_6
void FASTCALL call_1_15_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_3_7
#define init_1_15_3_7
void FASTCALL call_1_15_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_3_8
#define init_1_15_3_8
void FASTCALL call_1_15_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_3_9
#define init_1_15_3_9
void FASTCALL call_1_15_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_3_10
#define init_1_15_3_10
void FASTCALL call_1_15_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_3_11
#define init_1_15_3_11
void FASTCALL call_1_15_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_3_12
#define init_1_15_3_12
void FASTCALL call_1_15_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_3_13
#define init_1_15_3_13
void FASTCALL call_1_15_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_3_14
#define init_1_15_3_14
void FASTCALL call_1_15_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_3_15
#define init_1_15_3_15
void FASTCALL call_1_15_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_4_0
#define init_1_15_4_0
void FASTCALL call_1_15_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_4_1
#define init_1_15_4_1
void FASTCALL call_1_15_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_4_2
#define init_1_15_4_2
void FASTCALL call_1_15_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_4_3
#define init_1_15_4_3
void FASTCALL call_1_15_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_4_4
#define init_1_15_4_4
void FASTCALL call_1_15_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_4_5
#define init_1_15_4_5
void FASTCALL call_1_15_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_4_6
#define init_1_15_4_6
void FASTCALL call_1_15_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_4_7
#define init_1_15_4_7
void FASTCALL call_1_15_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_4_8
#define init_1_15_4_8
void FASTCALL call_1_15_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_4_9
#define init_1_15_4_9
void FASTCALL call_1_15_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_4_10
#define init_1_15_4_10
void FASTCALL call_1_15_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_4_11
#define init_1_15_4_11
void FASTCALL call_1_15_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_4_12
#define init_1_15_4_12
void FASTCALL call_1_15_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_4_13
#define init_1_15_4_13
void FASTCALL call_1_15_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_4_14
#define init_1_15_4_14
void FASTCALL call_1_15_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_4_15
#define init_1_15_4_15
void FASTCALL call_1_15_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_5_0
#define init_1_15_5_0
void FASTCALL call_1_15_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_5_1
#define init_1_15_5_1
void FASTCALL call_1_15_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_5_2
#define init_1_15_5_2
void FASTCALL call_1_15_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_5_3
#define init_1_15_5_3
void FASTCALL call_1_15_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_5_4
#define init_1_15_5_4
void FASTCALL call_1_15_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_5_5
#define init_1_15_5_5
void FASTCALL call_1_15_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_5_6
#define init_1_15_5_6
void FASTCALL call_1_15_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_5_7
#define init_1_15_5_7
void FASTCALL call_1_15_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_5_8
#define init_1_15_5_8
void FASTCALL call_1_15_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_5_9
#define init_1_15_5_9
void FASTCALL call_1_15_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_5_10
#define init_1_15_5_10
void FASTCALL call_1_15_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_5_11
#define init_1_15_5_11
void FASTCALL call_1_15_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_5_12
#define init_1_15_5_12
void FASTCALL call_1_15_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_5_13
#define init_1_15_5_13
void FASTCALL call_1_15_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_5_14
#define init_1_15_5_14
void FASTCALL call_1_15_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_5_15
#define init_1_15_5_15
void FASTCALL call_1_15_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_6_0
#define init_1_15_6_0
void FASTCALL call_1_15_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_6_1
#define init_1_15_6_1
void FASTCALL call_1_15_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_6_2
#define init_1_15_6_2
void FASTCALL call_1_15_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_6_3
#define init_1_15_6_3
void FASTCALL call_1_15_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_6_4
#define init_1_15_6_4
void FASTCALL call_1_15_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_6_5
#define init_1_15_6_5
void FASTCALL call_1_15_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_6_6
#define init_1_15_6_6
void FASTCALL call_1_15_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_6_7
#define init_1_15_6_7
void FASTCALL call_1_15_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_6_8
#define init_1_15_6_8
void FASTCALL call_1_15_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_6_9
#define init_1_15_6_9
void FASTCALL call_1_15_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_6_10
#define init_1_15_6_10
void FASTCALL call_1_15_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_6_11
#define init_1_15_6_11
void FASTCALL call_1_15_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_6_12
#define init_1_15_6_12
void FASTCALL call_1_15_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_6_13
#define init_1_15_6_13
void FASTCALL call_1_15_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_6_14
#define init_1_15_6_14
void FASTCALL call_1_15_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_6_15
#define init_1_15_6_15
void FASTCALL call_1_15_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_7_0
#define init_1_15_7_0
void FASTCALL call_1_15_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_7_1
#define init_1_15_7_1
void FASTCALL call_1_15_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_7_2
#define init_1_15_7_2
void FASTCALL call_1_15_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_7_3
#define init_1_15_7_3
void FASTCALL call_1_15_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_7_4
#define init_1_15_7_4
void FASTCALL call_1_15_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_7_5
#define init_1_15_7_5
void FASTCALL call_1_15_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_7_6
#define init_1_15_7_6
void FASTCALL call_1_15_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_7_7
#define init_1_15_7_7
void FASTCALL call_1_15_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_7_8
#define init_1_15_7_8
void FASTCALL call_1_15_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_7_9
#define init_1_15_7_9
void FASTCALL call_1_15_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_7_10
#define init_1_15_7_10
void FASTCALL call_1_15_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_7_11
#define init_1_15_7_11
void FASTCALL call_1_15_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_7_12
#define init_1_15_7_12
void FASTCALL call_1_15_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_7_13
#define init_1_15_7_13
void FASTCALL call_1_15_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_7_14
#define init_1_15_7_14
void FASTCALL call_1_15_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_7_15
#define init_1_15_7_15
void FASTCALL call_1_15_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_8_0
#define init_1_15_8_0
void FASTCALL call_1_15_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_8_1
#define init_1_15_8_1
void FASTCALL call_1_15_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_8_2
#define init_1_15_8_2
void FASTCALL call_1_15_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_8_3
#define init_1_15_8_3
void FASTCALL call_1_15_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_8_4
#define init_1_15_8_4
void FASTCALL call_1_15_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_8_5
#define init_1_15_8_5
void FASTCALL call_1_15_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_8_6
#define init_1_15_8_6
void FASTCALL call_1_15_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_8_7
#define init_1_15_8_7
void FASTCALL call_1_15_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_8_8
#define init_1_15_8_8
void FASTCALL call_1_15_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_8_9
#define init_1_15_8_9
void FASTCALL call_1_15_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_8_10
#define init_1_15_8_10
void FASTCALL call_1_15_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_8_11
#define init_1_15_8_11
void FASTCALL call_1_15_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_8_12
#define init_1_15_8_12
void FASTCALL call_1_15_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_8_13
#define init_1_15_8_13
void FASTCALL call_1_15_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_8_14
#define init_1_15_8_14
void FASTCALL call_1_15_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_8_15
#define init_1_15_8_15
void FASTCALL call_1_15_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_9_0
#define init_1_15_9_0
void FASTCALL call_1_15_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_9_1
#define init_1_15_9_1
void FASTCALL call_1_15_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_9_2
#define init_1_15_9_2
void FASTCALL call_1_15_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_9_3
#define init_1_15_9_3
void FASTCALL call_1_15_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_9_4
#define init_1_15_9_4
void FASTCALL call_1_15_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_9_5
#define init_1_15_9_5
void FASTCALL call_1_15_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_9_6
#define init_1_15_9_6
void FASTCALL call_1_15_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_9_7
#define init_1_15_9_7
void FASTCALL call_1_15_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_9_8
#define init_1_15_9_8
void FASTCALL call_1_15_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_9_9
#define init_1_15_9_9
void FASTCALL call_1_15_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_9_10
#define init_1_15_9_10
void FASTCALL call_1_15_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_9_11
#define init_1_15_9_11
void FASTCALL call_1_15_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_9_12
#define init_1_15_9_12
void FASTCALL call_1_15_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_9_13
#define init_1_15_9_13
void FASTCALL call_1_15_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_9_14
#define init_1_15_9_14
void FASTCALL call_1_15_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_9_15
#define init_1_15_9_15
void FASTCALL call_1_15_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_10_0
#define init_1_15_10_0
void FASTCALL call_1_15_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_10_1
#define init_1_15_10_1
void FASTCALL call_1_15_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_10_2
#define init_1_15_10_2
void FASTCALL call_1_15_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_10_3
#define init_1_15_10_3
void FASTCALL call_1_15_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_10_4
#define init_1_15_10_4
void FASTCALL call_1_15_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_10_5
#define init_1_15_10_5
void FASTCALL call_1_15_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_10_6
#define init_1_15_10_6
void FASTCALL call_1_15_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_10_7
#define init_1_15_10_7
void FASTCALL call_1_15_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_10_8
#define init_1_15_10_8
void FASTCALL call_1_15_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_10_9
#define init_1_15_10_9
void FASTCALL call_1_15_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_10_10
#define init_1_15_10_10
void FASTCALL call_1_15_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_10_11
#define init_1_15_10_11
void FASTCALL call_1_15_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_10_12
#define init_1_15_10_12
void FASTCALL call_1_15_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_10_13
#define init_1_15_10_13
void FASTCALL call_1_15_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_10_14
#define init_1_15_10_14
void FASTCALL call_1_15_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_10_15
#define init_1_15_10_15
void FASTCALL call_1_15_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_11_0
#define init_1_15_11_0
void FASTCALL call_1_15_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_11_1
#define init_1_15_11_1
void FASTCALL call_1_15_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_11_2
#define init_1_15_11_2
void FASTCALL call_1_15_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_11_3
#define init_1_15_11_3
void FASTCALL call_1_15_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_11_4
#define init_1_15_11_4
void FASTCALL call_1_15_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_11_5
#define init_1_15_11_5
void FASTCALL call_1_15_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_11_6
#define init_1_15_11_6
void FASTCALL call_1_15_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_11_7
#define init_1_15_11_7
void FASTCALL call_1_15_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_11_8
#define init_1_15_11_8
void FASTCALL call_1_15_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_11_9
#define init_1_15_11_9
void FASTCALL call_1_15_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_11_10
#define init_1_15_11_10
void FASTCALL call_1_15_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_11_11
#define init_1_15_11_11
void FASTCALL call_1_15_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_11_12
#define init_1_15_11_12
void FASTCALL call_1_15_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_11_13
#define init_1_15_11_13
void FASTCALL call_1_15_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_11_14
#define init_1_15_11_14
void FASTCALL call_1_15_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_11_15
#define init_1_15_11_15
void FASTCALL call_1_15_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_12_0
#define init_1_15_12_0
void FASTCALL call_1_15_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_12_1
#define init_1_15_12_1
void FASTCALL call_1_15_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_12_2
#define init_1_15_12_2
void FASTCALL call_1_15_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_12_3
#define init_1_15_12_3
void FASTCALL call_1_15_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_12_4
#define init_1_15_12_4
void FASTCALL call_1_15_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_12_5
#define init_1_15_12_5
void FASTCALL call_1_15_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_12_6
#define init_1_15_12_6
void FASTCALL call_1_15_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_12_7
#define init_1_15_12_7
void FASTCALL call_1_15_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_12_8
#define init_1_15_12_8
void FASTCALL call_1_15_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_12_9
#define init_1_15_12_9
void FASTCALL call_1_15_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_12_10
#define init_1_15_12_10
void FASTCALL call_1_15_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_12_11
#define init_1_15_12_11
void FASTCALL call_1_15_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_12_12
#define init_1_15_12_12
void FASTCALL call_1_15_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_12_13
#define init_1_15_12_13
void FASTCALL call_1_15_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_12_14
#define init_1_15_12_14
void FASTCALL call_1_15_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_12_15
#define init_1_15_12_15
void FASTCALL call_1_15_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_13_0
#define init_1_15_13_0
void FASTCALL call_1_15_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_13_1
#define init_1_15_13_1
void FASTCALL call_1_15_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_13_2
#define init_1_15_13_2
void FASTCALL call_1_15_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_13_3
#define init_1_15_13_3
void FASTCALL call_1_15_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_13_4
#define init_1_15_13_4
void FASTCALL call_1_15_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_13_5
#define init_1_15_13_5
void FASTCALL call_1_15_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_13_6
#define init_1_15_13_6
void FASTCALL call_1_15_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_13_7
#define init_1_15_13_7
void FASTCALL call_1_15_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_13_8
#define init_1_15_13_8
void FASTCALL call_1_15_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_13_9
#define init_1_15_13_9
void FASTCALL call_1_15_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_13_10
#define init_1_15_13_10
void FASTCALL call_1_15_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_13_11
#define init_1_15_13_11
void FASTCALL call_1_15_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_13_12
#define init_1_15_13_12
void FASTCALL call_1_15_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_13_13
#define init_1_15_13_13
void FASTCALL call_1_15_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_13_14
#define init_1_15_13_14
void FASTCALL call_1_15_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_13_15
#define init_1_15_13_15
void FASTCALL call_1_15_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_14_0
#define init_1_15_14_0
void FASTCALL call_1_15_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_14_1
#define init_1_15_14_1
void FASTCALL call_1_15_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_14_2
#define init_1_15_14_2
void FASTCALL call_1_15_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_14_3
#define init_1_15_14_3
void FASTCALL call_1_15_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_14_4
#define init_1_15_14_4
void FASTCALL call_1_15_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_14_5
#define init_1_15_14_5
void FASTCALL call_1_15_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_14_6
#define init_1_15_14_6
void FASTCALL call_1_15_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_14_7
#define init_1_15_14_7
void FASTCALL call_1_15_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_14_8
#define init_1_15_14_8
void FASTCALL call_1_15_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_14_9
#define init_1_15_14_9
void FASTCALL call_1_15_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_14_10
#define init_1_15_14_10
void FASTCALL call_1_15_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_14_11
#define init_1_15_14_11
void FASTCALL call_1_15_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_14_12
#define init_1_15_14_12
void FASTCALL call_1_15_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_14_13
#define init_1_15_14_13
void FASTCALL call_1_15_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_14_14
#define init_1_15_14_14
void FASTCALL call_1_15_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_14_15
#define init_1_15_14_15
void FASTCALL call_1_15_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_15_0
#define init_1_15_15_0
void FASTCALL call_1_15_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_15_1
#define init_1_15_15_1
void FASTCALL call_1_15_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_15_2
#define init_1_15_15_2
void FASTCALL call_1_15_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_15_3
#define init_1_15_15_3
void FASTCALL call_1_15_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_15_4
#define init_1_15_15_4
void FASTCALL call_1_15_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_15_5
#define init_1_15_15_5
void FASTCALL call_1_15_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_15_6
#define init_1_15_15_6
void FASTCALL call_1_15_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_15_7
#define init_1_15_15_7
void FASTCALL call_1_15_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_15_8
#define init_1_15_15_8
void FASTCALL call_1_15_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_15_9
#define init_1_15_15_9
void FASTCALL call_1_15_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_15_10
#define init_1_15_15_10
void FASTCALL call_1_15_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_15_11
#define init_1_15_15_11
void FASTCALL call_1_15_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_15_12
#define init_1_15_15_12
void FASTCALL call_1_15_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_15_13
#define init_1_15_15_13
void FASTCALL call_1_15_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_15_14
#define init_1_15_15_14
void FASTCALL call_1_15_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_1_15_15_15
#define init_1_15_15_15
void FASTCALL call_1_15_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_0_0
#define init_2_0_0_0
void FASTCALL call_2_0_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_0_1
#define init_2_0_0_1
void FASTCALL call_2_0_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_0_2
#define init_2_0_0_2
void FASTCALL call_2_0_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_0_3
#define init_2_0_0_3
void FASTCALL call_2_0_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_0_4
#define init_2_0_0_4
void FASTCALL call_2_0_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_0_5
#define init_2_0_0_5
void FASTCALL call_2_0_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_0_6
#define init_2_0_0_6
void FASTCALL call_2_0_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_0_7
#define init_2_0_0_7
void FASTCALL call_2_0_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_0_8
#define init_2_0_0_8
void FASTCALL call_2_0_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_0_9
#define init_2_0_0_9
void FASTCALL call_2_0_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_0_10
#define init_2_0_0_10
void FASTCALL call_2_0_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_0_11
#define init_2_0_0_11
void FASTCALL call_2_0_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_0_12
#define init_2_0_0_12
void FASTCALL call_2_0_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_0_13
#define init_2_0_0_13
void FASTCALL call_2_0_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_0_14
#define init_2_0_0_14
void FASTCALL call_2_0_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_0_15
#define init_2_0_0_15
void FASTCALL call_2_0_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_1_0
#define init_2_0_1_0
void FASTCALL call_2_0_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_1_1
#define init_2_0_1_1
void FASTCALL call_2_0_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_1_2
#define init_2_0_1_2
void FASTCALL call_2_0_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_1_3
#define init_2_0_1_3
void FASTCALL call_2_0_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_1_4
#define init_2_0_1_4
void FASTCALL call_2_0_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_1_5
#define init_2_0_1_5
void FASTCALL call_2_0_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_1_6
#define init_2_0_1_6
void FASTCALL call_2_0_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_1_7
#define init_2_0_1_7
void FASTCALL call_2_0_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_1_8
#define init_2_0_1_8
void FASTCALL call_2_0_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_1_9
#define init_2_0_1_9
void FASTCALL call_2_0_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_1_10
#define init_2_0_1_10
void FASTCALL call_2_0_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_1_11
#define init_2_0_1_11
void FASTCALL call_2_0_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_1_12
#define init_2_0_1_12
void FASTCALL call_2_0_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_1_13
#define init_2_0_1_13
void FASTCALL call_2_0_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_1_14
#define init_2_0_1_14
void FASTCALL call_2_0_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_1_15
#define init_2_0_1_15
void FASTCALL call_2_0_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_2_0
#define init_2_0_2_0
void FASTCALL call_2_0_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_2_1
#define init_2_0_2_1
void FASTCALL call_2_0_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_2_2
#define init_2_0_2_2
void FASTCALL call_2_0_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_2_3
#define init_2_0_2_3
void FASTCALL call_2_0_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_2_4
#define init_2_0_2_4
void FASTCALL call_2_0_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_2_5
#define init_2_0_2_5
void FASTCALL call_2_0_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_2_6
#define init_2_0_2_6
void FASTCALL call_2_0_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_2_7
#define init_2_0_2_7
void FASTCALL call_2_0_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_2_8
#define init_2_0_2_8
void FASTCALL call_2_0_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_2_9
#define init_2_0_2_9
void FASTCALL call_2_0_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_2_10
#define init_2_0_2_10
void FASTCALL call_2_0_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_2_11
#define init_2_0_2_11
void FASTCALL call_2_0_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_2_12
#define init_2_0_2_12
void FASTCALL call_2_0_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_2_13
#define init_2_0_2_13
void FASTCALL call_2_0_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_2_14
#define init_2_0_2_14
void FASTCALL call_2_0_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_2_15
#define init_2_0_2_15
void FASTCALL call_2_0_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_3_0
#define init_2_0_3_0
void FASTCALL call_2_0_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_3_1
#define init_2_0_3_1
void FASTCALL call_2_0_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_3_2
#define init_2_0_3_2
void FASTCALL call_2_0_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_3_3
#define init_2_0_3_3
void FASTCALL call_2_0_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_3_4
#define init_2_0_3_4
void FASTCALL call_2_0_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_3_5
#define init_2_0_3_5
void FASTCALL call_2_0_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_3_6
#define init_2_0_3_6
void FASTCALL call_2_0_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_3_7
#define init_2_0_3_7
void FASTCALL call_2_0_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_3_8
#define init_2_0_3_8
void FASTCALL call_2_0_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_3_9
#define init_2_0_3_9
void FASTCALL call_2_0_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_3_10
#define init_2_0_3_10
void FASTCALL call_2_0_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_3_11
#define init_2_0_3_11
void FASTCALL call_2_0_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_3_12
#define init_2_0_3_12
void FASTCALL call_2_0_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_3_13
#define init_2_0_3_13
void FASTCALL call_2_0_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_3_14
#define init_2_0_3_14
void FASTCALL call_2_0_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_3_15
#define init_2_0_3_15
void FASTCALL call_2_0_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_4_0
#define init_2_0_4_0
void FASTCALL call_2_0_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_4_1
#define init_2_0_4_1
void FASTCALL call_2_0_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_4_2
#define init_2_0_4_2
void FASTCALL call_2_0_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_4_3
#define init_2_0_4_3
void FASTCALL call_2_0_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_4_4
#define init_2_0_4_4
void FASTCALL call_2_0_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_4_5
#define init_2_0_4_5
void FASTCALL call_2_0_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_4_6
#define init_2_0_4_6
void FASTCALL call_2_0_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_4_7
#define init_2_0_4_7
void FASTCALL call_2_0_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_4_8
#define init_2_0_4_8
void FASTCALL call_2_0_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_4_9
#define init_2_0_4_9
void FASTCALL call_2_0_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_4_10
#define init_2_0_4_10
void FASTCALL call_2_0_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_4_11
#define init_2_0_4_11
void FASTCALL call_2_0_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_4_12
#define init_2_0_4_12
void FASTCALL call_2_0_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_4_13
#define init_2_0_4_13
void FASTCALL call_2_0_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_4_14
#define init_2_0_4_14
void FASTCALL call_2_0_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_4_15
#define init_2_0_4_15
void FASTCALL call_2_0_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_5_0
#define init_2_0_5_0
void FASTCALL call_2_0_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_5_1
#define init_2_0_5_1
void FASTCALL call_2_0_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_5_2
#define init_2_0_5_2
void FASTCALL call_2_0_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_5_3
#define init_2_0_5_3
void FASTCALL call_2_0_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_5_4
#define init_2_0_5_4
void FASTCALL call_2_0_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_5_5
#define init_2_0_5_5
void FASTCALL call_2_0_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_5_6
#define init_2_0_5_6
void FASTCALL call_2_0_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_5_7
#define init_2_0_5_7
void FASTCALL call_2_0_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_5_8
#define init_2_0_5_8
void FASTCALL call_2_0_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_5_9
#define init_2_0_5_9
void FASTCALL call_2_0_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_5_10
#define init_2_0_5_10
void FASTCALL call_2_0_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_5_11
#define init_2_0_5_11
void FASTCALL call_2_0_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_5_12
#define init_2_0_5_12
void FASTCALL call_2_0_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_5_13
#define init_2_0_5_13
void FASTCALL call_2_0_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_5_14
#define init_2_0_5_14
void FASTCALL call_2_0_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_5_15
#define init_2_0_5_15
void FASTCALL call_2_0_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_6_0
#define init_2_0_6_0
void FASTCALL call_2_0_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_6_1
#define init_2_0_6_1
void FASTCALL call_2_0_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_6_2
#define init_2_0_6_2
void FASTCALL call_2_0_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_6_3
#define init_2_0_6_3
void FASTCALL call_2_0_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_6_4
#define init_2_0_6_4
void FASTCALL call_2_0_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_6_5
#define init_2_0_6_5
void FASTCALL call_2_0_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_6_6
#define init_2_0_6_6
void FASTCALL call_2_0_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_6_7
#define init_2_0_6_7
void FASTCALL call_2_0_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_6_8
#define init_2_0_6_8
void FASTCALL call_2_0_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_6_9
#define init_2_0_6_9
void FASTCALL call_2_0_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_6_10
#define init_2_0_6_10
void FASTCALL call_2_0_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_6_11
#define init_2_0_6_11
void FASTCALL call_2_0_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_6_12
#define init_2_0_6_12
void FASTCALL call_2_0_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_6_13
#define init_2_0_6_13
void FASTCALL call_2_0_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_6_14
#define init_2_0_6_14
void FASTCALL call_2_0_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_6_15
#define init_2_0_6_15
void FASTCALL call_2_0_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_7_0
#define init_2_0_7_0
void FASTCALL call_2_0_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_7_1
#define init_2_0_7_1
void FASTCALL call_2_0_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_7_2
#define init_2_0_7_2
void FASTCALL call_2_0_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_7_3
#define init_2_0_7_3
void FASTCALL call_2_0_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_7_4
#define init_2_0_7_4
void FASTCALL call_2_0_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_7_5
#define init_2_0_7_5
void FASTCALL call_2_0_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_7_6
#define init_2_0_7_6
void FASTCALL call_2_0_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_7_7
#define init_2_0_7_7
void FASTCALL call_2_0_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_7_8
#define init_2_0_7_8
void FASTCALL call_2_0_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_7_9
#define init_2_0_7_9
void FASTCALL call_2_0_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_7_10
#define init_2_0_7_10
void FASTCALL call_2_0_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_7_11
#define init_2_0_7_11
void FASTCALL call_2_0_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_7_12
#define init_2_0_7_12
void FASTCALL call_2_0_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_7_13
#define init_2_0_7_13
void FASTCALL call_2_0_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_7_14
#define init_2_0_7_14
void FASTCALL call_2_0_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_7_15
#define init_2_0_7_15
void FASTCALL call_2_0_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_8_0
#define init_2_0_8_0
void FASTCALL call_2_0_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_8_1
#define init_2_0_8_1
void FASTCALL call_2_0_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_8_2
#define init_2_0_8_2
void FASTCALL call_2_0_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_8_3
#define init_2_0_8_3
void FASTCALL call_2_0_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_8_4
#define init_2_0_8_4
void FASTCALL call_2_0_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_8_5
#define init_2_0_8_5
void FASTCALL call_2_0_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_8_6
#define init_2_0_8_6
void FASTCALL call_2_0_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_8_7
#define init_2_0_8_7
void FASTCALL call_2_0_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_8_8
#define init_2_0_8_8
void FASTCALL call_2_0_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_8_9
#define init_2_0_8_9
void FASTCALL call_2_0_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_8_10
#define init_2_0_8_10
void FASTCALL call_2_0_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_8_11
#define init_2_0_8_11
void FASTCALL call_2_0_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_8_12
#define init_2_0_8_12
void FASTCALL call_2_0_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_8_13
#define init_2_0_8_13
void FASTCALL call_2_0_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_8_14
#define init_2_0_8_14
void FASTCALL call_2_0_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_8_15
#define init_2_0_8_15
void FASTCALL call_2_0_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_9_0
#define init_2_0_9_0
void FASTCALL call_2_0_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_9_1
#define init_2_0_9_1
void FASTCALL call_2_0_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_9_2
#define init_2_0_9_2
void FASTCALL call_2_0_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_9_3
#define init_2_0_9_3
void FASTCALL call_2_0_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_9_4
#define init_2_0_9_4
void FASTCALL call_2_0_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_9_5
#define init_2_0_9_5
void FASTCALL call_2_0_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_9_6
#define init_2_0_9_6
void FASTCALL call_2_0_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_9_7
#define init_2_0_9_7
void FASTCALL call_2_0_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_9_8
#define init_2_0_9_8
void FASTCALL call_2_0_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_9_9
#define init_2_0_9_9
void FASTCALL call_2_0_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_9_10
#define init_2_0_9_10
void FASTCALL call_2_0_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_9_11
#define init_2_0_9_11
void FASTCALL call_2_0_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_9_12
#define init_2_0_9_12
void FASTCALL call_2_0_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_9_13
#define init_2_0_9_13
void FASTCALL call_2_0_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_9_14
#define init_2_0_9_14
void FASTCALL call_2_0_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_9_15
#define init_2_0_9_15
void FASTCALL call_2_0_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_10_0
#define init_2_0_10_0
void FASTCALL call_2_0_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_10_1
#define init_2_0_10_1
void FASTCALL call_2_0_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_10_2
#define init_2_0_10_2
void FASTCALL call_2_0_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_10_3
#define init_2_0_10_3
void FASTCALL call_2_0_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_10_4
#define init_2_0_10_4
void FASTCALL call_2_0_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_10_5
#define init_2_0_10_5
void FASTCALL call_2_0_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_10_6
#define init_2_0_10_6
void FASTCALL call_2_0_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_10_7
#define init_2_0_10_7
void FASTCALL call_2_0_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_10_8
#define init_2_0_10_8
void FASTCALL call_2_0_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_10_9
#define init_2_0_10_9
void FASTCALL call_2_0_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_10_10
#define init_2_0_10_10
void FASTCALL call_2_0_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_10_11
#define init_2_0_10_11
void FASTCALL call_2_0_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_10_12
#define init_2_0_10_12
void FASTCALL call_2_0_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_10_13
#define init_2_0_10_13
void FASTCALL call_2_0_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_10_14
#define init_2_0_10_14
void FASTCALL call_2_0_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_10_15
#define init_2_0_10_15
void FASTCALL call_2_0_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_11_0
#define init_2_0_11_0
void FASTCALL call_2_0_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_11_1
#define init_2_0_11_1
void FASTCALL call_2_0_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_11_2
#define init_2_0_11_2
void FASTCALL call_2_0_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_11_3
#define init_2_0_11_3
void FASTCALL call_2_0_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_11_4
#define init_2_0_11_4
void FASTCALL call_2_0_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_11_5
#define init_2_0_11_5
void FASTCALL call_2_0_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_11_6
#define init_2_0_11_6
void FASTCALL call_2_0_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_11_7
#define init_2_0_11_7
void FASTCALL call_2_0_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_11_8
#define init_2_0_11_8
void FASTCALL call_2_0_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_11_9
#define init_2_0_11_9
void FASTCALL call_2_0_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_11_10
#define init_2_0_11_10
void FASTCALL call_2_0_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_11_11
#define init_2_0_11_11
void FASTCALL call_2_0_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_11_12
#define init_2_0_11_12
void FASTCALL call_2_0_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_11_13
#define init_2_0_11_13
void FASTCALL call_2_0_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_11_14
#define init_2_0_11_14
void FASTCALL call_2_0_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_11_15
#define init_2_0_11_15
void FASTCALL call_2_0_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_12_0
#define init_2_0_12_0
void FASTCALL call_2_0_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_12_1
#define init_2_0_12_1
void FASTCALL call_2_0_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_12_2
#define init_2_0_12_2
void FASTCALL call_2_0_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_12_3
#define init_2_0_12_3
void FASTCALL call_2_0_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_12_4
#define init_2_0_12_4
void FASTCALL call_2_0_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_12_5
#define init_2_0_12_5
void FASTCALL call_2_0_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_12_6
#define init_2_0_12_6
void FASTCALL call_2_0_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_12_7
#define init_2_0_12_7
void FASTCALL call_2_0_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_12_8
#define init_2_0_12_8
void FASTCALL call_2_0_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_12_9
#define init_2_0_12_9
void FASTCALL call_2_0_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_12_10
#define init_2_0_12_10
void FASTCALL call_2_0_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_12_11
#define init_2_0_12_11
void FASTCALL call_2_0_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_12_12
#define init_2_0_12_12
void FASTCALL call_2_0_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_12_13
#define init_2_0_12_13
void FASTCALL call_2_0_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_12_14
#define init_2_0_12_14
void FASTCALL call_2_0_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_12_15
#define init_2_0_12_15
void FASTCALL call_2_0_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_13_0
#define init_2_0_13_0
void FASTCALL call_2_0_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_13_1
#define init_2_0_13_1
void FASTCALL call_2_0_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_13_2
#define init_2_0_13_2
void FASTCALL call_2_0_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_13_3
#define init_2_0_13_3
void FASTCALL call_2_0_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_13_4
#define init_2_0_13_4
void FASTCALL call_2_0_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_13_5
#define init_2_0_13_5
void FASTCALL call_2_0_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_13_6
#define init_2_0_13_6
void FASTCALL call_2_0_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_13_7
#define init_2_0_13_7
void FASTCALL call_2_0_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_13_8
#define init_2_0_13_8
void FASTCALL call_2_0_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_13_9
#define init_2_0_13_9
void FASTCALL call_2_0_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_13_10
#define init_2_0_13_10
void FASTCALL call_2_0_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_13_11
#define init_2_0_13_11
void FASTCALL call_2_0_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_13_12
#define init_2_0_13_12
void FASTCALL call_2_0_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_13_13
#define init_2_0_13_13
void FASTCALL call_2_0_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_13_14
#define init_2_0_13_14
void FASTCALL call_2_0_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_13_15
#define init_2_0_13_15
void FASTCALL call_2_0_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_14_0
#define init_2_0_14_0
void FASTCALL call_2_0_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_14_1
#define init_2_0_14_1
void FASTCALL call_2_0_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_14_2
#define init_2_0_14_2
void FASTCALL call_2_0_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_14_3
#define init_2_0_14_3
void FASTCALL call_2_0_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_14_4
#define init_2_0_14_4
void FASTCALL call_2_0_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_14_5
#define init_2_0_14_5
void FASTCALL call_2_0_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_14_6
#define init_2_0_14_6
void FASTCALL call_2_0_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_14_7
#define init_2_0_14_7
void FASTCALL call_2_0_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_14_8
#define init_2_0_14_8
void FASTCALL call_2_0_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_14_9
#define init_2_0_14_9
void FASTCALL call_2_0_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_14_10
#define init_2_0_14_10
void FASTCALL call_2_0_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_14_11
#define init_2_0_14_11
void FASTCALL call_2_0_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_14_12
#define init_2_0_14_12
void FASTCALL call_2_0_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_14_13
#define init_2_0_14_13
void FASTCALL call_2_0_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_14_14
#define init_2_0_14_14
void FASTCALL call_2_0_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_14_15
#define init_2_0_14_15
void FASTCALL call_2_0_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_15_0
#define init_2_0_15_0
void FASTCALL call_2_0_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_15_1
#define init_2_0_15_1
void FASTCALL call_2_0_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_15_2
#define init_2_0_15_2
void FASTCALL call_2_0_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_15_3
#define init_2_0_15_3
void FASTCALL call_2_0_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_15_4
#define init_2_0_15_4
void FASTCALL call_2_0_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_15_5
#define init_2_0_15_5
void FASTCALL call_2_0_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_15_6
#define init_2_0_15_6
void FASTCALL call_2_0_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_15_7
#define init_2_0_15_7
void FASTCALL call_2_0_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_15_8
#define init_2_0_15_8
void FASTCALL call_2_0_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_15_9
#define init_2_0_15_9
void FASTCALL call_2_0_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_15_10
#define init_2_0_15_10
void FASTCALL call_2_0_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_15_11
#define init_2_0_15_11
void FASTCALL call_2_0_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_15_12
#define init_2_0_15_12
void FASTCALL call_2_0_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_15_13
#define init_2_0_15_13
void FASTCALL call_2_0_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_15_14
#define init_2_0_15_14
void FASTCALL call_2_0_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_0_15_15
#define init_2_0_15_15
void FASTCALL call_2_0_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_0_0
#define init_2_1_0_0
void FASTCALL call_2_1_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_0_1
#define init_2_1_0_1
void FASTCALL call_2_1_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_0_2
#define init_2_1_0_2
void FASTCALL call_2_1_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_0_3
#define init_2_1_0_3
void FASTCALL call_2_1_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_0_4
#define init_2_1_0_4
void FASTCALL call_2_1_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_0_5
#define init_2_1_0_5
void FASTCALL call_2_1_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_0_6
#define init_2_1_0_6
void FASTCALL call_2_1_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_0_7
#define init_2_1_0_7
void FASTCALL call_2_1_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_0_8
#define init_2_1_0_8
void FASTCALL call_2_1_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_0_9
#define init_2_1_0_9
void FASTCALL call_2_1_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_0_10
#define init_2_1_0_10
void FASTCALL call_2_1_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_0_11
#define init_2_1_0_11
void FASTCALL call_2_1_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_0_12
#define init_2_1_0_12
void FASTCALL call_2_1_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_0_13
#define init_2_1_0_13
void FASTCALL call_2_1_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_0_14
#define init_2_1_0_14
void FASTCALL call_2_1_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_0_15
#define init_2_1_0_15
void FASTCALL call_2_1_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_1_0
#define init_2_1_1_0
void FASTCALL call_2_1_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_1_1
#define init_2_1_1_1
void FASTCALL call_2_1_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_1_2
#define init_2_1_1_2
void FASTCALL call_2_1_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_1_3
#define init_2_1_1_3
void FASTCALL call_2_1_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_1_4
#define init_2_1_1_4
void FASTCALL call_2_1_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_1_5
#define init_2_1_1_5
void FASTCALL call_2_1_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_1_6
#define init_2_1_1_6
void FASTCALL call_2_1_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_1_7
#define init_2_1_1_7
void FASTCALL call_2_1_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_1_8
#define init_2_1_1_8
void FASTCALL call_2_1_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_1_9
#define init_2_1_1_9
void FASTCALL call_2_1_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_1_10
#define init_2_1_1_10
void FASTCALL call_2_1_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_1_11
#define init_2_1_1_11
void FASTCALL call_2_1_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_1_12
#define init_2_1_1_12
void FASTCALL call_2_1_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_1_13
#define init_2_1_1_13
void FASTCALL call_2_1_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_1_14
#define init_2_1_1_14
void FASTCALL call_2_1_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_1_15
#define init_2_1_1_15
void FASTCALL call_2_1_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_2_0
#define init_2_1_2_0
void FASTCALL call_2_1_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_2_1
#define init_2_1_2_1
void FASTCALL call_2_1_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_2_2
#define init_2_1_2_2
void FASTCALL call_2_1_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_2_3
#define init_2_1_2_3
void FASTCALL call_2_1_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_2_4
#define init_2_1_2_4
void FASTCALL call_2_1_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_2_5
#define init_2_1_2_5
void FASTCALL call_2_1_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_2_6
#define init_2_1_2_6
void FASTCALL call_2_1_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_2_7
#define init_2_1_2_7
void FASTCALL call_2_1_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_2_8
#define init_2_1_2_8
void FASTCALL call_2_1_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_2_9
#define init_2_1_2_9
void FASTCALL call_2_1_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_2_10
#define init_2_1_2_10
void FASTCALL call_2_1_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_2_11
#define init_2_1_2_11
void FASTCALL call_2_1_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_2_12
#define init_2_1_2_12
void FASTCALL call_2_1_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_2_13
#define init_2_1_2_13
void FASTCALL call_2_1_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_2_14
#define init_2_1_2_14
void FASTCALL call_2_1_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_2_15
#define init_2_1_2_15
void FASTCALL call_2_1_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_3_0
#define init_2_1_3_0
void FASTCALL call_2_1_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_3_1
#define init_2_1_3_1
void FASTCALL call_2_1_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_3_2
#define init_2_1_3_2
void FASTCALL call_2_1_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_3_3
#define init_2_1_3_3
void FASTCALL call_2_1_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_3_4
#define init_2_1_3_4
void FASTCALL call_2_1_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_3_5
#define init_2_1_3_5
void FASTCALL call_2_1_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_3_6
#define init_2_1_3_6
void FASTCALL call_2_1_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_3_7
#define init_2_1_3_7
void FASTCALL call_2_1_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_3_8
#define init_2_1_3_8
void FASTCALL call_2_1_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_3_9
#define init_2_1_3_9
void FASTCALL call_2_1_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_3_10
#define init_2_1_3_10
void FASTCALL call_2_1_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_3_11
#define init_2_1_3_11
void FASTCALL call_2_1_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_3_12
#define init_2_1_3_12
void FASTCALL call_2_1_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_3_13
#define init_2_1_3_13
void FASTCALL call_2_1_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_3_14
#define init_2_1_3_14
void FASTCALL call_2_1_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_3_15
#define init_2_1_3_15
void FASTCALL call_2_1_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_4_0
#define init_2_1_4_0
void FASTCALL call_2_1_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_4_1
#define init_2_1_4_1
void FASTCALL call_2_1_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_4_2
#define init_2_1_4_2
void FASTCALL call_2_1_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_4_3
#define init_2_1_4_3
void FASTCALL call_2_1_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_4_4
#define init_2_1_4_4
void FASTCALL call_2_1_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_4_5
#define init_2_1_4_5
void FASTCALL call_2_1_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_4_6
#define init_2_1_4_6
void FASTCALL call_2_1_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_4_7
#define init_2_1_4_7
void FASTCALL call_2_1_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_4_8
#define init_2_1_4_8
void FASTCALL call_2_1_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_4_9
#define init_2_1_4_9
void FASTCALL call_2_1_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_4_10
#define init_2_1_4_10
void FASTCALL call_2_1_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_4_11
#define init_2_1_4_11
void FASTCALL call_2_1_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_4_12
#define init_2_1_4_12
void FASTCALL call_2_1_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_4_13
#define init_2_1_4_13
void FASTCALL call_2_1_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_4_14
#define init_2_1_4_14
void FASTCALL call_2_1_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_4_15
#define init_2_1_4_15
void FASTCALL call_2_1_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_5_0
#define init_2_1_5_0
void FASTCALL call_2_1_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_5_1
#define init_2_1_5_1
void FASTCALL call_2_1_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_5_2
#define init_2_1_5_2
void FASTCALL call_2_1_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_5_3
#define init_2_1_5_3
void FASTCALL call_2_1_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_5_4
#define init_2_1_5_4
void FASTCALL call_2_1_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_5_5
#define init_2_1_5_5
void FASTCALL call_2_1_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_5_6
#define init_2_1_5_6
void FASTCALL call_2_1_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_5_7
#define init_2_1_5_7
void FASTCALL call_2_1_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_5_8
#define init_2_1_5_8
void FASTCALL call_2_1_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_5_9
#define init_2_1_5_9
void FASTCALL call_2_1_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_5_10
#define init_2_1_5_10
void FASTCALL call_2_1_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_5_11
#define init_2_1_5_11
void FASTCALL call_2_1_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_5_12
#define init_2_1_5_12
void FASTCALL call_2_1_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_5_13
#define init_2_1_5_13
void FASTCALL call_2_1_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_5_14
#define init_2_1_5_14
void FASTCALL call_2_1_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_5_15
#define init_2_1_5_15
void FASTCALL call_2_1_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_6_0
#define init_2_1_6_0
void FASTCALL call_2_1_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_6_1
#define init_2_1_6_1
void FASTCALL call_2_1_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_6_2
#define init_2_1_6_2
void FASTCALL call_2_1_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_6_3
#define init_2_1_6_3
void FASTCALL call_2_1_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_6_4
#define init_2_1_6_4
void FASTCALL call_2_1_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_6_5
#define init_2_1_6_5
void FASTCALL call_2_1_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_6_6
#define init_2_1_6_6
void FASTCALL call_2_1_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_6_7
#define init_2_1_6_7
void FASTCALL call_2_1_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_6_8
#define init_2_1_6_8
void FASTCALL call_2_1_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_6_9
#define init_2_1_6_9
void FASTCALL call_2_1_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_6_10
#define init_2_1_6_10
void FASTCALL call_2_1_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_6_11
#define init_2_1_6_11
void FASTCALL call_2_1_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_6_12
#define init_2_1_6_12
void FASTCALL call_2_1_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_6_13
#define init_2_1_6_13
void FASTCALL call_2_1_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_6_14
#define init_2_1_6_14
void FASTCALL call_2_1_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_6_15
#define init_2_1_6_15
void FASTCALL call_2_1_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_7_0
#define init_2_1_7_0
void FASTCALL call_2_1_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_7_1
#define init_2_1_7_1
void FASTCALL call_2_1_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_7_2
#define init_2_1_7_2
void FASTCALL call_2_1_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_7_3
#define init_2_1_7_3
void FASTCALL call_2_1_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_7_4
#define init_2_1_7_4
void FASTCALL call_2_1_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_7_5
#define init_2_1_7_5
void FASTCALL call_2_1_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_7_6
#define init_2_1_7_6
void FASTCALL call_2_1_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_7_7
#define init_2_1_7_7
void FASTCALL call_2_1_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_7_8
#define init_2_1_7_8
void FASTCALL call_2_1_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_7_9
#define init_2_1_7_9
void FASTCALL call_2_1_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_7_10
#define init_2_1_7_10
void FASTCALL call_2_1_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_7_11
#define init_2_1_7_11
void FASTCALL call_2_1_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_7_12
#define init_2_1_7_12
void FASTCALL call_2_1_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_7_13
#define init_2_1_7_13
void FASTCALL call_2_1_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_7_14
#define init_2_1_7_14
void FASTCALL call_2_1_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_7_15
#define init_2_1_7_15
void FASTCALL call_2_1_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_8_0
#define init_2_1_8_0
void FASTCALL call_2_1_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_8_1
#define init_2_1_8_1
void FASTCALL call_2_1_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_8_2
#define init_2_1_8_2
void FASTCALL call_2_1_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_8_3
#define init_2_1_8_3
void FASTCALL call_2_1_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_8_4
#define init_2_1_8_4
void FASTCALL call_2_1_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_8_5
#define init_2_1_8_5
void FASTCALL call_2_1_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_8_6
#define init_2_1_8_6
void FASTCALL call_2_1_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_8_7
#define init_2_1_8_7
void FASTCALL call_2_1_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_8_8
#define init_2_1_8_8
void FASTCALL call_2_1_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_8_9
#define init_2_1_8_9
void FASTCALL call_2_1_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_8_10
#define init_2_1_8_10
void FASTCALL call_2_1_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_8_11
#define init_2_1_8_11
void FASTCALL call_2_1_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_8_12
#define init_2_1_8_12
void FASTCALL call_2_1_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_8_13
#define init_2_1_8_13
void FASTCALL call_2_1_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_8_14
#define init_2_1_8_14
void FASTCALL call_2_1_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_8_15
#define init_2_1_8_15
void FASTCALL call_2_1_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_9_0
#define init_2_1_9_0
void FASTCALL call_2_1_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_9_1
#define init_2_1_9_1
void FASTCALL call_2_1_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_9_2
#define init_2_1_9_2
void FASTCALL call_2_1_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_9_3
#define init_2_1_9_3
void FASTCALL call_2_1_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_9_4
#define init_2_1_9_4
void FASTCALL call_2_1_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_9_5
#define init_2_1_9_5
void FASTCALL call_2_1_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_9_6
#define init_2_1_9_6
void FASTCALL call_2_1_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_9_7
#define init_2_1_9_7
void FASTCALL call_2_1_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_9_8
#define init_2_1_9_8
void FASTCALL call_2_1_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_9_9
#define init_2_1_9_9
void FASTCALL call_2_1_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_9_10
#define init_2_1_9_10
void FASTCALL call_2_1_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_9_11
#define init_2_1_9_11
void FASTCALL call_2_1_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_9_12
#define init_2_1_9_12
void FASTCALL call_2_1_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_9_13
#define init_2_1_9_13
void FASTCALL call_2_1_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_9_14
#define init_2_1_9_14
void FASTCALL call_2_1_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_9_15
#define init_2_1_9_15
void FASTCALL call_2_1_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_10_0
#define init_2_1_10_0
void FASTCALL call_2_1_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_10_1
#define init_2_1_10_1
void FASTCALL call_2_1_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_10_2
#define init_2_1_10_2
void FASTCALL call_2_1_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_10_3
#define init_2_1_10_3
void FASTCALL call_2_1_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_10_4
#define init_2_1_10_4
void FASTCALL call_2_1_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_10_5
#define init_2_1_10_5
void FASTCALL call_2_1_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_10_6
#define init_2_1_10_6
void FASTCALL call_2_1_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_10_7
#define init_2_1_10_7
void FASTCALL call_2_1_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_10_8
#define init_2_1_10_8
void FASTCALL call_2_1_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_10_9
#define init_2_1_10_9
void FASTCALL call_2_1_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_10_10
#define init_2_1_10_10
void FASTCALL call_2_1_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_10_11
#define init_2_1_10_11
void FASTCALL call_2_1_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_10_12
#define init_2_1_10_12
void FASTCALL call_2_1_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_10_13
#define init_2_1_10_13
void FASTCALL call_2_1_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_10_14
#define init_2_1_10_14
void FASTCALL call_2_1_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_10_15
#define init_2_1_10_15
void FASTCALL call_2_1_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_11_0
#define init_2_1_11_0
void FASTCALL call_2_1_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_11_1
#define init_2_1_11_1
void FASTCALL call_2_1_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_11_2
#define init_2_1_11_2
void FASTCALL call_2_1_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_11_3
#define init_2_1_11_3
void FASTCALL call_2_1_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_11_4
#define init_2_1_11_4
void FASTCALL call_2_1_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_11_5
#define init_2_1_11_5
void FASTCALL call_2_1_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_11_6
#define init_2_1_11_6
void FASTCALL call_2_1_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_11_7
#define init_2_1_11_7
void FASTCALL call_2_1_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_11_8
#define init_2_1_11_8
void FASTCALL call_2_1_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_11_9
#define init_2_1_11_9
void FASTCALL call_2_1_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_11_10
#define init_2_1_11_10
void FASTCALL call_2_1_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_11_11
#define init_2_1_11_11
void FASTCALL call_2_1_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_11_12
#define init_2_1_11_12
void FASTCALL call_2_1_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_11_13
#define init_2_1_11_13
void FASTCALL call_2_1_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_11_14
#define init_2_1_11_14
void FASTCALL call_2_1_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_11_15
#define init_2_1_11_15
void FASTCALL call_2_1_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_12_0
#define init_2_1_12_0
void FASTCALL call_2_1_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_12_1
#define init_2_1_12_1
void FASTCALL call_2_1_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_12_2
#define init_2_1_12_2
void FASTCALL call_2_1_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_12_3
#define init_2_1_12_3
void FASTCALL call_2_1_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_12_4
#define init_2_1_12_4
void FASTCALL call_2_1_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_12_5
#define init_2_1_12_5
void FASTCALL call_2_1_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_12_6
#define init_2_1_12_6
void FASTCALL call_2_1_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_12_7
#define init_2_1_12_7
void FASTCALL call_2_1_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_12_8
#define init_2_1_12_8
void FASTCALL call_2_1_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_12_9
#define init_2_1_12_9
void FASTCALL call_2_1_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_12_10
#define init_2_1_12_10
void FASTCALL call_2_1_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_12_11
#define init_2_1_12_11
void FASTCALL call_2_1_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_12_12
#define init_2_1_12_12
void FASTCALL call_2_1_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_12_13
#define init_2_1_12_13
void FASTCALL call_2_1_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_12_14
#define init_2_1_12_14
void FASTCALL call_2_1_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_12_15
#define init_2_1_12_15
void FASTCALL call_2_1_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_13_0
#define init_2_1_13_0
void FASTCALL call_2_1_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_13_1
#define init_2_1_13_1
void FASTCALL call_2_1_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_13_2
#define init_2_1_13_2
void FASTCALL call_2_1_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_13_3
#define init_2_1_13_3
void FASTCALL call_2_1_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_13_4
#define init_2_1_13_4
void FASTCALL call_2_1_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_13_5
#define init_2_1_13_5
void FASTCALL call_2_1_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_13_6
#define init_2_1_13_6
void FASTCALL call_2_1_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_13_7
#define init_2_1_13_7
void FASTCALL call_2_1_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_13_8
#define init_2_1_13_8
void FASTCALL call_2_1_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_13_9
#define init_2_1_13_9
void FASTCALL call_2_1_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_13_10
#define init_2_1_13_10
void FASTCALL call_2_1_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_13_11
#define init_2_1_13_11
void FASTCALL call_2_1_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_13_12
#define init_2_1_13_12
void FASTCALL call_2_1_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_13_13
#define init_2_1_13_13
void FASTCALL call_2_1_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_13_14
#define init_2_1_13_14
void FASTCALL call_2_1_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_13_15
#define init_2_1_13_15
void FASTCALL call_2_1_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_14_0
#define init_2_1_14_0
void FASTCALL call_2_1_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_14_1
#define init_2_1_14_1
void FASTCALL call_2_1_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_14_2
#define init_2_1_14_2
void FASTCALL call_2_1_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_14_3
#define init_2_1_14_3
void FASTCALL call_2_1_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_14_4
#define init_2_1_14_4
void FASTCALL call_2_1_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_14_5
#define init_2_1_14_5
void FASTCALL call_2_1_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_14_6
#define init_2_1_14_6
void FASTCALL call_2_1_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_14_7
#define init_2_1_14_7
void FASTCALL call_2_1_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_14_8
#define init_2_1_14_8
void FASTCALL call_2_1_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_14_9
#define init_2_1_14_9
void FASTCALL call_2_1_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_14_10
#define init_2_1_14_10
void FASTCALL call_2_1_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_14_11
#define init_2_1_14_11
void FASTCALL call_2_1_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_14_12
#define init_2_1_14_12
void FASTCALL call_2_1_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_14_13
#define init_2_1_14_13
void FASTCALL call_2_1_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_14_14
#define init_2_1_14_14
void FASTCALL call_2_1_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_14_15
#define init_2_1_14_15
void FASTCALL call_2_1_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_15_0
#define init_2_1_15_0
void FASTCALL call_2_1_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_15_1
#define init_2_1_15_1
void FASTCALL call_2_1_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_15_2
#define init_2_1_15_2
void FASTCALL call_2_1_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_15_3
#define init_2_1_15_3
void FASTCALL call_2_1_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_15_4
#define init_2_1_15_4
void FASTCALL call_2_1_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_15_5
#define init_2_1_15_5
void FASTCALL call_2_1_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_15_6
#define init_2_1_15_6
void FASTCALL call_2_1_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_15_7
#define init_2_1_15_7
void FASTCALL call_2_1_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_15_8
#define init_2_1_15_8
void FASTCALL call_2_1_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_15_9
#define init_2_1_15_9
void FASTCALL call_2_1_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_15_10
#define init_2_1_15_10
void FASTCALL call_2_1_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_15_11
#define init_2_1_15_11
void FASTCALL call_2_1_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_15_12
#define init_2_1_15_12
void FASTCALL call_2_1_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_15_13
#define init_2_1_15_13
void FASTCALL call_2_1_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_15_14
#define init_2_1_15_14
void FASTCALL call_2_1_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_1_15_15
#define init_2_1_15_15
void FASTCALL call_2_1_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_0_0
#define init_2_2_0_0
void FASTCALL call_2_2_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_0_1
#define init_2_2_0_1
void FASTCALL call_2_2_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_0_2
#define init_2_2_0_2
void FASTCALL call_2_2_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_0_3
#define init_2_2_0_3
void FASTCALL call_2_2_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_0_4
#define init_2_2_0_4
void FASTCALL call_2_2_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_0_5
#define init_2_2_0_5
void FASTCALL call_2_2_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_0_6
#define init_2_2_0_6
void FASTCALL call_2_2_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_0_7
#define init_2_2_0_7
void FASTCALL call_2_2_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_0_8
#define init_2_2_0_8
void FASTCALL call_2_2_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_0_9
#define init_2_2_0_9
void FASTCALL call_2_2_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_0_10
#define init_2_2_0_10
void FASTCALL call_2_2_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_0_11
#define init_2_2_0_11
void FASTCALL call_2_2_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_0_12
#define init_2_2_0_12
void FASTCALL call_2_2_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_0_13
#define init_2_2_0_13
void FASTCALL call_2_2_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_0_14
#define init_2_2_0_14
void FASTCALL call_2_2_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_0_15
#define init_2_2_0_15
void FASTCALL call_2_2_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_1_0
#define init_2_2_1_0
void FASTCALL call_2_2_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_1_1
#define init_2_2_1_1
void FASTCALL call_2_2_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_1_2
#define init_2_2_1_2
void FASTCALL call_2_2_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_1_3
#define init_2_2_1_3
void FASTCALL call_2_2_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_1_4
#define init_2_2_1_4
void FASTCALL call_2_2_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_1_5
#define init_2_2_1_5
void FASTCALL call_2_2_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_1_6
#define init_2_2_1_6
void FASTCALL call_2_2_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_1_7
#define init_2_2_1_7
void FASTCALL call_2_2_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_1_8
#define init_2_2_1_8
void FASTCALL call_2_2_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_1_9
#define init_2_2_1_9
void FASTCALL call_2_2_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_1_10
#define init_2_2_1_10
void FASTCALL call_2_2_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_1_11
#define init_2_2_1_11
void FASTCALL call_2_2_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_1_12
#define init_2_2_1_12
void FASTCALL call_2_2_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_1_13
#define init_2_2_1_13
void FASTCALL call_2_2_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_1_14
#define init_2_2_1_14
void FASTCALL call_2_2_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_1_15
#define init_2_2_1_15
void FASTCALL call_2_2_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_2_0
#define init_2_2_2_0
void FASTCALL call_2_2_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_2_1
#define init_2_2_2_1
void FASTCALL call_2_2_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_2_2
#define init_2_2_2_2
void FASTCALL call_2_2_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_2_3
#define init_2_2_2_3
void FASTCALL call_2_2_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_2_4
#define init_2_2_2_4
void FASTCALL call_2_2_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_2_5
#define init_2_2_2_5
void FASTCALL call_2_2_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_2_6
#define init_2_2_2_6
void FASTCALL call_2_2_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_2_7
#define init_2_2_2_7
void FASTCALL call_2_2_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_2_8
#define init_2_2_2_8
void FASTCALL call_2_2_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_2_9
#define init_2_2_2_9
void FASTCALL call_2_2_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_2_10
#define init_2_2_2_10
void FASTCALL call_2_2_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_2_11
#define init_2_2_2_11
void FASTCALL call_2_2_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_2_12
#define init_2_2_2_12
void FASTCALL call_2_2_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_2_13
#define init_2_2_2_13
void FASTCALL call_2_2_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_2_14
#define init_2_2_2_14
void FASTCALL call_2_2_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_2_15
#define init_2_2_2_15
void FASTCALL call_2_2_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_3_0
#define init_2_2_3_0
void FASTCALL call_2_2_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_3_1
#define init_2_2_3_1
void FASTCALL call_2_2_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_3_2
#define init_2_2_3_2
void FASTCALL call_2_2_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_3_3
#define init_2_2_3_3
void FASTCALL call_2_2_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_3_4
#define init_2_2_3_4
void FASTCALL call_2_2_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_3_5
#define init_2_2_3_5
void FASTCALL call_2_2_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_3_6
#define init_2_2_3_6
void FASTCALL call_2_2_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_3_7
#define init_2_2_3_7
void FASTCALL call_2_2_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_3_8
#define init_2_2_3_8
void FASTCALL call_2_2_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_3_9
#define init_2_2_3_9
void FASTCALL call_2_2_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_3_10
#define init_2_2_3_10
void FASTCALL call_2_2_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_3_11
#define init_2_2_3_11
void FASTCALL call_2_2_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_3_12
#define init_2_2_3_12
void FASTCALL call_2_2_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_3_13
#define init_2_2_3_13
void FASTCALL call_2_2_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_3_14
#define init_2_2_3_14
void FASTCALL call_2_2_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_3_15
#define init_2_2_3_15
void FASTCALL call_2_2_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_4_0
#define init_2_2_4_0
void FASTCALL call_2_2_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_4_1
#define init_2_2_4_1
void FASTCALL call_2_2_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_4_2
#define init_2_2_4_2
void FASTCALL call_2_2_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_4_3
#define init_2_2_4_3
void FASTCALL call_2_2_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_4_4
#define init_2_2_4_4
void FASTCALL call_2_2_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_4_5
#define init_2_2_4_5
void FASTCALL call_2_2_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_4_6
#define init_2_2_4_6
void FASTCALL call_2_2_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_4_7
#define init_2_2_4_7
void FASTCALL call_2_2_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_4_8
#define init_2_2_4_8
void FASTCALL call_2_2_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_4_9
#define init_2_2_4_9
void FASTCALL call_2_2_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_4_10
#define init_2_2_4_10
void FASTCALL call_2_2_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_4_11
#define init_2_2_4_11
void FASTCALL call_2_2_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_4_12
#define init_2_2_4_12
void FASTCALL call_2_2_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_4_13
#define init_2_2_4_13
void FASTCALL call_2_2_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_4_14
#define init_2_2_4_14
void FASTCALL call_2_2_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_4_15
#define init_2_2_4_15
void FASTCALL call_2_2_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_5_0
#define init_2_2_5_0
void FASTCALL call_2_2_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_5_1
#define init_2_2_5_1
void FASTCALL call_2_2_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_5_2
#define init_2_2_5_2
void FASTCALL call_2_2_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_5_3
#define init_2_2_5_3
void FASTCALL call_2_2_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_5_4
#define init_2_2_5_4
void FASTCALL call_2_2_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_5_5
#define init_2_2_5_5
void FASTCALL call_2_2_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_5_6
#define init_2_2_5_6
void FASTCALL call_2_2_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_5_7
#define init_2_2_5_7
void FASTCALL call_2_2_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_5_8
#define init_2_2_5_8
void FASTCALL call_2_2_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_5_9
#define init_2_2_5_9
void FASTCALL call_2_2_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_5_10
#define init_2_2_5_10
void FASTCALL call_2_2_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_5_11
#define init_2_2_5_11
void FASTCALL call_2_2_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_5_12
#define init_2_2_5_12
void FASTCALL call_2_2_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_5_13
#define init_2_2_5_13
void FASTCALL call_2_2_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_5_14
#define init_2_2_5_14
void FASTCALL call_2_2_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_5_15
#define init_2_2_5_15
void FASTCALL call_2_2_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_6_0
#define init_2_2_6_0
void FASTCALL call_2_2_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_6_1
#define init_2_2_6_1
void FASTCALL call_2_2_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_6_2
#define init_2_2_6_2
void FASTCALL call_2_2_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_6_3
#define init_2_2_6_3
void FASTCALL call_2_2_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_6_4
#define init_2_2_6_4
void FASTCALL call_2_2_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_6_5
#define init_2_2_6_5
void FASTCALL call_2_2_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_6_6
#define init_2_2_6_6
void FASTCALL call_2_2_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_6_7
#define init_2_2_6_7
void FASTCALL call_2_2_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_6_8
#define init_2_2_6_8
void FASTCALL call_2_2_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_6_9
#define init_2_2_6_9
void FASTCALL call_2_2_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_6_10
#define init_2_2_6_10
void FASTCALL call_2_2_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_6_11
#define init_2_2_6_11
void FASTCALL call_2_2_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_6_12
#define init_2_2_6_12
void FASTCALL call_2_2_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_6_13
#define init_2_2_6_13
void FASTCALL call_2_2_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_6_14
#define init_2_2_6_14
void FASTCALL call_2_2_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_6_15
#define init_2_2_6_15
void FASTCALL call_2_2_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_7_0
#define init_2_2_7_0
void FASTCALL call_2_2_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_7_1
#define init_2_2_7_1
void FASTCALL call_2_2_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_7_2
#define init_2_2_7_2
void FASTCALL call_2_2_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_7_3
#define init_2_2_7_3
void FASTCALL call_2_2_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_7_4
#define init_2_2_7_4
void FASTCALL call_2_2_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_7_5
#define init_2_2_7_5
void FASTCALL call_2_2_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_7_6
#define init_2_2_7_6
void FASTCALL call_2_2_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_7_7
#define init_2_2_7_7
void FASTCALL call_2_2_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_7_8
#define init_2_2_7_8
void FASTCALL call_2_2_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_7_9
#define init_2_2_7_9
void FASTCALL call_2_2_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_7_10
#define init_2_2_7_10
void FASTCALL call_2_2_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_7_11
#define init_2_2_7_11
void FASTCALL call_2_2_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_7_12
#define init_2_2_7_12
void FASTCALL call_2_2_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_7_13
#define init_2_2_7_13
void FASTCALL call_2_2_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_7_14
#define init_2_2_7_14
void FASTCALL call_2_2_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_7_15
#define init_2_2_7_15
void FASTCALL call_2_2_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_8_0
#define init_2_2_8_0
void FASTCALL call_2_2_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_8_1
#define init_2_2_8_1
void FASTCALL call_2_2_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_8_2
#define init_2_2_8_2
void FASTCALL call_2_2_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_8_3
#define init_2_2_8_3
void FASTCALL call_2_2_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_8_4
#define init_2_2_8_4
void FASTCALL call_2_2_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_8_5
#define init_2_2_8_5
void FASTCALL call_2_2_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_8_6
#define init_2_2_8_6
void FASTCALL call_2_2_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_8_7
#define init_2_2_8_7
void FASTCALL call_2_2_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_8_8
#define init_2_2_8_8
void FASTCALL call_2_2_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_8_9
#define init_2_2_8_9
void FASTCALL call_2_2_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_8_10
#define init_2_2_8_10
void FASTCALL call_2_2_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_8_11
#define init_2_2_8_11
void FASTCALL call_2_2_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_8_12
#define init_2_2_8_12
void FASTCALL call_2_2_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_8_13
#define init_2_2_8_13
void FASTCALL call_2_2_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_8_14
#define init_2_2_8_14
void FASTCALL call_2_2_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_8_15
#define init_2_2_8_15
void FASTCALL call_2_2_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_9_0
#define init_2_2_9_0
void FASTCALL call_2_2_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_9_1
#define init_2_2_9_1
void FASTCALL call_2_2_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_9_2
#define init_2_2_9_2
void FASTCALL call_2_2_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_9_3
#define init_2_2_9_3
void FASTCALL call_2_2_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_9_4
#define init_2_2_9_4
void FASTCALL call_2_2_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_9_5
#define init_2_2_9_5
void FASTCALL call_2_2_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_9_6
#define init_2_2_9_6
void FASTCALL call_2_2_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_9_7
#define init_2_2_9_7
void FASTCALL call_2_2_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_9_8
#define init_2_2_9_8
void FASTCALL call_2_2_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_9_9
#define init_2_2_9_9
void FASTCALL call_2_2_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_9_10
#define init_2_2_9_10
void FASTCALL call_2_2_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_9_11
#define init_2_2_9_11
void FASTCALL call_2_2_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_9_12
#define init_2_2_9_12
void FASTCALL call_2_2_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_9_13
#define init_2_2_9_13
void FASTCALL call_2_2_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_9_14
#define init_2_2_9_14
void FASTCALL call_2_2_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_9_15
#define init_2_2_9_15
void FASTCALL call_2_2_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_10_0
#define init_2_2_10_0
void FASTCALL call_2_2_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_10_1
#define init_2_2_10_1
void FASTCALL call_2_2_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_10_2
#define init_2_2_10_2
void FASTCALL call_2_2_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_10_3
#define init_2_2_10_3
void FASTCALL call_2_2_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_10_4
#define init_2_2_10_4
void FASTCALL call_2_2_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_10_5
#define init_2_2_10_5
void FASTCALL call_2_2_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_10_6
#define init_2_2_10_6
void FASTCALL call_2_2_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_10_7
#define init_2_2_10_7
void FASTCALL call_2_2_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_10_8
#define init_2_2_10_8
void FASTCALL call_2_2_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_10_9
#define init_2_2_10_9
void FASTCALL call_2_2_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_10_10
#define init_2_2_10_10
void FASTCALL call_2_2_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_10_11
#define init_2_2_10_11
void FASTCALL call_2_2_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_10_12
#define init_2_2_10_12
void FASTCALL call_2_2_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_10_13
#define init_2_2_10_13
void FASTCALL call_2_2_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_10_14
#define init_2_2_10_14
void FASTCALL call_2_2_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_10_15
#define init_2_2_10_15
void FASTCALL call_2_2_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_11_0
#define init_2_2_11_0
void FASTCALL call_2_2_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_11_1
#define init_2_2_11_1
void FASTCALL call_2_2_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_11_2
#define init_2_2_11_2
void FASTCALL call_2_2_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_11_3
#define init_2_2_11_3
void FASTCALL call_2_2_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_11_4
#define init_2_2_11_4
void FASTCALL call_2_2_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_11_5
#define init_2_2_11_5
void FASTCALL call_2_2_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_11_6
#define init_2_2_11_6
void FASTCALL call_2_2_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_11_7
#define init_2_2_11_7
void FASTCALL call_2_2_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_11_8
#define init_2_2_11_8
void FASTCALL call_2_2_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_11_9
#define init_2_2_11_9
void FASTCALL call_2_2_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_11_10
#define init_2_2_11_10
void FASTCALL call_2_2_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_11_11
#define init_2_2_11_11
void FASTCALL call_2_2_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_11_12
#define init_2_2_11_12
void FASTCALL call_2_2_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_11_13
#define init_2_2_11_13
void FASTCALL call_2_2_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_11_14
#define init_2_2_11_14
void FASTCALL call_2_2_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_11_15
#define init_2_2_11_15
void FASTCALL call_2_2_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_12_0
#define init_2_2_12_0
void FASTCALL call_2_2_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_12_1
#define init_2_2_12_1
void FASTCALL call_2_2_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_12_2
#define init_2_2_12_2
void FASTCALL call_2_2_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_12_3
#define init_2_2_12_3
void FASTCALL call_2_2_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_12_4
#define init_2_2_12_4
void FASTCALL call_2_2_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_12_5
#define init_2_2_12_5
void FASTCALL call_2_2_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_12_6
#define init_2_2_12_6
void FASTCALL call_2_2_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_12_7
#define init_2_2_12_7
void FASTCALL call_2_2_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_12_8
#define init_2_2_12_8
void FASTCALL call_2_2_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_12_9
#define init_2_2_12_9
void FASTCALL call_2_2_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_12_10
#define init_2_2_12_10
void FASTCALL call_2_2_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_12_11
#define init_2_2_12_11
void FASTCALL call_2_2_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_12_12
#define init_2_2_12_12
void FASTCALL call_2_2_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_12_13
#define init_2_2_12_13
void FASTCALL call_2_2_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_12_14
#define init_2_2_12_14
void FASTCALL call_2_2_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_12_15
#define init_2_2_12_15
void FASTCALL call_2_2_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_13_0
#define init_2_2_13_0
void FASTCALL call_2_2_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_13_1
#define init_2_2_13_1
void FASTCALL call_2_2_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_13_2
#define init_2_2_13_2
void FASTCALL call_2_2_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_13_3
#define init_2_2_13_3
void FASTCALL call_2_2_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_13_4
#define init_2_2_13_4
void FASTCALL call_2_2_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_13_5
#define init_2_2_13_5
void FASTCALL call_2_2_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_13_6
#define init_2_2_13_6
void FASTCALL call_2_2_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_13_7
#define init_2_2_13_7
void FASTCALL call_2_2_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_13_8
#define init_2_2_13_8
void FASTCALL call_2_2_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_13_9
#define init_2_2_13_9
void FASTCALL call_2_2_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_13_10
#define init_2_2_13_10
void FASTCALL call_2_2_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_13_11
#define init_2_2_13_11
void FASTCALL call_2_2_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_13_12
#define init_2_2_13_12
void FASTCALL call_2_2_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_13_13
#define init_2_2_13_13
void FASTCALL call_2_2_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_13_14
#define init_2_2_13_14
void FASTCALL call_2_2_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_13_15
#define init_2_2_13_15
void FASTCALL call_2_2_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_14_0
#define init_2_2_14_0
void FASTCALL call_2_2_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_14_1
#define init_2_2_14_1
void FASTCALL call_2_2_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_14_2
#define init_2_2_14_2
void FASTCALL call_2_2_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_14_3
#define init_2_2_14_3
void FASTCALL call_2_2_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_14_4
#define init_2_2_14_4
void FASTCALL call_2_2_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_14_5
#define init_2_2_14_5
void FASTCALL call_2_2_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_14_6
#define init_2_2_14_6
void FASTCALL call_2_2_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_14_7
#define init_2_2_14_7
void FASTCALL call_2_2_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_14_8
#define init_2_2_14_8
void FASTCALL call_2_2_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_14_9
#define init_2_2_14_9
void FASTCALL call_2_2_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_14_10
#define init_2_2_14_10
void FASTCALL call_2_2_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_14_11
#define init_2_2_14_11
void FASTCALL call_2_2_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_14_12
#define init_2_2_14_12
void FASTCALL call_2_2_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_14_13
#define init_2_2_14_13
void FASTCALL call_2_2_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_14_14
#define init_2_2_14_14
void FASTCALL call_2_2_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_14_15
#define init_2_2_14_15
void FASTCALL call_2_2_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_15_0
#define init_2_2_15_0
void FASTCALL call_2_2_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_15_1
#define init_2_2_15_1
void FASTCALL call_2_2_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_15_2
#define init_2_2_15_2
void FASTCALL call_2_2_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_15_3
#define init_2_2_15_3
void FASTCALL call_2_2_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_15_4
#define init_2_2_15_4
void FASTCALL call_2_2_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_15_5
#define init_2_2_15_5
void FASTCALL call_2_2_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_15_6
#define init_2_2_15_6
void FASTCALL call_2_2_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_15_7
#define init_2_2_15_7
void FASTCALL call_2_2_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_15_8
#define init_2_2_15_8
void FASTCALL call_2_2_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_15_9
#define init_2_2_15_9
void FASTCALL call_2_2_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_15_10
#define init_2_2_15_10
void FASTCALL call_2_2_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_15_11
#define init_2_2_15_11
void FASTCALL call_2_2_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_15_12
#define init_2_2_15_12
void FASTCALL call_2_2_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_15_13
#define init_2_2_15_13
void FASTCALL call_2_2_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_15_14
#define init_2_2_15_14
void FASTCALL call_2_2_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_2_15_15
#define init_2_2_15_15
void FASTCALL call_2_2_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_0_0
#define init_2_3_0_0
void FASTCALL call_2_3_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_0_1
#define init_2_3_0_1
void FASTCALL call_2_3_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_0_2
#define init_2_3_0_2
void FASTCALL call_2_3_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_0_3
#define init_2_3_0_3
void FASTCALL call_2_3_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_0_4
#define init_2_3_0_4
void FASTCALL call_2_3_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_0_5
#define init_2_3_0_5
void FASTCALL call_2_3_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_0_6
#define init_2_3_0_6
void FASTCALL call_2_3_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_0_7
#define init_2_3_0_7
void FASTCALL call_2_3_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_0_8
#define init_2_3_0_8
void FASTCALL call_2_3_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_0_9
#define init_2_3_0_9
void FASTCALL call_2_3_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_0_10
#define init_2_3_0_10
void FASTCALL call_2_3_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_0_11
#define init_2_3_0_11
void FASTCALL call_2_3_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_0_12
#define init_2_3_0_12
void FASTCALL call_2_3_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_0_13
#define init_2_3_0_13
void FASTCALL call_2_3_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_0_14
#define init_2_3_0_14
void FASTCALL call_2_3_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_0_15
#define init_2_3_0_15
void FASTCALL call_2_3_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_1_0
#define init_2_3_1_0
void FASTCALL call_2_3_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_1_1
#define init_2_3_1_1
void FASTCALL call_2_3_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_1_2
#define init_2_3_1_2
void FASTCALL call_2_3_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_1_3
#define init_2_3_1_3
void FASTCALL call_2_3_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_1_4
#define init_2_3_1_4
void FASTCALL call_2_3_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_1_5
#define init_2_3_1_5
void FASTCALL call_2_3_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_1_6
#define init_2_3_1_6
void FASTCALL call_2_3_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_1_7
#define init_2_3_1_7
void FASTCALL call_2_3_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_1_8
#define init_2_3_1_8
void FASTCALL call_2_3_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_1_9
#define init_2_3_1_9
void FASTCALL call_2_3_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_1_10
#define init_2_3_1_10
void FASTCALL call_2_3_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_1_11
#define init_2_3_1_11
void FASTCALL call_2_3_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_1_12
#define init_2_3_1_12
void FASTCALL call_2_3_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_1_13
#define init_2_3_1_13
void FASTCALL call_2_3_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_1_14
#define init_2_3_1_14
void FASTCALL call_2_3_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_1_15
#define init_2_3_1_15
void FASTCALL call_2_3_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_2_0
#define init_2_3_2_0
void FASTCALL call_2_3_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_2_1
#define init_2_3_2_1
void FASTCALL call_2_3_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_2_2
#define init_2_3_2_2
void FASTCALL call_2_3_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_2_3
#define init_2_3_2_3
void FASTCALL call_2_3_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_2_4
#define init_2_3_2_4
void FASTCALL call_2_3_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_2_5
#define init_2_3_2_5
void FASTCALL call_2_3_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_2_6
#define init_2_3_2_6
void FASTCALL call_2_3_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_2_7
#define init_2_3_2_7
void FASTCALL call_2_3_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_2_8
#define init_2_3_2_8
void FASTCALL call_2_3_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_2_9
#define init_2_3_2_9
void FASTCALL call_2_3_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_2_10
#define init_2_3_2_10
void FASTCALL call_2_3_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_2_11
#define init_2_3_2_11
void FASTCALL call_2_3_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_2_12
#define init_2_3_2_12
void FASTCALL call_2_3_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_2_13
#define init_2_3_2_13
void FASTCALL call_2_3_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_2_14
#define init_2_3_2_14
void FASTCALL call_2_3_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_2_15
#define init_2_3_2_15
void FASTCALL call_2_3_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_3_0
#define init_2_3_3_0
void FASTCALL call_2_3_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_3_1
#define init_2_3_3_1
void FASTCALL call_2_3_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_3_2
#define init_2_3_3_2
void FASTCALL call_2_3_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_3_3
#define init_2_3_3_3
void FASTCALL call_2_3_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_3_4
#define init_2_3_3_4
void FASTCALL call_2_3_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_3_5
#define init_2_3_3_5
void FASTCALL call_2_3_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_3_6
#define init_2_3_3_6
void FASTCALL call_2_3_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_3_7
#define init_2_3_3_7
void FASTCALL call_2_3_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_3_8
#define init_2_3_3_8
void FASTCALL call_2_3_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_3_9
#define init_2_3_3_9
void FASTCALL call_2_3_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_3_10
#define init_2_3_3_10
void FASTCALL call_2_3_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_3_11
#define init_2_3_3_11
void FASTCALL call_2_3_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_3_12
#define init_2_3_3_12
void FASTCALL call_2_3_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_3_13
#define init_2_3_3_13
void FASTCALL call_2_3_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_3_14
#define init_2_3_3_14
void FASTCALL call_2_3_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_3_15
#define init_2_3_3_15
void FASTCALL call_2_3_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_4_0
#define init_2_3_4_0
void FASTCALL call_2_3_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_4_1
#define init_2_3_4_1
void FASTCALL call_2_3_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_4_2
#define init_2_3_4_2
void FASTCALL call_2_3_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_4_3
#define init_2_3_4_3
void FASTCALL call_2_3_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_4_4
#define init_2_3_4_4
void FASTCALL call_2_3_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_4_5
#define init_2_3_4_5
void FASTCALL call_2_3_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_4_6
#define init_2_3_4_6
void FASTCALL call_2_3_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_4_7
#define init_2_3_4_7
void FASTCALL call_2_3_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_4_8
#define init_2_3_4_8
void FASTCALL call_2_3_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_4_9
#define init_2_3_4_9
void FASTCALL call_2_3_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_4_10
#define init_2_3_4_10
void FASTCALL call_2_3_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_4_11
#define init_2_3_4_11
void FASTCALL call_2_3_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_4_12
#define init_2_3_4_12
void FASTCALL call_2_3_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_4_13
#define init_2_3_4_13
void FASTCALL call_2_3_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_4_14
#define init_2_3_4_14
void FASTCALL call_2_3_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_4_15
#define init_2_3_4_15
void FASTCALL call_2_3_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_5_0
#define init_2_3_5_0
void FASTCALL call_2_3_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_5_1
#define init_2_3_5_1
void FASTCALL call_2_3_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_5_2
#define init_2_3_5_2
void FASTCALL call_2_3_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_5_3
#define init_2_3_5_3
void FASTCALL call_2_3_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_5_4
#define init_2_3_5_4
void FASTCALL call_2_3_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_5_5
#define init_2_3_5_5
void FASTCALL call_2_3_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_5_6
#define init_2_3_5_6
void FASTCALL call_2_3_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_5_7
#define init_2_3_5_7
void FASTCALL call_2_3_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_5_8
#define init_2_3_5_8
void FASTCALL call_2_3_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_5_9
#define init_2_3_5_9
void FASTCALL call_2_3_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_5_10
#define init_2_3_5_10
void FASTCALL call_2_3_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_5_11
#define init_2_3_5_11
void FASTCALL call_2_3_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_5_12
#define init_2_3_5_12
void FASTCALL call_2_3_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_5_13
#define init_2_3_5_13
void FASTCALL call_2_3_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_5_14
#define init_2_3_5_14
void FASTCALL call_2_3_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_5_15
#define init_2_3_5_15
void FASTCALL call_2_3_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_6_0
#define init_2_3_6_0
void FASTCALL call_2_3_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_6_1
#define init_2_3_6_1
void FASTCALL call_2_3_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_6_2
#define init_2_3_6_2
void FASTCALL call_2_3_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_6_3
#define init_2_3_6_3
void FASTCALL call_2_3_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_6_4
#define init_2_3_6_4
void FASTCALL call_2_3_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_6_5
#define init_2_3_6_5
void FASTCALL call_2_3_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_6_6
#define init_2_3_6_6
void FASTCALL call_2_3_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_6_7
#define init_2_3_6_7
void FASTCALL call_2_3_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_6_8
#define init_2_3_6_8
void FASTCALL call_2_3_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_6_9
#define init_2_3_6_9
void FASTCALL call_2_3_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_6_10
#define init_2_3_6_10
void FASTCALL call_2_3_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_6_11
#define init_2_3_6_11
void FASTCALL call_2_3_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_6_12
#define init_2_3_6_12
void FASTCALL call_2_3_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_6_13
#define init_2_3_6_13
void FASTCALL call_2_3_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_6_14
#define init_2_3_6_14
void FASTCALL call_2_3_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_6_15
#define init_2_3_6_15
void FASTCALL call_2_3_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_7_0
#define init_2_3_7_0
void FASTCALL call_2_3_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_7_1
#define init_2_3_7_1
void FASTCALL call_2_3_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_7_2
#define init_2_3_7_2
void FASTCALL call_2_3_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_7_3
#define init_2_3_7_3
void FASTCALL call_2_3_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_7_4
#define init_2_3_7_4
void FASTCALL call_2_3_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_7_5
#define init_2_3_7_5
void FASTCALL call_2_3_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_7_6
#define init_2_3_7_6
void FASTCALL call_2_3_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_7_7
#define init_2_3_7_7
void FASTCALL call_2_3_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_7_8
#define init_2_3_7_8
void FASTCALL call_2_3_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_7_9
#define init_2_3_7_9
void FASTCALL call_2_3_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_7_10
#define init_2_3_7_10
void FASTCALL call_2_3_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_7_11
#define init_2_3_7_11
void FASTCALL call_2_3_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_7_12
#define init_2_3_7_12
void FASTCALL call_2_3_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_7_13
#define init_2_3_7_13
void FASTCALL call_2_3_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_7_14
#define init_2_3_7_14
void FASTCALL call_2_3_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_7_15
#define init_2_3_7_15
void FASTCALL call_2_3_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_8_0
#define init_2_3_8_0
void FASTCALL call_2_3_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_8_1
#define init_2_3_8_1
void FASTCALL call_2_3_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_8_2
#define init_2_3_8_2
void FASTCALL call_2_3_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_8_3
#define init_2_3_8_3
void FASTCALL call_2_3_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_8_4
#define init_2_3_8_4
void FASTCALL call_2_3_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_8_5
#define init_2_3_8_5
void FASTCALL call_2_3_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_8_6
#define init_2_3_8_6
void FASTCALL call_2_3_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_8_7
#define init_2_3_8_7
void FASTCALL call_2_3_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_8_8
#define init_2_3_8_8
void FASTCALL call_2_3_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_8_9
#define init_2_3_8_9
void FASTCALL call_2_3_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_8_10
#define init_2_3_8_10
void FASTCALL call_2_3_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_8_11
#define init_2_3_8_11
void FASTCALL call_2_3_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_8_12
#define init_2_3_8_12
void FASTCALL call_2_3_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_8_13
#define init_2_3_8_13
void FASTCALL call_2_3_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_8_14
#define init_2_3_8_14
void FASTCALL call_2_3_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_8_15
#define init_2_3_8_15
void FASTCALL call_2_3_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_9_0
#define init_2_3_9_0
void FASTCALL call_2_3_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_9_1
#define init_2_3_9_1
void FASTCALL call_2_3_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_9_2
#define init_2_3_9_2
void FASTCALL call_2_3_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_9_3
#define init_2_3_9_3
void FASTCALL call_2_3_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_9_4
#define init_2_3_9_4
void FASTCALL call_2_3_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_9_5
#define init_2_3_9_5
void FASTCALL call_2_3_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_9_6
#define init_2_3_9_6
void FASTCALL call_2_3_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_9_7
#define init_2_3_9_7
void FASTCALL call_2_3_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_9_8
#define init_2_3_9_8
void FASTCALL call_2_3_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_9_9
#define init_2_3_9_9
void FASTCALL call_2_3_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_9_10
#define init_2_3_9_10
void FASTCALL call_2_3_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_9_11
#define init_2_3_9_11
void FASTCALL call_2_3_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_9_12
#define init_2_3_9_12
void FASTCALL call_2_3_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_9_13
#define init_2_3_9_13
void FASTCALL call_2_3_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_9_14
#define init_2_3_9_14
void FASTCALL call_2_3_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_9_15
#define init_2_3_9_15
void FASTCALL call_2_3_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_10_0
#define init_2_3_10_0
void FASTCALL call_2_3_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_10_1
#define init_2_3_10_1
void FASTCALL call_2_3_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_10_2
#define init_2_3_10_2
void FASTCALL call_2_3_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_10_3
#define init_2_3_10_3
void FASTCALL call_2_3_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_10_4
#define init_2_3_10_4
void FASTCALL call_2_3_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_10_5
#define init_2_3_10_5
void FASTCALL call_2_3_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_10_6
#define init_2_3_10_6
void FASTCALL call_2_3_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_10_7
#define init_2_3_10_7
void FASTCALL call_2_3_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_10_8
#define init_2_3_10_8
void FASTCALL call_2_3_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_10_9
#define init_2_3_10_9
void FASTCALL call_2_3_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_10_10
#define init_2_3_10_10
void FASTCALL call_2_3_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_10_11
#define init_2_3_10_11
void FASTCALL call_2_3_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_10_12
#define init_2_3_10_12
void FASTCALL call_2_3_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_10_13
#define init_2_3_10_13
void FASTCALL call_2_3_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_10_14
#define init_2_3_10_14
void FASTCALL call_2_3_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_10_15
#define init_2_3_10_15
void FASTCALL call_2_3_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_11_0
#define init_2_3_11_0
void FASTCALL call_2_3_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_11_1
#define init_2_3_11_1
void FASTCALL call_2_3_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_11_2
#define init_2_3_11_2
void FASTCALL call_2_3_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_11_3
#define init_2_3_11_3
void FASTCALL call_2_3_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_11_4
#define init_2_3_11_4
void FASTCALL call_2_3_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_11_5
#define init_2_3_11_5
void FASTCALL call_2_3_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_11_6
#define init_2_3_11_6
void FASTCALL call_2_3_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_11_7
#define init_2_3_11_7
void FASTCALL call_2_3_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_11_8
#define init_2_3_11_8
void FASTCALL call_2_3_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_11_9
#define init_2_3_11_9
void FASTCALL call_2_3_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_11_10
#define init_2_3_11_10
void FASTCALL call_2_3_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_11_11
#define init_2_3_11_11
void FASTCALL call_2_3_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_11_12
#define init_2_3_11_12
void FASTCALL call_2_3_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_11_13
#define init_2_3_11_13
void FASTCALL call_2_3_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_11_14
#define init_2_3_11_14
void FASTCALL call_2_3_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_11_15
#define init_2_3_11_15
void FASTCALL call_2_3_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_12_0
#define init_2_3_12_0
void FASTCALL call_2_3_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_12_1
#define init_2_3_12_1
void FASTCALL call_2_3_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_12_2
#define init_2_3_12_2
void FASTCALL call_2_3_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_12_3
#define init_2_3_12_3
void FASTCALL call_2_3_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_12_4
#define init_2_3_12_4
void FASTCALL call_2_3_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_12_5
#define init_2_3_12_5
void FASTCALL call_2_3_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_12_6
#define init_2_3_12_6
void FASTCALL call_2_3_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_12_7
#define init_2_3_12_7
void FASTCALL call_2_3_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_12_8
#define init_2_3_12_8
void FASTCALL call_2_3_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_12_9
#define init_2_3_12_9
void FASTCALL call_2_3_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_12_10
#define init_2_3_12_10
void FASTCALL call_2_3_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_12_11
#define init_2_3_12_11
void FASTCALL call_2_3_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_12_12
#define init_2_3_12_12
void FASTCALL call_2_3_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_12_13
#define init_2_3_12_13
void FASTCALL call_2_3_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_12_14
#define init_2_3_12_14
void FASTCALL call_2_3_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_12_15
#define init_2_3_12_15
void FASTCALL call_2_3_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_13_0
#define init_2_3_13_0
void FASTCALL call_2_3_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_13_1
#define init_2_3_13_1
void FASTCALL call_2_3_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_13_2
#define init_2_3_13_2
void FASTCALL call_2_3_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_13_3
#define init_2_3_13_3
void FASTCALL call_2_3_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_13_4
#define init_2_3_13_4
void FASTCALL call_2_3_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_13_5
#define init_2_3_13_5
void FASTCALL call_2_3_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_13_6
#define init_2_3_13_6
void FASTCALL call_2_3_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_13_7
#define init_2_3_13_7
void FASTCALL call_2_3_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_13_8
#define init_2_3_13_8
void FASTCALL call_2_3_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_13_9
#define init_2_3_13_9
void FASTCALL call_2_3_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_13_10
#define init_2_3_13_10
void FASTCALL call_2_3_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_13_11
#define init_2_3_13_11
void FASTCALL call_2_3_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_13_12
#define init_2_3_13_12
void FASTCALL call_2_3_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_13_13
#define init_2_3_13_13
void FASTCALL call_2_3_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_13_14
#define init_2_3_13_14
void FASTCALL call_2_3_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_13_15
#define init_2_3_13_15
void FASTCALL call_2_3_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_14_0
#define init_2_3_14_0
void FASTCALL call_2_3_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_14_1
#define init_2_3_14_1
void FASTCALL call_2_3_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_14_2
#define init_2_3_14_2
void FASTCALL call_2_3_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_14_3
#define init_2_3_14_3
void FASTCALL call_2_3_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_14_4
#define init_2_3_14_4
void FASTCALL call_2_3_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_14_5
#define init_2_3_14_5
void FASTCALL call_2_3_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_14_6
#define init_2_3_14_6
void FASTCALL call_2_3_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_14_7
#define init_2_3_14_7
void FASTCALL call_2_3_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_14_8
#define init_2_3_14_8
void FASTCALL call_2_3_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_14_9
#define init_2_3_14_9
void FASTCALL call_2_3_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_14_10
#define init_2_3_14_10
void FASTCALL call_2_3_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_14_11
#define init_2_3_14_11
void FASTCALL call_2_3_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_14_12
#define init_2_3_14_12
void FASTCALL call_2_3_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_14_13
#define init_2_3_14_13
void FASTCALL call_2_3_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_14_14
#define init_2_3_14_14
void FASTCALL call_2_3_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_14_15
#define init_2_3_14_15
void FASTCALL call_2_3_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_15_0
#define init_2_3_15_0
void FASTCALL call_2_3_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_15_1
#define init_2_3_15_1
void FASTCALL call_2_3_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_15_2
#define init_2_3_15_2
void FASTCALL call_2_3_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_15_3
#define init_2_3_15_3
void FASTCALL call_2_3_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_15_4
#define init_2_3_15_4
void FASTCALL call_2_3_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_15_5
#define init_2_3_15_5
void FASTCALL call_2_3_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_15_6
#define init_2_3_15_6
void FASTCALL call_2_3_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_15_7
#define init_2_3_15_7
void FASTCALL call_2_3_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_15_8
#define init_2_3_15_8
void FASTCALL call_2_3_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_15_9
#define init_2_3_15_9
void FASTCALL call_2_3_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_15_10
#define init_2_3_15_10
void FASTCALL call_2_3_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_15_11
#define init_2_3_15_11
void FASTCALL call_2_3_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_15_12
#define init_2_3_15_12
void FASTCALL call_2_3_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_15_13
#define init_2_3_15_13
void FASTCALL call_2_3_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_15_14
#define init_2_3_15_14
void FASTCALL call_2_3_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_3_15_15
#define init_2_3_15_15
void FASTCALL call_2_3_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_0_0
#define init_2_4_0_0
void FASTCALL call_2_4_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_0_1
#define init_2_4_0_1
void FASTCALL call_2_4_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_0_2
#define init_2_4_0_2
void FASTCALL call_2_4_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_0_3
#define init_2_4_0_3
void FASTCALL call_2_4_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_0_4
#define init_2_4_0_4
void FASTCALL call_2_4_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_0_5
#define init_2_4_0_5
void FASTCALL call_2_4_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_0_6
#define init_2_4_0_6
void FASTCALL call_2_4_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_0_7
#define init_2_4_0_7
void FASTCALL call_2_4_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_0_8
#define init_2_4_0_8
void FASTCALL call_2_4_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_0_9
#define init_2_4_0_9
void FASTCALL call_2_4_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_0_10
#define init_2_4_0_10
void FASTCALL call_2_4_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_0_11
#define init_2_4_0_11
void FASTCALL call_2_4_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_0_12
#define init_2_4_0_12
void FASTCALL call_2_4_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_0_13
#define init_2_4_0_13
void FASTCALL call_2_4_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_0_14
#define init_2_4_0_14
void FASTCALL call_2_4_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_0_15
#define init_2_4_0_15
void FASTCALL call_2_4_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_1_0
#define init_2_4_1_0
void FASTCALL call_2_4_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_1_1
#define init_2_4_1_1
void FASTCALL call_2_4_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_1_2
#define init_2_4_1_2
void FASTCALL call_2_4_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_1_3
#define init_2_4_1_3
void FASTCALL call_2_4_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_1_4
#define init_2_4_1_4
void FASTCALL call_2_4_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_1_5
#define init_2_4_1_5
void FASTCALL call_2_4_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_1_6
#define init_2_4_1_6
void FASTCALL call_2_4_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_1_7
#define init_2_4_1_7
void FASTCALL call_2_4_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_1_8
#define init_2_4_1_8
void FASTCALL call_2_4_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_1_9
#define init_2_4_1_9
void FASTCALL call_2_4_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_1_10
#define init_2_4_1_10
void FASTCALL call_2_4_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_1_11
#define init_2_4_1_11
void FASTCALL call_2_4_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_1_12
#define init_2_4_1_12
void FASTCALL call_2_4_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_1_13
#define init_2_4_1_13
void FASTCALL call_2_4_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_1_14
#define init_2_4_1_14
void FASTCALL call_2_4_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_1_15
#define init_2_4_1_15
void FASTCALL call_2_4_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_2_0
#define init_2_4_2_0
void FASTCALL call_2_4_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_2_1
#define init_2_4_2_1
void FASTCALL call_2_4_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_2_2
#define init_2_4_2_2
void FASTCALL call_2_4_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_2_3
#define init_2_4_2_3
void FASTCALL call_2_4_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_2_4
#define init_2_4_2_4
void FASTCALL call_2_4_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_2_5
#define init_2_4_2_5
void FASTCALL call_2_4_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_2_6
#define init_2_4_2_6
void FASTCALL call_2_4_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_2_7
#define init_2_4_2_7
void FASTCALL call_2_4_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_2_8
#define init_2_4_2_8
void FASTCALL call_2_4_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_2_9
#define init_2_4_2_9
void FASTCALL call_2_4_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_2_10
#define init_2_4_2_10
void FASTCALL call_2_4_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_2_11
#define init_2_4_2_11
void FASTCALL call_2_4_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_2_12
#define init_2_4_2_12
void FASTCALL call_2_4_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_2_13
#define init_2_4_2_13
void FASTCALL call_2_4_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_2_14
#define init_2_4_2_14
void FASTCALL call_2_4_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_2_15
#define init_2_4_2_15
void FASTCALL call_2_4_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_3_0
#define init_2_4_3_0
void FASTCALL call_2_4_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_3_1
#define init_2_4_3_1
void FASTCALL call_2_4_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_3_2
#define init_2_4_3_2
void FASTCALL call_2_4_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_3_3
#define init_2_4_3_3
void FASTCALL call_2_4_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_3_4
#define init_2_4_3_4
void FASTCALL call_2_4_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_3_5
#define init_2_4_3_5
void FASTCALL call_2_4_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_3_6
#define init_2_4_3_6
void FASTCALL call_2_4_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_3_7
#define init_2_4_3_7
void FASTCALL call_2_4_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_3_8
#define init_2_4_3_8
void FASTCALL call_2_4_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_3_9
#define init_2_4_3_9
void FASTCALL call_2_4_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_3_10
#define init_2_4_3_10
void FASTCALL call_2_4_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_3_11
#define init_2_4_3_11
void FASTCALL call_2_4_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_3_12
#define init_2_4_3_12
void FASTCALL call_2_4_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_3_13
#define init_2_4_3_13
void FASTCALL call_2_4_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_3_14
#define init_2_4_3_14
void FASTCALL call_2_4_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_3_15
#define init_2_4_3_15
void FASTCALL call_2_4_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_4_0
#define init_2_4_4_0
void FASTCALL call_2_4_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_4_1
#define init_2_4_4_1
void FASTCALL call_2_4_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_4_2
#define init_2_4_4_2
void FASTCALL call_2_4_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_4_3
#define init_2_4_4_3
void FASTCALL call_2_4_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_4_4
#define init_2_4_4_4
void FASTCALL call_2_4_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_4_5
#define init_2_4_4_5
void FASTCALL call_2_4_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_4_6
#define init_2_4_4_6
void FASTCALL call_2_4_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_4_7
#define init_2_4_4_7
void FASTCALL call_2_4_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_4_8
#define init_2_4_4_8
void FASTCALL call_2_4_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_4_9
#define init_2_4_4_9
void FASTCALL call_2_4_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_4_10
#define init_2_4_4_10
void FASTCALL call_2_4_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_4_11
#define init_2_4_4_11
void FASTCALL call_2_4_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_4_12
#define init_2_4_4_12
void FASTCALL call_2_4_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_4_13
#define init_2_4_4_13
void FASTCALL call_2_4_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_4_14
#define init_2_4_4_14
void FASTCALL call_2_4_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_4_15
#define init_2_4_4_15
void FASTCALL call_2_4_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_5_0
#define init_2_4_5_0
void FASTCALL call_2_4_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_5_1
#define init_2_4_5_1
void FASTCALL call_2_4_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_5_2
#define init_2_4_5_2
void FASTCALL call_2_4_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_5_3
#define init_2_4_5_3
void FASTCALL call_2_4_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_5_4
#define init_2_4_5_4
void FASTCALL call_2_4_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_5_5
#define init_2_4_5_5
void FASTCALL call_2_4_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_5_6
#define init_2_4_5_6
void FASTCALL call_2_4_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_5_7
#define init_2_4_5_7
void FASTCALL call_2_4_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_5_8
#define init_2_4_5_8
void FASTCALL call_2_4_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_5_9
#define init_2_4_5_9
void FASTCALL call_2_4_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_5_10
#define init_2_4_5_10
void FASTCALL call_2_4_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_5_11
#define init_2_4_5_11
void FASTCALL call_2_4_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_5_12
#define init_2_4_5_12
void FASTCALL call_2_4_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_5_13
#define init_2_4_5_13
void FASTCALL call_2_4_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_5_14
#define init_2_4_5_14
void FASTCALL call_2_4_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_5_15
#define init_2_4_5_15
void FASTCALL call_2_4_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_6_0
#define init_2_4_6_0
void FASTCALL call_2_4_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_6_1
#define init_2_4_6_1
void FASTCALL call_2_4_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_6_2
#define init_2_4_6_2
void FASTCALL call_2_4_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_6_3
#define init_2_4_6_3
void FASTCALL call_2_4_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_6_4
#define init_2_4_6_4
void FASTCALL call_2_4_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_6_5
#define init_2_4_6_5
void FASTCALL call_2_4_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_6_6
#define init_2_4_6_6
void FASTCALL call_2_4_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_6_7
#define init_2_4_6_7
void FASTCALL call_2_4_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_6_8
#define init_2_4_6_8
void FASTCALL call_2_4_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_6_9
#define init_2_4_6_9
void FASTCALL call_2_4_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_6_10
#define init_2_4_6_10
void FASTCALL call_2_4_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_6_11
#define init_2_4_6_11
void FASTCALL call_2_4_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_6_12
#define init_2_4_6_12
void FASTCALL call_2_4_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_6_13
#define init_2_4_6_13
void FASTCALL call_2_4_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_6_14
#define init_2_4_6_14
void FASTCALL call_2_4_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_6_15
#define init_2_4_6_15
void FASTCALL call_2_4_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_7_0
#define init_2_4_7_0
void FASTCALL call_2_4_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_7_1
#define init_2_4_7_1
void FASTCALL call_2_4_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_7_2
#define init_2_4_7_2
void FASTCALL call_2_4_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_7_3
#define init_2_4_7_3
void FASTCALL call_2_4_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_7_4
#define init_2_4_7_4
void FASTCALL call_2_4_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_7_5
#define init_2_4_7_5
void FASTCALL call_2_4_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_7_6
#define init_2_4_7_6
void FASTCALL call_2_4_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_7_7
#define init_2_4_7_7
void FASTCALL call_2_4_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_7_8
#define init_2_4_7_8
void FASTCALL call_2_4_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_7_9
#define init_2_4_7_9
void FASTCALL call_2_4_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_7_10
#define init_2_4_7_10
void FASTCALL call_2_4_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_7_11
#define init_2_4_7_11
void FASTCALL call_2_4_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_7_12
#define init_2_4_7_12
void FASTCALL call_2_4_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_7_13
#define init_2_4_7_13
void FASTCALL call_2_4_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_7_14
#define init_2_4_7_14
void FASTCALL call_2_4_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_7_15
#define init_2_4_7_15
void FASTCALL call_2_4_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_8_0
#define init_2_4_8_0
void FASTCALL call_2_4_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_8_1
#define init_2_4_8_1
void FASTCALL call_2_4_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_8_2
#define init_2_4_8_2
void FASTCALL call_2_4_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_8_3
#define init_2_4_8_3
void FASTCALL call_2_4_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_8_4
#define init_2_4_8_4
void FASTCALL call_2_4_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_8_5
#define init_2_4_8_5
void FASTCALL call_2_4_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_8_6
#define init_2_4_8_6
void FASTCALL call_2_4_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_8_7
#define init_2_4_8_7
void FASTCALL call_2_4_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_8_8
#define init_2_4_8_8
void FASTCALL call_2_4_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_8_9
#define init_2_4_8_9
void FASTCALL call_2_4_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_8_10
#define init_2_4_8_10
void FASTCALL call_2_4_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_8_11
#define init_2_4_8_11
void FASTCALL call_2_4_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_8_12
#define init_2_4_8_12
void FASTCALL call_2_4_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_8_13
#define init_2_4_8_13
void FASTCALL call_2_4_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_8_14
#define init_2_4_8_14
void FASTCALL call_2_4_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_8_15
#define init_2_4_8_15
void FASTCALL call_2_4_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_9_0
#define init_2_4_9_0
void FASTCALL call_2_4_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_9_1
#define init_2_4_9_1
void FASTCALL call_2_4_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_9_2
#define init_2_4_9_2
void FASTCALL call_2_4_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_9_3
#define init_2_4_9_3
void FASTCALL call_2_4_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_9_4
#define init_2_4_9_4
void FASTCALL call_2_4_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_9_5
#define init_2_4_9_5
void FASTCALL call_2_4_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_9_6
#define init_2_4_9_6
void FASTCALL call_2_4_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_9_7
#define init_2_4_9_7
void FASTCALL call_2_4_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_9_8
#define init_2_4_9_8
void FASTCALL call_2_4_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_9_9
#define init_2_4_9_9
void FASTCALL call_2_4_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_9_10
#define init_2_4_9_10
void FASTCALL call_2_4_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_9_11
#define init_2_4_9_11
void FASTCALL call_2_4_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_9_12
#define init_2_4_9_12
void FASTCALL call_2_4_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_9_13
#define init_2_4_9_13
void FASTCALL call_2_4_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_9_14
#define init_2_4_9_14
void FASTCALL call_2_4_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_9_15
#define init_2_4_9_15
void FASTCALL call_2_4_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_10_0
#define init_2_4_10_0
void FASTCALL call_2_4_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_10_1
#define init_2_4_10_1
void FASTCALL call_2_4_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_10_2
#define init_2_4_10_2
void FASTCALL call_2_4_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_10_3
#define init_2_4_10_3
void FASTCALL call_2_4_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_10_4
#define init_2_4_10_4
void FASTCALL call_2_4_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_10_5
#define init_2_4_10_5
void FASTCALL call_2_4_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_10_6
#define init_2_4_10_6
void FASTCALL call_2_4_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_10_7
#define init_2_4_10_7
void FASTCALL call_2_4_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_10_8
#define init_2_4_10_8
void FASTCALL call_2_4_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_10_9
#define init_2_4_10_9
void FASTCALL call_2_4_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_10_10
#define init_2_4_10_10
void FASTCALL call_2_4_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_10_11
#define init_2_4_10_11
void FASTCALL call_2_4_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_10_12
#define init_2_4_10_12
void FASTCALL call_2_4_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_10_13
#define init_2_4_10_13
void FASTCALL call_2_4_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_10_14
#define init_2_4_10_14
void FASTCALL call_2_4_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_10_15
#define init_2_4_10_15
void FASTCALL call_2_4_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_11_0
#define init_2_4_11_0
void FASTCALL call_2_4_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_11_1
#define init_2_4_11_1
void FASTCALL call_2_4_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_11_2
#define init_2_4_11_2
void FASTCALL call_2_4_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_11_3
#define init_2_4_11_3
void FASTCALL call_2_4_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_11_4
#define init_2_4_11_4
void FASTCALL call_2_4_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_11_5
#define init_2_4_11_5
void FASTCALL call_2_4_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_11_6
#define init_2_4_11_6
void FASTCALL call_2_4_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_11_7
#define init_2_4_11_7
void FASTCALL call_2_4_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_11_8
#define init_2_4_11_8
void FASTCALL call_2_4_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_11_9
#define init_2_4_11_9
void FASTCALL call_2_4_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_11_10
#define init_2_4_11_10
void FASTCALL call_2_4_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_11_11
#define init_2_4_11_11
void FASTCALL call_2_4_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_11_12
#define init_2_4_11_12
void FASTCALL call_2_4_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_11_13
#define init_2_4_11_13
void FASTCALL call_2_4_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_11_14
#define init_2_4_11_14
void FASTCALL call_2_4_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_11_15
#define init_2_4_11_15
void FASTCALL call_2_4_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_12_0
#define init_2_4_12_0
void FASTCALL call_2_4_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_12_1
#define init_2_4_12_1
void FASTCALL call_2_4_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_12_2
#define init_2_4_12_2
void FASTCALL call_2_4_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_12_3
#define init_2_4_12_3
void FASTCALL call_2_4_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_12_4
#define init_2_4_12_4
void FASTCALL call_2_4_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_12_5
#define init_2_4_12_5
void FASTCALL call_2_4_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_12_6
#define init_2_4_12_6
void FASTCALL call_2_4_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_12_7
#define init_2_4_12_7
void FASTCALL call_2_4_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_12_8
#define init_2_4_12_8
void FASTCALL call_2_4_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_12_9
#define init_2_4_12_9
void FASTCALL call_2_4_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_12_10
#define init_2_4_12_10
void FASTCALL call_2_4_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_12_11
#define init_2_4_12_11
void FASTCALL call_2_4_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_12_12
#define init_2_4_12_12
void FASTCALL call_2_4_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_12_13
#define init_2_4_12_13
void FASTCALL call_2_4_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_12_14
#define init_2_4_12_14
void FASTCALL call_2_4_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_12_15
#define init_2_4_12_15
void FASTCALL call_2_4_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_13_0
#define init_2_4_13_0
void FASTCALL call_2_4_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_13_1
#define init_2_4_13_1
void FASTCALL call_2_4_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_13_2
#define init_2_4_13_2
void FASTCALL call_2_4_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_13_3
#define init_2_4_13_3
void FASTCALL call_2_4_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_13_4
#define init_2_4_13_4
void FASTCALL call_2_4_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_13_5
#define init_2_4_13_5
void FASTCALL call_2_4_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_13_6
#define init_2_4_13_6
void FASTCALL call_2_4_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_13_7
#define init_2_4_13_7
void FASTCALL call_2_4_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_13_8
#define init_2_4_13_8
void FASTCALL call_2_4_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_13_9
#define init_2_4_13_9
void FASTCALL call_2_4_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_13_10
#define init_2_4_13_10
void FASTCALL call_2_4_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_13_11
#define init_2_4_13_11
void FASTCALL call_2_4_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_13_12
#define init_2_4_13_12
void FASTCALL call_2_4_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_13_13
#define init_2_4_13_13
void FASTCALL call_2_4_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_13_14
#define init_2_4_13_14
void FASTCALL call_2_4_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_13_15
#define init_2_4_13_15
void FASTCALL call_2_4_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_14_0
#define init_2_4_14_0
void FASTCALL call_2_4_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_14_1
#define init_2_4_14_1
void FASTCALL call_2_4_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_14_2
#define init_2_4_14_2
void FASTCALL call_2_4_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_14_3
#define init_2_4_14_3
void FASTCALL call_2_4_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_14_4
#define init_2_4_14_4
void FASTCALL call_2_4_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_14_5
#define init_2_4_14_5
void FASTCALL call_2_4_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_14_6
#define init_2_4_14_6
void FASTCALL call_2_4_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_14_7
#define init_2_4_14_7
void FASTCALL call_2_4_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_14_8
#define init_2_4_14_8
void FASTCALL call_2_4_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_14_9
#define init_2_4_14_9
void FASTCALL call_2_4_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_14_10
#define init_2_4_14_10
void FASTCALL call_2_4_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_14_11
#define init_2_4_14_11
void FASTCALL call_2_4_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_14_12
#define init_2_4_14_12
void FASTCALL call_2_4_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_14_13
#define init_2_4_14_13
void FASTCALL call_2_4_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_14_14
#define init_2_4_14_14
void FASTCALL call_2_4_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_14_15
#define init_2_4_14_15
void FASTCALL call_2_4_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_15_0
#define init_2_4_15_0
void FASTCALL call_2_4_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_15_1
#define init_2_4_15_1
void FASTCALL call_2_4_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_15_2
#define init_2_4_15_2
void FASTCALL call_2_4_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_15_3
#define init_2_4_15_3
void FASTCALL call_2_4_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_15_4
#define init_2_4_15_4
void FASTCALL call_2_4_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_15_5
#define init_2_4_15_5
void FASTCALL call_2_4_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_15_6
#define init_2_4_15_6
void FASTCALL call_2_4_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_15_7
#define init_2_4_15_7
void FASTCALL call_2_4_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_15_8
#define init_2_4_15_8
void FASTCALL call_2_4_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_15_9
#define init_2_4_15_9
void FASTCALL call_2_4_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_15_10
#define init_2_4_15_10
void FASTCALL call_2_4_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_15_11
#define init_2_4_15_11
void FASTCALL call_2_4_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_15_12
#define init_2_4_15_12
void FASTCALL call_2_4_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_15_13
#define init_2_4_15_13
void FASTCALL call_2_4_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_15_14
#define init_2_4_15_14
void FASTCALL call_2_4_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_4_15_15
#define init_2_4_15_15
void FASTCALL call_2_4_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_0_0
#define init_2_5_0_0
void FASTCALL call_2_5_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_0_1
#define init_2_5_0_1
void FASTCALL call_2_5_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_0_2
#define init_2_5_0_2
void FASTCALL call_2_5_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_0_3
#define init_2_5_0_3
void FASTCALL call_2_5_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_0_4
#define init_2_5_0_4
void FASTCALL call_2_5_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_0_5
#define init_2_5_0_5
void FASTCALL call_2_5_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_0_6
#define init_2_5_0_6
void FASTCALL call_2_5_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_0_7
#define init_2_5_0_7
void FASTCALL call_2_5_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_0_8
#define init_2_5_0_8
void FASTCALL call_2_5_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_0_9
#define init_2_5_0_9
void FASTCALL call_2_5_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_0_10
#define init_2_5_0_10
void FASTCALL call_2_5_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_0_11
#define init_2_5_0_11
void FASTCALL call_2_5_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_0_12
#define init_2_5_0_12
void FASTCALL call_2_5_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_0_13
#define init_2_5_0_13
void FASTCALL call_2_5_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_0_14
#define init_2_5_0_14
void FASTCALL call_2_5_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_0_15
#define init_2_5_0_15
void FASTCALL call_2_5_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_1_0
#define init_2_5_1_0
void FASTCALL call_2_5_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_1_1
#define init_2_5_1_1
void FASTCALL call_2_5_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_1_2
#define init_2_5_1_2
void FASTCALL call_2_5_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_1_3
#define init_2_5_1_3
void FASTCALL call_2_5_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_1_4
#define init_2_5_1_4
void FASTCALL call_2_5_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_1_5
#define init_2_5_1_5
void FASTCALL call_2_5_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_1_6
#define init_2_5_1_6
void FASTCALL call_2_5_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_1_7
#define init_2_5_1_7
void FASTCALL call_2_5_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_1_8
#define init_2_5_1_8
void FASTCALL call_2_5_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_1_9
#define init_2_5_1_9
void FASTCALL call_2_5_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_1_10
#define init_2_5_1_10
void FASTCALL call_2_5_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_1_11
#define init_2_5_1_11
void FASTCALL call_2_5_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_1_12
#define init_2_5_1_12
void FASTCALL call_2_5_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_1_13
#define init_2_5_1_13
void FASTCALL call_2_5_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_1_14
#define init_2_5_1_14
void FASTCALL call_2_5_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_1_15
#define init_2_5_1_15
void FASTCALL call_2_5_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_2_0
#define init_2_5_2_0
void FASTCALL call_2_5_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_2_1
#define init_2_5_2_1
void FASTCALL call_2_5_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_2_2
#define init_2_5_2_2
void FASTCALL call_2_5_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_2_3
#define init_2_5_2_3
void FASTCALL call_2_5_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_2_4
#define init_2_5_2_4
void FASTCALL call_2_5_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_2_5
#define init_2_5_2_5
void FASTCALL call_2_5_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_2_6
#define init_2_5_2_6
void FASTCALL call_2_5_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_2_7
#define init_2_5_2_7
void FASTCALL call_2_5_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_2_8
#define init_2_5_2_8
void FASTCALL call_2_5_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_2_9
#define init_2_5_2_9
void FASTCALL call_2_5_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_2_10
#define init_2_5_2_10
void FASTCALL call_2_5_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_2_11
#define init_2_5_2_11
void FASTCALL call_2_5_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_2_12
#define init_2_5_2_12
void FASTCALL call_2_5_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_2_13
#define init_2_5_2_13
void FASTCALL call_2_5_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_2_14
#define init_2_5_2_14
void FASTCALL call_2_5_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_2_15
#define init_2_5_2_15
void FASTCALL call_2_5_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_3_0
#define init_2_5_3_0
void FASTCALL call_2_5_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_3_1
#define init_2_5_3_1
void FASTCALL call_2_5_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_3_2
#define init_2_5_3_2
void FASTCALL call_2_5_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_3_3
#define init_2_5_3_3
void FASTCALL call_2_5_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_3_4
#define init_2_5_3_4
void FASTCALL call_2_5_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_3_5
#define init_2_5_3_5
void FASTCALL call_2_5_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_3_6
#define init_2_5_3_6
void FASTCALL call_2_5_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_3_7
#define init_2_5_3_7
void FASTCALL call_2_5_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_3_8
#define init_2_5_3_8
void FASTCALL call_2_5_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_3_9
#define init_2_5_3_9
void FASTCALL call_2_5_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_3_10
#define init_2_5_3_10
void FASTCALL call_2_5_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_3_11
#define init_2_5_3_11
void FASTCALL call_2_5_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_3_12
#define init_2_5_3_12
void FASTCALL call_2_5_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_3_13
#define init_2_5_3_13
void FASTCALL call_2_5_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_3_14
#define init_2_5_3_14
void FASTCALL call_2_5_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_3_15
#define init_2_5_3_15
void FASTCALL call_2_5_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_4_0
#define init_2_5_4_0
void FASTCALL call_2_5_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_4_1
#define init_2_5_4_1
void FASTCALL call_2_5_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_4_2
#define init_2_5_4_2
void FASTCALL call_2_5_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_4_3
#define init_2_5_4_3
void FASTCALL call_2_5_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_4_4
#define init_2_5_4_4
void FASTCALL call_2_5_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_4_5
#define init_2_5_4_5
void FASTCALL call_2_5_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_4_6
#define init_2_5_4_6
void FASTCALL call_2_5_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_4_7
#define init_2_5_4_7
void FASTCALL call_2_5_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_4_8
#define init_2_5_4_8
void FASTCALL call_2_5_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_4_9
#define init_2_5_4_9
void FASTCALL call_2_5_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_4_10
#define init_2_5_4_10
void FASTCALL call_2_5_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_4_11
#define init_2_5_4_11
void FASTCALL call_2_5_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_4_12
#define init_2_5_4_12
void FASTCALL call_2_5_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_4_13
#define init_2_5_4_13
void FASTCALL call_2_5_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_4_14
#define init_2_5_4_14
void FASTCALL call_2_5_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_4_15
#define init_2_5_4_15
void FASTCALL call_2_5_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_5_0
#define init_2_5_5_0
void FASTCALL call_2_5_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_5_1
#define init_2_5_5_1
void FASTCALL call_2_5_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_5_2
#define init_2_5_5_2
void FASTCALL call_2_5_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_5_3
#define init_2_5_5_3
void FASTCALL call_2_5_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_5_4
#define init_2_5_5_4
void FASTCALL call_2_5_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_5_5
#define init_2_5_5_5
void FASTCALL call_2_5_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_5_6
#define init_2_5_5_6
void FASTCALL call_2_5_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_5_7
#define init_2_5_5_7
void FASTCALL call_2_5_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_5_8
#define init_2_5_5_8
void FASTCALL call_2_5_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_5_9
#define init_2_5_5_9
void FASTCALL call_2_5_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_5_10
#define init_2_5_5_10
void FASTCALL call_2_5_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_5_11
#define init_2_5_5_11
void FASTCALL call_2_5_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_5_12
#define init_2_5_5_12
void FASTCALL call_2_5_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_5_13
#define init_2_5_5_13
void FASTCALL call_2_5_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_5_14
#define init_2_5_5_14
void FASTCALL call_2_5_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_5_15
#define init_2_5_5_15
void FASTCALL call_2_5_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_6_0
#define init_2_5_6_0
void FASTCALL call_2_5_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_6_1
#define init_2_5_6_1
void FASTCALL call_2_5_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_6_2
#define init_2_5_6_2
void FASTCALL call_2_5_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_6_3
#define init_2_5_6_3
void FASTCALL call_2_5_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_6_4
#define init_2_5_6_4
void FASTCALL call_2_5_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_6_5
#define init_2_5_6_5
void FASTCALL call_2_5_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_6_6
#define init_2_5_6_6
void FASTCALL call_2_5_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_6_7
#define init_2_5_6_7
void FASTCALL call_2_5_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_6_8
#define init_2_5_6_8
void FASTCALL call_2_5_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_6_9
#define init_2_5_6_9
void FASTCALL call_2_5_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_6_10
#define init_2_5_6_10
void FASTCALL call_2_5_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_6_11
#define init_2_5_6_11
void FASTCALL call_2_5_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_6_12
#define init_2_5_6_12
void FASTCALL call_2_5_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_6_13
#define init_2_5_6_13
void FASTCALL call_2_5_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_6_14
#define init_2_5_6_14
void FASTCALL call_2_5_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_6_15
#define init_2_5_6_15
void FASTCALL call_2_5_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_7_0
#define init_2_5_7_0
void FASTCALL call_2_5_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_7_1
#define init_2_5_7_1
void FASTCALL call_2_5_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_7_2
#define init_2_5_7_2
void FASTCALL call_2_5_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_7_3
#define init_2_5_7_3
void FASTCALL call_2_5_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_7_4
#define init_2_5_7_4
void FASTCALL call_2_5_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_7_5
#define init_2_5_7_5
void FASTCALL call_2_5_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_7_6
#define init_2_5_7_6
void FASTCALL call_2_5_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_7_7
#define init_2_5_7_7
void FASTCALL call_2_5_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_7_8
#define init_2_5_7_8
void FASTCALL call_2_5_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_7_9
#define init_2_5_7_9
void FASTCALL call_2_5_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_7_10
#define init_2_5_7_10
void FASTCALL call_2_5_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_7_11
#define init_2_5_7_11
void FASTCALL call_2_5_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_7_12
#define init_2_5_7_12
void FASTCALL call_2_5_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_7_13
#define init_2_5_7_13
void FASTCALL call_2_5_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_7_14
#define init_2_5_7_14
void FASTCALL call_2_5_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_7_15
#define init_2_5_7_15
void FASTCALL call_2_5_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_8_0
#define init_2_5_8_0
void FASTCALL call_2_5_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_8_1
#define init_2_5_8_1
void FASTCALL call_2_5_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_8_2
#define init_2_5_8_2
void FASTCALL call_2_5_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_8_3
#define init_2_5_8_3
void FASTCALL call_2_5_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_8_4
#define init_2_5_8_4
void FASTCALL call_2_5_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_8_5
#define init_2_5_8_5
void FASTCALL call_2_5_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_8_6
#define init_2_5_8_6
void FASTCALL call_2_5_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_8_7
#define init_2_5_8_7
void FASTCALL call_2_5_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_8_8
#define init_2_5_8_8
void FASTCALL call_2_5_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_8_9
#define init_2_5_8_9
void FASTCALL call_2_5_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_8_10
#define init_2_5_8_10
void FASTCALL call_2_5_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_8_11
#define init_2_5_8_11
void FASTCALL call_2_5_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_8_12
#define init_2_5_8_12
void FASTCALL call_2_5_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_8_13
#define init_2_5_8_13
void FASTCALL call_2_5_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_8_14
#define init_2_5_8_14
void FASTCALL call_2_5_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_8_15
#define init_2_5_8_15
void FASTCALL call_2_5_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_9_0
#define init_2_5_9_0
void FASTCALL call_2_5_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_9_1
#define init_2_5_9_1
void FASTCALL call_2_5_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_9_2
#define init_2_5_9_2
void FASTCALL call_2_5_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_9_3
#define init_2_5_9_3
void FASTCALL call_2_5_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_9_4
#define init_2_5_9_4
void FASTCALL call_2_5_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_9_5
#define init_2_5_9_5
void FASTCALL call_2_5_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_9_6
#define init_2_5_9_6
void FASTCALL call_2_5_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_9_7
#define init_2_5_9_7
void FASTCALL call_2_5_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_9_8
#define init_2_5_9_8
void FASTCALL call_2_5_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_9_9
#define init_2_5_9_9
void FASTCALL call_2_5_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_9_10
#define init_2_5_9_10
void FASTCALL call_2_5_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_9_11
#define init_2_5_9_11
void FASTCALL call_2_5_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_9_12
#define init_2_5_9_12
void FASTCALL call_2_5_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_9_13
#define init_2_5_9_13
void FASTCALL call_2_5_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_9_14
#define init_2_5_9_14
void FASTCALL call_2_5_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_9_15
#define init_2_5_9_15
void FASTCALL call_2_5_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_10_0
#define init_2_5_10_0
void FASTCALL call_2_5_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_10_1
#define init_2_5_10_1
void FASTCALL call_2_5_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_10_2
#define init_2_5_10_2
void FASTCALL call_2_5_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_10_3
#define init_2_5_10_3
void FASTCALL call_2_5_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_10_4
#define init_2_5_10_4
void FASTCALL call_2_5_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_10_5
#define init_2_5_10_5
void FASTCALL call_2_5_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_10_6
#define init_2_5_10_6
void FASTCALL call_2_5_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_10_7
#define init_2_5_10_7
void FASTCALL call_2_5_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_10_8
#define init_2_5_10_8
void FASTCALL call_2_5_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_10_9
#define init_2_5_10_9
void FASTCALL call_2_5_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_10_10
#define init_2_5_10_10
void FASTCALL call_2_5_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_10_11
#define init_2_5_10_11
void FASTCALL call_2_5_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_10_12
#define init_2_5_10_12
void FASTCALL call_2_5_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_10_13
#define init_2_5_10_13
void FASTCALL call_2_5_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_10_14
#define init_2_5_10_14
void FASTCALL call_2_5_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_10_15
#define init_2_5_10_15
void FASTCALL call_2_5_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_11_0
#define init_2_5_11_0
void FASTCALL call_2_5_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_11_1
#define init_2_5_11_1
void FASTCALL call_2_5_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_11_2
#define init_2_5_11_2
void FASTCALL call_2_5_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_11_3
#define init_2_5_11_3
void FASTCALL call_2_5_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_11_4
#define init_2_5_11_4
void FASTCALL call_2_5_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_11_5
#define init_2_5_11_5
void FASTCALL call_2_5_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_11_6
#define init_2_5_11_6
void FASTCALL call_2_5_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_11_7
#define init_2_5_11_7
void FASTCALL call_2_5_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_11_8
#define init_2_5_11_8
void FASTCALL call_2_5_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_11_9
#define init_2_5_11_9
void FASTCALL call_2_5_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_11_10
#define init_2_5_11_10
void FASTCALL call_2_5_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_11_11
#define init_2_5_11_11
void FASTCALL call_2_5_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_11_12
#define init_2_5_11_12
void FASTCALL call_2_5_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_11_13
#define init_2_5_11_13
void FASTCALL call_2_5_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_11_14
#define init_2_5_11_14
void FASTCALL call_2_5_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_11_15
#define init_2_5_11_15
void FASTCALL call_2_5_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_12_0
#define init_2_5_12_0
void FASTCALL call_2_5_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_12_1
#define init_2_5_12_1
void FASTCALL call_2_5_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_12_2
#define init_2_5_12_2
void FASTCALL call_2_5_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_12_3
#define init_2_5_12_3
void FASTCALL call_2_5_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_12_4
#define init_2_5_12_4
void FASTCALL call_2_5_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_12_5
#define init_2_5_12_5
void FASTCALL call_2_5_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_12_6
#define init_2_5_12_6
void FASTCALL call_2_5_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_12_7
#define init_2_5_12_7
void FASTCALL call_2_5_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_12_8
#define init_2_5_12_8
void FASTCALL call_2_5_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_12_9
#define init_2_5_12_9
void FASTCALL call_2_5_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_12_10
#define init_2_5_12_10
void FASTCALL call_2_5_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_12_11
#define init_2_5_12_11
void FASTCALL call_2_5_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_12_12
#define init_2_5_12_12
void FASTCALL call_2_5_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_12_13
#define init_2_5_12_13
void FASTCALL call_2_5_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_12_14
#define init_2_5_12_14
void FASTCALL call_2_5_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_12_15
#define init_2_5_12_15
void FASTCALL call_2_5_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_13_0
#define init_2_5_13_0
void FASTCALL call_2_5_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_13_1
#define init_2_5_13_1
void FASTCALL call_2_5_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_13_2
#define init_2_5_13_2
void FASTCALL call_2_5_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_13_3
#define init_2_5_13_3
void FASTCALL call_2_5_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_13_4
#define init_2_5_13_4
void FASTCALL call_2_5_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_13_5
#define init_2_5_13_5
void FASTCALL call_2_5_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_13_6
#define init_2_5_13_6
void FASTCALL call_2_5_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_13_7
#define init_2_5_13_7
void FASTCALL call_2_5_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_13_8
#define init_2_5_13_8
void FASTCALL call_2_5_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_13_9
#define init_2_5_13_9
void FASTCALL call_2_5_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_13_10
#define init_2_5_13_10
void FASTCALL call_2_5_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_13_11
#define init_2_5_13_11
void FASTCALL call_2_5_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_13_12
#define init_2_5_13_12
void FASTCALL call_2_5_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_13_13
#define init_2_5_13_13
void FASTCALL call_2_5_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_13_14
#define init_2_5_13_14
void FASTCALL call_2_5_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_13_15
#define init_2_5_13_15
void FASTCALL call_2_5_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_14_0
#define init_2_5_14_0
void FASTCALL call_2_5_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_14_1
#define init_2_5_14_1
void FASTCALL call_2_5_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_14_2
#define init_2_5_14_2
void FASTCALL call_2_5_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_14_3
#define init_2_5_14_3
void FASTCALL call_2_5_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_14_4
#define init_2_5_14_4
void FASTCALL call_2_5_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_14_5
#define init_2_5_14_5
void FASTCALL call_2_5_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_14_6
#define init_2_5_14_6
void FASTCALL call_2_5_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_14_7
#define init_2_5_14_7
void FASTCALL call_2_5_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_14_8
#define init_2_5_14_8
void FASTCALL call_2_5_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_14_9
#define init_2_5_14_9
void FASTCALL call_2_5_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_14_10
#define init_2_5_14_10
void FASTCALL call_2_5_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_14_11
#define init_2_5_14_11
void FASTCALL call_2_5_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_14_12
#define init_2_5_14_12
void FASTCALL call_2_5_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_14_13
#define init_2_5_14_13
void FASTCALL call_2_5_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_14_14
#define init_2_5_14_14
void FASTCALL call_2_5_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_14_15
#define init_2_5_14_15
void FASTCALL call_2_5_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_15_0
#define init_2_5_15_0
void FASTCALL call_2_5_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_15_1
#define init_2_5_15_1
void FASTCALL call_2_5_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_15_2
#define init_2_5_15_2
void FASTCALL call_2_5_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_15_3
#define init_2_5_15_3
void FASTCALL call_2_5_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_15_4
#define init_2_5_15_4
void FASTCALL call_2_5_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_15_5
#define init_2_5_15_5
void FASTCALL call_2_5_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_15_6
#define init_2_5_15_6
void FASTCALL call_2_5_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_15_7
#define init_2_5_15_7
void FASTCALL call_2_5_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_15_8
#define init_2_5_15_8
void FASTCALL call_2_5_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_15_9
#define init_2_5_15_9
void FASTCALL call_2_5_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_15_10
#define init_2_5_15_10
void FASTCALL call_2_5_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_15_11
#define init_2_5_15_11
void FASTCALL call_2_5_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_15_12
#define init_2_5_15_12
void FASTCALL call_2_5_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_15_13
#define init_2_5_15_13
void FASTCALL call_2_5_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_15_14
#define init_2_5_15_14
void FASTCALL call_2_5_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_5_15_15
#define init_2_5_15_15
void FASTCALL call_2_5_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_0_0
#define init_2_6_0_0
void FASTCALL call_2_6_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_0_1
#define init_2_6_0_1
void FASTCALL call_2_6_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_0_2
#define init_2_6_0_2
void FASTCALL call_2_6_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_0_3
#define init_2_6_0_3
void FASTCALL call_2_6_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_0_4
#define init_2_6_0_4
void FASTCALL call_2_6_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_0_5
#define init_2_6_0_5
void FASTCALL call_2_6_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_0_6
#define init_2_6_0_6
void FASTCALL call_2_6_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_0_7
#define init_2_6_0_7
void FASTCALL call_2_6_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_0_8
#define init_2_6_0_8
void FASTCALL call_2_6_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_0_9
#define init_2_6_0_9
void FASTCALL call_2_6_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_0_10
#define init_2_6_0_10
void FASTCALL call_2_6_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_0_11
#define init_2_6_0_11
void FASTCALL call_2_6_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_0_12
#define init_2_6_0_12
void FASTCALL call_2_6_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_0_13
#define init_2_6_0_13
void FASTCALL call_2_6_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_0_14
#define init_2_6_0_14
void FASTCALL call_2_6_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_0_15
#define init_2_6_0_15
void FASTCALL call_2_6_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_1_0
#define init_2_6_1_0
void FASTCALL call_2_6_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_1_1
#define init_2_6_1_1
void FASTCALL call_2_6_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_1_2
#define init_2_6_1_2
void FASTCALL call_2_6_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_1_3
#define init_2_6_1_3
void FASTCALL call_2_6_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_1_4
#define init_2_6_1_4
void FASTCALL call_2_6_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_1_5
#define init_2_6_1_5
void FASTCALL call_2_6_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_1_6
#define init_2_6_1_6
void FASTCALL call_2_6_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_1_7
#define init_2_6_1_7
void FASTCALL call_2_6_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_1_8
#define init_2_6_1_8
void FASTCALL call_2_6_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_1_9
#define init_2_6_1_9
void FASTCALL call_2_6_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_1_10
#define init_2_6_1_10
void FASTCALL call_2_6_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_1_11
#define init_2_6_1_11
void FASTCALL call_2_6_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_1_12
#define init_2_6_1_12
void FASTCALL call_2_6_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_1_13
#define init_2_6_1_13
void FASTCALL call_2_6_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_1_14
#define init_2_6_1_14
void FASTCALL call_2_6_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_1_15
#define init_2_6_1_15
void FASTCALL call_2_6_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_2_0
#define init_2_6_2_0
void FASTCALL call_2_6_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_2_1
#define init_2_6_2_1
void FASTCALL call_2_6_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_2_2
#define init_2_6_2_2
void FASTCALL call_2_6_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_2_3
#define init_2_6_2_3
void FASTCALL call_2_6_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_2_4
#define init_2_6_2_4
void FASTCALL call_2_6_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_2_5
#define init_2_6_2_5
void FASTCALL call_2_6_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_2_6
#define init_2_6_2_6
void FASTCALL call_2_6_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_2_7
#define init_2_6_2_7
void FASTCALL call_2_6_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_2_8
#define init_2_6_2_8
void FASTCALL call_2_6_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_2_9
#define init_2_6_2_9
void FASTCALL call_2_6_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_2_10
#define init_2_6_2_10
void FASTCALL call_2_6_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_2_11
#define init_2_6_2_11
void FASTCALL call_2_6_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_2_12
#define init_2_6_2_12
void FASTCALL call_2_6_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_2_13
#define init_2_6_2_13
void FASTCALL call_2_6_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_2_14
#define init_2_6_2_14
void FASTCALL call_2_6_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_2_15
#define init_2_6_2_15
void FASTCALL call_2_6_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_3_0
#define init_2_6_3_0
void FASTCALL call_2_6_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_3_1
#define init_2_6_3_1
void FASTCALL call_2_6_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_3_2
#define init_2_6_3_2
void FASTCALL call_2_6_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_3_3
#define init_2_6_3_3
void FASTCALL call_2_6_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_3_4
#define init_2_6_3_4
void FASTCALL call_2_6_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_3_5
#define init_2_6_3_5
void FASTCALL call_2_6_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_3_6
#define init_2_6_3_6
void FASTCALL call_2_6_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_3_7
#define init_2_6_3_7
void FASTCALL call_2_6_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_3_8
#define init_2_6_3_8
void FASTCALL call_2_6_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_3_9
#define init_2_6_3_9
void FASTCALL call_2_6_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_3_10
#define init_2_6_3_10
void FASTCALL call_2_6_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_3_11
#define init_2_6_3_11
void FASTCALL call_2_6_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_3_12
#define init_2_6_3_12
void FASTCALL call_2_6_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_3_13
#define init_2_6_3_13
void FASTCALL call_2_6_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_3_14
#define init_2_6_3_14
void FASTCALL call_2_6_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_3_15
#define init_2_6_3_15
void FASTCALL call_2_6_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_4_0
#define init_2_6_4_0
void FASTCALL call_2_6_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_4_1
#define init_2_6_4_1
void FASTCALL call_2_6_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_4_2
#define init_2_6_4_2
void FASTCALL call_2_6_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_4_3
#define init_2_6_4_3
void FASTCALL call_2_6_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_4_4
#define init_2_6_4_4
void FASTCALL call_2_6_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_4_5
#define init_2_6_4_5
void FASTCALL call_2_6_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_4_6
#define init_2_6_4_6
void FASTCALL call_2_6_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_4_7
#define init_2_6_4_7
void FASTCALL call_2_6_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_4_8
#define init_2_6_4_8
void FASTCALL call_2_6_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_4_9
#define init_2_6_4_9
void FASTCALL call_2_6_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_4_10
#define init_2_6_4_10
void FASTCALL call_2_6_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_4_11
#define init_2_6_4_11
void FASTCALL call_2_6_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_4_12
#define init_2_6_4_12
void FASTCALL call_2_6_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_4_13
#define init_2_6_4_13
void FASTCALL call_2_6_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_4_14
#define init_2_6_4_14
void FASTCALL call_2_6_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_4_15
#define init_2_6_4_15
void FASTCALL call_2_6_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_5_0
#define init_2_6_5_0
void FASTCALL call_2_6_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_5_1
#define init_2_6_5_1
void FASTCALL call_2_6_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_5_2
#define init_2_6_5_2
void FASTCALL call_2_6_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_5_3
#define init_2_6_5_3
void FASTCALL call_2_6_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_5_4
#define init_2_6_5_4
void FASTCALL call_2_6_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_5_5
#define init_2_6_5_5
void FASTCALL call_2_6_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_5_6
#define init_2_6_5_6
void FASTCALL call_2_6_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_5_7
#define init_2_6_5_7
void FASTCALL call_2_6_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_5_8
#define init_2_6_5_8
void FASTCALL call_2_6_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_5_9
#define init_2_6_5_9
void FASTCALL call_2_6_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_5_10
#define init_2_6_5_10
void FASTCALL call_2_6_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_5_11
#define init_2_6_5_11
void FASTCALL call_2_6_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_5_12
#define init_2_6_5_12
void FASTCALL call_2_6_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_5_13
#define init_2_6_5_13
void FASTCALL call_2_6_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_5_14
#define init_2_6_5_14
void FASTCALL call_2_6_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_5_15
#define init_2_6_5_15
void FASTCALL call_2_6_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_6_0
#define init_2_6_6_0
void FASTCALL call_2_6_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_6_1
#define init_2_6_6_1
void FASTCALL call_2_6_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_6_2
#define init_2_6_6_2
void FASTCALL call_2_6_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_6_3
#define init_2_6_6_3
void FASTCALL call_2_6_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_6_4
#define init_2_6_6_4
void FASTCALL call_2_6_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_6_5
#define init_2_6_6_5
void FASTCALL call_2_6_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_6_6
#define init_2_6_6_6
void FASTCALL call_2_6_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_6_7
#define init_2_6_6_7
void FASTCALL call_2_6_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_6_8
#define init_2_6_6_8
void FASTCALL call_2_6_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_6_9
#define init_2_6_6_9
void FASTCALL call_2_6_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_6_10
#define init_2_6_6_10
void FASTCALL call_2_6_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_6_11
#define init_2_6_6_11
void FASTCALL call_2_6_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_6_12
#define init_2_6_6_12
void FASTCALL call_2_6_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_6_13
#define init_2_6_6_13
void FASTCALL call_2_6_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_6_14
#define init_2_6_6_14
void FASTCALL call_2_6_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_6_15
#define init_2_6_6_15
void FASTCALL call_2_6_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_7_0
#define init_2_6_7_0
void FASTCALL call_2_6_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_7_1
#define init_2_6_7_1
void FASTCALL call_2_6_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_7_2
#define init_2_6_7_2
void FASTCALL call_2_6_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_7_3
#define init_2_6_7_3
void FASTCALL call_2_6_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_7_4
#define init_2_6_7_4
void FASTCALL call_2_6_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_7_5
#define init_2_6_7_5
void FASTCALL call_2_6_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_7_6
#define init_2_6_7_6
void FASTCALL call_2_6_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_7_7
#define init_2_6_7_7
void FASTCALL call_2_6_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_7_8
#define init_2_6_7_8
void FASTCALL call_2_6_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_7_9
#define init_2_6_7_9
void FASTCALL call_2_6_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_7_10
#define init_2_6_7_10
void FASTCALL call_2_6_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_7_11
#define init_2_6_7_11
void FASTCALL call_2_6_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_7_12
#define init_2_6_7_12
void FASTCALL call_2_6_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_7_13
#define init_2_6_7_13
void FASTCALL call_2_6_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_7_14
#define init_2_6_7_14
void FASTCALL call_2_6_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_7_15
#define init_2_6_7_15
void FASTCALL call_2_6_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_8_0
#define init_2_6_8_0
void FASTCALL call_2_6_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_8_1
#define init_2_6_8_1
void FASTCALL call_2_6_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_8_2
#define init_2_6_8_2
void FASTCALL call_2_6_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_8_3
#define init_2_6_8_3
void FASTCALL call_2_6_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_8_4
#define init_2_6_8_4
void FASTCALL call_2_6_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_8_5
#define init_2_6_8_5
void FASTCALL call_2_6_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_8_6
#define init_2_6_8_6
void FASTCALL call_2_6_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_8_7
#define init_2_6_8_7
void FASTCALL call_2_6_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_8_8
#define init_2_6_8_8
void FASTCALL call_2_6_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_8_9
#define init_2_6_8_9
void FASTCALL call_2_6_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_8_10
#define init_2_6_8_10
void FASTCALL call_2_6_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_8_11
#define init_2_6_8_11
void FASTCALL call_2_6_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_8_12
#define init_2_6_8_12
void FASTCALL call_2_6_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_8_13
#define init_2_6_8_13
void FASTCALL call_2_6_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_8_14
#define init_2_6_8_14
void FASTCALL call_2_6_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_8_15
#define init_2_6_8_15
void FASTCALL call_2_6_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_9_0
#define init_2_6_9_0
void FASTCALL call_2_6_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_9_1
#define init_2_6_9_1
void FASTCALL call_2_6_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_9_2
#define init_2_6_9_2
void FASTCALL call_2_6_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_9_3
#define init_2_6_9_3
void FASTCALL call_2_6_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_9_4
#define init_2_6_9_4
void FASTCALL call_2_6_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_9_5
#define init_2_6_9_5
void FASTCALL call_2_6_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_9_6
#define init_2_6_9_6
void FASTCALL call_2_6_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_9_7
#define init_2_6_9_7
void FASTCALL call_2_6_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_9_8
#define init_2_6_9_8
void FASTCALL call_2_6_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_9_9
#define init_2_6_9_9
void FASTCALL call_2_6_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_9_10
#define init_2_6_9_10
void FASTCALL call_2_6_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_9_11
#define init_2_6_9_11
void FASTCALL call_2_6_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_9_12
#define init_2_6_9_12
void FASTCALL call_2_6_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_9_13
#define init_2_6_9_13
void FASTCALL call_2_6_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_9_14
#define init_2_6_9_14
void FASTCALL call_2_6_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_9_15
#define init_2_6_9_15
void FASTCALL call_2_6_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_10_0
#define init_2_6_10_0
void FASTCALL call_2_6_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_10_1
#define init_2_6_10_1
void FASTCALL call_2_6_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_10_2
#define init_2_6_10_2
void FASTCALL call_2_6_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_10_3
#define init_2_6_10_3
void FASTCALL call_2_6_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_10_4
#define init_2_6_10_4
void FASTCALL call_2_6_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_10_5
#define init_2_6_10_5
void FASTCALL call_2_6_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_10_6
#define init_2_6_10_6
void FASTCALL call_2_6_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_10_7
#define init_2_6_10_7
void FASTCALL call_2_6_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_10_8
#define init_2_6_10_8
void FASTCALL call_2_6_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_10_9
#define init_2_6_10_9
void FASTCALL call_2_6_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_10_10
#define init_2_6_10_10
void FASTCALL call_2_6_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_10_11
#define init_2_6_10_11
void FASTCALL call_2_6_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_10_12
#define init_2_6_10_12
void FASTCALL call_2_6_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_10_13
#define init_2_6_10_13
void FASTCALL call_2_6_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_10_14
#define init_2_6_10_14
void FASTCALL call_2_6_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_10_15
#define init_2_6_10_15
void FASTCALL call_2_6_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_11_0
#define init_2_6_11_0
void FASTCALL call_2_6_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_11_1
#define init_2_6_11_1
void FASTCALL call_2_6_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_11_2
#define init_2_6_11_2
void FASTCALL call_2_6_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_11_3
#define init_2_6_11_3
void FASTCALL call_2_6_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_11_4
#define init_2_6_11_4
void FASTCALL call_2_6_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_11_5
#define init_2_6_11_5
void FASTCALL call_2_6_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_11_6
#define init_2_6_11_6
void FASTCALL call_2_6_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_11_7
#define init_2_6_11_7
void FASTCALL call_2_6_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_11_8
#define init_2_6_11_8
void FASTCALL call_2_6_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_11_9
#define init_2_6_11_9
void FASTCALL call_2_6_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_11_10
#define init_2_6_11_10
void FASTCALL call_2_6_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_11_11
#define init_2_6_11_11
void FASTCALL call_2_6_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_11_12
#define init_2_6_11_12
void FASTCALL call_2_6_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_11_13
#define init_2_6_11_13
void FASTCALL call_2_6_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_11_14
#define init_2_6_11_14
void FASTCALL call_2_6_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_11_15
#define init_2_6_11_15
void FASTCALL call_2_6_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_12_0
#define init_2_6_12_0
void FASTCALL call_2_6_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_12_1
#define init_2_6_12_1
void FASTCALL call_2_6_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_12_2
#define init_2_6_12_2
void FASTCALL call_2_6_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_12_3
#define init_2_6_12_3
void FASTCALL call_2_6_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_12_4
#define init_2_6_12_4
void FASTCALL call_2_6_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_12_5
#define init_2_6_12_5
void FASTCALL call_2_6_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_12_6
#define init_2_6_12_6
void FASTCALL call_2_6_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_12_7
#define init_2_6_12_7
void FASTCALL call_2_6_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_12_8
#define init_2_6_12_8
void FASTCALL call_2_6_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_12_9
#define init_2_6_12_9
void FASTCALL call_2_6_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_12_10
#define init_2_6_12_10
void FASTCALL call_2_6_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_12_11
#define init_2_6_12_11
void FASTCALL call_2_6_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_12_12
#define init_2_6_12_12
void FASTCALL call_2_6_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_12_13
#define init_2_6_12_13
void FASTCALL call_2_6_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_12_14
#define init_2_6_12_14
void FASTCALL call_2_6_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_12_15
#define init_2_6_12_15
void FASTCALL call_2_6_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_13_0
#define init_2_6_13_0
void FASTCALL call_2_6_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_13_1
#define init_2_6_13_1
void FASTCALL call_2_6_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_13_2
#define init_2_6_13_2
void FASTCALL call_2_6_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_13_3
#define init_2_6_13_3
void FASTCALL call_2_6_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_13_4
#define init_2_6_13_4
void FASTCALL call_2_6_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_13_5
#define init_2_6_13_5
void FASTCALL call_2_6_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_13_6
#define init_2_6_13_6
void FASTCALL call_2_6_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_13_7
#define init_2_6_13_7
void FASTCALL call_2_6_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_13_8
#define init_2_6_13_8
void FASTCALL call_2_6_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_13_9
#define init_2_6_13_9
void FASTCALL call_2_6_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_13_10
#define init_2_6_13_10
void FASTCALL call_2_6_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_13_11
#define init_2_6_13_11
void FASTCALL call_2_6_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_13_12
#define init_2_6_13_12
void FASTCALL call_2_6_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_13_13
#define init_2_6_13_13
void FASTCALL call_2_6_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_13_14
#define init_2_6_13_14
void FASTCALL call_2_6_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_13_15
#define init_2_6_13_15
void FASTCALL call_2_6_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_14_0
#define init_2_6_14_0
void FASTCALL call_2_6_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_14_1
#define init_2_6_14_1
void FASTCALL call_2_6_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_14_2
#define init_2_6_14_2
void FASTCALL call_2_6_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_14_3
#define init_2_6_14_3
void FASTCALL call_2_6_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_14_4
#define init_2_6_14_4
void FASTCALL call_2_6_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_14_5
#define init_2_6_14_5
void FASTCALL call_2_6_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_14_6
#define init_2_6_14_6
void FASTCALL call_2_6_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_14_7
#define init_2_6_14_7
void FASTCALL call_2_6_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_14_8
#define init_2_6_14_8
void FASTCALL call_2_6_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_14_9
#define init_2_6_14_9
void FASTCALL call_2_6_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_14_10
#define init_2_6_14_10
void FASTCALL call_2_6_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_14_11
#define init_2_6_14_11
void FASTCALL call_2_6_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_14_12
#define init_2_6_14_12
void FASTCALL call_2_6_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_14_13
#define init_2_6_14_13
void FASTCALL call_2_6_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_14_14
#define init_2_6_14_14
void FASTCALL call_2_6_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_14_15
#define init_2_6_14_15
void FASTCALL call_2_6_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_15_0
#define init_2_6_15_0
void FASTCALL call_2_6_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_15_1
#define init_2_6_15_1
void FASTCALL call_2_6_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_15_2
#define init_2_6_15_2
void FASTCALL call_2_6_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_15_3
#define init_2_6_15_3
void FASTCALL call_2_6_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_15_4
#define init_2_6_15_4
void FASTCALL call_2_6_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_15_5
#define init_2_6_15_5
void FASTCALL call_2_6_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_15_6
#define init_2_6_15_6
void FASTCALL call_2_6_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_15_7
#define init_2_6_15_7
void FASTCALL call_2_6_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_15_8
#define init_2_6_15_8
void FASTCALL call_2_6_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_15_9
#define init_2_6_15_9
void FASTCALL call_2_6_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_15_10
#define init_2_6_15_10
void FASTCALL call_2_6_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_15_11
#define init_2_6_15_11
void FASTCALL call_2_6_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_15_12
#define init_2_6_15_12
void FASTCALL call_2_6_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_15_13
#define init_2_6_15_13
void FASTCALL call_2_6_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_15_14
#define init_2_6_15_14
void FASTCALL call_2_6_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_6_15_15
#define init_2_6_15_15
void FASTCALL call_2_6_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_0_0
#define init_2_7_0_0
void FASTCALL call_2_7_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_0_1
#define init_2_7_0_1
void FASTCALL call_2_7_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_0_2
#define init_2_7_0_2
void FASTCALL call_2_7_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_0_3
#define init_2_7_0_3
void FASTCALL call_2_7_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_0_4
#define init_2_7_0_4
void FASTCALL call_2_7_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_0_5
#define init_2_7_0_5
void FASTCALL call_2_7_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_0_6
#define init_2_7_0_6
void FASTCALL call_2_7_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_0_7
#define init_2_7_0_7
void FASTCALL call_2_7_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_0_8
#define init_2_7_0_8
void FASTCALL call_2_7_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_0_9
#define init_2_7_0_9
void FASTCALL call_2_7_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_0_10
#define init_2_7_0_10
void FASTCALL call_2_7_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_0_11
#define init_2_7_0_11
void FASTCALL call_2_7_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_0_12
#define init_2_7_0_12
void FASTCALL call_2_7_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_0_13
#define init_2_7_0_13
void FASTCALL call_2_7_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_0_14
#define init_2_7_0_14
void FASTCALL call_2_7_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_0_15
#define init_2_7_0_15
void FASTCALL call_2_7_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_1_0
#define init_2_7_1_0
void FASTCALL call_2_7_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_1_1
#define init_2_7_1_1
void FASTCALL call_2_7_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_1_2
#define init_2_7_1_2
void FASTCALL call_2_7_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_1_3
#define init_2_7_1_3
void FASTCALL call_2_7_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_1_4
#define init_2_7_1_4
void FASTCALL call_2_7_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_1_5
#define init_2_7_1_5
void FASTCALL call_2_7_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_1_6
#define init_2_7_1_6
void FASTCALL call_2_7_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_1_7
#define init_2_7_1_7
void FASTCALL call_2_7_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_1_8
#define init_2_7_1_8
void FASTCALL call_2_7_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_1_9
#define init_2_7_1_9
void FASTCALL call_2_7_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_1_10
#define init_2_7_1_10
void FASTCALL call_2_7_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_1_11
#define init_2_7_1_11
void FASTCALL call_2_7_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_1_12
#define init_2_7_1_12
void FASTCALL call_2_7_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_1_13
#define init_2_7_1_13
void FASTCALL call_2_7_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_1_14
#define init_2_7_1_14
void FASTCALL call_2_7_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_1_15
#define init_2_7_1_15
void FASTCALL call_2_7_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_2_0
#define init_2_7_2_0
void FASTCALL call_2_7_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_2_1
#define init_2_7_2_1
void FASTCALL call_2_7_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_2_2
#define init_2_7_2_2
void FASTCALL call_2_7_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_2_3
#define init_2_7_2_3
void FASTCALL call_2_7_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_2_4
#define init_2_7_2_4
void FASTCALL call_2_7_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_2_5
#define init_2_7_2_5
void FASTCALL call_2_7_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_2_6
#define init_2_7_2_6
void FASTCALL call_2_7_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_2_7
#define init_2_7_2_7
void FASTCALL call_2_7_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_2_8
#define init_2_7_2_8
void FASTCALL call_2_7_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_2_9
#define init_2_7_2_9
void FASTCALL call_2_7_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_2_10
#define init_2_7_2_10
void FASTCALL call_2_7_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_2_11
#define init_2_7_2_11
void FASTCALL call_2_7_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_2_12
#define init_2_7_2_12
void FASTCALL call_2_7_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_2_13
#define init_2_7_2_13
void FASTCALL call_2_7_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_2_14
#define init_2_7_2_14
void FASTCALL call_2_7_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_2_15
#define init_2_7_2_15
void FASTCALL call_2_7_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_3_0
#define init_2_7_3_0
void FASTCALL call_2_7_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_3_1
#define init_2_7_3_1
void FASTCALL call_2_7_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_3_2
#define init_2_7_3_2
void FASTCALL call_2_7_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_3_3
#define init_2_7_3_3
void FASTCALL call_2_7_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_3_4
#define init_2_7_3_4
void FASTCALL call_2_7_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_3_5
#define init_2_7_3_5
void FASTCALL call_2_7_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_3_6
#define init_2_7_3_6
void FASTCALL call_2_7_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_3_7
#define init_2_7_3_7
void FASTCALL call_2_7_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_3_8
#define init_2_7_3_8
void FASTCALL call_2_7_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_3_9
#define init_2_7_3_9
void FASTCALL call_2_7_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_3_10
#define init_2_7_3_10
void FASTCALL call_2_7_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_3_11
#define init_2_7_3_11
void FASTCALL call_2_7_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_3_12
#define init_2_7_3_12
void FASTCALL call_2_7_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_3_13
#define init_2_7_3_13
void FASTCALL call_2_7_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_3_14
#define init_2_7_3_14
void FASTCALL call_2_7_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_3_15
#define init_2_7_3_15
void FASTCALL call_2_7_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_4_0
#define init_2_7_4_0
void FASTCALL call_2_7_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_4_1
#define init_2_7_4_1
void FASTCALL call_2_7_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_4_2
#define init_2_7_4_2
void FASTCALL call_2_7_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_4_3
#define init_2_7_4_3
void FASTCALL call_2_7_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_4_4
#define init_2_7_4_4
void FASTCALL call_2_7_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_4_5
#define init_2_7_4_5
void FASTCALL call_2_7_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_4_6
#define init_2_7_4_6
void FASTCALL call_2_7_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_4_7
#define init_2_7_4_7
void FASTCALL call_2_7_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_4_8
#define init_2_7_4_8
void FASTCALL call_2_7_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_4_9
#define init_2_7_4_9
void FASTCALL call_2_7_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_4_10
#define init_2_7_4_10
void FASTCALL call_2_7_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_4_11
#define init_2_7_4_11
void FASTCALL call_2_7_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_4_12
#define init_2_7_4_12
void FASTCALL call_2_7_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_4_13
#define init_2_7_4_13
void FASTCALL call_2_7_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_4_14
#define init_2_7_4_14
void FASTCALL call_2_7_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_4_15
#define init_2_7_4_15
void FASTCALL call_2_7_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_5_0
#define init_2_7_5_0
void FASTCALL call_2_7_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_5_1
#define init_2_7_5_1
void FASTCALL call_2_7_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_5_2
#define init_2_7_5_2
void FASTCALL call_2_7_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_5_3
#define init_2_7_5_3
void FASTCALL call_2_7_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_5_4
#define init_2_7_5_4
void FASTCALL call_2_7_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_5_5
#define init_2_7_5_5
void FASTCALL call_2_7_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_5_6
#define init_2_7_5_6
void FASTCALL call_2_7_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_5_7
#define init_2_7_5_7
void FASTCALL call_2_7_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_5_8
#define init_2_7_5_8
void FASTCALL call_2_7_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_5_9
#define init_2_7_5_9
void FASTCALL call_2_7_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_5_10
#define init_2_7_5_10
void FASTCALL call_2_7_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_5_11
#define init_2_7_5_11
void FASTCALL call_2_7_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_5_12
#define init_2_7_5_12
void FASTCALL call_2_7_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_5_13
#define init_2_7_5_13
void FASTCALL call_2_7_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_5_14
#define init_2_7_5_14
void FASTCALL call_2_7_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_5_15
#define init_2_7_5_15
void FASTCALL call_2_7_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_6_0
#define init_2_7_6_0
void FASTCALL call_2_7_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_6_1
#define init_2_7_6_1
void FASTCALL call_2_7_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_6_2
#define init_2_7_6_2
void FASTCALL call_2_7_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_6_3
#define init_2_7_6_3
void FASTCALL call_2_7_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_6_4
#define init_2_7_6_4
void FASTCALL call_2_7_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_6_5
#define init_2_7_6_5
void FASTCALL call_2_7_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_6_6
#define init_2_7_6_6
void FASTCALL call_2_7_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_6_7
#define init_2_7_6_7
void FASTCALL call_2_7_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_6_8
#define init_2_7_6_8
void FASTCALL call_2_7_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_6_9
#define init_2_7_6_9
void FASTCALL call_2_7_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_6_10
#define init_2_7_6_10
void FASTCALL call_2_7_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_6_11
#define init_2_7_6_11
void FASTCALL call_2_7_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_6_12
#define init_2_7_6_12
void FASTCALL call_2_7_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_6_13
#define init_2_7_6_13
void FASTCALL call_2_7_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_6_14
#define init_2_7_6_14
void FASTCALL call_2_7_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_6_15
#define init_2_7_6_15
void FASTCALL call_2_7_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_7_0
#define init_2_7_7_0
void FASTCALL call_2_7_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_7_1
#define init_2_7_7_1
void FASTCALL call_2_7_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_7_2
#define init_2_7_7_2
void FASTCALL call_2_7_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_7_3
#define init_2_7_7_3
void FASTCALL call_2_7_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_7_4
#define init_2_7_7_4
void FASTCALL call_2_7_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_7_5
#define init_2_7_7_5
void FASTCALL call_2_7_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_7_6
#define init_2_7_7_6
void FASTCALL call_2_7_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_7_7
#define init_2_7_7_7
void FASTCALL call_2_7_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_7_8
#define init_2_7_7_8
void FASTCALL call_2_7_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_7_9
#define init_2_7_7_9
void FASTCALL call_2_7_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_7_10
#define init_2_7_7_10
void FASTCALL call_2_7_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_7_11
#define init_2_7_7_11
void FASTCALL call_2_7_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_7_12
#define init_2_7_7_12
void FASTCALL call_2_7_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_7_13
#define init_2_7_7_13
void FASTCALL call_2_7_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_7_14
#define init_2_7_7_14
void FASTCALL call_2_7_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_7_15
#define init_2_7_7_15
void FASTCALL call_2_7_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_8_0
#define init_2_7_8_0
void FASTCALL call_2_7_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_8_1
#define init_2_7_8_1
void FASTCALL call_2_7_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_8_2
#define init_2_7_8_2
void FASTCALL call_2_7_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_8_3
#define init_2_7_8_3
void FASTCALL call_2_7_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_8_4
#define init_2_7_8_4
void FASTCALL call_2_7_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_8_5
#define init_2_7_8_5
void FASTCALL call_2_7_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_8_6
#define init_2_7_8_6
void FASTCALL call_2_7_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_8_7
#define init_2_7_8_7
void FASTCALL call_2_7_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_8_8
#define init_2_7_8_8
void FASTCALL call_2_7_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_8_9
#define init_2_7_8_9
void FASTCALL call_2_7_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_8_10
#define init_2_7_8_10
void FASTCALL call_2_7_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_8_11
#define init_2_7_8_11
void FASTCALL call_2_7_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_8_12
#define init_2_7_8_12
void FASTCALL call_2_7_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_8_13
#define init_2_7_8_13
void FASTCALL call_2_7_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_8_14
#define init_2_7_8_14
void FASTCALL call_2_7_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_8_15
#define init_2_7_8_15
void FASTCALL call_2_7_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_9_0
#define init_2_7_9_0
void FASTCALL call_2_7_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_9_1
#define init_2_7_9_1
void FASTCALL call_2_7_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_9_2
#define init_2_7_9_2
void FASTCALL call_2_7_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_9_3
#define init_2_7_9_3
void FASTCALL call_2_7_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_9_4
#define init_2_7_9_4
void FASTCALL call_2_7_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_9_5
#define init_2_7_9_5
void FASTCALL call_2_7_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_9_6
#define init_2_7_9_6
void FASTCALL call_2_7_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_9_7
#define init_2_7_9_7
void FASTCALL call_2_7_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_9_8
#define init_2_7_9_8
void FASTCALL call_2_7_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_9_9
#define init_2_7_9_9
void FASTCALL call_2_7_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_9_10
#define init_2_7_9_10
void FASTCALL call_2_7_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_9_11
#define init_2_7_9_11
void FASTCALL call_2_7_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_9_12
#define init_2_7_9_12
void FASTCALL call_2_7_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_9_13
#define init_2_7_9_13
void FASTCALL call_2_7_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_9_14
#define init_2_7_9_14
void FASTCALL call_2_7_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_9_15
#define init_2_7_9_15
void FASTCALL call_2_7_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_10_0
#define init_2_7_10_0
void FASTCALL call_2_7_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_10_1
#define init_2_7_10_1
void FASTCALL call_2_7_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_10_2
#define init_2_7_10_2
void FASTCALL call_2_7_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_10_3
#define init_2_7_10_3
void FASTCALL call_2_7_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_10_4
#define init_2_7_10_4
void FASTCALL call_2_7_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_10_5
#define init_2_7_10_5
void FASTCALL call_2_7_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_10_6
#define init_2_7_10_6
void FASTCALL call_2_7_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_10_7
#define init_2_7_10_7
void FASTCALL call_2_7_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_10_8
#define init_2_7_10_8
void FASTCALL call_2_7_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_10_9
#define init_2_7_10_9
void FASTCALL call_2_7_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_10_10
#define init_2_7_10_10
void FASTCALL call_2_7_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_10_11
#define init_2_7_10_11
void FASTCALL call_2_7_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_10_12
#define init_2_7_10_12
void FASTCALL call_2_7_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_10_13
#define init_2_7_10_13
void FASTCALL call_2_7_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_10_14
#define init_2_7_10_14
void FASTCALL call_2_7_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_10_15
#define init_2_7_10_15
void FASTCALL call_2_7_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_11_0
#define init_2_7_11_0
void FASTCALL call_2_7_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_11_1
#define init_2_7_11_1
void FASTCALL call_2_7_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_11_2
#define init_2_7_11_2
void FASTCALL call_2_7_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_11_3
#define init_2_7_11_3
void FASTCALL call_2_7_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_11_4
#define init_2_7_11_4
void FASTCALL call_2_7_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_11_5
#define init_2_7_11_5
void FASTCALL call_2_7_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_11_6
#define init_2_7_11_6
void FASTCALL call_2_7_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_11_7
#define init_2_7_11_7
void FASTCALL call_2_7_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_11_8
#define init_2_7_11_8
void FASTCALL call_2_7_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_11_9
#define init_2_7_11_9
void FASTCALL call_2_7_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_11_10
#define init_2_7_11_10
void FASTCALL call_2_7_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_11_11
#define init_2_7_11_11
void FASTCALL call_2_7_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_11_12
#define init_2_7_11_12
void FASTCALL call_2_7_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_11_13
#define init_2_7_11_13
void FASTCALL call_2_7_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_11_14
#define init_2_7_11_14
void FASTCALL call_2_7_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_11_15
#define init_2_7_11_15
void FASTCALL call_2_7_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_12_0
#define init_2_7_12_0
void FASTCALL call_2_7_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_12_1
#define init_2_7_12_1
void FASTCALL call_2_7_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_12_2
#define init_2_7_12_2
void FASTCALL call_2_7_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_12_3
#define init_2_7_12_3
void FASTCALL call_2_7_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_12_4
#define init_2_7_12_4
void FASTCALL call_2_7_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_12_5
#define init_2_7_12_5
void FASTCALL call_2_7_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_12_6
#define init_2_7_12_6
void FASTCALL call_2_7_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_12_7
#define init_2_7_12_7
void FASTCALL call_2_7_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_12_8
#define init_2_7_12_8
void FASTCALL call_2_7_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_12_9
#define init_2_7_12_9
void FASTCALL call_2_7_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_12_10
#define init_2_7_12_10
void FASTCALL call_2_7_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_12_11
#define init_2_7_12_11
void FASTCALL call_2_7_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_12_12
#define init_2_7_12_12
void FASTCALL call_2_7_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_12_13
#define init_2_7_12_13
void FASTCALL call_2_7_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_12_14
#define init_2_7_12_14
void FASTCALL call_2_7_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_12_15
#define init_2_7_12_15
void FASTCALL call_2_7_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_13_0
#define init_2_7_13_0
void FASTCALL call_2_7_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_13_1
#define init_2_7_13_1
void FASTCALL call_2_7_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_13_2
#define init_2_7_13_2
void FASTCALL call_2_7_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_13_3
#define init_2_7_13_3
void FASTCALL call_2_7_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_13_4
#define init_2_7_13_4
void FASTCALL call_2_7_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_13_5
#define init_2_7_13_5
void FASTCALL call_2_7_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_13_6
#define init_2_7_13_6
void FASTCALL call_2_7_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_13_7
#define init_2_7_13_7
void FASTCALL call_2_7_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_13_8
#define init_2_7_13_8
void FASTCALL call_2_7_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_13_9
#define init_2_7_13_9
void FASTCALL call_2_7_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_13_10
#define init_2_7_13_10
void FASTCALL call_2_7_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_13_11
#define init_2_7_13_11
void FASTCALL call_2_7_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_13_12
#define init_2_7_13_12
void FASTCALL call_2_7_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_13_13
#define init_2_7_13_13
void FASTCALL call_2_7_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_13_14
#define init_2_7_13_14
void FASTCALL call_2_7_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_13_15
#define init_2_7_13_15
void FASTCALL call_2_7_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_14_0
#define init_2_7_14_0
void FASTCALL call_2_7_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_14_1
#define init_2_7_14_1
void FASTCALL call_2_7_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_14_2
#define init_2_7_14_2
void FASTCALL call_2_7_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_14_3
#define init_2_7_14_3
void FASTCALL call_2_7_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_14_4
#define init_2_7_14_4
void FASTCALL call_2_7_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_14_5
#define init_2_7_14_5
void FASTCALL call_2_7_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_14_6
#define init_2_7_14_6
void FASTCALL call_2_7_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_14_7
#define init_2_7_14_7
void FASTCALL call_2_7_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_14_8
#define init_2_7_14_8
void FASTCALL call_2_7_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_14_9
#define init_2_7_14_9
void FASTCALL call_2_7_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_14_10
#define init_2_7_14_10
void FASTCALL call_2_7_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_14_11
#define init_2_7_14_11
void FASTCALL call_2_7_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_14_12
#define init_2_7_14_12
void FASTCALL call_2_7_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_14_13
#define init_2_7_14_13
void FASTCALL call_2_7_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_14_14
#define init_2_7_14_14
void FASTCALL call_2_7_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_14_15
#define init_2_7_14_15
void FASTCALL call_2_7_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_15_0
#define init_2_7_15_0
void FASTCALL call_2_7_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_15_1
#define init_2_7_15_1
void FASTCALL call_2_7_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_15_2
#define init_2_7_15_2
void FASTCALL call_2_7_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_15_3
#define init_2_7_15_3
void FASTCALL call_2_7_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_15_4
#define init_2_7_15_4
void FASTCALL call_2_7_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_15_5
#define init_2_7_15_5
void FASTCALL call_2_7_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_15_6
#define init_2_7_15_6
void FASTCALL call_2_7_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_15_7
#define init_2_7_15_7
void FASTCALL call_2_7_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_15_8
#define init_2_7_15_8
void FASTCALL call_2_7_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_15_9
#define init_2_7_15_9
void FASTCALL call_2_7_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_15_10
#define init_2_7_15_10
void FASTCALL call_2_7_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_15_11
#define init_2_7_15_11
void FASTCALL call_2_7_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_15_12
#define init_2_7_15_12
void FASTCALL call_2_7_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_15_13
#define init_2_7_15_13
void FASTCALL call_2_7_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_15_14
#define init_2_7_15_14
void FASTCALL call_2_7_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_7_15_15
#define init_2_7_15_15
void FASTCALL call_2_7_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_0_0
#define init_2_8_0_0
void FASTCALL call_2_8_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_0_1
#define init_2_8_0_1
void FASTCALL call_2_8_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_0_2
#define init_2_8_0_2
void FASTCALL call_2_8_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_0_3
#define init_2_8_0_3
void FASTCALL call_2_8_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_0_4
#define init_2_8_0_4
void FASTCALL call_2_8_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_0_5
#define init_2_8_0_5
void FASTCALL call_2_8_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_0_6
#define init_2_8_0_6
void FASTCALL call_2_8_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_0_7
#define init_2_8_0_7
void FASTCALL call_2_8_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_0_8
#define init_2_8_0_8
void FASTCALL call_2_8_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_0_9
#define init_2_8_0_9
void FASTCALL call_2_8_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_0_10
#define init_2_8_0_10
void FASTCALL call_2_8_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_0_11
#define init_2_8_0_11
void FASTCALL call_2_8_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_0_12
#define init_2_8_0_12
void FASTCALL call_2_8_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_0_13
#define init_2_8_0_13
void FASTCALL call_2_8_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_0_14
#define init_2_8_0_14
void FASTCALL call_2_8_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_0_15
#define init_2_8_0_15
void FASTCALL call_2_8_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_1_0
#define init_2_8_1_0
void FASTCALL call_2_8_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_1_1
#define init_2_8_1_1
void FASTCALL call_2_8_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_1_2
#define init_2_8_1_2
void FASTCALL call_2_8_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_1_3
#define init_2_8_1_3
void FASTCALL call_2_8_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_1_4
#define init_2_8_1_4
void FASTCALL call_2_8_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_1_5
#define init_2_8_1_5
void FASTCALL call_2_8_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_1_6
#define init_2_8_1_6
void FASTCALL call_2_8_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_1_7
#define init_2_8_1_7
void FASTCALL call_2_8_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_1_8
#define init_2_8_1_8
void FASTCALL call_2_8_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_1_9
#define init_2_8_1_9
void FASTCALL call_2_8_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_1_10
#define init_2_8_1_10
void FASTCALL call_2_8_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_1_11
#define init_2_8_1_11
void FASTCALL call_2_8_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_1_12
#define init_2_8_1_12
void FASTCALL call_2_8_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_1_13
#define init_2_8_1_13
void FASTCALL call_2_8_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_1_14
#define init_2_8_1_14
void FASTCALL call_2_8_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_1_15
#define init_2_8_1_15
void FASTCALL call_2_8_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_2_0
#define init_2_8_2_0
void FASTCALL call_2_8_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_2_1
#define init_2_8_2_1
void FASTCALL call_2_8_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_2_2
#define init_2_8_2_2
void FASTCALL call_2_8_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_2_3
#define init_2_8_2_3
void FASTCALL call_2_8_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_2_4
#define init_2_8_2_4
void FASTCALL call_2_8_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_2_5
#define init_2_8_2_5
void FASTCALL call_2_8_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_2_6
#define init_2_8_2_6
void FASTCALL call_2_8_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_2_7
#define init_2_8_2_7
void FASTCALL call_2_8_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_2_8
#define init_2_8_2_8
void FASTCALL call_2_8_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_2_9
#define init_2_8_2_9
void FASTCALL call_2_8_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_2_10
#define init_2_8_2_10
void FASTCALL call_2_8_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_2_11
#define init_2_8_2_11
void FASTCALL call_2_8_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_2_12
#define init_2_8_2_12
void FASTCALL call_2_8_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_2_13
#define init_2_8_2_13
void FASTCALL call_2_8_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_2_14
#define init_2_8_2_14
void FASTCALL call_2_8_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_2_15
#define init_2_8_2_15
void FASTCALL call_2_8_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_3_0
#define init_2_8_3_0
void FASTCALL call_2_8_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_3_1
#define init_2_8_3_1
void FASTCALL call_2_8_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_3_2
#define init_2_8_3_2
void FASTCALL call_2_8_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_3_3
#define init_2_8_3_3
void FASTCALL call_2_8_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_3_4
#define init_2_8_3_4
void FASTCALL call_2_8_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_3_5
#define init_2_8_3_5
void FASTCALL call_2_8_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_3_6
#define init_2_8_3_6
void FASTCALL call_2_8_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_3_7
#define init_2_8_3_7
void FASTCALL call_2_8_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_3_8
#define init_2_8_3_8
void FASTCALL call_2_8_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_3_9
#define init_2_8_3_9
void FASTCALL call_2_8_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_3_10
#define init_2_8_3_10
void FASTCALL call_2_8_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_3_11
#define init_2_8_3_11
void FASTCALL call_2_8_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_3_12
#define init_2_8_3_12
void FASTCALL call_2_8_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_3_13
#define init_2_8_3_13
void FASTCALL call_2_8_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_3_14
#define init_2_8_3_14
void FASTCALL call_2_8_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_3_15
#define init_2_8_3_15
void FASTCALL call_2_8_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_4_0
#define init_2_8_4_0
void FASTCALL call_2_8_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_4_1
#define init_2_8_4_1
void FASTCALL call_2_8_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_4_2
#define init_2_8_4_2
void FASTCALL call_2_8_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_4_3
#define init_2_8_4_3
void FASTCALL call_2_8_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_4_4
#define init_2_8_4_4
void FASTCALL call_2_8_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_4_5
#define init_2_8_4_5
void FASTCALL call_2_8_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_4_6
#define init_2_8_4_6
void FASTCALL call_2_8_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_4_7
#define init_2_8_4_7
void FASTCALL call_2_8_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_4_8
#define init_2_8_4_8
void FASTCALL call_2_8_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_4_9
#define init_2_8_4_9
void FASTCALL call_2_8_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_4_10
#define init_2_8_4_10
void FASTCALL call_2_8_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_4_11
#define init_2_8_4_11
void FASTCALL call_2_8_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_4_12
#define init_2_8_4_12
void FASTCALL call_2_8_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_4_13
#define init_2_8_4_13
void FASTCALL call_2_8_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_4_14
#define init_2_8_4_14
void FASTCALL call_2_8_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_4_15
#define init_2_8_4_15
void FASTCALL call_2_8_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_5_0
#define init_2_8_5_0
void FASTCALL call_2_8_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_5_1
#define init_2_8_5_1
void FASTCALL call_2_8_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_5_2
#define init_2_8_5_2
void FASTCALL call_2_8_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_5_3
#define init_2_8_5_3
void FASTCALL call_2_8_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_5_4
#define init_2_8_5_4
void FASTCALL call_2_8_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_5_5
#define init_2_8_5_5
void FASTCALL call_2_8_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_5_6
#define init_2_8_5_6
void FASTCALL call_2_8_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_5_7
#define init_2_8_5_7
void FASTCALL call_2_8_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_5_8
#define init_2_8_5_8
void FASTCALL call_2_8_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_5_9
#define init_2_8_5_9
void FASTCALL call_2_8_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_5_10
#define init_2_8_5_10
void FASTCALL call_2_8_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_5_11
#define init_2_8_5_11
void FASTCALL call_2_8_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_5_12
#define init_2_8_5_12
void FASTCALL call_2_8_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_5_13
#define init_2_8_5_13
void FASTCALL call_2_8_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_5_14
#define init_2_8_5_14
void FASTCALL call_2_8_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_5_15
#define init_2_8_5_15
void FASTCALL call_2_8_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_6_0
#define init_2_8_6_0
void FASTCALL call_2_8_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_6_1
#define init_2_8_6_1
void FASTCALL call_2_8_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_6_2
#define init_2_8_6_2
void FASTCALL call_2_8_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_6_3
#define init_2_8_6_3
void FASTCALL call_2_8_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_6_4
#define init_2_8_6_4
void FASTCALL call_2_8_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_6_5
#define init_2_8_6_5
void FASTCALL call_2_8_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_6_6
#define init_2_8_6_6
void FASTCALL call_2_8_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_6_7
#define init_2_8_6_7
void FASTCALL call_2_8_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_6_8
#define init_2_8_6_8
void FASTCALL call_2_8_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_6_9
#define init_2_8_6_9
void FASTCALL call_2_8_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_6_10
#define init_2_8_6_10
void FASTCALL call_2_8_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_6_11
#define init_2_8_6_11
void FASTCALL call_2_8_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_6_12
#define init_2_8_6_12
void FASTCALL call_2_8_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_6_13
#define init_2_8_6_13
void FASTCALL call_2_8_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_6_14
#define init_2_8_6_14
void FASTCALL call_2_8_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_6_15
#define init_2_8_6_15
void FASTCALL call_2_8_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_7_0
#define init_2_8_7_0
void FASTCALL call_2_8_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_7_1
#define init_2_8_7_1
void FASTCALL call_2_8_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_7_2
#define init_2_8_7_2
void FASTCALL call_2_8_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_7_3
#define init_2_8_7_3
void FASTCALL call_2_8_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_7_4
#define init_2_8_7_4
void FASTCALL call_2_8_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_7_5
#define init_2_8_7_5
void FASTCALL call_2_8_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_7_6
#define init_2_8_7_6
void FASTCALL call_2_8_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_7_7
#define init_2_8_7_7
void FASTCALL call_2_8_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_7_8
#define init_2_8_7_8
void FASTCALL call_2_8_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_7_9
#define init_2_8_7_9
void FASTCALL call_2_8_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_7_10
#define init_2_8_7_10
void FASTCALL call_2_8_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_7_11
#define init_2_8_7_11
void FASTCALL call_2_8_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_7_12
#define init_2_8_7_12
void FASTCALL call_2_8_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_7_13
#define init_2_8_7_13
void FASTCALL call_2_8_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_7_14
#define init_2_8_7_14
void FASTCALL call_2_8_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_7_15
#define init_2_8_7_15
void FASTCALL call_2_8_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_8_0
#define init_2_8_8_0
void FASTCALL call_2_8_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_8_1
#define init_2_8_8_1
void FASTCALL call_2_8_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_8_2
#define init_2_8_8_2
void FASTCALL call_2_8_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_8_3
#define init_2_8_8_3
void FASTCALL call_2_8_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_8_4
#define init_2_8_8_4
void FASTCALL call_2_8_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_8_5
#define init_2_8_8_5
void FASTCALL call_2_8_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_8_6
#define init_2_8_8_6
void FASTCALL call_2_8_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_8_7
#define init_2_8_8_7
void FASTCALL call_2_8_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_8_8
#define init_2_8_8_8
void FASTCALL call_2_8_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_8_9
#define init_2_8_8_9
void FASTCALL call_2_8_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_8_10
#define init_2_8_8_10
void FASTCALL call_2_8_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_8_11
#define init_2_8_8_11
void FASTCALL call_2_8_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_8_12
#define init_2_8_8_12
void FASTCALL call_2_8_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_8_13
#define init_2_8_8_13
void FASTCALL call_2_8_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_8_14
#define init_2_8_8_14
void FASTCALL call_2_8_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_8_15
#define init_2_8_8_15
void FASTCALL call_2_8_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_9_0
#define init_2_8_9_0
void FASTCALL call_2_8_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_9_1
#define init_2_8_9_1
void FASTCALL call_2_8_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_9_2
#define init_2_8_9_2
void FASTCALL call_2_8_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_9_3
#define init_2_8_9_3
void FASTCALL call_2_8_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_9_4
#define init_2_8_9_4
void FASTCALL call_2_8_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_9_5
#define init_2_8_9_5
void FASTCALL call_2_8_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_9_6
#define init_2_8_9_6
void FASTCALL call_2_8_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_9_7
#define init_2_8_9_7
void FASTCALL call_2_8_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_9_8
#define init_2_8_9_8
void FASTCALL call_2_8_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_9_9
#define init_2_8_9_9
void FASTCALL call_2_8_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_9_10
#define init_2_8_9_10
void FASTCALL call_2_8_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_9_11
#define init_2_8_9_11
void FASTCALL call_2_8_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_9_12
#define init_2_8_9_12
void FASTCALL call_2_8_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_9_13
#define init_2_8_9_13
void FASTCALL call_2_8_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_9_14
#define init_2_8_9_14
void FASTCALL call_2_8_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_9_15
#define init_2_8_9_15
void FASTCALL call_2_8_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_10_0
#define init_2_8_10_0
void FASTCALL call_2_8_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_10_1
#define init_2_8_10_1
void FASTCALL call_2_8_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_10_2
#define init_2_8_10_2
void FASTCALL call_2_8_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_10_3
#define init_2_8_10_3
void FASTCALL call_2_8_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_10_4
#define init_2_8_10_4
void FASTCALL call_2_8_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_10_5
#define init_2_8_10_5
void FASTCALL call_2_8_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_10_6
#define init_2_8_10_6
void FASTCALL call_2_8_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_10_7
#define init_2_8_10_7
void FASTCALL call_2_8_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_10_8
#define init_2_8_10_8
void FASTCALL call_2_8_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_10_9
#define init_2_8_10_9
void FASTCALL call_2_8_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_10_10
#define init_2_8_10_10
void FASTCALL call_2_8_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_10_11
#define init_2_8_10_11
void FASTCALL call_2_8_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_10_12
#define init_2_8_10_12
void FASTCALL call_2_8_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_10_13
#define init_2_8_10_13
void FASTCALL call_2_8_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_10_14
#define init_2_8_10_14
void FASTCALL call_2_8_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_10_15
#define init_2_8_10_15
void FASTCALL call_2_8_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_11_0
#define init_2_8_11_0
void FASTCALL call_2_8_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_11_1
#define init_2_8_11_1
void FASTCALL call_2_8_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_11_2
#define init_2_8_11_2
void FASTCALL call_2_8_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_11_3
#define init_2_8_11_3
void FASTCALL call_2_8_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_11_4
#define init_2_8_11_4
void FASTCALL call_2_8_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_11_5
#define init_2_8_11_5
void FASTCALL call_2_8_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_11_6
#define init_2_8_11_6
void FASTCALL call_2_8_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_11_7
#define init_2_8_11_7
void FASTCALL call_2_8_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_11_8
#define init_2_8_11_8
void FASTCALL call_2_8_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_11_9
#define init_2_8_11_9
void FASTCALL call_2_8_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_11_10
#define init_2_8_11_10
void FASTCALL call_2_8_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_11_11
#define init_2_8_11_11
void FASTCALL call_2_8_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_11_12
#define init_2_8_11_12
void FASTCALL call_2_8_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_11_13
#define init_2_8_11_13
void FASTCALL call_2_8_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_11_14
#define init_2_8_11_14
void FASTCALL call_2_8_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_11_15
#define init_2_8_11_15
void FASTCALL call_2_8_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_12_0
#define init_2_8_12_0
void FASTCALL call_2_8_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_12_1
#define init_2_8_12_1
void FASTCALL call_2_8_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_12_2
#define init_2_8_12_2
void FASTCALL call_2_8_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_12_3
#define init_2_8_12_3
void FASTCALL call_2_8_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_12_4
#define init_2_8_12_4
void FASTCALL call_2_8_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_12_5
#define init_2_8_12_5
void FASTCALL call_2_8_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_12_6
#define init_2_8_12_6
void FASTCALL call_2_8_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_12_7
#define init_2_8_12_7
void FASTCALL call_2_8_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_12_8
#define init_2_8_12_8
void FASTCALL call_2_8_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_12_9
#define init_2_8_12_9
void FASTCALL call_2_8_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_12_10
#define init_2_8_12_10
void FASTCALL call_2_8_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_12_11
#define init_2_8_12_11
void FASTCALL call_2_8_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_12_12
#define init_2_8_12_12
void FASTCALL call_2_8_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_12_13
#define init_2_8_12_13
void FASTCALL call_2_8_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_12_14
#define init_2_8_12_14
void FASTCALL call_2_8_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_12_15
#define init_2_8_12_15
void FASTCALL call_2_8_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_13_0
#define init_2_8_13_0
void FASTCALL call_2_8_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_13_1
#define init_2_8_13_1
void FASTCALL call_2_8_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_13_2
#define init_2_8_13_2
void FASTCALL call_2_8_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_13_3
#define init_2_8_13_3
void FASTCALL call_2_8_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_13_4
#define init_2_8_13_4
void FASTCALL call_2_8_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_13_5
#define init_2_8_13_5
void FASTCALL call_2_8_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_13_6
#define init_2_8_13_6
void FASTCALL call_2_8_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_13_7
#define init_2_8_13_7
void FASTCALL call_2_8_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_13_8
#define init_2_8_13_8
void FASTCALL call_2_8_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_13_9
#define init_2_8_13_9
void FASTCALL call_2_8_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_13_10
#define init_2_8_13_10
void FASTCALL call_2_8_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_13_11
#define init_2_8_13_11
void FASTCALL call_2_8_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_13_12
#define init_2_8_13_12
void FASTCALL call_2_8_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_13_13
#define init_2_8_13_13
void FASTCALL call_2_8_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_13_14
#define init_2_8_13_14
void FASTCALL call_2_8_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_13_15
#define init_2_8_13_15
void FASTCALL call_2_8_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_14_0
#define init_2_8_14_0
void FASTCALL call_2_8_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_14_1
#define init_2_8_14_1
void FASTCALL call_2_8_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_14_2
#define init_2_8_14_2
void FASTCALL call_2_8_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_14_3
#define init_2_8_14_3
void FASTCALL call_2_8_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_14_4
#define init_2_8_14_4
void FASTCALL call_2_8_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_14_5
#define init_2_8_14_5
void FASTCALL call_2_8_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_14_6
#define init_2_8_14_6
void FASTCALL call_2_8_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_14_7
#define init_2_8_14_7
void FASTCALL call_2_8_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_14_8
#define init_2_8_14_8
void FASTCALL call_2_8_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_14_9
#define init_2_8_14_9
void FASTCALL call_2_8_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_14_10
#define init_2_8_14_10
void FASTCALL call_2_8_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_14_11
#define init_2_8_14_11
void FASTCALL call_2_8_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_14_12
#define init_2_8_14_12
void FASTCALL call_2_8_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_14_13
#define init_2_8_14_13
void FASTCALL call_2_8_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_14_14
#define init_2_8_14_14
void FASTCALL call_2_8_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_14_15
#define init_2_8_14_15
void FASTCALL call_2_8_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_15_0
#define init_2_8_15_0
void FASTCALL call_2_8_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_15_1
#define init_2_8_15_1
void FASTCALL call_2_8_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_15_2
#define init_2_8_15_2
void FASTCALL call_2_8_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_15_3
#define init_2_8_15_3
void FASTCALL call_2_8_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_15_4
#define init_2_8_15_4
void FASTCALL call_2_8_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_15_5
#define init_2_8_15_5
void FASTCALL call_2_8_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_15_6
#define init_2_8_15_6
void FASTCALL call_2_8_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_15_7
#define init_2_8_15_7
void FASTCALL call_2_8_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_15_8
#define init_2_8_15_8
void FASTCALL call_2_8_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_15_9
#define init_2_8_15_9
void FASTCALL call_2_8_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_15_10
#define init_2_8_15_10
void FASTCALL call_2_8_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_15_11
#define init_2_8_15_11
void FASTCALL call_2_8_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_15_12
#define init_2_8_15_12
void FASTCALL call_2_8_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_15_13
#define init_2_8_15_13
void FASTCALL call_2_8_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_15_14
#define init_2_8_15_14
void FASTCALL call_2_8_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_8_15_15
#define init_2_8_15_15
void FASTCALL call_2_8_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_0_0
#define init_2_9_0_0
void FASTCALL call_2_9_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_0_1
#define init_2_9_0_1
void FASTCALL call_2_9_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_0_2
#define init_2_9_0_2
void FASTCALL call_2_9_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_0_3
#define init_2_9_0_3
void FASTCALL call_2_9_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_0_4
#define init_2_9_0_4
void FASTCALL call_2_9_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_0_5
#define init_2_9_0_5
void FASTCALL call_2_9_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_0_6
#define init_2_9_0_6
void FASTCALL call_2_9_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_0_7
#define init_2_9_0_7
void FASTCALL call_2_9_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_0_8
#define init_2_9_0_8
void FASTCALL call_2_9_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_0_9
#define init_2_9_0_9
void FASTCALL call_2_9_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_0_10
#define init_2_9_0_10
void FASTCALL call_2_9_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_0_11
#define init_2_9_0_11
void FASTCALL call_2_9_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_0_12
#define init_2_9_0_12
void FASTCALL call_2_9_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_0_13
#define init_2_9_0_13
void FASTCALL call_2_9_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_0_14
#define init_2_9_0_14
void FASTCALL call_2_9_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_0_15
#define init_2_9_0_15
void FASTCALL call_2_9_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_1_0
#define init_2_9_1_0
void FASTCALL call_2_9_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_1_1
#define init_2_9_1_1
void FASTCALL call_2_9_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_1_2
#define init_2_9_1_2
void FASTCALL call_2_9_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_1_3
#define init_2_9_1_3
void FASTCALL call_2_9_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_1_4
#define init_2_9_1_4
void FASTCALL call_2_9_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_1_5
#define init_2_9_1_5
void FASTCALL call_2_9_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_1_6
#define init_2_9_1_6
void FASTCALL call_2_9_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_1_7
#define init_2_9_1_7
void FASTCALL call_2_9_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_1_8
#define init_2_9_1_8
void FASTCALL call_2_9_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_1_9
#define init_2_9_1_9
void FASTCALL call_2_9_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_1_10
#define init_2_9_1_10
void FASTCALL call_2_9_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_1_11
#define init_2_9_1_11
void FASTCALL call_2_9_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_1_12
#define init_2_9_1_12
void FASTCALL call_2_9_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_1_13
#define init_2_9_1_13
void FASTCALL call_2_9_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_1_14
#define init_2_9_1_14
void FASTCALL call_2_9_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_1_15
#define init_2_9_1_15
void FASTCALL call_2_9_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_2_0
#define init_2_9_2_0
void FASTCALL call_2_9_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_2_1
#define init_2_9_2_1
void FASTCALL call_2_9_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_2_2
#define init_2_9_2_2
void FASTCALL call_2_9_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_2_3
#define init_2_9_2_3
void FASTCALL call_2_9_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_2_4
#define init_2_9_2_4
void FASTCALL call_2_9_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_2_5
#define init_2_9_2_5
void FASTCALL call_2_9_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_2_6
#define init_2_9_2_6
void FASTCALL call_2_9_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_2_7
#define init_2_9_2_7
void FASTCALL call_2_9_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_2_8
#define init_2_9_2_8
void FASTCALL call_2_9_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_2_9
#define init_2_9_2_9
void FASTCALL call_2_9_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_2_10
#define init_2_9_2_10
void FASTCALL call_2_9_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_2_11
#define init_2_9_2_11
void FASTCALL call_2_9_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_2_12
#define init_2_9_2_12
void FASTCALL call_2_9_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_2_13
#define init_2_9_2_13
void FASTCALL call_2_9_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_2_14
#define init_2_9_2_14
void FASTCALL call_2_9_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_2_15
#define init_2_9_2_15
void FASTCALL call_2_9_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_3_0
#define init_2_9_3_0
void FASTCALL call_2_9_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_3_1
#define init_2_9_3_1
void FASTCALL call_2_9_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_3_2
#define init_2_9_3_2
void FASTCALL call_2_9_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_3_3
#define init_2_9_3_3
void FASTCALL call_2_9_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_3_4
#define init_2_9_3_4
void FASTCALL call_2_9_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_3_5
#define init_2_9_3_5
void FASTCALL call_2_9_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_3_6
#define init_2_9_3_6
void FASTCALL call_2_9_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_3_7
#define init_2_9_3_7
void FASTCALL call_2_9_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_3_8
#define init_2_9_3_8
void FASTCALL call_2_9_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_3_9
#define init_2_9_3_9
void FASTCALL call_2_9_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_3_10
#define init_2_9_3_10
void FASTCALL call_2_9_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_3_11
#define init_2_9_3_11
void FASTCALL call_2_9_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_3_12
#define init_2_9_3_12
void FASTCALL call_2_9_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_3_13
#define init_2_9_3_13
void FASTCALL call_2_9_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_3_14
#define init_2_9_3_14
void FASTCALL call_2_9_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_3_15
#define init_2_9_3_15
void FASTCALL call_2_9_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_4_0
#define init_2_9_4_0
void FASTCALL call_2_9_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_4_1
#define init_2_9_4_1
void FASTCALL call_2_9_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_4_2
#define init_2_9_4_2
void FASTCALL call_2_9_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_4_3
#define init_2_9_4_3
void FASTCALL call_2_9_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_4_4
#define init_2_9_4_4
void FASTCALL call_2_9_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_4_5
#define init_2_9_4_5
void FASTCALL call_2_9_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_4_6
#define init_2_9_4_6
void FASTCALL call_2_9_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_4_7
#define init_2_9_4_7
void FASTCALL call_2_9_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_4_8
#define init_2_9_4_8
void FASTCALL call_2_9_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_4_9
#define init_2_9_4_9
void FASTCALL call_2_9_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_4_10
#define init_2_9_4_10
void FASTCALL call_2_9_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_4_11
#define init_2_9_4_11
void FASTCALL call_2_9_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_4_12
#define init_2_9_4_12
void FASTCALL call_2_9_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_4_13
#define init_2_9_4_13
void FASTCALL call_2_9_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_4_14
#define init_2_9_4_14
void FASTCALL call_2_9_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_4_15
#define init_2_9_4_15
void FASTCALL call_2_9_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_5_0
#define init_2_9_5_0
void FASTCALL call_2_9_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_5_1
#define init_2_9_5_1
void FASTCALL call_2_9_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_5_2
#define init_2_9_5_2
void FASTCALL call_2_9_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_5_3
#define init_2_9_5_3
void FASTCALL call_2_9_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_5_4
#define init_2_9_5_4
void FASTCALL call_2_9_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_5_5
#define init_2_9_5_5
void FASTCALL call_2_9_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_5_6
#define init_2_9_5_6
void FASTCALL call_2_9_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_5_7
#define init_2_9_5_7
void FASTCALL call_2_9_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_5_8
#define init_2_9_5_8
void FASTCALL call_2_9_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_5_9
#define init_2_9_5_9
void FASTCALL call_2_9_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_5_10
#define init_2_9_5_10
void FASTCALL call_2_9_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_5_11
#define init_2_9_5_11
void FASTCALL call_2_9_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_5_12
#define init_2_9_5_12
void FASTCALL call_2_9_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_5_13
#define init_2_9_5_13
void FASTCALL call_2_9_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_5_14
#define init_2_9_5_14
void FASTCALL call_2_9_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_5_15
#define init_2_9_5_15
void FASTCALL call_2_9_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_6_0
#define init_2_9_6_0
void FASTCALL call_2_9_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_6_1
#define init_2_9_6_1
void FASTCALL call_2_9_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_6_2
#define init_2_9_6_2
void FASTCALL call_2_9_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_6_3
#define init_2_9_6_3
void FASTCALL call_2_9_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_6_4
#define init_2_9_6_4
void FASTCALL call_2_9_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_6_5
#define init_2_9_6_5
void FASTCALL call_2_9_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_6_6
#define init_2_9_6_6
void FASTCALL call_2_9_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_6_7
#define init_2_9_6_7
void FASTCALL call_2_9_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_6_8
#define init_2_9_6_8
void FASTCALL call_2_9_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_6_9
#define init_2_9_6_9
void FASTCALL call_2_9_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_6_10
#define init_2_9_6_10
void FASTCALL call_2_9_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_6_11
#define init_2_9_6_11
void FASTCALL call_2_9_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_6_12
#define init_2_9_6_12
void FASTCALL call_2_9_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_6_13
#define init_2_9_6_13
void FASTCALL call_2_9_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_6_14
#define init_2_9_6_14
void FASTCALL call_2_9_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_6_15
#define init_2_9_6_15
void FASTCALL call_2_9_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_7_0
#define init_2_9_7_0
void FASTCALL call_2_9_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_7_1
#define init_2_9_7_1
void FASTCALL call_2_9_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_7_2
#define init_2_9_7_2
void FASTCALL call_2_9_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_7_3
#define init_2_9_7_3
void FASTCALL call_2_9_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_7_4
#define init_2_9_7_4
void FASTCALL call_2_9_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_7_5
#define init_2_9_7_5
void FASTCALL call_2_9_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_7_6
#define init_2_9_7_6
void FASTCALL call_2_9_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_7_7
#define init_2_9_7_7
void FASTCALL call_2_9_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_7_8
#define init_2_9_7_8
void FASTCALL call_2_9_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_7_9
#define init_2_9_7_9
void FASTCALL call_2_9_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_7_10
#define init_2_9_7_10
void FASTCALL call_2_9_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_7_11
#define init_2_9_7_11
void FASTCALL call_2_9_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_7_12
#define init_2_9_7_12
void FASTCALL call_2_9_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_7_13
#define init_2_9_7_13
void FASTCALL call_2_9_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_7_14
#define init_2_9_7_14
void FASTCALL call_2_9_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_7_15
#define init_2_9_7_15
void FASTCALL call_2_9_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_8_0
#define init_2_9_8_0
void FASTCALL call_2_9_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_8_1
#define init_2_9_8_1
void FASTCALL call_2_9_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_8_2
#define init_2_9_8_2
void FASTCALL call_2_9_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_8_3
#define init_2_9_8_3
void FASTCALL call_2_9_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_8_4
#define init_2_9_8_4
void FASTCALL call_2_9_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_8_5
#define init_2_9_8_5
void FASTCALL call_2_9_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_8_6
#define init_2_9_8_6
void FASTCALL call_2_9_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_8_7
#define init_2_9_8_7
void FASTCALL call_2_9_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_8_8
#define init_2_9_8_8
void FASTCALL call_2_9_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_8_9
#define init_2_9_8_9
void FASTCALL call_2_9_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_8_10
#define init_2_9_8_10
void FASTCALL call_2_9_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_8_11
#define init_2_9_8_11
void FASTCALL call_2_9_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_8_12
#define init_2_9_8_12
void FASTCALL call_2_9_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_8_13
#define init_2_9_8_13
void FASTCALL call_2_9_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_8_14
#define init_2_9_8_14
void FASTCALL call_2_9_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_8_15
#define init_2_9_8_15
void FASTCALL call_2_9_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_9_0
#define init_2_9_9_0
void FASTCALL call_2_9_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_9_1
#define init_2_9_9_1
void FASTCALL call_2_9_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_9_2
#define init_2_9_9_2
void FASTCALL call_2_9_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_9_3
#define init_2_9_9_3
void FASTCALL call_2_9_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_9_4
#define init_2_9_9_4
void FASTCALL call_2_9_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_9_5
#define init_2_9_9_5
void FASTCALL call_2_9_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_9_6
#define init_2_9_9_6
void FASTCALL call_2_9_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_9_7
#define init_2_9_9_7
void FASTCALL call_2_9_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_9_8
#define init_2_9_9_8
void FASTCALL call_2_9_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_9_9
#define init_2_9_9_9
void FASTCALL call_2_9_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_9_10
#define init_2_9_9_10
void FASTCALL call_2_9_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_9_11
#define init_2_9_9_11
void FASTCALL call_2_9_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_9_12
#define init_2_9_9_12
void FASTCALL call_2_9_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_9_13
#define init_2_9_9_13
void FASTCALL call_2_9_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_9_14
#define init_2_9_9_14
void FASTCALL call_2_9_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_9_15
#define init_2_9_9_15
void FASTCALL call_2_9_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_10_0
#define init_2_9_10_0
void FASTCALL call_2_9_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_10_1
#define init_2_9_10_1
void FASTCALL call_2_9_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_10_2
#define init_2_9_10_2
void FASTCALL call_2_9_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_10_3
#define init_2_9_10_3
void FASTCALL call_2_9_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_10_4
#define init_2_9_10_4
void FASTCALL call_2_9_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_10_5
#define init_2_9_10_5
void FASTCALL call_2_9_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_10_6
#define init_2_9_10_6
void FASTCALL call_2_9_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_10_7
#define init_2_9_10_7
void FASTCALL call_2_9_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_10_8
#define init_2_9_10_8
void FASTCALL call_2_9_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_10_9
#define init_2_9_10_9
void FASTCALL call_2_9_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_10_10
#define init_2_9_10_10
void FASTCALL call_2_9_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_10_11
#define init_2_9_10_11
void FASTCALL call_2_9_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_10_12
#define init_2_9_10_12
void FASTCALL call_2_9_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_10_13
#define init_2_9_10_13
void FASTCALL call_2_9_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_10_14
#define init_2_9_10_14
void FASTCALL call_2_9_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_10_15
#define init_2_9_10_15
void FASTCALL call_2_9_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_11_0
#define init_2_9_11_0
void FASTCALL call_2_9_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_11_1
#define init_2_9_11_1
void FASTCALL call_2_9_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_11_2
#define init_2_9_11_2
void FASTCALL call_2_9_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_11_3
#define init_2_9_11_3
void FASTCALL call_2_9_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_11_4
#define init_2_9_11_4
void FASTCALL call_2_9_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_11_5
#define init_2_9_11_5
void FASTCALL call_2_9_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_11_6
#define init_2_9_11_6
void FASTCALL call_2_9_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_11_7
#define init_2_9_11_7
void FASTCALL call_2_9_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_11_8
#define init_2_9_11_8
void FASTCALL call_2_9_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_11_9
#define init_2_9_11_9
void FASTCALL call_2_9_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_11_10
#define init_2_9_11_10
void FASTCALL call_2_9_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_11_11
#define init_2_9_11_11
void FASTCALL call_2_9_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_11_12
#define init_2_9_11_12
void FASTCALL call_2_9_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_11_13
#define init_2_9_11_13
void FASTCALL call_2_9_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_11_14
#define init_2_9_11_14
void FASTCALL call_2_9_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_11_15
#define init_2_9_11_15
void FASTCALL call_2_9_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_12_0
#define init_2_9_12_0
void FASTCALL call_2_9_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_12_1
#define init_2_9_12_1
void FASTCALL call_2_9_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_12_2
#define init_2_9_12_2
void FASTCALL call_2_9_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_12_3
#define init_2_9_12_3
void FASTCALL call_2_9_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_12_4
#define init_2_9_12_4
void FASTCALL call_2_9_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_12_5
#define init_2_9_12_5
void FASTCALL call_2_9_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_12_6
#define init_2_9_12_6
void FASTCALL call_2_9_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_12_7
#define init_2_9_12_7
void FASTCALL call_2_9_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_12_8
#define init_2_9_12_8
void FASTCALL call_2_9_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_12_9
#define init_2_9_12_9
void FASTCALL call_2_9_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_12_10
#define init_2_9_12_10
void FASTCALL call_2_9_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_12_11
#define init_2_9_12_11
void FASTCALL call_2_9_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_12_12
#define init_2_9_12_12
void FASTCALL call_2_9_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_12_13
#define init_2_9_12_13
void FASTCALL call_2_9_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_12_14
#define init_2_9_12_14
void FASTCALL call_2_9_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_12_15
#define init_2_9_12_15
void FASTCALL call_2_9_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_13_0
#define init_2_9_13_0
void FASTCALL call_2_9_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_13_1
#define init_2_9_13_1
void FASTCALL call_2_9_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_13_2
#define init_2_9_13_2
void FASTCALL call_2_9_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_13_3
#define init_2_9_13_3
void FASTCALL call_2_9_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_13_4
#define init_2_9_13_4
void FASTCALL call_2_9_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_13_5
#define init_2_9_13_5
void FASTCALL call_2_9_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_13_6
#define init_2_9_13_6
void FASTCALL call_2_9_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_13_7
#define init_2_9_13_7
void FASTCALL call_2_9_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_13_8
#define init_2_9_13_8
void FASTCALL call_2_9_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_13_9
#define init_2_9_13_9
void FASTCALL call_2_9_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_13_10
#define init_2_9_13_10
void FASTCALL call_2_9_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_13_11
#define init_2_9_13_11
void FASTCALL call_2_9_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_13_12
#define init_2_9_13_12
void FASTCALL call_2_9_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_13_13
#define init_2_9_13_13
void FASTCALL call_2_9_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_13_14
#define init_2_9_13_14
void FASTCALL call_2_9_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_13_15
#define init_2_9_13_15
void FASTCALL call_2_9_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_14_0
#define init_2_9_14_0
void FASTCALL call_2_9_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_14_1
#define init_2_9_14_1
void FASTCALL call_2_9_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_14_2
#define init_2_9_14_2
void FASTCALL call_2_9_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_14_3
#define init_2_9_14_3
void FASTCALL call_2_9_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_14_4
#define init_2_9_14_4
void FASTCALL call_2_9_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_14_5
#define init_2_9_14_5
void FASTCALL call_2_9_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_14_6
#define init_2_9_14_6
void FASTCALL call_2_9_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_14_7
#define init_2_9_14_7
void FASTCALL call_2_9_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_14_8
#define init_2_9_14_8
void FASTCALL call_2_9_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_14_9
#define init_2_9_14_9
void FASTCALL call_2_9_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_14_10
#define init_2_9_14_10
void FASTCALL call_2_9_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_14_11
#define init_2_9_14_11
void FASTCALL call_2_9_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_14_12
#define init_2_9_14_12
void FASTCALL call_2_9_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_14_13
#define init_2_9_14_13
void FASTCALL call_2_9_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_14_14
#define init_2_9_14_14
void FASTCALL call_2_9_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_14_15
#define init_2_9_14_15
void FASTCALL call_2_9_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_15_0
#define init_2_9_15_0
void FASTCALL call_2_9_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_15_1
#define init_2_9_15_1
void FASTCALL call_2_9_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_15_2
#define init_2_9_15_2
void FASTCALL call_2_9_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_15_3
#define init_2_9_15_3
void FASTCALL call_2_9_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_15_4
#define init_2_9_15_4
void FASTCALL call_2_9_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_15_5
#define init_2_9_15_5
void FASTCALL call_2_9_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_15_6
#define init_2_9_15_6
void FASTCALL call_2_9_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_15_7
#define init_2_9_15_7
void FASTCALL call_2_9_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_15_8
#define init_2_9_15_8
void FASTCALL call_2_9_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_15_9
#define init_2_9_15_9
void FASTCALL call_2_9_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_15_10
#define init_2_9_15_10
void FASTCALL call_2_9_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_15_11
#define init_2_9_15_11
void FASTCALL call_2_9_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_15_12
#define init_2_9_15_12
void FASTCALL call_2_9_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_15_13
#define init_2_9_15_13
void FASTCALL call_2_9_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_15_14
#define init_2_9_15_14
void FASTCALL call_2_9_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_9_15_15
#define init_2_9_15_15
void FASTCALL call_2_9_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_0_0
#define init_2_10_0_0
void FASTCALL call_2_10_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_0_1
#define init_2_10_0_1
void FASTCALL call_2_10_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_0_2
#define init_2_10_0_2
void FASTCALL call_2_10_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_0_3
#define init_2_10_0_3
void FASTCALL call_2_10_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_0_4
#define init_2_10_0_4
void FASTCALL call_2_10_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_0_5
#define init_2_10_0_5
void FASTCALL call_2_10_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_0_6
#define init_2_10_0_6
void FASTCALL call_2_10_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_0_7
#define init_2_10_0_7
void FASTCALL call_2_10_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_0_8
#define init_2_10_0_8
void FASTCALL call_2_10_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_0_9
#define init_2_10_0_9
void FASTCALL call_2_10_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_0_10
#define init_2_10_0_10
void FASTCALL call_2_10_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_0_11
#define init_2_10_0_11
void FASTCALL call_2_10_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_0_12
#define init_2_10_0_12
void FASTCALL call_2_10_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_0_13
#define init_2_10_0_13
void FASTCALL call_2_10_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_0_14
#define init_2_10_0_14
void FASTCALL call_2_10_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_0_15
#define init_2_10_0_15
void FASTCALL call_2_10_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_1_0
#define init_2_10_1_0
void FASTCALL call_2_10_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_1_1
#define init_2_10_1_1
void FASTCALL call_2_10_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_1_2
#define init_2_10_1_2
void FASTCALL call_2_10_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_1_3
#define init_2_10_1_3
void FASTCALL call_2_10_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_1_4
#define init_2_10_1_4
void FASTCALL call_2_10_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_1_5
#define init_2_10_1_5
void FASTCALL call_2_10_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_1_6
#define init_2_10_1_6
void FASTCALL call_2_10_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_1_7
#define init_2_10_1_7
void FASTCALL call_2_10_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_1_8
#define init_2_10_1_8
void FASTCALL call_2_10_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_1_9
#define init_2_10_1_9
void FASTCALL call_2_10_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_1_10
#define init_2_10_1_10
void FASTCALL call_2_10_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_1_11
#define init_2_10_1_11
void FASTCALL call_2_10_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_1_12
#define init_2_10_1_12
void FASTCALL call_2_10_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_1_13
#define init_2_10_1_13
void FASTCALL call_2_10_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_1_14
#define init_2_10_1_14
void FASTCALL call_2_10_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_1_15
#define init_2_10_1_15
void FASTCALL call_2_10_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_2_0
#define init_2_10_2_0
void FASTCALL call_2_10_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_2_1
#define init_2_10_2_1
void FASTCALL call_2_10_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_2_2
#define init_2_10_2_2
void FASTCALL call_2_10_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_2_3
#define init_2_10_2_3
void FASTCALL call_2_10_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_2_4
#define init_2_10_2_4
void FASTCALL call_2_10_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_2_5
#define init_2_10_2_5
void FASTCALL call_2_10_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_2_6
#define init_2_10_2_6
void FASTCALL call_2_10_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_2_7
#define init_2_10_2_7
void FASTCALL call_2_10_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_2_8
#define init_2_10_2_8
void FASTCALL call_2_10_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_2_9
#define init_2_10_2_9
void FASTCALL call_2_10_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_2_10
#define init_2_10_2_10
void FASTCALL call_2_10_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_2_11
#define init_2_10_2_11
void FASTCALL call_2_10_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_2_12
#define init_2_10_2_12
void FASTCALL call_2_10_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_2_13
#define init_2_10_2_13
void FASTCALL call_2_10_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_2_14
#define init_2_10_2_14
void FASTCALL call_2_10_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_2_15
#define init_2_10_2_15
void FASTCALL call_2_10_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_3_0
#define init_2_10_3_0
void FASTCALL call_2_10_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_3_1
#define init_2_10_3_1
void FASTCALL call_2_10_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_3_2
#define init_2_10_3_2
void FASTCALL call_2_10_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_3_3
#define init_2_10_3_3
void FASTCALL call_2_10_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_3_4
#define init_2_10_3_4
void FASTCALL call_2_10_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_3_5
#define init_2_10_3_5
void FASTCALL call_2_10_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_3_6
#define init_2_10_3_6
void FASTCALL call_2_10_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_3_7
#define init_2_10_3_7
void FASTCALL call_2_10_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_3_8
#define init_2_10_3_8
void FASTCALL call_2_10_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_3_9
#define init_2_10_3_9
void FASTCALL call_2_10_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_3_10
#define init_2_10_3_10
void FASTCALL call_2_10_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_3_11
#define init_2_10_3_11
void FASTCALL call_2_10_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_3_12
#define init_2_10_3_12
void FASTCALL call_2_10_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_3_13
#define init_2_10_3_13
void FASTCALL call_2_10_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_3_14
#define init_2_10_3_14
void FASTCALL call_2_10_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_3_15
#define init_2_10_3_15
void FASTCALL call_2_10_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_4_0
#define init_2_10_4_0
void FASTCALL call_2_10_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_4_1
#define init_2_10_4_1
void FASTCALL call_2_10_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_4_2
#define init_2_10_4_2
void FASTCALL call_2_10_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_4_3
#define init_2_10_4_3
void FASTCALL call_2_10_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_4_4
#define init_2_10_4_4
void FASTCALL call_2_10_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_4_5
#define init_2_10_4_5
void FASTCALL call_2_10_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_4_6
#define init_2_10_4_6
void FASTCALL call_2_10_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_4_7
#define init_2_10_4_7
void FASTCALL call_2_10_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_4_8
#define init_2_10_4_8
void FASTCALL call_2_10_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_4_9
#define init_2_10_4_9
void FASTCALL call_2_10_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_4_10
#define init_2_10_4_10
void FASTCALL call_2_10_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_4_11
#define init_2_10_4_11
void FASTCALL call_2_10_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_4_12
#define init_2_10_4_12
void FASTCALL call_2_10_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_4_13
#define init_2_10_4_13
void FASTCALL call_2_10_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_4_14
#define init_2_10_4_14
void FASTCALL call_2_10_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_4_15
#define init_2_10_4_15
void FASTCALL call_2_10_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_5_0
#define init_2_10_5_0
void FASTCALL call_2_10_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_5_1
#define init_2_10_5_1
void FASTCALL call_2_10_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_5_2
#define init_2_10_5_2
void FASTCALL call_2_10_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_5_3
#define init_2_10_5_3
void FASTCALL call_2_10_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_5_4
#define init_2_10_5_4
void FASTCALL call_2_10_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_5_5
#define init_2_10_5_5
void FASTCALL call_2_10_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_5_6
#define init_2_10_5_6
void FASTCALL call_2_10_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_5_7
#define init_2_10_5_7
void FASTCALL call_2_10_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_5_8
#define init_2_10_5_8
void FASTCALL call_2_10_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_5_9
#define init_2_10_5_9
void FASTCALL call_2_10_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_5_10
#define init_2_10_5_10
void FASTCALL call_2_10_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_5_11
#define init_2_10_5_11
void FASTCALL call_2_10_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_5_12
#define init_2_10_5_12
void FASTCALL call_2_10_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_5_13
#define init_2_10_5_13
void FASTCALL call_2_10_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_5_14
#define init_2_10_5_14
void FASTCALL call_2_10_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_5_15
#define init_2_10_5_15
void FASTCALL call_2_10_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_6_0
#define init_2_10_6_0
void FASTCALL call_2_10_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_6_1
#define init_2_10_6_1
void FASTCALL call_2_10_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_6_2
#define init_2_10_6_2
void FASTCALL call_2_10_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_6_3
#define init_2_10_6_3
void FASTCALL call_2_10_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_6_4
#define init_2_10_6_4
void FASTCALL call_2_10_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_6_5
#define init_2_10_6_5
void FASTCALL call_2_10_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_6_6
#define init_2_10_6_6
void FASTCALL call_2_10_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_6_7
#define init_2_10_6_7
void FASTCALL call_2_10_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_6_8
#define init_2_10_6_8
void FASTCALL call_2_10_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_6_9
#define init_2_10_6_9
void FASTCALL call_2_10_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_6_10
#define init_2_10_6_10
void FASTCALL call_2_10_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_6_11
#define init_2_10_6_11
void FASTCALL call_2_10_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_6_12
#define init_2_10_6_12
void FASTCALL call_2_10_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_6_13
#define init_2_10_6_13
void FASTCALL call_2_10_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_6_14
#define init_2_10_6_14
void FASTCALL call_2_10_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_6_15
#define init_2_10_6_15
void FASTCALL call_2_10_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_7_0
#define init_2_10_7_0
void FASTCALL call_2_10_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_7_1
#define init_2_10_7_1
void FASTCALL call_2_10_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_7_2
#define init_2_10_7_2
void FASTCALL call_2_10_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_7_3
#define init_2_10_7_3
void FASTCALL call_2_10_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_7_4
#define init_2_10_7_4
void FASTCALL call_2_10_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_7_5
#define init_2_10_7_5
void FASTCALL call_2_10_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_7_6
#define init_2_10_7_6
void FASTCALL call_2_10_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_7_7
#define init_2_10_7_7
void FASTCALL call_2_10_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_7_8
#define init_2_10_7_8
void FASTCALL call_2_10_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_7_9
#define init_2_10_7_9
void FASTCALL call_2_10_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_7_10
#define init_2_10_7_10
void FASTCALL call_2_10_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_7_11
#define init_2_10_7_11
void FASTCALL call_2_10_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_7_12
#define init_2_10_7_12
void FASTCALL call_2_10_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_7_13
#define init_2_10_7_13
void FASTCALL call_2_10_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_7_14
#define init_2_10_7_14
void FASTCALL call_2_10_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_7_15
#define init_2_10_7_15
void FASTCALL call_2_10_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_8_0
#define init_2_10_8_0
void FASTCALL call_2_10_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_8_1
#define init_2_10_8_1
void FASTCALL call_2_10_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_8_2
#define init_2_10_8_2
void FASTCALL call_2_10_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_8_3
#define init_2_10_8_3
void FASTCALL call_2_10_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_8_4
#define init_2_10_8_4
void FASTCALL call_2_10_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_8_5
#define init_2_10_8_5
void FASTCALL call_2_10_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_8_6
#define init_2_10_8_6
void FASTCALL call_2_10_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_8_7
#define init_2_10_8_7
void FASTCALL call_2_10_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_8_8
#define init_2_10_8_8
void FASTCALL call_2_10_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_8_9
#define init_2_10_8_9
void FASTCALL call_2_10_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_8_10
#define init_2_10_8_10
void FASTCALL call_2_10_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_8_11
#define init_2_10_8_11
void FASTCALL call_2_10_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_8_12
#define init_2_10_8_12
void FASTCALL call_2_10_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_8_13
#define init_2_10_8_13
void FASTCALL call_2_10_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_8_14
#define init_2_10_8_14
void FASTCALL call_2_10_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_8_15
#define init_2_10_8_15
void FASTCALL call_2_10_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_9_0
#define init_2_10_9_0
void FASTCALL call_2_10_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_9_1
#define init_2_10_9_1
void FASTCALL call_2_10_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_9_2
#define init_2_10_9_2
void FASTCALL call_2_10_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_9_3
#define init_2_10_9_3
void FASTCALL call_2_10_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_9_4
#define init_2_10_9_4
void FASTCALL call_2_10_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_9_5
#define init_2_10_9_5
void FASTCALL call_2_10_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_9_6
#define init_2_10_9_6
void FASTCALL call_2_10_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_9_7
#define init_2_10_9_7
void FASTCALL call_2_10_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_9_8
#define init_2_10_9_8
void FASTCALL call_2_10_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_9_9
#define init_2_10_9_9
void FASTCALL call_2_10_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_9_10
#define init_2_10_9_10
void FASTCALL call_2_10_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_9_11
#define init_2_10_9_11
void FASTCALL call_2_10_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_9_12
#define init_2_10_9_12
void FASTCALL call_2_10_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_9_13
#define init_2_10_9_13
void FASTCALL call_2_10_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_9_14
#define init_2_10_9_14
void FASTCALL call_2_10_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_9_15
#define init_2_10_9_15
void FASTCALL call_2_10_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_10_0
#define init_2_10_10_0
void FASTCALL call_2_10_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_10_1
#define init_2_10_10_1
void FASTCALL call_2_10_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_10_2
#define init_2_10_10_2
void FASTCALL call_2_10_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_10_3
#define init_2_10_10_3
void FASTCALL call_2_10_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_10_4
#define init_2_10_10_4
void FASTCALL call_2_10_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_10_5
#define init_2_10_10_5
void FASTCALL call_2_10_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_10_6
#define init_2_10_10_6
void FASTCALL call_2_10_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_10_7
#define init_2_10_10_7
void FASTCALL call_2_10_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_10_8
#define init_2_10_10_8
void FASTCALL call_2_10_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_10_9
#define init_2_10_10_9
void FASTCALL call_2_10_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_10_10
#define init_2_10_10_10
void FASTCALL call_2_10_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_10_11
#define init_2_10_10_11
void FASTCALL call_2_10_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_10_12
#define init_2_10_10_12
void FASTCALL call_2_10_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_10_13
#define init_2_10_10_13
void FASTCALL call_2_10_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_10_14
#define init_2_10_10_14
void FASTCALL call_2_10_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_10_15
#define init_2_10_10_15
void FASTCALL call_2_10_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_11_0
#define init_2_10_11_0
void FASTCALL call_2_10_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_11_1
#define init_2_10_11_1
void FASTCALL call_2_10_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_11_2
#define init_2_10_11_2
void FASTCALL call_2_10_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_11_3
#define init_2_10_11_3
void FASTCALL call_2_10_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_11_4
#define init_2_10_11_4
void FASTCALL call_2_10_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_11_5
#define init_2_10_11_5
void FASTCALL call_2_10_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_11_6
#define init_2_10_11_6
void FASTCALL call_2_10_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_11_7
#define init_2_10_11_7
void FASTCALL call_2_10_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_11_8
#define init_2_10_11_8
void FASTCALL call_2_10_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_11_9
#define init_2_10_11_9
void FASTCALL call_2_10_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_11_10
#define init_2_10_11_10
void FASTCALL call_2_10_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_11_11
#define init_2_10_11_11
void FASTCALL call_2_10_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_11_12
#define init_2_10_11_12
void FASTCALL call_2_10_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_11_13
#define init_2_10_11_13
void FASTCALL call_2_10_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_11_14
#define init_2_10_11_14
void FASTCALL call_2_10_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_11_15
#define init_2_10_11_15
void FASTCALL call_2_10_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_12_0
#define init_2_10_12_0
void FASTCALL call_2_10_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_12_1
#define init_2_10_12_1
void FASTCALL call_2_10_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_12_2
#define init_2_10_12_2
void FASTCALL call_2_10_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_12_3
#define init_2_10_12_3
void FASTCALL call_2_10_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_12_4
#define init_2_10_12_4
void FASTCALL call_2_10_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_12_5
#define init_2_10_12_5
void FASTCALL call_2_10_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_12_6
#define init_2_10_12_6
void FASTCALL call_2_10_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_12_7
#define init_2_10_12_7
void FASTCALL call_2_10_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_12_8
#define init_2_10_12_8
void FASTCALL call_2_10_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_12_9
#define init_2_10_12_9
void FASTCALL call_2_10_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_12_10
#define init_2_10_12_10
void FASTCALL call_2_10_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_12_11
#define init_2_10_12_11
void FASTCALL call_2_10_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_12_12
#define init_2_10_12_12
void FASTCALL call_2_10_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_12_13
#define init_2_10_12_13
void FASTCALL call_2_10_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_12_14
#define init_2_10_12_14
void FASTCALL call_2_10_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_12_15
#define init_2_10_12_15
void FASTCALL call_2_10_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_13_0
#define init_2_10_13_0
void FASTCALL call_2_10_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_13_1
#define init_2_10_13_1
void FASTCALL call_2_10_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_13_2
#define init_2_10_13_2
void FASTCALL call_2_10_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_13_3
#define init_2_10_13_3
void FASTCALL call_2_10_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_13_4
#define init_2_10_13_4
void FASTCALL call_2_10_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_13_5
#define init_2_10_13_5
void FASTCALL call_2_10_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_13_6
#define init_2_10_13_6
void FASTCALL call_2_10_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_13_7
#define init_2_10_13_7
void FASTCALL call_2_10_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_13_8
#define init_2_10_13_8
void FASTCALL call_2_10_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_13_9
#define init_2_10_13_9
void FASTCALL call_2_10_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_13_10
#define init_2_10_13_10
void FASTCALL call_2_10_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_13_11
#define init_2_10_13_11
void FASTCALL call_2_10_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_13_12
#define init_2_10_13_12
void FASTCALL call_2_10_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_13_13
#define init_2_10_13_13
void FASTCALL call_2_10_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_13_14
#define init_2_10_13_14
void FASTCALL call_2_10_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_13_15
#define init_2_10_13_15
void FASTCALL call_2_10_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_14_0
#define init_2_10_14_0
void FASTCALL call_2_10_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_14_1
#define init_2_10_14_1
void FASTCALL call_2_10_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_14_2
#define init_2_10_14_2
void FASTCALL call_2_10_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_14_3
#define init_2_10_14_3
void FASTCALL call_2_10_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_14_4
#define init_2_10_14_4
void FASTCALL call_2_10_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_14_5
#define init_2_10_14_5
void FASTCALL call_2_10_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_14_6
#define init_2_10_14_6
void FASTCALL call_2_10_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_14_7
#define init_2_10_14_7
void FASTCALL call_2_10_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_14_8
#define init_2_10_14_8
void FASTCALL call_2_10_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_14_9
#define init_2_10_14_9
void FASTCALL call_2_10_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_14_10
#define init_2_10_14_10
void FASTCALL call_2_10_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_14_11
#define init_2_10_14_11
void FASTCALL call_2_10_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_14_12
#define init_2_10_14_12
void FASTCALL call_2_10_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_14_13
#define init_2_10_14_13
void FASTCALL call_2_10_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_14_14
#define init_2_10_14_14
void FASTCALL call_2_10_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_14_15
#define init_2_10_14_15
void FASTCALL call_2_10_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_15_0
#define init_2_10_15_0
void FASTCALL call_2_10_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_15_1
#define init_2_10_15_1
void FASTCALL call_2_10_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_15_2
#define init_2_10_15_2
void FASTCALL call_2_10_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_15_3
#define init_2_10_15_3
void FASTCALL call_2_10_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_15_4
#define init_2_10_15_4
void FASTCALL call_2_10_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_15_5
#define init_2_10_15_5
void FASTCALL call_2_10_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_15_6
#define init_2_10_15_6
void FASTCALL call_2_10_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_15_7
#define init_2_10_15_7
void FASTCALL call_2_10_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_15_8
#define init_2_10_15_8
void FASTCALL call_2_10_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_15_9
#define init_2_10_15_9
void FASTCALL call_2_10_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_15_10
#define init_2_10_15_10
void FASTCALL call_2_10_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_15_11
#define init_2_10_15_11
void FASTCALL call_2_10_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_15_12
#define init_2_10_15_12
void FASTCALL call_2_10_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_15_13
#define init_2_10_15_13
void FASTCALL call_2_10_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_15_14
#define init_2_10_15_14
void FASTCALL call_2_10_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_10_15_15
#define init_2_10_15_15
void FASTCALL call_2_10_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_0_0
#define init_2_11_0_0
void FASTCALL call_2_11_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_0_1
#define init_2_11_0_1
void FASTCALL call_2_11_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_0_2
#define init_2_11_0_2
void FASTCALL call_2_11_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_0_3
#define init_2_11_0_3
void FASTCALL call_2_11_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_0_4
#define init_2_11_0_4
void FASTCALL call_2_11_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_0_5
#define init_2_11_0_5
void FASTCALL call_2_11_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_0_6
#define init_2_11_0_6
void FASTCALL call_2_11_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_0_7
#define init_2_11_0_7
void FASTCALL call_2_11_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_0_8
#define init_2_11_0_8
void FASTCALL call_2_11_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_0_9
#define init_2_11_0_9
void FASTCALL call_2_11_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_0_10
#define init_2_11_0_10
void FASTCALL call_2_11_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_0_11
#define init_2_11_0_11
void FASTCALL call_2_11_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_0_12
#define init_2_11_0_12
void FASTCALL call_2_11_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_0_13
#define init_2_11_0_13
void FASTCALL call_2_11_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_0_14
#define init_2_11_0_14
void FASTCALL call_2_11_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_0_15
#define init_2_11_0_15
void FASTCALL call_2_11_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_1_0
#define init_2_11_1_0
void FASTCALL call_2_11_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_1_1
#define init_2_11_1_1
void FASTCALL call_2_11_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_1_2
#define init_2_11_1_2
void FASTCALL call_2_11_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_1_3
#define init_2_11_1_3
void FASTCALL call_2_11_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_1_4
#define init_2_11_1_4
void FASTCALL call_2_11_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_1_5
#define init_2_11_1_5
void FASTCALL call_2_11_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_1_6
#define init_2_11_1_6
void FASTCALL call_2_11_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_1_7
#define init_2_11_1_7
void FASTCALL call_2_11_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_1_8
#define init_2_11_1_8
void FASTCALL call_2_11_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_1_9
#define init_2_11_1_9
void FASTCALL call_2_11_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_1_10
#define init_2_11_1_10
void FASTCALL call_2_11_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_1_11
#define init_2_11_1_11
void FASTCALL call_2_11_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_1_12
#define init_2_11_1_12
void FASTCALL call_2_11_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_1_13
#define init_2_11_1_13
void FASTCALL call_2_11_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_1_14
#define init_2_11_1_14
void FASTCALL call_2_11_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_1_15
#define init_2_11_1_15
void FASTCALL call_2_11_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_2_0
#define init_2_11_2_0
void FASTCALL call_2_11_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_2_1
#define init_2_11_2_1
void FASTCALL call_2_11_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_2_2
#define init_2_11_2_2
void FASTCALL call_2_11_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_2_3
#define init_2_11_2_3
void FASTCALL call_2_11_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_2_4
#define init_2_11_2_4
void FASTCALL call_2_11_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_2_5
#define init_2_11_2_5
void FASTCALL call_2_11_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_2_6
#define init_2_11_2_6
void FASTCALL call_2_11_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_2_7
#define init_2_11_2_7
void FASTCALL call_2_11_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_2_8
#define init_2_11_2_8
void FASTCALL call_2_11_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_2_9
#define init_2_11_2_9
void FASTCALL call_2_11_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_2_10
#define init_2_11_2_10
void FASTCALL call_2_11_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_2_11
#define init_2_11_2_11
void FASTCALL call_2_11_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_2_12
#define init_2_11_2_12
void FASTCALL call_2_11_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_2_13
#define init_2_11_2_13
void FASTCALL call_2_11_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_2_14
#define init_2_11_2_14
void FASTCALL call_2_11_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_2_15
#define init_2_11_2_15
void FASTCALL call_2_11_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_3_0
#define init_2_11_3_0
void FASTCALL call_2_11_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_3_1
#define init_2_11_3_1
void FASTCALL call_2_11_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_3_2
#define init_2_11_3_2
void FASTCALL call_2_11_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_3_3
#define init_2_11_3_3
void FASTCALL call_2_11_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_3_4
#define init_2_11_3_4
void FASTCALL call_2_11_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_3_5
#define init_2_11_3_5
void FASTCALL call_2_11_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_3_6
#define init_2_11_3_6
void FASTCALL call_2_11_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_3_7
#define init_2_11_3_7
void FASTCALL call_2_11_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_3_8
#define init_2_11_3_8
void FASTCALL call_2_11_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_3_9
#define init_2_11_3_9
void FASTCALL call_2_11_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_3_10
#define init_2_11_3_10
void FASTCALL call_2_11_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_3_11
#define init_2_11_3_11
void FASTCALL call_2_11_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_3_12
#define init_2_11_3_12
void FASTCALL call_2_11_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_3_13
#define init_2_11_3_13
void FASTCALL call_2_11_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_3_14
#define init_2_11_3_14
void FASTCALL call_2_11_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_3_15
#define init_2_11_3_15
void FASTCALL call_2_11_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_4_0
#define init_2_11_4_0
void FASTCALL call_2_11_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_4_1
#define init_2_11_4_1
void FASTCALL call_2_11_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_4_2
#define init_2_11_4_2
void FASTCALL call_2_11_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_4_3
#define init_2_11_4_3
void FASTCALL call_2_11_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_4_4
#define init_2_11_4_4
void FASTCALL call_2_11_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_4_5
#define init_2_11_4_5
void FASTCALL call_2_11_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_4_6
#define init_2_11_4_6
void FASTCALL call_2_11_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_4_7
#define init_2_11_4_7
void FASTCALL call_2_11_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_4_8
#define init_2_11_4_8
void FASTCALL call_2_11_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_4_9
#define init_2_11_4_9
void FASTCALL call_2_11_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_4_10
#define init_2_11_4_10
void FASTCALL call_2_11_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_4_11
#define init_2_11_4_11
void FASTCALL call_2_11_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_4_12
#define init_2_11_4_12
void FASTCALL call_2_11_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_4_13
#define init_2_11_4_13
void FASTCALL call_2_11_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_4_14
#define init_2_11_4_14
void FASTCALL call_2_11_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_4_15
#define init_2_11_4_15
void FASTCALL call_2_11_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_5_0
#define init_2_11_5_0
void FASTCALL call_2_11_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_5_1
#define init_2_11_5_1
void FASTCALL call_2_11_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_5_2
#define init_2_11_5_2
void FASTCALL call_2_11_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_5_3
#define init_2_11_5_3
void FASTCALL call_2_11_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_5_4
#define init_2_11_5_4
void FASTCALL call_2_11_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_5_5
#define init_2_11_5_5
void FASTCALL call_2_11_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_5_6
#define init_2_11_5_6
void FASTCALL call_2_11_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_5_7
#define init_2_11_5_7
void FASTCALL call_2_11_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_5_8
#define init_2_11_5_8
void FASTCALL call_2_11_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_5_9
#define init_2_11_5_9
void FASTCALL call_2_11_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_5_10
#define init_2_11_5_10
void FASTCALL call_2_11_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_5_11
#define init_2_11_5_11
void FASTCALL call_2_11_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_5_12
#define init_2_11_5_12
void FASTCALL call_2_11_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_5_13
#define init_2_11_5_13
void FASTCALL call_2_11_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_5_14
#define init_2_11_5_14
void FASTCALL call_2_11_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_5_15
#define init_2_11_5_15
void FASTCALL call_2_11_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_6_0
#define init_2_11_6_0
void FASTCALL call_2_11_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_6_1
#define init_2_11_6_1
void FASTCALL call_2_11_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_6_2
#define init_2_11_6_2
void FASTCALL call_2_11_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_6_3
#define init_2_11_6_3
void FASTCALL call_2_11_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_6_4
#define init_2_11_6_4
void FASTCALL call_2_11_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_6_5
#define init_2_11_6_5
void FASTCALL call_2_11_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_6_6
#define init_2_11_6_6
void FASTCALL call_2_11_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_6_7
#define init_2_11_6_7
void FASTCALL call_2_11_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_6_8
#define init_2_11_6_8
void FASTCALL call_2_11_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_6_9
#define init_2_11_6_9
void FASTCALL call_2_11_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_6_10
#define init_2_11_6_10
void FASTCALL call_2_11_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_6_11
#define init_2_11_6_11
void FASTCALL call_2_11_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_6_12
#define init_2_11_6_12
void FASTCALL call_2_11_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_6_13
#define init_2_11_6_13
void FASTCALL call_2_11_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_6_14
#define init_2_11_6_14
void FASTCALL call_2_11_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_6_15
#define init_2_11_6_15
void FASTCALL call_2_11_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_7_0
#define init_2_11_7_0
void FASTCALL call_2_11_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_7_1
#define init_2_11_7_1
void FASTCALL call_2_11_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_7_2
#define init_2_11_7_2
void FASTCALL call_2_11_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_7_3
#define init_2_11_7_3
void FASTCALL call_2_11_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_7_4
#define init_2_11_7_4
void FASTCALL call_2_11_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_7_5
#define init_2_11_7_5
void FASTCALL call_2_11_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_7_6
#define init_2_11_7_6
void FASTCALL call_2_11_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_7_7
#define init_2_11_7_7
void FASTCALL call_2_11_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_7_8
#define init_2_11_7_8
void FASTCALL call_2_11_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_7_9
#define init_2_11_7_9
void FASTCALL call_2_11_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_7_10
#define init_2_11_7_10
void FASTCALL call_2_11_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_7_11
#define init_2_11_7_11
void FASTCALL call_2_11_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_7_12
#define init_2_11_7_12
void FASTCALL call_2_11_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_7_13
#define init_2_11_7_13
void FASTCALL call_2_11_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_7_14
#define init_2_11_7_14
void FASTCALL call_2_11_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_7_15
#define init_2_11_7_15
void FASTCALL call_2_11_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_8_0
#define init_2_11_8_0
void FASTCALL call_2_11_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_8_1
#define init_2_11_8_1
void FASTCALL call_2_11_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_8_2
#define init_2_11_8_2
void FASTCALL call_2_11_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_8_3
#define init_2_11_8_3
void FASTCALL call_2_11_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_8_4
#define init_2_11_8_4
void FASTCALL call_2_11_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_8_5
#define init_2_11_8_5
void FASTCALL call_2_11_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_8_6
#define init_2_11_8_6
void FASTCALL call_2_11_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_8_7
#define init_2_11_8_7
void FASTCALL call_2_11_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_8_8
#define init_2_11_8_8
void FASTCALL call_2_11_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_8_9
#define init_2_11_8_9
void FASTCALL call_2_11_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_8_10
#define init_2_11_8_10
void FASTCALL call_2_11_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_8_11
#define init_2_11_8_11
void FASTCALL call_2_11_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_8_12
#define init_2_11_8_12
void FASTCALL call_2_11_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_8_13
#define init_2_11_8_13
void FASTCALL call_2_11_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_8_14
#define init_2_11_8_14
void FASTCALL call_2_11_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_8_15
#define init_2_11_8_15
void FASTCALL call_2_11_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_9_0
#define init_2_11_9_0
void FASTCALL call_2_11_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_9_1
#define init_2_11_9_1
void FASTCALL call_2_11_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_9_2
#define init_2_11_9_2
void FASTCALL call_2_11_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_9_3
#define init_2_11_9_3
void FASTCALL call_2_11_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_9_4
#define init_2_11_9_4
void FASTCALL call_2_11_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_9_5
#define init_2_11_9_5
void FASTCALL call_2_11_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_9_6
#define init_2_11_9_6
void FASTCALL call_2_11_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_9_7
#define init_2_11_9_7
void FASTCALL call_2_11_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_9_8
#define init_2_11_9_8
void FASTCALL call_2_11_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_9_9
#define init_2_11_9_9
void FASTCALL call_2_11_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_9_10
#define init_2_11_9_10
void FASTCALL call_2_11_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_9_11
#define init_2_11_9_11
void FASTCALL call_2_11_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_9_12
#define init_2_11_9_12
void FASTCALL call_2_11_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_9_13
#define init_2_11_9_13
void FASTCALL call_2_11_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_9_14
#define init_2_11_9_14
void FASTCALL call_2_11_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_9_15
#define init_2_11_9_15
void FASTCALL call_2_11_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_10_0
#define init_2_11_10_0
void FASTCALL call_2_11_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_10_1
#define init_2_11_10_1
void FASTCALL call_2_11_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_10_2
#define init_2_11_10_2
void FASTCALL call_2_11_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_10_3
#define init_2_11_10_3
void FASTCALL call_2_11_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_10_4
#define init_2_11_10_4
void FASTCALL call_2_11_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_10_5
#define init_2_11_10_5
void FASTCALL call_2_11_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_10_6
#define init_2_11_10_6
void FASTCALL call_2_11_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_10_7
#define init_2_11_10_7
void FASTCALL call_2_11_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_10_8
#define init_2_11_10_8
void FASTCALL call_2_11_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_10_9
#define init_2_11_10_9
void FASTCALL call_2_11_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_10_10
#define init_2_11_10_10
void FASTCALL call_2_11_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_10_11
#define init_2_11_10_11
void FASTCALL call_2_11_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_10_12
#define init_2_11_10_12
void FASTCALL call_2_11_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_10_13
#define init_2_11_10_13
void FASTCALL call_2_11_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_10_14
#define init_2_11_10_14
void FASTCALL call_2_11_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_10_15
#define init_2_11_10_15
void FASTCALL call_2_11_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_11_0
#define init_2_11_11_0
void FASTCALL call_2_11_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_11_1
#define init_2_11_11_1
void FASTCALL call_2_11_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_11_2
#define init_2_11_11_2
void FASTCALL call_2_11_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_11_3
#define init_2_11_11_3
void FASTCALL call_2_11_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_11_4
#define init_2_11_11_4
void FASTCALL call_2_11_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_11_5
#define init_2_11_11_5
void FASTCALL call_2_11_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_11_6
#define init_2_11_11_6
void FASTCALL call_2_11_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_11_7
#define init_2_11_11_7
void FASTCALL call_2_11_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_11_8
#define init_2_11_11_8
void FASTCALL call_2_11_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_11_9
#define init_2_11_11_9
void FASTCALL call_2_11_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_11_10
#define init_2_11_11_10
void FASTCALL call_2_11_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_11_11
#define init_2_11_11_11
void FASTCALL call_2_11_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_11_12
#define init_2_11_11_12
void FASTCALL call_2_11_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_11_13
#define init_2_11_11_13
void FASTCALL call_2_11_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_11_14
#define init_2_11_11_14
void FASTCALL call_2_11_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_11_15
#define init_2_11_11_15
void FASTCALL call_2_11_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_12_0
#define init_2_11_12_0
void FASTCALL call_2_11_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_12_1
#define init_2_11_12_1
void FASTCALL call_2_11_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_12_2
#define init_2_11_12_2
void FASTCALL call_2_11_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_12_3
#define init_2_11_12_3
void FASTCALL call_2_11_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_12_4
#define init_2_11_12_4
void FASTCALL call_2_11_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_12_5
#define init_2_11_12_5
void FASTCALL call_2_11_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_12_6
#define init_2_11_12_6
void FASTCALL call_2_11_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_12_7
#define init_2_11_12_7
void FASTCALL call_2_11_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_12_8
#define init_2_11_12_8
void FASTCALL call_2_11_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_12_9
#define init_2_11_12_9
void FASTCALL call_2_11_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_12_10
#define init_2_11_12_10
void FASTCALL call_2_11_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_12_11
#define init_2_11_12_11
void FASTCALL call_2_11_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_12_12
#define init_2_11_12_12
void FASTCALL call_2_11_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_12_13
#define init_2_11_12_13
void FASTCALL call_2_11_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_12_14
#define init_2_11_12_14
void FASTCALL call_2_11_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_12_15
#define init_2_11_12_15
void FASTCALL call_2_11_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_13_0
#define init_2_11_13_0
void FASTCALL call_2_11_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_13_1
#define init_2_11_13_1
void FASTCALL call_2_11_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_13_2
#define init_2_11_13_2
void FASTCALL call_2_11_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_13_3
#define init_2_11_13_3
void FASTCALL call_2_11_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_13_4
#define init_2_11_13_4
void FASTCALL call_2_11_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_13_5
#define init_2_11_13_5
void FASTCALL call_2_11_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_13_6
#define init_2_11_13_6
void FASTCALL call_2_11_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_13_7
#define init_2_11_13_7
void FASTCALL call_2_11_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_13_8
#define init_2_11_13_8
void FASTCALL call_2_11_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_13_9
#define init_2_11_13_9
void FASTCALL call_2_11_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_13_10
#define init_2_11_13_10
void FASTCALL call_2_11_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_13_11
#define init_2_11_13_11
void FASTCALL call_2_11_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_13_12
#define init_2_11_13_12
void FASTCALL call_2_11_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_13_13
#define init_2_11_13_13
void FASTCALL call_2_11_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_13_14
#define init_2_11_13_14
void FASTCALL call_2_11_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_13_15
#define init_2_11_13_15
void FASTCALL call_2_11_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_14_0
#define init_2_11_14_0
void FASTCALL call_2_11_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_14_1
#define init_2_11_14_1
void FASTCALL call_2_11_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_14_2
#define init_2_11_14_2
void FASTCALL call_2_11_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_14_3
#define init_2_11_14_3
void FASTCALL call_2_11_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_14_4
#define init_2_11_14_4
void FASTCALL call_2_11_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_14_5
#define init_2_11_14_5
void FASTCALL call_2_11_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_14_6
#define init_2_11_14_6
void FASTCALL call_2_11_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_14_7
#define init_2_11_14_7
void FASTCALL call_2_11_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_14_8
#define init_2_11_14_8
void FASTCALL call_2_11_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_14_9
#define init_2_11_14_9
void FASTCALL call_2_11_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_14_10
#define init_2_11_14_10
void FASTCALL call_2_11_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_14_11
#define init_2_11_14_11
void FASTCALL call_2_11_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_14_12
#define init_2_11_14_12
void FASTCALL call_2_11_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_14_13
#define init_2_11_14_13
void FASTCALL call_2_11_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_14_14
#define init_2_11_14_14
void FASTCALL call_2_11_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_14_15
#define init_2_11_14_15
void FASTCALL call_2_11_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_15_0
#define init_2_11_15_0
void FASTCALL call_2_11_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_15_1
#define init_2_11_15_1
void FASTCALL call_2_11_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_15_2
#define init_2_11_15_2
void FASTCALL call_2_11_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_15_3
#define init_2_11_15_3
void FASTCALL call_2_11_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_15_4
#define init_2_11_15_4
void FASTCALL call_2_11_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_15_5
#define init_2_11_15_5
void FASTCALL call_2_11_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_15_6
#define init_2_11_15_6
void FASTCALL call_2_11_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_15_7
#define init_2_11_15_7
void FASTCALL call_2_11_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_15_8
#define init_2_11_15_8
void FASTCALL call_2_11_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_15_9
#define init_2_11_15_9
void FASTCALL call_2_11_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_15_10
#define init_2_11_15_10
void FASTCALL call_2_11_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_15_11
#define init_2_11_15_11
void FASTCALL call_2_11_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_15_12
#define init_2_11_15_12
void FASTCALL call_2_11_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_15_13
#define init_2_11_15_13
void FASTCALL call_2_11_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_15_14
#define init_2_11_15_14
void FASTCALL call_2_11_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_11_15_15
#define init_2_11_15_15
void FASTCALL call_2_11_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_0_0
#define init_2_12_0_0
void FASTCALL call_2_12_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_0_1
#define init_2_12_0_1
void FASTCALL call_2_12_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_0_2
#define init_2_12_0_2
void FASTCALL call_2_12_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_0_3
#define init_2_12_0_3
void FASTCALL call_2_12_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_0_4
#define init_2_12_0_4
void FASTCALL call_2_12_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_0_5
#define init_2_12_0_5
void FASTCALL call_2_12_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_0_6
#define init_2_12_0_6
void FASTCALL call_2_12_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_0_7
#define init_2_12_0_7
void FASTCALL call_2_12_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_0_8
#define init_2_12_0_8
void FASTCALL call_2_12_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_0_9
#define init_2_12_0_9
void FASTCALL call_2_12_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_0_10
#define init_2_12_0_10
void FASTCALL call_2_12_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_0_11
#define init_2_12_0_11
void FASTCALL call_2_12_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_0_12
#define init_2_12_0_12
void FASTCALL call_2_12_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_0_13
#define init_2_12_0_13
void FASTCALL call_2_12_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_0_14
#define init_2_12_0_14
void FASTCALL call_2_12_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_0_15
#define init_2_12_0_15
void FASTCALL call_2_12_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_1_0
#define init_2_12_1_0
void FASTCALL call_2_12_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_1_1
#define init_2_12_1_1
void FASTCALL call_2_12_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_1_2
#define init_2_12_1_2
void FASTCALL call_2_12_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_1_3
#define init_2_12_1_3
void FASTCALL call_2_12_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_1_4
#define init_2_12_1_4
void FASTCALL call_2_12_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_1_5
#define init_2_12_1_5
void FASTCALL call_2_12_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_1_6
#define init_2_12_1_6
void FASTCALL call_2_12_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_1_7
#define init_2_12_1_7
void FASTCALL call_2_12_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_1_8
#define init_2_12_1_8
void FASTCALL call_2_12_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_1_9
#define init_2_12_1_9
void FASTCALL call_2_12_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_1_10
#define init_2_12_1_10
void FASTCALL call_2_12_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_1_11
#define init_2_12_1_11
void FASTCALL call_2_12_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_1_12
#define init_2_12_1_12
void FASTCALL call_2_12_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_1_13
#define init_2_12_1_13
void FASTCALL call_2_12_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_1_14
#define init_2_12_1_14
void FASTCALL call_2_12_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_1_15
#define init_2_12_1_15
void FASTCALL call_2_12_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_2_0
#define init_2_12_2_0
void FASTCALL call_2_12_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_2_1
#define init_2_12_2_1
void FASTCALL call_2_12_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_2_2
#define init_2_12_2_2
void FASTCALL call_2_12_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_2_3
#define init_2_12_2_3
void FASTCALL call_2_12_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_2_4
#define init_2_12_2_4
void FASTCALL call_2_12_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_2_5
#define init_2_12_2_5
void FASTCALL call_2_12_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_2_6
#define init_2_12_2_6
void FASTCALL call_2_12_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_2_7
#define init_2_12_2_7
void FASTCALL call_2_12_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_2_8
#define init_2_12_2_8
void FASTCALL call_2_12_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_2_9
#define init_2_12_2_9
void FASTCALL call_2_12_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_2_10
#define init_2_12_2_10
void FASTCALL call_2_12_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_2_11
#define init_2_12_2_11
void FASTCALL call_2_12_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_2_12
#define init_2_12_2_12
void FASTCALL call_2_12_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_2_13
#define init_2_12_2_13
void FASTCALL call_2_12_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_2_14
#define init_2_12_2_14
void FASTCALL call_2_12_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_2_15
#define init_2_12_2_15
void FASTCALL call_2_12_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_3_0
#define init_2_12_3_0
void FASTCALL call_2_12_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_3_1
#define init_2_12_3_1
void FASTCALL call_2_12_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_3_2
#define init_2_12_3_2
void FASTCALL call_2_12_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_3_3
#define init_2_12_3_3
void FASTCALL call_2_12_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_3_4
#define init_2_12_3_4
void FASTCALL call_2_12_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_3_5
#define init_2_12_3_5
void FASTCALL call_2_12_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_3_6
#define init_2_12_3_6
void FASTCALL call_2_12_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_3_7
#define init_2_12_3_7
void FASTCALL call_2_12_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_3_8
#define init_2_12_3_8
void FASTCALL call_2_12_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_3_9
#define init_2_12_3_9
void FASTCALL call_2_12_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_3_10
#define init_2_12_3_10
void FASTCALL call_2_12_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_3_11
#define init_2_12_3_11
void FASTCALL call_2_12_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_3_12
#define init_2_12_3_12
void FASTCALL call_2_12_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_3_13
#define init_2_12_3_13
void FASTCALL call_2_12_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_3_14
#define init_2_12_3_14
void FASTCALL call_2_12_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_3_15
#define init_2_12_3_15
void FASTCALL call_2_12_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_4_0
#define init_2_12_4_0
void FASTCALL call_2_12_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_4_1
#define init_2_12_4_1
void FASTCALL call_2_12_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_4_2
#define init_2_12_4_2
void FASTCALL call_2_12_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_4_3
#define init_2_12_4_3
void FASTCALL call_2_12_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_4_4
#define init_2_12_4_4
void FASTCALL call_2_12_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_4_5
#define init_2_12_4_5
void FASTCALL call_2_12_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_4_6
#define init_2_12_4_6
void FASTCALL call_2_12_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_4_7
#define init_2_12_4_7
void FASTCALL call_2_12_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_4_8
#define init_2_12_4_8
void FASTCALL call_2_12_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_4_9
#define init_2_12_4_9
void FASTCALL call_2_12_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_4_10
#define init_2_12_4_10
void FASTCALL call_2_12_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_4_11
#define init_2_12_4_11
void FASTCALL call_2_12_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_4_12
#define init_2_12_4_12
void FASTCALL call_2_12_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_4_13
#define init_2_12_4_13
void FASTCALL call_2_12_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_4_14
#define init_2_12_4_14
void FASTCALL call_2_12_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_4_15
#define init_2_12_4_15
void FASTCALL call_2_12_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_5_0
#define init_2_12_5_0
void FASTCALL call_2_12_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_5_1
#define init_2_12_5_1
void FASTCALL call_2_12_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_5_2
#define init_2_12_5_2
void FASTCALL call_2_12_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_5_3
#define init_2_12_5_3
void FASTCALL call_2_12_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_5_4
#define init_2_12_5_4
void FASTCALL call_2_12_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_5_5
#define init_2_12_5_5
void FASTCALL call_2_12_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_5_6
#define init_2_12_5_6
void FASTCALL call_2_12_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_5_7
#define init_2_12_5_7
void FASTCALL call_2_12_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_5_8
#define init_2_12_5_8
void FASTCALL call_2_12_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_5_9
#define init_2_12_5_9
void FASTCALL call_2_12_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_5_10
#define init_2_12_5_10
void FASTCALL call_2_12_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_5_11
#define init_2_12_5_11
void FASTCALL call_2_12_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_5_12
#define init_2_12_5_12
void FASTCALL call_2_12_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_5_13
#define init_2_12_5_13
void FASTCALL call_2_12_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_5_14
#define init_2_12_5_14
void FASTCALL call_2_12_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_5_15
#define init_2_12_5_15
void FASTCALL call_2_12_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_6_0
#define init_2_12_6_0
void FASTCALL call_2_12_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_6_1
#define init_2_12_6_1
void FASTCALL call_2_12_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_6_2
#define init_2_12_6_2
void FASTCALL call_2_12_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_6_3
#define init_2_12_6_3
void FASTCALL call_2_12_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_6_4
#define init_2_12_6_4
void FASTCALL call_2_12_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_6_5
#define init_2_12_6_5
void FASTCALL call_2_12_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_6_6
#define init_2_12_6_6
void FASTCALL call_2_12_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_6_7
#define init_2_12_6_7
void FASTCALL call_2_12_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_6_8
#define init_2_12_6_8
void FASTCALL call_2_12_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_6_9
#define init_2_12_6_9
void FASTCALL call_2_12_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_6_10
#define init_2_12_6_10
void FASTCALL call_2_12_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_6_11
#define init_2_12_6_11
void FASTCALL call_2_12_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_6_12
#define init_2_12_6_12
void FASTCALL call_2_12_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_6_13
#define init_2_12_6_13
void FASTCALL call_2_12_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_6_14
#define init_2_12_6_14
void FASTCALL call_2_12_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_6_15
#define init_2_12_6_15
void FASTCALL call_2_12_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_7_0
#define init_2_12_7_0
void FASTCALL call_2_12_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_7_1
#define init_2_12_7_1
void FASTCALL call_2_12_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_7_2
#define init_2_12_7_2
void FASTCALL call_2_12_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_7_3
#define init_2_12_7_3
void FASTCALL call_2_12_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_7_4
#define init_2_12_7_4
void FASTCALL call_2_12_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_7_5
#define init_2_12_7_5
void FASTCALL call_2_12_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_7_6
#define init_2_12_7_6
void FASTCALL call_2_12_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_7_7
#define init_2_12_7_7
void FASTCALL call_2_12_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_7_8
#define init_2_12_7_8
void FASTCALL call_2_12_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_7_9
#define init_2_12_7_9
void FASTCALL call_2_12_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_7_10
#define init_2_12_7_10
void FASTCALL call_2_12_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_7_11
#define init_2_12_7_11
void FASTCALL call_2_12_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_7_12
#define init_2_12_7_12
void FASTCALL call_2_12_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_7_13
#define init_2_12_7_13
void FASTCALL call_2_12_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_7_14
#define init_2_12_7_14
void FASTCALL call_2_12_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_7_15
#define init_2_12_7_15
void FASTCALL call_2_12_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_8_0
#define init_2_12_8_0
void FASTCALL call_2_12_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_8_1
#define init_2_12_8_1
void FASTCALL call_2_12_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_8_2
#define init_2_12_8_2
void FASTCALL call_2_12_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_8_3
#define init_2_12_8_3
void FASTCALL call_2_12_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_8_4
#define init_2_12_8_4
void FASTCALL call_2_12_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_8_5
#define init_2_12_8_5
void FASTCALL call_2_12_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_8_6
#define init_2_12_8_6
void FASTCALL call_2_12_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_8_7
#define init_2_12_8_7
void FASTCALL call_2_12_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_8_8
#define init_2_12_8_8
void FASTCALL call_2_12_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_8_9
#define init_2_12_8_9
void FASTCALL call_2_12_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_8_10
#define init_2_12_8_10
void FASTCALL call_2_12_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_8_11
#define init_2_12_8_11
void FASTCALL call_2_12_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_8_12
#define init_2_12_8_12
void FASTCALL call_2_12_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_8_13
#define init_2_12_8_13
void FASTCALL call_2_12_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_8_14
#define init_2_12_8_14
void FASTCALL call_2_12_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_8_15
#define init_2_12_8_15
void FASTCALL call_2_12_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_9_0
#define init_2_12_9_0
void FASTCALL call_2_12_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_9_1
#define init_2_12_9_1
void FASTCALL call_2_12_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_9_2
#define init_2_12_9_2
void FASTCALL call_2_12_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_9_3
#define init_2_12_9_3
void FASTCALL call_2_12_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_9_4
#define init_2_12_9_4
void FASTCALL call_2_12_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_9_5
#define init_2_12_9_5
void FASTCALL call_2_12_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_9_6
#define init_2_12_9_6
void FASTCALL call_2_12_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_9_7
#define init_2_12_9_7
void FASTCALL call_2_12_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_9_8
#define init_2_12_9_8
void FASTCALL call_2_12_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_9_9
#define init_2_12_9_9
void FASTCALL call_2_12_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_9_10
#define init_2_12_9_10
void FASTCALL call_2_12_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_9_11
#define init_2_12_9_11
void FASTCALL call_2_12_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_9_12
#define init_2_12_9_12
void FASTCALL call_2_12_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_9_13
#define init_2_12_9_13
void FASTCALL call_2_12_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_9_14
#define init_2_12_9_14
void FASTCALL call_2_12_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_9_15
#define init_2_12_9_15
void FASTCALL call_2_12_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_10_0
#define init_2_12_10_0
void FASTCALL call_2_12_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_10_1
#define init_2_12_10_1
void FASTCALL call_2_12_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_10_2
#define init_2_12_10_2
void FASTCALL call_2_12_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_10_3
#define init_2_12_10_3
void FASTCALL call_2_12_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_10_4
#define init_2_12_10_4
void FASTCALL call_2_12_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_10_5
#define init_2_12_10_5
void FASTCALL call_2_12_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_10_6
#define init_2_12_10_6
void FASTCALL call_2_12_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_10_7
#define init_2_12_10_7
void FASTCALL call_2_12_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_10_8
#define init_2_12_10_8
void FASTCALL call_2_12_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_10_9
#define init_2_12_10_9
void FASTCALL call_2_12_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_10_10
#define init_2_12_10_10
void FASTCALL call_2_12_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_10_11
#define init_2_12_10_11
void FASTCALL call_2_12_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_10_12
#define init_2_12_10_12
void FASTCALL call_2_12_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_10_13
#define init_2_12_10_13
void FASTCALL call_2_12_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_10_14
#define init_2_12_10_14
void FASTCALL call_2_12_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_10_15
#define init_2_12_10_15
void FASTCALL call_2_12_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_11_0
#define init_2_12_11_0
void FASTCALL call_2_12_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_11_1
#define init_2_12_11_1
void FASTCALL call_2_12_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_11_2
#define init_2_12_11_2
void FASTCALL call_2_12_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_11_3
#define init_2_12_11_3
void FASTCALL call_2_12_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_11_4
#define init_2_12_11_4
void FASTCALL call_2_12_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_11_5
#define init_2_12_11_5
void FASTCALL call_2_12_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_11_6
#define init_2_12_11_6
void FASTCALL call_2_12_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_11_7
#define init_2_12_11_7
void FASTCALL call_2_12_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_11_8
#define init_2_12_11_8
void FASTCALL call_2_12_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_11_9
#define init_2_12_11_9
void FASTCALL call_2_12_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_11_10
#define init_2_12_11_10
void FASTCALL call_2_12_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_11_11
#define init_2_12_11_11
void FASTCALL call_2_12_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_11_12
#define init_2_12_11_12
void FASTCALL call_2_12_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_11_13
#define init_2_12_11_13
void FASTCALL call_2_12_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_11_14
#define init_2_12_11_14
void FASTCALL call_2_12_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_11_15
#define init_2_12_11_15
void FASTCALL call_2_12_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_12_0
#define init_2_12_12_0
void FASTCALL call_2_12_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_12_1
#define init_2_12_12_1
void FASTCALL call_2_12_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_12_2
#define init_2_12_12_2
void FASTCALL call_2_12_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_12_3
#define init_2_12_12_3
void FASTCALL call_2_12_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_12_4
#define init_2_12_12_4
void FASTCALL call_2_12_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_12_5
#define init_2_12_12_5
void FASTCALL call_2_12_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_12_6
#define init_2_12_12_6
void FASTCALL call_2_12_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_12_7
#define init_2_12_12_7
void FASTCALL call_2_12_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_12_8
#define init_2_12_12_8
void FASTCALL call_2_12_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_12_9
#define init_2_12_12_9
void FASTCALL call_2_12_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_12_10
#define init_2_12_12_10
void FASTCALL call_2_12_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_12_11
#define init_2_12_12_11
void FASTCALL call_2_12_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_12_12
#define init_2_12_12_12
void FASTCALL call_2_12_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_12_13
#define init_2_12_12_13
void FASTCALL call_2_12_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_12_14
#define init_2_12_12_14
void FASTCALL call_2_12_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_12_15
#define init_2_12_12_15
void FASTCALL call_2_12_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_13_0
#define init_2_12_13_0
void FASTCALL call_2_12_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_13_1
#define init_2_12_13_1
void FASTCALL call_2_12_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_13_2
#define init_2_12_13_2
void FASTCALL call_2_12_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_13_3
#define init_2_12_13_3
void FASTCALL call_2_12_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_13_4
#define init_2_12_13_4
void FASTCALL call_2_12_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_13_5
#define init_2_12_13_5
void FASTCALL call_2_12_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_13_6
#define init_2_12_13_6
void FASTCALL call_2_12_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_13_7
#define init_2_12_13_7
void FASTCALL call_2_12_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_13_8
#define init_2_12_13_8
void FASTCALL call_2_12_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_13_9
#define init_2_12_13_9
void FASTCALL call_2_12_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_13_10
#define init_2_12_13_10
void FASTCALL call_2_12_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_13_11
#define init_2_12_13_11
void FASTCALL call_2_12_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_13_12
#define init_2_12_13_12
void FASTCALL call_2_12_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_13_13
#define init_2_12_13_13
void FASTCALL call_2_12_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_13_14
#define init_2_12_13_14
void FASTCALL call_2_12_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_13_15
#define init_2_12_13_15
void FASTCALL call_2_12_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_14_0
#define init_2_12_14_0
void FASTCALL call_2_12_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_14_1
#define init_2_12_14_1
void FASTCALL call_2_12_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_14_2
#define init_2_12_14_2
void FASTCALL call_2_12_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_14_3
#define init_2_12_14_3
void FASTCALL call_2_12_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_14_4
#define init_2_12_14_4
void FASTCALL call_2_12_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_14_5
#define init_2_12_14_5
void FASTCALL call_2_12_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_14_6
#define init_2_12_14_6
void FASTCALL call_2_12_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_14_7
#define init_2_12_14_7
void FASTCALL call_2_12_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_14_8
#define init_2_12_14_8
void FASTCALL call_2_12_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_14_9
#define init_2_12_14_9
void FASTCALL call_2_12_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_14_10
#define init_2_12_14_10
void FASTCALL call_2_12_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_14_11
#define init_2_12_14_11
void FASTCALL call_2_12_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_14_12
#define init_2_12_14_12
void FASTCALL call_2_12_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_14_13
#define init_2_12_14_13
void FASTCALL call_2_12_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_14_14
#define init_2_12_14_14
void FASTCALL call_2_12_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_14_15
#define init_2_12_14_15
void FASTCALL call_2_12_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_15_0
#define init_2_12_15_0
void FASTCALL call_2_12_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_15_1
#define init_2_12_15_1
void FASTCALL call_2_12_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_15_2
#define init_2_12_15_2
void FASTCALL call_2_12_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_15_3
#define init_2_12_15_3
void FASTCALL call_2_12_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_15_4
#define init_2_12_15_4
void FASTCALL call_2_12_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_15_5
#define init_2_12_15_5
void FASTCALL call_2_12_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_15_6
#define init_2_12_15_6
void FASTCALL call_2_12_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_15_7
#define init_2_12_15_7
void FASTCALL call_2_12_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_15_8
#define init_2_12_15_8
void FASTCALL call_2_12_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_15_9
#define init_2_12_15_9
void FASTCALL call_2_12_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_15_10
#define init_2_12_15_10
void FASTCALL call_2_12_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_15_11
#define init_2_12_15_11
void FASTCALL call_2_12_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_15_12
#define init_2_12_15_12
void FASTCALL call_2_12_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_15_13
#define init_2_12_15_13
void FASTCALL call_2_12_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_15_14
#define init_2_12_15_14
void FASTCALL call_2_12_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_12_15_15
#define init_2_12_15_15
void FASTCALL call_2_12_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_0_0
#define init_2_13_0_0
void FASTCALL call_2_13_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_0_1
#define init_2_13_0_1
void FASTCALL call_2_13_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_0_2
#define init_2_13_0_2
void FASTCALL call_2_13_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_0_3
#define init_2_13_0_3
void FASTCALL call_2_13_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_0_4
#define init_2_13_0_4
void FASTCALL call_2_13_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_0_5
#define init_2_13_0_5
void FASTCALL call_2_13_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_0_6
#define init_2_13_0_6
void FASTCALL call_2_13_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_0_7
#define init_2_13_0_7
void FASTCALL call_2_13_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_0_8
#define init_2_13_0_8
void FASTCALL call_2_13_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_0_9
#define init_2_13_0_9
void FASTCALL call_2_13_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_0_10
#define init_2_13_0_10
void FASTCALL call_2_13_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_0_11
#define init_2_13_0_11
void FASTCALL call_2_13_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_0_12
#define init_2_13_0_12
void FASTCALL call_2_13_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_0_13
#define init_2_13_0_13
void FASTCALL call_2_13_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_0_14
#define init_2_13_0_14
void FASTCALL call_2_13_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_0_15
#define init_2_13_0_15
void FASTCALL call_2_13_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_1_0
#define init_2_13_1_0
void FASTCALL call_2_13_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_1_1
#define init_2_13_1_1
void FASTCALL call_2_13_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_1_2
#define init_2_13_1_2
void FASTCALL call_2_13_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_1_3
#define init_2_13_1_3
void FASTCALL call_2_13_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_1_4
#define init_2_13_1_4
void FASTCALL call_2_13_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_1_5
#define init_2_13_1_5
void FASTCALL call_2_13_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_1_6
#define init_2_13_1_6
void FASTCALL call_2_13_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_1_7
#define init_2_13_1_7
void FASTCALL call_2_13_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_1_8
#define init_2_13_1_8
void FASTCALL call_2_13_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_1_9
#define init_2_13_1_9
void FASTCALL call_2_13_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_1_10
#define init_2_13_1_10
void FASTCALL call_2_13_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_1_11
#define init_2_13_1_11
void FASTCALL call_2_13_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_1_12
#define init_2_13_1_12
void FASTCALL call_2_13_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_1_13
#define init_2_13_1_13
void FASTCALL call_2_13_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_1_14
#define init_2_13_1_14
void FASTCALL call_2_13_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_1_15
#define init_2_13_1_15
void FASTCALL call_2_13_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_2_0
#define init_2_13_2_0
void FASTCALL call_2_13_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_2_1
#define init_2_13_2_1
void FASTCALL call_2_13_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_2_2
#define init_2_13_2_2
void FASTCALL call_2_13_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_2_3
#define init_2_13_2_3
void FASTCALL call_2_13_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_2_4
#define init_2_13_2_4
void FASTCALL call_2_13_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_2_5
#define init_2_13_2_5
void FASTCALL call_2_13_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_2_6
#define init_2_13_2_6
void FASTCALL call_2_13_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_2_7
#define init_2_13_2_7
void FASTCALL call_2_13_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_2_8
#define init_2_13_2_8
void FASTCALL call_2_13_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_2_9
#define init_2_13_2_9
void FASTCALL call_2_13_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_2_10
#define init_2_13_2_10
void FASTCALL call_2_13_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_2_11
#define init_2_13_2_11
void FASTCALL call_2_13_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_2_12
#define init_2_13_2_12
void FASTCALL call_2_13_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_2_13
#define init_2_13_2_13
void FASTCALL call_2_13_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_2_14
#define init_2_13_2_14
void FASTCALL call_2_13_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_2_15
#define init_2_13_2_15
void FASTCALL call_2_13_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_3_0
#define init_2_13_3_0
void FASTCALL call_2_13_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_3_1
#define init_2_13_3_1
void FASTCALL call_2_13_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_3_2
#define init_2_13_3_2
void FASTCALL call_2_13_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_3_3
#define init_2_13_3_3
void FASTCALL call_2_13_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_3_4
#define init_2_13_3_4
void FASTCALL call_2_13_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_3_5
#define init_2_13_3_5
void FASTCALL call_2_13_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_3_6
#define init_2_13_3_6
void FASTCALL call_2_13_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_3_7
#define init_2_13_3_7
void FASTCALL call_2_13_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_3_8
#define init_2_13_3_8
void FASTCALL call_2_13_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_3_9
#define init_2_13_3_9
void FASTCALL call_2_13_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_3_10
#define init_2_13_3_10
void FASTCALL call_2_13_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_3_11
#define init_2_13_3_11
void FASTCALL call_2_13_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_3_12
#define init_2_13_3_12
void FASTCALL call_2_13_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_3_13
#define init_2_13_3_13
void FASTCALL call_2_13_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_3_14
#define init_2_13_3_14
void FASTCALL call_2_13_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_3_15
#define init_2_13_3_15
void FASTCALL call_2_13_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_4_0
#define init_2_13_4_0
void FASTCALL call_2_13_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_4_1
#define init_2_13_4_1
void FASTCALL call_2_13_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_4_2
#define init_2_13_4_2
void FASTCALL call_2_13_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_4_3
#define init_2_13_4_3
void FASTCALL call_2_13_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_4_4
#define init_2_13_4_4
void FASTCALL call_2_13_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_4_5
#define init_2_13_4_5
void FASTCALL call_2_13_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_4_6
#define init_2_13_4_6
void FASTCALL call_2_13_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_4_7
#define init_2_13_4_7
void FASTCALL call_2_13_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_4_8
#define init_2_13_4_8
void FASTCALL call_2_13_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_4_9
#define init_2_13_4_9
void FASTCALL call_2_13_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_4_10
#define init_2_13_4_10
void FASTCALL call_2_13_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_4_11
#define init_2_13_4_11
void FASTCALL call_2_13_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_4_12
#define init_2_13_4_12
void FASTCALL call_2_13_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_4_13
#define init_2_13_4_13
void FASTCALL call_2_13_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_4_14
#define init_2_13_4_14
void FASTCALL call_2_13_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_4_15
#define init_2_13_4_15
void FASTCALL call_2_13_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_5_0
#define init_2_13_5_0
void FASTCALL call_2_13_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_5_1
#define init_2_13_5_1
void FASTCALL call_2_13_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_5_2
#define init_2_13_5_2
void FASTCALL call_2_13_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_5_3
#define init_2_13_5_3
void FASTCALL call_2_13_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_5_4
#define init_2_13_5_4
void FASTCALL call_2_13_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_5_5
#define init_2_13_5_5
void FASTCALL call_2_13_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_5_6
#define init_2_13_5_6
void FASTCALL call_2_13_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_5_7
#define init_2_13_5_7
void FASTCALL call_2_13_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_5_8
#define init_2_13_5_8
void FASTCALL call_2_13_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_5_9
#define init_2_13_5_9
void FASTCALL call_2_13_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_5_10
#define init_2_13_5_10
void FASTCALL call_2_13_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_5_11
#define init_2_13_5_11
void FASTCALL call_2_13_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_5_12
#define init_2_13_5_12
void FASTCALL call_2_13_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_5_13
#define init_2_13_5_13
void FASTCALL call_2_13_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_5_14
#define init_2_13_5_14
void FASTCALL call_2_13_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_5_15
#define init_2_13_5_15
void FASTCALL call_2_13_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_6_0
#define init_2_13_6_0
void FASTCALL call_2_13_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_6_1
#define init_2_13_6_1
void FASTCALL call_2_13_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_6_2
#define init_2_13_6_2
void FASTCALL call_2_13_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_6_3
#define init_2_13_6_3
void FASTCALL call_2_13_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_6_4
#define init_2_13_6_4
void FASTCALL call_2_13_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_6_5
#define init_2_13_6_5
void FASTCALL call_2_13_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_6_6
#define init_2_13_6_6
void FASTCALL call_2_13_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_6_7
#define init_2_13_6_7
void FASTCALL call_2_13_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_6_8
#define init_2_13_6_8
void FASTCALL call_2_13_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_6_9
#define init_2_13_6_9
void FASTCALL call_2_13_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_6_10
#define init_2_13_6_10
void FASTCALL call_2_13_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_6_11
#define init_2_13_6_11
void FASTCALL call_2_13_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_6_12
#define init_2_13_6_12
void FASTCALL call_2_13_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_6_13
#define init_2_13_6_13
void FASTCALL call_2_13_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_6_14
#define init_2_13_6_14
void FASTCALL call_2_13_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_6_15
#define init_2_13_6_15
void FASTCALL call_2_13_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_7_0
#define init_2_13_7_0
void FASTCALL call_2_13_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_7_1
#define init_2_13_7_1
void FASTCALL call_2_13_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_7_2
#define init_2_13_7_2
void FASTCALL call_2_13_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_7_3
#define init_2_13_7_3
void FASTCALL call_2_13_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_7_4
#define init_2_13_7_4
void FASTCALL call_2_13_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_7_5
#define init_2_13_7_5
void FASTCALL call_2_13_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_7_6
#define init_2_13_7_6
void FASTCALL call_2_13_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_7_7
#define init_2_13_7_7
void FASTCALL call_2_13_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_7_8
#define init_2_13_7_8
void FASTCALL call_2_13_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_7_9
#define init_2_13_7_9
void FASTCALL call_2_13_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_7_10
#define init_2_13_7_10
void FASTCALL call_2_13_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_7_11
#define init_2_13_7_11
void FASTCALL call_2_13_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_7_12
#define init_2_13_7_12
void FASTCALL call_2_13_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_7_13
#define init_2_13_7_13
void FASTCALL call_2_13_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_7_14
#define init_2_13_7_14
void FASTCALL call_2_13_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_7_15
#define init_2_13_7_15
void FASTCALL call_2_13_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_8_0
#define init_2_13_8_0
void FASTCALL call_2_13_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_8_1
#define init_2_13_8_1
void FASTCALL call_2_13_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_8_2
#define init_2_13_8_2
void FASTCALL call_2_13_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_8_3
#define init_2_13_8_3
void FASTCALL call_2_13_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_8_4
#define init_2_13_8_4
void FASTCALL call_2_13_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_8_5
#define init_2_13_8_5
void FASTCALL call_2_13_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_8_6
#define init_2_13_8_6
void FASTCALL call_2_13_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_8_7
#define init_2_13_8_7
void FASTCALL call_2_13_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_8_8
#define init_2_13_8_8
void FASTCALL call_2_13_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_8_9
#define init_2_13_8_9
void FASTCALL call_2_13_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_8_10
#define init_2_13_8_10
void FASTCALL call_2_13_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_8_11
#define init_2_13_8_11
void FASTCALL call_2_13_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_8_12
#define init_2_13_8_12
void FASTCALL call_2_13_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_8_13
#define init_2_13_8_13
void FASTCALL call_2_13_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_8_14
#define init_2_13_8_14
void FASTCALL call_2_13_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_8_15
#define init_2_13_8_15
void FASTCALL call_2_13_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_9_0
#define init_2_13_9_0
void FASTCALL call_2_13_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_9_1
#define init_2_13_9_1
void FASTCALL call_2_13_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_9_2
#define init_2_13_9_2
void FASTCALL call_2_13_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_9_3
#define init_2_13_9_3
void FASTCALL call_2_13_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_9_4
#define init_2_13_9_4
void FASTCALL call_2_13_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_9_5
#define init_2_13_9_5
void FASTCALL call_2_13_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_9_6
#define init_2_13_9_6
void FASTCALL call_2_13_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_9_7
#define init_2_13_9_7
void FASTCALL call_2_13_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_9_8
#define init_2_13_9_8
void FASTCALL call_2_13_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_9_9
#define init_2_13_9_9
void FASTCALL call_2_13_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_9_10
#define init_2_13_9_10
void FASTCALL call_2_13_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_9_11
#define init_2_13_9_11
void FASTCALL call_2_13_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_9_12
#define init_2_13_9_12
void FASTCALL call_2_13_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_9_13
#define init_2_13_9_13
void FASTCALL call_2_13_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_9_14
#define init_2_13_9_14
void FASTCALL call_2_13_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_9_15
#define init_2_13_9_15
void FASTCALL call_2_13_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_10_0
#define init_2_13_10_0
void FASTCALL call_2_13_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_10_1
#define init_2_13_10_1
void FASTCALL call_2_13_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_10_2
#define init_2_13_10_2
void FASTCALL call_2_13_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_10_3
#define init_2_13_10_3
void FASTCALL call_2_13_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_10_4
#define init_2_13_10_4
void FASTCALL call_2_13_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_10_5
#define init_2_13_10_5
void FASTCALL call_2_13_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_10_6
#define init_2_13_10_6
void FASTCALL call_2_13_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_10_7
#define init_2_13_10_7
void FASTCALL call_2_13_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_10_8
#define init_2_13_10_8
void FASTCALL call_2_13_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_10_9
#define init_2_13_10_9
void FASTCALL call_2_13_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_10_10
#define init_2_13_10_10
void FASTCALL call_2_13_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_10_11
#define init_2_13_10_11
void FASTCALL call_2_13_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_10_12
#define init_2_13_10_12
void FASTCALL call_2_13_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_10_13
#define init_2_13_10_13
void FASTCALL call_2_13_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_10_14
#define init_2_13_10_14
void FASTCALL call_2_13_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_10_15
#define init_2_13_10_15
void FASTCALL call_2_13_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_11_0
#define init_2_13_11_0
void FASTCALL call_2_13_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_11_1
#define init_2_13_11_1
void FASTCALL call_2_13_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_11_2
#define init_2_13_11_2
void FASTCALL call_2_13_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_11_3
#define init_2_13_11_3
void FASTCALL call_2_13_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_11_4
#define init_2_13_11_4
void FASTCALL call_2_13_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_11_5
#define init_2_13_11_5
void FASTCALL call_2_13_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_11_6
#define init_2_13_11_6
void FASTCALL call_2_13_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_11_7
#define init_2_13_11_7
void FASTCALL call_2_13_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_11_8
#define init_2_13_11_8
void FASTCALL call_2_13_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_11_9
#define init_2_13_11_9
void FASTCALL call_2_13_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_11_10
#define init_2_13_11_10
void FASTCALL call_2_13_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_11_11
#define init_2_13_11_11
void FASTCALL call_2_13_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_11_12
#define init_2_13_11_12
void FASTCALL call_2_13_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_11_13
#define init_2_13_11_13
void FASTCALL call_2_13_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_11_14
#define init_2_13_11_14
void FASTCALL call_2_13_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_11_15
#define init_2_13_11_15
void FASTCALL call_2_13_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_12_0
#define init_2_13_12_0
void FASTCALL call_2_13_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_12_1
#define init_2_13_12_1
void FASTCALL call_2_13_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_12_2
#define init_2_13_12_2
void FASTCALL call_2_13_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_12_3
#define init_2_13_12_3
void FASTCALL call_2_13_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_12_4
#define init_2_13_12_4
void FASTCALL call_2_13_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_12_5
#define init_2_13_12_5
void FASTCALL call_2_13_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_12_6
#define init_2_13_12_6
void FASTCALL call_2_13_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_12_7
#define init_2_13_12_7
void FASTCALL call_2_13_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_12_8
#define init_2_13_12_8
void FASTCALL call_2_13_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_12_9
#define init_2_13_12_9
void FASTCALL call_2_13_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_12_10
#define init_2_13_12_10
void FASTCALL call_2_13_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_12_11
#define init_2_13_12_11
void FASTCALL call_2_13_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_12_12
#define init_2_13_12_12
void FASTCALL call_2_13_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_12_13
#define init_2_13_12_13
void FASTCALL call_2_13_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_12_14
#define init_2_13_12_14
void FASTCALL call_2_13_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_12_15
#define init_2_13_12_15
void FASTCALL call_2_13_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_13_0
#define init_2_13_13_0
void FASTCALL call_2_13_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_13_1
#define init_2_13_13_1
void FASTCALL call_2_13_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_13_2
#define init_2_13_13_2
void FASTCALL call_2_13_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_13_3
#define init_2_13_13_3
void FASTCALL call_2_13_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_13_4
#define init_2_13_13_4
void FASTCALL call_2_13_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_13_5
#define init_2_13_13_5
void FASTCALL call_2_13_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_13_6
#define init_2_13_13_6
void FASTCALL call_2_13_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_13_7
#define init_2_13_13_7
void FASTCALL call_2_13_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_13_8
#define init_2_13_13_8
void FASTCALL call_2_13_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_13_9
#define init_2_13_13_9
void FASTCALL call_2_13_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_13_10
#define init_2_13_13_10
void FASTCALL call_2_13_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_13_11
#define init_2_13_13_11
void FASTCALL call_2_13_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_13_12
#define init_2_13_13_12
void FASTCALL call_2_13_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_13_13
#define init_2_13_13_13
void FASTCALL call_2_13_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_13_14
#define init_2_13_13_14
void FASTCALL call_2_13_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_13_15
#define init_2_13_13_15
void FASTCALL call_2_13_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_14_0
#define init_2_13_14_0
void FASTCALL call_2_13_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_14_1
#define init_2_13_14_1
void FASTCALL call_2_13_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_14_2
#define init_2_13_14_2
void FASTCALL call_2_13_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_14_3
#define init_2_13_14_3
void FASTCALL call_2_13_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_14_4
#define init_2_13_14_4
void FASTCALL call_2_13_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_14_5
#define init_2_13_14_5
void FASTCALL call_2_13_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_14_6
#define init_2_13_14_6
void FASTCALL call_2_13_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_14_7
#define init_2_13_14_7
void FASTCALL call_2_13_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_14_8
#define init_2_13_14_8
void FASTCALL call_2_13_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_14_9
#define init_2_13_14_9
void FASTCALL call_2_13_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_14_10
#define init_2_13_14_10
void FASTCALL call_2_13_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_14_11
#define init_2_13_14_11
void FASTCALL call_2_13_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_14_12
#define init_2_13_14_12
void FASTCALL call_2_13_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_14_13
#define init_2_13_14_13
void FASTCALL call_2_13_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_14_14
#define init_2_13_14_14
void FASTCALL call_2_13_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_14_15
#define init_2_13_14_15
void FASTCALL call_2_13_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_15_0
#define init_2_13_15_0
void FASTCALL call_2_13_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_15_1
#define init_2_13_15_1
void FASTCALL call_2_13_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_15_2
#define init_2_13_15_2
void FASTCALL call_2_13_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_15_3
#define init_2_13_15_3
void FASTCALL call_2_13_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_15_4
#define init_2_13_15_4
void FASTCALL call_2_13_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_15_5
#define init_2_13_15_5
void FASTCALL call_2_13_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_15_6
#define init_2_13_15_6
void FASTCALL call_2_13_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_15_7
#define init_2_13_15_7
void FASTCALL call_2_13_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_15_8
#define init_2_13_15_8
void FASTCALL call_2_13_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_15_9
#define init_2_13_15_9
void FASTCALL call_2_13_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_15_10
#define init_2_13_15_10
void FASTCALL call_2_13_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_15_11
#define init_2_13_15_11
void FASTCALL call_2_13_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_15_12
#define init_2_13_15_12
void FASTCALL call_2_13_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_15_13
#define init_2_13_15_13
void FASTCALL call_2_13_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_15_14
#define init_2_13_15_14
void FASTCALL call_2_13_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_13_15_15
#define init_2_13_15_15
void FASTCALL call_2_13_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_0_0
#define init_2_14_0_0
void FASTCALL call_2_14_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_0_1
#define init_2_14_0_1
void FASTCALL call_2_14_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_0_2
#define init_2_14_0_2
void FASTCALL call_2_14_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_0_3
#define init_2_14_0_3
void FASTCALL call_2_14_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_0_4
#define init_2_14_0_4
void FASTCALL call_2_14_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_0_5
#define init_2_14_0_5
void FASTCALL call_2_14_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_0_6
#define init_2_14_0_6
void FASTCALL call_2_14_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_0_7
#define init_2_14_0_7
void FASTCALL call_2_14_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_0_8
#define init_2_14_0_8
void FASTCALL call_2_14_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_0_9
#define init_2_14_0_9
void FASTCALL call_2_14_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_0_10
#define init_2_14_0_10
void FASTCALL call_2_14_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_0_11
#define init_2_14_0_11
void FASTCALL call_2_14_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_0_12
#define init_2_14_0_12
void FASTCALL call_2_14_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_0_13
#define init_2_14_0_13
void FASTCALL call_2_14_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_0_14
#define init_2_14_0_14
void FASTCALL call_2_14_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_0_15
#define init_2_14_0_15
void FASTCALL call_2_14_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_1_0
#define init_2_14_1_0
void FASTCALL call_2_14_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_1_1
#define init_2_14_1_1
void FASTCALL call_2_14_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_1_2
#define init_2_14_1_2
void FASTCALL call_2_14_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_1_3
#define init_2_14_1_3
void FASTCALL call_2_14_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_1_4
#define init_2_14_1_4
void FASTCALL call_2_14_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_1_5
#define init_2_14_1_5
void FASTCALL call_2_14_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_1_6
#define init_2_14_1_6
void FASTCALL call_2_14_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_1_7
#define init_2_14_1_7
void FASTCALL call_2_14_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_1_8
#define init_2_14_1_8
void FASTCALL call_2_14_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_1_9
#define init_2_14_1_9
void FASTCALL call_2_14_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_1_10
#define init_2_14_1_10
void FASTCALL call_2_14_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_1_11
#define init_2_14_1_11
void FASTCALL call_2_14_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_1_12
#define init_2_14_1_12
void FASTCALL call_2_14_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_1_13
#define init_2_14_1_13
void FASTCALL call_2_14_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_1_14
#define init_2_14_1_14
void FASTCALL call_2_14_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_1_15
#define init_2_14_1_15
void FASTCALL call_2_14_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_2_0
#define init_2_14_2_0
void FASTCALL call_2_14_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_2_1
#define init_2_14_2_1
void FASTCALL call_2_14_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_2_2
#define init_2_14_2_2
void FASTCALL call_2_14_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_2_3
#define init_2_14_2_3
void FASTCALL call_2_14_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_2_4
#define init_2_14_2_4
void FASTCALL call_2_14_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_2_5
#define init_2_14_2_5
void FASTCALL call_2_14_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_2_6
#define init_2_14_2_6
void FASTCALL call_2_14_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_2_7
#define init_2_14_2_7
void FASTCALL call_2_14_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_2_8
#define init_2_14_2_8
void FASTCALL call_2_14_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_2_9
#define init_2_14_2_9
void FASTCALL call_2_14_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_2_10
#define init_2_14_2_10
void FASTCALL call_2_14_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_2_11
#define init_2_14_2_11
void FASTCALL call_2_14_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_2_12
#define init_2_14_2_12
void FASTCALL call_2_14_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_2_13
#define init_2_14_2_13
void FASTCALL call_2_14_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_2_14
#define init_2_14_2_14
void FASTCALL call_2_14_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_2_15
#define init_2_14_2_15
void FASTCALL call_2_14_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_3_0
#define init_2_14_3_0
void FASTCALL call_2_14_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_3_1
#define init_2_14_3_1
void FASTCALL call_2_14_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_3_2
#define init_2_14_3_2
void FASTCALL call_2_14_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_3_3
#define init_2_14_3_3
void FASTCALL call_2_14_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_3_4
#define init_2_14_3_4
void FASTCALL call_2_14_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_3_5
#define init_2_14_3_5
void FASTCALL call_2_14_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_3_6
#define init_2_14_3_6
void FASTCALL call_2_14_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_3_7
#define init_2_14_3_7
void FASTCALL call_2_14_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_3_8
#define init_2_14_3_8
void FASTCALL call_2_14_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_3_9
#define init_2_14_3_9
void FASTCALL call_2_14_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_3_10
#define init_2_14_3_10
void FASTCALL call_2_14_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_3_11
#define init_2_14_3_11
void FASTCALL call_2_14_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_3_12
#define init_2_14_3_12
void FASTCALL call_2_14_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_3_13
#define init_2_14_3_13
void FASTCALL call_2_14_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_3_14
#define init_2_14_3_14
void FASTCALL call_2_14_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_3_15
#define init_2_14_3_15
void FASTCALL call_2_14_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_4_0
#define init_2_14_4_0
void FASTCALL call_2_14_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_4_1
#define init_2_14_4_1
void FASTCALL call_2_14_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_4_2
#define init_2_14_4_2
void FASTCALL call_2_14_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_4_3
#define init_2_14_4_3
void FASTCALL call_2_14_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_4_4
#define init_2_14_4_4
void FASTCALL call_2_14_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_4_5
#define init_2_14_4_5
void FASTCALL call_2_14_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_4_6
#define init_2_14_4_6
void FASTCALL call_2_14_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_4_7
#define init_2_14_4_7
void FASTCALL call_2_14_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_4_8
#define init_2_14_4_8
void FASTCALL call_2_14_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_4_9
#define init_2_14_4_9
void FASTCALL call_2_14_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_4_10
#define init_2_14_4_10
void FASTCALL call_2_14_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_4_11
#define init_2_14_4_11
void FASTCALL call_2_14_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_4_12
#define init_2_14_4_12
void FASTCALL call_2_14_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_4_13
#define init_2_14_4_13
void FASTCALL call_2_14_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_4_14
#define init_2_14_4_14
void FASTCALL call_2_14_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_4_15
#define init_2_14_4_15
void FASTCALL call_2_14_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_5_0
#define init_2_14_5_0
void FASTCALL call_2_14_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_5_1
#define init_2_14_5_1
void FASTCALL call_2_14_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_5_2
#define init_2_14_5_2
void FASTCALL call_2_14_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_5_3
#define init_2_14_5_3
void FASTCALL call_2_14_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_5_4
#define init_2_14_5_4
void FASTCALL call_2_14_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_5_5
#define init_2_14_5_5
void FASTCALL call_2_14_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_5_6
#define init_2_14_5_6
void FASTCALL call_2_14_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_5_7
#define init_2_14_5_7
void FASTCALL call_2_14_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_5_8
#define init_2_14_5_8
void FASTCALL call_2_14_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_5_9
#define init_2_14_5_9
void FASTCALL call_2_14_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_5_10
#define init_2_14_5_10
void FASTCALL call_2_14_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_5_11
#define init_2_14_5_11
void FASTCALL call_2_14_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_5_12
#define init_2_14_5_12
void FASTCALL call_2_14_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_5_13
#define init_2_14_5_13
void FASTCALL call_2_14_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_5_14
#define init_2_14_5_14
void FASTCALL call_2_14_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_5_15
#define init_2_14_5_15
void FASTCALL call_2_14_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_6_0
#define init_2_14_6_0
void FASTCALL call_2_14_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_6_1
#define init_2_14_6_1
void FASTCALL call_2_14_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_6_2
#define init_2_14_6_2
void FASTCALL call_2_14_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_6_3
#define init_2_14_6_3
void FASTCALL call_2_14_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_6_4
#define init_2_14_6_4
void FASTCALL call_2_14_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_6_5
#define init_2_14_6_5
void FASTCALL call_2_14_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_6_6
#define init_2_14_6_6
void FASTCALL call_2_14_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_6_7
#define init_2_14_6_7
void FASTCALL call_2_14_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_6_8
#define init_2_14_6_8
void FASTCALL call_2_14_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_6_9
#define init_2_14_6_9
void FASTCALL call_2_14_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_6_10
#define init_2_14_6_10
void FASTCALL call_2_14_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_6_11
#define init_2_14_6_11
void FASTCALL call_2_14_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_6_12
#define init_2_14_6_12
void FASTCALL call_2_14_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_6_13
#define init_2_14_6_13
void FASTCALL call_2_14_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_6_14
#define init_2_14_6_14
void FASTCALL call_2_14_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_6_15
#define init_2_14_6_15
void FASTCALL call_2_14_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_7_0
#define init_2_14_7_0
void FASTCALL call_2_14_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_7_1
#define init_2_14_7_1
void FASTCALL call_2_14_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_7_2
#define init_2_14_7_2
void FASTCALL call_2_14_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_7_3
#define init_2_14_7_3
void FASTCALL call_2_14_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_7_4
#define init_2_14_7_4
void FASTCALL call_2_14_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_7_5
#define init_2_14_7_5
void FASTCALL call_2_14_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_7_6
#define init_2_14_7_6
void FASTCALL call_2_14_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_7_7
#define init_2_14_7_7
void FASTCALL call_2_14_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_7_8
#define init_2_14_7_8
void FASTCALL call_2_14_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_7_9
#define init_2_14_7_9
void FASTCALL call_2_14_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_7_10
#define init_2_14_7_10
void FASTCALL call_2_14_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_7_11
#define init_2_14_7_11
void FASTCALL call_2_14_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_7_12
#define init_2_14_7_12
void FASTCALL call_2_14_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_7_13
#define init_2_14_7_13
void FASTCALL call_2_14_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_7_14
#define init_2_14_7_14
void FASTCALL call_2_14_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_7_15
#define init_2_14_7_15
void FASTCALL call_2_14_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_8_0
#define init_2_14_8_0
void FASTCALL call_2_14_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_8_1
#define init_2_14_8_1
void FASTCALL call_2_14_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_8_2
#define init_2_14_8_2
void FASTCALL call_2_14_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_8_3
#define init_2_14_8_3
void FASTCALL call_2_14_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_8_4
#define init_2_14_8_4
void FASTCALL call_2_14_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_8_5
#define init_2_14_8_5
void FASTCALL call_2_14_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_8_6
#define init_2_14_8_6
void FASTCALL call_2_14_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_8_7
#define init_2_14_8_7
void FASTCALL call_2_14_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_8_8
#define init_2_14_8_8
void FASTCALL call_2_14_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_8_9
#define init_2_14_8_9
void FASTCALL call_2_14_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_8_10
#define init_2_14_8_10
void FASTCALL call_2_14_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_8_11
#define init_2_14_8_11
void FASTCALL call_2_14_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_8_12
#define init_2_14_8_12
void FASTCALL call_2_14_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_8_13
#define init_2_14_8_13
void FASTCALL call_2_14_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_8_14
#define init_2_14_8_14
void FASTCALL call_2_14_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_8_15
#define init_2_14_8_15
void FASTCALL call_2_14_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_9_0
#define init_2_14_9_0
void FASTCALL call_2_14_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_9_1
#define init_2_14_9_1
void FASTCALL call_2_14_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_9_2
#define init_2_14_9_2
void FASTCALL call_2_14_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_9_3
#define init_2_14_9_3
void FASTCALL call_2_14_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_9_4
#define init_2_14_9_4
void FASTCALL call_2_14_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_9_5
#define init_2_14_9_5
void FASTCALL call_2_14_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_9_6
#define init_2_14_9_6
void FASTCALL call_2_14_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_9_7
#define init_2_14_9_7
void FASTCALL call_2_14_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_9_8
#define init_2_14_9_8
void FASTCALL call_2_14_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_9_9
#define init_2_14_9_9
void FASTCALL call_2_14_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_9_10
#define init_2_14_9_10
void FASTCALL call_2_14_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_9_11
#define init_2_14_9_11
void FASTCALL call_2_14_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_9_12
#define init_2_14_9_12
void FASTCALL call_2_14_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_9_13
#define init_2_14_9_13
void FASTCALL call_2_14_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_9_14
#define init_2_14_9_14
void FASTCALL call_2_14_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_9_15
#define init_2_14_9_15
void FASTCALL call_2_14_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_10_0
#define init_2_14_10_0
void FASTCALL call_2_14_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_10_1
#define init_2_14_10_1
void FASTCALL call_2_14_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_10_2
#define init_2_14_10_2
void FASTCALL call_2_14_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_10_3
#define init_2_14_10_3
void FASTCALL call_2_14_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_10_4
#define init_2_14_10_4
void FASTCALL call_2_14_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_10_5
#define init_2_14_10_5
void FASTCALL call_2_14_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_10_6
#define init_2_14_10_6
void FASTCALL call_2_14_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_10_7
#define init_2_14_10_7
void FASTCALL call_2_14_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_10_8
#define init_2_14_10_8
void FASTCALL call_2_14_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_10_9
#define init_2_14_10_9
void FASTCALL call_2_14_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_10_10
#define init_2_14_10_10
void FASTCALL call_2_14_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_10_11
#define init_2_14_10_11
void FASTCALL call_2_14_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_10_12
#define init_2_14_10_12
void FASTCALL call_2_14_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_10_13
#define init_2_14_10_13
void FASTCALL call_2_14_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_10_14
#define init_2_14_10_14
void FASTCALL call_2_14_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_10_15
#define init_2_14_10_15
void FASTCALL call_2_14_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_11_0
#define init_2_14_11_0
void FASTCALL call_2_14_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_11_1
#define init_2_14_11_1
void FASTCALL call_2_14_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_11_2
#define init_2_14_11_2
void FASTCALL call_2_14_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_11_3
#define init_2_14_11_3
void FASTCALL call_2_14_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_11_4
#define init_2_14_11_4
void FASTCALL call_2_14_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_11_5
#define init_2_14_11_5
void FASTCALL call_2_14_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_11_6
#define init_2_14_11_6
void FASTCALL call_2_14_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_11_7
#define init_2_14_11_7
void FASTCALL call_2_14_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_11_8
#define init_2_14_11_8
void FASTCALL call_2_14_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_11_9
#define init_2_14_11_9
void FASTCALL call_2_14_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_11_10
#define init_2_14_11_10
void FASTCALL call_2_14_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_11_11
#define init_2_14_11_11
void FASTCALL call_2_14_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_11_12
#define init_2_14_11_12
void FASTCALL call_2_14_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_11_13
#define init_2_14_11_13
void FASTCALL call_2_14_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_11_14
#define init_2_14_11_14
void FASTCALL call_2_14_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_11_15
#define init_2_14_11_15
void FASTCALL call_2_14_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_12_0
#define init_2_14_12_0
void FASTCALL call_2_14_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_12_1
#define init_2_14_12_1
void FASTCALL call_2_14_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_12_2
#define init_2_14_12_2
void FASTCALL call_2_14_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_12_3
#define init_2_14_12_3
void FASTCALL call_2_14_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_12_4
#define init_2_14_12_4
void FASTCALL call_2_14_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_12_5
#define init_2_14_12_5
void FASTCALL call_2_14_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_12_6
#define init_2_14_12_6
void FASTCALL call_2_14_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_12_7
#define init_2_14_12_7
void FASTCALL call_2_14_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_12_8
#define init_2_14_12_8
void FASTCALL call_2_14_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_12_9
#define init_2_14_12_9
void FASTCALL call_2_14_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_12_10
#define init_2_14_12_10
void FASTCALL call_2_14_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_12_11
#define init_2_14_12_11
void FASTCALL call_2_14_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_12_12
#define init_2_14_12_12
void FASTCALL call_2_14_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_12_13
#define init_2_14_12_13
void FASTCALL call_2_14_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_12_14
#define init_2_14_12_14
void FASTCALL call_2_14_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_12_15
#define init_2_14_12_15
void FASTCALL call_2_14_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_13_0
#define init_2_14_13_0
void FASTCALL call_2_14_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_13_1
#define init_2_14_13_1
void FASTCALL call_2_14_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_13_2
#define init_2_14_13_2
void FASTCALL call_2_14_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_13_3
#define init_2_14_13_3
void FASTCALL call_2_14_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_13_4
#define init_2_14_13_4
void FASTCALL call_2_14_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_13_5
#define init_2_14_13_5
void FASTCALL call_2_14_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_13_6
#define init_2_14_13_6
void FASTCALL call_2_14_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_13_7
#define init_2_14_13_7
void FASTCALL call_2_14_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_13_8
#define init_2_14_13_8
void FASTCALL call_2_14_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_13_9
#define init_2_14_13_9
void FASTCALL call_2_14_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_13_10
#define init_2_14_13_10
void FASTCALL call_2_14_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_13_11
#define init_2_14_13_11
void FASTCALL call_2_14_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_13_12
#define init_2_14_13_12
void FASTCALL call_2_14_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_13_13
#define init_2_14_13_13
void FASTCALL call_2_14_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_13_14
#define init_2_14_13_14
void FASTCALL call_2_14_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_13_15
#define init_2_14_13_15
void FASTCALL call_2_14_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_14_0
#define init_2_14_14_0
void FASTCALL call_2_14_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_14_1
#define init_2_14_14_1
void FASTCALL call_2_14_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_14_2
#define init_2_14_14_2
void FASTCALL call_2_14_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_14_3
#define init_2_14_14_3
void FASTCALL call_2_14_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_14_4
#define init_2_14_14_4
void FASTCALL call_2_14_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_14_5
#define init_2_14_14_5
void FASTCALL call_2_14_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_14_6
#define init_2_14_14_6
void FASTCALL call_2_14_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_14_7
#define init_2_14_14_7
void FASTCALL call_2_14_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_14_8
#define init_2_14_14_8
void FASTCALL call_2_14_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_14_9
#define init_2_14_14_9
void FASTCALL call_2_14_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_14_10
#define init_2_14_14_10
void FASTCALL call_2_14_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_14_11
#define init_2_14_14_11
void FASTCALL call_2_14_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_14_12
#define init_2_14_14_12
void FASTCALL call_2_14_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_14_13
#define init_2_14_14_13
void FASTCALL call_2_14_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_14_14
#define init_2_14_14_14
void FASTCALL call_2_14_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_14_15
#define init_2_14_14_15
void FASTCALL call_2_14_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_15_0
#define init_2_14_15_0
void FASTCALL call_2_14_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_15_1
#define init_2_14_15_1
void FASTCALL call_2_14_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_15_2
#define init_2_14_15_2
void FASTCALL call_2_14_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_15_3
#define init_2_14_15_3
void FASTCALL call_2_14_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_15_4
#define init_2_14_15_4
void FASTCALL call_2_14_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_15_5
#define init_2_14_15_5
void FASTCALL call_2_14_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_15_6
#define init_2_14_15_6
void FASTCALL call_2_14_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_15_7
#define init_2_14_15_7
void FASTCALL call_2_14_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_15_8
#define init_2_14_15_8
void FASTCALL call_2_14_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_15_9
#define init_2_14_15_9
void FASTCALL call_2_14_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_15_10
#define init_2_14_15_10
void FASTCALL call_2_14_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_15_11
#define init_2_14_15_11
void FASTCALL call_2_14_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_15_12
#define init_2_14_15_12
void FASTCALL call_2_14_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_15_13
#define init_2_14_15_13
void FASTCALL call_2_14_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_15_14
#define init_2_14_15_14
void FASTCALL call_2_14_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_14_15_15
#define init_2_14_15_15
void FASTCALL call_2_14_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_0_0
#define init_2_15_0_0
void FASTCALL call_2_15_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_0_1
#define init_2_15_0_1
void FASTCALL call_2_15_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_0_2
#define init_2_15_0_2
void FASTCALL call_2_15_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_0_3
#define init_2_15_0_3
void FASTCALL call_2_15_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_0_4
#define init_2_15_0_4
void FASTCALL call_2_15_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_0_5
#define init_2_15_0_5
void FASTCALL call_2_15_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_0_6
#define init_2_15_0_6
void FASTCALL call_2_15_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_0_7
#define init_2_15_0_7
void FASTCALL call_2_15_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_0_8
#define init_2_15_0_8
void FASTCALL call_2_15_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_0_9
#define init_2_15_0_9
void FASTCALL call_2_15_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_0_10
#define init_2_15_0_10
void FASTCALL call_2_15_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_0_11
#define init_2_15_0_11
void FASTCALL call_2_15_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_0_12
#define init_2_15_0_12
void FASTCALL call_2_15_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_0_13
#define init_2_15_0_13
void FASTCALL call_2_15_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_0_14
#define init_2_15_0_14
void FASTCALL call_2_15_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_0_15
#define init_2_15_0_15
void FASTCALL call_2_15_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_1_0
#define init_2_15_1_0
void FASTCALL call_2_15_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_1_1
#define init_2_15_1_1
void FASTCALL call_2_15_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_1_2
#define init_2_15_1_2
void FASTCALL call_2_15_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_1_3
#define init_2_15_1_3
void FASTCALL call_2_15_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_1_4
#define init_2_15_1_4
void FASTCALL call_2_15_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_1_5
#define init_2_15_1_5
void FASTCALL call_2_15_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_1_6
#define init_2_15_1_6
void FASTCALL call_2_15_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_1_7
#define init_2_15_1_7
void FASTCALL call_2_15_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_1_8
#define init_2_15_1_8
void FASTCALL call_2_15_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_1_9
#define init_2_15_1_9
void FASTCALL call_2_15_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_1_10
#define init_2_15_1_10
void FASTCALL call_2_15_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_1_11
#define init_2_15_1_11
void FASTCALL call_2_15_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_1_12
#define init_2_15_1_12
void FASTCALL call_2_15_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_1_13
#define init_2_15_1_13
void FASTCALL call_2_15_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_1_14
#define init_2_15_1_14
void FASTCALL call_2_15_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_1_15
#define init_2_15_1_15
void FASTCALL call_2_15_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_2_0
#define init_2_15_2_0
void FASTCALL call_2_15_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_2_1
#define init_2_15_2_1
void FASTCALL call_2_15_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_2_2
#define init_2_15_2_2
void FASTCALL call_2_15_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_2_3
#define init_2_15_2_3
void FASTCALL call_2_15_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_2_4
#define init_2_15_2_4
void FASTCALL call_2_15_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_2_5
#define init_2_15_2_5
void FASTCALL call_2_15_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_2_6
#define init_2_15_2_6
void FASTCALL call_2_15_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_2_7
#define init_2_15_2_7
void FASTCALL call_2_15_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_2_8
#define init_2_15_2_8
void FASTCALL call_2_15_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_2_9
#define init_2_15_2_9
void FASTCALL call_2_15_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_2_10
#define init_2_15_2_10
void FASTCALL call_2_15_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_2_11
#define init_2_15_2_11
void FASTCALL call_2_15_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_2_12
#define init_2_15_2_12
void FASTCALL call_2_15_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_2_13
#define init_2_15_2_13
void FASTCALL call_2_15_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_2_14
#define init_2_15_2_14
void FASTCALL call_2_15_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_2_15
#define init_2_15_2_15
void FASTCALL call_2_15_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_3_0
#define init_2_15_3_0
void FASTCALL call_2_15_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_3_1
#define init_2_15_3_1
void FASTCALL call_2_15_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_3_2
#define init_2_15_3_2
void FASTCALL call_2_15_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_3_3
#define init_2_15_3_3
void FASTCALL call_2_15_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_3_4
#define init_2_15_3_4
void FASTCALL call_2_15_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_3_5
#define init_2_15_3_5
void FASTCALL call_2_15_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_3_6
#define init_2_15_3_6
void FASTCALL call_2_15_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_3_7
#define init_2_15_3_7
void FASTCALL call_2_15_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_3_8
#define init_2_15_3_8
void FASTCALL call_2_15_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_3_9
#define init_2_15_3_9
void FASTCALL call_2_15_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_3_10
#define init_2_15_3_10
void FASTCALL call_2_15_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_3_11
#define init_2_15_3_11
void FASTCALL call_2_15_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_3_12
#define init_2_15_3_12
void FASTCALL call_2_15_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_3_13
#define init_2_15_3_13
void FASTCALL call_2_15_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_3_14
#define init_2_15_3_14
void FASTCALL call_2_15_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_3_15
#define init_2_15_3_15
void FASTCALL call_2_15_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_4_0
#define init_2_15_4_0
void FASTCALL call_2_15_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_4_1
#define init_2_15_4_1
void FASTCALL call_2_15_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_4_2
#define init_2_15_4_2
void FASTCALL call_2_15_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_4_3
#define init_2_15_4_3
void FASTCALL call_2_15_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_4_4
#define init_2_15_4_4
void FASTCALL call_2_15_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_4_5
#define init_2_15_4_5
void FASTCALL call_2_15_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_4_6
#define init_2_15_4_6
void FASTCALL call_2_15_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_4_7
#define init_2_15_4_7
void FASTCALL call_2_15_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_4_8
#define init_2_15_4_8
void FASTCALL call_2_15_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_4_9
#define init_2_15_4_9
void FASTCALL call_2_15_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_4_10
#define init_2_15_4_10
void FASTCALL call_2_15_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_4_11
#define init_2_15_4_11
void FASTCALL call_2_15_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_4_12
#define init_2_15_4_12
void FASTCALL call_2_15_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_4_13
#define init_2_15_4_13
void FASTCALL call_2_15_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_4_14
#define init_2_15_4_14
void FASTCALL call_2_15_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_4_15
#define init_2_15_4_15
void FASTCALL call_2_15_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_5_0
#define init_2_15_5_0
void FASTCALL call_2_15_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_5_1
#define init_2_15_5_1
void FASTCALL call_2_15_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_5_2
#define init_2_15_5_2
void FASTCALL call_2_15_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_5_3
#define init_2_15_5_3
void FASTCALL call_2_15_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_5_4
#define init_2_15_5_4
void FASTCALL call_2_15_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_5_5
#define init_2_15_5_5
void FASTCALL call_2_15_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_5_6
#define init_2_15_5_6
void FASTCALL call_2_15_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_5_7
#define init_2_15_5_7
void FASTCALL call_2_15_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_5_8
#define init_2_15_5_8
void FASTCALL call_2_15_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_5_9
#define init_2_15_5_9
void FASTCALL call_2_15_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_5_10
#define init_2_15_5_10
void FASTCALL call_2_15_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_5_11
#define init_2_15_5_11
void FASTCALL call_2_15_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_5_12
#define init_2_15_5_12
void FASTCALL call_2_15_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_5_13
#define init_2_15_5_13
void FASTCALL call_2_15_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_5_14
#define init_2_15_5_14
void FASTCALL call_2_15_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_5_15
#define init_2_15_5_15
void FASTCALL call_2_15_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_6_0
#define init_2_15_6_0
void FASTCALL call_2_15_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_6_1
#define init_2_15_6_1
void FASTCALL call_2_15_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_6_2
#define init_2_15_6_2
void FASTCALL call_2_15_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_6_3
#define init_2_15_6_3
void FASTCALL call_2_15_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_6_4
#define init_2_15_6_4
void FASTCALL call_2_15_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_6_5
#define init_2_15_6_5
void FASTCALL call_2_15_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_6_6
#define init_2_15_6_6
void FASTCALL call_2_15_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_6_7
#define init_2_15_6_7
void FASTCALL call_2_15_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_6_8
#define init_2_15_6_8
void FASTCALL call_2_15_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_6_9
#define init_2_15_6_9
void FASTCALL call_2_15_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_6_10
#define init_2_15_6_10
void FASTCALL call_2_15_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_6_11
#define init_2_15_6_11
void FASTCALL call_2_15_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_6_12
#define init_2_15_6_12
void FASTCALL call_2_15_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_6_13
#define init_2_15_6_13
void FASTCALL call_2_15_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_6_14
#define init_2_15_6_14
void FASTCALL call_2_15_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_6_15
#define init_2_15_6_15
void FASTCALL call_2_15_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_7_0
#define init_2_15_7_0
void FASTCALL call_2_15_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_7_1
#define init_2_15_7_1
void FASTCALL call_2_15_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_7_2
#define init_2_15_7_2
void FASTCALL call_2_15_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_7_3
#define init_2_15_7_3
void FASTCALL call_2_15_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_7_4
#define init_2_15_7_4
void FASTCALL call_2_15_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_7_5
#define init_2_15_7_5
void FASTCALL call_2_15_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_7_6
#define init_2_15_7_6
void FASTCALL call_2_15_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_7_7
#define init_2_15_7_7
void FASTCALL call_2_15_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_7_8
#define init_2_15_7_8
void FASTCALL call_2_15_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_7_9
#define init_2_15_7_9
void FASTCALL call_2_15_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_7_10
#define init_2_15_7_10
void FASTCALL call_2_15_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_7_11
#define init_2_15_7_11
void FASTCALL call_2_15_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_7_12
#define init_2_15_7_12
void FASTCALL call_2_15_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_7_13
#define init_2_15_7_13
void FASTCALL call_2_15_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_7_14
#define init_2_15_7_14
void FASTCALL call_2_15_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_7_15
#define init_2_15_7_15
void FASTCALL call_2_15_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_8_0
#define init_2_15_8_0
void FASTCALL call_2_15_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_8_1
#define init_2_15_8_1
void FASTCALL call_2_15_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_8_2
#define init_2_15_8_2
void FASTCALL call_2_15_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_8_3
#define init_2_15_8_3
void FASTCALL call_2_15_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_8_4
#define init_2_15_8_4
void FASTCALL call_2_15_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_8_5
#define init_2_15_8_5
void FASTCALL call_2_15_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_8_6
#define init_2_15_8_6
void FASTCALL call_2_15_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_8_7
#define init_2_15_8_7
void FASTCALL call_2_15_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_8_8
#define init_2_15_8_8
void FASTCALL call_2_15_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_8_9
#define init_2_15_8_9
void FASTCALL call_2_15_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_8_10
#define init_2_15_8_10
void FASTCALL call_2_15_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_8_11
#define init_2_15_8_11
void FASTCALL call_2_15_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_8_12
#define init_2_15_8_12
void FASTCALL call_2_15_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_8_13
#define init_2_15_8_13
void FASTCALL call_2_15_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_8_14
#define init_2_15_8_14
void FASTCALL call_2_15_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_8_15
#define init_2_15_8_15
void FASTCALL call_2_15_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_9_0
#define init_2_15_9_0
void FASTCALL call_2_15_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_9_1
#define init_2_15_9_1
void FASTCALL call_2_15_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_9_2
#define init_2_15_9_2
void FASTCALL call_2_15_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_9_3
#define init_2_15_9_3
void FASTCALL call_2_15_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_9_4
#define init_2_15_9_4
void FASTCALL call_2_15_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_9_5
#define init_2_15_9_5
void FASTCALL call_2_15_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_9_6
#define init_2_15_9_6
void FASTCALL call_2_15_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_9_7
#define init_2_15_9_7
void FASTCALL call_2_15_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_9_8
#define init_2_15_9_8
void FASTCALL call_2_15_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_9_9
#define init_2_15_9_9
void FASTCALL call_2_15_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_9_10
#define init_2_15_9_10
void FASTCALL call_2_15_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_9_11
#define init_2_15_9_11
void FASTCALL call_2_15_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_9_12
#define init_2_15_9_12
void FASTCALL call_2_15_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_9_13
#define init_2_15_9_13
void FASTCALL call_2_15_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_9_14
#define init_2_15_9_14
void FASTCALL call_2_15_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_9_15
#define init_2_15_9_15
void FASTCALL call_2_15_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_10_0
#define init_2_15_10_0
void FASTCALL call_2_15_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_10_1
#define init_2_15_10_1
void FASTCALL call_2_15_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_10_2
#define init_2_15_10_2
void FASTCALL call_2_15_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_10_3
#define init_2_15_10_3
void FASTCALL call_2_15_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_10_4
#define init_2_15_10_4
void FASTCALL call_2_15_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_10_5
#define init_2_15_10_5
void FASTCALL call_2_15_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_10_6
#define init_2_15_10_6
void FASTCALL call_2_15_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_10_7
#define init_2_15_10_7
void FASTCALL call_2_15_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_10_8
#define init_2_15_10_8
void FASTCALL call_2_15_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_10_9
#define init_2_15_10_9
void FASTCALL call_2_15_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_10_10
#define init_2_15_10_10
void FASTCALL call_2_15_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_10_11
#define init_2_15_10_11
void FASTCALL call_2_15_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_10_12
#define init_2_15_10_12
void FASTCALL call_2_15_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_10_13
#define init_2_15_10_13
void FASTCALL call_2_15_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_10_14
#define init_2_15_10_14
void FASTCALL call_2_15_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_10_15
#define init_2_15_10_15
void FASTCALL call_2_15_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_11_0
#define init_2_15_11_0
void FASTCALL call_2_15_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_11_1
#define init_2_15_11_1
void FASTCALL call_2_15_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_11_2
#define init_2_15_11_2
void FASTCALL call_2_15_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_11_3
#define init_2_15_11_3
void FASTCALL call_2_15_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_11_4
#define init_2_15_11_4
void FASTCALL call_2_15_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_11_5
#define init_2_15_11_5
void FASTCALL call_2_15_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_11_6
#define init_2_15_11_6
void FASTCALL call_2_15_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_11_7
#define init_2_15_11_7
void FASTCALL call_2_15_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_11_8
#define init_2_15_11_8
void FASTCALL call_2_15_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_11_9
#define init_2_15_11_9
void FASTCALL call_2_15_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_11_10
#define init_2_15_11_10
void FASTCALL call_2_15_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_11_11
#define init_2_15_11_11
void FASTCALL call_2_15_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_11_12
#define init_2_15_11_12
void FASTCALL call_2_15_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_11_13
#define init_2_15_11_13
void FASTCALL call_2_15_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_11_14
#define init_2_15_11_14
void FASTCALL call_2_15_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_11_15
#define init_2_15_11_15
void FASTCALL call_2_15_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_12_0
#define init_2_15_12_0
void FASTCALL call_2_15_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_12_1
#define init_2_15_12_1
void FASTCALL call_2_15_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_12_2
#define init_2_15_12_2
void FASTCALL call_2_15_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_12_3
#define init_2_15_12_3
void FASTCALL call_2_15_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_12_4
#define init_2_15_12_4
void FASTCALL call_2_15_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_12_5
#define init_2_15_12_5
void FASTCALL call_2_15_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_12_6
#define init_2_15_12_6
void FASTCALL call_2_15_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_12_7
#define init_2_15_12_7
void FASTCALL call_2_15_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_12_8
#define init_2_15_12_8
void FASTCALL call_2_15_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_12_9
#define init_2_15_12_9
void FASTCALL call_2_15_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_12_10
#define init_2_15_12_10
void FASTCALL call_2_15_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_12_11
#define init_2_15_12_11
void FASTCALL call_2_15_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_12_12
#define init_2_15_12_12
void FASTCALL call_2_15_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_12_13
#define init_2_15_12_13
void FASTCALL call_2_15_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_12_14
#define init_2_15_12_14
void FASTCALL call_2_15_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_12_15
#define init_2_15_12_15
void FASTCALL call_2_15_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_13_0
#define init_2_15_13_0
void FASTCALL call_2_15_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_13_1
#define init_2_15_13_1
void FASTCALL call_2_15_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_13_2
#define init_2_15_13_2
void FASTCALL call_2_15_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_13_3
#define init_2_15_13_3
void FASTCALL call_2_15_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_13_4
#define init_2_15_13_4
void FASTCALL call_2_15_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_13_5
#define init_2_15_13_5
void FASTCALL call_2_15_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_13_6
#define init_2_15_13_6
void FASTCALL call_2_15_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_13_7
#define init_2_15_13_7
void FASTCALL call_2_15_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_13_8
#define init_2_15_13_8
void FASTCALL call_2_15_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_13_9
#define init_2_15_13_9
void FASTCALL call_2_15_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_13_10
#define init_2_15_13_10
void FASTCALL call_2_15_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_13_11
#define init_2_15_13_11
void FASTCALL call_2_15_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_13_12
#define init_2_15_13_12
void FASTCALL call_2_15_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_13_13
#define init_2_15_13_13
void FASTCALL call_2_15_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_13_14
#define init_2_15_13_14
void FASTCALL call_2_15_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_13_15
#define init_2_15_13_15
void FASTCALL call_2_15_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_14_0
#define init_2_15_14_0
void FASTCALL call_2_15_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_14_1
#define init_2_15_14_1
void FASTCALL call_2_15_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_14_2
#define init_2_15_14_2
void FASTCALL call_2_15_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_14_3
#define init_2_15_14_3
void FASTCALL call_2_15_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_14_4
#define init_2_15_14_4
void FASTCALL call_2_15_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_14_5
#define init_2_15_14_5
void FASTCALL call_2_15_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_14_6
#define init_2_15_14_6
void FASTCALL call_2_15_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_14_7
#define init_2_15_14_7
void FASTCALL call_2_15_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_14_8
#define init_2_15_14_8
void FASTCALL call_2_15_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_14_9
#define init_2_15_14_9
void FASTCALL call_2_15_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_14_10
#define init_2_15_14_10
void FASTCALL call_2_15_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_14_11
#define init_2_15_14_11
void FASTCALL call_2_15_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_14_12
#define init_2_15_14_12
void FASTCALL call_2_15_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_14_13
#define init_2_15_14_13
void FASTCALL call_2_15_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_14_14
#define init_2_15_14_14
void FASTCALL call_2_15_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_14_15
#define init_2_15_14_15
void FASTCALL call_2_15_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_15_0
#define init_2_15_15_0
void FASTCALL call_2_15_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_15_1
#define init_2_15_15_1
void FASTCALL call_2_15_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_15_2
#define init_2_15_15_2
void FASTCALL call_2_15_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_15_3
#define init_2_15_15_3
void FASTCALL call_2_15_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_15_4
#define init_2_15_15_4
void FASTCALL call_2_15_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_15_5
#define init_2_15_15_5
void FASTCALL call_2_15_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_15_6
#define init_2_15_15_6
void FASTCALL call_2_15_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_15_7
#define init_2_15_15_7
void FASTCALL call_2_15_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_15_8
#define init_2_15_15_8
void FASTCALL call_2_15_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_15_9
#define init_2_15_15_9
void FASTCALL call_2_15_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_15_10
#define init_2_15_15_10
void FASTCALL call_2_15_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_15_11
#define init_2_15_15_11
void FASTCALL call_2_15_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_15_12
#define init_2_15_15_12
void FASTCALL call_2_15_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_15_13
#define init_2_15_15_13
void FASTCALL call_2_15_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_15_14
#define init_2_15_15_14
void FASTCALL call_2_15_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_2_15_15_15
#define init_2_15_15_15
void FASTCALL call_2_15_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_0_0
#define init_3_0_0_0
void FASTCALL call_3_0_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_0_1
#define init_3_0_0_1
void FASTCALL call_3_0_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_0_2
#define init_3_0_0_2
void FASTCALL call_3_0_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_0_3
#define init_3_0_0_3
void FASTCALL call_3_0_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_0_4
#define init_3_0_0_4
void FASTCALL call_3_0_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_0_5
#define init_3_0_0_5
void FASTCALL call_3_0_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_0_6
#define init_3_0_0_6
void FASTCALL call_3_0_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_0_7
#define init_3_0_0_7
void FASTCALL call_3_0_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_0_8
#define init_3_0_0_8
void FASTCALL call_3_0_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_0_9
#define init_3_0_0_9
void FASTCALL call_3_0_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_0_10
#define init_3_0_0_10
void FASTCALL call_3_0_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_0_11
#define init_3_0_0_11
void FASTCALL call_3_0_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_0_12
#define init_3_0_0_12
void FASTCALL call_3_0_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_0_13
#define init_3_0_0_13
void FASTCALL call_3_0_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_0_14
#define init_3_0_0_14
void FASTCALL call_3_0_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_0_15
#define init_3_0_0_15
void FASTCALL call_3_0_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_1_0
#define init_3_0_1_0
void FASTCALL call_3_0_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_1_1
#define init_3_0_1_1
void FASTCALL call_3_0_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_1_2
#define init_3_0_1_2
void FASTCALL call_3_0_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_1_3
#define init_3_0_1_3
void FASTCALL call_3_0_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_1_4
#define init_3_0_1_4
void FASTCALL call_3_0_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_1_5
#define init_3_0_1_5
void FASTCALL call_3_0_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_1_6
#define init_3_0_1_6
void FASTCALL call_3_0_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_1_7
#define init_3_0_1_7
void FASTCALL call_3_0_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_1_8
#define init_3_0_1_8
void FASTCALL call_3_0_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_1_9
#define init_3_0_1_9
void FASTCALL call_3_0_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_1_10
#define init_3_0_1_10
void FASTCALL call_3_0_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_1_11
#define init_3_0_1_11
void FASTCALL call_3_0_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_1_12
#define init_3_0_1_12
void FASTCALL call_3_0_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_1_13
#define init_3_0_1_13
void FASTCALL call_3_0_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_1_14
#define init_3_0_1_14
void FASTCALL call_3_0_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_1_15
#define init_3_0_1_15
void FASTCALL call_3_0_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_2_0
#define init_3_0_2_0
void FASTCALL call_3_0_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_2_1
#define init_3_0_2_1
void FASTCALL call_3_0_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_2_2
#define init_3_0_2_2
void FASTCALL call_3_0_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_2_3
#define init_3_0_2_3
void FASTCALL call_3_0_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_2_4
#define init_3_0_2_4
void FASTCALL call_3_0_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_2_5
#define init_3_0_2_5
void FASTCALL call_3_0_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_2_6
#define init_3_0_2_6
void FASTCALL call_3_0_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_2_7
#define init_3_0_2_7
void FASTCALL call_3_0_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_2_8
#define init_3_0_2_8
void FASTCALL call_3_0_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_2_9
#define init_3_0_2_9
void FASTCALL call_3_0_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_2_10
#define init_3_0_2_10
void FASTCALL call_3_0_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_2_11
#define init_3_0_2_11
void FASTCALL call_3_0_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_2_12
#define init_3_0_2_12
void FASTCALL call_3_0_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_2_13
#define init_3_0_2_13
void FASTCALL call_3_0_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_2_14
#define init_3_0_2_14
void FASTCALL call_3_0_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_2_15
#define init_3_0_2_15
void FASTCALL call_3_0_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_3_0
#define init_3_0_3_0
void FASTCALL call_3_0_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_3_1
#define init_3_0_3_1
void FASTCALL call_3_0_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_3_2
#define init_3_0_3_2
void FASTCALL call_3_0_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_3_3
#define init_3_0_3_3
void FASTCALL call_3_0_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_3_4
#define init_3_0_3_4
void FASTCALL call_3_0_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_3_5
#define init_3_0_3_5
void FASTCALL call_3_0_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_3_6
#define init_3_0_3_6
void FASTCALL call_3_0_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_3_7
#define init_3_0_3_7
void FASTCALL call_3_0_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_3_8
#define init_3_0_3_8
void FASTCALL call_3_0_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_3_9
#define init_3_0_3_9
void FASTCALL call_3_0_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_3_10
#define init_3_0_3_10
void FASTCALL call_3_0_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_3_11
#define init_3_0_3_11
void FASTCALL call_3_0_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_3_12
#define init_3_0_3_12
void FASTCALL call_3_0_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_3_13
#define init_3_0_3_13
void FASTCALL call_3_0_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_3_14
#define init_3_0_3_14
void FASTCALL call_3_0_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_3_15
#define init_3_0_3_15
void FASTCALL call_3_0_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_4_0
#define init_3_0_4_0
void FASTCALL call_3_0_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_4_1
#define init_3_0_4_1
void FASTCALL call_3_0_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_4_2
#define init_3_0_4_2
void FASTCALL call_3_0_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_4_3
#define init_3_0_4_3
void FASTCALL call_3_0_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_4_4
#define init_3_0_4_4
void FASTCALL call_3_0_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_4_5
#define init_3_0_4_5
void FASTCALL call_3_0_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_4_6
#define init_3_0_4_6
void FASTCALL call_3_0_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_4_7
#define init_3_0_4_7
void FASTCALL call_3_0_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_4_8
#define init_3_0_4_8
void FASTCALL call_3_0_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_4_9
#define init_3_0_4_9
void FASTCALL call_3_0_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_4_10
#define init_3_0_4_10
void FASTCALL call_3_0_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_4_11
#define init_3_0_4_11
void FASTCALL call_3_0_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_4_12
#define init_3_0_4_12
void FASTCALL call_3_0_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_4_13
#define init_3_0_4_13
void FASTCALL call_3_0_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_4_14
#define init_3_0_4_14
void FASTCALL call_3_0_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_4_15
#define init_3_0_4_15
void FASTCALL call_3_0_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_5_0
#define init_3_0_5_0
void FASTCALL call_3_0_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_5_1
#define init_3_0_5_1
void FASTCALL call_3_0_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_5_2
#define init_3_0_5_2
void FASTCALL call_3_0_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_5_3
#define init_3_0_5_3
void FASTCALL call_3_0_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_5_4
#define init_3_0_5_4
void FASTCALL call_3_0_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_5_5
#define init_3_0_5_5
void FASTCALL call_3_0_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_5_6
#define init_3_0_5_6
void FASTCALL call_3_0_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_5_7
#define init_3_0_5_7
void FASTCALL call_3_0_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_5_8
#define init_3_0_5_8
void FASTCALL call_3_0_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_5_9
#define init_3_0_5_9
void FASTCALL call_3_0_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_5_10
#define init_3_0_5_10
void FASTCALL call_3_0_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_5_11
#define init_3_0_5_11
void FASTCALL call_3_0_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_5_12
#define init_3_0_5_12
void FASTCALL call_3_0_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_5_13
#define init_3_0_5_13
void FASTCALL call_3_0_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_5_14
#define init_3_0_5_14
void FASTCALL call_3_0_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_5_15
#define init_3_0_5_15
void FASTCALL call_3_0_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_6_0
#define init_3_0_6_0
void FASTCALL call_3_0_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_6_1
#define init_3_0_6_1
void FASTCALL call_3_0_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_6_2
#define init_3_0_6_2
void FASTCALL call_3_0_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_6_3
#define init_3_0_6_3
void FASTCALL call_3_0_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_6_4
#define init_3_0_6_4
void FASTCALL call_3_0_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_6_5
#define init_3_0_6_5
void FASTCALL call_3_0_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_6_6
#define init_3_0_6_6
void FASTCALL call_3_0_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_6_7
#define init_3_0_6_7
void FASTCALL call_3_0_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_6_8
#define init_3_0_6_8
void FASTCALL call_3_0_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_6_9
#define init_3_0_6_9
void FASTCALL call_3_0_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_6_10
#define init_3_0_6_10
void FASTCALL call_3_0_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_6_11
#define init_3_0_6_11
void FASTCALL call_3_0_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_6_12
#define init_3_0_6_12
void FASTCALL call_3_0_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_6_13
#define init_3_0_6_13
void FASTCALL call_3_0_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_6_14
#define init_3_0_6_14
void FASTCALL call_3_0_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_6_15
#define init_3_0_6_15
void FASTCALL call_3_0_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_7_0
#define init_3_0_7_0
void FASTCALL call_3_0_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_7_1
#define init_3_0_7_1
void FASTCALL call_3_0_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_7_2
#define init_3_0_7_2
void FASTCALL call_3_0_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_7_3
#define init_3_0_7_3
void FASTCALL call_3_0_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_7_4
#define init_3_0_7_4
void FASTCALL call_3_0_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_7_5
#define init_3_0_7_5
void FASTCALL call_3_0_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_7_6
#define init_3_0_7_6
void FASTCALL call_3_0_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_7_7
#define init_3_0_7_7
void FASTCALL call_3_0_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_7_8
#define init_3_0_7_8
void FASTCALL call_3_0_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_7_9
#define init_3_0_7_9
void FASTCALL call_3_0_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_7_10
#define init_3_0_7_10
void FASTCALL call_3_0_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_7_11
#define init_3_0_7_11
void FASTCALL call_3_0_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_7_12
#define init_3_0_7_12
void FASTCALL call_3_0_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_7_13
#define init_3_0_7_13
void FASTCALL call_3_0_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_7_14
#define init_3_0_7_14
void FASTCALL call_3_0_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_7_15
#define init_3_0_7_15
void FASTCALL call_3_0_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_8_0
#define init_3_0_8_0
void FASTCALL call_3_0_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_8_1
#define init_3_0_8_1
void FASTCALL call_3_0_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_8_2
#define init_3_0_8_2
void FASTCALL call_3_0_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_8_3
#define init_3_0_8_3
void FASTCALL call_3_0_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_8_4
#define init_3_0_8_4
void FASTCALL call_3_0_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_8_5
#define init_3_0_8_5
void FASTCALL call_3_0_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_8_6
#define init_3_0_8_6
void FASTCALL call_3_0_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_8_7
#define init_3_0_8_7
void FASTCALL call_3_0_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_8_8
#define init_3_0_8_8
void FASTCALL call_3_0_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_8_9
#define init_3_0_8_9
void FASTCALL call_3_0_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_8_10
#define init_3_0_8_10
void FASTCALL call_3_0_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_8_11
#define init_3_0_8_11
void FASTCALL call_3_0_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_8_12
#define init_3_0_8_12
void FASTCALL call_3_0_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_8_13
#define init_3_0_8_13
void FASTCALL call_3_0_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_8_14
#define init_3_0_8_14
void FASTCALL call_3_0_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_8_15
#define init_3_0_8_15
void FASTCALL call_3_0_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_9_0
#define init_3_0_9_0
void FASTCALL call_3_0_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_9_1
#define init_3_0_9_1
void FASTCALL call_3_0_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_9_2
#define init_3_0_9_2
void FASTCALL call_3_0_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_9_3
#define init_3_0_9_3
void FASTCALL call_3_0_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_9_4
#define init_3_0_9_4
void FASTCALL call_3_0_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_9_5
#define init_3_0_9_5
void FASTCALL call_3_0_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_9_6
#define init_3_0_9_6
void FASTCALL call_3_0_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_9_7
#define init_3_0_9_7
void FASTCALL call_3_0_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_9_8
#define init_3_0_9_8
void FASTCALL call_3_0_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_9_9
#define init_3_0_9_9
void FASTCALL call_3_0_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_9_10
#define init_3_0_9_10
void FASTCALL call_3_0_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_9_11
#define init_3_0_9_11
void FASTCALL call_3_0_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_9_12
#define init_3_0_9_12
void FASTCALL call_3_0_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_9_13
#define init_3_0_9_13
void FASTCALL call_3_0_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_9_14
#define init_3_0_9_14
void FASTCALL call_3_0_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_9_15
#define init_3_0_9_15
void FASTCALL call_3_0_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_10_0
#define init_3_0_10_0
void FASTCALL call_3_0_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_10_1
#define init_3_0_10_1
void FASTCALL call_3_0_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_10_2
#define init_3_0_10_2
void FASTCALL call_3_0_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_10_3
#define init_3_0_10_3
void FASTCALL call_3_0_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_10_4
#define init_3_0_10_4
void FASTCALL call_3_0_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_10_5
#define init_3_0_10_5
void FASTCALL call_3_0_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_10_6
#define init_3_0_10_6
void FASTCALL call_3_0_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_10_7
#define init_3_0_10_7
void FASTCALL call_3_0_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_10_8
#define init_3_0_10_8
void FASTCALL call_3_0_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_10_9
#define init_3_0_10_9
void FASTCALL call_3_0_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_10_10
#define init_3_0_10_10
void FASTCALL call_3_0_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_10_11
#define init_3_0_10_11
void FASTCALL call_3_0_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_10_12
#define init_3_0_10_12
void FASTCALL call_3_0_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_10_13
#define init_3_0_10_13
void FASTCALL call_3_0_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_10_14
#define init_3_0_10_14
void FASTCALL call_3_0_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_10_15
#define init_3_0_10_15
void FASTCALL call_3_0_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_11_0
#define init_3_0_11_0
void FASTCALL call_3_0_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_11_1
#define init_3_0_11_1
void FASTCALL call_3_0_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_11_2
#define init_3_0_11_2
void FASTCALL call_3_0_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_11_3
#define init_3_0_11_3
void FASTCALL call_3_0_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_11_4
#define init_3_0_11_4
void FASTCALL call_3_0_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_11_5
#define init_3_0_11_5
void FASTCALL call_3_0_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_11_6
#define init_3_0_11_6
void FASTCALL call_3_0_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_11_7
#define init_3_0_11_7
void FASTCALL call_3_0_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_11_8
#define init_3_0_11_8
void FASTCALL call_3_0_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_11_9
#define init_3_0_11_9
void FASTCALL call_3_0_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_11_10
#define init_3_0_11_10
void FASTCALL call_3_0_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_11_11
#define init_3_0_11_11
void FASTCALL call_3_0_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_11_12
#define init_3_0_11_12
void FASTCALL call_3_0_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_11_13
#define init_3_0_11_13
void FASTCALL call_3_0_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_11_14
#define init_3_0_11_14
void FASTCALL call_3_0_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_11_15
#define init_3_0_11_15
void FASTCALL call_3_0_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_12_0
#define init_3_0_12_0
void FASTCALL call_3_0_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_12_1
#define init_3_0_12_1
void FASTCALL call_3_0_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_12_2
#define init_3_0_12_2
void FASTCALL call_3_0_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_12_3
#define init_3_0_12_3
void FASTCALL call_3_0_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_12_4
#define init_3_0_12_4
void FASTCALL call_3_0_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_12_5
#define init_3_0_12_5
void FASTCALL call_3_0_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_12_6
#define init_3_0_12_6
void FASTCALL call_3_0_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_12_7
#define init_3_0_12_7
void FASTCALL call_3_0_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_12_8
#define init_3_0_12_8
void FASTCALL call_3_0_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_12_9
#define init_3_0_12_9
void FASTCALL call_3_0_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_12_10
#define init_3_0_12_10
void FASTCALL call_3_0_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_12_11
#define init_3_0_12_11
void FASTCALL call_3_0_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_12_12
#define init_3_0_12_12
void FASTCALL call_3_0_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_12_13
#define init_3_0_12_13
void FASTCALL call_3_0_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_12_14
#define init_3_0_12_14
void FASTCALL call_3_0_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_12_15
#define init_3_0_12_15
void FASTCALL call_3_0_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_13_0
#define init_3_0_13_0
void FASTCALL call_3_0_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_13_1
#define init_3_0_13_1
void FASTCALL call_3_0_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_13_2
#define init_3_0_13_2
void FASTCALL call_3_0_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_13_3
#define init_3_0_13_3
void FASTCALL call_3_0_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_13_4
#define init_3_0_13_4
void FASTCALL call_3_0_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_13_5
#define init_3_0_13_5
void FASTCALL call_3_0_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_13_6
#define init_3_0_13_6
void FASTCALL call_3_0_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_13_7
#define init_3_0_13_7
void FASTCALL call_3_0_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_13_8
#define init_3_0_13_8
void FASTCALL call_3_0_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_13_9
#define init_3_0_13_9
void FASTCALL call_3_0_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_13_10
#define init_3_0_13_10
void FASTCALL call_3_0_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_13_11
#define init_3_0_13_11
void FASTCALL call_3_0_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_13_12
#define init_3_0_13_12
void FASTCALL call_3_0_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_13_13
#define init_3_0_13_13
void FASTCALL call_3_0_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_13_14
#define init_3_0_13_14
void FASTCALL call_3_0_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_13_15
#define init_3_0_13_15
void FASTCALL call_3_0_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_14_0
#define init_3_0_14_0
void FASTCALL call_3_0_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_14_1
#define init_3_0_14_1
void FASTCALL call_3_0_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_14_2
#define init_3_0_14_2
void FASTCALL call_3_0_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_14_3
#define init_3_0_14_3
void FASTCALL call_3_0_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_14_4
#define init_3_0_14_4
void FASTCALL call_3_0_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_14_5
#define init_3_0_14_5
void FASTCALL call_3_0_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_14_6
#define init_3_0_14_6
void FASTCALL call_3_0_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_14_7
#define init_3_0_14_7
void FASTCALL call_3_0_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_14_8
#define init_3_0_14_8
void FASTCALL call_3_0_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_14_9
#define init_3_0_14_9
void FASTCALL call_3_0_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_14_10
#define init_3_0_14_10
void FASTCALL call_3_0_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_14_11
#define init_3_0_14_11
void FASTCALL call_3_0_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_14_12
#define init_3_0_14_12
void FASTCALL call_3_0_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_14_13
#define init_3_0_14_13
void FASTCALL call_3_0_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_14_14
#define init_3_0_14_14
void FASTCALL call_3_0_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_14_15
#define init_3_0_14_15
void FASTCALL call_3_0_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_15_0
#define init_3_0_15_0
void FASTCALL call_3_0_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_15_1
#define init_3_0_15_1
void FASTCALL call_3_0_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_15_2
#define init_3_0_15_2
void FASTCALL call_3_0_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_15_3
#define init_3_0_15_3
void FASTCALL call_3_0_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_15_4
#define init_3_0_15_4
void FASTCALL call_3_0_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_15_5
#define init_3_0_15_5
void FASTCALL call_3_0_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_15_6
#define init_3_0_15_6
void FASTCALL call_3_0_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_15_7
#define init_3_0_15_7
void FASTCALL call_3_0_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_15_8
#define init_3_0_15_8
void FASTCALL call_3_0_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_15_9
#define init_3_0_15_9
void FASTCALL call_3_0_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_15_10
#define init_3_0_15_10
void FASTCALL call_3_0_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_15_11
#define init_3_0_15_11
void FASTCALL call_3_0_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_15_12
#define init_3_0_15_12
void FASTCALL call_3_0_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_15_13
#define init_3_0_15_13
void FASTCALL call_3_0_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_15_14
#define init_3_0_15_14
void FASTCALL call_3_0_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_0_15_15
#define init_3_0_15_15
void FASTCALL call_3_0_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_0_0
#define init_3_1_0_0
void FASTCALL call_3_1_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_0_1
#define init_3_1_0_1
void FASTCALL call_3_1_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_0_2
#define init_3_1_0_2
void FASTCALL call_3_1_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_0_3
#define init_3_1_0_3
void FASTCALL call_3_1_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_0_4
#define init_3_1_0_4
void FASTCALL call_3_1_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_0_5
#define init_3_1_0_5
void FASTCALL call_3_1_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_0_6
#define init_3_1_0_6
void FASTCALL call_3_1_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_0_7
#define init_3_1_0_7
void FASTCALL call_3_1_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_0_8
#define init_3_1_0_8
void FASTCALL call_3_1_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_0_9
#define init_3_1_0_9
void FASTCALL call_3_1_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_0_10
#define init_3_1_0_10
void FASTCALL call_3_1_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_0_11
#define init_3_1_0_11
void FASTCALL call_3_1_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_0_12
#define init_3_1_0_12
void FASTCALL call_3_1_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_0_13
#define init_3_1_0_13
void FASTCALL call_3_1_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_0_14
#define init_3_1_0_14
void FASTCALL call_3_1_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_0_15
#define init_3_1_0_15
void FASTCALL call_3_1_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_1_0
#define init_3_1_1_0
void FASTCALL call_3_1_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_1_1
#define init_3_1_1_1
void FASTCALL call_3_1_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_1_2
#define init_3_1_1_2
void FASTCALL call_3_1_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_1_3
#define init_3_1_1_3
void FASTCALL call_3_1_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_1_4
#define init_3_1_1_4
void FASTCALL call_3_1_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_1_5
#define init_3_1_1_5
void FASTCALL call_3_1_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_1_6
#define init_3_1_1_6
void FASTCALL call_3_1_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_1_7
#define init_3_1_1_7
void FASTCALL call_3_1_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_1_8
#define init_3_1_1_8
void FASTCALL call_3_1_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_1_9
#define init_3_1_1_9
void FASTCALL call_3_1_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_1_10
#define init_3_1_1_10
void FASTCALL call_3_1_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_1_11
#define init_3_1_1_11
void FASTCALL call_3_1_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_1_12
#define init_3_1_1_12
void FASTCALL call_3_1_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_1_13
#define init_3_1_1_13
void FASTCALL call_3_1_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_1_14
#define init_3_1_1_14
void FASTCALL call_3_1_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_1_15
#define init_3_1_1_15
void FASTCALL call_3_1_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_2_0
#define init_3_1_2_0
void FASTCALL call_3_1_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_2_1
#define init_3_1_2_1
void FASTCALL call_3_1_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_2_2
#define init_3_1_2_2
void FASTCALL call_3_1_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_2_3
#define init_3_1_2_3
void FASTCALL call_3_1_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_2_4
#define init_3_1_2_4
void FASTCALL call_3_1_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_2_5
#define init_3_1_2_5
void FASTCALL call_3_1_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_2_6
#define init_3_1_2_6
void FASTCALL call_3_1_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_2_7
#define init_3_1_2_7
void FASTCALL call_3_1_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_2_8
#define init_3_1_2_8
void FASTCALL call_3_1_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_2_9
#define init_3_1_2_9
void FASTCALL call_3_1_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_2_10
#define init_3_1_2_10
void FASTCALL call_3_1_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_2_11
#define init_3_1_2_11
void FASTCALL call_3_1_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_2_12
#define init_3_1_2_12
void FASTCALL call_3_1_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_2_13
#define init_3_1_2_13
void FASTCALL call_3_1_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_2_14
#define init_3_1_2_14
void FASTCALL call_3_1_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_2_15
#define init_3_1_2_15
void FASTCALL call_3_1_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_3_0
#define init_3_1_3_0
void FASTCALL call_3_1_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_3_1
#define init_3_1_3_1
void FASTCALL call_3_1_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_3_2
#define init_3_1_3_2
void FASTCALL call_3_1_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_3_3
#define init_3_1_3_3
void FASTCALL call_3_1_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_3_4
#define init_3_1_3_4
void FASTCALL call_3_1_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_3_5
#define init_3_1_3_5
void FASTCALL call_3_1_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_3_6
#define init_3_1_3_6
void FASTCALL call_3_1_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_3_7
#define init_3_1_3_7
void FASTCALL call_3_1_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_3_8
#define init_3_1_3_8
void FASTCALL call_3_1_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_3_9
#define init_3_1_3_9
void FASTCALL call_3_1_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_3_10
#define init_3_1_3_10
void FASTCALL call_3_1_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_3_11
#define init_3_1_3_11
void FASTCALL call_3_1_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_3_12
#define init_3_1_3_12
void FASTCALL call_3_1_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_3_13
#define init_3_1_3_13
void FASTCALL call_3_1_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_3_14
#define init_3_1_3_14
void FASTCALL call_3_1_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_3_15
#define init_3_1_3_15
void FASTCALL call_3_1_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_4_0
#define init_3_1_4_0
void FASTCALL call_3_1_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_4_1
#define init_3_1_4_1
void FASTCALL call_3_1_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_4_2
#define init_3_1_4_2
void FASTCALL call_3_1_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_4_3
#define init_3_1_4_3
void FASTCALL call_3_1_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_4_4
#define init_3_1_4_4
void FASTCALL call_3_1_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_4_5
#define init_3_1_4_5
void FASTCALL call_3_1_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_4_6
#define init_3_1_4_6
void FASTCALL call_3_1_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_4_7
#define init_3_1_4_7
void FASTCALL call_3_1_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_4_8
#define init_3_1_4_8
void FASTCALL call_3_1_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_4_9
#define init_3_1_4_9
void FASTCALL call_3_1_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_4_10
#define init_3_1_4_10
void FASTCALL call_3_1_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_4_11
#define init_3_1_4_11
void FASTCALL call_3_1_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_4_12
#define init_3_1_4_12
void FASTCALL call_3_1_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_4_13
#define init_3_1_4_13
void FASTCALL call_3_1_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_4_14
#define init_3_1_4_14
void FASTCALL call_3_1_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_4_15
#define init_3_1_4_15
void FASTCALL call_3_1_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_5_0
#define init_3_1_5_0
void FASTCALL call_3_1_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_5_1
#define init_3_1_5_1
void FASTCALL call_3_1_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_5_2
#define init_3_1_5_2
void FASTCALL call_3_1_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_5_3
#define init_3_1_5_3
void FASTCALL call_3_1_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_5_4
#define init_3_1_5_4
void FASTCALL call_3_1_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_5_5
#define init_3_1_5_5
void FASTCALL call_3_1_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_5_6
#define init_3_1_5_6
void FASTCALL call_3_1_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_5_7
#define init_3_1_5_7
void FASTCALL call_3_1_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_5_8
#define init_3_1_5_8
void FASTCALL call_3_1_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_5_9
#define init_3_1_5_9
void FASTCALL call_3_1_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_5_10
#define init_3_1_5_10
void FASTCALL call_3_1_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_5_11
#define init_3_1_5_11
void FASTCALL call_3_1_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_5_12
#define init_3_1_5_12
void FASTCALL call_3_1_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_5_13
#define init_3_1_5_13
void FASTCALL call_3_1_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_5_14
#define init_3_1_5_14
void FASTCALL call_3_1_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_5_15
#define init_3_1_5_15
void FASTCALL call_3_1_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_6_0
#define init_3_1_6_0
void FASTCALL call_3_1_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_6_1
#define init_3_1_6_1
void FASTCALL call_3_1_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_6_2
#define init_3_1_6_2
void FASTCALL call_3_1_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_6_3
#define init_3_1_6_3
void FASTCALL call_3_1_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_6_4
#define init_3_1_6_4
void FASTCALL call_3_1_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_6_5
#define init_3_1_6_5
void FASTCALL call_3_1_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_6_6
#define init_3_1_6_6
void FASTCALL call_3_1_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_6_7
#define init_3_1_6_7
void FASTCALL call_3_1_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_6_8
#define init_3_1_6_8
void FASTCALL call_3_1_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_6_9
#define init_3_1_6_9
void FASTCALL call_3_1_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_6_10
#define init_3_1_6_10
void FASTCALL call_3_1_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_6_11
#define init_3_1_6_11
void FASTCALL call_3_1_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_6_12
#define init_3_1_6_12
void FASTCALL call_3_1_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_6_13
#define init_3_1_6_13
void FASTCALL call_3_1_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_6_14
#define init_3_1_6_14
void FASTCALL call_3_1_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_6_15
#define init_3_1_6_15
void FASTCALL call_3_1_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_7_0
#define init_3_1_7_0
void FASTCALL call_3_1_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_7_1
#define init_3_1_7_1
void FASTCALL call_3_1_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_7_2
#define init_3_1_7_2
void FASTCALL call_3_1_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_7_3
#define init_3_1_7_3
void FASTCALL call_3_1_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_7_4
#define init_3_1_7_4
void FASTCALL call_3_1_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_7_5
#define init_3_1_7_5
void FASTCALL call_3_1_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_7_6
#define init_3_1_7_6
void FASTCALL call_3_1_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_7_7
#define init_3_1_7_7
void FASTCALL call_3_1_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_7_8
#define init_3_1_7_8
void FASTCALL call_3_1_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_7_9
#define init_3_1_7_9
void FASTCALL call_3_1_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_7_10
#define init_3_1_7_10
void FASTCALL call_3_1_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_7_11
#define init_3_1_7_11
void FASTCALL call_3_1_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_7_12
#define init_3_1_7_12
void FASTCALL call_3_1_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_7_13
#define init_3_1_7_13
void FASTCALL call_3_1_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_7_14
#define init_3_1_7_14
void FASTCALL call_3_1_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_7_15
#define init_3_1_7_15
void FASTCALL call_3_1_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_8_0
#define init_3_1_8_0
void FASTCALL call_3_1_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_8_1
#define init_3_1_8_1
void FASTCALL call_3_1_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_8_2
#define init_3_1_8_2
void FASTCALL call_3_1_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_8_3
#define init_3_1_8_3
void FASTCALL call_3_1_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_8_4
#define init_3_1_8_4
void FASTCALL call_3_1_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_8_5
#define init_3_1_8_5
void FASTCALL call_3_1_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_8_6
#define init_3_1_8_6
void FASTCALL call_3_1_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_8_7
#define init_3_1_8_7
void FASTCALL call_3_1_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_8_8
#define init_3_1_8_8
void FASTCALL call_3_1_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_8_9
#define init_3_1_8_9
void FASTCALL call_3_1_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_8_10
#define init_3_1_8_10
void FASTCALL call_3_1_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_8_11
#define init_3_1_8_11
void FASTCALL call_3_1_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_8_12
#define init_3_1_8_12
void FASTCALL call_3_1_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_8_13
#define init_3_1_8_13
void FASTCALL call_3_1_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_8_14
#define init_3_1_8_14
void FASTCALL call_3_1_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_8_15
#define init_3_1_8_15
void FASTCALL call_3_1_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_9_0
#define init_3_1_9_0
void FASTCALL call_3_1_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_9_1
#define init_3_1_9_1
void FASTCALL call_3_1_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_9_2
#define init_3_1_9_2
void FASTCALL call_3_1_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_9_3
#define init_3_1_9_3
void FASTCALL call_3_1_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_9_4
#define init_3_1_9_4
void FASTCALL call_3_1_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_9_5
#define init_3_1_9_5
void FASTCALL call_3_1_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_9_6
#define init_3_1_9_6
void FASTCALL call_3_1_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_9_7
#define init_3_1_9_7
void FASTCALL call_3_1_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_9_8
#define init_3_1_9_8
void FASTCALL call_3_1_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_9_9
#define init_3_1_9_9
void FASTCALL call_3_1_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_9_10
#define init_3_1_9_10
void FASTCALL call_3_1_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_9_11
#define init_3_1_9_11
void FASTCALL call_3_1_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_9_12
#define init_3_1_9_12
void FASTCALL call_3_1_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_9_13
#define init_3_1_9_13
void FASTCALL call_3_1_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_9_14
#define init_3_1_9_14
void FASTCALL call_3_1_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_9_15
#define init_3_1_9_15
void FASTCALL call_3_1_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_10_0
#define init_3_1_10_0
void FASTCALL call_3_1_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_10_1
#define init_3_1_10_1
void FASTCALL call_3_1_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_10_2
#define init_3_1_10_2
void FASTCALL call_3_1_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_10_3
#define init_3_1_10_3
void FASTCALL call_3_1_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_10_4
#define init_3_1_10_4
void FASTCALL call_3_1_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_10_5
#define init_3_1_10_5
void FASTCALL call_3_1_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_10_6
#define init_3_1_10_6
void FASTCALL call_3_1_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_10_7
#define init_3_1_10_7
void FASTCALL call_3_1_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_10_8
#define init_3_1_10_8
void FASTCALL call_3_1_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_10_9
#define init_3_1_10_9
void FASTCALL call_3_1_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_10_10
#define init_3_1_10_10
void FASTCALL call_3_1_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_10_11
#define init_3_1_10_11
void FASTCALL call_3_1_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_10_12
#define init_3_1_10_12
void FASTCALL call_3_1_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_10_13
#define init_3_1_10_13
void FASTCALL call_3_1_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_10_14
#define init_3_1_10_14
void FASTCALL call_3_1_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_10_15
#define init_3_1_10_15
void FASTCALL call_3_1_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_11_0
#define init_3_1_11_0
void FASTCALL call_3_1_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_11_1
#define init_3_1_11_1
void FASTCALL call_3_1_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_11_2
#define init_3_1_11_2
void FASTCALL call_3_1_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_11_3
#define init_3_1_11_3
void FASTCALL call_3_1_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_11_4
#define init_3_1_11_4
void FASTCALL call_3_1_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_11_5
#define init_3_1_11_5
void FASTCALL call_3_1_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_11_6
#define init_3_1_11_6
void FASTCALL call_3_1_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_11_7
#define init_3_1_11_7
void FASTCALL call_3_1_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_11_8
#define init_3_1_11_8
void FASTCALL call_3_1_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_11_9
#define init_3_1_11_9
void FASTCALL call_3_1_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_11_10
#define init_3_1_11_10
void FASTCALL call_3_1_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_11_11
#define init_3_1_11_11
void FASTCALL call_3_1_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_11_12
#define init_3_1_11_12
void FASTCALL call_3_1_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_11_13
#define init_3_1_11_13
void FASTCALL call_3_1_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_11_14
#define init_3_1_11_14
void FASTCALL call_3_1_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_11_15
#define init_3_1_11_15
void FASTCALL call_3_1_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_12_0
#define init_3_1_12_0
void FASTCALL call_3_1_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_12_1
#define init_3_1_12_1
void FASTCALL call_3_1_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_12_2
#define init_3_1_12_2
void FASTCALL call_3_1_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_12_3
#define init_3_1_12_3
void FASTCALL call_3_1_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_12_4
#define init_3_1_12_4
void FASTCALL call_3_1_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_12_5
#define init_3_1_12_5
void FASTCALL call_3_1_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_12_6
#define init_3_1_12_6
void FASTCALL call_3_1_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_12_7
#define init_3_1_12_7
void FASTCALL call_3_1_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_12_8
#define init_3_1_12_8
void FASTCALL call_3_1_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_12_9
#define init_3_1_12_9
void FASTCALL call_3_1_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_12_10
#define init_3_1_12_10
void FASTCALL call_3_1_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_12_11
#define init_3_1_12_11
void FASTCALL call_3_1_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_12_12
#define init_3_1_12_12
void FASTCALL call_3_1_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_12_13
#define init_3_1_12_13
void FASTCALL call_3_1_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_12_14
#define init_3_1_12_14
void FASTCALL call_3_1_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_12_15
#define init_3_1_12_15
void FASTCALL call_3_1_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_13_0
#define init_3_1_13_0
void FASTCALL call_3_1_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_13_1
#define init_3_1_13_1
void FASTCALL call_3_1_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_13_2
#define init_3_1_13_2
void FASTCALL call_3_1_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_13_3
#define init_3_1_13_3
void FASTCALL call_3_1_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_13_4
#define init_3_1_13_4
void FASTCALL call_3_1_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_13_5
#define init_3_1_13_5
void FASTCALL call_3_1_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_13_6
#define init_3_1_13_6
void FASTCALL call_3_1_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_13_7
#define init_3_1_13_7
void FASTCALL call_3_1_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_13_8
#define init_3_1_13_8
void FASTCALL call_3_1_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_13_9
#define init_3_1_13_9
void FASTCALL call_3_1_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_13_10
#define init_3_1_13_10
void FASTCALL call_3_1_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_13_11
#define init_3_1_13_11
void FASTCALL call_3_1_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_13_12
#define init_3_1_13_12
void FASTCALL call_3_1_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_13_13
#define init_3_1_13_13
void FASTCALL call_3_1_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_13_14
#define init_3_1_13_14
void FASTCALL call_3_1_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_13_15
#define init_3_1_13_15
void FASTCALL call_3_1_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_14_0
#define init_3_1_14_0
void FASTCALL call_3_1_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_14_1
#define init_3_1_14_1
void FASTCALL call_3_1_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_14_2
#define init_3_1_14_2
void FASTCALL call_3_1_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_14_3
#define init_3_1_14_3
void FASTCALL call_3_1_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_14_4
#define init_3_1_14_4
void FASTCALL call_3_1_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_14_5
#define init_3_1_14_5
void FASTCALL call_3_1_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_14_6
#define init_3_1_14_6
void FASTCALL call_3_1_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_14_7
#define init_3_1_14_7
void FASTCALL call_3_1_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_14_8
#define init_3_1_14_8
void FASTCALL call_3_1_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_14_9
#define init_3_1_14_9
void FASTCALL call_3_1_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_14_10
#define init_3_1_14_10
void FASTCALL call_3_1_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_14_11
#define init_3_1_14_11
void FASTCALL call_3_1_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_14_12
#define init_3_1_14_12
void FASTCALL call_3_1_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_14_13
#define init_3_1_14_13
void FASTCALL call_3_1_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_14_14
#define init_3_1_14_14
void FASTCALL call_3_1_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_14_15
#define init_3_1_14_15
void FASTCALL call_3_1_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_15_0
#define init_3_1_15_0
void FASTCALL call_3_1_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_15_1
#define init_3_1_15_1
void FASTCALL call_3_1_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_15_2
#define init_3_1_15_2
void FASTCALL call_3_1_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_15_3
#define init_3_1_15_3
void FASTCALL call_3_1_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_15_4
#define init_3_1_15_4
void FASTCALL call_3_1_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_15_5
#define init_3_1_15_5
void FASTCALL call_3_1_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_15_6
#define init_3_1_15_6
void FASTCALL call_3_1_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_15_7
#define init_3_1_15_7
void FASTCALL call_3_1_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_15_8
#define init_3_1_15_8
void FASTCALL call_3_1_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_15_9
#define init_3_1_15_9
void FASTCALL call_3_1_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_15_10
#define init_3_1_15_10
void FASTCALL call_3_1_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_15_11
#define init_3_1_15_11
void FASTCALL call_3_1_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_15_12
#define init_3_1_15_12
void FASTCALL call_3_1_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_15_13
#define init_3_1_15_13
void FASTCALL call_3_1_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_15_14
#define init_3_1_15_14
void FASTCALL call_3_1_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_1_15_15
#define init_3_1_15_15
void FASTCALL call_3_1_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_0_0
#define init_3_2_0_0
void FASTCALL call_3_2_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_0_1
#define init_3_2_0_1
void FASTCALL call_3_2_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_0_2
#define init_3_2_0_2
void FASTCALL call_3_2_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_0_3
#define init_3_2_0_3
void FASTCALL call_3_2_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_0_4
#define init_3_2_0_4
void FASTCALL call_3_2_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_0_5
#define init_3_2_0_5
void FASTCALL call_3_2_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_0_6
#define init_3_2_0_6
void FASTCALL call_3_2_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_0_7
#define init_3_2_0_7
void FASTCALL call_3_2_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_0_8
#define init_3_2_0_8
void FASTCALL call_3_2_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_0_9
#define init_3_2_0_9
void FASTCALL call_3_2_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_0_10
#define init_3_2_0_10
void FASTCALL call_3_2_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_0_11
#define init_3_2_0_11
void FASTCALL call_3_2_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_0_12
#define init_3_2_0_12
void FASTCALL call_3_2_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_0_13
#define init_3_2_0_13
void FASTCALL call_3_2_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_0_14
#define init_3_2_0_14
void FASTCALL call_3_2_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_0_15
#define init_3_2_0_15
void FASTCALL call_3_2_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_1_0
#define init_3_2_1_0
void FASTCALL call_3_2_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_1_1
#define init_3_2_1_1
void FASTCALL call_3_2_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_1_2
#define init_3_2_1_2
void FASTCALL call_3_2_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_1_3
#define init_3_2_1_3
void FASTCALL call_3_2_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_1_4
#define init_3_2_1_4
void FASTCALL call_3_2_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_1_5
#define init_3_2_1_5
void FASTCALL call_3_2_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_1_6
#define init_3_2_1_6
void FASTCALL call_3_2_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_1_7
#define init_3_2_1_7
void FASTCALL call_3_2_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_1_8
#define init_3_2_1_8
void FASTCALL call_3_2_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_1_9
#define init_3_2_1_9
void FASTCALL call_3_2_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_1_10
#define init_3_2_1_10
void FASTCALL call_3_2_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_1_11
#define init_3_2_1_11
void FASTCALL call_3_2_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_1_12
#define init_3_2_1_12
void FASTCALL call_3_2_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_1_13
#define init_3_2_1_13
void FASTCALL call_3_2_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_1_14
#define init_3_2_1_14
void FASTCALL call_3_2_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_1_15
#define init_3_2_1_15
void FASTCALL call_3_2_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_2_0
#define init_3_2_2_0
void FASTCALL call_3_2_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_2_1
#define init_3_2_2_1
void FASTCALL call_3_2_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_2_2
#define init_3_2_2_2
void FASTCALL call_3_2_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_2_3
#define init_3_2_2_3
void FASTCALL call_3_2_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_2_4
#define init_3_2_2_4
void FASTCALL call_3_2_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_2_5
#define init_3_2_2_5
void FASTCALL call_3_2_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_2_6
#define init_3_2_2_6
void FASTCALL call_3_2_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_2_7
#define init_3_2_2_7
void FASTCALL call_3_2_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_2_8
#define init_3_2_2_8
void FASTCALL call_3_2_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_2_9
#define init_3_2_2_9
void FASTCALL call_3_2_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_2_10
#define init_3_2_2_10
void FASTCALL call_3_2_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_2_11
#define init_3_2_2_11
void FASTCALL call_3_2_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_2_12
#define init_3_2_2_12
void FASTCALL call_3_2_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_2_13
#define init_3_2_2_13
void FASTCALL call_3_2_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_2_14
#define init_3_2_2_14
void FASTCALL call_3_2_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_2_15
#define init_3_2_2_15
void FASTCALL call_3_2_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_3_0
#define init_3_2_3_0
void FASTCALL call_3_2_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_3_1
#define init_3_2_3_1
void FASTCALL call_3_2_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_3_2
#define init_3_2_3_2
void FASTCALL call_3_2_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_3_3
#define init_3_2_3_3
void FASTCALL call_3_2_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_3_4
#define init_3_2_3_4
void FASTCALL call_3_2_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_3_5
#define init_3_2_3_5
void FASTCALL call_3_2_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_3_6
#define init_3_2_3_6
void FASTCALL call_3_2_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_3_7
#define init_3_2_3_7
void FASTCALL call_3_2_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_3_8
#define init_3_2_3_8
void FASTCALL call_3_2_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_3_9
#define init_3_2_3_9
void FASTCALL call_3_2_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_3_10
#define init_3_2_3_10
void FASTCALL call_3_2_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_3_11
#define init_3_2_3_11
void FASTCALL call_3_2_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_3_12
#define init_3_2_3_12
void FASTCALL call_3_2_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_3_13
#define init_3_2_3_13
void FASTCALL call_3_2_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_3_14
#define init_3_2_3_14
void FASTCALL call_3_2_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_3_15
#define init_3_2_3_15
void FASTCALL call_3_2_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_4_0
#define init_3_2_4_0
void FASTCALL call_3_2_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_4_1
#define init_3_2_4_1
void FASTCALL call_3_2_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_4_2
#define init_3_2_4_2
void FASTCALL call_3_2_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_4_3
#define init_3_2_4_3
void FASTCALL call_3_2_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_4_4
#define init_3_2_4_4
void FASTCALL call_3_2_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_4_5
#define init_3_2_4_5
void FASTCALL call_3_2_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_4_6
#define init_3_2_4_6
void FASTCALL call_3_2_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_4_7
#define init_3_2_4_7
void FASTCALL call_3_2_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_4_8
#define init_3_2_4_8
void FASTCALL call_3_2_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_4_9
#define init_3_2_4_9
void FASTCALL call_3_2_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_4_10
#define init_3_2_4_10
void FASTCALL call_3_2_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_4_11
#define init_3_2_4_11
void FASTCALL call_3_2_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_4_12
#define init_3_2_4_12
void FASTCALL call_3_2_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_4_13
#define init_3_2_4_13
void FASTCALL call_3_2_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_4_14
#define init_3_2_4_14
void FASTCALL call_3_2_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_4_15
#define init_3_2_4_15
void FASTCALL call_3_2_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_5_0
#define init_3_2_5_0
void FASTCALL call_3_2_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_5_1
#define init_3_2_5_1
void FASTCALL call_3_2_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_5_2
#define init_3_2_5_2
void FASTCALL call_3_2_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_5_3
#define init_3_2_5_3
void FASTCALL call_3_2_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_5_4
#define init_3_2_5_4
void FASTCALL call_3_2_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_5_5
#define init_3_2_5_5
void FASTCALL call_3_2_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_5_6
#define init_3_2_5_6
void FASTCALL call_3_2_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_5_7
#define init_3_2_5_7
void FASTCALL call_3_2_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_5_8
#define init_3_2_5_8
void FASTCALL call_3_2_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_5_9
#define init_3_2_5_9
void FASTCALL call_3_2_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_5_10
#define init_3_2_5_10
void FASTCALL call_3_2_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_5_11
#define init_3_2_5_11
void FASTCALL call_3_2_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_5_12
#define init_3_2_5_12
void FASTCALL call_3_2_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_5_13
#define init_3_2_5_13
void FASTCALL call_3_2_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_5_14
#define init_3_2_5_14
void FASTCALL call_3_2_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_5_15
#define init_3_2_5_15
void FASTCALL call_3_2_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_6_0
#define init_3_2_6_0
void FASTCALL call_3_2_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_6_1
#define init_3_2_6_1
void FASTCALL call_3_2_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_6_2
#define init_3_2_6_2
void FASTCALL call_3_2_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_6_3
#define init_3_2_6_3
void FASTCALL call_3_2_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_6_4
#define init_3_2_6_4
void FASTCALL call_3_2_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_6_5
#define init_3_2_6_5
void FASTCALL call_3_2_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_6_6
#define init_3_2_6_6
void FASTCALL call_3_2_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_6_7
#define init_3_2_6_7
void FASTCALL call_3_2_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_6_8
#define init_3_2_6_8
void FASTCALL call_3_2_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_6_9
#define init_3_2_6_9
void FASTCALL call_3_2_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_6_10
#define init_3_2_6_10
void FASTCALL call_3_2_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_6_11
#define init_3_2_6_11
void FASTCALL call_3_2_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_6_12
#define init_3_2_6_12
void FASTCALL call_3_2_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_6_13
#define init_3_2_6_13
void FASTCALL call_3_2_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_6_14
#define init_3_2_6_14
void FASTCALL call_3_2_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_6_15
#define init_3_2_6_15
void FASTCALL call_3_2_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_7_0
#define init_3_2_7_0
void FASTCALL call_3_2_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_7_1
#define init_3_2_7_1
void FASTCALL call_3_2_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_7_2
#define init_3_2_7_2
void FASTCALL call_3_2_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_7_3
#define init_3_2_7_3
void FASTCALL call_3_2_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_7_4
#define init_3_2_7_4
void FASTCALL call_3_2_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_7_5
#define init_3_2_7_5
void FASTCALL call_3_2_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_7_6
#define init_3_2_7_6
void FASTCALL call_3_2_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_7_7
#define init_3_2_7_7
void FASTCALL call_3_2_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_7_8
#define init_3_2_7_8
void FASTCALL call_3_2_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_7_9
#define init_3_2_7_9
void FASTCALL call_3_2_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_7_10
#define init_3_2_7_10
void FASTCALL call_3_2_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_7_11
#define init_3_2_7_11
void FASTCALL call_3_2_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_7_12
#define init_3_2_7_12
void FASTCALL call_3_2_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_7_13
#define init_3_2_7_13
void FASTCALL call_3_2_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_7_14
#define init_3_2_7_14
void FASTCALL call_3_2_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_7_15
#define init_3_2_7_15
void FASTCALL call_3_2_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_8_0
#define init_3_2_8_0
void FASTCALL call_3_2_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_8_1
#define init_3_2_8_1
void FASTCALL call_3_2_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_8_2
#define init_3_2_8_2
void FASTCALL call_3_2_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_8_3
#define init_3_2_8_3
void FASTCALL call_3_2_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_8_4
#define init_3_2_8_4
void FASTCALL call_3_2_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_8_5
#define init_3_2_8_5
void FASTCALL call_3_2_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_8_6
#define init_3_2_8_6
void FASTCALL call_3_2_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_8_7
#define init_3_2_8_7
void FASTCALL call_3_2_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_8_8
#define init_3_2_8_8
void FASTCALL call_3_2_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_8_9
#define init_3_2_8_9
void FASTCALL call_3_2_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_8_10
#define init_3_2_8_10
void FASTCALL call_3_2_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_8_11
#define init_3_2_8_11
void FASTCALL call_3_2_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_8_12
#define init_3_2_8_12
void FASTCALL call_3_2_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_8_13
#define init_3_2_8_13
void FASTCALL call_3_2_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_8_14
#define init_3_2_8_14
void FASTCALL call_3_2_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_8_15
#define init_3_2_8_15
void FASTCALL call_3_2_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_9_0
#define init_3_2_9_0
void FASTCALL call_3_2_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_9_1
#define init_3_2_9_1
void FASTCALL call_3_2_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_9_2
#define init_3_2_9_2
void FASTCALL call_3_2_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_9_3
#define init_3_2_9_3
void FASTCALL call_3_2_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_9_4
#define init_3_2_9_4
void FASTCALL call_3_2_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_9_5
#define init_3_2_9_5
void FASTCALL call_3_2_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_9_6
#define init_3_2_9_6
void FASTCALL call_3_2_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_9_7
#define init_3_2_9_7
void FASTCALL call_3_2_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_9_8
#define init_3_2_9_8
void FASTCALL call_3_2_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_9_9
#define init_3_2_9_9
void FASTCALL call_3_2_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_9_10
#define init_3_2_9_10
void FASTCALL call_3_2_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_9_11
#define init_3_2_9_11
void FASTCALL call_3_2_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_9_12
#define init_3_2_9_12
void FASTCALL call_3_2_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_9_13
#define init_3_2_9_13
void FASTCALL call_3_2_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_9_14
#define init_3_2_9_14
void FASTCALL call_3_2_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_9_15
#define init_3_2_9_15
void FASTCALL call_3_2_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_10_0
#define init_3_2_10_0
void FASTCALL call_3_2_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_10_1
#define init_3_2_10_1
void FASTCALL call_3_2_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_10_2
#define init_3_2_10_2
void FASTCALL call_3_2_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_10_3
#define init_3_2_10_3
void FASTCALL call_3_2_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_10_4
#define init_3_2_10_4
void FASTCALL call_3_2_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_10_5
#define init_3_2_10_5
void FASTCALL call_3_2_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_10_6
#define init_3_2_10_6
void FASTCALL call_3_2_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_10_7
#define init_3_2_10_7
void FASTCALL call_3_2_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_10_8
#define init_3_2_10_8
void FASTCALL call_3_2_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_10_9
#define init_3_2_10_9
void FASTCALL call_3_2_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_10_10
#define init_3_2_10_10
void FASTCALL call_3_2_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_10_11
#define init_3_2_10_11
void FASTCALL call_3_2_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_10_12
#define init_3_2_10_12
void FASTCALL call_3_2_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_10_13
#define init_3_2_10_13
void FASTCALL call_3_2_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_10_14
#define init_3_2_10_14
void FASTCALL call_3_2_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_10_15
#define init_3_2_10_15
void FASTCALL call_3_2_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_11_0
#define init_3_2_11_0
void FASTCALL call_3_2_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_11_1
#define init_3_2_11_1
void FASTCALL call_3_2_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_11_2
#define init_3_2_11_2
void FASTCALL call_3_2_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_11_3
#define init_3_2_11_3
void FASTCALL call_3_2_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_11_4
#define init_3_2_11_4
void FASTCALL call_3_2_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_11_5
#define init_3_2_11_5
void FASTCALL call_3_2_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_11_6
#define init_3_2_11_6
void FASTCALL call_3_2_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_11_7
#define init_3_2_11_7
void FASTCALL call_3_2_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_11_8
#define init_3_2_11_8
void FASTCALL call_3_2_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_11_9
#define init_3_2_11_9
void FASTCALL call_3_2_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_11_10
#define init_3_2_11_10
void FASTCALL call_3_2_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_11_11
#define init_3_2_11_11
void FASTCALL call_3_2_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_11_12
#define init_3_2_11_12
void FASTCALL call_3_2_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_11_13
#define init_3_2_11_13
void FASTCALL call_3_2_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_11_14
#define init_3_2_11_14
void FASTCALL call_3_2_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_11_15
#define init_3_2_11_15
void FASTCALL call_3_2_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_12_0
#define init_3_2_12_0
void FASTCALL call_3_2_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_12_1
#define init_3_2_12_1
void FASTCALL call_3_2_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_12_2
#define init_3_2_12_2
void FASTCALL call_3_2_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_12_3
#define init_3_2_12_3
void FASTCALL call_3_2_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_12_4
#define init_3_2_12_4
void FASTCALL call_3_2_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_12_5
#define init_3_2_12_5
void FASTCALL call_3_2_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_12_6
#define init_3_2_12_6
void FASTCALL call_3_2_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_12_7
#define init_3_2_12_7
void FASTCALL call_3_2_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_12_8
#define init_3_2_12_8
void FASTCALL call_3_2_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_12_9
#define init_3_2_12_9
void FASTCALL call_3_2_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_12_10
#define init_3_2_12_10
void FASTCALL call_3_2_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_12_11
#define init_3_2_12_11
void FASTCALL call_3_2_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_12_12
#define init_3_2_12_12
void FASTCALL call_3_2_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_12_13
#define init_3_2_12_13
void FASTCALL call_3_2_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_12_14
#define init_3_2_12_14
void FASTCALL call_3_2_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_12_15
#define init_3_2_12_15
void FASTCALL call_3_2_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_13_0
#define init_3_2_13_0
void FASTCALL call_3_2_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_13_1
#define init_3_2_13_1
void FASTCALL call_3_2_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_13_2
#define init_3_2_13_2
void FASTCALL call_3_2_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_13_3
#define init_3_2_13_3
void FASTCALL call_3_2_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_13_4
#define init_3_2_13_4
void FASTCALL call_3_2_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_13_5
#define init_3_2_13_5
void FASTCALL call_3_2_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_13_6
#define init_3_2_13_6
void FASTCALL call_3_2_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_13_7
#define init_3_2_13_7
void FASTCALL call_3_2_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_13_8
#define init_3_2_13_8
void FASTCALL call_3_2_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_13_9
#define init_3_2_13_9
void FASTCALL call_3_2_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_13_10
#define init_3_2_13_10
void FASTCALL call_3_2_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_13_11
#define init_3_2_13_11
void FASTCALL call_3_2_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_13_12
#define init_3_2_13_12
void FASTCALL call_3_2_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_13_13
#define init_3_2_13_13
void FASTCALL call_3_2_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_13_14
#define init_3_2_13_14
void FASTCALL call_3_2_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_13_15
#define init_3_2_13_15
void FASTCALL call_3_2_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_14_0
#define init_3_2_14_0
void FASTCALL call_3_2_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_14_1
#define init_3_2_14_1
void FASTCALL call_3_2_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_14_2
#define init_3_2_14_2
void FASTCALL call_3_2_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_14_3
#define init_3_2_14_3
void FASTCALL call_3_2_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_14_4
#define init_3_2_14_4
void FASTCALL call_3_2_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_14_5
#define init_3_2_14_5
void FASTCALL call_3_2_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_14_6
#define init_3_2_14_6
void FASTCALL call_3_2_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_14_7
#define init_3_2_14_7
void FASTCALL call_3_2_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_14_8
#define init_3_2_14_8
void FASTCALL call_3_2_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_14_9
#define init_3_2_14_9
void FASTCALL call_3_2_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_14_10
#define init_3_2_14_10
void FASTCALL call_3_2_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_14_11
#define init_3_2_14_11
void FASTCALL call_3_2_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_14_12
#define init_3_2_14_12
void FASTCALL call_3_2_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_14_13
#define init_3_2_14_13
void FASTCALL call_3_2_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_14_14
#define init_3_2_14_14
void FASTCALL call_3_2_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_14_15
#define init_3_2_14_15
void FASTCALL call_3_2_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_15_0
#define init_3_2_15_0
void FASTCALL call_3_2_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_15_1
#define init_3_2_15_1
void FASTCALL call_3_2_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_15_2
#define init_3_2_15_2
void FASTCALL call_3_2_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_15_3
#define init_3_2_15_3
void FASTCALL call_3_2_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_15_4
#define init_3_2_15_4
void FASTCALL call_3_2_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_15_5
#define init_3_2_15_5
void FASTCALL call_3_2_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_15_6
#define init_3_2_15_6
void FASTCALL call_3_2_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_15_7
#define init_3_2_15_7
void FASTCALL call_3_2_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_15_8
#define init_3_2_15_8
void FASTCALL call_3_2_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_15_9
#define init_3_2_15_9
void FASTCALL call_3_2_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_15_10
#define init_3_2_15_10
void FASTCALL call_3_2_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_15_11
#define init_3_2_15_11
void FASTCALL call_3_2_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_15_12
#define init_3_2_15_12
void FASTCALL call_3_2_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_15_13
#define init_3_2_15_13
void FASTCALL call_3_2_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_15_14
#define init_3_2_15_14
void FASTCALL call_3_2_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_2_15_15
#define init_3_2_15_15
void FASTCALL call_3_2_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_0_0
#define init_3_3_0_0
void FASTCALL call_3_3_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_0_1
#define init_3_3_0_1
void FASTCALL call_3_3_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_0_2
#define init_3_3_0_2
void FASTCALL call_3_3_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_0_3
#define init_3_3_0_3
void FASTCALL call_3_3_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_0_4
#define init_3_3_0_4
void FASTCALL call_3_3_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_0_5
#define init_3_3_0_5
void FASTCALL call_3_3_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_0_6
#define init_3_3_0_6
void FASTCALL call_3_3_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_0_7
#define init_3_3_0_7
void FASTCALL call_3_3_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_0_8
#define init_3_3_0_8
void FASTCALL call_3_3_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_0_9
#define init_3_3_0_9
void FASTCALL call_3_3_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_0_10
#define init_3_3_0_10
void FASTCALL call_3_3_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_0_11
#define init_3_3_0_11
void FASTCALL call_3_3_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_0_12
#define init_3_3_0_12
void FASTCALL call_3_3_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_0_13
#define init_3_3_0_13
void FASTCALL call_3_3_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_0_14
#define init_3_3_0_14
void FASTCALL call_3_3_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_0_15
#define init_3_3_0_15
void FASTCALL call_3_3_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_1_0
#define init_3_3_1_0
void FASTCALL call_3_3_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_1_1
#define init_3_3_1_1
void FASTCALL call_3_3_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_1_2
#define init_3_3_1_2
void FASTCALL call_3_3_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_1_3
#define init_3_3_1_3
void FASTCALL call_3_3_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_1_4
#define init_3_3_1_4
void FASTCALL call_3_3_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_1_5
#define init_3_3_1_5
void FASTCALL call_3_3_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_1_6
#define init_3_3_1_6
void FASTCALL call_3_3_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_1_7
#define init_3_3_1_7
void FASTCALL call_3_3_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_1_8
#define init_3_3_1_8
void FASTCALL call_3_3_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_1_9
#define init_3_3_1_9
void FASTCALL call_3_3_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_1_10
#define init_3_3_1_10
void FASTCALL call_3_3_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_1_11
#define init_3_3_1_11
void FASTCALL call_3_3_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_1_12
#define init_3_3_1_12
void FASTCALL call_3_3_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_1_13
#define init_3_3_1_13
void FASTCALL call_3_3_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_1_14
#define init_3_3_1_14
void FASTCALL call_3_3_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_1_15
#define init_3_3_1_15
void FASTCALL call_3_3_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_2_0
#define init_3_3_2_0
void FASTCALL call_3_3_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_2_1
#define init_3_3_2_1
void FASTCALL call_3_3_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_2_2
#define init_3_3_2_2
void FASTCALL call_3_3_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_2_3
#define init_3_3_2_3
void FASTCALL call_3_3_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_2_4
#define init_3_3_2_4
void FASTCALL call_3_3_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_2_5
#define init_3_3_2_5
void FASTCALL call_3_3_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_2_6
#define init_3_3_2_6
void FASTCALL call_3_3_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_2_7
#define init_3_3_2_7
void FASTCALL call_3_3_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_2_8
#define init_3_3_2_8
void FASTCALL call_3_3_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_2_9
#define init_3_3_2_9
void FASTCALL call_3_3_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_2_10
#define init_3_3_2_10
void FASTCALL call_3_3_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_2_11
#define init_3_3_2_11
void FASTCALL call_3_3_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_2_12
#define init_3_3_2_12
void FASTCALL call_3_3_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_2_13
#define init_3_3_2_13
void FASTCALL call_3_3_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_2_14
#define init_3_3_2_14
void FASTCALL call_3_3_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_2_15
#define init_3_3_2_15
void FASTCALL call_3_3_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_3_0
#define init_3_3_3_0
void FASTCALL call_3_3_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_3_1
#define init_3_3_3_1
void FASTCALL call_3_3_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_3_2
#define init_3_3_3_2
void FASTCALL call_3_3_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_3_3
#define init_3_3_3_3
void FASTCALL call_3_3_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_3_4
#define init_3_3_3_4
void FASTCALL call_3_3_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_3_5
#define init_3_3_3_5
void FASTCALL call_3_3_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_3_6
#define init_3_3_3_6
void FASTCALL call_3_3_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_3_7
#define init_3_3_3_7
void FASTCALL call_3_3_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_3_8
#define init_3_3_3_8
void FASTCALL call_3_3_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_3_9
#define init_3_3_3_9
void FASTCALL call_3_3_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_3_10
#define init_3_3_3_10
void FASTCALL call_3_3_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_3_11
#define init_3_3_3_11
void FASTCALL call_3_3_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_3_12
#define init_3_3_3_12
void FASTCALL call_3_3_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_3_13
#define init_3_3_3_13
void FASTCALL call_3_3_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_3_14
#define init_3_3_3_14
void FASTCALL call_3_3_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_3_15
#define init_3_3_3_15
void FASTCALL call_3_3_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_4_0
#define init_3_3_4_0
void FASTCALL call_3_3_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_4_1
#define init_3_3_4_1
void FASTCALL call_3_3_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_4_2
#define init_3_3_4_2
void FASTCALL call_3_3_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_4_3
#define init_3_3_4_3
void FASTCALL call_3_3_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_4_4
#define init_3_3_4_4
void FASTCALL call_3_3_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_4_5
#define init_3_3_4_5
void FASTCALL call_3_3_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_4_6
#define init_3_3_4_6
void FASTCALL call_3_3_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_4_7
#define init_3_3_4_7
void FASTCALL call_3_3_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_4_8
#define init_3_3_4_8
void FASTCALL call_3_3_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_4_9
#define init_3_3_4_9
void FASTCALL call_3_3_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_4_10
#define init_3_3_4_10
void FASTCALL call_3_3_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_4_11
#define init_3_3_4_11
void FASTCALL call_3_3_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_4_12
#define init_3_3_4_12
void FASTCALL call_3_3_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_4_13
#define init_3_3_4_13
void FASTCALL call_3_3_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_4_14
#define init_3_3_4_14
void FASTCALL call_3_3_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_4_15
#define init_3_3_4_15
void FASTCALL call_3_3_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_5_0
#define init_3_3_5_0
void FASTCALL call_3_3_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_5_1
#define init_3_3_5_1
void FASTCALL call_3_3_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_5_2
#define init_3_3_5_2
void FASTCALL call_3_3_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_5_3
#define init_3_3_5_3
void FASTCALL call_3_3_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_5_4
#define init_3_3_5_4
void FASTCALL call_3_3_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_5_5
#define init_3_3_5_5
void FASTCALL call_3_3_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_5_6
#define init_3_3_5_6
void FASTCALL call_3_3_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_5_7
#define init_3_3_5_7
void FASTCALL call_3_3_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_5_8
#define init_3_3_5_8
void FASTCALL call_3_3_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_5_9
#define init_3_3_5_9
void FASTCALL call_3_3_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_5_10
#define init_3_3_5_10
void FASTCALL call_3_3_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_5_11
#define init_3_3_5_11
void FASTCALL call_3_3_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_5_12
#define init_3_3_5_12
void FASTCALL call_3_3_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_5_13
#define init_3_3_5_13
void FASTCALL call_3_3_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_5_14
#define init_3_3_5_14
void FASTCALL call_3_3_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_5_15
#define init_3_3_5_15
void FASTCALL call_3_3_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_6_0
#define init_3_3_6_0
void FASTCALL call_3_3_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_6_1
#define init_3_3_6_1
void FASTCALL call_3_3_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_6_2
#define init_3_3_6_2
void FASTCALL call_3_3_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_6_3
#define init_3_3_6_3
void FASTCALL call_3_3_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_6_4
#define init_3_3_6_4
void FASTCALL call_3_3_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_6_5
#define init_3_3_6_5
void FASTCALL call_3_3_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_6_6
#define init_3_3_6_6
void FASTCALL call_3_3_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_6_7
#define init_3_3_6_7
void FASTCALL call_3_3_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_6_8
#define init_3_3_6_8
void FASTCALL call_3_3_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_6_9
#define init_3_3_6_9
void FASTCALL call_3_3_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_6_10
#define init_3_3_6_10
void FASTCALL call_3_3_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_6_11
#define init_3_3_6_11
void FASTCALL call_3_3_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_6_12
#define init_3_3_6_12
void FASTCALL call_3_3_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_6_13
#define init_3_3_6_13
void FASTCALL call_3_3_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_6_14
#define init_3_3_6_14
void FASTCALL call_3_3_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_6_15
#define init_3_3_6_15
void FASTCALL call_3_3_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_7_0
#define init_3_3_7_0
void FASTCALL call_3_3_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_7_1
#define init_3_3_7_1
void FASTCALL call_3_3_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_7_2
#define init_3_3_7_2
void FASTCALL call_3_3_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_7_3
#define init_3_3_7_3
void FASTCALL call_3_3_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_7_4
#define init_3_3_7_4
void FASTCALL call_3_3_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_7_5
#define init_3_3_7_5
void FASTCALL call_3_3_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_7_6
#define init_3_3_7_6
void FASTCALL call_3_3_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_7_7
#define init_3_3_7_7
void FASTCALL call_3_3_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_7_8
#define init_3_3_7_8
void FASTCALL call_3_3_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_7_9
#define init_3_3_7_9
void FASTCALL call_3_3_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_7_10
#define init_3_3_7_10
void FASTCALL call_3_3_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_7_11
#define init_3_3_7_11
void FASTCALL call_3_3_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_7_12
#define init_3_3_7_12
void FASTCALL call_3_3_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_7_13
#define init_3_3_7_13
void FASTCALL call_3_3_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_7_14
#define init_3_3_7_14
void FASTCALL call_3_3_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_7_15
#define init_3_3_7_15
void FASTCALL call_3_3_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_8_0
#define init_3_3_8_0
void FASTCALL call_3_3_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_8_1
#define init_3_3_8_1
void FASTCALL call_3_3_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_8_2
#define init_3_3_8_2
void FASTCALL call_3_3_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_8_3
#define init_3_3_8_3
void FASTCALL call_3_3_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_8_4
#define init_3_3_8_4
void FASTCALL call_3_3_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_8_5
#define init_3_3_8_5
void FASTCALL call_3_3_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_8_6
#define init_3_3_8_6
void FASTCALL call_3_3_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_8_7
#define init_3_3_8_7
void FASTCALL call_3_3_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_8_8
#define init_3_3_8_8
void FASTCALL call_3_3_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_8_9
#define init_3_3_8_9
void FASTCALL call_3_3_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_8_10
#define init_3_3_8_10
void FASTCALL call_3_3_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_8_11
#define init_3_3_8_11
void FASTCALL call_3_3_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_8_12
#define init_3_3_8_12
void FASTCALL call_3_3_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_8_13
#define init_3_3_8_13
void FASTCALL call_3_3_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_8_14
#define init_3_3_8_14
void FASTCALL call_3_3_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_8_15
#define init_3_3_8_15
void FASTCALL call_3_3_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_9_0
#define init_3_3_9_0
void FASTCALL call_3_3_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_9_1
#define init_3_3_9_1
void FASTCALL call_3_3_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_9_2
#define init_3_3_9_2
void FASTCALL call_3_3_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_9_3
#define init_3_3_9_3
void FASTCALL call_3_3_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_9_4
#define init_3_3_9_4
void FASTCALL call_3_3_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_9_5
#define init_3_3_9_5
void FASTCALL call_3_3_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_9_6
#define init_3_3_9_6
void FASTCALL call_3_3_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_9_7
#define init_3_3_9_7
void FASTCALL call_3_3_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_9_8
#define init_3_3_9_8
void FASTCALL call_3_3_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_9_9
#define init_3_3_9_9
void FASTCALL call_3_3_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_9_10
#define init_3_3_9_10
void FASTCALL call_3_3_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_9_11
#define init_3_3_9_11
void FASTCALL call_3_3_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_9_12
#define init_3_3_9_12
void FASTCALL call_3_3_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_9_13
#define init_3_3_9_13
void FASTCALL call_3_3_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_9_14
#define init_3_3_9_14
void FASTCALL call_3_3_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_9_15
#define init_3_3_9_15
void FASTCALL call_3_3_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_10_0
#define init_3_3_10_0
void FASTCALL call_3_3_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_10_1
#define init_3_3_10_1
void FASTCALL call_3_3_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_10_2
#define init_3_3_10_2
void FASTCALL call_3_3_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_10_3
#define init_3_3_10_3
void FASTCALL call_3_3_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_10_4
#define init_3_3_10_4
void FASTCALL call_3_3_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_10_5
#define init_3_3_10_5
void FASTCALL call_3_3_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_10_6
#define init_3_3_10_6
void FASTCALL call_3_3_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_10_7
#define init_3_3_10_7
void FASTCALL call_3_3_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_10_8
#define init_3_3_10_8
void FASTCALL call_3_3_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_10_9
#define init_3_3_10_9
void FASTCALL call_3_3_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_10_10
#define init_3_3_10_10
void FASTCALL call_3_3_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_10_11
#define init_3_3_10_11
void FASTCALL call_3_3_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_10_12
#define init_3_3_10_12
void FASTCALL call_3_3_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_10_13
#define init_3_3_10_13
void FASTCALL call_3_3_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_10_14
#define init_3_3_10_14
void FASTCALL call_3_3_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_10_15
#define init_3_3_10_15
void FASTCALL call_3_3_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_11_0
#define init_3_3_11_0
void FASTCALL call_3_3_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_11_1
#define init_3_3_11_1
void FASTCALL call_3_3_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_11_2
#define init_3_3_11_2
void FASTCALL call_3_3_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_11_3
#define init_3_3_11_3
void FASTCALL call_3_3_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_11_4
#define init_3_3_11_4
void FASTCALL call_3_3_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_11_5
#define init_3_3_11_5
void FASTCALL call_3_3_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_11_6
#define init_3_3_11_6
void FASTCALL call_3_3_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_11_7
#define init_3_3_11_7
void FASTCALL call_3_3_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_11_8
#define init_3_3_11_8
void FASTCALL call_3_3_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_11_9
#define init_3_3_11_9
void FASTCALL call_3_3_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_11_10
#define init_3_3_11_10
void FASTCALL call_3_3_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_11_11
#define init_3_3_11_11
void FASTCALL call_3_3_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_11_12
#define init_3_3_11_12
void FASTCALL call_3_3_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_11_13
#define init_3_3_11_13
void FASTCALL call_3_3_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_11_14
#define init_3_3_11_14
void FASTCALL call_3_3_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_11_15
#define init_3_3_11_15
void FASTCALL call_3_3_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_12_0
#define init_3_3_12_0
void FASTCALL call_3_3_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_12_1
#define init_3_3_12_1
void FASTCALL call_3_3_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_12_2
#define init_3_3_12_2
void FASTCALL call_3_3_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_12_3
#define init_3_3_12_3
void FASTCALL call_3_3_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_12_4
#define init_3_3_12_4
void FASTCALL call_3_3_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_12_5
#define init_3_3_12_5
void FASTCALL call_3_3_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_12_6
#define init_3_3_12_6
void FASTCALL call_3_3_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_12_7
#define init_3_3_12_7
void FASTCALL call_3_3_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_12_8
#define init_3_3_12_8
void FASTCALL call_3_3_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_12_9
#define init_3_3_12_9
void FASTCALL call_3_3_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_12_10
#define init_3_3_12_10
void FASTCALL call_3_3_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_12_11
#define init_3_3_12_11
void FASTCALL call_3_3_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_12_12
#define init_3_3_12_12
void FASTCALL call_3_3_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_12_13
#define init_3_3_12_13
void FASTCALL call_3_3_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_12_14
#define init_3_3_12_14
void FASTCALL call_3_3_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_12_15
#define init_3_3_12_15
void FASTCALL call_3_3_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_13_0
#define init_3_3_13_0
void FASTCALL call_3_3_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_13_1
#define init_3_3_13_1
void FASTCALL call_3_3_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_13_2
#define init_3_3_13_2
void FASTCALL call_3_3_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_13_3
#define init_3_3_13_3
void FASTCALL call_3_3_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_13_4
#define init_3_3_13_4
void FASTCALL call_3_3_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_13_5
#define init_3_3_13_5
void FASTCALL call_3_3_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_13_6
#define init_3_3_13_6
void FASTCALL call_3_3_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_13_7
#define init_3_3_13_7
void FASTCALL call_3_3_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_13_8
#define init_3_3_13_8
void FASTCALL call_3_3_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_13_9
#define init_3_3_13_9
void FASTCALL call_3_3_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_13_10
#define init_3_3_13_10
void FASTCALL call_3_3_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_13_11
#define init_3_3_13_11
void FASTCALL call_3_3_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_13_12
#define init_3_3_13_12
void FASTCALL call_3_3_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_13_13
#define init_3_3_13_13
void FASTCALL call_3_3_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_13_14
#define init_3_3_13_14
void FASTCALL call_3_3_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_13_15
#define init_3_3_13_15
void FASTCALL call_3_3_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_14_0
#define init_3_3_14_0
void FASTCALL call_3_3_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_14_1
#define init_3_3_14_1
void FASTCALL call_3_3_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_14_2
#define init_3_3_14_2
void FASTCALL call_3_3_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_14_3
#define init_3_3_14_3
void FASTCALL call_3_3_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_14_4
#define init_3_3_14_4
void FASTCALL call_3_3_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_14_5
#define init_3_3_14_5
void FASTCALL call_3_3_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_14_6
#define init_3_3_14_6
void FASTCALL call_3_3_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_14_7
#define init_3_3_14_7
void FASTCALL call_3_3_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_14_8
#define init_3_3_14_8
void FASTCALL call_3_3_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_14_9
#define init_3_3_14_9
void FASTCALL call_3_3_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_14_10
#define init_3_3_14_10
void FASTCALL call_3_3_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_14_11
#define init_3_3_14_11
void FASTCALL call_3_3_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_14_12
#define init_3_3_14_12
void FASTCALL call_3_3_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_14_13
#define init_3_3_14_13
void FASTCALL call_3_3_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_14_14
#define init_3_3_14_14
void FASTCALL call_3_3_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_14_15
#define init_3_3_14_15
void FASTCALL call_3_3_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_15_0
#define init_3_3_15_0
void FASTCALL call_3_3_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_15_1
#define init_3_3_15_1
void FASTCALL call_3_3_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_15_2
#define init_3_3_15_2
void FASTCALL call_3_3_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_15_3
#define init_3_3_15_3
void FASTCALL call_3_3_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_15_4
#define init_3_3_15_4
void FASTCALL call_3_3_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_15_5
#define init_3_3_15_5
void FASTCALL call_3_3_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_15_6
#define init_3_3_15_6
void FASTCALL call_3_3_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_15_7
#define init_3_3_15_7
void FASTCALL call_3_3_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_15_8
#define init_3_3_15_8
void FASTCALL call_3_3_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_15_9
#define init_3_3_15_9
void FASTCALL call_3_3_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_15_10
#define init_3_3_15_10
void FASTCALL call_3_3_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_15_11
#define init_3_3_15_11
void FASTCALL call_3_3_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_15_12
#define init_3_3_15_12
void FASTCALL call_3_3_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_15_13
#define init_3_3_15_13
void FASTCALL call_3_3_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_15_14
#define init_3_3_15_14
void FASTCALL call_3_3_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_3_15_15
#define init_3_3_15_15
void FASTCALL call_3_3_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_0_0
#define init_3_4_0_0
void FASTCALL call_3_4_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_0_1
#define init_3_4_0_1
void FASTCALL call_3_4_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_0_2
#define init_3_4_0_2
void FASTCALL call_3_4_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_0_3
#define init_3_4_0_3
void FASTCALL call_3_4_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_0_4
#define init_3_4_0_4
void FASTCALL call_3_4_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_0_5
#define init_3_4_0_5
void FASTCALL call_3_4_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_0_6
#define init_3_4_0_6
void FASTCALL call_3_4_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_0_7
#define init_3_4_0_7
void FASTCALL call_3_4_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_0_8
#define init_3_4_0_8
void FASTCALL call_3_4_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_0_9
#define init_3_4_0_9
void FASTCALL call_3_4_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_0_10
#define init_3_4_0_10
void FASTCALL call_3_4_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_0_11
#define init_3_4_0_11
void FASTCALL call_3_4_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_0_12
#define init_3_4_0_12
void FASTCALL call_3_4_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_0_13
#define init_3_4_0_13
void FASTCALL call_3_4_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_0_14
#define init_3_4_0_14
void FASTCALL call_3_4_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_0_15
#define init_3_4_0_15
void FASTCALL call_3_4_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_1_0
#define init_3_4_1_0
void FASTCALL call_3_4_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_1_1
#define init_3_4_1_1
void FASTCALL call_3_4_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_1_2
#define init_3_4_1_2
void FASTCALL call_3_4_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_1_3
#define init_3_4_1_3
void FASTCALL call_3_4_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_1_4
#define init_3_4_1_4
void FASTCALL call_3_4_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_1_5
#define init_3_4_1_5
void FASTCALL call_3_4_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_1_6
#define init_3_4_1_6
void FASTCALL call_3_4_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_1_7
#define init_3_4_1_7
void FASTCALL call_3_4_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_1_8
#define init_3_4_1_8
void FASTCALL call_3_4_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_1_9
#define init_3_4_1_9
void FASTCALL call_3_4_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_1_10
#define init_3_4_1_10
void FASTCALL call_3_4_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_1_11
#define init_3_4_1_11
void FASTCALL call_3_4_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_1_12
#define init_3_4_1_12
void FASTCALL call_3_4_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_1_13
#define init_3_4_1_13
void FASTCALL call_3_4_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_1_14
#define init_3_4_1_14
void FASTCALL call_3_4_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_1_15
#define init_3_4_1_15
void FASTCALL call_3_4_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_2_0
#define init_3_4_2_0
void FASTCALL call_3_4_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_2_1
#define init_3_4_2_1
void FASTCALL call_3_4_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_2_2
#define init_3_4_2_2
void FASTCALL call_3_4_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_2_3
#define init_3_4_2_3
void FASTCALL call_3_4_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_2_4
#define init_3_4_2_4
void FASTCALL call_3_4_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_2_5
#define init_3_4_2_5
void FASTCALL call_3_4_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_2_6
#define init_3_4_2_6
void FASTCALL call_3_4_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_2_7
#define init_3_4_2_7
void FASTCALL call_3_4_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_2_8
#define init_3_4_2_8
void FASTCALL call_3_4_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_2_9
#define init_3_4_2_9
void FASTCALL call_3_4_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_2_10
#define init_3_4_2_10
void FASTCALL call_3_4_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_2_11
#define init_3_4_2_11
void FASTCALL call_3_4_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_2_12
#define init_3_4_2_12
void FASTCALL call_3_4_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_2_13
#define init_3_4_2_13
void FASTCALL call_3_4_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_2_14
#define init_3_4_2_14
void FASTCALL call_3_4_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_2_15
#define init_3_4_2_15
void FASTCALL call_3_4_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_3_0
#define init_3_4_3_0
void FASTCALL call_3_4_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_3_1
#define init_3_4_3_1
void FASTCALL call_3_4_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_3_2
#define init_3_4_3_2
void FASTCALL call_3_4_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_3_3
#define init_3_4_3_3
void FASTCALL call_3_4_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_3_4
#define init_3_4_3_4
void FASTCALL call_3_4_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_3_5
#define init_3_4_3_5
void FASTCALL call_3_4_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_3_6
#define init_3_4_3_6
void FASTCALL call_3_4_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_3_7
#define init_3_4_3_7
void FASTCALL call_3_4_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_3_8
#define init_3_4_3_8
void FASTCALL call_3_4_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_3_9
#define init_3_4_3_9
void FASTCALL call_3_4_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_3_10
#define init_3_4_3_10
void FASTCALL call_3_4_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_3_11
#define init_3_4_3_11
void FASTCALL call_3_4_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_3_12
#define init_3_4_3_12
void FASTCALL call_3_4_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_3_13
#define init_3_4_3_13
void FASTCALL call_3_4_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_3_14
#define init_3_4_3_14
void FASTCALL call_3_4_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_3_15
#define init_3_4_3_15
void FASTCALL call_3_4_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_4_0
#define init_3_4_4_0
void FASTCALL call_3_4_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_4_1
#define init_3_4_4_1
void FASTCALL call_3_4_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_4_2
#define init_3_4_4_2
void FASTCALL call_3_4_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_4_3
#define init_3_4_4_3
void FASTCALL call_3_4_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_4_4
#define init_3_4_4_4
void FASTCALL call_3_4_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_4_5
#define init_3_4_4_5
void FASTCALL call_3_4_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_4_6
#define init_3_4_4_6
void FASTCALL call_3_4_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_4_7
#define init_3_4_4_7
void FASTCALL call_3_4_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_4_8
#define init_3_4_4_8
void FASTCALL call_3_4_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_4_9
#define init_3_4_4_9
void FASTCALL call_3_4_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_4_10
#define init_3_4_4_10
void FASTCALL call_3_4_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_4_11
#define init_3_4_4_11
void FASTCALL call_3_4_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_4_12
#define init_3_4_4_12
void FASTCALL call_3_4_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_4_13
#define init_3_4_4_13
void FASTCALL call_3_4_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_4_14
#define init_3_4_4_14
void FASTCALL call_3_4_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_4_15
#define init_3_4_4_15
void FASTCALL call_3_4_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_5_0
#define init_3_4_5_0
void FASTCALL call_3_4_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_5_1
#define init_3_4_5_1
void FASTCALL call_3_4_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_5_2
#define init_3_4_5_2
void FASTCALL call_3_4_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_5_3
#define init_3_4_5_3
void FASTCALL call_3_4_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_5_4
#define init_3_4_5_4
void FASTCALL call_3_4_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_5_5
#define init_3_4_5_5
void FASTCALL call_3_4_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_5_6
#define init_3_4_5_6
void FASTCALL call_3_4_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_5_7
#define init_3_4_5_7
void FASTCALL call_3_4_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_5_8
#define init_3_4_5_8
void FASTCALL call_3_4_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_5_9
#define init_3_4_5_9
void FASTCALL call_3_4_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_5_10
#define init_3_4_5_10
void FASTCALL call_3_4_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_5_11
#define init_3_4_5_11
void FASTCALL call_3_4_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_5_12
#define init_3_4_5_12
void FASTCALL call_3_4_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_5_13
#define init_3_4_5_13
void FASTCALL call_3_4_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_5_14
#define init_3_4_5_14
void FASTCALL call_3_4_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_5_15
#define init_3_4_5_15
void FASTCALL call_3_4_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_6_0
#define init_3_4_6_0
void FASTCALL call_3_4_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_6_1
#define init_3_4_6_1
void FASTCALL call_3_4_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_6_2
#define init_3_4_6_2
void FASTCALL call_3_4_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_6_3
#define init_3_4_6_3
void FASTCALL call_3_4_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_6_4
#define init_3_4_6_4
void FASTCALL call_3_4_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_6_5
#define init_3_4_6_5
void FASTCALL call_3_4_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_6_6
#define init_3_4_6_6
void FASTCALL call_3_4_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_6_7
#define init_3_4_6_7
void FASTCALL call_3_4_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_6_8
#define init_3_4_6_8
void FASTCALL call_3_4_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_6_9
#define init_3_4_6_9
void FASTCALL call_3_4_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_6_10
#define init_3_4_6_10
void FASTCALL call_3_4_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_6_11
#define init_3_4_6_11
void FASTCALL call_3_4_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_6_12
#define init_3_4_6_12
void FASTCALL call_3_4_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_6_13
#define init_3_4_6_13
void FASTCALL call_3_4_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_6_14
#define init_3_4_6_14
void FASTCALL call_3_4_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_6_15
#define init_3_4_6_15
void FASTCALL call_3_4_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_7_0
#define init_3_4_7_0
void FASTCALL call_3_4_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_7_1
#define init_3_4_7_1
void FASTCALL call_3_4_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_7_2
#define init_3_4_7_2
void FASTCALL call_3_4_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_7_3
#define init_3_4_7_3
void FASTCALL call_3_4_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_7_4
#define init_3_4_7_4
void FASTCALL call_3_4_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_7_5
#define init_3_4_7_5
void FASTCALL call_3_4_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_7_6
#define init_3_4_7_6
void FASTCALL call_3_4_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_7_7
#define init_3_4_7_7
void FASTCALL call_3_4_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_7_8
#define init_3_4_7_8
void FASTCALL call_3_4_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_7_9
#define init_3_4_7_9
void FASTCALL call_3_4_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_7_10
#define init_3_4_7_10
void FASTCALL call_3_4_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_7_11
#define init_3_4_7_11
void FASTCALL call_3_4_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_7_12
#define init_3_4_7_12
void FASTCALL call_3_4_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_7_13
#define init_3_4_7_13
void FASTCALL call_3_4_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_7_14
#define init_3_4_7_14
void FASTCALL call_3_4_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_7_15
#define init_3_4_7_15
void FASTCALL call_3_4_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_8_0
#define init_3_4_8_0
void FASTCALL call_3_4_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_8_1
#define init_3_4_8_1
void FASTCALL call_3_4_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_8_2
#define init_3_4_8_2
void FASTCALL call_3_4_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_8_3
#define init_3_4_8_3
void FASTCALL call_3_4_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_8_4
#define init_3_4_8_4
void FASTCALL call_3_4_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_8_5
#define init_3_4_8_5
void FASTCALL call_3_4_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_8_6
#define init_3_4_8_6
void FASTCALL call_3_4_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_8_7
#define init_3_4_8_7
void FASTCALL call_3_4_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_8_8
#define init_3_4_8_8
void FASTCALL call_3_4_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_8_9
#define init_3_4_8_9
void FASTCALL call_3_4_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_8_10
#define init_3_4_8_10
void FASTCALL call_3_4_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_8_11
#define init_3_4_8_11
void FASTCALL call_3_4_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_8_12
#define init_3_4_8_12
void FASTCALL call_3_4_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_8_13
#define init_3_4_8_13
void FASTCALL call_3_4_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_8_14
#define init_3_4_8_14
void FASTCALL call_3_4_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_8_15
#define init_3_4_8_15
void FASTCALL call_3_4_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_9_0
#define init_3_4_9_0
void FASTCALL call_3_4_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_9_1
#define init_3_4_9_1
void FASTCALL call_3_4_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_9_2
#define init_3_4_9_2
void FASTCALL call_3_4_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_9_3
#define init_3_4_9_3
void FASTCALL call_3_4_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_9_4
#define init_3_4_9_4
void FASTCALL call_3_4_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_9_5
#define init_3_4_9_5
void FASTCALL call_3_4_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_9_6
#define init_3_4_9_6
void FASTCALL call_3_4_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_9_7
#define init_3_4_9_7
void FASTCALL call_3_4_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_9_8
#define init_3_4_9_8
void FASTCALL call_3_4_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_9_9
#define init_3_4_9_9
void FASTCALL call_3_4_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_9_10
#define init_3_4_9_10
void FASTCALL call_3_4_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_9_11
#define init_3_4_9_11
void FASTCALL call_3_4_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_9_12
#define init_3_4_9_12
void FASTCALL call_3_4_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_9_13
#define init_3_4_9_13
void FASTCALL call_3_4_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_9_14
#define init_3_4_9_14
void FASTCALL call_3_4_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_9_15
#define init_3_4_9_15
void FASTCALL call_3_4_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_10_0
#define init_3_4_10_0
void FASTCALL call_3_4_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_10_1
#define init_3_4_10_1
void FASTCALL call_3_4_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_10_2
#define init_3_4_10_2
void FASTCALL call_3_4_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_10_3
#define init_3_4_10_3
void FASTCALL call_3_4_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_10_4
#define init_3_4_10_4
void FASTCALL call_3_4_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_10_5
#define init_3_4_10_5
void FASTCALL call_3_4_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_10_6
#define init_3_4_10_6
void FASTCALL call_3_4_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_10_7
#define init_3_4_10_7
void FASTCALL call_3_4_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_10_8
#define init_3_4_10_8
void FASTCALL call_3_4_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_10_9
#define init_3_4_10_9
void FASTCALL call_3_4_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_10_10
#define init_3_4_10_10
void FASTCALL call_3_4_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_10_11
#define init_3_4_10_11
void FASTCALL call_3_4_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_10_12
#define init_3_4_10_12
void FASTCALL call_3_4_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_10_13
#define init_3_4_10_13
void FASTCALL call_3_4_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_10_14
#define init_3_4_10_14
void FASTCALL call_3_4_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_10_15
#define init_3_4_10_15
void FASTCALL call_3_4_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_11_0
#define init_3_4_11_0
void FASTCALL call_3_4_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_11_1
#define init_3_4_11_1
void FASTCALL call_3_4_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_11_2
#define init_3_4_11_2
void FASTCALL call_3_4_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_11_3
#define init_3_4_11_3
void FASTCALL call_3_4_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_11_4
#define init_3_4_11_4
void FASTCALL call_3_4_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_11_5
#define init_3_4_11_5
void FASTCALL call_3_4_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_11_6
#define init_3_4_11_6
void FASTCALL call_3_4_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_11_7
#define init_3_4_11_7
void FASTCALL call_3_4_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_11_8
#define init_3_4_11_8
void FASTCALL call_3_4_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_11_9
#define init_3_4_11_9
void FASTCALL call_3_4_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_11_10
#define init_3_4_11_10
void FASTCALL call_3_4_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_11_11
#define init_3_4_11_11
void FASTCALL call_3_4_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_11_12
#define init_3_4_11_12
void FASTCALL call_3_4_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_11_13
#define init_3_4_11_13
void FASTCALL call_3_4_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_11_14
#define init_3_4_11_14
void FASTCALL call_3_4_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_11_15
#define init_3_4_11_15
void FASTCALL call_3_4_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_12_0
#define init_3_4_12_0
void FASTCALL call_3_4_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_12_1
#define init_3_4_12_1
void FASTCALL call_3_4_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_12_2
#define init_3_4_12_2
void FASTCALL call_3_4_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_12_3
#define init_3_4_12_3
void FASTCALL call_3_4_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_12_4
#define init_3_4_12_4
void FASTCALL call_3_4_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_12_5
#define init_3_4_12_5
void FASTCALL call_3_4_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_12_6
#define init_3_4_12_6
void FASTCALL call_3_4_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_12_7
#define init_3_4_12_7
void FASTCALL call_3_4_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_12_8
#define init_3_4_12_8
void FASTCALL call_3_4_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_12_9
#define init_3_4_12_9
void FASTCALL call_3_4_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_12_10
#define init_3_4_12_10
void FASTCALL call_3_4_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_12_11
#define init_3_4_12_11
void FASTCALL call_3_4_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_12_12
#define init_3_4_12_12
void FASTCALL call_3_4_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_12_13
#define init_3_4_12_13
void FASTCALL call_3_4_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_12_14
#define init_3_4_12_14
void FASTCALL call_3_4_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_12_15
#define init_3_4_12_15
void FASTCALL call_3_4_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_13_0
#define init_3_4_13_0
void FASTCALL call_3_4_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_13_1
#define init_3_4_13_1
void FASTCALL call_3_4_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_13_2
#define init_3_4_13_2
void FASTCALL call_3_4_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_13_3
#define init_3_4_13_3
void FASTCALL call_3_4_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_13_4
#define init_3_4_13_4
void FASTCALL call_3_4_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_13_5
#define init_3_4_13_5
void FASTCALL call_3_4_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_13_6
#define init_3_4_13_6
void FASTCALL call_3_4_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_13_7
#define init_3_4_13_7
void FASTCALL call_3_4_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_13_8
#define init_3_4_13_8
void FASTCALL call_3_4_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_13_9
#define init_3_4_13_9
void FASTCALL call_3_4_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_13_10
#define init_3_4_13_10
void FASTCALL call_3_4_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_13_11
#define init_3_4_13_11
void FASTCALL call_3_4_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_13_12
#define init_3_4_13_12
void FASTCALL call_3_4_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_13_13
#define init_3_4_13_13
void FASTCALL call_3_4_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_13_14
#define init_3_4_13_14
void FASTCALL call_3_4_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_13_15
#define init_3_4_13_15
void FASTCALL call_3_4_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_14_0
#define init_3_4_14_0
void FASTCALL call_3_4_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_14_1
#define init_3_4_14_1
void FASTCALL call_3_4_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_14_2
#define init_3_4_14_2
void FASTCALL call_3_4_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_14_3
#define init_3_4_14_3
void FASTCALL call_3_4_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_14_4
#define init_3_4_14_4
void FASTCALL call_3_4_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_14_5
#define init_3_4_14_5
void FASTCALL call_3_4_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_14_6
#define init_3_4_14_6
void FASTCALL call_3_4_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_14_7
#define init_3_4_14_7
void FASTCALL call_3_4_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_14_8
#define init_3_4_14_8
void FASTCALL call_3_4_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_14_9
#define init_3_4_14_9
void FASTCALL call_3_4_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_14_10
#define init_3_4_14_10
void FASTCALL call_3_4_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_14_11
#define init_3_4_14_11
void FASTCALL call_3_4_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_14_12
#define init_3_4_14_12
void FASTCALL call_3_4_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_14_13
#define init_3_4_14_13
void FASTCALL call_3_4_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_14_14
#define init_3_4_14_14
void FASTCALL call_3_4_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_14_15
#define init_3_4_14_15
void FASTCALL call_3_4_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_15_0
#define init_3_4_15_0
void FASTCALL call_3_4_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_15_1
#define init_3_4_15_1
void FASTCALL call_3_4_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_15_2
#define init_3_4_15_2
void FASTCALL call_3_4_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_15_3
#define init_3_4_15_3
void FASTCALL call_3_4_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_15_4
#define init_3_4_15_4
void FASTCALL call_3_4_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_15_5
#define init_3_4_15_5
void FASTCALL call_3_4_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_15_6
#define init_3_4_15_6
void FASTCALL call_3_4_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_15_7
#define init_3_4_15_7
void FASTCALL call_3_4_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_15_8
#define init_3_4_15_8
void FASTCALL call_3_4_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_15_9
#define init_3_4_15_9
void FASTCALL call_3_4_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_15_10
#define init_3_4_15_10
void FASTCALL call_3_4_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_15_11
#define init_3_4_15_11
void FASTCALL call_3_4_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_15_12
#define init_3_4_15_12
void FASTCALL call_3_4_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_15_13
#define init_3_4_15_13
void FASTCALL call_3_4_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_15_14
#define init_3_4_15_14
void FASTCALL call_3_4_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_4_15_15
#define init_3_4_15_15
void FASTCALL call_3_4_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_0_0
#define init_3_5_0_0
void FASTCALL call_3_5_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_0_1
#define init_3_5_0_1
void FASTCALL call_3_5_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_0_2
#define init_3_5_0_2
void FASTCALL call_3_5_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_0_3
#define init_3_5_0_3
void FASTCALL call_3_5_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_0_4
#define init_3_5_0_4
void FASTCALL call_3_5_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_0_5
#define init_3_5_0_5
void FASTCALL call_3_5_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_0_6
#define init_3_5_0_6
void FASTCALL call_3_5_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_0_7
#define init_3_5_0_7
void FASTCALL call_3_5_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_0_8
#define init_3_5_0_8
void FASTCALL call_3_5_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_0_9
#define init_3_5_0_9
void FASTCALL call_3_5_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_0_10
#define init_3_5_0_10
void FASTCALL call_3_5_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_0_11
#define init_3_5_0_11
void FASTCALL call_3_5_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_0_12
#define init_3_5_0_12
void FASTCALL call_3_5_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_0_13
#define init_3_5_0_13
void FASTCALL call_3_5_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_0_14
#define init_3_5_0_14
void FASTCALL call_3_5_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_0_15
#define init_3_5_0_15
void FASTCALL call_3_5_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_1_0
#define init_3_5_1_0
void FASTCALL call_3_5_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_1_1
#define init_3_5_1_1
void FASTCALL call_3_5_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_1_2
#define init_3_5_1_2
void FASTCALL call_3_5_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_1_3
#define init_3_5_1_3
void FASTCALL call_3_5_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_1_4
#define init_3_5_1_4
void FASTCALL call_3_5_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_1_5
#define init_3_5_1_5
void FASTCALL call_3_5_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_1_6
#define init_3_5_1_6
void FASTCALL call_3_5_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_1_7
#define init_3_5_1_7
void FASTCALL call_3_5_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_1_8
#define init_3_5_1_8
void FASTCALL call_3_5_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_1_9
#define init_3_5_1_9
void FASTCALL call_3_5_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_1_10
#define init_3_5_1_10
void FASTCALL call_3_5_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_1_11
#define init_3_5_1_11
void FASTCALL call_3_5_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_1_12
#define init_3_5_1_12
void FASTCALL call_3_5_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_1_13
#define init_3_5_1_13
void FASTCALL call_3_5_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_1_14
#define init_3_5_1_14
void FASTCALL call_3_5_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_1_15
#define init_3_5_1_15
void FASTCALL call_3_5_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_2_0
#define init_3_5_2_0
void FASTCALL call_3_5_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_2_1
#define init_3_5_2_1
void FASTCALL call_3_5_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_2_2
#define init_3_5_2_2
void FASTCALL call_3_5_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_2_3
#define init_3_5_2_3
void FASTCALL call_3_5_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_2_4
#define init_3_5_2_4
void FASTCALL call_3_5_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_2_5
#define init_3_5_2_5
void FASTCALL call_3_5_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_2_6
#define init_3_5_2_6
void FASTCALL call_3_5_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_2_7
#define init_3_5_2_7
void FASTCALL call_3_5_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_2_8
#define init_3_5_2_8
void FASTCALL call_3_5_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_2_9
#define init_3_5_2_9
void FASTCALL call_3_5_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_2_10
#define init_3_5_2_10
void FASTCALL call_3_5_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_2_11
#define init_3_5_2_11
void FASTCALL call_3_5_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_2_12
#define init_3_5_2_12
void FASTCALL call_3_5_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_2_13
#define init_3_5_2_13
void FASTCALL call_3_5_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_2_14
#define init_3_5_2_14
void FASTCALL call_3_5_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_2_15
#define init_3_5_2_15
void FASTCALL call_3_5_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_3_0
#define init_3_5_3_0
void FASTCALL call_3_5_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_3_1
#define init_3_5_3_1
void FASTCALL call_3_5_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_3_2
#define init_3_5_3_2
void FASTCALL call_3_5_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_3_3
#define init_3_5_3_3
void FASTCALL call_3_5_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_3_4
#define init_3_5_3_4
void FASTCALL call_3_5_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_3_5
#define init_3_5_3_5
void FASTCALL call_3_5_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_3_6
#define init_3_5_3_6
void FASTCALL call_3_5_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_3_7
#define init_3_5_3_7
void FASTCALL call_3_5_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_3_8
#define init_3_5_3_8
void FASTCALL call_3_5_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_3_9
#define init_3_5_3_9
void FASTCALL call_3_5_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_3_10
#define init_3_5_3_10
void FASTCALL call_3_5_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_3_11
#define init_3_5_3_11
void FASTCALL call_3_5_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_3_12
#define init_3_5_3_12
void FASTCALL call_3_5_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_3_13
#define init_3_5_3_13
void FASTCALL call_3_5_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_3_14
#define init_3_5_3_14
void FASTCALL call_3_5_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_3_15
#define init_3_5_3_15
void FASTCALL call_3_5_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_4_0
#define init_3_5_4_0
void FASTCALL call_3_5_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_4_1
#define init_3_5_4_1
void FASTCALL call_3_5_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_4_2
#define init_3_5_4_2
void FASTCALL call_3_5_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_4_3
#define init_3_5_4_3
void FASTCALL call_3_5_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_4_4
#define init_3_5_4_4
void FASTCALL call_3_5_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_4_5
#define init_3_5_4_5
void FASTCALL call_3_5_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_4_6
#define init_3_5_4_6
void FASTCALL call_3_5_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_4_7
#define init_3_5_4_7
void FASTCALL call_3_5_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_4_8
#define init_3_5_4_8
void FASTCALL call_3_5_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_4_9
#define init_3_5_4_9
void FASTCALL call_3_5_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_4_10
#define init_3_5_4_10
void FASTCALL call_3_5_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_4_11
#define init_3_5_4_11
void FASTCALL call_3_5_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_4_12
#define init_3_5_4_12
void FASTCALL call_3_5_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_4_13
#define init_3_5_4_13
void FASTCALL call_3_5_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_4_14
#define init_3_5_4_14
void FASTCALL call_3_5_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_4_15
#define init_3_5_4_15
void FASTCALL call_3_5_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_5_0
#define init_3_5_5_0
void FASTCALL call_3_5_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_5_1
#define init_3_5_5_1
void FASTCALL call_3_5_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_5_2
#define init_3_5_5_2
void FASTCALL call_3_5_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_5_3
#define init_3_5_5_3
void FASTCALL call_3_5_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_5_4
#define init_3_5_5_4
void FASTCALL call_3_5_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_5_5
#define init_3_5_5_5
void FASTCALL call_3_5_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_5_6
#define init_3_5_5_6
void FASTCALL call_3_5_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_5_7
#define init_3_5_5_7
void FASTCALL call_3_5_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_5_8
#define init_3_5_5_8
void FASTCALL call_3_5_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_5_9
#define init_3_5_5_9
void FASTCALL call_3_5_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_5_10
#define init_3_5_5_10
void FASTCALL call_3_5_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_5_11
#define init_3_5_5_11
void FASTCALL call_3_5_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_5_12
#define init_3_5_5_12
void FASTCALL call_3_5_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_5_13
#define init_3_5_5_13
void FASTCALL call_3_5_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_5_14
#define init_3_5_5_14
void FASTCALL call_3_5_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_5_15
#define init_3_5_5_15
void FASTCALL call_3_5_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_6_0
#define init_3_5_6_0
void FASTCALL call_3_5_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_6_1
#define init_3_5_6_1
void FASTCALL call_3_5_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_6_2
#define init_3_5_6_2
void FASTCALL call_3_5_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_6_3
#define init_3_5_6_3
void FASTCALL call_3_5_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_6_4
#define init_3_5_6_4
void FASTCALL call_3_5_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_6_5
#define init_3_5_6_5
void FASTCALL call_3_5_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_6_6
#define init_3_5_6_6
void FASTCALL call_3_5_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_6_7
#define init_3_5_6_7
void FASTCALL call_3_5_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_6_8
#define init_3_5_6_8
void FASTCALL call_3_5_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_6_9
#define init_3_5_6_9
void FASTCALL call_3_5_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_6_10
#define init_3_5_6_10
void FASTCALL call_3_5_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_6_11
#define init_3_5_6_11
void FASTCALL call_3_5_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_6_12
#define init_3_5_6_12
void FASTCALL call_3_5_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_6_13
#define init_3_5_6_13
void FASTCALL call_3_5_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_6_14
#define init_3_5_6_14
void FASTCALL call_3_5_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_6_15
#define init_3_5_6_15
void FASTCALL call_3_5_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_7_0
#define init_3_5_7_0
void FASTCALL call_3_5_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_7_1
#define init_3_5_7_1
void FASTCALL call_3_5_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_7_2
#define init_3_5_7_2
void FASTCALL call_3_5_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_7_3
#define init_3_5_7_3
void FASTCALL call_3_5_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_7_4
#define init_3_5_7_4
void FASTCALL call_3_5_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_7_5
#define init_3_5_7_5
void FASTCALL call_3_5_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_7_6
#define init_3_5_7_6
void FASTCALL call_3_5_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_7_7
#define init_3_5_7_7
void FASTCALL call_3_5_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_7_8
#define init_3_5_7_8
void FASTCALL call_3_5_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_7_9
#define init_3_5_7_9
void FASTCALL call_3_5_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_7_10
#define init_3_5_7_10
void FASTCALL call_3_5_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_7_11
#define init_3_5_7_11
void FASTCALL call_3_5_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_7_12
#define init_3_5_7_12
void FASTCALL call_3_5_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_7_13
#define init_3_5_7_13
void FASTCALL call_3_5_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_7_14
#define init_3_5_7_14
void FASTCALL call_3_5_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_7_15
#define init_3_5_7_15
void FASTCALL call_3_5_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_8_0
#define init_3_5_8_0
void FASTCALL call_3_5_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_8_1
#define init_3_5_8_1
void FASTCALL call_3_5_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_8_2
#define init_3_5_8_2
void FASTCALL call_3_5_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_8_3
#define init_3_5_8_3
void FASTCALL call_3_5_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_8_4
#define init_3_5_8_4
void FASTCALL call_3_5_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_8_5
#define init_3_5_8_5
void FASTCALL call_3_5_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_8_6
#define init_3_5_8_6
void FASTCALL call_3_5_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_8_7
#define init_3_5_8_7
void FASTCALL call_3_5_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_8_8
#define init_3_5_8_8
void FASTCALL call_3_5_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_8_9
#define init_3_5_8_9
void FASTCALL call_3_5_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_8_10
#define init_3_5_8_10
void FASTCALL call_3_5_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_8_11
#define init_3_5_8_11
void FASTCALL call_3_5_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_8_12
#define init_3_5_8_12
void FASTCALL call_3_5_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_8_13
#define init_3_5_8_13
void FASTCALL call_3_5_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_8_14
#define init_3_5_8_14
void FASTCALL call_3_5_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_8_15
#define init_3_5_8_15
void FASTCALL call_3_5_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_9_0
#define init_3_5_9_0
void FASTCALL call_3_5_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_9_1
#define init_3_5_9_1
void FASTCALL call_3_5_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_9_2
#define init_3_5_9_2
void FASTCALL call_3_5_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_9_3
#define init_3_5_9_3
void FASTCALL call_3_5_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_9_4
#define init_3_5_9_4
void FASTCALL call_3_5_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_9_5
#define init_3_5_9_5
void FASTCALL call_3_5_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_9_6
#define init_3_5_9_6
void FASTCALL call_3_5_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_9_7
#define init_3_5_9_7
void FASTCALL call_3_5_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_9_8
#define init_3_5_9_8
void FASTCALL call_3_5_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_9_9
#define init_3_5_9_9
void FASTCALL call_3_5_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_9_10
#define init_3_5_9_10
void FASTCALL call_3_5_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_9_11
#define init_3_5_9_11
void FASTCALL call_3_5_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_9_12
#define init_3_5_9_12
void FASTCALL call_3_5_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_9_13
#define init_3_5_9_13
void FASTCALL call_3_5_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_9_14
#define init_3_5_9_14
void FASTCALL call_3_5_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_9_15
#define init_3_5_9_15
void FASTCALL call_3_5_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_10_0
#define init_3_5_10_0
void FASTCALL call_3_5_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_10_1
#define init_3_5_10_1
void FASTCALL call_3_5_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_10_2
#define init_3_5_10_2
void FASTCALL call_3_5_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_10_3
#define init_3_5_10_3
void FASTCALL call_3_5_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_10_4
#define init_3_5_10_4
void FASTCALL call_3_5_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_10_5
#define init_3_5_10_5
void FASTCALL call_3_5_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_10_6
#define init_3_5_10_6
void FASTCALL call_3_5_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_10_7
#define init_3_5_10_7
void FASTCALL call_3_5_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_10_8
#define init_3_5_10_8
void FASTCALL call_3_5_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_10_9
#define init_3_5_10_9
void FASTCALL call_3_5_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_10_10
#define init_3_5_10_10
void FASTCALL call_3_5_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_10_11
#define init_3_5_10_11
void FASTCALL call_3_5_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_10_12
#define init_3_5_10_12
void FASTCALL call_3_5_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_10_13
#define init_3_5_10_13
void FASTCALL call_3_5_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_10_14
#define init_3_5_10_14
void FASTCALL call_3_5_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_10_15
#define init_3_5_10_15
void FASTCALL call_3_5_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_11_0
#define init_3_5_11_0
void FASTCALL call_3_5_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_11_1
#define init_3_5_11_1
void FASTCALL call_3_5_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_11_2
#define init_3_5_11_2
void FASTCALL call_3_5_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_11_3
#define init_3_5_11_3
void FASTCALL call_3_5_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_11_4
#define init_3_5_11_4
void FASTCALL call_3_5_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_11_5
#define init_3_5_11_5
void FASTCALL call_3_5_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_11_6
#define init_3_5_11_6
void FASTCALL call_3_5_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_11_7
#define init_3_5_11_7
void FASTCALL call_3_5_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_11_8
#define init_3_5_11_8
void FASTCALL call_3_5_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_11_9
#define init_3_5_11_9
void FASTCALL call_3_5_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_11_10
#define init_3_5_11_10
void FASTCALL call_3_5_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_11_11
#define init_3_5_11_11
void FASTCALL call_3_5_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_11_12
#define init_3_5_11_12
void FASTCALL call_3_5_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_11_13
#define init_3_5_11_13
void FASTCALL call_3_5_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_11_14
#define init_3_5_11_14
void FASTCALL call_3_5_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_11_15
#define init_3_5_11_15
void FASTCALL call_3_5_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_12_0
#define init_3_5_12_0
void FASTCALL call_3_5_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_12_1
#define init_3_5_12_1
void FASTCALL call_3_5_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_12_2
#define init_3_5_12_2
void FASTCALL call_3_5_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_12_3
#define init_3_5_12_3
void FASTCALL call_3_5_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_12_4
#define init_3_5_12_4
void FASTCALL call_3_5_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_12_5
#define init_3_5_12_5
void FASTCALL call_3_5_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_12_6
#define init_3_5_12_6
void FASTCALL call_3_5_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_12_7
#define init_3_5_12_7
void FASTCALL call_3_5_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_12_8
#define init_3_5_12_8
void FASTCALL call_3_5_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_12_9
#define init_3_5_12_9
void FASTCALL call_3_5_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_12_10
#define init_3_5_12_10
void FASTCALL call_3_5_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_12_11
#define init_3_5_12_11
void FASTCALL call_3_5_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_12_12
#define init_3_5_12_12
void FASTCALL call_3_5_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_12_13
#define init_3_5_12_13
void FASTCALL call_3_5_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_12_14
#define init_3_5_12_14
void FASTCALL call_3_5_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_12_15
#define init_3_5_12_15
void FASTCALL call_3_5_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_13_0
#define init_3_5_13_0
void FASTCALL call_3_5_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_13_1
#define init_3_5_13_1
void FASTCALL call_3_5_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_13_2
#define init_3_5_13_2
void FASTCALL call_3_5_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_13_3
#define init_3_5_13_3
void FASTCALL call_3_5_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_13_4
#define init_3_5_13_4
void FASTCALL call_3_5_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_13_5
#define init_3_5_13_5
void FASTCALL call_3_5_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_13_6
#define init_3_5_13_6
void FASTCALL call_3_5_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_13_7
#define init_3_5_13_7
void FASTCALL call_3_5_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_13_8
#define init_3_5_13_8
void FASTCALL call_3_5_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_13_9
#define init_3_5_13_9
void FASTCALL call_3_5_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_13_10
#define init_3_5_13_10
void FASTCALL call_3_5_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_13_11
#define init_3_5_13_11
void FASTCALL call_3_5_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_13_12
#define init_3_5_13_12
void FASTCALL call_3_5_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_13_13
#define init_3_5_13_13
void FASTCALL call_3_5_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_13_14
#define init_3_5_13_14
void FASTCALL call_3_5_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_13_15
#define init_3_5_13_15
void FASTCALL call_3_5_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_14_0
#define init_3_5_14_0
void FASTCALL call_3_5_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_14_1
#define init_3_5_14_1
void FASTCALL call_3_5_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_14_2
#define init_3_5_14_2
void FASTCALL call_3_5_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_14_3
#define init_3_5_14_3
void FASTCALL call_3_5_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_14_4
#define init_3_5_14_4
void FASTCALL call_3_5_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_14_5
#define init_3_5_14_5
void FASTCALL call_3_5_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_14_6
#define init_3_5_14_6
void FASTCALL call_3_5_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_14_7
#define init_3_5_14_7
void FASTCALL call_3_5_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_14_8
#define init_3_5_14_8
void FASTCALL call_3_5_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_14_9
#define init_3_5_14_9
void FASTCALL call_3_5_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_14_10
#define init_3_5_14_10
void FASTCALL call_3_5_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_14_11
#define init_3_5_14_11
void FASTCALL call_3_5_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_14_12
#define init_3_5_14_12
void FASTCALL call_3_5_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_14_13
#define init_3_5_14_13
void FASTCALL call_3_5_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_14_14
#define init_3_5_14_14
void FASTCALL call_3_5_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_14_15
#define init_3_5_14_15
void FASTCALL call_3_5_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_15_0
#define init_3_5_15_0
void FASTCALL call_3_5_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_15_1
#define init_3_5_15_1
void FASTCALL call_3_5_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_15_2
#define init_3_5_15_2
void FASTCALL call_3_5_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_15_3
#define init_3_5_15_3
void FASTCALL call_3_5_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_15_4
#define init_3_5_15_4
void FASTCALL call_3_5_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_15_5
#define init_3_5_15_5
void FASTCALL call_3_5_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_15_6
#define init_3_5_15_6
void FASTCALL call_3_5_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_15_7
#define init_3_5_15_7
void FASTCALL call_3_5_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_15_8
#define init_3_5_15_8
void FASTCALL call_3_5_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_15_9
#define init_3_5_15_9
void FASTCALL call_3_5_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_15_10
#define init_3_5_15_10
void FASTCALL call_3_5_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_15_11
#define init_3_5_15_11
void FASTCALL call_3_5_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_15_12
#define init_3_5_15_12
void FASTCALL call_3_5_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_15_13
#define init_3_5_15_13
void FASTCALL call_3_5_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_15_14
#define init_3_5_15_14
void FASTCALL call_3_5_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_5_15_15
#define init_3_5_15_15
void FASTCALL call_3_5_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_0_0
#define init_3_6_0_0
void FASTCALL call_3_6_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_0_1
#define init_3_6_0_1
void FASTCALL call_3_6_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_0_2
#define init_3_6_0_2
void FASTCALL call_3_6_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_0_3
#define init_3_6_0_3
void FASTCALL call_3_6_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_0_4
#define init_3_6_0_4
void FASTCALL call_3_6_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_0_5
#define init_3_6_0_5
void FASTCALL call_3_6_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_0_6
#define init_3_6_0_6
void FASTCALL call_3_6_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_0_7
#define init_3_6_0_7
void FASTCALL call_3_6_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_0_8
#define init_3_6_0_8
void FASTCALL call_3_6_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_0_9
#define init_3_6_0_9
void FASTCALL call_3_6_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_0_10
#define init_3_6_0_10
void FASTCALL call_3_6_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_0_11
#define init_3_6_0_11
void FASTCALL call_3_6_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_0_12
#define init_3_6_0_12
void FASTCALL call_3_6_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_0_13
#define init_3_6_0_13
void FASTCALL call_3_6_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_0_14
#define init_3_6_0_14
void FASTCALL call_3_6_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_0_15
#define init_3_6_0_15
void FASTCALL call_3_6_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_1_0
#define init_3_6_1_0
void FASTCALL call_3_6_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_1_1
#define init_3_6_1_1
void FASTCALL call_3_6_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_1_2
#define init_3_6_1_2
void FASTCALL call_3_6_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_1_3
#define init_3_6_1_3
void FASTCALL call_3_6_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_1_4
#define init_3_6_1_4
void FASTCALL call_3_6_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_1_5
#define init_3_6_1_5
void FASTCALL call_3_6_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_1_6
#define init_3_6_1_6
void FASTCALL call_3_6_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_1_7
#define init_3_6_1_7
void FASTCALL call_3_6_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_1_8
#define init_3_6_1_8
void FASTCALL call_3_6_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_1_9
#define init_3_6_1_9
void FASTCALL call_3_6_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_1_10
#define init_3_6_1_10
void FASTCALL call_3_6_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_1_11
#define init_3_6_1_11
void FASTCALL call_3_6_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_1_12
#define init_3_6_1_12
void FASTCALL call_3_6_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_1_13
#define init_3_6_1_13
void FASTCALL call_3_6_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_1_14
#define init_3_6_1_14
void FASTCALL call_3_6_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_1_15
#define init_3_6_1_15
void FASTCALL call_3_6_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_2_0
#define init_3_6_2_0
void FASTCALL call_3_6_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_2_1
#define init_3_6_2_1
void FASTCALL call_3_6_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_2_2
#define init_3_6_2_2
void FASTCALL call_3_6_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_2_3
#define init_3_6_2_3
void FASTCALL call_3_6_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_2_4
#define init_3_6_2_4
void FASTCALL call_3_6_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_2_5
#define init_3_6_2_5
void FASTCALL call_3_6_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_2_6
#define init_3_6_2_6
void FASTCALL call_3_6_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_2_7
#define init_3_6_2_7
void FASTCALL call_3_6_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_2_8
#define init_3_6_2_8
void FASTCALL call_3_6_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_2_9
#define init_3_6_2_9
void FASTCALL call_3_6_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_2_10
#define init_3_6_2_10
void FASTCALL call_3_6_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_2_11
#define init_3_6_2_11
void FASTCALL call_3_6_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_2_12
#define init_3_6_2_12
void FASTCALL call_3_6_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_2_13
#define init_3_6_2_13
void FASTCALL call_3_6_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_2_14
#define init_3_6_2_14
void FASTCALL call_3_6_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_2_15
#define init_3_6_2_15
void FASTCALL call_3_6_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_3_0
#define init_3_6_3_0
void FASTCALL call_3_6_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_3_1
#define init_3_6_3_1
void FASTCALL call_3_6_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_3_2
#define init_3_6_3_2
void FASTCALL call_3_6_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_3_3
#define init_3_6_3_3
void FASTCALL call_3_6_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_3_4
#define init_3_6_3_4
void FASTCALL call_3_6_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_3_5
#define init_3_6_3_5
void FASTCALL call_3_6_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_3_6
#define init_3_6_3_6
void FASTCALL call_3_6_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_3_7
#define init_3_6_3_7
void FASTCALL call_3_6_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_3_8
#define init_3_6_3_8
void FASTCALL call_3_6_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_3_9
#define init_3_6_3_9
void FASTCALL call_3_6_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_3_10
#define init_3_6_3_10
void FASTCALL call_3_6_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_3_11
#define init_3_6_3_11
void FASTCALL call_3_6_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_3_12
#define init_3_6_3_12
void FASTCALL call_3_6_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_3_13
#define init_3_6_3_13
void FASTCALL call_3_6_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_3_14
#define init_3_6_3_14
void FASTCALL call_3_6_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_3_15
#define init_3_6_3_15
void FASTCALL call_3_6_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_4_0
#define init_3_6_4_0
void FASTCALL call_3_6_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_4_1
#define init_3_6_4_1
void FASTCALL call_3_6_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_4_2
#define init_3_6_4_2
void FASTCALL call_3_6_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_4_3
#define init_3_6_4_3
void FASTCALL call_3_6_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_4_4
#define init_3_6_4_4
void FASTCALL call_3_6_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_4_5
#define init_3_6_4_5
void FASTCALL call_3_6_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_4_6
#define init_3_6_4_6
void FASTCALL call_3_6_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_4_7
#define init_3_6_4_7
void FASTCALL call_3_6_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_4_8
#define init_3_6_4_8
void FASTCALL call_3_6_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_4_9
#define init_3_6_4_9
void FASTCALL call_3_6_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_4_10
#define init_3_6_4_10
void FASTCALL call_3_6_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_4_11
#define init_3_6_4_11
void FASTCALL call_3_6_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_4_12
#define init_3_6_4_12
void FASTCALL call_3_6_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_4_13
#define init_3_6_4_13
void FASTCALL call_3_6_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_4_14
#define init_3_6_4_14
void FASTCALL call_3_6_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_4_15
#define init_3_6_4_15
void FASTCALL call_3_6_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_5_0
#define init_3_6_5_0
void FASTCALL call_3_6_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_5_1
#define init_3_6_5_1
void FASTCALL call_3_6_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_5_2
#define init_3_6_5_2
void FASTCALL call_3_6_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_5_3
#define init_3_6_5_3
void FASTCALL call_3_6_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_5_4
#define init_3_6_5_4
void FASTCALL call_3_6_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_5_5
#define init_3_6_5_5
void FASTCALL call_3_6_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_5_6
#define init_3_6_5_6
void FASTCALL call_3_6_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_5_7
#define init_3_6_5_7
void FASTCALL call_3_6_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_5_8
#define init_3_6_5_8
void FASTCALL call_3_6_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_5_9
#define init_3_6_5_9
void FASTCALL call_3_6_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_5_10
#define init_3_6_5_10
void FASTCALL call_3_6_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_5_11
#define init_3_6_5_11
void FASTCALL call_3_6_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_5_12
#define init_3_6_5_12
void FASTCALL call_3_6_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_5_13
#define init_3_6_5_13
void FASTCALL call_3_6_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_5_14
#define init_3_6_5_14
void FASTCALL call_3_6_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_5_15
#define init_3_6_5_15
void FASTCALL call_3_6_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_6_0
#define init_3_6_6_0
void FASTCALL call_3_6_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_6_1
#define init_3_6_6_1
void FASTCALL call_3_6_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_6_2
#define init_3_6_6_2
void FASTCALL call_3_6_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_6_3
#define init_3_6_6_3
void FASTCALL call_3_6_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_6_4
#define init_3_6_6_4
void FASTCALL call_3_6_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_6_5
#define init_3_6_6_5
void FASTCALL call_3_6_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_6_6
#define init_3_6_6_6
void FASTCALL call_3_6_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_6_7
#define init_3_6_6_7
void FASTCALL call_3_6_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_6_8
#define init_3_6_6_8
void FASTCALL call_3_6_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_6_9
#define init_3_6_6_9
void FASTCALL call_3_6_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_6_10
#define init_3_6_6_10
void FASTCALL call_3_6_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_6_11
#define init_3_6_6_11
void FASTCALL call_3_6_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_6_12
#define init_3_6_6_12
void FASTCALL call_3_6_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_6_13
#define init_3_6_6_13
void FASTCALL call_3_6_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_6_14
#define init_3_6_6_14
void FASTCALL call_3_6_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_6_15
#define init_3_6_6_15
void FASTCALL call_3_6_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_7_0
#define init_3_6_7_0
void FASTCALL call_3_6_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_7_1
#define init_3_6_7_1
void FASTCALL call_3_6_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_7_2
#define init_3_6_7_2
void FASTCALL call_3_6_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_7_3
#define init_3_6_7_3
void FASTCALL call_3_6_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_7_4
#define init_3_6_7_4
void FASTCALL call_3_6_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_7_5
#define init_3_6_7_5
void FASTCALL call_3_6_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_7_6
#define init_3_6_7_6
void FASTCALL call_3_6_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_7_7
#define init_3_6_7_7
void FASTCALL call_3_6_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_7_8
#define init_3_6_7_8
void FASTCALL call_3_6_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_7_9
#define init_3_6_7_9
void FASTCALL call_3_6_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_7_10
#define init_3_6_7_10
void FASTCALL call_3_6_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_7_11
#define init_3_6_7_11
void FASTCALL call_3_6_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_7_12
#define init_3_6_7_12
void FASTCALL call_3_6_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_7_13
#define init_3_6_7_13
void FASTCALL call_3_6_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_7_14
#define init_3_6_7_14
void FASTCALL call_3_6_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_7_15
#define init_3_6_7_15
void FASTCALL call_3_6_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_8_0
#define init_3_6_8_0
void FASTCALL call_3_6_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_8_1
#define init_3_6_8_1
void FASTCALL call_3_6_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_8_2
#define init_3_6_8_2
void FASTCALL call_3_6_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_8_3
#define init_3_6_8_3
void FASTCALL call_3_6_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_8_4
#define init_3_6_8_4
void FASTCALL call_3_6_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_8_5
#define init_3_6_8_5
void FASTCALL call_3_6_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_8_6
#define init_3_6_8_6
void FASTCALL call_3_6_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_8_7
#define init_3_6_8_7
void FASTCALL call_3_6_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_8_8
#define init_3_6_8_8
void FASTCALL call_3_6_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_8_9
#define init_3_6_8_9
void FASTCALL call_3_6_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_8_10
#define init_3_6_8_10
void FASTCALL call_3_6_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_8_11
#define init_3_6_8_11
void FASTCALL call_3_6_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_8_12
#define init_3_6_8_12
void FASTCALL call_3_6_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_8_13
#define init_3_6_8_13
void FASTCALL call_3_6_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_8_14
#define init_3_6_8_14
void FASTCALL call_3_6_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_8_15
#define init_3_6_8_15
void FASTCALL call_3_6_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_9_0
#define init_3_6_9_0
void FASTCALL call_3_6_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_9_1
#define init_3_6_9_1
void FASTCALL call_3_6_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_9_2
#define init_3_6_9_2
void FASTCALL call_3_6_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_9_3
#define init_3_6_9_3
void FASTCALL call_3_6_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_9_4
#define init_3_6_9_4
void FASTCALL call_3_6_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_9_5
#define init_3_6_9_5
void FASTCALL call_3_6_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_9_6
#define init_3_6_9_6
void FASTCALL call_3_6_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_9_7
#define init_3_6_9_7
void FASTCALL call_3_6_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_9_8
#define init_3_6_9_8
void FASTCALL call_3_6_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_9_9
#define init_3_6_9_9
void FASTCALL call_3_6_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_9_10
#define init_3_6_9_10
void FASTCALL call_3_6_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_9_11
#define init_3_6_9_11
void FASTCALL call_3_6_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_9_12
#define init_3_6_9_12
void FASTCALL call_3_6_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_9_13
#define init_3_6_9_13
void FASTCALL call_3_6_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_9_14
#define init_3_6_9_14
void FASTCALL call_3_6_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_9_15
#define init_3_6_9_15
void FASTCALL call_3_6_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_10_0
#define init_3_6_10_0
void FASTCALL call_3_6_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_10_1
#define init_3_6_10_1
void FASTCALL call_3_6_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_10_2
#define init_3_6_10_2
void FASTCALL call_3_6_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_10_3
#define init_3_6_10_3
void FASTCALL call_3_6_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_10_4
#define init_3_6_10_4
void FASTCALL call_3_6_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_10_5
#define init_3_6_10_5
void FASTCALL call_3_6_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_10_6
#define init_3_6_10_6
void FASTCALL call_3_6_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_10_7
#define init_3_6_10_7
void FASTCALL call_3_6_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_10_8
#define init_3_6_10_8
void FASTCALL call_3_6_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_10_9
#define init_3_6_10_9
void FASTCALL call_3_6_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_10_10
#define init_3_6_10_10
void FASTCALL call_3_6_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_10_11
#define init_3_6_10_11
void FASTCALL call_3_6_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_10_12
#define init_3_6_10_12
void FASTCALL call_3_6_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_10_13
#define init_3_6_10_13
void FASTCALL call_3_6_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_10_14
#define init_3_6_10_14
void FASTCALL call_3_6_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_10_15
#define init_3_6_10_15
void FASTCALL call_3_6_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_11_0
#define init_3_6_11_0
void FASTCALL call_3_6_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_11_1
#define init_3_6_11_1
void FASTCALL call_3_6_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_11_2
#define init_3_6_11_2
void FASTCALL call_3_6_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_11_3
#define init_3_6_11_3
void FASTCALL call_3_6_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_11_4
#define init_3_6_11_4
void FASTCALL call_3_6_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_11_5
#define init_3_6_11_5
void FASTCALL call_3_6_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_11_6
#define init_3_6_11_6
void FASTCALL call_3_6_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_11_7
#define init_3_6_11_7
void FASTCALL call_3_6_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_11_8
#define init_3_6_11_8
void FASTCALL call_3_6_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_11_9
#define init_3_6_11_9
void FASTCALL call_3_6_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_11_10
#define init_3_6_11_10
void FASTCALL call_3_6_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_11_11
#define init_3_6_11_11
void FASTCALL call_3_6_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_11_12
#define init_3_6_11_12
void FASTCALL call_3_6_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_11_13
#define init_3_6_11_13
void FASTCALL call_3_6_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_11_14
#define init_3_6_11_14
void FASTCALL call_3_6_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_11_15
#define init_3_6_11_15
void FASTCALL call_3_6_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_12_0
#define init_3_6_12_0
void FASTCALL call_3_6_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_12_1
#define init_3_6_12_1
void FASTCALL call_3_6_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_12_2
#define init_3_6_12_2
void FASTCALL call_3_6_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_12_3
#define init_3_6_12_3
void FASTCALL call_3_6_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_12_4
#define init_3_6_12_4
void FASTCALL call_3_6_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_12_5
#define init_3_6_12_5
void FASTCALL call_3_6_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_12_6
#define init_3_6_12_6
void FASTCALL call_3_6_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_12_7
#define init_3_6_12_7
void FASTCALL call_3_6_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_12_8
#define init_3_6_12_8
void FASTCALL call_3_6_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_12_9
#define init_3_6_12_9
void FASTCALL call_3_6_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_12_10
#define init_3_6_12_10
void FASTCALL call_3_6_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_12_11
#define init_3_6_12_11
void FASTCALL call_3_6_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_12_12
#define init_3_6_12_12
void FASTCALL call_3_6_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_12_13
#define init_3_6_12_13
void FASTCALL call_3_6_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_12_14
#define init_3_6_12_14
void FASTCALL call_3_6_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_12_15
#define init_3_6_12_15
void FASTCALL call_3_6_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_13_0
#define init_3_6_13_0
void FASTCALL call_3_6_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_13_1
#define init_3_6_13_1
void FASTCALL call_3_6_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_13_2
#define init_3_6_13_2
void FASTCALL call_3_6_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_13_3
#define init_3_6_13_3
void FASTCALL call_3_6_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_13_4
#define init_3_6_13_4
void FASTCALL call_3_6_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_13_5
#define init_3_6_13_5
void FASTCALL call_3_6_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_13_6
#define init_3_6_13_6
void FASTCALL call_3_6_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_13_7
#define init_3_6_13_7
void FASTCALL call_3_6_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_13_8
#define init_3_6_13_8
void FASTCALL call_3_6_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_13_9
#define init_3_6_13_9
void FASTCALL call_3_6_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_13_10
#define init_3_6_13_10
void FASTCALL call_3_6_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_13_11
#define init_3_6_13_11
void FASTCALL call_3_6_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_13_12
#define init_3_6_13_12
void FASTCALL call_3_6_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_13_13
#define init_3_6_13_13
void FASTCALL call_3_6_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_13_14
#define init_3_6_13_14
void FASTCALL call_3_6_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_13_15
#define init_3_6_13_15
void FASTCALL call_3_6_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_14_0
#define init_3_6_14_0
void FASTCALL call_3_6_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_14_1
#define init_3_6_14_1
void FASTCALL call_3_6_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_14_2
#define init_3_6_14_2
void FASTCALL call_3_6_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_14_3
#define init_3_6_14_3
void FASTCALL call_3_6_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_14_4
#define init_3_6_14_4
void FASTCALL call_3_6_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_14_5
#define init_3_6_14_5
void FASTCALL call_3_6_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_14_6
#define init_3_6_14_6
void FASTCALL call_3_6_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_14_7
#define init_3_6_14_7
void FASTCALL call_3_6_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_14_8
#define init_3_6_14_8
void FASTCALL call_3_6_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_14_9
#define init_3_6_14_9
void FASTCALL call_3_6_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_14_10
#define init_3_6_14_10
void FASTCALL call_3_6_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_14_11
#define init_3_6_14_11
void FASTCALL call_3_6_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_14_12
#define init_3_6_14_12
void FASTCALL call_3_6_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_14_13
#define init_3_6_14_13
void FASTCALL call_3_6_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_14_14
#define init_3_6_14_14
void FASTCALL call_3_6_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_14_15
#define init_3_6_14_15
void FASTCALL call_3_6_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_15_0
#define init_3_6_15_0
void FASTCALL call_3_6_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_15_1
#define init_3_6_15_1
void FASTCALL call_3_6_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_15_2
#define init_3_6_15_2
void FASTCALL call_3_6_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_15_3
#define init_3_6_15_3
void FASTCALL call_3_6_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_15_4
#define init_3_6_15_4
void FASTCALL call_3_6_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_15_5
#define init_3_6_15_5
void FASTCALL call_3_6_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_15_6
#define init_3_6_15_6
void FASTCALL call_3_6_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_15_7
#define init_3_6_15_7
void FASTCALL call_3_6_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_15_8
#define init_3_6_15_8
void FASTCALL call_3_6_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_15_9
#define init_3_6_15_9
void FASTCALL call_3_6_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_15_10
#define init_3_6_15_10
void FASTCALL call_3_6_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_15_11
#define init_3_6_15_11
void FASTCALL call_3_6_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_15_12
#define init_3_6_15_12
void FASTCALL call_3_6_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_15_13
#define init_3_6_15_13
void FASTCALL call_3_6_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_15_14
#define init_3_6_15_14
void FASTCALL call_3_6_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_6_15_15
#define init_3_6_15_15
void FASTCALL call_3_6_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_0_0
#define init_3_7_0_0
void FASTCALL call_3_7_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_0_1
#define init_3_7_0_1
void FASTCALL call_3_7_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_0_2
#define init_3_7_0_2
void FASTCALL call_3_7_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_0_3
#define init_3_7_0_3
void FASTCALL call_3_7_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_0_4
#define init_3_7_0_4
void FASTCALL call_3_7_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_0_5
#define init_3_7_0_5
void FASTCALL call_3_7_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_0_6
#define init_3_7_0_6
void FASTCALL call_3_7_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_0_7
#define init_3_7_0_7
void FASTCALL call_3_7_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_0_8
#define init_3_7_0_8
void FASTCALL call_3_7_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_0_9
#define init_3_7_0_9
void FASTCALL call_3_7_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_0_10
#define init_3_7_0_10
void FASTCALL call_3_7_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_0_11
#define init_3_7_0_11
void FASTCALL call_3_7_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_0_12
#define init_3_7_0_12
void FASTCALL call_3_7_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_0_13
#define init_3_7_0_13
void FASTCALL call_3_7_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_0_14
#define init_3_7_0_14
void FASTCALL call_3_7_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_0_15
#define init_3_7_0_15
void FASTCALL call_3_7_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_1_0
#define init_3_7_1_0
void FASTCALL call_3_7_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_1_1
#define init_3_7_1_1
void FASTCALL call_3_7_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_1_2
#define init_3_7_1_2
void FASTCALL call_3_7_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_1_3
#define init_3_7_1_3
void FASTCALL call_3_7_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_1_4
#define init_3_7_1_4
void FASTCALL call_3_7_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_1_5
#define init_3_7_1_5
void FASTCALL call_3_7_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_1_6
#define init_3_7_1_6
void FASTCALL call_3_7_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_1_7
#define init_3_7_1_7
void FASTCALL call_3_7_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_1_8
#define init_3_7_1_8
void FASTCALL call_3_7_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_1_9
#define init_3_7_1_9
void FASTCALL call_3_7_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_1_10
#define init_3_7_1_10
void FASTCALL call_3_7_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_1_11
#define init_3_7_1_11
void FASTCALL call_3_7_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_1_12
#define init_3_7_1_12
void FASTCALL call_3_7_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_1_13
#define init_3_7_1_13
void FASTCALL call_3_7_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_1_14
#define init_3_7_1_14
void FASTCALL call_3_7_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_1_15
#define init_3_7_1_15
void FASTCALL call_3_7_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_2_0
#define init_3_7_2_0
void FASTCALL call_3_7_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_2_1
#define init_3_7_2_1
void FASTCALL call_3_7_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_2_2
#define init_3_7_2_2
void FASTCALL call_3_7_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_2_3
#define init_3_7_2_3
void FASTCALL call_3_7_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_2_4
#define init_3_7_2_4
void FASTCALL call_3_7_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_2_5
#define init_3_7_2_5
void FASTCALL call_3_7_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_2_6
#define init_3_7_2_6
void FASTCALL call_3_7_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_2_7
#define init_3_7_2_7
void FASTCALL call_3_7_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_2_8
#define init_3_7_2_8
void FASTCALL call_3_7_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_2_9
#define init_3_7_2_9
void FASTCALL call_3_7_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_2_10
#define init_3_7_2_10
void FASTCALL call_3_7_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_2_11
#define init_3_7_2_11
void FASTCALL call_3_7_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_2_12
#define init_3_7_2_12
void FASTCALL call_3_7_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_2_13
#define init_3_7_2_13
void FASTCALL call_3_7_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_2_14
#define init_3_7_2_14
void FASTCALL call_3_7_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_2_15
#define init_3_7_2_15
void FASTCALL call_3_7_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_3_0
#define init_3_7_3_0
void FASTCALL call_3_7_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_3_1
#define init_3_7_3_1
void FASTCALL call_3_7_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_3_2
#define init_3_7_3_2
void FASTCALL call_3_7_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_3_3
#define init_3_7_3_3
void FASTCALL call_3_7_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_3_4
#define init_3_7_3_4
void FASTCALL call_3_7_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_3_5
#define init_3_7_3_5
void FASTCALL call_3_7_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_3_6
#define init_3_7_3_6
void FASTCALL call_3_7_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_3_7
#define init_3_7_3_7
void FASTCALL call_3_7_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_3_8
#define init_3_7_3_8
void FASTCALL call_3_7_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_3_9
#define init_3_7_3_9
void FASTCALL call_3_7_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_3_10
#define init_3_7_3_10
void FASTCALL call_3_7_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_3_11
#define init_3_7_3_11
void FASTCALL call_3_7_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_3_12
#define init_3_7_3_12
void FASTCALL call_3_7_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_3_13
#define init_3_7_3_13
void FASTCALL call_3_7_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_3_14
#define init_3_7_3_14
void FASTCALL call_3_7_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_3_15
#define init_3_7_3_15
void FASTCALL call_3_7_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_4_0
#define init_3_7_4_0
void FASTCALL call_3_7_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_4_1
#define init_3_7_4_1
void FASTCALL call_3_7_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_4_2
#define init_3_7_4_2
void FASTCALL call_3_7_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_4_3
#define init_3_7_4_3
void FASTCALL call_3_7_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_4_4
#define init_3_7_4_4
void FASTCALL call_3_7_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_4_5
#define init_3_7_4_5
void FASTCALL call_3_7_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_4_6
#define init_3_7_4_6
void FASTCALL call_3_7_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_4_7
#define init_3_7_4_7
void FASTCALL call_3_7_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_4_8
#define init_3_7_4_8
void FASTCALL call_3_7_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_4_9
#define init_3_7_4_9
void FASTCALL call_3_7_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_4_10
#define init_3_7_4_10
void FASTCALL call_3_7_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_4_11
#define init_3_7_4_11
void FASTCALL call_3_7_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_4_12
#define init_3_7_4_12
void FASTCALL call_3_7_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_4_13
#define init_3_7_4_13
void FASTCALL call_3_7_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_4_14
#define init_3_7_4_14
void FASTCALL call_3_7_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_4_15
#define init_3_7_4_15
void FASTCALL call_3_7_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_5_0
#define init_3_7_5_0
void FASTCALL call_3_7_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_5_1
#define init_3_7_5_1
void FASTCALL call_3_7_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_5_2
#define init_3_7_5_2
void FASTCALL call_3_7_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_5_3
#define init_3_7_5_3
void FASTCALL call_3_7_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_5_4
#define init_3_7_5_4
void FASTCALL call_3_7_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_5_5
#define init_3_7_5_5
void FASTCALL call_3_7_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_5_6
#define init_3_7_5_6
void FASTCALL call_3_7_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_5_7
#define init_3_7_5_7
void FASTCALL call_3_7_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_5_8
#define init_3_7_5_8
void FASTCALL call_3_7_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_5_9
#define init_3_7_5_9
void FASTCALL call_3_7_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_5_10
#define init_3_7_5_10
void FASTCALL call_3_7_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_5_11
#define init_3_7_5_11
void FASTCALL call_3_7_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_5_12
#define init_3_7_5_12
void FASTCALL call_3_7_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_5_13
#define init_3_7_5_13
void FASTCALL call_3_7_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_5_14
#define init_3_7_5_14
void FASTCALL call_3_7_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_5_15
#define init_3_7_5_15
void FASTCALL call_3_7_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_6_0
#define init_3_7_6_0
void FASTCALL call_3_7_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_6_1
#define init_3_7_6_1
void FASTCALL call_3_7_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_6_2
#define init_3_7_6_2
void FASTCALL call_3_7_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_6_3
#define init_3_7_6_3
void FASTCALL call_3_7_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_6_4
#define init_3_7_6_4
void FASTCALL call_3_7_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_6_5
#define init_3_7_6_5
void FASTCALL call_3_7_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_6_6
#define init_3_7_6_6
void FASTCALL call_3_7_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_6_7
#define init_3_7_6_7
void FASTCALL call_3_7_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_6_8
#define init_3_7_6_8
void FASTCALL call_3_7_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_6_9
#define init_3_7_6_9
void FASTCALL call_3_7_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_6_10
#define init_3_7_6_10
void FASTCALL call_3_7_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_6_11
#define init_3_7_6_11
void FASTCALL call_3_7_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_6_12
#define init_3_7_6_12
void FASTCALL call_3_7_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_6_13
#define init_3_7_6_13
void FASTCALL call_3_7_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_6_14
#define init_3_7_6_14
void FASTCALL call_3_7_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_6_15
#define init_3_7_6_15
void FASTCALL call_3_7_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_7_0
#define init_3_7_7_0
void FASTCALL call_3_7_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_7_1
#define init_3_7_7_1
void FASTCALL call_3_7_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_7_2
#define init_3_7_7_2
void FASTCALL call_3_7_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_7_3
#define init_3_7_7_3
void FASTCALL call_3_7_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_7_4
#define init_3_7_7_4
void FASTCALL call_3_7_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_7_5
#define init_3_7_7_5
void FASTCALL call_3_7_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_7_6
#define init_3_7_7_6
void FASTCALL call_3_7_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_7_7
#define init_3_7_7_7
void FASTCALL call_3_7_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_7_8
#define init_3_7_7_8
void FASTCALL call_3_7_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_7_9
#define init_3_7_7_9
void FASTCALL call_3_7_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_7_10
#define init_3_7_7_10
void FASTCALL call_3_7_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_7_11
#define init_3_7_7_11
void FASTCALL call_3_7_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_7_12
#define init_3_7_7_12
void FASTCALL call_3_7_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_7_13
#define init_3_7_7_13
void FASTCALL call_3_7_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_7_14
#define init_3_7_7_14
void FASTCALL call_3_7_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_7_15
#define init_3_7_7_15
void FASTCALL call_3_7_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_8_0
#define init_3_7_8_0
void FASTCALL call_3_7_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_8_1
#define init_3_7_8_1
void FASTCALL call_3_7_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_8_2
#define init_3_7_8_2
void FASTCALL call_3_7_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_8_3
#define init_3_7_8_3
void FASTCALL call_3_7_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_8_4
#define init_3_7_8_4
void FASTCALL call_3_7_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_8_5
#define init_3_7_8_5
void FASTCALL call_3_7_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_8_6
#define init_3_7_8_6
void FASTCALL call_3_7_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_8_7
#define init_3_7_8_7
void FASTCALL call_3_7_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_8_8
#define init_3_7_8_8
void FASTCALL call_3_7_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_8_9
#define init_3_7_8_9
void FASTCALL call_3_7_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_8_10
#define init_3_7_8_10
void FASTCALL call_3_7_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_8_11
#define init_3_7_8_11
void FASTCALL call_3_7_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_8_12
#define init_3_7_8_12
void FASTCALL call_3_7_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_8_13
#define init_3_7_8_13
void FASTCALL call_3_7_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_8_14
#define init_3_7_8_14
void FASTCALL call_3_7_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_8_15
#define init_3_7_8_15
void FASTCALL call_3_7_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_9_0
#define init_3_7_9_0
void FASTCALL call_3_7_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_9_1
#define init_3_7_9_1
void FASTCALL call_3_7_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_9_2
#define init_3_7_9_2
void FASTCALL call_3_7_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_9_3
#define init_3_7_9_3
void FASTCALL call_3_7_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_9_4
#define init_3_7_9_4
void FASTCALL call_3_7_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_9_5
#define init_3_7_9_5
void FASTCALL call_3_7_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_9_6
#define init_3_7_9_6
void FASTCALL call_3_7_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_9_7
#define init_3_7_9_7
void FASTCALL call_3_7_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_9_8
#define init_3_7_9_8
void FASTCALL call_3_7_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_9_9
#define init_3_7_9_9
void FASTCALL call_3_7_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_9_10
#define init_3_7_9_10
void FASTCALL call_3_7_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_9_11
#define init_3_7_9_11
void FASTCALL call_3_7_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_9_12
#define init_3_7_9_12
void FASTCALL call_3_7_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_9_13
#define init_3_7_9_13
void FASTCALL call_3_7_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_9_14
#define init_3_7_9_14
void FASTCALL call_3_7_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_9_15
#define init_3_7_9_15
void FASTCALL call_3_7_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_10_0
#define init_3_7_10_0
void FASTCALL call_3_7_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_10_1
#define init_3_7_10_1
void FASTCALL call_3_7_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_10_2
#define init_3_7_10_2
void FASTCALL call_3_7_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_10_3
#define init_3_7_10_3
void FASTCALL call_3_7_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_10_4
#define init_3_7_10_4
void FASTCALL call_3_7_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_10_5
#define init_3_7_10_5
void FASTCALL call_3_7_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_10_6
#define init_3_7_10_6
void FASTCALL call_3_7_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_10_7
#define init_3_7_10_7
void FASTCALL call_3_7_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_10_8
#define init_3_7_10_8
void FASTCALL call_3_7_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_10_9
#define init_3_7_10_9
void FASTCALL call_3_7_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_10_10
#define init_3_7_10_10
void FASTCALL call_3_7_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_10_11
#define init_3_7_10_11
void FASTCALL call_3_7_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_10_12
#define init_3_7_10_12
void FASTCALL call_3_7_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_10_13
#define init_3_7_10_13
void FASTCALL call_3_7_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_10_14
#define init_3_7_10_14
void FASTCALL call_3_7_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_10_15
#define init_3_7_10_15
void FASTCALL call_3_7_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_11_0
#define init_3_7_11_0
void FASTCALL call_3_7_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_11_1
#define init_3_7_11_1
void FASTCALL call_3_7_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_11_2
#define init_3_7_11_2
void FASTCALL call_3_7_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_11_3
#define init_3_7_11_3
void FASTCALL call_3_7_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_11_4
#define init_3_7_11_4
void FASTCALL call_3_7_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_11_5
#define init_3_7_11_5
void FASTCALL call_3_7_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_11_6
#define init_3_7_11_6
void FASTCALL call_3_7_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_11_7
#define init_3_7_11_7
void FASTCALL call_3_7_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_11_8
#define init_3_7_11_8
void FASTCALL call_3_7_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_11_9
#define init_3_7_11_9
void FASTCALL call_3_7_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_11_10
#define init_3_7_11_10
void FASTCALL call_3_7_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_11_11
#define init_3_7_11_11
void FASTCALL call_3_7_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_11_12
#define init_3_7_11_12
void FASTCALL call_3_7_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_11_13
#define init_3_7_11_13
void FASTCALL call_3_7_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_11_14
#define init_3_7_11_14
void FASTCALL call_3_7_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_11_15
#define init_3_7_11_15
void FASTCALL call_3_7_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_12_0
#define init_3_7_12_0
void FASTCALL call_3_7_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_12_1
#define init_3_7_12_1
void FASTCALL call_3_7_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_12_2
#define init_3_7_12_2
void FASTCALL call_3_7_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_12_3
#define init_3_7_12_3
void FASTCALL call_3_7_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_12_4
#define init_3_7_12_4
void FASTCALL call_3_7_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_12_5
#define init_3_7_12_5
void FASTCALL call_3_7_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_12_6
#define init_3_7_12_6
void FASTCALL call_3_7_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_12_7
#define init_3_7_12_7
void FASTCALL call_3_7_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_12_8
#define init_3_7_12_8
void FASTCALL call_3_7_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_12_9
#define init_3_7_12_9
void FASTCALL call_3_7_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_12_10
#define init_3_7_12_10
void FASTCALL call_3_7_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_12_11
#define init_3_7_12_11
void FASTCALL call_3_7_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_12_12
#define init_3_7_12_12
void FASTCALL call_3_7_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_12_13
#define init_3_7_12_13
void FASTCALL call_3_7_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_12_14
#define init_3_7_12_14
void FASTCALL call_3_7_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_12_15
#define init_3_7_12_15
void FASTCALL call_3_7_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_13_0
#define init_3_7_13_0
void FASTCALL call_3_7_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_13_1
#define init_3_7_13_1
void FASTCALL call_3_7_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_13_2
#define init_3_7_13_2
void FASTCALL call_3_7_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_13_3
#define init_3_7_13_3
void FASTCALL call_3_7_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_13_4
#define init_3_7_13_4
void FASTCALL call_3_7_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_13_5
#define init_3_7_13_5
void FASTCALL call_3_7_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_13_6
#define init_3_7_13_6
void FASTCALL call_3_7_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_13_7
#define init_3_7_13_7
void FASTCALL call_3_7_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_13_8
#define init_3_7_13_8
void FASTCALL call_3_7_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_13_9
#define init_3_7_13_9
void FASTCALL call_3_7_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_13_10
#define init_3_7_13_10
void FASTCALL call_3_7_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_13_11
#define init_3_7_13_11
void FASTCALL call_3_7_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_13_12
#define init_3_7_13_12
void FASTCALL call_3_7_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_13_13
#define init_3_7_13_13
void FASTCALL call_3_7_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_13_14
#define init_3_7_13_14
void FASTCALL call_3_7_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_13_15
#define init_3_7_13_15
void FASTCALL call_3_7_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_14_0
#define init_3_7_14_0
void FASTCALL call_3_7_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_14_1
#define init_3_7_14_1
void FASTCALL call_3_7_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_14_2
#define init_3_7_14_2
void FASTCALL call_3_7_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_14_3
#define init_3_7_14_3
void FASTCALL call_3_7_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_14_4
#define init_3_7_14_4
void FASTCALL call_3_7_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_14_5
#define init_3_7_14_5
void FASTCALL call_3_7_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_14_6
#define init_3_7_14_6
void FASTCALL call_3_7_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_14_7
#define init_3_7_14_7
void FASTCALL call_3_7_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_14_8
#define init_3_7_14_8
void FASTCALL call_3_7_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_14_9
#define init_3_7_14_9
void FASTCALL call_3_7_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_14_10
#define init_3_7_14_10
void FASTCALL call_3_7_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_14_11
#define init_3_7_14_11
void FASTCALL call_3_7_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_14_12
#define init_3_7_14_12
void FASTCALL call_3_7_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_14_13
#define init_3_7_14_13
void FASTCALL call_3_7_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_14_14
#define init_3_7_14_14
void FASTCALL call_3_7_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_14_15
#define init_3_7_14_15
void FASTCALL call_3_7_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_15_0
#define init_3_7_15_0
void FASTCALL call_3_7_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_15_1
#define init_3_7_15_1
void FASTCALL call_3_7_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_15_2
#define init_3_7_15_2
void FASTCALL call_3_7_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_15_3
#define init_3_7_15_3
void FASTCALL call_3_7_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_15_4
#define init_3_7_15_4
void FASTCALL call_3_7_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_15_5
#define init_3_7_15_5
void FASTCALL call_3_7_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_15_6
#define init_3_7_15_6
void FASTCALL call_3_7_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_15_7
#define init_3_7_15_7
void FASTCALL call_3_7_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_15_8
#define init_3_7_15_8
void FASTCALL call_3_7_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_15_9
#define init_3_7_15_9
void FASTCALL call_3_7_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_15_10
#define init_3_7_15_10
void FASTCALL call_3_7_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_15_11
#define init_3_7_15_11
void FASTCALL call_3_7_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_15_12
#define init_3_7_15_12
void FASTCALL call_3_7_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_15_13
#define init_3_7_15_13
void FASTCALL call_3_7_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_15_14
#define init_3_7_15_14
void FASTCALL call_3_7_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_7_15_15
#define init_3_7_15_15
void FASTCALL call_3_7_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_0_0
#define init_3_8_0_0
void FASTCALL call_3_8_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_0_1
#define init_3_8_0_1
void FASTCALL call_3_8_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_0_2
#define init_3_8_0_2
void FASTCALL call_3_8_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_0_3
#define init_3_8_0_3
void FASTCALL call_3_8_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_0_4
#define init_3_8_0_4
void FASTCALL call_3_8_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_0_5
#define init_3_8_0_5
void FASTCALL call_3_8_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_0_6
#define init_3_8_0_6
void FASTCALL call_3_8_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_0_7
#define init_3_8_0_7
void FASTCALL call_3_8_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_0_8
#define init_3_8_0_8
void FASTCALL call_3_8_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_0_9
#define init_3_8_0_9
void FASTCALL call_3_8_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_0_10
#define init_3_8_0_10
void FASTCALL call_3_8_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_0_11
#define init_3_8_0_11
void FASTCALL call_3_8_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_0_12
#define init_3_8_0_12
void FASTCALL call_3_8_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_0_13
#define init_3_8_0_13
void FASTCALL call_3_8_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_0_14
#define init_3_8_0_14
void FASTCALL call_3_8_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_0_15
#define init_3_8_0_15
void FASTCALL call_3_8_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_1_0
#define init_3_8_1_0
void FASTCALL call_3_8_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_1_1
#define init_3_8_1_1
void FASTCALL call_3_8_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_1_2
#define init_3_8_1_2
void FASTCALL call_3_8_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_1_3
#define init_3_8_1_3
void FASTCALL call_3_8_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_1_4
#define init_3_8_1_4
void FASTCALL call_3_8_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_1_5
#define init_3_8_1_5
void FASTCALL call_3_8_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_1_6
#define init_3_8_1_6
void FASTCALL call_3_8_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_1_7
#define init_3_8_1_7
void FASTCALL call_3_8_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_1_8
#define init_3_8_1_8
void FASTCALL call_3_8_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_1_9
#define init_3_8_1_9
void FASTCALL call_3_8_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_1_10
#define init_3_8_1_10
void FASTCALL call_3_8_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_1_11
#define init_3_8_1_11
void FASTCALL call_3_8_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_1_12
#define init_3_8_1_12
void FASTCALL call_3_8_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_1_13
#define init_3_8_1_13
void FASTCALL call_3_8_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_1_14
#define init_3_8_1_14
void FASTCALL call_3_8_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_1_15
#define init_3_8_1_15
void FASTCALL call_3_8_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_2_0
#define init_3_8_2_0
void FASTCALL call_3_8_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_2_1
#define init_3_8_2_1
void FASTCALL call_3_8_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_2_2
#define init_3_8_2_2
void FASTCALL call_3_8_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_2_3
#define init_3_8_2_3
void FASTCALL call_3_8_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_2_4
#define init_3_8_2_4
void FASTCALL call_3_8_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_2_5
#define init_3_8_2_5
void FASTCALL call_3_8_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_2_6
#define init_3_8_2_6
void FASTCALL call_3_8_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_2_7
#define init_3_8_2_7
void FASTCALL call_3_8_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_2_8
#define init_3_8_2_8
void FASTCALL call_3_8_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_2_9
#define init_3_8_2_9
void FASTCALL call_3_8_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_2_10
#define init_3_8_2_10
void FASTCALL call_3_8_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_2_11
#define init_3_8_2_11
void FASTCALL call_3_8_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_2_12
#define init_3_8_2_12
void FASTCALL call_3_8_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_2_13
#define init_3_8_2_13
void FASTCALL call_3_8_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_2_14
#define init_3_8_2_14
void FASTCALL call_3_8_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_2_15
#define init_3_8_2_15
void FASTCALL call_3_8_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_3_0
#define init_3_8_3_0
void FASTCALL call_3_8_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_3_1
#define init_3_8_3_1
void FASTCALL call_3_8_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_3_2
#define init_3_8_3_2
void FASTCALL call_3_8_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_3_3
#define init_3_8_3_3
void FASTCALL call_3_8_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_3_4
#define init_3_8_3_4
void FASTCALL call_3_8_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_3_5
#define init_3_8_3_5
void FASTCALL call_3_8_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_3_6
#define init_3_8_3_6
void FASTCALL call_3_8_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_3_7
#define init_3_8_3_7
void FASTCALL call_3_8_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_3_8
#define init_3_8_3_8
void FASTCALL call_3_8_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_3_9
#define init_3_8_3_9
void FASTCALL call_3_8_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_3_10
#define init_3_8_3_10
void FASTCALL call_3_8_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_3_11
#define init_3_8_3_11
void FASTCALL call_3_8_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_3_12
#define init_3_8_3_12
void FASTCALL call_3_8_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_3_13
#define init_3_8_3_13
void FASTCALL call_3_8_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_3_14
#define init_3_8_3_14
void FASTCALL call_3_8_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_3_15
#define init_3_8_3_15
void FASTCALL call_3_8_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_4_0
#define init_3_8_4_0
void FASTCALL call_3_8_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_4_1
#define init_3_8_4_1
void FASTCALL call_3_8_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_4_2
#define init_3_8_4_2
void FASTCALL call_3_8_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_4_3
#define init_3_8_4_3
void FASTCALL call_3_8_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_4_4
#define init_3_8_4_4
void FASTCALL call_3_8_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_4_5
#define init_3_8_4_5
void FASTCALL call_3_8_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_4_6
#define init_3_8_4_6
void FASTCALL call_3_8_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_4_7
#define init_3_8_4_7
void FASTCALL call_3_8_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_4_8
#define init_3_8_4_8
void FASTCALL call_3_8_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_4_9
#define init_3_8_4_9
void FASTCALL call_3_8_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_4_10
#define init_3_8_4_10
void FASTCALL call_3_8_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_4_11
#define init_3_8_4_11
void FASTCALL call_3_8_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_4_12
#define init_3_8_4_12
void FASTCALL call_3_8_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_4_13
#define init_3_8_4_13
void FASTCALL call_3_8_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_4_14
#define init_3_8_4_14
void FASTCALL call_3_8_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_4_15
#define init_3_8_4_15
void FASTCALL call_3_8_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_5_0
#define init_3_8_5_0
void FASTCALL call_3_8_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_5_1
#define init_3_8_5_1
void FASTCALL call_3_8_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_5_2
#define init_3_8_5_2
void FASTCALL call_3_8_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_5_3
#define init_3_8_5_3
void FASTCALL call_3_8_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_5_4
#define init_3_8_5_4
void FASTCALL call_3_8_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_5_5
#define init_3_8_5_5
void FASTCALL call_3_8_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_5_6
#define init_3_8_5_6
void FASTCALL call_3_8_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_5_7
#define init_3_8_5_7
void FASTCALL call_3_8_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_5_8
#define init_3_8_5_8
void FASTCALL call_3_8_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_5_9
#define init_3_8_5_9
void FASTCALL call_3_8_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_5_10
#define init_3_8_5_10
void FASTCALL call_3_8_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_5_11
#define init_3_8_5_11
void FASTCALL call_3_8_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_5_12
#define init_3_8_5_12
void FASTCALL call_3_8_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_5_13
#define init_3_8_5_13
void FASTCALL call_3_8_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_5_14
#define init_3_8_5_14
void FASTCALL call_3_8_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_5_15
#define init_3_8_5_15
void FASTCALL call_3_8_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_6_0
#define init_3_8_6_0
void FASTCALL call_3_8_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_6_1
#define init_3_8_6_1
void FASTCALL call_3_8_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_6_2
#define init_3_8_6_2
void FASTCALL call_3_8_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_6_3
#define init_3_8_6_3
void FASTCALL call_3_8_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_6_4
#define init_3_8_6_4
void FASTCALL call_3_8_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_6_5
#define init_3_8_6_5
void FASTCALL call_3_8_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_6_6
#define init_3_8_6_6
void FASTCALL call_3_8_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_6_7
#define init_3_8_6_7
void FASTCALL call_3_8_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_6_8
#define init_3_8_6_8
void FASTCALL call_3_8_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_6_9
#define init_3_8_6_9
void FASTCALL call_3_8_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_6_10
#define init_3_8_6_10
void FASTCALL call_3_8_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_6_11
#define init_3_8_6_11
void FASTCALL call_3_8_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_6_12
#define init_3_8_6_12
void FASTCALL call_3_8_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_6_13
#define init_3_8_6_13
void FASTCALL call_3_8_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_6_14
#define init_3_8_6_14
void FASTCALL call_3_8_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_6_15
#define init_3_8_6_15
void FASTCALL call_3_8_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_7_0
#define init_3_8_7_0
void FASTCALL call_3_8_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_7_1
#define init_3_8_7_1
void FASTCALL call_3_8_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_7_2
#define init_3_8_7_2
void FASTCALL call_3_8_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_7_3
#define init_3_8_7_3
void FASTCALL call_3_8_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_7_4
#define init_3_8_7_4
void FASTCALL call_3_8_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_7_5
#define init_3_8_7_5
void FASTCALL call_3_8_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_7_6
#define init_3_8_7_6
void FASTCALL call_3_8_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_7_7
#define init_3_8_7_7
void FASTCALL call_3_8_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_7_8
#define init_3_8_7_8
void FASTCALL call_3_8_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_7_9
#define init_3_8_7_9
void FASTCALL call_3_8_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_7_10
#define init_3_8_7_10
void FASTCALL call_3_8_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_7_11
#define init_3_8_7_11
void FASTCALL call_3_8_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_7_12
#define init_3_8_7_12
void FASTCALL call_3_8_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_7_13
#define init_3_8_7_13
void FASTCALL call_3_8_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_7_14
#define init_3_8_7_14
void FASTCALL call_3_8_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_7_15
#define init_3_8_7_15
void FASTCALL call_3_8_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_8_0
#define init_3_8_8_0
void FASTCALL call_3_8_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_8_1
#define init_3_8_8_1
void FASTCALL call_3_8_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_8_2
#define init_3_8_8_2
void FASTCALL call_3_8_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_8_3
#define init_3_8_8_3
void FASTCALL call_3_8_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_8_4
#define init_3_8_8_4
void FASTCALL call_3_8_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_8_5
#define init_3_8_8_5
void FASTCALL call_3_8_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_8_6
#define init_3_8_8_6
void FASTCALL call_3_8_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_8_7
#define init_3_8_8_7
void FASTCALL call_3_8_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_8_8
#define init_3_8_8_8
void FASTCALL call_3_8_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_8_9
#define init_3_8_8_9
void FASTCALL call_3_8_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_8_10
#define init_3_8_8_10
void FASTCALL call_3_8_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_8_11
#define init_3_8_8_11
void FASTCALL call_3_8_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_8_12
#define init_3_8_8_12
void FASTCALL call_3_8_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_8_13
#define init_3_8_8_13
void FASTCALL call_3_8_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_8_14
#define init_3_8_8_14
void FASTCALL call_3_8_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_8_15
#define init_3_8_8_15
void FASTCALL call_3_8_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_9_0
#define init_3_8_9_0
void FASTCALL call_3_8_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_9_1
#define init_3_8_9_1
void FASTCALL call_3_8_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_9_2
#define init_3_8_9_2
void FASTCALL call_3_8_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_9_3
#define init_3_8_9_3
void FASTCALL call_3_8_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_9_4
#define init_3_8_9_4
void FASTCALL call_3_8_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_9_5
#define init_3_8_9_5
void FASTCALL call_3_8_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_9_6
#define init_3_8_9_6
void FASTCALL call_3_8_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_9_7
#define init_3_8_9_7
void FASTCALL call_3_8_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_9_8
#define init_3_8_9_8
void FASTCALL call_3_8_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_9_9
#define init_3_8_9_9
void FASTCALL call_3_8_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_9_10
#define init_3_8_9_10
void FASTCALL call_3_8_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_9_11
#define init_3_8_9_11
void FASTCALL call_3_8_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_9_12
#define init_3_8_9_12
void FASTCALL call_3_8_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_9_13
#define init_3_8_9_13
void FASTCALL call_3_8_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_9_14
#define init_3_8_9_14
void FASTCALL call_3_8_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_9_15
#define init_3_8_9_15
void FASTCALL call_3_8_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_10_0
#define init_3_8_10_0
void FASTCALL call_3_8_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_10_1
#define init_3_8_10_1
void FASTCALL call_3_8_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_10_2
#define init_3_8_10_2
void FASTCALL call_3_8_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_10_3
#define init_3_8_10_3
void FASTCALL call_3_8_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_10_4
#define init_3_8_10_4
void FASTCALL call_3_8_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_10_5
#define init_3_8_10_5
void FASTCALL call_3_8_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_10_6
#define init_3_8_10_6
void FASTCALL call_3_8_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_10_7
#define init_3_8_10_7
void FASTCALL call_3_8_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_10_8
#define init_3_8_10_8
void FASTCALL call_3_8_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_10_9
#define init_3_8_10_9
void FASTCALL call_3_8_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_10_10
#define init_3_8_10_10
void FASTCALL call_3_8_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_10_11
#define init_3_8_10_11
void FASTCALL call_3_8_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_10_12
#define init_3_8_10_12
void FASTCALL call_3_8_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_10_13
#define init_3_8_10_13
void FASTCALL call_3_8_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_10_14
#define init_3_8_10_14
void FASTCALL call_3_8_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_10_15
#define init_3_8_10_15
void FASTCALL call_3_8_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_11_0
#define init_3_8_11_0
void FASTCALL call_3_8_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_11_1
#define init_3_8_11_1
void FASTCALL call_3_8_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_11_2
#define init_3_8_11_2
void FASTCALL call_3_8_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_11_3
#define init_3_8_11_3
void FASTCALL call_3_8_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_11_4
#define init_3_8_11_4
void FASTCALL call_3_8_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_11_5
#define init_3_8_11_5
void FASTCALL call_3_8_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_11_6
#define init_3_8_11_6
void FASTCALL call_3_8_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_11_7
#define init_3_8_11_7
void FASTCALL call_3_8_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_11_8
#define init_3_8_11_8
void FASTCALL call_3_8_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_11_9
#define init_3_8_11_9
void FASTCALL call_3_8_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_11_10
#define init_3_8_11_10
void FASTCALL call_3_8_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_11_11
#define init_3_8_11_11
void FASTCALL call_3_8_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_11_12
#define init_3_8_11_12
void FASTCALL call_3_8_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_11_13
#define init_3_8_11_13
void FASTCALL call_3_8_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_11_14
#define init_3_8_11_14
void FASTCALL call_3_8_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_11_15
#define init_3_8_11_15
void FASTCALL call_3_8_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_12_0
#define init_3_8_12_0
void FASTCALL call_3_8_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_12_1
#define init_3_8_12_1
void FASTCALL call_3_8_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_12_2
#define init_3_8_12_2
void FASTCALL call_3_8_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_12_3
#define init_3_8_12_3
void FASTCALL call_3_8_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_12_4
#define init_3_8_12_4
void FASTCALL call_3_8_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_12_5
#define init_3_8_12_5
void FASTCALL call_3_8_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_12_6
#define init_3_8_12_6
void FASTCALL call_3_8_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_12_7
#define init_3_8_12_7
void FASTCALL call_3_8_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_12_8
#define init_3_8_12_8
void FASTCALL call_3_8_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_12_9
#define init_3_8_12_9
void FASTCALL call_3_8_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_12_10
#define init_3_8_12_10
void FASTCALL call_3_8_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_12_11
#define init_3_8_12_11
void FASTCALL call_3_8_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_12_12
#define init_3_8_12_12
void FASTCALL call_3_8_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_12_13
#define init_3_8_12_13
void FASTCALL call_3_8_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_12_14
#define init_3_8_12_14
void FASTCALL call_3_8_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_12_15
#define init_3_8_12_15
void FASTCALL call_3_8_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_13_0
#define init_3_8_13_0
void FASTCALL call_3_8_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_13_1
#define init_3_8_13_1
void FASTCALL call_3_8_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_13_2
#define init_3_8_13_2
void FASTCALL call_3_8_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_13_3
#define init_3_8_13_3
void FASTCALL call_3_8_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_13_4
#define init_3_8_13_4
void FASTCALL call_3_8_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_13_5
#define init_3_8_13_5
void FASTCALL call_3_8_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_13_6
#define init_3_8_13_6
void FASTCALL call_3_8_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_13_7
#define init_3_8_13_7
void FASTCALL call_3_8_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_13_8
#define init_3_8_13_8
void FASTCALL call_3_8_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_13_9
#define init_3_8_13_9
void FASTCALL call_3_8_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_13_10
#define init_3_8_13_10
void FASTCALL call_3_8_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_13_11
#define init_3_8_13_11
void FASTCALL call_3_8_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_13_12
#define init_3_8_13_12
void FASTCALL call_3_8_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_13_13
#define init_3_8_13_13
void FASTCALL call_3_8_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_13_14
#define init_3_8_13_14
void FASTCALL call_3_8_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_13_15
#define init_3_8_13_15
void FASTCALL call_3_8_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_14_0
#define init_3_8_14_0
void FASTCALL call_3_8_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_14_1
#define init_3_8_14_1
void FASTCALL call_3_8_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_14_2
#define init_3_8_14_2
void FASTCALL call_3_8_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_14_3
#define init_3_8_14_3
void FASTCALL call_3_8_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_14_4
#define init_3_8_14_4
void FASTCALL call_3_8_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_14_5
#define init_3_8_14_5
void FASTCALL call_3_8_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_14_6
#define init_3_8_14_6
void FASTCALL call_3_8_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_14_7
#define init_3_8_14_7
void FASTCALL call_3_8_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_14_8
#define init_3_8_14_8
void FASTCALL call_3_8_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_14_9
#define init_3_8_14_9
void FASTCALL call_3_8_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_14_10
#define init_3_8_14_10
void FASTCALL call_3_8_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_14_11
#define init_3_8_14_11
void FASTCALL call_3_8_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_14_12
#define init_3_8_14_12
void FASTCALL call_3_8_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_14_13
#define init_3_8_14_13
void FASTCALL call_3_8_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_14_14
#define init_3_8_14_14
void FASTCALL call_3_8_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_14_15
#define init_3_8_14_15
void FASTCALL call_3_8_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_15_0
#define init_3_8_15_0
void FASTCALL call_3_8_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_15_1
#define init_3_8_15_1
void FASTCALL call_3_8_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_15_2
#define init_3_8_15_2
void FASTCALL call_3_8_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_15_3
#define init_3_8_15_3
void FASTCALL call_3_8_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_15_4
#define init_3_8_15_4
void FASTCALL call_3_8_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_15_5
#define init_3_8_15_5
void FASTCALL call_3_8_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_15_6
#define init_3_8_15_6
void FASTCALL call_3_8_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_15_7
#define init_3_8_15_7
void FASTCALL call_3_8_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_15_8
#define init_3_8_15_8
void FASTCALL call_3_8_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_15_9
#define init_3_8_15_9
void FASTCALL call_3_8_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_15_10
#define init_3_8_15_10
void FASTCALL call_3_8_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_15_11
#define init_3_8_15_11
void FASTCALL call_3_8_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_15_12
#define init_3_8_15_12
void FASTCALL call_3_8_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_15_13
#define init_3_8_15_13
void FASTCALL call_3_8_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_15_14
#define init_3_8_15_14
void FASTCALL call_3_8_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_8_15_15
#define init_3_8_15_15
void FASTCALL call_3_8_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_0_0
#define init_3_9_0_0
void FASTCALL call_3_9_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_0_1
#define init_3_9_0_1
void FASTCALL call_3_9_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_0_2
#define init_3_9_0_2
void FASTCALL call_3_9_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_0_3
#define init_3_9_0_3
void FASTCALL call_3_9_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_0_4
#define init_3_9_0_4
void FASTCALL call_3_9_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_0_5
#define init_3_9_0_5
void FASTCALL call_3_9_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_0_6
#define init_3_9_0_6
void FASTCALL call_3_9_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_0_7
#define init_3_9_0_7
void FASTCALL call_3_9_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_0_8
#define init_3_9_0_8
void FASTCALL call_3_9_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_0_9
#define init_3_9_0_9
void FASTCALL call_3_9_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_0_10
#define init_3_9_0_10
void FASTCALL call_3_9_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_0_11
#define init_3_9_0_11
void FASTCALL call_3_9_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_0_12
#define init_3_9_0_12
void FASTCALL call_3_9_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_0_13
#define init_3_9_0_13
void FASTCALL call_3_9_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_0_14
#define init_3_9_0_14
void FASTCALL call_3_9_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_0_15
#define init_3_9_0_15
void FASTCALL call_3_9_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_1_0
#define init_3_9_1_0
void FASTCALL call_3_9_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_1_1
#define init_3_9_1_1
void FASTCALL call_3_9_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_1_2
#define init_3_9_1_2
void FASTCALL call_3_9_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_1_3
#define init_3_9_1_3
void FASTCALL call_3_9_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_1_4
#define init_3_9_1_4
void FASTCALL call_3_9_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_1_5
#define init_3_9_1_5
void FASTCALL call_3_9_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_1_6
#define init_3_9_1_6
void FASTCALL call_3_9_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_1_7
#define init_3_9_1_7
void FASTCALL call_3_9_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_1_8
#define init_3_9_1_8
void FASTCALL call_3_9_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_1_9
#define init_3_9_1_9
void FASTCALL call_3_9_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_1_10
#define init_3_9_1_10
void FASTCALL call_3_9_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_1_11
#define init_3_9_1_11
void FASTCALL call_3_9_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_1_12
#define init_3_9_1_12
void FASTCALL call_3_9_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_1_13
#define init_3_9_1_13
void FASTCALL call_3_9_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_1_14
#define init_3_9_1_14
void FASTCALL call_3_9_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_1_15
#define init_3_9_1_15
void FASTCALL call_3_9_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_2_0
#define init_3_9_2_0
void FASTCALL call_3_9_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_2_1
#define init_3_9_2_1
void FASTCALL call_3_9_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_2_2
#define init_3_9_2_2
void FASTCALL call_3_9_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_2_3
#define init_3_9_2_3
void FASTCALL call_3_9_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_2_4
#define init_3_9_2_4
void FASTCALL call_3_9_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_2_5
#define init_3_9_2_5
void FASTCALL call_3_9_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_2_6
#define init_3_9_2_6
void FASTCALL call_3_9_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_2_7
#define init_3_9_2_7
void FASTCALL call_3_9_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_2_8
#define init_3_9_2_8
void FASTCALL call_3_9_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_2_9
#define init_3_9_2_9
void FASTCALL call_3_9_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_2_10
#define init_3_9_2_10
void FASTCALL call_3_9_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_2_11
#define init_3_9_2_11
void FASTCALL call_3_9_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_2_12
#define init_3_9_2_12
void FASTCALL call_3_9_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_2_13
#define init_3_9_2_13
void FASTCALL call_3_9_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_2_14
#define init_3_9_2_14
void FASTCALL call_3_9_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_2_15
#define init_3_9_2_15
void FASTCALL call_3_9_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_3_0
#define init_3_9_3_0
void FASTCALL call_3_9_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_3_1
#define init_3_9_3_1
void FASTCALL call_3_9_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_3_2
#define init_3_9_3_2
void FASTCALL call_3_9_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_3_3
#define init_3_9_3_3
void FASTCALL call_3_9_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_3_4
#define init_3_9_3_4
void FASTCALL call_3_9_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_3_5
#define init_3_9_3_5
void FASTCALL call_3_9_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_3_6
#define init_3_9_3_6
void FASTCALL call_3_9_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_3_7
#define init_3_9_3_7
void FASTCALL call_3_9_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_3_8
#define init_3_9_3_8
void FASTCALL call_3_9_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_3_9
#define init_3_9_3_9
void FASTCALL call_3_9_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_3_10
#define init_3_9_3_10
void FASTCALL call_3_9_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_3_11
#define init_3_9_3_11
void FASTCALL call_3_9_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_3_12
#define init_3_9_3_12
void FASTCALL call_3_9_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_3_13
#define init_3_9_3_13
void FASTCALL call_3_9_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_3_14
#define init_3_9_3_14
void FASTCALL call_3_9_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_3_15
#define init_3_9_3_15
void FASTCALL call_3_9_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_4_0
#define init_3_9_4_0
void FASTCALL call_3_9_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_4_1
#define init_3_9_4_1
void FASTCALL call_3_9_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_4_2
#define init_3_9_4_2
void FASTCALL call_3_9_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_4_3
#define init_3_9_4_3
void FASTCALL call_3_9_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_4_4
#define init_3_9_4_4
void FASTCALL call_3_9_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_4_5
#define init_3_9_4_5
void FASTCALL call_3_9_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_4_6
#define init_3_9_4_6
void FASTCALL call_3_9_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_4_7
#define init_3_9_4_7
void FASTCALL call_3_9_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_4_8
#define init_3_9_4_8
void FASTCALL call_3_9_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_4_9
#define init_3_9_4_9
void FASTCALL call_3_9_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_4_10
#define init_3_9_4_10
void FASTCALL call_3_9_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_4_11
#define init_3_9_4_11
void FASTCALL call_3_9_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_4_12
#define init_3_9_4_12
void FASTCALL call_3_9_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_4_13
#define init_3_9_4_13
void FASTCALL call_3_9_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_4_14
#define init_3_9_4_14
void FASTCALL call_3_9_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_4_15
#define init_3_9_4_15
void FASTCALL call_3_9_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_5_0
#define init_3_9_5_0
void FASTCALL call_3_9_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_5_1
#define init_3_9_5_1
void FASTCALL call_3_9_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_5_2
#define init_3_9_5_2
void FASTCALL call_3_9_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_5_3
#define init_3_9_5_3
void FASTCALL call_3_9_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_5_4
#define init_3_9_5_4
void FASTCALL call_3_9_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_5_5
#define init_3_9_5_5
void FASTCALL call_3_9_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_5_6
#define init_3_9_5_6
void FASTCALL call_3_9_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_5_7
#define init_3_9_5_7
void FASTCALL call_3_9_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_5_8
#define init_3_9_5_8
void FASTCALL call_3_9_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_5_9
#define init_3_9_5_9
void FASTCALL call_3_9_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_5_10
#define init_3_9_5_10
void FASTCALL call_3_9_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_5_11
#define init_3_9_5_11
void FASTCALL call_3_9_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_5_12
#define init_3_9_5_12
void FASTCALL call_3_9_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_5_13
#define init_3_9_5_13
void FASTCALL call_3_9_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_5_14
#define init_3_9_5_14
void FASTCALL call_3_9_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_5_15
#define init_3_9_5_15
void FASTCALL call_3_9_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_6_0
#define init_3_9_6_0
void FASTCALL call_3_9_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_6_1
#define init_3_9_6_1
void FASTCALL call_3_9_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_6_2
#define init_3_9_6_2
void FASTCALL call_3_9_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_6_3
#define init_3_9_6_3
void FASTCALL call_3_9_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_6_4
#define init_3_9_6_4
void FASTCALL call_3_9_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_6_5
#define init_3_9_6_5
void FASTCALL call_3_9_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_6_6
#define init_3_9_6_6
void FASTCALL call_3_9_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_6_7
#define init_3_9_6_7
void FASTCALL call_3_9_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_6_8
#define init_3_9_6_8
void FASTCALL call_3_9_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_6_9
#define init_3_9_6_9
void FASTCALL call_3_9_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_6_10
#define init_3_9_6_10
void FASTCALL call_3_9_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_6_11
#define init_3_9_6_11
void FASTCALL call_3_9_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_6_12
#define init_3_9_6_12
void FASTCALL call_3_9_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_6_13
#define init_3_9_6_13
void FASTCALL call_3_9_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_6_14
#define init_3_9_6_14
void FASTCALL call_3_9_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_6_15
#define init_3_9_6_15
void FASTCALL call_3_9_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_7_0
#define init_3_9_7_0
void FASTCALL call_3_9_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_7_1
#define init_3_9_7_1
void FASTCALL call_3_9_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_7_2
#define init_3_9_7_2
void FASTCALL call_3_9_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_7_3
#define init_3_9_7_3
void FASTCALL call_3_9_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_7_4
#define init_3_9_7_4
void FASTCALL call_3_9_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_7_5
#define init_3_9_7_5
void FASTCALL call_3_9_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_7_6
#define init_3_9_7_6
void FASTCALL call_3_9_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_7_7
#define init_3_9_7_7
void FASTCALL call_3_9_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_7_8
#define init_3_9_7_8
void FASTCALL call_3_9_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_7_9
#define init_3_9_7_9
void FASTCALL call_3_9_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_7_10
#define init_3_9_7_10
void FASTCALL call_3_9_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_7_11
#define init_3_9_7_11
void FASTCALL call_3_9_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_7_12
#define init_3_9_7_12
void FASTCALL call_3_9_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_7_13
#define init_3_9_7_13
void FASTCALL call_3_9_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_7_14
#define init_3_9_7_14
void FASTCALL call_3_9_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_7_15
#define init_3_9_7_15
void FASTCALL call_3_9_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_8_0
#define init_3_9_8_0
void FASTCALL call_3_9_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_8_1
#define init_3_9_8_1
void FASTCALL call_3_9_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_8_2
#define init_3_9_8_2
void FASTCALL call_3_9_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_8_3
#define init_3_9_8_3
void FASTCALL call_3_9_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_8_4
#define init_3_9_8_4
void FASTCALL call_3_9_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_8_5
#define init_3_9_8_5
void FASTCALL call_3_9_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_8_6
#define init_3_9_8_6
void FASTCALL call_3_9_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_8_7
#define init_3_9_8_7
void FASTCALL call_3_9_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_8_8
#define init_3_9_8_8
void FASTCALL call_3_9_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_8_9
#define init_3_9_8_9
void FASTCALL call_3_9_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_8_10
#define init_3_9_8_10
void FASTCALL call_3_9_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_8_11
#define init_3_9_8_11
void FASTCALL call_3_9_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_8_12
#define init_3_9_8_12
void FASTCALL call_3_9_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_8_13
#define init_3_9_8_13
void FASTCALL call_3_9_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_8_14
#define init_3_9_8_14
void FASTCALL call_3_9_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_8_15
#define init_3_9_8_15
void FASTCALL call_3_9_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_9_0
#define init_3_9_9_0
void FASTCALL call_3_9_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_9_1
#define init_3_9_9_1
void FASTCALL call_3_9_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_9_2
#define init_3_9_9_2
void FASTCALL call_3_9_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_9_3
#define init_3_9_9_3
void FASTCALL call_3_9_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_9_4
#define init_3_9_9_4
void FASTCALL call_3_9_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_9_5
#define init_3_9_9_5
void FASTCALL call_3_9_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_9_6
#define init_3_9_9_6
void FASTCALL call_3_9_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_9_7
#define init_3_9_9_7
void FASTCALL call_3_9_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_9_8
#define init_3_9_9_8
void FASTCALL call_3_9_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_9_9
#define init_3_9_9_9
void FASTCALL call_3_9_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_9_10
#define init_3_9_9_10
void FASTCALL call_3_9_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_9_11
#define init_3_9_9_11
void FASTCALL call_3_9_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_9_12
#define init_3_9_9_12
void FASTCALL call_3_9_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_9_13
#define init_3_9_9_13
void FASTCALL call_3_9_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_9_14
#define init_3_9_9_14
void FASTCALL call_3_9_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_9_15
#define init_3_9_9_15
void FASTCALL call_3_9_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_10_0
#define init_3_9_10_0
void FASTCALL call_3_9_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_10_1
#define init_3_9_10_1
void FASTCALL call_3_9_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_10_2
#define init_3_9_10_2
void FASTCALL call_3_9_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_10_3
#define init_3_9_10_3
void FASTCALL call_3_9_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_10_4
#define init_3_9_10_4
void FASTCALL call_3_9_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_10_5
#define init_3_9_10_5
void FASTCALL call_3_9_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_10_6
#define init_3_9_10_6
void FASTCALL call_3_9_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_10_7
#define init_3_9_10_7
void FASTCALL call_3_9_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_10_8
#define init_3_9_10_8
void FASTCALL call_3_9_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_10_9
#define init_3_9_10_9
void FASTCALL call_3_9_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_10_10
#define init_3_9_10_10
void FASTCALL call_3_9_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_10_11
#define init_3_9_10_11
void FASTCALL call_3_9_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_10_12
#define init_3_9_10_12
void FASTCALL call_3_9_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_10_13
#define init_3_9_10_13
void FASTCALL call_3_9_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_10_14
#define init_3_9_10_14
void FASTCALL call_3_9_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_10_15
#define init_3_9_10_15
void FASTCALL call_3_9_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_11_0
#define init_3_9_11_0
void FASTCALL call_3_9_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_11_1
#define init_3_9_11_1
void FASTCALL call_3_9_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_11_2
#define init_3_9_11_2
void FASTCALL call_3_9_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_11_3
#define init_3_9_11_3
void FASTCALL call_3_9_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_11_4
#define init_3_9_11_4
void FASTCALL call_3_9_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_11_5
#define init_3_9_11_5
void FASTCALL call_3_9_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_11_6
#define init_3_9_11_6
void FASTCALL call_3_9_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_11_7
#define init_3_9_11_7
void FASTCALL call_3_9_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_11_8
#define init_3_9_11_8
void FASTCALL call_3_9_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_11_9
#define init_3_9_11_9
void FASTCALL call_3_9_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_11_10
#define init_3_9_11_10
void FASTCALL call_3_9_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_11_11
#define init_3_9_11_11
void FASTCALL call_3_9_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_11_12
#define init_3_9_11_12
void FASTCALL call_3_9_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_11_13
#define init_3_9_11_13
void FASTCALL call_3_9_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_11_14
#define init_3_9_11_14
void FASTCALL call_3_9_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_11_15
#define init_3_9_11_15
void FASTCALL call_3_9_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_12_0
#define init_3_9_12_0
void FASTCALL call_3_9_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_12_1
#define init_3_9_12_1
void FASTCALL call_3_9_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_12_2
#define init_3_9_12_2
void FASTCALL call_3_9_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_12_3
#define init_3_9_12_3
void FASTCALL call_3_9_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_12_4
#define init_3_9_12_4
void FASTCALL call_3_9_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_12_5
#define init_3_9_12_5
void FASTCALL call_3_9_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_12_6
#define init_3_9_12_6
void FASTCALL call_3_9_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_12_7
#define init_3_9_12_7
void FASTCALL call_3_9_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_12_8
#define init_3_9_12_8
void FASTCALL call_3_9_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_12_9
#define init_3_9_12_9
void FASTCALL call_3_9_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_12_10
#define init_3_9_12_10
void FASTCALL call_3_9_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_12_11
#define init_3_9_12_11
void FASTCALL call_3_9_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_12_12
#define init_3_9_12_12
void FASTCALL call_3_9_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_12_13
#define init_3_9_12_13
void FASTCALL call_3_9_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_12_14
#define init_3_9_12_14
void FASTCALL call_3_9_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_12_15
#define init_3_9_12_15
void FASTCALL call_3_9_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_13_0
#define init_3_9_13_0
void FASTCALL call_3_9_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_13_1
#define init_3_9_13_1
void FASTCALL call_3_9_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_13_2
#define init_3_9_13_2
void FASTCALL call_3_9_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_13_3
#define init_3_9_13_3
void FASTCALL call_3_9_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_13_4
#define init_3_9_13_4
void FASTCALL call_3_9_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_13_5
#define init_3_9_13_5
void FASTCALL call_3_9_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_13_6
#define init_3_9_13_6
void FASTCALL call_3_9_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_13_7
#define init_3_9_13_7
void FASTCALL call_3_9_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_13_8
#define init_3_9_13_8
void FASTCALL call_3_9_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_13_9
#define init_3_9_13_9
void FASTCALL call_3_9_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_13_10
#define init_3_9_13_10
void FASTCALL call_3_9_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_13_11
#define init_3_9_13_11
void FASTCALL call_3_9_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_13_12
#define init_3_9_13_12
void FASTCALL call_3_9_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_13_13
#define init_3_9_13_13
void FASTCALL call_3_9_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_13_14
#define init_3_9_13_14
void FASTCALL call_3_9_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_13_15
#define init_3_9_13_15
void FASTCALL call_3_9_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_14_0
#define init_3_9_14_0
void FASTCALL call_3_9_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_14_1
#define init_3_9_14_1
void FASTCALL call_3_9_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_14_2
#define init_3_9_14_2
void FASTCALL call_3_9_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_14_3
#define init_3_9_14_3
void FASTCALL call_3_9_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_14_4
#define init_3_9_14_4
void FASTCALL call_3_9_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_14_5
#define init_3_9_14_5
void FASTCALL call_3_9_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_14_6
#define init_3_9_14_6
void FASTCALL call_3_9_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_14_7
#define init_3_9_14_7
void FASTCALL call_3_9_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_14_8
#define init_3_9_14_8
void FASTCALL call_3_9_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_14_9
#define init_3_9_14_9
void FASTCALL call_3_9_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_14_10
#define init_3_9_14_10
void FASTCALL call_3_9_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_14_11
#define init_3_9_14_11
void FASTCALL call_3_9_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_14_12
#define init_3_9_14_12
void FASTCALL call_3_9_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_14_13
#define init_3_9_14_13
void FASTCALL call_3_9_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_14_14
#define init_3_9_14_14
void FASTCALL call_3_9_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_14_15
#define init_3_9_14_15
void FASTCALL call_3_9_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_15_0
#define init_3_9_15_0
void FASTCALL call_3_9_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_15_1
#define init_3_9_15_1
void FASTCALL call_3_9_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_15_2
#define init_3_9_15_2
void FASTCALL call_3_9_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_15_3
#define init_3_9_15_3
void FASTCALL call_3_9_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_15_4
#define init_3_9_15_4
void FASTCALL call_3_9_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_15_5
#define init_3_9_15_5
void FASTCALL call_3_9_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_15_6
#define init_3_9_15_6
void FASTCALL call_3_9_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_15_7
#define init_3_9_15_7
void FASTCALL call_3_9_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_15_8
#define init_3_9_15_8
void FASTCALL call_3_9_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_15_9
#define init_3_9_15_9
void FASTCALL call_3_9_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_15_10
#define init_3_9_15_10
void FASTCALL call_3_9_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_15_11
#define init_3_9_15_11
void FASTCALL call_3_9_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_15_12
#define init_3_9_15_12
void FASTCALL call_3_9_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_15_13
#define init_3_9_15_13
void FASTCALL call_3_9_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_15_14
#define init_3_9_15_14
void FASTCALL call_3_9_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_9_15_15
#define init_3_9_15_15
void FASTCALL call_3_9_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_0_0
#define init_3_10_0_0
void FASTCALL call_3_10_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_0_1
#define init_3_10_0_1
void FASTCALL call_3_10_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_0_2
#define init_3_10_0_2
void FASTCALL call_3_10_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_0_3
#define init_3_10_0_3
void FASTCALL call_3_10_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_0_4
#define init_3_10_0_4
void FASTCALL call_3_10_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_0_5
#define init_3_10_0_5
void FASTCALL call_3_10_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_0_6
#define init_3_10_0_6
void FASTCALL call_3_10_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_0_7
#define init_3_10_0_7
void FASTCALL call_3_10_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_0_8
#define init_3_10_0_8
void FASTCALL call_3_10_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_0_9
#define init_3_10_0_9
void FASTCALL call_3_10_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_0_10
#define init_3_10_0_10
void FASTCALL call_3_10_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_0_11
#define init_3_10_0_11
void FASTCALL call_3_10_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_0_12
#define init_3_10_0_12
void FASTCALL call_3_10_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_0_13
#define init_3_10_0_13
void FASTCALL call_3_10_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_0_14
#define init_3_10_0_14
void FASTCALL call_3_10_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_0_15
#define init_3_10_0_15
void FASTCALL call_3_10_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_1_0
#define init_3_10_1_0
void FASTCALL call_3_10_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_1_1
#define init_3_10_1_1
void FASTCALL call_3_10_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_1_2
#define init_3_10_1_2
void FASTCALL call_3_10_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_1_3
#define init_3_10_1_3
void FASTCALL call_3_10_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_1_4
#define init_3_10_1_4
void FASTCALL call_3_10_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_1_5
#define init_3_10_1_5
void FASTCALL call_3_10_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_1_6
#define init_3_10_1_6
void FASTCALL call_3_10_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_1_7
#define init_3_10_1_7
void FASTCALL call_3_10_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_1_8
#define init_3_10_1_8
void FASTCALL call_3_10_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_1_9
#define init_3_10_1_9
void FASTCALL call_3_10_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_1_10
#define init_3_10_1_10
void FASTCALL call_3_10_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_1_11
#define init_3_10_1_11
void FASTCALL call_3_10_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_1_12
#define init_3_10_1_12
void FASTCALL call_3_10_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_1_13
#define init_3_10_1_13
void FASTCALL call_3_10_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_1_14
#define init_3_10_1_14
void FASTCALL call_3_10_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_1_15
#define init_3_10_1_15
void FASTCALL call_3_10_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_2_0
#define init_3_10_2_0
void FASTCALL call_3_10_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_2_1
#define init_3_10_2_1
void FASTCALL call_3_10_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_2_2
#define init_3_10_2_2
void FASTCALL call_3_10_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_2_3
#define init_3_10_2_3
void FASTCALL call_3_10_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_2_4
#define init_3_10_2_4
void FASTCALL call_3_10_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_2_5
#define init_3_10_2_5
void FASTCALL call_3_10_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_2_6
#define init_3_10_2_6
void FASTCALL call_3_10_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_2_7
#define init_3_10_2_7
void FASTCALL call_3_10_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_2_8
#define init_3_10_2_8
void FASTCALL call_3_10_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_2_9
#define init_3_10_2_9
void FASTCALL call_3_10_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_2_10
#define init_3_10_2_10
void FASTCALL call_3_10_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_2_11
#define init_3_10_2_11
void FASTCALL call_3_10_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_2_12
#define init_3_10_2_12
void FASTCALL call_3_10_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_2_13
#define init_3_10_2_13
void FASTCALL call_3_10_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_2_14
#define init_3_10_2_14
void FASTCALL call_3_10_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_2_15
#define init_3_10_2_15
void FASTCALL call_3_10_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_3_0
#define init_3_10_3_0
void FASTCALL call_3_10_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_3_1
#define init_3_10_3_1
void FASTCALL call_3_10_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_3_2
#define init_3_10_3_2
void FASTCALL call_3_10_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_3_3
#define init_3_10_3_3
void FASTCALL call_3_10_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_3_4
#define init_3_10_3_4
void FASTCALL call_3_10_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_3_5
#define init_3_10_3_5
void FASTCALL call_3_10_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_3_6
#define init_3_10_3_6
void FASTCALL call_3_10_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_3_7
#define init_3_10_3_7
void FASTCALL call_3_10_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_3_8
#define init_3_10_3_8
void FASTCALL call_3_10_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_3_9
#define init_3_10_3_9
void FASTCALL call_3_10_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_3_10
#define init_3_10_3_10
void FASTCALL call_3_10_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_3_11
#define init_3_10_3_11
void FASTCALL call_3_10_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_3_12
#define init_3_10_3_12
void FASTCALL call_3_10_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_3_13
#define init_3_10_3_13
void FASTCALL call_3_10_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_3_14
#define init_3_10_3_14
void FASTCALL call_3_10_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_3_15
#define init_3_10_3_15
void FASTCALL call_3_10_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_4_0
#define init_3_10_4_0
void FASTCALL call_3_10_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_4_1
#define init_3_10_4_1
void FASTCALL call_3_10_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_4_2
#define init_3_10_4_2
void FASTCALL call_3_10_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_4_3
#define init_3_10_4_3
void FASTCALL call_3_10_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_4_4
#define init_3_10_4_4
void FASTCALL call_3_10_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_4_5
#define init_3_10_4_5
void FASTCALL call_3_10_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_4_6
#define init_3_10_4_6
void FASTCALL call_3_10_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_4_7
#define init_3_10_4_7
void FASTCALL call_3_10_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_4_8
#define init_3_10_4_8
void FASTCALL call_3_10_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_4_9
#define init_3_10_4_9
void FASTCALL call_3_10_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_4_10
#define init_3_10_4_10
void FASTCALL call_3_10_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_4_11
#define init_3_10_4_11
void FASTCALL call_3_10_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_4_12
#define init_3_10_4_12
void FASTCALL call_3_10_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_4_13
#define init_3_10_4_13
void FASTCALL call_3_10_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_4_14
#define init_3_10_4_14
void FASTCALL call_3_10_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_4_15
#define init_3_10_4_15
void FASTCALL call_3_10_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_5_0
#define init_3_10_5_0
void FASTCALL call_3_10_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_5_1
#define init_3_10_5_1
void FASTCALL call_3_10_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_5_2
#define init_3_10_5_2
void FASTCALL call_3_10_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_5_3
#define init_3_10_5_3
void FASTCALL call_3_10_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_5_4
#define init_3_10_5_4
void FASTCALL call_3_10_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_5_5
#define init_3_10_5_5
void FASTCALL call_3_10_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_5_6
#define init_3_10_5_6
void FASTCALL call_3_10_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_5_7
#define init_3_10_5_7
void FASTCALL call_3_10_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_5_8
#define init_3_10_5_8
void FASTCALL call_3_10_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_5_9
#define init_3_10_5_9
void FASTCALL call_3_10_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_5_10
#define init_3_10_5_10
void FASTCALL call_3_10_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_5_11
#define init_3_10_5_11
void FASTCALL call_3_10_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_5_12
#define init_3_10_5_12
void FASTCALL call_3_10_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_5_13
#define init_3_10_5_13
void FASTCALL call_3_10_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_5_14
#define init_3_10_5_14
void FASTCALL call_3_10_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_5_15
#define init_3_10_5_15
void FASTCALL call_3_10_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_6_0
#define init_3_10_6_0
void FASTCALL call_3_10_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_6_1
#define init_3_10_6_1
void FASTCALL call_3_10_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_6_2
#define init_3_10_6_2
void FASTCALL call_3_10_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_6_3
#define init_3_10_6_3
void FASTCALL call_3_10_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_6_4
#define init_3_10_6_4
void FASTCALL call_3_10_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_6_5
#define init_3_10_6_5
void FASTCALL call_3_10_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_6_6
#define init_3_10_6_6
void FASTCALL call_3_10_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_6_7
#define init_3_10_6_7
void FASTCALL call_3_10_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_6_8
#define init_3_10_6_8
void FASTCALL call_3_10_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_6_9
#define init_3_10_6_9
void FASTCALL call_3_10_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_6_10
#define init_3_10_6_10
void FASTCALL call_3_10_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_6_11
#define init_3_10_6_11
void FASTCALL call_3_10_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_6_12
#define init_3_10_6_12
void FASTCALL call_3_10_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_6_13
#define init_3_10_6_13
void FASTCALL call_3_10_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_6_14
#define init_3_10_6_14
void FASTCALL call_3_10_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_6_15
#define init_3_10_6_15
void FASTCALL call_3_10_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_7_0
#define init_3_10_7_0
void FASTCALL call_3_10_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_7_1
#define init_3_10_7_1
void FASTCALL call_3_10_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_7_2
#define init_3_10_7_2
void FASTCALL call_3_10_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_7_3
#define init_3_10_7_3
void FASTCALL call_3_10_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_7_4
#define init_3_10_7_4
void FASTCALL call_3_10_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_7_5
#define init_3_10_7_5
void FASTCALL call_3_10_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_7_6
#define init_3_10_7_6
void FASTCALL call_3_10_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_7_7
#define init_3_10_7_7
void FASTCALL call_3_10_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_7_8
#define init_3_10_7_8
void FASTCALL call_3_10_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_7_9
#define init_3_10_7_9
void FASTCALL call_3_10_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_7_10
#define init_3_10_7_10
void FASTCALL call_3_10_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_7_11
#define init_3_10_7_11
void FASTCALL call_3_10_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_7_12
#define init_3_10_7_12
void FASTCALL call_3_10_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_7_13
#define init_3_10_7_13
void FASTCALL call_3_10_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_7_14
#define init_3_10_7_14
void FASTCALL call_3_10_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_7_15
#define init_3_10_7_15
void FASTCALL call_3_10_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_8_0
#define init_3_10_8_0
void FASTCALL call_3_10_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_8_1
#define init_3_10_8_1
void FASTCALL call_3_10_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_8_2
#define init_3_10_8_2
void FASTCALL call_3_10_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_8_3
#define init_3_10_8_3
void FASTCALL call_3_10_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_8_4
#define init_3_10_8_4
void FASTCALL call_3_10_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_8_5
#define init_3_10_8_5
void FASTCALL call_3_10_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_8_6
#define init_3_10_8_6
void FASTCALL call_3_10_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_8_7
#define init_3_10_8_7
void FASTCALL call_3_10_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_8_8
#define init_3_10_8_8
void FASTCALL call_3_10_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_8_9
#define init_3_10_8_9
void FASTCALL call_3_10_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_8_10
#define init_3_10_8_10
void FASTCALL call_3_10_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_8_11
#define init_3_10_8_11
void FASTCALL call_3_10_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_8_12
#define init_3_10_8_12
void FASTCALL call_3_10_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_8_13
#define init_3_10_8_13
void FASTCALL call_3_10_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_8_14
#define init_3_10_8_14
void FASTCALL call_3_10_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_8_15
#define init_3_10_8_15
void FASTCALL call_3_10_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_9_0
#define init_3_10_9_0
void FASTCALL call_3_10_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_9_1
#define init_3_10_9_1
void FASTCALL call_3_10_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_9_2
#define init_3_10_9_2
void FASTCALL call_3_10_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_9_3
#define init_3_10_9_3
void FASTCALL call_3_10_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_9_4
#define init_3_10_9_4
void FASTCALL call_3_10_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_9_5
#define init_3_10_9_5
void FASTCALL call_3_10_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_9_6
#define init_3_10_9_6
void FASTCALL call_3_10_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_9_7
#define init_3_10_9_7
void FASTCALL call_3_10_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_9_8
#define init_3_10_9_8
void FASTCALL call_3_10_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_9_9
#define init_3_10_9_9
void FASTCALL call_3_10_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_9_10
#define init_3_10_9_10
void FASTCALL call_3_10_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_9_11
#define init_3_10_9_11
void FASTCALL call_3_10_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_9_12
#define init_3_10_9_12
void FASTCALL call_3_10_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_9_13
#define init_3_10_9_13
void FASTCALL call_3_10_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_9_14
#define init_3_10_9_14
void FASTCALL call_3_10_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_9_15
#define init_3_10_9_15
void FASTCALL call_3_10_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_10_0
#define init_3_10_10_0
void FASTCALL call_3_10_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_10_1
#define init_3_10_10_1
void FASTCALL call_3_10_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_10_2
#define init_3_10_10_2
void FASTCALL call_3_10_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_10_3
#define init_3_10_10_3
void FASTCALL call_3_10_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_10_4
#define init_3_10_10_4
void FASTCALL call_3_10_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_10_5
#define init_3_10_10_5
void FASTCALL call_3_10_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_10_6
#define init_3_10_10_6
void FASTCALL call_3_10_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_10_7
#define init_3_10_10_7
void FASTCALL call_3_10_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_10_8
#define init_3_10_10_8
void FASTCALL call_3_10_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_10_9
#define init_3_10_10_9
void FASTCALL call_3_10_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_10_10
#define init_3_10_10_10
void FASTCALL call_3_10_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_10_11
#define init_3_10_10_11
void FASTCALL call_3_10_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_10_12
#define init_3_10_10_12
void FASTCALL call_3_10_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_10_13
#define init_3_10_10_13
void FASTCALL call_3_10_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_10_14
#define init_3_10_10_14
void FASTCALL call_3_10_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_10_15
#define init_3_10_10_15
void FASTCALL call_3_10_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_11_0
#define init_3_10_11_0
void FASTCALL call_3_10_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_11_1
#define init_3_10_11_1
void FASTCALL call_3_10_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_11_2
#define init_3_10_11_2
void FASTCALL call_3_10_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_11_3
#define init_3_10_11_3
void FASTCALL call_3_10_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_11_4
#define init_3_10_11_4
void FASTCALL call_3_10_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_11_5
#define init_3_10_11_5
void FASTCALL call_3_10_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_11_6
#define init_3_10_11_6
void FASTCALL call_3_10_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_11_7
#define init_3_10_11_7
void FASTCALL call_3_10_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_11_8
#define init_3_10_11_8
void FASTCALL call_3_10_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_11_9
#define init_3_10_11_9
void FASTCALL call_3_10_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_11_10
#define init_3_10_11_10
void FASTCALL call_3_10_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_11_11
#define init_3_10_11_11
void FASTCALL call_3_10_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_11_12
#define init_3_10_11_12
void FASTCALL call_3_10_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_11_13
#define init_3_10_11_13
void FASTCALL call_3_10_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_11_14
#define init_3_10_11_14
void FASTCALL call_3_10_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_11_15
#define init_3_10_11_15
void FASTCALL call_3_10_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_12_0
#define init_3_10_12_0
void FASTCALL call_3_10_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_12_1
#define init_3_10_12_1
void FASTCALL call_3_10_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_12_2
#define init_3_10_12_2
void FASTCALL call_3_10_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_12_3
#define init_3_10_12_3
void FASTCALL call_3_10_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_12_4
#define init_3_10_12_4
void FASTCALL call_3_10_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_12_5
#define init_3_10_12_5
void FASTCALL call_3_10_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_12_6
#define init_3_10_12_6
void FASTCALL call_3_10_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_12_7
#define init_3_10_12_7
void FASTCALL call_3_10_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_12_8
#define init_3_10_12_8
void FASTCALL call_3_10_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_12_9
#define init_3_10_12_9
void FASTCALL call_3_10_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_12_10
#define init_3_10_12_10
void FASTCALL call_3_10_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_12_11
#define init_3_10_12_11
void FASTCALL call_3_10_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_12_12
#define init_3_10_12_12
void FASTCALL call_3_10_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_12_13
#define init_3_10_12_13
void FASTCALL call_3_10_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_12_14
#define init_3_10_12_14
void FASTCALL call_3_10_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_12_15
#define init_3_10_12_15
void FASTCALL call_3_10_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_13_0
#define init_3_10_13_0
void FASTCALL call_3_10_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_13_1
#define init_3_10_13_1
void FASTCALL call_3_10_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_13_2
#define init_3_10_13_2
void FASTCALL call_3_10_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_13_3
#define init_3_10_13_3
void FASTCALL call_3_10_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_13_4
#define init_3_10_13_4
void FASTCALL call_3_10_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_13_5
#define init_3_10_13_5
void FASTCALL call_3_10_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_13_6
#define init_3_10_13_6
void FASTCALL call_3_10_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_13_7
#define init_3_10_13_7
void FASTCALL call_3_10_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_13_8
#define init_3_10_13_8
void FASTCALL call_3_10_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_13_9
#define init_3_10_13_9
void FASTCALL call_3_10_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_13_10
#define init_3_10_13_10
void FASTCALL call_3_10_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_13_11
#define init_3_10_13_11
void FASTCALL call_3_10_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_13_12
#define init_3_10_13_12
void FASTCALL call_3_10_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_13_13
#define init_3_10_13_13
void FASTCALL call_3_10_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_13_14
#define init_3_10_13_14
void FASTCALL call_3_10_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_13_15
#define init_3_10_13_15
void FASTCALL call_3_10_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_14_0
#define init_3_10_14_0
void FASTCALL call_3_10_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_14_1
#define init_3_10_14_1
void FASTCALL call_3_10_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_14_2
#define init_3_10_14_2
void FASTCALL call_3_10_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_14_3
#define init_3_10_14_3
void FASTCALL call_3_10_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_14_4
#define init_3_10_14_4
void FASTCALL call_3_10_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_14_5
#define init_3_10_14_5
void FASTCALL call_3_10_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_14_6
#define init_3_10_14_6
void FASTCALL call_3_10_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_14_7
#define init_3_10_14_7
void FASTCALL call_3_10_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_14_8
#define init_3_10_14_8
void FASTCALL call_3_10_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_14_9
#define init_3_10_14_9
void FASTCALL call_3_10_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_14_10
#define init_3_10_14_10
void FASTCALL call_3_10_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_14_11
#define init_3_10_14_11
void FASTCALL call_3_10_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_14_12
#define init_3_10_14_12
void FASTCALL call_3_10_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_14_13
#define init_3_10_14_13
void FASTCALL call_3_10_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_14_14
#define init_3_10_14_14
void FASTCALL call_3_10_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_14_15
#define init_3_10_14_15
void FASTCALL call_3_10_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_15_0
#define init_3_10_15_0
void FASTCALL call_3_10_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_15_1
#define init_3_10_15_1
void FASTCALL call_3_10_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_15_2
#define init_3_10_15_2
void FASTCALL call_3_10_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_15_3
#define init_3_10_15_3
void FASTCALL call_3_10_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_15_4
#define init_3_10_15_4
void FASTCALL call_3_10_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_15_5
#define init_3_10_15_5
void FASTCALL call_3_10_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_15_6
#define init_3_10_15_6
void FASTCALL call_3_10_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_15_7
#define init_3_10_15_7
void FASTCALL call_3_10_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_15_8
#define init_3_10_15_8
void FASTCALL call_3_10_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_15_9
#define init_3_10_15_9
void FASTCALL call_3_10_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_15_10
#define init_3_10_15_10
void FASTCALL call_3_10_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_15_11
#define init_3_10_15_11
void FASTCALL call_3_10_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_15_12
#define init_3_10_15_12
void FASTCALL call_3_10_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_15_13
#define init_3_10_15_13
void FASTCALL call_3_10_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_15_14
#define init_3_10_15_14
void FASTCALL call_3_10_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_10_15_15
#define init_3_10_15_15
void FASTCALL call_3_10_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_0_0
#define init_3_11_0_0
void FASTCALL call_3_11_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_0_1
#define init_3_11_0_1
void FASTCALL call_3_11_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_0_2
#define init_3_11_0_2
void FASTCALL call_3_11_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_0_3
#define init_3_11_0_3
void FASTCALL call_3_11_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_0_4
#define init_3_11_0_4
void FASTCALL call_3_11_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_0_5
#define init_3_11_0_5
void FASTCALL call_3_11_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_0_6
#define init_3_11_0_6
void FASTCALL call_3_11_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_0_7
#define init_3_11_0_7
void FASTCALL call_3_11_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_0_8
#define init_3_11_0_8
void FASTCALL call_3_11_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_0_9
#define init_3_11_0_9
void FASTCALL call_3_11_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_0_10
#define init_3_11_0_10
void FASTCALL call_3_11_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_0_11
#define init_3_11_0_11
void FASTCALL call_3_11_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_0_12
#define init_3_11_0_12
void FASTCALL call_3_11_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_0_13
#define init_3_11_0_13
void FASTCALL call_3_11_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_0_14
#define init_3_11_0_14
void FASTCALL call_3_11_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_0_15
#define init_3_11_0_15
void FASTCALL call_3_11_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_1_0
#define init_3_11_1_0
void FASTCALL call_3_11_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_1_1
#define init_3_11_1_1
void FASTCALL call_3_11_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_1_2
#define init_3_11_1_2
void FASTCALL call_3_11_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_1_3
#define init_3_11_1_3
void FASTCALL call_3_11_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_1_4
#define init_3_11_1_4
void FASTCALL call_3_11_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_1_5
#define init_3_11_1_5
void FASTCALL call_3_11_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_1_6
#define init_3_11_1_6
void FASTCALL call_3_11_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_1_7
#define init_3_11_1_7
void FASTCALL call_3_11_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_1_8
#define init_3_11_1_8
void FASTCALL call_3_11_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_1_9
#define init_3_11_1_9
void FASTCALL call_3_11_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_1_10
#define init_3_11_1_10
void FASTCALL call_3_11_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_1_11
#define init_3_11_1_11
void FASTCALL call_3_11_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_1_12
#define init_3_11_1_12
void FASTCALL call_3_11_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_1_13
#define init_3_11_1_13
void FASTCALL call_3_11_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_1_14
#define init_3_11_1_14
void FASTCALL call_3_11_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_1_15
#define init_3_11_1_15
void FASTCALL call_3_11_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_2_0
#define init_3_11_2_0
void FASTCALL call_3_11_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_2_1
#define init_3_11_2_1
void FASTCALL call_3_11_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_2_2
#define init_3_11_2_2
void FASTCALL call_3_11_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_2_3
#define init_3_11_2_3
void FASTCALL call_3_11_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_2_4
#define init_3_11_2_4
void FASTCALL call_3_11_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_2_5
#define init_3_11_2_5
void FASTCALL call_3_11_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_2_6
#define init_3_11_2_6
void FASTCALL call_3_11_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_2_7
#define init_3_11_2_7
void FASTCALL call_3_11_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_2_8
#define init_3_11_2_8
void FASTCALL call_3_11_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_2_9
#define init_3_11_2_9
void FASTCALL call_3_11_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_2_10
#define init_3_11_2_10
void FASTCALL call_3_11_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_2_11
#define init_3_11_2_11
void FASTCALL call_3_11_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_2_12
#define init_3_11_2_12
void FASTCALL call_3_11_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_2_13
#define init_3_11_2_13
void FASTCALL call_3_11_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_2_14
#define init_3_11_2_14
void FASTCALL call_3_11_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_2_15
#define init_3_11_2_15
void FASTCALL call_3_11_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_3_0
#define init_3_11_3_0
void FASTCALL call_3_11_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_3_1
#define init_3_11_3_1
void FASTCALL call_3_11_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_3_2
#define init_3_11_3_2
void FASTCALL call_3_11_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_3_3
#define init_3_11_3_3
void FASTCALL call_3_11_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_3_4
#define init_3_11_3_4
void FASTCALL call_3_11_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_3_5
#define init_3_11_3_5
void FASTCALL call_3_11_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_3_6
#define init_3_11_3_6
void FASTCALL call_3_11_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_3_7
#define init_3_11_3_7
void FASTCALL call_3_11_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_3_8
#define init_3_11_3_8
void FASTCALL call_3_11_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_3_9
#define init_3_11_3_9
void FASTCALL call_3_11_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_3_10
#define init_3_11_3_10
void FASTCALL call_3_11_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_3_11
#define init_3_11_3_11
void FASTCALL call_3_11_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_3_12
#define init_3_11_3_12
void FASTCALL call_3_11_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_3_13
#define init_3_11_3_13
void FASTCALL call_3_11_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_3_14
#define init_3_11_3_14
void FASTCALL call_3_11_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_3_15
#define init_3_11_3_15
void FASTCALL call_3_11_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_4_0
#define init_3_11_4_0
void FASTCALL call_3_11_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_4_1
#define init_3_11_4_1
void FASTCALL call_3_11_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_4_2
#define init_3_11_4_2
void FASTCALL call_3_11_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_4_3
#define init_3_11_4_3
void FASTCALL call_3_11_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_4_4
#define init_3_11_4_4
void FASTCALL call_3_11_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_4_5
#define init_3_11_4_5
void FASTCALL call_3_11_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_4_6
#define init_3_11_4_6
void FASTCALL call_3_11_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_4_7
#define init_3_11_4_7
void FASTCALL call_3_11_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_4_8
#define init_3_11_4_8
void FASTCALL call_3_11_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_4_9
#define init_3_11_4_9
void FASTCALL call_3_11_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_4_10
#define init_3_11_4_10
void FASTCALL call_3_11_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_4_11
#define init_3_11_4_11
void FASTCALL call_3_11_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_4_12
#define init_3_11_4_12
void FASTCALL call_3_11_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_4_13
#define init_3_11_4_13
void FASTCALL call_3_11_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_4_14
#define init_3_11_4_14
void FASTCALL call_3_11_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_4_15
#define init_3_11_4_15
void FASTCALL call_3_11_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_5_0
#define init_3_11_5_0
void FASTCALL call_3_11_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_5_1
#define init_3_11_5_1
void FASTCALL call_3_11_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_5_2
#define init_3_11_5_2
void FASTCALL call_3_11_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_5_3
#define init_3_11_5_3
void FASTCALL call_3_11_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_5_4
#define init_3_11_5_4
void FASTCALL call_3_11_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_5_5
#define init_3_11_5_5
void FASTCALL call_3_11_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_5_6
#define init_3_11_5_6
void FASTCALL call_3_11_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_5_7
#define init_3_11_5_7
void FASTCALL call_3_11_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_5_8
#define init_3_11_5_8
void FASTCALL call_3_11_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_5_9
#define init_3_11_5_9
void FASTCALL call_3_11_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_5_10
#define init_3_11_5_10
void FASTCALL call_3_11_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_5_11
#define init_3_11_5_11
void FASTCALL call_3_11_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_5_12
#define init_3_11_5_12
void FASTCALL call_3_11_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_5_13
#define init_3_11_5_13
void FASTCALL call_3_11_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_5_14
#define init_3_11_5_14
void FASTCALL call_3_11_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_5_15
#define init_3_11_5_15
void FASTCALL call_3_11_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_6_0
#define init_3_11_6_0
void FASTCALL call_3_11_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_6_1
#define init_3_11_6_1
void FASTCALL call_3_11_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_6_2
#define init_3_11_6_2
void FASTCALL call_3_11_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_6_3
#define init_3_11_6_3
void FASTCALL call_3_11_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_6_4
#define init_3_11_6_4
void FASTCALL call_3_11_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_6_5
#define init_3_11_6_5
void FASTCALL call_3_11_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_6_6
#define init_3_11_6_6
void FASTCALL call_3_11_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_6_7
#define init_3_11_6_7
void FASTCALL call_3_11_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_6_8
#define init_3_11_6_8
void FASTCALL call_3_11_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_6_9
#define init_3_11_6_9
void FASTCALL call_3_11_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_6_10
#define init_3_11_6_10
void FASTCALL call_3_11_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_6_11
#define init_3_11_6_11
void FASTCALL call_3_11_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_6_12
#define init_3_11_6_12
void FASTCALL call_3_11_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_6_13
#define init_3_11_6_13
void FASTCALL call_3_11_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_6_14
#define init_3_11_6_14
void FASTCALL call_3_11_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_6_15
#define init_3_11_6_15
void FASTCALL call_3_11_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_7_0
#define init_3_11_7_0
void FASTCALL call_3_11_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_7_1
#define init_3_11_7_1
void FASTCALL call_3_11_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_7_2
#define init_3_11_7_2
void FASTCALL call_3_11_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_7_3
#define init_3_11_7_3
void FASTCALL call_3_11_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_7_4
#define init_3_11_7_4
void FASTCALL call_3_11_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_7_5
#define init_3_11_7_5
void FASTCALL call_3_11_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_7_6
#define init_3_11_7_6
void FASTCALL call_3_11_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_7_7
#define init_3_11_7_7
void FASTCALL call_3_11_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_7_8
#define init_3_11_7_8
void FASTCALL call_3_11_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_7_9
#define init_3_11_7_9
void FASTCALL call_3_11_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_7_10
#define init_3_11_7_10
void FASTCALL call_3_11_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_7_11
#define init_3_11_7_11
void FASTCALL call_3_11_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_7_12
#define init_3_11_7_12
void FASTCALL call_3_11_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_7_13
#define init_3_11_7_13
void FASTCALL call_3_11_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_7_14
#define init_3_11_7_14
void FASTCALL call_3_11_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_7_15
#define init_3_11_7_15
void FASTCALL call_3_11_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_8_0
#define init_3_11_8_0
void FASTCALL call_3_11_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_8_1
#define init_3_11_8_1
void FASTCALL call_3_11_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_8_2
#define init_3_11_8_2
void FASTCALL call_3_11_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_8_3
#define init_3_11_8_3
void FASTCALL call_3_11_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_8_4
#define init_3_11_8_4
void FASTCALL call_3_11_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_8_5
#define init_3_11_8_5
void FASTCALL call_3_11_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_8_6
#define init_3_11_8_6
void FASTCALL call_3_11_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_8_7
#define init_3_11_8_7
void FASTCALL call_3_11_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_8_8
#define init_3_11_8_8
void FASTCALL call_3_11_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_8_9
#define init_3_11_8_9
void FASTCALL call_3_11_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_8_10
#define init_3_11_8_10
void FASTCALL call_3_11_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_8_11
#define init_3_11_8_11
void FASTCALL call_3_11_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_8_12
#define init_3_11_8_12
void FASTCALL call_3_11_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_8_13
#define init_3_11_8_13
void FASTCALL call_3_11_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_8_14
#define init_3_11_8_14
void FASTCALL call_3_11_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_8_15
#define init_3_11_8_15
void FASTCALL call_3_11_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_9_0
#define init_3_11_9_0
void FASTCALL call_3_11_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_9_1
#define init_3_11_9_1
void FASTCALL call_3_11_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_9_2
#define init_3_11_9_2
void FASTCALL call_3_11_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_9_3
#define init_3_11_9_3
void FASTCALL call_3_11_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_9_4
#define init_3_11_9_4
void FASTCALL call_3_11_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_9_5
#define init_3_11_9_5
void FASTCALL call_3_11_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_9_6
#define init_3_11_9_6
void FASTCALL call_3_11_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_9_7
#define init_3_11_9_7
void FASTCALL call_3_11_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_9_8
#define init_3_11_9_8
void FASTCALL call_3_11_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_9_9
#define init_3_11_9_9
void FASTCALL call_3_11_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_9_10
#define init_3_11_9_10
void FASTCALL call_3_11_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_9_11
#define init_3_11_9_11
void FASTCALL call_3_11_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_9_12
#define init_3_11_9_12
void FASTCALL call_3_11_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_9_13
#define init_3_11_9_13
void FASTCALL call_3_11_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_9_14
#define init_3_11_9_14
void FASTCALL call_3_11_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_9_15
#define init_3_11_9_15
void FASTCALL call_3_11_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_10_0
#define init_3_11_10_0
void FASTCALL call_3_11_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_10_1
#define init_3_11_10_1
void FASTCALL call_3_11_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_10_2
#define init_3_11_10_2
void FASTCALL call_3_11_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_10_3
#define init_3_11_10_3
void FASTCALL call_3_11_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_10_4
#define init_3_11_10_4
void FASTCALL call_3_11_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_10_5
#define init_3_11_10_5
void FASTCALL call_3_11_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_10_6
#define init_3_11_10_6
void FASTCALL call_3_11_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_10_7
#define init_3_11_10_7
void FASTCALL call_3_11_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_10_8
#define init_3_11_10_8
void FASTCALL call_3_11_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_10_9
#define init_3_11_10_9
void FASTCALL call_3_11_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_10_10
#define init_3_11_10_10
void FASTCALL call_3_11_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_10_11
#define init_3_11_10_11
void FASTCALL call_3_11_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_10_12
#define init_3_11_10_12
void FASTCALL call_3_11_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_10_13
#define init_3_11_10_13
void FASTCALL call_3_11_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_10_14
#define init_3_11_10_14
void FASTCALL call_3_11_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_10_15
#define init_3_11_10_15
void FASTCALL call_3_11_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_11_0
#define init_3_11_11_0
void FASTCALL call_3_11_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_11_1
#define init_3_11_11_1
void FASTCALL call_3_11_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_11_2
#define init_3_11_11_2
void FASTCALL call_3_11_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_11_3
#define init_3_11_11_3
void FASTCALL call_3_11_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_11_4
#define init_3_11_11_4
void FASTCALL call_3_11_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_11_5
#define init_3_11_11_5
void FASTCALL call_3_11_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_11_6
#define init_3_11_11_6
void FASTCALL call_3_11_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_11_7
#define init_3_11_11_7
void FASTCALL call_3_11_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_11_8
#define init_3_11_11_8
void FASTCALL call_3_11_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_11_9
#define init_3_11_11_9
void FASTCALL call_3_11_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_11_10
#define init_3_11_11_10
void FASTCALL call_3_11_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_11_11
#define init_3_11_11_11
void FASTCALL call_3_11_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_11_12
#define init_3_11_11_12
void FASTCALL call_3_11_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_11_13
#define init_3_11_11_13
void FASTCALL call_3_11_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_11_14
#define init_3_11_11_14
void FASTCALL call_3_11_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_11_15
#define init_3_11_11_15
void FASTCALL call_3_11_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_12_0
#define init_3_11_12_0
void FASTCALL call_3_11_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_12_1
#define init_3_11_12_1
void FASTCALL call_3_11_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_12_2
#define init_3_11_12_2
void FASTCALL call_3_11_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_12_3
#define init_3_11_12_3
void FASTCALL call_3_11_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_12_4
#define init_3_11_12_4
void FASTCALL call_3_11_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_12_5
#define init_3_11_12_5
void FASTCALL call_3_11_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_12_6
#define init_3_11_12_6
void FASTCALL call_3_11_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_12_7
#define init_3_11_12_7
void FASTCALL call_3_11_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_12_8
#define init_3_11_12_8
void FASTCALL call_3_11_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_12_9
#define init_3_11_12_9
void FASTCALL call_3_11_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_12_10
#define init_3_11_12_10
void FASTCALL call_3_11_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_12_11
#define init_3_11_12_11
void FASTCALL call_3_11_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_12_12
#define init_3_11_12_12
void FASTCALL call_3_11_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_12_13
#define init_3_11_12_13
void FASTCALL call_3_11_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_12_14
#define init_3_11_12_14
void FASTCALL call_3_11_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_12_15
#define init_3_11_12_15
void FASTCALL call_3_11_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_13_0
#define init_3_11_13_0
void FASTCALL call_3_11_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_13_1
#define init_3_11_13_1
void FASTCALL call_3_11_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_13_2
#define init_3_11_13_2
void FASTCALL call_3_11_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_13_3
#define init_3_11_13_3
void FASTCALL call_3_11_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_13_4
#define init_3_11_13_4
void FASTCALL call_3_11_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_13_5
#define init_3_11_13_5
void FASTCALL call_3_11_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_13_6
#define init_3_11_13_6
void FASTCALL call_3_11_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_13_7
#define init_3_11_13_7
void FASTCALL call_3_11_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_13_8
#define init_3_11_13_8
void FASTCALL call_3_11_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_13_9
#define init_3_11_13_9
void FASTCALL call_3_11_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_13_10
#define init_3_11_13_10
void FASTCALL call_3_11_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_13_11
#define init_3_11_13_11
void FASTCALL call_3_11_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_13_12
#define init_3_11_13_12
void FASTCALL call_3_11_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_13_13
#define init_3_11_13_13
void FASTCALL call_3_11_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_13_14
#define init_3_11_13_14
void FASTCALL call_3_11_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_13_15
#define init_3_11_13_15
void FASTCALL call_3_11_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_14_0
#define init_3_11_14_0
void FASTCALL call_3_11_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_14_1
#define init_3_11_14_1
void FASTCALL call_3_11_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_14_2
#define init_3_11_14_2
void FASTCALL call_3_11_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_14_3
#define init_3_11_14_3
void FASTCALL call_3_11_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_14_4
#define init_3_11_14_4
void FASTCALL call_3_11_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_14_5
#define init_3_11_14_5
void FASTCALL call_3_11_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_14_6
#define init_3_11_14_6
void FASTCALL call_3_11_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_14_7
#define init_3_11_14_7
void FASTCALL call_3_11_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_14_8
#define init_3_11_14_8
void FASTCALL call_3_11_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_14_9
#define init_3_11_14_9
void FASTCALL call_3_11_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_14_10
#define init_3_11_14_10
void FASTCALL call_3_11_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_14_11
#define init_3_11_14_11
void FASTCALL call_3_11_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_14_12
#define init_3_11_14_12
void FASTCALL call_3_11_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_14_13
#define init_3_11_14_13
void FASTCALL call_3_11_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_14_14
#define init_3_11_14_14
void FASTCALL call_3_11_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_14_15
#define init_3_11_14_15
void FASTCALL call_3_11_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_15_0
#define init_3_11_15_0
void FASTCALL call_3_11_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_15_1
#define init_3_11_15_1
void FASTCALL call_3_11_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_15_2
#define init_3_11_15_2
void FASTCALL call_3_11_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_15_3
#define init_3_11_15_3
void FASTCALL call_3_11_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_15_4
#define init_3_11_15_4
void FASTCALL call_3_11_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_15_5
#define init_3_11_15_5
void FASTCALL call_3_11_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_15_6
#define init_3_11_15_6
void FASTCALL call_3_11_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_15_7
#define init_3_11_15_7
void FASTCALL call_3_11_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_15_8
#define init_3_11_15_8
void FASTCALL call_3_11_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_15_9
#define init_3_11_15_9
void FASTCALL call_3_11_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_15_10
#define init_3_11_15_10
void FASTCALL call_3_11_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_15_11
#define init_3_11_15_11
void FASTCALL call_3_11_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_15_12
#define init_3_11_15_12
void FASTCALL call_3_11_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_15_13
#define init_3_11_15_13
void FASTCALL call_3_11_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_15_14
#define init_3_11_15_14
void FASTCALL call_3_11_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_11_15_15
#define init_3_11_15_15
void FASTCALL call_3_11_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_0_0
#define init_3_12_0_0
void FASTCALL call_3_12_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_0_1
#define init_3_12_0_1
void FASTCALL call_3_12_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_0_2
#define init_3_12_0_2
void FASTCALL call_3_12_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_0_3
#define init_3_12_0_3
void FASTCALL call_3_12_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_0_4
#define init_3_12_0_4
void FASTCALL call_3_12_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_0_5
#define init_3_12_0_5
void FASTCALL call_3_12_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_0_6
#define init_3_12_0_6
void FASTCALL call_3_12_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_0_7
#define init_3_12_0_7
void FASTCALL call_3_12_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_0_8
#define init_3_12_0_8
void FASTCALL call_3_12_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_0_9
#define init_3_12_0_9
void FASTCALL call_3_12_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_0_10
#define init_3_12_0_10
void FASTCALL call_3_12_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_0_11
#define init_3_12_0_11
void FASTCALL call_3_12_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_0_12
#define init_3_12_0_12
void FASTCALL call_3_12_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_0_13
#define init_3_12_0_13
void FASTCALL call_3_12_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_0_14
#define init_3_12_0_14
void FASTCALL call_3_12_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_0_15
#define init_3_12_0_15
void FASTCALL call_3_12_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_1_0
#define init_3_12_1_0
void FASTCALL call_3_12_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_1_1
#define init_3_12_1_1
void FASTCALL call_3_12_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_1_2
#define init_3_12_1_2
void FASTCALL call_3_12_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_1_3
#define init_3_12_1_3
void FASTCALL call_3_12_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_1_4
#define init_3_12_1_4
void FASTCALL call_3_12_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_1_5
#define init_3_12_1_5
void FASTCALL call_3_12_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_1_6
#define init_3_12_1_6
void FASTCALL call_3_12_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_1_7
#define init_3_12_1_7
void FASTCALL call_3_12_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_1_8
#define init_3_12_1_8
void FASTCALL call_3_12_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_1_9
#define init_3_12_1_9
void FASTCALL call_3_12_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_1_10
#define init_3_12_1_10
void FASTCALL call_3_12_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_1_11
#define init_3_12_1_11
void FASTCALL call_3_12_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_1_12
#define init_3_12_1_12
void FASTCALL call_3_12_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_1_13
#define init_3_12_1_13
void FASTCALL call_3_12_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_1_14
#define init_3_12_1_14
void FASTCALL call_3_12_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_1_15
#define init_3_12_1_15
void FASTCALL call_3_12_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_2_0
#define init_3_12_2_0
void FASTCALL call_3_12_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_2_1
#define init_3_12_2_1
void FASTCALL call_3_12_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_2_2
#define init_3_12_2_2
void FASTCALL call_3_12_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_2_3
#define init_3_12_2_3
void FASTCALL call_3_12_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_2_4
#define init_3_12_2_4
void FASTCALL call_3_12_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_2_5
#define init_3_12_2_5
void FASTCALL call_3_12_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_2_6
#define init_3_12_2_6
void FASTCALL call_3_12_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_2_7
#define init_3_12_2_7
void FASTCALL call_3_12_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_2_8
#define init_3_12_2_8
void FASTCALL call_3_12_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_2_9
#define init_3_12_2_9
void FASTCALL call_3_12_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_2_10
#define init_3_12_2_10
void FASTCALL call_3_12_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_2_11
#define init_3_12_2_11
void FASTCALL call_3_12_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_2_12
#define init_3_12_2_12
void FASTCALL call_3_12_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_2_13
#define init_3_12_2_13
void FASTCALL call_3_12_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_2_14
#define init_3_12_2_14
void FASTCALL call_3_12_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_2_15
#define init_3_12_2_15
void FASTCALL call_3_12_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_3_0
#define init_3_12_3_0
void FASTCALL call_3_12_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_3_1
#define init_3_12_3_1
void FASTCALL call_3_12_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_3_2
#define init_3_12_3_2
void FASTCALL call_3_12_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_3_3
#define init_3_12_3_3
void FASTCALL call_3_12_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_3_4
#define init_3_12_3_4
void FASTCALL call_3_12_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_3_5
#define init_3_12_3_5
void FASTCALL call_3_12_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_3_6
#define init_3_12_3_6
void FASTCALL call_3_12_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_3_7
#define init_3_12_3_7
void FASTCALL call_3_12_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_3_8
#define init_3_12_3_8
void FASTCALL call_3_12_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_3_9
#define init_3_12_3_9
void FASTCALL call_3_12_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_3_10
#define init_3_12_3_10
void FASTCALL call_3_12_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_3_11
#define init_3_12_3_11
void FASTCALL call_3_12_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_3_12
#define init_3_12_3_12
void FASTCALL call_3_12_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_3_13
#define init_3_12_3_13
void FASTCALL call_3_12_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_3_14
#define init_3_12_3_14
void FASTCALL call_3_12_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_3_15
#define init_3_12_3_15
void FASTCALL call_3_12_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_4_0
#define init_3_12_4_0
void FASTCALL call_3_12_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_4_1
#define init_3_12_4_1
void FASTCALL call_3_12_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_4_2
#define init_3_12_4_2
void FASTCALL call_3_12_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_4_3
#define init_3_12_4_3
void FASTCALL call_3_12_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_4_4
#define init_3_12_4_4
void FASTCALL call_3_12_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_4_5
#define init_3_12_4_5
void FASTCALL call_3_12_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_4_6
#define init_3_12_4_6
void FASTCALL call_3_12_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_4_7
#define init_3_12_4_7
void FASTCALL call_3_12_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_4_8
#define init_3_12_4_8
void FASTCALL call_3_12_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_4_9
#define init_3_12_4_9
void FASTCALL call_3_12_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_4_10
#define init_3_12_4_10
void FASTCALL call_3_12_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_4_11
#define init_3_12_4_11
void FASTCALL call_3_12_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_4_12
#define init_3_12_4_12
void FASTCALL call_3_12_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_4_13
#define init_3_12_4_13
void FASTCALL call_3_12_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_4_14
#define init_3_12_4_14
void FASTCALL call_3_12_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_4_15
#define init_3_12_4_15
void FASTCALL call_3_12_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_5_0
#define init_3_12_5_0
void FASTCALL call_3_12_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_5_1
#define init_3_12_5_1
void FASTCALL call_3_12_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_5_2
#define init_3_12_5_2
void FASTCALL call_3_12_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_5_3
#define init_3_12_5_3
void FASTCALL call_3_12_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_5_4
#define init_3_12_5_4
void FASTCALL call_3_12_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_5_5
#define init_3_12_5_5
void FASTCALL call_3_12_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_5_6
#define init_3_12_5_6
void FASTCALL call_3_12_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_5_7
#define init_3_12_5_7
void FASTCALL call_3_12_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_5_8
#define init_3_12_5_8
void FASTCALL call_3_12_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_5_9
#define init_3_12_5_9
void FASTCALL call_3_12_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_5_10
#define init_3_12_5_10
void FASTCALL call_3_12_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_5_11
#define init_3_12_5_11
void FASTCALL call_3_12_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_5_12
#define init_3_12_5_12
void FASTCALL call_3_12_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_5_13
#define init_3_12_5_13
void FASTCALL call_3_12_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_5_14
#define init_3_12_5_14
void FASTCALL call_3_12_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_5_15
#define init_3_12_5_15
void FASTCALL call_3_12_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_6_0
#define init_3_12_6_0
void FASTCALL call_3_12_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_6_1
#define init_3_12_6_1
void FASTCALL call_3_12_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_6_2
#define init_3_12_6_2
void FASTCALL call_3_12_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_6_3
#define init_3_12_6_3
void FASTCALL call_3_12_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_6_4
#define init_3_12_6_4
void FASTCALL call_3_12_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_6_5
#define init_3_12_6_5
void FASTCALL call_3_12_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_6_6
#define init_3_12_6_6
void FASTCALL call_3_12_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_6_7
#define init_3_12_6_7
void FASTCALL call_3_12_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_6_8
#define init_3_12_6_8
void FASTCALL call_3_12_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_6_9
#define init_3_12_6_9
void FASTCALL call_3_12_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_6_10
#define init_3_12_6_10
void FASTCALL call_3_12_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_6_11
#define init_3_12_6_11
void FASTCALL call_3_12_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_6_12
#define init_3_12_6_12
void FASTCALL call_3_12_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_6_13
#define init_3_12_6_13
void FASTCALL call_3_12_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_6_14
#define init_3_12_6_14
void FASTCALL call_3_12_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_6_15
#define init_3_12_6_15
void FASTCALL call_3_12_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_7_0
#define init_3_12_7_0
void FASTCALL call_3_12_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_7_1
#define init_3_12_7_1
void FASTCALL call_3_12_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_7_2
#define init_3_12_7_2
void FASTCALL call_3_12_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_7_3
#define init_3_12_7_3
void FASTCALL call_3_12_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_7_4
#define init_3_12_7_4
void FASTCALL call_3_12_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_7_5
#define init_3_12_7_5
void FASTCALL call_3_12_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_7_6
#define init_3_12_7_6
void FASTCALL call_3_12_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_7_7
#define init_3_12_7_7
void FASTCALL call_3_12_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_7_8
#define init_3_12_7_8
void FASTCALL call_3_12_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_7_9
#define init_3_12_7_9
void FASTCALL call_3_12_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_7_10
#define init_3_12_7_10
void FASTCALL call_3_12_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_7_11
#define init_3_12_7_11
void FASTCALL call_3_12_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_7_12
#define init_3_12_7_12
void FASTCALL call_3_12_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_7_13
#define init_3_12_7_13
void FASTCALL call_3_12_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_7_14
#define init_3_12_7_14
void FASTCALL call_3_12_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_7_15
#define init_3_12_7_15
void FASTCALL call_3_12_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_8_0
#define init_3_12_8_0
void FASTCALL call_3_12_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_8_1
#define init_3_12_8_1
void FASTCALL call_3_12_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_8_2
#define init_3_12_8_2
void FASTCALL call_3_12_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_8_3
#define init_3_12_8_3
void FASTCALL call_3_12_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_8_4
#define init_3_12_8_4
void FASTCALL call_3_12_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_8_5
#define init_3_12_8_5
void FASTCALL call_3_12_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_8_6
#define init_3_12_8_6
void FASTCALL call_3_12_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_8_7
#define init_3_12_8_7
void FASTCALL call_3_12_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_8_8
#define init_3_12_8_8
void FASTCALL call_3_12_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_8_9
#define init_3_12_8_9
void FASTCALL call_3_12_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_8_10
#define init_3_12_8_10
void FASTCALL call_3_12_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_8_11
#define init_3_12_8_11
void FASTCALL call_3_12_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_8_12
#define init_3_12_8_12
void FASTCALL call_3_12_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_8_13
#define init_3_12_8_13
void FASTCALL call_3_12_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_8_14
#define init_3_12_8_14
void FASTCALL call_3_12_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_8_15
#define init_3_12_8_15
void FASTCALL call_3_12_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_9_0
#define init_3_12_9_0
void FASTCALL call_3_12_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_9_1
#define init_3_12_9_1
void FASTCALL call_3_12_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_9_2
#define init_3_12_9_2
void FASTCALL call_3_12_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_9_3
#define init_3_12_9_3
void FASTCALL call_3_12_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_9_4
#define init_3_12_9_4
void FASTCALL call_3_12_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_9_5
#define init_3_12_9_5
void FASTCALL call_3_12_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_9_6
#define init_3_12_9_6
void FASTCALL call_3_12_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_9_7
#define init_3_12_9_7
void FASTCALL call_3_12_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_9_8
#define init_3_12_9_8
void FASTCALL call_3_12_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_9_9
#define init_3_12_9_9
void FASTCALL call_3_12_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_9_10
#define init_3_12_9_10
void FASTCALL call_3_12_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_9_11
#define init_3_12_9_11
void FASTCALL call_3_12_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_9_12
#define init_3_12_9_12
void FASTCALL call_3_12_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_9_13
#define init_3_12_9_13
void FASTCALL call_3_12_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_9_14
#define init_3_12_9_14
void FASTCALL call_3_12_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_9_15
#define init_3_12_9_15
void FASTCALL call_3_12_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_10_0
#define init_3_12_10_0
void FASTCALL call_3_12_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_10_1
#define init_3_12_10_1
void FASTCALL call_3_12_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_10_2
#define init_3_12_10_2
void FASTCALL call_3_12_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_10_3
#define init_3_12_10_3
void FASTCALL call_3_12_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_10_4
#define init_3_12_10_4
void FASTCALL call_3_12_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_10_5
#define init_3_12_10_5
void FASTCALL call_3_12_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_10_6
#define init_3_12_10_6
void FASTCALL call_3_12_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_10_7
#define init_3_12_10_7
void FASTCALL call_3_12_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_10_8
#define init_3_12_10_8
void FASTCALL call_3_12_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_10_9
#define init_3_12_10_9
void FASTCALL call_3_12_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_10_10
#define init_3_12_10_10
void FASTCALL call_3_12_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_10_11
#define init_3_12_10_11
void FASTCALL call_3_12_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_10_12
#define init_3_12_10_12
void FASTCALL call_3_12_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_10_13
#define init_3_12_10_13
void FASTCALL call_3_12_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_10_14
#define init_3_12_10_14
void FASTCALL call_3_12_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_10_15
#define init_3_12_10_15
void FASTCALL call_3_12_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_11_0
#define init_3_12_11_0
void FASTCALL call_3_12_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_11_1
#define init_3_12_11_1
void FASTCALL call_3_12_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_11_2
#define init_3_12_11_2
void FASTCALL call_3_12_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_11_3
#define init_3_12_11_3
void FASTCALL call_3_12_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_11_4
#define init_3_12_11_4
void FASTCALL call_3_12_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_11_5
#define init_3_12_11_5
void FASTCALL call_3_12_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_11_6
#define init_3_12_11_6
void FASTCALL call_3_12_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_11_7
#define init_3_12_11_7
void FASTCALL call_3_12_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_11_8
#define init_3_12_11_8
void FASTCALL call_3_12_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_11_9
#define init_3_12_11_9
void FASTCALL call_3_12_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_11_10
#define init_3_12_11_10
void FASTCALL call_3_12_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_11_11
#define init_3_12_11_11
void FASTCALL call_3_12_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_11_12
#define init_3_12_11_12
void FASTCALL call_3_12_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_11_13
#define init_3_12_11_13
void FASTCALL call_3_12_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_11_14
#define init_3_12_11_14
void FASTCALL call_3_12_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_11_15
#define init_3_12_11_15
void FASTCALL call_3_12_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_12_0
#define init_3_12_12_0
void FASTCALL call_3_12_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_12_1
#define init_3_12_12_1
void FASTCALL call_3_12_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_12_2
#define init_3_12_12_2
void FASTCALL call_3_12_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_12_3
#define init_3_12_12_3
void FASTCALL call_3_12_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_12_4
#define init_3_12_12_4
void FASTCALL call_3_12_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_12_5
#define init_3_12_12_5
void FASTCALL call_3_12_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_12_6
#define init_3_12_12_6
void FASTCALL call_3_12_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_12_7
#define init_3_12_12_7
void FASTCALL call_3_12_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_12_8
#define init_3_12_12_8
void FASTCALL call_3_12_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_12_9
#define init_3_12_12_9
void FASTCALL call_3_12_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_12_10
#define init_3_12_12_10
void FASTCALL call_3_12_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_12_11
#define init_3_12_12_11
void FASTCALL call_3_12_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_12_12
#define init_3_12_12_12
void FASTCALL call_3_12_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_12_13
#define init_3_12_12_13
void FASTCALL call_3_12_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_12_14
#define init_3_12_12_14
void FASTCALL call_3_12_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_12_15
#define init_3_12_12_15
void FASTCALL call_3_12_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_13_0
#define init_3_12_13_0
void FASTCALL call_3_12_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_13_1
#define init_3_12_13_1
void FASTCALL call_3_12_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_13_2
#define init_3_12_13_2
void FASTCALL call_3_12_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_13_3
#define init_3_12_13_3
void FASTCALL call_3_12_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_13_4
#define init_3_12_13_4
void FASTCALL call_3_12_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_13_5
#define init_3_12_13_5
void FASTCALL call_3_12_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_13_6
#define init_3_12_13_6
void FASTCALL call_3_12_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_13_7
#define init_3_12_13_7
void FASTCALL call_3_12_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_13_8
#define init_3_12_13_8
void FASTCALL call_3_12_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_13_9
#define init_3_12_13_9
void FASTCALL call_3_12_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_13_10
#define init_3_12_13_10
void FASTCALL call_3_12_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_13_11
#define init_3_12_13_11
void FASTCALL call_3_12_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_13_12
#define init_3_12_13_12
void FASTCALL call_3_12_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_13_13
#define init_3_12_13_13
void FASTCALL call_3_12_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_13_14
#define init_3_12_13_14
void FASTCALL call_3_12_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_13_15
#define init_3_12_13_15
void FASTCALL call_3_12_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_14_0
#define init_3_12_14_0
void FASTCALL call_3_12_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_14_1
#define init_3_12_14_1
void FASTCALL call_3_12_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_14_2
#define init_3_12_14_2
void FASTCALL call_3_12_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_14_3
#define init_3_12_14_3
void FASTCALL call_3_12_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_14_4
#define init_3_12_14_4
void FASTCALL call_3_12_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_14_5
#define init_3_12_14_5
void FASTCALL call_3_12_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_14_6
#define init_3_12_14_6
void FASTCALL call_3_12_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_14_7
#define init_3_12_14_7
void FASTCALL call_3_12_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_14_8
#define init_3_12_14_8
void FASTCALL call_3_12_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_14_9
#define init_3_12_14_9
void FASTCALL call_3_12_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_14_10
#define init_3_12_14_10
void FASTCALL call_3_12_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_14_11
#define init_3_12_14_11
void FASTCALL call_3_12_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_14_12
#define init_3_12_14_12
void FASTCALL call_3_12_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_14_13
#define init_3_12_14_13
void FASTCALL call_3_12_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_14_14
#define init_3_12_14_14
void FASTCALL call_3_12_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_14_15
#define init_3_12_14_15
void FASTCALL call_3_12_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_15_0
#define init_3_12_15_0
void FASTCALL call_3_12_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_15_1
#define init_3_12_15_1
void FASTCALL call_3_12_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_15_2
#define init_3_12_15_2
void FASTCALL call_3_12_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_15_3
#define init_3_12_15_3
void FASTCALL call_3_12_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_15_4
#define init_3_12_15_4
void FASTCALL call_3_12_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_15_5
#define init_3_12_15_5
void FASTCALL call_3_12_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_15_6
#define init_3_12_15_6
void FASTCALL call_3_12_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_15_7
#define init_3_12_15_7
void FASTCALL call_3_12_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_15_8
#define init_3_12_15_8
void FASTCALL call_3_12_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_15_9
#define init_3_12_15_9
void FASTCALL call_3_12_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_15_10
#define init_3_12_15_10
void FASTCALL call_3_12_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_15_11
#define init_3_12_15_11
void FASTCALL call_3_12_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_15_12
#define init_3_12_15_12
void FASTCALL call_3_12_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_15_13
#define init_3_12_15_13
void FASTCALL call_3_12_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_15_14
#define init_3_12_15_14
void FASTCALL call_3_12_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_12_15_15
#define init_3_12_15_15
void FASTCALL call_3_12_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_0_0
#define init_3_13_0_0
void FASTCALL call_3_13_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_0_1
#define init_3_13_0_1
void FASTCALL call_3_13_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_0_2
#define init_3_13_0_2
void FASTCALL call_3_13_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_0_3
#define init_3_13_0_3
void FASTCALL call_3_13_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_0_4
#define init_3_13_0_4
void FASTCALL call_3_13_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_0_5
#define init_3_13_0_5
void FASTCALL call_3_13_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_0_6
#define init_3_13_0_6
void FASTCALL call_3_13_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_0_7
#define init_3_13_0_7
void FASTCALL call_3_13_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_0_8
#define init_3_13_0_8
void FASTCALL call_3_13_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_0_9
#define init_3_13_0_9
void FASTCALL call_3_13_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_0_10
#define init_3_13_0_10
void FASTCALL call_3_13_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_0_11
#define init_3_13_0_11
void FASTCALL call_3_13_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_0_12
#define init_3_13_0_12
void FASTCALL call_3_13_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_0_13
#define init_3_13_0_13
void FASTCALL call_3_13_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_0_14
#define init_3_13_0_14
void FASTCALL call_3_13_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_0_15
#define init_3_13_0_15
void FASTCALL call_3_13_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_1_0
#define init_3_13_1_0
void FASTCALL call_3_13_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_1_1
#define init_3_13_1_1
void FASTCALL call_3_13_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_1_2
#define init_3_13_1_2
void FASTCALL call_3_13_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_1_3
#define init_3_13_1_3
void FASTCALL call_3_13_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_1_4
#define init_3_13_1_4
void FASTCALL call_3_13_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_1_5
#define init_3_13_1_5
void FASTCALL call_3_13_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_1_6
#define init_3_13_1_6
void FASTCALL call_3_13_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_1_7
#define init_3_13_1_7
void FASTCALL call_3_13_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_1_8
#define init_3_13_1_8
void FASTCALL call_3_13_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_1_9
#define init_3_13_1_9
void FASTCALL call_3_13_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_1_10
#define init_3_13_1_10
void FASTCALL call_3_13_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_1_11
#define init_3_13_1_11
void FASTCALL call_3_13_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_1_12
#define init_3_13_1_12
void FASTCALL call_3_13_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_1_13
#define init_3_13_1_13
void FASTCALL call_3_13_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_1_14
#define init_3_13_1_14
void FASTCALL call_3_13_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_1_15
#define init_3_13_1_15
void FASTCALL call_3_13_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_2_0
#define init_3_13_2_0
void FASTCALL call_3_13_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_2_1
#define init_3_13_2_1
void FASTCALL call_3_13_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_2_2
#define init_3_13_2_2
void FASTCALL call_3_13_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_2_3
#define init_3_13_2_3
void FASTCALL call_3_13_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_2_4
#define init_3_13_2_4
void FASTCALL call_3_13_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_2_5
#define init_3_13_2_5
void FASTCALL call_3_13_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_2_6
#define init_3_13_2_6
void FASTCALL call_3_13_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_2_7
#define init_3_13_2_7
void FASTCALL call_3_13_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_2_8
#define init_3_13_2_8
void FASTCALL call_3_13_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_2_9
#define init_3_13_2_9
void FASTCALL call_3_13_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_2_10
#define init_3_13_2_10
void FASTCALL call_3_13_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_2_11
#define init_3_13_2_11
void FASTCALL call_3_13_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_2_12
#define init_3_13_2_12
void FASTCALL call_3_13_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_2_13
#define init_3_13_2_13
void FASTCALL call_3_13_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_2_14
#define init_3_13_2_14
void FASTCALL call_3_13_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_2_15
#define init_3_13_2_15
void FASTCALL call_3_13_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_3_0
#define init_3_13_3_0
void FASTCALL call_3_13_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_3_1
#define init_3_13_3_1
void FASTCALL call_3_13_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_3_2
#define init_3_13_3_2
void FASTCALL call_3_13_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_3_3
#define init_3_13_3_3
void FASTCALL call_3_13_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_3_4
#define init_3_13_3_4
void FASTCALL call_3_13_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_3_5
#define init_3_13_3_5
void FASTCALL call_3_13_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_3_6
#define init_3_13_3_6
void FASTCALL call_3_13_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_3_7
#define init_3_13_3_7
void FASTCALL call_3_13_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_3_8
#define init_3_13_3_8
void FASTCALL call_3_13_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_3_9
#define init_3_13_3_9
void FASTCALL call_3_13_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_3_10
#define init_3_13_3_10
void FASTCALL call_3_13_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_3_11
#define init_3_13_3_11
void FASTCALL call_3_13_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_3_12
#define init_3_13_3_12
void FASTCALL call_3_13_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_3_13
#define init_3_13_3_13
void FASTCALL call_3_13_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_3_14
#define init_3_13_3_14
void FASTCALL call_3_13_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_3_15
#define init_3_13_3_15
void FASTCALL call_3_13_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_4_0
#define init_3_13_4_0
void FASTCALL call_3_13_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_4_1
#define init_3_13_4_1
void FASTCALL call_3_13_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_4_2
#define init_3_13_4_2
void FASTCALL call_3_13_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_4_3
#define init_3_13_4_3
void FASTCALL call_3_13_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_4_4
#define init_3_13_4_4
void FASTCALL call_3_13_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_4_5
#define init_3_13_4_5
void FASTCALL call_3_13_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_4_6
#define init_3_13_4_6
void FASTCALL call_3_13_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_4_7
#define init_3_13_4_7
void FASTCALL call_3_13_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_4_8
#define init_3_13_4_8
void FASTCALL call_3_13_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_4_9
#define init_3_13_4_9
void FASTCALL call_3_13_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_4_10
#define init_3_13_4_10
void FASTCALL call_3_13_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_4_11
#define init_3_13_4_11
void FASTCALL call_3_13_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_4_12
#define init_3_13_4_12
void FASTCALL call_3_13_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_4_13
#define init_3_13_4_13
void FASTCALL call_3_13_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_4_14
#define init_3_13_4_14
void FASTCALL call_3_13_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_4_15
#define init_3_13_4_15
void FASTCALL call_3_13_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_5_0
#define init_3_13_5_0
void FASTCALL call_3_13_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_5_1
#define init_3_13_5_1
void FASTCALL call_3_13_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_5_2
#define init_3_13_5_2
void FASTCALL call_3_13_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_5_3
#define init_3_13_5_3
void FASTCALL call_3_13_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_5_4
#define init_3_13_5_4
void FASTCALL call_3_13_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_5_5
#define init_3_13_5_5
void FASTCALL call_3_13_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_5_6
#define init_3_13_5_6
void FASTCALL call_3_13_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_5_7
#define init_3_13_5_7
void FASTCALL call_3_13_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_5_8
#define init_3_13_5_8
void FASTCALL call_3_13_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_5_9
#define init_3_13_5_9
void FASTCALL call_3_13_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_5_10
#define init_3_13_5_10
void FASTCALL call_3_13_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_5_11
#define init_3_13_5_11
void FASTCALL call_3_13_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_5_12
#define init_3_13_5_12
void FASTCALL call_3_13_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_5_13
#define init_3_13_5_13
void FASTCALL call_3_13_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_5_14
#define init_3_13_5_14
void FASTCALL call_3_13_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_5_15
#define init_3_13_5_15
void FASTCALL call_3_13_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_6_0
#define init_3_13_6_0
void FASTCALL call_3_13_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_6_1
#define init_3_13_6_1
void FASTCALL call_3_13_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_6_2
#define init_3_13_6_2
void FASTCALL call_3_13_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_6_3
#define init_3_13_6_3
void FASTCALL call_3_13_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_6_4
#define init_3_13_6_4
void FASTCALL call_3_13_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_6_5
#define init_3_13_6_5
void FASTCALL call_3_13_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_6_6
#define init_3_13_6_6
void FASTCALL call_3_13_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_6_7
#define init_3_13_6_7
void FASTCALL call_3_13_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_6_8
#define init_3_13_6_8
void FASTCALL call_3_13_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_6_9
#define init_3_13_6_9
void FASTCALL call_3_13_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_6_10
#define init_3_13_6_10
void FASTCALL call_3_13_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_6_11
#define init_3_13_6_11
void FASTCALL call_3_13_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_6_12
#define init_3_13_6_12
void FASTCALL call_3_13_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_6_13
#define init_3_13_6_13
void FASTCALL call_3_13_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_6_14
#define init_3_13_6_14
void FASTCALL call_3_13_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_6_15
#define init_3_13_6_15
void FASTCALL call_3_13_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_7_0
#define init_3_13_7_0
void FASTCALL call_3_13_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_7_1
#define init_3_13_7_1
void FASTCALL call_3_13_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_7_2
#define init_3_13_7_2
void FASTCALL call_3_13_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_7_3
#define init_3_13_7_3
void FASTCALL call_3_13_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_7_4
#define init_3_13_7_4
void FASTCALL call_3_13_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_7_5
#define init_3_13_7_5
void FASTCALL call_3_13_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_7_6
#define init_3_13_7_6
void FASTCALL call_3_13_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_7_7
#define init_3_13_7_7
void FASTCALL call_3_13_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_7_8
#define init_3_13_7_8
void FASTCALL call_3_13_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_7_9
#define init_3_13_7_9
void FASTCALL call_3_13_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_7_10
#define init_3_13_7_10
void FASTCALL call_3_13_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_7_11
#define init_3_13_7_11
void FASTCALL call_3_13_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_7_12
#define init_3_13_7_12
void FASTCALL call_3_13_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_7_13
#define init_3_13_7_13
void FASTCALL call_3_13_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_7_14
#define init_3_13_7_14
void FASTCALL call_3_13_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_7_15
#define init_3_13_7_15
void FASTCALL call_3_13_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_8_0
#define init_3_13_8_0
void FASTCALL call_3_13_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_8_1
#define init_3_13_8_1
void FASTCALL call_3_13_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_8_2
#define init_3_13_8_2
void FASTCALL call_3_13_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_8_3
#define init_3_13_8_3
void FASTCALL call_3_13_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_8_4
#define init_3_13_8_4
void FASTCALL call_3_13_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_8_5
#define init_3_13_8_5
void FASTCALL call_3_13_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_8_6
#define init_3_13_8_6
void FASTCALL call_3_13_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_8_7
#define init_3_13_8_7
void FASTCALL call_3_13_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_8_8
#define init_3_13_8_8
void FASTCALL call_3_13_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_8_9
#define init_3_13_8_9
void FASTCALL call_3_13_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_8_10
#define init_3_13_8_10
void FASTCALL call_3_13_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_8_11
#define init_3_13_8_11
void FASTCALL call_3_13_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_8_12
#define init_3_13_8_12
void FASTCALL call_3_13_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_8_13
#define init_3_13_8_13
void FASTCALL call_3_13_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_8_14
#define init_3_13_8_14
void FASTCALL call_3_13_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_8_15
#define init_3_13_8_15
void FASTCALL call_3_13_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_9_0
#define init_3_13_9_0
void FASTCALL call_3_13_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_9_1
#define init_3_13_9_1
void FASTCALL call_3_13_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_9_2
#define init_3_13_9_2
void FASTCALL call_3_13_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_9_3
#define init_3_13_9_3
void FASTCALL call_3_13_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_9_4
#define init_3_13_9_4
void FASTCALL call_3_13_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_9_5
#define init_3_13_9_5
void FASTCALL call_3_13_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_9_6
#define init_3_13_9_6
void FASTCALL call_3_13_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_9_7
#define init_3_13_9_7
void FASTCALL call_3_13_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_9_8
#define init_3_13_9_8
void FASTCALL call_3_13_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_9_9
#define init_3_13_9_9
void FASTCALL call_3_13_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_9_10
#define init_3_13_9_10
void FASTCALL call_3_13_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_9_11
#define init_3_13_9_11
void FASTCALL call_3_13_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_9_12
#define init_3_13_9_12
void FASTCALL call_3_13_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_9_13
#define init_3_13_9_13
void FASTCALL call_3_13_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_9_14
#define init_3_13_9_14
void FASTCALL call_3_13_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_9_15
#define init_3_13_9_15
void FASTCALL call_3_13_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_10_0
#define init_3_13_10_0
void FASTCALL call_3_13_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_10_1
#define init_3_13_10_1
void FASTCALL call_3_13_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_10_2
#define init_3_13_10_2
void FASTCALL call_3_13_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_10_3
#define init_3_13_10_3
void FASTCALL call_3_13_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_10_4
#define init_3_13_10_4
void FASTCALL call_3_13_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_10_5
#define init_3_13_10_5
void FASTCALL call_3_13_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_10_6
#define init_3_13_10_6
void FASTCALL call_3_13_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_10_7
#define init_3_13_10_7
void FASTCALL call_3_13_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_10_8
#define init_3_13_10_8
void FASTCALL call_3_13_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_10_9
#define init_3_13_10_9
void FASTCALL call_3_13_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_10_10
#define init_3_13_10_10
void FASTCALL call_3_13_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_10_11
#define init_3_13_10_11
void FASTCALL call_3_13_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_10_12
#define init_3_13_10_12
void FASTCALL call_3_13_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_10_13
#define init_3_13_10_13
void FASTCALL call_3_13_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_10_14
#define init_3_13_10_14
void FASTCALL call_3_13_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_10_15
#define init_3_13_10_15
void FASTCALL call_3_13_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_11_0
#define init_3_13_11_0
void FASTCALL call_3_13_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_11_1
#define init_3_13_11_1
void FASTCALL call_3_13_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_11_2
#define init_3_13_11_2
void FASTCALL call_3_13_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_11_3
#define init_3_13_11_3
void FASTCALL call_3_13_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_11_4
#define init_3_13_11_4
void FASTCALL call_3_13_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_11_5
#define init_3_13_11_5
void FASTCALL call_3_13_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_11_6
#define init_3_13_11_6
void FASTCALL call_3_13_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_11_7
#define init_3_13_11_7
void FASTCALL call_3_13_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_11_8
#define init_3_13_11_8
void FASTCALL call_3_13_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_11_9
#define init_3_13_11_9
void FASTCALL call_3_13_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_11_10
#define init_3_13_11_10
void FASTCALL call_3_13_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_11_11
#define init_3_13_11_11
void FASTCALL call_3_13_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_11_12
#define init_3_13_11_12
void FASTCALL call_3_13_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_11_13
#define init_3_13_11_13
void FASTCALL call_3_13_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_11_14
#define init_3_13_11_14
void FASTCALL call_3_13_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_11_15
#define init_3_13_11_15
void FASTCALL call_3_13_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_12_0
#define init_3_13_12_0
void FASTCALL call_3_13_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_12_1
#define init_3_13_12_1
void FASTCALL call_3_13_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_12_2
#define init_3_13_12_2
void FASTCALL call_3_13_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_12_3
#define init_3_13_12_3
void FASTCALL call_3_13_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_12_4
#define init_3_13_12_4
void FASTCALL call_3_13_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_12_5
#define init_3_13_12_5
void FASTCALL call_3_13_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_12_6
#define init_3_13_12_6
void FASTCALL call_3_13_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_12_7
#define init_3_13_12_7
void FASTCALL call_3_13_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_12_8
#define init_3_13_12_8
void FASTCALL call_3_13_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_12_9
#define init_3_13_12_9
void FASTCALL call_3_13_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_12_10
#define init_3_13_12_10
void FASTCALL call_3_13_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_12_11
#define init_3_13_12_11
void FASTCALL call_3_13_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_12_12
#define init_3_13_12_12
void FASTCALL call_3_13_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_12_13
#define init_3_13_12_13
void FASTCALL call_3_13_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_12_14
#define init_3_13_12_14
void FASTCALL call_3_13_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_12_15
#define init_3_13_12_15
void FASTCALL call_3_13_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_13_0
#define init_3_13_13_0
void FASTCALL call_3_13_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_13_1
#define init_3_13_13_1
void FASTCALL call_3_13_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_13_2
#define init_3_13_13_2
void FASTCALL call_3_13_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_13_3
#define init_3_13_13_3
void FASTCALL call_3_13_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_13_4
#define init_3_13_13_4
void FASTCALL call_3_13_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_13_5
#define init_3_13_13_5
void FASTCALL call_3_13_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_13_6
#define init_3_13_13_6
void FASTCALL call_3_13_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_13_7
#define init_3_13_13_7
void FASTCALL call_3_13_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_13_8
#define init_3_13_13_8
void FASTCALL call_3_13_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_13_9
#define init_3_13_13_9
void FASTCALL call_3_13_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_13_10
#define init_3_13_13_10
void FASTCALL call_3_13_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_13_11
#define init_3_13_13_11
void FASTCALL call_3_13_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_13_12
#define init_3_13_13_12
void FASTCALL call_3_13_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_13_13
#define init_3_13_13_13
void FASTCALL call_3_13_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_13_14
#define init_3_13_13_14
void FASTCALL call_3_13_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_13_15
#define init_3_13_13_15
void FASTCALL call_3_13_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_14_0
#define init_3_13_14_0
void FASTCALL call_3_13_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_14_1
#define init_3_13_14_1
void FASTCALL call_3_13_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_14_2
#define init_3_13_14_2
void FASTCALL call_3_13_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_14_3
#define init_3_13_14_3
void FASTCALL call_3_13_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_14_4
#define init_3_13_14_4
void FASTCALL call_3_13_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_14_5
#define init_3_13_14_5
void FASTCALL call_3_13_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_14_6
#define init_3_13_14_6
void FASTCALL call_3_13_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_14_7
#define init_3_13_14_7
void FASTCALL call_3_13_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_14_8
#define init_3_13_14_8
void FASTCALL call_3_13_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_14_9
#define init_3_13_14_9
void FASTCALL call_3_13_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_14_10
#define init_3_13_14_10
void FASTCALL call_3_13_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_14_11
#define init_3_13_14_11
void FASTCALL call_3_13_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_14_12
#define init_3_13_14_12
void FASTCALL call_3_13_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_14_13
#define init_3_13_14_13
void FASTCALL call_3_13_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_14_14
#define init_3_13_14_14
void FASTCALL call_3_13_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_14_15
#define init_3_13_14_15
void FASTCALL call_3_13_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_15_0
#define init_3_13_15_0
void FASTCALL call_3_13_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_15_1
#define init_3_13_15_1
void FASTCALL call_3_13_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_15_2
#define init_3_13_15_2
void FASTCALL call_3_13_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_15_3
#define init_3_13_15_3
void FASTCALL call_3_13_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_15_4
#define init_3_13_15_4
void FASTCALL call_3_13_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_15_5
#define init_3_13_15_5
void FASTCALL call_3_13_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_15_6
#define init_3_13_15_6
void FASTCALL call_3_13_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_15_7
#define init_3_13_15_7
void FASTCALL call_3_13_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_15_8
#define init_3_13_15_8
void FASTCALL call_3_13_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_15_9
#define init_3_13_15_9
void FASTCALL call_3_13_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_15_10
#define init_3_13_15_10
void FASTCALL call_3_13_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_15_11
#define init_3_13_15_11
void FASTCALL call_3_13_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_15_12
#define init_3_13_15_12
void FASTCALL call_3_13_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_15_13
#define init_3_13_15_13
void FASTCALL call_3_13_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_15_14
#define init_3_13_15_14
void FASTCALL call_3_13_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_13_15_15
#define init_3_13_15_15
void FASTCALL call_3_13_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_0_0
#define init_3_14_0_0
void FASTCALL call_3_14_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_0_1
#define init_3_14_0_1
void FASTCALL call_3_14_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_0_2
#define init_3_14_0_2
void FASTCALL call_3_14_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_0_3
#define init_3_14_0_3
void FASTCALL call_3_14_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_0_4
#define init_3_14_0_4
void FASTCALL call_3_14_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_0_5
#define init_3_14_0_5
void FASTCALL call_3_14_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_0_6
#define init_3_14_0_6
void FASTCALL call_3_14_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_0_7
#define init_3_14_0_7
void FASTCALL call_3_14_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_0_8
#define init_3_14_0_8
void FASTCALL call_3_14_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_0_9
#define init_3_14_0_9
void FASTCALL call_3_14_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_0_10
#define init_3_14_0_10
void FASTCALL call_3_14_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_0_11
#define init_3_14_0_11
void FASTCALL call_3_14_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_0_12
#define init_3_14_0_12
void FASTCALL call_3_14_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_0_13
#define init_3_14_0_13
void FASTCALL call_3_14_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_0_14
#define init_3_14_0_14
void FASTCALL call_3_14_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_0_15
#define init_3_14_0_15
void FASTCALL call_3_14_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_1_0
#define init_3_14_1_0
void FASTCALL call_3_14_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_1_1
#define init_3_14_1_1
void FASTCALL call_3_14_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_1_2
#define init_3_14_1_2
void FASTCALL call_3_14_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_1_3
#define init_3_14_1_3
void FASTCALL call_3_14_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_1_4
#define init_3_14_1_4
void FASTCALL call_3_14_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_1_5
#define init_3_14_1_5
void FASTCALL call_3_14_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_1_6
#define init_3_14_1_6
void FASTCALL call_3_14_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_1_7
#define init_3_14_1_7
void FASTCALL call_3_14_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_1_8
#define init_3_14_1_8
void FASTCALL call_3_14_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_1_9
#define init_3_14_1_9
void FASTCALL call_3_14_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_1_10
#define init_3_14_1_10
void FASTCALL call_3_14_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_1_11
#define init_3_14_1_11
void FASTCALL call_3_14_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_1_12
#define init_3_14_1_12
void FASTCALL call_3_14_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_1_13
#define init_3_14_1_13
void FASTCALL call_3_14_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_1_14
#define init_3_14_1_14
void FASTCALL call_3_14_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_1_15
#define init_3_14_1_15
void FASTCALL call_3_14_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_2_0
#define init_3_14_2_0
void FASTCALL call_3_14_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_2_1
#define init_3_14_2_1
void FASTCALL call_3_14_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_2_2
#define init_3_14_2_2
void FASTCALL call_3_14_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_2_3
#define init_3_14_2_3
void FASTCALL call_3_14_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_2_4
#define init_3_14_2_4
void FASTCALL call_3_14_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_2_5
#define init_3_14_2_5
void FASTCALL call_3_14_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_2_6
#define init_3_14_2_6
void FASTCALL call_3_14_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_2_7
#define init_3_14_2_7
void FASTCALL call_3_14_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_2_8
#define init_3_14_2_8
void FASTCALL call_3_14_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_2_9
#define init_3_14_2_9
void FASTCALL call_3_14_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_2_10
#define init_3_14_2_10
void FASTCALL call_3_14_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_2_11
#define init_3_14_2_11
void FASTCALL call_3_14_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_2_12
#define init_3_14_2_12
void FASTCALL call_3_14_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_2_13
#define init_3_14_2_13
void FASTCALL call_3_14_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_2_14
#define init_3_14_2_14
void FASTCALL call_3_14_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_2_15
#define init_3_14_2_15
void FASTCALL call_3_14_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_3_0
#define init_3_14_3_0
void FASTCALL call_3_14_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_3_1
#define init_3_14_3_1
void FASTCALL call_3_14_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_3_2
#define init_3_14_3_2
void FASTCALL call_3_14_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_3_3
#define init_3_14_3_3
void FASTCALL call_3_14_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_3_4
#define init_3_14_3_4
void FASTCALL call_3_14_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_3_5
#define init_3_14_3_5
void FASTCALL call_3_14_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_3_6
#define init_3_14_3_6
void FASTCALL call_3_14_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_3_7
#define init_3_14_3_7
void FASTCALL call_3_14_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_3_8
#define init_3_14_3_8
void FASTCALL call_3_14_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_3_9
#define init_3_14_3_9
void FASTCALL call_3_14_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_3_10
#define init_3_14_3_10
void FASTCALL call_3_14_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_3_11
#define init_3_14_3_11
void FASTCALL call_3_14_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_3_12
#define init_3_14_3_12
void FASTCALL call_3_14_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_3_13
#define init_3_14_3_13
void FASTCALL call_3_14_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_3_14
#define init_3_14_3_14
void FASTCALL call_3_14_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_3_15
#define init_3_14_3_15
void FASTCALL call_3_14_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_4_0
#define init_3_14_4_0
void FASTCALL call_3_14_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_4_1
#define init_3_14_4_1
void FASTCALL call_3_14_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_4_2
#define init_3_14_4_2
void FASTCALL call_3_14_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_4_3
#define init_3_14_4_3
void FASTCALL call_3_14_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_4_4
#define init_3_14_4_4
void FASTCALL call_3_14_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_4_5
#define init_3_14_4_5
void FASTCALL call_3_14_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_4_6
#define init_3_14_4_6
void FASTCALL call_3_14_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_4_7
#define init_3_14_4_7
void FASTCALL call_3_14_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_4_8
#define init_3_14_4_8
void FASTCALL call_3_14_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_4_9
#define init_3_14_4_9
void FASTCALL call_3_14_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_4_10
#define init_3_14_4_10
void FASTCALL call_3_14_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_4_11
#define init_3_14_4_11
void FASTCALL call_3_14_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_4_12
#define init_3_14_4_12
void FASTCALL call_3_14_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_4_13
#define init_3_14_4_13
void FASTCALL call_3_14_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_4_14
#define init_3_14_4_14
void FASTCALL call_3_14_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_4_15
#define init_3_14_4_15
void FASTCALL call_3_14_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_5_0
#define init_3_14_5_0
void FASTCALL call_3_14_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_5_1
#define init_3_14_5_1
void FASTCALL call_3_14_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_5_2
#define init_3_14_5_2
void FASTCALL call_3_14_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_5_3
#define init_3_14_5_3
void FASTCALL call_3_14_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_5_4
#define init_3_14_5_4
void FASTCALL call_3_14_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_5_5
#define init_3_14_5_5
void FASTCALL call_3_14_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_5_6
#define init_3_14_5_6
void FASTCALL call_3_14_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_5_7
#define init_3_14_5_7
void FASTCALL call_3_14_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_5_8
#define init_3_14_5_8
void FASTCALL call_3_14_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_5_9
#define init_3_14_5_9
void FASTCALL call_3_14_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_5_10
#define init_3_14_5_10
void FASTCALL call_3_14_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_5_11
#define init_3_14_5_11
void FASTCALL call_3_14_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_5_12
#define init_3_14_5_12
void FASTCALL call_3_14_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_5_13
#define init_3_14_5_13
void FASTCALL call_3_14_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_5_14
#define init_3_14_5_14
void FASTCALL call_3_14_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_5_15
#define init_3_14_5_15
void FASTCALL call_3_14_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_6_0
#define init_3_14_6_0
void FASTCALL call_3_14_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_6_1
#define init_3_14_6_1
void FASTCALL call_3_14_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_6_2
#define init_3_14_6_2
void FASTCALL call_3_14_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_6_3
#define init_3_14_6_3
void FASTCALL call_3_14_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_6_4
#define init_3_14_6_4
void FASTCALL call_3_14_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_6_5
#define init_3_14_6_5
void FASTCALL call_3_14_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_6_6
#define init_3_14_6_6
void FASTCALL call_3_14_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_6_7
#define init_3_14_6_7
void FASTCALL call_3_14_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_6_8
#define init_3_14_6_8
void FASTCALL call_3_14_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_6_9
#define init_3_14_6_9
void FASTCALL call_3_14_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_6_10
#define init_3_14_6_10
void FASTCALL call_3_14_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_6_11
#define init_3_14_6_11
void FASTCALL call_3_14_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_6_12
#define init_3_14_6_12
void FASTCALL call_3_14_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_6_13
#define init_3_14_6_13
void FASTCALL call_3_14_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_6_14
#define init_3_14_6_14
void FASTCALL call_3_14_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_6_15
#define init_3_14_6_15
void FASTCALL call_3_14_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_7_0
#define init_3_14_7_0
void FASTCALL call_3_14_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_7_1
#define init_3_14_7_1
void FASTCALL call_3_14_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_7_2
#define init_3_14_7_2
void FASTCALL call_3_14_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_7_3
#define init_3_14_7_3
void FASTCALL call_3_14_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_7_4
#define init_3_14_7_4
void FASTCALL call_3_14_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_7_5
#define init_3_14_7_5
void FASTCALL call_3_14_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_7_6
#define init_3_14_7_6
void FASTCALL call_3_14_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_7_7
#define init_3_14_7_7
void FASTCALL call_3_14_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_7_8
#define init_3_14_7_8
void FASTCALL call_3_14_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_7_9
#define init_3_14_7_9
void FASTCALL call_3_14_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_7_10
#define init_3_14_7_10
void FASTCALL call_3_14_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_7_11
#define init_3_14_7_11
void FASTCALL call_3_14_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_7_12
#define init_3_14_7_12
void FASTCALL call_3_14_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_7_13
#define init_3_14_7_13
void FASTCALL call_3_14_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_7_14
#define init_3_14_7_14
void FASTCALL call_3_14_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_7_15
#define init_3_14_7_15
void FASTCALL call_3_14_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_8_0
#define init_3_14_8_0
void FASTCALL call_3_14_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_8_1
#define init_3_14_8_1
void FASTCALL call_3_14_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_8_2
#define init_3_14_8_2
void FASTCALL call_3_14_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_8_3
#define init_3_14_8_3
void FASTCALL call_3_14_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_8_4
#define init_3_14_8_4
void FASTCALL call_3_14_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_8_5
#define init_3_14_8_5
void FASTCALL call_3_14_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_8_6
#define init_3_14_8_6
void FASTCALL call_3_14_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_8_7
#define init_3_14_8_7
void FASTCALL call_3_14_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_8_8
#define init_3_14_8_8
void FASTCALL call_3_14_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_8_9
#define init_3_14_8_9
void FASTCALL call_3_14_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_8_10
#define init_3_14_8_10
void FASTCALL call_3_14_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_8_11
#define init_3_14_8_11
void FASTCALL call_3_14_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_8_12
#define init_3_14_8_12
void FASTCALL call_3_14_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_8_13
#define init_3_14_8_13
void FASTCALL call_3_14_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_8_14
#define init_3_14_8_14
void FASTCALL call_3_14_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_8_15
#define init_3_14_8_15
void FASTCALL call_3_14_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_9_0
#define init_3_14_9_0
void FASTCALL call_3_14_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_9_1
#define init_3_14_9_1
void FASTCALL call_3_14_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_9_2
#define init_3_14_9_2
void FASTCALL call_3_14_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_9_3
#define init_3_14_9_3
void FASTCALL call_3_14_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_9_4
#define init_3_14_9_4
void FASTCALL call_3_14_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_9_5
#define init_3_14_9_5
void FASTCALL call_3_14_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_9_6
#define init_3_14_9_6
void FASTCALL call_3_14_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_9_7
#define init_3_14_9_7
void FASTCALL call_3_14_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_9_8
#define init_3_14_9_8
void FASTCALL call_3_14_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_9_9
#define init_3_14_9_9
void FASTCALL call_3_14_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_9_10
#define init_3_14_9_10
void FASTCALL call_3_14_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_9_11
#define init_3_14_9_11
void FASTCALL call_3_14_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_9_12
#define init_3_14_9_12
void FASTCALL call_3_14_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_9_13
#define init_3_14_9_13
void FASTCALL call_3_14_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_9_14
#define init_3_14_9_14
void FASTCALL call_3_14_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_9_15
#define init_3_14_9_15
void FASTCALL call_3_14_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_10_0
#define init_3_14_10_0
void FASTCALL call_3_14_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_10_1
#define init_3_14_10_1
void FASTCALL call_3_14_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_10_2
#define init_3_14_10_2
void FASTCALL call_3_14_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_10_3
#define init_3_14_10_3
void FASTCALL call_3_14_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_10_4
#define init_3_14_10_4
void FASTCALL call_3_14_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_10_5
#define init_3_14_10_5
void FASTCALL call_3_14_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_10_6
#define init_3_14_10_6
void FASTCALL call_3_14_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_10_7
#define init_3_14_10_7
void FASTCALL call_3_14_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_10_8
#define init_3_14_10_8
void FASTCALL call_3_14_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_10_9
#define init_3_14_10_9
void FASTCALL call_3_14_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_10_10
#define init_3_14_10_10
void FASTCALL call_3_14_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_10_11
#define init_3_14_10_11
void FASTCALL call_3_14_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_10_12
#define init_3_14_10_12
void FASTCALL call_3_14_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_10_13
#define init_3_14_10_13
void FASTCALL call_3_14_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_10_14
#define init_3_14_10_14
void FASTCALL call_3_14_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_10_15
#define init_3_14_10_15
void FASTCALL call_3_14_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_11_0
#define init_3_14_11_0
void FASTCALL call_3_14_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_11_1
#define init_3_14_11_1
void FASTCALL call_3_14_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_11_2
#define init_3_14_11_2
void FASTCALL call_3_14_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_11_3
#define init_3_14_11_3
void FASTCALL call_3_14_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_11_4
#define init_3_14_11_4
void FASTCALL call_3_14_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_11_5
#define init_3_14_11_5
void FASTCALL call_3_14_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_11_6
#define init_3_14_11_6
void FASTCALL call_3_14_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_11_7
#define init_3_14_11_7
void FASTCALL call_3_14_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_11_8
#define init_3_14_11_8
void FASTCALL call_3_14_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_11_9
#define init_3_14_11_9
void FASTCALL call_3_14_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_11_10
#define init_3_14_11_10
void FASTCALL call_3_14_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_11_11
#define init_3_14_11_11
void FASTCALL call_3_14_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_11_12
#define init_3_14_11_12
void FASTCALL call_3_14_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_11_13
#define init_3_14_11_13
void FASTCALL call_3_14_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_11_14
#define init_3_14_11_14
void FASTCALL call_3_14_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_11_15
#define init_3_14_11_15
void FASTCALL call_3_14_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_12_0
#define init_3_14_12_0
void FASTCALL call_3_14_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_12_1
#define init_3_14_12_1
void FASTCALL call_3_14_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_12_2
#define init_3_14_12_2
void FASTCALL call_3_14_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_12_3
#define init_3_14_12_3
void FASTCALL call_3_14_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_12_4
#define init_3_14_12_4
void FASTCALL call_3_14_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_12_5
#define init_3_14_12_5
void FASTCALL call_3_14_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_12_6
#define init_3_14_12_6
void FASTCALL call_3_14_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_12_7
#define init_3_14_12_7
void FASTCALL call_3_14_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_12_8
#define init_3_14_12_8
void FASTCALL call_3_14_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_12_9
#define init_3_14_12_9
void FASTCALL call_3_14_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_12_10
#define init_3_14_12_10
void FASTCALL call_3_14_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_12_11
#define init_3_14_12_11
void FASTCALL call_3_14_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_12_12
#define init_3_14_12_12
void FASTCALL call_3_14_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_12_13
#define init_3_14_12_13
void FASTCALL call_3_14_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_12_14
#define init_3_14_12_14
void FASTCALL call_3_14_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_12_15
#define init_3_14_12_15
void FASTCALL call_3_14_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_13_0
#define init_3_14_13_0
void FASTCALL call_3_14_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_13_1
#define init_3_14_13_1
void FASTCALL call_3_14_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_13_2
#define init_3_14_13_2
void FASTCALL call_3_14_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_13_3
#define init_3_14_13_3
void FASTCALL call_3_14_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_13_4
#define init_3_14_13_4
void FASTCALL call_3_14_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_13_5
#define init_3_14_13_5
void FASTCALL call_3_14_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_13_6
#define init_3_14_13_6
void FASTCALL call_3_14_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_13_7
#define init_3_14_13_7
void FASTCALL call_3_14_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_13_8
#define init_3_14_13_8
void FASTCALL call_3_14_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_13_9
#define init_3_14_13_9
void FASTCALL call_3_14_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_13_10
#define init_3_14_13_10
void FASTCALL call_3_14_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_13_11
#define init_3_14_13_11
void FASTCALL call_3_14_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_13_12
#define init_3_14_13_12
void FASTCALL call_3_14_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_13_13
#define init_3_14_13_13
void FASTCALL call_3_14_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_13_14
#define init_3_14_13_14
void FASTCALL call_3_14_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_13_15
#define init_3_14_13_15
void FASTCALL call_3_14_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_14_0
#define init_3_14_14_0
void FASTCALL call_3_14_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_14_1
#define init_3_14_14_1
void FASTCALL call_3_14_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_14_2
#define init_3_14_14_2
void FASTCALL call_3_14_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_14_3
#define init_3_14_14_3
void FASTCALL call_3_14_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_14_4
#define init_3_14_14_4
void FASTCALL call_3_14_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_14_5
#define init_3_14_14_5
void FASTCALL call_3_14_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_14_6
#define init_3_14_14_6
void FASTCALL call_3_14_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_14_7
#define init_3_14_14_7
void FASTCALL call_3_14_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_14_8
#define init_3_14_14_8
void FASTCALL call_3_14_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_14_9
#define init_3_14_14_9
void FASTCALL call_3_14_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_14_10
#define init_3_14_14_10
void FASTCALL call_3_14_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_14_11
#define init_3_14_14_11
void FASTCALL call_3_14_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_14_12
#define init_3_14_14_12
void FASTCALL call_3_14_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_14_13
#define init_3_14_14_13
void FASTCALL call_3_14_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_14_14
#define init_3_14_14_14
void FASTCALL call_3_14_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_14_15
#define init_3_14_14_15
void FASTCALL call_3_14_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_15_0
#define init_3_14_15_0
void FASTCALL call_3_14_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_15_1
#define init_3_14_15_1
void FASTCALL call_3_14_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_15_2
#define init_3_14_15_2
void FASTCALL call_3_14_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_15_3
#define init_3_14_15_3
void FASTCALL call_3_14_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_15_4
#define init_3_14_15_4
void FASTCALL call_3_14_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_15_5
#define init_3_14_15_5
void FASTCALL call_3_14_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_15_6
#define init_3_14_15_6
void FASTCALL call_3_14_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_15_7
#define init_3_14_15_7
void FASTCALL call_3_14_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_15_8
#define init_3_14_15_8
void FASTCALL call_3_14_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_15_9
#define init_3_14_15_9
void FASTCALL call_3_14_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_15_10
#define init_3_14_15_10
void FASTCALL call_3_14_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_15_11
#define init_3_14_15_11
void FASTCALL call_3_14_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_15_12
#define init_3_14_15_12
void FASTCALL call_3_14_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_15_13
#define init_3_14_15_13
void FASTCALL call_3_14_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_15_14
#define init_3_14_15_14
void FASTCALL call_3_14_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_14_15_15
#define init_3_14_15_15
void FASTCALL call_3_14_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_0_0
#define init_3_15_0_0
void FASTCALL call_3_15_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_0_1
#define init_3_15_0_1
void FASTCALL call_3_15_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_0_2
#define init_3_15_0_2
void FASTCALL call_3_15_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_0_3
#define init_3_15_0_3
void FASTCALL call_3_15_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_0_4
#define init_3_15_0_4
void FASTCALL call_3_15_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_0_5
#define init_3_15_0_5
void FASTCALL call_3_15_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_0_6
#define init_3_15_0_6
void FASTCALL call_3_15_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_0_7
#define init_3_15_0_7
void FASTCALL call_3_15_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_0_8
#define init_3_15_0_8
void FASTCALL call_3_15_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_0_9
#define init_3_15_0_9
void FASTCALL call_3_15_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_0_10
#define init_3_15_0_10
void FASTCALL call_3_15_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_0_11
#define init_3_15_0_11
void FASTCALL call_3_15_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_0_12
#define init_3_15_0_12
void FASTCALL call_3_15_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_0_13
#define init_3_15_0_13
void FASTCALL call_3_15_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_0_14
#define init_3_15_0_14
void FASTCALL call_3_15_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_0_15
#define init_3_15_0_15
void FASTCALL call_3_15_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_1_0
#define init_3_15_1_0
void FASTCALL call_3_15_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_1_1
#define init_3_15_1_1
void FASTCALL call_3_15_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_1_2
#define init_3_15_1_2
void FASTCALL call_3_15_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_1_3
#define init_3_15_1_3
void FASTCALL call_3_15_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_1_4
#define init_3_15_1_4
void FASTCALL call_3_15_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_1_5
#define init_3_15_1_5
void FASTCALL call_3_15_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_1_6
#define init_3_15_1_6
void FASTCALL call_3_15_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_1_7
#define init_3_15_1_7
void FASTCALL call_3_15_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_1_8
#define init_3_15_1_8
void FASTCALL call_3_15_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_1_9
#define init_3_15_1_9
void FASTCALL call_3_15_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_1_10
#define init_3_15_1_10
void FASTCALL call_3_15_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_1_11
#define init_3_15_1_11
void FASTCALL call_3_15_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_1_12
#define init_3_15_1_12
void FASTCALL call_3_15_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_1_13
#define init_3_15_1_13
void FASTCALL call_3_15_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_1_14
#define init_3_15_1_14
void FASTCALL call_3_15_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_1_15
#define init_3_15_1_15
void FASTCALL call_3_15_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_2_0
#define init_3_15_2_0
void FASTCALL call_3_15_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_2_1
#define init_3_15_2_1
void FASTCALL call_3_15_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_2_2
#define init_3_15_2_2
void FASTCALL call_3_15_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_2_3
#define init_3_15_2_3
void FASTCALL call_3_15_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_2_4
#define init_3_15_2_4
void FASTCALL call_3_15_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_2_5
#define init_3_15_2_5
void FASTCALL call_3_15_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_2_6
#define init_3_15_2_6
void FASTCALL call_3_15_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_2_7
#define init_3_15_2_7
void FASTCALL call_3_15_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_2_8
#define init_3_15_2_8
void FASTCALL call_3_15_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_2_9
#define init_3_15_2_9
void FASTCALL call_3_15_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_2_10
#define init_3_15_2_10
void FASTCALL call_3_15_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_2_11
#define init_3_15_2_11
void FASTCALL call_3_15_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_2_12
#define init_3_15_2_12
void FASTCALL call_3_15_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_2_13
#define init_3_15_2_13
void FASTCALL call_3_15_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_2_14
#define init_3_15_2_14
void FASTCALL call_3_15_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_2_15
#define init_3_15_2_15
void FASTCALL call_3_15_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_3_0
#define init_3_15_3_0
void FASTCALL call_3_15_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_3_1
#define init_3_15_3_1
void FASTCALL call_3_15_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_3_2
#define init_3_15_3_2
void FASTCALL call_3_15_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_3_3
#define init_3_15_3_3
void FASTCALL call_3_15_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_3_4
#define init_3_15_3_4
void FASTCALL call_3_15_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_3_5
#define init_3_15_3_5
void FASTCALL call_3_15_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_3_6
#define init_3_15_3_6
void FASTCALL call_3_15_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_3_7
#define init_3_15_3_7
void FASTCALL call_3_15_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_3_8
#define init_3_15_3_8
void FASTCALL call_3_15_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_3_9
#define init_3_15_3_9
void FASTCALL call_3_15_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_3_10
#define init_3_15_3_10
void FASTCALL call_3_15_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_3_11
#define init_3_15_3_11
void FASTCALL call_3_15_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_3_12
#define init_3_15_3_12
void FASTCALL call_3_15_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_3_13
#define init_3_15_3_13
void FASTCALL call_3_15_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_3_14
#define init_3_15_3_14
void FASTCALL call_3_15_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_3_15
#define init_3_15_3_15
void FASTCALL call_3_15_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_4_0
#define init_3_15_4_0
void FASTCALL call_3_15_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_4_1
#define init_3_15_4_1
void FASTCALL call_3_15_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_4_2
#define init_3_15_4_2
void FASTCALL call_3_15_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_4_3
#define init_3_15_4_3
void FASTCALL call_3_15_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_4_4
#define init_3_15_4_4
void FASTCALL call_3_15_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_4_5
#define init_3_15_4_5
void FASTCALL call_3_15_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_4_6
#define init_3_15_4_6
void FASTCALL call_3_15_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_4_7
#define init_3_15_4_7
void FASTCALL call_3_15_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_4_8
#define init_3_15_4_8
void FASTCALL call_3_15_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_4_9
#define init_3_15_4_9
void FASTCALL call_3_15_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_4_10
#define init_3_15_4_10
void FASTCALL call_3_15_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_4_11
#define init_3_15_4_11
void FASTCALL call_3_15_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_4_12
#define init_3_15_4_12
void FASTCALL call_3_15_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_4_13
#define init_3_15_4_13
void FASTCALL call_3_15_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_4_14
#define init_3_15_4_14
void FASTCALL call_3_15_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_4_15
#define init_3_15_4_15
void FASTCALL call_3_15_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_5_0
#define init_3_15_5_0
void FASTCALL call_3_15_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_5_1
#define init_3_15_5_1
void FASTCALL call_3_15_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_5_2
#define init_3_15_5_2
void FASTCALL call_3_15_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_5_3
#define init_3_15_5_3
void FASTCALL call_3_15_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_5_4
#define init_3_15_5_4
void FASTCALL call_3_15_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_5_5
#define init_3_15_5_5
void FASTCALL call_3_15_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_5_6
#define init_3_15_5_6
void FASTCALL call_3_15_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_5_7
#define init_3_15_5_7
void FASTCALL call_3_15_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_5_8
#define init_3_15_5_8
void FASTCALL call_3_15_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_5_9
#define init_3_15_5_9
void FASTCALL call_3_15_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_5_10
#define init_3_15_5_10
void FASTCALL call_3_15_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_5_11
#define init_3_15_5_11
void FASTCALL call_3_15_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_5_12
#define init_3_15_5_12
void FASTCALL call_3_15_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_5_13
#define init_3_15_5_13
void FASTCALL call_3_15_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_5_14
#define init_3_15_5_14
void FASTCALL call_3_15_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_5_15
#define init_3_15_5_15
void FASTCALL call_3_15_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_6_0
#define init_3_15_6_0
void FASTCALL call_3_15_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_6_1
#define init_3_15_6_1
void FASTCALL call_3_15_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_6_2
#define init_3_15_6_2
void FASTCALL call_3_15_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_6_3
#define init_3_15_6_3
void FASTCALL call_3_15_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_6_4
#define init_3_15_6_4
void FASTCALL call_3_15_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_6_5
#define init_3_15_6_5
void FASTCALL call_3_15_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_6_6
#define init_3_15_6_6
void FASTCALL call_3_15_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_6_7
#define init_3_15_6_7
void FASTCALL call_3_15_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_6_8
#define init_3_15_6_8
void FASTCALL call_3_15_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_6_9
#define init_3_15_6_9
void FASTCALL call_3_15_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_6_10
#define init_3_15_6_10
void FASTCALL call_3_15_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_6_11
#define init_3_15_6_11
void FASTCALL call_3_15_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_6_12
#define init_3_15_6_12
void FASTCALL call_3_15_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_6_13
#define init_3_15_6_13
void FASTCALL call_3_15_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_6_14
#define init_3_15_6_14
void FASTCALL call_3_15_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_6_15
#define init_3_15_6_15
void FASTCALL call_3_15_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_7_0
#define init_3_15_7_0
void FASTCALL call_3_15_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_7_1
#define init_3_15_7_1
void FASTCALL call_3_15_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_7_2
#define init_3_15_7_2
void FASTCALL call_3_15_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_7_3
#define init_3_15_7_3
void FASTCALL call_3_15_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_7_4
#define init_3_15_7_4
void FASTCALL call_3_15_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_7_5
#define init_3_15_7_5
void FASTCALL call_3_15_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_7_6
#define init_3_15_7_6
void FASTCALL call_3_15_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_7_7
#define init_3_15_7_7
void FASTCALL call_3_15_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_7_8
#define init_3_15_7_8
void FASTCALL call_3_15_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_7_9
#define init_3_15_7_9
void FASTCALL call_3_15_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_7_10
#define init_3_15_7_10
void FASTCALL call_3_15_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_7_11
#define init_3_15_7_11
void FASTCALL call_3_15_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_7_12
#define init_3_15_7_12
void FASTCALL call_3_15_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_7_13
#define init_3_15_7_13
void FASTCALL call_3_15_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_7_14
#define init_3_15_7_14
void FASTCALL call_3_15_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_7_15
#define init_3_15_7_15
void FASTCALL call_3_15_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_8_0
#define init_3_15_8_0
void FASTCALL call_3_15_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_8_1
#define init_3_15_8_1
void FASTCALL call_3_15_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_8_2
#define init_3_15_8_2
void FASTCALL call_3_15_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_8_3
#define init_3_15_8_3
void FASTCALL call_3_15_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_8_4
#define init_3_15_8_4
void FASTCALL call_3_15_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_8_5
#define init_3_15_8_5
void FASTCALL call_3_15_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_8_6
#define init_3_15_8_6
void FASTCALL call_3_15_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_8_7
#define init_3_15_8_7
void FASTCALL call_3_15_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_8_8
#define init_3_15_8_8
void FASTCALL call_3_15_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_8_9
#define init_3_15_8_9
void FASTCALL call_3_15_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_8_10
#define init_3_15_8_10
void FASTCALL call_3_15_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_8_11
#define init_3_15_8_11
void FASTCALL call_3_15_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_8_12
#define init_3_15_8_12
void FASTCALL call_3_15_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_8_13
#define init_3_15_8_13
void FASTCALL call_3_15_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_8_14
#define init_3_15_8_14
void FASTCALL call_3_15_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_8_15
#define init_3_15_8_15
void FASTCALL call_3_15_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_9_0
#define init_3_15_9_0
void FASTCALL call_3_15_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_9_1
#define init_3_15_9_1
void FASTCALL call_3_15_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_9_2
#define init_3_15_9_2
void FASTCALL call_3_15_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_9_3
#define init_3_15_9_3
void FASTCALL call_3_15_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_9_4
#define init_3_15_9_4
void FASTCALL call_3_15_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_9_5
#define init_3_15_9_5
void FASTCALL call_3_15_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_9_6
#define init_3_15_9_6
void FASTCALL call_3_15_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_9_7
#define init_3_15_9_7
void FASTCALL call_3_15_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_9_8
#define init_3_15_9_8
void FASTCALL call_3_15_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_9_9
#define init_3_15_9_9
void FASTCALL call_3_15_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_9_10
#define init_3_15_9_10
void FASTCALL call_3_15_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_9_11
#define init_3_15_9_11
void FASTCALL call_3_15_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_9_12
#define init_3_15_9_12
void FASTCALL call_3_15_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_9_13
#define init_3_15_9_13
void FASTCALL call_3_15_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_9_14
#define init_3_15_9_14
void FASTCALL call_3_15_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_9_15
#define init_3_15_9_15
void FASTCALL call_3_15_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_10_0
#define init_3_15_10_0
void FASTCALL call_3_15_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_10_1
#define init_3_15_10_1
void FASTCALL call_3_15_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_10_2
#define init_3_15_10_2
void FASTCALL call_3_15_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_10_3
#define init_3_15_10_3
void FASTCALL call_3_15_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_10_4
#define init_3_15_10_4
void FASTCALL call_3_15_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_10_5
#define init_3_15_10_5
void FASTCALL call_3_15_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_10_6
#define init_3_15_10_6
void FASTCALL call_3_15_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_10_7
#define init_3_15_10_7
void FASTCALL call_3_15_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_10_8
#define init_3_15_10_8
void FASTCALL call_3_15_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_10_9
#define init_3_15_10_9
void FASTCALL call_3_15_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_10_10
#define init_3_15_10_10
void FASTCALL call_3_15_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_10_11
#define init_3_15_10_11
void FASTCALL call_3_15_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_10_12
#define init_3_15_10_12
void FASTCALL call_3_15_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_10_13
#define init_3_15_10_13
void FASTCALL call_3_15_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_10_14
#define init_3_15_10_14
void FASTCALL call_3_15_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_10_15
#define init_3_15_10_15
void FASTCALL call_3_15_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_11_0
#define init_3_15_11_0
void FASTCALL call_3_15_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_11_1
#define init_3_15_11_1
void FASTCALL call_3_15_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_11_2
#define init_3_15_11_2
void FASTCALL call_3_15_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_11_3
#define init_3_15_11_3
void FASTCALL call_3_15_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_11_4
#define init_3_15_11_4
void FASTCALL call_3_15_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_11_5
#define init_3_15_11_5
void FASTCALL call_3_15_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_11_6
#define init_3_15_11_6
void FASTCALL call_3_15_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_11_7
#define init_3_15_11_7
void FASTCALL call_3_15_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_11_8
#define init_3_15_11_8
void FASTCALL call_3_15_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_11_9
#define init_3_15_11_9
void FASTCALL call_3_15_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_11_10
#define init_3_15_11_10
void FASTCALL call_3_15_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_11_11
#define init_3_15_11_11
void FASTCALL call_3_15_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_11_12
#define init_3_15_11_12
void FASTCALL call_3_15_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_11_13
#define init_3_15_11_13
void FASTCALL call_3_15_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_11_14
#define init_3_15_11_14
void FASTCALL call_3_15_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_11_15
#define init_3_15_11_15
void FASTCALL call_3_15_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_12_0
#define init_3_15_12_0
void FASTCALL call_3_15_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_12_1
#define init_3_15_12_1
void FASTCALL call_3_15_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_12_2
#define init_3_15_12_2
void FASTCALL call_3_15_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_12_3
#define init_3_15_12_3
void FASTCALL call_3_15_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_12_4
#define init_3_15_12_4
void FASTCALL call_3_15_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_12_5
#define init_3_15_12_5
void FASTCALL call_3_15_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_12_6
#define init_3_15_12_6
void FASTCALL call_3_15_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_12_7
#define init_3_15_12_7
void FASTCALL call_3_15_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_12_8
#define init_3_15_12_8
void FASTCALL call_3_15_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_12_9
#define init_3_15_12_9
void FASTCALL call_3_15_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_12_10
#define init_3_15_12_10
void FASTCALL call_3_15_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_12_11
#define init_3_15_12_11
void FASTCALL call_3_15_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_12_12
#define init_3_15_12_12
void FASTCALL call_3_15_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_12_13
#define init_3_15_12_13
void FASTCALL call_3_15_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_12_14
#define init_3_15_12_14
void FASTCALL call_3_15_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_12_15
#define init_3_15_12_15
void FASTCALL call_3_15_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_13_0
#define init_3_15_13_0
void FASTCALL call_3_15_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_13_1
#define init_3_15_13_1
void FASTCALL call_3_15_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_13_2
#define init_3_15_13_2
void FASTCALL call_3_15_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_13_3
#define init_3_15_13_3
void FASTCALL call_3_15_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_13_4
#define init_3_15_13_4
void FASTCALL call_3_15_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_13_5
#define init_3_15_13_5
void FASTCALL call_3_15_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_13_6
#define init_3_15_13_6
void FASTCALL call_3_15_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_13_7
#define init_3_15_13_7
void FASTCALL call_3_15_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_13_8
#define init_3_15_13_8
void FASTCALL call_3_15_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_13_9
#define init_3_15_13_9
void FASTCALL call_3_15_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_13_10
#define init_3_15_13_10
void FASTCALL call_3_15_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_13_11
#define init_3_15_13_11
void FASTCALL call_3_15_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_13_12
#define init_3_15_13_12
void FASTCALL call_3_15_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_13_13
#define init_3_15_13_13
void FASTCALL call_3_15_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_13_14
#define init_3_15_13_14
void FASTCALL call_3_15_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_13_15
#define init_3_15_13_15
void FASTCALL call_3_15_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_14_0
#define init_3_15_14_0
void FASTCALL call_3_15_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_14_1
#define init_3_15_14_1
void FASTCALL call_3_15_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_14_2
#define init_3_15_14_2
void FASTCALL call_3_15_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_14_3
#define init_3_15_14_3
void FASTCALL call_3_15_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_14_4
#define init_3_15_14_4
void FASTCALL call_3_15_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_14_5
#define init_3_15_14_5
void FASTCALL call_3_15_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_14_6
#define init_3_15_14_6
void FASTCALL call_3_15_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_14_7
#define init_3_15_14_7
void FASTCALL call_3_15_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_14_8
#define init_3_15_14_8
void FASTCALL call_3_15_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_14_9
#define init_3_15_14_9
void FASTCALL call_3_15_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_14_10
#define init_3_15_14_10
void FASTCALL call_3_15_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_14_11
#define init_3_15_14_11
void FASTCALL call_3_15_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_14_12
#define init_3_15_14_12
void FASTCALL call_3_15_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_14_13
#define init_3_15_14_13
void FASTCALL call_3_15_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_14_14
#define init_3_15_14_14
void FASTCALL call_3_15_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_14_15
#define init_3_15_14_15
void FASTCALL call_3_15_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_15_0
#define init_3_15_15_0
void FASTCALL call_3_15_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_15_1
#define init_3_15_15_1
void FASTCALL call_3_15_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_15_2
#define init_3_15_15_2
void FASTCALL call_3_15_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_15_3
#define init_3_15_15_3
void FASTCALL call_3_15_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_15_4
#define init_3_15_15_4
void FASTCALL call_3_15_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_15_5
#define init_3_15_15_5
void FASTCALL call_3_15_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_15_6
#define init_3_15_15_6
void FASTCALL call_3_15_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_15_7
#define init_3_15_15_7
void FASTCALL call_3_15_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_15_8
#define init_3_15_15_8
void FASTCALL call_3_15_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_15_9
#define init_3_15_15_9
void FASTCALL call_3_15_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_15_10
#define init_3_15_15_10
void FASTCALL call_3_15_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_15_11
#define init_3_15_15_11
void FASTCALL call_3_15_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_15_12
#define init_3_15_15_12
void FASTCALL call_3_15_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_15_13
#define init_3_15_15_13
void FASTCALL call_3_15_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_15_14
#define init_3_15_15_14
void FASTCALL call_3_15_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_3_15_15_15
#define init_3_15_15_15
void FASTCALL call_3_15_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_0_0
#define init_4_0_0_0
void FASTCALL call_4_0_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_0_1
#define init_4_0_0_1
void FASTCALL call_4_0_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_0_2
#define init_4_0_0_2
void FASTCALL call_4_0_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_0_3
#define init_4_0_0_3
void FASTCALL call_4_0_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_0_4
#define init_4_0_0_4
void FASTCALL call_4_0_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_0_5
#define init_4_0_0_5
void FASTCALL call_4_0_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_0_6
#define init_4_0_0_6
void FASTCALL call_4_0_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_0_7
#define init_4_0_0_7
void FASTCALL call_4_0_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_0_8
#define init_4_0_0_8
void FASTCALL call_4_0_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_0_9
#define init_4_0_0_9
void FASTCALL call_4_0_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_0_10
#define init_4_0_0_10
void FASTCALL call_4_0_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_0_11
#define init_4_0_0_11
void FASTCALL call_4_0_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_0_12
#define init_4_0_0_12
void FASTCALL call_4_0_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_0_13
#define init_4_0_0_13
void FASTCALL call_4_0_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_0_14
#define init_4_0_0_14
void FASTCALL call_4_0_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_0_15
#define init_4_0_0_15
void FASTCALL call_4_0_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_1_0
#define init_4_0_1_0
void FASTCALL call_4_0_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_1_1
#define init_4_0_1_1
void FASTCALL call_4_0_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_1_2
#define init_4_0_1_2
void FASTCALL call_4_0_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_1_3
#define init_4_0_1_3
void FASTCALL call_4_0_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_1_4
#define init_4_0_1_4
void FASTCALL call_4_0_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_1_5
#define init_4_0_1_5
void FASTCALL call_4_0_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_1_6
#define init_4_0_1_6
void FASTCALL call_4_0_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_1_7
#define init_4_0_1_7
void FASTCALL call_4_0_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_1_8
#define init_4_0_1_8
void FASTCALL call_4_0_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_1_9
#define init_4_0_1_9
void FASTCALL call_4_0_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_1_10
#define init_4_0_1_10
void FASTCALL call_4_0_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_1_11
#define init_4_0_1_11
void FASTCALL call_4_0_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_1_12
#define init_4_0_1_12
void FASTCALL call_4_0_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_1_13
#define init_4_0_1_13
void FASTCALL call_4_0_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_1_14
#define init_4_0_1_14
void FASTCALL call_4_0_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_1_15
#define init_4_0_1_15
void FASTCALL call_4_0_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_2_0
#define init_4_0_2_0
void FASTCALL call_4_0_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_2_1
#define init_4_0_2_1
void FASTCALL call_4_0_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_2_2
#define init_4_0_2_2
void FASTCALL call_4_0_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_2_3
#define init_4_0_2_3
void FASTCALL call_4_0_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_2_4
#define init_4_0_2_4
void FASTCALL call_4_0_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_2_5
#define init_4_0_2_5
void FASTCALL call_4_0_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_2_6
#define init_4_0_2_6
void FASTCALL call_4_0_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_2_7
#define init_4_0_2_7
void FASTCALL call_4_0_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_2_8
#define init_4_0_2_8
void FASTCALL call_4_0_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_2_9
#define init_4_0_2_9
void FASTCALL call_4_0_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_2_10
#define init_4_0_2_10
void FASTCALL call_4_0_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_2_11
#define init_4_0_2_11
void FASTCALL call_4_0_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_2_12
#define init_4_0_2_12
void FASTCALL call_4_0_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_2_13
#define init_4_0_2_13
void FASTCALL call_4_0_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_2_14
#define init_4_0_2_14
void FASTCALL call_4_0_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_2_15
#define init_4_0_2_15
void FASTCALL call_4_0_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_3_0
#define init_4_0_3_0
void FASTCALL call_4_0_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_3_1
#define init_4_0_3_1
void FASTCALL call_4_0_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_3_2
#define init_4_0_3_2
void FASTCALL call_4_0_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_3_3
#define init_4_0_3_3
void FASTCALL call_4_0_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_3_4
#define init_4_0_3_4
void FASTCALL call_4_0_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_3_5
#define init_4_0_3_5
void FASTCALL call_4_0_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_3_6
#define init_4_0_3_6
void FASTCALL call_4_0_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_3_7
#define init_4_0_3_7
void FASTCALL call_4_0_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_3_8
#define init_4_0_3_8
void FASTCALL call_4_0_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_3_9
#define init_4_0_3_9
void FASTCALL call_4_0_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_3_10
#define init_4_0_3_10
void FASTCALL call_4_0_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_3_11
#define init_4_0_3_11
void FASTCALL call_4_0_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_3_12
#define init_4_0_3_12
void FASTCALL call_4_0_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_3_13
#define init_4_0_3_13
void FASTCALL call_4_0_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_3_14
#define init_4_0_3_14
void FASTCALL call_4_0_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_3_15
#define init_4_0_3_15
void FASTCALL call_4_0_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_4_0
#define init_4_0_4_0
void FASTCALL call_4_0_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_4_1
#define init_4_0_4_1
void FASTCALL call_4_0_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_4_2
#define init_4_0_4_2
void FASTCALL call_4_0_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_4_3
#define init_4_0_4_3
void FASTCALL call_4_0_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_4_4
#define init_4_0_4_4
void FASTCALL call_4_0_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_4_5
#define init_4_0_4_5
void FASTCALL call_4_0_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_4_6
#define init_4_0_4_6
void FASTCALL call_4_0_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_4_7
#define init_4_0_4_7
void FASTCALL call_4_0_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_4_8
#define init_4_0_4_8
void FASTCALL call_4_0_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_4_9
#define init_4_0_4_9
void FASTCALL call_4_0_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_4_10
#define init_4_0_4_10
void FASTCALL call_4_0_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_4_11
#define init_4_0_4_11
void FASTCALL call_4_0_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_4_12
#define init_4_0_4_12
void FASTCALL call_4_0_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_4_13
#define init_4_0_4_13
void FASTCALL call_4_0_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_4_14
#define init_4_0_4_14
void FASTCALL call_4_0_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_4_15
#define init_4_0_4_15
void FASTCALL call_4_0_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_5_0
#define init_4_0_5_0
void FASTCALL call_4_0_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_5_1
#define init_4_0_5_1
void FASTCALL call_4_0_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_5_2
#define init_4_0_5_2
void FASTCALL call_4_0_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_5_3
#define init_4_0_5_3
void FASTCALL call_4_0_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_5_4
#define init_4_0_5_4
void FASTCALL call_4_0_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_5_5
#define init_4_0_5_5
void FASTCALL call_4_0_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_5_6
#define init_4_0_5_6
void FASTCALL call_4_0_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_5_7
#define init_4_0_5_7
void FASTCALL call_4_0_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_5_8
#define init_4_0_5_8
void FASTCALL call_4_0_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_5_9
#define init_4_0_5_9
void FASTCALL call_4_0_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_5_10
#define init_4_0_5_10
void FASTCALL call_4_0_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_5_11
#define init_4_0_5_11
void FASTCALL call_4_0_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_5_12
#define init_4_0_5_12
void FASTCALL call_4_0_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_5_13
#define init_4_0_5_13
void FASTCALL call_4_0_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_5_14
#define init_4_0_5_14
void FASTCALL call_4_0_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_5_15
#define init_4_0_5_15
void FASTCALL call_4_0_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_6_0
#define init_4_0_6_0
void FASTCALL call_4_0_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_6_1
#define init_4_0_6_1
void FASTCALL call_4_0_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_6_2
#define init_4_0_6_2
void FASTCALL call_4_0_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_6_3
#define init_4_0_6_3
void FASTCALL call_4_0_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_6_4
#define init_4_0_6_4
void FASTCALL call_4_0_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_6_5
#define init_4_0_6_5
void FASTCALL call_4_0_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_6_6
#define init_4_0_6_6
void FASTCALL call_4_0_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_6_7
#define init_4_0_6_7
void FASTCALL call_4_0_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_6_8
#define init_4_0_6_8
void FASTCALL call_4_0_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_6_9
#define init_4_0_6_9
void FASTCALL call_4_0_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_6_10
#define init_4_0_6_10
void FASTCALL call_4_0_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_6_11
#define init_4_0_6_11
void FASTCALL call_4_0_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_6_12
#define init_4_0_6_12
void FASTCALL call_4_0_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_6_13
#define init_4_0_6_13
void FASTCALL call_4_0_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_6_14
#define init_4_0_6_14
void FASTCALL call_4_0_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_6_15
#define init_4_0_6_15
void FASTCALL call_4_0_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_7_0
#define init_4_0_7_0
void FASTCALL call_4_0_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_7_1
#define init_4_0_7_1
void FASTCALL call_4_0_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_7_2
#define init_4_0_7_2
void FASTCALL call_4_0_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_7_3
#define init_4_0_7_3
void FASTCALL call_4_0_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_7_4
#define init_4_0_7_4
void FASTCALL call_4_0_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_7_5
#define init_4_0_7_5
void FASTCALL call_4_0_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_7_6
#define init_4_0_7_6
void FASTCALL call_4_0_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_7_7
#define init_4_0_7_7
void FASTCALL call_4_0_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_7_8
#define init_4_0_7_8
void FASTCALL call_4_0_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_7_9
#define init_4_0_7_9
void FASTCALL call_4_0_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_7_10
#define init_4_0_7_10
void FASTCALL call_4_0_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_7_11
#define init_4_0_7_11
void FASTCALL call_4_0_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_7_12
#define init_4_0_7_12
void FASTCALL call_4_0_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_7_13
#define init_4_0_7_13
void FASTCALL call_4_0_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_7_14
#define init_4_0_7_14
void FASTCALL call_4_0_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_7_15
#define init_4_0_7_15
void FASTCALL call_4_0_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_8_0
#define init_4_0_8_0
void FASTCALL call_4_0_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_8_1
#define init_4_0_8_1
void FASTCALL call_4_0_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_8_2
#define init_4_0_8_2
void FASTCALL call_4_0_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_8_3
#define init_4_0_8_3
void FASTCALL call_4_0_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_8_4
#define init_4_0_8_4
void FASTCALL call_4_0_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_8_5
#define init_4_0_8_5
void FASTCALL call_4_0_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_8_6
#define init_4_0_8_6
void FASTCALL call_4_0_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_8_7
#define init_4_0_8_7
void FASTCALL call_4_0_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_8_8
#define init_4_0_8_8
void FASTCALL call_4_0_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_8_9
#define init_4_0_8_9
void FASTCALL call_4_0_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_8_10
#define init_4_0_8_10
void FASTCALL call_4_0_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_8_11
#define init_4_0_8_11
void FASTCALL call_4_0_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_8_12
#define init_4_0_8_12
void FASTCALL call_4_0_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_8_13
#define init_4_0_8_13
void FASTCALL call_4_0_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_8_14
#define init_4_0_8_14
void FASTCALL call_4_0_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_8_15
#define init_4_0_8_15
void FASTCALL call_4_0_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_9_0
#define init_4_0_9_0
void FASTCALL call_4_0_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_9_1
#define init_4_0_9_1
void FASTCALL call_4_0_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_9_2
#define init_4_0_9_2
void FASTCALL call_4_0_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_9_3
#define init_4_0_9_3
void FASTCALL call_4_0_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_9_4
#define init_4_0_9_4
void FASTCALL call_4_0_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_9_5
#define init_4_0_9_5
void FASTCALL call_4_0_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_9_6
#define init_4_0_9_6
void FASTCALL call_4_0_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_9_7
#define init_4_0_9_7
void FASTCALL call_4_0_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_9_8
#define init_4_0_9_8
void FASTCALL call_4_0_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_9_9
#define init_4_0_9_9
void FASTCALL call_4_0_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_9_10
#define init_4_0_9_10
void FASTCALL call_4_0_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_9_11
#define init_4_0_9_11
void FASTCALL call_4_0_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_9_12
#define init_4_0_9_12
void FASTCALL call_4_0_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_9_13
#define init_4_0_9_13
void FASTCALL call_4_0_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_9_14
#define init_4_0_9_14
void FASTCALL call_4_0_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_9_15
#define init_4_0_9_15
void FASTCALL call_4_0_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_10_0
#define init_4_0_10_0
void FASTCALL call_4_0_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_10_1
#define init_4_0_10_1
void FASTCALL call_4_0_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_10_2
#define init_4_0_10_2
void FASTCALL call_4_0_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_10_3
#define init_4_0_10_3
void FASTCALL call_4_0_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_10_4
#define init_4_0_10_4
void FASTCALL call_4_0_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_10_5
#define init_4_0_10_5
void FASTCALL call_4_0_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_10_6
#define init_4_0_10_6
void FASTCALL call_4_0_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_10_7
#define init_4_0_10_7
void FASTCALL call_4_0_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_10_8
#define init_4_0_10_8
void FASTCALL call_4_0_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_10_9
#define init_4_0_10_9
void FASTCALL call_4_0_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_10_10
#define init_4_0_10_10
void FASTCALL call_4_0_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_10_11
#define init_4_0_10_11
void FASTCALL call_4_0_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_10_12
#define init_4_0_10_12
void FASTCALL call_4_0_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_10_13
#define init_4_0_10_13
void FASTCALL call_4_0_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_10_14
#define init_4_0_10_14
void FASTCALL call_4_0_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_10_15
#define init_4_0_10_15
void FASTCALL call_4_0_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_11_0
#define init_4_0_11_0
void FASTCALL call_4_0_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_11_1
#define init_4_0_11_1
void FASTCALL call_4_0_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_11_2
#define init_4_0_11_2
void FASTCALL call_4_0_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_11_3
#define init_4_0_11_3
void FASTCALL call_4_0_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_11_4
#define init_4_0_11_4
void FASTCALL call_4_0_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_11_5
#define init_4_0_11_5
void FASTCALL call_4_0_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_11_6
#define init_4_0_11_6
void FASTCALL call_4_0_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_11_7
#define init_4_0_11_7
void FASTCALL call_4_0_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_11_8
#define init_4_0_11_8
void FASTCALL call_4_0_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_11_9
#define init_4_0_11_9
void FASTCALL call_4_0_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_11_10
#define init_4_0_11_10
void FASTCALL call_4_0_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_11_11
#define init_4_0_11_11
void FASTCALL call_4_0_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_11_12
#define init_4_0_11_12
void FASTCALL call_4_0_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_11_13
#define init_4_0_11_13
void FASTCALL call_4_0_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_11_14
#define init_4_0_11_14
void FASTCALL call_4_0_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_11_15
#define init_4_0_11_15
void FASTCALL call_4_0_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_12_0
#define init_4_0_12_0
void FASTCALL call_4_0_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_12_1
#define init_4_0_12_1
void FASTCALL call_4_0_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_12_2
#define init_4_0_12_2
void FASTCALL call_4_0_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_12_3
#define init_4_0_12_3
void FASTCALL call_4_0_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_12_4
#define init_4_0_12_4
void FASTCALL call_4_0_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_12_5
#define init_4_0_12_5
void FASTCALL call_4_0_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_12_6
#define init_4_0_12_6
void FASTCALL call_4_0_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_12_7
#define init_4_0_12_7
void FASTCALL call_4_0_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_12_8
#define init_4_0_12_8
void FASTCALL call_4_0_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_12_9
#define init_4_0_12_9
void FASTCALL call_4_0_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_12_10
#define init_4_0_12_10
void FASTCALL call_4_0_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_12_11
#define init_4_0_12_11
void FASTCALL call_4_0_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_12_12
#define init_4_0_12_12
void FASTCALL call_4_0_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_12_13
#define init_4_0_12_13
void FASTCALL call_4_0_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_12_14
#define init_4_0_12_14
void FASTCALL call_4_0_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_12_15
#define init_4_0_12_15
void FASTCALL call_4_0_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_13_0
#define init_4_0_13_0
void FASTCALL call_4_0_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_13_1
#define init_4_0_13_1
void FASTCALL call_4_0_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_13_2
#define init_4_0_13_2
void FASTCALL call_4_0_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_13_3
#define init_4_0_13_3
void FASTCALL call_4_0_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_13_4
#define init_4_0_13_4
void FASTCALL call_4_0_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_13_5
#define init_4_0_13_5
void FASTCALL call_4_0_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_13_6
#define init_4_0_13_6
void FASTCALL call_4_0_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_13_7
#define init_4_0_13_7
void FASTCALL call_4_0_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_13_8
#define init_4_0_13_8
void FASTCALL call_4_0_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_13_9
#define init_4_0_13_9
void FASTCALL call_4_0_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_13_10
#define init_4_0_13_10
void FASTCALL call_4_0_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_13_11
#define init_4_0_13_11
void FASTCALL call_4_0_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_13_12
#define init_4_0_13_12
void FASTCALL call_4_0_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_13_13
#define init_4_0_13_13
void FASTCALL call_4_0_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_13_14
#define init_4_0_13_14
void FASTCALL call_4_0_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_13_15
#define init_4_0_13_15
void FASTCALL call_4_0_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_14_0
#define init_4_0_14_0
void FASTCALL call_4_0_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_14_1
#define init_4_0_14_1
void FASTCALL call_4_0_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_14_2
#define init_4_0_14_2
void FASTCALL call_4_0_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_14_3
#define init_4_0_14_3
void FASTCALL call_4_0_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_14_4
#define init_4_0_14_4
void FASTCALL call_4_0_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_14_5
#define init_4_0_14_5
void FASTCALL call_4_0_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_14_6
#define init_4_0_14_6
void FASTCALL call_4_0_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_14_7
#define init_4_0_14_7
void FASTCALL call_4_0_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_14_8
#define init_4_0_14_8
void FASTCALL call_4_0_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_14_9
#define init_4_0_14_9
void FASTCALL call_4_0_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_14_10
#define init_4_0_14_10
void FASTCALL call_4_0_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_14_11
#define init_4_0_14_11
void FASTCALL call_4_0_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_14_12
#define init_4_0_14_12
void FASTCALL call_4_0_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_14_13
#define init_4_0_14_13
void FASTCALL call_4_0_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_14_14
#define init_4_0_14_14
void FASTCALL call_4_0_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_14_15
#define init_4_0_14_15
void FASTCALL call_4_0_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_15_0
#define init_4_0_15_0
void FASTCALL call_4_0_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_15_1
#define init_4_0_15_1
void FASTCALL call_4_0_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_15_2
#define init_4_0_15_2
void FASTCALL call_4_0_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_15_3
#define init_4_0_15_3
void FASTCALL call_4_0_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_15_4
#define init_4_0_15_4
void FASTCALL call_4_0_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_15_5
#define init_4_0_15_5
void FASTCALL call_4_0_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_15_6
#define init_4_0_15_6
void FASTCALL call_4_0_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_15_7
#define init_4_0_15_7
void FASTCALL call_4_0_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_15_8
#define init_4_0_15_8
void FASTCALL call_4_0_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_15_9
#define init_4_0_15_9
void FASTCALL call_4_0_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_15_10
#define init_4_0_15_10
void FASTCALL call_4_0_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_15_11
#define init_4_0_15_11
void FASTCALL call_4_0_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_15_12
#define init_4_0_15_12
void FASTCALL call_4_0_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_15_13
#define init_4_0_15_13
void FASTCALL call_4_0_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_15_14
#define init_4_0_15_14
void FASTCALL call_4_0_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_0_15_15
#define init_4_0_15_15
void FASTCALL call_4_0_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_0_0
#define init_4_1_0_0
void FASTCALL call_4_1_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_0_1
#define init_4_1_0_1
void FASTCALL call_4_1_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_0_2
#define init_4_1_0_2
void FASTCALL call_4_1_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_0_3
#define init_4_1_0_3
void FASTCALL call_4_1_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_0_4
#define init_4_1_0_4
void FASTCALL call_4_1_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_0_5
#define init_4_1_0_5
void FASTCALL call_4_1_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_0_6
#define init_4_1_0_6
void FASTCALL call_4_1_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_0_7
#define init_4_1_0_7
void FASTCALL call_4_1_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_0_8
#define init_4_1_0_8
void FASTCALL call_4_1_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_0_9
#define init_4_1_0_9
void FASTCALL call_4_1_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_0_10
#define init_4_1_0_10
void FASTCALL call_4_1_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_0_11
#define init_4_1_0_11
void FASTCALL call_4_1_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_0_12
#define init_4_1_0_12
void FASTCALL call_4_1_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_0_13
#define init_4_1_0_13
void FASTCALL call_4_1_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_0_14
#define init_4_1_0_14
void FASTCALL call_4_1_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_0_15
#define init_4_1_0_15
void FASTCALL call_4_1_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_1_0
#define init_4_1_1_0
void FASTCALL call_4_1_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_1_1
#define init_4_1_1_1
void FASTCALL call_4_1_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_1_2
#define init_4_1_1_2
void FASTCALL call_4_1_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_1_3
#define init_4_1_1_3
void FASTCALL call_4_1_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_1_4
#define init_4_1_1_4
void FASTCALL call_4_1_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_1_5
#define init_4_1_1_5
void FASTCALL call_4_1_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_1_6
#define init_4_1_1_6
void FASTCALL call_4_1_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_1_7
#define init_4_1_1_7
void FASTCALL call_4_1_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_1_8
#define init_4_1_1_8
void FASTCALL call_4_1_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_1_9
#define init_4_1_1_9
void FASTCALL call_4_1_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_1_10
#define init_4_1_1_10
void FASTCALL call_4_1_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_1_11
#define init_4_1_1_11
void FASTCALL call_4_1_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_1_12
#define init_4_1_1_12
void FASTCALL call_4_1_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_1_13
#define init_4_1_1_13
void FASTCALL call_4_1_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_1_14
#define init_4_1_1_14
void FASTCALL call_4_1_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_1_15
#define init_4_1_1_15
void FASTCALL call_4_1_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_2_0
#define init_4_1_2_0
void FASTCALL call_4_1_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_2_1
#define init_4_1_2_1
void FASTCALL call_4_1_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_2_2
#define init_4_1_2_2
void FASTCALL call_4_1_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_2_3
#define init_4_1_2_3
void FASTCALL call_4_1_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_2_4
#define init_4_1_2_4
void FASTCALL call_4_1_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_2_5
#define init_4_1_2_5
void FASTCALL call_4_1_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_2_6
#define init_4_1_2_6
void FASTCALL call_4_1_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_2_7
#define init_4_1_2_7
void FASTCALL call_4_1_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_2_8
#define init_4_1_2_8
void FASTCALL call_4_1_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_2_9
#define init_4_1_2_9
void FASTCALL call_4_1_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_2_10
#define init_4_1_2_10
void FASTCALL call_4_1_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_2_11
#define init_4_1_2_11
void FASTCALL call_4_1_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_2_12
#define init_4_1_2_12
void FASTCALL call_4_1_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_2_13
#define init_4_1_2_13
void FASTCALL call_4_1_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_2_14
#define init_4_1_2_14
void FASTCALL call_4_1_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_2_15
#define init_4_1_2_15
void FASTCALL call_4_1_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_3_0
#define init_4_1_3_0
void FASTCALL call_4_1_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_3_1
#define init_4_1_3_1
void FASTCALL call_4_1_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_3_2
#define init_4_1_3_2
void FASTCALL call_4_1_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_3_3
#define init_4_1_3_3
void FASTCALL call_4_1_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_3_4
#define init_4_1_3_4
void FASTCALL call_4_1_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_3_5
#define init_4_1_3_5
void FASTCALL call_4_1_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_3_6
#define init_4_1_3_6
void FASTCALL call_4_1_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_3_7
#define init_4_1_3_7
void FASTCALL call_4_1_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_3_8
#define init_4_1_3_8
void FASTCALL call_4_1_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_3_9
#define init_4_1_3_9
void FASTCALL call_4_1_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_3_10
#define init_4_1_3_10
void FASTCALL call_4_1_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_3_11
#define init_4_1_3_11
void FASTCALL call_4_1_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_3_12
#define init_4_1_3_12
void FASTCALL call_4_1_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_3_13
#define init_4_1_3_13
void FASTCALL call_4_1_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_3_14
#define init_4_1_3_14
void FASTCALL call_4_1_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_3_15
#define init_4_1_3_15
void FASTCALL call_4_1_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_4_0
#define init_4_1_4_0
void FASTCALL call_4_1_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_4_1
#define init_4_1_4_1
void FASTCALL call_4_1_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_4_2
#define init_4_1_4_2
void FASTCALL call_4_1_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_4_3
#define init_4_1_4_3
void FASTCALL call_4_1_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_4_4
#define init_4_1_4_4
void FASTCALL call_4_1_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_4_5
#define init_4_1_4_5
void FASTCALL call_4_1_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_4_6
#define init_4_1_4_6
void FASTCALL call_4_1_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_4_7
#define init_4_1_4_7
void FASTCALL call_4_1_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_4_8
#define init_4_1_4_8
void FASTCALL call_4_1_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_4_9
#define init_4_1_4_9
void FASTCALL call_4_1_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_4_10
#define init_4_1_4_10
void FASTCALL call_4_1_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_4_11
#define init_4_1_4_11
void FASTCALL call_4_1_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_4_12
#define init_4_1_4_12
void FASTCALL call_4_1_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_4_13
#define init_4_1_4_13
void FASTCALL call_4_1_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_4_14
#define init_4_1_4_14
void FASTCALL call_4_1_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_4_15
#define init_4_1_4_15
void FASTCALL call_4_1_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_5_0
#define init_4_1_5_0
void FASTCALL call_4_1_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_5_1
#define init_4_1_5_1
void FASTCALL call_4_1_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_5_2
#define init_4_1_5_2
void FASTCALL call_4_1_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_5_3
#define init_4_1_5_3
void FASTCALL call_4_1_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_5_4
#define init_4_1_5_4
void FASTCALL call_4_1_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_5_5
#define init_4_1_5_5
void FASTCALL call_4_1_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_5_6
#define init_4_1_5_6
void FASTCALL call_4_1_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_5_7
#define init_4_1_5_7
void FASTCALL call_4_1_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_5_8
#define init_4_1_5_8
void FASTCALL call_4_1_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_5_9
#define init_4_1_5_9
void FASTCALL call_4_1_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_5_10
#define init_4_1_5_10
void FASTCALL call_4_1_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_5_11
#define init_4_1_5_11
void FASTCALL call_4_1_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_5_12
#define init_4_1_5_12
void FASTCALL call_4_1_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_5_13
#define init_4_1_5_13
void FASTCALL call_4_1_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_5_14
#define init_4_1_5_14
void FASTCALL call_4_1_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_5_15
#define init_4_1_5_15
void FASTCALL call_4_1_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_6_0
#define init_4_1_6_0
void FASTCALL call_4_1_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_6_1
#define init_4_1_6_1
void FASTCALL call_4_1_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_6_2
#define init_4_1_6_2
void FASTCALL call_4_1_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_6_3
#define init_4_1_6_3
void FASTCALL call_4_1_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_6_4
#define init_4_1_6_4
void FASTCALL call_4_1_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_6_5
#define init_4_1_6_5
void FASTCALL call_4_1_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_6_6
#define init_4_1_6_6
void FASTCALL call_4_1_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_6_7
#define init_4_1_6_7
void FASTCALL call_4_1_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_6_8
#define init_4_1_6_8
void FASTCALL call_4_1_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_6_9
#define init_4_1_6_9
void FASTCALL call_4_1_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_6_10
#define init_4_1_6_10
void FASTCALL call_4_1_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_6_11
#define init_4_1_6_11
void FASTCALL call_4_1_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_6_12
#define init_4_1_6_12
void FASTCALL call_4_1_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_6_13
#define init_4_1_6_13
void FASTCALL call_4_1_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_6_14
#define init_4_1_6_14
void FASTCALL call_4_1_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_6_15
#define init_4_1_6_15
void FASTCALL call_4_1_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_7_0
#define init_4_1_7_0
void FASTCALL call_4_1_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_7_1
#define init_4_1_7_1
void FASTCALL call_4_1_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_7_2
#define init_4_1_7_2
void FASTCALL call_4_1_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_7_3
#define init_4_1_7_3
void FASTCALL call_4_1_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_7_4
#define init_4_1_7_4
void FASTCALL call_4_1_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_7_5
#define init_4_1_7_5
void FASTCALL call_4_1_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_7_6
#define init_4_1_7_6
void FASTCALL call_4_1_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_7_7
#define init_4_1_7_7
void FASTCALL call_4_1_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_7_8
#define init_4_1_7_8
void FASTCALL call_4_1_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_7_9
#define init_4_1_7_9
void FASTCALL call_4_1_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_7_10
#define init_4_1_7_10
void FASTCALL call_4_1_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_7_11
#define init_4_1_7_11
void FASTCALL call_4_1_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_7_12
#define init_4_1_7_12
void FASTCALL call_4_1_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_7_13
#define init_4_1_7_13
void FASTCALL call_4_1_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_7_14
#define init_4_1_7_14
void FASTCALL call_4_1_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_7_15
#define init_4_1_7_15
void FASTCALL call_4_1_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_8_0
#define init_4_1_8_0
void FASTCALL call_4_1_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_8_1
#define init_4_1_8_1
void FASTCALL call_4_1_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_8_2
#define init_4_1_8_2
void FASTCALL call_4_1_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_8_3
#define init_4_1_8_3
void FASTCALL call_4_1_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_8_4
#define init_4_1_8_4
void FASTCALL call_4_1_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_8_5
#define init_4_1_8_5
void FASTCALL call_4_1_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_8_6
#define init_4_1_8_6
void FASTCALL call_4_1_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_8_7
#define init_4_1_8_7
void FASTCALL call_4_1_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_8_8
#define init_4_1_8_8
void FASTCALL call_4_1_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_8_9
#define init_4_1_8_9
void FASTCALL call_4_1_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_8_10
#define init_4_1_8_10
void FASTCALL call_4_1_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_8_11
#define init_4_1_8_11
void FASTCALL call_4_1_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_8_12
#define init_4_1_8_12
void FASTCALL call_4_1_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_8_13
#define init_4_1_8_13
void FASTCALL call_4_1_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_8_14
#define init_4_1_8_14
void FASTCALL call_4_1_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_8_15
#define init_4_1_8_15
void FASTCALL call_4_1_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_9_0
#define init_4_1_9_0
void FASTCALL call_4_1_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_9_1
#define init_4_1_9_1
void FASTCALL call_4_1_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_9_2
#define init_4_1_9_2
void FASTCALL call_4_1_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_9_3
#define init_4_1_9_3
void FASTCALL call_4_1_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_9_4
#define init_4_1_9_4
void FASTCALL call_4_1_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_9_5
#define init_4_1_9_5
void FASTCALL call_4_1_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_9_6
#define init_4_1_9_6
void FASTCALL call_4_1_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_9_7
#define init_4_1_9_7
void FASTCALL call_4_1_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_9_8
#define init_4_1_9_8
void FASTCALL call_4_1_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_9_9
#define init_4_1_9_9
void FASTCALL call_4_1_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_9_10
#define init_4_1_9_10
void FASTCALL call_4_1_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_9_11
#define init_4_1_9_11
void FASTCALL call_4_1_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_9_12
#define init_4_1_9_12
void FASTCALL call_4_1_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_9_13
#define init_4_1_9_13
void FASTCALL call_4_1_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_9_14
#define init_4_1_9_14
void FASTCALL call_4_1_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_9_15
#define init_4_1_9_15
void FASTCALL call_4_1_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_10_0
#define init_4_1_10_0
void FASTCALL call_4_1_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_10_1
#define init_4_1_10_1
void FASTCALL call_4_1_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_10_2
#define init_4_1_10_2
void FASTCALL call_4_1_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_10_3
#define init_4_1_10_3
void FASTCALL call_4_1_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_10_4
#define init_4_1_10_4
void FASTCALL call_4_1_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_10_5
#define init_4_1_10_5
void FASTCALL call_4_1_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_10_6
#define init_4_1_10_6
void FASTCALL call_4_1_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_10_7
#define init_4_1_10_7
void FASTCALL call_4_1_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_10_8
#define init_4_1_10_8
void FASTCALL call_4_1_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_10_9
#define init_4_1_10_9
void FASTCALL call_4_1_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_10_10
#define init_4_1_10_10
void FASTCALL call_4_1_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_10_11
#define init_4_1_10_11
void FASTCALL call_4_1_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_10_12
#define init_4_1_10_12
void FASTCALL call_4_1_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_10_13
#define init_4_1_10_13
void FASTCALL call_4_1_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_10_14
#define init_4_1_10_14
void FASTCALL call_4_1_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_10_15
#define init_4_1_10_15
void FASTCALL call_4_1_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_11_0
#define init_4_1_11_0
void FASTCALL call_4_1_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_11_1
#define init_4_1_11_1
void FASTCALL call_4_1_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_11_2
#define init_4_1_11_2
void FASTCALL call_4_1_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_11_3
#define init_4_1_11_3
void FASTCALL call_4_1_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_11_4
#define init_4_1_11_4
void FASTCALL call_4_1_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_11_5
#define init_4_1_11_5
void FASTCALL call_4_1_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_11_6
#define init_4_1_11_6
void FASTCALL call_4_1_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_11_7
#define init_4_1_11_7
void FASTCALL call_4_1_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_11_8
#define init_4_1_11_8
void FASTCALL call_4_1_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_11_9
#define init_4_1_11_9
void FASTCALL call_4_1_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_11_10
#define init_4_1_11_10
void FASTCALL call_4_1_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_11_11
#define init_4_1_11_11
void FASTCALL call_4_1_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_11_12
#define init_4_1_11_12
void FASTCALL call_4_1_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_11_13
#define init_4_1_11_13
void FASTCALL call_4_1_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_11_14
#define init_4_1_11_14
void FASTCALL call_4_1_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_11_15
#define init_4_1_11_15
void FASTCALL call_4_1_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_12_0
#define init_4_1_12_0
void FASTCALL call_4_1_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_12_1
#define init_4_1_12_1
void FASTCALL call_4_1_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_12_2
#define init_4_1_12_2
void FASTCALL call_4_1_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_12_3
#define init_4_1_12_3
void FASTCALL call_4_1_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_12_4
#define init_4_1_12_4
void FASTCALL call_4_1_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_12_5
#define init_4_1_12_5
void FASTCALL call_4_1_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_12_6
#define init_4_1_12_6
void FASTCALL call_4_1_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_12_7
#define init_4_1_12_7
void FASTCALL call_4_1_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_12_8
#define init_4_1_12_8
void FASTCALL call_4_1_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_12_9
#define init_4_1_12_9
void FASTCALL call_4_1_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_12_10
#define init_4_1_12_10
void FASTCALL call_4_1_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_12_11
#define init_4_1_12_11
void FASTCALL call_4_1_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_12_12
#define init_4_1_12_12
void FASTCALL call_4_1_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_12_13
#define init_4_1_12_13
void FASTCALL call_4_1_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_12_14
#define init_4_1_12_14
void FASTCALL call_4_1_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_12_15
#define init_4_1_12_15
void FASTCALL call_4_1_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_13_0
#define init_4_1_13_0
void FASTCALL call_4_1_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_13_1
#define init_4_1_13_1
void FASTCALL call_4_1_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_13_2
#define init_4_1_13_2
void FASTCALL call_4_1_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_13_3
#define init_4_1_13_3
void FASTCALL call_4_1_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_13_4
#define init_4_1_13_4
void FASTCALL call_4_1_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_13_5
#define init_4_1_13_5
void FASTCALL call_4_1_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_13_6
#define init_4_1_13_6
void FASTCALL call_4_1_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_13_7
#define init_4_1_13_7
void FASTCALL call_4_1_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_13_8
#define init_4_1_13_8
void FASTCALL call_4_1_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_13_9
#define init_4_1_13_9
void FASTCALL call_4_1_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_13_10
#define init_4_1_13_10
void FASTCALL call_4_1_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_13_11
#define init_4_1_13_11
void FASTCALL call_4_1_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_13_12
#define init_4_1_13_12
void FASTCALL call_4_1_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_13_13
#define init_4_1_13_13
void FASTCALL call_4_1_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_13_14
#define init_4_1_13_14
void FASTCALL call_4_1_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_13_15
#define init_4_1_13_15
void FASTCALL call_4_1_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_14_0
#define init_4_1_14_0
void FASTCALL call_4_1_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_14_1
#define init_4_1_14_1
void FASTCALL call_4_1_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_14_2
#define init_4_1_14_2
void FASTCALL call_4_1_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_14_3
#define init_4_1_14_3
void FASTCALL call_4_1_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_14_4
#define init_4_1_14_4
void FASTCALL call_4_1_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_14_5
#define init_4_1_14_5
void FASTCALL call_4_1_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_14_6
#define init_4_1_14_6
void FASTCALL call_4_1_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_14_7
#define init_4_1_14_7
void FASTCALL call_4_1_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_14_8
#define init_4_1_14_8
void FASTCALL call_4_1_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_14_9
#define init_4_1_14_9
void FASTCALL call_4_1_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_14_10
#define init_4_1_14_10
void FASTCALL call_4_1_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_14_11
#define init_4_1_14_11
void FASTCALL call_4_1_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_14_12
#define init_4_1_14_12
void FASTCALL call_4_1_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_14_13
#define init_4_1_14_13
void FASTCALL call_4_1_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_14_14
#define init_4_1_14_14
void FASTCALL call_4_1_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_14_15
#define init_4_1_14_15
void FASTCALL call_4_1_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_15_0
#define init_4_1_15_0
void FASTCALL call_4_1_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_15_1
#define init_4_1_15_1
void FASTCALL call_4_1_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_15_2
#define init_4_1_15_2
void FASTCALL call_4_1_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_15_3
#define init_4_1_15_3
void FASTCALL call_4_1_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_15_4
#define init_4_1_15_4
void FASTCALL call_4_1_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_15_5
#define init_4_1_15_5
void FASTCALL call_4_1_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_15_6
#define init_4_1_15_6
void FASTCALL call_4_1_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_15_7
#define init_4_1_15_7
void FASTCALL call_4_1_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_15_8
#define init_4_1_15_8
void FASTCALL call_4_1_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_15_9
#define init_4_1_15_9
void FASTCALL call_4_1_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_15_10
#define init_4_1_15_10
void FASTCALL call_4_1_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_15_11
#define init_4_1_15_11
void FASTCALL call_4_1_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_15_12
#define init_4_1_15_12
void FASTCALL call_4_1_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_15_13
#define init_4_1_15_13
void FASTCALL call_4_1_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_15_14
#define init_4_1_15_14
void FASTCALL call_4_1_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_1_15_15
#define init_4_1_15_15
void FASTCALL call_4_1_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_0_0
#define init_4_2_0_0
void FASTCALL call_4_2_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_0_1
#define init_4_2_0_1
void FASTCALL call_4_2_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_0_2
#define init_4_2_0_2
void FASTCALL call_4_2_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_0_3
#define init_4_2_0_3
void FASTCALL call_4_2_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_0_4
#define init_4_2_0_4
void FASTCALL call_4_2_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_0_5
#define init_4_2_0_5
void FASTCALL call_4_2_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_0_6
#define init_4_2_0_6
void FASTCALL call_4_2_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_0_7
#define init_4_2_0_7
void FASTCALL call_4_2_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_0_8
#define init_4_2_0_8
void FASTCALL call_4_2_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_0_9
#define init_4_2_0_9
void FASTCALL call_4_2_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_0_10
#define init_4_2_0_10
void FASTCALL call_4_2_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_0_11
#define init_4_2_0_11
void FASTCALL call_4_2_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_0_12
#define init_4_2_0_12
void FASTCALL call_4_2_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_0_13
#define init_4_2_0_13
void FASTCALL call_4_2_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_0_14
#define init_4_2_0_14
void FASTCALL call_4_2_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_0_15
#define init_4_2_0_15
void FASTCALL call_4_2_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_1_0
#define init_4_2_1_0
void FASTCALL call_4_2_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_1_1
#define init_4_2_1_1
void FASTCALL call_4_2_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_1_2
#define init_4_2_1_2
void FASTCALL call_4_2_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_1_3
#define init_4_2_1_3
void FASTCALL call_4_2_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_1_4
#define init_4_2_1_4
void FASTCALL call_4_2_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_1_5
#define init_4_2_1_5
void FASTCALL call_4_2_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_1_6
#define init_4_2_1_6
void FASTCALL call_4_2_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_1_7
#define init_4_2_1_7
void FASTCALL call_4_2_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_1_8
#define init_4_2_1_8
void FASTCALL call_4_2_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_1_9
#define init_4_2_1_9
void FASTCALL call_4_2_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_1_10
#define init_4_2_1_10
void FASTCALL call_4_2_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_1_11
#define init_4_2_1_11
void FASTCALL call_4_2_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_1_12
#define init_4_2_1_12
void FASTCALL call_4_2_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_1_13
#define init_4_2_1_13
void FASTCALL call_4_2_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_1_14
#define init_4_2_1_14
void FASTCALL call_4_2_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_1_15
#define init_4_2_1_15
void FASTCALL call_4_2_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_2_0
#define init_4_2_2_0
void FASTCALL call_4_2_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_2_1
#define init_4_2_2_1
void FASTCALL call_4_2_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_2_2
#define init_4_2_2_2
void FASTCALL call_4_2_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_2_3
#define init_4_2_2_3
void FASTCALL call_4_2_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_2_4
#define init_4_2_2_4
void FASTCALL call_4_2_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_2_5
#define init_4_2_2_5
void FASTCALL call_4_2_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_2_6
#define init_4_2_2_6
void FASTCALL call_4_2_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_2_7
#define init_4_2_2_7
void FASTCALL call_4_2_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_2_8
#define init_4_2_2_8
void FASTCALL call_4_2_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_2_9
#define init_4_2_2_9
void FASTCALL call_4_2_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_2_10
#define init_4_2_2_10
void FASTCALL call_4_2_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_2_11
#define init_4_2_2_11
void FASTCALL call_4_2_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_2_12
#define init_4_2_2_12
void FASTCALL call_4_2_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_2_13
#define init_4_2_2_13
void FASTCALL call_4_2_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_2_14
#define init_4_2_2_14
void FASTCALL call_4_2_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_2_15
#define init_4_2_2_15
void FASTCALL call_4_2_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_3_0
#define init_4_2_3_0
void FASTCALL call_4_2_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_3_1
#define init_4_2_3_1
void FASTCALL call_4_2_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_3_2
#define init_4_2_3_2
void FASTCALL call_4_2_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_3_3
#define init_4_2_3_3
void FASTCALL call_4_2_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_3_4
#define init_4_2_3_4
void FASTCALL call_4_2_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_3_5
#define init_4_2_3_5
void FASTCALL call_4_2_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_3_6
#define init_4_2_3_6
void FASTCALL call_4_2_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_3_7
#define init_4_2_3_7
void FASTCALL call_4_2_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_3_8
#define init_4_2_3_8
void FASTCALL call_4_2_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_3_9
#define init_4_2_3_9
void FASTCALL call_4_2_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_3_10
#define init_4_2_3_10
void FASTCALL call_4_2_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_3_11
#define init_4_2_3_11
void FASTCALL call_4_2_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_3_12
#define init_4_2_3_12
void FASTCALL call_4_2_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_3_13
#define init_4_2_3_13
void FASTCALL call_4_2_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_3_14
#define init_4_2_3_14
void FASTCALL call_4_2_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_3_15
#define init_4_2_3_15
void FASTCALL call_4_2_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_4_0
#define init_4_2_4_0
void FASTCALL call_4_2_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_4_1
#define init_4_2_4_1
void FASTCALL call_4_2_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_4_2
#define init_4_2_4_2
void FASTCALL call_4_2_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_4_3
#define init_4_2_4_3
void FASTCALL call_4_2_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_4_4
#define init_4_2_4_4
void FASTCALL call_4_2_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_4_5
#define init_4_2_4_5
void FASTCALL call_4_2_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_4_6
#define init_4_2_4_6
void FASTCALL call_4_2_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_4_7
#define init_4_2_4_7
void FASTCALL call_4_2_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_4_8
#define init_4_2_4_8
void FASTCALL call_4_2_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_4_9
#define init_4_2_4_9
void FASTCALL call_4_2_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_4_10
#define init_4_2_4_10
void FASTCALL call_4_2_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_4_11
#define init_4_2_4_11
void FASTCALL call_4_2_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_4_12
#define init_4_2_4_12
void FASTCALL call_4_2_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_4_13
#define init_4_2_4_13
void FASTCALL call_4_2_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_4_14
#define init_4_2_4_14
void FASTCALL call_4_2_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_4_15
#define init_4_2_4_15
void FASTCALL call_4_2_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_5_0
#define init_4_2_5_0
void FASTCALL call_4_2_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_5_1
#define init_4_2_5_1
void FASTCALL call_4_2_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_5_2
#define init_4_2_5_2
void FASTCALL call_4_2_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_5_3
#define init_4_2_5_3
void FASTCALL call_4_2_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_5_4
#define init_4_2_5_4
void FASTCALL call_4_2_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_5_5
#define init_4_2_5_5
void FASTCALL call_4_2_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_5_6
#define init_4_2_5_6
void FASTCALL call_4_2_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_5_7
#define init_4_2_5_7
void FASTCALL call_4_2_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_5_8
#define init_4_2_5_8
void FASTCALL call_4_2_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_5_9
#define init_4_2_5_9
void FASTCALL call_4_2_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_5_10
#define init_4_2_5_10
void FASTCALL call_4_2_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_5_11
#define init_4_2_5_11
void FASTCALL call_4_2_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_5_12
#define init_4_2_5_12
void FASTCALL call_4_2_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_5_13
#define init_4_2_5_13
void FASTCALL call_4_2_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_5_14
#define init_4_2_5_14
void FASTCALL call_4_2_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_5_15
#define init_4_2_5_15
void FASTCALL call_4_2_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_6_0
#define init_4_2_6_0
void FASTCALL call_4_2_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_6_1
#define init_4_2_6_1
void FASTCALL call_4_2_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_6_2
#define init_4_2_6_2
void FASTCALL call_4_2_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_6_3
#define init_4_2_6_3
void FASTCALL call_4_2_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_6_4
#define init_4_2_6_4
void FASTCALL call_4_2_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_6_5
#define init_4_2_6_5
void FASTCALL call_4_2_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_6_6
#define init_4_2_6_6
void FASTCALL call_4_2_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_6_7
#define init_4_2_6_7
void FASTCALL call_4_2_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_6_8
#define init_4_2_6_8
void FASTCALL call_4_2_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_6_9
#define init_4_2_6_9
void FASTCALL call_4_2_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_6_10
#define init_4_2_6_10
void FASTCALL call_4_2_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_6_11
#define init_4_2_6_11
void FASTCALL call_4_2_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_6_12
#define init_4_2_6_12
void FASTCALL call_4_2_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_6_13
#define init_4_2_6_13
void FASTCALL call_4_2_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_6_14
#define init_4_2_6_14
void FASTCALL call_4_2_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_6_15
#define init_4_2_6_15
void FASTCALL call_4_2_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_7_0
#define init_4_2_7_0
void FASTCALL call_4_2_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_7_1
#define init_4_2_7_1
void FASTCALL call_4_2_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_7_2
#define init_4_2_7_2
void FASTCALL call_4_2_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_7_3
#define init_4_2_7_3
void FASTCALL call_4_2_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_7_4
#define init_4_2_7_4
void FASTCALL call_4_2_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_7_5
#define init_4_2_7_5
void FASTCALL call_4_2_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_7_6
#define init_4_2_7_6
void FASTCALL call_4_2_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_7_7
#define init_4_2_7_7
void FASTCALL call_4_2_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_7_8
#define init_4_2_7_8
void FASTCALL call_4_2_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_7_9
#define init_4_2_7_9
void FASTCALL call_4_2_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_7_10
#define init_4_2_7_10
void FASTCALL call_4_2_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_7_11
#define init_4_2_7_11
void FASTCALL call_4_2_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_7_12
#define init_4_2_7_12
void FASTCALL call_4_2_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_7_13
#define init_4_2_7_13
void FASTCALL call_4_2_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_7_14
#define init_4_2_7_14
void FASTCALL call_4_2_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_7_15
#define init_4_2_7_15
void FASTCALL call_4_2_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_8_0
#define init_4_2_8_0
void FASTCALL call_4_2_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_8_1
#define init_4_2_8_1
void FASTCALL call_4_2_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_8_2
#define init_4_2_8_2
void FASTCALL call_4_2_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_8_3
#define init_4_2_8_3
void FASTCALL call_4_2_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_8_4
#define init_4_2_8_4
void FASTCALL call_4_2_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_8_5
#define init_4_2_8_5
void FASTCALL call_4_2_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_8_6
#define init_4_2_8_6
void FASTCALL call_4_2_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_8_7
#define init_4_2_8_7
void FASTCALL call_4_2_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_8_8
#define init_4_2_8_8
void FASTCALL call_4_2_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_8_9
#define init_4_2_8_9
void FASTCALL call_4_2_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_8_10
#define init_4_2_8_10
void FASTCALL call_4_2_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_8_11
#define init_4_2_8_11
void FASTCALL call_4_2_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_8_12
#define init_4_2_8_12
void FASTCALL call_4_2_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_8_13
#define init_4_2_8_13
void FASTCALL call_4_2_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_8_14
#define init_4_2_8_14
void FASTCALL call_4_2_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_8_15
#define init_4_2_8_15
void FASTCALL call_4_2_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_9_0
#define init_4_2_9_0
void FASTCALL call_4_2_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_9_1
#define init_4_2_9_1
void FASTCALL call_4_2_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_9_2
#define init_4_2_9_2
void FASTCALL call_4_2_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_9_3
#define init_4_2_9_3
void FASTCALL call_4_2_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_9_4
#define init_4_2_9_4
void FASTCALL call_4_2_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_9_5
#define init_4_2_9_5
void FASTCALL call_4_2_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_9_6
#define init_4_2_9_6
void FASTCALL call_4_2_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_9_7
#define init_4_2_9_7
void FASTCALL call_4_2_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_9_8
#define init_4_2_9_8
void FASTCALL call_4_2_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_9_9
#define init_4_2_9_9
void FASTCALL call_4_2_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_9_10
#define init_4_2_9_10
void FASTCALL call_4_2_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_9_11
#define init_4_2_9_11
void FASTCALL call_4_2_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_9_12
#define init_4_2_9_12
void FASTCALL call_4_2_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_9_13
#define init_4_2_9_13
void FASTCALL call_4_2_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_9_14
#define init_4_2_9_14
void FASTCALL call_4_2_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_9_15
#define init_4_2_9_15
void FASTCALL call_4_2_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_10_0
#define init_4_2_10_0
void FASTCALL call_4_2_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_10_1
#define init_4_2_10_1
void FASTCALL call_4_2_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_10_2
#define init_4_2_10_2
void FASTCALL call_4_2_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_10_3
#define init_4_2_10_3
void FASTCALL call_4_2_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_10_4
#define init_4_2_10_4
void FASTCALL call_4_2_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_10_5
#define init_4_2_10_5
void FASTCALL call_4_2_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_10_6
#define init_4_2_10_6
void FASTCALL call_4_2_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_10_7
#define init_4_2_10_7
void FASTCALL call_4_2_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_10_8
#define init_4_2_10_8
void FASTCALL call_4_2_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_10_9
#define init_4_2_10_9
void FASTCALL call_4_2_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_10_10
#define init_4_2_10_10
void FASTCALL call_4_2_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_10_11
#define init_4_2_10_11
void FASTCALL call_4_2_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_10_12
#define init_4_2_10_12
void FASTCALL call_4_2_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_10_13
#define init_4_2_10_13
void FASTCALL call_4_2_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_10_14
#define init_4_2_10_14
void FASTCALL call_4_2_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_10_15
#define init_4_2_10_15
void FASTCALL call_4_2_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_11_0
#define init_4_2_11_0
void FASTCALL call_4_2_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_11_1
#define init_4_2_11_1
void FASTCALL call_4_2_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_11_2
#define init_4_2_11_2
void FASTCALL call_4_2_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_11_3
#define init_4_2_11_3
void FASTCALL call_4_2_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_11_4
#define init_4_2_11_4
void FASTCALL call_4_2_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_11_5
#define init_4_2_11_5
void FASTCALL call_4_2_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_11_6
#define init_4_2_11_6
void FASTCALL call_4_2_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_11_7
#define init_4_2_11_7
void FASTCALL call_4_2_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_11_8
#define init_4_2_11_8
void FASTCALL call_4_2_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_11_9
#define init_4_2_11_9
void FASTCALL call_4_2_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_11_10
#define init_4_2_11_10
void FASTCALL call_4_2_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_11_11
#define init_4_2_11_11
void FASTCALL call_4_2_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_11_12
#define init_4_2_11_12
void FASTCALL call_4_2_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_11_13
#define init_4_2_11_13
void FASTCALL call_4_2_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_11_14
#define init_4_2_11_14
void FASTCALL call_4_2_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_11_15
#define init_4_2_11_15
void FASTCALL call_4_2_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_12_0
#define init_4_2_12_0
void FASTCALL call_4_2_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_12_1
#define init_4_2_12_1
void FASTCALL call_4_2_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_12_2
#define init_4_2_12_2
void FASTCALL call_4_2_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_12_3
#define init_4_2_12_3
void FASTCALL call_4_2_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_12_4
#define init_4_2_12_4
void FASTCALL call_4_2_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_12_5
#define init_4_2_12_5
void FASTCALL call_4_2_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_12_6
#define init_4_2_12_6
void FASTCALL call_4_2_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_12_7
#define init_4_2_12_7
void FASTCALL call_4_2_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_12_8
#define init_4_2_12_8
void FASTCALL call_4_2_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_12_9
#define init_4_2_12_9
void FASTCALL call_4_2_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_12_10
#define init_4_2_12_10
void FASTCALL call_4_2_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_12_11
#define init_4_2_12_11
void FASTCALL call_4_2_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_12_12
#define init_4_2_12_12
void FASTCALL call_4_2_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_12_13
#define init_4_2_12_13
void FASTCALL call_4_2_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_12_14
#define init_4_2_12_14
void FASTCALL call_4_2_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_12_15
#define init_4_2_12_15
void FASTCALL call_4_2_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_13_0
#define init_4_2_13_0
void FASTCALL call_4_2_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_13_1
#define init_4_2_13_1
void FASTCALL call_4_2_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_13_2
#define init_4_2_13_2
void FASTCALL call_4_2_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_13_3
#define init_4_2_13_3
void FASTCALL call_4_2_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_13_4
#define init_4_2_13_4
void FASTCALL call_4_2_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_13_5
#define init_4_2_13_5
void FASTCALL call_4_2_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_13_6
#define init_4_2_13_6
void FASTCALL call_4_2_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_13_7
#define init_4_2_13_7
void FASTCALL call_4_2_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_13_8
#define init_4_2_13_8
void FASTCALL call_4_2_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_13_9
#define init_4_2_13_9
void FASTCALL call_4_2_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_13_10
#define init_4_2_13_10
void FASTCALL call_4_2_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_13_11
#define init_4_2_13_11
void FASTCALL call_4_2_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_13_12
#define init_4_2_13_12
void FASTCALL call_4_2_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_13_13
#define init_4_2_13_13
void FASTCALL call_4_2_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_13_14
#define init_4_2_13_14
void FASTCALL call_4_2_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_13_15
#define init_4_2_13_15
void FASTCALL call_4_2_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_14_0
#define init_4_2_14_0
void FASTCALL call_4_2_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_14_1
#define init_4_2_14_1
void FASTCALL call_4_2_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_14_2
#define init_4_2_14_2
void FASTCALL call_4_2_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_14_3
#define init_4_2_14_3
void FASTCALL call_4_2_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_14_4
#define init_4_2_14_4
void FASTCALL call_4_2_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_14_5
#define init_4_2_14_5
void FASTCALL call_4_2_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_14_6
#define init_4_2_14_6
void FASTCALL call_4_2_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_14_7
#define init_4_2_14_7
void FASTCALL call_4_2_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_14_8
#define init_4_2_14_8
void FASTCALL call_4_2_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_14_9
#define init_4_2_14_9
void FASTCALL call_4_2_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_14_10
#define init_4_2_14_10
void FASTCALL call_4_2_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_14_11
#define init_4_2_14_11
void FASTCALL call_4_2_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_14_12
#define init_4_2_14_12
void FASTCALL call_4_2_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_14_13
#define init_4_2_14_13
void FASTCALL call_4_2_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_14_14
#define init_4_2_14_14
void FASTCALL call_4_2_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_14_15
#define init_4_2_14_15
void FASTCALL call_4_2_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_15_0
#define init_4_2_15_0
void FASTCALL call_4_2_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_15_1
#define init_4_2_15_1
void FASTCALL call_4_2_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_15_2
#define init_4_2_15_2
void FASTCALL call_4_2_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_15_3
#define init_4_2_15_3
void FASTCALL call_4_2_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_15_4
#define init_4_2_15_4
void FASTCALL call_4_2_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_15_5
#define init_4_2_15_5
void FASTCALL call_4_2_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_15_6
#define init_4_2_15_6
void FASTCALL call_4_2_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_15_7
#define init_4_2_15_7
void FASTCALL call_4_2_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_15_8
#define init_4_2_15_8
void FASTCALL call_4_2_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_15_9
#define init_4_2_15_9
void FASTCALL call_4_2_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_15_10
#define init_4_2_15_10
void FASTCALL call_4_2_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_15_11
#define init_4_2_15_11
void FASTCALL call_4_2_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_15_12
#define init_4_2_15_12
void FASTCALL call_4_2_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_15_13
#define init_4_2_15_13
void FASTCALL call_4_2_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_15_14
#define init_4_2_15_14
void FASTCALL call_4_2_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_2_15_15
#define init_4_2_15_15
void FASTCALL call_4_2_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_0_0
#define init_4_3_0_0
void FASTCALL call_4_3_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_0_1
#define init_4_3_0_1
void FASTCALL call_4_3_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_0_2
#define init_4_3_0_2
void FASTCALL call_4_3_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_0_3
#define init_4_3_0_3
void FASTCALL call_4_3_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_0_4
#define init_4_3_0_4
void FASTCALL call_4_3_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_0_5
#define init_4_3_0_5
void FASTCALL call_4_3_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_0_6
#define init_4_3_0_6
void FASTCALL call_4_3_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_0_7
#define init_4_3_0_7
void FASTCALL call_4_3_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_0_8
#define init_4_3_0_8
void FASTCALL call_4_3_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_0_9
#define init_4_3_0_9
void FASTCALL call_4_3_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_0_10
#define init_4_3_0_10
void FASTCALL call_4_3_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_0_11
#define init_4_3_0_11
void FASTCALL call_4_3_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_0_12
#define init_4_3_0_12
void FASTCALL call_4_3_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_0_13
#define init_4_3_0_13
void FASTCALL call_4_3_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_0_14
#define init_4_3_0_14
void FASTCALL call_4_3_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_0_15
#define init_4_3_0_15
void FASTCALL call_4_3_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_1_0
#define init_4_3_1_0
void FASTCALL call_4_3_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_1_1
#define init_4_3_1_1
void FASTCALL call_4_3_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_1_2
#define init_4_3_1_2
void FASTCALL call_4_3_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_1_3
#define init_4_3_1_3
void FASTCALL call_4_3_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_1_4
#define init_4_3_1_4
void FASTCALL call_4_3_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_1_5
#define init_4_3_1_5
void FASTCALL call_4_3_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_1_6
#define init_4_3_1_6
void FASTCALL call_4_3_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_1_7
#define init_4_3_1_7
void FASTCALL call_4_3_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_1_8
#define init_4_3_1_8
void FASTCALL call_4_3_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_1_9
#define init_4_3_1_9
void FASTCALL call_4_3_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_1_10
#define init_4_3_1_10
void FASTCALL call_4_3_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_1_11
#define init_4_3_1_11
void FASTCALL call_4_3_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_1_12
#define init_4_3_1_12
void FASTCALL call_4_3_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_1_13
#define init_4_3_1_13
void FASTCALL call_4_3_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_1_14
#define init_4_3_1_14
void FASTCALL call_4_3_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_1_15
#define init_4_3_1_15
void FASTCALL call_4_3_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_2_0
#define init_4_3_2_0
void FASTCALL call_4_3_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_2_1
#define init_4_3_2_1
void FASTCALL call_4_3_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_2_2
#define init_4_3_2_2
void FASTCALL call_4_3_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_2_3
#define init_4_3_2_3
void FASTCALL call_4_3_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_2_4
#define init_4_3_2_4
void FASTCALL call_4_3_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_2_5
#define init_4_3_2_5
void FASTCALL call_4_3_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_2_6
#define init_4_3_2_6
void FASTCALL call_4_3_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_2_7
#define init_4_3_2_7
void FASTCALL call_4_3_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_2_8
#define init_4_3_2_8
void FASTCALL call_4_3_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_2_9
#define init_4_3_2_9
void FASTCALL call_4_3_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_2_10
#define init_4_3_2_10
void FASTCALL call_4_3_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_2_11
#define init_4_3_2_11
void FASTCALL call_4_3_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_2_12
#define init_4_3_2_12
void FASTCALL call_4_3_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_2_13
#define init_4_3_2_13
void FASTCALL call_4_3_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_2_14
#define init_4_3_2_14
void FASTCALL call_4_3_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_2_15
#define init_4_3_2_15
void FASTCALL call_4_3_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_3_0
#define init_4_3_3_0
void FASTCALL call_4_3_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_3_1
#define init_4_3_3_1
void FASTCALL call_4_3_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_3_2
#define init_4_3_3_2
void FASTCALL call_4_3_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_3_3
#define init_4_3_3_3
void FASTCALL call_4_3_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_3_4
#define init_4_3_3_4
void FASTCALL call_4_3_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_3_5
#define init_4_3_3_5
void FASTCALL call_4_3_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_3_6
#define init_4_3_3_6
void FASTCALL call_4_3_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_3_7
#define init_4_3_3_7
void FASTCALL call_4_3_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_3_8
#define init_4_3_3_8
void FASTCALL call_4_3_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_3_9
#define init_4_3_3_9
void FASTCALL call_4_3_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_3_10
#define init_4_3_3_10
void FASTCALL call_4_3_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_3_11
#define init_4_3_3_11
void FASTCALL call_4_3_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_3_12
#define init_4_3_3_12
void FASTCALL call_4_3_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_3_13
#define init_4_3_3_13
void FASTCALL call_4_3_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_3_14
#define init_4_3_3_14
void FASTCALL call_4_3_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_3_15
#define init_4_3_3_15
void FASTCALL call_4_3_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_4_0
#define init_4_3_4_0
void FASTCALL call_4_3_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_4_1
#define init_4_3_4_1
void FASTCALL call_4_3_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_4_2
#define init_4_3_4_2
void FASTCALL call_4_3_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_4_3
#define init_4_3_4_3
void FASTCALL call_4_3_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_4_4
#define init_4_3_4_4
void FASTCALL call_4_3_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_4_5
#define init_4_3_4_5
void FASTCALL call_4_3_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_4_6
#define init_4_3_4_6
void FASTCALL call_4_3_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_4_7
#define init_4_3_4_7
void FASTCALL call_4_3_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_4_8
#define init_4_3_4_8
void FASTCALL call_4_3_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_4_9
#define init_4_3_4_9
void FASTCALL call_4_3_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_4_10
#define init_4_3_4_10
void FASTCALL call_4_3_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_4_11
#define init_4_3_4_11
void FASTCALL call_4_3_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_4_12
#define init_4_3_4_12
void FASTCALL call_4_3_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_4_13
#define init_4_3_4_13
void FASTCALL call_4_3_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_4_14
#define init_4_3_4_14
void FASTCALL call_4_3_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_4_15
#define init_4_3_4_15
void FASTCALL call_4_3_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_5_0
#define init_4_3_5_0
void FASTCALL call_4_3_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_5_1
#define init_4_3_5_1
void FASTCALL call_4_3_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_5_2
#define init_4_3_5_2
void FASTCALL call_4_3_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_5_3
#define init_4_3_5_3
void FASTCALL call_4_3_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_5_4
#define init_4_3_5_4
void FASTCALL call_4_3_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_5_5
#define init_4_3_5_5
void FASTCALL call_4_3_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_5_6
#define init_4_3_5_6
void FASTCALL call_4_3_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_5_7
#define init_4_3_5_7
void FASTCALL call_4_3_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_5_8
#define init_4_3_5_8
void FASTCALL call_4_3_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_5_9
#define init_4_3_5_9
void FASTCALL call_4_3_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_5_10
#define init_4_3_5_10
void FASTCALL call_4_3_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_5_11
#define init_4_3_5_11
void FASTCALL call_4_3_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_5_12
#define init_4_3_5_12
void FASTCALL call_4_3_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_5_13
#define init_4_3_5_13
void FASTCALL call_4_3_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_5_14
#define init_4_3_5_14
void FASTCALL call_4_3_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_5_15
#define init_4_3_5_15
void FASTCALL call_4_3_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_6_0
#define init_4_3_6_0
void FASTCALL call_4_3_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_6_1
#define init_4_3_6_1
void FASTCALL call_4_3_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_6_2
#define init_4_3_6_2
void FASTCALL call_4_3_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_6_3
#define init_4_3_6_3
void FASTCALL call_4_3_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_6_4
#define init_4_3_6_4
void FASTCALL call_4_3_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_6_5
#define init_4_3_6_5
void FASTCALL call_4_3_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_6_6
#define init_4_3_6_6
void FASTCALL call_4_3_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_6_7
#define init_4_3_6_7
void FASTCALL call_4_3_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_6_8
#define init_4_3_6_8
void FASTCALL call_4_3_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_6_9
#define init_4_3_6_9
void FASTCALL call_4_3_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_6_10
#define init_4_3_6_10
void FASTCALL call_4_3_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_6_11
#define init_4_3_6_11
void FASTCALL call_4_3_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_6_12
#define init_4_3_6_12
void FASTCALL call_4_3_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_6_13
#define init_4_3_6_13
void FASTCALL call_4_3_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_6_14
#define init_4_3_6_14
void FASTCALL call_4_3_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_6_15
#define init_4_3_6_15
void FASTCALL call_4_3_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_7_0
#define init_4_3_7_0
void FASTCALL call_4_3_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_7_1
#define init_4_3_7_1
void FASTCALL call_4_3_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_7_2
#define init_4_3_7_2
void FASTCALL call_4_3_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_7_3
#define init_4_3_7_3
void FASTCALL call_4_3_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_7_4
#define init_4_3_7_4
void FASTCALL call_4_3_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_7_5
#define init_4_3_7_5
void FASTCALL call_4_3_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_7_6
#define init_4_3_7_6
void FASTCALL call_4_3_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_7_7
#define init_4_3_7_7
void FASTCALL call_4_3_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_7_8
#define init_4_3_7_8
void FASTCALL call_4_3_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_7_9
#define init_4_3_7_9
void FASTCALL call_4_3_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_7_10
#define init_4_3_7_10
void FASTCALL call_4_3_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_7_11
#define init_4_3_7_11
void FASTCALL call_4_3_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_7_12
#define init_4_3_7_12
void FASTCALL call_4_3_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_7_13
#define init_4_3_7_13
void FASTCALL call_4_3_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_7_14
#define init_4_3_7_14
void FASTCALL call_4_3_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_7_15
#define init_4_3_7_15
void FASTCALL call_4_3_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_8_0
#define init_4_3_8_0
void FASTCALL call_4_3_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_8_1
#define init_4_3_8_1
void FASTCALL call_4_3_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_8_2
#define init_4_3_8_2
void FASTCALL call_4_3_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_8_3
#define init_4_3_8_3
void FASTCALL call_4_3_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_8_4
#define init_4_3_8_4
void FASTCALL call_4_3_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_8_5
#define init_4_3_8_5
void FASTCALL call_4_3_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_8_6
#define init_4_3_8_6
void FASTCALL call_4_3_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_8_7
#define init_4_3_8_7
void FASTCALL call_4_3_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_8_8
#define init_4_3_8_8
void FASTCALL call_4_3_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_8_9
#define init_4_3_8_9
void FASTCALL call_4_3_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_8_10
#define init_4_3_8_10
void FASTCALL call_4_3_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_8_11
#define init_4_3_8_11
void FASTCALL call_4_3_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_8_12
#define init_4_3_8_12
void FASTCALL call_4_3_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_8_13
#define init_4_3_8_13
void FASTCALL call_4_3_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_8_14
#define init_4_3_8_14
void FASTCALL call_4_3_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_8_15
#define init_4_3_8_15
void FASTCALL call_4_3_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_9_0
#define init_4_3_9_0
void FASTCALL call_4_3_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_9_1
#define init_4_3_9_1
void FASTCALL call_4_3_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_9_2
#define init_4_3_9_2
void FASTCALL call_4_3_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_9_3
#define init_4_3_9_3
void FASTCALL call_4_3_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_9_4
#define init_4_3_9_4
void FASTCALL call_4_3_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_9_5
#define init_4_3_9_5
void FASTCALL call_4_3_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_9_6
#define init_4_3_9_6
void FASTCALL call_4_3_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_9_7
#define init_4_3_9_7
void FASTCALL call_4_3_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_9_8
#define init_4_3_9_8
void FASTCALL call_4_3_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_9_9
#define init_4_3_9_9
void FASTCALL call_4_3_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_9_10
#define init_4_3_9_10
void FASTCALL call_4_3_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_9_11
#define init_4_3_9_11
void FASTCALL call_4_3_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_9_12
#define init_4_3_9_12
void FASTCALL call_4_3_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_9_13
#define init_4_3_9_13
void FASTCALL call_4_3_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_9_14
#define init_4_3_9_14
void FASTCALL call_4_3_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_9_15
#define init_4_3_9_15
void FASTCALL call_4_3_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_10_0
#define init_4_3_10_0
void FASTCALL call_4_3_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_10_1
#define init_4_3_10_1
void FASTCALL call_4_3_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_10_2
#define init_4_3_10_2
void FASTCALL call_4_3_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_10_3
#define init_4_3_10_3
void FASTCALL call_4_3_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_10_4
#define init_4_3_10_4
void FASTCALL call_4_3_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_10_5
#define init_4_3_10_5
void FASTCALL call_4_3_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_10_6
#define init_4_3_10_6
void FASTCALL call_4_3_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_10_7
#define init_4_3_10_7
void FASTCALL call_4_3_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_10_8
#define init_4_3_10_8
void FASTCALL call_4_3_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_10_9
#define init_4_3_10_9
void FASTCALL call_4_3_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_10_10
#define init_4_3_10_10
void FASTCALL call_4_3_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_10_11
#define init_4_3_10_11
void FASTCALL call_4_3_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_10_12
#define init_4_3_10_12
void FASTCALL call_4_3_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_10_13
#define init_4_3_10_13
void FASTCALL call_4_3_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_10_14
#define init_4_3_10_14
void FASTCALL call_4_3_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_10_15
#define init_4_3_10_15
void FASTCALL call_4_3_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_11_0
#define init_4_3_11_0
void FASTCALL call_4_3_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_11_1
#define init_4_3_11_1
void FASTCALL call_4_3_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_11_2
#define init_4_3_11_2
void FASTCALL call_4_3_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_11_3
#define init_4_3_11_3
void FASTCALL call_4_3_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_11_4
#define init_4_3_11_4
void FASTCALL call_4_3_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_11_5
#define init_4_3_11_5
void FASTCALL call_4_3_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_11_6
#define init_4_3_11_6
void FASTCALL call_4_3_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_11_7
#define init_4_3_11_7
void FASTCALL call_4_3_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_11_8
#define init_4_3_11_8
void FASTCALL call_4_3_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_11_9
#define init_4_3_11_9
void FASTCALL call_4_3_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_11_10
#define init_4_3_11_10
void FASTCALL call_4_3_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_11_11
#define init_4_3_11_11
void FASTCALL call_4_3_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_11_12
#define init_4_3_11_12
void FASTCALL call_4_3_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_11_13
#define init_4_3_11_13
void FASTCALL call_4_3_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_11_14
#define init_4_3_11_14
void FASTCALL call_4_3_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_11_15
#define init_4_3_11_15
void FASTCALL call_4_3_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_12_0
#define init_4_3_12_0
void FASTCALL call_4_3_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_12_1
#define init_4_3_12_1
void FASTCALL call_4_3_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_12_2
#define init_4_3_12_2
void FASTCALL call_4_3_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_12_3
#define init_4_3_12_3
void FASTCALL call_4_3_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_12_4
#define init_4_3_12_4
void FASTCALL call_4_3_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_12_5
#define init_4_3_12_5
void FASTCALL call_4_3_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_12_6
#define init_4_3_12_6
void FASTCALL call_4_3_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_12_7
#define init_4_3_12_7
void FASTCALL call_4_3_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_12_8
#define init_4_3_12_8
void FASTCALL call_4_3_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_12_9
#define init_4_3_12_9
void FASTCALL call_4_3_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_12_10
#define init_4_3_12_10
void FASTCALL call_4_3_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_12_11
#define init_4_3_12_11
void FASTCALL call_4_3_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_12_12
#define init_4_3_12_12
void FASTCALL call_4_3_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_12_13
#define init_4_3_12_13
void FASTCALL call_4_3_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_12_14
#define init_4_3_12_14
void FASTCALL call_4_3_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_12_15
#define init_4_3_12_15
void FASTCALL call_4_3_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_13_0
#define init_4_3_13_0
void FASTCALL call_4_3_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_13_1
#define init_4_3_13_1
void FASTCALL call_4_3_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_13_2
#define init_4_3_13_2
void FASTCALL call_4_3_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_13_3
#define init_4_3_13_3
void FASTCALL call_4_3_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_13_4
#define init_4_3_13_4
void FASTCALL call_4_3_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_13_5
#define init_4_3_13_5
void FASTCALL call_4_3_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_13_6
#define init_4_3_13_6
void FASTCALL call_4_3_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_13_7
#define init_4_3_13_7
void FASTCALL call_4_3_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_13_8
#define init_4_3_13_8
void FASTCALL call_4_3_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_13_9
#define init_4_3_13_9
void FASTCALL call_4_3_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_13_10
#define init_4_3_13_10
void FASTCALL call_4_3_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_13_11
#define init_4_3_13_11
void FASTCALL call_4_3_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_13_12
#define init_4_3_13_12
void FASTCALL call_4_3_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_13_13
#define init_4_3_13_13
void FASTCALL call_4_3_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_13_14
#define init_4_3_13_14
void FASTCALL call_4_3_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_13_15
#define init_4_3_13_15
void FASTCALL call_4_3_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_14_0
#define init_4_3_14_0
void FASTCALL call_4_3_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_14_1
#define init_4_3_14_1
void FASTCALL call_4_3_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_14_2
#define init_4_3_14_2
void FASTCALL call_4_3_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_14_3
#define init_4_3_14_3
void FASTCALL call_4_3_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_14_4
#define init_4_3_14_4
void FASTCALL call_4_3_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_14_5
#define init_4_3_14_5
void FASTCALL call_4_3_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_14_6
#define init_4_3_14_6
void FASTCALL call_4_3_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_14_7
#define init_4_3_14_7
void FASTCALL call_4_3_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_14_8
#define init_4_3_14_8
void FASTCALL call_4_3_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_14_9
#define init_4_3_14_9
void FASTCALL call_4_3_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_14_10
#define init_4_3_14_10
void FASTCALL call_4_3_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_14_11
#define init_4_3_14_11
void FASTCALL call_4_3_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_14_12
#define init_4_3_14_12
void FASTCALL call_4_3_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_14_13
#define init_4_3_14_13
void FASTCALL call_4_3_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_14_14
#define init_4_3_14_14
void FASTCALL call_4_3_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_14_15
#define init_4_3_14_15
void FASTCALL call_4_3_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_15_0
#define init_4_3_15_0
void FASTCALL call_4_3_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_15_1
#define init_4_3_15_1
void FASTCALL call_4_3_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_15_2
#define init_4_3_15_2
void FASTCALL call_4_3_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_15_3
#define init_4_3_15_3
void FASTCALL call_4_3_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_15_4
#define init_4_3_15_4
void FASTCALL call_4_3_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_15_5
#define init_4_3_15_5
void FASTCALL call_4_3_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_15_6
#define init_4_3_15_6
void FASTCALL call_4_3_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_15_7
#define init_4_3_15_7
void FASTCALL call_4_3_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_15_8
#define init_4_3_15_8
void FASTCALL call_4_3_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_15_9
#define init_4_3_15_9
void FASTCALL call_4_3_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_15_10
#define init_4_3_15_10
void FASTCALL call_4_3_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_15_11
#define init_4_3_15_11
void FASTCALL call_4_3_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_15_12
#define init_4_3_15_12
void FASTCALL call_4_3_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_15_13
#define init_4_3_15_13
void FASTCALL call_4_3_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_15_14
#define init_4_3_15_14
void FASTCALL call_4_3_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_3_15_15
#define init_4_3_15_15
void FASTCALL call_4_3_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_0_0
#define init_4_4_0_0
void FASTCALL call_4_4_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_0_1
#define init_4_4_0_1
void FASTCALL call_4_4_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_0_2
#define init_4_4_0_2
void FASTCALL call_4_4_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_0_3
#define init_4_4_0_3
void FASTCALL call_4_4_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_0_4
#define init_4_4_0_4
void FASTCALL call_4_4_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_0_5
#define init_4_4_0_5
void FASTCALL call_4_4_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_0_6
#define init_4_4_0_6
void FASTCALL call_4_4_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_0_7
#define init_4_4_0_7
void FASTCALL call_4_4_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_0_8
#define init_4_4_0_8
void FASTCALL call_4_4_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_0_9
#define init_4_4_0_9
void FASTCALL call_4_4_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_0_10
#define init_4_4_0_10
void FASTCALL call_4_4_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_0_11
#define init_4_4_0_11
void FASTCALL call_4_4_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_0_12
#define init_4_4_0_12
void FASTCALL call_4_4_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_0_13
#define init_4_4_0_13
void FASTCALL call_4_4_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_0_14
#define init_4_4_0_14
void FASTCALL call_4_4_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_0_15
#define init_4_4_0_15
void FASTCALL call_4_4_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_1_0
#define init_4_4_1_0
void FASTCALL call_4_4_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_1_1
#define init_4_4_1_1
void FASTCALL call_4_4_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_1_2
#define init_4_4_1_2
void FASTCALL call_4_4_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_1_3
#define init_4_4_1_3
void FASTCALL call_4_4_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_1_4
#define init_4_4_1_4
void FASTCALL call_4_4_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_1_5
#define init_4_4_1_5
void FASTCALL call_4_4_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_1_6
#define init_4_4_1_6
void FASTCALL call_4_4_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_1_7
#define init_4_4_1_7
void FASTCALL call_4_4_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_1_8
#define init_4_4_1_8
void FASTCALL call_4_4_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_1_9
#define init_4_4_1_9
void FASTCALL call_4_4_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_1_10
#define init_4_4_1_10
void FASTCALL call_4_4_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_1_11
#define init_4_4_1_11
void FASTCALL call_4_4_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_1_12
#define init_4_4_1_12
void FASTCALL call_4_4_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_1_13
#define init_4_4_1_13
void FASTCALL call_4_4_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_1_14
#define init_4_4_1_14
void FASTCALL call_4_4_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_1_15
#define init_4_4_1_15
void FASTCALL call_4_4_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_2_0
#define init_4_4_2_0
void FASTCALL call_4_4_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_2_1
#define init_4_4_2_1
void FASTCALL call_4_4_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_2_2
#define init_4_4_2_2
void FASTCALL call_4_4_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_2_3
#define init_4_4_2_3
void FASTCALL call_4_4_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_2_4
#define init_4_4_2_4
void FASTCALL call_4_4_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_2_5
#define init_4_4_2_5
void FASTCALL call_4_4_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_2_6
#define init_4_4_2_6
void FASTCALL call_4_4_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_2_7
#define init_4_4_2_7
void FASTCALL call_4_4_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_2_8
#define init_4_4_2_8
void FASTCALL call_4_4_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_2_9
#define init_4_4_2_9
void FASTCALL call_4_4_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_2_10
#define init_4_4_2_10
void FASTCALL call_4_4_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_2_11
#define init_4_4_2_11
void FASTCALL call_4_4_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_2_12
#define init_4_4_2_12
void FASTCALL call_4_4_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_2_13
#define init_4_4_2_13
void FASTCALL call_4_4_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_2_14
#define init_4_4_2_14
void FASTCALL call_4_4_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_2_15
#define init_4_4_2_15
void FASTCALL call_4_4_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_3_0
#define init_4_4_3_0
void FASTCALL call_4_4_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_3_1
#define init_4_4_3_1
void FASTCALL call_4_4_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_3_2
#define init_4_4_3_2
void FASTCALL call_4_4_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_3_3
#define init_4_4_3_3
void FASTCALL call_4_4_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_3_4
#define init_4_4_3_4
void FASTCALL call_4_4_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_3_5
#define init_4_4_3_5
void FASTCALL call_4_4_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_3_6
#define init_4_4_3_6
void FASTCALL call_4_4_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_3_7
#define init_4_4_3_7
void FASTCALL call_4_4_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_3_8
#define init_4_4_3_8
void FASTCALL call_4_4_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_3_9
#define init_4_4_3_9
void FASTCALL call_4_4_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_3_10
#define init_4_4_3_10
void FASTCALL call_4_4_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_3_11
#define init_4_4_3_11
void FASTCALL call_4_4_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_3_12
#define init_4_4_3_12
void FASTCALL call_4_4_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_3_13
#define init_4_4_3_13
void FASTCALL call_4_4_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_3_14
#define init_4_4_3_14
void FASTCALL call_4_4_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_3_15
#define init_4_4_3_15
void FASTCALL call_4_4_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_4_0
#define init_4_4_4_0
void FASTCALL call_4_4_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_4_1
#define init_4_4_4_1
void FASTCALL call_4_4_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_4_2
#define init_4_4_4_2
void FASTCALL call_4_4_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_4_3
#define init_4_4_4_3
void FASTCALL call_4_4_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_4_4
#define init_4_4_4_4
void FASTCALL call_4_4_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_4_5
#define init_4_4_4_5
void FASTCALL call_4_4_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_4_6
#define init_4_4_4_6
void FASTCALL call_4_4_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_4_7
#define init_4_4_4_7
void FASTCALL call_4_4_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_4_8
#define init_4_4_4_8
void FASTCALL call_4_4_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_4_9
#define init_4_4_4_9
void FASTCALL call_4_4_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_4_10
#define init_4_4_4_10
void FASTCALL call_4_4_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_4_11
#define init_4_4_4_11
void FASTCALL call_4_4_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_4_12
#define init_4_4_4_12
void FASTCALL call_4_4_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_4_13
#define init_4_4_4_13
void FASTCALL call_4_4_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_4_14
#define init_4_4_4_14
void FASTCALL call_4_4_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_4_15
#define init_4_4_4_15
void FASTCALL call_4_4_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_5_0
#define init_4_4_5_0
void FASTCALL call_4_4_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_5_1
#define init_4_4_5_1
void FASTCALL call_4_4_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_5_2
#define init_4_4_5_2
void FASTCALL call_4_4_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_5_3
#define init_4_4_5_3
void FASTCALL call_4_4_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_5_4
#define init_4_4_5_4
void FASTCALL call_4_4_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_5_5
#define init_4_4_5_5
void FASTCALL call_4_4_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_5_6
#define init_4_4_5_6
void FASTCALL call_4_4_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_5_7
#define init_4_4_5_7
void FASTCALL call_4_4_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_5_8
#define init_4_4_5_8
void FASTCALL call_4_4_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_5_9
#define init_4_4_5_9
void FASTCALL call_4_4_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_5_10
#define init_4_4_5_10
void FASTCALL call_4_4_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_5_11
#define init_4_4_5_11
void FASTCALL call_4_4_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_5_12
#define init_4_4_5_12
void FASTCALL call_4_4_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_5_13
#define init_4_4_5_13
void FASTCALL call_4_4_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_5_14
#define init_4_4_5_14
void FASTCALL call_4_4_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_5_15
#define init_4_4_5_15
void FASTCALL call_4_4_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_6_0
#define init_4_4_6_0
void FASTCALL call_4_4_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_6_1
#define init_4_4_6_1
void FASTCALL call_4_4_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_6_2
#define init_4_4_6_2
void FASTCALL call_4_4_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_6_3
#define init_4_4_6_3
void FASTCALL call_4_4_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_6_4
#define init_4_4_6_4
void FASTCALL call_4_4_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_6_5
#define init_4_4_6_5
void FASTCALL call_4_4_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_6_6
#define init_4_4_6_6
void FASTCALL call_4_4_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_6_7
#define init_4_4_6_7
void FASTCALL call_4_4_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_6_8
#define init_4_4_6_8
void FASTCALL call_4_4_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_6_9
#define init_4_4_6_9
void FASTCALL call_4_4_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_6_10
#define init_4_4_6_10
void FASTCALL call_4_4_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_6_11
#define init_4_4_6_11
void FASTCALL call_4_4_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_6_12
#define init_4_4_6_12
void FASTCALL call_4_4_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_6_13
#define init_4_4_6_13
void FASTCALL call_4_4_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_6_14
#define init_4_4_6_14
void FASTCALL call_4_4_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_6_15
#define init_4_4_6_15
void FASTCALL call_4_4_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_7_0
#define init_4_4_7_0
void FASTCALL call_4_4_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_7_1
#define init_4_4_7_1
void FASTCALL call_4_4_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_7_2
#define init_4_4_7_2
void FASTCALL call_4_4_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_7_3
#define init_4_4_7_3
void FASTCALL call_4_4_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_7_4
#define init_4_4_7_4
void FASTCALL call_4_4_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_7_5
#define init_4_4_7_5
void FASTCALL call_4_4_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_7_6
#define init_4_4_7_6
void FASTCALL call_4_4_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_7_7
#define init_4_4_7_7
void FASTCALL call_4_4_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_7_8
#define init_4_4_7_8
void FASTCALL call_4_4_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_7_9
#define init_4_4_7_9
void FASTCALL call_4_4_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_7_10
#define init_4_4_7_10
void FASTCALL call_4_4_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_7_11
#define init_4_4_7_11
void FASTCALL call_4_4_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_7_12
#define init_4_4_7_12
void FASTCALL call_4_4_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_7_13
#define init_4_4_7_13
void FASTCALL call_4_4_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_7_14
#define init_4_4_7_14
void FASTCALL call_4_4_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_7_15
#define init_4_4_7_15
void FASTCALL call_4_4_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_8_0
#define init_4_4_8_0
void FASTCALL call_4_4_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_8_1
#define init_4_4_8_1
void FASTCALL call_4_4_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_8_2
#define init_4_4_8_2
void FASTCALL call_4_4_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_8_3
#define init_4_4_8_3
void FASTCALL call_4_4_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_8_4
#define init_4_4_8_4
void FASTCALL call_4_4_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_8_5
#define init_4_4_8_5
void FASTCALL call_4_4_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_8_6
#define init_4_4_8_6
void FASTCALL call_4_4_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_8_7
#define init_4_4_8_7
void FASTCALL call_4_4_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_8_8
#define init_4_4_8_8
void FASTCALL call_4_4_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_8_9
#define init_4_4_8_9
void FASTCALL call_4_4_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_8_10
#define init_4_4_8_10
void FASTCALL call_4_4_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_8_11
#define init_4_4_8_11
void FASTCALL call_4_4_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_8_12
#define init_4_4_8_12
void FASTCALL call_4_4_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_8_13
#define init_4_4_8_13
void FASTCALL call_4_4_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_8_14
#define init_4_4_8_14
void FASTCALL call_4_4_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_8_15
#define init_4_4_8_15
void FASTCALL call_4_4_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_9_0
#define init_4_4_9_0
void FASTCALL call_4_4_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_9_1
#define init_4_4_9_1
void FASTCALL call_4_4_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_9_2
#define init_4_4_9_2
void FASTCALL call_4_4_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_9_3
#define init_4_4_9_3
void FASTCALL call_4_4_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_9_4
#define init_4_4_9_4
void FASTCALL call_4_4_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_9_5
#define init_4_4_9_5
void FASTCALL call_4_4_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_9_6
#define init_4_4_9_6
void FASTCALL call_4_4_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_9_7
#define init_4_4_9_7
void FASTCALL call_4_4_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_9_8
#define init_4_4_9_8
void FASTCALL call_4_4_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_9_9
#define init_4_4_9_9
void FASTCALL call_4_4_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_9_10
#define init_4_4_9_10
void FASTCALL call_4_4_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_9_11
#define init_4_4_9_11
void FASTCALL call_4_4_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_9_12
#define init_4_4_9_12
void FASTCALL call_4_4_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_9_13
#define init_4_4_9_13
void FASTCALL call_4_4_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_9_14
#define init_4_4_9_14
void FASTCALL call_4_4_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_9_15
#define init_4_4_9_15
void FASTCALL call_4_4_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_10_0
#define init_4_4_10_0
void FASTCALL call_4_4_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_10_1
#define init_4_4_10_1
void FASTCALL call_4_4_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_10_2
#define init_4_4_10_2
void FASTCALL call_4_4_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_10_3
#define init_4_4_10_3
void FASTCALL call_4_4_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_10_4
#define init_4_4_10_4
void FASTCALL call_4_4_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_10_5
#define init_4_4_10_5
void FASTCALL call_4_4_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_10_6
#define init_4_4_10_6
void FASTCALL call_4_4_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_10_7
#define init_4_4_10_7
void FASTCALL call_4_4_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_10_8
#define init_4_4_10_8
void FASTCALL call_4_4_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_10_9
#define init_4_4_10_9
void FASTCALL call_4_4_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_10_10
#define init_4_4_10_10
void FASTCALL call_4_4_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_10_11
#define init_4_4_10_11
void FASTCALL call_4_4_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_10_12
#define init_4_4_10_12
void FASTCALL call_4_4_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_10_13
#define init_4_4_10_13
void FASTCALL call_4_4_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_10_14
#define init_4_4_10_14
void FASTCALL call_4_4_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_10_15
#define init_4_4_10_15
void FASTCALL call_4_4_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_11_0
#define init_4_4_11_0
void FASTCALL call_4_4_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_11_1
#define init_4_4_11_1
void FASTCALL call_4_4_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_11_2
#define init_4_4_11_2
void FASTCALL call_4_4_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_11_3
#define init_4_4_11_3
void FASTCALL call_4_4_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_11_4
#define init_4_4_11_4
void FASTCALL call_4_4_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_11_5
#define init_4_4_11_5
void FASTCALL call_4_4_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_11_6
#define init_4_4_11_6
void FASTCALL call_4_4_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_11_7
#define init_4_4_11_7
void FASTCALL call_4_4_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_11_8
#define init_4_4_11_8
void FASTCALL call_4_4_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_11_9
#define init_4_4_11_9
void FASTCALL call_4_4_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_11_10
#define init_4_4_11_10
void FASTCALL call_4_4_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_11_11
#define init_4_4_11_11
void FASTCALL call_4_4_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_11_12
#define init_4_4_11_12
void FASTCALL call_4_4_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_11_13
#define init_4_4_11_13
void FASTCALL call_4_4_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_11_14
#define init_4_4_11_14
void FASTCALL call_4_4_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_11_15
#define init_4_4_11_15
void FASTCALL call_4_4_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_12_0
#define init_4_4_12_0
void FASTCALL call_4_4_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_12_1
#define init_4_4_12_1
void FASTCALL call_4_4_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_12_2
#define init_4_4_12_2
void FASTCALL call_4_4_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_12_3
#define init_4_4_12_3
void FASTCALL call_4_4_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_12_4
#define init_4_4_12_4
void FASTCALL call_4_4_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_12_5
#define init_4_4_12_5
void FASTCALL call_4_4_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_12_6
#define init_4_4_12_6
void FASTCALL call_4_4_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_12_7
#define init_4_4_12_7
void FASTCALL call_4_4_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_12_8
#define init_4_4_12_8
void FASTCALL call_4_4_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_12_9
#define init_4_4_12_9
void FASTCALL call_4_4_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_12_10
#define init_4_4_12_10
void FASTCALL call_4_4_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_12_11
#define init_4_4_12_11
void FASTCALL call_4_4_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_12_12
#define init_4_4_12_12
void FASTCALL call_4_4_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_12_13
#define init_4_4_12_13
void FASTCALL call_4_4_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_12_14
#define init_4_4_12_14
void FASTCALL call_4_4_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_12_15
#define init_4_4_12_15
void FASTCALL call_4_4_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_13_0
#define init_4_4_13_0
void FASTCALL call_4_4_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_13_1
#define init_4_4_13_1
void FASTCALL call_4_4_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_13_2
#define init_4_4_13_2
void FASTCALL call_4_4_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_13_3
#define init_4_4_13_3
void FASTCALL call_4_4_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_13_4
#define init_4_4_13_4
void FASTCALL call_4_4_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_13_5
#define init_4_4_13_5
void FASTCALL call_4_4_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_13_6
#define init_4_4_13_6
void FASTCALL call_4_4_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_13_7
#define init_4_4_13_7
void FASTCALL call_4_4_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_13_8
#define init_4_4_13_8
void FASTCALL call_4_4_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_13_9
#define init_4_4_13_9
void FASTCALL call_4_4_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_13_10
#define init_4_4_13_10
void FASTCALL call_4_4_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_13_11
#define init_4_4_13_11
void FASTCALL call_4_4_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_13_12
#define init_4_4_13_12
void FASTCALL call_4_4_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_13_13
#define init_4_4_13_13
void FASTCALL call_4_4_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_13_14
#define init_4_4_13_14
void FASTCALL call_4_4_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_13_15
#define init_4_4_13_15
void FASTCALL call_4_4_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_14_0
#define init_4_4_14_0
void FASTCALL call_4_4_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_14_1
#define init_4_4_14_1
void FASTCALL call_4_4_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_14_2
#define init_4_4_14_2
void FASTCALL call_4_4_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_14_3
#define init_4_4_14_3
void FASTCALL call_4_4_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_14_4
#define init_4_4_14_4
void FASTCALL call_4_4_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_14_5
#define init_4_4_14_5
void FASTCALL call_4_4_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_14_6
#define init_4_4_14_6
void FASTCALL call_4_4_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_14_7
#define init_4_4_14_7
void FASTCALL call_4_4_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_14_8
#define init_4_4_14_8
void FASTCALL call_4_4_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_14_9
#define init_4_4_14_9
void FASTCALL call_4_4_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_14_10
#define init_4_4_14_10
void FASTCALL call_4_4_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_14_11
#define init_4_4_14_11
void FASTCALL call_4_4_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_14_12
#define init_4_4_14_12
void FASTCALL call_4_4_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_14_13
#define init_4_4_14_13
void FASTCALL call_4_4_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_14_14
#define init_4_4_14_14
void FASTCALL call_4_4_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_14_15
#define init_4_4_14_15
void FASTCALL call_4_4_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_15_0
#define init_4_4_15_0
void FASTCALL call_4_4_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_15_1
#define init_4_4_15_1
void FASTCALL call_4_4_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_15_2
#define init_4_4_15_2
void FASTCALL call_4_4_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_15_3
#define init_4_4_15_3
void FASTCALL call_4_4_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_15_4
#define init_4_4_15_4
void FASTCALL call_4_4_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_15_5
#define init_4_4_15_5
void FASTCALL call_4_4_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_15_6
#define init_4_4_15_6
void FASTCALL call_4_4_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_15_7
#define init_4_4_15_7
void FASTCALL call_4_4_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_15_8
#define init_4_4_15_8
void FASTCALL call_4_4_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_15_9
#define init_4_4_15_9
void FASTCALL call_4_4_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_15_10
#define init_4_4_15_10
void FASTCALL call_4_4_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_15_11
#define init_4_4_15_11
void FASTCALL call_4_4_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_15_12
#define init_4_4_15_12
void FASTCALL call_4_4_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_15_13
#define init_4_4_15_13
void FASTCALL call_4_4_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_15_14
#define init_4_4_15_14
void FASTCALL call_4_4_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_4_15_15
#define init_4_4_15_15
void FASTCALL call_4_4_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_0_0
#define init_4_5_0_0
void FASTCALL call_4_5_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_0_1
#define init_4_5_0_1
void FASTCALL call_4_5_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_0_2
#define init_4_5_0_2
void FASTCALL call_4_5_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_0_3
#define init_4_5_0_3
void FASTCALL call_4_5_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_0_4
#define init_4_5_0_4
void FASTCALL call_4_5_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_0_5
#define init_4_5_0_5
void FASTCALL call_4_5_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_0_6
#define init_4_5_0_6
void FASTCALL call_4_5_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_0_7
#define init_4_5_0_7
void FASTCALL call_4_5_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_0_8
#define init_4_5_0_8
void FASTCALL call_4_5_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_0_9
#define init_4_5_0_9
void FASTCALL call_4_5_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_0_10
#define init_4_5_0_10
void FASTCALL call_4_5_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_0_11
#define init_4_5_0_11
void FASTCALL call_4_5_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_0_12
#define init_4_5_0_12
void FASTCALL call_4_5_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_0_13
#define init_4_5_0_13
void FASTCALL call_4_5_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_0_14
#define init_4_5_0_14
void FASTCALL call_4_5_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_0_15
#define init_4_5_0_15
void FASTCALL call_4_5_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_1_0
#define init_4_5_1_0
void FASTCALL call_4_5_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_1_1
#define init_4_5_1_1
void FASTCALL call_4_5_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_1_2
#define init_4_5_1_2
void FASTCALL call_4_5_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_1_3
#define init_4_5_1_3
void FASTCALL call_4_5_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_1_4
#define init_4_5_1_4
void FASTCALL call_4_5_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_1_5
#define init_4_5_1_5
void FASTCALL call_4_5_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_1_6
#define init_4_5_1_6
void FASTCALL call_4_5_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_1_7
#define init_4_5_1_7
void FASTCALL call_4_5_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_1_8
#define init_4_5_1_8
void FASTCALL call_4_5_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_1_9
#define init_4_5_1_9
void FASTCALL call_4_5_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_1_10
#define init_4_5_1_10
void FASTCALL call_4_5_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_1_11
#define init_4_5_1_11
void FASTCALL call_4_5_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_1_12
#define init_4_5_1_12
void FASTCALL call_4_5_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_1_13
#define init_4_5_1_13
void FASTCALL call_4_5_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_1_14
#define init_4_5_1_14
void FASTCALL call_4_5_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_1_15
#define init_4_5_1_15
void FASTCALL call_4_5_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_2_0
#define init_4_5_2_0
void FASTCALL call_4_5_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_2_1
#define init_4_5_2_1
void FASTCALL call_4_5_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_2_2
#define init_4_5_2_2
void FASTCALL call_4_5_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_2_3
#define init_4_5_2_3
void FASTCALL call_4_5_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_2_4
#define init_4_5_2_4
void FASTCALL call_4_5_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_2_5
#define init_4_5_2_5
void FASTCALL call_4_5_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_2_6
#define init_4_5_2_6
void FASTCALL call_4_5_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_2_7
#define init_4_5_2_7
void FASTCALL call_4_5_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_2_8
#define init_4_5_2_8
void FASTCALL call_4_5_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_2_9
#define init_4_5_2_9
void FASTCALL call_4_5_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_2_10
#define init_4_5_2_10
void FASTCALL call_4_5_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_2_11
#define init_4_5_2_11
void FASTCALL call_4_5_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_2_12
#define init_4_5_2_12
void FASTCALL call_4_5_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_2_13
#define init_4_5_2_13
void FASTCALL call_4_5_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_2_14
#define init_4_5_2_14
void FASTCALL call_4_5_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_2_15
#define init_4_5_2_15
void FASTCALL call_4_5_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_3_0
#define init_4_5_3_0
void FASTCALL call_4_5_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_3_1
#define init_4_5_3_1
void FASTCALL call_4_5_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_3_2
#define init_4_5_3_2
void FASTCALL call_4_5_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_3_3
#define init_4_5_3_3
void FASTCALL call_4_5_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_3_4
#define init_4_5_3_4
void FASTCALL call_4_5_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_3_5
#define init_4_5_3_5
void FASTCALL call_4_5_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_3_6
#define init_4_5_3_6
void FASTCALL call_4_5_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_3_7
#define init_4_5_3_7
void FASTCALL call_4_5_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_3_8
#define init_4_5_3_8
void FASTCALL call_4_5_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_3_9
#define init_4_5_3_9
void FASTCALL call_4_5_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_3_10
#define init_4_5_3_10
void FASTCALL call_4_5_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_3_11
#define init_4_5_3_11
void FASTCALL call_4_5_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_3_12
#define init_4_5_3_12
void FASTCALL call_4_5_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_3_13
#define init_4_5_3_13
void FASTCALL call_4_5_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_3_14
#define init_4_5_3_14
void FASTCALL call_4_5_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_3_15
#define init_4_5_3_15
void FASTCALL call_4_5_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_4_0
#define init_4_5_4_0
void FASTCALL call_4_5_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_4_1
#define init_4_5_4_1
void FASTCALL call_4_5_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_4_2
#define init_4_5_4_2
void FASTCALL call_4_5_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_4_3
#define init_4_5_4_3
void FASTCALL call_4_5_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_4_4
#define init_4_5_4_4
void FASTCALL call_4_5_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_4_5
#define init_4_5_4_5
void FASTCALL call_4_5_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_4_6
#define init_4_5_4_6
void FASTCALL call_4_5_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_4_7
#define init_4_5_4_7
void FASTCALL call_4_5_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_4_8
#define init_4_5_4_8
void FASTCALL call_4_5_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_4_9
#define init_4_5_4_9
void FASTCALL call_4_5_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_4_10
#define init_4_5_4_10
void FASTCALL call_4_5_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_4_11
#define init_4_5_4_11
void FASTCALL call_4_5_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_4_12
#define init_4_5_4_12
void FASTCALL call_4_5_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_4_13
#define init_4_5_4_13
void FASTCALL call_4_5_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_4_14
#define init_4_5_4_14
void FASTCALL call_4_5_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_4_15
#define init_4_5_4_15
void FASTCALL call_4_5_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_5_0
#define init_4_5_5_0
void FASTCALL call_4_5_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_5_1
#define init_4_5_5_1
void FASTCALL call_4_5_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_5_2
#define init_4_5_5_2
void FASTCALL call_4_5_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_5_3
#define init_4_5_5_3
void FASTCALL call_4_5_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_5_4
#define init_4_5_5_4
void FASTCALL call_4_5_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_5_5
#define init_4_5_5_5
void FASTCALL call_4_5_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_5_6
#define init_4_5_5_6
void FASTCALL call_4_5_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_5_7
#define init_4_5_5_7
void FASTCALL call_4_5_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_5_8
#define init_4_5_5_8
void FASTCALL call_4_5_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_5_9
#define init_4_5_5_9
void FASTCALL call_4_5_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_5_10
#define init_4_5_5_10
void FASTCALL call_4_5_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_5_11
#define init_4_5_5_11
void FASTCALL call_4_5_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_5_12
#define init_4_5_5_12
void FASTCALL call_4_5_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_5_13
#define init_4_5_5_13
void FASTCALL call_4_5_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_5_14
#define init_4_5_5_14
void FASTCALL call_4_5_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_5_15
#define init_4_5_5_15
void FASTCALL call_4_5_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_6_0
#define init_4_5_6_0
void FASTCALL call_4_5_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_6_1
#define init_4_5_6_1
void FASTCALL call_4_5_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_6_2
#define init_4_5_6_2
void FASTCALL call_4_5_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_6_3
#define init_4_5_6_3
void FASTCALL call_4_5_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_6_4
#define init_4_5_6_4
void FASTCALL call_4_5_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_6_5
#define init_4_5_6_5
void FASTCALL call_4_5_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_6_6
#define init_4_5_6_6
void FASTCALL call_4_5_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_6_7
#define init_4_5_6_7
void FASTCALL call_4_5_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_6_8
#define init_4_5_6_8
void FASTCALL call_4_5_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_6_9
#define init_4_5_6_9
void FASTCALL call_4_5_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_6_10
#define init_4_5_6_10
void FASTCALL call_4_5_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_6_11
#define init_4_5_6_11
void FASTCALL call_4_5_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_6_12
#define init_4_5_6_12
void FASTCALL call_4_5_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_6_13
#define init_4_5_6_13
void FASTCALL call_4_5_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_6_14
#define init_4_5_6_14
void FASTCALL call_4_5_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_6_15
#define init_4_5_6_15
void FASTCALL call_4_5_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_7_0
#define init_4_5_7_0
void FASTCALL call_4_5_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_7_1
#define init_4_5_7_1
void FASTCALL call_4_5_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_7_2
#define init_4_5_7_2
void FASTCALL call_4_5_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_7_3
#define init_4_5_7_3
void FASTCALL call_4_5_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_7_4
#define init_4_5_7_4
void FASTCALL call_4_5_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_7_5
#define init_4_5_7_5
void FASTCALL call_4_5_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_7_6
#define init_4_5_7_6
void FASTCALL call_4_5_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_7_7
#define init_4_5_7_7
void FASTCALL call_4_5_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_7_8
#define init_4_5_7_8
void FASTCALL call_4_5_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_7_9
#define init_4_5_7_9
void FASTCALL call_4_5_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_7_10
#define init_4_5_7_10
void FASTCALL call_4_5_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_7_11
#define init_4_5_7_11
void FASTCALL call_4_5_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_7_12
#define init_4_5_7_12
void FASTCALL call_4_5_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_7_13
#define init_4_5_7_13
void FASTCALL call_4_5_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_7_14
#define init_4_5_7_14
void FASTCALL call_4_5_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_7_15
#define init_4_5_7_15
void FASTCALL call_4_5_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_8_0
#define init_4_5_8_0
void FASTCALL call_4_5_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_8_1
#define init_4_5_8_1
void FASTCALL call_4_5_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_8_2
#define init_4_5_8_2
void FASTCALL call_4_5_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_8_3
#define init_4_5_8_3
void FASTCALL call_4_5_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_8_4
#define init_4_5_8_4
void FASTCALL call_4_5_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_8_5
#define init_4_5_8_5
void FASTCALL call_4_5_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_8_6
#define init_4_5_8_6
void FASTCALL call_4_5_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_8_7
#define init_4_5_8_7
void FASTCALL call_4_5_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_8_8
#define init_4_5_8_8
void FASTCALL call_4_5_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_8_9
#define init_4_5_8_9
void FASTCALL call_4_5_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_8_10
#define init_4_5_8_10
void FASTCALL call_4_5_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_8_11
#define init_4_5_8_11
void FASTCALL call_4_5_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_8_12
#define init_4_5_8_12
void FASTCALL call_4_5_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_8_13
#define init_4_5_8_13
void FASTCALL call_4_5_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_8_14
#define init_4_5_8_14
void FASTCALL call_4_5_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_8_15
#define init_4_5_8_15
void FASTCALL call_4_5_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_9_0
#define init_4_5_9_0
void FASTCALL call_4_5_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_9_1
#define init_4_5_9_1
void FASTCALL call_4_5_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_9_2
#define init_4_5_9_2
void FASTCALL call_4_5_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_9_3
#define init_4_5_9_3
void FASTCALL call_4_5_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_9_4
#define init_4_5_9_4
void FASTCALL call_4_5_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_9_5
#define init_4_5_9_5
void FASTCALL call_4_5_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_9_6
#define init_4_5_9_6
void FASTCALL call_4_5_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_9_7
#define init_4_5_9_7
void FASTCALL call_4_5_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_9_8
#define init_4_5_9_8
void FASTCALL call_4_5_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_9_9
#define init_4_5_9_9
void FASTCALL call_4_5_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_9_10
#define init_4_5_9_10
void FASTCALL call_4_5_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_9_11
#define init_4_5_9_11
void FASTCALL call_4_5_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_9_12
#define init_4_5_9_12
void FASTCALL call_4_5_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_9_13
#define init_4_5_9_13
void FASTCALL call_4_5_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_9_14
#define init_4_5_9_14
void FASTCALL call_4_5_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_9_15
#define init_4_5_9_15
void FASTCALL call_4_5_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_10_0
#define init_4_5_10_0
void FASTCALL call_4_5_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_10_1
#define init_4_5_10_1
void FASTCALL call_4_5_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_10_2
#define init_4_5_10_2
void FASTCALL call_4_5_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_10_3
#define init_4_5_10_3
void FASTCALL call_4_5_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_10_4
#define init_4_5_10_4
void FASTCALL call_4_5_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_10_5
#define init_4_5_10_5
void FASTCALL call_4_5_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_10_6
#define init_4_5_10_6
void FASTCALL call_4_5_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_10_7
#define init_4_5_10_7
void FASTCALL call_4_5_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_10_8
#define init_4_5_10_8
void FASTCALL call_4_5_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_10_9
#define init_4_5_10_9
void FASTCALL call_4_5_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_10_10
#define init_4_5_10_10
void FASTCALL call_4_5_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_10_11
#define init_4_5_10_11
void FASTCALL call_4_5_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_10_12
#define init_4_5_10_12
void FASTCALL call_4_5_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_10_13
#define init_4_5_10_13
void FASTCALL call_4_5_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_10_14
#define init_4_5_10_14
void FASTCALL call_4_5_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_10_15
#define init_4_5_10_15
void FASTCALL call_4_5_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_11_0
#define init_4_5_11_0
void FASTCALL call_4_5_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_11_1
#define init_4_5_11_1
void FASTCALL call_4_5_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_11_2
#define init_4_5_11_2
void FASTCALL call_4_5_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_11_3
#define init_4_5_11_3
void FASTCALL call_4_5_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_11_4
#define init_4_5_11_4
void FASTCALL call_4_5_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_11_5
#define init_4_5_11_5
void FASTCALL call_4_5_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_11_6
#define init_4_5_11_6
void FASTCALL call_4_5_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_11_7
#define init_4_5_11_7
void FASTCALL call_4_5_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_11_8
#define init_4_5_11_8
void FASTCALL call_4_5_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_11_9
#define init_4_5_11_9
void FASTCALL call_4_5_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_11_10
#define init_4_5_11_10
void FASTCALL call_4_5_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_11_11
#define init_4_5_11_11
void FASTCALL call_4_5_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_11_12
#define init_4_5_11_12
void FASTCALL call_4_5_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_11_13
#define init_4_5_11_13
void FASTCALL call_4_5_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_11_14
#define init_4_5_11_14
void FASTCALL call_4_5_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_11_15
#define init_4_5_11_15
void FASTCALL call_4_5_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_12_0
#define init_4_5_12_0
void FASTCALL call_4_5_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_12_1
#define init_4_5_12_1
void FASTCALL call_4_5_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_12_2
#define init_4_5_12_2
void FASTCALL call_4_5_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_12_3
#define init_4_5_12_3
void FASTCALL call_4_5_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_12_4
#define init_4_5_12_4
void FASTCALL call_4_5_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_12_5
#define init_4_5_12_5
void FASTCALL call_4_5_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_12_6
#define init_4_5_12_6
void FASTCALL call_4_5_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_12_7
#define init_4_5_12_7
void FASTCALL call_4_5_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_12_8
#define init_4_5_12_8
void FASTCALL call_4_5_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_12_9
#define init_4_5_12_9
void FASTCALL call_4_5_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_12_10
#define init_4_5_12_10
void FASTCALL call_4_5_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_12_11
#define init_4_5_12_11
void FASTCALL call_4_5_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_12_12
#define init_4_5_12_12
void FASTCALL call_4_5_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_12_13
#define init_4_5_12_13
void FASTCALL call_4_5_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_12_14
#define init_4_5_12_14
void FASTCALL call_4_5_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_12_15
#define init_4_5_12_15
void FASTCALL call_4_5_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_13_0
#define init_4_5_13_0
void FASTCALL call_4_5_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_13_1
#define init_4_5_13_1
void FASTCALL call_4_5_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_13_2
#define init_4_5_13_2
void FASTCALL call_4_5_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_13_3
#define init_4_5_13_3
void FASTCALL call_4_5_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_13_4
#define init_4_5_13_4
void FASTCALL call_4_5_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_13_5
#define init_4_5_13_5
void FASTCALL call_4_5_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_13_6
#define init_4_5_13_6
void FASTCALL call_4_5_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_13_7
#define init_4_5_13_7
void FASTCALL call_4_5_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_13_8
#define init_4_5_13_8
void FASTCALL call_4_5_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_13_9
#define init_4_5_13_9
void FASTCALL call_4_5_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_13_10
#define init_4_5_13_10
void FASTCALL call_4_5_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_13_11
#define init_4_5_13_11
void FASTCALL call_4_5_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_13_12
#define init_4_5_13_12
void FASTCALL call_4_5_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_13_13
#define init_4_5_13_13
void FASTCALL call_4_5_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_13_14
#define init_4_5_13_14
void FASTCALL call_4_5_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_13_15
#define init_4_5_13_15
void FASTCALL call_4_5_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_14_0
#define init_4_5_14_0
void FASTCALL call_4_5_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_14_1
#define init_4_5_14_1
void FASTCALL call_4_5_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_14_2
#define init_4_5_14_2
void FASTCALL call_4_5_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_14_3
#define init_4_5_14_3
void FASTCALL call_4_5_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_14_4
#define init_4_5_14_4
void FASTCALL call_4_5_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_14_5
#define init_4_5_14_5
void FASTCALL call_4_5_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_14_6
#define init_4_5_14_6
void FASTCALL call_4_5_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_14_7
#define init_4_5_14_7
void FASTCALL call_4_5_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_14_8
#define init_4_5_14_8
void FASTCALL call_4_5_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_14_9
#define init_4_5_14_9
void FASTCALL call_4_5_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_14_10
#define init_4_5_14_10
void FASTCALL call_4_5_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_14_11
#define init_4_5_14_11
void FASTCALL call_4_5_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_14_12
#define init_4_5_14_12
void FASTCALL call_4_5_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_14_13
#define init_4_5_14_13
void FASTCALL call_4_5_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_14_14
#define init_4_5_14_14
void FASTCALL call_4_5_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_14_15
#define init_4_5_14_15
void FASTCALL call_4_5_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_15_0
#define init_4_5_15_0
void FASTCALL call_4_5_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_15_1
#define init_4_5_15_1
void FASTCALL call_4_5_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_15_2
#define init_4_5_15_2
void FASTCALL call_4_5_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_15_3
#define init_4_5_15_3
void FASTCALL call_4_5_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_15_4
#define init_4_5_15_4
void FASTCALL call_4_5_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_15_5
#define init_4_5_15_5
void FASTCALL call_4_5_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_15_6
#define init_4_5_15_6
void FASTCALL call_4_5_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_15_7
#define init_4_5_15_7
void FASTCALL call_4_5_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_15_8
#define init_4_5_15_8
void FASTCALL call_4_5_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_15_9
#define init_4_5_15_9
void FASTCALL call_4_5_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_15_10
#define init_4_5_15_10
void FASTCALL call_4_5_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_15_11
#define init_4_5_15_11
void FASTCALL call_4_5_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_15_12
#define init_4_5_15_12
void FASTCALL call_4_5_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_15_13
#define init_4_5_15_13
void FASTCALL call_4_5_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_15_14
#define init_4_5_15_14
void FASTCALL call_4_5_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_5_15_15
#define init_4_5_15_15
void FASTCALL call_4_5_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_0_0
#define init_4_6_0_0
void FASTCALL call_4_6_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_0_1
#define init_4_6_0_1
void FASTCALL call_4_6_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_0_2
#define init_4_6_0_2
void FASTCALL call_4_6_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_0_3
#define init_4_6_0_3
void FASTCALL call_4_6_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_0_4
#define init_4_6_0_4
void FASTCALL call_4_6_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_0_5
#define init_4_6_0_5
void FASTCALL call_4_6_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_0_6
#define init_4_6_0_6
void FASTCALL call_4_6_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_0_7
#define init_4_6_0_7
void FASTCALL call_4_6_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_0_8
#define init_4_6_0_8
void FASTCALL call_4_6_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_0_9
#define init_4_6_0_9
void FASTCALL call_4_6_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_0_10
#define init_4_6_0_10
void FASTCALL call_4_6_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_0_11
#define init_4_6_0_11
void FASTCALL call_4_6_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_0_12
#define init_4_6_0_12
void FASTCALL call_4_6_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_0_13
#define init_4_6_0_13
void FASTCALL call_4_6_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_0_14
#define init_4_6_0_14
void FASTCALL call_4_6_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_0_15
#define init_4_6_0_15
void FASTCALL call_4_6_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_1_0
#define init_4_6_1_0
void FASTCALL call_4_6_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_1_1
#define init_4_6_1_1
void FASTCALL call_4_6_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_1_2
#define init_4_6_1_2
void FASTCALL call_4_6_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_1_3
#define init_4_6_1_3
void FASTCALL call_4_6_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_1_4
#define init_4_6_1_4
void FASTCALL call_4_6_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_1_5
#define init_4_6_1_5
void FASTCALL call_4_6_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_1_6
#define init_4_6_1_6
void FASTCALL call_4_6_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_1_7
#define init_4_6_1_7
void FASTCALL call_4_6_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_1_8
#define init_4_6_1_8
void FASTCALL call_4_6_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_1_9
#define init_4_6_1_9
void FASTCALL call_4_6_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_1_10
#define init_4_6_1_10
void FASTCALL call_4_6_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_1_11
#define init_4_6_1_11
void FASTCALL call_4_6_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_1_12
#define init_4_6_1_12
void FASTCALL call_4_6_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_1_13
#define init_4_6_1_13
void FASTCALL call_4_6_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_1_14
#define init_4_6_1_14
void FASTCALL call_4_6_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_1_15
#define init_4_6_1_15
void FASTCALL call_4_6_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_2_0
#define init_4_6_2_0
void FASTCALL call_4_6_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_2_1
#define init_4_6_2_1
void FASTCALL call_4_6_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_2_2
#define init_4_6_2_2
void FASTCALL call_4_6_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_2_3
#define init_4_6_2_3
void FASTCALL call_4_6_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_2_4
#define init_4_6_2_4
void FASTCALL call_4_6_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_2_5
#define init_4_6_2_5
void FASTCALL call_4_6_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_2_6
#define init_4_6_2_6
void FASTCALL call_4_6_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_2_7
#define init_4_6_2_7
void FASTCALL call_4_6_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_2_8
#define init_4_6_2_8
void FASTCALL call_4_6_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_2_9
#define init_4_6_2_9
void FASTCALL call_4_6_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_2_10
#define init_4_6_2_10
void FASTCALL call_4_6_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_2_11
#define init_4_6_2_11
void FASTCALL call_4_6_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_2_12
#define init_4_6_2_12
void FASTCALL call_4_6_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_2_13
#define init_4_6_2_13
void FASTCALL call_4_6_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_2_14
#define init_4_6_2_14
void FASTCALL call_4_6_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_2_15
#define init_4_6_2_15
void FASTCALL call_4_6_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_3_0
#define init_4_6_3_0
void FASTCALL call_4_6_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_3_1
#define init_4_6_3_1
void FASTCALL call_4_6_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_3_2
#define init_4_6_3_2
void FASTCALL call_4_6_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_3_3
#define init_4_6_3_3
void FASTCALL call_4_6_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_3_4
#define init_4_6_3_4
void FASTCALL call_4_6_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_3_5
#define init_4_6_3_5
void FASTCALL call_4_6_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_3_6
#define init_4_6_3_6
void FASTCALL call_4_6_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_3_7
#define init_4_6_3_7
void FASTCALL call_4_6_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_3_8
#define init_4_6_3_8
void FASTCALL call_4_6_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_3_9
#define init_4_6_3_9
void FASTCALL call_4_6_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_3_10
#define init_4_6_3_10
void FASTCALL call_4_6_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_3_11
#define init_4_6_3_11
void FASTCALL call_4_6_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_3_12
#define init_4_6_3_12
void FASTCALL call_4_6_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_3_13
#define init_4_6_3_13
void FASTCALL call_4_6_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_3_14
#define init_4_6_3_14
void FASTCALL call_4_6_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_3_15
#define init_4_6_3_15
void FASTCALL call_4_6_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_4_0
#define init_4_6_4_0
void FASTCALL call_4_6_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_4_1
#define init_4_6_4_1
void FASTCALL call_4_6_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_4_2
#define init_4_6_4_2
void FASTCALL call_4_6_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_4_3
#define init_4_6_4_3
void FASTCALL call_4_6_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_4_4
#define init_4_6_4_4
void FASTCALL call_4_6_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_4_5
#define init_4_6_4_5
void FASTCALL call_4_6_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_4_6
#define init_4_6_4_6
void FASTCALL call_4_6_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_4_7
#define init_4_6_4_7
void FASTCALL call_4_6_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_4_8
#define init_4_6_4_8
void FASTCALL call_4_6_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_4_9
#define init_4_6_4_9
void FASTCALL call_4_6_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_4_10
#define init_4_6_4_10
void FASTCALL call_4_6_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_4_11
#define init_4_6_4_11
void FASTCALL call_4_6_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_4_12
#define init_4_6_4_12
void FASTCALL call_4_6_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_4_13
#define init_4_6_4_13
void FASTCALL call_4_6_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_4_14
#define init_4_6_4_14
void FASTCALL call_4_6_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_4_15
#define init_4_6_4_15
void FASTCALL call_4_6_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_5_0
#define init_4_6_5_0
void FASTCALL call_4_6_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_5_1
#define init_4_6_5_1
void FASTCALL call_4_6_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_5_2
#define init_4_6_5_2
void FASTCALL call_4_6_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_5_3
#define init_4_6_5_3
void FASTCALL call_4_6_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_5_4
#define init_4_6_5_4
void FASTCALL call_4_6_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_5_5
#define init_4_6_5_5
void FASTCALL call_4_6_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_5_6
#define init_4_6_5_6
void FASTCALL call_4_6_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_5_7
#define init_4_6_5_7
void FASTCALL call_4_6_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_5_8
#define init_4_6_5_8
void FASTCALL call_4_6_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_5_9
#define init_4_6_5_9
void FASTCALL call_4_6_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_5_10
#define init_4_6_5_10
void FASTCALL call_4_6_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_5_11
#define init_4_6_5_11
void FASTCALL call_4_6_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_5_12
#define init_4_6_5_12
void FASTCALL call_4_6_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_5_13
#define init_4_6_5_13
void FASTCALL call_4_6_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_5_14
#define init_4_6_5_14
void FASTCALL call_4_6_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_5_15
#define init_4_6_5_15
void FASTCALL call_4_6_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_6_0
#define init_4_6_6_0
void FASTCALL call_4_6_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_6_1
#define init_4_6_6_1
void FASTCALL call_4_6_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_6_2
#define init_4_6_6_2
void FASTCALL call_4_6_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_6_3
#define init_4_6_6_3
void FASTCALL call_4_6_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_6_4
#define init_4_6_6_4
void FASTCALL call_4_6_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_6_5
#define init_4_6_6_5
void FASTCALL call_4_6_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_6_6
#define init_4_6_6_6
void FASTCALL call_4_6_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_6_7
#define init_4_6_6_7
void FASTCALL call_4_6_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_6_8
#define init_4_6_6_8
void FASTCALL call_4_6_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_6_9
#define init_4_6_6_9
void FASTCALL call_4_6_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_6_10
#define init_4_6_6_10
void FASTCALL call_4_6_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_6_11
#define init_4_6_6_11
void FASTCALL call_4_6_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_6_12
#define init_4_6_6_12
void FASTCALL call_4_6_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_6_13
#define init_4_6_6_13
void FASTCALL call_4_6_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_6_14
#define init_4_6_6_14
void FASTCALL call_4_6_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_6_15
#define init_4_6_6_15
void FASTCALL call_4_6_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_7_0
#define init_4_6_7_0
void FASTCALL call_4_6_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_7_1
#define init_4_6_7_1
void FASTCALL call_4_6_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_7_2
#define init_4_6_7_2
void FASTCALL call_4_6_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_7_3
#define init_4_6_7_3
void FASTCALL call_4_6_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_7_4
#define init_4_6_7_4
void FASTCALL call_4_6_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_7_5
#define init_4_6_7_5
void FASTCALL call_4_6_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_7_6
#define init_4_6_7_6
void FASTCALL call_4_6_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_7_7
#define init_4_6_7_7
void FASTCALL call_4_6_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_7_8
#define init_4_6_7_8
void FASTCALL call_4_6_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_7_9
#define init_4_6_7_9
void FASTCALL call_4_6_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_7_10
#define init_4_6_7_10
void FASTCALL call_4_6_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_7_11
#define init_4_6_7_11
void FASTCALL call_4_6_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_7_12
#define init_4_6_7_12
void FASTCALL call_4_6_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_7_13
#define init_4_6_7_13
void FASTCALL call_4_6_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_7_14
#define init_4_6_7_14
void FASTCALL call_4_6_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_7_15
#define init_4_6_7_15
void FASTCALL call_4_6_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_8_0
#define init_4_6_8_0
void FASTCALL call_4_6_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_8_1
#define init_4_6_8_1
void FASTCALL call_4_6_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_8_2
#define init_4_6_8_2
void FASTCALL call_4_6_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_8_3
#define init_4_6_8_3
void FASTCALL call_4_6_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_8_4
#define init_4_6_8_4
void FASTCALL call_4_6_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_8_5
#define init_4_6_8_5
void FASTCALL call_4_6_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_8_6
#define init_4_6_8_6
void FASTCALL call_4_6_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_8_7
#define init_4_6_8_7
void FASTCALL call_4_6_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_8_8
#define init_4_6_8_8
void FASTCALL call_4_6_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_8_9
#define init_4_6_8_9
void FASTCALL call_4_6_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_8_10
#define init_4_6_8_10
void FASTCALL call_4_6_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_8_11
#define init_4_6_8_11
void FASTCALL call_4_6_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_8_12
#define init_4_6_8_12
void FASTCALL call_4_6_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_8_13
#define init_4_6_8_13
void FASTCALL call_4_6_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_8_14
#define init_4_6_8_14
void FASTCALL call_4_6_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_8_15
#define init_4_6_8_15
void FASTCALL call_4_6_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_9_0
#define init_4_6_9_0
void FASTCALL call_4_6_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_9_1
#define init_4_6_9_1
void FASTCALL call_4_6_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_9_2
#define init_4_6_9_2
void FASTCALL call_4_6_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_9_3
#define init_4_6_9_3
void FASTCALL call_4_6_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_9_4
#define init_4_6_9_4
void FASTCALL call_4_6_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_9_5
#define init_4_6_9_5
void FASTCALL call_4_6_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_9_6
#define init_4_6_9_6
void FASTCALL call_4_6_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_9_7
#define init_4_6_9_7
void FASTCALL call_4_6_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_9_8
#define init_4_6_9_8
void FASTCALL call_4_6_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_9_9
#define init_4_6_9_9
void FASTCALL call_4_6_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_9_10
#define init_4_6_9_10
void FASTCALL call_4_6_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_9_11
#define init_4_6_9_11
void FASTCALL call_4_6_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_9_12
#define init_4_6_9_12
void FASTCALL call_4_6_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_9_13
#define init_4_6_9_13
void FASTCALL call_4_6_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_9_14
#define init_4_6_9_14
void FASTCALL call_4_6_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_9_15
#define init_4_6_9_15
void FASTCALL call_4_6_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_10_0
#define init_4_6_10_0
void FASTCALL call_4_6_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_10_1
#define init_4_6_10_1
void FASTCALL call_4_6_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_10_2
#define init_4_6_10_2
void FASTCALL call_4_6_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_10_3
#define init_4_6_10_3
void FASTCALL call_4_6_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_10_4
#define init_4_6_10_4
void FASTCALL call_4_6_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_10_5
#define init_4_6_10_5
void FASTCALL call_4_6_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_10_6
#define init_4_6_10_6
void FASTCALL call_4_6_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_10_7
#define init_4_6_10_7
void FASTCALL call_4_6_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_10_8
#define init_4_6_10_8
void FASTCALL call_4_6_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_10_9
#define init_4_6_10_9
void FASTCALL call_4_6_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_10_10
#define init_4_6_10_10
void FASTCALL call_4_6_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_10_11
#define init_4_6_10_11
void FASTCALL call_4_6_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_10_12
#define init_4_6_10_12
void FASTCALL call_4_6_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_10_13
#define init_4_6_10_13
void FASTCALL call_4_6_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_10_14
#define init_4_6_10_14
void FASTCALL call_4_6_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_10_15
#define init_4_6_10_15
void FASTCALL call_4_6_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_11_0
#define init_4_6_11_0
void FASTCALL call_4_6_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_11_1
#define init_4_6_11_1
void FASTCALL call_4_6_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_11_2
#define init_4_6_11_2
void FASTCALL call_4_6_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_11_3
#define init_4_6_11_3
void FASTCALL call_4_6_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_11_4
#define init_4_6_11_4
void FASTCALL call_4_6_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_11_5
#define init_4_6_11_5
void FASTCALL call_4_6_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_11_6
#define init_4_6_11_6
void FASTCALL call_4_6_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_11_7
#define init_4_6_11_7
void FASTCALL call_4_6_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_11_8
#define init_4_6_11_8
void FASTCALL call_4_6_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_11_9
#define init_4_6_11_9
void FASTCALL call_4_6_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_11_10
#define init_4_6_11_10
void FASTCALL call_4_6_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_11_11
#define init_4_6_11_11
void FASTCALL call_4_6_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_11_12
#define init_4_6_11_12
void FASTCALL call_4_6_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_11_13
#define init_4_6_11_13
void FASTCALL call_4_6_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_11_14
#define init_4_6_11_14
void FASTCALL call_4_6_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_11_15
#define init_4_6_11_15
void FASTCALL call_4_6_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_12_0
#define init_4_6_12_0
void FASTCALL call_4_6_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_12_1
#define init_4_6_12_1
void FASTCALL call_4_6_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_12_2
#define init_4_6_12_2
void FASTCALL call_4_6_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_12_3
#define init_4_6_12_3
void FASTCALL call_4_6_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_12_4
#define init_4_6_12_4
void FASTCALL call_4_6_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_12_5
#define init_4_6_12_5
void FASTCALL call_4_6_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_12_6
#define init_4_6_12_6
void FASTCALL call_4_6_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_12_7
#define init_4_6_12_7
void FASTCALL call_4_6_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_12_8
#define init_4_6_12_8
void FASTCALL call_4_6_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_12_9
#define init_4_6_12_9
void FASTCALL call_4_6_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_12_10
#define init_4_6_12_10
void FASTCALL call_4_6_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_12_11
#define init_4_6_12_11
void FASTCALL call_4_6_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_12_12
#define init_4_6_12_12
void FASTCALL call_4_6_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_12_13
#define init_4_6_12_13
void FASTCALL call_4_6_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_12_14
#define init_4_6_12_14
void FASTCALL call_4_6_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_12_15
#define init_4_6_12_15
void FASTCALL call_4_6_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_13_0
#define init_4_6_13_0
void FASTCALL call_4_6_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_13_1
#define init_4_6_13_1
void FASTCALL call_4_6_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_13_2
#define init_4_6_13_2
void FASTCALL call_4_6_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_13_3
#define init_4_6_13_3
void FASTCALL call_4_6_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_13_4
#define init_4_6_13_4
void FASTCALL call_4_6_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_13_5
#define init_4_6_13_5
void FASTCALL call_4_6_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_13_6
#define init_4_6_13_6
void FASTCALL call_4_6_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_13_7
#define init_4_6_13_7
void FASTCALL call_4_6_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_13_8
#define init_4_6_13_8
void FASTCALL call_4_6_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_13_9
#define init_4_6_13_9
void FASTCALL call_4_6_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_13_10
#define init_4_6_13_10
void FASTCALL call_4_6_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_13_11
#define init_4_6_13_11
void FASTCALL call_4_6_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_13_12
#define init_4_6_13_12
void FASTCALL call_4_6_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_13_13
#define init_4_6_13_13
void FASTCALL call_4_6_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_13_14
#define init_4_6_13_14
void FASTCALL call_4_6_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_13_15
#define init_4_6_13_15
void FASTCALL call_4_6_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_14_0
#define init_4_6_14_0
void FASTCALL call_4_6_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_14_1
#define init_4_6_14_1
void FASTCALL call_4_6_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_14_2
#define init_4_6_14_2
void FASTCALL call_4_6_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_14_3
#define init_4_6_14_3
void FASTCALL call_4_6_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_14_4
#define init_4_6_14_4
void FASTCALL call_4_6_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_14_5
#define init_4_6_14_5
void FASTCALL call_4_6_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_14_6
#define init_4_6_14_6
void FASTCALL call_4_6_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_14_7
#define init_4_6_14_7
void FASTCALL call_4_6_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_14_8
#define init_4_6_14_8
void FASTCALL call_4_6_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_14_9
#define init_4_6_14_9
void FASTCALL call_4_6_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_14_10
#define init_4_6_14_10
void FASTCALL call_4_6_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_14_11
#define init_4_6_14_11
void FASTCALL call_4_6_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_14_12
#define init_4_6_14_12
void FASTCALL call_4_6_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_14_13
#define init_4_6_14_13
void FASTCALL call_4_6_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_14_14
#define init_4_6_14_14
void FASTCALL call_4_6_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_14_15
#define init_4_6_14_15
void FASTCALL call_4_6_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_15_0
#define init_4_6_15_0
void FASTCALL call_4_6_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_15_1
#define init_4_6_15_1
void FASTCALL call_4_6_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_15_2
#define init_4_6_15_2
void FASTCALL call_4_6_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_15_3
#define init_4_6_15_3
void FASTCALL call_4_6_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_15_4
#define init_4_6_15_4
void FASTCALL call_4_6_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_15_5
#define init_4_6_15_5
void FASTCALL call_4_6_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_15_6
#define init_4_6_15_6
void FASTCALL call_4_6_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_15_7
#define init_4_6_15_7
void FASTCALL call_4_6_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_15_8
#define init_4_6_15_8
void FASTCALL call_4_6_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_15_9
#define init_4_6_15_9
void FASTCALL call_4_6_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_15_10
#define init_4_6_15_10
void FASTCALL call_4_6_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_15_11
#define init_4_6_15_11
void FASTCALL call_4_6_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_15_12
#define init_4_6_15_12
void FASTCALL call_4_6_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_15_13
#define init_4_6_15_13
void FASTCALL call_4_6_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_15_14
#define init_4_6_15_14
void FASTCALL call_4_6_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_6_15_15
#define init_4_6_15_15
void FASTCALL call_4_6_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_0_0
#define init_4_7_0_0
void FASTCALL call_4_7_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_0_1
#define init_4_7_0_1
void FASTCALL call_4_7_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_0_2
#define init_4_7_0_2
void FASTCALL call_4_7_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_0_3
#define init_4_7_0_3
void FASTCALL call_4_7_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_0_4
#define init_4_7_0_4
void FASTCALL call_4_7_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_0_5
#define init_4_7_0_5
void FASTCALL call_4_7_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_0_6
#define init_4_7_0_6
void FASTCALL call_4_7_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_0_7
#define init_4_7_0_7
void FASTCALL call_4_7_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_0_8
#define init_4_7_0_8
void FASTCALL call_4_7_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_0_9
#define init_4_7_0_9
void FASTCALL call_4_7_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_0_10
#define init_4_7_0_10
void FASTCALL call_4_7_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_0_11
#define init_4_7_0_11
void FASTCALL call_4_7_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_0_12
#define init_4_7_0_12
void FASTCALL call_4_7_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_0_13
#define init_4_7_0_13
void FASTCALL call_4_7_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_0_14
#define init_4_7_0_14
void FASTCALL call_4_7_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_0_15
#define init_4_7_0_15
void FASTCALL call_4_7_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_1_0
#define init_4_7_1_0
void FASTCALL call_4_7_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_1_1
#define init_4_7_1_1
void FASTCALL call_4_7_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_1_2
#define init_4_7_1_2
void FASTCALL call_4_7_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_1_3
#define init_4_7_1_3
void FASTCALL call_4_7_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_1_4
#define init_4_7_1_4
void FASTCALL call_4_7_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_1_5
#define init_4_7_1_5
void FASTCALL call_4_7_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_1_6
#define init_4_7_1_6
void FASTCALL call_4_7_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_1_7
#define init_4_7_1_7
void FASTCALL call_4_7_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_1_8
#define init_4_7_1_8
void FASTCALL call_4_7_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_1_9
#define init_4_7_1_9
void FASTCALL call_4_7_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_1_10
#define init_4_7_1_10
void FASTCALL call_4_7_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_1_11
#define init_4_7_1_11
void FASTCALL call_4_7_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_1_12
#define init_4_7_1_12
void FASTCALL call_4_7_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_1_13
#define init_4_7_1_13
void FASTCALL call_4_7_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_1_14
#define init_4_7_1_14
void FASTCALL call_4_7_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_1_15
#define init_4_7_1_15
void FASTCALL call_4_7_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_2_0
#define init_4_7_2_0
void FASTCALL call_4_7_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_2_1
#define init_4_7_2_1
void FASTCALL call_4_7_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_2_2
#define init_4_7_2_2
void FASTCALL call_4_7_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_2_3
#define init_4_7_2_3
void FASTCALL call_4_7_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_2_4
#define init_4_7_2_4
void FASTCALL call_4_7_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_2_5
#define init_4_7_2_5
void FASTCALL call_4_7_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_2_6
#define init_4_7_2_6
void FASTCALL call_4_7_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_2_7
#define init_4_7_2_7
void FASTCALL call_4_7_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_2_8
#define init_4_7_2_8
void FASTCALL call_4_7_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_2_9
#define init_4_7_2_9
void FASTCALL call_4_7_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_2_10
#define init_4_7_2_10
void FASTCALL call_4_7_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_2_11
#define init_4_7_2_11
void FASTCALL call_4_7_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_2_12
#define init_4_7_2_12
void FASTCALL call_4_7_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_2_13
#define init_4_7_2_13
void FASTCALL call_4_7_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_2_14
#define init_4_7_2_14
void FASTCALL call_4_7_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_2_15
#define init_4_7_2_15
void FASTCALL call_4_7_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_3_0
#define init_4_7_3_0
void FASTCALL call_4_7_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_3_1
#define init_4_7_3_1
void FASTCALL call_4_7_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_3_2
#define init_4_7_3_2
void FASTCALL call_4_7_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_3_3
#define init_4_7_3_3
void FASTCALL call_4_7_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_3_4
#define init_4_7_3_4
void FASTCALL call_4_7_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_3_5
#define init_4_7_3_5
void FASTCALL call_4_7_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_3_6
#define init_4_7_3_6
void FASTCALL call_4_7_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_3_7
#define init_4_7_3_7
void FASTCALL call_4_7_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_3_8
#define init_4_7_3_8
void FASTCALL call_4_7_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_3_9
#define init_4_7_3_9
void FASTCALL call_4_7_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_3_10
#define init_4_7_3_10
void FASTCALL call_4_7_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_3_11
#define init_4_7_3_11
void FASTCALL call_4_7_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_3_12
#define init_4_7_3_12
void FASTCALL call_4_7_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_3_13
#define init_4_7_3_13
void FASTCALL call_4_7_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_3_14
#define init_4_7_3_14
void FASTCALL call_4_7_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_3_15
#define init_4_7_3_15
void FASTCALL call_4_7_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_4_0
#define init_4_7_4_0
void FASTCALL call_4_7_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_4_1
#define init_4_7_4_1
void FASTCALL call_4_7_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_4_2
#define init_4_7_4_2
void FASTCALL call_4_7_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_4_3
#define init_4_7_4_3
void FASTCALL call_4_7_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_4_4
#define init_4_7_4_4
void FASTCALL call_4_7_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_4_5
#define init_4_7_4_5
void FASTCALL call_4_7_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_4_6
#define init_4_7_4_6
void FASTCALL call_4_7_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_4_7
#define init_4_7_4_7
void FASTCALL call_4_7_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_4_8
#define init_4_7_4_8
void FASTCALL call_4_7_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_4_9
#define init_4_7_4_9
void FASTCALL call_4_7_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_4_10
#define init_4_7_4_10
void FASTCALL call_4_7_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_4_11
#define init_4_7_4_11
void FASTCALL call_4_7_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_4_12
#define init_4_7_4_12
void FASTCALL call_4_7_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_4_13
#define init_4_7_4_13
void FASTCALL call_4_7_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_4_14
#define init_4_7_4_14
void FASTCALL call_4_7_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_4_15
#define init_4_7_4_15
void FASTCALL call_4_7_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_5_0
#define init_4_7_5_0
void FASTCALL call_4_7_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_5_1
#define init_4_7_5_1
void FASTCALL call_4_7_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_5_2
#define init_4_7_5_2
void FASTCALL call_4_7_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_5_3
#define init_4_7_5_3
void FASTCALL call_4_7_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_5_4
#define init_4_7_5_4
void FASTCALL call_4_7_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_5_5
#define init_4_7_5_5
void FASTCALL call_4_7_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_5_6
#define init_4_7_5_6
void FASTCALL call_4_7_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_5_7
#define init_4_7_5_7
void FASTCALL call_4_7_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_5_8
#define init_4_7_5_8
void FASTCALL call_4_7_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_5_9
#define init_4_7_5_9
void FASTCALL call_4_7_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_5_10
#define init_4_7_5_10
void FASTCALL call_4_7_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_5_11
#define init_4_7_5_11
void FASTCALL call_4_7_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_5_12
#define init_4_7_5_12
void FASTCALL call_4_7_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_5_13
#define init_4_7_5_13
void FASTCALL call_4_7_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_5_14
#define init_4_7_5_14
void FASTCALL call_4_7_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_5_15
#define init_4_7_5_15
void FASTCALL call_4_7_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_6_0
#define init_4_7_6_0
void FASTCALL call_4_7_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_6_1
#define init_4_7_6_1
void FASTCALL call_4_7_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_6_2
#define init_4_7_6_2
void FASTCALL call_4_7_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_6_3
#define init_4_7_6_3
void FASTCALL call_4_7_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_6_4
#define init_4_7_6_4
void FASTCALL call_4_7_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_6_5
#define init_4_7_6_5
void FASTCALL call_4_7_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_6_6
#define init_4_7_6_6
void FASTCALL call_4_7_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_6_7
#define init_4_7_6_7
void FASTCALL call_4_7_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_6_8
#define init_4_7_6_8
void FASTCALL call_4_7_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_6_9
#define init_4_7_6_9
void FASTCALL call_4_7_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_6_10
#define init_4_7_6_10
void FASTCALL call_4_7_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_6_11
#define init_4_7_6_11
void FASTCALL call_4_7_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_6_12
#define init_4_7_6_12
void FASTCALL call_4_7_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_6_13
#define init_4_7_6_13
void FASTCALL call_4_7_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_6_14
#define init_4_7_6_14
void FASTCALL call_4_7_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_6_15
#define init_4_7_6_15
void FASTCALL call_4_7_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_7_0
#define init_4_7_7_0
void FASTCALL call_4_7_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_7_1
#define init_4_7_7_1
void FASTCALL call_4_7_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_7_2
#define init_4_7_7_2
void FASTCALL call_4_7_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_7_3
#define init_4_7_7_3
void FASTCALL call_4_7_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_7_4
#define init_4_7_7_4
void FASTCALL call_4_7_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_7_5
#define init_4_7_7_5
void FASTCALL call_4_7_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_7_6
#define init_4_7_7_6
void FASTCALL call_4_7_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_7_7
#define init_4_7_7_7
void FASTCALL call_4_7_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_7_8
#define init_4_7_7_8
void FASTCALL call_4_7_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_7_9
#define init_4_7_7_9
void FASTCALL call_4_7_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_7_10
#define init_4_7_7_10
void FASTCALL call_4_7_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_7_11
#define init_4_7_7_11
void FASTCALL call_4_7_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_7_12
#define init_4_7_7_12
void FASTCALL call_4_7_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_7_13
#define init_4_7_7_13
void FASTCALL call_4_7_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_7_14
#define init_4_7_7_14
void FASTCALL call_4_7_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_7_15
#define init_4_7_7_15
void FASTCALL call_4_7_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_8_0
#define init_4_7_8_0
void FASTCALL call_4_7_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_8_1
#define init_4_7_8_1
void FASTCALL call_4_7_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_8_2
#define init_4_7_8_2
void FASTCALL call_4_7_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_8_3
#define init_4_7_8_3
void FASTCALL call_4_7_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_8_4
#define init_4_7_8_4
void FASTCALL call_4_7_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_8_5
#define init_4_7_8_5
void FASTCALL call_4_7_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_8_6
#define init_4_7_8_6
void FASTCALL call_4_7_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_8_7
#define init_4_7_8_7
void FASTCALL call_4_7_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_8_8
#define init_4_7_8_8
void FASTCALL call_4_7_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_8_9
#define init_4_7_8_9
void FASTCALL call_4_7_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_8_10
#define init_4_7_8_10
void FASTCALL call_4_7_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_8_11
#define init_4_7_8_11
void FASTCALL call_4_7_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_8_12
#define init_4_7_8_12
void FASTCALL call_4_7_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_8_13
#define init_4_7_8_13
void FASTCALL call_4_7_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_8_14
#define init_4_7_8_14
void FASTCALL call_4_7_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_8_15
#define init_4_7_8_15
void FASTCALL call_4_7_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_9_0
#define init_4_7_9_0
void FASTCALL call_4_7_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_9_1
#define init_4_7_9_1
void FASTCALL call_4_7_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_9_2
#define init_4_7_9_2
void FASTCALL call_4_7_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_9_3
#define init_4_7_9_3
void FASTCALL call_4_7_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_9_4
#define init_4_7_9_4
void FASTCALL call_4_7_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_9_5
#define init_4_7_9_5
void FASTCALL call_4_7_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_9_6
#define init_4_7_9_6
void FASTCALL call_4_7_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_9_7
#define init_4_7_9_7
void FASTCALL call_4_7_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_9_8
#define init_4_7_9_8
void FASTCALL call_4_7_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_9_9
#define init_4_7_9_9
void FASTCALL call_4_7_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_9_10
#define init_4_7_9_10
void FASTCALL call_4_7_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_9_11
#define init_4_7_9_11
void FASTCALL call_4_7_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_9_12
#define init_4_7_9_12
void FASTCALL call_4_7_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_9_13
#define init_4_7_9_13
void FASTCALL call_4_7_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_9_14
#define init_4_7_9_14
void FASTCALL call_4_7_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_9_15
#define init_4_7_9_15
void FASTCALL call_4_7_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_10_0
#define init_4_7_10_0
void FASTCALL call_4_7_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_10_1
#define init_4_7_10_1
void FASTCALL call_4_7_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_10_2
#define init_4_7_10_2
void FASTCALL call_4_7_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_10_3
#define init_4_7_10_3
void FASTCALL call_4_7_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_10_4
#define init_4_7_10_4
void FASTCALL call_4_7_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_10_5
#define init_4_7_10_5
void FASTCALL call_4_7_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_10_6
#define init_4_7_10_6
void FASTCALL call_4_7_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_10_7
#define init_4_7_10_7
void FASTCALL call_4_7_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_10_8
#define init_4_7_10_8
void FASTCALL call_4_7_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_10_9
#define init_4_7_10_9
void FASTCALL call_4_7_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_10_10
#define init_4_7_10_10
void FASTCALL call_4_7_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_10_11
#define init_4_7_10_11
void FASTCALL call_4_7_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_10_12
#define init_4_7_10_12
void FASTCALL call_4_7_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_10_13
#define init_4_7_10_13
void FASTCALL call_4_7_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_10_14
#define init_4_7_10_14
void FASTCALL call_4_7_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_10_15
#define init_4_7_10_15
void FASTCALL call_4_7_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_11_0
#define init_4_7_11_0
void FASTCALL call_4_7_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_11_1
#define init_4_7_11_1
void FASTCALL call_4_7_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_11_2
#define init_4_7_11_2
void FASTCALL call_4_7_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_11_3
#define init_4_7_11_3
void FASTCALL call_4_7_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_11_4
#define init_4_7_11_4
void FASTCALL call_4_7_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_11_5
#define init_4_7_11_5
void FASTCALL call_4_7_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_11_6
#define init_4_7_11_6
void FASTCALL call_4_7_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_11_7
#define init_4_7_11_7
void FASTCALL call_4_7_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_11_8
#define init_4_7_11_8
void FASTCALL call_4_7_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_11_9
#define init_4_7_11_9
void FASTCALL call_4_7_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_11_10
#define init_4_7_11_10
void FASTCALL call_4_7_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_11_11
#define init_4_7_11_11
void FASTCALL call_4_7_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_11_12
#define init_4_7_11_12
void FASTCALL call_4_7_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_11_13
#define init_4_7_11_13
void FASTCALL call_4_7_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_11_14
#define init_4_7_11_14
void FASTCALL call_4_7_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_11_15
#define init_4_7_11_15
void FASTCALL call_4_7_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_12_0
#define init_4_7_12_0
void FASTCALL call_4_7_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_12_1
#define init_4_7_12_1
void FASTCALL call_4_7_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_12_2
#define init_4_7_12_2
void FASTCALL call_4_7_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_12_3
#define init_4_7_12_3
void FASTCALL call_4_7_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_12_4
#define init_4_7_12_4
void FASTCALL call_4_7_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_12_5
#define init_4_7_12_5
void FASTCALL call_4_7_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_12_6
#define init_4_7_12_6
void FASTCALL call_4_7_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_12_7
#define init_4_7_12_7
void FASTCALL call_4_7_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_12_8
#define init_4_7_12_8
void FASTCALL call_4_7_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_12_9
#define init_4_7_12_9
void FASTCALL call_4_7_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_12_10
#define init_4_7_12_10
void FASTCALL call_4_7_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_12_11
#define init_4_7_12_11
void FASTCALL call_4_7_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_12_12
#define init_4_7_12_12
void FASTCALL call_4_7_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_12_13
#define init_4_7_12_13
void FASTCALL call_4_7_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_12_14
#define init_4_7_12_14
void FASTCALL call_4_7_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_12_15
#define init_4_7_12_15
void FASTCALL call_4_7_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_13_0
#define init_4_7_13_0
void FASTCALL call_4_7_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_13_1
#define init_4_7_13_1
void FASTCALL call_4_7_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_13_2
#define init_4_7_13_2
void FASTCALL call_4_7_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_13_3
#define init_4_7_13_3
void FASTCALL call_4_7_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_13_4
#define init_4_7_13_4
void FASTCALL call_4_7_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_13_5
#define init_4_7_13_5
void FASTCALL call_4_7_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_13_6
#define init_4_7_13_6
void FASTCALL call_4_7_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_13_7
#define init_4_7_13_7
void FASTCALL call_4_7_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_13_8
#define init_4_7_13_8
void FASTCALL call_4_7_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_13_9
#define init_4_7_13_9
void FASTCALL call_4_7_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_13_10
#define init_4_7_13_10
void FASTCALL call_4_7_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_13_11
#define init_4_7_13_11
void FASTCALL call_4_7_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_13_12
#define init_4_7_13_12
void FASTCALL call_4_7_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_13_13
#define init_4_7_13_13
void FASTCALL call_4_7_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_13_14
#define init_4_7_13_14
void FASTCALL call_4_7_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_13_15
#define init_4_7_13_15
void FASTCALL call_4_7_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_14_0
#define init_4_7_14_0
void FASTCALL call_4_7_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_14_1
#define init_4_7_14_1
void FASTCALL call_4_7_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_14_2
#define init_4_7_14_2
void FASTCALL call_4_7_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_14_3
#define init_4_7_14_3
void FASTCALL call_4_7_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_14_4
#define init_4_7_14_4
void FASTCALL call_4_7_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_14_5
#define init_4_7_14_5
void FASTCALL call_4_7_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_14_6
#define init_4_7_14_6
void FASTCALL call_4_7_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_14_7
#define init_4_7_14_7
void FASTCALL call_4_7_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_14_8
#define init_4_7_14_8
void FASTCALL call_4_7_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_14_9
#define init_4_7_14_9
void FASTCALL call_4_7_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_14_10
#define init_4_7_14_10
void FASTCALL call_4_7_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_14_11
#define init_4_7_14_11
void FASTCALL call_4_7_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_14_12
#define init_4_7_14_12
void FASTCALL call_4_7_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_14_13
#define init_4_7_14_13
void FASTCALL call_4_7_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_14_14
#define init_4_7_14_14
void FASTCALL call_4_7_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_14_15
#define init_4_7_14_15
void FASTCALL call_4_7_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_15_0
#define init_4_7_15_0
void FASTCALL call_4_7_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_15_1
#define init_4_7_15_1
void FASTCALL call_4_7_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_15_2
#define init_4_7_15_2
void FASTCALL call_4_7_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_15_3
#define init_4_7_15_3
void FASTCALL call_4_7_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_15_4
#define init_4_7_15_4
void FASTCALL call_4_7_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_15_5
#define init_4_7_15_5
void FASTCALL call_4_7_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_15_6
#define init_4_7_15_6
void FASTCALL call_4_7_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_15_7
#define init_4_7_15_7
void FASTCALL call_4_7_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_15_8
#define init_4_7_15_8
void FASTCALL call_4_7_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_15_9
#define init_4_7_15_9
void FASTCALL call_4_7_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_15_10
#define init_4_7_15_10
void FASTCALL call_4_7_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_15_11
#define init_4_7_15_11
void FASTCALL call_4_7_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_15_12
#define init_4_7_15_12
void FASTCALL call_4_7_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_15_13
#define init_4_7_15_13
void FASTCALL call_4_7_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_15_14
#define init_4_7_15_14
void FASTCALL call_4_7_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_7_15_15
#define init_4_7_15_15
void FASTCALL call_4_7_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_0_0
#define init_4_8_0_0
void FASTCALL call_4_8_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_0_1
#define init_4_8_0_1
void FASTCALL call_4_8_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_0_2
#define init_4_8_0_2
void FASTCALL call_4_8_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_0_3
#define init_4_8_0_3
void FASTCALL call_4_8_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_0_4
#define init_4_8_0_4
void FASTCALL call_4_8_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_0_5
#define init_4_8_0_5
void FASTCALL call_4_8_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_0_6
#define init_4_8_0_6
void FASTCALL call_4_8_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_0_7
#define init_4_8_0_7
void FASTCALL call_4_8_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_0_8
#define init_4_8_0_8
void FASTCALL call_4_8_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_0_9
#define init_4_8_0_9
void FASTCALL call_4_8_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_0_10
#define init_4_8_0_10
void FASTCALL call_4_8_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_0_11
#define init_4_8_0_11
void FASTCALL call_4_8_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_0_12
#define init_4_8_0_12
void FASTCALL call_4_8_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_0_13
#define init_4_8_0_13
void FASTCALL call_4_8_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_0_14
#define init_4_8_0_14
void FASTCALL call_4_8_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_0_15
#define init_4_8_0_15
void FASTCALL call_4_8_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_1_0
#define init_4_8_1_0
void FASTCALL call_4_8_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_1_1
#define init_4_8_1_1
void FASTCALL call_4_8_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_1_2
#define init_4_8_1_2
void FASTCALL call_4_8_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_1_3
#define init_4_8_1_3
void FASTCALL call_4_8_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_1_4
#define init_4_8_1_4
void FASTCALL call_4_8_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_1_5
#define init_4_8_1_5
void FASTCALL call_4_8_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_1_6
#define init_4_8_1_6
void FASTCALL call_4_8_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_1_7
#define init_4_8_1_7
void FASTCALL call_4_8_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_1_8
#define init_4_8_1_8
void FASTCALL call_4_8_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_1_9
#define init_4_8_1_9
void FASTCALL call_4_8_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_1_10
#define init_4_8_1_10
void FASTCALL call_4_8_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_1_11
#define init_4_8_1_11
void FASTCALL call_4_8_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_1_12
#define init_4_8_1_12
void FASTCALL call_4_8_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_1_13
#define init_4_8_1_13
void FASTCALL call_4_8_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_1_14
#define init_4_8_1_14
void FASTCALL call_4_8_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_1_15
#define init_4_8_1_15
void FASTCALL call_4_8_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_2_0
#define init_4_8_2_0
void FASTCALL call_4_8_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_2_1
#define init_4_8_2_1
void FASTCALL call_4_8_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_2_2
#define init_4_8_2_2
void FASTCALL call_4_8_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_2_3
#define init_4_8_2_3
void FASTCALL call_4_8_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_2_4
#define init_4_8_2_4
void FASTCALL call_4_8_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_2_5
#define init_4_8_2_5
void FASTCALL call_4_8_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_2_6
#define init_4_8_2_6
void FASTCALL call_4_8_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_2_7
#define init_4_8_2_7
void FASTCALL call_4_8_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_2_8
#define init_4_8_2_8
void FASTCALL call_4_8_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_2_9
#define init_4_8_2_9
void FASTCALL call_4_8_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_2_10
#define init_4_8_2_10
void FASTCALL call_4_8_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_2_11
#define init_4_8_2_11
void FASTCALL call_4_8_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_2_12
#define init_4_8_2_12
void FASTCALL call_4_8_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_2_13
#define init_4_8_2_13
void FASTCALL call_4_8_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_2_14
#define init_4_8_2_14
void FASTCALL call_4_8_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_2_15
#define init_4_8_2_15
void FASTCALL call_4_8_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_3_0
#define init_4_8_3_0
void FASTCALL call_4_8_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_3_1
#define init_4_8_3_1
void FASTCALL call_4_8_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_3_2
#define init_4_8_3_2
void FASTCALL call_4_8_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_3_3
#define init_4_8_3_3
void FASTCALL call_4_8_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_3_4
#define init_4_8_3_4
void FASTCALL call_4_8_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_3_5
#define init_4_8_3_5
void FASTCALL call_4_8_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_3_6
#define init_4_8_3_6
void FASTCALL call_4_8_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_3_7
#define init_4_8_3_7
void FASTCALL call_4_8_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_3_8
#define init_4_8_3_8
void FASTCALL call_4_8_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_3_9
#define init_4_8_3_9
void FASTCALL call_4_8_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_3_10
#define init_4_8_3_10
void FASTCALL call_4_8_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_3_11
#define init_4_8_3_11
void FASTCALL call_4_8_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_3_12
#define init_4_8_3_12
void FASTCALL call_4_8_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_3_13
#define init_4_8_3_13
void FASTCALL call_4_8_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_3_14
#define init_4_8_3_14
void FASTCALL call_4_8_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_3_15
#define init_4_8_3_15
void FASTCALL call_4_8_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_4_0
#define init_4_8_4_0
void FASTCALL call_4_8_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_4_1
#define init_4_8_4_1
void FASTCALL call_4_8_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_4_2
#define init_4_8_4_2
void FASTCALL call_4_8_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_4_3
#define init_4_8_4_3
void FASTCALL call_4_8_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_4_4
#define init_4_8_4_4
void FASTCALL call_4_8_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_4_5
#define init_4_8_4_5
void FASTCALL call_4_8_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_4_6
#define init_4_8_4_6
void FASTCALL call_4_8_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_4_7
#define init_4_8_4_7
void FASTCALL call_4_8_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_4_8
#define init_4_8_4_8
void FASTCALL call_4_8_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_4_9
#define init_4_8_4_9
void FASTCALL call_4_8_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_4_10
#define init_4_8_4_10
void FASTCALL call_4_8_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_4_11
#define init_4_8_4_11
void FASTCALL call_4_8_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_4_12
#define init_4_8_4_12
void FASTCALL call_4_8_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_4_13
#define init_4_8_4_13
void FASTCALL call_4_8_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_4_14
#define init_4_8_4_14
void FASTCALL call_4_8_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_4_15
#define init_4_8_4_15
void FASTCALL call_4_8_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_5_0
#define init_4_8_5_0
void FASTCALL call_4_8_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_5_1
#define init_4_8_5_1
void FASTCALL call_4_8_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_5_2
#define init_4_8_5_2
void FASTCALL call_4_8_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_5_3
#define init_4_8_5_3
void FASTCALL call_4_8_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_5_4
#define init_4_8_5_4
void FASTCALL call_4_8_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_5_5
#define init_4_8_5_5
void FASTCALL call_4_8_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_5_6
#define init_4_8_5_6
void FASTCALL call_4_8_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_5_7
#define init_4_8_5_7
void FASTCALL call_4_8_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_5_8
#define init_4_8_5_8
void FASTCALL call_4_8_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_5_9
#define init_4_8_5_9
void FASTCALL call_4_8_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_5_10
#define init_4_8_5_10
void FASTCALL call_4_8_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_5_11
#define init_4_8_5_11
void FASTCALL call_4_8_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_5_12
#define init_4_8_5_12
void FASTCALL call_4_8_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_5_13
#define init_4_8_5_13
void FASTCALL call_4_8_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_5_14
#define init_4_8_5_14
void FASTCALL call_4_8_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_5_15
#define init_4_8_5_15
void FASTCALL call_4_8_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_6_0
#define init_4_8_6_0
void FASTCALL call_4_8_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_6_1
#define init_4_8_6_1
void FASTCALL call_4_8_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_6_2
#define init_4_8_6_2
void FASTCALL call_4_8_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_6_3
#define init_4_8_6_3
void FASTCALL call_4_8_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_6_4
#define init_4_8_6_4
void FASTCALL call_4_8_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_6_5
#define init_4_8_6_5
void FASTCALL call_4_8_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_6_6
#define init_4_8_6_6
void FASTCALL call_4_8_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_6_7
#define init_4_8_6_7
void FASTCALL call_4_8_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_6_8
#define init_4_8_6_8
void FASTCALL call_4_8_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_6_9
#define init_4_8_6_9
void FASTCALL call_4_8_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_6_10
#define init_4_8_6_10
void FASTCALL call_4_8_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_6_11
#define init_4_8_6_11
void FASTCALL call_4_8_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_6_12
#define init_4_8_6_12
void FASTCALL call_4_8_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_6_13
#define init_4_8_6_13
void FASTCALL call_4_8_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_6_14
#define init_4_8_6_14
void FASTCALL call_4_8_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_6_15
#define init_4_8_6_15
void FASTCALL call_4_8_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_7_0
#define init_4_8_7_0
void FASTCALL call_4_8_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_7_1
#define init_4_8_7_1
void FASTCALL call_4_8_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_7_2
#define init_4_8_7_2
void FASTCALL call_4_8_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_7_3
#define init_4_8_7_3
void FASTCALL call_4_8_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_7_4
#define init_4_8_7_4
void FASTCALL call_4_8_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_7_5
#define init_4_8_7_5
void FASTCALL call_4_8_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_7_6
#define init_4_8_7_6
void FASTCALL call_4_8_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_7_7
#define init_4_8_7_7
void FASTCALL call_4_8_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_7_8
#define init_4_8_7_8
void FASTCALL call_4_8_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_7_9
#define init_4_8_7_9
void FASTCALL call_4_8_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_7_10
#define init_4_8_7_10
void FASTCALL call_4_8_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_7_11
#define init_4_8_7_11
void FASTCALL call_4_8_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_7_12
#define init_4_8_7_12
void FASTCALL call_4_8_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_7_13
#define init_4_8_7_13
void FASTCALL call_4_8_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_7_14
#define init_4_8_7_14
void FASTCALL call_4_8_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_7_15
#define init_4_8_7_15
void FASTCALL call_4_8_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_8_0
#define init_4_8_8_0
void FASTCALL call_4_8_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_8_1
#define init_4_8_8_1
void FASTCALL call_4_8_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_8_2
#define init_4_8_8_2
void FASTCALL call_4_8_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_8_3
#define init_4_8_8_3
void FASTCALL call_4_8_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_8_4
#define init_4_8_8_4
void FASTCALL call_4_8_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_8_5
#define init_4_8_8_5
void FASTCALL call_4_8_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_8_6
#define init_4_8_8_6
void FASTCALL call_4_8_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_8_7
#define init_4_8_8_7
void FASTCALL call_4_8_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_8_8
#define init_4_8_8_8
void FASTCALL call_4_8_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_8_9
#define init_4_8_8_9
void FASTCALL call_4_8_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_8_10
#define init_4_8_8_10
void FASTCALL call_4_8_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_8_11
#define init_4_8_8_11
void FASTCALL call_4_8_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_8_12
#define init_4_8_8_12
void FASTCALL call_4_8_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_8_13
#define init_4_8_8_13
void FASTCALL call_4_8_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_8_14
#define init_4_8_8_14
void FASTCALL call_4_8_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_8_15
#define init_4_8_8_15
void FASTCALL call_4_8_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_9_0
#define init_4_8_9_0
void FASTCALL call_4_8_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_9_1
#define init_4_8_9_1
void FASTCALL call_4_8_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_9_2
#define init_4_8_9_2
void FASTCALL call_4_8_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_9_3
#define init_4_8_9_3
void FASTCALL call_4_8_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_9_4
#define init_4_8_9_4
void FASTCALL call_4_8_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_9_5
#define init_4_8_9_5
void FASTCALL call_4_8_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_9_6
#define init_4_8_9_6
void FASTCALL call_4_8_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_9_7
#define init_4_8_9_7
void FASTCALL call_4_8_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_9_8
#define init_4_8_9_8
void FASTCALL call_4_8_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_9_9
#define init_4_8_9_9
void FASTCALL call_4_8_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_9_10
#define init_4_8_9_10
void FASTCALL call_4_8_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_9_11
#define init_4_8_9_11
void FASTCALL call_4_8_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_9_12
#define init_4_8_9_12
void FASTCALL call_4_8_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_9_13
#define init_4_8_9_13
void FASTCALL call_4_8_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_9_14
#define init_4_8_9_14
void FASTCALL call_4_8_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_9_15
#define init_4_8_9_15
void FASTCALL call_4_8_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_10_0
#define init_4_8_10_0
void FASTCALL call_4_8_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_10_1
#define init_4_8_10_1
void FASTCALL call_4_8_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_10_2
#define init_4_8_10_2
void FASTCALL call_4_8_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_10_3
#define init_4_8_10_3
void FASTCALL call_4_8_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_10_4
#define init_4_8_10_4
void FASTCALL call_4_8_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_10_5
#define init_4_8_10_5
void FASTCALL call_4_8_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_10_6
#define init_4_8_10_6
void FASTCALL call_4_8_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_10_7
#define init_4_8_10_7
void FASTCALL call_4_8_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_10_8
#define init_4_8_10_8
void FASTCALL call_4_8_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_10_9
#define init_4_8_10_9
void FASTCALL call_4_8_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_10_10
#define init_4_8_10_10
void FASTCALL call_4_8_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_10_11
#define init_4_8_10_11
void FASTCALL call_4_8_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_10_12
#define init_4_8_10_12
void FASTCALL call_4_8_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_10_13
#define init_4_8_10_13
void FASTCALL call_4_8_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_10_14
#define init_4_8_10_14
void FASTCALL call_4_8_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_10_15
#define init_4_8_10_15
void FASTCALL call_4_8_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_11_0
#define init_4_8_11_0
void FASTCALL call_4_8_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_11_1
#define init_4_8_11_1
void FASTCALL call_4_8_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_11_2
#define init_4_8_11_2
void FASTCALL call_4_8_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_11_3
#define init_4_8_11_3
void FASTCALL call_4_8_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_11_4
#define init_4_8_11_4
void FASTCALL call_4_8_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_11_5
#define init_4_8_11_5
void FASTCALL call_4_8_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_11_6
#define init_4_8_11_6
void FASTCALL call_4_8_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_11_7
#define init_4_8_11_7
void FASTCALL call_4_8_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_11_8
#define init_4_8_11_8
void FASTCALL call_4_8_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_11_9
#define init_4_8_11_9
void FASTCALL call_4_8_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_11_10
#define init_4_8_11_10
void FASTCALL call_4_8_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_11_11
#define init_4_8_11_11
void FASTCALL call_4_8_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_11_12
#define init_4_8_11_12
void FASTCALL call_4_8_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_11_13
#define init_4_8_11_13
void FASTCALL call_4_8_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_11_14
#define init_4_8_11_14
void FASTCALL call_4_8_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_11_15
#define init_4_8_11_15
void FASTCALL call_4_8_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_12_0
#define init_4_8_12_0
void FASTCALL call_4_8_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_12_1
#define init_4_8_12_1
void FASTCALL call_4_8_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_12_2
#define init_4_8_12_2
void FASTCALL call_4_8_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_12_3
#define init_4_8_12_3
void FASTCALL call_4_8_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_12_4
#define init_4_8_12_4
void FASTCALL call_4_8_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_12_5
#define init_4_8_12_5
void FASTCALL call_4_8_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_12_6
#define init_4_8_12_6
void FASTCALL call_4_8_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_12_7
#define init_4_8_12_7
void FASTCALL call_4_8_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_12_8
#define init_4_8_12_8
void FASTCALL call_4_8_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_12_9
#define init_4_8_12_9
void FASTCALL call_4_8_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_12_10
#define init_4_8_12_10
void FASTCALL call_4_8_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_12_11
#define init_4_8_12_11
void FASTCALL call_4_8_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_12_12
#define init_4_8_12_12
void FASTCALL call_4_8_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_12_13
#define init_4_8_12_13
void FASTCALL call_4_8_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_12_14
#define init_4_8_12_14
void FASTCALL call_4_8_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_12_15
#define init_4_8_12_15
void FASTCALL call_4_8_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_13_0
#define init_4_8_13_0
void FASTCALL call_4_8_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_13_1
#define init_4_8_13_1
void FASTCALL call_4_8_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_13_2
#define init_4_8_13_2
void FASTCALL call_4_8_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_13_3
#define init_4_8_13_3
void FASTCALL call_4_8_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_13_4
#define init_4_8_13_4
void FASTCALL call_4_8_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_13_5
#define init_4_8_13_5
void FASTCALL call_4_8_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_13_6
#define init_4_8_13_6
void FASTCALL call_4_8_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_13_7
#define init_4_8_13_7
void FASTCALL call_4_8_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_13_8
#define init_4_8_13_8
void FASTCALL call_4_8_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_13_9
#define init_4_8_13_9
void FASTCALL call_4_8_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_13_10
#define init_4_8_13_10
void FASTCALL call_4_8_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_13_11
#define init_4_8_13_11
void FASTCALL call_4_8_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_13_12
#define init_4_8_13_12
void FASTCALL call_4_8_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_13_13
#define init_4_8_13_13
void FASTCALL call_4_8_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_13_14
#define init_4_8_13_14
void FASTCALL call_4_8_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_13_15
#define init_4_8_13_15
void FASTCALL call_4_8_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_14_0
#define init_4_8_14_0
void FASTCALL call_4_8_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_14_1
#define init_4_8_14_1
void FASTCALL call_4_8_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_14_2
#define init_4_8_14_2
void FASTCALL call_4_8_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_14_3
#define init_4_8_14_3
void FASTCALL call_4_8_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_14_4
#define init_4_8_14_4
void FASTCALL call_4_8_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_14_5
#define init_4_8_14_5
void FASTCALL call_4_8_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_14_6
#define init_4_8_14_6
void FASTCALL call_4_8_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_14_7
#define init_4_8_14_7
void FASTCALL call_4_8_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_14_8
#define init_4_8_14_8
void FASTCALL call_4_8_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_14_9
#define init_4_8_14_9
void FASTCALL call_4_8_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_14_10
#define init_4_8_14_10
void FASTCALL call_4_8_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_14_11
#define init_4_8_14_11
void FASTCALL call_4_8_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_14_12
#define init_4_8_14_12
void FASTCALL call_4_8_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_14_13
#define init_4_8_14_13
void FASTCALL call_4_8_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_14_14
#define init_4_8_14_14
void FASTCALL call_4_8_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_14_15
#define init_4_8_14_15
void FASTCALL call_4_8_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_15_0
#define init_4_8_15_0
void FASTCALL call_4_8_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_15_1
#define init_4_8_15_1
void FASTCALL call_4_8_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_15_2
#define init_4_8_15_2
void FASTCALL call_4_8_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_15_3
#define init_4_8_15_3
void FASTCALL call_4_8_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_15_4
#define init_4_8_15_4
void FASTCALL call_4_8_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_15_5
#define init_4_8_15_5
void FASTCALL call_4_8_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_15_6
#define init_4_8_15_6
void FASTCALL call_4_8_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_15_7
#define init_4_8_15_7
void FASTCALL call_4_8_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_15_8
#define init_4_8_15_8
void FASTCALL call_4_8_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_15_9
#define init_4_8_15_9
void FASTCALL call_4_8_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_15_10
#define init_4_8_15_10
void FASTCALL call_4_8_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_15_11
#define init_4_8_15_11
void FASTCALL call_4_8_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_15_12
#define init_4_8_15_12
void FASTCALL call_4_8_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_15_13
#define init_4_8_15_13
void FASTCALL call_4_8_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_15_14
#define init_4_8_15_14
void FASTCALL call_4_8_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_8_15_15
#define init_4_8_15_15
void FASTCALL call_4_8_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_0_0
#define init_4_9_0_0
void FASTCALL call_4_9_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_0_1
#define init_4_9_0_1
void FASTCALL call_4_9_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_0_2
#define init_4_9_0_2
void FASTCALL call_4_9_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_0_3
#define init_4_9_0_3
void FASTCALL call_4_9_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_0_4
#define init_4_9_0_4
void FASTCALL call_4_9_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_0_5
#define init_4_9_0_5
void FASTCALL call_4_9_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_0_6
#define init_4_9_0_6
void FASTCALL call_4_9_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_0_7
#define init_4_9_0_7
void FASTCALL call_4_9_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_0_8
#define init_4_9_0_8
void FASTCALL call_4_9_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_0_9
#define init_4_9_0_9
void FASTCALL call_4_9_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_0_10
#define init_4_9_0_10
void FASTCALL call_4_9_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_0_11
#define init_4_9_0_11
void FASTCALL call_4_9_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_0_12
#define init_4_9_0_12
void FASTCALL call_4_9_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_0_13
#define init_4_9_0_13
void FASTCALL call_4_9_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_0_14
#define init_4_9_0_14
void FASTCALL call_4_9_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_0_15
#define init_4_9_0_15
void FASTCALL call_4_9_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_1_0
#define init_4_9_1_0
void FASTCALL call_4_9_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_1_1
#define init_4_9_1_1
void FASTCALL call_4_9_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_1_2
#define init_4_9_1_2
void FASTCALL call_4_9_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_1_3
#define init_4_9_1_3
void FASTCALL call_4_9_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_1_4
#define init_4_9_1_4
void FASTCALL call_4_9_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_1_5
#define init_4_9_1_5
void FASTCALL call_4_9_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_1_6
#define init_4_9_1_6
void FASTCALL call_4_9_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_1_7
#define init_4_9_1_7
void FASTCALL call_4_9_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_1_8
#define init_4_9_1_8
void FASTCALL call_4_9_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_1_9
#define init_4_9_1_9
void FASTCALL call_4_9_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_1_10
#define init_4_9_1_10
void FASTCALL call_4_9_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_1_11
#define init_4_9_1_11
void FASTCALL call_4_9_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_1_12
#define init_4_9_1_12
void FASTCALL call_4_9_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_1_13
#define init_4_9_1_13
void FASTCALL call_4_9_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_1_14
#define init_4_9_1_14
void FASTCALL call_4_9_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_1_15
#define init_4_9_1_15
void FASTCALL call_4_9_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_2_0
#define init_4_9_2_0
void FASTCALL call_4_9_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_2_1
#define init_4_9_2_1
void FASTCALL call_4_9_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_2_2
#define init_4_9_2_2
void FASTCALL call_4_9_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_2_3
#define init_4_9_2_3
void FASTCALL call_4_9_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_2_4
#define init_4_9_2_4
void FASTCALL call_4_9_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_2_5
#define init_4_9_2_5
void FASTCALL call_4_9_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_2_6
#define init_4_9_2_6
void FASTCALL call_4_9_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_2_7
#define init_4_9_2_7
void FASTCALL call_4_9_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_2_8
#define init_4_9_2_8
void FASTCALL call_4_9_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_2_9
#define init_4_9_2_9
void FASTCALL call_4_9_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_2_10
#define init_4_9_2_10
void FASTCALL call_4_9_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_2_11
#define init_4_9_2_11
void FASTCALL call_4_9_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_2_12
#define init_4_9_2_12
void FASTCALL call_4_9_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_2_13
#define init_4_9_2_13
void FASTCALL call_4_9_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_2_14
#define init_4_9_2_14
void FASTCALL call_4_9_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_2_15
#define init_4_9_2_15
void FASTCALL call_4_9_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_3_0
#define init_4_9_3_0
void FASTCALL call_4_9_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_3_1
#define init_4_9_3_1
void FASTCALL call_4_9_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_3_2
#define init_4_9_3_2
void FASTCALL call_4_9_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_3_3
#define init_4_9_3_3
void FASTCALL call_4_9_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_3_4
#define init_4_9_3_4
void FASTCALL call_4_9_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_3_5
#define init_4_9_3_5
void FASTCALL call_4_9_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_3_6
#define init_4_9_3_6
void FASTCALL call_4_9_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_3_7
#define init_4_9_3_7
void FASTCALL call_4_9_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_3_8
#define init_4_9_3_8
void FASTCALL call_4_9_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_3_9
#define init_4_9_3_9
void FASTCALL call_4_9_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_3_10
#define init_4_9_3_10
void FASTCALL call_4_9_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_3_11
#define init_4_9_3_11
void FASTCALL call_4_9_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_3_12
#define init_4_9_3_12
void FASTCALL call_4_9_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_3_13
#define init_4_9_3_13
void FASTCALL call_4_9_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_3_14
#define init_4_9_3_14
void FASTCALL call_4_9_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_3_15
#define init_4_9_3_15
void FASTCALL call_4_9_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_4_0
#define init_4_9_4_0
void FASTCALL call_4_9_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_4_1
#define init_4_9_4_1
void FASTCALL call_4_9_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_4_2
#define init_4_9_4_2
void FASTCALL call_4_9_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_4_3
#define init_4_9_4_3
void FASTCALL call_4_9_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_4_4
#define init_4_9_4_4
void FASTCALL call_4_9_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_4_5
#define init_4_9_4_5
void FASTCALL call_4_9_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_4_6
#define init_4_9_4_6
void FASTCALL call_4_9_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_4_7
#define init_4_9_4_7
void FASTCALL call_4_9_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_4_8
#define init_4_9_4_8
void FASTCALL call_4_9_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_4_9
#define init_4_9_4_9
void FASTCALL call_4_9_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_4_10
#define init_4_9_4_10
void FASTCALL call_4_9_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_4_11
#define init_4_9_4_11
void FASTCALL call_4_9_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_4_12
#define init_4_9_4_12
void FASTCALL call_4_9_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_4_13
#define init_4_9_4_13
void FASTCALL call_4_9_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_4_14
#define init_4_9_4_14
void FASTCALL call_4_9_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_4_15
#define init_4_9_4_15
void FASTCALL call_4_9_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_5_0
#define init_4_9_5_0
void FASTCALL call_4_9_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_5_1
#define init_4_9_5_1
void FASTCALL call_4_9_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_5_2
#define init_4_9_5_2
void FASTCALL call_4_9_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_5_3
#define init_4_9_5_3
void FASTCALL call_4_9_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_5_4
#define init_4_9_5_4
void FASTCALL call_4_9_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_5_5
#define init_4_9_5_5
void FASTCALL call_4_9_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_5_6
#define init_4_9_5_6
void FASTCALL call_4_9_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_5_7
#define init_4_9_5_7
void FASTCALL call_4_9_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_5_8
#define init_4_9_5_8
void FASTCALL call_4_9_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_5_9
#define init_4_9_5_9
void FASTCALL call_4_9_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_5_10
#define init_4_9_5_10
void FASTCALL call_4_9_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_5_11
#define init_4_9_5_11
void FASTCALL call_4_9_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_5_12
#define init_4_9_5_12
void FASTCALL call_4_9_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_5_13
#define init_4_9_5_13
void FASTCALL call_4_9_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_5_14
#define init_4_9_5_14
void FASTCALL call_4_9_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_5_15
#define init_4_9_5_15
void FASTCALL call_4_9_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_6_0
#define init_4_9_6_0
void FASTCALL call_4_9_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_6_1
#define init_4_9_6_1
void FASTCALL call_4_9_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_6_2
#define init_4_9_6_2
void FASTCALL call_4_9_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_6_3
#define init_4_9_6_3
void FASTCALL call_4_9_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_6_4
#define init_4_9_6_4
void FASTCALL call_4_9_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_6_5
#define init_4_9_6_5
void FASTCALL call_4_9_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_6_6
#define init_4_9_6_6
void FASTCALL call_4_9_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_6_7
#define init_4_9_6_7
void FASTCALL call_4_9_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_6_8
#define init_4_9_6_8
void FASTCALL call_4_9_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_6_9
#define init_4_9_6_9
void FASTCALL call_4_9_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_6_10
#define init_4_9_6_10
void FASTCALL call_4_9_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_6_11
#define init_4_9_6_11
void FASTCALL call_4_9_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_6_12
#define init_4_9_6_12
void FASTCALL call_4_9_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_6_13
#define init_4_9_6_13
void FASTCALL call_4_9_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_6_14
#define init_4_9_6_14
void FASTCALL call_4_9_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_6_15
#define init_4_9_6_15
void FASTCALL call_4_9_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_7_0
#define init_4_9_7_0
void FASTCALL call_4_9_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_7_1
#define init_4_9_7_1
void FASTCALL call_4_9_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_7_2
#define init_4_9_7_2
void FASTCALL call_4_9_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_7_3
#define init_4_9_7_3
void FASTCALL call_4_9_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_7_4
#define init_4_9_7_4
void FASTCALL call_4_9_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_7_5
#define init_4_9_7_5
void FASTCALL call_4_9_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_7_6
#define init_4_9_7_6
void FASTCALL call_4_9_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_7_7
#define init_4_9_7_7
void FASTCALL call_4_9_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_7_8
#define init_4_9_7_8
void FASTCALL call_4_9_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_7_9
#define init_4_9_7_9
void FASTCALL call_4_9_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_7_10
#define init_4_9_7_10
void FASTCALL call_4_9_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_7_11
#define init_4_9_7_11
void FASTCALL call_4_9_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_7_12
#define init_4_9_7_12
void FASTCALL call_4_9_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_7_13
#define init_4_9_7_13
void FASTCALL call_4_9_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_7_14
#define init_4_9_7_14
void FASTCALL call_4_9_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_7_15
#define init_4_9_7_15
void FASTCALL call_4_9_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_8_0
#define init_4_9_8_0
void FASTCALL call_4_9_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_8_1
#define init_4_9_8_1
void FASTCALL call_4_9_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_8_2
#define init_4_9_8_2
void FASTCALL call_4_9_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_8_3
#define init_4_9_8_3
void FASTCALL call_4_9_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_8_4
#define init_4_9_8_4
void FASTCALL call_4_9_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_8_5
#define init_4_9_8_5
void FASTCALL call_4_9_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_8_6
#define init_4_9_8_6
void FASTCALL call_4_9_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_8_7
#define init_4_9_8_7
void FASTCALL call_4_9_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_8_8
#define init_4_9_8_8
void FASTCALL call_4_9_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_8_9
#define init_4_9_8_9
void FASTCALL call_4_9_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_8_10
#define init_4_9_8_10
void FASTCALL call_4_9_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_8_11
#define init_4_9_8_11
void FASTCALL call_4_9_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_8_12
#define init_4_9_8_12
void FASTCALL call_4_9_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_8_13
#define init_4_9_8_13
void FASTCALL call_4_9_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_8_14
#define init_4_9_8_14
void FASTCALL call_4_9_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_8_15
#define init_4_9_8_15
void FASTCALL call_4_9_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_9_0
#define init_4_9_9_0
void FASTCALL call_4_9_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_9_1
#define init_4_9_9_1
void FASTCALL call_4_9_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_9_2
#define init_4_9_9_2
void FASTCALL call_4_9_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_9_3
#define init_4_9_9_3
void FASTCALL call_4_9_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_9_4
#define init_4_9_9_4
void FASTCALL call_4_9_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_9_5
#define init_4_9_9_5
void FASTCALL call_4_9_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_9_6
#define init_4_9_9_6
void FASTCALL call_4_9_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_9_7
#define init_4_9_9_7
void FASTCALL call_4_9_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_9_8
#define init_4_9_9_8
void FASTCALL call_4_9_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_9_9
#define init_4_9_9_9
void FASTCALL call_4_9_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_9_10
#define init_4_9_9_10
void FASTCALL call_4_9_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_9_11
#define init_4_9_9_11
void FASTCALL call_4_9_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_9_12
#define init_4_9_9_12
void FASTCALL call_4_9_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_9_13
#define init_4_9_9_13
void FASTCALL call_4_9_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_9_14
#define init_4_9_9_14
void FASTCALL call_4_9_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_9_15
#define init_4_9_9_15
void FASTCALL call_4_9_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_10_0
#define init_4_9_10_0
void FASTCALL call_4_9_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_10_1
#define init_4_9_10_1
void FASTCALL call_4_9_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_10_2
#define init_4_9_10_2
void FASTCALL call_4_9_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_10_3
#define init_4_9_10_3
void FASTCALL call_4_9_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_10_4
#define init_4_9_10_4
void FASTCALL call_4_9_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_10_5
#define init_4_9_10_5
void FASTCALL call_4_9_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_10_6
#define init_4_9_10_6
void FASTCALL call_4_9_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_10_7
#define init_4_9_10_7
void FASTCALL call_4_9_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_10_8
#define init_4_9_10_8
void FASTCALL call_4_9_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_10_9
#define init_4_9_10_9
void FASTCALL call_4_9_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_10_10
#define init_4_9_10_10
void FASTCALL call_4_9_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_10_11
#define init_4_9_10_11
void FASTCALL call_4_9_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_10_12
#define init_4_9_10_12
void FASTCALL call_4_9_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_10_13
#define init_4_9_10_13
void FASTCALL call_4_9_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_10_14
#define init_4_9_10_14
void FASTCALL call_4_9_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_10_15
#define init_4_9_10_15
void FASTCALL call_4_9_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_11_0
#define init_4_9_11_0
void FASTCALL call_4_9_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_11_1
#define init_4_9_11_1
void FASTCALL call_4_9_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_11_2
#define init_4_9_11_2
void FASTCALL call_4_9_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_11_3
#define init_4_9_11_3
void FASTCALL call_4_9_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_11_4
#define init_4_9_11_4
void FASTCALL call_4_9_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_11_5
#define init_4_9_11_5
void FASTCALL call_4_9_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_11_6
#define init_4_9_11_6
void FASTCALL call_4_9_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_11_7
#define init_4_9_11_7
void FASTCALL call_4_9_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_11_8
#define init_4_9_11_8
void FASTCALL call_4_9_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_11_9
#define init_4_9_11_9
void FASTCALL call_4_9_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_11_10
#define init_4_9_11_10
void FASTCALL call_4_9_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_11_11
#define init_4_9_11_11
void FASTCALL call_4_9_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_11_12
#define init_4_9_11_12
void FASTCALL call_4_9_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_11_13
#define init_4_9_11_13
void FASTCALL call_4_9_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_11_14
#define init_4_9_11_14
void FASTCALL call_4_9_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_11_15
#define init_4_9_11_15
void FASTCALL call_4_9_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_12_0
#define init_4_9_12_0
void FASTCALL call_4_9_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_12_1
#define init_4_9_12_1
void FASTCALL call_4_9_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_12_2
#define init_4_9_12_2
void FASTCALL call_4_9_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_12_3
#define init_4_9_12_3
void FASTCALL call_4_9_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_12_4
#define init_4_9_12_4
void FASTCALL call_4_9_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_12_5
#define init_4_9_12_5
void FASTCALL call_4_9_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_12_6
#define init_4_9_12_6
void FASTCALL call_4_9_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_12_7
#define init_4_9_12_7
void FASTCALL call_4_9_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_12_8
#define init_4_9_12_8
void FASTCALL call_4_9_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_12_9
#define init_4_9_12_9
void FASTCALL call_4_9_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_12_10
#define init_4_9_12_10
void FASTCALL call_4_9_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_12_11
#define init_4_9_12_11
void FASTCALL call_4_9_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_12_12
#define init_4_9_12_12
void FASTCALL call_4_9_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_12_13
#define init_4_9_12_13
void FASTCALL call_4_9_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_12_14
#define init_4_9_12_14
void FASTCALL call_4_9_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_12_15
#define init_4_9_12_15
void FASTCALL call_4_9_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_13_0
#define init_4_9_13_0
void FASTCALL call_4_9_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_13_1
#define init_4_9_13_1
void FASTCALL call_4_9_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_13_2
#define init_4_9_13_2
void FASTCALL call_4_9_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_13_3
#define init_4_9_13_3
void FASTCALL call_4_9_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_13_4
#define init_4_9_13_4
void FASTCALL call_4_9_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_13_5
#define init_4_9_13_5
void FASTCALL call_4_9_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_13_6
#define init_4_9_13_6
void FASTCALL call_4_9_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_13_7
#define init_4_9_13_7
void FASTCALL call_4_9_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_13_8
#define init_4_9_13_8
void FASTCALL call_4_9_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_13_9
#define init_4_9_13_9
void FASTCALL call_4_9_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_13_10
#define init_4_9_13_10
void FASTCALL call_4_9_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_13_11
#define init_4_9_13_11
void FASTCALL call_4_9_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_13_12
#define init_4_9_13_12
void FASTCALL call_4_9_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_13_13
#define init_4_9_13_13
void FASTCALL call_4_9_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_13_14
#define init_4_9_13_14
void FASTCALL call_4_9_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_13_15
#define init_4_9_13_15
void FASTCALL call_4_9_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_14_0
#define init_4_9_14_0
void FASTCALL call_4_9_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_14_1
#define init_4_9_14_1
void FASTCALL call_4_9_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_14_2
#define init_4_9_14_2
void FASTCALL call_4_9_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_14_3
#define init_4_9_14_3
void FASTCALL call_4_9_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_14_4
#define init_4_9_14_4
void FASTCALL call_4_9_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_14_5
#define init_4_9_14_5
void FASTCALL call_4_9_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_14_6
#define init_4_9_14_6
void FASTCALL call_4_9_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_14_7
#define init_4_9_14_7
void FASTCALL call_4_9_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_14_8
#define init_4_9_14_8
void FASTCALL call_4_9_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_14_9
#define init_4_9_14_9
void FASTCALL call_4_9_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_14_10
#define init_4_9_14_10
void FASTCALL call_4_9_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_14_11
#define init_4_9_14_11
void FASTCALL call_4_9_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_14_12
#define init_4_9_14_12
void FASTCALL call_4_9_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_14_13
#define init_4_9_14_13
void FASTCALL call_4_9_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_14_14
#define init_4_9_14_14
void FASTCALL call_4_9_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_14_15
#define init_4_9_14_15
void FASTCALL call_4_9_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_15_0
#define init_4_9_15_0
void FASTCALL call_4_9_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_15_1
#define init_4_9_15_1
void FASTCALL call_4_9_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_15_2
#define init_4_9_15_2
void FASTCALL call_4_9_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_15_3
#define init_4_9_15_3
void FASTCALL call_4_9_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_15_4
#define init_4_9_15_4
void FASTCALL call_4_9_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_15_5
#define init_4_9_15_5
void FASTCALL call_4_9_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_15_6
#define init_4_9_15_6
void FASTCALL call_4_9_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_15_7
#define init_4_9_15_7
void FASTCALL call_4_9_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_15_8
#define init_4_9_15_8
void FASTCALL call_4_9_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_15_9
#define init_4_9_15_9
void FASTCALL call_4_9_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_15_10
#define init_4_9_15_10
void FASTCALL call_4_9_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_15_11
#define init_4_9_15_11
void FASTCALL call_4_9_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_15_12
#define init_4_9_15_12
void FASTCALL call_4_9_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_15_13
#define init_4_9_15_13
void FASTCALL call_4_9_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_15_14
#define init_4_9_15_14
void FASTCALL call_4_9_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_9_15_15
#define init_4_9_15_15
void FASTCALL call_4_9_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_0_0
#define init_4_10_0_0
void FASTCALL call_4_10_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_0_1
#define init_4_10_0_1
void FASTCALL call_4_10_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_0_2
#define init_4_10_0_2
void FASTCALL call_4_10_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_0_3
#define init_4_10_0_3
void FASTCALL call_4_10_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_0_4
#define init_4_10_0_4
void FASTCALL call_4_10_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_0_5
#define init_4_10_0_5
void FASTCALL call_4_10_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_0_6
#define init_4_10_0_6
void FASTCALL call_4_10_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_0_7
#define init_4_10_0_7
void FASTCALL call_4_10_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_0_8
#define init_4_10_0_8
void FASTCALL call_4_10_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_0_9
#define init_4_10_0_9
void FASTCALL call_4_10_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_0_10
#define init_4_10_0_10
void FASTCALL call_4_10_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_0_11
#define init_4_10_0_11
void FASTCALL call_4_10_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_0_12
#define init_4_10_0_12
void FASTCALL call_4_10_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_0_13
#define init_4_10_0_13
void FASTCALL call_4_10_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_0_14
#define init_4_10_0_14
void FASTCALL call_4_10_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_0_15
#define init_4_10_0_15
void FASTCALL call_4_10_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_1_0
#define init_4_10_1_0
void FASTCALL call_4_10_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_1_1
#define init_4_10_1_1
void FASTCALL call_4_10_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_1_2
#define init_4_10_1_2
void FASTCALL call_4_10_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_1_3
#define init_4_10_1_3
void FASTCALL call_4_10_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_1_4
#define init_4_10_1_4
void FASTCALL call_4_10_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_1_5
#define init_4_10_1_5
void FASTCALL call_4_10_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_1_6
#define init_4_10_1_6
void FASTCALL call_4_10_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_1_7
#define init_4_10_1_7
void FASTCALL call_4_10_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_1_8
#define init_4_10_1_8
void FASTCALL call_4_10_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_1_9
#define init_4_10_1_9
void FASTCALL call_4_10_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_1_10
#define init_4_10_1_10
void FASTCALL call_4_10_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_1_11
#define init_4_10_1_11
void FASTCALL call_4_10_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_1_12
#define init_4_10_1_12
void FASTCALL call_4_10_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_1_13
#define init_4_10_1_13
void FASTCALL call_4_10_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_1_14
#define init_4_10_1_14
void FASTCALL call_4_10_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_1_15
#define init_4_10_1_15
void FASTCALL call_4_10_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_2_0
#define init_4_10_2_0
void FASTCALL call_4_10_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_2_1
#define init_4_10_2_1
void FASTCALL call_4_10_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_2_2
#define init_4_10_2_2
void FASTCALL call_4_10_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_2_3
#define init_4_10_2_3
void FASTCALL call_4_10_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_2_4
#define init_4_10_2_4
void FASTCALL call_4_10_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_2_5
#define init_4_10_2_5
void FASTCALL call_4_10_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_2_6
#define init_4_10_2_6
void FASTCALL call_4_10_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_2_7
#define init_4_10_2_7
void FASTCALL call_4_10_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_2_8
#define init_4_10_2_8
void FASTCALL call_4_10_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_2_9
#define init_4_10_2_9
void FASTCALL call_4_10_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_2_10
#define init_4_10_2_10
void FASTCALL call_4_10_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_2_11
#define init_4_10_2_11
void FASTCALL call_4_10_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_2_12
#define init_4_10_2_12
void FASTCALL call_4_10_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_2_13
#define init_4_10_2_13
void FASTCALL call_4_10_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_2_14
#define init_4_10_2_14
void FASTCALL call_4_10_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_2_15
#define init_4_10_2_15
void FASTCALL call_4_10_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_3_0
#define init_4_10_3_0
void FASTCALL call_4_10_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_3_1
#define init_4_10_3_1
void FASTCALL call_4_10_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_3_2
#define init_4_10_3_2
void FASTCALL call_4_10_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_3_3
#define init_4_10_3_3
void FASTCALL call_4_10_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_3_4
#define init_4_10_3_4
void FASTCALL call_4_10_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_3_5
#define init_4_10_3_5
void FASTCALL call_4_10_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_3_6
#define init_4_10_3_6
void FASTCALL call_4_10_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_3_7
#define init_4_10_3_7
void FASTCALL call_4_10_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_3_8
#define init_4_10_3_8
void FASTCALL call_4_10_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_3_9
#define init_4_10_3_9
void FASTCALL call_4_10_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_3_10
#define init_4_10_3_10
void FASTCALL call_4_10_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_3_11
#define init_4_10_3_11
void FASTCALL call_4_10_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_3_12
#define init_4_10_3_12
void FASTCALL call_4_10_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_3_13
#define init_4_10_3_13
void FASTCALL call_4_10_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_3_14
#define init_4_10_3_14
void FASTCALL call_4_10_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_3_15
#define init_4_10_3_15
void FASTCALL call_4_10_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_4_0
#define init_4_10_4_0
void FASTCALL call_4_10_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_4_1
#define init_4_10_4_1
void FASTCALL call_4_10_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_4_2
#define init_4_10_4_2
void FASTCALL call_4_10_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_4_3
#define init_4_10_4_3
void FASTCALL call_4_10_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_4_4
#define init_4_10_4_4
void FASTCALL call_4_10_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_4_5
#define init_4_10_4_5
void FASTCALL call_4_10_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_4_6
#define init_4_10_4_6
void FASTCALL call_4_10_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_4_7
#define init_4_10_4_7
void FASTCALL call_4_10_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_4_8
#define init_4_10_4_8
void FASTCALL call_4_10_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_4_9
#define init_4_10_4_9
void FASTCALL call_4_10_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_4_10
#define init_4_10_4_10
void FASTCALL call_4_10_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_4_11
#define init_4_10_4_11
void FASTCALL call_4_10_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_4_12
#define init_4_10_4_12
void FASTCALL call_4_10_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_4_13
#define init_4_10_4_13
void FASTCALL call_4_10_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_4_14
#define init_4_10_4_14
void FASTCALL call_4_10_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_4_15
#define init_4_10_4_15
void FASTCALL call_4_10_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_5_0
#define init_4_10_5_0
void FASTCALL call_4_10_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_5_1
#define init_4_10_5_1
void FASTCALL call_4_10_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_5_2
#define init_4_10_5_2
void FASTCALL call_4_10_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_5_3
#define init_4_10_5_3
void FASTCALL call_4_10_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_5_4
#define init_4_10_5_4
void FASTCALL call_4_10_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_5_5
#define init_4_10_5_5
void FASTCALL call_4_10_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_5_6
#define init_4_10_5_6
void FASTCALL call_4_10_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_5_7
#define init_4_10_5_7
void FASTCALL call_4_10_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_5_8
#define init_4_10_5_8
void FASTCALL call_4_10_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_5_9
#define init_4_10_5_9
void FASTCALL call_4_10_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_5_10
#define init_4_10_5_10
void FASTCALL call_4_10_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_5_11
#define init_4_10_5_11
void FASTCALL call_4_10_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_5_12
#define init_4_10_5_12
void FASTCALL call_4_10_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_5_13
#define init_4_10_5_13
void FASTCALL call_4_10_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_5_14
#define init_4_10_5_14
void FASTCALL call_4_10_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_5_15
#define init_4_10_5_15
void FASTCALL call_4_10_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_6_0
#define init_4_10_6_0
void FASTCALL call_4_10_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_6_1
#define init_4_10_6_1
void FASTCALL call_4_10_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_6_2
#define init_4_10_6_2
void FASTCALL call_4_10_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_6_3
#define init_4_10_6_3
void FASTCALL call_4_10_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_6_4
#define init_4_10_6_4
void FASTCALL call_4_10_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_6_5
#define init_4_10_6_5
void FASTCALL call_4_10_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_6_6
#define init_4_10_6_6
void FASTCALL call_4_10_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_6_7
#define init_4_10_6_7
void FASTCALL call_4_10_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_6_8
#define init_4_10_6_8
void FASTCALL call_4_10_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_6_9
#define init_4_10_6_9
void FASTCALL call_4_10_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_6_10
#define init_4_10_6_10
void FASTCALL call_4_10_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_6_11
#define init_4_10_6_11
void FASTCALL call_4_10_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_6_12
#define init_4_10_6_12
void FASTCALL call_4_10_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_6_13
#define init_4_10_6_13
void FASTCALL call_4_10_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_6_14
#define init_4_10_6_14
void FASTCALL call_4_10_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_6_15
#define init_4_10_6_15
void FASTCALL call_4_10_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_7_0
#define init_4_10_7_0
void FASTCALL call_4_10_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_7_1
#define init_4_10_7_1
void FASTCALL call_4_10_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_7_2
#define init_4_10_7_2
void FASTCALL call_4_10_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_7_3
#define init_4_10_7_3
void FASTCALL call_4_10_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_7_4
#define init_4_10_7_4
void FASTCALL call_4_10_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_7_5
#define init_4_10_7_5
void FASTCALL call_4_10_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_7_6
#define init_4_10_7_6
void FASTCALL call_4_10_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_7_7
#define init_4_10_7_7
void FASTCALL call_4_10_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_7_8
#define init_4_10_7_8
void FASTCALL call_4_10_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_7_9
#define init_4_10_7_9
void FASTCALL call_4_10_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_7_10
#define init_4_10_7_10
void FASTCALL call_4_10_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_7_11
#define init_4_10_7_11
void FASTCALL call_4_10_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_7_12
#define init_4_10_7_12
void FASTCALL call_4_10_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_7_13
#define init_4_10_7_13
void FASTCALL call_4_10_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_7_14
#define init_4_10_7_14
void FASTCALL call_4_10_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_7_15
#define init_4_10_7_15
void FASTCALL call_4_10_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_8_0
#define init_4_10_8_0
void FASTCALL call_4_10_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_8_1
#define init_4_10_8_1
void FASTCALL call_4_10_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_8_2
#define init_4_10_8_2
void FASTCALL call_4_10_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_8_3
#define init_4_10_8_3
void FASTCALL call_4_10_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_8_4
#define init_4_10_8_4
void FASTCALL call_4_10_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_8_5
#define init_4_10_8_5
void FASTCALL call_4_10_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_8_6
#define init_4_10_8_6
void FASTCALL call_4_10_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_8_7
#define init_4_10_8_7
void FASTCALL call_4_10_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_8_8
#define init_4_10_8_8
void FASTCALL call_4_10_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_8_9
#define init_4_10_8_9
void FASTCALL call_4_10_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_8_10
#define init_4_10_8_10
void FASTCALL call_4_10_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_8_11
#define init_4_10_8_11
void FASTCALL call_4_10_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_8_12
#define init_4_10_8_12
void FASTCALL call_4_10_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_8_13
#define init_4_10_8_13
void FASTCALL call_4_10_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_8_14
#define init_4_10_8_14
void FASTCALL call_4_10_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_8_15
#define init_4_10_8_15
void FASTCALL call_4_10_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_9_0
#define init_4_10_9_0
void FASTCALL call_4_10_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_9_1
#define init_4_10_9_1
void FASTCALL call_4_10_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_9_2
#define init_4_10_9_2
void FASTCALL call_4_10_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_9_3
#define init_4_10_9_3
void FASTCALL call_4_10_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_9_4
#define init_4_10_9_4
void FASTCALL call_4_10_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_9_5
#define init_4_10_9_5
void FASTCALL call_4_10_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_9_6
#define init_4_10_9_6
void FASTCALL call_4_10_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_9_7
#define init_4_10_9_7
void FASTCALL call_4_10_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_9_8
#define init_4_10_9_8
void FASTCALL call_4_10_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_9_9
#define init_4_10_9_9
void FASTCALL call_4_10_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_9_10
#define init_4_10_9_10
void FASTCALL call_4_10_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_9_11
#define init_4_10_9_11
void FASTCALL call_4_10_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_9_12
#define init_4_10_9_12
void FASTCALL call_4_10_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_9_13
#define init_4_10_9_13
void FASTCALL call_4_10_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_9_14
#define init_4_10_9_14
void FASTCALL call_4_10_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_9_15
#define init_4_10_9_15
void FASTCALL call_4_10_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_10_0
#define init_4_10_10_0
void FASTCALL call_4_10_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_10_1
#define init_4_10_10_1
void FASTCALL call_4_10_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_10_2
#define init_4_10_10_2
void FASTCALL call_4_10_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_10_3
#define init_4_10_10_3
void FASTCALL call_4_10_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_10_4
#define init_4_10_10_4
void FASTCALL call_4_10_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_10_5
#define init_4_10_10_5
void FASTCALL call_4_10_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_10_6
#define init_4_10_10_6
void FASTCALL call_4_10_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_10_7
#define init_4_10_10_7
void FASTCALL call_4_10_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_10_8
#define init_4_10_10_8
void FASTCALL call_4_10_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_10_9
#define init_4_10_10_9
void FASTCALL call_4_10_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_10_10
#define init_4_10_10_10
void FASTCALL call_4_10_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_10_11
#define init_4_10_10_11
void FASTCALL call_4_10_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_10_12
#define init_4_10_10_12
void FASTCALL call_4_10_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_10_13
#define init_4_10_10_13
void FASTCALL call_4_10_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_10_14
#define init_4_10_10_14
void FASTCALL call_4_10_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_10_15
#define init_4_10_10_15
void FASTCALL call_4_10_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_11_0
#define init_4_10_11_0
void FASTCALL call_4_10_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_11_1
#define init_4_10_11_1
void FASTCALL call_4_10_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_11_2
#define init_4_10_11_2
void FASTCALL call_4_10_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_11_3
#define init_4_10_11_3
void FASTCALL call_4_10_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_11_4
#define init_4_10_11_4
void FASTCALL call_4_10_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_11_5
#define init_4_10_11_5
void FASTCALL call_4_10_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_11_6
#define init_4_10_11_6
void FASTCALL call_4_10_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_11_7
#define init_4_10_11_7
void FASTCALL call_4_10_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_11_8
#define init_4_10_11_8
void FASTCALL call_4_10_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_11_9
#define init_4_10_11_9
void FASTCALL call_4_10_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_11_10
#define init_4_10_11_10
void FASTCALL call_4_10_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_11_11
#define init_4_10_11_11
void FASTCALL call_4_10_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_11_12
#define init_4_10_11_12
void FASTCALL call_4_10_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_11_13
#define init_4_10_11_13
void FASTCALL call_4_10_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_11_14
#define init_4_10_11_14
void FASTCALL call_4_10_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_11_15
#define init_4_10_11_15
void FASTCALL call_4_10_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_12_0
#define init_4_10_12_0
void FASTCALL call_4_10_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_12_1
#define init_4_10_12_1
void FASTCALL call_4_10_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_12_2
#define init_4_10_12_2
void FASTCALL call_4_10_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_12_3
#define init_4_10_12_3
void FASTCALL call_4_10_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_12_4
#define init_4_10_12_4
void FASTCALL call_4_10_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_12_5
#define init_4_10_12_5
void FASTCALL call_4_10_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_12_6
#define init_4_10_12_6
void FASTCALL call_4_10_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_12_7
#define init_4_10_12_7
void FASTCALL call_4_10_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_12_8
#define init_4_10_12_8
void FASTCALL call_4_10_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_12_9
#define init_4_10_12_9
void FASTCALL call_4_10_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_12_10
#define init_4_10_12_10
void FASTCALL call_4_10_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_12_11
#define init_4_10_12_11
void FASTCALL call_4_10_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_12_12
#define init_4_10_12_12
void FASTCALL call_4_10_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_12_13
#define init_4_10_12_13
void FASTCALL call_4_10_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_12_14
#define init_4_10_12_14
void FASTCALL call_4_10_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_12_15
#define init_4_10_12_15
void FASTCALL call_4_10_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_13_0
#define init_4_10_13_0
void FASTCALL call_4_10_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_13_1
#define init_4_10_13_1
void FASTCALL call_4_10_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_13_2
#define init_4_10_13_2
void FASTCALL call_4_10_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_13_3
#define init_4_10_13_3
void FASTCALL call_4_10_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_13_4
#define init_4_10_13_4
void FASTCALL call_4_10_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_13_5
#define init_4_10_13_5
void FASTCALL call_4_10_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_13_6
#define init_4_10_13_6
void FASTCALL call_4_10_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_13_7
#define init_4_10_13_7
void FASTCALL call_4_10_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_13_8
#define init_4_10_13_8
void FASTCALL call_4_10_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_13_9
#define init_4_10_13_9
void FASTCALL call_4_10_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_13_10
#define init_4_10_13_10
void FASTCALL call_4_10_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_13_11
#define init_4_10_13_11
void FASTCALL call_4_10_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_13_12
#define init_4_10_13_12
void FASTCALL call_4_10_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_13_13
#define init_4_10_13_13
void FASTCALL call_4_10_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_13_14
#define init_4_10_13_14
void FASTCALL call_4_10_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_13_15
#define init_4_10_13_15
void FASTCALL call_4_10_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_14_0
#define init_4_10_14_0
void FASTCALL call_4_10_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_14_1
#define init_4_10_14_1
void FASTCALL call_4_10_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_14_2
#define init_4_10_14_2
void FASTCALL call_4_10_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_14_3
#define init_4_10_14_3
void FASTCALL call_4_10_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_14_4
#define init_4_10_14_4
void FASTCALL call_4_10_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_14_5
#define init_4_10_14_5
void FASTCALL call_4_10_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_14_6
#define init_4_10_14_6
void FASTCALL call_4_10_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_14_7
#define init_4_10_14_7
void FASTCALL call_4_10_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_14_8
#define init_4_10_14_8
void FASTCALL call_4_10_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_14_9
#define init_4_10_14_9
void FASTCALL call_4_10_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_14_10
#define init_4_10_14_10
void FASTCALL call_4_10_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_14_11
#define init_4_10_14_11
void FASTCALL call_4_10_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_14_12
#define init_4_10_14_12
void FASTCALL call_4_10_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_14_13
#define init_4_10_14_13
void FASTCALL call_4_10_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_14_14
#define init_4_10_14_14
void FASTCALL call_4_10_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_14_15
#define init_4_10_14_15
void FASTCALL call_4_10_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_15_0
#define init_4_10_15_0
void FASTCALL call_4_10_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_15_1
#define init_4_10_15_1
void FASTCALL call_4_10_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_15_2
#define init_4_10_15_2
void FASTCALL call_4_10_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_15_3
#define init_4_10_15_3
void FASTCALL call_4_10_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_15_4
#define init_4_10_15_4
void FASTCALL call_4_10_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_15_5
#define init_4_10_15_5
void FASTCALL call_4_10_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_15_6
#define init_4_10_15_6
void FASTCALL call_4_10_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_15_7
#define init_4_10_15_7
void FASTCALL call_4_10_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_15_8
#define init_4_10_15_8
void FASTCALL call_4_10_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_15_9
#define init_4_10_15_9
void FASTCALL call_4_10_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_15_10
#define init_4_10_15_10
void FASTCALL call_4_10_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_15_11
#define init_4_10_15_11
void FASTCALL call_4_10_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_15_12
#define init_4_10_15_12
void FASTCALL call_4_10_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_15_13
#define init_4_10_15_13
void FASTCALL call_4_10_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_15_14
#define init_4_10_15_14
void FASTCALL call_4_10_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_10_15_15
#define init_4_10_15_15
void FASTCALL call_4_10_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_0_0
#define init_4_11_0_0
void FASTCALL call_4_11_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_0_1
#define init_4_11_0_1
void FASTCALL call_4_11_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_0_2
#define init_4_11_0_2
void FASTCALL call_4_11_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_0_3
#define init_4_11_0_3
void FASTCALL call_4_11_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_0_4
#define init_4_11_0_4
void FASTCALL call_4_11_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_0_5
#define init_4_11_0_5
void FASTCALL call_4_11_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_0_6
#define init_4_11_0_6
void FASTCALL call_4_11_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_0_7
#define init_4_11_0_7
void FASTCALL call_4_11_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_0_8
#define init_4_11_0_8
void FASTCALL call_4_11_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_0_9
#define init_4_11_0_9
void FASTCALL call_4_11_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_0_10
#define init_4_11_0_10
void FASTCALL call_4_11_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_0_11
#define init_4_11_0_11
void FASTCALL call_4_11_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_0_12
#define init_4_11_0_12
void FASTCALL call_4_11_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_0_13
#define init_4_11_0_13
void FASTCALL call_4_11_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_0_14
#define init_4_11_0_14
void FASTCALL call_4_11_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_0_15
#define init_4_11_0_15
void FASTCALL call_4_11_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_1_0
#define init_4_11_1_0
void FASTCALL call_4_11_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_1_1
#define init_4_11_1_1
void FASTCALL call_4_11_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_1_2
#define init_4_11_1_2
void FASTCALL call_4_11_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_1_3
#define init_4_11_1_3
void FASTCALL call_4_11_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_1_4
#define init_4_11_1_4
void FASTCALL call_4_11_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_1_5
#define init_4_11_1_5
void FASTCALL call_4_11_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_1_6
#define init_4_11_1_6
void FASTCALL call_4_11_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_1_7
#define init_4_11_1_7
void FASTCALL call_4_11_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_1_8
#define init_4_11_1_8
void FASTCALL call_4_11_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_1_9
#define init_4_11_1_9
void FASTCALL call_4_11_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_1_10
#define init_4_11_1_10
void FASTCALL call_4_11_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_1_11
#define init_4_11_1_11
void FASTCALL call_4_11_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_1_12
#define init_4_11_1_12
void FASTCALL call_4_11_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_1_13
#define init_4_11_1_13
void FASTCALL call_4_11_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_1_14
#define init_4_11_1_14
void FASTCALL call_4_11_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_1_15
#define init_4_11_1_15
void FASTCALL call_4_11_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_2_0
#define init_4_11_2_0
void FASTCALL call_4_11_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_2_1
#define init_4_11_2_1
void FASTCALL call_4_11_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_2_2
#define init_4_11_2_2
void FASTCALL call_4_11_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_2_3
#define init_4_11_2_3
void FASTCALL call_4_11_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_2_4
#define init_4_11_2_4
void FASTCALL call_4_11_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_2_5
#define init_4_11_2_5
void FASTCALL call_4_11_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_2_6
#define init_4_11_2_6
void FASTCALL call_4_11_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_2_7
#define init_4_11_2_7
void FASTCALL call_4_11_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_2_8
#define init_4_11_2_8
void FASTCALL call_4_11_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_2_9
#define init_4_11_2_9
void FASTCALL call_4_11_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_2_10
#define init_4_11_2_10
void FASTCALL call_4_11_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_2_11
#define init_4_11_2_11
void FASTCALL call_4_11_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_2_12
#define init_4_11_2_12
void FASTCALL call_4_11_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_2_13
#define init_4_11_2_13
void FASTCALL call_4_11_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_2_14
#define init_4_11_2_14
void FASTCALL call_4_11_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_2_15
#define init_4_11_2_15
void FASTCALL call_4_11_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_3_0
#define init_4_11_3_0
void FASTCALL call_4_11_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_3_1
#define init_4_11_3_1
void FASTCALL call_4_11_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_3_2
#define init_4_11_3_2
void FASTCALL call_4_11_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_3_3
#define init_4_11_3_3
void FASTCALL call_4_11_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_3_4
#define init_4_11_3_4
void FASTCALL call_4_11_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_3_5
#define init_4_11_3_5
void FASTCALL call_4_11_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_3_6
#define init_4_11_3_6
void FASTCALL call_4_11_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_3_7
#define init_4_11_3_7
void FASTCALL call_4_11_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_3_8
#define init_4_11_3_8
void FASTCALL call_4_11_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_3_9
#define init_4_11_3_9
void FASTCALL call_4_11_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_3_10
#define init_4_11_3_10
void FASTCALL call_4_11_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_3_11
#define init_4_11_3_11
void FASTCALL call_4_11_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_3_12
#define init_4_11_3_12
void FASTCALL call_4_11_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_3_13
#define init_4_11_3_13
void FASTCALL call_4_11_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_3_14
#define init_4_11_3_14
void FASTCALL call_4_11_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_3_15
#define init_4_11_3_15
void FASTCALL call_4_11_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_4_0
#define init_4_11_4_0
void FASTCALL call_4_11_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_4_1
#define init_4_11_4_1
void FASTCALL call_4_11_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_4_2
#define init_4_11_4_2
void FASTCALL call_4_11_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_4_3
#define init_4_11_4_3
void FASTCALL call_4_11_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_4_4
#define init_4_11_4_4
void FASTCALL call_4_11_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_4_5
#define init_4_11_4_5
void FASTCALL call_4_11_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_4_6
#define init_4_11_4_6
void FASTCALL call_4_11_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_4_7
#define init_4_11_4_7
void FASTCALL call_4_11_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_4_8
#define init_4_11_4_8
void FASTCALL call_4_11_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_4_9
#define init_4_11_4_9
void FASTCALL call_4_11_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_4_10
#define init_4_11_4_10
void FASTCALL call_4_11_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_4_11
#define init_4_11_4_11
void FASTCALL call_4_11_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_4_12
#define init_4_11_4_12
void FASTCALL call_4_11_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_4_13
#define init_4_11_4_13
void FASTCALL call_4_11_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_4_14
#define init_4_11_4_14
void FASTCALL call_4_11_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_4_15
#define init_4_11_4_15
void FASTCALL call_4_11_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_5_0
#define init_4_11_5_0
void FASTCALL call_4_11_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_5_1
#define init_4_11_5_1
void FASTCALL call_4_11_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_5_2
#define init_4_11_5_2
void FASTCALL call_4_11_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_5_3
#define init_4_11_5_3
void FASTCALL call_4_11_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_5_4
#define init_4_11_5_4
void FASTCALL call_4_11_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_5_5
#define init_4_11_5_5
void FASTCALL call_4_11_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_5_6
#define init_4_11_5_6
void FASTCALL call_4_11_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_5_7
#define init_4_11_5_7
void FASTCALL call_4_11_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_5_8
#define init_4_11_5_8
void FASTCALL call_4_11_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_5_9
#define init_4_11_5_9
void FASTCALL call_4_11_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_5_10
#define init_4_11_5_10
void FASTCALL call_4_11_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_5_11
#define init_4_11_5_11
void FASTCALL call_4_11_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_5_12
#define init_4_11_5_12
void FASTCALL call_4_11_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_5_13
#define init_4_11_5_13
void FASTCALL call_4_11_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_5_14
#define init_4_11_5_14
void FASTCALL call_4_11_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_5_15
#define init_4_11_5_15
void FASTCALL call_4_11_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_6_0
#define init_4_11_6_0
void FASTCALL call_4_11_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_6_1
#define init_4_11_6_1
void FASTCALL call_4_11_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_6_2
#define init_4_11_6_2
void FASTCALL call_4_11_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_6_3
#define init_4_11_6_3
void FASTCALL call_4_11_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_6_4
#define init_4_11_6_4
void FASTCALL call_4_11_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_6_5
#define init_4_11_6_5
void FASTCALL call_4_11_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_6_6
#define init_4_11_6_6
void FASTCALL call_4_11_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_6_7
#define init_4_11_6_7
void FASTCALL call_4_11_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_6_8
#define init_4_11_6_8
void FASTCALL call_4_11_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_6_9
#define init_4_11_6_9
void FASTCALL call_4_11_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_6_10
#define init_4_11_6_10
void FASTCALL call_4_11_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_6_11
#define init_4_11_6_11
void FASTCALL call_4_11_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_6_12
#define init_4_11_6_12
void FASTCALL call_4_11_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_6_13
#define init_4_11_6_13
void FASTCALL call_4_11_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_6_14
#define init_4_11_6_14
void FASTCALL call_4_11_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_6_15
#define init_4_11_6_15
void FASTCALL call_4_11_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_7_0
#define init_4_11_7_0
void FASTCALL call_4_11_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_7_1
#define init_4_11_7_1
void FASTCALL call_4_11_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_7_2
#define init_4_11_7_2
void FASTCALL call_4_11_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_7_3
#define init_4_11_7_3
void FASTCALL call_4_11_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_7_4
#define init_4_11_7_4
void FASTCALL call_4_11_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_7_5
#define init_4_11_7_5
void FASTCALL call_4_11_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_7_6
#define init_4_11_7_6
void FASTCALL call_4_11_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_7_7
#define init_4_11_7_7
void FASTCALL call_4_11_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_7_8
#define init_4_11_7_8
void FASTCALL call_4_11_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_7_9
#define init_4_11_7_9
void FASTCALL call_4_11_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_7_10
#define init_4_11_7_10
void FASTCALL call_4_11_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_7_11
#define init_4_11_7_11
void FASTCALL call_4_11_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_7_12
#define init_4_11_7_12
void FASTCALL call_4_11_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_7_13
#define init_4_11_7_13
void FASTCALL call_4_11_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_7_14
#define init_4_11_7_14
void FASTCALL call_4_11_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_7_15
#define init_4_11_7_15
void FASTCALL call_4_11_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_8_0
#define init_4_11_8_0
void FASTCALL call_4_11_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_8_1
#define init_4_11_8_1
void FASTCALL call_4_11_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_8_2
#define init_4_11_8_2
void FASTCALL call_4_11_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_8_3
#define init_4_11_8_3
void FASTCALL call_4_11_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_8_4
#define init_4_11_8_4
void FASTCALL call_4_11_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_8_5
#define init_4_11_8_5
void FASTCALL call_4_11_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_8_6
#define init_4_11_8_6
void FASTCALL call_4_11_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_8_7
#define init_4_11_8_7
void FASTCALL call_4_11_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_8_8
#define init_4_11_8_8
void FASTCALL call_4_11_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_8_9
#define init_4_11_8_9
void FASTCALL call_4_11_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_8_10
#define init_4_11_8_10
void FASTCALL call_4_11_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_8_11
#define init_4_11_8_11
void FASTCALL call_4_11_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_8_12
#define init_4_11_8_12
void FASTCALL call_4_11_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_8_13
#define init_4_11_8_13
void FASTCALL call_4_11_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_8_14
#define init_4_11_8_14
void FASTCALL call_4_11_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_8_15
#define init_4_11_8_15
void FASTCALL call_4_11_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_9_0
#define init_4_11_9_0
void FASTCALL call_4_11_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_9_1
#define init_4_11_9_1
void FASTCALL call_4_11_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_9_2
#define init_4_11_9_2
void FASTCALL call_4_11_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_9_3
#define init_4_11_9_3
void FASTCALL call_4_11_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_9_4
#define init_4_11_9_4
void FASTCALL call_4_11_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_9_5
#define init_4_11_9_5
void FASTCALL call_4_11_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_9_6
#define init_4_11_9_6
void FASTCALL call_4_11_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_9_7
#define init_4_11_9_7
void FASTCALL call_4_11_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_9_8
#define init_4_11_9_8
void FASTCALL call_4_11_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_9_9
#define init_4_11_9_9
void FASTCALL call_4_11_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_9_10
#define init_4_11_9_10
void FASTCALL call_4_11_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_9_11
#define init_4_11_9_11
void FASTCALL call_4_11_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_9_12
#define init_4_11_9_12
void FASTCALL call_4_11_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_9_13
#define init_4_11_9_13
void FASTCALL call_4_11_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_9_14
#define init_4_11_9_14
void FASTCALL call_4_11_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_9_15
#define init_4_11_9_15
void FASTCALL call_4_11_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_10_0
#define init_4_11_10_0
void FASTCALL call_4_11_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_10_1
#define init_4_11_10_1
void FASTCALL call_4_11_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_10_2
#define init_4_11_10_2
void FASTCALL call_4_11_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_10_3
#define init_4_11_10_3
void FASTCALL call_4_11_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_10_4
#define init_4_11_10_4
void FASTCALL call_4_11_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_10_5
#define init_4_11_10_5
void FASTCALL call_4_11_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_10_6
#define init_4_11_10_6
void FASTCALL call_4_11_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_10_7
#define init_4_11_10_7
void FASTCALL call_4_11_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_10_8
#define init_4_11_10_8
void FASTCALL call_4_11_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_10_9
#define init_4_11_10_9
void FASTCALL call_4_11_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_10_10
#define init_4_11_10_10
void FASTCALL call_4_11_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_10_11
#define init_4_11_10_11
void FASTCALL call_4_11_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_10_12
#define init_4_11_10_12
void FASTCALL call_4_11_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_10_13
#define init_4_11_10_13
void FASTCALL call_4_11_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_10_14
#define init_4_11_10_14
void FASTCALL call_4_11_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_10_15
#define init_4_11_10_15
void FASTCALL call_4_11_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_11_0
#define init_4_11_11_0
void FASTCALL call_4_11_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_11_1
#define init_4_11_11_1
void FASTCALL call_4_11_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_11_2
#define init_4_11_11_2
void FASTCALL call_4_11_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_11_3
#define init_4_11_11_3
void FASTCALL call_4_11_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_11_4
#define init_4_11_11_4
void FASTCALL call_4_11_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_11_5
#define init_4_11_11_5
void FASTCALL call_4_11_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_11_6
#define init_4_11_11_6
void FASTCALL call_4_11_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_11_7
#define init_4_11_11_7
void FASTCALL call_4_11_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_11_8
#define init_4_11_11_8
void FASTCALL call_4_11_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_11_9
#define init_4_11_11_9
void FASTCALL call_4_11_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_11_10
#define init_4_11_11_10
void FASTCALL call_4_11_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_11_11
#define init_4_11_11_11
void FASTCALL call_4_11_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_11_12
#define init_4_11_11_12
void FASTCALL call_4_11_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_11_13
#define init_4_11_11_13
void FASTCALL call_4_11_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_11_14
#define init_4_11_11_14
void FASTCALL call_4_11_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_11_15
#define init_4_11_11_15
void FASTCALL call_4_11_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_12_0
#define init_4_11_12_0
void FASTCALL call_4_11_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_12_1
#define init_4_11_12_1
void FASTCALL call_4_11_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_12_2
#define init_4_11_12_2
void FASTCALL call_4_11_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_12_3
#define init_4_11_12_3
void FASTCALL call_4_11_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_12_4
#define init_4_11_12_4
void FASTCALL call_4_11_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_12_5
#define init_4_11_12_5
void FASTCALL call_4_11_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_12_6
#define init_4_11_12_6
void FASTCALL call_4_11_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_12_7
#define init_4_11_12_7
void FASTCALL call_4_11_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_12_8
#define init_4_11_12_8
void FASTCALL call_4_11_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_12_9
#define init_4_11_12_9
void FASTCALL call_4_11_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_12_10
#define init_4_11_12_10
void FASTCALL call_4_11_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_12_11
#define init_4_11_12_11
void FASTCALL call_4_11_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_12_12
#define init_4_11_12_12
void FASTCALL call_4_11_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_12_13
#define init_4_11_12_13
void FASTCALL call_4_11_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_12_14
#define init_4_11_12_14
void FASTCALL call_4_11_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_12_15
#define init_4_11_12_15
void FASTCALL call_4_11_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_13_0
#define init_4_11_13_0
void FASTCALL call_4_11_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_13_1
#define init_4_11_13_1
void FASTCALL call_4_11_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_13_2
#define init_4_11_13_2
void FASTCALL call_4_11_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_13_3
#define init_4_11_13_3
void FASTCALL call_4_11_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_13_4
#define init_4_11_13_4
void FASTCALL call_4_11_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_13_5
#define init_4_11_13_5
void FASTCALL call_4_11_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_13_6
#define init_4_11_13_6
void FASTCALL call_4_11_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_13_7
#define init_4_11_13_7
void FASTCALL call_4_11_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_13_8
#define init_4_11_13_8
void FASTCALL call_4_11_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_13_9
#define init_4_11_13_9
void FASTCALL call_4_11_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_13_10
#define init_4_11_13_10
void FASTCALL call_4_11_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_13_11
#define init_4_11_13_11
void FASTCALL call_4_11_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_13_12
#define init_4_11_13_12
void FASTCALL call_4_11_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_13_13
#define init_4_11_13_13
void FASTCALL call_4_11_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_13_14
#define init_4_11_13_14
void FASTCALL call_4_11_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_13_15
#define init_4_11_13_15
void FASTCALL call_4_11_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_14_0
#define init_4_11_14_0
void FASTCALL call_4_11_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_14_1
#define init_4_11_14_1
void FASTCALL call_4_11_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_14_2
#define init_4_11_14_2
void FASTCALL call_4_11_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_14_3
#define init_4_11_14_3
void FASTCALL call_4_11_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_14_4
#define init_4_11_14_4
void FASTCALL call_4_11_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_14_5
#define init_4_11_14_5
void FASTCALL call_4_11_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_14_6
#define init_4_11_14_6
void FASTCALL call_4_11_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_14_7
#define init_4_11_14_7
void FASTCALL call_4_11_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_14_8
#define init_4_11_14_8
void FASTCALL call_4_11_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_14_9
#define init_4_11_14_9
void FASTCALL call_4_11_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_14_10
#define init_4_11_14_10
void FASTCALL call_4_11_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_14_11
#define init_4_11_14_11
void FASTCALL call_4_11_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_14_12
#define init_4_11_14_12
void FASTCALL call_4_11_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_14_13
#define init_4_11_14_13
void FASTCALL call_4_11_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_14_14
#define init_4_11_14_14
void FASTCALL call_4_11_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_14_15
#define init_4_11_14_15
void FASTCALL call_4_11_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_15_0
#define init_4_11_15_0
void FASTCALL call_4_11_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_15_1
#define init_4_11_15_1
void FASTCALL call_4_11_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_15_2
#define init_4_11_15_2
void FASTCALL call_4_11_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_15_3
#define init_4_11_15_3
void FASTCALL call_4_11_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_15_4
#define init_4_11_15_4
void FASTCALL call_4_11_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_15_5
#define init_4_11_15_5
void FASTCALL call_4_11_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_15_6
#define init_4_11_15_6
void FASTCALL call_4_11_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_15_7
#define init_4_11_15_7
void FASTCALL call_4_11_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_15_8
#define init_4_11_15_8
void FASTCALL call_4_11_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_15_9
#define init_4_11_15_9
void FASTCALL call_4_11_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_15_10
#define init_4_11_15_10
void FASTCALL call_4_11_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_15_11
#define init_4_11_15_11
void FASTCALL call_4_11_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_15_12
#define init_4_11_15_12
void FASTCALL call_4_11_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_15_13
#define init_4_11_15_13
void FASTCALL call_4_11_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_15_14
#define init_4_11_15_14
void FASTCALL call_4_11_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_11_15_15
#define init_4_11_15_15
void FASTCALL call_4_11_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_0_0
#define init_4_12_0_0
void FASTCALL call_4_12_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_0_1
#define init_4_12_0_1
void FASTCALL call_4_12_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_0_2
#define init_4_12_0_2
void FASTCALL call_4_12_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_0_3
#define init_4_12_0_3
void FASTCALL call_4_12_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_0_4
#define init_4_12_0_4
void FASTCALL call_4_12_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_0_5
#define init_4_12_0_5
void FASTCALL call_4_12_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_0_6
#define init_4_12_0_6
void FASTCALL call_4_12_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_0_7
#define init_4_12_0_7
void FASTCALL call_4_12_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_0_8
#define init_4_12_0_8
void FASTCALL call_4_12_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_0_9
#define init_4_12_0_9
void FASTCALL call_4_12_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_0_10
#define init_4_12_0_10
void FASTCALL call_4_12_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_0_11
#define init_4_12_0_11
void FASTCALL call_4_12_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_0_12
#define init_4_12_0_12
void FASTCALL call_4_12_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_0_13
#define init_4_12_0_13
void FASTCALL call_4_12_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_0_14
#define init_4_12_0_14
void FASTCALL call_4_12_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_0_15
#define init_4_12_0_15
void FASTCALL call_4_12_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_1_0
#define init_4_12_1_0
void FASTCALL call_4_12_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_1_1
#define init_4_12_1_1
void FASTCALL call_4_12_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_1_2
#define init_4_12_1_2
void FASTCALL call_4_12_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_1_3
#define init_4_12_1_3
void FASTCALL call_4_12_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_1_4
#define init_4_12_1_4
void FASTCALL call_4_12_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_1_5
#define init_4_12_1_5
void FASTCALL call_4_12_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_1_6
#define init_4_12_1_6
void FASTCALL call_4_12_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_1_7
#define init_4_12_1_7
void FASTCALL call_4_12_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_1_8
#define init_4_12_1_8
void FASTCALL call_4_12_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_1_9
#define init_4_12_1_9
void FASTCALL call_4_12_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_1_10
#define init_4_12_1_10
void FASTCALL call_4_12_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_1_11
#define init_4_12_1_11
void FASTCALL call_4_12_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_1_12
#define init_4_12_1_12
void FASTCALL call_4_12_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_1_13
#define init_4_12_1_13
void FASTCALL call_4_12_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_1_14
#define init_4_12_1_14
void FASTCALL call_4_12_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_1_15
#define init_4_12_1_15
void FASTCALL call_4_12_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_2_0
#define init_4_12_2_0
void FASTCALL call_4_12_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_2_1
#define init_4_12_2_1
void FASTCALL call_4_12_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_2_2
#define init_4_12_2_2
void FASTCALL call_4_12_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_2_3
#define init_4_12_2_3
void FASTCALL call_4_12_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_2_4
#define init_4_12_2_4
void FASTCALL call_4_12_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_2_5
#define init_4_12_2_5
void FASTCALL call_4_12_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_2_6
#define init_4_12_2_6
void FASTCALL call_4_12_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_2_7
#define init_4_12_2_7
void FASTCALL call_4_12_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_2_8
#define init_4_12_2_8
void FASTCALL call_4_12_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_2_9
#define init_4_12_2_9
void FASTCALL call_4_12_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_2_10
#define init_4_12_2_10
void FASTCALL call_4_12_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_2_11
#define init_4_12_2_11
void FASTCALL call_4_12_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_2_12
#define init_4_12_2_12
void FASTCALL call_4_12_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_2_13
#define init_4_12_2_13
void FASTCALL call_4_12_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_2_14
#define init_4_12_2_14
void FASTCALL call_4_12_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_2_15
#define init_4_12_2_15
void FASTCALL call_4_12_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_3_0
#define init_4_12_3_0
void FASTCALL call_4_12_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_3_1
#define init_4_12_3_1
void FASTCALL call_4_12_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_3_2
#define init_4_12_3_2
void FASTCALL call_4_12_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_3_3
#define init_4_12_3_3
void FASTCALL call_4_12_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_3_4
#define init_4_12_3_4
void FASTCALL call_4_12_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_3_5
#define init_4_12_3_5
void FASTCALL call_4_12_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_3_6
#define init_4_12_3_6
void FASTCALL call_4_12_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_3_7
#define init_4_12_3_7
void FASTCALL call_4_12_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_3_8
#define init_4_12_3_8
void FASTCALL call_4_12_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_3_9
#define init_4_12_3_9
void FASTCALL call_4_12_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_3_10
#define init_4_12_3_10
void FASTCALL call_4_12_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_3_11
#define init_4_12_3_11
void FASTCALL call_4_12_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_3_12
#define init_4_12_3_12
void FASTCALL call_4_12_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_3_13
#define init_4_12_3_13
void FASTCALL call_4_12_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_3_14
#define init_4_12_3_14
void FASTCALL call_4_12_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_3_15
#define init_4_12_3_15
void FASTCALL call_4_12_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_4_0
#define init_4_12_4_0
void FASTCALL call_4_12_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_4_1
#define init_4_12_4_1
void FASTCALL call_4_12_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_4_2
#define init_4_12_4_2
void FASTCALL call_4_12_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_4_3
#define init_4_12_4_3
void FASTCALL call_4_12_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_4_4
#define init_4_12_4_4
void FASTCALL call_4_12_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_4_5
#define init_4_12_4_5
void FASTCALL call_4_12_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_4_6
#define init_4_12_4_6
void FASTCALL call_4_12_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_4_7
#define init_4_12_4_7
void FASTCALL call_4_12_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_4_8
#define init_4_12_4_8
void FASTCALL call_4_12_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_4_9
#define init_4_12_4_9
void FASTCALL call_4_12_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_4_10
#define init_4_12_4_10
void FASTCALL call_4_12_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_4_11
#define init_4_12_4_11
void FASTCALL call_4_12_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_4_12
#define init_4_12_4_12
void FASTCALL call_4_12_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_4_13
#define init_4_12_4_13
void FASTCALL call_4_12_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_4_14
#define init_4_12_4_14
void FASTCALL call_4_12_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_4_15
#define init_4_12_4_15
void FASTCALL call_4_12_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_5_0
#define init_4_12_5_0
void FASTCALL call_4_12_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_5_1
#define init_4_12_5_1
void FASTCALL call_4_12_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_5_2
#define init_4_12_5_2
void FASTCALL call_4_12_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_5_3
#define init_4_12_5_3
void FASTCALL call_4_12_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_5_4
#define init_4_12_5_4
void FASTCALL call_4_12_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_5_5
#define init_4_12_5_5
void FASTCALL call_4_12_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_5_6
#define init_4_12_5_6
void FASTCALL call_4_12_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_5_7
#define init_4_12_5_7
void FASTCALL call_4_12_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_5_8
#define init_4_12_5_8
void FASTCALL call_4_12_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_5_9
#define init_4_12_5_9
void FASTCALL call_4_12_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_5_10
#define init_4_12_5_10
void FASTCALL call_4_12_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_5_11
#define init_4_12_5_11
void FASTCALL call_4_12_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_5_12
#define init_4_12_5_12
void FASTCALL call_4_12_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_5_13
#define init_4_12_5_13
void FASTCALL call_4_12_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_5_14
#define init_4_12_5_14
void FASTCALL call_4_12_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_5_15
#define init_4_12_5_15
void FASTCALL call_4_12_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_6_0
#define init_4_12_6_0
void FASTCALL call_4_12_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_6_1
#define init_4_12_6_1
void FASTCALL call_4_12_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_6_2
#define init_4_12_6_2
void FASTCALL call_4_12_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_6_3
#define init_4_12_6_3
void FASTCALL call_4_12_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_6_4
#define init_4_12_6_4
void FASTCALL call_4_12_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_6_5
#define init_4_12_6_5
void FASTCALL call_4_12_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_6_6
#define init_4_12_6_6
void FASTCALL call_4_12_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_6_7
#define init_4_12_6_7
void FASTCALL call_4_12_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_6_8
#define init_4_12_6_8
void FASTCALL call_4_12_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_6_9
#define init_4_12_6_9
void FASTCALL call_4_12_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_6_10
#define init_4_12_6_10
void FASTCALL call_4_12_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_6_11
#define init_4_12_6_11
void FASTCALL call_4_12_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_6_12
#define init_4_12_6_12
void FASTCALL call_4_12_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_6_13
#define init_4_12_6_13
void FASTCALL call_4_12_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_6_14
#define init_4_12_6_14
void FASTCALL call_4_12_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_6_15
#define init_4_12_6_15
void FASTCALL call_4_12_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_7_0
#define init_4_12_7_0
void FASTCALL call_4_12_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_7_1
#define init_4_12_7_1
void FASTCALL call_4_12_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_7_2
#define init_4_12_7_2
void FASTCALL call_4_12_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_7_3
#define init_4_12_7_3
void FASTCALL call_4_12_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_7_4
#define init_4_12_7_4
void FASTCALL call_4_12_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_7_5
#define init_4_12_7_5
void FASTCALL call_4_12_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_7_6
#define init_4_12_7_6
void FASTCALL call_4_12_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_7_7
#define init_4_12_7_7
void FASTCALL call_4_12_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_7_8
#define init_4_12_7_8
void FASTCALL call_4_12_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_7_9
#define init_4_12_7_9
void FASTCALL call_4_12_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_7_10
#define init_4_12_7_10
void FASTCALL call_4_12_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_7_11
#define init_4_12_7_11
void FASTCALL call_4_12_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_7_12
#define init_4_12_7_12
void FASTCALL call_4_12_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_7_13
#define init_4_12_7_13
void FASTCALL call_4_12_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_7_14
#define init_4_12_7_14
void FASTCALL call_4_12_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_7_15
#define init_4_12_7_15
void FASTCALL call_4_12_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_8_0
#define init_4_12_8_0
void FASTCALL call_4_12_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_8_1
#define init_4_12_8_1
void FASTCALL call_4_12_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_8_2
#define init_4_12_8_2
void FASTCALL call_4_12_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_8_3
#define init_4_12_8_3
void FASTCALL call_4_12_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_8_4
#define init_4_12_8_4
void FASTCALL call_4_12_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_8_5
#define init_4_12_8_5
void FASTCALL call_4_12_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_8_6
#define init_4_12_8_6
void FASTCALL call_4_12_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_8_7
#define init_4_12_8_7
void FASTCALL call_4_12_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_8_8
#define init_4_12_8_8
void FASTCALL call_4_12_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_8_9
#define init_4_12_8_9
void FASTCALL call_4_12_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_8_10
#define init_4_12_8_10
void FASTCALL call_4_12_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_8_11
#define init_4_12_8_11
void FASTCALL call_4_12_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_8_12
#define init_4_12_8_12
void FASTCALL call_4_12_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_8_13
#define init_4_12_8_13
void FASTCALL call_4_12_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_8_14
#define init_4_12_8_14
void FASTCALL call_4_12_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_8_15
#define init_4_12_8_15
void FASTCALL call_4_12_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_9_0
#define init_4_12_9_0
void FASTCALL call_4_12_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_9_1
#define init_4_12_9_1
void FASTCALL call_4_12_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_9_2
#define init_4_12_9_2
void FASTCALL call_4_12_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_9_3
#define init_4_12_9_3
void FASTCALL call_4_12_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_9_4
#define init_4_12_9_4
void FASTCALL call_4_12_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_9_5
#define init_4_12_9_5
void FASTCALL call_4_12_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_9_6
#define init_4_12_9_6
void FASTCALL call_4_12_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_9_7
#define init_4_12_9_7
void FASTCALL call_4_12_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_9_8
#define init_4_12_9_8
void FASTCALL call_4_12_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_9_9
#define init_4_12_9_9
void FASTCALL call_4_12_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_9_10
#define init_4_12_9_10
void FASTCALL call_4_12_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_9_11
#define init_4_12_9_11
void FASTCALL call_4_12_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_9_12
#define init_4_12_9_12
void FASTCALL call_4_12_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_9_13
#define init_4_12_9_13
void FASTCALL call_4_12_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_9_14
#define init_4_12_9_14
void FASTCALL call_4_12_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_9_15
#define init_4_12_9_15
void FASTCALL call_4_12_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_10_0
#define init_4_12_10_0
void FASTCALL call_4_12_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_10_1
#define init_4_12_10_1
void FASTCALL call_4_12_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_10_2
#define init_4_12_10_2
void FASTCALL call_4_12_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_10_3
#define init_4_12_10_3
void FASTCALL call_4_12_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_10_4
#define init_4_12_10_4
void FASTCALL call_4_12_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_10_5
#define init_4_12_10_5
void FASTCALL call_4_12_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_10_6
#define init_4_12_10_6
void FASTCALL call_4_12_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_10_7
#define init_4_12_10_7
void FASTCALL call_4_12_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_10_8
#define init_4_12_10_8
void FASTCALL call_4_12_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_10_9
#define init_4_12_10_9
void FASTCALL call_4_12_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_10_10
#define init_4_12_10_10
void FASTCALL call_4_12_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_10_11
#define init_4_12_10_11
void FASTCALL call_4_12_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_10_12
#define init_4_12_10_12
void FASTCALL call_4_12_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_10_13
#define init_4_12_10_13
void FASTCALL call_4_12_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_10_14
#define init_4_12_10_14
void FASTCALL call_4_12_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_10_15
#define init_4_12_10_15
void FASTCALL call_4_12_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_11_0
#define init_4_12_11_0
void FASTCALL call_4_12_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_11_1
#define init_4_12_11_1
void FASTCALL call_4_12_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_11_2
#define init_4_12_11_2
void FASTCALL call_4_12_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_11_3
#define init_4_12_11_3
void FASTCALL call_4_12_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_11_4
#define init_4_12_11_4
void FASTCALL call_4_12_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_11_5
#define init_4_12_11_5
void FASTCALL call_4_12_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_11_6
#define init_4_12_11_6
void FASTCALL call_4_12_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_11_7
#define init_4_12_11_7
void FASTCALL call_4_12_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_11_8
#define init_4_12_11_8
void FASTCALL call_4_12_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_11_9
#define init_4_12_11_9
void FASTCALL call_4_12_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_11_10
#define init_4_12_11_10
void FASTCALL call_4_12_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_11_11
#define init_4_12_11_11
void FASTCALL call_4_12_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_11_12
#define init_4_12_11_12
void FASTCALL call_4_12_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_11_13
#define init_4_12_11_13
void FASTCALL call_4_12_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_11_14
#define init_4_12_11_14
void FASTCALL call_4_12_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_11_15
#define init_4_12_11_15
void FASTCALL call_4_12_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_12_0
#define init_4_12_12_0
void FASTCALL call_4_12_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_12_1
#define init_4_12_12_1
void FASTCALL call_4_12_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_12_2
#define init_4_12_12_2
void FASTCALL call_4_12_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_12_3
#define init_4_12_12_3
void FASTCALL call_4_12_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_12_4
#define init_4_12_12_4
void FASTCALL call_4_12_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_12_5
#define init_4_12_12_5
void FASTCALL call_4_12_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_12_6
#define init_4_12_12_6
void FASTCALL call_4_12_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_12_7
#define init_4_12_12_7
void FASTCALL call_4_12_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_12_8
#define init_4_12_12_8
void FASTCALL call_4_12_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_12_9
#define init_4_12_12_9
void FASTCALL call_4_12_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_12_10
#define init_4_12_12_10
void FASTCALL call_4_12_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_12_11
#define init_4_12_12_11
void FASTCALL call_4_12_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_12_12
#define init_4_12_12_12
void FASTCALL call_4_12_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_12_13
#define init_4_12_12_13
void FASTCALL call_4_12_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_12_14
#define init_4_12_12_14
void FASTCALL call_4_12_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_12_15
#define init_4_12_12_15
void FASTCALL call_4_12_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_13_0
#define init_4_12_13_0
void FASTCALL call_4_12_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_13_1
#define init_4_12_13_1
void FASTCALL call_4_12_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_13_2
#define init_4_12_13_2
void FASTCALL call_4_12_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_13_3
#define init_4_12_13_3
void FASTCALL call_4_12_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_13_4
#define init_4_12_13_4
void FASTCALL call_4_12_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_13_5
#define init_4_12_13_5
void FASTCALL call_4_12_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_13_6
#define init_4_12_13_6
void FASTCALL call_4_12_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_13_7
#define init_4_12_13_7
void FASTCALL call_4_12_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_13_8
#define init_4_12_13_8
void FASTCALL call_4_12_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_13_9
#define init_4_12_13_9
void FASTCALL call_4_12_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_13_10
#define init_4_12_13_10
void FASTCALL call_4_12_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_13_11
#define init_4_12_13_11
void FASTCALL call_4_12_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_13_12
#define init_4_12_13_12
void FASTCALL call_4_12_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_13_13
#define init_4_12_13_13
void FASTCALL call_4_12_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_13_14
#define init_4_12_13_14
void FASTCALL call_4_12_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_13_15
#define init_4_12_13_15
void FASTCALL call_4_12_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_14_0
#define init_4_12_14_0
void FASTCALL call_4_12_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_14_1
#define init_4_12_14_1
void FASTCALL call_4_12_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_14_2
#define init_4_12_14_2
void FASTCALL call_4_12_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_14_3
#define init_4_12_14_3
void FASTCALL call_4_12_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_14_4
#define init_4_12_14_4
void FASTCALL call_4_12_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_14_5
#define init_4_12_14_5
void FASTCALL call_4_12_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_14_6
#define init_4_12_14_6
void FASTCALL call_4_12_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_14_7
#define init_4_12_14_7
void FASTCALL call_4_12_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_14_8
#define init_4_12_14_8
void FASTCALL call_4_12_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_14_9
#define init_4_12_14_9
void FASTCALL call_4_12_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_14_10
#define init_4_12_14_10
void FASTCALL call_4_12_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_14_11
#define init_4_12_14_11
void FASTCALL call_4_12_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_14_12
#define init_4_12_14_12
void FASTCALL call_4_12_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_14_13
#define init_4_12_14_13
void FASTCALL call_4_12_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_14_14
#define init_4_12_14_14
void FASTCALL call_4_12_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_14_15
#define init_4_12_14_15
void FASTCALL call_4_12_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_15_0
#define init_4_12_15_0
void FASTCALL call_4_12_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_15_1
#define init_4_12_15_1
void FASTCALL call_4_12_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_15_2
#define init_4_12_15_2
void FASTCALL call_4_12_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_15_3
#define init_4_12_15_3
void FASTCALL call_4_12_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_15_4
#define init_4_12_15_4
void FASTCALL call_4_12_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_15_5
#define init_4_12_15_5
void FASTCALL call_4_12_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_15_6
#define init_4_12_15_6
void FASTCALL call_4_12_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_15_7
#define init_4_12_15_7
void FASTCALL call_4_12_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_15_8
#define init_4_12_15_8
void FASTCALL call_4_12_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_15_9
#define init_4_12_15_9
void FASTCALL call_4_12_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_15_10
#define init_4_12_15_10
void FASTCALL call_4_12_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_15_11
#define init_4_12_15_11
void FASTCALL call_4_12_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_15_12
#define init_4_12_15_12
void FASTCALL call_4_12_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_15_13
#define init_4_12_15_13
void FASTCALL call_4_12_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_15_14
#define init_4_12_15_14
void FASTCALL call_4_12_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_12_15_15
#define init_4_12_15_15
void FASTCALL call_4_12_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_0_0
#define init_4_13_0_0
void FASTCALL call_4_13_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_0_1
#define init_4_13_0_1
void FASTCALL call_4_13_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_0_2
#define init_4_13_0_2
void FASTCALL call_4_13_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_0_3
#define init_4_13_0_3
void FASTCALL call_4_13_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_0_4
#define init_4_13_0_4
void FASTCALL call_4_13_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_0_5
#define init_4_13_0_5
void FASTCALL call_4_13_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_0_6
#define init_4_13_0_6
void FASTCALL call_4_13_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_0_7
#define init_4_13_0_7
void FASTCALL call_4_13_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_0_8
#define init_4_13_0_8
void FASTCALL call_4_13_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_0_9
#define init_4_13_0_9
void FASTCALL call_4_13_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_0_10
#define init_4_13_0_10
void FASTCALL call_4_13_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_0_11
#define init_4_13_0_11
void FASTCALL call_4_13_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_0_12
#define init_4_13_0_12
void FASTCALL call_4_13_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_0_13
#define init_4_13_0_13
void FASTCALL call_4_13_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_0_14
#define init_4_13_0_14
void FASTCALL call_4_13_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_0_15
#define init_4_13_0_15
void FASTCALL call_4_13_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_1_0
#define init_4_13_1_0
void FASTCALL call_4_13_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_1_1
#define init_4_13_1_1
void FASTCALL call_4_13_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_1_2
#define init_4_13_1_2
void FASTCALL call_4_13_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_1_3
#define init_4_13_1_3
void FASTCALL call_4_13_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_1_4
#define init_4_13_1_4
void FASTCALL call_4_13_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_1_5
#define init_4_13_1_5
void FASTCALL call_4_13_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_1_6
#define init_4_13_1_6
void FASTCALL call_4_13_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_1_7
#define init_4_13_1_7
void FASTCALL call_4_13_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_1_8
#define init_4_13_1_8
void FASTCALL call_4_13_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_1_9
#define init_4_13_1_9
void FASTCALL call_4_13_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_1_10
#define init_4_13_1_10
void FASTCALL call_4_13_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_1_11
#define init_4_13_1_11
void FASTCALL call_4_13_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_1_12
#define init_4_13_1_12
void FASTCALL call_4_13_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_1_13
#define init_4_13_1_13
void FASTCALL call_4_13_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_1_14
#define init_4_13_1_14
void FASTCALL call_4_13_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_1_15
#define init_4_13_1_15
void FASTCALL call_4_13_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_2_0
#define init_4_13_2_0
void FASTCALL call_4_13_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_2_1
#define init_4_13_2_1
void FASTCALL call_4_13_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_2_2
#define init_4_13_2_2
void FASTCALL call_4_13_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_2_3
#define init_4_13_2_3
void FASTCALL call_4_13_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_2_4
#define init_4_13_2_4
void FASTCALL call_4_13_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_2_5
#define init_4_13_2_5
void FASTCALL call_4_13_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_2_6
#define init_4_13_2_6
void FASTCALL call_4_13_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_2_7
#define init_4_13_2_7
void FASTCALL call_4_13_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_2_8
#define init_4_13_2_8
void FASTCALL call_4_13_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_2_9
#define init_4_13_2_9
void FASTCALL call_4_13_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_2_10
#define init_4_13_2_10
void FASTCALL call_4_13_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_2_11
#define init_4_13_2_11
void FASTCALL call_4_13_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_2_12
#define init_4_13_2_12
void FASTCALL call_4_13_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_2_13
#define init_4_13_2_13
void FASTCALL call_4_13_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_2_14
#define init_4_13_2_14
void FASTCALL call_4_13_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_2_15
#define init_4_13_2_15
void FASTCALL call_4_13_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_3_0
#define init_4_13_3_0
void FASTCALL call_4_13_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_3_1
#define init_4_13_3_1
void FASTCALL call_4_13_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_3_2
#define init_4_13_3_2
void FASTCALL call_4_13_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_3_3
#define init_4_13_3_3
void FASTCALL call_4_13_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_3_4
#define init_4_13_3_4
void FASTCALL call_4_13_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_3_5
#define init_4_13_3_5
void FASTCALL call_4_13_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_3_6
#define init_4_13_3_6
void FASTCALL call_4_13_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_3_7
#define init_4_13_3_7
void FASTCALL call_4_13_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_3_8
#define init_4_13_3_8
void FASTCALL call_4_13_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_3_9
#define init_4_13_3_9
void FASTCALL call_4_13_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_3_10
#define init_4_13_3_10
void FASTCALL call_4_13_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_3_11
#define init_4_13_3_11
void FASTCALL call_4_13_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_3_12
#define init_4_13_3_12
void FASTCALL call_4_13_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_3_13
#define init_4_13_3_13
void FASTCALL call_4_13_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_3_14
#define init_4_13_3_14
void FASTCALL call_4_13_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_3_15
#define init_4_13_3_15
void FASTCALL call_4_13_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_4_0
#define init_4_13_4_0
void FASTCALL call_4_13_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_4_1
#define init_4_13_4_1
void FASTCALL call_4_13_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_4_2
#define init_4_13_4_2
void FASTCALL call_4_13_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_4_3
#define init_4_13_4_3
void FASTCALL call_4_13_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_4_4
#define init_4_13_4_4
void FASTCALL call_4_13_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_4_5
#define init_4_13_4_5
void FASTCALL call_4_13_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_4_6
#define init_4_13_4_6
void FASTCALL call_4_13_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_4_7
#define init_4_13_4_7
void FASTCALL call_4_13_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_4_8
#define init_4_13_4_8
void FASTCALL call_4_13_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_4_9
#define init_4_13_4_9
void FASTCALL call_4_13_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_4_10
#define init_4_13_4_10
void FASTCALL call_4_13_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_4_11
#define init_4_13_4_11
void FASTCALL call_4_13_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_4_12
#define init_4_13_4_12
void FASTCALL call_4_13_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_4_13
#define init_4_13_4_13
void FASTCALL call_4_13_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_4_14
#define init_4_13_4_14
void FASTCALL call_4_13_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_4_15
#define init_4_13_4_15
void FASTCALL call_4_13_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_5_0
#define init_4_13_5_0
void FASTCALL call_4_13_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_5_1
#define init_4_13_5_1
void FASTCALL call_4_13_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_5_2
#define init_4_13_5_2
void FASTCALL call_4_13_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_5_3
#define init_4_13_5_3
void FASTCALL call_4_13_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_5_4
#define init_4_13_5_4
void FASTCALL call_4_13_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_5_5
#define init_4_13_5_5
void FASTCALL call_4_13_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_5_6
#define init_4_13_5_6
void FASTCALL call_4_13_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_5_7
#define init_4_13_5_7
void FASTCALL call_4_13_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_5_8
#define init_4_13_5_8
void FASTCALL call_4_13_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_5_9
#define init_4_13_5_9
void FASTCALL call_4_13_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_5_10
#define init_4_13_5_10
void FASTCALL call_4_13_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_5_11
#define init_4_13_5_11
void FASTCALL call_4_13_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_5_12
#define init_4_13_5_12
void FASTCALL call_4_13_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_5_13
#define init_4_13_5_13
void FASTCALL call_4_13_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_5_14
#define init_4_13_5_14
void FASTCALL call_4_13_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_5_15
#define init_4_13_5_15
void FASTCALL call_4_13_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_6_0
#define init_4_13_6_0
void FASTCALL call_4_13_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_6_1
#define init_4_13_6_1
void FASTCALL call_4_13_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_6_2
#define init_4_13_6_2
void FASTCALL call_4_13_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_6_3
#define init_4_13_6_3
void FASTCALL call_4_13_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_6_4
#define init_4_13_6_4
void FASTCALL call_4_13_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_6_5
#define init_4_13_6_5
void FASTCALL call_4_13_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_6_6
#define init_4_13_6_6
void FASTCALL call_4_13_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_6_7
#define init_4_13_6_7
void FASTCALL call_4_13_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_6_8
#define init_4_13_6_8
void FASTCALL call_4_13_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_6_9
#define init_4_13_6_9
void FASTCALL call_4_13_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_6_10
#define init_4_13_6_10
void FASTCALL call_4_13_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_6_11
#define init_4_13_6_11
void FASTCALL call_4_13_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_6_12
#define init_4_13_6_12
void FASTCALL call_4_13_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_6_13
#define init_4_13_6_13
void FASTCALL call_4_13_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_6_14
#define init_4_13_6_14
void FASTCALL call_4_13_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_6_15
#define init_4_13_6_15
void FASTCALL call_4_13_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_7_0
#define init_4_13_7_0
void FASTCALL call_4_13_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_7_1
#define init_4_13_7_1
void FASTCALL call_4_13_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_7_2
#define init_4_13_7_2
void FASTCALL call_4_13_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_7_3
#define init_4_13_7_3
void FASTCALL call_4_13_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_7_4
#define init_4_13_7_4
void FASTCALL call_4_13_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_7_5
#define init_4_13_7_5
void FASTCALL call_4_13_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_7_6
#define init_4_13_7_6
void FASTCALL call_4_13_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_7_7
#define init_4_13_7_7
void FASTCALL call_4_13_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_7_8
#define init_4_13_7_8
void FASTCALL call_4_13_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_7_9
#define init_4_13_7_9
void FASTCALL call_4_13_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_7_10
#define init_4_13_7_10
void FASTCALL call_4_13_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_7_11
#define init_4_13_7_11
void FASTCALL call_4_13_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_7_12
#define init_4_13_7_12
void FASTCALL call_4_13_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_7_13
#define init_4_13_7_13
void FASTCALL call_4_13_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_7_14
#define init_4_13_7_14
void FASTCALL call_4_13_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_7_15
#define init_4_13_7_15
void FASTCALL call_4_13_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_8_0
#define init_4_13_8_0
void FASTCALL call_4_13_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_8_1
#define init_4_13_8_1
void FASTCALL call_4_13_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_8_2
#define init_4_13_8_2
void FASTCALL call_4_13_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_8_3
#define init_4_13_8_3
void FASTCALL call_4_13_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_8_4
#define init_4_13_8_4
void FASTCALL call_4_13_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_8_5
#define init_4_13_8_5
void FASTCALL call_4_13_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_8_6
#define init_4_13_8_6
void FASTCALL call_4_13_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_8_7
#define init_4_13_8_7
void FASTCALL call_4_13_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_8_8
#define init_4_13_8_8
void FASTCALL call_4_13_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_8_9
#define init_4_13_8_9
void FASTCALL call_4_13_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_8_10
#define init_4_13_8_10
void FASTCALL call_4_13_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_8_11
#define init_4_13_8_11
void FASTCALL call_4_13_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_8_12
#define init_4_13_8_12
void FASTCALL call_4_13_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_8_13
#define init_4_13_8_13
void FASTCALL call_4_13_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_8_14
#define init_4_13_8_14
void FASTCALL call_4_13_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_8_15
#define init_4_13_8_15
void FASTCALL call_4_13_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_9_0
#define init_4_13_9_0
void FASTCALL call_4_13_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_9_1
#define init_4_13_9_1
void FASTCALL call_4_13_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_9_2
#define init_4_13_9_2
void FASTCALL call_4_13_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_9_3
#define init_4_13_9_3
void FASTCALL call_4_13_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_9_4
#define init_4_13_9_4
void FASTCALL call_4_13_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_9_5
#define init_4_13_9_5
void FASTCALL call_4_13_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_9_6
#define init_4_13_9_6
void FASTCALL call_4_13_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_9_7
#define init_4_13_9_7
void FASTCALL call_4_13_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_9_8
#define init_4_13_9_8
void FASTCALL call_4_13_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_9_9
#define init_4_13_9_9
void FASTCALL call_4_13_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_9_10
#define init_4_13_9_10
void FASTCALL call_4_13_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_9_11
#define init_4_13_9_11
void FASTCALL call_4_13_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_9_12
#define init_4_13_9_12
void FASTCALL call_4_13_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_9_13
#define init_4_13_9_13
void FASTCALL call_4_13_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_9_14
#define init_4_13_9_14
void FASTCALL call_4_13_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_9_15
#define init_4_13_9_15
void FASTCALL call_4_13_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_10_0
#define init_4_13_10_0
void FASTCALL call_4_13_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_10_1
#define init_4_13_10_1
void FASTCALL call_4_13_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_10_2
#define init_4_13_10_2
void FASTCALL call_4_13_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_10_3
#define init_4_13_10_3
void FASTCALL call_4_13_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_10_4
#define init_4_13_10_4
void FASTCALL call_4_13_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_10_5
#define init_4_13_10_5
void FASTCALL call_4_13_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_10_6
#define init_4_13_10_6
void FASTCALL call_4_13_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_10_7
#define init_4_13_10_7
void FASTCALL call_4_13_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_10_8
#define init_4_13_10_8
void FASTCALL call_4_13_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_10_9
#define init_4_13_10_9
void FASTCALL call_4_13_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_10_10
#define init_4_13_10_10
void FASTCALL call_4_13_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_10_11
#define init_4_13_10_11
void FASTCALL call_4_13_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_10_12
#define init_4_13_10_12
void FASTCALL call_4_13_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_10_13
#define init_4_13_10_13
void FASTCALL call_4_13_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_10_14
#define init_4_13_10_14
void FASTCALL call_4_13_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_10_15
#define init_4_13_10_15
void FASTCALL call_4_13_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_11_0
#define init_4_13_11_0
void FASTCALL call_4_13_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_11_1
#define init_4_13_11_1
void FASTCALL call_4_13_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_11_2
#define init_4_13_11_2
void FASTCALL call_4_13_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_11_3
#define init_4_13_11_3
void FASTCALL call_4_13_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_11_4
#define init_4_13_11_4
void FASTCALL call_4_13_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_11_5
#define init_4_13_11_5
void FASTCALL call_4_13_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_11_6
#define init_4_13_11_6
void FASTCALL call_4_13_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_11_7
#define init_4_13_11_7
void FASTCALL call_4_13_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_11_8
#define init_4_13_11_8
void FASTCALL call_4_13_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_11_9
#define init_4_13_11_9
void FASTCALL call_4_13_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_11_10
#define init_4_13_11_10
void FASTCALL call_4_13_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_11_11
#define init_4_13_11_11
void FASTCALL call_4_13_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_11_12
#define init_4_13_11_12
void FASTCALL call_4_13_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_11_13
#define init_4_13_11_13
void FASTCALL call_4_13_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_11_14
#define init_4_13_11_14
void FASTCALL call_4_13_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_11_15
#define init_4_13_11_15
void FASTCALL call_4_13_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_12_0
#define init_4_13_12_0
void FASTCALL call_4_13_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_12_1
#define init_4_13_12_1
void FASTCALL call_4_13_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_12_2
#define init_4_13_12_2
void FASTCALL call_4_13_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_12_3
#define init_4_13_12_3
void FASTCALL call_4_13_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_12_4
#define init_4_13_12_4
void FASTCALL call_4_13_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_12_5
#define init_4_13_12_5
void FASTCALL call_4_13_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_12_6
#define init_4_13_12_6
void FASTCALL call_4_13_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_12_7
#define init_4_13_12_7
void FASTCALL call_4_13_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_12_8
#define init_4_13_12_8
void FASTCALL call_4_13_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_12_9
#define init_4_13_12_9
void FASTCALL call_4_13_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_12_10
#define init_4_13_12_10
void FASTCALL call_4_13_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_12_11
#define init_4_13_12_11
void FASTCALL call_4_13_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_12_12
#define init_4_13_12_12
void FASTCALL call_4_13_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_12_13
#define init_4_13_12_13
void FASTCALL call_4_13_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_12_14
#define init_4_13_12_14
void FASTCALL call_4_13_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_12_15
#define init_4_13_12_15
void FASTCALL call_4_13_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_13_0
#define init_4_13_13_0
void FASTCALL call_4_13_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_13_1
#define init_4_13_13_1
void FASTCALL call_4_13_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_13_2
#define init_4_13_13_2
void FASTCALL call_4_13_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_13_3
#define init_4_13_13_3
void FASTCALL call_4_13_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_13_4
#define init_4_13_13_4
void FASTCALL call_4_13_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_13_5
#define init_4_13_13_5
void FASTCALL call_4_13_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_13_6
#define init_4_13_13_6
void FASTCALL call_4_13_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_13_7
#define init_4_13_13_7
void FASTCALL call_4_13_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_13_8
#define init_4_13_13_8
void FASTCALL call_4_13_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_13_9
#define init_4_13_13_9
void FASTCALL call_4_13_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_13_10
#define init_4_13_13_10
void FASTCALL call_4_13_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_13_11
#define init_4_13_13_11
void FASTCALL call_4_13_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_13_12
#define init_4_13_13_12
void FASTCALL call_4_13_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_13_13
#define init_4_13_13_13
void FASTCALL call_4_13_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_13_14
#define init_4_13_13_14
void FASTCALL call_4_13_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_13_15
#define init_4_13_13_15
void FASTCALL call_4_13_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_14_0
#define init_4_13_14_0
void FASTCALL call_4_13_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_14_1
#define init_4_13_14_1
void FASTCALL call_4_13_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_14_2
#define init_4_13_14_2
void FASTCALL call_4_13_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_14_3
#define init_4_13_14_3
void FASTCALL call_4_13_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_14_4
#define init_4_13_14_4
void FASTCALL call_4_13_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_14_5
#define init_4_13_14_5
void FASTCALL call_4_13_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_14_6
#define init_4_13_14_6
void FASTCALL call_4_13_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_14_7
#define init_4_13_14_7
void FASTCALL call_4_13_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_14_8
#define init_4_13_14_8
void FASTCALL call_4_13_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_14_9
#define init_4_13_14_9
void FASTCALL call_4_13_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_14_10
#define init_4_13_14_10
void FASTCALL call_4_13_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_14_11
#define init_4_13_14_11
void FASTCALL call_4_13_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_14_12
#define init_4_13_14_12
void FASTCALL call_4_13_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_14_13
#define init_4_13_14_13
void FASTCALL call_4_13_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_14_14
#define init_4_13_14_14
void FASTCALL call_4_13_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_14_15
#define init_4_13_14_15
void FASTCALL call_4_13_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_15_0
#define init_4_13_15_0
void FASTCALL call_4_13_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_15_1
#define init_4_13_15_1
void FASTCALL call_4_13_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_15_2
#define init_4_13_15_2
void FASTCALL call_4_13_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_15_3
#define init_4_13_15_3
void FASTCALL call_4_13_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_15_4
#define init_4_13_15_4
void FASTCALL call_4_13_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_15_5
#define init_4_13_15_5
void FASTCALL call_4_13_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_15_6
#define init_4_13_15_6
void FASTCALL call_4_13_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_15_7
#define init_4_13_15_7
void FASTCALL call_4_13_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_15_8
#define init_4_13_15_8
void FASTCALL call_4_13_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_15_9
#define init_4_13_15_9
void FASTCALL call_4_13_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_15_10
#define init_4_13_15_10
void FASTCALL call_4_13_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_15_11
#define init_4_13_15_11
void FASTCALL call_4_13_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_15_12
#define init_4_13_15_12
void FASTCALL call_4_13_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_15_13
#define init_4_13_15_13
void FASTCALL call_4_13_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_15_14
#define init_4_13_15_14
void FASTCALL call_4_13_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_13_15_15
#define init_4_13_15_15
void FASTCALL call_4_13_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_0_0
#define init_4_14_0_0
void FASTCALL call_4_14_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_0_1
#define init_4_14_0_1
void FASTCALL call_4_14_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_0_2
#define init_4_14_0_2
void FASTCALL call_4_14_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_0_3
#define init_4_14_0_3
void FASTCALL call_4_14_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_0_4
#define init_4_14_0_4
void FASTCALL call_4_14_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_0_5
#define init_4_14_0_5
void FASTCALL call_4_14_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_0_6
#define init_4_14_0_6
void FASTCALL call_4_14_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_0_7
#define init_4_14_0_7
void FASTCALL call_4_14_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_0_8
#define init_4_14_0_8
void FASTCALL call_4_14_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_0_9
#define init_4_14_0_9
void FASTCALL call_4_14_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_0_10
#define init_4_14_0_10
void FASTCALL call_4_14_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_0_11
#define init_4_14_0_11
void FASTCALL call_4_14_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_0_12
#define init_4_14_0_12
void FASTCALL call_4_14_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_0_13
#define init_4_14_0_13
void FASTCALL call_4_14_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_0_14
#define init_4_14_0_14
void FASTCALL call_4_14_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_0_15
#define init_4_14_0_15
void FASTCALL call_4_14_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_1_0
#define init_4_14_1_0
void FASTCALL call_4_14_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_1_1
#define init_4_14_1_1
void FASTCALL call_4_14_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_1_2
#define init_4_14_1_2
void FASTCALL call_4_14_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_1_3
#define init_4_14_1_3
void FASTCALL call_4_14_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_1_4
#define init_4_14_1_4
void FASTCALL call_4_14_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_1_5
#define init_4_14_1_5
void FASTCALL call_4_14_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_1_6
#define init_4_14_1_6
void FASTCALL call_4_14_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_1_7
#define init_4_14_1_7
void FASTCALL call_4_14_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_1_8
#define init_4_14_1_8
void FASTCALL call_4_14_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_1_9
#define init_4_14_1_9
void FASTCALL call_4_14_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_1_10
#define init_4_14_1_10
void FASTCALL call_4_14_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_1_11
#define init_4_14_1_11
void FASTCALL call_4_14_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_1_12
#define init_4_14_1_12
void FASTCALL call_4_14_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_1_13
#define init_4_14_1_13
void FASTCALL call_4_14_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_1_14
#define init_4_14_1_14
void FASTCALL call_4_14_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_1_15
#define init_4_14_1_15
void FASTCALL call_4_14_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_2_0
#define init_4_14_2_0
void FASTCALL call_4_14_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_2_1
#define init_4_14_2_1
void FASTCALL call_4_14_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_2_2
#define init_4_14_2_2
void FASTCALL call_4_14_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_2_3
#define init_4_14_2_3
void FASTCALL call_4_14_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_2_4
#define init_4_14_2_4
void FASTCALL call_4_14_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_2_5
#define init_4_14_2_5
void FASTCALL call_4_14_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_2_6
#define init_4_14_2_6
void FASTCALL call_4_14_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_2_7
#define init_4_14_2_7
void FASTCALL call_4_14_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_2_8
#define init_4_14_2_8
void FASTCALL call_4_14_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_2_9
#define init_4_14_2_9
void FASTCALL call_4_14_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_2_10
#define init_4_14_2_10
void FASTCALL call_4_14_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_2_11
#define init_4_14_2_11
void FASTCALL call_4_14_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_2_12
#define init_4_14_2_12
void FASTCALL call_4_14_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_2_13
#define init_4_14_2_13
void FASTCALL call_4_14_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_2_14
#define init_4_14_2_14
void FASTCALL call_4_14_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_2_15
#define init_4_14_2_15
void FASTCALL call_4_14_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_3_0
#define init_4_14_3_0
void FASTCALL call_4_14_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_3_1
#define init_4_14_3_1
void FASTCALL call_4_14_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_3_2
#define init_4_14_3_2
void FASTCALL call_4_14_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_3_3
#define init_4_14_3_3
void FASTCALL call_4_14_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_3_4
#define init_4_14_3_4
void FASTCALL call_4_14_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_3_5
#define init_4_14_3_5
void FASTCALL call_4_14_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_3_6
#define init_4_14_3_6
void FASTCALL call_4_14_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_3_7
#define init_4_14_3_7
void FASTCALL call_4_14_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_3_8
#define init_4_14_3_8
void FASTCALL call_4_14_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_3_9
#define init_4_14_3_9
void FASTCALL call_4_14_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_3_10
#define init_4_14_3_10
void FASTCALL call_4_14_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_3_11
#define init_4_14_3_11
void FASTCALL call_4_14_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_3_12
#define init_4_14_3_12
void FASTCALL call_4_14_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_3_13
#define init_4_14_3_13
void FASTCALL call_4_14_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_3_14
#define init_4_14_3_14
void FASTCALL call_4_14_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_3_15
#define init_4_14_3_15
void FASTCALL call_4_14_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_4_0
#define init_4_14_4_0
void FASTCALL call_4_14_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_4_1
#define init_4_14_4_1
void FASTCALL call_4_14_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_4_2
#define init_4_14_4_2
void FASTCALL call_4_14_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_4_3
#define init_4_14_4_3
void FASTCALL call_4_14_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_4_4
#define init_4_14_4_4
void FASTCALL call_4_14_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_4_5
#define init_4_14_4_5
void FASTCALL call_4_14_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_4_6
#define init_4_14_4_6
void FASTCALL call_4_14_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_4_7
#define init_4_14_4_7
void FASTCALL call_4_14_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_4_8
#define init_4_14_4_8
void FASTCALL call_4_14_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_4_9
#define init_4_14_4_9
void FASTCALL call_4_14_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_4_10
#define init_4_14_4_10
void FASTCALL call_4_14_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_4_11
#define init_4_14_4_11
void FASTCALL call_4_14_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_4_12
#define init_4_14_4_12
void FASTCALL call_4_14_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_4_13
#define init_4_14_4_13
void FASTCALL call_4_14_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_4_14
#define init_4_14_4_14
void FASTCALL call_4_14_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_4_15
#define init_4_14_4_15
void FASTCALL call_4_14_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_5_0
#define init_4_14_5_0
void FASTCALL call_4_14_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_5_1
#define init_4_14_5_1
void FASTCALL call_4_14_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_5_2
#define init_4_14_5_2
void FASTCALL call_4_14_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_5_3
#define init_4_14_5_3
void FASTCALL call_4_14_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_5_4
#define init_4_14_5_4
void FASTCALL call_4_14_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_5_5
#define init_4_14_5_5
void FASTCALL call_4_14_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_5_6
#define init_4_14_5_6
void FASTCALL call_4_14_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_5_7
#define init_4_14_5_7
void FASTCALL call_4_14_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_5_8
#define init_4_14_5_8
void FASTCALL call_4_14_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_5_9
#define init_4_14_5_9
void FASTCALL call_4_14_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_5_10
#define init_4_14_5_10
void FASTCALL call_4_14_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_5_11
#define init_4_14_5_11
void FASTCALL call_4_14_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_5_12
#define init_4_14_5_12
void FASTCALL call_4_14_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_5_13
#define init_4_14_5_13
void FASTCALL call_4_14_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_5_14
#define init_4_14_5_14
void FASTCALL call_4_14_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_5_15
#define init_4_14_5_15
void FASTCALL call_4_14_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_6_0
#define init_4_14_6_0
void FASTCALL call_4_14_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_6_1
#define init_4_14_6_1
void FASTCALL call_4_14_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_6_2
#define init_4_14_6_2
void FASTCALL call_4_14_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_6_3
#define init_4_14_6_3
void FASTCALL call_4_14_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_6_4
#define init_4_14_6_4
void FASTCALL call_4_14_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_6_5
#define init_4_14_6_5
void FASTCALL call_4_14_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_6_6
#define init_4_14_6_6
void FASTCALL call_4_14_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_6_7
#define init_4_14_6_7
void FASTCALL call_4_14_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_6_8
#define init_4_14_6_8
void FASTCALL call_4_14_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_6_9
#define init_4_14_6_9
void FASTCALL call_4_14_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_6_10
#define init_4_14_6_10
void FASTCALL call_4_14_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_6_11
#define init_4_14_6_11
void FASTCALL call_4_14_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_6_12
#define init_4_14_6_12
void FASTCALL call_4_14_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_6_13
#define init_4_14_6_13
void FASTCALL call_4_14_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_6_14
#define init_4_14_6_14
void FASTCALL call_4_14_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_6_15
#define init_4_14_6_15
void FASTCALL call_4_14_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_7_0
#define init_4_14_7_0
void FASTCALL call_4_14_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_7_1
#define init_4_14_7_1
void FASTCALL call_4_14_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_7_2
#define init_4_14_7_2
void FASTCALL call_4_14_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_7_3
#define init_4_14_7_3
void FASTCALL call_4_14_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_7_4
#define init_4_14_7_4
void FASTCALL call_4_14_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_7_5
#define init_4_14_7_5
void FASTCALL call_4_14_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_7_6
#define init_4_14_7_6
void FASTCALL call_4_14_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_7_7
#define init_4_14_7_7
void FASTCALL call_4_14_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_7_8
#define init_4_14_7_8
void FASTCALL call_4_14_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_7_9
#define init_4_14_7_9
void FASTCALL call_4_14_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_7_10
#define init_4_14_7_10
void FASTCALL call_4_14_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_7_11
#define init_4_14_7_11
void FASTCALL call_4_14_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_7_12
#define init_4_14_7_12
void FASTCALL call_4_14_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_7_13
#define init_4_14_7_13
void FASTCALL call_4_14_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_7_14
#define init_4_14_7_14
void FASTCALL call_4_14_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_7_15
#define init_4_14_7_15
void FASTCALL call_4_14_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_8_0
#define init_4_14_8_0
void FASTCALL call_4_14_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_8_1
#define init_4_14_8_1
void FASTCALL call_4_14_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_8_2
#define init_4_14_8_2
void FASTCALL call_4_14_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_8_3
#define init_4_14_8_3
void FASTCALL call_4_14_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_8_4
#define init_4_14_8_4
void FASTCALL call_4_14_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_8_5
#define init_4_14_8_5
void FASTCALL call_4_14_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_8_6
#define init_4_14_8_6
void FASTCALL call_4_14_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_8_7
#define init_4_14_8_7
void FASTCALL call_4_14_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_8_8
#define init_4_14_8_8
void FASTCALL call_4_14_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_8_9
#define init_4_14_8_9
void FASTCALL call_4_14_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_8_10
#define init_4_14_8_10
void FASTCALL call_4_14_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_8_11
#define init_4_14_8_11
void FASTCALL call_4_14_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_8_12
#define init_4_14_8_12
void FASTCALL call_4_14_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_8_13
#define init_4_14_8_13
void FASTCALL call_4_14_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_8_14
#define init_4_14_8_14
void FASTCALL call_4_14_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_8_15
#define init_4_14_8_15
void FASTCALL call_4_14_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_9_0
#define init_4_14_9_0
void FASTCALL call_4_14_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_9_1
#define init_4_14_9_1
void FASTCALL call_4_14_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_9_2
#define init_4_14_9_2
void FASTCALL call_4_14_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_9_3
#define init_4_14_9_3
void FASTCALL call_4_14_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_9_4
#define init_4_14_9_4
void FASTCALL call_4_14_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_9_5
#define init_4_14_9_5
void FASTCALL call_4_14_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_9_6
#define init_4_14_9_6
void FASTCALL call_4_14_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_9_7
#define init_4_14_9_7
void FASTCALL call_4_14_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_9_8
#define init_4_14_9_8
void FASTCALL call_4_14_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_9_9
#define init_4_14_9_9
void FASTCALL call_4_14_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_9_10
#define init_4_14_9_10
void FASTCALL call_4_14_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_9_11
#define init_4_14_9_11
void FASTCALL call_4_14_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_9_12
#define init_4_14_9_12
void FASTCALL call_4_14_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_9_13
#define init_4_14_9_13
void FASTCALL call_4_14_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_9_14
#define init_4_14_9_14
void FASTCALL call_4_14_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_9_15
#define init_4_14_9_15
void FASTCALL call_4_14_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_10_0
#define init_4_14_10_0
void FASTCALL call_4_14_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_10_1
#define init_4_14_10_1
void FASTCALL call_4_14_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_10_2
#define init_4_14_10_2
void FASTCALL call_4_14_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_10_3
#define init_4_14_10_3
void FASTCALL call_4_14_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_10_4
#define init_4_14_10_4
void FASTCALL call_4_14_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_10_5
#define init_4_14_10_5
void FASTCALL call_4_14_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_10_6
#define init_4_14_10_6
void FASTCALL call_4_14_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_10_7
#define init_4_14_10_7
void FASTCALL call_4_14_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_10_8
#define init_4_14_10_8
void FASTCALL call_4_14_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_10_9
#define init_4_14_10_9
void FASTCALL call_4_14_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_10_10
#define init_4_14_10_10
void FASTCALL call_4_14_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_10_11
#define init_4_14_10_11
void FASTCALL call_4_14_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_10_12
#define init_4_14_10_12
void FASTCALL call_4_14_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_10_13
#define init_4_14_10_13
void FASTCALL call_4_14_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_10_14
#define init_4_14_10_14
void FASTCALL call_4_14_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_10_15
#define init_4_14_10_15
void FASTCALL call_4_14_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_11_0
#define init_4_14_11_0
void FASTCALL call_4_14_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_11_1
#define init_4_14_11_1
void FASTCALL call_4_14_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_11_2
#define init_4_14_11_2
void FASTCALL call_4_14_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_11_3
#define init_4_14_11_3
void FASTCALL call_4_14_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_11_4
#define init_4_14_11_4
void FASTCALL call_4_14_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_11_5
#define init_4_14_11_5
void FASTCALL call_4_14_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_11_6
#define init_4_14_11_6
void FASTCALL call_4_14_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_11_7
#define init_4_14_11_7
void FASTCALL call_4_14_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_11_8
#define init_4_14_11_8
void FASTCALL call_4_14_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_11_9
#define init_4_14_11_9
void FASTCALL call_4_14_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_11_10
#define init_4_14_11_10
void FASTCALL call_4_14_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_11_11
#define init_4_14_11_11
void FASTCALL call_4_14_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_11_12
#define init_4_14_11_12
void FASTCALL call_4_14_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_11_13
#define init_4_14_11_13
void FASTCALL call_4_14_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_11_14
#define init_4_14_11_14
void FASTCALL call_4_14_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_11_15
#define init_4_14_11_15
void FASTCALL call_4_14_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_12_0
#define init_4_14_12_0
void FASTCALL call_4_14_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_12_1
#define init_4_14_12_1
void FASTCALL call_4_14_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_12_2
#define init_4_14_12_2
void FASTCALL call_4_14_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_12_3
#define init_4_14_12_3
void FASTCALL call_4_14_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_12_4
#define init_4_14_12_4
void FASTCALL call_4_14_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_12_5
#define init_4_14_12_5
void FASTCALL call_4_14_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_12_6
#define init_4_14_12_6
void FASTCALL call_4_14_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_12_7
#define init_4_14_12_7
void FASTCALL call_4_14_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_12_8
#define init_4_14_12_8
void FASTCALL call_4_14_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_12_9
#define init_4_14_12_9
void FASTCALL call_4_14_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_12_10
#define init_4_14_12_10
void FASTCALL call_4_14_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_12_11
#define init_4_14_12_11
void FASTCALL call_4_14_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_12_12
#define init_4_14_12_12
void FASTCALL call_4_14_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_12_13
#define init_4_14_12_13
void FASTCALL call_4_14_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_12_14
#define init_4_14_12_14
void FASTCALL call_4_14_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_12_15
#define init_4_14_12_15
void FASTCALL call_4_14_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_13_0
#define init_4_14_13_0
void FASTCALL call_4_14_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_13_1
#define init_4_14_13_1
void FASTCALL call_4_14_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_13_2
#define init_4_14_13_2
void FASTCALL call_4_14_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_13_3
#define init_4_14_13_3
void FASTCALL call_4_14_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_13_4
#define init_4_14_13_4
void FASTCALL call_4_14_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_13_5
#define init_4_14_13_5
void FASTCALL call_4_14_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_13_6
#define init_4_14_13_6
void FASTCALL call_4_14_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_13_7
#define init_4_14_13_7
void FASTCALL call_4_14_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_13_8
#define init_4_14_13_8
void FASTCALL call_4_14_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_13_9
#define init_4_14_13_9
void FASTCALL call_4_14_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_13_10
#define init_4_14_13_10
void FASTCALL call_4_14_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_13_11
#define init_4_14_13_11
void FASTCALL call_4_14_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_13_12
#define init_4_14_13_12
void FASTCALL call_4_14_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_13_13
#define init_4_14_13_13
void FASTCALL call_4_14_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_13_14
#define init_4_14_13_14
void FASTCALL call_4_14_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_13_15
#define init_4_14_13_15
void FASTCALL call_4_14_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_14_0
#define init_4_14_14_0
void FASTCALL call_4_14_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_14_1
#define init_4_14_14_1
void FASTCALL call_4_14_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_14_2
#define init_4_14_14_2
void FASTCALL call_4_14_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_14_3
#define init_4_14_14_3
void FASTCALL call_4_14_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_14_4
#define init_4_14_14_4
void FASTCALL call_4_14_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_14_5
#define init_4_14_14_5
void FASTCALL call_4_14_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_14_6
#define init_4_14_14_6
void FASTCALL call_4_14_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_14_7
#define init_4_14_14_7
void FASTCALL call_4_14_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_14_8
#define init_4_14_14_8
void FASTCALL call_4_14_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_14_9
#define init_4_14_14_9
void FASTCALL call_4_14_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_14_10
#define init_4_14_14_10
void FASTCALL call_4_14_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_14_11
#define init_4_14_14_11
void FASTCALL call_4_14_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_14_12
#define init_4_14_14_12
void FASTCALL call_4_14_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_14_13
#define init_4_14_14_13
void FASTCALL call_4_14_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_14_14
#define init_4_14_14_14
void FASTCALL call_4_14_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_14_15
#define init_4_14_14_15
void FASTCALL call_4_14_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_15_0
#define init_4_14_15_0
void FASTCALL call_4_14_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_15_1
#define init_4_14_15_1
void FASTCALL call_4_14_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_15_2
#define init_4_14_15_2
void FASTCALL call_4_14_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_15_3
#define init_4_14_15_3
void FASTCALL call_4_14_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_15_4
#define init_4_14_15_4
void FASTCALL call_4_14_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_15_5
#define init_4_14_15_5
void FASTCALL call_4_14_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_15_6
#define init_4_14_15_6
void FASTCALL call_4_14_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_15_7
#define init_4_14_15_7
void FASTCALL call_4_14_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_15_8
#define init_4_14_15_8
void FASTCALL call_4_14_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_15_9
#define init_4_14_15_9
void FASTCALL call_4_14_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_15_10
#define init_4_14_15_10
void FASTCALL call_4_14_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_15_11
#define init_4_14_15_11
void FASTCALL call_4_14_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_15_12
#define init_4_14_15_12
void FASTCALL call_4_14_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_15_13
#define init_4_14_15_13
void FASTCALL call_4_14_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_15_14
#define init_4_14_15_14
void FASTCALL call_4_14_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_14_15_15
#define init_4_14_15_15
void FASTCALL call_4_14_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_0_0
#define init_4_15_0_0
void FASTCALL call_4_15_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_0_1
#define init_4_15_0_1
void FASTCALL call_4_15_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_0_2
#define init_4_15_0_2
void FASTCALL call_4_15_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_0_3
#define init_4_15_0_3
void FASTCALL call_4_15_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_0_4
#define init_4_15_0_4
void FASTCALL call_4_15_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_0_5
#define init_4_15_0_5
void FASTCALL call_4_15_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_0_6
#define init_4_15_0_6
void FASTCALL call_4_15_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_0_7
#define init_4_15_0_7
void FASTCALL call_4_15_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_0_8
#define init_4_15_0_8
void FASTCALL call_4_15_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_0_9
#define init_4_15_0_9
void FASTCALL call_4_15_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_0_10
#define init_4_15_0_10
void FASTCALL call_4_15_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_0_11
#define init_4_15_0_11
void FASTCALL call_4_15_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_0_12
#define init_4_15_0_12
void FASTCALL call_4_15_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_0_13
#define init_4_15_0_13
void FASTCALL call_4_15_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_0_14
#define init_4_15_0_14
void FASTCALL call_4_15_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_0_15
#define init_4_15_0_15
void FASTCALL call_4_15_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_1_0
#define init_4_15_1_0
void FASTCALL call_4_15_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_1_1
#define init_4_15_1_1
void FASTCALL call_4_15_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_1_2
#define init_4_15_1_2
void FASTCALL call_4_15_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_1_3
#define init_4_15_1_3
void FASTCALL call_4_15_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_1_4
#define init_4_15_1_4
void FASTCALL call_4_15_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_1_5
#define init_4_15_1_5
void FASTCALL call_4_15_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_1_6
#define init_4_15_1_6
void FASTCALL call_4_15_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_1_7
#define init_4_15_1_7
void FASTCALL call_4_15_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_1_8
#define init_4_15_1_8
void FASTCALL call_4_15_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_1_9
#define init_4_15_1_9
void FASTCALL call_4_15_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_1_10
#define init_4_15_1_10
void FASTCALL call_4_15_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_1_11
#define init_4_15_1_11
void FASTCALL call_4_15_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_1_12
#define init_4_15_1_12
void FASTCALL call_4_15_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_1_13
#define init_4_15_1_13
void FASTCALL call_4_15_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_1_14
#define init_4_15_1_14
void FASTCALL call_4_15_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_1_15
#define init_4_15_1_15
void FASTCALL call_4_15_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_2_0
#define init_4_15_2_0
void FASTCALL call_4_15_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_2_1
#define init_4_15_2_1
void FASTCALL call_4_15_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_2_2
#define init_4_15_2_2
void FASTCALL call_4_15_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_2_3
#define init_4_15_2_3
void FASTCALL call_4_15_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_2_4
#define init_4_15_2_4
void FASTCALL call_4_15_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_2_5
#define init_4_15_2_5
void FASTCALL call_4_15_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_2_6
#define init_4_15_2_6
void FASTCALL call_4_15_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_2_7
#define init_4_15_2_7
void FASTCALL call_4_15_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_2_8
#define init_4_15_2_8
void FASTCALL call_4_15_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_2_9
#define init_4_15_2_9
void FASTCALL call_4_15_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_2_10
#define init_4_15_2_10
void FASTCALL call_4_15_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_2_11
#define init_4_15_2_11
void FASTCALL call_4_15_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_2_12
#define init_4_15_2_12
void FASTCALL call_4_15_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_2_13
#define init_4_15_2_13
void FASTCALL call_4_15_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_2_14
#define init_4_15_2_14
void FASTCALL call_4_15_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_2_15
#define init_4_15_2_15
void FASTCALL call_4_15_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_3_0
#define init_4_15_3_0
void FASTCALL call_4_15_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_3_1
#define init_4_15_3_1
void FASTCALL call_4_15_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_3_2
#define init_4_15_3_2
void FASTCALL call_4_15_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_3_3
#define init_4_15_3_3
void FASTCALL call_4_15_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_3_4
#define init_4_15_3_4
void FASTCALL call_4_15_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_3_5
#define init_4_15_3_5
void FASTCALL call_4_15_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_3_6
#define init_4_15_3_6
void FASTCALL call_4_15_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_3_7
#define init_4_15_3_7
void FASTCALL call_4_15_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_3_8
#define init_4_15_3_8
void FASTCALL call_4_15_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_3_9
#define init_4_15_3_9
void FASTCALL call_4_15_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_3_10
#define init_4_15_3_10
void FASTCALL call_4_15_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_3_11
#define init_4_15_3_11
void FASTCALL call_4_15_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_3_12
#define init_4_15_3_12
void FASTCALL call_4_15_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_3_13
#define init_4_15_3_13
void FASTCALL call_4_15_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_3_14
#define init_4_15_3_14
void FASTCALL call_4_15_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_3_15
#define init_4_15_3_15
void FASTCALL call_4_15_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_4_0
#define init_4_15_4_0
void FASTCALL call_4_15_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_4_1
#define init_4_15_4_1
void FASTCALL call_4_15_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_4_2
#define init_4_15_4_2
void FASTCALL call_4_15_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_4_3
#define init_4_15_4_3
void FASTCALL call_4_15_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_4_4
#define init_4_15_4_4
void FASTCALL call_4_15_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_4_5
#define init_4_15_4_5
void FASTCALL call_4_15_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_4_6
#define init_4_15_4_6
void FASTCALL call_4_15_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_4_7
#define init_4_15_4_7
void FASTCALL call_4_15_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_4_8
#define init_4_15_4_8
void FASTCALL call_4_15_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_4_9
#define init_4_15_4_9
void FASTCALL call_4_15_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_4_10
#define init_4_15_4_10
void FASTCALL call_4_15_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_4_11
#define init_4_15_4_11
void FASTCALL call_4_15_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_4_12
#define init_4_15_4_12
void FASTCALL call_4_15_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_4_13
#define init_4_15_4_13
void FASTCALL call_4_15_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_4_14
#define init_4_15_4_14
void FASTCALL call_4_15_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_4_15
#define init_4_15_4_15
void FASTCALL call_4_15_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_5_0
#define init_4_15_5_0
void FASTCALL call_4_15_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_5_1
#define init_4_15_5_1
void FASTCALL call_4_15_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_5_2
#define init_4_15_5_2
void FASTCALL call_4_15_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_5_3
#define init_4_15_5_3
void FASTCALL call_4_15_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_5_4
#define init_4_15_5_4
void FASTCALL call_4_15_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_5_5
#define init_4_15_5_5
void FASTCALL call_4_15_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_5_6
#define init_4_15_5_6
void FASTCALL call_4_15_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_5_7
#define init_4_15_5_7
void FASTCALL call_4_15_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_5_8
#define init_4_15_5_8
void FASTCALL call_4_15_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_5_9
#define init_4_15_5_9
void FASTCALL call_4_15_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_5_10
#define init_4_15_5_10
void FASTCALL call_4_15_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_5_11
#define init_4_15_5_11
void FASTCALL call_4_15_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_5_12
#define init_4_15_5_12
void FASTCALL call_4_15_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_5_13
#define init_4_15_5_13
void FASTCALL call_4_15_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_5_14
#define init_4_15_5_14
void FASTCALL call_4_15_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_5_15
#define init_4_15_5_15
void FASTCALL call_4_15_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_6_0
#define init_4_15_6_0
void FASTCALL call_4_15_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_6_1
#define init_4_15_6_1
void FASTCALL call_4_15_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_6_2
#define init_4_15_6_2
void FASTCALL call_4_15_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_6_3
#define init_4_15_6_3
void FASTCALL call_4_15_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_6_4
#define init_4_15_6_4
void FASTCALL call_4_15_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_6_5
#define init_4_15_6_5
void FASTCALL call_4_15_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_6_6
#define init_4_15_6_6
void FASTCALL call_4_15_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_6_7
#define init_4_15_6_7
void FASTCALL call_4_15_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_6_8
#define init_4_15_6_8
void FASTCALL call_4_15_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_6_9
#define init_4_15_6_9
void FASTCALL call_4_15_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_6_10
#define init_4_15_6_10
void FASTCALL call_4_15_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_6_11
#define init_4_15_6_11
void FASTCALL call_4_15_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_6_12
#define init_4_15_6_12
void FASTCALL call_4_15_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_6_13
#define init_4_15_6_13
void FASTCALL call_4_15_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_6_14
#define init_4_15_6_14
void FASTCALL call_4_15_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_6_15
#define init_4_15_6_15
void FASTCALL call_4_15_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_7_0
#define init_4_15_7_0
void FASTCALL call_4_15_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_7_1
#define init_4_15_7_1
void FASTCALL call_4_15_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_7_2
#define init_4_15_7_2
void FASTCALL call_4_15_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_7_3
#define init_4_15_7_3
void FASTCALL call_4_15_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_7_4
#define init_4_15_7_4
void FASTCALL call_4_15_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_7_5
#define init_4_15_7_5
void FASTCALL call_4_15_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_7_6
#define init_4_15_7_6
void FASTCALL call_4_15_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_7_7
#define init_4_15_7_7
void FASTCALL call_4_15_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_7_8
#define init_4_15_7_8
void FASTCALL call_4_15_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_7_9
#define init_4_15_7_9
void FASTCALL call_4_15_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_7_10
#define init_4_15_7_10
void FASTCALL call_4_15_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_7_11
#define init_4_15_7_11
void FASTCALL call_4_15_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_7_12
#define init_4_15_7_12
void FASTCALL call_4_15_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_7_13
#define init_4_15_7_13
void FASTCALL call_4_15_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_7_14
#define init_4_15_7_14
void FASTCALL call_4_15_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_7_15
#define init_4_15_7_15
void FASTCALL call_4_15_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_8_0
#define init_4_15_8_0
void FASTCALL call_4_15_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_8_1
#define init_4_15_8_1
void FASTCALL call_4_15_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_8_2
#define init_4_15_8_2
void FASTCALL call_4_15_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_8_3
#define init_4_15_8_3
void FASTCALL call_4_15_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_8_4
#define init_4_15_8_4
void FASTCALL call_4_15_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_8_5
#define init_4_15_8_5
void FASTCALL call_4_15_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_8_6
#define init_4_15_8_6
void FASTCALL call_4_15_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_8_7
#define init_4_15_8_7
void FASTCALL call_4_15_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_8_8
#define init_4_15_8_8
void FASTCALL call_4_15_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_8_9
#define init_4_15_8_9
void FASTCALL call_4_15_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_8_10
#define init_4_15_8_10
void FASTCALL call_4_15_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_8_11
#define init_4_15_8_11
void FASTCALL call_4_15_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_8_12
#define init_4_15_8_12
void FASTCALL call_4_15_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_8_13
#define init_4_15_8_13
void FASTCALL call_4_15_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_8_14
#define init_4_15_8_14
void FASTCALL call_4_15_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_8_15
#define init_4_15_8_15
void FASTCALL call_4_15_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_9_0
#define init_4_15_9_0
void FASTCALL call_4_15_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_9_1
#define init_4_15_9_1
void FASTCALL call_4_15_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_9_2
#define init_4_15_9_2
void FASTCALL call_4_15_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_9_3
#define init_4_15_9_3
void FASTCALL call_4_15_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_9_4
#define init_4_15_9_4
void FASTCALL call_4_15_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_9_5
#define init_4_15_9_5
void FASTCALL call_4_15_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_9_6
#define init_4_15_9_6
void FASTCALL call_4_15_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_9_7
#define init_4_15_9_7
void FASTCALL call_4_15_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_9_8
#define init_4_15_9_8
void FASTCALL call_4_15_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_9_9
#define init_4_15_9_9
void FASTCALL call_4_15_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_9_10
#define init_4_15_9_10
void FASTCALL call_4_15_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_9_11
#define init_4_15_9_11
void FASTCALL call_4_15_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_9_12
#define init_4_15_9_12
void FASTCALL call_4_15_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_9_13
#define init_4_15_9_13
void FASTCALL call_4_15_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_9_14
#define init_4_15_9_14
void FASTCALL call_4_15_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_9_15
#define init_4_15_9_15
void FASTCALL call_4_15_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_10_0
#define init_4_15_10_0
void FASTCALL call_4_15_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_10_1
#define init_4_15_10_1
void FASTCALL call_4_15_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_10_2
#define init_4_15_10_2
void FASTCALL call_4_15_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_10_3
#define init_4_15_10_3
void FASTCALL call_4_15_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_10_4
#define init_4_15_10_4
void FASTCALL call_4_15_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_10_5
#define init_4_15_10_5
void FASTCALL call_4_15_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_10_6
#define init_4_15_10_6
void FASTCALL call_4_15_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_10_7
#define init_4_15_10_7
void FASTCALL call_4_15_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_10_8
#define init_4_15_10_8
void FASTCALL call_4_15_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_10_9
#define init_4_15_10_9
void FASTCALL call_4_15_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_10_10
#define init_4_15_10_10
void FASTCALL call_4_15_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_10_11
#define init_4_15_10_11
void FASTCALL call_4_15_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_10_12
#define init_4_15_10_12
void FASTCALL call_4_15_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_10_13
#define init_4_15_10_13
void FASTCALL call_4_15_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_10_14
#define init_4_15_10_14
void FASTCALL call_4_15_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_10_15
#define init_4_15_10_15
void FASTCALL call_4_15_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_11_0
#define init_4_15_11_0
void FASTCALL call_4_15_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_11_1
#define init_4_15_11_1
void FASTCALL call_4_15_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_11_2
#define init_4_15_11_2
void FASTCALL call_4_15_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_11_3
#define init_4_15_11_3
void FASTCALL call_4_15_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_11_4
#define init_4_15_11_4
void FASTCALL call_4_15_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_11_5
#define init_4_15_11_5
void FASTCALL call_4_15_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_11_6
#define init_4_15_11_6
void FASTCALL call_4_15_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_11_7
#define init_4_15_11_7
void FASTCALL call_4_15_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_11_8
#define init_4_15_11_8
void FASTCALL call_4_15_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_11_9
#define init_4_15_11_9
void FASTCALL call_4_15_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_11_10
#define init_4_15_11_10
void FASTCALL call_4_15_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_11_11
#define init_4_15_11_11
void FASTCALL call_4_15_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_11_12
#define init_4_15_11_12
void FASTCALL call_4_15_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_11_13
#define init_4_15_11_13
void FASTCALL call_4_15_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_11_14
#define init_4_15_11_14
void FASTCALL call_4_15_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_11_15
#define init_4_15_11_15
void FASTCALL call_4_15_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_12_0
#define init_4_15_12_0
void FASTCALL call_4_15_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_12_1
#define init_4_15_12_1
void FASTCALL call_4_15_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_12_2
#define init_4_15_12_2
void FASTCALL call_4_15_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_12_3
#define init_4_15_12_3
void FASTCALL call_4_15_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_12_4
#define init_4_15_12_4
void FASTCALL call_4_15_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_12_5
#define init_4_15_12_5
void FASTCALL call_4_15_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_12_6
#define init_4_15_12_6
void FASTCALL call_4_15_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_12_7
#define init_4_15_12_7
void FASTCALL call_4_15_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_12_8
#define init_4_15_12_8
void FASTCALL call_4_15_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_12_9
#define init_4_15_12_9
void FASTCALL call_4_15_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_12_10
#define init_4_15_12_10
void FASTCALL call_4_15_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_12_11
#define init_4_15_12_11
void FASTCALL call_4_15_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_12_12
#define init_4_15_12_12
void FASTCALL call_4_15_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_12_13
#define init_4_15_12_13
void FASTCALL call_4_15_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_12_14
#define init_4_15_12_14
void FASTCALL call_4_15_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_12_15
#define init_4_15_12_15
void FASTCALL call_4_15_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_13_0
#define init_4_15_13_0
void FASTCALL call_4_15_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_13_1
#define init_4_15_13_1
void FASTCALL call_4_15_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_13_2
#define init_4_15_13_2
void FASTCALL call_4_15_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_13_3
#define init_4_15_13_3
void FASTCALL call_4_15_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_13_4
#define init_4_15_13_4
void FASTCALL call_4_15_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_13_5
#define init_4_15_13_5
void FASTCALL call_4_15_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_13_6
#define init_4_15_13_6
void FASTCALL call_4_15_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_13_7
#define init_4_15_13_7
void FASTCALL call_4_15_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_13_8
#define init_4_15_13_8
void FASTCALL call_4_15_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_13_9
#define init_4_15_13_9
void FASTCALL call_4_15_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_13_10
#define init_4_15_13_10
void FASTCALL call_4_15_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_13_11
#define init_4_15_13_11
void FASTCALL call_4_15_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_13_12
#define init_4_15_13_12
void FASTCALL call_4_15_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_13_13
#define init_4_15_13_13
void FASTCALL call_4_15_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_13_14
#define init_4_15_13_14
void FASTCALL call_4_15_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_13_15
#define init_4_15_13_15
void FASTCALL call_4_15_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_14_0
#define init_4_15_14_0
void FASTCALL call_4_15_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_14_1
#define init_4_15_14_1
void FASTCALL call_4_15_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_14_2
#define init_4_15_14_2
void FASTCALL call_4_15_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_14_3
#define init_4_15_14_3
void FASTCALL call_4_15_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_14_4
#define init_4_15_14_4
void FASTCALL call_4_15_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_14_5
#define init_4_15_14_5
void FASTCALL call_4_15_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_14_6
#define init_4_15_14_6
void FASTCALL call_4_15_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_14_7
#define init_4_15_14_7
void FASTCALL call_4_15_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_14_8
#define init_4_15_14_8
void FASTCALL call_4_15_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_14_9
#define init_4_15_14_9
void FASTCALL call_4_15_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_14_10
#define init_4_15_14_10
void FASTCALL call_4_15_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_14_11
#define init_4_15_14_11
void FASTCALL call_4_15_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_14_12
#define init_4_15_14_12
void FASTCALL call_4_15_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_14_13
#define init_4_15_14_13
void FASTCALL call_4_15_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_14_14
#define init_4_15_14_14
void FASTCALL call_4_15_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_14_15
#define init_4_15_14_15
void FASTCALL call_4_15_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_15_0
#define init_4_15_15_0
void FASTCALL call_4_15_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_15_1
#define init_4_15_15_1
void FASTCALL call_4_15_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_15_2
#define init_4_15_15_2
void FASTCALL call_4_15_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_15_3
#define init_4_15_15_3
void FASTCALL call_4_15_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_15_4
#define init_4_15_15_4
void FASTCALL call_4_15_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_15_5
#define init_4_15_15_5
void FASTCALL call_4_15_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_15_6
#define init_4_15_15_6
void FASTCALL call_4_15_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_15_7
#define init_4_15_15_7
void FASTCALL call_4_15_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_15_8
#define init_4_15_15_8
void FASTCALL call_4_15_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_15_9
#define init_4_15_15_9
void FASTCALL call_4_15_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_15_10
#define init_4_15_15_10
void FASTCALL call_4_15_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_15_11
#define init_4_15_15_11
void FASTCALL call_4_15_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_15_12
#define init_4_15_15_12
void FASTCALL call_4_15_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_15_13
#define init_4_15_15_13
void FASTCALL call_4_15_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_15_14
#define init_4_15_15_14
void FASTCALL call_4_15_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_4_15_15_15
#define init_4_15_15_15
void FASTCALL call_4_15_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_0_0
#define init_5_0_0_0
void FASTCALL call_5_0_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_0_1
#define init_5_0_0_1
void FASTCALL call_5_0_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_0_2
#define init_5_0_0_2
void FASTCALL call_5_0_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_0_3
#define init_5_0_0_3
void FASTCALL call_5_0_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_0_4
#define init_5_0_0_4
void FASTCALL call_5_0_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_0_5
#define init_5_0_0_5
void FASTCALL call_5_0_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_0_6
#define init_5_0_0_6
void FASTCALL call_5_0_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_0_7
#define init_5_0_0_7
void FASTCALL call_5_0_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_0_8
#define init_5_0_0_8
void FASTCALL call_5_0_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_0_9
#define init_5_0_0_9
void FASTCALL call_5_0_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_0_10
#define init_5_0_0_10
void FASTCALL call_5_0_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_0_11
#define init_5_0_0_11
void FASTCALL call_5_0_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_0_12
#define init_5_0_0_12
void FASTCALL call_5_0_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_0_13
#define init_5_0_0_13
void FASTCALL call_5_0_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_0_14
#define init_5_0_0_14
void FASTCALL call_5_0_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_0_15
#define init_5_0_0_15
void FASTCALL call_5_0_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_1_0
#define init_5_0_1_0
void FASTCALL call_5_0_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_1_1
#define init_5_0_1_1
void FASTCALL call_5_0_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_1_2
#define init_5_0_1_2
void FASTCALL call_5_0_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_1_3
#define init_5_0_1_3
void FASTCALL call_5_0_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_1_4
#define init_5_0_1_4
void FASTCALL call_5_0_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_1_5
#define init_5_0_1_5
void FASTCALL call_5_0_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_1_6
#define init_5_0_1_6
void FASTCALL call_5_0_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_1_7
#define init_5_0_1_7
void FASTCALL call_5_0_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_1_8
#define init_5_0_1_8
void FASTCALL call_5_0_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_1_9
#define init_5_0_1_9
void FASTCALL call_5_0_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_1_10
#define init_5_0_1_10
void FASTCALL call_5_0_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_1_11
#define init_5_0_1_11
void FASTCALL call_5_0_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_1_12
#define init_5_0_1_12
void FASTCALL call_5_0_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_1_13
#define init_5_0_1_13
void FASTCALL call_5_0_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_1_14
#define init_5_0_1_14
void FASTCALL call_5_0_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_1_15
#define init_5_0_1_15
void FASTCALL call_5_0_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_2_0
#define init_5_0_2_0
void FASTCALL call_5_0_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_2_1
#define init_5_0_2_1
void FASTCALL call_5_0_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_2_2
#define init_5_0_2_2
void FASTCALL call_5_0_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_2_3
#define init_5_0_2_3
void FASTCALL call_5_0_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_2_4
#define init_5_0_2_4
void FASTCALL call_5_0_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_2_5
#define init_5_0_2_5
void FASTCALL call_5_0_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_2_6
#define init_5_0_2_6
void FASTCALL call_5_0_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_2_7
#define init_5_0_2_7
void FASTCALL call_5_0_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_2_8
#define init_5_0_2_8
void FASTCALL call_5_0_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_2_9
#define init_5_0_2_9
void FASTCALL call_5_0_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_2_10
#define init_5_0_2_10
void FASTCALL call_5_0_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_2_11
#define init_5_0_2_11
void FASTCALL call_5_0_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_2_12
#define init_5_0_2_12
void FASTCALL call_5_0_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_2_13
#define init_5_0_2_13
void FASTCALL call_5_0_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_2_14
#define init_5_0_2_14
void FASTCALL call_5_0_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_2_15
#define init_5_0_2_15
void FASTCALL call_5_0_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_3_0
#define init_5_0_3_0
void FASTCALL call_5_0_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_3_1
#define init_5_0_3_1
void FASTCALL call_5_0_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_3_2
#define init_5_0_3_2
void FASTCALL call_5_0_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_3_3
#define init_5_0_3_3
void FASTCALL call_5_0_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_3_4
#define init_5_0_3_4
void FASTCALL call_5_0_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_3_5
#define init_5_0_3_5
void FASTCALL call_5_0_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_3_6
#define init_5_0_3_6
void FASTCALL call_5_0_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_3_7
#define init_5_0_3_7
void FASTCALL call_5_0_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_3_8
#define init_5_0_3_8
void FASTCALL call_5_0_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_3_9
#define init_5_0_3_9
void FASTCALL call_5_0_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_3_10
#define init_5_0_3_10
void FASTCALL call_5_0_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_3_11
#define init_5_0_3_11
void FASTCALL call_5_0_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_3_12
#define init_5_0_3_12
void FASTCALL call_5_0_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_3_13
#define init_5_0_3_13
void FASTCALL call_5_0_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_3_14
#define init_5_0_3_14
void FASTCALL call_5_0_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_3_15
#define init_5_0_3_15
void FASTCALL call_5_0_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_4_0
#define init_5_0_4_0
void FASTCALL call_5_0_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_4_1
#define init_5_0_4_1
void FASTCALL call_5_0_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_4_2
#define init_5_0_4_2
void FASTCALL call_5_0_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_4_3
#define init_5_0_4_3
void FASTCALL call_5_0_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_4_4
#define init_5_0_4_4
void FASTCALL call_5_0_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_4_5
#define init_5_0_4_5
void FASTCALL call_5_0_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_4_6
#define init_5_0_4_6
void FASTCALL call_5_0_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_4_7
#define init_5_0_4_7
void FASTCALL call_5_0_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_4_8
#define init_5_0_4_8
void FASTCALL call_5_0_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_4_9
#define init_5_0_4_9
void FASTCALL call_5_0_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_4_10
#define init_5_0_4_10
void FASTCALL call_5_0_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_4_11
#define init_5_0_4_11
void FASTCALL call_5_0_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_4_12
#define init_5_0_4_12
void FASTCALL call_5_0_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_4_13
#define init_5_0_4_13
void FASTCALL call_5_0_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_4_14
#define init_5_0_4_14
void FASTCALL call_5_0_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_4_15
#define init_5_0_4_15
void FASTCALL call_5_0_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_5_0
#define init_5_0_5_0
void FASTCALL call_5_0_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_5_1
#define init_5_0_5_1
void FASTCALL call_5_0_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_5_2
#define init_5_0_5_2
void FASTCALL call_5_0_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_5_3
#define init_5_0_5_3
void FASTCALL call_5_0_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_5_4
#define init_5_0_5_4
void FASTCALL call_5_0_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_5_5
#define init_5_0_5_5
void FASTCALL call_5_0_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_5_6
#define init_5_0_5_6
void FASTCALL call_5_0_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_5_7
#define init_5_0_5_7
void FASTCALL call_5_0_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_5_8
#define init_5_0_5_8
void FASTCALL call_5_0_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_5_9
#define init_5_0_5_9
void FASTCALL call_5_0_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_5_10
#define init_5_0_5_10
void FASTCALL call_5_0_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_5_11
#define init_5_0_5_11
void FASTCALL call_5_0_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_5_12
#define init_5_0_5_12
void FASTCALL call_5_0_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_5_13
#define init_5_0_5_13
void FASTCALL call_5_0_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_5_14
#define init_5_0_5_14
void FASTCALL call_5_0_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_5_15
#define init_5_0_5_15
void FASTCALL call_5_0_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_6_0
#define init_5_0_6_0
void FASTCALL call_5_0_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_6_1
#define init_5_0_6_1
void FASTCALL call_5_0_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_6_2
#define init_5_0_6_2
void FASTCALL call_5_0_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_6_3
#define init_5_0_6_3
void FASTCALL call_5_0_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_6_4
#define init_5_0_6_4
void FASTCALL call_5_0_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_6_5
#define init_5_0_6_5
void FASTCALL call_5_0_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_6_6
#define init_5_0_6_6
void FASTCALL call_5_0_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_6_7
#define init_5_0_6_7
void FASTCALL call_5_0_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_6_8
#define init_5_0_6_8
void FASTCALL call_5_0_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_6_9
#define init_5_0_6_9
void FASTCALL call_5_0_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_6_10
#define init_5_0_6_10
void FASTCALL call_5_0_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_6_11
#define init_5_0_6_11
void FASTCALL call_5_0_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_6_12
#define init_5_0_6_12
void FASTCALL call_5_0_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_6_13
#define init_5_0_6_13
void FASTCALL call_5_0_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_6_14
#define init_5_0_6_14
void FASTCALL call_5_0_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_6_15
#define init_5_0_6_15
void FASTCALL call_5_0_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_7_0
#define init_5_0_7_0
void FASTCALL call_5_0_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_7_1
#define init_5_0_7_1
void FASTCALL call_5_0_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_7_2
#define init_5_0_7_2
void FASTCALL call_5_0_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_7_3
#define init_5_0_7_3
void FASTCALL call_5_0_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_7_4
#define init_5_0_7_4
void FASTCALL call_5_0_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_7_5
#define init_5_0_7_5
void FASTCALL call_5_0_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_7_6
#define init_5_0_7_6
void FASTCALL call_5_0_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_7_7
#define init_5_0_7_7
void FASTCALL call_5_0_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_7_8
#define init_5_0_7_8
void FASTCALL call_5_0_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_7_9
#define init_5_0_7_9
void FASTCALL call_5_0_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_7_10
#define init_5_0_7_10
void FASTCALL call_5_0_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_7_11
#define init_5_0_7_11
void FASTCALL call_5_0_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_7_12
#define init_5_0_7_12
void FASTCALL call_5_0_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_7_13
#define init_5_0_7_13
void FASTCALL call_5_0_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_7_14
#define init_5_0_7_14
void FASTCALL call_5_0_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_7_15
#define init_5_0_7_15
void FASTCALL call_5_0_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_8_0
#define init_5_0_8_0
void FASTCALL call_5_0_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_8_1
#define init_5_0_8_1
void FASTCALL call_5_0_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_8_2
#define init_5_0_8_2
void FASTCALL call_5_0_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_8_3
#define init_5_0_8_3
void FASTCALL call_5_0_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_8_4
#define init_5_0_8_4
void FASTCALL call_5_0_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_8_5
#define init_5_0_8_5
void FASTCALL call_5_0_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_8_6
#define init_5_0_8_6
void FASTCALL call_5_0_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_8_7
#define init_5_0_8_7
void FASTCALL call_5_0_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_8_8
#define init_5_0_8_8
void FASTCALL call_5_0_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_8_9
#define init_5_0_8_9
void FASTCALL call_5_0_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_8_10
#define init_5_0_8_10
void FASTCALL call_5_0_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_8_11
#define init_5_0_8_11
void FASTCALL call_5_0_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_8_12
#define init_5_0_8_12
void FASTCALL call_5_0_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_8_13
#define init_5_0_8_13
void FASTCALL call_5_0_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_8_14
#define init_5_0_8_14
void FASTCALL call_5_0_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_8_15
#define init_5_0_8_15
void FASTCALL call_5_0_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_9_0
#define init_5_0_9_0
void FASTCALL call_5_0_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_9_1
#define init_5_0_9_1
void FASTCALL call_5_0_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_9_2
#define init_5_0_9_2
void FASTCALL call_5_0_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_9_3
#define init_5_0_9_3
void FASTCALL call_5_0_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_9_4
#define init_5_0_9_4
void FASTCALL call_5_0_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_9_5
#define init_5_0_9_5
void FASTCALL call_5_0_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_9_6
#define init_5_0_9_6
void FASTCALL call_5_0_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_9_7
#define init_5_0_9_7
void FASTCALL call_5_0_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_9_8
#define init_5_0_9_8
void FASTCALL call_5_0_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_9_9
#define init_5_0_9_9
void FASTCALL call_5_0_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_9_10
#define init_5_0_9_10
void FASTCALL call_5_0_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_9_11
#define init_5_0_9_11
void FASTCALL call_5_0_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_9_12
#define init_5_0_9_12
void FASTCALL call_5_0_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_9_13
#define init_5_0_9_13
void FASTCALL call_5_0_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_9_14
#define init_5_0_9_14
void FASTCALL call_5_0_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_9_15
#define init_5_0_9_15
void FASTCALL call_5_0_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_10_0
#define init_5_0_10_0
void FASTCALL call_5_0_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_10_1
#define init_5_0_10_1
void FASTCALL call_5_0_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_10_2
#define init_5_0_10_2
void FASTCALL call_5_0_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_10_3
#define init_5_0_10_3
void FASTCALL call_5_0_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_10_4
#define init_5_0_10_4
void FASTCALL call_5_0_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_10_5
#define init_5_0_10_5
void FASTCALL call_5_0_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_10_6
#define init_5_0_10_6
void FASTCALL call_5_0_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_10_7
#define init_5_0_10_7
void FASTCALL call_5_0_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_10_8
#define init_5_0_10_8
void FASTCALL call_5_0_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_10_9
#define init_5_0_10_9
void FASTCALL call_5_0_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_10_10
#define init_5_0_10_10
void FASTCALL call_5_0_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_10_11
#define init_5_0_10_11
void FASTCALL call_5_0_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_10_12
#define init_5_0_10_12
void FASTCALL call_5_0_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_10_13
#define init_5_0_10_13
void FASTCALL call_5_0_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_10_14
#define init_5_0_10_14
void FASTCALL call_5_0_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_10_15
#define init_5_0_10_15
void FASTCALL call_5_0_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_11_0
#define init_5_0_11_0
void FASTCALL call_5_0_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_11_1
#define init_5_0_11_1
void FASTCALL call_5_0_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_11_2
#define init_5_0_11_2
void FASTCALL call_5_0_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_11_3
#define init_5_0_11_3
void FASTCALL call_5_0_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_11_4
#define init_5_0_11_4
void FASTCALL call_5_0_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_11_5
#define init_5_0_11_5
void FASTCALL call_5_0_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_11_6
#define init_5_0_11_6
void FASTCALL call_5_0_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_11_7
#define init_5_0_11_7
void FASTCALL call_5_0_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_11_8
#define init_5_0_11_8
void FASTCALL call_5_0_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_11_9
#define init_5_0_11_9
void FASTCALL call_5_0_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_11_10
#define init_5_0_11_10
void FASTCALL call_5_0_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_11_11
#define init_5_0_11_11
void FASTCALL call_5_0_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_11_12
#define init_5_0_11_12
void FASTCALL call_5_0_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_11_13
#define init_5_0_11_13
void FASTCALL call_5_0_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_11_14
#define init_5_0_11_14
void FASTCALL call_5_0_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_11_15
#define init_5_0_11_15
void FASTCALL call_5_0_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_12_0
#define init_5_0_12_0
void FASTCALL call_5_0_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_12_1
#define init_5_0_12_1
void FASTCALL call_5_0_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_12_2
#define init_5_0_12_2
void FASTCALL call_5_0_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_12_3
#define init_5_0_12_3
void FASTCALL call_5_0_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_12_4
#define init_5_0_12_4
void FASTCALL call_5_0_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_12_5
#define init_5_0_12_5
void FASTCALL call_5_0_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_12_6
#define init_5_0_12_6
void FASTCALL call_5_0_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_12_7
#define init_5_0_12_7
void FASTCALL call_5_0_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_12_8
#define init_5_0_12_8
void FASTCALL call_5_0_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_12_9
#define init_5_0_12_9
void FASTCALL call_5_0_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_12_10
#define init_5_0_12_10
void FASTCALL call_5_0_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_12_11
#define init_5_0_12_11
void FASTCALL call_5_0_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_12_12
#define init_5_0_12_12
void FASTCALL call_5_0_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_12_13
#define init_5_0_12_13
void FASTCALL call_5_0_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_12_14
#define init_5_0_12_14
void FASTCALL call_5_0_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_12_15
#define init_5_0_12_15
void FASTCALL call_5_0_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_13_0
#define init_5_0_13_0
void FASTCALL call_5_0_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_13_1
#define init_5_0_13_1
void FASTCALL call_5_0_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_13_2
#define init_5_0_13_2
void FASTCALL call_5_0_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_13_3
#define init_5_0_13_3
void FASTCALL call_5_0_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_13_4
#define init_5_0_13_4
void FASTCALL call_5_0_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_13_5
#define init_5_0_13_5
void FASTCALL call_5_0_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_13_6
#define init_5_0_13_6
void FASTCALL call_5_0_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_13_7
#define init_5_0_13_7
void FASTCALL call_5_0_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_13_8
#define init_5_0_13_8
void FASTCALL call_5_0_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_13_9
#define init_5_0_13_9
void FASTCALL call_5_0_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_13_10
#define init_5_0_13_10
void FASTCALL call_5_0_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_13_11
#define init_5_0_13_11
void FASTCALL call_5_0_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_13_12
#define init_5_0_13_12
void FASTCALL call_5_0_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_13_13
#define init_5_0_13_13
void FASTCALL call_5_0_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_13_14
#define init_5_0_13_14
void FASTCALL call_5_0_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_13_15
#define init_5_0_13_15
void FASTCALL call_5_0_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_14_0
#define init_5_0_14_0
void FASTCALL call_5_0_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_14_1
#define init_5_0_14_1
void FASTCALL call_5_0_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_14_2
#define init_5_0_14_2
void FASTCALL call_5_0_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_14_3
#define init_5_0_14_3
void FASTCALL call_5_0_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_14_4
#define init_5_0_14_4
void FASTCALL call_5_0_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_14_5
#define init_5_0_14_5
void FASTCALL call_5_0_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_14_6
#define init_5_0_14_6
void FASTCALL call_5_0_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_14_7
#define init_5_0_14_7
void FASTCALL call_5_0_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_14_8
#define init_5_0_14_8
void FASTCALL call_5_0_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_14_9
#define init_5_0_14_9
void FASTCALL call_5_0_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_14_10
#define init_5_0_14_10
void FASTCALL call_5_0_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_14_11
#define init_5_0_14_11
void FASTCALL call_5_0_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_14_12
#define init_5_0_14_12
void FASTCALL call_5_0_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_14_13
#define init_5_0_14_13
void FASTCALL call_5_0_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_14_14
#define init_5_0_14_14
void FASTCALL call_5_0_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_14_15
#define init_5_0_14_15
void FASTCALL call_5_0_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_15_0
#define init_5_0_15_0
void FASTCALL call_5_0_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_15_1
#define init_5_0_15_1
void FASTCALL call_5_0_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_15_2
#define init_5_0_15_2
void FASTCALL call_5_0_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_15_3
#define init_5_0_15_3
void FASTCALL call_5_0_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_15_4
#define init_5_0_15_4
void FASTCALL call_5_0_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_15_5
#define init_5_0_15_5
void FASTCALL call_5_0_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_15_6
#define init_5_0_15_6
void FASTCALL call_5_0_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_15_7
#define init_5_0_15_7
void FASTCALL call_5_0_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_15_8
#define init_5_0_15_8
void FASTCALL call_5_0_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_15_9
#define init_5_0_15_9
void FASTCALL call_5_0_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_15_10
#define init_5_0_15_10
void FASTCALL call_5_0_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_15_11
#define init_5_0_15_11
void FASTCALL call_5_0_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_15_12
#define init_5_0_15_12
void FASTCALL call_5_0_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_15_13
#define init_5_0_15_13
void FASTCALL call_5_0_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_15_14
#define init_5_0_15_14
void FASTCALL call_5_0_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_0_15_15
#define init_5_0_15_15
void FASTCALL call_5_0_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_0_0
#define init_5_1_0_0
void FASTCALL call_5_1_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_0_1
#define init_5_1_0_1
void FASTCALL call_5_1_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_0_2
#define init_5_1_0_2
void FASTCALL call_5_1_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_0_3
#define init_5_1_0_3
void FASTCALL call_5_1_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_0_4
#define init_5_1_0_4
void FASTCALL call_5_1_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_0_5
#define init_5_1_0_5
void FASTCALL call_5_1_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_0_6
#define init_5_1_0_6
void FASTCALL call_5_1_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_0_7
#define init_5_1_0_7
void FASTCALL call_5_1_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_0_8
#define init_5_1_0_8
void FASTCALL call_5_1_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_0_9
#define init_5_1_0_9
void FASTCALL call_5_1_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_0_10
#define init_5_1_0_10
void FASTCALL call_5_1_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_0_11
#define init_5_1_0_11
void FASTCALL call_5_1_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_0_12
#define init_5_1_0_12
void FASTCALL call_5_1_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_0_13
#define init_5_1_0_13
void FASTCALL call_5_1_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_0_14
#define init_5_1_0_14
void FASTCALL call_5_1_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_0_15
#define init_5_1_0_15
void FASTCALL call_5_1_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_1_0
#define init_5_1_1_0
void FASTCALL call_5_1_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_1_1
#define init_5_1_1_1
void FASTCALL call_5_1_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_1_2
#define init_5_1_1_2
void FASTCALL call_5_1_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_1_3
#define init_5_1_1_3
void FASTCALL call_5_1_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_1_4
#define init_5_1_1_4
void FASTCALL call_5_1_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_1_5
#define init_5_1_1_5
void FASTCALL call_5_1_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_1_6
#define init_5_1_1_6
void FASTCALL call_5_1_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_1_7
#define init_5_1_1_7
void FASTCALL call_5_1_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_1_8
#define init_5_1_1_8
void FASTCALL call_5_1_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_1_9
#define init_5_1_1_9
void FASTCALL call_5_1_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_1_10
#define init_5_1_1_10
void FASTCALL call_5_1_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_1_11
#define init_5_1_1_11
void FASTCALL call_5_1_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_1_12
#define init_5_1_1_12
void FASTCALL call_5_1_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_1_13
#define init_5_1_1_13
void FASTCALL call_5_1_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_1_14
#define init_5_1_1_14
void FASTCALL call_5_1_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_1_15
#define init_5_1_1_15
void FASTCALL call_5_1_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_2_0
#define init_5_1_2_0
void FASTCALL call_5_1_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_2_1
#define init_5_1_2_1
void FASTCALL call_5_1_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_2_2
#define init_5_1_2_2
void FASTCALL call_5_1_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_2_3
#define init_5_1_2_3
void FASTCALL call_5_1_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_2_4
#define init_5_1_2_4
void FASTCALL call_5_1_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_2_5
#define init_5_1_2_5
void FASTCALL call_5_1_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_2_6
#define init_5_1_2_6
void FASTCALL call_5_1_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_2_7
#define init_5_1_2_7
void FASTCALL call_5_1_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_2_8
#define init_5_1_2_8
void FASTCALL call_5_1_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_2_9
#define init_5_1_2_9
void FASTCALL call_5_1_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_2_10
#define init_5_1_2_10
void FASTCALL call_5_1_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_2_11
#define init_5_1_2_11
void FASTCALL call_5_1_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_2_12
#define init_5_1_2_12
void FASTCALL call_5_1_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_2_13
#define init_5_1_2_13
void FASTCALL call_5_1_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_2_14
#define init_5_1_2_14
void FASTCALL call_5_1_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_2_15
#define init_5_1_2_15
void FASTCALL call_5_1_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_3_0
#define init_5_1_3_0
void FASTCALL call_5_1_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_3_1
#define init_5_1_3_1
void FASTCALL call_5_1_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_3_2
#define init_5_1_3_2
void FASTCALL call_5_1_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_3_3
#define init_5_1_3_3
void FASTCALL call_5_1_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_3_4
#define init_5_1_3_4
void FASTCALL call_5_1_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_3_5
#define init_5_1_3_5
void FASTCALL call_5_1_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_3_6
#define init_5_1_3_6
void FASTCALL call_5_1_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_3_7
#define init_5_1_3_7
void FASTCALL call_5_1_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_3_8
#define init_5_1_3_8
void FASTCALL call_5_1_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_3_9
#define init_5_1_3_9
void FASTCALL call_5_1_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_3_10
#define init_5_1_3_10
void FASTCALL call_5_1_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_3_11
#define init_5_1_3_11
void FASTCALL call_5_1_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_3_12
#define init_5_1_3_12
void FASTCALL call_5_1_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_3_13
#define init_5_1_3_13
void FASTCALL call_5_1_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_3_14
#define init_5_1_3_14
void FASTCALL call_5_1_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_3_15
#define init_5_1_3_15
void FASTCALL call_5_1_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_4_0
#define init_5_1_4_0
void FASTCALL call_5_1_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_4_1
#define init_5_1_4_1
void FASTCALL call_5_1_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_4_2
#define init_5_1_4_2
void FASTCALL call_5_1_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_4_3
#define init_5_1_4_3
void FASTCALL call_5_1_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_4_4
#define init_5_1_4_4
void FASTCALL call_5_1_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_4_5
#define init_5_1_4_5
void FASTCALL call_5_1_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_4_6
#define init_5_1_4_6
void FASTCALL call_5_1_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_4_7
#define init_5_1_4_7
void FASTCALL call_5_1_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_4_8
#define init_5_1_4_8
void FASTCALL call_5_1_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_4_9
#define init_5_1_4_9
void FASTCALL call_5_1_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_4_10
#define init_5_1_4_10
void FASTCALL call_5_1_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_4_11
#define init_5_1_4_11
void FASTCALL call_5_1_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_4_12
#define init_5_1_4_12
void FASTCALL call_5_1_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_4_13
#define init_5_1_4_13
void FASTCALL call_5_1_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_4_14
#define init_5_1_4_14
void FASTCALL call_5_1_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_4_15
#define init_5_1_4_15
void FASTCALL call_5_1_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_5_0
#define init_5_1_5_0
void FASTCALL call_5_1_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_5_1
#define init_5_1_5_1
void FASTCALL call_5_1_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_5_2
#define init_5_1_5_2
void FASTCALL call_5_1_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_5_3
#define init_5_1_5_3
void FASTCALL call_5_1_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_5_4
#define init_5_1_5_4
void FASTCALL call_5_1_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_5_5
#define init_5_1_5_5
void FASTCALL call_5_1_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_5_6
#define init_5_1_5_6
void FASTCALL call_5_1_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_5_7
#define init_5_1_5_7
void FASTCALL call_5_1_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_5_8
#define init_5_1_5_8
void FASTCALL call_5_1_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_5_9
#define init_5_1_5_9
void FASTCALL call_5_1_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_5_10
#define init_5_1_5_10
void FASTCALL call_5_1_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_5_11
#define init_5_1_5_11
void FASTCALL call_5_1_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_5_12
#define init_5_1_5_12
void FASTCALL call_5_1_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_5_13
#define init_5_1_5_13
void FASTCALL call_5_1_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_5_14
#define init_5_1_5_14
void FASTCALL call_5_1_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_5_15
#define init_5_1_5_15
void FASTCALL call_5_1_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_6_0
#define init_5_1_6_0
void FASTCALL call_5_1_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_6_1
#define init_5_1_6_1
void FASTCALL call_5_1_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_6_2
#define init_5_1_6_2
void FASTCALL call_5_1_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_6_3
#define init_5_1_6_3
void FASTCALL call_5_1_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_6_4
#define init_5_1_6_4
void FASTCALL call_5_1_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_6_5
#define init_5_1_6_5
void FASTCALL call_5_1_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_6_6
#define init_5_1_6_6
void FASTCALL call_5_1_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_6_7
#define init_5_1_6_7
void FASTCALL call_5_1_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_6_8
#define init_5_1_6_8
void FASTCALL call_5_1_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_6_9
#define init_5_1_6_9
void FASTCALL call_5_1_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_6_10
#define init_5_1_6_10
void FASTCALL call_5_1_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_6_11
#define init_5_1_6_11
void FASTCALL call_5_1_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_6_12
#define init_5_1_6_12
void FASTCALL call_5_1_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_6_13
#define init_5_1_6_13
void FASTCALL call_5_1_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_6_14
#define init_5_1_6_14
void FASTCALL call_5_1_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_6_15
#define init_5_1_6_15
void FASTCALL call_5_1_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_7_0
#define init_5_1_7_0
void FASTCALL call_5_1_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_7_1
#define init_5_1_7_1
void FASTCALL call_5_1_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_7_2
#define init_5_1_7_2
void FASTCALL call_5_1_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_7_3
#define init_5_1_7_3
void FASTCALL call_5_1_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_7_4
#define init_5_1_7_4
void FASTCALL call_5_1_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_7_5
#define init_5_1_7_5
void FASTCALL call_5_1_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_7_6
#define init_5_1_7_6
void FASTCALL call_5_1_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_7_7
#define init_5_1_7_7
void FASTCALL call_5_1_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_7_8
#define init_5_1_7_8
void FASTCALL call_5_1_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_7_9
#define init_5_1_7_9
void FASTCALL call_5_1_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_7_10
#define init_5_1_7_10
void FASTCALL call_5_1_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_7_11
#define init_5_1_7_11
void FASTCALL call_5_1_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_7_12
#define init_5_1_7_12
void FASTCALL call_5_1_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_7_13
#define init_5_1_7_13
void FASTCALL call_5_1_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_7_14
#define init_5_1_7_14
void FASTCALL call_5_1_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_7_15
#define init_5_1_7_15
void FASTCALL call_5_1_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_8_0
#define init_5_1_8_0
void FASTCALL call_5_1_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_8_1
#define init_5_1_8_1
void FASTCALL call_5_1_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_8_2
#define init_5_1_8_2
void FASTCALL call_5_1_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_8_3
#define init_5_1_8_3
void FASTCALL call_5_1_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_8_4
#define init_5_1_8_4
void FASTCALL call_5_1_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_8_5
#define init_5_1_8_5
void FASTCALL call_5_1_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_8_6
#define init_5_1_8_6
void FASTCALL call_5_1_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_8_7
#define init_5_1_8_7
void FASTCALL call_5_1_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_8_8
#define init_5_1_8_8
void FASTCALL call_5_1_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_8_9
#define init_5_1_8_9
void FASTCALL call_5_1_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_8_10
#define init_5_1_8_10
void FASTCALL call_5_1_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_8_11
#define init_5_1_8_11
void FASTCALL call_5_1_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_8_12
#define init_5_1_8_12
void FASTCALL call_5_1_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_8_13
#define init_5_1_8_13
void FASTCALL call_5_1_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_8_14
#define init_5_1_8_14
void FASTCALL call_5_1_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_8_15
#define init_5_1_8_15
void FASTCALL call_5_1_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_9_0
#define init_5_1_9_0
void FASTCALL call_5_1_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_9_1
#define init_5_1_9_1
void FASTCALL call_5_1_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_9_2
#define init_5_1_9_2
void FASTCALL call_5_1_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_9_3
#define init_5_1_9_3
void FASTCALL call_5_1_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_9_4
#define init_5_1_9_4
void FASTCALL call_5_1_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_9_5
#define init_5_1_9_5
void FASTCALL call_5_1_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_9_6
#define init_5_1_9_6
void FASTCALL call_5_1_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_9_7
#define init_5_1_9_7
void FASTCALL call_5_1_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_9_8
#define init_5_1_9_8
void FASTCALL call_5_1_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_9_9
#define init_5_1_9_9
void FASTCALL call_5_1_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_9_10
#define init_5_1_9_10
void FASTCALL call_5_1_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_9_11
#define init_5_1_9_11
void FASTCALL call_5_1_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_9_12
#define init_5_1_9_12
void FASTCALL call_5_1_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_9_13
#define init_5_1_9_13
void FASTCALL call_5_1_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_9_14
#define init_5_1_9_14
void FASTCALL call_5_1_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_9_15
#define init_5_1_9_15
void FASTCALL call_5_1_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_10_0
#define init_5_1_10_0
void FASTCALL call_5_1_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_10_1
#define init_5_1_10_1
void FASTCALL call_5_1_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_10_2
#define init_5_1_10_2
void FASTCALL call_5_1_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_10_3
#define init_5_1_10_3
void FASTCALL call_5_1_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_10_4
#define init_5_1_10_4
void FASTCALL call_5_1_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_10_5
#define init_5_1_10_5
void FASTCALL call_5_1_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_10_6
#define init_5_1_10_6
void FASTCALL call_5_1_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_10_7
#define init_5_1_10_7
void FASTCALL call_5_1_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_10_8
#define init_5_1_10_8
void FASTCALL call_5_1_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_10_9
#define init_5_1_10_9
void FASTCALL call_5_1_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_10_10
#define init_5_1_10_10
void FASTCALL call_5_1_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_10_11
#define init_5_1_10_11
void FASTCALL call_5_1_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_10_12
#define init_5_1_10_12
void FASTCALL call_5_1_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_10_13
#define init_5_1_10_13
void FASTCALL call_5_1_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_10_14
#define init_5_1_10_14
void FASTCALL call_5_1_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_10_15
#define init_5_1_10_15
void FASTCALL call_5_1_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_11_0
#define init_5_1_11_0
void FASTCALL call_5_1_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_11_1
#define init_5_1_11_1
void FASTCALL call_5_1_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_11_2
#define init_5_1_11_2
void FASTCALL call_5_1_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_11_3
#define init_5_1_11_3
void FASTCALL call_5_1_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_11_4
#define init_5_1_11_4
void FASTCALL call_5_1_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_11_5
#define init_5_1_11_5
void FASTCALL call_5_1_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_11_6
#define init_5_1_11_6
void FASTCALL call_5_1_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_11_7
#define init_5_1_11_7
void FASTCALL call_5_1_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_11_8
#define init_5_1_11_8
void FASTCALL call_5_1_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_11_9
#define init_5_1_11_9
void FASTCALL call_5_1_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_11_10
#define init_5_1_11_10
void FASTCALL call_5_1_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_11_11
#define init_5_1_11_11
void FASTCALL call_5_1_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_11_12
#define init_5_1_11_12
void FASTCALL call_5_1_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_11_13
#define init_5_1_11_13
void FASTCALL call_5_1_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_11_14
#define init_5_1_11_14
void FASTCALL call_5_1_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_11_15
#define init_5_1_11_15
void FASTCALL call_5_1_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_12_0
#define init_5_1_12_0
void FASTCALL call_5_1_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_12_1
#define init_5_1_12_1
void FASTCALL call_5_1_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_12_2
#define init_5_1_12_2
void FASTCALL call_5_1_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_12_3
#define init_5_1_12_3
void FASTCALL call_5_1_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_12_4
#define init_5_1_12_4
void FASTCALL call_5_1_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_12_5
#define init_5_1_12_5
void FASTCALL call_5_1_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_12_6
#define init_5_1_12_6
void FASTCALL call_5_1_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_12_7
#define init_5_1_12_7
void FASTCALL call_5_1_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_12_8
#define init_5_1_12_8
void FASTCALL call_5_1_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_12_9
#define init_5_1_12_9
void FASTCALL call_5_1_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_12_10
#define init_5_1_12_10
void FASTCALL call_5_1_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_12_11
#define init_5_1_12_11
void FASTCALL call_5_1_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_12_12
#define init_5_1_12_12
void FASTCALL call_5_1_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_12_13
#define init_5_1_12_13
void FASTCALL call_5_1_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_12_14
#define init_5_1_12_14
void FASTCALL call_5_1_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_12_15
#define init_5_1_12_15
void FASTCALL call_5_1_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_13_0
#define init_5_1_13_0
void FASTCALL call_5_1_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_13_1
#define init_5_1_13_1
void FASTCALL call_5_1_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_13_2
#define init_5_1_13_2
void FASTCALL call_5_1_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_13_3
#define init_5_1_13_3
void FASTCALL call_5_1_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_13_4
#define init_5_1_13_4
void FASTCALL call_5_1_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_13_5
#define init_5_1_13_5
void FASTCALL call_5_1_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_13_6
#define init_5_1_13_6
void FASTCALL call_5_1_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_13_7
#define init_5_1_13_7
void FASTCALL call_5_1_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_13_8
#define init_5_1_13_8
void FASTCALL call_5_1_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_13_9
#define init_5_1_13_9
void FASTCALL call_5_1_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_13_10
#define init_5_1_13_10
void FASTCALL call_5_1_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_13_11
#define init_5_1_13_11
void FASTCALL call_5_1_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_13_12
#define init_5_1_13_12
void FASTCALL call_5_1_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_13_13
#define init_5_1_13_13
void FASTCALL call_5_1_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_13_14
#define init_5_1_13_14
void FASTCALL call_5_1_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_13_15
#define init_5_1_13_15
void FASTCALL call_5_1_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_14_0
#define init_5_1_14_0
void FASTCALL call_5_1_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_14_1
#define init_5_1_14_1
void FASTCALL call_5_1_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_14_2
#define init_5_1_14_2
void FASTCALL call_5_1_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_14_3
#define init_5_1_14_3
void FASTCALL call_5_1_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_14_4
#define init_5_1_14_4
void FASTCALL call_5_1_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_14_5
#define init_5_1_14_5
void FASTCALL call_5_1_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_14_6
#define init_5_1_14_6
void FASTCALL call_5_1_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_14_7
#define init_5_1_14_7
void FASTCALL call_5_1_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_14_8
#define init_5_1_14_8
void FASTCALL call_5_1_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_14_9
#define init_5_1_14_9
void FASTCALL call_5_1_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_14_10
#define init_5_1_14_10
void FASTCALL call_5_1_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_14_11
#define init_5_1_14_11
void FASTCALL call_5_1_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_14_12
#define init_5_1_14_12
void FASTCALL call_5_1_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_14_13
#define init_5_1_14_13
void FASTCALL call_5_1_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_14_14
#define init_5_1_14_14
void FASTCALL call_5_1_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_14_15
#define init_5_1_14_15
void FASTCALL call_5_1_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_15_0
#define init_5_1_15_0
void FASTCALL call_5_1_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_15_1
#define init_5_1_15_1
void FASTCALL call_5_1_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_15_2
#define init_5_1_15_2
void FASTCALL call_5_1_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_15_3
#define init_5_1_15_3
void FASTCALL call_5_1_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_15_4
#define init_5_1_15_4
void FASTCALL call_5_1_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_15_5
#define init_5_1_15_5
void FASTCALL call_5_1_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_15_6
#define init_5_1_15_6
void FASTCALL call_5_1_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_15_7
#define init_5_1_15_7
void FASTCALL call_5_1_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_15_8
#define init_5_1_15_8
void FASTCALL call_5_1_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_15_9
#define init_5_1_15_9
void FASTCALL call_5_1_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_15_10
#define init_5_1_15_10
void FASTCALL call_5_1_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_15_11
#define init_5_1_15_11
void FASTCALL call_5_1_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_15_12
#define init_5_1_15_12
void FASTCALL call_5_1_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_15_13
#define init_5_1_15_13
void FASTCALL call_5_1_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_15_14
#define init_5_1_15_14
void FASTCALL call_5_1_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_1_15_15
#define init_5_1_15_15
void FASTCALL call_5_1_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_0_0
#define init_5_2_0_0
void FASTCALL call_5_2_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_0_1
#define init_5_2_0_1
void FASTCALL call_5_2_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_0_2
#define init_5_2_0_2
void FASTCALL call_5_2_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_0_3
#define init_5_2_0_3
void FASTCALL call_5_2_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_0_4
#define init_5_2_0_4
void FASTCALL call_5_2_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_0_5
#define init_5_2_0_5
void FASTCALL call_5_2_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_0_6
#define init_5_2_0_6
void FASTCALL call_5_2_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_0_7
#define init_5_2_0_7
void FASTCALL call_5_2_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_0_8
#define init_5_2_0_8
void FASTCALL call_5_2_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_0_9
#define init_5_2_0_9
void FASTCALL call_5_2_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_0_10
#define init_5_2_0_10
void FASTCALL call_5_2_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_0_11
#define init_5_2_0_11
void FASTCALL call_5_2_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_0_12
#define init_5_2_0_12
void FASTCALL call_5_2_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_0_13
#define init_5_2_0_13
void FASTCALL call_5_2_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_0_14
#define init_5_2_0_14
void FASTCALL call_5_2_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_0_15
#define init_5_2_0_15
void FASTCALL call_5_2_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_1_0
#define init_5_2_1_0
void FASTCALL call_5_2_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_1_1
#define init_5_2_1_1
void FASTCALL call_5_2_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_1_2
#define init_5_2_1_2
void FASTCALL call_5_2_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_1_3
#define init_5_2_1_3
void FASTCALL call_5_2_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_1_4
#define init_5_2_1_4
void FASTCALL call_5_2_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_1_5
#define init_5_2_1_5
void FASTCALL call_5_2_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_1_6
#define init_5_2_1_6
void FASTCALL call_5_2_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_1_7
#define init_5_2_1_7
void FASTCALL call_5_2_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_1_8
#define init_5_2_1_8
void FASTCALL call_5_2_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_1_9
#define init_5_2_1_9
void FASTCALL call_5_2_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_1_10
#define init_5_2_1_10
void FASTCALL call_5_2_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_1_11
#define init_5_2_1_11
void FASTCALL call_5_2_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_1_12
#define init_5_2_1_12
void FASTCALL call_5_2_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_1_13
#define init_5_2_1_13
void FASTCALL call_5_2_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_1_14
#define init_5_2_1_14
void FASTCALL call_5_2_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_1_15
#define init_5_2_1_15
void FASTCALL call_5_2_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_2_0
#define init_5_2_2_0
void FASTCALL call_5_2_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_2_1
#define init_5_2_2_1
void FASTCALL call_5_2_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_2_2
#define init_5_2_2_2
void FASTCALL call_5_2_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_2_3
#define init_5_2_2_3
void FASTCALL call_5_2_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_2_4
#define init_5_2_2_4
void FASTCALL call_5_2_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_2_5
#define init_5_2_2_5
void FASTCALL call_5_2_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_2_6
#define init_5_2_2_6
void FASTCALL call_5_2_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_2_7
#define init_5_2_2_7
void FASTCALL call_5_2_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_2_8
#define init_5_2_2_8
void FASTCALL call_5_2_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_2_9
#define init_5_2_2_9
void FASTCALL call_5_2_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_2_10
#define init_5_2_2_10
void FASTCALL call_5_2_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_2_11
#define init_5_2_2_11
void FASTCALL call_5_2_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_2_12
#define init_5_2_2_12
void FASTCALL call_5_2_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_2_13
#define init_5_2_2_13
void FASTCALL call_5_2_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_2_14
#define init_5_2_2_14
void FASTCALL call_5_2_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_2_15
#define init_5_2_2_15
void FASTCALL call_5_2_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_3_0
#define init_5_2_3_0
void FASTCALL call_5_2_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_3_1
#define init_5_2_3_1
void FASTCALL call_5_2_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_3_2
#define init_5_2_3_2
void FASTCALL call_5_2_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_3_3
#define init_5_2_3_3
void FASTCALL call_5_2_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_3_4
#define init_5_2_3_4
void FASTCALL call_5_2_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_3_5
#define init_5_2_3_5
void FASTCALL call_5_2_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_3_6
#define init_5_2_3_6
void FASTCALL call_5_2_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_3_7
#define init_5_2_3_7
void FASTCALL call_5_2_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_3_8
#define init_5_2_3_8
void FASTCALL call_5_2_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_3_9
#define init_5_2_3_9
void FASTCALL call_5_2_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_3_10
#define init_5_2_3_10
void FASTCALL call_5_2_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_3_11
#define init_5_2_3_11
void FASTCALL call_5_2_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_3_12
#define init_5_2_3_12
void FASTCALL call_5_2_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_3_13
#define init_5_2_3_13
void FASTCALL call_5_2_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_3_14
#define init_5_2_3_14
void FASTCALL call_5_2_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_3_15
#define init_5_2_3_15
void FASTCALL call_5_2_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_4_0
#define init_5_2_4_0
void FASTCALL call_5_2_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_4_1
#define init_5_2_4_1
void FASTCALL call_5_2_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_4_2
#define init_5_2_4_2
void FASTCALL call_5_2_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_4_3
#define init_5_2_4_3
void FASTCALL call_5_2_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_4_4
#define init_5_2_4_4
void FASTCALL call_5_2_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_4_5
#define init_5_2_4_5
void FASTCALL call_5_2_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_4_6
#define init_5_2_4_6
void FASTCALL call_5_2_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_4_7
#define init_5_2_4_7
void FASTCALL call_5_2_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_4_8
#define init_5_2_4_8
void FASTCALL call_5_2_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_4_9
#define init_5_2_4_9
void FASTCALL call_5_2_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_4_10
#define init_5_2_4_10
void FASTCALL call_5_2_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_4_11
#define init_5_2_4_11
void FASTCALL call_5_2_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_4_12
#define init_5_2_4_12
void FASTCALL call_5_2_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_4_13
#define init_5_2_4_13
void FASTCALL call_5_2_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_4_14
#define init_5_2_4_14
void FASTCALL call_5_2_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_4_15
#define init_5_2_4_15
void FASTCALL call_5_2_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_5_0
#define init_5_2_5_0
void FASTCALL call_5_2_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_5_1
#define init_5_2_5_1
void FASTCALL call_5_2_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_5_2
#define init_5_2_5_2
void FASTCALL call_5_2_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_5_3
#define init_5_2_5_3
void FASTCALL call_5_2_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_5_4
#define init_5_2_5_4
void FASTCALL call_5_2_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_5_5
#define init_5_2_5_5
void FASTCALL call_5_2_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_5_6
#define init_5_2_5_6
void FASTCALL call_5_2_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_5_7
#define init_5_2_5_7
void FASTCALL call_5_2_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_5_8
#define init_5_2_5_8
void FASTCALL call_5_2_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_5_9
#define init_5_2_5_9
void FASTCALL call_5_2_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_5_10
#define init_5_2_5_10
void FASTCALL call_5_2_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_5_11
#define init_5_2_5_11
void FASTCALL call_5_2_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_5_12
#define init_5_2_5_12
void FASTCALL call_5_2_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_5_13
#define init_5_2_5_13
void FASTCALL call_5_2_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_5_14
#define init_5_2_5_14
void FASTCALL call_5_2_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_5_15
#define init_5_2_5_15
void FASTCALL call_5_2_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_6_0
#define init_5_2_6_0
void FASTCALL call_5_2_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_6_1
#define init_5_2_6_1
void FASTCALL call_5_2_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_6_2
#define init_5_2_6_2
void FASTCALL call_5_2_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_6_3
#define init_5_2_6_3
void FASTCALL call_5_2_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_6_4
#define init_5_2_6_4
void FASTCALL call_5_2_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_6_5
#define init_5_2_6_5
void FASTCALL call_5_2_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_6_6
#define init_5_2_6_6
void FASTCALL call_5_2_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_6_7
#define init_5_2_6_7
void FASTCALL call_5_2_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_6_8
#define init_5_2_6_8
void FASTCALL call_5_2_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_6_9
#define init_5_2_6_9
void FASTCALL call_5_2_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_6_10
#define init_5_2_6_10
void FASTCALL call_5_2_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_6_11
#define init_5_2_6_11
void FASTCALL call_5_2_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_6_12
#define init_5_2_6_12
void FASTCALL call_5_2_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_6_13
#define init_5_2_6_13
void FASTCALL call_5_2_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_6_14
#define init_5_2_6_14
void FASTCALL call_5_2_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_6_15
#define init_5_2_6_15
void FASTCALL call_5_2_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_7_0
#define init_5_2_7_0
void FASTCALL call_5_2_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_7_1
#define init_5_2_7_1
void FASTCALL call_5_2_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_7_2
#define init_5_2_7_2
void FASTCALL call_5_2_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_7_3
#define init_5_2_7_3
void FASTCALL call_5_2_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_7_4
#define init_5_2_7_4
void FASTCALL call_5_2_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_7_5
#define init_5_2_7_5
void FASTCALL call_5_2_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_7_6
#define init_5_2_7_6
void FASTCALL call_5_2_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_7_7
#define init_5_2_7_7
void FASTCALL call_5_2_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_7_8
#define init_5_2_7_8
void FASTCALL call_5_2_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_7_9
#define init_5_2_7_9
void FASTCALL call_5_2_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_7_10
#define init_5_2_7_10
void FASTCALL call_5_2_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_7_11
#define init_5_2_7_11
void FASTCALL call_5_2_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_7_12
#define init_5_2_7_12
void FASTCALL call_5_2_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_7_13
#define init_5_2_7_13
void FASTCALL call_5_2_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_7_14
#define init_5_2_7_14
void FASTCALL call_5_2_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_7_15
#define init_5_2_7_15
void FASTCALL call_5_2_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_8_0
#define init_5_2_8_0
void FASTCALL call_5_2_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_8_1
#define init_5_2_8_1
void FASTCALL call_5_2_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_8_2
#define init_5_2_8_2
void FASTCALL call_5_2_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_8_3
#define init_5_2_8_3
void FASTCALL call_5_2_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_8_4
#define init_5_2_8_4
void FASTCALL call_5_2_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_8_5
#define init_5_2_8_5
void FASTCALL call_5_2_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_8_6
#define init_5_2_8_6
void FASTCALL call_5_2_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_8_7
#define init_5_2_8_7
void FASTCALL call_5_2_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_8_8
#define init_5_2_8_8
void FASTCALL call_5_2_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_8_9
#define init_5_2_8_9
void FASTCALL call_5_2_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_8_10
#define init_5_2_8_10
void FASTCALL call_5_2_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_8_11
#define init_5_2_8_11
void FASTCALL call_5_2_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_8_12
#define init_5_2_8_12
void FASTCALL call_5_2_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_8_13
#define init_5_2_8_13
void FASTCALL call_5_2_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_8_14
#define init_5_2_8_14
void FASTCALL call_5_2_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_8_15
#define init_5_2_8_15
void FASTCALL call_5_2_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_9_0
#define init_5_2_9_0
void FASTCALL call_5_2_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_9_1
#define init_5_2_9_1
void FASTCALL call_5_2_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_9_2
#define init_5_2_9_2
void FASTCALL call_5_2_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_9_3
#define init_5_2_9_3
void FASTCALL call_5_2_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_9_4
#define init_5_2_9_4
void FASTCALL call_5_2_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_9_5
#define init_5_2_9_5
void FASTCALL call_5_2_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_9_6
#define init_5_2_9_6
void FASTCALL call_5_2_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_9_7
#define init_5_2_9_7
void FASTCALL call_5_2_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_9_8
#define init_5_2_9_8
void FASTCALL call_5_2_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_9_9
#define init_5_2_9_9
void FASTCALL call_5_2_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_9_10
#define init_5_2_9_10
void FASTCALL call_5_2_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_9_11
#define init_5_2_9_11
void FASTCALL call_5_2_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_9_12
#define init_5_2_9_12
void FASTCALL call_5_2_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_9_13
#define init_5_2_9_13
void FASTCALL call_5_2_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_9_14
#define init_5_2_9_14
void FASTCALL call_5_2_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_9_15
#define init_5_2_9_15
void FASTCALL call_5_2_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_10_0
#define init_5_2_10_0
void FASTCALL call_5_2_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_10_1
#define init_5_2_10_1
void FASTCALL call_5_2_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_10_2
#define init_5_2_10_2
void FASTCALL call_5_2_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_10_3
#define init_5_2_10_3
void FASTCALL call_5_2_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_10_4
#define init_5_2_10_4
void FASTCALL call_5_2_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_10_5
#define init_5_2_10_5
void FASTCALL call_5_2_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_10_6
#define init_5_2_10_6
void FASTCALL call_5_2_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_10_7
#define init_5_2_10_7
void FASTCALL call_5_2_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_10_8
#define init_5_2_10_8
void FASTCALL call_5_2_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_10_9
#define init_5_2_10_9
void FASTCALL call_5_2_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_10_10
#define init_5_2_10_10
void FASTCALL call_5_2_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_10_11
#define init_5_2_10_11
void FASTCALL call_5_2_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_10_12
#define init_5_2_10_12
void FASTCALL call_5_2_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_10_13
#define init_5_2_10_13
void FASTCALL call_5_2_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_10_14
#define init_5_2_10_14
void FASTCALL call_5_2_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_10_15
#define init_5_2_10_15
void FASTCALL call_5_2_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_11_0
#define init_5_2_11_0
void FASTCALL call_5_2_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_11_1
#define init_5_2_11_1
void FASTCALL call_5_2_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_11_2
#define init_5_2_11_2
void FASTCALL call_5_2_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_11_3
#define init_5_2_11_3
void FASTCALL call_5_2_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_11_4
#define init_5_2_11_4
void FASTCALL call_5_2_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_11_5
#define init_5_2_11_5
void FASTCALL call_5_2_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_11_6
#define init_5_2_11_6
void FASTCALL call_5_2_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_11_7
#define init_5_2_11_7
void FASTCALL call_5_2_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_11_8
#define init_5_2_11_8
void FASTCALL call_5_2_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_11_9
#define init_5_2_11_9
void FASTCALL call_5_2_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_11_10
#define init_5_2_11_10
void FASTCALL call_5_2_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_11_11
#define init_5_2_11_11
void FASTCALL call_5_2_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_11_12
#define init_5_2_11_12
void FASTCALL call_5_2_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_11_13
#define init_5_2_11_13
void FASTCALL call_5_2_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_11_14
#define init_5_2_11_14
void FASTCALL call_5_2_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_11_15
#define init_5_2_11_15
void FASTCALL call_5_2_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_12_0
#define init_5_2_12_0
void FASTCALL call_5_2_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_12_1
#define init_5_2_12_1
void FASTCALL call_5_2_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_12_2
#define init_5_2_12_2
void FASTCALL call_5_2_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_12_3
#define init_5_2_12_3
void FASTCALL call_5_2_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_12_4
#define init_5_2_12_4
void FASTCALL call_5_2_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_12_5
#define init_5_2_12_5
void FASTCALL call_5_2_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_12_6
#define init_5_2_12_6
void FASTCALL call_5_2_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_12_7
#define init_5_2_12_7
void FASTCALL call_5_2_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_12_8
#define init_5_2_12_8
void FASTCALL call_5_2_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_12_9
#define init_5_2_12_9
void FASTCALL call_5_2_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_12_10
#define init_5_2_12_10
void FASTCALL call_5_2_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_12_11
#define init_5_2_12_11
void FASTCALL call_5_2_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_12_12
#define init_5_2_12_12
void FASTCALL call_5_2_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_12_13
#define init_5_2_12_13
void FASTCALL call_5_2_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_12_14
#define init_5_2_12_14
void FASTCALL call_5_2_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_12_15
#define init_5_2_12_15
void FASTCALL call_5_2_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_13_0
#define init_5_2_13_0
void FASTCALL call_5_2_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_13_1
#define init_5_2_13_1
void FASTCALL call_5_2_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_13_2
#define init_5_2_13_2
void FASTCALL call_5_2_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_13_3
#define init_5_2_13_3
void FASTCALL call_5_2_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_13_4
#define init_5_2_13_4
void FASTCALL call_5_2_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_13_5
#define init_5_2_13_5
void FASTCALL call_5_2_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_13_6
#define init_5_2_13_6
void FASTCALL call_5_2_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_13_7
#define init_5_2_13_7
void FASTCALL call_5_2_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_13_8
#define init_5_2_13_8
void FASTCALL call_5_2_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_13_9
#define init_5_2_13_9
void FASTCALL call_5_2_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_13_10
#define init_5_2_13_10
void FASTCALL call_5_2_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_13_11
#define init_5_2_13_11
void FASTCALL call_5_2_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_13_12
#define init_5_2_13_12
void FASTCALL call_5_2_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_13_13
#define init_5_2_13_13
void FASTCALL call_5_2_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_13_14
#define init_5_2_13_14
void FASTCALL call_5_2_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_13_15
#define init_5_2_13_15
void FASTCALL call_5_2_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_14_0
#define init_5_2_14_0
void FASTCALL call_5_2_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_14_1
#define init_5_2_14_1
void FASTCALL call_5_2_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_14_2
#define init_5_2_14_2
void FASTCALL call_5_2_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_14_3
#define init_5_2_14_3
void FASTCALL call_5_2_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_14_4
#define init_5_2_14_4
void FASTCALL call_5_2_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_14_5
#define init_5_2_14_5
void FASTCALL call_5_2_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_14_6
#define init_5_2_14_6
void FASTCALL call_5_2_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_14_7
#define init_5_2_14_7
void FASTCALL call_5_2_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_14_8
#define init_5_2_14_8
void FASTCALL call_5_2_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_14_9
#define init_5_2_14_9
void FASTCALL call_5_2_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_14_10
#define init_5_2_14_10
void FASTCALL call_5_2_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_14_11
#define init_5_2_14_11
void FASTCALL call_5_2_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_14_12
#define init_5_2_14_12
void FASTCALL call_5_2_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_14_13
#define init_5_2_14_13
void FASTCALL call_5_2_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_14_14
#define init_5_2_14_14
void FASTCALL call_5_2_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_14_15
#define init_5_2_14_15
void FASTCALL call_5_2_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_15_0
#define init_5_2_15_0
void FASTCALL call_5_2_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_15_1
#define init_5_2_15_1
void FASTCALL call_5_2_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_15_2
#define init_5_2_15_2
void FASTCALL call_5_2_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_15_3
#define init_5_2_15_3
void FASTCALL call_5_2_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_15_4
#define init_5_2_15_4
void FASTCALL call_5_2_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_15_5
#define init_5_2_15_5
void FASTCALL call_5_2_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_15_6
#define init_5_2_15_6
void FASTCALL call_5_2_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_15_7
#define init_5_2_15_7
void FASTCALL call_5_2_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_15_8
#define init_5_2_15_8
void FASTCALL call_5_2_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_15_9
#define init_5_2_15_9
void FASTCALL call_5_2_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_15_10
#define init_5_2_15_10
void FASTCALL call_5_2_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_15_11
#define init_5_2_15_11
void FASTCALL call_5_2_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_15_12
#define init_5_2_15_12
void FASTCALL call_5_2_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_15_13
#define init_5_2_15_13
void FASTCALL call_5_2_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_15_14
#define init_5_2_15_14
void FASTCALL call_5_2_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_2_15_15
#define init_5_2_15_15
void FASTCALL call_5_2_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_0_0
#define init_5_3_0_0
void FASTCALL call_5_3_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_0_1
#define init_5_3_0_1
void FASTCALL call_5_3_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_0_2
#define init_5_3_0_2
void FASTCALL call_5_3_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_0_3
#define init_5_3_0_3
void FASTCALL call_5_3_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_0_4
#define init_5_3_0_4
void FASTCALL call_5_3_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_0_5
#define init_5_3_0_5
void FASTCALL call_5_3_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_0_6
#define init_5_3_0_6
void FASTCALL call_5_3_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_0_7
#define init_5_3_0_7
void FASTCALL call_5_3_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_0_8
#define init_5_3_0_8
void FASTCALL call_5_3_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_0_9
#define init_5_3_0_9
void FASTCALL call_5_3_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_0_10
#define init_5_3_0_10
void FASTCALL call_5_3_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_0_11
#define init_5_3_0_11
void FASTCALL call_5_3_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_0_12
#define init_5_3_0_12
void FASTCALL call_5_3_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_0_13
#define init_5_3_0_13
void FASTCALL call_5_3_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_0_14
#define init_5_3_0_14
void FASTCALL call_5_3_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_0_15
#define init_5_3_0_15
void FASTCALL call_5_3_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_1_0
#define init_5_3_1_0
void FASTCALL call_5_3_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_1_1
#define init_5_3_1_1
void FASTCALL call_5_3_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_1_2
#define init_5_3_1_2
void FASTCALL call_5_3_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_1_3
#define init_5_3_1_3
void FASTCALL call_5_3_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_1_4
#define init_5_3_1_4
void FASTCALL call_5_3_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_1_5
#define init_5_3_1_5
void FASTCALL call_5_3_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_1_6
#define init_5_3_1_6
void FASTCALL call_5_3_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_1_7
#define init_5_3_1_7
void FASTCALL call_5_3_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_1_8
#define init_5_3_1_8
void FASTCALL call_5_3_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_1_9
#define init_5_3_1_9
void FASTCALL call_5_3_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_1_10
#define init_5_3_1_10
void FASTCALL call_5_3_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_1_11
#define init_5_3_1_11
void FASTCALL call_5_3_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_1_12
#define init_5_3_1_12
void FASTCALL call_5_3_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_1_13
#define init_5_3_1_13
void FASTCALL call_5_3_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_1_14
#define init_5_3_1_14
void FASTCALL call_5_3_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_1_15
#define init_5_3_1_15
void FASTCALL call_5_3_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_2_0
#define init_5_3_2_0
void FASTCALL call_5_3_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_2_1
#define init_5_3_2_1
void FASTCALL call_5_3_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_2_2
#define init_5_3_2_2
void FASTCALL call_5_3_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_2_3
#define init_5_3_2_3
void FASTCALL call_5_3_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_2_4
#define init_5_3_2_4
void FASTCALL call_5_3_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_2_5
#define init_5_3_2_5
void FASTCALL call_5_3_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_2_6
#define init_5_3_2_6
void FASTCALL call_5_3_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_2_7
#define init_5_3_2_7
void FASTCALL call_5_3_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_2_8
#define init_5_3_2_8
void FASTCALL call_5_3_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_2_9
#define init_5_3_2_9
void FASTCALL call_5_3_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_2_10
#define init_5_3_2_10
void FASTCALL call_5_3_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_2_11
#define init_5_3_2_11
void FASTCALL call_5_3_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_2_12
#define init_5_3_2_12
void FASTCALL call_5_3_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_2_13
#define init_5_3_2_13
void FASTCALL call_5_3_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_2_14
#define init_5_3_2_14
void FASTCALL call_5_3_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_2_15
#define init_5_3_2_15
void FASTCALL call_5_3_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_3_0
#define init_5_3_3_0
void FASTCALL call_5_3_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_3_1
#define init_5_3_3_1
void FASTCALL call_5_3_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_3_2
#define init_5_3_3_2
void FASTCALL call_5_3_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_3_3
#define init_5_3_3_3
void FASTCALL call_5_3_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_3_4
#define init_5_3_3_4
void FASTCALL call_5_3_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_3_5
#define init_5_3_3_5
void FASTCALL call_5_3_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_3_6
#define init_5_3_3_6
void FASTCALL call_5_3_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_3_7
#define init_5_3_3_7
void FASTCALL call_5_3_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_3_8
#define init_5_3_3_8
void FASTCALL call_5_3_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_3_9
#define init_5_3_3_9
void FASTCALL call_5_3_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_3_10
#define init_5_3_3_10
void FASTCALL call_5_3_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_3_11
#define init_5_3_3_11
void FASTCALL call_5_3_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_3_12
#define init_5_3_3_12
void FASTCALL call_5_3_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_3_13
#define init_5_3_3_13
void FASTCALL call_5_3_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_3_14
#define init_5_3_3_14
void FASTCALL call_5_3_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_3_15
#define init_5_3_3_15
void FASTCALL call_5_3_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_4_0
#define init_5_3_4_0
void FASTCALL call_5_3_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_4_1
#define init_5_3_4_1
void FASTCALL call_5_3_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_4_2
#define init_5_3_4_2
void FASTCALL call_5_3_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_4_3
#define init_5_3_4_3
void FASTCALL call_5_3_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_4_4
#define init_5_3_4_4
void FASTCALL call_5_3_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_4_5
#define init_5_3_4_5
void FASTCALL call_5_3_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_4_6
#define init_5_3_4_6
void FASTCALL call_5_3_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_4_7
#define init_5_3_4_7
void FASTCALL call_5_3_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_4_8
#define init_5_3_4_8
void FASTCALL call_5_3_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_4_9
#define init_5_3_4_9
void FASTCALL call_5_3_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_4_10
#define init_5_3_4_10
void FASTCALL call_5_3_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_4_11
#define init_5_3_4_11
void FASTCALL call_5_3_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_4_12
#define init_5_3_4_12
void FASTCALL call_5_3_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_4_13
#define init_5_3_4_13
void FASTCALL call_5_3_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_4_14
#define init_5_3_4_14
void FASTCALL call_5_3_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_4_15
#define init_5_3_4_15
void FASTCALL call_5_3_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_5_0
#define init_5_3_5_0
void FASTCALL call_5_3_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_5_1
#define init_5_3_5_1
void FASTCALL call_5_3_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_5_2
#define init_5_3_5_2
void FASTCALL call_5_3_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_5_3
#define init_5_3_5_3
void FASTCALL call_5_3_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_5_4
#define init_5_3_5_4
void FASTCALL call_5_3_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_5_5
#define init_5_3_5_5
void FASTCALL call_5_3_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_5_6
#define init_5_3_5_6
void FASTCALL call_5_3_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_5_7
#define init_5_3_5_7
void FASTCALL call_5_3_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_5_8
#define init_5_3_5_8
void FASTCALL call_5_3_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_5_9
#define init_5_3_5_9
void FASTCALL call_5_3_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_5_10
#define init_5_3_5_10
void FASTCALL call_5_3_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_5_11
#define init_5_3_5_11
void FASTCALL call_5_3_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_5_12
#define init_5_3_5_12
void FASTCALL call_5_3_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_5_13
#define init_5_3_5_13
void FASTCALL call_5_3_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_5_14
#define init_5_3_5_14
void FASTCALL call_5_3_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_5_15
#define init_5_3_5_15
void FASTCALL call_5_3_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_6_0
#define init_5_3_6_0
void FASTCALL call_5_3_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_6_1
#define init_5_3_6_1
void FASTCALL call_5_3_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_6_2
#define init_5_3_6_2
void FASTCALL call_5_3_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_6_3
#define init_5_3_6_3
void FASTCALL call_5_3_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_6_4
#define init_5_3_6_4
void FASTCALL call_5_3_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_6_5
#define init_5_3_6_5
void FASTCALL call_5_3_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_6_6
#define init_5_3_6_6
void FASTCALL call_5_3_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_6_7
#define init_5_3_6_7
void FASTCALL call_5_3_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_6_8
#define init_5_3_6_8
void FASTCALL call_5_3_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_6_9
#define init_5_3_6_9
void FASTCALL call_5_3_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_6_10
#define init_5_3_6_10
void FASTCALL call_5_3_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_6_11
#define init_5_3_6_11
void FASTCALL call_5_3_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_6_12
#define init_5_3_6_12
void FASTCALL call_5_3_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_6_13
#define init_5_3_6_13
void FASTCALL call_5_3_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_6_14
#define init_5_3_6_14
void FASTCALL call_5_3_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_6_15
#define init_5_3_6_15
void FASTCALL call_5_3_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_7_0
#define init_5_3_7_0
void FASTCALL call_5_3_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_7_1
#define init_5_3_7_1
void FASTCALL call_5_3_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_7_2
#define init_5_3_7_2
void FASTCALL call_5_3_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_7_3
#define init_5_3_7_3
void FASTCALL call_5_3_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_7_4
#define init_5_3_7_4
void FASTCALL call_5_3_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_7_5
#define init_5_3_7_5
void FASTCALL call_5_3_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_7_6
#define init_5_3_7_6
void FASTCALL call_5_3_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_7_7
#define init_5_3_7_7
void FASTCALL call_5_3_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_7_8
#define init_5_3_7_8
void FASTCALL call_5_3_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_7_9
#define init_5_3_7_9
void FASTCALL call_5_3_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_7_10
#define init_5_3_7_10
void FASTCALL call_5_3_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_7_11
#define init_5_3_7_11
void FASTCALL call_5_3_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_7_12
#define init_5_3_7_12
void FASTCALL call_5_3_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_7_13
#define init_5_3_7_13
void FASTCALL call_5_3_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_7_14
#define init_5_3_7_14
void FASTCALL call_5_3_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_7_15
#define init_5_3_7_15
void FASTCALL call_5_3_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_8_0
#define init_5_3_8_0
void FASTCALL call_5_3_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_8_1
#define init_5_3_8_1
void FASTCALL call_5_3_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_8_2
#define init_5_3_8_2
void FASTCALL call_5_3_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_8_3
#define init_5_3_8_3
void FASTCALL call_5_3_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_8_4
#define init_5_3_8_4
void FASTCALL call_5_3_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_8_5
#define init_5_3_8_5
void FASTCALL call_5_3_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_8_6
#define init_5_3_8_6
void FASTCALL call_5_3_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_8_7
#define init_5_3_8_7
void FASTCALL call_5_3_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_8_8
#define init_5_3_8_8
void FASTCALL call_5_3_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_8_9
#define init_5_3_8_9
void FASTCALL call_5_3_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_8_10
#define init_5_3_8_10
void FASTCALL call_5_3_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_8_11
#define init_5_3_8_11
void FASTCALL call_5_3_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_8_12
#define init_5_3_8_12
void FASTCALL call_5_3_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_8_13
#define init_5_3_8_13
void FASTCALL call_5_3_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_8_14
#define init_5_3_8_14
void FASTCALL call_5_3_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_8_15
#define init_5_3_8_15
void FASTCALL call_5_3_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_9_0
#define init_5_3_9_0
void FASTCALL call_5_3_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_9_1
#define init_5_3_9_1
void FASTCALL call_5_3_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_9_2
#define init_5_3_9_2
void FASTCALL call_5_3_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_9_3
#define init_5_3_9_3
void FASTCALL call_5_3_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_9_4
#define init_5_3_9_4
void FASTCALL call_5_3_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_9_5
#define init_5_3_9_5
void FASTCALL call_5_3_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_9_6
#define init_5_3_9_6
void FASTCALL call_5_3_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_9_7
#define init_5_3_9_7
void FASTCALL call_5_3_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_9_8
#define init_5_3_9_8
void FASTCALL call_5_3_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_9_9
#define init_5_3_9_9
void FASTCALL call_5_3_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_9_10
#define init_5_3_9_10
void FASTCALL call_5_3_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_9_11
#define init_5_3_9_11
void FASTCALL call_5_3_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_9_12
#define init_5_3_9_12
void FASTCALL call_5_3_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_9_13
#define init_5_3_9_13
void FASTCALL call_5_3_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_9_14
#define init_5_3_9_14
void FASTCALL call_5_3_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_9_15
#define init_5_3_9_15
void FASTCALL call_5_3_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_10_0
#define init_5_3_10_0
void FASTCALL call_5_3_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_10_1
#define init_5_3_10_1
void FASTCALL call_5_3_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_10_2
#define init_5_3_10_2
void FASTCALL call_5_3_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_10_3
#define init_5_3_10_3
void FASTCALL call_5_3_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_10_4
#define init_5_3_10_4
void FASTCALL call_5_3_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_10_5
#define init_5_3_10_5
void FASTCALL call_5_3_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_10_6
#define init_5_3_10_6
void FASTCALL call_5_3_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_10_7
#define init_5_3_10_7
void FASTCALL call_5_3_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_10_8
#define init_5_3_10_8
void FASTCALL call_5_3_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_10_9
#define init_5_3_10_9
void FASTCALL call_5_3_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_10_10
#define init_5_3_10_10
void FASTCALL call_5_3_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_10_11
#define init_5_3_10_11
void FASTCALL call_5_3_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_10_12
#define init_5_3_10_12
void FASTCALL call_5_3_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_10_13
#define init_5_3_10_13
void FASTCALL call_5_3_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_10_14
#define init_5_3_10_14
void FASTCALL call_5_3_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_10_15
#define init_5_3_10_15
void FASTCALL call_5_3_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_11_0
#define init_5_3_11_0
void FASTCALL call_5_3_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_11_1
#define init_5_3_11_1
void FASTCALL call_5_3_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_11_2
#define init_5_3_11_2
void FASTCALL call_5_3_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_11_3
#define init_5_3_11_3
void FASTCALL call_5_3_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_11_4
#define init_5_3_11_4
void FASTCALL call_5_3_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_11_5
#define init_5_3_11_5
void FASTCALL call_5_3_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_11_6
#define init_5_3_11_6
void FASTCALL call_5_3_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_11_7
#define init_5_3_11_7
void FASTCALL call_5_3_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_11_8
#define init_5_3_11_8
void FASTCALL call_5_3_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_11_9
#define init_5_3_11_9
void FASTCALL call_5_3_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_11_10
#define init_5_3_11_10
void FASTCALL call_5_3_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_11_11
#define init_5_3_11_11
void FASTCALL call_5_3_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_11_12
#define init_5_3_11_12
void FASTCALL call_5_3_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_11_13
#define init_5_3_11_13
void FASTCALL call_5_3_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_11_14
#define init_5_3_11_14
void FASTCALL call_5_3_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_11_15
#define init_5_3_11_15
void FASTCALL call_5_3_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_12_0
#define init_5_3_12_0
void FASTCALL call_5_3_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_12_1
#define init_5_3_12_1
void FASTCALL call_5_3_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_12_2
#define init_5_3_12_2
void FASTCALL call_5_3_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_12_3
#define init_5_3_12_3
void FASTCALL call_5_3_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_12_4
#define init_5_3_12_4
void FASTCALL call_5_3_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_12_5
#define init_5_3_12_5
void FASTCALL call_5_3_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_12_6
#define init_5_3_12_6
void FASTCALL call_5_3_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_12_7
#define init_5_3_12_7
void FASTCALL call_5_3_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_12_8
#define init_5_3_12_8
void FASTCALL call_5_3_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_12_9
#define init_5_3_12_9
void FASTCALL call_5_3_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_12_10
#define init_5_3_12_10
void FASTCALL call_5_3_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_12_11
#define init_5_3_12_11
void FASTCALL call_5_3_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_12_12
#define init_5_3_12_12
void FASTCALL call_5_3_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_12_13
#define init_5_3_12_13
void FASTCALL call_5_3_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_12_14
#define init_5_3_12_14
void FASTCALL call_5_3_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_12_15
#define init_5_3_12_15
void FASTCALL call_5_3_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_13_0
#define init_5_3_13_0
void FASTCALL call_5_3_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_13_1
#define init_5_3_13_1
void FASTCALL call_5_3_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_13_2
#define init_5_3_13_2
void FASTCALL call_5_3_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_13_3
#define init_5_3_13_3
void FASTCALL call_5_3_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_13_4
#define init_5_3_13_4
void FASTCALL call_5_3_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_13_5
#define init_5_3_13_5
void FASTCALL call_5_3_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_13_6
#define init_5_3_13_6
void FASTCALL call_5_3_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_13_7
#define init_5_3_13_7
void FASTCALL call_5_3_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_13_8
#define init_5_3_13_8
void FASTCALL call_5_3_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_13_9
#define init_5_3_13_9
void FASTCALL call_5_3_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_13_10
#define init_5_3_13_10
void FASTCALL call_5_3_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_13_11
#define init_5_3_13_11
void FASTCALL call_5_3_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_13_12
#define init_5_3_13_12
void FASTCALL call_5_3_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_13_13
#define init_5_3_13_13
void FASTCALL call_5_3_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_13_14
#define init_5_3_13_14
void FASTCALL call_5_3_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_13_15
#define init_5_3_13_15
void FASTCALL call_5_3_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_14_0
#define init_5_3_14_0
void FASTCALL call_5_3_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_14_1
#define init_5_3_14_1
void FASTCALL call_5_3_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_14_2
#define init_5_3_14_2
void FASTCALL call_5_3_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_14_3
#define init_5_3_14_3
void FASTCALL call_5_3_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_14_4
#define init_5_3_14_4
void FASTCALL call_5_3_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_14_5
#define init_5_3_14_5
void FASTCALL call_5_3_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_14_6
#define init_5_3_14_6
void FASTCALL call_5_3_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_14_7
#define init_5_3_14_7
void FASTCALL call_5_3_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_14_8
#define init_5_3_14_8
void FASTCALL call_5_3_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_14_9
#define init_5_3_14_9
void FASTCALL call_5_3_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_14_10
#define init_5_3_14_10
void FASTCALL call_5_3_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_14_11
#define init_5_3_14_11
void FASTCALL call_5_3_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_14_12
#define init_5_3_14_12
void FASTCALL call_5_3_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_14_13
#define init_5_3_14_13
void FASTCALL call_5_3_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_14_14
#define init_5_3_14_14
void FASTCALL call_5_3_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_14_15
#define init_5_3_14_15
void FASTCALL call_5_3_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_15_0
#define init_5_3_15_0
void FASTCALL call_5_3_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_15_1
#define init_5_3_15_1
void FASTCALL call_5_3_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_15_2
#define init_5_3_15_2
void FASTCALL call_5_3_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_15_3
#define init_5_3_15_3
void FASTCALL call_5_3_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_15_4
#define init_5_3_15_4
void FASTCALL call_5_3_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_15_5
#define init_5_3_15_5
void FASTCALL call_5_3_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_15_6
#define init_5_3_15_6
void FASTCALL call_5_3_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_15_7
#define init_5_3_15_7
void FASTCALL call_5_3_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_15_8
#define init_5_3_15_8
void FASTCALL call_5_3_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_15_9
#define init_5_3_15_9
void FASTCALL call_5_3_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_15_10
#define init_5_3_15_10
void FASTCALL call_5_3_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_15_11
#define init_5_3_15_11
void FASTCALL call_5_3_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_15_12
#define init_5_3_15_12
void FASTCALL call_5_3_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_15_13
#define init_5_3_15_13
void FASTCALL call_5_3_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_15_14
#define init_5_3_15_14
void FASTCALL call_5_3_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_3_15_15
#define init_5_3_15_15
void FASTCALL call_5_3_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_0_0
#define init_5_4_0_0
void FASTCALL call_5_4_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_0_1
#define init_5_4_0_1
void FASTCALL call_5_4_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_0_2
#define init_5_4_0_2
void FASTCALL call_5_4_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_0_3
#define init_5_4_0_3
void FASTCALL call_5_4_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_0_4
#define init_5_4_0_4
void FASTCALL call_5_4_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_0_5
#define init_5_4_0_5
void FASTCALL call_5_4_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_0_6
#define init_5_4_0_6
void FASTCALL call_5_4_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_0_7
#define init_5_4_0_7
void FASTCALL call_5_4_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_0_8
#define init_5_4_0_8
void FASTCALL call_5_4_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_0_9
#define init_5_4_0_9
void FASTCALL call_5_4_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_0_10
#define init_5_4_0_10
void FASTCALL call_5_4_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_0_11
#define init_5_4_0_11
void FASTCALL call_5_4_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_0_12
#define init_5_4_0_12
void FASTCALL call_5_4_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_0_13
#define init_5_4_0_13
void FASTCALL call_5_4_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_0_14
#define init_5_4_0_14
void FASTCALL call_5_4_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_0_15
#define init_5_4_0_15
void FASTCALL call_5_4_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_1_0
#define init_5_4_1_0
void FASTCALL call_5_4_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_1_1
#define init_5_4_1_1
void FASTCALL call_5_4_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_1_2
#define init_5_4_1_2
void FASTCALL call_5_4_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_1_3
#define init_5_4_1_3
void FASTCALL call_5_4_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_1_4
#define init_5_4_1_4
void FASTCALL call_5_4_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_1_5
#define init_5_4_1_5
void FASTCALL call_5_4_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_1_6
#define init_5_4_1_6
void FASTCALL call_5_4_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_1_7
#define init_5_4_1_7
void FASTCALL call_5_4_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_1_8
#define init_5_4_1_8
void FASTCALL call_5_4_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_1_9
#define init_5_4_1_9
void FASTCALL call_5_4_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_1_10
#define init_5_4_1_10
void FASTCALL call_5_4_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_1_11
#define init_5_4_1_11
void FASTCALL call_5_4_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_1_12
#define init_5_4_1_12
void FASTCALL call_5_4_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_1_13
#define init_5_4_1_13
void FASTCALL call_5_4_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_1_14
#define init_5_4_1_14
void FASTCALL call_5_4_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_1_15
#define init_5_4_1_15
void FASTCALL call_5_4_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_2_0
#define init_5_4_2_0
void FASTCALL call_5_4_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_2_1
#define init_5_4_2_1
void FASTCALL call_5_4_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_2_2
#define init_5_4_2_2
void FASTCALL call_5_4_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_2_3
#define init_5_4_2_3
void FASTCALL call_5_4_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_2_4
#define init_5_4_2_4
void FASTCALL call_5_4_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_2_5
#define init_5_4_2_5
void FASTCALL call_5_4_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_2_6
#define init_5_4_2_6
void FASTCALL call_5_4_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_2_7
#define init_5_4_2_7
void FASTCALL call_5_4_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_2_8
#define init_5_4_2_8
void FASTCALL call_5_4_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_2_9
#define init_5_4_2_9
void FASTCALL call_5_4_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_2_10
#define init_5_4_2_10
void FASTCALL call_5_4_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_2_11
#define init_5_4_2_11
void FASTCALL call_5_4_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_2_12
#define init_5_4_2_12
void FASTCALL call_5_4_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_2_13
#define init_5_4_2_13
void FASTCALL call_5_4_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_2_14
#define init_5_4_2_14
void FASTCALL call_5_4_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_2_15
#define init_5_4_2_15
void FASTCALL call_5_4_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_3_0
#define init_5_4_3_0
void FASTCALL call_5_4_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_3_1
#define init_5_4_3_1
void FASTCALL call_5_4_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_3_2
#define init_5_4_3_2
void FASTCALL call_5_4_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_3_3
#define init_5_4_3_3
void FASTCALL call_5_4_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_3_4
#define init_5_4_3_4
void FASTCALL call_5_4_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_3_5
#define init_5_4_3_5
void FASTCALL call_5_4_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_3_6
#define init_5_4_3_6
void FASTCALL call_5_4_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_3_7
#define init_5_4_3_7
void FASTCALL call_5_4_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_3_8
#define init_5_4_3_8
void FASTCALL call_5_4_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_3_9
#define init_5_4_3_9
void FASTCALL call_5_4_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_3_10
#define init_5_4_3_10
void FASTCALL call_5_4_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_3_11
#define init_5_4_3_11
void FASTCALL call_5_4_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_3_12
#define init_5_4_3_12
void FASTCALL call_5_4_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_3_13
#define init_5_4_3_13
void FASTCALL call_5_4_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_3_14
#define init_5_4_3_14
void FASTCALL call_5_4_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_3_15
#define init_5_4_3_15
void FASTCALL call_5_4_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_4_0
#define init_5_4_4_0
void FASTCALL call_5_4_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_4_1
#define init_5_4_4_1
void FASTCALL call_5_4_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_4_2
#define init_5_4_4_2
void FASTCALL call_5_4_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_4_3
#define init_5_4_4_3
void FASTCALL call_5_4_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_4_4
#define init_5_4_4_4
void FASTCALL call_5_4_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_4_5
#define init_5_4_4_5
void FASTCALL call_5_4_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_4_6
#define init_5_4_4_6
void FASTCALL call_5_4_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_4_7
#define init_5_4_4_7
void FASTCALL call_5_4_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_4_8
#define init_5_4_4_8
void FASTCALL call_5_4_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_4_9
#define init_5_4_4_9
void FASTCALL call_5_4_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_4_10
#define init_5_4_4_10
void FASTCALL call_5_4_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_4_11
#define init_5_4_4_11
void FASTCALL call_5_4_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_4_12
#define init_5_4_4_12
void FASTCALL call_5_4_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_4_13
#define init_5_4_4_13
void FASTCALL call_5_4_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_4_14
#define init_5_4_4_14
void FASTCALL call_5_4_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_4_15
#define init_5_4_4_15
void FASTCALL call_5_4_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_5_0
#define init_5_4_5_0
void FASTCALL call_5_4_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_5_1
#define init_5_4_5_1
void FASTCALL call_5_4_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_5_2
#define init_5_4_5_2
void FASTCALL call_5_4_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_5_3
#define init_5_4_5_3
void FASTCALL call_5_4_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_5_4
#define init_5_4_5_4
void FASTCALL call_5_4_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_5_5
#define init_5_4_5_5
void FASTCALL call_5_4_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_5_6
#define init_5_4_5_6
void FASTCALL call_5_4_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_5_7
#define init_5_4_5_7
void FASTCALL call_5_4_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_5_8
#define init_5_4_5_8
void FASTCALL call_5_4_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_5_9
#define init_5_4_5_9
void FASTCALL call_5_4_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_5_10
#define init_5_4_5_10
void FASTCALL call_5_4_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_5_11
#define init_5_4_5_11
void FASTCALL call_5_4_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_5_12
#define init_5_4_5_12
void FASTCALL call_5_4_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_5_13
#define init_5_4_5_13
void FASTCALL call_5_4_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_5_14
#define init_5_4_5_14
void FASTCALL call_5_4_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_5_15
#define init_5_4_5_15
void FASTCALL call_5_4_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_6_0
#define init_5_4_6_0
void FASTCALL call_5_4_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_6_1
#define init_5_4_6_1
void FASTCALL call_5_4_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_6_2
#define init_5_4_6_2
void FASTCALL call_5_4_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_6_3
#define init_5_4_6_3
void FASTCALL call_5_4_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_6_4
#define init_5_4_6_4
void FASTCALL call_5_4_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_6_5
#define init_5_4_6_5
void FASTCALL call_5_4_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_6_6
#define init_5_4_6_6
void FASTCALL call_5_4_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_6_7
#define init_5_4_6_7
void FASTCALL call_5_4_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_6_8
#define init_5_4_6_8
void FASTCALL call_5_4_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_6_9
#define init_5_4_6_9
void FASTCALL call_5_4_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_6_10
#define init_5_4_6_10
void FASTCALL call_5_4_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_6_11
#define init_5_4_6_11
void FASTCALL call_5_4_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_6_12
#define init_5_4_6_12
void FASTCALL call_5_4_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_6_13
#define init_5_4_6_13
void FASTCALL call_5_4_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_6_14
#define init_5_4_6_14
void FASTCALL call_5_4_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_6_15
#define init_5_4_6_15
void FASTCALL call_5_4_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_7_0
#define init_5_4_7_0
void FASTCALL call_5_4_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_7_1
#define init_5_4_7_1
void FASTCALL call_5_4_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_7_2
#define init_5_4_7_2
void FASTCALL call_5_4_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_7_3
#define init_5_4_7_3
void FASTCALL call_5_4_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_7_4
#define init_5_4_7_4
void FASTCALL call_5_4_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_7_5
#define init_5_4_7_5
void FASTCALL call_5_4_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_7_6
#define init_5_4_7_6
void FASTCALL call_5_4_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_7_7
#define init_5_4_7_7
void FASTCALL call_5_4_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_7_8
#define init_5_4_7_8
void FASTCALL call_5_4_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_7_9
#define init_5_4_7_9
void FASTCALL call_5_4_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_7_10
#define init_5_4_7_10
void FASTCALL call_5_4_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_7_11
#define init_5_4_7_11
void FASTCALL call_5_4_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_7_12
#define init_5_4_7_12
void FASTCALL call_5_4_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_7_13
#define init_5_4_7_13
void FASTCALL call_5_4_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_7_14
#define init_5_4_7_14
void FASTCALL call_5_4_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_7_15
#define init_5_4_7_15
void FASTCALL call_5_4_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_8_0
#define init_5_4_8_0
void FASTCALL call_5_4_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_8_1
#define init_5_4_8_1
void FASTCALL call_5_4_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_8_2
#define init_5_4_8_2
void FASTCALL call_5_4_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_8_3
#define init_5_4_8_3
void FASTCALL call_5_4_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_8_4
#define init_5_4_8_4
void FASTCALL call_5_4_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_8_5
#define init_5_4_8_5
void FASTCALL call_5_4_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_8_6
#define init_5_4_8_6
void FASTCALL call_5_4_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_8_7
#define init_5_4_8_7
void FASTCALL call_5_4_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_8_8
#define init_5_4_8_8
void FASTCALL call_5_4_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_8_9
#define init_5_4_8_9
void FASTCALL call_5_4_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_8_10
#define init_5_4_8_10
void FASTCALL call_5_4_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_8_11
#define init_5_4_8_11
void FASTCALL call_5_4_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_8_12
#define init_5_4_8_12
void FASTCALL call_5_4_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_8_13
#define init_5_4_8_13
void FASTCALL call_5_4_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_8_14
#define init_5_4_8_14
void FASTCALL call_5_4_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_8_15
#define init_5_4_8_15
void FASTCALL call_5_4_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_9_0
#define init_5_4_9_0
void FASTCALL call_5_4_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_9_1
#define init_5_4_9_1
void FASTCALL call_5_4_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_9_2
#define init_5_4_9_2
void FASTCALL call_5_4_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_9_3
#define init_5_4_9_3
void FASTCALL call_5_4_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_9_4
#define init_5_4_9_4
void FASTCALL call_5_4_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_9_5
#define init_5_4_9_5
void FASTCALL call_5_4_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_9_6
#define init_5_4_9_6
void FASTCALL call_5_4_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_9_7
#define init_5_4_9_7
void FASTCALL call_5_4_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_9_8
#define init_5_4_9_8
void FASTCALL call_5_4_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_9_9
#define init_5_4_9_9
void FASTCALL call_5_4_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_9_10
#define init_5_4_9_10
void FASTCALL call_5_4_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_9_11
#define init_5_4_9_11
void FASTCALL call_5_4_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_9_12
#define init_5_4_9_12
void FASTCALL call_5_4_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_9_13
#define init_5_4_9_13
void FASTCALL call_5_4_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_9_14
#define init_5_4_9_14
void FASTCALL call_5_4_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_9_15
#define init_5_4_9_15
void FASTCALL call_5_4_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_10_0
#define init_5_4_10_0
void FASTCALL call_5_4_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_10_1
#define init_5_4_10_1
void FASTCALL call_5_4_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_10_2
#define init_5_4_10_2
void FASTCALL call_5_4_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_10_3
#define init_5_4_10_3
void FASTCALL call_5_4_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_10_4
#define init_5_4_10_4
void FASTCALL call_5_4_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_10_5
#define init_5_4_10_5
void FASTCALL call_5_4_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_10_6
#define init_5_4_10_6
void FASTCALL call_5_4_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_10_7
#define init_5_4_10_7
void FASTCALL call_5_4_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_10_8
#define init_5_4_10_8
void FASTCALL call_5_4_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_10_9
#define init_5_4_10_9
void FASTCALL call_5_4_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_10_10
#define init_5_4_10_10
void FASTCALL call_5_4_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_10_11
#define init_5_4_10_11
void FASTCALL call_5_4_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_10_12
#define init_5_4_10_12
void FASTCALL call_5_4_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_10_13
#define init_5_4_10_13
void FASTCALL call_5_4_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_10_14
#define init_5_4_10_14
void FASTCALL call_5_4_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_10_15
#define init_5_4_10_15
void FASTCALL call_5_4_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_11_0
#define init_5_4_11_0
void FASTCALL call_5_4_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_11_1
#define init_5_4_11_1
void FASTCALL call_5_4_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_11_2
#define init_5_4_11_2
void FASTCALL call_5_4_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_11_3
#define init_5_4_11_3
void FASTCALL call_5_4_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_11_4
#define init_5_4_11_4
void FASTCALL call_5_4_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_11_5
#define init_5_4_11_5
void FASTCALL call_5_4_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_11_6
#define init_5_4_11_6
void FASTCALL call_5_4_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_11_7
#define init_5_4_11_7
void FASTCALL call_5_4_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_11_8
#define init_5_4_11_8
void FASTCALL call_5_4_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_11_9
#define init_5_4_11_9
void FASTCALL call_5_4_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_11_10
#define init_5_4_11_10
void FASTCALL call_5_4_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_11_11
#define init_5_4_11_11
void FASTCALL call_5_4_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_11_12
#define init_5_4_11_12
void FASTCALL call_5_4_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_11_13
#define init_5_4_11_13
void FASTCALL call_5_4_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_11_14
#define init_5_4_11_14
void FASTCALL call_5_4_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_11_15
#define init_5_4_11_15
void FASTCALL call_5_4_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_12_0
#define init_5_4_12_0
void FASTCALL call_5_4_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_12_1
#define init_5_4_12_1
void FASTCALL call_5_4_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_12_2
#define init_5_4_12_2
void FASTCALL call_5_4_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_12_3
#define init_5_4_12_3
void FASTCALL call_5_4_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_12_4
#define init_5_4_12_4
void FASTCALL call_5_4_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_12_5
#define init_5_4_12_5
void FASTCALL call_5_4_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_12_6
#define init_5_4_12_6
void FASTCALL call_5_4_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_12_7
#define init_5_4_12_7
void FASTCALL call_5_4_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_12_8
#define init_5_4_12_8
void FASTCALL call_5_4_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_12_9
#define init_5_4_12_9
void FASTCALL call_5_4_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_12_10
#define init_5_4_12_10
void FASTCALL call_5_4_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_12_11
#define init_5_4_12_11
void FASTCALL call_5_4_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_12_12
#define init_5_4_12_12
void FASTCALL call_5_4_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_12_13
#define init_5_4_12_13
void FASTCALL call_5_4_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_12_14
#define init_5_4_12_14
void FASTCALL call_5_4_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_12_15
#define init_5_4_12_15
void FASTCALL call_5_4_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_13_0
#define init_5_4_13_0
void FASTCALL call_5_4_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_13_1
#define init_5_4_13_1
void FASTCALL call_5_4_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_13_2
#define init_5_4_13_2
void FASTCALL call_5_4_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_13_3
#define init_5_4_13_3
void FASTCALL call_5_4_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_13_4
#define init_5_4_13_4
void FASTCALL call_5_4_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_13_5
#define init_5_4_13_5
void FASTCALL call_5_4_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_13_6
#define init_5_4_13_6
void FASTCALL call_5_4_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_13_7
#define init_5_4_13_7
void FASTCALL call_5_4_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_13_8
#define init_5_4_13_8
void FASTCALL call_5_4_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_13_9
#define init_5_4_13_9
void FASTCALL call_5_4_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_13_10
#define init_5_4_13_10
void FASTCALL call_5_4_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_13_11
#define init_5_4_13_11
void FASTCALL call_5_4_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_13_12
#define init_5_4_13_12
void FASTCALL call_5_4_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_13_13
#define init_5_4_13_13
void FASTCALL call_5_4_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_13_14
#define init_5_4_13_14
void FASTCALL call_5_4_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_13_15
#define init_5_4_13_15
void FASTCALL call_5_4_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_14_0
#define init_5_4_14_0
void FASTCALL call_5_4_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_14_1
#define init_5_4_14_1
void FASTCALL call_5_4_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_14_2
#define init_5_4_14_2
void FASTCALL call_5_4_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_14_3
#define init_5_4_14_3
void FASTCALL call_5_4_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_14_4
#define init_5_4_14_4
void FASTCALL call_5_4_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_14_5
#define init_5_4_14_5
void FASTCALL call_5_4_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_14_6
#define init_5_4_14_6
void FASTCALL call_5_4_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_14_7
#define init_5_4_14_7
void FASTCALL call_5_4_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_14_8
#define init_5_4_14_8
void FASTCALL call_5_4_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_14_9
#define init_5_4_14_9
void FASTCALL call_5_4_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_14_10
#define init_5_4_14_10
void FASTCALL call_5_4_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_14_11
#define init_5_4_14_11
void FASTCALL call_5_4_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_14_12
#define init_5_4_14_12
void FASTCALL call_5_4_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_14_13
#define init_5_4_14_13
void FASTCALL call_5_4_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_14_14
#define init_5_4_14_14
void FASTCALL call_5_4_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_14_15
#define init_5_4_14_15
void FASTCALL call_5_4_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_15_0
#define init_5_4_15_0
void FASTCALL call_5_4_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_15_1
#define init_5_4_15_1
void FASTCALL call_5_4_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_15_2
#define init_5_4_15_2
void FASTCALL call_5_4_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_15_3
#define init_5_4_15_3
void FASTCALL call_5_4_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_15_4
#define init_5_4_15_4
void FASTCALL call_5_4_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_15_5
#define init_5_4_15_5
void FASTCALL call_5_4_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_15_6
#define init_5_4_15_6
void FASTCALL call_5_4_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_15_7
#define init_5_4_15_7
void FASTCALL call_5_4_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_15_8
#define init_5_4_15_8
void FASTCALL call_5_4_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_15_9
#define init_5_4_15_9
void FASTCALL call_5_4_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_15_10
#define init_5_4_15_10
void FASTCALL call_5_4_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_15_11
#define init_5_4_15_11
void FASTCALL call_5_4_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_15_12
#define init_5_4_15_12
void FASTCALL call_5_4_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_15_13
#define init_5_4_15_13
void FASTCALL call_5_4_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_15_14
#define init_5_4_15_14
void FASTCALL call_5_4_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_4_15_15
#define init_5_4_15_15
void FASTCALL call_5_4_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_0_0
#define init_5_5_0_0
void FASTCALL call_5_5_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_0_1
#define init_5_5_0_1
void FASTCALL call_5_5_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_0_2
#define init_5_5_0_2
void FASTCALL call_5_5_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_0_3
#define init_5_5_0_3
void FASTCALL call_5_5_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_0_4
#define init_5_5_0_4
void FASTCALL call_5_5_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_0_5
#define init_5_5_0_5
void FASTCALL call_5_5_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_0_6
#define init_5_5_0_6
void FASTCALL call_5_5_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_0_7
#define init_5_5_0_7
void FASTCALL call_5_5_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_0_8
#define init_5_5_0_8
void FASTCALL call_5_5_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_0_9
#define init_5_5_0_9
void FASTCALL call_5_5_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_0_10
#define init_5_5_0_10
void FASTCALL call_5_5_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_0_11
#define init_5_5_0_11
void FASTCALL call_5_5_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_0_12
#define init_5_5_0_12
void FASTCALL call_5_5_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_0_13
#define init_5_5_0_13
void FASTCALL call_5_5_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_0_14
#define init_5_5_0_14
void FASTCALL call_5_5_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_0_15
#define init_5_5_0_15
void FASTCALL call_5_5_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_1_0
#define init_5_5_1_0
void FASTCALL call_5_5_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_1_1
#define init_5_5_1_1
void FASTCALL call_5_5_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_1_2
#define init_5_5_1_2
void FASTCALL call_5_5_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_1_3
#define init_5_5_1_3
void FASTCALL call_5_5_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_1_4
#define init_5_5_1_4
void FASTCALL call_5_5_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_1_5
#define init_5_5_1_5
void FASTCALL call_5_5_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_1_6
#define init_5_5_1_6
void FASTCALL call_5_5_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_1_7
#define init_5_5_1_7
void FASTCALL call_5_5_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_1_8
#define init_5_5_1_8
void FASTCALL call_5_5_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_1_9
#define init_5_5_1_9
void FASTCALL call_5_5_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_1_10
#define init_5_5_1_10
void FASTCALL call_5_5_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_1_11
#define init_5_5_1_11
void FASTCALL call_5_5_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_1_12
#define init_5_5_1_12
void FASTCALL call_5_5_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_1_13
#define init_5_5_1_13
void FASTCALL call_5_5_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_1_14
#define init_5_5_1_14
void FASTCALL call_5_5_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_1_15
#define init_5_5_1_15
void FASTCALL call_5_5_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_2_0
#define init_5_5_2_0
void FASTCALL call_5_5_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_2_1
#define init_5_5_2_1
void FASTCALL call_5_5_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_2_2
#define init_5_5_2_2
void FASTCALL call_5_5_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_2_3
#define init_5_5_2_3
void FASTCALL call_5_5_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_2_4
#define init_5_5_2_4
void FASTCALL call_5_5_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_2_5
#define init_5_5_2_5
void FASTCALL call_5_5_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_2_6
#define init_5_5_2_6
void FASTCALL call_5_5_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_2_7
#define init_5_5_2_7
void FASTCALL call_5_5_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_2_8
#define init_5_5_2_8
void FASTCALL call_5_5_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_2_9
#define init_5_5_2_9
void FASTCALL call_5_5_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_2_10
#define init_5_5_2_10
void FASTCALL call_5_5_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_2_11
#define init_5_5_2_11
void FASTCALL call_5_5_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_2_12
#define init_5_5_2_12
void FASTCALL call_5_5_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_2_13
#define init_5_5_2_13
void FASTCALL call_5_5_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_2_14
#define init_5_5_2_14
void FASTCALL call_5_5_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_2_15
#define init_5_5_2_15
void FASTCALL call_5_5_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_3_0
#define init_5_5_3_0
void FASTCALL call_5_5_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_3_1
#define init_5_5_3_1
void FASTCALL call_5_5_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_3_2
#define init_5_5_3_2
void FASTCALL call_5_5_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_3_3
#define init_5_5_3_3
void FASTCALL call_5_5_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_3_4
#define init_5_5_3_4
void FASTCALL call_5_5_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_3_5
#define init_5_5_3_5
void FASTCALL call_5_5_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_3_6
#define init_5_5_3_6
void FASTCALL call_5_5_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_3_7
#define init_5_5_3_7
void FASTCALL call_5_5_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_3_8
#define init_5_5_3_8
void FASTCALL call_5_5_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_3_9
#define init_5_5_3_9
void FASTCALL call_5_5_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_3_10
#define init_5_5_3_10
void FASTCALL call_5_5_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_3_11
#define init_5_5_3_11
void FASTCALL call_5_5_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_3_12
#define init_5_5_3_12
void FASTCALL call_5_5_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_3_13
#define init_5_5_3_13
void FASTCALL call_5_5_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_3_14
#define init_5_5_3_14
void FASTCALL call_5_5_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_3_15
#define init_5_5_3_15
void FASTCALL call_5_5_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_4_0
#define init_5_5_4_0
void FASTCALL call_5_5_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_4_1
#define init_5_5_4_1
void FASTCALL call_5_5_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_4_2
#define init_5_5_4_2
void FASTCALL call_5_5_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_4_3
#define init_5_5_4_3
void FASTCALL call_5_5_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_4_4
#define init_5_5_4_4
void FASTCALL call_5_5_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_4_5
#define init_5_5_4_5
void FASTCALL call_5_5_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_4_6
#define init_5_5_4_6
void FASTCALL call_5_5_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_4_7
#define init_5_5_4_7
void FASTCALL call_5_5_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_4_8
#define init_5_5_4_8
void FASTCALL call_5_5_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_4_9
#define init_5_5_4_9
void FASTCALL call_5_5_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_4_10
#define init_5_5_4_10
void FASTCALL call_5_5_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_4_11
#define init_5_5_4_11
void FASTCALL call_5_5_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_4_12
#define init_5_5_4_12
void FASTCALL call_5_5_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_4_13
#define init_5_5_4_13
void FASTCALL call_5_5_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_4_14
#define init_5_5_4_14
void FASTCALL call_5_5_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_4_15
#define init_5_5_4_15
void FASTCALL call_5_5_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_5_0
#define init_5_5_5_0
void FASTCALL call_5_5_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_5_1
#define init_5_5_5_1
void FASTCALL call_5_5_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_5_2
#define init_5_5_5_2
void FASTCALL call_5_5_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_5_3
#define init_5_5_5_3
void FASTCALL call_5_5_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_5_4
#define init_5_5_5_4
void FASTCALL call_5_5_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_5_5
#define init_5_5_5_5
void FASTCALL call_5_5_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_5_6
#define init_5_5_5_6
void FASTCALL call_5_5_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_5_7
#define init_5_5_5_7
void FASTCALL call_5_5_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_5_8
#define init_5_5_5_8
void FASTCALL call_5_5_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_5_9
#define init_5_5_5_9
void FASTCALL call_5_5_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_5_10
#define init_5_5_5_10
void FASTCALL call_5_5_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_5_11
#define init_5_5_5_11
void FASTCALL call_5_5_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_5_12
#define init_5_5_5_12
void FASTCALL call_5_5_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_5_13
#define init_5_5_5_13
void FASTCALL call_5_5_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_5_14
#define init_5_5_5_14
void FASTCALL call_5_5_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_5_15
#define init_5_5_5_15
void FASTCALL call_5_5_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_6_0
#define init_5_5_6_0
void FASTCALL call_5_5_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_6_1
#define init_5_5_6_1
void FASTCALL call_5_5_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_6_2
#define init_5_5_6_2
void FASTCALL call_5_5_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_6_3
#define init_5_5_6_3
void FASTCALL call_5_5_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_6_4
#define init_5_5_6_4
void FASTCALL call_5_5_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_6_5
#define init_5_5_6_5
void FASTCALL call_5_5_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_6_6
#define init_5_5_6_6
void FASTCALL call_5_5_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_6_7
#define init_5_5_6_7
void FASTCALL call_5_5_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_6_8
#define init_5_5_6_8
void FASTCALL call_5_5_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_6_9
#define init_5_5_6_9
void FASTCALL call_5_5_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_6_10
#define init_5_5_6_10
void FASTCALL call_5_5_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_6_11
#define init_5_5_6_11
void FASTCALL call_5_5_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_6_12
#define init_5_5_6_12
void FASTCALL call_5_5_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_6_13
#define init_5_5_6_13
void FASTCALL call_5_5_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_6_14
#define init_5_5_6_14
void FASTCALL call_5_5_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_6_15
#define init_5_5_6_15
void FASTCALL call_5_5_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_7_0
#define init_5_5_7_0
void FASTCALL call_5_5_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_7_1
#define init_5_5_7_1
void FASTCALL call_5_5_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_7_2
#define init_5_5_7_2
void FASTCALL call_5_5_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_7_3
#define init_5_5_7_3
void FASTCALL call_5_5_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_7_4
#define init_5_5_7_4
void FASTCALL call_5_5_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_7_5
#define init_5_5_7_5
void FASTCALL call_5_5_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_7_6
#define init_5_5_7_6
void FASTCALL call_5_5_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_7_7
#define init_5_5_7_7
void FASTCALL call_5_5_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_7_8
#define init_5_5_7_8
void FASTCALL call_5_5_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_7_9
#define init_5_5_7_9
void FASTCALL call_5_5_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_7_10
#define init_5_5_7_10
void FASTCALL call_5_5_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_7_11
#define init_5_5_7_11
void FASTCALL call_5_5_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_7_12
#define init_5_5_7_12
void FASTCALL call_5_5_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_7_13
#define init_5_5_7_13
void FASTCALL call_5_5_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_7_14
#define init_5_5_7_14
void FASTCALL call_5_5_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_7_15
#define init_5_5_7_15
void FASTCALL call_5_5_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_8_0
#define init_5_5_8_0
void FASTCALL call_5_5_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_8_1
#define init_5_5_8_1
void FASTCALL call_5_5_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_8_2
#define init_5_5_8_2
void FASTCALL call_5_5_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_8_3
#define init_5_5_8_3
void FASTCALL call_5_5_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_8_4
#define init_5_5_8_4
void FASTCALL call_5_5_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_8_5
#define init_5_5_8_5
void FASTCALL call_5_5_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_8_6
#define init_5_5_8_6
void FASTCALL call_5_5_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_8_7
#define init_5_5_8_7
void FASTCALL call_5_5_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_8_8
#define init_5_5_8_8
void FASTCALL call_5_5_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_8_9
#define init_5_5_8_9
void FASTCALL call_5_5_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_8_10
#define init_5_5_8_10
void FASTCALL call_5_5_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_8_11
#define init_5_5_8_11
void FASTCALL call_5_5_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_8_12
#define init_5_5_8_12
void FASTCALL call_5_5_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_8_13
#define init_5_5_8_13
void FASTCALL call_5_5_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_8_14
#define init_5_5_8_14
void FASTCALL call_5_5_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_8_15
#define init_5_5_8_15
void FASTCALL call_5_5_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_9_0
#define init_5_5_9_0
void FASTCALL call_5_5_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_9_1
#define init_5_5_9_1
void FASTCALL call_5_5_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_9_2
#define init_5_5_9_2
void FASTCALL call_5_5_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_9_3
#define init_5_5_9_3
void FASTCALL call_5_5_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_9_4
#define init_5_5_9_4
void FASTCALL call_5_5_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_9_5
#define init_5_5_9_5
void FASTCALL call_5_5_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_9_6
#define init_5_5_9_6
void FASTCALL call_5_5_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_9_7
#define init_5_5_9_7
void FASTCALL call_5_5_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_9_8
#define init_5_5_9_8
void FASTCALL call_5_5_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_9_9
#define init_5_5_9_9
void FASTCALL call_5_5_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_9_10
#define init_5_5_9_10
void FASTCALL call_5_5_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_9_11
#define init_5_5_9_11
void FASTCALL call_5_5_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_9_12
#define init_5_5_9_12
void FASTCALL call_5_5_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_9_13
#define init_5_5_9_13
void FASTCALL call_5_5_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_9_14
#define init_5_5_9_14
void FASTCALL call_5_5_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_9_15
#define init_5_5_9_15
void FASTCALL call_5_5_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_10_0
#define init_5_5_10_0
void FASTCALL call_5_5_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_10_1
#define init_5_5_10_1
void FASTCALL call_5_5_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_10_2
#define init_5_5_10_2
void FASTCALL call_5_5_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_10_3
#define init_5_5_10_3
void FASTCALL call_5_5_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_10_4
#define init_5_5_10_4
void FASTCALL call_5_5_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_10_5
#define init_5_5_10_5
void FASTCALL call_5_5_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_10_6
#define init_5_5_10_6
void FASTCALL call_5_5_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_10_7
#define init_5_5_10_7
void FASTCALL call_5_5_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_10_8
#define init_5_5_10_8
void FASTCALL call_5_5_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_10_9
#define init_5_5_10_9
void FASTCALL call_5_5_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_10_10
#define init_5_5_10_10
void FASTCALL call_5_5_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_10_11
#define init_5_5_10_11
void FASTCALL call_5_5_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_10_12
#define init_5_5_10_12
void FASTCALL call_5_5_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_10_13
#define init_5_5_10_13
void FASTCALL call_5_5_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_10_14
#define init_5_5_10_14
void FASTCALL call_5_5_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_10_15
#define init_5_5_10_15
void FASTCALL call_5_5_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_11_0
#define init_5_5_11_0
void FASTCALL call_5_5_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_11_1
#define init_5_5_11_1
void FASTCALL call_5_5_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_11_2
#define init_5_5_11_2
void FASTCALL call_5_5_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_11_3
#define init_5_5_11_3
void FASTCALL call_5_5_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_11_4
#define init_5_5_11_4
void FASTCALL call_5_5_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_11_5
#define init_5_5_11_5
void FASTCALL call_5_5_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_11_6
#define init_5_5_11_6
void FASTCALL call_5_5_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_11_7
#define init_5_5_11_7
void FASTCALL call_5_5_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_11_8
#define init_5_5_11_8
void FASTCALL call_5_5_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_11_9
#define init_5_5_11_9
void FASTCALL call_5_5_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_11_10
#define init_5_5_11_10
void FASTCALL call_5_5_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_11_11
#define init_5_5_11_11
void FASTCALL call_5_5_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_11_12
#define init_5_5_11_12
void FASTCALL call_5_5_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_11_13
#define init_5_5_11_13
void FASTCALL call_5_5_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_11_14
#define init_5_5_11_14
void FASTCALL call_5_5_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_11_15
#define init_5_5_11_15
void FASTCALL call_5_5_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_12_0
#define init_5_5_12_0
void FASTCALL call_5_5_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_12_1
#define init_5_5_12_1
void FASTCALL call_5_5_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_12_2
#define init_5_5_12_2
void FASTCALL call_5_5_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_12_3
#define init_5_5_12_3
void FASTCALL call_5_5_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_12_4
#define init_5_5_12_4
void FASTCALL call_5_5_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_12_5
#define init_5_5_12_5
void FASTCALL call_5_5_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_12_6
#define init_5_5_12_6
void FASTCALL call_5_5_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_12_7
#define init_5_5_12_7
void FASTCALL call_5_5_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_12_8
#define init_5_5_12_8
void FASTCALL call_5_5_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_12_9
#define init_5_5_12_9
void FASTCALL call_5_5_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_12_10
#define init_5_5_12_10
void FASTCALL call_5_5_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_12_11
#define init_5_5_12_11
void FASTCALL call_5_5_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_12_12
#define init_5_5_12_12
void FASTCALL call_5_5_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_12_13
#define init_5_5_12_13
void FASTCALL call_5_5_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_12_14
#define init_5_5_12_14
void FASTCALL call_5_5_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_12_15
#define init_5_5_12_15
void FASTCALL call_5_5_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_13_0
#define init_5_5_13_0
void FASTCALL call_5_5_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_13_1
#define init_5_5_13_1
void FASTCALL call_5_5_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_13_2
#define init_5_5_13_2
void FASTCALL call_5_5_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_13_3
#define init_5_5_13_3
void FASTCALL call_5_5_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_13_4
#define init_5_5_13_4
void FASTCALL call_5_5_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_13_5
#define init_5_5_13_5
void FASTCALL call_5_5_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_13_6
#define init_5_5_13_6
void FASTCALL call_5_5_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_13_7
#define init_5_5_13_7
void FASTCALL call_5_5_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_13_8
#define init_5_5_13_8
void FASTCALL call_5_5_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_13_9
#define init_5_5_13_9
void FASTCALL call_5_5_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_13_10
#define init_5_5_13_10
void FASTCALL call_5_5_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_13_11
#define init_5_5_13_11
void FASTCALL call_5_5_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_13_12
#define init_5_5_13_12
void FASTCALL call_5_5_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_13_13
#define init_5_5_13_13
void FASTCALL call_5_5_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_13_14
#define init_5_5_13_14
void FASTCALL call_5_5_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_13_15
#define init_5_5_13_15
void FASTCALL call_5_5_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_14_0
#define init_5_5_14_0
void FASTCALL call_5_5_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_14_1
#define init_5_5_14_1
void FASTCALL call_5_5_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_14_2
#define init_5_5_14_2
void FASTCALL call_5_5_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_14_3
#define init_5_5_14_3
void FASTCALL call_5_5_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_14_4
#define init_5_5_14_4
void FASTCALL call_5_5_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_14_5
#define init_5_5_14_5
void FASTCALL call_5_5_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_14_6
#define init_5_5_14_6
void FASTCALL call_5_5_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_14_7
#define init_5_5_14_7
void FASTCALL call_5_5_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_14_8
#define init_5_5_14_8
void FASTCALL call_5_5_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_14_9
#define init_5_5_14_9
void FASTCALL call_5_5_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_14_10
#define init_5_5_14_10
void FASTCALL call_5_5_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_14_11
#define init_5_5_14_11
void FASTCALL call_5_5_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_14_12
#define init_5_5_14_12
void FASTCALL call_5_5_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_14_13
#define init_5_5_14_13
void FASTCALL call_5_5_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_14_14
#define init_5_5_14_14
void FASTCALL call_5_5_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_14_15
#define init_5_5_14_15
void FASTCALL call_5_5_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_15_0
#define init_5_5_15_0
void FASTCALL call_5_5_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_15_1
#define init_5_5_15_1
void FASTCALL call_5_5_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_15_2
#define init_5_5_15_2
void FASTCALL call_5_5_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_15_3
#define init_5_5_15_3
void FASTCALL call_5_5_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_15_4
#define init_5_5_15_4
void FASTCALL call_5_5_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_15_5
#define init_5_5_15_5
void FASTCALL call_5_5_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_15_6
#define init_5_5_15_6
void FASTCALL call_5_5_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_15_7
#define init_5_5_15_7
void FASTCALL call_5_5_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_15_8
#define init_5_5_15_8
void FASTCALL call_5_5_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_15_9
#define init_5_5_15_9
void FASTCALL call_5_5_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_15_10
#define init_5_5_15_10
void FASTCALL call_5_5_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_15_11
#define init_5_5_15_11
void FASTCALL call_5_5_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_15_12
#define init_5_5_15_12
void FASTCALL call_5_5_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_15_13
#define init_5_5_15_13
void FASTCALL call_5_5_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_15_14
#define init_5_5_15_14
void FASTCALL call_5_5_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_5_15_15
#define init_5_5_15_15
void FASTCALL call_5_5_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_0_0
#define init_5_6_0_0
void FASTCALL call_5_6_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_0_1
#define init_5_6_0_1
void FASTCALL call_5_6_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_0_2
#define init_5_6_0_2
void FASTCALL call_5_6_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_0_3
#define init_5_6_0_3
void FASTCALL call_5_6_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_0_4
#define init_5_6_0_4
void FASTCALL call_5_6_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_0_5
#define init_5_6_0_5
void FASTCALL call_5_6_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_0_6
#define init_5_6_0_6
void FASTCALL call_5_6_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_0_7
#define init_5_6_0_7
void FASTCALL call_5_6_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_0_8
#define init_5_6_0_8
void FASTCALL call_5_6_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_0_9
#define init_5_6_0_9
void FASTCALL call_5_6_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_0_10
#define init_5_6_0_10
void FASTCALL call_5_6_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_0_11
#define init_5_6_0_11
void FASTCALL call_5_6_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_0_12
#define init_5_6_0_12
void FASTCALL call_5_6_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_0_13
#define init_5_6_0_13
void FASTCALL call_5_6_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_0_14
#define init_5_6_0_14
void FASTCALL call_5_6_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_0_15
#define init_5_6_0_15
void FASTCALL call_5_6_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_1_0
#define init_5_6_1_0
void FASTCALL call_5_6_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_1_1
#define init_5_6_1_1
void FASTCALL call_5_6_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_1_2
#define init_5_6_1_2
void FASTCALL call_5_6_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_1_3
#define init_5_6_1_3
void FASTCALL call_5_6_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_1_4
#define init_5_6_1_4
void FASTCALL call_5_6_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_1_5
#define init_5_6_1_5
void FASTCALL call_5_6_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_1_6
#define init_5_6_1_6
void FASTCALL call_5_6_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_1_7
#define init_5_6_1_7
void FASTCALL call_5_6_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_1_8
#define init_5_6_1_8
void FASTCALL call_5_6_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_1_9
#define init_5_6_1_9
void FASTCALL call_5_6_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_1_10
#define init_5_6_1_10
void FASTCALL call_5_6_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_1_11
#define init_5_6_1_11
void FASTCALL call_5_6_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_1_12
#define init_5_6_1_12
void FASTCALL call_5_6_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_1_13
#define init_5_6_1_13
void FASTCALL call_5_6_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_1_14
#define init_5_6_1_14
void FASTCALL call_5_6_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_1_15
#define init_5_6_1_15
void FASTCALL call_5_6_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_2_0
#define init_5_6_2_0
void FASTCALL call_5_6_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_2_1
#define init_5_6_2_1
void FASTCALL call_5_6_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_2_2
#define init_5_6_2_2
void FASTCALL call_5_6_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_2_3
#define init_5_6_2_3
void FASTCALL call_5_6_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_2_4
#define init_5_6_2_4
void FASTCALL call_5_6_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_2_5
#define init_5_6_2_5
void FASTCALL call_5_6_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_2_6
#define init_5_6_2_6
void FASTCALL call_5_6_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_2_7
#define init_5_6_2_7
void FASTCALL call_5_6_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_2_8
#define init_5_6_2_8
void FASTCALL call_5_6_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_2_9
#define init_5_6_2_9
void FASTCALL call_5_6_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_2_10
#define init_5_6_2_10
void FASTCALL call_5_6_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_2_11
#define init_5_6_2_11
void FASTCALL call_5_6_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_2_12
#define init_5_6_2_12
void FASTCALL call_5_6_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_2_13
#define init_5_6_2_13
void FASTCALL call_5_6_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_2_14
#define init_5_6_2_14
void FASTCALL call_5_6_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_2_15
#define init_5_6_2_15
void FASTCALL call_5_6_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_3_0
#define init_5_6_3_0
void FASTCALL call_5_6_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_3_1
#define init_5_6_3_1
void FASTCALL call_5_6_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_3_2
#define init_5_6_3_2
void FASTCALL call_5_6_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_3_3
#define init_5_6_3_3
void FASTCALL call_5_6_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_3_4
#define init_5_6_3_4
void FASTCALL call_5_6_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_3_5
#define init_5_6_3_5
void FASTCALL call_5_6_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_3_6
#define init_5_6_3_6
void FASTCALL call_5_6_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_3_7
#define init_5_6_3_7
void FASTCALL call_5_6_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_3_8
#define init_5_6_3_8
void FASTCALL call_5_6_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_3_9
#define init_5_6_3_9
void FASTCALL call_5_6_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_3_10
#define init_5_6_3_10
void FASTCALL call_5_6_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_3_11
#define init_5_6_3_11
void FASTCALL call_5_6_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_3_12
#define init_5_6_3_12
void FASTCALL call_5_6_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_3_13
#define init_5_6_3_13
void FASTCALL call_5_6_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_3_14
#define init_5_6_3_14
void FASTCALL call_5_6_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_3_15
#define init_5_6_3_15
void FASTCALL call_5_6_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_4_0
#define init_5_6_4_0
void FASTCALL call_5_6_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_4_1
#define init_5_6_4_1
void FASTCALL call_5_6_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_4_2
#define init_5_6_4_2
void FASTCALL call_5_6_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_4_3
#define init_5_6_4_3
void FASTCALL call_5_6_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_4_4
#define init_5_6_4_4
void FASTCALL call_5_6_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_4_5
#define init_5_6_4_5
void FASTCALL call_5_6_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_4_6
#define init_5_6_4_6
void FASTCALL call_5_6_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_4_7
#define init_5_6_4_7
void FASTCALL call_5_6_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_4_8
#define init_5_6_4_8
void FASTCALL call_5_6_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_4_9
#define init_5_6_4_9
void FASTCALL call_5_6_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_4_10
#define init_5_6_4_10
void FASTCALL call_5_6_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_4_11
#define init_5_6_4_11
void FASTCALL call_5_6_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_4_12
#define init_5_6_4_12
void FASTCALL call_5_6_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_4_13
#define init_5_6_4_13
void FASTCALL call_5_6_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_4_14
#define init_5_6_4_14
void FASTCALL call_5_6_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_4_15
#define init_5_6_4_15
void FASTCALL call_5_6_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_5_0
#define init_5_6_5_0
void FASTCALL call_5_6_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_5_1
#define init_5_6_5_1
void FASTCALL call_5_6_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_5_2
#define init_5_6_5_2
void FASTCALL call_5_6_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_5_3
#define init_5_6_5_3
void FASTCALL call_5_6_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_5_4
#define init_5_6_5_4
void FASTCALL call_5_6_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_5_5
#define init_5_6_5_5
void FASTCALL call_5_6_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_5_6
#define init_5_6_5_6
void FASTCALL call_5_6_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_5_7
#define init_5_6_5_7
void FASTCALL call_5_6_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_5_8
#define init_5_6_5_8
void FASTCALL call_5_6_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_5_9
#define init_5_6_5_9
void FASTCALL call_5_6_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_5_10
#define init_5_6_5_10
void FASTCALL call_5_6_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_5_11
#define init_5_6_5_11
void FASTCALL call_5_6_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_5_12
#define init_5_6_5_12
void FASTCALL call_5_6_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_5_13
#define init_5_6_5_13
void FASTCALL call_5_6_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_5_14
#define init_5_6_5_14
void FASTCALL call_5_6_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_5_15
#define init_5_6_5_15
void FASTCALL call_5_6_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_6_0
#define init_5_6_6_0
void FASTCALL call_5_6_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_6_1
#define init_5_6_6_1
void FASTCALL call_5_6_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_6_2
#define init_5_6_6_2
void FASTCALL call_5_6_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_6_3
#define init_5_6_6_3
void FASTCALL call_5_6_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_6_4
#define init_5_6_6_4
void FASTCALL call_5_6_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_6_5
#define init_5_6_6_5
void FASTCALL call_5_6_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_6_6
#define init_5_6_6_6
void FASTCALL call_5_6_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_6_7
#define init_5_6_6_7
void FASTCALL call_5_6_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_6_8
#define init_5_6_6_8
void FASTCALL call_5_6_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_6_9
#define init_5_6_6_9
void FASTCALL call_5_6_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_6_10
#define init_5_6_6_10
void FASTCALL call_5_6_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_6_11
#define init_5_6_6_11
void FASTCALL call_5_6_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_6_12
#define init_5_6_6_12
void FASTCALL call_5_6_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_6_13
#define init_5_6_6_13
void FASTCALL call_5_6_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_6_14
#define init_5_6_6_14
void FASTCALL call_5_6_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_6_15
#define init_5_6_6_15
void FASTCALL call_5_6_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_7_0
#define init_5_6_7_0
void FASTCALL call_5_6_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_7_1
#define init_5_6_7_1
void FASTCALL call_5_6_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_7_2
#define init_5_6_7_2
void FASTCALL call_5_6_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_7_3
#define init_5_6_7_3
void FASTCALL call_5_6_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_7_4
#define init_5_6_7_4
void FASTCALL call_5_6_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_7_5
#define init_5_6_7_5
void FASTCALL call_5_6_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_7_6
#define init_5_6_7_6
void FASTCALL call_5_6_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_7_7
#define init_5_6_7_7
void FASTCALL call_5_6_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_7_8
#define init_5_6_7_8
void FASTCALL call_5_6_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_7_9
#define init_5_6_7_9
void FASTCALL call_5_6_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_7_10
#define init_5_6_7_10
void FASTCALL call_5_6_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_7_11
#define init_5_6_7_11
void FASTCALL call_5_6_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_7_12
#define init_5_6_7_12
void FASTCALL call_5_6_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_7_13
#define init_5_6_7_13
void FASTCALL call_5_6_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_7_14
#define init_5_6_7_14
void FASTCALL call_5_6_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_7_15
#define init_5_6_7_15
void FASTCALL call_5_6_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_8_0
#define init_5_6_8_0
void FASTCALL call_5_6_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_8_1
#define init_5_6_8_1
void FASTCALL call_5_6_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_8_2
#define init_5_6_8_2
void FASTCALL call_5_6_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_8_3
#define init_5_6_8_3
void FASTCALL call_5_6_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_8_4
#define init_5_6_8_4
void FASTCALL call_5_6_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_8_5
#define init_5_6_8_5
void FASTCALL call_5_6_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_8_6
#define init_5_6_8_6
void FASTCALL call_5_6_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_8_7
#define init_5_6_8_7
void FASTCALL call_5_6_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_8_8
#define init_5_6_8_8
void FASTCALL call_5_6_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_8_9
#define init_5_6_8_9
void FASTCALL call_5_6_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_8_10
#define init_5_6_8_10
void FASTCALL call_5_6_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_8_11
#define init_5_6_8_11
void FASTCALL call_5_6_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_8_12
#define init_5_6_8_12
void FASTCALL call_5_6_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_8_13
#define init_5_6_8_13
void FASTCALL call_5_6_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_8_14
#define init_5_6_8_14
void FASTCALL call_5_6_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_8_15
#define init_5_6_8_15
void FASTCALL call_5_6_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_9_0
#define init_5_6_9_0
void FASTCALL call_5_6_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_9_1
#define init_5_6_9_1
void FASTCALL call_5_6_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_9_2
#define init_5_6_9_2
void FASTCALL call_5_6_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_9_3
#define init_5_6_9_3
void FASTCALL call_5_6_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_9_4
#define init_5_6_9_4
void FASTCALL call_5_6_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_9_5
#define init_5_6_9_5
void FASTCALL call_5_6_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_9_6
#define init_5_6_9_6
void FASTCALL call_5_6_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_9_7
#define init_5_6_9_7
void FASTCALL call_5_6_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_9_8
#define init_5_6_9_8
void FASTCALL call_5_6_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_9_9
#define init_5_6_9_9
void FASTCALL call_5_6_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_9_10
#define init_5_6_9_10
void FASTCALL call_5_6_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_9_11
#define init_5_6_9_11
void FASTCALL call_5_6_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_9_12
#define init_5_6_9_12
void FASTCALL call_5_6_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_9_13
#define init_5_6_9_13
void FASTCALL call_5_6_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_9_14
#define init_5_6_9_14
void FASTCALL call_5_6_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_9_15
#define init_5_6_9_15
void FASTCALL call_5_6_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_10_0
#define init_5_6_10_0
void FASTCALL call_5_6_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_10_1
#define init_5_6_10_1
void FASTCALL call_5_6_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_10_2
#define init_5_6_10_2
void FASTCALL call_5_6_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_10_3
#define init_5_6_10_3
void FASTCALL call_5_6_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_10_4
#define init_5_6_10_4
void FASTCALL call_5_6_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_10_5
#define init_5_6_10_5
void FASTCALL call_5_6_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_10_6
#define init_5_6_10_6
void FASTCALL call_5_6_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_10_7
#define init_5_6_10_7
void FASTCALL call_5_6_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_10_8
#define init_5_6_10_8
void FASTCALL call_5_6_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_10_9
#define init_5_6_10_9
void FASTCALL call_5_6_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_10_10
#define init_5_6_10_10
void FASTCALL call_5_6_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_10_11
#define init_5_6_10_11
void FASTCALL call_5_6_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_10_12
#define init_5_6_10_12
void FASTCALL call_5_6_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_10_13
#define init_5_6_10_13
void FASTCALL call_5_6_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_10_14
#define init_5_6_10_14
void FASTCALL call_5_6_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_10_15
#define init_5_6_10_15
void FASTCALL call_5_6_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_11_0
#define init_5_6_11_0
void FASTCALL call_5_6_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_11_1
#define init_5_6_11_1
void FASTCALL call_5_6_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_11_2
#define init_5_6_11_2
void FASTCALL call_5_6_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_11_3
#define init_5_6_11_3
void FASTCALL call_5_6_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_11_4
#define init_5_6_11_4
void FASTCALL call_5_6_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_11_5
#define init_5_6_11_5
void FASTCALL call_5_6_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_11_6
#define init_5_6_11_6
void FASTCALL call_5_6_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_11_7
#define init_5_6_11_7
void FASTCALL call_5_6_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_11_8
#define init_5_6_11_8
void FASTCALL call_5_6_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_11_9
#define init_5_6_11_9
void FASTCALL call_5_6_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_11_10
#define init_5_6_11_10
void FASTCALL call_5_6_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_11_11
#define init_5_6_11_11
void FASTCALL call_5_6_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_11_12
#define init_5_6_11_12
void FASTCALL call_5_6_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_11_13
#define init_5_6_11_13
void FASTCALL call_5_6_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_11_14
#define init_5_6_11_14
void FASTCALL call_5_6_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_11_15
#define init_5_6_11_15
void FASTCALL call_5_6_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_12_0
#define init_5_6_12_0
void FASTCALL call_5_6_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_12_1
#define init_5_6_12_1
void FASTCALL call_5_6_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_12_2
#define init_5_6_12_2
void FASTCALL call_5_6_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_12_3
#define init_5_6_12_3
void FASTCALL call_5_6_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_12_4
#define init_5_6_12_4
void FASTCALL call_5_6_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_12_5
#define init_5_6_12_5
void FASTCALL call_5_6_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_12_6
#define init_5_6_12_6
void FASTCALL call_5_6_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_12_7
#define init_5_6_12_7
void FASTCALL call_5_6_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_12_8
#define init_5_6_12_8
void FASTCALL call_5_6_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_12_9
#define init_5_6_12_9
void FASTCALL call_5_6_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_12_10
#define init_5_6_12_10
void FASTCALL call_5_6_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_12_11
#define init_5_6_12_11
void FASTCALL call_5_6_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_12_12
#define init_5_6_12_12
void FASTCALL call_5_6_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_12_13
#define init_5_6_12_13
void FASTCALL call_5_6_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_12_14
#define init_5_6_12_14
void FASTCALL call_5_6_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_12_15
#define init_5_6_12_15
void FASTCALL call_5_6_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_13_0
#define init_5_6_13_0
void FASTCALL call_5_6_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_13_1
#define init_5_6_13_1
void FASTCALL call_5_6_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_13_2
#define init_5_6_13_2
void FASTCALL call_5_6_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_13_3
#define init_5_6_13_3
void FASTCALL call_5_6_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_13_4
#define init_5_6_13_4
void FASTCALL call_5_6_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_13_5
#define init_5_6_13_5
void FASTCALL call_5_6_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_13_6
#define init_5_6_13_6
void FASTCALL call_5_6_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_13_7
#define init_5_6_13_7
void FASTCALL call_5_6_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_13_8
#define init_5_6_13_8
void FASTCALL call_5_6_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_13_9
#define init_5_6_13_9
void FASTCALL call_5_6_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_13_10
#define init_5_6_13_10
void FASTCALL call_5_6_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_13_11
#define init_5_6_13_11
void FASTCALL call_5_6_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_13_12
#define init_5_6_13_12
void FASTCALL call_5_6_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_13_13
#define init_5_6_13_13
void FASTCALL call_5_6_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_13_14
#define init_5_6_13_14
void FASTCALL call_5_6_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_13_15
#define init_5_6_13_15
void FASTCALL call_5_6_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_14_0
#define init_5_6_14_0
void FASTCALL call_5_6_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_14_1
#define init_5_6_14_1
void FASTCALL call_5_6_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_14_2
#define init_5_6_14_2
void FASTCALL call_5_6_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_14_3
#define init_5_6_14_3
void FASTCALL call_5_6_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_14_4
#define init_5_6_14_4
void FASTCALL call_5_6_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_14_5
#define init_5_6_14_5
void FASTCALL call_5_6_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_14_6
#define init_5_6_14_6
void FASTCALL call_5_6_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_14_7
#define init_5_6_14_7
void FASTCALL call_5_6_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_14_8
#define init_5_6_14_8
void FASTCALL call_5_6_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_14_9
#define init_5_6_14_9
void FASTCALL call_5_6_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_14_10
#define init_5_6_14_10
void FASTCALL call_5_6_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_14_11
#define init_5_6_14_11
void FASTCALL call_5_6_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_14_12
#define init_5_6_14_12
void FASTCALL call_5_6_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_14_13
#define init_5_6_14_13
void FASTCALL call_5_6_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_14_14
#define init_5_6_14_14
void FASTCALL call_5_6_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_14_15
#define init_5_6_14_15
void FASTCALL call_5_6_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_15_0
#define init_5_6_15_0
void FASTCALL call_5_6_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_15_1
#define init_5_6_15_1
void FASTCALL call_5_6_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_15_2
#define init_5_6_15_2
void FASTCALL call_5_6_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_15_3
#define init_5_6_15_3
void FASTCALL call_5_6_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_15_4
#define init_5_6_15_4
void FASTCALL call_5_6_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_15_5
#define init_5_6_15_5
void FASTCALL call_5_6_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_15_6
#define init_5_6_15_6
void FASTCALL call_5_6_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_15_7
#define init_5_6_15_7
void FASTCALL call_5_6_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_15_8
#define init_5_6_15_8
void FASTCALL call_5_6_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_15_9
#define init_5_6_15_9
void FASTCALL call_5_6_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_15_10
#define init_5_6_15_10
void FASTCALL call_5_6_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_15_11
#define init_5_6_15_11
void FASTCALL call_5_6_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_15_12
#define init_5_6_15_12
void FASTCALL call_5_6_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_15_13
#define init_5_6_15_13
void FASTCALL call_5_6_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_15_14
#define init_5_6_15_14
void FASTCALL call_5_6_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_6_15_15
#define init_5_6_15_15
void FASTCALL call_5_6_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_0_0
#define init_5_7_0_0
void FASTCALL call_5_7_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_0_1
#define init_5_7_0_1
void FASTCALL call_5_7_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_0_2
#define init_5_7_0_2
void FASTCALL call_5_7_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_0_3
#define init_5_7_0_3
void FASTCALL call_5_7_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_0_4
#define init_5_7_0_4
void FASTCALL call_5_7_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_0_5
#define init_5_7_0_5
void FASTCALL call_5_7_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_0_6
#define init_5_7_0_6
void FASTCALL call_5_7_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_0_7
#define init_5_7_0_7
void FASTCALL call_5_7_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_0_8
#define init_5_7_0_8
void FASTCALL call_5_7_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_0_9
#define init_5_7_0_9
void FASTCALL call_5_7_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_0_10
#define init_5_7_0_10
void FASTCALL call_5_7_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_0_11
#define init_5_7_0_11
void FASTCALL call_5_7_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_0_12
#define init_5_7_0_12
void FASTCALL call_5_7_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_0_13
#define init_5_7_0_13
void FASTCALL call_5_7_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_0_14
#define init_5_7_0_14
void FASTCALL call_5_7_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_0_15
#define init_5_7_0_15
void FASTCALL call_5_7_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_1_0
#define init_5_7_1_0
void FASTCALL call_5_7_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_1_1
#define init_5_7_1_1
void FASTCALL call_5_7_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_1_2
#define init_5_7_1_2
void FASTCALL call_5_7_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_1_3
#define init_5_7_1_3
void FASTCALL call_5_7_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_1_4
#define init_5_7_1_4
void FASTCALL call_5_7_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_1_5
#define init_5_7_1_5
void FASTCALL call_5_7_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_1_6
#define init_5_7_1_6
void FASTCALL call_5_7_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_1_7
#define init_5_7_1_7
void FASTCALL call_5_7_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_1_8
#define init_5_7_1_8
void FASTCALL call_5_7_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_1_9
#define init_5_7_1_9
void FASTCALL call_5_7_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_1_10
#define init_5_7_1_10
void FASTCALL call_5_7_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_1_11
#define init_5_7_1_11
void FASTCALL call_5_7_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_1_12
#define init_5_7_1_12
void FASTCALL call_5_7_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_1_13
#define init_5_7_1_13
void FASTCALL call_5_7_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_1_14
#define init_5_7_1_14
void FASTCALL call_5_7_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_1_15
#define init_5_7_1_15
void FASTCALL call_5_7_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_2_0
#define init_5_7_2_0
void FASTCALL call_5_7_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_2_1
#define init_5_7_2_1
void FASTCALL call_5_7_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_2_2
#define init_5_7_2_2
void FASTCALL call_5_7_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_2_3
#define init_5_7_2_3
void FASTCALL call_5_7_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_2_4
#define init_5_7_2_4
void FASTCALL call_5_7_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_2_5
#define init_5_7_2_5
void FASTCALL call_5_7_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_2_6
#define init_5_7_2_6
void FASTCALL call_5_7_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_2_7
#define init_5_7_2_7
void FASTCALL call_5_7_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_2_8
#define init_5_7_2_8
void FASTCALL call_5_7_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_2_9
#define init_5_7_2_9
void FASTCALL call_5_7_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_2_10
#define init_5_7_2_10
void FASTCALL call_5_7_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_2_11
#define init_5_7_2_11
void FASTCALL call_5_7_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_2_12
#define init_5_7_2_12
void FASTCALL call_5_7_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_2_13
#define init_5_7_2_13
void FASTCALL call_5_7_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_2_14
#define init_5_7_2_14
void FASTCALL call_5_7_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_2_15
#define init_5_7_2_15
void FASTCALL call_5_7_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_3_0
#define init_5_7_3_0
void FASTCALL call_5_7_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_3_1
#define init_5_7_3_1
void FASTCALL call_5_7_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_3_2
#define init_5_7_3_2
void FASTCALL call_5_7_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_3_3
#define init_5_7_3_3
void FASTCALL call_5_7_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_3_4
#define init_5_7_3_4
void FASTCALL call_5_7_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_3_5
#define init_5_7_3_5
void FASTCALL call_5_7_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_3_6
#define init_5_7_3_6
void FASTCALL call_5_7_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_3_7
#define init_5_7_3_7
void FASTCALL call_5_7_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_3_8
#define init_5_7_3_8
void FASTCALL call_5_7_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_3_9
#define init_5_7_3_9
void FASTCALL call_5_7_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_3_10
#define init_5_7_3_10
void FASTCALL call_5_7_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_3_11
#define init_5_7_3_11
void FASTCALL call_5_7_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_3_12
#define init_5_7_3_12
void FASTCALL call_5_7_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_3_13
#define init_5_7_3_13
void FASTCALL call_5_7_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_3_14
#define init_5_7_3_14
void FASTCALL call_5_7_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_3_15
#define init_5_7_3_15
void FASTCALL call_5_7_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_4_0
#define init_5_7_4_0
void FASTCALL call_5_7_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_4_1
#define init_5_7_4_1
void FASTCALL call_5_7_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_4_2
#define init_5_7_4_2
void FASTCALL call_5_7_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_4_3
#define init_5_7_4_3
void FASTCALL call_5_7_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_4_4
#define init_5_7_4_4
void FASTCALL call_5_7_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_4_5
#define init_5_7_4_5
void FASTCALL call_5_7_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_4_6
#define init_5_7_4_6
void FASTCALL call_5_7_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_4_7
#define init_5_7_4_7
void FASTCALL call_5_7_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_4_8
#define init_5_7_4_8
void FASTCALL call_5_7_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_4_9
#define init_5_7_4_9
void FASTCALL call_5_7_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_4_10
#define init_5_7_4_10
void FASTCALL call_5_7_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_4_11
#define init_5_7_4_11
void FASTCALL call_5_7_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_4_12
#define init_5_7_4_12
void FASTCALL call_5_7_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_4_13
#define init_5_7_4_13
void FASTCALL call_5_7_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_4_14
#define init_5_7_4_14
void FASTCALL call_5_7_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_4_15
#define init_5_7_4_15
void FASTCALL call_5_7_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_5_0
#define init_5_7_5_0
void FASTCALL call_5_7_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_5_1
#define init_5_7_5_1
void FASTCALL call_5_7_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_5_2
#define init_5_7_5_2
void FASTCALL call_5_7_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_5_3
#define init_5_7_5_3
void FASTCALL call_5_7_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_5_4
#define init_5_7_5_4
void FASTCALL call_5_7_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_5_5
#define init_5_7_5_5
void FASTCALL call_5_7_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_5_6
#define init_5_7_5_6
void FASTCALL call_5_7_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_5_7
#define init_5_7_5_7
void FASTCALL call_5_7_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_5_8
#define init_5_7_5_8
void FASTCALL call_5_7_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_5_9
#define init_5_7_5_9
void FASTCALL call_5_7_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_5_10
#define init_5_7_5_10
void FASTCALL call_5_7_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_5_11
#define init_5_7_5_11
void FASTCALL call_5_7_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_5_12
#define init_5_7_5_12
void FASTCALL call_5_7_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_5_13
#define init_5_7_5_13
void FASTCALL call_5_7_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_5_14
#define init_5_7_5_14
void FASTCALL call_5_7_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_5_15
#define init_5_7_5_15
void FASTCALL call_5_7_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_6_0
#define init_5_7_6_0
void FASTCALL call_5_7_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_6_1
#define init_5_7_6_1
void FASTCALL call_5_7_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_6_2
#define init_5_7_6_2
void FASTCALL call_5_7_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_6_3
#define init_5_7_6_3
void FASTCALL call_5_7_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_6_4
#define init_5_7_6_4
void FASTCALL call_5_7_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_6_5
#define init_5_7_6_5
void FASTCALL call_5_7_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_6_6
#define init_5_7_6_6
void FASTCALL call_5_7_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_6_7
#define init_5_7_6_7
void FASTCALL call_5_7_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_6_8
#define init_5_7_6_8
void FASTCALL call_5_7_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_6_9
#define init_5_7_6_9
void FASTCALL call_5_7_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_6_10
#define init_5_7_6_10
void FASTCALL call_5_7_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_6_11
#define init_5_7_6_11
void FASTCALL call_5_7_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_6_12
#define init_5_7_6_12
void FASTCALL call_5_7_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_6_13
#define init_5_7_6_13
void FASTCALL call_5_7_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_6_14
#define init_5_7_6_14
void FASTCALL call_5_7_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_6_15
#define init_5_7_6_15
void FASTCALL call_5_7_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_7_0
#define init_5_7_7_0
void FASTCALL call_5_7_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_7_1
#define init_5_7_7_1
void FASTCALL call_5_7_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_7_2
#define init_5_7_7_2
void FASTCALL call_5_7_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_7_3
#define init_5_7_7_3
void FASTCALL call_5_7_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_7_4
#define init_5_7_7_4
void FASTCALL call_5_7_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_7_5
#define init_5_7_7_5
void FASTCALL call_5_7_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_7_6
#define init_5_7_7_6
void FASTCALL call_5_7_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_7_7
#define init_5_7_7_7
void FASTCALL call_5_7_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_7_8
#define init_5_7_7_8
void FASTCALL call_5_7_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_7_9
#define init_5_7_7_9
void FASTCALL call_5_7_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_7_10
#define init_5_7_7_10
void FASTCALL call_5_7_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_7_11
#define init_5_7_7_11
void FASTCALL call_5_7_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_7_12
#define init_5_7_7_12
void FASTCALL call_5_7_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_7_13
#define init_5_7_7_13
void FASTCALL call_5_7_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_7_14
#define init_5_7_7_14
void FASTCALL call_5_7_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_7_15
#define init_5_7_7_15
void FASTCALL call_5_7_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_8_0
#define init_5_7_8_0
void FASTCALL call_5_7_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_8_1
#define init_5_7_8_1
void FASTCALL call_5_7_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_8_2
#define init_5_7_8_2
void FASTCALL call_5_7_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_8_3
#define init_5_7_8_3
void FASTCALL call_5_7_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_8_4
#define init_5_7_8_4
void FASTCALL call_5_7_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_8_5
#define init_5_7_8_5
void FASTCALL call_5_7_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_8_6
#define init_5_7_8_6
void FASTCALL call_5_7_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_8_7
#define init_5_7_8_7
void FASTCALL call_5_7_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_8_8
#define init_5_7_8_8
void FASTCALL call_5_7_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_8_9
#define init_5_7_8_9
void FASTCALL call_5_7_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_8_10
#define init_5_7_8_10
void FASTCALL call_5_7_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_8_11
#define init_5_7_8_11
void FASTCALL call_5_7_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_8_12
#define init_5_7_8_12
void FASTCALL call_5_7_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_8_13
#define init_5_7_8_13
void FASTCALL call_5_7_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_8_14
#define init_5_7_8_14
void FASTCALL call_5_7_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_8_15
#define init_5_7_8_15
void FASTCALL call_5_7_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_9_0
#define init_5_7_9_0
void FASTCALL call_5_7_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_9_1
#define init_5_7_9_1
void FASTCALL call_5_7_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_9_2
#define init_5_7_9_2
void FASTCALL call_5_7_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_9_3
#define init_5_7_9_3
void FASTCALL call_5_7_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_9_4
#define init_5_7_9_4
void FASTCALL call_5_7_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_9_5
#define init_5_7_9_5
void FASTCALL call_5_7_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_9_6
#define init_5_7_9_6
void FASTCALL call_5_7_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_9_7
#define init_5_7_9_7
void FASTCALL call_5_7_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_9_8
#define init_5_7_9_8
void FASTCALL call_5_7_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_9_9
#define init_5_7_9_9
void FASTCALL call_5_7_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_9_10
#define init_5_7_9_10
void FASTCALL call_5_7_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_9_11
#define init_5_7_9_11
void FASTCALL call_5_7_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_9_12
#define init_5_7_9_12
void FASTCALL call_5_7_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_9_13
#define init_5_7_9_13
void FASTCALL call_5_7_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_9_14
#define init_5_7_9_14
void FASTCALL call_5_7_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_9_15
#define init_5_7_9_15
void FASTCALL call_5_7_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_10_0
#define init_5_7_10_0
void FASTCALL call_5_7_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_10_1
#define init_5_7_10_1
void FASTCALL call_5_7_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_10_2
#define init_5_7_10_2
void FASTCALL call_5_7_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_10_3
#define init_5_7_10_3
void FASTCALL call_5_7_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_10_4
#define init_5_7_10_4
void FASTCALL call_5_7_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_10_5
#define init_5_7_10_5
void FASTCALL call_5_7_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_10_6
#define init_5_7_10_6
void FASTCALL call_5_7_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_10_7
#define init_5_7_10_7
void FASTCALL call_5_7_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_10_8
#define init_5_7_10_8
void FASTCALL call_5_7_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_10_9
#define init_5_7_10_9
void FASTCALL call_5_7_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_10_10
#define init_5_7_10_10
void FASTCALL call_5_7_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_10_11
#define init_5_7_10_11
void FASTCALL call_5_7_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_10_12
#define init_5_7_10_12
void FASTCALL call_5_7_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_10_13
#define init_5_7_10_13
void FASTCALL call_5_7_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_10_14
#define init_5_7_10_14
void FASTCALL call_5_7_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_10_15
#define init_5_7_10_15
void FASTCALL call_5_7_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_11_0
#define init_5_7_11_0
void FASTCALL call_5_7_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_11_1
#define init_5_7_11_1
void FASTCALL call_5_7_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_11_2
#define init_5_7_11_2
void FASTCALL call_5_7_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_11_3
#define init_5_7_11_3
void FASTCALL call_5_7_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_11_4
#define init_5_7_11_4
void FASTCALL call_5_7_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_11_5
#define init_5_7_11_5
void FASTCALL call_5_7_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_11_6
#define init_5_7_11_6
void FASTCALL call_5_7_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_11_7
#define init_5_7_11_7
void FASTCALL call_5_7_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_11_8
#define init_5_7_11_8
void FASTCALL call_5_7_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_11_9
#define init_5_7_11_9
void FASTCALL call_5_7_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_11_10
#define init_5_7_11_10
void FASTCALL call_5_7_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_11_11
#define init_5_7_11_11
void FASTCALL call_5_7_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_11_12
#define init_5_7_11_12
void FASTCALL call_5_7_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_11_13
#define init_5_7_11_13
void FASTCALL call_5_7_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_11_14
#define init_5_7_11_14
void FASTCALL call_5_7_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_11_15
#define init_5_7_11_15
void FASTCALL call_5_7_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_12_0
#define init_5_7_12_0
void FASTCALL call_5_7_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_12_1
#define init_5_7_12_1
void FASTCALL call_5_7_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_12_2
#define init_5_7_12_2
void FASTCALL call_5_7_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_12_3
#define init_5_7_12_3
void FASTCALL call_5_7_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_12_4
#define init_5_7_12_4
void FASTCALL call_5_7_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_12_5
#define init_5_7_12_5
void FASTCALL call_5_7_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_12_6
#define init_5_7_12_6
void FASTCALL call_5_7_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_12_7
#define init_5_7_12_7
void FASTCALL call_5_7_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_12_8
#define init_5_7_12_8
void FASTCALL call_5_7_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_12_9
#define init_5_7_12_9
void FASTCALL call_5_7_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_12_10
#define init_5_7_12_10
void FASTCALL call_5_7_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_12_11
#define init_5_7_12_11
void FASTCALL call_5_7_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_12_12
#define init_5_7_12_12
void FASTCALL call_5_7_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_12_13
#define init_5_7_12_13
void FASTCALL call_5_7_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_12_14
#define init_5_7_12_14
void FASTCALL call_5_7_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_12_15
#define init_5_7_12_15
void FASTCALL call_5_7_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_13_0
#define init_5_7_13_0
void FASTCALL call_5_7_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_13_1
#define init_5_7_13_1
void FASTCALL call_5_7_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_13_2
#define init_5_7_13_2
void FASTCALL call_5_7_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_13_3
#define init_5_7_13_3
void FASTCALL call_5_7_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_13_4
#define init_5_7_13_4
void FASTCALL call_5_7_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_13_5
#define init_5_7_13_5
void FASTCALL call_5_7_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_13_6
#define init_5_7_13_6
void FASTCALL call_5_7_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_13_7
#define init_5_7_13_7
void FASTCALL call_5_7_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_13_8
#define init_5_7_13_8
void FASTCALL call_5_7_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_13_9
#define init_5_7_13_9
void FASTCALL call_5_7_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_13_10
#define init_5_7_13_10
void FASTCALL call_5_7_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_13_11
#define init_5_7_13_11
void FASTCALL call_5_7_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_13_12
#define init_5_7_13_12
void FASTCALL call_5_7_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_13_13
#define init_5_7_13_13
void FASTCALL call_5_7_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_13_14
#define init_5_7_13_14
void FASTCALL call_5_7_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_13_15
#define init_5_7_13_15
void FASTCALL call_5_7_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_14_0
#define init_5_7_14_0
void FASTCALL call_5_7_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_14_1
#define init_5_7_14_1
void FASTCALL call_5_7_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_14_2
#define init_5_7_14_2
void FASTCALL call_5_7_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_14_3
#define init_5_7_14_3
void FASTCALL call_5_7_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_14_4
#define init_5_7_14_4
void FASTCALL call_5_7_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_14_5
#define init_5_7_14_5
void FASTCALL call_5_7_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_14_6
#define init_5_7_14_6
void FASTCALL call_5_7_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_14_7
#define init_5_7_14_7
void FASTCALL call_5_7_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_14_8
#define init_5_7_14_8
void FASTCALL call_5_7_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_14_9
#define init_5_7_14_9
void FASTCALL call_5_7_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_14_10
#define init_5_7_14_10
void FASTCALL call_5_7_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_14_11
#define init_5_7_14_11
void FASTCALL call_5_7_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_14_12
#define init_5_7_14_12
void FASTCALL call_5_7_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_14_13
#define init_5_7_14_13
void FASTCALL call_5_7_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_14_14
#define init_5_7_14_14
void FASTCALL call_5_7_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_14_15
#define init_5_7_14_15
void FASTCALL call_5_7_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_15_0
#define init_5_7_15_0
void FASTCALL call_5_7_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_15_1
#define init_5_7_15_1
void FASTCALL call_5_7_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_15_2
#define init_5_7_15_2
void FASTCALL call_5_7_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_15_3
#define init_5_7_15_3
void FASTCALL call_5_7_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_15_4
#define init_5_7_15_4
void FASTCALL call_5_7_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_15_5
#define init_5_7_15_5
void FASTCALL call_5_7_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_15_6
#define init_5_7_15_6
void FASTCALL call_5_7_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_15_7
#define init_5_7_15_7
void FASTCALL call_5_7_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_15_8
#define init_5_7_15_8
void FASTCALL call_5_7_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_15_9
#define init_5_7_15_9
void FASTCALL call_5_7_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_15_10
#define init_5_7_15_10
void FASTCALL call_5_7_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_15_11
#define init_5_7_15_11
void FASTCALL call_5_7_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_15_12
#define init_5_7_15_12
void FASTCALL call_5_7_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_15_13
#define init_5_7_15_13
void FASTCALL call_5_7_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_15_14
#define init_5_7_15_14
void FASTCALL call_5_7_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_7_15_15
#define init_5_7_15_15
void FASTCALL call_5_7_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_0_0
#define init_5_8_0_0
void FASTCALL call_5_8_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_0_1
#define init_5_8_0_1
void FASTCALL call_5_8_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_0_2
#define init_5_8_0_2
void FASTCALL call_5_8_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_0_3
#define init_5_8_0_3
void FASTCALL call_5_8_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_0_4
#define init_5_8_0_4
void FASTCALL call_5_8_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_0_5
#define init_5_8_0_5
void FASTCALL call_5_8_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_0_6
#define init_5_8_0_6
void FASTCALL call_5_8_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_0_7
#define init_5_8_0_7
void FASTCALL call_5_8_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_0_8
#define init_5_8_0_8
void FASTCALL call_5_8_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_0_9
#define init_5_8_0_9
void FASTCALL call_5_8_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_0_10
#define init_5_8_0_10
void FASTCALL call_5_8_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_0_11
#define init_5_8_0_11
void FASTCALL call_5_8_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_0_12
#define init_5_8_0_12
void FASTCALL call_5_8_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_0_13
#define init_5_8_0_13
void FASTCALL call_5_8_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_0_14
#define init_5_8_0_14
void FASTCALL call_5_8_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_0_15
#define init_5_8_0_15
void FASTCALL call_5_8_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_1_0
#define init_5_8_1_0
void FASTCALL call_5_8_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_1_1
#define init_5_8_1_1
void FASTCALL call_5_8_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_1_2
#define init_5_8_1_2
void FASTCALL call_5_8_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_1_3
#define init_5_8_1_3
void FASTCALL call_5_8_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_1_4
#define init_5_8_1_4
void FASTCALL call_5_8_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_1_5
#define init_5_8_1_5
void FASTCALL call_5_8_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_1_6
#define init_5_8_1_6
void FASTCALL call_5_8_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_1_7
#define init_5_8_1_7
void FASTCALL call_5_8_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_1_8
#define init_5_8_1_8
void FASTCALL call_5_8_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_1_9
#define init_5_8_1_9
void FASTCALL call_5_8_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_1_10
#define init_5_8_1_10
void FASTCALL call_5_8_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_1_11
#define init_5_8_1_11
void FASTCALL call_5_8_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_1_12
#define init_5_8_1_12
void FASTCALL call_5_8_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_1_13
#define init_5_8_1_13
void FASTCALL call_5_8_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_1_14
#define init_5_8_1_14
void FASTCALL call_5_8_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_1_15
#define init_5_8_1_15
void FASTCALL call_5_8_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_2_0
#define init_5_8_2_0
void FASTCALL call_5_8_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_2_1
#define init_5_8_2_1
void FASTCALL call_5_8_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_2_2
#define init_5_8_2_2
void FASTCALL call_5_8_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_2_3
#define init_5_8_2_3
void FASTCALL call_5_8_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_2_4
#define init_5_8_2_4
void FASTCALL call_5_8_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_2_5
#define init_5_8_2_5
void FASTCALL call_5_8_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_2_6
#define init_5_8_2_6
void FASTCALL call_5_8_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_2_7
#define init_5_8_2_7
void FASTCALL call_5_8_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_2_8
#define init_5_8_2_8
void FASTCALL call_5_8_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_2_9
#define init_5_8_2_9
void FASTCALL call_5_8_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_2_10
#define init_5_8_2_10
void FASTCALL call_5_8_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_2_11
#define init_5_8_2_11
void FASTCALL call_5_8_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_2_12
#define init_5_8_2_12
void FASTCALL call_5_8_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_2_13
#define init_5_8_2_13
void FASTCALL call_5_8_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_2_14
#define init_5_8_2_14
void FASTCALL call_5_8_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_2_15
#define init_5_8_2_15
void FASTCALL call_5_8_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_3_0
#define init_5_8_3_0
void FASTCALL call_5_8_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_3_1
#define init_5_8_3_1
void FASTCALL call_5_8_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_3_2
#define init_5_8_3_2
void FASTCALL call_5_8_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_3_3
#define init_5_8_3_3
void FASTCALL call_5_8_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_3_4
#define init_5_8_3_4
void FASTCALL call_5_8_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_3_5
#define init_5_8_3_5
void FASTCALL call_5_8_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_3_6
#define init_5_8_3_6
void FASTCALL call_5_8_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_3_7
#define init_5_8_3_7
void FASTCALL call_5_8_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_3_8
#define init_5_8_3_8
void FASTCALL call_5_8_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_3_9
#define init_5_8_3_9
void FASTCALL call_5_8_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_3_10
#define init_5_8_3_10
void FASTCALL call_5_8_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_3_11
#define init_5_8_3_11
void FASTCALL call_5_8_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_3_12
#define init_5_8_3_12
void FASTCALL call_5_8_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_3_13
#define init_5_8_3_13
void FASTCALL call_5_8_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_3_14
#define init_5_8_3_14
void FASTCALL call_5_8_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_3_15
#define init_5_8_3_15
void FASTCALL call_5_8_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_4_0
#define init_5_8_4_0
void FASTCALL call_5_8_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_4_1
#define init_5_8_4_1
void FASTCALL call_5_8_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_4_2
#define init_5_8_4_2
void FASTCALL call_5_8_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_4_3
#define init_5_8_4_3
void FASTCALL call_5_8_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_4_4
#define init_5_8_4_4
void FASTCALL call_5_8_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_4_5
#define init_5_8_4_5
void FASTCALL call_5_8_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_4_6
#define init_5_8_4_6
void FASTCALL call_5_8_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_4_7
#define init_5_8_4_7
void FASTCALL call_5_8_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_4_8
#define init_5_8_4_8
void FASTCALL call_5_8_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_4_9
#define init_5_8_4_9
void FASTCALL call_5_8_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_4_10
#define init_5_8_4_10
void FASTCALL call_5_8_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_4_11
#define init_5_8_4_11
void FASTCALL call_5_8_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_4_12
#define init_5_8_4_12
void FASTCALL call_5_8_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_4_13
#define init_5_8_4_13
void FASTCALL call_5_8_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_4_14
#define init_5_8_4_14
void FASTCALL call_5_8_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_4_15
#define init_5_8_4_15
void FASTCALL call_5_8_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_5_0
#define init_5_8_5_0
void FASTCALL call_5_8_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_5_1
#define init_5_8_5_1
void FASTCALL call_5_8_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_5_2
#define init_5_8_5_2
void FASTCALL call_5_8_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_5_3
#define init_5_8_5_3
void FASTCALL call_5_8_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_5_4
#define init_5_8_5_4
void FASTCALL call_5_8_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_5_5
#define init_5_8_5_5
void FASTCALL call_5_8_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_5_6
#define init_5_8_5_6
void FASTCALL call_5_8_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_5_7
#define init_5_8_5_7
void FASTCALL call_5_8_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_5_8
#define init_5_8_5_8
void FASTCALL call_5_8_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_5_9
#define init_5_8_5_9
void FASTCALL call_5_8_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_5_10
#define init_5_8_5_10
void FASTCALL call_5_8_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_5_11
#define init_5_8_5_11
void FASTCALL call_5_8_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_5_12
#define init_5_8_5_12
void FASTCALL call_5_8_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_5_13
#define init_5_8_5_13
void FASTCALL call_5_8_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_5_14
#define init_5_8_5_14
void FASTCALL call_5_8_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_5_15
#define init_5_8_5_15
void FASTCALL call_5_8_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_6_0
#define init_5_8_6_0
void FASTCALL call_5_8_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_6_1
#define init_5_8_6_1
void FASTCALL call_5_8_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_6_2
#define init_5_8_6_2
void FASTCALL call_5_8_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_6_3
#define init_5_8_6_3
void FASTCALL call_5_8_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_6_4
#define init_5_8_6_4
void FASTCALL call_5_8_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_6_5
#define init_5_8_6_5
void FASTCALL call_5_8_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_6_6
#define init_5_8_6_6
void FASTCALL call_5_8_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_6_7
#define init_5_8_6_7
void FASTCALL call_5_8_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_6_8
#define init_5_8_6_8
void FASTCALL call_5_8_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_6_9
#define init_5_8_6_9
void FASTCALL call_5_8_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_6_10
#define init_5_8_6_10
void FASTCALL call_5_8_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_6_11
#define init_5_8_6_11
void FASTCALL call_5_8_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_6_12
#define init_5_8_6_12
void FASTCALL call_5_8_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_6_13
#define init_5_8_6_13
void FASTCALL call_5_8_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_6_14
#define init_5_8_6_14
void FASTCALL call_5_8_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_6_15
#define init_5_8_6_15
void FASTCALL call_5_8_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_7_0
#define init_5_8_7_0
void FASTCALL call_5_8_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_7_1
#define init_5_8_7_1
void FASTCALL call_5_8_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_7_2
#define init_5_8_7_2
void FASTCALL call_5_8_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_7_3
#define init_5_8_7_3
void FASTCALL call_5_8_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_7_4
#define init_5_8_7_4
void FASTCALL call_5_8_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_7_5
#define init_5_8_7_5
void FASTCALL call_5_8_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_7_6
#define init_5_8_7_6
void FASTCALL call_5_8_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_7_7
#define init_5_8_7_7
void FASTCALL call_5_8_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_7_8
#define init_5_8_7_8
void FASTCALL call_5_8_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_7_9
#define init_5_8_7_9
void FASTCALL call_5_8_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_7_10
#define init_5_8_7_10
void FASTCALL call_5_8_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_7_11
#define init_5_8_7_11
void FASTCALL call_5_8_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_7_12
#define init_5_8_7_12
void FASTCALL call_5_8_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_7_13
#define init_5_8_7_13
void FASTCALL call_5_8_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_7_14
#define init_5_8_7_14
void FASTCALL call_5_8_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_7_15
#define init_5_8_7_15
void FASTCALL call_5_8_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_8_0
#define init_5_8_8_0
void FASTCALL call_5_8_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_8_1
#define init_5_8_8_1
void FASTCALL call_5_8_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_8_2
#define init_5_8_8_2
void FASTCALL call_5_8_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_8_3
#define init_5_8_8_3
void FASTCALL call_5_8_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_8_4
#define init_5_8_8_4
void FASTCALL call_5_8_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_8_5
#define init_5_8_8_5
void FASTCALL call_5_8_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_8_6
#define init_5_8_8_6
void FASTCALL call_5_8_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_8_7
#define init_5_8_8_7
void FASTCALL call_5_8_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_8_8
#define init_5_8_8_8
void FASTCALL call_5_8_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_8_9
#define init_5_8_8_9
void FASTCALL call_5_8_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_8_10
#define init_5_8_8_10
void FASTCALL call_5_8_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_8_11
#define init_5_8_8_11
void FASTCALL call_5_8_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_8_12
#define init_5_8_8_12
void FASTCALL call_5_8_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_8_13
#define init_5_8_8_13
void FASTCALL call_5_8_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_8_14
#define init_5_8_8_14
void FASTCALL call_5_8_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_8_15
#define init_5_8_8_15
void FASTCALL call_5_8_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_9_0
#define init_5_8_9_0
void FASTCALL call_5_8_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_9_1
#define init_5_8_9_1
void FASTCALL call_5_8_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_9_2
#define init_5_8_9_2
void FASTCALL call_5_8_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_9_3
#define init_5_8_9_3
void FASTCALL call_5_8_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_9_4
#define init_5_8_9_4
void FASTCALL call_5_8_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_9_5
#define init_5_8_9_5
void FASTCALL call_5_8_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_9_6
#define init_5_8_9_6
void FASTCALL call_5_8_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_9_7
#define init_5_8_9_7
void FASTCALL call_5_8_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_9_8
#define init_5_8_9_8
void FASTCALL call_5_8_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_9_9
#define init_5_8_9_9
void FASTCALL call_5_8_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_9_10
#define init_5_8_9_10
void FASTCALL call_5_8_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_9_11
#define init_5_8_9_11
void FASTCALL call_5_8_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_9_12
#define init_5_8_9_12
void FASTCALL call_5_8_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_9_13
#define init_5_8_9_13
void FASTCALL call_5_8_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_9_14
#define init_5_8_9_14
void FASTCALL call_5_8_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_9_15
#define init_5_8_9_15
void FASTCALL call_5_8_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_10_0
#define init_5_8_10_0
void FASTCALL call_5_8_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_10_1
#define init_5_8_10_1
void FASTCALL call_5_8_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_10_2
#define init_5_8_10_2
void FASTCALL call_5_8_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_10_3
#define init_5_8_10_3
void FASTCALL call_5_8_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_10_4
#define init_5_8_10_4
void FASTCALL call_5_8_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_10_5
#define init_5_8_10_5
void FASTCALL call_5_8_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_10_6
#define init_5_8_10_6
void FASTCALL call_5_8_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_10_7
#define init_5_8_10_7
void FASTCALL call_5_8_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_10_8
#define init_5_8_10_8
void FASTCALL call_5_8_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_10_9
#define init_5_8_10_9
void FASTCALL call_5_8_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_10_10
#define init_5_8_10_10
void FASTCALL call_5_8_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_10_11
#define init_5_8_10_11
void FASTCALL call_5_8_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_10_12
#define init_5_8_10_12
void FASTCALL call_5_8_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_10_13
#define init_5_8_10_13
void FASTCALL call_5_8_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_10_14
#define init_5_8_10_14
void FASTCALL call_5_8_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_10_15
#define init_5_8_10_15
void FASTCALL call_5_8_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_11_0
#define init_5_8_11_0
void FASTCALL call_5_8_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_11_1
#define init_5_8_11_1
void FASTCALL call_5_8_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_11_2
#define init_5_8_11_2
void FASTCALL call_5_8_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_11_3
#define init_5_8_11_3
void FASTCALL call_5_8_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_11_4
#define init_5_8_11_4
void FASTCALL call_5_8_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_11_5
#define init_5_8_11_5
void FASTCALL call_5_8_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_11_6
#define init_5_8_11_6
void FASTCALL call_5_8_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_11_7
#define init_5_8_11_7
void FASTCALL call_5_8_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_11_8
#define init_5_8_11_8
void FASTCALL call_5_8_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_11_9
#define init_5_8_11_9
void FASTCALL call_5_8_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_11_10
#define init_5_8_11_10
void FASTCALL call_5_8_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_11_11
#define init_5_8_11_11
void FASTCALL call_5_8_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_11_12
#define init_5_8_11_12
void FASTCALL call_5_8_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_11_13
#define init_5_8_11_13
void FASTCALL call_5_8_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_11_14
#define init_5_8_11_14
void FASTCALL call_5_8_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_11_15
#define init_5_8_11_15
void FASTCALL call_5_8_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_12_0
#define init_5_8_12_0
void FASTCALL call_5_8_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_12_1
#define init_5_8_12_1
void FASTCALL call_5_8_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_12_2
#define init_5_8_12_2
void FASTCALL call_5_8_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_12_3
#define init_5_8_12_3
void FASTCALL call_5_8_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_12_4
#define init_5_8_12_4
void FASTCALL call_5_8_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_12_5
#define init_5_8_12_5
void FASTCALL call_5_8_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_12_6
#define init_5_8_12_6
void FASTCALL call_5_8_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_12_7
#define init_5_8_12_7
void FASTCALL call_5_8_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_12_8
#define init_5_8_12_8
void FASTCALL call_5_8_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_12_9
#define init_5_8_12_9
void FASTCALL call_5_8_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_12_10
#define init_5_8_12_10
void FASTCALL call_5_8_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_12_11
#define init_5_8_12_11
void FASTCALL call_5_8_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_12_12
#define init_5_8_12_12
void FASTCALL call_5_8_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_12_13
#define init_5_8_12_13
void FASTCALL call_5_8_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_12_14
#define init_5_8_12_14
void FASTCALL call_5_8_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_12_15
#define init_5_8_12_15
void FASTCALL call_5_8_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_13_0
#define init_5_8_13_0
void FASTCALL call_5_8_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_13_1
#define init_5_8_13_1
void FASTCALL call_5_8_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_13_2
#define init_5_8_13_2
void FASTCALL call_5_8_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_13_3
#define init_5_8_13_3
void FASTCALL call_5_8_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_13_4
#define init_5_8_13_4
void FASTCALL call_5_8_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_13_5
#define init_5_8_13_5
void FASTCALL call_5_8_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_13_6
#define init_5_8_13_6
void FASTCALL call_5_8_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_13_7
#define init_5_8_13_7
void FASTCALL call_5_8_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_13_8
#define init_5_8_13_8
void FASTCALL call_5_8_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_13_9
#define init_5_8_13_9
void FASTCALL call_5_8_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_13_10
#define init_5_8_13_10
void FASTCALL call_5_8_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_13_11
#define init_5_8_13_11
void FASTCALL call_5_8_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_13_12
#define init_5_8_13_12
void FASTCALL call_5_8_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_13_13
#define init_5_8_13_13
void FASTCALL call_5_8_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_13_14
#define init_5_8_13_14
void FASTCALL call_5_8_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_13_15
#define init_5_8_13_15
void FASTCALL call_5_8_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_14_0
#define init_5_8_14_0
void FASTCALL call_5_8_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_14_1
#define init_5_8_14_1
void FASTCALL call_5_8_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_14_2
#define init_5_8_14_2
void FASTCALL call_5_8_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_14_3
#define init_5_8_14_3
void FASTCALL call_5_8_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_14_4
#define init_5_8_14_4
void FASTCALL call_5_8_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_14_5
#define init_5_8_14_5
void FASTCALL call_5_8_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_14_6
#define init_5_8_14_6
void FASTCALL call_5_8_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_14_7
#define init_5_8_14_7
void FASTCALL call_5_8_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_14_8
#define init_5_8_14_8
void FASTCALL call_5_8_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_14_9
#define init_5_8_14_9
void FASTCALL call_5_8_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_14_10
#define init_5_8_14_10
void FASTCALL call_5_8_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_14_11
#define init_5_8_14_11
void FASTCALL call_5_8_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_14_12
#define init_5_8_14_12
void FASTCALL call_5_8_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_14_13
#define init_5_8_14_13
void FASTCALL call_5_8_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_14_14
#define init_5_8_14_14
void FASTCALL call_5_8_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_14_15
#define init_5_8_14_15
void FASTCALL call_5_8_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_15_0
#define init_5_8_15_0
void FASTCALL call_5_8_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_15_1
#define init_5_8_15_1
void FASTCALL call_5_8_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_15_2
#define init_5_8_15_2
void FASTCALL call_5_8_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_15_3
#define init_5_8_15_3
void FASTCALL call_5_8_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_15_4
#define init_5_8_15_4
void FASTCALL call_5_8_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_15_5
#define init_5_8_15_5
void FASTCALL call_5_8_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_15_6
#define init_5_8_15_6
void FASTCALL call_5_8_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_15_7
#define init_5_8_15_7
void FASTCALL call_5_8_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_15_8
#define init_5_8_15_8
void FASTCALL call_5_8_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_15_9
#define init_5_8_15_9
void FASTCALL call_5_8_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_15_10
#define init_5_8_15_10
void FASTCALL call_5_8_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_15_11
#define init_5_8_15_11
void FASTCALL call_5_8_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_15_12
#define init_5_8_15_12
void FASTCALL call_5_8_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_15_13
#define init_5_8_15_13
void FASTCALL call_5_8_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_15_14
#define init_5_8_15_14
void FASTCALL call_5_8_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_8_15_15
#define init_5_8_15_15
void FASTCALL call_5_8_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_0_0
#define init_5_9_0_0
void FASTCALL call_5_9_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_0_1
#define init_5_9_0_1
void FASTCALL call_5_9_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_0_2
#define init_5_9_0_2
void FASTCALL call_5_9_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_0_3
#define init_5_9_0_3
void FASTCALL call_5_9_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_0_4
#define init_5_9_0_4
void FASTCALL call_5_9_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_0_5
#define init_5_9_0_5
void FASTCALL call_5_9_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_0_6
#define init_5_9_0_6
void FASTCALL call_5_9_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_0_7
#define init_5_9_0_7
void FASTCALL call_5_9_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_0_8
#define init_5_9_0_8
void FASTCALL call_5_9_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_0_9
#define init_5_9_0_9
void FASTCALL call_5_9_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_0_10
#define init_5_9_0_10
void FASTCALL call_5_9_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_0_11
#define init_5_9_0_11
void FASTCALL call_5_9_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_0_12
#define init_5_9_0_12
void FASTCALL call_5_9_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_0_13
#define init_5_9_0_13
void FASTCALL call_5_9_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_0_14
#define init_5_9_0_14
void FASTCALL call_5_9_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_0_15
#define init_5_9_0_15
void FASTCALL call_5_9_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_1_0
#define init_5_9_1_0
void FASTCALL call_5_9_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_1_1
#define init_5_9_1_1
void FASTCALL call_5_9_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_1_2
#define init_5_9_1_2
void FASTCALL call_5_9_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_1_3
#define init_5_9_1_3
void FASTCALL call_5_9_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_1_4
#define init_5_9_1_4
void FASTCALL call_5_9_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_1_5
#define init_5_9_1_5
void FASTCALL call_5_9_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_1_6
#define init_5_9_1_6
void FASTCALL call_5_9_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_1_7
#define init_5_9_1_7
void FASTCALL call_5_9_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_1_8
#define init_5_9_1_8
void FASTCALL call_5_9_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_1_9
#define init_5_9_1_9
void FASTCALL call_5_9_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_1_10
#define init_5_9_1_10
void FASTCALL call_5_9_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_1_11
#define init_5_9_1_11
void FASTCALL call_5_9_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_1_12
#define init_5_9_1_12
void FASTCALL call_5_9_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_1_13
#define init_5_9_1_13
void FASTCALL call_5_9_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_1_14
#define init_5_9_1_14
void FASTCALL call_5_9_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_1_15
#define init_5_9_1_15
void FASTCALL call_5_9_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_2_0
#define init_5_9_2_0
void FASTCALL call_5_9_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_2_1
#define init_5_9_2_1
void FASTCALL call_5_9_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_2_2
#define init_5_9_2_2
void FASTCALL call_5_9_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_2_3
#define init_5_9_2_3
void FASTCALL call_5_9_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_2_4
#define init_5_9_2_4
void FASTCALL call_5_9_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_2_5
#define init_5_9_2_5
void FASTCALL call_5_9_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_2_6
#define init_5_9_2_6
void FASTCALL call_5_9_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_2_7
#define init_5_9_2_7
void FASTCALL call_5_9_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_2_8
#define init_5_9_2_8
void FASTCALL call_5_9_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_2_9
#define init_5_9_2_9
void FASTCALL call_5_9_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_2_10
#define init_5_9_2_10
void FASTCALL call_5_9_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_2_11
#define init_5_9_2_11
void FASTCALL call_5_9_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_2_12
#define init_5_9_2_12
void FASTCALL call_5_9_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_2_13
#define init_5_9_2_13
void FASTCALL call_5_9_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_2_14
#define init_5_9_2_14
void FASTCALL call_5_9_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_2_15
#define init_5_9_2_15
void FASTCALL call_5_9_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_3_0
#define init_5_9_3_0
void FASTCALL call_5_9_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_3_1
#define init_5_9_3_1
void FASTCALL call_5_9_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_3_2
#define init_5_9_3_2
void FASTCALL call_5_9_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_3_3
#define init_5_9_3_3
void FASTCALL call_5_9_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_3_4
#define init_5_9_3_4
void FASTCALL call_5_9_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_3_5
#define init_5_9_3_5
void FASTCALL call_5_9_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_3_6
#define init_5_9_3_6
void FASTCALL call_5_9_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_3_7
#define init_5_9_3_7
void FASTCALL call_5_9_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_3_8
#define init_5_9_3_8
void FASTCALL call_5_9_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_3_9
#define init_5_9_3_9
void FASTCALL call_5_9_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_3_10
#define init_5_9_3_10
void FASTCALL call_5_9_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_3_11
#define init_5_9_3_11
void FASTCALL call_5_9_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_3_12
#define init_5_9_3_12
void FASTCALL call_5_9_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_3_13
#define init_5_9_3_13
void FASTCALL call_5_9_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_3_14
#define init_5_9_3_14
void FASTCALL call_5_9_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_3_15
#define init_5_9_3_15
void FASTCALL call_5_9_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_4_0
#define init_5_9_4_0
void FASTCALL call_5_9_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_4_1
#define init_5_9_4_1
void FASTCALL call_5_9_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_4_2
#define init_5_9_4_2
void FASTCALL call_5_9_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_4_3
#define init_5_9_4_3
void FASTCALL call_5_9_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_4_4
#define init_5_9_4_4
void FASTCALL call_5_9_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_4_5
#define init_5_9_4_5
void FASTCALL call_5_9_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_4_6
#define init_5_9_4_6
void FASTCALL call_5_9_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_4_7
#define init_5_9_4_7
void FASTCALL call_5_9_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_4_8
#define init_5_9_4_8
void FASTCALL call_5_9_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_4_9
#define init_5_9_4_9
void FASTCALL call_5_9_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_4_10
#define init_5_9_4_10
void FASTCALL call_5_9_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_4_11
#define init_5_9_4_11
void FASTCALL call_5_9_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_4_12
#define init_5_9_4_12
void FASTCALL call_5_9_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_4_13
#define init_5_9_4_13
void FASTCALL call_5_9_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_4_14
#define init_5_9_4_14
void FASTCALL call_5_9_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_4_15
#define init_5_9_4_15
void FASTCALL call_5_9_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_5_0
#define init_5_9_5_0
void FASTCALL call_5_9_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_5_1
#define init_5_9_5_1
void FASTCALL call_5_9_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_5_2
#define init_5_9_5_2
void FASTCALL call_5_9_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_5_3
#define init_5_9_5_3
void FASTCALL call_5_9_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_5_4
#define init_5_9_5_4
void FASTCALL call_5_9_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_5_5
#define init_5_9_5_5
void FASTCALL call_5_9_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_5_6
#define init_5_9_5_6
void FASTCALL call_5_9_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_5_7
#define init_5_9_5_7
void FASTCALL call_5_9_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_5_8
#define init_5_9_5_8
void FASTCALL call_5_9_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_5_9
#define init_5_9_5_9
void FASTCALL call_5_9_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_5_10
#define init_5_9_5_10
void FASTCALL call_5_9_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_5_11
#define init_5_9_5_11
void FASTCALL call_5_9_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_5_12
#define init_5_9_5_12
void FASTCALL call_5_9_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_5_13
#define init_5_9_5_13
void FASTCALL call_5_9_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_5_14
#define init_5_9_5_14
void FASTCALL call_5_9_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_5_15
#define init_5_9_5_15
void FASTCALL call_5_9_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_6_0
#define init_5_9_6_0
void FASTCALL call_5_9_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_6_1
#define init_5_9_6_1
void FASTCALL call_5_9_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_6_2
#define init_5_9_6_2
void FASTCALL call_5_9_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_6_3
#define init_5_9_6_3
void FASTCALL call_5_9_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_6_4
#define init_5_9_6_4
void FASTCALL call_5_9_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_6_5
#define init_5_9_6_5
void FASTCALL call_5_9_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_6_6
#define init_5_9_6_6
void FASTCALL call_5_9_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_6_7
#define init_5_9_6_7
void FASTCALL call_5_9_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_6_8
#define init_5_9_6_8
void FASTCALL call_5_9_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_6_9
#define init_5_9_6_9
void FASTCALL call_5_9_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_6_10
#define init_5_9_6_10
void FASTCALL call_5_9_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_6_11
#define init_5_9_6_11
void FASTCALL call_5_9_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_6_12
#define init_5_9_6_12
void FASTCALL call_5_9_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_6_13
#define init_5_9_6_13
void FASTCALL call_5_9_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_6_14
#define init_5_9_6_14
void FASTCALL call_5_9_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_6_15
#define init_5_9_6_15
void FASTCALL call_5_9_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_7_0
#define init_5_9_7_0
void FASTCALL call_5_9_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_7_1
#define init_5_9_7_1
void FASTCALL call_5_9_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_7_2
#define init_5_9_7_2
void FASTCALL call_5_9_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_7_3
#define init_5_9_7_3
void FASTCALL call_5_9_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_7_4
#define init_5_9_7_4
void FASTCALL call_5_9_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_7_5
#define init_5_9_7_5
void FASTCALL call_5_9_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_7_6
#define init_5_9_7_6
void FASTCALL call_5_9_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_7_7
#define init_5_9_7_7
void FASTCALL call_5_9_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_7_8
#define init_5_9_7_8
void FASTCALL call_5_9_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_7_9
#define init_5_9_7_9
void FASTCALL call_5_9_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_7_10
#define init_5_9_7_10
void FASTCALL call_5_9_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_7_11
#define init_5_9_7_11
void FASTCALL call_5_9_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_7_12
#define init_5_9_7_12
void FASTCALL call_5_9_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_7_13
#define init_5_9_7_13
void FASTCALL call_5_9_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_7_14
#define init_5_9_7_14
void FASTCALL call_5_9_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_7_15
#define init_5_9_7_15
void FASTCALL call_5_9_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_8_0
#define init_5_9_8_0
void FASTCALL call_5_9_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_8_1
#define init_5_9_8_1
void FASTCALL call_5_9_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_8_2
#define init_5_9_8_2
void FASTCALL call_5_9_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_8_3
#define init_5_9_8_3
void FASTCALL call_5_9_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_8_4
#define init_5_9_8_4
void FASTCALL call_5_9_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_8_5
#define init_5_9_8_5
void FASTCALL call_5_9_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_8_6
#define init_5_9_8_6
void FASTCALL call_5_9_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_8_7
#define init_5_9_8_7
void FASTCALL call_5_9_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_8_8
#define init_5_9_8_8
void FASTCALL call_5_9_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_8_9
#define init_5_9_8_9
void FASTCALL call_5_9_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_8_10
#define init_5_9_8_10
void FASTCALL call_5_9_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_8_11
#define init_5_9_8_11
void FASTCALL call_5_9_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_8_12
#define init_5_9_8_12
void FASTCALL call_5_9_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_8_13
#define init_5_9_8_13
void FASTCALL call_5_9_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_8_14
#define init_5_9_8_14
void FASTCALL call_5_9_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_8_15
#define init_5_9_8_15
void FASTCALL call_5_9_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_9_0
#define init_5_9_9_0
void FASTCALL call_5_9_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_9_1
#define init_5_9_9_1
void FASTCALL call_5_9_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_9_2
#define init_5_9_9_2
void FASTCALL call_5_9_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_9_3
#define init_5_9_9_3
void FASTCALL call_5_9_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_9_4
#define init_5_9_9_4
void FASTCALL call_5_9_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_9_5
#define init_5_9_9_5
void FASTCALL call_5_9_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_9_6
#define init_5_9_9_6
void FASTCALL call_5_9_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_9_7
#define init_5_9_9_7
void FASTCALL call_5_9_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_9_8
#define init_5_9_9_8
void FASTCALL call_5_9_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_9_9
#define init_5_9_9_9
void FASTCALL call_5_9_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_9_10
#define init_5_9_9_10
void FASTCALL call_5_9_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_9_11
#define init_5_9_9_11
void FASTCALL call_5_9_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_9_12
#define init_5_9_9_12
void FASTCALL call_5_9_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_9_13
#define init_5_9_9_13
void FASTCALL call_5_9_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_9_14
#define init_5_9_9_14
void FASTCALL call_5_9_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_9_15
#define init_5_9_9_15
void FASTCALL call_5_9_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_10_0
#define init_5_9_10_0
void FASTCALL call_5_9_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_10_1
#define init_5_9_10_1
void FASTCALL call_5_9_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_10_2
#define init_5_9_10_2
void FASTCALL call_5_9_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_10_3
#define init_5_9_10_3
void FASTCALL call_5_9_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_10_4
#define init_5_9_10_4
void FASTCALL call_5_9_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_10_5
#define init_5_9_10_5
void FASTCALL call_5_9_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_10_6
#define init_5_9_10_6
void FASTCALL call_5_9_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_10_7
#define init_5_9_10_7
void FASTCALL call_5_9_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_10_8
#define init_5_9_10_8
void FASTCALL call_5_9_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_10_9
#define init_5_9_10_9
void FASTCALL call_5_9_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_10_10
#define init_5_9_10_10
void FASTCALL call_5_9_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_10_11
#define init_5_9_10_11
void FASTCALL call_5_9_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_10_12
#define init_5_9_10_12
void FASTCALL call_5_9_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_10_13
#define init_5_9_10_13
void FASTCALL call_5_9_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_10_14
#define init_5_9_10_14
void FASTCALL call_5_9_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_10_15
#define init_5_9_10_15
void FASTCALL call_5_9_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_11_0
#define init_5_9_11_0
void FASTCALL call_5_9_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_11_1
#define init_5_9_11_1
void FASTCALL call_5_9_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_11_2
#define init_5_9_11_2
void FASTCALL call_5_9_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_11_3
#define init_5_9_11_3
void FASTCALL call_5_9_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_11_4
#define init_5_9_11_4
void FASTCALL call_5_9_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_11_5
#define init_5_9_11_5
void FASTCALL call_5_9_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_11_6
#define init_5_9_11_6
void FASTCALL call_5_9_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_11_7
#define init_5_9_11_7
void FASTCALL call_5_9_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_11_8
#define init_5_9_11_8
void FASTCALL call_5_9_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_11_9
#define init_5_9_11_9
void FASTCALL call_5_9_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_11_10
#define init_5_9_11_10
void FASTCALL call_5_9_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_11_11
#define init_5_9_11_11
void FASTCALL call_5_9_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_11_12
#define init_5_9_11_12
void FASTCALL call_5_9_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_11_13
#define init_5_9_11_13
void FASTCALL call_5_9_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_11_14
#define init_5_9_11_14
void FASTCALL call_5_9_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_11_15
#define init_5_9_11_15
void FASTCALL call_5_9_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_12_0
#define init_5_9_12_0
void FASTCALL call_5_9_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_12_1
#define init_5_9_12_1
void FASTCALL call_5_9_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_12_2
#define init_5_9_12_2
void FASTCALL call_5_9_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_12_3
#define init_5_9_12_3
void FASTCALL call_5_9_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_12_4
#define init_5_9_12_4
void FASTCALL call_5_9_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_12_5
#define init_5_9_12_5
void FASTCALL call_5_9_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_12_6
#define init_5_9_12_6
void FASTCALL call_5_9_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_12_7
#define init_5_9_12_7
void FASTCALL call_5_9_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_12_8
#define init_5_9_12_8
void FASTCALL call_5_9_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_12_9
#define init_5_9_12_9
void FASTCALL call_5_9_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_12_10
#define init_5_9_12_10
void FASTCALL call_5_9_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_12_11
#define init_5_9_12_11
void FASTCALL call_5_9_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_12_12
#define init_5_9_12_12
void FASTCALL call_5_9_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_12_13
#define init_5_9_12_13
void FASTCALL call_5_9_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_12_14
#define init_5_9_12_14
void FASTCALL call_5_9_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_12_15
#define init_5_9_12_15
void FASTCALL call_5_9_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_13_0
#define init_5_9_13_0
void FASTCALL call_5_9_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_13_1
#define init_5_9_13_1
void FASTCALL call_5_9_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_13_2
#define init_5_9_13_2
void FASTCALL call_5_9_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_13_3
#define init_5_9_13_3
void FASTCALL call_5_9_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_13_4
#define init_5_9_13_4
void FASTCALL call_5_9_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_13_5
#define init_5_9_13_5
void FASTCALL call_5_9_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_13_6
#define init_5_9_13_6
void FASTCALL call_5_9_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_13_7
#define init_5_9_13_7
void FASTCALL call_5_9_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_13_8
#define init_5_9_13_8
void FASTCALL call_5_9_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_13_9
#define init_5_9_13_9
void FASTCALL call_5_9_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_13_10
#define init_5_9_13_10
void FASTCALL call_5_9_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_13_11
#define init_5_9_13_11
void FASTCALL call_5_9_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_13_12
#define init_5_9_13_12
void FASTCALL call_5_9_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_13_13
#define init_5_9_13_13
void FASTCALL call_5_9_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_13_14
#define init_5_9_13_14
void FASTCALL call_5_9_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_13_15
#define init_5_9_13_15
void FASTCALL call_5_9_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_14_0
#define init_5_9_14_0
void FASTCALL call_5_9_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_14_1
#define init_5_9_14_1
void FASTCALL call_5_9_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_14_2
#define init_5_9_14_2
void FASTCALL call_5_9_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_14_3
#define init_5_9_14_3
void FASTCALL call_5_9_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_14_4
#define init_5_9_14_4
void FASTCALL call_5_9_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_14_5
#define init_5_9_14_5
void FASTCALL call_5_9_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_14_6
#define init_5_9_14_6
void FASTCALL call_5_9_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_14_7
#define init_5_9_14_7
void FASTCALL call_5_9_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_14_8
#define init_5_9_14_8
void FASTCALL call_5_9_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_14_9
#define init_5_9_14_9
void FASTCALL call_5_9_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_14_10
#define init_5_9_14_10
void FASTCALL call_5_9_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_14_11
#define init_5_9_14_11
void FASTCALL call_5_9_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_14_12
#define init_5_9_14_12
void FASTCALL call_5_9_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_14_13
#define init_5_9_14_13
void FASTCALL call_5_9_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_14_14
#define init_5_9_14_14
void FASTCALL call_5_9_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_14_15
#define init_5_9_14_15
void FASTCALL call_5_9_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_15_0
#define init_5_9_15_0
void FASTCALL call_5_9_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_15_1
#define init_5_9_15_1
void FASTCALL call_5_9_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_15_2
#define init_5_9_15_2
void FASTCALL call_5_9_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_15_3
#define init_5_9_15_3
void FASTCALL call_5_9_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_15_4
#define init_5_9_15_4
void FASTCALL call_5_9_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_15_5
#define init_5_9_15_5
void FASTCALL call_5_9_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_15_6
#define init_5_9_15_6
void FASTCALL call_5_9_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_15_7
#define init_5_9_15_7
void FASTCALL call_5_9_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_15_8
#define init_5_9_15_8
void FASTCALL call_5_9_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_15_9
#define init_5_9_15_9
void FASTCALL call_5_9_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_15_10
#define init_5_9_15_10
void FASTCALL call_5_9_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_15_11
#define init_5_9_15_11
void FASTCALL call_5_9_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_15_12
#define init_5_9_15_12
void FASTCALL call_5_9_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_15_13
#define init_5_9_15_13
void FASTCALL call_5_9_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_15_14
#define init_5_9_15_14
void FASTCALL call_5_9_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_9_15_15
#define init_5_9_15_15
void FASTCALL call_5_9_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_0_0
#define init_5_10_0_0
void FASTCALL call_5_10_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_0_1
#define init_5_10_0_1
void FASTCALL call_5_10_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_0_2
#define init_5_10_0_2
void FASTCALL call_5_10_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_0_3
#define init_5_10_0_3
void FASTCALL call_5_10_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_0_4
#define init_5_10_0_4
void FASTCALL call_5_10_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_0_5
#define init_5_10_0_5
void FASTCALL call_5_10_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_0_6
#define init_5_10_0_6
void FASTCALL call_5_10_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_0_7
#define init_5_10_0_7
void FASTCALL call_5_10_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_0_8
#define init_5_10_0_8
void FASTCALL call_5_10_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_0_9
#define init_5_10_0_9
void FASTCALL call_5_10_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_0_10
#define init_5_10_0_10
void FASTCALL call_5_10_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_0_11
#define init_5_10_0_11
void FASTCALL call_5_10_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_0_12
#define init_5_10_0_12
void FASTCALL call_5_10_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_0_13
#define init_5_10_0_13
void FASTCALL call_5_10_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_0_14
#define init_5_10_0_14
void FASTCALL call_5_10_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_0_15
#define init_5_10_0_15
void FASTCALL call_5_10_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_1_0
#define init_5_10_1_0
void FASTCALL call_5_10_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_1_1
#define init_5_10_1_1
void FASTCALL call_5_10_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_1_2
#define init_5_10_1_2
void FASTCALL call_5_10_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_1_3
#define init_5_10_1_3
void FASTCALL call_5_10_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_1_4
#define init_5_10_1_4
void FASTCALL call_5_10_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_1_5
#define init_5_10_1_5
void FASTCALL call_5_10_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_1_6
#define init_5_10_1_6
void FASTCALL call_5_10_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_1_7
#define init_5_10_1_7
void FASTCALL call_5_10_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_1_8
#define init_5_10_1_8
void FASTCALL call_5_10_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_1_9
#define init_5_10_1_9
void FASTCALL call_5_10_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_1_10
#define init_5_10_1_10
void FASTCALL call_5_10_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_1_11
#define init_5_10_1_11
void FASTCALL call_5_10_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_1_12
#define init_5_10_1_12
void FASTCALL call_5_10_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_1_13
#define init_5_10_1_13
void FASTCALL call_5_10_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_1_14
#define init_5_10_1_14
void FASTCALL call_5_10_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_1_15
#define init_5_10_1_15
void FASTCALL call_5_10_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_2_0
#define init_5_10_2_0
void FASTCALL call_5_10_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_2_1
#define init_5_10_2_1
void FASTCALL call_5_10_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_2_2
#define init_5_10_2_2
void FASTCALL call_5_10_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_2_3
#define init_5_10_2_3
void FASTCALL call_5_10_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_2_4
#define init_5_10_2_4
void FASTCALL call_5_10_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_2_5
#define init_5_10_2_5
void FASTCALL call_5_10_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_2_6
#define init_5_10_2_6
void FASTCALL call_5_10_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_2_7
#define init_5_10_2_7
void FASTCALL call_5_10_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_2_8
#define init_5_10_2_8
void FASTCALL call_5_10_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_2_9
#define init_5_10_2_9
void FASTCALL call_5_10_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_2_10
#define init_5_10_2_10
void FASTCALL call_5_10_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_2_11
#define init_5_10_2_11
void FASTCALL call_5_10_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_2_12
#define init_5_10_2_12
void FASTCALL call_5_10_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_2_13
#define init_5_10_2_13
void FASTCALL call_5_10_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_2_14
#define init_5_10_2_14
void FASTCALL call_5_10_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_2_15
#define init_5_10_2_15
void FASTCALL call_5_10_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_3_0
#define init_5_10_3_0
void FASTCALL call_5_10_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_3_1
#define init_5_10_3_1
void FASTCALL call_5_10_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_3_2
#define init_5_10_3_2
void FASTCALL call_5_10_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_3_3
#define init_5_10_3_3
void FASTCALL call_5_10_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_3_4
#define init_5_10_3_4
void FASTCALL call_5_10_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_3_5
#define init_5_10_3_5
void FASTCALL call_5_10_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_3_6
#define init_5_10_3_6
void FASTCALL call_5_10_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_3_7
#define init_5_10_3_7
void FASTCALL call_5_10_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_3_8
#define init_5_10_3_8
void FASTCALL call_5_10_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_3_9
#define init_5_10_3_9
void FASTCALL call_5_10_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_3_10
#define init_5_10_3_10
void FASTCALL call_5_10_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_3_11
#define init_5_10_3_11
void FASTCALL call_5_10_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_3_12
#define init_5_10_3_12
void FASTCALL call_5_10_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_3_13
#define init_5_10_3_13
void FASTCALL call_5_10_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_3_14
#define init_5_10_3_14
void FASTCALL call_5_10_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_3_15
#define init_5_10_3_15
void FASTCALL call_5_10_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_4_0
#define init_5_10_4_0
void FASTCALL call_5_10_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_4_1
#define init_5_10_4_1
void FASTCALL call_5_10_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_4_2
#define init_5_10_4_2
void FASTCALL call_5_10_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_4_3
#define init_5_10_4_3
void FASTCALL call_5_10_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_4_4
#define init_5_10_4_4
void FASTCALL call_5_10_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_4_5
#define init_5_10_4_5
void FASTCALL call_5_10_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_4_6
#define init_5_10_4_6
void FASTCALL call_5_10_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_4_7
#define init_5_10_4_7
void FASTCALL call_5_10_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_4_8
#define init_5_10_4_8
void FASTCALL call_5_10_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_4_9
#define init_5_10_4_9
void FASTCALL call_5_10_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_4_10
#define init_5_10_4_10
void FASTCALL call_5_10_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_4_11
#define init_5_10_4_11
void FASTCALL call_5_10_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_4_12
#define init_5_10_4_12
void FASTCALL call_5_10_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_4_13
#define init_5_10_4_13
void FASTCALL call_5_10_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_4_14
#define init_5_10_4_14
void FASTCALL call_5_10_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_4_15
#define init_5_10_4_15
void FASTCALL call_5_10_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_5_0
#define init_5_10_5_0
void FASTCALL call_5_10_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_5_1
#define init_5_10_5_1
void FASTCALL call_5_10_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_5_2
#define init_5_10_5_2
void FASTCALL call_5_10_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_5_3
#define init_5_10_5_3
void FASTCALL call_5_10_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_5_4
#define init_5_10_5_4
void FASTCALL call_5_10_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_5_5
#define init_5_10_5_5
void FASTCALL call_5_10_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_5_6
#define init_5_10_5_6
void FASTCALL call_5_10_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_5_7
#define init_5_10_5_7
void FASTCALL call_5_10_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_5_8
#define init_5_10_5_8
void FASTCALL call_5_10_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_5_9
#define init_5_10_5_9
void FASTCALL call_5_10_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_5_10
#define init_5_10_5_10
void FASTCALL call_5_10_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_5_11
#define init_5_10_5_11
void FASTCALL call_5_10_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_5_12
#define init_5_10_5_12
void FASTCALL call_5_10_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_5_13
#define init_5_10_5_13
void FASTCALL call_5_10_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_5_14
#define init_5_10_5_14
void FASTCALL call_5_10_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_5_15
#define init_5_10_5_15
void FASTCALL call_5_10_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_6_0
#define init_5_10_6_0
void FASTCALL call_5_10_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_6_1
#define init_5_10_6_1
void FASTCALL call_5_10_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_6_2
#define init_5_10_6_2
void FASTCALL call_5_10_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_6_3
#define init_5_10_6_3
void FASTCALL call_5_10_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_6_4
#define init_5_10_6_4
void FASTCALL call_5_10_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_6_5
#define init_5_10_6_5
void FASTCALL call_5_10_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_6_6
#define init_5_10_6_6
void FASTCALL call_5_10_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_6_7
#define init_5_10_6_7
void FASTCALL call_5_10_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_6_8
#define init_5_10_6_8
void FASTCALL call_5_10_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_6_9
#define init_5_10_6_9
void FASTCALL call_5_10_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_6_10
#define init_5_10_6_10
void FASTCALL call_5_10_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_6_11
#define init_5_10_6_11
void FASTCALL call_5_10_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_6_12
#define init_5_10_6_12
void FASTCALL call_5_10_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_6_13
#define init_5_10_6_13
void FASTCALL call_5_10_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_6_14
#define init_5_10_6_14
void FASTCALL call_5_10_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_6_15
#define init_5_10_6_15
void FASTCALL call_5_10_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_7_0
#define init_5_10_7_0
void FASTCALL call_5_10_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_7_1
#define init_5_10_7_1
void FASTCALL call_5_10_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_7_2
#define init_5_10_7_2
void FASTCALL call_5_10_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_7_3
#define init_5_10_7_3
void FASTCALL call_5_10_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_7_4
#define init_5_10_7_4
void FASTCALL call_5_10_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_7_5
#define init_5_10_7_5
void FASTCALL call_5_10_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_7_6
#define init_5_10_7_6
void FASTCALL call_5_10_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_7_7
#define init_5_10_7_7
void FASTCALL call_5_10_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_7_8
#define init_5_10_7_8
void FASTCALL call_5_10_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_7_9
#define init_5_10_7_9
void FASTCALL call_5_10_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_7_10
#define init_5_10_7_10
void FASTCALL call_5_10_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_7_11
#define init_5_10_7_11
void FASTCALL call_5_10_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_7_12
#define init_5_10_7_12
void FASTCALL call_5_10_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_7_13
#define init_5_10_7_13
void FASTCALL call_5_10_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_7_14
#define init_5_10_7_14
void FASTCALL call_5_10_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_7_15
#define init_5_10_7_15
void FASTCALL call_5_10_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_8_0
#define init_5_10_8_0
void FASTCALL call_5_10_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_8_1
#define init_5_10_8_1
void FASTCALL call_5_10_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_8_2
#define init_5_10_8_2
void FASTCALL call_5_10_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_8_3
#define init_5_10_8_3
void FASTCALL call_5_10_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_8_4
#define init_5_10_8_4
void FASTCALL call_5_10_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_8_5
#define init_5_10_8_5
void FASTCALL call_5_10_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_8_6
#define init_5_10_8_6
void FASTCALL call_5_10_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_8_7
#define init_5_10_8_7
void FASTCALL call_5_10_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_8_8
#define init_5_10_8_8
void FASTCALL call_5_10_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_8_9
#define init_5_10_8_9
void FASTCALL call_5_10_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_8_10
#define init_5_10_8_10
void FASTCALL call_5_10_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_8_11
#define init_5_10_8_11
void FASTCALL call_5_10_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_8_12
#define init_5_10_8_12
void FASTCALL call_5_10_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_8_13
#define init_5_10_8_13
void FASTCALL call_5_10_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_8_14
#define init_5_10_8_14
void FASTCALL call_5_10_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_8_15
#define init_5_10_8_15
void FASTCALL call_5_10_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_9_0
#define init_5_10_9_0
void FASTCALL call_5_10_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_9_1
#define init_5_10_9_1
void FASTCALL call_5_10_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_9_2
#define init_5_10_9_2
void FASTCALL call_5_10_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_9_3
#define init_5_10_9_3
void FASTCALL call_5_10_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_9_4
#define init_5_10_9_4
void FASTCALL call_5_10_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_9_5
#define init_5_10_9_5
void FASTCALL call_5_10_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_9_6
#define init_5_10_9_6
void FASTCALL call_5_10_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_9_7
#define init_5_10_9_7
void FASTCALL call_5_10_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_9_8
#define init_5_10_9_8
void FASTCALL call_5_10_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_9_9
#define init_5_10_9_9
void FASTCALL call_5_10_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_9_10
#define init_5_10_9_10
void FASTCALL call_5_10_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_9_11
#define init_5_10_9_11
void FASTCALL call_5_10_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_9_12
#define init_5_10_9_12
void FASTCALL call_5_10_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_9_13
#define init_5_10_9_13
void FASTCALL call_5_10_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_9_14
#define init_5_10_9_14
void FASTCALL call_5_10_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_9_15
#define init_5_10_9_15
void FASTCALL call_5_10_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_10_0
#define init_5_10_10_0
void FASTCALL call_5_10_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_10_1
#define init_5_10_10_1
void FASTCALL call_5_10_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_10_2
#define init_5_10_10_2
void FASTCALL call_5_10_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_10_3
#define init_5_10_10_3
void FASTCALL call_5_10_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_10_4
#define init_5_10_10_4
void FASTCALL call_5_10_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_10_5
#define init_5_10_10_5
void FASTCALL call_5_10_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_10_6
#define init_5_10_10_6
void FASTCALL call_5_10_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_10_7
#define init_5_10_10_7
void FASTCALL call_5_10_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_10_8
#define init_5_10_10_8
void FASTCALL call_5_10_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_10_9
#define init_5_10_10_9
void FASTCALL call_5_10_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_10_10
#define init_5_10_10_10
void FASTCALL call_5_10_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_10_11
#define init_5_10_10_11
void FASTCALL call_5_10_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_10_12
#define init_5_10_10_12
void FASTCALL call_5_10_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_10_13
#define init_5_10_10_13
void FASTCALL call_5_10_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_10_14
#define init_5_10_10_14
void FASTCALL call_5_10_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_10_15
#define init_5_10_10_15
void FASTCALL call_5_10_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_11_0
#define init_5_10_11_0
void FASTCALL call_5_10_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_11_1
#define init_5_10_11_1
void FASTCALL call_5_10_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_11_2
#define init_5_10_11_2
void FASTCALL call_5_10_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_11_3
#define init_5_10_11_3
void FASTCALL call_5_10_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_11_4
#define init_5_10_11_4
void FASTCALL call_5_10_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_11_5
#define init_5_10_11_5
void FASTCALL call_5_10_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_11_6
#define init_5_10_11_6
void FASTCALL call_5_10_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_11_7
#define init_5_10_11_7
void FASTCALL call_5_10_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_11_8
#define init_5_10_11_8
void FASTCALL call_5_10_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_11_9
#define init_5_10_11_9
void FASTCALL call_5_10_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_11_10
#define init_5_10_11_10
void FASTCALL call_5_10_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_11_11
#define init_5_10_11_11
void FASTCALL call_5_10_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_11_12
#define init_5_10_11_12
void FASTCALL call_5_10_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_11_13
#define init_5_10_11_13
void FASTCALL call_5_10_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_11_14
#define init_5_10_11_14
void FASTCALL call_5_10_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_11_15
#define init_5_10_11_15
void FASTCALL call_5_10_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_12_0
#define init_5_10_12_0
void FASTCALL call_5_10_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_12_1
#define init_5_10_12_1
void FASTCALL call_5_10_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_12_2
#define init_5_10_12_2
void FASTCALL call_5_10_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_12_3
#define init_5_10_12_3
void FASTCALL call_5_10_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_12_4
#define init_5_10_12_4
void FASTCALL call_5_10_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_12_5
#define init_5_10_12_5
void FASTCALL call_5_10_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_12_6
#define init_5_10_12_6
void FASTCALL call_5_10_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_12_7
#define init_5_10_12_7
void FASTCALL call_5_10_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_12_8
#define init_5_10_12_8
void FASTCALL call_5_10_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_12_9
#define init_5_10_12_9
void FASTCALL call_5_10_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_12_10
#define init_5_10_12_10
void FASTCALL call_5_10_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_12_11
#define init_5_10_12_11
void FASTCALL call_5_10_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_12_12
#define init_5_10_12_12
void FASTCALL call_5_10_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_12_13
#define init_5_10_12_13
void FASTCALL call_5_10_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_12_14
#define init_5_10_12_14
void FASTCALL call_5_10_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_12_15
#define init_5_10_12_15
void FASTCALL call_5_10_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_13_0
#define init_5_10_13_0
void FASTCALL call_5_10_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_13_1
#define init_5_10_13_1
void FASTCALL call_5_10_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_13_2
#define init_5_10_13_2
void FASTCALL call_5_10_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_13_3
#define init_5_10_13_3
void FASTCALL call_5_10_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_13_4
#define init_5_10_13_4
void FASTCALL call_5_10_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_13_5
#define init_5_10_13_5
void FASTCALL call_5_10_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_13_6
#define init_5_10_13_6
void FASTCALL call_5_10_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_13_7
#define init_5_10_13_7
void FASTCALL call_5_10_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_13_8
#define init_5_10_13_8
void FASTCALL call_5_10_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_13_9
#define init_5_10_13_9
void FASTCALL call_5_10_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_13_10
#define init_5_10_13_10
void FASTCALL call_5_10_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_13_11
#define init_5_10_13_11
void FASTCALL call_5_10_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_13_12
#define init_5_10_13_12
void FASTCALL call_5_10_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_13_13
#define init_5_10_13_13
void FASTCALL call_5_10_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_13_14
#define init_5_10_13_14
void FASTCALL call_5_10_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_13_15
#define init_5_10_13_15
void FASTCALL call_5_10_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_14_0
#define init_5_10_14_0
void FASTCALL call_5_10_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_14_1
#define init_5_10_14_1
void FASTCALL call_5_10_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_14_2
#define init_5_10_14_2
void FASTCALL call_5_10_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_14_3
#define init_5_10_14_3
void FASTCALL call_5_10_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_14_4
#define init_5_10_14_4
void FASTCALL call_5_10_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_14_5
#define init_5_10_14_5
void FASTCALL call_5_10_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_14_6
#define init_5_10_14_6
void FASTCALL call_5_10_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_14_7
#define init_5_10_14_7
void FASTCALL call_5_10_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_14_8
#define init_5_10_14_8
void FASTCALL call_5_10_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_14_9
#define init_5_10_14_9
void FASTCALL call_5_10_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_14_10
#define init_5_10_14_10
void FASTCALL call_5_10_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_14_11
#define init_5_10_14_11
void FASTCALL call_5_10_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_14_12
#define init_5_10_14_12
void FASTCALL call_5_10_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_14_13
#define init_5_10_14_13
void FASTCALL call_5_10_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_14_14
#define init_5_10_14_14
void FASTCALL call_5_10_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_14_15
#define init_5_10_14_15
void FASTCALL call_5_10_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_15_0
#define init_5_10_15_0
void FASTCALL call_5_10_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_15_1
#define init_5_10_15_1
void FASTCALL call_5_10_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_15_2
#define init_5_10_15_2
void FASTCALL call_5_10_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_15_3
#define init_5_10_15_3
void FASTCALL call_5_10_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_15_4
#define init_5_10_15_4
void FASTCALL call_5_10_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_15_5
#define init_5_10_15_5
void FASTCALL call_5_10_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_15_6
#define init_5_10_15_6
void FASTCALL call_5_10_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_15_7
#define init_5_10_15_7
void FASTCALL call_5_10_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_15_8
#define init_5_10_15_8
void FASTCALL call_5_10_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_15_9
#define init_5_10_15_9
void FASTCALL call_5_10_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_15_10
#define init_5_10_15_10
void FASTCALL call_5_10_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_15_11
#define init_5_10_15_11
void FASTCALL call_5_10_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_15_12
#define init_5_10_15_12
void FASTCALL call_5_10_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_15_13
#define init_5_10_15_13
void FASTCALL call_5_10_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_15_14
#define init_5_10_15_14
void FASTCALL call_5_10_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_10_15_15
#define init_5_10_15_15
void FASTCALL call_5_10_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_0_0
#define init_5_11_0_0
void FASTCALL call_5_11_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_0_1
#define init_5_11_0_1
void FASTCALL call_5_11_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_0_2
#define init_5_11_0_2
void FASTCALL call_5_11_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_0_3
#define init_5_11_0_3
void FASTCALL call_5_11_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_0_4
#define init_5_11_0_4
void FASTCALL call_5_11_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_0_5
#define init_5_11_0_5
void FASTCALL call_5_11_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_0_6
#define init_5_11_0_6
void FASTCALL call_5_11_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_0_7
#define init_5_11_0_7
void FASTCALL call_5_11_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_0_8
#define init_5_11_0_8
void FASTCALL call_5_11_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_0_9
#define init_5_11_0_9
void FASTCALL call_5_11_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_0_10
#define init_5_11_0_10
void FASTCALL call_5_11_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_0_11
#define init_5_11_0_11
void FASTCALL call_5_11_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_0_12
#define init_5_11_0_12
void FASTCALL call_5_11_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_0_13
#define init_5_11_0_13
void FASTCALL call_5_11_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_0_14
#define init_5_11_0_14
void FASTCALL call_5_11_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_0_15
#define init_5_11_0_15
void FASTCALL call_5_11_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_1_0
#define init_5_11_1_0
void FASTCALL call_5_11_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_1_1
#define init_5_11_1_1
void FASTCALL call_5_11_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_1_2
#define init_5_11_1_2
void FASTCALL call_5_11_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_1_3
#define init_5_11_1_3
void FASTCALL call_5_11_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_1_4
#define init_5_11_1_4
void FASTCALL call_5_11_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_1_5
#define init_5_11_1_5
void FASTCALL call_5_11_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_1_6
#define init_5_11_1_6
void FASTCALL call_5_11_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_1_7
#define init_5_11_1_7
void FASTCALL call_5_11_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_1_8
#define init_5_11_1_8
void FASTCALL call_5_11_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_1_9
#define init_5_11_1_9
void FASTCALL call_5_11_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_1_10
#define init_5_11_1_10
void FASTCALL call_5_11_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_1_11
#define init_5_11_1_11
void FASTCALL call_5_11_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_1_12
#define init_5_11_1_12
void FASTCALL call_5_11_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_1_13
#define init_5_11_1_13
void FASTCALL call_5_11_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_1_14
#define init_5_11_1_14
void FASTCALL call_5_11_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_1_15
#define init_5_11_1_15
void FASTCALL call_5_11_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_2_0
#define init_5_11_2_0
void FASTCALL call_5_11_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_2_1
#define init_5_11_2_1
void FASTCALL call_5_11_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_2_2
#define init_5_11_2_2
void FASTCALL call_5_11_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_2_3
#define init_5_11_2_3
void FASTCALL call_5_11_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_2_4
#define init_5_11_2_4
void FASTCALL call_5_11_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_2_5
#define init_5_11_2_5
void FASTCALL call_5_11_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_2_6
#define init_5_11_2_6
void FASTCALL call_5_11_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_2_7
#define init_5_11_2_7
void FASTCALL call_5_11_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_2_8
#define init_5_11_2_8
void FASTCALL call_5_11_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_2_9
#define init_5_11_2_9
void FASTCALL call_5_11_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_2_10
#define init_5_11_2_10
void FASTCALL call_5_11_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_2_11
#define init_5_11_2_11
void FASTCALL call_5_11_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_2_12
#define init_5_11_2_12
void FASTCALL call_5_11_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_2_13
#define init_5_11_2_13
void FASTCALL call_5_11_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_2_14
#define init_5_11_2_14
void FASTCALL call_5_11_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_2_15
#define init_5_11_2_15
void FASTCALL call_5_11_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_3_0
#define init_5_11_3_0
void FASTCALL call_5_11_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_3_1
#define init_5_11_3_1
void FASTCALL call_5_11_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_3_2
#define init_5_11_3_2
void FASTCALL call_5_11_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_3_3
#define init_5_11_3_3
void FASTCALL call_5_11_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_3_4
#define init_5_11_3_4
void FASTCALL call_5_11_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_3_5
#define init_5_11_3_5
void FASTCALL call_5_11_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_3_6
#define init_5_11_3_6
void FASTCALL call_5_11_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_3_7
#define init_5_11_3_7
void FASTCALL call_5_11_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_3_8
#define init_5_11_3_8
void FASTCALL call_5_11_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_3_9
#define init_5_11_3_9
void FASTCALL call_5_11_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_3_10
#define init_5_11_3_10
void FASTCALL call_5_11_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_3_11
#define init_5_11_3_11
void FASTCALL call_5_11_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_3_12
#define init_5_11_3_12
void FASTCALL call_5_11_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_3_13
#define init_5_11_3_13
void FASTCALL call_5_11_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_3_14
#define init_5_11_3_14
void FASTCALL call_5_11_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_3_15
#define init_5_11_3_15
void FASTCALL call_5_11_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_4_0
#define init_5_11_4_0
void FASTCALL call_5_11_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_4_1
#define init_5_11_4_1
void FASTCALL call_5_11_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_4_2
#define init_5_11_4_2
void FASTCALL call_5_11_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_4_3
#define init_5_11_4_3
void FASTCALL call_5_11_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_4_4
#define init_5_11_4_4
void FASTCALL call_5_11_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_4_5
#define init_5_11_4_5
void FASTCALL call_5_11_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_4_6
#define init_5_11_4_6
void FASTCALL call_5_11_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_4_7
#define init_5_11_4_7
void FASTCALL call_5_11_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_4_8
#define init_5_11_4_8
void FASTCALL call_5_11_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_4_9
#define init_5_11_4_9
void FASTCALL call_5_11_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_4_10
#define init_5_11_4_10
void FASTCALL call_5_11_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_4_11
#define init_5_11_4_11
void FASTCALL call_5_11_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_4_12
#define init_5_11_4_12
void FASTCALL call_5_11_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_4_13
#define init_5_11_4_13
void FASTCALL call_5_11_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_4_14
#define init_5_11_4_14
void FASTCALL call_5_11_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_4_15
#define init_5_11_4_15
void FASTCALL call_5_11_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_5_0
#define init_5_11_5_0
void FASTCALL call_5_11_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_5_1
#define init_5_11_5_1
void FASTCALL call_5_11_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_5_2
#define init_5_11_5_2
void FASTCALL call_5_11_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_5_3
#define init_5_11_5_3
void FASTCALL call_5_11_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_5_4
#define init_5_11_5_4
void FASTCALL call_5_11_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_5_5
#define init_5_11_5_5
void FASTCALL call_5_11_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_5_6
#define init_5_11_5_6
void FASTCALL call_5_11_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_5_7
#define init_5_11_5_7
void FASTCALL call_5_11_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_5_8
#define init_5_11_5_8
void FASTCALL call_5_11_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_5_9
#define init_5_11_5_9
void FASTCALL call_5_11_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_5_10
#define init_5_11_5_10
void FASTCALL call_5_11_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_5_11
#define init_5_11_5_11
void FASTCALL call_5_11_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_5_12
#define init_5_11_5_12
void FASTCALL call_5_11_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_5_13
#define init_5_11_5_13
void FASTCALL call_5_11_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_5_14
#define init_5_11_5_14
void FASTCALL call_5_11_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_5_15
#define init_5_11_5_15
void FASTCALL call_5_11_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_6_0
#define init_5_11_6_0
void FASTCALL call_5_11_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_6_1
#define init_5_11_6_1
void FASTCALL call_5_11_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_6_2
#define init_5_11_6_2
void FASTCALL call_5_11_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_6_3
#define init_5_11_6_3
void FASTCALL call_5_11_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_6_4
#define init_5_11_6_4
void FASTCALL call_5_11_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_6_5
#define init_5_11_6_5
void FASTCALL call_5_11_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_6_6
#define init_5_11_6_6
void FASTCALL call_5_11_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_6_7
#define init_5_11_6_7
void FASTCALL call_5_11_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_6_8
#define init_5_11_6_8
void FASTCALL call_5_11_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_6_9
#define init_5_11_6_9
void FASTCALL call_5_11_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_6_10
#define init_5_11_6_10
void FASTCALL call_5_11_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_6_11
#define init_5_11_6_11
void FASTCALL call_5_11_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_6_12
#define init_5_11_6_12
void FASTCALL call_5_11_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_6_13
#define init_5_11_6_13
void FASTCALL call_5_11_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_6_14
#define init_5_11_6_14
void FASTCALL call_5_11_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_6_15
#define init_5_11_6_15
void FASTCALL call_5_11_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_7_0
#define init_5_11_7_0
void FASTCALL call_5_11_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_7_1
#define init_5_11_7_1
void FASTCALL call_5_11_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_7_2
#define init_5_11_7_2
void FASTCALL call_5_11_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_7_3
#define init_5_11_7_3
void FASTCALL call_5_11_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_7_4
#define init_5_11_7_4
void FASTCALL call_5_11_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_7_5
#define init_5_11_7_5
void FASTCALL call_5_11_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_7_6
#define init_5_11_7_6
void FASTCALL call_5_11_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_7_7
#define init_5_11_7_7
void FASTCALL call_5_11_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_7_8
#define init_5_11_7_8
void FASTCALL call_5_11_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_7_9
#define init_5_11_7_9
void FASTCALL call_5_11_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_7_10
#define init_5_11_7_10
void FASTCALL call_5_11_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_7_11
#define init_5_11_7_11
void FASTCALL call_5_11_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_7_12
#define init_5_11_7_12
void FASTCALL call_5_11_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_7_13
#define init_5_11_7_13
void FASTCALL call_5_11_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_7_14
#define init_5_11_7_14
void FASTCALL call_5_11_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_7_15
#define init_5_11_7_15
void FASTCALL call_5_11_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_8_0
#define init_5_11_8_0
void FASTCALL call_5_11_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_8_1
#define init_5_11_8_1
void FASTCALL call_5_11_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_8_2
#define init_5_11_8_2
void FASTCALL call_5_11_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_8_3
#define init_5_11_8_3
void FASTCALL call_5_11_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_8_4
#define init_5_11_8_4
void FASTCALL call_5_11_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_8_5
#define init_5_11_8_5
void FASTCALL call_5_11_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_8_6
#define init_5_11_8_6
void FASTCALL call_5_11_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_8_7
#define init_5_11_8_7
void FASTCALL call_5_11_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_8_8
#define init_5_11_8_8
void FASTCALL call_5_11_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_8_9
#define init_5_11_8_9
void FASTCALL call_5_11_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_8_10
#define init_5_11_8_10
void FASTCALL call_5_11_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_8_11
#define init_5_11_8_11
void FASTCALL call_5_11_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_8_12
#define init_5_11_8_12
void FASTCALL call_5_11_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_8_13
#define init_5_11_8_13
void FASTCALL call_5_11_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_8_14
#define init_5_11_8_14
void FASTCALL call_5_11_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_8_15
#define init_5_11_8_15
void FASTCALL call_5_11_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_9_0
#define init_5_11_9_0
void FASTCALL call_5_11_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_9_1
#define init_5_11_9_1
void FASTCALL call_5_11_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_9_2
#define init_5_11_9_2
void FASTCALL call_5_11_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_9_3
#define init_5_11_9_3
void FASTCALL call_5_11_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_9_4
#define init_5_11_9_4
void FASTCALL call_5_11_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_9_5
#define init_5_11_9_5
void FASTCALL call_5_11_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_9_6
#define init_5_11_9_6
void FASTCALL call_5_11_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_9_7
#define init_5_11_9_7
void FASTCALL call_5_11_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_9_8
#define init_5_11_9_8
void FASTCALL call_5_11_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_9_9
#define init_5_11_9_9
void FASTCALL call_5_11_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_9_10
#define init_5_11_9_10
void FASTCALL call_5_11_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_9_11
#define init_5_11_9_11
void FASTCALL call_5_11_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_9_12
#define init_5_11_9_12
void FASTCALL call_5_11_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_9_13
#define init_5_11_9_13
void FASTCALL call_5_11_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_9_14
#define init_5_11_9_14
void FASTCALL call_5_11_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_9_15
#define init_5_11_9_15
void FASTCALL call_5_11_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_10_0
#define init_5_11_10_0
void FASTCALL call_5_11_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_10_1
#define init_5_11_10_1
void FASTCALL call_5_11_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_10_2
#define init_5_11_10_2
void FASTCALL call_5_11_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_10_3
#define init_5_11_10_3
void FASTCALL call_5_11_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_10_4
#define init_5_11_10_4
void FASTCALL call_5_11_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_10_5
#define init_5_11_10_5
void FASTCALL call_5_11_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_10_6
#define init_5_11_10_6
void FASTCALL call_5_11_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_10_7
#define init_5_11_10_7
void FASTCALL call_5_11_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_10_8
#define init_5_11_10_8
void FASTCALL call_5_11_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_10_9
#define init_5_11_10_9
void FASTCALL call_5_11_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_10_10
#define init_5_11_10_10
void FASTCALL call_5_11_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_10_11
#define init_5_11_10_11
void FASTCALL call_5_11_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_10_12
#define init_5_11_10_12
void FASTCALL call_5_11_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_10_13
#define init_5_11_10_13
void FASTCALL call_5_11_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_10_14
#define init_5_11_10_14
void FASTCALL call_5_11_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_10_15
#define init_5_11_10_15
void FASTCALL call_5_11_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_11_0
#define init_5_11_11_0
void FASTCALL call_5_11_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_11_1
#define init_5_11_11_1
void FASTCALL call_5_11_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_11_2
#define init_5_11_11_2
void FASTCALL call_5_11_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_11_3
#define init_5_11_11_3
void FASTCALL call_5_11_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_11_4
#define init_5_11_11_4
void FASTCALL call_5_11_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_11_5
#define init_5_11_11_5
void FASTCALL call_5_11_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_11_6
#define init_5_11_11_6
void FASTCALL call_5_11_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_11_7
#define init_5_11_11_7
void FASTCALL call_5_11_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_11_8
#define init_5_11_11_8
void FASTCALL call_5_11_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_11_9
#define init_5_11_11_9
void FASTCALL call_5_11_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_11_10
#define init_5_11_11_10
void FASTCALL call_5_11_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_11_11
#define init_5_11_11_11
void FASTCALL call_5_11_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_11_12
#define init_5_11_11_12
void FASTCALL call_5_11_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_11_13
#define init_5_11_11_13
void FASTCALL call_5_11_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_11_14
#define init_5_11_11_14
void FASTCALL call_5_11_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_11_15
#define init_5_11_11_15
void FASTCALL call_5_11_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_12_0
#define init_5_11_12_0
void FASTCALL call_5_11_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_12_1
#define init_5_11_12_1
void FASTCALL call_5_11_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_12_2
#define init_5_11_12_2
void FASTCALL call_5_11_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_12_3
#define init_5_11_12_3
void FASTCALL call_5_11_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_12_4
#define init_5_11_12_4
void FASTCALL call_5_11_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_12_5
#define init_5_11_12_5
void FASTCALL call_5_11_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_12_6
#define init_5_11_12_6
void FASTCALL call_5_11_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_12_7
#define init_5_11_12_7
void FASTCALL call_5_11_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_12_8
#define init_5_11_12_8
void FASTCALL call_5_11_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_12_9
#define init_5_11_12_9
void FASTCALL call_5_11_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_12_10
#define init_5_11_12_10
void FASTCALL call_5_11_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_12_11
#define init_5_11_12_11
void FASTCALL call_5_11_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_12_12
#define init_5_11_12_12
void FASTCALL call_5_11_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_12_13
#define init_5_11_12_13
void FASTCALL call_5_11_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_12_14
#define init_5_11_12_14
void FASTCALL call_5_11_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_12_15
#define init_5_11_12_15
void FASTCALL call_5_11_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_13_0
#define init_5_11_13_0
void FASTCALL call_5_11_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_13_1
#define init_5_11_13_1
void FASTCALL call_5_11_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_13_2
#define init_5_11_13_2
void FASTCALL call_5_11_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_13_3
#define init_5_11_13_3
void FASTCALL call_5_11_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_13_4
#define init_5_11_13_4
void FASTCALL call_5_11_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_13_5
#define init_5_11_13_5
void FASTCALL call_5_11_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_13_6
#define init_5_11_13_6
void FASTCALL call_5_11_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_13_7
#define init_5_11_13_7
void FASTCALL call_5_11_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_13_8
#define init_5_11_13_8
void FASTCALL call_5_11_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_13_9
#define init_5_11_13_9
void FASTCALL call_5_11_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_13_10
#define init_5_11_13_10
void FASTCALL call_5_11_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_13_11
#define init_5_11_13_11
void FASTCALL call_5_11_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_13_12
#define init_5_11_13_12
void FASTCALL call_5_11_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_13_13
#define init_5_11_13_13
void FASTCALL call_5_11_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_13_14
#define init_5_11_13_14
void FASTCALL call_5_11_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_13_15
#define init_5_11_13_15
void FASTCALL call_5_11_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_14_0
#define init_5_11_14_0
void FASTCALL call_5_11_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_14_1
#define init_5_11_14_1
void FASTCALL call_5_11_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_14_2
#define init_5_11_14_2
void FASTCALL call_5_11_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_14_3
#define init_5_11_14_3
void FASTCALL call_5_11_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_14_4
#define init_5_11_14_4
void FASTCALL call_5_11_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_14_5
#define init_5_11_14_5
void FASTCALL call_5_11_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_14_6
#define init_5_11_14_6
void FASTCALL call_5_11_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_14_7
#define init_5_11_14_7
void FASTCALL call_5_11_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_14_8
#define init_5_11_14_8
void FASTCALL call_5_11_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_14_9
#define init_5_11_14_9
void FASTCALL call_5_11_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_14_10
#define init_5_11_14_10
void FASTCALL call_5_11_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_14_11
#define init_5_11_14_11
void FASTCALL call_5_11_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_14_12
#define init_5_11_14_12
void FASTCALL call_5_11_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_14_13
#define init_5_11_14_13
void FASTCALL call_5_11_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_14_14
#define init_5_11_14_14
void FASTCALL call_5_11_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_14_15
#define init_5_11_14_15
void FASTCALL call_5_11_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_15_0
#define init_5_11_15_0
void FASTCALL call_5_11_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_15_1
#define init_5_11_15_1
void FASTCALL call_5_11_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_15_2
#define init_5_11_15_2
void FASTCALL call_5_11_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_15_3
#define init_5_11_15_3
void FASTCALL call_5_11_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_15_4
#define init_5_11_15_4
void FASTCALL call_5_11_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_15_5
#define init_5_11_15_5
void FASTCALL call_5_11_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_15_6
#define init_5_11_15_6
void FASTCALL call_5_11_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_15_7
#define init_5_11_15_7
void FASTCALL call_5_11_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_15_8
#define init_5_11_15_8
void FASTCALL call_5_11_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_15_9
#define init_5_11_15_9
void FASTCALL call_5_11_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_15_10
#define init_5_11_15_10
void FASTCALL call_5_11_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_15_11
#define init_5_11_15_11
void FASTCALL call_5_11_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_15_12
#define init_5_11_15_12
void FASTCALL call_5_11_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_15_13
#define init_5_11_15_13
void FASTCALL call_5_11_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_15_14
#define init_5_11_15_14
void FASTCALL call_5_11_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_11_15_15
#define init_5_11_15_15
void FASTCALL call_5_11_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_0_0
#define init_5_12_0_0
void FASTCALL call_5_12_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_0_1
#define init_5_12_0_1
void FASTCALL call_5_12_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_0_2
#define init_5_12_0_2
void FASTCALL call_5_12_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_0_3
#define init_5_12_0_3
void FASTCALL call_5_12_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_0_4
#define init_5_12_0_4
void FASTCALL call_5_12_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_0_5
#define init_5_12_0_5
void FASTCALL call_5_12_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_0_6
#define init_5_12_0_6
void FASTCALL call_5_12_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_0_7
#define init_5_12_0_7
void FASTCALL call_5_12_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_0_8
#define init_5_12_0_8
void FASTCALL call_5_12_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_0_9
#define init_5_12_0_9
void FASTCALL call_5_12_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_0_10
#define init_5_12_0_10
void FASTCALL call_5_12_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_0_11
#define init_5_12_0_11
void FASTCALL call_5_12_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_0_12
#define init_5_12_0_12
void FASTCALL call_5_12_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_0_13
#define init_5_12_0_13
void FASTCALL call_5_12_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_0_14
#define init_5_12_0_14
void FASTCALL call_5_12_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_0_15
#define init_5_12_0_15
void FASTCALL call_5_12_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_1_0
#define init_5_12_1_0
void FASTCALL call_5_12_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_1_1
#define init_5_12_1_1
void FASTCALL call_5_12_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_1_2
#define init_5_12_1_2
void FASTCALL call_5_12_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_1_3
#define init_5_12_1_3
void FASTCALL call_5_12_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_1_4
#define init_5_12_1_4
void FASTCALL call_5_12_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_1_5
#define init_5_12_1_5
void FASTCALL call_5_12_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_1_6
#define init_5_12_1_6
void FASTCALL call_5_12_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_1_7
#define init_5_12_1_7
void FASTCALL call_5_12_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_1_8
#define init_5_12_1_8
void FASTCALL call_5_12_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_1_9
#define init_5_12_1_9
void FASTCALL call_5_12_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_1_10
#define init_5_12_1_10
void FASTCALL call_5_12_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_1_11
#define init_5_12_1_11
void FASTCALL call_5_12_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_1_12
#define init_5_12_1_12
void FASTCALL call_5_12_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_1_13
#define init_5_12_1_13
void FASTCALL call_5_12_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_1_14
#define init_5_12_1_14
void FASTCALL call_5_12_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_1_15
#define init_5_12_1_15
void FASTCALL call_5_12_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_2_0
#define init_5_12_2_0
void FASTCALL call_5_12_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_2_1
#define init_5_12_2_1
void FASTCALL call_5_12_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_2_2
#define init_5_12_2_2
void FASTCALL call_5_12_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_2_3
#define init_5_12_2_3
void FASTCALL call_5_12_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_2_4
#define init_5_12_2_4
void FASTCALL call_5_12_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_2_5
#define init_5_12_2_5
void FASTCALL call_5_12_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_2_6
#define init_5_12_2_6
void FASTCALL call_5_12_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_2_7
#define init_5_12_2_7
void FASTCALL call_5_12_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_2_8
#define init_5_12_2_8
void FASTCALL call_5_12_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_2_9
#define init_5_12_2_9
void FASTCALL call_5_12_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_2_10
#define init_5_12_2_10
void FASTCALL call_5_12_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_2_11
#define init_5_12_2_11
void FASTCALL call_5_12_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_2_12
#define init_5_12_2_12
void FASTCALL call_5_12_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_2_13
#define init_5_12_2_13
void FASTCALL call_5_12_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_2_14
#define init_5_12_2_14
void FASTCALL call_5_12_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_2_15
#define init_5_12_2_15
void FASTCALL call_5_12_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_3_0
#define init_5_12_3_0
void FASTCALL call_5_12_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_3_1
#define init_5_12_3_1
void FASTCALL call_5_12_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_3_2
#define init_5_12_3_2
void FASTCALL call_5_12_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_3_3
#define init_5_12_3_3
void FASTCALL call_5_12_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_3_4
#define init_5_12_3_4
void FASTCALL call_5_12_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_3_5
#define init_5_12_3_5
void FASTCALL call_5_12_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_3_6
#define init_5_12_3_6
void FASTCALL call_5_12_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_3_7
#define init_5_12_3_7
void FASTCALL call_5_12_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_3_8
#define init_5_12_3_8
void FASTCALL call_5_12_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_3_9
#define init_5_12_3_9
void FASTCALL call_5_12_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_3_10
#define init_5_12_3_10
void FASTCALL call_5_12_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_3_11
#define init_5_12_3_11
void FASTCALL call_5_12_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_3_12
#define init_5_12_3_12
void FASTCALL call_5_12_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_3_13
#define init_5_12_3_13
void FASTCALL call_5_12_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_3_14
#define init_5_12_3_14
void FASTCALL call_5_12_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_3_15
#define init_5_12_3_15
void FASTCALL call_5_12_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_4_0
#define init_5_12_4_0
void FASTCALL call_5_12_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_4_1
#define init_5_12_4_1
void FASTCALL call_5_12_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_4_2
#define init_5_12_4_2
void FASTCALL call_5_12_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_4_3
#define init_5_12_4_3
void FASTCALL call_5_12_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_4_4
#define init_5_12_4_4
void FASTCALL call_5_12_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_4_5
#define init_5_12_4_5
void FASTCALL call_5_12_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_4_6
#define init_5_12_4_6
void FASTCALL call_5_12_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_4_7
#define init_5_12_4_7
void FASTCALL call_5_12_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_4_8
#define init_5_12_4_8
void FASTCALL call_5_12_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_4_9
#define init_5_12_4_9
void FASTCALL call_5_12_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_4_10
#define init_5_12_4_10
void FASTCALL call_5_12_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_4_11
#define init_5_12_4_11
void FASTCALL call_5_12_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_4_12
#define init_5_12_4_12
void FASTCALL call_5_12_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_4_13
#define init_5_12_4_13
void FASTCALL call_5_12_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_4_14
#define init_5_12_4_14
void FASTCALL call_5_12_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_4_15
#define init_5_12_4_15
void FASTCALL call_5_12_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_5_0
#define init_5_12_5_0
void FASTCALL call_5_12_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_5_1
#define init_5_12_5_1
void FASTCALL call_5_12_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_5_2
#define init_5_12_5_2
void FASTCALL call_5_12_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_5_3
#define init_5_12_5_3
void FASTCALL call_5_12_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_5_4
#define init_5_12_5_4
void FASTCALL call_5_12_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_5_5
#define init_5_12_5_5
void FASTCALL call_5_12_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_5_6
#define init_5_12_5_6
void FASTCALL call_5_12_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_5_7
#define init_5_12_5_7
void FASTCALL call_5_12_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_5_8
#define init_5_12_5_8
void FASTCALL call_5_12_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_5_9
#define init_5_12_5_9
void FASTCALL call_5_12_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_5_10
#define init_5_12_5_10
void FASTCALL call_5_12_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_5_11
#define init_5_12_5_11
void FASTCALL call_5_12_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_5_12
#define init_5_12_5_12
void FASTCALL call_5_12_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_5_13
#define init_5_12_5_13
void FASTCALL call_5_12_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_5_14
#define init_5_12_5_14
void FASTCALL call_5_12_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_5_15
#define init_5_12_5_15
void FASTCALL call_5_12_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_6_0
#define init_5_12_6_0
void FASTCALL call_5_12_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_6_1
#define init_5_12_6_1
void FASTCALL call_5_12_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_6_2
#define init_5_12_6_2
void FASTCALL call_5_12_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_6_3
#define init_5_12_6_3
void FASTCALL call_5_12_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_6_4
#define init_5_12_6_4
void FASTCALL call_5_12_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_6_5
#define init_5_12_6_5
void FASTCALL call_5_12_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_6_6
#define init_5_12_6_6
void FASTCALL call_5_12_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_6_7
#define init_5_12_6_7
void FASTCALL call_5_12_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_6_8
#define init_5_12_6_8
void FASTCALL call_5_12_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_6_9
#define init_5_12_6_9
void FASTCALL call_5_12_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_6_10
#define init_5_12_6_10
void FASTCALL call_5_12_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_6_11
#define init_5_12_6_11
void FASTCALL call_5_12_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_6_12
#define init_5_12_6_12
void FASTCALL call_5_12_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_6_13
#define init_5_12_6_13
void FASTCALL call_5_12_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_6_14
#define init_5_12_6_14
void FASTCALL call_5_12_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_6_15
#define init_5_12_6_15
void FASTCALL call_5_12_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_7_0
#define init_5_12_7_0
void FASTCALL call_5_12_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_7_1
#define init_5_12_7_1
void FASTCALL call_5_12_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_7_2
#define init_5_12_7_2
void FASTCALL call_5_12_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_7_3
#define init_5_12_7_3
void FASTCALL call_5_12_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_7_4
#define init_5_12_7_4
void FASTCALL call_5_12_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_7_5
#define init_5_12_7_5
void FASTCALL call_5_12_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_7_6
#define init_5_12_7_6
void FASTCALL call_5_12_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_7_7
#define init_5_12_7_7
void FASTCALL call_5_12_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_7_8
#define init_5_12_7_8
void FASTCALL call_5_12_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_7_9
#define init_5_12_7_9
void FASTCALL call_5_12_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_7_10
#define init_5_12_7_10
void FASTCALL call_5_12_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_7_11
#define init_5_12_7_11
void FASTCALL call_5_12_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_7_12
#define init_5_12_7_12
void FASTCALL call_5_12_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_7_13
#define init_5_12_7_13
void FASTCALL call_5_12_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_7_14
#define init_5_12_7_14
void FASTCALL call_5_12_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_7_15
#define init_5_12_7_15
void FASTCALL call_5_12_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_8_0
#define init_5_12_8_0
void FASTCALL call_5_12_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_8_1
#define init_5_12_8_1
void FASTCALL call_5_12_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_8_2
#define init_5_12_8_2
void FASTCALL call_5_12_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_8_3
#define init_5_12_8_3
void FASTCALL call_5_12_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_8_4
#define init_5_12_8_4
void FASTCALL call_5_12_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_8_5
#define init_5_12_8_5
void FASTCALL call_5_12_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_8_6
#define init_5_12_8_6
void FASTCALL call_5_12_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_8_7
#define init_5_12_8_7
void FASTCALL call_5_12_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_8_8
#define init_5_12_8_8
void FASTCALL call_5_12_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_8_9
#define init_5_12_8_9
void FASTCALL call_5_12_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_8_10
#define init_5_12_8_10
void FASTCALL call_5_12_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_8_11
#define init_5_12_8_11
void FASTCALL call_5_12_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_8_12
#define init_5_12_8_12
void FASTCALL call_5_12_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_8_13
#define init_5_12_8_13
void FASTCALL call_5_12_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_8_14
#define init_5_12_8_14
void FASTCALL call_5_12_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_8_15
#define init_5_12_8_15
void FASTCALL call_5_12_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_9_0
#define init_5_12_9_0
void FASTCALL call_5_12_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_9_1
#define init_5_12_9_1
void FASTCALL call_5_12_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_9_2
#define init_5_12_9_2
void FASTCALL call_5_12_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_9_3
#define init_5_12_9_3
void FASTCALL call_5_12_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_9_4
#define init_5_12_9_4
void FASTCALL call_5_12_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_9_5
#define init_5_12_9_5
void FASTCALL call_5_12_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_9_6
#define init_5_12_9_6
void FASTCALL call_5_12_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_9_7
#define init_5_12_9_7
void FASTCALL call_5_12_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_9_8
#define init_5_12_9_8
void FASTCALL call_5_12_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_9_9
#define init_5_12_9_9
void FASTCALL call_5_12_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_9_10
#define init_5_12_9_10
void FASTCALL call_5_12_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_9_11
#define init_5_12_9_11
void FASTCALL call_5_12_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_9_12
#define init_5_12_9_12
void FASTCALL call_5_12_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_9_13
#define init_5_12_9_13
void FASTCALL call_5_12_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_9_14
#define init_5_12_9_14
void FASTCALL call_5_12_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_9_15
#define init_5_12_9_15
void FASTCALL call_5_12_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_10_0
#define init_5_12_10_0
void FASTCALL call_5_12_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_10_1
#define init_5_12_10_1
void FASTCALL call_5_12_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_10_2
#define init_5_12_10_2
void FASTCALL call_5_12_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_10_3
#define init_5_12_10_3
void FASTCALL call_5_12_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_10_4
#define init_5_12_10_4
void FASTCALL call_5_12_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_10_5
#define init_5_12_10_5
void FASTCALL call_5_12_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_10_6
#define init_5_12_10_6
void FASTCALL call_5_12_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_10_7
#define init_5_12_10_7
void FASTCALL call_5_12_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_10_8
#define init_5_12_10_8
void FASTCALL call_5_12_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_10_9
#define init_5_12_10_9
void FASTCALL call_5_12_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_10_10
#define init_5_12_10_10
void FASTCALL call_5_12_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_10_11
#define init_5_12_10_11
void FASTCALL call_5_12_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_10_12
#define init_5_12_10_12
void FASTCALL call_5_12_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_10_13
#define init_5_12_10_13
void FASTCALL call_5_12_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_10_14
#define init_5_12_10_14
void FASTCALL call_5_12_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_10_15
#define init_5_12_10_15
void FASTCALL call_5_12_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_11_0
#define init_5_12_11_0
void FASTCALL call_5_12_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_11_1
#define init_5_12_11_1
void FASTCALL call_5_12_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_11_2
#define init_5_12_11_2
void FASTCALL call_5_12_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_11_3
#define init_5_12_11_3
void FASTCALL call_5_12_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_11_4
#define init_5_12_11_4
void FASTCALL call_5_12_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_11_5
#define init_5_12_11_5
void FASTCALL call_5_12_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_11_6
#define init_5_12_11_6
void FASTCALL call_5_12_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_11_7
#define init_5_12_11_7
void FASTCALL call_5_12_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_11_8
#define init_5_12_11_8
void FASTCALL call_5_12_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_11_9
#define init_5_12_11_9
void FASTCALL call_5_12_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_11_10
#define init_5_12_11_10
void FASTCALL call_5_12_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_11_11
#define init_5_12_11_11
void FASTCALL call_5_12_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_11_12
#define init_5_12_11_12
void FASTCALL call_5_12_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_11_13
#define init_5_12_11_13
void FASTCALL call_5_12_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_11_14
#define init_5_12_11_14
void FASTCALL call_5_12_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_11_15
#define init_5_12_11_15
void FASTCALL call_5_12_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_12_0
#define init_5_12_12_0
void FASTCALL call_5_12_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_12_1
#define init_5_12_12_1
void FASTCALL call_5_12_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_12_2
#define init_5_12_12_2
void FASTCALL call_5_12_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_12_3
#define init_5_12_12_3
void FASTCALL call_5_12_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_12_4
#define init_5_12_12_4
void FASTCALL call_5_12_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_12_5
#define init_5_12_12_5
void FASTCALL call_5_12_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_12_6
#define init_5_12_12_6
void FASTCALL call_5_12_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_12_7
#define init_5_12_12_7
void FASTCALL call_5_12_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_12_8
#define init_5_12_12_8
void FASTCALL call_5_12_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_12_9
#define init_5_12_12_9
void FASTCALL call_5_12_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_12_10
#define init_5_12_12_10
void FASTCALL call_5_12_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_12_11
#define init_5_12_12_11
void FASTCALL call_5_12_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_12_12
#define init_5_12_12_12
void FASTCALL call_5_12_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_12_13
#define init_5_12_12_13
void FASTCALL call_5_12_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_12_14
#define init_5_12_12_14
void FASTCALL call_5_12_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_12_15
#define init_5_12_12_15
void FASTCALL call_5_12_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_13_0
#define init_5_12_13_0
void FASTCALL call_5_12_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_13_1
#define init_5_12_13_1
void FASTCALL call_5_12_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_13_2
#define init_5_12_13_2
void FASTCALL call_5_12_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_13_3
#define init_5_12_13_3
void FASTCALL call_5_12_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_13_4
#define init_5_12_13_4
void FASTCALL call_5_12_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_13_5
#define init_5_12_13_5
void FASTCALL call_5_12_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_13_6
#define init_5_12_13_6
void FASTCALL call_5_12_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_13_7
#define init_5_12_13_7
void FASTCALL call_5_12_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_13_8
#define init_5_12_13_8
void FASTCALL call_5_12_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_13_9
#define init_5_12_13_9
void FASTCALL call_5_12_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_13_10
#define init_5_12_13_10
void FASTCALL call_5_12_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_13_11
#define init_5_12_13_11
void FASTCALL call_5_12_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_13_12
#define init_5_12_13_12
void FASTCALL call_5_12_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_13_13
#define init_5_12_13_13
void FASTCALL call_5_12_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_13_14
#define init_5_12_13_14
void FASTCALL call_5_12_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_13_15
#define init_5_12_13_15
void FASTCALL call_5_12_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_14_0
#define init_5_12_14_0
void FASTCALL call_5_12_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_14_1
#define init_5_12_14_1
void FASTCALL call_5_12_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_14_2
#define init_5_12_14_2
void FASTCALL call_5_12_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_14_3
#define init_5_12_14_3
void FASTCALL call_5_12_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_14_4
#define init_5_12_14_4
void FASTCALL call_5_12_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_14_5
#define init_5_12_14_5
void FASTCALL call_5_12_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_14_6
#define init_5_12_14_6
void FASTCALL call_5_12_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_14_7
#define init_5_12_14_7
void FASTCALL call_5_12_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_14_8
#define init_5_12_14_8
void FASTCALL call_5_12_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_14_9
#define init_5_12_14_9
void FASTCALL call_5_12_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_14_10
#define init_5_12_14_10
void FASTCALL call_5_12_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_14_11
#define init_5_12_14_11
void FASTCALL call_5_12_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_14_12
#define init_5_12_14_12
void FASTCALL call_5_12_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_14_13
#define init_5_12_14_13
void FASTCALL call_5_12_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_14_14
#define init_5_12_14_14
void FASTCALL call_5_12_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_14_15
#define init_5_12_14_15
void FASTCALL call_5_12_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_15_0
#define init_5_12_15_0
void FASTCALL call_5_12_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_15_1
#define init_5_12_15_1
void FASTCALL call_5_12_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_15_2
#define init_5_12_15_2
void FASTCALL call_5_12_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_15_3
#define init_5_12_15_3
void FASTCALL call_5_12_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_15_4
#define init_5_12_15_4
void FASTCALL call_5_12_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_15_5
#define init_5_12_15_5
void FASTCALL call_5_12_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_15_6
#define init_5_12_15_6
void FASTCALL call_5_12_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_15_7
#define init_5_12_15_7
void FASTCALL call_5_12_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_15_8
#define init_5_12_15_8
void FASTCALL call_5_12_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_15_9
#define init_5_12_15_9
void FASTCALL call_5_12_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_15_10
#define init_5_12_15_10
void FASTCALL call_5_12_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_15_11
#define init_5_12_15_11
void FASTCALL call_5_12_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_15_12
#define init_5_12_15_12
void FASTCALL call_5_12_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_15_13
#define init_5_12_15_13
void FASTCALL call_5_12_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_15_14
#define init_5_12_15_14
void FASTCALL call_5_12_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_12_15_15
#define init_5_12_15_15
void FASTCALL call_5_12_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_0_0
#define init_5_13_0_0
void FASTCALL call_5_13_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_0_1
#define init_5_13_0_1
void FASTCALL call_5_13_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_0_2
#define init_5_13_0_2
void FASTCALL call_5_13_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_0_3
#define init_5_13_0_3
void FASTCALL call_5_13_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_0_4
#define init_5_13_0_4
void FASTCALL call_5_13_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_0_5
#define init_5_13_0_5
void FASTCALL call_5_13_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_0_6
#define init_5_13_0_6
void FASTCALL call_5_13_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_0_7
#define init_5_13_0_7
void FASTCALL call_5_13_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_0_8
#define init_5_13_0_8
void FASTCALL call_5_13_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_0_9
#define init_5_13_0_9
void FASTCALL call_5_13_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_0_10
#define init_5_13_0_10
void FASTCALL call_5_13_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_0_11
#define init_5_13_0_11
void FASTCALL call_5_13_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_0_12
#define init_5_13_0_12
void FASTCALL call_5_13_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_0_13
#define init_5_13_0_13
void FASTCALL call_5_13_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_0_14
#define init_5_13_0_14
void FASTCALL call_5_13_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_0_15
#define init_5_13_0_15
void FASTCALL call_5_13_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_1_0
#define init_5_13_1_0
void FASTCALL call_5_13_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_1_1
#define init_5_13_1_1
void FASTCALL call_5_13_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_1_2
#define init_5_13_1_2
void FASTCALL call_5_13_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_1_3
#define init_5_13_1_3
void FASTCALL call_5_13_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_1_4
#define init_5_13_1_4
void FASTCALL call_5_13_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_1_5
#define init_5_13_1_5
void FASTCALL call_5_13_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_1_6
#define init_5_13_1_6
void FASTCALL call_5_13_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_1_7
#define init_5_13_1_7
void FASTCALL call_5_13_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_1_8
#define init_5_13_1_8
void FASTCALL call_5_13_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_1_9
#define init_5_13_1_9
void FASTCALL call_5_13_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_1_10
#define init_5_13_1_10
void FASTCALL call_5_13_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_1_11
#define init_5_13_1_11
void FASTCALL call_5_13_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_1_12
#define init_5_13_1_12
void FASTCALL call_5_13_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_1_13
#define init_5_13_1_13
void FASTCALL call_5_13_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_1_14
#define init_5_13_1_14
void FASTCALL call_5_13_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_1_15
#define init_5_13_1_15
void FASTCALL call_5_13_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_2_0
#define init_5_13_2_0
void FASTCALL call_5_13_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_2_1
#define init_5_13_2_1
void FASTCALL call_5_13_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_2_2
#define init_5_13_2_2
void FASTCALL call_5_13_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_2_3
#define init_5_13_2_3
void FASTCALL call_5_13_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_2_4
#define init_5_13_2_4
void FASTCALL call_5_13_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_2_5
#define init_5_13_2_5
void FASTCALL call_5_13_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_2_6
#define init_5_13_2_6
void FASTCALL call_5_13_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_2_7
#define init_5_13_2_7
void FASTCALL call_5_13_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_2_8
#define init_5_13_2_8
void FASTCALL call_5_13_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_2_9
#define init_5_13_2_9
void FASTCALL call_5_13_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_2_10
#define init_5_13_2_10
void FASTCALL call_5_13_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_2_11
#define init_5_13_2_11
void FASTCALL call_5_13_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_2_12
#define init_5_13_2_12
void FASTCALL call_5_13_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_2_13
#define init_5_13_2_13
void FASTCALL call_5_13_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_2_14
#define init_5_13_2_14
void FASTCALL call_5_13_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_2_15
#define init_5_13_2_15
void FASTCALL call_5_13_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_3_0
#define init_5_13_3_0
void FASTCALL call_5_13_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_3_1
#define init_5_13_3_1
void FASTCALL call_5_13_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_3_2
#define init_5_13_3_2
void FASTCALL call_5_13_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_3_3
#define init_5_13_3_3
void FASTCALL call_5_13_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_3_4
#define init_5_13_3_4
void FASTCALL call_5_13_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_3_5
#define init_5_13_3_5
void FASTCALL call_5_13_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_3_6
#define init_5_13_3_6
void FASTCALL call_5_13_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_3_7
#define init_5_13_3_7
void FASTCALL call_5_13_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_3_8
#define init_5_13_3_8
void FASTCALL call_5_13_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_3_9
#define init_5_13_3_9
void FASTCALL call_5_13_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_3_10
#define init_5_13_3_10
void FASTCALL call_5_13_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_3_11
#define init_5_13_3_11
void FASTCALL call_5_13_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_3_12
#define init_5_13_3_12
void FASTCALL call_5_13_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_3_13
#define init_5_13_3_13
void FASTCALL call_5_13_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_3_14
#define init_5_13_3_14
void FASTCALL call_5_13_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_3_15
#define init_5_13_3_15
void FASTCALL call_5_13_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_4_0
#define init_5_13_4_0
void FASTCALL call_5_13_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_4_1
#define init_5_13_4_1
void FASTCALL call_5_13_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_4_2
#define init_5_13_4_2
void FASTCALL call_5_13_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_4_3
#define init_5_13_4_3
void FASTCALL call_5_13_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_4_4
#define init_5_13_4_4
void FASTCALL call_5_13_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_4_5
#define init_5_13_4_5
void FASTCALL call_5_13_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_4_6
#define init_5_13_4_6
void FASTCALL call_5_13_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_4_7
#define init_5_13_4_7
void FASTCALL call_5_13_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_4_8
#define init_5_13_4_8
void FASTCALL call_5_13_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_4_9
#define init_5_13_4_9
void FASTCALL call_5_13_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_4_10
#define init_5_13_4_10
void FASTCALL call_5_13_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_4_11
#define init_5_13_4_11
void FASTCALL call_5_13_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_4_12
#define init_5_13_4_12
void FASTCALL call_5_13_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_4_13
#define init_5_13_4_13
void FASTCALL call_5_13_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_4_14
#define init_5_13_4_14
void FASTCALL call_5_13_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_4_15
#define init_5_13_4_15
void FASTCALL call_5_13_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_5_0
#define init_5_13_5_0
void FASTCALL call_5_13_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_5_1
#define init_5_13_5_1
void FASTCALL call_5_13_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_5_2
#define init_5_13_5_2
void FASTCALL call_5_13_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_5_3
#define init_5_13_5_3
void FASTCALL call_5_13_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_5_4
#define init_5_13_5_4
void FASTCALL call_5_13_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_5_5
#define init_5_13_5_5
void FASTCALL call_5_13_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_5_6
#define init_5_13_5_6
void FASTCALL call_5_13_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_5_7
#define init_5_13_5_7
void FASTCALL call_5_13_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_5_8
#define init_5_13_5_8
void FASTCALL call_5_13_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_5_9
#define init_5_13_5_9
void FASTCALL call_5_13_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_5_10
#define init_5_13_5_10
void FASTCALL call_5_13_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_5_11
#define init_5_13_5_11
void FASTCALL call_5_13_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_5_12
#define init_5_13_5_12
void FASTCALL call_5_13_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_5_13
#define init_5_13_5_13
void FASTCALL call_5_13_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_5_14
#define init_5_13_5_14
void FASTCALL call_5_13_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_5_15
#define init_5_13_5_15
void FASTCALL call_5_13_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_6_0
#define init_5_13_6_0
void FASTCALL call_5_13_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_6_1
#define init_5_13_6_1
void FASTCALL call_5_13_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_6_2
#define init_5_13_6_2
void FASTCALL call_5_13_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_6_3
#define init_5_13_6_3
void FASTCALL call_5_13_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_6_4
#define init_5_13_6_4
void FASTCALL call_5_13_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_6_5
#define init_5_13_6_5
void FASTCALL call_5_13_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_6_6
#define init_5_13_6_6
void FASTCALL call_5_13_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_6_7
#define init_5_13_6_7
void FASTCALL call_5_13_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_6_8
#define init_5_13_6_8
void FASTCALL call_5_13_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_6_9
#define init_5_13_6_9
void FASTCALL call_5_13_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_6_10
#define init_5_13_6_10
void FASTCALL call_5_13_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_6_11
#define init_5_13_6_11
void FASTCALL call_5_13_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_6_12
#define init_5_13_6_12
void FASTCALL call_5_13_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_6_13
#define init_5_13_6_13
void FASTCALL call_5_13_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_6_14
#define init_5_13_6_14
void FASTCALL call_5_13_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_6_15
#define init_5_13_6_15
void FASTCALL call_5_13_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_7_0
#define init_5_13_7_0
void FASTCALL call_5_13_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_7_1
#define init_5_13_7_1
void FASTCALL call_5_13_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_7_2
#define init_5_13_7_2
void FASTCALL call_5_13_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_7_3
#define init_5_13_7_3
void FASTCALL call_5_13_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_7_4
#define init_5_13_7_4
void FASTCALL call_5_13_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_7_5
#define init_5_13_7_5
void FASTCALL call_5_13_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_7_6
#define init_5_13_7_6
void FASTCALL call_5_13_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_7_7
#define init_5_13_7_7
void FASTCALL call_5_13_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_7_8
#define init_5_13_7_8
void FASTCALL call_5_13_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_7_9
#define init_5_13_7_9
void FASTCALL call_5_13_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_7_10
#define init_5_13_7_10
void FASTCALL call_5_13_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_7_11
#define init_5_13_7_11
void FASTCALL call_5_13_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_7_12
#define init_5_13_7_12
void FASTCALL call_5_13_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_7_13
#define init_5_13_7_13
void FASTCALL call_5_13_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_7_14
#define init_5_13_7_14
void FASTCALL call_5_13_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_7_15
#define init_5_13_7_15
void FASTCALL call_5_13_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_8_0
#define init_5_13_8_0
void FASTCALL call_5_13_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_8_1
#define init_5_13_8_1
void FASTCALL call_5_13_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_8_2
#define init_5_13_8_2
void FASTCALL call_5_13_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_8_3
#define init_5_13_8_3
void FASTCALL call_5_13_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_8_4
#define init_5_13_8_4
void FASTCALL call_5_13_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_8_5
#define init_5_13_8_5
void FASTCALL call_5_13_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_8_6
#define init_5_13_8_6
void FASTCALL call_5_13_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_8_7
#define init_5_13_8_7
void FASTCALL call_5_13_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_8_8
#define init_5_13_8_8
void FASTCALL call_5_13_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_8_9
#define init_5_13_8_9
void FASTCALL call_5_13_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_8_10
#define init_5_13_8_10
void FASTCALL call_5_13_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_8_11
#define init_5_13_8_11
void FASTCALL call_5_13_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_8_12
#define init_5_13_8_12
void FASTCALL call_5_13_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_8_13
#define init_5_13_8_13
void FASTCALL call_5_13_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_8_14
#define init_5_13_8_14
void FASTCALL call_5_13_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_8_15
#define init_5_13_8_15
void FASTCALL call_5_13_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_9_0
#define init_5_13_9_0
void FASTCALL call_5_13_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_9_1
#define init_5_13_9_1
void FASTCALL call_5_13_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_9_2
#define init_5_13_9_2
void FASTCALL call_5_13_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_9_3
#define init_5_13_9_3
void FASTCALL call_5_13_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_9_4
#define init_5_13_9_4
void FASTCALL call_5_13_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_9_5
#define init_5_13_9_5
void FASTCALL call_5_13_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_9_6
#define init_5_13_9_6
void FASTCALL call_5_13_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_9_7
#define init_5_13_9_7
void FASTCALL call_5_13_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_9_8
#define init_5_13_9_8
void FASTCALL call_5_13_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_9_9
#define init_5_13_9_9
void FASTCALL call_5_13_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_9_10
#define init_5_13_9_10
void FASTCALL call_5_13_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_9_11
#define init_5_13_9_11
void FASTCALL call_5_13_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_9_12
#define init_5_13_9_12
void FASTCALL call_5_13_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_9_13
#define init_5_13_9_13
void FASTCALL call_5_13_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_9_14
#define init_5_13_9_14
void FASTCALL call_5_13_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_9_15
#define init_5_13_9_15
void FASTCALL call_5_13_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_10_0
#define init_5_13_10_0
void FASTCALL call_5_13_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_10_1
#define init_5_13_10_1
void FASTCALL call_5_13_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_10_2
#define init_5_13_10_2
void FASTCALL call_5_13_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_10_3
#define init_5_13_10_3
void FASTCALL call_5_13_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_10_4
#define init_5_13_10_4
void FASTCALL call_5_13_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_10_5
#define init_5_13_10_5
void FASTCALL call_5_13_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_10_6
#define init_5_13_10_6
void FASTCALL call_5_13_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_10_7
#define init_5_13_10_7
void FASTCALL call_5_13_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_10_8
#define init_5_13_10_8
void FASTCALL call_5_13_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_10_9
#define init_5_13_10_9
void FASTCALL call_5_13_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_10_10
#define init_5_13_10_10
void FASTCALL call_5_13_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_10_11
#define init_5_13_10_11
void FASTCALL call_5_13_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_10_12
#define init_5_13_10_12
void FASTCALL call_5_13_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_10_13
#define init_5_13_10_13
void FASTCALL call_5_13_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_10_14
#define init_5_13_10_14
void FASTCALL call_5_13_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_10_15
#define init_5_13_10_15
void FASTCALL call_5_13_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_11_0
#define init_5_13_11_0
void FASTCALL call_5_13_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_11_1
#define init_5_13_11_1
void FASTCALL call_5_13_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_11_2
#define init_5_13_11_2
void FASTCALL call_5_13_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_11_3
#define init_5_13_11_3
void FASTCALL call_5_13_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_11_4
#define init_5_13_11_4
void FASTCALL call_5_13_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_11_5
#define init_5_13_11_5
void FASTCALL call_5_13_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_11_6
#define init_5_13_11_6
void FASTCALL call_5_13_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_11_7
#define init_5_13_11_7
void FASTCALL call_5_13_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_11_8
#define init_5_13_11_8
void FASTCALL call_5_13_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_11_9
#define init_5_13_11_9
void FASTCALL call_5_13_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_11_10
#define init_5_13_11_10
void FASTCALL call_5_13_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_11_11
#define init_5_13_11_11
void FASTCALL call_5_13_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_11_12
#define init_5_13_11_12
void FASTCALL call_5_13_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_11_13
#define init_5_13_11_13
void FASTCALL call_5_13_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_11_14
#define init_5_13_11_14
void FASTCALL call_5_13_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_11_15
#define init_5_13_11_15
void FASTCALL call_5_13_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_12_0
#define init_5_13_12_0
void FASTCALL call_5_13_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_12_1
#define init_5_13_12_1
void FASTCALL call_5_13_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_12_2
#define init_5_13_12_2
void FASTCALL call_5_13_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_12_3
#define init_5_13_12_3
void FASTCALL call_5_13_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_12_4
#define init_5_13_12_4
void FASTCALL call_5_13_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_12_5
#define init_5_13_12_5
void FASTCALL call_5_13_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_12_6
#define init_5_13_12_6
void FASTCALL call_5_13_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_12_7
#define init_5_13_12_7
void FASTCALL call_5_13_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_12_8
#define init_5_13_12_8
void FASTCALL call_5_13_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_12_9
#define init_5_13_12_9
void FASTCALL call_5_13_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_12_10
#define init_5_13_12_10
void FASTCALL call_5_13_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_12_11
#define init_5_13_12_11
void FASTCALL call_5_13_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_12_12
#define init_5_13_12_12
void FASTCALL call_5_13_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_12_13
#define init_5_13_12_13
void FASTCALL call_5_13_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_12_14
#define init_5_13_12_14
void FASTCALL call_5_13_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_12_15
#define init_5_13_12_15
void FASTCALL call_5_13_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_13_0
#define init_5_13_13_0
void FASTCALL call_5_13_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_13_1
#define init_5_13_13_1
void FASTCALL call_5_13_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_13_2
#define init_5_13_13_2
void FASTCALL call_5_13_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_13_3
#define init_5_13_13_3
void FASTCALL call_5_13_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_13_4
#define init_5_13_13_4
void FASTCALL call_5_13_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_13_5
#define init_5_13_13_5
void FASTCALL call_5_13_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_13_6
#define init_5_13_13_6
void FASTCALL call_5_13_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_13_7
#define init_5_13_13_7
void FASTCALL call_5_13_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_13_8
#define init_5_13_13_8
void FASTCALL call_5_13_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_13_9
#define init_5_13_13_9
void FASTCALL call_5_13_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_13_10
#define init_5_13_13_10
void FASTCALL call_5_13_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_13_11
#define init_5_13_13_11
void FASTCALL call_5_13_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_13_12
#define init_5_13_13_12
void FASTCALL call_5_13_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_13_13
#define init_5_13_13_13
void FASTCALL call_5_13_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_13_14
#define init_5_13_13_14
void FASTCALL call_5_13_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_13_15
#define init_5_13_13_15
void FASTCALL call_5_13_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_14_0
#define init_5_13_14_0
void FASTCALL call_5_13_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_14_1
#define init_5_13_14_1
void FASTCALL call_5_13_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_14_2
#define init_5_13_14_2
void FASTCALL call_5_13_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_14_3
#define init_5_13_14_3
void FASTCALL call_5_13_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_14_4
#define init_5_13_14_4
void FASTCALL call_5_13_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_14_5
#define init_5_13_14_5
void FASTCALL call_5_13_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_14_6
#define init_5_13_14_6
void FASTCALL call_5_13_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_14_7
#define init_5_13_14_7
void FASTCALL call_5_13_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_14_8
#define init_5_13_14_8
void FASTCALL call_5_13_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_14_9
#define init_5_13_14_9
void FASTCALL call_5_13_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_14_10
#define init_5_13_14_10
void FASTCALL call_5_13_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_14_11
#define init_5_13_14_11
void FASTCALL call_5_13_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_14_12
#define init_5_13_14_12
void FASTCALL call_5_13_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_14_13
#define init_5_13_14_13
void FASTCALL call_5_13_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_14_14
#define init_5_13_14_14
void FASTCALL call_5_13_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_14_15
#define init_5_13_14_15
void FASTCALL call_5_13_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_15_0
#define init_5_13_15_0
void FASTCALL call_5_13_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_15_1
#define init_5_13_15_1
void FASTCALL call_5_13_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_15_2
#define init_5_13_15_2
void FASTCALL call_5_13_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_15_3
#define init_5_13_15_3
void FASTCALL call_5_13_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_15_4
#define init_5_13_15_4
void FASTCALL call_5_13_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_15_5
#define init_5_13_15_5
void FASTCALL call_5_13_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_15_6
#define init_5_13_15_6
void FASTCALL call_5_13_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_15_7
#define init_5_13_15_7
void FASTCALL call_5_13_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_15_8
#define init_5_13_15_8
void FASTCALL call_5_13_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_15_9
#define init_5_13_15_9
void FASTCALL call_5_13_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_15_10
#define init_5_13_15_10
void FASTCALL call_5_13_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_15_11
#define init_5_13_15_11
void FASTCALL call_5_13_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_15_12
#define init_5_13_15_12
void FASTCALL call_5_13_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_15_13
#define init_5_13_15_13
void FASTCALL call_5_13_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_15_14
#define init_5_13_15_14
void FASTCALL call_5_13_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_13_15_15
#define init_5_13_15_15
void FASTCALL call_5_13_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_0_0
#define init_5_14_0_0
void FASTCALL call_5_14_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_0_1
#define init_5_14_0_1
void FASTCALL call_5_14_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_0_2
#define init_5_14_0_2
void FASTCALL call_5_14_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_0_3
#define init_5_14_0_3
void FASTCALL call_5_14_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_0_4
#define init_5_14_0_4
void FASTCALL call_5_14_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_0_5
#define init_5_14_0_5
void FASTCALL call_5_14_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_0_6
#define init_5_14_0_6
void FASTCALL call_5_14_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_0_7
#define init_5_14_0_7
void FASTCALL call_5_14_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_0_8
#define init_5_14_0_8
void FASTCALL call_5_14_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_0_9
#define init_5_14_0_9
void FASTCALL call_5_14_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_0_10
#define init_5_14_0_10
void FASTCALL call_5_14_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_0_11
#define init_5_14_0_11
void FASTCALL call_5_14_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_0_12
#define init_5_14_0_12
void FASTCALL call_5_14_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_0_13
#define init_5_14_0_13
void FASTCALL call_5_14_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_0_14
#define init_5_14_0_14
void FASTCALL call_5_14_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_0_15
#define init_5_14_0_15
void FASTCALL call_5_14_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_1_0
#define init_5_14_1_0
void FASTCALL call_5_14_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_1_1
#define init_5_14_1_1
void FASTCALL call_5_14_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_1_2
#define init_5_14_1_2
void FASTCALL call_5_14_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_1_3
#define init_5_14_1_3
void FASTCALL call_5_14_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_1_4
#define init_5_14_1_4
void FASTCALL call_5_14_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_1_5
#define init_5_14_1_5
void FASTCALL call_5_14_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_1_6
#define init_5_14_1_6
void FASTCALL call_5_14_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_1_7
#define init_5_14_1_7
void FASTCALL call_5_14_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_1_8
#define init_5_14_1_8
void FASTCALL call_5_14_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_1_9
#define init_5_14_1_9
void FASTCALL call_5_14_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_1_10
#define init_5_14_1_10
void FASTCALL call_5_14_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_1_11
#define init_5_14_1_11
void FASTCALL call_5_14_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_1_12
#define init_5_14_1_12
void FASTCALL call_5_14_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_1_13
#define init_5_14_1_13
void FASTCALL call_5_14_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_1_14
#define init_5_14_1_14
void FASTCALL call_5_14_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_1_15
#define init_5_14_1_15
void FASTCALL call_5_14_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_2_0
#define init_5_14_2_0
void FASTCALL call_5_14_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_2_1
#define init_5_14_2_1
void FASTCALL call_5_14_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_2_2
#define init_5_14_2_2
void FASTCALL call_5_14_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_2_3
#define init_5_14_2_3
void FASTCALL call_5_14_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_2_4
#define init_5_14_2_4
void FASTCALL call_5_14_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_2_5
#define init_5_14_2_5
void FASTCALL call_5_14_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_2_6
#define init_5_14_2_6
void FASTCALL call_5_14_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_2_7
#define init_5_14_2_7
void FASTCALL call_5_14_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_2_8
#define init_5_14_2_8
void FASTCALL call_5_14_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_2_9
#define init_5_14_2_9
void FASTCALL call_5_14_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_2_10
#define init_5_14_2_10
void FASTCALL call_5_14_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_2_11
#define init_5_14_2_11
void FASTCALL call_5_14_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_2_12
#define init_5_14_2_12
void FASTCALL call_5_14_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_2_13
#define init_5_14_2_13
void FASTCALL call_5_14_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_2_14
#define init_5_14_2_14
void FASTCALL call_5_14_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_2_15
#define init_5_14_2_15
void FASTCALL call_5_14_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_3_0
#define init_5_14_3_0
void FASTCALL call_5_14_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_3_1
#define init_5_14_3_1
void FASTCALL call_5_14_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_3_2
#define init_5_14_3_2
void FASTCALL call_5_14_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_3_3
#define init_5_14_3_3
void FASTCALL call_5_14_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_3_4
#define init_5_14_3_4
void FASTCALL call_5_14_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_3_5
#define init_5_14_3_5
void FASTCALL call_5_14_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_3_6
#define init_5_14_3_6
void FASTCALL call_5_14_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_3_7
#define init_5_14_3_7
void FASTCALL call_5_14_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_3_8
#define init_5_14_3_8
void FASTCALL call_5_14_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_3_9
#define init_5_14_3_9
void FASTCALL call_5_14_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_3_10
#define init_5_14_3_10
void FASTCALL call_5_14_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_3_11
#define init_5_14_3_11
void FASTCALL call_5_14_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_3_12
#define init_5_14_3_12
void FASTCALL call_5_14_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_3_13
#define init_5_14_3_13
void FASTCALL call_5_14_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_3_14
#define init_5_14_3_14
void FASTCALL call_5_14_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_3_15
#define init_5_14_3_15
void FASTCALL call_5_14_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_4_0
#define init_5_14_4_0
void FASTCALL call_5_14_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_4_1
#define init_5_14_4_1
void FASTCALL call_5_14_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_4_2
#define init_5_14_4_2
void FASTCALL call_5_14_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_4_3
#define init_5_14_4_3
void FASTCALL call_5_14_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_4_4
#define init_5_14_4_4
void FASTCALL call_5_14_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_4_5
#define init_5_14_4_5
void FASTCALL call_5_14_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_4_6
#define init_5_14_4_6
void FASTCALL call_5_14_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_4_7
#define init_5_14_4_7
void FASTCALL call_5_14_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_4_8
#define init_5_14_4_8
void FASTCALL call_5_14_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_4_9
#define init_5_14_4_9
void FASTCALL call_5_14_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_4_10
#define init_5_14_4_10
void FASTCALL call_5_14_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_4_11
#define init_5_14_4_11
void FASTCALL call_5_14_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_4_12
#define init_5_14_4_12
void FASTCALL call_5_14_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_4_13
#define init_5_14_4_13
void FASTCALL call_5_14_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_4_14
#define init_5_14_4_14
void FASTCALL call_5_14_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_4_15
#define init_5_14_4_15
void FASTCALL call_5_14_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_5_0
#define init_5_14_5_0
void FASTCALL call_5_14_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_5_1
#define init_5_14_5_1
void FASTCALL call_5_14_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_5_2
#define init_5_14_5_2
void FASTCALL call_5_14_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_5_3
#define init_5_14_5_3
void FASTCALL call_5_14_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_5_4
#define init_5_14_5_4
void FASTCALL call_5_14_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_5_5
#define init_5_14_5_5
void FASTCALL call_5_14_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_5_6
#define init_5_14_5_6
void FASTCALL call_5_14_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_5_7
#define init_5_14_5_7
void FASTCALL call_5_14_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_5_8
#define init_5_14_5_8
void FASTCALL call_5_14_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_5_9
#define init_5_14_5_9
void FASTCALL call_5_14_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_5_10
#define init_5_14_5_10
void FASTCALL call_5_14_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_5_11
#define init_5_14_5_11
void FASTCALL call_5_14_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_5_12
#define init_5_14_5_12
void FASTCALL call_5_14_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_5_13
#define init_5_14_5_13
void FASTCALL call_5_14_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_5_14
#define init_5_14_5_14
void FASTCALL call_5_14_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_5_15
#define init_5_14_5_15
void FASTCALL call_5_14_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_6_0
#define init_5_14_6_0
void FASTCALL call_5_14_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_6_1
#define init_5_14_6_1
void FASTCALL call_5_14_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_6_2
#define init_5_14_6_2
void FASTCALL call_5_14_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_6_3
#define init_5_14_6_3
void FASTCALL call_5_14_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_6_4
#define init_5_14_6_4
void FASTCALL call_5_14_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_6_5
#define init_5_14_6_5
void FASTCALL call_5_14_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_6_6
#define init_5_14_6_6
void FASTCALL call_5_14_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_6_7
#define init_5_14_6_7
void FASTCALL call_5_14_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_6_8
#define init_5_14_6_8
void FASTCALL call_5_14_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_6_9
#define init_5_14_6_9
void FASTCALL call_5_14_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_6_10
#define init_5_14_6_10
void FASTCALL call_5_14_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_6_11
#define init_5_14_6_11
void FASTCALL call_5_14_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_6_12
#define init_5_14_6_12
void FASTCALL call_5_14_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_6_13
#define init_5_14_6_13
void FASTCALL call_5_14_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_6_14
#define init_5_14_6_14
void FASTCALL call_5_14_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_6_15
#define init_5_14_6_15
void FASTCALL call_5_14_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_7_0
#define init_5_14_7_0
void FASTCALL call_5_14_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_7_1
#define init_5_14_7_1
void FASTCALL call_5_14_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_7_2
#define init_5_14_7_2
void FASTCALL call_5_14_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_7_3
#define init_5_14_7_3
void FASTCALL call_5_14_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_7_4
#define init_5_14_7_4
void FASTCALL call_5_14_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_7_5
#define init_5_14_7_5
void FASTCALL call_5_14_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_7_6
#define init_5_14_7_6
void FASTCALL call_5_14_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_7_7
#define init_5_14_7_7
void FASTCALL call_5_14_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_7_8
#define init_5_14_7_8
void FASTCALL call_5_14_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_7_9
#define init_5_14_7_9
void FASTCALL call_5_14_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_7_10
#define init_5_14_7_10
void FASTCALL call_5_14_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_7_11
#define init_5_14_7_11
void FASTCALL call_5_14_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_7_12
#define init_5_14_7_12
void FASTCALL call_5_14_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_7_13
#define init_5_14_7_13
void FASTCALL call_5_14_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_7_14
#define init_5_14_7_14
void FASTCALL call_5_14_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_7_15
#define init_5_14_7_15
void FASTCALL call_5_14_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_8_0
#define init_5_14_8_0
void FASTCALL call_5_14_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_8_1
#define init_5_14_8_1
void FASTCALL call_5_14_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_8_2
#define init_5_14_8_2
void FASTCALL call_5_14_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_8_3
#define init_5_14_8_3
void FASTCALL call_5_14_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_8_4
#define init_5_14_8_4
void FASTCALL call_5_14_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_8_5
#define init_5_14_8_5
void FASTCALL call_5_14_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_8_6
#define init_5_14_8_6
void FASTCALL call_5_14_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_8_7
#define init_5_14_8_7
void FASTCALL call_5_14_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_8_8
#define init_5_14_8_8
void FASTCALL call_5_14_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_8_9
#define init_5_14_8_9
void FASTCALL call_5_14_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_8_10
#define init_5_14_8_10
void FASTCALL call_5_14_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_8_11
#define init_5_14_8_11
void FASTCALL call_5_14_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_8_12
#define init_5_14_8_12
void FASTCALL call_5_14_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_8_13
#define init_5_14_8_13
void FASTCALL call_5_14_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_8_14
#define init_5_14_8_14
void FASTCALL call_5_14_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_8_15
#define init_5_14_8_15
void FASTCALL call_5_14_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_9_0
#define init_5_14_9_0
void FASTCALL call_5_14_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_9_1
#define init_5_14_9_1
void FASTCALL call_5_14_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_9_2
#define init_5_14_9_2
void FASTCALL call_5_14_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_9_3
#define init_5_14_9_3
void FASTCALL call_5_14_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_9_4
#define init_5_14_9_4
void FASTCALL call_5_14_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_9_5
#define init_5_14_9_5
void FASTCALL call_5_14_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_9_6
#define init_5_14_9_6
void FASTCALL call_5_14_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_9_7
#define init_5_14_9_7
void FASTCALL call_5_14_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_9_8
#define init_5_14_9_8
void FASTCALL call_5_14_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_9_9
#define init_5_14_9_9
void FASTCALL call_5_14_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_9_10
#define init_5_14_9_10
void FASTCALL call_5_14_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_9_11
#define init_5_14_9_11
void FASTCALL call_5_14_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_9_12
#define init_5_14_9_12
void FASTCALL call_5_14_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_9_13
#define init_5_14_9_13
void FASTCALL call_5_14_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_9_14
#define init_5_14_9_14
void FASTCALL call_5_14_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_9_15
#define init_5_14_9_15
void FASTCALL call_5_14_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_10_0
#define init_5_14_10_0
void FASTCALL call_5_14_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_10_1
#define init_5_14_10_1
void FASTCALL call_5_14_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_10_2
#define init_5_14_10_2
void FASTCALL call_5_14_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_10_3
#define init_5_14_10_3
void FASTCALL call_5_14_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_10_4
#define init_5_14_10_4
void FASTCALL call_5_14_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_10_5
#define init_5_14_10_5
void FASTCALL call_5_14_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_10_6
#define init_5_14_10_6
void FASTCALL call_5_14_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_10_7
#define init_5_14_10_7
void FASTCALL call_5_14_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_10_8
#define init_5_14_10_8
void FASTCALL call_5_14_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_10_9
#define init_5_14_10_9
void FASTCALL call_5_14_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_10_10
#define init_5_14_10_10
void FASTCALL call_5_14_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_10_11
#define init_5_14_10_11
void FASTCALL call_5_14_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_10_12
#define init_5_14_10_12
void FASTCALL call_5_14_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_10_13
#define init_5_14_10_13
void FASTCALL call_5_14_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_10_14
#define init_5_14_10_14
void FASTCALL call_5_14_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_10_15
#define init_5_14_10_15
void FASTCALL call_5_14_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_11_0
#define init_5_14_11_0
void FASTCALL call_5_14_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_11_1
#define init_5_14_11_1
void FASTCALL call_5_14_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_11_2
#define init_5_14_11_2
void FASTCALL call_5_14_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_11_3
#define init_5_14_11_3
void FASTCALL call_5_14_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_11_4
#define init_5_14_11_4
void FASTCALL call_5_14_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_11_5
#define init_5_14_11_5
void FASTCALL call_5_14_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_11_6
#define init_5_14_11_6
void FASTCALL call_5_14_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_11_7
#define init_5_14_11_7
void FASTCALL call_5_14_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_11_8
#define init_5_14_11_8
void FASTCALL call_5_14_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_11_9
#define init_5_14_11_9
void FASTCALL call_5_14_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_11_10
#define init_5_14_11_10
void FASTCALL call_5_14_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_11_11
#define init_5_14_11_11
void FASTCALL call_5_14_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_11_12
#define init_5_14_11_12
void FASTCALL call_5_14_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_11_13
#define init_5_14_11_13
void FASTCALL call_5_14_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_11_14
#define init_5_14_11_14
void FASTCALL call_5_14_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_11_15
#define init_5_14_11_15
void FASTCALL call_5_14_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_12_0
#define init_5_14_12_0
void FASTCALL call_5_14_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_12_1
#define init_5_14_12_1
void FASTCALL call_5_14_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_12_2
#define init_5_14_12_2
void FASTCALL call_5_14_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_12_3
#define init_5_14_12_3
void FASTCALL call_5_14_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_12_4
#define init_5_14_12_4
void FASTCALL call_5_14_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_12_5
#define init_5_14_12_5
void FASTCALL call_5_14_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_12_6
#define init_5_14_12_6
void FASTCALL call_5_14_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_12_7
#define init_5_14_12_7
void FASTCALL call_5_14_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_12_8
#define init_5_14_12_8
void FASTCALL call_5_14_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_12_9
#define init_5_14_12_9
void FASTCALL call_5_14_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_12_10
#define init_5_14_12_10
void FASTCALL call_5_14_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_12_11
#define init_5_14_12_11
void FASTCALL call_5_14_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_12_12
#define init_5_14_12_12
void FASTCALL call_5_14_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_12_13
#define init_5_14_12_13
void FASTCALL call_5_14_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_12_14
#define init_5_14_12_14
void FASTCALL call_5_14_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_12_15
#define init_5_14_12_15
void FASTCALL call_5_14_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_13_0
#define init_5_14_13_0
void FASTCALL call_5_14_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_13_1
#define init_5_14_13_1
void FASTCALL call_5_14_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_13_2
#define init_5_14_13_2
void FASTCALL call_5_14_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_13_3
#define init_5_14_13_3
void FASTCALL call_5_14_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_13_4
#define init_5_14_13_4
void FASTCALL call_5_14_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_13_5
#define init_5_14_13_5
void FASTCALL call_5_14_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_13_6
#define init_5_14_13_6
void FASTCALL call_5_14_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_13_7
#define init_5_14_13_7
void FASTCALL call_5_14_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_13_8
#define init_5_14_13_8
void FASTCALL call_5_14_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_13_9
#define init_5_14_13_9
void FASTCALL call_5_14_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_13_10
#define init_5_14_13_10
void FASTCALL call_5_14_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_13_11
#define init_5_14_13_11
void FASTCALL call_5_14_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_13_12
#define init_5_14_13_12
void FASTCALL call_5_14_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_13_13
#define init_5_14_13_13
void FASTCALL call_5_14_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_13_14
#define init_5_14_13_14
void FASTCALL call_5_14_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_13_15
#define init_5_14_13_15
void FASTCALL call_5_14_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_14_0
#define init_5_14_14_0
void FASTCALL call_5_14_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_14_1
#define init_5_14_14_1
void FASTCALL call_5_14_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_14_2
#define init_5_14_14_2
void FASTCALL call_5_14_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_14_3
#define init_5_14_14_3
void FASTCALL call_5_14_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_14_4
#define init_5_14_14_4
void FASTCALL call_5_14_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_14_5
#define init_5_14_14_5
void FASTCALL call_5_14_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_14_6
#define init_5_14_14_6
void FASTCALL call_5_14_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_14_7
#define init_5_14_14_7
void FASTCALL call_5_14_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_14_8
#define init_5_14_14_8
void FASTCALL call_5_14_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_14_9
#define init_5_14_14_9
void FASTCALL call_5_14_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_14_10
#define init_5_14_14_10
void FASTCALL call_5_14_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_14_11
#define init_5_14_14_11
void FASTCALL call_5_14_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_14_12
#define init_5_14_14_12
void FASTCALL call_5_14_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_14_13
#define init_5_14_14_13
void FASTCALL call_5_14_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_14_14
#define init_5_14_14_14
void FASTCALL call_5_14_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_14_15
#define init_5_14_14_15
void FASTCALL call_5_14_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_15_0
#define init_5_14_15_0
void FASTCALL call_5_14_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_15_1
#define init_5_14_15_1
void FASTCALL call_5_14_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_15_2
#define init_5_14_15_2
void FASTCALL call_5_14_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_15_3
#define init_5_14_15_3
void FASTCALL call_5_14_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_15_4
#define init_5_14_15_4
void FASTCALL call_5_14_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_15_5
#define init_5_14_15_5
void FASTCALL call_5_14_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_15_6
#define init_5_14_15_6
void FASTCALL call_5_14_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_15_7
#define init_5_14_15_7
void FASTCALL call_5_14_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_15_8
#define init_5_14_15_8
void FASTCALL call_5_14_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_15_9
#define init_5_14_15_9
void FASTCALL call_5_14_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_15_10
#define init_5_14_15_10
void FASTCALL call_5_14_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_15_11
#define init_5_14_15_11
void FASTCALL call_5_14_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_15_12
#define init_5_14_15_12
void FASTCALL call_5_14_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_15_13
#define init_5_14_15_13
void FASTCALL call_5_14_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_15_14
#define init_5_14_15_14
void FASTCALL call_5_14_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_14_15_15
#define init_5_14_15_15
void FASTCALL call_5_14_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_0_0
#define init_5_15_0_0
void FASTCALL call_5_15_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_0_1
#define init_5_15_0_1
void FASTCALL call_5_15_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_0_2
#define init_5_15_0_2
void FASTCALL call_5_15_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_0_3
#define init_5_15_0_3
void FASTCALL call_5_15_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_0_4
#define init_5_15_0_4
void FASTCALL call_5_15_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_0_5
#define init_5_15_0_5
void FASTCALL call_5_15_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_0_6
#define init_5_15_0_6
void FASTCALL call_5_15_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_0_7
#define init_5_15_0_7
void FASTCALL call_5_15_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_0_8
#define init_5_15_0_8
void FASTCALL call_5_15_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_0_9
#define init_5_15_0_9
void FASTCALL call_5_15_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_0_10
#define init_5_15_0_10
void FASTCALL call_5_15_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_0_11
#define init_5_15_0_11
void FASTCALL call_5_15_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_0_12
#define init_5_15_0_12
void FASTCALL call_5_15_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_0_13
#define init_5_15_0_13
void FASTCALL call_5_15_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_0_14
#define init_5_15_0_14
void FASTCALL call_5_15_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_0_15
#define init_5_15_0_15
void FASTCALL call_5_15_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_1_0
#define init_5_15_1_0
void FASTCALL call_5_15_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_1_1
#define init_5_15_1_1
void FASTCALL call_5_15_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_1_2
#define init_5_15_1_2
void FASTCALL call_5_15_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_1_3
#define init_5_15_1_3
void FASTCALL call_5_15_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_1_4
#define init_5_15_1_4
void FASTCALL call_5_15_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_1_5
#define init_5_15_1_5
void FASTCALL call_5_15_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_1_6
#define init_5_15_1_6
void FASTCALL call_5_15_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_1_7
#define init_5_15_1_7
void FASTCALL call_5_15_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_1_8
#define init_5_15_1_8
void FASTCALL call_5_15_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_1_9
#define init_5_15_1_9
void FASTCALL call_5_15_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_1_10
#define init_5_15_1_10
void FASTCALL call_5_15_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_1_11
#define init_5_15_1_11
void FASTCALL call_5_15_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_1_12
#define init_5_15_1_12
void FASTCALL call_5_15_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_1_13
#define init_5_15_1_13
void FASTCALL call_5_15_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_1_14
#define init_5_15_1_14
void FASTCALL call_5_15_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_1_15
#define init_5_15_1_15
void FASTCALL call_5_15_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_2_0
#define init_5_15_2_0
void FASTCALL call_5_15_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_2_1
#define init_5_15_2_1
void FASTCALL call_5_15_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_2_2
#define init_5_15_2_2
void FASTCALL call_5_15_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_2_3
#define init_5_15_2_3
void FASTCALL call_5_15_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_2_4
#define init_5_15_2_4
void FASTCALL call_5_15_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_2_5
#define init_5_15_2_5
void FASTCALL call_5_15_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_2_6
#define init_5_15_2_6
void FASTCALL call_5_15_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_2_7
#define init_5_15_2_7
void FASTCALL call_5_15_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_2_8
#define init_5_15_2_8
void FASTCALL call_5_15_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_2_9
#define init_5_15_2_9
void FASTCALL call_5_15_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_2_10
#define init_5_15_2_10
void FASTCALL call_5_15_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_2_11
#define init_5_15_2_11
void FASTCALL call_5_15_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_2_12
#define init_5_15_2_12
void FASTCALL call_5_15_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_2_13
#define init_5_15_2_13
void FASTCALL call_5_15_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_2_14
#define init_5_15_2_14
void FASTCALL call_5_15_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_2_15
#define init_5_15_2_15
void FASTCALL call_5_15_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_3_0
#define init_5_15_3_0
void FASTCALL call_5_15_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_3_1
#define init_5_15_3_1
void FASTCALL call_5_15_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_3_2
#define init_5_15_3_2
void FASTCALL call_5_15_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_3_3
#define init_5_15_3_3
void FASTCALL call_5_15_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_3_4
#define init_5_15_3_4
void FASTCALL call_5_15_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_3_5
#define init_5_15_3_5
void FASTCALL call_5_15_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_3_6
#define init_5_15_3_6
void FASTCALL call_5_15_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_3_7
#define init_5_15_3_7
void FASTCALL call_5_15_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_3_8
#define init_5_15_3_8
void FASTCALL call_5_15_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_3_9
#define init_5_15_3_9
void FASTCALL call_5_15_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_3_10
#define init_5_15_3_10
void FASTCALL call_5_15_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_3_11
#define init_5_15_3_11
void FASTCALL call_5_15_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_3_12
#define init_5_15_3_12
void FASTCALL call_5_15_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_3_13
#define init_5_15_3_13
void FASTCALL call_5_15_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_3_14
#define init_5_15_3_14
void FASTCALL call_5_15_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_3_15
#define init_5_15_3_15
void FASTCALL call_5_15_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_4_0
#define init_5_15_4_0
void FASTCALL call_5_15_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_4_1
#define init_5_15_4_1
void FASTCALL call_5_15_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_4_2
#define init_5_15_4_2
void FASTCALL call_5_15_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_4_3
#define init_5_15_4_3
void FASTCALL call_5_15_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_4_4
#define init_5_15_4_4
void FASTCALL call_5_15_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_4_5
#define init_5_15_4_5
void FASTCALL call_5_15_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_4_6
#define init_5_15_4_6
void FASTCALL call_5_15_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_4_7
#define init_5_15_4_7
void FASTCALL call_5_15_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_4_8
#define init_5_15_4_8
void FASTCALL call_5_15_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_4_9
#define init_5_15_4_9
void FASTCALL call_5_15_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_4_10
#define init_5_15_4_10
void FASTCALL call_5_15_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_4_11
#define init_5_15_4_11
void FASTCALL call_5_15_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_4_12
#define init_5_15_4_12
void FASTCALL call_5_15_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_4_13
#define init_5_15_4_13
void FASTCALL call_5_15_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_4_14
#define init_5_15_4_14
void FASTCALL call_5_15_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_4_15
#define init_5_15_4_15
void FASTCALL call_5_15_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_5_0
#define init_5_15_5_0
void FASTCALL call_5_15_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_5_1
#define init_5_15_5_1
void FASTCALL call_5_15_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_5_2
#define init_5_15_5_2
void FASTCALL call_5_15_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_5_3
#define init_5_15_5_3
void FASTCALL call_5_15_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_5_4
#define init_5_15_5_4
void FASTCALL call_5_15_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_5_5
#define init_5_15_5_5
void FASTCALL call_5_15_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_5_6
#define init_5_15_5_6
void FASTCALL call_5_15_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_5_7
#define init_5_15_5_7
void FASTCALL call_5_15_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_5_8
#define init_5_15_5_8
void FASTCALL call_5_15_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_5_9
#define init_5_15_5_9
void FASTCALL call_5_15_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_5_10
#define init_5_15_5_10
void FASTCALL call_5_15_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_5_11
#define init_5_15_5_11
void FASTCALL call_5_15_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_5_12
#define init_5_15_5_12
void FASTCALL call_5_15_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_5_13
#define init_5_15_5_13
void FASTCALL call_5_15_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_5_14
#define init_5_15_5_14
void FASTCALL call_5_15_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_5_15
#define init_5_15_5_15
void FASTCALL call_5_15_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_6_0
#define init_5_15_6_0
void FASTCALL call_5_15_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_6_1
#define init_5_15_6_1
void FASTCALL call_5_15_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_6_2
#define init_5_15_6_2
void FASTCALL call_5_15_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_6_3
#define init_5_15_6_3
void FASTCALL call_5_15_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_6_4
#define init_5_15_6_4
void FASTCALL call_5_15_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_6_5
#define init_5_15_6_5
void FASTCALL call_5_15_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_6_6
#define init_5_15_6_6
void FASTCALL call_5_15_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_6_7
#define init_5_15_6_7
void FASTCALL call_5_15_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_6_8
#define init_5_15_6_8
void FASTCALL call_5_15_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_6_9
#define init_5_15_6_9
void FASTCALL call_5_15_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_6_10
#define init_5_15_6_10
void FASTCALL call_5_15_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_6_11
#define init_5_15_6_11
void FASTCALL call_5_15_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_6_12
#define init_5_15_6_12
void FASTCALL call_5_15_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_6_13
#define init_5_15_6_13
void FASTCALL call_5_15_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_6_14
#define init_5_15_6_14
void FASTCALL call_5_15_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_6_15
#define init_5_15_6_15
void FASTCALL call_5_15_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_7_0
#define init_5_15_7_0
void FASTCALL call_5_15_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_7_1
#define init_5_15_7_1
void FASTCALL call_5_15_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_7_2
#define init_5_15_7_2
void FASTCALL call_5_15_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_7_3
#define init_5_15_7_3
void FASTCALL call_5_15_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_7_4
#define init_5_15_7_4
void FASTCALL call_5_15_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_7_5
#define init_5_15_7_5
void FASTCALL call_5_15_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_7_6
#define init_5_15_7_6
void FASTCALL call_5_15_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_7_7
#define init_5_15_7_7
void FASTCALL call_5_15_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_7_8
#define init_5_15_7_8
void FASTCALL call_5_15_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_7_9
#define init_5_15_7_9
void FASTCALL call_5_15_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_7_10
#define init_5_15_7_10
void FASTCALL call_5_15_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_7_11
#define init_5_15_7_11
void FASTCALL call_5_15_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_7_12
#define init_5_15_7_12
void FASTCALL call_5_15_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_7_13
#define init_5_15_7_13
void FASTCALL call_5_15_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_7_14
#define init_5_15_7_14
void FASTCALL call_5_15_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_7_15
#define init_5_15_7_15
void FASTCALL call_5_15_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_8_0
#define init_5_15_8_0
void FASTCALL call_5_15_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_8_1
#define init_5_15_8_1
void FASTCALL call_5_15_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_8_2
#define init_5_15_8_2
void FASTCALL call_5_15_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_8_3
#define init_5_15_8_3
void FASTCALL call_5_15_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_8_4
#define init_5_15_8_4
void FASTCALL call_5_15_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_8_5
#define init_5_15_8_5
void FASTCALL call_5_15_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_8_6
#define init_5_15_8_6
void FASTCALL call_5_15_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_8_7
#define init_5_15_8_7
void FASTCALL call_5_15_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_8_8
#define init_5_15_8_8
void FASTCALL call_5_15_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_8_9
#define init_5_15_8_9
void FASTCALL call_5_15_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_8_10
#define init_5_15_8_10
void FASTCALL call_5_15_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_8_11
#define init_5_15_8_11
void FASTCALL call_5_15_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_8_12
#define init_5_15_8_12
void FASTCALL call_5_15_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_8_13
#define init_5_15_8_13
void FASTCALL call_5_15_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_8_14
#define init_5_15_8_14
void FASTCALL call_5_15_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_8_15
#define init_5_15_8_15
void FASTCALL call_5_15_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_9_0
#define init_5_15_9_0
void FASTCALL call_5_15_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_9_1
#define init_5_15_9_1
void FASTCALL call_5_15_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_9_2
#define init_5_15_9_2
void FASTCALL call_5_15_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_9_3
#define init_5_15_9_3
void FASTCALL call_5_15_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_9_4
#define init_5_15_9_4
void FASTCALL call_5_15_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_9_5
#define init_5_15_9_5
void FASTCALL call_5_15_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_9_6
#define init_5_15_9_6
void FASTCALL call_5_15_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_9_7
#define init_5_15_9_7
void FASTCALL call_5_15_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_9_8
#define init_5_15_9_8
void FASTCALL call_5_15_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_9_9
#define init_5_15_9_9
void FASTCALL call_5_15_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_9_10
#define init_5_15_9_10
void FASTCALL call_5_15_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_9_11
#define init_5_15_9_11
void FASTCALL call_5_15_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_9_12
#define init_5_15_9_12
void FASTCALL call_5_15_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_9_13
#define init_5_15_9_13
void FASTCALL call_5_15_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_9_14
#define init_5_15_9_14
void FASTCALL call_5_15_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_9_15
#define init_5_15_9_15
void FASTCALL call_5_15_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_10_0
#define init_5_15_10_0
void FASTCALL call_5_15_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_10_1
#define init_5_15_10_1
void FASTCALL call_5_15_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_10_2
#define init_5_15_10_2
void FASTCALL call_5_15_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_10_3
#define init_5_15_10_3
void FASTCALL call_5_15_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_10_4
#define init_5_15_10_4
void FASTCALL call_5_15_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_10_5
#define init_5_15_10_5
void FASTCALL call_5_15_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_10_6
#define init_5_15_10_6
void FASTCALL call_5_15_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_10_7
#define init_5_15_10_7
void FASTCALL call_5_15_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_10_8
#define init_5_15_10_8
void FASTCALL call_5_15_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_10_9
#define init_5_15_10_9
void FASTCALL call_5_15_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_10_10
#define init_5_15_10_10
void FASTCALL call_5_15_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_10_11
#define init_5_15_10_11
void FASTCALL call_5_15_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_10_12
#define init_5_15_10_12
void FASTCALL call_5_15_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_10_13
#define init_5_15_10_13
void FASTCALL call_5_15_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_10_14
#define init_5_15_10_14
void FASTCALL call_5_15_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_10_15
#define init_5_15_10_15
void FASTCALL call_5_15_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_11_0
#define init_5_15_11_0
void FASTCALL call_5_15_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_11_1
#define init_5_15_11_1
void FASTCALL call_5_15_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_11_2
#define init_5_15_11_2
void FASTCALL call_5_15_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_11_3
#define init_5_15_11_3
void FASTCALL call_5_15_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_11_4
#define init_5_15_11_4
void FASTCALL call_5_15_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_11_5
#define init_5_15_11_5
void FASTCALL call_5_15_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_11_6
#define init_5_15_11_6
void FASTCALL call_5_15_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_11_7
#define init_5_15_11_7
void FASTCALL call_5_15_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_11_8
#define init_5_15_11_8
void FASTCALL call_5_15_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_11_9
#define init_5_15_11_9
void FASTCALL call_5_15_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_11_10
#define init_5_15_11_10
void FASTCALL call_5_15_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_11_11
#define init_5_15_11_11
void FASTCALL call_5_15_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_11_12
#define init_5_15_11_12
void FASTCALL call_5_15_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_11_13
#define init_5_15_11_13
void FASTCALL call_5_15_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_11_14
#define init_5_15_11_14
void FASTCALL call_5_15_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_11_15
#define init_5_15_11_15
void FASTCALL call_5_15_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_12_0
#define init_5_15_12_0
void FASTCALL call_5_15_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_12_1
#define init_5_15_12_1
void FASTCALL call_5_15_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_12_2
#define init_5_15_12_2
void FASTCALL call_5_15_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_12_3
#define init_5_15_12_3
void FASTCALL call_5_15_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_12_4
#define init_5_15_12_4
void FASTCALL call_5_15_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_12_5
#define init_5_15_12_5
void FASTCALL call_5_15_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_12_6
#define init_5_15_12_6
void FASTCALL call_5_15_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_12_7
#define init_5_15_12_7
void FASTCALL call_5_15_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_12_8
#define init_5_15_12_8
void FASTCALL call_5_15_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_12_9
#define init_5_15_12_9
void FASTCALL call_5_15_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_12_10
#define init_5_15_12_10
void FASTCALL call_5_15_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_12_11
#define init_5_15_12_11
void FASTCALL call_5_15_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_12_12
#define init_5_15_12_12
void FASTCALL call_5_15_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_12_13
#define init_5_15_12_13
void FASTCALL call_5_15_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_12_14
#define init_5_15_12_14
void FASTCALL call_5_15_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_12_15
#define init_5_15_12_15
void FASTCALL call_5_15_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_13_0
#define init_5_15_13_0
void FASTCALL call_5_15_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_13_1
#define init_5_15_13_1
void FASTCALL call_5_15_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_13_2
#define init_5_15_13_2
void FASTCALL call_5_15_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_13_3
#define init_5_15_13_3
void FASTCALL call_5_15_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_13_4
#define init_5_15_13_4
void FASTCALL call_5_15_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_13_5
#define init_5_15_13_5
void FASTCALL call_5_15_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_13_6
#define init_5_15_13_6
void FASTCALL call_5_15_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_13_7
#define init_5_15_13_7
void FASTCALL call_5_15_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_13_8
#define init_5_15_13_8
void FASTCALL call_5_15_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_13_9
#define init_5_15_13_9
void FASTCALL call_5_15_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_13_10
#define init_5_15_13_10
void FASTCALL call_5_15_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_13_11
#define init_5_15_13_11
void FASTCALL call_5_15_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_13_12
#define init_5_15_13_12
void FASTCALL call_5_15_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_13_13
#define init_5_15_13_13
void FASTCALL call_5_15_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_13_14
#define init_5_15_13_14
void FASTCALL call_5_15_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_13_15
#define init_5_15_13_15
void FASTCALL call_5_15_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_14_0
#define init_5_15_14_0
void FASTCALL call_5_15_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_14_1
#define init_5_15_14_1
void FASTCALL call_5_15_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_14_2
#define init_5_15_14_2
void FASTCALL call_5_15_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_14_3
#define init_5_15_14_3
void FASTCALL call_5_15_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_14_4
#define init_5_15_14_4
void FASTCALL call_5_15_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_14_5
#define init_5_15_14_5
void FASTCALL call_5_15_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_14_6
#define init_5_15_14_6
void FASTCALL call_5_15_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_14_7
#define init_5_15_14_7
void FASTCALL call_5_15_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_14_8
#define init_5_15_14_8
void FASTCALL call_5_15_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_14_9
#define init_5_15_14_9
void FASTCALL call_5_15_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_14_10
#define init_5_15_14_10
void FASTCALL call_5_15_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_14_11
#define init_5_15_14_11
void FASTCALL call_5_15_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_14_12
#define init_5_15_14_12
void FASTCALL call_5_15_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_14_13
#define init_5_15_14_13
void FASTCALL call_5_15_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_14_14
#define init_5_15_14_14
void FASTCALL call_5_15_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_14_15
#define init_5_15_14_15
void FASTCALL call_5_15_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_15_0
#define init_5_15_15_0
void FASTCALL call_5_15_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_15_1
#define init_5_15_15_1
void FASTCALL call_5_15_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_15_2
#define init_5_15_15_2
void FASTCALL call_5_15_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_15_3
#define init_5_15_15_3
void FASTCALL call_5_15_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_15_4
#define init_5_15_15_4
void FASTCALL call_5_15_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_15_5
#define init_5_15_15_5
void FASTCALL call_5_15_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_15_6
#define init_5_15_15_6
void FASTCALL call_5_15_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_15_7
#define init_5_15_15_7
void FASTCALL call_5_15_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_15_8
#define init_5_15_15_8
void FASTCALL call_5_15_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_15_9
#define init_5_15_15_9
void FASTCALL call_5_15_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_15_10
#define init_5_15_15_10
void FASTCALL call_5_15_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_15_11
#define init_5_15_15_11
void FASTCALL call_5_15_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_15_12
#define init_5_15_15_12
void FASTCALL call_5_15_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_15_13
#define init_5_15_15_13
void FASTCALL call_5_15_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_15_14
#define init_5_15_15_14
void FASTCALL call_5_15_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_5_15_15_15
#define init_5_15_15_15
void FASTCALL call_5_15_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_0_0
#define init_6_0_0_0
void FASTCALL call_6_0_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_0_1
#define init_6_0_0_1
void FASTCALL call_6_0_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_0_2
#define init_6_0_0_2
void FASTCALL call_6_0_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_0_3
#define init_6_0_0_3
void FASTCALL call_6_0_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_0_4
#define init_6_0_0_4
void FASTCALL call_6_0_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_0_5
#define init_6_0_0_5
void FASTCALL call_6_0_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_0_6
#define init_6_0_0_6
void FASTCALL call_6_0_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_0_7
#define init_6_0_0_7
void FASTCALL call_6_0_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_0_8
#define init_6_0_0_8
void FASTCALL call_6_0_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_0_9
#define init_6_0_0_9
void FASTCALL call_6_0_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_0_10
#define init_6_0_0_10
void FASTCALL call_6_0_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_0_11
#define init_6_0_0_11
void FASTCALL call_6_0_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_0_12
#define init_6_0_0_12
void FASTCALL call_6_0_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_0_13
#define init_6_0_0_13
void FASTCALL call_6_0_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_0_14
#define init_6_0_0_14
void FASTCALL call_6_0_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_0_15
#define init_6_0_0_15
void FASTCALL call_6_0_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_1_0
#define init_6_0_1_0
void FASTCALL call_6_0_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_1_1
#define init_6_0_1_1
void FASTCALL call_6_0_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_1_2
#define init_6_0_1_2
void FASTCALL call_6_0_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_1_3
#define init_6_0_1_3
void FASTCALL call_6_0_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_1_4
#define init_6_0_1_4
void FASTCALL call_6_0_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_1_5
#define init_6_0_1_5
void FASTCALL call_6_0_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_1_6
#define init_6_0_1_6
void FASTCALL call_6_0_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_1_7
#define init_6_0_1_7
void FASTCALL call_6_0_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_1_8
#define init_6_0_1_8
void FASTCALL call_6_0_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_1_9
#define init_6_0_1_9
void FASTCALL call_6_0_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_1_10
#define init_6_0_1_10
void FASTCALL call_6_0_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_1_11
#define init_6_0_1_11
void FASTCALL call_6_0_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_1_12
#define init_6_0_1_12
void FASTCALL call_6_0_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_1_13
#define init_6_0_1_13
void FASTCALL call_6_0_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_1_14
#define init_6_0_1_14
void FASTCALL call_6_0_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_1_15
#define init_6_0_1_15
void FASTCALL call_6_0_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_2_0
#define init_6_0_2_0
void FASTCALL call_6_0_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_2_1
#define init_6_0_2_1
void FASTCALL call_6_0_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_2_2
#define init_6_0_2_2
void FASTCALL call_6_0_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_2_3
#define init_6_0_2_3
void FASTCALL call_6_0_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_2_4
#define init_6_0_2_4
void FASTCALL call_6_0_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_2_5
#define init_6_0_2_5
void FASTCALL call_6_0_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_2_6
#define init_6_0_2_6
void FASTCALL call_6_0_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_2_7
#define init_6_0_2_7
void FASTCALL call_6_0_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_2_8
#define init_6_0_2_8
void FASTCALL call_6_0_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_2_9
#define init_6_0_2_9
void FASTCALL call_6_0_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_2_10
#define init_6_0_2_10
void FASTCALL call_6_0_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_2_11
#define init_6_0_2_11
void FASTCALL call_6_0_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_2_12
#define init_6_0_2_12
void FASTCALL call_6_0_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_2_13
#define init_6_0_2_13
void FASTCALL call_6_0_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_2_14
#define init_6_0_2_14
void FASTCALL call_6_0_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_2_15
#define init_6_0_2_15
void FASTCALL call_6_0_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_3_0
#define init_6_0_3_0
void FASTCALL call_6_0_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_3_1
#define init_6_0_3_1
void FASTCALL call_6_0_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_3_2
#define init_6_0_3_2
void FASTCALL call_6_0_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_3_3
#define init_6_0_3_3
void FASTCALL call_6_0_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_3_4
#define init_6_0_3_4
void FASTCALL call_6_0_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_3_5
#define init_6_0_3_5
void FASTCALL call_6_0_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_3_6
#define init_6_0_3_6
void FASTCALL call_6_0_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_3_7
#define init_6_0_3_7
void FASTCALL call_6_0_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_3_8
#define init_6_0_3_8
void FASTCALL call_6_0_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_3_9
#define init_6_0_3_9
void FASTCALL call_6_0_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_3_10
#define init_6_0_3_10
void FASTCALL call_6_0_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_3_11
#define init_6_0_3_11
void FASTCALL call_6_0_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_3_12
#define init_6_0_3_12
void FASTCALL call_6_0_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_3_13
#define init_6_0_3_13
void FASTCALL call_6_0_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_3_14
#define init_6_0_3_14
void FASTCALL call_6_0_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_3_15
#define init_6_0_3_15
void FASTCALL call_6_0_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_4_0
#define init_6_0_4_0
void FASTCALL call_6_0_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_4_1
#define init_6_0_4_1
void FASTCALL call_6_0_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_4_2
#define init_6_0_4_2
void FASTCALL call_6_0_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_4_3
#define init_6_0_4_3
void FASTCALL call_6_0_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_4_4
#define init_6_0_4_4
void FASTCALL call_6_0_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_4_5
#define init_6_0_4_5
void FASTCALL call_6_0_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_4_6
#define init_6_0_4_6
void FASTCALL call_6_0_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_4_7
#define init_6_0_4_7
void FASTCALL call_6_0_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_4_8
#define init_6_0_4_8
void FASTCALL call_6_0_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_4_9
#define init_6_0_4_9
void FASTCALL call_6_0_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_4_10
#define init_6_0_4_10
void FASTCALL call_6_0_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_4_11
#define init_6_0_4_11
void FASTCALL call_6_0_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_4_12
#define init_6_0_4_12
void FASTCALL call_6_0_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_4_13
#define init_6_0_4_13
void FASTCALL call_6_0_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_4_14
#define init_6_0_4_14
void FASTCALL call_6_0_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_4_15
#define init_6_0_4_15
void FASTCALL call_6_0_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_5_0
#define init_6_0_5_0
void FASTCALL call_6_0_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_5_1
#define init_6_0_5_1
void FASTCALL call_6_0_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_5_2
#define init_6_0_5_2
void FASTCALL call_6_0_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_5_3
#define init_6_0_5_3
void FASTCALL call_6_0_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_5_4
#define init_6_0_5_4
void FASTCALL call_6_0_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_5_5
#define init_6_0_5_5
void FASTCALL call_6_0_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_5_6
#define init_6_0_5_6
void FASTCALL call_6_0_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_5_7
#define init_6_0_5_7
void FASTCALL call_6_0_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_5_8
#define init_6_0_5_8
void FASTCALL call_6_0_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_5_9
#define init_6_0_5_9
void FASTCALL call_6_0_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_5_10
#define init_6_0_5_10
void FASTCALL call_6_0_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_5_11
#define init_6_0_5_11
void FASTCALL call_6_0_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_5_12
#define init_6_0_5_12
void FASTCALL call_6_0_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_5_13
#define init_6_0_5_13
void FASTCALL call_6_0_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_5_14
#define init_6_0_5_14
void FASTCALL call_6_0_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_5_15
#define init_6_0_5_15
void FASTCALL call_6_0_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_6_0
#define init_6_0_6_0
void FASTCALL call_6_0_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_6_1
#define init_6_0_6_1
void FASTCALL call_6_0_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_6_2
#define init_6_0_6_2
void FASTCALL call_6_0_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_6_3
#define init_6_0_6_3
void FASTCALL call_6_0_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_6_4
#define init_6_0_6_4
void FASTCALL call_6_0_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_6_5
#define init_6_0_6_5
void FASTCALL call_6_0_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_6_6
#define init_6_0_6_6
void FASTCALL call_6_0_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_6_7
#define init_6_0_6_7
void FASTCALL call_6_0_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_6_8
#define init_6_0_6_8
void FASTCALL call_6_0_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_6_9
#define init_6_0_6_9
void FASTCALL call_6_0_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_6_10
#define init_6_0_6_10
void FASTCALL call_6_0_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_6_11
#define init_6_0_6_11
void FASTCALL call_6_0_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_6_12
#define init_6_0_6_12
void FASTCALL call_6_0_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_6_13
#define init_6_0_6_13
void FASTCALL call_6_0_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_6_14
#define init_6_0_6_14
void FASTCALL call_6_0_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_6_15
#define init_6_0_6_15
void FASTCALL call_6_0_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_7_0
#define init_6_0_7_0
void FASTCALL call_6_0_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_7_1
#define init_6_0_7_1
void FASTCALL call_6_0_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_7_2
#define init_6_0_7_2
void FASTCALL call_6_0_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_7_3
#define init_6_0_7_3
void FASTCALL call_6_0_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_7_4
#define init_6_0_7_4
void FASTCALL call_6_0_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_7_5
#define init_6_0_7_5
void FASTCALL call_6_0_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_7_6
#define init_6_0_7_6
void FASTCALL call_6_0_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_7_7
#define init_6_0_7_7
void FASTCALL call_6_0_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_7_8
#define init_6_0_7_8
void FASTCALL call_6_0_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_7_9
#define init_6_0_7_9
void FASTCALL call_6_0_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_7_10
#define init_6_0_7_10
void FASTCALL call_6_0_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_7_11
#define init_6_0_7_11
void FASTCALL call_6_0_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_7_12
#define init_6_0_7_12
void FASTCALL call_6_0_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_7_13
#define init_6_0_7_13
void FASTCALL call_6_0_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_7_14
#define init_6_0_7_14
void FASTCALL call_6_0_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_7_15
#define init_6_0_7_15
void FASTCALL call_6_0_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_8_0
#define init_6_0_8_0
void FASTCALL call_6_0_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_8_1
#define init_6_0_8_1
void FASTCALL call_6_0_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_8_2
#define init_6_0_8_2
void FASTCALL call_6_0_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_8_3
#define init_6_0_8_3
void FASTCALL call_6_0_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_8_4
#define init_6_0_8_4
void FASTCALL call_6_0_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_8_5
#define init_6_0_8_5
void FASTCALL call_6_0_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_8_6
#define init_6_0_8_6
void FASTCALL call_6_0_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_8_7
#define init_6_0_8_7
void FASTCALL call_6_0_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_8_8
#define init_6_0_8_8
void FASTCALL call_6_0_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_8_9
#define init_6_0_8_9
void FASTCALL call_6_0_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_8_10
#define init_6_0_8_10
void FASTCALL call_6_0_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_8_11
#define init_6_0_8_11
void FASTCALL call_6_0_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_8_12
#define init_6_0_8_12
void FASTCALL call_6_0_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_8_13
#define init_6_0_8_13
void FASTCALL call_6_0_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_8_14
#define init_6_0_8_14
void FASTCALL call_6_0_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_8_15
#define init_6_0_8_15
void FASTCALL call_6_0_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_9_0
#define init_6_0_9_0
void FASTCALL call_6_0_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_9_1
#define init_6_0_9_1
void FASTCALL call_6_0_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_9_2
#define init_6_0_9_2
void FASTCALL call_6_0_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_9_3
#define init_6_0_9_3
void FASTCALL call_6_0_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_9_4
#define init_6_0_9_4
void FASTCALL call_6_0_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_9_5
#define init_6_0_9_5
void FASTCALL call_6_0_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_9_6
#define init_6_0_9_6
void FASTCALL call_6_0_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_9_7
#define init_6_0_9_7
void FASTCALL call_6_0_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_9_8
#define init_6_0_9_8
void FASTCALL call_6_0_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_9_9
#define init_6_0_9_9
void FASTCALL call_6_0_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_9_10
#define init_6_0_9_10
void FASTCALL call_6_0_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_9_11
#define init_6_0_9_11
void FASTCALL call_6_0_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_9_12
#define init_6_0_9_12
void FASTCALL call_6_0_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_9_13
#define init_6_0_9_13
void FASTCALL call_6_0_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_9_14
#define init_6_0_9_14
void FASTCALL call_6_0_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_9_15
#define init_6_0_9_15
void FASTCALL call_6_0_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_10_0
#define init_6_0_10_0
void FASTCALL call_6_0_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_10_1
#define init_6_0_10_1
void FASTCALL call_6_0_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_10_2
#define init_6_0_10_2
void FASTCALL call_6_0_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_10_3
#define init_6_0_10_3
void FASTCALL call_6_0_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_10_4
#define init_6_0_10_4
void FASTCALL call_6_0_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_10_5
#define init_6_0_10_5
void FASTCALL call_6_0_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_10_6
#define init_6_0_10_6
void FASTCALL call_6_0_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_10_7
#define init_6_0_10_7
void FASTCALL call_6_0_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_10_8
#define init_6_0_10_8
void FASTCALL call_6_0_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_10_9
#define init_6_0_10_9
void FASTCALL call_6_0_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_10_10
#define init_6_0_10_10
void FASTCALL call_6_0_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_10_11
#define init_6_0_10_11
void FASTCALL call_6_0_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_10_12
#define init_6_0_10_12
void FASTCALL call_6_0_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_10_13
#define init_6_0_10_13
void FASTCALL call_6_0_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_10_14
#define init_6_0_10_14
void FASTCALL call_6_0_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_10_15
#define init_6_0_10_15
void FASTCALL call_6_0_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_11_0
#define init_6_0_11_0
void FASTCALL call_6_0_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_11_1
#define init_6_0_11_1
void FASTCALL call_6_0_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_11_2
#define init_6_0_11_2
void FASTCALL call_6_0_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_11_3
#define init_6_0_11_3
void FASTCALL call_6_0_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_11_4
#define init_6_0_11_4
void FASTCALL call_6_0_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_11_5
#define init_6_0_11_5
void FASTCALL call_6_0_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_11_6
#define init_6_0_11_6
void FASTCALL call_6_0_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_11_7
#define init_6_0_11_7
void FASTCALL call_6_0_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_11_8
#define init_6_0_11_8
void FASTCALL call_6_0_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_11_9
#define init_6_0_11_9
void FASTCALL call_6_0_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_11_10
#define init_6_0_11_10
void FASTCALL call_6_0_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_11_11
#define init_6_0_11_11
void FASTCALL call_6_0_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_11_12
#define init_6_0_11_12
void FASTCALL call_6_0_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_11_13
#define init_6_0_11_13
void FASTCALL call_6_0_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_11_14
#define init_6_0_11_14
void FASTCALL call_6_0_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_11_15
#define init_6_0_11_15
void FASTCALL call_6_0_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_12_0
#define init_6_0_12_0
void FASTCALL call_6_0_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_12_1
#define init_6_0_12_1
void FASTCALL call_6_0_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_12_2
#define init_6_0_12_2
void FASTCALL call_6_0_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_12_3
#define init_6_0_12_3
void FASTCALL call_6_0_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_12_4
#define init_6_0_12_4
void FASTCALL call_6_0_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_12_5
#define init_6_0_12_5
void FASTCALL call_6_0_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_12_6
#define init_6_0_12_6
void FASTCALL call_6_0_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_12_7
#define init_6_0_12_7
void FASTCALL call_6_0_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_12_8
#define init_6_0_12_8
void FASTCALL call_6_0_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_12_9
#define init_6_0_12_9
void FASTCALL call_6_0_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_12_10
#define init_6_0_12_10
void FASTCALL call_6_0_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_12_11
#define init_6_0_12_11
void FASTCALL call_6_0_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_12_12
#define init_6_0_12_12
void FASTCALL call_6_0_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_12_13
#define init_6_0_12_13
void FASTCALL call_6_0_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_12_14
#define init_6_0_12_14
void FASTCALL call_6_0_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_12_15
#define init_6_0_12_15
void FASTCALL call_6_0_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_13_0
#define init_6_0_13_0
void FASTCALL call_6_0_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_13_1
#define init_6_0_13_1
void FASTCALL call_6_0_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_13_2
#define init_6_0_13_2
void FASTCALL call_6_0_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_13_3
#define init_6_0_13_3
void FASTCALL call_6_0_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_13_4
#define init_6_0_13_4
void FASTCALL call_6_0_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_13_5
#define init_6_0_13_5
void FASTCALL call_6_0_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_13_6
#define init_6_0_13_6
void FASTCALL call_6_0_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_13_7
#define init_6_0_13_7
void FASTCALL call_6_0_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_13_8
#define init_6_0_13_8
void FASTCALL call_6_0_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_13_9
#define init_6_0_13_9
void FASTCALL call_6_0_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_13_10
#define init_6_0_13_10
void FASTCALL call_6_0_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_13_11
#define init_6_0_13_11
void FASTCALL call_6_0_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_13_12
#define init_6_0_13_12
void FASTCALL call_6_0_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_13_13
#define init_6_0_13_13
void FASTCALL call_6_0_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_13_14
#define init_6_0_13_14
void FASTCALL call_6_0_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_13_15
#define init_6_0_13_15
void FASTCALL call_6_0_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_14_0
#define init_6_0_14_0
void FASTCALL call_6_0_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_14_1
#define init_6_0_14_1
void FASTCALL call_6_0_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_14_2
#define init_6_0_14_2
void FASTCALL call_6_0_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_14_3
#define init_6_0_14_3
void FASTCALL call_6_0_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_14_4
#define init_6_0_14_4
void FASTCALL call_6_0_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_14_5
#define init_6_0_14_5
void FASTCALL call_6_0_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_14_6
#define init_6_0_14_6
void FASTCALL call_6_0_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_14_7
#define init_6_0_14_7
void FASTCALL call_6_0_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_14_8
#define init_6_0_14_8
void FASTCALL call_6_0_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_14_9
#define init_6_0_14_9
void FASTCALL call_6_0_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_14_10
#define init_6_0_14_10
void FASTCALL call_6_0_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_14_11
#define init_6_0_14_11
void FASTCALL call_6_0_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_14_12
#define init_6_0_14_12
void FASTCALL call_6_0_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_14_13
#define init_6_0_14_13
void FASTCALL call_6_0_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_14_14
#define init_6_0_14_14
void FASTCALL call_6_0_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_14_15
#define init_6_0_14_15
void FASTCALL call_6_0_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_15_0
#define init_6_0_15_0
void FASTCALL call_6_0_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_15_1
#define init_6_0_15_1
void FASTCALL call_6_0_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_15_2
#define init_6_0_15_2
void FASTCALL call_6_0_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_15_3
#define init_6_0_15_3
void FASTCALL call_6_0_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_15_4
#define init_6_0_15_4
void FASTCALL call_6_0_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_15_5
#define init_6_0_15_5
void FASTCALL call_6_0_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_15_6
#define init_6_0_15_6
void FASTCALL call_6_0_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_15_7
#define init_6_0_15_7
void FASTCALL call_6_0_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_15_8
#define init_6_0_15_8
void FASTCALL call_6_0_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_15_9
#define init_6_0_15_9
void FASTCALL call_6_0_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_15_10
#define init_6_0_15_10
void FASTCALL call_6_0_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_15_11
#define init_6_0_15_11
void FASTCALL call_6_0_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_15_12
#define init_6_0_15_12
void FASTCALL call_6_0_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_15_13
#define init_6_0_15_13
void FASTCALL call_6_0_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_15_14
#define init_6_0_15_14
void FASTCALL call_6_0_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_0_15_15
#define init_6_0_15_15
void FASTCALL call_6_0_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_0_0
#define init_6_1_0_0
void FASTCALL call_6_1_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_0_1
#define init_6_1_0_1
void FASTCALL call_6_1_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_0_2
#define init_6_1_0_2
void FASTCALL call_6_1_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_0_3
#define init_6_1_0_3
void FASTCALL call_6_1_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_0_4
#define init_6_1_0_4
void FASTCALL call_6_1_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_0_5
#define init_6_1_0_5
void FASTCALL call_6_1_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_0_6
#define init_6_1_0_6
void FASTCALL call_6_1_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_0_7
#define init_6_1_0_7
void FASTCALL call_6_1_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_0_8
#define init_6_1_0_8
void FASTCALL call_6_1_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_0_9
#define init_6_1_0_9
void FASTCALL call_6_1_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_0_10
#define init_6_1_0_10
void FASTCALL call_6_1_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_0_11
#define init_6_1_0_11
void FASTCALL call_6_1_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_0_12
#define init_6_1_0_12
void FASTCALL call_6_1_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_0_13
#define init_6_1_0_13
void FASTCALL call_6_1_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_0_14
#define init_6_1_0_14
void FASTCALL call_6_1_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_0_15
#define init_6_1_0_15
void FASTCALL call_6_1_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_1_0
#define init_6_1_1_0
void FASTCALL call_6_1_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_1_1
#define init_6_1_1_1
void FASTCALL call_6_1_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_1_2
#define init_6_1_1_2
void FASTCALL call_6_1_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_1_3
#define init_6_1_1_3
void FASTCALL call_6_1_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_1_4
#define init_6_1_1_4
void FASTCALL call_6_1_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_1_5
#define init_6_1_1_5
void FASTCALL call_6_1_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_1_6
#define init_6_1_1_6
void FASTCALL call_6_1_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_1_7
#define init_6_1_1_7
void FASTCALL call_6_1_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_1_8
#define init_6_1_1_8
void FASTCALL call_6_1_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_1_9
#define init_6_1_1_9
void FASTCALL call_6_1_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_1_10
#define init_6_1_1_10
void FASTCALL call_6_1_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_1_11
#define init_6_1_1_11
void FASTCALL call_6_1_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_1_12
#define init_6_1_1_12
void FASTCALL call_6_1_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_1_13
#define init_6_1_1_13
void FASTCALL call_6_1_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_1_14
#define init_6_1_1_14
void FASTCALL call_6_1_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_1_15
#define init_6_1_1_15
void FASTCALL call_6_1_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_2_0
#define init_6_1_2_0
void FASTCALL call_6_1_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_2_1
#define init_6_1_2_1
void FASTCALL call_6_1_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_2_2
#define init_6_1_2_2
void FASTCALL call_6_1_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_2_3
#define init_6_1_2_3
void FASTCALL call_6_1_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_2_4
#define init_6_1_2_4
void FASTCALL call_6_1_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_2_5
#define init_6_1_2_5
void FASTCALL call_6_1_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_2_6
#define init_6_1_2_6
void FASTCALL call_6_1_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_2_7
#define init_6_1_2_7
void FASTCALL call_6_1_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_2_8
#define init_6_1_2_8
void FASTCALL call_6_1_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_2_9
#define init_6_1_2_9
void FASTCALL call_6_1_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_2_10
#define init_6_1_2_10
void FASTCALL call_6_1_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_2_11
#define init_6_1_2_11
void FASTCALL call_6_1_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_2_12
#define init_6_1_2_12
void FASTCALL call_6_1_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_2_13
#define init_6_1_2_13
void FASTCALL call_6_1_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_2_14
#define init_6_1_2_14
void FASTCALL call_6_1_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_2_15
#define init_6_1_2_15
void FASTCALL call_6_1_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_3_0
#define init_6_1_3_0
void FASTCALL call_6_1_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_3_1
#define init_6_1_3_1
void FASTCALL call_6_1_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_3_2
#define init_6_1_3_2
void FASTCALL call_6_1_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_3_3
#define init_6_1_3_3
void FASTCALL call_6_1_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_3_4
#define init_6_1_3_4
void FASTCALL call_6_1_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_3_5
#define init_6_1_3_5
void FASTCALL call_6_1_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_3_6
#define init_6_1_3_6
void FASTCALL call_6_1_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_3_7
#define init_6_1_3_7
void FASTCALL call_6_1_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_3_8
#define init_6_1_3_8
void FASTCALL call_6_1_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_3_9
#define init_6_1_3_9
void FASTCALL call_6_1_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_3_10
#define init_6_1_3_10
void FASTCALL call_6_1_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_3_11
#define init_6_1_3_11
void FASTCALL call_6_1_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_3_12
#define init_6_1_3_12
void FASTCALL call_6_1_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_3_13
#define init_6_1_3_13
void FASTCALL call_6_1_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_3_14
#define init_6_1_3_14
void FASTCALL call_6_1_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_3_15
#define init_6_1_3_15
void FASTCALL call_6_1_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_4_0
#define init_6_1_4_0
void FASTCALL call_6_1_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_4_1
#define init_6_1_4_1
void FASTCALL call_6_1_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_4_2
#define init_6_1_4_2
void FASTCALL call_6_1_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_4_3
#define init_6_1_4_3
void FASTCALL call_6_1_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_4_4
#define init_6_1_4_4
void FASTCALL call_6_1_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_4_5
#define init_6_1_4_5
void FASTCALL call_6_1_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_4_6
#define init_6_1_4_6
void FASTCALL call_6_1_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_4_7
#define init_6_1_4_7
void FASTCALL call_6_1_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_4_8
#define init_6_1_4_8
void FASTCALL call_6_1_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_4_9
#define init_6_1_4_9
void FASTCALL call_6_1_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_4_10
#define init_6_1_4_10
void FASTCALL call_6_1_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_4_11
#define init_6_1_4_11
void FASTCALL call_6_1_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_4_12
#define init_6_1_4_12
void FASTCALL call_6_1_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_4_13
#define init_6_1_4_13
void FASTCALL call_6_1_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_4_14
#define init_6_1_4_14
void FASTCALL call_6_1_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_4_15
#define init_6_1_4_15
void FASTCALL call_6_1_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_5_0
#define init_6_1_5_0
void FASTCALL call_6_1_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_5_1
#define init_6_1_5_1
void FASTCALL call_6_1_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_5_2
#define init_6_1_5_2
void FASTCALL call_6_1_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_5_3
#define init_6_1_5_3
void FASTCALL call_6_1_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_5_4
#define init_6_1_5_4
void FASTCALL call_6_1_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_5_5
#define init_6_1_5_5
void FASTCALL call_6_1_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_5_6
#define init_6_1_5_6
void FASTCALL call_6_1_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_5_7
#define init_6_1_5_7
void FASTCALL call_6_1_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_5_8
#define init_6_1_5_8
void FASTCALL call_6_1_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_5_9
#define init_6_1_5_9
void FASTCALL call_6_1_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_5_10
#define init_6_1_5_10
void FASTCALL call_6_1_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_5_11
#define init_6_1_5_11
void FASTCALL call_6_1_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_5_12
#define init_6_1_5_12
void FASTCALL call_6_1_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_5_13
#define init_6_1_5_13
void FASTCALL call_6_1_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_5_14
#define init_6_1_5_14
void FASTCALL call_6_1_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_5_15
#define init_6_1_5_15
void FASTCALL call_6_1_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_6_0
#define init_6_1_6_0
void FASTCALL call_6_1_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_6_1
#define init_6_1_6_1
void FASTCALL call_6_1_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_6_2
#define init_6_1_6_2
void FASTCALL call_6_1_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_6_3
#define init_6_1_6_3
void FASTCALL call_6_1_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_6_4
#define init_6_1_6_4
void FASTCALL call_6_1_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_6_5
#define init_6_1_6_5
void FASTCALL call_6_1_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_6_6
#define init_6_1_6_6
void FASTCALL call_6_1_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_6_7
#define init_6_1_6_7
void FASTCALL call_6_1_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_6_8
#define init_6_1_6_8
void FASTCALL call_6_1_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_6_9
#define init_6_1_6_9
void FASTCALL call_6_1_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_6_10
#define init_6_1_6_10
void FASTCALL call_6_1_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_6_11
#define init_6_1_6_11
void FASTCALL call_6_1_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_6_12
#define init_6_1_6_12
void FASTCALL call_6_1_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_6_13
#define init_6_1_6_13
void FASTCALL call_6_1_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_6_14
#define init_6_1_6_14
void FASTCALL call_6_1_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_6_15
#define init_6_1_6_15
void FASTCALL call_6_1_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_7_0
#define init_6_1_7_0
void FASTCALL call_6_1_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_7_1
#define init_6_1_7_1
void FASTCALL call_6_1_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_7_2
#define init_6_1_7_2
void FASTCALL call_6_1_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_7_3
#define init_6_1_7_3
void FASTCALL call_6_1_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_7_4
#define init_6_1_7_4
void FASTCALL call_6_1_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_7_5
#define init_6_1_7_5
void FASTCALL call_6_1_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_7_6
#define init_6_1_7_6
void FASTCALL call_6_1_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_7_7
#define init_6_1_7_7
void FASTCALL call_6_1_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_7_8
#define init_6_1_7_8
void FASTCALL call_6_1_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_7_9
#define init_6_1_7_9
void FASTCALL call_6_1_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_7_10
#define init_6_1_7_10
void FASTCALL call_6_1_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_7_11
#define init_6_1_7_11
void FASTCALL call_6_1_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_7_12
#define init_6_1_7_12
void FASTCALL call_6_1_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_7_13
#define init_6_1_7_13
void FASTCALL call_6_1_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_7_14
#define init_6_1_7_14
void FASTCALL call_6_1_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_7_15
#define init_6_1_7_15
void FASTCALL call_6_1_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_8_0
#define init_6_1_8_0
void FASTCALL call_6_1_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_8_1
#define init_6_1_8_1
void FASTCALL call_6_1_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_8_2
#define init_6_1_8_2
void FASTCALL call_6_1_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_8_3
#define init_6_1_8_3
void FASTCALL call_6_1_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_8_4
#define init_6_1_8_4
void FASTCALL call_6_1_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_8_5
#define init_6_1_8_5
void FASTCALL call_6_1_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_8_6
#define init_6_1_8_6
void FASTCALL call_6_1_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_8_7
#define init_6_1_8_7
void FASTCALL call_6_1_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_8_8
#define init_6_1_8_8
void FASTCALL call_6_1_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_8_9
#define init_6_1_8_9
void FASTCALL call_6_1_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_8_10
#define init_6_1_8_10
void FASTCALL call_6_1_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_8_11
#define init_6_1_8_11
void FASTCALL call_6_1_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_8_12
#define init_6_1_8_12
void FASTCALL call_6_1_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_8_13
#define init_6_1_8_13
void FASTCALL call_6_1_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_8_14
#define init_6_1_8_14
void FASTCALL call_6_1_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_8_15
#define init_6_1_8_15
void FASTCALL call_6_1_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_9_0
#define init_6_1_9_0
void FASTCALL call_6_1_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_9_1
#define init_6_1_9_1
void FASTCALL call_6_1_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_9_2
#define init_6_1_9_2
void FASTCALL call_6_1_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_9_3
#define init_6_1_9_3
void FASTCALL call_6_1_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_9_4
#define init_6_1_9_4
void FASTCALL call_6_1_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_9_5
#define init_6_1_9_5
void FASTCALL call_6_1_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_9_6
#define init_6_1_9_6
void FASTCALL call_6_1_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_9_7
#define init_6_1_9_7
void FASTCALL call_6_1_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_9_8
#define init_6_1_9_8
void FASTCALL call_6_1_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_9_9
#define init_6_1_9_9
void FASTCALL call_6_1_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_9_10
#define init_6_1_9_10
void FASTCALL call_6_1_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_9_11
#define init_6_1_9_11
void FASTCALL call_6_1_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_9_12
#define init_6_1_9_12
void FASTCALL call_6_1_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_9_13
#define init_6_1_9_13
void FASTCALL call_6_1_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_9_14
#define init_6_1_9_14
void FASTCALL call_6_1_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_9_15
#define init_6_1_9_15
void FASTCALL call_6_1_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_10_0
#define init_6_1_10_0
void FASTCALL call_6_1_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_10_1
#define init_6_1_10_1
void FASTCALL call_6_1_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_10_2
#define init_6_1_10_2
void FASTCALL call_6_1_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_10_3
#define init_6_1_10_3
void FASTCALL call_6_1_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_10_4
#define init_6_1_10_4
void FASTCALL call_6_1_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_10_5
#define init_6_1_10_5
void FASTCALL call_6_1_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_10_6
#define init_6_1_10_6
void FASTCALL call_6_1_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_10_7
#define init_6_1_10_7
void FASTCALL call_6_1_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_10_8
#define init_6_1_10_8
void FASTCALL call_6_1_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_10_9
#define init_6_1_10_9
void FASTCALL call_6_1_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_10_10
#define init_6_1_10_10
void FASTCALL call_6_1_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_10_11
#define init_6_1_10_11
void FASTCALL call_6_1_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_10_12
#define init_6_1_10_12
void FASTCALL call_6_1_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_10_13
#define init_6_1_10_13
void FASTCALL call_6_1_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_10_14
#define init_6_1_10_14
void FASTCALL call_6_1_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_10_15
#define init_6_1_10_15
void FASTCALL call_6_1_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_11_0
#define init_6_1_11_0
void FASTCALL call_6_1_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_11_1
#define init_6_1_11_1
void FASTCALL call_6_1_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_11_2
#define init_6_1_11_2
void FASTCALL call_6_1_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_11_3
#define init_6_1_11_3
void FASTCALL call_6_1_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_11_4
#define init_6_1_11_4
void FASTCALL call_6_1_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_11_5
#define init_6_1_11_5
void FASTCALL call_6_1_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_11_6
#define init_6_1_11_6
void FASTCALL call_6_1_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_11_7
#define init_6_1_11_7
void FASTCALL call_6_1_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_11_8
#define init_6_1_11_8
void FASTCALL call_6_1_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_11_9
#define init_6_1_11_9
void FASTCALL call_6_1_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_11_10
#define init_6_1_11_10
void FASTCALL call_6_1_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_11_11
#define init_6_1_11_11
void FASTCALL call_6_1_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_11_12
#define init_6_1_11_12
void FASTCALL call_6_1_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_11_13
#define init_6_1_11_13
void FASTCALL call_6_1_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_11_14
#define init_6_1_11_14
void FASTCALL call_6_1_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_11_15
#define init_6_1_11_15
void FASTCALL call_6_1_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_12_0
#define init_6_1_12_0
void FASTCALL call_6_1_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_12_1
#define init_6_1_12_1
void FASTCALL call_6_1_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_12_2
#define init_6_1_12_2
void FASTCALL call_6_1_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_12_3
#define init_6_1_12_3
void FASTCALL call_6_1_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_12_4
#define init_6_1_12_4
void FASTCALL call_6_1_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_12_5
#define init_6_1_12_5
void FASTCALL call_6_1_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_12_6
#define init_6_1_12_6
void FASTCALL call_6_1_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_12_7
#define init_6_1_12_7
void FASTCALL call_6_1_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_12_8
#define init_6_1_12_8
void FASTCALL call_6_1_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_12_9
#define init_6_1_12_9
void FASTCALL call_6_1_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_12_10
#define init_6_1_12_10
void FASTCALL call_6_1_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_12_11
#define init_6_1_12_11
void FASTCALL call_6_1_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_12_12
#define init_6_1_12_12
void FASTCALL call_6_1_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_12_13
#define init_6_1_12_13
void FASTCALL call_6_1_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_12_14
#define init_6_1_12_14
void FASTCALL call_6_1_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_12_15
#define init_6_1_12_15
void FASTCALL call_6_1_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_13_0
#define init_6_1_13_0
void FASTCALL call_6_1_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_13_1
#define init_6_1_13_1
void FASTCALL call_6_1_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_13_2
#define init_6_1_13_2
void FASTCALL call_6_1_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_13_3
#define init_6_1_13_3
void FASTCALL call_6_1_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_13_4
#define init_6_1_13_4
void FASTCALL call_6_1_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_13_5
#define init_6_1_13_5
void FASTCALL call_6_1_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_13_6
#define init_6_1_13_6
void FASTCALL call_6_1_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_13_7
#define init_6_1_13_7
void FASTCALL call_6_1_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_13_8
#define init_6_1_13_8
void FASTCALL call_6_1_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_13_9
#define init_6_1_13_9
void FASTCALL call_6_1_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_13_10
#define init_6_1_13_10
void FASTCALL call_6_1_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_13_11
#define init_6_1_13_11
void FASTCALL call_6_1_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_13_12
#define init_6_1_13_12
void FASTCALL call_6_1_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_13_13
#define init_6_1_13_13
void FASTCALL call_6_1_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_13_14
#define init_6_1_13_14
void FASTCALL call_6_1_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_13_15
#define init_6_1_13_15
void FASTCALL call_6_1_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_14_0
#define init_6_1_14_0
void FASTCALL call_6_1_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_14_1
#define init_6_1_14_1
void FASTCALL call_6_1_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_14_2
#define init_6_1_14_2
void FASTCALL call_6_1_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_14_3
#define init_6_1_14_3
void FASTCALL call_6_1_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_14_4
#define init_6_1_14_4
void FASTCALL call_6_1_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_14_5
#define init_6_1_14_5
void FASTCALL call_6_1_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_14_6
#define init_6_1_14_6
void FASTCALL call_6_1_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_14_7
#define init_6_1_14_7
void FASTCALL call_6_1_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_14_8
#define init_6_1_14_8
void FASTCALL call_6_1_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_14_9
#define init_6_1_14_9
void FASTCALL call_6_1_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_14_10
#define init_6_1_14_10
void FASTCALL call_6_1_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_14_11
#define init_6_1_14_11
void FASTCALL call_6_1_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_14_12
#define init_6_1_14_12
void FASTCALL call_6_1_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_14_13
#define init_6_1_14_13
void FASTCALL call_6_1_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_14_14
#define init_6_1_14_14
void FASTCALL call_6_1_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_14_15
#define init_6_1_14_15
void FASTCALL call_6_1_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_15_0
#define init_6_1_15_0
void FASTCALL call_6_1_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_15_1
#define init_6_1_15_1
void FASTCALL call_6_1_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_15_2
#define init_6_1_15_2
void FASTCALL call_6_1_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_15_3
#define init_6_1_15_3
void FASTCALL call_6_1_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_15_4
#define init_6_1_15_4
void FASTCALL call_6_1_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_15_5
#define init_6_1_15_5
void FASTCALL call_6_1_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_15_6
#define init_6_1_15_6
void FASTCALL call_6_1_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_15_7
#define init_6_1_15_7
void FASTCALL call_6_1_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_15_8
#define init_6_1_15_8
void FASTCALL call_6_1_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_15_9
#define init_6_1_15_9
void FASTCALL call_6_1_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_15_10
#define init_6_1_15_10
void FASTCALL call_6_1_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_15_11
#define init_6_1_15_11
void FASTCALL call_6_1_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_15_12
#define init_6_1_15_12
void FASTCALL call_6_1_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_15_13
#define init_6_1_15_13
void FASTCALL call_6_1_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_15_14
#define init_6_1_15_14
void FASTCALL call_6_1_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_1_15_15
#define init_6_1_15_15
void FASTCALL call_6_1_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_0_0
#define init_6_2_0_0
void FASTCALL call_6_2_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_0_1
#define init_6_2_0_1
void FASTCALL call_6_2_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_0_2
#define init_6_2_0_2
void FASTCALL call_6_2_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_0_3
#define init_6_2_0_3
void FASTCALL call_6_2_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_0_4
#define init_6_2_0_4
void FASTCALL call_6_2_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_0_5
#define init_6_2_0_5
void FASTCALL call_6_2_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_0_6
#define init_6_2_0_6
void FASTCALL call_6_2_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_0_7
#define init_6_2_0_7
void FASTCALL call_6_2_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_0_8
#define init_6_2_0_8
void FASTCALL call_6_2_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_0_9
#define init_6_2_0_9
void FASTCALL call_6_2_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_0_10
#define init_6_2_0_10
void FASTCALL call_6_2_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_0_11
#define init_6_2_0_11
void FASTCALL call_6_2_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_0_12
#define init_6_2_0_12
void FASTCALL call_6_2_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_0_13
#define init_6_2_0_13
void FASTCALL call_6_2_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_0_14
#define init_6_2_0_14
void FASTCALL call_6_2_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_0_15
#define init_6_2_0_15
void FASTCALL call_6_2_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_1_0
#define init_6_2_1_0
void FASTCALL call_6_2_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_1_1
#define init_6_2_1_1
void FASTCALL call_6_2_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_1_2
#define init_6_2_1_2
void FASTCALL call_6_2_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_1_3
#define init_6_2_1_3
void FASTCALL call_6_2_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_1_4
#define init_6_2_1_4
void FASTCALL call_6_2_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_1_5
#define init_6_2_1_5
void FASTCALL call_6_2_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_1_6
#define init_6_2_1_6
void FASTCALL call_6_2_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_1_7
#define init_6_2_1_7
void FASTCALL call_6_2_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_1_8
#define init_6_2_1_8
void FASTCALL call_6_2_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_1_9
#define init_6_2_1_9
void FASTCALL call_6_2_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_1_10
#define init_6_2_1_10
void FASTCALL call_6_2_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_1_11
#define init_6_2_1_11
void FASTCALL call_6_2_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_1_12
#define init_6_2_1_12
void FASTCALL call_6_2_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_1_13
#define init_6_2_1_13
void FASTCALL call_6_2_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_1_14
#define init_6_2_1_14
void FASTCALL call_6_2_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_1_15
#define init_6_2_1_15
void FASTCALL call_6_2_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_2_0
#define init_6_2_2_0
void FASTCALL call_6_2_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_2_1
#define init_6_2_2_1
void FASTCALL call_6_2_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_2_2
#define init_6_2_2_2
void FASTCALL call_6_2_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_2_3
#define init_6_2_2_3
void FASTCALL call_6_2_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_2_4
#define init_6_2_2_4
void FASTCALL call_6_2_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_2_5
#define init_6_2_2_5
void FASTCALL call_6_2_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_2_6
#define init_6_2_2_6
void FASTCALL call_6_2_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_2_7
#define init_6_2_2_7
void FASTCALL call_6_2_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_2_8
#define init_6_2_2_8
void FASTCALL call_6_2_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_2_9
#define init_6_2_2_9
void FASTCALL call_6_2_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_2_10
#define init_6_2_2_10
void FASTCALL call_6_2_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_2_11
#define init_6_2_2_11
void FASTCALL call_6_2_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_2_12
#define init_6_2_2_12
void FASTCALL call_6_2_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_2_13
#define init_6_2_2_13
void FASTCALL call_6_2_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_2_14
#define init_6_2_2_14
void FASTCALL call_6_2_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_2_15
#define init_6_2_2_15
void FASTCALL call_6_2_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_3_0
#define init_6_2_3_0
void FASTCALL call_6_2_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_3_1
#define init_6_2_3_1
void FASTCALL call_6_2_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_3_2
#define init_6_2_3_2
void FASTCALL call_6_2_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_3_3
#define init_6_2_3_3
void FASTCALL call_6_2_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_3_4
#define init_6_2_3_4
void FASTCALL call_6_2_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_3_5
#define init_6_2_3_5
void FASTCALL call_6_2_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_3_6
#define init_6_2_3_6
void FASTCALL call_6_2_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_3_7
#define init_6_2_3_7
void FASTCALL call_6_2_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_3_8
#define init_6_2_3_8
void FASTCALL call_6_2_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_3_9
#define init_6_2_3_9
void FASTCALL call_6_2_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_3_10
#define init_6_2_3_10
void FASTCALL call_6_2_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_3_11
#define init_6_2_3_11
void FASTCALL call_6_2_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_3_12
#define init_6_2_3_12
void FASTCALL call_6_2_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_3_13
#define init_6_2_3_13
void FASTCALL call_6_2_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_3_14
#define init_6_2_3_14
void FASTCALL call_6_2_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_3_15
#define init_6_2_3_15
void FASTCALL call_6_2_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_4_0
#define init_6_2_4_0
void FASTCALL call_6_2_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_4_1
#define init_6_2_4_1
void FASTCALL call_6_2_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_4_2
#define init_6_2_4_2
void FASTCALL call_6_2_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_4_3
#define init_6_2_4_3
void FASTCALL call_6_2_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_4_4
#define init_6_2_4_4
void FASTCALL call_6_2_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_4_5
#define init_6_2_4_5
void FASTCALL call_6_2_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_4_6
#define init_6_2_4_6
void FASTCALL call_6_2_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_4_7
#define init_6_2_4_7
void FASTCALL call_6_2_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_4_8
#define init_6_2_4_8
void FASTCALL call_6_2_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_4_9
#define init_6_2_4_9
void FASTCALL call_6_2_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_4_10
#define init_6_2_4_10
void FASTCALL call_6_2_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_4_11
#define init_6_2_4_11
void FASTCALL call_6_2_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_4_12
#define init_6_2_4_12
void FASTCALL call_6_2_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_4_13
#define init_6_2_4_13
void FASTCALL call_6_2_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_4_14
#define init_6_2_4_14
void FASTCALL call_6_2_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_4_15
#define init_6_2_4_15
void FASTCALL call_6_2_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_5_0
#define init_6_2_5_0
void FASTCALL call_6_2_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_5_1
#define init_6_2_5_1
void FASTCALL call_6_2_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_5_2
#define init_6_2_5_2
void FASTCALL call_6_2_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_5_3
#define init_6_2_5_3
void FASTCALL call_6_2_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_5_4
#define init_6_2_5_4
void FASTCALL call_6_2_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_5_5
#define init_6_2_5_5
void FASTCALL call_6_2_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_5_6
#define init_6_2_5_6
void FASTCALL call_6_2_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_5_7
#define init_6_2_5_7
void FASTCALL call_6_2_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_5_8
#define init_6_2_5_8
void FASTCALL call_6_2_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_5_9
#define init_6_2_5_9
void FASTCALL call_6_2_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_5_10
#define init_6_2_5_10
void FASTCALL call_6_2_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_5_11
#define init_6_2_5_11
void FASTCALL call_6_2_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_5_12
#define init_6_2_5_12
void FASTCALL call_6_2_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_5_13
#define init_6_2_5_13
void FASTCALL call_6_2_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_5_14
#define init_6_2_5_14
void FASTCALL call_6_2_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_5_15
#define init_6_2_5_15
void FASTCALL call_6_2_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_6_0
#define init_6_2_6_0
void FASTCALL call_6_2_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_6_1
#define init_6_2_6_1
void FASTCALL call_6_2_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_6_2
#define init_6_2_6_2
void FASTCALL call_6_2_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_6_3
#define init_6_2_6_3
void FASTCALL call_6_2_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_6_4
#define init_6_2_6_4
void FASTCALL call_6_2_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_6_5
#define init_6_2_6_5
void FASTCALL call_6_2_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_6_6
#define init_6_2_6_6
void FASTCALL call_6_2_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_6_7
#define init_6_2_6_7
void FASTCALL call_6_2_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_6_8
#define init_6_2_6_8
void FASTCALL call_6_2_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_6_9
#define init_6_2_6_9
void FASTCALL call_6_2_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_6_10
#define init_6_2_6_10
void FASTCALL call_6_2_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_6_11
#define init_6_2_6_11
void FASTCALL call_6_2_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_6_12
#define init_6_2_6_12
void FASTCALL call_6_2_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_6_13
#define init_6_2_6_13
void FASTCALL call_6_2_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_6_14
#define init_6_2_6_14
void FASTCALL call_6_2_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_6_15
#define init_6_2_6_15
void FASTCALL call_6_2_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_7_0
#define init_6_2_7_0
void FASTCALL call_6_2_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_7_1
#define init_6_2_7_1
void FASTCALL call_6_2_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_7_2
#define init_6_2_7_2
void FASTCALL call_6_2_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_7_3
#define init_6_2_7_3
void FASTCALL call_6_2_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_7_4
#define init_6_2_7_4
void FASTCALL call_6_2_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_7_5
#define init_6_2_7_5
void FASTCALL call_6_2_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_7_6
#define init_6_2_7_6
void FASTCALL call_6_2_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_7_7
#define init_6_2_7_7
void FASTCALL call_6_2_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_7_8
#define init_6_2_7_8
void FASTCALL call_6_2_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_7_9
#define init_6_2_7_9
void FASTCALL call_6_2_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_7_10
#define init_6_2_7_10
void FASTCALL call_6_2_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_7_11
#define init_6_2_7_11
void FASTCALL call_6_2_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_7_12
#define init_6_2_7_12
void FASTCALL call_6_2_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_7_13
#define init_6_2_7_13
void FASTCALL call_6_2_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_7_14
#define init_6_2_7_14
void FASTCALL call_6_2_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_7_15
#define init_6_2_7_15
void FASTCALL call_6_2_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_8_0
#define init_6_2_8_0
void FASTCALL call_6_2_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_8_1
#define init_6_2_8_1
void FASTCALL call_6_2_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_8_2
#define init_6_2_8_2
void FASTCALL call_6_2_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_8_3
#define init_6_2_8_3
void FASTCALL call_6_2_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_8_4
#define init_6_2_8_4
void FASTCALL call_6_2_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_8_5
#define init_6_2_8_5
void FASTCALL call_6_2_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_8_6
#define init_6_2_8_6
void FASTCALL call_6_2_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_8_7
#define init_6_2_8_7
void FASTCALL call_6_2_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_8_8
#define init_6_2_8_8
void FASTCALL call_6_2_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_8_9
#define init_6_2_8_9
void FASTCALL call_6_2_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_8_10
#define init_6_2_8_10
void FASTCALL call_6_2_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_8_11
#define init_6_2_8_11
void FASTCALL call_6_2_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_8_12
#define init_6_2_8_12
void FASTCALL call_6_2_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_8_13
#define init_6_2_8_13
void FASTCALL call_6_2_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_8_14
#define init_6_2_8_14
void FASTCALL call_6_2_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_8_15
#define init_6_2_8_15
void FASTCALL call_6_2_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_9_0
#define init_6_2_9_0
void FASTCALL call_6_2_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_9_1
#define init_6_2_9_1
void FASTCALL call_6_2_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_9_2
#define init_6_2_9_2
void FASTCALL call_6_2_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_9_3
#define init_6_2_9_3
void FASTCALL call_6_2_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_9_4
#define init_6_2_9_4
void FASTCALL call_6_2_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_9_5
#define init_6_2_9_5
void FASTCALL call_6_2_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_9_6
#define init_6_2_9_6
void FASTCALL call_6_2_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_9_7
#define init_6_2_9_7
void FASTCALL call_6_2_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_9_8
#define init_6_2_9_8
void FASTCALL call_6_2_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_9_9
#define init_6_2_9_9
void FASTCALL call_6_2_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_9_10
#define init_6_2_9_10
void FASTCALL call_6_2_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_9_11
#define init_6_2_9_11
void FASTCALL call_6_2_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_9_12
#define init_6_2_9_12
void FASTCALL call_6_2_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_9_13
#define init_6_2_9_13
void FASTCALL call_6_2_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_9_14
#define init_6_2_9_14
void FASTCALL call_6_2_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_9_15
#define init_6_2_9_15
void FASTCALL call_6_2_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_10_0
#define init_6_2_10_0
void FASTCALL call_6_2_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_10_1
#define init_6_2_10_1
void FASTCALL call_6_2_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_10_2
#define init_6_2_10_2
void FASTCALL call_6_2_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_10_3
#define init_6_2_10_3
void FASTCALL call_6_2_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_10_4
#define init_6_2_10_4
void FASTCALL call_6_2_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_10_5
#define init_6_2_10_5
void FASTCALL call_6_2_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_10_6
#define init_6_2_10_6
void FASTCALL call_6_2_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_10_7
#define init_6_2_10_7
void FASTCALL call_6_2_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_10_8
#define init_6_2_10_8
void FASTCALL call_6_2_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_10_9
#define init_6_2_10_9
void FASTCALL call_6_2_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_10_10
#define init_6_2_10_10
void FASTCALL call_6_2_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_10_11
#define init_6_2_10_11
void FASTCALL call_6_2_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_10_12
#define init_6_2_10_12
void FASTCALL call_6_2_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_10_13
#define init_6_2_10_13
void FASTCALL call_6_2_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_10_14
#define init_6_2_10_14
void FASTCALL call_6_2_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_10_15
#define init_6_2_10_15
void FASTCALL call_6_2_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_11_0
#define init_6_2_11_0
void FASTCALL call_6_2_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_11_1
#define init_6_2_11_1
void FASTCALL call_6_2_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_11_2
#define init_6_2_11_2
void FASTCALL call_6_2_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_11_3
#define init_6_2_11_3
void FASTCALL call_6_2_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_11_4
#define init_6_2_11_4
void FASTCALL call_6_2_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_11_5
#define init_6_2_11_5
void FASTCALL call_6_2_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_11_6
#define init_6_2_11_6
void FASTCALL call_6_2_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_11_7
#define init_6_2_11_7
void FASTCALL call_6_2_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_11_8
#define init_6_2_11_8
void FASTCALL call_6_2_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_11_9
#define init_6_2_11_9
void FASTCALL call_6_2_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_11_10
#define init_6_2_11_10
void FASTCALL call_6_2_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_11_11
#define init_6_2_11_11
void FASTCALL call_6_2_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_11_12
#define init_6_2_11_12
void FASTCALL call_6_2_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_11_13
#define init_6_2_11_13
void FASTCALL call_6_2_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_11_14
#define init_6_2_11_14
void FASTCALL call_6_2_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_11_15
#define init_6_2_11_15
void FASTCALL call_6_2_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_12_0
#define init_6_2_12_0
void FASTCALL call_6_2_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_12_1
#define init_6_2_12_1
void FASTCALL call_6_2_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_12_2
#define init_6_2_12_2
void FASTCALL call_6_2_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_12_3
#define init_6_2_12_3
void FASTCALL call_6_2_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_12_4
#define init_6_2_12_4
void FASTCALL call_6_2_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_12_5
#define init_6_2_12_5
void FASTCALL call_6_2_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_12_6
#define init_6_2_12_6
void FASTCALL call_6_2_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_12_7
#define init_6_2_12_7
void FASTCALL call_6_2_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_12_8
#define init_6_2_12_8
void FASTCALL call_6_2_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_12_9
#define init_6_2_12_9
void FASTCALL call_6_2_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_12_10
#define init_6_2_12_10
void FASTCALL call_6_2_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_12_11
#define init_6_2_12_11
void FASTCALL call_6_2_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_12_12
#define init_6_2_12_12
void FASTCALL call_6_2_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_12_13
#define init_6_2_12_13
void FASTCALL call_6_2_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_12_14
#define init_6_2_12_14
void FASTCALL call_6_2_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_12_15
#define init_6_2_12_15
void FASTCALL call_6_2_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_13_0
#define init_6_2_13_0
void FASTCALL call_6_2_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_13_1
#define init_6_2_13_1
void FASTCALL call_6_2_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_13_2
#define init_6_2_13_2
void FASTCALL call_6_2_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_13_3
#define init_6_2_13_3
void FASTCALL call_6_2_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_13_4
#define init_6_2_13_4
void FASTCALL call_6_2_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_13_5
#define init_6_2_13_5
void FASTCALL call_6_2_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_13_6
#define init_6_2_13_6
void FASTCALL call_6_2_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_13_7
#define init_6_2_13_7
void FASTCALL call_6_2_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_13_8
#define init_6_2_13_8
void FASTCALL call_6_2_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_13_9
#define init_6_2_13_9
void FASTCALL call_6_2_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_13_10
#define init_6_2_13_10
void FASTCALL call_6_2_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_13_11
#define init_6_2_13_11
void FASTCALL call_6_2_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_13_12
#define init_6_2_13_12
void FASTCALL call_6_2_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_13_13
#define init_6_2_13_13
void FASTCALL call_6_2_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_13_14
#define init_6_2_13_14
void FASTCALL call_6_2_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_13_15
#define init_6_2_13_15
void FASTCALL call_6_2_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_14_0
#define init_6_2_14_0
void FASTCALL call_6_2_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_14_1
#define init_6_2_14_1
void FASTCALL call_6_2_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_14_2
#define init_6_2_14_2
void FASTCALL call_6_2_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_14_3
#define init_6_2_14_3
void FASTCALL call_6_2_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_14_4
#define init_6_2_14_4
void FASTCALL call_6_2_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_14_5
#define init_6_2_14_5
void FASTCALL call_6_2_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_14_6
#define init_6_2_14_6
void FASTCALL call_6_2_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_14_7
#define init_6_2_14_7
void FASTCALL call_6_2_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_14_8
#define init_6_2_14_8
void FASTCALL call_6_2_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_14_9
#define init_6_2_14_9
void FASTCALL call_6_2_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_14_10
#define init_6_2_14_10
void FASTCALL call_6_2_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_14_11
#define init_6_2_14_11
void FASTCALL call_6_2_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_14_12
#define init_6_2_14_12
void FASTCALL call_6_2_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_14_13
#define init_6_2_14_13
void FASTCALL call_6_2_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_14_14
#define init_6_2_14_14
void FASTCALL call_6_2_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_14_15
#define init_6_2_14_15
void FASTCALL call_6_2_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_15_0
#define init_6_2_15_0
void FASTCALL call_6_2_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_15_1
#define init_6_2_15_1
void FASTCALL call_6_2_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_15_2
#define init_6_2_15_2
void FASTCALL call_6_2_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_15_3
#define init_6_2_15_3
void FASTCALL call_6_2_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_15_4
#define init_6_2_15_4
void FASTCALL call_6_2_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_15_5
#define init_6_2_15_5
void FASTCALL call_6_2_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_15_6
#define init_6_2_15_6
void FASTCALL call_6_2_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_15_7
#define init_6_2_15_7
void FASTCALL call_6_2_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_15_8
#define init_6_2_15_8
void FASTCALL call_6_2_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_15_9
#define init_6_2_15_9
void FASTCALL call_6_2_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_15_10
#define init_6_2_15_10
void FASTCALL call_6_2_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_15_11
#define init_6_2_15_11
void FASTCALL call_6_2_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_15_12
#define init_6_2_15_12
void FASTCALL call_6_2_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_15_13
#define init_6_2_15_13
void FASTCALL call_6_2_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_15_14
#define init_6_2_15_14
void FASTCALL call_6_2_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_2_15_15
#define init_6_2_15_15
void FASTCALL call_6_2_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_0_0
#define init_6_3_0_0
void FASTCALL call_6_3_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_0_1
#define init_6_3_0_1
void FASTCALL call_6_3_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_0_2
#define init_6_3_0_2
void FASTCALL call_6_3_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_0_3
#define init_6_3_0_3
void FASTCALL call_6_3_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_0_4
#define init_6_3_0_4
void FASTCALL call_6_3_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_0_5
#define init_6_3_0_5
void FASTCALL call_6_3_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_0_6
#define init_6_3_0_6
void FASTCALL call_6_3_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_0_7
#define init_6_3_0_7
void FASTCALL call_6_3_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_0_8
#define init_6_3_0_8
void FASTCALL call_6_3_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_0_9
#define init_6_3_0_9
void FASTCALL call_6_3_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_0_10
#define init_6_3_0_10
void FASTCALL call_6_3_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_0_11
#define init_6_3_0_11
void FASTCALL call_6_3_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_0_12
#define init_6_3_0_12
void FASTCALL call_6_3_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_0_13
#define init_6_3_0_13
void FASTCALL call_6_3_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_0_14
#define init_6_3_0_14
void FASTCALL call_6_3_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_0_15
#define init_6_3_0_15
void FASTCALL call_6_3_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_1_0
#define init_6_3_1_0
void FASTCALL call_6_3_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_1_1
#define init_6_3_1_1
void FASTCALL call_6_3_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_1_2
#define init_6_3_1_2
void FASTCALL call_6_3_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_1_3
#define init_6_3_1_3
void FASTCALL call_6_3_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_1_4
#define init_6_3_1_4
void FASTCALL call_6_3_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_1_5
#define init_6_3_1_5
void FASTCALL call_6_3_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_1_6
#define init_6_3_1_6
void FASTCALL call_6_3_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_1_7
#define init_6_3_1_7
void FASTCALL call_6_3_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_1_8
#define init_6_3_1_8
void FASTCALL call_6_3_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_1_9
#define init_6_3_1_9
void FASTCALL call_6_3_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_1_10
#define init_6_3_1_10
void FASTCALL call_6_3_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_1_11
#define init_6_3_1_11
void FASTCALL call_6_3_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_1_12
#define init_6_3_1_12
void FASTCALL call_6_3_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_1_13
#define init_6_3_1_13
void FASTCALL call_6_3_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_1_14
#define init_6_3_1_14
void FASTCALL call_6_3_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_1_15
#define init_6_3_1_15
void FASTCALL call_6_3_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_2_0
#define init_6_3_2_0
void FASTCALL call_6_3_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_2_1
#define init_6_3_2_1
void FASTCALL call_6_3_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_2_2
#define init_6_3_2_2
void FASTCALL call_6_3_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_2_3
#define init_6_3_2_3
void FASTCALL call_6_3_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_2_4
#define init_6_3_2_4
void FASTCALL call_6_3_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_2_5
#define init_6_3_2_5
void FASTCALL call_6_3_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_2_6
#define init_6_3_2_6
void FASTCALL call_6_3_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_2_7
#define init_6_3_2_7
void FASTCALL call_6_3_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_2_8
#define init_6_3_2_8
void FASTCALL call_6_3_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_2_9
#define init_6_3_2_9
void FASTCALL call_6_3_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_2_10
#define init_6_3_2_10
void FASTCALL call_6_3_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_2_11
#define init_6_3_2_11
void FASTCALL call_6_3_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_2_12
#define init_6_3_2_12
void FASTCALL call_6_3_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_2_13
#define init_6_3_2_13
void FASTCALL call_6_3_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_2_14
#define init_6_3_2_14
void FASTCALL call_6_3_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_2_15
#define init_6_3_2_15
void FASTCALL call_6_3_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_3_0
#define init_6_3_3_0
void FASTCALL call_6_3_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_3_1
#define init_6_3_3_1
void FASTCALL call_6_3_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_3_2
#define init_6_3_3_2
void FASTCALL call_6_3_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_3_3
#define init_6_3_3_3
void FASTCALL call_6_3_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_3_4
#define init_6_3_3_4
void FASTCALL call_6_3_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_3_5
#define init_6_3_3_5
void FASTCALL call_6_3_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_3_6
#define init_6_3_3_6
void FASTCALL call_6_3_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_3_7
#define init_6_3_3_7
void FASTCALL call_6_3_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_3_8
#define init_6_3_3_8
void FASTCALL call_6_3_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_3_9
#define init_6_3_3_9
void FASTCALL call_6_3_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_3_10
#define init_6_3_3_10
void FASTCALL call_6_3_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_3_11
#define init_6_3_3_11
void FASTCALL call_6_3_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_3_12
#define init_6_3_3_12
void FASTCALL call_6_3_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_3_13
#define init_6_3_3_13
void FASTCALL call_6_3_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_3_14
#define init_6_3_3_14
void FASTCALL call_6_3_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_3_15
#define init_6_3_3_15
void FASTCALL call_6_3_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_4_0
#define init_6_3_4_0
void FASTCALL call_6_3_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_4_1
#define init_6_3_4_1
void FASTCALL call_6_3_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_4_2
#define init_6_3_4_2
void FASTCALL call_6_3_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_4_3
#define init_6_3_4_3
void FASTCALL call_6_3_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_4_4
#define init_6_3_4_4
void FASTCALL call_6_3_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_4_5
#define init_6_3_4_5
void FASTCALL call_6_3_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_4_6
#define init_6_3_4_6
void FASTCALL call_6_3_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_4_7
#define init_6_3_4_7
void FASTCALL call_6_3_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_4_8
#define init_6_3_4_8
void FASTCALL call_6_3_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_4_9
#define init_6_3_4_9
void FASTCALL call_6_3_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_4_10
#define init_6_3_4_10
void FASTCALL call_6_3_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_4_11
#define init_6_3_4_11
void FASTCALL call_6_3_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_4_12
#define init_6_3_4_12
void FASTCALL call_6_3_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_4_13
#define init_6_3_4_13
void FASTCALL call_6_3_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_4_14
#define init_6_3_4_14
void FASTCALL call_6_3_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_4_15
#define init_6_3_4_15
void FASTCALL call_6_3_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_5_0
#define init_6_3_5_0
void FASTCALL call_6_3_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_5_1
#define init_6_3_5_1
void FASTCALL call_6_3_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_5_2
#define init_6_3_5_2
void FASTCALL call_6_3_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_5_3
#define init_6_3_5_3
void FASTCALL call_6_3_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_5_4
#define init_6_3_5_4
void FASTCALL call_6_3_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_5_5
#define init_6_3_5_5
void FASTCALL call_6_3_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_5_6
#define init_6_3_5_6
void FASTCALL call_6_3_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_5_7
#define init_6_3_5_7
void FASTCALL call_6_3_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_5_8
#define init_6_3_5_8
void FASTCALL call_6_3_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_5_9
#define init_6_3_5_9
void FASTCALL call_6_3_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_5_10
#define init_6_3_5_10
void FASTCALL call_6_3_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_5_11
#define init_6_3_5_11
void FASTCALL call_6_3_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_5_12
#define init_6_3_5_12
void FASTCALL call_6_3_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_5_13
#define init_6_3_5_13
void FASTCALL call_6_3_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_5_14
#define init_6_3_5_14
void FASTCALL call_6_3_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_5_15
#define init_6_3_5_15
void FASTCALL call_6_3_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_6_0
#define init_6_3_6_0
void FASTCALL call_6_3_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_6_1
#define init_6_3_6_1
void FASTCALL call_6_3_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_6_2
#define init_6_3_6_2
void FASTCALL call_6_3_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_6_3
#define init_6_3_6_3
void FASTCALL call_6_3_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_6_4
#define init_6_3_6_4
void FASTCALL call_6_3_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_6_5
#define init_6_3_6_5
void FASTCALL call_6_3_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_6_6
#define init_6_3_6_6
void FASTCALL call_6_3_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_6_7
#define init_6_3_6_7
void FASTCALL call_6_3_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_6_8
#define init_6_3_6_8
void FASTCALL call_6_3_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_6_9
#define init_6_3_6_9
void FASTCALL call_6_3_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_6_10
#define init_6_3_6_10
void FASTCALL call_6_3_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_6_11
#define init_6_3_6_11
void FASTCALL call_6_3_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_6_12
#define init_6_3_6_12
void FASTCALL call_6_3_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_6_13
#define init_6_3_6_13
void FASTCALL call_6_3_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_6_14
#define init_6_3_6_14
void FASTCALL call_6_3_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_6_15
#define init_6_3_6_15
void FASTCALL call_6_3_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_7_0
#define init_6_3_7_0
void FASTCALL call_6_3_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_7_1
#define init_6_3_7_1
void FASTCALL call_6_3_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_7_2
#define init_6_3_7_2
void FASTCALL call_6_3_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_7_3
#define init_6_3_7_3
void FASTCALL call_6_3_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_7_4
#define init_6_3_7_4
void FASTCALL call_6_3_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_7_5
#define init_6_3_7_5
void FASTCALL call_6_3_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_7_6
#define init_6_3_7_6
void FASTCALL call_6_3_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_7_7
#define init_6_3_7_7
void FASTCALL call_6_3_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_7_8
#define init_6_3_7_8
void FASTCALL call_6_3_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_7_9
#define init_6_3_7_9
void FASTCALL call_6_3_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_7_10
#define init_6_3_7_10
void FASTCALL call_6_3_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_7_11
#define init_6_3_7_11
void FASTCALL call_6_3_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_7_12
#define init_6_3_7_12
void FASTCALL call_6_3_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_7_13
#define init_6_3_7_13
void FASTCALL call_6_3_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_7_14
#define init_6_3_7_14
void FASTCALL call_6_3_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_7_15
#define init_6_3_7_15
void FASTCALL call_6_3_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_8_0
#define init_6_3_8_0
void FASTCALL call_6_3_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_8_1
#define init_6_3_8_1
void FASTCALL call_6_3_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_8_2
#define init_6_3_8_2
void FASTCALL call_6_3_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_8_3
#define init_6_3_8_3
void FASTCALL call_6_3_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_8_4
#define init_6_3_8_4
void FASTCALL call_6_3_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_8_5
#define init_6_3_8_5
void FASTCALL call_6_3_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_8_6
#define init_6_3_8_6
void FASTCALL call_6_3_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_8_7
#define init_6_3_8_7
void FASTCALL call_6_3_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_8_8
#define init_6_3_8_8
void FASTCALL call_6_3_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_8_9
#define init_6_3_8_9
void FASTCALL call_6_3_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_8_10
#define init_6_3_8_10
void FASTCALL call_6_3_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_8_11
#define init_6_3_8_11
void FASTCALL call_6_3_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_8_12
#define init_6_3_8_12
void FASTCALL call_6_3_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_8_13
#define init_6_3_8_13
void FASTCALL call_6_3_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_8_14
#define init_6_3_8_14
void FASTCALL call_6_3_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_8_15
#define init_6_3_8_15
void FASTCALL call_6_3_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_9_0
#define init_6_3_9_0
void FASTCALL call_6_3_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_9_1
#define init_6_3_9_1
void FASTCALL call_6_3_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_9_2
#define init_6_3_9_2
void FASTCALL call_6_3_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_9_3
#define init_6_3_9_3
void FASTCALL call_6_3_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_9_4
#define init_6_3_9_4
void FASTCALL call_6_3_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_9_5
#define init_6_3_9_5
void FASTCALL call_6_3_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_9_6
#define init_6_3_9_6
void FASTCALL call_6_3_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_9_7
#define init_6_3_9_7
void FASTCALL call_6_3_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_9_8
#define init_6_3_9_8
void FASTCALL call_6_3_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_9_9
#define init_6_3_9_9
void FASTCALL call_6_3_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_9_10
#define init_6_3_9_10
void FASTCALL call_6_3_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_9_11
#define init_6_3_9_11
void FASTCALL call_6_3_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_9_12
#define init_6_3_9_12
void FASTCALL call_6_3_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_9_13
#define init_6_3_9_13
void FASTCALL call_6_3_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_9_14
#define init_6_3_9_14
void FASTCALL call_6_3_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_9_15
#define init_6_3_9_15
void FASTCALL call_6_3_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_10_0
#define init_6_3_10_0
void FASTCALL call_6_3_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_10_1
#define init_6_3_10_1
void FASTCALL call_6_3_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_10_2
#define init_6_3_10_2
void FASTCALL call_6_3_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_10_3
#define init_6_3_10_3
void FASTCALL call_6_3_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_10_4
#define init_6_3_10_4
void FASTCALL call_6_3_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_10_5
#define init_6_3_10_5
void FASTCALL call_6_3_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_10_6
#define init_6_3_10_6
void FASTCALL call_6_3_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_10_7
#define init_6_3_10_7
void FASTCALL call_6_3_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_10_8
#define init_6_3_10_8
void FASTCALL call_6_3_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_10_9
#define init_6_3_10_9
void FASTCALL call_6_3_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_10_10
#define init_6_3_10_10
void FASTCALL call_6_3_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_10_11
#define init_6_3_10_11
void FASTCALL call_6_3_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_10_12
#define init_6_3_10_12
void FASTCALL call_6_3_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_10_13
#define init_6_3_10_13
void FASTCALL call_6_3_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_10_14
#define init_6_3_10_14
void FASTCALL call_6_3_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_10_15
#define init_6_3_10_15
void FASTCALL call_6_3_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_11_0
#define init_6_3_11_0
void FASTCALL call_6_3_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_11_1
#define init_6_3_11_1
void FASTCALL call_6_3_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_11_2
#define init_6_3_11_2
void FASTCALL call_6_3_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_11_3
#define init_6_3_11_3
void FASTCALL call_6_3_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_11_4
#define init_6_3_11_4
void FASTCALL call_6_3_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_11_5
#define init_6_3_11_5
void FASTCALL call_6_3_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_11_6
#define init_6_3_11_6
void FASTCALL call_6_3_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_11_7
#define init_6_3_11_7
void FASTCALL call_6_3_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_11_8
#define init_6_3_11_8
void FASTCALL call_6_3_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_11_9
#define init_6_3_11_9
void FASTCALL call_6_3_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_11_10
#define init_6_3_11_10
void FASTCALL call_6_3_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_11_11
#define init_6_3_11_11
void FASTCALL call_6_3_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_11_12
#define init_6_3_11_12
void FASTCALL call_6_3_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_11_13
#define init_6_3_11_13
void FASTCALL call_6_3_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_11_14
#define init_6_3_11_14
void FASTCALL call_6_3_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_11_15
#define init_6_3_11_15
void FASTCALL call_6_3_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_12_0
#define init_6_3_12_0
void FASTCALL call_6_3_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_12_1
#define init_6_3_12_1
void FASTCALL call_6_3_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_12_2
#define init_6_3_12_2
void FASTCALL call_6_3_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_12_3
#define init_6_3_12_3
void FASTCALL call_6_3_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_12_4
#define init_6_3_12_4
void FASTCALL call_6_3_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_12_5
#define init_6_3_12_5
void FASTCALL call_6_3_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_12_6
#define init_6_3_12_6
void FASTCALL call_6_3_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_12_7
#define init_6_3_12_7
void FASTCALL call_6_3_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_12_8
#define init_6_3_12_8
void FASTCALL call_6_3_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_12_9
#define init_6_3_12_9
void FASTCALL call_6_3_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_12_10
#define init_6_3_12_10
void FASTCALL call_6_3_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_12_11
#define init_6_3_12_11
void FASTCALL call_6_3_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_12_12
#define init_6_3_12_12
void FASTCALL call_6_3_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_12_13
#define init_6_3_12_13
void FASTCALL call_6_3_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_12_14
#define init_6_3_12_14
void FASTCALL call_6_3_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_12_15
#define init_6_3_12_15
void FASTCALL call_6_3_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_13_0
#define init_6_3_13_0
void FASTCALL call_6_3_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_13_1
#define init_6_3_13_1
void FASTCALL call_6_3_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_13_2
#define init_6_3_13_2
void FASTCALL call_6_3_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_13_3
#define init_6_3_13_3
void FASTCALL call_6_3_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_13_4
#define init_6_3_13_4
void FASTCALL call_6_3_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_13_5
#define init_6_3_13_5
void FASTCALL call_6_3_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_13_6
#define init_6_3_13_6
void FASTCALL call_6_3_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_13_7
#define init_6_3_13_7
void FASTCALL call_6_3_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_13_8
#define init_6_3_13_8
void FASTCALL call_6_3_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_13_9
#define init_6_3_13_9
void FASTCALL call_6_3_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_13_10
#define init_6_3_13_10
void FASTCALL call_6_3_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_13_11
#define init_6_3_13_11
void FASTCALL call_6_3_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_13_12
#define init_6_3_13_12
void FASTCALL call_6_3_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_13_13
#define init_6_3_13_13
void FASTCALL call_6_3_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_13_14
#define init_6_3_13_14
void FASTCALL call_6_3_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_13_15
#define init_6_3_13_15
void FASTCALL call_6_3_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_14_0
#define init_6_3_14_0
void FASTCALL call_6_3_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_14_1
#define init_6_3_14_1
void FASTCALL call_6_3_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_14_2
#define init_6_3_14_2
void FASTCALL call_6_3_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_14_3
#define init_6_3_14_3
void FASTCALL call_6_3_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_14_4
#define init_6_3_14_4
void FASTCALL call_6_3_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_14_5
#define init_6_3_14_5
void FASTCALL call_6_3_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_14_6
#define init_6_3_14_6
void FASTCALL call_6_3_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_14_7
#define init_6_3_14_7
void FASTCALL call_6_3_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_14_8
#define init_6_3_14_8
void FASTCALL call_6_3_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_14_9
#define init_6_3_14_9
void FASTCALL call_6_3_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_14_10
#define init_6_3_14_10
void FASTCALL call_6_3_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_14_11
#define init_6_3_14_11
void FASTCALL call_6_3_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_14_12
#define init_6_3_14_12
void FASTCALL call_6_3_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_14_13
#define init_6_3_14_13
void FASTCALL call_6_3_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_14_14
#define init_6_3_14_14
void FASTCALL call_6_3_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_14_15
#define init_6_3_14_15
void FASTCALL call_6_3_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_15_0
#define init_6_3_15_0
void FASTCALL call_6_3_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_15_1
#define init_6_3_15_1
void FASTCALL call_6_3_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_15_2
#define init_6_3_15_2
void FASTCALL call_6_3_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_15_3
#define init_6_3_15_3
void FASTCALL call_6_3_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_15_4
#define init_6_3_15_4
void FASTCALL call_6_3_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_15_5
#define init_6_3_15_5
void FASTCALL call_6_3_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_15_6
#define init_6_3_15_6
void FASTCALL call_6_3_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_15_7
#define init_6_3_15_7
void FASTCALL call_6_3_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_15_8
#define init_6_3_15_8
void FASTCALL call_6_3_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_15_9
#define init_6_3_15_9
void FASTCALL call_6_3_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_15_10
#define init_6_3_15_10
void FASTCALL call_6_3_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_15_11
#define init_6_3_15_11
void FASTCALL call_6_3_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_15_12
#define init_6_3_15_12
void FASTCALL call_6_3_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_15_13
#define init_6_3_15_13
void FASTCALL call_6_3_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_15_14
#define init_6_3_15_14
void FASTCALL call_6_3_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_3_15_15
#define init_6_3_15_15
void FASTCALL call_6_3_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_0_0
#define init_6_4_0_0
void FASTCALL call_6_4_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_0_1
#define init_6_4_0_1
void FASTCALL call_6_4_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_0_2
#define init_6_4_0_2
void FASTCALL call_6_4_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_0_3
#define init_6_4_0_3
void FASTCALL call_6_4_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_0_4
#define init_6_4_0_4
void FASTCALL call_6_4_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_0_5
#define init_6_4_0_5
void FASTCALL call_6_4_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_0_6
#define init_6_4_0_6
void FASTCALL call_6_4_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_0_7
#define init_6_4_0_7
void FASTCALL call_6_4_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_0_8
#define init_6_4_0_8
void FASTCALL call_6_4_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_0_9
#define init_6_4_0_9
void FASTCALL call_6_4_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_0_10
#define init_6_4_0_10
void FASTCALL call_6_4_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_0_11
#define init_6_4_0_11
void FASTCALL call_6_4_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_0_12
#define init_6_4_0_12
void FASTCALL call_6_4_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_0_13
#define init_6_4_0_13
void FASTCALL call_6_4_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_0_14
#define init_6_4_0_14
void FASTCALL call_6_4_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_0_15
#define init_6_4_0_15
void FASTCALL call_6_4_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_1_0
#define init_6_4_1_0
void FASTCALL call_6_4_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_1_1
#define init_6_4_1_1
void FASTCALL call_6_4_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_1_2
#define init_6_4_1_2
void FASTCALL call_6_4_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_1_3
#define init_6_4_1_3
void FASTCALL call_6_4_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_1_4
#define init_6_4_1_4
void FASTCALL call_6_4_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_1_5
#define init_6_4_1_5
void FASTCALL call_6_4_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_1_6
#define init_6_4_1_6
void FASTCALL call_6_4_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_1_7
#define init_6_4_1_7
void FASTCALL call_6_4_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_1_8
#define init_6_4_1_8
void FASTCALL call_6_4_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_1_9
#define init_6_4_1_9
void FASTCALL call_6_4_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_1_10
#define init_6_4_1_10
void FASTCALL call_6_4_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_1_11
#define init_6_4_1_11
void FASTCALL call_6_4_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_1_12
#define init_6_4_1_12
void FASTCALL call_6_4_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_1_13
#define init_6_4_1_13
void FASTCALL call_6_4_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_1_14
#define init_6_4_1_14
void FASTCALL call_6_4_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_1_15
#define init_6_4_1_15
void FASTCALL call_6_4_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_2_0
#define init_6_4_2_0
void FASTCALL call_6_4_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_2_1
#define init_6_4_2_1
void FASTCALL call_6_4_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_2_2
#define init_6_4_2_2
void FASTCALL call_6_4_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_2_3
#define init_6_4_2_3
void FASTCALL call_6_4_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_2_4
#define init_6_4_2_4
void FASTCALL call_6_4_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_2_5
#define init_6_4_2_5
void FASTCALL call_6_4_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_2_6
#define init_6_4_2_6
void FASTCALL call_6_4_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_2_7
#define init_6_4_2_7
void FASTCALL call_6_4_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_2_8
#define init_6_4_2_8
void FASTCALL call_6_4_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_2_9
#define init_6_4_2_9
void FASTCALL call_6_4_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_2_10
#define init_6_4_2_10
void FASTCALL call_6_4_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_2_11
#define init_6_4_2_11
void FASTCALL call_6_4_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_2_12
#define init_6_4_2_12
void FASTCALL call_6_4_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_2_13
#define init_6_4_2_13
void FASTCALL call_6_4_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_2_14
#define init_6_4_2_14
void FASTCALL call_6_4_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_2_15
#define init_6_4_2_15
void FASTCALL call_6_4_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_3_0
#define init_6_4_3_0
void FASTCALL call_6_4_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_3_1
#define init_6_4_3_1
void FASTCALL call_6_4_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_3_2
#define init_6_4_3_2
void FASTCALL call_6_4_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_3_3
#define init_6_4_3_3
void FASTCALL call_6_4_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_3_4
#define init_6_4_3_4
void FASTCALL call_6_4_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_3_5
#define init_6_4_3_5
void FASTCALL call_6_4_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_3_6
#define init_6_4_3_6
void FASTCALL call_6_4_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_3_7
#define init_6_4_3_7
void FASTCALL call_6_4_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_3_8
#define init_6_4_3_8
void FASTCALL call_6_4_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_3_9
#define init_6_4_3_9
void FASTCALL call_6_4_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_3_10
#define init_6_4_3_10
void FASTCALL call_6_4_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_3_11
#define init_6_4_3_11
void FASTCALL call_6_4_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_3_12
#define init_6_4_3_12
void FASTCALL call_6_4_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_3_13
#define init_6_4_3_13
void FASTCALL call_6_4_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_3_14
#define init_6_4_3_14
void FASTCALL call_6_4_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_3_15
#define init_6_4_3_15
void FASTCALL call_6_4_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_4_0
#define init_6_4_4_0
void FASTCALL call_6_4_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_4_1
#define init_6_4_4_1
void FASTCALL call_6_4_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_4_2
#define init_6_4_4_2
void FASTCALL call_6_4_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_4_3
#define init_6_4_4_3
void FASTCALL call_6_4_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_4_4
#define init_6_4_4_4
void FASTCALL call_6_4_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_4_5
#define init_6_4_4_5
void FASTCALL call_6_4_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_4_6
#define init_6_4_4_6
void FASTCALL call_6_4_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_4_7
#define init_6_4_4_7
void FASTCALL call_6_4_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_4_8
#define init_6_4_4_8
void FASTCALL call_6_4_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_4_9
#define init_6_4_4_9
void FASTCALL call_6_4_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_4_10
#define init_6_4_4_10
void FASTCALL call_6_4_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_4_11
#define init_6_4_4_11
void FASTCALL call_6_4_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_4_12
#define init_6_4_4_12
void FASTCALL call_6_4_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_4_13
#define init_6_4_4_13
void FASTCALL call_6_4_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_4_14
#define init_6_4_4_14
void FASTCALL call_6_4_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_4_15
#define init_6_4_4_15
void FASTCALL call_6_4_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_5_0
#define init_6_4_5_0
void FASTCALL call_6_4_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_5_1
#define init_6_4_5_1
void FASTCALL call_6_4_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_5_2
#define init_6_4_5_2
void FASTCALL call_6_4_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_5_3
#define init_6_4_5_3
void FASTCALL call_6_4_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_5_4
#define init_6_4_5_4
void FASTCALL call_6_4_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_5_5
#define init_6_4_5_5
void FASTCALL call_6_4_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_5_6
#define init_6_4_5_6
void FASTCALL call_6_4_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_5_7
#define init_6_4_5_7
void FASTCALL call_6_4_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_5_8
#define init_6_4_5_8
void FASTCALL call_6_4_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_5_9
#define init_6_4_5_9
void FASTCALL call_6_4_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_5_10
#define init_6_4_5_10
void FASTCALL call_6_4_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_5_11
#define init_6_4_5_11
void FASTCALL call_6_4_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_5_12
#define init_6_4_5_12
void FASTCALL call_6_4_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_5_13
#define init_6_4_5_13
void FASTCALL call_6_4_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_5_14
#define init_6_4_5_14
void FASTCALL call_6_4_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_5_15
#define init_6_4_5_15
void FASTCALL call_6_4_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_6_0
#define init_6_4_6_0
void FASTCALL call_6_4_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_6_1
#define init_6_4_6_1
void FASTCALL call_6_4_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_6_2
#define init_6_4_6_2
void FASTCALL call_6_4_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_6_3
#define init_6_4_6_3
void FASTCALL call_6_4_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_6_4
#define init_6_4_6_4
void FASTCALL call_6_4_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_6_5
#define init_6_4_6_5
void FASTCALL call_6_4_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_6_6
#define init_6_4_6_6
void FASTCALL call_6_4_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_6_7
#define init_6_4_6_7
void FASTCALL call_6_4_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_6_8
#define init_6_4_6_8
void FASTCALL call_6_4_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_6_9
#define init_6_4_6_9
void FASTCALL call_6_4_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_6_10
#define init_6_4_6_10
void FASTCALL call_6_4_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_6_11
#define init_6_4_6_11
void FASTCALL call_6_4_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_6_12
#define init_6_4_6_12
void FASTCALL call_6_4_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_6_13
#define init_6_4_6_13
void FASTCALL call_6_4_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_6_14
#define init_6_4_6_14
void FASTCALL call_6_4_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_6_15
#define init_6_4_6_15
void FASTCALL call_6_4_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_7_0
#define init_6_4_7_0
void FASTCALL call_6_4_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_7_1
#define init_6_4_7_1
void FASTCALL call_6_4_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_7_2
#define init_6_4_7_2
void FASTCALL call_6_4_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_7_3
#define init_6_4_7_3
void FASTCALL call_6_4_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_7_4
#define init_6_4_7_4
void FASTCALL call_6_4_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_7_5
#define init_6_4_7_5
void FASTCALL call_6_4_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_7_6
#define init_6_4_7_6
void FASTCALL call_6_4_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_7_7
#define init_6_4_7_7
void FASTCALL call_6_4_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_7_8
#define init_6_4_7_8
void FASTCALL call_6_4_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_7_9
#define init_6_4_7_9
void FASTCALL call_6_4_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_7_10
#define init_6_4_7_10
void FASTCALL call_6_4_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_7_11
#define init_6_4_7_11
void FASTCALL call_6_4_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_7_12
#define init_6_4_7_12
void FASTCALL call_6_4_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_7_13
#define init_6_4_7_13
void FASTCALL call_6_4_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_7_14
#define init_6_4_7_14
void FASTCALL call_6_4_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_7_15
#define init_6_4_7_15
void FASTCALL call_6_4_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_8_0
#define init_6_4_8_0
void FASTCALL call_6_4_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_8_1
#define init_6_4_8_1
void FASTCALL call_6_4_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_8_2
#define init_6_4_8_2
void FASTCALL call_6_4_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_8_3
#define init_6_4_8_3
void FASTCALL call_6_4_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_8_4
#define init_6_4_8_4
void FASTCALL call_6_4_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_8_5
#define init_6_4_8_5
void FASTCALL call_6_4_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_8_6
#define init_6_4_8_6
void FASTCALL call_6_4_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_8_7
#define init_6_4_8_7
void FASTCALL call_6_4_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_8_8
#define init_6_4_8_8
void FASTCALL call_6_4_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_8_9
#define init_6_4_8_9
void FASTCALL call_6_4_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_8_10
#define init_6_4_8_10
void FASTCALL call_6_4_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_8_11
#define init_6_4_8_11
void FASTCALL call_6_4_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_8_12
#define init_6_4_8_12
void FASTCALL call_6_4_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_8_13
#define init_6_4_8_13
void FASTCALL call_6_4_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_8_14
#define init_6_4_8_14
void FASTCALL call_6_4_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_8_15
#define init_6_4_8_15
void FASTCALL call_6_4_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_9_0
#define init_6_4_9_0
void FASTCALL call_6_4_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_9_1
#define init_6_4_9_1
void FASTCALL call_6_4_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_9_2
#define init_6_4_9_2
void FASTCALL call_6_4_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_9_3
#define init_6_4_9_3
void FASTCALL call_6_4_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_9_4
#define init_6_4_9_4
void FASTCALL call_6_4_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_9_5
#define init_6_4_9_5
void FASTCALL call_6_4_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_9_6
#define init_6_4_9_6
void FASTCALL call_6_4_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_9_7
#define init_6_4_9_7
void FASTCALL call_6_4_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_9_8
#define init_6_4_9_8
void FASTCALL call_6_4_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_9_9
#define init_6_4_9_9
void FASTCALL call_6_4_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_9_10
#define init_6_4_9_10
void FASTCALL call_6_4_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_9_11
#define init_6_4_9_11
void FASTCALL call_6_4_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_9_12
#define init_6_4_9_12
void FASTCALL call_6_4_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_9_13
#define init_6_4_9_13
void FASTCALL call_6_4_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_9_14
#define init_6_4_9_14
void FASTCALL call_6_4_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_9_15
#define init_6_4_9_15
void FASTCALL call_6_4_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_10_0
#define init_6_4_10_0
void FASTCALL call_6_4_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_10_1
#define init_6_4_10_1
void FASTCALL call_6_4_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_10_2
#define init_6_4_10_2
void FASTCALL call_6_4_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_10_3
#define init_6_4_10_3
void FASTCALL call_6_4_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_10_4
#define init_6_4_10_4
void FASTCALL call_6_4_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_10_5
#define init_6_4_10_5
void FASTCALL call_6_4_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_10_6
#define init_6_4_10_6
void FASTCALL call_6_4_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_10_7
#define init_6_4_10_7
void FASTCALL call_6_4_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_10_8
#define init_6_4_10_8
void FASTCALL call_6_4_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_10_9
#define init_6_4_10_9
void FASTCALL call_6_4_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_10_10
#define init_6_4_10_10
void FASTCALL call_6_4_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_10_11
#define init_6_4_10_11
void FASTCALL call_6_4_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_10_12
#define init_6_4_10_12
void FASTCALL call_6_4_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_10_13
#define init_6_4_10_13
void FASTCALL call_6_4_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_10_14
#define init_6_4_10_14
void FASTCALL call_6_4_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_10_15
#define init_6_4_10_15
void FASTCALL call_6_4_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_11_0
#define init_6_4_11_0
void FASTCALL call_6_4_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_11_1
#define init_6_4_11_1
void FASTCALL call_6_4_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_11_2
#define init_6_4_11_2
void FASTCALL call_6_4_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_11_3
#define init_6_4_11_3
void FASTCALL call_6_4_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_11_4
#define init_6_4_11_4
void FASTCALL call_6_4_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_11_5
#define init_6_4_11_5
void FASTCALL call_6_4_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_11_6
#define init_6_4_11_6
void FASTCALL call_6_4_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_11_7
#define init_6_4_11_7
void FASTCALL call_6_4_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_11_8
#define init_6_4_11_8
void FASTCALL call_6_4_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_11_9
#define init_6_4_11_9
void FASTCALL call_6_4_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_11_10
#define init_6_4_11_10
void FASTCALL call_6_4_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_11_11
#define init_6_4_11_11
void FASTCALL call_6_4_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_11_12
#define init_6_4_11_12
void FASTCALL call_6_4_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_11_13
#define init_6_4_11_13
void FASTCALL call_6_4_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_11_14
#define init_6_4_11_14
void FASTCALL call_6_4_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_11_15
#define init_6_4_11_15
void FASTCALL call_6_4_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_12_0
#define init_6_4_12_0
void FASTCALL call_6_4_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_12_1
#define init_6_4_12_1
void FASTCALL call_6_4_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_12_2
#define init_6_4_12_2
void FASTCALL call_6_4_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_12_3
#define init_6_4_12_3
void FASTCALL call_6_4_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_12_4
#define init_6_4_12_4
void FASTCALL call_6_4_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_12_5
#define init_6_4_12_5
void FASTCALL call_6_4_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_12_6
#define init_6_4_12_6
void FASTCALL call_6_4_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_12_7
#define init_6_4_12_7
void FASTCALL call_6_4_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_12_8
#define init_6_4_12_8
void FASTCALL call_6_4_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_12_9
#define init_6_4_12_9
void FASTCALL call_6_4_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_12_10
#define init_6_4_12_10
void FASTCALL call_6_4_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_12_11
#define init_6_4_12_11
void FASTCALL call_6_4_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_12_12
#define init_6_4_12_12
void FASTCALL call_6_4_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_12_13
#define init_6_4_12_13
void FASTCALL call_6_4_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_12_14
#define init_6_4_12_14
void FASTCALL call_6_4_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_12_15
#define init_6_4_12_15
void FASTCALL call_6_4_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_13_0
#define init_6_4_13_0
void FASTCALL call_6_4_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_13_1
#define init_6_4_13_1
void FASTCALL call_6_4_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_13_2
#define init_6_4_13_2
void FASTCALL call_6_4_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_13_3
#define init_6_4_13_3
void FASTCALL call_6_4_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_13_4
#define init_6_4_13_4
void FASTCALL call_6_4_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_13_5
#define init_6_4_13_5
void FASTCALL call_6_4_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_13_6
#define init_6_4_13_6
void FASTCALL call_6_4_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_13_7
#define init_6_4_13_7
void FASTCALL call_6_4_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_13_8
#define init_6_4_13_8
void FASTCALL call_6_4_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_13_9
#define init_6_4_13_9
void FASTCALL call_6_4_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_13_10
#define init_6_4_13_10
void FASTCALL call_6_4_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_13_11
#define init_6_4_13_11
void FASTCALL call_6_4_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_13_12
#define init_6_4_13_12
void FASTCALL call_6_4_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_13_13
#define init_6_4_13_13
void FASTCALL call_6_4_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_13_14
#define init_6_4_13_14
void FASTCALL call_6_4_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_13_15
#define init_6_4_13_15
void FASTCALL call_6_4_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_14_0
#define init_6_4_14_0
void FASTCALL call_6_4_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_14_1
#define init_6_4_14_1
void FASTCALL call_6_4_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_14_2
#define init_6_4_14_2
void FASTCALL call_6_4_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_14_3
#define init_6_4_14_3
void FASTCALL call_6_4_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_14_4
#define init_6_4_14_4
void FASTCALL call_6_4_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_14_5
#define init_6_4_14_5
void FASTCALL call_6_4_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_14_6
#define init_6_4_14_6
void FASTCALL call_6_4_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_14_7
#define init_6_4_14_7
void FASTCALL call_6_4_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_14_8
#define init_6_4_14_8
void FASTCALL call_6_4_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_14_9
#define init_6_4_14_9
void FASTCALL call_6_4_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_14_10
#define init_6_4_14_10
void FASTCALL call_6_4_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_14_11
#define init_6_4_14_11
void FASTCALL call_6_4_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_14_12
#define init_6_4_14_12
void FASTCALL call_6_4_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_14_13
#define init_6_4_14_13
void FASTCALL call_6_4_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_14_14
#define init_6_4_14_14
void FASTCALL call_6_4_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_14_15
#define init_6_4_14_15
void FASTCALL call_6_4_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_15_0
#define init_6_4_15_0
void FASTCALL call_6_4_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_15_1
#define init_6_4_15_1
void FASTCALL call_6_4_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_15_2
#define init_6_4_15_2
void FASTCALL call_6_4_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_15_3
#define init_6_4_15_3
void FASTCALL call_6_4_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_15_4
#define init_6_4_15_4
void FASTCALL call_6_4_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_15_5
#define init_6_4_15_5
void FASTCALL call_6_4_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_15_6
#define init_6_4_15_6
void FASTCALL call_6_4_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_15_7
#define init_6_4_15_7
void FASTCALL call_6_4_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_15_8
#define init_6_4_15_8
void FASTCALL call_6_4_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_15_9
#define init_6_4_15_9
void FASTCALL call_6_4_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_15_10
#define init_6_4_15_10
void FASTCALL call_6_4_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_15_11
#define init_6_4_15_11
void FASTCALL call_6_4_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_15_12
#define init_6_4_15_12
void FASTCALL call_6_4_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_15_13
#define init_6_4_15_13
void FASTCALL call_6_4_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_15_14
#define init_6_4_15_14
void FASTCALL call_6_4_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_4_15_15
#define init_6_4_15_15
void FASTCALL call_6_4_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_0_0
#define init_6_5_0_0
void FASTCALL call_6_5_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_0_1
#define init_6_5_0_1
void FASTCALL call_6_5_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_0_2
#define init_6_5_0_2
void FASTCALL call_6_5_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_0_3
#define init_6_5_0_3
void FASTCALL call_6_5_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_0_4
#define init_6_5_0_4
void FASTCALL call_6_5_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_0_5
#define init_6_5_0_5
void FASTCALL call_6_5_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_0_6
#define init_6_5_0_6
void FASTCALL call_6_5_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_0_7
#define init_6_5_0_7
void FASTCALL call_6_5_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_0_8
#define init_6_5_0_8
void FASTCALL call_6_5_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_0_9
#define init_6_5_0_9
void FASTCALL call_6_5_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_0_10
#define init_6_5_0_10
void FASTCALL call_6_5_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_0_11
#define init_6_5_0_11
void FASTCALL call_6_5_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_0_12
#define init_6_5_0_12
void FASTCALL call_6_5_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_0_13
#define init_6_5_0_13
void FASTCALL call_6_5_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_0_14
#define init_6_5_0_14
void FASTCALL call_6_5_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_0_15
#define init_6_5_0_15
void FASTCALL call_6_5_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_1_0
#define init_6_5_1_0
void FASTCALL call_6_5_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_1_1
#define init_6_5_1_1
void FASTCALL call_6_5_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_1_2
#define init_6_5_1_2
void FASTCALL call_6_5_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_1_3
#define init_6_5_1_3
void FASTCALL call_6_5_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_1_4
#define init_6_5_1_4
void FASTCALL call_6_5_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_1_5
#define init_6_5_1_5
void FASTCALL call_6_5_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_1_6
#define init_6_5_1_6
void FASTCALL call_6_5_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_1_7
#define init_6_5_1_7
void FASTCALL call_6_5_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_1_8
#define init_6_5_1_8
void FASTCALL call_6_5_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_1_9
#define init_6_5_1_9
void FASTCALL call_6_5_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_1_10
#define init_6_5_1_10
void FASTCALL call_6_5_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_1_11
#define init_6_5_1_11
void FASTCALL call_6_5_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_1_12
#define init_6_5_1_12
void FASTCALL call_6_5_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_1_13
#define init_6_5_1_13
void FASTCALL call_6_5_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_1_14
#define init_6_5_1_14
void FASTCALL call_6_5_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_1_15
#define init_6_5_1_15
void FASTCALL call_6_5_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_2_0
#define init_6_5_2_0
void FASTCALL call_6_5_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_2_1
#define init_6_5_2_1
void FASTCALL call_6_5_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_2_2
#define init_6_5_2_2
void FASTCALL call_6_5_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_2_3
#define init_6_5_2_3
void FASTCALL call_6_5_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_2_4
#define init_6_5_2_4
void FASTCALL call_6_5_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_2_5
#define init_6_5_2_5
void FASTCALL call_6_5_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_2_6
#define init_6_5_2_6
void FASTCALL call_6_5_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_2_7
#define init_6_5_2_7
void FASTCALL call_6_5_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_2_8
#define init_6_5_2_8
void FASTCALL call_6_5_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_2_9
#define init_6_5_2_9
void FASTCALL call_6_5_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_2_10
#define init_6_5_2_10
void FASTCALL call_6_5_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_2_11
#define init_6_5_2_11
void FASTCALL call_6_5_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_2_12
#define init_6_5_2_12
void FASTCALL call_6_5_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_2_13
#define init_6_5_2_13
void FASTCALL call_6_5_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_2_14
#define init_6_5_2_14
void FASTCALL call_6_5_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_2_15
#define init_6_5_2_15
void FASTCALL call_6_5_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_3_0
#define init_6_5_3_0
void FASTCALL call_6_5_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_3_1
#define init_6_5_3_1
void FASTCALL call_6_5_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_3_2
#define init_6_5_3_2
void FASTCALL call_6_5_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_3_3
#define init_6_5_3_3
void FASTCALL call_6_5_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_3_4
#define init_6_5_3_4
void FASTCALL call_6_5_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_3_5
#define init_6_5_3_5
void FASTCALL call_6_5_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_3_6
#define init_6_5_3_6
void FASTCALL call_6_5_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_3_7
#define init_6_5_3_7
void FASTCALL call_6_5_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_3_8
#define init_6_5_3_8
void FASTCALL call_6_5_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_3_9
#define init_6_5_3_9
void FASTCALL call_6_5_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_3_10
#define init_6_5_3_10
void FASTCALL call_6_5_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_3_11
#define init_6_5_3_11
void FASTCALL call_6_5_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_3_12
#define init_6_5_3_12
void FASTCALL call_6_5_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_3_13
#define init_6_5_3_13
void FASTCALL call_6_5_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_3_14
#define init_6_5_3_14
void FASTCALL call_6_5_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_3_15
#define init_6_5_3_15
void FASTCALL call_6_5_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_4_0
#define init_6_5_4_0
void FASTCALL call_6_5_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_4_1
#define init_6_5_4_1
void FASTCALL call_6_5_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_4_2
#define init_6_5_4_2
void FASTCALL call_6_5_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_4_3
#define init_6_5_4_3
void FASTCALL call_6_5_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_4_4
#define init_6_5_4_4
void FASTCALL call_6_5_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_4_5
#define init_6_5_4_5
void FASTCALL call_6_5_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_4_6
#define init_6_5_4_6
void FASTCALL call_6_5_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_4_7
#define init_6_5_4_7
void FASTCALL call_6_5_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_4_8
#define init_6_5_4_8
void FASTCALL call_6_5_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_4_9
#define init_6_5_4_9
void FASTCALL call_6_5_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_4_10
#define init_6_5_4_10
void FASTCALL call_6_5_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_4_11
#define init_6_5_4_11
void FASTCALL call_6_5_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_4_12
#define init_6_5_4_12
void FASTCALL call_6_5_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_4_13
#define init_6_5_4_13
void FASTCALL call_6_5_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_4_14
#define init_6_5_4_14
void FASTCALL call_6_5_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_4_15
#define init_6_5_4_15
void FASTCALL call_6_5_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_5_0
#define init_6_5_5_0
void FASTCALL call_6_5_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_5_1
#define init_6_5_5_1
void FASTCALL call_6_5_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_5_2
#define init_6_5_5_2
void FASTCALL call_6_5_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_5_3
#define init_6_5_5_3
void FASTCALL call_6_5_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_5_4
#define init_6_5_5_4
void FASTCALL call_6_5_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_5_5
#define init_6_5_5_5
void FASTCALL call_6_5_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_5_6
#define init_6_5_5_6
void FASTCALL call_6_5_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_5_7
#define init_6_5_5_7
void FASTCALL call_6_5_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_5_8
#define init_6_5_5_8
void FASTCALL call_6_5_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_5_9
#define init_6_5_5_9
void FASTCALL call_6_5_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_5_10
#define init_6_5_5_10
void FASTCALL call_6_5_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_5_11
#define init_6_5_5_11
void FASTCALL call_6_5_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_5_12
#define init_6_5_5_12
void FASTCALL call_6_5_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_5_13
#define init_6_5_5_13
void FASTCALL call_6_5_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_5_14
#define init_6_5_5_14
void FASTCALL call_6_5_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_5_15
#define init_6_5_5_15
void FASTCALL call_6_5_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_6_0
#define init_6_5_6_0
void FASTCALL call_6_5_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_6_1
#define init_6_5_6_1
void FASTCALL call_6_5_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_6_2
#define init_6_5_6_2
void FASTCALL call_6_5_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_6_3
#define init_6_5_6_3
void FASTCALL call_6_5_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_6_4
#define init_6_5_6_4
void FASTCALL call_6_5_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_6_5
#define init_6_5_6_5
void FASTCALL call_6_5_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_6_6
#define init_6_5_6_6
void FASTCALL call_6_5_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_6_7
#define init_6_5_6_7
void FASTCALL call_6_5_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_6_8
#define init_6_5_6_8
void FASTCALL call_6_5_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_6_9
#define init_6_5_6_9
void FASTCALL call_6_5_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_6_10
#define init_6_5_6_10
void FASTCALL call_6_5_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_6_11
#define init_6_5_6_11
void FASTCALL call_6_5_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_6_12
#define init_6_5_6_12
void FASTCALL call_6_5_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_6_13
#define init_6_5_6_13
void FASTCALL call_6_5_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_6_14
#define init_6_5_6_14
void FASTCALL call_6_5_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_6_15
#define init_6_5_6_15
void FASTCALL call_6_5_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_7_0
#define init_6_5_7_0
void FASTCALL call_6_5_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_7_1
#define init_6_5_7_1
void FASTCALL call_6_5_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_7_2
#define init_6_5_7_2
void FASTCALL call_6_5_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_7_3
#define init_6_5_7_3
void FASTCALL call_6_5_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_7_4
#define init_6_5_7_4
void FASTCALL call_6_5_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_7_5
#define init_6_5_7_5
void FASTCALL call_6_5_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_7_6
#define init_6_5_7_6
void FASTCALL call_6_5_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_7_7
#define init_6_5_7_7
void FASTCALL call_6_5_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_7_8
#define init_6_5_7_8
void FASTCALL call_6_5_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_7_9
#define init_6_5_7_9
void FASTCALL call_6_5_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_7_10
#define init_6_5_7_10
void FASTCALL call_6_5_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_7_11
#define init_6_5_7_11
void FASTCALL call_6_5_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_7_12
#define init_6_5_7_12
void FASTCALL call_6_5_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_7_13
#define init_6_5_7_13
void FASTCALL call_6_5_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_7_14
#define init_6_5_7_14
void FASTCALL call_6_5_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_7_15
#define init_6_5_7_15
void FASTCALL call_6_5_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_8_0
#define init_6_5_8_0
void FASTCALL call_6_5_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_8_1
#define init_6_5_8_1
void FASTCALL call_6_5_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_8_2
#define init_6_5_8_2
void FASTCALL call_6_5_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_8_3
#define init_6_5_8_3
void FASTCALL call_6_5_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_8_4
#define init_6_5_8_4
void FASTCALL call_6_5_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_8_5
#define init_6_5_8_5
void FASTCALL call_6_5_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_8_6
#define init_6_5_8_6
void FASTCALL call_6_5_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_8_7
#define init_6_5_8_7
void FASTCALL call_6_5_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_8_8
#define init_6_5_8_8
void FASTCALL call_6_5_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_8_9
#define init_6_5_8_9
void FASTCALL call_6_5_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_8_10
#define init_6_5_8_10
void FASTCALL call_6_5_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_8_11
#define init_6_5_8_11
void FASTCALL call_6_5_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_8_12
#define init_6_5_8_12
void FASTCALL call_6_5_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_8_13
#define init_6_5_8_13
void FASTCALL call_6_5_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_8_14
#define init_6_5_8_14
void FASTCALL call_6_5_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_8_15
#define init_6_5_8_15
void FASTCALL call_6_5_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_9_0
#define init_6_5_9_0
void FASTCALL call_6_5_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_9_1
#define init_6_5_9_1
void FASTCALL call_6_5_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_9_2
#define init_6_5_9_2
void FASTCALL call_6_5_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_9_3
#define init_6_5_9_3
void FASTCALL call_6_5_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_9_4
#define init_6_5_9_4
void FASTCALL call_6_5_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_9_5
#define init_6_5_9_5
void FASTCALL call_6_5_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_9_6
#define init_6_5_9_6
void FASTCALL call_6_5_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_9_7
#define init_6_5_9_7
void FASTCALL call_6_5_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_9_8
#define init_6_5_9_8
void FASTCALL call_6_5_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_9_9
#define init_6_5_9_9
void FASTCALL call_6_5_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_9_10
#define init_6_5_9_10
void FASTCALL call_6_5_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_9_11
#define init_6_5_9_11
void FASTCALL call_6_5_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_9_12
#define init_6_5_9_12
void FASTCALL call_6_5_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_9_13
#define init_6_5_9_13
void FASTCALL call_6_5_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_9_14
#define init_6_5_9_14
void FASTCALL call_6_5_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_9_15
#define init_6_5_9_15
void FASTCALL call_6_5_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_10_0
#define init_6_5_10_0
void FASTCALL call_6_5_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_10_1
#define init_6_5_10_1
void FASTCALL call_6_5_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_10_2
#define init_6_5_10_2
void FASTCALL call_6_5_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_10_3
#define init_6_5_10_3
void FASTCALL call_6_5_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_10_4
#define init_6_5_10_4
void FASTCALL call_6_5_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_10_5
#define init_6_5_10_5
void FASTCALL call_6_5_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_10_6
#define init_6_5_10_6
void FASTCALL call_6_5_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_10_7
#define init_6_5_10_7
void FASTCALL call_6_5_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_10_8
#define init_6_5_10_8
void FASTCALL call_6_5_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_10_9
#define init_6_5_10_9
void FASTCALL call_6_5_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_10_10
#define init_6_5_10_10
void FASTCALL call_6_5_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_10_11
#define init_6_5_10_11
void FASTCALL call_6_5_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_10_12
#define init_6_5_10_12
void FASTCALL call_6_5_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_10_13
#define init_6_5_10_13
void FASTCALL call_6_5_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_10_14
#define init_6_5_10_14
void FASTCALL call_6_5_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_10_15
#define init_6_5_10_15
void FASTCALL call_6_5_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_11_0
#define init_6_5_11_0
void FASTCALL call_6_5_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_11_1
#define init_6_5_11_1
void FASTCALL call_6_5_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_11_2
#define init_6_5_11_2
void FASTCALL call_6_5_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_11_3
#define init_6_5_11_3
void FASTCALL call_6_5_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_11_4
#define init_6_5_11_4
void FASTCALL call_6_5_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_11_5
#define init_6_5_11_5
void FASTCALL call_6_5_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_11_6
#define init_6_5_11_6
void FASTCALL call_6_5_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_11_7
#define init_6_5_11_7
void FASTCALL call_6_5_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_11_8
#define init_6_5_11_8
void FASTCALL call_6_5_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_11_9
#define init_6_5_11_9
void FASTCALL call_6_5_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_11_10
#define init_6_5_11_10
void FASTCALL call_6_5_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_11_11
#define init_6_5_11_11
void FASTCALL call_6_5_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_11_12
#define init_6_5_11_12
void FASTCALL call_6_5_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_11_13
#define init_6_5_11_13
void FASTCALL call_6_5_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_11_14
#define init_6_5_11_14
void FASTCALL call_6_5_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_11_15
#define init_6_5_11_15
void FASTCALL call_6_5_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_12_0
#define init_6_5_12_0
void FASTCALL call_6_5_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_12_1
#define init_6_5_12_1
void FASTCALL call_6_5_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_12_2
#define init_6_5_12_2
void FASTCALL call_6_5_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_12_3
#define init_6_5_12_3
void FASTCALL call_6_5_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_12_4
#define init_6_5_12_4
void FASTCALL call_6_5_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_12_5
#define init_6_5_12_5
void FASTCALL call_6_5_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_12_6
#define init_6_5_12_6
void FASTCALL call_6_5_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_12_7
#define init_6_5_12_7
void FASTCALL call_6_5_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_12_8
#define init_6_5_12_8
void FASTCALL call_6_5_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_12_9
#define init_6_5_12_9
void FASTCALL call_6_5_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_12_10
#define init_6_5_12_10
void FASTCALL call_6_5_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_12_11
#define init_6_5_12_11
void FASTCALL call_6_5_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_12_12
#define init_6_5_12_12
void FASTCALL call_6_5_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_12_13
#define init_6_5_12_13
void FASTCALL call_6_5_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_12_14
#define init_6_5_12_14
void FASTCALL call_6_5_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_12_15
#define init_6_5_12_15
void FASTCALL call_6_5_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_13_0
#define init_6_5_13_0
void FASTCALL call_6_5_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_13_1
#define init_6_5_13_1
void FASTCALL call_6_5_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_13_2
#define init_6_5_13_2
void FASTCALL call_6_5_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_13_3
#define init_6_5_13_3
void FASTCALL call_6_5_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_13_4
#define init_6_5_13_4
void FASTCALL call_6_5_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_13_5
#define init_6_5_13_5
void FASTCALL call_6_5_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_13_6
#define init_6_5_13_6
void FASTCALL call_6_5_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_13_7
#define init_6_5_13_7
void FASTCALL call_6_5_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_13_8
#define init_6_5_13_8
void FASTCALL call_6_5_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_13_9
#define init_6_5_13_9
void FASTCALL call_6_5_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_13_10
#define init_6_5_13_10
void FASTCALL call_6_5_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_13_11
#define init_6_5_13_11
void FASTCALL call_6_5_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_13_12
#define init_6_5_13_12
void FASTCALL call_6_5_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_13_13
#define init_6_5_13_13
void FASTCALL call_6_5_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_13_14
#define init_6_5_13_14
void FASTCALL call_6_5_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_13_15
#define init_6_5_13_15
void FASTCALL call_6_5_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_14_0
#define init_6_5_14_0
void FASTCALL call_6_5_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_14_1
#define init_6_5_14_1
void FASTCALL call_6_5_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_14_2
#define init_6_5_14_2
void FASTCALL call_6_5_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_14_3
#define init_6_5_14_3
void FASTCALL call_6_5_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_14_4
#define init_6_5_14_4
void FASTCALL call_6_5_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_14_5
#define init_6_5_14_5
void FASTCALL call_6_5_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_14_6
#define init_6_5_14_6
void FASTCALL call_6_5_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_14_7
#define init_6_5_14_7
void FASTCALL call_6_5_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_14_8
#define init_6_5_14_8
void FASTCALL call_6_5_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_14_9
#define init_6_5_14_9
void FASTCALL call_6_5_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_14_10
#define init_6_5_14_10
void FASTCALL call_6_5_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_14_11
#define init_6_5_14_11
void FASTCALL call_6_5_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_14_12
#define init_6_5_14_12
void FASTCALL call_6_5_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_14_13
#define init_6_5_14_13
void FASTCALL call_6_5_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_14_14
#define init_6_5_14_14
void FASTCALL call_6_5_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_14_15
#define init_6_5_14_15
void FASTCALL call_6_5_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_15_0
#define init_6_5_15_0
void FASTCALL call_6_5_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_15_1
#define init_6_5_15_1
void FASTCALL call_6_5_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_15_2
#define init_6_5_15_2
void FASTCALL call_6_5_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_15_3
#define init_6_5_15_3
void FASTCALL call_6_5_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_15_4
#define init_6_5_15_4
void FASTCALL call_6_5_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_15_5
#define init_6_5_15_5
void FASTCALL call_6_5_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_15_6
#define init_6_5_15_6
void FASTCALL call_6_5_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_15_7
#define init_6_5_15_7
void FASTCALL call_6_5_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_15_8
#define init_6_5_15_8
void FASTCALL call_6_5_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_15_9
#define init_6_5_15_9
void FASTCALL call_6_5_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_15_10
#define init_6_5_15_10
void FASTCALL call_6_5_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_15_11
#define init_6_5_15_11
void FASTCALL call_6_5_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_15_12
#define init_6_5_15_12
void FASTCALL call_6_5_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_15_13
#define init_6_5_15_13
void FASTCALL call_6_5_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_15_14
#define init_6_5_15_14
void FASTCALL call_6_5_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_5_15_15
#define init_6_5_15_15
void FASTCALL call_6_5_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_0_0
#define init_6_6_0_0
void FASTCALL call_6_6_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_0_1
#define init_6_6_0_1
void FASTCALL call_6_6_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_0_2
#define init_6_6_0_2
void FASTCALL call_6_6_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_0_3
#define init_6_6_0_3
void FASTCALL call_6_6_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_0_4
#define init_6_6_0_4
void FASTCALL call_6_6_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_0_5
#define init_6_6_0_5
void FASTCALL call_6_6_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_0_6
#define init_6_6_0_6
void FASTCALL call_6_6_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_0_7
#define init_6_6_0_7
void FASTCALL call_6_6_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_0_8
#define init_6_6_0_8
void FASTCALL call_6_6_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_0_9
#define init_6_6_0_9
void FASTCALL call_6_6_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_0_10
#define init_6_6_0_10
void FASTCALL call_6_6_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_0_11
#define init_6_6_0_11
void FASTCALL call_6_6_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_0_12
#define init_6_6_0_12
void FASTCALL call_6_6_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_0_13
#define init_6_6_0_13
void FASTCALL call_6_6_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_0_14
#define init_6_6_0_14
void FASTCALL call_6_6_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_0_15
#define init_6_6_0_15
void FASTCALL call_6_6_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_1_0
#define init_6_6_1_0
void FASTCALL call_6_6_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_1_1
#define init_6_6_1_1
void FASTCALL call_6_6_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_1_2
#define init_6_6_1_2
void FASTCALL call_6_6_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_1_3
#define init_6_6_1_3
void FASTCALL call_6_6_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_1_4
#define init_6_6_1_4
void FASTCALL call_6_6_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_1_5
#define init_6_6_1_5
void FASTCALL call_6_6_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_1_6
#define init_6_6_1_6
void FASTCALL call_6_6_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_1_7
#define init_6_6_1_7
void FASTCALL call_6_6_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_1_8
#define init_6_6_1_8
void FASTCALL call_6_6_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_1_9
#define init_6_6_1_9
void FASTCALL call_6_6_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_1_10
#define init_6_6_1_10
void FASTCALL call_6_6_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_1_11
#define init_6_6_1_11
void FASTCALL call_6_6_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_1_12
#define init_6_6_1_12
void FASTCALL call_6_6_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_1_13
#define init_6_6_1_13
void FASTCALL call_6_6_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_1_14
#define init_6_6_1_14
void FASTCALL call_6_6_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_1_15
#define init_6_6_1_15
void FASTCALL call_6_6_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_2_0
#define init_6_6_2_0
void FASTCALL call_6_6_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_2_1
#define init_6_6_2_1
void FASTCALL call_6_6_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_2_2
#define init_6_6_2_2
void FASTCALL call_6_6_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_2_3
#define init_6_6_2_3
void FASTCALL call_6_6_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_2_4
#define init_6_6_2_4
void FASTCALL call_6_6_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_2_5
#define init_6_6_2_5
void FASTCALL call_6_6_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_2_6
#define init_6_6_2_6
void FASTCALL call_6_6_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_2_7
#define init_6_6_2_7
void FASTCALL call_6_6_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_2_8
#define init_6_6_2_8
void FASTCALL call_6_6_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_2_9
#define init_6_6_2_9
void FASTCALL call_6_6_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_2_10
#define init_6_6_2_10
void FASTCALL call_6_6_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_2_11
#define init_6_6_2_11
void FASTCALL call_6_6_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_2_12
#define init_6_6_2_12
void FASTCALL call_6_6_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_2_13
#define init_6_6_2_13
void FASTCALL call_6_6_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_2_14
#define init_6_6_2_14
void FASTCALL call_6_6_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_2_15
#define init_6_6_2_15
void FASTCALL call_6_6_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_3_0
#define init_6_6_3_0
void FASTCALL call_6_6_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_3_1
#define init_6_6_3_1
void FASTCALL call_6_6_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_3_2
#define init_6_6_3_2
void FASTCALL call_6_6_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_3_3
#define init_6_6_3_3
void FASTCALL call_6_6_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_3_4
#define init_6_6_3_4
void FASTCALL call_6_6_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_3_5
#define init_6_6_3_5
void FASTCALL call_6_6_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_3_6
#define init_6_6_3_6
void FASTCALL call_6_6_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_3_7
#define init_6_6_3_7
void FASTCALL call_6_6_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_3_8
#define init_6_6_3_8
void FASTCALL call_6_6_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_3_9
#define init_6_6_3_9
void FASTCALL call_6_6_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_3_10
#define init_6_6_3_10
void FASTCALL call_6_6_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_3_11
#define init_6_6_3_11
void FASTCALL call_6_6_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_3_12
#define init_6_6_3_12
void FASTCALL call_6_6_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_3_13
#define init_6_6_3_13
void FASTCALL call_6_6_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_3_14
#define init_6_6_3_14
void FASTCALL call_6_6_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_3_15
#define init_6_6_3_15
void FASTCALL call_6_6_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_4_0
#define init_6_6_4_0
void FASTCALL call_6_6_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_4_1
#define init_6_6_4_1
void FASTCALL call_6_6_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_4_2
#define init_6_6_4_2
void FASTCALL call_6_6_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_4_3
#define init_6_6_4_3
void FASTCALL call_6_6_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_4_4
#define init_6_6_4_4
void FASTCALL call_6_6_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_4_5
#define init_6_6_4_5
void FASTCALL call_6_6_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_4_6
#define init_6_6_4_6
void FASTCALL call_6_6_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_4_7
#define init_6_6_4_7
void FASTCALL call_6_6_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_4_8
#define init_6_6_4_8
void FASTCALL call_6_6_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_4_9
#define init_6_6_4_9
void FASTCALL call_6_6_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_4_10
#define init_6_6_4_10
void FASTCALL call_6_6_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_4_11
#define init_6_6_4_11
void FASTCALL call_6_6_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_4_12
#define init_6_6_4_12
void FASTCALL call_6_6_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_4_13
#define init_6_6_4_13
void FASTCALL call_6_6_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_4_14
#define init_6_6_4_14
void FASTCALL call_6_6_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_4_15
#define init_6_6_4_15
void FASTCALL call_6_6_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_5_0
#define init_6_6_5_0
void FASTCALL call_6_6_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_5_1
#define init_6_6_5_1
void FASTCALL call_6_6_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_5_2
#define init_6_6_5_2
void FASTCALL call_6_6_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_5_3
#define init_6_6_5_3
void FASTCALL call_6_6_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_5_4
#define init_6_6_5_4
void FASTCALL call_6_6_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_5_5
#define init_6_6_5_5
void FASTCALL call_6_6_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_5_6
#define init_6_6_5_6
void FASTCALL call_6_6_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_5_7
#define init_6_6_5_7
void FASTCALL call_6_6_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_5_8
#define init_6_6_5_8
void FASTCALL call_6_6_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_5_9
#define init_6_6_5_9
void FASTCALL call_6_6_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_5_10
#define init_6_6_5_10
void FASTCALL call_6_6_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_5_11
#define init_6_6_5_11
void FASTCALL call_6_6_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_5_12
#define init_6_6_5_12
void FASTCALL call_6_6_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_5_13
#define init_6_6_5_13
void FASTCALL call_6_6_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_5_14
#define init_6_6_5_14
void FASTCALL call_6_6_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_5_15
#define init_6_6_5_15
void FASTCALL call_6_6_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_6_0
#define init_6_6_6_0
void FASTCALL call_6_6_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_6_1
#define init_6_6_6_1
void FASTCALL call_6_6_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_6_2
#define init_6_6_6_2
void FASTCALL call_6_6_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_6_3
#define init_6_6_6_3
void FASTCALL call_6_6_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_6_4
#define init_6_6_6_4
void FASTCALL call_6_6_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_6_5
#define init_6_6_6_5
void FASTCALL call_6_6_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_6_6
#define init_6_6_6_6
void FASTCALL call_6_6_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_6_7
#define init_6_6_6_7
void FASTCALL call_6_6_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_6_8
#define init_6_6_6_8
void FASTCALL call_6_6_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_6_9
#define init_6_6_6_9
void FASTCALL call_6_6_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_6_10
#define init_6_6_6_10
void FASTCALL call_6_6_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_6_11
#define init_6_6_6_11
void FASTCALL call_6_6_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_6_12
#define init_6_6_6_12
void FASTCALL call_6_6_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_6_13
#define init_6_6_6_13
void FASTCALL call_6_6_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_6_14
#define init_6_6_6_14
void FASTCALL call_6_6_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_6_15
#define init_6_6_6_15
void FASTCALL call_6_6_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_7_0
#define init_6_6_7_0
void FASTCALL call_6_6_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_7_1
#define init_6_6_7_1
void FASTCALL call_6_6_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_7_2
#define init_6_6_7_2
void FASTCALL call_6_6_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_7_3
#define init_6_6_7_3
void FASTCALL call_6_6_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_7_4
#define init_6_6_7_4
void FASTCALL call_6_6_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_7_5
#define init_6_6_7_5
void FASTCALL call_6_6_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_7_6
#define init_6_6_7_6
void FASTCALL call_6_6_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_7_7
#define init_6_6_7_7
void FASTCALL call_6_6_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_7_8
#define init_6_6_7_8
void FASTCALL call_6_6_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_7_9
#define init_6_6_7_9
void FASTCALL call_6_6_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_7_10
#define init_6_6_7_10
void FASTCALL call_6_6_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_7_11
#define init_6_6_7_11
void FASTCALL call_6_6_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_7_12
#define init_6_6_7_12
void FASTCALL call_6_6_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_7_13
#define init_6_6_7_13
void FASTCALL call_6_6_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_7_14
#define init_6_6_7_14
void FASTCALL call_6_6_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_7_15
#define init_6_6_7_15
void FASTCALL call_6_6_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_8_0
#define init_6_6_8_0
void FASTCALL call_6_6_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_8_1
#define init_6_6_8_1
void FASTCALL call_6_6_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_8_2
#define init_6_6_8_2
void FASTCALL call_6_6_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_8_3
#define init_6_6_8_3
void FASTCALL call_6_6_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_8_4
#define init_6_6_8_4
void FASTCALL call_6_6_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_8_5
#define init_6_6_8_5
void FASTCALL call_6_6_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_8_6
#define init_6_6_8_6
void FASTCALL call_6_6_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_8_7
#define init_6_6_8_7
void FASTCALL call_6_6_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_8_8
#define init_6_6_8_8
void FASTCALL call_6_6_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_8_9
#define init_6_6_8_9
void FASTCALL call_6_6_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_8_10
#define init_6_6_8_10
void FASTCALL call_6_6_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_8_11
#define init_6_6_8_11
void FASTCALL call_6_6_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_8_12
#define init_6_6_8_12
void FASTCALL call_6_6_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_8_13
#define init_6_6_8_13
void FASTCALL call_6_6_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_8_14
#define init_6_6_8_14
void FASTCALL call_6_6_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_8_15
#define init_6_6_8_15
void FASTCALL call_6_6_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_9_0
#define init_6_6_9_0
void FASTCALL call_6_6_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_9_1
#define init_6_6_9_1
void FASTCALL call_6_6_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_9_2
#define init_6_6_9_2
void FASTCALL call_6_6_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_9_3
#define init_6_6_9_3
void FASTCALL call_6_6_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_9_4
#define init_6_6_9_4
void FASTCALL call_6_6_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_9_5
#define init_6_6_9_5
void FASTCALL call_6_6_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_9_6
#define init_6_6_9_6
void FASTCALL call_6_6_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_9_7
#define init_6_6_9_7
void FASTCALL call_6_6_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_9_8
#define init_6_6_9_8
void FASTCALL call_6_6_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_9_9
#define init_6_6_9_9
void FASTCALL call_6_6_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_9_10
#define init_6_6_9_10
void FASTCALL call_6_6_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_9_11
#define init_6_6_9_11
void FASTCALL call_6_6_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_9_12
#define init_6_6_9_12
void FASTCALL call_6_6_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_9_13
#define init_6_6_9_13
void FASTCALL call_6_6_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_9_14
#define init_6_6_9_14
void FASTCALL call_6_6_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_9_15
#define init_6_6_9_15
void FASTCALL call_6_6_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_10_0
#define init_6_6_10_0
void FASTCALL call_6_6_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_10_1
#define init_6_6_10_1
void FASTCALL call_6_6_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_10_2
#define init_6_6_10_2
void FASTCALL call_6_6_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_10_3
#define init_6_6_10_3
void FASTCALL call_6_6_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_10_4
#define init_6_6_10_4
void FASTCALL call_6_6_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_10_5
#define init_6_6_10_5
void FASTCALL call_6_6_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_10_6
#define init_6_6_10_6
void FASTCALL call_6_6_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_10_7
#define init_6_6_10_7
void FASTCALL call_6_6_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_10_8
#define init_6_6_10_8
void FASTCALL call_6_6_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_10_9
#define init_6_6_10_9
void FASTCALL call_6_6_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_10_10
#define init_6_6_10_10
void FASTCALL call_6_6_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_10_11
#define init_6_6_10_11
void FASTCALL call_6_6_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_10_12
#define init_6_6_10_12
void FASTCALL call_6_6_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_10_13
#define init_6_6_10_13
void FASTCALL call_6_6_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_10_14
#define init_6_6_10_14
void FASTCALL call_6_6_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_10_15
#define init_6_6_10_15
void FASTCALL call_6_6_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_11_0
#define init_6_6_11_0
void FASTCALL call_6_6_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_11_1
#define init_6_6_11_1
void FASTCALL call_6_6_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_11_2
#define init_6_6_11_2
void FASTCALL call_6_6_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_11_3
#define init_6_6_11_3
void FASTCALL call_6_6_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_11_4
#define init_6_6_11_4
void FASTCALL call_6_6_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_11_5
#define init_6_6_11_5
void FASTCALL call_6_6_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_11_6
#define init_6_6_11_6
void FASTCALL call_6_6_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_11_7
#define init_6_6_11_7
void FASTCALL call_6_6_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_11_8
#define init_6_6_11_8
void FASTCALL call_6_6_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_11_9
#define init_6_6_11_9
void FASTCALL call_6_6_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_11_10
#define init_6_6_11_10
void FASTCALL call_6_6_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_11_11
#define init_6_6_11_11
void FASTCALL call_6_6_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_11_12
#define init_6_6_11_12
void FASTCALL call_6_6_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_11_13
#define init_6_6_11_13
void FASTCALL call_6_6_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_11_14
#define init_6_6_11_14
void FASTCALL call_6_6_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_11_15
#define init_6_6_11_15
void FASTCALL call_6_6_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_12_0
#define init_6_6_12_0
void FASTCALL call_6_6_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_12_1
#define init_6_6_12_1
void FASTCALL call_6_6_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_12_2
#define init_6_6_12_2
void FASTCALL call_6_6_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_12_3
#define init_6_6_12_3
void FASTCALL call_6_6_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_12_4
#define init_6_6_12_4
void FASTCALL call_6_6_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_12_5
#define init_6_6_12_5
void FASTCALL call_6_6_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_12_6
#define init_6_6_12_6
void FASTCALL call_6_6_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_12_7
#define init_6_6_12_7
void FASTCALL call_6_6_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_12_8
#define init_6_6_12_8
void FASTCALL call_6_6_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_12_9
#define init_6_6_12_9
void FASTCALL call_6_6_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_12_10
#define init_6_6_12_10
void FASTCALL call_6_6_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_12_11
#define init_6_6_12_11
void FASTCALL call_6_6_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_12_12
#define init_6_6_12_12
void FASTCALL call_6_6_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_12_13
#define init_6_6_12_13
void FASTCALL call_6_6_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_12_14
#define init_6_6_12_14
void FASTCALL call_6_6_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_12_15
#define init_6_6_12_15
void FASTCALL call_6_6_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_13_0
#define init_6_6_13_0
void FASTCALL call_6_6_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_13_1
#define init_6_6_13_1
void FASTCALL call_6_6_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_13_2
#define init_6_6_13_2
void FASTCALL call_6_6_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_13_3
#define init_6_6_13_3
void FASTCALL call_6_6_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_13_4
#define init_6_6_13_4
void FASTCALL call_6_6_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_13_5
#define init_6_6_13_5
void FASTCALL call_6_6_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_13_6
#define init_6_6_13_6
void FASTCALL call_6_6_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_13_7
#define init_6_6_13_7
void FASTCALL call_6_6_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_13_8
#define init_6_6_13_8
void FASTCALL call_6_6_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_13_9
#define init_6_6_13_9
void FASTCALL call_6_6_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_13_10
#define init_6_6_13_10
void FASTCALL call_6_6_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_13_11
#define init_6_6_13_11
void FASTCALL call_6_6_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_13_12
#define init_6_6_13_12
void FASTCALL call_6_6_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_13_13
#define init_6_6_13_13
void FASTCALL call_6_6_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_13_14
#define init_6_6_13_14
void FASTCALL call_6_6_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_13_15
#define init_6_6_13_15
void FASTCALL call_6_6_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_14_0
#define init_6_6_14_0
void FASTCALL call_6_6_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_14_1
#define init_6_6_14_1
void FASTCALL call_6_6_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_14_2
#define init_6_6_14_2
void FASTCALL call_6_6_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_14_3
#define init_6_6_14_3
void FASTCALL call_6_6_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_14_4
#define init_6_6_14_4
void FASTCALL call_6_6_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_14_5
#define init_6_6_14_5
void FASTCALL call_6_6_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_14_6
#define init_6_6_14_6
void FASTCALL call_6_6_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_14_7
#define init_6_6_14_7
void FASTCALL call_6_6_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_14_8
#define init_6_6_14_8
void FASTCALL call_6_6_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_14_9
#define init_6_6_14_9
void FASTCALL call_6_6_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_14_10
#define init_6_6_14_10
void FASTCALL call_6_6_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_14_11
#define init_6_6_14_11
void FASTCALL call_6_6_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_14_12
#define init_6_6_14_12
void FASTCALL call_6_6_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_14_13
#define init_6_6_14_13
void FASTCALL call_6_6_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_14_14
#define init_6_6_14_14
void FASTCALL call_6_6_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_14_15
#define init_6_6_14_15
void FASTCALL call_6_6_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_15_0
#define init_6_6_15_0
void FASTCALL call_6_6_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_15_1
#define init_6_6_15_1
void FASTCALL call_6_6_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_15_2
#define init_6_6_15_2
void FASTCALL call_6_6_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_15_3
#define init_6_6_15_3
void FASTCALL call_6_6_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_15_4
#define init_6_6_15_4
void FASTCALL call_6_6_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_15_5
#define init_6_6_15_5
void FASTCALL call_6_6_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_15_6
#define init_6_6_15_6
void FASTCALL call_6_6_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_15_7
#define init_6_6_15_7
void FASTCALL call_6_6_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_15_8
#define init_6_6_15_8
void FASTCALL call_6_6_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_15_9
#define init_6_6_15_9
void FASTCALL call_6_6_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_15_10
#define init_6_6_15_10
void FASTCALL call_6_6_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_15_11
#define init_6_6_15_11
void FASTCALL call_6_6_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_15_12
#define init_6_6_15_12
void FASTCALL call_6_6_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_15_13
#define init_6_6_15_13
void FASTCALL call_6_6_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_15_14
#define init_6_6_15_14
void FASTCALL call_6_6_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_6_15_15
#define init_6_6_15_15
void FASTCALL call_6_6_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_0_0
#define init_6_7_0_0
void FASTCALL call_6_7_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_0_1
#define init_6_7_0_1
void FASTCALL call_6_7_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_0_2
#define init_6_7_0_2
void FASTCALL call_6_7_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_0_3
#define init_6_7_0_3
void FASTCALL call_6_7_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_0_4
#define init_6_7_0_4
void FASTCALL call_6_7_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_0_5
#define init_6_7_0_5
void FASTCALL call_6_7_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_0_6
#define init_6_7_0_6
void FASTCALL call_6_7_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_0_7
#define init_6_7_0_7
void FASTCALL call_6_7_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_0_8
#define init_6_7_0_8
void FASTCALL call_6_7_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_0_9
#define init_6_7_0_9
void FASTCALL call_6_7_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_0_10
#define init_6_7_0_10
void FASTCALL call_6_7_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_0_11
#define init_6_7_0_11
void FASTCALL call_6_7_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_0_12
#define init_6_7_0_12
void FASTCALL call_6_7_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_0_13
#define init_6_7_0_13
void FASTCALL call_6_7_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_0_14
#define init_6_7_0_14
void FASTCALL call_6_7_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_0_15
#define init_6_7_0_15
void FASTCALL call_6_7_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_1_0
#define init_6_7_1_0
void FASTCALL call_6_7_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_1_1
#define init_6_7_1_1
void FASTCALL call_6_7_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_1_2
#define init_6_7_1_2
void FASTCALL call_6_7_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_1_3
#define init_6_7_1_3
void FASTCALL call_6_7_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_1_4
#define init_6_7_1_4
void FASTCALL call_6_7_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_1_5
#define init_6_7_1_5
void FASTCALL call_6_7_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_1_6
#define init_6_7_1_6
void FASTCALL call_6_7_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_1_7
#define init_6_7_1_7
void FASTCALL call_6_7_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_1_8
#define init_6_7_1_8
void FASTCALL call_6_7_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_1_9
#define init_6_7_1_9
void FASTCALL call_6_7_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_1_10
#define init_6_7_1_10
void FASTCALL call_6_7_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_1_11
#define init_6_7_1_11
void FASTCALL call_6_7_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_1_12
#define init_6_7_1_12
void FASTCALL call_6_7_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_1_13
#define init_6_7_1_13
void FASTCALL call_6_7_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_1_14
#define init_6_7_1_14
void FASTCALL call_6_7_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_1_15
#define init_6_7_1_15
void FASTCALL call_6_7_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_2_0
#define init_6_7_2_0
void FASTCALL call_6_7_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_2_1
#define init_6_7_2_1
void FASTCALL call_6_7_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_2_2
#define init_6_7_2_2
void FASTCALL call_6_7_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_2_3
#define init_6_7_2_3
void FASTCALL call_6_7_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_2_4
#define init_6_7_2_4
void FASTCALL call_6_7_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_2_5
#define init_6_7_2_5
void FASTCALL call_6_7_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_2_6
#define init_6_7_2_6
void FASTCALL call_6_7_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_2_7
#define init_6_7_2_7
void FASTCALL call_6_7_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_2_8
#define init_6_7_2_8
void FASTCALL call_6_7_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_2_9
#define init_6_7_2_9
void FASTCALL call_6_7_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_2_10
#define init_6_7_2_10
void FASTCALL call_6_7_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_2_11
#define init_6_7_2_11
void FASTCALL call_6_7_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_2_12
#define init_6_7_2_12
void FASTCALL call_6_7_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_2_13
#define init_6_7_2_13
void FASTCALL call_6_7_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_2_14
#define init_6_7_2_14
void FASTCALL call_6_7_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_2_15
#define init_6_7_2_15
void FASTCALL call_6_7_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_3_0
#define init_6_7_3_0
void FASTCALL call_6_7_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_3_1
#define init_6_7_3_1
void FASTCALL call_6_7_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_3_2
#define init_6_7_3_2
void FASTCALL call_6_7_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_3_3
#define init_6_7_3_3
void FASTCALL call_6_7_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_3_4
#define init_6_7_3_4
void FASTCALL call_6_7_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_3_5
#define init_6_7_3_5
void FASTCALL call_6_7_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_3_6
#define init_6_7_3_6
void FASTCALL call_6_7_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_3_7
#define init_6_7_3_7
void FASTCALL call_6_7_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_3_8
#define init_6_7_3_8
void FASTCALL call_6_7_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_3_9
#define init_6_7_3_9
void FASTCALL call_6_7_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_3_10
#define init_6_7_3_10
void FASTCALL call_6_7_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_3_11
#define init_6_7_3_11
void FASTCALL call_6_7_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_3_12
#define init_6_7_3_12
void FASTCALL call_6_7_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_3_13
#define init_6_7_3_13
void FASTCALL call_6_7_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_3_14
#define init_6_7_3_14
void FASTCALL call_6_7_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_3_15
#define init_6_7_3_15
void FASTCALL call_6_7_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_4_0
#define init_6_7_4_0
void FASTCALL call_6_7_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_4_1
#define init_6_7_4_1
void FASTCALL call_6_7_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_4_2
#define init_6_7_4_2
void FASTCALL call_6_7_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_4_3
#define init_6_7_4_3
void FASTCALL call_6_7_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_4_4
#define init_6_7_4_4
void FASTCALL call_6_7_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_4_5
#define init_6_7_4_5
void FASTCALL call_6_7_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_4_6
#define init_6_7_4_6
void FASTCALL call_6_7_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_4_7
#define init_6_7_4_7
void FASTCALL call_6_7_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_4_8
#define init_6_7_4_8
void FASTCALL call_6_7_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_4_9
#define init_6_7_4_9
void FASTCALL call_6_7_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_4_10
#define init_6_7_4_10
void FASTCALL call_6_7_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_4_11
#define init_6_7_4_11
void FASTCALL call_6_7_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_4_12
#define init_6_7_4_12
void FASTCALL call_6_7_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_4_13
#define init_6_7_4_13
void FASTCALL call_6_7_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_4_14
#define init_6_7_4_14
void FASTCALL call_6_7_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_4_15
#define init_6_7_4_15
void FASTCALL call_6_7_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_5_0
#define init_6_7_5_0
void FASTCALL call_6_7_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_5_1
#define init_6_7_5_1
void FASTCALL call_6_7_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_5_2
#define init_6_7_5_2
void FASTCALL call_6_7_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_5_3
#define init_6_7_5_3
void FASTCALL call_6_7_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_5_4
#define init_6_7_5_4
void FASTCALL call_6_7_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_5_5
#define init_6_7_5_5
void FASTCALL call_6_7_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_5_6
#define init_6_7_5_6
void FASTCALL call_6_7_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_5_7
#define init_6_7_5_7
void FASTCALL call_6_7_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_5_8
#define init_6_7_5_8
void FASTCALL call_6_7_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_5_9
#define init_6_7_5_9
void FASTCALL call_6_7_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_5_10
#define init_6_7_5_10
void FASTCALL call_6_7_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_5_11
#define init_6_7_5_11
void FASTCALL call_6_7_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_5_12
#define init_6_7_5_12
void FASTCALL call_6_7_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_5_13
#define init_6_7_5_13
void FASTCALL call_6_7_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_5_14
#define init_6_7_5_14
void FASTCALL call_6_7_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_5_15
#define init_6_7_5_15
void FASTCALL call_6_7_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_6_0
#define init_6_7_6_0
void FASTCALL call_6_7_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_6_1
#define init_6_7_6_1
void FASTCALL call_6_7_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_6_2
#define init_6_7_6_2
void FASTCALL call_6_7_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_6_3
#define init_6_7_6_3
void FASTCALL call_6_7_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_6_4
#define init_6_7_6_4
void FASTCALL call_6_7_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_6_5
#define init_6_7_6_5
void FASTCALL call_6_7_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_6_6
#define init_6_7_6_6
void FASTCALL call_6_7_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_6_7
#define init_6_7_6_7
void FASTCALL call_6_7_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_6_8
#define init_6_7_6_8
void FASTCALL call_6_7_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_6_9
#define init_6_7_6_9
void FASTCALL call_6_7_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_6_10
#define init_6_7_6_10
void FASTCALL call_6_7_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_6_11
#define init_6_7_6_11
void FASTCALL call_6_7_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_6_12
#define init_6_7_6_12
void FASTCALL call_6_7_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_6_13
#define init_6_7_6_13
void FASTCALL call_6_7_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_6_14
#define init_6_7_6_14
void FASTCALL call_6_7_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_6_15
#define init_6_7_6_15
void FASTCALL call_6_7_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_7_0
#define init_6_7_7_0
void FASTCALL call_6_7_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_7_1
#define init_6_7_7_1
void FASTCALL call_6_7_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_7_2
#define init_6_7_7_2
void FASTCALL call_6_7_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_7_3
#define init_6_7_7_3
void FASTCALL call_6_7_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_7_4
#define init_6_7_7_4
void FASTCALL call_6_7_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_7_5
#define init_6_7_7_5
void FASTCALL call_6_7_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_7_6
#define init_6_7_7_6
void FASTCALL call_6_7_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_7_7
#define init_6_7_7_7
void FASTCALL call_6_7_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_7_8
#define init_6_7_7_8
void FASTCALL call_6_7_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_7_9
#define init_6_7_7_9
void FASTCALL call_6_7_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_7_10
#define init_6_7_7_10
void FASTCALL call_6_7_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_7_11
#define init_6_7_7_11
void FASTCALL call_6_7_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_7_12
#define init_6_7_7_12
void FASTCALL call_6_7_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_7_13
#define init_6_7_7_13
void FASTCALL call_6_7_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_7_14
#define init_6_7_7_14
void FASTCALL call_6_7_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_7_15
#define init_6_7_7_15
void FASTCALL call_6_7_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_8_0
#define init_6_7_8_0
void FASTCALL call_6_7_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_8_1
#define init_6_7_8_1
void FASTCALL call_6_7_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_8_2
#define init_6_7_8_2
void FASTCALL call_6_7_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_8_3
#define init_6_7_8_3
void FASTCALL call_6_7_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_8_4
#define init_6_7_8_4
void FASTCALL call_6_7_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_8_5
#define init_6_7_8_5
void FASTCALL call_6_7_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_8_6
#define init_6_7_8_6
void FASTCALL call_6_7_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_8_7
#define init_6_7_8_7
void FASTCALL call_6_7_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_8_8
#define init_6_7_8_8
void FASTCALL call_6_7_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_8_9
#define init_6_7_8_9
void FASTCALL call_6_7_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_8_10
#define init_6_7_8_10
void FASTCALL call_6_7_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_8_11
#define init_6_7_8_11
void FASTCALL call_6_7_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_8_12
#define init_6_7_8_12
void FASTCALL call_6_7_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_8_13
#define init_6_7_8_13
void FASTCALL call_6_7_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_8_14
#define init_6_7_8_14
void FASTCALL call_6_7_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_8_15
#define init_6_7_8_15
void FASTCALL call_6_7_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_9_0
#define init_6_7_9_0
void FASTCALL call_6_7_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_9_1
#define init_6_7_9_1
void FASTCALL call_6_7_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_9_2
#define init_6_7_9_2
void FASTCALL call_6_7_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_9_3
#define init_6_7_9_3
void FASTCALL call_6_7_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_9_4
#define init_6_7_9_4
void FASTCALL call_6_7_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_9_5
#define init_6_7_9_5
void FASTCALL call_6_7_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_9_6
#define init_6_7_9_6
void FASTCALL call_6_7_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_9_7
#define init_6_7_9_7
void FASTCALL call_6_7_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_9_8
#define init_6_7_9_8
void FASTCALL call_6_7_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_9_9
#define init_6_7_9_9
void FASTCALL call_6_7_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_9_10
#define init_6_7_9_10
void FASTCALL call_6_7_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_9_11
#define init_6_7_9_11
void FASTCALL call_6_7_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_9_12
#define init_6_7_9_12
void FASTCALL call_6_7_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_9_13
#define init_6_7_9_13
void FASTCALL call_6_7_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_9_14
#define init_6_7_9_14
void FASTCALL call_6_7_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_9_15
#define init_6_7_9_15
void FASTCALL call_6_7_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_10_0
#define init_6_7_10_0
void FASTCALL call_6_7_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_10_1
#define init_6_7_10_1
void FASTCALL call_6_7_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_10_2
#define init_6_7_10_2
void FASTCALL call_6_7_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_10_3
#define init_6_7_10_3
void FASTCALL call_6_7_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_10_4
#define init_6_7_10_4
void FASTCALL call_6_7_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_10_5
#define init_6_7_10_5
void FASTCALL call_6_7_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_10_6
#define init_6_7_10_6
void FASTCALL call_6_7_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_10_7
#define init_6_7_10_7
void FASTCALL call_6_7_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_10_8
#define init_6_7_10_8
void FASTCALL call_6_7_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_10_9
#define init_6_7_10_9
void FASTCALL call_6_7_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_10_10
#define init_6_7_10_10
void FASTCALL call_6_7_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_10_11
#define init_6_7_10_11
void FASTCALL call_6_7_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_10_12
#define init_6_7_10_12
void FASTCALL call_6_7_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_10_13
#define init_6_7_10_13
void FASTCALL call_6_7_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_10_14
#define init_6_7_10_14
void FASTCALL call_6_7_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_10_15
#define init_6_7_10_15
void FASTCALL call_6_7_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_11_0
#define init_6_7_11_0
void FASTCALL call_6_7_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_11_1
#define init_6_7_11_1
void FASTCALL call_6_7_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_11_2
#define init_6_7_11_2
void FASTCALL call_6_7_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_11_3
#define init_6_7_11_3
void FASTCALL call_6_7_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_11_4
#define init_6_7_11_4
void FASTCALL call_6_7_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_11_5
#define init_6_7_11_5
void FASTCALL call_6_7_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_11_6
#define init_6_7_11_6
void FASTCALL call_6_7_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_11_7
#define init_6_7_11_7
void FASTCALL call_6_7_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_11_8
#define init_6_7_11_8
void FASTCALL call_6_7_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_11_9
#define init_6_7_11_9
void FASTCALL call_6_7_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_11_10
#define init_6_7_11_10
void FASTCALL call_6_7_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_11_11
#define init_6_7_11_11
void FASTCALL call_6_7_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_11_12
#define init_6_7_11_12
void FASTCALL call_6_7_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_11_13
#define init_6_7_11_13
void FASTCALL call_6_7_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_11_14
#define init_6_7_11_14
void FASTCALL call_6_7_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_11_15
#define init_6_7_11_15
void FASTCALL call_6_7_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_12_0
#define init_6_7_12_0
void FASTCALL call_6_7_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_12_1
#define init_6_7_12_1
void FASTCALL call_6_7_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_12_2
#define init_6_7_12_2
void FASTCALL call_6_7_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_12_3
#define init_6_7_12_3
void FASTCALL call_6_7_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_12_4
#define init_6_7_12_4
void FASTCALL call_6_7_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_12_5
#define init_6_7_12_5
void FASTCALL call_6_7_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_12_6
#define init_6_7_12_6
void FASTCALL call_6_7_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_12_7
#define init_6_7_12_7
void FASTCALL call_6_7_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_12_8
#define init_6_7_12_8
void FASTCALL call_6_7_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_12_9
#define init_6_7_12_9
void FASTCALL call_6_7_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_12_10
#define init_6_7_12_10
void FASTCALL call_6_7_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_12_11
#define init_6_7_12_11
void FASTCALL call_6_7_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_12_12
#define init_6_7_12_12
void FASTCALL call_6_7_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_12_13
#define init_6_7_12_13
void FASTCALL call_6_7_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_12_14
#define init_6_7_12_14
void FASTCALL call_6_7_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_12_15
#define init_6_7_12_15
void FASTCALL call_6_7_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_13_0
#define init_6_7_13_0
void FASTCALL call_6_7_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_13_1
#define init_6_7_13_1
void FASTCALL call_6_7_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_13_2
#define init_6_7_13_2
void FASTCALL call_6_7_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_13_3
#define init_6_7_13_3
void FASTCALL call_6_7_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_13_4
#define init_6_7_13_4
void FASTCALL call_6_7_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_13_5
#define init_6_7_13_5
void FASTCALL call_6_7_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_13_6
#define init_6_7_13_6
void FASTCALL call_6_7_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_13_7
#define init_6_7_13_7
void FASTCALL call_6_7_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_13_8
#define init_6_7_13_8
void FASTCALL call_6_7_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_13_9
#define init_6_7_13_9
void FASTCALL call_6_7_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_13_10
#define init_6_7_13_10
void FASTCALL call_6_7_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_13_11
#define init_6_7_13_11
void FASTCALL call_6_7_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_13_12
#define init_6_7_13_12
void FASTCALL call_6_7_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_13_13
#define init_6_7_13_13
void FASTCALL call_6_7_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_13_14
#define init_6_7_13_14
void FASTCALL call_6_7_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_13_15
#define init_6_7_13_15
void FASTCALL call_6_7_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_14_0
#define init_6_7_14_0
void FASTCALL call_6_7_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_14_1
#define init_6_7_14_1
void FASTCALL call_6_7_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_14_2
#define init_6_7_14_2
void FASTCALL call_6_7_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_14_3
#define init_6_7_14_3
void FASTCALL call_6_7_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_14_4
#define init_6_7_14_4
void FASTCALL call_6_7_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_14_5
#define init_6_7_14_5
void FASTCALL call_6_7_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_14_6
#define init_6_7_14_6
void FASTCALL call_6_7_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_14_7
#define init_6_7_14_7
void FASTCALL call_6_7_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_14_8
#define init_6_7_14_8
void FASTCALL call_6_7_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_14_9
#define init_6_7_14_9
void FASTCALL call_6_7_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_14_10
#define init_6_7_14_10
void FASTCALL call_6_7_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_14_11
#define init_6_7_14_11
void FASTCALL call_6_7_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_14_12
#define init_6_7_14_12
void FASTCALL call_6_7_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_14_13
#define init_6_7_14_13
void FASTCALL call_6_7_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_14_14
#define init_6_7_14_14
void FASTCALL call_6_7_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_14_15
#define init_6_7_14_15
void FASTCALL call_6_7_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_15_0
#define init_6_7_15_0
void FASTCALL call_6_7_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_15_1
#define init_6_7_15_1
void FASTCALL call_6_7_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_15_2
#define init_6_7_15_2
void FASTCALL call_6_7_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_15_3
#define init_6_7_15_3
void FASTCALL call_6_7_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_15_4
#define init_6_7_15_4
void FASTCALL call_6_7_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_15_5
#define init_6_7_15_5
void FASTCALL call_6_7_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_15_6
#define init_6_7_15_6
void FASTCALL call_6_7_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_15_7
#define init_6_7_15_7
void FASTCALL call_6_7_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_15_8
#define init_6_7_15_8
void FASTCALL call_6_7_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_15_9
#define init_6_7_15_9
void FASTCALL call_6_7_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_15_10
#define init_6_7_15_10
void FASTCALL call_6_7_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_15_11
#define init_6_7_15_11
void FASTCALL call_6_7_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_15_12
#define init_6_7_15_12
void FASTCALL call_6_7_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_15_13
#define init_6_7_15_13
void FASTCALL call_6_7_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_15_14
#define init_6_7_15_14
void FASTCALL call_6_7_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_7_15_15
#define init_6_7_15_15
void FASTCALL call_6_7_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_0_0
#define init_6_8_0_0
void FASTCALL call_6_8_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_0_1
#define init_6_8_0_1
void FASTCALL call_6_8_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_0_2
#define init_6_8_0_2
void FASTCALL call_6_8_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_0_3
#define init_6_8_0_3
void FASTCALL call_6_8_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_0_4
#define init_6_8_0_4
void FASTCALL call_6_8_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_0_5
#define init_6_8_0_5
void FASTCALL call_6_8_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_0_6
#define init_6_8_0_6
void FASTCALL call_6_8_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_0_7
#define init_6_8_0_7
void FASTCALL call_6_8_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_0_8
#define init_6_8_0_8
void FASTCALL call_6_8_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_0_9
#define init_6_8_0_9
void FASTCALL call_6_8_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_0_10
#define init_6_8_0_10
void FASTCALL call_6_8_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_0_11
#define init_6_8_0_11
void FASTCALL call_6_8_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_0_12
#define init_6_8_0_12
void FASTCALL call_6_8_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_0_13
#define init_6_8_0_13
void FASTCALL call_6_8_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_0_14
#define init_6_8_0_14
void FASTCALL call_6_8_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_0_15
#define init_6_8_0_15
void FASTCALL call_6_8_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_1_0
#define init_6_8_1_0
void FASTCALL call_6_8_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_1_1
#define init_6_8_1_1
void FASTCALL call_6_8_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_1_2
#define init_6_8_1_2
void FASTCALL call_6_8_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_1_3
#define init_6_8_1_3
void FASTCALL call_6_8_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_1_4
#define init_6_8_1_4
void FASTCALL call_6_8_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_1_5
#define init_6_8_1_5
void FASTCALL call_6_8_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_1_6
#define init_6_8_1_6
void FASTCALL call_6_8_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_1_7
#define init_6_8_1_7
void FASTCALL call_6_8_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_1_8
#define init_6_8_1_8
void FASTCALL call_6_8_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_1_9
#define init_6_8_1_9
void FASTCALL call_6_8_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_1_10
#define init_6_8_1_10
void FASTCALL call_6_8_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_1_11
#define init_6_8_1_11
void FASTCALL call_6_8_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_1_12
#define init_6_8_1_12
void FASTCALL call_6_8_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_1_13
#define init_6_8_1_13
void FASTCALL call_6_8_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_1_14
#define init_6_8_1_14
void FASTCALL call_6_8_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_1_15
#define init_6_8_1_15
void FASTCALL call_6_8_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_2_0
#define init_6_8_2_0
void FASTCALL call_6_8_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_2_1
#define init_6_8_2_1
void FASTCALL call_6_8_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_2_2
#define init_6_8_2_2
void FASTCALL call_6_8_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_2_3
#define init_6_8_2_3
void FASTCALL call_6_8_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_2_4
#define init_6_8_2_4
void FASTCALL call_6_8_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_2_5
#define init_6_8_2_5
void FASTCALL call_6_8_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_2_6
#define init_6_8_2_6
void FASTCALL call_6_8_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_2_7
#define init_6_8_2_7
void FASTCALL call_6_8_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_2_8
#define init_6_8_2_8
void FASTCALL call_6_8_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_2_9
#define init_6_8_2_9
void FASTCALL call_6_8_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_2_10
#define init_6_8_2_10
void FASTCALL call_6_8_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_2_11
#define init_6_8_2_11
void FASTCALL call_6_8_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_2_12
#define init_6_8_2_12
void FASTCALL call_6_8_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_2_13
#define init_6_8_2_13
void FASTCALL call_6_8_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_2_14
#define init_6_8_2_14
void FASTCALL call_6_8_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_2_15
#define init_6_8_2_15
void FASTCALL call_6_8_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_3_0
#define init_6_8_3_0
void FASTCALL call_6_8_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_3_1
#define init_6_8_3_1
void FASTCALL call_6_8_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_3_2
#define init_6_8_3_2
void FASTCALL call_6_8_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_3_3
#define init_6_8_3_3
void FASTCALL call_6_8_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_3_4
#define init_6_8_3_4
void FASTCALL call_6_8_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_3_5
#define init_6_8_3_5
void FASTCALL call_6_8_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_3_6
#define init_6_8_3_6
void FASTCALL call_6_8_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_3_7
#define init_6_8_3_7
void FASTCALL call_6_8_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_3_8
#define init_6_8_3_8
void FASTCALL call_6_8_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_3_9
#define init_6_8_3_9
void FASTCALL call_6_8_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_3_10
#define init_6_8_3_10
void FASTCALL call_6_8_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_3_11
#define init_6_8_3_11
void FASTCALL call_6_8_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_3_12
#define init_6_8_3_12
void FASTCALL call_6_8_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_3_13
#define init_6_8_3_13
void FASTCALL call_6_8_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_3_14
#define init_6_8_3_14
void FASTCALL call_6_8_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_3_15
#define init_6_8_3_15
void FASTCALL call_6_8_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_4_0
#define init_6_8_4_0
void FASTCALL call_6_8_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_4_1
#define init_6_8_4_1
void FASTCALL call_6_8_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_4_2
#define init_6_8_4_2
void FASTCALL call_6_8_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_4_3
#define init_6_8_4_3
void FASTCALL call_6_8_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_4_4
#define init_6_8_4_4
void FASTCALL call_6_8_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_4_5
#define init_6_8_4_5
void FASTCALL call_6_8_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_4_6
#define init_6_8_4_6
void FASTCALL call_6_8_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_4_7
#define init_6_8_4_7
void FASTCALL call_6_8_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_4_8
#define init_6_8_4_8
void FASTCALL call_6_8_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_4_9
#define init_6_8_4_9
void FASTCALL call_6_8_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_4_10
#define init_6_8_4_10
void FASTCALL call_6_8_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_4_11
#define init_6_8_4_11
void FASTCALL call_6_8_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_4_12
#define init_6_8_4_12
void FASTCALL call_6_8_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_4_13
#define init_6_8_4_13
void FASTCALL call_6_8_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_4_14
#define init_6_8_4_14
void FASTCALL call_6_8_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_4_15
#define init_6_8_4_15
void FASTCALL call_6_8_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_5_0
#define init_6_8_5_0
void FASTCALL call_6_8_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_5_1
#define init_6_8_5_1
void FASTCALL call_6_8_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_5_2
#define init_6_8_5_2
void FASTCALL call_6_8_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_5_3
#define init_6_8_5_3
void FASTCALL call_6_8_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_5_4
#define init_6_8_5_4
void FASTCALL call_6_8_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_5_5
#define init_6_8_5_5
void FASTCALL call_6_8_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_5_6
#define init_6_8_5_6
void FASTCALL call_6_8_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_5_7
#define init_6_8_5_7
void FASTCALL call_6_8_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_5_8
#define init_6_8_5_8
void FASTCALL call_6_8_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_5_9
#define init_6_8_5_9
void FASTCALL call_6_8_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_5_10
#define init_6_8_5_10
void FASTCALL call_6_8_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_5_11
#define init_6_8_5_11
void FASTCALL call_6_8_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_5_12
#define init_6_8_5_12
void FASTCALL call_6_8_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_5_13
#define init_6_8_5_13
void FASTCALL call_6_8_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_5_14
#define init_6_8_5_14
void FASTCALL call_6_8_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_5_15
#define init_6_8_5_15
void FASTCALL call_6_8_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_6_0
#define init_6_8_6_0
void FASTCALL call_6_8_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_6_1
#define init_6_8_6_1
void FASTCALL call_6_8_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_6_2
#define init_6_8_6_2
void FASTCALL call_6_8_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_6_3
#define init_6_8_6_3
void FASTCALL call_6_8_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_6_4
#define init_6_8_6_4
void FASTCALL call_6_8_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_6_5
#define init_6_8_6_5
void FASTCALL call_6_8_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_6_6
#define init_6_8_6_6
void FASTCALL call_6_8_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_6_7
#define init_6_8_6_7
void FASTCALL call_6_8_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_6_8
#define init_6_8_6_8
void FASTCALL call_6_8_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_6_9
#define init_6_8_6_9
void FASTCALL call_6_8_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_6_10
#define init_6_8_6_10
void FASTCALL call_6_8_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_6_11
#define init_6_8_6_11
void FASTCALL call_6_8_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_6_12
#define init_6_8_6_12
void FASTCALL call_6_8_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_6_13
#define init_6_8_6_13
void FASTCALL call_6_8_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_6_14
#define init_6_8_6_14
void FASTCALL call_6_8_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_6_15
#define init_6_8_6_15
void FASTCALL call_6_8_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_7_0
#define init_6_8_7_0
void FASTCALL call_6_8_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_7_1
#define init_6_8_7_1
void FASTCALL call_6_8_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_7_2
#define init_6_8_7_2
void FASTCALL call_6_8_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_7_3
#define init_6_8_7_3
void FASTCALL call_6_8_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_7_4
#define init_6_8_7_4
void FASTCALL call_6_8_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_7_5
#define init_6_8_7_5
void FASTCALL call_6_8_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_7_6
#define init_6_8_7_6
void FASTCALL call_6_8_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_7_7
#define init_6_8_7_7
void FASTCALL call_6_8_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_7_8
#define init_6_8_7_8
void FASTCALL call_6_8_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_7_9
#define init_6_8_7_9
void FASTCALL call_6_8_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_7_10
#define init_6_8_7_10
void FASTCALL call_6_8_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_7_11
#define init_6_8_7_11
void FASTCALL call_6_8_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_7_12
#define init_6_8_7_12
void FASTCALL call_6_8_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_7_13
#define init_6_8_7_13
void FASTCALL call_6_8_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_7_14
#define init_6_8_7_14
void FASTCALL call_6_8_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_7_15
#define init_6_8_7_15
void FASTCALL call_6_8_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_8_0
#define init_6_8_8_0
void FASTCALL call_6_8_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_8_1
#define init_6_8_8_1
void FASTCALL call_6_8_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_8_2
#define init_6_8_8_2
void FASTCALL call_6_8_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_8_3
#define init_6_8_8_3
void FASTCALL call_6_8_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_8_4
#define init_6_8_8_4
void FASTCALL call_6_8_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_8_5
#define init_6_8_8_5
void FASTCALL call_6_8_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_8_6
#define init_6_8_8_6
void FASTCALL call_6_8_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_8_7
#define init_6_8_8_7
void FASTCALL call_6_8_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_8_8
#define init_6_8_8_8
void FASTCALL call_6_8_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_8_9
#define init_6_8_8_9
void FASTCALL call_6_8_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_8_10
#define init_6_8_8_10
void FASTCALL call_6_8_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_8_11
#define init_6_8_8_11
void FASTCALL call_6_8_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_8_12
#define init_6_8_8_12
void FASTCALL call_6_8_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_8_13
#define init_6_8_8_13
void FASTCALL call_6_8_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_8_14
#define init_6_8_8_14
void FASTCALL call_6_8_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_8_15
#define init_6_8_8_15
void FASTCALL call_6_8_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_9_0
#define init_6_8_9_0
void FASTCALL call_6_8_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_9_1
#define init_6_8_9_1
void FASTCALL call_6_8_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_9_2
#define init_6_8_9_2
void FASTCALL call_6_8_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_9_3
#define init_6_8_9_3
void FASTCALL call_6_8_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_9_4
#define init_6_8_9_4
void FASTCALL call_6_8_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_9_5
#define init_6_8_9_5
void FASTCALL call_6_8_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_9_6
#define init_6_8_9_6
void FASTCALL call_6_8_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_9_7
#define init_6_8_9_7
void FASTCALL call_6_8_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_9_8
#define init_6_8_9_8
void FASTCALL call_6_8_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_9_9
#define init_6_8_9_9
void FASTCALL call_6_8_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_9_10
#define init_6_8_9_10
void FASTCALL call_6_8_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_9_11
#define init_6_8_9_11
void FASTCALL call_6_8_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_9_12
#define init_6_8_9_12
void FASTCALL call_6_8_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_9_13
#define init_6_8_9_13
void FASTCALL call_6_8_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_9_14
#define init_6_8_9_14
void FASTCALL call_6_8_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_9_15
#define init_6_8_9_15
void FASTCALL call_6_8_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_10_0
#define init_6_8_10_0
void FASTCALL call_6_8_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_10_1
#define init_6_8_10_1
void FASTCALL call_6_8_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_10_2
#define init_6_8_10_2
void FASTCALL call_6_8_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_10_3
#define init_6_8_10_3
void FASTCALL call_6_8_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_10_4
#define init_6_8_10_4
void FASTCALL call_6_8_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_10_5
#define init_6_8_10_5
void FASTCALL call_6_8_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_10_6
#define init_6_8_10_6
void FASTCALL call_6_8_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_10_7
#define init_6_8_10_7
void FASTCALL call_6_8_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_10_8
#define init_6_8_10_8
void FASTCALL call_6_8_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_10_9
#define init_6_8_10_9
void FASTCALL call_6_8_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_10_10
#define init_6_8_10_10
void FASTCALL call_6_8_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_10_11
#define init_6_8_10_11
void FASTCALL call_6_8_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_10_12
#define init_6_8_10_12
void FASTCALL call_6_8_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_10_13
#define init_6_8_10_13
void FASTCALL call_6_8_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_10_14
#define init_6_8_10_14
void FASTCALL call_6_8_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_10_15
#define init_6_8_10_15
void FASTCALL call_6_8_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_11_0
#define init_6_8_11_0
void FASTCALL call_6_8_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_11_1
#define init_6_8_11_1
void FASTCALL call_6_8_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_11_2
#define init_6_8_11_2
void FASTCALL call_6_8_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_11_3
#define init_6_8_11_3
void FASTCALL call_6_8_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_11_4
#define init_6_8_11_4
void FASTCALL call_6_8_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_11_5
#define init_6_8_11_5
void FASTCALL call_6_8_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_11_6
#define init_6_8_11_6
void FASTCALL call_6_8_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_11_7
#define init_6_8_11_7
void FASTCALL call_6_8_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_11_8
#define init_6_8_11_8
void FASTCALL call_6_8_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_11_9
#define init_6_8_11_9
void FASTCALL call_6_8_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_11_10
#define init_6_8_11_10
void FASTCALL call_6_8_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_11_11
#define init_6_8_11_11
void FASTCALL call_6_8_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_11_12
#define init_6_8_11_12
void FASTCALL call_6_8_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_11_13
#define init_6_8_11_13
void FASTCALL call_6_8_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_11_14
#define init_6_8_11_14
void FASTCALL call_6_8_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_11_15
#define init_6_8_11_15
void FASTCALL call_6_8_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_12_0
#define init_6_8_12_0
void FASTCALL call_6_8_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_12_1
#define init_6_8_12_1
void FASTCALL call_6_8_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_12_2
#define init_6_8_12_2
void FASTCALL call_6_8_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_12_3
#define init_6_8_12_3
void FASTCALL call_6_8_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_12_4
#define init_6_8_12_4
void FASTCALL call_6_8_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_12_5
#define init_6_8_12_5
void FASTCALL call_6_8_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_12_6
#define init_6_8_12_6
void FASTCALL call_6_8_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_12_7
#define init_6_8_12_7
void FASTCALL call_6_8_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_12_8
#define init_6_8_12_8
void FASTCALL call_6_8_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_12_9
#define init_6_8_12_9
void FASTCALL call_6_8_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_12_10
#define init_6_8_12_10
void FASTCALL call_6_8_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_12_11
#define init_6_8_12_11
void FASTCALL call_6_8_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_12_12
#define init_6_8_12_12
void FASTCALL call_6_8_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_12_13
#define init_6_8_12_13
void FASTCALL call_6_8_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_12_14
#define init_6_8_12_14
void FASTCALL call_6_8_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_12_15
#define init_6_8_12_15
void FASTCALL call_6_8_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_13_0
#define init_6_8_13_0
void FASTCALL call_6_8_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_13_1
#define init_6_8_13_1
void FASTCALL call_6_8_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_13_2
#define init_6_8_13_2
void FASTCALL call_6_8_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_13_3
#define init_6_8_13_3
void FASTCALL call_6_8_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_13_4
#define init_6_8_13_4
void FASTCALL call_6_8_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_13_5
#define init_6_8_13_5
void FASTCALL call_6_8_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_13_6
#define init_6_8_13_6
void FASTCALL call_6_8_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_13_7
#define init_6_8_13_7
void FASTCALL call_6_8_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_13_8
#define init_6_8_13_8
void FASTCALL call_6_8_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_13_9
#define init_6_8_13_9
void FASTCALL call_6_8_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_13_10
#define init_6_8_13_10
void FASTCALL call_6_8_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_13_11
#define init_6_8_13_11
void FASTCALL call_6_8_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_13_12
#define init_6_8_13_12
void FASTCALL call_6_8_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_13_13
#define init_6_8_13_13
void FASTCALL call_6_8_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_13_14
#define init_6_8_13_14
void FASTCALL call_6_8_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_13_15
#define init_6_8_13_15
void FASTCALL call_6_8_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_14_0
#define init_6_8_14_0
void FASTCALL call_6_8_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_14_1
#define init_6_8_14_1
void FASTCALL call_6_8_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_14_2
#define init_6_8_14_2
void FASTCALL call_6_8_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_14_3
#define init_6_8_14_3
void FASTCALL call_6_8_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_14_4
#define init_6_8_14_4
void FASTCALL call_6_8_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_14_5
#define init_6_8_14_5
void FASTCALL call_6_8_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_14_6
#define init_6_8_14_6
void FASTCALL call_6_8_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_14_7
#define init_6_8_14_7
void FASTCALL call_6_8_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_14_8
#define init_6_8_14_8
void FASTCALL call_6_8_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_14_9
#define init_6_8_14_9
void FASTCALL call_6_8_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_14_10
#define init_6_8_14_10
void FASTCALL call_6_8_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_14_11
#define init_6_8_14_11
void FASTCALL call_6_8_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_14_12
#define init_6_8_14_12
void FASTCALL call_6_8_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_14_13
#define init_6_8_14_13
void FASTCALL call_6_8_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_14_14
#define init_6_8_14_14
void FASTCALL call_6_8_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_14_15
#define init_6_8_14_15
void FASTCALL call_6_8_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_15_0
#define init_6_8_15_0
void FASTCALL call_6_8_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_15_1
#define init_6_8_15_1
void FASTCALL call_6_8_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_15_2
#define init_6_8_15_2
void FASTCALL call_6_8_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_15_3
#define init_6_8_15_3
void FASTCALL call_6_8_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_15_4
#define init_6_8_15_4
void FASTCALL call_6_8_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_15_5
#define init_6_8_15_5
void FASTCALL call_6_8_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_15_6
#define init_6_8_15_6
void FASTCALL call_6_8_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_15_7
#define init_6_8_15_7
void FASTCALL call_6_8_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_15_8
#define init_6_8_15_8
void FASTCALL call_6_8_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_15_9
#define init_6_8_15_9
void FASTCALL call_6_8_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_15_10
#define init_6_8_15_10
void FASTCALL call_6_8_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_15_11
#define init_6_8_15_11
void FASTCALL call_6_8_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_15_12
#define init_6_8_15_12
void FASTCALL call_6_8_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_15_13
#define init_6_8_15_13
void FASTCALL call_6_8_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_15_14
#define init_6_8_15_14
void FASTCALL call_6_8_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_8_15_15
#define init_6_8_15_15
void FASTCALL call_6_8_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_0_0
#define init_6_9_0_0
void FASTCALL call_6_9_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_0_1
#define init_6_9_0_1
void FASTCALL call_6_9_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_0_2
#define init_6_9_0_2
void FASTCALL call_6_9_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_0_3
#define init_6_9_0_3
void FASTCALL call_6_9_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_0_4
#define init_6_9_0_4
void FASTCALL call_6_9_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_0_5
#define init_6_9_0_5
void FASTCALL call_6_9_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_0_6
#define init_6_9_0_6
void FASTCALL call_6_9_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_0_7
#define init_6_9_0_7
void FASTCALL call_6_9_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_0_8
#define init_6_9_0_8
void FASTCALL call_6_9_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_0_9
#define init_6_9_0_9
void FASTCALL call_6_9_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_0_10
#define init_6_9_0_10
void FASTCALL call_6_9_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_0_11
#define init_6_9_0_11
void FASTCALL call_6_9_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_0_12
#define init_6_9_0_12
void FASTCALL call_6_9_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_0_13
#define init_6_9_0_13
void FASTCALL call_6_9_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_0_14
#define init_6_9_0_14
void FASTCALL call_6_9_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_0_15
#define init_6_9_0_15
void FASTCALL call_6_9_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_1_0
#define init_6_9_1_0
void FASTCALL call_6_9_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_1_1
#define init_6_9_1_1
void FASTCALL call_6_9_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_1_2
#define init_6_9_1_2
void FASTCALL call_6_9_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_1_3
#define init_6_9_1_3
void FASTCALL call_6_9_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_1_4
#define init_6_9_1_4
void FASTCALL call_6_9_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_1_5
#define init_6_9_1_5
void FASTCALL call_6_9_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_1_6
#define init_6_9_1_6
void FASTCALL call_6_9_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_1_7
#define init_6_9_1_7
void FASTCALL call_6_9_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_1_8
#define init_6_9_1_8
void FASTCALL call_6_9_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_1_9
#define init_6_9_1_9
void FASTCALL call_6_9_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_1_10
#define init_6_9_1_10
void FASTCALL call_6_9_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_1_11
#define init_6_9_1_11
void FASTCALL call_6_9_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_1_12
#define init_6_9_1_12
void FASTCALL call_6_9_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_1_13
#define init_6_9_1_13
void FASTCALL call_6_9_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_1_14
#define init_6_9_1_14
void FASTCALL call_6_9_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_1_15
#define init_6_9_1_15
void FASTCALL call_6_9_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_2_0
#define init_6_9_2_0
void FASTCALL call_6_9_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_2_1
#define init_6_9_2_1
void FASTCALL call_6_9_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_2_2
#define init_6_9_2_2
void FASTCALL call_6_9_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_2_3
#define init_6_9_2_3
void FASTCALL call_6_9_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_2_4
#define init_6_9_2_4
void FASTCALL call_6_9_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_2_5
#define init_6_9_2_5
void FASTCALL call_6_9_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_2_6
#define init_6_9_2_6
void FASTCALL call_6_9_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_2_7
#define init_6_9_2_7
void FASTCALL call_6_9_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_2_8
#define init_6_9_2_8
void FASTCALL call_6_9_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_2_9
#define init_6_9_2_9
void FASTCALL call_6_9_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_2_10
#define init_6_9_2_10
void FASTCALL call_6_9_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_2_11
#define init_6_9_2_11
void FASTCALL call_6_9_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_2_12
#define init_6_9_2_12
void FASTCALL call_6_9_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_2_13
#define init_6_9_2_13
void FASTCALL call_6_9_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_2_14
#define init_6_9_2_14
void FASTCALL call_6_9_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_2_15
#define init_6_9_2_15
void FASTCALL call_6_9_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_3_0
#define init_6_9_3_0
void FASTCALL call_6_9_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_3_1
#define init_6_9_3_1
void FASTCALL call_6_9_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_3_2
#define init_6_9_3_2
void FASTCALL call_6_9_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_3_3
#define init_6_9_3_3
void FASTCALL call_6_9_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_3_4
#define init_6_9_3_4
void FASTCALL call_6_9_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_3_5
#define init_6_9_3_5
void FASTCALL call_6_9_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_3_6
#define init_6_9_3_6
void FASTCALL call_6_9_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_3_7
#define init_6_9_3_7
void FASTCALL call_6_9_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_3_8
#define init_6_9_3_8
void FASTCALL call_6_9_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_3_9
#define init_6_9_3_9
void FASTCALL call_6_9_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_3_10
#define init_6_9_3_10
void FASTCALL call_6_9_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_3_11
#define init_6_9_3_11
void FASTCALL call_6_9_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_3_12
#define init_6_9_3_12
void FASTCALL call_6_9_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_3_13
#define init_6_9_3_13
void FASTCALL call_6_9_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_3_14
#define init_6_9_3_14
void FASTCALL call_6_9_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_3_15
#define init_6_9_3_15
void FASTCALL call_6_9_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_4_0
#define init_6_9_4_0
void FASTCALL call_6_9_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_4_1
#define init_6_9_4_1
void FASTCALL call_6_9_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_4_2
#define init_6_9_4_2
void FASTCALL call_6_9_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_4_3
#define init_6_9_4_3
void FASTCALL call_6_9_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_4_4
#define init_6_9_4_4
void FASTCALL call_6_9_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_4_5
#define init_6_9_4_5
void FASTCALL call_6_9_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_4_6
#define init_6_9_4_6
void FASTCALL call_6_9_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_4_7
#define init_6_9_4_7
void FASTCALL call_6_9_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_4_8
#define init_6_9_4_8
void FASTCALL call_6_9_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_4_9
#define init_6_9_4_9
void FASTCALL call_6_9_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_4_10
#define init_6_9_4_10
void FASTCALL call_6_9_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_4_11
#define init_6_9_4_11
void FASTCALL call_6_9_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_4_12
#define init_6_9_4_12
void FASTCALL call_6_9_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_4_13
#define init_6_9_4_13
void FASTCALL call_6_9_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_4_14
#define init_6_9_4_14
void FASTCALL call_6_9_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_4_15
#define init_6_9_4_15
void FASTCALL call_6_9_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_5_0
#define init_6_9_5_0
void FASTCALL call_6_9_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_5_1
#define init_6_9_5_1
void FASTCALL call_6_9_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_5_2
#define init_6_9_5_2
void FASTCALL call_6_9_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_5_3
#define init_6_9_5_3
void FASTCALL call_6_9_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_5_4
#define init_6_9_5_4
void FASTCALL call_6_9_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_5_5
#define init_6_9_5_5
void FASTCALL call_6_9_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_5_6
#define init_6_9_5_6
void FASTCALL call_6_9_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_5_7
#define init_6_9_5_7
void FASTCALL call_6_9_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_5_8
#define init_6_9_5_8
void FASTCALL call_6_9_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_5_9
#define init_6_9_5_9
void FASTCALL call_6_9_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_5_10
#define init_6_9_5_10
void FASTCALL call_6_9_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_5_11
#define init_6_9_5_11
void FASTCALL call_6_9_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_5_12
#define init_6_9_5_12
void FASTCALL call_6_9_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_5_13
#define init_6_9_5_13
void FASTCALL call_6_9_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_5_14
#define init_6_9_5_14
void FASTCALL call_6_9_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_5_15
#define init_6_9_5_15
void FASTCALL call_6_9_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_6_0
#define init_6_9_6_0
void FASTCALL call_6_9_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_6_1
#define init_6_9_6_1
void FASTCALL call_6_9_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_6_2
#define init_6_9_6_2
void FASTCALL call_6_9_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_6_3
#define init_6_9_6_3
void FASTCALL call_6_9_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_6_4
#define init_6_9_6_4
void FASTCALL call_6_9_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_6_5
#define init_6_9_6_5
void FASTCALL call_6_9_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_6_6
#define init_6_9_6_6
void FASTCALL call_6_9_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_6_7
#define init_6_9_6_7
void FASTCALL call_6_9_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_6_8
#define init_6_9_6_8
void FASTCALL call_6_9_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_6_9
#define init_6_9_6_9
void FASTCALL call_6_9_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_6_10
#define init_6_9_6_10
void FASTCALL call_6_9_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_6_11
#define init_6_9_6_11
void FASTCALL call_6_9_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_6_12
#define init_6_9_6_12
void FASTCALL call_6_9_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_6_13
#define init_6_9_6_13
void FASTCALL call_6_9_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_6_14
#define init_6_9_6_14
void FASTCALL call_6_9_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_6_15
#define init_6_9_6_15
void FASTCALL call_6_9_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_7_0
#define init_6_9_7_0
void FASTCALL call_6_9_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_7_1
#define init_6_9_7_1
void FASTCALL call_6_9_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_7_2
#define init_6_9_7_2
void FASTCALL call_6_9_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_7_3
#define init_6_9_7_3
void FASTCALL call_6_9_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_7_4
#define init_6_9_7_4
void FASTCALL call_6_9_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_7_5
#define init_6_9_7_5
void FASTCALL call_6_9_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_7_6
#define init_6_9_7_6
void FASTCALL call_6_9_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_7_7
#define init_6_9_7_7
void FASTCALL call_6_9_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_7_8
#define init_6_9_7_8
void FASTCALL call_6_9_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_7_9
#define init_6_9_7_9
void FASTCALL call_6_9_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_7_10
#define init_6_9_7_10
void FASTCALL call_6_9_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_7_11
#define init_6_9_7_11
void FASTCALL call_6_9_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_7_12
#define init_6_9_7_12
void FASTCALL call_6_9_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_7_13
#define init_6_9_7_13
void FASTCALL call_6_9_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_7_14
#define init_6_9_7_14
void FASTCALL call_6_9_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_7_15
#define init_6_9_7_15
void FASTCALL call_6_9_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_8_0
#define init_6_9_8_0
void FASTCALL call_6_9_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_8_1
#define init_6_9_8_1
void FASTCALL call_6_9_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_8_2
#define init_6_9_8_2
void FASTCALL call_6_9_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_8_3
#define init_6_9_8_3
void FASTCALL call_6_9_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_8_4
#define init_6_9_8_4
void FASTCALL call_6_9_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_8_5
#define init_6_9_8_5
void FASTCALL call_6_9_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_8_6
#define init_6_9_8_6
void FASTCALL call_6_9_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_8_7
#define init_6_9_8_7
void FASTCALL call_6_9_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_8_8
#define init_6_9_8_8
void FASTCALL call_6_9_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_8_9
#define init_6_9_8_9
void FASTCALL call_6_9_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_8_10
#define init_6_9_8_10
void FASTCALL call_6_9_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_8_11
#define init_6_9_8_11
void FASTCALL call_6_9_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_8_12
#define init_6_9_8_12
void FASTCALL call_6_9_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_8_13
#define init_6_9_8_13
void FASTCALL call_6_9_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_8_14
#define init_6_9_8_14
void FASTCALL call_6_9_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_8_15
#define init_6_9_8_15
void FASTCALL call_6_9_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_9_0
#define init_6_9_9_0
void FASTCALL call_6_9_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_9_1
#define init_6_9_9_1
void FASTCALL call_6_9_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_9_2
#define init_6_9_9_2
void FASTCALL call_6_9_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_9_3
#define init_6_9_9_3
void FASTCALL call_6_9_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_9_4
#define init_6_9_9_4
void FASTCALL call_6_9_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_9_5
#define init_6_9_9_5
void FASTCALL call_6_9_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_9_6
#define init_6_9_9_6
void FASTCALL call_6_9_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_9_7
#define init_6_9_9_7
void FASTCALL call_6_9_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_9_8
#define init_6_9_9_8
void FASTCALL call_6_9_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_9_9
#define init_6_9_9_9
void FASTCALL call_6_9_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_9_10
#define init_6_9_9_10
void FASTCALL call_6_9_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_9_11
#define init_6_9_9_11
void FASTCALL call_6_9_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_9_12
#define init_6_9_9_12
void FASTCALL call_6_9_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_9_13
#define init_6_9_9_13
void FASTCALL call_6_9_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_9_14
#define init_6_9_9_14
void FASTCALL call_6_9_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_9_15
#define init_6_9_9_15
void FASTCALL call_6_9_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_10_0
#define init_6_9_10_0
void FASTCALL call_6_9_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_10_1
#define init_6_9_10_1
void FASTCALL call_6_9_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_10_2
#define init_6_9_10_2
void FASTCALL call_6_9_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_10_3
#define init_6_9_10_3
void FASTCALL call_6_9_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_10_4
#define init_6_9_10_4
void FASTCALL call_6_9_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_10_5
#define init_6_9_10_5
void FASTCALL call_6_9_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_10_6
#define init_6_9_10_6
void FASTCALL call_6_9_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_10_7
#define init_6_9_10_7
void FASTCALL call_6_9_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_10_8
#define init_6_9_10_8
void FASTCALL call_6_9_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_10_9
#define init_6_9_10_9
void FASTCALL call_6_9_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_10_10
#define init_6_9_10_10
void FASTCALL call_6_9_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_10_11
#define init_6_9_10_11
void FASTCALL call_6_9_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_10_12
#define init_6_9_10_12
void FASTCALL call_6_9_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_10_13
#define init_6_9_10_13
void FASTCALL call_6_9_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_10_14
#define init_6_9_10_14
void FASTCALL call_6_9_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_10_15
#define init_6_9_10_15
void FASTCALL call_6_9_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_11_0
#define init_6_9_11_0
void FASTCALL call_6_9_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_11_1
#define init_6_9_11_1
void FASTCALL call_6_9_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_11_2
#define init_6_9_11_2
void FASTCALL call_6_9_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_11_3
#define init_6_9_11_3
void FASTCALL call_6_9_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_11_4
#define init_6_9_11_4
void FASTCALL call_6_9_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_11_5
#define init_6_9_11_5
void FASTCALL call_6_9_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_11_6
#define init_6_9_11_6
void FASTCALL call_6_9_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_11_7
#define init_6_9_11_7
void FASTCALL call_6_9_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_11_8
#define init_6_9_11_8
void FASTCALL call_6_9_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_11_9
#define init_6_9_11_9
void FASTCALL call_6_9_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_11_10
#define init_6_9_11_10
void FASTCALL call_6_9_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_11_11
#define init_6_9_11_11
void FASTCALL call_6_9_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_11_12
#define init_6_9_11_12
void FASTCALL call_6_9_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_11_13
#define init_6_9_11_13
void FASTCALL call_6_9_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_11_14
#define init_6_9_11_14
void FASTCALL call_6_9_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_11_15
#define init_6_9_11_15
void FASTCALL call_6_9_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_12_0
#define init_6_9_12_0
void FASTCALL call_6_9_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_12_1
#define init_6_9_12_1
void FASTCALL call_6_9_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_12_2
#define init_6_9_12_2
void FASTCALL call_6_9_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_12_3
#define init_6_9_12_3
void FASTCALL call_6_9_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_12_4
#define init_6_9_12_4
void FASTCALL call_6_9_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_12_5
#define init_6_9_12_5
void FASTCALL call_6_9_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_12_6
#define init_6_9_12_6
void FASTCALL call_6_9_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_12_7
#define init_6_9_12_7
void FASTCALL call_6_9_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_12_8
#define init_6_9_12_8
void FASTCALL call_6_9_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_12_9
#define init_6_9_12_9
void FASTCALL call_6_9_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_12_10
#define init_6_9_12_10
void FASTCALL call_6_9_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_12_11
#define init_6_9_12_11
void FASTCALL call_6_9_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_12_12
#define init_6_9_12_12
void FASTCALL call_6_9_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_12_13
#define init_6_9_12_13
void FASTCALL call_6_9_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_12_14
#define init_6_9_12_14
void FASTCALL call_6_9_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_12_15
#define init_6_9_12_15
void FASTCALL call_6_9_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_13_0
#define init_6_9_13_0
void FASTCALL call_6_9_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_13_1
#define init_6_9_13_1
void FASTCALL call_6_9_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_13_2
#define init_6_9_13_2
void FASTCALL call_6_9_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_13_3
#define init_6_9_13_3
void FASTCALL call_6_9_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_13_4
#define init_6_9_13_4
void FASTCALL call_6_9_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_13_5
#define init_6_9_13_5
void FASTCALL call_6_9_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_13_6
#define init_6_9_13_6
void FASTCALL call_6_9_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_13_7
#define init_6_9_13_7
void FASTCALL call_6_9_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_13_8
#define init_6_9_13_8
void FASTCALL call_6_9_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_13_9
#define init_6_9_13_9
void FASTCALL call_6_9_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_13_10
#define init_6_9_13_10
void FASTCALL call_6_9_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_13_11
#define init_6_9_13_11
void FASTCALL call_6_9_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_13_12
#define init_6_9_13_12
void FASTCALL call_6_9_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_13_13
#define init_6_9_13_13
void FASTCALL call_6_9_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_13_14
#define init_6_9_13_14
void FASTCALL call_6_9_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_13_15
#define init_6_9_13_15
void FASTCALL call_6_9_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_14_0
#define init_6_9_14_0
void FASTCALL call_6_9_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_14_1
#define init_6_9_14_1
void FASTCALL call_6_9_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_14_2
#define init_6_9_14_2
void FASTCALL call_6_9_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_14_3
#define init_6_9_14_3
void FASTCALL call_6_9_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_14_4
#define init_6_9_14_4
void FASTCALL call_6_9_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_14_5
#define init_6_9_14_5
void FASTCALL call_6_9_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_14_6
#define init_6_9_14_6
void FASTCALL call_6_9_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_14_7
#define init_6_9_14_7
void FASTCALL call_6_9_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_14_8
#define init_6_9_14_8
void FASTCALL call_6_9_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_14_9
#define init_6_9_14_9
void FASTCALL call_6_9_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_14_10
#define init_6_9_14_10
void FASTCALL call_6_9_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_14_11
#define init_6_9_14_11
void FASTCALL call_6_9_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_14_12
#define init_6_9_14_12
void FASTCALL call_6_9_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_14_13
#define init_6_9_14_13
void FASTCALL call_6_9_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_14_14
#define init_6_9_14_14
void FASTCALL call_6_9_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_14_15
#define init_6_9_14_15
void FASTCALL call_6_9_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_15_0
#define init_6_9_15_0
void FASTCALL call_6_9_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_15_1
#define init_6_9_15_1
void FASTCALL call_6_9_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_15_2
#define init_6_9_15_2
void FASTCALL call_6_9_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_15_3
#define init_6_9_15_3
void FASTCALL call_6_9_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_15_4
#define init_6_9_15_4
void FASTCALL call_6_9_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_15_5
#define init_6_9_15_5
void FASTCALL call_6_9_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_15_6
#define init_6_9_15_6
void FASTCALL call_6_9_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_15_7
#define init_6_9_15_7
void FASTCALL call_6_9_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_15_8
#define init_6_9_15_8
void FASTCALL call_6_9_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_15_9
#define init_6_9_15_9
void FASTCALL call_6_9_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_15_10
#define init_6_9_15_10
void FASTCALL call_6_9_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_15_11
#define init_6_9_15_11
void FASTCALL call_6_9_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_15_12
#define init_6_9_15_12
void FASTCALL call_6_9_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_15_13
#define init_6_9_15_13
void FASTCALL call_6_9_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_15_14
#define init_6_9_15_14
void FASTCALL call_6_9_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_9_15_15
#define init_6_9_15_15
void FASTCALL call_6_9_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_0_0
#define init_6_10_0_0
void FASTCALL call_6_10_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_0_1
#define init_6_10_0_1
void FASTCALL call_6_10_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_0_2
#define init_6_10_0_2
void FASTCALL call_6_10_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_0_3
#define init_6_10_0_3
void FASTCALL call_6_10_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_0_4
#define init_6_10_0_4
void FASTCALL call_6_10_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_0_5
#define init_6_10_0_5
void FASTCALL call_6_10_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_0_6
#define init_6_10_0_6
void FASTCALL call_6_10_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_0_7
#define init_6_10_0_7
void FASTCALL call_6_10_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_0_8
#define init_6_10_0_8
void FASTCALL call_6_10_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_0_9
#define init_6_10_0_9
void FASTCALL call_6_10_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_0_10
#define init_6_10_0_10
void FASTCALL call_6_10_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_0_11
#define init_6_10_0_11
void FASTCALL call_6_10_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_0_12
#define init_6_10_0_12
void FASTCALL call_6_10_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_0_13
#define init_6_10_0_13
void FASTCALL call_6_10_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_0_14
#define init_6_10_0_14
void FASTCALL call_6_10_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_0_15
#define init_6_10_0_15
void FASTCALL call_6_10_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_1_0
#define init_6_10_1_0
void FASTCALL call_6_10_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_1_1
#define init_6_10_1_1
void FASTCALL call_6_10_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_1_2
#define init_6_10_1_2
void FASTCALL call_6_10_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_1_3
#define init_6_10_1_3
void FASTCALL call_6_10_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_1_4
#define init_6_10_1_4
void FASTCALL call_6_10_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_1_5
#define init_6_10_1_5
void FASTCALL call_6_10_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_1_6
#define init_6_10_1_6
void FASTCALL call_6_10_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_1_7
#define init_6_10_1_7
void FASTCALL call_6_10_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_1_8
#define init_6_10_1_8
void FASTCALL call_6_10_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_1_9
#define init_6_10_1_9
void FASTCALL call_6_10_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_1_10
#define init_6_10_1_10
void FASTCALL call_6_10_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_1_11
#define init_6_10_1_11
void FASTCALL call_6_10_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_1_12
#define init_6_10_1_12
void FASTCALL call_6_10_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_1_13
#define init_6_10_1_13
void FASTCALL call_6_10_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_1_14
#define init_6_10_1_14
void FASTCALL call_6_10_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_1_15
#define init_6_10_1_15
void FASTCALL call_6_10_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_2_0
#define init_6_10_2_0
void FASTCALL call_6_10_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_2_1
#define init_6_10_2_1
void FASTCALL call_6_10_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_2_2
#define init_6_10_2_2
void FASTCALL call_6_10_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_2_3
#define init_6_10_2_3
void FASTCALL call_6_10_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_2_4
#define init_6_10_2_4
void FASTCALL call_6_10_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_2_5
#define init_6_10_2_5
void FASTCALL call_6_10_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_2_6
#define init_6_10_2_6
void FASTCALL call_6_10_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_2_7
#define init_6_10_2_7
void FASTCALL call_6_10_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_2_8
#define init_6_10_2_8
void FASTCALL call_6_10_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_2_9
#define init_6_10_2_9
void FASTCALL call_6_10_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_2_10
#define init_6_10_2_10
void FASTCALL call_6_10_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_2_11
#define init_6_10_2_11
void FASTCALL call_6_10_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_2_12
#define init_6_10_2_12
void FASTCALL call_6_10_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_2_13
#define init_6_10_2_13
void FASTCALL call_6_10_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_2_14
#define init_6_10_2_14
void FASTCALL call_6_10_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_2_15
#define init_6_10_2_15
void FASTCALL call_6_10_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_3_0
#define init_6_10_3_0
void FASTCALL call_6_10_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_3_1
#define init_6_10_3_1
void FASTCALL call_6_10_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_3_2
#define init_6_10_3_2
void FASTCALL call_6_10_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_3_3
#define init_6_10_3_3
void FASTCALL call_6_10_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_3_4
#define init_6_10_3_4
void FASTCALL call_6_10_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_3_5
#define init_6_10_3_5
void FASTCALL call_6_10_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_3_6
#define init_6_10_3_6
void FASTCALL call_6_10_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_3_7
#define init_6_10_3_7
void FASTCALL call_6_10_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_3_8
#define init_6_10_3_8
void FASTCALL call_6_10_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_3_9
#define init_6_10_3_9
void FASTCALL call_6_10_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_3_10
#define init_6_10_3_10
void FASTCALL call_6_10_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_3_11
#define init_6_10_3_11
void FASTCALL call_6_10_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_3_12
#define init_6_10_3_12
void FASTCALL call_6_10_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_3_13
#define init_6_10_3_13
void FASTCALL call_6_10_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_3_14
#define init_6_10_3_14
void FASTCALL call_6_10_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_3_15
#define init_6_10_3_15
void FASTCALL call_6_10_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_4_0
#define init_6_10_4_0
void FASTCALL call_6_10_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_4_1
#define init_6_10_4_1
void FASTCALL call_6_10_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_4_2
#define init_6_10_4_2
void FASTCALL call_6_10_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_4_3
#define init_6_10_4_3
void FASTCALL call_6_10_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_4_4
#define init_6_10_4_4
void FASTCALL call_6_10_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_4_5
#define init_6_10_4_5
void FASTCALL call_6_10_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_4_6
#define init_6_10_4_6
void FASTCALL call_6_10_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_4_7
#define init_6_10_4_7
void FASTCALL call_6_10_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_4_8
#define init_6_10_4_8
void FASTCALL call_6_10_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_4_9
#define init_6_10_4_9
void FASTCALL call_6_10_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_4_10
#define init_6_10_4_10
void FASTCALL call_6_10_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_4_11
#define init_6_10_4_11
void FASTCALL call_6_10_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_4_12
#define init_6_10_4_12
void FASTCALL call_6_10_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_4_13
#define init_6_10_4_13
void FASTCALL call_6_10_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_4_14
#define init_6_10_4_14
void FASTCALL call_6_10_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_4_15
#define init_6_10_4_15
void FASTCALL call_6_10_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_5_0
#define init_6_10_5_0
void FASTCALL call_6_10_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_5_1
#define init_6_10_5_1
void FASTCALL call_6_10_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_5_2
#define init_6_10_5_2
void FASTCALL call_6_10_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_5_3
#define init_6_10_5_3
void FASTCALL call_6_10_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_5_4
#define init_6_10_5_4
void FASTCALL call_6_10_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_5_5
#define init_6_10_5_5
void FASTCALL call_6_10_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_5_6
#define init_6_10_5_6
void FASTCALL call_6_10_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_5_7
#define init_6_10_5_7
void FASTCALL call_6_10_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_5_8
#define init_6_10_5_8
void FASTCALL call_6_10_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_5_9
#define init_6_10_5_9
void FASTCALL call_6_10_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_5_10
#define init_6_10_5_10
void FASTCALL call_6_10_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_5_11
#define init_6_10_5_11
void FASTCALL call_6_10_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_5_12
#define init_6_10_5_12
void FASTCALL call_6_10_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_5_13
#define init_6_10_5_13
void FASTCALL call_6_10_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_5_14
#define init_6_10_5_14
void FASTCALL call_6_10_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_5_15
#define init_6_10_5_15
void FASTCALL call_6_10_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_6_0
#define init_6_10_6_0
void FASTCALL call_6_10_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_6_1
#define init_6_10_6_1
void FASTCALL call_6_10_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_6_2
#define init_6_10_6_2
void FASTCALL call_6_10_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_6_3
#define init_6_10_6_3
void FASTCALL call_6_10_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_6_4
#define init_6_10_6_4
void FASTCALL call_6_10_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_6_5
#define init_6_10_6_5
void FASTCALL call_6_10_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_6_6
#define init_6_10_6_6
void FASTCALL call_6_10_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_6_7
#define init_6_10_6_7
void FASTCALL call_6_10_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_6_8
#define init_6_10_6_8
void FASTCALL call_6_10_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_6_9
#define init_6_10_6_9
void FASTCALL call_6_10_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_6_10
#define init_6_10_6_10
void FASTCALL call_6_10_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_6_11
#define init_6_10_6_11
void FASTCALL call_6_10_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_6_12
#define init_6_10_6_12
void FASTCALL call_6_10_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_6_13
#define init_6_10_6_13
void FASTCALL call_6_10_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_6_14
#define init_6_10_6_14
void FASTCALL call_6_10_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_6_15
#define init_6_10_6_15
void FASTCALL call_6_10_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_7_0
#define init_6_10_7_0
void FASTCALL call_6_10_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_7_1
#define init_6_10_7_1
void FASTCALL call_6_10_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_7_2
#define init_6_10_7_2
void FASTCALL call_6_10_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_7_3
#define init_6_10_7_3
void FASTCALL call_6_10_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_7_4
#define init_6_10_7_4
void FASTCALL call_6_10_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_7_5
#define init_6_10_7_5
void FASTCALL call_6_10_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_7_6
#define init_6_10_7_6
void FASTCALL call_6_10_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_7_7
#define init_6_10_7_7
void FASTCALL call_6_10_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_7_8
#define init_6_10_7_8
void FASTCALL call_6_10_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_7_9
#define init_6_10_7_9
void FASTCALL call_6_10_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_7_10
#define init_6_10_7_10
void FASTCALL call_6_10_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_7_11
#define init_6_10_7_11
void FASTCALL call_6_10_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_7_12
#define init_6_10_7_12
void FASTCALL call_6_10_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_7_13
#define init_6_10_7_13
void FASTCALL call_6_10_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_7_14
#define init_6_10_7_14
void FASTCALL call_6_10_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_7_15
#define init_6_10_7_15
void FASTCALL call_6_10_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_8_0
#define init_6_10_8_0
void FASTCALL call_6_10_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_8_1
#define init_6_10_8_1
void FASTCALL call_6_10_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_8_2
#define init_6_10_8_2
void FASTCALL call_6_10_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_8_3
#define init_6_10_8_3
void FASTCALL call_6_10_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_8_4
#define init_6_10_8_4
void FASTCALL call_6_10_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_8_5
#define init_6_10_8_5
void FASTCALL call_6_10_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_8_6
#define init_6_10_8_6
void FASTCALL call_6_10_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_8_7
#define init_6_10_8_7
void FASTCALL call_6_10_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_8_8
#define init_6_10_8_8
void FASTCALL call_6_10_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_8_9
#define init_6_10_8_9
void FASTCALL call_6_10_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_8_10
#define init_6_10_8_10
void FASTCALL call_6_10_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_8_11
#define init_6_10_8_11
void FASTCALL call_6_10_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_8_12
#define init_6_10_8_12
void FASTCALL call_6_10_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_8_13
#define init_6_10_8_13
void FASTCALL call_6_10_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_8_14
#define init_6_10_8_14
void FASTCALL call_6_10_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_8_15
#define init_6_10_8_15
void FASTCALL call_6_10_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_9_0
#define init_6_10_9_0
void FASTCALL call_6_10_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_9_1
#define init_6_10_9_1
void FASTCALL call_6_10_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_9_2
#define init_6_10_9_2
void FASTCALL call_6_10_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_9_3
#define init_6_10_9_3
void FASTCALL call_6_10_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_9_4
#define init_6_10_9_4
void FASTCALL call_6_10_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_9_5
#define init_6_10_9_5
void FASTCALL call_6_10_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_9_6
#define init_6_10_9_6
void FASTCALL call_6_10_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_9_7
#define init_6_10_9_7
void FASTCALL call_6_10_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_9_8
#define init_6_10_9_8
void FASTCALL call_6_10_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_9_9
#define init_6_10_9_9
void FASTCALL call_6_10_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_9_10
#define init_6_10_9_10
void FASTCALL call_6_10_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_9_11
#define init_6_10_9_11
void FASTCALL call_6_10_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_9_12
#define init_6_10_9_12
void FASTCALL call_6_10_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_9_13
#define init_6_10_9_13
void FASTCALL call_6_10_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_9_14
#define init_6_10_9_14
void FASTCALL call_6_10_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_9_15
#define init_6_10_9_15
void FASTCALL call_6_10_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_10_0
#define init_6_10_10_0
void FASTCALL call_6_10_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_10_1
#define init_6_10_10_1
void FASTCALL call_6_10_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_10_2
#define init_6_10_10_2
void FASTCALL call_6_10_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_10_3
#define init_6_10_10_3
void FASTCALL call_6_10_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_10_4
#define init_6_10_10_4
void FASTCALL call_6_10_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_10_5
#define init_6_10_10_5
void FASTCALL call_6_10_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_10_6
#define init_6_10_10_6
void FASTCALL call_6_10_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_10_7
#define init_6_10_10_7
void FASTCALL call_6_10_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_10_8
#define init_6_10_10_8
void FASTCALL call_6_10_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_10_9
#define init_6_10_10_9
void FASTCALL call_6_10_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_10_10
#define init_6_10_10_10
void FASTCALL call_6_10_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_10_11
#define init_6_10_10_11
void FASTCALL call_6_10_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_10_12
#define init_6_10_10_12
void FASTCALL call_6_10_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_10_13
#define init_6_10_10_13
void FASTCALL call_6_10_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_10_14
#define init_6_10_10_14
void FASTCALL call_6_10_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_10_15
#define init_6_10_10_15
void FASTCALL call_6_10_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_11_0
#define init_6_10_11_0
void FASTCALL call_6_10_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_11_1
#define init_6_10_11_1
void FASTCALL call_6_10_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_11_2
#define init_6_10_11_2
void FASTCALL call_6_10_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_11_3
#define init_6_10_11_3
void FASTCALL call_6_10_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_11_4
#define init_6_10_11_4
void FASTCALL call_6_10_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_11_5
#define init_6_10_11_5
void FASTCALL call_6_10_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_11_6
#define init_6_10_11_6
void FASTCALL call_6_10_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_11_7
#define init_6_10_11_7
void FASTCALL call_6_10_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_11_8
#define init_6_10_11_8
void FASTCALL call_6_10_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_11_9
#define init_6_10_11_9
void FASTCALL call_6_10_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_11_10
#define init_6_10_11_10
void FASTCALL call_6_10_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_11_11
#define init_6_10_11_11
void FASTCALL call_6_10_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_11_12
#define init_6_10_11_12
void FASTCALL call_6_10_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_11_13
#define init_6_10_11_13
void FASTCALL call_6_10_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_11_14
#define init_6_10_11_14
void FASTCALL call_6_10_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_11_15
#define init_6_10_11_15
void FASTCALL call_6_10_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_12_0
#define init_6_10_12_0
void FASTCALL call_6_10_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_12_1
#define init_6_10_12_1
void FASTCALL call_6_10_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_12_2
#define init_6_10_12_2
void FASTCALL call_6_10_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_12_3
#define init_6_10_12_3
void FASTCALL call_6_10_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_12_4
#define init_6_10_12_4
void FASTCALL call_6_10_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_12_5
#define init_6_10_12_5
void FASTCALL call_6_10_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_12_6
#define init_6_10_12_6
void FASTCALL call_6_10_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_12_7
#define init_6_10_12_7
void FASTCALL call_6_10_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_12_8
#define init_6_10_12_8
void FASTCALL call_6_10_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_12_9
#define init_6_10_12_9
void FASTCALL call_6_10_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_12_10
#define init_6_10_12_10
void FASTCALL call_6_10_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_12_11
#define init_6_10_12_11
void FASTCALL call_6_10_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_12_12
#define init_6_10_12_12
void FASTCALL call_6_10_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_12_13
#define init_6_10_12_13
void FASTCALL call_6_10_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_12_14
#define init_6_10_12_14
void FASTCALL call_6_10_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_12_15
#define init_6_10_12_15
void FASTCALL call_6_10_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_13_0
#define init_6_10_13_0
void FASTCALL call_6_10_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_13_1
#define init_6_10_13_1
void FASTCALL call_6_10_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_13_2
#define init_6_10_13_2
void FASTCALL call_6_10_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_13_3
#define init_6_10_13_3
void FASTCALL call_6_10_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_13_4
#define init_6_10_13_4
void FASTCALL call_6_10_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_13_5
#define init_6_10_13_5
void FASTCALL call_6_10_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_13_6
#define init_6_10_13_6
void FASTCALL call_6_10_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_13_7
#define init_6_10_13_7
void FASTCALL call_6_10_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_13_8
#define init_6_10_13_8
void FASTCALL call_6_10_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_13_9
#define init_6_10_13_9
void FASTCALL call_6_10_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_13_10
#define init_6_10_13_10
void FASTCALL call_6_10_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_13_11
#define init_6_10_13_11
void FASTCALL call_6_10_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_13_12
#define init_6_10_13_12
void FASTCALL call_6_10_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_13_13
#define init_6_10_13_13
void FASTCALL call_6_10_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_13_14
#define init_6_10_13_14
void FASTCALL call_6_10_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_13_15
#define init_6_10_13_15
void FASTCALL call_6_10_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_14_0
#define init_6_10_14_0
void FASTCALL call_6_10_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_14_1
#define init_6_10_14_1
void FASTCALL call_6_10_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_14_2
#define init_6_10_14_2
void FASTCALL call_6_10_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_14_3
#define init_6_10_14_3
void FASTCALL call_6_10_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_14_4
#define init_6_10_14_4
void FASTCALL call_6_10_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_14_5
#define init_6_10_14_5
void FASTCALL call_6_10_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_14_6
#define init_6_10_14_6
void FASTCALL call_6_10_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_14_7
#define init_6_10_14_7
void FASTCALL call_6_10_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_14_8
#define init_6_10_14_8
void FASTCALL call_6_10_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_14_9
#define init_6_10_14_9
void FASTCALL call_6_10_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_14_10
#define init_6_10_14_10
void FASTCALL call_6_10_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_14_11
#define init_6_10_14_11
void FASTCALL call_6_10_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_14_12
#define init_6_10_14_12
void FASTCALL call_6_10_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_14_13
#define init_6_10_14_13
void FASTCALL call_6_10_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_14_14
#define init_6_10_14_14
void FASTCALL call_6_10_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_14_15
#define init_6_10_14_15
void FASTCALL call_6_10_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_15_0
#define init_6_10_15_0
void FASTCALL call_6_10_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_15_1
#define init_6_10_15_1
void FASTCALL call_6_10_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_15_2
#define init_6_10_15_2
void FASTCALL call_6_10_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_15_3
#define init_6_10_15_3
void FASTCALL call_6_10_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_15_4
#define init_6_10_15_4
void FASTCALL call_6_10_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_15_5
#define init_6_10_15_5
void FASTCALL call_6_10_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_15_6
#define init_6_10_15_6
void FASTCALL call_6_10_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_15_7
#define init_6_10_15_7
void FASTCALL call_6_10_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_15_8
#define init_6_10_15_8
void FASTCALL call_6_10_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_15_9
#define init_6_10_15_9
void FASTCALL call_6_10_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_15_10
#define init_6_10_15_10
void FASTCALL call_6_10_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_15_11
#define init_6_10_15_11
void FASTCALL call_6_10_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_15_12
#define init_6_10_15_12
void FASTCALL call_6_10_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_15_13
#define init_6_10_15_13
void FASTCALL call_6_10_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_15_14
#define init_6_10_15_14
void FASTCALL call_6_10_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_10_15_15
#define init_6_10_15_15
void FASTCALL call_6_10_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_0_0
#define init_6_11_0_0
void FASTCALL call_6_11_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_0_1
#define init_6_11_0_1
void FASTCALL call_6_11_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_0_2
#define init_6_11_0_2
void FASTCALL call_6_11_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_0_3
#define init_6_11_0_3
void FASTCALL call_6_11_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_0_4
#define init_6_11_0_4
void FASTCALL call_6_11_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_0_5
#define init_6_11_0_5
void FASTCALL call_6_11_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_0_6
#define init_6_11_0_6
void FASTCALL call_6_11_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_0_7
#define init_6_11_0_7
void FASTCALL call_6_11_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_0_8
#define init_6_11_0_8
void FASTCALL call_6_11_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_0_9
#define init_6_11_0_9
void FASTCALL call_6_11_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_0_10
#define init_6_11_0_10
void FASTCALL call_6_11_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_0_11
#define init_6_11_0_11
void FASTCALL call_6_11_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_0_12
#define init_6_11_0_12
void FASTCALL call_6_11_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_0_13
#define init_6_11_0_13
void FASTCALL call_6_11_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_0_14
#define init_6_11_0_14
void FASTCALL call_6_11_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_0_15
#define init_6_11_0_15
void FASTCALL call_6_11_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_1_0
#define init_6_11_1_0
void FASTCALL call_6_11_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_1_1
#define init_6_11_1_1
void FASTCALL call_6_11_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_1_2
#define init_6_11_1_2
void FASTCALL call_6_11_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_1_3
#define init_6_11_1_3
void FASTCALL call_6_11_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_1_4
#define init_6_11_1_4
void FASTCALL call_6_11_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_1_5
#define init_6_11_1_5
void FASTCALL call_6_11_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_1_6
#define init_6_11_1_6
void FASTCALL call_6_11_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_1_7
#define init_6_11_1_7
void FASTCALL call_6_11_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_1_8
#define init_6_11_1_8
void FASTCALL call_6_11_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_1_9
#define init_6_11_1_9
void FASTCALL call_6_11_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_1_10
#define init_6_11_1_10
void FASTCALL call_6_11_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_1_11
#define init_6_11_1_11
void FASTCALL call_6_11_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_1_12
#define init_6_11_1_12
void FASTCALL call_6_11_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_1_13
#define init_6_11_1_13
void FASTCALL call_6_11_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_1_14
#define init_6_11_1_14
void FASTCALL call_6_11_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_1_15
#define init_6_11_1_15
void FASTCALL call_6_11_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_2_0
#define init_6_11_2_0
void FASTCALL call_6_11_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_2_1
#define init_6_11_2_1
void FASTCALL call_6_11_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_2_2
#define init_6_11_2_2
void FASTCALL call_6_11_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_2_3
#define init_6_11_2_3
void FASTCALL call_6_11_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_2_4
#define init_6_11_2_4
void FASTCALL call_6_11_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_2_5
#define init_6_11_2_5
void FASTCALL call_6_11_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_2_6
#define init_6_11_2_6
void FASTCALL call_6_11_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_2_7
#define init_6_11_2_7
void FASTCALL call_6_11_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_2_8
#define init_6_11_2_8
void FASTCALL call_6_11_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_2_9
#define init_6_11_2_9
void FASTCALL call_6_11_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_2_10
#define init_6_11_2_10
void FASTCALL call_6_11_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_2_11
#define init_6_11_2_11
void FASTCALL call_6_11_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_2_12
#define init_6_11_2_12
void FASTCALL call_6_11_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_2_13
#define init_6_11_2_13
void FASTCALL call_6_11_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_2_14
#define init_6_11_2_14
void FASTCALL call_6_11_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_2_15
#define init_6_11_2_15
void FASTCALL call_6_11_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_3_0
#define init_6_11_3_0
void FASTCALL call_6_11_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_3_1
#define init_6_11_3_1
void FASTCALL call_6_11_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_3_2
#define init_6_11_3_2
void FASTCALL call_6_11_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_3_3
#define init_6_11_3_3
void FASTCALL call_6_11_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_3_4
#define init_6_11_3_4
void FASTCALL call_6_11_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_3_5
#define init_6_11_3_5
void FASTCALL call_6_11_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_3_6
#define init_6_11_3_6
void FASTCALL call_6_11_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_3_7
#define init_6_11_3_7
void FASTCALL call_6_11_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_3_8
#define init_6_11_3_8
void FASTCALL call_6_11_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_3_9
#define init_6_11_3_9
void FASTCALL call_6_11_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_3_10
#define init_6_11_3_10
void FASTCALL call_6_11_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_3_11
#define init_6_11_3_11
void FASTCALL call_6_11_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_3_12
#define init_6_11_3_12
void FASTCALL call_6_11_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_3_13
#define init_6_11_3_13
void FASTCALL call_6_11_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_3_14
#define init_6_11_3_14
void FASTCALL call_6_11_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_3_15
#define init_6_11_3_15
void FASTCALL call_6_11_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_4_0
#define init_6_11_4_0
void FASTCALL call_6_11_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_4_1
#define init_6_11_4_1
void FASTCALL call_6_11_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_4_2
#define init_6_11_4_2
void FASTCALL call_6_11_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_4_3
#define init_6_11_4_3
void FASTCALL call_6_11_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_4_4
#define init_6_11_4_4
void FASTCALL call_6_11_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_4_5
#define init_6_11_4_5
void FASTCALL call_6_11_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_4_6
#define init_6_11_4_6
void FASTCALL call_6_11_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_4_7
#define init_6_11_4_7
void FASTCALL call_6_11_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_4_8
#define init_6_11_4_8
void FASTCALL call_6_11_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_4_9
#define init_6_11_4_9
void FASTCALL call_6_11_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_4_10
#define init_6_11_4_10
void FASTCALL call_6_11_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_4_11
#define init_6_11_4_11
void FASTCALL call_6_11_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_4_12
#define init_6_11_4_12
void FASTCALL call_6_11_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_4_13
#define init_6_11_4_13
void FASTCALL call_6_11_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_4_14
#define init_6_11_4_14
void FASTCALL call_6_11_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_4_15
#define init_6_11_4_15
void FASTCALL call_6_11_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_5_0
#define init_6_11_5_0
void FASTCALL call_6_11_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_5_1
#define init_6_11_5_1
void FASTCALL call_6_11_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_5_2
#define init_6_11_5_2
void FASTCALL call_6_11_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_5_3
#define init_6_11_5_3
void FASTCALL call_6_11_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_5_4
#define init_6_11_5_4
void FASTCALL call_6_11_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_5_5
#define init_6_11_5_5
void FASTCALL call_6_11_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_5_6
#define init_6_11_5_6
void FASTCALL call_6_11_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_5_7
#define init_6_11_5_7
void FASTCALL call_6_11_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_5_8
#define init_6_11_5_8
void FASTCALL call_6_11_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_5_9
#define init_6_11_5_9
void FASTCALL call_6_11_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_5_10
#define init_6_11_5_10
void FASTCALL call_6_11_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_5_11
#define init_6_11_5_11
void FASTCALL call_6_11_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_5_12
#define init_6_11_5_12
void FASTCALL call_6_11_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_5_13
#define init_6_11_5_13
void FASTCALL call_6_11_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_5_14
#define init_6_11_5_14
void FASTCALL call_6_11_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_5_15
#define init_6_11_5_15
void FASTCALL call_6_11_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_6_0
#define init_6_11_6_0
void FASTCALL call_6_11_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_6_1
#define init_6_11_6_1
void FASTCALL call_6_11_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_6_2
#define init_6_11_6_2
void FASTCALL call_6_11_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_6_3
#define init_6_11_6_3
void FASTCALL call_6_11_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_6_4
#define init_6_11_6_4
void FASTCALL call_6_11_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_6_5
#define init_6_11_6_5
void FASTCALL call_6_11_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_6_6
#define init_6_11_6_6
void FASTCALL call_6_11_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_6_7
#define init_6_11_6_7
void FASTCALL call_6_11_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_6_8
#define init_6_11_6_8
void FASTCALL call_6_11_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_6_9
#define init_6_11_6_9
void FASTCALL call_6_11_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_6_10
#define init_6_11_6_10
void FASTCALL call_6_11_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_6_11
#define init_6_11_6_11
void FASTCALL call_6_11_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_6_12
#define init_6_11_6_12
void FASTCALL call_6_11_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_6_13
#define init_6_11_6_13
void FASTCALL call_6_11_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_6_14
#define init_6_11_6_14
void FASTCALL call_6_11_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_6_15
#define init_6_11_6_15
void FASTCALL call_6_11_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_7_0
#define init_6_11_7_0
void FASTCALL call_6_11_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_7_1
#define init_6_11_7_1
void FASTCALL call_6_11_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_7_2
#define init_6_11_7_2
void FASTCALL call_6_11_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_7_3
#define init_6_11_7_3
void FASTCALL call_6_11_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_7_4
#define init_6_11_7_4
void FASTCALL call_6_11_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_7_5
#define init_6_11_7_5
void FASTCALL call_6_11_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_7_6
#define init_6_11_7_6
void FASTCALL call_6_11_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_7_7
#define init_6_11_7_7
void FASTCALL call_6_11_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_7_8
#define init_6_11_7_8
void FASTCALL call_6_11_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_7_9
#define init_6_11_7_9
void FASTCALL call_6_11_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_7_10
#define init_6_11_7_10
void FASTCALL call_6_11_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_7_11
#define init_6_11_7_11
void FASTCALL call_6_11_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_7_12
#define init_6_11_7_12
void FASTCALL call_6_11_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_7_13
#define init_6_11_7_13
void FASTCALL call_6_11_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_7_14
#define init_6_11_7_14
void FASTCALL call_6_11_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_7_15
#define init_6_11_7_15
void FASTCALL call_6_11_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_8_0
#define init_6_11_8_0
void FASTCALL call_6_11_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_8_1
#define init_6_11_8_1
void FASTCALL call_6_11_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_8_2
#define init_6_11_8_2
void FASTCALL call_6_11_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_8_3
#define init_6_11_8_3
void FASTCALL call_6_11_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_8_4
#define init_6_11_8_4
void FASTCALL call_6_11_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_8_5
#define init_6_11_8_5
void FASTCALL call_6_11_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_8_6
#define init_6_11_8_6
void FASTCALL call_6_11_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_8_7
#define init_6_11_8_7
void FASTCALL call_6_11_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_8_8
#define init_6_11_8_8
void FASTCALL call_6_11_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_8_9
#define init_6_11_8_9
void FASTCALL call_6_11_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_8_10
#define init_6_11_8_10
void FASTCALL call_6_11_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_8_11
#define init_6_11_8_11
void FASTCALL call_6_11_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_8_12
#define init_6_11_8_12
void FASTCALL call_6_11_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_8_13
#define init_6_11_8_13
void FASTCALL call_6_11_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_8_14
#define init_6_11_8_14
void FASTCALL call_6_11_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_8_15
#define init_6_11_8_15
void FASTCALL call_6_11_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_9_0
#define init_6_11_9_0
void FASTCALL call_6_11_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_9_1
#define init_6_11_9_1
void FASTCALL call_6_11_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_9_2
#define init_6_11_9_2
void FASTCALL call_6_11_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_9_3
#define init_6_11_9_3
void FASTCALL call_6_11_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_9_4
#define init_6_11_9_4
void FASTCALL call_6_11_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_9_5
#define init_6_11_9_5
void FASTCALL call_6_11_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_9_6
#define init_6_11_9_6
void FASTCALL call_6_11_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_9_7
#define init_6_11_9_7
void FASTCALL call_6_11_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_9_8
#define init_6_11_9_8
void FASTCALL call_6_11_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_9_9
#define init_6_11_9_9
void FASTCALL call_6_11_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_9_10
#define init_6_11_9_10
void FASTCALL call_6_11_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_9_11
#define init_6_11_9_11
void FASTCALL call_6_11_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_9_12
#define init_6_11_9_12
void FASTCALL call_6_11_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_9_13
#define init_6_11_9_13
void FASTCALL call_6_11_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_9_14
#define init_6_11_9_14
void FASTCALL call_6_11_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_9_15
#define init_6_11_9_15
void FASTCALL call_6_11_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_10_0
#define init_6_11_10_0
void FASTCALL call_6_11_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_10_1
#define init_6_11_10_1
void FASTCALL call_6_11_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_10_2
#define init_6_11_10_2
void FASTCALL call_6_11_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_10_3
#define init_6_11_10_3
void FASTCALL call_6_11_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_10_4
#define init_6_11_10_4
void FASTCALL call_6_11_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_10_5
#define init_6_11_10_5
void FASTCALL call_6_11_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_10_6
#define init_6_11_10_6
void FASTCALL call_6_11_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_10_7
#define init_6_11_10_7
void FASTCALL call_6_11_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_10_8
#define init_6_11_10_8
void FASTCALL call_6_11_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_10_9
#define init_6_11_10_9
void FASTCALL call_6_11_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_10_10
#define init_6_11_10_10
void FASTCALL call_6_11_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_10_11
#define init_6_11_10_11
void FASTCALL call_6_11_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_10_12
#define init_6_11_10_12
void FASTCALL call_6_11_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_10_13
#define init_6_11_10_13
void FASTCALL call_6_11_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_10_14
#define init_6_11_10_14
void FASTCALL call_6_11_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_10_15
#define init_6_11_10_15
void FASTCALL call_6_11_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_11_0
#define init_6_11_11_0
void FASTCALL call_6_11_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_11_1
#define init_6_11_11_1
void FASTCALL call_6_11_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_11_2
#define init_6_11_11_2
void FASTCALL call_6_11_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_11_3
#define init_6_11_11_3
void FASTCALL call_6_11_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_11_4
#define init_6_11_11_4
void FASTCALL call_6_11_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_11_5
#define init_6_11_11_5
void FASTCALL call_6_11_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_11_6
#define init_6_11_11_6
void FASTCALL call_6_11_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_11_7
#define init_6_11_11_7
void FASTCALL call_6_11_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_11_8
#define init_6_11_11_8
void FASTCALL call_6_11_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_11_9
#define init_6_11_11_9
void FASTCALL call_6_11_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_11_10
#define init_6_11_11_10
void FASTCALL call_6_11_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_11_11
#define init_6_11_11_11
void FASTCALL call_6_11_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_11_12
#define init_6_11_11_12
void FASTCALL call_6_11_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_11_13
#define init_6_11_11_13
void FASTCALL call_6_11_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_11_14
#define init_6_11_11_14
void FASTCALL call_6_11_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_11_15
#define init_6_11_11_15
void FASTCALL call_6_11_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_12_0
#define init_6_11_12_0
void FASTCALL call_6_11_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_12_1
#define init_6_11_12_1
void FASTCALL call_6_11_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_12_2
#define init_6_11_12_2
void FASTCALL call_6_11_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_12_3
#define init_6_11_12_3
void FASTCALL call_6_11_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_12_4
#define init_6_11_12_4
void FASTCALL call_6_11_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_12_5
#define init_6_11_12_5
void FASTCALL call_6_11_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_12_6
#define init_6_11_12_6
void FASTCALL call_6_11_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_12_7
#define init_6_11_12_7
void FASTCALL call_6_11_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_12_8
#define init_6_11_12_8
void FASTCALL call_6_11_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_12_9
#define init_6_11_12_9
void FASTCALL call_6_11_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_12_10
#define init_6_11_12_10
void FASTCALL call_6_11_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_12_11
#define init_6_11_12_11
void FASTCALL call_6_11_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_12_12
#define init_6_11_12_12
void FASTCALL call_6_11_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_12_13
#define init_6_11_12_13
void FASTCALL call_6_11_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_12_14
#define init_6_11_12_14
void FASTCALL call_6_11_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_12_15
#define init_6_11_12_15
void FASTCALL call_6_11_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_13_0
#define init_6_11_13_0
void FASTCALL call_6_11_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_13_1
#define init_6_11_13_1
void FASTCALL call_6_11_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_13_2
#define init_6_11_13_2
void FASTCALL call_6_11_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_13_3
#define init_6_11_13_3
void FASTCALL call_6_11_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_13_4
#define init_6_11_13_4
void FASTCALL call_6_11_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_13_5
#define init_6_11_13_5
void FASTCALL call_6_11_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_13_6
#define init_6_11_13_6
void FASTCALL call_6_11_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_13_7
#define init_6_11_13_7
void FASTCALL call_6_11_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_13_8
#define init_6_11_13_8
void FASTCALL call_6_11_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_13_9
#define init_6_11_13_9
void FASTCALL call_6_11_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_13_10
#define init_6_11_13_10
void FASTCALL call_6_11_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_13_11
#define init_6_11_13_11
void FASTCALL call_6_11_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_13_12
#define init_6_11_13_12
void FASTCALL call_6_11_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_13_13
#define init_6_11_13_13
void FASTCALL call_6_11_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_13_14
#define init_6_11_13_14
void FASTCALL call_6_11_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_13_15
#define init_6_11_13_15
void FASTCALL call_6_11_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_14_0
#define init_6_11_14_0
void FASTCALL call_6_11_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_14_1
#define init_6_11_14_1
void FASTCALL call_6_11_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_14_2
#define init_6_11_14_2
void FASTCALL call_6_11_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_14_3
#define init_6_11_14_3
void FASTCALL call_6_11_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_14_4
#define init_6_11_14_4
void FASTCALL call_6_11_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_14_5
#define init_6_11_14_5
void FASTCALL call_6_11_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_14_6
#define init_6_11_14_6
void FASTCALL call_6_11_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_14_7
#define init_6_11_14_7
void FASTCALL call_6_11_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_14_8
#define init_6_11_14_8
void FASTCALL call_6_11_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_14_9
#define init_6_11_14_9
void FASTCALL call_6_11_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_14_10
#define init_6_11_14_10
void FASTCALL call_6_11_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_14_11
#define init_6_11_14_11
void FASTCALL call_6_11_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_14_12
#define init_6_11_14_12
void FASTCALL call_6_11_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_14_13
#define init_6_11_14_13
void FASTCALL call_6_11_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_14_14
#define init_6_11_14_14
void FASTCALL call_6_11_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_14_15
#define init_6_11_14_15
void FASTCALL call_6_11_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_15_0
#define init_6_11_15_0
void FASTCALL call_6_11_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_15_1
#define init_6_11_15_1
void FASTCALL call_6_11_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_15_2
#define init_6_11_15_2
void FASTCALL call_6_11_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_15_3
#define init_6_11_15_3
void FASTCALL call_6_11_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_15_4
#define init_6_11_15_4
void FASTCALL call_6_11_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_15_5
#define init_6_11_15_5
void FASTCALL call_6_11_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_15_6
#define init_6_11_15_6
void FASTCALL call_6_11_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_15_7
#define init_6_11_15_7
void FASTCALL call_6_11_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_15_8
#define init_6_11_15_8
void FASTCALL call_6_11_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_15_9
#define init_6_11_15_9
void FASTCALL call_6_11_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_15_10
#define init_6_11_15_10
void FASTCALL call_6_11_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_15_11
#define init_6_11_15_11
void FASTCALL call_6_11_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_15_12
#define init_6_11_15_12
void FASTCALL call_6_11_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_15_13
#define init_6_11_15_13
void FASTCALL call_6_11_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_15_14
#define init_6_11_15_14
void FASTCALL call_6_11_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_11_15_15
#define init_6_11_15_15
void FASTCALL call_6_11_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_0_0
#define init_6_12_0_0
void FASTCALL call_6_12_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_0_1
#define init_6_12_0_1
void FASTCALL call_6_12_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_0_2
#define init_6_12_0_2
void FASTCALL call_6_12_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_0_3
#define init_6_12_0_3
void FASTCALL call_6_12_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_0_4
#define init_6_12_0_4
void FASTCALL call_6_12_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_0_5
#define init_6_12_0_5
void FASTCALL call_6_12_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_0_6
#define init_6_12_0_6
void FASTCALL call_6_12_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_0_7
#define init_6_12_0_7
void FASTCALL call_6_12_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_0_8
#define init_6_12_0_8
void FASTCALL call_6_12_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_0_9
#define init_6_12_0_9
void FASTCALL call_6_12_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_0_10
#define init_6_12_0_10
void FASTCALL call_6_12_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_0_11
#define init_6_12_0_11
void FASTCALL call_6_12_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_0_12
#define init_6_12_0_12
void FASTCALL call_6_12_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_0_13
#define init_6_12_0_13
void FASTCALL call_6_12_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_0_14
#define init_6_12_0_14
void FASTCALL call_6_12_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_0_15
#define init_6_12_0_15
void FASTCALL call_6_12_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_1_0
#define init_6_12_1_0
void FASTCALL call_6_12_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_1_1
#define init_6_12_1_1
void FASTCALL call_6_12_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_1_2
#define init_6_12_1_2
void FASTCALL call_6_12_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_1_3
#define init_6_12_1_3
void FASTCALL call_6_12_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_1_4
#define init_6_12_1_4
void FASTCALL call_6_12_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_1_5
#define init_6_12_1_5
void FASTCALL call_6_12_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_1_6
#define init_6_12_1_6
void FASTCALL call_6_12_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_1_7
#define init_6_12_1_7
void FASTCALL call_6_12_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_1_8
#define init_6_12_1_8
void FASTCALL call_6_12_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_1_9
#define init_6_12_1_9
void FASTCALL call_6_12_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_1_10
#define init_6_12_1_10
void FASTCALL call_6_12_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_1_11
#define init_6_12_1_11
void FASTCALL call_6_12_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_1_12
#define init_6_12_1_12
void FASTCALL call_6_12_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_1_13
#define init_6_12_1_13
void FASTCALL call_6_12_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_1_14
#define init_6_12_1_14
void FASTCALL call_6_12_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_1_15
#define init_6_12_1_15
void FASTCALL call_6_12_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_2_0
#define init_6_12_2_0
void FASTCALL call_6_12_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_2_1
#define init_6_12_2_1
void FASTCALL call_6_12_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_2_2
#define init_6_12_2_2
void FASTCALL call_6_12_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_2_3
#define init_6_12_2_3
void FASTCALL call_6_12_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_2_4
#define init_6_12_2_4
void FASTCALL call_6_12_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_2_5
#define init_6_12_2_5
void FASTCALL call_6_12_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_2_6
#define init_6_12_2_6
void FASTCALL call_6_12_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_2_7
#define init_6_12_2_7
void FASTCALL call_6_12_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_2_8
#define init_6_12_2_8
void FASTCALL call_6_12_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_2_9
#define init_6_12_2_9
void FASTCALL call_6_12_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_2_10
#define init_6_12_2_10
void FASTCALL call_6_12_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_2_11
#define init_6_12_2_11
void FASTCALL call_6_12_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_2_12
#define init_6_12_2_12
void FASTCALL call_6_12_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_2_13
#define init_6_12_2_13
void FASTCALL call_6_12_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_2_14
#define init_6_12_2_14
void FASTCALL call_6_12_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_2_15
#define init_6_12_2_15
void FASTCALL call_6_12_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_3_0
#define init_6_12_3_0
void FASTCALL call_6_12_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_3_1
#define init_6_12_3_1
void FASTCALL call_6_12_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_3_2
#define init_6_12_3_2
void FASTCALL call_6_12_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_3_3
#define init_6_12_3_3
void FASTCALL call_6_12_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_3_4
#define init_6_12_3_4
void FASTCALL call_6_12_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_3_5
#define init_6_12_3_5
void FASTCALL call_6_12_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_3_6
#define init_6_12_3_6
void FASTCALL call_6_12_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_3_7
#define init_6_12_3_7
void FASTCALL call_6_12_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_3_8
#define init_6_12_3_8
void FASTCALL call_6_12_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_3_9
#define init_6_12_3_9
void FASTCALL call_6_12_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_3_10
#define init_6_12_3_10
void FASTCALL call_6_12_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_3_11
#define init_6_12_3_11
void FASTCALL call_6_12_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_3_12
#define init_6_12_3_12
void FASTCALL call_6_12_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_3_13
#define init_6_12_3_13
void FASTCALL call_6_12_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_3_14
#define init_6_12_3_14
void FASTCALL call_6_12_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_3_15
#define init_6_12_3_15
void FASTCALL call_6_12_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_4_0
#define init_6_12_4_0
void FASTCALL call_6_12_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_4_1
#define init_6_12_4_1
void FASTCALL call_6_12_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_4_2
#define init_6_12_4_2
void FASTCALL call_6_12_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_4_3
#define init_6_12_4_3
void FASTCALL call_6_12_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_4_4
#define init_6_12_4_4
void FASTCALL call_6_12_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_4_5
#define init_6_12_4_5
void FASTCALL call_6_12_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_4_6
#define init_6_12_4_6
void FASTCALL call_6_12_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_4_7
#define init_6_12_4_7
void FASTCALL call_6_12_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_4_8
#define init_6_12_4_8
void FASTCALL call_6_12_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_4_9
#define init_6_12_4_9
void FASTCALL call_6_12_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_4_10
#define init_6_12_4_10
void FASTCALL call_6_12_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_4_11
#define init_6_12_4_11
void FASTCALL call_6_12_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_4_12
#define init_6_12_4_12
void FASTCALL call_6_12_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_4_13
#define init_6_12_4_13
void FASTCALL call_6_12_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_4_14
#define init_6_12_4_14
void FASTCALL call_6_12_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_4_15
#define init_6_12_4_15
void FASTCALL call_6_12_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_5_0
#define init_6_12_5_0
void FASTCALL call_6_12_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_5_1
#define init_6_12_5_1
void FASTCALL call_6_12_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_5_2
#define init_6_12_5_2
void FASTCALL call_6_12_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_5_3
#define init_6_12_5_3
void FASTCALL call_6_12_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_5_4
#define init_6_12_5_4
void FASTCALL call_6_12_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_5_5
#define init_6_12_5_5
void FASTCALL call_6_12_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_5_6
#define init_6_12_5_6
void FASTCALL call_6_12_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_5_7
#define init_6_12_5_7
void FASTCALL call_6_12_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_5_8
#define init_6_12_5_8
void FASTCALL call_6_12_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_5_9
#define init_6_12_5_9
void FASTCALL call_6_12_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_5_10
#define init_6_12_5_10
void FASTCALL call_6_12_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_5_11
#define init_6_12_5_11
void FASTCALL call_6_12_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_5_12
#define init_6_12_5_12
void FASTCALL call_6_12_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_5_13
#define init_6_12_5_13
void FASTCALL call_6_12_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_5_14
#define init_6_12_5_14
void FASTCALL call_6_12_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_5_15
#define init_6_12_5_15
void FASTCALL call_6_12_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_6_0
#define init_6_12_6_0
void FASTCALL call_6_12_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_6_1
#define init_6_12_6_1
void FASTCALL call_6_12_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_6_2
#define init_6_12_6_2
void FASTCALL call_6_12_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_6_3
#define init_6_12_6_3
void FASTCALL call_6_12_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_6_4
#define init_6_12_6_4
void FASTCALL call_6_12_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_6_5
#define init_6_12_6_5
void FASTCALL call_6_12_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_6_6
#define init_6_12_6_6
void FASTCALL call_6_12_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_6_7
#define init_6_12_6_7
void FASTCALL call_6_12_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_6_8
#define init_6_12_6_8
void FASTCALL call_6_12_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_6_9
#define init_6_12_6_9
void FASTCALL call_6_12_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_6_10
#define init_6_12_6_10
void FASTCALL call_6_12_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_6_11
#define init_6_12_6_11
void FASTCALL call_6_12_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_6_12
#define init_6_12_6_12
void FASTCALL call_6_12_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_6_13
#define init_6_12_6_13
void FASTCALL call_6_12_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_6_14
#define init_6_12_6_14
void FASTCALL call_6_12_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_6_15
#define init_6_12_6_15
void FASTCALL call_6_12_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_7_0
#define init_6_12_7_0
void FASTCALL call_6_12_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_7_1
#define init_6_12_7_1
void FASTCALL call_6_12_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_7_2
#define init_6_12_7_2
void FASTCALL call_6_12_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_7_3
#define init_6_12_7_3
void FASTCALL call_6_12_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_7_4
#define init_6_12_7_4
void FASTCALL call_6_12_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_7_5
#define init_6_12_7_5
void FASTCALL call_6_12_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_7_6
#define init_6_12_7_6
void FASTCALL call_6_12_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_7_7
#define init_6_12_7_7
void FASTCALL call_6_12_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_7_8
#define init_6_12_7_8
void FASTCALL call_6_12_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_7_9
#define init_6_12_7_9
void FASTCALL call_6_12_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_7_10
#define init_6_12_7_10
void FASTCALL call_6_12_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_7_11
#define init_6_12_7_11
void FASTCALL call_6_12_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_7_12
#define init_6_12_7_12
void FASTCALL call_6_12_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_7_13
#define init_6_12_7_13
void FASTCALL call_6_12_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_7_14
#define init_6_12_7_14
void FASTCALL call_6_12_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_7_15
#define init_6_12_7_15
void FASTCALL call_6_12_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_8_0
#define init_6_12_8_0
void FASTCALL call_6_12_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_8_1
#define init_6_12_8_1
void FASTCALL call_6_12_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_8_2
#define init_6_12_8_2
void FASTCALL call_6_12_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_8_3
#define init_6_12_8_3
void FASTCALL call_6_12_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_8_4
#define init_6_12_8_4
void FASTCALL call_6_12_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_8_5
#define init_6_12_8_5
void FASTCALL call_6_12_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_8_6
#define init_6_12_8_6
void FASTCALL call_6_12_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_8_7
#define init_6_12_8_7
void FASTCALL call_6_12_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_8_8
#define init_6_12_8_8
void FASTCALL call_6_12_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_8_9
#define init_6_12_8_9
void FASTCALL call_6_12_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_8_10
#define init_6_12_8_10
void FASTCALL call_6_12_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_8_11
#define init_6_12_8_11
void FASTCALL call_6_12_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_8_12
#define init_6_12_8_12
void FASTCALL call_6_12_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_8_13
#define init_6_12_8_13
void FASTCALL call_6_12_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_8_14
#define init_6_12_8_14
void FASTCALL call_6_12_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_8_15
#define init_6_12_8_15
void FASTCALL call_6_12_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_9_0
#define init_6_12_9_0
void FASTCALL call_6_12_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_9_1
#define init_6_12_9_1
void FASTCALL call_6_12_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_9_2
#define init_6_12_9_2
void FASTCALL call_6_12_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_9_3
#define init_6_12_9_3
void FASTCALL call_6_12_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_9_4
#define init_6_12_9_4
void FASTCALL call_6_12_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_9_5
#define init_6_12_9_5
void FASTCALL call_6_12_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_9_6
#define init_6_12_9_6
void FASTCALL call_6_12_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_9_7
#define init_6_12_9_7
void FASTCALL call_6_12_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_9_8
#define init_6_12_9_8
void FASTCALL call_6_12_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_9_9
#define init_6_12_9_9
void FASTCALL call_6_12_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_9_10
#define init_6_12_9_10
void FASTCALL call_6_12_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_9_11
#define init_6_12_9_11
void FASTCALL call_6_12_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_9_12
#define init_6_12_9_12
void FASTCALL call_6_12_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_9_13
#define init_6_12_9_13
void FASTCALL call_6_12_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_9_14
#define init_6_12_9_14
void FASTCALL call_6_12_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_9_15
#define init_6_12_9_15
void FASTCALL call_6_12_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_10_0
#define init_6_12_10_0
void FASTCALL call_6_12_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_10_1
#define init_6_12_10_1
void FASTCALL call_6_12_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_10_2
#define init_6_12_10_2
void FASTCALL call_6_12_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_10_3
#define init_6_12_10_3
void FASTCALL call_6_12_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_10_4
#define init_6_12_10_4
void FASTCALL call_6_12_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_10_5
#define init_6_12_10_5
void FASTCALL call_6_12_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_10_6
#define init_6_12_10_6
void FASTCALL call_6_12_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_10_7
#define init_6_12_10_7
void FASTCALL call_6_12_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_10_8
#define init_6_12_10_8
void FASTCALL call_6_12_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_10_9
#define init_6_12_10_9
void FASTCALL call_6_12_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_10_10
#define init_6_12_10_10
void FASTCALL call_6_12_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_10_11
#define init_6_12_10_11
void FASTCALL call_6_12_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_10_12
#define init_6_12_10_12
void FASTCALL call_6_12_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_10_13
#define init_6_12_10_13
void FASTCALL call_6_12_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_10_14
#define init_6_12_10_14
void FASTCALL call_6_12_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_10_15
#define init_6_12_10_15
void FASTCALL call_6_12_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_11_0
#define init_6_12_11_0
void FASTCALL call_6_12_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_11_1
#define init_6_12_11_1
void FASTCALL call_6_12_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_11_2
#define init_6_12_11_2
void FASTCALL call_6_12_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_11_3
#define init_6_12_11_3
void FASTCALL call_6_12_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_11_4
#define init_6_12_11_4
void FASTCALL call_6_12_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_11_5
#define init_6_12_11_5
void FASTCALL call_6_12_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_11_6
#define init_6_12_11_6
void FASTCALL call_6_12_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_11_7
#define init_6_12_11_7
void FASTCALL call_6_12_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_11_8
#define init_6_12_11_8
void FASTCALL call_6_12_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_11_9
#define init_6_12_11_9
void FASTCALL call_6_12_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_11_10
#define init_6_12_11_10
void FASTCALL call_6_12_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_11_11
#define init_6_12_11_11
void FASTCALL call_6_12_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_11_12
#define init_6_12_11_12
void FASTCALL call_6_12_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_11_13
#define init_6_12_11_13
void FASTCALL call_6_12_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_11_14
#define init_6_12_11_14
void FASTCALL call_6_12_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_11_15
#define init_6_12_11_15
void FASTCALL call_6_12_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_12_0
#define init_6_12_12_0
void FASTCALL call_6_12_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_12_1
#define init_6_12_12_1
void FASTCALL call_6_12_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_12_2
#define init_6_12_12_2
void FASTCALL call_6_12_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_12_3
#define init_6_12_12_3
void FASTCALL call_6_12_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_12_4
#define init_6_12_12_4
void FASTCALL call_6_12_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_12_5
#define init_6_12_12_5
void FASTCALL call_6_12_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_12_6
#define init_6_12_12_6
void FASTCALL call_6_12_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_12_7
#define init_6_12_12_7
void FASTCALL call_6_12_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_12_8
#define init_6_12_12_8
void FASTCALL call_6_12_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_12_9
#define init_6_12_12_9
void FASTCALL call_6_12_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_12_10
#define init_6_12_12_10
void FASTCALL call_6_12_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_12_11
#define init_6_12_12_11
void FASTCALL call_6_12_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_12_12
#define init_6_12_12_12
void FASTCALL call_6_12_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_12_13
#define init_6_12_12_13
void FASTCALL call_6_12_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_12_14
#define init_6_12_12_14
void FASTCALL call_6_12_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_12_15
#define init_6_12_12_15
void FASTCALL call_6_12_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_13_0
#define init_6_12_13_0
void FASTCALL call_6_12_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_13_1
#define init_6_12_13_1
void FASTCALL call_6_12_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_13_2
#define init_6_12_13_2
void FASTCALL call_6_12_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_13_3
#define init_6_12_13_3
void FASTCALL call_6_12_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_13_4
#define init_6_12_13_4
void FASTCALL call_6_12_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_13_5
#define init_6_12_13_5
void FASTCALL call_6_12_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_13_6
#define init_6_12_13_6
void FASTCALL call_6_12_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_13_7
#define init_6_12_13_7
void FASTCALL call_6_12_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_13_8
#define init_6_12_13_8
void FASTCALL call_6_12_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_13_9
#define init_6_12_13_9
void FASTCALL call_6_12_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_13_10
#define init_6_12_13_10
void FASTCALL call_6_12_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_13_11
#define init_6_12_13_11
void FASTCALL call_6_12_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_13_12
#define init_6_12_13_12
void FASTCALL call_6_12_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_13_13
#define init_6_12_13_13
void FASTCALL call_6_12_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_13_14
#define init_6_12_13_14
void FASTCALL call_6_12_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_13_15
#define init_6_12_13_15
void FASTCALL call_6_12_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_14_0
#define init_6_12_14_0
void FASTCALL call_6_12_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_14_1
#define init_6_12_14_1
void FASTCALL call_6_12_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_14_2
#define init_6_12_14_2
void FASTCALL call_6_12_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_14_3
#define init_6_12_14_3
void FASTCALL call_6_12_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_14_4
#define init_6_12_14_4
void FASTCALL call_6_12_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_14_5
#define init_6_12_14_5
void FASTCALL call_6_12_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_14_6
#define init_6_12_14_6
void FASTCALL call_6_12_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_14_7
#define init_6_12_14_7
void FASTCALL call_6_12_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_14_8
#define init_6_12_14_8
void FASTCALL call_6_12_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_14_9
#define init_6_12_14_9
void FASTCALL call_6_12_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_14_10
#define init_6_12_14_10
void FASTCALL call_6_12_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_14_11
#define init_6_12_14_11
void FASTCALL call_6_12_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_14_12
#define init_6_12_14_12
void FASTCALL call_6_12_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_14_13
#define init_6_12_14_13
void FASTCALL call_6_12_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_14_14
#define init_6_12_14_14
void FASTCALL call_6_12_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_14_15
#define init_6_12_14_15
void FASTCALL call_6_12_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_15_0
#define init_6_12_15_0
void FASTCALL call_6_12_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_15_1
#define init_6_12_15_1
void FASTCALL call_6_12_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_15_2
#define init_6_12_15_2
void FASTCALL call_6_12_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_15_3
#define init_6_12_15_3
void FASTCALL call_6_12_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_15_4
#define init_6_12_15_4
void FASTCALL call_6_12_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_15_5
#define init_6_12_15_5
void FASTCALL call_6_12_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_15_6
#define init_6_12_15_6
void FASTCALL call_6_12_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_15_7
#define init_6_12_15_7
void FASTCALL call_6_12_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_15_8
#define init_6_12_15_8
void FASTCALL call_6_12_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_15_9
#define init_6_12_15_9
void FASTCALL call_6_12_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_15_10
#define init_6_12_15_10
void FASTCALL call_6_12_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_15_11
#define init_6_12_15_11
void FASTCALL call_6_12_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_15_12
#define init_6_12_15_12
void FASTCALL call_6_12_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_15_13
#define init_6_12_15_13
void FASTCALL call_6_12_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_15_14
#define init_6_12_15_14
void FASTCALL call_6_12_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_12_15_15
#define init_6_12_15_15
void FASTCALL call_6_12_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_0_0
#define init_6_13_0_0
void FASTCALL call_6_13_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_0_1
#define init_6_13_0_1
void FASTCALL call_6_13_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_0_2
#define init_6_13_0_2
void FASTCALL call_6_13_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_0_3
#define init_6_13_0_3
void FASTCALL call_6_13_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_0_4
#define init_6_13_0_4
void FASTCALL call_6_13_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_0_5
#define init_6_13_0_5
void FASTCALL call_6_13_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_0_6
#define init_6_13_0_6
void FASTCALL call_6_13_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_0_7
#define init_6_13_0_7
void FASTCALL call_6_13_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_0_8
#define init_6_13_0_8
void FASTCALL call_6_13_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_0_9
#define init_6_13_0_9
void FASTCALL call_6_13_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_0_10
#define init_6_13_0_10
void FASTCALL call_6_13_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_0_11
#define init_6_13_0_11
void FASTCALL call_6_13_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_0_12
#define init_6_13_0_12
void FASTCALL call_6_13_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_0_13
#define init_6_13_0_13
void FASTCALL call_6_13_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_0_14
#define init_6_13_0_14
void FASTCALL call_6_13_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_0_15
#define init_6_13_0_15
void FASTCALL call_6_13_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_1_0
#define init_6_13_1_0
void FASTCALL call_6_13_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_1_1
#define init_6_13_1_1
void FASTCALL call_6_13_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_1_2
#define init_6_13_1_2
void FASTCALL call_6_13_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_1_3
#define init_6_13_1_3
void FASTCALL call_6_13_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_1_4
#define init_6_13_1_4
void FASTCALL call_6_13_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_1_5
#define init_6_13_1_5
void FASTCALL call_6_13_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_1_6
#define init_6_13_1_6
void FASTCALL call_6_13_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_1_7
#define init_6_13_1_7
void FASTCALL call_6_13_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_1_8
#define init_6_13_1_8
void FASTCALL call_6_13_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_1_9
#define init_6_13_1_9
void FASTCALL call_6_13_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_1_10
#define init_6_13_1_10
void FASTCALL call_6_13_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_1_11
#define init_6_13_1_11
void FASTCALL call_6_13_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_1_12
#define init_6_13_1_12
void FASTCALL call_6_13_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_1_13
#define init_6_13_1_13
void FASTCALL call_6_13_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_1_14
#define init_6_13_1_14
void FASTCALL call_6_13_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_1_15
#define init_6_13_1_15
void FASTCALL call_6_13_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_2_0
#define init_6_13_2_0
void FASTCALL call_6_13_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_2_1
#define init_6_13_2_1
void FASTCALL call_6_13_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_2_2
#define init_6_13_2_2
void FASTCALL call_6_13_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_2_3
#define init_6_13_2_3
void FASTCALL call_6_13_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_2_4
#define init_6_13_2_4
void FASTCALL call_6_13_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_2_5
#define init_6_13_2_5
void FASTCALL call_6_13_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_2_6
#define init_6_13_2_6
void FASTCALL call_6_13_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_2_7
#define init_6_13_2_7
void FASTCALL call_6_13_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_2_8
#define init_6_13_2_8
void FASTCALL call_6_13_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_2_9
#define init_6_13_2_9
void FASTCALL call_6_13_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_2_10
#define init_6_13_2_10
void FASTCALL call_6_13_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_2_11
#define init_6_13_2_11
void FASTCALL call_6_13_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_2_12
#define init_6_13_2_12
void FASTCALL call_6_13_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_2_13
#define init_6_13_2_13
void FASTCALL call_6_13_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_2_14
#define init_6_13_2_14
void FASTCALL call_6_13_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_2_15
#define init_6_13_2_15
void FASTCALL call_6_13_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_3_0
#define init_6_13_3_0
void FASTCALL call_6_13_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_3_1
#define init_6_13_3_1
void FASTCALL call_6_13_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_3_2
#define init_6_13_3_2
void FASTCALL call_6_13_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_3_3
#define init_6_13_3_3
void FASTCALL call_6_13_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_3_4
#define init_6_13_3_4
void FASTCALL call_6_13_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_3_5
#define init_6_13_3_5
void FASTCALL call_6_13_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_3_6
#define init_6_13_3_6
void FASTCALL call_6_13_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_3_7
#define init_6_13_3_7
void FASTCALL call_6_13_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_3_8
#define init_6_13_3_8
void FASTCALL call_6_13_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_3_9
#define init_6_13_3_9
void FASTCALL call_6_13_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_3_10
#define init_6_13_3_10
void FASTCALL call_6_13_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_3_11
#define init_6_13_3_11
void FASTCALL call_6_13_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_3_12
#define init_6_13_3_12
void FASTCALL call_6_13_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_3_13
#define init_6_13_3_13
void FASTCALL call_6_13_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_3_14
#define init_6_13_3_14
void FASTCALL call_6_13_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_3_15
#define init_6_13_3_15
void FASTCALL call_6_13_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_4_0
#define init_6_13_4_0
void FASTCALL call_6_13_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_4_1
#define init_6_13_4_1
void FASTCALL call_6_13_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_4_2
#define init_6_13_4_2
void FASTCALL call_6_13_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_4_3
#define init_6_13_4_3
void FASTCALL call_6_13_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_4_4
#define init_6_13_4_4
void FASTCALL call_6_13_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_4_5
#define init_6_13_4_5
void FASTCALL call_6_13_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_4_6
#define init_6_13_4_6
void FASTCALL call_6_13_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_4_7
#define init_6_13_4_7
void FASTCALL call_6_13_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_4_8
#define init_6_13_4_8
void FASTCALL call_6_13_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_4_9
#define init_6_13_4_9
void FASTCALL call_6_13_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_4_10
#define init_6_13_4_10
void FASTCALL call_6_13_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_4_11
#define init_6_13_4_11
void FASTCALL call_6_13_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_4_12
#define init_6_13_4_12
void FASTCALL call_6_13_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_4_13
#define init_6_13_4_13
void FASTCALL call_6_13_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_4_14
#define init_6_13_4_14
void FASTCALL call_6_13_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_4_15
#define init_6_13_4_15
void FASTCALL call_6_13_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_5_0
#define init_6_13_5_0
void FASTCALL call_6_13_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_5_1
#define init_6_13_5_1
void FASTCALL call_6_13_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_5_2
#define init_6_13_5_2
void FASTCALL call_6_13_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_5_3
#define init_6_13_5_3
void FASTCALL call_6_13_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_5_4
#define init_6_13_5_4
void FASTCALL call_6_13_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_5_5
#define init_6_13_5_5
void FASTCALL call_6_13_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_5_6
#define init_6_13_5_6
void FASTCALL call_6_13_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_5_7
#define init_6_13_5_7
void FASTCALL call_6_13_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_5_8
#define init_6_13_5_8
void FASTCALL call_6_13_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_5_9
#define init_6_13_5_9
void FASTCALL call_6_13_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_5_10
#define init_6_13_5_10
void FASTCALL call_6_13_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_5_11
#define init_6_13_5_11
void FASTCALL call_6_13_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_5_12
#define init_6_13_5_12
void FASTCALL call_6_13_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_5_13
#define init_6_13_5_13
void FASTCALL call_6_13_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_5_14
#define init_6_13_5_14
void FASTCALL call_6_13_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_5_15
#define init_6_13_5_15
void FASTCALL call_6_13_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_6_0
#define init_6_13_6_0
void FASTCALL call_6_13_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_6_1
#define init_6_13_6_1
void FASTCALL call_6_13_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_6_2
#define init_6_13_6_2
void FASTCALL call_6_13_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_6_3
#define init_6_13_6_3
void FASTCALL call_6_13_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_6_4
#define init_6_13_6_4
void FASTCALL call_6_13_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_6_5
#define init_6_13_6_5
void FASTCALL call_6_13_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_6_6
#define init_6_13_6_6
void FASTCALL call_6_13_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_6_7
#define init_6_13_6_7
void FASTCALL call_6_13_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_6_8
#define init_6_13_6_8
void FASTCALL call_6_13_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_6_9
#define init_6_13_6_9
void FASTCALL call_6_13_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_6_10
#define init_6_13_6_10
void FASTCALL call_6_13_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_6_11
#define init_6_13_6_11
void FASTCALL call_6_13_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_6_12
#define init_6_13_6_12
void FASTCALL call_6_13_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_6_13
#define init_6_13_6_13
void FASTCALL call_6_13_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_6_14
#define init_6_13_6_14
void FASTCALL call_6_13_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_6_15
#define init_6_13_6_15
void FASTCALL call_6_13_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_7_0
#define init_6_13_7_0
void FASTCALL call_6_13_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_7_1
#define init_6_13_7_1
void FASTCALL call_6_13_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_7_2
#define init_6_13_7_2
void FASTCALL call_6_13_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_7_3
#define init_6_13_7_3
void FASTCALL call_6_13_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_7_4
#define init_6_13_7_4
void FASTCALL call_6_13_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_7_5
#define init_6_13_7_5
void FASTCALL call_6_13_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_7_6
#define init_6_13_7_6
void FASTCALL call_6_13_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_7_7
#define init_6_13_7_7
void FASTCALL call_6_13_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_7_8
#define init_6_13_7_8
void FASTCALL call_6_13_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_7_9
#define init_6_13_7_9
void FASTCALL call_6_13_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_7_10
#define init_6_13_7_10
void FASTCALL call_6_13_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_7_11
#define init_6_13_7_11
void FASTCALL call_6_13_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_7_12
#define init_6_13_7_12
void FASTCALL call_6_13_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_7_13
#define init_6_13_7_13
void FASTCALL call_6_13_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_7_14
#define init_6_13_7_14
void FASTCALL call_6_13_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_7_15
#define init_6_13_7_15
void FASTCALL call_6_13_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_8_0
#define init_6_13_8_0
void FASTCALL call_6_13_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_8_1
#define init_6_13_8_1
void FASTCALL call_6_13_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_8_2
#define init_6_13_8_2
void FASTCALL call_6_13_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_8_3
#define init_6_13_8_3
void FASTCALL call_6_13_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_8_4
#define init_6_13_8_4
void FASTCALL call_6_13_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_8_5
#define init_6_13_8_5
void FASTCALL call_6_13_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_8_6
#define init_6_13_8_6
void FASTCALL call_6_13_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_8_7
#define init_6_13_8_7
void FASTCALL call_6_13_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_8_8
#define init_6_13_8_8
void FASTCALL call_6_13_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_8_9
#define init_6_13_8_9
void FASTCALL call_6_13_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_8_10
#define init_6_13_8_10
void FASTCALL call_6_13_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_8_11
#define init_6_13_8_11
void FASTCALL call_6_13_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_8_12
#define init_6_13_8_12
void FASTCALL call_6_13_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_8_13
#define init_6_13_8_13
void FASTCALL call_6_13_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_8_14
#define init_6_13_8_14
void FASTCALL call_6_13_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_8_15
#define init_6_13_8_15
void FASTCALL call_6_13_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_9_0
#define init_6_13_9_0
void FASTCALL call_6_13_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_9_1
#define init_6_13_9_1
void FASTCALL call_6_13_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_9_2
#define init_6_13_9_2
void FASTCALL call_6_13_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_9_3
#define init_6_13_9_3
void FASTCALL call_6_13_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_9_4
#define init_6_13_9_4
void FASTCALL call_6_13_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_9_5
#define init_6_13_9_5
void FASTCALL call_6_13_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_9_6
#define init_6_13_9_6
void FASTCALL call_6_13_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_9_7
#define init_6_13_9_7
void FASTCALL call_6_13_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_9_8
#define init_6_13_9_8
void FASTCALL call_6_13_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_9_9
#define init_6_13_9_9
void FASTCALL call_6_13_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_9_10
#define init_6_13_9_10
void FASTCALL call_6_13_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_9_11
#define init_6_13_9_11
void FASTCALL call_6_13_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_9_12
#define init_6_13_9_12
void FASTCALL call_6_13_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_9_13
#define init_6_13_9_13
void FASTCALL call_6_13_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_9_14
#define init_6_13_9_14
void FASTCALL call_6_13_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_9_15
#define init_6_13_9_15
void FASTCALL call_6_13_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_10_0
#define init_6_13_10_0
void FASTCALL call_6_13_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_10_1
#define init_6_13_10_1
void FASTCALL call_6_13_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_10_2
#define init_6_13_10_2
void FASTCALL call_6_13_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_10_3
#define init_6_13_10_3
void FASTCALL call_6_13_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_10_4
#define init_6_13_10_4
void FASTCALL call_6_13_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_10_5
#define init_6_13_10_5
void FASTCALL call_6_13_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_10_6
#define init_6_13_10_6
void FASTCALL call_6_13_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_10_7
#define init_6_13_10_7
void FASTCALL call_6_13_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_10_8
#define init_6_13_10_8
void FASTCALL call_6_13_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_10_9
#define init_6_13_10_9
void FASTCALL call_6_13_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_10_10
#define init_6_13_10_10
void FASTCALL call_6_13_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_10_11
#define init_6_13_10_11
void FASTCALL call_6_13_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_10_12
#define init_6_13_10_12
void FASTCALL call_6_13_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_10_13
#define init_6_13_10_13
void FASTCALL call_6_13_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_10_14
#define init_6_13_10_14
void FASTCALL call_6_13_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_10_15
#define init_6_13_10_15
void FASTCALL call_6_13_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_11_0
#define init_6_13_11_0
void FASTCALL call_6_13_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_11_1
#define init_6_13_11_1
void FASTCALL call_6_13_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_11_2
#define init_6_13_11_2
void FASTCALL call_6_13_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_11_3
#define init_6_13_11_3
void FASTCALL call_6_13_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_11_4
#define init_6_13_11_4
void FASTCALL call_6_13_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_11_5
#define init_6_13_11_5
void FASTCALL call_6_13_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_11_6
#define init_6_13_11_6
void FASTCALL call_6_13_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_11_7
#define init_6_13_11_7
void FASTCALL call_6_13_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_11_8
#define init_6_13_11_8
void FASTCALL call_6_13_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_11_9
#define init_6_13_11_9
void FASTCALL call_6_13_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_11_10
#define init_6_13_11_10
void FASTCALL call_6_13_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_11_11
#define init_6_13_11_11
void FASTCALL call_6_13_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_11_12
#define init_6_13_11_12
void FASTCALL call_6_13_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_11_13
#define init_6_13_11_13
void FASTCALL call_6_13_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_11_14
#define init_6_13_11_14
void FASTCALL call_6_13_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_11_15
#define init_6_13_11_15
void FASTCALL call_6_13_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_12_0
#define init_6_13_12_0
void FASTCALL call_6_13_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_12_1
#define init_6_13_12_1
void FASTCALL call_6_13_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_12_2
#define init_6_13_12_2
void FASTCALL call_6_13_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_12_3
#define init_6_13_12_3
void FASTCALL call_6_13_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_12_4
#define init_6_13_12_4
void FASTCALL call_6_13_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_12_5
#define init_6_13_12_5
void FASTCALL call_6_13_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_12_6
#define init_6_13_12_6
void FASTCALL call_6_13_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_12_7
#define init_6_13_12_7
void FASTCALL call_6_13_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_12_8
#define init_6_13_12_8
void FASTCALL call_6_13_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_12_9
#define init_6_13_12_9
void FASTCALL call_6_13_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_12_10
#define init_6_13_12_10
void FASTCALL call_6_13_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_12_11
#define init_6_13_12_11
void FASTCALL call_6_13_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_12_12
#define init_6_13_12_12
void FASTCALL call_6_13_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_12_13
#define init_6_13_12_13
void FASTCALL call_6_13_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_12_14
#define init_6_13_12_14
void FASTCALL call_6_13_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_12_15
#define init_6_13_12_15
void FASTCALL call_6_13_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_13_0
#define init_6_13_13_0
void FASTCALL call_6_13_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_13_1
#define init_6_13_13_1
void FASTCALL call_6_13_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_13_2
#define init_6_13_13_2
void FASTCALL call_6_13_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_13_3
#define init_6_13_13_3
void FASTCALL call_6_13_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_13_4
#define init_6_13_13_4
void FASTCALL call_6_13_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_13_5
#define init_6_13_13_5
void FASTCALL call_6_13_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_13_6
#define init_6_13_13_6
void FASTCALL call_6_13_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_13_7
#define init_6_13_13_7
void FASTCALL call_6_13_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_13_8
#define init_6_13_13_8
void FASTCALL call_6_13_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_13_9
#define init_6_13_13_9
void FASTCALL call_6_13_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_13_10
#define init_6_13_13_10
void FASTCALL call_6_13_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_13_11
#define init_6_13_13_11
void FASTCALL call_6_13_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_13_12
#define init_6_13_13_12
void FASTCALL call_6_13_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_13_13
#define init_6_13_13_13
void FASTCALL call_6_13_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_13_14
#define init_6_13_13_14
void FASTCALL call_6_13_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_13_15
#define init_6_13_13_15
void FASTCALL call_6_13_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_14_0
#define init_6_13_14_0
void FASTCALL call_6_13_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_14_1
#define init_6_13_14_1
void FASTCALL call_6_13_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_14_2
#define init_6_13_14_2
void FASTCALL call_6_13_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_14_3
#define init_6_13_14_3
void FASTCALL call_6_13_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_14_4
#define init_6_13_14_4
void FASTCALL call_6_13_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_14_5
#define init_6_13_14_5
void FASTCALL call_6_13_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_14_6
#define init_6_13_14_6
void FASTCALL call_6_13_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_14_7
#define init_6_13_14_7
void FASTCALL call_6_13_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_14_8
#define init_6_13_14_8
void FASTCALL call_6_13_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_14_9
#define init_6_13_14_9
void FASTCALL call_6_13_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_14_10
#define init_6_13_14_10
void FASTCALL call_6_13_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_14_11
#define init_6_13_14_11
void FASTCALL call_6_13_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_14_12
#define init_6_13_14_12
void FASTCALL call_6_13_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_14_13
#define init_6_13_14_13
void FASTCALL call_6_13_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_14_14
#define init_6_13_14_14
void FASTCALL call_6_13_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_14_15
#define init_6_13_14_15
void FASTCALL call_6_13_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_15_0
#define init_6_13_15_0
void FASTCALL call_6_13_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_15_1
#define init_6_13_15_1
void FASTCALL call_6_13_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_15_2
#define init_6_13_15_2
void FASTCALL call_6_13_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_15_3
#define init_6_13_15_3
void FASTCALL call_6_13_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_15_4
#define init_6_13_15_4
void FASTCALL call_6_13_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_15_5
#define init_6_13_15_5
void FASTCALL call_6_13_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_15_6
#define init_6_13_15_6
void FASTCALL call_6_13_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_15_7
#define init_6_13_15_7
void FASTCALL call_6_13_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_15_8
#define init_6_13_15_8
void FASTCALL call_6_13_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_15_9
#define init_6_13_15_9
void FASTCALL call_6_13_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_15_10
#define init_6_13_15_10
void FASTCALL call_6_13_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_15_11
#define init_6_13_15_11
void FASTCALL call_6_13_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_15_12
#define init_6_13_15_12
void FASTCALL call_6_13_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_15_13
#define init_6_13_15_13
void FASTCALL call_6_13_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_15_14
#define init_6_13_15_14
void FASTCALL call_6_13_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_13_15_15
#define init_6_13_15_15
void FASTCALL call_6_13_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_0_0
#define init_6_14_0_0
void FASTCALL call_6_14_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_0_1
#define init_6_14_0_1
void FASTCALL call_6_14_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_0_2
#define init_6_14_0_2
void FASTCALL call_6_14_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_0_3
#define init_6_14_0_3
void FASTCALL call_6_14_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_0_4
#define init_6_14_0_4
void FASTCALL call_6_14_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_0_5
#define init_6_14_0_5
void FASTCALL call_6_14_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_0_6
#define init_6_14_0_6
void FASTCALL call_6_14_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_0_7
#define init_6_14_0_7
void FASTCALL call_6_14_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_0_8
#define init_6_14_0_8
void FASTCALL call_6_14_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_0_9
#define init_6_14_0_9
void FASTCALL call_6_14_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_0_10
#define init_6_14_0_10
void FASTCALL call_6_14_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_0_11
#define init_6_14_0_11
void FASTCALL call_6_14_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_0_12
#define init_6_14_0_12
void FASTCALL call_6_14_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_0_13
#define init_6_14_0_13
void FASTCALL call_6_14_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_0_14
#define init_6_14_0_14
void FASTCALL call_6_14_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_0_15
#define init_6_14_0_15
void FASTCALL call_6_14_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_1_0
#define init_6_14_1_0
void FASTCALL call_6_14_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_1_1
#define init_6_14_1_1
void FASTCALL call_6_14_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_1_2
#define init_6_14_1_2
void FASTCALL call_6_14_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_1_3
#define init_6_14_1_3
void FASTCALL call_6_14_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_1_4
#define init_6_14_1_4
void FASTCALL call_6_14_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_1_5
#define init_6_14_1_5
void FASTCALL call_6_14_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_1_6
#define init_6_14_1_6
void FASTCALL call_6_14_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_1_7
#define init_6_14_1_7
void FASTCALL call_6_14_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_1_8
#define init_6_14_1_8
void FASTCALL call_6_14_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_1_9
#define init_6_14_1_9
void FASTCALL call_6_14_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_1_10
#define init_6_14_1_10
void FASTCALL call_6_14_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_1_11
#define init_6_14_1_11
void FASTCALL call_6_14_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_1_12
#define init_6_14_1_12
void FASTCALL call_6_14_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_1_13
#define init_6_14_1_13
void FASTCALL call_6_14_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_1_14
#define init_6_14_1_14
void FASTCALL call_6_14_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_1_15
#define init_6_14_1_15
void FASTCALL call_6_14_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_2_0
#define init_6_14_2_0
void FASTCALL call_6_14_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_2_1
#define init_6_14_2_1
void FASTCALL call_6_14_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_2_2
#define init_6_14_2_2
void FASTCALL call_6_14_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_2_3
#define init_6_14_2_3
void FASTCALL call_6_14_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_2_4
#define init_6_14_2_4
void FASTCALL call_6_14_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_2_5
#define init_6_14_2_5
void FASTCALL call_6_14_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_2_6
#define init_6_14_2_6
void FASTCALL call_6_14_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_2_7
#define init_6_14_2_7
void FASTCALL call_6_14_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_2_8
#define init_6_14_2_8
void FASTCALL call_6_14_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_2_9
#define init_6_14_2_9
void FASTCALL call_6_14_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_2_10
#define init_6_14_2_10
void FASTCALL call_6_14_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_2_11
#define init_6_14_2_11
void FASTCALL call_6_14_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_2_12
#define init_6_14_2_12
void FASTCALL call_6_14_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_2_13
#define init_6_14_2_13
void FASTCALL call_6_14_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_2_14
#define init_6_14_2_14
void FASTCALL call_6_14_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_2_15
#define init_6_14_2_15
void FASTCALL call_6_14_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_3_0
#define init_6_14_3_0
void FASTCALL call_6_14_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_3_1
#define init_6_14_3_1
void FASTCALL call_6_14_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_3_2
#define init_6_14_3_2
void FASTCALL call_6_14_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_3_3
#define init_6_14_3_3
void FASTCALL call_6_14_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_3_4
#define init_6_14_3_4
void FASTCALL call_6_14_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_3_5
#define init_6_14_3_5
void FASTCALL call_6_14_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_3_6
#define init_6_14_3_6
void FASTCALL call_6_14_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_3_7
#define init_6_14_3_7
void FASTCALL call_6_14_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_3_8
#define init_6_14_3_8
void FASTCALL call_6_14_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_3_9
#define init_6_14_3_9
void FASTCALL call_6_14_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_3_10
#define init_6_14_3_10
void FASTCALL call_6_14_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_3_11
#define init_6_14_3_11
void FASTCALL call_6_14_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_3_12
#define init_6_14_3_12
void FASTCALL call_6_14_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_3_13
#define init_6_14_3_13
void FASTCALL call_6_14_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_3_14
#define init_6_14_3_14
void FASTCALL call_6_14_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_3_15
#define init_6_14_3_15
void FASTCALL call_6_14_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_4_0
#define init_6_14_4_0
void FASTCALL call_6_14_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_4_1
#define init_6_14_4_1
void FASTCALL call_6_14_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_4_2
#define init_6_14_4_2
void FASTCALL call_6_14_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_4_3
#define init_6_14_4_3
void FASTCALL call_6_14_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_4_4
#define init_6_14_4_4
void FASTCALL call_6_14_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_4_5
#define init_6_14_4_5
void FASTCALL call_6_14_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_4_6
#define init_6_14_4_6
void FASTCALL call_6_14_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_4_7
#define init_6_14_4_7
void FASTCALL call_6_14_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_4_8
#define init_6_14_4_8
void FASTCALL call_6_14_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_4_9
#define init_6_14_4_9
void FASTCALL call_6_14_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_4_10
#define init_6_14_4_10
void FASTCALL call_6_14_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_4_11
#define init_6_14_4_11
void FASTCALL call_6_14_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_4_12
#define init_6_14_4_12
void FASTCALL call_6_14_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_4_13
#define init_6_14_4_13
void FASTCALL call_6_14_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_4_14
#define init_6_14_4_14
void FASTCALL call_6_14_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_4_15
#define init_6_14_4_15
void FASTCALL call_6_14_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_5_0
#define init_6_14_5_0
void FASTCALL call_6_14_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_5_1
#define init_6_14_5_1
void FASTCALL call_6_14_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_5_2
#define init_6_14_5_2
void FASTCALL call_6_14_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_5_3
#define init_6_14_5_3
void FASTCALL call_6_14_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_5_4
#define init_6_14_5_4
void FASTCALL call_6_14_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_5_5
#define init_6_14_5_5
void FASTCALL call_6_14_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_5_6
#define init_6_14_5_6
void FASTCALL call_6_14_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_5_7
#define init_6_14_5_7
void FASTCALL call_6_14_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_5_8
#define init_6_14_5_8
void FASTCALL call_6_14_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_5_9
#define init_6_14_5_9
void FASTCALL call_6_14_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_5_10
#define init_6_14_5_10
void FASTCALL call_6_14_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_5_11
#define init_6_14_5_11
void FASTCALL call_6_14_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_5_12
#define init_6_14_5_12
void FASTCALL call_6_14_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_5_13
#define init_6_14_5_13
void FASTCALL call_6_14_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_5_14
#define init_6_14_5_14
void FASTCALL call_6_14_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_5_15
#define init_6_14_5_15
void FASTCALL call_6_14_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_6_0
#define init_6_14_6_0
void FASTCALL call_6_14_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_6_1
#define init_6_14_6_1
void FASTCALL call_6_14_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_6_2
#define init_6_14_6_2
void FASTCALL call_6_14_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_6_3
#define init_6_14_6_3
void FASTCALL call_6_14_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_6_4
#define init_6_14_6_4
void FASTCALL call_6_14_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_6_5
#define init_6_14_6_5
void FASTCALL call_6_14_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_6_6
#define init_6_14_6_6
void FASTCALL call_6_14_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_6_7
#define init_6_14_6_7
void FASTCALL call_6_14_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_6_8
#define init_6_14_6_8
void FASTCALL call_6_14_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_6_9
#define init_6_14_6_9
void FASTCALL call_6_14_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_6_10
#define init_6_14_6_10
void FASTCALL call_6_14_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_6_11
#define init_6_14_6_11
void FASTCALL call_6_14_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_6_12
#define init_6_14_6_12
void FASTCALL call_6_14_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_6_13
#define init_6_14_6_13
void FASTCALL call_6_14_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_6_14
#define init_6_14_6_14
void FASTCALL call_6_14_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_6_15
#define init_6_14_6_15
void FASTCALL call_6_14_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_7_0
#define init_6_14_7_0
void FASTCALL call_6_14_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_7_1
#define init_6_14_7_1
void FASTCALL call_6_14_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_7_2
#define init_6_14_7_2
void FASTCALL call_6_14_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_7_3
#define init_6_14_7_3
void FASTCALL call_6_14_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_7_4
#define init_6_14_7_4
void FASTCALL call_6_14_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_7_5
#define init_6_14_7_5
void FASTCALL call_6_14_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_7_6
#define init_6_14_7_6
void FASTCALL call_6_14_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_7_7
#define init_6_14_7_7
void FASTCALL call_6_14_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_7_8
#define init_6_14_7_8
void FASTCALL call_6_14_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_7_9
#define init_6_14_7_9
void FASTCALL call_6_14_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_7_10
#define init_6_14_7_10
void FASTCALL call_6_14_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_7_11
#define init_6_14_7_11
void FASTCALL call_6_14_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_7_12
#define init_6_14_7_12
void FASTCALL call_6_14_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_7_13
#define init_6_14_7_13
void FASTCALL call_6_14_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_7_14
#define init_6_14_7_14
void FASTCALL call_6_14_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_7_15
#define init_6_14_7_15
void FASTCALL call_6_14_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_8_0
#define init_6_14_8_0
void FASTCALL call_6_14_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_8_1
#define init_6_14_8_1
void FASTCALL call_6_14_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_8_2
#define init_6_14_8_2
void FASTCALL call_6_14_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_8_3
#define init_6_14_8_3
void FASTCALL call_6_14_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_8_4
#define init_6_14_8_4
void FASTCALL call_6_14_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_8_5
#define init_6_14_8_5
void FASTCALL call_6_14_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_8_6
#define init_6_14_8_6
void FASTCALL call_6_14_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_8_7
#define init_6_14_8_7
void FASTCALL call_6_14_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_8_8
#define init_6_14_8_8
void FASTCALL call_6_14_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_8_9
#define init_6_14_8_9
void FASTCALL call_6_14_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_8_10
#define init_6_14_8_10
void FASTCALL call_6_14_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_8_11
#define init_6_14_8_11
void FASTCALL call_6_14_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_8_12
#define init_6_14_8_12
void FASTCALL call_6_14_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_8_13
#define init_6_14_8_13
void FASTCALL call_6_14_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_8_14
#define init_6_14_8_14
void FASTCALL call_6_14_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_8_15
#define init_6_14_8_15
void FASTCALL call_6_14_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_9_0
#define init_6_14_9_0
void FASTCALL call_6_14_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_9_1
#define init_6_14_9_1
void FASTCALL call_6_14_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_9_2
#define init_6_14_9_2
void FASTCALL call_6_14_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_9_3
#define init_6_14_9_3
void FASTCALL call_6_14_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_9_4
#define init_6_14_9_4
void FASTCALL call_6_14_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_9_5
#define init_6_14_9_5
void FASTCALL call_6_14_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_9_6
#define init_6_14_9_6
void FASTCALL call_6_14_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_9_7
#define init_6_14_9_7
void FASTCALL call_6_14_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_9_8
#define init_6_14_9_8
void FASTCALL call_6_14_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_9_9
#define init_6_14_9_9
void FASTCALL call_6_14_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_9_10
#define init_6_14_9_10
void FASTCALL call_6_14_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_9_11
#define init_6_14_9_11
void FASTCALL call_6_14_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_9_12
#define init_6_14_9_12
void FASTCALL call_6_14_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_9_13
#define init_6_14_9_13
void FASTCALL call_6_14_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_9_14
#define init_6_14_9_14
void FASTCALL call_6_14_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_9_15
#define init_6_14_9_15
void FASTCALL call_6_14_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_10_0
#define init_6_14_10_0
void FASTCALL call_6_14_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_10_1
#define init_6_14_10_1
void FASTCALL call_6_14_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_10_2
#define init_6_14_10_2
void FASTCALL call_6_14_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_10_3
#define init_6_14_10_3
void FASTCALL call_6_14_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_10_4
#define init_6_14_10_4
void FASTCALL call_6_14_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_10_5
#define init_6_14_10_5
void FASTCALL call_6_14_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_10_6
#define init_6_14_10_6
void FASTCALL call_6_14_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_10_7
#define init_6_14_10_7
void FASTCALL call_6_14_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_10_8
#define init_6_14_10_8
void FASTCALL call_6_14_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_10_9
#define init_6_14_10_9
void FASTCALL call_6_14_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_10_10
#define init_6_14_10_10
void FASTCALL call_6_14_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_10_11
#define init_6_14_10_11
void FASTCALL call_6_14_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_10_12
#define init_6_14_10_12
void FASTCALL call_6_14_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_10_13
#define init_6_14_10_13
void FASTCALL call_6_14_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_10_14
#define init_6_14_10_14
void FASTCALL call_6_14_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_10_15
#define init_6_14_10_15
void FASTCALL call_6_14_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_11_0
#define init_6_14_11_0
void FASTCALL call_6_14_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_11_1
#define init_6_14_11_1
void FASTCALL call_6_14_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_11_2
#define init_6_14_11_2
void FASTCALL call_6_14_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_11_3
#define init_6_14_11_3
void FASTCALL call_6_14_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_11_4
#define init_6_14_11_4
void FASTCALL call_6_14_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_11_5
#define init_6_14_11_5
void FASTCALL call_6_14_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_11_6
#define init_6_14_11_6
void FASTCALL call_6_14_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_11_7
#define init_6_14_11_7
void FASTCALL call_6_14_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_11_8
#define init_6_14_11_8
void FASTCALL call_6_14_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_11_9
#define init_6_14_11_9
void FASTCALL call_6_14_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_11_10
#define init_6_14_11_10
void FASTCALL call_6_14_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_11_11
#define init_6_14_11_11
void FASTCALL call_6_14_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_11_12
#define init_6_14_11_12
void FASTCALL call_6_14_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_11_13
#define init_6_14_11_13
void FASTCALL call_6_14_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_11_14
#define init_6_14_11_14
void FASTCALL call_6_14_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_11_15
#define init_6_14_11_15
void FASTCALL call_6_14_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_12_0
#define init_6_14_12_0
void FASTCALL call_6_14_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_12_1
#define init_6_14_12_1
void FASTCALL call_6_14_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_12_2
#define init_6_14_12_2
void FASTCALL call_6_14_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_12_3
#define init_6_14_12_3
void FASTCALL call_6_14_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_12_4
#define init_6_14_12_4
void FASTCALL call_6_14_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_12_5
#define init_6_14_12_5
void FASTCALL call_6_14_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_12_6
#define init_6_14_12_6
void FASTCALL call_6_14_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_12_7
#define init_6_14_12_7
void FASTCALL call_6_14_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_12_8
#define init_6_14_12_8
void FASTCALL call_6_14_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_12_9
#define init_6_14_12_9
void FASTCALL call_6_14_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_12_10
#define init_6_14_12_10
void FASTCALL call_6_14_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_12_11
#define init_6_14_12_11
void FASTCALL call_6_14_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_12_12
#define init_6_14_12_12
void FASTCALL call_6_14_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_12_13
#define init_6_14_12_13
void FASTCALL call_6_14_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_12_14
#define init_6_14_12_14
void FASTCALL call_6_14_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_12_15
#define init_6_14_12_15
void FASTCALL call_6_14_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_13_0
#define init_6_14_13_0
void FASTCALL call_6_14_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_13_1
#define init_6_14_13_1
void FASTCALL call_6_14_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_13_2
#define init_6_14_13_2
void FASTCALL call_6_14_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_13_3
#define init_6_14_13_3
void FASTCALL call_6_14_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_13_4
#define init_6_14_13_4
void FASTCALL call_6_14_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_13_5
#define init_6_14_13_5
void FASTCALL call_6_14_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_13_6
#define init_6_14_13_6
void FASTCALL call_6_14_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_13_7
#define init_6_14_13_7
void FASTCALL call_6_14_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_13_8
#define init_6_14_13_8
void FASTCALL call_6_14_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_13_9
#define init_6_14_13_9
void FASTCALL call_6_14_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_13_10
#define init_6_14_13_10
void FASTCALL call_6_14_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_13_11
#define init_6_14_13_11
void FASTCALL call_6_14_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_13_12
#define init_6_14_13_12
void FASTCALL call_6_14_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_13_13
#define init_6_14_13_13
void FASTCALL call_6_14_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_13_14
#define init_6_14_13_14
void FASTCALL call_6_14_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_13_15
#define init_6_14_13_15
void FASTCALL call_6_14_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_14_0
#define init_6_14_14_0
void FASTCALL call_6_14_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_14_1
#define init_6_14_14_1
void FASTCALL call_6_14_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_14_2
#define init_6_14_14_2
void FASTCALL call_6_14_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_14_3
#define init_6_14_14_3
void FASTCALL call_6_14_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_14_4
#define init_6_14_14_4
void FASTCALL call_6_14_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_14_5
#define init_6_14_14_5
void FASTCALL call_6_14_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_14_6
#define init_6_14_14_6
void FASTCALL call_6_14_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_14_7
#define init_6_14_14_7
void FASTCALL call_6_14_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_14_8
#define init_6_14_14_8
void FASTCALL call_6_14_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_14_9
#define init_6_14_14_9
void FASTCALL call_6_14_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_14_10
#define init_6_14_14_10
void FASTCALL call_6_14_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_14_11
#define init_6_14_14_11
void FASTCALL call_6_14_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_14_12
#define init_6_14_14_12
void FASTCALL call_6_14_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_14_13
#define init_6_14_14_13
void FASTCALL call_6_14_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_14_14
#define init_6_14_14_14
void FASTCALL call_6_14_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_14_15
#define init_6_14_14_15
void FASTCALL call_6_14_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_15_0
#define init_6_14_15_0
void FASTCALL call_6_14_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_15_1
#define init_6_14_15_1
void FASTCALL call_6_14_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_15_2
#define init_6_14_15_2
void FASTCALL call_6_14_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_15_3
#define init_6_14_15_3
void FASTCALL call_6_14_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_15_4
#define init_6_14_15_4
void FASTCALL call_6_14_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_15_5
#define init_6_14_15_5
void FASTCALL call_6_14_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_15_6
#define init_6_14_15_6
void FASTCALL call_6_14_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_15_7
#define init_6_14_15_7
void FASTCALL call_6_14_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_15_8
#define init_6_14_15_8
void FASTCALL call_6_14_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_15_9
#define init_6_14_15_9
void FASTCALL call_6_14_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_15_10
#define init_6_14_15_10
void FASTCALL call_6_14_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_15_11
#define init_6_14_15_11
void FASTCALL call_6_14_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_15_12
#define init_6_14_15_12
void FASTCALL call_6_14_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_15_13
#define init_6_14_15_13
void FASTCALL call_6_14_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_15_14
#define init_6_14_15_14
void FASTCALL call_6_14_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_14_15_15
#define init_6_14_15_15
void FASTCALL call_6_14_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_0_0
#define init_6_15_0_0
void FASTCALL call_6_15_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_0_1
#define init_6_15_0_1
void FASTCALL call_6_15_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_0_2
#define init_6_15_0_2
void FASTCALL call_6_15_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_0_3
#define init_6_15_0_3
void FASTCALL call_6_15_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_0_4
#define init_6_15_0_4
void FASTCALL call_6_15_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_0_5
#define init_6_15_0_5
void FASTCALL call_6_15_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_0_6
#define init_6_15_0_6
void FASTCALL call_6_15_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_0_7
#define init_6_15_0_7
void FASTCALL call_6_15_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_0_8
#define init_6_15_0_8
void FASTCALL call_6_15_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_0_9
#define init_6_15_0_9
void FASTCALL call_6_15_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_0_10
#define init_6_15_0_10
void FASTCALL call_6_15_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_0_11
#define init_6_15_0_11
void FASTCALL call_6_15_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_0_12
#define init_6_15_0_12
void FASTCALL call_6_15_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_0_13
#define init_6_15_0_13
void FASTCALL call_6_15_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_0_14
#define init_6_15_0_14
void FASTCALL call_6_15_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_0_15
#define init_6_15_0_15
void FASTCALL call_6_15_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_1_0
#define init_6_15_1_0
void FASTCALL call_6_15_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_1_1
#define init_6_15_1_1
void FASTCALL call_6_15_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_1_2
#define init_6_15_1_2
void FASTCALL call_6_15_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_1_3
#define init_6_15_1_3
void FASTCALL call_6_15_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_1_4
#define init_6_15_1_4
void FASTCALL call_6_15_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_1_5
#define init_6_15_1_5
void FASTCALL call_6_15_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_1_6
#define init_6_15_1_6
void FASTCALL call_6_15_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_1_7
#define init_6_15_1_7
void FASTCALL call_6_15_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_1_8
#define init_6_15_1_8
void FASTCALL call_6_15_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_1_9
#define init_6_15_1_9
void FASTCALL call_6_15_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_1_10
#define init_6_15_1_10
void FASTCALL call_6_15_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_1_11
#define init_6_15_1_11
void FASTCALL call_6_15_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_1_12
#define init_6_15_1_12
void FASTCALL call_6_15_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_1_13
#define init_6_15_1_13
void FASTCALL call_6_15_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_1_14
#define init_6_15_1_14
void FASTCALL call_6_15_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_1_15
#define init_6_15_1_15
void FASTCALL call_6_15_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_2_0
#define init_6_15_2_0
void FASTCALL call_6_15_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_2_1
#define init_6_15_2_1
void FASTCALL call_6_15_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_2_2
#define init_6_15_2_2
void FASTCALL call_6_15_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_2_3
#define init_6_15_2_3
void FASTCALL call_6_15_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_2_4
#define init_6_15_2_4
void FASTCALL call_6_15_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_2_5
#define init_6_15_2_5
void FASTCALL call_6_15_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_2_6
#define init_6_15_2_6
void FASTCALL call_6_15_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_2_7
#define init_6_15_2_7
void FASTCALL call_6_15_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_2_8
#define init_6_15_2_8
void FASTCALL call_6_15_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_2_9
#define init_6_15_2_9
void FASTCALL call_6_15_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_2_10
#define init_6_15_2_10
void FASTCALL call_6_15_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_2_11
#define init_6_15_2_11
void FASTCALL call_6_15_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_2_12
#define init_6_15_2_12
void FASTCALL call_6_15_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_2_13
#define init_6_15_2_13
void FASTCALL call_6_15_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_2_14
#define init_6_15_2_14
void FASTCALL call_6_15_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_2_15
#define init_6_15_2_15
void FASTCALL call_6_15_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_3_0
#define init_6_15_3_0
void FASTCALL call_6_15_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_3_1
#define init_6_15_3_1
void FASTCALL call_6_15_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_3_2
#define init_6_15_3_2
void FASTCALL call_6_15_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_3_3
#define init_6_15_3_3
void FASTCALL call_6_15_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_3_4
#define init_6_15_3_4
void FASTCALL call_6_15_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_3_5
#define init_6_15_3_5
void FASTCALL call_6_15_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_3_6
#define init_6_15_3_6
void FASTCALL call_6_15_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_3_7
#define init_6_15_3_7
void FASTCALL call_6_15_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_3_8
#define init_6_15_3_8
void FASTCALL call_6_15_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_3_9
#define init_6_15_3_9
void FASTCALL call_6_15_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_3_10
#define init_6_15_3_10
void FASTCALL call_6_15_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_3_11
#define init_6_15_3_11
void FASTCALL call_6_15_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_3_12
#define init_6_15_3_12
void FASTCALL call_6_15_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_3_13
#define init_6_15_3_13
void FASTCALL call_6_15_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_3_14
#define init_6_15_3_14
void FASTCALL call_6_15_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_3_15
#define init_6_15_3_15
void FASTCALL call_6_15_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_4_0
#define init_6_15_4_0
void FASTCALL call_6_15_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_4_1
#define init_6_15_4_1
void FASTCALL call_6_15_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_4_2
#define init_6_15_4_2
void FASTCALL call_6_15_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_4_3
#define init_6_15_4_3
void FASTCALL call_6_15_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_4_4
#define init_6_15_4_4
void FASTCALL call_6_15_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_4_5
#define init_6_15_4_5
void FASTCALL call_6_15_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_4_6
#define init_6_15_4_6
void FASTCALL call_6_15_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_4_7
#define init_6_15_4_7
void FASTCALL call_6_15_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_4_8
#define init_6_15_4_8
void FASTCALL call_6_15_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_4_9
#define init_6_15_4_9
void FASTCALL call_6_15_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_4_10
#define init_6_15_4_10
void FASTCALL call_6_15_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_4_11
#define init_6_15_4_11
void FASTCALL call_6_15_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_4_12
#define init_6_15_4_12
void FASTCALL call_6_15_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_4_13
#define init_6_15_4_13
void FASTCALL call_6_15_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_4_14
#define init_6_15_4_14
void FASTCALL call_6_15_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_4_15
#define init_6_15_4_15
void FASTCALL call_6_15_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_5_0
#define init_6_15_5_0
void FASTCALL call_6_15_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_5_1
#define init_6_15_5_1
void FASTCALL call_6_15_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_5_2
#define init_6_15_5_2
void FASTCALL call_6_15_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_5_3
#define init_6_15_5_3
void FASTCALL call_6_15_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_5_4
#define init_6_15_5_4
void FASTCALL call_6_15_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_5_5
#define init_6_15_5_5
void FASTCALL call_6_15_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_5_6
#define init_6_15_5_6
void FASTCALL call_6_15_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_5_7
#define init_6_15_5_7
void FASTCALL call_6_15_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_5_8
#define init_6_15_5_8
void FASTCALL call_6_15_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_5_9
#define init_6_15_5_9
void FASTCALL call_6_15_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_5_10
#define init_6_15_5_10
void FASTCALL call_6_15_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_5_11
#define init_6_15_5_11
void FASTCALL call_6_15_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_5_12
#define init_6_15_5_12
void FASTCALL call_6_15_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_5_13
#define init_6_15_5_13
void FASTCALL call_6_15_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_5_14
#define init_6_15_5_14
void FASTCALL call_6_15_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_5_15
#define init_6_15_5_15
void FASTCALL call_6_15_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_6_0
#define init_6_15_6_0
void FASTCALL call_6_15_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_6_1
#define init_6_15_6_1
void FASTCALL call_6_15_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_6_2
#define init_6_15_6_2
void FASTCALL call_6_15_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_6_3
#define init_6_15_6_3
void FASTCALL call_6_15_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_6_4
#define init_6_15_6_4
void FASTCALL call_6_15_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_6_5
#define init_6_15_6_5
void FASTCALL call_6_15_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_6_6
#define init_6_15_6_6
void FASTCALL call_6_15_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_6_7
#define init_6_15_6_7
void FASTCALL call_6_15_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_6_8
#define init_6_15_6_8
void FASTCALL call_6_15_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_6_9
#define init_6_15_6_9
void FASTCALL call_6_15_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_6_10
#define init_6_15_6_10
void FASTCALL call_6_15_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_6_11
#define init_6_15_6_11
void FASTCALL call_6_15_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_6_12
#define init_6_15_6_12
void FASTCALL call_6_15_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_6_13
#define init_6_15_6_13
void FASTCALL call_6_15_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_6_14
#define init_6_15_6_14
void FASTCALL call_6_15_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_6_15
#define init_6_15_6_15
void FASTCALL call_6_15_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_7_0
#define init_6_15_7_0
void FASTCALL call_6_15_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_7_1
#define init_6_15_7_1
void FASTCALL call_6_15_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_7_2
#define init_6_15_7_2
void FASTCALL call_6_15_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_7_3
#define init_6_15_7_3
void FASTCALL call_6_15_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_7_4
#define init_6_15_7_4
void FASTCALL call_6_15_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_7_5
#define init_6_15_7_5
void FASTCALL call_6_15_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_7_6
#define init_6_15_7_6
void FASTCALL call_6_15_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_7_7
#define init_6_15_7_7
void FASTCALL call_6_15_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_7_8
#define init_6_15_7_8
void FASTCALL call_6_15_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_7_9
#define init_6_15_7_9
void FASTCALL call_6_15_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_7_10
#define init_6_15_7_10
void FASTCALL call_6_15_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_7_11
#define init_6_15_7_11
void FASTCALL call_6_15_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_7_12
#define init_6_15_7_12
void FASTCALL call_6_15_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_7_13
#define init_6_15_7_13
void FASTCALL call_6_15_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_7_14
#define init_6_15_7_14
void FASTCALL call_6_15_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_7_15
#define init_6_15_7_15
void FASTCALL call_6_15_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_8_0
#define init_6_15_8_0
void FASTCALL call_6_15_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_8_1
#define init_6_15_8_1
void FASTCALL call_6_15_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_8_2
#define init_6_15_8_2
void FASTCALL call_6_15_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_8_3
#define init_6_15_8_3
void FASTCALL call_6_15_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_8_4
#define init_6_15_8_4
void FASTCALL call_6_15_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_8_5
#define init_6_15_8_5
void FASTCALL call_6_15_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_8_6
#define init_6_15_8_6
void FASTCALL call_6_15_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_8_7
#define init_6_15_8_7
void FASTCALL call_6_15_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_8_8
#define init_6_15_8_8
void FASTCALL call_6_15_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_8_9
#define init_6_15_8_9
void FASTCALL call_6_15_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_8_10
#define init_6_15_8_10
void FASTCALL call_6_15_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_8_11
#define init_6_15_8_11
void FASTCALL call_6_15_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_8_12
#define init_6_15_8_12
void FASTCALL call_6_15_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_8_13
#define init_6_15_8_13
void FASTCALL call_6_15_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_8_14
#define init_6_15_8_14
void FASTCALL call_6_15_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_8_15
#define init_6_15_8_15
void FASTCALL call_6_15_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_9_0
#define init_6_15_9_0
void FASTCALL call_6_15_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_9_1
#define init_6_15_9_1
void FASTCALL call_6_15_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_9_2
#define init_6_15_9_2
void FASTCALL call_6_15_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_9_3
#define init_6_15_9_3
void FASTCALL call_6_15_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_9_4
#define init_6_15_9_4
void FASTCALL call_6_15_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_9_5
#define init_6_15_9_5
void FASTCALL call_6_15_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_9_6
#define init_6_15_9_6
void FASTCALL call_6_15_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_9_7
#define init_6_15_9_7
void FASTCALL call_6_15_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_9_8
#define init_6_15_9_8
void FASTCALL call_6_15_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_9_9
#define init_6_15_9_9
void FASTCALL call_6_15_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_9_10
#define init_6_15_9_10
void FASTCALL call_6_15_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_9_11
#define init_6_15_9_11
void FASTCALL call_6_15_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_9_12
#define init_6_15_9_12
void FASTCALL call_6_15_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_9_13
#define init_6_15_9_13
void FASTCALL call_6_15_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_9_14
#define init_6_15_9_14
void FASTCALL call_6_15_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_9_15
#define init_6_15_9_15
void FASTCALL call_6_15_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_10_0
#define init_6_15_10_0
void FASTCALL call_6_15_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_10_1
#define init_6_15_10_1
void FASTCALL call_6_15_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_10_2
#define init_6_15_10_2
void FASTCALL call_6_15_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_10_3
#define init_6_15_10_3
void FASTCALL call_6_15_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_10_4
#define init_6_15_10_4
void FASTCALL call_6_15_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_10_5
#define init_6_15_10_5
void FASTCALL call_6_15_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_10_6
#define init_6_15_10_6
void FASTCALL call_6_15_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_10_7
#define init_6_15_10_7
void FASTCALL call_6_15_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_10_8
#define init_6_15_10_8
void FASTCALL call_6_15_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_10_9
#define init_6_15_10_9
void FASTCALL call_6_15_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_10_10
#define init_6_15_10_10
void FASTCALL call_6_15_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_10_11
#define init_6_15_10_11
void FASTCALL call_6_15_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_10_12
#define init_6_15_10_12
void FASTCALL call_6_15_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_10_13
#define init_6_15_10_13
void FASTCALL call_6_15_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_10_14
#define init_6_15_10_14
void FASTCALL call_6_15_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_10_15
#define init_6_15_10_15
void FASTCALL call_6_15_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_11_0
#define init_6_15_11_0
void FASTCALL call_6_15_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_11_1
#define init_6_15_11_1
void FASTCALL call_6_15_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_11_2
#define init_6_15_11_2
void FASTCALL call_6_15_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_11_3
#define init_6_15_11_3
void FASTCALL call_6_15_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_11_4
#define init_6_15_11_4
void FASTCALL call_6_15_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_11_5
#define init_6_15_11_5
void FASTCALL call_6_15_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_11_6
#define init_6_15_11_6
void FASTCALL call_6_15_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_11_7
#define init_6_15_11_7
void FASTCALL call_6_15_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_11_8
#define init_6_15_11_8
void FASTCALL call_6_15_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_11_9
#define init_6_15_11_9
void FASTCALL call_6_15_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_11_10
#define init_6_15_11_10
void FASTCALL call_6_15_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_11_11
#define init_6_15_11_11
void FASTCALL call_6_15_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_11_12
#define init_6_15_11_12
void FASTCALL call_6_15_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_11_13
#define init_6_15_11_13
void FASTCALL call_6_15_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_11_14
#define init_6_15_11_14
void FASTCALL call_6_15_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_11_15
#define init_6_15_11_15
void FASTCALL call_6_15_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_12_0
#define init_6_15_12_0
void FASTCALL call_6_15_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_12_1
#define init_6_15_12_1
void FASTCALL call_6_15_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_12_2
#define init_6_15_12_2
void FASTCALL call_6_15_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_12_3
#define init_6_15_12_3
void FASTCALL call_6_15_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_12_4
#define init_6_15_12_4
void FASTCALL call_6_15_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_12_5
#define init_6_15_12_5
void FASTCALL call_6_15_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_12_6
#define init_6_15_12_6
void FASTCALL call_6_15_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_12_7
#define init_6_15_12_7
void FASTCALL call_6_15_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_12_8
#define init_6_15_12_8
void FASTCALL call_6_15_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_12_9
#define init_6_15_12_9
void FASTCALL call_6_15_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_12_10
#define init_6_15_12_10
void FASTCALL call_6_15_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_12_11
#define init_6_15_12_11
void FASTCALL call_6_15_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_12_12
#define init_6_15_12_12
void FASTCALL call_6_15_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_12_13
#define init_6_15_12_13
void FASTCALL call_6_15_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_12_14
#define init_6_15_12_14
void FASTCALL call_6_15_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_12_15
#define init_6_15_12_15
void FASTCALL call_6_15_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_13_0
#define init_6_15_13_0
void FASTCALL call_6_15_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_13_1
#define init_6_15_13_1
void FASTCALL call_6_15_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_13_2
#define init_6_15_13_2
void FASTCALL call_6_15_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_13_3
#define init_6_15_13_3
void FASTCALL call_6_15_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_13_4
#define init_6_15_13_4
void FASTCALL call_6_15_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_13_5
#define init_6_15_13_5
void FASTCALL call_6_15_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_13_6
#define init_6_15_13_6
void FASTCALL call_6_15_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_13_7
#define init_6_15_13_7
void FASTCALL call_6_15_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_13_8
#define init_6_15_13_8
void FASTCALL call_6_15_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_13_9
#define init_6_15_13_9
void FASTCALL call_6_15_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_13_10
#define init_6_15_13_10
void FASTCALL call_6_15_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_13_11
#define init_6_15_13_11
void FASTCALL call_6_15_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_13_12
#define init_6_15_13_12
void FASTCALL call_6_15_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_13_13
#define init_6_15_13_13
void FASTCALL call_6_15_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_13_14
#define init_6_15_13_14
void FASTCALL call_6_15_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_13_15
#define init_6_15_13_15
void FASTCALL call_6_15_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_14_0
#define init_6_15_14_0
void FASTCALL call_6_15_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_14_1
#define init_6_15_14_1
void FASTCALL call_6_15_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_14_2
#define init_6_15_14_2
void FASTCALL call_6_15_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_14_3
#define init_6_15_14_3
void FASTCALL call_6_15_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_14_4
#define init_6_15_14_4
void FASTCALL call_6_15_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_14_5
#define init_6_15_14_5
void FASTCALL call_6_15_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_14_6
#define init_6_15_14_6
void FASTCALL call_6_15_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_14_7
#define init_6_15_14_7
void FASTCALL call_6_15_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_14_8
#define init_6_15_14_8
void FASTCALL call_6_15_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_14_9
#define init_6_15_14_9
void FASTCALL call_6_15_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_14_10
#define init_6_15_14_10
void FASTCALL call_6_15_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_14_11
#define init_6_15_14_11
void FASTCALL call_6_15_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_14_12
#define init_6_15_14_12
void FASTCALL call_6_15_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_14_13
#define init_6_15_14_13
void FASTCALL call_6_15_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_14_14
#define init_6_15_14_14
void FASTCALL call_6_15_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_14_15
#define init_6_15_14_15
void FASTCALL call_6_15_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_15_0
#define init_6_15_15_0
void FASTCALL call_6_15_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_15_1
#define init_6_15_15_1
void FASTCALL call_6_15_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_15_2
#define init_6_15_15_2
void FASTCALL call_6_15_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_15_3
#define init_6_15_15_3
void FASTCALL call_6_15_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_15_4
#define init_6_15_15_4
void FASTCALL call_6_15_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_15_5
#define init_6_15_15_5
void FASTCALL call_6_15_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_15_6
#define init_6_15_15_6
void FASTCALL call_6_15_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_15_7
#define init_6_15_15_7
void FASTCALL call_6_15_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_15_8
#define init_6_15_15_8
void FASTCALL call_6_15_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_15_9
#define init_6_15_15_9
void FASTCALL call_6_15_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_15_10
#define init_6_15_15_10
void FASTCALL call_6_15_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_15_11
#define init_6_15_15_11
void FASTCALL call_6_15_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_15_12
#define init_6_15_15_12
void FASTCALL call_6_15_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_15_13
#define init_6_15_15_13
void FASTCALL call_6_15_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_15_14
#define init_6_15_15_14
void FASTCALL call_6_15_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_6_15_15_15
#define init_6_15_15_15
void FASTCALL call_6_15_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_0_0
#define init_7_0_0_0
void FASTCALL call_7_0_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_0_1
#define init_7_0_0_1
void FASTCALL call_7_0_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_0_2
#define init_7_0_0_2
void FASTCALL call_7_0_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_0_3
#define init_7_0_0_3
void FASTCALL call_7_0_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_0_4
#define init_7_0_0_4
void FASTCALL call_7_0_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_0_5
#define init_7_0_0_5
void FASTCALL call_7_0_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_0_6
#define init_7_0_0_6
void FASTCALL call_7_0_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_0_7
#define init_7_0_0_7
void FASTCALL call_7_0_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_0_8
#define init_7_0_0_8
void FASTCALL call_7_0_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_0_9
#define init_7_0_0_9
void FASTCALL call_7_0_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_0_10
#define init_7_0_0_10
void FASTCALL call_7_0_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_0_11
#define init_7_0_0_11
void FASTCALL call_7_0_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_0_12
#define init_7_0_0_12
void FASTCALL call_7_0_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_0_13
#define init_7_0_0_13
void FASTCALL call_7_0_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_0_14
#define init_7_0_0_14
void FASTCALL call_7_0_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_0_15
#define init_7_0_0_15
void FASTCALL call_7_0_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_1_0
#define init_7_0_1_0
void FASTCALL call_7_0_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_1_1
#define init_7_0_1_1
void FASTCALL call_7_0_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_1_2
#define init_7_0_1_2
void FASTCALL call_7_0_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_1_3
#define init_7_0_1_3
void FASTCALL call_7_0_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_1_4
#define init_7_0_1_4
void FASTCALL call_7_0_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_1_5
#define init_7_0_1_5
void FASTCALL call_7_0_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_1_6
#define init_7_0_1_6
void FASTCALL call_7_0_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_1_7
#define init_7_0_1_7
void FASTCALL call_7_0_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_1_8
#define init_7_0_1_8
void FASTCALL call_7_0_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_1_9
#define init_7_0_1_9
void FASTCALL call_7_0_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_1_10
#define init_7_0_1_10
void FASTCALL call_7_0_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_1_11
#define init_7_0_1_11
void FASTCALL call_7_0_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_1_12
#define init_7_0_1_12
void FASTCALL call_7_0_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_1_13
#define init_7_0_1_13
void FASTCALL call_7_0_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_1_14
#define init_7_0_1_14
void FASTCALL call_7_0_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_1_15
#define init_7_0_1_15
void FASTCALL call_7_0_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_2_0
#define init_7_0_2_0
void FASTCALL call_7_0_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_2_1
#define init_7_0_2_1
void FASTCALL call_7_0_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_2_2
#define init_7_0_2_2
void FASTCALL call_7_0_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_2_3
#define init_7_0_2_3
void FASTCALL call_7_0_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_2_4
#define init_7_0_2_4
void FASTCALL call_7_0_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_2_5
#define init_7_0_2_5
void FASTCALL call_7_0_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_2_6
#define init_7_0_2_6
void FASTCALL call_7_0_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_2_7
#define init_7_0_2_7
void FASTCALL call_7_0_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_2_8
#define init_7_0_2_8
void FASTCALL call_7_0_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_2_9
#define init_7_0_2_9
void FASTCALL call_7_0_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_2_10
#define init_7_0_2_10
void FASTCALL call_7_0_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_2_11
#define init_7_0_2_11
void FASTCALL call_7_0_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_2_12
#define init_7_0_2_12
void FASTCALL call_7_0_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_2_13
#define init_7_0_2_13
void FASTCALL call_7_0_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_2_14
#define init_7_0_2_14
void FASTCALL call_7_0_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_2_15
#define init_7_0_2_15
void FASTCALL call_7_0_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_3_0
#define init_7_0_3_0
void FASTCALL call_7_0_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_3_1
#define init_7_0_3_1
void FASTCALL call_7_0_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_3_2
#define init_7_0_3_2
void FASTCALL call_7_0_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_3_3
#define init_7_0_3_3
void FASTCALL call_7_0_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_3_4
#define init_7_0_3_4
void FASTCALL call_7_0_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_3_5
#define init_7_0_3_5
void FASTCALL call_7_0_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_3_6
#define init_7_0_3_6
void FASTCALL call_7_0_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_3_7
#define init_7_0_3_7
void FASTCALL call_7_0_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_3_8
#define init_7_0_3_8
void FASTCALL call_7_0_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_3_9
#define init_7_0_3_9
void FASTCALL call_7_0_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_3_10
#define init_7_0_3_10
void FASTCALL call_7_0_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_3_11
#define init_7_0_3_11
void FASTCALL call_7_0_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_3_12
#define init_7_0_3_12
void FASTCALL call_7_0_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_3_13
#define init_7_0_3_13
void FASTCALL call_7_0_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_3_14
#define init_7_0_3_14
void FASTCALL call_7_0_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_3_15
#define init_7_0_3_15
void FASTCALL call_7_0_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_4_0
#define init_7_0_4_0
void FASTCALL call_7_0_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_4_1
#define init_7_0_4_1
void FASTCALL call_7_0_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_4_2
#define init_7_0_4_2
void FASTCALL call_7_0_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_4_3
#define init_7_0_4_3
void FASTCALL call_7_0_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_4_4
#define init_7_0_4_4
void FASTCALL call_7_0_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_4_5
#define init_7_0_4_5
void FASTCALL call_7_0_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_4_6
#define init_7_0_4_6
void FASTCALL call_7_0_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_4_7
#define init_7_0_4_7
void FASTCALL call_7_0_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_4_8
#define init_7_0_4_8
void FASTCALL call_7_0_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_4_9
#define init_7_0_4_9
void FASTCALL call_7_0_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_4_10
#define init_7_0_4_10
void FASTCALL call_7_0_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_4_11
#define init_7_0_4_11
void FASTCALL call_7_0_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_4_12
#define init_7_0_4_12
void FASTCALL call_7_0_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_4_13
#define init_7_0_4_13
void FASTCALL call_7_0_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_4_14
#define init_7_0_4_14
void FASTCALL call_7_0_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_4_15
#define init_7_0_4_15
void FASTCALL call_7_0_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_5_0
#define init_7_0_5_0
void FASTCALL call_7_0_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_5_1
#define init_7_0_5_1
void FASTCALL call_7_0_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_5_2
#define init_7_0_5_2
void FASTCALL call_7_0_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_5_3
#define init_7_0_5_3
void FASTCALL call_7_0_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_5_4
#define init_7_0_5_4
void FASTCALL call_7_0_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_5_5
#define init_7_0_5_5
void FASTCALL call_7_0_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_5_6
#define init_7_0_5_6
void FASTCALL call_7_0_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_5_7
#define init_7_0_5_7
void FASTCALL call_7_0_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_5_8
#define init_7_0_5_8
void FASTCALL call_7_0_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_5_9
#define init_7_0_5_9
void FASTCALL call_7_0_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_5_10
#define init_7_0_5_10
void FASTCALL call_7_0_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_5_11
#define init_7_0_5_11
void FASTCALL call_7_0_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_5_12
#define init_7_0_5_12
void FASTCALL call_7_0_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_5_13
#define init_7_0_5_13
void FASTCALL call_7_0_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_5_14
#define init_7_0_5_14
void FASTCALL call_7_0_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_5_15
#define init_7_0_5_15
void FASTCALL call_7_0_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_6_0
#define init_7_0_6_0
void FASTCALL call_7_0_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_6_1
#define init_7_0_6_1
void FASTCALL call_7_0_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_6_2
#define init_7_0_6_2
void FASTCALL call_7_0_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_6_3
#define init_7_0_6_3
void FASTCALL call_7_0_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_6_4
#define init_7_0_6_4
void FASTCALL call_7_0_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_6_5
#define init_7_0_6_5
void FASTCALL call_7_0_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_6_6
#define init_7_0_6_6
void FASTCALL call_7_0_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_6_7
#define init_7_0_6_7
void FASTCALL call_7_0_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_6_8
#define init_7_0_6_8
void FASTCALL call_7_0_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_6_9
#define init_7_0_6_9
void FASTCALL call_7_0_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_6_10
#define init_7_0_6_10
void FASTCALL call_7_0_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_6_11
#define init_7_0_6_11
void FASTCALL call_7_0_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_6_12
#define init_7_0_6_12
void FASTCALL call_7_0_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_6_13
#define init_7_0_6_13
void FASTCALL call_7_0_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_6_14
#define init_7_0_6_14
void FASTCALL call_7_0_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_6_15
#define init_7_0_6_15
void FASTCALL call_7_0_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_7_0
#define init_7_0_7_0
void FASTCALL call_7_0_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_7_1
#define init_7_0_7_1
void FASTCALL call_7_0_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_7_2
#define init_7_0_7_2
void FASTCALL call_7_0_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_7_3
#define init_7_0_7_3
void FASTCALL call_7_0_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_7_4
#define init_7_0_7_4
void FASTCALL call_7_0_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_7_5
#define init_7_0_7_5
void FASTCALL call_7_0_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_7_6
#define init_7_0_7_6
void FASTCALL call_7_0_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_7_7
#define init_7_0_7_7
void FASTCALL call_7_0_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_7_8
#define init_7_0_7_8
void FASTCALL call_7_0_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_7_9
#define init_7_0_7_9
void FASTCALL call_7_0_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_7_10
#define init_7_0_7_10
void FASTCALL call_7_0_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_7_11
#define init_7_0_7_11
void FASTCALL call_7_0_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_7_12
#define init_7_0_7_12
void FASTCALL call_7_0_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_7_13
#define init_7_0_7_13
void FASTCALL call_7_0_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_7_14
#define init_7_0_7_14
void FASTCALL call_7_0_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_7_15
#define init_7_0_7_15
void FASTCALL call_7_0_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_8_0
#define init_7_0_8_0
void FASTCALL call_7_0_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_8_1
#define init_7_0_8_1
void FASTCALL call_7_0_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_8_2
#define init_7_0_8_2
void FASTCALL call_7_0_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_8_3
#define init_7_0_8_3
void FASTCALL call_7_0_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_8_4
#define init_7_0_8_4
void FASTCALL call_7_0_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_8_5
#define init_7_0_8_5
void FASTCALL call_7_0_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_8_6
#define init_7_0_8_6
void FASTCALL call_7_0_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_8_7
#define init_7_0_8_7
void FASTCALL call_7_0_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_8_8
#define init_7_0_8_8
void FASTCALL call_7_0_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_8_9
#define init_7_0_8_9
void FASTCALL call_7_0_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_8_10
#define init_7_0_8_10
void FASTCALL call_7_0_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_8_11
#define init_7_0_8_11
void FASTCALL call_7_0_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_8_12
#define init_7_0_8_12
void FASTCALL call_7_0_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_8_13
#define init_7_0_8_13
void FASTCALL call_7_0_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_8_14
#define init_7_0_8_14
void FASTCALL call_7_0_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_8_15
#define init_7_0_8_15
void FASTCALL call_7_0_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_9_0
#define init_7_0_9_0
void FASTCALL call_7_0_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_9_1
#define init_7_0_9_1
void FASTCALL call_7_0_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_9_2
#define init_7_0_9_2
void FASTCALL call_7_0_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_9_3
#define init_7_0_9_3
void FASTCALL call_7_0_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_9_4
#define init_7_0_9_4
void FASTCALL call_7_0_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_9_5
#define init_7_0_9_5
void FASTCALL call_7_0_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_9_6
#define init_7_0_9_6
void FASTCALL call_7_0_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_9_7
#define init_7_0_9_7
void FASTCALL call_7_0_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_9_8
#define init_7_0_9_8
void FASTCALL call_7_0_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_9_9
#define init_7_0_9_9
void FASTCALL call_7_0_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_9_10
#define init_7_0_9_10
void FASTCALL call_7_0_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_9_11
#define init_7_0_9_11
void FASTCALL call_7_0_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_9_12
#define init_7_0_9_12
void FASTCALL call_7_0_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_9_13
#define init_7_0_9_13
void FASTCALL call_7_0_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_9_14
#define init_7_0_9_14
void FASTCALL call_7_0_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_9_15
#define init_7_0_9_15
void FASTCALL call_7_0_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_10_0
#define init_7_0_10_0
void FASTCALL call_7_0_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_10_1
#define init_7_0_10_1
void FASTCALL call_7_0_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_10_2
#define init_7_0_10_2
void FASTCALL call_7_0_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_10_3
#define init_7_0_10_3
void FASTCALL call_7_0_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_10_4
#define init_7_0_10_4
void FASTCALL call_7_0_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_10_5
#define init_7_0_10_5
void FASTCALL call_7_0_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_10_6
#define init_7_0_10_6
void FASTCALL call_7_0_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_10_7
#define init_7_0_10_7
void FASTCALL call_7_0_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_10_8
#define init_7_0_10_8
void FASTCALL call_7_0_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_10_9
#define init_7_0_10_9
void FASTCALL call_7_0_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_10_10
#define init_7_0_10_10
void FASTCALL call_7_0_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_10_11
#define init_7_0_10_11
void FASTCALL call_7_0_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_10_12
#define init_7_0_10_12
void FASTCALL call_7_0_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_10_13
#define init_7_0_10_13
void FASTCALL call_7_0_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_10_14
#define init_7_0_10_14
void FASTCALL call_7_0_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_10_15
#define init_7_0_10_15
void FASTCALL call_7_0_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_11_0
#define init_7_0_11_0
void FASTCALL call_7_0_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_11_1
#define init_7_0_11_1
void FASTCALL call_7_0_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_11_2
#define init_7_0_11_2
void FASTCALL call_7_0_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_11_3
#define init_7_0_11_3
void FASTCALL call_7_0_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_11_4
#define init_7_0_11_4
void FASTCALL call_7_0_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_11_5
#define init_7_0_11_5
void FASTCALL call_7_0_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_11_6
#define init_7_0_11_6
void FASTCALL call_7_0_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_11_7
#define init_7_0_11_7
void FASTCALL call_7_0_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_11_8
#define init_7_0_11_8
void FASTCALL call_7_0_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_11_9
#define init_7_0_11_9
void FASTCALL call_7_0_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_11_10
#define init_7_0_11_10
void FASTCALL call_7_0_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_11_11
#define init_7_0_11_11
void FASTCALL call_7_0_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_11_12
#define init_7_0_11_12
void FASTCALL call_7_0_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_11_13
#define init_7_0_11_13
void FASTCALL call_7_0_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_11_14
#define init_7_0_11_14
void FASTCALL call_7_0_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_11_15
#define init_7_0_11_15
void FASTCALL call_7_0_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_12_0
#define init_7_0_12_0
void FASTCALL call_7_0_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_12_1
#define init_7_0_12_1
void FASTCALL call_7_0_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_12_2
#define init_7_0_12_2
void FASTCALL call_7_0_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_12_3
#define init_7_0_12_3
void FASTCALL call_7_0_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_12_4
#define init_7_0_12_4
void FASTCALL call_7_0_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_12_5
#define init_7_0_12_5
void FASTCALL call_7_0_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_12_6
#define init_7_0_12_6
void FASTCALL call_7_0_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_12_7
#define init_7_0_12_7
void FASTCALL call_7_0_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_12_8
#define init_7_0_12_8
void FASTCALL call_7_0_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_12_9
#define init_7_0_12_9
void FASTCALL call_7_0_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_12_10
#define init_7_0_12_10
void FASTCALL call_7_0_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_12_11
#define init_7_0_12_11
void FASTCALL call_7_0_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_12_12
#define init_7_0_12_12
void FASTCALL call_7_0_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_12_13
#define init_7_0_12_13
void FASTCALL call_7_0_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_12_14
#define init_7_0_12_14
void FASTCALL call_7_0_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_12_15
#define init_7_0_12_15
void FASTCALL call_7_0_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_13_0
#define init_7_0_13_0
void FASTCALL call_7_0_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_13_1
#define init_7_0_13_1
void FASTCALL call_7_0_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_13_2
#define init_7_0_13_2
void FASTCALL call_7_0_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_13_3
#define init_7_0_13_3
void FASTCALL call_7_0_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_13_4
#define init_7_0_13_4
void FASTCALL call_7_0_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_13_5
#define init_7_0_13_5
void FASTCALL call_7_0_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_13_6
#define init_7_0_13_6
void FASTCALL call_7_0_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_13_7
#define init_7_0_13_7
void FASTCALL call_7_0_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_13_8
#define init_7_0_13_8
void FASTCALL call_7_0_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_13_9
#define init_7_0_13_9
void FASTCALL call_7_0_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_13_10
#define init_7_0_13_10
void FASTCALL call_7_0_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_13_11
#define init_7_0_13_11
void FASTCALL call_7_0_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_13_12
#define init_7_0_13_12
void FASTCALL call_7_0_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_13_13
#define init_7_0_13_13
void FASTCALL call_7_0_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_13_14
#define init_7_0_13_14
void FASTCALL call_7_0_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_13_15
#define init_7_0_13_15
void FASTCALL call_7_0_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_14_0
#define init_7_0_14_0
void FASTCALL call_7_0_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_14_1
#define init_7_0_14_1
void FASTCALL call_7_0_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_14_2
#define init_7_0_14_2
void FASTCALL call_7_0_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_14_3
#define init_7_0_14_3
void FASTCALL call_7_0_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_14_4
#define init_7_0_14_4
void FASTCALL call_7_0_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_14_5
#define init_7_0_14_5
void FASTCALL call_7_0_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_14_6
#define init_7_0_14_6
void FASTCALL call_7_0_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_14_7
#define init_7_0_14_7
void FASTCALL call_7_0_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_14_8
#define init_7_0_14_8
void FASTCALL call_7_0_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_14_9
#define init_7_0_14_9
void FASTCALL call_7_0_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_14_10
#define init_7_0_14_10
void FASTCALL call_7_0_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_14_11
#define init_7_0_14_11
void FASTCALL call_7_0_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_14_12
#define init_7_0_14_12
void FASTCALL call_7_0_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_14_13
#define init_7_0_14_13
void FASTCALL call_7_0_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_14_14
#define init_7_0_14_14
void FASTCALL call_7_0_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_14_15
#define init_7_0_14_15
void FASTCALL call_7_0_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_15_0
#define init_7_0_15_0
void FASTCALL call_7_0_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_15_1
#define init_7_0_15_1
void FASTCALL call_7_0_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_15_2
#define init_7_0_15_2
void FASTCALL call_7_0_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_15_3
#define init_7_0_15_3
void FASTCALL call_7_0_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_15_4
#define init_7_0_15_4
void FASTCALL call_7_0_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_15_5
#define init_7_0_15_5
void FASTCALL call_7_0_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_15_6
#define init_7_0_15_6
void FASTCALL call_7_0_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_15_7
#define init_7_0_15_7
void FASTCALL call_7_0_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_15_8
#define init_7_0_15_8
void FASTCALL call_7_0_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_15_9
#define init_7_0_15_9
void FASTCALL call_7_0_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_15_10
#define init_7_0_15_10
void FASTCALL call_7_0_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_15_11
#define init_7_0_15_11
void FASTCALL call_7_0_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_15_12
#define init_7_0_15_12
void FASTCALL call_7_0_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_15_13
#define init_7_0_15_13
void FASTCALL call_7_0_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_15_14
#define init_7_0_15_14
void FASTCALL call_7_0_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_0_15_15
#define init_7_0_15_15
void FASTCALL call_7_0_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_0_0
#define init_7_1_0_0
void FASTCALL call_7_1_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_0_1
#define init_7_1_0_1
void FASTCALL call_7_1_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_0_2
#define init_7_1_0_2
void FASTCALL call_7_1_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_0_3
#define init_7_1_0_3
void FASTCALL call_7_1_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_0_4
#define init_7_1_0_4
void FASTCALL call_7_1_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_0_5
#define init_7_1_0_5
void FASTCALL call_7_1_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_0_6
#define init_7_1_0_6
void FASTCALL call_7_1_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_0_7
#define init_7_1_0_7
void FASTCALL call_7_1_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_0_8
#define init_7_1_0_8
void FASTCALL call_7_1_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_0_9
#define init_7_1_0_9
void FASTCALL call_7_1_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_0_10
#define init_7_1_0_10
void FASTCALL call_7_1_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_0_11
#define init_7_1_0_11
void FASTCALL call_7_1_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_0_12
#define init_7_1_0_12
void FASTCALL call_7_1_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_0_13
#define init_7_1_0_13
void FASTCALL call_7_1_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_0_14
#define init_7_1_0_14
void FASTCALL call_7_1_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_0_15
#define init_7_1_0_15
void FASTCALL call_7_1_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_1_0
#define init_7_1_1_0
void FASTCALL call_7_1_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_1_1
#define init_7_1_1_1
void FASTCALL call_7_1_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_1_2
#define init_7_1_1_2
void FASTCALL call_7_1_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_1_3
#define init_7_1_1_3
void FASTCALL call_7_1_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_1_4
#define init_7_1_1_4
void FASTCALL call_7_1_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_1_5
#define init_7_1_1_5
void FASTCALL call_7_1_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_1_6
#define init_7_1_1_6
void FASTCALL call_7_1_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_1_7
#define init_7_1_1_7
void FASTCALL call_7_1_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_1_8
#define init_7_1_1_8
void FASTCALL call_7_1_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_1_9
#define init_7_1_1_9
void FASTCALL call_7_1_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_1_10
#define init_7_1_1_10
void FASTCALL call_7_1_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_1_11
#define init_7_1_1_11
void FASTCALL call_7_1_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_1_12
#define init_7_1_1_12
void FASTCALL call_7_1_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_1_13
#define init_7_1_1_13
void FASTCALL call_7_1_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_1_14
#define init_7_1_1_14
void FASTCALL call_7_1_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_1_15
#define init_7_1_1_15
void FASTCALL call_7_1_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_2_0
#define init_7_1_2_0
void FASTCALL call_7_1_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_2_1
#define init_7_1_2_1
void FASTCALL call_7_1_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_2_2
#define init_7_1_2_2
void FASTCALL call_7_1_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_2_3
#define init_7_1_2_3
void FASTCALL call_7_1_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_2_4
#define init_7_1_2_4
void FASTCALL call_7_1_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_2_5
#define init_7_1_2_5
void FASTCALL call_7_1_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_2_6
#define init_7_1_2_6
void FASTCALL call_7_1_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_2_7
#define init_7_1_2_7
void FASTCALL call_7_1_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_2_8
#define init_7_1_2_8
void FASTCALL call_7_1_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_2_9
#define init_7_1_2_9
void FASTCALL call_7_1_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_2_10
#define init_7_1_2_10
void FASTCALL call_7_1_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_2_11
#define init_7_1_2_11
void FASTCALL call_7_1_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_2_12
#define init_7_1_2_12
void FASTCALL call_7_1_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_2_13
#define init_7_1_2_13
void FASTCALL call_7_1_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_2_14
#define init_7_1_2_14
void FASTCALL call_7_1_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_2_15
#define init_7_1_2_15
void FASTCALL call_7_1_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_3_0
#define init_7_1_3_0
void FASTCALL call_7_1_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_3_1
#define init_7_1_3_1
void FASTCALL call_7_1_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_3_2
#define init_7_1_3_2
void FASTCALL call_7_1_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_3_3
#define init_7_1_3_3
void FASTCALL call_7_1_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_3_4
#define init_7_1_3_4
void FASTCALL call_7_1_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_3_5
#define init_7_1_3_5
void FASTCALL call_7_1_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_3_6
#define init_7_1_3_6
void FASTCALL call_7_1_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_3_7
#define init_7_1_3_7
void FASTCALL call_7_1_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_3_8
#define init_7_1_3_8
void FASTCALL call_7_1_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_3_9
#define init_7_1_3_9
void FASTCALL call_7_1_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_3_10
#define init_7_1_3_10
void FASTCALL call_7_1_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_3_11
#define init_7_1_3_11
void FASTCALL call_7_1_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_3_12
#define init_7_1_3_12
void FASTCALL call_7_1_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_3_13
#define init_7_1_3_13
void FASTCALL call_7_1_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_3_14
#define init_7_1_3_14
void FASTCALL call_7_1_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_3_15
#define init_7_1_3_15
void FASTCALL call_7_1_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_4_0
#define init_7_1_4_0
void FASTCALL call_7_1_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_4_1
#define init_7_1_4_1
void FASTCALL call_7_1_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_4_2
#define init_7_1_4_2
void FASTCALL call_7_1_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_4_3
#define init_7_1_4_3
void FASTCALL call_7_1_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_4_4
#define init_7_1_4_4
void FASTCALL call_7_1_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_4_5
#define init_7_1_4_5
void FASTCALL call_7_1_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_4_6
#define init_7_1_4_6
void FASTCALL call_7_1_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_4_7
#define init_7_1_4_7
void FASTCALL call_7_1_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_4_8
#define init_7_1_4_8
void FASTCALL call_7_1_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_4_9
#define init_7_1_4_9
void FASTCALL call_7_1_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_4_10
#define init_7_1_4_10
void FASTCALL call_7_1_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_4_11
#define init_7_1_4_11
void FASTCALL call_7_1_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_4_12
#define init_7_1_4_12
void FASTCALL call_7_1_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_4_13
#define init_7_1_4_13
void FASTCALL call_7_1_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_4_14
#define init_7_1_4_14
void FASTCALL call_7_1_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_4_15
#define init_7_1_4_15
void FASTCALL call_7_1_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_5_0
#define init_7_1_5_0
void FASTCALL call_7_1_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_5_1
#define init_7_1_5_1
void FASTCALL call_7_1_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_5_2
#define init_7_1_5_2
void FASTCALL call_7_1_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_5_3
#define init_7_1_5_3
void FASTCALL call_7_1_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_5_4
#define init_7_1_5_4
void FASTCALL call_7_1_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_5_5
#define init_7_1_5_5
void FASTCALL call_7_1_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_5_6
#define init_7_1_5_6
void FASTCALL call_7_1_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_5_7
#define init_7_1_5_7
void FASTCALL call_7_1_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_5_8
#define init_7_1_5_8
void FASTCALL call_7_1_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_5_9
#define init_7_1_5_9
void FASTCALL call_7_1_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_5_10
#define init_7_1_5_10
void FASTCALL call_7_1_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_5_11
#define init_7_1_5_11
void FASTCALL call_7_1_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_5_12
#define init_7_1_5_12
void FASTCALL call_7_1_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_5_13
#define init_7_1_5_13
void FASTCALL call_7_1_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_5_14
#define init_7_1_5_14
void FASTCALL call_7_1_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_5_15
#define init_7_1_5_15
void FASTCALL call_7_1_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_6_0
#define init_7_1_6_0
void FASTCALL call_7_1_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_6_1
#define init_7_1_6_1
void FASTCALL call_7_1_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_6_2
#define init_7_1_6_2
void FASTCALL call_7_1_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_6_3
#define init_7_1_6_3
void FASTCALL call_7_1_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_6_4
#define init_7_1_6_4
void FASTCALL call_7_1_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_6_5
#define init_7_1_6_5
void FASTCALL call_7_1_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_6_6
#define init_7_1_6_6
void FASTCALL call_7_1_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_6_7
#define init_7_1_6_7
void FASTCALL call_7_1_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_6_8
#define init_7_1_6_8
void FASTCALL call_7_1_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_6_9
#define init_7_1_6_9
void FASTCALL call_7_1_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_6_10
#define init_7_1_6_10
void FASTCALL call_7_1_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_6_11
#define init_7_1_6_11
void FASTCALL call_7_1_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_6_12
#define init_7_1_6_12
void FASTCALL call_7_1_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_6_13
#define init_7_1_6_13
void FASTCALL call_7_1_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_6_14
#define init_7_1_6_14
void FASTCALL call_7_1_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_6_15
#define init_7_1_6_15
void FASTCALL call_7_1_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_7_0
#define init_7_1_7_0
void FASTCALL call_7_1_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_7_1
#define init_7_1_7_1
void FASTCALL call_7_1_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_7_2
#define init_7_1_7_2
void FASTCALL call_7_1_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_7_3
#define init_7_1_7_3
void FASTCALL call_7_1_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_7_4
#define init_7_1_7_4
void FASTCALL call_7_1_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_7_5
#define init_7_1_7_5
void FASTCALL call_7_1_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_7_6
#define init_7_1_7_6
void FASTCALL call_7_1_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_7_7
#define init_7_1_7_7
void FASTCALL call_7_1_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_7_8
#define init_7_1_7_8
void FASTCALL call_7_1_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_7_9
#define init_7_1_7_9
void FASTCALL call_7_1_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_7_10
#define init_7_1_7_10
void FASTCALL call_7_1_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_7_11
#define init_7_1_7_11
void FASTCALL call_7_1_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_7_12
#define init_7_1_7_12
void FASTCALL call_7_1_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_7_13
#define init_7_1_7_13
void FASTCALL call_7_1_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_7_14
#define init_7_1_7_14
void FASTCALL call_7_1_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_7_15
#define init_7_1_7_15
void FASTCALL call_7_1_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_8_0
#define init_7_1_8_0
void FASTCALL call_7_1_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_8_1
#define init_7_1_8_1
void FASTCALL call_7_1_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_8_2
#define init_7_1_8_2
void FASTCALL call_7_1_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_8_3
#define init_7_1_8_3
void FASTCALL call_7_1_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_8_4
#define init_7_1_8_4
void FASTCALL call_7_1_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_8_5
#define init_7_1_8_5
void FASTCALL call_7_1_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_8_6
#define init_7_1_8_6
void FASTCALL call_7_1_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_8_7
#define init_7_1_8_7
void FASTCALL call_7_1_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_8_8
#define init_7_1_8_8
void FASTCALL call_7_1_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_8_9
#define init_7_1_8_9
void FASTCALL call_7_1_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_8_10
#define init_7_1_8_10
void FASTCALL call_7_1_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_8_11
#define init_7_1_8_11
void FASTCALL call_7_1_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_8_12
#define init_7_1_8_12
void FASTCALL call_7_1_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_8_13
#define init_7_1_8_13
void FASTCALL call_7_1_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_8_14
#define init_7_1_8_14
void FASTCALL call_7_1_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_8_15
#define init_7_1_8_15
void FASTCALL call_7_1_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_9_0
#define init_7_1_9_0
void FASTCALL call_7_1_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_9_1
#define init_7_1_9_1
void FASTCALL call_7_1_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_9_2
#define init_7_1_9_2
void FASTCALL call_7_1_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_9_3
#define init_7_1_9_3
void FASTCALL call_7_1_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_9_4
#define init_7_1_9_4
void FASTCALL call_7_1_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_9_5
#define init_7_1_9_5
void FASTCALL call_7_1_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_9_6
#define init_7_1_9_6
void FASTCALL call_7_1_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_9_7
#define init_7_1_9_7
void FASTCALL call_7_1_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_9_8
#define init_7_1_9_8
void FASTCALL call_7_1_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_9_9
#define init_7_1_9_9
void FASTCALL call_7_1_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_9_10
#define init_7_1_9_10
void FASTCALL call_7_1_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_9_11
#define init_7_1_9_11
void FASTCALL call_7_1_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_9_12
#define init_7_1_9_12
void FASTCALL call_7_1_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_9_13
#define init_7_1_9_13
void FASTCALL call_7_1_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_9_14
#define init_7_1_9_14
void FASTCALL call_7_1_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_9_15
#define init_7_1_9_15
void FASTCALL call_7_1_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_10_0
#define init_7_1_10_0
void FASTCALL call_7_1_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_10_1
#define init_7_1_10_1
void FASTCALL call_7_1_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_10_2
#define init_7_1_10_2
void FASTCALL call_7_1_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_10_3
#define init_7_1_10_3
void FASTCALL call_7_1_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_10_4
#define init_7_1_10_4
void FASTCALL call_7_1_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_10_5
#define init_7_1_10_5
void FASTCALL call_7_1_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_10_6
#define init_7_1_10_6
void FASTCALL call_7_1_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_10_7
#define init_7_1_10_7
void FASTCALL call_7_1_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_10_8
#define init_7_1_10_8
void FASTCALL call_7_1_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_10_9
#define init_7_1_10_9
void FASTCALL call_7_1_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_10_10
#define init_7_1_10_10
void FASTCALL call_7_1_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_10_11
#define init_7_1_10_11
void FASTCALL call_7_1_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_10_12
#define init_7_1_10_12
void FASTCALL call_7_1_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_10_13
#define init_7_1_10_13
void FASTCALL call_7_1_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_10_14
#define init_7_1_10_14
void FASTCALL call_7_1_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_10_15
#define init_7_1_10_15
void FASTCALL call_7_1_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_11_0
#define init_7_1_11_0
void FASTCALL call_7_1_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_11_1
#define init_7_1_11_1
void FASTCALL call_7_1_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_11_2
#define init_7_1_11_2
void FASTCALL call_7_1_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_11_3
#define init_7_1_11_3
void FASTCALL call_7_1_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_11_4
#define init_7_1_11_4
void FASTCALL call_7_1_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_11_5
#define init_7_1_11_5
void FASTCALL call_7_1_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_11_6
#define init_7_1_11_6
void FASTCALL call_7_1_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_11_7
#define init_7_1_11_7
void FASTCALL call_7_1_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_11_8
#define init_7_1_11_8
void FASTCALL call_7_1_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_11_9
#define init_7_1_11_9
void FASTCALL call_7_1_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_11_10
#define init_7_1_11_10
void FASTCALL call_7_1_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_11_11
#define init_7_1_11_11
void FASTCALL call_7_1_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_11_12
#define init_7_1_11_12
void FASTCALL call_7_1_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_11_13
#define init_7_1_11_13
void FASTCALL call_7_1_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_11_14
#define init_7_1_11_14
void FASTCALL call_7_1_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_11_15
#define init_7_1_11_15
void FASTCALL call_7_1_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_12_0
#define init_7_1_12_0
void FASTCALL call_7_1_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_12_1
#define init_7_1_12_1
void FASTCALL call_7_1_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_12_2
#define init_7_1_12_2
void FASTCALL call_7_1_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_12_3
#define init_7_1_12_3
void FASTCALL call_7_1_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_12_4
#define init_7_1_12_4
void FASTCALL call_7_1_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_12_5
#define init_7_1_12_5
void FASTCALL call_7_1_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_12_6
#define init_7_1_12_6
void FASTCALL call_7_1_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_12_7
#define init_7_1_12_7
void FASTCALL call_7_1_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_12_8
#define init_7_1_12_8
void FASTCALL call_7_1_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_12_9
#define init_7_1_12_9
void FASTCALL call_7_1_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_12_10
#define init_7_1_12_10
void FASTCALL call_7_1_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_12_11
#define init_7_1_12_11
void FASTCALL call_7_1_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_12_12
#define init_7_1_12_12
void FASTCALL call_7_1_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_12_13
#define init_7_1_12_13
void FASTCALL call_7_1_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_12_14
#define init_7_1_12_14
void FASTCALL call_7_1_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_12_15
#define init_7_1_12_15
void FASTCALL call_7_1_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_13_0
#define init_7_1_13_0
void FASTCALL call_7_1_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_13_1
#define init_7_1_13_1
void FASTCALL call_7_1_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_13_2
#define init_7_1_13_2
void FASTCALL call_7_1_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_13_3
#define init_7_1_13_3
void FASTCALL call_7_1_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_13_4
#define init_7_1_13_4
void FASTCALL call_7_1_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_13_5
#define init_7_1_13_5
void FASTCALL call_7_1_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_13_6
#define init_7_1_13_6
void FASTCALL call_7_1_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_13_7
#define init_7_1_13_7
void FASTCALL call_7_1_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_13_8
#define init_7_1_13_8
void FASTCALL call_7_1_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_13_9
#define init_7_1_13_9
void FASTCALL call_7_1_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_13_10
#define init_7_1_13_10
void FASTCALL call_7_1_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_13_11
#define init_7_1_13_11
void FASTCALL call_7_1_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_13_12
#define init_7_1_13_12
void FASTCALL call_7_1_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_13_13
#define init_7_1_13_13
void FASTCALL call_7_1_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_13_14
#define init_7_1_13_14
void FASTCALL call_7_1_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_13_15
#define init_7_1_13_15
void FASTCALL call_7_1_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_14_0
#define init_7_1_14_0
void FASTCALL call_7_1_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_14_1
#define init_7_1_14_1
void FASTCALL call_7_1_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_14_2
#define init_7_1_14_2
void FASTCALL call_7_1_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_14_3
#define init_7_1_14_3
void FASTCALL call_7_1_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_14_4
#define init_7_1_14_4
void FASTCALL call_7_1_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_14_5
#define init_7_1_14_5
void FASTCALL call_7_1_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_14_6
#define init_7_1_14_6
void FASTCALL call_7_1_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_14_7
#define init_7_1_14_7
void FASTCALL call_7_1_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_14_8
#define init_7_1_14_8
void FASTCALL call_7_1_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_14_9
#define init_7_1_14_9
void FASTCALL call_7_1_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_14_10
#define init_7_1_14_10
void FASTCALL call_7_1_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_14_11
#define init_7_1_14_11
void FASTCALL call_7_1_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_14_12
#define init_7_1_14_12
void FASTCALL call_7_1_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_14_13
#define init_7_1_14_13
void FASTCALL call_7_1_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_14_14
#define init_7_1_14_14
void FASTCALL call_7_1_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_14_15
#define init_7_1_14_15
void FASTCALL call_7_1_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_15_0
#define init_7_1_15_0
void FASTCALL call_7_1_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_15_1
#define init_7_1_15_1
void FASTCALL call_7_1_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_15_2
#define init_7_1_15_2
void FASTCALL call_7_1_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_15_3
#define init_7_1_15_3
void FASTCALL call_7_1_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_15_4
#define init_7_1_15_4
void FASTCALL call_7_1_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_15_5
#define init_7_1_15_5
void FASTCALL call_7_1_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_15_6
#define init_7_1_15_6
void FASTCALL call_7_1_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_15_7
#define init_7_1_15_7
void FASTCALL call_7_1_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_15_8
#define init_7_1_15_8
void FASTCALL call_7_1_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_15_9
#define init_7_1_15_9
void FASTCALL call_7_1_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_15_10
#define init_7_1_15_10
void FASTCALL call_7_1_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_15_11
#define init_7_1_15_11
void FASTCALL call_7_1_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_15_12
#define init_7_1_15_12
void FASTCALL call_7_1_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_15_13
#define init_7_1_15_13
void FASTCALL call_7_1_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_15_14
#define init_7_1_15_14
void FASTCALL call_7_1_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_1_15_15
#define init_7_1_15_15
void FASTCALL call_7_1_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_0_0
#define init_7_2_0_0
void FASTCALL call_7_2_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_0_1
#define init_7_2_0_1
void FASTCALL call_7_2_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_0_2
#define init_7_2_0_2
void FASTCALL call_7_2_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_0_3
#define init_7_2_0_3
void FASTCALL call_7_2_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_0_4
#define init_7_2_0_4
void FASTCALL call_7_2_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_0_5
#define init_7_2_0_5
void FASTCALL call_7_2_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_0_6
#define init_7_2_0_6
void FASTCALL call_7_2_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_0_7
#define init_7_2_0_7
void FASTCALL call_7_2_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_0_8
#define init_7_2_0_8
void FASTCALL call_7_2_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_0_9
#define init_7_2_0_9
void FASTCALL call_7_2_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_0_10
#define init_7_2_0_10
void FASTCALL call_7_2_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_0_11
#define init_7_2_0_11
void FASTCALL call_7_2_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_0_12
#define init_7_2_0_12
void FASTCALL call_7_2_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_0_13
#define init_7_2_0_13
void FASTCALL call_7_2_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_0_14
#define init_7_2_0_14
void FASTCALL call_7_2_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_0_15
#define init_7_2_0_15
void FASTCALL call_7_2_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_1_0
#define init_7_2_1_0
void FASTCALL call_7_2_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_1_1
#define init_7_2_1_1
void FASTCALL call_7_2_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_1_2
#define init_7_2_1_2
void FASTCALL call_7_2_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_1_3
#define init_7_2_1_3
void FASTCALL call_7_2_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_1_4
#define init_7_2_1_4
void FASTCALL call_7_2_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_1_5
#define init_7_2_1_5
void FASTCALL call_7_2_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_1_6
#define init_7_2_1_6
void FASTCALL call_7_2_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_1_7
#define init_7_2_1_7
void FASTCALL call_7_2_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_1_8
#define init_7_2_1_8
void FASTCALL call_7_2_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_1_9
#define init_7_2_1_9
void FASTCALL call_7_2_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_1_10
#define init_7_2_1_10
void FASTCALL call_7_2_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_1_11
#define init_7_2_1_11
void FASTCALL call_7_2_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_1_12
#define init_7_2_1_12
void FASTCALL call_7_2_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_1_13
#define init_7_2_1_13
void FASTCALL call_7_2_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_1_14
#define init_7_2_1_14
void FASTCALL call_7_2_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_1_15
#define init_7_2_1_15
void FASTCALL call_7_2_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_2_0
#define init_7_2_2_0
void FASTCALL call_7_2_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_2_1
#define init_7_2_2_1
void FASTCALL call_7_2_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_2_2
#define init_7_2_2_2
void FASTCALL call_7_2_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_2_3
#define init_7_2_2_3
void FASTCALL call_7_2_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_2_4
#define init_7_2_2_4
void FASTCALL call_7_2_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_2_5
#define init_7_2_2_5
void FASTCALL call_7_2_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_2_6
#define init_7_2_2_6
void FASTCALL call_7_2_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_2_7
#define init_7_2_2_7
void FASTCALL call_7_2_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_2_8
#define init_7_2_2_8
void FASTCALL call_7_2_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_2_9
#define init_7_2_2_9
void FASTCALL call_7_2_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_2_10
#define init_7_2_2_10
void FASTCALL call_7_2_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_2_11
#define init_7_2_2_11
void FASTCALL call_7_2_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_2_12
#define init_7_2_2_12
void FASTCALL call_7_2_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_2_13
#define init_7_2_2_13
void FASTCALL call_7_2_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_2_14
#define init_7_2_2_14
void FASTCALL call_7_2_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_2_15
#define init_7_2_2_15
void FASTCALL call_7_2_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_3_0
#define init_7_2_3_0
void FASTCALL call_7_2_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_3_1
#define init_7_2_3_1
void FASTCALL call_7_2_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_3_2
#define init_7_2_3_2
void FASTCALL call_7_2_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_3_3
#define init_7_2_3_3
void FASTCALL call_7_2_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_3_4
#define init_7_2_3_4
void FASTCALL call_7_2_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_3_5
#define init_7_2_3_5
void FASTCALL call_7_2_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_3_6
#define init_7_2_3_6
void FASTCALL call_7_2_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_3_7
#define init_7_2_3_7
void FASTCALL call_7_2_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_3_8
#define init_7_2_3_8
void FASTCALL call_7_2_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_3_9
#define init_7_2_3_9
void FASTCALL call_7_2_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_3_10
#define init_7_2_3_10
void FASTCALL call_7_2_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_3_11
#define init_7_2_3_11
void FASTCALL call_7_2_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_3_12
#define init_7_2_3_12
void FASTCALL call_7_2_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_3_13
#define init_7_2_3_13
void FASTCALL call_7_2_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_3_14
#define init_7_2_3_14
void FASTCALL call_7_2_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_3_15
#define init_7_2_3_15
void FASTCALL call_7_2_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_4_0
#define init_7_2_4_0
void FASTCALL call_7_2_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_4_1
#define init_7_2_4_1
void FASTCALL call_7_2_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_4_2
#define init_7_2_4_2
void FASTCALL call_7_2_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_4_3
#define init_7_2_4_3
void FASTCALL call_7_2_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_4_4
#define init_7_2_4_4
void FASTCALL call_7_2_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_4_5
#define init_7_2_4_5
void FASTCALL call_7_2_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_4_6
#define init_7_2_4_6
void FASTCALL call_7_2_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_4_7
#define init_7_2_4_7
void FASTCALL call_7_2_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_4_8
#define init_7_2_4_8
void FASTCALL call_7_2_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_4_9
#define init_7_2_4_9
void FASTCALL call_7_2_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_4_10
#define init_7_2_4_10
void FASTCALL call_7_2_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_4_11
#define init_7_2_4_11
void FASTCALL call_7_2_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_4_12
#define init_7_2_4_12
void FASTCALL call_7_2_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_4_13
#define init_7_2_4_13
void FASTCALL call_7_2_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_4_14
#define init_7_2_4_14
void FASTCALL call_7_2_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_4_15
#define init_7_2_4_15
void FASTCALL call_7_2_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_5_0
#define init_7_2_5_0
void FASTCALL call_7_2_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_5_1
#define init_7_2_5_1
void FASTCALL call_7_2_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_5_2
#define init_7_2_5_2
void FASTCALL call_7_2_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_5_3
#define init_7_2_5_3
void FASTCALL call_7_2_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_5_4
#define init_7_2_5_4
void FASTCALL call_7_2_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_5_5
#define init_7_2_5_5
void FASTCALL call_7_2_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_5_6
#define init_7_2_5_6
void FASTCALL call_7_2_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_5_7
#define init_7_2_5_7
void FASTCALL call_7_2_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_5_8
#define init_7_2_5_8
void FASTCALL call_7_2_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_5_9
#define init_7_2_5_9
void FASTCALL call_7_2_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_5_10
#define init_7_2_5_10
void FASTCALL call_7_2_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_5_11
#define init_7_2_5_11
void FASTCALL call_7_2_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_5_12
#define init_7_2_5_12
void FASTCALL call_7_2_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_5_13
#define init_7_2_5_13
void FASTCALL call_7_2_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_5_14
#define init_7_2_5_14
void FASTCALL call_7_2_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_5_15
#define init_7_2_5_15
void FASTCALL call_7_2_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_6_0
#define init_7_2_6_0
void FASTCALL call_7_2_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_6_1
#define init_7_2_6_1
void FASTCALL call_7_2_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_6_2
#define init_7_2_6_2
void FASTCALL call_7_2_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_6_3
#define init_7_2_6_3
void FASTCALL call_7_2_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_6_4
#define init_7_2_6_4
void FASTCALL call_7_2_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_6_5
#define init_7_2_6_5
void FASTCALL call_7_2_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_6_6
#define init_7_2_6_6
void FASTCALL call_7_2_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_6_7
#define init_7_2_6_7
void FASTCALL call_7_2_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_6_8
#define init_7_2_6_8
void FASTCALL call_7_2_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_6_9
#define init_7_2_6_9
void FASTCALL call_7_2_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_6_10
#define init_7_2_6_10
void FASTCALL call_7_2_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_6_11
#define init_7_2_6_11
void FASTCALL call_7_2_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_6_12
#define init_7_2_6_12
void FASTCALL call_7_2_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_6_13
#define init_7_2_6_13
void FASTCALL call_7_2_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_6_14
#define init_7_2_6_14
void FASTCALL call_7_2_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_6_15
#define init_7_2_6_15
void FASTCALL call_7_2_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_7_0
#define init_7_2_7_0
void FASTCALL call_7_2_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_7_1
#define init_7_2_7_1
void FASTCALL call_7_2_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_7_2
#define init_7_2_7_2
void FASTCALL call_7_2_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_7_3
#define init_7_2_7_3
void FASTCALL call_7_2_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_7_4
#define init_7_2_7_4
void FASTCALL call_7_2_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_7_5
#define init_7_2_7_5
void FASTCALL call_7_2_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_7_6
#define init_7_2_7_6
void FASTCALL call_7_2_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_7_7
#define init_7_2_7_7
void FASTCALL call_7_2_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_7_8
#define init_7_2_7_8
void FASTCALL call_7_2_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_7_9
#define init_7_2_7_9
void FASTCALL call_7_2_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_7_10
#define init_7_2_7_10
void FASTCALL call_7_2_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_7_11
#define init_7_2_7_11
void FASTCALL call_7_2_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_7_12
#define init_7_2_7_12
void FASTCALL call_7_2_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_7_13
#define init_7_2_7_13
void FASTCALL call_7_2_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_7_14
#define init_7_2_7_14
void FASTCALL call_7_2_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_7_15
#define init_7_2_7_15
void FASTCALL call_7_2_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_8_0
#define init_7_2_8_0
void FASTCALL call_7_2_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_8_1
#define init_7_2_8_1
void FASTCALL call_7_2_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_8_2
#define init_7_2_8_2
void FASTCALL call_7_2_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_8_3
#define init_7_2_8_3
void FASTCALL call_7_2_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_8_4
#define init_7_2_8_4
void FASTCALL call_7_2_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_8_5
#define init_7_2_8_5
void FASTCALL call_7_2_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_8_6
#define init_7_2_8_6
void FASTCALL call_7_2_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_8_7
#define init_7_2_8_7
void FASTCALL call_7_2_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_8_8
#define init_7_2_8_8
void FASTCALL call_7_2_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_8_9
#define init_7_2_8_9
void FASTCALL call_7_2_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_8_10
#define init_7_2_8_10
void FASTCALL call_7_2_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_8_11
#define init_7_2_8_11
void FASTCALL call_7_2_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_8_12
#define init_7_2_8_12
void FASTCALL call_7_2_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_8_13
#define init_7_2_8_13
void FASTCALL call_7_2_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_8_14
#define init_7_2_8_14
void FASTCALL call_7_2_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_8_15
#define init_7_2_8_15
void FASTCALL call_7_2_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_9_0
#define init_7_2_9_0
void FASTCALL call_7_2_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_9_1
#define init_7_2_9_1
void FASTCALL call_7_2_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_9_2
#define init_7_2_9_2
void FASTCALL call_7_2_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_9_3
#define init_7_2_9_3
void FASTCALL call_7_2_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_9_4
#define init_7_2_9_4
void FASTCALL call_7_2_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_9_5
#define init_7_2_9_5
void FASTCALL call_7_2_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_9_6
#define init_7_2_9_6
void FASTCALL call_7_2_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_9_7
#define init_7_2_9_7
void FASTCALL call_7_2_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_9_8
#define init_7_2_9_8
void FASTCALL call_7_2_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_9_9
#define init_7_2_9_9
void FASTCALL call_7_2_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_9_10
#define init_7_2_9_10
void FASTCALL call_7_2_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_9_11
#define init_7_2_9_11
void FASTCALL call_7_2_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_9_12
#define init_7_2_9_12
void FASTCALL call_7_2_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_9_13
#define init_7_2_9_13
void FASTCALL call_7_2_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_9_14
#define init_7_2_9_14
void FASTCALL call_7_2_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_9_15
#define init_7_2_9_15
void FASTCALL call_7_2_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_10_0
#define init_7_2_10_0
void FASTCALL call_7_2_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_10_1
#define init_7_2_10_1
void FASTCALL call_7_2_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_10_2
#define init_7_2_10_2
void FASTCALL call_7_2_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_10_3
#define init_7_2_10_3
void FASTCALL call_7_2_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_10_4
#define init_7_2_10_4
void FASTCALL call_7_2_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_10_5
#define init_7_2_10_5
void FASTCALL call_7_2_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_10_6
#define init_7_2_10_6
void FASTCALL call_7_2_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_10_7
#define init_7_2_10_7
void FASTCALL call_7_2_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_10_8
#define init_7_2_10_8
void FASTCALL call_7_2_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_10_9
#define init_7_2_10_9
void FASTCALL call_7_2_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_10_10
#define init_7_2_10_10
void FASTCALL call_7_2_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_10_11
#define init_7_2_10_11
void FASTCALL call_7_2_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_10_12
#define init_7_2_10_12
void FASTCALL call_7_2_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_10_13
#define init_7_2_10_13
void FASTCALL call_7_2_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_10_14
#define init_7_2_10_14
void FASTCALL call_7_2_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_10_15
#define init_7_2_10_15
void FASTCALL call_7_2_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_11_0
#define init_7_2_11_0
void FASTCALL call_7_2_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_11_1
#define init_7_2_11_1
void FASTCALL call_7_2_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_11_2
#define init_7_2_11_2
void FASTCALL call_7_2_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_11_3
#define init_7_2_11_3
void FASTCALL call_7_2_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_11_4
#define init_7_2_11_4
void FASTCALL call_7_2_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_11_5
#define init_7_2_11_5
void FASTCALL call_7_2_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_11_6
#define init_7_2_11_6
void FASTCALL call_7_2_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_11_7
#define init_7_2_11_7
void FASTCALL call_7_2_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_11_8
#define init_7_2_11_8
void FASTCALL call_7_2_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_11_9
#define init_7_2_11_9
void FASTCALL call_7_2_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_11_10
#define init_7_2_11_10
void FASTCALL call_7_2_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_11_11
#define init_7_2_11_11
void FASTCALL call_7_2_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_11_12
#define init_7_2_11_12
void FASTCALL call_7_2_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_11_13
#define init_7_2_11_13
void FASTCALL call_7_2_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_11_14
#define init_7_2_11_14
void FASTCALL call_7_2_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_11_15
#define init_7_2_11_15
void FASTCALL call_7_2_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_12_0
#define init_7_2_12_0
void FASTCALL call_7_2_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_12_1
#define init_7_2_12_1
void FASTCALL call_7_2_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_12_2
#define init_7_2_12_2
void FASTCALL call_7_2_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_12_3
#define init_7_2_12_3
void FASTCALL call_7_2_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_12_4
#define init_7_2_12_4
void FASTCALL call_7_2_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_12_5
#define init_7_2_12_5
void FASTCALL call_7_2_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_12_6
#define init_7_2_12_6
void FASTCALL call_7_2_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_12_7
#define init_7_2_12_7
void FASTCALL call_7_2_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_12_8
#define init_7_2_12_8
void FASTCALL call_7_2_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_12_9
#define init_7_2_12_9
void FASTCALL call_7_2_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_12_10
#define init_7_2_12_10
void FASTCALL call_7_2_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_12_11
#define init_7_2_12_11
void FASTCALL call_7_2_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_12_12
#define init_7_2_12_12
void FASTCALL call_7_2_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_12_13
#define init_7_2_12_13
void FASTCALL call_7_2_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_12_14
#define init_7_2_12_14
void FASTCALL call_7_2_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_12_15
#define init_7_2_12_15
void FASTCALL call_7_2_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_13_0
#define init_7_2_13_0
void FASTCALL call_7_2_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_13_1
#define init_7_2_13_1
void FASTCALL call_7_2_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_13_2
#define init_7_2_13_2
void FASTCALL call_7_2_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_13_3
#define init_7_2_13_3
void FASTCALL call_7_2_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_13_4
#define init_7_2_13_4
void FASTCALL call_7_2_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_13_5
#define init_7_2_13_5
void FASTCALL call_7_2_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_13_6
#define init_7_2_13_6
void FASTCALL call_7_2_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_13_7
#define init_7_2_13_7
void FASTCALL call_7_2_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_13_8
#define init_7_2_13_8
void FASTCALL call_7_2_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_13_9
#define init_7_2_13_9
void FASTCALL call_7_2_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_13_10
#define init_7_2_13_10
void FASTCALL call_7_2_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_13_11
#define init_7_2_13_11
void FASTCALL call_7_2_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_13_12
#define init_7_2_13_12
void FASTCALL call_7_2_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_13_13
#define init_7_2_13_13
void FASTCALL call_7_2_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_13_14
#define init_7_2_13_14
void FASTCALL call_7_2_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_13_15
#define init_7_2_13_15
void FASTCALL call_7_2_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_14_0
#define init_7_2_14_0
void FASTCALL call_7_2_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_14_1
#define init_7_2_14_1
void FASTCALL call_7_2_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_14_2
#define init_7_2_14_2
void FASTCALL call_7_2_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_14_3
#define init_7_2_14_3
void FASTCALL call_7_2_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_14_4
#define init_7_2_14_4
void FASTCALL call_7_2_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_14_5
#define init_7_2_14_5
void FASTCALL call_7_2_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_14_6
#define init_7_2_14_6
void FASTCALL call_7_2_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_14_7
#define init_7_2_14_7
void FASTCALL call_7_2_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_14_8
#define init_7_2_14_8
void FASTCALL call_7_2_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_14_9
#define init_7_2_14_9
void FASTCALL call_7_2_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_14_10
#define init_7_2_14_10
void FASTCALL call_7_2_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_14_11
#define init_7_2_14_11
void FASTCALL call_7_2_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_14_12
#define init_7_2_14_12
void FASTCALL call_7_2_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_14_13
#define init_7_2_14_13
void FASTCALL call_7_2_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_14_14
#define init_7_2_14_14
void FASTCALL call_7_2_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_14_15
#define init_7_2_14_15
void FASTCALL call_7_2_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_15_0
#define init_7_2_15_0
void FASTCALL call_7_2_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_15_1
#define init_7_2_15_1
void FASTCALL call_7_2_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_15_2
#define init_7_2_15_2
void FASTCALL call_7_2_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_15_3
#define init_7_2_15_3
void FASTCALL call_7_2_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_15_4
#define init_7_2_15_4
void FASTCALL call_7_2_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_15_5
#define init_7_2_15_5
void FASTCALL call_7_2_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_15_6
#define init_7_2_15_6
void FASTCALL call_7_2_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_15_7
#define init_7_2_15_7
void FASTCALL call_7_2_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_15_8
#define init_7_2_15_8
void FASTCALL call_7_2_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_15_9
#define init_7_2_15_9
void FASTCALL call_7_2_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_15_10
#define init_7_2_15_10
void FASTCALL call_7_2_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_15_11
#define init_7_2_15_11
void FASTCALL call_7_2_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_15_12
#define init_7_2_15_12
void FASTCALL call_7_2_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_15_13
#define init_7_2_15_13
void FASTCALL call_7_2_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_15_14
#define init_7_2_15_14
void FASTCALL call_7_2_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_2_15_15
#define init_7_2_15_15
void FASTCALL call_7_2_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_0_0
#define init_7_3_0_0
void FASTCALL call_7_3_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_0_1
#define init_7_3_0_1
void FASTCALL call_7_3_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_0_2
#define init_7_3_0_2
void FASTCALL call_7_3_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_0_3
#define init_7_3_0_3
void FASTCALL call_7_3_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_0_4
#define init_7_3_0_4
void FASTCALL call_7_3_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_0_5
#define init_7_3_0_5
void FASTCALL call_7_3_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_0_6
#define init_7_3_0_6
void FASTCALL call_7_3_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_0_7
#define init_7_3_0_7
void FASTCALL call_7_3_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_0_8
#define init_7_3_0_8
void FASTCALL call_7_3_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_0_9
#define init_7_3_0_9
void FASTCALL call_7_3_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_0_10
#define init_7_3_0_10
void FASTCALL call_7_3_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_0_11
#define init_7_3_0_11
void FASTCALL call_7_3_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_0_12
#define init_7_3_0_12
void FASTCALL call_7_3_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_0_13
#define init_7_3_0_13
void FASTCALL call_7_3_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_0_14
#define init_7_3_0_14
void FASTCALL call_7_3_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_0_15
#define init_7_3_0_15
void FASTCALL call_7_3_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_1_0
#define init_7_3_1_0
void FASTCALL call_7_3_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_1_1
#define init_7_3_1_1
void FASTCALL call_7_3_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_1_2
#define init_7_3_1_2
void FASTCALL call_7_3_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_1_3
#define init_7_3_1_3
void FASTCALL call_7_3_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_1_4
#define init_7_3_1_4
void FASTCALL call_7_3_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_1_5
#define init_7_3_1_5
void FASTCALL call_7_3_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_1_6
#define init_7_3_1_6
void FASTCALL call_7_3_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_1_7
#define init_7_3_1_7
void FASTCALL call_7_3_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_1_8
#define init_7_3_1_8
void FASTCALL call_7_3_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_1_9
#define init_7_3_1_9
void FASTCALL call_7_3_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_1_10
#define init_7_3_1_10
void FASTCALL call_7_3_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_1_11
#define init_7_3_1_11
void FASTCALL call_7_3_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_1_12
#define init_7_3_1_12
void FASTCALL call_7_3_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_1_13
#define init_7_3_1_13
void FASTCALL call_7_3_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_1_14
#define init_7_3_1_14
void FASTCALL call_7_3_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_1_15
#define init_7_3_1_15
void FASTCALL call_7_3_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_2_0
#define init_7_3_2_0
void FASTCALL call_7_3_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_2_1
#define init_7_3_2_1
void FASTCALL call_7_3_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_2_2
#define init_7_3_2_2
void FASTCALL call_7_3_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_2_3
#define init_7_3_2_3
void FASTCALL call_7_3_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_2_4
#define init_7_3_2_4
void FASTCALL call_7_3_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_2_5
#define init_7_3_2_5
void FASTCALL call_7_3_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_2_6
#define init_7_3_2_6
void FASTCALL call_7_3_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_2_7
#define init_7_3_2_7
void FASTCALL call_7_3_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_2_8
#define init_7_3_2_8
void FASTCALL call_7_3_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_2_9
#define init_7_3_2_9
void FASTCALL call_7_3_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_2_10
#define init_7_3_2_10
void FASTCALL call_7_3_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_2_11
#define init_7_3_2_11
void FASTCALL call_7_3_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_2_12
#define init_7_3_2_12
void FASTCALL call_7_3_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_2_13
#define init_7_3_2_13
void FASTCALL call_7_3_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_2_14
#define init_7_3_2_14
void FASTCALL call_7_3_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_2_15
#define init_7_3_2_15
void FASTCALL call_7_3_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_3_0
#define init_7_3_3_0
void FASTCALL call_7_3_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_3_1
#define init_7_3_3_1
void FASTCALL call_7_3_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_3_2
#define init_7_3_3_2
void FASTCALL call_7_3_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_3_3
#define init_7_3_3_3
void FASTCALL call_7_3_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_3_4
#define init_7_3_3_4
void FASTCALL call_7_3_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_3_5
#define init_7_3_3_5
void FASTCALL call_7_3_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_3_6
#define init_7_3_3_6
void FASTCALL call_7_3_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_3_7
#define init_7_3_3_7
void FASTCALL call_7_3_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_3_8
#define init_7_3_3_8
void FASTCALL call_7_3_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_3_9
#define init_7_3_3_9
void FASTCALL call_7_3_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_3_10
#define init_7_3_3_10
void FASTCALL call_7_3_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_3_11
#define init_7_3_3_11
void FASTCALL call_7_3_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_3_12
#define init_7_3_3_12
void FASTCALL call_7_3_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_3_13
#define init_7_3_3_13
void FASTCALL call_7_3_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_3_14
#define init_7_3_3_14
void FASTCALL call_7_3_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_3_15
#define init_7_3_3_15
void FASTCALL call_7_3_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_4_0
#define init_7_3_4_0
void FASTCALL call_7_3_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_4_1
#define init_7_3_4_1
void FASTCALL call_7_3_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_4_2
#define init_7_3_4_2
void FASTCALL call_7_3_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_4_3
#define init_7_3_4_3
void FASTCALL call_7_3_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_4_4
#define init_7_3_4_4
void FASTCALL call_7_3_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_4_5
#define init_7_3_4_5
void FASTCALL call_7_3_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_4_6
#define init_7_3_4_6
void FASTCALL call_7_3_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_4_7
#define init_7_3_4_7
void FASTCALL call_7_3_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_4_8
#define init_7_3_4_8
void FASTCALL call_7_3_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_4_9
#define init_7_3_4_9
void FASTCALL call_7_3_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_4_10
#define init_7_3_4_10
void FASTCALL call_7_3_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_4_11
#define init_7_3_4_11
void FASTCALL call_7_3_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_4_12
#define init_7_3_4_12
void FASTCALL call_7_3_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_4_13
#define init_7_3_4_13
void FASTCALL call_7_3_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_4_14
#define init_7_3_4_14
void FASTCALL call_7_3_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_4_15
#define init_7_3_4_15
void FASTCALL call_7_3_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_5_0
#define init_7_3_5_0
void FASTCALL call_7_3_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_5_1
#define init_7_3_5_1
void FASTCALL call_7_3_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_5_2
#define init_7_3_5_2
void FASTCALL call_7_3_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_5_3
#define init_7_3_5_3
void FASTCALL call_7_3_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_5_4
#define init_7_3_5_4
void FASTCALL call_7_3_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_5_5
#define init_7_3_5_5
void FASTCALL call_7_3_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_5_6
#define init_7_3_5_6
void FASTCALL call_7_3_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_5_7
#define init_7_3_5_7
void FASTCALL call_7_3_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_5_8
#define init_7_3_5_8
void FASTCALL call_7_3_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_5_9
#define init_7_3_5_9
void FASTCALL call_7_3_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_5_10
#define init_7_3_5_10
void FASTCALL call_7_3_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_5_11
#define init_7_3_5_11
void FASTCALL call_7_3_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_5_12
#define init_7_3_5_12
void FASTCALL call_7_3_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_5_13
#define init_7_3_5_13
void FASTCALL call_7_3_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_5_14
#define init_7_3_5_14
void FASTCALL call_7_3_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_5_15
#define init_7_3_5_15
void FASTCALL call_7_3_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_6_0
#define init_7_3_6_0
void FASTCALL call_7_3_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_6_1
#define init_7_3_6_1
void FASTCALL call_7_3_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_6_2
#define init_7_3_6_2
void FASTCALL call_7_3_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_6_3
#define init_7_3_6_3
void FASTCALL call_7_3_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_6_4
#define init_7_3_6_4
void FASTCALL call_7_3_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_6_5
#define init_7_3_6_5
void FASTCALL call_7_3_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_6_6
#define init_7_3_6_6
void FASTCALL call_7_3_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_6_7
#define init_7_3_6_7
void FASTCALL call_7_3_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_6_8
#define init_7_3_6_8
void FASTCALL call_7_3_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_6_9
#define init_7_3_6_9
void FASTCALL call_7_3_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_6_10
#define init_7_3_6_10
void FASTCALL call_7_3_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_6_11
#define init_7_3_6_11
void FASTCALL call_7_3_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_6_12
#define init_7_3_6_12
void FASTCALL call_7_3_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_6_13
#define init_7_3_6_13
void FASTCALL call_7_3_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_6_14
#define init_7_3_6_14
void FASTCALL call_7_3_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_6_15
#define init_7_3_6_15
void FASTCALL call_7_3_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_7_0
#define init_7_3_7_0
void FASTCALL call_7_3_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_7_1
#define init_7_3_7_1
void FASTCALL call_7_3_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_7_2
#define init_7_3_7_2
void FASTCALL call_7_3_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_7_3
#define init_7_3_7_3
void FASTCALL call_7_3_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_7_4
#define init_7_3_7_4
void FASTCALL call_7_3_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_7_5
#define init_7_3_7_5
void FASTCALL call_7_3_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_7_6
#define init_7_3_7_6
void FASTCALL call_7_3_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_7_7
#define init_7_3_7_7
void FASTCALL call_7_3_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_7_8
#define init_7_3_7_8
void FASTCALL call_7_3_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_7_9
#define init_7_3_7_9
void FASTCALL call_7_3_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_7_10
#define init_7_3_7_10
void FASTCALL call_7_3_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_7_11
#define init_7_3_7_11
void FASTCALL call_7_3_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_7_12
#define init_7_3_7_12
void FASTCALL call_7_3_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_7_13
#define init_7_3_7_13
void FASTCALL call_7_3_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_7_14
#define init_7_3_7_14
void FASTCALL call_7_3_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_7_15
#define init_7_3_7_15
void FASTCALL call_7_3_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_8_0
#define init_7_3_8_0
void FASTCALL call_7_3_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_8_1
#define init_7_3_8_1
void FASTCALL call_7_3_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_8_2
#define init_7_3_8_2
void FASTCALL call_7_3_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_8_3
#define init_7_3_8_3
void FASTCALL call_7_3_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_8_4
#define init_7_3_8_4
void FASTCALL call_7_3_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_8_5
#define init_7_3_8_5
void FASTCALL call_7_3_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_8_6
#define init_7_3_8_6
void FASTCALL call_7_3_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_8_7
#define init_7_3_8_7
void FASTCALL call_7_3_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_8_8
#define init_7_3_8_8
void FASTCALL call_7_3_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_8_9
#define init_7_3_8_9
void FASTCALL call_7_3_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_8_10
#define init_7_3_8_10
void FASTCALL call_7_3_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_8_11
#define init_7_3_8_11
void FASTCALL call_7_3_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_8_12
#define init_7_3_8_12
void FASTCALL call_7_3_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_8_13
#define init_7_3_8_13
void FASTCALL call_7_3_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_8_14
#define init_7_3_8_14
void FASTCALL call_7_3_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_8_15
#define init_7_3_8_15
void FASTCALL call_7_3_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_9_0
#define init_7_3_9_0
void FASTCALL call_7_3_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_9_1
#define init_7_3_9_1
void FASTCALL call_7_3_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_9_2
#define init_7_3_9_2
void FASTCALL call_7_3_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_9_3
#define init_7_3_9_3
void FASTCALL call_7_3_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_9_4
#define init_7_3_9_4
void FASTCALL call_7_3_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_9_5
#define init_7_3_9_5
void FASTCALL call_7_3_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_9_6
#define init_7_3_9_6
void FASTCALL call_7_3_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_9_7
#define init_7_3_9_7
void FASTCALL call_7_3_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_9_8
#define init_7_3_9_8
void FASTCALL call_7_3_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_9_9
#define init_7_3_9_9
void FASTCALL call_7_3_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_9_10
#define init_7_3_9_10
void FASTCALL call_7_3_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_9_11
#define init_7_3_9_11
void FASTCALL call_7_3_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_9_12
#define init_7_3_9_12
void FASTCALL call_7_3_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_9_13
#define init_7_3_9_13
void FASTCALL call_7_3_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_9_14
#define init_7_3_9_14
void FASTCALL call_7_3_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_9_15
#define init_7_3_9_15
void FASTCALL call_7_3_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_10_0
#define init_7_3_10_0
void FASTCALL call_7_3_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_10_1
#define init_7_3_10_1
void FASTCALL call_7_3_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_10_2
#define init_7_3_10_2
void FASTCALL call_7_3_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_10_3
#define init_7_3_10_3
void FASTCALL call_7_3_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_10_4
#define init_7_3_10_4
void FASTCALL call_7_3_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_10_5
#define init_7_3_10_5
void FASTCALL call_7_3_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_10_6
#define init_7_3_10_6
void FASTCALL call_7_3_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_10_7
#define init_7_3_10_7
void FASTCALL call_7_3_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_10_8
#define init_7_3_10_8
void FASTCALL call_7_3_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_10_9
#define init_7_3_10_9
void FASTCALL call_7_3_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_10_10
#define init_7_3_10_10
void FASTCALL call_7_3_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_10_11
#define init_7_3_10_11
void FASTCALL call_7_3_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_10_12
#define init_7_3_10_12
void FASTCALL call_7_3_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_10_13
#define init_7_3_10_13
void FASTCALL call_7_3_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_10_14
#define init_7_3_10_14
void FASTCALL call_7_3_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_10_15
#define init_7_3_10_15
void FASTCALL call_7_3_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_11_0
#define init_7_3_11_0
void FASTCALL call_7_3_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_11_1
#define init_7_3_11_1
void FASTCALL call_7_3_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_11_2
#define init_7_3_11_2
void FASTCALL call_7_3_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_11_3
#define init_7_3_11_3
void FASTCALL call_7_3_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_11_4
#define init_7_3_11_4
void FASTCALL call_7_3_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_11_5
#define init_7_3_11_5
void FASTCALL call_7_3_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_11_6
#define init_7_3_11_6
void FASTCALL call_7_3_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_11_7
#define init_7_3_11_7
void FASTCALL call_7_3_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_11_8
#define init_7_3_11_8
void FASTCALL call_7_3_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_11_9
#define init_7_3_11_9
void FASTCALL call_7_3_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_11_10
#define init_7_3_11_10
void FASTCALL call_7_3_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_11_11
#define init_7_3_11_11
void FASTCALL call_7_3_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_11_12
#define init_7_3_11_12
void FASTCALL call_7_3_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_11_13
#define init_7_3_11_13
void FASTCALL call_7_3_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_11_14
#define init_7_3_11_14
void FASTCALL call_7_3_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_11_15
#define init_7_3_11_15
void FASTCALL call_7_3_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_12_0
#define init_7_3_12_0
void FASTCALL call_7_3_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_12_1
#define init_7_3_12_1
void FASTCALL call_7_3_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_12_2
#define init_7_3_12_2
void FASTCALL call_7_3_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_12_3
#define init_7_3_12_3
void FASTCALL call_7_3_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_12_4
#define init_7_3_12_4
void FASTCALL call_7_3_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_12_5
#define init_7_3_12_5
void FASTCALL call_7_3_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_12_6
#define init_7_3_12_6
void FASTCALL call_7_3_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_12_7
#define init_7_3_12_7
void FASTCALL call_7_3_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_12_8
#define init_7_3_12_8
void FASTCALL call_7_3_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_12_9
#define init_7_3_12_9
void FASTCALL call_7_3_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_12_10
#define init_7_3_12_10
void FASTCALL call_7_3_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_12_11
#define init_7_3_12_11
void FASTCALL call_7_3_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_12_12
#define init_7_3_12_12
void FASTCALL call_7_3_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_12_13
#define init_7_3_12_13
void FASTCALL call_7_3_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_12_14
#define init_7_3_12_14
void FASTCALL call_7_3_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_12_15
#define init_7_3_12_15
void FASTCALL call_7_3_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_13_0
#define init_7_3_13_0
void FASTCALL call_7_3_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_13_1
#define init_7_3_13_1
void FASTCALL call_7_3_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_13_2
#define init_7_3_13_2
void FASTCALL call_7_3_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_13_3
#define init_7_3_13_3
void FASTCALL call_7_3_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_13_4
#define init_7_3_13_4
void FASTCALL call_7_3_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_13_5
#define init_7_3_13_5
void FASTCALL call_7_3_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_13_6
#define init_7_3_13_6
void FASTCALL call_7_3_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_13_7
#define init_7_3_13_7
void FASTCALL call_7_3_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_13_8
#define init_7_3_13_8
void FASTCALL call_7_3_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_13_9
#define init_7_3_13_9
void FASTCALL call_7_3_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_13_10
#define init_7_3_13_10
void FASTCALL call_7_3_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_13_11
#define init_7_3_13_11
void FASTCALL call_7_3_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_13_12
#define init_7_3_13_12
void FASTCALL call_7_3_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_13_13
#define init_7_3_13_13
void FASTCALL call_7_3_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_13_14
#define init_7_3_13_14
void FASTCALL call_7_3_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_13_15
#define init_7_3_13_15
void FASTCALL call_7_3_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_14_0
#define init_7_3_14_0
void FASTCALL call_7_3_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_14_1
#define init_7_3_14_1
void FASTCALL call_7_3_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_14_2
#define init_7_3_14_2
void FASTCALL call_7_3_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_14_3
#define init_7_3_14_3
void FASTCALL call_7_3_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_14_4
#define init_7_3_14_4
void FASTCALL call_7_3_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_14_5
#define init_7_3_14_5
void FASTCALL call_7_3_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_14_6
#define init_7_3_14_6
void FASTCALL call_7_3_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_14_7
#define init_7_3_14_7
void FASTCALL call_7_3_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_14_8
#define init_7_3_14_8
void FASTCALL call_7_3_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_14_9
#define init_7_3_14_9
void FASTCALL call_7_3_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_14_10
#define init_7_3_14_10
void FASTCALL call_7_3_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_14_11
#define init_7_3_14_11
void FASTCALL call_7_3_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_14_12
#define init_7_3_14_12
void FASTCALL call_7_3_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_14_13
#define init_7_3_14_13
void FASTCALL call_7_3_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_14_14
#define init_7_3_14_14
void FASTCALL call_7_3_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_14_15
#define init_7_3_14_15
void FASTCALL call_7_3_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_15_0
#define init_7_3_15_0
void FASTCALL call_7_3_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_15_1
#define init_7_3_15_1
void FASTCALL call_7_3_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_15_2
#define init_7_3_15_2
void FASTCALL call_7_3_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_15_3
#define init_7_3_15_3
void FASTCALL call_7_3_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_15_4
#define init_7_3_15_4
void FASTCALL call_7_3_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_15_5
#define init_7_3_15_5
void FASTCALL call_7_3_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_15_6
#define init_7_3_15_6
void FASTCALL call_7_3_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_15_7
#define init_7_3_15_7
void FASTCALL call_7_3_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_15_8
#define init_7_3_15_8
void FASTCALL call_7_3_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_15_9
#define init_7_3_15_9
void FASTCALL call_7_3_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_15_10
#define init_7_3_15_10
void FASTCALL call_7_3_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_15_11
#define init_7_3_15_11
void FASTCALL call_7_3_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_15_12
#define init_7_3_15_12
void FASTCALL call_7_3_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_15_13
#define init_7_3_15_13
void FASTCALL call_7_3_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_15_14
#define init_7_3_15_14
void FASTCALL call_7_3_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_3_15_15
#define init_7_3_15_15
void FASTCALL call_7_3_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_0_0
#define init_7_4_0_0
void FASTCALL call_7_4_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_0_1
#define init_7_4_0_1
void FASTCALL call_7_4_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_0_2
#define init_7_4_0_2
void FASTCALL call_7_4_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_0_3
#define init_7_4_0_3
void FASTCALL call_7_4_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_0_4
#define init_7_4_0_4
void FASTCALL call_7_4_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_0_5
#define init_7_4_0_5
void FASTCALL call_7_4_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_0_6
#define init_7_4_0_6
void FASTCALL call_7_4_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_0_7
#define init_7_4_0_7
void FASTCALL call_7_4_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_0_8
#define init_7_4_0_8
void FASTCALL call_7_4_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_0_9
#define init_7_4_0_9
void FASTCALL call_7_4_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_0_10
#define init_7_4_0_10
void FASTCALL call_7_4_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_0_11
#define init_7_4_0_11
void FASTCALL call_7_4_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_0_12
#define init_7_4_0_12
void FASTCALL call_7_4_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_0_13
#define init_7_4_0_13
void FASTCALL call_7_4_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_0_14
#define init_7_4_0_14
void FASTCALL call_7_4_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_0_15
#define init_7_4_0_15
void FASTCALL call_7_4_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_1_0
#define init_7_4_1_0
void FASTCALL call_7_4_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_1_1
#define init_7_4_1_1
void FASTCALL call_7_4_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_1_2
#define init_7_4_1_2
void FASTCALL call_7_4_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_1_3
#define init_7_4_1_3
void FASTCALL call_7_4_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_1_4
#define init_7_4_1_4
void FASTCALL call_7_4_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_1_5
#define init_7_4_1_5
void FASTCALL call_7_4_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_1_6
#define init_7_4_1_6
void FASTCALL call_7_4_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_1_7
#define init_7_4_1_7
void FASTCALL call_7_4_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_1_8
#define init_7_4_1_8
void FASTCALL call_7_4_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_1_9
#define init_7_4_1_9
void FASTCALL call_7_4_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_1_10
#define init_7_4_1_10
void FASTCALL call_7_4_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_1_11
#define init_7_4_1_11
void FASTCALL call_7_4_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_1_12
#define init_7_4_1_12
void FASTCALL call_7_4_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_1_13
#define init_7_4_1_13
void FASTCALL call_7_4_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_1_14
#define init_7_4_1_14
void FASTCALL call_7_4_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_1_15
#define init_7_4_1_15
void FASTCALL call_7_4_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_2_0
#define init_7_4_2_0
void FASTCALL call_7_4_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_2_1
#define init_7_4_2_1
void FASTCALL call_7_4_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_2_2
#define init_7_4_2_2
void FASTCALL call_7_4_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_2_3
#define init_7_4_2_3
void FASTCALL call_7_4_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_2_4
#define init_7_4_2_4
void FASTCALL call_7_4_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_2_5
#define init_7_4_2_5
void FASTCALL call_7_4_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_2_6
#define init_7_4_2_6
void FASTCALL call_7_4_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_2_7
#define init_7_4_2_7
void FASTCALL call_7_4_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_2_8
#define init_7_4_2_8
void FASTCALL call_7_4_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_2_9
#define init_7_4_2_9
void FASTCALL call_7_4_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_2_10
#define init_7_4_2_10
void FASTCALL call_7_4_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_2_11
#define init_7_4_2_11
void FASTCALL call_7_4_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_2_12
#define init_7_4_2_12
void FASTCALL call_7_4_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_2_13
#define init_7_4_2_13
void FASTCALL call_7_4_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_2_14
#define init_7_4_2_14
void FASTCALL call_7_4_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_2_15
#define init_7_4_2_15
void FASTCALL call_7_4_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_3_0
#define init_7_4_3_0
void FASTCALL call_7_4_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_3_1
#define init_7_4_3_1
void FASTCALL call_7_4_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_3_2
#define init_7_4_3_2
void FASTCALL call_7_4_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_3_3
#define init_7_4_3_3
void FASTCALL call_7_4_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_3_4
#define init_7_4_3_4
void FASTCALL call_7_4_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_3_5
#define init_7_4_3_5
void FASTCALL call_7_4_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_3_6
#define init_7_4_3_6
void FASTCALL call_7_4_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_3_7
#define init_7_4_3_7
void FASTCALL call_7_4_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_3_8
#define init_7_4_3_8
void FASTCALL call_7_4_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_3_9
#define init_7_4_3_9
void FASTCALL call_7_4_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_3_10
#define init_7_4_3_10
void FASTCALL call_7_4_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_3_11
#define init_7_4_3_11
void FASTCALL call_7_4_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_3_12
#define init_7_4_3_12
void FASTCALL call_7_4_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_3_13
#define init_7_4_3_13
void FASTCALL call_7_4_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_3_14
#define init_7_4_3_14
void FASTCALL call_7_4_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_3_15
#define init_7_4_3_15
void FASTCALL call_7_4_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_4_0
#define init_7_4_4_0
void FASTCALL call_7_4_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_4_1
#define init_7_4_4_1
void FASTCALL call_7_4_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_4_2
#define init_7_4_4_2
void FASTCALL call_7_4_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_4_3
#define init_7_4_4_3
void FASTCALL call_7_4_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_4_4
#define init_7_4_4_4
void FASTCALL call_7_4_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_4_5
#define init_7_4_4_5
void FASTCALL call_7_4_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_4_6
#define init_7_4_4_6
void FASTCALL call_7_4_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_4_7
#define init_7_4_4_7
void FASTCALL call_7_4_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_4_8
#define init_7_4_4_8
void FASTCALL call_7_4_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_4_9
#define init_7_4_4_9
void FASTCALL call_7_4_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_4_10
#define init_7_4_4_10
void FASTCALL call_7_4_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_4_11
#define init_7_4_4_11
void FASTCALL call_7_4_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_4_12
#define init_7_4_4_12
void FASTCALL call_7_4_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_4_13
#define init_7_4_4_13
void FASTCALL call_7_4_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_4_14
#define init_7_4_4_14
void FASTCALL call_7_4_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_4_15
#define init_7_4_4_15
void FASTCALL call_7_4_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_5_0
#define init_7_4_5_0
void FASTCALL call_7_4_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_5_1
#define init_7_4_5_1
void FASTCALL call_7_4_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_5_2
#define init_7_4_5_2
void FASTCALL call_7_4_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_5_3
#define init_7_4_5_3
void FASTCALL call_7_4_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_5_4
#define init_7_4_5_4
void FASTCALL call_7_4_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_5_5
#define init_7_4_5_5
void FASTCALL call_7_4_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_5_6
#define init_7_4_5_6
void FASTCALL call_7_4_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_5_7
#define init_7_4_5_7
void FASTCALL call_7_4_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_5_8
#define init_7_4_5_8
void FASTCALL call_7_4_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_5_9
#define init_7_4_5_9
void FASTCALL call_7_4_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_5_10
#define init_7_4_5_10
void FASTCALL call_7_4_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_5_11
#define init_7_4_5_11
void FASTCALL call_7_4_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_5_12
#define init_7_4_5_12
void FASTCALL call_7_4_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_5_13
#define init_7_4_5_13
void FASTCALL call_7_4_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_5_14
#define init_7_4_5_14
void FASTCALL call_7_4_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_5_15
#define init_7_4_5_15
void FASTCALL call_7_4_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_6_0
#define init_7_4_6_0
void FASTCALL call_7_4_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_6_1
#define init_7_4_6_1
void FASTCALL call_7_4_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_6_2
#define init_7_4_6_2
void FASTCALL call_7_4_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_6_3
#define init_7_4_6_3
void FASTCALL call_7_4_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_6_4
#define init_7_4_6_4
void FASTCALL call_7_4_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_6_5
#define init_7_4_6_5
void FASTCALL call_7_4_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_6_6
#define init_7_4_6_6
void FASTCALL call_7_4_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_6_7
#define init_7_4_6_7
void FASTCALL call_7_4_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_6_8
#define init_7_4_6_8
void FASTCALL call_7_4_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_6_9
#define init_7_4_6_9
void FASTCALL call_7_4_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_6_10
#define init_7_4_6_10
void FASTCALL call_7_4_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_6_11
#define init_7_4_6_11
void FASTCALL call_7_4_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_6_12
#define init_7_4_6_12
void FASTCALL call_7_4_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_6_13
#define init_7_4_6_13
void FASTCALL call_7_4_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_6_14
#define init_7_4_6_14
void FASTCALL call_7_4_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_6_15
#define init_7_4_6_15
void FASTCALL call_7_4_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_7_0
#define init_7_4_7_0
void FASTCALL call_7_4_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_7_1
#define init_7_4_7_1
void FASTCALL call_7_4_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_7_2
#define init_7_4_7_2
void FASTCALL call_7_4_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_7_3
#define init_7_4_7_3
void FASTCALL call_7_4_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_7_4
#define init_7_4_7_4
void FASTCALL call_7_4_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_7_5
#define init_7_4_7_5
void FASTCALL call_7_4_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_7_6
#define init_7_4_7_6
void FASTCALL call_7_4_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_7_7
#define init_7_4_7_7
void FASTCALL call_7_4_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_7_8
#define init_7_4_7_8
void FASTCALL call_7_4_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_7_9
#define init_7_4_7_9
void FASTCALL call_7_4_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_7_10
#define init_7_4_7_10
void FASTCALL call_7_4_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_7_11
#define init_7_4_7_11
void FASTCALL call_7_4_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_7_12
#define init_7_4_7_12
void FASTCALL call_7_4_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_7_13
#define init_7_4_7_13
void FASTCALL call_7_4_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_7_14
#define init_7_4_7_14
void FASTCALL call_7_4_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_7_15
#define init_7_4_7_15
void FASTCALL call_7_4_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_8_0
#define init_7_4_8_0
void FASTCALL call_7_4_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_8_1
#define init_7_4_8_1
void FASTCALL call_7_4_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_8_2
#define init_7_4_8_2
void FASTCALL call_7_4_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_8_3
#define init_7_4_8_3
void FASTCALL call_7_4_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_8_4
#define init_7_4_8_4
void FASTCALL call_7_4_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_8_5
#define init_7_4_8_5
void FASTCALL call_7_4_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_8_6
#define init_7_4_8_6
void FASTCALL call_7_4_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_8_7
#define init_7_4_8_7
void FASTCALL call_7_4_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_8_8
#define init_7_4_8_8
void FASTCALL call_7_4_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_8_9
#define init_7_4_8_9
void FASTCALL call_7_4_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_8_10
#define init_7_4_8_10
void FASTCALL call_7_4_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_8_11
#define init_7_4_8_11
void FASTCALL call_7_4_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_8_12
#define init_7_4_8_12
void FASTCALL call_7_4_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_8_13
#define init_7_4_8_13
void FASTCALL call_7_4_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_8_14
#define init_7_4_8_14
void FASTCALL call_7_4_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_8_15
#define init_7_4_8_15
void FASTCALL call_7_4_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_9_0
#define init_7_4_9_0
void FASTCALL call_7_4_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_9_1
#define init_7_4_9_1
void FASTCALL call_7_4_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_9_2
#define init_7_4_9_2
void FASTCALL call_7_4_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_9_3
#define init_7_4_9_3
void FASTCALL call_7_4_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_9_4
#define init_7_4_9_4
void FASTCALL call_7_4_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_9_5
#define init_7_4_9_5
void FASTCALL call_7_4_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_9_6
#define init_7_4_9_6
void FASTCALL call_7_4_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_9_7
#define init_7_4_9_7
void FASTCALL call_7_4_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_9_8
#define init_7_4_9_8
void FASTCALL call_7_4_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_9_9
#define init_7_4_9_9
void FASTCALL call_7_4_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_9_10
#define init_7_4_9_10
void FASTCALL call_7_4_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_9_11
#define init_7_4_9_11
void FASTCALL call_7_4_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_9_12
#define init_7_4_9_12
void FASTCALL call_7_4_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_9_13
#define init_7_4_9_13
void FASTCALL call_7_4_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_9_14
#define init_7_4_9_14
void FASTCALL call_7_4_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_9_15
#define init_7_4_9_15
void FASTCALL call_7_4_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_10_0
#define init_7_4_10_0
void FASTCALL call_7_4_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_10_1
#define init_7_4_10_1
void FASTCALL call_7_4_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_10_2
#define init_7_4_10_2
void FASTCALL call_7_4_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_10_3
#define init_7_4_10_3
void FASTCALL call_7_4_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_10_4
#define init_7_4_10_4
void FASTCALL call_7_4_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_10_5
#define init_7_4_10_5
void FASTCALL call_7_4_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_10_6
#define init_7_4_10_6
void FASTCALL call_7_4_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_10_7
#define init_7_4_10_7
void FASTCALL call_7_4_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_10_8
#define init_7_4_10_8
void FASTCALL call_7_4_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_10_9
#define init_7_4_10_9
void FASTCALL call_7_4_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_10_10
#define init_7_4_10_10
void FASTCALL call_7_4_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_10_11
#define init_7_4_10_11
void FASTCALL call_7_4_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_10_12
#define init_7_4_10_12
void FASTCALL call_7_4_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_10_13
#define init_7_4_10_13
void FASTCALL call_7_4_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_10_14
#define init_7_4_10_14
void FASTCALL call_7_4_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_10_15
#define init_7_4_10_15
void FASTCALL call_7_4_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_11_0
#define init_7_4_11_0
void FASTCALL call_7_4_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_11_1
#define init_7_4_11_1
void FASTCALL call_7_4_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_11_2
#define init_7_4_11_2
void FASTCALL call_7_4_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_11_3
#define init_7_4_11_3
void FASTCALL call_7_4_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_11_4
#define init_7_4_11_4
void FASTCALL call_7_4_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_11_5
#define init_7_4_11_5
void FASTCALL call_7_4_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_11_6
#define init_7_4_11_6
void FASTCALL call_7_4_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_11_7
#define init_7_4_11_7
void FASTCALL call_7_4_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_11_8
#define init_7_4_11_8
void FASTCALL call_7_4_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_11_9
#define init_7_4_11_9
void FASTCALL call_7_4_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_11_10
#define init_7_4_11_10
void FASTCALL call_7_4_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_11_11
#define init_7_4_11_11
void FASTCALL call_7_4_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_11_12
#define init_7_4_11_12
void FASTCALL call_7_4_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_11_13
#define init_7_4_11_13
void FASTCALL call_7_4_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_11_14
#define init_7_4_11_14
void FASTCALL call_7_4_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_11_15
#define init_7_4_11_15
void FASTCALL call_7_4_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_12_0
#define init_7_4_12_0
void FASTCALL call_7_4_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_12_1
#define init_7_4_12_1
void FASTCALL call_7_4_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_12_2
#define init_7_4_12_2
void FASTCALL call_7_4_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_12_3
#define init_7_4_12_3
void FASTCALL call_7_4_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_12_4
#define init_7_4_12_4
void FASTCALL call_7_4_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_12_5
#define init_7_4_12_5
void FASTCALL call_7_4_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_12_6
#define init_7_4_12_6
void FASTCALL call_7_4_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_12_7
#define init_7_4_12_7
void FASTCALL call_7_4_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_12_8
#define init_7_4_12_8
void FASTCALL call_7_4_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_12_9
#define init_7_4_12_9
void FASTCALL call_7_4_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_12_10
#define init_7_4_12_10
void FASTCALL call_7_4_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_12_11
#define init_7_4_12_11
void FASTCALL call_7_4_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_12_12
#define init_7_4_12_12
void FASTCALL call_7_4_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_12_13
#define init_7_4_12_13
void FASTCALL call_7_4_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_12_14
#define init_7_4_12_14
void FASTCALL call_7_4_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_12_15
#define init_7_4_12_15
void FASTCALL call_7_4_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_13_0
#define init_7_4_13_0
void FASTCALL call_7_4_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_13_1
#define init_7_4_13_1
void FASTCALL call_7_4_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_13_2
#define init_7_4_13_2
void FASTCALL call_7_4_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_13_3
#define init_7_4_13_3
void FASTCALL call_7_4_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_13_4
#define init_7_4_13_4
void FASTCALL call_7_4_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_13_5
#define init_7_4_13_5
void FASTCALL call_7_4_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_13_6
#define init_7_4_13_6
void FASTCALL call_7_4_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_13_7
#define init_7_4_13_7
void FASTCALL call_7_4_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_13_8
#define init_7_4_13_8
void FASTCALL call_7_4_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_13_9
#define init_7_4_13_9
void FASTCALL call_7_4_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_13_10
#define init_7_4_13_10
void FASTCALL call_7_4_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_13_11
#define init_7_4_13_11
void FASTCALL call_7_4_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_13_12
#define init_7_4_13_12
void FASTCALL call_7_4_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_13_13
#define init_7_4_13_13
void FASTCALL call_7_4_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_13_14
#define init_7_4_13_14
void FASTCALL call_7_4_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_13_15
#define init_7_4_13_15
void FASTCALL call_7_4_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_14_0
#define init_7_4_14_0
void FASTCALL call_7_4_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_14_1
#define init_7_4_14_1
void FASTCALL call_7_4_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_14_2
#define init_7_4_14_2
void FASTCALL call_7_4_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_14_3
#define init_7_4_14_3
void FASTCALL call_7_4_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_14_4
#define init_7_4_14_4
void FASTCALL call_7_4_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_14_5
#define init_7_4_14_5
void FASTCALL call_7_4_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_14_6
#define init_7_4_14_6
void FASTCALL call_7_4_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_14_7
#define init_7_4_14_7
void FASTCALL call_7_4_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_14_8
#define init_7_4_14_8
void FASTCALL call_7_4_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_14_9
#define init_7_4_14_9
void FASTCALL call_7_4_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_14_10
#define init_7_4_14_10
void FASTCALL call_7_4_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_14_11
#define init_7_4_14_11
void FASTCALL call_7_4_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_14_12
#define init_7_4_14_12
void FASTCALL call_7_4_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_14_13
#define init_7_4_14_13
void FASTCALL call_7_4_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_14_14
#define init_7_4_14_14
void FASTCALL call_7_4_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_14_15
#define init_7_4_14_15
void FASTCALL call_7_4_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_15_0
#define init_7_4_15_0
void FASTCALL call_7_4_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_15_1
#define init_7_4_15_1
void FASTCALL call_7_4_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_15_2
#define init_7_4_15_2
void FASTCALL call_7_4_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_15_3
#define init_7_4_15_3
void FASTCALL call_7_4_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_15_4
#define init_7_4_15_4
void FASTCALL call_7_4_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_15_5
#define init_7_4_15_5
void FASTCALL call_7_4_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_15_6
#define init_7_4_15_6
void FASTCALL call_7_4_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_15_7
#define init_7_4_15_7
void FASTCALL call_7_4_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_15_8
#define init_7_4_15_8
void FASTCALL call_7_4_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_15_9
#define init_7_4_15_9
void FASTCALL call_7_4_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_15_10
#define init_7_4_15_10
void FASTCALL call_7_4_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_15_11
#define init_7_4_15_11
void FASTCALL call_7_4_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_15_12
#define init_7_4_15_12
void FASTCALL call_7_4_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_15_13
#define init_7_4_15_13
void FASTCALL call_7_4_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_15_14
#define init_7_4_15_14
void FASTCALL call_7_4_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_4_15_15
#define init_7_4_15_15
void FASTCALL call_7_4_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_0_0
#define init_7_5_0_0
void FASTCALL call_7_5_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_0_1
#define init_7_5_0_1
void FASTCALL call_7_5_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_0_2
#define init_7_5_0_2
void FASTCALL call_7_5_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_0_3
#define init_7_5_0_3
void FASTCALL call_7_5_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_0_4
#define init_7_5_0_4
void FASTCALL call_7_5_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_0_5
#define init_7_5_0_5
void FASTCALL call_7_5_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_0_6
#define init_7_5_0_6
void FASTCALL call_7_5_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_0_7
#define init_7_5_0_7
void FASTCALL call_7_5_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_0_8
#define init_7_5_0_8
void FASTCALL call_7_5_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_0_9
#define init_7_5_0_9
void FASTCALL call_7_5_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_0_10
#define init_7_5_0_10
void FASTCALL call_7_5_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_0_11
#define init_7_5_0_11
void FASTCALL call_7_5_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_0_12
#define init_7_5_0_12
void FASTCALL call_7_5_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_0_13
#define init_7_5_0_13
void FASTCALL call_7_5_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_0_14
#define init_7_5_0_14
void FASTCALL call_7_5_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_0_15
#define init_7_5_0_15
void FASTCALL call_7_5_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_1_0
#define init_7_5_1_0
void FASTCALL call_7_5_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_1_1
#define init_7_5_1_1
void FASTCALL call_7_5_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_1_2
#define init_7_5_1_2
void FASTCALL call_7_5_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_1_3
#define init_7_5_1_3
void FASTCALL call_7_5_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_1_4
#define init_7_5_1_4
void FASTCALL call_7_5_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_1_5
#define init_7_5_1_5
void FASTCALL call_7_5_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_1_6
#define init_7_5_1_6
void FASTCALL call_7_5_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_1_7
#define init_7_5_1_7
void FASTCALL call_7_5_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_1_8
#define init_7_5_1_8
void FASTCALL call_7_5_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_1_9
#define init_7_5_1_9
void FASTCALL call_7_5_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_1_10
#define init_7_5_1_10
void FASTCALL call_7_5_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_1_11
#define init_7_5_1_11
void FASTCALL call_7_5_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_1_12
#define init_7_5_1_12
void FASTCALL call_7_5_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_1_13
#define init_7_5_1_13
void FASTCALL call_7_5_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_1_14
#define init_7_5_1_14
void FASTCALL call_7_5_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_1_15
#define init_7_5_1_15
void FASTCALL call_7_5_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_2_0
#define init_7_5_2_0
void FASTCALL call_7_5_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_2_1
#define init_7_5_2_1
void FASTCALL call_7_5_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_2_2
#define init_7_5_2_2
void FASTCALL call_7_5_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_2_3
#define init_7_5_2_3
void FASTCALL call_7_5_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_2_4
#define init_7_5_2_4
void FASTCALL call_7_5_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_2_5
#define init_7_5_2_5
void FASTCALL call_7_5_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_2_6
#define init_7_5_2_6
void FASTCALL call_7_5_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_2_7
#define init_7_5_2_7
void FASTCALL call_7_5_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_2_8
#define init_7_5_2_8
void FASTCALL call_7_5_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_2_9
#define init_7_5_2_9
void FASTCALL call_7_5_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_2_10
#define init_7_5_2_10
void FASTCALL call_7_5_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_2_11
#define init_7_5_2_11
void FASTCALL call_7_5_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_2_12
#define init_7_5_2_12
void FASTCALL call_7_5_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_2_13
#define init_7_5_2_13
void FASTCALL call_7_5_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_2_14
#define init_7_5_2_14
void FASTCALL call_7_5_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_2_15
#define init_7_5_2_15
void FASTCALL call_7_5_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_3_0
#define init_7_5_3_0
void FASTCALL call_7_5_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_3_1
#define init_7_5_3_1
void FASTCALL call_7_5_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_3_2
#define init_7_5_3_2
void FASTCALL call_7_5_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_3_3
#define init_7_5_3_3
void FASTCALL call_7_5_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_3_4
#define init_7_5_3_4
void FASTCALL call_7_5_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_3_5
#define init_7_5_3_5
void FASTCALL call_7_5_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_3_6
#define init_7_5_3_6
void FASTCALL call_7_5_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_3_7
#define init_7_5_3_7
void FASTCALL call_7_5_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_3_8
#define init_7_5_3_8
void FASTCALL call_7_5_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_3_9
#define init_7_5_3_9
void FASTCALL call_7_5_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_3_10
#define init_7_5_3_10
void FASTCALL call_7_5_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_3_11
#define init_7_5_3_11
void FASTCALL call_7_5_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_3_12
#define init_7_5_3_12
void FASTCALL call_7_5_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_3_13
#define init_7_5_3_13
void FASTCALL call_7_5_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_3_14
#define init_7_5_3_14
void FASTCALL call_7_5_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_3_15
#define init_7_5_3_15
void FASTCALL call_7_5_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_4_0
#define init_7_5_4_0
void FASTCALL call_7_5_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_4_1
#define init_7_5_4_1
void FASTCALL call_7_5_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_4_2
#define init_7_5_4_2
void FASTCALL call_7_5_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_4_3
#define init_7_5_4_3
void FASTCALL call_7_5_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_4_4
#define init_7_5_4_4
void FASTCALL call_7_5_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_4_5
#define init_7_5_4_5
void FASTCALL call_7_5_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_4_6
#define init_7_5_4_6
void FASTCALL call_7_5_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_4_7
#define init_7_5_4_7
void FASTCALL call_7_5_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_4_8
#define init_7_5_4_8
void FASTCALL call_7_5_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_4_9
#define init_7_5_4_9
void FASTCALL call_7_5_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_4_10
#define init_7_5_4_10
void FASTCALL call_7_5_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_4_11
#define init_7_5_4_11
void FASTCALL call_7_5_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_4_12
#define init_7_5_4_12
void FASTCALL call_7_5_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_4_13
#define init_7_5_4_13
void FASTCALL call_7_5_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_4_14
#define init_7_5_4_14
void FASTCALL call_7_5_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_4_15
#define init_7_5_4_15
void FASTCALL call_7_5_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_5_0
#define init_7_5_5_0
void FASTCALL call_7_5_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_5_1
#define init_7_5_5_1
void FASTCALL call_7_5_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_5_2
#define init_7_5_5_2
void FASTCALL call_7_5_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_5_3
#define init_7_5_5_3
void FASTCALL call_7_5_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_5_4
#define init_7_5_5_4
void FASTCALL call_7_5_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_5_5
#define init_7_5_5_5
void FASTCALL call_7_5_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_5_6
#define init_7_5_5_6
void FASTCALL call_7_5_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_5_7
#define init_7_5_5_7
void FASTCALL call_7_5_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_5_8
#define init_7_5_5_8
void FASTCALL call_7_5_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_5_9
#define init_7_5_5_9
void FASTCALL call_7_5_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_5_10
#define init_7_5_5_10
void FASTCALL call_7_5_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_5_11
#define init_7_5_5_11
void FASTCALL call_7_5_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_5_12
#define init_7_5_5_12
void FASTCALL call_7_5_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_5_13
#define init_7_5_5_13
void FASTCALL call_7_5_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_5_14
#define init_7_5_5_14
void FASTCALL call_7_5_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_5_15
#define init_7_5_5_15
void FASTCALL call_7_5_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_6_0
#define init_7_5_6_0
void FASTCALL call_7_5_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_6_1
#define init_7_5_6_1
void FASTCALL call_7_5_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_6_2
#define init_7_5_6_2
void FASTCALL call_7_5_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_6_3
#define init_7_5_6_3
void FASTCALL call_7_5_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_6_4
#define init_7_5_6_4
void FASTCALL call_7_5_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_6_5
#define init_7_5_6_5
void FASTCALL call_7_5_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_6_6
#define init_7_5_6_6
void FASTCALL call_7_5_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_6_7
#define init_7_5_6_7
void FASTCALL call_7_5_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_6_8
#define init_7_5_6_8
void FASTCALL call_7_5_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_6_9
#define init_7_5_6_9
void FASTCALL call_7_5_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_6_10
#define init_7_5_6_10
void FASTCALL call_7_5_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_6_11
#define init_7_5_6_11
void FASTCALL call_7_5_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_6_12
#define init_7_5_6_12
void FASTCALL call_7_5_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_6_13
#define init_7_5_6_13
void FASTCALL call_7_5_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_6_14
#define init_7_5_6_14
void FASTCALL call_7_5_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_6_15
#define init_7_5_6_15
void FASTCALL call_7_5_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_7_0
#define init_7_5_7_0
void FASTCALL call_7_5_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_7_1
#define init_7_5_7_1
void FASTCALL call_7_5_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_7_2
#define init_7_5_7_2
void FASTCALL call_7_5_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_7_3
#define init_7_5_7_3
void FASTCALL call_7_5_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_7_4
#define init_7_5_7_4
void FASTCALL call_7_5_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_7_5
#define init_7_5_7_5
void FASTCALL call_7_5_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_7_6
#define init_7_5_7_6
void FASTCALL call_7_5_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_7_7
#define init_7_5_7_7
void FASTCALL call_7_5_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_7_8
#define init_7_5_7_8
void FASTCALL call_7_5_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_7_9
#define init_7_5_7_9
void FASTCALL call_7_5_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_7_10
#define init_7_5_7_10
void FASTCALL call_7_5_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_7_11
#define init_7_5_7_11
void FASTCALL call_7_5_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_7_12
#define init_7_5_7_12
void FASTCALL call_7_5_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_7_13
#define init_7_5_7_13
void FASTCALL call_7_5_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_7_14
#define init_7_5_7_14
void FASTCALL call_7_5_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_7_15
#define init_7_5_7_15
void FASTCALL call_7_5_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_8_0
#define init_7_5_8_0
void FASTCALL call_7_5_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_8_1
#define init_7_5_8_1
void FASTCALL call_7_5_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_8_2
#define init_7_5_8_2
void FASTCALL call_7_5_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_8_3
#define init_7_5_8_3
void FASTCALL call_7_5_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_8_4
#define init_7_5_8_4
void FASTCALL call_7_5_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_8_5
#define init_7_5_8_5
void FASTCALL call_7_5_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_8_6
#define init_7_5_8_6
void FASTCALL call_7_5_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_8_7
#define init_7_5_8_7
void FASTCALL call_7_5_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_8_8
#define init_7_5_8_8
void FASTCALL call_7_5_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_8_9
#define init_7_5_8_9
void FASTCALL call_7_5_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_8_10
#define init_7_5_8_10
void FASTCALL call_7_5_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_8_11
#define init_7_5_8_11
void FASTCALL call_7_5_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_8_12
#define init_7_5_8_12
void FASTCALL call_7_5_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_8_13
#define init_7_5_8_13
void FASTCALL call_7_5_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_8_14
#define init_7_5_8_14
void FASTCALL call_7_5_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_8_15
#define init_7_5_8_15
void FASTCALL call_7_5_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_9_0
#define init_7_5_9_0
void FASTCALL call_7_5_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_9_1
#define init_7_5_9_1
void FASTCALL call_7_5_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_9_2
#define init_7_5_9_2
void FASTCALL call_7_5_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_9_3
#define init_7_5_9_3
void FASTCALL call_7_5_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_9_4
#define init_7_5_9_4
void FASTCALL call_7_5_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_9_5
#define init_7_5_9_5
void FASTCALL call_7_5_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_9_6
#define init_7_5_9_6
void FASTCALL call_7_5_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_9_7
#define init_7_5_9_7
void FASTCALL call_7_5_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_9_8
#define init_7_5_9_8
void FASTCALL call_7_5_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_9_9
#define init_7_5_9_9
void FASTCALL call_7_5_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_9_10
#define init_7_5_9_10
void FASTCALL call_7_5_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_9_11
#define init_7_5_9_11
void FASTCALL call_7_5_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_9_12
#define init_7_5_9_12
void FASTCALL call_7_5_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_9_13
#define init_7_5_9_13
void FASTCALL call_7_5_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_9_14
#define init_7_5_9_14
void FASTCALL call_7_5_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_9_15
#define init_7_5_9_15
void FASTCALL call_7_5_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_10_0
#define init_7_5_10_0
void FASTCALL call_7_5_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_10_1
#define init_7_5_10_1
void FASTCALL call_7_5_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_10_2
#define init_7_5_10_2
void FASTCALL call_7_5_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_10_3
#define init_7_5_10_3
void FASTCALL call_7_5_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_10_4
#define init_7_5_10_4
void FASTCALL call_7_5_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_10_5
#define init_7_5_10_5
void FASTCALL call_7_5_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_10_6
#define init_7_5_10_6
void FASTCALL call_7_5_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_10_7
#define init_7_5_10_7
void FASTCALL call_7_5_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_10_8
#define init_7_5_10_8
void FASTCALL call_7_5_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_10_9
#define init_7_5_10_9
void FASTCALL call_7_5_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_10_10
#define init_7_5_10_10
void FASTCALL call_7_5_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_10_11
#define init_7_5_10_11
void FASTCALL call_7_5_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_10_12
#define init_7_5_10_12
void FASTCALL call_7_5_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_10_13
#define init_7_5_10_13
void FASTCALL call_7_5_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_10_14
#define init_7_5_10_14
void FASTCALL call_7_5_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_10_15
#define init_7_5_10_15
void FASTCALL call_7_5_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_11_0
#define init_7_5_11_0
void FASTCALL call_7_5_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_11_1
#define init_7_5_11_1
void FASTCALL call_7_5_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_11_2
#define init_7_5_11_2
void FASTCALL call_7_5_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_11_3
#define init_7_5_11_3
void FASTCALL call_7_5_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_11_4
#define init_7_5_11_4
void FASTCALL call_7_5_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_11_5
#define init_7_5_11_5
void FASTCALL call_7_5_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_11_6
#define init_7_5_11_6
void FASTCALL call_7_5_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_11_7
#define init_7_5_11_7
void FASTCALL call_7_5_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_11_8
#define init_7_5_11_8
void FASTCALL call_7_5_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_11_9
#define init_7_5_11_9
void FASTCALL call_7_5_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_11_10
#define init_7_5_11_10
void FASTCALL call_7_5_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_11_11
#define init_7_5_11_11
void FASTCALL call_7_5_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_11_12
#define init_7_5_11_12
void FASTCALL call_7_5_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_11_13
#define init_7_5_11_13
void FASTCALL call_7_5_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_11_14
#define init_7_5_11_14
void FASTCALL call_7_5_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_11_15
#define init_7_5_11_15
void FASTCALL call_7_5_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_12_0
#define init_7_5_12_0
void FASTCALL call_7_5_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_12_1
#define init_7_5_12_1
void FASTCALL call_7_5_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_12_2
#define init_7_5_12_2
void FASTCALL call_7_5_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_12_3
#define init_7_5_12_3
void FASTCALL call_7_5_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_12_4
#define init_7_5_12_4
void FASTCALL call_7_5_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_12_5
#define init_7_5_12_5
void FASTCALL call_7_5_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_12_6
#define init_7_5_12_6
void FASTCALL call_7_5_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_12_7
#define init_7_5_12_7
void FASTCALL call_7_5_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_12_8
#define init_7_5_12_8
void FASTCALL call_7_5_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_12_9
#define init_7_5_12_9
void FASTCALL call_7_5_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_12_10
#define init_7_5_12_10
void FASTCALL call_7_5_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_12_11
#define init_7_5_12_11
void FASTCALL call_7_5_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_12_12
#define init_7_5_12_12
void FASTCALL call_7_5_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_12_13
#define init_7_5_12_13
void FASTCALL call_7_5_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_12_14
#define init_7_5_12_14
void FASTCALL call_7_5_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_12_15
#define init_7_5_12_15
void FASTCALL call_7_5_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_13_0
#define init_7_5_13_0
void FASTCALL call_7_5_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_13_1
#define init_7_5_13_1
void FASTCALL call_7_5_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_13_2
#define init_7_5_13_2
void FASTCALL call_7_5_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_13_3
#define init_7_5_13_3
void FASTCALL call_7_5_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_13_4
#define init_7_5_13_4
void FASTCALL call_7_5_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_13_5
#define init_7_5_13_5
void FASTCALL call_7_5_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_13_6
#define init_7_5_13_6
void FASTCALL call_7_5_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_13_7
#define init_7_5_13_7
void FASTCALL call_7_5_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_13_8
#define init_7_5_13_8
void FASTCALL call_7_5_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_13_9
#define init_7_5_13_9
void FASTCALL call_7_5_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_13_10
#define init_7_5_13_10
void FASTCALL call_7_5_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_13_11
#define init_7_5_13_11
void FASTCALL call_7_5_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_13_12
#define init_7_5_13_12
void FASTCALL call_7_5_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_13_13
#define init_7_5_13_13
void FASTCALL call_7_5_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_13_14
#define init_7_5_13_14
void FASTCALL call_7_5_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_13_15
#define init_7_5_13_15
void FASTCALL call_7_5_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_14_0
#define init_7_5_14_0
void FASTCALL call_7_5_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_14_1
#define init_7_5_14_1
void FASTCALL call_7_5_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_14_2
#define init_7_5_14_2
void FASTCALL call_7_5_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_14_3
#define init_7_5_14_3
void FASTCALL call_7_5_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_14_4
#define init_7_5_14_4
void FASTCALL call_7_5_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_14_5
#define init_7_5_14_5
void FASTCALL call_7_5_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_14_6
#define init_7_5_14_6
void FASTCALL call_7_5_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_14_7
#define init_7_5_14_7
void FASTCALL call_7_5_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_14_8
#define init_7_5_14_8
void FASTCALL call_7_5_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_14_9
#define init_7_5_14_9
void FASTCALL call_7_5_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_14_10
#define init_7_5_14_10
void FASTCALL call_7_5_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_14_11
#define init_7_5_14_11
void FASTCALL call_7_5_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_14_12
#define init_7_5_14_12
void FASTCALL call_7_5_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_14_13
#define init_7_5_14_13
void FASTCALL call_7_5_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_14_14
#define init_7_5_14_14
void FASTCALL call_7_5_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_14_15
#define init_7_5_14_15
void FASTCALL call_7_5_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_15_0
#define init_7_5_15_0
void FASTCALL call_7_5_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_15_1
#define init_7_5_15_1
void FASTCALL call_7_5_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_15_2
#define init_7_5_15_2
void FASTCALL call_7_5_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_15_3
#define init_7_5_15_3
void FASTCALL call_7_5_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_15_4
#define init_7_5_15_4
void FASTCALL call_7_5_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_15_5
#define init_7_5_15_5
void FASTCALL call_7_5_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_15_6
#define init_7_5_15_6
void FASTCALL call_7_5_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_15_7
#define init_7_5_15_7
void FASTCALL call_7_5_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_15_8
#define init_7_5_15_8
void FASTCALL call_7_5_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_15_9
#define init_7_5_15_9
void FASTCALL call_7_5_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_15_10
#define init_7_5_15_10
void FASTCALL call_7_5_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_15_11
#define init_7_5_15_11
void FASTCALL call_7_5_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_15_12
#define init_7_5_15_12
void FASTCALL call_7_5_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_15_13
#define init_7_5_15_13
void FASTCALL call_7_5_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_15_14
#define init_7_5_15_14
void FASTCALL call_7_5_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_5_15_15
#define init_7_5_15_15
void FASTCALL call_7_5_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_0_0
#define init_7_6_0_0
void FASTCALL call_7_6_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_0_1
#define init_7_6_0_1
void FASTCALL call_7_6_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_0_2
#define init_7_6_0_2
void FASTCALL call_7_6_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_0_3
#define init_7_6_0_3
void FASTCALL call_7_6_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_0_4
#define init_7_6_0_4
void FASTCALL call_7_6_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_0_5
#define init_7_6_0_5
void FASTCALL call_7_6_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_0_6
#define init_7_6_0_6
void FASTCALL call_7_6_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_0_7
#define init_7_6_0_7
void FASTCALL call_7_6_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_0_8
#define init_7_6_0_8
void FASTCALL call_7_6_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_0_9
#define init_7_6_0_9
void FASTCALL call_7_6_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_0_10
#define init_7_6_0_10
void FASTCALL call_7_6_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_0_11
#define init_7_6_0_11
void FASTCALL call_7_6_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_0_12
#define init_7_6_0_12
void FASTCALL call_7_6_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_0_13
#define init_7_6_0_13
void FASTCALL call_7_6_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_0_14
#define init_7_6_0_14
void FASTCALL call_7_6_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_0_15
#define init_7_6_0_15
void FASTCALL call_7_6_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_1_0
#define init_7_6_1_0
void FASTCALL call_7_6_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_1_1
#define init_7_6_1_1
void FASTCALL call_7_6_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_1_2
#define init_7_6_1_2
void FASTCALL call_7_6_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_1_3
#define init_7_6_1_3
void FASTCALL call_7_6_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_1_4
#define init_7_6_1_4
void FASTCALL call_7_6_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_1_5
#define init_7_6_1_5
void FASTCALL call_7_6_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_1_6
#define init_7_6_1_6
void FASTCALL call_7_6_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_1_7
#define init_7_6_1_7
void FASTCALL call_7_6_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_1_8
#define init_7_6_1_8
void FASTCALL call_7_6_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_1_9
#define init_7_6_1_9
void FASTCALL call_7_6_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_1_10
#define init_7_6_1_10
void FASTCALL call_7_6_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_1_11
#define init_7_6_1_11
void FASTCALL call_7_6_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_1_12
#define init_7_6_1_12
void FASTCALL call_7_6_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_1_13
#define init_7_6_1_13
void FASTCALL call_7_6_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_1_14
#define init_7_6_1_14
void FASTCALL call_7_6_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_1_15
#define init_7_6_1_15
void FASTCALL call_7_6_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_2_0
#define init_7_6_2_0
void FASTCALL call_7_6_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_2_1
#define init_7_6_2_1
void FASTCALL call_7_6_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_2_2
#define init_7_6_2_2
void FASTCALL call_7_6_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_2_3
#define init_7_6_2_3
void FASTCALL call_7_6_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_2_4
#define init_7_6_2_4
void FASTCALL call_7_6_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_2_5
#define init_7_6_2_5
void FASTCALL call_7_6_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_2_6
#define init_7_6_2_6
void FASTCALL call_7_6_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_2_7
#define init_7_6_2_7
void FASTCALL call_7_6_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_2_8
#define init_7_6_2_8
void FASTCALL call_7_6_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_2_9
#define init_7_6_2_9
void FASTCALL call_7_6_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_2_10
#define init_7_6_2_10
void FASTCALL call_7_6_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_2_11
#define init_7_6_2_11
void FASTCALL call_7_6_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_2_12
#define init_7_6_2_12
void FASTCALL call_7_6_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_2_13
#define init_7_6_2_13
void FASTCALL call_7_6_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_2_14
#define init_7_6_2_14
void FASTCALL call_7_6_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_2_15
#define init_7_6_2_15
void FASTCALL call_7_6_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_3_0
#define init_7_6_3_0
void FASTCALL call_7_6_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_3_1
#define init_7_6_3_1
void FASTCALL call_7_6_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_3_2
#define init_7_6_3_2
void FASTCALL call_7_6_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_3_3
#define init_7_6_3_3
void FASTCALL call_7_6_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_3_4
#define init_7_6_3_4
void FASTCALL call_7_6_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_3_5
#define init_7_6_3_5
void FASTCALL call_7_6_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_3_6
#define init_7_6_3_6
void FASTCALL call_7_6_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_3_7
#define init_7_6_3_7
void FASTCALL call_7_6_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_3_8
#define init_7_6_3_8
void FASTCALL call_7_6_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_3_9
#define init_7_6_3_9
void FASTCALL call_7_6_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_3_10
#define init_7_6_3_10
void FASTCALL call_7_6_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_3_11
#define init_7_6_3_11
void FASTCALL call_7_6_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_3_12
#define init_7_6_3_12
void FASTCALL call_7_6_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_3_13
#define init_7_6_3_13
void FASTCALL call_7_6_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_3_14
#define init_7_6_3_14
void FASTCALL call_7_6_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_3_15
#define init_7_6_3_15
void FASTCALL call_7_6_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_4_0
#define init_7_6_4_0
void FASTCALL call_7_6_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_4_1
#define init_7_6_4_1
void FASTCALL call_7_6_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_4_2
#define init_7_6_4_2
void FASTCALL call_7_6_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_4_3
#define init_7_6_4_3
void FASTCALL call_7_6_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_4_4
#define init_7_6_4_4
void FASTCALL call_7_6_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_4_5
#define init_7_6_4_5
void FASTCALL call_7_6_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_4_6
#define init_7_6_4_6
void FASTCALL call_7_6_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_4_7
#define init_7_6_4_7
void FASTCALL call_7_6_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_4_8
#define init_7_6_4_8
void FASTCALL call_7_6_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_4_9
#define init_7_6_4_9
void FASTCALL call_7_6_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_4_10
#define init_7_6_4_10
void FASTCALL call_7_6_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_4_11
#define init_7_6_4_11
void FASTCALL call_7_6_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_4_12
#define init_7_6_4_12
void FASTCALL call_7_6_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_4_13
#define init_7_6_4_13
void FASTCALL call_7_6_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_4_14
#define init_7_6_4_14
void FASTCALL call_7_6_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_4_15
#define init_7_6_4_15
void FASTCALL call_7_6_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_5_0
#define init_7_6_5_0
void FASTCALL call_7_6_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_5_1
#define init_7_6_5_1
void FASTCALL call_7_6_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_5_2
#define init_7_6_5_2
void FASTCALL call_7_6_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_5_3
#define init_7_6_5_3
void FASTCALL call_7_6_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_5_4
#define init_7_6_5_4
void FASTCALL call_7_6_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_5_5
#define init_7_6_5_5
void FASTCALL call_7_6_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_5_6
#define init_7_6_5_6
void FASTCALL call_7_6_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_5_7
#define init_7_6_5_7
void FASTCALL call_7_6_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_5_8
#define init_7_6_5_8
void FASTCALL call_7_6_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_5_9
#define init_7_6_5_9
void FASTCALL call_7_6_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_5_10
#define init_7_6_5_10
void FASTCALL call_7_6_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_5_11
#define init_7_6_5_11
void FASTCALL call_7_6_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_5_12
#define init_7_6_5_12
void FASTCALL call_7_6_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_5_13
#define init_7_6_5_13
void FASTCALL call_7_6_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_5_14
#define init_7_6_5_14
void FASTCALL call_7_6_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_5_15
#define init_7_6_5_15
void FASTCALL call_7_6_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_6_0
#define init_7_6_6_0
void FASTCALL call_7_6_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_6_1
#define init_7_6_6_1
void FASTCALL call_7_6_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_6_2
#define init_7_6_6_2
void FASTCALL call_7_6_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_6_3
#define init_7_6_6_3
void FASTCALL call_7_6_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_6_4
#define init_7_6_6_4
void FASTCALL call_7_6_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_6_5
#define init_7_6_6_5
void FASTCALL call_7_6_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_6_6
#define init_7_6_6_6
void FASTCALL call_7_6_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_6_7
#define init_7_6_6_7
void FASTCALL call_7_6_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_6_8
#define init_7_6_6_8
void FASTCALL call_7_6_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_6_9
#define init_7_6_6_9
void FASTCALL call_7_6_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_6_10
#define init_7_6_6_10
void FASTCALL call_7_6_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_6_11
#define init_7_6_6_11
void FASTCALL call_7_6_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_6_12
#define init_7_6_6_12
void FASTCALL call_7_6_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_6_13
#define init_7_6_6_13
void FASTCALL call_7_6_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_6_14
#define init_7_6_6_14
void FASTCALL call_7_6_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_6_15
#define init_7_6_6_15
void FASTCALL call_7_6_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_7_0
#define init_7_6_7_0
void FASTCALL call_7_6_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_7_1
#define init_7_6_7_1
void FASTCALL call_7_6_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_7_2
#define init_7_6_7_2
void FASTCALL call_7_6_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_7_3
#define init_7_6_7_3
void FASTCALL call_7_6_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_7_4
#define init_7_6_7_4
void FASTCALL call_7_6_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_7_5
#define init_7_6_7_5
void FASTCALL call_7_6_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_7_6
#define init_7_6_7_6
void FASTCALL call_7_6_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_7_7
#define init_7_6_7_7
void FASTCALL call_7_6_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_7_8
#define init_7_6_7_8
void FASTCALL call_7_6_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_7_9
#define init_7_6_7_9
void FASTCALL call_7_6_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_7_10
#define init_7_6_7_10
void FASTCALL call_7_6_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_7_11
#define init_7_6_7_11
void FASTCALL call_7_6_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_7_12
#define init_7_6_7_12
void FASTCALL call_7_6_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_7_13
#define init_7_6_7_13
void FASTCALL call_7_6_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_7_14
#define init_7_6_7_14
void FASTCALL call_7_6_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_7_15
#define init_7_6_7_15
void FASTCALL call_7_6_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_8_0
#define init_7_6_8_0
void FASTCALL call_7_6_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_8_1
#define init_7_6_8_1
void FASTCALL call_7_6_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_8_2
#define init_7_6_8_2
void FASTCALL call_7_6_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_8_3
#define init_7_6_8_3
void FASTCALL call_7_6_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_8_4
#define init_7_6_8_4
void FASTCALL call_7_6_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_8_5
#define init_7_6_8_5
void FASTCALL call_7_6_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_8_6
#define init_7_6_8_6
void FASTCALL call_7_6_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_8_7
#define init_7_6_8_7
void FASTCALL call_7_6_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_8_8
#define init_7_6_8_8
void FASTCALL call_7_6_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_8_9
#define init_7_6_8_9
void FASTCALL call_7_6_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_8_10
#define init_7_6_8_10
void FASTCALL call_7_6_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_8_11
#define init_7_6_8_11
void FASTCALL call_7_6_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_8_12
#define init_7_6_8_12
void FASTCALL call_7_6_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_8_13
#define init_7_6_8_13
void FASTCALL call_7_6_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_8_14
#define init_7_6_8_14
void FASTCALL call_7_6_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_8_15
#define init_7_6_8_15
void FASTCALL call_7_6_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_9_0
#define init_7_6_9_0
void FASTCALL call_7_6_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_9_1
#define init_7_6_9_1
void FASTCALL call_7_6_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_9_2
#define init_7_6_9_2
void FASTCALL call_7_6_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_9_3
#define init_7_6_9_3
void FASTCALL call_7_6_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_9_4
#define init_7_6_9_4
void FASTCALL call_7_6_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_9_5
#define init_7_6_9_5
void FASTCALL call_7_6_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_9_6
#define init_7_6_9_6
void FASTCALL call_7_6_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_9_7
#define init_7_6_9_7
void FASTCALL call_7_6_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_9_8
#define init_7_6_9_8
void FASTCALL call_7_6_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_9_9
#define init_7_6_9_9
void FASTCALL call_7_6_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_9_10
#define init_7_6_9_10
void FASTCALL call_7_6_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_9_11
#define init_7_6_9_11
void FASTCALL call_7_6_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_9_12
#define init_7_6_9_12
void FASTCALL call_7_6_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_9_13
#define init_7_6_9_13
void FASTCALL call_7_6_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_9_14
#define init_7_6_9_14
void FASTCALL call_7_6_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_9_15
#define init_7_6_9_15
void FASTCALL call_7_6_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_10_0
#define init_7_6_10_0
void FASTCALL call_7_6_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_10_1
#define init_7_6_10_1
void FASTCALL call_7_6_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_10_2
#define init_7_6_10_2
void FASTCALL call_7_6_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_10_3
#define init_7_6_10_3
void FASTCALL call_7_6_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_10_4
#define init_7_6_10_4
void FASTCALL call_7_6_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_10_5
#define init_7_6_10_5
void FASTCALL call_7_6_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_10_6
#define init_7_6_10_6
void FASTCALL call_7_6_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_10_7
#define init_7_6_10_7
void FASTCALL call_7_6_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_10_8
#define init_7_6_10_8
void FASTCALL call_7_6_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_10_9
#define init_7_6_10_9
void FASTCALL call_7_6_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_10_10
#define init_7_6_10_10
void FASTCALL call_7_6_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_10_11
#define init_7_6_10_11
void FASTCALL call_7_6_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_10_12
#define init_7_6_10_12
void FASTCALL call_7_6_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_10_13
#define init_7_6_10_13
void FASTCALL call_7_6_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_10_14
#define init_7_6_10_14
void FASTCALL call_7_6_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_10_15
#define init_7_6_10_15
void FASTCALL call_7_6_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_11_0
#define init_7_6_11_0
void FASTCALL call_7_6_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_11_1
#define init_7_6_11_1
void FASTCALL call_7_6_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_11_2
#define init_7_6_11_2
void FASTCALL call_7_6_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_11_3
#define init_7_6_11_3
void FASTCALL call_7_6_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_11_4
#define init_7_6_11_4
void FASTCALL call_7_6_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_11_5
#define init_7_6_11_5
void FASTCALL call_7_6_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_11_6
#define init_7_6_11_6
void FASTCALL call_7_6_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_11_7
#define init_7_6_11_7
void FASTCALL call_7_6_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_11_8
#define init_7_6_11_8
void FASTCALL call_7_6_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_11_9
#define init_7_6_11_9
void FASTCALL call_7_6_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_11_10
#define init_7_6_11_10
void FASTCALL call_7_6_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_11_11
#define init_7_6_11_11
void FASTCALL call_7_6_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_11_12
#define init_7_6_11_12
void FASTCALL call_7_6_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_11_13
#define init_7_6_11_13
void FASTCALL call_7_6_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_11_14
#define init_7_6_11_14
void FASTCALL call_7_6_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_11_15
#define init_7_6_11_15
void FASTCALL call_7_6_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_12_0
#define init_7_6_12_0
void FASTCALL call_7_6_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_12_1
#define init_7_6_12_1
void FASTCALL call_7_6_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_12_2
#define init_7_6_12_2
void FASTCALL call_7_6_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_12_3
#define init_7_6_12_3
void FASTCALL call_7_6_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_12_4
#define init_7_6_12_4
void FASTCALL call_7_6_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_12_5
#define init_7_6_12_5
void FASTCALL call_7_6_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_12_6
#define init_7_6_12_6
void FASTCALL call_7_6_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_12_7
#define init_7_6_12_7
void FASTCALL call_7_6_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_12_8
#define init_7_6_12_8
void FASTCALL call_7_6_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_12_9
#define init_7_6_12_9
void FASTCALL call_7_6_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_12_10
#define init_7_6_12_10
void FASTCALL call_7_6_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_12_11
#define init_7_6_12_11
void FASTCALL call_7_6_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_12_12
#define init_7_6_12_12
void FASTCALL call_7_6_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_12_13
#define init_7_6_12_13
void FASTCALL call_7_6_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_12_14
#define init_7_6_12_14
void FASTCALL call_7_6_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_12_15
#define init_7_6_12_15
void FASTCALL call_7_6_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_13_0
#define init_7_6_13_0
void FASTCALL call_7_6_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_13_1
#define init_7_6_13_1
void FASTCALL call_7_6_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_13_2
#define init_7_6_13_2
void FASTCALL call_7_6_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_13_3
#define init_7_6_13_3
void FASTCALL call_7_6_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_13_4
#define init_7_6_13_4
void FASTCALL call_7_6_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_13_5
#define init_7_6_13_5
void FASTCALL call_7_6_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_13_6
#define init_7_6_13_6
void FASTCALL call_7_6_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_13_7
#define init_7_6_13_7
void FASTCALL call_7_6_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_13_8
#define init_7_6_13_8
void FASTCALL call_7_6_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_13_9
#define init_7_6_13_9
void FASTCALL call_7_6_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_13_10
#define init_7_6_13_10
void FASTCALL call_7_6_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_13_11
#define init_7_6_13_11
void FASTCALL call_7_6_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_13_12
#define init_7_6_13_12
void FASTCALL call_7_6_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_13_13
#define init_7_6_13_13
void FASTCALL call_7_6_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_13_14
#define init_7_6_13_14
void FASTCALL call_7_6_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_13_15
#define init_7_6_13_15
void FASTCALL call_7_6_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_14_0
#define init_7_6_14_0
void FASTCALL call_7_6_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_14_1
#define init_7_6_14_1
void FASTCALL call_7_6_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_14_2
#define init_7_6_14_2
void FASTCALL call_7_6_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_14_3
#define init_7_6_14_3
void FASTCALL call_7_6_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_14_4
#define init_7_6_14_4
void FASTCALL call_7_6_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_14_5
#define init_7_6_14_5
void FASTCALL call_7_6_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_14_6
#define init_7_6_14_6
void FASTCALL call_7_6_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_14_7
#define init_7_6_14_7
void FASTCALL call_7_6_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_14_8
#define init_7_6_14_8
void FASTCALL call_7_6_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_14_9
#define init_7_6_14_9
void FASTCALL call_7_6_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_14_10
#define init_7_6_14_10
void FASTCALL call_7_6_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_14_11
#define init_7_6_14_11
void FASTCALL call_7_6_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_14_12
#define init_7_6_14_12
void FASTCALL call_7_6_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_14_13
#define init_7_6_14_13
void FASTCALL call_7_6_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_14_14
#define init_7_6_14_14
void FASTCALL call_7_6_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_14_15
#define init_7_6_14_15
void FASTCALL call_7_6_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_15_0
#define init_7_6_15_0
void FASTCALL call_7_6_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_15_1
#define init_7_6_15_1
void FASTCALL call_7_6_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_15_2
#define init_7_6_15_2
void FASTCALL call_7_6_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_15_3
#define init_7_6_15_3
void FASTCALL call_7_6_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_15_4
#define init_7_6_15_4
void FASTCALL call_7_6_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_15_5
#define init_7_6_15_5
void FASTCALL call_7_6_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_15_6
#define init_7_6_15_6
void FASTCALL call_7_6_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_15_7
#define init_7_6_15_7
void FASTCALL call_7_6_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_15_8
#define init_7_6_15_8
void FASTCALL call_7_6_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_15_9
#define init_7_6_15_9
void FASTCALL call_7_6_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_15_10
#define init_7_6_15_10
void FASTCALL call_7_6_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_15_11
#define init_7_6_15_11
void FASTCALL call_7_6_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_15_12
#define init_7_6_15_12
void FASTCALL call_7_6_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_15_13
#define init_7_6_15_13
void FASTCALL call_7_6_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_15_14
#define init_7_6_15_14
void FASTCALL call_7_6_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_6_15_15
#define init_7_6_15_15
void FASTCALL call_7_6_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_0_0
#define init_7_7_0_0
void FASTCALL call_7_7_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_0_1
#define init_7_7_0_1
void FASTCALL call_7_7_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_0_2
#define init_7_7_0_2
void FASTCALL call_7_7_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_0_3
#define init_7_7_0_3
void FASTCALL call_7_7_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_0_4
#define init_7_7_0_4
void FASTCALL call_7_7_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_0_5
#define init_7_7_0_5
void FASTCALL call_7_7_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_0_6
#define init_7_7_0_6
void FASTCALL call_7_7_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_0_7
#define init_7_7_0_7
void FASTCALL call_7_7_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_0_8
#define init_7_7_0_8
void FASTCALL call_7_7_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_0_9
#define init_7_7_0_9
void FASTCALL call_7_7_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_0_10
#define init_7_7_0_10
void FASTCALL call_7_7_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_0_11
#define init_7_7_0_11
void FASTCALL call_7_7_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_0_12
#define init_7_7_0_12
void FASTCALL call_7_7_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_0_13
#define init_7_7_0_13
void FASTCALL call_7_7_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_0_14
#define init_7_7_0_14
void FASTCALL call_7_7_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_0_15
#define init_7_7_0_15
void FASTCALL call_7_7_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_1_0
#define init_7_7_1_0
void FASTCALL call_7_7_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_1_1
#define init_7_7_1_1
void FASTCALL call_7_7_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_1_2
#define init_7_7_1_2
void FASTCALL call_7_7_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_1_3
#define init_7_7_1_3
void FASTCALL call_7_7_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_1_4
#define init_7_7_1_4
void FASTCALL call_7_7_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_1_5
#define init_7_7_1_5
void FASTCALL call_7_7_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_1_6
#define init_7_7_1_6
void FASTCALL call_7_7_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_1_7
#define init_7_7_1_7
void FASTCALL call_7_7_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_1_8
#define init_7_7_1_8
void FASTCALL call_7_7_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_1_9
#define init_7_7_1_9
void FASTCALL call_7_7_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_1_10
#define init_7_7_1_10
void FASTCALL call_7_7_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_1_11
#define init_7_7_1_11
void FASTCALL call_7_7_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_1_12
#define init_7_7_1_12
void FASTCALL call_7_7_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_1_13
#define init_7_7_1_13
void FASTCALL call_7_7_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_1_14
#define init_7_7_1_14
void FASTCALL call_7_7_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_1_15
#define init_7_7_1_15
void FASTCALL call_7_7_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_2_0
#define init_7_7_2_0
void FASTCALL call_7_7_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_2_1
#define init_7_7_2_1
void FASTCALL call_7_7_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_2_2
#define init_7_7_2_2
void FASTCALL call_7_7_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_2_3
#define init_7_7_2_3
void FASTCALL call_7_7_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_2_4
#define init_7_7_2_4
void FASTCALL call_7_7_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_2_5
#define init_7_7_2_5
void FASTCALL call_7_7_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_2_6
#define init_7_7_2_6
void FASTCALL call_7_7_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_2_7
#define init_7_7_2_7
void FASTCALL call_7_7_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_2_8
#define init_7_7_2_8
void FASTCALL call_7_7_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_2_9
#define init_7_7_2_9
void FASTCALL call_7_7_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_2_10
#define init_7_7_2_10
void FASTCALL call_7_7_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_2_11
#define init_7_7_2_11
void FASTCALL call_7_7_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_2_12
#define init_7_7_2_12
void FASTCALL call_7_7_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_2_13
#define init_7_7_2_13
void FASTCALL call_7_7_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_2_14
#define init_7_7_2_14
void FASTCALL call_7_7_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_2_15
#define init_7_7_2_15
void FASTCALL call_7_7_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_3_0
#define init_7_7_3_0
void FASTCALL call_7_7_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_3_1
#define init_7_7_3_1
void FASTCALL call_7_7_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_3_2
#define init_7_7_3_2
void FASTCALL call_7_7_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_3_3
#define init_7_7_3_3
void FASTCALL call_7_7_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_3_4
#define init_7_7_3_4
void FASTCALL call_7_7_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_3_5
#define init_7_7_3_5
void FASTCALL call_7_7_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_3_6
#define init_7_7_3_6
void FASTCALL call_7_7_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_3_7
#define init_7_7_3_7
void FASTCALL call_7_7_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_3_8
#define init_7_7_3_8
void FASTCALL call_7_7_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_3_9
#define init_7_7_3_9
void FASTCALL call_7_7_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_3_10
#define init_7_7_3_10
void FASTCALL call_7_7_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_3_11
#define init_7_7_3_11
void FASTCALL call_7_7_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_3_12
#define init_7_7_3_12
void FASTCALL call_7_7_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_3_13
#define init_7_7_3_13
void FASTCALL call_7_7_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_3_14
#define init_7_7_3_14
void FASTCALL call_7_7_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_3_15
#define init_7_7_3_15
void FASTCALL call_7_7_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_4_0
#define init_7_7_4_0
void FASTCALL call_7_7_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_4_1
#define init_7_7_4_1
void FASTCALL call_7_7_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_4_2
#define init_7_7_4_2
void FASTCALL call_7_7_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_4_3
#define init_7_7_4_3
void FASTCALL call_7_7_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_4_4
#define init_7_7_4_4
void FASTCALL call_7_7_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_4_5
#define init_7_7_4_5
void FASTCALL call_7_7_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_4_6
#define init_7_7_4_6
void FASTCALL call_7_7_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_4_7
#define init_7_7_4_7
void FASTCALL call_7_7_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_4_8
#define init_7_7_4_8
void FASTCALL call_7_7_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_4_9
#define init_7_7_4_9
void FASTCALL call_7_7_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_4_10
#define init_7_7_4_10
void FASTCALL call_7_7_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_4_11
#define init_7_7_4_11
void FASTCALL call_7_7_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_4_12
#define init_7_7_4_12
void FASTCALL call_7_7_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_4_13
#define init_7_7_4_13
void FASTCALL call_7_7_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_4_14
#define init_7_7_4_14
void FASTCALL call_7_7_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_4_15
#define init_7_7_4_15
void FASTCALL call_7_7_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_5_0
#define init_7_7_5_0
void FASTCALL call_7_7_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_5_1
#define init_7_7_5_1
void FASTCALL call_7_7_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_5_2
#define init_7_7_5_2
void FASTCALL call_7_7_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_5_3
#define init_7_7_5_3
void FASTCALL call_7_7_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_5_4
#define init_7_7_5_4
void FASTCALL call_7_7_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_5_5
#define init_7_7_5_5
void FASTCALL call_7_7_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_5_6
#define init_7_7_5_6
void FASTCALL call_7_7_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_5_7
#define init_7_7_5_7
void FASTCALL call_7_7_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_5_8
#define init_7_7_5_8
void FASTCALL call_7_7_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_5_9
#define init_7_7_5_9
void FASTCALL call_7_7_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_5_10
#define init_7_7_5_10
void FASTCALL call_7_7_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_5_11
#define init_7_7_5_11
void FASTCALL call_7_7_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_5_12
#define init_7_7_5_12
void FASTCALL call_7_7_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_5_13
#define init_7_7_5_13
void FASTCALL call_7_7_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_5_14
#define init_7_7_5_14
void FASTCALL call_7_7_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_5_15
#define init_7_7_5_15
void FASTCALL call_7_7_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_6_0
#define init_7_7_6_0
void FASTCALL call_7_7_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_6_1
#define init_7_7_6_1
void FASTCALL call_7_7_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_6_2
#define init_7_7_6_2
void FASTCALL call_7_7_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_6_3
#define init_7_7_6_3
void FASTCALL call_7_7_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_6_4
#define init_7_7_6_4
void FASTCALL call_7_7_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_6_5
#define init_7_7_6_5
void FASTCALL call_7_7_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_6_6
#define init_7_7_6_6
void FASTCALL call_7_7_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_6_7
#define init_7_7_6_7
void FASTCALL call_7_7_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_6_8
#define init_7_7_6_8
void FASTCALL call_7_7_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_6_9
#define init_7_7_6_9
void FASTCALL call_7_7_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_6_10
#define init_7_7_6_10
void FASTCALL call_7_7_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_6_11
#define init_7_7_6_11
void FASTCALL call_7_7_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_6_12
#define init_7_7_6_12
void FASTCALL call_7_7_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_6_13
#define init_7_7_6_13
void FASTCALL call_7_7_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_6_14
#define init_7_7_6_14
void FASTCALL call_7_7_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_6_15
#define init_7_7_6_15
void FASTCALL call_7_7_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_7_0
#define init_7_7_7_0
void FASTCALL call_7_7_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_7_1
#define init_7_7_7_1
void FASTCALL call_7_7_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_7_2
#define init_7_7_7_2
void FASTCALL call_7_7_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_7_3
#define init_7_7_7_3
void FASTCALL call_7_7_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_7_4
#define init_7_7_7_4
void FASTCALL call_7_7_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_7_5
#define init_7_7_7_5
void FASTCALL call_7_7_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_7_6
#define init_7_7_7_6
void FASTCALL call_7_7_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_7_7
#define init_7_7_7_7
void FASTCALL call_7_7_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_7_8
#define init_7_7_7_8
void FASTCALL call_7_7_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_7_9
#define init_7_7_7_9
void FASTCALL call_7_7_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_7_10
#define init_7_7_7_10
void FASTCALL call_7_7_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_7_11
#define init_7_7_7_11
void FASTCALL call_7_7_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_7_12
#define init_7_7_7_12
void FASTCALL call_7_7_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_7_13
#define init_7_7_7_13
void FASTCALL call_7_7_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_7_14
#define init_7_7_7_14
void FASTCALL call_7_7_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_7_15
#define init_7_7_7_15
void FASTCALL call_7_7_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_8_0
#define init_7_7_8_0
void FASTCALL call_7_7_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_8_1
#define init_7_7_8_1
void FASTCALL call_7_7_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_8_2
#define init_7_7_8_2
void FASTCALL call_7_7_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_8_3
#define init_7_7_8_3
void FASTCALL call_7_7_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_8_4
#define init_7_7_8_4
void FASTCALL call_7_7_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_8_5
#define init_7_7_8_5
void FASTCALL call_7_7_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_8_6
#define init_7_7_8_6
void FASTCALL call_7_7_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_8_7
#define init_7_7_8_7
void FASTCALL call_7_7_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_8_8
#define init_7_7_8_8
void FASTCALL call_7_7_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_8_9
#define init_7_7_8_9
void FASTCALL call_7_7_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_8_10
#define init_7_7_8_10
void FASTCALL call_7_7_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_8_11
#define init_7_7_8_11
void FASTCALL call_7_7_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_8_12
#define init_7_7_8_12
void FASTCALL call_7_7_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_8_13
#define init_7_7_8_13
void FASTCALL call_7_7_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_8_14
#define init_7_7_8_14
void FASTCALL call_7_7_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_8_15
#define init_7_7_8_15
void FASTCALL call_7_7_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_9_0
#define init_7_7_9_0
void FASTCALL call_7_7_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_9_1
#define init_7_7_9_1
void FASTCALL call_7_7_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_9_2
#define init_7_7_9_2
void FASTCALL call_7_7_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_9_3
#define init_7_7_9_3
void FASTCALL call_7_7_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_9_4
#define init_7_7_9_4
void FASTCALL call_7_7_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_9_5
#define init_7_7_9_5
void FASTCALL call_7_7_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_9_6
#define init_7_7_9_6
void FASTCALL call_7_7_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_9_7
#define init_7_7_9_7
void FASTCALL call_7_7_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_9_8
#define init_7_7_9_8
void FASTCALL call_7_7_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_9_9
#define init_7_7_9_9
void FASTCALL call_7_7_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_9_10
#define init_7_7_9_10
void FASTCALL call_7_7_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_9_11
#define init_7_7_9_11
void FASTCALL call_7_7_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_9_12
#define init_7_7_9_12
void FASTCALL call_7_7_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_9_13
#define init_7_7_9_13
void FASTCALL call_7_7_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_9_14
#define init_7_7_9_14
void FASTCALL call_7_7_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_9_15
#define init_7_7_9_15
void FASTCALL call_7_7_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_10_0
#define init_7_7_10_0
void FASTCALL call_7_7_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_10_1
#define init_7_7_10_1
void FASTCALL call_7_7_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_10_2
#define init_7_7_10_2
void FASTCALL call_7_7_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_10_3
#define init_7_7_10_3
void FASTCALL call_7_7_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_10_4
#define init_7_7_10_4
void FASTCALL call_7_7_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_10_5
#define init_7_7_10_5
void FASTCALL call_7_7_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_10_6
#define init_7_7_10_6
void FASTCALL call_7_7_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_10_7
#define init_7_7_10_7
void FASTCALL call_7_7_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_10_8
#define init_7_7_10_8
void FASTCALL call_7_7_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_10_9
#define init_7_7_10_9
void FASTCALL call_7_7_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_10_10
#define init_7_7_10_10
void FASTCALL call_7_7_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_10_11
#define init_7_7_10_11
void FASTCALL call_7_7_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_10_12
#define init_7_7_10_12
void FASTCALL call_7_7_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_10_13
#define init_7_7_10_13
void FASTCALL call_7_7_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_10_14
#define init_7_7_10_14
void FASTCALL call_7_7_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_10_15
#define init_7_7_10_15
void FASTCALL call_7_7_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_11_0
#define init_7_7_11_0
void FASTCALL call_7_7_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_11_1
#define init_7_7_11_1
void FASTCALL call_7_7_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_11_2
#define init_7_7_11_2
void FASTCALL call_7_7_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_11_3
#define init_7_7_11_3
void FASTCALL call_7_7_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_11_4
#define init_7_7_11_4
void FASTCALL call_7_7_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_11_5
#define init_7_7_11_5
void FASTCALL call_7_7_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_11_6
#define init_7_7_11_6
void FASTCALL call_7_7_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_11_7
#define init_7_7_11_7
void FASTCALL call_7_7_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_11_8
#define init_7_7_11_8
void FASTCALL call_7_7_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_11_9
#define init_7_7_11_9
void FASTCALL call_7_7_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_11_10
#define init_7_7_11_10
void FASTCALL call_7_7_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_11_11
#define init_7_7_11_11
void FASTCALL call_7_7_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_11_12
#define init_7_7_11_12
void FASTCALL call_7_7_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_11_13
#define init_7_7_11_13
void FASTCALL call_7_7_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_11_14
#define init_7_7_11_14
void FASTCALL call_7_7_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_11_15
#define init_7_7_11_15
void FASTCALL call_7_7_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_12_0
#define init_7_7_12_0
void FASTCALL call_7_7_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_12_1
#define init_7_7_12_1
void FASTCALL call_7_7_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_12_2
#define init_7_7_12_2
void FASTCALL call_7_7_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_12_3
#define init_7_7_12_3
void FASTCALL call_7_7_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_12_4
#define init_7_7_12_4
void FASTCALL call_7_7_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_12_5
#define init_7_7_12_5
void FASTCALL call_7_7_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_12_6
#define init_7_7_12_6
void FASTCALL call_7_7_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_12_7
#define init_7_7_12_7
void FASTCALL call_7_7_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_12_8
#define init_7_7_12_8
void FASTCALL call_7_7_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_12_9
#define init_7_7_12_9
void FASTCALL call_7_7_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_12_10
#define init_7_7_12_10
void FASTCALL call_7_7_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_12_11
#define init_7_7_12_11
void FASTCALL call_7_7_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_12_12
#define init_7_7_12_12
void FASTCALL call_7_7_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_12_13
#define init_7_7_12_13
void FASTCALL call_7_7_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_12_14
#define init_7_7_12_14
void FASTCALL call_7_7_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_12_15
#define init_7_7_12_15
void FASTCALL call_7_7_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_13_0
#define init_7_7_13_0
void FASTCALL call_7_7_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_13_1
#define init_7_7_13_1
void FASTCALL call_7_7_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_13_2
#define init_7_7_13_2
void FASTCALL call_7_7_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_13_3
#define init_7_7_13_3
void FASTCALL call_7_7_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_13_4
#define init_7_7_13_4
void FASTCALL call_7_7_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_13_5
#define init_7_7_13_5
void FASTCALL call_7_7_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_13_6
#define init_7_7_13_6
void FASTCALL call_7_7_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_13_7
#define init_7_7_13_7
void FASTCALL call_7_7_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_13_8
#define init_7_7_13_8
void FASTCALL call_7_7_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_13_9
#define init_7_7_13_9
void FASTCALL call_7_7_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_13_10
#define init_7_7_13_10
void FASTCALL call_7_7_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_13_11
#define init_7_7_13_11
void FASTCALL call_7_7_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_13_12
#define init_7_7_13_12
void FASTCALL call_7_7_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_13_13
#define init_7_7_13_13
void FASTCALL call_7_7_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_13_14
#define init_7_7_13_14
void FASTCALL call_7_7_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_13_15
#define init_7_7_13_15
void FASTCALL call_7_7_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_14_0
#define init_7_7_14_0
void FASTCALL call_7_7_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_14_1
#define init_7_7_14_1
void FASTCALL call_7_7_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_14_2
#define init_7_7_14_2
void FASTCALL call_7_7_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_14_3
#define init_7_7_14_3
void FASTCALL call_7_7_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_14_4
#define init_7_7_14_4
void FASTCALL call_7_7_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_14_5
#define init_7_7_14_5
void FASTCALL call_7_7_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_14_6
#define init_7_7_14_6
void FASTCALL call_7_7_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_14_7
#define init_7_7_14_7
void FASTCALL call_7_7_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_14_8
#define init_7_7_14_8
void FASTCALL call_7_7_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_14_9
#define init_7_7_14_9
void FASTCALL call_7_7_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_14_10
#define init_7_7_14_10
void FASTCALL call_7_7_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_14_11
#define init_7_7_14_11
void FASTCALL call_7_7_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_14_12
#define init_7_7_14_12
void FASTCALL call_7_7_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_14_13
#define init_7_7_14_13
void FASTCALL call_7_7_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_14_14
#define init_7_7_14_14
void FASTCALL call_7_7_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_14_15
#define init_7_7_14_15
void FASTCALL call_7_7_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_15_0
#define init_7_7_15_0
void FASTCALL call_7_7_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_15_1
#define init_7_7_15_1
void FASTCALL call_7_7_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_15_2
#define init_7_7_15_2
void FASTCALL call_7_7_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_15_3
#define init_7_7_15_3
void FASTCALL call_7_7_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_15_4
#define init_7_7_15_4
void FASTCALL call_7_7_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_15_5
#define init_7_7_15_5
void FASTCALL call_7_7_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_15_6
#define init_7_7_15_6
void FASTCALL call_7_7_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_15_7
#define init_7_7_15_7
void FASTCALL call_7_7_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_15_8
#define init_7_7_15_8
void FASTCALL call_7_7_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_15_9
#define init_7_7_15_9
void FASTCALL call_7_7_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_15_10
#define init_7_7_15_10
void FASTCALL call_7_7_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_15_11
#define init_7_7_15_11
void FASTCALL call_7_7_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_15_12
#define init_7_7_15_12
void FASTCALL call_7_7_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_15_13
#define init_7_7_15_13
void FASTCALL call_7_7_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_15_14
#define init_7_7_15_14
void FASTCALL call_7_7_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_7_15_15
#define init_7_7_15_15
void FASTCALL call_7_7_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_0_0
#define init_7_8_0_0
void FASTCALL call_7_8_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_0_1
#define init_7_8_0_1
void FASTCALL call_7_8_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_0_2
#define init_7_8_0_2
void FASTCALL call_7_8_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_0_3
#define init_7_8_0_3
void FASTCALL call_7_8_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_0_4
#define init_7_8_0_4
void FASTCALL call_7_8_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_0_5
#define init_7_8_0_5
void FASTCALL call_7_8_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_0_6
#define init_7_8_0_6
void FASTCALL call_7_8_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_0_7
#define init_7_8_0_7
void FASTCALL call_7_8_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_0_8
#define init_7_8_0_8
void FASTCALL call_7_8_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_0_9
#define init_7_8_0_9
void FASTCALL call_7_8_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_0_10
#define init_7_8_0_10
void FASTCALL call_7_8_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_0_11
#define init_7_8_0_11
void FASTCALL call_7_8_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_0_12
#define init_7_8_0_12
void FASTCALL call_7_8_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_0_13
#define init_7_8_0_13
void FASTCALL call_7_8_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_0_14
#define init_7_8_0_14
void FASTCALL call_7_8_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_0_15
#define init_7_8_0_15
void FASTCALL call_7_8_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_1_0
#define init_7_8_1_0
void FASTCALL call_7_8_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_1_1
#define init_7_8_1_1
void FASTCALL call_7_8_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_1_2
#define init_7_8_1_2
void FASTCALL call_7_8_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_1_3
#define init_7_8_1_3
void FASTCALL call_7_8_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_1_4
#define init_7_8_1_4
void FASTCALL call_7_8_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_1_5
#define init_7_8_1_5
void FASTCALL call_7_8_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_1_6
#define init_7_8_1_6
void FASTCALL call_7_8_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_1_7
#define init_7_8_1_7
void FASTCALL call_7_8_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_1_8
#define init_7_8_1_8
void FASTCALL call_7_8_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_1_9
#define init_7_8_1_9
void FASTCALL call_7_8_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_1_10
#define init_7_8_1_10
void FASTCALL call_7_8_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_1_11
#define init_7_8_1_11
void FASTCALL call_7_8_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_1_12
#define init_7_8_1_12
void FASTCALL call_7_8_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_1_13
#define init_7_8_1_13
void FASTCALL call_7_8_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_1_14
#define init_7_8_1_14
void FASTCALL call_7_8_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_1_15
#define init_7_8_1_15
void FASTCALL call_7_8_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_2_0
#define init_7_8_2_0
void FASTCALL call_7_8_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_2_1
#define init_7_8_2_1
void FASTCALL call_7_8_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_2_2
#define init_7_8_2_2
void FASTCALL call_7_8_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_2_3
#define init_7_8_2_3
void FASTCALL call_7_8_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_2_4
#define init_7_8_2_4
void FASTCALL call_7_8_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_2_5
#define init_7_8_2_5
void FASTCALL call_7_8_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_2_6
#define init_7_8_2_6
void FASTCALL call_7_8_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_2_7
#define init_7_8_2_7
void FASTCALL call_7_8_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_2_8
#define init_7_8_2_8
void FASTCALL call_7_8_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_2_9
#define init_7_8_2_9
void FASTCALL call_7_8_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_2_10
#define init_7_8_2_10
void FASTCALL call_7_8_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_2_11
#define init_7_8_2_11
void FASTCALL call_7_8_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_2_12
#define init_7_8_2_12
void FASTCALL call_7_8_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_2_13
#define init_7_8_2_13
void FASTCALL call_7_8_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_2_14
#define init_7_8_2_14
void FASTCALL call_7_8_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_2_15
#define init_7_8_2_15
void FASTCALL call_7_8_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_3_0
#define init_7_8_3_0
void FASTCALL call_7_8_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_3_1
#define init_7_8_3_1
void FASTCALL call_7_8_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_3_2
#define init_7_8_3_2
void FASTCALL call_7_8_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_3_3
#define init_7_8_3_3
void FASTCALL call_7_8_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_3_4
#define init_7_8_3_4
void FASTCALL call_7_8_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_3_5
#define init_7_8_3_5
void FASTCALL call_7_8_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_3_6
#define init_7_8_3_6
void FASTCALL call_7_8_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_3_7
#define init_7_8_3_7
void FASTCALL call_7_8_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_3_8
#define init_7_8_3_8
void FASTCALL call_7_8_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_3_9
#define init_7_8_3_9
void FASTCALL call_7_8_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_3_10
#define init_7_8_3_10
void FASTCALL call_7_8_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_3_11
#define init_7_8_3_11
void FASTCALL call_7_8_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_3_12
#define init_7_8_3_12
void FASTCALL call_7_8_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_3_13
#define init_7_8_3_13
void FASTCALL call_7_8_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_3_14
#define init_7_8_3_14
void FASTCALL call_7_8_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_3_15
#define init_7_8_3_15
void FASTCALL call_7_8_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_4_0
#define init_7_8_4_0
void FASTCALL call_7_8_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_4_1
#define init_7_8_4_1
void FASTCALL call_7_8_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_4_2
#define init_7_8_4_2
void FASTCALL call_7_8_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_4_3
#define init_7_8_4_3
void FASTCALL call_7_8_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_4_4
#define init_7_8_4_4
void FASTCALL call_7_8_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_4_5
#define init_7_8_4_5
void FASTCALL call_7_8_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_4_6
#define init_7_8_4_6
void FASTCALL call_7_8_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_4_7
#define init_7_8_4_7
void FASTCALL call_7_8_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_4_8
#define init_7_8_4_8
void FASTCALL call_7_8_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_4_9
#define init_7_8_4_9
void FASTCALL call_7_8_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_4_10
#define init_7_8_4_10
void FASTCALL call_7_8_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_4_11
#define init_7_8_4_11
void FASTCALL call_7_8_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_4_12
#define init_7_8_4_12
void FASTCALL call_7_8_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_4_13
#define init_7_8_4_13
void FASTCALL call_7_8_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_4_14
#define init_7_8_4_14
void FASTCALL call_7_8_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_4_15
#define init_7_8_4_15
void FASTCALL call_7_8_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_5_0
#define init_7_8_5_0
void FASTCALL call_7_8_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_5_1
#define init_7_8_5_1
void FASTCALL call_7_8_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_5_2
#define init_7_8_5_2
void FASTCALL call_7_8_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_5_3
#define init_7_8_5_3
void FASTCALL call_7_8_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_5_4
#define init_7_8_5_4
void FASTCALL call_7_8_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_5_5
#define init_7_8_5_5
void FASTCALL call_7_8_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_5_6
#define init_7_8_5_6
void FASTCALL call_7_8_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_5_7
#define init_7_8_5_7
void FASTCALL call_7_8_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_5_8
#define init_7_8_5_8
void FASTCALL call_7_8_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_5_9
#define init_7_8_5_9
void FASTCALL call_7_8_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_5_10
#define init_7_8_5_10
void FASTCALL call_7_8_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_5_11
#define init_7_8_5_11
void FASTCALL call_7_8_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_5_12
#define init_7_8_5_12
void FASTCALL call_7_8_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_5_13
#define init_7_8_5_13
void FASTCALL call_7_8_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_5_14
#define init_7_8_5_14
void FASTCALL call_7_8_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_5_15
#define init_7_8_5_15
void FASTCALL call_7_8_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_6_0
#define init_7_8_6_0
void FASTCALL call_7_8_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_6_1
#define init_7_8_6_1
void FASTCALL call_7_8_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_6_2
#define init_7_8_6_2
void FASTCALL call_7_8_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_6_3
#define init_7_8_6_3
void FASTCALL call_7_8_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_6_4
#define init_7_8_6_4
void FASTCALL call_7_8_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_6_5
#define init_7_8_6_5
void FASTCALL call_7_8_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_6_6
#define init_7_8_6_6
void FASTCALL call_7_8_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_6_7
#define init_7_8_6_7
void FASTCALL call_7_8_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_6_8
#define init_7_8_6_8
void FASTCALL call_7_8_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_6_9
#define init_7_8_6_9
void FASTCALL call_7_8_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_6_10
#define init_7_8_6_10
void FASTCALL call_7_8_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_6_11
#define init_7_8_6_11
void FASTCALL call_7_8_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_6_12
#define init_7_8_6_12
void FASTCALL call_7_8_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_6_13
#define init_7_8_6_13
void FASTCALL call_7_8_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_6_14
#define init_7_8_6_14
void FASTCALL call_7_8_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_6_15
#define init_7_8_6_15
void FASTCALL call_7_8_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_7_0
#define init_7_8_7_0
void FASTCALL call_7_8_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_7_1
#define init_7_8_7_1
void FASTCALL call_7_8_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_7_2
#define init_7_8_7_2
void FASTCALL call_7_8_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_7_3
#define init_7_8_7_3
void FASTCALL call_7_8_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_7_4
#define init_7_8_7_4
void FASTCALL call_7_8_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_7_5
#define init_7_8_7_5
void FASTCALL call_7_8_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_7_6
#define init_7_8_7_6
void FASTCALL call_7_8_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_7_7
#define init_7_8_7_7
void FASTCALL call_7_8_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_7_8
#define init_7_8_7_8
void FASTCALL call_7_8_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_7_9
#define init_7_8_7_9
void FASTCALL call_7_8_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_7_10
#define init_7_8_7_10
void FASTCALL call_7_8_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_7_11
#define init_7_8_7_11
void FASTCALL call_7_8_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_7_12
#define init_7_8_7_12
void FASTCALL call_7_8_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_7_13
#define init_7_8_7_13
void FASTCALL call_7_8_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_7_14
#define init_7_8_7_14
void FASTCALL call_7_8_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_7_15
#define init_7_8_7_15
void FASTCALL call_7_8_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_8_0
#define init_7_8_8_0
void FASTCALL call_7_8_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_8_1
#define init_7_8_8_1
void FASTCALL call_7_8_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_8_2
#define init_7_8_8_2
void FASTCALL call_7_8_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_8_3
#define init_7_8_8_3
void FASTCALL call_7_8_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_8_4
#define init_7_8_8_4
void FASTCALL call_7_8_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_8_5
#define init_7_8_8_5
void FASTCALL call_7_8_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_8_6
#define init_7_8_8_6
void FASTCALL call_7_8_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_8_7
#define init_7_8_8_7
void FASTCALL call_7_8_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_8_8
#define init_7_8_8_8
void FASTCALL call_7_8_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_8_9
#define init_7_8_8_9
void FASTCALL call_7_8_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_8_10
#define init_7_8_8_10
void FASTCALL call_7_8_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_8_11
#define init_7_8_8_11
void FASTCALL call_7_8_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_8_12
#define init_7_8_8_12
void FASTCALL call_7_8_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_8_13
#define init_7_8_8_13
void FASTCALL call_7_8_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_8_14
#define init_7_8_8_14
void FASTCALL call_7_8_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_8_15
#define init_7_8_8_15
void FASTCALL call_7_8_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_9_0
#define init_7_8_9_0
void FASTCALL call_7_8_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_9_1
#define init_7_8_9_1
void FASTCALL call_7_8_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_9_2
#define init_7_8_9_2
void FASTCALL call_7_8_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_9_3
#define init_7_8_9_3
void FASTCALL call_7_8_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_9_4
#define init_7_8_9_4
void FASTCALL call_7_8_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_9_5
#define init_7_8_9_5
void FASTCALL call_7_8_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_9_6
#define init_7_8_9_6
void FASTCALL call_7_8_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_9_7
#define init_7_8_9_7
void FASTCALL call_7_8_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_9_8
#define init_7_8_9_8
void FASTCALL call_7_8_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_9_9
#define init_7_8_9_9
void FASTCALL call_7_8_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_9_10
#define init_7_8_9_10
void FASTCALL call_7_8_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_9_11
#define init_7_8_9_11
void FASTCALL call_7_8_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_9_12
#define init_7_8_9_12
void FASTCALL call_7_8_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_9_13
#define init_7_8_9_13
void FASTCALL call_7_8_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_9_14
#define init_7_8_9_14
void FASTCALL call_7_8_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_9_15
#define init_7_8_9_15
void FASTCALL call_7_8_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_10_0
#define init_7_8_10_0
void FASTCALL call_7_8_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_10_1
#define init_7_8_10_1
void FASTCALL call_7_8_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_10_2
#define init_7_8_10_2
void FASTCALL call_7_8_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_10_3
#define init_7_8_10_3
void FASTCALL call_7_8_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_10_4
#define init_7_8_10_4
void FASTCALL call_7_8_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_10_5
#define init_7_8_10_5
void FASTCALL call_7_8_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_10_6
#define init_7_8_10_6
void FASTCALL call_7_8_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_10_7
#define init_7_8_10_7
void FASTCALL call_7_8_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_10_8
#define init_7_8_10_8
void FASTCALL call_7_8_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_10_9
#define init_7_8_10_9
void FASTCALL call_7_8_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_10_10
#define init_7_8_10_10
void FASTCALL call_7_8_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_10_11
#define init_7_8_10_11
void FASTCALL call_7_8_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_10_12
#define init_7_8_10_12
void FASTCALL call_7_8_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_10_13
#define init_7_8_10_13
void FASTCALL call_7_8_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_10_14
#define init_7_8_10_14
void FASTCALL call_7_8_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_10_15
#define init_7_8_10_15
void FASTCALL call_7_8_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_11_0
#define init_7_8_11_0
void FASTCALL call_7_8_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_11_1
#define init_7_8_11_1
void FASTCALL call_7_8_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_11_2
#define init_7_8_11_2
void FASTCALL call_7_8_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_11_3
#define init_7_8_11_3
void FASTCALL call_7_8_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_11_4
#define init_7_8_11_4
void FASTCALL call_7_8_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_11_5
#define init_7_8_11_5
void FASTCALL call_7_8_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_11_6
#define init_7_8_11_6
void FASTCALL call_7_8_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_11_7
#define init_7_8_11_7
void FASTCALL call_7_8_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_11_8
#define init_7_8_11_8
void FASTCALL call_7_8_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_11_9
#define init_7_8_11_9
void FASTCALL call_7_8_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_11_10
#define init_7_8_11_10
void FASTCALL call_7_8_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_11_11
#define init_7_8_11_11
void FASTCALL call_7_8_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_11_12
#define init_7_8_11_12
void FASTCALL call_7_8_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_11_13
#define init_7_8_11_13
void FASTCALL call_7_8_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_11_14
#define init_7_8_11_14
void FASTCALL call_7_8_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_11_15
#define init_7_8_11_15
void FASTCALL call_7_8_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_12_0
#define init_7_8_12_0
void FASTCALL call_7_8_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_12_1
#define init_7_8_12_1
void FASTCALL call_7_8_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_12_2
#define init_7_8_12_2
void FASTCALL call_7_8_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_12_3
#define init_7_8_12_3
void FASTCALL call_7_8_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_12_4
#define init_7_8_12_4
void FASTCALL call_7_8_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_12_5
#define init_7_8_12_5
void FASTCALL call_7_8_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_12_6
#define init_7_8_12_6
void FASTCALL call_7_8_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_12_7
#define init_7_8_12_7
void FASTCALL call_7_8_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_12_8
#define init_7_8_12_8
void FASTCALL call_7_8_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_12_9
#define init_7_8_12_9
void FASTCALL call_7_8_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_12_10
#define init_7_8_12_10
void FASTCALL call_7_8_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_12_11
#define init_7_8_12_11
void FASTCALL call_7_8_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_12_12
#define init_7_8_12_12
void FASTCALL call_7_8_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_12_13
#define init_7_8_12_13
void FASTCALL call_7_8_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_12_14
#define init_7_8_12_14
void FASTCALL call_7_8_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_12_15
#define init_7_8_12_15
void FASTCALL call_7_8_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_13_0
#define init_7_8_13_0
void FASTCALL call_7_8_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_13_1
#define init_7_8_13_1
void FASTCALL call_7_8_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_13_2
#define init_7_8_13_2
void FASTCALL call_7_8_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_13_3
#define init_7_8_13_3
void FASTCALL call_7_8_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_13_4
#define init_7_8_13_4
void FASTCALL call_7_8_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_13_5
#define init_7_8_13_5
void FASTCALL call_7_8_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_13_6
#define init_7_8_13_6
void FASTCALL call_7_8_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_13_7
#define init_7_8_13_7
void FASTCALL call_7_8_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_13_8
#define init_7_8_13_8
void FASTCALL call_7_8_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_13_9
#define init_7_8_13_9
void FASTCALL call_7_8_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_13_10
#define init_7_8_13_10
void FASTCALL call_7_8_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_13_11
#define init_7_8_13_11
void FASTCALL call_7_8_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_13_12
#define init_7_8_13_12
void FASTCALL call_7_8_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_13_13
#define init_7_8_13_13
void FASTCALL call_7_8_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_13_14
#define init_7_8_13_14
void FASTCALL call_7_8_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_13_15
#define init_7_8_13_15
void FASTCALL call_7_8_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_14_0
#define init_7_8_14_0
void FASTCALL call_7_8_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_14_1
#define init_7_8_14_1
void FASTCALL call_7_8_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_14_2
#define init_7_8_14_2
void FASTCALL call_7_8_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_14_3
#define init_7_8_14_3
void FASTCALL call_7_8_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_14_4
#define init_7_8_14_4
void FASTCALL call_7_8_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_14_5
#define init_7_8_14_5
void FASTCALL call_7_8_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_14_6
#define init_7_8_14_6
void FASTCALL call_7_8_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_14_7
#define init_7_8_14_7
void FASTCALL call_7_8_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_14_8
#define init_7_8_14_8
void FASTCALL call_7_8_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_14_9
#define init_7_8_14_9
void FASTCALL call_7_8_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_14_10
#define init_7_8_14_10
void FASTCALL call_7_8_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_14_11
#define init_7_8_14_11
void FASTCALL call_7_8_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_14_12
#define init_7_8_14_12
void FASTCALL call_7_8_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_14_13
#define init_7_8_14_13
void FASTCALL call_7_8_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_14_14
#define init_7_8_14_14
void FASTCALL call_7_8_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_14_15
#define init_7_8_14_15
void FASTCALL call_7_8_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_15_0
#define init_7_8_15_0
void FASTCALL call_7_8_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_15_1
#define init_7_8_15_1
void FASTCALL call_7_8_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_15_2
#define init_7_8_15_2
void FASTCALL call_7_8_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_15_3
#define init_7_8_15_3
void FASTCALL call_7_8_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_15_4
#define init_7_8_15_4
void FASTCALL call_7_8_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_15_5
#define init_7_8_15_5
void FASTCALL call_7_8_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_15_6
#define init_7_8_15_6
void FASTCALL call_7_8_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_15_7
#define init_7_8_15_7
void FASTCALL call_7_8_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_15_8
#define init_7_8_15_8
void FASTCALL call_7_8_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_15_9
#define init_7_8_15_9
void FASTCALL call_7_8_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_15_10
#define init_7_8_15_10
void FASTCALL call_7_8_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_15_11
#define init_7_8_15_11
void FASTCALL call_7_8_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_15_12
#define init_7_8_15_12
void FASTCALL call_7_8_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_15_13
#define init_7_8_15_13
void FASTCALL call_7_8_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_15_14
#define init_7_8_15_14
void FASTCALL call_7_8_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_8_15_15
#define init_7_8_15_15
void FASTCALL call_7_8_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_0_0
#define init_7_9_0_0
void FASTCALL call_7_9_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_0_1
#define init_7_9_0_1
void FASTCALL call_7_9_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_0_2
#define init_7_9_0_2
void FASTCALL call_7_9_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_0_3
#define init_7_9_0_3
void FASTCALL call_7_9_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_0_4
#define init_7_9_0_4
void FASTCALL call_7_9_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_0_5
#define init_7_9_0_5
void FASTCALL call_7_9_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_0_6
#define init_7_9_0_6
void FASTCALL call_7_9_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_0_7
#define init_7_9_0_7
void FASTCALL call_7_9_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_0_8
#define init_7_9_0_8
void FASTCALL call_7_9_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_0_9
#define init_7_9_0_9
void FASTCALL call_7_9_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_0_10
#define init_7_9_0_10
void FASTCALL call_7_9_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_0_11
#define init_7_9_0_11
void FASTCALL call_7_9_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_0_12
#define init_7_9_0_12
void FASTCALL call_7_9_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_0_13
#define init_7_9_0_13
void FASTCALL call_7_9_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_0_14
#define init_7_9_0_14
void FASTCALL call_7_9_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_0_15
#define init_7_9_0_15
void FASTCALL call_7_9_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_1_0
#define init_7_9_1_0
void FASTCALL call_7_9_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_1_1
#define init_7_9_1_1
void FASTCALL call_7_9_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_1_2
#define init_7_9_1_2
void FASTCALL call_7_9_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_1_3
#define init_7_9_1_3
void FASTCALL call_7_9_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_1_4
#define init_7_9_1_4
void FASTCALL call_7_9_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_1_5
#define init_7_9_1_5
void FASTCALL call_7_9_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_1_6
#define init_7_9_1_6
void FASTCALL call_7_9_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_1_7
#define init_7_9_1_7
void FASTCALL call_7_9_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_1_8
#define init_7_9_1_8
void FASTCALL call_7_9_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_1_9
#define init_7_9_1_9
void FASTCALL call_7_9_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_1_10
#define init_7_9_1_10
void FASTCALL call_7_9_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_1_11
#define init_7_9_1_11
void FASTCALL call_7_9_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_1_12
#define init_7_9_1_12
void FASTCALL call_7_9_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_1_13
#define init_7_9_1_13
void FASTCALL call_7_9_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_1_14
#define init_7_9_1_14
void FASTCALL call_7_9_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_1_15
#define init_7_9_1_15
void FASTCALL call_7_9_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_2_0
#define init_7_9_2_0
void FASTCALL call_7_9_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_2_1
#define init_7_9_2_1
void FASTCALL call_7_9_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_2_2
#define init_7_9_2_2
void FASTCALL call_7_9_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_2_3
#define init_7_9_2_3
void FASTCALL call_7_9_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_2_4
#define init_7_9_2_4
void FASTCALL call_7_9_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_2_5
#define init_7_9_2_5
void FASTCALL call_7_9_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_2_6
#define init_7_9_2_6
void FASTCALL call_7_9_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_2_7
#define init_7_9_2_7
void FASTCALL call_7_9_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_2_8
#define init_7_9_2_8
void FASTCALL call_7_9_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_2_9
#define init_7_9_2_9
void FASTCALL call_7_9_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_2_10
#define init_7_9_2_10
void FASTCALL call_7_9_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_2_11
#define init_7_9_2_11
void FASTCALL call_7_9_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_2_12
#define init_7_9_2_12
void FASTCALL call_7_9_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_2_13
#define init_7_9_2_13
void FASTCALL call_7_9_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_2_14
#define init_7_9_2_14
void FASTCALL call_7_9_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_2_15
#define init_7_9_2_15
void FASTCALL call_7_9_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_3_0
#define init_7_9_3_0
void FASTCALL call_7_9_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_3_1
#define init_7_9_3_1
void FASTCALL call_7_9_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_3_2
#define init_7_9_3_2
void FASTCALL call_7_9_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_3_3
#define init_7_9_3_3
void FASTCALL call_7_9_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_3_4
#define init_7_9_3_4
void FASTCALL call_7_9_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_3_5
#define init_7_9_3_5
void FASTCALL call_7_9_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_3_6
#define init_7_9_3_6
void FASTCALL call_7_9_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_3_7
#define init_7_9_3_7
void FASTCALL call_7_9_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_3_8
#define init_7_9_3_8
void FASTCALL call_7_9_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_3_9
#define init_7_9_3_9
void FASTCALL call_7_9_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_3_10
#define init_7_9_3_10
void FASTCALL call_7_9_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_3_11
#define init_7_9_3_11
void FASTCALL call_7_9_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_3_12
#define init_7_9_3_12
void FASTCALL call_7_9_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_3_13
#define init_7_9_3_13
void FASTCALL call_7_9_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_3_14
#define init_7_9_3_14
void FASTCALL call_7_9_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_3_15
#define init_7_9_3_15
void FASTCALL call_7_9_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_4_0
#define init_7_9_4_0
void FASTCALL call_7_9_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_4_1
#define init_7_9_4_1
void FASTCALL call_7_9_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_4_2
#define init_7_9_4_2
void FASTCALL call_7_9_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_4_3
#define init_7_9_4_3
void FASTCALL call_7_9_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_4_4
#define init_7_9_4_4
void FASTCALL call_7_9_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_4_5
#define init_7_9_4_5
void FASTCALL call_7_9_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_4_6
#define init_7_9_4_6
void FASTCALL call_7_9_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_4_7
#define init_7_9_4_7
void FASTCALL call_7_9_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_4_8
#define init_7_9_4_8
void FASTCALL call_7_9_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_4_9
#define init_7_9_4_9
void FASTCALL call_7_9_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_4_10
#define init_7_9_4_10
void FASTCALL call_7_9_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_4_11
#define init_7_9_4_11
void FASTCALL call_7_9_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_4_12
#define init_7_9_4_12
void FASTCALL call_7_9_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_4_13
#define init_7_9_4_13
void FASTCALL call_7_9_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_4_14
#define init_7_9_4_14
void FASTCALL call_7_9_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_4_15
#define init_7_9_4_15
void FASTCALL call_7_9_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_5_0
#define init_7_9_5_0
void FASTCALL call_7_9_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_5_1
#define init_7_9_5_1
void FASTCALL call_7_9_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_5_2
#define init_7_9_5_2
void FASTCALL call_7_9_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_5_3
#define init_7_9_5_3
void FASTCALL call_7_9_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_5_4
#define init_7_9_5_4
void FASTCALL call_7_9_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_5_5
#define init_7_9_5_5
void FASTCALL call_7_9_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_5_6
#define init_7_9_5_6
void FASTCALL call_7_9_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_5_7
#define init_7_9_5_7
void FASTCALL call_7_9_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_5_8
#define init_7_9_5_8
void FASTCALL call_7_9_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_5_9
#define init_7_9_5_9
void FASTCALL call_7_9_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_5_10
#define init_7_9_5_10
void FASTCALL call_7_9_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_5_11
#define init_7_9_5_11
void FASTCALL call_7_9_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_5_12
#define init_7_9_5_12
void FASTCALL call_7_9_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_5_13
#define init_7_9_5_13
void FASTCALL call_7_9_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_5_14
#define init_7_9_5_14
void FASTCALL call_7_9_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_5_15
#define init_7_9_5_15
void FASTCALL call_7_9_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_6_0
#define init_7_9_6_0
void FASTCALL call_7_9_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_6_1
#define init_7_9_6_1
void FASTCALL call_7_9_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_6_2
#define init_7_9_6_2
void FASTCALL call_7_9_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_6_3
#define init_7_9_6_3
void FASTCALL call_7_9_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_6_4
#define init_7_9_6_4
void FASTCALL call_7_9_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_6_5
#define init_7_9_6_5
void FASTCALL call_7_9_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_6_6
#define init_7_9_6_6
void FASTCALL call_7_9_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_6_7
#define init_7_9_6_7
void FASTCALL call_7_9_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_6_8
#define init_7_9_6_8
void FASTCALL call_7_9_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_6_9
#define init_7_9_6_9
void FASTCALL call_7_9_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_6_10
#define init_7_9_6_10
void FASTCALL call_7_9_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_6_11
#define init_7_9_6_11
void FASTCALL call_7_9_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_6_12
#define init_7_9_6_12
void FASTCALL call_7_9_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_6_13
#define init_7_9_6_13
void FASTCALL call_7_9_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_6_14
#define init_7_9_6_14
void FASTCALL call_7_9_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_6_15
#define init_7_9_6_15
void FASTCALL call_7_9_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_7_0
#define init_7_9_7_0
void FASTCALL call_7_9_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_7_1
#define init_7_9_7_1
void FASTCALL call_7_9_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_7_2
#define init_7_9_7_2
void FASTCALL call_7_9_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_7_3
#define init_7_9_7_3
void FASTCALL call_7_9_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_7_4
#define init_7_9_7_4
void FASTCALL call_7_9_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_7_5
#define init_7_9_7_5
void FASTCALL call_7_9_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_7_6
#define init_7_9_7_6
void FASTCALL call_7_9_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_7_7
#define init_7_9_7_7
void FASTCALL call_7_9_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_7_8
#define init_7_9_7_8
void FASTCALL call_7_9_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_7_9
#define init_7_9_7_9
void FASTCALL call_7_9_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_7_10
#define init_7_9_7_10
void FASTCALL call_7_9_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_7_11
#define init_7_9_7_11
void FASTCALL call_7_9_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_7_12
#define init_7_9_7_12
void FASTCALL call_7_9_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_7_13
#define init_7_9_7_13
void FASTCALL call_7_9_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_7_14
#define init_7_9_7_14
void FASTCALL call_7_9_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_7_15
#define init_7_9_7_15
void FASTCALL call_7_9_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_8_0
#define init_7_9_8_0
void FASTCALL call_7_9_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_8_1
#define init_7_9_8_1
void FASTCALL call_7_9_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_8_2
#define init_7_9_8_2
void FASTCALL call_7_9_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_8_3
#define init_7_9_8_3
void FASTCALL call_7_9_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_8_4
#define init_7_9_8_4
void FASTCALL call_7_9_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_8_5
#define init_7_9_8_5
void FASTCALL call_7_9_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_8_6
#define init_7_9_8_6
void FASTCALL call_7_9_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_8_7
#define init_7_9_8_7
void FASTCALL call_7_9_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_8_8
#define init_7_9_8_8
void FASTCALL call_7_9_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_8_9
#define init_7_9_8_9
void FASTCALL call_7_9_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_8_10
#define init_7_9_8_10
void FASTCALL call_7_9_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_8_11
#define init_7_9_8_11
void FASTCALL call_7_9_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_8_12
#define init_7_9_8_12
void FASTCALL call_7_9_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_8_13
#define init_7_9_8_13
void FASTCALL call_7_9_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_8_14
#define init_7_9_8_14
void FASTCALL call_7_9_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_8_15
#define init_7_9_8_15
void FASTCALL call_7_9_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_9_0
#define init_7_9_9_0
void FASTCALL call_7_9_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_9_1
#define init_7_9_9_1
void FASTCALL call_7_9_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_9_2
#define init_7_9_9_2
void FASTCALL call_7_9_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_9_3
#define init_7_9_9_3
void FASTCALL call_7_9_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_9_4
#define init_7_9_9_4
void FASTCALL call_7_9_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_9_5
#define init_7_9_9_5
void FASTCALL call_7_9_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_9_6
#define init_7_9_9_6
void FASTCALL call_7_9_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_9_7
#define init_7_9_9_7
void FASTCALL call_7_9_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_9_8
#define init_7_9_9_8
void FASTCALL call_7_9_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_9_9
#define init_7_9_9_9
void FASTCALL call_7_9_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_9_10
#define init_7_9_9_10
void FASTCALL call_7_9_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_9_11
#define init_7_9_9_11
void FASTCALL call_7_9_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_9_12
#define init_7_9_9_12
void FASTCALL call_7_9_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_9_13
#define init_7_9_9_13
void FASTCALL call_7_9_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_9_14
#define init_7_9_9_14
void FASTCALL call_7_9_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_9_15
#define init_7_9_9_15
void FASTCALL call_7_9_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_10_0
#define init_7_9_10_0
void FASTCALL call_7_9_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_10_1
#define init_7_9_10_1
void FASTCALL call_7_9_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_10_2
#define init_7_9_10_2
void FASTCALL call_7_9_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_10_3
#define init_7_9_10_3
void FASTCALL call_7_9_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_10_4
#define init_7_9_10_4
void FASTCALL call_7_9_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_10_5
#define init_7_9_10_5
void FASTCALL call_7_9_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_10_6
#define init_7_9_10_6
void FASTCALL call_7_9_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_10_7
#define init_7_9_10_7
void FASTCALL call_7_9_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_10_8
#define init_7_9_10_8
void FASTCALL call_7_9_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_10_9
#define init_7_9_10_9
void FASTCALL call_7_9_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_10_10
#define init_7_9_10_10
void FASTCALL call_7_9_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_10_11
#define init_7_9_10_11
void FASTCALL call_7_9_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_10_12
#define init_7_9_10_12
void FASTCALL call_7_9_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_10_13
#define init_7_9_10_13
void FASTCALL call_7_9_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_10_14
#define init_7_9_10_14
void FASTCALL call_7_9_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_10_15
#define init_7_9_10_15
void FASTCALL call_7_9_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_11_0
#define init_7_9_11_0
void FASTCALL call_7_9_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_11_1
#define init_7_9_11_1
void FASTCALL call_7_9_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_11_2
#define init_7_9_11_2
void FASTCALL call_7_9_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_11_3
#define init_7_9_11_3
void FASTCALL call_7_9_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_11_4
#define init_7_9_11_4
void FASTCALL call_7_9_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_11_5
#define init_7_9_11_5
void FASTCALL call_7_9_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_11_6
#define init_7_9_11_6
void FASTCALL call_7_9_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_11_7
#define init_7_9_11_7
void FASTCALL call_7_9_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_11_8
#define init_7_9_11_8
void FASTCALL call_7_9_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_11_9
#define init_7_9_11_9
void FASTCALL call_7_9_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_11_10
#define init_7_9_11_10
void FASTCALL call_7_9_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_11_11
#define init_7_9_11_11
void FASTCALL call_7_9_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_11_12
#define init_7_9_11_12
void FASTCALL call_7_9_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_11_13
#define init_7_9_11_13
void FASTCALL call_7_9_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_11_14
#define init_7_9_11_14
void FASTCALL call_7_9_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_11_15
#define init_7_9_11_15
void FASTCALL call_7_9_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_12_0
#define init_7_9_12_0
void FASTCALL call_7_9_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_12_1
#define init_7_9_12_1
void FASTCALL call_7_9_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_12_2
#define init_7_9_12_2
void FASTCALL call_7_9_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_12_3
#define init_7_9_12_3
void FASTCALL call_7_9_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_12_4
#define init_7_9_12_4
void FASTCALL call_7_9_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_12_5
#define init_7_9_12_5
void FASTCALL call_7_9_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_12_6
#define init_7_9_12_6
void FASTCALL call_7_9_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_12_7
#define init_7_9_12_7
void FASTCALL call_7_9_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_12_8
#define init_7_9_12_8
void FASTCALL call_7_9_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_12_9
#define init_7_9_12_9
void FASTCALL call_7_9_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_12_10
#define init_7_9_12_10
void FASTCALL call_7_9_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_12_11
#define init_7_9_12_11
void FASTCALL call_7_9_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_12_12
#define init_7_9_12_12
void FASTCALL call_7_9_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_12_13
#define init_7_9_12_13
void FASTCALL call_7_9_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_12_14
#define init_7_9_12_14
void FASTCALL call_7_9_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_12_15
#define init_7_9_12_15
void FASTCALL call_7_9_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_13_0
#define init_7_9_13_0
void FASTCALL call_7_9_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_13_1
#define init_7_9_13_1
void FASTCALL call_7_9_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_13_2
#define init_7_9_13_2
void FASTCALL call_7_9_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_13_3
#define init_7_9_13_3
void FASTCALL call_7_9_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_13_4
#define init_7_9_13_4
void FASTCALL call_7_9_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_13_5
#define init_7_9_13_5
void FASTCALL call_7_9_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_13_6
#define init_7_9_13_6
void FASTCALL call_7_9_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_13_7
#define init_7_9_13_7
void FASTCALL call_7_9_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_13_8
#define init_7_9_13_8
void FASTCALL call_7_9_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_13_9
#define init_7_9_13_9
void FASTCALL call_7_9_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_13_10
#define init_7_9_13_10
void FASTCALL call_7_9_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_13_11
#define init_7_9_13_11
void FASTCALL call_7_9_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_13_12
#define init_7_9_13_12
void FASTCALL call_7_9_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_13_13
#define init_7_9_13_13
void FASTCALL call_7_9_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_13_14
#define init_7_9_13_14
void FASTCALL call_7_9_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_13_15
#define init_7_9_13_15
void FASTCALL call_7_9_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_14_0
#define init_7_9_14_0
void FASTCALL call_7_9_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_14_1
#define init_7_9_14_1
void FASTCALL call_7_9_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_14_2
#define init_7_9_14_2
void FASTCALL call_7_9_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_14_3
#define init_7_9_14_3
void FASTCALL call_7_9_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_14_4
#define init_7_9_14_4
void FASTCALL call_7_9_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_14_5
#define init_7_9_14_5
void FASTCALL call_7_9_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_14_6
#define init_7_9_14_6
void FASTCALL call_7_9_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_14_7
#define init_7_9_14_7
void FASTCALL call_7_9_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_14_8
#define init_7_9_14_8
void FASTCALL call_7_9_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_14_9
#define init_7_9_14_9
void FASTCALL call_7_9_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_14_10
#define init_7_9_14_10
void FASTCALL call_7_9_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_14_11
#define init_7_9_14_11
void FASTCALL call_7_9_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_14_12
#define init_7_9_14_12
void FASTCALL call_7_9_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_14_13
#define init_7_9_14_13
void FASTCALL call_7_9_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_14_14
#define init_7_9_14_14
void FASTCALL call_7_9_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_14_15
#define init_7_9_14_15
void FASTCALL call_7_9_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_15_0
#define init_7_9_15_0
void FASTCALL call_7_9_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_15_1
#define init_7_9_15_1
void FASTCALL call_7_9_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_15_2
#define init_7_9_15_2
void FASTCALL call_7_9_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_15_3
#define init_7_9_15_3
void FASTCALL call_7_9_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_15_4
#define init_7_9_15_4
void FASTCALL call_7_9_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_15_5
#define init_7_9_15_5
void FASTCALL call_7_9_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_15_6
#define init_7_9_15_6
void FASTCALL call_7_9_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_15_7
#define init_7_9_15_7
void FASTCALL call_7_9_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_15_8
#define init_7_9_15_8
void FASTCALL call_7_9_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_15_9
#define init_7_9_15_9
void FASTCALL call_7_9_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_15_10
#define init_7_9_15_10
void FASTCALL call_7_9_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_15_11
#define init_7_9_15_11
void FASTCALL call_7_9_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_15_12
#define init_7_9_15_12
void FASTCALL call_7_9_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_15_13
#define init_7_9_15_13
void FASTCALL call_7_9_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_15_14
#define init_7_9_15_14
void FASTCALL call_7_9_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_9_15_15
#define init_7_9_15_15
void FASTCALL call_7_9_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_0_0
#define init_7_10_0_0
void FASTCALL call_7_10_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_0_1
#define init_7_10_0_1
void FASTCALL call_7_10_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_0_2
#define init_7_10_0_2
void FASTCALL call_7_10_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_0_3
#define init_7_10_0_3
void FASTCALL call_7_10_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_0_4
#define init_7_10_0_4
void FASTCALL call_7_10_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_0_5
#define init_7_10_0_5
void FASTCALL call_7_10_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_0_6
#define init_7_10_0_6
void FASTCALL call_7_10_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_0_7
#define init_7_10_0_7
void FASTCALL call_7_10_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_0_8
#define init_7_10_0_8
void FASTCALL call_7_10_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_0_9
#define init_7_10_0_9
void FASTCALL call_7_10_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_0_10
#define init_7_10_0_10
void FASTCALL call_7_10_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_0_11
#define init_7_10_0_11
void FASTCALL call_7_10_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_0_12
#define init_7_10_0_12
void FASTCALL call_7_10_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_0_13
#define init_7_10_0_13
void FASTCALL call_7_10_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_0_14
#define init_7_10_0_14
void FASTCALL call_7_10_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_0_15
#define init_7_10_0_15
void FASTCALL call_7_10_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_1_0
#define init_7_10_1_0
void FASTCALL call_7_10_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_1_1
#define init_7_10_1_1
void FASTCALL call_7_10_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_1_2
#define init_7_10_1_2
void FASTCALL call_7_10_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_1_3
#define init_7_10_1_3
void FASTCALL call_7_10_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_1_4
#define init_7_10_1_4
void FASTCALL call_7_10_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_1_5
#define init_7_10_1_5
void FASTCALL call_7_10_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_1_6
#define init_7_10_1_6
void FASTCALL call_7_10_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_1_7
#define init_7_10_1_7
void FASTCALL call_7_10_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_1_8
#define init_7_10_1_8
void FASTCALL call_7_10_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_1_9
#define init_7_10_1_9
void FASTCALL call_7_10_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_1_10
#define init_7_10_1_10
void FASTCALL call_7_10_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_1_11
#define init_7_10_1_11
void FASTCALL call_7_10_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_1_12
#define init_7_10_1_12
void FASTCALL call_7_10_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_1_13
#define init_7_10_1_13
void FASTCALL call_7_10_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_1_14
#define init_7_10_1_14
void FASTCALL call_7_10_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_1_15
#define init_7_10_1_15
void FASTCALL call_7_10_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_2_0
#define init_7_10_2_0
void FASTCALL call_7_10_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_2_1
#define init_7_10_2_1
void FASTCALL call_7_10_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_2_2
#define init_7_10_2_2
void FASTCALL call_7_10_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_2_3
#define init_7_10_2_3
void FASTCALL call_7_10_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_2_4
#define init_7_10_2_4
void FASTCALL call_7_10_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_2_5
#define init_7_10_2_5
void FASTCALL call_7_10_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_2_6
#define init_7_10_2_6
void FASTCALL call_7_10_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_2_7
#define init_7_10_2_7
void FASTCALL call_7_10_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_2_8
#define init_7_10_2_8
void FASTCALL call_7_10_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_2_9
#define init_7_10_2_9
void FASTCALL call_7_10_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_2_10
#define init_7_10_2_10
void FASTCALL call_7_10_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_2_11
#define init_7_10_2_11
void FASTCALL call_7_10_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_2_12
#define init_7_10_2_12
void FASTCALL call_7_10_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_2_13
#define init_7_10_2_13
void FASTCALL call_7_10_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_2_14
#define init_7_10_2_14
void FASTCALL call_7_10_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_2_15
#define init_7_10_2_15
void FASTCALL call_7_10_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_3_0
#define init_7_10_3_0
void FASTCALL call_7_10_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_3_1
#define init_7_10_3_1
void FASTCALL call_7_10_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_3_2
#define init_7_10_3_2
void FASTCALL call_7_10_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_3_3
#define init_7_10_3_3
void FASTCALL call_7_10_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_3_4
#define init_7_10_3_4
void FASTCALL call_7_10_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_3_5
#define init_7_10_3_5
void FASTCALL call_7_10_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_3_6
#define init_7_10_3_6
void FASTCALL call_7_10_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_3_7
#define init_7_10_3_7
void FASTCALL call_7_10_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_3_8
#define init_7_10_3_8
void FASTCALL call_7_10_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_3_9
#define init_7_10_3_9
void FASTCALL call_7_10_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_3_10
#define init_7_10_3_10
void FASTCALL call_7_10_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_3_11
#define init_7_10_3_11
void FASTCALL call_7_10_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_3_12
#define init_7_10_3_12
void FASTCALL call_7_10_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_3_13
#define init_7_10_3_13
void FASTCALL call_7_10_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_3_14
#define init_7_10_3_14
void FASTCALL call_7_10_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_3_15
#define init_7_10_3_15
void FASTCALL call_7_10_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_4_0
#define init_7_10_4_0
void FASTCALL call_7_10_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_4_1
#define init_7_10_4_1
void FASTCALL call_7_10_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_4_2
#define init_7_10_4_2
void FASTCALL call_7_10_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_4_3
#define init_7_10_4_3
void FASTCALL call_7_10_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_4_4
#define init_7_10_4_4
void FASTCALL call_7_10_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_4_5
#define init_7_10_4_5
void FASTCALL call_7_10_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_4_6
#define init_7_10_4_6
void FASTCALL call_7_10_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_4_7
#define init_7_10_4_7
void FASTCALL call_7_10_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_4_8
#define init_7_10_4_8
void FASTCALL call_7_10_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_4_9
#define init_7_10_4_9
void FASTCALL call_7_10_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_4_10
#define init_7_10_4_10
void FASTCALL call_7_10_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_4_11
#define init_7_10_4_11
void FASTCALL call_7_10_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_4_12
#define init_7_10_4_12
void FASTCALL call_7_10_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_4_13
#define init_7_10_4_13
void FASTCALL call_7_10_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_4_14
#define init_7_10_4_14
void FASTCALL call_7_10_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_4_15
#define init_7_10_4_15
void FASTCALL call_7_10_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_5_0
#define init_7_10_5_0
void FASTCALL call_7_10_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_5_1
#define init_7_10_5_1
void FASTCALL call_7_10_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_5_2
#define init_7_10_5_2
void FASTCALL call_7_10_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_5_3
#define init_7_10_5_3
void FASTCALL call_7_10_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_5_4
#define init_7_10_5_4
void FASTCALL call_7_10_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_5_5
#define init_7_10_5_5
void FASTCALL call_7_10_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_5_6
#define init_7_10_5_6
void FASTCALL call_7_10_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_5_7
#define init_7_10_5_7
void FASTCALL call_7_10_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_5_8
#define init_7_10_5_8
void FASTCALL call_7_10_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_5_9
#define init_7_10_5_9
void FASTCALL call_7_10_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_5_10
#define init_7_10_5_10
void FASTCALL call_7_10_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_5_11
#define init_7_10_5_11
void FASTCALL call_7_10_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_5_12
#define init_7_10_5_12
void FASTCALL call_7_10_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_5_13
#define init_7_10_5_13
void FASTCALL call_7_10_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_5_14
#define init_7_10_5_14
void FASTCALL call_7_10_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_5_15
#define init_7_10_5_15
void FASTCALL call_7_10_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_6_0
#define init_7_10_6_0
void FASTCALL call_7_10_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_6_1
#define init_7_10_6_1
void FASTCALL call_7_10_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_6_2
#define init_7_10_6_2
void FASTCALL call_7_10_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_6_3
#define init_7_10_6_3
void FASTCALL call_7_10_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_6_4
#define init_7_10_6_4
void FASTCALL call_7_10_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_6_5
#define init_7_10_6_5
void FASTCALL call_7_10_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_6_6
#define init_7_10_6_6
void FASTCALL call_7_10_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_6_7
#define init_7_10_6_7
void FASTCALL call_7_10_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_6_8
#define init_7_10_6_8
void FASTCALL call_7_10_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_6_9
#define init_7_10_6_9
void FASTCALL call_7_10_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_6_10
#define init_7_10_6_10
void FASTCALL call_7_10_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_6_11
#define init_7_10_6_11
void FASTCALL call_7_10_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_6_12
#define init_7_10_6_12
void FASTCALL call_7_10_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_6_13
#define init_7_10_6_13
void FASTCALL call_7_10_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_6_14
#define init_7_10_6_14
void FASTCALL call_7_10_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_6_15
#define init_7_10_6_15
void FASTCALL call_7_10_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_7_0
#define init_7_10_7_0
void FASTCALL call_7_10_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_7_1
#define init_7_10_7_1
void FASTCALL call_7_10_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_7_2
#define init_7_10_7_2
void FASTCALL call_7_10_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_7_3
#define init_7_10_7_3
void FASTCALL call_7_10_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_7_4
#define init_7_10_7_4
void FASTCALL call_7_10_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_7_5
#define init_7_10_7_5
void FASTCALL call_7_10_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_7_6
#define init_7_10_7_6
void FASTCALL call_7_10_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_7_7
#define init_7_10_7_7
void FASTCALL call_7_10_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_7_8
#define init_7_10_7_8
void FASTCALL call_7_10_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_7_9
#define init_7_10_7_9
void FASTCALL call_7_10_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_7_10
#define init_7_10_7_10
void FASTCALL call_7_10_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_7_11
#define init_7_10_7_11
void FASTCALL call_7_10_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_7_12
#define init_7_10_7_12
void FASTCALL call_7_10_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_7_13
#define init_7_10_7_13
void FASTCALL call_7_10_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_7_14
#define init_7_10_7_14
void FASTCALL call_7_10_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_7_15
#define init_7_10_7_15
void FASTCALL call_7_10_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_8_0
#define init_7_10_8_0
void FASTCALL call_7_10_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_8_1
#define init_7_10_8_1
void FASTCALL call_7_10_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_8_2
#define init_7_10_8_2
void FASTCALL call_7_10_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_8_3
#define init_7_10_8_3
void FASTCALL call_7_10_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_8_4
#define init_7_10_8_4
void FASTCALL call_7_10_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_8_5
#define init_7_10_8_5
void FASTCALL call_7_10_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_8_6
#define init_7_10_8_6
void FASTCALL call_7_10_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_8_7
#define init_7_10_8_7
void FASTCALL call_7_10_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_8_8
#define init_7_10_8_8
void FASTCALL call_7_10_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_8_9
#define init_7_10_8_9
void FASTCALL call_7_10_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_8_10
#define init_7_10_8_10
void FASTCALL call_7_10_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_8_11
#define init_7_10_8_11
void FASTCALL call_7_10_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_8_12
#define init_7_10_8_12
void FASTCALL call_7_10_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_8_13
#define init_7_10_8_13
void FASTCALL call_7_10_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_8_14
#define init_7_10_8_14
void FASTCALL call_7_10_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_8_15
#define init_7_10_8_15
void FASTCALL call_7_10_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_9_0
#define init_7_10_9_0
void FASTCALL call_7_10_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_9_1
#define init_7_10_9_1
void FASTCALL call_7_10_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_9_2
#define init_7_10_9_2
void FASTCALL call_7_10_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_9_3
#define init_7_10_9_3
void FASTCALL call_7_10_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_9_4
#define init_7_10_9_4
void FASTCALL call_7_10_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_9_5
#define init_7_10_9_5
void FASTCALL call_7_10_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_9_6
#define init_7_10_9_6
void FASTCALL call_7_10_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_9_7
#define init_7_10_9_7
void FASTCALL call_7_10_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_9_8
#define init_7_10_9_8
void FASTCALL call_7_10_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_9_9
#define init_7_10_9_9
void FASTCALL call_7_10_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_9_10
#define init_7_10_9_10
void FASTCALL call_7_10_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_9_11
#define init_7_10_9_11
void FASTCALL call_7_10_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_9_12
#define init_7_10_9_12
void FASTCALL call_7_10_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_9_13
#define init_7_10_9_13
void FASTCALL call_7_10_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_9_14
#define init_7_10_9_14
void FASTCALL call_7_10_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_9_15
#define init_7_10_9_15
void FASTCALL call_7_10_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_10_0
#define init_7_10_10_0
void FASTCALL call_7_10_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_10_1
#define init_7_10_10_1
void FASTCALL call_7_10_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_10_2
#define init_7_10_10_2
void FASTCALL call_7_10_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_10_3
#define init_7_10_10_3
void FASTCALL call_7_10_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_10_4
#define init_7_10_10_4
void FASTCALL call_7_10_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_10_5
#define init_7_10_10_5
void FASTCALL call_7_10_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_10_6
#define init_7_10_10_6
void FASTCALL call_7_10_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_10_7
#define init_7_10_10_7
void FASTCALL call_7_10_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_10_8
#define init_7_10_10_8
void FASTCALL call_7_10_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_10_9
#define init_7_10_10_9
void FASTCALL call_7_10_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_10_10
#define init_7_10_10_10
void FASTCALL call_7_10_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_10_11
#define init_7_10_10_11
void FASTCALL call_7_10_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_10_12
#define init_7_10_10_12
void FASTCALL call_7_10_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_10_13
#define init_7_10_10_13
void FASTCALL call_7_10_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_10_14
#define init_7_10_10_14
void FASTCALL call_7_10_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_10_15
#define init_7_10_10_15
void FASTCALL call_7_10_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_11_0
#define init_7_10_11_0
void FASTCALL call_7_10_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_11_1
#define init_7_10_11_1
void FASTCALL call_7_10_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_11_2
#define init_7_10_11_2
void FASTCALL call_7_10_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_11_3
#define init_7_10_11_3
void FASTCALL call_7_10_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_11_4
#define init_7_10_11_4
void FASTCALL call_7_10_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_11_5
#define init_7_10_11_5
void FASTCALL call_7_10_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_11_6
#define init_7_10_11_6
void FASTCALL call_7_10_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_11_7
#define init_7_10_11_7
void FASTCALL call_7_10_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_11_8
#define init_7_10_11_8
void FASTCALL call_7_10_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_11_9
#define init_7_10_11_9
void FASTCALL call_7_10_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_11_10
#define init_7_10_11_10
void FASTCALL call_7_10_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_11_11
#define init_7_10_11_11
void FASTCALL call_7_10_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_11_12
#define init_7_10_11_12
void FASTCALL call_7_10_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_11_13
#define init_7_10_11_13
void FASTCALL call_7_10_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_11_14
#define init_7_10_11_14
void FASTCALL call_7_10_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_11_15
#define init_7_10_11_15
void FASTCALL call_7_10_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_12_0
#define init_7_10_12_0
void FASTCALL call_7_10_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_12_1
#define init_7_10_12_1
void FASTCALL call_7_10_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_12_2
#define init_7_10_12_2
void FASTCALL call_7_10_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_12_3
#define init_7_10_12_3
void FASTCALL call_7_10_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_12_4
#define init_7_10_12_4
void FASTCALL call_7_10_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_12_5
#define init_7_10_12_5
void FASTCALL call_7_10_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_12_6
#define init_7_10_12_6
void FASTCALL call_7_10_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_12_7
#define init_7_10_12_7
void FASTCALL call_7_10_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_12_8
#define init_7_10_12_8
void FASTCALL call_7_10_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_12_9
#define init_7_10_12_9
void FASTCALL call_7_10_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_12_10
#define init_7_10_12_10
void FASTCALL call_7_10_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_12_11
#define init_7_10_12_11
void FASTCALL call_7_10_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_12_12
#define init_7_10_12_12
void FASTCALL call_7_10_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_12_13
#define init_7_10_12_13
void FASTCALL call_7_10_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_12_14
#define init_7_10_12_14
void FASTCALL call_7_10_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_12_15
#define init_7_10_12_15
void FASTCALL call_7_10_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_13_0
#define init_7_10_13_0
void FASTCALL call_7_10_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_13_1
#define init_7_10_13_1
void FASTCALL call_7_10_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_13_2
#define init_7_10_13_2
void FASTCALL call_7_10_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_13_3
#define init_7_10_13_3
void FASTCALL call_7_10_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_13_4
#define init_7_10_13_4
void FASTCALL call_7_10_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_13_5
#define init_7_10_13_5
void FASTCALL call_7_10_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_13_6
#define init_7_10_13_6
void FASTCALL call_7_10_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_13_7
#define init_7_10_13_7
void FASTCALL call_7_10_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_13_8
#define init_7_10_13_8
void FASTCALL call_7_10_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_13_9
#define init_7_10_13_9
void FASTCALL call_7_10_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_13_10
#define init_7_10_13_10
void FASTCALL call_7_10_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_13_11
#define init_7_10_13_11
void FASTCALL call_7_10_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_13_12
#define init_7_10_13_12
void FASTCALL call_7_10_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_13_13
#define init_7_10_13_13
void FASTCALL call_7_10_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_13_14
#define init_7_10_13_14
void FASTCALL call_7_10_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_13_15
#define init_7_10_13_15
void FASTCALL call_7_10_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_14_0
#define init_7_10_14_0
void FASTCALL call_7_10_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_14_1
#define init_7_10_14_1
void FASTCALL call_7_10_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_14_2
#define init_7_10_14_2
void FASTCALL call_7_10_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_14_3
#define init_7_10_14_3
void FASTCALL call_7_10_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_14_4
#define init_7_10_14_4
void FASTCALL call_7_10_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_14_5
#define init_7_10_14_5
void FASTCALL call_7_10_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_14_6
#define init_7_10_14_6
void FASTCALL call_7_10_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_14_7
#define init_7_10_14_7
void FASTCALL call_7_10_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_14_8
#define init_7_10_14_8
void FASTCALL call_7_10_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_14_9
#define init_7_10_14_9
void FASTCALL call_7_10_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_14_10
#define init_7_10_14_10
void FASTCALL call_7_10_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_14_11
#define init_7_10_14_11
void FASTCALL call_7_10_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_14_12
#define init_7_10_14_12
void FASTCALL call_7_10_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_14_13
#define init_7_10_14_13
void FASTCALL call_7_10_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_14_14
#define init_7_10_14_14
void FASTCALL call_7_10_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_14_15
#define init_7_10_14_15
void FASTCALL call_7_10_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_15_0
#define init_7_10_15_0
void FASTCALL call_7_10_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_15_1
#define init_7_10_15_1
void FASTCALL call_7_10_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_15_2
#define init_7_10_15_2
void FASTCALL call_7_10_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_15_3
#define init_7_10_15_3
void FASTCALL call_7_10_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_15_4
#define init_7_10_15_4
void FASTCALL call_7_10_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_15_5
#define init_7_10_15_5
void FASTCALL call_7_10_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_15_6
#define init_7_10_15_6
void FASTCALL call_7_10_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_15_7
#define init_7_10_15_7
void FASTCALL call_7_10_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_15_8
#define init_7_10_15_8
void FASTCALL call_7_10_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_15_9
#define init_7_10_15_9
void FASTCALL call_7_10_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_15_10
#define init_7_10_15_10
void FASTCALL call_7_10_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_15_11
#define init_7_10_15_11
void FASTCALL call_7_10_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_15_12
#define init_7_10_15_12
void FASTCALL call_7_10_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_15_13
#define init_7_10_15_13
void FASTCALL call_7_10_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_15_14
#define init_7_10_15_14
void FASTCALL call_7_10_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_10_15_15
#define init_7_10_15_15
void FASTCALL call_7_10_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_0_0
#define init_7_11_0_0
void FASTCALL call_7_11_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_0_1
#define init_7_11_0_1
void FASTCALL call_7_11_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_0_2
#define init_7_11_0_2
void FASTCALL call_7_11_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_0_3
#define init_7_11_0_3
void FASTCALL call_7_11_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_0_4
#define init_7_11_0_4
void FASTCALL call_7_11_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_0_5
#define init_7_11_0_5
void FASTCALL call_7_11_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_0_6
#define init_7_11_0_6
void FASTCALL call_7_11_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_0_7
#define init_7_11_0_7
void FASTCALL call_7_11_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_0_8
#define init_7_11_0_8
void FASTCALL call_7_11_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_0_9
#define init_7_11_0_9
void FASTCALL call_7_11_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_0_10
#define init_7_11_0_10
void FASTCALL call_7_11_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_0_11
#define init_7_11_0_11
void FASTCALL call_7_11_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_0_12
#define init_7_11_0_12
void FASTCALL call_7_11_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_0_13
#define init_7_11_0_13
void FASTCALL call_7_11_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_0_14
#define init_7_11_0_14
void FASTCALL call_7_11_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_0_15
#define init_7_11_0_15
void FASTCALL call_7_11_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_1_0
#define init_7_11_1_0
void FASTCALL call_7_11_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_1_1
#define init_7_11_1_1
void FASTCALL call_7_11_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_1_2
#define init_7_11_1_2
void FASTCALL call_7_11_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_1_3
#define init_7_11_1_3
void FASTCALL call_7_11_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_1_4
#define init_7_11_1_4
void FASTCALL call_7_11_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_1_5
#define init_7_11_1_5
void FASTCALL call_7_11_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_1_6
#define init_7_11_1_6
void FASTCALL call_7_11_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_1_7
#define init_7_11_1_7
void FASTCALL call_7_11_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_1_8
#define init_7_11_1_8
void FASTCALL call_7_11_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_1_9
#define init_7_11_1_9
void FASTCALL call_7_11_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_1_10
#define init_7_11_1_10
void FASTCALL call_7_11_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_1_11
#define init_7_11_1_11
void FASTCALL call_7_11_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_1_12
#define init_7_11_1_12
void FASTCALL call_7_11_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_1_13
#define init_7_11_1_13
void FASTCALL call_7_11_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_1_14
#define init_7_11_1_14
void FASTCALL call_7_11_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_1_15
#define init_7_11_1_15
void FASTCALL call_7_11_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_2_0
#define init_7_11_2_0
void FASTCALL call_7_11_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_2_1
#define init_7_11_2_1
void FASTCALL call_7_11_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_2_2
#define init_7_11_2_2
void FASTCALL call_7_11_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_2_3
#define init_7_11_2_3
void FASTCALL call_7_11_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_2_4
#define init_7_11_2_4
void FASTCALL call_7_11_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_2_5
#define init_7_11_2_5
void FASTCALL call_7_11_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_2_6
#define init_7_11_2_6
void FASTCALL call_7_11_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_2_7
#define init_7_11_2_7
void FASTCALL call_7_11_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_2_8
#define init_7_11_2_8
void FASTCALL call_7_11_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_2_9
#define init_7_11_2_9
void FASTCALL call_7_11_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_2_10
#define init_7_11_2_10
void FASTCALL call_7_11_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_2_11
#define init_7_11_2_11
void FASTCALL call_7_11_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_2_12
#define init_7_11_2_12
void FASTCALL call_7_11_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_2_13
#define init_7_11_2_13
void FASTCALL call_7_11_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_2_14
#define init_7_11_2_14
void FASTCALL call_7_11_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_2_15
#define init_7_11_2_15
void FASTCALL call_7_11_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_3_0
#define init_7_11_3_0
void FASTCALL call_7_11_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_3_1
#define init_7_11_3_1
void FASTCALL call_7_11_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_3_2
#define init_7_11_3_2
void FASTCALL call_7_11_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_3_3
#define init_7_11_3_3
void FASTCALL call_7_11_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_3_4
#define init_7_11_3_4
void FASTCALL call_7_11_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_3_5
#define init_7_11_3_5
void FASTCALL call_7_11_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_3_6
#define init_7_11_3_6
void FASTCALL call_7_11_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_3_7
#define init_7_11_3_7
void FASTCALL call_7_11_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_3_8
#define init_7_11_3_8
void FASTCALL call_7_11_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_3_9
#define init_7_11_3_9
void FASTCALL call_7_11_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_3_10
#define init_7_11_3_10
void FASTCALL call_7_11_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_3_11
#define init_7_11_3_11
void FASTCALL call_7_11_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_3_12
#define init_7_11_3_12
void FASTCALL call_7_11_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_3_13
#define init_7_11_3_13
void FASTCALL call_7_11_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_3_14
#define init_7_11_3_14
void FASTCALL call_7_11_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_3_15
#define init_7_11_3_15
void FASTCALL call_7_11_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_4_0
#define init_7_11_4_0
void FASTCALL call_7_11_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_4_1
#define init_7_11_4_1
void FASTCALL call_7_11_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_4_2
#define init_7_11_4_2
void FASTCALL call_7_11_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_4_3
#define init_7_11_4_3
void FASTCALL call_7_11_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_4_4
#define init_7_11_4_4
void FASTCALL call_7_11_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_4_5
#define init_7_11_4_5
void FASTCALL call_7_11_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_4_6
#define init_7_11_4_6
void FASTCALL call_7_11_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_4_7
#define init_7_11_4_7
void FASTCALL call_7_11_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_4_8
#define init_7_11_4_8
void FASTCALL call_7_11_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_4_9
#define init_7_11_4_9
void FASTCALL call_7_11_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_4_10
#define init_7_11_4_10
void FASTCALL call_7_11_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_4_11
#define init_7_11_4_11
void FASTCALL call_7_11_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_4_12
#define init_7_11_4_12
void FASTCALL call_7_11_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_4_13
#define init_7_11_4_13
void FASTCALL call_7_11_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_4_14
#define init_7_11_4_14
void FASTCALL call_7_11_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_4_15
#define init_7_11_4_15
void FASTCALL call_7_11_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_5_0
#define init_7_11_5_0
void FASTCALL call_7_11_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_5_1
#define init_7_11_5_1
void FASTCALL call_7_11_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_5_2
#define init_7_11_5_2
void FASTCALL call_7_11_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_5_3
#define init_7_11_5_3
void FASTCALL call_7_11_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_5_4
#define init_7_11_5_4
void FASTCALL call_7_11_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_5_5
#define init_7_11_5_5
void FASTCALL call_7_11_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_5_6
#define init_7_11_5_6
void FASTCALL call_7_11_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_5_7
#define init_7_11_5_7
void FASTCALL call_7_11_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_5_8
#define init_7_11_5_8
void FASTCALL call_7_11_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_5_9
#define init_7_11_5_9
void FASTCALL call_7_11_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_5_10
#define init_7_11_5_10
void FASTCALL call_7_11_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_5_11
#define init_7_11_5_11
void FASTCALL call_7_11_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_5_12
#define init_7_11_5_12
void FASTCALL call_7_11_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_5_13
#define init_7_11_5_13
void FASTCALL call_7_11_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_5_14
#define init_7_11_5_14
void FASTCALL call_7_11_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_5_15
#define init_7_11_5_15
void FASTCALL call_7_11_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_6_0
#define init_7_11_6_0
void FASTCALL call_7_11_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_6_1
#define init_7_11_6_1
void FASTCALL call_7_11_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_6_2
#define init_7_11_6_2
void FASTCALL call_7_11_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_6_3
#define init_7_11_6_3
void FASTCALL call_7_11_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_6_4
#define init_7_11_6_4
void FASTCALL call_7_11_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_6_5
#define init_7_11_6_5
void FASTCALL call_7_11_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_6_6
#define init_7_11_6_6
void FASTCALL call_7_11_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_6_7
#define init_7_11_6_7
void FASTCALL call_7_11_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_6_8
#define init_7_11_6_8
void FASTCALL call_7_11_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_6_9
#define init_7_11_6_9
void FASTCALL call_7_11_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_6_10
#define init_7_11_6_10
void FASTCALL call_7_11_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_6_11
#define init_7_11_6_11
void FASTCALL call_7_11_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_6_12
#define init_7_11_6_12
void FASTCALL call_7_11_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_6_13
#define init_7_11_6_13
void FASTCALL call_7_11_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_6_14
#define init_7_11_6_14
void FASTCALL call_7_11_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_6_15
#define init_7_11_6_15
void FASTCALL call_7_11_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_7_0
#define init_7_11_7_0
void FASTCALL call_7_11_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_7_1
#define init_7_11_7_1
void FASTCALL call_7_11_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_7_2
#define init_7_11_7_2
void FASTCALL call_7_11_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_7_3
#define init_7_11_7_3
void FASTCALL call_7_11_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_7_4
#define init_7_11_7_4
void FASTCALL call_7_11_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_7_5
#define init_7_11_7_5
void FASTCALL call_7_11_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_7_6
#define init_7_11_7_6
void FASTCALL call_7_11_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_7_7
#define init_7_11_7_7
void FASTCALL call_7_11_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_7_8
#define init_7_11_7_8
void FASTCALL call_7_11_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_7_9
#define init_7_11_7_9
void FASTCALL call_7_11_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_7_10
#define init_7_11_7_10
void FASTCALL call_7_11_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_7_11
#define init_7_11_7_11
void FASTCALL call_7_11_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_7_12
#define init_7_11_7_12
void FASTCALL call_7_11_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_7_13
#define init_7_11_7_13
void FASTCALL call_7_11_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_7_14
#define init_7_11_7_14
void FASTCALL call_7_11_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_7_15
#define init_7_11_7_15
void FASTCALL call_7_11_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_8_0
#define init_7_11_8_0
void FASTCALL call_7_11_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_8_1
#define init_7_11_8_1
void FASTCALL call_7_11_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_8_2
#define init_7_11_8_2
void FASTCALL call_7_11_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_8_3
#define init_7_11_8_3
void FASTCALL call_7_11_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_8_4
#define init_7_11_8_4
void FASTCALL call_7_11_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_8_5
#define init_7_11_8_5
void FASTCALL call_7_11_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_8_6
#define init_7_11_8_6
void FASTCALL call_7_11_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_8_7
#define init_7_11_8_7
void FASTCALL call_7_11_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_8_8
#define init_7_11_8_8
void FASTCALL call_7_11_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_8_9
#define init_7_11_8_9
void FASTCALL call_7_11_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_8_10
#define init_7_11_8_10
void FASTCALL call_7_11_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_8_11
#define init_7_11_8_11
void FASTCALL call_7_11_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_8_12
#define init_7_11_8_12
void FASTCALL call_7_11_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_8_13
#define init_7_11_8_13
void FASTCALL call_7_11_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_8_14
#define init_7_11_8_14
void FASTCALL call_7_11_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_8_15
#define init_7_11_8_15
void FASTCALL call_7_11_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_9_0
#define init_7_11_9_0
void FASTCALL call_7_11_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_9_1
#define init_7_11_9_1
void FASTCALL call_7_11_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_9_2
#define init_7_11_9_2
void FASTCALL call_7_11_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_9_3
#define init_7_11_9_3
void FASTCALL call_7_11_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_9_4
#define init_7_11_9_4
void FASTCALL call_7_11_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_9_5
#define init_7_11_9_5
void FASTCALL call_7_11_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_9_6
#define init_7_11_9_6
void FASTCALL call_7_11_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_9_7
#define init_7_11_9_7
void FASTCALL call_7_11_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_9_8
#define init_7_11_9_8
void FASTCALL call_7_11_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_9_9
#define init_7_11_9_9
void FASTCALL call_7_11_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_9_10
#define init_7_11_9_10
void FASTCALL call_7_11_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_9_11
#define init_7_11_9_11
void FASTCALL call_7_11_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_9_12
#define init_7_11_9_12
void FASTCALL call_7_11_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_9_13
#define init_7_11_9_13
void FASTCALL call_7_11_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_9_14
#define init_7_11_9_14
void FASTCALL call_7_11_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_9_15
#define init_7_11_9_15
void FASTCALL call_7_11_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_10_0
#define init_7_11_10_0
void FASTCALL call_7_11_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_10_1
#define init_7_11_10_1
void FASTCALL call_7_11_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_10_2
#define init_7_11_10_2
void FASTCALL call_7_11_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_10_3
#define init_7_11_10_3
void FASTCALL call_7_11_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_10_4
#define init_7_11_10_4
void FASTCALL call_7_11_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_10_5
#define init_7_11_10_5
void FASTCALL call_7_11_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_10_6
#define init_7_11_10_6
void FASTCALL call_7_11_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_10_7
#define init_7_11_10_7
void FASTCALL call_7_11_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_10_8
#define init_7_11_10_8
void FASTCALL call_7_11_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_10_9
#define init_7_11_10_9
void FASTCALL call_7_11_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_10_10
#define init_7_11_10_10
void FASTCALL call_7_11_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_10_11
#define init_7_11_10_11
void FASTCALL call_7_11_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_10_12
#define init_7_11_10_12
void FASTCALL call_7_11_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_10_13
#define init_7_11_10_13
void FASTCALL call_7_11_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_10_14
#define init_7_11_10_14
void FASTCALL call_7_11_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_10_15
#define init_7_11_10_15
void FASTCALL call_7_11_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_11_0
#define init_7_11_11_0
void FASTCALL call_7_11_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_11_1
#define init_7_11_11_1
void FASTCALL call_7_11_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_11_2
#define init_7_11_11_2
void FASTCALL call_7_11_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_11_3
#define init_7_11_11_3
void FASTCALL call_7_11_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_11_4
#define init_7_11_11_4
void FASTCALL call_7_11_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_11_5
#define init_7_11_11_5
void FASTCALL call_7_11_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_11_6
#define init_7_11_11_6
void FASTCALL call_7_11_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_11_7
#define init_7_11_11_7
void FASTCALL call_7_11_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_11_8
#define init_7_11_11_8
void FASTCALL call_7_11_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_11_9
#define init_7_11_11_9
void FASTCALL call_7_11_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_11_10
#define init_7_11_11_10
void FASTCALL call_7_11_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_11_11
#define init_7_11_11_11
void FASTCALL call_7_11_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_11_12
#define init_7_11_11_12
void FASTCALL call_7_11_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_11_13
#define init_7_11_11_13
void FASTCALL call_7_11_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_11_14
#define init_7_11_11_14
void FASTCALL call_7_11_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_11_15
#define init_7_11_11_15
void FASTCALL call_7_11_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_12_0
#define init_7_11_12_0
void FASTCALL call_7_11_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_12_1
#define init_7_11_12_1
void FASTCALL call_7_11_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_12_2
#define init_7_11_12_2
void FASTCALL call_7_11_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_12_3
#define init_7_11_12_3
void FASTCALL call_7_11_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_12_4
#define init_7_11_12_4
void FASTCALL call_7_11_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_12_5
#define init_7_11_12_5
void FASTCALL call_7_11_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_12_6
#define init_7_11_12_6
void FASTCALL call_7_11_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_12_7
#define init_7_11_12_7
void FASTCALL call_7_11_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_12_8
#define init_7_11_12_8
void FASTCALL call_7_11_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_12_9
#define init_7_11_12_9
void FASTCALL call_7_11_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_12_10
#define init_7_11_12_10
void FASTCALL call_7_11_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_12_11
#define init_7_11_12_11
void FASTCALL call_7_11_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_12_12
#define init_7_11_12_12
void FASTCALL call_7_11_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_12_13
#define init_7_11_12_13
void FASTCALL call_7_11_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_12_14
#define init_7_11_12_14
void FASTCALL call_7_11_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_12_15
#define init_7_11_12_15
void FASTCALL call_7_11_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_13_0
#define init_7_11_13_0
void FASTCALL call_7_11_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_13_1
#define init_7_11_13_1
void FASTCALL call_7_11_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_13_2
#define init_7_11_13_2
void FASTCALL call_7_11_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_13_3
#define init_7_11_13_3
void FASTCALL call_7_11_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_13_4
#define init_7_11_13_4
void FASTCALL call_7_11_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_13_5
#define init_7_11_13_5
void FASTCALL call_7_11_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_13_6
#define init_7_11_13_6
void FASTCALL call_7_11_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_13_7
#define init_7_11_13_7
void FASTCALL call_7_11_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_13_8
#define init_7_11_13_8
void FASTCALL call_7_11_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_13_9
#define init_7_11_13_9
void FASTCALL call_7_11_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_13_10
#define init_7_11_13_10
void FASTCALL call_7_11_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_13_11
#define init_7_11_13_11
void FASTCALL call_7_11_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_13_12
#define init_7_11_13_12
void FASTCALL call_7_11_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_13_13
#define init_7_11_13_13
void FASTCALL call_7_11_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_13_14
#define init_7_11_13_14
void FASTCALL call_7_11_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_13_15
#define init_7_11_13_15
void FASTCALL call_7_11_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_14_0
#define init_7_11_14_0
void FASTCALL call_7_11_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_14_1
#define init_7_11_14_1
void FASTCALL call_7_11_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_14_2
#define init_7_11_14_2
void FASTCALL call_7_11_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_14_3
#define init_7_11_14_3
void FASTCALL call_7_11_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_14_4
#define init_7_11_14_4
void FASTCALL call_7_11_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_14_5
#define init_7_11_14_5
void FASTCALL call_7_11_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_14_6
#define init_7_11_14_6
void FASTCALL call_7_11_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_14_7
#define init_7_11_14_7
void FASTCALL call_7_11_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_14_8
#define init_7_11_14_8
void FASTCALL call_7_11_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_14_9
#define init_7_11_14_9
void FASTCALL call_7_11_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_14_10
#define init_7_11_14_10
void FASTCALL call_7_11_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_14_11
#define init_7_11_14_11
void FASTCALL call_7_11_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_14_12
#define init_7_11_14_12
void FASTCALL call_7_11_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_14_13
#define init_7_11_14_13
void FASTCALL call_7_11_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_14_14
#define init_7_11_14_14
void FASTCALL call_7_11_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_14_15
#define init_7_11_14_15
void FASTCALL call_7_11_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_15_0
#define init_7_11_15_0
void FASTCALL call_7_11_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_15_1
#define init_7_11_15_1
void FASTCALL call_7_11_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_15_2
#define init_7_11_15_2
void FASTCALL call_7_11_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_15_3
#define init_7_11_15_3
void FASTCALL call_7_11_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_15_4
#define init_7_11_15_4
void FASTCALL call_7_11_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_15_5
#define init_7_11_15_5
void FASTCALL call_7_11_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_15_6
#define init_7_11_15_6
void FASTCALL call_7_11_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_15_7
#define init_7_11_15_7
void FASTCALL call_7_11_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_15_8
#define init_7_11_15_8
void FASTCALL call_7_11_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_15_9
#define init_7_11_15_9
void FASTCALL call_7_11_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_15_10
#define init_7_11_15_10
void FASTCALL call_7_11_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_15_11
#define init_7_11_15_11
void FASTCALL call_7_11_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_15_12
#define init_7_11_15_12
void FASTCALL call_7_11_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_15_13
#define init_7_11_15_13
void FASTCALL call_7_11_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_15_14
#define init_7_11_15_14
void FASTCALL call_7_11_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_11_15_15
#define init_7_11_15_15
void FASTCALL call_7_11_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_0_0
#define init_7_12_0_0
void FASTCALL call_7_12_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_0_1
#define init_7_12_0_1
void FASTCALL call_7_12_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_0_2
#define init_7_12_0_2
void FASTCALL call_7_12_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_0_3
#define init_7_12_0_3
void FASTCALL call_7_12_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_0_4
#define init_7_12_0_4
void FASTCALL call_7_12_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_0_5
#define init_7_12_0_5
void FASTCALL call_7_12_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_0_6
#define init_7_12_0_6
void FASTCALL call_7_12_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_0_7
#define init_7_12_0_7
void FASTCALL call_7_12_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_0_8
#define init_7_12_0_8
void FASTCALL call_7_12_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_0_9
#define init_7_12_0_9
void FASTCALL call_7_12_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_0_10
#define init_7_12_0_10
void FASTCALL call_7_12_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_0_11
#define init_7_12_0_11
void FASTCALL call_7_12_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_0_12
#define init_7_12_0_12
void FASTCALL call_7_12_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_0_13
#define init_7_12_0_13
void FASTCALL call_7_12_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_0_14
#define init_7_12_0_14
void FASTCALL call_7_12_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_0_15
#define init_7_12_0_15
void FASTCALL call_7_12_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_1_0
#define init_7_12_1_0
void FASTCALL call_7_12_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_1_1
#define init_7_12_1_1
void FASTCALL call_7_12_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_1_2
#define init_7_12_1_2
void FASTCALL call_7_12_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_1_3
#define init_7_12_1_3
void FASTCALL call_7_12_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_1_4
#define init_7_12_1_4
void FASTCALL call_7_12_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_1_5
#define init_7_12_1_5
void FASTCALL call_7_12_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_1_6
#define init_7_12_1_6
void FASTCALL call_7_12_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_1_7
#define init_7_12_1_7
void FASTCALL call_7_12_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_1_8
#define init_7_12_1_8
void FASTCALL call_7_12_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_1_9
#define init_7_12_1_9
void FASTCALL call_7_12_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_1_10
#define init_7_12_1_10
void FASTCALL call_7_12_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_1_11
#define init_7_12_1_11
void FASTCALL call_7_12_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_1_12
#define init_7_12_1_12
void FASTCALL call_7_12_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_1_13
#define init_7_12_1_13
void FASTCALL call_7_12_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_1_14
#define init_7_12_1_14
void FASTCALL call_7_12_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_1_15
#define init_7_12_1_15
void FASTCALL call_7_12_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_2_0
#define init_7_12_2_0
void FASTCALL call_7_12_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_2_1
#define init_7_12_2_1
void FASTCALL call_7_12_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_2_2
#define init_7_12_2_2
void FASTCALL call_7_12_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_2_3
#define init_7_12_2_3
void FASTCALL call_7_12_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_2_4
#define init_7_12_2_4
void FASTCALL call_7_12_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_2_5
#define init_7_12_2_5
void FASTCALL call_7_12_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_2_6
#define init_7_12_2_6
void FASTCALL call_7_12_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_2_7
#define init_7_12_2_7
void FASTCALL call_7_12_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_2_8
#define init_7_12_2_8
void FASTCALL call_7_12_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_2_9
#define init_7_12_2_9
void FASTCALL call_7_12_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_2_10
#define init_7_12_2_10
void FASTCALL call_7_12_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_2_11
#define init_7_12_2_11
void FASTCALL call_7_12_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_2_12
#define init_7_12_2_12
void FASTCALL call_7_12_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_2_13
#define init_7_12_2_13
void FASTCALL call_7_12_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_2_14
#define init_7_12_2_14
void FASTCALL call_7_12_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_2_15
#define init_7_12_2_15
void FASTCALL call_7_12_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_3_0
#define init_7_12_3_0
void FASTCALL call_7_12_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_3_1
#define init_7_12_3_1
void FASTCALL call_7_12_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_3_2
#define init_7_12_3_2
void FASTCALL call_7_12_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_3_3
#define init_7_12_3_3
void FASTCALL call_7_12_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_3_4
#define init_7_12_3_4
void FASTCALL call_7_12_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_3_5
#define init_7_12_3_5
void FASTCALL call_7_12_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_3_6
#define init_7_12_3_6
void FASTCALL call_7_12_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_3_7
#define init_7_12_3_7
void FASTCALL call_7_12_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_3_8
#define init_7_12_3_8
void FASTCALL call_7_12_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_3_9
#define init_7_12_3_9
void FASTCALL call_7_12_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_3_10
#define init_7_12_3_10
void FASTCALL call_7_12_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_3_11
#define init_7_12_3_11
void FASTCALL call_7_12_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_3_12
#define init_7_12_3_12
void FASTCALL call_7_12_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_3_13
#define init_7_12_3_13
void FASTCALL call_7_12_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_3_14
#define init_7_12_3_14
void FASTCALL call_7_12_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_3_15
#define init_7_12_3_15
void FASTCALL call_7_12_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_4_0
#define init_7_12_4_0
void FASTCALL call_7_12_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_4_1
#define init_7_12_4_1
void FASTCALL call_7_12_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_4_2
#define init_7_12_4_2
void FASTCALL call_7_12_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_4_3
#define init_7_12_4_3
void FASTCALL call_7_12_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_4_4
#define init_7_12_4_4
void FASTCALL call_7_12_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_4_5
#define init_7_12_4_5
void FASTCALL call_7_12_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_4_6
#define init_7_12_4_6
void FASTCALL call_7_12_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_4_7
#define init_7_12_4_7
void FASTCALL call_7_12_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_4_8
#define init_7_12_4_8
void FASTCALL call_7_12_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_4_9
#define init_7_12_4_9
void FASTCALL call_7_12_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_4_10
#define init_7_12_4_10
void FASTCALL call_7_12_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_4_11
#define init_7_12_4_11
void FASTCALL call_7_12_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_4_12
#define init_7_12_4_12
void FASTCALL call_7_12_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_4_13
#define init_7_12_4_13
void FASTCALL call_7_12_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_4_14
#define init_7_12_4_14
void FASTCALL call_7_12_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_4_15
#define init_7_12_4_15
void FASTCALL call_7_12_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_5_0
#define init_7_12_5_0
void FASTCALL call_7_12_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_5_1
#define init_7_12_5_1
void FASTCALL call_7_12_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_5_2
#define init_7_12_5_2
void FASTCALL call_7_12_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_5_3
#define init_7_12_5_3
void FASTCALL call_7_12_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_5_4
#define init_7_12_5_4
void FASTCALL call_7_12_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_5_5
#define init_7_12_5_5
void FASTCALL call_7_12_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_5_6
#define init_7_12_5_6
void FASTCALL call_7_12_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_5_7
#define init_7_12_5_7
void FASTCALL call_7_12_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_5_8
#define init_7_12_5_8
void FASTCALL call_7_12_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_5_9
#define init_7_12_5_9
void FASTCALL call_7_12_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_5_10
#define init_7_12_5_10
void FASTCALL call_7_12_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_5_11
#define init_7_12_5_11
void FASTCALL call_7_12_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_5_12
#define init_7_12_5_12
void FASTCALL call_7_12_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_5_13
#define init_7_12_5_13
void FASTCALL call_7_12_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_5_14
#define init_7_12_5_14
void FASTCALL call_7_12_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_5_15
#define init_7_12_5_15
void FASTCALL call_7_12_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_6_0
#define init_7_12_6_0
void FASTCALL call_7_12_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_6_1
#define init_7_12_6_1
void FASTCALL call_7_12_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_6_2
#define init_7_12_6_2
void FASTCALL call_7_12_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_6_3
#define init_7_12_6_3
void FASTCALL call_7_12_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_6_4
#define init_7_12_6_4
void FASTCALL call_7_12_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_6_5
#define init_7_12_6_5
void FASTCALL call_7_12_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_6_6
#define init_7_12_6_6
void FASTCALL call_7_12_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_6_7
#define init_7_12_6_7
void FASTCALL call_7_12_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_6_8
#define init_7_12_6_8
void FASTCALL call_7_12_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_6_9
#define init_7_12_6_9
void FASTCALL call_7_12_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_6_10
#define init_7_12_6_10
void FASTCALL call_7_12_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_6_11
#define init_7_12_6_11
void FASTCALL call_7_12_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_6_12
#define init_7_12_6_12
void FASTCALL call_7_12_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_6_13
#define init_7_12_6_13
void FASTCALL call_7_12_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_6_14
#define init_7_12_6_14
void FASTCALL call_7_12_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_6_15
#define init_7_12_6_15
void FASTCALL call_7_12_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_7_0
#define init_7_12_7_0
void FASTCALL call_7_12_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_7_1
#define init_7_12_7_1
void FASTCALL call_7_12_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_7_2
#define init_7_12_7_2
void FASTCALL call_7_12_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_7_3
#define init_7_12_7_3
void FASTCALL call_7_12_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_7_4
#define init_7_12_7_4
void FASTCALL call_7_12_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_7_5
#define init_7_12_7_5
void FASTCALL call_7_12_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_7_6
#define init_7_12_7_6
void FASTCALL call_7_12_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_7_7
#define init_7_12_7_7
void FASTCALL call_7_12_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_7_8
#define init_7_12_7_8
void FASTCALL call_7_12_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_7_9
#define init_7_12_7_9
void FASTCALL call_7_12_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_7_10
#define init_7_12_7_10
void FASTCALL call_7_12_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_7_11
#define init_7_12_7_11
void FASTCALL call_7_12_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_7_12
#define init_7_12_7_12
void FASTCALL call_7_12_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_7_13
#define init_7_12_7_13
void FASTCALL call_7_12_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_7_14
#define init_7_12_7_14
void FASTCALL call_7_12_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_7_15
#define init_7_12_7_15
void FASTCALL call_7_12_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_8_0
#define init_7_12_8_0
void FASTCALL call_7_12_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_8_1
#define init_7_12_8_1
void FASTCALL call_7_12_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_8_2
#define init_7_12_8_2
void FASTCALL call_7_12_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_8_3
#define init_7_12_8_3
void FASTCALL call_7_12_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_8_4
#define init_7_12_8_4
void FASTCALL call_7_12_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_8_5
#define init_7_12_8_5
void FASTCALL call_7_12_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_8_6
#define init_7_12_8_6
void FASTCALL call_7_12_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_8_7
#define init_7_12_8_7
void FASTCALL call_7_12_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_8_8
#define init_7_12_8_8
void FASTCALL call_7_12_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_8_9
#define init_7_12_8_9
void FASTCALL call_7_12_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_8_10
#define init_7_12_8_10
void FASTCALL call_7_12_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_8_11
#define init_7_12_8_11
void FASTCALL call_7_12_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_8_12
#define init_7_12_8_12
void FASTCALL call_7_12_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_8_13
#define init_7_12_8_13
void FASTCALL call_7_12_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_8_14
#define init_7_12_8_14
void FASTCALL call_7_12_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_8_15
#define init_7_12_8_15
void FASTCALL call_7_12_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_9_0
#define init_7_12_9_0
void FASTCALL call_7_12_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_9_1
#define init_7_12_9_1
void FASTCALL call_7_12_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_9_2
#define init_7_12_9_2
void FASTCALL call_7_12_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_9_3
#define init_7_12_9_3
void FASTCALL call_7_12_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_9_4
#define init_7_12_9_4
void FASTCALL call_7_12_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_9_5
#define init_7_12_9_5
void FASTCALL call_7_12_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_9_6
#define init_7_12_9_6
void FASTCALL call_7_12_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_9_7
#define init_7_12_9_7
void FASTCALL call_7_12_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_9_8
#define init_7_12_9_8
void FASTCALL call_7_12_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_9_9
#define init_7_12_9_9
void FASTCALL call_7_12_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_9_10
#define init_7_12_9_10
void FASTCALL call_7_12_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_9_11
#define init_7_12_9_11
void FASTCALL call_7_12_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_9_12
#define init_7_12_9_12
void FASTCALL call_7_12_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_9_13
#define init_7_12_9_13
void FASTCALL call_7_12_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_9_14
#define init_7_12_9_14
void FASTCALL call_7_12_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_9_15
#define init_7_12_9_15
void FASTCALL call_7_12_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_10_0
#define init_7_12_10_0
void FASTCALL call_7_12_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_10_1
#define init_7_12_10_1
void FASTCALL call_7_12_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_10_2
#define init_7_12_10_2
void FASTCALL call_7_12_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_10_3
#define init_7_12_10_3
void FASTCALL call_7_12_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_10_4
#define init_7_12_10_4
void FASTCALL call_7_12_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_10_5
#define init_7_12_10_5
void FASTCALL call_7_12_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_10_6
#define init_7_12_10_6
void FASTCALL call_7_12_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_10_7
#define init_7_12_10_7
void FASTCALL call_7_12_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_10_8
#define init_7_12_10_8
void FASTCALL call_7_12_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_10_9
#define init_7_12_10_9
void FASTCALL call_7_12_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_10_10
#define init_7_12_10_10
void FASTCALL call_7_12_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_10_11
#define init_7_12_10_11
void FASTCALL call_7_12_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_10_12
#define init_7_12_10_12
void FASTCALL call_7_12_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_10_13
#define init_7_12_10_13
void FASTCALL call_7_12_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_10_14
#define init_7_12_10_14
void FASTCALL call_7_12_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_10_15
#define init_7_12_10_15
void FASTCALL call_7_12_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_11_0
#define init_7_12_11_0
void FASTCALL call_7_12_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_11_1
#define init_7_12_11_1
void FASTCALL call_7_12_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_11_2
#define init_7_12_11_2
void FASTCALL call_7_12_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_11_3
#define init_7_12_11_3
void FASTCALL call_7_12_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_11_4
#define init_7_12_11_4
void FASTCALL call_7_12_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_11_5
#define init_7_12_11_5
void FASTCALL call_7_12_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_11_6
#define init_7_12_11_6
void FASTCALL call_7_12_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_11_7
#define init_7_12_11_7
void FASTCALL call_7_12_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_11_8
#define init_7_12_11_8
void FASTCALL call_7_12_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_11_9
#define init_7_12_11_9
void FASTCALL call_7_12_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_11_10
#define init_7_12_11_10
void FASTCALL call_7_12_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_11_11
#define init_7_12_11_11
void FASTCALL call_7_12_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_11_12
#define init_7_12_11_12
void FASTCALL call_7_12_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_11_13
#define init_7_12_11_13
void FASTCALL call_7_12_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_11_14
#define init_7_12_11_14
void FASTCALL call_7_12_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_11_15
#define init_7_12_11_15
void FASTCALL call_7_12_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_12_0
#define init_7_12_12_0
void FASTCALL call_7_12_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_12_1
#define init_7_12_12_1
void FASTCALL call_7_12_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_12_2
#define init_7_12_12_2
void FASTCALL call_7_12_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_12_3
#define init_7_12_12_3
void FASTCALL call_7_12_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_12_4
#define init_7_12_12_4
void FASTCALL call_7_12_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_12_5
#define init_7_12_12_5
void FASTCALL call_7_12_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_12_6
#define init_7_12_12_6
void FASTCALL call_7_12_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_12_7
#define init_7_12_12_7
void FASTCALL call_7_12_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_12_8
#define init_7_12_12_8
void FASTCALL call_7_12_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_12_9
#define init_7_12_12_9
void FASTCALL call_7_12_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_12_10
#define init_7_12_12_10
void FASTCALL call_7_12_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_12_11
#define init_7_12_12_11
void FASTCALL call_7_12_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_12_12
#define init_7_12_12_12
void FASTCALL call_7_12_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_12_13
#define init_7_12_12_13
void FASTCALL call_7_12_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_12_14
#define init_7_12_12_14
void FASTCALL call_7_12_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_12_15
#define init_7_12_12_15
void FASTCALL call_7_12_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_13_0
#define init_7_12_13_0
void FASTCALL call_7_12_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_13_1
#define init_7_12_13_1
void FASTCALL call_7_12_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_13_2
#define init_7_12_13_2
void FASTCALL call_7_12_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_13_3
#define init_7_12_13_3
void FASTCALL call_7_12_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_13_4
#define init_7_12_13_4
void FASTCALL call_7_12_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_13_5
#define init_7_12_13_5
void FASTCALL call_7_12_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_13_6
#define init_7_12_13_6
void FASTCALL call_7_12_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_13_7
#define init_7_12_13_7
void FASTCALL call_7_12_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_13_8
#define init_7_12_13_8
void FASTCALL call_7_12_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_13_9
#define init_7_12_13_9
void FASTCALL call_7_12_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_13_10
#define init_7_12_13_10
void FASTCALL call_7_12_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_13_11
#define init_7_12_13_11
void FASTCALL call_7_12_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_13_12
#define init_7_12_13_12
void FASTCALL call_7_12_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_13_13
#define init_7_12_13_13
void FASTCALL call_7_12_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_13_14
#define init_7_12_13_14
void FASTCALL call_7_12_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_13_15
#define init_7_12_13_15
void FASTCALL call_7_12_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_14_0
#define init_7_12_14_0
void FASTCALL call_7_12_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_14_1
#define init_7_12_14_1
void FASTCALL call_7_12_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_14_2
#define init_7_12_14_2
void FASTCALL call_7_12_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_14_3
#define init_7_12_14_3
void FASTCALL call_7_12_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_14_4
#define init_7_12_14_4
void FASTCALL call_7_12_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_14_5
#define init_7_12_14_5
void FASTCALL call_7_12_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_14_6
#define init_7_12_14_6
void FASTCALL call_7_12_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_14_7
#define init_7_12_14_7
void FASTCALL call_7_12_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_14_8
#define init_7_12_14_8
void FASTCALL call_7_12_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_14_9
#define init_7_12_14_9
void FASTCALL call_7_12_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_14_10
#define init_7_12_14_10
void FASTCALL call_7_12_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_14_11
#define init_7_12_14_11
void FASTCALL call_7_12_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_14_12
#define init_7_12_14_12
void FASTCALL call_7_12_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_14_13
#define init_7_12_14_13
void FASTCALL call_7_12_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_14_14
#define init_7_12_14_14
void FASTCALL call_7_12_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_14_15
#define init_7_12_14_15
void FASTCALL call_7_12_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_15_0
#define init_7_12_15_0
void FASTCALL call_7_12_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_15_1
#define init_7_12_15_1
void FASTCALL call_7_12_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_15_2
#define init_7_12_15_2
void FASTCALL call_7_12_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_15_3
#define init_7_12_15_3
void FASTCALL call_7_12_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_15_4
#define init_7_12_15_4
void FASTCALL call_7_12_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_15_5
#define init_7_12_15_5
void FASTCALL call_7_12_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_15_6
#define init_7_12_15_6
void FASTCALL call_7_12_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_15_7
#define init_7_12_15_7
void FASTCALL call_7_12_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_15_8
#define init_7_12_15_8
void FASTCALL call_7_12_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_15_9
#define init_7_12_15_9
void FASTCALL call_7_12_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_15_10
#define init_7_12_15_10
void FASTCALL call_7_12_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_15_11
#define init_7_12_15_11
void FASTCALL call_7_12_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_15_12
#define init_7_12_15_12
void FASTCALL call_7_12_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_15_13
#define init_7_12_15_13
void FASTCALL call_7_12_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_15_14
#define init_7_12_15_14
void FASTCALL call_7_12_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_12_15_15
#define init_7_12_15_15
void FASTCALL call_7_12_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_0_0
#define init_7_13_0_0
void FASTCALL call_7_13_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_0_1
#define init_7_13_0_1
void FASTCALL call_7_13_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_0_2
#define init_7_13_0_2
void FASTCALL call_7_13_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_0_3
#define init_7_13_0_3
void FASTCALL call_7_13_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_0_4
#define init_7_13_0_4
void FASTCALL call_7_13_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_0_5
#define init_7_13_0_5
void FASTCALL call_7_13_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_0_6
#define init_7_13_0_6
void FASTCALL call_7_13_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_0_7
#define init_7_13_0_7
void FASTCALL call_7_13_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_0_8
#define init_7_13_0_8
void FASTCALL call_7_13_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_0_9
#define init_7_13_0_9
void FASTCALL call_7_13_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_0_10
#define init_7_13_0_10
void FASTCALL call_7_13_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_0_11
#define init_7_13_0_11
void FASTCALL call_7_13_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_0_12
#define init_7_13_0_12
void FASTCALL call_7_13_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_0_13
#define init_7_13_0_13
void FASTCALL call_7_13_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_0_14
#define init_7_13_0_14
void FASTCALL call_7_13_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_0_15
#define init_7_13_0_15
void FASTCALL call_7_13_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_1_0
#define init_7_13_1_0
void FASTCALL call_7_13_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_1_1
#define init_7_13_1_1
void FASTCALL call_7_13_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_1_2
#define init_7_13_1_2
void FASTCALL call_7_13_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_1_3
#define init_7_13_1_3
void FASTCALL call_7_13_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_1_4
#define init_7_13_1_4
void FASTCALL call_7_13_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_1_5
#define init_7_13_1_5
void FASTCALL call_7_13_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_1_6
#define init_7_13_1_6
void FASTCALL call_7_13_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_1_7
#define init_7_13_1_7
void FASTCALL call_7_13_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_1_8
#define init_7_13_1_8
void FASTCALL call_7_13_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_1_9
#define init_7_13_1_9
void FASTCALL call_7_13_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_1_10
#define init_7_13_1_10
void FASTCALL call_7_13_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_1_11
#define init_7_13_1_11
void FASTCALL call_7_13_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_1_12
#define init_7_13_1_12
void FASTCALL call_7_13_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_1_13
#define init_7_13_1_13
void FASTCALL call_7_13_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_1_14
#define init_7_13_1_14
void FASTCALL call_7_13_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_1_15
#define init_7_13_1_15
void FASTCALL call_7_13_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_2_0
#define init_7_13_2_0
void FASTCALL call_7_13_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_2_1
#define init_7_13_2_1
void FASTCALL call_7_13_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_2_2
#define init_7_13_2_2
void FASTCALL call_7_13_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_2_3
#define init_7_13_2_3
void FASTCALL call_7_13_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_2_4
#define init_7_13_2_4
void FASTCALL call_7_13_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_2_5
#define init_7_13_2_5
void FASTCALL call_7_13_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_2_6
#define init_7_13_2_6
void FASTCALL call_7_13_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_2_7
#define init_7_13_2_7
void FASTCALL call_7_13_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_2_8
#define init_7_13_2_8
void FASTCALL call_7_13_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_2_9
#define init_7_13_2_9
void FASTCALL call_7_13_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_2_10
#define init_7_13_2_10
void FASTCALL call_7_13_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_2_11
#define init_7_13_2_11
void FASTCALL call_7_13_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_2_12
#define init_7_13_2_12
void FASTCALL call_7_13_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_2_13
#define init_7_13_2_13
void FASTCALL call_7_13_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_2_14
#define init_7_13_2_14
void FASTCALL call_7_13_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_2_15
#define init_7_13_2_15
void FASTCALL call_7_13_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_3_0
#define init_7_13_3_0
void FASTCALL call_7_13_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_3_1
#define init_7_13_3_1
void FASTCALL call_7_13_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_3_2
#define init_7_13_3_2
void FASTCALL call_7_13_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_3_3
#define init_7_13_3_3
void FASTCALL call_7_13_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_3_4
#define init_7_13_3_4
void FASTCALL call_7_13_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_3_5
#define init_7_13_3_5
void FASTCALL call_7_13_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_3_6
#define init_7_13_3_6
void FASTCALL call_7_13_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_3_7
#define init_7_13_3_7
void FASTCALL call_7_13_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_3_8
#define init_7_13_3_8
void FASTCALL call_7_13_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_3_9
#define init_7_13_3_9
void FASTCALL call_7_13_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_3_10
#define init_7_13_3_10
void FASTCALL call_7_13_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_3_11
#define init_7_13_3_11
void FASTCALL call_7_13_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_3_12
#define init_7_13_3_12
void FASTCALL call_7_13_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_3_13
#define init_7_13_3_13
void FASTCALL call_7_13_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_3_14
#define init_7_13_3_14
void FASTCALL call_7_13_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_3_15
#define init_7_13_3_15
void FASTCALL call_7_13_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_4_0
#define init_7_13_4_0
void FASTCALL call_7_13_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_4_1
#define init_7_13_4_1
void FASTCALL call_7_13_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_4_2
#define init_7_13_4_2
void FASTCALL call_7_13_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_4_3
#define init_7_13_4_3
void FASTCALL call_7_13_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_4_4
#define init_7_13_4_4
void FASTCALL call_7_13_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_4_5
#define init_7_13_4_5
void FASTCALL call_7_13_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_4_6
#define init_7_13_4_6
void FASTCALL call_7_13_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_4_7
#define init_7_13_4_7
void FASTCALL call_7_13_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_4_8
#define init_7_13_4_8
void FASTCALL call_7_13_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_4_9
#define init_7_13_4_9
void FASTCALL call_7_13_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_4_10
#define init_7_13_4_10
void FASTCALL call_7_13_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_4_11
#define init_7_13_4_11
void FASTCALL call_7_13_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_4_12
#define init_7_13_4_12
void FASTCALL call_7_13_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_4_13
#define init_7_13_4_13
void FASTCALL call_7_13_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_4_14
#define init_7_13_4_14
void FASTCALL call_7_13_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_4_15
#define init_7_13_4_15
void FASTCALL call_7_13_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_5_0
#define init_7_13_5_0
void FASTCALL call_7_13_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_5_1
#define init_7_13_5_1
void FASTCALL call_7_13_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_5_2
#define init_7_13_5_2
void FASTCALL call_7_13_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_5_3
#define init_7_13_5_3
void FASTCALL call_7_13_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_5_4
#define init_7_13_5_4
void FASTCALL call_7_13_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_5_5
#define init_7_13_5_5
void FASTCALL call_7_13_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_5_6
#define init_7_13_5_6
void FASTCALL call_7_13_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_5_7
#define init_7_13_5_7
void FASTCALL call_7_13_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_5_8
#define init_7_13_5_8
void FASTCALL call_7_13_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_5_9
#define init_7_13_5_9
void FASTCALL call_7_13_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_5_10
#define init_7_13_5_10
void FASTCALL call_7_13_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_5_11
#define init_7_13_5_11
void FASTCALL call_7_13_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_5_12
#define init_7_13_5_12
void FASTCALL call_7_13_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_5_13
#define init_7_13_5_13
void FASTCALL call_7_13_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_5_14
#define init_7_13_5_14
void FASTCALL call_7_13_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_5_15
#define init_7_13_5_15
void FASTCALL call_7_13_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_6_0
#define init_7_13_6_0
void FASTCALL call_7_13_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_6_1
#define init_7_13_6_1
void FASTCALL call_7_13_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_6_2
#define init_7_13_6_2
void FASTCALL call_7_13_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_6_3
#define init_7_13_6_3
void FASTCALL call_7_13_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_6_4
#define init_7_13_6_4
void FASTCALL call_7_13_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_6_5
#define init_7_13_6_5
void FASTCALL call_7_13_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_6_6
#define init_7_13_6_6
void FASTCALL call_7_13_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_6_7
#define init_7_13_6_7
void FASTCALL call_7_13_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_6_8
#define init_7_13_6_8
void FASTCALL call_7_13_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_6_9
#define init_7_13_6_9
void FASTCALL call_7_13_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_6_10
#define init_7_13_6_10
void FASTCALL call_7_13_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_6_11
#define init_7_13_6_11
void FASTCALL call_7_13_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_6_12
#define init_7_13_6_12
void FASTCALL call_7_13_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_6_13
#define init_7_13_6_13
void FASTCALL call_7_13_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_6_14
#define init_7_13_6_14
void FASTCALL call_7_13_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_6_15
#define init_7_13_6_15
void FASTCALL call_7_13_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_7_0
#define init_7_13_7_0
void FASTCALL call_7_13_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_7_1
#define init_7_13_7_1
void FASTCALL call_7_13_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_7_2
#define init_7_13_7_2
void FASTCALL call_7_13_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_7_3
#define init_7_13_7_3
void FASTCALL call_7_13_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_7_4
#define init_7_13_7_4
void FASTCALL call_7_13_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_7_5
#define init_7_13_7_5
void FASTCALL call_7_13_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_7_6
#define init_7_13_7_6
void FASTCALL call_7_13_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_7_7
#define init_7_13_7_7
void FASTCALL call_7_13_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_7_8
#define init_7_13_7_8
void FASTCALL call_7_13_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_7_9
#define init_7_13_7_9
void FASTCALL call_7_13_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_7_10
#define init_7_13_7_10
void FASTCALL call_7_13_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_7_11
#define init_7_13_7_11
void FASTCALL call_7_13_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_7_12
#define init_7_13_7_12
void FASTCALL call_7_13_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_7_13
#define init_7_13_7_13
void FASTCALL call_7_13_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_7_14
#define init_7_13_7_14
void FASTCALL call_7_13_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_7_15
#define init_7_13_7_15
void FASTCALL call_7_13_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_8_0
#define init_7_13_8_0
void FASTCALL call_7_13_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_8_1
#define init_7_13_8_1
void FASTCALL call_7_13_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_8_2
#define init_7_13_8_2
void FASTCALL call_7_13_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_8_3
#define init_7_13_8_3
void FASTCALL call_7_13_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_8_4
#define init_7_13_8_4
void FASTCALL call_7_13_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_8_5
#define init_7_13_8_5
void FASTCALL call_7_13_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_8_6
#define init_7_13_8_6
void FASTCALL call_7_13_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_8_7
#define init_7_13_8_7
void FASTCALL call_7_13_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_8_8
#define init_7_13_8_8
void FASTCALL call_7_13_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_8_9
#define init_7_13_8_9
void FASTCALL call_7_13_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_8_10
#define init_7_13_8_10
void FASTCALL call_7_13_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_8_11
#define init_7_13_8_11
void FASTCALL call_7_13_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_8_12
#define init_7_13_8_12
void FASTCALL call_7_13_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_8_13
#define init_7_13_8_13
void FASTCALL call_7_13_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_8_14
#define init_7_13_8_14
void FASTCALL call_7_13_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_8_15
#define init_7_13_8_15
void FASTCALL call_7_13_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_9_0
#define init_7_13_9_0
void FASTCALL call_7_13_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_9_1
#define init_7_13_9_1
void FASTCALL call_7_13_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_9_2
#define init_7_13_9_2
void FASTCALL call_7_13_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_9_3
#define init_7_13_9_3
void FASTCALL call_7_13_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_9_4
#define init_7_13_9_4
void FASTCALL call_7_13_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_9_5
#define init_7_13_9_5
void FASTCALL call_7_13_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_9_6
#define init_7_13_9_6
void FASTCALL call_7_13_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_9_7
#define init_7_13_9_7
void FASTCALL call_7_13_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_9_8
#define init_7_13_9_8
void FASTCALL call_7_13_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_9_9
#define init_7_13_9_9
void FASTCALL call_7_13_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_9_10
#define init_7_13_9_10
void FASTCALL call_7_13_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_9_11
#define init_7_13_9_11
void FASTCALL call_7_13_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_9_12
#define init_7_13_9_12
void FASTCALL call_7_13_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_9_13
#define init_7_13_9_13
void FASTCALL call_7_13_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_9_14
#define init_7_13_9_14
void FASTCALL call_7_13_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_9_15
#define init_7_13_9_15
void FASTCALL call_7_13_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_10_0
#define init_7_13_10_0
void FASTCALL call_7_13_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_10_1
#define init_7_13_10_1
void FASTCALL call_7_13_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_10_2
#define init_7_13_10_2
void FASTCALL call_7_13_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_10_3
#define init_7_13_10_3
void FASTCALL call_7_13_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_10_4
#define init_7_13_10_4
void FASTCALL call_7_13_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_10_5
#define init_7_13_10_5
void FASTCALL call_7_13_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_10_6
#define init_7_13_10_6
void FASTCALL call_7_13_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_10_7
#define init_7_13_10_7
void FASTCALL call_7_13_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_10_8
#define init_7_13_10_8
void FASTCALL call_7_13_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_10_9
#define init_7_13_10_9
void FASTCALL call_7_13_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_10_10
#define init_7_13_10_10
void FASTCALL call_7_13_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_10_11
#define init_7_13_10_11
void FASTCALL call_7_13_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_10_12
#define init_7_13_10_12
void FASTCALL call_7_13_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_10_13
#define init_7_13_10_13
void FASTCALL call_7_13_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_10_14
#define init_7_13_10_14
void FASTCALL call_7_13_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_10_15
#define init_7_13_10_15
void FASTCALL call_7_13_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_11_0
#define init_7_13_11_0
void FASTCALL call_7_13_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_11_1
#define init_7_13_11_1
void FASTCALL call_7_13_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_11_2
#define init_7_13_11_2
void FASTCALL call_7_13_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_11_3
#define init_7_13_11_3
void FASTCALL call_7_13_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_11_4
#define init_7_13_11_4
void FASTCALL call_7_13_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_11_5
#define init_7_13_11_5
void FASTCALL call_7_13_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_11_6
#define init_7_13_11_6
void FASTCALL call_7_13_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_11_7
#define init_7_13_11_7
void FASTCALL call_7_13_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_11_8
#define init_7_13_11_8
void FASTCALL call_7_13_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_11_9
#define init_7_13_11_9
void FASTCALL call_7_13_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_11_10
#define init_7_13_11_10
void FASTCALL call_7_13_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_11_11
#define init_7_13_11_11
void FASTCALL call_7_13_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_11_12
#define init_7_13_11_12
void FASTCALL call_7_13_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_11_13
#define init_7_13_11_13
void FASTCALL call_7_13_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_11_14
#define init_7_13_11_14
void FASTCALL call_7_13_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_11_15
#define init_7_13_11_15
void FASTCALL call_7_13_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_12_0
#define init_7_13_12_0
void FASTCALL call_7_13_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_12_1
#define init_7_13_12_1
void FASTCALL call_7_13_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_12_2
#define init_7_13_12_2
void FASTCALL call_7_13_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_12_3
#define init_7_13_12_3
void FASTCALL call_7_13_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_12_4
#define init_7_13_12_4
void FASTCALL call_7_13_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_12_5
#define init_7_13_12_5
void FASTCALL call_7_13_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_12_6
#define init_7_13_12_6
void FASTCALL call_7_13_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_12_7
#define init_7_13_12_7
void FASTCALL call_7_13_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_12_8
#define init_7_13_12_8
void FASTCALL call_7_13_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_12_9
#define init_7_13_12_9
void FASTCALL call_7_13_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_12_10
#define init_7_13_12_10
void FASTCALL call_7_13_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_12_11
#define init_7_13_12_11
void FASTCALL call_7_13_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_12_12
#define init_7_13_12_12
void FASTCALL call_7_13_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_12_13
#define init_7_13_12_13
void FASTCALL call_7_13_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_12_14
#define init_7_13_12_14
void FASTCALL call_7_13_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_12_15
#define init_7_13_12_15
void FASTCALL call_7_13_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_13_0
#define init_7_13_13_0
void FASTCALL call_7_13_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_13_1
#define init_7_13_13_1
void FASTCALL call_7_13_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_13_2
#define init_7_13_13_2
void FASTCALL call_7_13_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_13_3
#define init_7_13_13_3
void FASTCALL call_7_13_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_13_4
#define init_7_13_13_4
void FASTCALL call_7_13_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_13_5
#define init_7_13_13_5
void FASTCALL call_7_13_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_13_6
#define init_7_13_13_6
void FASTCALL call_7_13_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_13_7
#define init_7_13_13_7
void FASTCALL call_7_13_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_13_8
#define init_7_13_13_8
void FASTCALL call_7_13_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_13_9
#define init_7_13_13_9
void FASTCALL call_7_13_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_13_10
#define init_7_13_13_10
void FASTCALL call_7_13_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_13_11
#define init_7_13_13_11
void FASTCALL call_7_13_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_13_12
#define init_7_13_13_12
void FASTCALL call_7_13_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_13_13
#define init_7_13_13_13
void FASTCALL call_7_13_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_13_14
#define init_7_13_13_14
void FASTCALL call_7_13_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_13_15
#define init_7_13_13_15
void FASTCALL call_7_13_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_14_0
#define init_7_13_14_0
void FASTCALL call_7_13_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_14_1
#define init_7_13_14_1
void FASTCALL call_7_13_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_14_2
#define init_7_13_14_2
void FASTCALL call_7_13_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_14_3
#define init_7_13_14_3
void FASTCALL call_7_13_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_14_4
#define init_7_13_14_4
void FASTCALL call_7_13_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_14_5
#define init_7_13_14_5
void FASTCALL call_7_13_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_14_6
#define init_7_13_14_6
void FASTCALL call_7_13_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_14_7
#define init_7_13_14_7
void FASTCALL call_7_13_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_14_8
#define init_7_13_14_8
void FASTCALL call_7_13_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_14_9
#define init_7_13_14_9
void FASTCALL call_7_13_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_14_10
#define init_7_13_14_10
void FASTCALL call_7_13_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_14_11
#define init_7_13_14_11
void FASTCALL call_7_13_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_14_12
#define init_7_13_14_12
void FASTCALL call_7_13_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_14_13
#define init_7_13_14_13
void FASTCALL call_7_13_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_14_14
#define init_7_13_14_14
void FASTCALL call_7_13_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_14_15
#define init_7_13_14_15
void FASTCALL call_7_13_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_15_0
#define init_7_13_15_0
void FASTCALL call_7_13_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_15_1
#define init_7_13_15_1
void FASTCALL call_7_13_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_15_2
#define init_7_13_15_2
void FASTCALL call_7_13_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_15_3
#define init_7_13_15_3
void FASTCALL call_7_13_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_15_4
#define init_7_13_15_4
void FASTCALL call_7_13_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_15_5
#define init_7_13_15_5
void FASTCALL call_7_13_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_15_6
#define init_7_13_15_6
void FASTCALL call_7_13_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_15_7
#define init_7_13_15_7
void FASTCALL call_7_13_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_15_8
#define init_7_13_15_8
void FASTCALL call_7_13_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_15_9
#define init_7_13_15_9
void FASTCALL call_7_13_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_15_10
#define init_7_13_15_10
void FASTCALL call_7_13_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_15_11
#define init_7_13_15_11
void FASTCALL call_7_13_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_15_12
#define init_7_13_15_12
void FASTCALL call_7_13_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_15_13
#define init_7_13_15_13
void FASTCALL call_7_13_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_15_14
#define init_7_13_15_14
void FASTCALL call_7_13_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_13_15_15
#define init_7_13_15_15
void FASTCALL call_7_13_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_0_0
#define init_7_14_0_0
void FASTCALL call_7_14_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_0_1
#define init_7_14_0_1
void FASTCALL call_7_14_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_0_2
#define init_7_14_0_2
void FASTCALL call_7_14_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_0_3
#define init_7_14_0_3
void FASTCALL call_7_14_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_0_4
#define init_7_14_0_4
void FASTCALL call_7_14_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_0_5
#define init_7_14_0_5
void FASTCALL call_7_14_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_0_6
#define init_7_14_0_6
void FASTCALL call_7_14_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_0_7
#define init_7_14_0_7
void FASTCALL call_7_14_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_0_8
#define init_7_14_0_8
void FASTCALL call_7_14_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_0_9
#define init_7_14_0_9
void FASTCALL call_7_14_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_0_10
#define init_7_14_0_10
void FASTCALL call_7_14_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_0_11
#define init_7_14_0_11
void FASTCALL call_7_14_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_0_12
#define init_7_14_0_12
void FASTCALL call_7_14_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_0_13
#define init_7_14_0_13
void FASTCALL call_7_14_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_0_14
#define init_7_14_0_14
void FASTCALL call_7_14_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_0_15
#define init_7_14_0_15
void FASTCALL call_7_14_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_1_0
#define init_7_14_1_0
void FASTCALL call_7_14_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_1_1
#define init_7_14_1_1
void FASTCALL call_7_14_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_1_2
#define init_7_14_1_2
void FASTCALL call_7_14_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_1_3
#define init_7_14_1_3
void FASTCALL call_7_14_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_1_4
#define init_7_14_1_4
void FASTCALL call_7_14_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_1_5
#define init_7_14_1_5
void FASTCALL call_7_14_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_1_6
#define init_7_14_1_6
void FASTCALL call_7_14_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_1_7
#define init_7_14_1_7
void FASTCALL call_7_14_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_1_8
#define init_7_14_1_8
void FASTCALL call_7_14_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_1_9
#define init_7_14_1_9
void FASTCALL call_7_14_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_1_10
#define init_7_14_1_10
void FASTCALL call_7_14_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_1_11
#define init_7_14_1_11
void FASTCALL call_7_14_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_1_12
#define init_7_14_1_12
void FASTCALL call_7_14_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_1_13
#define init_7_14_1_13
void FASTCALL call_7_14_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_1_14
#define init_7_14_1_14
void FASTCALL call_7_14_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_1_15
#define init_7_14_1_15
void FASTCALL call_7_14_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_2_0
#define init_7_14_2_0
void FASTCALL call_7_14_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_2_1
#define init_7_14_2_1
void FASTCALL call_7_14_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_2_2
#define init_7_14_2_2
void FASTCALL call_7_14_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_2_3
#define init_7_14_2_3
void FASTCALL call_7_14_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_2_4
#define init_7_14_2_4
void FASTCALL call_7_14_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_2_5
#define init_7_14_2_5
void FASTCALL call_7_14_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_2_6
#define init_7_14_2_6
void FASTCALL call_7_14_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_2_7
#define init_7_14_2_7
void FASTCALL call_7_14_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_2_8
#define init_7_14_2_8
void FASTCALL call_7_14_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_2_9
#define init_7_14_2_9
void FASTCALL call_7_14_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_2_10
#define init_7_14_2_10
void FASTCALL call_7_14_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_2_11
#define init_7_14_2_11
void FASTCALL call_7_14_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_2_12
#define init_7_14_2_12
void FASTCALL call_7_14_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_2_13
#define init_7_14_2_13
void FASTCALL call_7_14_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_2_14
#define init_7_14_2_14
void FASTCALL call_7_14_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_2_15
#define init_7_14_2_15
void FASTCALL call_7_14_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_3_0
#define init_7_14_3_0
void FASTCALL call_7_14_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_3_1
#define init_7_14_3_1
void FASTCALL call_7_14_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_3_2
#define init_7_14_3_2
void FASTCALL call_7_14_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_3_3
#define init_7_14_3_3
void FASTCALL call_7_14_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_3_4
#define init_7_14_3_4
void FASTCALL call_7_14_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_3_5
#define init_7_14_3_5
void FASTCALL call_7_14_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_3_6
#define init_7_14_3_6
void FASTCALL call_7_14_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_3_7
#define init_7_14_3_7
void FASTCALL call_7_14_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_3_8
#define init_7_14_3_8
void FASTCALL call_7_14_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_3_9
#define init_7_14_3_9
void FASTCALL call_7_14_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_3_10
#define init_7_14_3_10
void FASTCALL call_7_14_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_3_11
#define init_7_14_3_11
void FASTCALL call_7_14_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_3_12
#define init_7_14_3_12
void FASTCALL call_7_14_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_3_13
#define init_7_14_3_13
void FASTCALL call_7_14_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_3_14
#define init_7_14_3_14
void FASTCALL call_7_14_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_3_15
#define init_7_14_3_15
void FASTCALL call_7_14_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_4_0
#define init_7_14_4_0
void FASTCALL call_7_14_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_4_1
#define init_7_14_4_1
void FASTCALL call_7_14_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_4_2
#define init_7_14_4_2
void FASTCALL call_7_14_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_4_3
#define init_7_14_4_3
void FASTCALL call_7_14_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_4_4
#define init_7_14_4_4
void FASTCALL call_7_14_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_4_5
#define init_7_14_4_5
void FASTCALL call_7_14_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_4_6
#define init_7_14_4_6
void FASTCALL call_7_14_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_4_7
#define init_7_14_4_7
void FASTCALL call_7_14_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_4_8
#define init_7_14_4_8
void FASTCALL call_7_14_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_4_9
#define init_7_14_4_9
void FASTCALL call_7_14_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_4_10
#define init_7_14_4_10
void FASTCALL call_7_14_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_4_11
#define init_7_14_4_11
void FASTCALL call_7_14_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_4_12
#define init_7_14_4_12
void FASTCALL call_7_14_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_4_13
#define init_7_14_4_13
void FASTCALL call_7_14_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_4_14
#define init_7_14_4_14
void FASTCALL call_7_14_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_4_15
#define init_7_14_4_15
void FASTCALL call_7_14_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_5_0
#define init_7_14_5_0
void FASTCALL call_7_14_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_5_1
#define init_7_14_5_1
void FASTCALL call_7_14_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_5_2
#define init_7_14_5_2
void FASTCALL call_7_14_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_5_3
#define init_7_14_5_3
void FASTCALL call_7_14_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_5_4
#define init_7_14_5_4
void FASTCALL call_7_14_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_5_5
#define init_7_14_5_5
void FASTCALL call_7_14_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_5_6
#define init_7_14_5_6
void FASTCALL call_7_14_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_5_7
#define init_7_14_5_7
void FASTCALL call_7_14_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_5_8
#define init_7_14_5_8
void FASTCALL call_7_14_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_5_9
#define init_7_14_5_9
void FASTCALL call_7_14_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_5_10
#define init_7_14_5_10
void FASTCALL call_7_14_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_5_11
#define init_7_14_5_11
void FASTCALL call_7_14_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_5_12
#define init_7_14_5_12
void FASTCALL call_7_14_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_5_13
#define init_7_14_5_13
void FASTCALL call_7_14_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_5_14
#define init_7_14_5_14
void FASTCALL call_7_14_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_5_15
#define init_7_14_5_15
void FASTCALL call_7_14_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_6_0
#define init_7_14_6_0
void FASTCALL call_7_14_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_6_1
#define init_7_14_6_1
void FASTCALL call_7_14_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_6_2
#define init_7_14_6_2
void FASTCALL call_7_14_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_6_3
#define init_7_14_6_3
void FASTCALL call_7_14_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_6_4
#define init_7_14_6_4
void FASTCALL call_7_14_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_6_5
#define init_7_14_6_5
void FASTCALL call_7_14_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_6_6
#define init_7_14_6_6
void FASTCALL call_7_14_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_6_7
#define init_7_14_6_7
void FASTCALL call_7_14_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_6_8
#define init_7_14_6_8
void FASTCALL call_7_14_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_6_9
#define init_7_14_6_9
void FASTCALL call_7_14_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_6_10
#define init_7_14_6_10
void FASTCALL call_7_14_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_6_11
#define init_7_14_6_11
void FASTCALL call_7_14_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_6_12
#define init_7_14_6_12
void FASTCALL call_7_14_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_6_13
#define init_7_14_6_13
void FASTCALL call_7_14_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_6_14
#define init_7_14_6_14
void FASTCALL call_7_14_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_6_15
#define init_7_14_6_15
void FASTCALL call_7_14_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_7_0
#define init_7_14_7_0
void FASTCALL call_7_14_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_7_1
#define init_7_14_7_1
void FASTCALL call_7_14_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_7_2
#define init_7_14_7_2
void FASTCALL call_7_14_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_7_3
#define init_7_14_7_3
void FASTCALL call_7_14_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_7_4
#define init_7_14_7_4
void FASTCALL call_7_14_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_7_5
#define init_7_14_7_5
void FASTCALL call_7_14_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_7_6
#define init_7_14_7_6
void FASTCALL call_7_14_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_7_7
#define init_7_14_7_7
void FASTCALL call_7_14_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_7_8
#define init_7_14_7_8
void FASTCALL call_7_14_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_7_9
#define init_7_14_7_9
void FASTCALL call_7_14_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_7_10
#define init_7_14_7_10
void FASTCALL call_7_14_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_7_11
#define init_7_14_7_11
void FASTCALL call_7_14_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_7_12
#define init_7_14_7_12
void FASTCALL call_7_14_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_7_13
#define init_7_14_7_13
void FASTCALL call_7_14_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_7_14
#define init_7_14_7_14
void FASTCALL call_7_14_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_7_15
#define init_7_14_7_15
void FASTCALL call_7_14_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_8_0
#define init_7_14_8_0
void FASTCALL call_7_14_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_8_1
#define init_7_14_8_1
void FASTCALL call_7_14_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_8_2
#define init_7_14_8_2
void FASTCALL call_7_14_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_8_3
#define init_7_14_8_3
void FASTCALL call_7_14_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_8_4
#define init_7_14_8_4
void FASTCALL call_7_14_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_8_5
#define init_7_14_8_5
void FASTCALL call_7_14_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_8_6
#define init_7_14_8_6
void FASTCALL call_7_14_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_8_7
#define init_7_14_8_7
void FASTCALL call_7_14_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_8_8
#define init_7_14_8_8
void FASTCALL call_7_14_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_8_9
#define init_7_14_8_9
void FASTCALL call_7_14_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_8_10
#define init_7_14_8_10
void FASTCALL call_7_14_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_8_11
#define init_7_14_8_11
void FASTCALL call_7_14_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_8_12
#define init_7_14_8_12
void FASTCALL call_7_14_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_8_13
#define init_7_14_8_13
void FASTCALL call_7_14_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_8_14
#define init_7_14_8_14
void FASTCALL call_7_14_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_8_15
#define init_7_14_8_15
void FASTCALL call_7_14_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_9_0
#define init_7_14_9_0
void FASTCALL call_7_14_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_9_1
#define init_7_14_9_1
void FASTCALL call_7_14_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_9_2
#define init_7_14_9_2
void FASTCALL call_7_14_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_9_3
#define init_7_14_9_3
void FASTCALL call_7_14_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_9_4
#define init_7_14_9_4
void FASTCALL call_7_14_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_9_5
#define init_7_14_9_5
void FASTCALL call_7_14_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_9_6
#define init_7_14_9_6
void FASTCALL call_7_14_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_9_7
#define init_7_14_9_7
void FASTCALL call_7_14_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_9_8
#define init_7_14_9_8
void FASTCALL call_7_14_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_9_9
#define init_7_14_9_9
void FASTCALL call_7_14_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_9_10
#define init_7_14_9_10
void FASTCALL call_7_14_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_9_11
#define init_7_14_9_11
void FASTCALL call_7_14_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_9_12
#define init_7_14_9_12
void FASTCALL call_7_14_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_9_13
#define init_7_14_9_13
void FASTCALL call_7_14_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_9_14
#define init_7_14_9_14
void FASTCALL call_7_14_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_9_15
#define init_7_14_9_15
void FASTCALL call_7_14_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_10_0
#define init_7_14_10_0
void FASTCALL call_7_14_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_10_1
#define init_7_14_10_1
void FASTCALL call_7_14_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_10_2
#define init_7_14_10_2
void FASTCALL call_7_14_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_10_3
#define init_7_14_10_3
void FASTCALL call_7_14_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_10_4
#define init_7_14_10_4
void FASTCALL call_7_14_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_10_5
#define init_7_14_10_5
void FASTCALL call_7_14_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_10_6
#define init_7_14_10_6
void FASTCALL call_7_14_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_10_7
#define init_7_14_10_7
void FASTCALL call_7_14_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_10_8
#define init_7_14_10_8
void FASTCALL call_7_14_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_10_9
#define init_7_14_10_9
void FASTCALL call_7_14_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_10_10
#define init_7_14_10_10
void FASTCALL call_7_14_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_10_11
#define init_7_14_10_11
void FASTCALL call_7_14_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_10_12
#define init_7_14_10_12
void FASTCALL call_7_14_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_10_13
#define init_7_14_10_13
void FASTCALL call_7_14_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_10_14
#define init_7_14_10_14
void FASTCALL call_7_14_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_10_15
#define init_7_14_10_15
void FASTCALL call_7_14_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_11_0
#define init_7_14_11_0
void FASTCALL call_7_14_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_11_1
#define init_7_14_11_1
void FASTCALL call_7_14_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_11_2
#define init_7_14_11_2
void FASTCALL call_7_14_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_11_3
#define init_7_14_11_3
void FASTCALL call_7_14_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_11_4
#define init_7_14_11_4
void FASTCALL call_7_14_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_11_5
#define init_7_14_11_5
void FASTCALL call_7_14_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_11_6
#define init_7_14_11_6
void FASTCALL call_7_14_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_11_7
#define init_7_14_11_7
void FASTCALL call_7_14_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_11_8
#define init_7_14_11_8
void FASTCALL call_7_14_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_11_9
#define init_7_14_11_9
void FASTCALL call_7_14_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_11_10
#define init_7_14_11_10
void FASTCALL call_7_14_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_11_11
#define init_7_14_11_11
void FASTCALL call_7_14_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_11_12
#define init_7_14_11_12
void FASTCALL call_7_14_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_11_13
#define init_7_14_11_13
void FASTCALL call_7_14_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_11_14
#define init_7_14_11_14
void FASTCALL call_7_14_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_11_15
#define init_7_14_11_15
void FASTCALL call_7_14_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_12_0
#define init_7_14_12_0
void FASTCALL call_7_14_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_12_1
#define init_7_14_12_1
void FASTCALL call_7_14_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_12_2
#define init_7_14_12_2
void FASTCALL call_7_14_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_12_3
#define init_7_14_12_3
void FASTCALL call_7_14_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_12_4
#define init_7_14_12_4
void FASTCALL call_7_14_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_12_5
#define init_7_14_12_5
void FASTCALL call_7_14_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_12_6
#define init_7_14_12_6
void FASTCALL call_7_14_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_12_7
#define init_7_14_12_7
void FASTCALL call_7_14_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_12_8
#define init_7_14_12_8
void FASTCALL call_7_14_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_12_9
#define init_7_14_12_9
void FASTCALL call_7_14_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_12_10
#define init_7_14_12_10
void FASTCALL call_7_14_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_12_11
#define init_7_14_12_11
void FASTCALL call_7_14_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_12_12
#define init_7_14_12_12
void FASTCALL call_7_14_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_12_13
#define init_7_14_12_13
void FASTCALL call_7_14_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_12_14
#define init_7_14_12_14
void FASTCALL call_7_14_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_12_15
#define init_7_14_12_15
void FASTCALL call_7_14_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_13_0
#define init_7_14_13_0
void FASTCALL call_7_14_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_13_1
#define init_7_14_13_1
void FASTCALL call_7_14_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_13_2
#define init_7_14_13_2
void FASTCALL call_7_14_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_13_3
#define init_7_14_13_3
void FASTCALL call_7_14_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_13_4
#define init_7_14_13_4
void FASTCALL call_7_14_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_13_5
#define init_7_14_13_5
void FASTCALL call_7_14_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_13_6
#define init_7_14_13_6
void FASTCALL call_7_14_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_13_7
#define init_7_14_13_7
void FASTCALL call_7_14_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_13_8
#define init_7_14_13_8
void FASTCALL call_7_14_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_13_9
#define init_7_14_13_9
void FASTCALL call_7_14_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_13_10
#define init_7_14_13_10
void FASTCALL call_7_14_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_13_11
#define init_7_14_13_11
void FASTCALL call_7_14_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_13_12
#define init_7_14_13_12
void FASTCALL call_7_14_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_13_13
#define init_7_14_13_13
void FASTCALL call_7_14_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_13_14
#define init_7_14_13_14
void FASTCALL call_7_14_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_13_15
#define init_7_14_13_15
void FASTCALL call_7_14_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_14_0
#define init_7_14_14_0
void FASTCALL call_7_14_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_14_1
#define init_7_14_14_1
void FASTCALL call_7_14_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_14_2
#define init_7_14_14_2
void FASTCALL call_7_14_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_14_3
#define init_7_14_14_3
void FASTCALL call_7_14_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_14_4
#define init_7_14_14_4
void FASTCALL call_7_14_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_14_5
#define init_7_14_14_5
void FASTCALL call_7_14_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_14_6
#define init_7_14_14_6
void FASTCALL call_7_14_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_14_7
#define init_7_14_14_7
void FASTCALL call_7_14_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_14_8
#define init_7_14_14_8
void FASTCALL call_7_14_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_14_9
#define init_7_14_14_9
void FASTCALL call_7_14_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_14_10
#define init_7_14_14_10
void FASTCALL call_7_14_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_14_11
#define init_7_14_14_11
void FASTCALL call_7_14_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_14_12
#define init_7_14_14_12
void FASTCALL call_7_14_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_14_13
#define init_7_14_14_13
void FASTCALL call_7_14_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_14_14
#define init_7_14_14_14
void FASTCALL call_7_14_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_14_15
#define init_7_14_14_15
void FASTCALL call_7_14_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_15_0
#define init_7_14_15_0
void FASTCALL call_7_14_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_15_1
#define init_7_14_15_1
void FASTCALL call_7_14_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_15_2
#define init_7_14_15_2
void FASTCALL call_7_14_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_15_3
#define init_7_14_15_3
void FASTCALL call_7_14_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_15_4
#define init_7_14_15_4
void FASTCALL call_7_14_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_15_5
#define init_7_14_15_5
void FASTCALL call_7_14_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_15_6
#define init_7_14_15_6
void FASTCALL call_7_14_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_15_7
#define init_7_14_15_7
void FASTCALL call_7_14_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_15_8
#define init_7_14_15_8
void FASTCALL call_7_14_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_15_9
#define init_7_14_15_9
void FASTCALL call_7_14_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_15_10
#define init_7_14_15_10
void FASTCALL call_7_14_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_15_11
#define init_7_14_15_11
void FASTCALL call_7_14_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_15_12
#define init_7_14_15_12
void FASTCALL call_7_14_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_15_13
#define init_7_14_15_13
void FASTCALL call_7_14_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_15_14
#define init_7_14_15_14
void FASTCALL call_7_14_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_14_15_15
#define init_7_14_15_15
void FASTCALL call_7_14_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_0_0
#define init_7_15_0_0
void FASTCALL call_7_15_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_0_1
#define init_7_15_0_1
void FASTCALL call_7_15_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_0_2
#define init_7_15_0_2
void FASTCALL call_7_15_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_0_3
#define init_7_15_0_3
void FASTCALL call_7_15_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_0_4
#define init_7_15_0_4
void FASTCALL call_7_15_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_0_5
#define init_7_15_0_5
void FASTCALL call_7_15_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_0_6
#define init_7_15_0_6
void FASTCALL call_7_15_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_0_7
#define init_7_15_0_7
void FASTCALL call_7_15_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_0_8
#define init_7_15_0_8
void FASTCALL call_7_15_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_0_9
#define init_7_15_0_9
void FASTCALL call_7_15_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_0_10
#define init_7_15_0_10
void FASTCALL call_7_15_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_0_11
#define init_7_15_0_11
void FASTCALL call_7_15_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_0_12
#define init_7_15_0_12
void FASTCALL call_7_15_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_0_13
#define init_7_15_0_13
void FASTCALL call_7_15_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_0_14
#define init_7_15_0_14
void FASTCALL call_7_15_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_0_15
#define init_7_15_0_15
void FASTCALL call_7_15_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_1_0
#define init_7_15_1_0
void FASTCALL call_7_15_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_1_1
#define init_7_15_1_1
void FASTCALL call_7_15_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_1_2
#define init_7_15_1_2
void FASTCALL call_7_15_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_1_3
#define init_7_15_1_3
void FASTCALL call_7_15_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_1_4
#define init_7_15_1_4
void FASTCALL call_7_15_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_1_5
#define init_7_15_1_5
void FASTCALL call_7_15_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_1_6
#define init_7_15_1_6
void FASTCALL call_7_15_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_1_7
#define init_7_15_1_7
void FASTCALL call_7_15_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_1_8
#define init_7_15_1_8
void FASTCALL call_7_15_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_1_9
#define init_7_15_1_9
void FASTCALL call_7_15_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_1_10
#define init_7_15_1_10
void FASTCALL call_7_15_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_1_11
#define init_7_15_1_11
void FASTCALL call_7_15_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_1_12
#define init_7_15_1_12
void FASTCALL call_7_15_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_1_13
#define init_7_15_1_13
void FASTCALL call_7_15_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_1_14
#define init_7_15_1_14
void FASTCALL call_7_15_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_1_15
#define init_7_15_1_15
void FASTCALL call_7_15_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_2_0
#define init_7_15_2_0
void FASTCALL call_7_15_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_2_1
#define init_7_15_2_1
void FASTCALL call_7_15_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_2_2
#define init_7_15_2_2
void FASTCALL call_7_15_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_2_3
#define init_7_15_2_3
void FASTCALL call_7_15_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_2_4
#define init_7_15_2_4
void FASTCALL call_7_15_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_2_5
#define init_7_15_2_5
void FASTCALL call_7_15_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_2_6
#define init_7_15_2_6
void FASTCALL call_7_15_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_2_7
#define init_7_15_2_7
void FASTCALL call_7_15_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_2_8
#define init_7_15_2_8
void FASTCALL call_7_15_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_2_9
#define init_7_15_2_9
void FASTCALL call_7_15_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_2_10
#define init_7_15_2_10
void FASTCALL call_7_15_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_2_11
#define init_7_15_2_11
void FASTCALL call_7_15_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_2_12
#define init_7_15_2_12
void FASTCALL call_7_15_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_2_13
#define init_7_15_2_13
void FASTCALL call_7_15_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_2_14
#define init_7_15_2_14
void FASTCALL call_7_15_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_2_15
#define init_7_15_2_15
void FASTCALL call_7_15_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_3_0
#define init_7_15_3_0
void FASTCALL call_7_15_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_3_1
#define init_7_15_3_1
void FASTCALL call_7_15_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_3_2
#define init_7_15_3_2
void FASTCALL call_7_15_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_3_3
#define init_7_15_3_3
void FASTCALL call_7_15_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_3_4
#define init_7_15_3_4
void FASTCALL call_7_15_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_3_5
#define init_7_15_3_5
void FASTCALL call_7_15_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_3_6
#define init_7_15_3_6
void FASTCALL call_7_15_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_3_7
#define init_7_15_3_7
void FASTCALL call_7_15_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_3_8
#define init_7_15_3_8
void FASTCALL call_7_15_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_3_9
#define init_7_15_3_9
void FASTCALL call_7_15_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_3_10
#define init_7_15_3_10
void FASTCALL call_7_15_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_3_11
#define init_7_15_3_11
void FASTCALL call_7_15_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_3_12
#define init_7_15_3_12
void FASTCALL call_7_15_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_3_13
#define init_7_15_3_13
void FASTCALL call_7_15_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_3_14
#define init_7_15_3_14
void FASTCALL call_7_15_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_3_15
#define init_7_15_3_15
void FASTCALL call_7_15_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_4_0
#define init_7_15_4_0
void FASTCALL call_7_15_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_4_1
#define init_7_15_4_1
void FASTCALL call_7_15_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_4_2
#define init_7_15_4_2
void FASTCALL call_7_15_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_4_3
#define init_7_15_4_3
void FASTCALL call_7_15_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_4_4
#define init_7_15_4_4
void FASTCALL call_7_15_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_4_5
#define init_7_15_4_5
void FASTCALL call_7_15_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_4_6
#define init_7_15_4_6
void FASTCALL call_7_15_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_4_7
#define init_7_15_4_7
void FASTCALL call_7_15_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_4_8
#define init_7_15_4_8
void FASTCALL call_7_15_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_4_9
#define init_7_15_4_9
void FASTCALL call_7_15_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_4_10
#define init_7_15_4_10
void FASTCALL call_7_15_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_4_11
#define init_7_15_4_11
void FASTCALL call_7_15_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_4_12
#define init_7_15_4_12
void FASTCALL call_7_15_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_4_13
#define init_7_15_4_13
void FASTCALL call_7_15_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_4_14
#define init_7_15_4_14
void FASTCALL call_7_15_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_4_15
#define init_7_15_4_15
void FASTCALL call_7_15_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_5_0
#define init_7_15_5_0
void FASTCALL call_7_15_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_5_1
#define init_7_15_5_1
void FASTCALL call_7_15_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_5_2
#define init_7_15_5_2
void FASTCALL call_7_15_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_5_3
#define init_7_15_5_3
void FASTCALL call_7_15_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_5_4
#define init_7_15_5_4
void FASTCALL call_7_15_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_5_5
#define init_7_15_5_5
void FASTCALL call_7_15_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_5_6
#define init_7_15_5_6
void FASTCALL call_7_15_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_5_7
#define init_7_15_5_7
void FASTCALL call_7_15_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_5_8
#define init_7_15_5_8
void FASTCALL call_7_15_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_5_9
#define init_7_15_5_9
void FASTCALL call_7_15_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_5_10
#define init_7_15_5_10
void FASTCALL call_7_15_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_5_11
#define init_7_15_5_11
void FASTCALL call_7_15_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_5_12
#define init_7_15_5_12
void FASTCALL call_7_15_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_5_13
#define init_7_15_5_13
void FASTCALL call_7_15_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_5_14
#define init_7_15_5_14
void FASTCALL call_7_15_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_5_15
#define init_7_15_5_15
void FASTCALL call_7_15_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_6_0
#define init_7_15_6_0
void FASTCALL call_7_15_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_6_1
#define init_7_15_6_1
void FASTCALL call_7_15_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_6_2
#define init_7_15_6_2
void FASTCALL call_7_15_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_6_3
#define init_7_15_6_3
void FASTCALL call_7_15_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_6_4
#define init_7_15_6_4
void FASTCALL call_7_15_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_6_5
#define init_7_15_6_5
void FASTCALL call_7_15_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_6_6
#define init_7_15_6_6
void FASTCALL call_7_15_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_6_7
#define init_7_15_6_7
void FASTCALL call_7_15_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_6_8
#define init_7_15_6_8
void FASTCALL call_7_15_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_6_9
#define init_7_15_6_9
void FASTCALL call_7_15_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_6_10
#define init_7_15_6_10
void FASTCALL call_7_15_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_6_11
#define init_7_15_6_11
void FASTCALL call_7_15_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_6_12
#define init_7_15_6_12
void FASTCALL call_7_15_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_6_13
#define init_7_15_6_13
void FASTCALL call_7_15_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_6_14
#define init_7_15_6_14
void FASTCALL call_7_15_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_6_15
#define init_7_15_6_15
void FASTCALL call_7_15_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_7_0
#define init_7_15_7_0
void FASTCALL call_7_15_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_7_1
#define init_7_15_7_1
void FASTCALL call_7_15_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_7_2
#define init_7_15_7_2
void FASTCALL call_7_15_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_7_3
#define init_7_15_7_3
void FASTCALL call_7_15_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_7_4
#define init_7_15_7_4
void FASTCALL call_7_15_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_7_5
#define init_7_15_7_5
void FASTCALL call_7_15_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_7_6
#define init_7_15_7_6
void FASTCALL call_7_15_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_7_7
#define init_7_15_7_7
void FASTCALL call_7_15_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_7_8
#define init_7_15_7_8
void FASTCALL call_7_15_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_7_9
#define init_7_15_7_9
void FASTCALL call_7_15_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_7_10
#define init_7_15_7_10
void FASTCALL call_7_15_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_7_11
#define init_7_15_7_11
void FASTCALL call_7_15_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_7_12
#define init_7_15_7_12
void FASTCALL call_7_15_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_7_13
#define init_7_15_7_13
void FASTCALL call_7_15_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_7_14
#define init_7_15_7_14
void FASTCALL call_7_15_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_7_15
#define init_7_15_7_15
void FASTCALL call_7_15_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_8_0
#define init_7_15_8_0
void FASTCALL call_7_15_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_8_1
#define init_7_15_8_1
void FASTCALL call_7_15_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_8_2
#define init_7_15_8_2
void FASTCALL call_7_15_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_8_3
#define init_7_15_8_3
void FASTCALL call_7_15_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_8_4
#define init_7_15_8_4
void FASTCALL call_7_15_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_8_5
#define init_7_15_8_5
void FASTCALL call_7_15_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_8_6
#define init_7_15_8_6
void FASTCALL call_7_15_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_8_7
#define init_7_15_8_7
void FASTCALL call_7_15_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_8_8
#define init_7_15_8_8
void FASTCALL call_7_15_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_8_9
#define init_7_15_8_9
void FASTCALL call_7_15_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_8_10
#define init_7_15_8_10
void FASTCALL call_7_15_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_8_11
#define init_7_15_8_11
void FASTCALL call_7_15_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_8_12
#define init_7_15_8_12
void FASTCALL call_7_15_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_8_13
#define init_7_15_8_13
void FASTCALL call_7_15_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_8_14
#define init_7_15_8_14
void FASTCALL call_7_15_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_8_15
#define init_7_15_8_15
void FASTCALL call_7_15_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_9_0
#define init_7_15_9_0
void FASTCALL call_7_15_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_9_1
#define init_7_15_9_1
void FASTCALL call_7_15_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_9_2
#define init_7_15_9_2
void FASTCALL call_7_15_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_9_3
#define init_7_15_9_3
void FASTCALL call_7_15_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_9_4
#define init_7_15_9_4
void FASTCALL call_7_15_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_9_5
#define init_7_15_9_5
void FASTCALL call_7_15_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_9_6
#define init_7_15_9_6
void FASTCALL call_7_15_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_9_7
#define init_7_15_9_7
void FASTCALL call_7_15_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_9_8
#define init_7_15_9_8
void FASTCALL call_7_15_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_9_9
#define init_7_15_9_9
void FASTCALL call_7_15_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_9_10
#define init_7_15_9_10
void FASTCALL call_7_15_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_9_11
#define init_7_15_9_11
void FASTCALL call_7_15_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_9_12
#define init_7_15_9_12
void FASTCALL call_7_15_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_9_13
#define init_7_15_9_13
void FASTCALL call_7_15_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_9_14
#define init_7_15_9_14
void FASTCALL call_7_15_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_9_15
#define init_7_15_9_15
void FASTCALL call_7_15_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_10_0
#define init_7_15_10_0
void FASTCALL call_7_15_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_10_1
#define init_7_15_10_1
void FASTCALL call_7_15_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_10_2
#define init_7_15_10_2
void FASTCALL call_7_15_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_10_3
#define init_7_15_10_3
void FASTCALL call_7_15_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_10_4
#define init_7_15_10_4
void FASTCALL call_7_15_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_10_5
#define init_7_15_10_5
void FASTCALL call_7_15_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_10_6
#define init_7_15_10_6
void FASTCALL call_7_15_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_10_7
#define init_7_15_10_7
void FASTCALL call_7_15_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_10_8
#define init_7_15_10_8
void FASTCALL call_7_15_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_10_9
#define init_7_15_10_9
void FASTCALL call_7_15_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_10_10
#define init_7_15_10_10
void FASTCALL call_7_15_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_10_11
#define init_7_15_10_11
void FASTCALL call_7_15_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_10_12
#define init_7_15_10_12
void FASTCALL call_7_15_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_10_13
#define init_7_15_10_13
void FASTCALL call_7_15_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_10_14
#define init_7_15_10_14
void FASTCALL call_7_15_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_10_15
#define init_7_15_10_15
void FASTCALL call_7_15_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_11_0
#define init_7_15_11_0
void FASTCALL call_7_15_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_11_1
#define init_7_15_11_1
void FASTCALL call_7_15_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_11_2
#define init_7_15_11_2
void FASTCALL call_7_15_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_11_3
#define init_7_15_11_3
void FASTCALL call_7_15_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_11_4
#define init_7_15_11_4
void FASTCALL call_7_15_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_11_5
#define init_7_15_11_5
void FASTCALL call_7_15_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_11_6
#define init_7_15_11_6
void FASTCALL call_7_15_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_11_7
#define init_7_15_11_7
void FASTCALL call_7_15_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_11_8
#define init_7_15_11_8
void FASTCALL call_7_15_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_11_9
#define init_7_15_11_9
void FASTCALL call_7_15_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_11_10
#define init_7_15_11_10
void FASTCALL call_7_15_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_11_11
#define init_7_15_11_11
void FASTCALL call_7_15_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_11_12
#define init_7_15_11_12
void FASTCALL call_7_15_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_11_13
#define init_7_15_11_13
void FASTCALL call_7_15_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_11_14
#define init_7_15_11_14
void FASTCALL call_7_15_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_11_15
#define init_7_15_11_15
void FASTCALL call_7_15_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_12_0
#define init_7_15_12_0
void FASTCALL call_7_15_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_12_1
#define init_7_15_12_1
void FASTCALL call_7_15_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_12_2
#define init_7_15_12_2
void FASTCALL call_7_15_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_12_3
#define init_7_15_12_3
void FASTCALL call_7_15_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_12_4
#define init_7_15_12_4
void FASTCALL call_7_15_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_12_5
#define init_7_15_12_5
void FASTCALL call_7_15_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_12_6
#define init_7_15_12_6
void FASTCALL call_7_15_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_12_7
#define init_7_15_12_7
void FASTCALL call_7_15_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_12_8
#define init_7_15_12_8
void FASTCALL call_7_15_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_12_9
#define init_7_15_12_9
void FASTCALL call_7_15_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_12_10
#define init_7_15_12_10
void FASTCALL call_7_15_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_12_11
#define init_7_15_12_11
void FASTCALL call_7_15_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_12_12
#define init_7_15_12_12
void FASTCALL call_7_15_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_12_13
#define init_7_15_12_13
void FASTCALL call_7_15_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_12_14
#define init_7_15_12_14
void FASTCALL call_7_15_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_12_15
#define init_7_15_12_15
void FASTCALL call_7_15_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_13_0
#define init_7_15_13_0
void FASTCALL call_7_15_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_13_1
#define init_7_15_13_1
void FASTCALL call_7_15_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_13_2
#define init_7_15_13_2
void FASTCALL call_7_15_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_13_3
#define init_7_15_13_3
void FASTCALL call_7_15_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_13_4
#define init_7_15_13_4
void FASTCALL call_7_15_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_13_5
#define init_7_15_13_5
void FASTCALL call_7_15_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_13_6
#define init_7_15_13_6
void FASTCALL call_7_15_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_13_7
#define init_7_15_13_7
void FASTCALL call_7_15_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_13_8
#define init_7_15_13_8
void FASTCALL call_7_15_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_13_9
#define init_7_15_13_9
void FASTCALL call_7_15_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_13_10
#define init_7_15_13_10
void FASTCALL call_7_15_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_13_11
#define init_7_15_13_11
void FASTCALL call_7_15_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_13_12
#define init_7_15_13_12
void FASTCALL call_7_15_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_13_13
#define init_7_15_13_13
void FASTCALL call_7_15_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_13_14
#define init_7_15_13_14
void FASTCALL call_7_15_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_13_15
#define init_7_15_13_15
void FASTCALL call_7_15_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_14_0
#define init_7_15_14_0
void FASTCALL call_7_15_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_14_1
#define init_7_15_14_1
void FASTCALL call_7_15_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_14_2
#define init_7_15_14_2
void FASTCALL call_7_15_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_14_3
#define init_7_15_14_3
void FASTCALL call_7_15_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_14_4
#define init_7_15_14_4
void FASTCALL call_7_15_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_14_5
#define init_7_15_14_5
void FASTCALL call_7_15_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_14_6
#define init_7_15_14_6
void FASTCALL call_7_15_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_14_7
#define init_7_15_14_7
void FASTCALL call_7_15_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_14_8
#define init_7_15_14_8
void FASTCALL call_7_15_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_14_9
#define init_7_15_14_9
void FASTCALL call_7_15_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_14_10
#define init_7_15_14_10
void FASTCALL call_7_15_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_14_11
#define init_7_15_14_11
void FASTCALL call_7_15_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_14_12
#define init_7_15_14_12
void FASTCALL call_7_15_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_14_13
#define init_7_15_14_13
void FASTCALL call_7_15_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_14_14
#define init_7_15_14_14
void FASTCALL call_7_15_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_14_15
#define init_7_15_14_15
void FASTCALL call_7_15_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_15_0
#define init_7_15_15_0
void FASTCALL call_7_15_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_15_1
#define init_7_15_15_1
void FASTCALL call_7_15_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_15_2
#define init_7_15_15_2
void FASTCALL call_7_15_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_15_3
#define init_7_15_15_3
void FASTCALL call_7_15_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_15_4
#define init_7_15_15_4
void FASTCALL call_7_15_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_15_5
#define init_7_15_15_5
void FASTCALL call_7_15_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_15_6
#define init_7_15_15_6
void FASTCALL call_7_15_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_15_7
#define init_7_15_15_7
void FASTCALL call_7_15_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_15_8
#define init_7_15_15_8
void FASTCALL call_7_15_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_15_9
#define init_7_15_15_9
void FASTCALL call_7_15_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_15_10
#define init_7_15_15_10
void FASTCALL call_7_15_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_15_11
#define init_7_15_15_11
void FASTCALL call_7_15_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_15_12
#define init_7_15_15_12
void FASTCALL call_7_15_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_15_13
#define init_7_15_15_13
void FASTCALL call_7_15_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_15_14
#define init_7_15_15_14
void FASTCALL call_7_15_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_7_15_15_15
#define init_7_15_15_15
void FASTCALL call_7_15_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_0_0
#define init_8_0_0_0
void FASTCALL call_8_0_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_0_1
#define init_8_0_0_1
void FASTCALL call_8_0_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_0_2
#define init_8_0_0_2
void FASTCALL call_8_0_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_0_3
#define init_8_0_0_3
void FASTCALL call_8_0_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_0_4
#define init_8_0_0_4
void FASTCALL call_8_0_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_0_5
#define init_8_0_0_5
void FASTCALL call_8_0_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_0_6
#define init_8_0_0_6
void FASTCALL call_8_0_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_0_7
#define init_8_0_0_7
void FASTCALL call_8_0_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_0_8
#define init_8_0_0_8
void FASTCALL call_8_0_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_0_9
#define init_8_0_0_9
void FASTCALL call_8_0_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_0_10
#define init_8_0_0_10
void FASTCALL call_8_0_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_0_11
#define init_8_0_0_11
void FASTCALL call_8_0_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_0_12
#define init_8_0_0_12
void FASTCALL call_8_0_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_0_13
#define init_8_0_0_13
void FASTCALL call_8_0_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_0_14
#define init_8_0_0_14
void FASTCALL call_8_0_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_0_15
#define init_8_0_0_15
void FASTCALL call_8_0_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_1_0
#define init_8_0_1_0
void FASTCALL call_8_0_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_1_1
#define init_8_0_1_1
void FASTCALL call_8_0_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_1_2
#define init_8_0_1_2
void FASTCALL call_8_0_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_1_3
#define init_8_0_1_3
void FASTCALL call_8_0_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_1_4
#define init_8_0_1_4
void FASTCALL call_8_0_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_1_5
#define init_8_0_1_5
void FASTCALL call_8_0_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_1_6
#define init_8_0_1_6
void FASTCALL call_8_0_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_1_7
#define init_8_0_1_7
void FASTCALL call_8_0_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_1_8
#define init_8_0_1_8
void FASTCALL call_8_0_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_1_9
#define init_8_0_1_9
void FASTCALL call_8_0_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_1_10
#define init_8_0_1_10
void FASTCALL call_8_0_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_1_11
#define init_8_0_1_11
void FASTCALL call_8_0_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_1_12
#define init_8_0_1_12
void FASTCALL call_8_0_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_1_13
#define init_8_0_1_13
void FASTCALL call_8_0_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_1_14
#define init_8_0_1_14
void FASTCALL call_8_0_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_1_15
#define init_8_0_1_15
void FASTCALL call_8_0_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_2_0
#define init_8_0_2_0
void FASTCALL call_8_0_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_2_1
#define init_8_0_2_1
void FASTCALL call_8_0_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_2_2
#define init_8_0_2_2
void FASTCALL call_8_0_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_2_3
#define init_8_0_2_3
void FASTCALL call_8_0_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_2_4
#define init_8_0_2_4
void FASTCALL call_8_0_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_2_5
#define init_8_0_2_5
void FASTCALL call_8_0_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_2_6
#define init_8_0_2_6
void FASTCALL call_8_0_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_2_7
#define init_8_0_2_7
void FASTCALL call_8_0_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_2_8
#define init_8_0_2_8
void FASTCALL call_8_0_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_2_9
#define init_8_0_2_9
void FASTCALL call_8_0_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_2_10
#define init_8_0_2_10
void FASTCALL call_8_0_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_2_11
#define init_8_0_2_11
void FASTCALL call_8_0_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_2_12
#define init_8_0_2_12
void FASTCALL call_8_0_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_2_13
#define init_8_0_2_13
void FASTCALL call_8_0_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_2_14
#define init_8_0_2_14
void FASTCALL call_8_0_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_2_15
#define init_8_0_2_15
void FASTCALL call_8_0_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_3_0
#define init_8_0_3_0
void FASTCALL call_8_0_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_3_1
#define init_8_0_3_1
void FASTCALL call_8_0_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_3_2
#define init_8_0_3_2
void FASTCALL call_8_0_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_3_3
#define init_8_0_3_3
void FASTCALL call_8_0_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_3_4
#define init_8_0_3_4
void FASTCALL call_8_0_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_3_5
#define init_8_0_3_5
void FASTCALL call_8_0_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_3_6
#define init_8_0_3_6
void FASTCALL call_8_0_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_3_7
#define init_8_0_3_7
void FASTCALL call_8_0_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_3_8
#define init_8_0_3_8
void FASTCALL call_8_0_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_3_9
#define init_8_0_3_9
void FASTCALL call_8_0_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_3_10
#define init_8_0_3_10
void FASTCALL call_8_0_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_3_11
#define init_8_0_3_11
void FASTCALL call_8_0_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_3_12
#define init_8_0_3_12
void FASTCALL call_8_0_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_3_13
#define init_8_0_3_13
void FASTCALL call_8_0_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_3_14
#define init_8_0_3_14
void FASTCALL call_8_0_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_3_15
#define init_8_0_3_15
void FASTCALL call_8_0_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_4_0
#define init_8_0_4_0
void FASTCALL call_8_0_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_4_1
#define init_8_0_4_1
void FASTCALL call_8_0_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_4_2
#define init_8_0_4_2
void FASTCALL call_8_0_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_4_3
#define init_8_0_4_3
void FASTCALL call_8_0_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_4_4
#define init_8_0_4_4
void FASTCALL call_8_0_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_4_5
#define init_8_0_4_5
void FASTCALL call_8_0_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_4_6
#define init_8_0_4_6
void FASTCALL call_8_0_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_4_7
#define init_8_0_4_7
void FASTCALL call_8_0_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_4_8
#define init_8_0_4_8
void FASTCALL call_8_0_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_4_9
#define init_8_0_4_9
void FASTCALL call_8_0_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_4_10
#define init_8_0_4_10
void FASTCALL call_8_0_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_4_11
#define init_8_0_4_11
void FASTCALL call_8_0_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_4_12
#define init_8_0_4_12
void FASTCALL call_8_0_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_4_13
#define init_8_0_4_13
void FASTCALL call_8_0_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_4_14
#define init_8_0_4_14
void FASTCALL call_8_0_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_4_15
#define init_8_0_4_15
void FASTCALL call_8_0_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_5_0
#define init_8_0_5_0
void FASTCALL call_8_0_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_5_1
#define init_8_0_5_1
void FASTCALL call_8_0_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_5_2
#define init_8_0_5_2
void FASTCALL call_8_0_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_5_3
#define init_8_0_5_3
void FASTCALL call_8_0_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_5_4
#define init_8_0_5_4
void FASTCALL call_8_0_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_5_5
#define init_8_0_5_5
void FASTCALL call_8_0_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_5_6
#define init_8_0_5_6
void FASTCALL call_8_0_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_5_7
#define init_8_0_5_7
void FASTCALL call_8_0_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_5_8
#define init_8_0_5_8
void FASTCALL call_8_0_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_5_9
#define init_8_0_5_9
void FASTCALL call_8_0_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_5_10
#define init_8_0_5_10
void FASTCALL call_8_0_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_5_11
#define init_8_0_5_11
void FASTCALL call_8_0_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_5_12
#define init_8_0_5_12
void FASTCALL call_8_0_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_5_13
#define init_8_0_5_13
void FASTCALL call_8_0_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_5_14
#define init_8_0_5_14
void FASTCALL call_8_0_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_5_15
#define init_8_0_5_15
void FASTCALL call_8_0_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_6_0
#define init_8_0_6_0
void FASTCALL call_8_0_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_6_1
#define init_8_0_6_1
void FASTCALL call_8_0_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_6_2
#define init_8_0_6_2
void FASTCALL call_8_0_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_6_3
#define init_8_0_6_3
void FASTCALL call_8_0_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_6_4
#define init_8_0_6_4
void FASTCALL call_8_0_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_6_5
#define init_8_0_6_5
void FASTCALL call_8_0_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_6_6
#define init_8_0_6_6
void FASTCALL call_8_0_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_6_7
#define init_8_0_6_7
void FASTCALL call_8_0_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_6_8
#define init_8_0_6_8
void FASTCALL call_8_0_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_6_9
#define init_8_0_6_9
void FASTCALL call_8_0_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_6_10
#define init_8_0_6_10
void FASTCALL call_8_0_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_6_11
#define init_8_0_6_11
void FASTCALL call_8_0_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_6_12
#define init_8_0_6_12
void FASTCALL call_8_0_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_6_13
#define init_8_0_6_13
void FASTCALL call_8_0_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_6_14
#define init_8_0_6_14
void FASTCALL call_8_0_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_6_15
#define init_8_0_6_15
void FASTCALL call_8_0_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_7_0
#define init_8_0_7_0
void FASTCALL call_8_0_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_7_1
#define init_8_0_7_1
void FASTCALL call_8_0_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_7_2
#define init_8_0_7_2
void FASTCALL call_8_0_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_7_3
#define init_8_0_7_3
void FASTCALL call_8_0_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_7_4
#define init_8_0_7_4
void FASTCALL call_8_0_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_7_5
#define init_8_0_7_5
void FASTCALL call_8_0_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_7_6
#define init_8_0_7_6
void FASTCALL call_8_0_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_7_7
#define init_8_0_7_7
void FASTCALL call_8_0_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_7_8
#define init_8_0_7_8
void FASTCALL call_8_0_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_7_9
#define init_8_0_7_9
void FASTCALL call_8_0_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_7_10
#define init_8_0_7_10
void FASTCALL call_8_0_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_7_11
#define init_8_0_7_11
void FASTCALL call_8_0_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_7_12
#define init_8_0_7_12
void FASTCALL call_8_0_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_7_13
#define init_8_0_7_13
void FASTCALL call_8_0_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_7_14
#define init_8_0_7_14
void FASTCALL call_8_0_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_7_15
#define init_8_0_7_15
void FASTCALL call_8_0_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_8_0
#define init_8_0_8_0
void FASTCALL call_8_0_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_8_1
#define init_8_0_8_1
void FASTCALL call_8_0_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_8_2
#define init_8_0_8_2
void FASTCALL call_8_0_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_8_3
#define init_8_0_8_3
void FASTCALL call_8_0_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_8_4
#define init_8_0_8_4
void FASTCALL call_8_0_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_8_5
#define init_8_0_8_5
void FASTCALL call_8_0_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_8_6
#define init_8_0_8_6
void FASTCALL call_8_0_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_8_7
#define init_8_0_8_7
void FASTCALL call_8_0_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_8_8
#define init_8_0_8_8
void FASTCALL call_8_0_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_8_9
#define init_8_0_8_9
void FASTCALL call_8_0_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_8_10
#define init_8_0_8_10
void FASTCALL call_8_0_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_8_11
#define init_8_0_8_11
void FASTCALL call_8_0_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_8_12
#define init_8_0_8_12
void FASTCALL call_8_0_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_8_13
#define init_8_0_8_13
void FASTCALL call_8_0_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_8_14
#define init_8_0_8_14
void FASTCALL call_8_0_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_8_15
#define init_8_0_8_15
void FASTCALL call_8_0_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_9_0
#define init_8_0_9_0
void FASTCALL call_8_0_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_9_1
#define init_8_0_9_1
void FASTCALL call_8_0_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_9_2
#define init_8_0_9_2
void FASTCALL call_8_0_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_9_3
#define init_8_0_9_3
void FASTCALL call_8_0_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_9_4
#define init_8_0_9_4
void FASTCALL call_8_0_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_9_5
#define init_8_0_9_5
void FASTCALL call_8_0_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_9_6
#define init_8_0_9_6
void FASTCALL call_8_0_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_9_7
#define init_8_0_9_7
void FASTCALL call_8_0_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_9_8
#define init_8_0_9_8
void FASTCALL call_8_0_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_9_9
#define init_8_0_9_9
void FASTCALL call_8_0_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_9_10
#define init_8_0_9_10
void FASTCALL call_8_0_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_9_11
#define init_8_0_9_11
void FASTCALL call_8_0_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_9_12
#define init_8_0_9_12
void FASTCALL call_8_0_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_9_13
#define init_8_0_9_13
void FASTCALL call_8_0_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_9_14
#define init_8_0_9_14
void FASTCALL call_8_0_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_9_15
#define init_8_0_9_15
void FASTCALL call_8_0_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_10_0
#define init_8_0_10_0
void FASTCALL call_8_0_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_10_1
#define init_8_0_10_1
void FASTCALL call_8_0_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_10_2
#define init_8_0_10_2
void FASTCALL call_8_0_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_10_3
#define init_8_0_10_3
void FASTCALL call_8_0_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_10_4
#define init_8_0_10_4
void FASTCALL call_8_0_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_10_5
#define init_8_0_10_5
void FASTCALL call_8_0_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_10_6
#define init_8_0_10_6
void FASTCALL call_8_0_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_10_7
#define init_8_0_10_7
void FASTCALL call_8_0_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_10_8
#define init_8_0_10_8
void FASTCALL call_8_0_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_10_9
#define init_8_0_10_9
void FASTCALL call_8_0_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_10_10
#define init_8_0_10_10
void FASTCALL call_8_0_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_10_11
#define init_8_0_10_11
void FASTCALL call_8_0_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_10_12
#define init_8_0_10_12
void FASTCALL call_8_0_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_10_13
#define init_8_0_10_13
void FASTCALL call_8_0_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_10_14
#define init_8_0_10_14
void FASTCALL call_8_0_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_10_15
#define init_8_0_10_15
void FASTCALL call_8_0_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_11_0
#define init_8_0_11_0
void FASTCALL call_8_0_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_11_1
#define init_8_0_11_1
void FASTCALL call_8_0_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_11_2
#define init_8_0_11_2
void FASTCALL call_8_0_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_11_3
#define init_8_0_11_3
void FASTCALL call_8_0_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_11_4
#define init_8_0_11_4
void FASTCALL call_8_0_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_11_5
#define init_8_0_11_5
void FASTCALL call_8_0_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_11_6
#define init_8_0_11_6
void FASTCALL call_8_0_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_11_7
#define init_8_0_11_7
void FASTCALL call_8_0_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_11_8
#define init_8_0_11_8
void FASTCALL call_8_0_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_11_9
#define init_8_0_11_9
void FASTCALL call_8_0_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_11_10
#define init_8_0_11_10
void FASTCALL call_8_0_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_11_11
#define init_8_0_11_11
void FASTCALL call_8_0_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_11_12
#define init_8_0_11_12
void FASTCALL call_8_0_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_11_13
#define init_8_0_11_13
void FASTCALL call_8_0_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_11_14
#define init_8_0_11_14
void FASTCALL call_8_0_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_11_15
#define init_8_0_11_15
void FASTCALL call_8_0_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_12_0
#define init_8_0_12_0
void FASTCALL call_8_0_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_12_1
#define init_8_0_12_1
void FASTCALL call_8_0_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_12_2
#define init_8_0_12_2
void FASTCALL call_8_0_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_12_3
#define init_8_0_12_3
void FASTCALL call_8_0_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_12_4
#define init_8_0_12_4
void FASTCALL call_8_0_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_12_5
#define init_8_0_12_5
void FASTCALL call_8_0_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_12_6
#define init_8_0_12_6
void FASTCALL call_8_0_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_12_7
#define init_8_0_12_7
void FASTCALL call_8_0_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_12_8
#define init_8_0_12_8
void FASTCALL call_8_0_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_12_9
#define init_8_0_12_9
void FASTCALL call_8_0_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_12_10
#define init_8_0_12_10
void FASTCALL call_8_0_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_12_11
#define init_8_0_12_11
void FASTCALL call_8_0_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_12_12
#define init_8_0_12_12
void FASTCALL call_8_0_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_12_13
#define init_8_0_12_13
void FASTCALL call_8_0_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_12_14
#define init_8_0_12_14
void FASTCALL call_8_0_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_12_15
#define init_8_0_12_15
void FASTCALL call_8_0_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_13_0
#define init_8_0_13_0
void FASTCALL call_8_0_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_13_1
#define init_8_0_13_1
void FASTCALL call_8_0_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_13_2
#define init_8_0_13_2
void FASTCALL call_8_0_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_13_3
#define init_8_0_13_3
void FASTCALL call_8_0_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_13_4
#define init_8_0_13_4
void FASTCALL call_8_0_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_13_5
#define init_8_0_13_5
void FASTCALL call_8_0_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_13_6
#define init_8_0_13_6
void FASTCALL call_8_0_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_13_7
#define init_8_0_13_7
void FASTCALL call_8_0_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_13_8
#define init_8_0_13_8
void FASTCALL call_8_0_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_13_9
#define init_8_0_13_9
void FASTCALL call_8_0_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_13_10
#define init_8_0_13_10
void FASTCALL call_8_0_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_13_11
#define init_8_0_13_11
void FASTCALL call_8_0_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_13_12
#define init_8_0_13_12
void FASTCALL call_8_0_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_13_13
#define init_8_0_13_13
void FASTCALL call_8_0_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_13_14
#define init_8_0_13_14
void FASTCALL call_8_0_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_13_15
#define init_8_0_13_15
void FASTCALL call_8_0_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_14_0
#define init_8_0_14_0
void FASTCALL call_8_0_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_14_1
#define init_8_0_14_1
void FASTCALL call_8_0_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_14_2
#define init_8_0_14_2
void FASTCALL call_8_0_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_14_3
#define init_8_0_14_3
void FASTCALL call_8_0_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_14_4
#define init_8_0_14_4
void FASTCALL call_8_0_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_14_5
#define init_8_0_14_5
void FASTCALL call_8_0_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_14_6
#define init_8_0_14_6
void FASTCALL call_8_0_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_14_7
#define init_8_0_14_7
void FASTCALL call_8_0_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_14_8
#define init_8_0_14_8
void FASTCALL call_8_0_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_14_9
#define init_8_0_14_9
void FASTCALL call_8_0_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_14_10
#define init_8_0_14_10
void FASTCALL call_8_0_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_14_11
#define init_8_0_14_11
void FASTCALL call_8_0_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_14_12
#define init_8_0_14_12
void FASTCALL call_8_0_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_14_13
#define init_8_0_14_13
void FASTCALL call_8_0_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_14_14
#define init_8_0_14_14
void FASTCALL call_8_0_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_14_15
#define init_8_0_14_15
void FASTCALL call_8_0_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_15_0
#define init_8_0_15_0
void FASTCALL call_8_0_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_15_1
#define init_8_0_15_1
void FASTCALL call_8_0_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_15_2
#define init_8_0_15_2
void FASTCALL call_8_0_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_15_3
#define init_8_0_15_3
void FASTCALL call_8_0_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_15_4
#define init_8_0_15_4
void FASTCALL call_8_0_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_15_5
#define init_8_0_15_5
void FASTCALL call_8_0_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_15_6
#define init_8_0_15_6
void FASTCALL call_8_0_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_15_7
#define init_8_0_15_7
void FASTCALL call_8_0_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_15_8
#define init_8_0_15_8
void FASTCALL call_8_0_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_15_9
#define init_8_0_15_9
void FASTCALL call_8_0_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_15_10
#define init_8_0_15_10
void FASTCALL call_8_0_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_15_11
#define init_8_0_15_11
void FASTCALL call_8_0_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_15_12
#define init_8_0_15_12
void FASTCALL call_8_0_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_15_13
#define init_8_0_15_13
void FASTCALL call_8_0_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_15_14
#define init_8_0_15_14
void FASTCALL call_8_0_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_0_15_15
#define init_8_0_15_15
void FASTCALL call_8_0_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_0_0
#define init_8_1_0_0
void FASTCALL call_8_1_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_0_1
#define init_8_1_0_1
void FASTCALL call_8_1_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_0_2
#define init_8_1_0_2
void FASTCALL call_8_1_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_0_3
#define init_8_1_0_3
void FASTCALL call_8_1_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_0_4
#define init_8_1_0_4
void FASTCALL call_8_1_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_0_5
#define init_8_1_0_5
void FASTCALL call_8_1_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_0_6
#define init_8_1_0_6
void FASTCALL call_8_1_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_0_7
#define init_8_1_0_7
void FASTCALL call_8_1_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_0_8
#define init_8_1_0_8
void FASTCALL call_8_1_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_0_9
#define init_8_1_0_9
void FASTCALL call_8_1_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_0_10
#define init_8_1_0_10
void FASTCALL call_8_1_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_0_11
#define init_8_1_0_11
void FASTCALL call_8_1_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_0_12
#define init_8_1_0_12
void FASTCALL call_8_1_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_0_13
#define init_8_1_0_13
void FASTCALL call_8_1_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_0_14
#define init_8_1_0_14
void FASTCALL call_8_1_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_0_15
#define init_8_1_0_15
void FASTCALL call_8_1_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_1_0
#define init_8_1_1_0
void FASTCALL call_8_1_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_1_1
#define init_8_1_1_1
void FASTCALL call_8_1_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_1_2
#define init_8_1_1_2
void FASTCALL call_8_1_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_1_3
#define init_8_1_1_3
void FASTCALL call_8_1_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_1_4
#define init_8_1_1_4
void FASTCALL call_8_1_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_1_5
#define init_8_1_1_5
void FASTCALL call_8_1_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_1_6
#define init_8_1_1_6
void FASTCALL call_8_1_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_1_7
#define init_8_1_1_7
void FASTCALL call_8_1_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_1_8
#define init_8_1_1_8
void FASTCALL call_8_1_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_1_9
#define init_8_1_1_9
void FASTCALL call_8_1_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_1_10
#define init_8_1_1_10
void FASTCALL call_8_1_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_1_11
#define init_8_1_1_11
void FASTCALL call_8_1_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_1_12
#define init_8_1_1_12
void FASTCALL call_8_1_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_1_13
#define init_8_1_1_13
void FASTCALL call_8_1_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_1_14
#define init_8_1_1_14
void FASTCALL call_8_1_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_1_15
#define init_8_1_1_15
void FASTCALL call_8_1_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_2_0
#define init_8_1_2_0
void FASTCALL call_8_1_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_2_1
#define init_8_1_2_1
void FASTCALL call_8_1_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_2_2
#define init_8_1_2_2
void FASTCALL call_8_1_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_2_3
#define init_8_1_2_3
void FASTCALL call_8_1_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_2_4
#define init_8_1_2_4
void FASTCALL call_8_1_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_2_5
#define init_8_1_2_5
void FASTCALL call_8_1_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_2_6
#define init_8_1_2_6
void FASTCALL call_8_1_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_2_7
#define init_8_1_2_7
void FASTCALL call_8_1_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_2_8
#define init_8_1_2_8
void FASTCALL call_8_1_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_2_9
#define init_8_1_2_9
void FASTCALL call_8_1_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_2_10
#define init_8_1_2_10
void FASTCALL call_8_1_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_2_11
#define init_8_1_2_11
void FASTCALL call_8_1_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_2_12
#define init_8_1_2_12
void FASTCALL call_8_1_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_2_13
#define init_8_1_2_13
void FASTCALL call_8_1_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_2_14
#define init_8_1_2_14
void FASTCALL call_8_1_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_2_15
#define init_8_1_2_15
void FASTCALL call_8_1_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_3_0
#define init_8_1_3_0
void FASTCALL call_8_1_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_3_1
#define init_8_1_3_1
void FASTCALL call_8_1_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_3_2
#define init_8_1_3_2
void FASTCALL call_8_1_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_3_3
#define init_8_1_3_3
void FASTCALL call_8_1_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_3_4
#define init_8_1_3_4
void FASTCALL call_8_1_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_3_5
#define init_8_1_3_5
void FASTCALL call_8_1_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_3_6
#define init_8_1_3_6
void FASTCALL call_8_1_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_3_7
#define init_8_1_3_7
void FASTCALL call_8_1_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_3_8
#define init_8_1_3_8
void FASTCALL call_8_1_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_3_9
#define init_8_1_3_9
void FASTCALL call_8_1_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_3_10
#define init_8_1_3_10
void FASTCALL call_8_1_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_3_11
#define init_8_1_3_11
void FASTCALL call_8_1_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_3_12
#define init_8_1_3_12
void FASTCALL call_8_1_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_3_13
#define init_8_1_3_13
void FASTCALL call_8_1_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_3_14
#define init_8_1_3_14
void FASTCALL call_8_1_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_3_15
#define init_8_1_3_15
void FASTCALL call_8_1_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_4_0
#define init_8_1_4_0
void FASTCALL call_8_1_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_4_1
#define init_8_1_4_1
void FASTCALL call_8_1_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_4_2
#define init_8_1_4_2
void FASTCALL call_8_1_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_4_3
#define init_8_1_4_3
void FASTCALL call_8_1_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_4_4
#define init_8_1_4_4
void FASTCALL call_8_1_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_4_5
#define init_8_1_4_5
void FASTCALL call_8_1_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_4_6
#define init_8_1_4_6
void FASTCALL call_8_1_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_4_7
#define init_8_1_4_7
void FASTCALL call_8_1_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_4_8
#define init_8_1_4_8
void FASTCALL call_8_1_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_4_9
#define init_8_1_4_9
void FASTCALL call_8_1_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_4_10
#define init_8_1_4_10
void FASTCALL call_8_1_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_4_11
#define init_8_1_4_11
void FASTCALL call_8_1_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_4_12
#define init_8_1_4_12
void FASTCALL call_8_1_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_4_13
#define init_8_1_4_13
void FASTCALL call_8_1_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_4_14
#define init_8_1_4_14
void FASTCALL call_8_1_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_4_15
#define init_8_1_4_15
void FASTCALL call_8_1_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_5_0
#define init_8_1_5_0
void FASTCALL call_8_1_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_5_1
#define init_8_1_5_1
void FASTCALL call_8_1_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_5_2
#define init_8_1_5_2
void FASTCALL call_8_1_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_5_3
#define init_8_1_5_3
void FASTCALL call_8_1_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_5_4
#define init_8_1_5_4
void FASTCALL call_8_1_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_5_5
#define init_8_1_5_5
void FASTCALL call_8_1_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_5_6
#define init_8_1_5_6
void FASTCALL call_8_1_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_5_7
#define init_8_1_5_7
void FASTCALL call_8_1_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_5_8
#define init_8_1_5_8
void FASTCALL call_8_1_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_5_9
#define init_8_1_5_9
void FASTCALL call_8_1_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_5_10
#define init_8_1_5_10
void FASTCALL call_8_1_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_5_11
#define init_8_1_5_11
void FASTCALL call_8_1_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_5_12
#define init_8_1_5_12
void FASTCALL call_8_1_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_5_13
#define init_8_1_5_13
void FASTCALL call_8_1_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_5_14
#define init_8_1_5_14
void FASTCALL call_8_1_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_5_15
#define init_8_1_5_15
void FASTCALL call_8_1_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_6_0
#define init_8_1_6_0
void FASTCALL call_8_1_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_6_1
#define init_8_1_6_1
void FASTCALL call_8_1_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_6_2
#define init_8_1_6_2
void FASTCALL call_8_1_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_6_3
#define init_8_1_6_3
void FASTCALL call_8_1_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_6_4
#define init_8_1_6_4
void FASTCALL call_8_1_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_6_5
#define init_8_1_6_5
void FASTCALL call_8_1_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_6_6
#define init_8_1_6_6
void FASTCALL call_8_1_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_6_7
#define init_8_1_6_7
void FASTCALL call_8_1_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_6_8
#define init_8_1_6_8
void FASTCALL call_8_1_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_6_9
#define init_8_1_6_9
void FASTCALL call_8_1_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_6_10
#define init_8_1_6_10
void FASTCALL call_8_1_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_6_11
#define init_8_1_6_11
void FASTCALL call_8_1_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_6_12
#define init_8_1_6_12
void FASTCALL call_8_1_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_6_13
#define init_8_1_6_13
void FASTCALL call_8_1_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_6_14
#define init_8_1_6_14
void FASTCALL call_8_1_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_6_15
#define init_8_1_6_15
void FASTCALL call_8_1_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_7_0
#define init_8_1_7_0
void FASTCALL call_8_1_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_7_1
#define init_8_1_7_1
void FASTCALL call_8_1_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_7_2
#define init_8_1_7_2
void FASTCALL call_8_1_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_7_3
#define init_8_1_7_3
void FASTCALL call_8_1_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_7_4
#define init_8_1_7_4
void FASTCALL call_8_1_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_7_5
#define init_8_1_7_5
void FASTCALL call_8_1_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_7_6
#define init_8_1_7_6
void FASTCALL call_8_1_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_7_7
#define init_8_1_7_7
void FASTCALL call_8_1_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_7_8
#define init_8_1_7_8
void FASTCALL call_8_1_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_7_9
#define init_8_1_7_9
void FASTCALL call_8_1_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_7_10
#define init_8_1_7_10
void FASTCALL call_8_1_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_7_11
#define init_8_1_7_11
void FASTCALL call_8_1_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_7_12
#define init_8_1_7_12
void FASTCALL call_8_1_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_7_13
#define init_8_1_7_13
void FASTCALL call_8_1_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_7_14
#define init_8_1_7_14
void FASTCALL call_8_1_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_7_15
#define init_8_1_7_15
void FASTCALL call_8_1_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_8_0
#define init_8_1_8_0
void FASTCALL call_8_1_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_8_1
#define init_8_1_8_1
void FASTCALL call_8_1_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_8_2
#define init_8_1_8_2
void FASTCALL call_8_1_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_8_3
#define init_8_1_8_3
void FASTCALL call_8_1_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_8_4
#define init_8_1_8_4
void FASTCALL call_8_1_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_8_5
#define init_8_1_8_5
void FASTCALL call_8_1_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_8_6
#define init_8_1_8_6
void FASTCALL call_8_1_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_8_7
#define init_8_1_8_7
void FASTCALL call_8_1_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_8_8
#define init_8_1_8_8
void FASTCALL call_8_1_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_8_9
#define init_8_1_8_9
void FASTCALL call_8_1_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_8_10
#define init_8_1_8_10
void FASTCALL call_8_1_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_8_11
#define init_8_1_8_11
void FASTCALL call_8_1_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_8_12
#define init_8_1_8_12
void FASTCALL call_8_1_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_8_13
#define init_8_1_8_13
void FASTCALL call_8_1_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_8_14
#define init_8_1_8_14
void FASTCALL call_8_1_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_8_15
#define init_8_1_8_15
void FASTCALL call_8_1_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_9_0
#define init_8_1_9_0
void FASTCALL call_8_1_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_9_1
#define init_8_1_9_1
void FASTCALL call_8_1_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_9_2
#define init_8_1_9_2
void FASTCALL call_8_1_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_9_3
#define init_8_1_9_3
void FASTCALL call_8_1_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_9_4
#define init_8_1_9_4
void FASTCALL call_8_1_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_9_5
#define init_8_1_9_5
void FASTCALL call_8_1_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_9_6
#define init_8_1_9_6
void FASTCALL call_8_1_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_9_7
#define init_8_1_9_7
void FASTCALL call_8_1_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_9_8
#define init_8_1_9_8
void FASTCALL call_8_1_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_9_9
#define init_8_1_9_9
void FASTCALL call_8_1_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_9_10
#define init_8_1_9_10
void FASTCALL call_8_1_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_9_11
#define init_8_1_9_11
void FASTCALL call_8_1_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_9_12
#define init_8_1_9_12
void FASTCALL call_8_1_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_9_13
#define init_8_1_9_13
void FASTCALL call_8_1_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_9_14
#define init_8_1_9_14
void FASTCALL call_8_1_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_9_15
#define init_8_1_9_15
void FASTCALL call_8_1_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_10_0
#define init_8_1_10_0
void FASTCALL call_8_1_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_10_1
#define init_8_1_10_1
void FASTCALL call_8_1_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_10_2
#define init_8_1_10_2
void FASTCALL call_8_1_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_10_3
#define init_8_1_10_3
void FASTCALL call_8_1_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_10_4
#define init_8_1_10_4
void FASTCALL call_8_1_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_10_5
#define init_8_1_10_5
void FASTCALL call_8_1_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_10_6
#define init_8_1_10_6
void FASTCALL call_8_1_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_10_7
#define init_8_1_10_7
void FASTCALL call_8_1_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_10_8
#define init_8_1_10_8
void FASTCALL call_8_1_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_10_9
#define init_8_1_10_9
void FASTCALL call_8_1_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_10_10
#define init_8_1_10_10
void FASTCALL call_8_1_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_10_11
#define init_8_1_10_11
void FASTCALL call_8_1_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_10_12
#define init_8_1_10_12
void FASTCALL call_8_1_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_10_13
#define init_8_1_10_13
void FASTCALL call_8_1_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_10_14
#define init_8_1_10_14
void FASTCALL call_8_1_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_10_15
#define init_8_1_10_15
void FASTCALL call_8_1_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_11_0
#define init_8_1_11_0
void FASTCALL call_8_1_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_11_1
#define init_8_1_11_1
void FASTCALL call_8_1_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_11_2
#define init_8_1_11_2
void FASTCALL call_8_1_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_11_3
#define init_8_1_11_3
void FASTCALL call_8_1_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_11_4
#define init_8_1_11_4
void FASTCALL call_8_1_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_11_5
#define init_8_1_11_5
void FASTCALL call_8_1_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_11_6
#define init_8_1_11_6
void FASTCALL call_8_1_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_11_7
#define init_8_1_11_7
void FASTCALL call_8_1_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_11_8
#define init_8_1_11_8
void FASTCALL call_8_1_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_11_9
#define init_8_1_11_9
void FASTCALL call_8_1_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_11_10
#define init_8_1_11_10
void FASTCALL call_8_1_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_11_11
#define init_8_1_11_11
void FASTCALL call_8_1_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_11_12
#define init_8_1_11_12
void FASTCALL call_8_1_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_11_13
#define init_8_1_11_13
void FASTCALL call_8_1_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_11_14
#define init_8_1_11_14
void FASTCALL call_8_1_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_11_15
#define init_8_1_11_15
void FASTCALL call_8_1_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_12_0
#define init_8_1_12_0
void FASTCALL call_8_1_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_12_1
#define init_8_1_12_1
void FASTCALL call_8_1_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_12_2
#define init_8_1_12_2
void FASTCALL call_8_1_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_12_3
#define init_8_1_12_3
void FASTCALL call_8_1_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_12_4
#define init_8_1_12_4
void FASTCALL call_8_1_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_12_5
#define init_8_1_12_5
void FASTCALL call_8_1_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_12_6
#define init_8_1_12_6
void FASTCALL call_8_1_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_12_7
#define init_8_1_12_7
void FASTCALL call_8_1_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_12_8
#define init_8_1_12_8
void FASTCALL call_8_1_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_12_9
#define init_8_1_12_9
void FASTCALL call_8_1_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_12_10
#define init_8_1_12_10
void FASTCALL call_8_1_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_12_11
#define init_8_1_12_11
void FASTCALL call_8_1_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_12_12
#define init_8_1_12_12
void FASTCALL call_8_1_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_12_13
#define init_8_1_12_13
void FASTCALL call_8_1_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_12_14
#define init_8_1_12_14
void FASTCALL call_8_1_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_12_15
#define init_8_1_12_15
void FASTCALL call_8_1_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_13_0
#define init_8_1_13_0
void FASTCALL call_8_1_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_13_1
#define init_8_1_13_1
void FASTCALL call_8_1_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_13_2
#define init_8_1_13_2
void FASTCALL call_8_1_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_13_3
#define init_8_1_13_3
void FASTCALL call_8_1_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_13_4
#define init_8_1_13_4
void FASTCALL call_8_1_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_13_5
#define init_8_1_13_5
void FASTCALL call_8_1_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_13_6
#define init_8_1_13_6
void FASTCALL call_8_1_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_13_7
#define init_8_1_13_7
void FASTCALL call_8_1_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_13_8
#define init_8_1_13_8
void FASTCALL call_8_1_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_13_9
#define init_8_1_13_9
void FASTCALL call_8_1_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_13_10
#define init_8_1_13_10
void FASTCALL call_8_1_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_13_11
#define init_8_1_13_11
void FASTCALL call_8_1_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_13_12
#define init_8_1_13_12
void FASTCALL call_8_1_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_13_13
#define init_8_1_13_13
void FASTCALL call_8_1_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_13_14
#define init_8_1_13_14
void FASTCALL call_8_1_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_13_15
#define init_8_1_13_15
void FASTCALL call_8_1_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_14_0
#define init_8_1_14_0
void FASTCALL call_8_1_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_14_1
#define init_8_1_14_1
void FASTCALL call_8_1_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_14_2
#define init_8_1_14_2
void FASTCALL call_8_1_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_14_3
#define init_8_1_14_3
void FASTCALL call_8_1_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_14_4
#define init_8_1_14_4
void FASTCALL call_8_1_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_14_5
#define init_8_1_14_5
void FASTCALL call_8_1_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_14_6
#define init_8_1_14_6
void FASTCALL call_8_1_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_14_7
#define init_8_1_14_7
void FASTCALL call_8_1_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_14_8
#define init_8_1_14_8
void FASTCALL call_8_1_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_14_9
#define init_8_1_14_9
void FASTCALL call_8_1_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_14_10
#define init_8_1_14_10
void FASTCALL call_8_1_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_14_11
#define init_8_1_14_11
void FASTCALL call_8_1_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_14_12
#define init_8_1_14_12
void FASTCALL call_8_1_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_14_13
#define init_8_1_14_13
void FASTCALL call_8_1_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_14_14
#define init_8_1_14_14
void FASTCALL call_8_1_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_14_15
#define init_8_1_14_15
void FASTCALL call_8_1_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_15_0
#define init_8_1_15_0
void FASTCALL call_8_1_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_15_1
#define init_8_1_15_1
void FASTCALL call_8_1_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_15_2
#define init_8_1_15_2
void FASTCALL call_8_1_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_15_3
#define init_8_1_15_3
void FASTCALL call_8_1_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_15_4
#define init_8_1_15_4
void FASTCALL call_8_1_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_15_5
#define init_8_1_15_5
void FASTCALL call_8_1_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_15_6
#define init_8_1_15_6
void FASTCALL call_8_1_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_15_7
#define init_8_1_15_7
void FASTCALL call_8_1_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_15_8
#define init_8_1_15_8
void FASTCALL call_8_1_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_15_9
#define init_8_1_15_9
void FASTCALL call_8_1_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_15_10
#define init_8_1_15_10
void FASTCALL call_8_1_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_15_11
#define init_8_1_15_11
void FASTCALL call_8_1_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_15_12
#define init_8_1_15_12
void FASTCALL call_8_1_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_15_13
#define init_8_1_15_13
void FASTCALL call_8_1_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_15_14
#define init_8_1_15_14
void FASTCALL call_8_1_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_1_15_15
#define init_8_1_15_15
void FASTCALL call_8_1_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_0_0
#define init_8_2_0_0
void FASTCALL call_8_2_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_0_1
#define init_8_2_0_1
void FASTCALL call_8_2_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_0_2
#define init_8_2_0_2
void FASTCALL call_8_2_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_0_3
#define init_8_2_0_3
void FASTCALL call_8_2_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_0_4
#define init_8_2_0_4
void FASTCALL call_8_2_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_0_5
#define init_8_2_0_5
void FASTCALL call_8_2_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_0_6
#define init_8_2_0_6
void FASTCALL call_8_2_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_0_7
#define init_8_2_0_7
void FASTCALL call_8_2_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_0_8
#define init_8_2_0_8
void FASTCALL call_8_2_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_0_9
#define init_8_2_0_9
void FASTCALL call_8_2_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_0_10
#define init_8_2_0_10
void FASTCALL call_8_2_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_0_11
#define init_8_2_0_11
void FASTCALL call_8_2_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_0_12
#define init_8_2_0_12
void FASTCALL call_8_2_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_0_13
#define init_8_2_0_13
void FASTCALL call_8_2_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_0_14
#define init_8_2_0_14
void FASTCALL call_8_2_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_0_15
#define init_8_2_0_15
void FASTCALL call_8_2_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_1_0
#define init_8_2_1_0
void FASTCALL call_8_2_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_1_1
#define init_8_2_1_1
void FASTCALL call_8_2_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_1_2
#define init_8_2_1_2
void FASTCALL call_8_2_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_1_3
#define init_8_2_1_3
void FASTCALL call_8_2_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_1_4
#define init_8_2_1_4
void FASTCALL call_8_2_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_1_5
#define init_8_2_1_5
void FASTCALL call_8_2_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_1_6
#define init_8_2_1_6
void FASTCALL call_8_2_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_1_7
#define init_8_2_1_7
void FASTCALL call_8_2_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_1_8
#define init_8_2_1_8
void FASTCALL call_8_2_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_1_9
#define init_8_2_1_9
void FASTCALL call_8_2_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_1_10
#define init_8_2_1_10
void FASTCALL call_8_2_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_1_11
#define init_8_2_1_11
void FASTCALL call_8_2_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_1_12
#define init_8_2_1_12
void FASTCALL call_8_2_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_1_13
#define init_8_2_1_13
void FASTCALL call_8_2_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_1_14
#define init_8_2_1_14
void FASTCALL call_8_2_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_1_15
#define init_8_2_1_15
void FASTCALL call_8_2_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_2_0
#define init_8_2_2_0
void FASTCALL call_8_2_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_2_1
#define init_8_2_2_1
void FASTCALL call_8_2_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_2_2
#define init_8_2_2_2
void FASTCALL call_8_2_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_2_3
#define init_8_2_2_3
void FASTCALL call_8_2_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_2_4
#define init_8_2_2_4
void FASTCALL call_8_2_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_2_5
#define init_8_2_2_5
void FASTCALL call_8_2_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_2_6
#define init_8_2_2_6
void FASTCALL call_8_2_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_2_7
#define init_8_2_2_7
void FASTCALL call_8_2_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_2_8
#define init_8_2_2_8
void FASTCALL call_8_2_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_2_9
#define init_8_2_2_9
void FASTCALL call_8_2_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_2_10
#define init_8_2_2_10
void FASTCALL call_8_2_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_2_11
#define init_8_2_2_11
void FASTCALL call_8_2_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_2_12
#define init_8_2_2_12
void FASTCALL call_8_2_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_2_13
#define init_8_2_2_13
void FASTCALL call_8_2_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_2_14
#define init_8_2_2_14
void FASTCALL call_8_2_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_2_15
#define init_8_2_2_15
void FASTCALL call_8_2_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_3_0
#define init_8_2_3_0
void FASTCALL call_8_2_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_3_1
#define init_8_2_3_1
void FASTCALL call_8_2_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_3_2
#define init_8_2_3_2
void FASTCALL call_8_2_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_3_3
#define init_8_2_3_3
void FASTCALL call_8_2_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_3_4
#define init_8_2_3_4
void FASTCALL call_8_2_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_3_5
#define init_8_2_3_5
void FASTCALL call_8_2_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_3_6
#define init_8_2_3_6
void FASTCALL call_8_2_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_3_7
#define init_8_2_3_7
void FASTCALL call_8_2_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_3_8
#define init_8_2_3_8
void FASTCALL call_8_2_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_3_9
#define init_8_2_3_9
void FASTCALL call_8_2_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_3_10
#define init_8_2_3_10
void FASTCALL call_8_2_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_3_11
#define init_8_2_3_11
void FASTCALL call_8_2_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_3_12
#define init_8_2_3_12
void FASTCALL call_8_2_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_3_13
#define init_8_2_3_13
void FASTCALL call_8_2_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_3_14
#define init_8_2_3_14
void FASTCALL call_8_2_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_3_15
#define init_8_2_3_15
void FASTCALL call_8_2_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_4_0
#define init_8_2_4_0
void FASTCALL call_8_2_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_4_1
#define init_8_2_4_1
void FASTCALL call_8_2_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_4_2
#define init_8_2_4_2
void FASTCALL call_8_2_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_4_3
#define init_8_2_4_3
void FASTCALL call_8_2_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_4_4
#define init_8_2_4_4
void FASTCALL call_8_2_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_4_5
#define init_8_2_4_5
void FASTCALL call_8_2_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_4_6
#define init_8_2_4_6
void FASTCALL call_8_2_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_4_7
#define init_8_2_4_7
void FASTCALL call_8_2_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_4_8
#define init_8_2_4_8
void FASTCALL call_8_2_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_4_9
#define init_8_2_4_9
void FASTCALL call_8_2_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_4_10
#define init_8_2_4_10
void FASTCALL call_8_2_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_4_11
#define init_8_2_4_11
void FASTCALL call_8_2_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_4_12
#define init_8_2_4_12
void FASTCALL call_8_2_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_4_13
#define init_8_2_4_13
void FASTCALL call_8_2_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_4_14
#define init_8_2_4_14
void FASTCALL call_8_2_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_4_15
#define init_8_2_4_15
void FASTCALL call_8_2_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_5_0
#define init_8_2_5_0
void FASTCALL call_8_2_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_5_1
#define init_8_2_5_1
void FASTCALL call_8_2_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_5_2
#define init_8_2_5_2
void FASTCALL call_8_2_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_5_3
#define init_8_2_5_3
void FASTCALL call_8_2_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_5_4
#define init_8_2_5_4
void FASTCALL call_8_2_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_5_5
#define init_8_2_5_5
void FASTCALL call_8_2_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_5_6
#define init_8_2_5_6
void FASTCALL call_8_2_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_5_7
#define init_8_2_5_7
void FASTCALL call_8_2_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_5_8
#define init_8_2_5_8
void FASTCALL call_8_2_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_5_9
#define init_8_2_5_9
void FASTCALL call_8_2_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_5_10
#define init_8_2_5_10
void FASTCALL call_8_2_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_5_11
#define init_8_2_5_11
void FASTCALL call_8_2_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_5_12
#define init_8_2_5_12
void FASTCALL call_8_2_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_5_13
#define init_8_2_5_13
void FASTCALL call_8_2_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_5_14
#define init_8_2_5_14
void FASTCALL call_8_2_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_5_15
#define init_8_2_5_15
void FASTCALL call_8_2_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_6_0
#define init_8_2_6_0
void FASTCALL call_8_2_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_6_1
#define init_8_2_6_1
void FASTCALL call_8_2_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_6_2
#define init_8_2_6_2
void FASTCALL call_8_2_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_6_3
#define init_8_2_6_3
void FASTCALL call_8_2_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_6_4
#define init_8_2_6_4
void FASTCALL call_8_2_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_6_5
#define init_8_2_6_5
void FASTCALL call_8_2_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_6_6
#define init_8_2_6_6
void FASTCALL call_8_2_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_6_7
#define init_8_2_6_7
void FASTCALL call_8_2_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_6_8
#define init_8_2_6_8
void FASTCALL call_8_2_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_6_9
#define init_8_2_6_9
void FASTCALL call_8_2_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_6_10
#define init_8_2_6_10
void FASTCALL call_8_2_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_6_11
#define init_8_2_6_11
void FASTCALL call_8_2_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_6_12
#define init_8_2_6_12
void FASTCALL call_8_2_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_6_13
#define init_8_2_6_13
void FASTCALL call_8_2_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_6_14
#define init_8_2_6_14
void FASTCALL call_8_2_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_6_15
#define init_8_2_6_15
void FASTCALL call_8_2_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_7_0
#define init_8_2_7_0
void FASTCALL call_8_2_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_7_1
#define init_8_2_7_1
void FASTCALL call_8_2_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_7_2
#define init_8_2_7_2
void FASTCALL call_8_2_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_7_3
#define init_8_2_7_3
void FASTCALL call_8_2_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_7_4
#define init_8_2_7_4
void FASTCALL call_8_2_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_7_5
#define init_8_2_7_5
void FASTCALL call_8_2_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_7_6
#define init_8_2_7_6
void FASTCALL call_8_2_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_7_7
#define init_8_2_7_7
void FASTCALL call_8_2_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_7_8
#define init_8_2_7_8
void FASTCALL call_8_2_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_7_9
#define init_8_2_7_9
void FASTCALL call_8_2_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_7_10
#define init_8_2_7_10
void FASTCALL call_8_2_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_7_11
#define init_8_2_7_11
void FASTCALL call_8_2_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_7_12
#define init_8_2_7_12
void FASTCALL call_8_2_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_7_13
#define init_8_2_7_13
void FASTCALL call_8_2_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_7_14
#define init_8_2_7_14
void FASTCALL call_8_2_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_7_15
#define init_8_2_7_15
void FASTCALL call_8_2_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_8_0
#define init_8_2_8_0
void FASTCALL call_8_2_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_8_1
#define init_8_2_8_1
void FASTCALL call_8_2_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_8_2
#define init_8_2_8_2
void FASTCALL call_8_2_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_8_3
#define init_8_2_8_3
void FASTCALL call_8_2_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_8_4
#define init_8_2_8_4
void FASTCALL call_8_2_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_8_5
#define init_8_2_8_5
void FASTCALL call_8_2_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_8_6
#define init_8_2_8_6
void FASTCALL call_8_2_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_8_7
#define init_8_2_8_7
void FASTCALL call_8_2_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_8_8
#define init_8_2_8_8
void FASTCALL call_8_2_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_8_9
#define init_8_2_8_9
void FASTCALL call_8_2_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_8_10
#define init_8_2_8_10
void FASTCALL call_8_2_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_8_11
#define init_8_2_8_11
void FASTCALL call_8_2_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_8_12
#define init_8_2_8_12
void FASTCALL call_8_2_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_8_13
#define init_8_2_8_13
void FASTCALL call_8_2_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_8_14
#define init_8_2_8_14
void FASTCALL call_8_2_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_8_15
#define init_8_2_8_15
void FASTCALL call_8_2_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_9_0
#define init_8_2_9_0
void FASTCALL call_8_2_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_9_1
#define init_8_2_9_1
void FASTCALL call_8_2_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_9_2
#define init_8_2_9_2
void FASTCALL call_8_2_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_9_3
#define init_8_2_9_3
void FASTCALL call_8_2_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_9_4
#define init_8_2_9_4
void FASTCALL call_8_2_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_9_5
#define init_8_2_9_5
void FASTCALL call_8_2_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_9_6
#define init_8_2_9_6
void FASTCALL call_8_2_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_9_7
#define init_8_2_9_7
void FASTCALL call_8_2_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_9_8
#define init_8_2_9_8
void FASTCALL call_8_2_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_9_9
#define init_8_2_9_9
void FASTCALL call_8_2_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_9_10
#define init_8_2_9_10
void FASTCALL call_8_2_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_9_11
#define init_8_2_9_11
void FASTCALL call_8_2_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_9_12
#define init_8_2_9_12
void FASTCALL call_8_2_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_9_13
#define init_8_2_9_13
void FASTCALL call_8_2_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_9_14
#define init_8_2_9_14
void FASTCALL call_8_2_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_9_15
#define init_8_2_9_15
void FASTCALL call_8_2_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_10_0
#define init_8_2_10_0
void FASTCALL call_8_2_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_10_1
#define init_8_2_10_1
void FASTCALL call_8_2_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_10_2
#define init_8_2_10_2
void FASTCALL call_8_2_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_10_3
#define init_8_2_10_3
void FASTCALL call_8_2_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_10_4
#define init_8_2_10_4
void FASTCALL call_8_2_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_10_5
#define init_8_2_10_5
void FASTCALL call_8_2_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_10_6
#define init_8_2_10_6
void FASTCALL call_8_2_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_10_7
#define init_8_2_10_7
void FASTCALL call_8_2_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_10_8
#define init_8_2_10_8
void FASTCALL call_8_2_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_10_9
#define init_8_2_10_9
void FASTCALL call_8_2_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_10_10
#define init_8_2_10_10
void FASTCALL call_8_2_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_10_11
#define init_8_2_10_11
void FASTCALL call_8_2_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_10_12
#define init_8_2_10_12
void FASTCALL call_8_2_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_10_13
#define init_8_2_10_13
void FASTCALL call_8_2_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_10_14
#define init_8_2_10_14
void FASTCALL call_8_2_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_10_15
#define init_8_2_10_15
void FASTCALL call_8_2_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_11_0
#define init_8_2_11_0
void FASTCALL call_8_2_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_11_1
#define init_8_2_11_1
void FASTCALL call_8_2_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_11_2
#define init_8_2_11_2
void FASTCALL call_8_2_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_11_3
#define init_8_2_11_3
void FASTCALL call_8_2_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_11_4
#define init_8_2_11_4
void FASTCALL call_8_2_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_11_5
#define init_8_2_11_5
void FASTCALL call_8_2_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_11_6
#define init_8_2_11_6
void FASTCALL call_8_2_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_11_7
#define init_8_2_11_7
void FASTCALL call_8_2_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_11_8
#define init_8_2_11_8
void FASTCALL call_8_2_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_11_9
#define init_8_2_11_9
void FASTCALL call_8_2_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_11_10
#define init_8_2_11_10
void FASTCALL call_8_2_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_11_11
#define init_8_2_11_11
void FASTCALL call_8_2_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_11_12
#define init_8_2_11_12
void FASTCALL call_8_2_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_11_13
#define init_8_2_11_13
void FASTCALL call_8_2_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_11_14
#define init_8_2_11_14
void FASTCALL call_8_2_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_11_15
#define init_8_2_11_15
void FASTCALL call_8_2_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_12_0
#define init_8_2_12_0
void FASTCALL call_8_2_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_12_1
#define init_8_2_12_1
void FASTCALL call_8_2_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_12_2
#define init_8_2_12_2
void FASTCALL call_8_2_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_12_3
#define init_8_2_12_3
void FASTCALL call_8_2_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_12_4
#define init_8_2_12_4
void FASTCALL call_8_2_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_12_5
#define init_8_2_12_5
void FASTCALL call_8_2_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_12_6
#define init_8_2_12_6
void FASTCALL call_8_2_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_12_7
#define init_8_2_12_7
void FASTCALL call_8_2_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_12_8
#define init_8_2_12_8
void FASTCALL call_8_2_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_12_9
#define init_8_2_12_9
void FASTCALL call_8_2_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_12_10
#define init_8_2_12_10
void FASTCALL call_8_2_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_12_11
#define init_8_2_12_11
void FASTCALL call_8_2_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_12_12
#define init_8_2_12_12
void FASTCALL call_8_2_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_12_13
#define init_8_2_12_13
void FASTCALL call_8_2_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_12_14
#define init_8_2_12_14
void FASTCALL call_8_2_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_12_15
#define init_8_2_12_15
void FASTCALL call_8_2_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_13_0
#define init_8_2_13_0
void FASTCALL call_8_2_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_13_1
#define init_8_2_13_1
void FASTCALL call_8_2_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_13_2
#define init_8_2_13_2
void FASTCALL call_8_2_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_13_3
#define init_8_2_13_3
void FASTCALL call_8_2_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_13_4
#define init_8_2_13_4
void FASTCALL call_8_2_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_13_5
#define init_8_2_13_5
void FASTCALL call_8_2_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_13_6
#define init_8_2_13_6
void FASTCALL call_8_2_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_13_7
#define init_8_2_13_7
void FASTCALL call_8_2_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_13_8
#define init_8_2_13_8
void FASTCALL call_8_2_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_13_9
#define init_8_2_13_9
void FASTCALL call_8_2_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_13_10
#define init_8_2_13_10
void FASTCALL call_8_2_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_13_11
#define init_8_2_13_11
void FASTCALL call_8_2_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_13_12
#define init_8_2_13_12
void FASTCALL call_8_2_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_13_13
#define init_8_2_13_13
void FASTCALL call_8_2_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_13_14
#define init_8_2_13_14
void FASTCALL call_8_2_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_13_15
#define init_8_2_13_15
void FASTCALL call_8_2_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_14_0
#define init_8_2_14_0
void FASTCALL call_8_2_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_14_1
#define init_8_2_14_1
void FASTCALL call_8_2_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_14_2
#define init_8_2_14_2
void FASTCALL call_8_2_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_14_3
#define init_8_2_14_3
void FASTCALL call_8_2_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_14_4
#define init_8_2_14_4
void FASTCALL call_8_2_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_14_5
#define init_8_2_14_5
void FASTCALL call_8_2_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_14_6
#define init_8_2_14_6
void FASTCALL call_8_2_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_14_7
#define init_8_2_14_7
void FASTCALL call_8_2_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_14_8
#define init_8_2_14_8
void FASTCALL call_8_2_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_14_9
#define init_8_2_14_9
void FASTCALL call_8_2_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_14_10
#define init_8_2_14_10
void FASTCALL call_8_2_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_14_11
#define init_8_2_14_11
void FASTCALL call_8_2_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_14_12
#define init_8_2_14_12
void FASTCALL call_8_2_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_14_13
#define init_8_2_14_13
void FASTCALL call_8_2_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_14_14
#define init_8_2_14_14
void FASTCALL call_8_2_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_14_15
#define init_8_2_14_15
void FASTCALL call_8_2_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_15_0
#define init_8_2_15_0
void FASTCALL call_8_2_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_15_1
#define init_8_2_15_1
void FASTCALL call_8_2_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_15_2
#define init_8_2_15_2
void FASTCALL call_8_2_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_15_3
#define init_8_2_15_3
void FASTCALL call_8_2_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_15_4
#define init_8_2_15_4
void FASTCALL call_8_2_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_15_5
#define init_8_2_15_5
void FASTCALL call_8_2_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_15_6
#define init_8_2_15_6
void FASTCALL call_8_2_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_15_7
#define init_8_2_15_7
void FASTCALL call_8_2_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_15_8
#define init_8_2_15_8
void FASTCALL call_8_2_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_15_9
#define init_8_2_15_9
void FASTCALL call_8_2_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_15_10
#define init_8_2_15_10
void FASTCALL call_8_2_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_15_11
#define init_8_2_15_11
void FASTCALL call_8_2_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_15_12
#define init_8_2_15_12
void FASTCALL call_8_2_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_15_13
#define init_8_2_15_13
void FASTCALL call_8_2_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_15_14
#define init_8_2_15_14
void FASTCALL call_8_2_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_2_15_15
#define init_8_2_15_15
void FASTCALL call_8_2_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_0_0
#define init_8_3_0_0
void FASTCALL call_8_3_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_0_1
#define init_8_3_0_1
void FASTCALL call_8_3_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_0_2
#define init_8_3_0_2
void FASTCALL call_8_3_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_0_3
#define init_8_3_0_3
void FASTCALL call_8_3_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_0_4
#define init_8_3_0_4
void FASTCALL call_8_3_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_0_5
#define init_8_3_0_5
void FASTCALL call_8_3_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_0_6
#define init_8_3_0_6
void FASTCALL call_8_3_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_0_7
#define init_8_3_0_7
void FASTCALL call_8_3_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_0_8
#define init_8_3_0_8
void FASTCALL call_8_3_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_0_9
#define init_8_3_0_9
void FASTCALL call_8_3_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_0_10
#define init_8_3_0_10
void FASTCALL call_8_3_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_0_11
#define init_8_3_0_11
void FASTCALL call_8_3_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_0_12
#define init_8_3_0_12
void FASTCALL call_8_3_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_0_13
#define init_8_3_0_13
void FASTCALL call_8_3_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_0_14
#define init_8_3_0_14
void FASTCALL call_8_3_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_0_15
#define init_8_3_0_15
void FASTCALL call_8_3_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_1_0
#define init_8_3_1_0
void FASTCALL call_8_3_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_1_1
#define init_8_3_1_1
void FASTCALL call_8_3_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_1_2
#define init_8_3_1_2
void FASTCALL call_8_3_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_1_3
#define init_8_3_1_3
void FASTCALL call_8_3_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_1_4
#define init_8_3_1_4
void FASTCALL call_8_3_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_1_5
#define init_8_3_1_5
void FASTCALL call_8_3_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_1_6
#define init_8_3_1_6
void FASTCALL call_8_3_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_1_7
#define init_8_3_1_7
void FASTCALL call_8_3_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_1_8
#define init_8_3_1_8
void FASTCALL call_8_3_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_1_9
#define init_8_3_1_9
void FASTCALL call_8_3_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_1_10
#define init_8_3_1_10
void FASTCALL call_8_3_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_1_11
#define init_8_3_1_11
void FASTCALL call_8_3_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_1_12
#define init_8_3_1_12
void FASTCALL call_8_3_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_1_13
#define init_8_3_1_13
void FASTCALL call_8_3_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_1_14
#define init_8_3_1_14
void FASTCALL call_8_3_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_1_15
#define init_8_3_1_15
void FASTCALL call_8_3_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_2_0
#define init_8_3_2_0
void FASTCALL call_8_3_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_2_1
#define init_8_3_2_1
void FASTCALL call_8_3_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_2_2
#define init_8_3_2_2
void FASTCALL call_8_3_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_2_3
#define init_8_3_2_3
void FASTCALL call_8_3_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_2_4
#define init_8_3_2_4
void FASTCALL call_8_3_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_2_5
#define init_8_3_2_5
void FASTCALL call_8_3_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_2_6
#define init_8_3_2_6
void FASTCALL call_8_3_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_2_7
#define init_8_3_2_7
void FASTCALL call_8_3_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_2_8
#define init_8_3_2_8
void FASTCALL call_8_3_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_2_9
#define init_8_3_2_9
void FASTCALL call_8_3_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_2_10
#define init_8_3_2_10
void FASTCALL call_8_3_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_2_11
#define init_8_3_2_11
void FASTCALL call_8_3_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_2_12
#define init_8_3_2_12
void FASTCALL call_8_3_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_2_13
#define init_8_3_2_13
void FASTCALL call_8_3_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_2_14
#define init_8_3_2_14
void FASTCALL call_8_3_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_2_15
#define init_8_3_2_15
void FASTCALL call_8_3_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_3_0
#define init_8_3_3_0
void FASTCALL call_8_3_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_3_1
#define init_8_3_3_1
void FASTCALL call_8_3_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_3_2
#define init_8_3_3_2
void FASTCALL call_8_3_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_3_3
#define init_8_3_3_3
void FASTCALL call_8_3_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_3_4
#define init_8_3_3_4
void FASTCALL call_8_3_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_3_5
#define init_8_3_3_5
void FASTCALL call_8_3_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_3_6
#define init_8_3_3_6
void FASTCALL call_8_3_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_3_7
#define init_8_3_3_7
void FASTCALL call_8_3_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_3_8
#define init_8_3_3_8
void FASTCALL call_8_3_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_3_9
#define init_8_3_3_9
void FASTCALL call_8_3_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_3_10
#define init_8_3_3_10
void FASTCALL call_8_3_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_3_11
#define init_8_3_3_11
void FASTCALL call_8_3_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_3_12
#define init_8_3_3_12
void FASTCALL call_8_3_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_3_13
#define init_8_3_3_13
void FASTCALL call_8_3_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_3_14
#define init_8_3_3_14
void FASTCALL call_8_3_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_3_15
#define init_8_3_3_15
void FASTCALL call_8_3_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_4_0
#define init_8_3_4_0
void FASTCALL call_8_3_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_4_1
#define init_8_3_4_1
void FASTCALL call_8_3_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_4_2
#define init_8_3_4_2
void FASTCALL call_8_3_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_4_3
#define init_8_3_4_3
void FASTCALL call_8_3_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_4_4
#define init_8_3_4_4
void FASTCALL call_8_3_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_4_5
#define init_8_3_4_5
void FASTCALL call_8_3_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_4_6
#define init_8_3_4_6
void FASTCALL call_8_3_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_4_7
#define init_8_3_4_7
void FASTCALL call_8_3_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_4_8
#define init_8_3_4_8
void FASTCALL call_8_3_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_4_9
#define init_8_3_4_9
void FASTCALL call_8_3_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_4_10
#define init_8_3_4_10
void FASTCALL call_8_3_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_4_11
#define init_8_3_4_11
void FASTCALL call_8_3_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_4_12
#define init_8_3_4_12
void FASTCALL call_8_3_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_4_13
#define init_8_3_4_13
void FASTCALL call_8_3_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_4_14
#define init_8_3_4_14
void FASTCALL call_8_3_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_4_15
#define init_8_3_4_15
void FASTCALL call_8_3_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_5_0
#define init_8_3_5_0
void FASTCALL call_8_3_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_5_1
#define init_8_3_5_1
void FASTCALL call_8_3_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_5_2
#define init_8_3_5_2
void FASTCALL call_8_3_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_5_3
#define init_8_3_5_3
void FASTCALL call_8_3_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_5_4
#define init_8_3_5_4
void FASTCALL call_8_3_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_5_5
#define init_8_3_5_5
void FASTCALL call_8_3_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_5_6
#define init_8_3_5_6
void FASTCALL call_8_3_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_5_7
#define init_8_3_5_7
void FASTCALL call_8_3_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_5_8
#define init_8_3_5_8
void FASTCALL call_8_3_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_5_9
#define init_8_3_5_9
void FASTCALL call_8_3_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_5_10
#define init_8_3_5_10
void FASTCALL call_8_3_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_5_11
#define init_8_3_5_11
void FASTCALL call_8_3_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_5_12
#define init_8_3_5_12
void FASTCALL call_8_3_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_5_13
#define init_8_3_5_13
void FASTCALL call_8_3_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_5_14
#define init_8_3_5_14
void FASTCALL call_8_3_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_5_15
#define init_8_3_5_15
void FASTCALL call_8_3_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_6_0
#define init_8_3_6_0
void FASTCALL call_8_3_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_6_1
#define init_8_3_6_1
void FASTCALL call_8_3_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_6_2
#define init_8_3_6_2
void FASTCALL call_8_3_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_6_3
#define init_8_3_6_3
void FASTCALL call_8_3_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_6_4
#define init_8_3_6_4
void FASTCALL call_8_3_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_6_5
#define init_8_3_6_5
void FASTCALL call_8_3_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_6_6
#define init_8_3_6_6
void FASTCALL call_8_3_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_6_7
#define init_8_3_6_7
void FASTCALL call_8_3_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_6_8
#define init_8_3_6_8
void FASTCALL call_8_3_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_6_9
#define init_8_3_6_9
void FASTCALL call_8_3_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_6_10
#define init_8_3_6_10
void FASTCALL call_8_3_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_6_11
#define init_8_3_6_11
void FASTCALL call_8_3_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_6_12
#define init_8_3_6_12
void FASTCALL call_8_3_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_6_13
#define init_8_3_6_13
void FASTCALL call_8_3_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_6_14
#define init_8_3_6_14
void FASTCALL call_8_3_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_6_15
#define init_8_3_6_15
void FASTCALL call_8_3_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_7_0
#define init_8_3_7_0
void FASTCALL call_8_3_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_7_1
#define init_8_3_7_1
void FASTCALL call_8_3_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_7_2
#define init_8_3_7_2
void FASTCALL call_8_3_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_7_3
#define init_8_3_7_3
void FASTCALL call_8_3_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_7_4
#define init_8_3_7_4
void FASTCALL call_8_3_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_7_5
#define init_8_3_7_5
void FASTCALL call_8_3_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_7_6
#define init_8_3_7_6
void FASTCALL call_8_3_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_7_7
#define init_8_3_7_7
void FASTCALL call_8_3_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_7_8
#define init_8_3_7_8
void FASTCALL call_8_3_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_7_9
#define init_8_3_7_9
void FASTCALL call_8_3_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_7_10
#define init_8_3_7_10
void FASTCALL call_8_3_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_7_11
#define init_8_3_7_11
void FASTCALL call_8_3_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_7_12
#define init_8_3_7_12
void FASTCALL call_8_3_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_7_13
#define init_8_3_7_13
void FASTCALL call_8_3_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_7_14
#define init_8_3_7_14
void FASTCALL call_8_3_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_7_15
#define init_8_3_7_15
void FASTCALL call_8_3_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_8_0
#define init_8_3_8_0
void FASTCALL call_8_3_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_8_1
#define init_8_3_8_1
void FASTCALL call_8_3_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_8_2
#define init_8_3_8_2
void FASTCALL call_8_3_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_8_3
#define init_8_3_8_3
void FASTCALL call_8_3_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_8_4
#define init_8_3_8_4
void FASTCALL call_8_3_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_8_5
#define init_8_3_8_5
void FASTCALL call_8_3_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_8_6
#define init_8_3_8_6
void FASTCALL call_8_3_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_8_7
#define init_8_3_8_7
void FASTCALL call_8_3_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_8_8
#define init_8_3_8_8
void FASTCALL call_8_3_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_8_9
#define init_8_3_8_9
void FASTCALL call_8_3_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_8_10
#define init_8_3_8_10
void FASTCALL call_8_3_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_8_11
#define init_8_3_8_11
void FASTCALL call_8_3_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_8_12
#define init_8_3_8_12
void FASTCALL call_8_3_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_8_13
#define init_8_3_8_13
void FASTCALL call_8_3_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_8_14
#define init_8_3_8_14
void FASTCALL call_8_3_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_8_15
#define init_8_3_8_15
void FASTCALL call_8_3_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_9_0
#define init_8_3_9_0
void FASTCALL call_8_3_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_9_1
#define init_8_3_9_1
void FASTCALL call_8_3_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_9_2
#define init_8_3_9_2
void FASTCALL call_8_3_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_9_3
#define init_8_3_9_3
void FASTCALL call_8_3_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_9_4
#define init_8_3_9_4
void FASTCALL call_8_3_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_9_5
#define init_8_3_9_5
void FASTCALL call_8_3_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_9_6
#define init_8_3_9_6
void FASTCALL call_8_3_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_9_7
#define init_8_3_9_7
void FASTCALL call_8_3_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_9_8
#define init_8_3_9_8
void FASTCALL call_8_3_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_9_9
#define init_8_3_9_9
void FASTCALL call_8_3_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_9_10
#define init_8_3_9_10
void FASTCALL call_8_3_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_9_11
#define init_8_3_9_11
void FASTCALL call_8_3_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_9_12
#define init_8_3_9_12
void FASTCALL call_8_3_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_9_13
#define init_8_3_9_13
void FASTCALL call_8_3_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_9_14
#define init_8_3_9_14
void FASTCALL call_8_3_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_9_15
#define init_8_3_9_15
void FASTCALL call_8_3_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_10_0
#define init_8_3_10_0
void FASTCALL call_8_3_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_10_1
#define init_8_3_10_1
void FASTCALL call_8_3_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_10_2
#define init_8_3_10_2
void FASTCALL call_8_3_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_10_3
#define init_8_3_10_3
void FASTCALL call_8_3_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_10_4
#define init_8_3_10_4
void FASTCALL call_8_3_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_10_5
#define init_8_3_10_5
void FASTCALL call_8_3_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_10_6
#define init_8_3_10_6
void FASTCALL call_8_3_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_10_7
#define init_8_3_10_7
void FASTCALL call_8_3_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_10_8
#define init_8_3_10_8
void FASTCALL call_8_3_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_10_9
#define init_8_3_10_9
void FASTCALL call_8_3_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_10_10
#define init_8_3_10_10
void FASTCALL call_8_3_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_10_11
#define init_8_3_10_11
void FASTCALL call_8_3_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_10_12
#define init_8_3_10_12
void FASTCALL call_8_3_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_10_13
#define init_8_3_10_13
void FASTCALL call_8_3_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_10_14
#define init_8_3_10_14
void FASTCALL call_8_3_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_10_15
#define init_8_3_10_15
void FASTCALL call_8_3_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_11_0
#define init_8_3_11_0
void FASTCALL call_8_3_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_11_1
#define init_8_3_11_1
void FASTCALL call_8_3_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_11_2
#define init_8_3_11_2
void FASTCALL call_8_3_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_11_3
#define init_8_3_11_3
void FASTCALL call_8_3_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_11_4
#define init_8_3_11_4
void FASTCALL call_8_3_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_11_5
#define init_8_3_11_5
void FASTCALL call_8_3_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_11_6
#define init_8_3_11_6
void FASTCALL call_8_3_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_11_7
#define init_8_3_11_7
void FASTCALL call_8_3_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_11_8
#define init_8_3_11_8
void FASTCALL call_8_3_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_11_9
#define init_8_3_11_9
void FASTCALL call_8_3_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_11_10
#define init_8_3_11_10
void FASTCALL call_8_3_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_11_11
#define init_8_3_11_11
void FASTCALL call_8_3_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_11_12
#define init_8_3_11_12
void FASTCALL call_8_3_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_11_13
#define init_8_3_11_13
void FASTCALL call_8_3_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_11_14
#define init_8_3_11_14
void FASTCALL call_8_3_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_11_15
#define init_8_3_11_15
void FASTCALL call_8_3_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_12_0
#define init_8_3_12_0
void FASTCALL call_8_3_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_12_1
#define init_8_3_12_1
void FASTCALL call_8_3_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_12_2
#define init_8_3_12_2
void FASTCALL call_8_3_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_12_3
#define init_8_3_12_3
void FASTCALL call_8_3_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_12_4
#define init_8_3_12_4
void FASTCALL call_8_3_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_12_5
#define init_8_3_12_5
void FASTCALL call_8_3_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_12_6
#define init_8_3_12_6
void FASTCALL call_8_3_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_12_7
#define init_8_3_12_7
void FASTCALL call_8_3_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_12_8
#define init_8_3_12_8
void FASTCALL call_8_3_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_12_9
#define init_8_3_12_9
void FASTCALL call_8_3_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_12_10
#define init_8_3_12_10
void FASTCALL call_8_3_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_12_11
#define init_8_3_12_11
void FASTCALL call_8_3_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_12_12
#define init_8_3_12_12
void FASTCALL call_8_3_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_12_13
#define init_8_3_12_13
void FASTCALL call_8_3_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_12_14
#define init_8_3_12_14
void FASTCALL call_8_3_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_12_15
#define init_8_3_12_15
void FASTCALL call_8_3_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_13_0
#define init_8_3_13_0
void FASTCALL call_8_3_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_13_1
#define init_8_3_13_1
void FASTCALL call_8_3_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_13_2
#define init_8_3_13_2
void FASTCALL call_8_3_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_13_3
#define init_8_3_13_3
void FASTCALL call_8_3_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_13_4
#define init_8_3_13_4
void FASTCALL call_8_3_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_13_5
#define init_8_3_13_5
void FASTCALL call_8_3_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_13_6
#define init_8_3_13_6
void FASTCALL call_8_3_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_13_7
#define init_8_3_13_7
void FASTCALL call_8_3_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_13_8
#define init_8_3_13_8
void FASTCALL call_8_3_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_13_9
#define init_8_3_13_9
void FASTCALL call_8_3_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_13_10
#define init_8_3_13_10
void FASTCALL call_8_3_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_13_11
#define init_8_3_13_11
void FASTCALL call_8_3_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_13_12
#define init_8_3_13_12
void FASTCALL call_8_3_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_13_13
#define init_8_3_13_13
void FASTCALL call_8_3_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_13_14
#define init_8_3_13_14
void FASTCALL call_8_3_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_13_15
#define init_8_3_13_15
void FASTCALL call_8_3_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_14_0
#define init_8_3_14_0
void FASTCALL call_8_3_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_14_1
#define init_8_3_14_1
void FASTCALL call_8_3_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_14_2
#define init_8_3_14_2
void FASTCALL call_8_3_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_14_3
#define init_8_3_14_3
void FASTCALL call_8_3_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_14_4
#define init_8_3_14_4
void FASTCALL call_8_3_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_14_5
#define init_8_3_14_5
void FASTCALL call_8_3_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_14_6
#define init_8_3_14_6
void FASTCALL call_8_3_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_14_7
#define init_8_3_14_7
void FASTCALL call_8_3_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_14_8
#define init_8_3_14_8
void FASTCALL call_8_3_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_14_9
#define init_8_3_14_9
void FASTCALL call_8_3_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_14_10
#define init_8_3_14_10
void FASTCALL call_8_3_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_14_11
#define init_8_3_14_11
void FASTCALL call_8_3_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_14_12
#define init_8_3_14_12
void FASTCALL call_8_3_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_14_13
#define init_8_3_14_13
void FASTCALL call_8_3_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_14_14
#define init_8_3_14_14
void FASTCALL call_8_3_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_14_15
#define init_8_3_14_15
void FASTCALL call_8_3_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_15_0
#define init_8_3_15_0
void FASTCALL call_8_3_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_15_1
#define init_8_3_15_1
void FASTCALL call_8_3_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_15_2
#define init_8_3_15_2
void FASTCALL call_8_3_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_15_3
#define init_8_3_15_3
void FASTCALL call_8_3_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_15_4
#define init_8_3_15_4
void FASTCALL call_8_3_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_15_5
#define init_8_3_15_5
void FASTCALL call_8_3_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_15_6
#define init_8_3_15_6
void FASTCALL call_8_3_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_15_7
#define init_8_3_15_7
void FASTCALL call_8_3_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_15_8
#define init_8_3_15_8
void FASTCALL call_8_3_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_15_9
#define init_8_3_15_9
void FASTCALL call_8_3_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_15_10
#define init_8_3_15_10
void FASTCALL call_8_3_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_15_11
#define init_8_3_15_11
void FASTCALL call_8_3_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_15_12
#define init_8_3_15_12
void FASTCALL call_8_3_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_15_13
#define init_8_3_15_13
void FASTCALL call_8_3_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_15_14
#define init_8_3_15_14
void FASTCALL call_8_3_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_3_15_15
#define init_8_3_15_15
void FASTCALL call_8_3_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_0_0
#define init_8_4_0_0
void FASTCALL call_8_4_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_0_1
#define init_8_4_0_1
void FASTCALL call_8_4_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_0_2
#define init_8_4_0_2
void FASTCALL call_8_4_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_0_3
#define init_8_4_0_3
void FASTCALL call_8_4_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_0_4
#define init_8_4_0_4
void FASTCALL call_8_4_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_0_5
#define init_8_4_0_5
void FASTCALL call_8_4_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_0_6
#define init_8_4_0_6
void FASTCALL call_8_4_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_0_7
#define init_8_4_0_7
void FASTCALL call_8_4_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_0_8
#define init_8_4_0_8
void FASTCALL call_8_4_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_0_9
#define init_8_4_0_9
void FASTCALL call_8_4_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_0_10
#define init_8_4_0_10
void FASTCALL call_8_4_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_0_11
#define init_8_4_0_11
void FASTCALL call_8_4_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_0_12
#define init_8_4_0_12
void FASTCALL call_8_4_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_0_13
#define init_8_4_0_13
void FASTCALL call_8_4_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_0_14
#define init_8_4_0_14
void FASTCALL call_8_4_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_0_15
#define init_8_4_0_15
void FASTCALL call_8_4_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_1_0
#define init_8_4_1_0
void FASTCALL call_8_4_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_1_1
#define init_8_4_1_1
void FASTCALL call_8_4_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_1_2
#define init_8_4_1_2
void FASTCALL call_8_4_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_1_3
#define init_8_4_1_3
void FASTCALL call_8_4_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_1_4
#define init_8_4_1_4
void FASTCALL call_8_4_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_1_5
#define init_8_4_1_5
void FASTCALL call_8_4_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_1_6
#define init_8_4_1_6
void FASTCALL call_8_4_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_1_7
#define init_8_4_1_7
void FASTCALL call_8_4_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_1_8
#define init_8_4_1_8
void FASTCALL call_8_4_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_1_9
#define init_8_4_1_9
void FASTCALL call_8_4_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_1_10
#define init_8_4_1_10
void FASTCALL call_8_4_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_1_11
#define init_8_4_1_11
void FASTCALL call_8_4_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_1_12
#define init_8_4_1_12
void FASTCALL call_8_4_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_1_13
#define init_8_4_1_13
void FASTCALL call_8_4_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_1_14
#define init_8_4_1_14
void FASTCALL call_8_4_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_1_15
#define init_8_4_1_15
void FASTCALL call_8_4_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_2_0
#define init_8_4_2_0
void FASTCALL call_8_4_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_2_1
#define init_8_4_2_1
void FASTCALL call_8_4_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_2_2
#define init_8_4_2_2
void FASTCALL call_8_4_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_2_3
#define init_8_4_2_3
void FASTCALL call_8_4_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_2_4
#define init_8_4_2_4
void FASTCALL call_8_4_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_2_5
#define init_8_4_2_5
void FASTCALL call_8_4_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_2_6
#define init_8_4_2_6
void FASTCALL call_8_4_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_2_7
#define init_8_4_2_7
void FASTCALL call_8_4_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_2_8
#define init_8_4_2_8
void FASTCALL call_8_4_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_2_9
#define init_8_4_2_9
void FASTCALL call_8_4_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_2_10
#define init_8_4_2_10
void FASTCALL call_8_4_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_2_11
#define init_8_4_2_11
void FASTCALL call_8_4_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_2_12
#define init_8_4_2_12
void FASTCALL call_8_4_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_2_13
#define init_8_4_2_13
void FASTCALL call_8_4_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_2_14
#define init_8_4_2_14
void FASTCALL call_8_4_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_2_15
#define init_8_4_2_15
void FASTCALL call_8_4_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_3_0
#define init_8_4_3_0
void FASTCALL call_8_4_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_3_1
#define init_8_4_3_1
void FASTCALL call_8_4_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_3_2
#define init_8_4_3_2
void FASTCALL call_8_4_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_3_3
#define init_8_4_3_3
void FASTCALL call_8_4_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_3_4
#define init_8_4_3_4
void FASTCALL call_8_4_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_3_5
#define init_8_4_3_5
void FASTCALL call_8_4_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_3_6
#define init_8_4_3_6
void FASTCALL call_8_4_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_3_7
#define init_8_4_3_7
void FASTCALL call_8_4_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_3_8
#define init_8_4_3_8
void FASTCALL call_8_4_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_3_9
#define init_8_4_3_9
void FASTCALL call_8_4_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_3_10
#define init_8_4_3_10
void FASTCALL call_8_4_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_3_11
#define init_8_4_3_11
void FASTCALL call_8_4_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_3_12
#define init_8_4_3_12
void FASTCALL call_8_4_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_3_13
#define init_8_4_3_13
void FASTCALL call_8_4_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_3_14
#define init_8_4_3_14
void FASTCALL call_8_4_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_3_15
#define init_8_4_3_15
void FASTCALL call_8_4_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_4_0
#define init_8_4_4_0
void FASTCALL call_8_4_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_4_1
#define init_8_4_4_1
void FASTCALL call_8_4_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_4_2
#define init_8_4_4_2
void FASTCALL call_8_4_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_4_3
#define init_8_4_4_3
void FASTCALL call_8_4_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_4_4
#define init_8_4_4_4
void FASTCALL call_8_4_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_4_5
#define init_8_4_4_5
void FASTCALL call_8_4_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_4_6
#define init_8_4_4_6
void FASTCALL call_8_4_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_4_7
#define init_8_4_4_7
void FASTCALL call_8_4_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_4_8
#define init_8_4_4_8
void FASTCALL call_8_4_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_4_9
#define init_8_4_4_9
void FASTCALL call_8_4_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_4_10
#define init_8_4_4_10
void FASTCALL call_8_4_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_4_11
#define init_8_4_4_11
void FASTCALL call_8_4_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_4_12
#define init_8_4_4_12
void FASTCALL call_8_4_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_4_13
#define init_8_4_4_13
void FASTCALL call_8_4_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_4_14
#define init_8_4_4_14
void FASTCALL call_8_4_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_4_15
#define init_8_4_4_15
void FASTCALL call_8_4_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_5_0
#define init_8_4_5_0
void FASTCALL call_8_4_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_5_1
#define init_8_4_5_1
void FASTCALL call_8_4_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_5_2
#define init_8_4_5_2
void FASTCALL call_8_4_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_5_3
#define init_8_4_5_3
void FASTCALL call_8_4_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_5_4
#define init_8_4_5_4
void FASTCALL call_8_4_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_5_5
#define init_8_4_5_5
void FASTCALL call_8_4_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_5_6
#define init_8_4_5_6
void FASTCALL call_8_4_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_5_7
#define init_8_4_5_7
void FASTCALL call_8_4_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_5_8
#define init_8_4_5_8
void FASTCALL call_8_4_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_5_9
#define init_8_4_5_9
void FASTCALL call_8_4_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_5_10
#define init_8_4_5_10
void FASTCALL call_8_4_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_5_11
#define init_8_4_5_11
void FASTCALL call_8_4_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_5_12
#define init_8_4_5_12
void FASTCALL call_8_4_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_5_13
#define init_8_4_5_13
void FASTCALL call_8_4_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_5_14
#define init_8_4_5_14
void FASTCALL call_8_4_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_5_15
#define init_8_4_5_15
void FASTCALL call_8_4_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_6_0
#define init_8_4_6_0
void FASTCALL call_8_4_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_6_1
#define init_8_4_6_1
void FASTCALL call_8_4_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_6_2
#define init_8_4_6_2
void FASTCALL call_8_4_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_6_3
#define init_8_4_6_3
void FASTCALL call_8_4_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_6_4
#define init_8_4_6_4
void FASTCALL call_8_4_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_6_5
#define init_8_4_6_5
void FASTCALL call_8_4_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_6_6
#define init_8_4_6_6
void FASTCALL call_8_4_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_6_7
#define init_8_4_6_7
void FASTCALL call_8_4_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_6_8
#define init_8_4_6_8
void FASTCALL call_8_4_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_6_9
#define init_8_4_6_9
void FASTCALL call_8_4_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_6_10
#define init_8_4_6_10
void FASTCALL call_8_4_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_6_11
#define init_8_4_6_11
void FASTCALL call_8_4_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_6_12
#define init_8_4_6_12
void FASTCALL call_8_4_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_6_13
#define init_8_4_6_13
void FASTCALL call_8_4_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_6_14
#define init_8_4_6_14
void FASTCALL call_8_4_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_6_15
#define init_8_4_6_15
void FASTCALL call_8_4_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_7_0
#define init_8_4_7_0
void FASTCALL call_8_4_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_7_1
#define init_8_4_7_1
void FASTCALL call_8_4_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_7_2
#define init_8_4_7_2
void FASTCALL call_8_4_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_7_3
#define init_8_4_7_3
void FASTCALL call_8_4_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_7_4
#define init_8_4_7_4
void FASTCALL call_8_4_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_7_5
#define init_8_4_7_5
void FASTCALL call_8_4_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_7_6
#define init_8_4_7_6
void FASTCALL call_8_4_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_7_7
#define init_8_4_7_7
void FASTCALL call_8_4_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_7_8
#define init_8_4_7_8
void FASTCALL call_8_4_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_7_9
#define init_8_4_7_9
void FASTCALL call_8_4_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_7_10
#define init_8_4_7_10
void FASTCALL call_8_4_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_7_11
#define init_8_4_7_11
void FASTCALL call_8_4_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_7_12
#define init_8_4_7_12
void FASTCALL call_8_4_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_7_13
#define init_8_4_7_13
void FASTCALL call_8_4_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_7_14
#define init_8_4_7_14
void FASTCALL call_8_4_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_7_15
#define init_8_4_7_15
void FASTCALL call_8_4_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_8_0
#define init_8_4_8_0
void FASTCALL call_8_4_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_8_1
#define init_8_4_8_1
void FASTCALL call_8_4_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_8_2
#define init_8_4_8_2
void FASTCALL call_8_4_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_8_3
#define init_8_4_8_3
void FASTCALL call_8_4_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_8_4
#define init_8_4_8_4
void FASTCALL call_8_4_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_8_5
#define init_8_4_8_5
void FASTCALL call_8_4_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_8_6
#define init_8_4_8_6
void FASTCALL call_8_4_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_8_7
#define init_8_4_8_7
void FASTCALL call_8_4_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_8_8
#define init_8_4_8_8
void FASTCALL call_8_4_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_8_9
#define init_8_4_8_9
void FASTCALL call_8_4_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_8_10
#define init_8_4_8_10
void FASTCALL call_8_4_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_8_11
#define init_8_4_8_11
void FASTCALL call_8_4_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_8_12
#define init_8_4_8_12
void FASTCALL call_8_4_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_8_13
#define init_8_4_8_13
void FASTCALL call_8_4_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_8_14
#define init_8_4_8_14
void FASTCALL call_8_4_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_8_15
#define init_8_4_8_15
void FASTCALL call_8_4_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_9_0
#define init_8_4_9_0
void FASTCALL call_8_4_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_9_1
#define init_8_4_9_1
void FASTCALL call_8_4_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_9_2
#define init_8_4_9_2
void FASTCALL call_8_4_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_9_3
#define init_8_4_9_3
void FASTCALL call_8_4_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_9_4
#define init_8_4_9_4
void FASTCALL call_8_4_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_9_5
#define init_8_4_9_5
void FASTCALL call_8_4_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_9_6
#define init_8_4_9_6
void FASTCALL call_8_4_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_9_7
#define init_8_4_9_7
void FASTCALL call_8_4_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_9_8
#define init_8_4_9_8
void FASTCALL call_8_4_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_9_9
#define init_8_4_9_9
void FASTCALL call_8_4_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_9_10
#define init_8_4_9_10
void FASTCALL call_8_4_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_9_11
#define init_8_4_9_11
void FASTCALL call_8_4_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_9_12
#define init_8_4_9_12
void FASTCALL call_8_4_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_9_13
#define init_8_4_9_13
void FASTCALL call_8_4_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_9_14
#define init_8_4_9_14
void FASTCALL call_8_4_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_9_15
#define init_8_4_9_15
void FASTCALL call_8_4_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_10_0
#define init_8_4_10_0
void FASTCALL call_8_4_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_10_1
#define init_8_4_10_1
void FASTCALL call_8_4_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_10_2
#define init_8_4_10_2
void FASTCALL call_8_4_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_10_3
#define init_8_4_10_3
void FASTCALL call_8_4_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_10_4
#define init_8_4_10_4
void FASTCALL call_8_4_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_10_5
#define init_8_4_10_5
void FASTCALL call_8_4_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_10_6
#define init_8_4_10_6
void FASTCALL call_8_4_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_10_7
#define init_8_4_10_7
void FASTCALL call_8_4_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_10_8
#define init_8_4_10_8
void FASTCALL call_8_4_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_10_9
#define init_8_4_10_9
void FASTCALL call_8_4_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_10_10
#define init_8_4_10_10
void FASTCALL call_8_4_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_10_11
#define init_8_4_10_11
void FASTCALL call_8_4_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_10_12
#define init_8_4_10_12
void FASTCALL call_8_4_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_10_13
#define init_8_4_10_13
void FASTCALL call_8_4_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_10_14
#define init_8_4_10_14
void FASTCALL call_8_4_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_10_15
#define init_8_4_10_15
void FASTCALL call_8_4_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_11_0
#define init_8_4_11_0
void FASTCALL call_8_4_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_11_1
#define init_8_4_11_1
void FASTCALL call_8_4_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_11_2
#define init_8_4_11_2
void FASTCALL call_8_4_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_11_3
#define init_8_4_11_3
void FASTCALL call_8_4_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_11_4
#define init_8_4_11_4
void FASTCALL call_8_4_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_11_5
#define init_8_4_11_5
void FASTCALL call_8_4_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_11_6
#define init_8_4_11_6
void FASTCALL call_8_4_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_11_7
#define init_8_4_11_7
void FASTCALL call_8_4_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_11_8
#define init_8_4_11_8
void FASTCALL call_8_4_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_11_9
#define init_8_4_11_9
void FASTCALL call_8_4_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_11_10
#define init_8_4_11_10
void FASTCALL call_8_4_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_11_11
#define init_8_4_11_11
void FASTCALL call_8_4_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_11_12
#define init_8_4_11_12
void FASTCALL call_8_4_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_11_13
#define init_8_4_11_13
void FASTCALL call_8_4_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_11_14
#define init_8_4_11_14
void FASTCALL call_8_4_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_11_15
#define init_8_4_11_15
void FASTCALL call_8_4_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_12_0
#define init_8_4_12_0
void FASTCALL call_8_4_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_12_1
#define init_8_4_12_1
void FASTCALL call_8_4_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_12_2
#define init_8_4_12_2
void FASTCALL call_8_4_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_12_3
#define init_8_4_12_3
void FASTCALL call_8_4_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_12_4
#define init_8_4_12_4
void FASTCALL call_8_4_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_12_5
#define init_8_4_12_5
void FASTCALL call_8_4_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_12_6
#define init_8_4_12_6
void FASTCALL call_8_4_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_12_7
#define init_8_4_12_7
void FASTCALL call_8_4_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_12_8
#define init_8_4_12_8
void FASTCALL call_8_4_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_12_9
#define init_8_4_12_9
void FASTCALL call_8_4_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_12_10
#define init_8_4_12_10
void FASTCALL call_8_4_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_12_11
#define init_8_4_12_11
void FASTCALL call_8_4_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_12_12
#define init_8_4_12_12
void FASTCALL call_8_4_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_12_13
#define init_8_4_12_13
void FASTCALL call_8_4_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_12_14
#define init_8_4_12_14
void FASTCALL call_8_4_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_12_15
#define init_8_4_12_15
void FASTCALL call_8_4_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_13_0
#define init_8_4_13_0
void FASTCALL call_8_4_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_13_1
#define init_8_4_13_1
void FASTCALL call_8_4_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_13_2
#define init_8_4_13_2
void FASTCALL call_8_4_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_13_3
#define init_8_4_13_3
void FASTCALL call_8_4_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_13_4
#define init_8_4_13_4
void FASTCALL call_8_4_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_13_5
#define init_8_4_13_5
void FASTCALL call_8_4_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_13_6
#define init_8_4_13_6
void FASTCALL call_8_4_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_13_7
#define init_8_4_13_7
void FASTCALL call_8_4_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_13_8
#define init_8_4_13_8
void FASTCALL call_8_4_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_13_9
#define init_8_4_13_9
void FASTCALL call_8_4_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_13_10
#define init_8_4_13_10
void FASTCALL call_8_4_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_13_11
#define init_8_4_13_11
void FASTCALL call_8_4_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_13_12
#define init_8_4_13_12
void FASTCALL call_8_4_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_13_13
#define init_8_4_13_13
void FASTCALL call_8_4_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_13_14
#define init_8_4_13_14
void FASTCALL call_8_4_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_13_15
#define init_8_4_13_15
void FASTCALL call_8_4_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_14_0
#define init_8_4_14_0
void FASTCALL call_8_4_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_14_1
#define init_8_4_14_1
void FASTCALL call_8_4_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_14_2
#define init_8_4_14_2
void FASTCALL call_8_4_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_14_3
#define init_8_4_14_3
void FASTCALL call_8_4_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_14_4
#define init_8_4_14_4
void FASTCALL call_8_4_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_14_5
#define init_8_4_14_5
void FASTCALL call_8_4_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_14_6
#define init_8_4_14_6
void FASTCALL call_8_4_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_14_7
#define init_8_4_14_7
void FASTCALL call_8_4_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_14_8
#define init_8_4_14_8
void FASTCALL call_8_4_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_14_9
#define init_8_4_14_9
void FASTCALL call_8_4_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_14_10
#define init_8_4_14_10
void FASTCALL call_8_4_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_14_11
#define init_8_4_14_11
void FASTCALL call_8_4_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_14_12
#define init_8_4_14_12
void FASTCALL call_8_4_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_14_13
#define init_8_4_14_13
void FASTCALL call_8_4_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_14_14
#define init_8_4_14_14
void FASTCALL call_8_4_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_14_15
#define init_8_4_14_15
void FASTCALL call_8_4_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_15_0
#define init_8_4_15_0
void FASTCALL call_8_4_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_15_1
#define init_8_4_15_1
void FASTCALL call_8_4_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_15_2
#define init_8_4_15_2
void FASTCALL call_8_4_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_15_3
#define init_8_4_15_3
void FASTCALL call_8_4_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_15_4
#define init_8_4_15_4
void FASTCALL call_8_4_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_15_5
#define init_8_4_15_5
void FASTCALL call_8_4_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_15_6
#define init_8_4_15_6
void FASTCALL call_8_4_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_15_7
#define init_8_4_15_7
void FASTCALL call_8_4_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_15_8
#define init_8_4_15_8
void FASTCALL call_8_4_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_15_9
#define init_8_4_15_9
void FASTCALL call_8_4_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_15_10
#define init_8_4_15_10
void FASTCALL call_8_4_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_15_11
#define init_8_4_15_11
void FASTCALL call_8_4_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_15_12
#define init_8_4_15_12
void FASTCALL call_8_4_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_15_13
#define init_8_4_15_13
void FASTCALL call_8_4_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_15_14
#define init_8_4_15_14
void FASTCALL call_8_4_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_4_15_15
#define init_8_4_15_15
void FASTCALL call_8_4_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_0_0
#define init_8_5_0_0
void FASTCALL call_8_5_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_0_1
#define init_8_5_0_1
void FASTCALL call_8_5_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_0_2
#define init_8_5_0_2
void FASTCALL call_8_5_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_0_3
#define init_8_5_0_3
void FASTCALL call_8_5_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_0_4
#define init_8_5_0_4
void FASTCALL call_8_5_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_0_5
#define init_8_5_0_5
void FASTCALL call_8_5_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_0_6
#define init_8_5_0_6
void FASTCALL call_8_5_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_0_7
#define init_8_5_0_7
void FASTCALL call_8_5_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_0_8
#define init_8_5_0_8
void FASTCALL call_8_5_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_0_9
#define init_8_5_0_9
void FASTCALL call_8_5_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_0_10
#define init_8_5_0_10
void FASTCALL call_8_5_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_0_11
#define init_8_5_0_11
void FASTCALL call_8_5_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_0_12
#define init_8_5_0_12
void FASTCALL call_8_5_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_0_13
#define init_8_5_0_13
void FASTCALL call_8_5_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_0_14
#define init_8_5_0_14
void FASTCALL call_8_5_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_0_15
#define init_8_5_0_15
void FASTCALL call_8_5_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_1_0
#define init_8_5_1_0
void FASTCALL call_8_5_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_1_1
#define init_8_5_1_1
void FASTCALL call_8_5_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_1_2
#define init_8_5_1_2
void FASTCALL call_8_5_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_1_3
#define init_8_5_1_3
void FASTCALL call_8_5_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_1_4
#define init_8_5_1_4
void FASTCALL call_8_5_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_1_5
#define init_8_5_1_5
void FASTCALL call_8_5_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_1_6
#define init_8_5_1_6
void FASTCALL call_8_5_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_1_7
#define init_8_5_1_7
void FASTCALL call_8_5_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_1_8
#define init_8_5_1_8
void FASTCALL call_8_5_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_1_9
#define init_8_5_1_9
void FASTCALL call_8_5_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_1_10
#define init_8_5_1_10
void FASTCALL call_8_5_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_1_11
#define init_8_5_1_11
void FASTCALL call_8_5_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_1_12
#define init_8_5_1_12
void FASTCALL call_8_5_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_1_13
#define init_8_5_1_13
void FASTCALL call_8_5_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_1_14
#define init_8_5_1_14
void FASTCALL call_8_5_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_1_15
#define init_8_5_1_15
void FASTCALL call_8_5_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_2_0
#define init_8_5_2_0
void FASTCALL call_8_5_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_2_1
#define init_8_5_2_1
void FASTCALL call_8_5_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_2_2
#define init_8_5_2_2
void FASTCALL call_8_5_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_2_3
#define init_8_5_2_3
void FASTCALL call_8_5_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_2_4
#define init_8_5_2_4
void FASTCALL call_8_5_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_2_5
#define init_8_5_2_5
void FASTCALL call_8_5_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_2_6
#define init_8_5_2_6
void FASTCALL call_8_5_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_2_7
#define init_8_5_2_7
void FASTCALL call_8_5_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_2_8
#define init_8_5_2_8
void FASTCALL call_8_5_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_2_9
#define init_8_5_2_9
void FASTCALL call_8_5_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_2_10
#define init_8_5_2_10
void FASTCALL call_8_5_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_2_11
#define init_8_5_2_11
void FASTCALL call_8_5_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_2_12
#define init_8_5_2_12
void FASTCALL call_8_5_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_2_13
#define init_8_5_2_13
void FASTCALL call_8_5_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_2_14
#define init_8_5_2_14
void FASTCALL call_8_5_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_2_15
#define init_8_5_2_15
void FASTCALL call_8_5_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_3_0
#define init_8_5_3_0
void FASTCALL call_8_5_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_3_1
#define init_8_5_3_1
void FASTCALL call_8_5_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_3_2
#define init_8_5_3_2
void FASTCALL call_8_5_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_3_3
#define init_8_5_3_3
void FASTCALL call_8_5_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_3_4
#define init_8_5_3_4
void FASTCALL call_8_5_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_3_5
#define init_8_5_3_5
void FASTCALL call_8_5_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_3_6
#define init_8_5_3_6
void FASTCALL call_8_5_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_3_7
#define init_8_5_3_7
void FASTCALL call_8_5_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_3_8
#define init_8_5_3_8
void FASTCALL call_8_5_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_3_9
#define init_8_5_3_9
void FASTCALL call_8_5_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_3_10
#define init_8_5_3_10
void FASTCALL call_8_5_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_3_11
#define init_8_5_3_11
void FASTCALL call_8_5_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_3_12
#define init_8_5_3_12
void FASTCALL call_8_5_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_3_13
#define init_8_5_3_13
void FASTCALL call_8_5_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_3_14
#define init_8_5_3_14
void FASTCALL call_8_5_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_3_15
#define init_8_5_3_15
void FASTCALL call_8_5_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_4_0
#define init_8_5_4_0
void FASTCALL call_8_5_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_4_1
#define init_8_5_4_1
void FASTCALL call_8_5_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_4_2
#define init_8_5_4_2
void FASTCALL call_8_5_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_4_3
#define init_8_5_4_3
void FASTCALL call_8_5_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_4_4
#define init_8_5_4_4
void FASTCALL call_8_5_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_4_5
#define init_8_5_4_5
void FASTCALL call_8_5_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_4_6
#define init_8_5_4_6
void FASTCALL call_8_5_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_4_7
#define init_8_5_4_7
void FASTCALL call_8_5_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_4_8
#define init_8_5_4_8
void FASTCALL call_8_5_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_4_9
#define init_8_5_4_9
void FASTCALL call_8_5_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_4_10
#define init_8_5_4_10
void FASTCALL call_8_5_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_4_11
#define init_8_5_4_11
void FASTCALL call_8_5_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_4_12
#define init_8_5_4_12
void FASTCALL call_8_5_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_4_13
#define init_8_5_4_13
void FASTCALL call_8_5_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_4_14
#define init_8_5_4_14
void FASTCALL call_8_5_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_4_15
#define init_8_5_4_15
void FASTCALL call_8_5_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_5_0
#define init_8_5_5_0
void FASTCALL call_8_5_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_5_1
#define init_8_5_5_1
void FASTCALL call_8_5_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_5_2
#define init_8_5_5_2
void FASTCALL call_8_5_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_5_3
#define init_8_5_5_3
void FASTCALL call_8_5_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_5_4
#define init_8_5_5_4
void FASTCALL call_8_5_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_5_5
#define init_8_5_5_5
void FASTCALL call_8_5_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_5_6
#define init_8_5_5_6
void FASTCALL call_8_5_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_5_7
#define init_8_5_5_7
void FASTCALL call_8_5_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_5_8
#define init_8_5_5_8
void FASTCALL call_8_5_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_5_9
#define init_8_5_5_9
void FASTCALL call_8_5_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_5_10
#define init_8_5_5_10
void FASTCALL call_8_5_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_5_11
#define init_8_5_5_11
void FASTCALL call_8_5_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_5_12
#define init_8_5_5_12
void FASTCALL call_8_5_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_5_13
#define init_8_5_5_13
void FASTCALL call_8_5_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_5_14
#define init_8_5_5_14
void FASTCALL call_8_5_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_5_15
#define init_8_5_5_15
void FASTCALL call_8_5_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_6_0
#define init_8_5_6_0
void FASTCALL call_8_5_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_6_1
#define init_8_5_6_1
void FASTCALL call_8_5_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_6_2
#define init_8_5_6_2
void FASTCALL call_8_5_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_6_3
#define init_8_5_6_3
void FASTCALL call_8_5_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_6_4
#define init_8_5_6_4
void FASTCALL call_8_5_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_6_5
#define init_8_5_6_5
void FASTCALL call_8_5_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_6_6
#define init_8_5_6_6
void FASTCALL call_8_5_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_6_7
#define init_8_5_6_7
void FASTCALL call_8_5_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_6_8
#define init_8_5_6_8
void FASTCALL call_8_5_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_6_9
#define init_8_5_6_9
void FASTCALL call_8_5_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_6_10
#define init_8_5_6_10
void FASTCALL call_8_5_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_6_11
#define init_8_5_6_11
void FASTCALL call_8_5_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_6_12
#define init_8_5_6_12
void FASTCALL call_8_5_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_6_13
#define init_8_5_6_13
void FASTCALL call_8_5_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_6_14
#define init_8_5_6_14
void FASTCALL call_8_5_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_6_15
#define init_8_5_6_15
void FASTCALL call_8_5_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_7_0
#define init_8_5_7_0
void FASTCALL call_8_5_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_7_1
#define init_8_5_7_1
void FASTCALL call_8_5_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_7_2
#define init_8_5_7_2
void FASTCALL call_8_5_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_7_3
#define init_8_5_7_3
void FASTCALL call_8_5_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_7_4
#define init_8_5_7_4
void FASTCALL call_8_5_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_7_5
#define init_8_5_7_5
void FASTCALL call_8_5_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_7_6
#define init_8_5_7_6
void FASTCALL call_8_5_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_7_7
#define init_8_5_7_7
void FASTCALL call_8_5_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_7_8
#define init_8_5_7_8
void FASTCALL call_8_5_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_7_9
#define init_8_5_7_9
void FASTCALL call_8_5_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_7_10
#define init_8_5_7_10
void FASTCALL call_8_5_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_7_11
#define init_8_5_7_11
void FASTCALL call_8_5_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_7_12
#define init_8_5_7_12
void FASTCALL call_8_5_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_7_13
#define init_8_5_7_13
void FASTCALL call_8_5_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_7_14
#define init_8_5_7_14
void FASTCALL call_8_5_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_7_15
#define init_8_5_7_15
void FASTCALL call_8_5_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_8_0
#define init_8_5_8_0
void FASTCALL call_8_5_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_8_1
#define init_8_5_8_1
void FASTCALL call_8_5_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_8_2
#define init_8_5_8_2
void FASTCALL call_8_5_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_8_3
#define init_8_5_8_3
void FASTCALL call_8_5_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_8_4
#define init_8_5_8_4
void FASTCALL call_8_5_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_8_5
#define init_8_5_8_5
void FASTCALL call_8_5_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_8_6
#define init_8_5_8_6
void FASTCALL call_8_5_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_8_7
#define init_8_5_8_7
void FASTCALL call_8_5_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_8_8
#define init_8_5_8_8
void FASTCALL call_8_5_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_8_9
#define init_8_5_8_9
void FASTCALL call_8_5_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_8_10
#define init_8_5_8_10
void FASTCALL call_8_5_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_8_11
#define init_8_5_8_11
void FASTCALL call_8_5_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_8_12
#define init_8_5_8_12
void FASTCALL call_8_5_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_8_13
#define init_8_5_8_13
void FASTCALL call_8_5_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_8_14
#define init_8_5_8_14
void FASTCALL call_8_5_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_8_15
#define init_8_5_8_15
void FASTCALL call_8_5_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_9_0
#define init_8_5_9_0
void FASTCALL call_8_5_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_9_1
#define init_8_5_9_1
void FASTCALL call_8_5_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_9_2
#define init_8_5_9_2
void FASTCALL call_8_5_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_9_3
#define init_8_5_9_3
void FASTCALL call_8_5_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_9_4
#define init_8_5_9_4
void FASTCALL call_8_5_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_9_5
#define init_8_5_9_5
void FASTCALL call_8_5_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_9_6
#define init_8_5_9_6
void FASTCALL call_8_5_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_9_7
#define init_8_5_9_7
void FASTCALL call_8_5_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_9_8
#define init_8_5_9_8
void FASTCALL call_8_5_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_9_9
#define init_8_5_9_9
void FASTCALL call_8_5_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_9_10
#define init_8_5_9_10
void FASTCALL call_8_5_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_9_11
#define init_8_5_9_11
void FASTCALL call_8_5_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_9_12
#define init_8_5_9_12
void FASTCALL call_8_5_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_9_13
#define init_8_5_9_13
void FASTCALL call_8_5_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_9_14
#define init_8_5_9_14
void FASTCALL call_8_5_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_9_15
#define init_8_5_9_15
void FASTCALL call_8_5_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_10_0
#define init_8_5_10_0
void FASTCALL call_8_5_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_10_1
#define init_8_5_10_1
void FASTCALL call_8_5_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_10_2
#define init_8_5_10_2
void FASTCALL call_8_5_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_10_3
#define init_8_5_10_3
void FASTCALL call_8_5_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_10_4
#define init_8_5_10_4
void FASTCALL call_8_5_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_10_5
#define init_8_5_10_5
void FASTCALL call_8_5_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_10_6
#define init_8_5_10_6
void FASTCALL call_8_5_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_10_7
#define init_8_5_10_7
void FASTCALL call_8_5_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_10_8
#define init_8_5_10_8
void FASTCALL call_8_5_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_10_9
#define init_8_5_10_9
void FASTCALL call_8_5_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_10_10
#define init_8_5_10_10
void FASTCALL call_8_5_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_10_11
#define init_8_5_10_11
void FASTCALL call_8_5_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_10_12
#define init_8_5_10_12
void FASTCALL call_8_5_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_10_13
#define init_8_5_10_13
void FASTCALL call_8_5_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_10_14
#define init_8_5_10_14
void FASTCALL call_8_5_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_10_15
#define init_8_5_10_15
void FASTCALL call_8_5_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_11_0
#define init_8_5_11_0
void FASTCALL call_8_5_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_11_1
#define init_8_5_11_1
void FASTCALL call_8_5_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_11_2
#define init_8_5_11_2
void FASTCALL call_8_5_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_11_3
#define init_8_5_11_3
void FASTCALL call_8_5_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_11_4
#define init_8_5_11_4
void FASTCALL call_8_5_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_11_5
#define init_8_5_11_5
void FASTCALL call_8_5_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_11_6
#define init_8_5_11_6
void FASTCALL call_8_5_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_11_7
#define init_8_5_11_7
void FASTCALL call_8_5_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_11_8
#define init_8_5_11_8
void FASTCALL call_8_5_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_11_9
#define init_8_5_11_9
void FASTCALL call_8_5_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_11_10
#define init_8_5_11_10
void FASTCALL call_8_5_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_11_11
#define init_8_5_11_11
void FASTCALL call_8_5_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_11_12
#define init_8_5_11_12
void FASTCALL call_8_5_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_11_13
#define init_8_5_11_13
void FASTCALL call_8_5_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_11_14
#define init_8_5_11_14
void FASTCALL call_8_5_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_11_15
#define init_8_5_11_15
void FASTCALL call_8_5_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_12_0
#define init_8_5_12_0
void FASTCALL call_8_5_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_12_1
#define init_8_5_12_1
void FASTCALL call_8_5_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_12_2
#define init_8_5_12_2
void FASTCALL call_8_5_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_12_3
#define init_8_5_12_3
void FASTCALL call_8_5_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_12_4
#define init_8_5_12_4
void FASTCALL call_8_5_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_12_5
#define init_8_5_12_5
void FASTCALL call_8_5_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_12_6
#define init_8_5_12_6
void FASTCALL call_8_5_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_12_7
#define init_8_5_12_7
void FASTCALL call_8_5_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_12_8
#define init_8_5_12_8
void FASTCALL call_8_5_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_12_9
#define init_8_5_12_9
void FASTCALL call_8_5_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_12_10
#define init_8_5_12_10
void FASTCALL call_8_5_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_12_11
#define init_8_5_12_11
void FASTCALL call_8_5_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_12_12
#define init_8_5_12_12
void FASTCALL call_8_5_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_12_13
#define init_8_5_12_13
void FASTCALL call_8_5_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_12_14
#define init_8_5_12_14
void FASTCALL call_8_5_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_12_15
#define init_8_5_12_15
void FASTCALL call_8_5_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_13_0
#define init_8_5_13_0
void FASTCALL call_8_5_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_13_1
#define init_8_5_13_1
void FASTCALL call_8_5_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_13_2
#define init_8_5_13_2
void FASTCALL call_8_5_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_13_3
#define init_8_5_13_3
void FASTCALL call_8_5_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_13_4
#define init_8_5_13_4
void FASTCALL call_8_5_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_13_5
#define init_8_5_13_5
void FASTCALL call_8_5_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_13_6
#define init_8_5_13_6
void FASTCALL call_8_5_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_13_7
#define init_8_5_13_7
void FASTCALL call_8_5_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_13_8
#define init_8_5_13_8
void FASTCALL call_8_5_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_13_9
#define init_8_5_13_9
void FASTCALL call_8_5_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_13_10
#define init_8_5_13_10
void FASTCALL call_8_5_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_13_11
#define init_8_5_13_11
void FASTCALL call_8_5_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_13_12
#define init_8_5_13_12
void FASTCALL call_8_5_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_13_13
#define init_8_5_13_13
void FASTCALL call_8_5_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_13_14
#define init_8_5_13_14
void FASTCALL call_8_5_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_13_15
#define init_8_5_13_15
void FASTCALL call_8_5_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_14_0
#define init_8_5_14_0
void FASTCALL call_8_5_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_14_1
#define init_8_5_14_1
void FASTCALL call_8_5_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_14_2
#define init_8_5_14_2
void FASTCALL call_8_5_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_14_3
#define init_8_5_14_3
void FASTCALL call_8_5_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_14_4
#define init_8_5_14_4
void FASTCALL call_8_5_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_14_5
#define init_8_5_14_5
void FASTCALL call_8_5_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_14_6
#define init_8_5_14_6
void FASTCALL call_8_5_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_14_7
#define init_8_5_14_7
void FASTCALL call_8_5_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_14_8
#define init_8_5_14_8
void FASTCALL call_8_5_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_14_9
#define init_8_5_14_9
void FASTCALL call_8_5_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_14_10
#define init_8_5_14_10
void FASTCALL call_8_5_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_14_11
#define init_8_5_14_11
void FASTCALL call_8_5_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_14_12
#define init_8_5_14_12
void FASTCALL call_8_5_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_14_13
#define init_8_5_14_13
void FASTCALL call_8_5_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_14_14
#define init_8_5_14_14
void FASTCALL call_8_5_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_14_15
#define init_8_5_14_15
void FASTCALL call_8_5_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_15_0
#define init_8_5_15_0
void FASTCALL call_8_5_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_15_1
#define init_8_5_15_1
void FASTCALL call_8_5_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_15_2
#define init_8_5_15_2
void FASTCALL call_8_5_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_15_3
#define init_8_5_15_3
void FASTCALL call_8_5_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_15_4
#define init_8_5_15_4
void FASTCALL call_8_5_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_15_5
#define init_8_5_15_5
void FASTCALL call_8_5_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_15_6
#define init_8_5_15_6
void FASTCALL call_8_5_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_15_7
#define init_8_5_15_7
void FASTCALL call_8_5_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_15_8
#define init_8_5_15_8
void FASTCALL call_8_5_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_15_9
#define init_8_5_15_9
void FASTCALL call_8_5_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_15_10
#define init_8_5_15_10
void FASTCALL call_8_5_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_15_11
#define init_8_5_15_11
void FASTCALL call_8_5_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_15_12
#define init_8_5_15_12
void FASTCALL call_8_5_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_15_13
#define init_8_5_15_13
void FASTCALL call_8_5_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_15_14
#define init_8_5_15_14
void FASTCALL call_8_5_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_5_15_15
#define init_8_5_15_15
void FASTCALL call_8_5_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_0_0
#define init_8_6_0_0
void FASTCALL call_8_6_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_0_1
#define init_8_6_0_1
void FASTCALL call_8_6_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_0_2
#define init_8_6_0_2
void FASTCALL call_8_6_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_0_3
#define init_8_6_0_3
void FASTCALL call_8_6_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_0_4
#define init_8_6_0_4
void FASTCALL call_8_6_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_0_5
#define init_8_6_0_5
void FASTCALL call_8_6_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_0_6
#define init_8_6_0_6
void FASTCALL call_8_6_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_0_7
#define init_8_6_0_7
void FASTCALL call_8_6_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_0_8
#define init_8_6_0_8
void FASTCALL call_8_6_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_0_9
#define init_8_6_0_9
void FASTCALL call_8_6_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_0_10
#define init_8_6_0_10
void FASTCALL call_8_6_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_0_11
#define init_8_6_0_11
void FASTCALL call_8_6_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_0_12
#define init_8_6_0_12
void FASTCALL call_8_6_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_0_13
#define init_8_6_0_13
void FASTCALL call_8_6_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_0_14
#define init_8_6_0_14
void FASTCALL call_8_6_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_0_15
#define init_8_6_0_15
void FASTCALL call_8_6_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_1_0
#define init_8_6_1_0
void FASTCALL call_8_6_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_1_1
#define init_8_6_1_1
void FASTCALL call_8_6_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_1_2
#define init_8_6_1_2
void FASTCALL call_8_6_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_1_3
#define init_8_6_1_3
void FASTCALL call_8_6_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_1_4
#define init_8_6_1_4
void FASTCALL call_8_6_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_1_5
#define init_8_6_1_5
void FASTCALL call_8_6_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_1_6
#define init_8_6_1_6
void FASTCALL call_8_6_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_1_7
#define init_8_6_1_7
void FASTCALL call_8_6_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_1_8
#define init_8_6_1_8
void FASTCALL call_8_6_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_1_9
#define init_8_6_1_9
void FASTCALL call_8_6_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_1_10
#define init_8_6_1_10
void FASTCALL call_8_6_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_1_11
#define init_8_6_1_11
void FASTCALL call_8_6_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_1_12
#define init_8_6_1_12
void FASTCALL call_8_6_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_1_13
#define init_8_6_1_13
void FASTCALL call_8_6_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_1_14
#define init_8_6_1_14
void FASTCALL call_8_6_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_1_15
#define init_8_6_1_15
void FASTCALL call_8_6_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_2_0
#define init_8_6_2_0
void FASTCALL call_8_6_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_2_1
#define init_8_6_2_1
void FASTCALL call_8_6_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_2_2
#define init_8_6_2_2
void FASTCALL call_8_6_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_2_3
#define init_8_6_2_3
void FASTCALL call_8_6_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_2_4
#define init_8_6_2_4
void FASTCALL call_8_6_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_2_5
#define init_8_6_2_5
void FASTCALL call_8_6_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_2_6
#define init_8_6_2_6
void FASTCALL call_8_6_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_2_7
#define init_8_6_2_7
void FASTCALL call_8_6_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_2_8
#define init_8_6_2_8
void FASTCALL call_8_6_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_2_9
#define init_8_6_2_9
void FASTCALL call_8_6_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_2_10
#define init_8_6_2_10
void FASTCALL call_8_6_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_2_11
#define init_8_6_2_11
void FASTCALL call_8_6_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_2_12
#define init_8_6_2_12
void FASTCALL call_8_6_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_2_13
#define init_8_6_2_13
void FASTCALL call_8_6_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_2_14
#define init_8_6_2_14
void FASTCALL call_8_6_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_2_15
#define init_8_6_2_15
void FASTCALL call_8_6_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_3_0
#define init_8_6_3_0
void FASTCALL call_8_6_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_3_1
#define init_8_6_3_1
void FASTCALL call_8_6_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_3_2
#define init_8_6_3_2
void FASTCALL call_8_6_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_3_3
#define init_8_6_3_3
void FASTCALL call_8_6_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_3_4
#define init_8_6_3_4
void FASTCALL call_8_6_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_3_5
#define init_8_6_3_5
void FASTCALL call_8_6_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_3_6
#define init_8_6_3_6
void FASTCALL call_8_6_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_3_7
#define init_8_6_3_7
void FASTCALL call_8_6_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_3_8
#define init_8_6_3_8
void FASTCALL call_8_6_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_3_9
#define init_8_6_3_9
void FASTCALL call_8_6_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_3_10
#define init_8_6_3_10
void FASTCALL call_8_6_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_3_11
#define init_8_6_3_11
void FASTCALL call_8_6_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_3_12
#define init_8_6_3_12
void FASTCALL call_8_6_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_3_13
#define init_8_6_3_13
void FASTCALL call_8_6_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_3_14
#define init_8_6_3_14
void FASTCALL call_8_6_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_3_15
#define init_8_6_3_15
void FASTCALL call_8_6_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_4_0
#define init_8_6_4_0
void FASTCALL call_8_6_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_4_1
#define init_8_6_4_1
void FASTCALL call_8_6_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_4_2
#define init_8_6_4_2
void FASTCALL call_8_6_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_4_3
#define init_8_6_4_3
void FASTCALL call_8_6_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_4_4
#define init_8_6_4_4
void FASTCALL call_8_6_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_4_5
#define init_8_6_4_5
void FASTCALL call_8_6_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_4_6
#define init_8_6_4_6
void FASTCALL call_8_6_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_4_7
#define init_8_6_4_7
void FASTCALL call_8_6_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_4_8
#define init_8_6_4_8
void FASTCALL call_8_6_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_4_9
#define init_8_6_4_9
void FASTCALL call_8_6_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_4_10
#define init_8_6_4_10
void FASTCALL call_8_6_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_4_11
#define init_8_6_4_11
void FASTCALL call_8_6_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_4_12
#define init_8_6_4_12
void FASTCALL call_8_6_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_4_13
#define init_8_6_4_13
void FASTCALL call_8_6_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_4_14
#define init_8_6_4_14
void FASTCALL call_8_6_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_4_15
#define init_8_6_4_15
void FASTCALL call_8_6_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_5_0
#define init_8_6_5_0
void FASTCALL call_8_6_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_5_1
#define init_8_6_5_1
void FASTCALL call_8_6_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_5_2
#define init_8_6_5_2
void FASTCALL call_8_6_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_5_3
#define init_8_6_5_3
void FASTCALL call_8_6_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_5_4
#define init_8_6_5_4
void FASTCALL call_8_6_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_5_5
#define init_8_6_5_5
void FASTCALL call_8_6_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_5_6
#define init_8_6_5_6
void FASTCALL call_8_6_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_5_7
#define init_8_6_5_7
void FASTCALL call_8_6_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_5_8
#define init_8_6_5_8
void FASTCALL call_8_6_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_5_9
#define init_8_6_5_9
void FASTCALL call_8_6_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_5_10
#define init_8_6_5_10
void FASTCALL call_8_6_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_5_11
#define init_8_6_5_11
void FASTCALL call_8_6_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_5_12
#define init_8_6_5_12
void FASTCALL call_8_6_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_5_13
#define init_8_6_5_13
void FASTCALL call_8_6_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_5_14
#define init_8_6_5_14
void FASTCALL call_8_6_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_5_15
#define init_8_6_5_15
void FASTCALL call_8_6_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_6_0
#define init_8_6_6_0
void FASTCALL call_8_6_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_6_1
#define init_8_6_6_1
void FASTCALL call_8_6_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_6_2
#define init_8_6_6_2
void FASTCALL call_8_6_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_6_3
#define init_8_6_6_3
void FASTCALL call_8_6_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_6_4
#define init_8_6_6_4
void FASTCALL call_8_6_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_6_5
#define init_8_6_6_5
void FASTCALL call_8_6_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_6_6
#define init_8_6_6_6
void FASTCALL call_8_6_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_6_7
#define init_8_6_6_7
void FASTCALL call_8_6_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_6_8
#define init_8_6_6_8
void FASTCALL call_8_6_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_6_9
#define init_8_6_6_9
void FASTCALL call_8_6_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_6_10
#define init_8_6_6_10
void FASTCALL call_8_6_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_6_11
#define init_8_6_6_11
void FASTCALL call_8_6_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_6_12
#define init_8_6_6_12
void FASTCALL call_8_6_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_6_13
#define init_8_6_6_13
void FASTCALL call_8_6_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_6_14
#define init_8_6_6_14
void FASTCALL call_8_6_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_6_15
#define init_8_6_6_15
void FASTCALL call_8_6_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_7_0
#define init_8_6_7_0
void FASTCALL call_8_6_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_7_1
#define init_8_6_7_1
void FASTCALL call_8_6_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_7_2
#define init_8_6_7_2
void FASTCALL call_8_6_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_7_3
#define init_8_6_7_3
void FASTCALL call_8_6_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_7_4
#define init_8_6_7_4
void FASTCALL call_8_6_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_7_5
#define init_8_6_7_5
void FASTCALL call_8_6_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_7_6
#define init_8_6_7_6
void FASTCALL call_8_6_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_7_7
#define init_8_6_7_7
void FASTCALL call_8_6_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_7_8
#define init_8_6_7_8
void FASTCALL call_8_6_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_7_9
#define init_8_6_7_9
void FASTCALL call_8_6_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_7_10
#define init_8_6_7_10
void FASTCALL call_8_6_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_7_11
#define init_8_6_7_11
void FASTCALL call_8_6_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_7_12
#define init_8_6_7_12
void FASTCALL call_8_6_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_7_13
#define init_8_6_7_13
void FASTCALL call_8_6_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_7_14
#define init_8_6_7_14
void FASTCALL call_8_6_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_7_15
#define init_8_6_7_15
void FASTCALL call_8_6_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_8_0
#define init_8_6_8_0
void FASTCALL call_8_6_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_8_1
#define init_8_6_8_1
void FASTCALL call_8_6_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_8_2
#define init_8_6_8_2
void FASTCALL call_8_6_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_8_3
#define init_8_6_8_3
void FASTCALL call_8_6_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_8_4
#define init_8_6_8_4
void FASTCALL call_8_6_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_8_5
#define init_8_6_8_5
void FASTCALL call_8_6_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_8_6
#define init_8_6_8_6
void FASTCALL call_8_6_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_8_7
#define init_8_6_8_7
void FASTCALL call_8_6_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_8_8
#define init_8_6_8_8
void FASTCALL call_8_6_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_8_9
#define init_8_6_8_9
void FASTCALL call_8_6_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_8_10
#define init_8_6_8_10
void FASTCALL call_8_6_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_8_11
#define init_8_6_8_11
void FASTCALL call_8_6_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_8_12
#define init_8_6_8_12
void FASTCALL call_8_6_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_8_13
#define init_8_6_8_13
void FASTCALL call_8_6_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_8_14
#define init_8_6_8_14
void FASTCALL call_8_6_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_8_15
#define init_8_6_8_15
void FASTCALL call_8_6_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_9_0
#define init_8_6_9_0
void FASTCALL call_8_6_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_9_1
#define init_8_6_9_1
void FASTCALL call_8_6_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_9_2
#define init_8_6_9_2
void FASTCALL call_8_6_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_9_3
#define init_8_6_9_3
void FASTCALL call_8_6_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_9_4
#define init_8_6_9_4
void FASTCALL call_8_6_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_9_5
#define init_8_6_9_5
void FASTCALL call_8_6_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_9_6
#define init_8_6_9_6
void FASTCALL call_8_6_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_9_7
#define init_8_6_9_7
void FASTCALL call_8_6_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_9_8
#define init_8_6_9_8
void FASTCALL call_8_6_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_9_9
#define init_8_6_9_9
void FASTCALL call_8_6_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_9_10
#define init_8_6_9_10
void FASTCALL call_8_6_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_9_11
#define init_8_6_9_11
void FASTCALL call_8_6_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_9_12
#define init_8_6_9_12
void FASTCALL call_8_6_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_9_13
#define init_8_6_9_13
void FASTCALL call_8_6_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_9_14
#define init_8_6_9_14
void FASTCALL call_8_6_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_9_15
#define init_8_6_9_15
void FASTCALL call_8_6_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_10_0
#define init_8_6_10_0
void FASTCALL call_8_6_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_10_1
#define init_8_6_10_1
void FASTCALL call_8_6_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_10_2
#define init_8_6_10_2
void FASTCALL call_8_6_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_10_3
#define init_8_6_10_3
void FASTCALL call_8_6_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_10_4
#define init_8_6_10_4
void FASTCALL call_8_6_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_10_5
#define init_8_6_10_5
void FASTCALL call_8_6_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_10_6
#define init_8_6_10_6
void FASTCALL call_8_6_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_10_7
#define init_8_6_10_7
void FASTCALL call_8_6_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_10_8
#define init_8_6_10_8
void FASTCALL call_8_6_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_10_9
#define init_8_6_10_9
void FASTCALL call_8_6_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_10_10
#define init_8_6_10_10
void FASTCALL call_8_6_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_10_11
#define init_8_6_10_11
void FASTCALL call_8_6_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_10_12
#define init_8_6_10_12
void FASTCALL call_8_6_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_10_13
#define init_8_6_10_13
void FASTCALL call_8_6_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_10_14
#define init_8_6_10_14
void FASTCALL call_8_6_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_10_15
#define init_8_6_10_15
void FASTCALL call_8_6_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_11_0
#define init_8_6_11_0
void FASTCALL call_8_6_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_11_1
#define init_8_6_11_1
void FASTCALL call_8_6_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_11_2
#define init_8_6_11_2
void FASTCALL call_8_6_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_11_3
#define init_8_6_11_3
void FASTCALL call_8_6_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_11_4
#define init_8_6_11_4
void FASTCALL call_8_6_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_11_5
#define init_8_6_11_5
void FASTCALL call_8_6_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_11_6
#define init_8_6_11_6
void FASTCALL call_8_6_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_11_7
#define init_8_6_11_7
void FASTCALL call_8_6_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_11_8
#define init_8_6_11_8
void FASTCALL call_8_6_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_11_9
#define init_8_6_11_9
void FASTCALL call_8_6_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_11_10
#define init_8_6_11_10
void FASTCALL call_8_6_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_11_11
#define init_8_6_11_11
void FASTCALL call_8_6_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_11_12
#define init_8_6_11_12
void FASTCALL call_8_6_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_11_13
#define init_8_6_11_13
void FASTCALL call_8_6_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_11_14
#define init_8_6_11_14
void FASTCALL call_8_6_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_11_15
#define init_8_6_11_15
void FASTCALL call_8_6_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_12_0
#define init_8_6_12_0
void FASTCALL call_8_6_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_12_1
#define init_8_6_12_1
void FASTCALL call_8_6_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_12_2
#define init_8_6_12_2
void FASTCALL call_8_6_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_12_3
#define init_8_6_12_3
void FASTCALL call_8_6_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_12_4
#define init_8_6_12_4
void FASTCALL call_8_6_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_12_5
#define init_8_6_12_5
void FASTCALL call_8_6_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_12_6
#define init_8_6_12_6
void FASTCALL call_8_6_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_12_7
#define init_8_6_12_7
void FASTCALL call_8_6_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_12_8
#define init_8_6_12_8
void FASTCALL call_8_6_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_12_9
#define init_8_6_12_9
void FASTCALL call_8_6_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_12_10
#define init_8_6_12_10
void FASTCALL call_8_6_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_12_11
#define init_8_6_12_11
void FASTCALL call_8_6_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_12_12
#define init_8_6_12_12
void FASTCALL call_8_6_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_12_13
#define init_8_6_12_13
void FASTCALL call_8_6_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_12_14
#define init_8_6_12_14
void FASTCALL call_8_6_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_12_15
#define init_8_6_12_15
void FASTCALL call_8_6_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_13_0
#define init_8_6_13_0
void FASTCALL call_8_6_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_13_1
#define init_8_6_13_1
void FASTCALL call_8_6_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_13_2
#define init_8_6_13_2
void FASTCALL call_8_6_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_13_3
#define init_8_6_13_3
void FASTCALL call_8_6_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_13_4
#define init_8_6_13_4
void FASTCALL call_8_6_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_13_5
#define init_8_6_13_5
void FASTCALL call_8_6_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_13_6
#define init_8_6_13_6
void FASTCALL call_8_6_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_13_7
#define init_8_6_13_7
void FASTCALL call_8_6_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_13_8
#define init_8_6_13_8
void FASTCALL call_8_6_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_13_9
#define init_8_6_13_9
void FASTCALL call_8_6_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_13_10
#define init_8_6_13_10
void FASTCALL call_8_6_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_13_11
#define init_8_6_13_11
void FASTCALL call_8_6_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_13_12
#define init_8_6_13_12
void FASTCALL call_8_6_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_13_13
#define init_8_6_13_13
void FASTCALL call_8_6_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_13_14
#define init_8_6_13_14
void FASTCALL call_8_6_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_13_15
#define init_8_6_13_15
void FASTCALL call_8_6_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_14_0
#define init_8_6_14_0
void FASTCALL call_8_6_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_14_1
#define init_8_6_14_1
void FASTCALL call_8_6_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_14_2
#define init_8_6_14_2
void FASTCALL call_8_6_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_14_3
#define init_8_6_14_3
void FASTCALL call_8_6_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_14_4
#define init_8_6_14_4
void FASTCALL call_8_6_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_14_5
#define init_8_6_14_5
void FASTCALL call_8_6_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_14_6
#define init_8_6_14_6
void FASTCALL call_8_6_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_14_7
#define init_8_6_14_7
void FASTCALL call_8_6_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_14_8
#define init_8_6_14_8
void FASTCALL call_8_6_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_14_9
#define init_8_6_14_9
void FASTCALL call_8_6_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_14_10
#define init_8_6_14_10
void FASTCALL call_8_6_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_14_11
#define init_8_6_14_11
void FASTCALL call_8_6_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_14_12
#define init_8_6_14_12
void FASTCALL call_8_6_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_14_13
#define init_8_6_14_13
void FASTCALL call_8_6_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_14_14
#define init_8_6_14_14
void FASTCALL call_8_6_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_14_15
#define init_8_6_14_15
void FASTCALL call_8_6_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_15_0
#define init_8_6_15_0
void FASTCALL call_8_6_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_15_1
#define init_8_6_15_1
void FASTCALL call_8_6_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_15_2
#define init_8_6_15_2
void FASTCALL call_8_6_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_15_3
#define init_8_6_15_3
void FASTCALL call_8_6_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_15_4
#define init_8_6_15_4
void FASTCALL call_8_6_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_15_5
#define init_8_6_15_5
void FASTCALL call_8_6_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_15_6
#define init_8_6_15_6
void FASTCALL call_8_6_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_15_7
#define init_8_6_15_7
void FASTCALL call_8_6_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_15_8
#define init_8_6_15_8
void FASTCALL call_8_6_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_15_9
#define init_8_6_15_9
void FASTCALL call_8_6_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_15_10
#define init_8_6_15_10
void FASTCALL call_8_6_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_15_11
#define init_8_6_15_11
void FASTCALL call_8_6_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_15_12
#define init_8_6_15_12
void FASTCALL call_8_6_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_15_13
#define init_8_6_15_13
void FASTCALL call_8_6_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_15_14
#define init_8_6_15_14
void FASTCALL call_8_6_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_6_15_15
#define init_8_6_15_15
void FASTCALL call_8_6_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_0_0
#define init_8_7_0_0
void FASTCALL call_8_7_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_0_1
#define init_8_7_0_1
void FASTCALL call_8_7_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_0_2
#define init_8_7_0_2
void FASTCALL call_8_7_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_0_3
#define init_8_7_0_3
void FASTCALL call_8_7_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_0_4
#define init_8_7_0_4
void FASTCALL call_8_7_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_0_5
#define init_8_7_0_5
void FASTCALL call_8_7_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_0_6
#define init_8_7_0_6
void FASTCALL call_8_7_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_0_7
#define init_8_7_0_7
void FASTCALL call_8_7_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_0_8
#define init_8_7_0_8
void FASTCALL call_8_7_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_0_9
#define init_8_7_0_9
void FASTCALL call_8_7_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_0_10
#define init_8_7_0_10
void FASTCALL call_8_7_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_0_11
#define init_8_7_0_11
void FASTCALL call_8_7_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_0_12
#define init_8_7_0_12
void FASTCALL call_8_7_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_0_13
#define init_8_7_0_13
void FASTCALL call_8_7_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_0_14
#define init_8_7_0_14
void FASTCALL call_8_7_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_0_15
#define init_8_7_0_15
void FASTCALL call_8_7_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_1_0
#define init_8_7_1_0
void FASTCALL call_8_7_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_1_1
#define init_8_7_1_1
void FASTCALL call_8_7_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_1_2
#define init_8_7_1_2
void FASTCALL call_8_7_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_1_3
#define init_8_7_1_3
void FASTCALL call_8_7_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_1_4
#define init_8_7_1_4
void FASTCALL call_8_7_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_1_5
#define init_8_7_1_5
void FASTCALL call_8_7_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_1_6
#define init_8_7_1_6
void FASTCALL call_8_7_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_1_7
#define init_8_7_1_7
void FASTCALL call_8_7_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_1_8
#define init_8_7_1_8
void FASTCALL call_8_7_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_1_9
#define init_8_7_1_9
void FASTCALL call_8_7_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_1_10
#define init_8_7_1_10
void FASTCALL call_8_7_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_1_11
#define init_8_7_1_11
void FASTCALL call_8_7_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_1_12
#define init_8_7_1_12
void FASTCALL call_8_7_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_1_13
#define init_8_7_1_13
void FASTCALL call_8_7_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_1_14
#define init_8_7_1_14
void FASTCALL call_8_7_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_1_15
#define init_8_7_1_15
void FASTCALL call_8_7_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_2_0
#define init_8_7_2_0
void FASTCALL call_8_7_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_2_1
#define init_8_7_2_1
void FASTCALL call_8_7_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_2_2
#define init_8_7_2_2
void FASTCALL call_8_7_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_2_3
#define init_8_7_2_3
void FASTCALL call_8_7_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_2_4
#define init_8_7_2_4
void FASTCALL call_8_7_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_2_5
#define init_8_7_2_5
void FASTCALL call_8_7_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_2_6
#define init_8_7_2_6
void FASTCALL call_8_7_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_2_7
#define init_8_7_2_7
void FASTCALL call_8_7_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_2_8
#define init_8_7_2_8
void FASTCALL call_8_7_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_2_9
#define init_8_7_2_9
void FASTCALL call_8_7_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_2_10
#define init_8_7_2_10
void FASTCALL call_8_7_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_2_11
#define init_8_7_2_11
void FASTCALL call_8_7_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_2_12
#define init_8_7_2_12
void FASTCALL call_8_7_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_2_13
#define init_8_7_2_13
void FASTCALL call_8_7_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_2_14
#define init_8_7_2_14
void FASTCALL call_8_7_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_2_15
#define init_8_7_2_15
void FASTCALL call_8_7_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_3_0
#define init_8_7_3_0
void FASTCALL call_8_7_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_3_1
#define init_8_7_3_1
void FASTCALL call_8_7_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_3_2
#define init_8_7_3_2
void FASTCALL call_8_7_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_3_3
#define init_8_7_3_3
void FASTCALL call_8_7_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_3_4
#define init_8_7_3_4
void FASTCALL call_8_7_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_3_5
#define init_8_7_3_5
void FASTCALL call_8_7_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_3_6
#define init_8_7_3_6
void FASTCALL call_8_7_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_3_7
#define init_8_7_3_7
void FASTCALL call_8_7_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_3_8
#define init_8_7_3_8
void FASTCALL call_8_7_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_3_9
#define init_8_7_3_9
void FASTCALL call_8_7_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_3_10
#define init_8_7_3_10
void FASTCALL call_8_7_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_3_11
#define init_8_7_3_11
void FASTCALL call_8_7_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_3_12
#define init_8_7_3_12
void FASTCALL call_8_7_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_3_13
#define init_8_7_3_13
void FASTCALL call_8_7_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_3_14
#define init_8_7_3_14
void FASTCALL call_8_7_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_3_15
#define init_8_7_3_15
void FASTCALL call_8_7_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_4_0
#define init_8_7_4_0
void FASTCALL call_8_7_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_4_1
#define init_8_7_4_1
void FASTCALL call_8_7_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_4_2
#define init_8_7_4_2
void FASTCALL call_8_7_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_4_3
#define init_8_7_4_3
void FASTCALL call_8_7_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_4_4
#define init_8_7_4_4
void FASTCALL call_8_7_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_4_5
#define init_8_7_4_5
void FASTCALL call_8_7_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_4_6
#define init_8_7_4_6
void FASTCALL call_8_7_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_4_7
#define init_8_7_4_7
void FASTCALL call_8_7_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_4_8
#define init_8_7_4_8
void FASTCALL call_8_7_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_4_9
#define init_8_7_4_9
void FASTCALL call_8_7_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_4_10
#define init_8_7_4_10
void FASTCALL call_8_7_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_4_11
#define init_8_7_4_11
void FASTCALL call_8_7_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_4_12
#define init_8_7_4_12
void FASTCALL call_8_7_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_4_13
#define init_8_7_4_13
void FASTCALL call_8_7_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_4_14
#define init_8_7_4_14
void FASTCALL call_8_7_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_4_15
#define init_8_7_4_15
void FASTCALL call_8_7_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_5_0
#define init_8_7_5_0
void FASTCALL call_8_7_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_5_1
#define init_8_7_5_1
void FASTCALL call_8_7_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_5_2
#define init_8_7_5_2
void FASTCALL call_8_7_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_5_3
#define init_8_7_5_3
void FASTCALL call_8_7_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_5_4
#define init_8_7_5_4
void FASTCALL call_8_7_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_5_5
#define init_8_7_5_5
void FASTCALL call_8_7_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_5_6
#define init_8_7_5_6
void FASTCALL call_8_7_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_5_7
#define init_8_7_5_7
void FASTCALL call_8_7_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_5_8
#define init_8_7_5_8
void FASTCALL call_8_7_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_5_9
#define init_8_7_5_9
void FASTCALL call_8_7_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_5_10
#define init_8_7_5_10
void FASTCALL call_8_7_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_5_11
#define init_8_7_5_11
void FASTCALL call_8_7_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_5_12
#define init_8_7_5_12
void FASTCALL call_8_7_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_5_13
#define init_8_7_5_13
void FASTCALL call_8_7_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_5_14
#define init_8_7_5_14
void FASTCALL call_8_7_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_5_15
#define init_8_7_5_15
void FASTCALL call_8_7_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_6_0
#define init_8_7_6_0
void FASTCALL call_8_7_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_6_1
#define init_8_7_6_1
void FASTCALL call_8_7_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_6_2
#define init_8_7_6_2
void FASTCALL call_8_7_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_6_3
#define init_8_7_6_3
void FASTCALL call_8_7_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_6_4
#define init_8_7_6_4
void FASTCALL call_8_7_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_6_5
#define init_8_7_6_5
void FASTCALL call_8_7_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_6_6
#define init_8_7_6_6
void FASTCALL call_8_7_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_6_7
#define init_8_7_6_7
void FASTCALL call_8_7_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_6_8
#define init_8_7_6_8
void FASTCALL call_8_7_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_6_9
#define init_8_7_6_9
void FASTCALL call_8_7_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_6_10
#define init_8_7_6_10
void FASTCALL call_8_7_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_6_11
#define init_8_7_6_11
void FASTCALL call_8_7_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_6_12
#define init_8_7_6_12
void FASTCALL call_8_7_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_6_13
#define init_8_7_6_13
void FASTCALL call_8_7_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_6_14
#define init_8_7_6_14
void FASTCALL call_8_7_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_6_15
#define init_8_7_6_15
void FASTCALL call_8_7_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_7_0
#define init_8_7_7_0
void FASTCALL call_8_7_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_7_1
#define init_8_7_7_1
void FASTCALL call_8_7_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_7_2
#define init_8_7_7_2
void FASTCALL call_8_7_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_7_3
#define init_8_7_7_3
void FASTCALL call_8_7_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_7_4
#define init_8_7_7_4
void FASTCALL call_8_7_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_7_5
#define init_8_7_7_5
void FASTCALL call_8_7_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_7_6
#define init_8_7_7_6
void FASTCALL call_8_7_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_7_7
#define init_8_7_7_7
void FASTCALL call_8_7_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_7_8
#define init_8_7_7_8
void FASTCALL call_8_7_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_7_9
#define init_8_7_7_9
void FASTCALL call_8_7_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_7_10
#define init_8_7_7_10
void FASTCALL call_8_7_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_7_11
#define init_8_7_7_11
void FASTCALL call_8_7_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_7_12
#define init_8_7_7_12
void FASTCALL call_8_7_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_7_13
#define init_8_7_7_13
void FASTCALL call_8_7_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_7_14
#define init_8_7_7_14
void FASTCALL call_8_7_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_7_15
#define init_8_7_7_15
void FASTCALL call_8_7_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_8_0
#define init_8_7_8_0
void FASTCALL call_8_7_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_8_1
#define init_8_7_8_1
void FASTCALL call_8_7_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_8_2
#define init_8_7_8_2
void FASTCALL call_8_7_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_8_3
#define init_8_7_8_3
void FASTCALL call_8_7_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_8_4
#define init_8_7_8_4
void FASTCALL call_8_7_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_8_5
#define init_8_7_8_5
void FASTCALL call_8_7_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_8_6
#define init_8_7_8_6
void FASTCALL call_8_7_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_8_7
#define init_8_7_8_7
void FASTCALL call_8_7_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_8_8
#define init_8_7_8_8
void FASTCALL call_8_7_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_8_9
#define init_8_7_8_9
void FASTCALL call_8_7_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_8_10
#define init_8_7_8_10
void FASTCALL call_8_7_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_8_11
#define init_8_7_8_11
void FASTCALL call_8_7_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_8_12
#define init_8_7_8_12
void FASTCALL call_8_7_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_8_13
#define init_8_7_8_13
void FASTCALL call_8_7_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_8_14
#define init_8_7_8_14
void FASTCALL call_8_7_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_8_15
#define init_8_7_8_15
void FASTCALL call_8_7_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_9_0
#define init_8_7_9_0
void FASTCALL call_8_7_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_9_1
#define init_8_7_9_1
void FASTCALL call_8_7_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_9_2
#define init_8_7_9_2
void FASTCALL call_8_7_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_9_3
#define init_8_7_9_3
void FASTCALL call_8_7_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_9_4
#define init_8_7_9_4
void FASTCALL call_8_7_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_9_5
#define init_8_7_9_5
void FASTCALL call_8_7_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_9_6
#define init_8_7_9_6
void FASTCALL call_8_7_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_9_7
#define init_8_7_9_7
void FASTCALL call_8_7_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_9_8
#define init_8_7_9_8
void FASTCALL call_8_7_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_9_9
#define init_8_7_9_9
void FASTCALL call_8_7_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_9_10
#define init_8_7_9_10
void FASTCALL call_8_7_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_9_11
#define init_8_7_9_11
void FASTCALL call_8_7_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_9_12
#define init_8_7_9_12
void FASTCALL call_8_7_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_9_13
#define init_8_7_9_13
void FASTCALL call_8_7_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_9_14
#define init_8_7_9_14
void FASTCALL call_8_7_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_9_15
#define init_8_7_9_15
void FASTCALL call_8_7_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_10_0
#define init_8_7_10_0
void FASTCALL call_8_7_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_10_1
#define init_8_7_10_1
void FASTCALL call_8_7_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_10_2
#define init_8_7_10_2
void FASTCALL call_8_7_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_10_3
#define init_8_7_10_3
void FASTCALL call_8_7_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_10_4
#define init_8_7_10_4
void FASTCALL call_8_7_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_10_5
#define init_8_7_10_5
void FASTCALL call_8_7_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_10_6
#define init_8_7_10_6
void FASTCALL call_8_7_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_10_7
#define init_8_7_10_7
void FASTCALL call_8_7_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_10_8
#define init_8_7_10_8
void FASTCALL call_8_7_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_10_9
#define init_8_7_10_9
void FASTCALL call_8_7_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_10_10
#define init_8_7_10_10
void FASTCALL call_8_7_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_10_11
#define init_8_7_10_11
void FASTCALL call_8_7_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_10_12
#define init_8_7_10_12
void FASTCALL call_8_7_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_10_13
#define init_8_7_10_13
void FASTCALL call_8_7_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_10_14
#define init_8_7_10_14
void FASTCALL call_8_7_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_10_15
#define init_8_7_10_15
void FASTCALL call_8_7_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_11_0
#define init_8_7_11_0
void FASTCALL call_8_7_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_11_1
#define init_8_7_11_1
void FASTCALL call_8_7_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_11_2
#define init_8_7_11_2
void FASTCALL call_8_7_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_11_3
#define init_8_7_11_3
void FASTCALL call_8_7_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_11_4
#define init_8_7_11_4
void FASTCALL call_8_7_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_11_5
#define init_8_7_11_5
void FASTCALL call_8_7_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_11_6
#define init_8_7_11_6
void FASTCALL call_8_7_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_11_7
#define init_8_7_11_7
void FASTCALL call_8_7_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_11_8
#define init_8_7_11_8
void FASTCALL call_8_7_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_11_9
#define init_8_7_11_9
void FASTCALL call_8_7_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_11_10
#define init_8_7_11_10
void FASTCALL call_8_7_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_11_11
#define init_8_7_11_11
void FASTCALL call_8_7_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_11_12
#define init_8_7_11_12
void FASTCALL call_8_7_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_11_13
#define init_8_7_11_13
void FASTCALL call_8_7_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_11_14
#define init_8_7_11_14
void FASTCALL call_8_7_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_11_15
#define init_8_7_11_15
void FASTCALL call_8_7_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_12_0
#define init_8_7_12_0
void FASTCALL call_8_7_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_12_1
#define init_8_7_12_1
void FASTCALL call_8_7_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_12_2
#define init_8_7_12_2
void FASTCALL call_8_7_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_12_3
#define init_8_7_12_3
void FASTCALL call_8_7_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_12_4
#define init_8_7_12_4
void FASTCALL call_8_7_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_12_5
#define init_8_7_12_5
void FASTCALL call_8_7_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_12_6
#define init_8_7_12_6
void FASTCALL call_8_7_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_12_7
#define init_8_7_12_7
void FASTCALL call_8_7_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_12_8
#define init_8_7_12_8
void FASTCALL call_8_7_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_12_9
#define init_8_7_12_9
void FASTCALL call_8_7_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_12_10
#define init_8_7_12_10
void FASTCALL call_8_7_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_12_11
#define init_8_7_12_11
void FASTCALL call_8_7_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_12_12
#define init_8_7_12_12
void FASTCALL call_8_7_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_12_13
#define init_8_7_12_13
void FASTCALL call_8_7_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_12_14
#define init_8_7_12_14
void FASTCALL call_8_7_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_12_15
#define init_8_7_12_15
void FASTCALL call_8_7_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_13_0
#define init_8_7_13_0
void FASTCALL call_8_7_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_13_1
#define init_8_7_13_1
void FASTCALL call_8_7_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_13_2
#define init_8_7_13_2
void FASTCALL call_8_7_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_13_3
#define init_8_7_13_3
void FASTCALL call_8_7_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_13_4
#define init_8_7_13_4
void FASTCALL call_8_7_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_13_5
#define init_8_7_13_5
void FASTCALL call_8_7_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_13_6
#define init_8_7_13_6
void FASTCALL call_8_7_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_13_7
#define init_8_7_13_7
void FASTCALL call_8_7_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_13_8
#define init_8_7_13_8
void FASTCALL call_8_7_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_13_9
#define init_8_7_13_9
void FASTCALL call_8_7_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_13_10
#define init_8_7_13_10
void FASTCALL call_8_7_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_13_11
#define init_8_7_13_11
void FASTCALL call_8_7_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_13_12
#define init_8_7_13_12
void FASTCALL call_8_7_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_13_13
#define init_8_7_13_13
void FASTCALL call_8_7_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_13_14
#define init_8_7_13_14
void FASTCALL call_8_7_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_13_15
#define init_8_7_13_15
void FASTCALL call_8_7_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_14_0
#define init_8_7_14_0
void FASTCALL call_8_7_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_14_1
#define init_8_7_14_1
void FASTCALL call_8_7_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_14_2
#define init_8_7_14_2
void FASTCALL call_8_7_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_14_3
#define init_8_7_14_3
void FASTCALL call_8_7_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_14_4
#define init_8_7_14_4
void FASTCALL call_8_7_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_14_5
#define init_8_7_14_5
void FASTCALL call_8_7_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_14_6
#define init_8_7_14_6
void FASTCALL call_8_7_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_14_7
#define init_8_7_14_7
void FASTCALL call_8_7_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_14_8
#define init_8_7_14_8
void FASTCALL call_8_7_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_14_9
#define init_8_7_14_9
void FASTCALL call_8_7_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_14_10
#define init_8_7_14_10
void FASTCALL call_8_7_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_14_11
#define init_8_7_14_11
void FASTCALL call_8_7_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_14_12
#define init_8_7_14_12
void FASTCALL call_8_7_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_14_13
#define init_8_7_14_13
void FASTCALL call_8_7_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_14_14
#define init_8_7_14_14
void FASTCALL call_8_7_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_14_15
#define init_8_7_14_15
void FASTCALL call_8_7_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_15_0
#define init_8_7_15_0
void FASTCALL call_8_7_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_15_1
#define init_8_7_15_1
void FASTCALL call_8_7_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_15_2
#define init_8_7_15_2
void FASTCALL call_8_7_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_15_3
#define init_8_7_15_3
void FASTCALL call_8_7_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_15_4
#define init_8_7_15_4
void FASTCALL call_8_7_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_15_5
#define init_8_7_15_5
void FASTCALL call_8_7_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_15_6
#define init_8_7_15_6
void FASTCALL call_8_7_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_15_7
#define init_8_7_15_7
void FASTCALL call_8_7_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_15_8
#define init_8_7_15_8
void FASTCALL call_8_7_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_15_9
#define init_8_7_15_9
void FASTCALL call_8_7_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_15_10
#define init_8_7_15_10
void FASTCALL call_8_7_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_15_11
#define init_8_7_15_11
void FASTCALL call_8_7_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_15_12
#define init_8_7_15_12
void FASTCALL call_8_7_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_15_13
#define init_8_7_15_13
void FASTCALL call_8_7_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_15_14
#define init_8_7_15_14
void FASTCALL call_8_7_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_7_15_15
#define init_8_7_15_15
void FASTCALL call_8_7_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_0_0
#define init_8_8_0_0
void FASTCALL call_8_8_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_0_1
#define init_8_8_0_1
void FASTCALL call_8_8_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_0_2
#define init_8_8_0_2
void FASTCALL call_8_8_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_0_3
#define init_8_8_0_3
void FASTCALL call_8_8_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_0_4
#define init_8_8_0_4
void FASTCALL call_8_8_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_0_5
#define init_8_8_0_5
void FASTCALL call_8_8_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_0_6
#define init_8_8_0_6
void FASTCALL call_8_8_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_0_7
#define init_8_8_0_7
void FASTCALL call_8_8_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_0_8
#define init_8_8_0_8
void FASTCALL call_8_8_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_0_9
#define init_8_8_0_9
void FASTCALL call_8_8_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_0_10
#define init_8_8_0_10
void FASTCALL call_8_8_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_0_11
#define init_8_8_0_11
void FASTCALL call_8_8_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_0_12
#define init_8_8_0_12
void FASTCALL call_8_8_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_0_13
#define init_8_8_0_13
void FASTCALL call_8_8_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_0_14
#define init_8_8_0_14
void FASTCALL call_8_8_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_0_15
#define init_8_8_0_15
void FASTCALL call_8_8_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_1_0
#define init_8_8_1_0
void FASTCALL call_8_8_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_1_1
#define init_8_8_1_1
void FASTCALL call_8_8_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_1_2
#define init_8_8_1_2
void FASTCALL call_8_8_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_1_3
#define init_8_8_1_3
void FASTCALL call_8_8_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_1_4
#define init_8_8_1_4
void FASTCALL call_8_8_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_1_5
#define init_8_8_1_5
void FASTCALL call_8_8_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_1_6
#define init_8_8_1_6
void FASTCALL call_8_8_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_1_7
#define init_8_8_1_7
void FASTCALL call_8_8_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_1_8
#define init_8_8_1_8
void FASTCALL call_8_8_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_1_9
#define init_8_8_1_9
void FASTCALL call_8_8_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_1_10
#define init_8_8_1_10
void FASTCALL call_8_8_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_1_11
#define init_8_8_1_11
void FASTCALL call_8_8_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_1_12
#define init_8_8_1_12
void FASTCALL call_8_8_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_1_13
#define init_8_8_1_13
void FASTCALL call_8_8_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_1_14
#define init_8_8_1_14
void FASTCALL call_8_8_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_1_15
#define init_8_8_1_15
void FASTCALL call_8_8_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_2_0
#define init_8_8_2_0
void FASTCALL call_8_8_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_2_1
#define init_8_8_2_1
void FASTCALL call_8_8_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_2_2
#define init_8_8_2_2
void FASTCALL call_8_8_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_2_3
#define init_8_8_2_3
void FASTCALL call_8_8_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_2_4
#define init_8_8_2_4
void FASTCALL call_8_8_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_2_5
#define init_8_8_2_5
void FASTCALL call_8_8_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_2_6
#define init_8_8_2_6
void FASTCALL call_8_8_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_2_7
#define init_8_8_2_7
void FASTCALL call_8_8_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_2_8
#define init_8_8_2_8
void FASTCALL call_8_8_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_2_9
#define init_8_8_2_9
void FASTCALL call_8_8_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_2_10
#define init_8_8_2_10
void FASTCALL call_8_8_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_2_11
#define init_8_8_2_11
void FASTCALL call_8_8_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_2_12
#define init_8_8_2_12
void FASTCALL call_8_8_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_2_13
#define init_8_8_2_13
void FASTCALL call_8_8_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_2_14
#define init_8_8_2_14
void FASTCALL call_8_8_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_2_15
#define init_8_8_2_15
void FASTCALL call_8_8_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_3_0
#define init_8_8_3_0
void FASTCALL call_8_8_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_3_1
#define init_8_8_3_1
void FASTCALL call_8_8_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_3_2
#define init_8_8_3_2
void FASTCALL call_8_8_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_3_3
#define init_8_8_3_3
void FASTCALL call_8_8_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_3_4
#define init_8_8_3_4
void FASTCALL call_8_8_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_3_5
#define init_8_8_3_5
void FASTCALL call_8_8_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_3_6
#define init_8_8_3_6
void FASTCALL call_8_8_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_3_7
#define init_8_8_3_7
void FASTCALL call_8_8_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_3_8
#define init_8_8_3_8
void FASTCALL call_8_8_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_3_9
#define init_8_8_3_9
void FASTCALL call_8_8_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_3_10
#define init_8_8_3_10
void FASTCALL call_8_8_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_3_11
#define init_8_8_3_11
void FASTCALL call_8_8_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_3_12
#define init_8_8_3_12
void FASTCALL call_8_8_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_3_13
#define init_8_8_3_13
void FASTCALL call_8_8_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_3_14
#define init_8_8_3_14
void FASTCALL call_8_8_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_3_15
#define init_8_8_3_15
void FASTCALL call_8_8_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_4_0
#define init_8_8_4_0
void FASTCALL call_8_8_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_4_1
#define init_8_8_4_1
void FASTCALL call_8_8_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_4_2
#define init_8_8_4_2
void FASTCALL call_8_8_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_4_3
#define init_8_8_4_3
void FASTCALL call_8_8_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_4_4
#define init_8_8_4_4
void FASTCALL call_8_8_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_4_5
#define init_8_8_4_5
void FASTCALL call_8_8_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_4_6
#define init_8_8_4_6
void FASTCALL call_8_8_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_4_7
#define init_8_8_4_7
void FASTCALL call_8_8_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_4_8
#define init_8_8_4_8
void FASTCALL call_8_8_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_4_9
#define init_8_8_4_9
void FASTCALL call_8_8_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_4_10
#define init_8_8_4_10
void FASTCALL call_8_8_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_4_11
#define init_8_8_4_11
void FASTCALL call_8_8_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_4_12
#define init_8_8_4_12
void FASTCALL call_8_8_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_4_13
#define init_8_8_4_13
void FASTCALL call_8_8_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_4_14
#define init_8_8_4_14
void FASTCALL call_8_8_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_4_15
#define init_8_8_4_15
void FASTCALL call_8_8_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_5_0
#define init_8_8_5_0
void FASTCALL call_8_8_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_5_1
#define init_8_8_5_1
void FASTCALL call_8_8_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_5_2
#define init_8_8_5_2
void FASTCALL call_8_8_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_5_3
#define init_8_8_5_3
void FASTCALL call_8_8_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_5_4
#define init_8_8_5_4
void FASTCALL call_8_8_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_5_5
#define init_8_8_5_5
void FASTCALL call_8_8_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_5_6
#define init_8_8_5_6
void FASTCALL call_8_8_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_5_7
#define init_8_8_5_7
void FASTCALL call_8_8_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_5_8
#define init_8_8_5_8
void FASTCALL call_8_8_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_5_9
#define init_8_8_5_9
void FASTCALL call_8_8_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_5_10
#define init_8_8_5_10
void FASTCALL call_8_8_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_5_11
#define init_8_8_5_11
void FASTCALL call_8_8_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_5_12
#define init_8_8_5_12
void FASTCALL call_8_8_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_5_13
#define init_8_8_5_13
void FASTCALL call_8_8_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_5_14
#define init_8_8_5_14
void FASTCALL call_8_8_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_5_15
#define init_8_8_5_15
void FASTCALL call_8_8_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_6_0
#define init_8_8_6_0
void FASTCALL call_8_8_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_6_1
#define init_8_8_6_1
void FASTCALL call_8_8_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_6_2
#define init_8_8_6_2
void FASTCALL call_8_8_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_6_3
#define init_8_8_6_3
void FASTCALL call_8_8_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_6_4
#define init_8_8_6_4
void FASTCALL call_8_8_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_6_5
#define init_8_8_6_5
void FASTCALL call_8_8_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_6_6
#define init_8_8_6_6
void FASTCALL call_8_8_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_6_7
#define init_8_8_6_7
void FASTCALL call_8_8_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_6_8
#define init_8_8_6_8
void FASTCALL call_8_8_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_6_9
#define init_8_8_6_9
void FASTCALL call_8_8_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_6_10
#define init_8_8_6_10
void FASTCALL call_8_8_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_6_11
#define init_8_8_6_11
void FASTCALL call_8_8_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_6_12
#define init_8_8_6_12
void FASTCALL call_8_8_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_6_13
#define init_8_8_6_13
void FASTCALL call_8_8_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_6_14
#define init_8_8_6_14
void FASTCALL call_8_8_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_6_15
#define init_8_8_6_15
void FASTCALL call_8_8_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_7_0
#define init_8_8_7_0
void FASTCALL call_8_8_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_7_1
#define init_8_8_7_1
void FASTCALL call_8_8_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_7_2
#define init_8_8_7_2
void FASTCALL call_8_8_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_7_3
#define init_8_8_7_3
void FASTCALL call_8_8_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_7_4
#define init_8_8_7_4
void FASTCALL call_8_8_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_7_5
#define init_8_8_7_5
void FASTCALL call_8_8_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_7_6
#define init_8_8_7_6
void FASTCALL call_8_8_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_7_7
#define init_8_8_7_7
void FASTCALL call_8_8_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_7_8
#define init_8_8_7_8
void FASTCALL call_8_8_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_7_9
#define init_8_8_7_9
void FASTCALL call_8_8_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_7_10
#define init_8_8_7_10
void FASTCALL call_8_8_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_7_11
#define init_8_8_7_11
void FASTCALL call_8_8_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_7_12
#define init_8_8_7_12
void FASTCALL call_8_8_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_7_13
#define init_8_8_7_13
void FASTCALL call_8_8_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_7_14
#define init_8_8_7_14
void FASTCALL call_8_8_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_7_15
#define init_8_8_7_15
void FASTCALL call_8_8_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_8_0
#define init_8_8_8_0
void FASTCALL call_8_8_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_8_1
#define init_8_8_8_1
void FASTCALL call_8_8_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_8_2
#define init_8_8_8_2
void FASTCALL call_8_8_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_8_3
#define init_8_8_8_3
void FASTCALL call_8_8_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_8_4
#define init_8_8_8_4
void FASTCALL call_8_8_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_8_5
#define init_8_8_8_5
void FASTCALL call_8_8_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_8_6
#define init_8_8_8_6
void FASTCALL call_8_8_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_8_7
#define init_8_8_8_7
void FASTCALL call_8_8_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_8_8
#define init_8_8_8_8
void FASTCALL call_8_8_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_8_9
#define init_8_8_8_9
void FASTCALL call_8_8_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_8_10
#define init_8_8_8_10
void FASTCALL call_8_8_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_8_11
#define init_8_8_8_11
void FASTCALL call_8_8_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_8_12
#define init_8_8_8_12
void FASTCALL call_8_8_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_8_13
#define init_8_8_8_13
void FASTCALL call_8_8_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_8_14
#define init_8_8_8_14
void FASTCALL call_8_8_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_8_15
#define init_8_8_8_15
void FASTCALL call_8_8_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_9_0
#define init_8_8_9_0
void FASTCALL call_8_8_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_9_1
#define init_8_8_9_1
void FASTCALL call_8_8_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_9_2
#define init_8_8_9_2
void FASTCALL call_8_8_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_9_3
#define init_8_8_9_3
void FASTCALL call_8_8_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_9_4
#define init_8_8_9_4
void FASTCALL call_8_8_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_9_5
#define init_8_8_9_5
void FASTCALL call_8_8_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_9_6
#define init_8_8_9_6
void FASTCALL call_8_8_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_9_7
#define init_8_8_9_7
void FASTCALL call_8_8_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_9_8
#define init_8_8_9_8
void FASTCALL call_8_8_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_9_9
#define init_8_8_9_9
void FASTCALL call_8_8_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_9_10
#define init_8_8_9_10
void FASTCALL call_8_8_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_9_11
#define init_8_8_9_11
void FASTCALL call_8_8_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_9_12
#define init_8_8_9_12
void FASTCALL call_8_8_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_9_13
#define init_8_8_9_13
void FASTCALL call_8_8_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_9_14
#define init_8_8_9_14
void FASTCALL call_8_8_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_9_15
#define init_8_8_9_15
void FASTCALL call_8_8_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_10_0
#define init_8_8_10_0
void FASTCALL call_8_8_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_10_1
#define init_8_8_10_1
void FASTCALL call_8_8_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_10_2
#define init_8_8_10_2
void FASTCALL call_8_8_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_10_3
#define init_8_8_10_3
void FASTCALL call_8_8_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_10_4
#define init_8_8_10_4
void FASTCALL call_8_8_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_10_5
#define init_8_8_10_5
void FASTCALL call_8_8_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_10_6
#define init_8_8_10_6
void FASTCALL call_8_8_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_10_7
#define init_8_8_10_7
void FASTCALL call_8_8_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_10_8
#define init_8_8_10_8
void FASTCALL call_8_8_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_10_9
#define init_8_8_10_9
void FASTCALL call_8_8_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_10_10
#define init_8_8_10_10
void FASTCALL call_8_8_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_10_11
#define init_8_8_10_11
void FASTCALL call_8_8_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_10_12
#define init_8_8_10_12
void FASTCALL call_8_8_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_10_13
#define init_8_8_10_13
void FASTCALL call_8_8_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_10_14
#define init_8_8_10_14
void FASTCALL call_8_8_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_10_15
#define init_8_8_10_15
void FASTCALL call_8_8_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_11_0
#define init_8_8_11_0
void FASTCALL call_8_8_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_11_1
#define init_8_8_11_1
void FASTCALL call_8_8_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_11_2
#define init_8_8_11_2
void FASTCALL call_8_8_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_11_3
#define init_8_8_11_3
void FASTCALL call_8_8_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_11_4
#define init_8_8_11_4
void FASTCALL call_8_8_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_11_5
#define init_8_8_11_5
void FASTCALL call_8_8_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_11_6
#define init_8_8_11_6
void FASTCALL call_8_8_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_11_7
#define init_8_8_11_7
void FASTCALL call_8_8_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_11_8
#define init_8_8_11_8
void FASTCALL call_8_8_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_11_9
#define init_8_8_11_9
void FASTCALL call_8_8_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_11_10
#define init_8_8_11_10
void FASTCALL call_8_8_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_11_11
#define init_8_8_11_11
void FASTCALL call_8_8_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_11_12
#define init_8_8_11_12
void FASTCALL call_8_8_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_11_13
#define init_8_8_11_13
void FASTCALL call_8_8_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_11_14
#define init_8_8_11_14
void FASTCALL call_8_8_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_11_15
#define init_8_8_11_15
void FASTCALL call_8_8_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_12_0
#define init_8_8_12_0
void FASTCALL call_8_8_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_12_1
#define init_8_8_12_1
void FASTCALL call_8_8_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_12_2
#define init_8_8_12_2
void FASTCALL call_8_8_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_12_3
#define init_8_8_12_3
void FASTCALL call_8_8_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_12_4
#define init_8_8_12_4
void FASTCALL call_8_8_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_12_5
#define init_8_8_12_5
void FASTCALL call_8_8_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_12_6
#define init_8_8_12_6
void FASTCALL call_8_8_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_12_7
#define init_8_8_12_7
void FASTCALL call_8_8_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_12_8
#define init_8_8_12_8
void FASTCALL call_8_8_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_12_9
#define init_8_8_12_9
void FASTCALL call_8_8_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_12_10
#define init_8_8_12_10
void FASTCALL call_8_8_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_12_11
#define init_8_8_12_11
void FASTCALL call_8_8_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_12_12
#define init_8_8_12_12
void FASTCALL call_8_8_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_12_13
#define init_8_8_12_13
void FASTCALL call_8_8_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_12_14
#define init_8_8_12_14
void FASTCALL call_8_8_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_12_15
#define init_8_8_12_15
void FASTCALL call_8_8_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_13_0
#define init_8_8_13_0
void FASTCALL call_8_8_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_13_1
#define init_8_8_13_1
void FASTCALL call_8_8_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_13_2
#define init_8_8_13_2
void FASTCALL call_8_8_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_13_3
#define init_8_8_13_3
void FASTCALL call_8_8_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_13_4
#define init_8_8_13_4
void FASTCALL call_8_8_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_13_5
#define init_8_8_13_5
void FASTCALL call_8_8_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_13_6
#define init_8_8_13_6
void FASTCALL call_8_8_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_13_7
#define init_8_8_13_7
void FASTCALL call_8_8_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_13_8
#define init_8_8_13_8
void FASTCALL call_8_8_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_13_9
#define init_8_8_13_9
void FASTCALL call_8_8_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_13_10
#define init_8_8_13_10
void FASTCALL call_8_8_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_13_11
#define init_8_8_13_11
void FASTCALL call_8_8_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_13_12
#define init_8_8_13_12
void FASTCALL call_8_8_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_13_13
#define init_8_8_13_13
void FASTCALL call_8_8_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_13_14
#define init_8_8_13_14
void FASTCALL call_8_8_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_13_15
#define init_8_8_13_15
void FASTCALL call_8_8_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_14_0
#define init_8_8_14_0
void FASTCALL call_8_8_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_14_1
#define init_8_8_14_1
void FASTCALL call_8_8_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_14_2
#define init_8_8_14_2
void FASTCALL call_8_8_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_14_3
#define init_8_8_14_3
void FASTCALL call_8_8_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_14_4
#define init_8_8_14_4
void FASTCALL call_8_8_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_14_5
#define init_8_8_14_5
void FASTCALL call_8_8_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_14_6
#define init_8_8_14_6
void FASTCALL call_8_8_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_14_7
#define init_8_8_14_7
void FASTCALL call_8_8_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_14_8
#define init_8_8_14_8
void FASTCALL call_8_8_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_14_9
#define init_8_8_14_9
void FASTCALL call_8_8_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_14_10
#define init_8_8_14_10
void FASTCALL call_8_8_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_14_11
#define init_8_8_14_11
void FASTCALL call_8_8_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_14_12
#define init_8_8_14_12
void FASTCALL call_8_8_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_14_13
#define init_8_8_14_13
void FASTCALL call_8_8_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_14_14
#define init_8_8_14_14
void FASTCALL call_8_8_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_14_15
#define init_8_8_14_15
void FASTCALL call_8_8_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_15_0
#define init_8_8_15_0
void FASTCALL call_8_8_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_15_1
#define init_8_8_15_1
void FASTCALL call_8_8_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_15_2
#define init_8_8_15_2
void FASTCALL call_8_8_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_15_3
#define init_8_8_15_3
void FASTCALL call_8_8_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_15_4
#define init_8_8_15_4
void FASTCALL call_8_8_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_15_5
#define init_8_8_15_5
void FASTCALL call_8_8_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_15_6
#define init_8_8_15_6
void FASTCALL call_8_8_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_15_7
#define init_8_8_15_7
void FASTCALL call_8_8_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_15_8
#define init_8_8_15_8
void FASTCALL call_8_8_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_15_9
#define init_8_8_15_9
void FASTCALL call_8_8_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_15_10
#define init_8_8_15_10
void FASTCALL call_8_8_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_15_11
#define init_8_8_15_11
void FASTCALL call_8_8_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_15_12
#define init_8_8_15_12
void FASTCALL call_8_8_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_15_13
#define init_8_8_15_13
void FASTCALL call_8_8_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_15_14
#define init_8_8_15_14
void FASTCALL call_8_8_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_8_15_15
#define init_8_8_15_15
void FASTCALL call_8_8_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_0_0
#define init_8_9_0_0
void FASTCALL call_8_9_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_0_1
#define init_8_9_0_1
void FASTCALL call_8_9_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_0_2
#define init_8_9_0_2
void FASTCALL call_8_9_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_0_3
#define init_8_9_0_3
void FASTCALL call_8_9_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_0_4
#define init_8_9_0_4
void FASTCALL call_8_9_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_0_5
#define init_8_9_0_5
void FASTCALL call_8_9_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_0_6
#define init_8_9_0_6
void FASTCALL call_8_9_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_0_7
#define init_8_9_0_7
void FASTCALL call_8_9_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_0_8
#define init_8_9_0_8
void FASTCALL call_8_9_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_0_9
#define init_8_9_0_9
void FASTCALL call_8_9_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_0_10
#define init_8_9_0_10
void FASTCALL call_8_9_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_0_11
#define init_8_9_0_11
void FASTCALL call_8_9_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_0_12
#define init_8_9_0_12
void FASTCALL call_8_9_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_0_13
#define init_8_9_0_13
void FASTCALL call_8_9_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_0_14
#define init_8_9_0_14
void FASTCALL call_8_9_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_0_15
#define init_8_9_0_15
void FASTCALL call_8_9_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_1_0
#define init_8_9_1_0
void FASTCALL call_8_9_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_1_1
#define init_8_9_1_1
void FASTCALL call_8_9_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_1_2
#define init_8_9_1_2
void FASTCALL call_8_9_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_1_3
#define init_8_9_1_3
void FASTCALL call_8_9_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_1_4
#define init_8_9_1_4
void FASTCALL call_8_9_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_1_5
#define init_8_9_1_5
void FASTCALL call_8_9_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_1_6
#define init_8_9_1_6
void FASTCALL call_8_9_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_1_7
#define init_8_9_1_7
void FASTCALL call_8_9_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_1_8
#define init_8_9_1_8
void FASTCALL call_8_9_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_1_9
#define init_8_9_1_9
void FASTCALL call_8_9_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_1_10
#define init_8_9_1_10
void FASTCALL call_8_9_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_1_11
#define init_8_9_1_11
void FASTCALL call_8_9_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_1_12
#define init_8_9_1_12
void FASTCALL call_8_9_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_1_13
#define init_8_9_1_13
void FASTCALL call_8_9_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_1_14
#define init_8_9_1_14
void FASTCALL call_8_9_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_1_15
#define init_8_9_1_15
void FASTCALL call_8_9_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_2_0
#define init_8_9_2_0
void FASTCALL call_8_9_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_2_1
#define init_8_9_2_1
void FASTCALL call_8_9_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_2_2
#define init_8_9_2_2
void FASTCALL call_8_9_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_2_3
#define init_8_9_2_3
void FASTCALL call_8_9_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_2_4
#define init_8_9_2_4
void FASTCALL call_8_9_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_2_5
#define init_8_9_2_5
void FASTCALL call_8_9_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_2_6
#define init_8_9_2_6
void FASTCALL call_8_9_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_2_7
#define init_8_9_2_7
void FASTCALL call_8_9_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_2_8
#define init_8_9_2_8
void FASTCALL call_8_9_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_2_9
#define init_8_9_2_9
void FASTCALL call_8_9_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_2_10
#define init_8_9_2_10
void FASTCALL call_8_9_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_2_11
#define init_8_9_2_11
void FASTCALL call_8_9_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_2_12
#define init_8_9_2_12
void FASTCALL call_8_9_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_2_13
#define init_8_9_2_13
void FASTCALL call_8_9_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_2_14
#define init_8_9_2_14
void FASTCALL call_8_9_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_2_15
#define init_8_9_2_15
void FASTCALL call_8_9_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_3_0
#define init_8_9_3_0
void FASTCALL call_8_9_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_3_1
#define init_8_9_3_1
void FASTCALL call_8_9_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_3_2
#define init_8_9_3_2
void FASTCALL call_8_9_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_3_3
#define init_8_9_3_3
void FASTCALL call_8_9_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_3_4
#define init_8_9_3_4
void FASTCALL call_8_9_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_3_5
#define init_8_9_3_5
void FASTCALL call_8_9_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_3_6
#define init_8_9_3_6
void FASTCALL call_8_9_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_3_7
#define init_8_9_3_7
void FASTCALL call_8_9_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_3_8
#define init_8_9_3_8
void FASTCALL call_8_9_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_3_9
#define init_8_9_3_9
void FASTCALL call_8_9_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_3_10
#define init_8_9_3_10
void FASTCALL call_8_9_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_3_11
#define init_8_9_3_11
void FASTCALL call_8_9_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_3_12
#define init_8_9_3_12
void FASTCALL call_8_9_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_3_13
#define init_8_9_3_13
void FASTCALL call_8_9_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_3_14
#define init_8_9_3_14
void FASTCALL call_8_9_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_3_15
#define init_8_9_3_15
void FASTCALL call_8_9_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_4_0
#define init_8_9_4_0
void FASTCALL call_8_9_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_4_1
#define init_8_9_4_1
void FASTCALL call_8_9_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_4_2
#define init_8_9_4_2
void FASTCALL call_8_9_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_4_3
#define init_8_9_4_3
void FASTCALL call_8_9_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_4_4
#define init_8_9_4_4
void FASTCALL call_8_9_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_4_5
#define init_8_9_4_5
void FASTCALL call_8_9_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_4_6
#define init_8_9_4_6
void FASTCALL call_8_9_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_4_7
#define init_8_9_4_7
void FASTCALL call_8_9_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_4_8
#define init_8_9_4_8
void FASTCALL call_8_9_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_4_9
#define init_8_9_4_9
void FASTCALL call_8_9_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_4_10
#define init_8_9_4_10
void FASTCALL call_8_9_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_4_11
#define init_8_9_4_11
void FASTCALL call_8_9_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_4_12
#define init_8_9_4_12
void FASTCALL call_8_9_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_4_13
#define init_8_9_4_13
void FASTCALL call_8_9_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_4_14
#define init_8_9_4_14
void FASTCALL call_8_9_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_4_15
#define init_8_9_4_15
void FASTCALL call_8_9_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_5_0
#define init_8_9_5_0
void FASTCALL call_8_9_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_5_1
#define init_8_9_5_1
void FASTCALL call_8_9_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_5_2
#define init_8_9_5_2
void FASTCALL call_8_9_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_5_3
#define init_8_9_5_3
void FASTCALL call_8_9_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_5_4
#define init_8_9_5_4
void FASTCALL call_8_9_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_5_5
#define init_8_9_5_5
void FASTCALL call_8_9_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_5_6
#define init_8_9_5_6
void FASTCALL call_8_9_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_5_7
#define init_8_9_5_7
void FASTCALL call_8_9_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_5_8
#define init_8_9_5_8
void FASTCALL call_8_9_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_5_9
#define init_8_9_5_9
void FASTCALL call_8_9_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_5_10
#define init_8_9_5_10
void FASTCALL call_8_9_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_5_11
#define init_8_9_5_11
void FASTCALL call_8_9_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_5_12
#define init_8_9_5_12
void FASTCALL call_8_9_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_5_13
#define init_8_9_5_13
void FASTCALL call_8_9_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_5_14
#define init_8_9_5_14
void FASTCALL call_8_9_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_5_15
#define init_8_9_5_15
void FASTCALL call_8_9_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_6_0
#define init_8_9_6_0
void FASTCALL call_8_9_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_6_1
#define init_8_9_6_1
void FASTCALL call_8_9_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_6_2
#define init_8_9_6_2
void FASTCALL call_8_9_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_6_3
#define init_8_9_6_3
void FASTCALL call_8_9_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_6_4
#define init_8_9_6_4
void FASTCALL call_8_9_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_6_5
#define init_8_9_6_5
void FASTCALL call_8_9_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_6_6
#define init_8_9_6_6
void FASTCALL call_8_9_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_6_7
#define init_8_9_6_7
void FASTCALL call_8_9_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_6_8
#define init_8_9_6_8
void FASTCALL call_8_9_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_6_9
#define init_8_9_6_9
void FASTCALL call_8_9_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_6_10
#define init_8_9_6_10
void FASTCALL call_8_9_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_6_11
#define init_8_9_6_11
void FASTCALL call_8_9_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_6_12
#define init_8_9_6_12
void FASTCALL call_8_9_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_6_13
#define init_8_9_6_13
void FASTCALL call_8_9_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_6_14
#define init_8_9_6_14
void FASTCALL call_8_9_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_6_15
#define init_8_9_6_15
void FASTCALL call_8_9_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_7_0
#define init_8_9_7_0
void FASTCALL call_8_9_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_7_1
#define init_8_9_7_1
void FASTCALL call_8_9_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_7_2
#define init_8_9_7_2
void FASTCALL call_8_9_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_7_3
#define init_8_9_7_3
void FASTCALL call_8_9_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_7_4
#define init_8_9_7_4
void FASTCALL call_8_9_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_7_5
#define init_8_9_7_5
void FASTCALL call_8_9_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_7_6
#define init_8_9_7_6
void FASTCALL call_8_9_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_7_7
#define init_8_9_7_7
void FASTCALL call_8_9_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_7_8
#define init_8_9_7_8
void FASTCALL call_8_9_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_7_9
#define init_8_9_7_9
void FASTCALL call_8_9_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_7_10
#define init_8_9_7_10
void FASTCALL call_8_9_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_7_11
#define init_8_9_7_11
void FASTCALL call_8_9_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_7_12
#define init_8_9_7_12
void FASTCALL call_8_9_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_7_13
#define init_8_9_7_13
void FASTCALL call_8_9_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_7_14
#define init_8_9_7_14
void FASTCALL call_8_9_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_7_15
#define init_8_9_7_15
void FASTCALL call_8_9_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_8_0
#define init_8_9_8_0
void FASTCALL call_8_9_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_8_1
#define init_8_9_8_1
void FASTCALL call_8_9_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_8_2
#define init_8_9_8_2
void FASTCALL call_8_9_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_8_3
#define init_8_9_8_3
void FASTCALL call_8_9_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_8_4
#define init_8_9_8_4
void FASTCALL call_8_9_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_8_5
#define init_8_9_8_5
void FASTCALL call_8_9_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_8_6
#define init_8_9_8_6
void FASTCALL call_8_9_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_8_7
#define init_8_9_8_7
void FASTCALL call_8_9_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_8_8
#define init_8_9_8_8
void FASTCALL call_8_9_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_8_9
#define init_8_9_8_9
void FASTCALL call_8_9_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_8_10
#define init_8_9_8_10
void FASTCALL call_8_9_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_8_11
#define init_8_9_8_11
void FASTCALL call_8_9_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_8_12
#define init_8_9_8_12
void FASTCALL call_8_9_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_8_13
#define init_8_9_8_13
void FASTCALL call_8_9_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_8_14
#define init_8_9_8_14
void FASTCALL call_8_9_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_8_15
#define init_8_9_8_15
void FASTCALL call_8_9_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_9_0
#define init_8_9_9_0
void FASTCALL call_8_9_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_9_1
#define init_8_9_9_1
void FASTCALL call_8_9_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_9_2
#define init_8_9_9_2
void FASTCALL call_8_9_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_9_3
#define init_8_9_9_3
void FASTCALL call_8_9_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_9_4
#define init_8_9_9_4
void FASTCALL call_8_9_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_9_5
#define init_8_9_9_5
void FASTCALL call_8_9_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_9_6
#define init_8_9_9_6
void FASTCALL call_8_9_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_9_7
#define init_8_9_9_7
void FASTCALL call_8_9_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_9_8
#define init_8_9_9_8
void FASTCALL call_8_9_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_9_9
#define init_8_9_9_9
void FASTCALL call_8_9_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_9_10
#define init_8_9_9_10
void FASTCALL call_8_9_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_9_11
#define init_8_9_9_11
void FASTCALL call_8_9_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_9_12
#define init_8_9_9_12
void FASTCALL call_8_9_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_9_13
#define init_8_9_9_13
void FASTCALL call_8_9_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_9_14
#define init_8_9_9_14
void FASTCALL call_8_9_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_9_15
#define init_8_9_9_15
void FASTCALL call_8_9_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_10_0
#define init_8_9_10_0
void FASTCALL call_8_9_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_10_1
#define init_8_9_10_1
void FASTCALL call_8_9_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_10_2
#define init_8_9_10_2
void FASTCALL call_8_9_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_10_3
#define init_8_9_10_3
void FASTCALL call_8_9_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_10_4
#define init_8_9_10_4
void FASTCALL call_8_9_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_10_5
#define init_8_9_10_5
void FASTCALL call_8_9_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_10_6
#define init_8_9_10_6
void FASTCALL call_8_9_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_10_7
#define init_8_9_10_7
void FASTCALL call_8_9_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_10_8
#define init_8_9_10_8
void FASTCALL call_8_9_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_10_9
#define init_8_9_10_9
void FASTCALL call_8_9_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_10_10
#define init_8_9_10_10
void FASTCALL call_8_9_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_10_11
#define init_8_9_10_11
void FASTCALL call_8_9_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_10_12
#define init_8_9_10_12
void FASTCALL call_8_9_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_10_13
#define init_8_9_10_13
void FASTCALL call_8_9_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_10_14
#define init_8_9_10_14
void FASTCALL call_8_9_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_10_15
#define init_8_9_10_15
void FASTCALL call_8_9_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_11_0
#define init_8_9_11_0
void FASTCALL call_8_9_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_11_1
#define init_8_9_11_1
void FASTCALL call_8_9_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_11_2
#define init_8_9_11_2
void FASTCALL call_8_9_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_11_3
#define init_8_9_11_3
void FASTCALL call_8_9_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_11_4
#define init_8_9_11_4
void FASTCALL call_8_9_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_11_5
#define init_8_9_11_5
void FASTCALL call_8_9_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_11_6
#define init_8_9_11_6
void FASTCALL call_8_9_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_11_7
#define init_8_9_11_7
void FASTCALL call_8_9_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_11_8
#define init_8_9_11_8
void FASTCALL call_8_9_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_11_9
#define init_8_9_11_9
void FASTCALL call_8_9_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_11_10
#define init_8_9_11_10
void FASTCALL call_8_9_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_11_11
#define init_8_9_11_11
void FASTCALL call_8_9_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_11_12
#define init_8_9_11_12
void FASTCALL call_8_9_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_11_13
#define init_8_9_11_13
void FASTCALL call_8_9_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_11_14
#define init_8_9_11_14
void FASTCALL call_8_9_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_11_15
#define init_8_9_11_15
void FASTCALL call_8_9_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_12_0
#define init_8_9_12_0
void FASTCALL call_8_9_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_12_1
#define init_8_9_12_1
void FASTCALL call_8_9_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_12_2
#define init_8_9_12_2
void FASTCALL call_8_9_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_12_3
#define init_8_9_12_3
void FASTCALL call_8_9_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_12_4
#define init_8_9_12_4
void FASTCALL call_8_9_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_12_5
#define init_8_9_12_5
void FASTCALL call_8_9_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_12_6
#define init_8_9_12_6
void FASTCALL call_8_9_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_12_7
#define init_8_9_12_7
void FASTCALL call_8_9_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_12_8
#define init_8_9_12_8
void FASTCALL call_8_9_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_12_9
#define init_8_9_12_9
void FASTCALL call_8_9_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_12_10
#define init_8_9_12_10
void FASTCALL call_8_9_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_12_11
#define init_8_9_12_11
void FASTCALL call_8_9_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_12_12
#define init_8_9_12_12
void FASTCALL call_8_9_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_12_13
#define init_8_9_12_13
void FASTCALL call_8_9_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_12_14
#define init_8_9_12_14
void FASTCALL call_8_9_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_12_15
#define init_8_9_12_15
void FASTCALL call_8_9_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_13_0
#define init_8_9_13_0
void FASTCALL call_8_9_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_13_1
#define init_8_9_13_1
void FASTCALL call_8_9_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_13_2
#define init_8_9_13_2
void FASTCALL call_8_9_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_13_3
#define init_8_9_13_3
void FASTCALL call_8_9_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_13_4
#define init_8_9_13_4
void FASTCALL call_8_9_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_13_5
#define init_8_9_13_5
void FASTCALL call_8_9_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_13_6
#define init_8_9_13_6
void FASTCALL call_8_9_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_13_7
#define init_8_9_13_7
void FASTCALL call_8_9_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_13_8
#define init_8_9_13_8
void FASTCALL call_8_9_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_13_9
#define init_8_9_13_9
void FASTCALL call_8_9_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_13_10
#define init_8_9_13_10
void FASTCALL call_8_9_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_13_11
#define init_8_9_13_11
void FASTCALL call_8_9_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_13_12
#define init_8_9_13_12
void FASTCALL call_8_9_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_13_13
#define init_8_9_13_13
void FASTCALL call_8_9_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_13_14
#define init_8_9_13_14
void FASTCALL call_8_9_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_13_15
#define init_8_9_13_15
void FASTCALL call_8_9_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_14_0
#define init_8_9_14_0
void FASTCALL call_8_9_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_14_1
#define init_8_9_14_1
void FASTCALL call_8_9_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_14_2
#define init_8_9_14_2
void FASTCALL call_8_9_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_14_3
#define init_8_9_14_3
void FASTCALL call_8_9_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_14_4
#define init_8_9_14_4
void FASTCALL call_8_9_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_14_5
#define init_8_9_14_5
void FASTCALL call_8_9_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_14_6
#define init_8_9_14_6
void FASTCALL call_8_9_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_14_7
#define init_8_9_14_7
void FASTCALL call_8_9_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_14_8
#define init_8_9_14_8
void FASTCALL call_8_9_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_14_9
#define init_8_9_14_9
void FASTCALL call_8_9_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_14_10
#define init_8_9_14_10
void FASTCALL call_8_9_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_14_11
#define init_8_9_14_11
void FASTCALL call_8_9_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_14_12
#define init_8_9_14_12
void FASTCALL call_8_9_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_14_13
#define init_8_9_14_13
void FASTCALL call_8_9_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_14_14
#define init_8_9_14_14
void FASTCALL call_8_9_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_14_15
#define init_8_9_14_15
void FASTCALL call_8_9_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_15_0
#define init_8_9_15_0
void FASTCALL call_8_9_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_15_1
#define init_8_9_15_1
void FASTCALL call_8_9_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_15_2
#define init_8_9_15_2
void FASTCALL call_8_9_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_15_3
#define init_8_9_15_3
void FASTCALL call_8_9_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_15_4
#define init_8_9_15_4
void FASTCALL call_8_9_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_15_5
#define init_8_9_15_5
void FASTCALL call_8_9_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_15_6
#define init_8_9_15_6
void FASTCALL call_8_9_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_15_7
#define init_8_9_15_7
void FASTCALL call_8_9_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_15_8
#define init_8_9_15_8
void FASTCALL call_8_9_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_15_9
#define init_8_9_15_9
void FASTCALL call_8_9_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_15_10
#define init_8_9_15_10
void FASTCALL call_8_9_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_15_11
#define init_8_9_15_11
void FASTCALL call_8_9_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_15_12
#define init_8_9_15_12
void FASTCALL call_8_9_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_15_13
#define init_8_9_15_13
void FASTCALL call_8_9_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_15_14
#define init_8_9_15_14
void FASTCALL call_8_9_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_9_15_15
#define init_8_9_15_15
void FASTCALL call_8_9_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_0_0
#define init_8_10_0_0
void FASTCALL call_8_10_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_0_1
#define init_8_10_0_1
void FASTCALL call_8_10_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_0_2
#define init_8_10_0_2
void FASTCALL call_8_10_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_0_3
#define init_8_10_0_3
void FASTCALL call_8_10_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_0_4
#define init_8_10_0_4
void FASTCALL call_8_10_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_0_5
#define init_8_10_0_5
void FASTCALL call_8_10_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_0_6
#define init_8_10_0_6
void FASTCALL call_8_10_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_0_7
#define init_8_10_0_7
void FASTCALL call_8_10_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_0_8
#define init_8_10_0_8
void FASTCALL call_8_10_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_0_9
#define init_8_10_0_9
void FASTCALL call_8_10_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_0_10
#define init_8_10_0_10
void FASTCALL call_8_10_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_0_11
#define init_8_10_0_11
void FASTCALL call_8_10_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_0_12
#define init_8_10_0_12
void FASTCALL call_8_10_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_0_13
#define init_8_10_0_13
void FASTCALL call_8_10_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_0_14
#define init_8_10_0_14
void FASTCALL call_8_10_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_0_15
#define init_8_10_0_15
void FASTCALL call_8_10_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_1_0
#define init_8_10_1_0
void FASTCALL call_8_10_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_1_1
#define init_8_10_1_1
void FASTCALL call_8_10_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_1_2
#define init_8_10_1_2
void FASTCALL call_8_10_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_1_3
#define init_8_10_1_3
void FASTCALL call_8_10_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_1_4
#define init_8_10_1_4
void FASTCALL call_8_10_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_1_5
#define init_8_10_1_5
void FASTCALL call_8_10_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_1_6
#define init_8_10_1_6
void FASTCALL call_8_10_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_1_7
#define init_8_10_1_7
void FASTCALL call_8_10_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_1_8
#define init_8_10_1_8
void FASTCALL call_8_10_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_1_9
#define init_8_10_1_9
void FASTCALL call_8_10_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_1_10
#define init_8_10_1_10
void FASTCALL call_8_10_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_1_11
#define init_8_10_1_11
void FASTCALL call_8_10_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_1_12
#define init_8_10_1_12
void FASTCALL call_8_10_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_1_13
#define init_8_10_1_13
void FASTCALL call_8_10_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_1_14
#define init_8_10_1_14
void FASTCALL call_8_10_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_1_15
#define init_8_10_1_15
void FASTCALL call_8_10_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_2_0
#define init_8_10_2_0
void FASTCALL call_8_10_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_2_1
#define init_8_10_2_1
void FASTCALL call_8_10_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_2_2
#define init_8_10_2_2
void FASTCALL call_8_10_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_2_3
#define init_8_10_2_3
void FASTCALL call_8_10_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_2_4
#define init_8_10_2_4
void FASTCALL call_8_10_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_2_5
#define init_8_10_2_5
void FASTCALL call_8_10_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_2_6
#define init_8_10_2_6
void FASTCALL call_8_10_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_2_7
#define init_8_10_2_7
void FASTCALL call_8_10_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_2_8
#define init_8_10_2_8
void FASTCALL call_8_10_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_2_9
#define init_8_10_2_9
void FASTCALL call_8_10_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_2_10
#define init_8_10_2_10
void FASTCALL call_8_10_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_2_11
#define init_8_10_2_11
void FASTCALL call_8_10_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_2_12
#define init_8_10_2_12
void FASTCALL call_8_10_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_2_13
#define init_8_10_2_13
void FASTCALL call_8_10_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_2_14
#define init_8_10_2_14
void FASTCALL call_8_10_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_2_15
#define init_8_10_2_15
void FASTCALL call_8_10_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_3_0
#define init_8_10_3_0
void FASTCALL call_8_10_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_3_1
#define init_8_10_3_1
void FASTCALL call_8_10_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_3_2
#define init_8_10_3_2
void FASTCALL call_8_10_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_3_3
#define init_8_10_3_3
void FASTCALL call_8_10_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_3_4
#define init_8_10_3_4
void FASTCALL call_8_10_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_3_5
#define init_8_10_3_5
void FASTCALL call_8_10_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_3_6
#define init_8_10_3_6
void FASTCALL call_8_10_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_3_7
#define init_8_10_3_7
void FASTCALL call_8_10_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_3_8
#define init_8_10_3_8
void FASTCALL call_8_10_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_3_9
#define init_8_10_3_9
void FASTCALL call_8_10_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_3_10
#define init_8_10_3_10
void FASTCALL call_8_10_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_3_11
#define init_8_10_3_11
void FASTCALL call_8_10_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_3_12
#define init_8_10_3_12
void FASTCALL call_8_10_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_3_13
#define init_8_10_3_13
void FASTCALL call_8_10_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_3_14
#define init_8_10_3_14
void FASTCALL call_8_10_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_3_15
#define init_8_10_3_15
void FASTCALL call_8_10_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_4_0
#define init_8_10_4_0
void FASTCALL call_8_10_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_4_1
#define init_8_10_4_1
void FASTCALL call_8_10_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_4_2
#define init_8_10_4_2
void FASTCALL call_8_10_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_4_3
#define init_8_10_4_3
void FASTCALL call_8_10_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_4_4
#define init_8_10_4_4
void FASTCALL call_8_10_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_4_5
#define init_8_10_4_5
void FASTCALL call_8_10_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_4_6
#define init_8_10_4_6
void FASTCALL call_8_10_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_4_7
#define init_8_10_4_7
void FASTCALL call_8_10_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_4_8
#define init_8_10_4_8
void FASTCALL call_8_10_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_4_9
#define init_8_10_4_9
void FASTCALL call_8_10_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_4_10
#define init_8_10_4_10
void FASTCALL call_8_10_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_4_11
#define init_8_10_4_11
void FASTCALL call_8_10_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_4_12
#define init_8_10_4_12
void FASTCALL call_8_10_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_4_13
#define init_8_10_4_13
void FASTCALL call_8_10_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_4_14
#define init_8_10_4_14
void FASTCALL call_8_10_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_4_15
#define init_8_10_4_15
void FASTCALL call_8_10_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_5_0
#define init_8_10_5_0
void FASTCALL call_8_10_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_5_1
#define init_8_10_5_1
void FASTCALL call_8_10_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_5_2
#define init_8_10_5_2
void FASTCALL call_8_10_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_5_3
#define init_8_10_5_3
void FASTCALL call_8_10_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_5_4
#define init_8_10_5_4
void FASTCALL call_8_10_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_5_5
#define init_8_10_5_5
void FASTCALL call_8_10_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_5_6
#define init_8_10_5_6
void FASTCALL call_8_10_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_5_7
#define init_8_10_5_7
void FASTCALL call_8_10_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_5_8
#define init_8_10_5_8
void FASTCALL call_8_10_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_5_9
#define init_8_10_5_9
void FASTCALL call_8_10_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_5_10
#define init_8_10_5_10
void FASTCALL call_8_10_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_5_11
#define init_8_10_5_11
void FASTCALL call_8_10_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_5_12
#define init_8_10_5_12
void FASTCALL call_8_10_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_5_13
#define init_8_10_5_13
void FASTCALL call_8_10_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_5_14
#define init_8_10_5_14
void FASTCALL call_8_10_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_5_15
#define init_8_10_5_15
void FASTCALL call_8_10_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_6_0
#define init_8_10_6_0
void FASTCALL call_8_10_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_6_1
#define init_8_10_6_1
void FASTCALL call_8_10_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_6_2
#define init_8_10_6_2
void FASTCALL call_8_10_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_6_3
#define init_8_10_6_3
void FASTCALL call_8_10_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_6_4
#define init_8_10_6_4
void FASTCALL call_8_10_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_6_5
#define init_8_10_6_5
void FASTCALL call_8_10_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_6_6
#define init_8_10_6_6
void FASTCALL call_8_10_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_6_7
#define init_8_10_6_7
void FASTCALL call_8_10_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_6_8
#define init_8_10_6_8
void FASTCALL call_8_10_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_6_9
#define init_8_10_6_9
void FASTCALL call_8_10_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_6_10
#define init_8_10_6_10
void FASTCALL call_8_10_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_6_11
#define init_8_10_6_11
void FASTCALL call_8_10_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_6_12
#define init_8_10_6_12
void FASTCALL call_8_10_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_6_13
#define init_8_10_6_13
void FASTCALL call_8_10_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_6_14
#define init_8_10_6_14
void FASTCALL call_8_10_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_6_15
#define init_8_10_6_15
void FASTCALL call_8_10_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_7_0
#define init_8_10_7_0
void FASTCALL call_8_10_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_7_1
#define init_8_10_7_1
void FASTCALL call_8_10_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_7_2
#define init_8_10_7_2
void FASTCALL call_8_10_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_7_3
#define init_8_10_7_3
void FASTCALL call_8_10_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_7_4
#define init_8_10_7_4
void FASTCALL call_8_10_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_7_5
#define init_8_10_7_5
void FASTCALL call_8_10_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_7_6
#define init_8_10_7_6
void FASTCALL call_8_10_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_7_7
#define init_8_10_7_7
void FASTCALL call_8_10_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_7_8
#define init_8_10_7_8
void FASTCALL call_8_10_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_7_9
#define init_8_10_7_9
void FASTCALL call_8_10_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_7_10
#define init_8_10_7_10
void FASTCALL call_8_10_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_7_11
#define init_8_10_7_11
void FASTCALL call_8_10_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_7_12
#define init_8_10_7_12
void FASTCALL call_8_10_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_7_13
#define init_8_10_7_13
void FASTCALL call_8_10_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_7_14
#define init_8_10_7_14
void FASTCALL call_8_10_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_7_15
#define init_8_10_7_15
void FASTCALL call_8_10_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_8_0
#define init_8_10_8_0
void FASTCALL call_8_10_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_8_1
#define init_8_10_8_1
void FASTCALL call_8_10_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_8_2
#define init_8_10_8_2
void FASTCALL call_8_10_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_8_3
#define init_8_10_8_3
void FASTCALL call_8_10_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_8_4
#define init_8_10_8_4
void FASTCALL call_8_10_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_8_5
#define init_8_10_8_5
void FASTCALL call_8_10_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_8_6
#define init_8_10_8_6
void FASTCALL call_8_10_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_8_7
#define init_8_10_8_7
void FASTCALL call_8_10_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_8_8
#define init_8_10_8_8
void FASTCALL call_8_10_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_8_9
#define init_8_10_8_9
void FASTCALL call_8_10_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_8_10
#define init_8_10_8_10
void FASTCALL call_8_10_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_8_11
#define init_8_10_8_11
void FASTCALL call_8_10_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_8_12
#define init_8_10_8_12
void FASTCALL call_8_10_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_8_13
#define init_8_10_8_13
void FASTCALL call_8_10_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_8_14
#define init_8_10_8_14
void FASTCALL call_8_10_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_8_15
#define init_8_10_8_15
void FASTCALL call_8_10_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_9_0
#define init_8_10_9_0
void FASTCALL call_8_10_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_9_1
#define init_8_10_9_1
void FASTCALL call_8_10_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_9_2
#define init_8_10_9_2
void FASTCALL call_8_10_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_9_3
#define init_8_10_9_3
void FASTCALL call_8_10_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_9_4
#define init_8_10_9_4
void FASTCALL call_8_10_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_9_5
#define init_8_10_9_5
void FASTCALL call_8_10_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_9_6
#define init_8_10_9_6
void FASTCALL call_8_10_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_9_7
#define init_8_10_9_7
void FASTCALL call_8_10_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_9_8
#define init_8_10_9_8
void FASTCALL call_8_10_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_9_9
#define init_8_10_9_9
void FASTCALL call_8_10_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_9_10
#define init_8_10_9_10
void FASTCALL call_8_10_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_9_11
#define init_8_10_9_11
void FASTCALL call_8_10_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_9_12
#define init_8_10_9_12
void FASTCALL call_8_10_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_9_13
#define init_8_10_9_13
void FASTCALL call_8_10_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_9_14
#define init_8_10_9_14
void FASTCALL call_8_10_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_9_15
#define init_8_10_9_15
void FASTCALL call_8_10_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_10_0
#define init_8_10_10_0
void FASTCALL call_8_10_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_10_1
#define init_8_10_10_1
void FASTCALL call_8_10_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_10_2
#define init_8_10_10_2
void FASTCALL call_8_10_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_10_3
#define init_8_10_10_3
void FASTCALL call_8_10_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_10_4
#define init_8_10_10_4
void FASTCALL call_8_10_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_10_5
#define init_8_10_10_5
void FASTCALL call_8_10_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_10_6
#define init_8_10_10_6
void FASTCALL call_8_10_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_10_7
#define init_8_10_10_7
void FASTCALL call_8_10_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_10_8
#define init_8_10_10_8
void FASTCALL call_8_10_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_10_9
#define init_8_10_10_9
void FASTCALL call_8_10_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_10_10
#define init_8_10_10_10
void FASTCALL call_8_10_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_10_11
#define init_8_10_10_11
void FASTCALL call_8_10_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_10_12
#define init_8_10_10_12
void FASTCALL call_8_10_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_10_13
#define init_8_10_10_13
void FASTCALL call_8_10_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_10_14
#define init_8_10_10_14
void FASTCALL call_8_10_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_10_15
#define init_8_10_10_15
void FASTCALL call_8_10_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_11_0
#define init_8_10_11_0
void FASTCALL call_8_10_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_11_1
#define init_8_10_11_1
void FASTCALL call_8_10_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_11_2
#define init_8_10_11_2
void FASTCALL call_8_10_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_11_3
#define init_8_10_11_3
void FASTCALL call_8_10_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_11_4
#define init_8_10_11_4
void FASTCALL call_8_10_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_11_5
#define init_8_10_11_5
void FASTCALL call_8_10_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_11_6
#define init_8_10_11_6
void FASTCALL call_8_10_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_11_7
#define init_8_10_11_7
void FASTCALL call_8_10_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_11_8
#define init_8_10_11_8
void FASTCALL call_8_10_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_11_9
#define init_8_10_11_9
void FASTCALL call_8_10_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_11_10
#define init_8_10_11_10
void FASTCALL call_8_10_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_11_11
#define init_8_10_11_11
void FASTCALL call_8_10_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_11_12
#define init_8_10_11_12
void FASTCALL call_8_10_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_11_13
#define init_8_10_11_13
void FASTCALL call_8_10_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_11_14
#define init_8_10_11_14
void FASTCALL call_8_10_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_11_15
#define init_8_10_11_15
void FASTCALL call_8_10_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_12_0
#define init_8_10_12_0
void FASTCALL call_8_10_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_12_1
#define init_8_10_12_1
void FASTCALL call_8_10_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_12_2
#define init_8_10_12_2
void FASTCALL call_8_10_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_12_3
#define init_8_10_12_3
void FASTCALL call_8_10_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_12_4
#define init_8_10_12_4
void FASTCALL call_8_10_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_12_5
#define init_8_10_12_5
void FASTCALL call_8_10_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_12_6
#define init_8_10_12_6
void FASTCALL call_8_10_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_12_7
#define init_8_10_12_7
void FASTCALL call_8_10_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_12_8
#define init_8_10_12_8
void FASTCALL call_8_10_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_12_9
#define init_8_10_12_9
void FASTCALL call_8_10_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_12_10
#define init_8_10_12_10
void FASTCALL call_8_10_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_12_11
#define init_8_10_12_11
void FASTCALL call_8_10_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_12_12
#define init_8_10_12_12
void FASTCALL call_8_10_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_12_13
#define init_8_10_12_13
void FASTCALL call_8_10_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_12_14
#define init_8_10_12_14
void FASTCALL call_8_10_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_12_15
#define init_8_10_12_15
void FASTCALL call_8_10_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_13_0
#define init_8_10_13_0
void FASTCALL call_8_10_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_13_1
#define init_8_10_13_1
void FASTCALL call_8_10_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_13_2
#define init_8_10_13_2
void FASTCALL call_8_10_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_13_3
#define init_8_10_13_3
void FASTCALL call_8_10_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_13_4
#define init_8_10_13_4
void FASTCALL call_8_10_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_13_5
#define init_8_10_13_5
void FASTCALL call_8_10_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_13_6
#define init_8_10_13_6
void FASTCALL call_8_10_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_13_7
#define init_8_10_13_7
void FASTCALL call_8_10_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_13_8
#define init_8_10_13_8
void FASTCALL call_8_10_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_13_9
#define init_8_10_13_9
void FASTCALL call_8_10_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_13_10
#define init_8_10_13_10
void FASTCALL call_8_10_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_13_11
#define init_8_10_13_11
void FASTCALL call_8_10_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_13_12
#define init_8_10_13_12
void FASTCALL call_8_10_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_13_13
#define init_8_10_13_13
void FASTCALL call_8_10_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_13_14
#define init_8_10_13_14
void FASTCALL call_8_10_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_13_15
#define init_8_10_13_15
void FASTCALL call_8_10_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_14_0
#define init_8_10_14_0
void FASTCALL call_8_10_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_14_1
#define init_8_10_14_1
void FASTCALL call_8_10_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_14_2
#define init_8_10_14_2
void FASTCALL call_8_10_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_14_3
#define init_8_10_14_3
void FASTCALL call_8_10_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_14_4
#define init_8_10_14_4
void FASTCALL call_8_10_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_14_5
#define init_8_10_14_5
void FASTCALL call_8_10_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_14_6
#define init_8_10_14_6
void FASTCALL call_8_10_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_14_7
#define init_8_10_14_7
void FASTCALL call_8_10_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_14_8
#define init_8_10_14_8
void FASTCALL call_8_10_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_14_9
#define init_8_10_14_9
void FASTCALL call_8_10_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_14_10
#define init_8_10_14_10
void FASTCALL call_8_10_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_14_11
#define init_8_10_14_11
void FASTCALL call_8_10_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_14_12
#define init_8_10_14_12
void FASTCALL call_8_10_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_14_13
#define init_8_10_14_13
void FASTCALL call_8_10_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_14_14
#define init_8_10_14_14
void FASTCALL call_8_10_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_14_15
#define init_8_10_14_15
void FASTCALL call_8_10_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_15_0
#define init_8_10_15_0
void FASTCALL call_8_10_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_15_1
#define init_8_10_15_1
void FASTCALL call_8_10_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_15_2
#define init_8_10_15_2
void FASTCALL call_8_10_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_15_3
#define init_8_10_15_3
void FASTCALL call_8_10_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_15_4
#define init_8_10_15_4
void FASTCALL call_8_10_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_15_5
#define init_8_10_15_5
void FASTCALL call_8_10_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_15_6
#define init_8_10_15_6
void FASTCALL call_8_10_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_15_7
#define init_8_10_15_7
void FASTCALL call_8_10_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_15_8
#define init_8_10_15_8
void FASTCALL call_8_10_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_15_9
#define init_8_10_15_9
void FASTCALL call_8_10_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_15_10
#define init_8_10_15_10
void FASTCALL call_8_10_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_15_11
#define init_8_10_15_11
void FASTCALL call_8_10_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_15_12
#define init_8_10_15_12
void FASTCALL call_8_10_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_15_13
#define init_8_10_15_13
void FASTCALL call_8_10_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_15_14
#define init_8_10_15_14
void FASTCALL call_8_10_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_10_15_15
#define init_8_10_15_15
void FASTCALL call_8_10_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_0_0
#define init_8_11_0_0
void FASTCALL call_8_11_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_0_1
#define init_8_11_0_1
void FASTCALL call_8_11_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_0_2
#define init_8_11_0_2
void FASTCALL call_8_11_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_0_3
#define init_8_11_0_3
void FASTCALL call_8_11_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_0_4
#define init_8_11_0_4
void FASTCALL call_8_11_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_0_5
#define init_8_11_0_5
void FASTCALL call_8_11_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_0_6
#define init_8_11_0_6
void FASTCALL call_8_11_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_0_7
#define init_8_11_0_7
void FASTCALL call_8_11_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_0_8
#define init_8_11_0_8
void FASTCALL call_8_11_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_0_9
#define init_8_11_0_9
void FASTCALL call_8_11_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_0_10
#define init_8_11_0_10
void FASTCALL call_8_11_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_0_11
#define init_8_11_0_11
void FASTCALL call_8_11_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_0_12
#define init_8_11_0_12
void FASTCALL call_8_11_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_0_13
#define init_8_11_0_13
void FASTCALL call_8_11_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_0_14
#define init_8_11_0_14
void FASTCALL call_8_11_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_0_15
#define init_8_11_0_15
void FASTCALL call_8_11_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_1_0
#define init_8_11_1_0
void FASTCALL call_8_11_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_1_1
#define init_8_11_1_1
void FASTCALL call_8_11_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_1_2
#define init_8_11_1_2
void FASTCALL call_8_11_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_1_3
#define init_8_11_1_3
void FASTCALL call_8_11_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_1_4
#define init_8_11_1_4
void FASTCALL call_8_11_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_1_5
#define init_8_11_1_5
void FASTCALL call_8_11_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_1_6
#define init_8_11_1_6
void FASTCALL call_8_11_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_1_7
#define init_8_11_1_7
void FASTCALL call_8_11_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_1_8
#define init_8_11_1_8
void FASTCALL call_8_11_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_1_9
#define init_8_11_1_9
void FASTCALL call_8_11_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_1_10
#define init_8_11_1_10
void FASTCALL call_8_11_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_1_11
#define init_8_11_1_11
void FASTCALL call_8_11_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_1_12
#define init_8_11_1_12
void FASTCALL call_8_11_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_1_13
#define init_8_11_1_13
void FASTCALL call_8_11_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_1_14
#define init_8_11_1_14
void FASTCALL call_8_11_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_1_15
#define init_8_11_1_15
void FASTCALL call_8_11_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_2_0
#define init_8_11_2_0
void FASTCALL call_8_11_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_2_1
#define init_8_11_2_1
void FASTCALL call_8_11_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_2_2
#define init_8_11_2_2
void FASTCALL call_8_11_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_2_3
#define init_8_11_2_3
void FASTCALL call_8_11_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_2_4
#define init_8_11_2_4
void FASTCALL call_8_11_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_2_5
#define init_8_11_2_5
void FASTCALL call_8_11_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_2_6
#define init_8_11_2_6
void FASTCALL call_8_11_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_2_7
#define init_8_11_2_7
void FASTCALL call_8_11_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_2_8
#define init_8_11_2_8
void FASTCALL call_8_11_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_2_9
#define init_8_11_2_9
void FASTCALL call_8_11_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_2_10
#define init_8_11_2_10
void FASTCALL call_8_11_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_2_11
#define init_8_11_2_11
void FASTCALL call_8_11_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_2_12
#define init_8_11_2_12
void FASTCALL call_8_11_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_2_13
#define init_8_11_2_13
void FASTCALL call_8_11_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_2_14
#define init_8_11_2_14
void FASTCALL call_8_11_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_2_15
#define init_8_11_2_15
void FASTCALL call_8_11_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_3_0
#define init_8_11_3_0
void FASTCALL call_8_11_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_3_1
#define init_8_11_3_1
void FASTCALL call_8_11_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_3_2
#define init_8_11_3_2
void FASTCALL call_8_11_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_3_3
#define init_8_11_3_3
void FASTCALL call_8_11_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_3_4
#define init_8_11_3_4
void FASTCALL call_8_11_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_3_5
#define init_8_11_3_5
void FASTCALL call_8_11_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_3_6
#define init_8_11_3_6
void FASTCALL call_8_11_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_3_7
#define init_8_11_3_7
void FASTCALL call_8_11_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_3_8
#define init_8_11_3_8
void FASTCALL call_8_11_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_3_9
#define init_8_11_3_9
void FASTCALL call_8_11_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_3_10
#define init_8_11_3_10
void FASTCALL call_8_11_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_3_11
#define init_8_11_3_11
void FASTCALL call_8_11_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_3_12
#define init_8_11_3_12
void FASTCALL call_8_11_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_3_13
#define init_8_11_3_13
void FASTCALL call_8_11_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_3_14
#define init_8_11_3_14
void FASTCALL call_8_11_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_3_15
#define init_8_11_3_15
void FASTCALL call_8_11_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_4_0
#define init_8_11_4_0
void FASTCALL call_8_11_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_4_1
#define init_8_11_4_1
void FASTCALL call_8_11_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_4_2
#define init_8_11_4_2
void FASTCALL call_8_11_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_4_3
#define init_8_11_4_3
void FASTCALL call_8_11_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_4_4
#define init_8_11_4_4
void FASTCALL call_8_11_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_4_5
#define init_8_11_4_5
void FASTCALL call_8_11_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_4_6
#define init_8_11_4_6
void FASTCALL call_8_11_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_4_7
#define init_8_11_4_7
void FASTCALL call_8_11_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_4_8
#define init_8_11_4_8
void FASTCALL call_8_11_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_4_9
#define init_8_11_4_9
void FASTCALL call_8_11_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_4_10
#define init_8_11_4_10
void FASTCALL call_8_11_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_4_11
#define init_8_11_4_11
void FASTCALL call_8_11_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_4_12
#define init_8_11_4_12
void FASTCALL call_8_11_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_4_13
#define init_8_11_4_13
void FASTCALL call_8_11_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_4_14
#define init_8_11_4_14
void FASTCALL call_8_11_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_4_15
#define init_8_11_4_15
void FASTCALL call_8_11_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_5_0
#define init_8_11_5_0
void FASTCALL call_8_11_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_5_1
#define init_8_11_5_1
void FASTCALL call_8_11_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_5_2
#define init_8_11_5_2
void FASTCALL call_8_11_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_5_3
#define init_8_11_5_3
void FASTCALL call_8_11_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_5_4
#define init_8_11_5_4
void FASTCALL call_8_11_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_5_5
#define init_8_11_5_5
void FASTCALL call_8_11_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_5_6
#define init_8_11_5_6
void FASTCALL call_8_11_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_5_7
#define init_8_11_5_7
void FASTCALL call_8_11_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_5_8
#define init_8_11_5_8
void FASTCALL call_8_11_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_5_9
#define init_8_11_5_9
void FASTCALL call_8_11_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_5_10
#define init_8_11_5_10
void FASTCALL call_8_11_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_5_11
#define init_8_11_5_11
void FASTCALL call_8_11_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_5_12
#define init_8_11_5_12
void FASTCALL call_8_11_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_5_13
#define init_8_11_5_13
void FASTCALL call_8_11_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_5_14
#define init_8_11_5_14
void FASTCALL call_8_11_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_5_15
#define init_8_11_5_15
void FASTCALL call_8_11_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_6_0
#define init_8_11_6_0
void FASTCALL call_8_11_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_6_1
#define init_8_11_6_1
void FASTCALL call_8_11_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_6_2
#define init_8_11_6_2
void FASTCALL call_8_11_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_6_3
#define init_8_11_6_3
void FASTCALL call_8_11_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_6_4
#define init_8_11_6_4
void FASTCALL call_8_11_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_6_5
#define init_8_11_6_5
void FASTCALL call_8_11_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_6_6
#define init_8_11_6_6
void FASTCALL call_8_11_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_6_7
#define init_8_11_6_7
void FASTCALL call_8_11_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_6_8
#define init_8_11_6_8
void FASTCALL call_8_11_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_6_9
#define init_8_11_6_9
void FASTCALL call_8_11_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_6_10
#define init_8_11_6_10
void FASTCALL call_8_11_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_6_11
#define init_8_11_6_11
void FASTCALL call_8_11_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_6_12
#define init_8_11_6_12
void FASTCALL call_8_11_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_6_13
#define init_8_11_6_13
void FASTCALL call_8_11_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_6_14
#define init_8_11_6_14
void FASTCALL call_8_11_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_6_15
#define init_8_11_6_15
void FASTCALL call_8_11_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_7_0
#define init_8_11_7_0
void FASTCALL call_8_11_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_7_1
#define init_8_11_7_1
void FASTCALL call_8_11_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_7_2
#define init_8_11_7_2
void FASTCALL call_8_11_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_7_3
#define init_8_11_7_3
void FASTCALL call_8_11_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_7_4
#define init_8_11_7_4
void FASTCALL call_8_11_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_7_5
#define init_8_11_7_5
void FASTCALL call_8_11_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_7_6
#define init_8_11_7_6
void FASTCALL call_8_11_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_7_7
#define init_8_11_7_7
void FASTCALL call_8_11_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_7_8
#define init_8_11_7_8
void FASTCALL call_8_11_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_7_9
#define init_8_11_7_9
void FASTCALL call_8_11_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_7_10
#define init_8_11_7_10
void FASTCALL call_8_11_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_7_11
#define init_8_11_7_11
void FASTCALL call_8_11_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_7_12
#define init_8_11_7_12
void FASTCALL call_8_11_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_7_13
#define init_8_11_7_13
void FASTCALL call_8_11_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_7_14
#define init_8_11_7_14
void FASTCALL call_8_11_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_7_15
#define init_8_11_7_15
void FASTCALL call_8_11_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_8_0
#define init_8_11_8_0
void FASTCALL call_8_11_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_8_1
#define init_8_11_8_1
void FASTCALL call_8_11_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_8_2
#define init_8_11_8_2
void FASTCALL call_8_11_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_8_3
#define init_8_11_8_3
void FASTCALL call_8_11_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_8_4
#define init_8_11_8_4
void FASTCALL call_8_11_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_8_5
#define init_8_11_8_5
void FASTCALL call_8_11_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_8_6
#define init_8_11_8_6
void FASTCALL call_8_11_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_8_7
#define init_8_11_8_7
void FASTCALL call_8_11_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_8_8
#define init_8_11_8_8
void FASTCALL call_8_11_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_8_9
#define init_8_11_8_9
void FASTCALL call_8_11_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_8_10
#define init_8_11_8_10
void FASTCALL call_8_11_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_8_11
#define init_8_11_8_11
void FASTCALL call_8_11_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_8_12
#define init_8_11_8_12
void FASTCALL call_8_11_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_8_13
#define init_8_11_8_13
void FASTCALL call_8_11_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_8_14
#define init_8_11_8_14
void FASTCALL call_8_11_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_8_15
#define init_8_11_8_15
void FASTCALL call_8_11_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_9_0
#define init_8_11_9_0
void FASTCALL call_8_11_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_9_1
#define init_8_11_9_1
void FASTCALL call_8_11_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_9_2
#define init_8_11_9_2
void FASTCALL call_8_11_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_9_3
#define init_8_11_9_3
void FASTCALL call_8_11_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_9_4
#define init_8_11_9_4
void FASTCALL call_8_11_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_9_5
#define init_8_11_9_5
void FASTCALL call_8_11_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_9_6
#define init_8_11_9_6
void FASTCALL call_8_11_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_9_7
#define init_8_11_9_7
void FASTCALL call_8_11_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_9_8
#define init_8_11_9_8
void FASTCALL call_8_11_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_9_9
#define init_8_11_9_9
void FASTCALL call_8_11_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_9_10
#define init_8_11_9_10
void FASTCALL call_8_11_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_9_11
#define init_8_11_9_11
void FASTCALL call_8_11_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_9_12
#define init_8_11_9_12
void FASTCALL call_8_11_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_9_13
#define init_8_11_9_13
void FASTCALL call_8_11_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_9_14
#define init_8_11_9_14
void FASTCALL call_8_11_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_9_15
#define init_8_11_9_15
void FASTCALL call_8_11_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_10_0
#define init_8_11_10_0
void FASTCALL call_8_11_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_10_1
#define init_8_11_10_1
void FASTCALL call_8_11_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_10_2
#define init_8_11_10_2
void FASTCALL call_8_11_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_10_3
#define init_8_11_10_3
void FASTCALL call_8_11_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_10_4
#define init_8_11_10_4
void FASTCALL call_8_11_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_10_5
#define init_8_11_10_5
void FASTCALL call_8_11_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_10_6
#define init_8_11_10_6
void FASTCALL call_8_11_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_10_7
#define init_8_11_10_7
void FASTCALL call_8_11_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_10_8
#define init_8_11_10_8
void FASTCALL call_8_11_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_10_9
#define init_8_11_10_9
void FASTCALL call_8_11_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_10_10
#define init_8_11_10_10
void FASTCALL call_8_11_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_10_11
#define init_8_11_10_11
void FASTCALL call_8_11_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_10_12
#define init_8_11_10_12
void FASTCALL call_8_11_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_10_13
#define init_8_11_10_13
void FASTCALL call_8_11_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_10_14
#define init_8_11_10_14
void FASTCALL call_8_11_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_10_15
#define init_8_11_10_15
void FASTCALL call_8_11_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_11_0
#define init_8_11_11_0
void FASTCALL call_8_11_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_11_1
#define init_8_11_11_1
void FASTCALL call_8_11_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_11_2
#define init_8_11_11_2
void FASTCALL call_8_11_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_11_3
#define init_8_11_11_3
void FASTCALL call_8_11_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_11_4
#define init_8_11_11_4
void FASTCALL call_8_11_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_11_5
#define init_8_11_11_5
void FASTCALL call_8_11_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_11_6
#define init_8_11_11_6
void FASTCALL call_8_11_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_11_7
#define init_8_11_11_7
void FASTCALL call_8_11_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_11_8
#define init_8_11_11_8
void FASTCALL call_8_11_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_11_9
#define init_8_11_11_9
void FASTCALL call_8_11_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_11_10
#define init_8_11_11_10
void FASTCALL call_8_11_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_11_11
#define init_8_11_11_11
void FASTCALL call_8_11_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_11_12
#define init_8_11_11_12
void FASTCALL call_8_11_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_11_13
#define init_8_11_11_13
void FASTCALL call_8_11_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_11_14
#define init_8_11_11_14
void FASTCALL call_8_11_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_11_15
#define init_8_11_11_15
void FASTCALL call_8_11_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_12_0
#define init_8_11_12_0
void FASTCALL call_8_11_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_12_1
#define init_8_11_12_1
void FASTCALL call_8_11_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_12_2
#define init_8_11_12_2
void FASTCALL call_8_11_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_12_3
#define init_8_11_12_3
void FASTCALL call_8_11_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_12_4
#define init_8_11_12_4
void FASTCALL call_8_11_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_12_5
#define init_8_11_12_5
void FASTCALL call_8_11_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_12_6
#define init_8_11_12_6
void FASTCALL call_8_11_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_12_7
#define init_8_11_12_7
void FASTCALL call_8_11_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_12_8
#define init_8_11_12_8
void FASTCALL call_8_11_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_12_9
#define init_8_11_12_9
void FASTCALL call_8_11_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_12_10
#define init_8_11_12_10
void FASTCALL call_8_11_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_12_11
#define init_8_11_12_11
void FASTCALL call_8_11_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_12_12
#define init_8_11_12_12
void FASTCALL call_8_11_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_12_13
#define init_8_11_12_13
void FASTCALL call_8_11_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_12_14
#define init_8_11_12_14
void FASTCALL call_8_11_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_12_15
#define init_8_11_12_15
void FASTCALL call_8_11_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_13_0
#define init_8_11_13_0
void FASTCALL call_8_11_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_13_1
#define init_8_11_13_1
void FASTCALL call_8_11_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_13_2
#define init_8_11_13_2
void FASTCALL call_8_11_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_13_3
#define init_8_11_13_3
void FASTCALL call_8_11_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_13_4
#define init_8_11_13_4
void FASTCALL call_8_11_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_13_5
#define init_8_11_13_5
void FASTCALL call_8_11_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_13_6
#define init_8_11_13_6
void FASTCALL call_8_11_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_13_7
#define init_8_11_13_7
void FASTCALL call_8_11_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_13_8
#define init_8_11_13_8
void FASTCALL call_8_11_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_13_9
#define init_8_11_13_9
void FASTCALL call_8_11_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_13_10
#define init_8_11_13_10
void FASTCALL call_8_11_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_13_11
#define init_8_11_13_11
void FASTCALL call_8_11_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_13_12
#define init_8_11_13_12
void FASTCALL call_8_11_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_13_13
#define init_8_11_13_13
void FASTCALL call_8_11_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_13_14
#define init_8_11_13_14
void FASTCALL call_8_11_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_13_15
#define init_8_11_13_15
void FASTCALL call_8_11_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_14_0
#define init_8_11_14_0
void FASTCALL call_8_11_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_14_1
#define init_8_11_14_1
void FASTCALL call_8_11_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_14_2
#define init_8_11_14_2
void FASTCALL call_8_11_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_14_3
#define init_8_11_14_3
void FASTCALL call_8_11_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_14_4
#define init_8_11_14_4
void FASTCALL call_8_11_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_14_5
#define init_8_11_14_5
void FASTCALL call_8_11_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_14_6
#define init_8_11_14_6
void FASTCALL call_8_11_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_14_7
#define init_8_11_14_7
void FASTCALL call_8_11_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_14_8
#define init_8_11_14_8
void FASTCALL call_8_11_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_14_9
#define init_8_11_14_9
void FASTCALL call_8_11_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_14_10
#define init_8_11_14_10
void FASTCALL call_8_11_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_14_11
#define init_8_11_14_11
void FASTCALL call_8_11_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_14_12
#define init_8_11_14_12
void FASTCALL call_8_11_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_14_13
#define init_8_11_14_13
void FASTCALL call_8_11_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_14_14
#define init_8_11_14_14
void FASTCALL call_8_11_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_14_15
#define init_8_11_14_15
void FASTCALL call_8_11_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_15_0
#define init_8_11_15_0
void FASTCALL call_8_11_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_15_1
#define init_8_11_15_1
void FASTCALL call_8_11_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_15_2
#define init_8_11_15_2
void FASTCALL call_8_11_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_15_3
#define init_8_11_15_3
void FASTCALL call_8_11_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_15_4
#define init_8_11_15_4
void FASTCALL call_8_11_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_15_5
#define init_8_11_15_5
void FASTCALL call_8_11_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_15_6
#define init_8_11_15_6
void FASTCALL call_8_11_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_15_7
#define init_8_11_15_7
void FASTCALL call_8_11_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_15_8
#define init_8_11_15_8
void FASTCALL call_8_11_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_15_9
#define init_8_11_15_9
void FASTCALL call_8_11_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_15_10
#define init_8_11_15_10
void FASTCALL call_8_11_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_15_11
#define init_8_11_15_11
void FASTCALL call_8_11_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_15_12
#define init_8_11_15_12
void FASTCALL call_8_11_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_15_13
#define init_8_11_15_13
void FASTCALL call_8_11_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_15_14
#define init_8_11_15_14
void FASTCALL call_8_11_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_11_15_15
#define init_8_11_15_15
void FASTCALL call_8_11_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_0_0
#define init_8_12_0_0
void FASTCALL call_8_12_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_0_1
#define init_8_12_0_1
void FASTCALL call_8_12_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_0_2
#define init_8_12_0_2
void FASTCALL call_8_12_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_0_3
#define init_8_12_0_3
void FASTCALL call_8_12_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_0_4
#define init_8_12_0_4
void FASTCALL call_8_12_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_0_5
#define init_8_12_0_5
void FASTCALL call_8_12_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_0_6
#define init_8_12_0_6
void FASTCALL call_8_12_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_0_7
#define init_8_12_0_7
void FASTCALL call_8_12_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_0_8
#define init_8_12_0_8
void FASTCALL call_8_12_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_0_9
#define init_8_12_0_9
void FASTCALL call_8_12_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_0_10
#define init_8_12_0_10
void FASTCALL call_8_12_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_0_11
#define init_8_12_0_11
void FASTCALL call_8_12_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_0_12
#define init_8_12_0_12
void FASTCALL call_8_12_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_0_13
#define init_8_12_0_13
void FASTCALL call_8_12_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_0_14
#define init_8_12_0_14
void FASTCALL call_8_12_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_0_15
#define init_8_12_0_15
void FASTCALL call_8_12_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_1_0
#define init_8_12_1_0
void FASTCALL call_8_12_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_1_1
#define init_8_12_1_1
void FASTCALL call_8_12_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_1_2
#define init_8_12_1_2
void FASTCALL call_8_12_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_1_3
#define init_8_12_1_3
void FASTCALL call_8_12_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_1_4
#define init_8_12_1_4
void FASTCALL call_8_12_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_1_5
#define init_8_12_1_5
void FASTCALL call_8_12_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_1_6
#define init_8_12_1_6
void FASTCALL call_8_12_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_1_7
#define init_8_12_1_7
void FASTCALL call_8_12_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_1_8
#define init_8_12_1_8
void FASTCALL call_8_12_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_1_9
#define init_8_12_1_9
void FASTCALL call_8_12_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_1_10
#define init_8_12_1_10
void FASTCALL call_8_12_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_1_11
#define init_8_12_1_11
void FASTCALL call_8_12_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_1_12
#define init_8_12_1_12
void FASTCALL call_8_12_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_1_13
#define init_8_12_1_13
void FASTCALL call_8_12_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_1_14
#define init_8_12_1_14
void FASTCALL call_8_12_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_1_15
#define init_8_12_1_15
void FASTCALL call_8_12_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_2_0
#define init_8_12_2_0
void FASTCALL call_8_12_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_2_1
#define init_8_12_2_1
void FASTCALL call_8_12_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_2_2
#define init_8_12_2_2
void FASTCALL call_8_12_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_2_3
#define init_8_12_2_3
void FASTCALL call_8_12_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_2_4
#define init_8_12_2_4
void FASTCALL call_8_12_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_2_5
#define init_8_12_2_5
void FASTCALL call_8_12_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_2_6
#define init_8_12_2_6
void FASTCALL call_8_12_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_2_7
#define init_8_12_2_7
void FASTCALL call_8_12_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_2_8
#define init_8_12_2_8
void FASTCALL call_8_12_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_2_9
#define init_8_12_2_9
void FASTCALL call_8_12_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_2_10
#define init_8_12_2_10
void FASTCALL call_8_12_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_2_11
#define init_8_12_2_11
void FASTCALL call_8_12_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_2_12
#define init_8_12_2_12
void FASTCALL call_8_12_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_2_13
#define init_8_12_2_13
void FASTCALL call_8_12_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_2_14
#define init_8_12_2_14
void FASTCALL call_8_12_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_2_15
#define init_8_12_2_15
void FASTCALL call_8_12_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_3_0
#define init_8_12_3_0
void FASTCALL call_8_12_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_3_1
#define init_8_12_3_1
void FASTCALL call_8_12_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_3_2
#define init_8_12_3_2
void FASTCALL call_8_12_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_3_3
#define init_8_12_3_3
void FASTCALL call_8_12_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_3_4
#define init_8_12_3_4
void FASTCALL call_8_12_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_3_5
#define init_8_12_3_5
void FASTCALL call_8_12_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_3_6
#define init_8_12_3_6
void FASTCALL call_8_12_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_3_7
#define init_8_12_3_7
void FASTCALL call_8_12_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_3_8
#define init_8_12_3_8
void FASTCALL call_8_12_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_3_9
#define init_8_12_3_9
void FASTCALL call_8_12_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_3_10
#define init_8_12_3_10
void FASTCALL call_8_12_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_3_11
#define init_8_12_3_11
void FASTCALL call_8_12_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_3_12
#define init_8_12_3_12
void FASTCALL call_8_12_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_3_13
#define init_8_12_3_13
void FASTCALL call_8_12_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_3_14
#define init_8_12_3_14
void FASTCALL call_8_12_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_3_15
#define init_8_12_3_15
void FASTCALL call_8_12_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_4_0
#define init_8_12_4_0
void FASTCALL call_8_12_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_4_1
#define init_8_12_4_1
void FASTCALL call_8_12_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_4_2
#define init_8_12_4_2
void FASTCALL call_8_12_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_4_3
#define init_8_12_4_3
void FASTCALL call_8_12_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_4_4
#define init_8_12_4_4
void FASTCALL call_8_12_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_4_5
#define init_8_12_4_5
void FASTCALL call_8_12_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_4_6
#define init_8_12_4_6
void FASTCALL call_8_12_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_4_7
#define init_8_12_4_7
void FASTCALL call_8_12_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_4_8
#define init_8_12_4_8
void FASTCALL call_8_12_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_4_9
#define init_8_12_4_9
void FASTCALL call_8_12_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_4_10
#define init_8_12_4_10
void FASTCALL call_8_12_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_4_11
#define init_8_12_4_11
void FASTCALL call_8_12_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_4_12
#define init_8_12_4_12
void FASTCALL call_8_12_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_4_13
#define init_8_12_4_13
void FASTCALL call_8_12_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_4_14
#define init_8_12_4_14
void FASTCALL call_8_12_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_4_15
#define init_8_12_4_15
void FASTCALL call_8_12_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_5_0
#define init_8_12_5_0
void FASTCALL call_8_12_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_5_1
#define init_8_12_5_1
void FASTCALL call_8_12_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_5_2
#define init_8_12_5_2
void FASTCALL call_8_12_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_5_3
#define init_8_12_5_3
void FASTCALL call_8_12_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_5_4
#define init_8_12_5_4
void FASTCALL call_8_12_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_5_5
#define init_8_12_5_5
void FASTCALL call_8_12_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_5_6
#define init_8_12_5_6
void FASTCALL call_8_12_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_5_7
#define init_8_12_5_7
void FASTCALL call_8_12_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_5_8
#define init_8_12_5_8
void FASTCALL call_8_12_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_5_9
#define init_8_12_5_9
void FASTCALL call_8_12_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_5_10
#define init_8_12_5_10
void FASTCALL call_8_12_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_5_11
#define init_8_12_5_11
void FASTCALL call_8_12_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_5_12
#define init_8_12_5_12
void FASTCALL call_8_12_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_5_13
#define init_8_12_5_13
void FASTCALL call_8_12_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_5_14
#define init_8_12_5_14
void FASTCALL call_8_12_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_5_15
#define init_8_12_5_15
void FASTCALL call_8_12_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_6_0
#define init_8_12_6_0
void FASTCALL call_8_12_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_6_1
#define init_8_12_6_1
void FASTCALL call_8_12_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_6_2
#define init_8_12_6_2
void FASTCALL call_8_12_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_6_3
#define init_8_12_6_3
void FASTCALL call_8_12_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_6_4
#define init_8_12_6_4
void FASTCALL call_8_12_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_6_5
#define init_8_12_6_5
void FASTCALL call_8_12_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_6_6
#define init_8_12_6_6
void FASTCALL call_8_12_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_6_7
#define init_8_12_6_7
void FASTCALL call_8_12_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_6_8
#define init_8_12_6_8
void FASTCALL call_8_12_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_6_9
#define init_8_12_6_9
void FASTCALL call_8_12_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_6_10
#define init_8_12_6_10
void FASTCALL call_8_12_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_6_11
#define init_8_12_6_11
void FASTCALL call_8_12_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_6_12
#define init_8_12_6_12
void FASTCALL call_8_12_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_6_13
#define init_8_12_6_13
void FASTCALL call_8_12_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_6_14
#define init_8_12_6_14
void FASTCALL call_8_12_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_6_15
#define init_8_12_6_15
void FASTCALL call_8_12_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_7_0
#define init_8_12_7_0
void FASTCALL call_8_12_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_7_1
#define init_8_12_7_1
void FASTCALL call_8_12_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_7_2
#define init_8_12_7_2
void FASTCALL call_8_12_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_7_3
#define init_8_12_7_3
void FASTCALL call_8_12_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_7_4
#define init_8_12_7_4
void FASTCALL call_8_12_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_7_5
#define init_8_12_7_5
void FASTCALL call_8_12_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_7_6
#define init_8_12_7_6
void FASTCALL call_8_12_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_7_7
#define init_8_12_7_7
void FASTCALL call_8_12_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_7_8
#define init_8_12_7_8
void FASTCALL call_8_12_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_7_9
#define init_8_12_7_9
void FASTCALL call_8_12_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_7_10
#define init_8_12_7_10
void FASTCALL call_8_12_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_7_11
#define init_8_12_7_11
void FASTCALL call_8_12_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_7_12
#define init_8_12_7_12
void FASTCALL call_8_12_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_7_13
#define init_8_12_7_13
void FASTCALL call_8_12_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_7_14
#define init_8_12_7_14
void FASTCALL call_8_12_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_7_15
#define init_8_12_7_15
void FASTCALL call_8_12_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_8_0
#define init_8_12_8_0
void FASTCALL call_8_12_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_8_1
#define init_8_12_8_1
void FASTCALL call_8_12_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_8_2
#define init_8_12_8_2
void FASTCALL call_8_12_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_8_3
#define init_8_12_8_3
void FASTCALL call_8_12_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_8_4
#define init_8_12_8_4
void FASTCALL call_8_12_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_8_5
#define init_8_12_8_5
void FASTCALL call_8_12_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_8_6
#define init_8_12_8_6
void FASTCALL call_8_12_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_8_7
#define init_8_12_8_7
void FASTCALL call_8_12_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_8_8
#define init_8_12_8_8
void FASTCALL call_8_12_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_8_9
#define init_8_12_8_9
void FASTCALL call_8_12_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_8_10
#define init_8_12_8_10
void FASTCALL call_8_12_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_8_11
#define init_8_12_8_11
void FASTCALL call_8_12_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_8_12
#define init_8_12_8_12
void FASTCALL call_8_12_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_8_13
#define init_8_12_8_13
void FASTCALL call_8_12_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_8_14
#define init_8_12_8_14
void FASTCALL call_8_12_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_8_15
#define init_8_12_8_15
void FASTCALL call_8_12_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_9_0
#define init_8_12_9_0
void FASTCALL call_8_12_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_9_1
#define init_8_12_9_1
void FASTCALL call_8_12_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_9_2
#define init_8_12_9_2
void FASTCALL call_8_12_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_9_3
#define init_8_12_9_3
void FASTCALL call_8_12_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_9_4
#define init_8_12_9_4
void FASTCALL call_8_12_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_9_5
#define init_8_12_9_5
void FASTCALL call_8_12_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_9_6
#define init_8_12_9_6
void FASTCALL call_8_12_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_9_7
#define init_8_12_9_7
void FASTCALL call_8_12_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_9_8
#define init_8_12_9_8
void FASTCALL call_8_12_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_9_9
#define init_8_12_9_9
void FASTCALL call_8_12_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_9_10
#define init_8_12_9_10
void FASTCALL call_8_12_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_9_11
#define init_8_12_9_11
void FASTCALL call_8_12_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_9_12
#define init_8_12_9_12
void FASTCALL call_8_12_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_9_13
#define init_8_12_9_13
void FASTCALL call_8_12_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_9_14
#define init_8_12_9_14
void FASTCALL call_8_12_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_9_15
#define init_8_12_9_15
void FASTCALL call_8_12_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_10_0
#define init_8_12_10_0
void FASTCALL call_8_12_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_10_1
#define init_8_12_10_1
void FASTCALL call_8_12_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_10_2
#define init_8_12_10_2
void FASTCALL call_8_12_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_10_3
#define init_8_12_10_3
void FASTCALL call_8_12_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_10_4
#define init_8_12_10_4
void FASTCALL call_8_12_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_10_5
#define init_8_12_10_5
void FASTCALL call_8_12_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_10_6
#define init_8_12_10_6
void FASTCALL call_8_12_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_10_7
#define init_8_12_10_7
void FASTCALL call_8_12_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_10_8
#define init_8_12_10_8
void FASTCALL call_8_12_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_10_9
#define init_8_12_10_9
void FASTCALL call_8_12_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_10_10
#define init_8_12_10_10
void FASTCALL call_8_12_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_10_11
#define init_8_12_10_11
void FASTCALL call_8_12_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_10_12
#define init_8_12_10_12
void FASTCALL call_8_12_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_10_13
#define init_8_12_10_13
void FASTCALL call_8_12_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_10_14
#define init_8_12_10_14
void FASTCALL call_8_12_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_10_15
#define init_8_12_10_15
void FASTCALL call_8_12_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_11_0
#define init_8_12_11_0
void FASTCALL call_8_12_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_11_1
#define init_8_12_11_1
void FASTCALL call_8_12_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_11_2
#define init_8_12_11_2
void FASTCALL call_8_12_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_11_3
#define init_8_12_11_3
void FASTCALL call_8_12_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_11_4
#define init_8_12_11_4
void FASTCALL call_8_12_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_11_5
#define init_8_12_11_5
void FASTCALL call_8_12_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_11_6
#define init_8_12_11_6
void FASTCALL call_8_12_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_11_7
#define init_8_12_11_7
void FASTCALL call_8_12_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_11_8
#define init_8_12_11_8
void FASTCALL call_8_12_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_11_9
#define init_8_12_11_9
void FASTCALL call_8_12_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_11_10
#define init_8_12_11_10
void FASTCALL call_8_12_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_11_11
#define init_8_12_11_11
void FASTCALL call_8_12_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_11_12
#define init_8_12_11_12
void FASTCALL call_8_12_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_11_13
#define init_8_12_11_13
void FASTCALL call_8_12_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_11_14
#define init_8_12_11_14
void FASTCALL call_8_12_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_11_15
#define init_8_12_11_15
void FASTCALL call_8_12_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_12_0
#define init_8_12_12_0
void FASTCALL call_8_12_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_12_1
#define init_8_12_12_1
void FASTCALL call_8_12_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_12_2
#define init_8_12_12_2
void FASTCALL call_8_12_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_12_3
#define init_8_12_12_3
void FASTCALL call_8_12_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_12_4
#define init_8_12_12_4
void FASTCALL call_8_12_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_12_5
#define init_8_12_12_5
void FASTCALL call_8_12_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_12_6
#define init_8_12_12_6
void FASTCALL call_8_12_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_12_7
#define init_8_12_12_7
void FASTCALL call_8_12_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_12_8
#define init_8_12_12_8
void FASTCALL call_8_12_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_12_9
#define init_8_12_12_9
void FASTCALL call_8_12_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_12_10
#define init_8_12_12_10
void FASTCALL call_8_12_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_12_11
#define init_8_12_12_11
void FASTCALL call_8_12_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_12_12
#define init_8_12_12_12
void FASTCALL call_8_12_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_12_13
#define init_8_12_12_13
void FASTCALL call_8_12_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_12_14
#define init_8_12_12_14
void FASTCALL call_8_12_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_12_15
#define init_8_12_12_15
void FASTCALL call_8_12_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_13_0
#define init_8_12_13_0
void FASTCALL call_8_12_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_13_1
#define init_8_12_13_1
void FASTCALL call_8_12_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_13_2
#define init_8_12_13_2
void FASTCALL call_8_12_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_13_3
#define init_8_12_13_3
void FASTCALL call_8_12_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_13_4
#define init_8_12_13_4
void FASTCALL call_8_12_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_13_5
#define init_8_12_13_5
void FASTCALL call_8_12_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_13_6
#define init_8_12_13_6
void FASTCALL call_8_12_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_13_7
#define init_8_12_13_7
void FASTCALL call_8_12_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_13_8
#define init_8_12_13_8
void FASTCALL call_8_12_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_13_9
#define init_8_12_13_9
void FASTCALL call_8_12_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_13_10
#define init_8_12_13_10
void FASTCALL call_8_12_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_13_11
#define init_8_12_13_11
void FASTCALL call_8_12_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_13_12
#define init_8_12_13_12
void FASTCALL call_8_12_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_13_13
#define init_8_12_13_13
void FASTCALL call_8_12_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_13_14
#define init_8_12_13_14
void FASTCALL call_8_12_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_13_15
#define init_8_12_13_15
void FASTCALL call_8_12_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_14_0
#define init_8_12_14_0
void FASTCALL call_8_12_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_14_1
#define init_8_12_14_1
void FASTCALL call_8_12_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_14_2
#define init_8_12_14_2
void FASTCALL call_8_12_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_14_3
#define init_8_12_14_3
void FASTCALL call_8_12_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_14_4
#define init_8_12_14_4
void FASTCALL call_8_12_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_14_5
#define init_8_12_14_5
void FASTCALL call_8_12_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_14_6
#define init_8_12_14_6
void FASTCALL call_8_12_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_14_7
#define init_8_12_14_7
void FASTCALL call_8_12_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_14_8
#define init_8_12_14_8
void FASTCALL call_8_12_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_14_9
#define init_8_12_14_9
void FASTCALL call_8_12_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_14_10
#define init_8_12_14_10
void FASTCALL call_8_12_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_14_11
#define init_8_12_14_11
void FASTCALL call_8_12_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_14_12
#define init_8_12_14_12
void FASTCALL call_8_12_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_14_13
#define init_8_12_14_13
void FASTCALL call_8_12_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_14_14
#define init_8_12_14_14
void FASTCALL call_8_12_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_14_15
#define init_8_12_14_15
void FASTCALL call_8_12_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_15_0
#define init_8_12_15_0
void FASTCALL call_8_12_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_15_1
#define init_8_12_15_1
void FASTCALL call_8_12_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_15_2
#define init_8_12_15_2
void FASTCALL call_8_12_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_15_3
#define init_8_12_15_3
void FASTCALL call_8_12_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_15_4
#define init_8_12_15_4
void FASTCALL call_8_12_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_15_5
#define init_8_12_15_5
void FASTCALL call_8_12_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_15_6
#define init_8_12_15_6
void FASTCALL call_8_12_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_15_7
#define init_8_12_15_7
void FASTCALL call_8_12_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_15_8
#define init_8_12_15_8
void FASTCALL call_8_12_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_15_9
#define init_8_12_15_9
void FASTCALL call_8_12_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_15_10
#define init_8_12_15_10
void FASTCALL call_8_12_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_15_11
#define init_8_12_15_11
void FASTCALL call_8_12_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_15_12
#define init_8_12_15_12
void FASTCALL call_8_12_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_15_13
#define init_8_12_15_13
void FASTCALL call_8_12_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_15_14
#define init_8_12_15_14
void FASTCALL call_8_12_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_12_15_15
#define init_8_12_15_15
void FASTCALL call_8_12_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_0_0
#define init_8_13_0_0
void FASTCALL call_8_13_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_0_1
#define init_8_13_0_1
void FASTCALL call_8_13_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_0_2
#define init_8_13_0_2
void FASTCALL call_8_13_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_0_3
#define init_8_13_0_3
void FASTCALL call_8_13_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_0_4
#define init_8_13_0_4
void FASTCALL call_8_13_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_0_5
#define init_8_13_0_5
void FASTCALL call_8_13_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_0_6
#define init_8_13_0_6
void FASTCALL call_8_13_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_0_7
#define init_8_13_0_7
void FASTCALL call_8_13_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_0_8
#define init_8_13_0_8
void FASTCALL call_8_13_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_0_9
#define init_8_13_0_9
void FASTCALL call_8_13_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_0_10
#define init_8_13_0_10
void FASTCALL call_8_13_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_0_11
#define init_8_13_0_11
void FASTCALL call_8_13_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_0_12
#define init_8_13_0_12
void FASTCALL call_8_13_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_0_13
#define init_8_13_0_13
void FASTCALL call_8_13_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_0_14
#define init_8_13_0_14
void FASTCALL call_8_13_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_0_15
#define init_8_13_0_15
void FASTCALL call_8_13_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_1_0
#define init_8_13_1_0
void FASTCALL call_8_13_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_1_1
#define init_8_13_1_1
void FASTCALL call_8_13_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_1_2
#define init_8_13_1_2
void FASTCALL call_8_13_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_1_3
#define init_8_13_1_3
void FASTCALL call_8_13_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_1_4
#define init_8_13_1_4
void FASTCALL call_8_13_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_1_5
#define init_8_13_1_5
void FASTCALL call_8_13_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_1_6
#define init_8_13_1_6
void FASTCALL call_8_13_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_1_7
#define init_8_13_1_7
void FASTCALL call_8_13_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_1_8
#define init_8_13_1_8
void FASTCALL call_8_13_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_1_9
#define init_8_13_1_9
void FASTCALL call_8_13_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_1_10
#define init_8_13_1_10
void FASTCALL call_8_13_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_1_11
#define init_8_13_1_11
void FASTCALL call_8_13_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_1_12
#define init_8_13_1_12
void FASTCALL call_8_13_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_1_13
#define init_8_13_1_13
void FASTCALL call_8_13_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_1_14
#define init_8_13_1_14
void FASTCALL call_8_13_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_1_15
#define init_8_13_1_15
void FASTCALL call_8_13_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_2_0
#define init_8_13_2_0
void FASTCALL call_8_13_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_2_1
#define init_8_13_2_1
void FASTCALL call_8_13_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_2_2
#define init_8_13_2_2
void FASTCALL call_8_13_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_2_3
#define init_8_13_2_3
void FASTCALL call_8_13_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_2_4
#define init_8_13_2_4
void FASTCALL call_8_13_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_2_5
#define init_8_13_2_5
void FASTCALL call_8_13_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_2_6
#define init_8_13_2_6
void FASTCALL call_8_13_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_2_7
#define init_8_13_2_7
void FASTCALL call_8_13_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_2_8
#define init_8_13_2_8
void FASTCALL call_8_13_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_2_9
#define init_8_13_2_9
void FASTCALL call_8_13_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_2_10
#define init_8_13_2_10
void FASTCALL call_8_13_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_2_11
#define init_8_13_2_11
void FASTCALL call_8_13_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_2_12
#define init_8_13_2_12
void FASTCALL call_8_13_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_2_13
#define init_8_13_2_13
void FASTCALL call_8_13_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_2_14
#define init_8_13_2_14
void FASTCALL call_8_13_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_2_15
#define init_8_13_2_15
void FASTCALL call_8_13_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_3_0
#define init_8_13_3_0
void FASTCALL call_8_13_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_3_1
#define init_8_13_3_1
void FASTCALL call_8_13_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_3_2
#define init_8_13_3_2
void FASTCALL call_8_13_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_3_3
#define init_8_13_3_3
void FASTCALL call_8_13_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_3_4
#define init_8_13_3_4
void FASTCALL call_8_13_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_3_5
#define init_8_13_3_5
void FASTCALL call_8_13_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_3_6
#define init_8_13_3_6
void FASTCALL call_8_13_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_3_7
#define init_8_13_3_7
void FASTCALL call_8_13_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_3_8
#define init_8_13_3_8
void FASTCALL call_8_13_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_3_9
#define init_8_13_3_9
void FASTCALL call_8_13_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_3_10
#define init_8_13_3_10
void FASTCALL call_8_13_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_3_11
#define init_8_13_3_11
void FASTCALL call_8_13_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_3_12
#define init_8_13_3_12
void FASTCALL call_8_13_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_3_13
#define init_8_13_3_13
void FASTCALL call_8_13_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_3_14
#define init_8_13_3_14
void FASTCALL call_8_13_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_3_15
#define init_8_13_3_15
void FASTCALL call_8_13_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_4_0
#define init_8_13_4_0
void FASTCALL call_8_13_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_4_1
#define init_8_13_4_1
void FASTCALL call_8_13_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_4_2
#define init_8_13_4_2
void FASTCALL call_8_13_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_4_3
#define init_8_13_4_3
void FASTCALL call_8_13_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_4_4
#define init_8_13_4_4
void FASTCALL call_8_13_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_4_5
#define init_8_13_4_5
void FASTCALL call_8_13_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_4_6
#define init_8_13_4_6
void FASTCALL call_8_13_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_4_7
#define init_8_13_4_7
void FASTCALL call_8_13_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_4_8
#define init_8_13_4_8
void FASTCALL call_8_13_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_4_9
#define init_8_13_4_9
void FASTCALL call_8_13_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_4_10
#define init_8_13_4_10
void FASTCALL call_8_13_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_4_11
#define init_8_13_4_11
void FASTCALL call_8_13_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_4_12
#define init_8_13_4_12
void FASTCALL call_8_13_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_4_13
#define init_8_13_4_13
void FASTCALL call_8_13_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_4_14
#define init_8_13_4_14
void FASTCALL call_8_13_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_4_15
#define init_8_13_4_15
void FASTCALL call_8_13_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_5_0
#define init_8_13_5_0
void FASTCALL call_8_13_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_5_1
#define init_8_13_5_1
void FASTCALL call_8_13_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_5_2
#define init_8_13_5_2
void FASTCALL call_8_13_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_5_3
#define init_8_13_5_3
void FASTCALL call_8_13_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_5_4
#define init_8_13_5_4
void FASTCALL call_8_13_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_5_5
#define init_8_13_5_5
void FASTCALL call_8_13_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_5_6
#define init_8_13_5_6
void FASTCALL call_8_13_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_5_7
#define init_8_13_5_7
void FASTCALL call_8_13_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_5_8
#define init_8_13_5_8
void FASTCALL call_8_13_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_5_9
#define init_8_13_5_9
void FASTCALL call_8_13_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_5_10
#define init_8_13_5_10
void FASTCALL call_8_13_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_5_11
#define init_8_13_5_11
void FASTCALL call_8_13_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_5_12
#define init_8_13_5_12
void FASTCALL call_8_13_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_5_13
#define init_8_13_5_13
void FASTCALL call_8_13_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_5_14
#define init_8_13_5_14
void FASTCALL call_8_13_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_5_15
#define init_8_13_5_15
void FASTCALL call_8_13_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_6_0
#define init_8_13_6_0
void FASTCALL call_8_13_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_6_1
#define init_8_13_6_1
void FASTCALL call_8_13_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_6_2
#define init_8_13_6_2
void FASTCALL call_8_13_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_6_3
#define init_8_13_6_3
void FASTCALL call_8_13_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_6_4
#define init_8_13_6_4
void FASTCALL call_8_13_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_6_5
#define init_8_13_6_5
void FASTCALL call_8_13_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_6_6
#define init_8_13_6_6
void FASTCALL call_8_13_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_6_7
#define init_8_13_6_7
void FASTCALL call_8_13_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_6_8
#define init_8_13_6_8
void FASTCALL call_8_13_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_6_9
#define init_8_13_6_9
void FASTCALL call_8_13_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_6_10
#define init_8_13_6_10
void FASTCALL call_8_13_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_6_11
#define init_8_13_6_11
void FASTCALL call_8_13_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_6_12
#define init_8_13_6_12
void FASTCALL call_8_13_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_6_13
#define init_8_13_6_13
void FASTCALL call_8_13_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_6_14
#define init_8_13_6_14
void FASTCALL call_8_13_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_6_15
#define init_8_13_6_15
void FASTCALL call_8_13_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_7_0
#define init_8_13_7_0
void FASTCALL call_8_13_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_7_1
#define init_8_13_7_1
void FASTCALL call_8_13_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_7_2
#define init_8_13_7_2
void FASTCALL call_8_13_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_7_3
#define init_8_13_7_3
void FASTCALL call_8_13_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_7_4
#define init_8_13_7_4
void FASTCALL call_8_13_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_7_5
#define init_8_13_7_5
void FASTCALL call_8_13_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_7_6
#define init_8_13_7_6
void FASTCALL call_8_13_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_7_7
#define init_8_13_7_7
void FASTCALL call_8_13_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_7_8
#define init_8_13_7_8
void FASTCALL call_8_13_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_7_9
#define init_8_13_7_9
void FASTCALL call_8_13_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_7_10
#define init_8_13_7_10
void FASTCALL call_8_13_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_7_11
#define init_8_13_7_11
void FASTCALL call_8_13_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_7_12
#define init_8_13_7_12
void FASTCALL call_8_13_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_7_13
#define init_8_13_7_13
void FASTCALL call_8_13_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_7_14
#define init_8_13_7_14
void FASTCALL call_8_13_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_7_15
#define init_8_13_7_15
void FASTCALL call_8_13_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_8_0
#define init_8_13_8_0
void FASTCALL call_8_13_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_8_1
#define init_8_13_8_1
void FASTCALL call_8_13_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_8_2
#define init_8_13_8_2
void FASTCALL call_8_13_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_8_3
#define init_8_13_8_3
void FASTCALL call_8_13_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_8_4
#define init_8_13_8_4
void FASTCALL call_8_13_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_8_5
#define init_8_13_8_5
void FASTCALL call_8_13_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_8_6
#define init_8_13_8_6
void FASTCALL call_8_13_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_8_7
#define init_8_13_8_7
void FASTCALL call_8_13_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_8_8
#define init_8_13_8_8
void FASTCALL call_8_13_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_8_9
#define init_8_13_8_9
void FASTCALL call_8_13_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_8_10
#define init_8_13_8_10
void FASTCALL call_8_13_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_8_11
#define init_8_13_8_11
void FASTCALL call_8_13_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_8_12
#define init_8_13_8_12
void FASTCALL call_8_13_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_8_13
#define init_8_13_8_13
void FASTCALL call_8_13_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_8_14
#define init_8_13_8_14
void FASTCALL call_8_13_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_8_15
#define init_8_13_8_15
void FASTCALL call_8_13_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_9_0
#define init_8_13_9_0
void FASTCALL call_8_13_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_9_1
#define init_8_13_9_1
void FASTCALL call_8_13_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_9_2
#define init_8_13_9_2
void FASTCALL call_8_13_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_9_3
#define init_8_13_9_3
void FASTCALL call_8_13_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_9_4
#define init_8_13_9_4
void FASTCALL call_8_13_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_9_5
#define init_8_13_9_5
void FASTCALL call_8_13_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_9_6
#define init_8_13_9_6
void FASTCALL call_8_13_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_9_7
#define init_8_13_9_7
void FASTCALL call_8_13_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_9_8
#define init_8_13_9_8
void FASTCALL call_8_13_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_9_9
#define init_8_13_9_9
void FASTCALL call_8_13_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_9_10
#define init_8_13_9_10
void FASTCALL call_8_13_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_9_11
#define init_8_13_9_11
void FASTCALL call_8_13_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_9_12
#define init_8_13_9_12
void FASTCALL call_8_13_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_9_13
#define init_8_13_9_13
void FASTCALL call_8_13_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_9_14
#define init_8_13_9_14
void FASTCALL call_8_13_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_9_15
#define init_8_13_9_15
void FASTCALL call_8_13_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_10_0
#define init_8_13_10_0
void FASTCALL call_8_13_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_10_1
#define init_8_13_10_1
void FASTCALL call_8_13_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_10_2
#define init_8_13_10_2
void FASTCALL call_8_13_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_10_3
#define init_8_13_10_3
void FASTCALL call_8_13_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_10_4
#define init_8_13_10_4
void FASTCALL call_8_13_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_10_5
#define init_8_13_10_5
void FASTCALL call_8_13_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_10_6
#define init_8_13_10_6
void FASTCALL call_8_13_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_10_7
#define init_8_13_10_7
void FASTCALL call_8_13_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_10_8
#define init_8_13_10_8
void FASTCALL call_8_13_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_10_9
#define init_8_13_10_9
void FASTCALL call_8_13_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_10_10
#define init_8_13_10_10
void FASTCALL call_8_13_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_10_11
#define init_8_13_10_11
void FASTCALL call_8_13_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_10_12
#define init_8_13_10_12
void FASTCALL call_8_13_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_10_13
#define init_8_13_10_13
void FASTCALL call_8_13_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_10_14
#define init_8_13_10_14
void FASTCALL call_8_13_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_10_15
#define init_8_13_10_15
void FASTCALL call_8_13_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_11_0
#define init_8_13_11_0
void FASTCALL call_8_13_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_11_1
#define init_8_13_11_1
void FASTCALL call_8_13_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_11_2
#define init_8_13_11_2
void FASTCALL call_8_13_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_11_3
#define init_8_13_11_3
void FASTCALL call_8_13_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_11_4
#define init_8_13_11_4
void FASTCALL call_8_13_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_11_5
#define init_8_13_11_5
void FASTCALL call_8_13_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_11_6
#define init_8_13_11_6
void FASTCALL call_8_13_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_11_7
#define init_8_13_11_7
void FASTCALL call_8_13_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_11_8
#define init_8_13_11_8
void FASTCALL call_8_13_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_11_9
#define init_8_13_11_9
void FASTCALL call_8_13_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_11_10
#define init_8_13_11_10
void FASTCALL call_8_13_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_11_11
#define init_8_13_11_11
void FASTCALL call_8_13_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_11_12
#define init_8_13_11_12
void FASTCALL call_8_13_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_11_13
#define init_8_13_11_13
void FASTCALL call_8_13_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_11_14
#define init_8_13_11_14
void FASTCALL call_8_13_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_11_15
#define init_8_13_11_15
void FASTCALL call_8_13_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_12_0
#define init_8_13_12_0
void FASTCALL call_8_13_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_12_1
#define init_8_13_12_1
void FASTCALL call_8_13_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_12_2
#define init_8_13_12_2
void FASTCALL call_8_13_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_12_3
#define init_8_13_12_3
void FASTCALL call_8_13_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_12_4
#define init_8_13_12_4
void FASTCALL call_8_13_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_12_5
#define init_8_13_12_5
void FASTCALL call_8_13_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_12_6
#define init_8_13_12_6
void FASTCALL call_8_13_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_12_7
#define init_8_13_12_7
void FASTCALL call_8_13_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_12_8
#define init_8_13_12_8
void FASTCALL call_8_13_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_12_9
#define init_8_13_12_9
void FASTCALL call_8_13_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_12_10
#define init_8_13_12_10
void FASTCALL call_8_13_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_12_11
#define init_8_13_12_11
void FASTCALL call_8_13_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_12_12
#define init_8_13_12_12
void FASTCALL call_8_13_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_12_13
#define init_8_13_12_13
void FASTCALL call_8_13_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_12_14
#define init_8_13_12_14
void FASTCALL call_8_13_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_12_15
#define init_8_13_12_15
void FASTCALL call_8_13_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_13_0
#define init_8_13_13_0
void FASTCALL call_8_13_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_13_1
#define init_8_13_13_1
void FASTCALL call_8_13_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_13_2
#define init_8_13_13_2
void FASTCALL call_8_13_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_13_3
#define init_8_13_13_3
void FASTCALL call_8_13_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_13_4
#define init_8_13_13_4
void FASTCALL call_8_13_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_13_5
#define init_8_13_13_5
void FASTCALL call_8_13_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_13_6
#define init_8_13_13_6
void FASTCALL call_8_13_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_13_7
#define init_8_13_13_7
void FASTCALL call_8_13_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_13_8
#define init_8_13_13_8
void FASTCALL call_8_13_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_13_9
#define init_8_13_13_9
void FASTCALL call_8_13_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_13_10
#define init_8_13_13_10
void FASTCALL call_8_13_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_13_11
#define init_8_13_13_11
void FASTCALL call_8_13_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_13_12
#define init_8_13_13_12
void FASTCALL call_8_13_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_13_13
#define init_8_13_13_13
void FASTCALL call_8_13_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_13_14
#define init_8_13_13_14
void FASTCALL call_8_13_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_13_15
#define init_8_13_13_15
void FASTCALL call_8_13_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_14_0
#define init_8_13_14_0
void FASTCALL call_8_13_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_14_1
#define init_8_13_14_1
void FASTCALL call_8_13_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_14_2
#define init_8_13_14_2
void FASTCALL call_8_13_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_14_3
#define init_8_13_14_3
void FASTCALL call_8_13_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_14_4
#define init_8_13_14_4
void FASTCALL call_8_13_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_14_5
#define init_8_13_14_5
void FASTCALL call_8_13_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_14_6
#define init_8_13_14_6
void FASTCALL call_8_13_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_14_7
#define init_8_13_14_7
void FASTCALL call_8_13_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_14_8
#define init_8_13_14_8
void FASTCALL call_8_13_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_14_9
#define init_8_13_14_9
void FASTCALL call_8_13_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_14_10
#define init_8_13_14_10
void FASTCALL call_8_13_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_14_11
#define init_8_13_14_11
void FASTCALL call_8_13_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_14_12
#define init_8_13_14_12
void FASTCALL call_8_13_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_14_13
#define init_8_13_14_13
void FASTCALL call_8_13_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_14_14
#define init_8_13_14_14
void FASTCALL call_8_13_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_14_15
#define init_8_13_14_15
void FASTCALL call_8_13_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_15_0
#define init_8_13_15_0
void FASTCALL call_8_13_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_15_1
#define init_8_13_15_1
void FASTCALL call_8_13_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_15_2
#define init_8_13_15_2
void FASTCALL call_8_13_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_15_3
#define init_8_13_15_3
void FASTCALL call_8_13_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_15_4
#define init_8_13_15_4
void FASTCALL call_8_13_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_15_5
#define init_8_13_15_5
void FASTCALL call_8_13_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_15_6
#define init_8_13_15_6
void FASTCALL call_8_13_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_15_7
#define init_8_13_15_7
void FASTCALL call_8_13_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_15_8
#define init_8_13_15_8
void FASTCALL call_8_13_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_15_9
#define init_8_13_15_9
void FASTCALL call_8_13_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_15_10
#define init_8_13_15_10
void FASTCALL call_8_13_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_15_11
#define init_8_13_15_11
void FASTCALL call_8_13_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_15_12
#define init_8_13_15_12
void FASTCALL call_8_13_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_15_13
#define init_8_13_15_13
void FASTCALL call_8_13_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_15_14
#define init_8_13_15_14
void FASTCALL call_8_13_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_13_15_15
#define init_8_13_15_15
void FASTCALL call_8_13_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_0_0
#define init_8_14_0_0
void FASTCALL call_8_14_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_0_1
#define init_8_14_0_1
void FASTCALL call_8_14_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_0_2
#define init_8_14_0_2
void FASTCALL call_8_14_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_0_3
#define init_8_14_0_3
void FASTCALL call_8_14_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_0_4
#define init_8_14_0_4
void FASTCALL call_8_14_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_0_5
#define init_8_14_0_5
void FASTCALL call_8_14_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_0_6
#define init_8_14_0_6
void FASTCALL call_8_14_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_0_7
#define init_8_14_0_7
void FASTCALL call_8_14_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_0_8
#define init_8_14_0_8
void FASTCALL call_8_14_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_0_9
#define init_8_14_0_9
void FASTCALL call_8_14_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_0_10
#define init_8_14_0_10
void FASTCALL call_8_14_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_0_11
#define init_8_14_0_11
void FASTCALL call_8_14_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_0_12
#define init_8_14_0_12
void FASTCALL call_8_14_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_0_13
#define init_8_14_0_13
void FASTCALL call_8_14_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_0_14
#define init_8_14_0_14
void FASTCALL call_8_14_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_0_15
#define init_8_14_0_15
void FASTCALL call_8_14_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_1_0
#define init_8_14_1_0
void FASTCALL call_8_14_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_1_1
#define init_8_14_1_1
void FASTCALL call_8_14_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_1_2
#define init_8_14_1_2
void FASTCALL call_8_14_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_1_3
#define init_8_14_1_3
void FASTCALL call_8_14_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_1_4
#define init_8_14_1_4
void FASTCALL call_8_14_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_1_5
#define init_8_14_1_5
void FASTCALL call_8_14_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_1_6
#define init_8_14_1_6
void FASTCALL call_8_14_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_1_7
#define init_8_14_1_7
void FASTCALL call_8_14_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_1_8
#define init_8_14_1_8
void FASTCALL call_8_14_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_1_9
#define init_8_14_1_9
void FASTCALL call_8_14_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_1_10
#define init_8_14_1_10
void FASTCALL call_8_14_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_1_11
#define init_8_14_1_11
void FASTCALL call_8_14_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_1_12
#define init_8_14_1_12
void FASTCALL call_8_14_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_1_13
#define init_8_14_1_13
void FASTCALL call_8_14_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_1_14
#define init_8_14_1_14
void FASTCALL call_8_14_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_1_15
#define init_8_14_1_15
void FASTCALL call_8_14_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_2_0
#define init_8_14_2_0
void FASTCALL call_8_14_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_2_1
#define init_8_14_2_1
void FASTCALL call_8_14_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_2_2
#define init_8_14_2_2
void FASTCALL call_8_14_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_2_3
#define init_8_14_2_3
void FASTCALL call_8_14_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_2_4
#define init_8_14_2_4
void FASTCALL call_8_14_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_2_5
#define init_8_14_2_5
void FASTCALL call_8_14_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_2_6
#define init_8_14_2_6
void FASTCALL call_8_14_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_2_7
#define init_8_14_2_7
void FASTCALL call_8_14_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_2_8
#define init_8_14_2_8
void FASTCALL call_8_14_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_2_9
#define init_8_14_2_9
void FASTCALL call_8_14_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_2_10
#define init_8_14_2_10
void FASTCALL call_8_14_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_2_11
#define init_8_14_2_11
void FASTCALL call_8_14_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_2_12
#define init_8_14_2_12
void FASTCALL call_8_14_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_2_13
#define init_8_14_2_13
void FASTCALL call_8_14_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_2_14
#define init_8_14_2_14
void FASTCALL call_8_14_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_2_15
#define init_8_14_2_15
void FASTCALL call_8_14_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_3_0
#define init_8_14_3_0
void FASTCALL call_8_14_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_3_1
#define init_8_14_3_1
void FASTCALL call_8_14_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_3_2
#define init_8_14_3_2
void FASTCALL call_8_14_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_3_3
#define init_8_14_3_3
void FASTCALL call_8_14_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_3_4
#define init_8_14_3_4
void FASTCALL call_8_14_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_3_5
#define init_8_14_3_5
void FASTCALL call_8_14_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_3_6
#define init_8_14_3_6
void FASTCALL call_8_14_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_3_7
#define init_8_14_3_7
void FASTCALL call_8_14_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_3_8
#define init_8_14_3_8
void FASTCALL call_8_14_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_3_9
#define init_8_14_3_9
void FASTCALL call_8_14_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_3_10
#define init_8_14_3_10
void FASTCALL call_8_14_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_3_11
#define init_8_14_3_11
void FASTCALL call_8_14_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_3_12
#define init_8_14_3_12
void FASTCALL call_8_14_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_3_13
#define init_8_14_3_13
void FASTCALL call_8_14_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_3_14
#define init_8_14_3_14
void FASTCALL call_8_14_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_3_15
#define init_8_14_3_15
void FASTCALL call_8_14_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_4_0
#define init_8_14_4_0
void FASTCALL call_8_14_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_4_1
#define init_8_14_4_1
void FASTCALL call_8_14_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_4_2
#define init_8_14_4_2
void FASTCALL call_8_14_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_4_3
#define init_8_14_4_3
void FASTCALL call_8_14_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_4_4
#define init_8_14_4_4
void FASTCALL call_8_14_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_4_5
#define init_8_14_4_5
void FASTCALL call_8_14_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_4_6
#define init_8_14_4_6
void FASTCALL call_8_14_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_4_7
#define init_8_14_4_7
void FASTCALL call_8_14_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_4_8
#define init_8_14_4_8
void FASTCALL call_8_14_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_4_9
#define init_8_14_4_9
void FASTCALL call_8_14_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_4_10
#define init_8_14_4_10
void FASTCALL call_8_14_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_4_11
#define init_8_14_4_11
void FASTCALL call_8_14_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_4_12
#define init_8_14_4_12
void FASTCALL call_8_14_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_4_13
#define init_8_14_4_13
void FASTCALL call_8_14_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_4_14
#define init_8_14_4_14
void FASTCALL call_8_14_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_4_15
#define init_8_14_4_15
void FASTCALL call_8_14_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_5_0
#define init_8_14_5_0
void FASTCALL call_8_14_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_5_1
#define init_8_14_5_1
void FASTCALL call_8_14_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_5_2
#define init_8_14_5_2
void FASTCALL call_8_14_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_5_3
#define init_8_14_5_3
void FASTCALL call_8_14_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_5_4
#define init_8_14_5_4
void FASTCALL call_8_14_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_5_5
#define init_8_14_5_5
void FASTCALL call_8_14_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_5_6
#define init_8_14_5_6
void FASTCALL call_8_14_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_5_7
#define init_8_14_5_7
void FASTCALL call_8_14_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_5_8
#define init_8_14_5_8
void FASTCALL call_8_14_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_5_9
#define init_8_14_5_9
void FASTCALL call_8_14_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_5_10
#define init_8_14_5_10
void FASTCALL call_8_14_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_5_11
#define init_8_14_5_11
void FASTCALL call_8_14_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_5_12
#define init_8_14_5_12
void FASTCALL call_8_14_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_5_13
#define init_8_14_5_13
void FASTCALL call_8_14_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_5_14
#define init_8_14_5_14
void FASTCALL call_8_14_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_5_15
#define init_8_14_5_15
void FASTCALL call_8_14_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_6_0
#define init_8_14_6_0
void FASTCALL call_8_14_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_6_1
#define init_8_14_6_1
void FASTCALL call_8_14_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_6_2
#define init_8_14_6_2
void FASTCALL call_8_14_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_6_3
#define init_8_14_6_3
void FASTCALL call_8_14_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_6_4
#define init_8_14_6_4
void FASTCALL call_8_14_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_6_5
#define init_8_14_6_5
void FASTCALL call_8_14_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_6_6
#define init_8_14_6_6
void FASTCALL call_8_14_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_6_7
#define init_8_14_6_7
void FASTCALL call_8_14_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_6_8
#define init_8_14_6_8
void FASTCALL call_8_14_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_6_9
#define init_8_14_6_9
void FASTCALL call_8_14_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_6_10
#define init_8_14_6_10
void FASTCALL call_8_14_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_6_11
#define init_8_14_6_11
void FASTCALL call_8_14_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_6_12
#define init_8_14_6_12
void FASTCALL call_8_14_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_6_13
#define init_8_14_6_13
void FASTCALL call_8_14_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_6_14
#define init_8_14_6_14
void FASTCALL call_8_14_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_6_15
#define init_8_14_6_15
void FASTCALL call_8_14_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_7_0
#define init_8_14_7_0
void FASTCALL call_8_14_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_7_1
#define init_8_14_7_1
void FASTCALL call_8_14_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_7_2
#define init_8_14_7_2
void FASTCALL call_8_14_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_7_3
#define init_8_14_7_3
void FASTCALL call_8_14_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_7_4
#define init_8_14_7_4
void FASTCALL call_8_14_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_7_5
#define init_8_14_7_5
void FASTCALL call_8_14_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_7_6
#define init_8_14_7_6
void FASTCALL call_8_14_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_7_7
#define init_8_14_7_7
void FASTCALL call_8_14_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_7_8
#define init_8_14_7_8
void FASTCALL call_8_14_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_7_9
#define init_8_14_7_9
void FASTCALL call_8_14_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_7_10
#define init_8_14_7_10
void FASTCALL call_8_14_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_7_11
#define init_8_14_7_11
void FASTCALL call_8_14_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_7_12
#define init_8_14_7_12
void FASTCALL call_8_14_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_7_13
#define init_8_14_7_13
void FASTCALL call_8_14_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_7_14
#define init_8_14_7_14
void FASTCALL call_8_14_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_7_15
#define init_8_14_7_15
void FASTCALL call_8_14_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_8_0
#define init_8_14_8_0
void FASTCALL call_8_14_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_8_1
#define init_8_14_8_1
void FASTCALL call_8_14_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_8_2
#define init_8_14_8_2
void FASTCALL call_8_14_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_8_3
#define init_8_14_8_3
void FASTCALL call_8_14_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_8_4
#define init_8_14_8_4
void FASTCALL call_8_14_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_8_5
#define init_8_14_8_5
void FASTCALL call_8_14_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_8_6
#define init_8_14_8_6
void FASTCALL call_8_14_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_8_7
#define init_8_14_8_7
void FASTCALL call_8_14_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_8_8
#define init_8_14_8_8
void FASTCALL call_8_14_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_8_9
#define init_8_14_8_9
void FASTCALL call_8_14_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_8_10
#define init_8_14_8_10
void FASTCALL call_8_14_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_8_11
#define init_8_14_8_11
void FASTCALL call_8_14_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_8_12
#define init_8_14_8_12
void FASTCALL call_8_14_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_8_13
#define init_8_14_8_13
void FASTCALL call_8_14_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_8_14
#define init_8_14_8_14
void FASTCALL call_8_14_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_8_15
#define init_8_14_8_15
void FASTCALL call_8_14_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_9_0
#define init_8_14_9_0
void FASTCALL call_8_14_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_9_1
#define init_8_14_9_1
void FASTCALL call_8_14_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_9_2
#define init_8_14_9_2
void FASTCALL call_8_14_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_9_3
#define init_8_14_9_3
void FASTCALL call_8_14_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_9_4
#define init_8_14_9_4
void FASTCALL call_8_14_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_9_5
#define init_8_14_9_5
void FASTCALL call_8_14_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_9_6
#define init_8_14_9_6
void FASTCALL call_8_14_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_9_7
#define init_8_14_9_7
void FASTCALL call_8_14_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_9_8
#define init_8_14_9_8
void FASTCALL call_8_14_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_9_9
#define init_8_14_9_9
void FASTCALL call_8_14_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_9_10
#define init_8_14_9_10
void FASTCALL call_8_14_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_9_11
#define init_8_14_9_11
void FASTCALL call_8_14_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_9_12
#define init_8_14_9_12
void FASTCALL call_8_14_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_9_13
#define init_8_14_9_13
void FASTCALL call_8_14_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_9_14
#define init_8_14_9_14
void FASTCALL call_8_14_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_9_15
#define init_8_14_9_15
void FASTCALL call_8_14_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_10_0
#define init_8_14_10_0
void FASTCALL call_8_14_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_10_1
#define init_8_14_10_1
void FASTCALL call_8_14_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_10_2
#define init_8_14_10_2
void FASTCALL call_8_14_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_10_3
#define init_8_14_10_3
void FASTCALL call_8_14_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_10_4
#define init_8_14_10_4
void FASTCALL call_8_14_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_10_5
#define init_8_14_10_5
void FASTCALL call_8_14_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_10_6
#define init_8_14_10_6
void FASTCALL call_8_14_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_10_7
#define init_8_14_10_7
void FASTCALL call_8_14_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_10_8
#define init_8_14_10_8
void FASTCALL call_8_14_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_10_9
#define init_8_14_10_9
void FASTCALL call_8_14_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_10_10
#define init_8_14_10_10
void FASTCALL call_8_14_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_10_11
#define init_8_14_10_11
void FASTCALL call_8_14_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_10_12
#define init_8_14_10_12
void FASTCALL call_8_14_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_10_13
#define init_8_14_10_13
void FASTCALL call_8_14_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_10_14
#define init_8_14_10_14
void FASTCALL call_8_14_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_10_15
#define init_8_14_10_15
void FASTCALL call_8_14_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_11_0
#define init_8_14_11_0
void FASTCALL call_8_14_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_11_1
#define init_8_14_11_1
void FASTCALL call_8_14_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_11_2
#define init_8_14_11_2
void FASTCALL call_8_14_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_11_3
#define init_8_14_11_3
void FASTCALL call_8_14_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_11_4
#define init_8_14_11_4
void FASTCALL call_8_14_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_11_5
#define init_8_14_11_5
void FASTCALL call_8_14_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_11_6
#define init_8_14_11_6
void FASTCALL call_8_14_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_11_7
#define init_8_14_11_7
void FASTCALL call_8_14_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_11_8
#define init_8_14_11_8
void FASTCALL call_8_14_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_11_9
#define init_8_14_11_9
void FASTCALL call_8_14_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_11_10
#define init_8_14_11_10
void FASTCALL call_8_14_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_11_11
#define init_8_14_11_11
void FASTCALL call_8_14_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_11_12
#define init_8_14_11_12
void FASTCALL call_8_14_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_11_13
#define init_8_14_11_13
void FASTCALL call_8_14_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_11_14
#define init_8_14_11_14
void FASTCALL call_8_14_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_11_15
#define init_8_14_11_15
void FASTCALL call_8_14_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_12_0
#define init_8_14_12_0
void FASTCALL call_8_14_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_12_1
#define init_8_14_12_1
void FASTCALL call_8_14_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_12_2
#define init_8_14_12_2
void FASTCALL call_8_14_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_12_3
#define init_8_14_12_3
void FASTCALL call_8_14_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_12_4
#define init_8_14_12_4
void FASTCALL call_8_14_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_12_5
#define init_8_14_12_5
void FASTCALL call_8_14_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_12_6
#define init_8_14_12_6
void FASTCALL call_8_14_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_12_7
#define init_8_14_12_7
void FASTCALL call_8_14_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_12_8
#define init_8_14_12_8
void FASTCALL call_8_14_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_12_9
#define init_8_14_12_9
void FASTCALL call_8_14_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_12_10
#define init_8_14_12_10
void FASTCALL call_8_14_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_12_11
#define init_8_14_12_11
void FASTCALL call_8_14_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_12_12
#define init_8_14_12_12
void FASTCALL call_8_14_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_12_13
#define init_8_14_12_13
void FASTCALL call_8_14_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_12_14
#define init_8_14_12_14
void FASTCALL call_8_14_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_12_15
#define init_8_14_12_15
void FASTCALL call_8_14_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_13_0
#define init_8_14_13_0
void FASTCALL call_8_14_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_13_1
#define init_8_14_13_1
void FASTCALL call_8_14_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_13_2
#define init_8_14_13_2
void FASTCALL call_8_14_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_13_3
#define init_8_14_13_3
void FASTCALL call_8_14_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_13_4
#define init_8_14_13_4
void FASTCALL call_8_14_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_13_5
#define init_8_14_13_5
void FASTCALL call_8_14_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_13_6
#define init_8_14_13_6
void FASTCALL call_8_14_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_13_7
#define init_8_14_13_7
void FASTCALL call_8_14_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_13_8
#define init_8_14_13_8
void FASTCALL call_8_14_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_13_9
#define init_8_14_13_9
void FASTCALL call_8_14_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_13_10
#define init_8_14_13_10
void FASTCALL call_8_14_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_13_11
#define init_8_14_13_11
void FASTCALL call_8_14_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_13_12
#define init_8_14_13_12
void FASTCALL call_8_14_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_13_13
#define init_8_14_13_13
void FASTCALL call_8_14_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_13_14
#define init_8_14_13_14
void FASTCALL call_8_14_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_13_15
#define init_8_14_13_15
void FASTCALL call_8_14_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_14_0
#define init_8_14_14_0
void FASTCALL call_8_14_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_14_1
#define init_8_14_14_1
void FASTCALL call_8_14_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_14_2
#define init_8_14_14_2
void FASTCALL call_8_14_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_14_3
#define init_8_14_14_3
void FASTCALL call_8_14_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_14_4
#define init_8_14_14_4
void FASTCALL call_8_14_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_14_5
#define init_8_14_14_5
void FASTCALL call_8_14_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_14_6
#define init_8_14_14_6
void FASTCALL call_8_14_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_14_7
#define init_8_14_14_7
void FASTCALL call_8_14_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_14_8
#define init_8_14_14_8
void FASTCALL call_8_14_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_14_9
#define init_8_14_14_9
void FASTCALL call_8_14_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_14_10
#define init_8_14_14_10
void FASTCALL call_8_14_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_14_11
#define init_8_14_14_11
void FASTCALL call_8_14_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_14_12
#define init_8_14_14_12
void FASTCALL call_8_14_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_14_13
#define init_8_14_14_13
void FASTCALL call_8_14_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_14_14
#define init_8_14_14_14
void FASTCALL call_8_14_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_14_15
#define init_8_14_14_15
void FASTCALL call_8_14_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_15_0
#define init_8_14_15_0
void FASTCALL call_8_14_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_15_1
#define init_8_14_15_1
void FASTCALL call_8_14_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_15_2
#define init_8_14_15_2
void FASTCALL call_8_14_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_15_3
#define init_8_14_15_3
void FASTCALL call_8_14_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_15_4
#define init_8_14_15_4
void FASTCALL call_8_14_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_15_5
#define init_8_14_15_5
void FASTCALL call_8_14_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_15_6
#define init_8_14_15_6
void FASTCALL call_8_14_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_15_7
#define init_8_14_15_7
void FASTCALL call_8_14_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_15_8
#define init_8_14_15_8
void FASTCALL call_8_14_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_15_9
#define init_8_14_15_9
void FASTCALL call_8_14_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_15_10
#define init_8_14_15_10
void FASTCALL call_8_14_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_15_11
#define init_8_14_15_11
void FASTCALL call_8_14_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_15_12
#define init_8_14_15_12
void FASTCALL call_8_14_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_15_13
#define init_8_14_15_13
void FASTCALL call_8_14_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_15_14
#define init_8_14_15_14
void FASTCALL call_8_14_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_14_15_15
#define init_8_14_15_15
void FASTCALL call_8_14_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_0_0
#define init_8_15_0_0
void FASTCALL call_8_15_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_0_1
#define init_8_15_0_1
void FASTCALL call_8_15_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_0_2
#define init_8_15_0_2
void FASTCALL call_8_15_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_0_3
#define init_8_15_0_3
void FASTCALL call_8_15_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_0_4
#define init_8_15_0_4
void FASTCALL call_8_15_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_0_5
#define init_8_15_0_5
void FASTCALL call_8_15_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_0_6
#define init_8_15_0_6
void FASTCALL call_8_15_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_0_7
#define init_8_15_0_7
void FASTCALL call_8_15_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_0_8
#define init_8_15_0_8
void FASTCALL call_8_15_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_0_9
#define init_8_15_0_9
void FASTCALL call_8_15_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_0_10
#define init_8_15_0_10
void FASTCALL call_8_15_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_0_11
#define init_8_15_0_11
void FASTCALL call_8_15_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_0_12
#define init_8_15_0_12
void FASTCALL call_8_15_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_0_13
#define init_8_15_0_13
void FASTCALL call_8_15_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_0_14
#define init_8_15_0_14
void FASTCALL call_8_15_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_0_15
#define init_8_15_0_15
void FASTCALL call_8_15_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_1_0
#define init_8_15_1_0
void FASTCALL call_8_15_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_1_1
#define init_8_15_1_1
void FASTCALL call_8_15_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_1_2
#define init_8_15_1_2
void FASTCALL call_8_15_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_1_3
#define init_8_15_1_3
void FASTCALL call_8_15_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_1_4
#define init_8_15_1_4
void FASTCALL call_8_15_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_1_5
#define init_8_15_1_5
void FASTCALL call_8_15_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_1_6
#define init_8_15_1_6
void FASTCALL call_8_15_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_1_7
#define init_8_15_1_7
void FASTCALL call_8_15_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_1_8
#define init_8_15_1_8
void FASTCALL call_8_15_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_1_9
#define init_8_15_1_9
void FASTCALL call_8_15_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_1_10
#define init_8_15_1_10
void FASTCALL call_8_15_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_1_11
#define init_8_15_1_11
void FASTCALL call_8_15_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_1_12
#define init_8_15_1_12
void FASTCALL call_8_15_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_1_13
#define init_8_15_1_13
void FASTCALL call_8_15_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_1_14
#define init_8_15_1_14
void FASTCALL call_8_15_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_1_15
#define init_8_15_1_15
void FASTCALL call_8_15_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_2_0
#define init_8_15_2_0
void FASTCALL call_8_15_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_2_1
#define init_8_15_2_1
void FASTCALL call_8_15_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_2_2
#define init_8_15_2_2
void FASTCALL call_8_15_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_2_3
#define init_8_15_2_3
void FASTCALL call_8_15_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_2_4
#define init_8_15_2_4
void FASTCALL call_8_15_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_2_5
#define init_8_15_2_5
void FASTCALL call_8_15_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_2_6
#define init_8_15_2_6
void FASTCALL call_8_15_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_2_7
#define init_8_15_2_7
void FASTCALL call_8_15_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_2_8
#define init_8_15_2_8
void FASTCALL call_8_15_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_2_9
#define init_8_15_2_9
void FASTCALL call_8_15_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_2_10
#define init_8_15_2_10
void FASTCALL call_8_15_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_2_11
#define init_8_15_2_11
void FASTCALL call_8_15_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_2_12
#define init_8_15_2_12
void FASTCALL call_8_15_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_2_13
#define init_8_15_2_13
void FASTCALL call_8_15_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_2_14
#define init_8_15_2_14
void FASTCALL call_8_15_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_2_15
#define init_8_15_2_15
void FASTCALL call_8_15_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_3_0
#define init_8_15_3_0
void FASTCALL call_8_15_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_3_1
#define init_8_15_3_1
void FASTCALL call_8_15_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_3_2
#define init_8_15_3_2
void FASTCALL call_8_15_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_3_3
#define init_8_15_3_3
void FASTCALL call_8_15_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_3_4
#define init_8_15_3_4
void FASTCALL call_8_15_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_3_5
#define init_8_15_3_5
void FASTCALL call_8_15_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_3_6
#define init_8_15_3_6
void FASTCALL call_8_15_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_3_7
#define init_8_15_3_7
void FASTCALL call_8_15_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_3_8
#define init_8_15_3_8
void FASTCALL call_8_15_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_3_9
#define init_8_15_3_9
void FASTCALL call_8_15_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_3_10
#define init_8_15_3_10
void FASTCALL call_8_15_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_3_11
#define init_8_15_3_11
void FASTCALL call_8_15_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_3_12
#define init_8_15_3_12
void FASTCALL call_8_15_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_3_13
#define init_8_15_3_13
void FASTCALL call_8_15_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_3_14
#define init_8_15_3_14
void FASTCALL call_8_15_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_3_15
#define init_8_15_3_15
void FASTCALL call_8_15_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_4_0
#define init_8_15_4_0
void FASTCALL call_8_15_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_4_1
#define init_8_15_4_1
void FASTCALL call_8_15_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_4_2
#define init_8_15_4_2
void FASTCALL call_8_15_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_4_3
#define init_8_15_4_3
void FASTCALL call_8_15_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_4_4
#define init_8_15_4_4
void FASTCALL call_8_15_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_4_5
#define init_8_15_4_5
void FASTCALL call_8_15_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_4_6
#define init_8_15_4_6
void FASTCALL call_8_15_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_4_7
#define init_8_15_4_7
void FASTCALL call_8_15_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_4_8
#define init_8_15_4_8
void FASTCALL call_8_15_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_4_9
#define init_8_15_4_9
void FASTCALL call_8_15_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_4_10
#define init_8_15_4_10
void FASTCALL call_8_15_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_4_11
#define init_8_15_4_11
void FASTCALL call_8_15_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_4_12
#define init_8_15_4_12
void FASTCALL call_8_15_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_4_13
#define init_8_15_4_13
void FASTCALL call_8_15_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_4_14
#define init_8_15_4_14
void FASTCALL call_8_15_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_4_15
#define init_8_15_4_15
void FASTCALL call_8_15_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_5_0
#define init_8_15_5_0
void FASTCALL call_8_15_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_5_1
#define init_8_15_5_1
void FASTCALL call_8_15_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_5_2
#define init_8_15_5_2
void FASTCALL call_8_15_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_5_3
#define init_8_15_5_3
void FASTCALL call_8_15_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_5_4
#define init_8_15_5_4
void FASTCALL call_8_15_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_5_5
#define init_8_15_5_5
void FASTCALL call_8_15_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_5_6
#define init_8_15_5_6
void FASTCALL call_8_15_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_5_7
#define init_8_15_5_7
void FASTCALL call_8_15_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_5_8
#define init_8_15_5_8
void FASTCALL call_8_15_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_5_9
#define init_8_15_5_9
void FASTCALL call_8_15_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_5_10
#define init_8_15_5_10
void FASTCALL call_8_15_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_5_11
#define init_8_15_5_11
void FASTCALL call_8_15_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_5_12
#define init_8_15_5_12
void FASTCALL call_8_15_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_5_13
#define init_8_15_5_13
void FASTCALL call_8_15_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_5_14
#define init_8_15_5_14
void FASTCALL call_8_15_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_5_15
#define init_8_15_5_15
void FASTCALL call_8_15_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_6_0
#define init_8_15_6_0
void FASTCALL call_8_15_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_6_1
#define init_8_15_6_1
void FASTCALL call_8_15_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_6_2
#define init_8_15_6_2
void FASTCALL call_8_15_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_6_3
#define init_8_15_6_3
void FASTCALL call_8_15_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_6_4
#define init_8_15_6_4
void FASTCALL call_8_15_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_6_5
#define init_8_15_6_5
void FASTCALL call_8_15_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_6_6
#define init_8_15_6_6
void FASTCALL call_8_15_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_6_7
#define init_8_15_6_7
void FASTCALL call_8_15_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_6_8
#define init_8_15_6_8
void FASTCALL call_8_15_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_6_9
#define init_8_15_6_9
void FASTCALL call_8_15_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_6_10
#define init_8_15_6_10
void FASTCALL call_8_15_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_6_11
#define init_8_15_6_11
void FASTCALL call_8_15_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_6_12
#define init_8_15_6_12
void FASTCALL call_8_15_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_6_13
#define init_8_15_6_13
void FASTCALL call_8_15_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_6_14
#define init_8_15_6_14
void FASTCALL call_8_15_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_6_15
#define init_8_15_6_15
void FASTCALL call_8_15_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_7_0
#define init_8_15_7_0
void FASTCALL call_8_15_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_7_1
#define init_8_15_7_1
void FASTCALL call_8_15_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_7_2
#define init_8_15_7_2
void FASTCALL call_8_15_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_7_3
#define init_8_15_7_3
void FASTCALL call_8_15_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_7_4
#define init_8_15_7_4
void FASTCALL call_8_15_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_7_5
#define init_8_15_7_5
void FASTCALL call_8_15_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_7_6
#define init_8_15_7_6
void FASTCALL call_8_15_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_7_7
#define init_8_15_7_7
void FASTCALL call_8_15_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_7_8
#define init_8_15_7_8
void FASTCALL call_8_15_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_7_9
#define init_8_15_7_9
void FASTCALL call_8_15_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_7_10
#define init_8_15_7_10
void FASTCALL call_8_15_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_7_11
#define init_8_15_7_11
void FASTCALL call_8_15_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_7_12
#define init_8_15_7_12
void FASTCALL call_8_15_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_7_13
#define init_8_15_7_13
void FASTCALL call_8_15_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_7_14
#define init_8_15_7_14
void FASTCALL call_8_15_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_7_15
#define init_8_15_7_15
void FASTCALL call_8_15_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_8_0
#define init_8_15_8_0
void FASTCALL call_8_15_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_8_1
#define init_8_15_8_1
void FASTCALL call_8_15_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_8_2
#define init_8_15_8_2
void FASTCALL call_8_15_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_8_3
#define init_8_15_8_3
void FASTCALL call_8_15_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_8_4
#define init_8_15_8_4
void FASTCALL call_8_15_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_8_5
#define init_8_15_8_5
void FASTCALL call_8_15_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_8_6
#define init_8_15_8_6
void FASTCALL call_8_15_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_8_7
#define init_8_15_8_7
void FASTCALL call_8_15_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_8_8
#define init_8_15_8_8
void FASTCALL call_8_15_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_8_9
#define init_8_15_8_9
void FASTCALL call_8_15_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_8_10
#define init_8_15_8_10
void FASTCALL call_8_15_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_8_11
#define init_8_15_8_11
void FASTCALL call_8_15_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_8_12
#define init_8_15_8_12
void FASTCALL call_8_15_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_8_13
#define init_8_15_8_13
void FASTCALL call_8_15_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_8_14
#define init_8_15_8_14
void FASTCALL call_8_15_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_8_15
#define init_8_15_8_15
void FASTCALL call_8_15_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_9_0
#define init_8_15_9_0
void FASTCALL call_8_15_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_9_1
#define init_8_15_9_1
void FASTCALL call_8_15_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_9_2
#define init_8_15_9_2
void FASTCALL call_8_15_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_9_3
#define init_8_15_9_3
void FASTCALL call_8_15_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_9_4
#define init_8_15_9_4
void FASTCALL call_8_15_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_9_5
#define init_8_15_9_5
void FASTCALL call_8_15_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_9_6
#define init_8_15_9_6
void FASTCALL call_8_15_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_9_7
#define init_8_15_9_7
void FASTCALL call_8_15_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_9_8
#define init_8_15_9_8
void FASTCALL call_8_15_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_9_9
#define init_8_15_9_9
void FASTCALL call_8_15_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_9_10
#define init_8_15_9_10
void FASTCALL call_8_15_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_9_11
#define init_8_15_9_11
void FASTCALL call_8_15_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_9_12
#define init_8_15_9_12
void FASTCALL call_8_15_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_9_13
#define init_8_15_9_13
void FASTCALL call_8_15_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_9_14
#define init_8_15_9_14
void FASTCALL call_8_15_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_9_15
#define init_8_15_9_15
void FASTCALL call_8_15_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_10_0
#define init_8_15_10_0
void FASTCALL call_8_15_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_10_1
#define init_8_15_10_1
void FASTCALL call_8_15_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_10_2
#define init_8_15_10_2
void FASTCALL call_8_15_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_10_3
#define init_8_15_10_3
void FASTCALL call_8_15_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_10_4
#define init_8_15_10_4
void FASTCALL call_8_15_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_10_5
#define init_8_15_10_5
void FASTCALL call_8_15_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_10_6
#define init_8_15_10_6
void FASTCALL call_8_15_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_10_7
#define init_8_15_10_7
void FASTCALL call_8_15_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_10_8
#define init_8_15_10_8
void FASTCALL call_8_15_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_10_9
#define init_8_15_10_9
void FASTCALL call_8_15_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_10_10
#define init_8_15_10_10
void FASTCALL call_8_15_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_10_11
#define init_8_15_10_11
void FASTCALL call_8_15_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_10_12
#define init_8_15_10_12
void FASTCALL call_8_15_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_10_13
#define init_8_15_10_13
void FASTCALL call_8_15_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_10_14
#define init_8_15_10_14
void FASTCALL call_8_15_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_10_15
#define init_8_15_10_15
void FASTCALL call_8_15_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_11_0
#define init_8_15_11_0
void FASTCALL call_8_15_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_11_1
#define init_8_15_11_1
void FASTCALL call_8_15_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_11_2
#define init_8_15_11_2
void FASTCALL call_8_15_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_11_3
#define init_8_15_11_3
void FASTCALL call_8_15_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_11_4
#define init_8_15_11_4
void FASTCALL call_8_15_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_11_5
#define init_8_15_11_5
void FASTCALL call_8_15_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_11_6
#define init_8_15_11_6
void FASTCALL call_8_15_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_11_7
#define init_8_15_11_7
void FASTCALL call_8_15_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_11_8
#define init_8_15_11_8
void FASTCALL call_8_15_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_11_9
#define init_8_15_11_9
void FASTCALL call_8_15_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_11_10
#define init_8_15_11_10
void FASTCALL call_8_15_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_11_11
#define init_8_15_11_11
void FASTCALL call_8_15_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_11_12
#define init_8_15_11_12
void FASTCALL call_8_15_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_11_13
#define init_8_15_11_13
void FASTCALL call_8_15_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_11_14
#define init_8_15_11_14
void FASTCALL call_8_15_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_11_15
#define init_8_15_11_15
void FASTCALL call_8_15_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_12_0
#define init_8_15_12_0
void FASTCALL call_8_15_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_12_1
#define init_8_15_12_1
void FASTCALL call_8_15_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_12_2
#define init_8_15_12_2
void FASTCALL call_8_15_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_12_3
#define init_8_15_12_3
void FASTCALL call_8_15_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_12_4
#define init_8_15_12_4
void FASTCALL call_8_15_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_12_5
#define init_8_15_12_5
void FASTCALL call_8_15_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_12_6
#define init_8_15_12_6
void FASTCALL call_8_15_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_12_7
#define init_8_15_12_7
void FASTCALL call_8_15_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_12_8
#define init_8_15_12_8
void FASTCALL call_8_15_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_12_9
#define init_8_15_12_9
void FASTCALL call_8_15_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_12_10
#define init_8_15_12_10
void FASTCALL call_8_15_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_12_11
#define init_8_15_12_11
void FASTCALL call_8_15_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_12_12
#define init_8_15_12_12
void FASTCALL call_8_15_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_12_13
#define init_8_15_12_13
void FASTCALL call_8_15_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_12_14
#define init_8_15_12_14
void FASTCALL call_8_15_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_12_15
#define init_8_15_12_15
void FASTCALL call_8_15_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_13_0
#define init_8_15_13_0
void FASTCALL call_8_15_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_13_1
#define init_8_15_13_1
void FASTCALL call_8_15_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_13_2
#define init_8_15_13_2
void FASTCALL call_8_15_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_13_3
#define init_8_15_13_3
void FASTCALL call_8_15_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_13_4
#define init_8_15_13_4
void FASTCALL call_8_15_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_13_5
#define init_8_15_13_5
void FASTCALL call_8_15_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_13_6
#define init_8_15_13_6
void FASTCALL call_8_15_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_13_7
#define init_8_15_13_7
void FASTCALL call_8_15_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_13_8
#define init_8_15_13_8
void FASTCALL call_8_15_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_13_9
#define init_8_15_13_9
void FASTCALL call_8_15_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_13_10
#define init_8_15_13_10
void FASTCALL call_8_15_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_13_11
#define init_8_15_13_11
void FASTCALL call_8_15_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_13_12
#define init_8_15_13_12
void FASTCALL call_8_15_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_13_13
#define init_8_15_13_13
void FASTCALL call_8_15_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_13_14
#define init_8_15_13_14
void FASTCALL call_8_15_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_13_15
#define init_8_15_13_15
void FASTCALL call_8_15_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_14_0
#define init_8_15_14_0
void FASTCALL call_8_15_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_14_1
#define init_8_15_14_1
void FASTCALL call_8_15_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_14_2
#define init_8_15_14_2
void FASTCALL call_8_15_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_14_3
#define init_8_15_14_3
void FASTCALL call_8_15_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_14_4
#define init_8_15_14_4
void FASTCALL call_8_15_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_14_5
#define init_8_15_14_5
void FASTCALL call_8_15_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_14_6
#define init_8_15_14_6
void FASTCALL call_8_15_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_14_7
#define init_8_15_14_7
void FASTCALL call_8_15_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_14_8
#define init_8_15_14_8
void FASTCALL call_8_15_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_14_9
#define init_8_15_14_9
void FASTCALL call_8_15_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_14_10
#define init_8_15_14_10
void FASTCALL call_8_15_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_14_11
#define init_8_15_14_11
void FASTCALL call_8_15_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_14_12
#define init_8_15_14_12
void FASTCALL call_8_15_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_14_13
#define init_8_15_14_13
void FASTCALL call_8_15_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_14_14
#define init_8_15_14_14
void FASTCALL call_8_15_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_14_15
#define init_8_15_14_15
void FASTCALL call_8_15_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_15_0
#define init_8_15_15_0
void FASTCALL call_8_15_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_15_1
#define init_8_15_15_1
void FASTCALL call_8_15_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_15_2
#define init_8_15_15_2
void FASTCALL call_8_15_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_15_3
#define init_8_15_15_3
void FASTCALL call_8_15_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_15_4
#define init_8_15_15_4
void FASTCALL call_8_15_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_15_5
#define init_8_15_15_5
void FASTCALL call_8_15_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_15_6
#define init_8_15_15_6
void FASTCALL call_8_15_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_15_7
#define init_8_15_15_7
void FASTCALL call_8_15_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_15_8
#define init_8_15_15_8
void FASTCALL call_8_15_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_15_9
#define init_8_15_15_9
void FASTCALL call_8_15_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_15_10
#define init_8_15_15_10
void FASTCALL call_8_15_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_15_11
#define init_8_15_15_11
void FASTCALL call_8_15_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_15_12
#define init_8_15_15_12
void FASTCALL call_8_15_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_15_13
#define init_8_15_15_13
void FASTCALL call_8_15_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_15_14
#define init_8_15_15_14
void FASTCALL call_8_15_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_8_15_15_15
#define init_8_15_15_15
void FASTCALL call_8_15_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_0_0
#define init_9_0_0_0
void FASTCALL call_9_0_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_0_1
#define init_9_0_0_1
void FASTCALL call_9_0_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_0_2
#define init_9_0_0_2
void FASTCALL call_9_0_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_0_3
#define init_9_0_0_3
void FASTCALL call_9_0_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_0_4
#define init_9_0_0_4
void FASTCALL call_9_0_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_0_5
#define init_9_0_0_5
void FASTCALL call_9_0_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_0_6
#define init_9_0_0_6
void FASTCALL call_9_0_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_0_7
#define init_9_0_0_7
void FASTCALL call_9_0_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_0_8
#define init_9_0_0_8
void FASTCALL call_9_0_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_0_9
#define init_9_0_0_9
void FASTCALL call_9_0_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_0_10
#define init_9_0_0_10
void FASTCALL call_9_0_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_0_11
#define init_9_0_0_11
void FASTCALL call_9_0_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_0_12
#define init_9_0_0_12
void FASTCALL call_9_0_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_0_13
#define init_9_0_0_13
void FASTCALL call_9_0_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_0_14
#define init_9_0_0_14
void FASTCALL call_9_0_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_0_15
#define init_9_0_0_15
void FASTCALL call_9_0_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_1_0
#define init_9_0_1_0
void FASTCALL call_9_0_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_1_1
#define init_9_0_1_1
void FASTCALL call_9_0_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_1_2
#define init_9_0_1_2
void FASTCALL call_9_0_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_1_3
#define init_9_0_1_3
void FASTCALL call_9_0_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_1_4
#define init_9_0_1_4
void FASTCALL call_9_0_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_1_5
#define init_9_0_1_5
void FASTCALL call_9_0_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_1_6
#define init_9_0_1_6
void FASTCALL call_9_0_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_1_7
#define init_9_0_1_7
void FASTCALL call_9_0_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_1_8
#define init_9_0_1_8
void FASTCALL call_9_0_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_1_9
#define init_9_0_1_9
void FASTCALL call_9_0_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_1_10
#define init_9_0_1_10
void FASTCALL call_9_0_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_1_11
#define init_9_0_1_11
void FASTCALL call_9_0_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_1_12
#define init_9_0_1_12
void FASTCALL call_9_0_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_1_13
#define init_9_0_1_13
void FASTCALL call_9_0_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_1_14
#define init_9_0_1_14
void FASTCALL call_9_0_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_1_15
#define init_9_0_1_15
void FASTCALL call_9_0_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_2_0
#define init_9_0_2_0
void FASTCALL call_9_0_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_2_1
#define init_9_0_2_1
void FASTCALL call_9_0_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_2_2
#define init_9_0_2_2
void FASTCALL call_9_0_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_2_3
#define init_9_0_2_3
void FASTCALL call_9_0_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_2_4
#define init_9_0_2_4
void FASTCALL call_9_0_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_2_5
#define init_9_0_2_5
void FASTCALL call_9_0_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_2_6
#define init_9_0_2_6
void FASTCALL call_9_0_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_2_7
#define init_9_0_2_7
void FASTCALL call_9_0_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_2_8
#define init_9_0_2_8
void FASTCALL call_9_0_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_2_9
#define init_9_0_2_9
void FASTCALL call_9_0_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_2_10
#define init_9_0_2_10
void FASTCALL call_9_0_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_2_11
#define init_9_0_2_11
void FASTCALL call_9_0_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_2_12
#define init_9_0_2_12
void FASTCALL call_9_0_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_2_13
#define init_9_0_2_13
void FASTCALL call_9_0_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_2_14
#define init_9_0_2_14
void FASTCALL call_9_0_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_2_15
#define init_9_0_2_15
void FASTCALL call_9_0_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_3_0
#define init_9_0_3_0
void FASTCALL call_9_0_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_3_1
#define init_9_0_3_1
void FASTCALL call_9_0_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_3_2
#define init_9_0_3_2
void FASTCALL call_9_0_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_3_3
#define init_9_0_3_3
void FASTCALL call_9_0_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_3_4
#define init_9_0_3_4
void FASTCALL call_9_0_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_3_5
#define init_9_0_3_5
void FASTCALL call_9_0_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_3_6
#define init_9_0_3_6
void FASTCALL call_9_0_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_3_7
#define init_9_0_3_7
void FASTCALL call_9_0_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_3_8
#define init_9_0_3_8
void FASTCALL call_9_0_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_3_9
#define init_9_0_3_9
void FASTCALL call_9_0_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_3_10
#define init_9_0_3_10
void FASTCALL call_9_0_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_3_11
#define init_9_0_3_11
void FASTCALL call_9_0_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_3_12
#define init_9_0_3_12
void FASTCALL call_9_0_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_3_13
#define init_9_0_3_13
void FASTCALL call_9_0_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_3_14
#define init_9_0_3_14
void FASTCALL call_9_0_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_3_15
#define init_9_0_3_15
void FASTCALL call_9_0_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_4_0
#define init_9_0_4_0
void FASTCALL call_9_0_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_4_1
#define init_9_0_4_1
void FASTCALL call_9_0_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_4_2
#define init_9_0_4_2
void FASTCALL call_9_0_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_4_3
#define init_9_0_4_3
void FASTCALL call_9_0_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_4_4
#define init_9_0_4_4
void FASTCALL call_9_0_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_4_5
#define init_9_0_4_5
void FASTCALL call_9_0_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_4_6
#define init_9_0_4_6
void FASTCALL call_9_0_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_4_7
#define init_9_0_4_7
void FASTCALL call_9_0_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_4_8
#define init_9_0_4_8
void FASTCALL call_9_0_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_4_9
#define init_9_0_4_9
void FASTCALL call_9_0_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_4_10
#define init_9_0_4_10
void FASTCALL call_9_0_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_4_11
#define init_9_0_4_11
void FASTCALL call_9_0_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_4_12
#define init_9_0_4_12
void FASTCALL call_9_0_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_4_13
#define init_9_0_4_13
void FASTCALL call_9_0_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_4_14
#define init_9_0_4_14
void FASTCALL call_9_0_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_4_15
#define init_9_0_4_15
void FASTCALL call_9_0_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_5_0
#define init_9_0_5_0
void FASTCALL call_9_0_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_5_1
#define init_9_0_5_1
void FASTCALL call_9_0_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_5_2
#define init_9_0_5_2
void FASTCALL call_9_0_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_5_3
#define init_9_0_5_3
void FASTCALL call_9_0_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_5_4
#define init_9_0_5_4
void FASTCALL call_9_0_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_5_5
#define init_9_0_5_5
void FASTCALL call_9_0_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_5_6
#define init_9_0_5_6
void FASTCALL call_9_0_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_5_7
#define init_9_0_5_7
void FASTCALL call_9_0_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_5_8
#define init_9_0_5_8
void FASTCALL call_9_0_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_5_9
#define init_9_0_5_9
void FASTCALL call_9_0_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_5_10
#define init_9_0_5_10
void FASTCALL call_9_0_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_5_11
#define init_9_0_5_11
void FASTCALL call_9_0_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_5_12
#define init_9_0_5_12
void FASTCALL call_9_0_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_5_13
#define init_9_0_5_13
void FASTCALL call_9_0_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_5_14
#define init_9_0_5_14
void FASTCALL call_9_0_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_5_15
#define init_9_0_5_15
void FASTCALL call_9_0_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_6_0
#define init_9_0_6_0
void FASTCALL call_9_0_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_6_1
#define init_9_0_6_1
void FASTCALL call_9_0_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_6_2
#define init_9_0_6_2
void FASTCALL call_9_0_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_6_3
#define init_9_0_6_3
void FASTCALL call_9_0_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_6_4
#define init_9_0_6_4
void FASTCALL call_9_0_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_6_5
#define init_9_0_6_5
void FASTCALL call_9_0_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_6_6
#define init_9_0_6_6
void FASTCALL call_9_0_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_6_7
#define init_9_0_6_7
void FASTCALL call_9_0_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_6_8
#define init_9_0_6_8
void FASTCALL call_9_0_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_6_9
#define init_9_0_6_9
void FASTCALL call_9_0_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_6_10
#define init_9_0_6_10
void FASTCALL call_9_0_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_6_11
#define init_9_0_6_11
void FASTCALL call_9_0_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_6_12
#define init_9_0_6_12
void FASTCALL call_9_0_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_6_13
#define init_9_0_6_13
void FASTCALL call_9_0_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_6_14
#define init_9_0_6_14
void FASTCALL call_9_0_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_6_15
#define init_9_0_6_15
void FASTCALL call_9_0_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_7_0
#define init_9_0_7_0
void FASTCALL call_9_0_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_7_1
#define init_9_0_7_1
void FASTCALL call_9_0_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_7_2
#define init_9_0_7_2
void FASTCALL call_9_0_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_7_3
#define init_9_0_7_3
void FASTCALL call_9_0_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_7_4
#define init_9_0_7_4
void FASTCALL call_9_0_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_7_5
#define init_9_0_7_5
void FASTCALL call_9_0_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_7_6
#define init_9_0_7_6
void FASTCALL call_9_0_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_7_7
#define init_9_0_7_7
void FASTCALL call_9_0_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_7_8
#define init_9_0_7_8
void FASTCALL call_9_0_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_7_9
#define init_9_0_7_9
void FASTCALL call_9_0_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_7_10
#define init_9_0_7_10
void FASTCALL call_9_0_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_7_11
#define init_9_0_7_11
void FASTCALL call_9_0_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_7_12
#define init_9_0_7_12
void FASTCALL call_9_0_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_7_13
#define init_9_0_7_13
void FASTCALL call_9_0_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_7_14
#define init_9_0_7_14
void FASTCALL call_9_0_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_7_15
#define init_9_0_7_15
void FASTCALL call_9_0_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_8_0
#define init_9_0_8_0
void FASTCALL call_9_0_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_8_1
#define init_9_0_8_1
void FASTCALL call_9_0_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_8_2
#define init_9_0_8_2
void FASTCALL call_9_0_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_8_3
#define init_9_0_8_3
void FASTCALL call_9_0_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_8_4
#define init_9_0_8_4
void FASTCALL call_9_0_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_8_5
#define init_9_0_8_5
void FASTCALL call_9_0_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_8_6
#define init_9_0_8_6
void FASTCALL call_9_0_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_8_7
#define init_9_0_8_7
void FASTCALL call_9_0_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_8_8
#define init_9_0_8_8
void FASTCALL call_9_0_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_8_9
#define init_9_0_8_9
void FASTCALL call_9_0_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_8_10
#define init_9_0_8_10
void FASTCALL call_9_0_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_8_11
#define init_9_0_8_11
void FASTCALL call_9_0_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_8_12
#define init_9_0_8_12
void FASTCALL call_9_0_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_8_13
#define init_9_0_8_13
void FASTCALL call_9_0_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_8_14
#define init_9_0_8_14
void FASTCALL call_9_0_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_8_15
#define init_9_0_8_15
void FASTCALL call_9_0_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_9_0
#define init_9_0_9_0
void FASTCALL call_9_0_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_9_1
#define init_9_0_9_1
void FASTCALL call_9_0_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_9_2
#define init_9_0_9_2
void FASTCALL call_9_0_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_9_3
#define init_9_0_9_3
void FASTCALL call_9_0_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_9_4
#define init_9_0_9_4
void FASTCALL call_9_0_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_9_5
#define init_9_0_9_5
void FASTCALL call_9_0_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_9_6
#define init_9_0_9_6
void FASTCALL call_9_0_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_9_7
#define init_9_0_9_7
void FASTCALL call_9_0_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_9_8
#define init_9_0_9_8
void FASTCALL call_9_0_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_9_9
#define init_9_0_9_9
void FASTCALL call_9_0_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_9_10
#define init_9_0_9_10
void FASTCALL call_9_0_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_9_11
#define init_9_0_9_11
void FASTCALL call_9_0_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_9_12
#define init_9_0_9_12
void FASTCALL call_9_0_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_9_13
#define init_9_0_9_13
void FASTCALL call_9_0_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_9_14
#define init_9_0_9_14
void FASTCALL call_9_0_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_9_15
#define init_9_0_9_15
void FASTCALL call_9_0_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_10_0
#define init_9_0_10_0
void FASTCALL call_9_0_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_10_1
#define init_9_0_10_1
void FASTCALL call_9_0_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_10_2
#define init_9_0_10_2
void FASTCALL call_9_0_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_10_3
#define init_9_0_10_3
void FASTCALL call_9_0_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_10_4
#define init_9_0_10_4
void FASTCALL call_9_0_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_10_5
#define init_9_0_10_5
void FASTCALL call_9_0_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_10_6
#define init_9_0_10_6
void FASTCALL call_9_0_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_10_7
#define init_9_0_10_7
void FASTCALL call_9_0_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_10_8
#define init_9_0_10_8
void FASTCALL call_9_0_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_10_9
#define init_9_0_10_9
void FASTCALL call_9_0_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_10_10
#define init_9_0_10_10
void FASTCALL call_9_0_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_10_11
#define init_9_0_10_11
void FASTCALL call_9_0_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_10_12
#define init_9_0_10_12
void FASTCALL call_9_0_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_10_13
#define init_9_0_10_13
void FASTCALL call_9_0_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_10_14
#define init_9_0_10_14
void FASTCALL call_9_0_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_10_15
#define init_9_0_10_15
void FASTCALL call_9_0_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_11_0
#define init_9_0_11_0
void FASTCALL call_9_0_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_11_1
#define init_9_0_11_1
void FASTCALL call_9_0_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_11_2
#define init_9_0_11_2
void FASTCALL call_9_0_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_11_3
#define init_9_0_11_3
void FASTCALL call_9_0_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_11_4
#define init_9_0_11_4
void FASTCALL call_9_0_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_11_5
#define init_9_0_11_5
void FASTCALL call_9_0_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_11_6
#define init_9_0_11_6
void FASTCALL call_9_0_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_11_7
#define init_9_0_11_7
void FASTCALL call_9_0_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_11_8
#define init_9_0_11_8
void FASTCALL call_9_0_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_11_9
#define init_9_0_11_9
void FASTCALL call_9_0_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_11_10
#define init_9_0_11_10
void FASTCALL call_9_0_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_11_11
#define init_9_0_11_11
void FASTCALL call_9_0_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_11_12
#define init_9_0_11_12
void FASTCALL call_9_0_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_11_13
#define init_9_0_11_13
void FASTCALL call_9_0_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_11_14
#define init_9_0_11_14
void FASTCALL call_9_0_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_11_15
#define init_9_0_11_15
void FASTCALL call_9_0_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_12_0
#define init_9_0_12_0
void FASTCALL call_9_0_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_12_1
#define init_9_0_12_1
void FASTCALL call_9_0_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_12_2
#define init_9_0_12_2
void FASTCALL call_9_0_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_12_3
#define init_9_0_12_3
void FASTCALL call_9_0_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_12_4
#define init_9_0_12_4
void FASTCALL call_9_0_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_12_5
#define init_9_0_12_5
void FASTCALL call_9_0_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_12_6
#define init_9_0_12_6
void FASTCALL call_9_0_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_12_7
#define init_9_0_12_7
void FASTCALL call_9_0_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_12_8
#define init_9_0_12_8
void FASTCALL call_9_0_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_12_9
#define init_9_0_12_9
void FASTCALL call_9_0_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_12_10
#define init_9_0_12_10
void FASTCALL call_9_0_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_12_11
#define init_9_0_12_11
void FASTCALL call_9_0_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_12_12
#define init_9_0_12_12
void FASTCALL call_9_0_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_12_13
#define init_9_0_12_13
void FASTCALL call_9_0_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_12_14
#define init_9_0_12_14
void FASTCALL call_9_0_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_12_15
#define init_9_0_12_15
void FASTCALL call_9_0_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_13_0
#define init_9_0_13_0
void FASTCALL call_9_0_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_13_1
#define init_9_0_13_1
void FASTCALL call_9_0_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_13_2
#define init_9_0_13_2
void FASTCALL call_9_0_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_13_3
#define init_9_0_13_3
void FASTCALL call_9_0_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_13_4
#define init_9_0_13_4
void FASTCALL call_9_0_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_13_5
#define init_9_0_13_5
void FASTCALL call_9_0_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_13_6
#define init_9_0_13_6
void FASTCALL call_9_0_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_13_7
#define init_9_0_13_7
void FASTCALL call_9_0_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_13_8
#define init_9_0_13_8
void FASTCALL call_9_0_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_13_9
#define init_9_0_13_9
void FASTCALL call_9_0_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_13_10
#define init_9_0_13_10
void FASTCALL call_9_0_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_13_11
#define init_9_0_13_11
void FASTCALL call_9_0_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_13_12
#define init_9_0_13_12
void FASTCALL call_9_0_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_13_13
#define init_9_0_13_13
void FASTCALL call_9_0_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_13_14
#define init_9_0_13_14
void FASTCALL call_9_0_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_13_15
#define init_9_0_13_15
void FASTCALL call_9_0_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_14_0
#define init_9_0_14_0
void FASTCALL call_9_0_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_14_1
#define init_9_0_14_1
void FASTCALL call_9_0_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_14_2
#define init_9_0_14_2
void FASTCALL call_9_0_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_14_3
#define init_9_0_14_3
void FASTCALL call_9_0_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_14_4
#define init_9_0_14_4
void FASTCALL call_9_0_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_14_5
#define init_9_0_14_5
void FASTCALL call_9_0_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_14_6
#define init_9_0_14_6
void FASTCALL call_9_0_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_14_7
#define init_9_0_14_7
void FASTCALL call_9_0_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_14_8
#define init_9_0_14_8
void FASTCALL call_9_0_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_14_9
#define init_9_0_14_9
void FASTCALL call_9_0_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_14_10
#define init_9_0_14_10
void FASTCALL call_9_0_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_14_11
#define init_9_0_14_11
void FASTCALL call_9_0_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_14_12
#define init_9_0_14_12
void FASTCALL call_9_0_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_14_13
#define init_9_0_14_13
void FASTCALL call_9_0_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_14_14
#define init_9_0_14_14
void FASTCALL call_9_0_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_14_15
#define init_9_0_14_15
void FASTCALL call_9_0_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_15_0
#define init_9_0_15_0
void FASTCALL call_9_0_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_15_1
#define init_9_0_15_1
void FASTCALL call_9_0_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_15_2
#define init_9_0_15_2
void FASTCALL call_9_0_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_15_3
#define init_9_0_15_3
void FASTCALL call_9_0_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_15_4
#define init_9_0_15_4
void FASTCALL call_9_0_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_15_5
#define init_9_0_15_5
void FASTCALL call_9_0_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_15_6
#define init_9_0_15_6
void FASTCALL call_9_0_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_15_7
#define init_9_0_15_7
void FASTCALL call_9_0_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_15_8
#define init_9_0_15_8
void FASTCALL call_9_0_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_15_9
#define init_9_0_15_9
void FASTCALL call_9_0_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_15_10
#define init_9_0_15_10
void FASTCALL call_9_0_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_15_11
#define init_9_0_15_11
void FASTCALL call_9_0_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_15_12
#define init_9_0_15_12
void FASTCALL call_9_0_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_15_13
#define init_9_0_15_13
void FASTCALL call_9_0_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_15_14
#define init_9_0_15_14
void FASTCALL call_9_0_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_0_15_15
#define init_9_0_15_15
void FASTCALL call_9_0_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_0_0
#define init_9_1_0_0
void FASTCALL call_9_1_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_0_1
#define init_9_1_0_1
void FASTCALL call_9_1_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_0_2
#define init_9_1_0_2
void FASTCALL call_9_1_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_0_3
#define init_9_1_0_3
void FASTCALL call_9_1_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_0_4
#define init_9_1_0_4
void FASTCALL call_9_1_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_0_5
#define init_9_1_0_5
void FASTCALL call_9_1_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_0_6
#define init_9_1_0_6
void FASTCALL call_9_1_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_0_7
#define init_9_1_0_7
void FASTCALL call_9_1_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_0_8
#define init_9_1_0_8
void FASTCALL call_9_1_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_0_9
#define init_9_1_0_9
void FASTCALL call_9_1_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_0_10
#define init_9_1_0_10
void FASTCALL call_9_1_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_0_11
#define init_9_1_0_11
void FASTCALL call_9_1_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_0_12
#define init_9_1_0_12
void FASTCALL call_9_1_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_0_13
#define init_9_1_0_13
void FASTCALL call_9_1_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_0_14
#define init_9_1_0_14
void FASTCALL call_9_1_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_0_15
#define init_9_1_0_15
void FASTCALL call_9_1_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_1_0
#define init_9_1_1_0
void FASTCALL call_9_1_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_1_1
#define init_9_1_1_1
void FASTCALL call_9_1_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_1_2
#define init_9_1_1_2
void FASTCALL call_9_1_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_1_3
#define init_9_1_1_3
void FASTCALL call_9_1_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_1_4
#define init_9_1_1_4
void FASTCALL call_9_1_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_1_5
#define init_9_1_1_5
void FASTCALL call_9_1_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_1_6
#define init_9_1_1_6
void FASTCALL call_9_1_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_1_7
#define init_9_1_1_7
void FASTCALL call_9_1_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_1_8
#define init_9_1_1_8
void FASTCALL call_9_1_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_1_9
#define init_9_1_1_9
void FASTCALL call_9_1_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_1_10
#define init_9_1_1_10
void FASTCALL call_9_1_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_1_11
#define init_9_1_1_11
void FASTCALL call_9_1_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_1_12
#define init_9_1_1_12
void FASTCALL call_9_1_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_1_13
#define init_9_1_1_13
void FASTCALL call_9_1_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_1_14
#define init_9_1_1_14
void FASTCALL call_9_1_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_1_15
#define init_9_1_1_15
void FASTCALL call_9_1_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_2_0
#define init_9_1_2_0
void FASTCALL call_9_1_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_2_1
#define init_9_1_2_1
void FASTCALL call_9_1_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_2_2
#define init_9_1_2_2
void FASTCALL call_9_1_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_2_3
#define init_9_1_2_3
void FASTCALL call_9_1_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_2_4
#define init_9_1_2_4
void FASTCALL call_9_1_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_2_5
#define init_9_1_2_5
void FASTCALL call_9_1_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_2_6
#define init_9_1_2_6
void FASTCALL call_9_1_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_2_7
#define init_9_1_2_7
void FASTCALL call_9_1_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_2_8
#define init_9_1_2_8
void FASTCALL call_9_1_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_2_9
#define init_9_1_2_9
void FASTCALL call_9_1_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_2_10
#define init_9_1_2_10
void FASTCALL call_9_1_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_2_11
#define init_9_1_2_11
void FASTCALL call_9_1_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_2_12
#define init_9_1_2_12
void FASTCALL call_9_1_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_2_13
#define init_9_1_2_13
void FASTCALL call_9_1_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_2_14
#define init_9_1_2_14
void FASTCALL call_9_1_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_2_15
#define init_9_1_2_15
void FASTCALL call_9_1_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_3_0
#define init_9_1_3_0
void FASTCALL call_9_1_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_3_1
#define init_9_1_3_1
void FASTCALL call_9_1_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_3_2
#define init_9_1_3_2
void FASTCALL call_9_1_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_3_3
#define init_9_1_3_3
void FASTCALL call_9_1_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_3_4
#define init_9_1_3_4
void FASTCALL call_9_1_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_3_5
#define init_9_1_3_5
void FASTCALL call_9_1_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_3_6
#define init_9_1_3_6
void FASTCALL call_9_1_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_3_7
#define init_9_1_3_7
void FASTCALL call_9_1_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_3_8
#define init_9_1_3_8
void FASTCALL call_9_1_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_3_9
#define init_9_1_3_9
void FASTCALL call_9_1_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_3_10
#define init_9_1_3_10
void FASTCALL call_9_1_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_3_11
#define init_9_1_3_11
void FASTCALL call_9_1_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_3_12
#define init_9_1_3_12
void FASTCALL call_9_1_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_3_13
#define init_9_1_3_13
void FASTCALL call_9_1_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_3_14
#define init_9_1_3_14
void FASTCALL call_9_1_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_3_15
#define init_9_1_3_15
void FASTCALL call_9_1_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_4_0
#define init_9_1_4_0
void FASTCALL call_9_1_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_4_1
#define init_9_1_4_1
void FASTCALL call_9_1_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_4_2
#define init_9_1_4_2
void FASTCALL call_9_1_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_4_3
#define init_9_1_4_3
void FASTCALL call_9_1_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_4_4
#define init_9_1_4_4
void FASTCALL call_9_1_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_4_5
#define init_9_1_4_5
void FASTCALL call_9_1_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_4_6
#define init_9_1_4_6
void FASTCALL call_9_1_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_4_7
#define init_9_1_4_7
void FASTCALL call_9_1_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_4_8
#define init_9_1_4_8
void FASTCALL call_9_1_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_4_9
#define init_9_1_4_9
void FASTCALL call_9_1_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_4_10
#define init_9_1_4_10
void FASTCALL call_9_1_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_4_11
#define init_9_1_4_11
void FASTCALL call_9_1_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_4_12
#define init_9_1_4_12
void FASTCALL call_9_1_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_4_13
#define init_9_1_4_13
void FASTCALL call_9_1_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_4_14
#define init_9_1_4_14
void FASTCALL call_9_1_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_4_15
#define init_9_1_4_15
void FASTCALL call_9_1_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_5_0
#define init_9_1_5_0
void FASTCALL call_9_1_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_5_1
#define init_9_1_5_1
void FASTCALL call_9_1_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_5_2
#define init_9_1_5_2
void FASTCALL call_9_1_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_5_3
#define init_9_1_5_3
void FASTCALL call_9_1_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_5_4
#define init_9_1_5_4
void FASTCALL call_9_1_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_5_5
#define init_9_1_5_5
void FASTCALL call_9_1_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_5_6
#define init_9_1_5_6
void FASTCALL call_9_1_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_5_7
#define init_9_1_5_7
void FASTCALL call_9_1_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_5_8
#define init_9_1_5_8
void FASTCALL call_9_1_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_5_9
#define init_9_1_5_9
void FASTCALL call_9_1_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_5_10
#define init_9_1_5_10
void FASTCALL call_9_1_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_5_11
#define init_9_1_5_11
void FASTCALL call_9_1_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_5_12
#define init_9_1_5_12
void FASTCALL call_9_1_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_5_13
#define init_9_1_5_13
void FASTCALL call_9_1_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_5_14
#define init_9_1_5_14
void FASTCALL call_9_1_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_5_15
#define init_9_1_5_15
void FASTCALL call_9_1_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_6_0
#define init_9_1_6_0
void FASTCALL call_9_1_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_6_1
#define init_9_1_6_1
void FASTCALL call_9_1_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_6_2
#define init_9_1_6_2
void FASTCALL call_9_1_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_6_3
#define init_9_1_6_3
void FASTCALL call_9_1_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_6_4
#define init_9_1_6_4
void FASTCALL call_9_1_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_6_5
#define init_9_1_6_5
void FASTCALL call_9_1_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_6_6
#define init_9_1_6_6
void FASTCALL call_9_1_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_6_7
#define init_9_1_6_7
void FASTCALL call_9_1_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_6_8
#define init_9_1_6_8
void FASTCALL call_9_1_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_6_9
#define init_9_1_6_9
void FASTCALL call_9_1_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_6_10
#define init_9_1_6_10
void FASTCALL call_9_1_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_6_11
#define init_9_1_6_11
void FASTCALL call_9_1_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_6_12
#define init_9_1_6_12
void FASTCALL call_9_1_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_6_13
#define init_9_1_6_13
void FASTCALL call_9_1_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_6_14
#define init_9_1_6_14
void FASTCALL call_9_1_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_6_15
#define init_9_1_6_15
void FASTCALL call_9_1_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_7_0
#define init_9_1_7_0
void FASTCALL call_9_1_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_7_1
#define init_9_1_7_1
void FASTCALL call_9_1_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_7_2
#define init_9_1_7_2
void FASTCALL call_9_1_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_7_3
#define init_9_1_7_3
void FASTCALL call_9_1_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_7_4
#define init_9_1_7_4
void FASTCALL call_9_1_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_7_5
#define init_9_1_7_5
void FASTCALL call_9_1_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_7_6
#define init_9_1_7_6
void FASTCALL call_9_1_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_7_7
#define init_9_1_7_7
void FASTCALL call_9_1_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_7_8
#define init_9_1_7_8
void FASTCALL call_9_1_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_7_9
#define init_9_1_7_9
void FASTCALL call_9_1_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_7_10
#define init_9_1_7_10
void FASTCALL call_9_1_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_7_11
#define init_9_1_7_11
void FASTCALL call_9_1_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_7_12
#define init_9_1_7_12
void FASTCALL call_9_1_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_7_13
#define init_9_1_7_13
void FASTCALL call_9_1_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_7_14
#define init_9_1_7_14
void FASTCALL call_9_1_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_7_15
#define init_9_1_7_15
void FASTCALL call_9_1_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_8_0
#define init_9_1_8_0
void FASTCALL call_9_1_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_8_1
#define init_9_1_8_1
void FASTCALL call_9_1_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_8_2
#define init_9_1_8_2
void FASTCALL call_9_1_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_8_3
#define init_9_1_8_3
void FASTCALL call_9_1_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_8_4
#define init_9_1_8_4
void FASTCALL call_9_1_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_8_5
#define init_9_1_8_5
void FASTCALL call_9_1_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_8_6
#define init_9_1_8_6
void FASTCALL call_9_1_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_8_7
#define init_9_1_8_7
void FASTCALL call_9_1_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_8_8
#define init_9_1_8_8
void FASTCALL call_9_1_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_8_9
#define init_9_1_8_9
void FASTCALL call_9_1_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_8_10
#define init_9_1_8_10
void FASTCALL call_9_1_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_8_11
#define init_9_1_8_11
void FASTCALL call_9_1_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_8_12
#define init_9_1_8_12
void FASTCALL call_9_1_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_8_13
#define init_9_1_8_13
void FASTCALL call_9_1_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_8_14
#define init_9_1_8_14
void FASTCALL call_9_1_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_8_15
#define init_9_1_8_15
void FASTCALL call_9_1_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_9_0
#define init_9_1_9_0
void FASTCALL call_9_1_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_9_1
#define init_9_1_9_1
void FASTCALL call_9_1_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_9_2
#define init_9_1_9_2
void FASTCALL call_9_1_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_9_3
#define init_9_1_9_3
void FASTCALL call_9_1_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_9_4
#define init_9_1_9_4
void FASTCALL call_9_1_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_9_5
#define init_9_1_9_5
void FASTCALL call_9_1_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_9_6
#define init_9_1_9_6
void FASTCALL call_9_1_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_9_7
#define init_9_1_9_7
void FASTCALL call_9_1_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_9_8
#define init_9_1_9_8
void FASTCALL call_9_1_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_9_9
#define init_9_1_9_9
void FASTCALL call_9_1_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_9_10
#define init_9_1_9_10
void FASTCALL call_9_1_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_9_11
#define init_9_1_9_11
void FASTCALL call_9_1_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_9_12
#define init_9_1_9_12
void FASTCALL call_9_1_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_9_13
#define init_9_1_9_13
void FASTCALL call_9_1_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_9_14
#define init_9_1_9_14
void FASTCALL call_9_1_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_9_15
#define init_9_1_9_15
void FASTCALL call_9_1_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_10_0
#define init_9_1_10_0
void FASTCALL call_9_1_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_10_1
#define init_9_1_10_1
void FASTCALL call_9_1_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_10_2
#define init_9_1_10_2
void FASTCALL call_9_1_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_10_3
#define init_9_1_10_3
void FASTCALL call_9_1_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_10_4
#define init_9_1_10_4
void FASTCALL call_9_1_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_10_5
#define init_9_1_10_5
void FASTCALL call_9_1_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_10_6
#define init_9_1_10_6
void FASTCALL call_9_1_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_10_7
#define init_9_1_10_7
void FASTCALL call_9_1_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_10_8
#define init_9_1_10_8
void FASTCALL call_9_1_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_10_9
#define init_9_1_10_9
void FASTCALL call_9_1_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_10_10
#define init_9_1_10_10
void FASTCALL call_9_1_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_10_11
#define init_9_1_10_11
void FASTCALL call_9_1_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_10_12
#define init_9_1_10_12
void FASTCALL call_9_1_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_10_13
#define init_9_1_10_13
void FASTCALL call_9_1_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_10_14
#define init_9_1_10_14
void FASTCALL call_9_1_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_10_15
#define init_9_1_10_15
void FASTCALL call_9_1_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_11_0
#define init_9_1_11_0
void FASTCALL call_9_1_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_11_1
#define init_9_1_11_1
void FASTCALL call_9_1_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_11_2
#define init_9_1_11_2
void FASTCALL call_9_1_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_11_3
#define init_9_1_11_3
void FASTCALL call_9_1_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_11_4
#define init_9_1_11_4
void FASTCALL call_9_1_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_11_5
#define init_9_1_11_5
void FASTCALL call_9_1_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_11_6
#define init_9_1_11_6
void FASTCALL call_9_1_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_11_7
#define init_9_1_11_7
void FASTCALL call_9_1_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_11_8
#define init_9_1_11_8
void FASTCALL call_9_1_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_11_9
#define init_9_1_11_9
void FASTCALL call_9_1_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_11_10
#define init_9_1_11_10
void FASTCALL call_9_1_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_11_11
#define init_9_1_11_11
void FASTCALL call_9_1_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_11_12
#define init_9_1_11_12
void FASTCALL call_9_1_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_11_13
#define init_9_1_11_13
void FASTCALL call_9_1_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_11_14
#define init_9_1_11_14
void FASTCALL call_9_1_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_11_15
#define init_9_1_11_15
void FASTCALL call_9_1_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_12_0
#define init_9_1_12_0
void FASTCALL call_9_1_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_12_1
#define init_9_1_12_1
void FASTCALL call_9_1_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_12_2
#define init_9_1_12_2
void FASTCALL call_9_1_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_12_3
#define init_9_1_12_3
void FASTCALL call_9_1_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_12_4
#define init_9_1_12_4
void FASTCALL call_9_1_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_12_5
#define init_9_1_12_5
void FASTCALL call_9_1_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_12_6
#define init_9_1_12_6
void FASTCALL call_9_1_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_12_7
#define init_9_1_12_7
void FASTCALL call_9_1_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_12_8
#define init_9_1_12_8
void FASTCALL call_9_1_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_12_9
#define init_9_1_12_9
void FASTCALL call_9_1_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_12_10
#define init_9_1_12_10
void FASTCALL call_9_1_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_12_11
#define init_9_1_12_11
void FASTCALL call_9_1_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_12_12
#define init_9_1_12_12
void FASTCALL call_9_1_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_12_13
#define init_9_1_12_13
void FASTCALL call_9_1_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_12_14
#define init_9_1_12_14
void FASTCALL call_9_1_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_12_15
#define init_9_1_12_15
void FASTCALL call_9_1_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_13_0
#define init_9_1_13_0
void FASTCALL call_9_1_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_13_1
#define init_9_1_13_1
void FASTCALL call_9_1_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_13_2
#define init_9_1_13_2
void FASTCALL call_9_1_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_13_3
#define init_9_1_13_3
void FASTCALL call_9_1_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_13_4
#define init_9_1_13_4
void FASTCALL call_9_1_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_13_5
#define init_9_1_13_5
void FASTCALL call_9_1_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_13_6
#define init_9_1_13_6
void FASTCALL call_9_1_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_13_7
#define init_9_1_13_7
void FASTCALL call_9_1_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_13_8
#define init_9_1_13_8
void FASTCALL call_9_1_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_13_9
#define init_9_1_13_9
void FASTCALL call_9_1_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_13_10
#define init_9_1_13_10
void FASTCALL call_9_1_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_13_11
#define init_9_1_13_11
void FASTCALL call_9_1_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_13_12
#define init_9_1_13_12
void FASTCALL call_9_1_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_13_13
#define init_9_1_13_13
void FASTCALL call_9_1_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_13_14
#define init_9_1_13_14
void FASTCALL call_9_1_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_13_15
#define init_9_1_13_15
void FASTCALL call_9_1_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_14_0
#define init_9_1_14_0
void FASTCALL call_9_1_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_14_1
#define init_9_1_14_1
void FASTCALL call_9_1_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_14_2
#define init_9_1_14_2
void FASTCALL call_9_1_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_14_3
#define init_9_1_14_3
void FASTCALL call_9_1_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_14_4
#define init_9_1_14_4
void FASTCALL call_9_1_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_14_5
#define init_9_1_14_5
void FASTCALL call_9_1_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_14_6
#define init_9_1_14_6
void FASTCALL call_9_1_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_14_7
#define init_9_1_14_7
void FASTCALL call_9_1_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_14_8
#define init_9_1_14_8
void FASTCALL call_9_1_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_14_9
#define init_9_1_14_9
void FASTCALL call_9_1_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_14_10
#define init_9_1_14_10
void FASTCALL call_9_1_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_14_11
#define init_9_1_14_11
void FASTCALL call_9_1_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_14_12
#define init_9_1_14_12
void FASTCALL call_9_1_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_14_13
#define init_9_1_14_13
void FASTCALL call_9_1_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_14_14
#define init_9_1_14_14
void FASTCALL call_9_1_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_14_15
#define init_9_1_14_15
void FASTCALL call_9_1_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_15_0
#define init_9_1_15_0
void FASTCALL call_9_1_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_15_1
#define init_9_1_15_1
void FASTCALL call_9_1_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_15_2
#define init_9_1_15_2
void FASTCALL call_9_1_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_15_3
#define init_9_1_15_3
void FASTCALL call_9_1_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_15_4
#define init_9_1_15_4
void FASTCALL call_9_1_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_15_5
#define init_9_1_15_5
void FASTCALL call_9_1_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_15_6
#define init_9_1_15_6
void FASTCALL call_9_1_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_15_7
#define init_9_1_15_7
void FASTCALL call_9_1_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_15_8
#define init_9_1_15_8
void FASTCALL call_9_1_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_15_9
#define init_9_1_15_9
void FASTCALL call_9_1_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_15_10
#define init_9_1_15_10
void FASTCALL call_9_1_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_15_11
#define init_9_1_15_11
void FASTCALL call_9_1_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_15_12
#define init_9_1_15_12
void FASTCALL call_9_1_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_15_13
#define init_9_1_15_13
void FASTCALL call_9_1_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_15_14
#define init_9_1_15_14
void FASTCALL call_9_1_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_1_15_15
#define init_9_1_15_15
void FASTCALL call_9_1_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_0_0
#define init_9_2_0_0
void FASTCALL call_9_2_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_0_1
#define init_9_2_0_1
void FASTCALL call_9_2_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_0_2
#define init_9_2_0_2
void FASTCALL call_9_2_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_0_3
#define init_9_2_0_3
void FASTCALL call_9_2_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_0_4
#define init_9_2_0_4
void FASTCALL call_9_2_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_0_5
#define init_9_2_0_5
void FASTCALL call_9_2_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_0_6
#define init_9_2_0_6
void FASTCALL call_9_2_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_0_7
#define init_9_2_0_7
void FASTCALL call_9_2_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_0_8
#define init_9_2_0_8
void FASTCALL call_9_2_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_0_9
#define init_9_2_0_9
void FASTCALL call_9_2_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_0_10
#define init_9_2_0_10
void FASTCALL call_9_2_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_0_11
#define init_9_2_0_11
void FASTCALL call_9_2_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_0_12
#define init_9_2_0_12
void FASTCALL call_9_2_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_0_13
#define init_9_2_0_13
void FASTCALL call_9_2_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_0_14
#define init_9_2_0_14
void FASTCALL call_9_2_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_0_15
#define init_9_2_0_15
void FASTCALL call_9_2_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_1_0
#define init_9_2_1_0
void FASTCALL call_9_2_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_1_1
#define init_9_2_1_1
void FASTCALL call_9_2_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_1_2
#define init_9_2_1_2
void FASTCALL call_9_2_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_1_3
#define init_9_2_1_3
void FASTCALL call_9_2_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_1_4
#define init_9_2_1_4
void FASTCALL call_9_2_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_1_5
#define init_9_2_1_5
void FASTCALL call_9_2_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_1_6
#define init_9_2_1_6
void FASTCALL call_9_2_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_1_7
#define init_9_2_1_7
void FASTCALL call_9_2_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_1_8
#define init_9_2_1_8
void FASTCALL call_9_2_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_1_9
#define init_9_2_1_9
void FASTCALL call_9_2_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_1_10
#define init_9_2_1_10
void FASTCALL call_9_2_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_1_11
#define init_9_2_1_11
void FASTCALL call_9_2_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_1_12
#define init_9_2_1_12
void FASTCALL call_9_2_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_1_13
#define init_9_2_1_13
void FASTCALL call_9_2_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_1_14
#define init_9_2_1_14
void FASTCALL call_9_2_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_1_15
#define init_9_2_1_15
void FASTCALL call_9_2_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_2_0
#define init_9_2_2_0
void FASTCALL call_9_2_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_2_1
#define init_9_2_2_1
void FASTCALL call_9_2_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_2_2
#define init_9_2_2_2
void FASTCALL call_9_2_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_2_3
#define init_9_2_2_3
void FASTCALL call_9_2_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_2_4
#define init_9_2_2_4
void FASTCALL call_9_2_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_2_5
#define init_9_2_2_5
void FASTCALL call_9_2_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_2_6
#define init_9_2_2_6
void FASTCALL call_9_2_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_2_7
#define init_9_2_2_7
void FASTCALL call_9_2_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_2_8
#define init_9_2_2_8
void FASTCALL call_9_2_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_2_9
#define init_9_2_2_9
void FASTCALL call_9_2_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_2_10
#define init_9_2_2_10
void FASTCALL call_9_2_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_2_11
#define init_9_2_2_11
void FASTCALL call_9_2_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_2_12
#define init_9_2_2_12
void FASTCALL call_9_2_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_2_13
#define init_9_2_2_13
void FASTCALL call_9_2_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_2_14
#define init_9_2_2_14
void FASTCALL call_9_2_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_2_15
#define init_9_2_2_15
void FASTCALL call_9_2_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_3_0
#define init_9_2_3_0
void FASTCALL call_9_2_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_3_1
#define init_9_2_3_1
void FASTCALL call_9_2_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_3_2
#define init_9_2_3_2
void FASTCALL call_9_2_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_3_3
#define init_9_2_3_3
void FASTCALL call_9_2_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_3_4
#define init_9_2_3_4
void FASTCALL call_9_2_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_3_5
#define init_9_2_3_5
void FASTCALL call_9_2_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_3_6
#define init_9_2_3_6
void FASTCALL call_9_2_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_3_7
#define init_9_2_3_7
void FASTCALL call_9_2_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_3_8
#define init_9_2_3_8
void FASTCALL call_9_2_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_3_9
#define init_9_2_3_9
void FASTCALL call_9_2_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_3_10
#define init_9_2_3_10
void FASTCALL call_9_2_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_3_11
#define init_9_2_3_11
void FASTCALL call_9_2_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_3_12
#define init_9_2_3_12
void FASTCALL call_9_2_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_3_13
#define init_9_2_3_13
void FASTCALL call_9_2_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_3_14
#define init_9_2_3_14
void FASTCALL call_9_2_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_3_15
#define init_9_2_3_15
void FASTCALL call_9_2_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_4_0
#define init_9_2_4_0
void FASTCALL call_9_2_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_4_1
#define init_9_2_4_1
void FASTCALL call_9_2_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_4_2
#define init_9_2_4_2
void FASTCALL call_9_2_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_4_3
#define init_9_2_4_3
void FASTCALL call_9_2_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_4_4
#define init_9_2_4_4
void FASTCALL call_9_2_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_4_5
#define init_9_2_4_5
void FASTCALL call_9_2_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_4_6
#define init_9_2_4_6
void FASTCALL call_9_2_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_4_7
#define init_9_2_4_7
void FASTCALL call_9_2_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_4_8
#define init_9_2_4_8
void FASTCALL call_9_2_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_4_9
#define init_9_2_4_9
void FASTCALL call_9_2_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_4_10
#define init_9_2_4_10
void FASTCALL call_9_2_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_4_11
#define init_9_2_4_11
void FASTCALL call_9_2_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_4_12
#define init_9_2_4_12
void FASTCALL call_9_2_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_4_13
#define init_9_2_4_13
void FASTCALL call_9_2_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_4_14
#define init_9_2_4_14
void FASTCALL call_9_2_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_4_15
#define init_9_2_4_15
void FASTCALL call_9_2_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_5_0
#define init_9_2_5_0
void FASTCALL call_9_2_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_5_1
#define init_9_2_5_1
void FASTCALL call_9_2_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_5_2
#define init_9_2_5_2
void FASTCALL call_9_2_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_5_3
#define init_9_2_5_3
void FASTCALL call_9_2_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_5_4
#define init_9_2_5_4
void FASTCALL call_9_2_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_5_5
#define init_9_2_5_5
void FASTCALL call_9_2_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_5_6
#define init_9_2_5_6
void FASTCALL call_9_2_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_5_7
#define init_9_2_5_7
void FASTCALL call_9_2_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_5_8
#define init_9_2_5_8
void FASTCALL call_9_2_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_5_9
#define init_9_2_5_9
void FASTCALL call_9_2_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_5_10
#define init_9_2_5_10
void FASTCALL call_9_2_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_5_11
#define init_9_2_5_11
void FASTCALL call_9_2_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_5_12
#define init_9_2_5_12
void FASTCALL call_9_2_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_5_13
#define init_9_2_5_13
void FASTCALL call_9_2_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_5_14
#define init_9_2_5_14
void FASTCALL call_9_2_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_5_15
#define init_9_2_5_15
void FASTCALL call_9_2_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_6_0
#define init_9_2_6_0
void FASTCALL call_9_2_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_6_1
#define init_9_2_6_1
void FASTCALL call_9_2_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_6_2
#define init_9_2_6_2
void FASTCALL call_9_2_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_6_3
#define init_9_2_6_3
void FASTCALL call_9_2_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_6_4
#define init_9_2_6_4
void FASTCALL call_9_2_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_6_5
#define init_9_2_6_5
void FASTCALL call_9_2_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_6_6
#define init_9_2_6_6
void FASTCALL call_9_2_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_6_7
#define init_9_2_6_7
void FASTCALL call_9_2_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_6_8
#define init_9_2_6_8
void FASTCALL call_9_2_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_6_9
#define init_9_2_6_9
void FASTCALL call_9_2_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_6_10
#define init_9_2_6_10
void FASTCALL call_9_2_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_6_11
#define init_9_2_6_11
void FASTCALL call_9_2_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_6_12
#define init_9_2_6_12
void FASTCALL call_9_2_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_6_13
#define init_9_2_6_13
void FASTCALL call_9_2_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_6_14
#define init_9_2_6_14
void FASTCALL call_9_2_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_6_15
#define init_9_2_6_15
void FASTCALL call_9_2_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_7_0
#define init_9_2_7_0
void FASTCALL call_9_2_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_7_1
#define init_9_2_7_1
void FASTCALL call_9_2_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_7_2
#define init_9_2_7_2
void FASTCALL call_9_2_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_7_3
#define init_9_2_7_3
void FASTCALL call_9_2_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_7_4
#define init_9_2_7_4
void FASTCALL call_9_2_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_7_5
#define init_9_2_7_5
void FASTCALL call_9_2_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_7_6
#define init_9_2_7_6
void FASTCALL call_9_2_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_7_7
#define init_9_2_7_7
void FASTCALL call_9_2_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_7_8
#define init_9_2_7_8
void FASTCALL call_9_2_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_7_9
#define init_9_2_7_9
void FASTCALL call_9_2_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_7_10
#define init_9_2_7_10
void FASTCALL call_9_2_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_7_11
#define init_9_2_7_11
void FASTCALL call_9_2_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_7_12
#define init_9_2_7_12
void FASTCALL call_9_2_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_7_13
#define init_9_2_7_13
void FASTCALL call_9_2_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_7_14
#define init_9_2_7_14
void FASTCALL call_9_2_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_7_15
#define init_9_2_7_15
void FASTCALL call_9_2_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_8_0
#define init_9_2_8_0
void FASTCALL call_9_2_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_8_1
#define init_9_2_8_1
void FASTCALL call_9_2_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_8_2
#define init_9_2_8_2
void FASTCALL call_9_2_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_8_3
#define init_9_2_8_3
void FASTCALL call_9_2_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_8_4
#define init_9_2_8_4
void FASTCALL call_9_2_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_8_5
#define init_9_2_8_5
void FASTCALL call_9_2_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_8_6
#define init_9_2_8_6
void FASTCALL call_9_2_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_8_7
#define init_9_2_8_7
void FASTCALL call_9_2_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_8_8
#define init_9_2_8_8
void FASTCALL call_9_2_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_8_9
#define init_9_2_8_9
void FASTCALL call_9_2_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_8_10
#define init_9_2_8_10
void FASTCALL call_9_2_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_8_11
#define init_9_2_8_11
void FASTCALL call_9_2_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_8_12
#define init_9_2_8_12
void FASTCALL call_9_2_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_8_13
#define init_9_2_8_13
void FASTCALL call_9_2_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_8_14
#define init_9_2_8_14
void FASTCALL call_9_2_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_8_15
#define init_9_2_8_15
void FASTCALL call_9_2_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_9_0
#define init_9_2_9_0
void FASTCALL call_9_2_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_9_1
#define init_9_2_9_1
void FASTCALL call_9_2_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_9_2
#define init_9_2_9_2
void FASTCALL call_9_2_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_9_3
#define init_9_2_9_3
void FASTCALL call_9_2_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_9_4
#define init_9_2_9_4
void FASTCALL call_9_2_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_9_5
#define init_9_2_9_5
void FASTCALL call_9_2_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_9_6
#define init_9_2_9_6
void FASTCALL call_9_2_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_9_7
#define init_9_2_9_7
void FASTCALL call_9_2_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_9_8
#define init_9_2_9_8
void FASTCALL call_9_2_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_9_9
#define init_9_2_9_9
void FASTCALL call_9_2_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_9_10
#define init_9_2_9_10
void FASTCALL call_9_2_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_9_11
#define init_9_2_9_11
void FASTCALL call_9_2_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_9_12
#define init_9_2_9_12
void FASTCALL call_9_2_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_9_13
#define init_9_2_9_13
void FASTCALL call_9_2_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_9_14
#define init_9_2_9_14
void FASTCALL call_9_2_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_9_15
#define init_9_2_9_15
void FASTCALL call_9_2_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_10_0
#define init_9_2_10_0
void FASTCALL call_9_2_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_10_1
#define init_9_2_10_1
void FASTCALL call_9_2_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_10_2
#define init_9_2_10_2
void FASTCALL call_9_2_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_10_3
#define init_9_2_10_3
void FASTCALL call_9_2_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_10_4
#define init_9_2_10_4
void FASTCALL call_9_2_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_10_5
#define init_9_2_10_5
void FASTCALL call_9_2_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_10_6
#define init_9_2_10_6
void FASTCALL call_9_2_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_10_7
#define init_9_2_10_7
void FASTCALL call_9_2_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_10_8
#define init_9_2_10_8
void FASTCALL call_9_2_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_10_9
#define init_9_2_10_9
void FASTCALL call_9_2_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_10_10
#define init_9_2_10_10
void FASTCALL call_9_2_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_10_11
#define init_9_2_10_11
void FASTCALL call_9_2_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_10_12
#define init_9_2_10_12
void FASTCALL call_9_2_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_10_13
#define init_9_2_10_13
void FASTCALL call_9_2_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_10_14
#define init_9_2_10_14
void FASTCALL call_9_2_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_10_15
#define init_9_2_10_15
void FASTCALL call_9_2_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_11_0
#define init_9_2_11_0
void FASTCALL call_9_2_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_11_1
#define init_9_2_11_1
void FASTCALL call_9_2_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_11_2
#define init_9_2_11_2
void FASTCALL call_9_2_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_11_3
#define init_9_2_11_3
void FASTCALL call_9_2_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_11_4
#define init_9_2_11_4
void FASTCALL call_9_2_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_11_5
#define init_9_2_11_5
void FASTCALL call_9_2_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_11_6
#define init_9_2_11_6
void FASTCALL call_9_2_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_11_7
#define init_9_2_11_7
void FASTCALL call_9_2_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_11_8
#define init_9_2_11_8
void FASTCALL call_9_2_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_11_9
#define init_9_2_11_9
void FASTCALL call_9_2_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_11_10
#define init_9_2_11_10
void FASTCALL call_9_2_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_11_11
#define init_9_2_11_11
void FASTCALL call_9_2_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_11_12
#define init_9_2_11_12
void FASTCALL call_9_2_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_11_13
#define init_9_2_11_13
void FASTCALL call_9_2_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_11_14
#define init_9_2_11_14
void FASTCALL call_9_2_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_11_15
#define init_9_2_11_15
void FASTCALL call_9_2_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_12_0
#define init_9_2_12_0
void FASTCALL call_9_2_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_12_1
#define init_9_2_12_1
void FASTCALL call_9_2_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_12_2
#define init_9_2_12_2
void FASTCALL call_9_2_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_12_3
#define init_9_2_12_3
void FASTCALL call_9_2_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_12_4
#define init_9_2_12_4
void FASTCALL call_9_2_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_12_5
#define init_9_2_12_5
void FASTCALL call_9_2_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_12_6
#define init_9_2_12_6
void FASTCALL call_9_2_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_12_7
#define init_9_2_12_7
void FASTCALL call_9_2_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_12_8
#define init_9_2_12_8
void FASTCALL call_9_2_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_12_9
#define init_9_2_12_9
void FASTCALL call_9_2_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_12_10
#define init_9_2_12_10
void FASTCALL call_9_2_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_12_11
#define init_9_2_12_11
void FASTCALL call_9_2_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_12_12
#define init_9_2_12_12
void FASTCALL call_9_2_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_12_13
#define init_9_2_12_13
void FASTCALL call_9_2_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_12_14
#define init_9_2_12_14
void FASTCALL call_9_2_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_12_15
#define init_9_2_12_15
void FASTCALL call_9_2_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_13_0
#define init_9_2_13_0
void FASTCALL call_9_2_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_13_1
#define init_9_2_13_1
void FASTCALL call_9_2_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_13_2
#define init_9_2_13_2
void FASTCALL call_9_2_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_13_3
#define init_9_2_13_3
void FASTCALL call_9_2_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_13_4
#define init_9_2_13_4
void FASTCALL call_9_2_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_13_5
#define init_9_2_13_5
void FASTCALL call_9_2_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_13_6
#define init_9_2_13_6
void FASTCALL call_9_2_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_13_7
#define init_9_2_13_7
void FASTCALL call_9_2_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_13_8
#define init_9_2_13_8
void FASTCALL call_9_2_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_13_9
#define init_9_2_13_9
void FASTCALL call_9_2_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_13_10
#define init_9_2_13_10
void FASTCALL call_9_2_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_13_11
#define init_9_2_13_11
void FASTCALL call_9_2_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_13_12
#define init_9_2_13_12
void FASTCALL call_9_2_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_13_13
#define init_9_2_13_13
void FASTCALL call_9_2_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_13_14
#define init_9_2_13_14
void FASTCALL call_9_2_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_13_15
#define init_9_2_13_15
void FASTCALL call_9_2_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_14_0
#define init_9_2_14_0
void FASTCALL call_9_2_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_14_1
#define init_9_2_14_1
void FASTCALL call_9_2_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_14_2
#define init_9_2_14_2
void FASTCALL call_9_2_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_14_3
#define init_9_2_14_3
void FASTCALL call_9_2_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_14_4
#define init_9_2_14_4
void FASTCALL call_9_2_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_14_5
#define init_9_2_14_5
void FASTCALL call_9_2_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_14_6
#define init_9_2_14_6
void FASTCALL call_9_2_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_14_7
#define init_9_2_14_7
void FASTCALL call_9_2_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_14_8
#define init_9_2_14_8
void FASTCALL call_9_2_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_14_9
#define init_9_2_14_9
void FASTCALL call_9_2_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_14_10
#define init_9_2_14_10
void FASTCALL call_9_2_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_14_11
#define init_9_2_14_11
void FASTCALL call_9_2_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_14_12
#define init_9_2_14_12
void FASTCALL call_9_2_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_14_13
#define init_9_2_14_13
void FASTCALL call_9_2_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_14_14
#define init_9_2_14_14
void FASTCALL call_9_2_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_14_15
#define init_9_2_14_15
void FASTCALL call_9_2_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_15_0
#define init_9_2_15_0
void FASTCALL call_9_2_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_15_1
#define init_9_2_15_1
void FASTCALL call_9_2_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_15_2
#define init_9_2_15_2
void FASTCALL call_9_2_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_15_3
#define init_9_2_15_3
void FASTCALL call_9_2_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_15_4
#define init_9_2_15_4
void FASTCALL call_9_2_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_15_5
#define init_9_2_15_5
void FASTCALL call_9_2_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_15_6
#define init_9_2_15_6
void FASTCALL call_9_2_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_15_7
#define init_9_2_15_7
void FASTCALL call_9_2_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_15_8
#define init_9_2_15_8
void FASTCALL call_9_2_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_15_9
#define init_9_2_15_9
void FASTCALL call_9_2_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_15_10
#define init_9_2_15_10
void FASTCALL call_9_2_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_15_11
#define init_9_2_15_11
void FASTCALL call_9_2_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_15_12
#define init_9_2_15_12
void FASTCALL call_9_2_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_15_13
#define init_9_2_15_13
void FASTCALL call_9_2_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_15_14
#define init_9_2_15_14
void FASTCALL call_9_2_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_2_15_15
#define init_9_2_15_15
void FASTCALL call_9_2_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_0_0
#define init_9_3_0_0
void FASTCALL call_9_3_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_0_1
#define init_9_3_0_1
void FASTCALL call_9_3_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_0_2
#define init_9_3_0_2
void FASTCALL call_9_3_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_0_3
#define init_9_3_0_3
void FASTCALL call_9_3_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_0_4
#define init_9_3_0_4
void FASTCALL call_9_3_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_0_5
#define init_9_3_0_5
void FASTCALL call_9_3_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_0_6
#define init_9_3_0_6
void FASTCALL call_9_3_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_0_7
#define init_9_3_0_7
void FASTCALL call_9_3_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_0_8
#define init_9_3_0_8
void FASTCALL call_9_3_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_0_9
#define init_9_3_0_9
void FASTCALL call_9_3_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_0_10
#define init_9_3_0_10
void FASTCALL call_9_3_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_0_11
#define init_9_3_0_11
void FASTCALL call_9_3_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_0_12
#define init_9_3_0_12
void FASTCALL call_9_3_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_0_13
#define init_9_3_0_13
void FASTCALL call_9_3_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_0_14
#define init_9_3_0_14
void FASTCALL call_9_3_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_0_15
#define init_9_3_0_15
void FASTCALL call_9_3_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_1_0
#define init_9_3_1_0
void FASTCALL call_9_3_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_1_1
#define init_9_3_1_1
void FASTCALL call_9_3_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_1_2
#define init_9_3_1_2
void FASTCALL call_9_3_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_1_3
#define init_9_3_1_3
void FASTCALL call_9_3_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_1_4
#define init_9_3_1_4
void FASTCALL call_9_3_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_1_5
#define init_9_3_1_5
void FASTCALL call_9_3_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_1_6
#define init_9_3_1_6
void FASTCALL call_9_3_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_1_7
#define init_9_3_1_7
void FASTCALL call_9_3_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_1_8
#define init_9_3_1_8
void FASTCALL call_9_3_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_1_9
#define init_9_3_1_9
void FASTCALL call_9_3_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_1_10
#define init_9_3_1_10
void FASTCALL call_9_3_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_1_11
#define init_9_3_1_11
void FASTCALL call_9_3_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_1_12
#define init_9_3_1_12
void FASTCALL call_9_3_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_1_13
#define init_9_3_1_13
void FASTCALL call_9_3_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_1_14
#define init_9_3_1_14
void FASTCALL call_9_3_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_1_15
#define init_9_3_1_15
void FASTCALL call_9_3_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_2_0
#define init_9_3_2_0
void FASTCALL call_9_3_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_2_1
#define init_9_3_2_1
void FASTCALL call_9_3_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_2_2
#define init_9_3_2_2
void FASTCALL call_9_3_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_2_3
#define init_9_3_2_3
void FASTCALL call_9_3_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_2_4
#define init_9_3_2_4
void FASTCALL call_9_3_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_2_5
#define init_9_3_2_5
void FASTCALL call_9_3_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_2_6
#define init_9_3_2_6
void FASTCALL call_9_3_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_2_7
#define init_9_3_2_7
void FASTCALL call_9_3_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_2_8
#define init_9_3_2_8
void FASTCALL call_9_3_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_2_9
#define init_9_3_2_9
void FASTCALL call_9_3_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_2_10
#define init_9_3_2_10
void FASTCALL call_9_3_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_2_11
#define init_9_3_2_11
void FASTCALL call_9_3_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_2_12
#define init_9_3_2_12
void FASTCALL call_9_3_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_2_13
#define init_9_3_2_13
void FASTCALL call_9_3_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_2_14
#define init_9_3_2_14
void FASTCALL call_9_3_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_2_15
#define init_9_3_2_15
void FASTCALL call_9_3_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_3_0
#define init_9_3_3_0
void FASTCALL call_9_3_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_3_1
#define init_9_3_3_1
void FASTCALL call_9_3_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_3_2
#define init_9_3_3_2
void FASTCALL call_9_3_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_3_3
#define init_9_3_3_3
void FASTCALL call_9_3_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_3_4
#define init_9_3_3_4
void FASTCALL call_9_3_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_3_5
#define init_9_3_3_5
void FASTCALL call_9_3_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_3_6
#define init_9_3_3_6
void FASTCALL call_9_3_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_3_7
#define init_9_3_3_7
void FASTCALL call_9_3_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_3_8
#define init_9_3_3_8
void FASTCALL call_9_3_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_3_9
#define init_9_3_3_9
void FASTCALL call_9_3_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_3_10
#define init_9_3_3_10
void FASTCALL call_9_3_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_3_11
#define init_9_3_3_11
void FASTCALL call_9_3_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_3_12
#define init_9_3_3_12
void FASTCALL call_9_3_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_3_13
#define init_9_3_3_13
void FASTCALL call_9_3_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_3_14
#define init_9_3_3_14
void FASTCALL call_9_3_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_3_15
#define init_9_3_3_15
void FASTCALL call_9_3_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_4_0
#define init_9_3_4_0
void FASTCALL call_9_3_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_4_1
#define init_9_3_4_1
void FASTCALL call_9_3_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_4_2
#define init_9_3_4_2
void FASTCALL call_9_3_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_4_3
#define init_9_3_4_3
void FASTCALL call_9_3_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_4_4
#define init_9_3_4_4
void FASTCALL call_9_3_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_4_5
#define init_9_3_4_5
void FASTCALL call_9_3_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_4_6
#define init_9_3_4_6
void FASTCALL call_9_3_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_4_7
#define init_9_3_4_7
void FASTCALL call_9_3_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_4_8
#define init_9_3_4_8
void FASTCALL call_9_3_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_4_9
#define init_9_3_4_9
void FASTCALL call_9_3_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_4_10
#define init_9_3_4_10
void FASTCALL call_9_3_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_4_11
#define init_9_3_4_11
void FASTCALL call_9_3_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_4_12
#define init_9_3_4_12
void FASTCALL call_9_3_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_4_13
#define init_9_3_4_13
void FASTCALL call_9_3_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_4_14
#define init_9_3_4_14
void FASTCALL call_9_3_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_4_15
#define init_9_3_4_15
void FASTCALL call_9_3_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_5_0
#define init_9_3_5_0
void FASTCALL call_9_3_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_5_1
#define init_9_3_5_1
void FASTCALL call_9_3_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_5_2
#define init_9_3_5_2
void FASTCALL call_9_3_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_5_3
#define init_9_3_5_3
void FASTCALL call_9_3_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_5_4
#define init_9_3_5_4
void FASTCALL call_9_3_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_5_5
#define init_9_3_5_5
void FASTCALL call_9_3_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_5_6
#define init_9_3_5_6
void FASTCALL call_9_3_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_5_7
#define init_9_3_5_7
void FASTCALL call_9_3_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_5_8
#define init_9_3_5_8
void FASTCALL call_9_3_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_5_9
#define init_9_3_5_9
void FASTCALL call_9_3_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_5_10
#define init_9_3_5_10
void FASTCALL call_9_3_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_5_11
#define init_9_3_5_11
void FASTCALL call_9_3_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_5_12
#define init_9_3_5_12
void FASTCALL call_9_3_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_5_13
#define init_9_3_5_13
void FASTCALL call_9_3_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_5_14
#define init_9_3_5_14
void FASTCALL call_9_3_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_5_15
#define init_9_3_5_15
void FASTCALL call_9_3_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_6_0
#define init_9_3_6_0
void FASTCALL call_9_3_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_6_1
#define init_9_3_6_1
void FASTCALL call_9_3_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_6_2
#define init_9_3_6_2
void FASTCALL call_9_3_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_6_3
#define init_9_3_6_3
void FASTCALL call_9_3_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_6_4
#define init_9_3_6_4
void FASTCALL call_9_3_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_6_5
#define init_9_3_6_5
void FASTCALL call_9_3_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_6_6
#define init_9_3_6_6
void FASTCALL call_9_3_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_6_7
#define init_9_3_6_7
void FASTCALL call_9_3_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_6_8
#define init_9_3_6_8
void FASTCALL call_9_3_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_6_9
#define init_9_3_6_9
void FASTCALL call_9_3_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_6_10
#define init_9_3_6_10
void FASTCALL call_9_3_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_6_11
#define init_9_3_6_11
void FASTCALL call_9_3_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_6_12
#define init_9_3_6_12
void FASTCALL call_9_3_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_6_13
#define init_9_3_6_13
void FASTCALL call_9_3_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_6_14
#define init_9_3_6_14
void FASTCALL call_9_3_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_6_15
#define init_9_3_6_15
void FASTCALL call_9_3_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_7_0
#define init_9_3_7_0
void FASTCALL call_9_3_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_7_1
#define init_9_3_7_1
void FASTCALL call_9_3_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_7_2
#define init_9_3_7_2
void FASTCALL call_9_3_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_7_3
#define init_9_3_7_3
void FASTCALL call_9_3_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_7_4
#define init_9_3_7_4
void FASTCALL call_9_3_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_7_5
#define init_9_3_7_5
void FASTCALL call_9_3_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_7_6
#define init_9_3_7_6
void FASTCALL call_9_3_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_7_7
#define init_9_3_7_7
void FASTCALL call_9_3_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_7_8
#define init_9_3_7_8
void FASTCALL call_9_3_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_7_9
#define init_9_3_7_9
void FASTCALL call_9_3_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_7_10
#define init_9_3_7_10
void FASTCALL call_9_3_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_7_11
#define init_9_3_7_11
void FASTCALL call_9_3_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_7_12
#define init_9_3_7_12
void FASTCALL call_9_3_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_7_13
#define init_9_3_7_13
void FASTCALL call_9_3_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_7_14
#define init_9_3_7_14
void FASTCALL call_9_3_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_7_15
#define init_9_3_7_15
void FASTCALL call_9_3_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_8_0
#define init_9_3_8_0
void FASTCALL call_9_3_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_8_1
#define init_9_3_8_1
void FASTCALL call_9_3_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_8_2
#define init_9_3_8_2
void FASTCALL call_9_3_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_8_3
#define init_9_3_8_3
void FASTCALL call_9_3_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_8_4
#define init_9_3_8_4
void FASTCALL call_9_3_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_8_5
#define init_9_3_8_5
void FASTCALL call_9_3_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_8_6
#define init_9_3_8_6
void FASTCALL call_9_3_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_8_7
#define init_9_3_8_7
void FASTCALL call_9_3_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_8_8
#define init_9_3_8_8
void FASTCALL call_9_3_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_8_9
#define init_9_3_8_9
void FASTCALL call_9_3_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_8_10
#define init_9_3_8_10
void FASTCALL call_9_3_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_8_11
#define init_9_3_8_11
void FASTCALL call_9_3_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_8_12
#define init_9_3_8_12
void FASTCALL call_9_3_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_8_13
#define init_9_3_8_13
void FASTCALL call_9_3_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_8_14
#define init_9_3_8_14
void FASTCALL call_9_3_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_8_15
#define init_9_3_8_15
void FASTCALL call_9_3_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_9_0
#define init_9_3_9_0
void FASTCALL call_9_3_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_9_1
#define init_9_3_9_1
void FASTCALL call_9_3_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_9_2
#define init_9_3_9_2
void FASTCALL call_9_3_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_9_3
#define init_9_3_9_3
void FASTCALL call_9_3_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_9_4
#define init_9_3_9_4
void FASTCALL call_9_3_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_9_5
#define init_9_3_9_5
void FASTCALL call_9_3_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_9_6
#define init_9_3_9_6
void FASTCALL call_9_3_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_9_7
#define init_9_3_9_7
void FASTCALL call_9_3_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_9_8
#define init_9_3_9_8
void FASTCALL call_9_3_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_9_9
#define init_9_3_9_9
void FASTCALL call_9_3_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_9_10
#define init_9_3_9_10
void FASTCALL call_9_3_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_9_11
#define init_9_3_9_11
void FASTCALL call_9_3_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_9_12
#define init_9_3_9_12
void FASTCALL call_9_3_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_9_13
#define init_9_3_9_13
void FASTCALL call_9_3_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_9_14
#define init_9_3_9_14
void FASTCALL call_9_3_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_9_15
#define init_9_3_9_15
void FASTCALL call_9_3_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_10_0
#define init_9_3_10_0
void FASTCALL call_9_3_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_10_1
#define init_9_3_10_1
void FASTCALL call_9_3_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_10_2
#define init_9_3_10_2
void FASTCALL call_9_3_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_10_3
#define init_9_3_10_3
void FASTCALL call_9_3_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_10_4
#define init_9_3_10_4
void FASTCALL call_9_3_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_10_5
#define init_9_3_10_5
void FASTCALL call_9_3_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_10_6
#define init_9_3_10_6
void FASTCALL call_9_3_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_10_7
#define init_9_3_10_7
void FASTCALL call_9_3_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_10_8
#define init_9_3_10_8
void FASTCALL call_9_3_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_10_9
#define init_9_3_10_9
void FASTCALL call_9_3_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_10_10
#define init_9_3_10_10
void FASTCALL call_9_3_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_10_11
#define init_9_3_10_11
void FASTCALL call_9_3_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_10_12
#define init_9_3_10_12
void FASTCALL call_9_3_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_10_13
#define init_9_3_10_13
void FASTCALL call_9_3_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_10_14
#define init_9_3_10_14
void FASTCALL call_9_3_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_10_15
#define init_9_3_10_15
void FASTCALL call_9_3_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_11_0
#define init_9_3_11_0
void FASTCALL call_9_3_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_11_1
#define init_9_3_11_1
void FASTCALL call_9_3_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_11_2
#define init_9_3_11_2
void FASTCALL call_9_3_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_11_3
#define init_9_3_11_3
void FASTCALL call_9_3_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_11_4
#define init_9_3_11_4
void FASTCALL call_9_3_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_11_5
#define init_9_3_11_5
void FASTCALL call_9_3_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_11_6
#define init_9_3_11_6
void FASTCALL call_9_3_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_11_7
#define init_9_3_11_7
void FASTCALL call_9_3_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_11_8
#define init_9_3_11_8
void FASTCALL call_9_3_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_11_9
#define init_9_3_11_9
void FASTCALL call_9_3_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_11_10
#define init_9_3_11_10
void FASTCALL call_9_3_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_11_11
#define init_9_3_11_11
void FASTCALL call_9_3_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_11_12
#define init_9_3_11_12
void FASTCALL call_9_3_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_11_13
#define init_9_3_11_13
void FASTCALL call_9_3_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_11_14
#define init_9_3_11_14
void FASTCALL call_9_3_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_11_15
#define init_9_3_11_15
void FASTCALL call_9_3_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_12_0
#define init_9_3_12_0
void FASTCALL call_9_3_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_12_1
#define init_9_3_12_1
void FASTCALL call_9_3_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_12_2
#define init_9_3_12_2
void FASTCALL call_9_3_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_12_3
#define init_9_3_12_3
void FASTCALL call_9_3_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_12_4
#define init_9_3_12_4
void FASTCALL call_9_3_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_12_5
#define init_9_3_12_5
void FASTCALL call_9_3_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_12_6
#define init_9_3_12_6
void FASTCALL call_9_3_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_12_7
#define init_9_3_12_7
void FASTCALL call_9_3_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_12_8
#define init_9_3_12_8
void FASTCALL call_9_3_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_12_9
#define init_9_3_12_9
void FASTCALL call_9_3_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_12_10
#define init_9_3_12_10
void FASTCALL call_9_3_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_12_11
#define init_9_3_12_11
void FASTCALL call_9_3_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_12_12
#define init_9_3_12_12
void FASTCALL call_9_3_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_12_13
#define init_9_3_12_13
void FASTCALL call_9_3_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_12_14
#define init_9_3_12_14
void FASTCALL call_9_3_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_12_15
#define init_9_3_12_15
void FASTCALL call_9_3_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_13_0
#define init_9_3_13_0
void FASTCALL call_9_3_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_13_1
#define init_9_3_13_1
void FASTCALL call_9_3_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_13_2
#define init_9_3_13_2
void FASTCALL call_9_3_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_13_3
#define init_9_3_13_3
void FASTCALL call_9_3_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_13_4
#define init_9_3_13_4
void FASTCALL call_9_3_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_13_5
#define init_9_3_13_5
void FASTCALL call_9_3_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_13_6
#define init_9_3_13_6
void FASTCALL call_9_3_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_13_7
#define init_9_3_13_7
void FASTCALL call_9_3_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_13_8
#define init_9_3_13_8
void FASTCALL call_9_3_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_13_9
#define init_9_3_13_9
void FASTCALL call_9_3_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_13_10
#define init_9_3_13_10
void FASTCALL call_9_3_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_13_11
#define init_9_3_13_11
void FASTCALL call_9_3_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_13_12
#define init_9_3_13_12
void FASTCALL call_9_3_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_13_13
#define init_9_3_13_13
void FASTCALL call_9_3_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_13_14
#define init_9_3_13_14
void FASTCALL call_9_3_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_13_15
#define init_9_3_13_15
void FASTCALL call_9_3_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_14_0
#define init_9_3_14_0
void FASTCALL call_9_3_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_14_1
#define init_9_3_14_1
void FASTCALL call_9_3_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_14_2
#define init_9_3_14_2
void FASTCALL call_9_3_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_14_3
#define init_9_3_14_3
void FASTCALL call_9_3_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_14_4
#define init_9_3_14_4
void FASTCALL call_9_3_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_14_5
#define init_9_3_14_5
void FASTCALL call_9_3_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_14_6
#define init_9_3_14_6
void FASTCALL call_9_3_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_14_7
#define init_9_3_14_7
void FASTCALL call_9_3_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_14_8
#define init_9_3_14_8
void FASTCALL call_9_3_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_14_9
#define init_9_3_14_9
void FASTCALL call_9_3_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_14_10
#define init_9_3_14_10
void FASTCALL call_9_3_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_14_11
#define init_9_3_14_11
void FASTCALL call_9_3_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_14_12
#define init_9_3_14_12
void FASTCALL call_9_3_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_14_13
#define init_9_3_14_13
void FASTCALL call_9_3_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_14_14
#define init_9_3_14_14
void FASTCALL call_9_3_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_14_15
#define init_9_3_14_15
void FASTCALL call_9_3_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_15_0
#define init_9_3_15_0
void FASTCALL call_9_3_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_15_1
#define init_9_3_15_1
void FASTCALL call_9_3_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_15_2
#define init_9_3_15_2
void FASTCALL call_9_3_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_15_3
#define init_9_3_15_3
void FASTCALL call_9_3_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_15_4
#define init_9_3_15_4
void FASTCALL call_9_3_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_15_5
#define init_9_3_15_5
void FASTCALL call_9_3_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_15_6
#define init_9_3_15_6
void FASTCALL call_9_3_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_15_7
#define init_9_3_15_7
void FASTCALL call_9_3_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_15_8
#define init_9_3_15_8
void FASTCALL call_9_3_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_15_9
#define init_9_3_15_9
void FASTCALL call_9_3_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_15_10
#define init_9_3_15_10
void FASTCALL call_9_3_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_15_11
#define init_9_3_15_11
void FASTCALL call_9_3_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_15_12
#define init_9_3_15_12
void FASTCALL call_9_3_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_15_13
#define init_9_3_15_13
void FASTCALL call_9_3_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_15_14
#define init_9_3_15_14
void FASTCALL call_9_3_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_3_15_15
#define init_9_3_15_15
void FASTCALL call_9_3_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_0_0
#define init_9_4_0_0
void FASTCALL call_9_4_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_0_1
#define init_9_4_0_1
void FASTCALL call_9_4_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_0_2
#define init_9_4_0_2
void FASTCALL call_9_4_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_0_3
#define init_9_4_0_3
void FASTCALL call_9_4_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_0_4
#define init_9_4_0_4
void FASTCALL call_9_4_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_0_5
#define init_9_4_0_5
void FASTCALL call_9_4_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_0_6
#define init_9_4_0_6
void FASTCALL call_9_4_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_0_7
#define init_9_4_0_7
void FASTCALL call_9_4_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_0_8
#define init_9_4_0_8
void FASTCALL call_9_4_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_0_9
#define init_9_4_0_9
void FASTCALL call_9_4_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_0_10
#define init_9_4_0_10
void FASTCALL call_9_4_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_0_11
#define init_9_4_0_11
void FASTCALL call_9_4_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_0_12
#define init_9_4_0_12
void FASTCALL call_9_4_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_0_13
#define init_9_4_0_13
void FASTCALL call_9_4_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_0_14
#define init_9_4_0_14
void FASTCALL call_9_4_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_0_15
#define init_9_4_0_15
void FASTCALL call_9_4_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_1_0
#define init_9_4_1_0
void FASTCALL call_9_4_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_1_1
#define init_9_4_1_1
void FASTCALL call_9_4_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_1_2
#define init_9_4_1_2
void FASTCALL call_9_4_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_1_3
#define init_9_4_1_3
void FASTCALL call_9_4_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_1_4
#define init_9_4_1_4
void FASTCALL call_9_4_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_1_5
#define init_9_4_1_5
void FASTCALL call_9_4_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_1_6
#define init_9_4_1_6
void FASTCALL call_9_4_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_1_7
#define init_9_4_1_7
void FASTCALL call_9_4_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_1_8
#define init_9_4_1_8
void FASTCALL call_9_4_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_1_9
#define init_9_4_1_9
void FASTCALL call_9_4_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_1_10
#define init_9_4_1_10
void FASTCALL call_9_4_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_1_11
#define init_9_4_1_11
void FASTCALL call_9_4_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_1_12
#define init_9_4_1_12
void FASTCALL call_9_4_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_1_13
#define init_9_4_1_13
void FASTCALL call_9_4_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_1_14
#define init_9_4_1_14
void FASTCALL call_9_4_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_1_15
#define init_9_4_1_15
void FASTCALL call_9_4_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_2_0
#define init_9_4_2_0
void FASTCALL call_9_4_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_2_1
#define init_9_4_2_1
void FASTCALL call_9_4_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_2_2
#define init_9_4_2_2
void FASTCALL call_9_4_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_2_3
#define init_9_4_2_3
void FASTCALL call_9_4_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_2_4
#define init_9_4_2_4
void FASTCALL call_9_4_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_2_5
#define init_9_4_2_5
void FASTCALL call_9_4_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_2_6
#define init_9_4_2_6
void FASTCALL call_9_4_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_2_7
#define init_9_4_2_7
void FASTCALL call_9_4_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_2_8
#define init_9_4_2_8
void FASTCALL call_9_4_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_2_9
#define init_9_4_2_9
void FASTCALL call_9_4_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_2_10
#define init_9_4_2_10
void FASTCALL call_9_4_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_2_11
#define init_9_4_2_11
void FASTCALL call_9_4_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_2_12
#define init_9_4_2_12
void FASTCALL call_9_4_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_2_13
#define init_9_4_2_13
void FASTCALL call_9_4_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_2_14
#define init_9_4_2_14
void FASTCALL call_9_4_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_2_15
#define init_9_4_2_15
void FASTCALL call_9_4_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_3_0
#define init_9_4_3_0
void FASTCALL call_9_4_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_3_1
#define init_9_4_3_1
void FASTCALL call_9_4_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_3_2
#define init_9_4_3_2
void FASTCALL call_9_4_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_3_3
#define init_9_4_3_3
void FASTCALL call_9_4_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_3_4
#define init_9_4_3_4
void FASTCALL call_9_4_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_3_5
#define init_9_4_3_5
void FASTCALL call_9_4_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_3_6
#define init_9_4_3_6
void FASTCALL call_9_4_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_3_7
#define init_9_4_3_7
void FASTCALL call_9_4_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_3_8
#define init_9_4_3_8
void FASTCALL call_9_4_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_3_9
#define init_9_4_3_9
void FASTCALL call_9_4_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_3_10
#define init_9_4_3_10
void FASTCALL call_9_4_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_3_11
#define init_9_4_3_11
void FASTCALL call_9_4_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_3_12
#define init_9_4_3_12
void FASTCALL call_9_4_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_3_13
#define init_9_4_3_13
void FASTCALL call_9_4_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_3_14
#define init_9_4_3_14
void FASTCALL call_9_4_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_3_15
#define init_9_4_3_15
void FASTCALL call_9_4_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_4_0
#define init_9_4_4_0
void FASTCALL call_9_4_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_4_1
#define init_9_4_4_1
void FASTCALL call_9_4_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_4_2
#define init_9_4_4_2
void FASTCALL call_9_4_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_4_3
#define init_9_4_4_3
void FASTCALL call_9_4_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_4_4
#define init_9_4_4_4
void FASTCALL call_9_4_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_4_5
#define init_9_4_4_5
void FASTCALL call_9_4_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_4_6
#define init_9_4_4_6
void FASTCALL call_9_4_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_4_7
#define init_9_4_4_7
void FASTCALL call_9_4_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_4_8
#define init_9_4_4_8
void FASTCALL call_9_4_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_4_9
#define init_9_4_4_9
void FASTCALL call_9_4_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_4_10
#define init_9_4_4_10
void FASTCALL call_9_4_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_4_11
#define init_9_4_4_11
void FASTCALL call_9_4_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_4_12
#define init_9_4_4_12
void FASTCALL call_9_4_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_4_13
#define init_9_4_4_13
void FASTCALL call_9_4_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_4_14
#define init_9_4_4_14
void FASTCALL call_9_4_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_4_15
#define init_9_4_4_15
void FASTCALL call_9_4_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_5_0
#define init_9_4_5_0
void FASTCALL call_9_4_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_5_1
#define init_9_4_5_1
void FASTCALL call_9_4_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_5_2
#define init_9_4_5_2
void FASTCALL call_9_4_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_5_3
#define init_9_4_5_3
void FASTCALL call_9_4_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_5_4
#define init_9_4_5_4
void FASTCALL call_9_4_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_5_5
#define init_9_4_5_5
void FASTCALL call_9_4_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_5_6
#define init_9_4_5_6
void FASTCALL call_9_4_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_5_7
#define init_9_4_5_7
void FASTCALL call_9_4_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_5_8
#define init_9_4_5_8
void FASTCALL call_9_4_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_5_9
#define init_9_4_5_9
void FASTCALL call_9_4_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_5_10
#define init_9_4_5_10
void FASTCALL call_9_4_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_5_11
#define init_9_4_5_11
void FASTCALL call_9_4_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_5_12
#define init_9_4_5_12
void FASTCALL call_9_4_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_5_13
#define init_9_4_5_13
void FASTCALL call_9_4_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_5_14
#define init_9_4_5_14
void FASTCALL call_9_4_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_5_15
#define init_9_4_5_15
void FASTCALL call_9_4_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_6_0
#define init_9_4_6_0
void FASTCALL call_9_4_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_6_1
#define init_9_4_6_1
void FASTCALL call_9_4_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_6_2
#define init_9_4_6_2
void FASTCALL call_9_4_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_6_3
#define init_9_4_6_3
void FASTCALL call_9_4_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_6_4
#define init_9_4_6_4
void FASTCALL call_9_4_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_6_5
#define init_9_4_6_5
void FASTCALL call_9_4_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_6_6
#define init_9_4_6_6
void FASTCALL call_9_4_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_6_7
#define init_9_4_6_7
void FASTCALL call_9_4_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_6_8
#define init_9_4_6_8
void FASTCALL call_9_4_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_6_9
#define init_9_4_6_9
void FASTCALL call_9_4_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_6_10
#define init_9_4_6_10
void FASTCALL call_9_4_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_6_11
#define init_9_4_6_11
void FASTCALL call_9_4_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_6_12
#define init_9_4_6_12
void FASTCALL call_9_4_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_6_13
#define init_9_4_6_13
void FASTCALL call_9_4_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_6_14
#define init_9_4_6_14
void FASTCALL call_9_4_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_6_15
#define init_9_4_6_15
void FASTCALL call_9_4_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_7_0
#define init_9_4_7_0
void FASTCALL call_9_4_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_7_1
#define init_9_4_7_1
void FASTCALL call_9_4_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_7_2
#define init_9_4_7_2
void FASTCALL call_9_4_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_7_3
#define init_9_4_7_3
void FASTCALL call_9_4_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_7_4
#define init_9_4_7_4
void FASTCALL call_9_4_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_7_5
#define init_9_4_7_5
void FASTCALL call_9_4_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_7_6
#define init_9_4_7_6
void FASTCALL call_9_4_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_7_7
#define init_9_4_7_7
void FASTCALL call_9_4_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_7_8
#define init_9_4_7_8
void FASTCALL call_9_4_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_7_9
#define init_9_4_7_9
void FASTCALL call_9_4_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_7_10
#define init_9_4_7_10
void FASTCALL call_9_4_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_7_11
#define init_9_4_7_11
void FASTCALL call_9_4_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_7_12
#define init_9_4_7_12
void FASTCALL call_9_4_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_7_13
#define init_9_4_7_13
void FASTCALL call_9_4_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_7_14
#define init_9_4_7_14
void FASTCALL call_9_4_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_7_15
#define init_9_4_7_15
void FASTCALL call_9_4_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_8_0
#define init_9_4_8_0
void FASTCALL call_9_4_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_8_1
#define init_9_4_8_1
void FASTCALL call_9_4_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_8_2
#define init_9_4_8_2
void FASTCALL call_9_4_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_8_3
#define init_9_4_8_3
void FASTCALL call_9_4_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_8_4
#define init_9_4_8_4
void FASTCALL call_9_4_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_8_5
#define init_9_4_8_5
void FASTCALL call_9_4_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_8_6
#define init_9_4_8_6
void FASTCALL call_9_4_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_8_7
#define init_9_4_8_7
void FASTCALL call_9_4_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_8_8
#define init_9_4_8_8
void FASTCALL call_9_4_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_8_9
#define init_9_4_8_9
void FASTCALL call_9_4_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_8_10
#define init_9_4_8_10
void FASTCALL call_9_4_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_8_11
#define init_9_4_8_11
void FASTCALL call_9_4_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_8_12
#define init_9_4_8_12
void FASTCALL call_9_4_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_8_13
#define init_9_4_8_13
void FASTCALL call_9_4_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_8_14
#define init_9_4_8_14
void FASTCALL call_9_4_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_8_15
#define init_9_4_8_15
void FASTCALL call_9_4_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_9_0
#define init_9_4_9_0
void FASTCALL call_9_4_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_9_1
#define init_9_4_9_1
void FASTCALL call_9_4_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_9_2
#define init_9_4_9_2
void FASTCALL call_9_4_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_9_3
#define init_9_4_9_3
void FASTCALL call_9_4_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_9_4
#define init_9_4_9_4
void FASTCALL call_9_4_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_9_5
#define init_9_4_9_5
void FASTCALL call_9_4_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_9_6
#define init_9_4_9_6
void FASTCALL call_9_4_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_9_7
#define init_9_4_9_7
void FASTCALL call_9_4_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_9_8
#define init_9_4_9_8
void FASTCALL call_9_4_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_9_9
#define init_9_4_9_9
void FASTCALL call_9_4_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_9_10
#define init_9_4_9_10
void FASTCALL call_9_4_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_9_11
#define init_9_4_9_11
void FASTCALL call_9_4_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_9_12
#define init_9_4_9_12
void FASTCALL call_9_4_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_9_13
#define init_9_4_9_13
void FASTCALL call_9_4_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_9_14
#define init_9_4_9_14
void FASTCALL call_9_4_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_9_15
#define init_9_4_9_15
void FASTCALL call_9_4_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_10_0
#define init_9_4_10_0
void FASTCALL call_9_4_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_10_1
#define init_9_4_10_1
void FASTCALL call_9_4_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_10_2
#define init_9_4_10_2
void FASTCALL call_9_4_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_10_3
#define init_9_4_10_3
void FASTCALL call_9_4_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_10_4
#define init_9_4_10_4
void FASTCALL call_9_4_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_10_5
#define init_9_4_10_5
void FASTCALL call_9_4_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_10_6
#define init_9_4_10_6
void FASTCALL call_9_4_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_10_7
#define init_9_4_10_7
void FASTCALL call_9_4_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_10_8
#define init_9_4_10_8
void FASTCALL call_9_4_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_10_9
#define init_9_4_10_9
void FASTCALL call_9_4_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_10_10
#define init_9_4_10_10
void FASTCALL call_9_4_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_10_11
#define init_9_4_10_11
void FASTCALL call_9_4_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_10_12
#define init_9_4_10_12
void FASTCALL call_9_4_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_10_13
#define init_9_4_10_13
void FASTCALL call_9_4_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_10_14
#define init_9_4_10_14
void FASTCALL call_9_4_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_10_15
#define init_9_4_10_15
void FASTCALL call_9_4_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_11_0
#define init_9_4_11_0
void FASTCALL call_9_4_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_11_1
#define init_9_4_11_1
void FASTCALL call_9_4_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_11_2
#define init_9_4_11_2
void FASTCALL call_9_4_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_11_3
#define init_9_4_11_3
void FASTCALL call_9_4_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_11_4
#define init_9_4_11_4
void FASTCALL call_9_4_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_11_5
#define init_9_4_11_5
void FASTCALL call_9_4_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_11_6
#define init_9_4_11_6
void FASTCALL call_9_4_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_11_7
#define init_9_4_11_7
void FASTCALL call_9_4_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_11_8
#define init_9_4_11_8
void FASTCALL call_9_4_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_11_9
#define init_9_4_11_9
void FASTCALL call_9_4_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_11_10
#define init_9_4_11_10
void FASTCALL call_9_4_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_11_11
#define init_9_4_11_11
void FASTCALL call_9_4_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_11_12
#define init_9_4_11_12
void FASTCALL call_9_4_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_11_13
#define init_9_4_11_13
void FASTCALL call_9_4_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_11_14
#define init_9_4_11_14
void FASTCALL call_9_4_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_11_15
#define init_9_4_11_15
void FASTCALL call_9_4_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_12_0
#define init_9_4_12_0
void FASTCALL call_9_4_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_12_1
#define init_9_4_12_1
void FASTCALL call_9_4_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_12_2
#define init_9_4_12_2
void FASTCALL call_9_4_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_12_3
#define init_9_4_12_3
void FASTCALL call_9_4_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_12_4
#define init_9_4_12_4
void FASTCALL call_9_4_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_12_5
#define init_9_4_12_5
void FASTCALL call_9_4_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_12_6
#define init_9_4_12_6
void FASTCALL call_9_4_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_12_7
#define init_9_4_12_7
void FASTCALL call_9_4_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_12_8
#define init_9_4_12_8
void FASTCALL call_9_4_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_12_9
#define init_9_4_12_9
void FASTCALL call_9_4_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_12_10
#define init_9_4_12_10
void FASTCALL call_9_4_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_12_11
#define init_9_4_12_11
void FASTCALL call_9_4_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_12_12
#define init_9_4_12_12
void FASTCALL call_9_4_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_12_13
#define init_9_4_12_13
void FASTCALL call_9_4_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_12_14
#define init_9_4_12_14
void FASTCALL call_9_4_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_12_15
#define init_9_4_12_15
void FASTCALL call_9_4_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_13_0
#define init_9_4_13_0
void FASTCALL call_9_4_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_13_1
#define init_9_4_13_1
void FASTCALL call_9_4_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_13_2
#define init_9_4_13_2
void FASTCALL call_9_4_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_13_3
#define init_9_4_13_3
void FASTCALL call_9_4_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_13_4
#define init_9_4_13_4
void FASTCALL call_9_4_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_13_5
#define init_9_4_13_5
void FASTCALL call_9_4_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_13_6
#define init_9_4_13_6
void FASTCALL call_9_4_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_13_7
#define init_9_4_13_7
void FASTCALL call_9_4_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_13_8
#define init_9_4_13_8
void FASTCALL call_9_4_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_13_9
#define init_9_4_13_9
void FASTCALL call_9_4_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_13_10
#define init_9_4_13_10
void FASTCALL call_9_4_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_13_11
#define init_9_4_13_11
void FASTCALL call_9_4_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_13_12
#define init_9_4_13_12
void FASTCALL call_9_4_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_13_13
#define init_9_4_13_13
void FASTCALL call_9_4_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_13_14
#define init_9_4_13_14
void FASTCALL call_9_4_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_13_15
#define init_9_4_13_15
void FASTCALL call_9_4_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_14_0
#define init_9_4_14_0
void FASTCALL call_9_4_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_14_1
#define init_9_4_14_1
void FASTCALL call_9_4_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_14_2
#define init_9_4_14_2
void FASTCALL call_9_4_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_14_3
#define init_9_4_14_3
void FASTCALL call_9_4_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_14_4
#define init_9_4_14_4
void FASTCALL call_9_4_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_14_5
#define init_9_4_14_5
void FASTCALL call_9_4_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_14_6
#define init_9_4_14_6
void FASTCALL call_9_4_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_14_7
#define init_9_4_14_7
void FASTCALL call_9_4_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_14_8
#define init_9_4_14_8
void FASTCALL call_9_4_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_14_9
#define init_9_4_14_9
void FASTCALL call_9_4_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_14_10
#define init_9_4_14_10
void FASTCALL call_9_4_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_14_11
#define init_9_4_14_11
void FASTCALL call_9_4_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_14_12
#define init_9_4_14_12
void FASTCALL call_9_4_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_14_13
#define init_9_4_14_13
void FASTCALL call_9_4_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_14_14
#define init_9_4_14_14
void FASTCALL call_9_4_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_14_15
#define init_9_4_14_15
void FASTCALL call_9_4_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_15_0
#define init_9_4_15_0
void FASTCALL call_9_4_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_15_1
#define init_9_4_15_1
void FASTCALL call_9_4_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_15_2
#define init_9_4_15_2
void FASTCALL call_9_4_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_15_3
#define init_9_4_15_3
void FASTCALL call_9_4_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_15_4
#define init_9_4_15_4
void FASTCALL call_9_4_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_15_5
#define init_9_4_15_5
void FASTCALL call_9_4_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_15_6
#define init_9_4_15_6
void FASTCALL call_9_4_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_15_7
#define init_9_4_15_7
void FASTCALL call_9_4_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_15_8
#define init_9_4_15_8
void FASTCALL call_9_4_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_15_9
#define init_9_4_15_9
void FASTCALL call_9_4_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_15_10
#define init_9_4_15_10
void FASTCALL call_9_4_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_15_11
#define init_9_4_15_11
void FASTCALL call_9_4_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_15_12
#define init_9_4_15_12
void FASTCALL call_9_4_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_15_13
#define init_9_4_15_13
void FASTCALL call_9_4_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_15_14
#define init_9_4_15_14
void FASTCALL call_9_4_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_4_15_15
#define init_9_4_15_15
void FASTCALL call_9_4_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_0_0
#define init_9_5_0_0
void FASTCALL call_9_5_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_0_1
#define init_9_5_0_1
void FASTCALL call_9_5_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_0_2
#define init_9_5_0_2
void FASTCALL call_9_5_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_0_3
#define init_9_5_0_3
void FASTCALL call_9_5_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_0_4
#define init_9_5_0_4
void FASTCALL call_9_5_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_0_5
#define init_9_5_0_5
void FASTCALL call_9_5_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_0_6
#define init_9_5_0_6
void FASTCALL call_9_5_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_0_7
#define init_9_5_0_7
void FASTCALL call_9_5_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_0_8
#define init_9_5_0_8
void FASTCALL call_9_5_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_0_9
#define init_9_5_0_9
void FASTCALL call_9_5_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_0_10
#define init_9_5_0_10
void FASTCALL call_9_5_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_0_11
#define init_9_5_0_11
void FASTCALL call_9_5_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_0_12
#define init_9_5_0_12
void FASTCALL call_9_5_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_0_13
#define init_9_5_0_13
void FASTCALL call_9_5_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_0_14
#define init_9_5_0_14
void FASTCALL call_9_5_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_0_15
#define init_9_5_0_15
void FASTCALL call_9_5_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_1_0
#define init_9_5_1_0
void FASTCALL call_9_5_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_1_1
#define init_9_5_1_1
void FASTCALL call_9_5_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_1_2
#define init_9_5_1_2
void FASTCALL call_9_5_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_1_3
#define init_9_5_1_3
void FASTCALL call_9_5_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_1_4
#define init_9_5_1_4
void FASTCALL call_9_5_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_1_5
#define init_9_5_1_5
void FASTCALL call_9_5_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_1_6
#define init_9_5_1_6
void FASTCALL call_9_5_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_1_7
#define init_9_5_1_7
void FASTCALL call_9_5_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_1_8
#define init_9_5_1_8
void FASTCALL call_9_5_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_1_9
#define init_9_5_1_9
void FASTCALL call_9_5_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_1_10
#define init_9_5_1_10
void FASTCALL call_9_5_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_1_11
#define init_9_5_1_11
void FASTCALL call_9_5_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_1_12
#define init_9_5_1_12
void FASTCALL call_9_5_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_1_13
#define init_9_5_1_13
void FASTCALL call_9_5_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_1_14
#define init_9_5_1_14
void FASTCALL call_9_5_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_1_15
#define init_9_5_1_15
void FASTCALL call_9_5_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_2_0
#define init_9_5_2_0
void FASTCALL call_9_5_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_2_1
#define init_9_5_2_1
void FASTCALL call_9_5_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_2_2
#define init_9_5_2_2
void FASTCALL call_9_5_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_2_3
#define init_9_5_2_3
void FASTCALL call_9_5_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_2_4
#define init_9_5_2_4
void FASTCALL call_9_5_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_2_5
#define init_9_5_2_5
void FASTCALL call_9_5_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_2_6
#define init_9_5_2_6
void FASTCALL call_9_5_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_2_7
#define init_9_5_2_7
void FASTCALL call_9_5_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_2_8
#define init_9_5_2_8
void FASTCALL call_9_5_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_2_9
#define init_9_5_2_9
void FASTCALL call_9_5_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_2_10
#define init_9_5_2_10
void FASTCALL call_9_5_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_2_11
#define init_9_5_2_11
void FASTCALL call_9_5_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_2_12
#define init_9_5_2_12
void FASTCALL call_9_5_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_2_13
#define init_9_5_2_13
void FASTCALL call_9_5_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_2_14
#define init_9_5_2_14
void FASTCALL call_9_5_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_2_15
#define init_9_5_2_15
void FASTCALL call_9_5_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_3_0
#define init_9_5_3_0
void FASTCALL call_9_5_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_3_1
#define init_9_5_3_1
void FASTCALL call_9_5_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_3_2
#define init_9_5_3_2
void FASTCALL call_9_5_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_3_3
#define init_9_5_3_3
void FASTCALL call_9_5_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_3_4
#define init_9_5_3_4
void FASTCALL call_9_5_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_3_5
#define init_9_5_3_5
void FASTCALL call_9_5_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_3_6
#define init_9_5_3_6
void FASTCALL call_9_5_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_3_7
#define init_9_5_3_7
void FASTCALL call_9_5_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_3_8
#define init_9_5_3_8
void FASTCALL call_9_5_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_3_9
#define init_9_5_3_9
void FASTCALL call_9_5_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_3_10
#define init_9_5_3_10
void FASTCALL call_9_5_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_3_11
#define init_9_5_3_11
void FASTCALL call_9_5_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_3_12
#define init_9_5_3_12
void FASTCALL call_9_5_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_3_13
#define init_9_5_3_13
void FASTCALL call_9_5_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_3_14
#define init_9_5_3_14
void FASTCALL call_9_5_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_3_15
#define init_9_5_3_15
void FASTCALL call_9_5_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_4_0
#define init_9_5_4_0
void FASTCALL call_9_5_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_4_1
#define init_9_5_4_1
void FASTCALL call_9_5_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_4_2
#define init_9_5_4_2
void FASTCALL call_9_5_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_4_3
#define init_9_5_4_3
void FASTCALL call_9_5_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_4_4
#define init_9_5_4_4
void FASTCALL call_9_5_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_4_5
#define init_9_5_4_5
void FASTCALL call_9_5_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_4_6
#define init_9_5_4_6
void FASTCALL call_9_5_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_4_7
#define init_9_5_4_7
void FASTCALL call_9_5_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_4_8
#define init_9_5_4_8
void FASTCALL call_9_5_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_4_9
#define init_9_5_4_9
void FASTCALL call_9_5_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_4_10
#define init_9_5_4_10
void FASTCALL call_9_5_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_4_11
#define init_9_5_4_11
void FASTCALL call_9_5_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_4_12
#define init_9_5_4_12
void FASTCALL call_9_5_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_4_13
#define init_9_5_4_13
void FASTCALL call_9_5_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_4_14
#define init_9_5_4_14
void FASTCALL call_9_5_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_4_15
#define init_9_5_4_15
void FASTCALL call_9_5_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_5_0
#define init_9_5_5_0
void FASTCALL call_9_5_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_5_1
#define init_9_5_5_1
void FASTCALL call_9_5_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_5_2
#define init_9_5_5_2
void FASTCALL call_9_5_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_5_3
#define init_9_5_5_3
void FASTCALL call_9_5_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_5_4
#define init_9_5_5_4
void FASTCALL call_9_5_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_5_5
#define init_9_5_5_5
void FASTCALL call_9_5_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_5_6
#define init_9_5_5_6
void FASTCALL call_9_5_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_5_7
#define init_9_5_5_7
void FASTCALL call_9_5_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_5_8
#define init_9_5_5_8
void FASTCALL call_9_5_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_5_9
#define init_9_5_5_9
void FASTCALL call_9_5_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_5_10
#define init_9_5_5_10
void FASTCALL call_9_5_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_5_11
#define init_9_5_5_11
void FASTCALL call_9_5_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_5_12
#define init_9_5_5_12
void FASTCALL call_9_5_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_5_13
#define init_9_5_5_13
void FASTCALL call_9_5_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_5_14
#define init_9_5_5_14
void FASTCALL call_9_5_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_5_15
#define init_9_5_5_15
void FASTCALL call_9_5_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_6_0
#define init_9_5_6_0
void FASTCALL call_9_5_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_6_1
#define init_9_5_6_1
void FASTCALL call_9_5_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_6_2
#define init_9_5_6_2
void FASTCALL call_9_5_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_6_3
#define init_9_5_6_3
void FASTCALL call_9_5_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_6_4
#define init_9_5_6_4
void FASTCALL call_9_5_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_6_5
#define init_9_5_6_5
void FASTCALL call_9_5_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_6_6
#define init_9_5_6_6
void FASTCALL call_9_5_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_6_7
#define init_9_5_6_7
void FASTCALL call_9_5_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_6_8
#define init_9_5_6_8
void FASTCALL call_9_5_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_6_9
#define init_9_5_6_9
void FASTCALL call_9_5_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_6_10
#define init_9_5_6_10
void FASTCALL call_9_5_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_6_11
#define init_9_5_6_11
void FASTCALL call_9_5_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_6_12
#define init_9_5_6_12
void FASTCALL call_9_5_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_6_13
#define init_9_5_6_13
void FASTCALL call_9_5_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_6_14
#define init_9_5_6_14
void FASTCALL call_9_5_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_6_15
#define init_9_5_6_15
void FASTCALL call_9_5_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_7_0
#define init_9_5_7_0
void FASTCALL call_9_5_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_7_1
#define init_9_5_7_1
void FASTCALL call_9_5_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_7_2
#define init_9_5_7_2
void FASTCALL call_9_5_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_7_3
#define init_9_5_7_3
void FASTCALL call_9_5_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_7_4
#define init_9_5_7_4
void FASTCALL call_9_5_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_7_5
#define init_9_5_7_5
void FASTCALL call_9_5_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_7_6
#define init_9_5_7_6
void FASTCALL call_9_5_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_7_7
#define init_9_5_7_7
void FASTCALL call_9_5_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_7_8
#define init_9_5_7_8
void FASTCALL call_9_5_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_7_9
#define init_9_5_7_9
void FASTCALL call_9_5_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_7_10
#define init_9_5_7_10
void FASTCALL call_9_5_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_7_11
#define init_9_5_7_11
void FASTCALL call_9_5_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_7_12
#define init_9_5_7_12
void FASTCALL call_9_5_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_7_13
#define init_9_5_7_13
void FASTCALL call_9_5_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_7_14
#define init_9_5_7_14
void FASTCALL call_9_5_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_7_15
#define init_9_5_7_15
void FASTCALL call_9_5_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_8_0
#define init_9_5_8_0
void FASTCALL call_9_5_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_8_1
#define init_9_5_8_1
void FASTCALL call_9_5_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_8_2
#define init_9_5_8_2
void FASTCALL call_9_5_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_8_3
#define init_9_5_8_3
void FASTCALL call_9_5_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_8_4
#define init_9_5_8_4
void FASTCALL call_9_5_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_8_5
#define init_9_5_8_5
void FASTCALL call_9_5_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_8_6
#define init_9_5_8_6
void FASTCALL call_9_5_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_8_7
#define init_9_5_8_7
void FASTCALL call_9_5_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_8_8
#define init_9_5_8_8
void FASTCALL call_9_5_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_8_9
#define init_9_5_8_9
void FASTCALL call_9_5_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_8_10
#define init_9_5_8_10
void FASTCALL call_9_5_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_8_11
#define init_9_5_8_11
void FASTCALL call_9_5_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_8_12
#define init_9_5_8_12
void FASTCALL call_9_5_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_8_13
#define init_9_5_8_13
void FASTCALL call_9_5_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_8_14
#define init_9_5_8_14
void FASTCALL call_9_5_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_8_15
#define init_9_5_8_15
void FASTCALL call_9_5_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_9_0
#define init_9_5_9_0
void FASTCALL call_9_5_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_9_1
#define init_9_5_9_1
void FASTCALL call_9_5_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_9_2
#define init_9_5_9_2
void FASTCALL call_9_5_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_9_3
#define init_9_5_9_3
void FASTCALL call_9_5_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_9_4
#define init_9_5_9_4
void FASTCALL call_9_5_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_9_5
#define init_9_5_9_5
void FASTCALL call_9_5_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_9_6
#define init_9_5_9_6
void FASTCALL call_9_5_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_9_7
#define init_9_5_9_7
void FASTCALL call_9_5_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_9_8
#define init_9_5_9_8
void FASTCALL call_9_5_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_9_9
#define init_9_5_9_9
void FASTCALL call_9_5_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_9_10
#define init_9_5_9_10
void FASTCALL call_9_5_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_9_11
#define init_9_5_9_11
void FASTCALL call_9_5_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_9_12
#define init_9_5_9_12
void FASTCALL call_9_5_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_9_13
#define init_9_5_9_13
void FASTCALL call_9_5_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_9_14
#define init_9_5_9_14
void FASTCALL call_9_5_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_9_15
#define init_9_5_9_15
void FASTCALL call_9_5_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_10_0
#define init_9_5_10_0
void FASTCALL call_9_5_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_10_1
#define init_9_5_10_1
void FASTCALL call_9_5_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_10_2
#define init_9_5_10_2
void FASTCALL call_9_5_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_10_3
#define init_9_5_10_3
void FASTCALL call_9_5_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_10_4
#define init_9_5_10_4
void FASTCALL call_9_5_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_10_5
#define init_9_5_10_5
void FASTCALL call_9_5_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_10_6
#define init_9_5_10_6
void FASTCALL call_9_5_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_10_7
#define init_9_5_10_7
void FASTCALL call_9_5_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_10_8
#define init_9_5_10_8
void FASTCALL call_9_5_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_10_9
#define init_9_5_10_9
void FASTCALL call_9_5_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_10_10
#define init_9_5_10_10
void FASTCALL call_9_5_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_10_11
#define init_9_5_10_11
void FASTCALL call_9_5_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_10_12
#define init_9_5_10_12
void FASTCALL call_9_5_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_10_13
#define init_9_5_10_13
void FASTCALL call_9_5_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_10_14
#define init_9_5_10_14
void FASTCALL call_9_5_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_10_15
#define init_9_5_10_15
void FASTCALL call_9_5_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_11_0
#define init_9_5_11_0
void FASTCALL call_9_5_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_11_1
#define init_9_5_11_1
void FASTCALL call_9_5_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_11_2
#define init_9_5_11_2
void FASTCALL call_9_5_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_11_3
#define init_9_5_11_3
void FASTCALL call_9_5_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_11_4
#define init_9_5_11_4
void FASTCALL call_9_5_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_11_5
#define init_9_5_11_5
void FASTCALL call_9_5_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_11_6
#define init_9_5_11_6
void FASTCALL call_9_5_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_11_7
#define init_9_5_11_7
void FASTCALL call_9_5_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_11_8
#define init_9_5_11_8
void FASTCALL call_9_5_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_11_9
#define init_9_5_11_9
void FASTCALL call_9_5_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_11_10
#define init_9_5_11_10
void FASTCALL call_9_5_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_11_11
#define init_9_5_11_11
void FASTCALL call_9_5_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_11_12
#define init_9_5_11_12
void FASTCALL call_9_5_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_11_13
#define init_9_5_11_13
void FASTCALL call_9_5_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_11_14
#define init_9_5_11_14
void FASTCALL call_9_5_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_11_15
#define init_9_5_11_15
void FASTCALL call_9_5_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_12_0
#define init_9_5_12_0
void FASTCALL call_9_5_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_12_1
#define init_9_5_12_1
void FASTCALL call_9_5_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_12_2
#define init_9_5_12_2
void FASTCALL call_9_5_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_12_3
#define init_9_5_12_3
void FASTCALL call_9_5_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_12_4
#define init_9_5_12_4
void FASTCALL call_9_5_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_12_5
#define init_9_5_12_5
void FASTCALL call_9_5_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_12_6
#define init_9_5_12_6
void FASTCALL call_9_5_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_12_7
#define init_9_5_12_7
void FASTCALL call_9_5_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_12_8
#define init_9_5_12_8
void FASTCALL call_9_5_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_12_9
#define init_9_5_12_9
void FASTCALL call_9_5_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_12_10
#define init_9_5_12_10
void FASTCALL call_9_5_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_12_11
#define init_9_5_12_11
void FASTCALL call_9_5_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_12_12
#define init_9_5_12_12
void FASTCALL call_9_5_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_12_13
#define init_9_5_12_13
void FASTCALL call_9_5_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_12_14
#define init_9_5_12_14
void FASTCALL call_9_5_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_12_15
#define init_9_5_12_15
void FASTCALL call_9_5_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_13_0
#define init_9_5_13_0
void FASTCALL call_9_5_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_13_1
#define init_9_5_13_1
void FASTCALL call_9_5_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_13_2
#define init_9_5_13_2
void FASTCALL call_9_5_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_13_3
#define init_9_5_13_3
void FASTCALL call_9_5_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_13_4
#define init_9_5_13_4
void FASTCALL call_9_5_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_13_5
#define init_9_5_13_5
void FASTCALL call_9_5_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_13_6
#define init_9_5_13_6
void FASTCALL call_9_5_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_13_7
#define init_9_5_13_7
void FASTCALL call_9_5_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_13_8
#define init_9_5_13_8
void FASTCALL call_9_5_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_13_9
#define init_9_5_13_9
void FASTCALL call_9_5_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_13_10
#define init_9_5_13_10
void FASTCALL call_9_5_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_13_11
#define init_9_5_13_11
void FASTCALL call_9_5_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_13_12
#define init_9_5_13_12
void FASTCALL call_9_5_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_13_13
#define init_9_5_13_13
void FASTCALL call_9_5_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_13_14
#define init_9_5_13_14
void FASTCALL call_9_5_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_13_15
#define init_9_5_13_15
void FASTCALL call_9_5_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_14_0
#define init_9_5_14_0
void FASTCALL call_9_5_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_14_1
#define init_9_5_14_1
void FASTCALL call_9_5_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_14_2
#define init_9_5_14_2
void FASTCALL call_9_5_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_14_3
#define init_9_5_14_3
void FASTCALL call_9_5_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_14_4
#define init_9_5_14_4
void FASTCALL call_9_5_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_14_5
#define init_9_5_14_5
void FASTCALL call_9_5_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_14_6
#define init_9_5_14_6
void FASTCALL call_9_5_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_14_7
#define init_9_5_14_7
void FASTCALL call_9_5_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_14_8
#define init_9_5_14_8
void FASTCALL call_9_5_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_14_9
#define init_9_5_14_9
void FASTCALL call_9_5_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_14_10
#define init_9_5_14_10
void FASTCALL call_9_5_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_14_11
#define init_9_5_14_11
void FASTCALL call_9_5_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_14_12
#define init_9_5_14_12
void FASTCALL call_9_5_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_14_13
#define init_9_5_14_13
void FASTCALL call_9_5_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_14_14
#define init_9_5_14_14
void FASTCALL call_9_5_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_14_15
#define init_9_5_14_15
void FASTCALL call_9_5_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_15_0
#define init_9_5_15_0
void FASTCALL call_9_5_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_15_1
#define init_9_5_15_1
void FASTCALL call_9_5_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_15_2
#define init_9_5_15_2
void FASTCALL call_9_5_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_15_3
#define init_9_5_15_3
void FASTCALL call_9_5_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_15_4
#define init_9_5_15_4
void FASTCALL call_9_5_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_15_5
#define init_9_5_15_5
void FASTCALL call_9_5_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_15_6
#define init_9_5_15_6
void FASTCALL call_9_5_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_15_7
#define init_9_5_15_7
void FASTCALL call_9_5_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_15_8
#define init_9_5_15_8
void FASTCALL call_9_5_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_15_9
#define init_9_5_15_9
void FASTCALL call_9_5_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_15_10
#define init_9_5_15_10
void FASTCALL call_9_5_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_15_11
#define init_9_5_15_11
void FASTCALL call_9_5_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_15_12
#define init_9_5_15_12
void FASTCALL call_9_5_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_15_13
#define init_9_5_15_13
void FASTCALL call_9_5_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_15_14
#define init_9_5_15_14
void FASTCALL call_9_5_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_5_15_15
#define init_9_5_15_15
void FASTCALL call_9_5_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_0_0
#define init_9_6_0_0
void FASTCALL call_9_6_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_0_1
#define init_9_6_0_1
void FASTCALL call_9_6_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_0_2
#define init_9_6_0_2
void FASTCALL call_9_6_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_0_3
#define init_9_6_0_3
void FASTCALL call_9_6_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_0_4
#define init_9_6_0_4
void FASTCALL call_9_6_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_0_5
#define init_9_6_0_5
void FASTCALL call_9_6_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_0_6
#define init_9_6_0_6
void FASTCALL call_9_6_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_0_7
#define init_9_6_0_7
void FASTCALL call_9_6_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_0_8
#define init_9_6_0_8
void FASTCALL call_9_6_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_0_9
#define init_9_6_0_9
void FASTCALL call_9_6_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_0_10
#define init_9_6_0_10
void FASTCALL call_9_6_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_0_11
#define init_9_6_0_11
void FASTCALL call_9_6_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_0_12
#define init_9_6_0_12
void FASTCALL call_9_6_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_0_13
#define init_9_6_0_13
void FASTCALL call_9_6_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_0_14
#define init_9_6_0_14
void FASTCALL call_9_6_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_0_15
#define init_9_6_0_15
void FASTCALL call_9_6_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_1_0
#define init_9_6_1_0
void FASTCALL call_9_6_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_1_1
#define init_9_6_1_1
void FASTCALL call_9_6_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_1_2
#define init_9_6_1_2
void FASTCALL call_9_6_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_1_3
#define init_9_6_1_3
void FASTCALL call_9_6_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_1_4
#define init_9_6_1_4
void FASTCALL call_9_6_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_1_5
#define init_9_6_1_5
void FASTCALL call_9_6_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_1_6
#define init_9_6_1_6
void FASTCALL call_9_6_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_1_7
#define init_9_6_1_7
void FASTCALL call_9_6_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_1_8
#define init_9_6_1_8
void FASTCALL call_9_6_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_1_9
#define init_9_6_1_9
void FASTCALL call_9_6_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_1_10
#define init_9_6_1_10
void FASTCALL call_9_6_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_1_11
#define init_9_6_1_11
void FASTCALL call_9_6_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_1_12
#define init_9_6_1_12
void FASTCALL call_9_6_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_1_13
#define init_9_6_1_13
void FASTCALL call_9_6_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_1_14
#define init_9_6_1_14
void FASTCALL call_9_6_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_1_15
#define init_9_6_1_15
void FASTCALL call_9_6_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_2_0
#define init_9_6_2_0
void FASTCALL call_9_6_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_2_1
#define init_9_6_2_1
void FASTCALL call_9_6_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_2_2
#define init_9_6_2_2
void FASTCALL call_9_6_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_2_3
#define init_9_6_2_3
void FASTCALL call_9_6_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_2_4
#define init_9_6_2_4
void FASTCALL call_9_6_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_2_5
#define init_9_6_2_5
void FASTCALL call_9_6_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_2_6
#define init_9_6_2_6
void FASTCALL call_9_6_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_2_7
#define init_9_6_2_7
void FASTCALL call_9_6_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_2_8
#define init_9_6_2_8
void FASTCALL call_9_6_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_2_9
#define init_9_6_2_9
void FASTCALL call_9_6_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_2_10
#define init_9_6_2_10
void FASTCALL call_9_6_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_2_11
#define init_9_6_2_11
void FASTCALL call_9_6_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_2_12
#define init_9_6_2_12
void FASTCALL call_9_6_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_2_13
#define init_9_6_2_13
void FASTCALL call_9_6_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_2_14
#define init_9_6_2_14
void FASTCALL call_9_6_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_2_15
#define init_9_6_2_15
void FASTCALL call_9_6_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_3_0
#define init_9_6_3_0
void FASTCALL call_9_6_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_3_1
#define init_9_6_3_1
void FASTCALL call_9_6_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_3_2
#define init_9_6_3_2
void FASTCALL call_9_6_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_3_3
#define init_9_6_3_3
void FASTCALL call_9_6_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_3_4
#define init_9_6_3_4
void FASTCALL call_9_6_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_3_5
#define init_9_6_3_5
void FASTCALL call_9_6_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_3_6
#define init_9_6_3_6
void FASTCALL call_9_6_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_3_7
#define init_9_6_3_7
void FASTCALL call_9_6_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_3_8
#define init_9_6_3_8
void FASTCALL call_9_6_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_3_9
#define init_9_6_3_9
void FASTCALL call_9_6_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_3_10
#define init_9_6_3_10
void FASTCALL call_9_6_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_3_11
#define init_9_6_3_11
void FASTCALL call_9_6_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_3_12
#define init_9_6_3_12
void FASTCALL call_9_6_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_3_13
#define init_9_6_3_13
void FASTCALL call_9_6_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_3_14
#define init_9_6_3_14
void FASTCALL call_9_6_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_3_15
#define init_9_6_3_15
void FASTCALL call_9_6_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_4_0
#define init_9_6_4_0
void FASTCALL call_9_6_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_4_1
#define init_9_6_4_1
void FASTCALL call_9_6_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_4_2
#define init_9_6_4_2
void FASTCALL call_9_6_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_4_3
#define init_9_6_4_3
void FASTCALL call_9_6_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_4_4
#define init_9_6_4_4
void FASTCALL call_9_6_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_4_5
#define init_9_6_4_5
void FASTCALL call_9_6_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_4_6
#define init_9_6_4_6
void FASTCALL call_9_6_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_4_7
#define init_9_6_4_7
void FASTCALL call_9_6_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_4_8
#define init_9_6_4_8
void FASTCALL call_9_6_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_4_9
#define init_9_6_4_9
void FASTCALL call_9_6_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_4_10
#define init_9_6_4_10
void FASTCALL call_9_6_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_4_11
#define init_9_6_4_11
void FASTCALL call_9_6_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_4_12
#define init_9_6_4_12
void FASTCALL call_9_6_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_4_13
#define init_9_6_4_13
void FASTCALL call_9_6_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_4_14
#define init_9_6_4_14
void FASTCALL call_9_6_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_4_15
#define init_9_6_4_15
void FASTCALL call_9_6_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_5_0
#define init_9_6_5_0
void FASTCALL call_9_6_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_5_1
#define init_9_6_5_1
void FASTCALL call_9_6_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_5_2
#define init_9_6_5_2
void FASTCALL call_9_6_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_5_3
#define init_9_6_5_3
void FASTCALL call_9_6_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_5_4
#define init_9_6_5_4
void FASTCALL call_9_6_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_5_5
#define init_9_6_5_5
void FASTCALL call_9_6_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_5_6
#define init_9_6_5_6
void FASTCALL call_9_6_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_5_7
#define init_9_6_5_7
void FASTCALL call_9_6_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_5_8
#define init_9_6_5_8
void FASTCALL call_9_6_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_5_9
#define init_9_6_5_9
void FASTCALL call_9_6_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_5_10
#define init_9_6_5_10
void FASTCALL call_9_6_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_5_11
#define init_9_6_5_11
void FASTCALL call_9_6_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_5_12
#define init_9_6_5_12
void FASTCALL call_9_6_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_5_13
#define init_9_6_5_13
void FASTCALL call_9_6_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_5_14
#define init_9_6_5_14
void FASTCALL call_9_6_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_5_15
#define init_9_6_5_15
void FASTCALL call_9_6_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_6_0
#define init_9_6_6_0
void FASTCALL call_9_6_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_6_1
#define init_9_6_6_1
void FASTCALL call_9_6_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_6_2
#define init_9_6_6_2
void FASTCALL call_9_6_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_6_3
#define init_9_6_6_3
void FASTCALL call_9_6_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_6_4
#define init_9_6_6_4
void FASTCALL call_9_6_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_6_5
#define init_9_6_6_5
void FASTCALL call_9_6_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_6_6
#define init_9_6_6_6
void FASTCALL call_9_6_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_6_7
#define init_9_6_6_7
void FASTCALL call_9_6_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_6_8
#define init_9_6_6_8
void FASTCALL call_9_6_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_6_9
#define init_9_6_6_9
void FASTCALL call_9_6_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_6_10
#define init_9_6_6_10
void FASTCALL call_9_6_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_6_11
#define init_9_6_6_11
void FASTCALL call_9_6_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_6_12
#define init_9_6_6_12
void FASTCALL call_9_6_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_6_13
#define init_9_6_6_13
void FASTCALL call_9_6_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_6_14
#define init_9_6_6_14
void FASTCALL call_9_6_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_6_15
#define init_9_6_6_15
void FASTCALL call_9_6_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_7_0
#define init_9_6_7_0
void FASTCALL call_9_6_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_7_1
#define init_9_6_7_1
void FASTCALL call_9_6_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_7_2
#define init_9_6_7_2
void FASTCALL call_9_6_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_7_3
#define init_9_6_7_3
void FASTCALL call_9_6_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_7_4
#define init_9_6_7_4
void FASTCALL call_9_6_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_7_5
#define init_9_6_7_5
void FASTCALL call_9_6_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_7_6
#define init_9_6_7_6
void FASTCALL call_9_6_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_7_7
#define init_9_6_7_7
void FASTCALL call_9_6_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_7_8
#define init_9_6_7_8
void FASTCALL call_9_6_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_7_9
#define init_9_6_7_9
void FASTCALL call_9_6_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_7_10
#define init_9_6_7_10
void FASTCALL call_9_6_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_7_11
#define init_9_6_7_11
void FASTCALL call_9_6_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_7_12
#define init_9_6_7_12
void FASTCALL call_9_6_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_7_13
#define init_9_6_7_13
void FASTCALL call_9_6_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_7_14
#define init_9_6_7_14
void FASTCALL call_9_6_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_7_15
#define init_9_6_7_15
void FASTCALL call_9_6_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_8_0
#define init_9_6_8_0
void FASTCALL call_9_6_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_8_1
#define init_9_6_8_1
void FASTCALL call_9_6_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_8_2
#define init_9_6_8_2
void FASTCALL call_9_6_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_8_3
#define init_9_6_8_3
void FASTCALL call_9_6_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_8_4
#define init_9_6_8_4
void FASTCALL call_9_6_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_8_5
#define init_9_6_8_5
void FASTCALL call_9_6_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_8_6
#define init_9_6_8_6
void FASTCALL call_9_6_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_8_7
#define init_9_6_8_7
void FASTCALL call_9_6_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_8_8
#define init_9_6_8_8
void FASTCALL call_9_6_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_8_9
#define init_9_6_8_9
void FASTCALL call_9_6_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_8_10
#define init_9_6_8_10
void FASTCALL call_9_6_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_8_11
#define init_9_6_8_11
void FASTCALL call_9_6_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_8_12
#define init_9_6_8_12
void FASTCALL call_9_6_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_8_13
#define init_9_6_8_13
void FASTCALL call_9_6_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_8_14
#define init_9_6_8_14
void FASTCALL call_9_6_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_8_15
#define init_9_6_8_15
void FASTCALL call_9_6_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_9_0
#define init_9_6_9_0
void FASTCALL call_9_6_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_9_1
#define init_9_6_9_1
void FASTCALL call_9_6_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_9_2
#define init_9_6_9_2
void FASTCALL call_9_6_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_9_3
#define init_9_6_9_3
void FASTCALL call_9_6_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_9_4
#define init_9_6_9_4
void FASTCALL call_9_6_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_9_5
#define init_9_6_9_5
void FASTCALL call_9_6_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_9_6
#define init_9_6_9_6
void FASTCALL call_9_6_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_9_7
#define init_9_6_9_7
void FASTCALL call_9_6_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_9_8
#define init_9_6_9_8
void FASTCALL call_9_6_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_9_9
#define init_9_6_9_9
void FASTCALL call_9_6_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_9_10
#define init_9_6_9_10
void FASTCALL call_9_6_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_9_11
#define init_9_6_9_11
void FASTCALL call_9_6_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_9_12
#define init_9_6_9_12
void FASTCALL call_9_6_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_9_13
#define init_9_6_9_13
void FASTCALL call_9_6_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_9_14
#define init_9_6_9_14
void FASTCALL call_9_6_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_9_15
#define init_9_6_9_15
void FASTCALL call_9_6_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_10_0
#define init_9_6_10_0
void FASTCALL call_9_6_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_10_1
#define init_9_6_10_1
void FASTCALL call_9_6_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_10_2
#define init_9_6_10_2
void FASTCALL call_9_6_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_10_3
#define init_9_6_10_3
void FASTCALL call_9_6_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_10_4
#define init_9_6_10_4
void FASTCALL call_9_6_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_10_5
#define init_9_6_10_5
void FASTCALL call_9_6_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_10_6
#define init_9_6_10_6
void FASTCALL call_9_6_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_10_7
#define init_9_6_10_7
void FASTCALL call_9_6_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_10_8
#define init_9_6_10_8
void FASTCALL call_9_6_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_10_9
#define init_9_6_10_9
void FASTCALL call_9_6_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_10_10
#define init_9_6_10_10
void FASTCALL call_9_6_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_10_11
#define init_9_6_10_11
void FASTCALL call_9_6_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_10_12
#define init_9_6_10_12
void FASTCALL call_9_6_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_10_13
#define init_9_6_10_13
void FASTCALL call_9_6_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_10_14
#define init_9_6_10_14
void FASTCALL call_9_6_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_10_15
#define init_9_6_10_15
void FASTCALL call_9_6_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_11_0
#define init_9_6_11_0
void FASTCALL call_9_6_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_11_1
#define init_9_6_11_1
void FASTCALL call_9_6_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_11_2
#define init_9_6_11_2
void FASTCALL call_9_6_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_11_3
#define init_9_6_11_3
void FASTCALL call_9_6_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_11_4
#define init_9_6_11_4
void FASTCALL call_9_6_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_11_5
#define init_9_6_11_5
void FASTCALL call_9_6_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_11_6
#define init_9_6_11_6
void FASTCALL call_9_6_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_11_7
#define init_9_6_11_7
void FASTCALL call_9_6_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_11_8
#define init_9_6_11_8
void FASTCALL call_9_6_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_11_9
#define init_9_6_11_9
void FASTCALL call_9_6_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_11_10
#define init_9_6_11_10
void FASTCALL call_9_6_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_11_11
#define init_9_6_11_11
void FASTCALL call_9_6_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_11_12
#define init_9_6_11_12
void FASTCALL call_9_6_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_11_13
#define init_9_6_11_13
void FASTCALL call_9_6_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_11_14
#define init_9_6_11_14
void FASTCALL call_9_6_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_11_15
#define init_9_6_11_15
void FASTCALL call_9_6_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_12_0
#define init_9_6_12_0
void FASTCALL call_9_6_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_12_1
#define init_9_6_12_1
void FASTCALL call_9_6_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_12_2
#define init_9_6_12_2
void FASTCALL call_9_6_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_12_3
#define init_9_6_12_3
void FASTCALL call_9_6_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_12_4
#define init_9_6_12_4
void FASTCALL call_9_6_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_12_5
#define init_9_6_12_5
void FASTCALL call_9_6_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_12_6
#define init_9_6_12_6
void FASTCALL call_9_6_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_12_7
#define init_9_6_12_7
void FASTCALL call_9_6_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_12_8
#define init_9_6_12_8
void FASTCALL call_9_6_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_12_9
#define init_9_6_12_9
void FASTCALL call_9_6_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_12_10
#define init_9_6_12_10
void FASTCALL call_9_6_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_12_11
#define init_9_6_12_11
void FASTCALL call_9_6_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_12_12
#define init_9_6_12_12
void FASTCALL call_9_6_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_12_13
#define init_9_6_12_13
void FASTCALL call_9_6_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_12_14
#define init_9_6_12_14
void FASTCALL call_9_6_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_12_15
#define init_9_6_12_15
void FASTCALL call_9_6_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_13_0
#define init_9_6_13_0
void FASTCALL call_9_6_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_13_1
#define init_9_6_13_1
void FASTCALL call_9_6_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_13_2
#define init_9_6_13_2
void FASTCALL call_9_6_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_13_3
#define init_9_6_13_3
void FASTCALL call_9_6_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_13_4
#define init_9_6_13_4
void FASTCALL call_9_6_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_13_5
#define init_9_6_13_5
void FASTCALL call_9_6_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_13_6
#define init_9_6_13_6
void FASTCALL call_9_6_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_13_7
#define init_9_6_13_7
void FASTCALL call_9_6_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_13_8
#define init_9_6_13_8
void FASTCALL call_9_6_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_13_9
#define init_9_6_13_9
void FASTCALL call_9_6_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_13_10
#define init_9_6_13_10
void FASTCALL call_9_6_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_13_11
#define init_9_6_13_11
void FASTCALL call_9_6_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_13_12
#define init_9_6_13_12
void FASTCALL call_9_6_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_13_13
#define init_9_6_13_13
void FASTCALL call_9_6_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_13_14
#define init_9_6_13_14
void FASTCALL call_9_6_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_13_15
#define init_9_6_13_15
void FASTCALL call_9_6_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_14_0
#define init_9_6_14_0
void FASTCALL call_9_6_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_14_1
#define init_9_6_14_1
void FASTCALL call_9_6_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_14_2
#define init_9_6_14_2
void FASTCALL call_9_6_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_14_3
#define init_9_6_14_3
void FASTCALL call_9_6_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_14_4
#define init_9_6_14_4
void FASTCALL call_9_6_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_14_5
#define init_9_6_14_5
void FASTCALL call_9_6_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_14_6
#define init_9_6_14_6
void FASTCALL call_9_6_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_14_7
#define init_9_6_14_7
void FASTCALL call_9_6_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_14_8
#define init_9_6_14_8
void FASTCALL call_9_6_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_14_9
#define init_9_6_14_9
void FASTCALL call_9_6_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_14_10
#define init_9_6_14_10
void FASTCALL call_9_6_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_14_11
#define init_9_6_14_11
void FASTCALL call_9_6_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_14_12
#define init_9_6_14_12
void FASTCALL call_9_6_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_14_13
#define init_9_6_14_13
void FASTCALL call_9_6_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_14_14
#define init_9_6_14_14
void FASTCALL call_9_6_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_14_15
#define init_9_6_14_15
void FASTCALL call_9_6_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_15_0
#define init_9_6_15_0
void FASTCALL call_9_6_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_15_1
#define init_9_6_15_1
void FASTCALL call_9_6_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_15_2
#define init_9_6_15_2
void FASTCALL call_9_6_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_15_3
#define init_9_6_15_3
void FASTCALL call_9_6_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_15_4
#define init_9_6_15_4
void FASTCALL call_9_6_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_15_5
#define init_9_6_15_5
void FASTCALL call_9_6_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_15_6
#define init_9_6_15_6
void FASTCALL call_9_6_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_15_7
#define init_9_6_15_7
void FASTCALL call_9_6_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_15_8
#define init_9_6_15_8
void FASTCALL call_9_6_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_15_9
#define init_9_6_15_9
void FASTCALL call_9_6_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_15_10
#define init_9_6_15_10
void FASTCALL call_9_6_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_15_11
#define init_9_6_15_11
void FASTCALL call_9_6_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_15_12
#define init_9_6_15_12
void FASTCALL call_9_6_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_15_13
#define init_9_6_15_13
void FASTCALL call_9_6_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_15_14
#define init_9_6_15_14
void FASTCALL call_9_6_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_6_15_15
#define init_9_6_15_15
void FASTCALL call_9_6_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_0_0
#define init_9_7_0_0
void FASTCALL call_9_7_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_0_1
#define init_9_7_0_1
void FASTCALL call_9_7_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_0_2
#define init_9_7_0_2
void FASTCALL call_9_7_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_0_3
#define init_9_7_0_3
void FASTCALL call_9_7_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_0_4
#define init_9_7_0_4
void FASTCALL call_9_7_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_0_5
#define init_9_7_0_5
void FASTCALL call_9_7_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_0_6
#define init_9_7_0_6
void FASTCALL call_9_7_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_0_7
#define init_9_7_0_7
void FASTCALL call_9_7_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_0_8
#define init_9_7_0_8
void FASTCALL call_9_7_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_0_9
#define init_9_7_0_9
void FASTCALL call_9_7_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_0_10
#define init_9_7_0_10
void FASTCALL call_9_7_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_0_11
#define init_9_7_0_11
void FASTCALL call_9_7_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_0_12
#define init_9_7_0_12
void FASTCALL call_9_7_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_0_13
#define init_9_7_0_13
void FASTCALL call_9_7_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_0_14
#define init_9_7_0_14
void FASTCALL call_9_7_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_0_15
#define init_9_7_0_15
void FASTCALL call_9_7_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_1_0
#define init_9_7_1_0
void FASTCALL call_9_7_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_1_1
#define init_9_7_1_1
void FASTCALL call_9_7_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_1_2
#define init_9_7_1_2
void FASTCALL call_9_7_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_1_3
#define init_9_7_1_3
void FASTCALL call_9_7_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_1_4
#define init_9_7_1_4
void FASTCALL call_9_7_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_1_5
#define init_9_7_1_5
void FASTCALL call_9_7_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_1_6
#define init_9_7_1_6
void FASTCALL call_9_7_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_1_7
#define init_9_7_1_7
void FASTCALL call_9_7_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_1_8
#define init_9_7_1_8
void FASTCALL call_9_7_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_1_9
#define init_9_7_1_9
void FASTCALL call_9_7_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_1_10
#define init_9_7_1_10
void FASTCALL call_9_7_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_1_11
#define init_9_7_1_11
void FASTCALL call_9_7_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_1_12
#define init_9_7_1_12
void FASTCALL call_9_7_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_1_13
#define init_9_7_1_13
void FASTCALL call_9_7_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_1_14
#define init_9_7_1_14
void FASTCALL call_9_7_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_1_15
#define init_9_7_1_15
void FASTCALL call_9_7_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_2_0
#define init_9_7_2_0
void FASTCALL call_9_7_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_2_1
#define init_9_7_2_1
void FASTCALL call_9_7_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_2_2
#define init_9_7_2_2
void FASTCALL call_9_7_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_2_3
#define init_9_7_2_3
void FASTCALL call_9_7_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_2_4
#define init_9_7_2_4
void FASTCALL call_9_7_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_2_5
#define init_9_7_2_5
void FASTCALL call_9_7_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_2_6
#define init_9_7_2_6
void FASTCALL call_9_7_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_2_7
#define init_9_7_2_7
void FASTCALL call_9_7_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_2_8
#define init_9_7_2_8
void FASTCALL call_9_7_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_2_9
#define init_9_7_2_9
void FASTCALL call_9_7_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_2_10
#define init_9_7_2_10
void FASTCALL call_9_7_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_2_11
#define init_9_7_2_11
void FASTCALL call_9_7_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_2_12
#define init_9_7_2_12
void FASTCALL call_9_7_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_2_13
#define init_9_7_2_13
void FASTCALL call_9_7_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_2_14
#define init_9_7_2_14
void FASTCALL call_9_7_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_2_15
#define init_9_7_2_15
void FASTCALL call_9_7_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_3_0
#define init_9_7_3_0
void FASTCALL call_9_7_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_3_1
#define init_9_7_3_1
void FASTCALL call_9_7_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_3_2
#define init_9_7_3_2
void FASTCALL call_9_7_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_3_3
#define init_9_7_3_3
void FASTCALL call_9_7_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_3_4
#define init_9_7_3_4
void FASTCALL call_9_7_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_3_5
#define init_9_7_3_5
void FASTCALL call_9_7_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_3_6
#define init_9_7_3_6
void FASTCALL call_9_7_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_3_7
#define init_9_7_3_7
void FASTCALL call_9_7_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_3_8
#define init_9_7_3_8
void FASTCALL call_9_7_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_3_9
#define init_9_7_3_9
void FASTCALL call_9_7_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_3_10
#define init_9_7_3_10
void FASTCALL call_9_7_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_3_11
#define init_9_7_3_11
void FASTCALL call_9_7_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_3_12
#define init_9_7_3_12
void FASTCALL call_9_7_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_3_13
#define init_9_7_3_13
void FASTCALL call_9_7_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_3_14
#define init_9_7_3_14
void FASTCALL call_9_7_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_3_15
#define init_9_7_3_15
void FASTCALL call_9_7_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_4_0
#define init_9_7_4_0
void FASTCALL call_9_7_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_4_1
#define init_9_7_4_1
void FASTCALL call_9_7_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_4_2
#define init_9_7_4_2
void FASTCALL call_9_7_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_4_3
#define init_9_7_4_3
void FASTCALL call_9_7_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_4_4
#define init_9_7_4_4
void FASTCALL call_9_7_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_4_5
#define init_9_7_4_5
void FASTCALL call_9_7_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_4_6
#define init_9_7_4_6
void FASTCALL call_9_7_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_4_7
#define init_9_7_4_7
void FASTCALL call_9_7_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_4_8
#define init_9_7_4_8
void FASTCALL call_9_7_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_4_9
#define init_9_7_4_9
void FASTCALL call_9_7_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_4_10
#define init_9_7_4_10
void FASTCALL call_9_7_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_4_11
#define init_9_7_4_11
void FASTCALL call_9_7_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_4_12
#define init_9_7_4_12
void FASTCALL call_9_7_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_4_13
#define init_9_7_4_13
void FASTCALL call_9_7_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_4_14
#define init_9_7_4_14
void FASTCALL call_9_7_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_4_15
#define init_9_7_4_15
void FASTCALL call_9_7_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_5_0
#define init_9_7_5_0
void FASTCALL call_9_7_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_5_1
#define init_9_7_5_1
void FASTCALL call_9_7_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_5_2
#define init_9_7_5_2
void FASTCALL call_9_7_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_5_3
#define init_9_7_5_3
void FASTCALL call_9_7_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_5_4
#define init_9_7_5_4
void FASTCALL call_9_7_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_5_5
#define init_9_7_5_5
void FASTCALL call_9_7_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_5_6
#define init_9_7_5_6
void FASTCALL call_9_7_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_5_7
#define init_9_7_5_7
void FASTCALL call_9_7_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_5_8
#define init_9_7_5_8
void FASTCALL call_9_7_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_5_9
#define init_9_7_5_9
void FASTCALL call_9_7_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_5_10
#define init_9_7_5_10
void FASTCALL call_9_7_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_5_11
#define init_9_7_5_11
void FASTCALL call_9_7_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_5_12
#define init_9_7_5_12
void FASTCALL call_9_7_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_5_13
#define init_9_7_5_13
void FASTCALL call_9_7_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_5_14
#define init_9_7_5_14
void FASTCALL call_9_7_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_5_15
#define init_9_7_5_15
void FASTCALL call_9_7_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_6_0
#define init_9_7_6_0
void FASTCALL call_9_7_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_6_1
#define init_9_7_6_1
void FASTCALL call_9_7_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_6_2
#define init_9_7_6_2
void FASTCALL call_9_7_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_6_3
#define init_9_7_6_3
void FASTCALL call_9_7_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_6_4
#define init_9_7_6_4
void FASTCALL call_9_7_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_6_5
#define init_9_7_6_5
void FASTCALL call_9_7_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_6_6
#define init_9_7_6_6
void FASTCALL call_9_7_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_6_7
#define init_9_7_6_7
void FASTCALL call_9_7_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_6_8
#define init_9_7_6_8
void FASTCALL call_9_7_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_6_9
#define init_9_7_6_9
void FASTCALL call_9_7_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_6_10
#define init_9_7_6_10
void FASTCALL call_9_7_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_6_11
#define init_9_7_6_11
void FASTCALL call_9_7_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_6_12
#define init_9_7_6_12
void FASTCALL call_9_7_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_6_13
#define init_9_7_6_13
void FASTCALL call_9_7_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_6_14
#define init_9_7_6_14
void FASTCALL call_9_7_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_6_15
#define init_9_7_6_15
void FASTCALL call_9_7_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_7_0
#define init_9_7_7_0
void FASTCALL call_9_7_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_7_1
#define init_9_7_7_1
void FASTCALL call_9_7_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_7_2
#define init_9_7_7_2
void FASTCALL call_9_7_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_7_3
#define init_9_7_7_3
void FASTCALL call_9_7_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_7_4
#define init_9_7_7_4
void FASTCALL call_9_7_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_7_5
#define init_9_7_7_5
void FASTCALL call_9_7_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_7_6
#define init_9_7_7_6
void FASTCALL call_9_7_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_7_7
#define init_9_7_7_7
void FASTCALL call_9_7_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_7_8
#define init_9_7_7_8
void FASTCALL call_9_7_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_7_9
#define init_9_7_7_9
void FASTCALL call_9_7_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_7_10
#define init_9_7_7_10
void FASTCALL call_9_7_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_7_11
#define init_9_7_7_11
void FASTCALL call_9_7_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_7_12
#define init_9_7_7_12
void FASTCALL call_9_7_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_7_13
#define init_9_7_7_13
void FASTCALL call_9_7_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_7_14
#define init_9_7_7_14
void FASTCALL call_9_7_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_7_15
#define init_9_7_7_15
void FASTCALL call_9_7_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_8_0
#define init_9_7_8_0
void FASTCALL call_9_7_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_8_1
#define init_9_7_8_1
void FASTCALL call_9_7_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_8_2
#define init_9_7_8_2
void FASTCALL call_9_7_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_8_3
#define init_9_7_8_3
void FASTCALL call_9_7_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_8_4
#define init_9_7_8_4
void FASTCALL call_9_7_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_8_5
#define init_9_7_8_5
void FASTCALL call_9_7_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_8_6
#define init_9_7_8_6
void FASTCALL call_9_7_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_8_7
#define init_9_7_8_7
void FASTCALL call_9_7_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_8_8
#define init_9_7_8_8
void FASTCALL call_9_7_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_8_9
#define init_9_7_8_9
void FASTCALL call_9_7_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_8_10
#define init_9_7_8_10
void FASTCALL call_9_7_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_8_11
#define init_9_7_8_11
void FASTCALL call_9_7_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_8_12
#define init_9_7_8_12
void FASTCALL call_9_7_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_8_13
#define init_9_7_8_13
void FASTCALL call_9_7_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_8_14
#define init_9_7_8_14
void FASTCALL call_9_7_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_8_15
#define init_9_7_8_15
void FASTCALL call_9_7_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_9_0
#define init_9_7_9_0
void FASTCALL call_9_7_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_9_1
#define init_9_7_9_1
void FASTCALL call_9_7_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_9_2
#define init_9_7_9_2
void FASTCALL call_9_7_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_9_3
#define init_9_7_9_3
void FASTCALL call_9_7_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_9_4
#define init_9_7_9_4
void FASTCALL call_9_7_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_9_5
#define init_9_7_9_5
void FASTCALL call_9_7_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_9_6
#define init_9_7_9_6
void FASTCALL call_9_7_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_9_7
#define init_9_7_9_7
void FASTCALL call_9_7_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_9_8
#define init_9_7_9_8
void FASTCALL call_9_7_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_9_9
#define init_9_7_9_9
void FASTCALL call_9_7_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_9_10
#define init_9_7_9_10
void FASTCALL call_9_7_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_9_11
#define init_9_7_9_11
void FASTCALL call_9_7_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_9_12
#define init_9_7_9_12
void FASTCALL call_9_7_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_9_13
#define init_9_7_9_13
void FASTCALL call_9_7_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_9_14
#define init_9_7_9_14
void FASTCALL call_9_7_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_9_15
#define init_9_7_9_15
void FASTCALL call_9_7_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_10_0
#define init_9_7_10_0
void FASTCALL call_9_7_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_10_1
#define init_9_7_10_1
void FASTCALL call_9_7_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_10_2
#define init_9_7_10_2
void FASTCALL call_9_7_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_10_3
#define init_9_7_10_3
void FASTCALL call_9_7_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_10_4
#define init_9_7_10_4
void FASTCALL call_9_7_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_10_5
#define init_9_7_10_5
void FASTCALL call_9_7_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_10_6
#define init_9_7_10_6
void FASTCALL call_9_7_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_10_7
#define init_9_7_10_7
void FASTCALL call_9_7_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_10_8
#define init_9_7_10_8
void FASTCALL call_9_7_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_10_9
#define init_9_7_10_9
void FASTCALL call_9_7_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_10_10
#define init_9_7_10_10
void FASTCALL call_9_7_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_10_11
#define init_9_7_10_11
void FASTCALL call_9_7_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_10_12
#define init_9_7_10_12
void FASTCALL call_9_7_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_10_13
#define init_9_7_10_13
void FASTCALL call_9_7_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_10_14
#define init_9_7_10_14
void FASTCALL call_9_7_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_10_15
#define init_9_7_10_15
void FASTCALL call_9_7_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_11_0
#define init_9_7_11_0
void FASTCALL call_9_7_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_11_1
#define init_9_7_11_1
void FASTCALL call_9_7_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_11_2
#define init_9_7_11_2
void FASTCALL call_9_7_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_11_3
#define init_9_7_11_3
void FASTCALL call_9_7_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_11_4
#define init_9_7_11_4
void FASTCALL call_9_7_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_11_5
#define init_9_7_11_5
void FASTCALL call_9_7_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_11_6
#define init_9_7_11_6
void FASTCALL call_9_7_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_11_7
#define init_9_7_11_7
void FASTCALL call_9_7_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_11_8
#define init_9_7_11_8
void FASTCALL call_9_7_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_11_9
#define init_9_7_11_9
void FASTCALL call_9_7_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_11_10
#define init_9_7_11_10
void FASTCALL call_9_7_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_11_11
#define init_9_7_11_11
void FASTCALL call_9_7_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_11_12
#define init_9_7_11_12
void FASTCALL call_9_7_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_11_13
#define init_9_7_11_13
void FASTCALL call_9_7_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_11_14
#define init_9_7_11_14
void FASTCALL call_9_7_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_11_15
#define init_9_7_11_15
void FASTCALL call_9_7_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_12_0
#define init_9_7_12_0
void FASTCALL call_9_7_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_12_1
#define init_9_7_12_1
void FASTCALL call_9_7_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_12_2
#define init_9_7_12_2
void FASTCALL call_9_7_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_12_3
#define init_9_7_12_3
void FASTCALL call_9_7_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_12_4
#define init_9_7_12_4
void FASTCALL call_9_7_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_12_5
#define init_9_7_12_5
void FASTCALL call_9_7_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_12_6
#define init_9_7_12_6
void FASTCALL call_9_7_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_12_7
#define init_9_7_12_7
void FASTCALL call_9_7_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_12_8
#define init_9_7_12_8
void FASTCALL call_9_7_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_12_9
#define init_9_7_12_9
void FASTCALL call_9_7_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_12_10
#define init_9_7_12_10
void FASTCALL call_9_7_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_12_11
#define init_9_7_12_11
void FASTCALL call_9_7_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_12_12
#define init_9_7_12_12
void FASTCALL call_9_7_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_12_13
#define init_9_7_12_13
void FASTCALL call_9_7_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_12_14
#define init_9_7_12_14
void FASTCALL call_9_7_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_12_15
#define init_9_7_12_15
void FASTCALL call_9_7_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_13_0
#define init_9_7_13_0
void FASTCALL call_9_7_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_13_1
#define init_9_7_13_1
void FASTCALL call_9_7_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_13_2
#define init_9_7_13_2
void FASTCALL call_9_7_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_13_3
#define init_9_7_13_3
void FASTCALL call_9_7_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_13_4
#define init_9_7_13_4
void FASTCALL call_9_7_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_13_5
#define init_9_7_13_5
void FASTCALL call_9_7_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_13_6
#define init_9_7_13_6
void FASTCALL call_9_7_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_13_7
#define init_9_7_13_7
void FASTCALL call_9_7_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_13_8
#define init_9_7_13_8
void FASTCALL call_9_7_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_13_9
#define init_9_7_13_9
void FASTCALL call_9_7_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_13_10
#define init_9_7_13_10
void FASTCALL call_9_7_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_13_11
#define init_9_7_13_11
void FASTCALL call_9_7_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_13_12
#define init_9_7_13_12
void FASTCALL call_9_7_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_13_13
#define init_9_7_13_13
void FASTCALL call_9_7_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_13_14
#define init_9_7_13_14
void FASTCALL call_9_7_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_13_15
#define init_9_7_13_15
void FASTCALL call_9_7_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_14_0
#define init_9_7_14_0
void FASTCALL call_9_7_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_14_1
#define init_9_7_14_1
void FASTCALL call_9_7_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_14_2
#define init_9_7_14_2
void FASTCALL call_9_7_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_14_3
#define init_9_7_14_3
void FASTCALL call_9_7_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_14_4
#define init_9_7_14_4
void FASTCALL call_9_7_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_14_5
#define init_9_7_14_5
void FASTCALL call_9_7_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_14_6
#define init_9_7_14_6
void FASTCALL call_9_7_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_14_7
#define init_9_7_14_7
void FASTCALL call_9_7_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_14_8
#define init_9_7_14_8
void FASTCALL call_9_7_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_14_9
#define init_9_7_14_9
void FASTCALL call_9_7_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_14_10
#define init_9_7_14_10
void FASTCALL call_9_7_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_14_11
#define init_9_7_14_11
void FASTCALL call_9_7_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_14_12
#define init_9_7_14_12
void FASTCALL call_9_7_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_14_13
#define init_9_7_14_13
void FASTCALL call_9_7_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_14_14
#define init_9_7_14_14
void FASTCALL call_9_7_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_14_15
#define init_9_7_14_15
void FASTCALL call_9_7_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_15_0
#define init_9_7_15_0
void FASTCALL call_9_7_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_15_1
#define init_9_7_15_1
void FASTCALL call_9_7_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_15_2
#define init_9_7_15_2
void FASTCALL call_9_7_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_15_3
#define init_9_7_15_3
void FASTCALL call_9_7_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_15_4
#define init_9_7_15_4
void FASTCALL call_9_7_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_15_5
#define init_9_7_15_5
void FASTCALL call_9_7_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_15_6
#define init_9_7_15_6
void FASTCALL call_9_7_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_15_7
#define init_9_7_15_7
void FASTCALL call_9_7_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_15_8
#define init_9_7_15_8
void FASTCALL call_9_7_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_15_9
#define init_9_7_15_9
void FASTCALL call_9_7_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_15_10
#define init_9_7_15_10
void FASTCALL call_9_7_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_15_11
#define init_9_7_15_11
void FASTCALL call_9_7_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_15_12
#define init_9_7_15_12
void FASTCALL call_9_7_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_15_13
#define init_9_7_15_13
void FASTCALL call_9_7_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_15_14
#define init_9_7_15_14
void FASTCALL call_9_7_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_7_15_15
#define init_9_7_15_15
void FASTCALL call_9_7_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_0_0
#define init_9_8_0_0
void FASTCALL call_9_8_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_0_1
#define init_9_8_0_1
void FASTCALL call_9_8_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_0_2
#define init_9_8_0_2
void FASTCALL call_9_8_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_0_3
#define init_9_8_0_3
void FASTCALL call_9_8_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_0_4
#define init_9_8_0_4
void FASTCALL call_9_8_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_0_5
#define init_9_8_0_5
void FASTCALL call_9_8_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_0_6
#define init_9_8_0_6
void FASTCALL call_9_8_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_0_7
#define init_9_8_0_7
void FASTCALL call_9_8_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_0_8
#define init_9_8_0_8
void FASTCALL call_9_8_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_0_9
#define init_9_8_0_9
void FASTCALL call_9_8_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_0_10
#define init_9_8_0_10
void FASTCALL call_9_8_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_0_11
#define init_9_8_0_11
void FASTCALL call_9_8_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_0_12
#define init_9_8_0_12
void FASTCALL call_9_8_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_0_13
#define init_9_8_0_13
void FASTCALL call_9_8_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_0_14
#define init_9_8_0_14
void FASTCALL call_9_8_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_0_15
#define init_9_8_0_15
void FASTCALL call_9_8_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_1_0
#define init_9_8_1_0
void FASTCALL call_9_8_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_1_1
#define init_9_8_1_1
void FASTCALL call_9_8_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_1_2
#define init_9_8_1_2
void FASTCALL call_9_8_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_1_3
#define init_9_8_1_3
void FASTCALL call_9_8_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_1_4
#define init_9_8_1_4
void FASTCALL call_9_8_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_1_5
#define init_9_8_1_5
void FASTCALL call_9_8_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_1_6
#define init_9_8_1_6
void FASTCALL call_9_8_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_1_7
#define init_9_8_1_7
void FASTCALL call_9_8_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_1_8
#define init_9_8_1_8
void FASTCALL call_9_8_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_1_9
#define init_9_8_1_9
void FASTCALL call_9_8_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_1_10
#define init_9_8_1_10
void FASTCALL call_9_8_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_1_11
#define init_9_8_1_11
void FASTCALL call_9_8_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_1_12
#define init_9_8_1_12
void FASTCALL call_9_8_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_1_13
#define init_9_8_1_13
void FASTCALL call_9_8_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_1_14
#define init_9_8_1_14
void FASTCALL call_9_8_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_1_15
#define init_9_8_1_15
void FASTCALL call_9_8_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_2_0
#define init_9_8_2_0
void FASTCALL call_9_8_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_2_1
#define init_9_8_2_1
void FASTCALL call_9_8_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_2_2
#define init_9_8_2_2
void FASTCALL call_9_8_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_2_3
#define init_9_8_2_3
void FASTCALL call_9_8_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_2_4
#define init_9_8_2_4
void FASTCALL call_9_8_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_2_5
#define init_9_8_2_5
void FASTCALL call_9_8_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_2_6
#define init_9_8_2_6
void FASTCALL call_9_8_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_2_7
#define init_9_8_2_7
void FASTCALL call_9_8_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_2_8
#define init_9_8_2_8
void FASTCALL call_9_8_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_2_9
#define init_9_8_2_9
void FASTCALL call_9_8_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_2_10
#define init_9_8_2_10
void FASTCALL call_9_8_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_2_11
#define init_9_8_2_11
void FASTCALL call_9_8_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_2_12
#define init_9_8_2_12
void FASTCALL call_9_8_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_2_13
#define init_9_8_2_13
void FASTCALL call_9_8_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_2_14
#define init_9_8_2_14
void FASTCALL call_9_8_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_2_15
#define init_9_8_2_15
void FASTCALL call_9_8_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_3_0
#define init_9_8_3_0
void FASTCALL call_9_8_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_3_1
#define init_9_8_3_1
void FASTCALL call_9_8_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_3_2
#define init_9_8_3_2
void FASTCALL call_9_8_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_3_3
#define init_9_8_3_3
void FASTCALL call_9_8_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_3_4
#define init_9_8_3_4
void FASTCALL call_9_8_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_3_5
#define init_9_8_3_5
void FASTCALL call_9_8_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_3_6
#define init_9_8_3_6
void FASTCALL call_9_8_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_3_7
#define init_9_8_3_7
void FASTCALL call_9_8_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_3_8
#define init_9_8_3_8
void FASTCALL call_9_8_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_3_9
#define init_9_8_3_9
void FASTCALL call_9_8_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_3_10
#define init_9_8_3_10
void FASTCALL call_9_8_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_3_11
#define init_9_8_3_11
void FASTCALL call_9_8_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_3_12
#define init_9_8_3_12
void FASTCALL call_9_8_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_3_13
#define init_9_8_3_13
void FASTCALL call_9_8_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_3_14
#define init_9_8_3_14
void FASTCALL call_9_8_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_3_15
#define init_9_8_3_15
void FASTCALL call_9_8_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_4_0
#define init_9_8_4_0
void FASTCALL call_9_8_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_4_1
#define init_9_8_4_1
void FASTCALL call_9_8_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_4_2
#define init_9_8_4_2
void FASTCALL call_9_8_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_4_3
#define init_9_8_4_3
void FASTCALL call_9_8_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_4_4
#define init_9_8_4_4
void FASTCALL call_9_8_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_4_5
#define init_9_8_4_5
void FASTCALL call_9_8_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_4_6
#define init_9_8_4_6
void FASTCALL call_9_8_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_4_7
#define init_9_8_4_7
void FASTCALL call_9_8_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_4_8
#define init_9_8_4_8
void FASTCALL call_9_8_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_4_9
#define init_9_8_4_9
void FASTCALL call_9_8_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_4_10
#define init_9_8_4_10
void FASTCALL call_9_8_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_4_11
#define init_9_8_4_11
void FASTCALL call_9_8_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_4_12
#define init_9_8_4_12
void FASTCALL call_9_8_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_4_13
#define init_9_8_4_13
void FASTCALL call_9_8_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_4_14
#define init_9_8_4_14
void FASTCALL call_9_8_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_4_15
#define init_9_8_4_15
void FASTCALL call_9_8_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_5_0
#define init_9_8_5_0
void FASTCALL call_9_8_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_5_1
#define init_9_8_5_1
void FASTCALL call_9_8_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_5_2
#define init_9_8_5_2
void FASTCALL call_9_8_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_5_3
#define init_9_8_5_3
void FASTCALL call_9_8_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_5_4
#define init_9_8_5_4
void FASTCALL call_9_8_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_5_5
#define init_9_8_5_5
void FASTCALL call_9_8_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_5_6
#define init_9_8_5_6
void FASTCALL call_9_8_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_5_7
#define init_9_8_5_7
void FASTCALL call_9_8_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_5_8
#define init_9_8_5_8
void FASTCALL call_9_8_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_5_9
#define init_9_8_5_9
void FASTCALL call_9_8_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_5_10
#define init_9_8_5_10
void FASTCALL call_9_8_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_5_11
#define init_9_8_5_11
void FASTCALL call_9_8_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_5_12
#define init_9_8_5_12
void FASTCALL call_9_8_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_5_13
#define init_9_8_5_13
void FASTCALL call_9_8_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_5_14
#define init_9_8_5_14
void FASTCALL call_9_8_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_5_15
#define init_9_8_5_15
void FASTCALL call_9_8_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_6_0
#define init_9_8_6_0
void FASTCALL call_9_8_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_6_1
#define init_9_8_6_1
void FASTCALL call_9_8_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_6_2
#define init_9_8_6_2
void FASTCALL call_9_8_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_6_3
#define init_9_8_6_3
void FASTCALL call_9_8_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_6_4
#define init_9_8_6_4
void FASTCALL call_9_8_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_6_5
#define init_9_8_6_5
void FASTCALL call_9_8_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_6_6
#define init_9_8_6_6
void FASTCALL call_9_8_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_6_7
#define init_9_8_6_7
void FASTCALL call_9_8_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_6_8
#define init_9_8_6_8
void FASTCALL call_9_8_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_6_9
#define init_9_8_6_9
void FASTCALL call_9_8_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_6_10
#define init_9_8_6_10
void FASTCALL call_9_8_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_6_11
#define init_9_8_6_11
void FASTCALL call_9_8_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_6_12
#define init_9_8_6_12
void FASTCALL call_9_8_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_6_13
#define init_9_8_6_13
void FASTCALL call_9_8_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_6_14
#define init_9_8_6_14
void FASTCALL call_9_8_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_6_15
#define init_9_8_6_15
void FASTCALL call_9_8_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_7_0
#define init_9_8_7_0
void FASTCALL call_9_8_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_7_1
#define init_9_8_7_1
void FASTCALL call_9_8_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_7_2
#define init_9_8_7_2
void FASTCALL call_9_8_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_7_3
#define init_9_8_7_3
void FASTCALL call_9_8_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_7_4
#define init_9_8_7_4
void FASTCALL call_9_8_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_7_5
#define init_9_8_7_5
void FASTCALL call_9_8_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_7_6
#define init_9_8_7_6
void FASTCALL call_9_8_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_7_7
#define init_9_8_7_7
void FASTCALL call_9_8_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_7_8
#define init_9_8_7_8
void FASTCALL call_9_8_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_7_9
#define init_9_8_7_9
void FASTCALL call_9_8_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_7_10
#define init_9_8_7_10
void FASTCALL call_9_8_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_7_11
#define init_9_8_7_11
void FASTCALL call_9_8_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_7_12
#define init_9_8_7_12
void FASTCALL call_9_8_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_7_13
#define init_9_8_7_13
void FASTCALL call_9_8_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_7_14
#define init_9_8_7_14
void FASTCALL call_9_8_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_7_15
#define init_9_8_7_15
void FASTCALL call_9_8_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_8_0
#define init_9_8_8_0
void FASTCALL call_9_8_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_8_1
#define init_9_8_8_1
void FASTCALL call_9_8_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_8_2
#define init_9_8_8_2
void FASTCALL call_9_8_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_8_3
#define init_9_8_8_3
void FASTCALL call_9_8_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_8_4
#define init_9_8_8_4
void FASTCALL call_9_8_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_8_5
#define init_9_8_8_5
void FASTCALL call_9_8_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_8_6
#define init_9_8_8_6
void FASTCALL call_9_8_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_8_7
#define init_9_8_8_7
void FASTCALL call_9_8_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_8_8
#define init_9_8_8_8
void FASTCALL call_9_8_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_8_9
#define init_9_8_8_9
void FASTCALL call_9_8_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_8_10
#define init_9_8_8_10
void FASTCALL call_9_8_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_8_11
#define init_9_8_8_11
void FASTCALL call_9_8_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_8_12
#define init_9_8_8_12
void FASTCALL call_9_8_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_8_13
#define init_9_8_8_13
void FASTCALL call_9_8_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_8_14
#define init_9_8_8_14
void FASTCALL call_9_8_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_8_15
#define init_9_8_8_15
void FASTCALL call_9_8_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_9_0
#define init_9_8_9_0
void FASTCALL call_9_8_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_9_1
#define init_9_8_9_1
void FASTCALL call_9_8_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_9_2
#define init_9_8_9_2
void FASTCALL call_9_8_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_9_3
#define init_9_8_9_3
void FASTCALL call_9_8_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_9_4
#define init_9_8_9_4
void FASTCALL call_9_8_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_9_5
#define init_9_8_9_5
void FASTCALL call_9_8_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_9_6
#define init_9_8_9_6
void FASTCALL call_9_8_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_9_7
#define init_9_8_9_7
void FASTCALL call_9_8_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_9_8
#define init_9_8_9_8
void FASTCALL call_9_8_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_9_9
#define init_9_8_9_9
void FASTCALL call_9_8_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_9_10
#define init_9_8_9_10
void FASTCALL call_9_8_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_9_11
#define init_9_8_9_11
void FASTCALL call_9_8_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_9_12
#define init_9_8_9_12
void FASTCALL call_9_8_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_9_13
#define init_9_8_9_13
void FASTCALL call_9_8_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_9_14
#define init_9_8_9_14
void FASTCALL call_9_8_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_9_15
#define init_9_8_9_15
void FASTCALL call_9_8_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_10_0
#define init_9_8_10_0
void FASTCALL call_9_8_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_10_1
#define init_9_8_10_1
void FASTCALL call_9_8_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_10_2
#define init_9_8_10_2
void FASTCALL call_9_8_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_10_3
#define init_9_8_10_3
void FASTCALL call_9_8_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_10_4
#define init_9_8_10_4
void FASTCALL call_9_8_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_10_5
#define init_9_8_10_5
void FASTCALL call_9_8_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_10_6
#define init_9_8_10_6
void FASTCALL call_9_8_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_10_7
#define init_9_8_10_7
void FASTCALL call_9_8_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_10_8
#define init_9_8_10_8
void FASTCALL call_9_8_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_10_9
#define init_9_8_10_9
void FASTCALL call_9_8_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_10_10
#define init_9_8_10_10
void FASTCALL call_9_8_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_10_11
#define init_9_8_10_11
void FASTCALL call_9_8_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_10_12
#define init_9_8_10_12
void FASTCALL call_9_8_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_10_13
#define init_9_8_10_13
void FASTCALL call_9_8_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_10_14
#define init_9_8_10_14
void FASTCALL call_9_8_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_10_15
#define init_9_8_10_15
void FASTCALL call_9_8_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_11_0
#define init_9_8_11_0
void FASTCALL call_9_8_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_11_1
#define init_9_8_11_1
void FASTCALL call_9_8_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_11_2
#define init_9_8_11_2
void FASTCALL call_9_8_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_11_3
#define init_9_8_11_3
void FASTCALL call_9_8_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_11_4
#define init_9_8_11_4
void FASTCALL call_9_8_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_11_5
#define init_9_8_11_5
void FASTCALL call_9_8_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_11_6
#define init_9_8_11_6
void FASTCALL call_9_8_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_11_7
#define init_9_8_11_7
void FASTCALL call_9_8_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_11_8
#define init_9_8_11_8
void FASTCALL call_9_8_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_11_9
#define init_9_8_11_9
void FASTCALL call_9_8_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_11_10
#define init_9_8_11_10
void FASTCALL call_9_8_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_11_11
#define init_9_8_11_11
void FASTCALL call_9_8_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_11_12
#define init_9_8_11_12
void FASTCALL call_9_8_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_11_13
#define init_9_8_11_13
void FASTCALL call_9_8_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_11_14
#define init_9_8_11_14
void FASTCALL call_9_8_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_11_15
#define init_9_8_11_15
void FASTCALL call_9_8_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_12_0
#define init_9_8_12_0
void FASTCALL call_9_8_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_12_1
#define init_9_8_12_1
void FASTCALL call_9_8_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_12_2
#define init_9_8_12_2
void FASTCALL call_9_8_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_12_3
#define init_9_8_12_3
void FASTCALL call_9_8_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_12_4
#define init_9_8_12_4
void FASTCALL call_9_8_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_12_5
#define init_9_8_12_5
void FASTCALL call_9_8_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_12_6
#define init_9_8_12_6
void FASTCALL call_9_8_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_12_7
#define init_9_8_12_7
void FASTCALL call_9_8_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_12_8
#define init_9_8_12_8
void FASTCALL call_9_8_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_12_9
#define init_9_8_12_9
void FASTCALL call_9_8_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_12_10
#define init_9_8_12_10
void FASTCALL call_9_8_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_12_11
#define init_9_8_12_11
void FASTCALL call_9_8_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_12_12
#define init_9_8_12_12
void FASTCALL call_9_8_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_12_13
#define init_9_8_12_13
void FASTCALL call_9_8_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_12_14
#define init_9_8_12_14
void FASTCALL call_9_8_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_12_15
#define init_9_8_12_15
void FASTCALL call_9_8_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_13_0
#define init_9_8_13_0
void FASTCALL call_9_8_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_13_1
#define init_9_8_13_1
void FASTCALL call_9_8_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_13_2
#define init_9_8_13_2
void FASTCALL call_9_8_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_13_3
#define init_9_8_13_3
void FASTCALL call_9_8_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_13_4
#define init_9_8_13_4
void FASTCALL call_9_8_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_13_5
#define init_9_8_13_5
void FASTCALL call_9_8_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_13_6
#define init_9_8_13_6
void FASTCALL call_9_8_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_13_7
#define init_9_8_13_7
void FASTCALL call_9_8_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_13_8
#define init_9_8_13_8
void FASTCALL call_9_8_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_13_9
#define init_9_8_13_9
void FASTCALL call_9_8_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_13_10
#define init_9_8_13_10
void FASTCALL call_9_8_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_13_11
#define init_9_8_13_11
void FASTCALL call_9_8_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_13_12
#define init_9_8_13_12
void FASTCALL call_9_8_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_13_13
#define init_9_8_13_13
void FASTCALL call_9_8_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_13_14
#define init_9_8_13_14
void FASTCALL call_9_8_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_13_15
#define init_9_8_13_15
void FASTCALL call_9_8_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_14_0
#define init_9_8_14_0
void FASTCALL call_9_8_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_14_1
#define init_9_8_14_1
void FASTCALL call_9_8_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_14_2
#define init_9_8_14_2
void FASTCALL call_9_8_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_14_3
#define init_9_8_14_3
void FASTCALL call_9_8_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_14_4
#define init_9_8_14_4
void FASTCALL call_9_8_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_14_5
#define init_9_8_14_5
void FASTCALL call_9_8_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_14_6
#define init_9_8_14_6
void FASTCALL call_9_8_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_14_7
#define init_9_8_14_7
void FASTCALL call_9_8_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_14_8
#define init_9_8_14_8
void FASTCALL call_9_8_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_14_9
#define init_9_8_14_9
void FASTCALL call_9_8_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_14_10
#define init_9_8_14_10
void FASTCALL call_9_8_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_14_11
#define init_9_8_14_11
void FASTCALL call_9_8_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_14_12
#define init_9_8_14_12
void FASTCALL call_9_8_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_14_13
#define init_9_8_14_13
void FASTCALL call_9_8_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_14_14
#define init_9_8_14_14
void FASTCALL call_9_8_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_14_15
#define init_9_8_14_15
void FASTCALL call_9_8_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_15_0
#define init_9_8_15_0
void FASTCALL call_9_8_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_15_1
#define init_9_8_15_1
void FASTCALL call_9_8_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_15_2
#define init_9_8_15_2
void FASTCALL call_9_8_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_15_3
#define init_9_8_15_3
void FASTCALL call_9_8_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_15_4
#define init_9_8_15_4
void FASTCALL call_9_8_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_15_5
#define init_9_8_15_5
void FASTCALL call_9_8_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_15_6
#define init_9_8_15_6
void FASTCALL call_9_8_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_15_7
#define init_9_8_15_7
void FASTCALL call_9_8_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_15_8
#define init_9_8_15_8
void FASTCALL call_9_8_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_15_9
#define init_9_8_15_9
void FASTCALL call_9_8_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_15_10
#define init_9_8_15_10
void FASTCALL call_9_8_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_15_11
#define init_9_8_15_11
void FASTCALL call_9_8_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_15_12
#define init_9_8_15_12
void FASTCALL call_9_8_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_15_13
#define init_9_8_15_13
void FASTCALL call_9_8_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_15_14
#define init_9_8_15_14
void FASTCALL call_9_8_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_8_15_15
#define init_9_8_15_15
void FASTCALL call_9_8_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_0_0
#define init_9_9_0_0
void FASTCALL call_9_9_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_0_1
#define init_9_9_0_1
void FASTCALL call_9_9_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_0_2
#define init_9_9_0_2
void FASTCALL call_9_9_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_0_3
#define init_9_9_0_3
void FASTCALL call_9_9_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_0_4
#define init_9_9_0_4
void FASTCALL call_9_9_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_0_5
#define init_9_9_0_5
void FASTCALL call_9_9_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_0_6
#define init_9_9_0_6
void FASTCALL call_9_9_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_0_7
#define init_9_9_0_7
void FASTCALL call_9_9_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_0_8
#define init_9_9_0_8
void FASTCALL call_9_9_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_0_9
#define init_9_9_0_9
void FASTCALL call_9_9_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_0_10
#define init_9_9_0_10
void FASTCALL call_9_9_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_0_11
#define init_9_9_0_11
void FASTCALL call_9_9_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_0_12
#define init_9_9_0_12
void FASTCALL call_9_9_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_0_13
#define init_9_9_0_13
void FASTCALL call_9_9_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_0_14
#define init_9_9_0_14
void FASTCALL call_9_9_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_0_15
#define init_9_9_0_15
void FASTCALL call_9_9_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_1_0
#define init_9_9_1_0
void FASTCALL call_9_9_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_1_1
#define init_9_9_1_1
void FASTCALL call_9_9_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_1_2
#define init_9_9_1_2
void FASTCALL call_9_9_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_1_3
#define init_9_9_1_3
void FASTCALL call_9_9_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_1_4
#define init_9_9_1_4
void FASTCALL call_9_9_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_1_5
#define init_9_9_1_5
void FASTCALL call_9_9_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_1_6
#define init_9_9_1_6
void FASTCALL call_9_9_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_1_7
#define init_9_9_1_7
void FASTCALL call_9_9_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_1_8
#define init_9_9_1_8
void FASTCALL call_9_9_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_1_9
#define init_9_9_1_9
void FASTCALL call_9_9_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_1_10
#define init_9_9_1_10
void FASTCALL call_9_9_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_1_11
#define init_9_9_1_11
void FASTCALL call_9_9_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_1_12
#define init_9_9_1_12
void FASTCALL call_9_9_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_1_13
#define init_9_9_1_13
void FASTCALL call_9_9_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_1_14
#define init_9_9_1_14
void FASTCALL call_9_9_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_1_15
#define init_9_9_1_15
void FASTCALL call_9_9_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_2_0
#define init_9_9_2_0
void FASTCALL call_9_9_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_2_1
#define init_9_9_2_1
void FASTCALL call_9_9_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_2_2
#define init_9_9_2_2
void FASTCALL call_9_9_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_2_3
#define init_9_9_2_3
void FASTCALL call_9_9_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_2_4
#define init_9_9_2_4
void FASTCALL call_9_9_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_2_5
#define init_9_9_2_5
void FASTCALL call_9_9_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_2_6
#define init_9_9_2_6
void FASTCALL call_9_9_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_2_7
#define init_9_9_2_7
void FASTCALL call_9_9_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_2_8
#define init_9_9_2_8
void FASTCALL call_9_9_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_2_9
#define init_9_9_2_9
void FASTCALL call_9_9_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_2_10
#define init_9_9_2_10
void FASTCALL call_9_9_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_2_11
#define init_9_9_2_11
void FASTCALL call_9_9_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_2_12
#define init_9_9_2_12
void FASTCALL call_9_9_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_2_13
#define init_9_9_2_13
void FASTCALL call_9_9_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_2_14
#define init_9_9_2_14
void FASTCALL call_9_9_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_2_15
#define init_9_9_2_15
void FASTCALL call_9_9_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_3_0
#define init_9_9_3_0
void FASTCALL call_9_9_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_3_1
#define init_9_9_3_1
void FASTCALL call_9_9_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_3_2
#define init_9_9_3_2
void FASTCALL call_9_9_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_3_3
#define init_9_9_3_3
void FASTCALL call_9_9_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_3_4
#define init_9_9_3_4
void FASTCALL call_9_9_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_3_5
#define init_9_9_3_5
void FASTCALL call_9_9_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_3_6
#define init_9_9_3_6
void FASTCALL call_9_9_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_3_7
#define init_9_9_3_7
void FASTCALL call_9_9_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_3_8
#define init_9_9_3_8
void FASTCALL call_9_9_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_3_9
#define init_9_9_3_9
void FASTCALL call_9_9_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_3_10
#define init_9_9_3_10
void FASTCALL call_9_9_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_3_11
#define init_9_9_3_11
void FASTCALL call_9_9_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_3_12
#define init_9_9_3_12
void FASTCALL call_9_9_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_3_13
#define init_9_9_3_13
void FASTCALL call_9_9_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_3_14
#define init_9_9_3_14
void FASTCALL call_9_9_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_3_15
#define init_9_9_3_15
void FASTCALL call_9_9_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_4_0
#define init_9_9_4_0
void FASTCALL call_9_9_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_4_1
#define init_9_9_4_1
void FASTCALL call_9_9_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_4_2
#define init_9_9_4_2
void FASTCALL call_9_9_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_4_3
#define init_9_9_4_3
void FASTCALL call_9_9_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_4_4
#define init_9_9_4_4
void FASTCALL call_9_9_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_4_5
#define init_9_9_4_5
void FASTCALL call_9_9_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_4_6
#define init_9_9_4_6
void FASTCALL call_9_9_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_4_7
#define init_9_9_4_7
void FASTCALL call_9_9_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_4_8
#define init_9_9_4_8
void FASTCALL call_9_9_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_4_9
#define init_9_9_4_9
void FASTCALL call_9_9_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_4_10
#define init_9_9_4_10
void FASTCALL call_9_9_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_4_11
#define init_9_9_4_11
void FASTCALL call_9_9_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_4_12
#define init_9_9_4_12
void FASTCALL call_9_9_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_4_13
#define init_9_9_4_13
void FASTCALL call_9_9_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_4_14
#define init_9_9_4_14
void FASTCALL call_9_9_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_4_15
#define init_9_9_4_15
void FASTCALL call_9_9_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_5_0
#define init_9_9_5_0
void FASTCALL call_9_9_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_5_1
#define init_9_9_5_1
void FASTCALL call_9_9_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_5_2
#define init_9_9_5_2
void FASTCALL call_9_9_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_5_3
#define init_9_9_5_3
void FASTCALL call_9_9_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_5_4
#define init_9_9_5_4
void FASTCALL call_9_9_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_5_5
#define init_9_9_5_5
void FASTCALL call_9_9_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_5_6
#define init_9_9_5_6
void FASTCALL call_9_9_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_5_7
#define init_9_9_5_7
void FASTCALL call_9_9_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_5_8
#define init_9_9_5_8
void FASTCALL call_9_9_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_5_9
#define init_9_9_5_9
void FASTCALL call_9_9_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_5_10
#define init_9_9_5_10
void FASTCALL call_9_9_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_5_11
#define init_9_9_5_11
void FASTCALL call_9_9_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_5_12
#define init_9_9_5_12
void FASTCALL call_9_9_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_5_13
#define init_9_9_5_13
void FASTCALL call_9_9_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_5_14
#define init_9_9_5_14
void FASTCALL call_9_9_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_5_15
#define init_9_9_5_15
void FASTCALL call_9_9_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_6_0
#define init_9_9_6_0
void FASTCALL call_9_9_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_6_1
#define init_9_9_6_1
void FASTCALL call_9_9_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_6_2
#define init_9_9_6_2
void FASTCALL call_9_9_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_6_3
#define init_9_9_6_3
void FASTCALL call_9_9_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_6_4
#define init_9_9_6_4
void FASTCALL call_9_9_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_6_5
#define init_9_9_6_5
void FASTCALL call_9_9_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_6_6
#define init_9_9_6_6
void FASTCALL call_9_9_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_6_7
#define init_9_9_6_7
void FASTCALL call_9_9_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_6_8
#define init_9_9_6_8
void FASTCALL call_9_9_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_6_9
#define init_9_9_6_9
void FASTCALL call_9_9_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_6_10
#define init_9_9_6_10
void FASTCALL call_9_9_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_6_11
#define init_9_9_6_11
void FASTCALL call_9_9_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_6_12
#define init_9_9_6_12
void FASTCALL call_9_9_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_6_13
#define init_9_9_6_13
void FASTCALL call_9_9_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_6_14
#define init_9_9_6_14
void FASTCALL call_9_9_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_6_15
#define init_9_9_6_15
void FASTCALL call_9_9_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_7_0
#define init_9_9_7_0
void FASTCALL call_9_9_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_7_1
#define init_9_9_7_1
void FASTCALL call_9_9_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_7_2
#define init_9_9_7_2
void FASTCALL call_9_9_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_7_3
#define init_9_9_7_3
void FASTCALL call_9_9_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_7_4
#define init_9_9_7_4
void FASTCALL call_9_9_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_7_5
#define init_9_9_7_5
void FASTCALL call_9_9_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_7_6
#define init_9_9_7_6
void FASTCALL call_9_9_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_7_7
#define init_9_9_7_7
void FASTCALL call_9_9_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_7_8
#define init_9_9_7_8
void FASTCALL call_9_9_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_7_9
#define init_9_9_7_9
void FASTCALL call_9_9_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_7_10
#define init_9_9_7_10
void FASTCALL call_9_9_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_7_11
#define init_9_9_7_11
void FASTCALL call_9_9_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_7_12
#define init_9_9_7_12
void FASTCALL call_9_9_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_7_13
#define init_9_9_7_13
void FASTCALL call_9_9_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_7_14
#define init_9_9_7_14
void FASTCALL call_9_9_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_7_15
#define init_9_9_7_15
void FASTCALL call_9_9_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_8_0
#define init_9_9_8_0
void FASTCALL call_9_9_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_8_1
#define init_9_9_8_1
void FASTCALL call_9_9_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_8_2
#define init_9_9_8_2
void FASTCALL call_9_9_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_8_3
#define init_9_9_8_3
void FASTCALL call_9_9_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_8_4
#define init_9_9_8_4
void FASTCALL call_9_9_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_8_5
#define init_9_9_8_5
void FASTCALL call_9_9_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_8_6
#define init_9_9_8_6
void FASTCALL call_9_9_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_8_7
#define init_9_9_8_7
void FASTCALL call_9_9_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_8_8
#define init_9_9_8_8
void FASTCALL call_9_9_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_8_9
#define init_9_9_8_9
void FASTCALL call_9_9_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_8_10
#define init_9_9_8_10
void FASTCALL call_9_9_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_8_11
#define init_9_9_8_11
void FASTCALL call_9_9_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_8_12
#define init_9_9_8_12
void FASTCALL call_9_9_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_8_13
#define init_9_9_8_13
void FASTCALL call_9_9_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_8_14
#define init_9_9_8_14
void FASTCALL call_9_9_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_8_15
#define init_9_9_8_15
void FASTCALL call_9_9_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_9_0
#define init_9_9_9_0
void FASTCALL call_9_9_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_9_1
#define init_9_9_9_1
void FASTCALL call_9_9_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_9_2
#define init_9_9_9_2
void FASTCALL call_9_9_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_9_3
#define init_9_9_9_3
void FASTCALL call_9_9_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_9_4
#define init_9_9_9_4
void FASTCALL call_9_9_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_9_5
#define init_9_9_9_5
void FASTCALL call_9_9_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_9_6
#define init_9_9_9_6
void FASTCALL call_9_9_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_9_7
#define init_9_9_9_7
void FASTCALL call_9_9_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_9_8
#define init_9_9_9_8
void FASTCALL call_9_9_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_9_9
#define init_9_9_9_9
void FASTCALL call_9_9_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_9_10
#define init_9_9_9_10
void FASTCALL call_9_9_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_9_11
#define init_9_9_9_11
void FASTCALL call_9_9_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_9_12
#define init_9_9_9_12
void FASTCALL call_9_9_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_9_13
#define init_9_9_9_13
void FASTCALL call_9_9_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_9_14
#define init_9_9_9_14
void FASTCALL call_9_9_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_9_15
#define init_9_9_9_15
void FASTCALL call_9_9_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_10_0
#define init_9_9_10_0
void FASTCALL call_9_9_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_10_1
#define init_9_9_10_1
void FASTCALL call_9_9_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_10_2
#define init_9_9_10_2
void FASTCALL call_9_9_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_10_3
#define init_9_9_10_3
void FASTCALL call_9_9_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_10_4
#define init_9_9_10_4
void FASTCALL call_9_9_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_10_5
#define init_9_9_10_5
void FASTCALL call_9_9_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_10_6
#define init_9_9_10_6
void FASTCALL call_9_9_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_10_7
#define init_9_9_10_7
void FASTCALL call_9_9_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_10_8
#define init_9_9_10_8
void FASTCALL call_9_9_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_10_9
#define init_9_9_10_9
void FASTCALL call_9_9_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_10_10
#define init_9_9_10_10
void FASTCALL call_9_9_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_10_11
#define init_9_9_10_11
void FASTCALL call_9_9_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_10_12
#define init_9_9_10_12
void FASTCALL call_9_9_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_10_13
#define init_9_9_10_13
void FASTCALL call_9_9_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_10_14
#define init_9_9_10_14
void FASTCALL call_9_9_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_10_15
#define init_9_9_10_15
void FASTCALL call_9_9_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_11_0
#define init_9_9_11_0
void FASTCALL call_9_9_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_11_1
#define init_9_9_11_1
void FASTCALL call_9_9_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_11_2
#define init_9_9_11_2
void FASTCALL call_9_9_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_11_3
#define init_9_9_11_3
void FASTCALL call_9_9_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_11_4
#define init_9_9_11_4
void FASTCALL call_9_9_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_11_5
#define init_9_9_11_5
void FASTCALL call_9_9_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_11_6
#define init_9_9_11_6
void FASTCALL call_9_9_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_11_7
#define init_9_9_11_7
void FASTCALL call_9_9_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_11_8
#define init_9_9_11_8
void FASTCALL call_9_9_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_11_9
#define init_9_9_11_9
void FASTCALL call_9_9_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_11_10
#define init_9_9_11_10
void FASTCALL call_9_9_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_11_11
#define init_9_9_11_11
void FASTCALL call_9_9_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_11_12
#define init_9_9_11_12
void FASTCALL call_9_9_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_11_13
#define init_9_9_11_13
void FASTCALL call_9_9_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_11_14
#define init_9_9_11_14
void FASTCALL call_9_9_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_11_15
#define init_9_9_11_15
void FASTCALL call_9_9_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_12_0
#define init_9_9_12_0
void FASTCALL call_9_9_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_12_1
#define init_9_9_12_1
void FASTCALL call_9_9_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_12_2
#define init_9_9_12_2
void FASTCALL call_9_9_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_12_3
#define init_9_9_12_3
void FASTCALL call_9_9_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_12_4
#define init_9_9_12_4
void FASTCALL call_9_9_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_12_5
#define init_9_9_12_5
void FASTCALL call_9_9_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_12_6
#define init_9_9_12_6
void FASTCALL call_9_9_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_12_7
#define init_9_9_12_7
void FASTCALL call_9_9_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_12_8
#define init_9_9_12_8
void FASTCALL call_9_9_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_12_9
#define init_9_9_12_9
void FASTCALL call_9_9_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_12_10
#define init_9_9_12_10
void FASTCALL call_9_9_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_12_11
#define init_9_9_12_11
void FASTCALL call_9_9_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_12_12
#define init_9_9_12_12
void FASTCALL call_9_9_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_12_13
#define init_9_9_12_13
void FASTCALL call_9_9_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_12_14
#define init_9_9_12_14
void FASTCALL call_9_9_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_12_15
#define init_9_9_12_15
void FASTCALL call_9_9_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_13_0
#define init_9_9_13_0
void FASTCALL call_9_9_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_13_1
#define init_9_9_13_1
void FASTCALL call_9_9_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_13_2
#define init_9_9_13_2
void FASTCALL call_9_9_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_13_3
#define init_9_9_13_3
void FASTCALL call_9_9_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_13_4
#define init_9_9_13_4
void FASTCALL call_9_9_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_13_5
#define init_9_9_13_5
void FASTCALL call_9_9_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_13_6
#define init_9_9_13_6
void FASTCALL call_9_9_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_13_7
#define init_9_9_13_7
void FASTCALL call_9_9_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_13_8
#define init_9_9_13_8
void FASTCALL call_9_9_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_13_9
#define init_9_9_13_9
void FASTCALL call_9_9_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_13_10
#define init_9_9_13_10
void FASTCALL call_9_9_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_13_11
#define init_9_9_13_11
void FASTCALL call_9_9_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_13_12
#define init_9_9_13_12
void FASTCALL call_9_9_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_13_13
#define init_9_9_13_13
void FASTCALL call_9_9_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_13_14
#define init_9_9_13_14
void FASTCALL call_9_9_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_13_15
#define init_9_9_13_15
void FASTCALL call_9_9_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_14_0
#define init_9_9_14_0
void FASTCALL call_9_9_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_14_1
#define init_9_9_14_1
void FASTCALL call_9_9_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_14_2
#define init_9_9_14_2
void FASTCALL call_9_9_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_14_3
#define init_9_9_14_3
void FASTCALL call_9_9_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_14_4
#define init_9_9_14_4
void FASTCALL call_9_9_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_14_5
#define init_9_9_14_5
void FASTCALL call_9_9_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_14_6
#define init_9_9_14_6
void FASTCALL call_9_9_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_14_7
#define init_9_9_14_7
void FASTCALL call_9_9_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_14_8
#define init_9_9_14_8
void FASTCALL call_9_9_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_14_9
#define init_9_9_14_9
void FASTCALL call_9_9_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_14_10
#define init_9_9_14_10
void FASTCALL call_9_9_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_14_11
#define init_9_9_14_11
void FASTCALL call_9_9_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_14_12
#define init_9_9_14_12
void FASTCALL call_9_9_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_14_13
#define init_9_9_14_13
void FASTCALL call_9_9_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_14_14
#define init_9_9_14_14
void FASTCALL call_9_9_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_14_15
#define init_9_9_14_15
void FASTCALL call_9_9_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_15_0
#define init_9_9_15_0
void FASTCALL call_9_9_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_15_1
#define init_9_9_15_1
void FASTCALL call_9_9_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_15_2
#define init_9_9_15_2
void FASTCALL call_9_9_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_15_3
#define init_9_9_15_3
void FASTCALL call_9_9_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_15_4
#define init_9_9_15_4
void FASTCALL call_9_9_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_15_5
#define init_9_9_15_5
void FASTCALL call_9_9_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_15_6
#define init_9_9_15_6
void FASTCALL call_9_9_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_15_7
#define init_9_9_15_7
void FASTCALL call_9_9_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_15_8
#define init_9_9_15_8
void FASTCALL call_9_9_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_15_9
#define init_9_9_15_9
void FASTCALL call_9_9_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_15_10
#define init_9_9_15_10
void FASTCALL call_9_9_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_15_11
#define init_9_9_15_11
void FASTCALL call_9_9_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_15_12
#define init_9_9_15_12
void FASTCALL call_9_9_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_15_13
#define init_9_9_15_13
void FASTCALL call_9_9_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_15_14
#define init_9_9_15_14
void FASTCALL call_9_9_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_9_15_15
#define init_9_9_15_15
void FASTCALL call_9_9_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_0_0
#define init_9_10_0_0
void FASTCALL call_9_10_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_0_1
#define init_9_10_0_1
void FASTCALL call_9_10_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_0_2
#define init_9_10_0_2
void FASTCALL call_9_10_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_0_3
#define init_9_10_0_3
void FASTCALL call_9_10_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_0_4
#define init_9_10_0_4
void FASTCALL call_9_10_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_0_5
#define init_9_10_0_5
void FASTCALL call_9_10_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_0_6
#define init_9_10_0_6
void FASTCALL call_9_10_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_0_7
#define init_9_10_0_7
void FASTCALL call_9_10_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_0_8
#define init_9_10_0_8
void FASTCALL call_9_10_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_0_9
#define init_9_10_0_9
void FASTCALL call_9_10_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_0_10
#define init_9_10_0_10
void FASTCALL call_9_10_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_0_11
#define init_9_10_0_11
void FASTCALL call_9_10_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_0_12
#define init_9_10_0_12
void FASTCALL call_9_10_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_0_13
#define init_9_10_0_13
void FASTCALL call_9_10_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_0_14
#define init_9_10_0_14
void FASTCALL call_9_10_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_0_15
#define init_9_10_0_15
void FASTCALL call_9_10_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_1_0
#define init_9_10_1_0
void FASTCALL call_9_10_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_1_1
#define init_9_10_1_1
void FASTCALL call_9_10_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_1_2
#define init_9_10_1_2
void FASTCALL call_9_10_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_1_3
#define init_9_10_1_3
void FASTCALL call_9_10_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_1_4
#define init_9_10_1_4
void FASTCALL call_9_10_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_1_5
#define init_9_10_1_5
void FASTCALL call_9_10_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_1_6
#define init_9_10_1_6
void FASTCALL call_9_10_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_1_7
#define init_9_10_1_7
void FASTCALL call_9_10_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_1_8
#define init_9_10_1_8
void FASTCALL call_9_10_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_1_9
#define init_9_10_1_9
void FASTCALL call_9_10_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_1_10
#define init_9_10_1_10
void FASTCALL call_9_10_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_1_11
#define init_9_10_1_11
void FASTCALL call_9_10_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_1_12
#define init_9_10_1_12
void FASTCALL call_9_10_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_1_13
#define init_9_10_1_13
void FASTCALL call_9_10_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_1_14
#define init_9_10_1_14
void FASTCALL call_9_10_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_1_15
#define init_9_10_1_15
void FASTCALL call_9_10_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_2_0
#define init_9_10_2_0
void FASTCALL call_9_10_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_2_1
#define init_9_10_2_1
void FASTCALL call_9_10_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_2_2
#define init_9_10_2_2
void FASTCALL call_9_10_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_2_3
#define init_9_10_2_3
void FASTCALL call_9_10_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_2_4
#define init_9_10_2_4
void FASTCALL call_9_10_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_2_5
#define init_9_10_2_5
void FASTCALL call_9_10_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_2_6
#define init_9_10_2_6
void FASTCALL call_9_10_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_2_7
#define init_9_10_2_7
void FASTCALL call_9_10_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_2_8
#define init_9_10_2_8
void FASTCALL call_9_10_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_2_9
#define init_9_10_2_9
void FASTCALL call_9_10_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_2_10
#define init_9_10_2_10
void FASTCALL call_9_10_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_2_11
#define init_9_10_2_11
void FASTCALL call_9_10_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_2_12
#define init_9_10_2_12
void FASTCALL call_9_10_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_2_13
#define init_9_10_2_13
void FASTCALL call_9_10_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_2_14
#define init_9_10_2_14
void FASTCALL call_9_10_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_2_15
#define init_9_10_2_15
void FASTCALL call_9_10_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_3_0
#define init_9_10_3_0
void FASTCALL call_9_10_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_3_1
#define init_9_10_3_1
void FASTCALL call_9_10_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_3_2
#define init_9_10_3_2
void FASTCALL call_9_10_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_3_3
#define init_9_10_3_3
void FASTCALL call_9_10_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_3_4
#define init_9_10_3_4
void FASTCALL call_9_10_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_3_5
#define init_9_10_3_5
void FASTCALL call_9_10_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_3_6
#define init_9_10_3_6
void FASTCALL call_9_10_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_3_7
#define init_9_10_3_7
void FASTCALL call_9_10_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_3_8
#define init_9_10_3_8
void FASTCALL call_9_10_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_3_9
#define init_9_10_3_9
void FASTCALL call_9_10_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_3_10
#define init_9_10_3_10
void FASTCALL call_9_10_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_3_11
#define init_9_10_3_11
void FASTCALL call_9_10_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_3_12
#define init_9_10_3_12
void FASTCALL call_9_10_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_3_13
#define init_9_10_3_13
void FASTCALL call_9_10_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_3_14
#define init_9_10_3_14
void FASTCALL call_9_10_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_3_15
#define init_9_10_3_15
void FASTCALL call_9_10_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_4_0
#define init_9_10_4_0
void FASTCALL call_9_10_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_4_1
#define init_9_10_4_1
void FASTCALL call_9_10_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_4_2
#define init_9_10_4_2
void FASTCALL call_9_10_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_4_3
#define init_9_10_4_3
void FASTCALL call_9_10_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_4_4
#define init_9_10_4_4
void FASTCALL call_9_10_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_4_5
#define init_9_10_4_5
void FASTCALL call_9_10_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_4_6
#define init_9_10_4_6
void FASTCALL call_9_10_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_4_7
#define init_9_10_4_7
void FASTCALL call_9_10_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_4_8
#define init_9_10_4_8
void FASTCALL call_9_10_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_4_9
#define init_9_10_4_9
void FASTCALL call_9_10_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_4_10
#define init_9_10_4_10
void FASTCALL call_9_10_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_4_11
#define init_9_10_4_11
void FASTCALL call_9_10_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_4_12
#define init_9_10_4_12
void FASTCALL call_9_10_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_4_13
#define init_9_10_4_13
void FASTCALL call_9_10_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_4_14
#define init_9_10_4_14
void FASTCALL call_9_10_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_4_15
#define init_9_10_4_15
void FASTCALL call_9_10_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_5_0
#define init_9_10_5_0
void FASTCALL call_9_10_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_5_1
#define init_9_10_5_1
void FASTCALL call_9_10_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_5_2
#define init_9_10_5_2
void FASTCALL call_9_10_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_5_3
#define init_9_10_5_3
void FASTCALL call_9_10_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_5_4
#define init_9_10_5_4
void FASTCALL call_9_10_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_5_5
#define init_9_10_5_5
void FASTCALL call_9_10_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_5_6
#define init_9_10_5_6
void FASTCALL call_9_10_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_5_7
#define init_9_10_5_7
void FASTCALL call_9_10_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_5_8
#define init_9_10_5_8
void FASTCALL call_9_10_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_5_9
#define init_9_10_5_9
void FASTCALL call_9_10_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_5_10
#define init_9_10_5_10
void FASTCALL call_9_10_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_5_11
#define init_9_10_5_11
void FASTCALL call_9_10_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_5_12
#define init_9_10_5_12
void FASTCALL call_9_10_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_5_13
#define init_9_10_5_13
void FASTCALL call_9_10_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_5_14
#define init_9_10_5_14
void FASTCALL call_9_10_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_5_15
#define init_9_10_5_15
void FASTCALL call_9_10_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_6_0
#define init_9_10_6_0
void FASTCALL call_9_10_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_6_1
#define init_9_10_6_1
void FASTCALL call_9_10_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_6_2
#define init_9_10_6_2
void FASTCALL call_9_10_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_6_3
#define init_9_10_6_3
void FASTCALL call_9_10_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_6_4
#define init_9_10_6_4
void FASTCALL call_9_10_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_6_5
#define init_9_10_6_5
void FASTCALL call_9_10_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_6_6
#define init_9_10_6_6
void FASTCALL call_9_10_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_6_7
#define init_9_10_6_7
void FASTCALL call_9_10_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_6_8
#define init_9_10_6_8
void FASTCALL call_9_10_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_6_9
#define init_9_10_6_9
void FASTCALL call_9_10_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_6_10
#define init_9_10_6_10
void FASTCALL call_9_10_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_6_11
#define init_9_10_6_11
void FASTCALL call_9_10_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_6_12
#define init_9_10_6_12
void FASTCALL call_9_10_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_6_13
#define init_9_10_6_13
void FASTCALL call_9_10_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_6_14
#define init_9_10_6_14
void FASTCALL call_9_10_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_6_15
#define init_9_10_6_15
void FASTCALL call_9_10_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_7_0
#define init_9_10_7_0
void FASTCALL call_9_10_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_7_1
#define init_9_10_7_1
void FASTCALL call_9_10_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_7_2
#define init_9_10_7_2
void FASTCALL call_9_10_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_7_3
#define init_9_10_7_3
void FASTCALL call_9_10_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_7_4
#define init_9_10_7_4
void FASTCALL call_9_10_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_7_5
#define init_9_10_7_5
void FASTCALL call_9_10_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_7_6
#define init_9_10_7_6
void FASTCALL call_9_10_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_7_7
#define init_9_10_7_7
void FASTCALL call_9_10_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_7_8
#define init_9_10_7_8
void FASTCALL call_9_10_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_7_9
#define init_9_10_7_9
void FASTCALL call_9_10_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_7_10
#define init_9_10_7_10
void FASTCALL call_9_10_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_7_11
#define init_9_10_7_11
void FASTCALL call_9_10_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_7_12
#define init_9_10_7_12
void FASTCALL call_9_10_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_7_13
#define init_9_10_7_13
void FASTCALL call_9_10_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_7_14
#define init_9_10_7_14
void FASTCALL call_9_10_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_7_15
#define init_9_10_7_15
void FASTCALL call_9_10_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_8_0
#define init_9_10_8_0
void FASTCALL call_9_10_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_8_1
#define init_9_10_8_1
void FASTCALL call_9_10_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_8_2
#define init_9_10_8_2
void FASTCALL call_9_10_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_8_3
#define init_9_10_8_3
void FASTCALL call_9_10_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_8_4
#define init_9_10_8_4
void FASTCALL call_9_10_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_8_5
#define init_9_10_8_5
void FASTCALL call_9_10_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_8_6
#define init_9_10_8_6
void FASTCALL call_9_10_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_8_7
#define init_9_10_8_7
void FASTCALL call_9_10_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_8_8
#define init_9_10_8_8
void FASTCALL call_9_10_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_8_9
#define init_9_10_8_9
void FASTCALL call_9_10_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_8_10
#define init_9_10_8_10
void FASTCALL call_9_10_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_8_11
#define init_9_10_8_11
void FASTCALL call_9_10_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_8_12
#define init_9_10_8_12
void FASTCALL call_9_10_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_8_13
#define init_9_10_8_13
void FASTCALL call_9_10_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_8_14
#define init_9_10_8_14
void FASTCALL call_9_10_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_8_15
#define init_9_10_8_15
void FASTCALL call_9_10_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_9_0
#define init_9_10_9_0
void FASTCALL call_9_10_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_9_1
#define init_9_10_9_1
void FASTCALL call_9_10_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_9_2
#define init_9_10_9_2
void FASTCALL call_9_10_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_9_3
#define init_9_10_9_3
void FASTCALL call_9_10_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_9_4
#define init_9_10_9_4
void FASTCALL call_9_10_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_9_5
#define init_9_10_9_5
void FASTCALL call_9_10_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_9_6
#define init_9_10_9_6
void FASTCALL call_9_10_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_9_7
#define init_9_10_9_7
void FASTCALL call_9_10_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_9_8
#define init_9_10_9_8
void FASTCALL call_9_10_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_9_9
#define init_9_10_9_9
void FASTCALL call_9_10_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_9_10
#define init_9_10_9_10
void FASTCALL call_9_10_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_9_11
#define init_9_10_9_11
void FASTCALL call_9_10_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_9_12
#define init_9_10_9_12
void FASTCALL call_9_10_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_9_13
#define init_9_10_9_13
void FASTCALL call_9_10_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_9_14
#define init_9_10_9_14
void FASTCALL call_9_10_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_9_15
#define init_9_10_9_15
void FASTCALL call_9_10_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_10_0
#define init_9_10_10_0
void FASTCALL call_9_10_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_10_1
#define init_9_10_10_1
void FASTCALL call_9_10_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_10_2
#define init_9_10_10_2
void FASTCALL call_9_10_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_10_3
#define init_9_10_10_3
void FASTCALL call_9_10_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_10_4
#define init_9_10_10_4
void FASTCALL call_9_10_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_10_5
#define init_9_10_10_5
void FASTCALL call_9_10_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_10_6
#define init_9_10_10_6
void FASTCALL call_9_10_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_10_7
#define init_9_10_10_7
void FASTCALL call_9_10_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_10_8
#define init_9_10_10_8
void FASTCALL call_9_10_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_10_9
#define init_9_10_10_9
void FASTCALL call_9_10_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_10_10
#define init_9_10_10_10
void FASTCALL call_9_10_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_10_11
#define init_9_10_10_11
void FASTCALL call_9_10_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_10_12
#define init_9_10_10_12
void FASTCALL call_9_10_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_10_13
#define init_9_10_10_13
void FASTCALL call_9_10_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_10_14
#define init_9_10_10_14
void FASTCALL call_9_10_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_10_15
#define init_9_10_10_15
void FASTCALL call_9_10_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_11_0
#define init_9_10_11_0
void FASTCALL call_9_10_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_11_1
#define init_9_10_11_1
void FASTCALL call_9_10_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_11_2
#define init_9_10_11_2
void FASTCALL call_9_10_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_11_3
#define init_9_10_11_3
void FASTCALL call_9_10_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_11_4
#define init_9_10_11_4
void FASTCALL call_9_10_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_11_5
#define init_9_10_11_5
void FASTCALL call_9_10_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_11_6
#define init_9_10_11_6
void FASTCALL call_9_10_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_11_7
#define init_9_10_11_7
void FASTCALL call_9_10_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_11_8
#define init_9_10_11_8
void FASTCALL call_9_10_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_11_9
#define init_9_10_11_9
void FASTCALL call_9_10_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_11_10
#define init_9_10_11_10
void FASTCALL call_9_10_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_11_11
#define init_9_10_11_11
void FASTCALL call_9_10_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_11_12
#define init_9_10_11_12
void FASTCALL call_9_10_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_11_13
#define init_9_10_11_13
void FASTCALL call_9_10_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_11_14
#define init_9_10_11_14
void FASTCALL call_9_10_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_11_15
#define init_9_10_11_15
void FASTCALL call_9_10_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_12_0
#define init_9_10_12_0
void FASTCALL call_9_10_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_12_1
#define init_9_10_12_1
void FASTCALL call_9_10_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_12_2
#define init_9_10_12_2
void FASTCALL call_9_10_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_12_3
#define init_9_10_12_3
void FASTCALL call_9_10_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_12_4
#define init_9_10_12_4
void FASTCALL call_9_10_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_12_5
#define init_9_10_12_5
void FASTCALL call_9_10_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_12_6
#define init_9_10_12_6
void FASTCALL call_9_10_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_12_7
#define init_9_10_12_7
void FASTCALL call_9_10_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_12_8
#define init_9_10_12_8
void FASTCALL call_9_10_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_12_9
#define init_9_10_12_9
void FASTCALL call_9_10_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_12_10
#define init_9_10_12_10
void FASTCALL call_9_10_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_12_11
#define init_9_10_12_11
void FASTCALL call_9_10_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_12_12
#define init_9_10_12_12
void FASTCALL call_9_10_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_12_13
#define init_9_10_12_13
void FASTCALL call_9_10_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_12_14
#define init_9_10_12_14
void FASTCALL call_9_10_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_12_15
#define init_9_10_12_15
void FASTCALL call_9_10_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_13_0
#define init_9_10_13_0
void FASTCALL call_9_10_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_13_1
#define init_9_10_13_1
void FASTCALL call_9_10_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_13_2
#define init_9_10_13_2
void FASTCALL call_9_10_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_13_3
#define init_9_10_13_3
void FASTCALL call_9_10_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_13_4
#define init_9_10_13_4
void FASTCALL call_9_10_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_13_5
#define init_9_10_13_5
void FASTCALL call_9_10_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_13_6
#define init_9_10_13_6
void FASTCALL call_9_10_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_13_7
#define init_9_10_13_7
void FASTCALL call_9_10_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_13_8
#define init_9_10_13_8
void FASTCALL call_9_10_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_13_9
#define init_9_10_13_9
void FASTCALL call_9_10_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_13_10
#define init_9_10_13_10
void FASTCALL call_9_10_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_13_11
#define init_9_10_13_11
void FASTCALL call_9_10_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_13_12
#define init_9_10_13_12
void FASTCALL call_9_10_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_13_13
#define init_9_10_13_13
void FASTCALL call_9_10_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_13_14
#define init_9_10_13_14
void FASTCALL call_9_10_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_13_15
#define init_9_10_13_15
void FASTCALL call_9_10_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_14_0
#define init_9_10_14_0
void FASTCALL call_9_10_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_14_1
#define init_9_10_14_1
void FASTCALL call_9_10_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_14_2
#define init_9_10_14_2
void FASTCALL call_9_10_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_14_3
#define init_9_10_14_3
void FASTCALL call_9_10_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_14_4
#define init_9_10_14_4
void FASTCALL call_9_10_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_14_5
#define init_9_10_14_5
void FASTCALL call_9_10_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_14_6
#define init_9_10_14_6
void FASTCALL call_9_10_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_14_7
#define init_9_10_14_7
void FASTCALL call_9_10_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_14_8
#define init_9_10_14_8
void FASTCALL call_9_10_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_14_9
#define init_9_10_14_9
void FASTCALL call_9_10_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_14_10
#define init_9_10_14_10
void FASTCALL call_9_10_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_14_11
#define init_9_10_14_11
void FASTCALL call_9_10_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_14_12
#define init_9_10_14_12
void FASTCALL call_9_10_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_14_13
#define init_9_10_14_13
void FASTCALL call_9_10_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_14_14
#define init_9_10_14_14
void FASTCALL call_9_10_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_14_15
#define init_9_10_14_15
void FASTCALL call_9_10_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_15_0
#define init_9_10_15_0
void FASTCALL call_9_10_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_15_1
#define init_9_10_15_1
void FASTCALL call_9_10_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_15_2
#define init_9_10_15_2
void FASTCALL call_9_10_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_15_3
#define init_9_10_15_3
void FASTCALL call_9_10_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_15_4
#define init_9_10_15_4
void FASTCALL call_9_10_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_15_5
#define init_9_10_15_5
void FASTCALL call_9_10_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_15_6
#define init_9_10_15_6
void FASTCALL call_9_10_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_15_7
#define init_9_10_15_7
void FASTCALL call_9_10_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_15_8
#define init_9_10_15_8
void FASTCALL call_9_10_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_15_9
#define init_9_10_15_9
void FASTCALL call_9_10_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_15_10
#define init_9_10_15_10
void FASTCALL call_9_10_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_15_11
#define init_9_10_15_11
void FASTCALL call_9_10_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_15_12
#define init_9_10_15_12
void FASTCALL call_9_10_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_15_13
#define init_9_10_15_13
void FASTCALL call_9_10_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_15_14
#define init_9_10_15_14
void FASTCALL call_9_10_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_10_15_15
#define init_9_10_15_15
void FASTCALL call_9_10_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_0_0
#define init_9_11_0_0
void FASTCALL call_9_11_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_0_1
#define init_9_11_0_1
void FASTCALL call_9_11_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_0_2
#define init_9_11_0_2
void FASTCALL call_9_11_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_0_3
#define init_9_11_0_3
void FASTCALL call_9_11_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_0_4
#define init_9_11_0_4
void FASTCALL call_9_11_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_0_5
#define init_9_11_0_5
void FASTCALL call_9_11_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_0_6
#define init_9_11_0_6
void FASTCALL call_9_11_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_0_7
#define init_9_11_0_7
void FASTCALL call_9_11_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_0_8
#define init_9_11_0_8
void FASTCALL call_9_11_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_0_9
#define init_9_11_0_9
void FASTCALL call_9_11_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_0_10
#define init_9_11_0_10
void FASTCALL call_9_11_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_0_11
#define init_9_11_0_11
void FASTCALL call_9_11_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_0_12
#define init_9_11_0_12
void FASTCALL call_9_11_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_0_13
#define init_9_11_0_13
void FASTCALL call_9_11_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_0_14
#define init_9_11_0_14
void FASTCALL call_9_11_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_0_15
#define init_9_11_0_15
void FASTCALL call_9_11_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_1_0
#define init_9_11_1_0
void FASTCALL call_9_11_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_1_1
#define init_9_11_1_1
void FASTCALL call_9_11_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_1_2
#define init_9_11_1_2
void FASTCALL call_9_11_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_1_3
#define init_9_11_1_3
void FASTCALL call_9_11_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_1_4
#define init_9_11_1_4
void FASTCALL call_9_11_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_1_5
#define init_9_11_1_5
void FASTCALL call_9_11_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_1_6
#define init_9_11_1_6
void FASTCALL call_9_11_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_1_7
#define init_9_11_1_7
void FASTCALL call_9_11_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_1_8
#define init_9_11_1_8
void FASTCALL call_9_11_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_1_9
#define init_9_11_1_9
void FASTCALL call_9_11_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_1_10
#define init_9_11_1_10
void FASTCALL call_9_11_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_1_11
#define init_9_11_1_11
void FASTCALL call_9_11_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_1_12
#define init_9_11_1_12
void FASTCALL call_9_11_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_1_13
#define init_9_11_1_13
void FASTCALL call_9_11_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_1_14
#define init_9_11_1_14
void FASTCALL call_9_11_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_1_15
#define init_9_11_1_15
void FASTCALL call_9_11_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_2_0
#define init_9_11_2_0
void FASTCALL call_9_11_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_2_1
#define init_9_11_2_1
void FASTCALL call_9_11_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_2_2
#define init_9_11_2_2
void FASTCALL call_9_11_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_2_3
#define init_9_11_2_3
void FASTCALL call_9_11_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_2_4
#define init_9_11_2_4
void FASTCALL call_9_11_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_2_5
#define init_9_11_2_5
void FASTCALL call_9_11_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_2_6
#define init_9_11_2_6
void FASTCALL call_9_11_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_2_7
#define init_9_11_2_7
void FASTCALL call_9_11_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_2_8
#define init_9_11_2_8
void FASTCALL call_9_11_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_2_9
#define init_9_11_2_9
void FASTCALL call_9_11_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_2_10
#define init_9_11_2_10
void FASTCALL call_9_11_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_2_11
#define init_9_11_2_11
void FASTCALL call_9_11_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_2_12
#define init_9_11_2_12
void FASTCALL call_9_11_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_2_13
#define init_9_11_2_13
void FASTCALL call_9_11_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_2_14
#define init_9_11_2_14
void FASTCALL call_9_11_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_2_15
#define init_9_11_2_15
void FASTCALL call_9_11_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_3_0
#define init_9_11_3_0
void FASTCALL call_9_11_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_3_1
#define init_9_11_3_1
void FASTCALL call_9_11_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_3_2
#define init_9_11_3_2
void FASTCALL call_9_11_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_3_3
#define init_9_11_3_3
void FASTCALL call_9_11_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_3_4
#define init_9_11_3_4
void FASTCALL call_9_11_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_3_5
#define init_9_11_3_5
void FASTCALL call_9_11_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_3_6
#define init_9_11_3_6
void FASTCALL call_9_11_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_3_7
#define init_9_11_3_7
void FASTCALL call_9_11_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_3_8
#define init_9_11_3_8
void FASTCALL call_9_11_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_3_9
#define init_9_11_3_9
void FASTCALL call_9_11_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_3_10
#define init_9_11_3_10
void FASTCALL call_9_11_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_3_11
#define init_9_11_3_11
void FASTCALL call_9_11_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_3_12
#define init_9_11_3_12
void FASTCALL call_9_11_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_3_13
#define init_9_11_3_13
void FASTCALL call_9_11_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_3_14
#define init_9_11_3_14
void FASTCALL call_9_11_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_3_15
#define init_9_11_3_15
void FASTCALL call_9_11_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_4_0
#define init_9_11_4_0
void FASTCALL call_9_11_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_4_1
#define init_9_11_4_1
void FASTCALL call_9_11_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_4_2
#define init_9_11_4_2
void FASTCALL call_9_11_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_4_3
#define init_9_11_4_3
void FASTCALL call_9_11_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_4_4
#define init_9_11_4_4
void FASTCALL call_9_11_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_4_5
#define init_9_11_4_5
void FASTCALL call_9_11_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_4_6
#define init_9_11_4_6
void FASTCALL call_9_11_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_4_7
#define init_9_11_4_7
void FASTCALL call_9_11_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_4_8
#define init_9_11_4_8
void FASTCALL call_9_11_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_4_9
#define init_9_11_4_9
void FASTCALL call_9_11_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_4_10
#define init_9_11_4_10
void FASTCALL call_9_11_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_4_11
#define init_9_11_4_11
void FASTCALL call_9_11_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_4_12
#define init_9_11_4_12
void FASTCALL call_9_11_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_4_13
#define init_9_11_4_13
void FASTCALL call_9_11_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_4_14
#define init_9_11_4_14
void FASTCALL call_9_11_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_4_15
#define init_9_11_4_15
void FASTCALL call_9_11_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_5_0
#define init_9_11_5_0
void FASTCALL call_9_11_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_5_1
#define init_9_11_5_1
void FASTCALL call_9_11_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_5_2
#define init_9_11_5_2
void FASTCALL call_9_11_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_5_3
#define init_9_11_5_3
void FASTCALL call_9_11_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_5_4
#define init_9_11_5_4
void FASTCALL call_9_11_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_5_5
#define init_9_11_5_5
void FASTCALL call_9_11_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_5_6
#define init_9_11_5_6
void FASTCALL call_9_11_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_5_7
#define init_9_11_5_7
void FASTCALL call_9_11_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_5_8
#define init_9_11_5_8
void FASTCALL call_9_11_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_5_9
#define init_9_11_5_9
void FASTCALL call_9_11_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_5_10
#define init_9_11_5_10
void FASTCALL call_9_11_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_5_11
#define init_9_11_5_11
void FASTCALL call_9_11_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_5_12
#define init_9_11_5_12
void FASTCALL call_9_11_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_5_13
#define init_9_11_5_13
void FASTCALL call_9_11_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_5_14
#define init_9_11_5_14
void FASTCALL call_9_11_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_5_15
#define init_9_11_5_15
void FASTCALL call_9_11_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_6_0
#define init_9_11_6_0
void FASTCALL call_9_11_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_6_1
#define init_9_11_6_1
void FASTCALL call_9_11_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_6_2
#define init_9_11_6_2
void FASTCALL call_9_11_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_6_3
#define init_9_11_6_3
void FASTCALL call_9_11_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_6_4
#define init_9_11_6_4
void FASTCALL call_9_11_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_6_5
#define init_9_11_6_5
void FASTCALL call_9_11_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_6_6
#define init_9_11_6_6
void FASTCALL call_9_11_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_6_7
#define init_9_11_6_7
void FASTCALL call_9_11_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_6_8
#define init_9_11_6_8
void FASTCALL call_9_11_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_6_9
#define init_9_11_6_9
void FASTCALL call_9_11_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_6_10
#define init_9_11_6_10
void FASTCALL call_9_11_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_6_11
#define init_9_11_6_11
void FASTCALL call_9_11_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_6_12
#define init_9_11_6_12
void FASTCALL call_9_11_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_6_13
#define init_9_11_6_13
void FASTCALL call_9_11_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_6_14
#define init_9_11_6_14
void FASTCALL call_9_11_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_6_15
#define init_9_11_6_15
void FASTCALL call_9_11_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_7_0
#define init_9_11_7_0
void FASTCALL call_9_11_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_7_1
#define init_9_11_7_1
void FASTCALL call_9_11_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_7_2
#define init_9_11_7_2
void FASTCALL call_9_11_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_7_3
#define init_9_11_7_3
void FASTCALL call_9_11_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_7_4
#define init_9_11_7_4
void FASTCALL call_9_11_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_7_5
#define init_9_11_7_5
void FASTCALL call_9_11_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_7_6
#define init_9_11_7_6
void FASTCALL call_9_11_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_7_7
#define init_9_11_7_7
void FASTCALL call_9_11_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_7_8
#define init_9_11_7_8
void FASTCALL call_9_11_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_7_9
#define init_9_11_7_9
void FASTCALL call_9_11_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_7_10
#define init_9_11_7_10
void FASTCALL call_9_11_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_7_11
#define init_9_11_7_11
void FASTCALL call_9_11_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_7_12
#define init_9_11_7_12
void FASTCALL call_9_11_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_7_13
#define init_9_11_7_13
void FASTCALL call_9_11_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_7_14
#define init_9_11_7_14
void FASTCALL call_9_11_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_7_15
#define init_9_11_7_15
void FASTCALL call_9_11_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_8_0
#define init_9_11_8_0
void FASTCALL call_9_11_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_8_1
#define init_9_11_8_1
void FASTCALL call_9_11_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_8_2
#define init_9_11_8_2
void FASTCALL call_9_11_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_8_3
#define init_9_11_8_3
void FASTCALL call_9_11_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_8_4
#define init_9_11_8_4
void FASTCALL call_9_11_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_8_5
#define init_9_11_8_5
void FASTCALL call_9_11_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_8_6
#define init_9_11_8_6
void FASTCALL call_9_11_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_8_7
#define init_9_11_8_7
void FASTCALL call_9_11_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_8_8
#define init_9_11_8_8
void FASTCALL call_9_11_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_8_9
#define init_9_11_8_9
void FASTCALL call_9_11_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_8_10
#define init_9_11_8_10
void FASTCALL call_9_11_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_8_11
#define init_9_11_8_11
void FASTCALL call_9_11_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_8_12
#define init_9_11_8_12
void FASTCALL call_9_11_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_8_13
#define init_9_11_8_13
void FASTCALL call_9_11_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_8_14
#define init_9_11_8_14
void FASTCALL call_9_11_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_8_15
#define init_9_11_8_15
void FASTCALL call_9_11_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_9_0
#define init_9_11_9_0
void FASTCALL call_9_11_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_9_1
#define init_9_11_9_1
void FASTCALL call_9_11_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_9_2
#define init_9_11_9_2
void FASTCALL call_9_11_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_9_3
#define init_9_11_9_3
void FASTCALL call_9_11_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_9_4
#define init_9_11_9_4
void FASTCALL call_9_11_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_9_5
#define init_9_11_9_5
void FASTCALL call_9_11_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_9_6
#define init_9_11_9_6
void FASTCALL call_9_11_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_9_7
#define init_9_11_9_7
void FASTCALL call_9_11_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_9_8
#define init_9_11_9_8
void FASTCALL call_9_11_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_9_9
#define init_9_11_9_9
void FASTCALL call_9_11_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_9_10
#define init_9_11_9_10
void FASTCALL call_9_11_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_9_11
#define init_9_11_9_11
void FASTCALL call_9_11_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_9_12
#define init_9_11_9_12
void FASTCALL call_9_11_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_9_13
#define init_9_11_9_13
void FASTCALL call_9_11_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_9_14
#define init_9_11_9_14
void FASTCALL call_9_11_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_9_15
#define init_9_11_9_15
void FASTCALL call_9_11_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_10_0
#define init_9_11_10_0
void FASTCALL call_9_11_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_10_1
#define init_9_11_10_1
void FASTCALL call_9_11_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_10_2
#define init_9_11_10_2
void FASTCALL call_9_11_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_10_3
#define init_9_11_10_3
void FASTCALL call_9_11_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_10_4
#define init_9_11_10_4
void FASTCALL call_9_11_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_10_5
#define init_9_11_10_5
void FASTCALL call_9_11_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_10_6
#define init_9_11_10_6
void FASTCALL call_9_11_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_10_7
#define init_9_11_10_7
void FASTCALL call_9_11_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_10_8
#define init_9_11_10_8
void FASTCALL call_9_11_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_10_9
#define init_9_11_10_9
void FASTCALL call_9_11_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_10_10
#define init_9_11_10_10
void FASTCALL call_9_11_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_10_11
#define init_9_11_10_11
void FASTCALL call_9_11_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_10_12
#define init_9_11_10_12
void FASTCALL call_9_11_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_10_13
#define init_9_11_10_13
void FASTCALL call_9_11_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_10_14
#define init_9_11_10_14
void FASTCALL call_9_11_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_10_15
#define init_9_11_10_15
void FASTCALL call_9_11_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_11_0
#define init_9_11_11_0
void FASTCALL call_9_11_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_11_1
#define init_9_11_11_1
void FASTCALL call_9_11_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_11_2
#define init_9_11_11_2
void FASTCALL call_9_11_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_11_3
#define init_9_11_11_3
void FASTCALL call_9_11_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_11_4
#define init_9_11_11_4
void FASTCALL call_9_11_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_11_5
#define init_9_11_11_5
void FASTCALL call_9_11_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_11_6
#define init_9_11_11_6
void FASTCALL call_9_11_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_11_7
#define init_9_11_11_7
void FASTCALL call_9_11_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_11_8
#define init_9_11_11_8
void FASTCALL call_9_11_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_11_9
#define init_9_11_11_9
void FASTCALL call_9_11_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_11_10
#define init_9_11_11_10
void FASTCALL call_9_11_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_11_11
#define init_9_11_11_11
void FASTCALL call_9_11_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_11_12
#define init_9_11_11_12
void FASTCALL call_9_11_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_11_13
#define init_9_11_11_13
void FASTCALL call_9_11_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_11_14
#define init_9_11_11_14
void FASTCALL call_9_11_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_11_15
#define init_9_11_11_15
void FASTCALL call_9_11_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_12_0
#define init_9_11_12_0
void FASTCALL call_9_11_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_12_1
#define init_9_11_12_1
void FASTCALL call_9_11_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_12_2
#define init_9_11_12_2
void FASTCALL call_9_11_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_12_3
#define init_9_11_12_3
void FASTCALL call_9_11_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_12_4
#define init_9_11_12_4
void FASTCALL call_9_11_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_12_5
#define init_9_11_12_5
void FASTCALL call_9_11_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_12_6
#define init_9_11_12_6
void FASTCALL call_9_11_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_12_7
#define init_9_11_12_7
void FASTCALL call_9_11_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_12_8
#define init_9_11_12_8
void FASTCALL call_9_11_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_12_9
#define init_9_11_12_9
void FASTCALL call_9_11_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_12_10
#define init_9_11_12_10
void FASTCALL call_9_11_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_12_11
#define init_9_11_12_11
void FASTCALL call_9_11_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_12_12
#define init_9_11_12_12
void FASTCALL call_9_11_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_12_13
#define init_9_11_12_13
void FASTCALL call_9_11_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_12_14
#define init_9_11_12_14
void FASTCALL call_9_11_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_12_15
#define init_9_11_12_15
void FASTCALL call_9_11_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_13_0
#define init_9_11_13_0
void FASTCALL call_9_11_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_13_1
#define init_9_11_13_1
void FASTCALL call_9_11_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_13_2
#define init_9_11_13_2
void FASTCALL call_9_11_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_13_3
#define init_9_11_13_3
void FASTCALL call_9_11_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_13_4
#define init_9_11_13_4
void FASTCALL call_9_11_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_13_5
#define init_9_11_13_5
void FASTCALL call_9_11_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_13_6
#define init_9_11_13_6
void FASTCALL call_9_11_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_13_7
#define init_9_11_13_7
void FASTCALL call_9_11_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_13_8
#define init_9_11_13_8
void FASTCALL call_9_11_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_13_9
#define init_9_11_13_9
void FASTCALL call_9_11_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_13_10
#define init_9_11_13_10
void FASTCALL call_9_11_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_13_11
#define init_9_11_13_11
void FASTCALL call_9_11_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_13_12
#define init_9_11_13_12
void FASTCALL call_9_11_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_13_13
#define init_9_11_13_13
void FASTCALL call_9_11_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_13_14
#define init_9_11_13_14
void FASTCALL call_9_11_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_13_15
#define init_9_11_13_15
void FASTCALL call_9_11_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_14_0
#define init_9_11_14_0
void FASTCALL call_9_11_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_14_1
#define init_9_11_14_1
void FASTCALL call_9_11_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_14_2
#define init_9_11_14_2
void FASTCALL call_9_11_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_14_3
#define init_9_11_14_3
void FASTCALL call_9_11_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_14_4
#define init_9_11_14_4
void FASTCALL call_9_11_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_14_5
#define init_9_11_14_5
void FASTCALL call_9_11_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_14_6
#define init_9_11_14_6
void FASTCALL call_9_11_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_14_7
#define init_9_11_14_7
void FASTCALL call_9_11_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_14_8
#define init_9_11_14_8
void FASTCALL call_9_11_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_14_9
#define init_9_11_14_9
void FASTCALL call_9_11_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_14_10
#define init_9_11_14_10
void FASTCALL call_9_11_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_14_11
#define init_9_11_14_11
void FASTCALL call_9_11_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_14_12
#define init_9_11_14_12
void FASTCALL call_9_11_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_14_13
#define init_9_11_14_13
void FASTCALL call_9_11_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_14_14
#define init_9_11_14_14
void FASTCALL call_9_11_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_14_15
#define init_9_11_14_15
void FASTCALL call_9_11_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_15_0
#define init_9_11_15_0
void FASTCALL call_9_11_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_15_1
#define init_9_11_15_1
void FASTCALL call_9_11_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_15_2
#define init_9_11_15_2
void FASTCALL call_9_11_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_15_3
#define init_9_11_15_3
void FASTCALL call_9_11_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_15_4
#define init_9_11_15_4
void FASTCALL call_9_11_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_15_5
#define init_9_11_15_5
void FASTCALL call_9_11_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_15_6
#define init_9_11_15_6
void FASTCALL call_9_11_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_15_7
#define init_9_11_15_7
void FASTCALL call_9_11_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_15_8
#define init_9_11_15_8
void FASTCALL call_9_11_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_15_9
#define init_9_11_15_9
void FASTCALL call_9_11_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_15_10
#define init_9_11_15_10
void FASTCALL call_9_11_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_15_11
#define init_9_11_15_11
void FASTCALL call_9_11_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_15_12
#define init_9_11_15_12
void FASTCALL call_9_11_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_15_13
#define init_9_11_15_13
void FASTCALL call_9_11_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_15_14
#define init_9_11_15_14
void FASTCALL call_9_11_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_11_15_15
#define init_9_11_15_15
void FASTCALL call_9_11_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_0_0
#define init_9_12_0_0
void FASTCALL call_9_12_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_0_1
#define init_9_12_0_1
void FASTCALL call_9_12_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_0_2
#define init_9_12_0_2
void FASTCALL call_9_12_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_0_3
#define init_9_12_0_3
void FASTCALL call_9_12_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_0_4
#define init_9_12_0_4
void FASTCALL call_9_12_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_0_5
#define init_9_12_0_5
void FASTCALL call_9_12_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_0_6
#define init_9_12_0_6
void FASTCALL call_9_12_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_0_7
#define init_9_12_0_7
void FASTCALL call_9_12_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_0_8
#define init_9_12_0_8
void FASTCALL call_9_12_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_0_9
#define init_9_12_0_9
void FASTCALL call_9_12_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_0_10
#define init_9_12_0_10
void FASTCALL call_9_12_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_0_11
#define init_9_12_0_11
void FASTCALL call_9_12_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_0_12
#define init_9_12_0_12
void FASTCALL call_9_12_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_0_13
#define init_9_12_0_13
void FASTCALL call_9_12_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_0_14
#define init_9_12_0_14
void FASTCALL call_9_12_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_0_15
#define init_9_12_0_15
void FASTCALL call_9_12_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_1_0
#define init_9_12_1_0
void FASTCALL call_9_12_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_1_1
#define init_9_12_1_1
void FASTCALL call_9_12_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_1_2
#define init_9_12_1_2
void FASTCALL call_9_12_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_1_3
#define init_9_12_1_3
void FASTCALL call_9_12_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_1_4
#define init_9_12_1_4
void FASTCALL call_9_12_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_1_5
#define init_9_12_1_5
void FASTCALL call_9_12_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_1_6
#define init_9_12_1_6
void FASTCALL call_9_12_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_1_7
#define init_9_12_1_7
void FASTCALL call_9_12_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_1_8
#define init_9_12_1_8
void FASTCALL call_9_12_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_1_9
#define init_9_12_1_9
void FASTCALL call_9_12_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_1_10
#define init_9_12_1_10
void FASTCALL call_9_12_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_1_11
#define init_9_12_1_11
void FASTCALL call_9_12_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_1_12
#define init_9_12_1_12
void FASTCALL call_9_12_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_1_13
#define init_9_12_1_13
void FASTCALL call_9_12_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_1_14
#define init_9_12_1_14
void FASTCALL call_9_12_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_1_15
#define init_9_12_1_15
void FASTCALL call_9_12_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_2_0
#define init_9_12_2_0
void FASTCALL call_9_12_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_2_1
#define init_9_12_2_1
void FASTCALL call_9_12_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_2_2
#define init_9_12_2_2
void FASTCALL call_9_12_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_2_3
#define init_9_12_2_3
void FASTCALL call_9_12_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_2_4
#define init_9_12_2_4
void FASTCALL call_9_12_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_2_5
#define init_9_12_2_5
void FASTCALL call_9_12_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_2_6
#define init_9_12_2_6
void FASTCALL call_9_12_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_2_7
#define init_9_12_2_7
void FASTCALL call_9_12_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_2_8
#define init_9_12_2_8
void FASTCALL call_9_12_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_2_9
#define init_9_12_2_9
void FASTCALL call_9_12_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_2_10
#define init_9_12_2_10
void FASTCALL call_9_12_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_2_11
#define init_9_12_2_11
void FASTCALL call_9_12_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_2_12
#define init_9_12_2_12
void FASTCALL call_9_12_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_2_13
#define init_9_12_2_13
void FASTCALL call_9_12_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_2_14
#define init_9_12_2_14
void FASTCALL call_9_12_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_2_15
#define init_9_12_2_15
void FASTCALL call_9_12_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_3_0
#define init_9_12_3_0
void FASTCALL call_9_12_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_3_1
#define init_9_12_3_1
void FASTCALL call_9_12_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_3_2
#define init_9_12_3_2
void FASTCALL call_9_12_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_3_3
#define init_9_12_3_3
void FASTCALL call_9_12_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_3_4
#define init_9_12_3_4
void FASTCALL call_9_12_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_3_5
#define init_9_12_3_5
void FASTCALL call_9_12_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_3_6
#define init_9_12_3_6
void FASTCALL call_9_12_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_3_7
#define init_9_12_3_7
void FASTCALL call_9_12_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_3_8
#define init_9_12_3_8
void FASTCALL call_9_12_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_3_9
#define init_9_12_3_9
void FASTCALL call_9_12_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_3_10
#define init_9_12_3_10
void FASTCALL call_9_12_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_3_11
#define init_9_12_3_11
void FASTCALL call_9_12_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_3_12
#define init_9_12_3_12
void FASTCALL call_9_12_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_3_13
#define init_9_12_3_13
void FASTCALL call_9_12_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_3_14
#define init_9_12_3_14
void FASTCALL call_9_12_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_3_15
#define init_9_12_3_15
void FASTCALL call_9_12_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_4_0
#define init_9_12_4_0
void FASTCALL call_9_12_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_4_1
#define init_9_12_4_1
void FASTCALL call_9_12_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_4_2
#define init_9_12_4_2
void FASTCALL call_9_12_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_4_3
#define init_9_12_4_3
void FASTCALL call_9_12_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_4_4
#define init_9_12_4_4
void FASTCALL call_9_12_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_4_5
#define init_9_12_4_5
void FASTCALL call_9_12_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_4_6
#define init_9_12_4_6
void FASTCALL call_9_12_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_4_7
#define init_9_12_4_7
void FASTCALL call_9_12_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_4_8
#define init_9_12_4_8
void FASTCALL call_9_12_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_4_9
#define init_9_12_4_9
void FASTCALL call_9_12_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_4_10
#define init_9_12_4_10
void FASTCALL call_9_12_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_4_11
#define init_9_12_4_11
void FASTCALL call_9_12_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_4_12
#define init_9_12_4_12
void FASTCALL call_9_12_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_4_13
#define init_9_12_4_13
void FASTCALL call_9_12_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_4_14
#define init_9_12_4_14
void FASTCALL call_9_12_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_4_15
#define init_9_12_4_15
void FASTCALL call_9_12_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_5_0
#define init_9_12_5_0
void FASTCALL call_9_12_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_5_1
#define init_9_12_5_1
void FASTCALL call_9_12_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_5_2
#define init_9_12_5_2
void FASTCALL call_9_12_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_5_3
#define init_9_12_5_3
void FASTCALL call_9_12_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_5_4
#define init_9_12_5_4
void FASTCALL call_9_12_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_5_5
#define init_9_12_5_5
void FASTCALL call_9_12_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_5_6
#define init_9_12_5_6
void FASTCALL call_9_12_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_5_7
#define init_9_12_5_7
void FASTCALL call_9_12_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_5_8
#define init_9_12_5_8
void FASTCALL call_9_12_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_5_9
#define init_9_12_5_9
void FASTCALL call_9_12_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_5_10
#define init_9_12_5_10
void FASTCALL call_9_12_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_5_11
#define init_9_12_5_11
void FASTCALL call_9_12_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_5_12
#define init_9_12_5_12
void FASTCALL call_9_12_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_5_13
#define init_9_12_5_13
void FASTCALL call_9_12_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_5_14
#define init_9_12_5_14
void FASTCALL call_9_12_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_5_15
#define init_9_12_5_15
void FASTCALL call_9_12_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_6_0
#define init_9_12_6_0
void FASTCALL call_9_12_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_6_1
#define init_9_12_6_1
void FASTCALL call_9_12_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_6_2
#define init_9_12_6_2
void FASTCALL call_9_12_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_6_3
#define init_9_12_6_3
void FASTCALL call_9_12_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_6_4
#define init_9_12_6_4
void FASTCALL call_9_12_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_6_5
#define init_9_12_6_5
void FASTCALL call_9_12_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_6_6
#define init_9_12_6_6
void FASTCALL call_9_12_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_6_7
#define init_9_12_6_7
void FASTCALL call_9_12_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_6_8
#define init_9_12_6_8
void FASTCALL call_9_12_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_6_9
#define init_9_12_6_9
void FASTCALL call_9_12_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_6_10
#define init_9_12_6_10
void FASTCALL call_9_12_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_6_11
#define init_9_12_6_11
void FASTCALL call_9_12_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_6_12
#define init_9_12_6_12
void FASTCALL call_9_12_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_6_13
#define init_9_12_6_13
void FASTCALL call_9_12_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_6_14
#define init_9_12_6_14
void FASTCALL call_9_12_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_6_15
#define init_9_12_6_15
void FASTCALL call_9_12_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_7_0
#define init_9_12_7_0
void FASTCALL call_9_12_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_7_1
#define init_9_12_7_1
void FASTCALL call_9_12_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_7_2
#define init_9_12_7_2
void FASTCALL call_9_12_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_7_3
#define init_9_12_7_3
void FASTCALL call_9_12_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_7_4
#define init_9_12_7_4
void FASTCALL call_9_12_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_7_5
#define init_9_12_7_5
void FASTCALL call_9_12_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_7_6
#define init_9_12_7_6
void FASTCALL call_9_12_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_7_7
#define init_9_12_7_7
void FASTCALL call_9_12_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_7_8
#define init_9_12_7_8
void FASTCALL call_9_12_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_7_9
#define init_9_12_7_9
void FASTCALL call_9_12_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_7_10
#define init_9_12_7_10
void FASTCALL call_9_12_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_7_11
#define init_9_12_7_11
void FASTCALL call_9_12_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_7_12
#define init_9_12_7_12
void FASTCALL call_9_12_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_7_13
#define init_9_12_7_13
void FASTCALL call_9_12_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_7_14
#define init_9_12_7_14
void FASTCALL call_9_12_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_7_15
#define init_9_12_7_15
void FASTCALL call_9_12_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_8_0
#define init_9_12_8_0
void FASTCALL call_9_12_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_8_1
#define init_9_12_8_1
void FASTCALL call_9_12_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_8_2
#define init_9_12_8_2
void FASTCALL call_9_12_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_8_3
#define init_9_12_8_3
void FASTCALL call_9_12_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_8_4
#define init_9_12_8_4
void FASTCALL call_9_12_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_8_5
#define init_9_12_8_5
void FASTCALL call_9_12_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_8_6
#define init_9_12_8_6
void FASTCALL call_9_12_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_8_7
#define init_9_12_8_7
void FASTCALL call_9_12_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_8_8
#define init_9_12_8_8
void FASTCALL call_9_12_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_8_9
#define init_9_12_8_9
void FASTCALL call_9_12_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_8_10
#define init_9_12_8_10
void FASTCALL call_9_12_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_8_11
#define init_9_12_8_11
void FASTCALL call_9_12_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_8_12
#define init_9_12_8_12
void FASTCALL call_9_12_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_8_13
#define init_9_12_8_13
void FASTCALL call_9_12_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_8_14
#define init_9_12_8_14
void FASTCALL call_9_12_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_8_15
#define init_9_12_8_15
void FASTCALL call_9_12_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_9_0
#define init_9_12_9_0
void FASTCALL call_9_12_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_9_1
#define init_9_12_9_1
void FASTCALL call_9_12_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_9_2
#define init_9_12_9_2
void FASTCALL call_9_12_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_9_3
#define init_9_12_9_3
void FASTCALL call_9_12_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_9_4
#define init_9_12_9_4
void FASTCALL call_9_12_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_9_5
#define init_9_12_9_5
void FASTCALL call_9_12_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_9_6
#define init_9_12_9_6
void FASTCALL call_9_12_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_9_7
#define init_9_12_9_7
void FASTCALL call_9_12_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_9_8
#define init_9_12_9_8
void FASTCALL call_9_12_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_9_9
#define init_9_12_9_9
void FASTCALL call_9_12_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_9_10
#define init_9_12_9_10
void FASTCALL call_9_12_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_9_11
#define init_9_12_9_11
void FASTCALL call_9_12_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_9_12
#define init_9_12_9_12
void FASTCALL call_9_12_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_9_13
#define init_9_12_9_13
void FASTCALL call_9_12_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_9_14
#define init_9_12_9_14
void FASTCALL call_9_12_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_9_15
#define init_9_12_9_15
void FASTCALL call_9_12_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_10_0
#define init_9_12_10_0
void FASTCALL call_9_12_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_10_1
#define init_9_12_10_1
void FASTCALL call_9_12_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_10_2
#define init_9_12_10_2
void FASTCALL call_9_12_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_10_3
#define init_9_12_10_3
void FASTCALL call_9_12_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_10_4
#define init_9_12_10_4
void FASTCALL call_9_12_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_10_5
#define init_9_12_10_5
void FASTCALL call_9_12_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_10_6
#define init_9_12_10_6
void FASTCALL call_9_12_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_10_7
#define init_9_12_10_7
void FASTCALL call_9_12_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_10_8
#define init_9_12_10_8
void FASTCALL call_9_12_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_10_9
#define init_9_12_10_9
void FASTCALL call_9_12_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_10_10
#define init_9_12_10_10
void FASTCALL call_9_12_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_10_11
#define init_9_12_10_11
void FASTCALL call_9_12_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_10_12
#define init_9_12_10_12
void FASTCALL call_9_12_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_10_13
#define init_9_12_10_13
void FASTCALL call_9_12_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_10_14
#define init_9_12_10_14
void FASTCALL call_9_12_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_10_15
#define init_9_12_10_15
void FASTCALL call_9_12_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_11_0
#define init_9_12_11_0
void FASTCALL call_9_12_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_11_1
#define init_9_12_11_1
void FASTCALL call_9_12_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_11_2
#define init_9_12_11_2
void FASTCALL call_9_12_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_11_3
#define init_9_12_11_3
void FASTCALL call_9_12_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_11_4
#define init_9_12_11_4
void FASTCALL call_9_12_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_11_5
#define init_9_12_11_5
void FASTCALL call_9_12_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_11_6
#define init_9_12_11_6
void FASTCALL call_9_12_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_11_7
#define init_9_12_11_7
void FASTCALL call_9_12_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_11_8
#define init_9_12_11_8
void FASTCALL call_9_12_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_11_9
#define init_9_12_11_9
void FASTCALL call_9_12_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_11_10
#define init_9_12_11_10
void FASTCALL call_9_12_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_11_11
#define init_9_12_11_11
void FASTCALL call_9_12_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_11_12
#define init_9_12_11_12
void FASTCALL call_9_12_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_11_13
#define init_9_12_11_13
void FASTCALL call_9_12_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_11_14
#define init_9_12_11_14
void FASTCALL call_9_12_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_11_15
#define init_9_12_11_15
void FASTCALL call_9_12_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_12_0
#define init_9_12_12_0
void FASTCALL call_9_12_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_12_1
#define init_9_12_12_1
void FASTCALL call_9_12_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_12_2
#define init_9_12_12_2
void FASTCALL call_9_12_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_12_3
#define init_9_12_12_3
void FASTCALL call_9_12_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_12_4
#define init_9_12_12_4
void FASTCALL call_9_12_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_12_5
#define init_9_12_12_5
void FASTCALL call_9_12_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_12_6
#define init_9_12_12_6
void FASTCALL call_9_12_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_12_7
#define init_9_12_12_7
void FASTCALL call_9_12_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_12_8
#define init_9_12_12_8
void FASTCALL call_9_12_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_12_9
#define init_9_12_12_9
void FASTCALL call_9_12_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_12_10
#define init_9_12_12_10
void FASTCALL call_9_12_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_12_11
#define init_9_12_12_11
void FASTCALL call_9_12_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_12_12
#define init_9_12_12_12
void FASTCALL call_9_12_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_12_13
#define init_9_12_12_13
void FASTCALL call_9_12_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_12_14
#define init_9_12_12_14
void FASTCALL call_9_12_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_12_15
#define init_9_12_12_15
void FASTCALL call_9_12_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_13_0
#define init_9_12_13_0
void FASTCALL call_9_12_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_13_1
#define init_9_12_13_1
void FASTCALL call_9_12_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_13_2
#define init_9_12_13_2
void FASTCALL call_9_12_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_13_3
#define init_9_12_13_3
void FASTCALL call_9_12_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_13_4
#define init_9_12_13_4
void FASTCALL call_9_12_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_13_5
#define init_9_12_13_5
void FASTCALL call_9_12_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_13_6
#define init_9_12_13_6
void FASTCALL call_9_12_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_13_7
#define init_9_12_13_7
void FASTCALL call_9_12_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_13_8
#define init_9_12_13_8
void FASTCALL call_9_12_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_13_9
#define init_9_12_13_9
void FASTCALL call_9_12_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_13_10
#define init_9_12_13_10
void FASTCALL call_9_12_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_13_11
#define init_9_12_13_11
void FASTCALL call_9_12_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_13_12
#define init_9_12_13_12
void FASTCALL call_9_12_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_13_13
#define init_9_12_13_13
void FASTCALL call_9_12_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_13_14
#define init_9_12_13_14
void FASTCALL call_9_12_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_13_15
#define init_9_12_13_15
void FASTCALL call_9_12_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_14_0
#define init_9_12_14_0
void FASTCALL call_9_12_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_14_1
#define init_9_12_14_1
void FASTCALL call_9_12_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_14_2
#define init_9_12_14_2
void FASTCALL call_9_12_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_14_3
#define init_9_12_14_3
void FASTCALL call_9_12_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_14_4
#define init_9_12_14_4
void FASTCALL call_9_12_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_14_5
#define init_9_12_14_5
void FASTCALL call_9_12_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_14_6
#define init_9_12_14_6
void FASTCALL call_9_12_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_14_7
#define init_9_12_14_7
void FASTCALL call_9_12_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_14_8
#define init_9_12_14_8
void FASTCALL call_9_12_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_14_9
#define init_9_12_14_9
void FASTCALL call_9_12_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_14_10
#define init_9_12_14_10
void FASTCALL call_9_12_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_14_11
#define init_9_12_14_11
void FASTCALL call_9_12_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_14_12
#define init_9_12_14_12
void FASTCALL call_9_12_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_14_13
#define init_9_12_14_13
void FASTCALL call_9_12_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_14_14
#define init_9_12_14_14
void FASTCALL call_9_12_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_14_15
#define init_9_12_14_15
void FASTCALL call_9_12_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_15_0
#define init_9_12_15_0
void FASTCALL call_9_12_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_15_1
#define init_9_12_15_1
void FASTCALL call_9_12_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_15_2
#define init_9_12_15_2
void FASTCALL call_9_12_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_15_3
#define init_9_12_15_3
void FASTCALL call_9_12_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_15_4
#define init_9_12_15_4
void FASTCALL call_9_12_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_15_5
#define init_9_12_15_5
void FASTCALL call_9_12_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_15_6
#define init_9_12_15_6
void FASTCALL call_9_12_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_15_7
#define init_9_12_15_7
void FASTCALL call_9_12_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_15_8
#define init_9_12_15_8
void FASTCALL call_9_12_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_15_9
#define init_9_12_15_9
void FASTCALL call_9_12_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_15_10
#define init_9_12_15_10
void FASTCALL call_9_12_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_15_11
#define init_9_12_15_11
void FASTCALL call_9_12_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_15_12
#define init_9_12_15_12
void FASTCALL call_9_12_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_15_13
#define init_9_12_15_13
void FASTCALL call_9_12_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_15_14
#define init_9_12_15_14
void FASTCALL call_9_12_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_12_15_15
#define init_9_12_15_15
void FASTCALL call_9_12_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_0_0
#define init_9_13_0_0
void FASTCALL call_9_13_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_0_1
#define init_9_13_0_1
void FASTCALL call_9_13_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_0_2
#define init_9_13_0_2
void FASTCALL call_9_13_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_0_3
#define init_9_13_0_3
void FASTCALL call_9_13_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_0_4
#define init_9_13_0_4
void FASTCALL call_9_13_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_0_5
#define init_9_13_0_5
void FASTCALL call_9_13_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_0_6
#define init_9_13_0_6
void FASTCALL call_9_13_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_0_7
#define init_9_13_0_7
void FASTCALL call_9_13_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_0_8
#define init_9_13_0_8
void FASTCALL call_9_13_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_0_9
#define init_9_13_0_9
void FASTCALL call_9_13_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_0_10
#define init_9_13_0_10
void FASTCALL call_9_13_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_0_11
#define init_9_13_0_11
void FASTCALL call_9_13_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_0_12
#define init_9_13_0_12
void FASTCALL call_9_13_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_0_13
#define init_9_13_0_13
void FASTCALL call_9_13_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_0_14
#define init_9_13_0_14
void FASTCALL call_9_13_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_0_15
#define init_9_13_0_15
void FASTCALL call_9_13_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_1_0
#define init_9_13_1_0
void FASTCALL call_9_13_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_1_1
#define init_9_13_1_1
void FASTCALL call_9_13_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_1_2
#define init_9_13_1_2
void FASTCALL call_9_13_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_1_3
#define init_9_13_1_3
void FASTCALL call_9_13_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_1_4
#define init_9_13_1_4
void FASTCALL call_9_13_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_1_5
#define init_9_13_1_5
void FASTCALL call_9_13_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_1_6
#define init_9_13_1_6
void FASTCALL call_9_13_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_1_7
#define init_9_13_1_7
void FASTCALL call_9_13_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_1_8
#define init_9_13_1_8
void FASTCALL call_9_13_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_1_9
#define init_9_13_1_9
void FASTCALL call_9_13_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_1_10
#define init_9_13_1_10
void FASTCALL call_9_13_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_1_11
#define init_9_13_1_11
void FASTCALL call_9_13_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_1_12
#define init_9_13_1_12
void FASTCALL call_9_13_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_1_13
#define init_9_13_1_13
void FASTCALL call_9_13_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_1_14
#define init_9_13_1_14
void FASTCALL call_9_13_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_1_15
#define init_9_13_1_15
void FASTCALL call_9_13_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_2_0
#define init_9_13_2_0
void FASTCALL call_9_13_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_2_1
#define init_9_13_2_1
void FASTCALL call_9_13_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_2_2
#define init_9_13_2_2
void FASTCALL call_9_13_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_2_3
#define init_9_13_2_3
void FASTCALL call_9_13_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_2_4
#define init_9_13_2_4
void FASTCALL call_9_13_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_2_5
#define init_9_13_2_5
void FASTCALL call_9_13_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_2_6
#define init_9_13_2_6
void FASTCALL call_9_13_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_2_7
#define init_9_13_2_7
void FASTCALL call_9_13_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_2_8
#define init_9_13_2_8
void FASTCALL call_9_13_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_2_9
#define init_9_13_2_9
void FASTCALL call_9_13_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_2_10
#define init_9_13_2_10
void FASTCALL call_9_13_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_2_11
#define init_9_13_2_11
void FASTCALL call_9_13_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_2_12
#define init_9_13_2_12
void FASTCALL call_9_13_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_2_13
#define init_9_13_2_13
void FASTCALL call_9_13_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_2_14
#define init_9_13_2_14
void FASTCALL call_9_13_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_2_15
#define init_9_13_2_15
void FASTCALL call_9_13_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_3_0
#define init_9_13_3_0
void FASTCALL call_9_13_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_3_1
#define init_9_13_3_1
void FASTCALL call_9_13_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_3_2
#define init_9_13_3_2
void FASTCALL call_9_13_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_3_3
#define init_9_13_3_3
void FASTCALL call_9_13_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_3_4
#define init_9_13_3_4
void FASTCALL call_9_13_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_3_5
#define init_9_13_3_5
void FASTCALL call_9_13_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_3_6
#define init_9_13_3_6
void FASTCALL call_9_13_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_3_7
#define init_9_13_3_7
void FASTCALL call_9_13_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_3_8
#define init_9_13_3_8
void FASTCALL call_9_13_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_3_9
#define init_9_13_3_9
void FASTCALL call_9_13_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_3_10
#define init_9_13_3_10
void FASTCALL call_9_13_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_3_11
#define init_9_13_3_11
void FASTCALL call_9_13_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_3_12
#define init_9_13_3_12
void FASTCALL call_9_13_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_3_13
#define init_9_13_3_13
void FASTCALL call_9_13_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_3_14
#define init_9_13_3_14
void FASTCALL call_9_13_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_3_15
#define init_9_13_3_15
void FASTCALL call_9_13_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_4_0
#define init_9_13_4_0
void FASTCALL call_9_13_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_4_1
#define init_9_13_4_1
void FASTCALL call_9_13_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_4_2
#define init_9_13_4_2
void FASTCALL call_9_13_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_4_3
#define init_9_13_4_3
void FASTCALL call_9_13_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_4_4
#define init_9_13_4_4
void FASTCALL call_9_13_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_4_5
#define init_9_13_4_5
void FASTCALL call_9_13_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_4_6
#define init_9_13_4_6
void FASTCALL call_9_13_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_4_7
#define init_9_13_4_7
void FASTCALL call_9_13_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_4_8
#define init_9_13_4_8
void FASTCALL call_9_13_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_4_9
#define init_9_13_4_9
void FASTCALL call_9_13_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_4_10
#define init_9_13_4_10
void FASTCALL call_9_13_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_4_11
#define init_9_13_4_11
void FASTCALL call_9_13_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_4_12
#define init_9_13_4_12
void FASTCALL call_9_13_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_4_13
#define init_9_13_4_13
void FASTCALL call_9_13_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_4_14
#define init_9_13_4_14
void FASTCALL call_9_13_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_4_15
#define init_9_13_4_15
void FASTCALL call_9_13_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_5_0
#define init_9_13_5_0
void FASTCALL call_9_13_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_5_1
#define init_9_13_5_1
void FASTCALL call_9_13_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_5_2
#define init_9_13_5_2
void FASTCALL call_9_13_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_5_3
#define init_9_13_5_3
void FASTCALL call_9_13_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_5_4
#define init_9_13_5_4
void FASTCALL call_9_13_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_5_5
#define init_9_13_5_5
void FASTCALL call_9_13_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_5_6
#define init_9_13_5_6
void FASTCALL call_9_13_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_5_7
#define init_9_13_5_7
void FASTCALL call_9_13_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_5_8
#define init_9_13_5_8
void FASTCALL call_9_13_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_5_9
#define init_9_13_5_9
void FASTCALL call_9_13_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_5_10
#define init_9_13_5_10
void FASTCALL call_9_13_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_5_11
#define init_9_13_5_11
void FASTCALL call_9_13_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_5_12
#define init_9_13_5_12
void FASTCALL call_9_13_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_5_13
#define init_9_13_5_13
void FASTCALL call_9_13_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_5_14
#define init_9_13_5_14
void FASTCALL call_9_13_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_5_15
#define init_9_13_5_15
void FASTCALL call_9_13_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_6_0
#define init_9_13_6_0
void FASTCALL call_9_13_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_6_1
#define init_9_13_6_1
void FASTCALL call_9_13_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_6_2
#define init_9_13_6_2
void FASTCALL call_9_13_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_6_3
#define init_9_13_6_3
void FASTCALL call_9_13_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_6_4
#define init_9_13_6_4
void FASTCALL call_9_13_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_6_5
#define init_9_13_6_5
void FASTCALL call_9_13_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_6_6
#define init_9_13_6_6
void FASTCALL call_9_13_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_6_7
#define init_9_13_6_7
void FASTCALL call_9_13_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_6_8
#define init_9_13_6_8
void FASTCALL call_9_13_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_6_9
#define init_9_13_6_9
void FASTCALL call_9_13_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_6_10
#define init_9_13_6_10
void FASTCALL call_9_13_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_6_11
#define init_9_13_6_11
void FASTCALL call_9_13_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_6_12
#define init_9_13_6_12
void FASTCALL call_9_13_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_6_13
#define init_9_13_6_13
void FASTCALL call_9_13_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_6_14
#define init_9_13_6_14
void FASTCALL call_9_13_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_6_15
#define init_9_13_6_15
void FASTCALL call_9_13_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_7_0
#define init_9_13_7_0
void FASTCALL call_9_13_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_7_1
#define init_9_13_7_1
void FASTCALL call_9_13_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_7_2
#define init_9_13_7_2
void FASTCALL call_9_13_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_7_3
#define init_9_13_7_3
void FASTCALL call_9_13_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_7_4
#define init_9_13_7_4
void FASTCALL call_9_13_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_7_5
#define init_9_13_7_5
void FASTCALL call_9_13_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_7_6
#define init_9_13_7_6
void FASTCALL call_9_13_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_7_7
#define init_9_13_7_7
void FASTCALL call_9_13_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_7_8
#define init_9_13_7_8
void FASTCALL call_9_13_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_7_9
#define init_9_13_7_9
void FASTCALL call_9_13_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_7_10
#define init_9_13_7_10
void FASTCALL call_9_13_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_7_11
#define init_9_13_7_11
void FASTCALL call_9_13_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_7_12
#define init_9_13_7_12
void FASTCALL call_9_13_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_7_13
#define init_9_13_7_13
void FASTCALL call_9_13_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_7_14
#define init_9_13_7_14
void FASTCALL call_9_13_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_7_15
#define init_9_13_7_15
void FASTCALL call_9_13_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_8_0
#define init_9_13_8_0
void FASTCALL call_9_13_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_8_1
#define init_9_13_8_1
void FASTCALL call_9_13_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_8_2
#define init_9_13_8_2
void FASTCALL call_9_13_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_8_3
#define init_9_13_8_3
void FASTCALL call_9_13_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_8_4
#define init_9_13_8_4
void FASTCALL call_9_13_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_8_5
#define init_9_13_8_5
void FASTCALL call_9_13_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_8_6
#define init_9_13_8_6
void FASTCALL call_9_13_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_8_7
#define init_9_13_8_7
void FASTCALL call_9_13_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_8_8
#define init_9_13_8_8
void FASTCALL call_9_13_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_8_9
#define init_9_13_8_9
void FASTCALL call_9_13_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_8_10
#define init_9_13_8_10
void FASTCALL call_9_13_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_8_11
#define init_9_13_8_11
void FASTCALL call_9_13_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_8_12
#define init_9_13_8_12
void FASTCALL call_9_13_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_8_13
#define init_9_13_8_13
void FASTCALL call_9_13_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_8_14
#define init_9_13_8_14
void FASTCALL call_9_13_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_8_15
#define init_9_13_8_15
void FASTCALL call_9_13_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_9_0
#define init_9_13_9_0
void FASTCALL call_9_13_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_9_1
#define init_9_13_9_1
void FASTCALL call_9_13_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_9_2
#define init_9_13_9_2
void FASTCALL call_9_13_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_9_3
#define init_9_13_9_3
void FASTCALL call_9_13_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_9_4
#define init_9_13_9_4
void FASTCALL call_9_13_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_9_5
#define init_9_13_9_5
void FASTCALL call_9_13_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_9_6
#define init_9_13_9_6
void FASTCALL call_9_13_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_9_7
#define init_9_13_9_7
void FASTCALL call_9_13_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_9_8
#define init_9_13_9_8
void FASTCALL call_9_13_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_9_9
#define init_9_13_9_9
void FASTCALL call_9_13_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_9_10
#define init_9_13_9_10
void FASTCALL call_9_13_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_9_11
#define init_9_13_9_11
void FASTCALL call_9_13_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_9_12
#define init_9_13_9_12
void FASTCALL call_9_13_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_9_13
#define init_9_13_9_13
void FASTCALL call_9_13_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_9_14
#define init_9_13_9_14
void FASTCALL call_9_13_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_9_15
#define init_9_13_9_15
void FASTCALL call_9_13_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_10_0
#define init_9_13_10_0
void FASTCALL call_9_13_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_10_1
#define init_9_13_10_1
void FASTCALL call_9_13_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_10_2
#define init_9_13_10_2
void FASTCALL call_9_13_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_10_3
#define init_9_13_10_3
void FASTCALL call_9_13_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_10_4
#define init_9_13_10_4
void FASTCALL call_9_13_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_10_5
#define init_9_13_10_5
void FASTCALL call_9_13_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_10_6
#define init_9_13_10_6
void FASTCALL call_9_13_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_10_7
#define init_9_13_10_7
void FASTCALL call_9_13_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_10_8
#define init_9_13_10_8
void FASTCALL call_9_13_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_10_9
#define init_9_13_10_9
void FASTCALL call_9_13_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_10_10
#define init_9_13_10_10
void FASTCALL call_9_13_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_10_11
#define init_9_13_10_11
void FASTCALL call_9_13_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_10_12
#define init_9_13_10_12
void FASTCALL call_9_13_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_10_13
#define init_9_13_10_13
void FASTCALL call_9_13_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_10_14
#define init_9_13_10_14
void FASTCALL call_9_13_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_10_15
#define init_9_13_10_15
void FASTCALL call_9_13_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_11_0
#define init_9_13_11_0
void FASTCALL call_9_13_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_11_1
#define init_9_13_11_1
void FASTCALL call_9_13_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_11_2
#define init_9_13_11_2
void FASTCALL call_9_13_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_11_3
#define init_9_13_11_3
void FASTCALL call_9_13_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_11_4
#define init_9_13_11_4
void FASTCALL call_9_13_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_11_5
#define init_9_13_11_5
void FASTCALL call_9_13_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_11_6
#define init_9_13_11_6
void FASTCALL call_9_13_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_11_7
#define init_9_13_11_7
void FASTCALL call_9_13_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_11_8
#define init_9_13_11_8
void FASTCALL call_9_13_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_11_9
#define init_9_13_11_9
void FASTCALL call_9_13_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_11_10
#define init_9_13_11_10
void FASTCALL call_9_13_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_11_11
#define init_9_13_11_11
void FASTCALL call_9_13_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_11_12
#define init_9_13_11_12
void FASTCALL call_9_13_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_11_13
#define init_9_13_11_13
void FASTCALL call_9_13_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_11_14
#define init_9_13_11_14
void FASTCALL call_9_13_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_11_15
#define init_9_13_11_15
void FASTCALL call_9_13_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_12_0
#define init_9_13_12_0
void FASTCALL call_9_13_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_12_1
#define init_9_13_12_1
void FASTCALL call_9_13_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_12_2
#define init_9_13_12_2
void FASTCALL call_9_13_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_12_3
#define init_9_13_12_3
void FASTCALL call_9_13_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_12_4
#define init_9_13_12_4
void FASTCALL call_9_13_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_12_5
#define init_9_13_12_5
void FASTCALL call_9_13_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_12_6
#define init_9_13_12_6
void FASTCALL call_9_13_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_12_7
#define init_9_13_12_7
void FASTCALL call_9_13_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_12_8
#define init_9_13_12_8
void FASTCALL call_9_13_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_12_9
#define init_9_13_12_9
void FASTCALL call_9_13_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_12_10
#define init_9_13_12_10
void FASTCALL call_9_13_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_12_11
#define init_9_13_12_11
void FASTCALL call_9_13_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_12_12
#define init_9_13_12_12
void FASTCALL call_9_13_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_12_13
#define init_9_13_12_13
void FASTCALL call_9_13_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_12_14
#define init_9_13_12_14
void FASTCALL call_9_13_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_12_15
#define init_9_13_12_15
void FASTCALL call_9_13_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_13_0
#define init_9_13_13_0
void FASTCALL call_9_13_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_13_1
#define init_9_13_13_1
void FASTCALL call_9_13_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_13_2
#define init_9_13_13_2
void FASTCALL call_9_13_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_13_3
#define init_9_13_13_3
void FASTCALL call_9_13_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_13_4
#define init_9_13_13_4
void FASTCALL call_9_13_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_13_5
#define init_9_13_13_5
void FASTCALL call_9_13_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_13_6
#define init_9_13_13_6
void FASTCALL call_9_13_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_13_7
#define init_9_13_13_7
void FASTCALL call_9_13_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_13_8
#define init_9_13_13_8
void FASTCALL call_9_13_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_13_9
#define init_9_13_13_9
void FASTCALL call_9_13_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_13_10
#define init_9_13_13_10
void FASTCALL call_9_13_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_13_11
#define init_9_13_13_11
void FASTCALL call_9_13_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_13_12
#define init_9_13_13_12
void FASTCALL call_9_13_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_13_13
#define init_9_13_13_13
void FASTCALL call_9_13_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_13_14
#define init_9_13_13_14
void FASTCALL call_9_13_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_13_15
#define init_9_13_13_15
void FASTCALL call_9_13_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_14_0
#define init_9_13_14_0
void FASTCALL call_9_13_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_14_1
#define init_9_13_14_1
void FASTCALL call_9_13_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_14_2
#define init_9_13_14_2
void FASTCALL call_9_13_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_14_3
#define init_9_13_14_3
void FASTCALL call_9_13_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_14_4
#define init_9_13_14_4
void FASTCALL call_9_13_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_14_5
#define init_9_13_14_5
void FASTCALL call_9_13_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_14_6
#define init_9_13_14_6
void FASTCALL call_9_13_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_14_7
#define init_9_13_14_7
void FASTCALL call_9_13_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_14_8
#define init_9_13_14_8
void FASTCALL call_9_13_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_14_9
#define init_9_13_14_9
void FASTCALL call_9_13_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_14_10
#define init_9_13_14_10
void FASTCALL call_9_13_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_14_11
#define init_9_13_14_11
void FASTCALL call_9_13_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_14_12
#define init_9_13_14_12
void FASTCALL call_9_13_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_14_13
#define init_9_13_14_13
void FASTCALL call_9_13_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_14_14
#define init_9_13_14_14
void FASTCALL call_9_13_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_14_15
#define init_9_13_14_15
void FASTCALL call_9_13_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_15_0
#define init_9_13_15_0
void FASTCALL call_9_13_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_15_1
#define init_9_13_15_1
void FASTCALL call_9_13_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_15_2
#define init_9_13_15_2
void FASTCALL call_9_13_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_15_3
#define init_9_13_15_3
void FASTCALL call_9_13_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_15_4
#define init_9_13_15_4
void FASTCALL call_9_13_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_15_5
#define init_9_13_15_5
void FASTCALL call_9_13_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_15_6
#define init_9_13_15_6
void FASTCALL call_9_13_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_15_7
#define init_9_13_15_7
void FASTCALL call_9_13_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_15_8
#define init_9_13_15_8
void FASTCALL call_9_13_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_15_9
#define init_9_13_15_9
void FASTCALL call_9_13_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_15_10
#define init_9_13_15_10
void FASTCALL call_9_13_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_15_11
#define init_9_13_15_11
void FASTCALL call_9_13_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_15_12
#define init_9_13_15_12
void FASTCALL call_9_13_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_15_13
#define init_9_13_15_13
void FASTCALL call_9_13_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_15_14
#define init_9_13_15_14
void FASTCALL call_9_13_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_13_15_15
#define init_9_13_15_15
void FASTCALL call_9_13_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_0_0
#define init_9_14_0_0
void FASTCALL call_9_14_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_0_1
#define init_9_14_0_1
void FASTCALL call_9_14_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_0_2
#define init_9_14_0_2
void FASTCALL call_9_14_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_0_3
#define init_9_14_0_3
void FASTCALL call_9_14_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_0_4
#define init_9_14_0_4
void FASTCALL call_9_14_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_0_5
#define init_9_14_0_5
void FASTCALL call_9_14_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_0_6
#define init_9_14_0_6
void FASTCALL call_9_14_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_0_7
#define init_9_14_0_7
void FASTCALL call_9_14_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_0_8
#define init_9_14_0_8
void FASTCALL call_9_14_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_0_9
#define init_9_14_0_9
void FASTCALL call_9_14_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_0_10
#define init_9_14_0_10
void FASTCALL call_9_14_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_0_11
#define init_9_14_0_11
void FASTCALL call_9_14_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_0_12
#define init_9_14_0_12
void FASTCALL call_9_14_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_0_13
#define init_9_14_0_13
void FASTCALL call_9_14_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_0_14
#define init_9_14_0_14
void FASTCALL call_9_14_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_0_15
#define init_9_14_0_15
void FASTCALL call_9_14_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_1_0
#define init_9_14_1_0
void FASTCALL call_9_14_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_1_1
#define init_9_14_1_1
void FASTCALL call_9_14_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_1_2
#define init_9_14_1_2
void FASTCALL call_9_14_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_1_3
#define init_9_14_1_3
void FASTCALL call_9_14_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_1_4
#define init_9_14_1_4
void FASTCALL call_9_14_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_1_5
#define init_9_14_1_5
void FASTCALL call_9_14_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_1_6
#define init_9_14_1_6
void FASTCALL call_9_14_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_1_7
#define init_9_14_1_7
void FASTCALL call_9_14_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_1_8
#define init_9_14_1_8
void FASTCALL call_9_14_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_1_9
#define init_9_14_1_9
void FASTCALL call_9_14_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_1_10
#define init_9_14_1_10
void FASTCALL call_9_14_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_1_11
#define init_9_14_1_11
void FASTCALL call_9_14_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_1_12
#define init_9_14_1_12
void FASTCALL call_9_14_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_1_13
#define init_9_14_1_13
void FASTCALL call_9_14_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_1_14
#define init_9_14_1_14
void FASTCALL call_9_14_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_1_15
#define init_9_14_1_15
void FASTCALL call_9_14_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_2_0
#define init_9_14_2_0
void FASTCALL call_9_14_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_2_1
#define init_9_14_2_1
void FASTCALL call_9_14_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_2_2
#define init_9_14_2_2
void FASTCALL call_9_14_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_2_3
#define init_9_14_2_3
void FASTCALL call_9_14_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_2_4
#define init_9_14_2_4
void FASTCALL call_9_14_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_2_5
#define init_9_14_2_5
void FASTCALL call_9_14_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_2_6
#define init_9_14_2_6
void FASTCALL call_9_14_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_2_7
#define init_9_14_2_7
void FASTCALL call_9_14_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_2_8
#define init_9_14_2_8
void FASTCALL call_9_14_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_2_9
#define init_9_14_2_9
void FASTCALL call_9_14_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_2_10
#define init_9_14_2_10
void FASTCALL call_9_14_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_2_11
#define init_9_14_2_11
void FASTCALL call_9_14_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_2_12
#define init_9_14_2_12
void FASTCALL call_9_14_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_2_13
#define init_9_14_2_13
void FASTCALL call_9_14_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_2_14
#define init_9_14_2_14
void FASTCALL call_9_14_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_2_15
#define init_9_14_2_15
void FASTCALL call_9_14_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_3_0
#define init_9_14_3_0
void FASTCALL call_9_14_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_3_1
#define init_9_14_3_1
void FASTCALL call_9_14_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_3_2
#define init_9_14_3_2
void FASTCALL call_9_14_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_3_3
#define init_9_14_3_3
void FASTCALL call_9_14_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_3_4
#define init_9_14_3_4
void FASTCALL call_9_14_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_3_5
#define init_9_14_3_5
void FASTCALL call_9_14_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_3_6
#define init_9_14_3_6
void FASTCALL call_9_14_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_3_7
#define init_9_14_3_7
void FASTCALL call_9_14_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_3_8
#define init_9_14_3_8
void FASTCALL call_9_14_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_3_9
#define init_9_14_3_9
void FASTCALL call_9_14_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_3_10
#define init_9_14_3_10
void FASTCALL call_9_14_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_3_11
#define init_9_14_3_11
void FASTCALL call_9_14_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_3_12
#define init_9_14_3_12
void FASTCALL call_9_14_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_3_13
#define init_9_14_3_13
void FASTCALL call_9_14_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_3_14
#define init_9_14_3_14
void FASTCALL call_9_14_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_3_15
#define init_9_14_3_15
void FASTCALL call_9_14_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_4_0
#define init_9_14_4_0
void FASTCALL call_9_14_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_4_1
#define init_9_14_4_1
void FASTCALL call_9_14_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_4_2
#define init_9_14_4_2
void FASTCALL call_9_14_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_4_3
#define init_9_14_4_3
void FASTCALL call_9_14_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_4_4
#define init_9_14_4_4
void FASTCALL call_9_14_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_4_5
#define init_9_14_4_5
void FASTCALL call_9_14_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_4_6
#define init_9_14_4_6
void FASTCALL call_9_14_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_4_7
#define init_9_14_4_7
void FASTCALL call_9_14_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_4_8
#define init_9_14_4_8
void FASTCALL call_9_14_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_4_9
#define init_9_14_4_9
void FASTCALL call_9_14_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_4_10
#define init_9_14_4_10
void FASTCALL call_9_14_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_4_11
#define init_9_14_4_11
void FASTCALL call_9_14_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_4_12
#define init_9_14_4_12
void FASTCALL call_9_14_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_4_13
#define init_9_14_4_13
void FASTCALL call_9_14_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_4_14
#define init_9_14_4_14
void FASTCALL call_9_14_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_4_15
#define init_9_14_4_15
void FASTCALL call_9_14_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_5_0
#define init_9_14_5_0
void FASTCALL call_9_14_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_5_1
#define init_9_14_5_1
void FASTCALL call_9_14_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_5_2
#define init_9_14_5_2
void FASTCALL call_9_14_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_5_3
#define init_9_14_5_3
void FASTCALL call_9_14_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_5_4
#define init_9_14_5_4
void FASTCALL call_9_14_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_5_5
#define init_9_14_5_5
void FASTCALL call_9_14_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_5_6
#define init_9_14_5_6
void FASTCALL call_9_14_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_5_7
#define init_9_14_5_7
void FASTCALL call_9_14_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_5_8
#define init_9_14_5_8
void FASTCALL call_9_14_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_5_9
#define init_9_14_5_9
void FASTCALL call_9_14_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_5_10
#define init_9_14_5_10
void FASTCALL call_9_14_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_5_11
#define init_9_14_5_11
void FASTCALL call_9_14_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_5_12
#define init_9_14_5_12
void FASTCALL call_9_14_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_5_13
#define init_9_14_5_13
void FASTCALL call_9_14_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_5_14
#define init_9_14_5_14
void FASTCALL call_9_14_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_5_15
#define init_9_14_5_15
void FASTCALL call_9_14_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_6_0
#define init_9_14_6_0
void FASTCALL call_9_14_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_6_1
#define init_9_14_6_1
void FASTCALL call_9_14_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_6_2
#define init_9_14_6_2
void FASTCALL call_9_14_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_6_3
#define init_9_14_6_3
void FASTCALL call_9_14_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_6_4
#define init_9_14_6_4
void FASTCALL call_9_14_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_6_5
#define init_9_14_6_5
void FASTCALL call_9_14_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_6_6
#define init_9_14_6_6
void FASTCALL call_9_14_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_6_7
#define init_9_14_6_7
void FASTCALL call_9_14_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_6_8
#define init_9_14_6_8
void FASTCALL call_9_14_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_6_9
#define init_9_14_6_9
void FASTCALL call_9_14_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_6_10
#define init_9_14_6_10
void FASTCALL call_9_14_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_6_11
#define init_9_14_6_11
void FASTCALL call_9_14_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_6_12
#define init_9_14_6_12
void FASTCALL call_9_14_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_6_13
#define init_9_14_6_13
void FASTCALL call_9_14_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_6_14
#define init_9_14_6_14
void FASTCALL call_9_14_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_6_15
#define init_9_14_6_15
void FASTCALL call_9_14_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_7_0
#define init_9_14_7_0
void FASTCALL call_9_14_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_7_1
#define init_9_14_7_1
void FASTCALL call_9_14_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_7_2
#define init_9_14_7_2
void FASTCALL call_9_14_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_7_3
#define init_9_14_7_3
void FASTCALL call_9_14_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_7_4
#define init_9_14_7_4
void FASTCALL call_9_14_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_7_5
#define init_9_14_7_5
void FASTCALL call_9_14_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_7_6
#define init_9_14_7_6
void FASTCALL call_9_14_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_7_7
#define init_9_14_7_7
void FASTCALL call_9_14_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_7_8
#define init_9_14_7_8
void FASTCALL call_9_14_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_7_9
#define init_9_14_7_9
void FASTCALL call_9_14_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_7_10
#define init_9_14_7_10
void FASTCALL call_9_14_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_7_11
#define init_9_14_7_11
void FASTCALL call_9_14_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_7_12
#define init_9_14_7_12
void FASTCALL call_9_14_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_7_13
#define init_9_14_7_13
void FASTCALL call_9_14_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_7_14
#define init_9_14_7_14
void FASTCALL call_9_14_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_7_15
#define init_9_14_7_15
void FASTCALL call_9_14_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_8_0
#define init_9_14_8_0
void FASTCALL call_9_14_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_8_1
#define init_9_14_8_1
void FASTCALL call_9_14_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_8_2
#define init_9_14_8_2
void FASTCALL call_9_14_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_8_3
#define init_9_14_8_3
void FASTCALL call_9_14_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_8_4
#define init_9_14_8_4
void FASTCALL call_9_14_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_8_5
#define init_9_14_8_5
void FASTCALL call_9_14_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_8_6
#define init_9_14_8_6
void FASTCALL call_9_14_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_8_7
#define init_9_14_8_7
void FASTCALL call_9_14_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_8_8
#define init_9_14_8_8
void FASTCALL call_9_14_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_8_9
#define init_9_14_8_9
void FASTCALL call_9_14_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_8_10
#define init_9_14_8_10
void FASTCALL call_9_14_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_8_11
#define init_9_14_8_11
void FASTCALL call_9_14_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_8_12
#define init_9_14_8_12
void FASTCALL call_9_14_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_8_13
#define init_9_14_8_13
void FASTCALL call_9_14_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_8_14
#define init_9_14_8_14
void FASTCALL call_9_14_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_8_15
#define init_9_14_8_15
void FASTCALL call_9_14_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_9_0
#define init_9_14_9_0
void FASTCALL call_9_14_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_9_1
#define init_9_14_9_1
void FASTCALL call_9_14_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_9_2
#define init_9_14_9_2
void FASTCALL call_9_14_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_9_3
#define init_9_14_9_3
void FASTCALL call_9_14_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_9_4
#define init_9_14_9_4
void FASTCALL call_9_14_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_9_5
#define init_9_14_9_5
void FASTCALL call_9_14_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_9_6
#define init_9_14_9_6
void FASTCALL call_9_14_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_9_7
#define init_9_14_9_7
void FASTCALL call_9_14_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_9_8
#define init_9_14_9_8
void FASTCALL call_9_14_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_9_9
#define init_9_14_9_9
void FASTCALL call_9_14_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_9_10
#define init_9_14_9_10
void FASTCALL call_9_14_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_9_11
#define init_9_14_9_11
void FASTCALL call_9_14_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_9_12
#define init_9_14_9_12
void FASTCALL call_9_14_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_9_13
#define init_9_14_9_13
void FASTCALL call_9_14_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_9_14
#define init_9_14_9_14
void FASTCALL call_9_14_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_9_15
#define init_9_14_9_15
void FASTCALL call_9_14_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_10_0
#define init_9_14_10_0
void FASTCALL call_9_14_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_10_1
#define init_9_14_10_1
void FASTCALL call_9_14_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_10_2
#define init_9_14_10_2
void FASTCALL call_9_14_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_10_3
#define init_9_14_10_3
void FASTCALL call_9_14_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_10_4
#define init_9_14_10_4
void FASTCALL call_9_14_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_10_5
#define init_9_14_10_5
void FASTCALL call_9_14_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_10_6
#define init_9_14_10_6
void FASTCALL call_9_14_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_10_7
#define init_9_14_10_7
void FASTCALL call_9_14_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_10_8
#define init_9_14_10_8
void FASTCALL call_9_14_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_10_9
#define init_9_14_10_9
void FASTCALL call_9_14_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_10_10
#define init_9_14_10_10
void FASTCALL call_9_14_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_10_11
#define init_9_14_10_11
void FASTCALL call_9_14_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_10_12
#define init_9_14_10_12
void FASTCALL call_9_14_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_10_13
#define init_9_14_10_13
void FASTCALL call_9_14_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_10_14
#define init_9_14_10_14
void FASTCALL call_9_14_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_10_15
#define init_9_14_10_15
void FASTCALL call_9_14_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_11_0
#define init_9_14_11_0
void FASTCALL call_9_14_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_11_1
#define init_9_14_11_1
void FASTCALL call_9_14_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_11_2
#define init_9_14_11_2
void FASTCALL call_9_14_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_11_3
#define init_9_14_11_3
void FASTCALL call_9_14_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_11_4
#define init_9_14_11_4
void FASTCALL call_9_14_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_11_5
#define init_9_14_11_5
void FASTCALL call_9_14_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_11_6
#define init_9_14_11_6
void FASTCALL call_9_14_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_11_7
#define init_9_14_11_7
void FASTCALL call_9_14_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_11_8
#define init_9_14_11_8
void FASTCALL call_9_14_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_11_9
#define init_9_14_11_9
void FASTCALL call_9_14_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_11_10
#define init_9_14_11_10
void FASTCALL call_9_14_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_11_11
#define init_9_14_11_11
void FASTCALL call_9_14_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_11_12
#define init_9_14_11_12
void FASTCALL call_9_14_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_11_13
#define init_9_14_11_13
void FASTCALL call_9_14_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_11_14
#define init_9_14_11_14
void FASTCALL call_9_14_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_11_15
#define init_9_14_11_15
void FASTCALL call_9_14_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_12_0
#define init_9_14_12_0
void FASTCALL call_9_14_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_12_1
#define init_9_14_12_1
void FASTCALL call_9_14_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_12_2
#define init_9_14_12_2
void FASTCALL call_9_14_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_12_3
#define init_9_14_12_3
void FASTCALL call_9_14_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_12_4
#define init_9_14_12_4
void FASTCALL call_9_14_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_12_5
#define init_9_14_12_5
void FASTCALL call_9_14_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_12_6
#define init_9_14_12_6
void FASTCALL call_9_14_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_12_7
#define init_9_14_12_7
void FASTCALL call_9_14_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_12_8
#define init_9_14_12_8
void FASTCALL call_9_14_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_12_9
#define init_9_14_12_9
void FASTCALL call_9_14_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_12_10
#define init_9_14_12_10
void FASTCALL call_9_14_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_12_11
#define init_9_14_12_11
void FASTCALL call_9_14_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_12_12
#define init_9_14_12_12
void FASTCALL call_9_14_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_12_13
#define init_9_14_12_13
void FASTCALL call_9_14_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_12_14
#define init_9_14_12_14
void FASTCALL call_9_14_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_12_15
#define init_9_14_12_15
void FASTCALL call_9_14_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_13_0
#define init_9_14_13_0
void FASTCALL call_9_14_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_13_1
#define init_9_14_13_1
void FASTCALL call_9_14_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_13_2
#define init_9_14_13_2
void FASTCALL call_9_14_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_13_3
#define init_9_14_13_3
void FASTCALL call_9_14_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_13_4
#define init_9_14_13_4
void FASTCALL call_9_14_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_13_5
#define init_9_14_13_5
void FASTCALL call_9_14_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_13_6
#define init_9_14_13_6
void FASTCALL call_9_14_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_13_7
#define init_9_14_13_7
void FASTCALL call_9_14_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_13_8
#define init_9_14_13_8
void FASTCALL call_9_14_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_13_9
#define init_9_14_13_9
void FASTCALL call_9_14_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_13_10
#define init_9_14_13_10
void FASTCALL call_9_14_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_13_11
#define init_9_14_13_11
void FASTCALL call_9_14_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_13_12
#define init_9_14_13_12
void FASTCALL call_9_14_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_13_13
#define init_9_14_13_13
void FASTCALL call_9_14_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_13_14
#define init_9_14_13_14
void FASTCALL call_9_14_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_13_15
#define init_9_14_13_15
void FASTCALL call_9_14_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_14_0
#define init_9_14_14_0
void FASTCALL call_9_14_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_14_1
#define init_9_14_14_1
void FASTCALL call_9_14_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_14_2
#define init_9_14_14_2
void FASTCALL call_9_14_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_14_3
#define init_9_14_14_3
void FASTCALL call_9_14_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_14_4
#define init_9_14_14_4
void FASTCALL call_9_14_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_14_5
#define init_9_14_14_5
void FASTCALL call_9_14_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_14_6
#define init_9_14_14_6
void FASTCALL call_9_14_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_14_7
#define init_9_14_14_7
void FASTCALL call_9_14_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_14_8
#define init_9_14_14_8
void FASTCALL call_9_14_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_14_9
#define init_9_14_14_9
void FASTCALL call_9_14_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_14_10
#define init_9_14_14_10
void FASTCALL call_9_14_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_14_11
#define init_9_14_14_11
void FASTCALL call_9_14_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_14_12
#define init_9_14_14_12
void FASTCALL call_9_14_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_14_13
#define init_9_14_14_13
void FASTCALL call_9_14_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_14_14
#define init_9_14_14_14
void FASTCALL call_9_14_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_14_15
#define init_9_14_14_15
void FASTCALL call_9_14_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_15_0
#define init_9_14_15_0
void FASTCALL call_9_14_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_15_1
#define init_9_14_15_1
void FASTCALL call_9_14_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_15_2
#define init_9_14_15_2
void FASTCALL call_9_14_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_15_3
#define init_9_14_15_3
void FASTCALL call_9_14_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_15_4
#define init_9_14_15_4
void FASTCALL call_9_14_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_15_5
#define init_9_14_15_5
void FASTCALL call_9_14_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_15_6
#define init_9_14_15_6
void FASTCALL call_9_14_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_15_7
#define init_9_14_15_7
void FASTCALL call_9_14_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_15_8
#define init_9_14_15_8
void FASTCALL call_9_14_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_15_9
#define init_9_14_15_9
void FASTCALL call_9_14_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_15_10
#define init_9_14_15_10
void FASTCALL call_9_14_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_15_11
#define init_9_14_15_11
void FASTCALL call_9_14_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_15_12
#define init_9_14_15_12
void FASTCALL call_9_14_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_15_13
#define init_9_14_15_13
void FASTCALL call_9_14_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_15_14
#define init_9_14_15_14
void FASTCALL call_9_14_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_14_15_15
#define init_9_14_15_15
void FASTCALL call_9_14_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_0_0
#define init_9_15_0_0
void FASTCALL call_9_15_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_0_1
#define init_9_15_0_1
void FASTCALL call_9_15_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_0_2
#define init_9_15_0_2
void FASTCALL call_9_15_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_0_3
#define init_9_15_0_3
void FASTCALL call_9_15_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_0_4
#define init_9_15_0_4
void FASTCALL call_9_15_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_0_5
#define init_9_15_0_5
void FASTCALL call_9_15_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_0_6
#define init_9_15_0_6
void FASTCALL call_9_15_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_0_7
#define init_9_15_0_7
void FASTCALL call_9_15_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_0_8
#define init_9_15_0_8
void FASTCALL call_9_15_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_0_9
#define init_9_15_0_9
void FASTCALL call_9_15_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_0_10
#define init_9_15_0_10
void FASTCALL call_9_15_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_0_11
#define init_9_15_0_11
void FASTCALL call_9_15_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_0_12
#define init_9_15_0_12
void FASTCALL call_9_15_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_0_13
#define init_9_15_0_13
void FASTCALL call_9_15_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_0_14
#define init_9_15_0_14
void FASTCALL call_9_15_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_0_15
#define init_9_15_0_15
void FASTCALL call_9_15_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_1_0
#define init_9_15_1_0
void FASTCALL call_9_15_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_1_1
#define init_9_15_1_1
void FASTCALL call_9_15_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_1_2
#define init_9_15_1_2
void FASTCALL call_9_15_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_1_3
#define init_9_15_1_3
void FASTCALL call_9_15_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_1_4
#define init_9_15_1_4
void FASTCALL call_9_15_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_1_5
#define init_9_15_1_5
void FASTCALL call_9_15_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_1_6
#define init_9_15_1_6
void FASTCALL call_9_15_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_1_7
#define init_9_15_1_7
void FASTCALL call_9_15_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_1_8
#define init_9_15_1_8
void FASTCALL call_9_15_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_1_9
#define init_9_15_1_9
void FASTCALL call_9_15_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_1_10
#define init_9_15_1_10
void FASTCALL call_9_15_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_1_11
#define init_9_15_1_11
void FASTCALL call_9_15_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_1_12
#define init_9_15_1_12
void FASTCALL call_9_15_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_1_13
#define init_9_15_1_13
void FASTCALL call_9_15_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_1_14
#define init_9_15_1_14
void FASTCALL call_9_15_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_1_15
#define init_9_15_1_15
void FASTCALL call_9_15_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_2_0
#define init_9_15_2_0
void FASTCALL call_9_15_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_2_1
#define init_9_15_2_1
void FASTCALL call_9_15_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_2_2
#define init_9_15_2_2
void FASTCALL call_9_15_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_2_3
#define init_9_15_2_3
void FASTCALL call_9_15_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_2_4
#define init_9_15_2_4
void FASTCALL call_9_15_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_2_5
#define init_9_15_2_5
void FASTCALL call_9_15_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_2_6
#define init_9_15_2_6
void FASTCALL call_9_15_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_2_7
#define init_9_15_2_7
void FASTCALL call_9_15_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_2_8
#define init_9_15_2_8
void FASTCALL call_9_15_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_2_9
#define init_9_15_2_9
void FASTCALL call_9_15_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_2_10
#define init_9_15_2_10
void FASTCALL call_9_15_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_2_11
#define init_9_15_2_11
void FASTCALL call_9_15_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_2_12
#define init_9_15_2_12
void FASTCALL call_9_15_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_2_13
#define init_9_15_2_13
void FASTCALL call_9_15_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_2_14
#define init_9_15_2_14
void FASTCALL call_9_15_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_2_15
#define init_9_15_2_15
void FASTCALL call_9_15_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_3_0
#define init_9_15_3_0
void FASTCALL call_9_15_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_3_1
#define init_9_15_3_1
void FASTCALL call_9_15_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_3_2
#define init_9_15_3_2
void FASTCALL call_9_15_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_3_3
#define init_9_15_3_3
void FASTCALL call_9_15_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_3_4
#define init_9_15_3_4
void FASTCALL call_9_15_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_3_5
#define init_9_15_3_5
void FASTCALL call_9_15_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_3_6
#define init_9_15_3_6
void FASTCALL call_9_15_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_3_7
#define init_9_15_3_7
void FASTCALL call_9_15_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_3_8
#define init_9_15_3_8
void FASTCALL call_9_15_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_3_9
#define init_9_15_3_9
void FASTCALL call_9_15_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_3_10
#define init_9_15_3_10
void FASTCALL call_9_15_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_3_11
#define init_9_15_3_11
void FASTCALL call_9_15_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_3_12
#define init_9_15_3_12
void FASTCALL call_9_15_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_3_13
#define init_9_15_3_13
void FASTCALL call_9_15_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_3_14
#define init_9_15_3_14
void FASTCALL call_9_15_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_3_15
#define init_9_15_3_15
void FASTCALL call_9_15_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_4_0
#define init_9_15_4_0
void FASTCALL call_9_15_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_4_1
#define init_9_15_4_1
void FASTCALL call_9_15_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_4_2
#define init_9_15_4_2
void FASTCALL call_9_15_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_4_3
#define init_9_15_4_3
void FASTCALL call_9_15_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_4_4
#define init_9_15_4_4
void FASTCALL call_9_15_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_4_5
#define init_9_15_4_5
void FASTCALL call_9_15_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_4_6
#define init_9_15_4_6
void FASTCALL call_9_15_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_4_7
#define init_9_15_4_7
void FASTCALL call_9_15_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_4_8
#define init_9_15_4_8
void FASTCALL call_9_15_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_4_9
#define init_9_15_4_9
void FASTCALL call_9_15_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_4_10
#define init_9_15_4_10
void FASTCALL call_9_15_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_4_11
#define init_9_15_4_11
void FASTCALL call_9_15_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_4_12
#define init_9_15_4_12
void FASTCALL call_9_15_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_4_13
#define init_9_15_4_13
void FASTCALL call_9_15_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_4_14
#define init_9_15_4_14
void FASTCALL call_9_15_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_4_15
#define init_9_15_4_15
void FASTCALL call_9_15_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_5_0
#define init_9_15_5_0
void FASTCALL call_9_15_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_5_1
#define init_9_15_5_1
void FASTCALL call_9_15_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_5_2
#define init_9_15_5_2
void FASTCALL call_9_15_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_5_3
#define init_9_15_5_3
void FASTCALL call_9_15_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_5_4
#define init_9_15_5_4
void FASTCALL call_9_15_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_5_5
#define init_9_15_5_5
void FASTCALL call_9_15_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_5_6
#define init_9_15_5_6
void FASTCALL call_9_15_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_5_7
#define init_9_15_5_7
void FASTCALL call_9_15_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_5_8
#define init_9_15_5_8
void FASTCALL call_9_15_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_5_9
#define init_9_15_5_9
void FASTCALL call_9_15_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_5_10
#define init_9_15_5_10
void FASTCALL call_9_15_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_5_11
#define init_9_15_5_11
void FASTCALL call_9_15_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_5_12
#define init_9_15_5_12
void FASTCALL call_9_15_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_5_13
#define init_9_15_5_13
void FASTCALL call_9_15_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_5_14
#define init_9_15_5_14
void FASTCALL call_9_15_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_5_15
#define init_9_15_5_15
void FASTCALL call_9_15_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_6_0
#define init_9_15_6_0
void FASTCALL call_9_15_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_6_1
#define init_9_15_6_1
void FASTCALL call_9_15_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_6_2
#define init_9_15_6_2
void FASTCALL call_9_15_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_6_3
#define init_9_15_6_3
void FASTCALL call_9_15_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_6_4
#define init_9_15_6_4
void FASTCALL call_9_15_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_6_5
#define init_9_15_6_5
void FASTCALL call_9_15_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_6_6
#define init_9_15_6_6
void FASTCALL call_9_15_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_6_7
#define init_9_15_6_7
void FASTCALL call_9_15_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_6_8
#define init_9_15_6_8
void FASTCALL call_9_15_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_6_9
#define init_9_15_6_9
void FASTCALL call_9_15_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_6_10
#define init_9_15_6_10
void FASTCALL call_9_15_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_6_11
#define init_9_15_6_11
void FASTCALL call_9_15_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_6_12
#define init_9_15_6_12
void FASTCALL call_9_15_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_6_13
#define init_9_15_6_13
void FASTCALL call_9_15_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_6_14
#define init_9_15_6_14
void FASTCALL call_9_15_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_6_15
#define init_9_15_6_15
void FASTCALL call_9_15_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_7_0
#define init_9_15_7_0
void FASTCALL call_9_15_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_7_1
#define init_9_15_7_1
void FASTCALL call_9_15_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_7_2
#define init_9_15_7_2
void FASTCALL call_9_15_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_7_3
#define init_9_15_7_3
void FASTCALL call_9_15_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_7_4
#define init_9_15_7_4
void FASTCALL call_9_15_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_7_5
#define init_9_15_7_5
void FASTCALL call_9_15_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_7_6
#define init_9_15_7_6
void FASTCALL call_9_15_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_7_7
#define init_9_15_7_7
void FASTCALL call_9_15_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_7_8
#define init_9_15_7_8
void FASTCALL call_9_15_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_7_9
#define init_9_15_7_9
void FASTCALL call_9_15_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_7_10
#define init_9_15_7_10
void FASTCALL call_9_15_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_7_11
#define init_9_15_7_11
void FASTCALL call_9_15_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_7_12
#define init_9_15_7_12
void FASTCALL call_9_15_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_7_13
#define init_9_15_7_13
void FASTCALL call_9_15_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_7_14
#define init_9_15_7_14
void FASTCALL call_9_15_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_7_15
#define init_9_15_7_15
void FASTCALL call_9_15_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_8_0
#define init_9_15_8_0
void FASTCALL call_9_15_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_8_1
#define init_9_15_8_1
void FASTCALL call_9_15_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_8_2
#define init_9_15_8_2
void FASTCALL call_9_15_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_8_3
#define init_9_15_8_3
void FASTCALL call_9_15_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_8_4
#define init_9_15_8_4
void FASTCALL call_9_15_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_8_5
#define init_9_15_8_5
void FASTCALL call_9_15_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_8_6
#define init_9_15_8_6
void FASTCALL call_9_15_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_8_7
#define init_9_15_8_7
void FASTCALL call_9_15_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_8_8
#define init_9_15_8_8
void FASTCALL call_9_15_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_8_9
#define init_9_15_8_9
void FASTCALL call_9_15_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_8_10
#define init_9_15_8_10
void FASTCALL call_9_15_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_8_11
#define init_9_15_8_11
void FASTCALL call_9_15_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_8_12
#define init_9_15_8_12
void FASTCALL call_9_15_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_8_13
#define init_9_15_8_13
void FASTCALL call_9_15_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_8_14
#define init_9_15_8_14
void FASTCALL call_9_15_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_8_15
#define init_9_15_8_15
void FASTCALL call_9_15_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_9_0
#define init_9_15_9_0
void FASTCALL call_9_15_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_9_1
#define init_9_15_9_1
void FASTCALL call_9_15_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_9_2
#define init_9_15_9_2
void FASTCALL call_9_15_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_9_3
#define init_9_15_9_3
void FASTCALL call_9_15_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_9_4
#define init_9_15_9_4
void FASTCALL call_9_15_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_9_5
#define init_9_15_9_5
void FASTCALL call_9_15_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_9_6
#define init_9_15_9_6
void FASTCALL call_9_15_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_9_7
#define init_9_15_9_7
void FASTCALL call_9_15_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_9_8
#define init_9_15_9_8
void FASTCALL call_9_15_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_9_9
#define init_9_15_9_9
void FASTCALL call_9_15_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_9_10
#define init_9_15_9_10
void FASTCALL call_9_15_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_9_11
#define init_9_15_9_11
void FASTCALL call_9_15_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_9_12
#define init_9_15_9_12
void FASTCALL call_9_15_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_9_13
#define init_9_15_9_13
void FASTCALL call_9_15_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_9_14
#define init_9_15_9_14
void FASTCALL call_9_15_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_9_15
#define init_9_15_9_15
void FASTCALL call_9_15_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_10_0
#define init_9_15_10_0
void FASTCALL call_9_15_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_10_1
#define init_9_15_10_1
void FASTCALL call_9_15_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_10_2
#define init_9_15_10_2
void FASTCALL call_9_15_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_10_3
#define init_9_15_10_3
void FASTCALL call_9_15_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_10_4
#define init_9_15_10_4
void FASTCALL call_9_15_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_10_5
#define init_9_15_10_5
void FASTCALL call_9_15_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_10_6
#define init_9_15_10_6
void FASTCALL call_9_15_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_10_7
#define init_9_15_10_7
void FASTCALL call_9_15_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_10_8
#define init_9_15_10_8
void FASTCALL call_9_15_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_10_9
#define init_9_15_10_9
void FASTCALL call_9_15_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_10_10
#define init_9_15_10_10
void FASTCALL call_9_15_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_10_11
#define init_9_15_10_11
void FASTCALL call_9_15_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_10_12
#define init_9_15_10_12
void FASTCALL call_9_15_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_10_13
#define init_9_15_10_13
void FASTCALL call_9_15_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_10_14
#define init_9_15_10_14
void FASTCALL call_9_15_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_10_15
#define init_9_15_10_15
void FASTCALL call_9_15_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_11_0
#define init_9_15_11_0
void FASTCALL call_9_15_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_11_1
#define init_9_15_11_1
void FASTCALL call_9_15_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_11_2
#define init_9_15_11_2
void FASTCALL call_9_15_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_11_3
#define init_9_15_11_3
void FASTCALL call_9_15_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_11_4
#define init_9_15_11_4
void FASTCALL call_9_15_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_11_5
#define init_9_15_11_5
void FASTCALL call_9_15_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_11_6
#define init_9_15_11_6
void FASTCALL call_9_15_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_11_7
#define init_9_15_11_7
void FASTCALL call_9_15_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_11_8
#define init_9_15_11_8
void FASTCALL call_9_15_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_11_9
#define init_9_15_11_9
void FASTCALL call_9_15_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_11_10
#define init_9_15_11_10
void FASTCALL call_9_15_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_11_11
#define init_9_15_11_11
void FASTCALL call_9_15_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_11_12
#define init_9_15_11_12
void FASTCALL call_9_15_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_11_13
#define init_9_15_11_13
void FASTCALL call_9_15_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_11_14
#define init_9_15_11_14
void FASTCALL call_9_15_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_11_15
#define init_9_15_11_15
void FASTCALL call_9_15_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_12_0
#define init_9_15_12_0
void FASTCALL call_9_15_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_12_1
#define init_9_15_12_1
void FASTCALL call_9_15_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_12_2
#define init_9_15_12_2
void FASTCALL call_9_15_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_12_3
#define init_9_15_12_3
void FASTCALL call_9_15_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_12_4
#define init_9_15_12_4
void FASTCALL call_9_15_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_12_5
#define init_9_15_12_5
void FASTCALL call_9_15_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_12_6
#define init_9_15_12_6
void FASTCALL call_9_15_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_12_7
#define init_9_15_12_7
void FASTCALL call_9_15_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_12_8
#define init_9_15_12_8
void FASTCALL call_9_15_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_12_9
#define init_9_15_12_9
void FASTCALL call_9_15_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_12_10
#define init_9_15_12_10
void FASTCALL call_9_15_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_12_11
#define init_9_15_12_11
void FASTCALL call_9_15_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_12_12
#define init_9_15_12_12
void FASTCALL call_9_15_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_12_13
#define init_9_15_12_13
void FASTCALL call_9_15_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_12_14
#define init_9_15_12_14
void FASTCALL call_9_15_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_12_15
#define init_9_15_12_15
void FASTCALL call_9_15_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_13_0
#define init_9_15_13_0
void FASTCALL call_9_15_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_13_1
#define init_9_15_13_1
void FASTCALL call_9_15_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_13_2
#define init_9_15_13_2
void FASTCALL call_9_15_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_13_3
#define init_9_15_13_3
void FASTCALL call_9_15_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_13_4
#define init_9_15_13_4
void FASTCALL call_9_15_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_13_5
#define init_9_15_13_5
void FASTCALL call_9_15_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_13_6
#define init_9_15_13_6
void FASTCALL call_9_15_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_13_7
#define init_9_15_13_7
void FASTCALL call_9_15_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_13_8
#define init_9_15_13_8
void FASTCALL call_9_15_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_13_9
#define init_9_15_13_9
void FASTCALL call_9_15_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_13_10
#define init_9_15_13_10
void FASTCALL call_9_15_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_13_11
#define init_9_15_13_11
void FASTCALL call_9_15_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_13_12
#define init_9_15_13_12
void FASTCALL call_9_15_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_13_13
#define init_9_15_13_13
void FASTCALL call_9_15_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_13_14
#define init_9_15_13_14
void FASTCALL call_9_15_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_13_15
#define init_9_15_13_15
void FASTCALL call_9_15_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_14_0
#define init_9_15_14_0
void FASTCALL call_9_15_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_14_1
#define init_9_15_14_1
void FASTCALL call_9_15_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_14_2
#define init_9_15_14_2
void FASTCALL call_9_15_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_14_3
#define init_9_15_14_3
void FASTCALL call_9_15_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_14_4
#define init_9_15_14_4
void FASTCALL call_9_15_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_14_5
#define init_9_15_14_5
void FASTCALL call_9_15_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_14_6
#define init_9_15_14_6
void FASTCALL call_9_15_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_14_7
#define init_9_15_14_7
void FASTCALL call_9_15_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_14_8
#define init_9_15_14_8
void FASTCALL call_9_15_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_14_9
#define init_9_15_14_9
void FASTCALL call_9_15_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_14_10
#define init_9_15_14_10
void FASTCALL call_9_15_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_14_11
#define init_9_15_14_11
void FASTCALL call_9_15_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_14_12
#define init_9_15_14_12
void FASTCALL call_9_15_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_14_13
#define init_9_15_14_13
void FASTCALL call_9_15_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_14_14
#define init_9_15_14_14
void FASTCALL call_9_15_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_14_15
#define init_9_15_14_15
void FASTCALL call_9_15_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_15_0
#define init_9_15_15_0
void FASTCALL call_9_15_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_15_1
#define init_9_15_15_1
void FASTCALL call_9_15_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_15_2
#define init_9_15_15_2
void FASTCALL call_9_15_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_15_3
#define init_9_15_15_3
void FASTCALL call_9_15_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_15_4
#define init_9_15_15_4
void FASTCALL call_9_15_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_15_5
#define init_9_15_15_5
void FASTCALL call_9_15_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_15_6
#define init_9_15_15_6
void FASTCALL call_9_15_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_15_7
#define init_9_15_15_7
void FASTCALL call_9_15_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_15_8
#define init_9_15_15_8
void FASTCALL call_9_15_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_15_9
#define init_9_15_15_9
void FASTCALL call_9_15_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_15_10
#define init_9_15_15_10
void FASTCALL call_9_15_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_15_11
#define init_9_15_15_11
void FASTCALL call_9_15_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_15_12
#define init_9_15_15_12
void FASTCALL call_9_15_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_15_13
#define init_9_15_15_13
void FASTCALL call_9_15_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_15_14
#define init_9_15_15_14
void FASTCALL call_9_15_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_9_15_15_15
#define init_9_15_15_15
void FASTCALL call_9_15_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_0_0
#define init_10_0_0_0
void FASTCALL call_10_0_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_0_1
#define init_10_0_0_1
void FASTCALL call_10_0_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_0_2
#define init_10_0_0_2
void FASTCALL call_10_0_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_0_3
#define init_10_0_0_3
void FASTCALL call_10_0_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_0_4
#define init_10_0_0_4
void FASTCALL call_10_0_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_0_5
#define init_10_0_0_5
void FASTCALL call_10_0_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_0_6
#define init_10_0_0_6
void FASTCALL call_10_0_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_0_7
#define init_10_0_0_7
void FASTCALL call_10_0_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_0_8
#define init_10_0_0_8
void FASTCALL call_10_0_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_0_9
#define init_10_0_0_9
void FASTCALL call_10_0_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_0_10
#define init_10_0_0_10
void FASTCALL call_10_0_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_0_11
#define init_10_0_0_11
void FASTCALL call_10_0_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_0_12
#define init_10_0_0_12
void FASTCALL call_10_0_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_0_13
#define init_10_0_0_13
void FASTCALL call_10_0_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_0_14
#define init_10_0_0_14
void FASTCALL call_10_0_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_0_15
#define init_10_0_0_15
void FASTCALL call_10_0_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_1_0
#define init_10_0_1_0
void FASTCALL call_10_0_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_1_1
#define init_10_0_1_1
void FASTCALL call_10_0_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_1_2
#define init_10_0_1_2
void FASTCALL call_10_0_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_1_3
#define init_10_0_1_3
void FASTCALL call_10_0_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_1_4
#define init_10_0_1_4
void FASTCALL call_10_0_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_1_5
#define init_10_0_1_5
void FASTCALL call_10_0_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_1_6
#define init_10_0_1_6
void FASTCALL call_10_0_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_1_7
#define init_10_0_1_7
void FASTCALL call_10_0_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_1_8
#define init_10_0_1_8
void FASTCALL call_10_0_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_1_9
#define init_10_0_1_9
void FASTCALL call_10_0_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_1_10
#define init_10_0_1_10
void FASTCALL call_10_0_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_1_11
#define init_10_0_1_11
void FASTCALL call_10_0_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_1_12
#define init_10_0_1_12
void FASTCALL call_10_0_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_1_13
#define init_10_0_1_13
void FASTCALL call_10_0_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_1_14
#define init_10_0_1_14
void FASTCALL call_10_0_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_1_15
#define init_10_0_1_15
void FASTCALL call_10_0_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_2_0
#define init_10_0_2_0
void FASTCALL call_10_0_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_2_1
#define init_10_0_2_1
void FASTCALL call_10_0_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_2_2
#define init_10_0_2_2
void FASTCALL call_10_0_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_2_3
#define init_10_0_2_3
void FASTCALL call_10_0_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_2_4
#define init_10_0_2_4
void FASTCALL call_10_0_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_2_5
#define init_10_0_2_5
void FASTCALL call_10_0_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_2_6
#define init_10_0_2_6
void FASTCALL call_10_0_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_2_7
#define init_10_0_2_7
void FASTCALL call_10_0_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_2_8
#define init_10_0_2_8
void FASTCALL call_10_0_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_2_9
#define init_10_0_2_9
void FASTCALL call_10_0_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_2_10
#define init_10_0_2_10
void FASTCALL call_10_0_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_2_11
#define init_10_0_2_11
void FASTCALL call_10_0_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_2_12
#define init_10_0_2_12
void FASTCALL call_10_0_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_2_13
#define init_10_0_2_13
void FASTCALL call_10_0_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_2_14
#define init_10_0_2_14
void FASTCALL call_10_0_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_2_15
#define init_10_0_2_15
void FASTCALL call_10_0_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_3_0
#define init_10_0_3_0
void FASTCALL call_10_0_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_3_1
#define init_10_0_3_1
void FASTCALL call_10_0_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_3_2
#define init_10_0_3_2
void FASTCALL call_10_0_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_3_3
#define init_10_0_3_3
void FASTCALL call_10_0_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_3_4
#define init_10_0_3_4
void FASTCALL call_10_0_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_3_5
#define init_10_0_3_5
void FASTCALL call_10_0_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_3_6
#define init_10_0_3_6
void FASTCALL call_10_0_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_3_7
#define init_10_0_3_7
void FASTCALL call_10_0_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_3_8
#define init_10_0_3_8
void FASTCALL call_10_0_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_3_9
#define init_10_0_3_9
void FASTCALL call_10_0_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_3_10
#define init_10_0_3_10
void FASTCALL call_10_0_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_3_11
#define init_10_0_3_11
void FASTCALL call_10_0_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_3_12
#define init_10_0_3_12
void FASTCALL call_10_0_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_3_13
#define init_10_0_3_13
void FASTCALL call_10_0_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_3_14
#define init_10_0_3_14
void FASTCALL call_10_0_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_3_15
#define init_10_0_3_15
void FASTCALL call_10_0_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_4_0
#define init_10_0_4_0
void FASTCALL call_10_0_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_4_1
#define init_10_0_4_1
void FASTCALL call_10_0_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_4_2
#define init_10_0_4_2
void FASTCALL call_10_0_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_4_3
#define init_10_0_4_3
void FASTCALL call_10_0_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_4_4
#define init_10_0_4_4
void FASTCALL call_10_0_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_4_5
#define init_10_0_4_5
void FASTCALL call_10_0_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_4_6
#define init_10_0_4_6
void FASTCALL call_10_0_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_4_7
#define init_10_0_4_7
void FASTCALL call_10_0_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_4_8
#define init_10_0_4_8
void FASTCALL call_10_0_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_4_9
#define init_10_0_4_9
void FASTCALL call_10_0_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_4_10
#define init_10_0_4_10
void FASTCALL call_10_0_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_4_11
#define init_10_0_4_11
void FASTCALL call_10_0_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_4_12
#define init_10_0_4_12
void FASTCALL call_10_0_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_4_13
#define init_10_0_4_13
void FASTCALL call_10_0_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_4_14
#define init_10_0_4_14
void FASTCALL call_10_0_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_4_15
#define init_10_0_4_15
void FASTCALL call_10_0_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_5_0
#define init_10_0_5_0
void FASTCALL call_10_0_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_5_1
#define init_10_0_5_1
void FASTCALL call_10_0_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_5_2
#define init_10_0_5_2
void FASTCALL call_10_0_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_5_3
#define init_10_0_5_3
void FASTCALL call_10_0_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_5_4
#define init_10_0_5_4
void FASTCALL call_10_0_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_5_5
#define init_10_0_5_5
void FASTCALL call_10_0_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_5_6
#define init_10_0_5_6
void FASTCALL call_10_0_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_5_7
#define init_10_0_5_7
void FASTCALL call_10_0_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_5_8
#define init_10_0_5_8
void FASTCALL call_10_0_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_5_9
#define init_10_0_5_9
void FASTCALL call_10_0_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_5_10
#define init_10_0_5_10
void FASTCALL call_10_0_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_5_11
#define init_10_0_5_11
void FASTCALL call_10_0_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_5_12
#define init_10_0_5_12
void FASTCALL call_10_0_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_5_13
#define init_10_0_5_13
void FASTCALL call_10_0_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_5_14
#define init_10_0_5_14
void FASTCALL call_10_0_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_5_15
#define init_10_0_5_15
void FASTCALL call_10_0_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_6_0
#define init_10_0_6_0
void FASTCALL call_10_0_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_6_1
#define init_10_0_6_1
void FASTCALL call_10_0_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_6_2
#define init_10_0_6_2
void FASTCALL call_10_0_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_6_3
#define init_10_0_6_3
void FASTCALL call_10_0_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_6_4
#define init_10_0_6_4
void FASTCALL call_10_0_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_6_5
#define init_10_0_6_5
void FASTCALL call_10_0_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_6_6
#define init_10_0_6_6
void FASTCALL call_10_0_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_6_7
#define init_10_0_6_7
void FASTCALL call_10_0_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_6_8
#define init_10_0_6_8
void FASTCALL call_10_0_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_6_9
#define init_10_0_6_9
void FASTCALL call_10_0_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_6_10
#define init_10_0_6_10
void FASTCALL call_10_0_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_6_11
#define init_10_0_6_11
void FASTCALL call_10_0_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_6_12
#define init_10_0_6_12
void FASTCALL call_10_0_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_6_13
#define init_10_0_6_13
void FASTCALL call_10_0_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_6_14
#define init_10_0_6_14
void FASTCALL call_10_0_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_6_15
#define init_10_0_6_15
void FASTCALL call_10_0_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_7_0
#define init_10_0_7_0
void FASTCALL call_10_0_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_7_1
#define init_10_0_7_1
void FASTCALL call_10_0_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_7_2
#define init_10_0_7_2
void FASTCALL call_10_0_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_7_3
#define init_10_0_7_3
void FASTCALL call_10_0_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_7_4
#define init_10_0_7_4
void FASTCALL call_10_0_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_7_5
#define init_10_0_7_5
void FASTCALL call_10_0_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_7_6
#define init_10_0_7_6
void FASTCALL call_10_0_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_7_7
#define init_10_0_7_7
void FASTCALL call_10_0_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_7_8
#define init_10_0_7_8
void FASTCALL call_10_0_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_7_9
#define init_10_0_7_9
void FASTCALL call_10_0_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_7_10
#define init_10_0_7_10
void FASTCALL call_10_0_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_7_11
#define init_10_0_7_11
void FASTCALL call_10_0_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_7_12
#define init_10_0_7_12
void FASTCALL call_10_0_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_7_13
#define init_10_0_7_13
void FASTCALL call_10_0_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_7_14
#define init_10_0_7_14
void FASTCALL call_10_0_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_7_15
#define init_10_0_7_15
void FASTCALL call_10_0_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_8_0
#define init_10_0_8_0
void FASTCALL call_10_0_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_8_1
#define init_10_0_8_1
void FASTCALL call_10_0_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_8_2
#define init_10_0_8_2
void FASTCALL call_10_0_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_8_3
#define init_10_0_8_3
void FASTCALL call_10_0_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_8_4
#define init_10_0_8_4
void FASTCALL call_10_0_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_8_5
#define init_10_0_8_5
void FASTCALL call_10_0_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_8_6
#define init_10_0_8_6
void FASTCALL call_10_0_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_8_7
#define init_10_0_8_7
void FASTCALL call_10_0_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_8_8
#define init_10_0_8_8
void FASTCALL call_10_0_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_8_9
#define init_10_0_8_9
void FASTCALL call_10_0_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_8_10
#define init_10_0_8_10
void FASTCALL call_10_0_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_8_11
#define init_10_0_8_11
void FASTCALL call_10_0_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_8_12
#define init_10_0_8_12
void FASTCALL call_10_0_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_8_13
#define init_10_0_8_13
void FASTCALL call_10_0_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_8_14
#define init_10_0_8_14
void FASTCALL call_10_0_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_8_15
#define init_10_0_8_15
void FASTCALL call_10_0_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_9_0
#define init_10_0_9_0
void FASTCALL call_10_0_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_9_1
#define init_10_0_9_1
void FASTCALL call_10_0_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_9_2
#define init_10_0_9_2
void FASTCALL call_10_0_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_9_3
#define init_10_0_9_3
void FASTCALL call_10_0_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_9_4
#define init_10_0_9_4
void FASTCALL call_10_0_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_9_5
#define init_10_0_9_5
void FASTCALL call_10_0_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_9_6
#define init_10_0_9_6
void FASTCALL call_10_0_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_9_7
#define init_10_0_9_7
void FASTCALL call_10_0_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_9_8
#define init_10_0_9_8
void FASTCALL call_10_0_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_9_9
#define init_10_0_9_9
void FASTCALL call_10_0_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_9_10
#define init_10_0_9_10
void FASTCALL call_10_0_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_9_11
#define init_10_0_9_11
void FASTCALL call_10_0_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_9_12
#define init_10_0_9_12
void FASTCALL call_10_0_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_9_13
#define init_10_0_9_13
void FASTCALL call_10_0_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_9_14
#define init_10_0_9_14
void FASTCALL call_10_0_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_9_15
#define init_10_0_9_15
void FASTCALL call_10_0_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_10_0
#define init_10_0_10_0
void FASTCALL call_10_0_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_10_1
#define init_10_0_10_1
void FASTCALL call_10_0_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_10_2
#define init_10_0_10_2
void FASTCALL call_10_0_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_10_3
#define init_10_0_10_3
void FASTCALL call_10_0_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_10_4
#define init_10_0_10_4
void FASTCALL call_10_0_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_10_5
#define init_10_0_10_5
void FASTCALL call_10_0_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_10_6
#define init_10_0_10_6
void FASTCALL call_10_0_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_10_7
#define init_10_0_10_7
void FASTCALL call_10_0_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_10_8
#define init_10_0_10_8
void FASTCALL call_10_0_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_10_9
#define init_10_0_10_9
void FASTCALL call_10_0_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_10_10
#define init_10_0_10_10
void FASTCALL call_10_0_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_10_11
#define init_10_0_10_11
void FASTCALL call_10_0_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_10_12
#define init_10_0_10_12
void FASTCALL call_10_0_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_10_13
#define init_10_0_10_13
void FASTCALL call_10_0_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_10_14
#define init_10_0_10_14
void FASTCALL call_10_0_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_10_15
#define init_10_0_10_15
void FASTCALL call_10_0_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_11_0
#define init_10_0_11_0
void FASTCALL call_10_0_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_11_1
#define init_10_0_11_1
void FASTCALL call_10_0_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_11_2
#define init_10_0_11_2
void FASTCALL call_10_0_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_11_3
#define init_10_0_11_3
void FASTCALL call_10_0_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_11_4
#define init_10_0_11_4
void FASTCALL call_10_0_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_11_5
#define init_10_0_11_5
void FASTCALL call_10_0_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_11_6
#define init_10_0_11_6
void FASTCALL call_10_0_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_11_7
#define init_10_0_11_7
void FASTCALL call_10_0_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_11_8
#define init_10_0_11_8
void FASTCALL call_10_0_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_11_9
#define init_10_0_11_9
void FASTCALL call_10_0_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_11_10
#define init_10_0_11_10
void FASTCALL call_10_0_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_11_11
#define init_10_0_11_11
void FASTCALL call_10_0_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_11_12
#define init_10_0_11_12
void FASTCALL call_10_0_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_11_13
#define init_10_0_11_13
void FASTCALL call_10_0_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_11_14
#define init_10_0_11_14
void FASTCALL call_10_0_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_11_15
#define init_10_0_11_15
void FASTCALL call_10_0_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_12_0
#define init_10_0_12_0
void FASTCALL call_10_0_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_12_1
#define init_10_0_12_1
void FASTCALL call_10_0_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_12_2
#define init_10_0_12_2
void FASTCALL call_10_0_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_12_3
#define init_10_0_12_3
void FASTCALL call_10_0_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_12_4
#define init_10_0_12_4
void FASTCALL call_10_0_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_12_5
#define init_10_0_12_5
void FASTCALL call_10_0_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_12_6
#define init_10_0_12_6
void FASTCALL call_10_0_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_12_7
#define init_10_0_12_7
void FASTCALL call_10_0_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_12_8
#define init_10_0_12_8
void FASTCALL call_10_0_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_12_9
#define init_10_0_12_9
void FASTCALL call_10_0_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_12_10
#define init_10_0_12_10
void FASTCALL call_10_0_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_12_11
#define init_10_0_12_11
void FASTCALL call_10_0_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_12_12
#define init_10_0_12_12
void FASTCALL call_10_0_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_12_13
#define init_10_0_12_13
void FASTCALL call_10_0_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_12_14
#define init_10_0_12_14
void FASTCALL call_10_0_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_12_15
#define init_10_0_12_15
void FASTCALL call_10_0_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_13_0
#define init_10_0_13_0
void FASTCALL call_10_0_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_13_1
#define init_10_0_13_1
void FASTCALL call_10_0_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_13_2
#define init_10_0_13_2
void FASTCALL call_10_0_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_13_3
#define init_10_0_13_3
void FASTCALL call_10_0_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_13_4
#define init_10_0_13_4
void FASTCALL call_10_0_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_13_5
#define init_10_0_13_5
void FASTCALL call_10_0_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_13_6
#define init_10_0_13_6
void FASTCALL call_10_0_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_13_7
#define init_10_0_13_7
void FASTCALL call_10_0_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_13_8
#define init_10_0_13_8
void FASTCALL call_10_0_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_13_9
#define init_10_0_13_9
void FASTCALL call_10_0_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_13_10
#define init_10_0_13_10
void FASTCALL call_10_0_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_13_11
#define init_10_0_13_11
void FASTCALL call_10_0_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_13_12
#define init_10_0_13_12
void FASTCALL call_10_0_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_13_13
#define init_10_0_13_13
void FASTCALL call_10_0_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_13_14
#define init_10_0_13_14
void FASTCALL call_10_0_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_13_15
#define init_10_0_13_15
void FASTCALL call_10_0_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_14_0
#define init_10_0_14_0
void FASTCALL call_10_0_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_14_1
#define init_10_0_14_1
void FASTCALL call_10_0_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_14_2
#define init_10_0_14_2
void FASTCALL call_10_0_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_14_3
#define init_10_0_14_3
void FASTCALL call_10_0_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_14_4
#define init_10_0_14_4
void FASTCALL call_10_0_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_14_5
#define init_10_0_14_5
void FASTCALL call_10_0_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_14_6
#define init_10_0_14_6
void FASTCALL call_10_0_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_14_7
#define init_10_0_14_7
void FASTCALL call_10_0_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_14_8
#define init_10_0_14_8
void FASTCALL call_10_0_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_14_9
#define init_10_0_14_9
void FASTCALL call_10_0_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_14_10
#define init_10_0_14_10
void FASTCALL call_10_0_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_14_11
#define init_10_0_14_11
void FASTCALL call_10_0_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_14_12
#define init_10_0_14_12
void FASTCALL call_10_0_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_14_13
#define init_10_0_14_13
void FASTCALL call_10_0_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_14_14
#define init_10_0_14_14
void FASTCALL call_10_0_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_14_15
#define init_10_0_14_15
void FASTCALL call_10_0_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_15_0
#define init_10_0_15_0
void FASTCALL call_10_0_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_15_1
#define init_10_0_15_1
void FASTCALL call_10_0_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_15_2
#define init_10_0_15_2
void FASTCALL call_10_0_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_15_3
#define init_10_0_15_3
void FASTCALL call_10_0_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_15_4
#define init_10_0_15_4
void FASTCALL call_10_0_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_15_5
#define init_10_0_15_5
void FASTCALL call_10_0_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_15_6
#define init_10_0_15_6
void FASTCALL call_10_0_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_15_7
#define init_10_0_15_7
void FASTCALL call_10_0_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_15_8
#define init_10_0_15_8
void FASTCALL call_10_0_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_15_9
#define init_10_0_15_9
void FASTCALL call_10_0_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_15_10
#define init_10_0_15_10
void FASTCALL call_10_0_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_15_11
#define init_10_0_15_11
void FASTCALL call_10_0_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_15_12
#define init_10_0_15_12
void FASTCALL call_10_0_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_15_13
#define init_10_0_15_13
void FASTCALL call_10_0_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_15_14
#define init_10_0_15_14
void FASTCALL call_10_0_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_0_15_15
#define init_10_0_15_15
void FASTCALL call_10_0_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_0_0
#define init_10_1_0_0
void FASTCALL call_10_1_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_0_1
#define init_10_1_0_1
void FASTCALL call_10_1_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_0_2
#define init_10_1_0_2
void FASTCALL call_10_1_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_0_3
#define init_10_1_0_3
void FASTCALL call_10_1_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_0_4
#define init_10_1_0_4
void FASTCALL call_10_1_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_0_5
#define init_10_1_0_5
void FASTCALL call_10_1_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_0_6
#define init_10_1_0_6
void FASTCALL call_10_1_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_0_7
#define init_10_1_0_7
void FASTCALL call_10_1_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_0_8
#define init_10_1_0_8
void FASTCALL call_10_1_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_0_9
#define init_10_1_0_9
void FASTCALL call_10_1_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_0_10
#define init_10_1_0_10
void FASTCALL call_10_1_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_0_11
#define init_10_1_0_11
void FASTCALL call_10_1_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_0_12
#define init_10_1_0_12
void FASTCALL call_10_1_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_0_13
#define init_10_1_0_13
void FASTCALL call_10_1_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_0_14
#define init_10_1_0_14
void FASTCALL call_10_1_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_0_15
#define init_10_1_0_15
void FASTCALL call_10_1_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_1_0
#define init_10_1_1_0
void FASTCALL call_10_1_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_1_1
#define init_10_1_1_1
void FASTCALL call_10_1_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_1_2
#define init_10_1_1_2
void FASTCALL call_10_1_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_1_3
#define init_10_1_1_3
void FASTCALL call_10_1_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_1_4
#define init_10_1_1_4
void FASTCALL call_10_1_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_1_5
#define init_10_1_1_5
void FASTCALL call_10_1_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_1_6
#define init_10_1_1_6
void FASTCALL call_10_1_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_1_7
#define init_10_1_1_7
void FASTCALL call_10_1_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_1_8
#define init_10_1_1_8
void FASTCALL call_10_1_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_1_9
#define init_10_1_1_9
void FASTCALL call_10_1_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_1_10
#define init_10_1_1_10
void FASTCALL call_10_1_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_1_11
#define init_10_1_1_11
void FASTCALL call_10_1_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_1_12
#define init_10_1_1_12
void FASTCALL call_10_1_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_1_13
#define init_10_1_1_13
void FASTCALL call_10_1_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_1_14
#define init_10_1_1_14
void FASTCALL call_10_1_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_1_15
#define init_10_1_1_15
void FASTCALL call_10_1_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_2_0
#define init_10_1_2_0
void FASTCALL call_10_1_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_2_1
#define init_10_1_2_1
void FASTCALL call_10_1_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_2_2
#define init_10_1_2_2
void FASTCALL call_10_1_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_2_3
#define init_10_1_2_3
void FASTCALL call_10_1_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_2_4
#define init_10_1_2_4
void FASTCALL call_10_1_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_2_5
#define init_10_1_2_5
void FASTCALL call_10_1_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_2_6
#define init_10_1_2_6
void FASTCALL call_10_1_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_2_7
#define init_10_1_2_7
void FASTCALL call_10_1_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_2_8
#define init_10_1_2_8
void FASTCALL call_10_1_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_2_9
#define init_10_1_2_9
void FASTCALL call_10_1_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_2_10
#define init_10_1_2_10
void FASTCALL call_10_1_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_2_11
#define init_10_1_2_11
void FASTCALL call_10_1_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_2_12
#define init_10_1_2_12
void FASTCALL call_10_1_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_2_13
#define init_10_1_2_13
void FASTCALL call_10_1_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_2_14
#define init_10_1_2_14
void FASTCALL call_10_1_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_2_15
#define init_10_1_2_15
void FASTCALL call_10_1_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_3_0
#define init_10_1_3_0
void FASTCALL call_10_1_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_3_1
#define init_10_1_3_1
void FASTCALL call_10_1_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_3_2
#define init_10_1_3_2
void FASTCALL call_10_1_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_3_3
#define init_10_1_3_3
void FASTCALL call_10_1_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_3_4
#define init_10_1_3_4
void FASTCALL call_10_1_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_3_5
#define init_10_1_3_5
void FASTCALL call_10_1_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_3_6
#define init_10_1_3_6
void FASTCALL call_10_1_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_3_7
#define init_10_1_3_7
void FASTCALL call_10_1_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_3_8
#define init_10_1_3_8
void FASTCALL call_10_1_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_3_9
#define init_10_1_3_9
void FASTCALL call_10_1_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_3_10
#define init_10_1_3_10
void FASTCALL call_10_1_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_3_11
#define init_10_1_3_11
void FASTCALL call_10_1_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_3_12
#define init_10_1_3_12
void FASTCALL call_10_1_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_3_13
#define init_10_1_3_13
void FASTCALL call_10_1_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_3_14
#define init_10_1_3_14
void FASTCALL call_10_1_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_3_15
#define init_10_1_3_15
void FASTCALL call_10_1_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_4_0
#define init_10_1_4_0
void FASTCALL call_10_1_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_4_1
#define init_10_1_4_1
void FASTCALL call_10_1_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_4_2
#define init_10_1_4_2
void FASTCALL call_10_1_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_4_3
#define init_10_1_4_3
void FASTCALL call_10_1_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_4_4
#define init_10_1_4_4
void FASTCALL call_10_1_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_4_5
#define init_10_1_4_5
void FASTCALL call_10_1_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_4_6
#define init_10_1_4_6
void FASTCALL call_10_1_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_4_7
#define init_10_1_4_7
void FASTCALL call_10_1_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_4_8
#define init_10_1_4_8
void FASTCALL call_10_1_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_4_9
#define init_10_1_4_9
void FASTCALL call_10_1_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_4_10
#define init_10_1_4_10
void FASTCALL call_10_1_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_4_11
#define init_10_1_4_11
void FASTCALL call_10_1_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_4_12
#define init_10_1_4_12
void FASTCALL call_10_1_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_4_13
#define init_10_1_4_13
void FASTCALL call_10_1_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_4_14
#define init_10_1_4_14
void FASTCALL call_10_1_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_4_15
#define init_10_1_4_15
void FASTCALL call_10_1_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_5_0
#define init_10_1_5_0
void FASTCALL call_10_1_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_5_1
#define init_10_1_5_1
void FASTCALL call_10_1_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_5_2
#define init_10_1_5_2
void FASTCALL call_10_1_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_5_3
#define init_10_1_5_3
void FASTCALL call_10_1_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_5_4
#define init_10_1_5_4
void FASTCALL call_10_1_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_5_5
#define init_10_1_5_5
void FASTCALL call_10_1_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_5_6
#define init_10_1_5_6
void FASTCALL call_10_1_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_5_7
#define init_10_1_5_7
void FASTCALL call_10_1_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_5_8
#define init_10_1_5_8
void FASTCALL call_10_1_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_5_9
#define init_10_1_5_9
void FASTCALL call_10_1_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_5_10
#define init_10_1_5_10
void FASTCALL call_10_1_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_5_11
#define init_10_1_5_11
void FASTCALL call_10_1_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_5_12
#define init_10_1_5_12
void FASTCALL call_10_1_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_5_13
#define init_10_1_5_13
void FASTCALL call_10_1_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_5_14
#define init_10_1_5_14
void FASTCALL call_10_1_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_5_15
#define init_10_1_5_15
void FASTCALL call_10_1_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_6_0
#define init_10_1_6_0
void FASTCALL call_10_1_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_6_1
#define init_10_1_6_1
void FASTCALL call_10_1_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_6_2
#define init_10_1_6_2
void FASTCALL call_10_1_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_6_3
#define init_10_1_6_3
void FASTCALL call_10_1_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_6_4
#define init_10_1_6_4
void FASTCALL call_10_1_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_6_5
#define init_10_1_6_5
void FASTCALL call_10_1_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_6_6
#define init_10_1_6_6
void FASTCALL call_10_1_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_6_7
#define init_10_1_6_7
void FASTCALL call_10_1_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_6_8
#define init_10_1_6_8
void FASTCALL call_10_1_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_6_9
#define init_10_1_6_9
void FASTCALL call_10_1_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_6_10
#define init_10_1_6_10
void FASTCALL call_10_1_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_6_11
#define init_10_1_6_11
void FASTCALL call_10_1_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_6_12
#define init_10_1_6_12
void FASTCALL call_10_1_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_6_13
#define init_10_1_6_13
void FASTCALL call_10_1_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_6_14
#define init_10_1_6_14
void FASTCALL call_10_1_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_6_15
#define init_10_1_6_15
void FASTCALL call_10_1_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_7_0
#define init_10_1_7_0
void FASTCALL call_10_1_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_7_1
#define init_10_1_7_1
void FASTCALL call_10_1_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_7_2
#define init_10_1_7_2
void FASTCALL call_10_1_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_7_3
#define init_10_1_7_3
void FASTCALL call_10_1_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_7_4
#define init_10_1_7_4
void FASTCALL call_10_1_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_7_5
#define init_10_1_7_5
void FASTCALL call_10_1_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_7_6
#define init_10_1_7_6
void FASTCALL call_10_1_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_7_7
#define init_10_1_7_7
void FASTCALL call_10_1_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_7_8
#define init_10_1_7_8
void FASTCALL call_10_1_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_7_9
#define init_10_1_7_9
void FASTCALL call_10_1_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_7_10
#define init_10_1_7_10
void FASTCALL call_10_1_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_7_11
#define init_10_1_7_11
void FASTCALL call_10_1_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_7_12
#define init_10_1_7_12
void FASTCALL call_10_1_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_7_13
#define init_10_1_7_13
void FASTCALL call_10_1_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_7_14
#define init_10_1_7_14
void FASTCALL call_10_1_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_7_15
#define init_10_1_7_15
void FASTCALL call_10_1_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_8_0
#define init_10_1_8_0
void FASTCALL call_10_1_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_8_1
#define init_10_1_8_1
void FASTCALL call_10_1_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_8_2
#define init_10_1_8_2
void FASTCALL call_10_1_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_8_3
#define init_10_1_8_3
void FASTCALL call_10_1_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_8_4
#define init_10_1_8_4
void FASTCALL call_10_1_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_8_5
#define init_10_1_8_5
void FASTCALL call_10_1_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_8_6
#define init_10_1_8_6
void FASTCALL call_10_1_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_8_7
#define init_10_1_8_7
void FASTCALL call_10_1_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_8_8
#define init_10_1_8_8
void FASTCALL call_10_1_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_8_9
#define init_10_1_8_9
void FASTCALL call_10_1_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_8_10
#define init_10_1_8_10
void FASTCALL call_10_1_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_8_11
#define init_10_1_8_11
void FASTCALL call_10_1_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_8_12
#define init_10_1_8_12
void FASTCALL call_10_1_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_8_13
#define init_10_1_8_13
void FASTCALL call_10_1_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_8_14
#define init_10_1_8_14
void FASTCALL call_10_1_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_8_15
#define init_10_1_8_15
void FASTCALL call_10_1_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_9_0
#define init_10_1_9_0
void FASTCALL call_10_1_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_9_1
#define init_10_1_9_1
void FASTCALL call_10_1_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_9_2
#define init_10_1_9_2
void FASTCALL call_10_1_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_9_3
#define init_10_1_9_3
void FASTCALL call_10_1_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_9_4
#define init_10_1_9_4
void FASTCALL call_10_1_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_9_5
#define init_10_1_9_5
void FASTCALL call_10_1_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_9_6
#define init_10_1_9_6
void FASTCALL call_10_1_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_9_7
#define init_10_1_9_7
void FASTCALL call_10_1_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_9_8
#define init_10_1_9_8
void FASTCALL call_10_1_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_9_9
#define init_10_1_9_9
void FASTCALL call_10_1_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_9_10
#define init_10_1_9_10
void FASTCALL call_10_1_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_9_11
#define init_10_1_9_11
void FASTCALL call_10_1_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_9_12
#define init_10_1_9_12
void FASTCALL call_10_1_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_9_13
#define init_10_1_9_13
void FASTCALL call_10_1_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_9_14
#define init_10_1_9_14
void FASTCALL call_10_1_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_9_15
#define init_10_1_9_15
void FASTCALL call_10_1_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_10_0
#define init_10_1_10_0
void FASTCALL call_10_1_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_10_1
#define init_10_1_10_1
void FASTCALL call_10_1_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_10_2
#define init_10_1_10_2
void FASTCALL call_10_1_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_10_3
#define init_10_1_10_3
void FASTCALL call_10_1_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_10_4
#define init_10_1_10_4
void FASTCALL call_10_1_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_10_5
#define init_10_1_10_5
void FASTCALL call_10_1_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_10_6
#define init_10_1_10_6
void FASTCALL call_10_1_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_10_7
#define init_10_1_10_7
void FASTCALL call_10_1_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_10_8
#define init_10_1_10_8
void FASTCALL call_10_1_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_10_9
#define init_10_1_10_9
void FASTCALL call_10_1_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_10_10
#define init_10_1_10_10
void FASTCALL call_10_1_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_10_11
#define init_10_1_10_11
void FASTCALL call_10_1_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_10_12
#define init_10_1_10_12
void FASTCALL call_10_1_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_10_13
#define init_10_1_10_13
void FASTCALL call_10_1_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_10_14
#define init_10_1_10_14
void FASTCALL call_10_1_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_10_15
#define init_10_1_10_15
void FASTCALL call_10_1_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_11_0
#define init_10_1_11_0
void FASTCALL call_10_1_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_11_1
#define init_10_1_11_1
void FASTCALL call_10_1_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_11_2
#define init_10_1_11_2
void FASTCALL call_10_1_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_11_3
#define init_10_1_11_3
void FASTCALL call_10_1_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_11_4
#define init_10_1_11_4
void FASTCALL call_10_1_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_11_5
#define init_10_1_11_5
void FASTCALL call_10_1_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_11_6
#define init_10_1_11_6
void FASTCALL call_10_1_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_11_7
#define init_10_1_11_7
void FASTCALL call_10_1_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_11_8
#define init_10_1_11_8
void FASTCALL call_10_1_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_11_9
#define init_10_1_11_9
void FASTCALL call_10_1_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_11_10
#define init_10_1_11_10
void FASTCALL call_10_1_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_11_11
#define init_10_1_11_11
void FASTCALL call_10_1_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_11_12
#define init_10_1_11_12
void FASTCALL call_10_1_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_11_13
#define init_10_1_11_13
void FASTCALL call_10_1_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_11_14
#define init_10_1_11_14
void FASTCALL call_10_1_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_11_15
#define init_10_1_11_15
void FASTCALL call_10_1_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_12_0
#define init_10_1_12_0
void FASTCALL call_10_1_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_12_1
#define init_10_1_12_1
void FASTCALL call_10_1_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_12_2
#define init_10_1_12_2
void FASTCALL call_10_1_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_12_3
#define init_10_1_12_3
void FASTCALL call_10_1_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_12_4
#define init_10_1_12_4
void FASTCALL call_10_1_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_12_5
#define init_10_1_12_5
void FASTCALL call_10_1_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_12_6
#define init_10_1_12_6
void FASTCALL call_10_1_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_12_7
#define init_10_1_12_7
void FASTCALL call_10_1_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_12_8
#define init_10_1_12_8
void FASTCALL call_10_1_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_12_9
#define init_10_1_12_9
void FASTCALL call_10_1_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_12_10
#define init_10_1_12_10
void FASTCALL call_10_1_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_12_11
#define init_10_1_12_11
void FASTCALL call_10_1_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_12_12
#define init_10_1_12_12
void FASTCALL call_10_1_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_12_13
#define init_10_1_12_13
void FASTCALL call_10_1_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_12_14
#define init_10_1_12_14
void FASTCALL call_10_1_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_12_15
#define init_10_1_12_15
void FASTCALL call_10_1_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_13_0
#define init_10_1_13_0
void FASTCALL call_10_1_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_13_1
#define init_10_1_13_1
void FASTCALL call_10_1_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_13_2
#define init_10_1_13_2
void FASTCALL call_10_1_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_13_3
#define init_10_1_13_3
void FASTCALL call_10_1_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_13_4
#define init_10_1_13_4
void FASTCALL call_10_1_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_13_5
#define init_10_1_13_5
void FASTCALL call_10_1_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_13_6
#define init_10_1_13_6
void FASTCALL call_10_1_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_13_7
#define init_10_1_13_7
void FASTCALL call_10_1_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_13_8
#define init_10_1_13_8
void FASTCALL call_10_1_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_13_9
#define init_10_1_13_9
void FASTCALL call_10_1_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_13_10
#define init_10_1_13_10
void FASTCALL call_10_1_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_13_11
#define init_10_1_13_11
void FASTCALL call_10_1_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_13_12
#define init_10_1_13_12
void FASTCALL call_10_1_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_13_13
#define init_10_1_13_13
void FASTCALL call_10_1_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_13_14
#define init_10_1_13_14
void FASTCALL call_10_1_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_13_15
#define init_10_1_13_15
void FASTCALL call_10_1_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_14_0
#define init_10_1_14_0
void FASTCALL call_10_1_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_14_1
#define init_10_1_14_1
void FASTCALL call_10_1_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_14_2
#define init_10_1_14_2
void FASTCALL call_10_1_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_14_3
#define init_10_1_14_3
void FASTCALL call_10_1_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_14_4
#define init_10_1_14_4
void FASTCALL call_10_1_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_14_5
#define init_10_1_14_5
void FASTCALL call_10_1_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_14_6
#define init_10_1_14_6
void FASTCALL call_10_1_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_14_7
#define init_10_1_14_7
void FASTCALL call_10_1_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_14_8
#define init_10_1_14_8
void FASTCALL call_10_1_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_14_9
#define init_10_1_14_9
void FASTCALL call_10_1_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_14_10
#define init_10_1_14_10
void FASTCALL call_10_1_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_14_11
#define init_10_1_14_11
void FASTCALL call_10_1_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_14_12
#define init_10_1_14_12
void FASTCALL call_10_1_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_14_13
#define init_10_1_14_13
void FASTCALL call_10_1_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_14_14
#define init_10_1_14_14
void FASTCALL call_10_1_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_14_15
#define init_10_1_14_15
void FASTCALL call_10_1_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_15_0
#define init_10_1_15_0
void FASTCALL call_10_1_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_15_1
#define init_10_1_15_1
void FASTCALL call_10_1_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_15_2
#define init_10_1_15_2
void FASTCALL call_10_1_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_15_3
#define init_10_1_15_3
void FASTCALL call_10_1_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_15_4
#define init_10_1_15_4
void FASTCALL call_10_1_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_15_5
#define init_10_1_15_5
void FASTCALL call_10_1_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_15_6
#define init_10_1_15_6
void FASTCALL call_10_1_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_15_7
#define init_10_1_15_7
void FASTCALL call_10_1_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_15_8
#define init_10_1_15_8
void FASTCALL call_10_1_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_15_9
#define init_10_1_15_9
void FASTCALL call_10_1_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_15_10
#define init_10_1_15_10
void FASTCALL call_10_1_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_15_11
#define init_10_1_15_11
void FASTCALL call_10_1_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_15_12
#define init_10_1_15_12
void FASTCALL call_10_1_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_15_13
#define init_10_1_15_13
void FASTCALL call_10_1_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_15_14
#define init_10_1_15_14
void FASTCALL call_10_1_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_1_15_15
#define init_10_1_15_15
void FASTCALL call_10_1_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_0_0
#define init_10_2_0_0
void FASTCALL call_10_2_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_0_1
#define init_10_2_0_1
void FASTCALL call_10_2_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_0_2
#define init_10_2_0_2
void FASTCALL call_10_2_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_0_3
#define init_10_2_0_3
void FASTCALL call_10_2_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_0_4
#define init_10_2_0_4
void FASTCALL call_10_2_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_0_5
#define init_10_2_0_5
void FASTCALL call_10_2_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_0_6
#define init_10_2_0_6
void FASTCALL call_10_2_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_0_7
#define init_10_2_0_7
void FASTCALL call_10_2_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_0_8
#define init_10_2_0_8
void FASTCALL call_10_2_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_0_9
#define init_10_2_0_9
void FASTCALL call_10_2_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_0_10
#define init_10_2_0_10
void FASTCALL call_10_2_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_0_11
#define init_10_2_0_11
void FASTCALL call_10_2_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_0_12
#define init_10_2_0_12
void FASTCALL call_10_2_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_0_13
#define init_10_2_0_13
void FASTCALL call_10_2_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_0_14
#define init_10_2_0_14
void FASTCALL call_10_2_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_0_15
#define init_10_2_0_15
void FASTCALL call_10_2_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_1_0
#define init_10_2_1_0
void FASTCALL call_10_2_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_1_1
#define init_10_2_1_1
void FASTCALL call_10_2_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_1_2
#define init_10_2_1_2
void FASTCALL call_10_2_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_1_3
#define init_10_2_1_3
void FASTCALL call_10_2_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_1_4
#define init_10_2_1_4
void FASTCALL call_10_2_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_1_5
#define init_10_2_1_5
void FASTCALL call_10_2_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_1_6
#define init_10_2_1_6
void FASTCALL call_10_2_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_1_7
#define init_10_2_1_7
void FASTCALL call_10_2_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_1_8
#define init_10_2_1_8
void FASTCALL call_10_2_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_1_9
#define init_10_2_1_9
void FASTCALL call_10_2_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_1_10
#define init_10_2_1_10
void FASTCALL call_10_2_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_1_11
#define init_10_2_1_11
void FASTCALL call_10_2_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_1_12
#define init_10_2_1_12
void FASTCALL call_10_2_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_1_13
#define init_10_2_1_13
void FASTCALL call_10_2_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_1_14
#define init_10_2_1_14
void FASTCALL call_10_2_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_1_15
#define init_10_2_1_15
void FASTCALL call_10_2_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_2_0
#define init_10_2_2_0
void FASTCALL call_10_2_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_2_1
#define init_10_2_2_1
void FASTCALL call_10_2_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_2_2
#define init_10_2_2_2
void FASTCALL call_10_2_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_2_3
#define init_10_2_2_3
void FASTCALL call_10_2_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_2_4
#define init_10_2_2_4
void FASTCALL call_10_2_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_2_5
#define init_10_2_2_5
void FASTCALL call_10_2_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_2_6
#define init_10_2_2_6
void FASTCALL call_10_2_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_2_7
#define init_10_2_2_7
void FASTCALL call_10_2_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_2_8
#define init_10_2_2_8
void FASTCALL call_10_2_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_2_9
#define init_10_2_2_9
void FASTCALL call_10_2_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_2_10
#define init_10_2_2_10
void FASTCALL call_10_2_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_2_11
#define init_10_2_2_11
void FASTCALL call_10_2_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_2_12
#define init_10_2_2_12
void FASTCALL call_10_2_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_2_13
#define init_10_2_2_13
void FASTCALL call_10_2_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_2_14
#define init_10_2_2_14
void FASTCALL call_10_2_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_2_15
#define init_10_2_2_15
void FASTCALL call_10_2_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_3_0
#define init_10_2_3_0
void FASTCALL call_10_2_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_3_1
#define init_10_2_3_1
void FASTCALL call_10_2_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_3_2
#define init_10_2_3_2
void FASTCALL call_10_2_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_3_3
#define init_10_2_3_3
void FASTCALL call_10_2_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_3_4
#define init_10_2_3_4
void FASTCALL call_10_2_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_3_5
#define init_10_2_3_5
void FASTCALL call_10_2_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_3_6
#define init_10_2_3_6
void FASTCALL call_10_2_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_3_7
#define init_10_2_3_7
void FASTCALL call_10_2_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_3_8
#define init_10_2_3_8
void FASTCALL call_10_2_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_3_9
#define init_10_2_3_9
void FASTCALL call_10_2_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_3_10
#define init_10_2_3_10
void FASTCALL call_10_2_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_3_11
#define init_10_2_3_11
void FASTCALL call_10_2_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_3_12
#define init_10_2_3_12
void FASTCALL call_10_2_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_3_13
#define init_10_2_3_13
void FASTCALL call_10_2_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_3_14
#define init_10_2_3_14
void FASTCALL call_10_2_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_3_15
#define init_10_2_3_15
void FASTCALL call_10_2_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_4_0
#define init_10_2_4_0
void FASTCALL call_10_2_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_4_1
#define init_10_2_4_1
void FASTCALL call_10_2_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_4_2
#define init_10_2_4_2
void FASTCALL call_10_2_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_4_3
#define init_10_2_4_3
void FASTCALL call_10_2_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_4_4
#define init_10_2_4_4
void FASTCALL call_10_2_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_4_5
#define init_10_2_4_5
void FASTCALL call_10_2_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_4_6
#define init_10_2_4_6
void FASTCALL call_10_2_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_4_7
#define init_10_2_4_7
void FASTCALL call_10_2_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_4_8
#define init_10_2_4_8
void FASTCALL call_10_2_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_4_9
#define init_10_2_4_9
void FASTCALL call_10_2_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_4_10
#define init_10_2_4_10
void FASTCALL call_10_2_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_4_11
#define init_10_2_4_11
void FASTCALL call_10_2_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_4_12
#define init_10_2_4_12
void FASTCALL call_10_2_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_4_13
#define init_10_2_4_13
void FASTCALL call_10_2_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_4_14
#define init_10_2_4_14
void FASTCALL call_10_2_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_4_15
#define init_10_2_4_15
void FASTCALL call_10_2_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_5_0
#define init_10_2_5_0
void FASTCALL call_10_2_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_5_1
#define init_10_2_5_1
void FASTCALL call_10_2_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_5_2
#define init_10_2_5_2
void FASTCALL call_10_2_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_5_3
#define init_10_2_5_3
void FASTCALL call_10_2_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_5_4
#define init_10_2_5_4
void FASTCALL call_10_2_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_5_5
#define init_10_2_5_5
void FASTCALL call_10_2_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_5_6
#define init_10_2_5_6
void FASTCALL call_10_2_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_5_7
#define init_10_2_5_7
void FASTCALL call_10_2_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_5_8
#define init_10_2_5_8
void FASTCALL call_10_2_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_5_9
#define init_10_2_5_9
void FASTCALL call_10_2_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_5_10
#define init_10_2_5_10
void FASTCALL call_10_2_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_5_11
#define init_10_2_5_11
void FASTCALL call_10_2_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_5_12
#define init_10_2_5_12
void FASTCALL call_10_2_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_5_13
#define init_10_2_5_13
void FASTCALL call_10_2_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_5_14
#define init_10_2_5_14
void FASTCALL call_10_2_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_5_15
#define init_10_2_5_15
void FASTCALL call_10_2_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_6_0
#define init_10_2_6_0
void FASTCALL call_10_2_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_6_1
#define init_10_2_6_1
void FASTCALL call_10_2_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_6_2
#define init_10_2_6_2
void FASTCALL call_10_2_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_6_3
#define init_10_2_6_3
void FASTCALL call_10_2_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_6_4
#define init_10_2_6_4
void FASTCALL call_10_2_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_6_5
#define init_10_2_6_5
void FASTCALL call_10_2_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_6_6
#define init_10_2_6_6
void FASTCALL call_10_2_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_6_7
#define init_10_2_6_7
void FASTCALL call_10_2_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_6_8
#define init_10_2_6_8
void FASTCALL call_10_2_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_6_9
#define init_10_2_6_9
void FASTCALL call_10_2_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_6_10
#define init_10_2_6_10
void FASTCALL call_10_2_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_6_11
#define init_10_2_6_11
void FASTCALL call_10_2_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_6_12
#define init_10_2_6_12
void FASTCALL call_10_2_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_6_13
#define init_10_2_6_13
void FASTCALL call_10_2_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_6_14
#define init_10_2_6_14
void FASTCALL call_10_2_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_6_15
#define init_10_2_6_15
void FASTCALL call_10_2_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_7_0
#define init_10_2_7_0
void FASTCALL call_10_2_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_7_1
#define init_10_2_7_1
void FASTCALL call_10_2_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_7_2
#define init_10_2_7_2
void FASTCALL call_10_2_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_7_3
#define init_10_2_7_3
void FASTCALL call_10_2_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_7_4
#define init_10_2_7_4
void FASTCALL call_10_2_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_7_5
#define init_10_2_7_5
void FASTCALL call_10_2_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_7_6
#define init_10_2_7_6
void FASTCALL call_10_2_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_7_7
#define init_10_2_7_7
void FASTCALL call_10_2_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_7_8
#define init_10_2_7_8
void FASTCALL call_10_2_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_7_9
#define init_10_2_7_9
void FASTCALL call_10_2_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_7_10
#define init_10_2_7_10
void FASTCALL call_10_2_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_7_11
#define init_10_2_7_11
void FASTCALL call_10_2_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_7_12
#define init_10_2_7_12
void FASTCALL call_10_2_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_7_13
#define init_10_2_7_13
void FASTCALL call_10_2_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_7_14
#define init_10_2_7_14
void FASTCALL call_10_2_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_7_15
#define init_10_2_7_15
void FASTCALL call_10_2_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_8_0
#define init_10_2_8_0
void FASTCALL call_10_2_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_8_1
#define init_10_2_8_1
void FASTCALL call_10_2_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_8_2
#define init_10_2_8_2
void FASTCALL call_10_2_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_8_3
#define init_10_2_8_3
void FASTCALL call_10_2_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_8_4
#define init_10_2_8_4
void FASTCALL call_10_2_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_8_5
#define init_10_2_8_5
void FASTCALL call_10_2_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_8_6
#define init_10_2_8_6
void FASTCALL call_10_2_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_8_7
#define init_10_2_8_7
void FASTCALL call_10_2_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_8_8
#define init_10_2_8_8
void FASTCALL call_10_2_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_8_9
#define init_10_2_8_9
void FASTCALL call_10_2_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_8_10
#define init_10_2_8_10
void FASTCALL call_10_2_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_8_11
#define init_10_2_8_11
void FASTCALL call_10_2_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_8_12
#define init_10_2_8_12
void FASTCALL call_10_2_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_8_13
#define init_10_2_8_13
void FASTCALL call_10_2_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_8_14
#define init_10_2_8_14
void FASTCALL call_10_2_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_8_15
#define init_10_2_8_15
void FASTCALL call_10_2_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_9_0
#define init_10_2_9_0
void FASTCALL call_10_2_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_9_1
#define init_10_2_9_1
void FASTCALL call_10_2_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_9_2
#define init_10_2_9_2
void FASTCALL call_10_2_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_9_3
#define init_10_2_9_3
void FASTCALL call_10_2_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_9_4
#define init_10_2_9_4
void FASTCALL call_10_2_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_9_5
#define init_10_2_9_5
void FASTCALL call_10_2_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_9_6
#define init_10_2_9_6
void FASTCALL call_10_2_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_9_7
#define init_10_2_9_7
void FASTCALL call_10_2_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_9_8
#define init_10_2_9_8
void FASTCALL call_10_2_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_9_9
#define init_10_2_9_9
void FASTCALL call_10_2_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_9_10
#define init_10_2_9_10
void FASTCALL call_10_2_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_9_11
#define init_10_2_9_11
void FASTCALL call_10_2_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_9_12
#define init_10_2_9_12
void FASTCALL call_10_2_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_9_13
#define init_10_2_9_13
void FASTCALL call_10_2_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_9_14
#define init_10_2_9_14
void FASTCALL call_10_2_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_9_15
#define init_10_2_9_15
void FASTCALL call_10_2_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_10_0
#define init_10_2_10_0
void FASTCALL call_10_2_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_10_1
#define init_10_2_10_1
void FASTCALL call_10_2_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_10_2
#define init_10_2_10_2
void FASTCALL call_10_2_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_10_3
#define init_10_2_10_3
void FASTCALL call_10_2_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_10_4
#define init_10_2_10_4
void FASTCALL call_10_2_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_10_5
#define init_10_2_10_5
void FASTCALL call_10_2_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_10_6
#define init_10_2_10_6
void FASTCALL call_10_2_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_10_7
#define init_10_2_10_7
void FASTCALL call_10_2_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_10_8
#define init_10_2_10_8
void FASTCALL call_10_2_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_10_9
#define init_10_2_10_9
void FASTCALL call_10_2_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_10_10
#define init_10_2_10_10
void FASTCALL call_10_2_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_10_11
#define init_10_2_10_11
void FASTCALL call_10_2_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_10_12
#define init_10_2_10_12
void FASTCALL call_10_2_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_10_13
#define init_10_2_10_13
void FASTCALL call_10_2_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_10_14
#define init_10_2_10_14
void FASTCALL call_10_2_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_10_15
#define init_10_2_10_15
void FASTCALL call_10_2_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_11_0
#define init_10_2_11_0
void FASTCALL call_10_2_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_11_1
#define init_10_2_11_1
void FASTCALL call_10_2_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_11_2
#define init_10_2_11_2
void FASTCALL call_10_2_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_11_3
#define init_10_2_11_3
void FASTCALL call_10_2_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_11_4
#define init_10_2_11_4
void FASTCALL call_10_2_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_11_5
#define init_10_2_11_5
void FASTCALL call_10_2_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_11_6
#define init_10_2_11_6
void FASTCALL call_10_2_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_11_7
#define init_10_2_11_7
void FASTCALL call_10_2_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_11_8
#define init_10_2_11_8
void FASTCALL call_10_2_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_11_9
#define init_10_2_11_9
void FASTCALL call_10_2_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_11_10
#define init_10_2_11_10
void FASTCALL call_10_2_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_11_11
#define init_10_2_11_11
void FASTCALL call_10_2_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_11_12
#define init_10_2_11_12
void FASTCALL call_10_2_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_11_13
#define init_10_2_11_13
void FASTCALL call_10_2_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_11_14
#define init_10_2_11_14
void FASTCALL call_10_2_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_11_15
#define init_10_2_11_15
void FASTCALL call_10_2_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_12_0
#define init_10_2_12_0
void FASTCALL call_10_2_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_12_1
#define init_10_2_12_1
void FASTCALL call_10_2_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_12_2
#define init_10_2_12_2
void FASTCALL call_10_2_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_12_3
#define init_10_2_12_3
void FASTCALL call_10_2_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_12_4
#define init_10_2_12_4
void FASTCALL call_10_2_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_12_5
#define init_10_2_12_5
void FASTCALL call_10_2_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_12_6
#define init_10_2_12_6
void FASTCALL call_10_2_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_12_7
#define init_10_2_12_7
void FASTCALL call_10_2_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_12_8
#define init_10_2_12_8
void FASTCALL call_10_2_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_12_9
#define init_10_2_12_9
void FASTCALL call_10_2_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_12_10
#define init_10_2_12_10
void FASTCALL call_10_2_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_12_11
#define init_10_2_12_11
void FASTCALL call_10_2_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_12_12
#define init_10_2_12_12
void FASTCALL call_10_2_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_12_13
#define init_10_2_12_13
void FASTCALL call_10_2_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_12_14
#define init_10_2_12_14
void FASTCALL call_10_2_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_12_15
#define init_10_2_12_15
void FASTCALL call_10_2_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_13_0
#define init_10_2_13_0
void FASTCALL call_10_2_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_13_1
#define init_10_2_13_1
void FASTCALL call_10_2_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_13_2
#define init_10_2_13_2
void FASTCALL call_10_2_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_13_3
#define init_10_2_13_3
void FASTCALL call_10_2_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_13_4
#define init_10_2_13_4
void FASTCALL call_10_2_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_13_5
#define init_10_2_13_5
void FASTCALL call_10_2_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_13_6
#define init_10_2_13_6
void FASTCALL call_10_2_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_13_7
#define init_10_2_13_7
void FASTCALL call_10_2_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_13_8
#define init_10_2_13_8
void FASTCALL call_10_2_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_13_9
#define init_10_2_13_9
void FASTCALL call_10_2_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_13_10
#define init_10_2_13_10
void FASTCALL call_10_2_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_13_11
#define init_10_2_13_11
void FASTCALL call_10_2_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_13_12
#define init_10_2_13_12
void FASTCALL call_10_2_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_13_13
#define init_10_2_13_13
void FASTCALL call_10_2_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_13_14
#define init_10_2_13_14
void FASTCALL call_10_2_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_13_15
#define init_10_2_13_15
void FASTCALL call_10_2_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_14_0
#define init_10_2_14_0
void FASTCALL call_10_2_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_14_1
#define init_10_2_14_1
void FASTCALL call_10_2_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_14_2
#define init_10_2_14_2
void FASTCALL call_10_2_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_14_3
#define init_10_2_14_3
void FASTCALL call_10_2_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_14_4
#define init_10_2_14_4
void FASTCALL call_10_2_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_14_5
#define init_10_2_14_5
void FASTCALL call_10_2_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_14_6
#define init_10_2_14_6
void FASTCALL call_10_2_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_14_7
#define init_10_2_14_7
void FASTCALL call_10_2_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_14_8
#define init_10_2_14_8
void FASTCALL call_10_2_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_14_9
#define init_10_2_14_9
void FASTCALL call_10_2_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_14_10
#define init_10_2_14_10
void FASTCALL call_10_2_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_14_11
#define init_10_2_14_11
void FASTCALL call_10_2_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_14_12
#define init_10_2_14_12
void FASTCALL call_10_2_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_14_13
#define init_10_2_14_13
void FASTCALL call_10_2_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_14_14
#define init_10_2_14_14
void FASTCALL call_10_2_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_14_15
#define init_10_2_14_15
void FASTCALL call_10_2_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_15_0
#define init_10_2_15_0
void FASTCALL call_10_2_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_15_1
#define init_10_2_15_1
void FASTCALL call_10_2_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_15_2
#define init_10_2_15_2
void FASTCALL call_10_2_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_15_3
#define init_10_2_15_3
void FASTCALL call_10_2_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_15_4
#define init_10_2_15_4
void FASTCALL call_10_2_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_15_5
#define init_10_2_15_5
void FASTCALL call_10_2_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_15_6
#define init_10_2_15_6
void FASTCALL call_10_2_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_15_7
#define init_10_2_15_7
void FASTCALL call_10_2_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_15_8
#define init_10_2_15_8
void FASTCALL call_10_2_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_15_9
#define init_10_2_15_9
void FASTCALL call_10_2_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_15_10
#define init_10_2_15_10
void FASTCALL call_10_2_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_15_11
#define init_10_2_15_11
void FASTCALL call_10_2_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_15_12
#define init_10_2_15_12
void FASTCALL call_10_2_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_15_13
#define init_10_2_15_13
void FASTCALL call_10_2_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_15_14
#define init_10_2_15_14
void FASTCALL call_10_2_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_2_15_15
#define init_10_2_15_15
void FASTCALL call_10_2_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_0_0
#define init_10_3_0_0
void FASTCALL call_10_3_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_0_1
#define init_10_3_0_1
void FASTCALL call_10_3_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_0_2
#define init_10_3_0_2
void FASTCALL call_10_3_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_0_3
#define init_10_3_0_3
void FASTCALL call_10_3_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_0_4
#define init_10_3_0_4
void FASTCALL call_10_3_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_0_5
#define init_10_3_0_5
void FASTCALL call_10_3_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_0_6
#define init_10_3_0_6
void FASTCALL call_10_3_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_0_7
#define init_10_3_0_7
void FASTCALL call_10_3_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_0_8
#define init_10_3_0_8
void FASTCALL call_10_3_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_0_9
#define init_10_3_0_9
void FASTCALL call_10_3_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_0_10
#define init_10_3_0_10
void FASTCALL call_10_3_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_0_11
#define init_10_3_0_11
void FASTCALL call_10_3_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_0_12
#define init_10_3_0_12
void FASTCALL call_10_3_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_0_13
#define init_10_3_0_13
void FASTCALL call_10_3_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_0_14
#define init_10_3_0_14
void FASTCALL call_10_3_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_0_15
#define init_10_3_0_15
void FASTCALL call_10_3_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_1_0
#define init_10_3_1_0
void FASTCALL call_10_3_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_1_1
#define init_10_3_1_1
void FASTCALL call_10_3_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_1_2
#define init_10_3_1_2
void FASTCALL call_10_3_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_1_3
#define init_10_3_1_3
void FASTCALL call_10_3_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_1_4
#define init_10_3_1_4
void FASTCALL call_10_3_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_1_5
#define init_10_3_1_5
void FASTCALL call_10_3_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_1_6
#define init_10_3_1_6
void FASTCALL call_10_3_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_1_7
#define init_10_3_1_7
void FASTCALL call_10_3_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_1_8
#define init_10_3_1_8
void FASTCALL call_10_3_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_1_9
#define init_10_3_1_9
void FASTCALL call_10_3_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_1_10
#define init_10_3_1_10
void FASTCALL call_10_3_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_1_11
#define init_10_3_1_11
void FASTCALL call_10_3_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_1_12
#define init_10_3_1_12
void FASTCALL call_10_3_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_1_13
#define init_10_3_1_13
void FASTCALL call_10_3_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_1_14
#define init_10_3_1_14
void FASTCALL call_10_3_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_1_15
#define init_10_3_1_15
void FASTCALL call_10_3_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_2_0
#define init_10_3_2_0
void FASTCALL call_10_3_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_2_1
#define init_10_3_2_1
void FASTCALL call_10_3_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_2_2
#define init_10_3_2_2
void FASTCALL call_10_3_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_2_3
#define init_10_3_2_3
void FASTCALL call_10_3_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_2_4
#define init_10_3_2_4
void FASTCALL call_10_3_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_2_5
#define init_10_3_2_5
void FASTCALL call_10_3_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_2_6
#define init_10_3_2_6
void FASTCALL call_10_3_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_2_7
#define init_10_3_2_7
void FASTCALL call_10_3_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_2_8
#define init_10_3_2_8
void FASTCALL call_10_3_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_2_9
#define init_10_3_2_9
void FASTCALL call_10_3_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_2_10
#define init_10_3_2_10
void FASTCALL call_10_3_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_2_11
#define init_10_3_2_11
void FASTCALL call_10_3_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_2_12
#define init_10_3_2_12
void FASTCALL call_10_3_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_2_13
#define init_10_3_2_13
void FASTCALL call_10_3_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_2_14
#define init_10_3_2_14
void FASTCALL call_10_3_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_2_15
#define init_10_3_2_15
void FASTCALL call_10_3_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_3_0
#define init_10_3_3_0
void FASTCALL call_10_3_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_3_1
#define init_10_3_3_1
void FASTCALL call_10_3_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_3_2
#define init_10_3_3_2
void FASTCALL call_10_3_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_3_3
#define init_10_3_3_3
void FASTCALL call_10_3_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_3_4
#define init_10_3_3_4
void FASTCALL call_10_3_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_3_5
#define init_10_3_3_5
void FASTCALL call_10_3_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_3_6
#define init_10_3_3_6
void FASTCALL call_10_3_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_3_7
#define init_10_3_3_7
void FASTCALL call_10_3_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_3_8
#define init_10_3_3_8
void FASTCALL call_10_3_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_3_9
#define init_10_3_3_9
void FASTCALL call_10_3_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_3_10
#define init_10_3_3_10
void FASTCALL call_10_3_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_3_11
#define init_10_3_3_11
void FASTCALL call_10_3_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_3_12
#define init_10_3_3_12
void FASTCALL call_10_3_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_3_13
#define init_10_3_3_13
void FASTCALL call_10_3_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_3_14
#define init_10_3_3_14
void FASTCALL call_10_3_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_3_15
#define init_10_3_3_15
void FASTCALL call_10_3_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_4_0
#define init_10_3_4_0
void FASTCALL call_10_3_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_4_1
#define init_10_3_4_1
void FASTCALL call_10_3_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_4_2
#define init_10_3_4_2
void FASTCALL call_10_3_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_4_3
#define init_10_3_4_3
void FASTCALL call_10_3_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_4_4
#define init_10_3_4_4
void FASTCALL call_10_3_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_4_5
#define init_10_3_4_5
void FASTCALL call_10_3_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_4_6
#define init_10_3_4_6
void FASTCALL call_10_3_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_4_7
#define init_10_3_4_7
void FASTCALL call_10_3_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_4_8
#define init_10_3_4_8
void FASTCALL call_10_3_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_4_9
#define init_10_3_4_9
void FASTCALL call_10_3_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_4_10
#define init_10_3_4_10
void FASTCALL call_10_3_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_4_11
#define init_10_3_4_11
void FASTCALL call_10_3_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_4_12
#define init_10_3_4_12
void FASTCALL call_10_3_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_4_13
#define init_10_3_4_13
void FASTCALL call_10_3_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_4_14
#define init_10_3_4_14
void FASTCALL call_10_3_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_4_15
#define init_10_3_4_15
void FASTCALL call_10_3_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_5_0
#define init_10_3_5_0
void FASTCALL call_10_3_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_5_1
#define init_10_3_5_1
void FASTCALL call_10_3_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_5_2
#define init_10_3_5_2
void FASTCALL call_10_3_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_5_3
#define init_10_3_5_3
void FASTCALL call_10_3_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_5_4
#define init_10_3_5_4
void FASTCALL call_10_3_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_5_5
#define init_10_3_5_5
void FASTCALL call_10_3_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_5_6
#define init_10_3_5_6
void FASTCALL call_10_3_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_5_7
#define init_10_3_5_7
void FASTCALL call_10_3_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_5_8
#define init_10_3_5_8
void FASTCALL call_10_3_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_5_9
#define init_10_3_5_9
void FASTCALL call_10_3_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_5_10
#define init_10_3_5_10
void FASTCALL call_10_3_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_5_11
#define init_10_3_5_11
void FASTCALL call_10_3_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_5_12
#define init_10_3_5_12
void FASTCALL call_10_3_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_5_13
#define init_10_3_5_13
void FASTCALL call_10_3_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_5_14
#define init_10_3_5_14
void FASTCALL call_10_3_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_5_15
#define init_10_3_5_15
void FASTCALL call_10_3_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_6_0
#define init_10_3_6_0
void FASTCALL call_10_3_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_6_1
#define init_10_3_6_1
void FASTCALL call_10_3_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_6_2
#define init_10_3_6_2
void FASTCALL call_10_3_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_6_3
#define init_10_3_6_3
void FASTCALL call_10_3_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_6_4
#define init_10_3_6_4
void FASTCALL call_10_3_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_6_5
#define init_10_3_6_5
void FASTCALL call_10_3_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_6_6
#define init_10_3_6_6
void FASTCALL call_10_3_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_6_7
#define init_10_3_6_7
void FASTCALL call_10_3_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_6_8
#define init_10_3_6_8
void FASTCALL call_10_3_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_6_9
#define init_10_3_6_9
void FASTCALL call_10_3_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_6_10
#define init_10_3_6_10
void FASTCALL call_10_3_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_6_11
#define init_10_3_6_11
void FASTCALL call_10_3_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_6_12
#define init_10_3_6_12
void FASTCALL call_10_3_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_6_13
#define init_10_3_6_13
void FASTCALL call_10_3_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_6_14
#define init_10_3_6_14
void FASTCALL call_10_3_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_6_15
#define init_10_3_6_15
void FASTCALL call_10_3_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_7_0
#define init_10_3_7_0
void FASTCALL call_10_3_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_7_1
#define init_10_3_7_1
void FASTCALL call_10_3_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_7_2
#define init_10_3_7_2
void FASTCALL call_10_3_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_7_3
#define init_10_3_7_3
void FASTCALL call_10_3_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_7_4
#define init_10_3_7_4
void FASTCALL call_10_3_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_7_5
#define init_10_3_7_5
void FASTCALL call_10_3_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_7_6
#define init_10_3_7_6
void FASTCALL call_10_3_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_7_7
#define init_10_3_7_7
void FASTCALL call_10_3_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_7_8
#define init_10_3_7_8
void FASTCALL call_10_3_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_7_9
#define init_10_3_7_9
void FASTCALL call_10_3_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_7_10
#define init_10_3_7_10
void FASTCALL call_10_3_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_7_11
#define init_10_3_7_11
void FASTCALL call_10_3_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_7_12
#define init_10_3_7_12
void FASTCALL call_10_3_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_7_13
#define init_10_3_7_13
void FASTCALL call_10_3_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_7_14
#define init_10_3_7_14
void FASTCALL call_10_3_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_7_15
#define init_10_3_7_15
void FASTCALL call_10_3_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_8_0
#define init_10_3_8_0
void FASTCALL call_10_3_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_8_1
#define init_10_3_8_1
void FASTCALL call_10_3_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_8_2
#define init_10_3_8_2
void FASTCALL call_10_3_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_8_3
#define init_10_3_8_3
void FASTCALL call_10_3_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_8_4
#define init_10_3_8_4
void FASTCALL call_10_3_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_8_5
#define init_10_3_8_5
void FASTCALL call_10_3_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_8_6
#define init_10_3_8_6
void FASTCALL call_10_3_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_8_7
#define init_10_3_8_7
void FASTCALL call_10_3_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_8_8
#define init_10_3_8_8
void FASTCALL call_10_3_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_8_9
#define init_10_3_8_9
void FASTCALL call_10_3_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_8_10
#define init_10_3_8_10
void FASTCALL call_10_3_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_8_11
#define init_10_3_8_11
void FASTCALL call_10_3_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_8_12
#define init_10_3_8_12
void FASTCALL call_10_3_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_8_13
#define init_10_3_8_13
void FASTCALL call_10_3_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_8_14
#define init_10_3_8_14
void FASTCALL call_10_3_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_8_15
#define init_10_3_8_15
void FASTCALL call_10_3_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_9_0
#define init_10_3_9_0
void FASTCALL call_10_3_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_9_1
#define init_10_3_9_1
void FASTCALL call_10_3_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_9_2
#define init_10_3_9_2
void FASTCALL call_10_3_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_9_3
#define init_10_3_9_3
void FASTCALL call_10_3_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_9_4
#define init_10_3_9_4
void FASTCALL call_10_3_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_9_5
#define init_10_3_9_5
void FASTCALL call_10_3_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_9_6
#define init_10_3_9_6
void FASTCALL call_10_3_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_9_7
#define init_10_3_9_7
void FASTCALL call_10_3_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_9_8
#define init_10_3_9_8
void FASTCALL call_10_3_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_9_9
#define init_10_3_9_9
void FASTCALL call_10_3_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_9_10
#define init_10_3_9_10
void FASTCALL call_10_3_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_9_11
#define init_10_3_9_11
void FASTCALL call_10_3_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_9_12
#define init_10_3_9_12
void FASTCALL call_10_3_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_9_13
#define init_10_3_9_13
void FASTCALL call_10_3_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_9_14
#define init_10_3_9_14
void FASTCALL call_10_3_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_9_15
#define init_10_3_9_15
void FASTCALL call_10_3_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_10_0
#define init_10_3_10_0
void FASTCALL call_10_3_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_10_1
#define init_10_3_10_1
void FASTCALL call_10_3_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_10_2
#define init_10_3_10_2
void FASTCALL call_10_3_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_10_3
#define init_10_3_10_3
void FASTCALL call_10_3_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_10_4
#define init_10_3_10_4
void FASTCALL call_10_3_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_10_5
#define init_10_3_10_5
void FASTCALL call_10_3_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_10_6
#define init_10_3_10_6
void FASTCALL call_10_3_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_10_7
#define init_10_3_10_7
void FASTCALL call_10_3_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_10_8
#define init_10_3_10_8
void FASTCALL call_10_3_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_10_9
#define init_10_3_10_9
void FASTCALL call_10_3_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_10_10
#define init_10_3_10_10
void FASTCALL call_10_3_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_10_11
#define init_10_3_10_11
void FASTCALL call_10_3_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_10_12
#define init_10_3_10_12
void FASTCALL call_10_3_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_10_13
#define init_10_3_10_13
void FASTCALL call_10_3_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_10_14
#define init_10_3_10_14
void FASTCALL call_10_3_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_10_15
#define init_10_3_10_15
void FASTCALL call_10_3_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_11_0
#define init_10_3_11_0
void FASTCALL call_10_3_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_11_1
#define init_10_3_11_1
void FASTCALL call_10_3_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_11_2
#define init_10_3_11_2
void FASTCALL call_10_3_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_11_3
#define init_10_3_11_3
void FASTCALL call_10_3_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_11_4
#define init_10_3_11_4
void FASTCALL call_10_3_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_11_5
#define init_10_3_11_5
void FASTCALL call_10_3_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_11_6
#define init_10_3_11_6
void FASTCALL call_10_3_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_11_7
#define init_10_3_11_7
void FASTCALL call_10_3_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_11_8
#define init_10_3_11_8
void FASTCALL call_10_3_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_11_9
#define init_10_3_11_9
void FASTCALL call_10_3_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_11_10
#define init_10_3_11_10
void FASTCALL call_10_3_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_11_11
#define init_10_3_11_11
void FASTCALL call_10_3_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_11_12
#define init_10_3_11_12
void FASTCALL call_10_3_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_11_13
#define init_10_3_11_13
void FASTCALL call_10_3_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_11_14
#define init_10_3_11_14
void FASTCALL call_10_3_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_11_15
#define init_10_3_11_15
void FASTCALL call_10_3_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_12_0
#define init_10_3_12_0
void FASTCALL call_10_3_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_12_1
#define init_10_3_12_1
void FASTCALL call_10_3_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_12_2
#define init_10_3_12_2
void FASTCALL call_10_3_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_12_3
#define init_10_3_12_3
void FASTCALL call_10_3_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_12_4
#define init_10_3_12_4
void FASTCALL call_10_3_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_12_5
#define init_10_3_12_5
void FASTCALL call_10_3_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_12_6
#define init_10_3_12_6
void FASTCALL call_10_3_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_12_7
#define init_10_3_12_7
void FASTCALL call_10_3_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_12_8
#define init_10_3_12_8
void FASTCALL call_10_3_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_12_9
#define init_10_3_12_9
void FASTCALL call_10_3_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_12_10
#define init_10_3_12_10
void FASTCALL call_10_3_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_12_11
#define init_10_3_12_11
void FASTCALL call_10_3_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_12_12
#define init_10_3_12_12
void FASTCALL call_10_3_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_12_13
#define init_10_3_12_13
void FASTCALL call_10_3_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_12_14
#define init_10_3_12_14
void FASTCALL call_10_3_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_12_15
#define init_10_3_12_15
void FASTCALL call_10_3_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_13_0
#define init_10_3_13_0
void FASTCALL call_10_3_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_13_1
#define init_10_3_13_1
void FASTCALL call_10_3_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_13_2
#define init_10_3_13_2
void FASTCALL call_10_3_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_13_3
#define init_10_3_13_3
void FASTCALL call_10_3_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_13_4
#define init_10_3_13_4
void FASTCALL call_10_3_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_13_5
#define init_10_3_13_5
void FASTCALL call_10_3_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_13_6
#define init_10_3_13_6
void FASTCALL call_10_3_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_13_7
#define init_10_3_13_7
void FASTCALL call_10_3_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_13_8
#define init_10_3_13_8
void FASTCALL call_10_3_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_13_9
#define init_10_3_13_9
void FASTCALL call_10_3_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_13_10
#define init_10_3_13_10
void FASTCALL call_10_3_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_13_11
#define init_10_3_13_11
void FASTCALL call_10_3_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_13_12
#define init_10_3_13_12
void FASTCALL call_10_3_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_13_13
#define init_10_3_13_13
void FASTCALL call_10_3_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_13_14
#define init_10_3_13_14
void FASTCALL call_10_3_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_13_15
#define init_10_3_13_15
void FASTCALL call_10_3_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_14_0
#define init_10_3_14_0
void FASTCALL call_10_3_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_14_1
#define init_10_3_14_1
void FASTCALL call_10_3_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_14_2
#define init_10_3_14_2
void FASTCALL call_10_3_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_14_3
#define init_10_3_14_3
void FASTCALL call_10_3_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_14_4
#define init_10_3_14_4
void FASTCALL call_10_3_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_14_5
#define init_10_3_14_5
void FASTCALL call_10_3_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_14_6
#define init_10_3_14_6
void FASTCALL call_10_3_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_14_7
#define init_10_3_14_7
void FASTCALL call_10_3_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_14_8
#define init_10_3_14_8
void FASTCALL call_10_3_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_14_9
#define init_10_3_14_9
void FASTCALL call_10_3_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_14_10
#define init_10_3_14_10
void FASTCALL call_10_3_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_14_11
#define init_10_3_14_11
void FASTCALL call_10_3_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_14_12
#define init_10_3_14_12
void FASTCALL call_10_3_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_14_13
#define init_10_3_14_13
void FASTCALL call_10_3_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_14_14
#define init_10_3_14_14
void FASTCALL call_10_3_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_14_15
#define init_10_3_14_15
void FASTCALL call_10_3_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_15_0
#define init_10_3_15_0
void FASTCALL call_10_3_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_15_1
#define init_10_3_15_1
void FASTCALL call_10_3_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_15_2
#define init_10_3_15_2
void FASTCALL call_10_3_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_15_3
#define init_10_3_15_3
void FASTCALL call_10_3_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_15_4
#define init_10_3_15_4
void FASTCALL call_10_3_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_15_5
#define init_10_3_15_5
void FASTCALL call_10_3_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_15_6
#define init_10_3_15_6
void FASTCALL call_10_3_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_15_7
#define init_10_3_15_7
void FASTCALL call_10_3_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_15_8
#define init_10_3_15_8
void FASTCALL call_10_3_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_15_9
#define init_10_3_15_9
void FASTCALL call_10_3_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_15_10
#define init_10_3_15_10
void FASTCALL call_10_3_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_15_11
#define init_10_3_15_11
void FASTCALL call_10_3_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_15_12
#define init_10_3_15_12
void FASTCALL call_10_3_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_15_13
#define init_10_3_15_13
void FASTCALL call_10_3_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_15_14
#define init_10_3_15_14
void FASTCALL call_10_3_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_3_15_15
#define init_10_3_15_15
void FASTCALL call_10_3_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_0_0
#define init_10_4_0_0
void FASTCALL call_10_4_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_0_1
#define init_10_4_0_1
void FASTCALL call_10_4_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_0_2
#define init_10_4_0_2
void FASTCALL call_10_4_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_0_3
#define init_10_4_0_3
void FASTCALL call_10_4_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_0_4
#define init_10_4_0_4
void FASTCALL call_10_4_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_0_5
#define init_10_4_0_5
void FASTCALL call_10_4_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_0_6
#define init_10_4_0_6
void FASTCALL call_10_4_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_0_7
#define init_10_4_0_7
void FASTCALL call_10_4_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_0_8
#define init_10_4_0_8
void FASTCALL call_10_4_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_0_9
#define init_10_4_0_9
void FASTCALL call_10_4_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_0_10
#define init_10_4_0_10
void FASTCALL call_10_4_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_0_11
#define init_10_4_0_11
void FASTCALL call_10_4_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_0_12
#define init_10_4_0_12
void FASTCALL call_10_4_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_0_13
#define init_10_4_0_13
void FASTCALL call_10_4_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_0_14
#define init_10_4_0_14
void FASTCALL call_10_4_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_0_15
#define init_10_4_0_15
void FASTCALL call_10_4_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_1_0
#define init_10_4_1_0
void FASTCALL call_10_4_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_1_1
#define init_10_4_1_1
void FASTCALL call_10_4_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_1_2
#define init_10_4_1_2
void FASTCALL call_10_4_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_1_3
#define init_10_4_1_3
void FASTCALL call_10_4_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_1_4
#define init_10_4_1_4
void FASTCALL call_10_4_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_1_5
#define init_10_4_1_5
void FASTCALL call_10_4_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_1_6
#define init_10_4_1_6
void FASTCALL call_10_4_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_1_7
#define init_10_4_1_7
void FASTCALL call_10_4_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_1_8
#define init_10_4_1_8
void FASTCALL call_10_4_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_1_9
#define init_10_4_1_9
void FASTCALL call_10_4_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_1_10
#define init_10_4_1_10
void FASTCALL call_10_4_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_1_11
#define init_10_4_1_11
void FASTCALL call_10_4_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_1_12
#define init_10_4_1_12
void FASTCALL call_10_4_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_1_13
#define init_10_4_1_13
void FASTCALL call_10_4_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_1_14
#define init_10_4_1_14
void FASTCALL call_10_4_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_1_15
#define init_10_4_1_15
void FASTCALL call_10_4_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_2_0
#define init_10_4_2_0
void FASTCALL call_10_4_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_2_1
#define init_10_4_2_1
void FASTCALL call_10_4_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_2_2
#define init_10_4_2_2
void FASTCALL call_10_4_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_2_3
#define init_10_4_2_3
void FASTCALL call_10_4_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_2_4
#define init_10_4_2_4
void FASTCALL call_10_4_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_2_5
#define init_10_4_2_5
void FASTCALL call_10_4_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_2_6
#define init_10_4_2_6
void FASTCALL call_10_4_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_2_7
#define init_10_4_2_7
void FASTCALL call_10_4_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_2_8
#define init_10_4_2_8
void FASTCALL call_10_4_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_2_9
#define init_10_4_2_9
void FASTCALL call_10_4_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_2_10
#define init_10_4_2_10
void FASTCALL call_10_4_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_2_11
#define init_10_4_2_11
void FASTCALL call_10_4_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_2_12
#define init_10_4_2_12
void FASTCALL call_10_4_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_2_13
#define init_10_4_2_13
void FASTCALL call_10_4_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_2_14
#define init_10_4_2_14
void FASTCALL call_10_4_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_2_15
#define init_10_4_2_15
void FASTCALL call_10_4_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_3_0
#define init_10_4_3_0
void FASTCALL call_10_4_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_3_1
#define init_10_4_3_1
void FASTCALL call_10_4_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_3_2
#define init_10_4_3_2
void FASTCALL call_10_4_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_3_3
#define init_10_4_3_3
void FASTCALL call_10_4_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_3_4
#define init_10_4_3_4
void FASTCALL call_10_4_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_3_5
#define init_10_4_3_5
void FASTCALL call_10_4_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_3_6
#define init_10_4_3_6
void FASTCALL call_10_4_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_3_7
#define init_10_4_3_7
void FASTCALL call_10_4_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_3_8
#define init_10_4_3_8
void FASTCALL call_10_4_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_3_9
#define init_10_4_3_9
void FASTCALL call_10_4_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_3_10
#define init_10_4_3_10
void FASTCALL call_10_4_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_3_11
#define init_10_4_3_11
void FASTCALL call_10_4_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_3_12
#define init_10_4_3_12
void FASTCALL call_10_4_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_3_13
#define init_10_4_3_13
void FASTCALL call_10_4_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_3_14
#define init_10_4_3_14
void FASTCALL call_10_4_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_3_15
#define init_10_4_3_15
void FASTCALL call_10_4_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_4_0
#define init_10_4_4_0
void FASTCALL call_10_4_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_4_1
#define init_10_4_4_1
void FASTCALL call_10_4_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_4_2
#define init_10_4_4_2
void FASTCALL call_10_4_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_4_3
#define init_10_4_4_3
void FASTCALL call_10_4_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_4_4
#define init_10_4_4_4
void FASTCALL call_10_4_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_4_5
#define init_10_4_4_5
void FASTCALL call_10_4_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_4_6
#define init_10_4_4_6
void FASTCALL call_10_4_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_4_7
#define init_10_4_4_7
void FASTCALL call_10_4_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_4_8
#define init_10_4_4_8
void FASTCALL call_10_4_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_4_9
#define init_10_4_4_9
void FASTCALL call_10_4_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_4_10
#define init_10_4_4_10
void FASTCALL call_10_4_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_4_11
#define init_10_4_4_11
void FASTCALL call_10_4_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_4_12
#define init_10_4_4_12
void FASTCALL call_10_4_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_4_13
#define init_10_4_4_13
void FASTCALL call_10_4_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_4_14
#define init_10_4_4_14
void FASTCALL call_10_4_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_4_15
#define init_10_4_4_15
void FASTCALL call_10_4_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_5_0
#define init_10_4_5_0
void FASTCALL call_10_4_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_5_1
#define init_10_4_5_1
void FASTCALL call_10_4_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_5_2
#define init_10_4_5_2
void FASTCALL call_10_4_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_5_3
#define init_10_4_5_3
void FASTCALL call_10_4_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_5_4
#define init_10_4_5_4
void FASTCALL call_10_4_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_5_5
#define init_10_4_5_5
void FASTCALL call_10_4_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_5_6
#define init_10_4_5_6
void FASTCALL call_10_4_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_5_7
#define init_10_4_5_7
void FASTCALL call_10_4_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_5_8
#define init_10_4_5_8
void FASTCALL call_10_4_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_5_9
#define init_10_4_5_9
void FASTCALL call_10_4_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_5_10
#define init_10_4_5_10
void FASTCALL call_10_4_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_5_11
#define init_10_4_5_11
void FASTCALL call_10_4_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_5_12
#define init_10_4_5_12
void FASTCALL call_10_4_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_5_13
#define init_10_4_5_13
void FASTCALL call_10_4_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_5_14
#define init_10_4_5_14
void FASTCALL call_10_4_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_5_15
#define init_10_4_5_15
void FASTCALL call_10_4_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_6_0
#define init_10_4_6_0
void FASTCALL call_10_4_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_6_1
#define init_10_4_6_1
void FASTCALL call_10_4_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_6_2
#define init_10_4_6_2
void FASTCALL call_10_4_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_6_3
#define init_10_4_6_3
void FASTCALL call_10_4_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_6_4
#define init_10_4_6_4
void FASTCALL call_10_4_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_6_5
#define init_10_4_6_5
void FASTCALL call_10_4_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_6_6
#define init_10_4_6_6
void FASTCALL call_10_4_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_6_7
#define init_10_4_6_7
void FASTCALL call_10_4_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_6_8
#define init_10_4_6_8
void FASTCALL call_10_4_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_6_9
#define init_10_4_6_9
void FASTCALL call_10_4_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_6_10
#define init_10_4_6_10
void FASTCALL call_10_4_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_6_11
#define init_10_4_6_11
void FASTCALL call_10_4_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_6_12
#define init_10_4_6_12
void FASTCALL call_10_4_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_6_13
#define init_10_4_6_13
void FASTCALL call_10_4_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_6_14
#define init_10_4_6_14
void FASTCALL call_10_4_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_6_15
#define init_10_4_6_15
void FASTCALL call_10_4_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_7_0
#define init_10_4_7_0
void FASTCALL call_10_4_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_7_1
#define init_10_4_7_1
void FASTCALL call_10_4_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_7_2
#define init_10_4_7_2
void FASTCALL call_10_4_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_7_3
#define init_10_4_7_3
void FASTCALL call_10_4_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_7_4
#define init_10_4_7_4
void FASTCALL call_10_4_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_7_5
#define init_10_4_7_5
void FASTCALL call_10_4_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_7_6
#define init_10_4_7_6
void FASTCALL call_10_4_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_7_7
#define init_10_4_7_7
void FASTCALL call_10_4_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_7_8
#define init_10_4_7_8
void FASTCALL call_10_4_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_7_9
#define init_10_4_7_9
void FASTCALL call_10_4_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_7_10
#define init_10_4_7_10
void FASTCALL call_10_4_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_7_11
#define init_10_4_7_11
void FASTCALL call_10_4_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_7_12
#define init_10_4_7_12
void FASTCALL call_10_4_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_7_13
#define init_10_4_7_13
void FASTCALL call_10_4_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_7_14
#define init_10_4_7_14
void FASTCALL call_10_4_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_7_15
#define init_10_4_7_15
void FASTCALL call_10_4_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_8_0
#define init_10_4_8_0
void FASTCALL call_10_4_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_8_1
#define init_10_4_8_1
void FASTCALL call_10_4_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_8_2
#define init_10_4_8_2
void FASTCALL call_10_4_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_8_3
#define init_10_4_8_3
void FASTCALL call_10_4_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_8_4
#define init_10_4_8_4
void FASTCALL call_10_4_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_8_5
#define init_10_4_8_5
void FASTCALL call_10_4_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_8_6
#define init_10_4_8_6
void FASTCALL call_10_4_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_8_7
#define init_10_4_8_7
void FASTCALL call_10_4_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_8_8
#define init_10_4_8_8
void FASTCALL call_10_4_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_8_9
#define init_10_4_8_9
void FASTCALL call_10_4_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_8_10
#define init_10_4_8_10
void FASTCALL call_10_4_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_8_11
#define init_10_4_8_11
void FASTCALL call_10_4_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_8_12
#define init_10_4_8_12
void FASTCALL call_10_4_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_8_13
#define init_10_4_8_13
void FASTCALL call_10_4_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_8_14
#define init_10_4_8_14
void FASTCALL call_10_4_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_8_15
#define init_10_4_8_15
void FASTCALL call_10_4_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_9_0
#define init_10_4_9_0
void FASTCALL call_10_4_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_9_1
#define init_10_4_9_1
void FASTCALL call_10_4_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_9_2
#define init_10_4_9_2
void FASTCALL call_10_4_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_9_3
#define init_10_4_9_3
void FASTCALL call_10_4_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_9_4
#define init_10_4_9_4
void FASTCALL call_10_4_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_9_5
#define init_10_4_9_5
void FASTCALL call_10_4_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_9_6
#define init_10_4_9_6
void FASTCALL call_10_4_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_9_7
#define init_10_4_9_7
void FASTCALL call_10_4_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_9_8
#define init_10_4_9_8
void FASTCALL call_10_4_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_9_9
#define init_10_4_9_9
void FASTCALL call_10_4_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_9_10
#define init_10_4_9_10
void FASTCALL call_10_4_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_9_11
#define init_10_4_9_11
void FASTCALL call_10_4_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_9_12
#define init_10_4_9_12
void FASTCALL call_10_4_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_9_13
#define init_10_4_9_13
void FASTCALL call_10_4_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_9_14
#define init_10_4_9_14
void FASTCALL call_10_4_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_9_15
#define init_10_4_9_15
void FASTCALL call_10_4_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_10_0
#define init_10_4_10_0
void FASTCALL call_10_4_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_10_1
#define init_10_4_10_1
void FASTCALL call_10_4_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_10_2
#define init_10_4_10_2
void FASTCALL call_10_4_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_10_3
#define init_10_4_10_3
void FASTCALL call_10_4_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_10_4
#define init_10_4_10_4
void FASTCALL call_10_4_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_10_5
#define init_10_4_10_5
void FASTCALL call_10_4_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_10_6
#define init_10_4_10_6
void FASTCALL call_10_4_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_10_7
#define init_10_4_10_7
void FASTCALL call_10_4_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_10_8
#define init_10_4_10_8
void FASTCALL call_10_4_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_10_9
#define init_10_4_10_9
void FASTCALL call_10_4_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_10_10
#define init_10_4_10_10
void FASTCALL call_10_4_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_10_11
#define init_10_4_10_11
void FASTCALL call_10_4_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_10_12
#define init_10_4_10_12
void FASTCALL call_10_4_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_10_13
#define init_10_4_10_13
void FASTCALL call_10_4_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_10_14
#define init_10_4_10_14
void FASTCALL call_10_4_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_10_15
#define init_10_4_10_15
void FASTCALL call_10_4_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_11_0
#define init_10_4_11_0
void FASTCALL call_10_4_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_11_1
#define init_10_4_11_1
void FASTCALL call_10_4_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_11_2
#define init_10_4_11_2
void FASTCALL call_10_4_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_11_3
#define init_10_4_11_3
void FASTCALL call_10_4_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_11_4
#define init_10_4_11_4
void FASTCALL call_10_4_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_11_5
#define init_10_4_11_5
void FASTCALL call_10_4_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_11_6
#define init_10_4_11_6
void FASTCALL call_10_4_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_11_7
#define init_10_4_11_7
void FASTCALL call_10_4_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_11_8
#define init_10_4_11_8
void FASTCALL call_10_4_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_11_9
#define init_10_4_11_9
void FASTCALL call_10_4_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_11_10
#define init_10_4_11_10
void FASTCALL call_10_4_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_11_11
#define init_10_4_11_11
void FASTCALL call_10_4_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_11_12
#define init_10_4_11_12
void FASTCALL call_10_4_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_11_13
#define init_10_4_11_13
void FASTCALL call_10_4_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_11_14
#define init_10_4_11_14
void FASTCALL call_10_4_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_11_15
#define init_10_4_11_15
void FASTCALL call_10_4_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_12_0
#define init_10_4_12_0
void FASTCALL call_10_4_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_12_1
#define init_10_4_12_1
void FASTCALL call_10_4_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_12_2
#define init_10_4_12_2
void FASTCALL call_10_4_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_12_3
#define init_10_4_12_3
void FASTCALL call_10_4_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_12_4
#define init_10_4_12_4
void FASTCALL call_10_4_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_12_5
#define init_10_4_12_5
void FASTCALL call_10_4_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_12_6
#define init_10_4_12_6
void FASTCALL call_10_4_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_12_7
#define init_10_4_12_7
void FASTCALL call_10_4_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_12_8
#define init_10_4_12_8
void FASTCALL call_10_4_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_12_9
#define init_10_4_12_9
void FASTCALL call_10_4_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_12_10
#define init_10_4_12_10
void FASTCALL call_10_4_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_12_11
#define init_10_4_12_11
void FASTCALL call_10_4_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_12_12
#define init_10_4_12_12
void FASTCALL call_10_4_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_12_13
#define init_10_4_12_13
void FASTCALL call_10_4_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_12_14
#define init_10_4_12_14
void FASTCALL call_10_4_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_12_15
#define init_10_4_12_15
void FASTCALL call_10_4_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_13_0
#define init_10_4_13_0
void FASTCALL call_10_4_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_13_1
#define init_10_4_13_1
void FASTCALL call_10_4_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_13_2
#define init_10_4_13_2
void FASTCALL call_10_4_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_13_3
#define init_10_4_13_3
void FASTCALL call_10_4_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_13_4
#define init_10_4_13_4
void FASTCALL call_10_4_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_13_5
#define init_10_4_13_5
void FASTCALL call_10_4_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_13_6
#define init_10_4_13_6
void FASTCALL call_10_4_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_13_7
#define init_10_4_13_7
void FASTCALL call_10_4_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_13_8
#define init_10_4_13_8
void FASTCALL call_10_4_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_13_9
#define init_10_4_13_9
void FASTCALL call_10_4_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_13_10
#define init_10_4_13_10
void FASTCALL call_10_4_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_13_11
#define init_10_4_13_11
void FASTCALL call_10_4_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_13_12
#define init_10_4_13_12
void FASTCALL call_10_4_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_13_13
#define init_10_4_13_13
void FASTCALL call_10_4_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_13_14
#define init_10_4_13_14
void FASTCALL call_10_4_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_13_15
#define init_10_4_13_15
void FASTCALL call_10_4_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_14_0
#define init_10_4_14_0
void FASTCALL call_10_4_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_14_1
#define init_10_4_14_1
void FASTCALL call_10_4_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_14_2
#define init_10_4_14_2
void FASTCALL call_10_4_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_14_3
#define init_10_4_14_3
void FASTCALL call_10_4_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_14_4
#define init_10_4_14_4
void FASTCALL call_10_4_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_14_5
#define init_10_4_14_5
void FASTCALL call_10_4_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_14_6
#define init_10_4_14_6
void FASTCALL call_10_4_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_14_7
#define init_10_4_14_7
void FASTCALL call_10_4_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_14_8
#define init_10_4_14_8
void FASTCALL call_10_4_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_14_9
#define init_10_4_14_9
void FASTCALL call_10_4_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_14_10
#define init_10_4_14_10
void FASTCALL call_10_4_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_14_11
#define init_10_4_14_11
void FASTCALL call_10_4_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_14_12
#define init_10_4_14_12
void FASTCALL call_10_4_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_14_13
#define init_10_4_14_13
void FASTCALL call_10_4_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_14_14
#define init_10_4_14_14
void FASTCALL call_10_4_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_14_15
#define init_10_4_14_15
void FASTCALL call_10_4_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_15_0
#define init_10_4_15_0
void FASTCALL call_10_4_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_15_1
#define init_10_4_15_1
void FASTCALL call_10_4_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_15_2
#define init_10_4_15_2
void FASTCALL call_10_4_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_15_3
#define init_10_4_15_3
void FASTCALL call_10_4_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_15_4
#define init_10_4_15_4
void FASTCALL call_10_4_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_15_5
#define init_10_4_15_5
void FASTCALL call_10_4_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_15_6
#define init_10_4_15_6
void FASTCALL call_10_4_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_15_7
#define init_10_4_15_7
void FASTCALL call_10_4_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_15_8
#define init_10_4_15_8
void FASTCALL call_10_4_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_15_9
#define init_10_4_15_9
void FASTCALL call_10_4_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_15_10
#define init_10_4_15_10
void FASTCALL call_10_4_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_15_11
#define init_10_4_15_11
void FASTCALL call_10_4_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_15_12
#define init_10_4_15_12
void FASTCALL call_10_4_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_15_13
#define init_10_4_15_13
void FASTCALL call_10_4_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_15_14
#define init_10_4_15_14
void FASTCALL call_10_4_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_4_15_15
#define init_10_4_15_15
void FASTCALL call_10_4_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_0_0
#define init_10_5_0_0
void FASTCALL call_10_5_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_0_1
#define init_10_5_0_1
void FASTCALL call_10_5_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_0_2
#define init_10_5_0_2
void FASTCALL call_10_5_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_0_3
#define init_10_5_0_3
void FASTCALL call_10_5_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_0_4
#define init_10_5_0_4
void FASTCALL call_10_5_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_0_5
#define init_10_5_0_5
void FASTCALL call_10_5_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_0_6
#define init_10_5_0_6
void FASTCALL call_10_5_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_0_7
#define init_10_5_0_7
void FASTCALL call_10_5_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_0_8
#define init_10_5_0_8
void FASTCALL call_10_5_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_0_9
#define init_10_5_0_9
void FASTCALL call_10_5_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_0_10
#define init_10_5_0_10
void FASTCALL call_10_5_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_0_11
#define init_10_5_0_11
void FASTCALL call_10_5_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_0_12
#define init_10_5_0_12
void FASTCALL call_10_5_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_0_13
#define init_10_5_0_13
void FASTCALL call_10_5_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_0_14
#define init_10_5_0_14
void FASTCALL call_10_5_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_0_15
#define init_10_5_0_15
void FASTCALL call_10_5_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_1_0
#define init_10_5_1_0
void FASTCALL call_10_5_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_1_1
#define init_10_5_1_1
void FASTCALL call_10_5_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_1_2
#define init_10_5_1_2
void FASTCALL call_10_5_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_1_3
#define init_10_5_1_3
void FASTCALL call_10_5_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_1_4
#define init_10_5_1_4
void FASTCALL call_10_5_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_1_5
#define init_10_5_1_5
void FASTCALL call_10_5_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_1_6
#define init_10_5_1_6
void FASTCALL call_10_5_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_1_7
#define init_10_5_1_7
void FASTCALL call_10_5_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_1_8
#define init_10_5_1_8
void FASTCALL call_10_5_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_1_9
#define init_10_5_1_9
void FASTCALL call_10_5_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_1_10
#define init_10_5_1_10
void FASTCALL call_10_5_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_1_11
#define init_10_5_1_11
void FASTCALL call_10_5_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_1_12
#define init_10_5_1_12
void FASTCALL call_10_5_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_1_13
#define init_10_5_1_13
void FASTCALL call_10_5_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_1_14
#define init_10_5_1_14
void FASTCALL call_10_5_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_1_15
#define init_10_5_1_15
void FASTCALL call_10_5_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_2_0
#define init_10_5_2_0
void FASTCALL call_10_5_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_2_1
#define init_10_5_2_1
void FASTCALL call_10_5_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_2_2
#define init_10_5_2_2
void FASTCALL call_10_5_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_2_3
#define init_10_5_2_3
void FASTCALL call_10_5_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_2_4
#define init_10_5_2_4
void FASTCALL call_10_5_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_2_5
#define init_10_5_2_5
void FASTCALL call_10_5_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_2_6
#define init_10_5_2_6
void FASTCALL call_10_5_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_2_7
#define init_10_5_2_7
void FASTCALL call_10_5_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_2_8
#define init_10_5_2_8
void FASTCALL call_10_5_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_2_9
#define init_10_5_2_9
void FASTCALL call_10_5_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_2_10
#define init_10_5_2_10
void FASTCALL call_10_5_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_2_11
#define init_10_5_2_11
void FASTCALL call_10_5_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_2_12
#define init_10_5_2_12
void FASTCALL call_10_5_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_2_13
#define init_10_5_2_13
void FASTCALL call_10_5_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_2_14
#define init_10_5_2_14
void FASTCALL call_10_5_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_2_15
#define init_10_5_2_15
void FASTCALL call_10_5_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_3_0
#define init_10_5_3_0
void FASTCALL call_10_5_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_3_1
#define init_10_5_3_1
void FASTCALL call_10_5_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_3_2
#define init_10_5_3_2
void FASTCALL call_10_5_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_3_3
#define init_10_5_3_3
void FASTCALL call_10_5_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_3_4
#define init_10_5_3_4
void FASTCALL call_10_5_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_3_5
#define init_10_5_3_5
void FASTCALL call_10_5_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_3_6
#define init_10_5_3_6
void FASTCALL call_10_5_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_3_7
#define init_10_5_3_7
void FASTCALL call_10_5_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_3_8
#define init_10_5_3_8
void FASTCALL call_10_5_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_3_9
#define init_10_5_3_9
void FASTCALL call_10_5_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_3_10
#define init_10_5_3_10
void FASTCALL call_10_5_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_3_11
#define init_10_5_3_11
void FASTCALL call_10_5_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_3_12
#define init_10_5_3_12
void FASTCALL call_10_5_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_3_13
#define init_10_5_3_13
void FASTCALL call_10_5_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_3_14
#define init_10_5_3_14
void FASTCALL call_10_5_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_3_15
#define init_10_5_3_15
void FASTCALL call_10_5_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_4_0
#define init_10_5_4_0
void FASTCALL call_10_5_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_4_1
#define init_10_5_4_1
void FASTCALL call_10_5_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_4_2
#define init_10_5_4_2
void FASTCALL call_10_5_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_4_3
#define init_10_5_4_3
void FASTCALL call_10_5_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_4_4
#define init_10_5_4_4
void FASTCALL call_10_5_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_4_5
#define init_10_5_4_5
void FASTCALL call_10_5_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_4_6
#define init_10_5_4_6
void FASTCALL call_10_5_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_4_7
#define init_10_5_4_7
void FASTCALL call_10_5_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_4_8
#define init_10_5_4_8
void FASTCALL call_10_5_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_4_9
#define init_10_5_4_9
void FASTCALL call_10_5_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_4_10
#define init_10_5_4_10
void FASTCALL call_10_5_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_4_11
#define init_10_5_4_11
void FASTCALL call_10_5_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_4_12
#define init_10_5_4_12
void FASTCALL call_10_5_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_4_13
#define init_10_5_4_13
void FASTCALL call_10_5_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_4_14
#define init_10_5_4_14
void FASTCALL call_10_5_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_4_15
#define init_10_5_4_15
void FASTCALL call_10_5_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_5_0
#define init_10_5_5_0
void FASTCALL call_10_5_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_5_1
#define init_10_5_5_1
void FASTCALL call_10_5_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_5_2
#define init_10_5_5_2
void FASTCALL call_10_5_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_5_3
#define init_10_5_5_3
void FASTCALL call_10_5_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_5_4
#define init_10_5_5_4
void FASTCALL call_10_5_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_5_5
#define init_10_5_5_5
void FASTCALL call_10_5_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_5_6
#define init_10_5_5_6
void FASTCALL call_10_5_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_5_7
#define init_10_5_5_7
void FASTCALL call_10_5_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_5_8
#define init_10_5_5_8
void FASTCALL call_10_5_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_5_9
#define init_10_5_5_9
void FASTCALL call_10_5_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_5_10
#define init_10_5_5_10
void FASTCALL call_10_5_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_5_11
#define init_10_5_5_11
void FASTCALL call_10_5_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_5_12
#define init_10_5_5_12
void FASTCALL call_10_5_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_5_13
#define init_10_5_5_13
void FASTCALL call_10_5_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_5_14
#define init_10_5_5_14
void FASTCALL call_10_5_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_5_15
#define init_10_5_5_15
void FASTCALL call_10_5_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_6_0
#define init_10_5_6_0
void FASTCALL call_10_5_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_6_1
#define init_10_5_6_1
void FASTCALL call_10_5_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_6_2
#define init_10_5_6_2
void FASTCALL call_10_5_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_6_3
#define init_10_5_6_3
void FASTCALL call_10_5_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_6_4
#define init_10_5_6_4
void FASTCALL call_10_5_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_6_5
#define init_10_5_6_5
void FASTCALL call_10_5_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_6_6
#define init_10_5_6_6
void FASTCALL call_10_5_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_6_7
#define init_10_5_6_7
void FASTCALL call_10_5_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_6_8
#define init_10_5_6_8
void FASTCALL call_10_5_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_6_9
#define init_10_5_6_9
void FASTCALL call_10_5_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_6_10
#define init_10_5_6_10
void FASTCALL call_10_5_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_6_11
#define init_10_5_6_11
void FASTCALL call_10_5_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_6_12
#define init_10_5_6_12
void FASTCALL call_10_5_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_6_13
#define init_10_5_6_13
void FASTCALL call_10_5_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_6_14
#define init_10_5_6_14
void FASTCALL call_10_5_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_6_15
#define init_10_5_6_15
void FASTCALL call_10_5_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_7_0
#define init_10_5_7_0
void FASTCALL call_10_5_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_7_1
#define init_10_5_7_1
void FASTCALL call_10_5_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_7_2
#define init_10_5_7_2
void FASTCALL call_10_5_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_7_3
#define init_10_5_7_3
void FASTCALL call_10_5_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_7_4
#define init_10_5_7_4
void FASTCALL call_10_5_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_7_5
#define init_10_5_7_5
void FASTCALL call_10_5_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_7_6
#define init_10_5_7_6
void FASTCALL call_10_5_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_7_7
#define init_10_5_7_7
void FASTCALL call_10_5_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_7_8
#define init_10_5_7_8
void FASTCALL call_10_5_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_7_9
#define init_10_5_7_9
void FASTCALL call_10_5_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_7_10
#define init_10_5_7_10
void FASTCALL call_10_5_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_7_11
#define init_10_5_7_11
void FASTCALL call_10_5_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_7_12
#define init_10_5_7_12
void FASTCALL call_10_5_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_7_13
#define init_10_5_7_13
void FASTCALL call_10_5_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_7_14
#define init_10_5_7_14
void FASTCALL call_10_5_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_7_15
#define init_10_5_7_15
void FASTCALL call_10_5_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_8_0
#define init_10_5_8_0
void FASTCALL call_10_5_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_8_1
#define init_10_5_8_1
void FASTCALL call_10_5_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_8_2
#define init_10_5_8_2
void FASTCALL call_10_5_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_8_3
#define init_10_5_8_3
void FASTCALL call_10_5_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_8_4
#define init_10_5_8_4
void FASTCALL call_10_5_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_8_5
#define init_10_5_8_5
void FASTCALL call_10_5_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_8_6
#define init_10_5_8_6
void FASTCALL call_10_5_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_8_7
#define init_10_5_8_7
void FASTCALL call_10_5_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_8_8
#define init_10_5_8_8
void FASTCALL call_10_5_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_8_9
#define init_10_5_8_9
void FASTCALL call_10_5_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_8_10
#define init_10_5_8_10
void FASTCALL call_10_5_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_8_11
#define init_10_5_8_11
void FASTCALL call_10_5_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_8_12
#define init_10_5_8_12
void FASTCALL call_10_5_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_8_13
#define init_10_5_8_13
void FASTCALL call_10_5_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_8_14
#define init_10_5_8_14
void FASTCALL call_10_5_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_8_15
#define init_10_5_8_15
void FASTCALL call_10_5_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_9_0
#define init_10_5_9_0
void FASTCALL call_10_5_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_9_1
#define init_10_5_9_1
void FASTCALL call_10_5_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_9_2
#define init_10_5_9_2
void FASTCALL call_10_5_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_9_3
#define init_10_5_9_3
void FASTCALL call_10_5_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_9_4
#define init_10_5_9_4
void FASTCALL call_10_5_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_9_5
#define init_10_5_9_5
void FASTCALL call_10_5_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_9_6
#define init_10_5_9_6
void FASTCALL call_10_5_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_9_7
#define init_10_5_9_7
void FASTCALL call_10_5_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_9_8
#define init_10_5_9_8
void FASTCALL call_10_5_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_9_9
#define init_10_5_9_9
void FASTCALL call_10_5_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_9_10
#define init_10_5_9_10
void FASTCALL call_10_5_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_9_11
#define init_10_5_9_11
void FASTCALL call_10_5_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_9_12
#define init_10_5_9_12
void FASTCALL call_10_5_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_9_13
#define init_10_5_9_13
void FASTCALL call_10_5_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_9_14
#define init_10_5_9_14
void FASTCALL call_10_5_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_9_15
#define init_10_5_9_15
void FASTCALL call_10_5_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_10_0
#define init_10_5_10_0
void FASTCALL call_10_5_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_10_1
#define init_10_5_10_1
void FASTCALL call_10_5_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_10_2
#define init_10_5_10_2
void FASTCALL call_10_5_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_10_3
#define init_10_5_10_3
void FASTCALL call_10_5_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_10_4
#define init_10_5_10_4
void FASTCALL call_10_5_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_10_5
#define init_10_5_10_5
void FASTCALL call_10_5_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_10_6
#define init_10_5_10_6
void FASTCALL call_10_5_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_10_7
#define init_10_5_10_7
void FASTCALL call_10_5_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_10_8
#define init_10_5_10_8
void FASTCALL call_10_5_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_10_9
#define init_10_5_10_9
void FASTCALL call_10_5_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_10_10
#define init_10_5_10_10
void FASTCALL call_10_5_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_10_11
#define init_10_5_10_11
void FASTCALL call_10_5_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_10_12
#define init_10_5_10_12
void FASTCALL call_10_5_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_10_13
#define init_10_5_10_13
void FASTCALL call_10_5_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_10_14
#define init_10_5_10_14
void FASTCALL call_10_5_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_10_15
#define init_10_5_10_15
void FASTCALL call_10_5_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_11_0
#define init_10_5_11_0
void FASTCALL call_10_5_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_11_1
#define init_10_5_11_1
void FASTCALL call_10_5_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_11_2
#define init_10_5_11_2
void FASTCALL call_10_5_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_11_3
#define init_10_5_11_3
void FASTCALL call_10_5_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_11_4
#define init_10_5_11_4
void FASTCALL call_10_5_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_11_5
#define init_10_5_11_5
void FASTCALL call_10_5_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_11_6
#define init_10_5_11_6
void FASTCALL call_10_5_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_11_7
#define init_10_5_11_7
void FASTCALL call_10_5_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_11_8
#define init_10_5_11_8
void FASTCALL call_10_5_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_11_9
#define init_10_5_11_9
void FASTCALL call_10_5_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_11_10
#define init_10_5_11_10
void FASTCALL call_10_5_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_11_11
#define init_10_5_11_11
void FASTCALL call_10_5_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_11_12
#define init_10_5_11_12
void FASTCALL call_10_5_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_11_13
#define init_10_5_11_13
void FASTCALL call_10_5_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_11_14
#define init_10_5_11_14
void FASTCALL call_10_5_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_11_15
#define init_10_5_11_15
void FASTCALL call_10_5_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_12_0
#define init_10_5_12_0
void FASTCALL call_10_5_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_12_1
#define init_10_5_12_1
void FASTCALL call_10_5_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_12_2
#define init_10_5_12_2
void FASTCALL call_10_5_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_12_3
#define init_10_5_12_3
void FASTCALL call_10_5_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_12_4
#define init_10_5_12_4
void FASTCALL call_10_5_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_12_5
#define init_10_5_12_5
void FASTCALL call_10_5_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_12_6
#define init_10_5_12_6
void FASTCALL call_10_5_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_12_7
#define init_10_5_12_7
void FASTCALL call_10_5_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_12_8
#define init_10_5_12_8
void FASTCALL call_10_5_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_12_9
#define init_10_5_12_9
void FASTCALL call_10_5_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_12_10
#define init_10_5_12_10
void FASTCALL call_10_5_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_12_11
#define init_10_5_12_11
void FASTCALL call_10_5_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_12_12
#define init_10_5_12_12
void FASTCALL call_10_5_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_12_13
#define init_10_5_12_13
void FASTCALL call_10_5_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_12_14
#define init_10_5_12_14
void FASTCALL call_10_5_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_12_15
#define init_10_5_12_15
void FASTCALL call_10_5_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_13_0
#define init_10_5_13_0
void FASTCALL call_10_5_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_13_1
#define init_10_5_13_1
void FASTCALL call_10_5_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_13_2
#define init_10_5_13_2
void FASTCALL call_10_5_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_13_3
#define init_10_5_13_3
void FASTCALL call_10_5_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_13_4
#define init_10_5_13_4
void FASTCALL call_10_5_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_13_5
#define init_10_5_13_5
void FASTCALL call_10_5_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_13_6
#define init_10_5_13_6
void FASTCALL call_10_5_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_13_7
#define init_10_5_13_7
void FASTCALL call_10_5_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_13_8
#define init_10_5_13_8
void FASTCALL call_10_5_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_13_9
#define init_10_5_13_9
void FASTCALL call_10_5_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_13_10
#define init_10_5_13_10
void FASTCALL call_10_5_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_13_11
#define init_10_5_13_11
void FASTCALL call_10_5_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_13_12
#define init_10_5_13_12
void FASTCALL call_10_5_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_13_13
#define init_10_5_13_13
void FASTCALL call_10_5_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_13_14
#define init_10_5_13_14
void FASTCALL call_10_5_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_13_15
#define init_10_5_13_15
void FASTCALL call_10_5_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_14_0
#define init_10_5_14_0
void FASTCALL call_10_5_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_14_1
#define init_10_5_14_1
void FASTCALL call_10_5_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_14_2
#define init_10_5_14_2
void FASTCALL call_10_5_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_14_3
#define init_10_5_14_3
void FASTCALL call_10_5_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_14_4
#define init_10_5_14_4
void FASTCALL call_10_5_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_14_5
#define init_10_5_14_5
void FASTCALL call_10_5_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_14_6
#define init_10_5_14_6
void FASTCALL call_10_5_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_14_7
#define init_10_5_14_7
void FASTCALL call_10_5_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_14_8
#define init_10_5_14_8
void FASTCALL call_10_5_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_14_9
#define init_10_5_14_9
void FASTCALL call_10_5_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_14_10
#define init_10_5_14_10
void FASTCALL call_10_5_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_14_11
#define init_10_5_14_11
void FASTCALL call_10_5_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_14_12
#define init_10_5_14_12
void FASTCALL call_10_5_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_14_13
#define init_10_5_14_13
void FASTCALL call_10_5_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_14_14
#define init_10_5_14_14
void FASTCALL call_10_5_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_14_15
#define init_10_5_14_15
void FASTCALL call_10_5_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_15_0
#define init_10_5_15_0
void FASTCALL call_10_5_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_15_1
#define init_10_5_15_1
void FASTCALL call_10_5_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_15_2
#define init_10_5_15_2
void FASTCALL call_10_5_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_15_3
#define init_10_5_15_3
void FASTCALL call_10_5_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_15_4
#define init_10_5_15_4
void FASTCALL call_10_5_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_15_5
#define init_10_5_15_5
void FASTCALL call_10_5_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_15_6
#define init_10_5_15_6
void FASTCALL call_10_5_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_15_7
#define init_10_5_15_7
void FASTCALL call_10_5_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_15_8
#define init_10_5_15_8
void FASTCALL call_10_5_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_15_9
#define init_10_5_15_9
void FASTCALL call_10_5_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_15_10
#define init_10_5_15_10
void FASTCALL call_10_5_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_15_11
#define init_10_5_15_11
void FASTCALL call_10_5_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_15_12
#define init_10_5_15_12
void FASTCALL call_10_5_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_15_13
#define init_10_5_15_13
void FASTCALL call_10_5_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_15_14
#define init_10_5_15_14
void FASTCALL call_10_5_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_5_15_15
#define init_10_5_15_15
void FASTCALL call_10_5_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_0_0
#define init_10_6_0_0
void FASTCALL call_10_6_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_0_1
#define init_10_6_0_1
void FASTCALL call_10_6_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_0_2
#define init_10_6_0_2
void FASTCALL call_10_6_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_0_3
#define init_10_6_0_3
void FASTCALL call_10_6_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_0_4
#define init_10_6_0_4
void FASTCALL call_10_6_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_0_5
#define init_10_6_0_5
void FASTCALL call_10_6_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_0_6
#define init_10_6_0_6
void FASTCALL call_10_6_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_0_7
#define init_10_6_0_7
void FASTCALL call_10_6_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_0_8
#define init_10_6_0_8
void FASTCALL call_10_6_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_0_9
#define init_10_6_0_9
void FASTCALL call_10_6_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_0_10
#define init_10_6_0_10
void FASTCALL call_10_6_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_0_11
#define init_10_6_0_11
void FASTCALL call_10_6_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_0_12
#define init_10_6_0_12
void FASTCALL call_10_6_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_0_13
#define init_10_6_0_13
void FASTCALL call_10_6_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_0_14
#define init_10_6_0_14
void FASTCALL call_10_6_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_0_15
#define init_10_6_0_15
void FASTCALL call_10_6_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_1_0
#define init_10_6_1_0
void FASTCALL call_10_6_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_1_1
#define init_10_6_1_1
void FASTCALL call_10_6_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_1_2
#define init_10_6_1_2
void FASTCALL call_10_6_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_1_3
#define init_10_6_1_3
void FASTCALL call_10_6_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_1_4
#define init_10_6_1_4
void FASTCALL call_10_6_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_1_5
#define init_10_6_1_5
void FASTCALL call_10_6_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_1_6
#define init_10_6_1_6
void FASTCALL call_10_6_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_1_7
#define init_10_6_1_7
void FASTCALL call_10_6_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_1_8
#define init_10_6_1_8
void FASTCALL call_10_6_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_1_9
#define init_10_6_1_9
void FASTCALL call_10_6_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_1_10
#define init_10_6_1_10
void FASTCALL call_10_6_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_1_11
#define init_10_6_1_11
void FASTCALL call_10_6_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_1_12
#define init_10_6_1_12
void FASTCALL call_10_6_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_1_13
#define init_10_6_1_13
void FASTCALL call_10_6_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_1_14
#define init_10_6_1_14
void FASTCALL call_10_6_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_1_15
#define init_10_6_1_15
void FASTCALL call_10_6_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_2_0
#define init_10_6_2_0
void FASTCALL call_10_6_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_2_1
#define init_10_6_2_1
void FASTCALL call_10_6_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_2_2
#define init_10_6_2_2
void FASTCALL call_10_6_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_2_3
#define init_10_6_2_3
void FASTCALL call_10_6_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_2_4
#define init_10_6_2_4
void FASTCALL call_10_6_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_2_5
#define init_10_6_2_5
void FASTCALL call_10_6_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_2_6
#define init_10_6_2_6
void FASTCALL call_10_6_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_2_7
#define init_10_6_2_7
void FASTCALL call_10_6_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_2_8
#define init_10_6_2_8
void FASTCALL call_10_6_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_2_9
#define init_10_6_2_9
void FASTCALL call_10_6_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_2_10
#define init_10_6_2_10
void FASTCALL call_10_6_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_2_11
#define init_10_6_2_11
void FASTCALL call_10_6_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_2_12
#define init_10_6_2_12
void FASTCALL call_10_6_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_2_13
#define init_10_6_2_13
void FASTCALL call_10_6_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_2_14
#define init_10_6_2_14
void FASTCALL call_10_6_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_2_15
#define init_10_6_2_15
void FASTCALL call_10_6_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_3_0
#define init_10_6_3_0
void FASTCALL call_10_6_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_3_1
#define init_10_6_3_1
void FASTCALL call_10_6_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_3_2
#define init_10_6_3_2
void FASTCALL call_10_6_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_3_3
#define init_10_6_3_3
void FASTCALL call_10_6_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_3_4
#define init_10_6_3_4
void FASTCALL call_10_6_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_3_5
#define init_10_6_3_5
void FASTCALL call_10_6_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_3_6
#define init_10_6_3_6
void FASTCALL call_10_6_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_3_7
#define init_10_6_3_7
void FASTCALL call_10_6_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_3_8
#define init_10_6_3_8
void FASTCALL call_10_6_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_3_9
#define init_10_6_3_9
void FASTCALL call_10_6_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_3_10
#define init_10_6_3_10
void FASTCALL call_10_6_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_3_11
#define init_10_6_3_11
void FASTCALL call_10_6_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_3_12
#define init_10_6_3_12
void FASTCALL call_10_6_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_3_13
#define init_10_6_3_13
void FASTCALL call_10_6_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_3_14
#define init_10_6_3_14
void FASTCALL call_10_6_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_3_15
#define init_10_6_3_15
void FASTCALL call_10_6_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_4_0
#define init_10_6_4_0
void FASTCALL call_10_6_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_4_1
#define init_10_6_4_1
void FASTCALL call_10_6_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_4_2
#define init_10_6_4_2
void FASTCALL call_10_6_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_4_3
#define init_10_6_4_3
void FASTCALL call_10_6_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_4_4
#define init_10_6_4_4
void FASTCALL call_10_6_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_4_5
#define init_10_6_4_5
void FASTCALL call_10_6_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_4_6
#define init_10_6_4_6
void FASTCALL call_10_6_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_4_7
#define init_10_6_4_7
void FASTCALL call_10_6_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_4_8
#define init_10_6_4_8
void FASTCALL call_10_6_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_4_9
#define init_10_6_4_9
void FASTCALL call_10_6_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_4_10
#define init_10_6_4_10
void FASTCALL call_10_6_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_4_11
#define init_10_6_4_11
void FASTCALL call_10_6_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_4_12
#define init_10_6_4_12
void FASTCALL call_10_6_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_4_13
#define init_10_6_4_13
void FASTCALL call_10_6_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_4_14
#define init_10_6_4_14
void FASTCALL call_10_6_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_4_15
#define init_10_6_4_15
void FASTCALL call_10_6_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_5_0
#define init_10_6_5_0
void FASTCALL call_10_6_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_5_1
#define init_10_6_5_1
void FASTCALL call_10_6_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_5_2
#define init_10_6_5_2
void FASTCALL call_10_6_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_5_3
#define init_10_6_5_3
void FASTCALL call_10_6_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_5_4
#define init_10_6_5_4
void FASTCALL call_10_6_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_5_5
#define init_10_6_5_5
void FASTCALL call_10_6_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_5_6
#define init_10_6_5_6
void FASTCALL call_10_6_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_5_7
#define init_10_6_5_7
void FASTCALL call_10_6_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_5_8
#define init_10_6_5_8
void FASTCALL call_10_6_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_5_9
#define init_10_6_5_9
void FASTCALL call_10_6_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_5_10
#define init_10_6_5_10
void FASTCALL call_10_6_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_5_11
#define init_10_6_5_11
void FASTCALL call_10_6_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_5_12
#define init_10_6_5_12
void FASTCALL call_10_6_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_5_13
#define init_10_6_5_13
void FASTCALL call_10_6_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_5_14
#define init_10_6_5_14
void FASTCALL call_10_6_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_5_15
#define init_10_6_5_15
void FASTCALL call_10_6_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_6_0
#define init_10_6_6_0
void FASTCALL call_10_6_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_6_1
#define init_10_6_6_1
void FASTCALL call_10_6_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_6_2
#define init_10_6_6_2
void FASTCALL call_10_6_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_6_3
#define init_10_6_6_3
void FASTCALL call_10_6_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_6_4
#define init_10_6_6_4
void FASTCALL call_10_6_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_6_5
#define init_10_6_6_5
void FASTCALL call_10_6_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_6_6
#define init_10_6_6_6
void FASTCALL call_10_6_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_6_7
#define init_10_6_6_7
void FASTCALL call_10_6_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_6_8
#define init_10_6_6_8
void FASTCALL call_10_6_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_6_9
#define init_10_6_6_9
void FASTCALL call_10_6_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_6_10
#define init_10_6_6_10
void FASTCALL call_10_6_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_6_11
#define init_10_6_6_11
void FASTCALL call_10_6_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_6_12
#define init_10_6_6_12
void FASTCALL call_10_6_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_6_13
#define init_10_6_6_13
void FASTCALL call_10_6_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_6_14
#define init_10_6_6_14
void FASTCALL call_10_6_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_6_15
#define init_10_6_6_15
void FASTCALL call_10_6_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_7_0
#define init_10_6_7_0
void FASTCALL call_10_6_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_7_1
#define init_10_6_7_1
void FASTCALL call_10_6_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_7_2
#define init_10_6_7_2
void FASTCALL call_10_6_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_7_3
#define init_10_6_7_3
void FASTCALL call_10_6_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_7_4
#define init_10_6_7_4
void FASTCALL call_10_6_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_7_5
#define init_10_6_7_5
void FASTCALL call_10_6_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_7_6
#define init_10_6_7_6
void FASTCALL call_10_6_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_7_7
#define init_10_6_7_7
void FASTCALL call_10_6_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_7_8
#define init_10_6_7_8
void FASTCALL call_10_6_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_7_9
#define init_10_6_7_9
void FASTCALL call_10_6_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_7_10
#define init_10_6_7_10
void FASTCALL call_10_6_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_7_11
#define init_10_6_7_11
void FASTCALL call_10_6_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_7_12
#define init_10_6_7_12
void FASTCALL call_10_6_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_7_13
#define init_10_6_7_13
void FASTCALL call_10_6_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_7_14
#define init_10_6_7_14
void FASTCALL call_10_6_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_7_15
#define init_10_6_7_15
void FASTCALL call_10_6_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_8_0
#define init_10_6_8_0
void FASTCALL call_10_6_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_8_1
#define init_10_6_8_1
void FASTCALL call_10_6_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_8_2
#define init_10_6_8_2
void FASTCALL call_10_6_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_8_3
#define init_10_6_8_3
void FASTCALL call_10_6_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_8_4
#define init_10_6_8_4
void FASTCALL call_10_6_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_8_5
#define init_10_6_8_5
void FASTCALL call_10_6_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_8_6
#define init_10_6_8_6
void FASTCALL call_10_6_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_8_7
#define init_10_6_8_7
void FASTCALL call_10_6_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_8_8
#define init_10_6_8_8
void FASTCALL call_10_6_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_8_9
#define init_10_6_8_9
void FASTCALL call_10_6_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_8_10
#define init_10_6_8_10
void FASTCALL call_10_6_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_8_11
#define init_10_6_8_11
void FASTCALL call_10_6_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_8_12
#define init_10_6_8_12
void FASTCALL call_10_6_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_8_13
#define init_10_6_8_13
void FASTCALL call_10_6_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_8_14
#define init_10_6_8_14
void FASTCALL call_10_6_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_8_15
#define init_10_6_8_15
void FASTCALL call_10_6_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_9_0
#define init_10_6_9_0
void FASTCALL call_10_6_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_9_1
#define init_10_6_9_1
void FASTCALL call_10_6_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_9_2
#define init_10_6_9_2
void FASTCALL call_10_6_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_9_3
#define init_10_6_9_3
void FASTCALL call_10_6_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_9_4
#define init_10_6_9_4
void FASTCALL call_10_6_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_9_5
#define init_10_6_9_5
void FASTCALL call_10_6_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_9_6
#define init_10_6_9_6
void FASTCALL call_10_6_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_9_7
#define init_10_6_9_7
void FASTCALL call_10_6_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_9_8
#define init_10_6_9_8
void FASTCALL call_10_6_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_9_9
#define init_10_6_9_9
void FASTCALL call_10_6_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_9_10
#define init_10_6_9_10
void FASTCALL call_10_6_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_9_11
#define init_10_6_9_11
void FASTCALL call_10_6_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_9_12
#define init_10_6_9_12
void FASTCALL call_10_6_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_9_13
#define init_10_6_9_13
void FASTCALL call_10_6_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_9_14
#define init_10_6_9_14
void FASTCALL call_10_6_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_9_15
#define init_10_6_9_15
void FASTCALL call_10_6_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_10_0
#define init_10_6_10_0
void FASTCALL call_10_6_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_10_1
#define init_10_6_10_1
void FASTCALL call_10_6_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_10_2
#define init_10_6_10_2
void FASTCALL call_10_6_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_10_3
#define init_10_6_10_3
void FASTCALL call_10_6_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_10_4
#define init_10_6_10_4
void FASTCALL call_10_6_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_10_5
#define init_10_6_10_5
void FASTCALL call_10_6_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_10_6
#define init_10_6_10_6
void FASTCALL call_10_6_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_10_7
#define init_10_6_10_7
void FASTCALL call_10_6_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_10_8
#define init_10_6_10_8
void FASTCALL call_10_6_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_10_9
#define init_10_6_10_9
void FASTCALL call_10_6_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_10_10
#define init_10_6_10_10
void FASTCALL call_10_6_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_10_11
#define init_10_6_10_11
void FASTCALL call_10_6_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_10_12
#define init_10_6_10_12
void FASTCALL call_10_6_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_10_13
#define init_10_6_10_13
void FASTCALL call_10_6_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_10_14
#define init_10_6_10_14
void FASTCALL call_10_6_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_10_15
#define init_10_6_10_15
void FASTCALL call_10_6_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_11_0
#define init_10_6_11_0
void FASTCALL call_10_6_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_11_1
#define init_10_6_11_1
void FASTCALL call_10_6_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_11_2
#define init_10_6_11_2
void FASTCALL call_10_6_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_11_3
#define init_10_6_11_3
void FASTCALL call_10_6_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_11_4
#define init_10_6_11_4
void FASTCALL call_10_6_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_11_5
#define init_10_6_11_5
void FASTCALL call_10_6_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_11_6
#define init_10_6_11_6
void FASTCALL call_10_6_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_11_7
#define init_10_6_11_7
void FASTCALL call_10_6_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_11_8
#define init_10_6_11_8
void FASTCALL call_10_6_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_11_9
#define init_10_6_11_9
void FASTCALL call_10_6_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_11_10
#define init_10_6_11_10
void FASTCALL call_10_6_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_11_11
#define init_10_6_11_11
void FASTCALL call_10_6_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_11_12
#define init_10_6_11_12
void FASTCALL call_10_6_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_11_13
#define init_10_6_11_13
void FASTCALL call_10_6_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_11_14
#define init_10_6_11_14
void FASTCALL call_10_6_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_11_15
#define init_10_6_11_15
void FASTCALL call_10_6_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_12_0
#define init_10_6_12_0
void FASTCALL call_10_6_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_12_1
#define init_10_6_12_1
void FASTCALL call_10_6_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_12_2
#define init_10_6_12_2
void FASTCALL call_10_6_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_12_3
#define init_10_6_12_3
void FASTCALL call_10_6_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_12_4
#define init_10_6_12_4
void FASTCALL call_10_6_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_12_5
#define init_10_6_12_5
void FASTCALL call_10_6_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_12_6
#define init_10_6_12_6
void FASTCALL call_10_6_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_12_7
#define init_10_6_12_7
void FASTCALL call_10_6_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_12_8
#define init_10_6_12_8
void FASTCALL call_10_6_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_12_9
#define init_10_6_12_9
void FASTCALL call_10_6_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_12_10
#define init_10_6_12_10
void FASTCALL call_10_6_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_12_11
#define init_10_6_12_11
void FASTCALL call_10_6_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_12_12
#define init_10_6_12_12
void FASTCALL call_10_6_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_12_13
#define init_10_6_12_13
void FASTCALL call_10_6_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_12_14
#define init_10_6_12_14
void FASTCALL call_10_6_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_12_15
#define init_10_6_12_15
void FASTCALL call_10_6_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_13_0
#define init_10_6_13_0
void FASTCALL call_10_6_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_13_1
#define init_10_6_13_1
void FASTCALL call_10_6_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_13_2
#define init_10_6_13_2
void FASTCALL call_10_6_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_13_3
#define init_10_6_13_3
void FASTCALL call_10_6_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_13_4
#define init_10_6_13_4
void FASTCALL call_10_6_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_13_5
#define init_10_6_13_5
void FASTCALL call_10_6_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_13_6
#define init_10_6_13_6
void FASTCALL call_10_6_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_13_7
#define init_10_6_13_7
void FASTCALL call_10_6_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_13_8
#define init_10_6_13_8
void FASTCALL call_10_6_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_13_9
#define init_10_6_13_9
void FASTCALL call_10_6_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_13_10
#define init_10_6_13_10
void FASTCALL call_10_6_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_13_11
#define init_10_6_13_11
void FASTCALL call_10_6_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_13_12
#define init_10_6_13_12
void FASTCALL call_10_6_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_13_13
#define init_10_6_13_13
void FASTCALL call_10_6_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_13_14
#define init_10_6_13_14
void FASTCALL call_10_6_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_13_15
#define init_10_6_13_15
void FASTCALL call_10_6_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_14_0
#define init_10_6_14_0
void FASTCALL call_10_6_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_14_1
#define init_10_6_14_1
void FASTCALL call_10_6_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_14_2
#define init_10_6_14_2
void FASTCALL call_10_6_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_14_3
#define init_10_6_14_3
void FASTCALL call_10_6_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_14_4
#define init_10_6_14_4
void FASTCALL call_10_6_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_14_5
#define init_10_6_14_5
void FASTCALL call_10_6_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_14_6
#define init_10_6_14_6
void FASTCALL call_10_6_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_14_7
#define init_10_6_14_7
void FASTCALL call_10_6_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_14_8
#define init_10_6_14_8
void FASTCALL call_10_6_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_14_9
#define init_10_6_14_9
void FASTCALL call_10_6_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_14_10
#define init_10_6_14_10
void FASTCALL call_10_6_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_14_11
#define init_10_6_14_11
void FASTCALL call_10_6_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_14_12
#define init_10_6_14_12
void FASTCALL call_10_6_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_14_13
#define init_10_6_14_13
void FASTCALL call_10_6_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_14_14
#define init_10_6_14_14
void FASTCALL call_10_6_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_14_15
#define init_10_6_14_15
void FASTCALL call_10_6_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_15_0
#define init_10_6_15_0
void FASTCALL call_10_6_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_15_1
#define init_10_6_15_1
void FASTCALL call_10_6_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_15_2
#define init_10_6_15_2
void FASTCALL call_10_6_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_15_3
#define init_10_6_15_3
void FASTCALL call_10_6_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_15_4
#define init_10_6_15_4
void FASTCALL call_10_6_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_15_5
#define init_10_6_15_5
void FASTCALL call_10_6_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_15_6
#define init_10_6_15_6
void FASTCALL call_10_6_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_15_7
#define init_10_6_15_7
void FASTCALL call_10_6_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_15_8
#define init_10_6_15_8
void FASTCALL call_10_6_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_15_9
#define init_10_6_15_9
void FASTCALL call_10_6_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_15_10
#define init_10_6_15_10
void FASTCALL call_10_6_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_15_11
#define init_10_6_15_11
void FASTCALL call_10_6_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_15_12
#define init_10_6_15_12
void FASTCALL call_10_6_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_15_13
#define init_10_6_15_13
void FASTCALL call_10_6_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_15_14
#define init_10_6_15_14
void FASTCALL call_10_6_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_6_15_15
#define init_10_6_15_15
void FASTCALL call_10_6_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_0_0
#define init_10_7_0_0
void FASTCALL call_10_7_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_0_1
#define init_10_7_0_1
void FASTCALL call_10_7_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_0_2
#define init_10_7_0_2
void FASTCALL call_10_7_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_0_3
#define init_10_7_0_3
void FASTCALL call_10_7_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_0_4
#define init_10_7_0_4
void FASTCALL call_10_7_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_0_5
#define init_10_7_0_5
void FASTCALL call_10_7_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_0_6
#define init_10_7_0_6
void FASTCALL call_10_7_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_0_7
#define init_10_7_0_7
void FASTCALL call_10_7_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_0_8
#define init_10_7_0_8
void FASTCALL call_10_7_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_0_9
#define init_10_7_0_9
void FASTCALL call_10_7_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_0_10
#define init_10_7_0_10
void FASTCALL call_10_7_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_0_11
#define init_10_7_0_11
void FASTCALL call_10_7_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_0_12
#define init_10_7_0_12
void FASTCALL call_10_7_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_0_13
#define init_10_7_0_13
void FASTCALL call_10_7_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_0_14
#define init_10_7_0_14
void FASTCALL call_10_7_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_0_15
#define init_10_7_0_15
void FASTCALL call_10_7_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_1_0
#define init_10_7_1_0
void FASTCALL call_10_7_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_1_1
#define init_10_7_1_1
void FASTCALL call_10_7_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_1_2
#define init_10_7_1_2
void FASTCALL call_10_7_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_1_3
#define init_10_7_1_3
void FASTCALL call_10_7_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_1_4
#define init_10_7_1_4
void FASTCALL call_10_7_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_1_5
#define init_10_7_1_5
void FASTCALL call_10_7_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_1_6
#define init_10_7_1_6
void FASTCALL call_10_7_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_1_7
#define init_10_7_1_7
void FASTCALL call_10_7_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_1_8
#define init_10_7_1_8
void FASTCALL call_10_7_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_1_9
#define init_10_7_1_9
void FASTCALL call_10_7_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_1_10
#define init_10_7_1_10
void FASTCALL call_10_7_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_1_11
#define init_10_7_1_11
void FASTCALL call_10_7_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_1_12
#define init_10_7_1_12
void FASTCALL call_10_7_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_1_13
#define init_10_7_1_13
void FASTCALL call_10_7_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_1_14
#define init_10_7_1_14
void FASTCALL call_10_7_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_1_15
#define init_10_7_1_15
void FASTCALL call_10_7_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_2_0
#define init_10_7_2_0
void FASTCALL call_10_7_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_2_1
#define init_10_7_2_1
void FASTCALL call_10_7_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_2_2
#define init_10_7_2_2
void FASTCALL call_10_7_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_2_3
#define init_10_7_2_3
void FASTCALL call_10_7_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_2_4
#define init_10_7_2_4
void FASTCALL call_10_7_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_2_5
#define init_10_7_2_5
void FASTCALL call_10_7_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_2_6
#define init_10_7_2_6
void FASTCALL call_10_7_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_2_7
#define init_10_7_2_7
void FASTCALL call_10_7_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_2_8
#define init_10_7_2_8
void FASTCALL call_10_7_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_2_9
#define init_10_7_2_9
void FASTCALL call_10_7_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_2_10
#define init_10_7_2_10
void FASTCALL call_10_7_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_2_11
#define init_10_7_2_11
void FASTCALL call_10_7_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_2_12
#define init_10_7_2_12
void FASTCALL call_10_7_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_2_13
#define init_10_7_2_13
void FASTCALL call_10_7_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_2_14
#define init_10_7_2_14
void FASTCALL call_10_7_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_2_15
#define init_10_7_2_15
void FASTCALL call_10_7_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_3_0
#define init_10_7_3_0
void FASTCALL call_10_7_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_3_1
#define init_10_7_3_1
void FASTCALL call_10_7_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_3_2
#define init_10_7_3_2
void FASTCALL call_10_7_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_3_3
#define init_10_7_3_3
void FASTCALL call_10_7_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_3_4
#define init_10_7_3_4
void FASTCALL call_10_7_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_3_5
#define init_10_7_3_5
void FASTCALL call_10_7_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_3_6
#define init_10_7_3_6
void FASTCALL call_10_7_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_3_7
#define init_10_7_3_7
void FASTCALL call_10_7_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_3_8
#define init_10_7_3_8
void FASTCALL call_10_7_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_3_9
#define init_10_7_3_9
void FASTCALL call_10_7_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_3_10
#define init_10_7_3_10
void FASTCALL call_10_7_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_3_11
#define init_10_7_3_11
void FASTCALL call_10_7_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_3_12
#define init_10_7_3_12
void FASTCALL call_10_7_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_3_13
#define init_10_7_3_13
void FASTCALL call_10_7_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_3_14
#define init_10_7_3_14
void FASTCALL call_10_7_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_3_15
#define init_10_7_3_15
void FASTCALL call_10_7_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_4_0
#define init_10_7_4_0
void FASTCALL call_10_7_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_4_1
#define init_10_7_4_1
void FASTCALL call_10_7_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_4_2
#define init_10_7_4_2
void FASTCALL call_10_7_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_4_3
#define init_10_7_4_3
void FASTCALL call_10_7_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_4_4
#define init_10_7_4_4
void FASTCALL call_10_7_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_4_5
#define init_10_7_4_5
void FASTCALL call_10_7_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_4_6
#define init_10_7_4_6
void FASTCALL call_10_7_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_4_7
#define init_10_7_4_7
void FASTCALL call_10_7_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_4_8
#define init_10_7_4_8
void FASTCALL call_10_7_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_4_9
#define init_10_7_4_9
void FASTCALL call_10_7_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_4_10
#define init_10_7_4_10
void FASTCALL call_10_7_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_4_11
#define init_10_7_4_11
void FASTCALL call_10_7_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_4_12
#define init_10_7_4_12
void FASTCALL call_10_7_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_4_13
#define init_10_7_4_13
void FASTCALL call_10_7_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_4_14
#define init_10_7_4_14
void FASTCALL call_10_7_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_4_15
#define init_10_7_4_15
void FASTCALL call_10_7_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_5_0
#define init_10_7_5_0
void FASTCALL call_10_7_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_5_1
#define init_10_7_5_1
void FASTCALL call_10_7_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_5_2
#define init_10_7_5_2
void FASTCALL call_10_7_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_5_3
#define init_10_7_5_3
void FASTCALL call_10_7_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_5_4
#define init_10_7_5_4
void FASTCALL call_10_7_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_5_5
#define init_10_7_5_5
void FASTCALL call_10_7_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_5_6
#define init_10_7_5_6
void FASTCALL call_10_7_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_5_7
#define init_10_7_5_7
void FASTCALL call_10_7_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_5_8
#define init_10_7_5_8
void FASTCALL call_10_7_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_5_9
#define init_10_7_5_9
void FASTCALL call_10_7_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_5_10
#define init_10_7_5_10
void FASTCALL call_10_7_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_5_11
#define init_10_7_5_11
void FASTCALL call_10_7_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_5_12
#define init_10_7_5_12
void FASTCALL call_10_7_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_5_13
#define init_10_7_5_13
void FASTCALL call_10_7_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_5_14
#define init_10_7_5_14
void FASTCALL call_10_7_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_5_15
#define init_10_7_5_15
void FASTCALL call_10_7_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_6_0
#define init_10_7_6_0
void FASTCALL call_10_7_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_6_1
#define init_10_7_6_1
void FASTCALL call_10_7_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_6_2
#define init_10_7_6_2
void FASTCALL call_10_7_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_6_3
#define init_10_7_6_3
void FASTCALL call_10_7_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_6_4
#define init_10_7_6_4
void FASTCALL call_10_7_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_6_5
#define init_10_7_6_5
void FASTCALL call_10_7_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_6_6
#define init_10_7_6_6
void FASTCALL call_10_7_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_6_7
#define init_10_7_6_7
void FASTCALL call_10_7_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_6_8
#define init_10_7_6_8
void FASTCALL call_10_7_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_6_9
#define init_10_7_6_9
void FASTCALL call_10_7_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_6_10
#define init_10_7_6_10
void FASTCALL call_10_7_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_6_11
#define init_10_7_6_11
void FASTCALL call_10_7_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_6_12
#define init_10_7_6_12
void FASTCALL call_10_7_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_6_13
#define init_10_7_6_13
void FASTCALL call_10_7_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_6_14
#define init_10_7_6_14
void FASTCALL call_10_7_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_6_15
#define init_10_7_6_15
void FASTCALL call_10_7_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_7_0
#define init_10_7_7_0
void FASTCALL call_10_7_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_7_1
#define init_10_7_7_1
void FASTCALL call_10_7_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_7_2
#define init_10_7_7_2
void FASTCALL call_10_7_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_7_3
#define init_10_7_7_3
void FASTCALL call_10_7_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_7_4
#define init_10_7_7_4
void FASTCALL call_10_7_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_7_5
#define init_10_7_7_5
void FASTCALL call_10_7_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_7_6
#define init_10_7_7_6
void FASTCALL call_10_7_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_7_7
#define init_10_7_7_7
void FASTCALL call_10_7_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_7_8
#define init_10_7_7_8
void FASTCALL call_10_7_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_7_9
#define init_10_7_7_9
void FASTCALL call_10_7_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_7_10
#define init_10_7_7_10
void FASTCALL call_10_7_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_7_11
#define init_10_7_7_11
void FASTCALL call_10_7_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_7_12
#define init_10_7_7_12
void FASTCALL call_10_7_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_7_13
#define init_10_7_7_13
void FASTCALL call_10_7_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_7_14
#define init_10_7_7_14
void FASTCALL call_10_7_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_7_15
#define init_10_7_7_15
void FASTCALL call_10_7_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_8_0
#define init_10_7_8_0
void FASTCALL call_10_7_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_8_1
#define init_10_7_8_1
void FASTCALL call_10_7_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_8_2
#define init_10_7_8_2
void FASTCALL call_10_7_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_8_3
#define init_10_7_8_3
void FASTCALL call_10_7_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_8_4
#define init_10_7_8_4
void FASTCALL call_10_7_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_8_5
#define init_10_7_8_5
void FASTCALL call_10_7_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_8_6
#define init_10_7_8_6
void FASTCALL call_10_7_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_8_7
#define init_10_7_8_7
void FASTCALL call_10_7_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_8_8
#define init_10_7_8_8
void FASTCALL call_10_7_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_8_9
#define init_10_7_8_9
void FASTCALL call_10_7_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_8_10
#define init_10_7_8_10
void FASTCALL call_10_7_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_8_11
#define init_10_7_8_11
void FASTCALL call_10_7_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_8_12
#define init_10_7_8_12
void FASTCALL call_10_7_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_8_13
#define init_10_7_8_13
void FASTCALL call_10_7_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_8_14
#define init_10_7_8_14
void FASTCALL call_10_7_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_8_15
#define init_10_7_8_15
void FASTCALL call_10_7_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_9_0
#define init_10_7_9_0
void FASTCALL call_10_7_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_9_1
#define init_10_7_9_1
void FASTCALL call_10_7_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_9_2
#define init_10_7_9_2
void FASTCALL call_10_7_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_9_3
#define init_10_7_9_3
void FASTCALL call_10_7_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_9_4
#define init_10_7_9_4
void FASTCALL call_10_7_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_9_5
#define init_10_7_9_5
void FASTCALL call_10_7_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_9_6
#define init_10_7_9_6
void FASTCALL call_10_7_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_9_7
#define init_10_7_9_7
void FASTCALL call_10_7_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_9_8
#define init_10_7_9_8
void FASTCALL call_10_7_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_9_9
#define init_10_7_9_9
void FASTCALL call_10_7_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_9_10
#define init_10_7_9_10
void FASTCALL call_10_7_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_9_11
#define init_10_7_9_11
void FASTCALL call_10_7_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_9_12
#define init_10_7_9_12
void FASTCALL call_10_7_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_9_13
#define init_10_7_9_13
void FASTCALL call_10_7_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_9_14
#define init_10_7_9_14
void FASTCALL call_10_7_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_9_15
#define init_10_7_9_15
void FASTCALL call_10_7_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_10_0
#define init_10_7_10_0
void FASTCALL call_10_7_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_10_1
#define init_10_7_10_1
void FASTCALL call_10_7_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_10_2
#define init_10_7_10_2
void FASTCALL call_10_7_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_10_3
#define init_10_7_10_3
void FASTCALL call_10_7_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_10_4
#define init_10_7_10_4
void FASTCALL call_10_7_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_10_5
#define init_10_7_10_5
void FASTCALL call_10_7_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_10_6
#define init_10_7_10_6
void FASTCALL call_10_7_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_10_7
#define init_10_7_10_7
void FASTCALL call_10_7_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_10_8
#define init_10_7_10_8
void FASTCALL call_10_7_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_10_9
#define init_10_7_10_9
void FASTCALL call_10_7_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_10_10
#define init_10_7_10_10
void FASTCALL call_10_7_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_10_11
#define init_10_7_10_11
void FASTCALL call_10_7_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_10_12
#define init_10_7_10_12
void FASTCALL call_10_7_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_10_13
#define init_10_7_10_13
void FASTCALL call_10_7_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_10_14
#define init_10_7_10_14
void FASTCALL call_10_7_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_10_15
#define init_10_7_10_15
void FASTCALL call_10_7_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_11_0
#define init_10_7_11_0
void FASTCALL call_10_7_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_11_1
#define init_10_7_11_1
void FASTCALL call_10_7_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_11_2
#define init_10_7_11_2
void FASTCALL call_10_7_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_11_3
#define init_10_7_11_3
void FASTCALL call_10_7_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_11_4
#define init_10_7_11_4
void FASTCALL call_10_7_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_11_5
#define init_10_7_11_5
void FASTCALL call_10_7_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_11_6
#define init_10_7_11_6
void FASTCALL call_10_7_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_11_7
#define init_10_7_11_7
void FASTCALL call_10_7_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_11_8
#define init_10_7_11_8
void FASTCALL call_10_7_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_11_9
#define init_10_7_11_9
void FASTCALL call_10_7_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_11_10
#define init_10_7_11_10
void FASTCALL call_10_7_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_11_11
#define init_10_7_11_11
void FASTCALL call_10_7_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_11_12
#define init_10_7_11_12
void FASTCALL call_10_7_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_11_13
#define init_10_7_11_13
void FASTCALL call_10_7_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_11_14
#define init_10_7_11_14
void FASTCALL call_10_7_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_11_15
#define init_10_7_11_15
void FASTCALL call_10_7_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_12_0
#define init_10_7_12_0
void FASTCALL call_10_7_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_12_1
#define init_10_7_12_1
void FASTCALL call_10_7_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_12_2
#define init_10_7_12_2
void FASTCALL call_10_7_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_12_3
#define init_10_7_12_3
void FASTCALL call_10_7_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_12_4
#define init_10_7_12_4
void FASTCALL call_10_7_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_12_5
#define init_10_7_12_5
void FASTCALL call_10_7_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_12_6
#define init_10_7_12_6
void FASTCALL call_10_7_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_12_7
#define init_10_7_12_7
void FASTCALL call_10_7_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_12_8
#define init_10_7_12_8
void FASTCALL call_10_7_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_12_9
#define init_10_7_12_9
void FASTCALL call_10_7_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_12_10
#define init_10_7_12_10
void FASTCALL call_10_7_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_12_11
#define init_10_7_12_11
void FASTCALL call_10_7_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_12_12
#define init_10_7_12_12
void FASTCALL call_10_7_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_12_13
#define init_10_7_12_13
void FASTCALL call_10_7_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_12_14
#define init_10_7_12_14
void FASTCALL call_10_7_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_12_15
#define init_10_7_12_15
void FASTCALL call_10_7_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_13_0
#define init_10_7_13_0
void FASTCALL call_10_7_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_13_1
#define init_10_7_13_1
void FASTCALL call_10_7_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_13_2
#define init_10_7_13_2
void FASTCALL call_10_7_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_13_3
#define init_10_7_13_3
void FASTCALL call_10_7_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_13_4
#define init_10_7_13_4
void FASTCALL call_10_7_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_13_5
#define init_10_7_13_5
void FASTCALL call_10_7_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_13_6
#define init_10_7_13_6
void FASTCALL call_10_7_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_13_7
#define init_10_7_13_7
void FASTCALL call_10_7_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_13_8
#define init_10_7_13_8
void FASTCALL call_10_7_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_13_9
#define init_10_7_13_9
void FASTCALL call_10_7_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_13_10
#define init_10_7_13_10
void FASTCALL call_10_7_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_13_11
#define init_10_7_13_11
void FASTCALL call_10_7_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_13_12
#define init_10_7_13_12
void FASTCALL call_10_7_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_13_13
#define init_10_7_13_13
void FASTCALL call_10_7_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_13_14
#define init_10_7_13_14
void FASTCALL call_10_7_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_13_15
#define init_10_7_13_15
void FASTCALL call_10_7_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_14_0
#define init_10_7_14_0
void FASTCALL call_10_7_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_14_1
#define init_10_7_14_1
void FASTCALL call_10_7_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_14_2
#define init_10_7_14_2
void FASTCALL call_10_7_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_14_3
#define init_10_7_14_3
void FASTCALL call_10_7_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_14_4
#define init_10_7_14_4
void FASTCALL call_10_7_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_14_5
#define init_10_7_14_5
void FASTCALL call_10_7_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_14_6
#define init_10_7_14_6
void FASTCALL call_10_7_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_14_7
#define init_10_7_14_7
void FASTCALL call_10_7_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_14_8
#define init_10_7_14_8
void FASTCALL call_10_7_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_14_9
#define init_10_7_14_9
void FASTCALL call_10_7_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_14_10
#define init_10_7_14_10
void FASTCALL call_10_7_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_14_11
#define init_10_7_14_11
void FASTCALL call_10_7_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_14_12
#define init_10_7_14_12
void FASTCALL call_10_7_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_14_13
#define init_10_7_14_13
void FASTCALL call_10_7_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_14_14
#define init_10_7_14_14
void FASTCALL call_10_7_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_14_15
#define init_10_7_14_15
void FASTCALL call_10_7_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_15_0
#define init_10_7_15_0
void FASTCALL call_10_7_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_15_1
#define init_10_7_15_1
void FASTCALL call_10_7_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_15_2
#define init_10_7_15_2
void FASTCALL call_10_7_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_15_3
#define init_10_7_15_3
void FASTCALL call_10_7_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_15_4
#define init_10_7_15_4
void FASTCALL call_10_7_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_15_5
#define init_10_7_15_5
void FASTCALL call_10_7_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_15_6
#define init_10_7_15_6
void FASTCALL call_10_7_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_15_7
#define init_10_7_15_7
void FASTCALL call_10_7_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_15_8
#define init_10_7_15_8
void FASTCALL call_10_7_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_15_9
#define init_10_7_15_9
void FASTCALL call_10_7_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_15_10
#define init_10_7_15_10
void FASTCALL call_10_7_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_15_11
#define init_10_7_15_11
void FASTCALL call_10_7_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_15_12
#define init_10_7_15_12
void FASTCALL call_10_7_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_15_13
#define init_10_7_15_13
void FASTCALL call_10_7_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_15_14
#define init_10_7_15_14
void FASTCALL call_10_7_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_7_15_15
#define init_10_7_15_15
void FASTCALL call_10_7_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_0_0
#define init_10_8_0_0
void FASTCALL call_10_8_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_0_1
#define init_10_8_0_1
void FASTCALL call_10_8_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_0_2
#define init_10_8_0_2
void FASTCALL call_10_8_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_0_3
#define init_10_8_0_3
void FASTCALL call_10_8_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_0_4
#define init_10_8_0_4
void FASTCALL call_10_8_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_0_5
#define init_10_8_0_5
void FASTCALL call_10_8_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_0_6
#define init_10_8_0_6
void FASTCALL call_10_8_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_0_7
#define init_10_8_0_7
void FASTCALL call_10_8_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_0_8
#define init_10_8_0_8
void FASTCALL call_10_8_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_0_9
#define init_10_8_0_9
void FASTCALL call_10_8_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_0_10
#define init_10_8_0_10
void FASTCALL call_10_8_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_0_11
#define init_10_8_0_11
void FASTCALL call_10_8_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_0_12
#define init_10_8_0_12
void FASTCALL call_10_8_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_0_13
#define init_10_8_0_13
void FASTCALL call_10_8_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_0_14
#define init_10_8_0_14
void FASTCALL call_10_8_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_0_15
#define init_10_8_0_15
void FASTCALL call_10_8_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_1_0
#define init_10_8_1_0
void FASTCALL call_10_8_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_1_1
#define init_10_8_1_1
void FASTCALL call_10_8_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_1_2
#define init_10_8_1_2
void FASTCALL call_10_8_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_1_3
#define init_10_8_1_3
void FASTCALL call_10_8_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_1_4
#define init_10_8_1_4
void FASTCALL call_10_8_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_1_5
#define init_10_8_1_5
void FASTCALL call_10_8_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_1_6
#define init_10_8_1_6
void FASTCALL call_10_8_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_1_7
#define init_10_8_1_7
void FASTCALL call_10_8_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_1_8
#define init_10_8_1_8
void FASTCALL call_10_8_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_1_9
#define init_10_8_1_9
void FASTCALL call_10_8_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_1_10
#define init_10_8_1_10
void FASTCALL call_10_8_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_1_11
#define init_10_8_1_11
void FASTCALL call_10_8_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_1_12
#define init_10_8_1_12
void FASTCALL call_10_8_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_1_13
#define init_10_8_1_13
void FASTCALL call_10_8_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_1_14
#define init_10_8_1_14
void FASTCALL call_10_8_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_1_15
#define init_10_8_1_15
void FASTCALL call_10_8_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_2_0
#define init_10_8_2_0
void FASTCALL call_10_8_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_2_1
#define init_10_8_2_1
void FASTCALL call_10_8_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_2_2
#define init_10_8_2_2
void FASTCALL call_10_8_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_2_3
#define init_10_8_2_3
void FASTCALL call_10_8_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_2_4
#define init_10_8_2_4
void FASTCALL call_10_8_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_2_5
#define init_10_8_2_5
void FASTCALL call_10_8_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_2_6
#define init_10_8_2_6
void FASTCALL call_10_8_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_2_7
#define init_10_8_2_7
void FASTCALL call_10_8_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_2_8
#define init_10_8_2_8
void FASTCALL call_10_8_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_2_9
#define init_10_8_2_9
void FASTCALL call_10_8_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_2_10
#define init_10_8_2_10
void FASTCALL call_10_8_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_2_11
#define init_10_8_2_11
void FASTCALL call_10_8_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_2_12
#define init_10_8_2_12
void FASTCALL call_10_8_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_2_13
#define init_10_8_2_13
void FASTCALL call_10_8_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_2_14
#define init_10_8_2_14
void FASTCALL call_10_8_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_2_15
#define init_10_8_2_15
void FASTCALL call_10_8_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_3_0
#define init_10_8_3_0
void FASTCALL call_10_8_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_3_1
#define init_10_8_3_1
void FASTCALL call_10_8_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_3_2
#define init_10_8_3_2
void FASTCALL call_10_8_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_3_3
#define init_10_8_3_3
void FASTCALL call_10_8_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_3_4
#define init_10_8_3_4
void FASTCALL call_10_8_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_3_5
#define init_10_8_3_5
void FASTCALL call_10_8_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_3_6
#define init_10_8_3_6
void FASTCALL call_10_8_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_3_7
#define init_10_8_3_7
void FASTCALL call_10_8_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_3_8
#define init_10_8_3_8
void FASTCALL call_10_8_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_3_9
#define init_10_8_3_9
void FASTCALL call_10_8_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_3_10
#define init_10_8_3_10
void FASTCALL call_10_8_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_3_11
#define init_10_8_3_11
void FASTCALL call_10_8_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_3_12
#define init_10_8_3_12
void FASTCALL call_10_8_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_3_13
#define init_10_8_3_13
void FASTCALL call_10_8_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_3_14
#define init_10_8_3_14
void FASTCALL call_10_8_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_3_15
#define init_10_8_3_15
void FASTCALL call_10_8_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_4_0
#define init_10_8_4_0
void FASTCALL call_10_8_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_4_1
#define init_10_8_4_1
void FASTCALL call_10_8_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_4_2
#define init_10_8_4_2
void FASTCALL call_10_8_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_4_3
#define init_10_8_4_3
void FASTCALL call_10_8_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_4_4
#define init_10_8_4_4
void FASTCALL call_10_8_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_4_5
#define init_10_8_4_5
void FASTCALL call_10_8_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_4_6
#define init_10_8_4_6
void FASTCALL call_10_8_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_4_7
#define init_10_8_4_7
void FASTCALL call_10_8_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_4_8
#define init_10_8_4_8
void FASTCALL call_10_8_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_4_9
#define init_10_8_4_9
void FASTCALL call_10_8_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_4_10
#define init_10_8_4_10
void FASTCALL call_10_8_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_4_11
#define init_10_8_4_11
void FASTCALL call_10_8_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_4_12
#define init_10_8_4_12
void FASTCALL call_10_8_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_4_13
#define init_10_8_4_13
void FASTCALL call_10_8_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_4_14
#define init_10_8_4_14
void FASTCALL call_10_8_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_4_15
#define init_10_8_4_15
void FASTCALL call_10_8_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_5_0
#define init_10_8_5_0
void FASTCALL call_10_8_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_5_1
#define init_10_8_5_1
void FASTCALL call_10_8_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_5_2
#define init_10_8_5_2
void FASTCALL call_10_8_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_5_3
#define init_10_8_5_3
void FASTCALL call_10_8_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_5_4
#define init_10_8_5_4
void FASTCALL call_10_8_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_5_5
#define init_10_8_5_5
void FASTCALL call_10_8_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_5_6
#define init_10_8_5_6
void FASTCALL call_10_8_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_5_7
#define init_10_8_5_7
void FASTCALL call_10_8_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_5_8
#define init_10_8_5_8
void FASTCALL call_10_8_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_5_9
#define init_10_8_5_9
void FASTCALL call_10_8_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_5_10
#define init_10_8_5_10
void FASTCALL call_10_8_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_5_11
#define init_10_8_5_11
void FASTCALL call_10_8_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_5_12
#define init_10_8_5_12
void FASTCALL call_10_8_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_5_13
#define init_10_8_5_13
void FASTCALL call_10_8_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_5_14
#define init_10_8_5_14
void FASTCALL call_10_8_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_5_15
#define init_10_8_5_15
void FASTCALL call_10_8_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_6_0
#define init_10_8_6_0
void FASTCALL call_10_8_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_6_1
#define init_10_8_6_1
void FASTCALL call_10_8_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_6_2
#define init_10_8_6_2
void FASTCALL call_10_8_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_6_3
#define init_10_8_6_3
void FASTCALL call_10_8_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_6_4
#define init_10_8_6_4
void FASTCALL call_10_8_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_6_5
#define init_10_8_6_5
void FASTCALL call_10_8_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_6_6
#define init_10_8_6_6
void FASTCALL call_10_8_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_6_7
#define init_10_8_6_7
void FASTCALL call_10_8_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_6_8
#define init_10_8_6_8
void FASTCALL call_10_8_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_6_9
#define init_10_8_6_9
void FASTCALL call_10_8_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_6_10
#define init_10_8_6_10
void FASTCALL call_10_8_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_6_11
#define init_10_8_6_11
void FASTCALL call_10_8_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_6_12
#define init_10_8_6_12
void FASTCALL call_10_8_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_6_13
#define init_10_8_6_13
void FASTCALL call_10_8_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_6_14
#define init_10_8_6_14
void FASTCALL call_10_8_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_6_15
#define init_10_8_6_15
void FASTCALL call_10_8_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_7_0
#define init_10_8_7_0
void FASTCALL call_10_8_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_7_1
#define init_10_8_7_1
void FASTCALL call_10_8_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_7_2
#define init_10_8_7_2
void FASTCALL call_10_8_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_7_3
#define init_10_8_7_3
void FASTCALL call_10_8_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_7_4
#define init_10_8_7_4
void FASTCALL call_10_8_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_7_5
#define init_10_8_7_5
void FASTCALL call_10_8_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_7_6
#define init_10_8_7_6
void FASTCALL call_10_8_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_7_7
#define init_10_8_7_7
void FASTCALL call_10_8_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_7_8
#define init_10_8_7_8
void FASTCALL call_10_8_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_7_9
#define init_10_8_7_9
void FASTCALL call_10_8_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_7_10
#define init_10_8_7_10
void FASTCALL call_10_8_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_7_11
#define init_10_8_7_11
void FASTCALL call_10_8_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_7_12
#define init_10_8_7_12
void FASTCALL call_10_8_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_7_13
#define init_10_8_7_13
void FASTCALL call_10_8_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_7_14
#define init_10_8_7_14
void FASTCALL call_10_8_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_7_15
#define init_10_8_7_15
void FASTCALL call_10_8_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_8_0
#define init_10_8_8_0
void FASTCALL call_10_8_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_8_1
#define init_10_8_8_1
void FASTCALL call_10_8_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_8_2
#define init_10_8_8_2
void FASTCALL call_10_8_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_8_3
#define init_10_8_8_3
void FASTCALL call_10_8_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_8_4
#define init_10_8_8_4
void FASTCALL call_10_8_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_8_5
#define init_10_8_8_5
void FASTCALL call_10_8_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_8_6
#define init_10_8_8_6
void FASTCALL call_10_8_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_8_7
#define init_10_8_8_7
void FASTCALL call_10_8_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_8_8
#define init_10_8_8_8
void FASTCALL call_10_8_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_8_9
#define init_10_8_8_9
void FASTCALL call_10_8_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_8_10
#define init_10_8_8_10
void FASTCALL call_10_8_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_8_11
#define init_10_8_8_11
void FASTCALL call_10_8_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_8_12
#define init_10_8_8_12
void FASTCALL call_10_8_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_8_13
#define init_10_8_8_13
void FASTCALL call_10_8_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_8_14
#define init_10_8_8_14
void FASTCALL call_10_8_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_8_15
#define init_10_8_8_15
void FASTCALL call_10_8_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_9_0
#define init_10_8_9_0
void FASTCALL call_10_8_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_9_1
#define init_10_8_9_1
void FASTCALL call_10_8_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_9_2
#define init_10_8_9_2
void FASTCALL call_10_8_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_9_3
#define init_10_8_9_3
void FASTCALL call_10_8_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_9_4
#define init_10_8_9_4
void FASTCALL call_10_8_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_9_5
#define init_10_8_9_5
void FASTCALL call_10_8_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_9_6
#define init_10_8_9_6
void FASTCALL call_10_8_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_9_7
#define init_10_8_9_7
void FASTCALL call_10_8_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_9_8
#define init_10_8_9_8
void FASTCALL call_10_8_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_9_9
#define init_10_8_9_9
void FASTCALL call_10_8_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_9_10
#define init_10_8_9_10
void FASTCALL call_10_8_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_9_11
#define init_10_8_9_11
void FASTCALL call_10_8_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_9_12
#define init_10_8_9_12
void FASTCALL call_10_8_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_9_13
#define init_10_8_9_13
void FASTCALL call_10_8_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_9_14
#define init_10_8_9_14
void FASTCALL call_10_8_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_9_15
#define init_10_8_9_15
void FASTCALL call_10_8_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_10_0
#define init_10_8_10_0
void FASTCALL call_10_8_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_10_1
#define init_10_8_10_1
void FASTCALL call_10_8_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_10_2
#define init_10_8_10_2
void FASTCALL call_10_8_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_10_3
#define init_10_8_10_3
void FASTCALL call_10_8_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_10_4
#define init_10_8_10_4
void FASTCALL call_10_8_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_10_5
#define init_10_8_10_5
void FASTCALL call_10_8_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_10_6
#define init_10_8_10_6
void FASTCALL call_10_8_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_10_7
#define init_10_8_10_7
void FASTCALL call_10_8_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_10_8
#define init_10_8_10_8
void FASTCALL call_10_8_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_10_9
#define init_10_8_10_9
void FASTCALL call_10_8_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_10_10
#define init_10_8_10_10
void FASTCALL call_10_8_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_10_11
#define init_10_8_10_11
void FASTCALL call_10_8_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_10_12
#define init_10_8_10_12
void FASTCALL call_10_8_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_10_13
#define init_10_8_10_13
void FASTCALL call_10_8_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_10_14
#define init_10_8_10_14
void FASTCALL call_10_8_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_10_15
#define init_10_8_10_15
void FASTCALL call_10_8_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_11_0
#define init_10_8_11_0
void FASTCALL call_10_8_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_11_1
#define init_10_8_11_1
void FASTCALL call_10_8_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_11_2
#define init_10_8_11_2
void FASTCALL call_10_8_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_11_3
#define init_10_8_11_3
void FASTCALL call_10_8_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_11_4
#define init_10_8_11_4
void FASTCALL call_10_8_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_11_5
#define init_10_8_11_5
void FASTCALL call_10_8_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_11_6
#define init_10_8_11_6
void FASTCALL call_10_8_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_11_7
#define init_10_8_11_7
void FASTCALL call_10_8_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_11_8
#define init_10_8_11_8
void FASTCALL call_10_8_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_11_9
#define init_10_8_11_9
void FASTCALL call_10_8_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_11_10
#define init_10_8_11_10
void FASTCALL call_10_8_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_11_11
#define init_10_8_11_11
void FASTCALL call_10_8_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_11_12
#define init_10_8_11_12
void FASTCALL call_10_8_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_11_13
#define init_10_8_11_13
void FASTCALL call_10_8_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_11_14
#define init_10_8_11_14
void FASTCALL call_10_8_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_11_15
#define init_10_8_11_15
void FASTCALL call_10_8_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_12_0
#define init_10_8_12_0
void FASTCALL call_10_8_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_12_1
#define init_10_8_12_1
void FASTCALL call_10_8_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_12_2
#define init_10_8_12_2
void FASTCALL call_10_8_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_12_3
#define init_10_8_12_3
void FASTCALL call_10_8_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_12_4
#define init_10_8_12_4
void FASTCALL call_10_8_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_12_5
#define init_10_8_12_5
void FASTCALL call_10_8_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_12_6
#define init_10_8_12_6
void FASTCALL call_10_8_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_12_7
#define init_10_8_12_7
void FASTCALL call_10_8_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_12_8
#define init_10_8_12_8
void FASTCALL call_10_8_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_12_9
#define init_10_8_12_9
void FASTCALL call_10_8_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_12_10
#define init_10_8_12_10
void FASTCALL call_10_8_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_12_11
#define init_10_8_12_11
void FASTCALL call_10_8_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_12_12
#define init_10_8_12_12
void FASTCALL call_10_8_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_12_13
#define init_10_8_12_13
void FASTCALL call_10_8_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_12_14
#define init_10_8_12_14
void FASTCALL call_10_8_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_12_15
#define init_10_8_12_15
void FASTCALL call_10_8_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_13_0
#define init_10_8_13_0
void FASTCALL call_10_8_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_13_1
#define init_10_8_13_1
void FASTCALL call_10_8_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_13_2
#define init_10_8_13_2
void FASTCALL call_10_8_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_13_3
#define init_10_8_13_3
void FASTCALL call_10_8_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_13_4
#define init_10_8_13_4
void FASTCALL call_10_8_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_13_5
#define init_10_8_13_5
void FASTCALL call_10_8_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_13_6
#define init_10_8_13_6
void FASTCALL call_10_8_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_13_7
#define init_10_8_13_7
void FASTCALL call_10_8_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_13_8
#define init_10_8_13_8
void FASTCALL call_10_8_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_13_9
#define init_10_8_13_9
void FASTCALL call_10_8_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_13_10
#define init_10_8_13_10
void FASTCALL call_10_8_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_13_11
#define init_10_8_13_11
void FASTCALL call_10_8_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_13_12
#define init_10_8_13_12
void FASTCALL call_10_8_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_13_13
#define init_10_8_13_13
void FASTCALL call_10_8_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_13_14
#define init_10_8_13_14
void FASTCALL call_10_8_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_13_15
#define init_10_8_13_15
void FASTCALL call_10_8_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_14_0
#define init_10_8_14_0
void FASTCALL call_10_8_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_14_1
#define init_10_8_14_1
void FASTCALL call_10_8_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_14_2
#define init_10_8_14_2
void FASTCALL call_10_8_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_14_3
#define init_10_8_14_3
void FASTCALL call_10_8_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_14_4
#define init_10_8_14_4
void FASTCALL call_10_8_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_14_5
#define init_10_8_14_5
void FASTCALL call_10_8_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_14_6
#define init_10_8_14_6
void FASTCALL call_10_8_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_14_7
#define init_10_8_14_7
void FASTCALL call_10_8_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_14_8
#define init_10_8_14_8
void FASTCALL call_10_8_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_14_9
#define init_10_8_14_9
void FASTCALL call_10_8_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_14_10
#define init_10_8_14_10
void FASTCALL call_10_8_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_14_11
#define init_10_8_14_11
void FASTCALL call_10_8_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_14_12
#define init_10_8_14_12
void FASTCALL call_10_8_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_14_13
#define init_10_8_14_13
void FASTCALL call_10_8_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_14_14
#define init_10_8_14_14
void FASTCALL call_10_8_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_14_15
#define init_10_8_14_15
void FASTCALL call_10_8_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_15_0
#define init_10_8_15_0
void FASTCALL call_10_8_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_15_1
#define init_10_8_15_1
void FASTCALL call_10_8_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_15_2
#define init_10_8_15_2
void FASTCALL call_10_8_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_15_3
#define init_10_8_15_3
void FASTCALL call_10_8_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_15_4
#define init_10_8_15_4
void FASTCALL call_10_8_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_15_5
#define init_10_8_15_5
void FASTCALL call_10_8_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_15_6
#define init_10_8_15_6
void FASTCALL call_10_8_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_15_7
#define init_10_8_15_7
void FASTCALL call_10_8_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_15_8
#define init_10_8_15_8
void FASTCALL call_10_8_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_15_9
#define init_10_8_15_9
void FASTCALL call_10_8_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_15_10
#define init_10_8_15_10
void FASTCALL call_10_8_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_15_11
#define init_10_8_15_11
void FASTCALL call_10_8_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_15_12
#define init_10_8_15_12
void FASTCALL call_10_8_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_15_13
#define init_10_8_15_13
void FASTCALL call_10_8_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_15_14
#define init_10_8_15_14
void FASTCALL call_10_8_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_8_15_15
#define init_10_8_15_15
void FASTCALL call_10_8_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_0_0
#define init_10_9_0_0
void FASTCALL call_10_9_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_0_1
#define init_10_9_0_1
void FASTCALL call_10_9_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_0_2
#define init_10_9_0_2
void FASTCALL call_10_9_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_0_3
#define init_10_9_0_3
void FASTCALL call_10_9_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_0_4
#define init_10_9_0_4
void FASTCALL call_10_9_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_0_5
#define init_10_9_0_5
void FASTCALL call_10_9_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_0_6
#define init_10_9_0_6
void FASTCALL call_10_9_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_0_7
#define init_10_9_0_7
void FASTCALL call_10_9_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_0_8
#define init_10_9_0_8
void FASTCALL call_10_9_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_0_9
#define init_10_9_0_9
void FASTCALL call_10_9_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_0_10
#define init_10_9_0_10
void FASTCALL call_10_9_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_0_11
#define init_10_9_0_11
void FASTCALL call_10_9_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_0_12
#define init_10_9_0_12
void FASTCALL call_10_9_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_0_13
#define init_10_9_0_13
void FASTCALL call_10_9_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_0_14
#define init_10_9_0_14
void FASTCALL call_10_9_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_0_15
#define init_10_9_0_15
void FASTCALL call_10_9_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_1_0
#define init_10_9_1_0
void FASTCALL call_10_9_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_1_1
#define init_10_9_1_1
void FASTCALL call_10_9_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_1_2
#define init_10_9_1_2
void FASTCALL call_10_9_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_1_3
#define init_10_9_1_3
void FASTCALL call_10_9_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_1_4
#define init_10_9_1_4
void FASTCALL call_10_9_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_1_5
#define init_10_9_1_5
void FASTCALL call_10_9_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_1_6
#define init_10_9_1_6
void FASTCALL call_10_9_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_1_7
#define init_10_9_1_7
void FASTCALL call_10_9_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_1_8
#define init_10_9_1_8
void FASTCALL call_10_9_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_1_9
#define init_10_9_1_9
void FASTCALL call_10_9_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_1_10
#define init_10_9_1_10
void FASTCALL call_10_9_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_1_11
#define init_10_9_1_11
void FASTCALL call_10_9_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_1_12
#define init_10_9_1_12
void FASTCALL call_10_9_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_1_13
#define init_10_9_1_13
void FASTCALL call_10_9_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_1_14
#define init_10_9_1_14
void FASTCALL call_10_9_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_1_15
#define init_10_9_1_15
void FASTCALL call_10_9_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_2_0
#define init_10_9_2_0
void FASTCALL call_10_9_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_2_1
#define init_10_9_2_1
void FASTCALL call_10_9_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_2_2
#define init_10_9_2_2
void FASTCALL call_10_9_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_2_3
#define init_10_9_2_3
void FASTCALL call_10_9_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_2_4
#define init_10_9_2_4
void FASTCALL call_10_9_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_2_5
#define init_10_9_2_5
void FASTCALL call_10_9_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_2_6
#define init_10_9_2_6
void FASTCALL call_10_9_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_2_7
#define init_10_9_2_7
void FASTCALL call_10_9_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_2_8
#define init_10_9_2_8
void FASTCALL call_10_9_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_2_9
#define init_10_9_2_9
void FASTCALL call_10_9_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_2_10
#define init_10_9_2_10
void FASTCALL call_10_9_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_2_11
#define init_10_9_2_11
void FASTCALL call_10_9_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_2_12
#define init_10_9_2_12
void FASTCALL call_10_9_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_2_13
#define init_10_9_2_13
void FASTCALL call_10_9_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_2_14
#define init_10_9_2_14
void FASTCALL call_10_9_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_2_15
#define init_10_9_2_15
void FASTCALL call_10_9_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_3_0
#define init_10_9_3_0
void FASTCALL call_10_9_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_3_1
#define init_10_9_3_1
void FASTCALL call_10_9_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_3_2
#define init_10_9_3_2
void FASTCALL call_10_9_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_3_3
#define init_10_9_3_3
void FASTCALL call_10_9_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_3_4
#define init_10_9_3_4
void FASTCALL call_10_9_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_3_5
#define init_10_9_3_5
void FASTCALL call_10_9_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_3_6
#define init_10_9_3_6
void FASTCALL call_10_9_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_3_7
#define init_10_9_3_7
void FASTCALL call_10_9_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_3_8
#define init_10_9_3_8
void FASTCALL call_10_9_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_3_9
#define init_10_9_3_9
void FASTCALL call_10_9_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_3_10
#define init_10_9_3_10
void FASTCALL call_10_9_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_3_11
#define init_10_9_3_11
void FASTCALL call_10_9_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_3_12
#define init_10_9_3_12
void FASTCALL call_10_9_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_3_13
#define init_10_9_3_13
void FASTCALL call_10_9_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_3_14
#define init_10_9_3_14
void FASTCALL call_10_9_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_3_15
#define init_10_9_3_15
void FASTCALL call_10_9_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_4_0
#define init_10_9_4_0
void FASTCALL call_10_9_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_4_1
#define init_10_9_4_1
void FASTCALL call_10_9_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_4_2
#define init_10_9_4_2
void FASTCALL call_10_9_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_4_3
#define init_10_9_4_3
void FASTCALL call_10_9_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_4_4
#define init_10_9_4_4
void FASTCALL call_10_9_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_4_5
#define init_10_9_4_5
void FASTCALL call_10_9_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_4_6
#define init_10_9_4_6
void FASTCALL call_10_9_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_4_7
#define init_10_9_4_7
void FASTCALL call_10_9_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_4_8
#define init_10_9_4_8
void FASTCALL call_10_9_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_4_9
#define init_10_9_4_9
void FASTCALL call_10_9_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_4_10
#define init_10_9_4_10
void FASTCALL call_10_9_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_4_11
#define init_10_9_4_11
void FASTCALL call_10_9_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_4_12
#define init_10_9_4_12
void FASTCALL call_10_9_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_4_13
#define init_10_9_4_13
void FASTCALL call_10_9_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_4_14
#define init_10_9_4_14
void FASTCALL call_10_9_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_4_15
#define init_10_9_4_15
void FASTCALL call_10_9_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_5_0
#define init_10_9_5_0
void FASTCALL call_10_9_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_5_1
#define init_10_9_5_1
void FASTCALL call_10_9_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_5_2
#define init_10_9_5_2
void FASTCALL call_10_9_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_5_3
#define init_10_9_5_3
void FASTCALL call_10_9_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_5_4
#define init_10_9_5_4
void FASTCALL call_10_9_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_5_5
#define init_10_9_5_5
void FASTCALL call_10_9_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_5_6
#define init_10_9_5_6
void FASTCALL call_10_9_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_5_7
#define init_10_9_5_7
void FASTCALL call_10_9_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_5_8
#define init_10_9_5_8
void FASTCALL call_10_9_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_5_9
#define init_10_9_5_9
void FASTCALL call_10_9_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_5_10
#define init_10_9_5_10
void FASTCALL call_10_9_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_5_11
#define init_10_9_5_11
void FASTCALL call_10_9_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_5_12
#define init_10_9_5_12
void FASTCALL call_10_9_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_5_13
#define init_10_9_5_13
void FASTCALL call_10_9_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_5_14
#define init_10_9_5_14
void FASTCALL call_10_9_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_5_15
#define init_10_9_5_15
void FASTCALL call_10_9_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_6_0
#define init_10_9_6_0
void FASTCALL call_10_9_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_6_1
#define init_10_9_6_1
void FASTCALL call_10_9_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_6_2
#define init_10_9_6_2
void FASTCALL call_10_9_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_6_3
#define init_10_9_6_3
void FASTCALL call_10_9_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_6_4
#define init_10_9_6_4
void FASTCALL call_10_9_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_6_5
#define init_10_9_6_5
void FASTCALL call_10_9_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_6_6
#define init_10_9_6_6
void FASTCALL call_10_9_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_6_7
#define init_10_9_6_7
void FASTCALL call_10_9_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_6_8
#define init_10_9_6_8
void FASTCALL call_10_9_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_6_9
#define init_10_9_6_9
void FASTCALL call_10_9_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_6_10
#define init_10_9_6_10
void FASTCALL call_10_9_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_6_11
#define init_10_9_6_11
void FASTCALL call_10_9_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_6_12
#define init_10_9_6_12
void FASTCALL call_10_9_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_6_13
#define init_10_9_6_13
void FASTCALL call_10_9_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_6_14
#define init_10_9_6_14
void FASTCALL call_10_9_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_6_15
#define init_10_9_6_15
void FASTCALL call_10_9_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_7_0
#define init_10_9_7_0
void FASTCALL call_10_9_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_7_1
#define init_10_9_7_1
void FASTCALL call_10_9_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_7_2
#define init_10_9_7_2
void FASTCALL call_10_9_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_7_3
#define init_10_9_7_3
void FASTCALL call_10_9_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_7_4
#define init_10_9_7_4
void FASTCALL call_10_9_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_7_5
#define init_10_9_7_5
void FASTCALL call_10_9_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_7_6
#define init_10_9_7_6
void FASTCALL call_10_9_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_7_7
#define init_10_9_7_7
void FASTCALL call_10_9_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_7_8
#define init_10_9_7_8
void FASTCALL call_10_9_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_7_9
#define init_10_9_7_9
void FASTCALL call_10_9_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_7_10
#define init_10_9_7_10
void FASTCALL call_10_9_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_7_11
#define init_10_9_7_11
void FASTCALL call_10_9_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_7_12
#define init_10_9_7_12
void FASTCALL call_10_9_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_7_13
#define init_10_9_7_13
void FASTCALL call_10_9_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_7_14
#define init_10_9_7_14
void FASTCALL call_10_9_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_7_15
#define init_10_9_7_15
void FASTCALL call_10_9_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_8_0
#define init_10_9_8_0
void FASTCALL call_10_9_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_8_1
#define init_10_9_8_1
void FASTCALL call_10_9_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_8_2
#define init_10_9_8_2
void FASTCALL call_10_9_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_8_3
#define init_10_9_8_3
void FASTCALL call_10_9_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_8_4
#define init_10_9_8_4
void FASTCALL call_10_9_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_8_5
#define init_10_9_8_5
void FASTCALL call_10_9_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_8_6
#define init_10_9_8_6
void FASTCALL call_10_9_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_8_7
#define init_10_9_8_7
void FASTCALL call_10_9_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_8_8
#define init_10_9_8_8
void FASTCALL call_10_9_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_8_9
#define init_10_9_8_9
void FASTCALL call_10_9_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_8_10
#define init_10_9_8_10
void FASTCALL call_10_9_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_8_11
#define init_10_9_8_11
void FASTCALL call_10_9_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_8_12
#define init_10_9_8_12
void FASTCALL call_10_9_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_8_13
#define init_10_9_8_13
void FASTCALL call_10_9_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_8_14
#define init_10_9_8_14
void FASTCALL call_10_9_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_8_15
#define init_10_9_8_15
void FASTCALL call_10_9_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_9_0
#define init_10_9_9_0
void FASTCALL call_10_9_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_9_1
#define init_10_9_9_1
void FASTCALL call_10_9_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_9_2
#define init_10_9_9_2
void FASTCALL call_10_9_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_9_3
#define init_10_9_9_3
void FASTCALL call_10_9_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_9_4
#define init_10_9_9_4
void FASTCALL call_10_9_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_9_5
#define init_10_9_9_5
void FASTCALL call_10_9_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_9_6
#define init_10_9_9_6
void FASTCALL call_10_9_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_9_7
#define init_10_9_9_7
void FASTCALL call_10_9_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_9_8
#define init_10_9_9_8
void FASTCALL call_10_9_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_9_9
#define init_10_9_9_9
void FASTCALL call_10_9_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_9_10
#define init_10_9_9_10
void FASTCALL call_10_9_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_9_11
#define init_10_9_9_11
void FASTCALL call_10_9_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_9_12
#define init_10_9_9_12
void FASTCALL call_10_9_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_9_13
#define init_10_9_9_13
void FASTCALL call_10_9_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_9_14
#define init_10_9_9_14
void FASTCALL call_10_9_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_9_15
#define init_10_9_9_15
void FASTCALL call_10_9_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_10_0
#define init_10_9_10_0
void FASTCALL call_10_9_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_10_1
#define init_10_9_10_1
void FASTCALL call_10_9_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_10_2
#define init_10_9_10_2
void FASTCALL call_10_9_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_10_3
#define init_10_9_10_3
void FASTCALL call_10_9_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_10_4
#define init_10_9_10_4
void FASTCALL call_10_9_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_10_5
#define init_10_9_10_5
void FASTCALL call_10_9_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_10_6
#define init_10_9_10_6
void FASTCALL call_10_9_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_10_7
#define init_10_9_10_7
void FASTCALL call_10_9_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_10_8
#define init_10_9_10_8
void FASTCALL call_10_9_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_10_9
#define init_10_9_10_9
void FASTCALL call_10_9_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_10_10
#define init_10_9_10_10
void FASTCALL call_10_9_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_10_11
#define init_10_9_10_11
void FASTCALL call_10_9_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_10_12
#define init_10_9_10_12
void FASTCALL call_10_9_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_10_13
#define init_10_9_10_13
void FASTCALL call_10_9_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_10_14
#define init_10_9_10_14
void FASTCALL call_10_9_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_10_15
#define init_10_9_10_15
void FASTCALL call_10_9_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_11_0
#define init_10_9_11_0
void FASTCALL call_10_9_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_11_1
#define init_10_9_11_1
void FASTCALL call_10_9_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_11_2
#define init_10_9_11_2
void FASTCALL call_10_9_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_11_3
#define init_10_9_11_3
void FASTCALL call_10_9_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_11_4
#define init_10_9_11_4
void FASTCALL call_10_9_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_11_5
#define init_10_9_11_5
void FASTCALL call_10_9_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_11_6
#define init_10_9_11_6
void FASTCALL call_10_9_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_11_7
#define init_10_9_11_7
void FASTCALL call_10_9_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_11_8
#define init_10_9_11_8
void FASTCALL call_10_9_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_11_9
#define init_10_9_11_9
void FASTCALL call_10_9_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_11_10
#define init_10_9_11_10
void FASTCALL call_10_9_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_11_11
#define init_10_9_11_11
void FASTCALL call_10_9_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_11_12
#define init_10_9_11_12
void FASTCALL call_10_9_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_11_13
#define init_10_9_11_13
void FASTCALL call_10_9_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_11_14
#define init_10_9_11_14
void FASTCALL call_10_9_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_11_15
#define init_10_9_11_15
void FASTCALL call_10_9_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_12_0
#define init_10_9_12_0
void FASTCALL call_10_9_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_12_1
#define init_10_9_12_1
void FASTCALL call_10_9_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_12_2
#define init_10_9_12_2
void FASTCALL call_10_9_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_12_3
#define init_10_9_12_3
void FASTCALL call_10_9_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_12_4
#define init_10_9_12_4
void FASTCALL call_10_9_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_12_5
#define init_10_9_12_5
void FASTCALL call_10_9_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_12_6
#define init_10_9_12_6
void FASTCALL call_10_9_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_12_7
#define init_10_9_12_7
void FASTCALL call_10_9_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_12_8
#define init_10_9_12_8
void FASTCALL call_10_9_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_12_9
#define init_10_9_12_9
void FASTCALL call_10_9_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_12_10
#define init_10_9_12_10
void FASTCALL call_10_9_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_12_11
#define init_10_9_12_11
void FASTCALL call_10_9_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_12_12
#define init_10_9_12_12
void FASTCALL call_10_9_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_12_13
#define init_10_9_12_13
void FASTCALL call_10_9_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_12_14
#define init_10_9_12_14
void FASTCALL call_10_9_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_12_15
#define init_10_9_12_15
void FASTCALL call_10_9_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_13_0
#define init_10_9_13_0
void FASTCALL call_10_9_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_13_1
#define init_10_9_13_1
void FASTCALL call_10_9_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_13_2
#define init_10_9_13_2
void FASTCALL call_10_9_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_13_3
#define init_10_9_13_3
void FASTCALL call_10_9_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_13_4
#define init_10_9_13_4
void FASTCALL call_10_9_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_13_5
#define init_10_9_13_5
void FASTCALL call_10_9_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_13_6
#define init_10_9_13_6
void FASTCALL call_10_9_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_13_7
#define init_10_9_13_7
void FASTCALL call_10_9_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_13_8
#define init_10_9_13_8
void FASTCALL call_10_9_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_13_9
#define init_10_9_13_9
void FASTCALL call_10_9_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_13_10
#define init_10_9_13_10
void FASTCALL call_10_9_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_13_11
#define init_10_9_13_11
void FASTCALL call_10_9_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_13_12
#define init_10_9_13_12
void FASTCALL call_10_9_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_13_13
#define init_10_9_13_13
void FASTCALL call_10_9_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_13_14
#define init_10_9_13_14
void FASTCALL call_10_9_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_13_15
#define init_10_9_13_15
void FASTCALL call_10_9_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_14_0
#define init_10_9_14_0
void FASTCALL call_10_9_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_14_1
#define init_10_9_14_1
void FASTCALL call_10_9_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_14_2
#define init_10_9_14_2
void FASTCALL call_10_9_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_14_3
#define init_10_9_14_3
void FASTCALL call_10_9_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_14_4
#define init_10_9_14_4
void FASTCALL call_10_9_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_14_5
#define init_10_9_14_5
void FASTCALL call_10_9_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_14_6
#define init_10_9_14_6
void FASTCALL call_10_9_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_14_7
#define init_10_9_14_7
void FASTCALL call_10_9_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_14_8
#define init_10_9_14_8
void FASTCALL call_10_9_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_14_9
#define init_10_9_14_9
void FASTCALL call_10_9_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_14_10
#define init_10_9_14_10
void FASTCALL call_10_9_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_14_11
#define init_10_9_14_11
void FASTCALL call_10_9_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_14_12
#define init_10_9_14_12
void FASTCALL call_10_9_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_14_13
#define init_10_9_14_13
void FASTCALL call_10_9_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_14_14
#define init_10_9_14_14
void FASTCALL call_10_9_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_14_15
#define init_10_9_14_15
void FASTCALL call_10_9_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_15_0
#define init_10_9_15_0
void FASTCALL call_10_9_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_15_1
#define init_10_9_15_1
void FASTCALL call_10_9_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_15_2
#define init_10_9_15_2
void FASTCALL call_10_9_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_15_3
#define init_10_9_15_3
void FASTCALL call_10_9_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_15_4
#define init_10_9_15_4
void FASTCALL call_10_9_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_15_5
#define init_10_9_15_5
void FASTCALL call_10_9_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_15_6
#define init_10_9_15_6
void FASTCALL call_10_9_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_15_7
#define init_10_9_15_7
void FASTCALL call_10_9_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_15_8
#define init_10_9_15_8
void FASTCALL call_10_9_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_15_9
#define init_10_9_15_9
void FASTCALL call_10_9_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_15_10
#define init_10_9_15_10
void FASTCALL call_10_9_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_15_11
#define init_10_9_15_11
void FASTCALL call_10_9_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_15_12
#define init_10_9_15_12
void FASTCALL call_10_9_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_15_13
#define init_10_9_15_13
void FASTCALL call_10_9_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_15_14
#define init_10_9_15_14
void FASTCALL call_10_9_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_9_15_15
#define init_10_9_15_15
void FASTCALL call_10_9_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_0_0
#define init_10_10_0_0
void FASTCALL call_10_10_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_0_1
#define init_10_10_0_1
void FASTCALL call_10_10_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_0_2
#define init_10_10_0_2
void FASTCALL call_10_10_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_0_3
#define init_10_10_0_3
void FASTCALL call_10_10_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_0_4
#define init_10_10_0_4
void FASTCALL call_10_10_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_0_5
#define init_10_10_0_5
void FASTCALL call_10_10_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_0_6
#define init_10_10_0_6
void FASTCALL call_10_10_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_0_7
#define init_10_10_0_7
void FASTCALL call_10_10_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_0_8
#define init_10_10_0_8
void FASTCALL call_10_10_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_0_9
#define init_10_10_0_9
void FASTCALL call_10_10_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_0_10
#define init_10_10_0_10
void FASTCALL call_10_10_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_0_11
#define init_10_10_0_11
void FASTCALL call_10_10_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_0_12
#define init_10_10_0_12
void FASTCALL call_10_10_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_0_13
#define init_10_10_0_13
void FASTCALL call_10_10_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_0_14
#define init_10_10_0_14
void FASTCALL call_10_10_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_0_15
#define init_10_10_0_15
void FASTCALL call_10_10_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_1_0
#define init_10_10_1_0
void FASTCALL call_10_10_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_1_1
#define init_10_10_1_1
void FASTCALL call_10_10_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_1_2
#define init_10_10_1_2
void FASTCALL call_10_10_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_1_3
#define init_10_10_1_3
void FASTCALL call_10_10_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_1_4
#define init_10_10_1_4
void FASTCALL call_10_10_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_1_5
#define init_10_10_1_5
void FASTCALL call_10_10_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_1_6
#define init_10_10_1_6
void FASTCALL call_10_10_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_1_7
#define init_10_10_1_7
void FASTCALL call_10_10_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_1_8
#define init_10_10_1_8
void FASTCALL call_10_10_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_1_9
#define init_10_10_1_9
void FASTCALL call_10_10_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_1_10
#define init_10_10_1_10
void FASTCALL call_10_10_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_1_11
#define init_10_10_1_11
void FASTCALL call_10_10_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_1_12
#define init_10_10_1_12
void FASTCALL call_10_10_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_1_13
#define init_10_10_1_13
void FASTCALL call_10_10_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_1_14
#define init_10_10_1_14
void FASTCALL call_10_10_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_1_15
#define init_10_10_1_15
void FASTCALL call_10_10_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_2_0
#define init_10_10_2_0
void FASTCALL call_10_10_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_2_1
#define init_10_10_2_1
void FASTCALL call_10_10_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_2_2
#define init_10_10_2_2
void FASTCALL call_10_10_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_2_3
#define init_10_10_2_3
void FASTCALL call_10_10_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_2_4
#define init_10_10_2_4
void FASTCALL call_10_10_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_2_5
#define init_10_10_2_5
void FASTCALL call_10_10_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_2_6
#define init_10_10_2_6
void FASTCALL call_10_10_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_2_7
#define init_10_10_2_7
void FASTCALL call_10_10_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_2_8
#define init_10_10_2_8
void FASTCALL call_10_10_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_2_9
#define init_10_10_2_9
void FASTCALL call_10_10_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_2_10
#define init_10_10_2_10
void FASTCALL call_10_10_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_2_11
#define init_10_10_2_11
void FASTCALL call_10_10_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_2_12
#define init_10_10_2_12
void FASTCALL call_10_10_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_2_13
#define init_10_10_2_13
void FASTCALL call_10_10_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_2_14
#define init_10_10_2_14
void FASTCALL call_10_10_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_2_15
#define init_10_10_2_15
void FASTCALL call_10_10_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_3_0
#define init_10_10_3_0
void FASTCALL call_10_10_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_3_1
#define init_10_10_3_1
void FASTCALL call_10_10_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_3_2
#define init_10_10_3_2
void FASTCALL call_10_10_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_3_3
#define init_10_10_3_3
void FASTCALL call_10_10_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_3_4
#define init_10_10_3_4
void FASTCALL call_10_10_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_3_5
#define init_10_10_3_5
void FASTCALL call_10_10_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_3_6
#define init_10_10_3_6
void FASTCALL call_10_10_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_3_7
#define init_10_10_3_7
void FASTCALL call_10_10_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_3_8
#define init_10_10_3_8
void FASTCALL call_10_10_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_3_9
#define init_10_10_3_9
void FASTCALL call_10_10_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_3_10
#define init_10_10_3_10
void FASTCALL call_10_10_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_3_11
#define init_10_10_3_11
void FASTCALL call_10_10_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_3_12
#define init_10_10_3_12
void FASTCALL call_10_10_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_3_13
#define init_10_10_3_13
void FASTCALL call_10_10_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_3_14
#define init_10_10_3_14
void FASTCALL call_10_10_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_3_15
#define init_10_10_3_15
void FASTCALL call_10_10_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_4_0
#define init_10_10_4_0
void FASTCALL call_10_10_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_4_1
#define init_10_10_4_1
void FASTCALL call_10_10_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_4_2
#define init_10_10_4_2
void FASTCALL call_10_10_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_4_3
#define init_10_10_4_3
void FASTCALL call_10_10_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_4_4
#define init_10_10_4_4
void FASTCALL call_10_10_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_4_5
#define init_10_10_4_5
void FASTCALL call_10_10_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_4_6
#define init_10_10_4_6
void FASTCALL call_10_10_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_4_7
#define init_10_10_4_7
void FASTCALL call_10_10_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_4_8
#define init_10_10_4_8
void FASTCALL call_10_10_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_4_9
#define init_10_10_4_9
void FASTCALL call_10_10_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_4_10
#define init_10_10_4_10
void FASTCALL call_10_10_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_4_11
#define init_10_10_4_11
void FASTCALL call_10_10_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_4_12
#define init_10_10_4_12
void FASTCALL call_10_10_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_4_13
#define init_10_10_4_13
void FASTCALL call_10_10_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_4_14
#define init_10_10_4_14
void FASTCALL call_10_10_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_4_15
#define init_10_10_4_15
void FASTCALL call_10_10_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_5_0
#define init_10_10_5_0
void FASTCALL call_10_10_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_5_1
#define init_10_10_5_1
void FASTCALL call_10_10_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_5_2
#define init_10_10_5_2
void FASTCALL call_10_10_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_5_3
#define init_10_10_5_3
void FASTCALL call_10_10_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_5_4
#define init_10_10_5_4
void FASTCALL call_10_10_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_5_5
#define init_10_10_5_5
void FASTCALL call_10_10_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_5_6
#define init_10_10_5_6
void FASTCALL call_10_10_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_5_7
#define init_10_10_5_7
void FASTCALL call_10_10_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_5_8
#define init_10_10_5_8
void FASTCALL call_10_10_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_5_9
#define init_10_10_5_9
void FASTCALL call_10_10_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_5_10
#define init_10_10_5_10
void FASTCALL call_10_10_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_5_11
#define init_10_10_5_11
void FASTCALL call_10_10_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_5_12
#define init_10_10_5_12
void FASTCALL call_10_10_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_5_13
#define init_10_10_5_13
void FASTCALL call_10_10_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_5_14
#define init_10_10_5_14
void FASTCALL call_10_10_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_5_15
#define init_10_10_5_15
void FASTCALL call_10_10_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_6_0
#define init_10_10_6_0
void FASTCALL call_10_10_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_6_1
#define init_10_10_6_1
void FASTCALL call_10_10_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_6_2
#define init_10_10_6_2
void FASTCALL call_10_10_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_6_3
#define init_10_10_6_3
void FASTCALL call_10_10_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_6_4
#define init_10_10_6_4
void FASTCALL call_10_10_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_6_5
#define init_10_10_6_5
void FASTCALL call_10_10_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_6_6
#define init_10_10_6_6
void FASTCALL call_10_10_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_6_7
#define init_10_10_6_7
void FASTCALL call_10_10_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_6_8
#define init_10_10_6_8
void FASTCALL call_10_10_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_6_9
#define init_10_10_6_9
void FASTCALL call_10_10_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_6_10
#define init_10_10_6_10
void FASTCALL call_10_10_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_6_11
#define init_10_10_6_11
void FASTCALL call_10_10_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_6_12
#define init_10_10_6_12
void FASTCALL call_10_10_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_6_13
#define init_10_10_6_13
void FASTCALL call_10_10_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_6_14
#define init_10_10_6_14
void FASTCALL call_10_10_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_6_15
#define init_10_10_6_15
void FASTCALL call_10_10_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_7_0
#define init_10_10_7_0
void FASTCALL call_10_10_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_7_1
#define init_10_10_7_1
void FASTCALL call_10_10_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_7_2
#define init_10_10_7_2
void FASTCALL call_10_10_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_7_3
#define init_10_10_7_3
void FASTCALL call_10_10_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_7_4
#define init_10_10_7_4
void FASTCALL call_10_10_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_7_5
#define init_10_10_7_5
void FASTCALL call_10_10_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_7_6
#define init_10_10_7_6
void FASTCALL call_10_10_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_7_7
#define init_10_10_7_7
void FASTCALL call_10_10_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_7_8
#define init_10_10_7_8
void FASTCALL call_10_10_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_7_9
#define init_10_10_7_9
void FASTCALL call_10_10_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_7_10
#define init_10_10_7_10
void FASTCALL call_10_10_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_7_11
#define init_10_10_7_11
void FASTCALL call_10_10_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_7_12
#define init_10_10_7_12
void FASTCALL call_10_10_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_7_13
#define init_10_10_7_13
void FASTCALL call_10_10_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_7_14
#define init_10_10_7_14
void FASTCALL call_10_10_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_7_15
#define init_10_10_7_15
void FASTCALL call_10_10_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_8_0
#define init_10_10_8_0
void FASTCALL call_10_10_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_8_1
#define init_10_10_8_1
void FASTCALL call_10_10_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_8_2
#define init_10_10_8_2
void FASTCALL call_10_10_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_8_3
#define init_10_10_8_3
void FASTCALL call_10_10_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_8_4
#define init_10_10_8_4
void FASTCALL call_10_10_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_8_5
#define init_10_10_8_5
void FASTCALL call_10_10_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_8_6
#define init_10_10_8_6
void FASTCALL call_10_10_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_8_7
#define init_10_10_8_7
void FASTCALL call_10_10_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_8_8
#define init_10_10_8_8
void FASTCALL call_10_10_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_8_9
#define init_10_10_8_9
void FASTCALL call_10_10_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_8_10
#define init_10_10_8_10
void FASTCALL call_10_10_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_8_11
#define init_10_10_8_11
void FASTCALL call_10_10_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_8_12
#define init_10_10_8_12
void FASTCALL call_10_10_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_8_13
#define init_10_10_8_13
void FASTCALL call_10_10_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_8_14
#define init_10_10_8_14
void FASTCALL call_10_10_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_8_15
#define init_10_10_8_15
void FASTCALL call_10_10_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_9_0
#define init_10_10_9_0
void FASTCALL call_10_10_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_9_1
#define init_10_10_9_1
void FASTCALL call_10_10_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_9_2
#define init_10_10_9_2
void FASTCALL call_10_10_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_9_3
#define init_10_10_9_3
void FASTCALL call_10_10_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_9_4
#define init_10_10_9_4
void FASTCALL call_10_10_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_9_5
#define init_10_10_9_5
void FASTCALL call_10_10_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_9_6
#define init_10_10_9_6
void FASTCALL call_10_10_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_9_7
#define init_10_10_9_7
void FASTCALL call_10_10_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_9_8
#define init_10_10_9_8
void FASTCALL call_10_10_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_9_9
#define init_10_10_9_9
void FASTCALL call_10_10_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_9_10
#define init_10_10_9_10
void FASTCALL call_10_10_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_9_11
#define init_10_10_9_11
void FASTCALL call_10_10_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_9_12
#define init_10_10_9_12
void FASTCALL call_10_10_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_9_13
#define init_10_10_9_13
void FASTCALL call_10_10_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_9_14
#define init_10_10_9_14
void FASTCALL call_10_10_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_9_15
#define init_10_10_9_15
void FASTCALL call_10_10_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_10_0
#define init_10_10_10_0
void FASTCALL call_10_10_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_10_1
#define init_10_10_10_1
void FASTCALL call_10_10_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_10_2
#define init_10_10_10_2
void FASTCALL call_10_10_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_10_3
#define init_10_10_10_3
void FASTCALL call_10_10_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_10_4
#define init_10_10_10_4
void FASTCALL call_10_10_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_10_5
#define init_10_10_10_5
void FASTCALL call_10_10_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_10_6
#define init_10_10_10_6
void FASTCALL call_10_10_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_10_7
#define init_10_10_10_7
void FASTCALL call_10_10_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_10_8
#define init_10_10_10_8
void FASTCALL call_10_10_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_10_9
#define init_10_10_10_9
void FASTCALL call_10_10_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_10_10
#define init_10_10_10_10
void FASTCALL call_10_10_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_10_11
#define init_10_10_10_11
void FASTCALL call_10_10_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_10_12
#define init_10_10_10_12
void FASTCALL call_10_10_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_10_13
#define init_10_10_10_13
void FASTCALL call_10_10_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_10_14
#define init_10_10_10_14
void FASTCALL call_10_10_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_10_15
#define init_10_10_10_15
void FASTCALL call_10_10_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_11_0
#define init_10_10_11_0
void FASTCALL call_10_10_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_11_1
#define init_10_10_11_1
void FASTCALL call_10_10_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_11_2
#define init_10_10_11_2
void FASTCALL call_10_10_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_11_3
#define init_10_10_11_3
void FASTCALL call_10_10_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_11_4
#define init_10_10_11_4
void FASTCALL call_10_10_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_11_5
#define init_10_10_11_5
void FASTCALL call_10_10_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_11_6
#define init_10_10_11_6
void FASTCALL call_10_10_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_11_7
#define init_10_10_11_7
void FASTCALL call_10_10_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_11_8
#define init_10_10_11_8
void FASTCALL call_10_10_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_11_9
#define init_10_10_11_9
void FASTCALL call_10_10_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_11_10
#define init_10_10_11_10
void FASTCALL call_10_10_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_11_11
#define init_10_10_11_11
void FASTCALL call_10_10_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_11_12
#define init_10_10_11_12
void FASTCALL call_10_10_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_11_13
#define init_10_10_11_13
void FASTCALL call_10_10_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_11_14
#define init_10_10_11_14
void FASTCALL call_10_10_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_11_15
#define init_10_10_11_15
void FASTCALL call_10_10_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_12_0
#define init_10_10_12_0
void FASTCALL call_10_10_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_12_1
#define init_10_10_12_1
void FASTCALL call_10_10_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_12_2
#define init_10_10_12_2
void FASTCALL call_10_10_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_12_3
#define init_10_10_12_3
void FASTCALL call_10_10_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_12_4
#define init_10_10_12_4
void FASTCALL call_10_10_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_12_5
#define init_10_10_12_5
void FASTCALL call_10_10_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_12_6
#define init_10_10_12_6
void FASTCALL call_10_10_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_12_7
#define init_10_10_12_7
void FASTCALL call_10_10_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_12_8
#define init_10_10_12_8
void FASTCALL call_10_10_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_12_9
#define init_10_10_12_9
void FASTCALL call_10_10_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_12_10
#define init_10_10_12_10
void FASTCALL call_10_10_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_12_11
#define init_10_10_12_11
void FASTCALL call_10_10_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_12_12
#define init_10_10_12_12
void FASTCALL call_10_10_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_12_13
#define init_10_10_12_13
void FASTCALL call_10_10_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_12_14
#define init_10_10_12_14
void FASTCALL call_10_10_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_12_15
#define init_10_10_12_15
void FASTCALL call_10_10_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_13_0
#define init_10_10_13_0
void FASTCALL call_10_10_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_13_1
#define init_10_10_13_1
void FASTCALL call_10_10_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_13_2
#define init_10_10_13_2
void FASTCALL call_10_10_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_13_3
#define init_10_10_13_3
void FASTCALL call_10_10_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_13_4
#define init_10_10_13_4
void FASTCALL call_10_10_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_13_5
#define init_10_10_13_5
void FASTCALL call_10_10_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_13_6
#define init_10_10_13_6
void FASTCALL call_10_10_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_13_7
#define init_10_10_13_7
void FASTCALL call_10_10_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_13_8
#define init_10_10_13_8
void FASTCALL call_10_10_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_13_9
#define init_10_10_13_9
void FASTCALL call_10_10_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_13_10
#define init_10_10_13_10
void FASTCALL call_10_10_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_13_11
#define init_10_10_13_11
void FASTCALL call_10_10_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_13_12
#define init_10_10_13_12
void FASTCALL call_10_10_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_13_13
#define init_10_10_13_13
void FASTCALL call_10_10_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_13_14
#define init_10_10_13_14
void FASTCALL call_10_10_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_13_15
#define init_10_10_13_15
void FASTCALL call_10_10_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_14_0
#define init_10_10_14_0
void FASTCALL call_10_10_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_14_1
#define init_10_10_14_1
void FASTCALL call_10_10_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_14_2
#define init_10_10_14_2
void FASTCALL call_10_10_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_14_3
#define init_10_10_14_3
void FASTCALL call_10_10_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_14_4
#define init_10_10_14_4
void FASTCALL call_10_10_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_14_5
#define init_10_10_14_5
void FASTCALL call_10_10_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_14_6
#define init_10_10_14_6
void FASTCALL call_10_10_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_14_7
#define init_10_10_14_7
void FASTCALL call_10_10_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_14_8
#define init_10_10_14_8
void FASTCALL call_10_10_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_14_9
#define init_10_10_14_9
void FASTCALL call_10_10_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_14_10
#define init_10_10_14_10
void FASTCALL call_10_10_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_14_11
#define init_10_10_14_11
void FASTCALL call_10_10_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_14_12
#define init_10_10_14_12
void FASTCALL call_10_10_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_14_13
#define init_10_10_14_13
void FASTCALL call_10_10_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_14_14
#define init_10_10_14_14
void FASTCALL call_10_10_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_14_15
#define init_10_10_14_15
void FASTCALL call_10_10_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_15_0
#define init_10_10_15_0
void FASTCALL call_10_10_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_15_1
#define init_10_10_15_1
void FASTCALL call_10_10_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_15_2
#define init_10_10_15_2
void FASTCALL call_10_10_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_15_3
#define init_10_10_15_3
void FASTCALL call_10_10_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_15_4
#define init_10_10_15_4
void FASTCALL call_10_10_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_15_5
#define init_10_10_15_5
void FASTCALL call_10_10_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_15_6
#define init_10_10_15_6
void FASTCALL call_10_10_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_15_7
#define init_10_10_15_7
void FASTCALL call_10_10_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_15_8
#define init_10_10_15_8
void FASTCALL call_10_10_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_15_9
#define init_10_10_15_9
void FASTCALL call_10_10_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_15_10
#define init_10_10_15_10
void FASTCALL call_10_10_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_15_11
#define init_10_10_15_11
void FASTCALL call_10_10_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_15_12
#define init_10_10_15_12
void FASTCALL call_10_10_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_15_13
#define init_10_10_15_13
void FASTCALL call_10_10_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_15_14
#define init_10_10_15_14
void FASTCALL call_10_10_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_10_15_15
#define init_10_10_15_15
void FASTCALL call_10_10_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_0_0
#define init_10_11_0_0
void FASTCALL call_10_11_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_0_1
#define init_10_11_0_1
void FASTCALL call_10_11_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_0_2
#define init_10_11_0_2
void FASTCALL call_10_11_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_0_3
#define init_10_11_0_3
void FASTCALL call_10_11_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_0_4
#define init_10_11_0_4
void FASTCALL call_10_11_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_0_5
#define init_10_11_0_5
void FASTCALL call_10_11_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_0_6
#define init_10_11_0_6
void FASTCALL call_10_11_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_0_7
#define init_10_11_0_7
void FASTCALL call_10_11_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_0_8
#define init_10_11_0_8
void FASTCALL call_10_11_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_0_9
#define init_10_11_0_9
void FASTCALL call_10_11_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_0_10
#define init_10_11_0_10
void FASTCALL call_10_11_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_0_11
#define init_10_11_0_11
void FASTCALL call_10_11_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_0_12
#define init_10_11_0_12
void FASTCALL call_10_11_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_0_13
#define init_10_11_0_13
void FASTCALL call_10_11_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_0_14
#define init_10_11_0_14
void FASTCALL call_10_11_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_0_15
#define init_10_11_0_15
void FASTCALL call_10_11_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_1_0
#define init_10_11_1_0
void FASTCALL call_10_11_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_1_1
#define init_10_11_1_1
void FASTCALL call_10_11_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_1_2
#define init_10_11_1_2
void FASTCALL call_10_11_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_1_3
#define init_10_11_1_3
void FASTCALL call_10_11_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_1_4
#define init_10_11_1_4
void FASTCALL call_10_11_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_1_5
#define init_10_11_1_5
void FASTCALL call_10_11_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_1_6
#define init_10_11_1_6
void FASTCALL call_10_11_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_1_7
#define init_10_11_1_7
void FASTCALL call_10_11_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_1_8
#define init_10_11_1_8
void FASTCALL call_10_11_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_1_9
#define init_10_11_1_9
void FASTCALL call_10_11_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_1_10
#define init_10_11_1_10
void FASTCALL call_10_11_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_1_11
#define init_10_11_1_11
void FASTCALL call_10_11_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_1_12
#define init_10_11_1_12
void FASTCALL call_10_11_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_1_13
#define init_10_11_1_13
void FASTCALL call_10_11_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_1_14
#define init_10_11_1_14
void FASTCALL call_10_11_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_1_15
#define init_10_11_1_15
void FASTCALL call_10_11_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_2_0
#define init_10_11_2_0
void FASTCALL call_10_11_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_2_1
#define init_10_11_2_1
void FASTCALL call_10_11_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_2_2
#define init_10_11_2_2
void FASTCALL call_10_11_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_2_3
#define init_10_11_2_3
void FASTCALL call_10_11_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_2_4
#define init_10_11_2_4
void FASTCALL call_10_11_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_2_5
#define init_10_11_2_5
void FASTCALL call_10_11_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_2_6
#define init_10_11_2_6
void FASTCALL call_10_11_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_2_7
#define init_10_11_2_7
void FASTCALL call_10_11_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_2_8
#define init_10_11_2_8
void FASTCALL call_10_11_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_2_9
#define init_10_11_2_9
void FASTCALL call_10_11_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_2_10
#define init_10_11_2_10
void FASTCALL call_10_11_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_2_11
#define init_10_11_2_11
void FASTCALL call_10_11_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_2_12
#define init_10_11_2_12
void FASTCALL call_10_11_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_2_13
#define init_10_11_2_13
void FASTCALL call_10_11_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_2_14
#define init_10_11_2_14
void FASTCALL call_10_11_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_2_15
#define init_10_11_2_15
void FASTCALL call_10_11_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_3_0
#define init_10_11_3_0
void FASTCALL call_10_11_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_3_1
#define init_10_11_3_1
void FASTCALL call_10_11_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_3_2
#define init_10_11_3_2
void FASTCALL call_10_11_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_3_3
#define init_10_11_3_3
void FASTCALL call_10_11_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_3_4
#define init_10_11_3_4
void FASTCALL call_10_11_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_3_5
#define init_10_11_3_5
void FASTCALL call_10_11_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_3_6
#define init_10_11_3_6
void FASTCALL call_10_11_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_3_7
#define init_10_11_3_7
void FASTCALL call_10_11_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_3_8
#define init_10_11_3_8
void FASTCALL call_10_11_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_3_9
#define init_10_11_3_9
void FASTCALL call_10_11_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_3_10
#define init_10_11_3_10
void FASTCALL call_10_11_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_3_11
#define init_10_11_3_11
void FASTCALL call_10_11_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_3_12
#define init_10_11_3_12
void FASTCALL call_10_11_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_3_13
#define init_10_11_3_13
void FASTCALL call_10_11_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_3_14
#define init_10_11_3_14
void FASTCALL call_10_11_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_3_15
#define init_10_11_3_15
void FASTCALL call_10_11_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_4_0
#define init_10_11_4_0
void FASTCALL call_10_11_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_4_1
#define init_10_11_4_1
void FASTCALL call_10_11_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_4_2
#define init_10_11_4_2
void FASTCALL call_10_11_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_4_3
#define init_10_11_4_3
void FASTCALL call_10_11_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_4_4
#define init_10_11_4_4
void FASTCALL call_10_11_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_4_5
#define init_10_11_4_5
void FASTCALL call_10_11_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_4_6
#define init_10_11_4_6
void FASTCALL call_10_11_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_4_7
#define init_10_11_4_7
void FASTCALL call_10_11_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_4_8
#define init_10_11_4_8
void FASTCALL call_10_11_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_4_9
#define init_10_11_4_9
void FASTCALL call_10_11_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_4_10
#define init_10_11_4_10
void FASTCALL call_10_11_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_4_11
#define init_10_11_4_11
void FASTCALL call_10_11_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_4_12
#define init_10_11_4_12
void FASTCALL call_10_11_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_4_13
#define init_10_11_4_13
void FASTCALL call_10_11_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_4_14
#define init_10_11_4_14
void FASTCALL call_10_11_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_4_15
#define init_10_11_4_15
void FASTCALL call_10_11_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_5_0
#define init_10_11_5_0
void FASTCALL call_10_11_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_5_1
#define init_10_11_5_1
void FASTCALL call_10_11_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_5_2
#define init_10_11_5_2
void FASTCALL call_10_11_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_5_3
#define init_10_11_5_3
void FASTCALL call_10_11_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_5_4
#define init_10_11_5_4
void FASTCALL call_10_11_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_5_5
#define init_10_11_5_5
void FASTCALL call_10_11_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_5_6
#define init_10_11_5_6
void FASTCALL call_10_11_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_5_7
#define init_10_11_5_7
void FASTCALL call_10_11_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_5_8
#define init_10_11_5_8
void FASTCALL call_10_11_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_5_9
#define init_10_11_5_9
void FASTCALL call_10_11_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_5_10
#define init_10_11_5_10
void FASTCALL call_10_11_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_5_11
#define init_10_11_5_11
void FASTCALL call_10_11_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_5_12
#define init_10_11_5_12
void FASTCALL call_10_11_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_5_13
#define init_10_11_5_13
void FASTCALL call_10_11_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_5_14
#define init_10_11_5_14
void FASTCALL call_10_11_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_5_15
#define init_10_11_5_15
void FASTCALL call_10_11_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_6_0
#define init_10_11_6_0
void FASTCALL call_10_11_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_6_1
#define init_10_11_6_1
void FASTCALL call_10_11_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_6_2
#define init_10_11_6_2
void FASTCALL call_10_11_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_6_3
#define init_10_11_6_3
void FASTCALL call_10_11_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_6_4
#define init_10_11_6_4
void FASTCALL call_10_11_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_6_5
#define init_10_11_6_5
void FASTCALL call_10_11_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_6_6
#define init_10_11_6_6
void FASTCALL call_10_11_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_6_7
#define init_10_11_6_7
void FASTCALL call_10_11_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_6_8
#define init_10_11_6_8
void FASTCALL call_10_11_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_6_9
#define init_10_11_6_9
void FASTCALL call_10_11_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_6_10
#define init_10_11_6_10
void FASTCALL call_10_11_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_6_11
#define init_10_11_6_11
void FASTCALL call_10_11_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_6_12
#define init_10_11_6_12
void FASTCALL call_10_11_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_6_13
#define init_10_11_6_13
void FASTCALL call_10_11_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_6_14
#define init_10_11_6_14
void FASTCALL call_10_11_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_6_15
#define init_10_11_6_15
void FASTCALL call_10_11_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_7_0
#define init_10_11_7_0
void FASTCALL call_10_11_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_7_1
#define init_10_11_7_1
void FASTCALL call_10_11_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_7_2
#define init_10_11_7_2
void FASTCALL call_10_11_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_7_3
#define init_10_11_7_3
void FASTCALL call_10_11_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_7_4
#define init_10_11_7_4
void FASTCALL call_10_11_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_7_5
#define init_10_11_7_5
void FASTCALL call_10_11_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_7_6
#define init_10_11_7_6
void FASTCALL call_10_11_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_7_7
#define init_10_11_7_7
void FASTCALL call_10_11_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_7_8
#define init_10_11_7_8
void FASTCALL call_10_11_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_7_9
#define init_10_11_7_9
void FASTCALL call_10_11_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_7_10
#define init_10_11_7_10
void FASTCALL call_10_11_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_7_11
#define init_10_11_7_11
void FASTCALL call_10_11_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_7_12
#define init_10_11_7_12
void FASTCALL call_10_11_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_7_13
#define init_10_11_7_13
void FASTCALL call_10_11_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_7_14
#define init_10_11_7_14
void FASTCALL call_10_11_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_7_15
#define init_10_11_7_15
void FASTCALL call_10_11_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_8_0
#define init_10_11_8_0
void FASTCALL call_10_11_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_8_1
#define init_10_11_8_1
void FASTCALL call_10_11_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_8_2
#define init_10_11_8_2
void FASTCALL call_10_11_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_8_3
#define init_10_11_8_3
void FASTCALL call_10_11_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_8_4
#define init_10_11_8_4
void FASTCALL call_10_11_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_8_5
#define init_10_11_8_5
void FASTCALL call_10_11_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_8_6
#define init_10_11_8_6
void FASTCALL call_10_11_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_8_7
#define init_10_11_8_7
void FASTCALL call_10_11_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_8_8
#define init_10_11_8_8
void FASTCALL call_10_11_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_8_9
#define init_10_11_8_9
void FASTCALL call_10_11_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_8_10
#define init_10_11_8_10
void FASTCALL call_10_11_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_8_11
#define init_10_11_8_11
void FASTCALL call_10_11_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_8_12
#define init_10_11_8_12
void FASTCALL call_10_11_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_8_13
#define init_10_11_8_13
void FASTCALL call_10_11_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_8_14
#define init_10_11_8_14
void FASTCALL call_10_11_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_8_15
#define init_10_11_8_15
void FASTCALL call_10_11_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_9_0
#define init_10_11_9_0
void FASTCALL call_10_11_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_9_1
#define init_10_11_9_1
void FASTCALL call_10_11_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_9_2
#define init_10_11_9_2
void FASTCALL call_10_11_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_9_3
#define init_10_11_9_3
void FASTCALL call_10_11_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_9_4
#define init_10_11_9_4
void FASTCALL call_10_11_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_9_5
#define init_10_11_9_5
void FASTCALL call_10_11_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_9_6
#define init_10_11_9_6
void FASTCALL call_10_11_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_9_7
#define init_10_11_9_7
void FASTCALL call_10_11_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_9_8
#define init_10_11_9_8
void FASTCALL call_10_11_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_9_9
#define init_10_11_9_9
void FASTCALL call_10_11_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_9_10
#define init_10_11_9_10
void FASTCALL call_10_11_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_9_11
#define init_10_11_9_11
void FASTCALL call_10_11_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_9_12
#define init_10_11_9_12
void FASTCALL call_10_11_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_9_13
#define init_10_11_9_13
void FASTCALL call_10_11_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_9_14
#define init_10_11_9_14
void FASTCALL call_10_11_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_9_15
#define init_10_11_9_15
void FASTCALL call_10_11_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_10_0
#define init_10_11_10_0
void FASTCALL call_10_11_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_10_1
#define init_10_11_10_1
void FASTCALL call_10_11_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_10_2
#define init_10_11_10_2
void FASTCALL call_10_11_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_10_3
#define init_10_11_10_3
void FASTCALL call_10_11_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_10_4
#define init_10_11_10_4
void FASTCALL call_10_11_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_10_5
#define init_10_11_10_5
void FASTCALL call_10_11_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_10_6
#define init_10_11_10_6
void FASTCALL call_10_11_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_10_7
#define init_10_11_10_7
void FASTCALL call_10_11_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_10_8
#define init_10_11_10_8
void FASTCALL call_10_11_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_10_9
#define init_10_11_10_9
void FASTCALL call_10_11_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_10_10
#define init_10_11_10_10
void FASTCALL call_10_11_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_10_11
#define init_10_11_10_11
void FASTCALL call_10_11_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_10_12
#define init_10_11_10_12
void FASTCALL call_10_11_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_10_13
#define init_10_11_10_13
void FASTCALL call_10_11_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_10_14
#define init_10_11_10_14
void FASTCALL call_10_11_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_10_15
#define init_10_11_10_15
void FASTCALL call_10_11_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_11_0
#define init_10_11_11_0
void FASTCALL call_10_11_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_11_1
#define init_10_11_11_1
void FASTCALL call_10_11_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_11_2
#define init_10_11_11_2
void FASTCALL call_10_11_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_11_3
#define init_10_11_11_3
void FASTCALL call_10_11_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_11_4
#define init_10_11_11_4
void FASTCALL call_10_11_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_11_5
#define init_10_11_11_5
void FASTCALL call_10_11_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_11_6
#define init_10_11_11_6
void FASTCALL call_10_11_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_11_7
#define init_10_11_11_7
void FASTCALL call_10_11_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_11_8
#define init_10_11_11_8
void FASTCALL call_10_11_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_11_9
#define init_10_11_11_9
void FASTCALL call_10_11_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_11_10
#define init_10_11_11_10
void FASTCALL call_10_11_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_11_11
#define init_10_11_11_11
void FASTCALL call_10_11_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_11_12
#define init_10_11_11_12
void FASTCALL call_10_11_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_11_13
#define init_10_11_11_13
void FASTCALL call_10_11_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_11_14
#define init_10_11_11_14
void FASTCALL call_10_11_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_11_15
#define init_10_11_11_15
void FASTCALL call_10_11_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_12_0
#define init_10_11_12_0
void FASTCALL call_10_11_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_12_1
#define init_10_11_12_1
void FASTCALL call_10_11_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_12_2
#define init_10_11_12_2
void FASTCALL call_10_11_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_12_3
#define init_10_11_12_3
void FASTCALL call_10_11_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_12_4
#define init_10_11_12_4
void FASTCALL call_10_11_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_12_5
#define init_10_11_12_5
void FASTCALL call_10_11_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_12_6
#define init_10_11_12_6
void FASTCALL call_10_11_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_12_7
#define init_10_11_12_7
void FASTCALL call_10_11_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_12_8
#define init_10_11_12_8
void FASTCALL call_10_11_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_12_9
#define init_10_11_12_9
void FASTCALL call_10_11_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_12_10
#define init_10_11_12_10
void FASTCALL call_10_11_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_12_11
#define init_10_11_12_11
void FASTCALL call_10_11_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_12_12
#define init_10_11_12_12
void FASTCALL call_10_11_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_12_13
#define init_10_11_12_13
void FASTCALL call_10_11_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_12_14
#define init_10_11_12_14
void FASTCALL call_10_11_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_12_15
#define init_10_11_12_15
void FASTCALL call_10_11_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_13_0
#define init_10_11_13_0
void FASTCALL call_10_11_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_13_1
#define init_10_11_13_1
void FASTCALL call_10_11_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_13_2
#define init_10_11_13_2
void FASTCALL call_10_11_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_13_3
#define init_10_11_13_3
void FASTCALL call_10_11_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_13_4
#define init_10_11_13_4
void FASTCALL call_10_11_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_13_5
#define init_10_11_13_5
void FASTCALL call_10_11_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_13_6
#define init_10_11_13_6
void FASTCALL call_10_11_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_13_7
#define init_10_11_13_7
void FASTCALL call_10_11_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_13_8
#define init_10_11_13_8
void FASTCALL call_10_11_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_13_9
#define init_10_11_13_9
void FASTCALL call_10_11_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_13_10
#define init_10_11_13_10
void FASTCALL call_10_11_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_13_11
#define init_10_11_13_11
void FASTCALL call_10_11_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_13_12
#define init_10_11_13_12
void FASTCALL call_10_11_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_13_13
#define init_10_11_13_13
void FASTCALL call_10_11_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_13_14
#define init_10_11_13_14
void FASTCALL call_10_11_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_13_15
#define init_10_11_13_15
void FASTCALL call_10_11_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_14_0
#define init_10_11_14_0
void FASTCALL call_10_11_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_14_1
#define init_10_11_14_1
void FASTCALL call_10_11_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_14_2
#define init_10_11_14_2
void FASTCALL call_10_11_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_14_3
#define init_10_11_14_3
void FASTCALL call_10_11_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_14_4
#define init_10_11_14_4
void FASTCALL call_10_11_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_14_5
#define init_10_11_14_5
void FASTCALL call_10_11_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_14_6
#define init_10_11_14_6
void FASTCALL call_10_11_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_14_7
#define init_10_11_14_7
void FASTCALL call_10_11_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_14_8
#define init_10_11_14_8
void FASTCALL call_10_11_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_14_9
#define init_10_11_14_9
void FASTCALL call_10_11_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_14_10
#define init_10_11_14_10
void FASTCALL call_10_11_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_14_11
#define init_10_11_14_11
void FASTCALL call_10_11_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_14_12
#define init_10_11_14_12
void FASTCALL call_10_11_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_14_13
#define init_10_11_14_13
void FASTCALL call_10_11_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_14_14
#define init_10_11_14_14
void FASTCALL call_10_11_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_14_15
#define init_10_11_14_15
void FASTCALL call_10_11_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_15_0
#define init_10_11_15_0
void FASTCALL call_10_11_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_15_1
#define init_10_11_15_1
void FASTCALL call_10_11_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_15_2
#define init_10_11_15_2
void FASTCALL call_10_11_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_15_3
#define init_10_11_15_3
void FASTCALL call_10_11_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_15_4
#define init_10_11_15_4
void FASTCALL call_10_11_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_15_5
#define init_10_11_15_5
void FASTCALL call_10_11_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_15_6
#define init_10_11_15_6
void FASTCALL call_10_11_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_15_7
#define init_10_11_15_7
void FASTCALL call_10_11_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_15_8
#define init_10_11_15_8
void FASTCALL call_10_11_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_15_9
#define init_10_11_15_9
void FASTCALL call_10_11_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_15_10
#define init_10_11_15_10
void FASTCALL call_10_11_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_15_11
#define init_10_11_15_11
void FASTCALL call_10_11_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_15_12
#define init_10_11_15_12
void FASTCALL call_10_11_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_15_13
#define init_10_11_15_13
void FASTCALL call_10_11_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_15_14
#define init_10_11_15_14
void FASTCALL call_10_11_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_11_15_15
#define init_10_11_15_15
void FASTCALL call_10_11_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_0_0
#define init_10_12_0_0
void FASTCALL call_10_12_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_0_1
#define init_10_12_0_1
void FASTCALL call_10_12_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_0_2
#define init_10_12_0_2
void FASTCALL call_10_12_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_0_3
#define init_10_12_0_3
void FASTCALL call_10_12_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_0_4
#define init_10_12_0_4
void FASTCALL call_10_12_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_0_5
#define init_10_12_0_5
void FASTCALL call_10_12_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_0_6
#define init_10_12_0_6
void FASTCALL call_10_12_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_0_7
#define init_10_12_0_7
void FASTCALL call_10_12_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_0_8
#define init_10_12_0_8
void FASTCALL call_10_12_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_0_9
#define init_10_12_0_9
void FASTCALL call_10_12_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_0_10
#define init_10_12_0_10
void FASTCALL call_10_12_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_0_11
#define init_10_12_0_11
void FASTCALL call_10_12_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_0_12
#define init_10_12_0_12
void FASTCALL call_10_12_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_0_13
#define init_10_12_0_13
void FASTCALL call_10_12_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_0_14
#define init_10_12_0_14
void FASTCALL call_10_12_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_0_15
#define init_10_12_0_15
void FASTCALL call_10_12_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_1_0
#define init_10_12_1_0
void FASTCALL call_10_12_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_1_1
#define init_10_12_1_1
void FASTCALL call_10_12_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_1_2
#define init_10_12_1_2
void FASTCALL call_10_12_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_1_3
#define init_10_12_1_3
void FASTCALL call_10_12_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_1_4
#define init_10_12_1_4
void FASTCALL call_10_12_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_1_5
#define init_10_12_1_5
void FASTCALL call_10_12_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_1_6
#define init_10_12_1_6
void FASTCALL call_10_12_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_1_7
#define init_10_12_1_7
void FASTCALL call_10_12_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_1_8
#define init_10_12_1_8
void FASTCALL call_10_12_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_1_9
#define init_10_12_1_9
void FASTCALL call_10_12_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_1_10
#define init_10_12_1_10
void FASTCALL call_10_12_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_1_11
#define init_10_12_1_11
void FASTCALL call_10_12_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_1_12
#define init_10_12_1_12
void FASTCALL call_10_12_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_1_13
#define init_10_12_1_13
void FASTCALL call_10_12_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_1_14
#define init_10_12_1_14
void FASTCALL call_10_12_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_1_15
#define init_10_12_1_15
void FASTCALL call_10_12_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_2_0
#define init_10_12_2_0
void FASTCALL call_10_12_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_2_1
#define init_10_12_2_1
void FASTCALL call_10_12_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_2_2
#define init_10_12_2_2
void FASTCALL call_10_12_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_2_3
#define init_10_12_2_3
void FASTCALL call_10_12_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_2_4
#define init_10_12_2_4
void FASTCALL call_10_12_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_2_5
#define init_10_12_2_5
void FASTCALL call_10_12_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_2_6
#define init_10_12_2_6
void FASTCALL call_10_12_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_2_7
#define init_10_12_2_7
void FASTCALL call_10_12_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_2_8
#define init_10_12_2_8
void FASTCALL call_10_12_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_2_9
#define init_10_12_2_9
void FASTCALL call_10_12_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_2_10
#define init_10_12_2_10
void FASTCALL call_10_12_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_2_11
#define init_10_12_2_11
void FASTCALL call_10_12_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_2_12
#define init_10_12_2_12
void FASTCALL call_10_12_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_2_13
#define init_10_12_2_13
void FASTCALL call_10_12_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_2_14
#define init_10_12_2_14
void FASTCALL call_10_12_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_2_15
#define init_10_12_2_15
void FASTCALL call_10_12_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_3_0
#define init_10_12_3_0
void FASTCALL call_10_12_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_3_1
#define init_10_12_3_1
void FASTCALL call_10_12_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_3_2
#define init_10_12_3_2
void FASTCALL call_10_12_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_3_3
#define init_10_12_3_3
void FASTCALL call_10_12_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_3_4
#define init_10_12_3_4
void FASTCALL call_10_12_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_3_5
#define init_10_12_3_5
void FASTCALL call_10_12_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_3_6
#define init_10_12_3_6
void FASTCALL call_10_12_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_3_7
#define init_10_12_3_7
void FASTCALL call_10_12_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_3_8
#define init_10_12_3_8
void FASTCALL call_10_12_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_3_9
#define init_10_12_3_9
void FASTCALL call_10_12_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_3_10
#define init_10_12_3_10
void FASTCALL call_10_12_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_3_11
#define init_10_12_3_11
void FASTCALL call_10_12_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_3_12
#define init_10_12_3_12
void FASTCALL call_10_12_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_3_13
#define init_10_12_3_13
void FASTCALL call_10_12_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_3_14
#define init_10_12_3_14
void FASTCALL call_10_12_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_3_15
#define init_10_12_3_15
void FASTCALL call_10_12_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_4_0
#define init_10_12_4_0
void FASTCALL call_10_12_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_4_1
#define init_10_12_4_1
void FASTCALL call_10_12_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_4_2
#define init_10_12_4_2
void FASTCALL call_10_12_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_4_3
#define init_10_12_4_3
void FASTCALL call_10_12_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_4_4
#define init_10_12_4_4
void FASTCALL call_10_12_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_4_5
#define init_10_12_4_5
void FASTCALL call_10_12_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_4_6
#define init_10_12_4_6
void FASTCALL call_10_12_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_4_7
#define init_10_12_4_7
void FASTCALL call_10_12_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_4_8
#define init_10_12_4_8
void FASTCALL call_10_12_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_4_9
#define init_10_12_4_9
void FASTCALL call_10_12_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_4_10
#define init_10_12_4_10
void FASTCALL call_10_12_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_4_11
#define init_10_12_4_11
void FASTCALL call_10_12_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_4_12
#define init_10_12_4_12
void FASTCALL call_10_12_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_4_13
#define init_10_12_4_13
void FASTCALL call_10_12_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_4_14
#define init_10_12_4_14
void FASTCALL call_10_12_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_4_15
#define init_10_12_4_15
void FASTCALL call_10_12_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_5_0
#define init_10_12_5_0
void FASTCALL call_10_12_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_5_1
#define init_10_12_5_1
void FASTCALL call_10_12_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_5_2
#define init_10_12_5_2
void FASTCALL call_10_12_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_5_3
#define init_10_12_5_3
void FASTCALL call_10_12_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_5_4
#define init_10_12_5_4
void FASTCALL call_10_12_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_5_5
#define init_10_12_5_5
void FASTCALL call_10_12_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_5_6
#define init_10_12_5_6
void FASTCALL call_10_12_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_5_7
#define init_10_12_5_7
void FASTCALL call_10_12_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_5_8
#define init_10_12_5_8
void FASTCALL call_10_12_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_5_9
#define init_10_12_5_9
void FASTCALL call_10_12_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_5_10
#define init_10_12_5_10
void FASTCALL call_10_12_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_5_11
#define init_10_12_5_11
void FASTCALL call_10_12_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_5_12
#define init_10_12_5_12
void FASTCALL call_10_12_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_5_13
#define init_10_12_5_13
void FASTCALL call_10_12_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_5_14
#define init_10_12_5_14
void FASTCALL call_10_12_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_5_15
#define init_10_12_5_15
void FASTCALL call_10_12_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_6_0
#define init_10_12_6_0
void FASTCALL call_10_12_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_6_1
#define init_10_12_6_1
void FASTCALL call_10_12_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_6_2
#define init_10_12_6_2
void FASTCALL call_10_12_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_6_3
#define init_10_12_6_3
void FASTCALL call_10_12_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_6_4
#define init_10_12_6_4
void FASTCALL call_10_12_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_6_5
#define init_10_12_6_5
void FASTCALL call_10_12_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_6_6
#define init_10_12_6_6
void FASTCALL call_10_12_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_6_7
#define init_10_12_6_7
void FASTCALL call_10_12_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_6_8
#define init_10_12_6_8
void FASTCALL call_10_12_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_6_9
#define init_10_12_6_9
void FASTCALL call_10_12_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_6_10
#define init_10_12_6_10
void FASTCALL call_10_12_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_6_11
#define init_10_12_6_11
void FASTCALL call_10_12_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_6_12
#define init_10_12_6_12
void FASTCALL call_10_12_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_6_13
#define init_10_12_6_13
void FASTCALL call_10_12_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_6_14
#define init_10_12_6_14
void FASTCALL call_10_12_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_6_15
#define init_10_12_6_15
void FASTCALL call_10_12_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_7_0
#define init_10_12_7_0
void FASTCALL call_10_12_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_7_1
#define init_10_12_7_1
void FASTCALL call_10_12_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_7_2
#define init_10_12_7_2
void FASTCALL call_10_12_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_7_3
#define init_10_12_7_3
void FASTCALL call_10_12_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_7_4
#define init_10_12_7_4
void FASTCALL call_10_12_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_7_5
#define init_10_12_7_5
void FASTCALL call_10_12_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_7_6
#define init_10_12_7_6
void FASTCALL call_10_12_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_7_7
#define init_10_12_7_7
void FASTCALL call_10_12_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_7_8
#define init_10_12_7_8
void FASTCALL call_10_12_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_7_9
#define init_10_12_7_9
void FASTCALL call_10_12_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_7_10
#define init_10_12_7_10
void FASTCALL call_10_12_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_7_11
#define init_10_12_7_11
void FASTCALL call_10_12_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_7_12
#define init_10_12_7_12
void FASTCALL call_10_12_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_7_13
#define init_10_12_7_13
void FASTCALL call_10_12_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_7_14
#define init_10_12_7_14
void FASTCALL call_10_12_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_7_15
#define init_10_12_7_15
void FASTCALL call_10_12_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_8_0
#define init_10_12_8_0
void FASTCALL call_10_12_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_8_1
#define init_10_12_8_1
void FASTCALL call_10_12_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_8_2
#define init_10_12_8_2
void FASTCALL call_10_12_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_8_3
#define init_10_12_8_3
void FASTCALL call_10_12_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_8_4
#define init_10_12_8_4
void FASTCALL call_10_12_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_8_5
#define init_10_12_8_5
void FASTCALL call_10_12_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_8_6
#define init_10_12_8_6
void FASTCALL call_10_12_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_8_7
#define init_10_12_8_7
void FASTCALL call_10_12_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_8_8
#define init_10_12_8_8
void FASTCALL call_10_12_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_8_9
#define init_10_12_8_9
void FASTCALL call_10_12_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_8_10
#define init_10_12_8_10
void FASTCALL call_10_12_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_8_11
#define init_10_12_8_11
void FASTCALL call_10_12_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_8_12
#define init_10_12_8_12
void FASTCALL call_10_12_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_8_13
#define init_10_12_8_13
void FASTCALL call_10_12_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_8_14
#define init_10_12_8_14
void FASTCALL call_10_12_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_8_15
#define init_10_12_8_15
void FASTCALL call_10_12_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_9_0
#define init_10_12_9_0
void FASTCALL call_10_12_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_9_1
#define init_10_12_9_1
void FASTCALL call_10_12_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_9_2
#define init_10_12_9_2
void FASTCALL call_10_12_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_9_3
#define init_10_12_9_3
void FASTCALL call_10_12_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_9_4
#define init_10_12_9_4
void FASTCALL call_10_12_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_9_5
#define init_10_12_9_5
void FASTCALL call_10_12_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_9_6
#define init_10_12_9_6
void FASTCALL call_10_12_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_9_7
#define init_10_12_9_7
void FASTCALL call_10_12_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_9_8
#define init_10_12_9_8
void FASTCALL call_10_12_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_9_9
#define init_10_12_9_9
void FASTCALL call_10_12_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_9_10
#define init_10_12_9_10
void FASTCALL call_10_12_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_9_11
#define init_10_12_9_11
void FASTCALL call_10_12_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_9_12
#define init_10_12_9_12
void FASTCALL call_10_12_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_9_13
#define init_10_12_9_13
void FASTCALL call_10_12_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_9_14
#define init_10_12_9_14
void FASTCALL call_10_12_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_9_15
#define init_10_12_9_15
void FASTCALL call_10_12_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_10_0
#define init_10_12_10_0
void FASTCALL call_10_12_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_10_1
#define init_10_12_10_1
void FASTCALL call_10_12_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_10_2
#define init_10_12_10_2
void FASTCALL call_10_12_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_10_3
#define init_10_12_10_3
void FASTCALL call_10_12_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_10_4
#define init_10_12_10_4
void FASTCALL call_10_12_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_10_5
#define init_10_12_10_5
void FASTCALL call_10_12_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_10_6
#define init_10_12_10_6
void FASTCALL call_10_12_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_10_7
#define init_10_12_10_7
void FASTCALL call_10_12_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_10_8
#define init_10_12_10_8
void FASTCALL call_10_12_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_10_9
#define init_10_12_10_9
void FASTCALL call_10_12_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_10_10
#define init_10_12_10_10
void FASTCALL call_10_12_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_10_11
#define init_10_12_10_11
void FASTCALL call_10_12_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_10_12
#define init_10_12_10_12
void FASTCALL call_10_12_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_10_13
#define init_10_12_10_13
void FASTCALL call_10_12_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_10_14
#define init_10_12_10_14
void FASTCALL call_10_12_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_10_15
#define init_10_12_10_15
void FASTCALL call_10_12_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_11_0
#define init_10_12_11_0
void FASTCALL call_10_12_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_11_1
#define init_10_12_11_1
void FASTCALL call_10_12_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_11_2
#define init_10_12_11_2
void FASTCALL call_10_12_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_11_3
#define init_10_12_11_3
void FASTCALL call_10_12_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_11_4
#define init_10_12_11_4
void FASTCALL call_10_12_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_11_5
#define init_10_12_11_5
void FASTCALL call_10_12_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_11_6
#define init_10_12_11_6
void FASTCALL call_10_12_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_11_7
#define init_10_12_11_7
void FASTCALL call_10_12_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_11_8
#define init_10_12_11_8
void FASTCALL call_10_12_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_11_9
#define init_10_12_11_9
void FASTCALL call_10_12_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_11_10
#define init_10_12_11_10
void FASTCALL call_10_12_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_11_11
#define init_10_12_11_11
void FASTCALL call_10_12_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_11_12
#define init_10_12_11_12
void FASTCALL call_10_12_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_11_13
#define init_10_12_11_13
void FASTCALL call_10_12_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_11_14
#define init_10_12_11_14
void FASTCALL call_10_12_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_11_15
#define init_10_12_11_15
void FASTCALL call_10_12_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_12_0
#define init_10_12_12_0
void FASTCALL call_10_12_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_12_1
#define init_10_12_12_1
void FASTCALL call_10_12_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_12_2
#define init_10_12_12_2
void FASTCALL call_10_12_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_12_3
#define init_10_12_12_3
void FASTCALL call_10_12_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_12_4
#define init_10_12_12_4
void FASTCALL call_10_12_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_12_5
#define init_10_12_12_5
void FASTCALL call_10_12_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_12_6
#define init_10_12_12_6
void FASTCALL call_10_12_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_12_7
#define init_10_12_12_7
void FASTCALL call_10_12_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_12_8
#define init_10_12_12_8
void FASTCALL call_10_12_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_12_9
#define init_10_12_12_9
void FASTCALL call_10_12_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_12_10
#define init_10_12_12_10
void FASTCALL call_10_12_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_12_11
#define init_10_12_12_11
void FASTCALL call_10_12_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_12_12
#define init_10_12_12_12
void FASTCALL call_10_12_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_12_13
#define init_10_12_12_13
void FASTCALL call_10_12_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_12_14
#define init_10_12_12_14
void FASTCALL call_10_12_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_12_15
#define init_10_12_12_15
void FASTCALL call_10_12_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_13_0
#define init_10_12_13_0
void FASTCALL call_10_12_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_13_1
#define init_10_12_13_1
void FASTCALL call_10_12_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_13_2
#define init_10_12_13_2
void FASTCALL call_10_12_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_13_3
#define init_10_12_13_3
void FASTCALL call_10_12_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_13_4
#define init_10_12_13_4
void FASTCALL call_10_12_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_13_5
#define init_10_12_13_5
void FASTCALL call_10_12_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_13_6
#define init_10_12_13_6
void FASTCALL call_10_12_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_13_7
#define init_10_12_13_7
void FASTCALL call_10_12_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_13_8
#define init_10_12_13_8
void FASTCALL call_10_12_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_13_9
#define init_10_12_13_9
void FASTCALL call_10_12_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_13_10
#define init_10_12_13_10
void FASTCALL call_10_12_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_13_11
#define init_10_12_13_11
void FASTCALL call_10_12_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_13_12
#define init_10_12_13_12
void FASTCALL call_10_12_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_13_13
#define init_10_12_13_13
void FASTCALL call_10_12_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_13_14
#define init_10_12_13_14
void FASTCALL call_10_12_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_13_15
#define init_10_12_13_15
void FASTCALL call_10_12_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_14_0
#define init_10_12_14_0
void FASTCALL call_10_12_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_14_1
#define init_10_12_14_1
void FASTCALL call_10_12_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_14_2
#define init_10_12_14_2
void FASTCALL call_10_12_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_14_3
#define init_10_12_14_3
void FASTCALL call_10_12_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_14_4
#define init_10_12_14_4
void FASTCALL call_10_12_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_14_5
#define init_10_12_14_5
void FASTCALL call_10_12_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_14_6
#define init_10_12_14_6
void FASTCALL call_10_12_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_14_7
#define init_10_12_14_7
void FASTCALL call_10_12_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_14_8
#define init_10_12_14_8
void FASTCALL call_10_12_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_14_9
#define init_10_12_14_9
void FASTCALL call_10_12_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_14_10
#define init_10_12_14_10
void FASTCALL call_10_12_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_14_11
#define init_10_12_14_11
void FASTCALL call_10_12_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_14_12
#define init_10_12_14_12
void FASTCALL call_10_12_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_14_13
#define init_10_12_14_13
void FASTCALL call_10_12_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_14_14
#define init_10_12_14_14
void FASTCALL call_10_12_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_14_15
#define init_10_12_14_15
void FASTCALL call_10_12_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_15_0
#define init_10_12_15_0
void FASTCALL call_10_12_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_15_1
#define init_10_12_15_1
void FASTCALL call_10_12_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_15_2
#define init_10_12_15_2
void FASTCALL call_10_12_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_15_3
#define init_10_12_15_3
void FASTCALL call_10_12_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_15_4
#define init_10_12_15_4
void FASTCALL call_10_12_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_15_5
#define init_10_12_15_5
void FASTCALL call_10_12_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_15_6
#define init_10_12_15_6
void FASTCALL call_10_12_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_15_7
#define init_10_12_15_7
void FASTCALL call_10_12_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_15_8
#define init_10_12_15_8
void FASTCALL call_10_12_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_15_9
#define init_10_12_15_9
void FASTCALL call_10_12_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_15_10
#define init_10_12_15_10
void FASTCALL call_10_12_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_15_11
#define init_10_12_15_11
void FASTCALL call_10_12_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_15_12
#define init_10_12_15_12
void FASTCALL call_10_12_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_15_13
#define init_10_12_15_13
void FASTCALL call_10_12_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_15_14
#define init_10_12_15_14
void FASTCALL call_10_12_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_12_15_15
#define init_10_12_15_15
void FASTCALL call_10_12_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_0_0
#define init_10_13_0_0
void FASTCALL call_10_13_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_0_1
#define init_10_13_0_1
void FASTCALL call_10_13_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_0_2
#define init_10_13_0_2
void FASTCALL call_10_13_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_0_3
#define init_10_13_0_3
void FASTCALL call_10_13_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_0_4
#define init_10_13_0_4
void FASTCALL call_10_13_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_0_5
#define init_10_13_0_5
void FASTCALL call_10_13_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_0_6
#define init_10_13_0_6
void FASTCALL call_10_13_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_0_7
#define init_10_13_0_7
void FASTCALL call_10_13_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_0_8
#define init_10_13_0_8
void FASTCALL call_10_13_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_0_9
#define init_10_13_0_9
void FASTCALL call_10_13_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_0_10
#define init_10_13_0_10
void FASTCALL call_10_13_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_0_11
#define init_10_13_0_11
void FASTCALL call_10_13_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_0_12
#define init_10_13_0_12
void FASTCALL call_10_13_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_0_13
#define init_10_13_0_13
void FASTCALL call_10_13_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_0_14
#define init_10_13_0_14
void FASTCALL call_10_13_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_0_15
#define init_10_13_0_15
void FASTCALL call_10_13_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_1_0
#define init_10_13_1_0
void FASTCALL call_10_13_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_1_1
#define init_10_13_1_1
void FASTCALL call_10_13_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_1_2
#define init_10_13_1_2
void FASTCALL call_10_13_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_1_3
#define init_10_13_1_3
void FASTCALL call_10_13_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_1_4
#define init_10_13_1_4
void FASTCALL call_10_13_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_1_5
#define init_10_13_1_5
void FASTCALL call_10_13_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_1_6
#define init_10_13_1_6
void FASTCALL call_10_13_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_1_7
#define init_10_13_1_7
void FASTCALL call_10_13_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_1_8
#define init_10_13_1_8
void FASTCALL call_10_13_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_1_9
#define init_10_13_1_9
void FASTCALL call_10_13_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_1_10
#define init_10_13_1_10
void FASTCALL call_10_13_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_1_11
#define init_10_13_1_11
void FASTCALL call_10_13_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_1_12
#define init_10_13_1_12
void FASTCALL call_10_13_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_1_13
#define init_10_13_1_13
void FASTCALL call_10_13_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_1_14
#define init_10_13_1_14
void FASTCALL call_10_13_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_1_15
#define init_10_13_1_15
void FASTCALL call_10_13_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_2_0
#define init_10_13_2_0
void FASTCALL call_10_13_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_2_1
#define init_10_13_2_1
void FASTCALL call_10_13_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_2_2
#define init_10_13_2_2
void FASTCALL call_10_13_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_2_3
#define init_10_13_2_3
void FASTCALL call_10_13_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_2_4
#define init_10_13_2_4
void FASTCALL call_10_13_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_2_5
#define init_10_13_2_5
void FASTCALL call_10_13_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_2_6
#define init_10_13_2_6
void FASTCALL call_10_13_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_2_7
#define init_10_13_2_7
void FASTCALL call_10_13_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_2_8
#define init_10_13_2_8
void FASTCALL call_10_13_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_2_9
#define init_10_13_2_9
void FASTCALL call_10_13_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_2_10
#define init_10_13_2_10
void FASTCALL call_10_13_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_2_11
#define init_10_13_2_11
void FASTCALL call_10_13_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_2_12
#define init_10_13_2_12
void FASTCALL call_10_13_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_2_13
#define init_10_13_2_13
void FASTCALL call_10_13_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_2_14
#define init_10_13_2_14
void FASTCALL call_10_13_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_2_15
#define init_10_13_2_15
void FASTCALL call_10_13_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_3_0
#define init_10_13_3_0
void FASTCALL call_10_13_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_3_1
#define init_10_13_3_1
void FASTCALL call_10_13_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_3_2
#define init_10_13_3_2
void FASTCALL call_10_13_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_3_3
#define init_10_13_3_3
void FASTCALL call_10_13_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_3_4
#define init_10_13_3_4
void FASTCALL call_10_13_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_3_5
#define init_10_13_3_5
void FASTCALL call_10_13_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_3_6
#define init_10_13_3_6
void FASTCALL call_10_13_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_3_7
#define init_10_13_3_7
void FASTCALL call_10_13_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_3_8
#define init_10_13_3_8
void FASTCALL call_10_13_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_3_9
#define init_10_13_3_9
void FASTCALL call_10_13_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_3_10
#define init_10_13_3_10
void FASTCALL call_10_13_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_3_11
#define init_10_13_3_11
void FASTCALL call_10_13_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_3_12
#define init_10_13_3_12
void FASTCALL call_10_13_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_3_13
#define init_10_13_3_13
void FASTCALL call_10_13_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_3_14
#define init_10_13_3_14
void FASTCALL call_10_13_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_3_15
#define init_10_13_3_15
void FASTCALL call_10_13_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_4_0
#define init_10_13_4_0
void FASTCALL call_10_13_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_4_1
#define init_10_13_4_1
void FASTCALL call_10_13_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_4_2
#define init_10_13_4_2
void FASTCALL call_10_13_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_4_3
#define init_10_13_4_3
void FASTCALL call_10_13_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_4_4
#define init_10_13_4_4
void FASTCALL call_10_13_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_4_5
#define init_10_13_4_5
void FASTCALL call_10_13_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_4_6
#define init_10_13_4_6
void FASTCALL call_10_13_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_4_7
#define init_10_13_4_7
void FASTCALL call_10_13_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_4_8
#define init_10_13_4_8
void FASTCALL call_10_13_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_4_9
#define init_10_13_4_9
void FASTCALL call_10_13_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_4_10
#define init_10_13_4_10
void FASTCALL call_10_13_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_4_11
#define init_10_13_4_11
void FASTCALL call_10_13_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_4_12
#define init_10_13_4_12
void FASTCALL call_10_13_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_4_13
#define init_10_13_4_13
void FASTCALL call_10_13_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_4_14
#define init_10_13_4_14
void FASTCALL call_10_13_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_4_15
#define init_10_13_4_15
void FASTCALL call_10_13_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_5_0
#define init_10_13_5_0
void FASTCALL call_10_13_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_5_1
#define init_10_13_5_1
void FASTCALL call_10_13_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_5_2
#define init_10_13_5_2
void FASTCALL call_10_13_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_5_3
#define init_10_13_5_3
void FASTCALL call_10_13_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_5_4
#define init_10_13_5_4
void FASTCALL call_10_13_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_5_5
#define init_10_13_5_5
void FASTCALL call_10_13_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_5_6
#define init_10_13_5_6
void FASTCALL call_10_13_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_5_7
#define init_10_13_5_7
void FASTCALL call_10_13_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_5_8
#define init_10_13_5_8
void FASTCALL call_10_13_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_5_9
#define init_10_13_5_9
void FASTCALL call_10_13_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_5_10
#define init_10_13_5_10
void FASTCALL call_10_13_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_5_11
#define init_10_13_5_11
void FASTCALL call_10_13_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_5_12
#define init_10_13_5_12
void FASTCALL call_10_13_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_5_13
#define init_10_13_5_13
void FASTCALL call_10_13_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_5_14
#define init_10_13_5_14
void FASTCALL call_10_13_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_5_15
#define init_10_13_5_15
void FASTCALL call_10_13_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_6_0
#define init_10_13_6_0
void FASTCALL call_10_13_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_6_1
#define init_10_13_6_1
void FASTCALL call_10_13_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_6_2
#define init_10_13_6_2
void FASTCALL call_10_13_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_6_3
#define init_10_13_6_3
void FASTCALL call_10_13_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_6_4
#define init_10_13_6_4
void FASTCALL call_10_13_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_6_5
#define init_10_13_6_5
void FASTCALL call_10_13_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_6_6
#define init_10_13_6_6
void FASTCALL call_10_13_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_6_7
#define init_10_13_6_7
void FASTCALL call_10_13_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_6_8
#define init_10_13_6_8
void FASTCALL call_10_13_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_6_9
#define init_10_13_6_9
void FASTCALL call_10_13_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_6_10
#define init_10_13_6_10
void FASTCALL call_10_13_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_6_11
#define init_10_13_6_11
void FASTCALL call_10_13_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_6_12
#define init_10_13_6_12
void FASTCALL call_10_13_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_6_13
#define init_10_13_6_13
void FASTCALL call_10_13_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_6_14
#define init_10_13_6_14
void FASTCALL call_10_13_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_6_15
#define init_10_13_6_15
void FASTCALL call_10_13_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_7_0
#define init_10_13_7_0
void FASTCALL call_10_13_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_7_1
#define init_10_13_7_1
void FASTCALL call_10_13_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_7_2
#define init_10_13_7_2
void FASTCALL call_10_13_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_7_3
#define init_10_13_7_3
void FASTCALL call_10_13_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_7_4
#define init_10_13_7_4
void FASTCALL call_10_13_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_7_5
#define init_10_13_7_5
void FASTCALL call_10_13_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_7_6
#define init_10_13_7_6
void FASTCALL call_10_13_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_7_7
#define init_10_13_7_7
void FASTCALL call_10_13_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_7_8
#define init_10_13_7_8
void FASTCALL call_10_13_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_7_9
#define init_10_13_7_9
void FASTCALL call_10_13_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_7_10
#define init_10_13_7_10
void FASTCALL call_10_13_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_7_11
#define init_10_13_7_11
void FASTCALL call_10_13_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_7_12
#define init_10_13_7_12
void FASTCALL call_10_13_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_7_13
#define init_10_13_7_13
void FASTCALL call_10_13_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_7_14
#define init_10_13_7_14
void FASTCALL call_10_13_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_7_15
#define init_10_13_7_15
void FASTCALL call_10_13_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_8_0
#define init_10_13_8_0
void FASTCALL call_10_13_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_8_1
#define init_10_13_8_1
void FASTCALL call_10_13_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_8_2
#define init_10_13_8_2
void FASTCALL call_10_13_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_8_3
#define init_10_13_8_3
void FASTCALL call_10_13_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_8_4
#define init_10_13_8_4
void FASTCALL call_10_13_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_8_5
#define init_10_13_8_5
void FASTCALL call_10_13_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_8_6
#define init_10_13_8_6
void FASTCALL call_10_13_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_8_7
#define init_10_13_8_7
void FASTCALL call_10_13_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_8_8
#define init_10_13_8_8
void FASTCALL call_10_13_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_8_9
#define init_10_13_8_9
void FASTCALL call_10_13_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_8_10
#define init_10_13_8_10
void FASTCALL call_10_13_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_8_11
#define init_10_13_8_11
void FASTCALL call_10_13_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_8_12
#define init_10_13_8_12
void FASTCALL call_10_13_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_8_13
#define init_10_13_8_13
void FASTCALL call_10_13_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_8_14
#define init_10_13_8_14
void FASTCALL call_10_13_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_8_15
#define init_10_13_8_15
void FASTCALL call_10_13_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_9_0
#define init_10_13_9_0
void FASTCALL call_10_13_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_9_1
#define init_10_13_9_1
void FASTCALL call_10_13_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_9_2
#define init_10_13_9_2
void FASTCALL call_10_13_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_9_3
#define init_10_13_9_3
void FASTCALL call_10_13_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_9_4
#define init_10_13_9_4
void FASTCALL call_10_13_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_9_5
#define init_10_13_9_5
void FASTCALL call_10_13_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_9_6
#define init_10_13_9_6
void FASTCALL call_10_13_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_9_7
#define init_10_13_9_7
void FASTCALL call_10_13_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_9_8
#define init_10_13_9_8
void FASTCALL call_10_13_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_9_9
#define init_10_13_9_9
void FASTCALL call_10_13_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_9_10
#define init_10_13_9_10
void FASTCALL call_10_13_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_9_11
#define init_10_13_9_11
void FASTCALL call_10_13_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_9_12
#define init_10_13_9_12
void FASTCALL call_10_13_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_9_13
#define init_10_13_9_13
void FASTCALL call_10_13_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_9_14
#define init_10_13_9_14
void FASTCALL call_10_13_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_9_15
#define init_10_13_9_15
void FASTCALL call_10_13_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_10_0
#define init_10_13_10_0
void FASTCALL call_10_13_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_10_1
#define init_10_13_10_1
void FASTCALL call_10_13_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_10_2
#define init_10_13_10_2
void FASTCALL call_10_13_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_10_3
#define init_10_13_10_3
void FASTCALL call_10_13_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_10_4
#define init_10_13_10_4
void FASTCALL call_10_13_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_10_5
#define init_10_13_10_5
void FASTCALL call_10_13_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_10_6
#define init_10_13_10_6
void FASTCALL call_10_13_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_10_7
#define init_10_13_10_7
void FASTCALL call_10_13_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_10_8
#define init_10_13_10_8
void FASTCALL call_10_13_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_10_9
#define init_10_13_10_9
void FASTCALL call_10_13_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_10_10
#define init_10_13_10_10
void FASTCALL call_10_13_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_10_11
#define init_10_13_10_11
void FASTCALL call_10_13_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_10_12
#define init_10_13_10_12
void FASTCALL call_10_13_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_10_13
#define init_10_13_10_13
void FASTCALL call_10_13_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_10_14
#define init_10_13_10_14
void FASTCALL call_10_13_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_10_15
#define init_10_13_10_15
void FASTCALL call_10_13_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_11_0
#define init_10_13_11_0
void FASTCALL call_10_13_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_11_1
#define init_10_13_11_1
void FASTCALL call_10_13_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_11_2
#define init_10_13_11_2
void FASTCALL call_10_13_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_11_3
#define init_10_13_11_3
void FASTCALL call_10_13_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_11_4
#define init_10_13_11_4
void FASTCALL call_10_13_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_11_5
#define init_10_13_11_5
void FASTCALL call_10_13_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_11_6
#define init_10_13_11_6
void FASTCALL call_10_13_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_11_7
#define init_10_13_11_7
void FASTCALL call_10_13_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_11_8
#define init_10_13_11_8
void FASTCALL call_10_13_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_11_9
#define init_10_13_11_9
void FASTCALL call_10_13_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_11_10
#define init_10_13_11_10
void FASTCALL call_10_13_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_11_11
#define init_10_13_11_11
void FASTCALL call_10_13_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_11_12
#define init_10_13_11_12
void FASTCALL call_10_13_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_11_13
#define init_10_13_11_13
void FASTCALL call_10_13_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_11_14
#define init_10_13_11_14
void FASTCALL call_10_13_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_11_15
#define init_10_13_11_15
void FASTCALL call_10_13_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_12_0
#define init_10_13_12_0
void FASTCALL call_10_13_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_12_1
#define init_10_13_12_1
void FASTCALL call_10_13_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_12_2
#define init_10_13_12_2
void FASTCALL call_10_13_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_12_3
#define init_10_13_12_3
void FASTCALL call_10_13_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_12_4
#define init_10_13_12_4
void FASTCALL call_10_13_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_12_5
#define init_10_13_12_5
void FASTCALL call_10_13_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_12_6
#define init_10_13_12_6
void FASTCALL call_10_13_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_12_7
#define init_10_13_12_7
void FASTCALL call_10_13_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_12_8
#define init_10_13_12_8
void FASTCALL call_10_13_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_12_9
#define init_10_13_12_9
void FASTCALL call_10_13_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_12_10
#define init_10_13_12_10
void FASTCALL call_10_13_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_12_11
#define init_10_13_12_11
void FASTCALL call_10_13_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_12_12
#define init_10_13_12_12
void FASTCALL call_10_13_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_12_13
#define init_10_13_12_13
void FASTCALL call_10_13_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_12_14
#define init_10_13_12_14
void FASTCALL call_10_13_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_12_15
#define init_10_13_12_15
void FASTCALL call_10_13_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_13_0
#define init_10_13_13_0
void FASTCALL call_10_13_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_13_1
#define init_10_13_13_1
void FASTCALL call_10_13_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_13_2
#define init_10_13_13_2
void FASTCALL call_10_13_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_13_3
#define init_10_13_13_3
void FASTCALL call_10_13_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_13_4
#define init_10_13_13_4
void FASTCALL call_10_13_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_13_5
#define init_10_13_13_5
void FASTCALL call_10_13_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_13_6
#define init_10_13_13_6
void FASTCALL call_10_13_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_13_7
#define init_10_13_13_7
void FASTCALL call_10_13_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_13_8
#define init_10_13_13_8
void FASTCALL call_10_13_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_13_9
#define init_10_13_13_9
void FASTCALL call_10_13_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_13_10
#define init_10_13_13_10
void FASTCALL call_10_13_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_13_11
#define init_10_13_13_11
void FASTCALL call_10_13_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_13_12
#define init_10_13_13_12
void FASTCALL call_10_13_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_13_13
#define init_10_13_13_13
void FASTCALL call_10_13_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_13_14
#define init_10_13_13_14
void FASTCALL call_10_13_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_13_15
#define init_10_13_13_15
void FASTCALL call_10_13_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_14_0
#define init_10_13_14_0
void FASTCALL call_10_13_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_14_1
#define init_10_13_14_1
void FASTCALL call_10_13_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_14_2
#define init_10_13_14_2
void FASTCALL call_10_13_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_14_3
#define init_10_13_14_3
void FASTCALL call_10_13_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_14_4
#define init_10_13_14_4
void FASTCALL call_10_13_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_14_5
#define init_10_13_14_5
void FASTCALL call_10_13_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_14_6
#define init_10_13_14_6
void FASTCALL call_10_13_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_14_7
#define init_10_13_14_7
void FASTCALL call_10_13_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_14_8
#define init_10_13_14_8
void FASTCALL call_10_13_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_14_9
#define init_10_13_14_9
void FASTCALL call_10_13_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_14_10
#define init_10_13_14_10
void FASTCALL call_10_13_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_14_11
#define init_10_13_14_11
void FASTCALL call_10_13_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_14_12
#define init_10_13_14_12
void FASTCALL call_10_13_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_14_13
#define init_10_13_14_13
void FASTCALL call_10_13_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_14_14
#define init_10_13_14_14
void FASTCALL call_10_13_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_14_15
#define init_10_13_14_15
void FASTCALL call_10_13_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_15_0
#define init_10_13_15_0
void FASTCALL call_10_13_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_15_1
#define init_10_13_15_1
void FASTCALL call_10_13_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_15_2
#define init_10_13_15_2
void FASTCALL call_10_13_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_15_3
#define init_10_13_15_3
void FASTCALL call_10_13_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_15_4
#define init_10_13_15_4
void FASTCALL call_10_13_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_15_5
#define init_10_13_15_5
void FASTCALL call_10_13_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_15_6
#define init_10_13_15_6
void FASTCALL call_10_13_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_15_7
#define init_10_13_15_7
void FASTCALL call_10_13_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_15_8
#define init_10_13_15_8
void FASTCALL call_10_13_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_15_9
#define init_10_13_15_9
void FASTCALL call_10_13_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_15_10
#define init_10_13_15_10
void FASTCALL call_10_13_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_15_11
#define init_10_13_15_11
void FASTCALL call_10_13_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_15_12
#define init_10_13_15_12
void FASTCALL call_10_13_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_15_13
#define init_10_13_15_13
void FASTCALL call_10_13_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_15_14
#define init_10_13_15_14
void FASTCALL call_10_13_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_13_15_15
#define init_10_13_15_15
void FASTCALL call_10_13_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_0_0
#define init_10_14_0_0
void FASTCALL call_10_14_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_0_1
#define init_10_14_0_1
void FASTCALL call_10_14_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_0_2
#define init_10_14_0_2
void FASTCALL call_10_14_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_0_3
#define init_10_14_0_3
void FASTCALL call_10_14_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_0_4
#define init_10_14_0_4
void FASTCALL call_10_14_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_0_5
#define init_10_14_0_5
void FASTCALL call_10_14_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_0_6
#define init_10_14_0_6
void FASTCALL call_10_14_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_0_7
#define init_10_14_0_7
void FASTCALL call_10_14_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_0_8
#define init_10_14_0_8
void FASTCALL call_10_14_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_0_9
#define init_10_14_0_9
void FASTCALL call_10_14_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_0_10
#define init_10_14_0_10
void FASTCALL call_10_14_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_0_11
#define init_10_14_0_11
void FASTCALL call_10_14_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_0_12
#define init_10_14_0_12
void FASTCALL call_10_14_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_0_13
#define init_10_14_0_13
void FASTCALL call_10_14_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_0_14
#define init_10_14_0_14
void FASTCALL call_10_14_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_0_15
#define init_10_14_0_15
void FASTCALL call_10_14_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_1_0
#define init_10_14_1_0
void FASTCALL call_10_14_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_1_1
#define init_10_14_1_1
void FASTCALL call_10_14_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_1_2
#define init_10_14_1_2
void FASTCALL call_10_14_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_1_3
#define init_10_14_1_3
void FASTCALL call_10_14_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_1_4
#define init_10_14_1_4
void FASTCALL call_10_14_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_1_5
#define init_10_14_1_5
void FASTCALL call_10_14_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_1_6
#define init_10_14_1_6
void FASTCALL call_10_14_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_1_7
#define init_10_14_1_7
void FASTCALL call_10_14_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_1_8
#define init_10_14_1_8
void FASTCALL call_10_14_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_1_9
#define init_10_14_1_9
void FASTCALL call_10_14_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_1_10
#define init_10_14_1_10
void FASTCALL call_10_14_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_1_11
#define init_10_14_1_11
void FASTCALL call_10_14_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_1_12
#define init_10_14_1_12
void FASTCALL call_10_14_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_1_13
#define init_10_14_1_13
void FASTCALL call_10_14_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_1_14
#define init_10_14_1_14
void FASTCALL call_10_14_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_1_15
#define init_10_14_1_15
void FASTCALL call_10_14_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_2_0
#define init_10_14_2_0
void FASTCALL call_10_14_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_2_1
#define init_10_14_2_1
void FASTCALL call_10_14_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_2_2
#define init_10_14_2_2
void FASTCALL call_10_14_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_2_3
#define init_10_14_2_3
void FASTCALL call_10_14_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_2_4
#define init_10_14_2_4
void FASTCALL call_10_14_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_2_5
#define init_10_14_2_5
void FASTCALL call_10_14_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_2_6
#define init_10_14_2_6
void FASTCALL call_10_14_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_2_7
#define init_10_14_2_7
void FASTCALL call_10_14_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_2_8
#define init_10_14_2_8
void FASTCALL call_10_14_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_2_9
#define init_10_14_2_9
void FASTCALL call_10_14_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_2_10
#define init_10_14_2_10
void FASTCALL call_10_14_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_2_11
#define init_10_14_2_11
void FASTCALL call_10_14_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_2_12
#define init_10_14_2_12
void FASTCALL call_10_14_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_2_13
#define init_10_14_2_13
void FASTCALL call_10_14_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_2_14
#define init_10_14_2_14
void FASTCALL call_10_14_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_2_15
#define init_10_14_2_15
void FASTCALL call_10_14_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_3_0
#define init_10_14_3_0
void FASTCALL call_10_14_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_3_1
#define init_10_14_3_1
void FASTCALL call_10_14_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_3_2
#define init_10_14_3_2
void FASTCALL call_10_14_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_3_3
#define init_10_14_3_3
void FASTCALL call_10_14_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_3_4
#define init_10_14_3_4
void FASTCALL call_10_14_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_3_5
#define init_10_14_3_5
void FASTCALL call_10_14_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_3_6
#define init_10_14_3_6
void FASTCALL call_10_14_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_3_7
#define init_10_14_3_7
void FASTCALL call_10_14_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_3_8
#define init_10_14_3_8
void FASTCALL call_10_14_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_3_9
#define init_10_14_3_9
void FASTCALL call_10_14_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_3_10
#define init_10_14_3_10
void FASTCALL call_10_14_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_3_11
#define init_10_14_3_11
void FASTCALL call_10_14_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_3_12
#define init_10_14_3_12
void FASTCALL call_10_14_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_3_13
#define init_10_14_3_13
void FASTCALL call_10_14_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_3_14
#define init_10_14_3_14
void FASTCALL call_10_14_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_3_15
#define init_10_14_3_15
void FASTCALL call_10_14_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_4_0
#define init_10_14_4_0
void FASTCALL call_10_14_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_4_1
#define init_10_14_4_1
void FASTCALL call_10_14_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_4_2
#define init_10_14_4_2
void FASTCALL call_10_14_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_4_3
#define init_10_14_4_3
void FASTCALL call_10_14_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_4_4
#define init_10_14_4_4
void FASTCALL call_10_14_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_4_5
#define init_10_14_4_5
void FASTCALL call_10_14_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_4_6
#define init_10_14_4_6
void FASTCALL call_10_14_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_4_7
#define init_10_14_4_7
void FASTCALL call_10_14_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_4_8
#define init_10_14_4_8
void FASTCALL call_10_14_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_4_9
#define init_10_14_4_9
void FASTCALL call_10_14_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_4_10
#define init_10_14_4_10
void FASTCALL call_10_14_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_4_11
#define init_10_14_4_11
void FASTCALL call_10_14_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_4_12
#define init_10_14_4_12
void FASTCALL call_10_14_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_4_13
#define init_10_14_4_13
void FASTCALL call_10_14_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_4_14
#define init_10_14_4_14
void FASTCALL call_10_14_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_4_15
#define init_10_14_4_15
void FASTCALL call_10_14_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_5_0
#define init_10_14_5_0
void FASTCALL call_10_14_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_5_1
#define init_10_14_5_1
void FASTCALL call_10_14_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_5_2
#define init_10_14_5_2
void FASTCALL call_10_14_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_5_3
#define init_10_14_5_3
void FASTCALL call_10_14_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_5_4
#define init_10_14_5_4
void FASTCALL call_10_14_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_5_5
#define init_10_14_5_5
void FASTCALL call_10_14_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_5_6
#define init_10_14_5_6
void FASTCALL call_10_14_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_5_7
#define init_10_14_5_7
void FASTCALL call_10_14_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_5_8
#define init_10_14_5_8
void FASTCALL call_10_14_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_5_9
#define init_10_14_5_9
void FASTCALL call_10_14_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_5_10
#define init_10_14_5_10
void FASTCALL call_10_14_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_5_11
#define init_10_14_5_11
void FASTCALL call_10_14_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_5_12
#define init_10_14_5_12
void FASTCALL call_10_14_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_5_13
#define init_10_14_5_13
void FASTCALL call_10_14_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_5_14
#define init_10_14_5_14
void FASTCALL call_10_14_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_5_15
#define init_10_14_5_15
void FASTCALL call_10_14_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_6_0
#define init_10_14_6_0
void FASTCALL call_10_14_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_6_1
#define init_10_14_6_1
void FASTCALL call_10_14_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_6_2
#define init_10_14_6_2
void FASTCALL call_10_14_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_6_3
#define init_10_14_6_3
void FASTCALL call_10_14_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_6_4
#define init_10_14_6_4
void FASTCALL call_10_14_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_6_5
#define init_10_14_6_5
void FASTCALL call_10_14_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_6_6
#define init_10_14_6_6
void FASTCALL call_10_14_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_6_7
#define init_10_14_6_7
void FASTCALL call_10_14_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_6_8
#define init_10_14_6_8
void FASTCALL call_10_14_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_6_9
#define init_10_14_6_9
void FASTCALL call_10_14_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_6_10
#define init_10_14_6_10
void FASTCALL call_10_14_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_6_11
#define init_10_14_6_11
void FASTCALL call_10_14_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_6_12
#define init_10_14_6_12
void FASTCALL call_10_14_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_6_13
#define init_10_14_6_13
void FASTCALL call_10_14_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_6_14
#define init_10_14_6_14
void FASTCALL call_10_14_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_6_15
#define init_10_14_6_15
void FASTCALL call_10_14_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_7_0
#define init_10_14_7_0
void FASTCALL call_10_14_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_7_1
#define init_10_14_7_1
void FASTCALL call_10_14_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_7_2
#define init_10_14_7_2
void FASTCALL call_10_14_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_7_3
#define init_10_14_7_3
void FASTCALL call_10_14_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_7_4
#define init_10_14_7_4
void FASTCALL call_10_14_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_7_5
#define init_10_14_7_5
void FASTCALL call_10_14_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_7_6
#define init_10_14_7_6
void FASTCALL call_10_14_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_7_7
#define init_10_14_7_7
void FASTCALL call_10_14_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_7_8
#define init_10_14_7_8
void FASTCALL call_10_14_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_7_9
#define init_10_14_7_9
void FASTCALL call_10_14_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_7_10
#define init_10_14_7_10
void FASTCALL call_10_14_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_7_11
#define init_10_14_7_11
void FASTCALL call_10_14_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_7_12
#define init_10_14_7_12
void FASTCALL call_10_14_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_7_13
#define init_10_14_7_13
void FASTCALL call_10_14_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_7_14
#define init_10_14_7_14
void FASTCALL call_10_14_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_7_15
#define init_10_14_7_15
void FASTCALL call_10_14_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_8_0
#define init_10_14_8_0
void FASTCALL call_10_14_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_8_1
#define init_10_14_8_1
void FASTCALL call_10_14_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_8_2
#define init_10_14_8_2
void FASTCALL call_10_14_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_8_3
#define init_10_14_8_3
void FASTCALL call_10_14_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_8_4
#define init_10_14_8_4
void FASTCALL call_10_14_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_8_5
#define init_10_14_8_5
void FASTCALL call_10_14_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_8_6
#define init_10_14_8_6
void FASTCALL call_10_14_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_8_7
#define init_10_14_8_7
void FASTCALL call_10_14_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_8_8
#define init_10_14_8_8
void FASTCALL call_10_14_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_8_9
#define init_10_14_8_9
void FASTCALL call_10_14_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_8_10
#define init_10_14_8_10
void FASTCALL call_10_14_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_8_11
#define init_10_14_8_11
void FASTCALL call_10_14_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_8_12
#define init_10_14_8_12
void FASTCALL call_10_14_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_8_13
#define init_10_14_8_13
void FASTCALL call_10_14_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_8_14
#define init_10_14_8_14
void FASTCALL call_10_14_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_8_15
#define init_10_14_8_15
void FASTCALL call_10_14_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_9_0
#define init_10_14_9_0
void FASTCALL call_10_14_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_9_1
#define init_10_14_9_1
void FASTCALL call_10_14_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_9_2
#define init_10_14_9_2
void FASTCALL call_10_14_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_9_3
#define init_10_14_9_3
void FASTCALL call_10_14_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_9_4
#define init_10_14_9_4
void FASTCALL call_10_14_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_9_5
#define init_10_14_9_5
void FASTCALL call_10_14_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_9_6
#define init_10_14_9_6
void FASTCALL call_10_14_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_9_7
#define init_10_14_9_7
void FASTCALL call_10_14_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_9_8
#define init_10_14_9_8
void FASTCALL call_10_14_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_9_9
#define init_10_14_9_9
void FASTCALL call_10_14_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_9_10
#define init_10_14_9_10
void FASTCALL call_10_14_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_9_11
#define init_10_14_9_11
void FASTCALL call_10_14_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_9_12
#define init_10_14_9_12
void FASTCALL call_10_14_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_9_13
#define init_10_14_9_13
void FASTCALL call_10_14_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_9_14
#define init_10_14_9_14
void FASTCALL call_10_14_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_9_15
#define init_10_14_9_15
void FASTCALL call_10_14_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_10_0
#define init_10_14_10_0
void FASTCALL call_10_14_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_10_1
#define init_10_14_10_1
void FASTCALL call_10_14_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_10_2
#define init_10_14_10_2
void FASTCALL call_10_14_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_10_3
#define init_10_14_10_3
void FASTCALL call_10_14_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_10_4
#define init_10_14_10_4
void FASTCALL call_10_14_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_10_5
#define init_10_14_10_5
void FASTCALL call_10_14_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_10_6
#define init_10_14_10_6
void FASTCALL call_10_14_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_10_7
#define init_10_14_10_7
void FASTCALL call_10_14_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_10_8
#define init_10_14_10_8
void FASTCALL call_10_14_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_10_9
#define init_10_14_10_9
void FASTCALL call_10_14_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_10_10
#define init_10_14_10_10
void FASTCALL call_10_14_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_10_11
#define init_10_14_10_11
void FASTCALL call_10_14_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_10_12
#define init_10_14_10_12
void FASTCALL call_10_14_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_10_13
#define init_10_14_10_13
void FASTCALL call_10_14_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_10_14
#define init_10_14_10_14
void FASTCALL call_10_14_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_10_15
#define init_10_14_10_15
void FASTCALL call_10_14_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_11_0
#define init_10_14_11_0
void FASTCALL call_10_14_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_11_1
#define init_10_14_11_1
void FASTCALL call_10_14_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_11_2
#define init_10_14_11_2
void FASTCALL call_10_14_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_11_3
#define init_10_14_11_3
void FASTCALL call_10_14_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_11_4
#define init_10_14_11_4
void FASTCALL call_10_14_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_11_5
#define init_10_14_11_5
void FASTCALL call_10_14_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_11_6
#define init_10_14_11_6
void FASTCALL call_10_14_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_11_7
#define init_10_14_11_7
void FASTCALL call_10_14_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_11_8
#define init_10_14_11_8
void FASTCALL call_10_14_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_11_9
#define init_10_14_11_9
void FASTCALL call_10_14_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_11_10
#define init_10_14_11_10
void FASTCALL call_10_14_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_11_11
#define init_10_14_11_11
void FASTCALL call_10_14_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_11_12
#define init_10_14_11_12
void FASTCALL call_10_14_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_11_13
#define init_10_14_11_13
void FASTCALL call_10_14_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_11_14
#define init_10_14_11_14
void FASTCALL call_10_14_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_11_15
#define init_10_14_11_15
void FASTCALL call_10_14_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_12_0
#define init_10_14_12_0
void FASTCALL call_10_14_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_12_1
#define init_10_14_12_1
void FASTCALL call_10_14_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_12_2
#define init_10_14_12_2
void FASTCALL call_10_14_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_12_3
#define init_10_14_12_3
void FASTCALL call_10_14_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_12_4
#define init_10_14_12_4
void FASTCALL call_10_14_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_12_5
#define init_10_14_12_5
void FASTCALL call_10_14_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_12_6
#define init_10_14_12_6
void FASTCALL call_10_14_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_12_7
#define init_10_14_12_7
void FASTCALL call_10_14_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_12_8
#define init_10_14_12_8
void FASTCALL call_10_14_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_12_9
#define init_10_14_12_9
void FASTCALL call_10_14_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_12_10
#define init_10_14_12_10
void FASTCALL call_10_14_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_12_11
#define init_10_14_12_11
void FASTCALL call_10_14_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_12_12
#define init_10_14_12_12
void FASTCALL call_10_14_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_12_13
#define init_10_14_12_13
void FASTCALL call_10_14_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_12_14
#define init_10_14_12_14
void FASTCALL call_10_14_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_12_15
#define init_10_14_12_15
void FASTCALL call_10_14_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_13_0
#define init_10_14_13_0
void FASTCALL call_10_14_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_13_1
#define init_10_14_13_1
void FASTCALL call_10_14_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_13_2
#define init_10_14_13_2
void FASTCALL call_10_14_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_13_3
#define init_10_14_13_3
void FASTCALL call_10_14_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_13_4
#define init_10_14_13_4
void FASTCALL call_10_14_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_13_5
#define init_10_14_13_5
void FASTCALL call_10_14_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_13_6
#define init_10_14_13_6
void FASTCALL call_10_14_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_13_7
#define init_10_14_13_7
void FASTCALL call_10_14_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_13_8
#define init_10_14_13_8
void FASTCALL call_10_14_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_13_9
#define init_10_14_13_9
void FASTCALL call_10_14_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_13_10
#define init_10_14_13_10
void FASTCALL call_10_14_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_13_11
#define init_10_14_13_11
void FASTCALL call_10_14_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_13_12
#define init_10_14_13_12
void FASTCALL call_10_14_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_13_13
#define init_10_14_13_13
void FASTCALL call_10_14_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_13_14
#define init_10_14_13_14
void FASTCALL call_10_14_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_13_15
#define init_10_14_13_15
void FASTCALL call_10_14_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_14_0
#define init_10_14_14_0
void FASTCALL call_10_14_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_14_1
#define init_10_14_14_1
void FASTCALL call_10_14_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_14_2
#define init_10_14_14_2
void FASTCALL call_10_14_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_14_3
#define init_10_14_14_3
void FASTCALL call_10_14_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_14_4
#define init_10_14_14_4
void FASTCALL call_10_14_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_14_5
#define init_10_14_14_5
void FASTCALL call_10_14_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_14_6
#define init_10_14_14_6
void FASTCALL call_10_14_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_14_7
#define init_10_14_14_7
void FASTCALL call_10_14_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_14_8
#define init_10_14_14_8
void FASTCALL call_10_14_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_14_9
#define init_10_14_14_9
void FASTCALL call_10_14_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_14_10
#define init_10_14_14_10
void FASTCALL call_10_14_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_14_11
#define init_10_14_14_11
void FASTCALL call_10_14_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_14_12
#define init_10_14_14_12
void FASTCALL call_10_14_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_14_13
#define init_10_14_14_13
void FASTCALL call_10_14_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_14_14
#define init_10_14_14_14
void FASTCALL call_10_14_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_14_15
#define init_10_14_14_15
void FASTCALL call_10_14_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_15_0
#define init_10_14_15_0
void FASTCALL call_10_14_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_15_1
#define init_10_14_15_1
void FASTCALL call_10_14_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_15_2
#define init_10_14_15_2
void FASTCALL call_10_14_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_15_3
#define init_10_14_15_3
void FASTCALL call_10_14_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_15_4
#define init_10_14_15_4
void FASTCALL call_10_14_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_15_5
#define init_10_14_15_5
void FASTCALL call_10_14_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_15_6
#define init_10_14_15_6
void FASTCALL call_10_14_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_15_7
#define init_10_14_15_7
void FASTCALL call_10_14_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_15_8
#define init_10_14_15_8
void FASTCALL call_10_14_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_15_9
#define init_10_14_15_9
void FASTCALL call_10_14_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_15_10
#define init_10_14_15_10
void FASTCALL call_10_14_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_15_11
#define init_10_14_15_11
void FASTCALL call_10_14_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_15_12
#define init_10_14_15_12
void FASTCALL call_10_14_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_15_13
#define init_10_14_15_13
void FASTCALL call_10_14_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_15_14
#define init_10_14_15_14
void FASTCALL call_10_14_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_14_15_15
#define init_10_14_15_15
void FASTCALL call_10_14_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_0_0
#define init_10_15_0_0
void FASTCALL call_10_15_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_0_1
#define init_10_15_0_1
void FASTCALL call_10_15_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_0_2
#define init_10_15_0_2
void FASTCALL call_10_15_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_0_3
#define init_10_15_0_3
void FASTCALL call_10_15_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_0_4
#define init_10_15_0_4
void FASTCALL call_10_15_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_0_5
#define init_10_15_0_5
void FASTCALL call_10_15_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_0_6
#define init_10_15_0_6
void FASTCALL call_10_15_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_0_7
#define init_10_15_0_7
void FASTCALL call_10_15_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_0_8
#define init_10_15_0_8
void FASTCALL call_10_15_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_0_9
#define init_10_15_0_9
void FASTCALL call_10_15_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_0_10
#define init_10_15_0_10
void FASTCALL call_10_15_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_0_11
#define init_10_15_0_11
void FASTCALL call_10_15_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_0_12
#define init_10_15_0_12
void FASTCALL call_10_15_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_0_13
#define init_10_15_0_13
void FASTCALL call_10_15_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_0_14
#define init_10_15_0_14
void FASTCALL call_10_15_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_0_15
#define init_10_15_0_15
void FASTCALL call_10_15_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_1_0
#define init_10_15_1_0
void FASTCALL call_10_15_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_1_1
#define init_10_15_1_1
void FASTCALL call_10_15_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_1_2
#define init_10_15_1_2
void FASTCALL call_10_15_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_1_3
#define init_10_15_1_3
void FASTCALL call_10_15_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_1_4
#define init_10_15_1_4
void FASTCALL call_10_15_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_1_5
#define init_10_15_1_5
void FASTCALL call_10_15_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_1_6
#define init_10_15_1_6
void FASTCALL call_10_15_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_1_7
#define init_10_15_1_7
void FASTCALL call_10_15_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_1_8
#define init_10_15_1_8
void FASTCALL call_10_15_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_1_9
#define init_10_15_1_9
void FASTCALL call_10_15_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_1_10
#define init_10_15_1_10
void FASTCALL call_10_15_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_1_11
#define init_10_15_1_11
void FASTCALL call_10_15_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_1_12
#define init_10_15_1_12
void FASTCALL call_10_15_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_1_13
#define init_10_15_1_13
void FASTCALL call_10_15_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_1_14
#define init_10_15_1_14
void FASTCALL call_10_15_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_1_15
#define init_10_15_1_15
void FASTCALL call_10_15_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_2_0
#define init_10_15_2_0
void FASTCALL call_10_15_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_2_1
#define init_10_15_2_1
void FASTCALL call_10_15_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_2_2
#define init_10_15_2_2
void FASTCALL call_10_15_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_2_3
#define init_10_15_2_3
void FASTCALL call_10_15_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_2_4
#define init_10_15_2_4
void FASTCALL call_10_15_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_2_5
#define init_10_15_2_5
void FASTCALL call_10_15_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_2_6
#define init_10_15_2_6
void FASTCALL call_10_15_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_2_7
#define init_10_15_2_7
void FASTCALL call_10_15_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_2_8
#define init_10_15_2_8
void FASTCALL call_10_15_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_2_9
#define init_10_15_2_9
void FASTCALL call_10_15_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_2_10
#define init_10_15_2_10
void FASTCALL call_10_15_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_2_11
#define init_10_15_2_11
void FASTCALL call_10_15_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_2_12
#define init_10_15_2_12
void FASTCALL call_10_15_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_2_13
#define init_10_15_2_13
void FASTCALL call_10_15_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_2_14
#define init_10_15_2_14
void FASTCALL call_10_15_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_2_15
#define init_10_15_2_15
void FASTCALL call_10_15_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_3_0
#define init_10_15_3_0
void FASTCALL call_10_15_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_3_1
#define init_10_15_3_1
void FASTCALL call_10_15_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_3_2
#define init_10_15_3_2
void FASTCALL call_10_15_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_3_3
#define init_10_15_3_3
void FASTCALL call_10_15_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_3_4
#define init_10_15_3_4
void FASTCALL call_10_15_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_3_5
#define init_10_15_3_5
void FASTCALL call_10_15_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_3_6
#define init_10_15_3_6
void FASTCALL call_10_15_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_3_7
#define init_10_15_3_7
void FASTCALL call_10_15_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_3_8
#define init_10_15_3_8
void FASTCALL call_10_15_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_3_9
#define init_10_15_3_9
void FASTCALL call_10_15_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_3_10
#define init_10_15_3_10
void FASTCALL call_10_15_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_3_11
#define init_10_15_3_11
void FASTCALL call_10_15_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_3_12
#define init_10_15_3_12
void FASTCALL call_10_15_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_3_13
#define init_10_15_3_13
void FASTCALL call_10_15_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_3_14
#define init_10_15_3_14
void FASTCALL call_10_15_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_3_15
#define init_10_15_3_15
void FASTCALL call_10_15_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_4_0
#define init_10_15_4_0
void FASTCALL call_10_15_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_4_1
#define init_10_15_4_1
void FASTCALL call_10_15_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_4_2
#define init_10_15_4_2
void FASTCALL call_10_15_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_4_3
#define init_10_15_4_3
void FASTCALL call_10_15_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_4_4
#define init_10_15_4_4
void FASTCALL call_10_15_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_4_5
#define init_10_15_4_5
void FASTCALL call_10_15_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_4_6
#define init_10_15_4_6
void FASTCALL call_10_15_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_4_7
#define init_10_15_4_7
void FASTCALL call_10_15_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_4_8
#define init_10_15_4_8
void FASTCALL call_10_15_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_4_9
#define init_10_15_4_9
void FASTCALL call_10_15_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_4_10
#define init_10_15_4_10
void FASTCALL call_10_15_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_4_11
#define init_10_15_4_11
void FASTCALL call_10_15_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_4_12
#define init_10_15_4_12
void FASTCALL call_10_15_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_4_13
#define init_10_15_4_13
void FASTCALL call_10_15_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_4_14
#define init_10_15_4_14
void FASTCALL call_10_15_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_4_15
#define init_10_15_4_15
void FASTCALL call_10_15_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_5_0
#define init_10_15_5_0
void FASTCALL call_10_15_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_5_1
#define init_10_15_5_1
void FASTCALL call_10_15_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_5_2
#define init_10_15_5_2
void FASTCALL call_10_15_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_5_3
#define init_10_15_5_3
void FASTCALL call_10_15_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_5_4
#define init_10_15_5_4
void FASTCALL call_10_15_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_5_5
#define init_10_15_5_5
void FASTCALL call_10_15_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_5_6
#define init_10_15_5_6
void FASTCALL call_10_15_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_5_7
#define init_10_15_5_7
void FASTCALL call_10_15_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_5_8
#define init_10_15_5_8
void FASTCALL call_10_15_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_5_9
#define init_10_15_5_9
void FASTCALL call_10_15_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_5_10
#define init_10_15_5_10
void FASTCALL call_10_15_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_5_11
#define init_10_15_5_11
void FASTCALL call_10_15_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_5_12
#define init_10_15_5_12
void FASTCALL call_10_15_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_5_13
#define init_10_15_5_13
void FASTCALL call_10_15_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_5_14
#define init_10_15_5_14
void FASTCALL call_10_15_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_5_15
#define init_10_15_5_15
void FASTCALL call_10_15_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_6_0
#define init_10_15_6_0
void FASTCALL call_10_15_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_6_1
#define init_10_15_6_1
void FASTCALL call_10_15_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_6_2
#define init_10_15_6_2
void FASTCALL call_10_15_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_6_3
#define init_10_15_6_3
void FASTCALL call_10_15_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_6_4
#define init_10_15_6_4
void FASTCALL call_10_15_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_6_5
#define init_10_15_6_5
void FASTCALL call_10_15_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_6_6
#define init_10_15_6_6
void FASTCALL call_10_15_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_6_7
#define init_10_15_6_7
void FASTCALL call_10_15_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_6_8
#define init_10_15_6_8
void FASTCALL call_10_15_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_6_9
#define init_10_15_6_9
void FASTCALL call_10_15_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_6_10
#define init_10_15_6_10
void FASTCALL call_10_15_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_6_11
#define init_10_15_6_11
void FASTCALL call_10_15_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_6_12
#define init_10_15_6_12
void FASTCALL call_10_15_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_6_13
#define init_10_15_6_13
void FASTCALL call_10_15_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_6_14
#define init_10_15_6_14
void FASTCALL call_10_15_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_6_15
#define init_10_15_6_15
void FASTCALL call_10_15_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_7_0
#define init_10_15_7_0
void FASTCALL call_10_15_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_7_1
#define init_10_15_7_1
void FASTCALL call_10_15_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_7_2
#define init_10_15_7_2
void FASTCALL call_10_15_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_7_3
#define init_10_15_7_3
void FASTCALL call_10_15_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_7_4
#define init_10_15_7_4
void FASTCALL call_10_15_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_7_5
#define init_10_15_7_5
void FASTCALL call_10_15_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_7_6
#define init_10_15_7_6
void FASTCALL call_10_15_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_7_7
#define init_10_15_7_7
void FASTCALL call_10_15_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_7_8
#define init_10_15_7_8
void FASTCALL call_10_15_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_7_9
#define init_10_15_7_9
void FASTCALL call_10_15_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_7_10
#define init_10_15_7_10
void FASTCALL call_10_15_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_7_11
#define init_10_15_7_11
void FASTCALL call_10_15_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_7_12
#define init_10_15_7_12
void FASTCALL call_10_15_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_7_13
#define init_10_15_7_13
void FASTCALL call_10_15_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_7_14
#define init_10_15_7_14
void FASTCALL call_10_15_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_7_15
#define init_10_15_7_15
void FASTCALL call_10_15_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_8_0
#define init_10_15_8_0
void FASTCALL call_10_15_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_8_1
#define init_10_15_8_1
void FASTCALL call_10_15_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_8_2
#define init_10_15_8_2
void FASTCALL call_10_15_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_8_3
#define init_10_15_8_3
void FASTCALL call_10_15_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_8_4
#define init_10_15_8_4
void FASTCALL call_10_15_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_8_5
#define init_10_15_8_5
void FASTCALL call_10_15_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_8_6
#define init_10_15_8_6
void FASTCALL call_10_15_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_8_7
#define init_10_15_8_7
void FASTCALL call_10_15_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_8_8
#define init_10_15_8_8
void FASTCALL call_10_15_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_8_9
#define init_10_15_8_9
void FASTCALL call_10_15_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_8_10
#define init_10_15_8_10
void FASTCALL call_10_15_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_8_11
#define init_10_15_8_11
void FASTCALL call_10_15_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_8_12
#define init_10_15_8_12
void FASTCALL call_10_15_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_8_13
#define init_10_15_8_13
void FASTCALL call_10_15_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_8_14
#define init_10_15_8_14
void FASTCALL call_10_15_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_8_15
#define init_10_15_8_15
void FASTCALL call_10_15_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_9_0
#define init_10_15_9_0
void FASTCALL call_10_15_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_9_1
#define init_10_15_9_1
void FASTCALL call_10_15_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_9_2
#define init_10_15_9_2
void FASTCALL call_10_15_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_9_3
#define init_10_15_9_3
void FASTCALL call_10_15_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_9_4
#define init_10_15_9_4
void FASTCALL call_10_15_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_9_5
#define init_10_15_9_5
void FASTCALL call_10_15_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_9_6
#define init_10_15_9_6
void FASTCALL call_10_15_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_9_7
#define init_10_15_9_7
void FASTCALL call_10_15_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_9_8
#define init_10_15_9_8
void FASTCALL call_10_15_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_9_9
#define init_10_15_9_9
void FASTCALL call_10_15_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_9_10
#define init_10_15_9_10
void FASTCALL call_10_15_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_9_11
#define init_10_15_9_11
void FASTCALL call_10_15_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_9_12
#define init_10_15_9_12
void FASTCALL call_10_15_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_9_13
#define init_10_15_9_13
void FASTCALL call_10_15_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_9_14
#define init_10_15_9_14
void FASTCALL call_10_15_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_9_15
#define init_10_15_9_15
void FASTCALL call_10_15_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_10_0
#define init_10_15_10_0
void FASTCALL call_10_15_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_10_1
#define init_10_15_10_1
void FASTCALL call_10_15_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_10_2
#define init_10_15_10_2
void FASTCALL call_10_15_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_10_3
#define init_10_15_10_3
void FASTCALL call_10_15_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_10_4
#define init_10_15_10_4
void FASTCALL call_10_15_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_10_5
#define init_10_15_10_5
void FASTCALL call_10_15_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_10_6
#define init_10_15_10_6
void FASTCALL call_10_15_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_10_7
#define init_10_15_10_7
void FASTCALL call_10_15_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_10_8
#define init_10_15_10_8
void FASTCALL call_10_15_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_10_9
#define init_10_15_10_9
void FASTCALL call_10_15_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_10_10
#define init_10_15_10_10
void FASTCALL call_10_15_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_10_11
#define init_10_15_10_11
void FASTCALL call_10_15_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_10_12
#define init_10_15_10_12
void FASTCALL call_10_15_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_10_13
#define init_10_15_10_13
void FASTCALL call_10_15_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_10_14
#define init_10_15_10_14
void FASTCALL call_10_15_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_10_15
#define init_10_15_10_15
void FASTCALL call_10_15_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_11_0
#define init_10_15_11_0
void FASTCALL call_10_15_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_11_1
#define init_10_15_11_1
void FASTCALL call_10_15_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_11_2
#define init_10_15_11_2
void FASTCALL call_10_15_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_11_3
#define init_10_15_11_3
void FASTCALL call_10_15_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_11_4
#define init_10_15_11_4
void FASTCALL call_10_15_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_11_5
#define init_10_15_11_5
void FASTCALL call_10_15_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_11_6
#define init_10_15_11_6
void FASTCALL call_10_15_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_11_7
#define init_10_15_11_7
void FASTCALL call_10_15_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_11_8
#define init_10_15_11_8
void FASTCALL call_10_15_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_11_9
#define init_10_15_11_9
void FASTCALL call_10_15_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_11_10
#define init_10_15_11_10
void FASTCALL call_10_15_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_11_11
#define init_10_15_11_11
void FASTCALL call_10_15_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_11_12
#define init_10_15_11_12
void FASTCALL call_10_15_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_11_13
#define init_10_15_11_13
void FASTCALL call_10_15_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_11_14
#define init_10_15_11_14
void FASTCALL call_10_15_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_11_15
#define init_10_15_11_15
void FASTCALL call_10_15_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_12_0
#define init_10_15_12_0
void FASTCALL call_10_15_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_12_1
#define init_10_15_12_1
void FASTCALL call_10_15_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_12_2
#define init_10_15_12_2
void FASTCALL call_10_15_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_12_3
#define init_10_15_12_3
void FASTCALL call_10_15_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_12_4
#define init_10_15_12_4
void FASTCALL call_10_15_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_12_5
#define init_10_15_12_5
void FASTCALL call_10_15_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_12_6
#define init_10_15_12_6
void FASTCALL call_10_15_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_12_7
#define init_10_15_12_7
void FASTCALL call_10_15_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_12_8
#define init_10_15_12_8
void FASTCALL call_10_15_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_12_9
#define init_10_15_12_9
void FASTCALL call_10_15_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_12_10
#define init_10_15_12_10
void FASTCALL call_10_15_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_12_11
#define init_10_15_12_11
void FASTCALL call_10_15_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_12_12
#define init_10_15_12_12
void FASTCALL call_10_15_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_12_13
#define init_10_15_12_13
void FASTCALL call_10_15_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_12_14
#define init_10_15_12_14
void FASTCALL call_10_15_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_12_15
#define init_10_15_12_15
void FASTCALL call_10_15_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_13_0
#define init_10_15_13_0
void FASTCALL call_10_15_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_13_1
#define init_10_15_13_1
void FASTCALL call_10_15_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_13_2
#define init_10_15_13_2
void FASTCALL call_10_15_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_13_3
#define init_10_15_13_3
void FASTCALL call_10_15_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_13_4
#define init_10_15_13_4
void FASTCALL call_10_15_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_13_5
#define init_10_15_13_5
void FASTCALL call_10_15_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_13_6
#define init_10_15_13_6
void FASTCALL call_10_15_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_13_7
#define init_10_15_13_7
void FASTCALL call_10_15_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_13_8
#define init_10_15_13_8
void FASTCALL call_10_15_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_13_9
#define init_10_15_13_9
void FASTCALL call_10_15_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_13_10
#define init_10_15_13_10
void FASTCALL call_10_15_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_13_11
#define init_10_15_13_11
void FASTCALL call_10_15_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_13_12
#define init_10_15_13_12
void FASTCALL call_10_15_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_13_13
#define init_10_15_13_13
void FASTCALL call_10_15_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_13_14
#define init_10_15_13_14
void FASTCALL call_10_15_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_13_15
#define init_10_15_13_15
void FASTCALL call_10_15_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_14_0
#define init_10_15_14_0
void FASTCALL call_10_15_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_14_1
#define init_10_15_14_1
void FASTCALL call_10_15_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_14_2
#define init_10_15_14_2
void FASTCALL call_10_15_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_14_3
#define init_10_15_14_3
void FASTCALL call_10_15_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_14_4
#define init_10_15_14_4
void FASTCALL call_10_15_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_14_5
#define init_10_15_14_5
void FASTCALL call_10_15_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_14_6
#define init_10_15_14_6
void FASTCALL call_10_15_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_14_7
#define init_10_15_14_7
void FASTCALL call_10_15_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_14_8
#define init_10_15_14_8
void FASTCALL call_10_15_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_14_9
#define init_10_15_14_9
void FASTCALL call_10_15_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_14_10
#define init_10_15_14_10
void FASTCALL call_10_15_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_14_11
#define init_10_15_14_11
void FASTCALL call_10_15_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_14_12
#define init_10_15_14_12
void FASTCALL call_10_15_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_14_13
#define init_10_15_14_13
void FASTCALL call_10_15_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_14_14
#define init_10_15_14_14
void FASTCALL call_10_15_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_14_15
#define init_10_15_14_15
void FASTCALL call_10_15_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_15_0
#define init_10_15_15_0
void FASTCALL call_10_15_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_15_1
#define init_10_15_15_1
void FASTCALL call_10_15_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_15_2
#define init_10_15_15_2
void FASTCALL call_10_15_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_15_3
#define init_10_15_15_3
void FASTCALL call_10_15_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_15_4
#define init_10_15_15_4
void FASTCALL call_10_15_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_15_5
#define init_10_15_15_5
void FASTCALL call_10_15_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_15_6
#define init_10_15_15_6
void FASTCALL call_10_15_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_15_7
#define init_10_15_15_7
void FASTCALL call_10_15_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_15_8
#define init_10_15_15_8
void FASTCALL call_10_15_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_15_9
#define init_10_15_15_9
void FASTCALL call_10_15_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_15_10
#define init_10_15_15_10
void FASTCALL call_10_15_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_15_11
#define init_10_15_15_11
void FASTCALL call_10_15_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_15_12
#define init_10_15_15_12
void FASTCALL call_10_15_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_15_13
#define init_10_15_15_13
void FASTCALL call_10_15_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_15_14
#define init_10_15_15_14
void FASTCALL call_10_15_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_10_15_15_15
#define init_10_15_15_15
void FASTCALL call_10_15_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_0_0
#define init_11_0_0_0
void FASTCALL call_11_0_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_0_1
#define init_11_0_0_1
void FASTCALL call_11_0_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_0_2
#define init_11_0_0_2
void FASTCALL call_11_0_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_0_3
#define init_11_0_0_3
void FASTCALL call_11_0_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_0_4
#define init_11_0_0_4
void FASTCALL call_11_0_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_0_5
#define init_11_0_0_5
void FASTCALL call_11_0_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_0_6
#define init_11_0_0_6
void FASTCALL call_11_0_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_0_7
#define init_11_0_0_7
void FASTCALL call_11_0_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_0_8
#define init_11_0_0_8
void FASTCALL call_11_0_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_0_9
#define init_11_0_0_9
void FASTCALL call_11_0_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_0_10
#define init_11_0_0_10
void FASTCALL call_11_0_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_0_11
#define init_11_0_0_11
void FASTCALL call_11_0_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_0_12
#define init_11_0_0_12
void FASTCALL call_11_0_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_0_13
#define init_11_0_0_13
void FASTCALL call_11_0_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_0_14
#define init_11_0_0_14
void FASTCALL call_11_0_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_0_15
#define init_11_0_0_15
void FASTCALL call_11_0_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_1_0
#define init_11_0_1_0
void FASTCALL call_11_0_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_1_1
#define init_11_0_1_1
void FASTCALL call_11_0_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_1_2
#define init_11_0_1_2
void FASTCALL call_11_0_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_1_3
#define init_11_0_1_3
void FASTCALL call_11_0_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_1_4
#define init_11_0_1_4
void FASTCALL call_11_0_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_1_5
#define init_11_0_1_5
void FASTCALL call_11_0_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_1_6
#define init_11_0_1_6
void FASTCALL call_11_0_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_1_7
#define init_11_0_1_7
void FASTCALL call_11_0_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_1_8
#define init_11_0_1_8
void FASTCALL call_11_0_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_1_9
#define init_11_0_1_9
void FASTCALL call_11_0_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_1_10
#define init_11_0_1_10
void FASTCALL call_11_0_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_1_11
#define init_11_0_1_11
void FASTCALL call_11_0_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_1_12
#define init_11_0_1_12
void FASTCALL call_11_0_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_1_13
#define init_11_0_1_13
void FASTCALL call_11_0_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_1_14
#define init_11_0_1_14
void FASTCALL call_11_0_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_1_15
#define init_11_0_1_15
void FASTCALL call_11_0_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_2_0
#define init_11_0_2_0
void FASTCALL call_11_0_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_2_1
#define init_11_0_2_1
void FASTCALL call_11_0_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_2_2
#define init_11_0_2_2
void FASTCALL call_11_0_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_2_3
#define init_11_0_2_3
void FASTCALL call_11_0_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_2_4
#define init_11_0_2_4
void FASTCALL call_11_0_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_2_5
#define init_11_0_2_5
void FASTCALL call_11_0_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_2_6
#define init_11_0_2_6
void FASTCALL call_11_0_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_2_7
#define init_11_0_2_7
void FASTCALL call_11_0_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_2_8
#define init_11_0_2_8
void FASTCALL call_11_0_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_2_9
#define init_11_0_2_9
void FASTCALL call_11_0_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_2_10
#define init_11_0_2_10
void FASTCALL call_11_0_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_2_11
#define init_11_0_2_11
void FASTCALL call_11_0_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_2_12
#define init_11_0_2_12
void FASTCALL call_11_0_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_2_13
#define init_11_0_2_13
void FASTCALL call_11_0_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_2_14
#define init_11_0_2_14
void FASTCALL call_11_0_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_2_15
#define init_11_0_2_15
void FASTCALL call_11_0_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_3_0
#define init_11_0_3_0
void FASTCALL call_11_0_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_3_1
#define init_11_0_3_1
void FASTCALL call_11_0_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_3_2
#define init_11_0_3_2
void FASTCALL call_11_0_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_3_3
#define init_11_0_3_3
void FASTCALL call_11_0_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_3_4
#define init_11_0_3_4
void FASTCALL call_11_0_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_3_5
#define init_11_0_3_5
void FASTCALL call_11_0_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_3_6
#define init_11_0_3_6
void FASTCALL call_11_0_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_3_7
#define init_11_0_3_7
void FASTCALL call_11_0_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_3_8
#define init_11_0_3_8
void FASTCALL call_11_0_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_3_9
#define init_11_0_3_9
void FASTCALL call_11_0_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_3_10
#define init_11_0_3_10
void FASTCALL call_11_0_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_3_11
#define init_11_0_3_11
void FASTCALL call_11_0_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_3_12
#define init_11_0_3_12
void FASTCALL call_11_0_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_3_13
#define init_11_0_3_13
void FASTCALL call_11_0_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_3_14
#define init_11_0_3_14
void FASTCALL call_11_0_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_3_15
#define init_11_0_3_15
void FASTCALL call_11_0_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_4_0
#define init_11_0_4_0
void FASTCALL call_11_0_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_4_1
#define init_11_0_4_1
void FASTCALL call_11_0_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_4_2
#define init_11_0_4_2
void FASTCALL call_11_0_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_4_3
#define init_11_0_4_3
void FASTCALL call_11_0_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_4_4
#define init_11_0_4_4
void FASTCALL call_11_0_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_4_5
#define init_11_0_4_5
void FASTCALL call_11_0_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_4_6
#define init_11_0_4_6
void FASTCALL call_11_0_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_4_7
#define init_11_0_4_7
void FASTCALL call_11_0_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_4_8
#define init_11_0_4_8
void FASTCALL call_11_0_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_4_9
#define init_11_0_4_9
void FASTCALL call_11_0_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_4_10
#define init_11_0_4_10
void FASTCALL call_11_0_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_4_11
#define init_11_0_4_11
void FASTCALL call_11_0_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_4_12
#define init_11_0_4_12
void FASTCALL call_11_0_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_4_13
#define init_11_0_4_13
void FASTCALL call_11_0_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_4_14
#define init_11_0_4_14
void FASTCALL call_11_0_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_4_15
#define init_11_0_4_15
void FASTCALL call_11_0_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_5_0
#define init_11_0_5_0
void FASTCALL call_11_0_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_5_1
#define init_11_0_5_1
void FASTCALL call_11_0_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_5_2
#define init_11_0_5_2
void FASTCALL call_11_0_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_5_3
#define init_11_0_5_3
void FASTCALL call_11_0_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_5_4
#define init_11_0_5_4
void FASTCALL call_11_0_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_5_5
#define init_11_0_5_5
void FASTCALL call_11_0_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_5_6
#define init_11_0_5_6
void FASTCALL call_11_0_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_5_7
#define init_11_0_5_7
void FASTCALL call_11_0_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_5_8
#define init_11_0_5_8
void FASTCALL call_11_0_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_5_9
#define init_11_0_5_9
void FASTCALL call_11_0_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_5_10
#define init_11_0_5_10
void FASTCALL call_11_0_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_5_11
#define init_11_0_5_11
void FASTCALL call_11_0_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_5_12
#define init_11_0_5_12
void FASTCALL call_11_0_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_5_13
#define init_11_0_5_13
void FASTCALL call_11_0_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_5_14
#define init_11_0_5_14
void FASTCALL call_11_0_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_5_15
#define init_11_0_5_15
void FASTCALL call_11_0_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_6_0
#define init_11_0_6_0
void FASTCALL call_11_0_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_6_1
#define init_11_0_6_1
void FASTCALL call_11_0_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_6_2
#define init_11_0_6_2
void FASTCALL call_11_0_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_6_3
#define init_11_0_6_3
void FASTCALL call_11_0_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_6_4
#define init_11_0_6_4
void FASTCALL call_11_0_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_6_5
#define init_11_0_6_5
void FASTCALL call_11_0_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_6_6
#define init_11_0_6_6
void FASTCALL call_11_0_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_6_7
#define init_11_0_6_7
void FASTCALL call_11_0_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_6_8
#define init_11_0_6_8
void FASTCALL call_11_0_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_6_9
#define init_11_0_6_9
void FASTCALL call_11_0_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_6_10
#define init_11_0_6_10
void FASTCALL call_11_0_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_6_11
#define init_11_0_6_11
void FASTCALL call_11_0_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_6_12
#define init_11_0_6_12
void FASTCALL call_11_0_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_6_13
#define init_11_0_6_13
void FASTCALL call_11_0_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_6_14
#define init_11_0_6_14
void FASTCALL call_11_0_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_6_15
#define init_11_0_6_15
void FASTCALL call_11_0_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_7_0
#define init_11_0_7_0
void FASTCALL call_11_0_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_7_1
#define init_11_0_7_1
void FASTCALL call_11_0_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_7_2
#define init_11_0_7_2
void FASTCALL call_11_0_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_7_3
#define init_11_0_7_3
void FASTCALL call_11_0_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_7_4
#define init_11_0_7_4
void FASTCALL call_11_0_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_7_5
#define init_11_0_7_5
void FASTCALL call_11_0_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_7_6
#define init_11_0_7_6
void FASTCALL call_11_0_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_7_7
#define init_11_0_7_7
void FASTCALL call_11_0_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_7_8
#define init_11_0_7_8
void FASTCALL call_11_0_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_7_9
#define init_11_0_7_9
void FASTCALL call_11_0_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_7_10
#define init_11_0_7_10
void FASTCALL call_11_0_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_7_11
#define init_11_0_7_11
void FASTCALL call_11_0_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_7_12
#define init_11_0_7_12
void FASTCALL call_11_0_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_7_13
#define init_11_0_7_13
void FASTCALL call_11_0_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_7_14
#define init_11_0_7_14
void FASTCALL call_11_0_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_7_15
#define init_11_0_7_15
void FASTCALL call_11_0_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_8_0
#define init_11_0_8_0
void FASTCALL call_11_0_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_8_1
#define init_11_0_8_1
void FASTCALL call_11_0_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_8_2
#define init_11_0_8_2
void FASTCALL call_11_0_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_8_3
#define init_11_0_8_3
void FASTCALL call_11_0_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_8_4
#define init_11_0_8_4
void FASTCALL call_11_0_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_8_5
#define init_11_0_8_5
void FASTCALL call_11_0_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_8_6
#define init_11_0_8_6
void FASTCALL call_11_0_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_8_7
#define init_11_0_8_7
void FASTCALL call_11_0_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_8_8
#define init_11_0_8_8
void FASTCALL call_11_0_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_8_9
#define init_11_0_8_9
void FASTCALL call_11_0_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_8_10
#define init_11_0_8_10
void FASTCALL call_11_0_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_8_11
#define init_11_0_8_11
void FASTCALL call_11_0_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_8_12
#define init_11_0_8_12
void FASTCALL call_11_0_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_8_13
#define init_11_0_8_13
void FASTCALL call_11_0_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_8_14
#define init_11_0_8_14
void FASTCALL call_11_0_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_8_15
#define init_11_0_8_15
void FASTCALL call_11_0_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_9_0
#define init_11_0_9_0
void FASTCALL call_11_0_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_9_1
#define init_11_0_9_1
void FASTCALL call_11_0_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_9_2
#define init_11_0_9_2
void FASTCALL call_11_0_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_9_3
#define init_11_0_9_3
void FASTCALL call_11_0_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_9_4
#define init_11_0_9_4
void FASTCALL call_11_0_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_9_5
#define init_11_0_9_5
void FASTCALL call_11_0_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_9_6
#define init_11_0_9_6
void FASTCALL call_11_0_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_9_7
#define init_11_0_9_7
void FASTCALL call_11_0_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_9_8
#define init_11_0_9_8
void FASTCALL call_11_0_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_9_9
#define init_11_0_9_9
void FASTCALL call_11_0_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_9_10
#define init_11_0_9_10
void FASTCALL call_11_0_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_9_11
#define init_11_0_9_11
void FASTCALL call_11_0_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_9_12
#define init_11_0_9_12
void FASTCALL call_11_0_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_9_13
#define init_11_0_9_13
void FASTCALL call_11_0_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_9_14
#define init_11_0_9_14
void FASTCALL call_11_0_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_9_15
#define init_11_0_9_15
void FASTCALL call_11_0_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_10_0
#define init_11_0_10_0
void FASTCALL call_11_0_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_10_1
#define init_11_0_10_1
void FASTCALL call_11_0_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_10_2
#define init_11_0_10_2
void FASTCALL call_11_0_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_10_3
#define init_11_0_10_3
void FASTCALL call_11_0_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_10_4
#define init_11_0_10_4
void FASTCALL call_11_0_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_10_5
#define init_11_0_10_5
void FASTCALL call_11_0_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_10_6
#define init_11_0_10_6
void FASTCALL call_11_0_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_10_7
#define init_11_0_10_7
void FASTCALL call_11_0_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_10_8
#define init_11_0_10_8
void FASTCALL call_11_0_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_10_9
#define init_11_0_10_9
void FASTCALL call_11_0_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_10_10
#define init_11_0_10_10
void FASTCALL call_11_0_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_10_11
#define init_11_0_10_11
void FASTCALL call_11_0_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_10_12
#define init_11_0_10_12
void FASTCALL call_11_0_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_10_13
#define init_11_0_10_13
void FASTCALL call_11_0_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_10_14
#define init_11_0_10_14
void FASTCALL call_11_0_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_10_15
#define init_11_0_10_15
void FASTCALL call_11_0_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_11_0
#define init_11_0_11_0
void FASTCALL call_11_0_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_11_1
#define init_11_0_11_1
void FASTCALL call_11_0_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_11_2
#define init_11_0_11_2
void FASTCALL call_11_0_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_11_3
#define init_11_0_11_3
void FASTCALL call_11_0_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_11_4
#define init_11_0_11_4
void FASTCALL call_11_0_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_11_5
#define init_11_0_11_5
void FASTCALL call_11_0_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_11_6
#define init_11_0_11_6
void FASTCALL call_11_0_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_11_7
#define init_11_0_11_7
void FASTCALL call_11_0_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_11_8
#define init_11_0_11_8
void FASTCALL call_11_0_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_11_9
#define init_11_0_11_9
void FASTCALL call_11_0_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_11_10
#define init_11_0_11_10
void FASTCALL call_11_0_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_11_11
#define init_11_0_11_11
void FASTCALL call_11_0_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_11_12
#define init_11_0_11_12
void FASTCALL call_11_0_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_11_13
#define init_11_0_11_13
void FASTCALL call_11_0_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_11_14
#define init_11_0_11_14
void FASTCALL call_11_0_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_11_15
#define init_11_0_11_15
void FASTCALL call_11_0_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_12_0
#define init_11_0_12_0
void FASTCALL call_11_0_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_12_1
#define init_11_0_12_1
void FASTCALL call_11_0_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_12_2
#define init_11_0_12_2
void FASTCALL call_11_0_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_12_3
#define init_11_0_12_3
void FASTCALL call_11_0_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_12_4
#define init_11_0_12_4
void FASTCALL call_11_0_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_12_5
#define init_11_0_12_5
void FASTCALL call_11_0_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_12_6
#define init_11_0_12_6
void FASTCALL call_11_0_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_12_7
#define init_11_0_12_7
void FASTCALL call_11_0_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_12_8
#define init_11_0_12_8
void FASTCALL call_11_0_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_12_9
#define init_11_0_12_9
void FASTCALL call_11_0_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_12_10
#define init_11_0_12_10
void FASTCALL call_11_0_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_12_11
#define init_11_0_12_11
void FASTCALL call_11_0_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_12_12
#define init_11_0_12_12
void FASTCALL call_11_0_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_12_13
#define init_11_0_12_13
void FASTCALL call_11_0_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_12_14
#define init_11_0_12_14
void FASTCALL call_11_0_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_12_15
#define init_11_0_12_15
void FASTCALL call_11_0_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_13_0
#define init_11_0_13_0
void FASTCALL call_11_0_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_13_1
#define init_11_0_13_1
void FASTCALL call_11_0_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_13_2
#define init_11_0_13_2
void FASTCALL call_11_0_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_13_3
#define init_11_0_13_3
void FASTCALL call_11_0_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_13_4
#define init_11_0_13_4
void FASTCALL call_11_0_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_13_5
#define init_11_0_13_5
void FASTCALL call_11_0_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_13_6
#define init_11_0_13_6
void FASTCALL call_11_0_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_13_7
#define init_11_0_13_7
void FASTCALL call_11_0_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_13_8
#define init_11_0_13_8
void FASTCALL call_11_0_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_13_9
#define init_11_0_13_9
void FASTCALL call_11_0_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_13_10
#define init_11_0_13_10
void FASTCALL call_11_0_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_13_11
#define init_11_0_13_11
void FASTCALL call_11_0_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_13_12
#define init_11_0_13_12
void FASTCALL call_11_0_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_13_13
#define init_11_0_13_13
void FASTCALL call_11_0_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_13_14
#define init_11_0_13_14
void FASTCALL call_11_0_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_13_15
#define init_11_0_13_15
void FASTCALL call_11_0_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_14_0
#define init_11_0_14_0
void FASTCALL call_11_0_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_14_1
#define init_11_0_14_1
void FASTCALL call_11_0_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_14_2
#define init_11_0_14_2
void FASTCALL call_11_0_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_14_3
#define init_11_0_14_3
void FASTCALL call_11_0_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_14_4
#define init_11_0_14_4
void FASTCALL call_11_0_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_14_5
#define init_11_0_14_5
void FASTCALL call_11_0_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_14_6
#define init_11_0_14_6
void FASTCALL call_11_0_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_14_7
#define init_11_0_14_7
void FASTCALL call_11_0_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_14_8
#define init_11_0_14_8
void FASTCALL call_11_0_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_14_9
#define init_11_0_14_9
void FASTCALL call_11_0_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_14_10
#define init_11_0_14_10
void FASTCALL call_11_0_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_14_11
#define init_11_0_14_11
void FASTCALL call_11_0_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_14_12
#define init_11_0_14_12
void FASTCALL call_11_0_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_14_13
#define init_11_0_14_13
void FASTCALL call_11_0_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_14_14
#define init_11_0_14_14
void FASTCALL call_11_0_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_14_15
#define init_11_0_14_15
void FASTCALL call_11_0_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_15_0
#define init_11_0_15_0
void FASTCALL call_11_0_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_15_1
#define init_11_0_15_1
void FASTCALL call_11_0_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_15_2
#define init_11_0_15_2
void FASTCALL call_11_0_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_15_3
#define init_11_0_15_3
void FASTCALL call_11_0_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_15_4
#define init_11_0_15_4
void FASTCALL call_11_0_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_15_5
#define init_11_0_15_5
void FASTCALL call_11_0_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_15_6
#define init_11_0_15_6
void FASTCALL call_11_0_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_15_7
#define init_11_0_15_7
void FASTCALL call_11_0_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_15_8
#define init_11_0_15_8
void FASTCALL call_11_0_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_15_9
#define init_11_0_15_9
void FASTCALL call_11_0_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_15_10
#define init_11_0_15_10
void FASTCALL call_11_0_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_15_11
#define init_11_0_15_11
void FASTCALL call_11_0_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_15_12
#define init_11_0_15_12
void FASTCALL call_11_0_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_15_13
#define init_11_0_15_13
void FASTCALL call_11_0_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_15_14
#define init_11_0_15_14
void FASTCALL call_11_0_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_0_15_15
#define init_11_0_15_15
void FASTCALL call_11_0_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_0_0
#define init_11_1_0_0
void FASTCALL call_11_1_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_0_1
#define init_11_1_0_1
void FASTCALL call_11_1_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_0_2
#define init_11_1_0_2
void FASTCALL call_11_1_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_0_3
#define init_11_1_0_3
void FASTCALL call_11_1_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_0_4
#define init_11_1_0_4
void FASTCALL call_11_1_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_0_5
#define init_11_1_0_5
void FASTCALL call_11_1_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_0_6
#define init_11_1_0_6
void FASTCALL call_11_1_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_0_7
#define init_11_1_0_7
void FASTCALL call_11_1_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_0_8
#define init_11_1_0_8
void FASTCALL call_11_1_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_0_9
#define init_11_1_0_9
void FASTCALL call_11_1_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_0_10
#define init_11_1_0_10
void FASTCALL call_11_1_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_0_11
#define init_11_1_0_11
void FASTCALL call_11_1_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_0_12
#define init_11_1_0_12
void FASTCALL call_11_1_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_0_13
#define init_11_1_0_13
void FASTCALL call_11_1_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_0_14
#define init_11_1_0_14
void FASTCALL call_11_1_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_0_15
#define init_11_1_0_15
void FASTCALL call_11_1_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_1_0
#define init_11_1_1_0
void FASTCALL call_11_1_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_1_1
#define init_11_1_1_1
void FASTCALL call_11_1_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_1_2
#define init_11_1_1_2
void FASTCALL call_11_1_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_1_3
#define init_11_1_1_3
void FASTCALL call_11_1_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_1_4
#define init_11_1_1_4
void FASTCALL call_11_1_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_1_5
#define init_11_1_1_5
void FASTCALL call_11_1_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_1_6
#define init_11_1_1_6
void FASTCALL call_11_1_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_1_7
#define init_11_1_1_7
void FASTCALL call_11_1_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_1_8
#define init_11_1_1_8
void FASTCALL call_11_1_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_1_9
#define init_11_1_1_9
void FASTCALL call_11_1_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_1_10
#define init_11_1_1_10
void FASTCALL call_11_1_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_1_11
#define init_11_1_1_11
void FASTCALL call_11_1_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_1_12
#define init_11_1_1_12
void FASTCALL call_11_1_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_1_13
#define init_11_1_1_13
void FASTCALL call_11_1_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_1_14
#define init_11_1_1_14
void FASTCALL call_11_1_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_1_15
#define init_11_1_1_15
void FASTCALL call_11_1_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_2_0
#define init_11_1_2_0
void FASTCALL call_11_1_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_2_1
#define init_11_1_2_1
void FASTCALL call_11_1_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_2_2
#define init_11_1_2_2
void FASTCALL call_11_1_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_2_3
#define init_11_1_2_3
void FASTCALL call_11_1_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_2_4
#define init_11_1_2_4
void FASTCALL call_11_1_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_2_5
#define init_11_1_2_5
void FASTCALL call_11_1_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_2_6
#define init_11_1_2_6
void FASTCALL call_11_1_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_2_7
#define init_11_1_2_7
void FASTCALL call_11_1_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_2_8
#define init_11_1_2_8
void FASTCALL call_11_1_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_2_9
#define init_11_1_2_9
void FASTCALL call_11_1_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_2_10
#define init_11_1_2_10
void FASTCALL call_11_1_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_2_11
#define init_11_1_2_11
void FASTCALL call_11_1_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_2_12
#define init_11_1_2_12
void FASTCALL call_11_1_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_2_13
#define init_11_1_2_13
void FASTCALL call_11_1_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_2_14
#define init_11_1_2_14
void FASTCALL call_11_1_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_2_15
#define init_11_1_2_15
void FASTCALL call_11_1_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_3_0
#define init_11_1_3_0
void FASTCALL call_11_1_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_3_1
#define init_11_1_3_1
void FASTCALL call_11_1_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_3_2
#define init_11_1_3_2
void FASTCALL call_11_1_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_3_3
#define init_11_1_3_3
void FASTCALL call_11_1_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_3_4
#define init_11_1_3_4
void FASTCALL call_11_1_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_3_5
#define init_11_1_3_5
void FASTCALL call_11_1_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_3_6
#define init_11_1_3_6
void FASTCALL call_11_1_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_3_7
#define init_11_1_3_7
void FASTCALL call_11_1_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_3_8
#define init_11_1_3_8
void FASTCALL call_11_1_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_3_9
#define init_11_1_3_9
void FASTCALL call_11_1_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_3_10
#define init_11_1_3_10
void FASTCALL call_11_1_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_3_11
#define init_11_1_3_11
void FASTCALL call_11_1_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_3_12
#define init_11_1_3_12
void FASTCALL call_11_1_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_3_13
#define init_11_1_3_13
void FASTCALL call_11_1_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_3_14
#define init_11_1_3_14
void FASTCALL call_11_1_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_3_15
#define init_11_1_3_15
void FASTCALL call_11_1_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_4_0
#define init_11_1_4_0
void FASTCALL call_11_1_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_4_1
#define init_11_1_4_1
void FASTCALL call_11_1_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_4_2
#define init_11_1_4_2
void FASTCALL call_11_1_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_4_3
#define init_11_1_4_3
void FASTCALL call_11_1_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_4_4
#define init_11_1_4_4
void FASTCALL call_11_1_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_4_5
#define init_11_1_4_5
void FASTCALL call_11_1_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_4_6
#define init_11_1_4_6
void FASTCALL call_11_1_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_4_7
#define init_11_1_4_7
void FASTCALL call_11_1_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_4_8
#define init_11_1_4_8
void FASTCALL call_11_1_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_4_9
#define init_11_1_4_9
void FASTCALL call_11_1_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_4_10
#define init_11_1_4_10
void FASTCALL call_11_1_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_4_11
#define init_11_1_4_11
void FASTCALL call_11_1_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_4_12
#define init_11_1_4_12
void FASTCALL call_11_1_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_4_13
#define init_11_1_4_13
void FASTCALL call_11_1_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_4_14
#define init_11_1_4_14
void FASTCALL call_11_1_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_4_15
#define init_11_1_4_15
void FASTCALL call_11_1_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_5_0
#define init_11_1_5_0
void FASTCALL call_11_1_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_5_1
#define init_11_1_5_1
void FASTCALL call_11_1_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_5_2
#define init_11_1_5_2
void FASTCALL call_11_1_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_5_3
#define init_11_1_5_3
void FASTCALL call_11_1_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_5_4
#define init_11_1_5_4
void FASTCALL call_11_1_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_5_5
#define init_11_1_5_5
void FASTCALL call_11_1_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_5_6
#define init_11_1_5_6
void FASTCALL call_11_1_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_5_7
#define init_11_1_5_7
void FASTCALL call_11_1_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_5_8
#define init_11_1_5_8
void FASTCALL call_11_1_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_5_9
#define init_11_1_5_9
void FASTCALL call_11_1_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_5_10
#define init_11_1_5_10
void FASTCALL call_11_1_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_5_11
#define init_11_1_5_11
void FASTCALL call_11_1_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_5_12
#define init_11_1_5_12
void FASTCALL call_11_1_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_5_13
#define init_11_1_5_13
void FASTCALL call_11_1_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_5_14
#define init_11_1_5_14
void FASTCALL call_11_1_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_5_15
#define init_11_1_5_15
void FASTCALL call_11_1_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_6_0
#define init_11_1_6_0
void FASTCALL call_11_1_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_6_1
#define init_11_1_6_1
void FASTCALL call_11_1_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_6_2
#define init_11_1_6_2
void FASTCALL call_11_1_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_6_3
#define init_11_1_6_3
void FASTCALL call_11_1_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_6_4
#define init_11_1_6_4
void FASTCALL call_11_1_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_6_5
#define init_11_1_6_5
void FASTCALL call_11_1_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_6_6
#define init_11_1_6_6
void FASTCALL call_11_1_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_6_7
#define init_11_1_6_7
void FASTCALL call_11_1_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_6_8
#define init_11_1_6_8
void FASTCALL call_11_1_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_6_9
#define init_11_1_6_9
void FASTCALL call_11_1_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_6_10
#define init_11_1_6_10
void FASTCALL call_11_1_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_6_11
#define init_11_1_6_11
void FASTCALL call_11_1_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_6_12
#define init_11_1_6_12
void FASTCALL call_11_1_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_6_13
#define init_11_1_6_13
void FASTCALL call_11_1_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_6_14
#define init_11_1_6_14
void FASTCALL call_11_1_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_6_15
#define init_11_1_6_15
void FASTCALL call_11_1_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_7_0
#define init_11_1_7_0
void FASTCALL call_11_1_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_7_1
#define init_11_1_7_1
void FASTCALL call_11_1_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_7_2
#define init_11_1_7_2
void FASTCALL call_11_1_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_7_3
#define init_11_1_7_3
void FASTCALL call_11_1_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_7_4
#define init_11_1_7_4
void FASTCALL call_11_1_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_7_5
#define init_11_1_7_5
void FASTCALL call_11_1_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_7_6
#define init_11_1_7_6
void FASTCALL call_11_1_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_7_7
#define init_11_1_7_7
void FASTCALL call_11_1_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_7_8
#define init_11_1_7_8
void FASTCALL call_11_1_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_7_9
#define init_11_1_7_9
void FASTCALL call_11_1_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_7_10
#define init_11_1_7_10
void FASTCALL call_11_1_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_7_11
#define init_11_1_7_11
void FASTCALL call_11_1_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_7_12
#define init_11_1_7_12
void FASTCALL call_11_1_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_7_13
#define init_11_1_7_13
void FASTCALL call_11_1_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_7_14
#define init_11_1_7_14
void FASTCALL call_11_1_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_7_15
#define init_11_1_7_15
void FASTCALL call_11_1_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_8_0
#define init_11_1_8_0
void FASTCALL call_11_1_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_8_1
#define init_11_1_8_1
void FASTCALL call_11_1_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_8_2
#define init_11_1_8_2
void FASTCALL call_11_1_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_8_3
#define init_11_1_8_3
void FASTCALL call_11_1_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_8_4
#define init_11_1_8_4
void FASTCALL call_11_1_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_8_5
#define init_11_1_8_5
void FASTCALL call_11_1_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_8_6
#define init_11_1_8_6
void FASTCALL call_11_1_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_8_7
#define init_11_1_8_7
void FASTCALL call_11_1_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_8_8
#define init_11_1_8_8
void FASTCALL call_11_1_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_8_9
#define init_11_1_8_9
void FASTCALL call_11_1_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_8_10
#define init_11_1_8_10
void FASTCALL call_11_1_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_8_11
#define init_11_1_8_11
void FASTCALL call_11_1_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_8_12
#define init_11_1_8_12
void FASTCALL call_11_1_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_8_13
#define init_11_1_8_13
void FASTCALL call_11_1_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_8_14
#define init_11_1_8_14
void FASTCALL call_11_1_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_8_15
#define init_11_1_8_15
void FASTCALL call_11_1_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_9_0
#define init_11_1_9_0
void FASTCALL call_11_1_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_9_1
#define init_11_1_9_1
void FASTCALL call_11_1_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_9_2
#define init_11_1_9_2
void FASTCALL call_11_1_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_9_3
#define init_11_1_9_3
void FASTCALL call_11_1_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_9_4
#define init_11_1_9_4
void FASTCALL call_11_1_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_9_5
#define init_11_1_9_5
void FASTCALL call_11_1_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_9_6
#define init_11_1_9_6
void FASTCALL call_11_1_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_9_7
#define init_11_1_9_7
void FASTCALL call_11_1_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_9_8
#define init_11_1_9_8
void FASTCALL call_11_1_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_9_9
#define init_11_1_9_9
void FASTCALL call_11_1_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_9_10
#define init_11_1_9_10
void FASTCALL call_11_1_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_9_11
#define init_11_1_9_11
void FASTCALL call_11_1_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_9_12
#define init_11_1_9_12
void FASTCALL call_11_1_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_9_13
#define init_11_1_9_13
void FASTCALL call_11_1_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_9_14
#define init_11_1_9_14
void FASTCALL call_11_1_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_9_15
#define init_11_1_9_15
void FASTCALL call_11_1_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_10_0
#define init_11_1_10_0
void FASTCALL call_11_1_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_10_1
#define init_11_1_10_1
void FASTCALL call_11_1_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_10_2
#define init_11_1_10_2
void FASTCALL call_11_1_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_10_3
#define init_11_1_10_3
void FASTCALL call_11_1_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_10_4
#define init_11_1_10_4
void FASTCALL call_11_1_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_10_5
#define init_11_1_10_5
void FASTCALL call_11_1_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_10_6
#define init_11_1_10_6
void FASTCALL call_11_1_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_10_7
#define init_11_1_10_7
void FASTCALL call_11_1_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_10_8
#define init_11_1_10_8
void FASTCALL call_11_1_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_10_9
#define init_11_1_10_9
void FASTCALL call_11_1_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_10_10
#define init_11_1_10_10
void FASTCALL call_11_1_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_10_11
#define init_11_1_10_11
void FASTCALL call_11_1_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_10_12
#define init_11_1_10_12
void FASTCALL call_11_1_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_10_13
#define init_11_1_10_13
void FASTCALL call_11_1_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_10_14
#define init_11_1_10_14
void FASTCALL call_11_1_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_10_15
#define init_11_1_10_15
void FASTCALL call_11_1_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_11_0
#define init_11_1_11_0
void FASTCALL call_11_1_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_11_1
#define init_11_1_11_1
void FASTCALL call_11_1_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_11_2
#define init_11_1_11_2
void FASTCALL call_11_1_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_11_3
#define init_11_1_11_3
void FASTCALL call_11_1_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_11_4
#define init_11_1_11_4
void FASTCALL call_11_1_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_11_5
#define init_11_1_11_5
void FASTCALL call_11_1_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_11_6
#define init_11_1_11_6
void FASTCALL call_11_1_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_11_7
#define init_11_1_11_7
void FASTCALL call_11_1_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_11_8
#define init_11_1_11_8
void FASTCALL call_11_1_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_11_9
#define init_11_1_11_9
void FASTCALL call_11_1_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_11_10
#define init_11_1_11_10
void FASTCALL call_11_1_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_11_11
#define init_11_1_11_11
void FASTCALL call_11_1_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_11_12
#define init_11_1_11_12
void FASTCALL call_11_1_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_11_13
#define init_11_1_11_13
void FASTCALL call_11_1_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_11_14
#define init_11_1_11_14
void FASTCALL call_11_1_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_11_15
#define init_11_1_11_15
void FASTCALL call_11_1_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_12_0
#define init_11_1_12_0
void FASTCALL call_11_1_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_12_1
#define init_11_1_12_1
void FASTCALL call_11_1_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_12_2
#define init_11_1_12_2
void FASTCALL call_11_1_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_12_3
#define init_11_1_12_3
void FASTCALL call_11_1_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_12_4
#define init_11_1_12_4
void FASTCALL call_11_1_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_12_5
#define init_11_1_12_5
void FASTCALL call_11_1_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_12_6
#define init_11_1_12_6
void FASTCALL call_11_1_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_12_7
#define init_11_1_12_7
void FASTCALL call_11_1_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_12_8
#define init_11_1_12_8
void FASTCALL call_11_1_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_12_9
#define init_11_1_12_9
void FASTCALL call_11_1_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_12_10
#define init_11_1_12_10
void FASTCALL call_11_1_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_12_11
#define init_11_1_12_11
void FASTCALL call_11_1_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_12_12
#define init_11_1_12_12
void FASTCALL call_11_1_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_12_13
#define init_11_1_12_13
void FASTCALL call_11_1_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_12_14
#define init_11_1_12_14
void FASTCALL call_11_1_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_12_15
#define init_11_1_12_15
void FASTCALL call_11_1_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_13_0
#define init_11_1_13_0
void FASTCALL call_11_1_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_13_1
#define init_11_1_13_1
void FASTCALL call_11_1_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_13_2
#define init_11_1_13_2
void FASTCALL call_11_1_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_13_3
#define init_11_1_13_3
void FASTCALL call_11_1_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_13_4
#define init_11_1_13_4
void FASTCALL call_11_1_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_13_5
#define init_11_1_13_5
void FASTCALL call_11_1_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_13_6
#define init_11_1_13_6
void FASTCALL call_11_1_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_13_7
#define init_11_1_13_7
void FASTCALL call_11_1_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_13_8
#define init_11_1_13_8
void FASTCALL call_11_1_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_13_9
#define init_11_1_13_9
void FASTCALL call_11_1_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_13_10
#define init_11_1_13_10
void FASTCALL call_11_1_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_13_11
#define init_11_1_13_11
void FASTCALL call_11_1_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_13_12
#define init_11_1_13_12
void FASTCALL call_11_1_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_13_13
#define init_11_1_13_13
void FASTCALL call_11_1_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_13_14
#define init_11_1_13_14
void FASTCALL call_11_1_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_13_15
#define init_11_1_13_15
void FASTCALL call_11_1_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_14_0
#define init_11_1_14_0
void FASTCALL call_11_1_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_14_1
#define init_11_1_14_1
void FASTCALL call_11_1_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_14_2
#define init_11_1_14_2
void FASTCALL call_11_1_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_14_3
#define init_11_1_14_3
void FASTCALL call_11_1_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_14_4
#define init_11_1_14_4
void FASTCALL call_11_1_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_14_5
#define init_11_1_14_5
void FASTCALL call_11_1_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_14_6
#define init_11_1_14_6
void FASTCALL call_11_1_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_14_7
#define init_11_1_14_7
void FASTCALL call_11_1_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_14_8
#define init_11_1_14_8
void FASTCALL call_11_1_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_14_9
#define init_11_1_14_9
void FASTCALL call_11_1_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_14_10
#define init_11_1_14_10
void FASTCALL call_11_1_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_14_11
#define init_11_1_14_11
void FASTCALL call_11_1_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_14_12
#define init_11_1_14_12
void FASTCALL call_11_1_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_14_13
#define init_11_1_14_13
void FASTCALL call_11_1_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_14_14
#define init_11_1_14_14
void FASTCALL call_11_1_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_14_15
#define init_11_1_14_15
void FASTCALL call_11_1_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_15_0
#define init_11_1_15_0
void FASTCALL call_11_1_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_15_1
#define init_11_1_15_1
void FASTCALL call_11_1_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_15_2
#define init_11_1_15_2
void FASTCALL call_11_1_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_15_3
#define init_11_1_15_3
void FASTCALL call_11_1_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_15_4
#define init_11_1_15_4
void FASTCALL call_11_1_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_15_5
#define init_11_1_15_5
void FASTCALL call_11_1_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_15_6
#define init_11_1_15_6
void FASTCALL call_11_1_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_15_7
#define init_11_1_15_7
void FASTCALL call_11_1_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_15_8
#define init_11_1_15_8
void FASTCALL call_11_1_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_15_9
#define init_11_1_15_9
void FASTCALL call_11_1_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_15_10
#define init_11_1_15_10
void FASTCALL call_11_1_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_15_11
#define init_11_1_15_11
void FASTCALL call_11_1_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_15_12
#define init_11_1_15_12
void FASTCALL call_11_1_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_15_13
#define init_11_1_15_13
void FASTCALL call_11_1_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_15_14
#define init_11_1_15_14
void FASTCALL call_11_1_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_1_15_15
#define init_11_1_15_15
void FASTCALL call_11_1_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_0_0
#define init_11_2_0_0
void FASTCALL call_11_2_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_0_1
#define init_11_2_0_1
void FASTCALL call_11_2_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_0_2
#define init_11_2_0_2
void FASTCALL call_11_2_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_0_3
#define init_11_2_0_3
void FASTCALL call_11_2_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_0_4
#define init_11_2_0_4
void FASTCALL call_11_2_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_0_5
#define init_11_2_0_5
void FASTCALL call_11_2_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_0_6
#define init_11_2_0_6
void FASTCALL call_11_2_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_0_7
#define init_11_2_0_7
void FASTCALL call_11_2_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_0_8
#define init_11_2_0_8
void FASTCALL call_11_2_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_0_9
#define init_11_2_0_9
void FASTCALL call_11_2_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_0_10
#define init_11_2_0_10
void FASTCALL call_11_2_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_0_11
#define init_11_2_0_11
void FASTCALL call_11_2_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_0_12
#define init_11_2_0_12
void FASTCALL call_11_2_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_0_13
#define init_11_2_0_13
void FASTCALL call_11_2_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_0_14
#define init_11_2_0_14
void FASTCALL call_11_2_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_0_15
#define init_11_2_0_15
void FASTCALL call_11_2_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_1_0
#define init_11_2_1_0
void FASTCALL call_11_2_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_1_1
#define init_11_2_1_1
void FASTCALL call_11_2_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_1_2
#define init_11_2_1_2
void FASTCALL call_11_2_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_1_3
#define init_11_2_1_3
void FASTCALL call_11_2_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_1_4
#define init_11_2_1_4
void FASTCALL call_11_2_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_1_5
#define init_11_2_1_5
void FASTCALL call_11_2_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_1_6
#define init_11_2_1_6
void FASTCALL call_11_2_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_1_7
#define init_11_2_1_7
void FASTCALL call_11_2_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_1_8
#define init_11_2_1_8
void FASTCALL call_11_2_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_1_9
#define init_11_2_1_9
void FASTCALL call_11_2_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_1_10
#define init_11_2_1_10
void FASTCALL call_11_2_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_1_11
#define init_11_2_1_11
void FASTCALL call_11_2_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_1_12
#define init_11_2_1_12
void FASTCALL call_11_2_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_1_13
#define init_11_2_1_13
void FASTCALL call_11_2_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_1_14
#define init_11_2_1_14
void FASTCALL call_11_2_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_1_15
#define init_11_2_1_15
void FASTCALL call_11_2_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_2_0
#define init_11_2_2_0
void FASTCALL call_11_2_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_2_1
#define init_11_2_2_1
void FASTCALL call_11_2_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_2_2
#define init_11_2_2_2
void FASTCALL call_11_2_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_2_3
#define init_11_2_2_3
void FASTCALL call_11_2_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_2_4
#define init_11_2_2_4
void FASTCALL call_11_2_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_2_5
#define init_11_2_2_5
void FASTCALL call_11_2_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_2_6
#define init_11_2_2_6
void FASTCALL call_11_2_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_2_7
#define init_11_2_2_7
void FASTCALL call_11_2_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_2_8
#define init_11_2_2_8
void FASTCALL call_11_2_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_2_9
#define init_11_2_2_9
void FASTCALL call_11_2_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_2_10
#define init_11_2_2_10
void FASTCALL call_11_2_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_2_11
#define init_11_2_2_11
void FASTCALL call_11_2_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_2_12
#define init_11_2_2_12
void FASTCALL call_11_2_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_2_13
#define init_11_2_2_13
void FASTCALL call_11_2_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_2_14
#define init_11_2_2_14
void FASTCALL call_11_2_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_2_15
#define init_11_2_2_15
void FASTCALL call_11_2_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_3_0
#define init_11_2_3_0
void FASTCALL call_11_2_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_3_1
#define init_11_2_3_1
void FASTCALL call_11_2_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_3_2
#define init_11_2_3_2
void FASTCALL call_11_2_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_3_3
#define init_11_2_3_3
void FASTCALL call_11_2_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_3_4
#define init_11_2_3_4
void FASTCALL call_11_2_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_3_5
#define init_11_2_3_5
void FASTCALL call_11_2_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_3_6
#define init_11_2_3_6
void FASTCALL call_11_2_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_3_7
#define init_11_2_3_7
void FASTCALL call_11_2_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_3_8
#define init_11_2_3_8
void FASTCALL call_11_2_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_3_9
#define init_11_2_3_9
void FASTCALL call_11_2_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_3_10
#define init_11_2_3_10
void FASTCALL call_11_2_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_3_11
#define init_11_2_3_11
void FASTCALL call_11_2_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_3_12
#define init_11_2_3_12
void FASTCALL call_11_2_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_3_13
#define init_11_2_3_13
void FASTCALL call_11_2_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_3_14
#define init_11_2_3_14
void FASTCALL call_11_2_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_3_15
#define init_11_2_3_15
void FASTCALL call_11_2_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_4_0
#define init_11_2_4_0
void FASTCALL call_11_2_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_4_1
#define init_11_2_4_1
void FASTCALL call_11_2_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_4_2
#define init_11_2_4_2
void FASTCALL call_11_2_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_4_3
#define init_11_2_4_3
void FASTCALL call_11_2_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_4_4
#define init_11_2_4_4
void FASTCALL call_11_2_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_4_5
#define init_11_2_4_5
void FASTCALL call_11_2_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_4_6
#define init_11_2_4_6
void FASTCALL call_11_2_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_4_7
#define init_11_2_4_7
void FASTCALL call_11_2_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_4_8
#define init_11_2_4_8
void FASTCALL call_11_2_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_4_9
#define init_11_2_4_9
void FASTCALL call_11_2_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_4_10
#define init_11_2_4_10
void FASTCALL call_11_2_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_4_11
#define init_11_2_4_11
void FASTCALL call_11_2_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_4_12
#define init_11_2_4_12
void FASTCALL call_11_2_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_4_13
#define init_11_2_4_13
void FASTCALL call_11_2_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_4_14
#define init_11_2_4_14
void FASTCALL call_11_2_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_4_15
#define init_11_2_4_15
void FASTCALL call_11_2_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_5_0
#define init_11_2_5_0
void FASTCALL call_11_2_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_5_1
#define init_11_2_5_1
void FASTCALL call_11_2_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_5_2
#define init_11_2_5_2
void FASTCALL call_11_2_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_5_3
#define init_11_2_5_3
void FASTCALL call_11_2_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_5_4
#define init_11_2_5_4
void FASTCALL call_11_2_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_5_5
#define init_11_2_5_5
void FASTCALL call_11_2_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_5_6
#define init_11_2_5_6
void FASTCALL call_11_2_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_5_7
#define init_11_2_5_7
void FASTCALL call_11_2_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_5_8
#define init_11_2_5_8
void FASTCALL call_11_2_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_5_9
#define init_11_2_5_9
void FASTCALL call_11_2_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_5_10
#define init_11_2_5_10
void FASTCALL call_11_2_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_5_11
#define init_11_2_5_11
void FASTCALL call_11_2_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_5_12
#define init_11_2_5_12
void FASTCALL call_11_2_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_5_13
#define init_11_2_5_13
void FASTCALL call_11_2_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_5_14
#define init_11_2_5_14
void FASTCALL call_11_2_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_5_15
#define init_11_2_5_15
void FASTCALL call_11_2_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_6_0
#define init_11_2_6_0
void FASTCALL call_11_2_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_6_1
#define init_11_2_6_1
void FASTCALL call_11_2_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_6_2
#define init_11_2_6_2
void FASTCALL call_11_2_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_6_3
#define init_11_2_6_3
void FASTCALL call_11_2_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_6_4
#define init_11_2_6_4
void FASTCALL call_11_2_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_6_5
#define init_11_2_6_5
void FASTCALL call_11_2_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_6_6
#define init_11_2_6_6
void FASTCALL call_11_2_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_6_7
#define init_11_2_6_7
void FASTCALL call_11_2_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_6_8
#define init_11_2_6_8
void FASTCALL call_11_2_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_6_9
#define init_11_2_6_9
void FASTCALL call_11_2_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_6_10
#define init_11_2_6_10
void FASTCALL call_11_2_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_6_11
#define init_11_2_6_11
void FASTCALL call_11_2_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_6_12
#define init_11_2_6_12
void FASTCALL call_11_2_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_6_13
#define init_11_2_6_13
void FASTCALL call_11_2_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_6_14
#define init_11_2_6_14
void FASTCALL call_11_2_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_6_15
#define init_11_2_6_15
void FASTCALL call_11_2_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_7_0
#define init_11_2_7_0
void FASTCALL call_11_2_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_7_1
#define init_11_2_7_1
void FASTCALL call_11_2_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_7_2
#define init_11_2_7_2
void FASTCALL call_11_2_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_7_3
#define init_11_2_7_3
void FASTCALL call_11_2_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_7_4
#define init_11_2_7_4
void FASTCALL call_11_2_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_7_5
#define init_11_2_7_5
void FASTCALL call_11_2_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_7_6
#define init_11_2_7_6
void FASTCALL call_11_2_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_7_7
#define init_11_2_7_7
void FASTCALL call_11_2_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_7_8
#define init_11_2_7_8
void FASTCALL call_11_2_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_7_9
#define init_11_2_7_9
void FASTCALL call_11_2_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_7_10
#define init_11_2_7_10
void FASTCALL call_11_2_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_7_11
#define init_11_2_7_11
void FASTCALL call_11_2_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_7_12
#define init_11_2_7_12
void FASTCALL call_11_2_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_7_13
#define init_11_2_7_13
void FASTCALL call_11_2_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_7_14
#define init_11_2_7_14
void FASTCALL call_11_2_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_7_15
#define init_11_2_7_15
void FASTCALL call_11_2_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_8_0
#define init_11_2_8_0
void FASTCALL call_11_2_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_8_1
#define init_11_2_8_1
void FASTCALL call_11_2_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_8_2
#define init_11_2_8_2
void FASTCALL call_11_2_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_8_3
#define init_11_2_8_3
void FASTCALL call_11_2_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_8_4
#define init_11_2_8_4
void FASTCALL call_11_2_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_8_5
#define init_11_2_8_5
void FASTCALL call_11_2_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_8_6
#define init_11_2_8_6
void FASTCALL call_11_2_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_8_7
#define init_11_2_8_7
void FASTCALL call_11_2_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_8_8
#define init_11_2_8_8
void FASTCALL call_11_2_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_8_9
#define init_11_2_8_9
void FASTCALL call_11_2_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_8_10
#define init_11_2_8_10
void FASTCALL call_11_2_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_8_11
#define init_11_2_8_11
void FASTCALL call_11_2_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_8_12
#define init_11_2_8_12
void FASTCALL call_11_2_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_8_13
#define init_11_2_8_13
void FASTCALL call_11_2_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_8_14
#define init_11_2_8_14
void FASTCALL call_11_2_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_8_15
#define init_11_2_8_15
void FASTCALL call_11_2_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_9_0
#define init_11_2_9_0
void FASTCALL call_11_2_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_9_1
#define init_11_2_9_1
void FASTCALL call_11_2_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_9_2
#define init_11_2_9_2
void FASTCALL call_11_2_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_9_3
#define init_11_2_9_3
void FASTCALL call_11_2_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_9_4
#define init_11_2_9_4
void FASTCALL call_11_2_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_9_5
#define init_11_2_9_5
void FASTCALL call_11_2_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_9_6
#define init_11_2_9_6
void FASTCALL call_11_2_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_9_7
#define init_11_2_9_7
void FASTCALL call_11_2_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_9_8
#define init_11_2_9_8
void FASTCALL call_11_2_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_9_9
#define init_11_2_9_9
void FASTCALL call_11_2_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_9_10
#define init_11_2_9_10
void FASTCALL call_11_2_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_9_11
#define init_11_2_9_11
void FASTCALL call_11_2_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_9_12
#define init_11_2_9_12
void FASTCALL call_11_2_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_9_13
#define init_11_2_9_13
void FASTCALL call_11_2_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_9_14
#define init_11_2_9_14
void FASTCALL call_11_2_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_9_15
#define init_11_2_9_15
void FASTCALL call_11_2_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_10_0
#define init_11_2_10_0
void FASTCALL call_11_2_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_10_1
#define init_11_2_10_1
void FASTCALL call_11_2_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_10_2
#define init_11_2_10_2
void FASTCALL call_11_2_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_10_3
#define init_11_2_10_3
void FASTCALL call_11_2_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_10_4
#define init_11_2_10_4
void FASTCALL call_11_2_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_10_5
#define init_11_2_10_5
void FASTCALL call_11_2_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_10_6
#define init_11_2_10_6
void FASTCALL call_11_2_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_10_7
#define init_11_2_10_7
void FASTCALL call_11_2_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_10_8
#define init_11_2_10_8
void FASTCALL call_11_2_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_10_9
#define init_11_2_10_9
void FASTCALL call_11_2_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_10_10
#define init_11_2_10_10
void FASTCALL call_11_2_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_10_11
#define init_11_2_10_11
void FASTCALL call_11_2_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_10_12
#define init_11_2_10_12
void FASTCALL call_11_2_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_10_13
#define init_11_2_10_13
void FASTCALL call_11_2_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_10_14
#define init_11_2_10_14
void FASTCALL call_11_2_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_10_15
#define init_11_2_10_15
void FASTCALL call_11_2_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_11_0
#define init_11_2_11_0
void FASTCALL call_11_2_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_11_1
#define init_11_2_11_1
void FASTCALL call_11_2_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_11_2
#define init_11_2_11_2
void FASTCALL call_11_2_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_11_3
#define init_11_2_11_3
void FASTCALL call_11_2_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_11_4
#define init_11_2_11_4
void FASTCALL call_11_2_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_11_5
#define init_11_2_11_5
void FASTCALL call_11_2_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_11_6
#define init_11_2_11_6
void FASTCALL call_11_2_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_11_7
#define init_11_2_11_7
void FASTCALL call_11_2_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_11_8
#define init_11_2_11_8
void FASTCALL call_11_2_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_11_9
#define init_11_2_11_9
void FASTCALL call_11_2_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_11_10
#define init_11_2_11_10
void FASTCALL call_11_2_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_11_11
#define init_11_2_11_11
void FASTCALL call_11_2_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_11_12
#define init_11_2_11_12
void FASTCALL call_11_2_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_11_13
#define init_11_2_11_13
void FASTCALL call_11_2_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_11_14
#define init_11_2_11_14
void FASTCALL call_11_2_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_11_15
#define init_11_2_11_15
void FASTCALL call_11_2_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_12_0
#define init_11_2_12_0
void FASTCALL call_11_2_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_12_1
#define init_11_2_12_1
void FASTCALL call_11_2_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_12_2
#define init_11_2_12_2
void FASTCALL call_11_2_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_12_3
#define init_11_2_12_3
void FASTCALL call_11_2_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_12_4
#define init_11_2_12_4
void FASTCALL call_11_2_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_12_5
#define init_11_2_12_5
void FASTCALL call_11_2_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_12_6
#define init_11_2_12_6
void FASTCALL call_11_2_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_12_7
#define init_11_2_12_7
void FASTCALL call_11_2_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_12_8
#define init_11_2_12_8
void FASTCALL call_11_2_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_12_9
#define init_11_2_12_9
void FASTCALL call_11_2_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_12_10
#define init_11_2_12_10
void FASTCALL call_11_2_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_12_11
#define init_11_2_12_11
void FASTCALL call_11_2_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_12_12
#define init_11_2_12_12
void FASTCALL call_11_2_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_12_13
#define init_11_2_12_13
void FASTCALL call_11_2_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_12_14
#define init_11_2_12_14
void FASTCALL call_11_2_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_12_15
#define init_11_2_12_15
void FASTCALL call_11_2_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_13_0
#define init_11_2_13_0
void FASTCALL call_11_2_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_13_1
#define init_11_2_13_1
void FASTCALL call_11_2_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_13_2
#define init_11_2_13_2
void FASTCALL call_11_2_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_13_3
#define init_11_2_13_3
void FASTCALL call_11_2_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_13_4
#define init_11_2_13_4
void FASTCALL call_11_2_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_13_5
#define init_11_2_13_5
void FASTCALL call_11_2_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_13_6
#define init_11_2_13_6
void FASTCALL call_11_2_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_13_7
#define init_11_2_13_7
void FASTCALL call_11_2_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_13_8
#define init_11_2_13_8
void FASTCALL call_11_2_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_13_9
#define init_11_2_13_9
void FASTCALL call_11_2_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_13_10
#define init_11_2_13_10
void FASTCALL call_11_2_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_13_11
#define init_11_2_13_11
void FASTCALL call_11_2_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_13_12
#define init_11_2_13_12
void FASTCALL call_11_2_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_13_13
#define init_11_2_13_13
void FASTCALL call_11_2_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_13_14
#define init_11_2_13_14
void FASTCALL call_11_2_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_13_15
#define init_11_2_13_15
void FASTCALL call_11_2_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_14_0
#define init_11_2_14_0
void FASTCALL call_11_2_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_14_1
#define init_11_2_14_1
void FASTCALL call_11_2_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_14_2
#define init_11_2_14_2
void FASTCALL call_11_2_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_14_3
#define init_11_2_14_3
void FASTCALL call_11_2_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_14_4
#define init_11_2_14_4
void FASTCALL call_11_2_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_14_5
#define init_11_2_14_5
void FASTCALL call_11_2_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_14_6
#define init_11_2_14_6
void FASTCALL call_11_2_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_14_7
#define init_11_2_14_7
void FASTCALL call_11_2_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_14_8
#define init_11_2_14_8
void FASTCALL call_11_2_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_14_9
#define init_11_2_14_9
void FASTCALL call_11_2_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_14_10
#define init_11_2_14_10
void FASTCALL call_11_2_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_14_11
#define init_11_2_14_11
void FASTCALL call_11_2_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_14_12
#define init_11_2_14_12
void FASTCALL call_11_2_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_14_13
#define init_11_2_14_13
void FASTCALL call_11_2_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_14_14
#define init_11_2_14_14
void FASTCALL call_11_2_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_14_15
#define init_11_2_14_15
void FASTCALL call_11_2_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_15_0
#define init_11_2_15_0
void FASTCALL call_11_2_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_15_1
#define init_11_2_15_1
void FASTCALL call_11_2_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_15_2
#define init_11_2_15_2
void FASTCALL call_11_2_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_15_3
#define init_11_2_15_3
void FASTCALL call_11_2_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_15_4
#define init_11_2_15_4
void FASTCALL call_11_2_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_15_5
#define init_11_2_15_5
void FASTCALL call_11_2_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_15_6
#define init_11_2_15_6
void FASTCALL call_11_2_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_15_7
#define init_11_2_15_7
void FASTCALL call_11_2_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_15_8
#define init_11_2_15_8
void FASTCALL call_11_2_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_15_9
#define init_11_2_15_9
void FASTCALL call_11_2_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_15_10
#define init_11_2_15_10
void FASTCALL call_11_2_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_15_11
#define init_11_2_15_11
void FASTCALL call_11_2_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_15_12
#define init_11_2_15_12
void FASTCALL call_11_2_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_15_13
#define init_11_2_15_13
void FASTCALL call_11_2_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_15_14
#define init_11_2_15_14
void FASTCALL call_11_2_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_2_15_15
#define init_11_2_15_15
void FASTCALL call_11_2_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_0_0
#define init_11_3_0_0
void FASTCALL call_11_3_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_0_1
#define init_11_3_0_1
void FASTCALL call_11_3_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_0_2
#define init_11_3_0_2
void FASTCALL call_11_3_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_0_3
#define init_11_3_0_3
void FASTCALL call_11_3_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_0_4
#define init_11_3_0_4
void FASTCALL call_11_3_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_0_5
#define init_11_3_0_5
void FASTCALL call_11_3_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_0_6
#define init_11_3_0_6
void FASTCALL call_11_3_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_0_7
#define init_11_3_0_7
void FASTCALL call_11_3_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_0_8
#define init_11_3_0_8
void FASTCALL call_11_3_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_0_9
#define init_11_3_0_9
void FASTCALL call_11_3_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_0_10
#define init_11_3_0_10
void FASTCALL call_11_3_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_0_11
#define init_11_3_0_11
void FASTCALL call_11_3_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_0_12
#define init_11_3_0_12
void FASTCALL call_11_3_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_0_13
#define init_11_3_0_13
void FASTCALL call_11_3_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_0_14
#define init_11_3_0_14
void FASTCALL call_11_3_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_0_15
#define init_11_3_0_15
void FASTCALL call_11_3_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_1_0
#define init_11_3_1_0
void FASTCALL call_11_3_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_1_1
#define init_11_3_1_1
void FASTCALL call_11_3_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_1_2
#define init_11_3_1_2
void FASTCALL call_11_3_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_1_3
#define init_11_3_1_3
void FASTCALL call_11_3_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_1_4
#define init_11_3_1_4
void FASTCALL call_11_3_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_1_5
#define init_11_3_1_5
void FASTCALL call_11_3_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_1_6
#define init_11_3_1_6
void FASTCALL call_11_3_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_1_7
#define init_11_3_1_7
void FASTCALL call_11_3_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_1_8
#define init_11_3_1_8
void FASTCALL call_11_3_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_1_9
#define init_11_3_1_9
void FASTCALL call_11_3_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_1_10
#define init_11_3_1_10
void FASTCALL call_11_3_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_1_11
#define init_11_3_1_11
void FASTCALL call_11_3_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_1_12
#define init_11_3_1_12
void FASTCALL call_11_3_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_1_13
#define init_11_3_1_13
void FASTCALL call_11_3_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_1_14
#define init_11_3_1_14
void FASTCALL call_11_3_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_1_15
#define init_11_3_1_15
void FASTCALL call_11_3_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_2_0
#define init_11_3_2_0
void FASTCALL call_11_3_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_2_1
#define init_11_3_2_1
void FASTCALL call_11_3_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_2_2
#define init_11_3_2_2
void FASTCALL call_11_3_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_2_3
#define init_11_3_2_3
void FASTCALL call_11_3_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_2_4
#define init_11_3_2_4
void FASTCALL call_11_3_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_2_5
#define init_11_3_2_5
void FASTCALL call_11_3_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_2_6
#define init_11_3_2_6
void FASTCALL call_11_3_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_2_7
#define init_11_3_2_7
void FASTCALL call_11_3_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_2_8
#define init_11_3_2_8
void FASTCALL call_11_3_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_2_9
#define init_11_3_2_9
void FASTCALL call_11_3_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_2_10
#define init_11_3_2_10
void FASTCALL call_11_3_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_2_11
#define init_11_3_2_11
void FASTCALL call_11_3_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_2_12
#define init_11_3_2_12
void FASTCALL call_11_3_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_2_13
#define init_11_3_2_13
void FASTCALL call_11_3_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_2_14
#define init_11_3_2_14
void FASTCALL call_11_3_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_2_15
#define init_11_3_2_15
void FASTCALL call_11_3_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_3_0
#define init_11_3_3_0
void FASTCALL call_11_3_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_3_1
#define init_11_3_3_1
void FASTCALL call_11_3_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_3_2
#define init_11_3_3_2
void FASTCALL call_11_3_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_3_3
#define init_11_3_3_3
void FASTCALL call_11_3_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_3_4
#define init_11_3_3_4
void FASTCALL call_11_3_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_3_5
#define init_11_3_3_5
void FASTCALL call_11_3_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_3_6
#define init_11_3_3_6
void FASTCALL call_11_3_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_3_7
#define init_11_3_3_7
void FASTCALL call_11_3_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_3_8
#define init_11_3_3_8
void FASTCALL call_11_3_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_3_9
#define init_11_3_3_9
void FASTCALL call_11_3_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_3_10
#define init_11_3_3_10
void FASTCALL call_11_3_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_3_11
#define init_11_3_3_11
void FASTCALL call_11_3_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_3_12
#define init_11_3_3_12
void FASTCALL call_11_3_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_3_13
#define init_11_3_3_13
void FASTCALL call_11_3_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_3_14
#define init_11_3_3_14
void FASTCALL call_11_3_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_3_15
#define init_11_3_3_15
void FASTCALL call_11_3_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_4_0
#define init_11_3_4_0
void FASTCALL call_11_3_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_4_1
#define init_11_3_4_1
void FASTCALL call_11_3_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_4_2
#define init_11_3_4_2
void FASTCALL call_11_3_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_4_3
#define init_11_3_4_3
void FASTCALL call_11_3_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_4_4
#define init_11_3_4_4
void FASTCALL call_11_3_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_4_5
#define init_11_3_4_5
void FASTCALL call_11_3_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_4_6
#define init_11_3_4_6
void FASTCALL call_11_3_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_4_7
#define init_11_3_4_7
void FASTCALL call_11_3_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_4_8
#define init_11_3_4_8
void FASTCALL call_11_3_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_4_9
#define init_11_3_4_9
void FASTCALL call_11_3_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_4_10
#define init_11_3_4_10
void FASTCALL call_11_3_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_4_11
#define init_11_3_4_11
void FASTCALL call_11_3_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_4_12
#define init_11_3_4_12
void FASTCALL call_11_3_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_4_13
#define init_11_3_4_13
void FASTCALL call_11_3_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_4_14
#define init_11_3_4_14
void FASTCALL call_11_3_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_4_15
#define init_11_3_4_15
void FASTCALL call_11_3_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_5_0
#define init_11_3_5_0
void FASTCALL call_11_3_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_5_1
#define init_11_3_5_1
void FASTCALL call_11_3_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_5_2
#define init_11_3_5_2
void FASTCALL call_11_3_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_5_3
#define init_11_3_5_3
void FASTCALL call_11_3_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_5_4
#define init_11_3_5_4
void FASTCALL call_11_3_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_5_5
#define init_11_3_5_5
void FASTCALL call_11_3_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_5_6
#define init_11_3_5_6
void FASTCALL call_11_3_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_5_7
#define init_11_3_5_7
void FASTCALL call_11_3_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_5_8
#define init_11_3_5_8
void FASTCALL call_11_3_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_5_9
#define init_11_3_5_9
void FASTCALL call_11_3_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_5_10
#define init_11_3_5_10
void FASTCALL call_11_3_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_5_11
#define init_11_3_5_11
void FASTCALL call_11_3_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_5_12
#define init_11_3_5_12
void FASTCALL call_11_3_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_5_13
#define init_11_3_5_13
void FASTCALL call_11_3_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_5_14
#define init_11_3_5_14
void FASTCALL call_11_3_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_5_15
#define init_11_3_5_15
void FASTCALL call_11_3_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_6_0
#define init_11_3_6_0
void FASTCALL call_11_3_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_6_1
#define init_11_3_6_1
void FASTCALL call_11_3_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_6_2
#define init_11_3_6_2
void FASTCALL call_11_3_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_6_3
#define init_11_3_6_3
void FASTCALL call_11_3_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_6_4
#define init_11_3_6_4
void FASTCALL call_11_3_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_6_5
#define init_11_3_6_5
void FASTCALL call_11_3_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_6_6
#define init_11_3_6_6
void FASTCALL call_11_3_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_6_7
#define init_11_3_6_7
void FASTCALL call_11_3_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_6_8
#define init_11_3_6_8
void FASTCALL call_11_3_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_6_9
#define init_11_3_6_9
void FASTCALL call_11_3_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_6_10
#define init_11_3_6_10
void FASTCALL call_11_3_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_6_11
#define init_11_3_6_11
void FASTCALL call_11_3_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_6_12
#define init_11_3_6_12
void FASTCALL call_11_3_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_6_13
#define init_11_3_6_13
void FASTCALL call_11_3_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_6_14
#define init_11_3_6_14
void FASTCALL call_11_3_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_6_15
#define init_11_3_6_15
void FASTCALL call_11_3_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_7_0
#define init_11_3_7_0
void FASTCALL call_11_3_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_7_1
#define init_11_3_7_1
void FASTCALL call_11_3_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_7_2
#define init_11_3_7_2
void FASTCALL call_11_3_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_7_3
#define init_11_3_7_3
void FASTCALL call_11_3_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_7_4
#define init_11_3_7_4
void FASTCALL call_11_3_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_7_5
#define init_11_3_7_5
void FASTCALL call_11_3_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_7_6
#define init_11_3_7_6
void FASTCALL call_11_3_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_7_7
#define init_11_3_7_7
void FASTCALL call_11_3_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_7_8
#define init_11_3_7_8
void FASTCALL call_11_3_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_7_9
#define init_11_3_7_9
void FASTCALL call_11_3_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_7_10
#define init_11_3_7_10
void FASTCALL call_11_3_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_7_11
#define init_11_3_7_11
void FASTCALL call_11_3_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_7_12
#define init_11_3_7_12
void FASTCALL call_11_3_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_7_13
#define init_11_3_7_13
void FASTCALL call_11_3_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_7_14
#define init_11_3_7_14
void FASTCALL call_11_3_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_7_15
#define init_11_3_7_15
void FASTCALL call_11_3_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_8_0
#define init_11_3_8_0
void FASTCALL call_11_3_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_8_1
#define init_11_3_8_1
void FASTCALL call_11_3_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_8_2
#define init_11_3_8_2
void FASTCALL call_11_3_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_8_3
#define init_11_3_8_3
void FASTCALL call_11_3_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_8_4
#define init_11_3_8_4
void FASTCALL call_11_3_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_8_5
#define init_11_3_8_5
void FASTCALL call_11_3_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_8_6
#define init_11_3_8_6
void FASTCALL call_11_3_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_8_7
#define init_11_3_8_7
void FASTCALL call_11_3_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_8_8
#define init_11_3_8_8
void FASTCALL call_11_3_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_8_9
#define init_11_3_8_9
void FASTCALL call_11_3_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_8_10
#define init_11_3_8_10
void FASTCALL call_11_3_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_8_11
#define init_11_3_8_11
void FASTCALL call_11_3_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_8_12
#define init_11_3_8_12
void FASTCALL call_11_3_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_8_13
#define init_11_3_8_13
void FASTCALL call_11_3_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_8_14
#define init_11_3_8_14
void FASTCALL call_11_3_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_8_15
#define init_11_3_8_15
void FASTCALL call_11_3_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_9_0
#define init_11_3_9_0
void FASTCALL call_11_3_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_9_1
#define init_11_3_9_1
void FASTCALL call_11_3_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_9_2
#define init_11_3_9_2
void FASTCALL call_11_3_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_9_3
#define init_11_3_9_3
void FASTCALL call_11_3_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_9_4
#define init_11_3_9_4
void FASTCALL call_11_3_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_9_5
#define init_11_3_9_5
void FASTCALL call_11_3_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_9_6
#define init_11_3_9_6
void FASTCALL call_11_3_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_9_7
#define init_11_3_9_7
void FASTCALL call_11_3_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_9_8
#define init_11_3_9_8
void FASTCALL call_11_3_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_9_9
#define init_11_3_9_9
void FASTCALL call_11_3_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_9_10
#define init_11_3_9_10
void FASTCALL call_11_3_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_9_11
#define init_11_3_9_11
void FASTCALL call_11_3_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_9_12
#define init_11_3_9_12
void FASTCALL call_11_3_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_9_13
#define init_11_3_9_13
void FASTCALL call_11_3_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_9_14
#define init_11_3_9_14
void FASTCALL call_11_3_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_9_15
#define init_11_3_9_15
void FASTCALL call_11_3_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_10_0
#define init_11_3_10_0
void FASTCALL call_11_3_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_10_1
#define init_11_3_10_1
void FASTCALL call_11_3_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_10_2
#define init_11_3_10_2
void FASTCALL call_11_3_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_10_3
#define init_11_3_10_3
void FASTCALL call_11_3_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_10_4
#define init_11_3_10_4
void FASTCALL call_11_3_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_10_5
#define init_11_3_10_5
void FASTCALL call_11_3_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_10_6
#define init_11_3_10_6
void FASTCALL call_11_3_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_10_7
#define init_11_3_10_7
void FASTCALL call_11_3_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_10_8
#define init_11_3_10_8
void FASTCALL call_11_3_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_10_9
#define init_11_3_10_9
void FASTCALL call_11_3_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_10_10
#define init_11_3_10_10
void FASTCALL call_11_3_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_10_11
#define init_11_3_10_11
void FASTCALL call_11_3_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_10_12
#define init_11_3_10_12
void FASTCALL call_11_3_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_10_13
#define init_11_3_10_13
void FASTCALL call_11_3_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_10_14
#define init_11_3_10_14
void FASTCALL call_11_3_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_10_15
#define init_11_3_10_15
void FASTCALL call_11_3_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_11_0
#define init_11_3_11_0
void FASTCALL call_11_3_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_11_1
#define init_11_3_11_1
void FASTCALL call_11_3_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_11_2
#define init_11_3_11_2
void FASTCALL call_11_3_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_11_3
#define init_11_3_11_3
void FASTCALL call_11_3_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_11_4
#define init_11_3_11_4
void FASTCALL call_11_3_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_11_5
#define init_11_3_11_5
void FASTCALL call_11_3_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_11_6
#define init_11_3_11_6
void FASTCALL call_11_3_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_11_7
#define init_11_3_11_7
void FASTCALL call_11_3_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_11_8
#define init_11_3_11_8
void FASTCALL call_11_3_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_11_9
#define init_11_3_11_9
void FASTCALL call_11_3_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_11_10
#define init_11_3_11_10
void FASTCALL call_11_3_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_11_11
#define init_11_3_11_11
void FASTCALL call_11_3_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_11_12
#define init_11_3_11_12
void FASTCALL call_11_3_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_11_13
#define init_11_3_11_13
void FASTCALL call_11_3_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_11_14
#define init_11_3_11_14
void FASTCALL call_11_3_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_11_15
#define init_11_3_11_15
void FASTCALL call_11_3_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_12_0
#define init_11_3_12_0
void FASTCALL call_11_3_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_12_1
#define init_11_3_12_1
void FASTCALL call_11_3_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_12_2
#define init_11_3_12_2
void FASTCALL call_11_3_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_12_3
#define init_11_3_12_3
void FASTCALL call_11_3_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_12_4
#define init_11_3_12_4
void FASTCALL call_11_3_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_12_5
#define init_11_3_12_5
void FASTCALL call_11_3_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_12_6
#define init_11_3_12_6
void FASTCALL call_11_3_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_12_7
#define init_11_3_12_7
void FASTCALL call_11_3_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_12_8
#define init_11_3_12_8
void FASTCALL call_11_3_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_12_9
#define init_11_3_12_9
void FASTCALL call_11_3_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_12_10
#define init_11_3_12_10
void FASTCALL call_11_3_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_12_11
#define init_11_3_12_11
void FASTCALL call_11_3_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_12_12
#define init_11_3_12_12
void FASTCALL call_11_3_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_12_13
#define init_11_3_12_13
void FASTCALL call_11_3_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_12_14
#define init_11_3_12_14
void FASTCALL call_11_3_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_12_15
#define init_11_3_12_15
void FASTCALL call_11_3_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_13_0
#define init_11_3_13_0
void FASTCALL call_11_3_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_13_1
#define init_11_3_13_1
void FASTCALL call_11_3_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_13_2
#define init_11_3_13_2
void FASTCALL call_11_3_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_13_3
#define init_11_3_13_3
void FASTCALL call_11_3_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_13_4
#define init_11_3_13_4
void FASTCALL call_11_3_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_13_5
#define init_11_3_13_5
void FASTCALL call_11_3_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_13_6
#define init_11_3_13_6
void FASTCALL call_11_3_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_13_7
#define init_11_3_13_7
void FASTCALL call_11_3_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_13_8
#define init_11_3_13_8
void FASTCALL call_11_3_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_13_9
#define init_11_3_13_9
void FASTCALL call_11_3_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_13_10
#define init_11_3_13_10
void FASTCALL call_11_3_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_13_11
#define init_11_3_13_11
void FASTCALL call_11_3_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_13_12
#define init_11_3_13_12
void FASTCALL call_11_3_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_13_13
#define init_11_3_13_13
void FASTCALL call_11_3_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_13_14
#define init_11_3_13_14
void FASTCALL call_11_3_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_13_15
#define init_11_3_13_15
void FASTCALL call_11_3_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_14_0
#define init_11_3_14_0
void FASTCALL call_11_3_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_14_1
#define init_11_3_14_1
void FASTCALL call_11_3_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_14_2
#define init_11_3_14_2
void FASTCALL call_11_3_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_14_3
#define init_11_3_14_3
void FASTCALL call_11_3_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_14_4
#define init_11_3_14_4
void FASTCALL call_11_3_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_14_5
#define init_11_3_14_5
void FASTCALL call_11_3_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_14_6
#define init_11_3_14_6
void FASTCALL call_11_3_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_14_7
#define init_11_3_14_7
void FASTCALL call_11_3_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_14_8
#define init_11_3_14_8
void FASTCALL call_11_3_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_14_9
#define init_11_3_14_9
void FASTCALL call_11_3_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_14_10
#define init_11_3_14_10
void FASTCALL call_11_3_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_14_11
#define init_11_3_14_11
void FASTCALL call_11_3_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_14_12
#define init_11_3_14_12
void FASTCALL call_11_3_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_14_13
#define init_11_3_14_13
void FASTCALL call_11_3_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_14_14
#define init_11_3_14_14
void FASTCALL call_11_3_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_14_15
#define init_11_3_14_15
void FASTCALL call_11_3_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_15_0
#define init_11_3_15_0
void FASTCALL call_11_3_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_15_1
#define init_11_3_15_1
void FASTCALL call_11_3_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_15_2
#define init_11_3_15_2
void FASTCALL call_11_3_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_15_3
#define init_11_3_15_3
void FASTCALL call_11_3_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_15_4
#define init_11_3_15_4
void FASTCALL call_11_3_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_15_5
#define init_11_3_15_5
void FASTCALL call_11_3_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_15_6
#define init_11_3_15_6
void FASTCALL call_11_3_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_15_7
#define init_11_3_15_7
void FASTCALL call_11_3_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_15_8
#define init_11_3_15_8
void FASTCALL call_11_3_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_15_9
#define init_11_3_15_9
void FASTCALL call_11_3_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_15_10
#define init_11_3_15_10
void FASTCALL call_11_3_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_15_11
#define init_11_3_15_11
void FASTCALL call_11_3_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_15_12
#define init_11_3_15_12
void FASTCALL call_11_3_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_15_13
#define init_11_3_15_13
void FASTCALL call_11_3_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_15_14
#define init_11_3_15_14
void FASTCALL call_11_3_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_3_15_15
#define init_11_3_15_15
void FASTCALL call_11_3_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_0_0
#define init_11_4_0_0
void FASTCALL call_11_4_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_0_1
#define init_11_4_0_1
void FASTCALL call_11_4_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_0_2
#define init_11_4_0_2
void FASTCALL call_11_4_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_0_3
#define init_11_4_0_3
void FASTCALL call_11_4_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_0_4
#define init_11_4_0_4
void FASTCALL call_11_4_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_0_5
#define init_11_4_0_5
void FASTCALL call_11_4_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_0_6
#define init_11_4_0_6
void FASTCALL call_11_4_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_0_7
#define init_11_4_0_7
void FASTCALL call_11_4_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_0_8
#define init_11_4_0_8
void FASTCALL call_11_4_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_0_9
#define init_11_4_0_9
void FASTCALL call_11_4_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_0_10
#define init_11_4_0_10
void FASTCALL call_11_4_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_0_11
#define init_11_4_0_11
void FASTCALL call_11_4_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_0_12
#define init_11_4_0_12
void FASTCALL call_11_4_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_0_13
#define init_11_4_0_13
void FASTCALL call_11_4_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_0_14
#define init_11_4_0_14
void FASTCALL call_11_4_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_0_15
#define init_11_4_0_15
void FASTCALL call_11_4_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_1_0
#define init_11_4_1_0
void FASTCALL call_11_4_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_1_1
#define init_11_4_1_1
void FASTCALL call_11_4_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_1_2
#define init_11_4_1_2
void FASTCALL call_11_4_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_1_3
#define init_11_4_1_3
void FASTCALL call_11_4_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_1_4
#define init_11_4_1_4
void FASTCALL call_11_4_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_1_5
#define init_11_4_1_5
void FASTCALL call_11_4_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_1_6
#define init_11_4_1_6
void FASTCALL call_11_4_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_1_7
#define init_11_4_1_7
void FASTCALL call_11_4_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_1_8
#define init_11_4_1_8
void FASTCALL call_11_4_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_1_9
#define init_11_4_1_9
void FASTCALL call_11_4_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_1_10
#define init_11_4_1_10
void FASTCALL call_11_4_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_1_11
#define init_11_4_1_11
void FASTCALL call_11_4_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_1_12
#define init_11_4_1_12
void FASTCALL call_11_4_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_1_13
#define init_11_4_1_13
void FASTCALL call_11_4_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_1_14
#define init_11_4_1_14
void FASTCALL call_11_4_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_1_15
#define init_11_4_1_15
void FASTCALL call_11_4_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_2_0
#define init_11_4_2_0
void FASTCALL call_11_4_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_2_1
#define init_11_4_2_1
void FASTCALL call_11_4_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_2_2
#define init_11_4_2_2
void FASTCALL call_11_4_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_2_3
#define init_11_4_2_3
void FASTCALL call_11_4_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_2_4
#define init_11_4_2_4
void FASTCALL call_11_4_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_2_5
#define init_11_4_2_5
void FASTCALL call_11_4_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_2_6
#define init_11_4_2_6
void FASTCALL call_11_4_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_2_7
#define init_11_4_2_7
void FASTCALL call_11_4_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_2_8
#define init_11_4_2_8
void FASTCALL call_11_4_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_2_9
#define init_11_4_2_9
void FASTCALL call_11_4_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_2_10
#define init_11_4_2_10
void FASTCALL call_11_4_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_2_11
#define init_11_4_2_11
void FASTCALL call_11_4_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_2_12
#define init_11_4_2_12
void FASTCALL call_11_4_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_2_13
#define init_11_4_2_13
void FASTCALL call_11_4_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_2_14
#define init_11_4_2_14
void FASTCALL call_11_4_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_2_15
#define init_11_4_2_15
void FASTCALL call_11_4_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_3_0
#define init_11_4_3_0
void FASTCALL call_11_4_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_3_1
#define init_11_4_3_1
void FASTCALL call_11_4_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_3_2
#define init_11_4_3_2
void FASTCALL call_11_4_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_3_3
#define init_11_4_3_3
void FASTCALL call_11_4_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_3_4
#define init_11_4_3_4
void FASTCALL call_11_4_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_3_5
#define init_11_4_3_5
void FASTCALL call_11_4_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_3_6
#define init_11_4_3_6
void FASTCALL call_11_4_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_3_7
#define init_11_4_3_7
void FASTCALL call_11_4_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_3_8
#define init_11_4_3_8
void FASTCALL call_11_4_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_3_9
#define init_11_4_3_9
void FASTCALL call_11_4_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_3_10
#define init_11_4_3_10
void FASTCALL call_11_4_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_3_11
#define init_11_4_3_11
void FASTCALL call_11_4_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_3_12
#define init_11_4_3_12
void FASTCALL call_11_4_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_3_13
#define init_11_4_3_13
void FASTCALL call_11_4_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_3_14
#define init_11_4_3_14
void FASTCALL call_11_4_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_3_15
#define init_11_4_3_15
void FASTCALL call_11_4_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_4_0
#define init_11_4_4_0
void FASTCALL call_11_4_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_4_1
#define init_11_4_4_1
void FASTCALL call_11_4_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_4_2
#define init_11_4_4_2
void FASTCALL call_11_4_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_4_3
#define init_11_4_4_3
void FASTCALL call_11_4_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_4_4
#define init_11_4_4_4
void FASTCALL call_11_4_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_4_5
#define init_11_4_4_5
void FASTCALL call_11_4_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_4_6
#define init_11_4_4_6
void FASTCALL call_11_4_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_4_7
#define init_11_4_4_7
void FASTCALL call_11_4_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_4_8
#define init_11_4_4_8
void FASTCALL call_11_4_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_4_9
#define init_11_4_4_9
void FASTCALL call_11_4_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_4_10
#define init_11_4_4_10
void FASTCALL call_11_4_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_4_11
#define init_11_4_4_11
void FASTCALL call_11_4_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_4_12
#define init_11_4_4_12
void FASTCALL call_11_4_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_4_13
#define init_11_4_4_13
void FASTCALL call_11_4_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_4_14
#define init_11_4_4_14
void FASTCALL call_11_4_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_4_15
#define init_11_4_4_15
void FASTCALL call_11_4_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_5_0
#define init_11_4_5_0
void FASTCALL call_11_4_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_5_1
#define init_11_4_5_1
void FASTCALL call_11_4_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_5_2
#define init_11_4_5_2
void FASTCALL call_11_4_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_5_3
#define init_11_4_5_3
void FASTCALL call_11_4_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_5_4
#define init_11_4_5_4
void FASTCALL call_11_4_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_5_5
#define init_11_4_5_5
void FASTCALL call_11_4_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_5_6
#define init_11_4_5_6
void FASTCALL call_11_4_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_5_7
#define init_11_4_5_7
void FASTCALL call_11_4_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_5_8
#define init_11_4_5_8
void FASTCALL call_11_4_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_5_9
#define init_11_4_5_9
void FASTCALL call_11_4_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_5_10
#define init_11_4_5_10
void FASTCALL call_11_4_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_5_11
#define init_11_4_5_11
void FASTCALL call_11_4_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_5_12
#define init_11_4_5_12
void FASTCALL call_11_4_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_5_13
#define init_11_4_5_13
void FASTCALL call_11_4_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_5_14
#define init_11_4_5_14
void FASTCALL call_11_4_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_5_15
#define init_11_4_5_15
void FASTCALL call_11_4_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_6_0
#define init_11_4_6_0
void FASTCALL call_11_4_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_6_1
#define init_11_4_6_1
void FASTCALL call_11_4_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_6_2
#define init_11_4_6_2
void FASTCALL call_11_4_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_6_3
#define init_11_4_6_3
void FASTCALL call_11_4_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_6_4
#define init_11_4_6_4
void FASTCALL call_11_4_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_6_5
#define init_11_4_6_5
void FASTCALL call_11_4_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_6_6
#define init_11_4_6_6
void FASTCALL call_11_4_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_6_7
#define init_11_4_6_7
void FASTCALL call_11_4_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_6_8
#define init_11_4_6_8
void FASTCALL call_11_4_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_6_9
#define init_11_4_6_9
void FASTCALL call_11_4_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_6_10
#define init_11_4_6_10
void FASTCALL call_11_4_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_6_11
#define init_11_4_6_11
void FASTCALL call_11_4_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_6_12
#define init_11_4_6_12
void FASTCALL call_11_4_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_6_13
#define init_11_4_6_13
void FASTCALL call_11_4_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_6_14
#define init_11_4_6_14
void FASTCALL call_11_4_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_6_15
#define init_11_4_6_15
void FASTCALL call_11_4_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_7_0
#define init_11_4_7_0
void FASTCALL call_11_4_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_7_1
#define init_11_4_7_1
void FASTCALL call_11_4_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_7_2
#define init_11_4_7_2
void FASTCALL call_11_4_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_7_3
#define init_11_4_7_3
void FASTCALL call_11_4_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_7_4
#define init_11_4_7_4
void FASTCALL call_11_4_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_7_5
#define init_11_4_7_5
void FASTCALL call_11_4_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_7_6
#define init_11_4_7_6
void FASTCALL call_11_4_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_7_7
#define init_11_4_7_7
void FASTCALL call_11_4_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_7_8
#define init_11_4_7_8
void FASTCALL call_11_4_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_7_9
#define init_11_4_7_9
void FASTCALL call_11_4_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_7_10
#define init_11_4_7_10
void FASTCALL call_11_4_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_7_11
#define init_11_4_7_11
void FASTCALL call_11_4_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_7_12
#define init_11_4_7_12
void FASTCALL call_11_4_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_7_13
#define init_11_4_7_13
void FASTCALL call_11_4_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_7_14
#define init_11_4_7_14
void FASTCALL call_11_4_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_7_15
#define init_11_4_7_15
void FASTCALL call_11_4_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_8_0
#define init_11_4_8_0
void FASTCALL call_11_4_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_8_1
#define init_11_4_8_1
void FASTCALL call_11_4_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_8_2
#define init_11_4_8_2
void FASTCALL call_11_4_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_8_3
#define init_11_4_8_3
void FASTCALL call_11_4_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_8_4
#define init_11_4_8_4
void FASTCALL call_11_4_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_8_5
#define init_11_4_8_5
void FASTCALL call_11_4_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_8_6
#define init_11_4_8_6
void FASTCALL call_11_4_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_8_7
#define init_11_4_8_7
void FASTCALL call_11_4_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_8_8
#define init_11_4_8_8
void FASTCALL call_11_4_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_8_9
#define init_11_4_8_9
void FASTCALL call_11_4_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_8_10
#define init_11_4_8_10
void FASTCALL call_11_4_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_8_11
#define init_11_4_8_11
void FASTCALL call_11_4_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_8_12
#define init_11_4_8_12
void FASTCALL call_11_4_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_8_13
#define init_11_4_8_13
void FASTCALL call_11_4_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_8_14
#define init_11_4_8_14
void FASTCALL call_11_4_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_8_15
#define init_11_4_8_15
void FASTCALL call_11_4_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_9_0
#define init_11_4_9_0
void FASTCALL call_11_4_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_9_1
#define init_11_4_9_1
void FASTCALL call_11_4_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_9_2
#define init_11_4_9_2
void FASTCALL call_11_4_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_9_3
#define init_11_4_9_3
void FASTCALL call_11_4_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_9_4
#define init_11_4_9_4
void FASTCALL call_11_4_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_9_5
#define init_11_4_9_5
void FASTCALL call_11_4_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_9_6
#define init_11_4_9_6
void FASTCALL call_11_4_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_9_7
#define init_11_4_9_7
void FASTCALL call_11_4_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_9_8
#define init_11_4_9_8
void FASTCALL call_11_4_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_9_9
#define init_11_4_9_9
void FASTCALL call_11_4_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_9_10
#define init_11_4_9_10
void FASTCALL call_11_4_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_9_11
#define init_11_4_9_11
void FASTCALL call_11_4_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_9_12
#define init_11_4_9_12
void FASTCALL call_11_4_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_9_13
#define init_11_4_9_13
void FASTCALL call_11_4_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_9_14
#define init_11_4_9_14
void FASTCALL call_11_4_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_9_15
#define init_11_4_9_15
void FASTCALL call_11_4_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_10_0
#define init_11_4_10_0
void FASTCALL call_11_4_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_10_1
#define init_11_4_10_1
void FASTCALL call_11_4_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_10_2
#define init_11_4_10_2
void FASTCALL call_11_4_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_10_3
#define init_11_4_10_3
void FASTCALL call_11_4_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_10_4
#define init_11_4_10_4
void FASTCALL call_11_4_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_10_5
#define init_11_4_10_5
void FASTCALL call_11_4_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_10_6
#define init_11_4_10_6
void FASTCALL call_11_4_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_10_7
#define init_11_4_10_7
void FASTCALL call_11_4_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_10_8
#define init_11_4_10_8
void FASTCALL call_11_4_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_10_9
#define init_11_4_10_9
void FASTCALL call_11_4_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_10_10
#define init_11_4_10_10
void FASTCALL call_11_4_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_10_11
#define init_11_4_10_11
void FASTCALL call_11_4_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_10_12
#define init_11_4_10_12
void FASTCALL call_11_4_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_10_13
#define init_11_4_10_13
void FASTCALL call_11_4_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_10_14
#define init_11_4_10_14
void FASTCALL call_11_4_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_10_15
#define init_11_4_10_15
void FASTCALL call_11_4_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_11_0
#define init_11_4_11_0
void FASTCALL call_11_4_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_11_1
#define init_11_4_11_1
void FASTCALL call_11_4_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_11_2
#define init_11_4_11_2
void FASTCALL call_11_4_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_11_3
#define init_11_4_11_3
void FASTCALL call_11_4_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_11_4
#define init_11_4_11_4
void FASTCALL call_11_4_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_11_5
#define init_11_4_11_5
void FASTCALL call_11_4_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_11_6
#define init_11_4_11_6
void FASTCALL call_11_4_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_11_7
#define init_11_4_11_7
void FASTCALL call_11_4_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_11_8
#define init_11_4_11_8
void FASTCALL call_11_4_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_11_9
#define init_11_4_11_9
void FASTCALL call_11_4_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_11_10
#define init_11_4_11_10
void FASTCALL call_11_4_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_11_11
#define init_11_4_11_11
void FASTCALL call_11_4_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_11_12
#define init_11_4_11_12
void FASTCALL call_11_4_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_11_13
#define init_11_4_11_13
void FASTCALL call_11_4_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_11_14
#define init_11_4_11_14
void FASTCALL call_11_4_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_11_15
#define init_11_4_11_15
void FASTCALL call_11_4_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_12_0
#define init_11_4_12_0
void FASTCALL call_11_4_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_12_1
#define init_11_4_12_1
void FASTCALL call_11_4_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_12_2
#define init_11_4_12_2
void FASTCALL call_11_4_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_12_3
#define init_11_4_12_3
void FASTCALL call_11_4_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_12_4
#define init_11_4_12_4
void FASTCALL call_11_4_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_12_5
#define init_11_4_12_5
void FASTCALL call_11_4_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_12_6
#define init_11_4_12_6
void FASTCALL call_11_4_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_12_7
#define init_11_4_12_7
void FASTCALL call_11_4_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_12_8
#define init_11_4_12_8
void FASTCALL call_11_4_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_12_9
#define init_11_4_12_9
void FASTCALL call_11_4_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_12_10
#define init_11_4_12_10
void FASTCALL call_11_4_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_12_11
#define init_11_4_12_11
void FASTCALL call_11_4_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_12_12
#define init_11_4_12_12
void FASTCALL call_11_4_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_12_13
#define init_11_4_12_13
void FASTCALL call_11_4_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_12_14
#define init_11_4_12_14
void FASTCALL call_11_4_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_12_15
#define init_11_4_12_15
void FASTCALL call_11_4_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_13_0
#define init_11_4_13_0
void FASTCALL call_11_4_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_13_1
#define init_11_4_13_1
void FASTCALL call_11_4_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_13_2
#define init_11_4_13_2
void FASTCALL call_11_4_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_13_3
#define init_11_4_13_3
void FASTCALL call_11_4_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_13_4
#define init_11_4_13_4
void FASTCALL call_11_4_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_13_5
#define init_11_4_13_5
void FASTCALL call_11_4_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_13_6
#define init_11_4_13_6
void FASTCALL call_11_4_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_13_7
#define init_11_4_13_7
void FASTCALL call_11_4_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_13_8
#define init_11_4_13_8
void FASTCALL call_11_4_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_13_9
#define init_11_4_13_9
void FASTCALL call_11_4_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_13_10
#define init_11_4_13_10
void FASTCALL call_11_4_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_13_11
#define init_11_4_13_11
void FASTCALL call_11_4_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_13_12
#define init_11_4_13_12
void FASTCALL call_11_4_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_13_13
#define init_11_4_13_13
void FASTCALL call_11_4_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_13_14
#define init_11_4_13_14
void FASTCALL call_11_4_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_13_15
#define init_11_4_13_15
void FASTCALL call_11_4_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_14_0
#define init_11_4_14_0
void FASTCALL call_11_4_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_14_1
#define init_11_4_14_1
void FASTCALL call_11_4_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_14_2
#define init_11_4_14_2
void FASTCALL call_11_4_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_14_3
#define init_11_4_14_3
void FASTCALL call_11_4_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_14_4
#define init_11_4_14_4
void FASTCALL call_11_4_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_14_5
#define init_11_4_14_5
void FASTCALL call_11_4_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_14_6
#define init_11_4_14_6
void FASTCALL call_11_4_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_14_7
#define init_11_4_14_7
void FASTCALL call_11_4_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_14_8
#define init_11_4_14_8
void FASTCALL call_11_4_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_14_9
#define init_11_4_14_9
void FASTCALL call_11_4_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_14_10
#define init_11_4_14_10
void FASTCALL call_11_4_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_14_11
#define init_11_4_14_11
void FASTCALL call_11_4_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_14_12
#define init_11_4_14_12
void FASTCALL call_11_4_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_14_13
#define init_11_4_14_13
void FASTCALL call_11_4_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_14_14
#define init_11_4_14_14
void FASTCALL call_11_4_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_14_15
#define init_11_4_14_15
void FASTCALL call_11_4_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_15_0
#define init_11_4_15_0
void FASTCALL call_11_4_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_15_1
#define init_11_4_15_1
void FASTCALL call_11_4_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_15_2
#define init_11_4_15_2
void FASTCALL call_11_4_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_15_3
#define init_11_4_15_3
void FASTCALL call_11_4_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_15_4
#define init_11_4_15_4
void FASTCALL call_11_4_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_15_5
#define init_11_4_15_5
void FASTCALL call_11_4_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_15_6
#define init_11_4_15_6
void FASTCALL call_11_4_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_15_7
#define init_11_4_15_7
void FASTCALL call_11_4_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_15_8
#define init_11_4_15_8
void FASTCALL call_11_4_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_15_9
#define init_11_4_15_9
void FASTCALL call_11_4_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_15_10
#define init_11_4_15_10
void FASTCALL call_11_4_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_15_11
#define init_11_4_15_11
void FASTCALL call_11_4_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_15_12
#define init_11_4_15_12
void FASTCALL call_11_4_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_15_13
#define init_11_4_15_13
void FASTCALL call_11_4_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_15_14
#define init_11_4_15_14
void FASTCALL call_11_4_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_4_15_15
#define init_11_4_15_15
void FASTCALL call_11_4_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_0_0
#define init_11_5_0_0
void FASTCALL call_11_5_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_0_1
#define init_11_5_0_1
void FASTCALL call_11_5_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_0_2
#define init_11_5_0_2
void FASTCALL call_11_5_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_0_3
#define init_11_5_0_3
void FASTCALL call_11_5_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_0_4
#define init_11_5_0_4
void FASTCALL call_11_5_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_0_5
#define init_11_5_0_5
void FASTCALL call_11_5_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_0_6
#define init_11_5_0_6
void FASTCALL call_11_5_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_0_7
#define init_11_5_0_7
void FASTCALL call_11_5_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_0_8
#define init_11_5_0_8
void FASTCALL call_11_5_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_0_9
#define init_11_5_0_9
void FASTCALL call_11_5_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_0_10
#define init_11_5_0_10
void FASTCALL call_11_5_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_0_11
#define init_11_5_0_11
void FASTCALL call_11_5_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_0_12
#define init_11_5_0_12
void FASTCALL call_11_5_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_0_13
#define init_11_5_0_13
void FASTCALL call_11_5_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_0_14
#define init_11_5_0_14
void FASTCALL call_11_5_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_0_15
#define init_11_5_0_15
void FASTCALL call_11_5_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_1_0
#define init_11_5_1_0
void FASTCALL call_11_5_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_1_1
#define init_11_5_1_1
void FASTCALL call_11_5_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_1_2
#define init_11_5_1_2
void FASTCALL call_11_5_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_1_3
#define init_11_5_1_3
void FASTCALL call_11_5_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_1_4
#define init_11_5_1_4
void FASTCALL call_11_5_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_1_5
#define init_11_5_1_5
void FASTCALL call_11_5_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_1_6
#define init_11_5_1_6
void FASTCALL call_11_5_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_1_7
#define init_11_5_1_7
void FASTCALL call_11_5_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_1_8
#define init_11_5_1_8
void FASTCALL call_11_5_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_1_9
#define init_11_5_1_9
void FASTCALL call_11_5_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_1_10
#define init_11_5_1_10
void FASTCALL call_11_5_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_1_11
#define init_11_5_1_11
void FASTCALL call_11_5_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_1_12
#define init_11_5_1_12
void FASTCALL call_11_5_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_1_13
#define init_11_5_1_13
void FASTCALL call_11_5_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_1_14
#define init_11_5_1_14
void FASTCALL call_11_5_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_1_15
#define init_11_5_1_15
void FASTCALL call_11_5_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_2_0
#define init_11_5_2_0
void FASTCALL call_11_5_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_2_1
#define init_11_5_2_1
void FASTCALL call_11_5_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_2_2
#define init_11_5_2_2
void FASTCALL call_11_5_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_2_3
#define init_11_5_2_3
void FASTCALL call_11_5_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_2_4
#define init_11_5_2_4
void FASTCALL call_11_5_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_2_5
#define init_11_5_2_5
void FASTCALL call_11_5_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_2_6
#define init_11_5_2_6
void FASTCALL call_11_5_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_2_7
#define init_11_5_2_7
void FASTCALL call_11_5_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_2_8
#define init_11_5_2_8
void FASTCALL call_11_5_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_2_9
#define init_11_5_2_9
void FASTCALL call_11_5_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_2_10
#define init_11_5_2_10
void FASTCALL call_11_5_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_2_11
#define init_11_5_2_11
void FASTCALL call_11_5_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_2_12
#define init_11_5_2_12
void FASTCALL call_11_5_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_2_13
#define init_11_5_2_13
void FASTCALL call_11_5_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_2_14
#define init_11_5_2_14
void FASTCALL call_11_5_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_2_15
#define init_11_5_2_15
void FASTCALL call_11_5_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_3_0
#define init_11_5_3_0
void FASTCALL call_11_5_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_3_1
#define init_11_5_3_1
void FASTCALL call_11_5_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_3_2
#define init_11_5_3_2
void FASTCALL call_11_5_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_3_3
#define init_11_5_3_3
void FASTCALL call_11_5_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_3_4
#define init_11_5_3_4
void FASTCALL call_11_5_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_3_5
#define init_11_5_3_5
void FASTCALL call_11_5_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_3_6
#define init_11_5_3_6
void FASTCALL call_11_5_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_3_7
#define init_11_5_3_7
void FASTCALL call_11_5_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_3_8
#define init_11_5_3_8
void FASTCALL call_11_5_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_3_9
#define init_11_5_3_9
void FASTCALL call_11_5_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_3_10
#define init_11_5_3_10
void FASTCALL call_11_5_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_3_11
#define init_11_5_3_11
void FASTCALL call_11_5_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_3_12
#define init_11_5_3_12
void FASTCALL call_11_5_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_3_13
#define init_11_5_3_13
void FASTCALL call_11_5_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_3_14
#define init_11_5_3_14
void FASTCALL call_11_5_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_3_15
#define init_11_5_3_15
void FASTCALL call_11_5_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_4_0
#define init_11_5_4_0
void FASTCALL call_11_5_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_4_1
#define init_11_5_4_1
void FASTCALL call_11_5_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_4_2
#define init_11_5_4_2
void FASTCALL call_11_5_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_4_3
#define init_11_5_4_3
void FASTCALL call_11_5_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_4_4
#define init_11_5_4_4
void FASTCALL call_11_5_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_4_5
#define init_11_5_4_5
void FASTCALL call_11_5_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_4_6
#define init_11_5_4_6
void FASTCALL call_11_5_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_4_7
#define init_11_5_4_7
void FASTCALL call_11_5_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_4_8
#define init_11_5_4_8
void FASTCALL call_11_5_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_4_9
#define init_11_5_4_9
void FASTCALL call_11_5_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_4_10
#define init_11_5_4_10
void FASTCALL call_11_5_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_4_11
#define init_11_5_4_11
void FASTCALL call_11_5_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_4_12
#define init_11_5_4_12
void FASTCALL call_11_5_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_4_13
#define init_11_5_4_13
void FASTCALL call_11_5_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_4_14
#define init_11_5_4_14
void FASTCALL call_11_5_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_4_15
#define init_11_5_4_15
void FASTCALL call_11_5_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_5_0
#define init_11_5_5_0
void FASTCALL call_11_5_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_5_1
#define init_11_5_5_1
void FASTCALL call_11_5_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_5_2
#define init_11_5_5_2
void FASTCALL call_11_5_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_5_3
#define init_11_5_5_3
void FASTCALL call_11_5_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_5_4
#define init_11_5_5_4
void FASTCALL call_11_5_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_5_5
#define init_11_5_5_5
void FASTCALL call_11_5_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_5_6
#define init_11_5_5_6
void FASTCALL call_11_5_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_5_7
#define init_11_5_5_7
void FASTCALL call_11_5_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_5_8
#define init_11_5_5_8
void FASTCALL call_11_5_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_5_9
#define init_11_5_5_9
void FASTCALL call_11_5_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_5_10
#define init_11_5_5_10
void FASTCALL call_11_5_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_5_11
#define init_11_5_5_11
void FASTCALL call_11_5_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_5_12
#define init_11_5_5_12
void FASTCALL call_11_5_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_5_13
#define init_11_5_5_13
void FASTCALL call_11_5_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_5_14
#define init_11_5_5_14
void FASTCALL call_11_5_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_5_15
#define init_11_5_5_15
void FASTCALL call_11_5_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_6_0
#define init_11_5_6_0
void FASTCALL call_11_5_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_6_1
#define init_11_5_6_1
void FASTCALL call_11_5_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_6_2
#define init_11_5_6_2
void FASTCALL call_11_5_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_6_3
#define init_11_5_6_3
void FASTCALL call_11_5_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_6_4
#define init_11_5_6_4
void FASTCALL call_11_5_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_6_5
#define init_11_5_6_5
void FASTCALL call_11_5_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_6_6
#define init_11_5_6_6
void FASTCALL call_11_5_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_6_7
#define init_11_5_6_7
void FASTCALL call_11_5_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_6_8
#define init_11_5_6_8
void FASTCALL call_11_5_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_6_9
#define init_11_5_6_9
void FASTCALL call_11_5_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_6_10
#define init_11_5_6_10
void FASTCALL call_11_5_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_6_11
#define init_11_5_6_11
void FASTCALL call_11_5_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_6_12
#define init_11_5_6_12
void FASTCALL call_11_5_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_6_13
#define init_11_5_6_13
void FASTCALL call_11_5_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_6_14
#define init_11_5_6_14
void FASTCALL call_11_5_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_6_15
#define init_11_5_6_15
void FASTCALL call_11_5_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_7_0
#define init_11_5_7_0
void FASTCALL call_11_5_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_7_1
#define init_11_5_7_1
void FASTCALL call_11_5_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_7_2
#define init_11_5_7_2
void FASTCALL call_11_5_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_7_3
#define init_11_5_7_3
void FASTCALL call_11_5_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_7_4
#define init_11_5_7_4
void FASTCALL call_11_5_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_7_5
#define init_11_5_7_5
void FASTCALL call_11_5_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_7_6
#define init_11_5_7_6
void FASTCALL call_11_5_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_7_7
#define init_11_5_7_7
void FASTCALL call_11_5_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_7_8
#define init_11_5_7_8
void FASTCALL call_11_5_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_7_9
#define init_11_5_7_9
void FASTCALL call_11_5_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_7_10
#define init_11_5_7_10
void FASTCALL call_11_5_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_7_11
#define init_11_5_7_11
void FASTCALL call_11_5_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_7_12
#define init_11_5_7_12
void FASTCALL call_11_5_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_7_13
#define init_11_5_7_13
void FASTCALL call_11_5_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_7_14
#define init_11_5_7_14
void FASTCALL call_11_5_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_7_15
#define init_11_5_7_15
void FASTCALL call_11_5_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_8_0
#define init_11_5_8_0
void FASTCALL call_11_5_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_8_1
#define init_11_5_8_1
void FASTCALL call_11_5_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_8_2
#define init_11_5_8_2
void FASTCALL call_11_5_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_8_3
#define init_11_5_8_3
void FASTCALL call_11_5_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_8_4
#define init_11_5_8_4
void FASTCALL call_11_5_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_8_5
#define init_11_5_8_5
void FASTCALL call_11_5_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_8_6
#define init_11_5_8_6
void FASTCALL call_11_5_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_8_7
#define init_11_5_8_7
void FASTCALL call_11_5_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_8_8
#define init_11_5_8_8
void FASTCALL call_11_5_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_8_9
#define init_11_5_8_9
void FASTCALL call_11_5_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_8_10
#define init_11_5_8_10
void FASTCALL call_11_5_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_8_11
#define init_11_5_8_11
void FASTCALL call_11_5_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_8_12
#define init_11_5_8_12
void FASTCALL call_11_5_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_8_13
#define init_11_5_8_13
void FASTCALL call_11_5_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_8_14
#define init_11_5_8_14
void FASTCALL call_11_5_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_8_15
#define init_11_5_8_15
void FASTCALL call_11_5_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_9_0
#define init_11_5_9_0
void FASTCALL call_11_5_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_9_1
#define init_11_5_9_1
void FASTCALL call_11_5_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_9_2
#define init_11_5_9_2
void FASTCALL call_11_5_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_9_3
#define init_11_5_9_3
void FASTCALL call_11_5_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_9_4
#define init_11_5_9_4
void FASTCALL call_11_5_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_9_5
#define init_11_5_9_5
void FASTCALL call_11_5_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_9_6
#define init_11_5_9_6
void FASTCALL call_11_5_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_9_7
#define init_11_5_9_7
void FASTCALL call_11_5_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_9_8
#define init_11_5_9_8
void FASTCALL call_11_5_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_9_9
#define init_11_5_9_9
void FASTCALL call_11_5_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_9_10
#define init_11_5_9_10
void FASTCALL call_11_5_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_9_11
#define init_11_5_9_11
void FASTCALL call_11_5_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_9_12
#define init_11_5_9_12
void FASTCALL call_11_5_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_9_13
#define init_11_5_9_13
void FASTCALL call_11_5_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_9_14
#define init_11_5_9_14
void FASTCALL call_11_5_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_9_15
#define init_11_5_9_15
void FASTCALL call_11_5_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_10_0
#define init_11_5_10_0
void FASTCALL call_11_5_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_10_1
#define init_11_5_10_1
void FASTCALL call_11_5_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_10_2
#define init_11_5_10_2
void FASTCALL call_11_5_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_10_3
#define init_11_5_10_3
void FASTCALL call_11_5_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_10_4
#define init_11_5_10_4
void FASTCALL call_11_5_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_10_5
#define init_11_5_10_5
void FASTCALL call_11_5_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_10_6
#define init_11_5_10_6
void FASTCALL call_11_5_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_10_7
#define init_11_5_10_7
void FASTCALL call_11_5_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_10_8
#define init_11_5_10_8
void FASTCALL call_11_5_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_10_9
#define init_11_5_10_9
void FASTCALL call_11_5_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_10_10
#define init_11_5_10_10
void FASTCALL call_11_5_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_10_11
#define init_11_5_10_11
void FASTCALL call_11_5_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_10_12
#define init_11_5_10_12
void FASTCALL call_11_5_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_10_13
#define init_11_5_10_13
void FASTCALL call_11_5_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_10_14
#define init_11_5_10_14
void FASTCALL call_11_5_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_10_15
#define init_11_5_10_15
void FASTCALL call_11_5_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_11_0
#define init_11_5_11_0
void FASTCALL call_11_5_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_11_1
#define init_11_5_11_1
void FASTCALL call_11_5_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_11_2
#define init_11_5_11_2
void FASTCALL call_11_5_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_11_3
#define init_11_5_11_3
void FASTCALL call_11_5_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_11_4
#define init_11_5_11_4
void FASTCALL call_11_5_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_11_5
#define init_11_5_11_5
void FASTCALL call_11_5_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_11_6
#define init_11_5_11_6
void FASTCALL call_11_5_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_11_7
#define init_11_5_11_7
void FASTCALL call_11_5_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_11_8
#define init_11_5_11_8
void FASTCALL call_11_5_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_11_9
#define init_11_5_11_9
void FASTCALL call_11_5_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_11_10
#define init_11_5_11_10
void FASTCALL call_11_5_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_11_11
#define init_11_5_11_11
void FASTCALL call_11_5_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_11_12
#define init_11_5_11_12
void FASTCALL call_11_5_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_11_13
#define init_11_5_11_13
void FASTCALL call_11_5_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_11_14
#define init_11_5_11_14
void FASTCALL call_11_5_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_11_15
#define init_11_5_11_15
void FASTCALL call_11_5_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_12_0
#define init_11_5_12_0
void FASTCALL call_11_5_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_12_1
#define init_11_5_12_1
void FASTCALL call_11_5_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_12_2
#define init_11_5_12_2
void FASTCALL call_11_5_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_12_3
#define init_11_5_12_3
void FASTCALL call_11_5_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_12_4
#define init_11_5_12_4
void FASTCALL call_11_5_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_12_5
#define init_11_5_12_5
void FASTCALL call_11_5_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_12_6
#define init_11_5_12_6
void FASTCALL call_11_5_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_12_7
#define init_11_5_12_7
void FASTCALL call_11_5_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_12_8
#define init_11_5_12_8
void FASTCALL call_11_5_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_12_9
#define init_11_5_12_9
void FASTCALL call_11_5_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_12_10
#define init_11_5_12_10
void FASTCALL call_11_5_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_12_11
#define init_11_5_12_11
void FASTCALL call_11_5_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_12_12
#define init_11_5_12_12
void FASTCALL call_11_5_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_12_13
#define init_11_5_12_13
void FASTCALL call_11_5_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_12_14
#define init_11_5_12_14
void FASTCALL call_11_5_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_12_15
#define init_11_5_12_15
void FASTCALL call_11_5_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_13_0
#define init_11_5_13_0
void FASTCALL call_11_5_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_13_1
#define init_11_5_13_1
void FASTCALL call_11_5_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_13_2
#define init_11_5_13_2
void FASTCALL call_11_5_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_13_3
#define init_11_5_13_3
void FASTCALL call_11_5_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_13_4
#define init_11_5_13_4
void FASTCALL call_11_5_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_13_5
#define init_11_5_13_5
void FASTCALL call_11_5_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_13_6
#define init_11_5_13_6
void FASTCALL call_11_5_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_13_7
#define init_11_5_13_7
void FASTCALL call_11_5_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_13_8
#define init_11_5_13_8
void FASTCALL call_11_5_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_13_9
#define init_11_5_13_9
void FASTCALL call_11_5_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_13_10
#define init_11_5_13_10
void FASTCALL call_11_5_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_13_11
#define init_11_5_13_11
void FASTCALL call_11_5_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_13_12
#define init_11_5_13_12
void FASTCALL call_11_5_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_13_13
#define init_11_5_13_13
void FASTCALL call_11_5_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_13_14
#define init_11_5_13_14
void FASTCALL call_11_5_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_13_15
#define init_11_5_13_15
void FASTCALL call_11_5_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_14_0
#define init_11_5_14_0
void FASTCALL call_11_5_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_14_1
#define init_11_5_14_1
void FASTCALL call_11_5_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_14_2
#define init_11_5_14_2
void FASTCALL call_11_5_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_14_3
#define init_11_5_14_3
void FASTCALL call_11_5_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_14_4
#define init_11_5_14_4
void FASTCALL call_11_5_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_14_5
#define init_11_5_14_5
void FASTCALL call_11_5_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_14_6
#define init_11_5_14_6
void FASTCALL call_11_5_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_14_7
#define init_11_5_14_7
void FASTCALL call_11_5_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_14_8
#define init_11_5_14_8
void FASTCALL call_11_5_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_14_9
#define init_11_5_14_9
void FASTCALL call_11_5_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_14_10
#define init_11_5_14_10
void FASTCALL call_11_5_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_14_11
#define init_11_5_14_11
void FASTCALL call_11_5_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_14_12
#define init_11_5_14_12
void FASTCALL call_11_5_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_14_13
#define init_11_5_14_13
void FASTCALL call_11_5_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_14_14
#define init_11_5_14_14
void FASTCALL call_11_5_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_14_15
#define init_11_5_14_15
void FASTCALL call_11_5_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_15_0
#define init_11_5_15_0
void FASTCALL call_11_5_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_15_1
#define init_11_5_15_1
void FASTCALL call_11_5_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_15_2
#define init_11_5_15_2
void FASTCALL call_11_5_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_15_3
#define init_11_5_15_3
void FASTCALL call_11_5_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_15_4
#define init_11_5_15_4
void FASTCALL call_11_5_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_15_5
#define init_11_5_15_5
void FASTCALL call_11_5_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_15_6
#define init_11_5_15_6
void FASTCALL call_11_5_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_15_7
#define init_11_5_15_7
void FASTCALL call_11_5_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_15_8
#define init_11_5_15_8
void FASTCALL call_11_5_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_15_9
#define init_11_5_15_9
void FASTCALL call_11_5_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_15_10
#define init_11_5_15_10
void FASTCALL call_11_5_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_15_11
#define init_11_5_15_11
void FASTCALL call_11_5_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_15_12
#define init_11_5_15_12
void FASTCALL call_11_5_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_15_13
#define init_11_5_15_13
void FASTCALL call_11_5_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_15_14
#define init_11_5_15_14
void FASTCALL call_11_5_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_5_15_15
#define init_11_5_15_15
void FASTCALL call_11_5_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_0_0
#define init_11_6_0_0
void FASTCALL call_11_6_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_0_1
#define init_11_6_0_1
void FASTCALL call_11_6_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_0_2
#define init_11_6_0_2
void FASTCALL call_11_6_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_0_3
#define init_11_6_0_3
void FASTCALL call_11_6_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_0_4
#define init_11_6_0_4
void FASTCALL call_11_6_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_0_5
#define init_11_6_0_5
void FASTCALL call_11_6_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_0_6
#define init_11_6_0_6
void FASTCALL call_11_6_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_0_7
#define init_11_6_0_7
void FASTCALL call_11_6_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_0_8
#define init_11_6_0_8
void FASTCALL call_11_6_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_0_9
#define init_11_6_0_9
void FASTCALL call_11_6_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_0_10
#define init_11_6_0_10
void FASTCALL call_11_6_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_0_11
#define init_11_6_0_11
void FASTCALL call_11_6_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_0_12
#define init_11_6_0_12
void FASTCALL call_11_6_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_0_13
#define init_11_6_0_13
void FASTCALL call_11_6_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_0_14
#define init_11_6_0_14
void FASTCALL call_11_6_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_0_15
#define init_11_6_0_15
void FASTCALL call_11_6_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_1_0
#define init_11_6_1_0
void FASTCALL call_11_6_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_1_1
#define init_11_6_1_1
void FASTCALL call_11_6_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_1_2
#define init_11_6_1_2
void FASTCALL call_11_6_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_1_3
#define init_11_6_1_3
void FASTCALL call_11_6_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_1_4
#define init_11_6_1_4
void FASTCALL call_11_6_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_1_5
#define init_11_6_1_5
void FASTCALL call_11_6_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_1_6
#define init_11_6_1_6
void FASTCALL call_11_6_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_1_7
#define init_11_6_1_7
void FASTCALL call_11_6_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_1_8
#define init_11_6_1_8
void FASTCALL call_11_6_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_1_9
#define init_11_6_1_9
void FASTCALL call_11_6_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_1_10
#define init_11_6_1_10
void FASTCALL call_11_6_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_1_11
#define init_11_6_1_11
void FASTCALL call_11_6_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_1_12
#define init_11_6_1_12
void FASTCALL call_11_6_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_1_13
#define init_11_6_1_13
void FASTCALL call_11_6_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_1_14
#define init_11_6_1_14
void FASTCALL call_11_6_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_1_15
#define init_11_6_1_15
void FASTCALL call_11_6_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_2_0
#define init_11_6_2_0
void FASTCALL call_11_6_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_2_1
#define init_11_6_2_1
void FASTCALL call_11_6_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_2_2
#define init_11_6_2_2
void FASTCALL call_11_6_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_2_3
#define init_11_6_2_3
void FASTCALL call_11_6_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_2_4
#define init_11_6_2_4
void FASTCALL call_11_6_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_2_5
#define init_11_6_2_5
void FASTCALL call_11_6_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_2_6
#define init_11_6_2_6
void FASTCALL call_11_6_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_2_7
#define init_11_6_2_7
void FASTCALL call_11_6_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_2_8
#define init_11_6_2_8
void FASTCALL call_11_6_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_2_9
#define init_11_6_2_9
void FASTCALL call_11_6_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_2_10
#define init_11_6_2_10
void FASTCALL call_11_6_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_2_11
#define init_11_6_2_11
void FASTCALL call_11_6_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_2_12
#define init_11_6_2_12
void FASTCALL call_11_6_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_2_13
#define init_11_6_2_13
void FASTCALL call_11_6_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_2_14
#define init_11_6_2_14
void FASTCALL call_11_6_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_2_15
#define init_11_6_2_15
void FASTCALL call_11_6_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_3_0
#define init_11_6_3_0
void FASTCALL call_11_6_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_3_1
#define init_11_6_3_1
void FASTCALL call_11_6_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_3_2
#define init_11_6_3_2
void FASTCALL call_11_6_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_3_3
#define init_11_6_3_3
void FASTCALL call_11_6_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_3_4
#define init_11_6_3_4
void FASTCALL call_11_6_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_3_5
#define init_11_6_3_5
void FASTCALL call_11_6_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_3_6
#define init_11_6_3_6
void FASTCALL call_11_6_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_3_7
#define init_11_6_3_7
void FASTCALL call_11_6_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_3_8
#define init_11_6_3_8
void FASTCALL call_11_6_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_3_9
#define init_11_6_3_9
void FASTCALL call_11_6_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_3_10
#define init_11_6_3_10
void FASTCALL call_11_6_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_3_11
#define init_11_6_3_11
void FASTCALL call_11_6_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_3_12
#define init_11_6_3_12
void FASTCALL call_11_6_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_3_13
#define init_11_6_3_13
void FASTCALL call_11_6_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_3_14
#define init_11_6_3_14
void FASTCALL call_11_6_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_3_15
#define init_11_6_3_15
void FASTCALL call_11_6_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_4_0
#define init_11_6_4_0
void FASTCALL call_11_6_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_4_1
#define init_11_6_4_1
void FASTCALL call_11_6_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_4_2
#define init_11_6_4_2
void FASTCALL call_11_6_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_4_3
#define init_11_6_4_3
void FASTCALL call_11_6_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_4_4
#define init_11_6_4_4
void FASTCALL call_11_6_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_4_5
#define init_11_6_4_5
void FASTCALL call_11_6_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_4_6
#define init_11_6_4_6
void FASTCALL call_11_6_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_4_7
#define init_11_6_4_7
void FASTCALL call_11_6_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_4_8
#define init_11_6_4_8
void FASTCALL call_11_6_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_4_9
#define init_11_6_4_9
void FASTCALL call_11_6_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_4_10
#define init_11_6_4_10
void FASTCALL call_11_6_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_4_11
#define init_11_6_4_11
void FASTCALL call_11_6_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_4_12
#define init_11_6_4_12
void FASTCALL call_11_6_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_4_13
#define init_11_6_4_13
void FASTCALL call_11_6_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_4_14
#define init_11_6_4_14
void FASTCALL call_11_6_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_4_15
#define init_11_6_4_15
void FASTCALL call_11_6_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_5_0
#define init_11_6_5_0
void FASTCALL call_11_6_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_5_1
#define init_11_6_5_1
void FASTCALL call_11_6_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_5_2
#define init_11_6_5_2
void FASTCALL call_11_6_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_5_3
#define init_11_6_5_3
void FASTCALL call_11_6_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_5_4
#define init_11_6_5_4
void FASTCALL call_11_6_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_5_5
#define init_11_6_5_5
void FASTCALL call_11_6_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_5_6
#define init_11_6_5_6
void FASTCALL call_11_6_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_5_7
#define init_11_6_5_7
void FASTCALL call_11_6_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_5_8
#define init_11_6_5_8
void FASTCALL call_11_6_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_5_9
#define init_11_6_5_9
void FASTCALL call_11_6_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_5_10
#define init_11_6_5_10
void FASTCALL call_11_6_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_5_11
#define init_11_6_5_11
void FASTCALL call_11_6_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_5_12
#define init_11_6_5_12
void FASTCALL call_11_6_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_5_13
#define init_11_6_5_13
void FASTCALL call_11_6_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_5_14
#define init_11_6_5_14
void FASTCALL call_11_6_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_5_15
#define init_11_6_5_15
void FASTCALL call_11_6_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_6_0
#define init_11_6_6_0
void FASTCALL call_11_6_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_6_1
#define init_11_6_6_1
void FASTCALL call_11_6_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_6_2
#define init_11_6_6_2
void FASTCALL call_11_6_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_6_3
#define init_11_6_6_3
void FASTCALL call_11_6_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_6_4
#define init_11_6_6_4
void FASTCALL call_11_6_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_6_5
#define init_11_6_6_5
void FASTCALL call_11_6_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_6_6
#define init_11_6_6_6
void FASTCALL call_11_6_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_6_7
#define init_11_6_6_7
void FASTCALL call_11_6_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_6_8
#define init_11_6_6_8
void FASTCALL call_11_6_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_6_9
#define init_11_6_6_9
void FASTCALL call_11_6_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_6_10
#define init_11_6_6_10
void FASTCALL call_11_6_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_6_11
#define init_11_6_6_11
void FASTCALL call_11_6_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_6_12
#define init_11_6_6_12
void FASTCALL call_11_6_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_6_13
#define init_11_6_6_13
void FASTCALL call_11_6_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_6_14
#define init_11_6_6_14
void FASTCALL call_11_6_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_6_15
#define init_11_6_6_15
void FASTCALL call_11_6_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_7_0
#define init_11_6_7_0
void FASTCALL call_11_6_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_7_1
#define init_11_6_7_1
void FASTCALL call_11_6_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_7_2
#define init_11_6_7_2
void FASTCALL call_11_6_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_7_3
#define init_11_6_7_3
void FASTCALL call_11_6_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_7_4
#define init_11_6_7_4
void FASTCALL call_11_6_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_7_5
#define init_11_6_7_5
void FASTCALL call_11_6_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_7_6
#define init_11_6_7_6
void FASTCALL call_11_6_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_7_7
#define init_11_6_7_7
void FASTCALL call_11_6_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_7_8
#define init_11_6_7_8
void FASTCALL call_11_6_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_7_9
#define init_11_6_7_9
void FASTCALL call_11_6_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_7_10
#define init_11_6_7_10
void FASTCALL call_11_6_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_7_11
#define init_11_6_7_11
void FASTCALL call_11_6_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_7_12
#define init_11_6_7_12
void FASTCALL call_11_6_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_7_13
#define init_11_6_7_13
void FASTCALL call_11_6_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_7_14
#define init_11_6_7_14
void FASTCALL call_11_6_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_7_15
#define init_11_6_7_15
void FASTCALL call_11_6_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_8_0
#define init_11_6_8_0
void FASTCALL call_11_6_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_8_1
#define init_11_6_8_1
void FASTCALL call_11_6_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_8_2
#define init_11_6_8_2
void FASTCALL call_11_6_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_8_3
#define init_11_6_8_3
void FASTCALL call_11_6_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_8_4
#define init_11_6_8_4
void FASTCALL call_11_6_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_8_5
#define init_11_6_8_5
void FASTCALL call_11_6_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_8_6
#define init_11_6_8_6
void FASTCALL call_11_6_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_8_7
#define init_11_6_8_7
void FASTCALL call_11_6_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_8_8
#define init_11_6_8_8
void FASTCALL call_11_6_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_8_9
#define init_11_6_8_9
void FASTCALL call_11_6_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_8_10
#define init_11_6_8_10
void FASTCALL call_11_6_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_8_11
#define init_11_6_8_11
void FASTCALL call_11_6_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_8_12
#define init_11_6_8_12
void FASTCALL call_11_6_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_8_13
#define init_11_6_8_13
void FASTCALL call_11_6_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_8_14
#define init_11_6_8_14
void FASTCALL call_11_6_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_8_15
#define init_11_6_8_15
void FASTCALL call_11_6_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_9_0
#define init_11_6_9_0
void FASTCALL call_11_6_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_9_1
#define init_11_6_9_1
void FASTCALL call_11_6_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_9_2
#define init_11_6_9_2
void FASTCALL call_11_6_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_9_3
#define init_11_6_9_3
void FASTCALL call_11_6_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_9_4
#define init_11_6_9_4
void FASTCALL call_11_6_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_9_5
#define init_11_6_9_5
void FASTCALL call_11_6_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_9_6
#define init_11_6_9_6
void FASTCALL call_11_6_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_9_7
#define init_11_6_9_7
void FASTCALL call_11_6_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_9_8
#define init_11_6_9_8
void FASTCALL call_11_6_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_9_9
#define init_11_6_9_9
void FASTCALL call_11_6_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_9_10
#define init_11_6_9_10
void FASTCALL call_11_6_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_9_11
#define init_11_6_9_11
void FASTCALL call_11_6_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_9_12
#define init_11_6_9_12
void FASTCALL call_11_6_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_9_13
#define init_11_6_9_13
void FASTCALL call_11_6_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_9_14
#define init_11_6_9_14
void FASTCALL call_11_6_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_9_15
#define init_11_6_9_15
void FASTCALL call_11_6_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_10_0
#define init_11_6_10_0
void FASTCALL call_11_6_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_10_1
#define init_11_6_10_1
void FASTCALL call_11_6_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_10_2
#define init_11_6_10_2
void FASTCALL call_11_6_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_10_3
#define init_11_6_10_3
void FASTCALL call_11_6_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_10_4
#define init_11_6_10_4
void FASTCALL call_11_6_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_10_5
#define init_11_6_10_5
void FASTCALL call_11_6_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_10_6
#define init_11_6_10_6
void FASTCALL call_11_6_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_10_7
#define init_11_6_10_7
void FASTCALL call_11_6_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_10_8
#define init_11_6_10_8
void FASTCALL call_11_6_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_10_9
#define init_11_6_10_9
void FASTCALL call_11_6_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_10_10
#define init_11_6_10_10
void FASTCALL call_11_6_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_10_11
#define init_11_6_10_11
void FASTCALL call_11_6_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_10_12
#define init_11_6_10_12
void FASTCALL call_11_6_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_10_13
#define init_11_6_10_13
void FASTCALL call_11_6_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_10_14
#define init_11_6_10_14
void FASTCALL call_11_6_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_10_15
#define init_11_6_10_15
void FASTCALL call_11_6_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_11_0
#define init_11_6_11_0
void FASTCALL call_11_6_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_11_1
#define init_11_6_11_1
void FASTCALL call_11_6_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_11_2
#define init_11_6_11_2
void FASTCALL call_11_6_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_11_3
#define init_11_6_11_3
void FASTCALL call_11_6_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_11_4
#define init_11_6_11_4
void FASTCALL call_11_6_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_11_5
#define init_11_6_11_5
void FASTCALL call_11_6_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_11_6
#define init_11_6_11_6
void FASTCALL call_11_6_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_11_7
#define init_11_6_11_7
void FASTCALL call_11_6_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_11_8
#define init_11_6_11_8
void FASTCALL call_11_6_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_11_9
#define init_11_6_11_9
void FASTCALL call_11_6_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_11_10
#define init_11_6_11_10
void FASTCALL call_11_6_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_11_11
#define init_11_6_11_11
void FASTCALL call_11_6_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_11_12
#define init_11_6_11_12
void FASTCALL call_11_6_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_11_13
#define init_11_6_11_13
void FASTCALL call_11_6_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_11_14
#define init_11_6_11_14
void FASTCALL call_11_6_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_11_15
#define init_11_6_11_15
void FASTCALL call_11_6_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_12_0
#define init_11_6_12_0
void FASTCALL call_11_6_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_12_1
#define init_11_6_12_1
void FASTCALL call_11_6_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_12_2
#define init_11_6_12_2
void FASTCALL call_11_6_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_12_3
#define init_11_6_12_3
void FASTCALL call_11_6_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_12_4
#define init_11_6_12_4
void FASTCALL call_11_6_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_12_5
#define init_11_6_12_5
void FASTCALL call_11_6_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_12_6
#define init_11_6_12_6
void FASTCALL call_11_6_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_12_7
#define init_11_6_12_7
void FASTCALL call_11_6_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_12_8
#define init_11_6_12_8
void FASTCALL call_11_6_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_12_9
#define init_11_6_12_9
void FASTCALL call_11_6_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_12_10
#define init_11_6_12_10
void FASTCALL call_11_6_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_12_11
#define init_11_6_12_11
void FASTCALL call_11_6_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_12_12
#define init_11_6_12_12
void FASTCALL call_11_6_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_12_13
#define init_11_6_12_13
void FASTCALL call_11_6_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_12_14
#define init_11_6_12_14
void FASTCALL call_11_6_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_12_15
#define init_11_6_12_15
void FASTCALL call_11_6_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_13_0
#define init_11_6_13_0
void FASTCALL call_11_6_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_13_1
#define init_11_6_13_1
void FASTCALL call_11_6_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_13_2
#define init_11_6_13_2
void FASTCALL call_11_6_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_13_3
#define init_11_6_13_3
void FASTCALL call_11_6_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_13_4
#define init_11_6_13_4
void FASTCALL call_11_6_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_13_5
#define init_11_6_13_5
void FASTCALL call_11_6_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_13_6
#define init_11_6_13_6
void FASTCALL call_11_6_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_13_7
#define init_11_6_13_7
void FASTCALL call_11_6_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_13_8
#define init_11_6_13_8
void FASTCALL call_11_6_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_13_9
#define init_11_6_13_9
void FASTCALL call_11_6_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_13_10
#define init_11_6_13_10
void FASTCALL call_11_6_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_13_11
#define init_11_6_13_11
void FASTCALL call_11_6_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_13_12
#define init_11_6_13_12
void FASTCALL call_11_6_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_13_13
#define init_11_6_13_13
void FASTCALL call_11_6_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_13_14
#define init_11_6_13_14
void FASTCALL call_11_6_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_13_15
#define init_11_6_13_15
void FASTCALL call_11_6_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_14_0
#define init_11_6_14_0
void FASTCALL call_11_6_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_14_1
#define init_11_6_14_1
void FASTCALL call_11_6_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_14_2
#define init_11_6_14_2
void FASTCALL call_11_6_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_14_3
#define init_11_6_14_3
void FASTCALL call_11_6_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_14_4
#define init_11_6_14_4
void FASTCALL call_11_6_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_14_5
#define init_11_6_14_5
void FASTCALL call_11_6_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_14_6
#define init_11_6_14_6
void FASTCALL call_11_6_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_14_7
#define init_11_6_14_7
void FASTCALL call_11_6_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_14_8
#define init_11_6_14_8
void FASTCALL call_11_6_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_14_9
#define init_11_6_14_9
void FASTCALL call_11_6_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_14_10
#define init_11_6_14_10
void FASTCALL call_11_6_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_14_11
#define init_11_6_14_11
void FASTCALL call_11_6_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_14_12
#define init_11_6_14_12
void FASTCALL call_11_6_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_14_13
#define init_11_6_14_13
void FASTCALL call_11_6_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_14_14
#define init_11_6_14_14
void FASTCALL call_11_6_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_14_15
#define init_11_6_14_15
void FASTCALL call_11_6_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_15_0
#define init_11_6_15_0
void FASTCALL call_11_6_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_15_1
#define init_11_6_15_1
void FASTCALL call_11_6_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_15_2
#define init_11_6_15_2
void FASTCALL call_11_6_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_15_3
#define init_11_6_15_3
void FASTCALL call_11_6_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_15_4
#define init_11_6_15_4
void FASTCALL call_11_6_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_15_5
#define init_11_6_15_5
void FASTCALL call_11_6_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_15_6
#define init_11_6_15_6
void FASTCALL call_11_6_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_15_7
#define init_11_6_15_7
void FASTCALL call_11_6_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_15_8
#define init_11_6_15_8
void FASTCALL call_11_6_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_15_9
#define init_11_6_15_9
void FASTCALL call_11_6_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_15_10
#define init_11_6_15_10
void FASTCALL call_11_6_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_15_11
#define init_11_6_15_11
void FASTCALL call_11_6_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_15_12
#define init_11_6_15_12
void FASTCALL call_11_6_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_15_13
#define init_11_6_15_13
void FASTCALL call_11_6_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_15_14
#define init_11_6_15_14
void FASTCALL call_11_6_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_6_15_15
#define init_11_6_15_15
void FASTCALL call_11_6_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_0_0
#define init_11_7_0_0
void FASTCALL call_11_7_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_0_1
#define init_11_7_0_1
void FASTCALL call_11_7_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_0_2
#define init_11_7_0_2
void FASTCALL call_11_7_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_0_3
#define init_11_7_0_3
void FASTCALL call_11_7_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_0_4
#define init_11_7_0_4
void FASTCALL call_11_7_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_0_5
#define init_11_7_0_5
void FASTCALL call_11_7_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_0_6
#define init_11_7_0_6
void FASTCALL call_11_7_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_0_7
#define init_11_7_0_7
void FASTCALL call_11_7_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_0_8
#define init_11_7_0_8
void FASTCALL call_11_7_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_0_9
#define init_11_7_0_9
void FASTCALL call_11_7_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_0_10
#define init_11_7_0_10
void FASTCALL call_11_7_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_0_11
#define init_11_7_0_11
void FASTCALL call_11_7_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_0_12
#define init_11_7_0_12
void FASTCALL call_11_7_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_0_13
#define init_11_7_0_13
void FASTCALL call_11_7_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_0_14
#define init_11_7_0_14
void FASTCALL call_11_7_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_0_15
#define init_11_7_0_15
void FASTCALL call_11_7_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_1_0
#define init_11_7_1_0
void FASTCALL call_11_7_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_1_1
#define init_11_7_1_1
void FASTCALL call_11_7_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_1_2
#define init_11_7_1_2
void FASTCALL call_11_7_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_1_3
#define init_11_7_1_3
void FASTCALL call_11_7_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_1_4
#define init_11_7_1_4
void FASTCALL call_11_7_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_1_5
#define init_11_7_1_5
void FASTCALL call_11_7_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_1_6
#define init_11_7_1_6
void FASTCALL call_11_7_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_1_7
#define init_11_7_1_7
void FASTCALL call_11_7_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_1_8
#define init_11_7_1_8
void FASTCALL call_11_7_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_1_9
#define init_11_7_1_9
void FASTCALL call_11_7_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_1_10
#define init_11_7_1_10
void FASTCALL call_11_7_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_1_11
#define init_11_7_1_11
void FASTCALL call_11_7_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_1_12
#define init_11_7_1_12
void FASTCALL call_11_7_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_1_13
#define init_11_7_1_13
void FASTCALL call_11_7_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_1_14
#define init_11_7_1_14
void FASTCALL call_11_7_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_1_15
#define init_11_7_1_15
void FASTCALL call_11_7_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_2_0
#define init_11_7_2_0
void FASTCALL call_11_7_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_2_1
#define init_11_7_2_1
void FASTCALL call_11_7_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_2_2
#define init_11_7_2_2
void FASTCALL call_11_7_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_2_3
#define init_11_7_2_3
void FASTCALL call_11_7_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_2_4
#define init_11_7_2_4
void FASTCALL call_11_7_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_2_5
#define init_11_7_2_5
void FASTCALL call_11_7_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_2_6
#define init_11_7_2_6
void FASTCALL call_11_7_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_2_7
#define init_11_7_2_7
void FASTCALL call_11_7_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_2_8
#define init_11_7_2_8
void FASTCALL call_11_7_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_2_9
#define init_11_7_2_9
void FASTCALL call_11_7_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_2_10
#define init_11_7_2_10
void FASTCALL call_11_7_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_2_11
#define init_11_7_2_11
void FASTCALL call_11_7_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_2_12
#define init_11_7_2_12
void FASTCALL call_11_7_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_2_13
#define init_11_7_2_13
void FASTCALL call_11_7_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_2_14
#define init_11_7_2_14
void FASTCALL call_11_7_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_2_15
#define init_11_7_2_15
void FASTCALL call_11_7_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_3_0
#define init_11_7_3_0
void FASTCALL call_11_7_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_3_1
#define init_11_7_3_1
void FASTCALL call_11_7_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_3_2
#define init_11_7_3_2
void FASTCALL call_11_7_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_3_3
#define init_11_7_3_3
void FASTCALL call_11_7_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_3_4
#define init_11_7_3_4
void FASTCALL call_11_7_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_3_5
#define init_11_7_3_5
void FASTCALL call_11_7_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_3_6
#define init_11_7_3_6
void FASTCALL call_11_7_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_3_7
#define init_11_7_3_7
void FASTCALL call_11_7_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_3_8
#define init_11_7_3_8
void FASTCALL call_11_7_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_3_9
#define init_11_7_3_9
void FASTCALL call_11_7_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_3_10
#define init_11_7_3_10
void FASTCALL call_11_7_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_3_11
#define init_11_7_3_11
void FASTCALL call_11_7_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_3_12
#define init_11_7_3_12
void FASTCALL call_11_7_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_3_13
#define init_11_7_3_13
void FASTCALL call_11_7_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_3_14
#define init_11_7_3_14
void FASTCALL call_11_7_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_3_15
#define init_11_7_3_15
void FASTCALL call_11_7_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_4_0
#define init_11_7_4_0
void FASTCALL call_11_7_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_4_1
#define init_11_7_4_1
void FASTCALL call_11_7_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_4_2
#define init_11_7_4_2
void FASTCALL call_11_7_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_4_3
#define init_11_7_4_3
void FASTCALL call_11_7_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_4_4
#define init_11_7_4_4
void FASTCALL call_11_7_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_4_5
#define init_11_7_4_5
void FASTCALL call_11_7_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_4_6
#define init_11_7_4_6
void FASTCALL call_11_7_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_4_7
#define init_11_7_4_7
void FASTCALL call_11_7_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_4_8
#define init_11_7_4_8
void FASTCALL call_11_7_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_4_9
#define init_11_7_4_9
void FASTCALL call_11_7_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_4_10
#define init_11_7_4_10
void FASTCALL call_11_7_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_4_11
#define init_11_7_4_11
void FASTCALL call_11_7_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_4_12
#define init_11_7_4_12
void FASTCALL call_11_7_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_4_13
#define init_11_7_4_13
void FASTCALL call_11_7_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_4_14
#define init_11_7_4_14
void FASTCALL call_11_7_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_4_15
#define init_11_7_4_15
void FASTCALL call_11_7_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_5_0
#define init_11_7_5_0
void FASTCALL call_11_7_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_5_1
#define init_11_7_5_1
void FASTCALL call_11_7_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_5_2
#define init_11_7_5_2
void FASTCALL call_11_7_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_5_3
#define init_11_7_5_3
void FASTCALL call_11_7_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_5_4
#define init_11_7_5_4
void FASTCALL call_11_7_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_5_5
#define init_11_7_5_5
void FASTCALL call_11_7_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_5_6
#define init_11_7_5_6
void FASTCALL call_11_7_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_5_7
#define init_11_7_5_7
void FASTCALL call_11_7_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_5_8
#define init_11_7_5_8
void FASTCALL call_11_7_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_5_9
#define init_11_7_5_9
void FASTCALL call_11_7_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_5_10
#define init_11_7_5_10
void FASTCALL call_11_7_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_5_11
#define init_11_7_5_11
void FASTCALL call_11_7_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_5_12
#define init_11_7_5_12
void FASTCALL call_11_7_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_5_13
#define init_11_7_5_13
void FASTCALL call_11_7_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_5_14
#define init_11_7_5_14
void FASTCALL call_11_7_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_5_15
#define init_11_7_5_15
void FASTCALL call_11_7_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_6_0
#define init_11_7_6_0
void FASTCALL call_11_7_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_6_1
#define init_11_7_6_1
void FASTCALL call_11_7_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_6_2
#define init_11_7_6_2
void FASTCALL call_11_7_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_6_3
#define init_11_7_6_3
void FASTCALL call_11_7_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_6_4
#define init_11_7_6_4
void FASTCALL call_11_7_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_6_5
#define init_11_7_6_5
void FASTCALL call_11_7_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_6_6
#define init_11_7_6_6
void FASTCALL call_11_7_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_6_7
#define init_11_7_6_7
void FASTCALL call_11_7_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_6_8
#define init_11_7_6_8
void FASTCALL call_11_7_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_6_9
#define init_11_7_6_9
void FASTCALL call_11_7_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_6_10
#define init_11_7_6_10
void FASTCALL call_11_7_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_6_11
#define init_11_7_6_11
void FASTCALL call_11_7_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_6_12
#define init_11_7_6_12
void FASTCALL call_11_7_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_6_13
#define init_11_7_6_13
void FASTCALL call_11_7_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_6_14
#define init_11_7_6_14
void FASTCALL call_11_7_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_6_15
#define init_11_7_6_15
void FASTCALL call_11_7_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_7_0
#define init_11_7_7_0
void FASTCALL call_11_7_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_7_1
#define init_11_7_7_1
void FASTCALL call_11_7_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_7_2
#define init_11_7_7_2
void FASTCALL call_11_7_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_7_3
#define init_11_7_7_3
void FASTCALL call_11_7_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_7_4
#define init_11_7_7_4
void FASTCALL call_11_7_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_7_5
#define init_11_7_7_5
void FASTCALL call_11_7_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_7_6
#define init_11_7_7_6
void FASTCALL call_11_7_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_7_7
#define init_11_7_7_7
void FASTCALL call_11_7_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_7_8
#define init_11_7_7_8
void FASTCALL call_11_7_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_7_9
#define init_11_7_7_9
void FASTCALL call_11_7_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_7_10
#define init_11_7_7_10
void FASTCALL call_11_7_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_7_11
#define init_11_7_7_11
void FASTCALL call_11_7_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_7_12
#define init_11_7_7_12
void FASTCALL call_11_7_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_7_13
#define init_11_7_7_13
void FASTCALL call_11_7_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_7_14
#define init_11_7_7_14
void FASTCALL call_11_7_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_7_15
#define init_11_7_7_15
void FASTCALL call_11_7_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_8_0
#define init_11_7_8_0
void FASTCALL call_11_7_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_8_1
#define init_11_7_8_1
void FASTCALL call_11_7_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_8_2
#define init_11_7_8_2
void FASTCALL call_11_7_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_8_3
#define init_11_7_8_3
void FASTCALL call_11_7_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_8_4
#define init_11_7_8_4
void FASTCALL call_11_7_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_8_5
#define init_11_7_8_5
void FASTCALL call_11_7_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_8_6
#define init_11_7_8_6
void FASTCALL call_11_7_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_8_7
#define init_11_7_8_7
void FASTCALL call_11_7_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_8_8
#define init_11_7_8_8
void FASTCALL call_11_7_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_8_9
#define init_11_7_8_9
void FASTCALL call_11_7_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_8_10
#define init_11_7_8_10
void FASTCALL call_11_7_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_8_11
#define init_11_7_8_11
void FASTCALL call_11_7_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_8_12
#define init_11_7_8_12
void FASTCALL call_11_7_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_8_13
#define init_11_7_8_13
void FASTCALL call_11_7_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_8_14
#define init_11_7_8_14
void FASTCALL call_11_7_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_8_15
#define init_11_7_8_15
void FASTCALL call_11_7_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_9_0
#define init_11_7_9_0
void FASTCALL call_11_7_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_9_1
#define init_11_7_9_1
void FASTCALL call_11_7_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_9_2
#define init_11_7_9_2
void FASTCALL call_11_7_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_9_3
#define init_11_7_9_3
void FASTCALL call_11_7_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_9_4
#define init_11_7_9_4
void FASTCALL call_11_7_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_9_5
#define init_11_7_9_5
void FASTCALL call_11_7_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_9_6
#define init_11_7_9_6
void FASTCALL call_11_7_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_9_7
#define init_11_7_9_7
void FASTCALL call_11_7_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_9_8
#define init_11_7_9_8
void FASTCALL call_11_7_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_9_9
#define init_11_7_9_9
void FASTCALL call_11_7_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_9_10
#define init_11_7_9_10
void FASTCALL call_11_7_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_9_11
#define init_11_7_9_11
void FASTCALL call_11_7_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_9_12
#define init_11_7_9_12
void FASTCALL call_11_7_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_9_13
#define init_11_7_9_13
void FASTCALL call_11_7_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_9_14
#define init_11_7_9_14
void FASTCALL call_11_7_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_9_15
#define init_11_7_9_15
void FASTCALL call_11_7_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_10_0
#define init_11_7_10_0
void FASTCALL call_11_7_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_10_1
#define init_11_7_10_1
void FASTCALL call_11_7_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_10_2
#define init_11_7_10_2
void FASTCALL call_11_7_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_10_3
#define init_11_7_10_3
void FASTCALL call_11_7_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_10_4
#define init_11_7_10_4
void FASTCALL call_11_7_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_10_5
#define init_11_7_10_5
void FASTCALL call_11_7_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_10_6
#define init_11_7_10_6
void FASTCALL call_11_7_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_10_7
#define init_11_7_10_7
void FASTCALL call_11_7_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_10_8
#define init_11_7_10_8
void FASTCALL call_11_7_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_10_9
#define init_11_7_10_9
void FASTCALL call_11_7_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_10_10
#define init_11_7_10_10
void FASTCALL call_11_7_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_10_11
#define init_11_7_10_11
void FASTCALL call_11_7_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_10_12
#define init_11_7_10_12
void FASTCALL call_11_7_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_10_13
#define init_11_7_10_13
void FASTCALL call_11_7_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_10_14
#define init_11_7_10_14
void FASTCALL call_11_7_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_10_15
#define init_11_7_10_15
void FASTCALL call_11_7_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_11_0
#define init_11_7_11_0
void FASTCALL call_11_7_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_11_1
#define init_11_7_11_1
void FASTCALL call_11_7_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_11_2
#define init_11_7_11_2
void FASTCALL call_11_7_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_11_3
#define init_11_7_11_3
void FASTCALL call_11_7_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_11_4
#define init_11_7_11_4
void FASTCALL call_11_7_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_11_5
#define init_11_7_11_5
void FASTCALL call_11_7_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_11_6
#define init_11_7_11_6
void FASTCALL call_11_7_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_11_7
#define init_11_7_11_7
void FASTCALL call_11_7_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_11_8
#define init_11_7_11_8
void FASTCALL call_11_7_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_11_9
#define init_11_7_11_9
void FASTCALL call_11_7_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_11_10
#define init_11_7_11_10
void FASTCALL call_11_7_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_11_11
#define init_11_7_11_11
void FASTCALL call_11_7_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_11_12
#define init_11_7_11_12
void FASTCALL call_11_7_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_11_13
#define init_11_7_11_13
void FASTCALL call_11_7_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_11_14
#define init_11_7_11_14
void FASTCALL call_11_7_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_11_15
#define init_11_7_11_15
void FASTCALL call_11_7_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_12_0
#define init_11_7_12_0
void FASTCALL call_11_7_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_12_1
#define init_11_7_12_1
void FASTCALL call_11_7_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_12_2
#define init_11_7_12_2
void FASTCALL call_11_7_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_12_3
#define init_11_7_12_3
void FASTCALL call_11_7_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_12_4
#define init_11_7_12_4
void FASTCALL call_11_7_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_12_5
#define init_11_7_12_5
void FASTCALL call_11_7_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_12_6
#define init_11_7_12_6
void FASTCALL call_11_7_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_12_7
#define init_11_7_12_7
void FASTCALL call_11_7_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_12_8
#define init_11_7_12_8
void FASTCALL call_11_7_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_12_9
#define init_11_7_12_9
void FASTCALL call_11_7_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_12_10
#define init_11_7_12_10
void FASTCALL call_11_7_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_12_11
#define init_11_7_12_11
void FASTCALL call_11_7_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_12_12
#define init_11_7_12_12
void FASTCALL call_11_7_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_12_13
#define init_11_7_12_13
void FASTCALL call_11_7_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_12_14
#define init_11_7_12_14
void FASTCALL call_11_7_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_12_15
#define init_11_7_12_15
void FASTCALL call_11_7_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_13_0
#define init_11_7_13_0
void FASTCALL call_11_7_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_13_1
#define init_11_7_13_1
void FASTCALL call_11_7_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_13_2
#define init_11_7_13_2
void FASTCALL call_11_7_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_13_3
#define init_11_7_13_3
void FASTCALL call_11_7_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_13_4
#define init_11_7_13_4
void FASTCALL call_11_7_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_13_5
#define init_11_7_13_5
void FASTCALL call_11_7_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_13_6
#define init_11_7_13_6
void FASTCALL call_11_7_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_13_7
#define init_11_7_13_7
void FASTCALL call_11_7_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_13_8
#define init_11_7_13_8
void FASTCALL call_11_7_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_13_9
#define init_11_7_13_9
void FASTCALL call_11_7_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_13_10
#define init_11_7_13_10
void FASTCALL call_11_7_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_13_11
#define init_11_7_13_11
void FASTCALL call_11_7_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_13_12
#define init_11_7_13_12
void FASTCALL call_11_7_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_13_13
#define init_11_7_13_13
void FASTCALL call_11_7_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_13_14
#define init_11_7_13_14
void FASTCALL call_11_7_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_13_15
#define init_11_7_13_15
void FASTCALL call_11_7_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_14_0
#define init_11_7_14_0
void FASTCALL call_11_7_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_14_1
#define init_11_7_14_1
void FASTCALL call_11_7_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_14_2
#define init_11_7_14_2
void FASTCALL call_11_7_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_14_3
#define init_11_7_14_3
void FASTCALL call_11_7_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_14_4
#define init_11_7_14_4
void FASTCALL call_11_7_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_14_5
#define init_11_7_14_5
void FASTCALL call_11_7_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_14_6
#define init_11_7_14_6
void FASTCALL call_11_7_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_14_7
#define init_11_7_14_7
void FASTCALL call_11_7_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_14_8
#define init_11_7_14_8
void FASTCALL call_11_7_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_14_9
#define init_11_7_14_9
void FASTCALL call_11_7_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_14_10
#define init_11_7_14_10
void FASTCALL call_11_7_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_14_11
#define init_11_7_14_11
void FASTCALL call_11_7_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_14_12
#define init_11_7_14_12
void FASTCALL call_11_7_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_14_13
#define init_11_7_14_13
void FASTCALL call_11_7_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_14_14
#define init_11_7_14_14
void FASTCALL call_11_7_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_14_15
#define init_11_7_14_15
void FASTCALL call_11_7_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_15_0
#define init_11_7_15_0
void FASTCALL call_11_7_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_15_1
#define init_11_7_15_1
void FASTCALL call_11_7_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_15_2
#define init_11_7_15_2
void FASTCALL call_11_7_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_15_3
#define init_11_7_15_3
void FASTCALL call_11_7_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_15_4
#define init_11_7_15_4
void FASTCALL call_11_7_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_15_5
#define init_11_7_15_5
void FASTCALL call_11_7_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_15_6
#define init_11_7_15_6
void FASTCALL call_11_7_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_15_7
#define init_11_7_15_7
void FASTCALL call_11_7_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_15_8
#define init_11_7_15_8
void FASTCALL call_11_7_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_15_9
#define init_11_7_15_9
void FASTCALL call_11_7_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_15_10
#define init_11_7_15_10
void FASTCALL call_11_7_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_15_11
#define init_11_7_15_11
void FASTCALL call_11_7_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_15_12
#define init_11_7_15_12
void FASTCALL call_11_7_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_15_13
#define init_11_7_15_13
void FASTCALL call_11_7_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_15_14
#define init_11_7_15_14
void FASTCALL call_11_7_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_7_15_15
#define init_11_7_15_15
void FASTCALL call_11_7_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_0_0
#define init_11_8_0_0
void FASTCALL call_11_8_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_0_1
#define init_11_8_0_1
void FASTCALL call_11_8_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_0_2
#define init_11_8_0_2
void FASTCALL call_11_8_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_0_3
#define init_11_8_0_3
void FASTCALL call_11_8_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_0_4
#define init_11_8_0_4
void FASTCALL call_11_8_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_0_5
#define init_11_8_0_5
void FASTCALL call_11_8_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_0_6
#define init_11_8_0_6
void FASTCALL call_11_8_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_0_7
#define init_11_8_0_7
void FASTCALL call_11_8_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_0_8
#define init_11_8_0_8
void FASTCALL call_11_8_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_0_9
#define init_11_8_0_9
void FASTCALL call_11_8_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_0_10
#define init_11_8_0_10
void FASTCALL call_11_8_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_0_11
#define init_11_8_0_11
void FASTCALL call_11_8_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_0_12
#define init_11_8_0_12
void FASTCALL call_11_8_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_0_13
#define init_11_8_0_13
void FASTCALL call_11_8_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_0_14
#define init_11_8_0_14
void FASTCALL call_11_8_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_0_15
#define init_11_8_0_15
void FASTCALL call_11_8_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_1_0
#define init_11_8_1_0
void FASTCALL call_11_8_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_1_1
#define init_11_8_1_1
void FASTCALL call_11_8_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_1_2
#define init_11_8_1_2
void FASTCALL call_11_8_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_1_3
#define init_11_8_1_3
void FASTCALL call_11_8_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_1_4
#define init_11_8_1_4
void FASTCALL call_11_8_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_1_5
#define init_11_8_1_5
void FASTCALL call_11_8_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_1_6
#define init_11_8_1_6
void FASTCALL call_11_8_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_1_7
#define init_11_8_1_7
void FASTCALL call_11_8_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_1_8
#define init_11_8_1_8
void FASTCALL call_11_8_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_1_9
#define init_11_8_1_9
void FASTCALL call_11_8_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_1_10
#define init_11_8_1_10
void FASTCALL call_11_8_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_1_11
#define init_11_8_1_11
void FASTCALL call_11_8_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_1_12
#define init_11_8_1_12
void FASTCALL call_11_8_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_1_13
#define init_11_8_1_13
void FASTCALL call_11_8_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_1_14
#define init_11_8_1_14
void FASTCALL call_11_8_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_1_15
#define init_11_8_1_15
void FASTCALL call_11_8_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_2_0
#define init_11_8_2_0
void FASTCALL call_11_8_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_2_1
#define init_11_8_2_1
void FASTCALL call_11_8_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_2_2
#define init_11_8_2_2
void FASTCALL call_11_8_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_2_3
#define init_11_8_2_3
void FASTCALL call_11_8_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_2_4
#define init_11_8_2_4
void FASTCALL call_11_8_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_2_5
#define init_11_8_2_5
void FASTCALL call_11_8_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_2_6
#define init_11_8_2_6
void FASTCALL call_11_8_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_2_7
#define init_11_8_2_7
void FASTCALL call_11_8_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_2_8
#define init_11_8_2_8
void FASTCALL call_11_8_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_2_9
#define init_11_8_2_9
void FASTCALL call_11_8_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_2_10
#define init_11_8_2_10
void FASTCALL call_11_8_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_2_11
#define init_11_8_2_11
void FASTCALL call_11_8_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_2_12
#define init_11_8_2_12
void FASTCALL call_11_8_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_2_13
#define init_11_8_2_13
void FASTCALL call_11_8_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_2_14
#define init_11_8_2_14
void FASTCALL call_11_8_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_2_15
#define init_11_8_2_15
void FASTCALL call_11_8_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_3_0
#define init_11_8_3_0
void FASTCALL call_11_8_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_3_1
#define init_11_8_3_1
void FASTCALL call_11_8_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_3_2
#define init_11_8_3_2
void FASTCALL call_11_8_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_3_3
#define init_11_8_3_3
void FASTCALL call_11_8_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_3_4
#define init_11_8_3_4
void FASTCALL call_11_8_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_3_5
#define init_11_8_3_5
void FASTCALL call_11_8_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_3_6
#define init_11_8_3_6
void FASTCALL call_11_8_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_3_7
#define init_11_8_3_7
void FASTCALL call_11_8_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_3_8
#define init_11_8_3_8
void FASTCALL call_11_8_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_3_9
#define init_11_8_3_9
void FASTCALL call_11_8_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_3_10
#define init_11_8_3_10
void FASTCALL call_11_8_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_3_11
#define init_11_8_3_11
void FASTCALL call_11_8_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_3_12
#define init_11_8_3_12
void FASTCALL call_11_8_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_3_13
#define init_11_8_3_13
void FASTCALL call_11_8_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_3_14
#define init_11_8_3_14
void FASTCALL call_11_8_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_3_15
#define init_11_8_3_15
void FASTCALL call_11_8_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_4_0
#define init_11_8_4_0
void FASTCALL call_11_8_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_4_1
#define init_11_8_4_1
void FASTCALL call_11_8_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_4_2
#define init_11_8_4_2
void FASTCALL call_11_8_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_4_3
#define init_11_8_4_3
void FASTCALL call_11_8_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_4_4
#define init_11_8_4_4
void FASTCALL call_11_8_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_4_5
#define init_11_8_4_5
void FASTCALL call_11_8_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_4_6
#define init_11_8_4_6
void FASTCALL call_11_8_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_4_7
#define init_11_8_4_7
void FASTCALL call_11_8_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_4_8
#define init_11_8_4_8
void FASTCALL call_11_8_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_4_9
#define init_11_8_4_9
void FASTCALL call_11_8_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_4_10
#define init_11_8_4_10
void FASTCALL call_11_8_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_4_11
#define init_11_8_4_11
void FASTCALL call_11_8_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_4_12
#define init_11_8_4_12
void FASTCALL call_11_8_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_4_13
#define init_11_8_4_13
void FASTCALL call_11_8_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_4_14
#define init_11_8_4_14
void FASTCALL call_11_8_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_4_15
#define init_11_8_4_15
void FASTCALL call_11_8_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_5_0
#define init_11_8_5_0
void FASTCALL call_11_8_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_5_1
#define init_11_8_5_1
void FASTCALL call_11_8_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_5_2
#define init_11_8_5_2
void FASTCALL call_11_8_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_5_3
#define init_11_8_5_3
void FASTCALL call_11_8_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_5_4
#define init_11_8_5_4
void FASTCALL call_11_8_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_5_5
#define init_11_8_5_5
void FASTCALL call_11_8_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_5_6
#define init_11_8_5_6
void FASTCALL call_11_8_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_5_7
#define init_11_8_5_7
void FASTCALL call_11_8_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_5_8
#define init_11_8_5_8
void FASTCALL call_11_8_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_5_9
#define init_11_8_5_9
void FASTCALL call_11_8_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_5_10
#define init_11_8_5_10
void FASTCALL call_11_8_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_5_11
#define init_11_8_5_11
void FASTCALL call_11_8_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_5_12
#define init_11_8_5_12
void FASTCALL call_11_8_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_5_13
#define init_11_8_5_13
void FASTCALL call_11_8_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_5_14
#define init_11_8_5_14
void FASTCALL call_11_8_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_5_15
#define init_11_8_5_15
void FASTCALL call_11_8_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_6_0
#define init_11_8_6_0
void FASTCALL call_11_8_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_6_1
#define init_11_8_6_1
void FASTCALL call_11_8_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_6_2
#define init_11_8_6_2
void FASTCALL call_11_8_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_6_3
#define init_11_8_6_3
void FASTCALL call_11_8_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_6_4
#define init_11_8_6_4
void FASTCALL call_11_8_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_6_5
#define init_11_8_6_5
void FASTCALL call_11_8_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_6_6
#define init_11_8_6_6
void FASTCALL call_11_8_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_6_7
#define init_11_8_6_7
void FASTCALL call_11_8_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_6_8
#define init_11_8_6_8
void FASTCALL call_11_8_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_6_9
#define init_11_8_6_9
void FASTCALL call_11_8_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_6_10
#define init_11_8_6_10
void FASTCALL call_11_8_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_6_11
#define init_11_8_6_11
void FASTCALL call_11_8_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_6_12
#define init_11_8_6_12
void FASTCALL call_11_8_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_6_13
#define init_11_8_6_13
void FASTCALL call_11_8_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_6_14
#define init_11_8_6_14
void FASTCALL call_11_8_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_6_15
#define init_11_8_6_15
void FASTCALL call_11_8_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_7_0
#define init_11_8_7_0
void FASTCALL call_11_8_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_7_1
#define init_11_8_7_1
void FASTCALL call_11_8_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_7_2
#define init_11_8_7_2
void FASTCALL call_11_8_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_7_3
#define init_11_8_7_3
void FASTCALL call_11_8_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_7_4
#define init_11_8_7_4
void FASTCALL call_11_8_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_7_5
#define init_11_8_7_5
void FASTCALL call_11_8_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_7_6
#define init_11_8_7_6
void FASTCALL call_11_8_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_7_7
#define init_11_8_7_7
void FASTCALL call_11_8_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_7_8
#define init_11_8_7_8
void FASTCALL call_11_8_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_7_9
#define init_11_8_7_9
void FASTCALL call_11_8_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_7_10
#define init_11_8_7_10
void FASTCALL call_11_8_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_7_11
#define init_11_8_7_11
void FASTCALL call_11_8_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_7_12
#define init_11_8_7_12
void FASTCALL call_11_8_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_7_13
#define init_11_8_7_13
void FASTCALL call_11_8_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_7_14
#define init_11_8_7_14
void FASTCALL call_11_8_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_7_15
#define init_11_8_7_15
void FASTCALL call_11_8_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_8_0
#define init_11_8_8_0
void FASTCALL call_11_8_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_8_1
#define init_11_8_8_1
void FASTCALL call_11_8_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_8_2
#define init_11_8_8_2
void FASTCALL call_11_8_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_8_3
#define init_11_8_8_3
void FASTCALL call_11_8_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_8_4
#define init_11_8_8_4
void FASTCALL call_11_8_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_8_5
#define init_11_8_8_5
void FASTCALL call_11_8_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_8_6
#define init_11_8_8_6
void FASTCALL call_11_8_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_8_7
#define init_11_8_8_7
void FASTCALL call_11_8_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_8_8
#define init_11_8_8_8
void FASTCALL call_11_8_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_8_9
#define init_11_8_8_9
void FASTCALL call_11_8_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_8_10
#define init_11_8_8_10
void FASTCALL call_11_8_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_8_11
#define init_11_8_8_11
void FASTCALL call_11_8_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_8_12
#define init_11_8_8_12
void FASTCALL call_11_8_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_8_13
#define init_11_8_8_13
void FASTCALL call_11_8_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_8_14
#define init_11_8_8_14
void FASTCALL call_11_8_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_8_15
#define init_11_8_8_15
void FASTCALL call_11_8_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_9_0
#define init_11_8_9_0
void FASTCALL call_11_8_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_9_1
#define init_11_8_9_1
void FASTCALL call_11_8_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_9_2
#define init_11_8_9_2
void FASTCALL call_11_8_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_9_3
#define init_11_8_9_3
void FASTCALL call_11_8_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_9_4
#define init_11_8_9_4
void FASTCALL call_11_8_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_9_5
#define init_11_8_9_5
void FASTCALL call_11_8_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_9_6
#define init_11_8_9_6
void FASTCALL call_11_8_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_9_7
#define init_11_8_9_7
void FASTCALL call_11_8_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_9_8
#define init_11_8_9_8
void FASTCALL call_11_8_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_9_9
#define init_11_8_9_9
void FASTCALL call_11_8_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_9_10
#define init_11_8_9_10
void FASTCALL call_11_8_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_9_11
#define init_11_8_9_11
void FASTCALL call_11_8_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_9_12
#define init_11_8_9_12
void FASTCALL call_11_8_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_9_13
#define init_11_8_9_13
void FASTCALL call_11_8_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_9_14
#define init_11_8_9_14
void FASTCALL call_11_8_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_9_15
#define init_11_8_9_15
void FASTCALL call_11_8_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_10_0
#define init_11_8_10_0
void FASTCALL call_11_8_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_10_1
#define init_11_8_10_1
void FASTCALL call_11_8_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_10_2
#define init_11_8_10_2
void FASTCALL call_11_8_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_10_3
#define init_11_8_10_3
void FASTCALL call_11_8_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_10_4
#define init_11_8_10_4
void FASTCALL call_11_8_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_10_5
#define init_11_8_10_5
void FASTCALL call_11_8_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_10_6
#define init_11_8_10_6
void FASTCALL call_11_8_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_10_7
#define init_11_8_10_7
void FASTCALL call_11_8_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_10_8
#define init_11_8_10_8
void FASTCALL call_11_8_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_10_9
#define init_11_8_10_9
void FASTCALL call_11_8_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_10_10
#define init_11_8_10_10
void FASTCALL call_11_8_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_10_11
#define init_11_8_10_11
void FASTCALL call_11_8_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_10_12
#define init_11_8_10_12
void FASTCALL call_11_8_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_10_13
#define init_11_8_10_13
void FASTCALL call_11_8_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_10_14
#define init_11_8_10_14
void FASTCALL call_11_8_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_10_15
#define init_11_8_10_15
void FASTCALL call_11_8_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_11_0
#define init_11_8_11_0
void FASTCALL call_11_8_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_11_1
#define init_11_8_11_1
void FASTCALL call_11_8_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_11_2
#define init_11_8_11_2
void FASTCALL call_11_8_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_11_3
#define init_11_8_11_3
void FASTCALL call_11_8_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_11_4
#define init_11_8_11_4
void FASTCALL call_11_8_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_11_5
#define init_11_8_11_5
void FASTCALL call_11_8_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_11_6
#define init_11_8_11_6
void FASTCALL call_11_8_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_11_7
#define init_11_8_11_7
void FASTCALL call_11_8_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_11_8
#define init_11_8_11_8
void FASTCALL call_11_8_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_11_9
#define init_11_8_11_9
void FASTCALL call_11_8_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_11_10
#define init_11_8_11_10
void FASTCALL call_11_8_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_11_11
#define init_11_8_11_11
void FASTCALL call_11_8_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_11_12
#define init_11_8_11_12
void FASTCALL call_11_8_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_11_13
#define init_11_8_11_13
void FASTCALL call_11_8_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_11_14
#define init_11_8_11_14
void FASTCALL call_11_8_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_11_15
#define init_11_8_11_15
void FASTCALL call_11_8_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_12_0
#define init_11_8_12_0
void FASTCALL call_11_8_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_12_1
#define init_11_8_12_1
void FASTCALL call_11_8_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_12_2
#define init_11_8_12_2
void FASTCALL call_11_8_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_12_3
#define init_11_8_12_3
void FASTCALL call_11_8_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_12_4
#define init_11_8_12_4
void FASTCALL call_11_8_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_12_5
#define init_11_8_12_5
void FASTCALL call_11_8_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_12_6
#define init_11_8_12_6
void FASTCALL call_11_8_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_12_7
#define init_11_8_12_7
void FASTCALL call_11_8_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_12_8
#define init_11_8_12_8
void FASTCALL call_11_8_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_12_9
#define init_11_8_12_9
void FASTCALL call_11_8_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_12_10
#define init_11_8_12_10
void FASTCALL call_11_8_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_12_11
#define init_11_8_12_11
void FASTCALL call_11_8_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_12_12
#define init_11_8_12_12
void FASTCALL call_11_8_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_12_13
#define init_11_8_12_13
void FASTCALL call_11_8_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_12_14
#define init_11_8_12_14
void FASTCALL call_11_8_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_12_15
#define init_11_8_12_15
void FASTCALL call_11_8_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_13_0
#define init_11_8_13_0
void FASTCALL call_11_8_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_13_1
#define init_11_8_13_1
void FASTCALL call_11_8_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_13_2
#define init_11_8_13_2
void FASTCALL call_11_8_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_13_3
#define init_11_8_13_3
void FASTCALL call_11_8_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_13_4
#define init_11_8_13_4
void FASTCALL call_11_8_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_13_5
#define init_11_8_13_5
void FASTCALL call_11_8_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_13_6
#define init_11_8_13_6
void FASTCALL call_11_8_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_13_7
#define init_11_8_13_7
void FASTCALL call_11_8_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_13_8
#define init_11_8_13_8
void FASTCALL call_11_8_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_13_9
#define init_11_8_13_9
void FASTCALL call_11_8_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_13_10
#define init_11_8_13_10
void FASTCALL call_11_8_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_13_11
#define init_11_8_13_11
void FASTCALL call_11_8_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_13_12
#define init_11_8_13_12
void FASTCALL call_11_8_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_13_13
#define init_11_8_13_13
void FASTCALL call_11_8_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_13_14
#define init_11_8_13_14
void FASTCALL call_11_8_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_13_15
#define init_11_8_13_15
void FASTCALL call_11_8_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_14_0
#define init_11_8_14_0
void FASTCALL call_11_8_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_14_1
#define init_11_8_14_1
void FASTCALL call_11_8_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_14_2
#define init_11_8_14_2
void FASTCALL call_11_8_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_14_3
#define init_11_8_14_3
void FASTCALL call_11_8_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_14_4
#define init_11_8_14_4
void FASTCALL call_11_8_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_14_5
#define init_11_8_14_5
void FASTCALL call_11_8_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_14_6
#define init_11_8_14_6
void FASTCALL call_11_8_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_14_7
#define init_11_8_14_7
void FASTCALL call_11_8_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_14_8
#define init_11_8_14_8
void FASTCALL call_11_8_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_14_9
#define init_11_8_14_9
void FASTCALL call_11_8_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_14_10
#define init_11_8_14_10
void FASTCALL call_11_8_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_14_11
#define init_11_8_14_11
void FASTCALL call_11_8_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_14_12
#define init_11_8_14_12
void FASTCALL call_11_8_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_14_13
#define init_11_8_14_13
void FASTCALL call_11_8_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_14_14
#define init_11_8_14_14
void FASTCALL call_11_8_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_14_15
#define init_11_8_14_15
void FASTCALL call_11_8_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_15_0
#define init_11_8_15_0
void FASTCALL call_11_8_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_15_1
#define init_11_8_15_1
void FASTCALL call_11_8_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_15_2
#define init_11_8_15_2
void FASTCALL call_11_8_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_15_3
#define init_11_8_15_3
void FASTCALL call_11_8_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_15_4
#define init_11_8_15_4
void FASTCALL call_11_8_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_15_5
#define init_11_8_15_5
void FASTCALL call_11_8_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_15_6
#define init_11_8_15_6
void FASTCALL call_11_8_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_15_7
#define init_11_8_15_7
void FASTCALL call_11_8_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_15_8
#define init_11_8_15_8
void FASTCALL call_11_8_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_15_9
#define init_11_8_15_9
void FASTCALL call_11_8_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_15_10
#define init_11_8_15_10
void FASTCALL call_11_8_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_15_11
#define init_11_8_15_11
void FASTCALL call_11_8_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_15_12
#define init_11_8_15_12
void FASTCALL call_11_8_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_15_13
#define init_11_8_15_13
void FASTCALL call_11_8_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_15_14
#define init_11_8_15_14
void FASTCALL call_11_8_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_8_15_15
#define init_11_8_15_15
void FASTCALL call_11_8_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_0_0
#define init_11_9_0_0
void FASTCALL call_11_9_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_0_1
#define init_11_9_0_1
void FASTCALL call_11_9_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_0_2
#define init_11_9_0_2
void FASTCALL call_11_9_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_0_3
#define init_11_9_0_3
void FASTCALL call_11_9_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_0_4
#define init_11_9_0_4
void FASTCALL call_11_9_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_0_5
#define init_11_9_0_5
void FASTCALL call_11_9_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_0_6
#define init_11_9_0_6
void FASTCALL call_11_9_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_0_7
#define init_11_9_0_7
void FASTCALL call_11_9_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_0_8
#define init_11_9_0_8
void FASTCALL call_11_9_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_0_9
#define init_11_9_0_9
void FASTCALL call_11_9_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_0_10
#define init_11_9_0_10
void FASTCALL call_11_9_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_0_11
#define init_11_9_0_11
void FASTCALL call_11_9_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_0_12
#define init_11_9_0_12
void FASTCALL call_11_9_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_0_13
#define init_11_9_0_13
void FASTCALL call_11_9_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_0_14
#define init_11_9_0_14
void FASTCALL call_11_9_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_0_15
#define init_11_9_0_15
void FASTCALL call_11_9_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_1_0
#define init_11_9_1_0
void FASTCALL call_11_9_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_1_1
#define init_11_9_1_1
void FASTCALL call_11_9_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_1_2
#define init_11_9_1_2
void FASTCALL call_11_9_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_1_3
#define init_11_9_1_3
void FASTCALL call_11_9_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_1_4
#define init_11_9_1_4
void FASTCALL call_11_9_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_1_5
#define init_11_9_1_5
void FASTCALL call_11_9_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_1_6
#define init_11_9_1_6
void FASTCALL call_11_9_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_1_7
#define init_11_9_1_7
void FASTCALL call_11_9_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_1_8
#define init_11_9_1_8
void FASTCALL call_11_9_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_1_9
#define init_11_9_1_9
void FASTCALL call_11_9_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_1_10
#define init_11_9_1_10
void FASTCALL call_11_9_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_1_11
#define init_11_9_1_11
void FASTCALL call_11_9_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_1_12
#define init_11_9_1_12
void FASTCALL call_11_9_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_1_13
#define init_11_9_1_13
void FASTCALL call_11_9_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_1_14
#define init_11_9_1_14
void FASTCALL call_11_9_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_1_15
#define init_11_9_1_15
void FASTCALL call_11_9_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_2_0
#define init_11_9_2_0
void FASTCALL call_11_9_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_2_1
#define init_11_9_2_1
void FASTCALL call_11_9_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_2_2
#define init_11_9_2_2
void FASTCALL call_11_9_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_2_3
#define init_11_9_2_3
void FASTCALL call_11_9_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_2_4
#define init_11_9_2_4
void FASTCALL call_11_9_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_2_5
#define init_11_9_2_5
void FASTCALL call_11_9_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_2_6
#define init_11_9_2_6
void FASTCALL call_11_9_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_2_7
#define init_11_9_2_7
void FASTCALL call_11_9_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_2_8
#define init_11_9_2_8
void FASTCALL call_11_9_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_2_9
#define init_11_9_2_9
void FASTCALL call_11_9_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_2_10
#define init_11_9_2_10
void FASTCALL call_11_9_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_2_11
#define init_11_9_2_11
void FASTCALL call_11_9_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_2_12
#define init_11_9_2_12
void FASTCALL call_11_9_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_2_13
#define init_11_9_2_13
void FASTCALL call_11_9_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_2_14
#define init_11_9_2_14
void FASTCALL call_11_9_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_2_15
#define init_11_9_2_15
void FASTCALL call_11_9_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_3_0
#define init_11_9_3_0
void FASTCALL call_11_9_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_3_1
#define init_11_9_3_1
void FASTCALL call_11_9_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_3_2
#define init_11_9_3_2
void FASTCALL call_11_9_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_3_3
#define init_11_9_3_3
void FASTCALL call_11_9_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_3_4
#define init_11_9_3_4
void FASTCALL call_11_9_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_3_5
#define init_11_9_3_5
void FASTCALL call_11_9_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_3_6
#define init_11_9_3_6
void FASTCALL call_11_9_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_3_7
#define init_11_9_3_7
void FASTCALL call_11_9_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_3_8
#define init_11_9_3_8
void FASTCALL call_11_9_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_3_9
#define init_11_9_3_9
void FASTCALL call_11_9_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_3_10
#define init_11_9_3_10
void FASTCALL call_11_9_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_3_11
#define init_11_9_3_11
void FASTCALL call_11_9_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_3_12
#define init_11_9_3_12
void FASTCALL call_11_9_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_3_13
#define init_11_9_3_13
void FASTCALL call_11_9_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_3_14
#define init_11_9_3_14
void FASTCALL call_11_9_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_3_15
#define init_11_9_3_15
void FASTCALL call_11_9_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_4_0
#define init_11_9_4_0
void FASTCALL call_11_9_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_4_1
#define init_11_9_4_1
void FASTCALL call_11_9_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_4_2
#define init_11_9_4_2
void FASTCALL call_11_9_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_4_3
#define init_11_9_4_3
void FASTCALL call_11_9_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_4_4
#define init_11_9_4_4
void FASTCALL call_11_9_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_4_5
#define init_11_9_4_5
void FASTCALL call_11_9_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_4_6
#define init_11_9_4_6
void FASTCALL call_11_9_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_4_7
#define init_11_9_4_7
void FASTCALL call_11_9_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_4_8
#define init_11_9_4_8
void FASTCALL call_11_9_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_4_9
#define init_11_9_4_9
void FASTCALL call_11_9_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_4_10
#define init_11_9_4_10
void FASTCALL call_11_9_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_4_11
#define init_11_9_4_11
void FASTCALL call_11_9_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_4_12
#define init_11_9_4_12
void FASTCALL call_11_9_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_4_13
#define init_11_9_4_13
void FASTCALL call_11_9_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_4_14
#define init_11_9_4_14
void FASTCALL call_11_9_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_4_15
#define init_11_9_4_15
void FASTCALL call_11_9_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_5_0
#define init_11_9_5_0
void FASTCALL call_11_9_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_5_1
#define init_11_9_5_1
void FASTCALL call_11_9_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_5_2
#define init_11_9_5_2
void FASTCALL call_11_9_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_5_3
#define init_11_9_5_3
void FASTCALL call_11_9_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_5_4
#define init_11_9_5_4
void FASTCALL call_11_9_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_5_5
#define init_11_9_5_5
void FASTCALL call_11_9_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_5_6
#define init_11_9_5_6
void FASTCALL call_11_9_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_5_7
#define init_11_9_5_7
void FASTCALL call_11_9_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_5_8
#define init_11_9_5_8
void FASTCALL call_11_9_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_5_9
#define init_11_9_5_9
void FASTCALL call_11_9_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_5_10
#define init_11_9_5_10
void FASTCALL call_11_9_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_5_11
#define init_11_9_5_11
void FASTCALL call_11_9_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_5_12
#define init_11_9_5_12
void FASTCALL call_11_9_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_5_13
#define init_11_9_5_13
void FASTCALL call_11_9_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_5_14
#define init_11_9_5_14
void FASTCALL call_11_9_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_5_15
#define init_11_9_5_15
void FASTCALL call_11_9_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_6_0
#define init_11_9_6_0
void FASTCALL call_11_9_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_6_1
#define init_11_9_6_1
void FASTCALL call_11_9_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_6_2
#define init_11_9_6_2
void FASTCALL call_11_9_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_6_3
#define init_11_9_6_3
void FASTCALL call_11_9_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_6_4
#define init_11_9_6_4
void FASTCALL call_11_9_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_6_5
#define init_11_9_6_5
void FASTCALL call_11_9_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_6_6
#define init_11_9_6_6
void FASTCALL call_11_9_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_6_7
#define init_11_9_6_7
void FASTCALL call_11_9_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_6_8
#define init_11_9_6_8
void FASTCALL call_11_9_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_6_9
#define init_11_9_6_9
void FASTCALL call_11_9_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_6_10
#define init_11_9_6_10
void FASTCALL call_11_9_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_6_11
#define init_11_9_6_11
void FASTCALL call_11_9_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_6_12
#define init_11_9_6_12
void FASTCALL call_11_9_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_6_13
#define init_11_9_6_13
void FASTCALL call_11_9_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_6_14
#define init_11_9_6_14
void FASTCALL call_11_9_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_6_15
#define init_11_9_6_15
void FASTCALL call_11_9_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_7_0
#define init_11_9_7_0
void FASTCALL call_11_9_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_7_1
#define init_11_9_7_1
void FASTCALL call_11_9_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_7_2
#define init_11_9_7_2
void FASTCALL call_11_9_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_7_3
#define init_11_9_7_3
void FASTCALL call_11_9_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_7_4
#define init_11_9_7_4
void FASTCALL call_11_9_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_7_5
#define init_11_9_7_5
void FASTCALL call_11_9_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_7_6
#define init_11_9_7_6
void FASTCALL call_11_9_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_7_7
#define init_11_9_7_7
void FASTCALL call_11_9_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_7_8
#define init_11_9_7_8
void FASTCALL call_11_9_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_7_9
#define init_11_9_7_9
void FASTCALL call_11_9_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_7_10
#define init_11_9_7_10
void FASTCALL call_11_9_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_7_11
#define init_11_9_7_11
void FASTCALL call_11_9_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_7_12
#define init_11_9_7_12
void FASTCALL call_11_9_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_7_13
#define init_11_9_7_13
void FASTCALL call_11_9_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_7_14
#define init_11_9_7_14
void FASTCALL call_11_9_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_7_15
#define init_11_9_7_15
void FASTCALL call_11_9_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_8_0
#define init_11_9_8_0
void FASTCALL call_11_9_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_8_1
#define init_11_9_8_1
void FASTCALL call_11_9_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_8_2
#define init_11_9_8_2
void FASTCALL call_11_9_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_8_3
#define init_11_9_8_3
void FASTCALL call_11_9_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_8_4
#define init_11_9_8_4
void FASTCALL call_11_9_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_8_5
#define init_11_9_8_5
void FASTCALL call_11_9_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_8_6
#define init_11_9_8_6
void FASTCALL call_11_9_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_8_7
#define init_11_9_8_7
void FASTCALL call_11_9_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_8_8
#define init_11_9_8_8
void FASTCALL call_11_9_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_8_9
#define init_11_9_8_9
void FASTCALL call_11_9_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_8_10
#define init_11_9_8_10
void FASTCALL call_11_9_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_8_11
#define init_11_9_8_11
void FASTCALL call_11_9_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_8_12
#define init_11_9_8_12
void FASTCALL call_11_9_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_8_13
#define init_11_9_8_13
void FASTCALL call_11_9_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_8_14
#define init_11_9_8_14
void FASTCALL call_11_9_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_8_15
#define init_11_9_8_15
void FASTCALL call_11_9_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_9_0
#define init_11_9_9_0
void FASTCALL call_11_9_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_9_1
#define init_11_9_9_1
void FASTCALL call_11_9_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_9_2
#define init_11_9_9_2
void FASTCALL call_11_9_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_9_3
#define init_11_9_9_3
void FASTCALL call_11_9_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_9_4
#define init_11_9_9_4
void FASTCALL call_11_9_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_9_5
#define init_11_9_9_5
void FASTCALL call_11_9_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_9_6
#define init_11_9_9_6
void FASTCALL call_11_9_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_9_7
#define init_11_9_9_7
void FASTCALL call_11_9_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_9_8
#define init_11_9_9_8
void FASTCALL call_11_9_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_9_9
#define init_11_9_9_9
void FASTCALL call_11_9_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_9_10
#define init_11_9_9_10
void FASTCALL call_11_9_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_9_11
#define init_11_9_9_11
void FASTCALL call_11_9_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_9_12
#define init_11_9_9_12
void FASTCALL call_11_9_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_9_13
#define init_11_9_9_13
void FASTCALL call_11_9_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_9_14
#define init_11_9_9_14
void FASTCALL call_11_9_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_9_15
#define init_11_9_9_15
void FASTCALL call_11_9_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_10_0
#define init_11_9_10_0
void FASTCALL call_11_9_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_10_1
#define init_11_9_10_1
void FASTCALL call_11_9_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_10_2
#define init_11_9_10_2
void FASTCALL call_11_9_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_10_3
#define init_11_9_10_3
void FASTCALL call_11_9_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_10_4
#define init_11_9_10_4
void FASTCALL call_11_9_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_10_5
#define init_11_9_10_5
void FASTCALL call_11_9_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_10_6
#define init_11_9_10_6
void FASTCALL call_11_9_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_10_7
#define init_11_9_10_7
void FASTCALL call_11_9_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_10_8
#define init_11_9_10_8
void FASTCALL call_11_9_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_10_9
#define init_11_9_10_9
void FASTCALL call_11_9_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_10_10
#define init_11_9_10_10
void FASTCALL call_11_9_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_10_11
#define init_11_9_10_11
void FASTCALL call_11_9_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_10_12
#define init_11_9_10_12
void FASTCALL call_11_9_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_10_13
#define init_11_9_10_13
void FASTCALL call_11_9_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_10_14
#define init_11_9_10_14
void FASTCALL call_11_9_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_10_15
#define init_11_9_10_15
void FASTCALL call_11_9_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_11_0
#define init_11_9_11_0
void FASTCALL call_11_9_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_11_1
#define init_11_9_11_1
void FASTCALL call_11_9_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_11_2
#define init_11_9_11_2
void FASTCALL call_11_9_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_11_3
#define init_11_9_11_3
void FASTCALL call_11_9_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_11_4
#define init_11_9_11_4
void FASTCALL call_11_9_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_11_5
#define init_11_9_11_5
void FASTCALL call_11_9_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_11_6
#define init_11_9_11_6
void FASTCALL call_11_9_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_11_7
#define init_11_9_11_7
void FASTCALL call_11_9_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_11_8
#define init_11_9_11_8
void FASTCALL call_11_9_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_11_9
#define init_11_9_11_9
void FASTCALL call_11_9_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_11_10
#define init_11_9_11_10
void FASTCALL call_11_9_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_11_11
#define init_11_9_11_11
void FASTCALL call_11_9_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_11_12
#define init_11_9_11_12
void FASTCALL call_11_9_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_11_13
#define init_11_9_11_13
void FASTCALL call_11_9_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_11_14
#define init_11_9_11_14
void FASTCALL call_11_9_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_11_15
#define init_11_9_11_15
void FASTCALL call_11_9_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_12_0
#define init_11_9_12_0
void FASTCALL call_11_9_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_12_1
#define init_11_9_12_1
void FASTCALL call_11_9_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_12_2
#define init_11_9_12_2
void FASTCALL call_11_9_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_12_3
#define init_11_9_12_3
void FASTCALL call_11_9_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_12_4
#define init_11_9_12_4
void FASTCALL call_11_9_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_12_5
#define init_11_9_12_5
void FASTCALL call_11_9_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_12_6
#define init_11_9_12_6
void FASTCALL call_11_9_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_12_7
#define init_11_9_12_7
void FASTCALL call_11_9_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_12_8
#define init_11_9_12_8
void FASTCALL call_11_9_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_12_9
#define init_11_9_12_9
void FASTCALL call_11_9_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_12_10
#define init_11_9_12_10
void FASTCALL call_11_9_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_12_11
#define init_11_9_12_11
void FASTCALL call_11_9_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_12_12
#define init_11_9_12_12
void FASTCALL call_11_9_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_12_13
#define init_11_9_12_13
void FASTCALL call_11_9_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_12_14
#define init_11_9_12_14
void FASTCALL call_11_9_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_12_15
#define init_11_9_12_15
void FASTCALL call_11_9_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_13_0
#define init_11_9_13_0
void FASTCALL call_11_9_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_13_1
#define init_11_9_13_1
void FASTCALL call_11_9_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_13_2
#define init_11_9_13_2
void FASTCALL call_11_9_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_13_3
#define init_11_9_13_3
void FASTCALL call_11_9_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_13_4
#define init_11_9_13_4
void FASTCALL call_11_9_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_13_5
#define init_11_9_13_5
void FASTCALL call_11_9_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_13_6
#define init_11_9_13_6
void FASTCALL call_11_9_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_13_7
#define init_11_9_13_7
void FASTCALL call_11_9_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_13_8
#define init_11_9_13_8
void FASTCALL call_11_9_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_13_9
#define init_11_9_13_9
void FASTCALL call_11_9_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_13_10
#define init_11_9_13_10
void FASTCALL call_11_9_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_13_11
#define init_11_9_13_11
void FASTCALL call_11_9_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_13_12
#define init_11_9_13_12
void FASTCALL call_11_9_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_13_13
#define init_11_9_13_13
void FASTCALL call_11_9_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_13_14
#define init_11_9_13_14
void FASTCALL call_11_9_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_13_15
#define init_11_9_13_15
void FASTCALL call_11_9_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_14_0
#define init_11_9_14_0
void FASTCALL call_11_9_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_14_1
#define init_11_9_14_1
void FASTCALL call_11_9_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_14_2
#define init_11_9_14_2
void FASTCALL call_11_9_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_14_3
#define init_11_9_14_3
void FASTCALL call_11_9_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_14_4
#define init_11_9_14_4
void FASTCALL call_11_9_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_14_5
#define init_11_9_14_5
void FASTCALL call_11_9_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_14_6
#define init_11_9_14_6
void FASTCALL call_11_9_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_14_7
#define init_11_9_14_7
void FASTCALL call_11_9_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_14_8
#define init_11_9_14_8
void FASTCALL call_11_9_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_14_9
#define init_11_9_14_9
void FASTCALL call_11_9_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_14_10
#define init_11_9_14_10
void FASTCALL call_11_9_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_14_11
#define init_11_9_14_11
void FASTCALL call_11_9_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_14_12
#define init_11_9_14_12
void FASTCALL call_11_9_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_14_13
#define init_11_9_14_13
void FASTCALL call_11_9_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_14_14
#define init_11_9_14_14
void FASTCALL call_11_9_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_14_15
#define init_11_9_14_15
void FASTCALL call_11_9_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_15_0
#define init_11_9_15_0
void FASTCALL call_11_9_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_15_1
#define init_11_9_15_1
void FASTCALL call_11_9_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_15_2
#define init_11_9_15_2
void FASTCALL call_11_9_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_15_3
#define init_11_9_15_3
void FASTCALL call_11_9_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_15_4
#define init_11_9_15_4
void FASTCALL call_11_9_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_15_5
#define init_11_9_15_5
void FASTCALL call_11_9_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_15_6
#define init_11_9_15_6
void FASTCALL call_11_9_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_15_7
#define init_11_9_15_7
void FASTCALL call_11_9_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_15_8
#define init_11_9_15_8
void FASTCALL call_11_9_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_15_9
#define init_11_9_15_9
void FASTCALL call_11_9_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_15_10
#define init_11_9_15_10
void FASTCALL call_11_9_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_15_11
#define init_11_9_15_11
void FASTCALL call_11_9_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_15_12
#define init_11_9_15_12
void FASTCALL call_11_9_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_15_13
#define init_11_9_15_13
void FASTCALL call_11_9_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_15_14
#define init_11_9_15_14
void FASTCALL call_11_9_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_9_15_15
#define init_11_9_15_15
void FASTCALL call_11_9_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_0_0
#define init_11_10_0_0
void FASTCALL call_11_10_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_0_1
#define init_11_10_0_1
void FASTCALL call_11_10_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_0_2
#define init_11_10_0_2
void FASTCALL call_11_10_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_0_3
#define init_11_10_0_3
void FASTCALL call_11_10_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_0_4
#define init_11_10_0_4
void FASTCALL call_11_10_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_0_5
#define init_11_10_0_5
void FASTCALL call_11_10_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_0_6
#define init_11_10_0_6
void FASTCALL call_11_10_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_0_7
#define init_11_10_0_7
void FASTCALL call_11_10_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_0_8
#define init_11_10_0_8
void FASTCALL call_11_10_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_0_9
#define init_11_10_0_9
void FASTCALL call_11_10_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_0_10
#define init_11_10_0_10
void FASTCALL call_11_10_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_0_11
#define init_11_10_0_11
void FASTCALL call_11_10_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_0_12
#define init_11_10_0_12
void FASTCALL call_11_10_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_0_13
#define init_11_10_0_13
void FASTCALL call_11_10_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_0_14
#define init_11_10_0_14
void FASTCALL call_11_10_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_0_15
#define init_11_10_0_15
void FASTCALL call_11_10_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_1_0
#define init_11_10_1_0
void FASTCALL call_11_10_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_1_1
#define init_11_10_1_1
void FASTCALL call_11_10_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_1_2
#define init_11_10_1_2
void FASTCALL call_11_10_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_1_3
#define init_11_10_1_3
void FASTCALL call_11_10_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_1_4
#define init_11_10_1_4
void FASTCALL call_11_10_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_1_5
#define init_11_10_1_5
void FASTCALL call_11_10_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_1_6
#define init_11_10_1_6
void FASTCALL call_11_10_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_1_7
#define init_11_10_1_7
void FASTCALL call_11_10_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_1_8
#define init_11_10_1_8
void FASTCALL call_11_10_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_1_9
#define init_11_10_1_9
void FASTCALL call_11_10_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_1_10
#define init_11_10_1_10
void FASTCALL call_11_10_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_1_11
#define init_11_10_1_11
void FASTCALL call_11_10_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_1_12
#define init_11_10_1_12
void FASTCALL call_11_10_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_1_13
#define init_11_10_1_13
void FASTCALL call_11_10_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_1_14
#define init_11_10_1_14
void FASTCALL call_11_10_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_1_15
#define init_11_10_1_15
void FASTCALL call_11_10_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_2_0
#define init_11_10_2_0
void FASTCALL call_11_10_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_2_1
#define init_11_10_2_1
void FASTCALL call_11_10_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_2_2
#define init_11_10_2_2
void FASTCALL call_11_10_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_2_3
#define init_11_10_2_3
void FASTCALL call_11_10_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_2_4
#define init_11_10_2_4
void FASTCALL call_11_10_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_2_5
#define init_11_10_2_5
void FASTCALL call_11_10_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_2_6
#define init_11_10_2_6
void FASTCALL call_11_10_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_2_7
#define init_11_10_2_7
void FASTCALL call_11_10_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_2_8
#define init_11_10_2_8
void FASTCALL call_11_10_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_2_9
#define init_11_10_2_9
void FASTCALL call_11_10_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_2_10
#define init_11_10_2_10
void FASTCALL call_11_10_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_2_11
#define init_11_10_2_11
void FASTCALL call_11_10_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_2_12
#define init_11_10_2_12
void FASTCALL call_11_10_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_2_13
#define init_11_10_2_13
void FASTCALL call_11_10_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_2_14
#define init_11_10_2_14
void FASTCALL call_11_10_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_2_15
#define init_11_10_2_15
void FASTCALL call_11_10_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_3_0
#define init_11_10_3_0
void FASTCALL call_11_10_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_3_1
#define init_11_10_3_1
void FASTCALL call_11_10_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_3_2
#define init_11_10_3_2
void FASTCALL call_11_10_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_3_3
#define init_11_10_3_3
void FASTCALL call_11_10_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_3_4
#define init_11_10_3_4
void FASTCALL call_11_10_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_3_5
#define init_11_10_3_5
void FASTCALL call_11_10_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_3_6
#define init_11_10_3_6
void FASTCALL call_11_10_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_3_7
#define init_11_10_3_7
void FASTCALL call_11_10_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_3_8
#define init_11_10_3_8
void FASTCALL call_11_10_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_3_9
#define init_11_10_3_9
void FASTCALL call_11_10_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_3_10
#define init_11_10_3_10
void FASTCALL call_11_10_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_3_11
#define init_11_10_3_11
void FASTCALL call_11_10_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_3_12
#define init_11_10_3_12
void FASTCALL call_11_10_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_3_13
#define init_11_10_3_13
void FASTCALL call_11_10_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_3_14
#define init_11_10_3_14
void FASTCALL call_11_10_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_3_15
#define init_11_10_3_15
void FASTCALL call_11_10_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_4_0
#define init_11_10_4_0
void FASTCALL call_11_10_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_4_1
#define init_11_10_4_1
void FASTCALL call_11_10_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_4_2
#define init_11_10_4_2
void FASTCALL call_11_10_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_4_3
#define init_11_10_4_3
void FASTCALL call_11_10_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_4_4
#define init_11_10_4_4
void FASTCALL call_11_10_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_4_5
#define init_11_10_4_5
void FASTCALL call_11_10_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_4_6
#define init_11_10_4_6
void FASTCALL call_11_10_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_4_7
#define init_11_10_4_7
void FASTCALL call_11_10_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_4_8
#define init_11_10_4_8
void FASTCALL call_11_10_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_4_9
#define init_11_10_4_9
void FASTCALL call_11_10_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_4_10
#define init_11_10_4_10
void FASTCALL call_11_10_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_4_11
#define init_11_10_4_11
void FASTCALL call_11_10_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_4_12
#define init_11_10_4_12
void FASTCALL call_11_10_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_4_13
#define init_11_10_4_13
void FASTCALL call_11_10_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_4_14
#define init_11_10_4_14
void FASTCALL call_11_10_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_4_15
#define init_11_10_4_15
void FASTCALL call_11_10_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_5_0
#define init_11_10_5_0
void FASTCALL call_11_10_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_5_1
#define init_11_10_5_1
void FASTCALL call_11_10_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_5_2
#define init_11_10_5_2
void FASTCALL call_11_10_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_5_3
#define init_11_10_5_3
void FASTCALL call_11_10_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_5_4
#define init_11_10_5_4
void FASTCALL call_11_10_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_5_5
#define init_11_10_5_5
void FASTCALL call_11_10_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_5_6
#define init_11_10_5_6
void FASTCALL call_11_10_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_5_7
#define init_11_10_5_7
void FASTCALL call_11_10_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_5_8
#define init_11_10_5_8
void FASTCALL call_11_10_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_5_9
#define init_11_10_5_9
void FASTCALL call_11_10_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_5_10
#define init_11_10_5_10
void FASTCALL call_11_10_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_5_11
#define init_11_10_5_11
void FASTCALL call_11_10_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_5_12
#define init_11_10_5_12
void FASTCALL call_11_10_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_5_13
#define init_11_10_5_13
void FASTCALL call_11_10_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_5_14
#define init_11_10_5_14
void FASTCALL call_11_10_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_5_15
#define init_11_10_5_15
void FASTCALL call_11_10_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_6_0
#define init_11_10_6_0
void FASTCALL call_11_10_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_6_1
#define init_11_10_6_1
void FASTCALL call_11_10_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_6_2
#define init_11_10_6_2
void FASTCALL call_11_10_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_6_3
#define init_11_10_6_3
void FASTCALL call_11_10_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_6_4
#define init_11_10_6_4
void FASTCALL call_11_10_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_6_5
#define init_11_10_6_5
void FASTCALL call_11_10_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_6_6
#define init_11_10_6_6
void FASTCALL call_11_10_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_6_7
#define init_11_10_6_7
void FASTCALL call_11_10_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_6_8
#define init_11_10_6_8
void FASTCALL call_11_10_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_6_9
#define init_11_10_6_9
void FASTCALL call_11_10_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_6_10
#define init_11_10_6_10
void FASTCALL call_11_10_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_6_11
#define init_11_10_6_11
void FASTCALL call_11_10_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_6_12
#define init_11_10_6_12
void FASTCALL call_11_10_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_6_13
#define init_11_10_6_13
void FASTCALL call_11_10_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_6_14
#define init_11_10_6_14
void FASTCALL call_11_10_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_6_15
#define init_11_10_6_15
void FASTCALL call_11_10_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_7_0
#define init_11_10_7_0
void FASTCALL call_11_10_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_7_1
#define init_11_10_7_1
void FASTCALL call_11_10_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_7_2
#define init_11_10_7_2
void FASTCALL call_11_10_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_7_3
#define init_11_10_7_3
void FASTCALL call_11_10_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_7_4
#define init_11_10_7_4
void FASTCALL call_11_10_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_7_5
#define init_11_10_7_5
void FASTCALL call_11_10_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_7_6
#define init_11_10_7_6
void FASTCALL call_11_10_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_7_7
#define init_11_10_7_7
void FASTCALL call_11_10_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_7_8
#define init_11_10_7_8
void FASTCALL call_11_10_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_7_9
#define init_11_10_7_9
void FASTCALL call_11_10_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_7_10
#define init_11_10_7_10
void FASTCALL call_11_10_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_7_11
#define init_11_10_7_11
void FASTCALL call_11_10_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_7_12
#define init_11_10_7_12
void FASTCALL call_11_10_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_7_13
#define init_11_10_7_13
void FASTCALL call_11_10_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_7_14
#define init_11_10_7_14
void FASTCALL call_11_10_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_7_15
#define init_11_10_7_15
void FASTCALL call_11_10_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_8_0
#define init_11_10_8_0
void FASTCALL call_11_10_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_8_1
#define init_11_10_8_1
void FASTCALL call_11_10_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_8_2
#define init_11_10_8_2
void FASTCALL call_11_10_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_8_3
#define init_11_10_8_3
void FASTCALL call_11_10_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_8_4
#define init_11_10_8_4
void FASTCALL call_11_10_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_8_5
#define init_11_10_8_5
void FASTCALL call_11_10_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_8_6
#define init_11_10_8_6
void FASTCALL call_11_10_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_8_7
#define init_11_10_8_7
void FASTCALL call_11_10_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_8_8
#define init_11_10_8_8
void FASTCALL call_11_10_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_8_9
#define init_11_10_8_9
void FASTCALL call_11_10_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_8_10
#define init_11_10_8_10
void FASTCALL call_11_10_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_8_11
#define init_11_10_8_11
void FASTCALL call_11_10_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_8_12
#define init_11_10_8_12
void FASTCALL call_11_10_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_8_13
#define init_11_10_8_13
void FASTCALL call_11_10_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_8_14
#define init_11_10_8_14
void FASTCALL call_11_10_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_8_15
#define init_11_10_8_15
void FASTCALL call_11_10_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_9_0
#define init_11_10_9_0
void FASTCALL call_11_10_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_9_1
#define init_11_10_9_1
void FASTCALL call_11_10_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_9_2
#define init_11_10_9_2
void FASTCALL call_11_10_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_9_3
#define init_11_10_9_3
void FASTCALL call_11_10_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_9_4
#define init_11_10_9_4
void FASTCALL call_11_10_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_9_5
#define init_11_10_9_5
void FASTCALL call_11_10_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_9_6
#define init_11_10_9_6
void FASTCALL call_11_10_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_9_7
#define init_11_10_9_7
void FASTCALL call_11_10_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_9_8
#define init_11_10_9_8
void FASTCALL call_11_10_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_9_9
#define init_11_10_9_9
void FASTCALL call_11_10_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_9_10
#define init_11_10_9_10
void FASTCALL call_11_10_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_9_11
#define init_11_10_9_11
void FASTCALL call_11_10_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_9_12
#define init_11_10_9_12
void FASTCALL call_11_10_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_9_13
#define init_11_10_9_13
void FASTCALL call_11_10_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_9_14
#define init_11_10_9_14
void FASTCALL call_11_10_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_9_15
#define init_11_10_9_15
void FASTCALL call_11_10_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_10_0
#define init_11_10_10_0
void FASTCALL call_11_10_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_10_1
#define init_11_10_10_1
void FASTCALL call_11_10_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_10_2
#define init_11_10_10_2
void FASTCALL call_11_10_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_10_3
#define init_11_10_10_3
void FASTCALL call_11_10_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_10_4
#define init_11_10_10_4
void FASTCALL call_11_10_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_10_5
#define init_11_10_10_5
void FASTCALL call_11_10_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_10_6
#define init_11_10_10_6
void FASTCALL call_11_10_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_10_7
#define init_11_10_10_7
void FASTCALL call_11_10_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_10_8
#define init_11_10_10_8
void FASTCALL call_11_10_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_10_9
#define init_11_10_10_9
void FASTCALL call_11_10_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_10_10
#define init_11_10_10_10
void FASTCALL call_11_10_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_10_11
#define init_11_10_10_11
void FASTCALL call_11_10_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_10_12
#define init_11_10_10_12
void FASTCALL call_11_10_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_10_13
#define init_11_10_10_13
void FASTCALL call_11_10_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_10_14
#define init_11_10_10_14
void FASTCALL call_11_10_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_10_15
#define init_11_10_10_15
void FASTCALL call_11_10_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_11_0
#define init_11_10_11_0
void FASTCALL call_11_10_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_11_1
#define init_11_10_11_1
void FASTCALL call_11_10_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_11_2
#define init_11_10_11_2
void FASTCALL call_11_10_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_11_3
#define init_11_10_11_3
void FASTCALL call_11_10_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_11_4
#define init_11_10_11_4
void FASTCALL call_11_10_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_11_5
#define init_11_10_11_5
void FASTCALL call_11_10_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_11_6
#define init_11_10_11_6
void FASTCALL call_11_10_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_11_7
#define init_11_10_11_7
void FASTCALL call_11_10_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_11_8
#define init_11_10_11_8
void FASTCALL call_11_10_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_11_9
#define init_11_10_11_9
void FASTCALL call_11_10_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_11_10
#define init_11_10_11_10
void FASTCALL call_11_10_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_11_11
#define init_11_10_11_11
void FASTCALL call_11_10_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_11_12
#define init_11_10_11_12
void FASTCALL call_11_10_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_11_13
#define init_11_10_11_13
void FASTCALL call_11_10_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_11_14
#define init_11_10_11_14
void FASTCALL call_11_10_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_11_15
#define init_11_10_11_15
void FASTCALL call_11_10_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_12_0
#define init_11_10_12_0
void FASTCALL call_11_10_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_12_1
#define init_11_10_12_1
void FASTCALL call_11_10_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_12_2
#define init_11_10_12_2
void FASTCALL call_11_10_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_12_3
#define init_11_10_12_3
void FASTCALL call_11_10_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_12_4
#define init_11_10_12_4
void FASTCALL call_11_10_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_12_5
#define init_11_10_12_5
void FASTCALL call_11_10_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_12_6
#define init_11_10_12_6
void FASTCALL call_11_10_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_12_7
#define init_11_10_12_7
void FASTCALL call_11_10_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_12_8
#define init_11_10_12_8
void FASTCALL call_11_10_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_12_9
#define init_11_10_12_9
void FASTCALL call_11_10_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_12_10
#define init_11_10_12_10
void FASTCALL call_11_10_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_12_11
#define init_11_10_12_11
void FASTCALL call_11_10_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_12_12
#define init_11_10_12_12
void FASTCALL call_11_10_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_12_13
#define init_11_10_12_13
void FASTCALL call_11_10_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_12_14
#define init_11_10_12_14
void FASTCALL call_11_10_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_12_15
#define init_11_10_12_15
void FASTCALL call_11_10_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_13_0
#define init_11_10_13_0
void FASTCALL call_11_10_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_13_1
#define init_11_10_13_1
void FASTCALL call_11_10_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_13_2
#define init_11_10_13_2
void FASTCALL call_11_10_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_13_3
#define init_11_10_13_3
void FASTCALL call_11_10_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_13_4
#define init_11_10_13_4
void FASTCALL call_11_10_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_13_5
#define init_11_10_13_5
void FASTCALL call_11_10_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_13_6
#define init_11_10_13_6
void FASTCALL call_11_10_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_13_7
#define init_11_10_13_7
void FASTCALL call_11_10_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_13_8
#define init_11_10_13_8
void FASTCALL call_11_10_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_13_9
#define init_11_10_13_9
void FASTCALL call_11_10_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_13_10
#define init_11_10_13_10
void FASTCALL call_11_10_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_13_11
#define init_11_10_13_11
void FASTCALL call_11_10_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_13_12
#define init_11_10_13_12
void FASTCALL call_11_10_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_13_13
#define init_11_10_13_13
void FASTCALL call_11_10_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_13_14
#define init_11_10_13_14
void FASTCALL call_11_10_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_13_15
#define init_11_10_13_15
void FASTCALL call_11_10_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_14_0
#define init_11_10_14_0
void FASTCALL call_11_10_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_14_1
#define init_11_10_14_1
void FASTCALL call_11_10_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_14_2
#define init_11_10_14_2
void FASTCALL call_11_10_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_14_3
#define init_11_10_14_3
void FASTCALL call_11_10_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_14_4
#define init_11_10_14_4
void FASTCALL call_11_10_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_14_5
#define init_11_10_14_5
void FASTCALL call_11_10_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_14_6
#define init_11_10_14_6
void FASTCALL call_11_10_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_14_7
#define init_11_10_14_7
void FASTCALL call_11_10_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_14_8
#define init_11_10_14_8
void FASTCALL call_11_10_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_14_9
#define init_11_10_14_9
void FASTCALL call_11_10_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_14_10
#define init_11_10_14_10
void FASTCALL call_11_10_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_14_11
#define init_11_10_14_11
void FASTCALL call_11_10_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_14_12
#define init_11_10_14_12
void FASTCALL call_11_10_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_14_13
#define init_11_10_14_13
void FASTCALL call_11_10_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_14_14
#define init_11_10_14_14
void FASTCALL call_11_10_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_14_15
#define init_11_10_14_15
void FASTCALL call_11_10_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_15_0
#define init_11_10_15_0
void FASTCALL call_11_10_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_15_1
#define init_11_10_15_1
void FASTCALL call_11_10_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_15_2
#define init_11_10_15_2
void FASTCALL call_11_10_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_15_3
#define init_11_10_15_3
void FASTCALL call_11_10_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_15_4
#define init_11_10_15_4
void FASTCALL call_11_10_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_15_5
#define init_11_10_15_5
void FASTCALL call_11_10_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_15_6
#define init_11_10_15_6
void FASTCALL call_11_10_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_15_7
#define init_11_10_15_7
void FASTCALL call_11_10_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_15_8
#define init_11_10_15_8
void FASTCALL call_11_10_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_15_9
#define init_11_10_15_9
void FASTCALL call_11_10_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_15_10
#define init_11_10_15_10
void FASTCALL call_11_10_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_15_11
#define init_11_10_15_11
void FASTCALL call_11_10_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_15_12
#define init_11_10_15_12
void FASTCALL call_11_10_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_15_13
#define init_11_10_15_13
void FASTCALL call_11_10_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_15_14
#define init_11_10_15_14
void FASTCALL call_11_10_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_10_15_15
#define init_11_10_15_15
void FASTCALL call_11_10_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_0_0
#define init_11_11_0_0
void FASTCALL call_11_11_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_0_1
#define init_11_11_0_1
void FASTCALL call_11_11_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_0_2
#define init_11_11_0_2
void FASTCALL call_11_11_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_0_3
#define init_11_11_0_3
void FASTCALL call_11_11_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_0_4
#define init_11_11_0_4
void FASTCALL call_11_11_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_0_5
#define init_11_11_0_5
void FASTCALL call_11_11_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_0_6
#define init_11_11_0_6
void FASTCALL call_11_11_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_0_7
#define init_11_11_0_7
void FASTCALL call_11_11_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_0_8
#define init_11_11_0_8
void FASTCALL call_11_11_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_0_9
#define init_11_11_0_9
void FASTCALL call_11_11_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_0_10
#define init_11_11_0_10
void FASTCALL call_11_11_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_0_11
#define init_11_11_0_11
void FASTCALL call_11_11_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_0_12
#define init_11_11_0_12
void FASTCALL call_11_11_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_0_13
#define init_11_11_0_13
void FASTCALL call_11_11_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_0_14
#define init_11_11_0_14
void FASTCALL call_11_11_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_0_15
#define init_11_11_0_15
void FASTCALL call_11_11_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_1_0
#define init_11_11_1_0
void FASTCALL call_11_11_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_1_1
#define init_11_11_1_1
void FASTCALL call_11_11_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_1_2
#define init_11_11_1_2
void FASTCALL call_11_11_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_1_3
#define init_11_11_1_3
void FASTCALL call_11_11_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_1_4
#define init_11_11_1_4
void FASTCALL call_11_11_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_1_5
#define init_11_11_1_5
void FASTCALL call_11_11_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_1_6
#define init_11_11_1_6
void FASTCALL call_11_11_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_1_7
#define init_11_11_1_7
void FASTCALL call_11_11_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_1_8
#define init_11_11_1_8
void FASTCALL call_11_11_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_1_9
#define init_11_11_1_9
void FASTCALL call_11_11_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_1_10
#define init_11_11_1_10
void FASTCALL call_11_11_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_1_11
#define init_11_11_1_11
void FASTCALL call_11_11_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_1_12
#define init_11_11_1_12
void FASTCALL call_11_11_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_1_13
#define init_11_11_1_13
void FASTCALL call_11_11_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_1_14
#define init_11_11_1_14
void FASTCALL call_11_11_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_1_15
#define init_11_11_1_15
void FASTCALL call_11_11_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_2_0
#define init_11_11_2_0
void FASTCALL call_11_11_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_2_1
#define init_11_11_2_1
void FASTCALL call_11_11_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_2_2
#define init_11_11_2_2
void FASTCALL call_11_11_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_2_3
#define init_11_11_2_3
void FASTCALL call_11_11_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_2_4
#define init_11_11_2_4
void FASTCALL call_11_11_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_2_5
#define init_11_11_2_5
void FASTCALL call_11_11_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_2_6
#define init_11_11_2_6
void FASTCALL call_11_11_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_2_7
#define init_11_11_2_7
void FASTCALL call_11_11_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_2_8
#define init_11_11_2_8
void FASTCALL call_11_11_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_2_9
#define init_11_11_2_9
void FASTCALL call_11_11_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_2_10
#define init_11_11_2_10
void FASTCALL call_11_11_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_2_11
#define init_11_11_2_11
void FASTCALL call_11_11_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_2_12
#define init_11_11_2_12
void FASTCALL call_11_11_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_2_13
#define init_11_11_2_13
void FASTCALL call_11_11_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_2_14
#define init_11_11_2_14
void FASTCALL call_11_11_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_2_15
#define init_11_11_2_15
void FASTCALL call_11_11_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_3_0
#define init_11_11_3_0
void FASTCALL call_11_11_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_3_1
#define init_11_11_3_1
void FASTCALL call_11_11_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_3_2
#define init_11_11_3_2
void FASTCALL call_11_11_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_3_3
#define init_11_11_3_3
void FASTCALL call_11_11_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_3_4
#define init_11_11_3_4
void FASTCALL call_11_11_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_3_5
#define init_11_11_3_5
void FASTCALL call_11_11_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_3_6
#define init_11_11_3_6
void FASTCALL call_11_11_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_3_7
#define init_11_11_3_7
void FASTCALL call_11_11_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_3_8
#define init_11_11_3_8
void FASTCALL call_11_11_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_3_9
#define init_11_11_3_9
void FASTCALL call_11_11_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_3_10
#define init_11_11_3_10
void FASTCALL call_11_11_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_3_11
#define init_11_11_3_11
void FASTCALL call_11_11_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_3_12
#define init_11_11_3_12
void FASTCALL call_11_11_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_3_13
#define init_11_11_3_13
void FASTCALL call_11_11_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_3_14
#define init_11_11_3_14
void FASTCALL call_11_11_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_3_15
#define init_11_11_3_15
void FASTCALL call_11_11_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_4_0
#define init_11_11_4_0
void FASTCALL call_11_11_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_4_1
#define init_11_11_4_1
void FASTCALL call_11_11_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_4_2
#define init_11_11_4_2
void FASTCALL call_11_11_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_4_3
#define init_11_11_4_3
void FASTCALL call_11_11_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_4_4
#define init_11_11_4_4
void FASTCALL call_11_11_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_4_5
#define init_11_11_4_5
void FASTCALL call_11_11_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_4_6
#define init_11_11_4_6
void FASTCALL call_11_11_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_4_7
#define init_11_11_4_7
void FASTCALL call_11_11_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_4_8
#define init_11_11_4_8
void FASTCALL call_11_11_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_4_9
#define init_11_11_4_9
void FASTCALL call_11_11_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_4_10
#define init_11_11_4_10
void FASTCALL call_11_11_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_4_11
#define init_11_11_4_11
void FASTCALL call_11_11_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_4_12
#define init_11_11_4_12
void FASTCALL call_11_11_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_4_13
#define init_11_11_4_13
void FASTCALL call_11_11_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_4_14
#define init_11_11_4_14
void FASTCALL call_11_11_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_4_15
#define init_11_11_4_15
void FASTCALL call_11_11_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_5_0
#define init_11_11_5_0
void FASTCALL call_11_11_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_5_1
#define init_11_11_5_1
void FASTCALL call_11_11_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_5_2
#define init_11_11_5_2
void FASTCALL call_11_11_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_5_3
#define init_11_11_5_3
void FASTCALL call_11_11_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_5_4
#define init_11_11_5_4
void FASTCALL call_11_11_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_5_5
#define init_11_11_5_5
void FASTCALL call_11_11_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_5_6
#define init_11_11_5_6
void FASTCALL call_11_11_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_5_7
#define init_11_11_5_7
void FASTCALL call_11_11_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_5_8
#define init_11_11_5_8
void FASTCALL call_11_11_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_5_9
#define init_11_11_5_9
void FASTCALL call_11_11_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_5_10
#define init_11_11_5_10
void FASTCALL call_11_11_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_5_11
#define init_11_11_5_11
void FASTCALL call_11_11_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_5_12
#define init_11_11_5_12
void FASTCALL call_11_11_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_5_13
#define init_11_11_5_13
void FASTCALL call_11_11_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_5_14
#define init_11_11_5_14
void FASTCALL call_11_11_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_5_15
#define init_11_11_5_15
void FASTCALL call_11_11_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_6_0
#define init_11_11_6_0
void FASTCALL call_11_11_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_6_1
#define init_11_11_6_1
void FASTCALL call_11_11_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_6_2
#define init_11_11_6_2
void FASTCALL call_11_11_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_6_3
#define init_11_11_6_3
void FASTCALL call_11_11_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_6_4
#define init_11_11_6_4
void FASTCALL call_11_11_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_6_5
#define init_11_11_6_5
void FASTCALL call_11_11_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_6_6
#define init_11_11_6_6
void FASTCALL call_11_11_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_6_7
#define init_11_11_6_7
void FASTCALL call_11_11_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_6_8
#define init_11_11_6_8
void FASTCALL call_11_11_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_6_9
#define init_11_11_6_9
void FASTCALL call_11_11_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_6_10
#define init_11_11_6_10
void FASTCALL call_11_11_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_6_11
#define init_11_11_6_11
void FASTCALL call_11_11_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_6_12
#define init_11_11_6_12
void FASTCALL call_11_11_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_6_13
#define init_11_11_6_13
void FASTCALL call_11_11_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_6_14
#define init_11_11_6_14
void FASTCALL call_11_11_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_6_15
#define init_11_11_6_15
void FASTCALL call_11_11_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_7_0
#define init_11_11_7_0
void FASTCALL call_11_11_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_7_1
#define init_11_11_7_1
void FASTCALL call_11_11_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_7_2
#define init_11_11_7_2
void FASTCALL call_11_11_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_7_3
#define init_11_11_7_3
void FASTCALL call_11_11_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_7_4
#define init_11_11_7_4
void FASTCALL call_11_11_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_7_5
#define init_11_11_7_5
void FASTCALL call_11_11_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_7_6
#define init_11_11_7_6
void FASTCALL call_11_11_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_7_7
#define init_11_11_7_7
void FASTCALL call_11_11_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_7_8
#define init_11_11_7_8
void FASTCALL call_11_11_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_7_9
#define init_11_11_7_9
void FASTCALL call_11_11_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_7_10
#define init_11_11_7_10
void FASTCALL call_11_11_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_7_11
#define init_11_11_7_11
void FASTCALL call_11_11_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_7_12
#define init_11_11_7_12
void FASTCALL call_11_11_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_7_13
#define init_11_11_7_13
void FASTCALL call_11_11_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_7_14
#define init_11_11_7_14
void FASTCALL call_11_11_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_7_15
#define init_11_11_7_15
void FASTCALL call_11_11_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_8_0
#define init_11_11_8_0
void FASTCALL call_11_11_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_8_1
#define init_11_11_8_1
void FASTCALL call_11_11_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_8_2
#define init_11_11_8_2
void FASTCALL call_11_11_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_8_3
#define init_11_11_8_3
void FASTCALL call_11_11_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_8_4
#define init_11_11_8_4
void FASTCALL call_11_11_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_8_5
#define init_11_11_8_5
void FASTCALL call_11_11_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_8_6
#define init_11_11_8_6
void FASTCALL call_11_11_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_8_7
#define init_11_11_8_7
void FASTCALL call_11_11_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_8_8
#define init_11_11_8_8
void FASTCALL call_11_11_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_8_9
#define init_11_11_8_9
void FASTCALL call_11_11_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_8_10
#define init_11_11_8_10
void FASTCALL call_11_11_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_8_11
#define init_11_11_8_11
void FASTCALL call_11_11_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_8_12
#define init_11_11_8_12
void FASTCALL call_11_11_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_8_13
#define init_11_11_8_13
void FASTCALL call_11_11_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_8_14
#define init_11_11_8_14
void FASTCALL call_11_11_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_8_15
#define init_11_11_8_15
void FASTCALL call_11_11_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_9_0
#define init_11_11_9_0
void FASTCALL call_11_11_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_9_1
#define init_11_11_9_1
void FASTCALL call_11_11_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_9_2
#define init_11_11_9_2
void FASTCALL call_11_11_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_9_3
#define init_11_11_9_3
void FASTCALL call_11_11_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_9_4
#define init_11_11_9_4
void FASTCALL call_11_11_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_9_5
#define init_11_11_9_5
void FASTCALL call_11_11_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_9_6
#define init_11_11_9_6
void FASTCALL call_11_11_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_9_7
#define init_11_11_9_7
void FASTCALL call_11_11_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_9_8
#define init_11_11_9_8
void FASTCALL call_11_11_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_9_9
#define init_11_11_9_9
void FASTCALL call_11_11_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_9_10
#define init_11_11_9_10
void FASTCALL call_11_11_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_9_11
#define init_11_11_9_11
void FASTCALL call_11_11_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_9_12
#define init_11_11_9_12
void FASTCALL call_11_11_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_9_13
#define init_11_11_9_13
void FASTCALL call_11_11_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_9_14
#define init_11_11_9_14
void FASTCALL call_11_11_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_9_15
#define init_11_11_9_15
void FASTCALL call_11_11_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_10_0
#define init_11_11_10_0
void FASTCALL call_11_11_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_10_1
#define init_11_11_10_1
void FASTCALL call_11_11_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_10_2
#define init_11_11_10_2
void FASTCALL call_11_11_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_10_3
#define init_11_11_10_3
void FASTCALL call_11_11_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_10_4
#define init_11_11_10_4
void FASTCALL call_11_11_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_10_5
#define init_11_11_10_5
void FASTCALL call_11_11_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_10_6
#define init_11_11_10_6
void FASTCALL call_11_11_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_10_7
#define init_11_11_10_7
void FASTCALL call_11_11_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_10_8
#define init_11_11_10_8
void FASTCALL call_11_11_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_10_9
#define init_11_11_10_9
void FASTCALL call_11_11_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_10_10
#define init_11_11_10_10
void FASTCALL call_11_11_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_10_11
#define init_11_11_10_11
void FASTCALL call_11_11_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_10_12
#define init_11_11_10_12
void FASTCALL call_11_11_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_10_13
#define init_11_11_10_13
void FASTCALL call_11_11_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_10_14
#define init_11_11_10_14
void FASTCALL call_11_11_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_10_15
#define init_11_11_10_15
void FASTCALL call_11_11_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_11_0
#define init_11_11_11_0
void FASTCALL call_11_11_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_11_1
#define init_11_11_11_1
void FASTCALL call_11_11_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_11_2
#define init_11_11_11_2
void FASTCALL call_11_11_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_11_3
#define init_11_11_11_3
void FASTCALL call_11_11_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_11_4
#define init_11_11_11_4
void FASTCALL call_11_11_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_11_5
#define init_11_11_11_5
void FASTCALL call_11_11_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_11_6
#define init_11_11_11_6
void FASTCALL call_11_11_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_11_7
#define init_11_11_11_7
void FASTCALL call_11_11_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_11_8
#define init_11_11_11_8
void FASTCALL call_11_11_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_11_9
#define init_11_11_11_9
void FASTCALL call_11_11_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_11_10
#define init_11_11_11_10
void FASTCALL call_11_11_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_11_11
#define init_11_11_11_11
void FASTCALL call_11_11_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_11_12
#define init_11_11_11_12
void FASTCALL call_11_11_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_11_13
#define init_11_11_11_13
void FASTCALL call_11_11_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_11_14
#define init_11_11_11_14
void FASTCALL call_11_11_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_11_15
#define init_11_11_11_15
void FASTCALL call_11_11_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_12_0
#define init_11_11_12_0
void FASTCALL call_11_11_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_12_1
#define init_11_11_12_1
void FASTCALL call_11_11_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_12_2
#define init_11_11_12_2
void FASTCALL call_11_11_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_12_3
#define init_11_11_12_3
void FASTCALL call_11_11_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_12_4
#define init_11_11_12_4
void FASTCALL call_11_11_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_12_5
#define init_11_11_12_5
void FASTCALL call_11_11_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_12_6
#define init_11_11_12_6
void FASTCALL call_11_11_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_12_7
#define init_11_11_12_7
void FASTCALL call_11_11_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_12_8
#define init_11_11_12_8
void FASTCALL call_11_11_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_12_9
#define init_11_11_12_9
void FASTCALL call_11_11_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_12_10
#define init_11_11_12_10
void FASTCALL call_11_11_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_12_11
#define init_11_11_12_11
void FASTCALL call_11_11_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_12_12
#define init_11_11_12_12
void FASTCALL call_11_11_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_12_13
#define init_11_11_12_13
void FASTCALL call_11_11_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_12_14
#define init_11_11_12_14
void FASTCALL call_11_11_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_12_15
#define init_11_11_12_15
void FASTCALL call_11_11_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_13_0
#define init_11_11_13_0
void FASTCALL call_11_11_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_13_1
#define init_11_11_13_1
void FASTCALL call_11_11_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_13_2
#define init_11_11_13_2
void FASTCALL call_11_11_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_13_3
#define init_11_11_13_3
void FASTCALL call_11_11_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_13_4
#define init_11_11_13_4
void FASTCALL call_11_11_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_13_5
#define init_11_11_13_5
void FASTCALL call_11_11_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_13_6
#define init_11_11_13_6
void FASTCALL call_11_11_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_13_7
#define init_11_11_13_7
void FASTCALL call_11_11_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_13_8
#define init_11_11_13_8
void FASTCALL call_11_11_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_13_9
#define init_11_11_13_9
void FASTCALL call_11_11_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_13_10
#define init_11_11_13_10
void FASTCALL call_11_11_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_13_11
#define init_11_11_13_11
void FASTCALL call_11_11_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_13_12
#define init_11_11_13_12
void FASTCALL call_11_11_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_13_13
#define init_11_11_13_13
void FASTCALL call_11_11_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_13_14
#define init_11_11_13_14
void FASTCALL call_11_11_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_13_15
#define init_11_11_13_15
void FASTCALL call_11_11_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_14_0
#define init_11_11_14_0
void FASTCALL call_11_11_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_14_1
#define init_11_11_14_1
void FASTCALL call_11_11_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_14_2
#define init_11_11_14_2
void FASTCALL call_11_11_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_14_3
#define init_11_11_14_3
void FASTCALL call_11_11_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_14_4
#define init_11_11_14_4
void FASTCALL call_11_11_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_14_5
#define init_11_11_14_5
void FASTCALL call_11_11_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_14_6
#define init_11_11_14_6
void FASTCALL call_11_11_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_14_7
#define init_11_11_14_7
void FASTCALL call_11_11_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_14_8
#define init_11_11_14_8
void FASTCALL call_11_11_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_14_9
#define init_11_11_14_9
void FASTCALL call_11_11_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_14_10
#define init_11_11_14_10
void FASTCALL call_11_11_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_14_11
#define init_11_11_14_11
void FASTCALL call_11_11_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_14_12
#define init_11_11_14_12
void FASTCALL call_11_11_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_14_13
#define init_11_11_14_13
void FASTCALL call_11_11_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_14_14
#define init_11_11_14_14
void FASTCALL call_11_11_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_14_15
#define init_11_11_14_15
void FASTCALL call_11_11_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_15_0
#define init_11_11_15_0
void FASTCALL call_11_11_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_15_1
#define init_11_11_15_1
void FASTCALL call_11_11_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_15_2
#define init_11_11_15_2
void FASTCALL call_11_11_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_15_3
#define init_11_11_15_3
void FASTCALL call_11_11_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_15_4
#define init_11_11_15_4
void FASTCALL call_11_11_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_15_5
#define init_11_11_15_5
void FASTCALL call_11_11_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_15_6
#define init_11_11_15_6
void FASTCALL call_11_11_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_15_7
#define init_11_11_15_7
void FASTCALL call_11_11_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_15_8
#define init_11_11_15_8
void FASTCALL call_11_11_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_15_9
#define init_11_11_15_9
void FASTCALL call_11_11_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_15_10
#define init_11_11_15_10
void FASTCALL call_11_11_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_15_11
#define init_11_11_15_11
void FASTCALL call_11_11_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_15_12
#define init_11_11_15_12
void FASTCALL call_11_11_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_15_13
#define init_11_11_15_13
void FASTCALL call_11_11_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_15_14
#define init_11_11_15_14
void FASTCALL call_11_11_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_11_15_15
#define init_11_11_15_15
void FASTCALL call_11_11_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_0_0
#define init_11_12_0_0
void FASTCALL call_11_12_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_0_1
#define init_11_12_0_1
void FASTCALL call_11_12_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_0_2
#define init_11_12_0_2
void FASTCALL call_11_12_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_0_3
#define init_11_12_0_3
void FASTCALL call_11_12_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_0_4
#define init_11_12_0_4
void FASTCALL call_11_12_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_0_5
#define init_11_12_0_5
void FASTCALL call_11_12_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_0_6
#define init_11_12_0_6
void FASTCALL call_11_12_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_0_7
#define init_11_12_0_7
void FASTCALL call_11_12_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_0_8
#define init_11_12_0_8
void FASTCALL call_11_12_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_0_9
#define init_11_12_0_9
void FASTCALL call_11_12_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_0_10
#define init_11_12_0_10
void FASTCALL call_11_12_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_0_11
#define init_11_12_0_11
void FASTCALL call_11_12_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_0_12
#define init_11_12_0_12
void FASTCALL call_11_12_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_0_13
#define init_11_12_0_13
void FASTCALL call_11_12_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_0_14
#define init_11_12_0_14
void FASTCALL call_11_12_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_0_15
#define init_11_12_0_15
void FASTCALL call_11_12_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_1_0
#define init_11_12_1_0
void FASTCALL call_11_12_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_1_1
#define init_11_12_1_1
void FASTCALL call_11_12_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_1_2
#define init_11_12_1_2
void FASTCALL call_11_12_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_1_3
#define init_11_12_1_3
void FASTCALL call_11_12_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_1_4
#define init_11_12_1_4
void FASTCALL call_11_12_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_1_5
#define init_11_12_1_5
void FASTCALL call_11_12_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_1_6
#define init_11_12_1_6
void FASTCALL call_11_12_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_1_7
#define init_11_12_1_7
void FASTCALL call_11_12_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_1_8
#define init_11_12_1_8
void FASTCALL call_11_12_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_1_9
#define init_11_12_1_9
void FASTCALL call_11_12_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_1_10
#define init_11_12_1_10
void FASTCALL call_11_12_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_1_11
#define init_11_12_1_11
void FASTCALL call_11_12_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_1_12
#define init_11_12_1_12
void FASTCALL call_11_12_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_1_13
#define init_11_12_1_13
void FASTCALL call_11_12_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_1_14
#define init_11_12_1_14
void FASTCALL call_11_12_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_1_15
#define init_11_12_1_15
void FASTCALL call_11_12_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_2_0
#define init_11_12_2_0
void FASTCALL call_11_12_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_2_1
#define init_11_12_2_1
void FASTCALL call_11_12_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_2_2
#define init_11_12_2_2
void FASTCALL call_11_12_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_2_3
#define init_11_12_2_3
void FASTCALL call_11_12_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_2_4
#define init_11_12_2_4
void FASTCALL call_11_12_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_2_5
#define init_11_12_2_5
void FASTCALL call_11_12_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_2_6
#define init_11_12_2_6
void FASTCALL call_11_12_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_2_7
#define init_11_12_2_7
void FASTCALL call_11_12_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_2_8
#define init_11_12_2_8
void FASTCALL call_11_12_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_2_9
#define init_11_12_2_9
void FASTCALL call_11_12_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_2_10
#define init_11_12_2_10
void FASTCALL call_11_12_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_2_11
#define init_11_12_2_11
void FASTCALL call_11_12_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_2_12
#define init_11_12_2_12
void FASTCALL call_11_12_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_2_13
#define init_11_12_2_13
void FASTCALL call_11_12_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_2_14
#define init_11_12_2_14
void FASTCALL call_11_12_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_2_15
#define init_11_12_2_15
void FASTCALL call_11_12_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_3_0
#define init_11_12_3_0
void FASTCALL call_11_12_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_3_1
#define init_11_12_3_1
void FASTCALL call_11_12_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_3_2
#define init_11_12_3_2
void FASTCALL call_11_12_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_3_3
#define init_11_12_3_3
void FASTCALL call_11_12_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_3_4
#define init_11_12_3_4
void FASTCALL call_11_12_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_3_5
#define init_11_12_3_5
void FASTCALL call_11_12_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_3_6
#define init_11_12_3_6
void FASTCALL call_11_12_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_3_7
#define init_11_12_3_7
void FASTCALL call_11_12_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_3_8
#define init_11_12_3_8
void FASTCALL call_11_12_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_3_9
#define init_11_12_3_9
void FASTCALL call_11_12_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_3_10
#define init_11_12_3_10
void FASTCALL call_11_12_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_3_11
#define init_11_12_3_11
void FASTCALL call_11_12_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_3_12
#define init_11_12_3_12
void FASTCALL call_11_12_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_3_13
#define init_11_12_3_13
void FASTCALL call_11_12_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_3_14
#define init_11_12_3_14
void FASTCALL call_11_12_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_3_15
#define init_11_12_3_15
void FASTCALL call_11_12_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_4_0
#define init_11_12_4_0
void FASTCALL call_11_12_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_4_1
#define init_11_12_4_1
void FASTCALL call_11_12_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_4_2
#define init_11_12_4_2
void FASTCALL call_11_12_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_4_3
#define init_11_12_4_3
void FASTCALL call_11_12_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_4_4
#define init_11_12_4_4
void FASTCALL call_11_12_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_4_5
#define init_11_12_4_5
void FASTCALL call_11_12_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_4_6
#define init_11_12_4_6
void FASTCALL call_11_12_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_4_7
#define init_11_12_4_7
void FASTCALL call_11_12_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_4_8
#define init_11_12_4_8
void FASTCALL call_11_12_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_4_9
#define init_11_12_4_9
void FASTCALL call_11_12_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_4_10
#define init_11_12_4_10
void FASTCALL call_11_12_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_4_11
#define init_11_12_4_11
void FASTCALL call_11_12_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_4_12
#define init_11_12_4_12
void FASTCALL call_11_12_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_4_13
#define init_11_12_4_13
void FASTCALL call_11_12_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_4_14
#define init_11_12_4_14
void FASTCALL call_11_12_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_4_15
#define init_11_12_4_15
void FASTCALL call_11_12_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_5_0
#define init_11_12_5_0
void FASTCALL call_11_12_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_5_1
#define init_11_12_5_1
void FASTCALL call_11_12_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_5_2
#define init_11_12_5_2
void FASTCALL call_11_12_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_5_3
#define init_11_12_5_3
void FASTCALL call_11_12_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_5_4
#define init_11_12_5_4
void FASTCALL call_11_12_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_5_5
#define init_11_12_5_5
void FASTCALL call_11_12_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_5_6
#define init_11_12_5_6
void FASTCALL call_11_12_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_5_7
#define init_11_12_5_7
void FASTCALL call_11_12_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_5_8
#define init_11_12_5_8
void FASTCALL call_11_12_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_5_9
#define init_11_12_5_9
void FASTCALL call_11_12_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_5_10
#define init_11_12_5_10
void FASTCALL call_11_12_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_5_11
#define init_11_12_5_11
void FASTCALL call_11_12_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_5_12
#define init_11_12_5_12
void FASTCALL call_11_12_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_5_13
#define init_11_12_5_13
void FASTCALL call_11_12_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_5_14
#define init_11_12_5_14
void FASTCALL call_11_12_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_5_15
#define init_11_12_5_15
void FASTCALL call_11_12_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_6_0
#define init_11_12_6_0
void FASTCALL call_11_12_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_6_1
#define init_11_12_6_1
void FASTCALL call_11_12_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_6_2
#define init_11_12_6_2
void FASTCALL call_11_12_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_6_3
#define init_11_12_6_3
void FASTCALL call_11_12_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_6_4
#define init_11_12_6_4
void FASTCALL call_11_12_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_6_5
#define init_11_12_6_5
void FASTCALL call_11_12_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_6_6
#define init_11_12_6_6
void FASTCALL call_11_12_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_6_7
#define init_11_12_6_7
void FASTCALL call_11_12_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_6_8
#define init_11_12_6_8
void FASTCALL call_11_12_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_6_9
#define init_11_12_6_9
void FASTCALL call_11_12_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_6_10
#define init_11_12_6_10
void FASTCALL call_11_12_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_6_11
#define init_11_12_6_11
void FASTCALL call_11_12_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_6_12
#define init_11_12_6_12
void FASTCALL call_11_12_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_6_13
#define init_11_12_6_13
void FASTCALL call_11_12_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_6_14
#define init_11_12_6_14
void FASTCALL call_11_12_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_6_15
#define init_11_12_6_15
void FASTCALL call_11_12_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_7_0
#define init_11_12_7_0
void FASTCALL call_11_12_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_7_1
#define init_11_12_7_1
void FASTCALL call_11_12_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_7_2
#define init_11_12_7_2
void FASTCALL call_11_12_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_7_3
#define init_11_12_7_3
void FASTCALL call_11_12_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_7_4
#define init_11_12_7_4
void FASTCALL call_11_12_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_7_5
#define init_11_12_7_5
void FASTCALL call_11_12_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_7_6
#define init_11_12_7_6
void FASTCALL call_11_12_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_7_7
#define init_11_12_7_7
void FASTCALL call_11_12_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_7_8
#define init_11_12_7_8
void FASTCALL call_11_12_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_7_9
#define init_11_12_7_9
void FASTCALL call_11_12_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_7_10
#define init_11_12_7_10
void FASTCALL call_11_12_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_7_11
#define init_11_12_7_11
void FASTCALL call_11_12_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_7_12
#define init_11_12_7_12
void FASTCALL call_11_12_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_7_13
#define init_11_12_7_13
void FASTCALL call_11_12_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_7_14
#define init_11_12_7_14
void FASTCALL call_11_12_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_7_15
#define init_11_12_7_15
void FASTCALL call_11_12_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_8_0
#define init_11_12_8_0
void FASTCALL call_11_12_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_8_1
#define init_11_12_8_1
void FASTCALL call_11_12_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_8_2
#define init_11_12_8_2
void FASTCALL call_11_12_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_8_3
#define init_11_12_8_3
void FASTCALL call_11_12_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_8_4
#define init_11_12_8_4
void FASTCALL call_11_12_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_8_5
#define init_11_12_8_5
void FASTCALL call_11_12_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_8_6
#define init_11_12_8_6
void FASTCALL call_11_12_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_8_7
#define init_11_12_8_7
void FASTCALL call_11_12_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_8_8
#define init_11_12_8_8
void FASTCALL call_11_12_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_8_9
#define init_11_12_8_9
void FASTCALL call_11_12_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_8_10
#define init_11_12_8_10
void FASTCALL call_11_12_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_8_11
#define init_11_12_8_11
void FASTCALL call_11_12_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_8_12
#define init_11_12_8_12
void FASTCALL call_11_12_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_8_13
#define init_11_12_8_13
void FASTCALL call_11_12_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_8_14
#define init_11_12_8_14
void FASTCALL call_11_12_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_8_15
#define init_11_12_8_15
void FASTCALL call_11_12_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_9_0
#define init_11_12_9_0
void FASTCALL call_11_12_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_9_1
#define init_11_12_9_1
void FASTCALL call_11_12_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_9_2
#define init_11_12_9_2
void FASTCALL call_11_12_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_9_3
#define init_11_12_9_3
void FASTCALL call_11_12_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_9_4
#define init_11_12_9_4
void FASTCALL call_11_12_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_9_5
#define init_11_12_9_5
void FASTCALL call_11_12_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_9_6
#define init_11_12_9_6
void FASTCALL call_11_12_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_9_7
#define init_11_12_9_7
void FASTCALL call_11_12_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_9_8
#define init_11_12_9_8
void FASTCALL call_11_12_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_9_9
#define init_11_12_9_9
void FASTCALL call_11_12_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_9_10
#define init_11_12_9_10
void FASTCALL call_11_12_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_9_11
#define init_11_12_9_11
void FASTCALL call_11_12_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_9_12
#define init_11_12_9_12
void FASTCALL call_11_12_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_9_13
#define init_11_12_9_13
void FASTCALL call_11_12_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_9_14
#define init_11_12_9_14
void FASTCALL call_11_12_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_9_15
#define init_11_12_9_15
void FASTCALL call_11_12_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_10_0
#define init_11_12_10_0
void FASTCALL call_11_12_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_10_1
#define init_11_12_10_1
void FASTCALL call_11_12_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_10_2
#define init_11_12_10_2
void FASTCALL call_11_12_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_10_3
#define init_11_12_10_3
void FASTCALL call_11_12_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_10_4
#define init_11_12_10_4
void FASTCALL call_11_12_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_10_5
#define init_11_12_10_5
void FASTCALL call_11_12_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_10_6
#define init_11_12_10_6
void FASTCALL call_11_12_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_10_7
#define init_11_12_10_7
void FASTCALL call_11_12_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_10_8
#define init_11_12_10_8
void FASTCALL call_11_12_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_10_9
#define init_11_12_10_9
void FASTCALL call_11_12_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_10_10
#define init_11_12_10_10
void FASTCALL call_11_12_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_10_11
#define init_11_12_10_11
void FASTCALL call_11_12_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_10_12
#define init_11_12_10_12
void FASTCALL call_11_12_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_10_13
#define init_11_12_10_13
void FASTCALL call_11_12_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_10_14
#define init_11_12_10_14
void FASTCALL call_11_12_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_10_15
#define init_11_12_10_15
void FASTCALL call_11_12_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_11_0
#define init_11_12_11_0
void FASTCALL call_11_12_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_11_1
#define init_11_12_11_1
void FASTCALL call_11_12_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_11_2
#define init_11_12_11_2
void FASTCALL call_11_12_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_11_3
#define init_11_12_11_3
void FASTCALL call_11_12_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_11_4
#define init_11_12_11_4
void FASTCALL call_11_12_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_11_5
#define init_11_12_11_5
void FASTCALL call_11_12_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_11_6
#define init_11_12_11_6
void FASTCALL call_11_12_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_11_7
#define init_11_12_11_7
void FASTCALL call_11_12_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_11_8
#define init_11_12_11_8
void FASTCALL call_11_12_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_11_9
#define init_11_12_11_9
void FASTCALL call_11_12_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_11_10
#define init_11_12_11_10
void FASTCALL call_11_12_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_11_11
#define init_11_12_11_11
void FASTCALL call_11_12_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_11_12
#define init_11_12_11_12
void FASTCALL call_11_12_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_11_13
#define init_11_12_11_13
void FASTCALL call_11_12_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_11_14
#define init_11_12_11_14
void FASTCALL call_11_12_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_11_15
#define init_11_12_11_15
void FASTCALL call_11_12_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_12_0
#define init_11_12_12_0
void FASTCALL call_11_12_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_12_1
#define init_11_12_12_1
void FASTCALL call_11_12_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_12_2
#define init_11_12_12_2
void FASTCALL call_11_12_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_12_3
#define init_11_12_12_3
void FASTCALL call_11_12_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_12_4
#define init_11_12_12_4
void FASTCALL call_11_12_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_12_5
#define init_11_12_12_5
void FASTCALL call_11_12_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_12_6
#define init_11_12_12_6
void FASTCALL call_11_12_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_12_7
#define init_11_12_12_7
void FASTCALL call_11_12_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_12_8
#define init_11_12_12_8
void FASTCALL call_11_12_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_12_9
#define init_11_12_12_9
void FASTCALL call_11_12_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_12_10
#define init_11_12_12_10
void FASTCALL call_11_12_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_12_11
#define init_11_12_12_11
void FASTCALL call_11_12_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_12_12
#define init_11_12_12_12
void FASTCALL call_11_12_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_12_13
#define init_11_12_12_13
void FASTCALL call_11_12_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_12_14
#define init_11_12_12_14
void FASTCALL call_11_12_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_12_15
#define init_11_12_12_15
void FASTCALL call_11_12_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_13_0
#define init_11_12_13_0
void FASTCALL call_11_12_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_13_1
#define init_11_12_13_1
void FASTCALL call_11_12_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_13_2
#define init_11_12_13_2
void FASTCALL call_11_12_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_13_3
#define init_11_12_13_3
void FASTCALL call_11_12_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_13_4
#define init_11_12_13_4
void FASTCALL call_11_12_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_13_5
#define init_11_12_13_5
void FASTCALL call_11_12_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_13_6
#define init_11_12_13_6
void FASTCALL call_11_12_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_13_7
#define init_11_12_13_7
void FASTCALL call_11_12_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_13_8
#define init_11_12_13_8
void FASTCALL call_11_12_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_13_9
#define init_11_12_13_9
void FASTCALL call_11_12_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_13_10
#define init_11_12_13_10
void FASTCALL call_11_12_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_13_11
#define init_11_12_13_11
void FASTCALL call_11_12_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_13_12
#define init_11_12_13_12
void FASTCALL call_11_12_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_13_13
#define init_11_12_13_13
void FASTCALL call_11_12_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_13_14
#define init_11_12_13_14
void FASTCALL call_11_12_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_13_15
#define init_11_12_13_15
void FASTCALL call_11_12_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_14_0
#define init_11_12_14_0
void FASTCALL call_11_12_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_14_1
#define init_11_12_14_1
void FASTCALL call_11_12_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_14_2
#define init_11_12_14_2
void FASTCALL call_11_12_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_14_3
#define init_11_12_14_3
void FASTCALL call_11_12_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_14_4
#define init_11_12_14_4
void FASTCALL call_11_12_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_14_5
#define init_11_12_14_5
void FASTCALL call_11_12_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_14_6
#define init_11_12_14_6
void FASTCALL call_11_12_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_14_7
#define init_11_12_14_7
void FASTCALL call_11_12_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_14_8
#define init_11_12_14_8
void FASTCALL call_11_12_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_14_9
#define init_11_12_14_9
void FASTCALL call_11_12_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_14_10
#define init_11_12_14_10
void FASTCALL call_11_12_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_14_11
#define init_11_12_14_11
void FASTCALL call_11_12_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_14_12
#define init_11_12_14_12
void FASTCALL call_11_12_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_14_13
#define init_11_12_14_13
void FASTCALL call_11_12_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_14_14
#define init_11_12_14_14
void FASTCALL call_11_12_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_14_15
#define init_11_12_14_15
void FASTCALL call_11_12_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_15_0
#define init_11_12_15_0
void FASTCALL call_11_12_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_15_1
#define init_11_12_15_1
void FASTCALL call_11_12_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_15_2
#define init_11_12_15_2
void FASTCALL call_11_12_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_15_3
#define init_11_12_15_3
void FASTCALL call_11_12_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_15_4
#define init_11_12_15_4
void FASTCALL call_11_12_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_15_5
#define init_11_12_15_5
void FASTCALL call_11_12_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_15_6
#define init_11_12_15_6
void FASTCALL call_11_12_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_15_7
#define init_11_12_15_7
void FASTCALL call_11_12_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_15_8
#define init_11_12_15_8
void FASTCALL call_11_12_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_15_9
#define init_11_12_15_9
void FASTCALL call_11_12_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_15_10
#define init_11_12_15_10
void FASTCALL call_11_12_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_15_11
#define init_11_12_15_11
void FASTCALL call_11_12_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_15_12
#define init_11_12_15_12
void FASTCALL call_11_12_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_15_13
#define init_11_12_15_13
void FASTCALL call_11_12_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_15_14
#define init_11_12_15_14
void FASTCALL call_11_12_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_12_15_15
#define init_11_12_15_15
void FASTCALL call_11_12_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_0_0
#define init_11_13_0_0
void FASTCALL call_11_13_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_0_1
#define init_11_13_0_1
void FASTCALL call_11_13_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_0_2
#define init_11_13_0_2
void FASTCALL call_11_13_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_0_3
#define init_11_13_0_3
void FASTCALL call_11_13_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_0_4
#define init_11_13_0_4
void FASTCALL call_11_13_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_0_5
#define init_11_13_0_5
void FASTCALL call_11_13_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_0_6
#define init_11_13_0_6
void FASTCALL call_11_13_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_0_7
#define init_11_13_0_7
void FASTCALL call_11_13_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_0_8
#define init_11_13_0_8
void FASTCALL call_11_13_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_0_9
#define init_11_13_0_9
void FASTCALL call_11_13_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_0_10
#define init_11_13_0_10
void FASTCALL call_11_13_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_0_11
#define init_11_13_0_11
void FASTCALL call_11_13_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_0_12
#define init_11_13_0_12
void FASTCALL call_11_13_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_0_13
#define init_11_13_0_13
void FASTCALL call_11_13_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_0_14
#define init_11_13_0_14
void FASTCALL call_11_13_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_0_15
#define init_11_13_0_15
void FASTCALL call_11_13_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_1_0
#define init_11_13_1_0
void FASTCALL call_11_13_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_1_1
#define init_11_13_1_1
void FASTCALL call_11_13_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_1_2
#define init_11_13_1_2
void FASTCALL call_11_13_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_1_3
#define init_11_13_1_3
void FASTCALL call_11_13_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_1_4
#define init_11_13_1_4
void FASTCALL call_11_13_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_1_5
#define init_11_13_1_5
void FASTCALL call_11_13_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_1_6
#define init_11_13_1_6
void FASTCALL call_11_13_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_1_7
#define init_11_13_1_7
void FASTCALL call_11_13_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_1_8
#define init_11_13_1_8
void FASTCALL call_11_13_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_1_9
#define init_11_13_1_9
void FASTCALL call_11_13_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_1_10
#define init_11_13_1_10
void FASTCALL call_11_13_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_1_11
#define init_11_13_1_11
void FASTCALL call_11_13_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_1_12
#define init_11_13_1_12
void FASTCALL call_11_13_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_1_13
#define init_11_13_1_13
void FASTCALL call_11_13_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_1_14
#define init_11_13_1_14
void FASTCALL call_11_13_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_1_15
#define init_11_13_1_15
void FASTCALL call_11_13_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_2_0
#define init_11_13_2_0
void FASTCALL call_11_13_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_2_1
#define init_11_13_2_1
void FASTCALL call_11_13_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_2_2
#define init_11_13_2_2
void FASTCALL call_11_13_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_2_3
#define init_11_13_2_3
void FASTCALL call_11_13_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_2_4
#define init_11_13_2_4
void FASTCALL call_11_13_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_2_5
#define init_11_13_2_5
void FASTCALL call_11_13_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_2_6
#define init_11_13_2_6
void FASTCALL call_11_13_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_2_7
#define init_11_13_2_7
void FASTCALL call_11_13_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_2_8
#define init_11_13_2_8
void FASTCALL call_11_13_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_2_9
#define init_11_13_2_9
void FASTCALL call_11_13_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_2_10
#define init_11_13_2_10
void FASTCALL call_11_13_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_2_11
#define init_11_13_2_11
void FASTCALL call_11_13_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_2_12
#define init_11_13_2_12
void FASTCALL call_11_13_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_2_13
#define init_11_13_2_13
void FASTCALL call_11_13_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_2_14
#define init_11_13_2_14
void FASTCALL call_11_13_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_2_15
#define init_11_13_2_15
void FASTCALL call_11_13_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_3_0
#define init_11_13_3_0
void FASTCALL call_11_13_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_3_1
#define init_11_13_3_1
void FASTCALL call_11_13_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_3_2
#define init_11_13_3_2
void FASTCALL call_11_13_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_3_3
#define init_11_13_3_3
void FASTCALL call_11_13_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_3_4
#define init_11_13_3_4
void FASTCALL call_11_13_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_3_5
#define init_11_13_3_5
void FASTCALL call_11_13_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_3_6
#define init_11_13_3_6
void FASTCALL call_11_13_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_3_7
#define init_11_13_3_7
void FASTCALL call_11_13_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_3_8
#define init_11_13_3_8
void FASTCALL call_11_13_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_3_9
#define init_11_13_3_9
void FASTCALL call_11_13_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_3_10
#define init_11_13_3_10
void FASTCALL call_11_13_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_3_11
#define init_11_13_3_11
void FASTCALL call_11_13_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_3_12
#define init_11_13_3_12
void FASTCALL call_11_13_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_3_13
#define init_11_13_3_13
void FASTCALL call_11_13_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_3_14
#define init_11_13_3_14
void FASTCALL call_11_13_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_3_15
#define init_11_13_3_15
void FASTCALL call_11_13_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_4_0
#define init_11_13_4_0
void FASTCALL call_11_13_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_4_1
#define init_11_13_4_1
void FASTCALL call_11_13_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_4_2
#define init_11_13_4_2
void FASTCALL call_11_13_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_4_3
#define init_11_13_4_3
void FASTCALL call_11_13_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_4_4
#define init_11_13_4_4
void FASTCALL call_11_13_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_4_5
#define init_11_13_4_5
void FASTCALL call_11_13_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_4_6
#define init_11_13_4_6
void FASTCALL call_11_13_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_4_7
#define init_11_13_4_7
void FASTCALL call_11_13_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_4_8
#define init_11_13_4_8
void FASTCALL call_11_13_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_4_9
#define init_11_13_4_9
void FASTCALL call_11_13_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_4_10
#define init_11_13_4_10
void FASTCALL call_11_13_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_4_11
#define init_11_13_4_11
void FASTCALL call_11_13_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_4_12
#define init_11_13_4_12
void FASTCALL call_11_13_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_4_13
#define init_11_13_4_13
void FASTCALL call_11_13_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_4_14
#define init_11_13_4_14
void FASTCALL call_11_13_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_4_15
#define init_11_13_4_15
void FASTCALL call_11_13_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_5_0
#define init_11_13_5_0
void FASTCALL call_11_13_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_5_1
#define init_11_13_5_1
void FASTCALL call_11_13_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_5_2
#define init_11_13_5_2
void FASTCALL call_11_13_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_5_3
#define init_11_13_5_3
void FASTCALL call_11_13_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_5_4
#define init_11_13_5_4
void FASTCALL call_11_13_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_5_5
#define init_11_13_5_5
void FASTCALL call_11_13_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_5_6
#define init_11_13_5_6
void FASTCALL call_11_13_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_5_7
#define init_11_13_5_7
void FASTCALL call_11_13_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_5_8
#define init_11_13_5_8
void FASTCALL call_11_13_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_5_9
#define init_11_13_5_9
void FASTCALL call_11_13_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_5_10
#define init_11_13_5_10
void FASTCALL call_11_13_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_5_11
#define init_11_13_5_11
void FASTCALL call_11_13_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_5_12
#define init_11_13_5_12
void FASTCALL call_11_13_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_5_13
#define init_11_13_5_13
void FASTCALL call_11_13_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_5_14
#define init_11_13_5_14
void FASTCALL call_11_13_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_5_15
#define init_11_13_5_15
void FASTCALL call_11_13_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_6_0
#define init_11_13_6_0
void FASTCALL call_11_13_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_6_1
#define init_11_13_6_1
void FASTCALL call_11_13_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_6_2
#define init_11_13_6_2
void FASTCALL call_11_13_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_6_3
#define init_11_13_6_3
void FASTCALL call_11_13_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_6_4
#define init_11_13_6_4
void FASTCALL call_11_13_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_6_5
#define init_11_13_6_5
void FASTCALL call_11_13_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_6_6
#define init_11_13_6_6
void FASTCALL call_11_13_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_6_7
#define init_11_13_6_7
void FASTCALL call_11_13_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_6_8
#define init_11_13_6_8
void FASTCALL call_11_13_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_6_9
#define init_11_13_6_9
void FASTCALL call_11_13_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_6_10
#define init_11_13_6_10
void FASTCALL call_11_13_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_6_11
#define init_11_13_6_11
void FASTCALL call_11_13_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_6_12
#define init_11_13_6_12
void FASTCALL call_11_13_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_6_13
#define init_11_13_6_13
void FASTCALL call_11_13_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_6_14
#define init_11_13_6_14
void FASTCALL call_11_13_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_6_15
#define init_11_13_6_15
void FASTCALL call_11_13_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_7_0
#define init_11_13_7_0
void FASTCALL call_11_13_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_7_1
#define init_11_13_7_1
void FASTCALL call_11_13_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_7_2
#define init_11_13_7_2
void FASTCALL call_11_13_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_7_3
#define init_11_13_7_3
void FASTCALL call_11_13_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_7_4
#define init_11_13_7_4
void FASTCALL call_11_13_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_7_5
#define init_11_13_7_5
void FASTCALL call_11_13_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_7_6
#define init_11_13_7_6
void FASTCALL call_11_13_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_7_7
#define init_11_13_7_7
void FASTCALL call_11_13_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_7_8
#define init_11_13_7_8
void FASTCALL call_11_13_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_7_9
#define init_11_13_7_9
void FASTCALL call_11_13_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_7_10
#define init_11_13_7_10
void FASTCALL call_11_13_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_7_11
#define init_11_13_7_11
void FASTCALL call_11_13_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_7_12
#define init_11_13_7_12
void FASTCALL call_11_13_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_7_13
#define init_11_13_7_13
void FASTCALL call_11_13_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_7_14
#define init_11_13_7_14
void FASTCALL call_11_13_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_7_15
#define init_11_13_7_15
void FASTCALL call_11_13_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_8_0
#define init_11_13_8_0
void FASTCALL call_11_13_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_8_1
#define init_11_13_8_1
void FASTCALL call_11_13_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_8_2
#define init_11_13_8_2
void FASTCALL call_11_13_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_8_3
#define init_11_13_8_3
void FASTCALL call_11_13_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_8_4
#define init_11_13_8_4
void FASTCALL call_11_13_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_8_5
#define init_11_13_8_5
void FASTCALL call_11_13_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_8_6
#define init_11_13_8_6
void FASTCALL call_11_13_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_8_7
#define init_11_13_8_7
void FASTCALL call_11_13_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_8_8
#define init_11_13_8_8
void FASTCALL call_11_13_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_8_9
#define init_11_13_8_9
void FASTCALL call_11_13_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_8_10
#define init_11_13_8_10
void FASTCALL call_11_13_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_8_11
#define init_11_13_8_11
void FASTCALL call_11_13_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_8_12
#define init_11_13_8_12
void FASTCALL call_11_13_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_8_13
#define init_11_13_8_13
void FASTCALL call_11_13_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_8_14
#define init_11_13_8_14
void FASTCALL call_11_13_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_8_15
#define init_11_13_8_15
void FASTCALL call_11_13_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_9_0
#define init_11_13_9_0
void FASTCALL call_11_13_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_9_1
#define init_11_13_9_1
void FASTCALL call_11_13_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_9_2
#define init_11_13_9_2
void FASTCALL call_11_13_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_9_3
#define init_11_13_9_3
void FASTCALL call_11_13_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_9_4
#define init_11_13_9_4
void FASTCALL call_11_13_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_9_5
#define init_11_13_9_5
void FASTCALL call_11_13_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_9_6
#define init_11_13_9_6
void FASTCALL call_11_13_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_9_7
#define init_11_13_9_7
void FASTCALL call_11_13_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_9_8
#define init_11_13_9_8
void FASTCALL call_11_13_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_9_9
#define init_11_13_9_9
void FASTCALL call_11_13_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_9_10
#define init_11_13_9_10
void FASTCALL call_11_13_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_9_11
#define init_11_13_9_11
void FASTCALL call_11_13_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_9_12
#define init_11_13_9_12
void FASTCALL call_11_13_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_9_13
#define init_11_13_9_13
void FASTCALL call_11_13_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_9_14
#define init_11_13_9_14
void FASTCALL call_11_13_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_9_15
#define init_11_13_9_15
void FASTCALL call_11_13_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_10_0
#define init_11_13_10_0
void FASTCALL call_11_13_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_10_1
#define init_11_13_10_1
void FASTCALL call_11_13_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_10_2
#define init_11_13_10_2
void FASTCALL call_11_13_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_10_3
#define init_11_13_10_3
void FASTCALL call_11_13_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_10_4
#define init_11_13_10_4
void FASTCALL call_11_13_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_10_5
#define init_11_13_10_5
void FASTCALL call_11_13_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_10_6
#define init_11_13_10_6
void FASTCALL call_11_13_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_10_7
#define init_11_13_10_7
void FASTCALL call_11_13_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_10_8
#define init_11_13_10_8
void FASTCALL call_11_13_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_10_9
#define init_11_13_10_9
void FASTCALL call_11_13_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_10_10
#define init_11_13_10_10
void FASTCALL call_11_13_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_10_11
#define init_11_13_10_11
void FASTCALL call_11_13_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_10_12
#define init_11_13_10_12
void FASTCALL call_11_13_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_10_13
#define init_11_13_10_13
void FASTCALL call_11_13_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_10_14
#define init_11_13_10_14
void FASTCALL call_11_13_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_10_15
#define init_11_13_10_15
void FASTCALL call_11_13_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_11_0
#define init_11_13_11_0
void FASTCALL call_11_13_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_11_1
#define init_11_13_11_1
void FASTCALL call_11_13_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_11_2
#define init_11_13_11_2
void FASTCALL call_11_13_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_11_3
#define init_11_13_11_3
void FASTCALL call_11_13_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_11_4
#define init_11_13_11_4
void FASTCALL call_11_13_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_11_5
#define init_11_13_11_5
void FASTCALL call_11_13_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_11_6
#define init_11_13_11_6
void FASTCALL call_11_13_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_11_7
#define init_11_13_11_7
void FASTCALL call_11_13_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_11_8
#define init_11_13_11_8
void FASTCALL call_11_13_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_11_9
#define init_11_13_11_9
void FASTCALL call_11_13_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_11_10
#define init_11_13_11_10
void FASTCALL call_11_13_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_11_11
#define init_11_13_11_11
void FASTCALL call_11_13_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_11_12
#define init_11_13_11_12
void FASTCALL call_11_13_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_11_13
#define init_11_13_11_13
void FASTCALL call_11_13_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_11_14
#define init_11_13_11_14
void FASTCALL call_11_13_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_11_15
#define init_11_13_11_15
void FASTCALL call_11_13_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_12_0
#define init_11_13_12_0
void FASTCALL call_11_13_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_12_1
#define init_11_13_12_1
void FASTCALL call_11_13_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_12_2
#define init_11_13_12_2
void FASTCALL call_11_13_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_12_3
#define init_11_13_12_3
void FASTCALL call_11_13_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_12_4
#define init_11_13_12_4
void FASTCALL call_11_13_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_12_5
#define init_11_13_12_5
void FASTCALL call_11_13_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_12_6
#define init_11_13_12_6
void FASTCALL call_11_13_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_12_7
#define init_11_13_12_7
void FASTCALL call_11_13_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_12_8
#define init_11_13_12_8
void FASTCALL call_11_13_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_12_9
#define init_11_13_12_9
void FASTCALL call_11_13_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_12_10
#define init_11_13_12_10
void FASTCALL call_11_13_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_12_11
#define init_11_13_12_11
void FASTCALL call_11_13_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_12_12
#define init_11_13_12_12
void FASTCALL call_11_13_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_12_13
#define init_11_13_12_13
void FASTCALL call_11_13_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_12_14
#define init_11_13_12_14
void FASTCALL call_11_13_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_12_15
#define init_11_13_12_15
void FASTCALL call_11_13_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_13_0
#define init_11_13_13_0
void FASTCALL call_11_13_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_13_1
#define init_11_13_13_1
void FASTCALL call_11_13_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_13_2
#define init_11_13_13_2
void FASTCALL call_11_13_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_13_3
#define init_11_13_13_3
void FASTCALL call_11_13_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_13_4
#define init_11_13_13_4
void FASTCALL call_11_13_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_13_5
#define init_11_13_13_5
void FASTCALL call_11_13_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_13_6
#define init_11_13_13_6
void FASTCALL call_11_13_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_13_7
#define init_11_13_13_7
void FASTCALL call_11_13_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_13_8
#define init_11_13_13_8
void FASTCALL call_11_13_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_13_9
#define init_11_13_13_9
void FASTCALL call_11_13_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_13_10
#define init_11_13_13_10
void FASTCALL call_11_13_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_13_11
#define init_11_13_13_11
void FASTCALL call_11_13_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_13_12
#define init_11_13_13_12
void FASTCALL call_11_13_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_13_13
#define init_11_13_13_13
void FASTCALL call_11_13_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_13_14
#define init_11_13_13_14
void FASTCALL call_11_13_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_13_15
#define init_11_13_13_15
void FASTCALL call_11_13_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_14_0
#define init_11_13_14_0
void FASTCALL call_11_13_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_14_1
#define init_11_13_14_1
void FASTCALL call_11_13_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_14_2
#define init_11_13_14_2
void FASTCALL call_11_13_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_14_3
#define init_11_13_14_3
void FASTCALL call_11_13_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_14_4
#define init_11_13_14_4
void FASTCALL call_11_13_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_14_5
#define init_11_13_14_5
void FASTCALL call_11_13_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_14_6
#define init_11_13_14_6
void FASTCALL call_11_13_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_14_7
#define init_11_13_14_7
void FASTCALL call_11_13_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_14_8
#define init_11_13_14_8
void FASTCALL call_11_13_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_14_9
#define init_11_13_14_9
void FASTCALL call_11_13_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_14_10
#define init_11_13_14_10
void FASTCALL call_11_13_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_14_11
#define init_11_13_14_11
void FASTCALL call_11_13_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_14_12
#define init_11_13_14_12
void FASTCALL call_11_13_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_14_13
#define init_11_13_14_13
void FASTCALL call_11_13_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_14_14
#define init_11_13_14_14
void FASTCALL call_11_13_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_14_15
#define init_11_13_14_15
void FASTCALL call_11_13_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_15_0
#define init_11_13_15_0
void FASTCALL call_11_13_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_15_1
#define init_11_13_15_1
void FASTCALL call_11_13_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_15_2
#define init_11_13_15_2
void FASTCALL call_11_13_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_15_3
#define init_11_13_15_3
void FASTCALL call_11_13_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_15_4
#define init_11_13_15_4
void FASTCALL call_11_13_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_15_5
#define init_11_13_15_5
void FASTCALL call_11_13_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_15_6
#define init_11_13_15_6
void FASTCALL call_11_13_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_15_7
#define init_11_13_15_7
void FASTCALL call_11_13_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_15_8
#define init_11_13_15_8
void FASTCALL call_11_13_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_15_9
#define init_11_13_15_9
void FASTCALL call_11_13_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_15_10
#define init_11_13_15_10
void FASTCALL call_11_13_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_15_11
#define init_11_13_15_11
void FASTCALL call_11_13_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_15_12
#define init_11_13_15_12
void FASTCALL call_11_13_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_15_13
#define init_11_13_15_13
void FASTCALL call_11_13_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_15_14
#define init_11_13_15_14
void FASTCALL call_11_13_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_13_15_15
#define init_11_13_15_15
void FASTCALL call_11_13_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_0_0
#define init_11_14_0_0
void FASTCALL call_11_14_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_0_1
#define init_11_14_0_1
void FASTCALL call_11_14_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_0_2
#define init_11_14_0_2
void FASTCALL call_11_14_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_0_3
#define init_11_14_0_3
void FASTCALL call_11_14_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_0_4
#define init_11_14_0_4
void FASTCALL call_11_14_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_0_5
#define init_11_14_0_5
void FASTCALL call_11_14_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_0_6
#define init_11_14_0_6
void FASTCALL call_11_14_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_0_7
#define init_11_14_0_7
void FASTCALL call_11_14_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_0_8
#define init_11_14_0_8
void FASTCALL call_11_14_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_0_9
#define init_11_14_0_9
void FASTCALL call_11_14_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_0_10
#define init_11_14_0_10
void FASTCALL call_11_14_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_0_11
#define init_11_14_0_11
void FASTCALL call_11_14_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_0_12
#define init_11_14_0_12
void FASTCALL call_11_14_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_0_13
#define init_11_14_0_13
void FASTCALL call_11_14_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_0_14
#define init_11_14_0_14
void FASTCALL call_11_14_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_0_15
#define init_11_14_0_15
void FASTCALL call_11_14_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_1_0
#define init_11_14_1_0
void FASTCALL call_11_14_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_1_1
#define init_11_14_1_1
void FASTCALL call_11_14_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_1_2
#define init_11_14_1_2
void FASTCALL call_11_14_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_1_3
#define init_11_14_1_3
void FASTCALL call_11_14_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_1_4
#define init_11_14_1_4
void FASTCALL call_11_14_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_1_5
#define init_11_14_1_5
void FASTCALL call_11_14_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_1_6
#define init_11_14_1_6
void FASTCALL call_11_14_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_1_7
#define init_11_14_1_7
void FASTCALL call_11_14_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_1_8
#define init_11_14_1_8
void FASTCALL call_11_14_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_1_9
#define init_11_14_1_9
void FASTCALL call_11_14_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_1_10
#define init_11_14_1_10
void FASTCALL call_11_14_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_1_11
#define init_11_14_1_11
void FASTCALL call_11_14_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_1_12
#define init_11_14_1_12
void FASTCALL call_11_14_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_1_13
#define init_11_14_1_13
void FASTCALL call_11_14_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_1_14
#define init_11_14_1_14
void FASTCALL call_11_14_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_1_15
#define init_11_14_1_15
void FASTCALL call_11_14_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_2_0
#define init_11_14_2_0
void FASTCALL call_11_14_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_2_1
#define init_11_14_2_1
void FASTCALL call_11_14_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_2_2
#define init_11_14_2_2
void FASTCALL call_11_14_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_2_3
#define init_11_14_2_3
void FASTCALL call_11_14_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_2_4
#define init_11_14_2_4
void FASTCALL call_11_14_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_2_5
#define init_11_14_2_5
void FASTCALL call_11_14_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_2_6
#define init_11_14_2_6
void FASTCALL call_11_14_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_2_7
#define init_11_14_2_7
void FASTCALL call_11_14_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_2_8
#define init_11_14_2_8
void FASTCALL call_11_14_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_2_9
#define init_11_14_2_9
void FASTCALL call_11_14_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_2_10
#define init_11_14_2_10
void FASTCALL call_11_14_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_2_11
#define init_11_14_2_11
void FASTCALL call_11_14_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_2_12
#define init_11_14_2_12
void FASTCALL call_11_14_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_2_13
#define init_11_14_2_13
void FASTCALL call_11_14_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_2_14
#define init_11_14_2_14
void FASTCALL call_11_14_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_2_15
#define init_11_14_2_15
void FASTCALL call_11_14_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_3_0
#define init_11_14_3_0
void FASTCALL call_11_14_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_3_1
#define init_11_14_3_1
void FASTCALL call_11_14_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_3_2
#define init_11_14_3_2
void FASTCALL call_11_14_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_3_3
#define init_11_14_3_3
void FASTCALL call_11_14_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_3_4
#define init_11_14_3_4
void FASTCALL call_11_14_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_3_5
#define init_11_14_3_5
void FASTCALL call_11_14_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_3_6
#define init_11_14_3_6
void FASTCALL call_11_14_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_3_7
#define init_11_14_3_7
void FASTCALL call_11_14_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_3_8
#define init_11_14_3_8
void FASTCALL call_11_14_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_3_9
#define init_11_14_3_9
void FASTCALL call_11_14_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_3_10
#define init_11_14_3_10
void FASTCALL call_11_14_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_3_11
#define init_11_14_3_11
void FASTCALL call_11_14_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_3_12
#define init_11_14_3_12
void FASTCALL call_11_14_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_3_13
#define init_11_14_3_13
void FASTCALL call_11_14_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_3_14
#define init_11_14_3_14
void FASTCALL call_11_14_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_3_15
#define init_11_14_3_15
void FASTCALL call_11_14_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_4_0
#define init_11_14_4_0
void FASTCALL call_11_14_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_4_1
#define init_11_14_4_1
void FASTCALL call_11_14_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_4_2
#define init_11_14_4_2
void FASTCALL call_11_14_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_4_3
#define init_11_14_4_3
void FASTCALL call_11_14_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_4_4
#define init_11_14_4_4
void FASTCALL call_11_14_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_4_5
#define init_11_14_4_5
void FASTCALL call_11_14_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_4_6
#define init_11_14_4_6
void FASTCALL call_11_14_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_4_7
#define init_11_14_4_7
void FASTCALL call_11_14_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_4_8
#define init_11_14_4_8
void FASTCALL call_11_14_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_4_9
#define init_11_14_4_9
void FASTCALL call_11_14_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_4_10
#define init_11_14_4_10
void FASTCALL call_11_14_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_4_11
#define init_11_14_4_11
void FASTCALL call_11_14_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_4_12
#define init_11_14_4_12
void FASTCALL call_11_14_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_4_13
#define init_11_14_4_13
void FASTCALL call_11_14_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_4_14
#define init_11_14_4_14
void FASTCALL call_11_14_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_4_15
#define init_11_14_4_15
void FASTCALL call_11_14_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_5_0
#define init_11_14_5_0
void FASTCALL call_11_14_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_5_1
#define init_11_14_5_1
void FASTCALL call_11_14_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_5_2
#define init_11_14_5_2
void FASTCALL call_11_14_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_5_3
#define init_11_14_5_3
void FASTCALL call_11_14_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_5_4
#define init_11_14_5_4
void FASTCALL call_11_14_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_5_5
#define init_11_14_5_5
void FASTCALL call_11_14_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_5_6
#define init_11_14_5_6
void FASTCALL call_11_14_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_5_7
#define init_11_14_5_7
void FASTCALL call_11_14_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_5_8
#define init_11_14_5_8
void FASTCALL call_11_14_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_5_9
#define init_11_14_5_9
void FASTCALL call_11_14_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_5_10
#define init_11_14_5_10
void FASTCALL call_11_14_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_5_11
#define init_11_14_5_11
void FASTCALL call_11_14_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_5_12
#define init_11_14_5_12
void FASTCALL call_11_14_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_5_13
#define init_11_14_5_13
void FASTCALL call_11_14_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_5_14
#define init_11_14_5_14
void FASTCALL call_11_14_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_5_15
#define init_11_14_5_15
void FASTCALL call_11_14_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_6_0
#define init_11_14_6_0
void FASTCALL call_11_14_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_6_1
#define init_11_14_6_1
void FASTCALL call_11_14_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_6_2
#define init_11_14_6_2
void FASTCALL call_11_14_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_6_3
#define init_11_14_6_3
void FASTCALL call_11_14_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_6_4
#define init_11_14_6_4
void FASTCALL call_11_14_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_6_5
#define init_11_14_6_5
void FASTCALL call_11_14_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_6_6
#define init_11_14_6_6
void FASTCALL call_11_14_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_6_7
#define init_11_14_6_7
void FASTCALL call_11_14_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_6_8
#define init_11_14_6_8
void FASTCALL call_11_14_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_6_9
#define init_11_14_6_9
void FASTCALL call_11_14_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_6_10
#define init_11_14_6_10
void FASTCALL call_11_14_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_6_11
#define init_11_14_6_11
void FASTCALL call_11_14_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_6_12
#define init_11_14_6_12
void FASTCALL call_11_14_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_6_13
#define init_11_14_6_13
void FASTCALL call_11_14_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_6_14
#define init_11_14_6_14
void FASTCALL call_11_14_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_6_15
#define init_11_14_6_15
void FASTCALL call_11_14_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_7_0
#define init_11_14_7_0
void FASTCALL call_11_14_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_7_1
#define init_11_14_7_1
void FASTCALL call_11_14_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_7_2
#define init_11_14_7_2
void FASTCALL call_11_14_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_7_3
#define init_11_14_7_3
void FASTCALL call_11_14_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_7_4
#define init_11_14_7_4
void FASTCALL call_11_14_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_7_5
#define init_11_14_7_5
void FASTCALL call_11_14_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_7_6
#define init_11_14_7_6
void FASTCALL call_11_14_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_7_7
#define init_11_14_7_7
void FASTCALL call_11_14_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_7_8
#define init_11_14_7_8
void FASTCALL call_11_14_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_7_9
#define init_11_14_7_9
void FASTCALL call_11_14_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_7_10
#define init_11_14_7_10
void FASTCALL call_11_14_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_7_11
#define init_11_14_7_11
void FASTCALL call_11_14_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_7_12
#define init_11_14_7_12
void FASTCALL call_11_14_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_7_13
#define init_11_14_7_13
void FASTCALL call_11_14_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_7_14
#define init_11_14_7_14
void FASTCALL call_11_14_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_7_15
#define init_11_14_7_15
void FASTCALL call_11_14_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_8_0
#define init_11_14_8_0
void FASTCALL call_11_14_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_8_1
#define init_11_14_8_1
void FASTCALL call_11_14_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_8_2
#define init_11_14_8_2
void FASTCALL call_11_14_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_8_3
#define init_11_14_8_3
void FASTCALL call_11_14_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_8_4
#define init_11_14_8_4
void FASTCALL call_11_14_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_8_5
#define init_11_14_8_5
void FASTCALL call_11_14_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_8_6
#define init_11_14_8_6
void FASTCALL call_11_14_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_8_7
#define init_11_14_8_7
void FASTCALL call_11_14_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_8_8
#define init_11_14_8_8
void FASTCALL call_11_14_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_8_9
#define init_11_14_8_9
void FASTCALL call_11_14_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_8_10
#define init_11_14_8_10
void FASTCALL call_11_14_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_8_11
#define init_11_14_8_11
void FASTCALL call_11_14_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_8_12
#define init_11_14_8_12
void FASTCALL call_11_14_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_8_13
#define init_11_14_8_13
void FASTCALL call_11_14_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_8_14
#define init_11_14_8_14
void FASTCALL call_11_14_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_8_15
#define init_11_14_8_15
void FASTCALL call_11_14_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_9_0
#define init_11_14_9_0
void FASTCALL call_11_14_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_9_1
#define init_11_14_9_1
void FASTCALL call_11_14_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_9_2
#define init_11_14_9_2
void FASTCALL call_11_14_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_9_3
#define init_11_14_9_3
void FASTCALL call_11_14_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_9_4
#define init_11_14_9_4
void FASTCALL call_11_14_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_9_5
#define init_11_14_9_5
void FASTCALL call_11_14_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_9_6
#define init_11_14_9_6
void FASTCALL call_11_14_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_9_7
#define init_11_14_9_7
void FASTCALL call_11_14_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_9_8
#define init_11_14_9_8
void FASTCALL call_11_14_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_9_9
#define init_11_14_9_9
void FASTCALL call_11_14_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_9_10
#define init_11_14_9_10
void FASTCALL call_11_14_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_9_11
#define init_11_14_9_11
void FASTCALL call_11_14_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_9_12
#define init_11_14_9_12
void FASTCALL call_11_14_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_9_13
#define init_11_14_9_13
void FASTCALL call_11_14_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_9_14
#define init_11_14_9_14
void FASTCALL call_11_14_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_9_15
#define init_11_14_9_15
void FASTCALL call_11_14_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_10_0
#define init_11_14_10_0
void FASTCALL call_11_14_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_10_1
#define init_11_14_10_1
void FASTCALL call_11_14_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_10_2
#define init_11_14_10_2
void FASTCALL call_11_14_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_10_3
#define init_11_14_10_3
void FASTCALL call_11_14_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_10_4
#define init_11_14_10_4
void FASTCALL call_11_14_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_10_5
#define init_11_14_10_5
void FASTCALL call_11_14_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_10_6
#define init_11_14_10_6
void FASTCALL call_11_14_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_10_7
#define init_11_14_10_7
void FASTCALL call_11_14_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_10_8
#define init_11_14_10_8
void FASTCALL call_11_14_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_10_9
#define init_11_14_10_9
void FASTCALL call_11_14_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_10_10
#define init_11_14_10_10
void FASTCALL call_11_14_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_10_11
#define init_11_14_10_11
void FASTCALL call_11_14_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_10_12
#define init_11_14_10_12
void FASTCALL call_11_14_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_10_13
#define init_11_14_10_13
void FASTCALL call_11_14_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_10_14
#define init_11_14_10_14
void FASTCALL call_11_14_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_10_15
#define init_11_14_10_15
void FASTCALL call_11_14_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_11_0
#define init_11_14_11_0
void FASTCALL call_11_14_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_11_1
#define init_11_14_11_1
void FASTCALL call_11_14_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_11_2
#define init_11_14_11_2
void FASTCALL call_11_14_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_11_3
#define init_11_14_11_3
void FASTCALL call_11_14_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_11_4
#define init_11_14_11_4
void FASTCALL call_11_14_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_11_5
#define init_11_14_11_5
void FASTCALL call_11_14_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_11_6
#define init_11_14_11_6
void FASTCALL call_11_14_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_11_7
#define init_11_14_11_7
void FASTCALL call_11_14_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_11_8
#define init_11_14_11_8
void FASTCALL call_11_14_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_11_9
#define init_11_14_11_9
void FASTCALL call_11_14_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_11_10
#define init_11_14_11_10
void FASTCALL call_11_14_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_11_11
#define init_11_14_11_11
void FASTCALL call_11_14_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_11_12
#define init_11_14_11_12
void FASTCALL call_11_14_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_11_13
#define init_11_14_11_13
void FASTCALL call_11_14_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_11_14
#define init_11_14_11_14
void FASTCALL call_11_14_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_11_15
#define init_11_14_11_15
void FASTCALL call_11_14_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_12_0
#define init_11_14_12_0
void FASTCALL call_11_14_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_12_1
#define init_11_14_12_1
void FASTCALL call_11_14_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_12_2
#define init_11_14_12_2
void FASTCALL call_11_14_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_12_3
#define init_11_14_12_3
void FASTCALL call_11_14_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_12_4
#define init_11_14_12_4
void FASTCALL call_11_14_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_12_5
#define init_11_14_12_5
void FASTCALL call_11_14_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_12_6
#define init_11_14_12_6
void FASTCALL call_11_14_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_12_7
#define init_11_14_12_7
void FASTCALL call_11_14_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_12_8
#define init_11_14_12_8
void FASTCALL call_11_14_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_12_9
#define init_11_14_12_9
void FASTCALL call_11_14_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_12_10
#define init_11_14_12_10
void FASTCALL call_11_14_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_12_11
#define init_11_14_12_11
void FASTCALL call_11_14_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_12_12
#define init_11_14_12_12
void FASTCALL call_11_14_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_12_13
#define init_11_14_12_13
void FASTCALL call_11_14_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_12_14
#define init_11_14_12_14
void FASTCALL call_11_14_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_12_15
#define init_11_14_12_15
void FASTCALL call_11_14_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_13_0
#define init_11_14_13_0
void FASTCALL call_11_14_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_13_1
#define init_11_14_13_1
void FASTCALL call_11_14_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_13_2
#define init_11_14_13_2
void FASTCALL call_11_14_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_13_3
#define init_11_14_13_3
void FASTCALL call_11_14_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_13_4
#define init_11_14_13_4
void FASTCALL call_11_14_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_13_5
#define init_11_14_13_5
void FASTCALL call_11_14_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_13_6
#define init_11_14_13_6
void FASTCALL call_11_14_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_13_7
#define init_11_14_13_7
void FASTCALL call_11_14_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_13_8
#define init_11_14_13_8
void FASTCALL call_11_14_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_13_9
#define init_11_14_13_9
void FASTCALL call_11_14_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_13_10
#define init_11_14_13_10
void FASTCALL call_11_14_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_13_11
#define init_11_14_13_11
void FASTCALL call_11_14_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_13_12
#define init_11_14_13_12
void FASTCALL call_11_14_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_13_13
#define init_11_14_13_13
void FASTCALL call_11_14_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_13_14
#define init_11_14_13_14
void FASTCALL call_11_14_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_13_15
#define init_11_14_13_15
void FASTCALL call_11_14_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_14_0
#define init_11_14_14_0
void FASTCALL call_11_14_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_14_1
#define init_11_14_14_1
void FASTCALL call_11_14_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_14_2
#define init_11_14_14_2
void FASTCALL call_11_14_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_14_3
#define init_11_14_14_3
void FASTCALL call_11_14_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_14_4
#define init_11_14_14_4
void FASTCALL call_11_14_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_14_5
#define init_11_14_14_5
void FASTCALL call_11_14_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_14_6
#define init_11_14_14_6
void FASTCALL call_11_14_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_14_7
#define init_11_14_14_7
void FASTCALL call_11_14_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_14_8
#define init_11_14_14_8
void FASTCALL call_11_14_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_14_9
#define init_11_14_14_9
void FASTCALL call_11_14_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_14_10
#define init_11_14_14_10
void FASTCALL call_11_14_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_14_11
#define init_11_14_14_11
void FASTCALL call_11_14_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_14_12
#define init_11_14_14_12
void FASTCALL call_11_14_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_14_13
#define init_11_14_14_13
void FASTCALL call_11_14_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_14_14
#define init_11_14_14_14
void FASTCALL call_11_14_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_14_15
#define init_11_14_14_15
void FASTCALL call_11_14_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_15_0
#define init_11_14_15_0
void FASTCALL call_11_14_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_15_1
#define init_11_14_15_1
void FASTCALL call_11_14_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_15_2
#define init_11_14_15_2
void FASTCALL call_11_14_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_15_3
#define init_11_14_15_3
void FASTCALL call_11_14_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_15_4
#define init_11_14_15_4
void FASTCALL call_11_14_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_15_5
#define init_11_14_15_5
void FASTCALL call_11_14_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_15_6
#define init_11_14_15_6
void FASTCALL call_11_14_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_15_7
#define init_11_14_15_7
void FASTCALL call_11_14_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_15_8
#define init_11_14_15_8
void FASTCALL call_11_14_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_15_9
#define init_11_14_15_9
void FASTCALL call_11_14_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_15_10
#define init_11_14_15_10
void FASTCALL call_11_14_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_15_11
#define init_11_14_15_11
void FASTCALL call_11_14_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_15_12
#define init_11_14_15_12
void FASTCALL call_11_14_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_15_13
#define init_11_14_15_13
void FASTCALL call_11_14_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_15_14
#define init_11_14_15_14
void FASTCALL call_11_14_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_14_15_15
#define init_11_14_15_15
void FASTCALL call_11_14_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_0_0
#define init_11_15_0_0
void FASTCALL call_11_15_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_0_1
#define init_11_15_0_1
void FASTCALL call_11_15_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_0_2
#define init_11_15_0_2
void FASTCALL call_11_15_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_0_3
#define init_11_15_0_3
void FASTCALL call_11_15_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_0_4
#define init_11_15_0_4
void FASTCALL call_11_15_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_0_5
#define init_11_15_0_5
void FASTCALL call_11_15_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_0_6
#define init_11_15_0_6
void FASTCALL call_11_15_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_0_7
#define init_11_15_0_7
void FASTCALL call_11_15_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_0_8
#define init_11_15_0_8
void FASTCALL call_11_15_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_0_9
#define init_11_15_0_9
void FASTCALL call_11_15_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_0_10
#define init_11_15_0_10
void FASTCALL call_11_15_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_0_11
#define init_11_15_0_11
void FASTCALL call_11_15_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_0_12
#define init_11_15_0_12
void FASTCALL call_11_15_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_0_13
#define init_11_15_0_13
void FASTCALL call_11_15_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_0_14
#define init_11_15_0_14
void FASTCALL call_11_15_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_0_15
#define init_11_15_0_15
void FASTCALL call_11_15_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_1_0
#define init_11_15_1_0
void FASTCALL call_11_15_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_1_1
#define init_11_15_1_1
void FASTCALL call_11_15_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_1_2
#define init_11_15_1_2
void FASTCALL call_11_15_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_1_3
#define init_11_15_1_3
void FASTCALL call_11_15_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_1_4
#define init_11_15_1_4
void FASTCALL call_11_15_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_1_5
#define init_11_15_1_5
void FASTCALL call_11_15_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_1_6
#define init_11_15_1_6
void FASTCALL call_11_15_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_1_7
#define init_11_15_1_7
void FASTCALL call_11_15_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_1_8
#define init_11_15_1_8
void FASTCALL call_11_15_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_1_9
#define init_11_15_1_9
void FASTCALL call_11_15_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_1_10
#define init_11_15_1_10
void FASTCALL call_11_15_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_1_11
#define init_11_15_1_11
void FASTCALL call_11_15_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_1_12
#define init_11_15_1_12
void FASTCALL call_11_15_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_1_13
#define init_11_15_1_13
void FASTCALL call_11_15_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_1_14
#define init_11_15_1_14
void FASTCALL call_11_15_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_1_15
#define init_11_15_1_15
void FASTCALL call_11_15_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_2_0
#define init_11_15_2_0
void FASTCALL call_11_15_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_2_1
#define init_11_15_2_1
void FASTCALL call_11_15_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_2_2
#define init_11_15_2_2
void FASTCALL call_11_15_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_2_3
#define init_11_15_2_3
void FASTCALL call_11_15_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_2_4
#define init_11_15_2_4
void FASTCALL call_11_15_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_2_5
#define init_11_15_2_5
void FASTCALL call_11_15_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_2_6
#define init_11_15_2_6
void FASTCALL call_11_15_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_2_7
#define init_11_15_2_7
void FASTCALL call_11_15_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_2_8
#define init_11_15_2_8
void FASTCALL call_11_15_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_2_9
#define init_11_15_2_9
void FASTCALL call_11_15_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_2_10
#define init_11_15_2_10
void FASTCALL call_11_15_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_2_11
#define init_11_15_2_11
void FASTCALL call_11_15_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_2_12
#define init_11_15_2_12
void FASTCALL call_11_15_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_2_13
#define init_11_15_2_13
void FASTCALL call_11_15_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_2_14
#define init_11_15_2_14
void FASTCALL call_11_15_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_2_15
#define init_11_15_2_15
void FASTCALL call_11_15_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_3_0
#define init_11_15_3_0
void FASTCALL call_11_15_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_3_1
#define init_11_15_3_1
void FASTCALL call_11_15_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_3_2
#define init_11_15_3_2
void FASTCALL call_11_15_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_3_3
#define init_11_15_3_3
void FASTCALL call_11_15_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_3_4
#define init_11_15_3_4
void FASTCALL call_11_15_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_3_5
#define init_11_15_3_5
void FASTCALL call_11_15_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_3_6
#define init_11_15_3_6
void FASTCALL call_11_15_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_3_7
#define init_11_15_3_7
void FASTCALL call_11_15_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_3_8
#define init_11_15_3_8
void FASTCALL call_11_15_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_3_9
#define init_11_15_3_9
void FASTCALL call_11_15_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_3_10
#define init_11_15_3_10
void FASTCALL call_11_15_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_3_11
#define init_11_15_3_11
void FASTCALL call_11_15_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_3_12
#define init_11_15_3_12
void FASTCALL call_11_15_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_3_13
#define init_11_15_3_13
void FASTCALL call_11_15_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_3_14
#define init_11_15_3_14
void FASTCALL call_11_15_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_3_15
#define init_11_15_3_15
void FASTCALL call_11_15_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_4_0
#define init_11_15_4_0
void FASTCALL call_11_15_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_4_1
#define init_11_15_4_1
void FASTCALL call_11_15_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_4_2
#define init_11_15_4_2
void FASTCALL call_11_15_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_4_3
#define init_11_15_4_3
void FASTCALL call_11_15_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_4_4
#define init_11_15_4_4
void FASTCALL call_11_15_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_4_5
#define init_11_15_4_5
void FASTCALL call_11_15_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_4_6
#define init_11_15_4_6
void FASTCALL call_11_15_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_4_7
#define init_11_15_4_7
void FASTCALL call_11_15_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_4_8
#define init_11_15_4_8
void FASTCALL call_11_15_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_4_9
#define init_11_15_4_9
void FASTCALL call_11_15_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_4_10
#define init_11_15_4_10
void FASTCALL call_11_15_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_4_11
#define init_11_15_4_11
void FASTCALL call_11_15_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_4_12
#define init_11_15_4_12
void FASTCALL call_11_15_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_4_13
#define init_11_15_4_13
void FASTCALL call_11_15_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_4_14
#define init_11_15_4_14
void FASTCALL call_11_15_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_4_15
#define init_11_15_4_15
void FASTCALL call_11_15_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_5_0
#define init_11_15_5_0
void FASTCALL call_11_15_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_5_1
#define init_11_15_5_1
void FASTCALL call_11_15_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_5_2
#define init_11_15_5_2
void FASTCALL call_11_15_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_5_3
#define init_11_15_5_3
void FASTCALL call_11_15_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_5_4
#define init_11_15_5_4
void FASTCALL call_11_15_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_5_5
#define init_11_15_5_5
void FASTCALL call_11_15_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_5_6
#define init_11_15_5_6
void FASTCALL call_11_15_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_5_7
#define init_11_15_5_7
void FASTCALL call_11_15_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_5_8
#define init_11_15_5_8
void FASTCALL call_11_15_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_5_9
#define init_11_15_5_9
void FASTCALL call_11_15_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_5_10
#define init_11_15_5_10
void FASTCALL call_11_15_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_5_11
#define init_11_15_5_11
void FASTCALL call_11_15_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_5_12
#define init_11_15_5_12
void FASTCALL call_11_15_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_5_13
#define init_11_15_5_13
void FASTCALL call_11_15_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_5_14
#define init_11_15_5_14
void FASTCALL call_11_15_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_5_15
#define init_11_15_5_15
void FASTCALL call_11_15_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_6_0
#define init_11_15_6_0
void FASTCALL call_11_15_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_6_1
#define init_11_15_6_1
void FASTCALL call_11_15_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_6_2
#define init_11_15_6_2
void FASTCALL call_11_15_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_6_3
#define init_11_15_6_3
void FASTCALL call_11_15_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_6_4
#define init_11_15_6_4
void FASTCALL call_11_15_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_6_5
#define init_11_15_6_5
void FASTCALL call_11_15_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_6_6
#define init_11_15_6_6
void FASTCALL call_11_15_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_6_7
#define init_11_15_6_7
void FASTCALL call_11_15_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_6_8
#define init_11_15_6_8
void FASTCALL call_11_15_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_6_9
#define init_11_15_6_9
void FASTCALL call_11_15_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_6_10
#define init_11_15_6_10
void FASTCALL call_11_15_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_6_11
#define init_11_15_6_11
void FASTCALL call_11_15_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_6_12
#define init_11_15_6_12
void FASTCALL call_11_15_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_6_13
#define init_11_15_6_13
void FASTCALL call_11_15_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_6_14
#define init_11_15_6_14
void FASTCALL call_11_15_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_6_15
#define init_11_15_6_15
void FASTCALL call_11_15_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_7_0
#define init_11_15_7_0
void FASTCALL call_11_15_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_7_1
#define init_11_15_7_1
void FASTCALL call_11_15_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_7_2
#define init_11_15_7_2
void FASTCALL call_11_15_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_7_3
#define init_11_15_7_3
void FASTCALL call_11_15_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_7_4
#define init_11_15_7_4
void FASTCALL call_11_15_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_7_5
#define init_11_15_7_5
void FASTCALL call_11_15_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_7_6
#define init_11_15_7_6
void FASTCALL call_11_15_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_7_7
#define init_11_15_7_7
void FASTCALL call_11_15_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_7_8
#define init_11_15_7_8
void FASTCALL call_11_15_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_7_9
#define init_11_15_7_9
void FASTCALL call_11_15_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_7_10
#define init_11_15_7_10
void FASTCALL call_11_15_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_7_11
#define init_11_15_7_11
void FASTCALL call_11_15_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_7_12
#define init_11_15_7_12
void FASTCALL call_11_15_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_7_13
#define init_11_15_7_13
void FASTCALL call_11_15_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_7_14
#define init_11_15_7_14
void FASTCALL call_11_15_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_7_15
#define init_11_15_7_15
void FASTCALL call_11_15_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_8_0
#define init_11_15_8_0
void FASTCALL call_11_15_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_8_1
#define init_11_15_8_1
void FASTCALL call_11_15_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_8_2
#define init_11_15_8_2
void FASTCALL call_11_15_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_8_3
#define init_11_15_8_3
void FASTCALL call_11_15_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_8_4
#define init_11_15_8_4
void FASTCALL call_11_15_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_8_5
#define init_11_15_8_5
void FASTCALL call_11_15_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_8_6
#define init_11_15_8_6
void FASTCALL call_11_15_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_8_7
#define init_11_15_8_7
void FASTCALL call_11_15_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_8_8
#define init_11_15_8_8
void FASTCALL call_11_15_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_8_9
#define init_11_15_8_9
void FASTCALL call_11_15_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_8_10
#define init_11_15_8_10
void FASTCALL call_11_15_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_8_11
#define init_11_15_8_11
void FASTCALL call_11_15_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_8_12
#define init_11_15_8_12
void FASTCALL call_11_15_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_8_13
#define init_11_15_8_13
void FASTCALL call_11_15_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_8_14
#define init_11_15_8_14
void FASTCALL call_11_15_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_8_15
#define init_11_15_8_15
void FASTCALL call_11_15_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_9_0
#define init_11_15_9_0
void FASTCALL call_11_15_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_9_1
#define init_11_15_9_1
void FASTCALL call_11_15_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_9_2
#define init_11_15_9_2
void FASTCALL call_11_15_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_9_3
#define init_11_15_9_3
void FASTCALL call_11_15_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_9_4
#define init_11_15_9_4
void FASTCALL call_11_15_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_9_5
#define init_11_15_9_5
void FASTCALL call_11_15_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_9_6
#define init_11_15_9_6
void FASTCALL call_11_15_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_9_7
#define init_11_15_9_7
void FASTCALL call_11_15_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_9_8
#define init_11_15_9_8
void FASTCALL call_11_15_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_9_9
#define init_11_15_9_9
void FASTCALL call_11_15_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_9_10
#define init_11_15_9_10
void FASTCALL call_11_15_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_9_11
#define init_11_15_9_11
void FASTCALL call_11_15_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_9_12
#define init_11_15_9_12
void FASTCALL call_11_15_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_9_13
#define init_11_15_9_13
void FASTCALL call_11_15_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_9_14
#define init_11_15_9_14
void FASTCALL call_11_15_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_9_15
#define init_11_15_9_15
void FASTCALL call_11_15_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_10_0
#define init_11_15_10_0
void FASTCALL call_11_15_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_10_1
#define init_11_15_10_1
void FASTCALL call_11_15_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_10_2
#define init_11_15_10_2
void FASTCALL call_11_15_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_10_3
#define init_11_15_10_3
void FASTCALL call_11_15_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_10_4
#define init_11_15_10_4
void FASTCALL call_11_15_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_10_5
#define init_11_15_10_5
void FASTCALL call_11_15_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_10_6
#define init_11_15_10_6
void FASTCALL call_11_15_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_10_7
#define init_11_15_10_7
void FASTCALL call_11_15_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_10_8
#define init_11_15_10_8
void FASTCALL call_11_15_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_10_9
#define init_11_15_10_9
void FASTCALL call_11_15_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_10_10
#define init_11_15_10_10
void FASTCALL call_11_15_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_10_11
#define init_11_15_10_11
void FASTCALL call_11_15_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_10_12
#define init_11_15_10_12
void FASTCALL call_11_15_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_10_13
#define init_11_15_10_13
void FASTCALL call_11_15_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_10_14
#define init_11_15_10_14
void FASTCALL call_11_15_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_10_15
#define init_11_15_10_15
void FASTCALL call_11_15_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_11_0
#define init_11_15_11_0
void FASTCALL call_11_15_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_11_1
#define init_11_15_11_1
void FASTCALL call_11_15_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_11_2
#define init_11_15_11_2
void FASTCALL call_11_15_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_11_3
#define init_11_15_11_3
void FASTCALL call_11_15_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_11_4
#define init_11_15_11_4
void FASTCALL call_11_15_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_11_5
#define init_11_15_11_5
void FASTCALL call_11_15_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_11_6
#define init_11_15_11_6
void FASTCALL call_11_15_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_11_7
#define init_11_15_11_7
void FASTCALL call_11_15_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_11_8
#define init_11_15_11_8
void FASTCALL call_11_15_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_11_9
#define init_11_15_11_9
void FASTCALL call_11_15_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_11_10
#define init_11_15_11_10
void FASTCALL call_11_15_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_11_11
#define init_11_15_11_11
void FASTCALL call_11_15_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_11_12
#define init_11_15_11_12
void FASTCALL call_11_15_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_11_13
#define init_11_15_11_13
void FASTCALL call_11_15_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_11_14
#define init_11_15_11_14
void FASTCALL call_11_15_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_11_15
#define init_11_15_11_15
void FASTCALL call_11_15_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_12_0
#define init_11_15_12_0
void FASTCALL call_11_15_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_12_1
#define init_11_15_12_1
void FASTCALL call_11_15_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_12_2
#define init_11_15_12_2
void FASTCALL call_11_15_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_12_3
#define init_11_15_12_3
void FASTCALL call_11_15_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_12_4
#define init_11_15_12_4
void FASTCALL call_11_15_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_12_5
#define init_11_15_12_5
void FASTCALL call_11_15_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_12_6
#define init_11_15_12_6
void FASTCALL call_11_15_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_12_7
#define init_11_15_12_7
void FASTCALL call_11_15_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_12_8
#define init_11_15_12_8
void FASTCALL call_11_15_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_12_9
#define init_11_15_12_9
void FASTCALL call_11_15_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_12_10
#define init_11_15_12_10
void FASTCALL call_11_15_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_12_11
#define init_11_15_12_11
void FASTCALL call_11_15_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_12_12
#define init_11_15_12_12
void FASTCALL call_11_15_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_12_13
#define init_11_15_12_13
void FASTCALL call_11_15_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_12_14
#define init_11_15_12_14
void FASTCALL call_11_15_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_12_15
#define init_11_15_12_15
void FASTCALL call_11_15_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_13_0
#define init_11_15_13_0
void FASTCALL call_11_15_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_13_1
#define init_11_15_13_1
void FASTCALL call_11_15_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_13_2
#define init_11_15_13_2
void FASTCALL call_11_15_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_13_3
#define init_11_15_13_3
void FASTCALL call_11_15_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_13_4
#define init_11_15_13_4
void FASTCALL call_11_15_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_13_5
#define init_11_15_13_5
void FASTCALL call_11_15_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_13_6
#define init_11_15_13_6
void FASTCALL call_11_15_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_13_7
#define init_11_15_13_7
void FASTCALL call_11_15_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_13_8
#define init_11_15_13_8
void FASTCALL call_11_15_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_13_9
#define init_11_15_13_9
void FASTCALL call_11_15_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_13_10
#define init_11_15_13_10
void FASTCALL call_11_15_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_13_11
#define init_11_15_13_11
void FASTCALL call_11_15_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_13_12
#define init_11_15_13_12
void FASTCALL call_11_15_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_13_13
#define init_11_15_13_13
void FASTCALL call_11_15_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_13_14
#define init_11_15_13_14
void FASTCALL call_11_15_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_13_15
#define init_11_15_13_15
void FASTCALL call_11_15_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_14_0
#define init_11_15_14_0
void FASTCALL call_11_15_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_14_1
#define init_11_15_14_1
void FASTCALL call_11_15_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_14_2
#define init_11_15_14_2
void FASTCALL call_11_15_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_14_3
#define init_11_15_14_3
void FASTCALL call_11_15_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_14_4
#define init_11_15_14_4
void FASTCALL call_11_15_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_14_5
#define init_11_15_14_5
void FASTCALL call_11_15_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_14_6
#define init_11_15_14_6
void FASTCALL call_11_15_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_14_7
#define init_11_15_14_7
void FASTCALL call_11_15_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_14_8
#define init_11_15_14_8
void FASTCALL call_11_15_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_14_9
#define init_11_15_14_9
void FASTCALL call_11_15_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_14_10
#define init_11_15_14_10
void FASTCALL call_11_15_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_14_11
#define init_11_15_14_11
void FASTCALL call_11_15_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_14_12
#define init_11_15_14_12
void FASTCALL call_11_15_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_14_13
#define init_11_15_14_13
void FASTCALL call_11_15_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_14_14
#define init_11_15_14_14
void FASTCALL call_11_15_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_14_15
#define init_11_15_14_15
void FASTCALL call_11_15_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_15_0
#define init_11_15_15_0
void FASTCALL call_11_15_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_15_1
#define init_11_15_15_1
void FASTCALL call_11_15_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_15_2
#define init_11_15_15_2
void FASTCALL call_11_15_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_15_3
#define init_11_15_15_3
void FASTCALL call_11_15_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_15_4
#define init_11_15_15_4
void FASTCALL call_11_15_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_15_5
#define init_11_15_15_5
void FASTCALL call_11_15_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_15_6
#define init_11_15_15_6
void FASTCALL call_11_15_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_15_7
#define init_11_15_15_7
void FASTCALL call_11_15_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_15_8
#define init_11_15_15_8
void FASTCALL call_11_15_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_15_9
#define init_11_15_15_9
void FASTCALL call_11_15_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_15_10
#define init_11_15_15_10
void FASTCALL call_11_15_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_15_11
#define init_11_15_15_11
void FASTCALL call_11_15_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_15_12
#define init_11_15_15_12
void FASTCALL call_11_15_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_15_13
#define init_11_15_15_13
void FASTCALL call_11_15_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_15_14
#define init_11_15_15_14
void FASTCALL call_11_15_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_11_15_15_15
#define init_11_15_15_15
void FASTCALL call_11_15_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_0_0
#define init_12_0_0_0
void FASTCALL call_12_0_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_0_1
#define init_12_0_0_1
void FASTCALL call_12_0_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_0_2
#define init_12_0_0_2
void FASTCALL call_12_0_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_0_3
#define init_12_0_0_3
void FASTCALL call_12_0_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_0_4
#define init_12_0_0_4
void FASTCALL call_12_0_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_0_5
#define init_12_0_0_5
void FASTCALL call_12_0_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_0_6
#define init_12_0_0_6
void FASTCALL call_12_0_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_0_7
#define init_12_0_0_7
void FASTCALL call_12_0_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_0_8
#define init_12_0_0_8
void FASTCALL call_12_0_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_0_9
#define init_12_0_0_9
void FASTCALL call_12_0_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_0_10
#define init_12_0_0_10
void FASTCALL call_12_0_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_0_11
#define init_12_0_0_11
void FASTCALL call_12_0_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_0_12
#define init_12_0_0_12
void FASTCALL call_12_0_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_0_13
#define init_12_0_0_13
void FASTCALL call_12_0_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_0_14
#define init_12_0_0_14
void FASTCALL call_12_0_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_0_15
#define init_12_0_0_15
void FASTCALL call_12_0_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_1_0
#define init_12_0_1_0
void FASTCALL call_12_0_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_1_1
#define init_12_0_1_1
void FASTCALL call_12_0_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_1_2
#define init_12_0_1_2
void FASTCALL call_12_0_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_1_3
#define init_12_0_1_3
void FASTCALL call_12_0_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_1_4
#define init_12_0_1_4
void FASTCALL call_12_0_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_1_5
#define init_12_0_1_5
void FASTCALL call_12_0_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_1_6
#define init_12_0_1_6
void FASTCALL call_12_0_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_1_7
#define init_12_0_1_7
void FASTCALL call_12_0_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_1_8
#define init_12_0_1_8
void FASTCALL call_12_0_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_1_9
#define init_12_0_1_9
void FASTCALL call_12_0_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_1_10
#define init_12_0_1_10
void FASTCALL call_12_0_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_1_11
#define init_12_0_1_11
void FASTCALL call_12_0_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_1_12
#define init_12_0_1_12
void FASTCALL call_12_0_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_1_13
#define init_12_0_1_13
void FASTCALL call_12_0_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_1_14
#define init_12_0_1_14
void FASTCALL call_12_0_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_1_15
#define init_12_0_1_15
void FASTCALL call_12_0_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_2_0
#define init_12_0_2_0
void FASTCALL call_12_0_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_2_1
#define init_12_0_2_1
void FASTCALL call_12_0_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_2_2
#define init_12_0_2_2
void FASTCALL call_12_0_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_2_3
#define init_12_0_2_3
void FASTCALL call_12_0_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_2_4
#define init_12_0_2_4
void FASTCALL call_12_0_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_2_5
#define init_12_0_2_5
void FASTCALL call_12_0_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_2_6
#define init_12_0_2_6
void FASTCALL call_12_0_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_2_7
#define init_12_0_2_7
void FASTCALL call_12_0_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_2_8
#define init_12_0_2_8
void FASTCALL call_12_0_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_2_9
#define init_12_0_2_9
void FASTCALL call_12_0_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_2_10
#define init_12_0_2_10
void FASTCALL call_12_0_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_2_11
#define init_12_0_2_11
void FASTCALL call_12_0_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_2_12
#define init_12_0_2_12
void FASTCALL call_12_0_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_2_13
#define init_12_0_2_13
void FASTCALL call_12_0_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_2_14
#define init_12_0_2_14
void FASTCALL call_12_0_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_2_15
#define init_12_0_2_15
void FASTCALL call_12_0_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_3_0
#define init_12_0_3_0
void FASTCALL call_12_0_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_3_1
#define init_12_0_3_1
void FASTCALL call_12_0_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_3_2
#define init_12_0_3_2
void FASTCALL call_12_0_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_3_3
#define init_12_0_3_3
void FASTCALL call_12_0_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_3_4
#define init_12_0_3_4
void FASTCALL call_12_0_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_3_5
#define init_12_0_3_5
void FASTCALL call_12_0_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_3_6
#define init_12_0_3_6
void FASTCALL call_12_0_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_3_7
#define init_12_0_3_7
void FASTCALL call_12_0_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_3_8
#define init_12_0_3_8
void FASTCALL call_12_0_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_3_9
#define init_12_0_3_9
void FASTCALL call_12_0_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_3_10
#define init_12_0_3_10
void FASTCALL call_12_0_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_3_11
#define init_12_0_3_11
void FASTCALL call_12_0_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_3_12
#define init_12_0_3_12
void FASTCALL call_12_0_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_3_13
#define init_12_0_3_13
void FASTCALL call_12_0_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_3_14
#define init_12_0_3_14
void FASTCALL call_12_0_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_3_15
#define init_12_0_3_15
void FASTCALL call_12_0_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_4_0
#define init_12_0_4_0
void FASTCALL call_12_0_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_4_1
#define init_12_0_4_1
void FASTCALL call_12_0_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_4_2
#define init_12_0_4_2
void FASTCALL call_12_0_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_4_3
#define init_12_0_4_3
void FASTCALL call_12_0_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_4_4
#define init_12_0_4_4
void FASTCALL call_12_0_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_4_5
#define init_12_0_4_5
void FASTCALL call_12_0_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_4_6
#define init_12_0_4_6
void FASTCALL call_12_0_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_4_7
#define init_12_0_4_7
void FASTCALL call_12_0_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_4_8
#define init_12_0_4_8
void FASTCALL call_12_0_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_4_9
#define init_12_0_4_9
void FASTCALL call_12_0_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_4_10
#define init_12_0_4_10
void FASTCALL call_12_0_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_4_11
#define init_12_0_4_11
void FASTCALL call_12_0_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_4_12
#define init_12_0_4_12
void FASTCALL call_12_0_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_4_13
#define init_12_0_4_13
void FASTCALL call_12_0_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_4_14
#define init_12_0_4_14
void FASTCALL call_12_0_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_4_15
#define init_12_0_4_15
void FASTCALL call_12_0_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_5_0
#define init_12_0_5_0
void FASTCALL call_12_0_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_5_1
#define init_12_0_5_1
void FASTCALL call_12_0_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_5_2
#define init_12_0_5_2
void FASTCALL call_12_0_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_5_3
#define init_12_0_5_3
void FASTCALL call_12_0_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_5_4
#define init_12_0_5_4
void FASTCALL call_12_0_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_5_5
#define init_12_0_5_5
void FASTCALL call_12_0_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_5_6
#define init_12_0_5_6
void FASTCALL call_12_0_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_5_7
#define init_12_0_5_7
void FASTCALL call_12_0_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_5_8
#define init_12_0_5_8
void FASTCALL call_12_0_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_5_9
#define init_12_0_5_9
void FASTCALL call_12_0_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_5_10
#define init_12_0_5_10
void FASTCALL call_12_0_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_5_11
#define init_12_0_5_11
void FASTCALL call_12_0_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_5_12
#define init_12_0_5_12
void FASTCALL call_12_0_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_5_13
#define init_12_0_5_13
void FASTCALL call_12_0_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_5_14
#define init_12_0_5_14
void FASTCALL call_12_0_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_5_15
#define init_12_0_5_15
void FASTCALL call_12_0_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_6_0
#define init_12_0_6_0
void FASTCALL call_12_0_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_6_1
#define init_12_0_6_1
void FASTCALL call_12_0_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_6_2
#define init_12_0_6_2
void FASTCALL call_12_0_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_6_3
#define init_12_0_6_3
void FASTCALL call_12_0_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_6_4
#define init_12_0_6_4
void FASTCALL call_12_0_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_6_5
#define init_12_0_6_5
void FASTCALL call_12_0_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_6_6
#define init_12_0_6_6
void FASTCALL call_12_0_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_6_7
#define init_12_0_6_7
void FASTCALL call_12_0_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_6_8
#define init_12_0_6_8
void FASTCALL call_12_0_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_6_9
#define init_12_0_6_9
void FASTCALL call_12_0_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_6_10
#define init_12_0_6_10
void FASTCALL call_12_0_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_6_11
#define init_12_0_6_11
void FASTCALL call_12_0_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_6_12
#define init_12_0_6_12
void FASTCALL call_12_0_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_6_13
#define init_12_0_6_13
void FASTCALL call_12_0_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_6_14
#define init_12_0_6_14
void FASTCALL call_12_0_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_6_15
#define init_12_0_6_15
void FASTCALL call_12_0_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_7_0
#define init_12_0_7_0
void FASTCALL call_12_0_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_7_1
#define init_12_0_7_1
void FASTCALL call_12_0_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_7_2
#define init_12_0_7_2
void FASTCALL call_12_0_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_7_3
#define init_12_0_7_3
void FASTCALL call_12_0_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_7_4
#define init_12_0_7_4
void FASTCALL call_12_0_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_7_5
#define init_12_0_7_5
void FASTCALL call_12_0_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_7_6
#define init_12_0_7_6
void FASTCALL call_12_0_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_7_7
#define init_12_0_7_7
void FASTCALL call_12_0_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_7_8
#define init_12_0_7_8
void FASTCALL call_12_0_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_7_9
#define init_12_0_7_9
void FASTCALL call_12_0_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_7_10
#define init_12_0_7_10
void FASTCALL call_12_0_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_7_11
#define init_12_0_7_11
void FASTCALL call_12_0_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_7_12
#define init_12_0_7_12
void FASTCALL call_12_0_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_7_13
#define init_12_0_7_13
void FASTCALL call_12_0_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_7_14
#define init_12_0_7_14
void FASTCALL call_12_0_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_7_15
#define init_12_0_7_15
void FASTCALL call_12_0_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_8_0
#define init_12_0_8_0
void FASTCALL call_12_0_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_8_1
#define init_12_0_8_1
void FASTCALL call_12_0_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_8_2
#define init_12_0_8_2
void FASTCALL call_12_0_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_8_3
#define init_12_0_8_3
void FASTCALL call_12_0_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_8_4
#define init_12_0_8_4
void FASTCALL call_12_0_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_8_5
#define init_12_0_8_5
void FASTCALL call_12_0_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_8_6
#define init_12_0_8_6
void FASTCALL call_12_0_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_8_7
#define init_12_0_8_7
void FASTCALL call_12_0_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_8_8
#define init_12_0_8_8
void FASTCALL call_12_0_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_8_9
#define init_12_0_8_9
void FASTCALL call_12_0_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_8_10
#define init_12_0_8_10
void FASTCALL call_12_0_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_8_11
#define init_12_0_8_11
void FASTCALL call_12_0_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_8_12
#define init_12_0_8_12
void FASTCALL call_12_0_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_8_13
#define init_12_0_8_13
void FASTCALL call_12_0_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_8_14
#define init_12_0_8_14
void FASTCALL call_12_0_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_8_15
#define init_12_0_8_15
void FASTCALL call_12_0_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_9_0
#define init_12_0_9_0
void FASTCALL call_12_0_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_9_1
#define init_12_0_9_1
void FASTCALL call_12_0_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_9_2
#define init_12_0_9_2
void FASTCALL call_12_0_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_9_3
#define init_12_0_9_3
void FASTCALL call_12_0_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_9_4
#define init_12_0_9_4
void FASTCALL call_12_0_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_9_5
#define init_12_0_9_5
void FASTCALL call_12_0_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_9_6
#define init_12_0_9_6
void FASTCALL call_12_0_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_9_7
#define init_12_0_9_7
void FASTCALL call_12_0_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_9_8
#define init_12_0_9_8
void FASTCALL call_12_0_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_9_9
#define init_12_0_9_9
void FASTCALL call_12_0_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_9_10
#define init_12_0_9_10
void FASTCALL call_12_0_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_9_11
#define init_12_0_9_11
void FASTCALL call_12_0_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_9_12
#define init_12_0_9_12
void FASTCALL call_12_0_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_9_13
#define init_12_0_9_13
void FASTCALL call_12_0_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_9_14
#define init_12_0_9_14
void FASTCALL call_12_0_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_9_15
#define init_12_0_9_15
void FASTCALL call_12_0_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_10_0
#define init_12_0_10_0
void FASTCALL call_12_0_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_10_1
#define init_12_0_10_1
void FASTCALL call_12_0_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_10_2
#define init_12_0_10_2
void FASTCALL call_12_0_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_10_3
#define init_12_0_10_3
void FASTCALL call_12_0_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_10_4
#define init_12_0_10_4
void FASTCALL call_12_0_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_10_5
#define init_12_0_10_5
void FASTCALL call_12_0_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_10_6
#define init_12_0_10_6
void FASTCALL call_12_0_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_10_7
#define init_12_0_10_7
void FASTCALL call_12_0_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_10_8
#define init_12_0_10_8
void FASTCALL call_12_0_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_10_9
#define init_12_0_10_9
void FASTCALL call_12_0_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_10_10
#define init_12_0_10_10
void FASTCALL call_12_0_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_10_11
#define init_12_0_10_11
void FASTCALL call_12_0_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_10_12
#define init_12_0_10_12
void FASTCALL call_12_0_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_10_13
#define init_12_0_10_13
void FASTCALL call_12_0_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_10_14
#define init_12_0_10_14
void FASTCALL call_12_0_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_10_15
#define init_12_0_10_15
void FASTCALL call_12_0_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_11_0
#define init_12_0_11_0
void FASTCALL call_12_0_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_11_1
#define init_12_0_11_1
void FASTCALL call_12_0_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_11_2
#define init_12_0_11_2
void FASTCALL call_12_0_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_11_3
#define init_12_0_11_3
void FASTCALL call_12_0_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_11_4
#define init_12_0_11_4
void FASTCALL call_12_0_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_11_5
#define init_12_0_11_5
void FASTCALL call_12_0_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_11_6
#define init_12_0_11_6
void FASTCALL call_12_0_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_11_7
#define init_12_0_11_7
void FASTCALL call_12_0_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_11_8
#define init_12_0_11_8
void FASTCALL call_12_0_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_11_9
#define init_12_0_11_9
void FASTCALL call_12_0_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_11_10
#define init_12_0_11_10
void FASTCALL call_12_0_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_11_11
#define init_12_0_11_11
void FASTCALL call_12_0_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_11_12
#define init_12_0_11_12
void FASTCALL call_12_0_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_11_13
#define init_12_0_11_13
void FASTCALL call_12_0_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_11_14
#define init_12_0_11_14
void FASTCALL call_12_0_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_11_15
#define init_12_0_11_15
void FASTCALL call_12_0_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_12_0
#define init_12_0_12_0
void FASTCALL call_12_0_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_12_1
#define init_12_0_12_1
void FASTCALL call_12_0_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_12_2
#define init_12_0_12_2
void FASTCALL call_12_0_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_12_3
#define init_12_0_12_3
void FASTCALL call_12_0_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_12_4
#define init_12_0_12_4
void FASTCALL call_12_0_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_12_5
#define init_12_0_12_5
void FASTCALL call_12_0_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_12_6
#define init_12_0_12_6
void FASTCALL call_12_0_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_12_7
#define init_12_0_12_7
void FASTCALL call_12_0_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_12_8
#define init_12_0_12_8
void FASTCALL call_12_0_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_12_9
#define init_12_0_12_9
void FASTCALL call_12_0_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_12_10
#define init_12_0_12_10
void FASTCALL call_12_0_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_12_11
#define init_12_0_12_11
void FASTCALL call_12_0_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_12_12
#define init_12_0_12_12
void FASTCALL call_12_0_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_12_13
#define init_12_0_12_13
void FASTCALL call_12_0_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_12_14
#define init_12_0_12_14
void FASTCALL call_12_0_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_12_15
#define init_12_0_12_15
void FASTCALL call_12_0_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_13_0
#define init_12_0_13_0
void FASTCALL call_12_0_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_13_1
#define init_12_0_13_1
void FASTCALL call_12_0_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_13_2
#define init_12_0_13_2
void FASTCALL call_12_0_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_13_3
#define init_12_0_13_3
void FASTCALL call_12_0_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_13_4
#define init_12_0_13_4
void FASTCALL call_12_0_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_13_5
#define init_12_0_13_5
void FASTCALL call_12_0_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_13_6
#define init_12_0_13_6
void FASTCALL call_12_0_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_13_7
#define init_12_0_13_7
void FASTCALL call_12_0_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_13_8
#define init_12_0_13_8
void FASTCALL call_12_0_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_13_9
#define init_12_0_13_9
void FASTCALL call_12_0_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_13_10
#define init_12_0_13_10
void FASTCALL call_12_0_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_13_11
#define init_12_0_13_11
void FASTCALL call_12_0_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_13_12
#define init_12_0_13_12
void FASTCALL call_12_0_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_13_13
#define init_12_0_13_13
void FASTCALL call_12_0_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_13_14
#define init_12_0_13_14
void FASTCALL call_12_0_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_13_15
#define init_12_0_13_15
void FASTCALL call_12_0_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_14_0
#define init_12_0_14_0
void FASTCALL call_12_0_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_14_1
#define init_12_0_14_1
void FASTCALL call_12_0_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_14_2
#define init_12_0_14_2
void FASTCALL call_12_0_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_14_3
#define init_12_0_14_3
void FASTCALL call_12_0_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_14_4
#define init_12_0_14_4
void FASTCALL call_12_0_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_14_5
#define init_12_0_14_5
void FASTCALL call_12_0_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_14_6
#define init_12_0_14_6
void FASTCALL call_12_0_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_14_7
#define init_12_0_14_7
void FASTCALL call_12_0_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_14_8
#define init_12_0_14_8
void FASTCALL call_12_0_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_14_9
#define init_12_0_14_9
void FASTCALL call_12_0_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_14_10
#define init_12_0_14_10
void FASTCALL call_12_0_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_14_11
#define init_12_0_14_11
void FASTCALL call_12_0_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_14_12
#define init_12_0_14_12
void FASTCALL call_12_0_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_14_13
#define init_12_0_14_13
void FASTCALL call_12_0_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_14_14
#define init_12_0_14_14
void FASTCALL call_12_0_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_14_15
#define init_12_0_14_15
void FASTCALL call_12_0_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_15_0
#define init_12_0_15_0
void FASTCALL call_12_0_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_15_1
#define init_12_0_15_1
void FASTCALL call_12_0_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_15_2
#define init_12_0_15_2
void FASTCALL call_12_0_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_15_3
#define init_12_0_15_3
void FASTCALL call_12_0_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_15_4
#define init_12_0_15_4
void FASTCALL call_12_0_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_15_5
#define init_12_0_15_5
void FASTCALL call_12_0_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_15_6
#define init_12_0_15_6
void FASTCALL call_12_0_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_15_7
#define init_12_0_15_7
void FASTCALL call_12_0_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_15_8
#define init_12_0_15_8
void FASTCALL call_12_0_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_15_9
#define init_12_0_15_9
void FASTCALL call_12_0_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_15_10
#define init_12_0_15_10
void FASTCALL call_12_0_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_15_11
#define init_12_0_15_11
void FASTCALL call_12_0_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_15_12
#define init_12_0_15_12
void FASTCALL call_12_0_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_15_13
#define init_12_0_15_13
void FASTCALL call_12_0_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_15_14
#define init_12_0_15_14
void FASTCALL call_12_0_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_0_15_15
#define init_12_0_15_15
void FASTCALL call_12_0_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_0_0
#define init_12_1_0_0
void FASTCALL call_12_1_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_0_1
#define init_12_1_0_1
void FASTCALL call_12_1_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_0_2
#define init_12_1_0_2
void FASTCALL call_12_1_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_0_3
#define init_12_1_0_3
void FASTCALL call_12_1_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_0_4
#define init_12_1_0_4
void FASTCALL call_12_1_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_0_5
#define init_12_1_0_5
void FASTCALL call_12_1_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_0_6
#define init_12_1_0_6
void FASTCALL call_12_1_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_0_7
#define init_12_1_0_7
void FASTCALL call_12_1_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_0_8
#define init_12_1_0_8
void FASTCALL call_12_1_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_0_9
#define init_12_1_0_9
void FASTCALL call_12_1_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_0_10
#define init_12_1_0_10
void FASTCALL call_12_1_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_0_11
#define init_12_1_0_11
void FASTCALL call_12_1_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_0_12
#define init_12_1_0_12
void FASTCALL call_12_1_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_0_13
#define init_12_1_0_13
void FASTCALL call_12_1_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_0_14
#define init_12_1_0_14
void FASTCALL call_12_1_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_0_15
#define init_12_1_0_15
void FASTCALL call_12_1_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_1_0
#define init_12_1_1_0
void FASTCALL call_12_1_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_1_1
#define init_12_1_1_1
void FASTCALL call_12_1_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_1_2
#define init_12_1_1_2
void FASTCALL call_12_1_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_1_3
#define init_12_1_1_3
void FASTCALL call_12_1_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_1_4
#define init_12_1_1_4
void FASTCALL call_12_1_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_1_5
#define init_12_1_1_5
void FASTCALL call_12_1_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_1_6
#define init_12_1_1_6
void FASTCALL call_12_1_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_1_7
#define init_12_1_1_7
void FASTCALL call_12_1_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_1_8
#define init_12_1_1_8
void FASTCALL call_12_1_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_1_9
#define init_12_1_1_9
void FASTCALL call_12_1_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_1_10
#define init_12_1_1_10
void FASTCALL call_12_1_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_1_11
#define init_12_1_1_11
void FASTCALL call_12_1_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_1_12
#define init_12_1_1_12
void FASTCALL call_12_1_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_1_13
#define init_12_1_1_13
void FASTCALL call_12_1_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_1_14
#define init_12_1_1_14
void FASTCALL call_12_1_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_1_15
#define init_12_1_1_15
void FASTCALL call_12_1_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_2_0
#define init_12_1_2_0
void FASTCALL call_12_1_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_2_1
#define init_12_1_2_1
void FASTCALL call_12_1_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_2_2
#define init_12_1_2_2
void FASTCALL call_12_1_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_2_3
#define init_12_1_2_3
void FASTCALL call_12_1_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_2_4
#define init_12_1_2_4
void FASTCALL call_12_1_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_2_5
#define init_12_1_2_5
void FASTCALL call_12_1_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_2_6
#define init_12_1_2_6
void FASTCALL call_12_1_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_2_7
#define init_12_1_2_7
void FASTCALL call_12_1_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_2_8
#define init_12_1_2_8
void FASTCALL call_12_1_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_2_9
#define init_12_1_2_9
void FASTCALL call_12_1_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_2_10
#define init_12_1_2_10
void FASTCALL call_12_1_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_2_11
#define init_12_1_2_11
void FASTCALL call_12_1_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_2_12
#define init_12_1_2_12
void FASTCALL call_12_1_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_2_13
#define init_12_1_2_13
void FASTCALL call_12_1_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_2_14
#define init_12_1_2_14
void FASTCALL call_12_1_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_2_15
#define init_12_1_2_15
void FASTCALL call_12_1_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_3_0
#define init_12_1_3_0
void FASTCALL call_12_1_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_3_1
#define init_12_1_3_1
void FASTCALL call_12_1_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_3_2
#define init_12_1_3_2
void FASTCALL call_12_1_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_3_3
#define init_12_1_3_3
void FASTCALL call_12_1_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_3_4
#define init_12_1_3_4
void FASTCALL call_12_1_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_3_5
#define init_12_1_3_5
void FASTCALL call_12_1_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_3_6
#define init_12_1_3_6
void FASTCALL call_12_1_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_3_7
#define init_12_1_3_7
void FASTCALL call_12_1_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_3_8
#define init_12_1_3_8
void FASTCALL call_12_1_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_3_9
#define init_12_1_3_9
void FASTCALL call_12_1_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_3_10
#define init_12_1_3_10
void FASTCALL call_12_1_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_3_11
#define init_12_1_3_11
void FASTCALL call_12_1_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_3_12
#define init_12_1_3_12
void FASTCALL call_12_1_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_3_13
#define init_12_1_3_13
void FASTCALL call_12_1_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_3_14
#define init_12_1_3_14
void FASTCALL call_12_1_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_3_15
#define init_12_1_3_15
void FASTCALL call_12_1_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_4_0
#define init_12_1_4_0
void FASTCALL call_12_1_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_4_1
#define init_12_1_4_1
void FASTCALL call_12_1_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_4_2
#define init_12_1_4_2
void FASTCALL call_12_1_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_4_3
#define init_12_1_4_3
void FASTCALL call_12_1_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_4_4
#define init_12_1_4_4
void FASTCALL call_12_1_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_4_5
#define init_12_1_4_5
void FASTCALL call_12_1_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_4_6
#define init_12_1_4_6
void FASTCALL call_12_1_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_4_7
#define init_12_1_4_7
void FASTCALL call_12_1_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_4_8
#define init_12_1_4_8
void FASTCALL call_12_1_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_4_9
#define init_12_1_4_9
void FASTCALL call_12_1_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_4_10
#define init_12_1_4_10
void FASTCALL call_12_1_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_4_11
#define init_12_1_4_11
void FASTCALL call_12_1_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_4_12
#define init_12_1_4_12
void FASTCALL call_12_1_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_4_13
#define init_12_1_4_13
void FASTCALL call_12_1_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_4_14
#define init_12_1_4_14
void FASTCALL call_12_1_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_4_15
#define init_12_1_4_15
void FASTCALL call_12_1_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_5_0
#define init_12_1_5_0
void FASTCALL call_12_1_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_5_1
#define init_12_1_5_1
void FASTCALL call_12_1_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_5_2
#define init_12_1_5_2
void FASTCALL call_12_1_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_5_3
#define init_12_1_5_3
void FASTCALL call_12_1_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_5_4
#define init_12_1_5_4
void FASTCALL call_12_1_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_5_5
#define init_12_1_5_5
void FASTCALL call_12_1_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_5_6
#define init_12_1_5_6
void FASTCALL call_12_1_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_5_7
#define init_12_1_5_7
void FASTCALL call_12_1_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_5_8
#define init_12_1_5_8
void FASTCALL call_12_1_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_5_9
#define init_12_1_5_9
void FASTCALL call_12_1_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_5_10
#define init_12_1_5_10
void FASTCALL call_12_1_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_5_11
#define init_12_1_5_11
void FASTCALL call_12_1_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_5_12
#define init_12_1_5_12
void FASTCALL call_12_1_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_5_13
#define init_12_1_5_13
void FASTCALL call_12_1_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_5_14
#define init_12_1_5_14
void FASTCALL call_12_1_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_5_15
#define init_12_1_5_15
void FASTCALL call_12_1_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_6_0
#define init_12_1_6_0
void FASTCALL call_12_1_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_6_1
#define init_12_1_6_1
void FASTCALL call_12_1_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_6_2
#define init_12_1_6_2
void FASTCALL call_12_1_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_6_3
#define init_12_1_6_3
void FASTCALL call_12_1_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_6_4
#define init_12_1_6_4
void FASTCALL call_12_1_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_6_5
#define init_12_1_6_5
void FASTCALL call_12_1_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_6_6
#define init_12_1_6_6
void FASTCALL call_12_1_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_6_7
#define init_12_1_6_7
void FASTCALL call_12_1_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_6_8
#define init_12_1_6_8
void FASTCALL call_12_1_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_6_9
#define init_12_1_6_9
void FASTCALL call_12_1_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_6_10
#define init_12_1_6_10
void FASTCALL call_12_1_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_6_11
#define init_12_1_6_11
void FASTCALL call_12_1_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_6_12
#define init_12_1_6_12
void FASTCALL call_12_1_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_6_13
#define init_12_1_6_13
void FASTCALL call_12_1_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_6_14
#define init_12_1_6_14
void FASTCALL call_12_1_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_6_15
#define init_12_1_6_15
void FASTCALL call_12_1_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_7_0
#define init_12_1_7_0
void FASTCALL call_12_1_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_7_1
#define init_12_1_7_1
void FASTCALL call_12_1_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_7_2
#define init_12_1_7_2
void FASTCALL call_12_1_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_7_3
#define init_12_1_7_3
void FASTCALL call_12_1_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_7_4
#define init_12_1_7_4
void FASTCALL call_12_1_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_7_5
#define init_12_1_7_5
void FASTCALL call_12_1_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_7_6
#define init_12_1_7_6
void FASTCALL call_12_1_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_7_7
#define init_12_1_7_7
void FASTCALL call_12_1_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_7_8
#define init_12_1_7_8
void FASTCALL call_12_1_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_7_9
#define init_12_1_7_9
void FASTCALL call_12_1_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_7_10
#define init_12_1_7_10
void FASTCALL call_12_1_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_7_11
#define init_12_1_7_11
void FASTCALL call_12_1_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_7_12
#define init_12_1_7_12
void FASTCALL call_12_1_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_7_13
#define init_12_1_7_13
void FASTCALL call_12_1_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_7_14
#define init_12_1_7_14
void FASTCALL call_12_1_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_7_15
#define init_12_1_7_15
void FASTCALL call_12_1_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_8_0
#define init_12_1_8_0
void FASTCALL call_12_1_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_8_1
#define init_12_1_8_1
void FASTCALL call_12_1_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_8_2
#define init_12_1_8_2
void FASTCALL call_12_1_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_8_3
#define init_12_1_8_3
void FASTCALL call_12_1_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_8_4
#define init_12_1_8_4
void FASTCALL call_12_1_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_8_5
#define init_12_1_8_5
void FASTCALL call_12_1_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_8_6
#define init_12_1_8_6
void FASTCALL call_12_1_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_8_7
#define init_12_1_8_7
void FASTCALL call_12_1_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_8_8
#define init_12_1_8_8
void FASTCALL call_12_1_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_8_9
#define init_12_1_8_9
void FASTCALL call_12_1_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_8_10
#define init_12_1_8_10
void FASTCALL call_12_1_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_8_11
#define init_12_1_8_11
void FASTCALL call_12_1_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_8_12
#define init_12_1_8_12
void FASTCALL call_12_1_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_8_13
#define init_12_1_8_13
void FASTCALL call_12_1_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_8_14
#define init_12_1_8_14
void FASTCALL call_12_1_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_8_15
#define init_12_1_8_15
void FASTCALL call_12_1_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_9_0
#define init_12_1_9_0
void FASTCALL call_12_1_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_9_1
#define init_12_1_9_1
void FASTCALL call_12_1_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_9_2
#define init_12_1_9_2
void FASTCALL call_12_1_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_9_3
#define init_12_1_9_3
void FASTCALL call_12_1_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_9_4
#define init_12_1_9_4
void FASTCALL call_12_1_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_9_5
#define init_12_1_9_5
void FASTCALL call_12_1_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_9_6
#define init_12_1_9_6
void FASTCALL call_12_1_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_9_7
#define init_12_1_9_7
void FASTCALL call_12_1_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_9_8
#define init_12_1_9_8
void FASTCALL call_12_1_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_9_9
#define init_12_1_9_9
void FASTCALL call_12_1_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_9_10
#define init_12_1_9_10
void FASTCALL call_12_1_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_9_11
#define init_12_1_9_11
void FASTCALL call_12_1_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_9_12
#define init_12_1_9_12
void FASTCALL call_12_1_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_9_13
#define init_12_1_9_13
void FASTCALL call_12_1_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_9_14
#define init_12_1_9_14
void FASTCALL call_12_1_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_9_15
#define init_12_1_9_15
void FASTCALL call_12_1_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_10_0
#define init_12_1_10_0
void FASTCALL call_12_1_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_10_1
#define init_12_1_10_1
void FASTCALL call_12_1_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_10_2
#define init_12_1_10_2
void FASTCALL call_12_1_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_10_3
#define init_12_1_10_3
void FASTCALL call_12_1_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_10_4
#define init_12_1_10_4
void FASTCALL call_12_1_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_10_5
#define init_12_1_10_5
void FASTCALL call_12_1_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_10_6
#define init_12_1_10_6
void FASTCALL call_12_1_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_10_7
#define init_12_1_10_7
void FASTCALL call_12_1_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_10_8
#define init_12_1_10_8
void FASTCALL call_12_1_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_10_9
#define init_12_1_10_9
void FASTCALL call_12_1_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_10_10
#define init_12_1_10_10
void FASTCALL call_12_1_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_10_11
#define init_12_1_10_11
void FASTCALL call_12_1_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_10_12
#define init_12_1_10_12
void FASTCALL call_12_1_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_10_13
#define init_12_1_10_13
void FASTCALL call_12_1_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_10_14
#define init_12_1_10_14
void FASTCALL call_12_1_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_10_15
#define init_12_1_10_15
void FASTCALL call_12_1_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_11_0
#define init_12_1_11_0
void FASTCALL call_12_1_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_11_1
#define init_12_1_11_1
void FASTCALL call_12_1_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_11_2
#define init_12_1_11_2
void FASTCALL call_12_1_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_11_3
#define init_12_1_11_3
void FASTCALL call_12_1_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_11_4
#define init_12_1_11_4
void FASTCALL call_12_1_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_11_5
#define init_12_1_11_5
void FASTCALL call_12_1_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_11_6
#define init_12_1_11_6
void FASTCALL call_12_1_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_11_7
#define init_12_1_11_7
void FASTCALL call_12_1_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_11_8
#define init_12_1_11_8
void FASTCALL call_12_1_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_11_9
#define init_12_1_11_9
void FASTCALL call_12_1_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_11_10
#define init_12_1_11_10
void FASTCALL call_12_1_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_11_11
#define init_12_1_11_11
void FASTCALL call_12_1_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_11_12
#define init_12_1_11_12
void FASTCALL call_12_1_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_11_13
#define init_12_1_11_13
void FASTCALL call_12_1_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_11_14
#define init_12_1_11_14
void FASTCALL call_12_1_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_11_15
#define init_12_1_11_15
void FASTCALL call_12_1_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_12_0
#define init_12_1_12_0
void FASTCALL call_12_1_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_12_1
#define init_12_1_12_1
void FASTCALL call_12_1_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_12_2
#define init_12_1_12_2
void FASTCALL call_12_1_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_12_3
#define init_12_1_12_3
void FASTCALL call_12_1_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_12_4
#define init_12_1_12_4
void FASTCALL call_12_1_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_12_5
#define init_12_1_12_5
void FASTCALL call_12_1_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_12_6
#define init_12_1_12_6
void FASTCALL call_12_1_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_12_7
#define init_12_1_12_7
void FASTCALL call_12_1_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_12_8
#define init_12_1_12_8
void FASTCALL call_12_1_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_12_9
#define init_12_1_12_9
void FASTCALL call_12_1_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_12_10
#define init_12_1_12_10
void FASTCALL call_12_1_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_12_11
#define init_12_1_12_11
void FASTCALL call_12_1_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_12_12
#define init_12_1_12_12
void FASTCALL call_12_1_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_12_13
#define init_12_1_12_13
void FASTCALL call_12_1_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_12_14
#define init_12_1_12_14
void FASTCALL call_12_1_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_12_15
#define init_12_1_12_15
void FASTCALL call_12_1_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_13_0
#define init_12_1_13_0
void FASTCALL call_12_1_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_13_1
#define init_12_1_13_1
void FASTCALL call_12_1_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_13_2
#define init_12_1_13_2
void FASTCALL call_12_1_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_13_3
#define init_12_1_13_3
void FASTCALL call_12_1_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_13_4
#define init_12_1_13_4
void FASTCALL call_12_1_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_13_5
#define init_12_1_13_5
void FASTCALL call_12_1_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_13_6
#define init_12_1_13_6
void FASTCALL call_12_1_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_13_7
#define init_12_1_13_7
void FASTCALL call_12_1_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_13_8
#define init_12_1_13_8
void FASTCALL call_12_1_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_13_9
#define init_12_1_13_9
void FASTCALL call_12_1_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_13_10
#define init_12_1_13_10
void FASTCALL call_12_1_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_13_11
#define init_12_1_13_11
void FASTCALL call_12_1_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_13_12
#define init_12_1_13_12
void FASTCALL call_12_1_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_13_13
#define init_12_1_13_13
void FASTCALL call_12_1_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_13_14
#define init_12_1_13_14
void FASTCALL call_12_1_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_13_15
#define init_12_1_13_15
void FASTCALL call_12_1_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_14_0
#define init_12_1_14_0
void FASTCALL call_12_1_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_14_1
#define init_12_1_14_1
void FASTCALL call_12_1_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_14_2
#define init_12_1_14_2
void FASTCALL call_12_1_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_14_3
#define init_12_1_14_3
void FASTCALL call_12_1_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_14_4
#define init_12_1_14_4
void FASTCALL call_12_1_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_14_5
#define init_12_1_14_5
void FASTCALL call_12_1_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_14_6
#define init_12_1_14_6
void FASTCALL call_12_1_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_14_7
#define init_12_1_14_7
void FASTCALL call_12_1_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_14_8
#define init_12_1_14_8
void FASTCALL call_12_1_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_14_9
#define init_12_1_14_9
void FASTCALL call_12_1_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_14_10
#define init_12_1_14_10
void FASTCALL call_12_1_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_14_11
#define init_12_1_14_11
void FASTCALL call_12_1_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_14_12
#define init_12_1_14_12
void FASTCALL call_12_1_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_14_13
#define init_12_1_14_13
void FASTCALL call_12_1_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_14_14
#define init_12_1_14_14
void FASTCALL call_12_1_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_14_15
#define init_12_1_14_15
void FASTCALL call_12_1_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_15_0
#define init_12_1_15_0
void FASTCALL call_12_1_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_15_1
#define init_12_1_15_1
void FASTCALL call_12_1_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_15_2
#define init_12_1_15_2
void FASTCALL call_12_1_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_15_3
#define init_12_1_15_3
void FASTCALL call_12_1_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_15_4
#define init_12_1_15_4
void FASTCALL call_12_1_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_15_5
#define init_12_1_15_5
void FASTCALL call_12_1_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_15_6
#define init_12_1_15_6
void FASTCALL call_12_1_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_15_7
#define init_12_1_15_7
void FASTCALL call_12_1_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_15_8
#define init_12_1_15_8
void FASTCALL call_12_1_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_15_9
#define init_12_1_15_9
void FASTCALL call_12_1_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_15_10
#define init_12_1_15_10
void FASTCALL call_12_1_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_15_11
#define init_12_1_15_11
void FASTCALL call_12_1_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_15_12
#define init_12_1_15_12
void FASTCALL call_12_1_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_15_13
#define init_12_1_15_13
void FASTCALL call_12_1_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_15_14
#define init_12_1_15_14
void FASTCALL call_12_1_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_1_15_15
#define init_12_1_15_15
void FASTCALL call_12_1_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_0_0
#define init_12_2_0_0
void FASTCALL call_12_2_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_0_1
#define init_12_2_0_1
void FASTCALL call_12_2_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_0_2
#define init_12_2_0_2
void FASTCALL call_12_2_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_0_3
#define init_12_2_0_3
void FASTCALL call_12_2_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_0_4
#define init_12_2_0_4
void FASTCALL call_12_2_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_0_5
#define init_12_2_0_5
void FASTCALL call_12_2_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_0_6
#define init_12_2_0_6
void FASTCALL call_12_2_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_0_7
#define init_12_2_0_7
void FASTCALL call_12_2_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_0_8
#define init_12_2_0_8
void FASTCALL call_12_2_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_0_9
#define init_12_2_0_9
void FASTCALL call_12_2_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_0_10
#define init_12_2_0_10
void FASTCALL call_12_2_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_0_11
#define init_12_2_0_11
void FASTCALL call_12_2_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_0_12
#define init_12_2_0_12
void FASTCALL call_12_2_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_0_13
#define init_12_2_0_13
void FASTCALL call_12_2_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_0_14
#define init_12_2_0_14
void FASTCALL call_12_2_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_0_15
#define init_12_2_0_15
void FASTCALL call_12_2_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_1_0
#define init_12_2_1_0
void FASTCALL call_12_2_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_1_1
#define init_12_2_1_1
void FASTCALL call_12_2_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_1_2
#define init_12_2_1_2
void FASTCALL call_12_2_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_1_3
#define init_12_2_1_3
void FASTCALL call_12_2_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_1_4
#define init_12_2_1_4
void FASTCALL call_12_2_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_1_5
#define init_12_2_1_5
void FASTCALL call_12_2_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_1_6
#define init_12_2_1_6
void FASTCALL call_12_2_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_1_7
#define init_12_2_1_7
void FASTCALL call_12_2_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_1_8
#define init_12_2_1_8
void FASTCALL call_12_2_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_1_9
#define init_12_2_1_9
void FASTCALL call_12_2_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_1_10
#define init_12_2_1_10
void FASTCALL call_12_2_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_1_11
#define init_12_2_1_11
void FASTCALL call_12_2_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_1_12
#define init_12_2_1_12
void FASTCALL call_12_2_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_1_13
#define init_12_2_1_13
void FASTCALL call_12_2_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_1_14
#define init_12_2_1_14
void FASTCALL call_12_2_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_1_15
#define init_12_2_1_15
void FASTCALL call_12_2_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_2_0
#define init_12_2_2_0
void FASTCALL call_12_2_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_2_1
#define init_12_2_2_1
void FASTCALL call_12_2_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_2_2
#define init_12_2_2_2
void FASTCALL call_12_2_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_2_3
#define init_12_2_2_3
void FASTCALL call_12_2_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_2_4
#define init_12_2_2_4
void FASTCALL call_12_2_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_2_5
#define init_12_2_2_5
void FASTCALL call_12_2_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_2_6
#define init_12_2_2_6
void FASTCALL call_12_2_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_2_7
#define init_12_2_2_7
void FASTCALL call_12_2_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_2_8
#define init_12_2_2_8
void FASTCALL call_12_2_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_2_9
#define init_12_2_2_9
void FASTCALL call_12_2_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_2_10
#define init_12_2_2_10
void FASTCALL call_12_2_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_2_11
#define init_12_2_2_11
void FASTCALL call_12_2_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_2_12
#define init_12_2_2_12
void FASTCALL call_12_2_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_2_13
#define init_12_2_2_13
void FASTCALL call_12_2_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_2_14
#define init_12_2_2_14
void FASTCALL call_12_2_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_2_15
#define init_12_2_2_15
void FASTCALL call_12_2_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_3_0
#define init_12_2_3_0
void FASTCALL call_12_2_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_3_1
#define init_12_2_3_1
void FASTCALL call_12_2_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_3_2
#define init_12_2_3_2
void FASTCALL call_12_2_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_3_3
#define init_12_2_3_3
void FASTCALL call_12_2_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_3_4
#define init_12_2_3_4
void FASTCALL call_12_2_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_3_5
#define init_12_2_3_5
void FASTCALL call_12_2_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_3_6
#define init_12_2_3_6
void FASTCALL call_12_2_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_3_7
#define init_12_2_3_7
void FASTCALL call_12_2_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_3_8
#define init_12_2_3_8
void FASTCALL call_12_2_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_3_9
#define init_12_2_3_9
void FASTCALL call_12_2_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_3_10
#define init_12_2_3_10
void FASTCALL call_12_2_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_3_11
#define init_12_2_3_11
void FASTCALL call_12_2_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_3_12
#define init_12_2_3_12
void FASTCALL call_12_2_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_3_13
#define init_12_2_3_13
void FASTCALL call_12_2_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_3_14
#define init_12_2_3_14
void FASTCALL call_12_2_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_3_15
#define init_12_2_3_15
void FASTCALL call_12_2_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_4_0
#define init_12_2_4_0
void FASTCALL call_12_2_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_4_1
#define init_12_2_4_1
void FASTCALL call_12_2_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_4_2
#define init_12_2_4_2
void FASTCALL call_12_2_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_4_3
#define init_12_2_4_3
void FASTCALL call_12_2_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_4_4
#define init_12_2_4_4
void FASTCALL call_12_2_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_4_5
#define init_12_2_4_5
void FASTCALL call_12_2_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_4_6
#define init_12_2_4_6
void FASTCALL call_12_2_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_4_7
#define init_12_2_4_7
void FASTCALL call_12_2_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_4_8
#define init_12_2_4_8
void FASTCALL call_12_2_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_4_9
#define init_12_2_4_9
void FASTCALL call_12_2_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_4_10
#define init_12_2_4_10
void FASTCALL call_12_2_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_4_11
#define init_12_2_4_11
void FASTCALL call_12_2_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_4_12
#define init_12_2_4_12
void FASTCALL call_12_2_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_4_13
#define init_12_2_4_13
void FASTCALL call_12_2_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_4_14
#define init_12_2_4_14
void FASTCALL call_12_2_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_4_15
#define init_12_2_4_15
void FASTCALL call_12_2_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_5_0
#define init_12_2_5_0
void FASTCALL call_12_2_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_5_1
#define init_12_2_5_1
void FASTCALL call_12_2_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_5_2
#define init_12_2_5_2
void FASTCALL call_12_2_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_5_3
#define init_12_2_5_3
void FASTCALL call_12_2_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_5_4
#define init_12_2_5_4
void FASTCALL call_12_2_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_5_5
#define init_12_2_5_5
void FASTCALL call_12_2_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_5_6
#define init_12_2_5_6
void FASTCALL call_12_2_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_5_7
#define init_12_2_5_7
void FASTCALL call_12_2_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_5_8
#define init_12_2_5_8
void FASTCALL call_12_2_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_5_9
#define init_12_2_5_9
void FASTCALL call_12_2_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_5_10
#define init_12_2_5_10
void FASTCALL call_12_2_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_5_11
#define init_12_2_5_11
void FASTCALL call_12_2_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_5_12
#define init_12_2_5_12
void FASTCALL call_12_2_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_5_13
#define init_12_2_5_13
void FASTCALL call_12_2_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_5_14
#define init_12_2_5_14
void FASTCALL call_12_2_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_5_15
#define init_12_2_5_15
void FASTCALL call_12_2_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_6_0
#define init_12_2_6_0
void FASTCALL call_12_2_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_6_1
#define init_12_2_6_1
void FASTCALL call_12_2_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_6_2
#define init_12_2_6_2
void FASTCALL call_12_2_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_6_3
#define init_12_2_6_3
void FASTCALL call_12_2_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_6_4
#define init_12_2_6_4
void FASTCALL call_12_2_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_6_5
#define init_12_2_6_5
void FASTCALL call_12_2_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_6_6
#define init_12_2_6_6
void FASTCALL call_12_2_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_6_7
#define init_12_2_6_7
void FASTCALL call_12_2_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_6_8
#define init_12_2_6_8
void FASTCALL call_12_2_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_6_9
#define init_12_2_6_9
void FASTCALL call_12_2_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_6_10
#define init_12_2_6_10
void FASTCALL call_12_2_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_6_11
#define init_12_2_6_11
void FASTCALL call_12_2_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_6_12
#define init_12_2_6_12
void FASTCALL call_12_2_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_6_13
#define init_12_2_6_13
void FASTCALL call_12_2_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_6_14
#define init_12_2_6_14
void FASTCALL call_12_2_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_6_15
#define init_12_2_6_15
void FASTCALL call_12_2_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_7_0
#define init_12_2_7_0
void FASTCALL call_12_2_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_7_1
#define init_12_2_7_1
void FASTCALL call_12_2_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_7_2
#define init_12_2_7_2
void FASTCALL call_12_2_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_7_3
#define init_12_2_7_3
void FASTCALL call_12_2_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_7_4
#define init_12_2_7_4
void FASTCALL call_12_2_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_7_5
#define init_12_2_7_5
void FASTCALL call_12_2_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_7_6
#define init_12_2_7_6
void FASTCALL call_12_2_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_7_7
#define init_12_2_7_7
void FASTCALL call_12_2_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_7_8
#define init_12_2_7_8
void FASTCALL call_12_2_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_7_9
#define init_12_2_7_9
void FASTCALL call_12_2_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_7_10
#define init_12_2_7_10
void FASTCALL call_12_2_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_7_11
#define init_12_2_7_11
void FASTCALL call_12_2_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_7_12
#define init_12_2_7_12
void FASTCALL call_12_2_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_7_13
#define init_12_2_7_13
void FASTCALL call_12_2_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_7_14
#define init_12_2_7_14
void FASTCALL call_12_2_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_7_15
#define init_12_2_7_15
void FASTCALL call_12_2_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_8_0
#define init_12_2_8_0
void FASTCALL call_12_2_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_8_1
#define init_12_2_8_1
void FASTCALL call_12_2_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_8_2
#define init_12_2_8_2
void FASTCALL call_12_2_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_8_3
#define init_12_2_8_3
void FASTCALL call_12_2_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_8_4
#define init_12_2_8_4
void FASTCALL call_12_2_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_8_5
#define init_12_2_8_5
void FASTCALL call_12_2_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_8_6
#define init_12_2_8_6
void FASTCALL call_12_2_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_8_7
#define init_12_2_8_7
void FASTCALL call_12_2_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_8_8
#define init_12_2_8_8
void FASTCALL call_12_2_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_8_9
#define init_12_2_8_9
void FASTCALL call_12_2_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_8_10
#define init_12_2_8_10
void FASTCALL call_12_2_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_8_11
#define init_12_2_8_11
void FASTCALL call_12_2_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_8_12
#define init_12_2_8_12
void FASTCALL call_12_2_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_8_13
#define init_12_2_8_13
void FASTCALL call_12_2_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_8_14
#define init_12_2_8_14
void FASTCALL call_12_2_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_8_15
#define init_12_2_8_15
void FASTCALL call_12_2_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_9_0
#define init_12_2_9_0
void FASTCALL call_12_2_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_9_1
#define init_12_2_9_1
void FASTCALL call_12_2_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_9_2
#define init_12_2_9_2
void FASTCALL call_12_2_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_9_3
#define init_12_2_9_3
void FASTCALL call_12_2_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_9_4
#define init_12_2_9_4
void FASTCALL call_12_2_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_9_5
#define init_12_2_9_5
void FASTCALL call_12_2_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_9_6
#define init_12_2_9_6
void FASTCALL call_12_2_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_9_7
#define init_12_2_9_7
void FASTCALL call_12_2_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_9_8
#define init_12_2_9_8
void FASTCALL call_12_2_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_9_9
#define init_12_2_9_9
void FASTCALL call_12_2_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_9_10
#define init_12_2_9_10
void FASTCALL call_12_2_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_9_11
#define init_12_2_9_11
void FASTCALL call_12_2_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_9_12
#define init_12_2_9_12
void FASTCALL call_12_2_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_9_13
#define init_12_2_9_13
void FASTCALL call_12_2_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_9_14
#define init_12_2_9_14
void FASTCALL call_12_2_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_9_15
#define init_12_2_9_15
void FASTCALL call_12_2_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_10_0
#define init_12_2_10_0
void FASTCALL call_12_2_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_10_1
#define init_12_2_10_1
void FASTCALL call_12_2_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_10_2
#define init_12_2_10_2
void FASTCALL call_12_2_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_10_3
#define init_12_2_10_3
void FASTCALL call_12_2_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_10_4
#define init_12_2_10_4
void FASTCALL call_12_2_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_10_5
#define init_12_2_10_5
void FASTCALL call_12_2_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_10_6
#define init_12_2_10_6
void FASTCALL call_12_2_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_10_7
#define init_12_2_10_7
void FASTCALL call_12_2_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_10_8
#define init_12_2_10_8
void FASTCALL call_12_2_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_10_9
#define init_12_2_10_9
void FASTCALL call_12_2_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_10_10
#define init_12_2_10_10
void FASTCALL call_12_2_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_10_11
#define init_12_2_10_11
void FASTCALL call_12_2_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_10_12
#define init_12_2_10_12
void FASTCALL call_12_2_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_10_13
#define init_12_2_10_13
void FASTCALL call_12_2_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_10_14
#define init_12_2_10_14
void FASTCALL call_12_2_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_10_15
#define init_12_2_10_15
void FASTCALL call_12_2_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_11_0
#define init_12_2_11_0
void FASTCALL call_12_2_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_11_1
#define init_12_2_11_1
void FASTCALL call_12_2_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_11_2
#define init_12_2_11_2
void FASTCALL call_12_2_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_11_3
#define init_12_2_11_3
void FASTCALL call_12_2_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_11_4
#define init_12_2_11_4
void FASTCALL call_12_2_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_11_5
#define init_12_2_11_5
void FASTCALL call_12_2_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_11_6
#define init_12_2_11_6
void FASTCALL call_12_2_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_11_7
#define init_12_2_11_7
void FASTCALL call_12_2_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_11_8
#define init_12_2_11_8
void FASTCALL call_12_2_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_11_9
#define init_12_2_11_9
void FASTCALL call_12_2_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_11_10
#define init_12_2_11_10
void FASTCALL call_12_2_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_11_11
#define init_12_2_11_11
void FASTCALL call_12_2_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_11_12
#define init_12_2_11_12
void FASTCALL call_12_2_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_11_13
#define init_12_2_11_13
void FASTCALL call_12_2_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_11_14
#define init_12_2_11_14
void FASTCALL call_12_2_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_11_15
#define init_12_2_11_15
void FASTCALL call_12_2_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_12_0
#define init_12_2_12_0
void FASTCALL call_12_2_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_12_1
#define init_12_2_12_1
void FASTCALL call_12_2_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_12_2
#define init_12_2_12_2
void FASTCALL call_12_2_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_12_3
#define init_12_2_12_3
void FASTCALL call_12_2_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_12_4
#define init_12_2_12_4
void FASTCALL call_12_2_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_12_5
#define init_12_2_12_5
void FASTCALL call_12_2_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_12_6
#define init_12_2_12_6
void FASTCALL call_12_2_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_12_7
#define init_12_2_12_7
void FASTCALL call_12_2_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_12_8
#define init_12_2_12_8
void FASTCALL call_12_2_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_12_9
#define init_12_2_12_9
void FASTCALL call_12_2_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_12_10
#define init_12_2_12_10
void FASTCALL call_12_2_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_12_11
#define init_12_2_12_11
void FASTCALL call_12_2_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_12_12
#define init_12_2_12_12
void FASTCALL call_12_2_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_12_13
#define init_12_2_12_13
void FASTCALL call_12_2_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_12_14
#define init_12_2_12_14
void FASTCALL call_12_2_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_12_15
#define init_12_2_12_15
void FASTCALL call_12_2_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_13_0
#define init_12_2_13_0
void FASTCALL call_12_2_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_13_1
#define init_12_2_13_1
void FASTCALL call_12_2_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_13_2
#define init_12_2_13_2
void FASTCALL call_12_2_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_13_3
#define init_12_2_13_3
void FASTCALL call_12_2_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_13_4
#define init_12_2_13_4
void FASTCALL call_12_2_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_13_5
#define init_12_2_13_5
void FASTCALL call_12_2_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_13_6
#define init_12_2_13_6
void FASTCALL call_12_2_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_13_7
#define init_12_2_13_7
void FASTCALL call_12_2_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_13_8
#define init_12_2_13_8
void FASTCALL call_12_2_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_13_9
#define init_12_2_13_9
void FASTCALL call_12_2_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_13_10
#define init_12_2_13_10
void FASTCALL call_12_2_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_13_11
#define init_12_2_13_11
void FASTCALL call_12_2_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_13_12
#define init_12_2_13_12
void FASTCALL call_12_2_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_13_13
#define init_12_2_13_13
void FASTCALL call_12_2_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_13_14
#define init_12_2_13_14
void FASTCALL call_12_2_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_13_15
#define init_12_2_13_15
void FASTCALL call_12_2_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_14_0
#define init_12_2_14_0
void FASTCALL call_12_2_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_14_1
#define init_12_2_14_1
void FASTCALL call_12_2_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_14_2
#define init_12_2_14_2
void FASTCALL call_12_2_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_14_3
#define init_12_2_14_3
void FASTCALL call_12_2_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_14_4
#define init_12_2_14_4
void FASTCALL call_12_2_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_14_5
#define init_12_2_14_5
void FASTCALL call_12_2_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_14_6
#define init_12_2_14_6
void FASTCALL call_12_2_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_14_7
#define init_12_2_14_7
void FASTCALL call_12_2_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_14_8
#define init_12_2_14_8
void FASTCALL call_12_2_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_14_9
#define init_12_2_14_9
void FASTCALL call_12_2_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_14_10
#define init_12_2_14_10
void FASTCALL call_12_2_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_14_11
#define init_12_2_14_11
void FASTCALL call_12_2_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_14_12
#define init_12_2_14_12
void FASTCALL call_12_2_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_14_13
#define init_12_2_14_13
void FASTCALL call_12_2_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_14_14
#define init_12_2_14_14
void FASTCALL call_12_2_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_14_15
#define init_12_2_14_15
void FASTCALL call_12_2_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_15_0
#define init_12_2_15_0
void FASTCALL call_12_2_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_15_1
#define init_12_2_15_1
void FASTCALL call_12_2_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_15_2
#define init_12_2_15_2
void FASTCALL call_12_2_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_15_3
#define init_12_2_15_3
void FASTCALL call_12_2_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_15_4
#define init_12_2_15_4
void FASTCALL call_12_2_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_15_5
#define init_12_2_15_5
void FASTCALL call_12_2_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_15_6
#define init_12_2_15_6
void FASTCALL call_12_2_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_15_7
#define init_12_2_15_7
void FASTCALL call_12_2_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_15_8
#define init_12_2_15_8
void FASTCALL call_12_2_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_15_9
#define init_12_2_15_9
void FASTCALL call_12_2_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_15_10
#define init_12_2_15_10
void FASTCALL call_12_2_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_15_11
#define init_12_2_15_11
void FASTCALL call_12_2_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_15_12
#define init_12_2_15_12
void FASTCALL call_12_2_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_15_13
#define init_12_2_15_13
void FASTCALL call_12_2_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_15_14
#define init_12_2_15_14
void FASTCALL call_12_2_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_2_15_15
#define init_12_2_15_15
void FASTCALL call_12_2_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_0_0
#define init_12_3_0_0
void FASTCALL call_12_3_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_0_1
#define init_12_3_0_1
void FASTCALL call_12_3_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_0_2
#define init_12_3_0_2
void FASTCALL call_12_3_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_0_3
#define init_12_3_0_3
void FASTCALL call_12_3_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_0_4
#define init_12_3_0_4
void FASTCALL call_12_3_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_0_5
#define init_12_3_0_5
void FASTCALL call_12_3_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_0_6
#define init_12_3_0_6
void FASTCALL call_12_3_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_0_7
#define init_12_3_0_7
void FASTCALL call_12_3_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_0_8
#define init_12_3_0_8
void FASTCALL call_12_3_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_0_9
#define init_12_3_0_9
void FASTCALL call_12_3_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_0_10
#define init_12_3_0_10
void FASTCALL call_12_3_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_0_11
#define init_12_3_0_11
void FASTCALL call_12_3_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_0_12
#define init_12_3_0_12
void FASTCALL call_12_3_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_0_13
#define init_12_3_0_13
void FASTCALL call_12_3_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_0_14
#define init_12_3_0_14
void FASTCALL call_12_3_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_0_15
#define init_12_3_0_15
void FASTCALL call_12_3_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_1_0
#define init_12_3_1_0
void FASTCALL call_12_3_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_1_1
#define init_12_3_1_1
void FASTCALL call_12_3_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_1_2
#define init_12_3_1_2
void FASTCALL call_12_3_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_1_3
#define init_12_3_1_3
void FASTCALL call_12_3_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_1_4
#define init_12_3_1_4
void FASTCALL call_12_3_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_1_5
#define init_12_3_1_5
void FASTCALL call_12_3_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_1_6
#define init_12_3_1_6
void FASTCALL call_12_3_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_1_7
#define init_12_3_1_7
void FASTCALL call_12_3_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_1_8
#define init_12_3_1_8
void FASTCALL call_12_3_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_1_9
#define init_12_3_1_9
void FASTCALL call_12_3_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_1_10
#define init_12_3_1_10
void FASTCALL call_12_3_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_1_11
#define init_12_3_1_11
void FASTCALL call_12_3_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_1_12
#define init_12_3_1_12
void FASTCALL call_12_3_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_1_13
#define init_12_3_1_13
void FASTCALL call_12_3_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_1_14
#define init_12_3_1_14
void FASTCALL call_12_3_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_1_15
#define init_12_3_1_15
void FASTCALL call_12_3_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_2_0
#define init_12_3_2_0
void FASTCALL call_12_3_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_2_1
#define init_12_3_2_1
void FASTCALL call_12_3_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_2_2
#define init_12_3_2_2
void FASTCALL call_12_3_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_2_3
#define init_12_3_2_3
void FASTCALL call_12_3_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_2_4
#define init_12_3_2_4
void FASTCALL call_12_3_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_2_5
#define init_12_3_2_5
void FASTCALL call_12_3_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_2_6
#define init_12_3_2_6
void FASTCALL call_12_3_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_2_7
#define init_12_3_2_7
void FASTCALL call_12_3_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_2_8
#define init_12_3_2_8
void FASTCALL call_12_3_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_2_9
#define init_12_3_2_9
void FASTCALL call_12_3_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_2_10
#define init_12_3_2_10
void FASTCALL call_12_3_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_2_11
#define init_12_3_2_11
void FASTCALL call_12_3_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_2_12
#define init_12_3_2_12
void FASTCALL call_12_3_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_2_13
#define init_12_3_2_13
void FASTCALL call_12_3_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_2_14
#define init_12_3_2_14
void FASTCALL call_12_3_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_2_15
#define init_12_3_2_15
void FASTCALL call_12_3_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_3_0
#define init_12_3_3_0
void FASTCALL call_12_3_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_3_1
#define init_12_3_3_1
void FASTCALL call_12_3_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_3_2
#define init_12_3_3_2
void FASTCALL call_12_3_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_3_3
#define init_12_3_3_3
void FASTCALL call_12_3_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_3_4
#define init_12_3_3_4
void FASTCALL call_12_3_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_3_5
#define init_12_3_3_5
void FASTCALL call_12_3_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_3_6
#define init_12_3_3_6
void FASTCALL call_12_3_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_3_7
#define init_12_3_3_7
void FASTCALL call_12_3_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_3_8
#define init_12_3_3_8
void FASTCALL call_12_3_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_3_9
#define init_12_3_3_9
void FASTCALL call_12_3_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_3_10
#define init_12_3_3_10
void FASTCALL call_12_3_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_3_11
#define init_12_3_3_11
void FASTCALL call_12_3_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_3_12
#define init_12_3_3_12
void FASTCALL call_12_3_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_3_13
#define init_12_3_3_13
void FASTCALL call_12_3_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_3_14
#define init_12_3_3_14
void FASTCALL call_12_3_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_3_15
#define init_12_3_3_15
void FASTCALL call_12_3_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_4_0
#define init_12_3_4_0
void FASTCALL call_12_3_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_4_1
#define init_12_3_4_1
void FASTCALL call_12_3_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_4_2
#define init_12_3_4_2
void FASTCALL call_12_3_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_4_3
#define init_12_3_4_3
void FASTCALL call_12_3_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_4_4
#define init_12_3_4_4
void FASTCALL call_12_3_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_4_5
#define init_12_3_4_5
void FASTCALL call_12_3_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_4_6
#define init_12_3_4_6
void FASTCALL call_12_3_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_4_7
#define init_12_3_4_7
void FASTCALL call_12_3_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_4_8
#define init_12_3_4_8
void FASTCALL call_12_3_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_4_9
#define init_12_3_4_9
void FASTCALL call_12_3_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_4_10
#define init_12_3_4_10
void FASTCALL call_12_3_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_4_11
#define init_12_3_4_11
void FASTCALL call_12_3_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_4_12
#define init_12_3_4_12
void FASTCALL call_12_3_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_4_13
#define init_12_3_4_13
void FASTCALL call_12_3_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_4_14
#define init_12_3_4_14
void FASTCALL call_12_3_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_4_15
#define init_12_3_4_15
void FASTCALL call_12_3_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_5_0
#define init_12_3_5_0
void FASTCALL call_12_3_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_5_1
#define init_12_3_5_1
void FASTCALL call_12_3_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_5_2
#define init_12_3_5_2
void FASTCALL call_12_3_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_5_3
#define init_12_3_5_3
void FASTCALL call_12_3_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_5_4
#define init_12_3_5_4
void FASTCALL call_12_3_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_5_5
#define init_12_3_5_5
void FASTCALL call_12_3_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_5_6
#define init_12_3_5_6
void FASTCALL call_12_3_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_5_7
#define init_12_3_5_7
void FASTCALL call_12_3_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_5_8
#define init_12_3_5_8
void FASTCALL call_12_3_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_5_9
#define init_12_3_5_9
void FASTCALL call_12_3_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_5_10
#define init_12_3_5_10
void FASTCALL call_12_3_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_5_11
#define init_12_3_5_11
void FASTCALL call_12_3_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_5_12
#define init_12_3_5_12
void FASTCALL call_12_3_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_5_13
#define init_12_3_5_13
void FASTCALL call_12_3_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_5_14
#define init_12_3_5_14
void FASTCALL call_12_3_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_5_15
#define init_12_3_5_15
void FASTCALL call_12_3_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_6_0
#define init_12_3_6_0
void FASTCALL call_12_3_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_6_1
#define init_12_3_6_1
void FASTCALL call_12_3_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_6_2
#define init_12_3_6_2
void FASTCALL call_12_3_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_6_3
#define init_12_3_6_3
void FASTCALL call_12_3_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_6_4
#define init_12_3_6_4
void FASTCALL call_12_3_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_6_5
#define init_12_3_6_5
void FASTCALL call_12_3_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_6_6
#define init_12_3_6_6
void FASTCALL call_12_3_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_6_7
#define init_12_3_6_7
void FASTCALL call_12_3_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_6_8
#define init_12_3_6_8
void FASTCALL call_12_3_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_6_9
#define init_12_3_6_9
void FASTCALL call_12_3_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_6_10
#define init_12_3_6_10
void FASTCALL call_12_3_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_6_11
#define init_12_3_6_11
void FASTCALL call_12_3_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_6_12
#define init_12_3_6_12
void FASTCALL call_12_3_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_6_13
#define init_12_3_6_13
void FASTCALL call_12_3_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_6_14
#define init_12_3_6_14
void FASTCALL call_12_3_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_6_15
#define init_12_3_6_15
void FASTCALL call_12_3_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_7_0
#define init_12_3_7_0
void FASTCALL call_12_3_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_7_1
#define init_12_3_7_1
void FASTCALL call_12_3_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_7_2
#define init_12_3_7_2
void FASTCALL call_12_3_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_7_3
#define init_12_3_7_3
void FASTCALL call_12_3_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_7_4
#define init_12_3_7_4
void FASTCALL call_12_3_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_7_5
#define init_12_3_7_5
void FASTCALL call_12_3_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_7_6
#define init_12_3_7_6
void FASTCALL call_12_3_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_7_7
#define init_12_3_7_7
void FASTCALL call_12_3_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_7_8
#define init_12_3_7_8
void FASTCALL call_12_3_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_7_9
#define init_12_3_7_9
void FASTCALL call_12_3_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_7_10
#define init_12_3_7_10
void FASTCALL call_12_3_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_7_11
#define init_12_3_7_11
void FASTCALL call_12_3_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_7_12
#define init_12_3_7_12
void FASTCALL call_12_3_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_7_13
#define init_12_3_7_13
void FASTCALL call_12_3_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_7_14
#define init_12_3_7_14
void FASTCALL call_12_3_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_7_15
#define init_12_3_7_15
void FASTCALL call_12_3_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_8_0
#define init_12_3_8_0
void FASTCALL call_12_3_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_8_1
#define init_12_3_8_1
void FASTCALL call_12_3_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_8_2
#define init_12_3_8_2
void FASTCALL call_12_3_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_8_3
#define init_12_3_8_3
void FASTCALL call_12_3_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_8_4
#define init_12_3_8_4
void FASTCALL call_12_3_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_8_5
#define init_12_3_8_5
void FASTCALL call_12_3_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_8_6
#define init_12_3_8_6
void FASTCALL call_12_3_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_8_7
#define init_12_3_8_7
void FASTCALL call_12_3_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_8_8
#define init_12_3_8_8
void FASTCALL call_12_3_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_8_9
#define init_12_3_8_9
void FASTCALL call_12_3_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_8_10
#define init_12_3_8_10
void FASTCALL call_12_3_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_8_11
#define init_12_3_8_11
void FASTCALL call_12_3_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_8_12
#define init_12_3_8_12
void FASTCALL call_12_3_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_8_13
#define init_12_3_8_13
void FASTCALL call_12_3_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_8_14
#define init_12_3_8_14
void FASTCALL call_12_3_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_8_15
#define init_12_3_8_15
void FASTCALL call_12_3_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_9_0
#define init_12_3_9_0
void FASTCALL call_12_3_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_9_1
#define init_12_3_9_1
void FASTCALL call_12_3_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_9_2
#define init_12_3_9_2
void FASTCALL call_12_3_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_9_3
#define init_12_3_9_3
void FASTCALL call_12_3_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_9_4
#define init_12_3_9_4
void FASTCALL call_12_3_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_9_5
#define init_12_3_9_5
void FASTCALL call_12_3_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_9_6
#define init_12_3_9_6
void FASTCALL call_12_3_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_9_7
#define init_12_3_9_7
void FASTCALL call_12_3_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_9_8
#define init_12_3_9_8
void FASTCALL call_12_3_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_9_9
#define init_12_3_9_9
void FASTCALL call_12_3_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_9_10
#define init_12_3_9_10
void FASTCALL call_12_3_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_9_11
#define init_12_3_9_11
void FASTCALL call_12_3_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_9_12
#define init_12_3_9_12
void FASTCALL call_12_3_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_9_13
#define init_12_3_9_13
void FASTCALL call_12_3_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_9_14
#define init_12_3_9_14
void FASTCALL call_12_3_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_9_15
#define init_12_3_9_15
void FASTCALL call_12_3_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_10_0
#define init_12_3_10_0
void FASTCALL call_12_3_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_10_1
#define init_12_3_10_1
void FASTCALL call_12_3_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_10_2
#define init_12_3_10_2
void FASTCALL call_12_3_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_10_3
#define init_12_3_10_3
void FASTCALL call_12_3_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_10_4
#define init_12_3_10_4
void FASTCALL call_12_3_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_10_5
#define init_12_3_10_5
void FASTCALL call_12_3_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_10_6
#define init_12_3_10_6
void FASTCALL call_12_3_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_10_7
#define init_12_3_10_7
void FASTCALL call_12_3_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_10_8
#define init_12_3_10_8
void FASTCALL call_12_3_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_10_9
#define init_12_3_10_9
void FASTCALL call_12_3_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_10_10
#define init_12_3_10_10
void FASTCALL call_12_3_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_10_11
#define init_12_3_10_11
void FASTCALL call_12_3_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_10_12
#define init_12_3_10_12
void FASTCALL call_12_3_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_10_13
#define init_12_3_10_13
void FASTCALL call_12_3_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_10_14
#define init_12_3_10_14
void FASTCALL call_12_3_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_10_15
#define init_12_3_10_15
void FASTCALL call_12_3_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_11_0
#define init_12_3_11_0
void FASTCALL call_12_3_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_11_1
#define init_12_3_11_1
void FASTCALL call_12_3_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_11_2
#define init_12_3_11_2
void FASTCALL call_12_3_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_11_3
#define init_12_3_11_3
void FASTCALL call_12_3_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_11_4
#define init_12_3_11_4
void FASTCALL call_12_3_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_11_5
#define init_12_3_11_5
void FASTCALL call_12_3_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_11_6
#define init_12_3_11_6
void FASTCALL call_12_3_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_11_7
#define init_12_3_11_7
void FASTCALL call_12_3_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_11_8
#define init_12_3_11_8
void FASTCALL call_12_3_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_11_9
#define init_12_3_11_9
void FASTCALL call_12_3_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_11_10
#define init_12_3_11_10
void FASTCALL call_12_3_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_11_11
#define init_12_3_11_11
void FASTCALL call_12_3_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_11_12
#define init_12_3_11_12
void FASTCALL call_12_3_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_11_13
#define init_12_3_11_13
void FASTCALL call_12_3_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_11_14
#define init_12_3_11_14
void FASTCALL call_12_3_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_11_15
#define init_12_3_11_15
void FASTCALL call_12_3_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_12_0
#define init_12_3_12_0
void FASTCALL call_12_3_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_12_1
#define init_12_3_12_1
void FASTCALL call_12_3_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_12_2
#define init_12_3_12_2
void FASTCALL call_12_3_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_12_3
#define init_12_3_12_3
void FASTCALL call_12_3_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_12_4
#define init_12_3_12_4
void FASTCALL call_12_3_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_12_5
#define init_12_3_12_5
void FASTCALL call_12_3_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_12_6
#define init_12_3_12_6
void FASTCALL call_12_3_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_12_7
#define init_12_3_12_7
void FASTCALL call_12_3_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_12_8
#define init_12_3_12_8
void FASTCALL call_12_3_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_12_9
#define init_12_3_12_9
void FASTCALL call_12_3_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_12_10
#define init_12_3_12_10
void FASTCALL call_12_3_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_12_11
#define init_12_3_12_11
void FASTCALL call_12_3_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_12_12
#define init_12_3_12_12
void FASTCALL call_12_3_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_12_13
#define init_12_3_12_13
void FASTCALL call_12_3_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_12_14
#define init_12_3_12_14
void FASTCALL call_12_3_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_12_15
#define init_12_3_12_15
void FASTCALL call_12_3_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_13_0
#define init_12_3_13_0
void FASTCALL call_12_3_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_13_1
#define init_12_3_13_1
void FASTCALL call_12_3_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_13_2
#define init_12_3_13_2
void FASTCALL call_12_3_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_13_3
#define init_12_3_13_3
void FASTCALL call_12_3_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_13_4
#define init_12_3_13_4
void FASTCALL call_12_3_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_13_5
#define init_12_3_13_5
void FASTCALL call_12_3_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_13_6
#define init_12_3_13_6
void FASTCALL call_12_3_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_13_7
#define init_12_3_13_7
void FASTCALL call_12_3_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_13_8
#define init_12_3_13_8
void FASTCALL call_12_3_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_13_9
#define init_12_3_13_9
void FASTCALL call_12_3_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_13_10
#define init_12_3_13_10
void FASTCALL call_12_3_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_13_11
#define init_12_3_13_11
void FASTCALL call_12_3_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_13_12
#define init_12_3_13_12
void FASTCALL call_12_3_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_13_13
#define init_12_3_13_13
void FASTCALL call_12_3_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_13_14
#define init_12_3_13_14
void FASTCALL call_12_3_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_13_15
#define init_12_3_13_15
void FASTCALL call_12_3_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_14_0
#define init_12_3_14_0
void FASTCALL call_12_3_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_14_1
#define init_12_3_14_1
void FASTCALL call_12_3_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_14_2
#define init_12_3_14_2
void FASTCALL call_12_3_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_14_3
#define init_12_3_14_3
void FASTCALL call_12_3_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_14_4
#define init_12_3_14_4
void FASTCALL call_12_3_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_14_5
#define init_12_3_14_5
void FASTCALL call_12_3_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_14_6
#define init_12_3_14_6
void FASTCALL call_12_3_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_14_7
#define init_12_3_14_7
void FASTCALL call_12_3_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_14_8
#define init_12_3_14_8
void FASTCALL call_12_3_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_14_9
#define init_12_3_14_9
void FASTCALL call_12_3_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_14_10
#define init_12_3_14_10
void FASTCALL call_12_3_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_14_11
#define init_12_3_14_11
void FASTCALL call_12_3_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_14_12
#define init_12_3_14_12
void FASTCALL call_12_3_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_14_13
#define init_12_3_14_13
void FASTCALL call_12_3_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_14_14
#define init_12_3_14_14
void FASTCALL call_12_3_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_14_15
#define init_12_3_14_15
void FASTCALL call_12_3_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_15_0
#define init_12_3_15_0
void FASTCALL call_12_3_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_15_1
#define init_12_3_15_1
void FASTCALL call_12_3_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_15_2
#define init_12_3_15_2
void FASTCALL call_12_3_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_15_3
#define init_12_3_15_3
void FASTCALL call_12_3_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_15_4
#define init_12_3_15_4
void FASTCALL call_12_3_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_15_5
#define init_12_3_15_5
void FASTCALL call_12_3_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_15_6
#define init_12_3_15_6
void FASTCALL call_12_3_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_15_7
#define init_12_3_15_7
void FASTCALL call_12_3_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_15_8
#define init_12_3_15_8
void FASTCALL call_12_3_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_15_9
#define init_12_3_15_9
void FASTCALL call_12_3_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_15_10
#define init_12_3_15_10
void FASTCALL call_12_3_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_15_11
#define init_12_3_15_11
void FASTCALL call_12_3_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_15_12
#define init_12_3_15_12
void FASTCALL call_12_3_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_15_13
#define init_12_3_15_13
void FASTCALL call_12_3_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_15_14
#define init_12_3_15_14
void FASTCALL call_12_3_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_3_15_15
#define init_12_3_15_15
void FASTCALL call_12_3_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_0_0
#define init_12_4_0_0
void FASTCALL call_12_4_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_0_1
#define init_12_4_0_1
void FASTCALL call_12_4_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_0_2
#define init_12_4_0_2
void FASTCALL call_12_4_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_0_3
#define init_12_4_0_3
void FASTCALL call_12_4_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_0_4
#define init_12_4_0_4
void FASTCALL call_12_4_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_0_5
#define init_12_4_0_5
void FASTCALL call_12_4_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_0_6
#define init_12_4_0_6
void FASTCALL call_12_4_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_0_7
#define init_12_4_0_7
void FASTCALL call_12_4_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_0_8
#define init_12_4_0_8
void FASTCALL call_12_4_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_0_9
#define init_12_4_0_9
void FASTCALL call_12_4_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_0_10
#define init_12_4_0_10
void FASTCALL call_12_4_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_0_11
#define init_12_4_0_11
void FASTCALL call_12_4_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_0_12
#define init_12_4_0_12
void FASTCALL call_12_4_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_0_13
#define init_12_4_0_13
void FASTCALL call_12_4_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_0_14
#define init_12_4_0_14
void FASTCALL call_12_4_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_0_15
#define init_12_4_0_15
void FASTCALL call_12_4_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_1_0
#define init_12_4_1_0
void FASTCALL call_12_4_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_1_1
#define init_12_4_1_1
void FASTCALL call_12_4_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_1_2
#define init_12_4_1_2
void FASTCALL call_12_4_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_1_3
#define init_12_4_1_3
void FASTCALL call_12_4_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_1_4
#define init_12_4_1_4
void FASTCALL call_12_4_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_1_5
#define init_12_4_1_5
void FASTCALL call_12_4_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_1_6
#define init_12_4_1_6
void FASTCALL call_12_4_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_1_7
#define init_12_4_1_7
void FASTCALL call_12_4_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_1_8
#define init_12_4_1_8
void FASTCALL call_12_4_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_1_9
#define init_12_4_1_9
void FASTCALL call_12_4_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_1_10
#define init_12_4_1_10
void FASTCALL call_12_4_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_1_11
#define init_12_4_1_11
void FASTCALL call_12_4_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_1_12
#define init_12_4_1_12
void FASTCALL call_12_4_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_1_13
#define init_12_4_1_13
void FASTCALL call_12_4_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_1_14
#define init_12_4_1_14
void FASTCALL call_12_4_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_1_15
#define init_12_4_1_15
void FASTCALL call_12_4_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_2_0
#define init_12_4_2_0
void FASTCALL call_12_4_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_2_1
#define init_12_4_2_1
void FASTCALL call_12_4_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_2_2
#define init_12_4_2_2
void FASTCALL call_12_4_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_2_3
#define init_12_4_2_3
void FASTCALL call_12_4_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_2_4
#define init_12_4_2_4
void FASTCALL call_12_4_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_2_5
#define init_12_4_2_5
void FASTCALL call_12_4_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_2_6
#define init_12_4_2_6
void FASTCALL call_12_4_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_2_7
#define init_12_4_2_7
void FASTCALL call_12_4_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_2_8
#define init_12_4_2_8
void FASTCALL call_12_4_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_2_9
#define init_12_4_2_9
void FASTCALL call_12_4_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_2_10
#define init_12_4_2_10
void FASTCALL call_12_4_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_2_11
#define init_12_4_2_11
void FASTCALL call_12_4_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_2_12
#define init_12_4_2_12
void FASTCALL call_12_4_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_2_13
#define init_12_4_2_13
void FASTCALL call_12_4_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_2_14
#define init_12_4_2_14
void FASTCALL call_12_4_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_2_15
#define init_12_4_2_15
void FASTCALL call_12_4_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_3_0
#define init_12_4_3_0
void FASTCALL call_12_4_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_3_1
#define init_12_4_3_1
void FASTCALL call_12_4_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_3_2
#define init_12_4_3_2
void FASTCALL call_12_4_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_3_3
#define init_12_4_3_3
void FASTCALL call_12_4_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_3_4
#define init_12_4_3_4
void FASTCALL call_12_4_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_3_5
#define init_12_4_3_5
void FASTCALL call_12_4_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_3_6
#define init_12_4_3_6
void FASTCALL call_12_4_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_3_7
#define init_12_4_3_7
void FASTCALL call_12_4_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_3_8
#define init_12_4_3_8
void FASTCALL call_12_4_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_3_9
#define init_12_4_3_9
void FASTCALL call_12_4_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_3_10
#define init_12_4_3_10
void FASTCALL call_12_4_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_3_11
#define init_12_4_3_11
void FASTCALL call_12_4_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_3_12
#define init_12_4_3_12
void FASTCALL call_12_4_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_3_13
#define init_12_4_3_13
void FASTCALL call_12_4_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_3_14
#define init_12_4_3_14
void FASTCALL call_12_4_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_3_15
#define init_12_4_3_15
void FASTCALL call_12_4_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_4_0
#define init_12_4_4_0
void FASTCALL call_12_4_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_4_1
#define init_12_4_4_1
void FASTCALL call_12_4_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_4_2
#define init_12_4_4_2
void FASTCALL call_12_4_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_4_3
#define init_12_4_4_3
void FASTCALL call_12_4_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_4_4
#define init_12_4_4_4
void FASTCALL call_12_4_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_4_5
#define init_12_4_4_5
void FASTCALL call_12_4_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_4_6
#define init_12_4_4_6
void FASTCALL call_12_4_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_4_7
#define init_12_4_4_7
void FASTCALL call_12_4_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_4_8
#define init_12_4_4_8
void FASTCALL call_12_4_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_4_9
#define init_12_4_4_9
void FASTCALL call_12_4_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_4_10
#define init_12_4_4_10
void FASTCALL call_12_4_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_4_11
#define init_12_4_4_11
void FASTCALL call_12_4_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_4_12
#define init_12_4_4_12
void FASTCALL call_12_4_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_4_13
#define init_12_4_4_13
void FASTCALL call_12_4_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_4_14
#define init_12_4_4_14
void FASTCALL call_12_4_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_4_15
#define init_12_4_4_15
void FASTCALL call_12_4_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_5_0
#define init_12_4_5_0
void FASTCALL call_12_4_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_5_1
#define init_12_4_5_1
void FASTCALL call_12_4_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_5_2
#define init_12_4_5_2
void FASTCALL call_12_4_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_5_3
#define init_12_4_5_3
void FASTCALL call_12_4_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_5_4
#define init_12_4_5_4
void FASTCALL call_12_4_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_5_5
#define init_12_4_5_5
void FASTCALL call_12_4_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_5_6
#define init_12_4_5_6
void FASTCALL call_12_4_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_5_7
#define init_12_4_5_7
void FASTCALL call_12_4_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_5_8
#define init_12_4_5_8
void FASTCALL call_12_4_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_5_9
#define init_12_4_5_9
void FASTCALL call_12_4_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_5_10
#define init_12_4_5_10
void FASTCALL call_12_4_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_5_11
#define init_12_4_5_11
void FASTCALL call_12_4_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_5_12
#define init_12_4_5_12
void FASTCALL call_12_4_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_5_13
#define init_12_4_5_13
void FASTCALL call_12_4_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_5_14
#define init_12_4_5_14
void FASTCALL call_12_4_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_5_15
#define init_12_4_5_15
void FASTCALL call_12_4_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_6_0
#define init_12_4_6_0
void FASTCALL call_12_4_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_6_1
#define init_12_4_6_1
void FASTCALL call_12_4_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_6_2
#define init_12_4_6_2
void FASTCALL call_12_4_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_6_3
#define init_12_4_6_3
void FASTCALL call_12_4_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_6_4
#define init_12_4_6_4
void FASTCALL call_12_4_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_6_5
#define init_12_4_6_5
void FASTCALL call_12_4_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_6_6
#define init_12_4_6_6
void FASTCALL call_12_4_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_6_7
#define init_12_4_6_7
void FASTCALL call_12_4_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_6_8
#define init_12_4_6_8
void FASTCALL call_12_4_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_6_9
#define init_12_4_6_9
void FASTCALL call_12_4_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_6_10
#define init_12_4_6_10
void FASTCALL call_12_4_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_6_11
#define init_12_4_6_11
void FASTCALL call_12_4_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_6_12
#define init_12_4_6_12
void FASTCALL call_12_4_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_6_13
#define init_12_4_6_13
void FASTCALL call_12_4_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_6_14
#define init_12_4_6_14
void FASTCALL call_12_4_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_6_15
#define init_12_4_6_15
void FASTCALL call_12_4_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_7_0
#define init_12_4_7_0
void FASTCALL call_12_4_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_7_1
#define init_12_4_7_1
void FASTCALL call_12_4_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_7_2
#define init_12_4_7_2
void FASTCALL call_12_4_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_7_3
#define init_12_4_7_3
void FASTCALL call_12_4_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_7_4
#define init_12_4_7_4
void FASTCALL call_12_4_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_7_5
#define init_12_4_7_5
void FASTCALL call_12_4_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_7_6
#define init_12_4_7_6
void FASTCALL call_12_4_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_7_7
#define init_12_4_7_7
void FASTCALL call_12_4_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_7_8
#define init_12_4_7_8
void FASTCALL call_12_4_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_7_9
#define init_12_4_7_9
void FASTCALL call_12_4_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_7_10
#define init_12_4_7_10
void FASTCALL call_12_4_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_7_11
#define init_12_4_7_11
void FASTCALL call_12_4_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_7_12
#define init_12_4_7_12
void FASTCALL call_12_4_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_7_13
#define init_12_4_7_13
void FASTCALL call_12_4_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_7_14
#define init_12_4_7_14
void FASTCALL call_12_4_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_7_15
#define init_12_4_7_15
void FASTCALL call_12_4_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_8_0
#define init_12_4_8_0
void FASTCALL call_12_4_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_8_1
#define init_12_4_8_1
void FASTCALL call_12_4_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_8_2
#define init_12_4_8_2
void FASTCALL call_12_4_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_8_3
#define init_12_4_8_3
void FASTCALL call_12_4_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_8_4
#define init_12_4_8_4
void FASTCALL call_12_4_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_8_5
#define init_12_4_8_5
void FASTCALL call_12_4_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_8_6
#define init_12_4_8_6
void FASTCALL call_12_4_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_8_7
#define init_12_4_8_7
void FASTCALL call_12_4_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_8_8
#define init_12_4_8_8
void FASTCALL call_12_4_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_8_9
#define init_12_4_8_9
void FASTCALL call_12_4_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_8_10
#define init_12_4_8_10
void FASTCALL call_12_4_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_8_11
#define init_12_4_8_11
void FASTCALL call_12_4_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_8_12
#define init_12_4_8_12
void FASTCALL call_12_4_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_8_13
#define init_12_4_8_13
void FASTCALL call_12_4_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_8_14
#define init_12_4_8_14
void FASTCALL call_12_4_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_8_15
#define init_12_4_8_15
void FASTCALL call_12_4_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_9_0
#define init_12_4_9_0
void FASTCALL call_12_4_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_9_1
#define init_12_4_9_1
void FASTCALL call_12_4_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_9_2
#define init_12_4_9_2
void FASTCALL call_12_4_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_9_3
#define init_12_4_9_3
void FASTCALL call_12_4_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_9_4
#define init_12_4_9_4
void FASTCALL call_12_4_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_9_5
#define init_12_4_9_5
void FASTCALL call_12_4_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_9_6
#define init_12_4_9_6
void FASTCALL call_12_4_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_9_7
#define init_12_4_9_7
void FASTCALL call_12_4_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_9_8
#define init_12_4_9_8
void FASTCALL call_12_4_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_9_9
#define init_12_4_9_9
void FASTCALL call_12_4_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_9_10
#define init_12_4_9_10
void FASTCALL call_12_4_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_9_11
#define init_12_4_9_11
void FASTCALL call_12_4_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_9_12
#define init_12_4_9_12
void FASTCALL call_12_4_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_9_13
#define init_12_4_9_13
void FASTCALL call_12_4_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_9_14
#define init_12_4_9_14
void FASTCALL call_12_4_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_9_15
#define init_12_4_9_15
void FASTCALL call_12_4_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_10_0
#define init_12_4_10_0
void FASTCALL call_12_4_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_10_1
#define init_12_4_10_1
void FASTCALL call_12_4_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_10_2
#define init_12_4_10_2
void FASTCALL call_12_4_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_10_3
#define init_12_4_10_3
void FASTCALL call_12_4_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_10_4
#define init_12_4_10_4
void FASTCALL call_12_4_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_10_5
#define init_12_4_10_5
void FASTCALL call_12_4_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_10_6
#define init_12_4_10_6
void FASTCALL call_12_4_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_10_7
#define init_12_4_10_7
void FASTCALL call_12_4_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_10_8
#define init_12_4_10_8
void FASTCALL call_12_4_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_10_9
#define init_12_4_10_9
void FASTCALL call_12_4_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_10_10
#define init_12_4_10_10
void FASTCALL call_12_4_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_10_11
#define init_12_4_10_11
void FASTCALL call_12_4_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_10_12
#define init_12_4_10_12
void FASTCALL call_12_4_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_10_13
#define init_12_4_10_13
void FASTCALL call_12_4_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_10_14
#define init_12_4_10_14
void FASTCALL call_12_4_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_10_15
#define init_12_4_10_15
void FASTCALL call_12_4_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_11_0
#define init_12_4_11_0
void FASTCALL call_12_4_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_11_1
#define init_12_4_11_1
void FASTCALL call_12_4_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_11_2
#define init_12_4_11_2
void FASTCALL call_12_4_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_11_3
#define init_12_4_11_3
void FASTCALL call_12_4_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_11_4
#define init_12_4_11_4
void FASTCALL call_12_4_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_11_5
#define init_12_4_11_5
void FASTCALL call_12_4_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_11_6
#define init_12_4_11_6
void FASTCALL call_12_4_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_11_7
#define init_12_4_11_7
void FASTCALL call_12_4_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_11_8
#define init_12_4_11_8
void FASTCALL call_12_4_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_11_9
#define init_12_4_11_9
void FASTCALL call_12_4_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_11_10
#define init_12_4_11_10
void FASTCALL call_12_4_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_11_11
#define init_12_4_11_11
void FASTCALL call_12_4_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_11_12
#define init_12_4_11_12
void FASTCALL call_12_4_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_11_13
#define init_12_4_11_13
void FASTCALL call_12_4_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_11_14
#define init_12_4_11_14
void FASTCALL call_12_4_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_11_15
#define init_12_4_11_15
void FASTCALL call_12_4_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_12_0
#define init_12_4_12_0
void FASTCALL call_12_4_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_12_1
#define init_12_4_12_1
void FASTCALL call_12_4_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_12_2
#define init_12_4_12_2
void FASTCALL call_12_4_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_12_3
#define init_12_4_12_3
void FASTCALL call_12_4_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_12_4
#define init_12_4_12_4
void FASTCALL call_12_4_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_12_5
#define init_12_4_12_5
void FASTCALL call_12_4_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_12_6
#define init_12_4_12_6
void FASTCALL call_12_4_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_12_7
#define init_12_4_12_7
void FASTCALL call_12_4_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_12_8
#define init_12_4_12_8
void FASTCALL call_12_4_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_12_9
#define init_12_4_12_9
void FASTCALL call_12_4_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_12_10
#define init_12_4_12_10
void FASTCALL call_12_4_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_12_11
#define init_12_4_12_11
void FASTCALL call_12_4_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_12_12
#define init_12_4_12_12
void FASTCALL call_12_4_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_12_13
#define init_12_4_12_13
void FASTCALL call_12_4_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_12_14
#define init_12_4_12_14
void FASTCALL call_12_4_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_12_15
#define init_12_4_12_15
void FASTCALL call_12_4_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_13_0
#define init_12_4_13_0
void FASTCALL call_12_4_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_13_1
#define init_12_4_13_1
void FASTCALL call_12_4_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_13_2
#define init_12_4_13_2
void FASTCALL call_12_4_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_13_3
#define init_12_4_13_3
void FASTCALL call_12_4_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_13_4
#define init_12_4_13_4
void FASTCALL call_12_4_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_13_5
#define init_12_4_13_5
void FASTCALL call_12_4_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_13_6
#define init_12_4_13_6
void FASTCALL call_12_4_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_13_7
#define init_12_4_13_7
void FASTCALL call_12_4_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_13_8
#define init_12_4_13_8
void FASTCALL call_12_4_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_13_9
#define init_12_4_13_9
void FASTCALL call_12_4_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_13_10
#define init_12_4_13_10
void FASTCALL call_12_4_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_13_11
#define init_12_4_13_11
void FASTCALL call_12_4_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_13_12
#define init_12_4_13_12
void FASTCALL call_12_4_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_13_13
#define init_12_4_13_13
void FASTCALL call_12_4_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_13_14
#define init_12_4_13_14
void FASTCALL call_12_4_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_13_15
#define init_12_4_13_15
void FASTCALL call_12_4_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_14_0
#define init_12_4_14_0
void FASTCALL call_12_4_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_14_1
#define init_12_4_14_1
void FASTCALL call_12_4_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_14_2
#define init_12_4_14_2
void FASTCALL call_12_4_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_14_3
#define init_12_4_14_3
void FASTCALL call_12_4_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_14_4
#define init_12_4_14_4
void FASTCALL call_12_4_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_14_5
#define init_12_4_14_5
void FASTCALL call_12_4_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_14_6
#define init_12_4_14_6
void FASTCALL call_12_4_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_14_7
#define init_12_4_14_7
void FASTCALL call_12_4_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_14_8
#define init_12_4_14_8
void FASTCALL call_12_4_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_14_9
#define init_12_4_14_9
void FASTCALL call_12_4_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_14_10
#define init_12_4_14_10
void FASTCALL call_12_4_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_14_11
#define init_12_4_14_11
void FASTCALL call_12_4_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_14_12
#define init_12_4_14_12
void FASTCALL call_12_4_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_14_13
#define init_12_4_14_13
void FASTCALL call_12_4_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_14_14
#define init_12_4_14_14
void FASTCALL call_12_4_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_14_15
#define init_12_4_14_15
void FASTCALL call_12_4_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_15_0
#define init_12_4_15_0
void FASTCALL call_12_4_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_15_1
#define init_12_4_15_1
void FASTCALL call_12_4_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_15_2
#define init_12_4_15_2
void FASTCALL call_12_4_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_15_3
#define init_12_4_15_3
void FASTCALL call_12_4_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_15_4
#define init_12_4_15_4
void FASTCALL call_12_4_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_15_5
#define init_12_4_15_5
void FASTCALL call_12_4_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_15_6
#define init_12_4_15_6
void FASTCALL call_12_4_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_15_7
#define init_12_4_15_7
void FASTCALL call_12_4_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_15_8
#define init_12_4_15_8
void FASTCALL call_12_4_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_15_9
#define init_12_4_15_9
void FASTCALL call_12_4_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_15_10
#define init_12_4_15_10
void FASTCALL call_12_4_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_15_11
#define init_12_4_15_11
void FASTCALL call_12_4_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_15_12
#define init_12_4_15_12
void FASTCALL call_12_4_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_15_13
#define init_12_4_15_13
void FASTCALL call_12_4_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_15_14
#define init_12_4_15_14
void FASTCALL call_12_4_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_4_15_15
#define init_12_4_15_15
void FASTCALL call_12_4_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_0_0
#define init_12_5_0_0
void FASTCALL call_12_5_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_0_1
#define init_12_5_0_1
void FASTCALL call_12_5_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_0_2
#define init_12_5_0_2
void FASTCALL call_12_5_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_0_3
#define init_12_5_0_3
void FASTCALL call_12_5_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_0_4
#define init_12_5_0_4
void FASTCALL call_12_5_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_0_5
#define init_12_5_0_5
void FASTCALL call_12_5_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_0_6
#define init_12_5_0_6
void FASTCALL call_12_5_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_0_7
#define init_12_5_0_7
void FASTCALL call_12_5_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_0_8
#define init_12_5_0_8
void FASTCALL call_12_5_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_0_9
#define init_12_5_0_9
void FASTCALL call_12_5_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_0_10
#define init_12_5_0_10
void FASTCALL call_12_5_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_0_11
#define init_12_5_0_11
void FASTCALL call_12_5_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_0_12
#define init_12_5_0_12
void FASTCALL call_12_5_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_0_13
#define init_12_5_0_13
void FASTCALL call_12_5_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_0_14
#define init_12_5_0_14
void FASTCALL call_12_5_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_0_15
#define init_12_5_0_15
void FASTCALL call_12_5_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_1_0
#define init_12_5_1_0
void FASTCALL call_12_5_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_1_1
#define init_12_5_1_1
void FASTCALL call_12_5_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_1_2
#define init_12_5_1_2
void FASTCALL call_12_5_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_1_3
#define init_12_5_1_3
void FASTCALL call_12_5_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_1_4
#define init_12_5_1_4
void FASTCALL call_12_5_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_1_5
#define init_12_5_1_5
void FASTCALL call_12_5_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_1_6
#define init_12_5_1_6
void FASTCALL call_12_5_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_1_7
#define init_12_5_1_7
void FASTCALL call_12_5_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_1_8
#define init_12_5_1_8
void FASTCALL call_12_5_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_1_9
#define init_12_5_1_9
void FASTCALL call_12_5_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_1_10
#define init_12_5_1_10
void FASTCALL call_12_5_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_1_11
#define init_12_5_1_11
void FASTCALL call_12_5_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_1_12
#define init_12_5_1_12
void FASTCALL call_12_5_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_1_13
#define init_12_5_1_13
void FASTCALL call_12_5_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_1_14
#define init_12_5_1_14
void FASTCALL call_12_5_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_1_15
#define init_12_5_1_15
void FASTCALL call_12_5_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_2_0
#define init_12_5_2_0
void FASTCALL call_12_5_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_2_1
#define init_12_5_2_1
void FASTCALL call_12_5_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_2_2
#define init_12_5_2_2
void FASTCALL call_12_5_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_2_3
#define init_12_5_2_3
void FASTCALL call_12_5_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_2_4
#define init_12_5_2_4
void FASTCALL call_12_5_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_2_5
#define init_12_5_2_5
void FASTCALL call_12_5_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_2_6
#define init_12_5_2_6
void FASTCALL call_12_5_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_2_7
#define init_12_5_2_7
void FASTCALL call_12_5_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_2_8
#define init_12_5_2_8
void FASTCALL call_12_5_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_2_9
#define init_12_5_2_9
void FASTCALL call_12_5_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_2_10
#define init_12_5_2_10
void FASTCALL call_12_5_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_2_11
#define init_12_5_2_11
void FASTCALL call_12_5_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_2_12
#define init_12_5_2_12
void FASTCALL call_12_5_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_2_13
#define init_12_5_2_13
void FASTCALL call_12_5_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_2_14
#define init_12_5_2_14
void FASTCALL call_12_5_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_2_15
#define init_12_5_2_15
void FASTCALL call_12_5_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_3_0
#define init_12_5_3_0
void FASTCALL call_12_5_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_3_1
#define init_12_5_3_1
void FASTCALL call_12_5_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_3_2
#define init_12_5_3_2
void FASTCALL call_12_5_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_3_3
#define init_12_5_3_3
void FASTCALL call_12_5_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_3_4
#define init_12_5_3_4
void FASTCALL call_12_5_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_3_5
#define init_12_5_3_5
void FASTCALL call_12_5_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_3_6
#define init_12_5_3_6
void FASTCALL call_12_5_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_3_7
#define init_12_5_3_7
void FASTCALL call_12_5_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_3_8
#define init_12_5_3_8
void FASTCALL call_12_5_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_3_9
#define init_12_5_3_9
void FASTCALL call_12_5_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_3_10
#define init_12_5_3_10
void FASTCALL call_12_5_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_3_11
#define init_12_5_3_11
void FASTCALL call_12_5_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_3_12
#define init_12_5_3_12
void FASTCALL call_12_5_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_3_13
#define init_12_5_3_13
void FASTCALL call_12_5_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_3_14
#define init_12_5_3_14
void FASTCALL call_12_5_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_3_15
#define init_12_5_3_15
void FASTCALL call_12_5_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_4_0
#define init_12_5_4_0
void FASTCALL call_12_5_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_4_1
#define init_12_5_4_1
void FASTCALL call_12_5_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_4_2
#define init_12_5_4_2
void FASTCALL call_12_5_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_4_3
#define init_12_5_4_3
void FASTCALL call_12_5_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_4_4
#define init_12_5_4_4
void FASTCALL call_12_5_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_4_5
#define init_12_5_4_5
void FASTCALL call_12_5_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_4_6
#define init_12_5_4_6
void FASTCALL call_12_5_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_4_7
#define init_12_5_4_7
void FASTCALL call_12_5_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_4_8
#define init_12_5_4_8
void FASTCALL call_12_5_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_4_9
#define init_12_5_4_9
void FASTCALL call_12_5_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_4_10
#define init_12_5_4_10
void FASTCALL call_12_5_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_4_11
#define init_12_5_4_11
void FASTCALL call_12_5_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_4_12
#define init_12_5_4_12
void FASTCALL call_12_5_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_4_13
#define init_12_5_4_13
void FASTCALL call_12_5_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_4_14
#define init_12_5_4_14
void FASTCALL call_12_5_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_4_15
#define init_12_5_4_15
void FASTCALL call_12_5_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_5_0
#define init_12_5_5_0
void FASTCALL call_12_5_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_5_1
#define init_12_5_5_1
void FASTCALL call_12_5_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_5_2
#define init_12_5_5_2
void FASTCALL call_12_5_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_5_3
#define init_12_5_5_3
void FASTCALL call_12_5_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_5_4
#define init_12_5_5_4
void FASTCALL call_12_5_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_5_5
#define init_12_5_5_5
void FASTCALL call_12_5_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_5_6
#define init_12_5_5_6
void FASTCALL call_12_5_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_5_7
#define init_12_5_5_7
void FASTCALL call_12_5_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_5_8
#define init_12_5_5_8
void FASTCALL call_12_5_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_5_9
#define init_12_5_5_9
void FASTCALL call_12_5_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_5_10
#define init_12_5_5_10
void FASTCALL call_12_5_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_5_11
#define init_12_5_5_11
void FASTCALL call_12_5_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_5_12
#define init_12_5_5_12
void FASTCALL call_12_5_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_5_13
#define init_12_5_5_13
void FASTCALL call_12_5_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_5_14
#define init_12_5_5_14
void FASTCALL call_12_5_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_5_15
#define init_12_5_5_15
void FASTCALL call_12_5_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_6_0
#define init_12_5_6_0
void FASTCALL call_12_5_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_6_1
#define init_12_5_6_1
void FASTCALL call_12_5_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_6_2
#define init_12_5_6_2
void FASTCALL call_12_5_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_6_3
#define init_12_5_6_3
void FASTCALL call_12_5_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_6_4
#define init_12_5_6_4
void FASTCALL call_12_5_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_6_5
#define init_12_5_6_5
void FASTCALL call_12_5_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_6_6
#define init_12_5_6_6
void FASTCALL call_12_5_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_6_7
#define init_12_5_6_7
void FASTCALL call_12_5_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_6_8
#define init_12_5_6_8
void FASTCALL call_12_5_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_6_9
#define init_12_5_6_9
void FASTCALL call_12_5_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_6_10
#define init_12_5_6_10
void FASTCALL call_12_5_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_6_11
#define init_12_5_6_11
void FASTCALL call_12_5_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_6_12
#define init_12_5_6_12
void FASTCALL call_12_5_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_6_13
#define init_12_5_6_13
void FASTCALL call_12_5_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_6_14
#define init_12_5_6_14
void FASTCALL call_12_5_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_6_15
#define init_12_5_6_15
void FASTCALL call_12_5_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_7_0
#define init_12_5_7_0
void FASTCALL call_12_5_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_7_1
#define init_12_5_7_1
void FASTCALL call_12_5_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_7_2
#define init_12_5_7_2
void FASTCALL call_12_5_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_7_3
#define init_12_5_7_3
void FASTCALL call_12_5_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_7_4
#define init_12_5_7_4
void FASTCALL call_12_5_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_7_5
#define init_12_5_7_5
void FASTCALL call_12_5_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_7_6
#define init_12_5_7_6
void FASTCALL call_12_5_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_7_7
#define init_12_5_7_7
void FASTCALL call_12_5_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_7_8
#define init_12_5_7_8
void FASTCALL call_12_5_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_7_9
#define init_12_5_7_9
void FASTCALL call_12_5_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_7_10
#define init_12_5_7_10
void FASTCALL call_12_5_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_7_11
#define init_12_5_7_11
void FASTCALL call_12_5_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_7_12
#define init_12_5_7_12
void FASTCALL call_12_5_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_7_13
#define init_12_5_7_13
void FASTCALL call_12_5_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_7_14
#define init_12_5_7_14
void FASTCALL call_12_5_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_7_15
#define init_12_5_7_15
void FASTCALL call_12_5_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_8_0
#define init_12_5_8_0
void FASTCALL call_12_5_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_8_1
#define init_12_5_8_1
void FASTCALL call_12_5_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_8_2
#define init_12_5_8_2
void FASTCALL call_12_5_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_8_3
#define init_12_5_8_3
void FASTCALL call_12_5_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_8_4
#define init_12_5_8_4
void FASTCALL call_12_5_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_8_5
#define init_12_5_8_5
void FASTCALL call_12_5_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_8_6
#define init_12_5_8_6
void FASTCALL call_12_5_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_8_7
#define init_12_5_8_7
void FASTCALL call_12_5_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_8_8
#define init_12_5_8_8
void FASTCALL call_12_5_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_8_9
#define init_12_5_8_9
void FASTCALL call_12_5_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_8_10
#define init_12_5_8_10
void FASTCALL call_12_5_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_8_11
#define init_12_5_8_11
void FASTCALL call_12_5_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_8_12
#define init_12_5_8_12
void FASTCALL call_12_5_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_8_13
#define init_12_5_8_13
void FASTCALL call_12_5_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_8_14
#define init_12_5_8_14
void FASTCALL call_12_5_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_8_15
#define init_12_5_8_15
void FASTCALL call_12_5_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_9_0
#define init_12_5_9_0
void FASTCALL call_12_5_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_9_1
#define init_12_5_9_1
void FASTCALL call_12_5_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_9_2
#define init_12_5_9_2
void FASTCALL call_12_5_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_9_3
#define init_12_5_9_3
void FASTCALL call_12_5_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_9_4
#define init_12_5_9_4
void FASTCALL call_12_5_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_9_5
#define init_12_5_9_5
void FASTCALL call_12_5_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_9_6
#define init_12_5_9_6
void FASTCALL call_12_5_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_9_7
#define init_12_5_9_7
void FASTCALL call_12_5_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_9_8
#define init_12_5_9_8
void FASTCALL call_12_5_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_9_9
#define init_12_5_9_9
void FASTCALL call_12_5_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_9_10
#define init_12_5_9_10
void FASTCALL call_12_5_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_9_11
#define init_12_5_9_11
void FASTCALL call_12_5_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_9_12
#define init_12_5_9_12
void FASTCALL call_12_5_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_9_13
#define init_12_5_9_13
void FASTCALL call_12_5_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_9_14
#define init_12_5_9_14
void FASTCALL call_12_5_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_9_15
#define init_12_5_9_15
void FASTCALL call_12_5_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_10_0
#define init_12_5_10_0
void FASTCALL call_12_5_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_10_1
#define init_12_5_10_1
void FASTCALL call_12_5_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_10_2
#define init_12_5_10_2
void FASTCALL call_12_5_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_10_3
#define init_12_5_10_3
void FASTCALL call_12_5_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_10_4
#define init_12_5_10_4
void FASTCALL call_12_5_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_10_5
#define init_12_5_10_5
void FASTCALL call_12_5_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_10_6
#define init_12_5_10_6
void FASTCALL call_12_5_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_10_7
#define init_12_5_10_7
void FASTCALL call_12_5_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_10_8
#define init_12_5_10_8
void FASTCALL call_12_5_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_10_9
#define init_12_5_10_9
void FASTCALL call_12_5_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_10_10
#define init_12_5_10_10
void FASTCALL call_12_5_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_10_11
#define init_12_5_10_11
void FASTCALL call_12_5_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_10_12
#define init_12_5_10_12
void FASTCALL call_12_5_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_10_13
#define init_12_5_10_13
void FASTCALL call_12_5_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_10_14
#define init_12_5_10_14
void FASTCALL call_12_5_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_10_15
#define init_12_5_10_15
void FASTCALL call_12_5_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_11_0
#define init_12_5_11_0
void FASTCALL call_12_5_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_11_1
#define init_12_5_11_1
void FASTCALL call_12_5_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_11_2
#define init_12_5_11_2
void FASTCALL call_12_5_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_11_3
#define init_12_5_11_3
void FASTCALL call_12_5_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_11_4
#define init_12_5_11_4
void FASTCALL call_12_5_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_11_5
#define init_12_5_11_5
void FASTCALL call_12_5_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_11_6
#define init_12_5_11_6
void FASTCALL call_12_5_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_11_7
#define init_12_5_11_7
void FASTCALL call_12_5_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_11_8
#define init_12_5_11_8
void FASTCALL call_12_5_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_11_9
#define init_12_5_11_9
void FASTCALL call_12_5_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_11_10
#define init_12_5_11_10
void FASTCALL call_12_5_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_11_11
#define init_12_5_11_11
void FASTCALL call_12_5_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_11_12
#define init_12_5_11_12
void FASTCALL call_12_5_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_11_13
#define init_12_5_11_13
void FASTCALL call_12_5_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_11_14
#define init_12_5_11_14
void FASTCALL call_12_5_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_11_15
#define init_12_5_11_15
void FASTCALL call_12_5_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_12_0
#define init_12_5_12_0
void FASTCALL call_12_5_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_12_1
#define init_12_5_12_1
void FASTCALL call_12_5_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_12_2
#define init_12_5_12_2
void FASTCALL call_12_5_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_12_3
#define init_12_5_12_3
void FASTCALL call_12_5_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_12_4
#define init_12_5_12_4
void FASTCALL call_12_5_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_12_5
#define init_12_5_12_5
void FASTCALL call_12_5_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_12_6
#define init_12_5_12_6
void FASTCALL call_12_5_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_12_7
#define init_12_5_12_7
void FASTCALL call_12_5_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_12_8
#define init_12_5_12_8
void FASTCALL call_12_5_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_12_9
#define init_12_5_12_9
void FASTCALL call_12_5_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_12_10
#define init_12_5_12_10
void FASTCALL call_12_5_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_12_11
#define init_12_5_12_11
void FASTCALL call_12_5_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_12_12
#define init_12_5_12_12
void FASTCALL call_12_5_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_12_13
#define init_12_5_12_13
void FASTCALL call_12_5_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_12_14
#define init_12_5_12_14
void FASTCALL call_12_5_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_12_15
#define init_12_5_12_15
void FASTCALL call_12_5_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_13_0
#define init_12_5_13_0
void FASTCALL call_12_5_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_13_1
#define init_12_5_13_1
void FASTCALL call_12_5_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_13_2
#define init_12_5_13_2
void FASTCALL call_12_5_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_13_3
#define init_12_5_13_3
void FASTCALL call_12_5_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_13_4
#define init_12_5_13_4
void FASTCALL call_12_5_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_13_5
#define init_12_5_13_5
void FASTCALL call_12_5_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_13_6
#define init_12_5_13_6
void FASTCALL call_12_5_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_13_7
#define init_12_5_13_7
void FASTCALL call_12_5_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_13_8
#define init_12_5_13_8
void FASTCALL call_12_5_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_13_9
#define init_12_5_13_9
void FASTCALL call_12_5_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_13_10
#define init_12_5_13_10
void FASTCALL call_12_5_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_13_11
#define init_12_5_13_11
void FASTCALL call_12_5_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_13_12
#define init_12_5_13_12
void FASTCALL call_12_5_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_13_13
#define init_12_5_13_13
void FASTCALL call_12_5_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_13_14
#define init_12_5_13_14
void FASTCALL call_12_5_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_13_15
#define init_12_5_13_15
void FASTCALL call_12_5_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_14_0
#define init_12_5_14_0
void FASTCALL call_12_5_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_14_1
#define init_12_5_14_1
void FASTCALL call_12_5_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_14_2
#define init_12_5_14_2
void FASTCALL call_12_5_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_14_3
#define init_12_5_14_3
void FASTCALL call_12_5_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_14_4
#define init_12_5_14_4
void FASTCALL call_12_5_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_14_5
#define init_12_5_14_5
void FASTCALL call_12_5_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_14_6
#define init_12_5_14_6
void FASTCALL call_12_5_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_14_7
#define init_12_5_14_7
void FASTCALL call_12_5_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_14_8
#define init_12_5_14_8
void FASTCALL call_12_5_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_14_9
#define init_12_5_14_9
void FASTCALL call_12_5_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_14_10
#define init_12_5_14_10
void FASTCALL call_12_5_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_14_11
#define init_12_5_14_11
void FASTCALL call_12_5_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_14_12
#define init_12_5_14_12
void FASTCALL call_12_5_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_14_13
#define init_12_5_14_13
void FASTCALL call_12_5_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_14_14
#define init_12_5_14_14
void FASTCALL call_12_5_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_14_15
#define init_12_5_14_15
void FASTCALL call_12_5_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_15_0
#define init_12_5_15_0
void FASTCALL call_12_5_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_15_1
#define init_12_5_15_1
void FASTCALL call_12_5_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_15_2
#define init_12_5_15_2
void FASTCALL call_12_5_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_15_3
#define init_12_5_15_3
void FASTCALL call_12_5_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_15_4
#define init_12_5_15_4
void FASTCALL call_12_5_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_15_5
#define init_12_5_15_5
void FASTCALL call_12_5_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_15_6
#define init_12_5_15_6
void FASTCALL call_12_5_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_15_7
#define init_12_5_15_7
void FASTCALL call_12_5_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_15_8
#define init_12_5_15_8
void FASTCALL call_12_5_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_15_9
#define init_12_5_15_9
void FASTCALL call_12_5_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_15_10
#define init_12_5_15_10
void FASTCALL call_12_5_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_15_11
#define init_12_5_15_11
void FASTCALL call_12_5_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_15_12
#define init_12_5_15_12
void FASTCALL call_12_5_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_15_13
#define init_12_5_15_13
void FASTCALL call_12_5_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_15_14
#define init_12_5_15_14
void FASTCALL call_12_5_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_5_15_15
#define init_12_5_15_15
void FASTCALL call_12_5_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_0_0
#define init_12_6_0_0
void FASTCALL call_12_6_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_0_1
#define init_12_6_0_1
void FASTCALL call_12_6_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_0_2
#define init_12_6_0_2
void FASTCALL call_12_6_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_0_3
#define init_12_6_0_3
void FASTCALL call_12_6_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_0_4
#define init_12_6_0_4
void FASTCALL call_12_6_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_0_5
#define init_12_6_0_5
void FASTCALL call_12_6_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_0_6
#define init_12_6_0_6
void FASTCALL call_12_6_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_0_7
#define init_12_6_0_7
void FASTCALL call_12_6_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_0_8
#define init_12_6_0_8
void FASTCALL call_12_6_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_0_9
#define init_12_6_0_9
void FASTCALL call_12_6_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_0_10
#define init_12_6_0_10
void FASTCALL call_12_6_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_0_11
#define init_12_6_0_11
void FASTCALL call_12_6_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_0_12
#define init_12_6_0_12
void FASTCALL call_12_6_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_0_13
#define init_12_6_0_13
void FASTCALL call_12_6_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_0_14
#define init_12_6_0_14
void FASTCALL call_12_6_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_0_15
#define init_12_6_0_15
void FASTCALL call_12_6_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_1_0
#define init_12_6_1_0
void FASTCALL call_12_6_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_1_1
#define init_12_6_1_1
void FASTCALL call_12_6_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_1_2
#define init_12_6_1_2
void FASTCALL call_12_6_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_1_3
#define init_12_6_1_3
void FASTCALL call_12_6_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_1_4
#define init_12_6_1_4
void FASTCALL call_12_6_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_1_5
#define init_12_6_1_5
void FASTCALL call_12_6_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_1_6
#define init_12_6_1_6
void FASTCALL call_12_6_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_1_7
#define init_12_6_1_7
void FASTCALL call_12_6_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_1_8
#define init_12_6_1_8
void FASTCALL call_12_6_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_1_9
#define init_12_6_1_9
void FASTCALL call_12_6_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_1_10
#define init_12_6_1_10
void FASTCALL call_12_6_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_1_11
#define init_12_6_1_11
void FASTCALL call_12_6_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_1_12
#define init_12_6_1_12
void FASTCALL call_12_6_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_1_13
#define init_12_6_1_13
void FASTCALL call_12_6_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_1_14
#define init_12_6_1_14
void FASTCALL call_12_6_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_1_15
#define init_12_6_1_15
void FASTCALL call_12_6_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_2_0
#define init_12_6_2_0
void FASTCALL call_12_6_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_2_1
#define init_12_6_2_1
void FASTCALL call_12_6_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_2_2
#define init_12_6_2_2
void FASTCALL call_12_6_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_2_3
#define init_12_6_2_3
void FASTCALL call_12_6_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_2_4
#define init_12_6_2_4
void FASTCALL call_12_6_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_2_5
#define init_12_6_2_5
void FASTCALL call_12_6_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_2_6
#define init_12_6_2_6
void FASTCALL call_12_6_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_2_7
#define init_12_6_2_7
void FASTCALL call_12_6_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_2_8
#define init_12_6_2_8
void FASTCALL call_12_6_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_2_9
#define init_12_6_2_9
void FASTCALL call_12_6_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_2_10
#define init_12_6_2_10
void FASTCALL call_12_6_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_2_11
#define init_12_6_2_11
void FASTCALL call_12_6_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_2_12
#define init_12_6_2_12
void FASTCALL call_12_6_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_2_13
#define init_12_6_2_13
void FASTCALL call_12_6_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_2_14
#define init_12_6_2_14
void FASTCALL call_12_6_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_2_15
#define init_12_6_2_15
void FASTCALL call_12_6_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_3_0
#define init_12_6_3_0
void FASTCALL call_12_6_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_3_1
#define init_12_6_3_1
void FASTCALL call_12_6_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_3_2
#define init_12_6_3_2
void FASTCALL call_12_6_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_3_3
#define init_12_6_3_3
void FASTCALL call_12_6_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_3_4
#define init_12_6_3_4
void FASTCALL call_12_6_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_3_5
#define init_12_6_3_5
void FASTCALL call_12_6_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_3_6
#define init_12_6_3_6
void FASTCALL call_12_6_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_3_7
#define init_12_6_3_7
void FASTCALL call_12_6_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_3_8
#define init_12_6_3_8
void FASTCALL call_12_6_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_3_9
#define init_12_6_3_9
void FASTCALL call_12_6_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_3_10
#define init_12_6_3_10
void FASTCALL call_12_6_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_3_11
#define init_12_6_3_11
void FASTCALL call_12_6_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_3_12
#define init_12_6_3_12
void FASTCALL call_12_6_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_3_13
#define init_12_6_3_13
void FASTCALL call_12_6_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_3_14
#define init_12_6_3_14
void FASTCALL call_12_6_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_3_15
#define init_12_6_3_15
void FASTCALL call_12_6_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_4_0
#define init_12_6_4_0
void FASTCALL call_12_6_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_4_1
#define init_12_6_4_1
void FASTCALL call_12_6_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_4_2
#define init_12_6_4_2
void FASTCALL call_12_6_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_4_3
#define init_12_6_4_3
void FASTCALL call_12_6_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_4_4
#define init_12_6_4_4
void FASTCALL call_12_6_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_4_5
#define init_12_6_4_5
void FASTCALL call_12_6_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_4_6
#define init_12_6_4_6
void FASTCALL call_12_6_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_4_7
#define init_12_6_4_7
void FASTCALL call_12_6_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_4_8
#define init_12_6_4_8
void FASTCALL call_12_6_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_4_9
#define init_12_6_4_9
void FASTCALL call_12_6_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_4_10
#define init_12_6_4_10
void FASTCALL call_12_6_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_4_11
#define init_12_6_4_11
void FASTCALL call_12_6_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_4_12
#define init_12_6_4_12
void FASTCALL call_12_6_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_4_13
#define init_12_6_4_13
void FASTCALL call_12_6_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_4_14
#define init_12_6_4_14
void FASTCALL call_12_6_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_4_15
#define init_12_6_4_15
void FASTCALL call_12_6_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_5_0
#define init_12_6_5_0
void FASTCALL call_12_6_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_5_1
#define init_12_6_5_1
void FASTCALL call_12_6_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_5_2
#define init_12_6_5_2
void FASTCALL call_12_6_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_5_3
#define init_12_6_5_3
void FASTCALL call_12_6_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_5_4
#define init_12_6_5_4
void FASTCALL call_12_6_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_5_5
#define init_12_6_5_5
void FASTCALL call_12_6_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_5_6
#define init_12_6_5_6
void FASTCALL call_12_6_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_5_7
#define init_12_6_5_7
void FASTCALL call_12_6_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_5_8
#define init_12_6_5_8
void FASTCALL call_12_6_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_5_9
#define init_12_6_5_9
void FASTCALL call_12_6_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_5_10
#define init_12_6_5_10
void FASTCALL call_12_6_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_5_11
#define init_12_6_5_11
void FASTCALL call_12_6_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_5_12
#define init_12_6_5_12
void FASTCALL call_12_6_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_5_13
#define init_12_6_5_13
void FASTCALL call_12_6_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_5_14
#define init_12_6_5_14
void FASTCALL call_12_6_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_5_15
#define init_12_6_5_15
void FASTCALL call_12_6_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_6_0
#define init_12_6_6_0
void FASTCALL call_12_6_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_6_1
#define init_12_6_6_1
void FASTCALL call_12_6_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_6_2
#define init_12_6_6_2
void FASTCALL call_12_6_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_6_3
#define init_12_6_6_3
void FASTCALL call_12_6_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_6_4
#define init_12_6_6_4
void FASTCALL call_12_6_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_6_5
#define init_12_6_6_5
void FASTCALL call_12_6_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_6_6
#define init_12_6_6_6
void FASTCALL call_12_6_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_6_7
#define init_12_6_6_7
void FASTCALL call_12_6_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_6_8
#define init_12_6_6_8
void FASTCALL call_12_6_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_6_9
#define init_12_6_6_9
void FASTCALL call_12_6_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_6_10
#define init_12_6_6_10
void FASTCALL call_12_6_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_6_11
#define init_12_6_6_11
void FASTCALL call_12_6_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_6_12
#define init_12_6_6_12
void FASTCALL call_12_6_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_6_13
#define init_12_6_6_13
void FASTCALL call_12_6_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_6_14
#define init_12_6_6_14
void FASTCALL call_12_6_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_6_15
#define init_12_6_6_15
void FASTCALL call_12_6_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_7_0
#define init_12_6_7_0
void FASTCALL call_12_6_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_7_1
#define init_12_6_7_1
void FASTCALL call_12_6_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_7_2
#define init_12_6_7_2
void FASTCALL call_12_6_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_7_3
#define init_12_6_7_3
void FASTCALL call_12_6_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_7_4
#define init_12_6_7_4
void FASTCALL call_12_6_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_7_5
#define init_12_6_7_5
void FASTCALL call_12_6_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_7_6
#define init_12_6_7_6
void FASTCALL call_12_6_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_7_7
#define init_12_6_7_7
void FASTCALL call_12_6_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_7_8
#define init_12_6_7_8
void FASTCALL call_12_6_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_7_9
#define init_12_6_7_9
void FASTCALL call_12_6_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_7_10
#define init_12_6_7_10
void FASTCALL call_12_6_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_7_11
#define init_12_6_7_11
void FASTCALL call_12_6_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_7_12
#define init_12_6_7_12
void FASTCALL call_12_6_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_7_13
#define init_12_6_7_13
void FASTCALL call_12_6_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_7_14
#define init_12_6_7_14
void FASTCALL call_12_6_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_7_15
#define init_12_6_7_15
void FASTCALL call_12_6_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_8_0
#define init_12_6_8_0
void FASTCALL call_12_6_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_8_1
#define init_12_6_8_1
void FASTCALL call_12_6_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_8_2
#define init_12_6_8_2
void FASTCALL call_12_6_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_8_3
#define init_12_6_8_3
void FASTCALL call_12_6_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_8_4
#define init_12_6_8_4
void FASTCALL call_12_6_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_8_5
#define init_12_6_8_5
void FASTCALL call_12_6_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_8_6
#define init_12_6_8_6
void FASTCALL call_12_6_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_8_7
#define init_12_6_8_7
void FASTCALL call_12_6_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_8_8
#define init_12_6_8_8
void FASTCALL call_12_6_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_8_9
#define init_12_6_8_9
void FASTCALL call_12_6_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_8_10
#define init_12_6_8_10
void FASTCALL call_12_6_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_8_11
#define init_12_6_8_11
void FASTCALL call_12_6_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_8_12
#define init_12_6_8_12
void FASTCALL call_12_6_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_8_13
#define init_12_6_8_13
void FASTCALL call_12_6_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_8_14
#define init_12_6_8_14
void FASTCALL call_12_6_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_8_15
#define init_12_6_8_15
void FASTCALL call_12_6_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_9_0
#define init_12_6_9_0
void FASTCALL call_12_6_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_9_1
#define init_12_6_9_1
void FASTCALL call_12_6_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_9_2
#define init_12_6_9_2
void FASTCALL call_12_6_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_9_3
#define init_12_6_9_3
void FASTCALL call_12_6_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_9_4
#define init_12_6_9_4
void FASTCALL call_12_6_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_9_5
#define init_12_6_9_5
void FASTCALL call_12_6_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_9_6
#define init_12_6_9_6
void FASTCALL call_12_6_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_9_7
#define init_12_6_9_7
void FASTCALL call_12_6_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_9_8
#define init_12_6_9_8
void FASTCALL call_12_6_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_9_9
#define init_12_6_9_9
void FASTCALL call_12_6_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_9_10
#define init_12_6_9_10
void FASTCALL call_12_6_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_9_11
#define init_12_6_9_11
void FASTCALL call_12_6_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_9_12
#define init_12_6_9_12
void FASTCALL call_12_6_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_9_13
#define init_12_6_9_13
void FASTCALL call_12_6_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_9_14
#define init_12_6_9_14
void FASTCALL call_12_6_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_9_15
#define init_12_6_9_15
void FASTCALL call_12_6_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_10_0
#define init_12_6_10_0
void FASTCALL call_12_6_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_10_1
#define init_12_6_10_1
void FASTCALL call_12_6_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_10_2
#define init_12_6_10_2
void FASTCALL call_12_6_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_10_3
#define init_12_6_10_3
void FASTCALL call_12_6_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_10_4
#define init_12_6_10_4
void FASTCALL call_12_6_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_10_5
#define init_12_6_10_5
void FASTCALL call_12_6_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_10_6
#define init_12_6_10_6
void FASTCALL call_12_6_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_10_7
#define init_12_6_10_7
void FASTCALL call_12_6_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_10_8
#define init_12_6_10_8
void FASTCALL call_12_6_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_10_9
#define init_12_6_10_9
void FASTCALL call_12_6_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_10_10
#define init_12_6_10_10
void FASTCALL call_12_6_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_10_11
#define init_12_6_10_11
void FASTCALL call_12_6_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_10_12
#define init_12_6_10_12
void FASTCALL call_12_6_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_10_13
#define init_12_6_10_13
void FASTCALL call_12_6_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_10_14
#define init_12_6_10_14
void FASTCALL call_12_6_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_10_15
#define init_12_6_10_15
void FASTCALL call_12_6_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_11_0
#define init_12_6_11_0
void FASTCALL call_12_6_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_11_1
#define init_12_6_11_1
void FASTCALL call_12_6_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_11_2
#define init_12_6_11_2
void FASTCALL call_12_6_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_11_3
#define init_12_6_11_3
void FASTCALL call_12_6_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_11_4
#define init_12_6_11_4
void FASTCALL call_12_6_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_11_5
#define init_12_6_11_5
void FASTCALL call_12_6_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_11_6
#define init_12_6_11_6
void FASTCALL call_12_6_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_11_7
#define init_12_6_11_7
void FASTCALL call_12_6_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_11_8
#define init_12_6_11_8
void FASTCALL call_12_6_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_11_9
#define init_12_6_11_9
void FASTCALL call_12_6_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_11_10
#define init_12_6_11_10
void FASTCALL call_12_6_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_11_11
#define init_12_6_11_11
void FASTCALL call_12_6_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_11_12
#define init_12_6_11_12
void FASTCALL call_12_6_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_11_13
#define init_12_6_11_13
void FASTCALL call_12_6_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_11_14
#define init_12_6_11_14
void FASTCALL call_12_6_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_11_15
#define init_12_6_11_15
void FASTCALL call_12_6_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_12_0
#define init_12_6_12_0
void FASTCALL call_12_6_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_12_1
#define init_12_6_12_1
void FASTCALL call_12_6_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_12_2
#define init_12_6_12_2
void FASTCALL call_12_6_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_12_3
#define init_12_6_12_3
void FASTCALL call_12_6_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_12_4
#define init_12_6_12_4
void FASTCALL call_12_6_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_12_5
#define init_12_6_12_5
void FASTCALL call_12_6_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_12_6
#define init_12_6_12_6
void FASTCALL call_12_6_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_12_7
#define init_12_6_12_7
void FASTCALL call_12_6_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_12_8
#define init_12_6_12_8
void FASTCALL call_12_6_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_12_9
#define init_12_6_12_9
void FASTCALL call_12_6_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_12_10
#define init_12_6_12_10
void FASTCALL call_12_6_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_12_11
#define init_12_6_12_11
void FASTCALL call_12_6_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_12_12
#define init_12_6_12_12
void FASTCALL call_12_6_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_12_13
#define init_12_6_12_13
void FASTCALL call_12_6_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_12_14
#define init_12_6_12_14
void FASTCALL call_12_6_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_12_15
#define init_12_6_12_15
void FASTCALL call_12_6_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_13_0
#define init_12_6_13_0
void FASTCALL call_12_6_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_13_1
#define init_12_6_13_1
void FASTCALL call_12_6_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_13_2
#define init_12_6_13_2
void FASTCALL call_12_6_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_13_3
#define init_12_6_13_3
void FASTCALL call_12_6_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_13_4
#define init_12_6_13_4
void FASTCALL call_12_6_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_13_5
#define init_12_6_13_5
void FASTCALL call_12_6_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_13_6
#define init_12_6_13_6
void FASTCALL call_12_6_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_13_7
#define init_12_6_13_7
void FASTCALL call_12_6_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_13_8
#define init_12_6_13_8
void FASTCALL call_12_6_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_13_9
#define init_12_6_13_9
void FASTCALL call_12_6_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_13_10
#define init_12_6_13_10
void FASTCALL call_12_6_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_13_11
#define init_12_6_13_11
void FASTCALL call_12_6_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_13_12
#define init_12_6_13_12
void FASTCALL call_12_6_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_13_13
#define init_12_6_13_13
void FASTCALL call_12_6_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_13_14
#define init_12_6_13_14
void FASTCALL call_12_6_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_13_15
#define init_12_6_13_15
void FASTCALL call_12_6_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_14_0
#define init_12_6_14_0
void FASTCALL call_12_6_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_14_1
#define init_12_6_14_1
void FASTCALL call_12_6_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_14_2
#define init_12_6_14_2
void FASTCALL call_12_6_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_14_3
#define init_12_6_14_3
void FASTCALL call_12_6_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_14_4
#define init_12_6_14_4
void FASTCALL call_12_6_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_14_5
#define init_12_6_14_5
void FASTCALL call_12_6_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_14_6
#define init_12_6_14_6
void FASTCALL call_12_6_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_14_7
#define init_12_6_14_7
void FASTCALL call_12_6_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_14_8
#define init_12_6_14_8
void FASTCALL call_12_6_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_14_9
#define init_12_6_14_9
void FASTCALL call_12_6_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_14_10
#define init_12_6_14_10
void FASTCALL call_12_6_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_14_11
#define init_12_6_14_11
void FASTCALL call_12_6_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_14_12
#define init_12_6_14_12
void FASTCALL call_12_6_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_14_13
#define init_12_6_14_13
void FASTCALL call_12_6_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_14_14
#define init_12_6_14_14
void FASTCALL call_12_6_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_14_15
#define init_12_6_14_15
void FASTCALL call_12_6_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_15_0
#define init_12_6_15_0
void FASTCALL call_12_6_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_15_1
#define init_12_6_15_1
void FASTCALL call_12_6_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_15_2
#define init_12_6_15_2
void FASTCALL call_12_6_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_15_3
#define init_12_6_15_3
void FASTCALL call_12_6_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_15_4
#define init_12_6_15_4
void FASTCALL call_12_6_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_15_5
#define init_12_6_15_5
void FASTCALL call_12_6_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_15_6
#define init_12_6_15_6
void FASTCALL call_12_6_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_15_7
#define init_12_6_15_7
void FASTCALL call_12_6_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_15_8
#define init_12_6_15_8
void FASTCALL call_12_6_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_15_9
#define init_12_6_15_9
void FASTCALL call_12_6_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_15_10
#define init_12_6_15_10
void FASTCALL call_12_6_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_15_11
#define init_12_6_15_11
void FASTCALL call_12_6_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_15_12
#define init_12_6_15_12
void FASTCALL call_12_6_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_15_13
#define init_12_6_15_13
void FASTCALL call_12_6_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_15_14
#define init_12_6_15_14
void FASTCALL call_12_6_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_6_15_15
#define init_12_6_15_15
void FASTCALL call_12_6_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_0_0
#define init_12_7_0_0
void FASTCALL call_12_7_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_0_1
#define init_12_7_0_1
void FASTCALL call_12_7_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_0_2
#define init_12_7_0_2
void FASTCALL call_12_7_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_0_3
#define init_12_7_0_3
void FASTCALL call_12_7_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_0_4
#define init_12_7_0_4
void FASTCALL call_12_7_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_0_5
#define init_12_7_0_5
void FASTCALL call_12_7_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_0_6
#define init_12_7_0_6
void FASTCALL call_12_7_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_0_7
#define init_12_7_0_7
void FASTCALL call_12_7_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_0_8
#define init_12_7_0_8
void FASTCALL call_12_7_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_0_9
#define init_12_7_0_9
void FASTCALL call_12_7_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_0_10
#define init_12_7_0_10
void FASTCALL call_12_7_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_0_11
#define init_12_7_0_11
void FASTCALL call_12_7_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_0_12
#define init_12_7_0_12
void FASTCALL call_12_7_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_0_13
#define init_12_7_0_13
void FASTCALL call_12_7_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_0_14
#define init_12_7_0_14
void FASTCALL call_12_7_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_0_15
#define init_12_7_0_15
void FASTCALL call_12_7_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_1_0
#define init_12_7_1_0
void FASTCALL call_12_7_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_1_1
#define init_12_7_1_1
void FASTCALL call_12_7_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_1_2
#define init_12_7_1_2
void FASTCALL call_12_7_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_1_3
#define init_12_7_1_3
void FASTCALL call_12_7_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_1_4
#define init_12_7_1_4
void FASTCALL call_12_7_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_1_5
#define init_12_7_1_5
void FASTCALL call_12_7_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_1_6
#define init_12_7_1_6
void FASTCALL call_12_7_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_1_7
#define init_12_7_1_7
void FASTCALL call_12_7_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_1_8
#define init_12_7_1_8
void FASTCALL call_12_7_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_1_9
#define init_12_7_1_9
void FASTCALL call_12_7_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_1_10
#define init_12_7_1_10
void FASTCALL call_12_7_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_1_11
#define init_12_7_1_11
void FASTCALL call_12_7_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_1_12
#define init_12_7_1_12
void FASTCALL call_12_7_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_1_13
#define init_12_7_1_13
void FASTCALL call_12_7_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_1_14
#define init_12_7_1_14
void FASTCALL call_12_7_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_1_15
#define init_12_7_1_15
void FASTCALL call_12_7_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_2_0
#define init_12_7_2_0
void FASTCALL call_12_7_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_2_1
#define init_12_7_2_1
void FASTCALL call_12_7_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_2_2
#define init_12_7_2_2
void FASTCALL call_12_7_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_2_3
#define init_12_7_2_3
void FASTCALL call_12_7_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_2_4
#define init_12_7_2_4
void FASTCALL call_12_7_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_2_5
#define init_12_7_2_5
void FASTCALL call_12_7_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_2_6
#define init_12_7_2_6
void FASTCALL call_12_7_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_2_7
#define init_12_7_2_7
void FASTCALL call_12_7_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_2_8
#define init_12_7_2_8
void FASTCALL call_12_7_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_2_9
#define init_12_7_2_9
void FASTCALL call_12_7_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_2_10
#define init_12_7_2_10
void FASTCALL call_12_7_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_2_11
#define init_12_7_2_11
void FASTCALL call_12_7_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_2_12
#define init_12_7_2_12
void FASTCALL call_12_7_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_2_13
#define init_12_7_2_13
void FASTCALL call_12_7_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_2_14
#define init_12_7_2_14
void FASTCALL call_12_7_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_2_15
#define init_12_7_2_15
void FASTCALL call_12_7_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_3_0
#define init_12_7_3_0
void FASTCALL call_12_7_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_3_1
#define init_12_7_3_1
void FASTCALL call_12_7_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_3_2
#define init_12_7_3_2
void FASTCALL call_12_7_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_3_3
#define init_12_7_3_3
void FASTCALL call_12_7_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_3_4
#define init_12_7_3_4
void FASTCALL call_12_7_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_3_5
#define init_12_7_3_5
void FASTCALL call_12_7_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_3_6
#define init_12_7_3_6
void FASTCALL call_12_7_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_3_7
#define init_12_7_3_7
void FASTCALL call_12_7_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_3_8
#define init_12_7_3_8
void FASTCALL call_12_7_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_3_9
#define init_12_7_3_9
void FASTCALL call_12_7_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_3_10
#define init_12_7_3_10
void FASTCALL call_12_7_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_3_11
#define init_12_7_3_11
void FASTCALL call_12_7_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_3_12
#define init_12_7_3_12
void FASTCALL call_12_7_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_3_13
#define init_12_7_3_13
void FASTCALL call_12_7_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_3_14
#define init_12_7_3_14
void FASTCALL call_12_7_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_3_15
#define init_12_7_3_15
void FASTCALL call_12_7_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_4_0
#define init_12_7_4_0
void FASTCALL call_12_7_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_4_1
#define init_12_7_4_1
void FASTCALL call_12_7_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_4_2
#define init_12_7_4_2
void FASTCALL call_12_7_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_4_3
#define init_12_7_4_3
void FASTCALL call_12_7_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_4_4
#define init_12_7_4_4
void FASTCALL call_12_7_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_4_5
#define init_12_7_4_5
void FASTCALL call_12_7_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_4_6
#define init_12_7_4_6
void FASTCALL call_12_7_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_4_7
#define init_12_7_4_7
void FASTCALL call_12_7_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_4_8
#define init_12_7_4_8
void FASTCALL call_12_7_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_4_9
#define init_12_7_4_9
void FASTCALL call_12_7_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_4_10
#define init_12_7_4_10
void FASTCALL call_12_7_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_4_11
#define init_12_7_4_11
void FASTCALL call_12_7_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_4_12
#define init_12_7_4_12
void FASTCALL call_12_7_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_4_13
#define init_12_7_4_13
void FASTCALL call_12_7_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_4_14
#define init_12_7_4_14
void FASTCALL call_12_7_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_4_15
#define init_12_7_4_15
void FASTCALL call_12_7_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_5_0
#define init_12_7_5_0
void FASTCALL call_12_7_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_5_1
#define init_12_7_5_1
void FASTCALL call_12_7_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_5_2
#define init_12_7_5_2
void FASTCALL call_12_7_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_5_3
#define init_12_7_5_3
void FASTCALL call_12_7_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_5_4
#define init_12_7_5_4
void FASTCALL call_12_7_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_5_5
#define init_12_7_5_5
void FASTCALL call_12_7_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_5_6
#define init_12_7_5_6
void FASTCALL call_12_7_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_5_7
#define init_12_7_5_7
void FASTCALL call_12_7_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_5_8
#define init_12_7_5_8
void FASTCALL call_12_7_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_5_9
#define init_12_7_5_9
void FASTCALL call_12_7_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_5_10
#define init_12_7_5_10
void FASTCALL call_12_7_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_5_11
#define init_12_7_5_11
void FASTCALL call_12_7_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_5_12
#define init_12_7_5_12
void FASTCALL call_12_7_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_5_13
#define init_12_7_5_13
void FASTCALL call_12_7_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_5_14
#define init_12_7_5_14
void FASTCALL call_12_7_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_5_15
#define init_12_7_5_15
void FASTCALL call_12_7_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_6_0
#define init_12_7_6_0
void FASTCALL call_12_7_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_6_1
#define init_12_7_6_1
void FASTCALL call_12_7_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_6_2
#define init_12_7_6_2
void FASTCALL call_12_7_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_6_3
#define init_12_7_6_3
void FASTCALL call_12_7_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_6_4
#define init_12_7_6_4
void FASTCALL call_12_7_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_6_5
#define init_12_7_6_5
void FASTCALL call_12_7_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_6_6
#define init_12_7_6_6
void FASTCALL call_12_7_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_6_7
#define init_12_7_6_7
void FASTCALL call_12_7_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_6_8
#define init_12_7_6_8
void FASTCALL call_12_7_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_6_9
#define init_12_7_6_9
void FASTCALL call_12_7_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_6_10
#define init_12_7_6_10
void FASTCALL call_12_7_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_6_11
#define init_12_7_6_11
void FASTCALL call_12_7_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_6_12
#define init_12_7_6_12
void FASTCALL call_12_7_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_6_13
#define init_12_7_6_13
void FASTCALL call_12_7_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_6_14
#define init_12_7_6_14
void FASTCALL call_12_7_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_6_15
#define init_12_7_6_15
void FASTCALL call_12_7_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_7_0
#define init_12_7_7_0
void FASTCALL call_12_7_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_7_1
#define init_12_7_7_1
void FASTCALL call_12_7_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_7_2
#define init_12_7_7_2
void FASTCALL call_12_7_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_7_3
#define init_12_7_7_3
void FASTCALL call_12_7_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_7_4
#define init_12_7_7_4
void FASTCALL call_12_7_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_7_5
#define init_12_7_7_5
void FASTCALL call_12_7_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_7_6
#define init_12_7_7_6
void FASTCALL call_12_7_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_7_7
#define init_12_7_7_7
void FASTCALL call_12_7_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_7_8
#define init_12_7_7_8
void FASTCALL call_12_7_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_7_9
#define init_12_7_7_9
void FASTCALL call_12_7_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_7_10
#define init_12_7_7_10
void FASTCALL call_12_7_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_7_11
#define init_12_7_7_11
void FASTCALL call_12_7_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_7_12
#define init_12_7_7_12
void FASTCALL call_12_7_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_7_13
#define init_12_7_7_13
void FASTCALL call_12_7_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_7_14
#define init_12_7_7_14
void FASTCALL call_12_7_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_7_15
#define init_12_7_7_15
void FASTCALL call_12_7_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_8_0
#define init_12_7_8_0
void FASTCALL call_12_7_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_8_1
#define init_12_7_8_1
void FASTCALL call_12_7_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_8_2
#define init_12_7_8_2
void FASTCALL call_12_7_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_8_3
#define init_12_7_8_3
void FASTCALL call_12_7_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_8_4
#define init_12_7_8_4
void FASTCALL call_12_7_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_8_5
#define init_12_7_8_5
void FASTCALL call_12_7_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_8_6
#define init_12_7_8_6
void FASTCALL call_12_7_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_8_7
#define init_12_7_8_7
void FASTCALL call_12_7_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_8_8
#define init_12_7_8_8
void FASTCALL call_12_7_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_8_9
#define init_12_7_8_9
void FASTCALL call_12_7_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_8_10
#define init_12_7_8_10
void FASTCALL call_12_7_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_8_11
#define init_12_7_8_11
void FASTCALL call_12_7_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_8_12
#define init_12_7_8_12
void FASTCALL call_12_7_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_8_13
#define init_12_7_8_13
void FASTCALL call_12_7_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_8_14
#define init_12_7_8_14
void FASTCALL call_12_7_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_8_15
#define init_12_7_8_15
void FASTCALL call_12_7_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_9_0
#define init_12_7_9_0
void FASTCALL call_12_7_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_9_1
#define init_12_7_9_1
void FASTCALL call_12_7_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_9_2
#define init_12_7_9_2
void FASTCALL call_12_7_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_9_3
#define init_12_7_9_3
void FASTCALL call_12_7_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_9_4
#define init_12_7_9_4
void FASTCALL call_12_7_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_9_5
#define init_12_7_9_5
void FASTCALL call_12_7_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_9_6
#define init_12_7_9_6
void FASTCALL call_12_7_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_9_7
#define init_12_7_9_7
void FASTCALL call_12_7_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_9_8
#define init_12_7_9_8
void FASTCALL call_12_7_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_9_9
#define init_12_7_9_9
void FASTCALL call_12_7_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_9_10
#define init_12_7_9_10
void FASTCALL call_12_7_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_9_11
#define init_12_7_9_11
void FASTCALL call_12_7_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_9_12
#define init_12_7_9_12
void FASTCALL call_12_7_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_9_13
#define init_12_7_9_13
void FASTCALL call_12_7_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_9_14
#define init_12_7_9_14
void FASTCALL call_12_7_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_9_15
#define init_12_7_9_15
void FASTCALL call_12_7_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_10_0
#define init_12_7_10_0
void FASTCALL call_12_7_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_10_1
#define init_12_7_10_1
void FASTCALL call_12_7_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_10_2
#define init_12_7_10_2
void FASTCALL call_12_7_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_10_3
#define init_12_7_10_3
void FASTCALL call_12_7_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_10_4
#define init_12_7_10_4
void FASTCALL call_12_7_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_10_5
#define init_12_7_10_5
void FASTCALL call_12_7_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_10_6
#define init_12_7_10_6
void FASTCALL call_12_7_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_10_7
#define init_12_7_10_7
void FASTCALL call_12_7_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_10_8
#define init_12_7_10_8
void FASTCALL call_12_7_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_10_9
#define init_12_7_10_9
void FASTCALL call_12_7_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_10_10
#define init_12_7_10_10
void FASTCALL call_12_7_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_10_11
#define init_12_7_10_11
void FASTCALL call_12_7_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_10_12
#define init_12_7_10_12
void FASTCALL call_12_7_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_10_13
#define init_12_7_10_13
void FASTCALL call_12_7_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_10_14
#define init_12_7_10_14
void FASTCALL call_12_7_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_10_15
#define init_12_7_10_15
void FASTCALL call_12_7_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_11_0
#define init_12_7_11_0
void FASTCALL call_12_7_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_11_1
#define init_12_7_11_1
void FASTCALL call_12_7_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_11_2
#define init_12_7_11_2
void FASTCALL call_12_7_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_11_3
#define init_12_7_11_3
void FASTCALL call_12_7_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_11_4
#define init_12_7_11_4
void FASTCALL call_12_7_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_11_5
#define init_12_7_11_5
void FASTCALL call_12_7_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_11_6
#define init_12_7_11_6
void FASTCALL call_12_7_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_11_7
#define init_12_7_11_7
void FASTCALL call_12_7_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_11_8
#define init_12_7_11_8
void FASTCALL call_12_7_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_11_9
#define init_12_7_11_9
void FASTCALL call_12_7_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_11_10
#define init_12_7_11_10
void FASTCALL call_12_7_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_11_11
#define init_12_7_11_11
void FASTCALL call_12_7_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_11_12
#define init_12_7_11_12
void FASTCALL call_12_7_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_11_13
#define init_12_7_11_13
void FASTCALL call_12_7_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_11_14
#define init_12_7_11_14
void FASTCALL call_12_7_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_11_15
#define init_12_7_11_15
void FASTCALL call_12_7_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_12_0
#define init_12_7_12_0
void FASTCALL call_12_7_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_12_1
#define init_12_7_12_1
void FASTCALL call_12_7_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_12_2
#define init_12_7_12_2
void FASTCALL call_12_7_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_12_3
#define init_12_7_12_3
void FASTCALL call_12_7_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_12_4
#define init_12_7_12_4
void FASTCALL call_12_7_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_12_5
#define init_12_7_12_5
void FASTCALL call_12_7_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_12_6
#define init_12_7_12_6
void FASTCALL call_12_7_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_12_7
#define init_12_7_12_7
void FASTCALL call_12_7_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_12_8
#define init_12_7_12_8
void FASTCALL call_12_7_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_12_9
#define init_12_7_12_9
void FASTCALL call_12_7_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_12_10
#define init_12_7_12_10
void FASTCALL call_12_7_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_12_11
#define init_12_7_12_11
void FASTCALL call_12_7_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_12_12
#define init_12_7_12_12
void FASTCALL call_12_7_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_12_13
#define init_12_7_12_13
void FASTCALL call_12_7_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_12_14
#define init_12_7_12_14
void FASTCALL call_12_7_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_12_15
#define init_12_7_12_15
void FASTCALL call_12_7_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_13_0
#define init_12_7_13_0
void FASTCALL call_12_7_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_13_1
#define init_12_7_13_1
void FASTCALL call_12_7_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_13_2
#define init_12_7_13_2
void FASTCALL call_12_7_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_13_3
#define init_12_7_13_3
void FASTCALL call_12_7_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_13_4
#define init_12_7_13_4
void FASTCALL call_12_7_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_13_5
#define init_12_7_13_5
void FASTCALL call_12_7_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_13_6
#define init_12_7_13_6
void FASTCALL call_12_7_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_13_7
#define init_12_7_13_7
void FASTCALL call_12_7_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_13_8
#define init_12_7_13_8
void FASTCALL call_12_7_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_13_9
#define init_12_7_13_9
void FASTCALL call_12_7_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_13_10
#define init_12_7_13_10
void FASTCALL call_12_7_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_13_11
#define init_12_7_13_11
void FASTCALL call_12_7_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_13_12
#define init_12_7_13_12
void FASTCALL call_12_7_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_13_13
#define init_12_7_13_13
void FASTCALL call_12_7_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_13_14
#define init_12_7_13_14
void FASTCALL call_12_7_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_13_15
#define init_12_7_13_15
void FASTCALL call_12_7_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_14_0
#define init_12_7_14_0
void FASTCALL call_12_7_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_14_1
#define init_12_7_14_1
void FASTCALL call_12_7_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_14_2
#define init_12_7_14_2
void FASTCALL call_12_7_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_14_3
#define init_12_7_14_3
void FASTCALL call_12_7_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_14_4
#define init_12_7_14_4
void FASTCALL call_12_7_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_14_5
#define init_12_7_14_5
void FASTCALL call_12_7_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_14_6
#define init_12_7_14_6
void FASTCALL call_12_7_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_14_7
#define init_12_7_14_7
void FASTCALL call_12_7_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_14_8
#define init_12_7_14_8
void FASTCALL call_12_7_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_14_9
#define init_12_7_14_9
void FASTCALL call_12_7_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_14_10
#define init_12_7_14_10
void FASTCALL call_12_7_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_14_11
#define init_12_7_14_11
void FASTCALL call_12_7_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_14_12
#define init_12_7_14_12
void FASTCALL call_12_7_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_14_13
#define init_12_7_14_13
void FASTCALL call_12_7_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_14_14
#define init_12_7_14_14
void FASTCALL call_12_7_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_14_15
#define init_12_7_14_15
void FASTCALL call_12_7_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_15_0
#define init_12_7_15_0
void FASTCALL call_12_7_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_15_1
#define init_12_7_15_1
void FASTCALL call_12_7_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_15_2
#define init_12_7_15_2
void FASTCALL call_12_7_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_15_3
#define init_12_7_15_3
void FASTCALL call_12_7_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_15_4
#define init_12_7_15_4
void FASTCALL call_12_7_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_15_5
#define init_12_7_15_5
void FASTCALL call_12_7_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_15_6
#define init_12_7_15_6
void FASTCALL call_12_7_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_15_7
#define init_12_7_15_7
void FASTCALL call_12_7_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_15_8
#define init_12_7_15_8
void FASTCALL call_12_7_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_15_9
#define init_12_7_15_9
void FASTCALL call_12_7_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_15_10
#define init_12_7_15_10
void FASTCALL call_12_7_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_15_11
#define init_12_7_15_11
void FASTCALL call_12_7_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_15_12
#define init_12_7_15_12
void FASTCALL call_12_7_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_15_13
#define init_12_7_15_13
void FASTCALL call_12_7_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_15_14
#define init_12_7_15_14
void FASTCALL call_12_7_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_7_15_15
#define init_12_7_15_15
void FASTCALL call_12_7_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_0_0
#define init_12_8_0_0
void FASTCALL call_12_8_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_0_1
#define init_12_8_0_1
void FASTCALL call_12_8_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_0_2
#define init_12_8_0_2
void FASTCALL call_12_8_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_0_3
#define init_12_8_0_3
void FASTCALL call_12_8_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_0_4
#define init_12_8_0_4
void FASTCALL call_12_8_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_0_5
#define init_12_8_0_5
void FASTCALL call_12_8_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_0_6
#define init_12_8_0_6
void FASTCALL call_12_8_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_0_7
#define init_12_8_0_7
void FASTCALL call_12_8_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_0_8
#define init_12_8_0_8
void FASTCALL call_12_8_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_0_9
#define init_12_8_0_9
void FASTCALL call_12_8_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_0_10
#define init_12_8_0_10
void FASTCALL call_12_8_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_0_11
#define init_12_8_0_11
void FASTCALL call_12_8_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_0_12
#define init_12_8_0_12
void FASTCALL call_12_8_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_0_13
#define init_12_8_0_13
void FASTCALL call_12_8_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_0_14
#define init_12_8_0_14
void FASTCALL call_12_8_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_0_15
#define init_12_8_0_15
void FASTCALL call_12_8_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_1_0
#define init_12_8_1_0
void FASTCALL call_12_8_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_1_1
#define init_12_8_1_1
void FASTCALL call_12_8_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_1_2
#define init_12_8_1_2
void FASTCALL call_12_8_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_1_3
#define init_12_8_1_3
void FASTCALL call_12_8_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_1_4
#define init_12_8_1_4
void FASTCALL call_12_8_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_1_5
#define init_12_8_1_5
void FASTCALL call_12_8_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_1_6
#define init_12_8_1_6
void FASTCALL call_12_8_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_1_7
#define init_12_8_1_7
void FASTCALL call_12_8_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_1_8
#define init_12_8_1_8
void FASTCALL call_12_8_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_1_9
#define init_12_8_1_9
void FASTCALL call_12_8_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_1_10
#define init_12_8_1_10
void FASTCALL call_12_8_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_1_11
#define init_12_8_1_11
void FASTCALL call_12_8_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_1_12
#define init_12_8_1_12
void FASTCALL call_12_8_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_1_13
#define init_12_8_1_13
void FASTCALL call_12_8_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_1_14
#define init_12_8_1_14
void FASTCALL call_12_8_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_1_15
#define init_12_8_1_15
void FASTCALL call_12_8_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_2_0
#define init_12_8_2_0
void FASTCALL call_12_8_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_2_1
#define init_12_8_2_1
void FASTCALL call_12_8_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_2_2
#define init_12_8_2_2
void FASTCALL call_12_8_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_2_3
#define init_12_8_2_3
void FASTCALL call_12_8_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_2_4
#define init_12_8_2_4
void FASTCALL call_12_8_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_2_5
#define init_12_8_2_5
void FASTCALL call_12_8_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_2_6
#define init_12_8_2_6
void FASTCALL call_12_8_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_2_7
#define init_12_8_2_7
void FASTCALL call_12_8_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_2_8
#define init_12_8_2_8
void FASTCALL call_12_8_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_2_9
#define init_12_8_2_9
void FASTCALL call_12_8_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_2_10
#define init_12_8_2_10
void FASTCALL call_12_8_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_2_11
#define init_12_8_2_11
void FASTCALL call_12_8_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_2_12
#define init_12_8_2_12
void FASTCALL call_12_8_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_2_13
#define init_12_8_2_13
void FASTCALL call_12_8_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_2_14
#define init_12_8_2_14
void FASTCALL call_12_8_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_2_15
#define init_12_8_2_15
void FASTCALL call_12_8_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_3_0
#define init_12_8_3_0
void FASTCALL call_12_8_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_3_1
#define init_12_8_3_1
void FASTCALL call_12_8_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_3_2
#define init_12_8_3_2
void FASTCALL call_12_8_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_3_3
#define init_12_8_3_3
void FASTCALL call_12_8_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_3_4
#define init_12_8_3_4
void FASTCALL call_12_8_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_3_5
#define init_12_8_3_5
void FASTCALL call_12_8_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_3_6
#define init_12_8_3_6
void FASTCALL call_12_8_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_3_7
#define init_12_8_3_7
void FASTCALL call_12_8_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_3_8
#define init_12_8_3_8
void FASTCALL call_12_8_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_3_9
#define init_12_8_3_9
void FASTCALL call_12_8_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_3_10
#define init_12_8_3_10
void FASTCALL call_12_8_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_3_11
#define init_12_8_3_11
void FASTCALL call_12_8_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_3_12
#define init_12_8_3_12
void FASTCALL call_12_8_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_3_13
#define init_12_8_3_13
void FASTCALL call_12_8_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_3_14
#define init_12_8_3_14
void FASTCALL call_12_8_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_3_15
#define init_12_8_3_15
void FASTCALL call_12_8_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_4_0
#define init_12_8_4_0
void FASTCALL call_12_8_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_4_1
#define init_12_8_4_1
void FASTCALL call_12_8_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_4_2
#define init_12_8_4_2
void FASTCALL call_12_8_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_4_3
#define init_12_8_4_3
void FASTCALL call_12_8_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_4_4
#define init_12_8_4_4
void FASTCALL call_12_8_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_4_5
#define init_12_8_4_5
void FASTCALL call_12_8_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_4_6
#define init_12_8_4_6
void FASTCALL call_12_8_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_4_7
#define init_12_8_4_7
void FASTCALL call_12_8_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_4_8
#define init_12_8_4_8
void FASTCALL call_12_8_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_4_9
#define init_12_8_4_9
void FASTCALL call_12_8_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_4_10
#define init_12_8_4_10
void FASTCALL call_12_8_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_4_11
#define init_12_8_4_11
void FASTCALL call_12_8_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_4_12
#define init_12_8_4_12
void FASTCALL call_12_8_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_4_13
#define init_12_8_4_13
void FASTCALL call_12_8_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_4_14
#define init_12_8_4_14
void FASTCALL call_12_8_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_4_15
#define init_12_8_4_15
void FASTCALL call_12_8_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_5_0
#define init_12_8_5_0
void FASTCALL call_12_8_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_5_1
#define init_12_8_5_1
void FASTCALL call_12_8_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_5_2
#define init_12_8_5_2
void FASTCALL call_12_8_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_5_3
#define init_12_8_5_3
void FASTCALL call_12_8_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_5_4
#define init_12_8_5_4
void FASTCALL call_12_8_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_5_5
#define init_12_8_5_5
void FASTCALL call_12_8_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_5_6
#define init_12_8_5_6
void FASTCALL call_12_8_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_5_7
#define init_12_8_5_7
void FASTCALL call_12_8_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_5_8
#define init_12_8_5_8
void FASTCALL call_12_8_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_5_9
#define init_12_8_5_9
void FASTCALL call_12_8_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_5_10
#define init_12_8_5_10
void FASTCALL call_12_8_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_5_11
#define init_12_8_5_11
void FASTCALL call_12_8_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_5_12
#define init_12_8_5_12
void FASTCALL call_12_8_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_5_13
#define init_12_8_5_13
void FASTCALL call_12_8_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_5_14
#define init_12_8_5_14
void FASTCALL call_12_8_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_5_15
#define init_12_8_5_15
void FASTCALL call_12_8_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_6_0
#define init_12_8_6_0
void FASTCALL call_12_8_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_6_1
#define init_12_8_6_1
void FASTCALL call_12_8_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_6_2
#define init_12_8_6_2
void FASTCALL call_12_8_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_6_3
#define init_12_8_6_3
void FASTCALL call_12_8_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_6_4
#define init_12_8_6_4
void FASTCALL call_12_8_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_6_5
#define init_12_8_6_5
void FASTCALL call_12_8_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_6_6
#define init_12_8_6_6
void FASTCALL call_12_8_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_6_7
#define init_12_8_6_7
void FASTCALL call_12_8_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_6_8
#define init_12_8_6_8
void FASTCALL call_12_8_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_6_9
#define init_12_8_6_9
void FASTCALL call_12_8_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_6_10
#define init_12_8_6_10
void FASTCALL call_12_8_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_6_11
#define init_12_8_6_11
void FASTCALL call_12_8_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_6_12
#define init_12_8_6_12
void FASTCALL call_12_8_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_6_13
#define init_12_8_6_13
void FASTCALL call_12_8_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_6_14
#define init_12_8_6_14
void FASTCALL call_12_8_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_6_15
#define init_12_8_6_15
void FASTCALL call_12_8_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_7_0
#define init_12_8_7_0
void FASTCALL call_12_8_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_7_1
#define init_12_8_7_1
void FASTCALL call_12_8_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_7_2
#define init_12_8_7_2
void FASTCALL call_12_8_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_7_3
#define init_12_8_7_3
void FASTCALL call_12_8_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_7_4
#define init_12_8_7_4
void FASTCALL call_12_8_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_7_5
#define init_12_8_7_5
void FASTCALL call_12_8_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_7_6
#define init_12_8_7_6
void FASTCALL call_12_8_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_7_7
#define init_12_8_7_7
void FASTCALL call_12_8_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_7_8
#define init_12_8_7_8
void FASTCALL call_12_8_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_7_9
#define init_12_8_7_9
void FASTCALL call_12_8_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_7_10
#define init_12_8_7_10
void FASTCALL call_12_8_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_7_11
#define init_12_8_7_11
void FASTCALL call_12_8_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_7_12
#define init_12_8_7_12
void FASTCALL call_12_8_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_7_13
#define init_12_8_7_13
void FASTCALL call_12_8_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_7_14
#define init_12_8_7_14
void FASTCALL call_12_8_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_7_15
#define init_12_8_7_15
void FASTCALL call_12_8_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_8_0
#define init_12_8_8_0
void FASTCALL call_12_8_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_8_1
#define init_12_8_8_1
void FASTCALL call_12_8_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_8_2
#define init_12_8_8_2
void FASTCALL call_12_8_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_8_3
#define init_12_8_8_3
void FASTCALL call_12_8_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_8_4
#define init_12_8_8_4
void FASTCALL call_12_8_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_8_5
#define init_12_8_8_5
void FASTCALL call_12_8_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_8_6
#define init_12_8_8_6
void FASTCALL call_12_8_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_8_7
#define init_12_8_8_7
void FASTCALL call_12_8_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_8_8
#define init_12_8_8_8
void FASTCALL call_12_8_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_8_9
#define init_12_8_8_9
void FASTCALL call_12_8_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_8_10
#define init_12_8_8_10
void FASTCALL call_12_8_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_8_11
#define init_12_8_8_11
void FASTCALL call_12_8_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_8_12
#define init_12_8_8_12
void FASTCALL call_12_8_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_8_13
#define init_12_8_8_13
void FASTCALL call_12_8_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_8_14
#define init_12_8_8_14
void FASTCALL call_12_8_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_8_15
#define init_12_8_8_15
void FASTCALL call_12_8_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_9_0
#define init_12_8_9_0
void FASTCALL call_12_8_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_9_1
#define init_12_8_9_1
void FASTCALL call_12_8_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_9_2
#define init_12_8_9_2
void FASTCALL call_12_8_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_9_3
#define init_12_8_9_3
void FASTCALL call_12_8_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_9_4
#define init_12_8_9_4
void FASTCALL call_12_8_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_9_5
#define init_12_8_9_5
void FASTCALL call_12_8_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_9_6
#define init_12_8_9_6
void FASTCALL call_12_8_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_9_7
#define init_12_8_9_7
void FASTCALL call_12_8_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_9_8
#define init_12_8_9_8
void FASTCALL call_12_8_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_9_9
#define init_12_8_9_9
void FASTCALL call_12_8_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_9_10
#define init_12_8_9_10
void FASTCALL call_12_8_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_9_11
#define init_12_8_9_11
void FASTCALL call_12_8_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_9_12
#define init_12_8_9_12
void FASTCALL call_12_8_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_9_13
#define init_12_8_9_13
void FASTCALL call_12_8_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_9_14
#define init_12_8_9_14
void FASTCALL call_12_8_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_9_15
#define init_12_8_9_15
void FASTCALL call_12_8_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_10_0
#define init_12_8_10_0
void FASTCALL call_12_8_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_10_1
#define init_12_8_10_1
void FASTCALL call_12_8_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_10_2
#define init_12_8_10_2
void FASTCALL call_12_8_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_10_3
#define init_12_8_10_3
void FASTCALL call_12_8_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_10_4
#define init_12_8_10_4
void FASTCALL call_12_8_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_10_5
#define init_12_8_10_5
void FASTCALL call_12_8_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_10_6
#define init_12_8_10_6
void FASTCALL call_12_8_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_10_7
#define init_12_8_10_7
void FASTCALL call_12_8_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_10_8
#define init_12_8_10_8
void FASTCALL call_12_8_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_10_9
#define init_12_8_10_9
void FASTCALL call_12_8_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_10_10
#define init_12_8_10_10
void FASTCALL call_12_8_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_10_11
#define init_12_8_10_11
void FASTCALL call_12_8_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_10_12
#define init_12_8_10_12
void FASTCALL call_12_8_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_10_13
#define init_12_8_10_13
void FASTCALL call_12_8_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_10_14
#define init_12_8_10_14
void FASTCALL call_12_8_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_10_15
#define init_12_8_10_15
void FASTCALL call_12_8_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_11_0
#define init_12_8_11_0
void FASTCALL call_12_8_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_11_1
#define init_12_8_11_1
void FASTCALL call_12_8_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_11_2
#define init_12_8_11_2
void FASTCALL call_12_8_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_11_3
#define init_12_8_11_3
void FASTCALL call_12_8_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_11_4
#define init_12_8_11_4
void FASTCALL call_12_8_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_11_5
#define init_12_8_11_5
void FASTCALL call_12_8_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_11_6
#define init_12_8_11_6
void FASTCALL call_12_8_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_11_7
#define init_12_8_11_7
void FASTCALL call_12_8_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_11_8
#define init_12_8_11_8
void FASTCALL call_12_8_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_11_9
#define init_12_8_11_9
void FASTCALL call_12_8_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_11_10
#define init_12_8_11_10
void FASTCALL call_12_8_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_11_11
#define init_12_8_11_11
void FASTCALL call_12_8_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_11_12
#define init_12_8_11_12
void FASTCALL call_12_8_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_11_13
#define init_12_8_11_13
void FASTCALL call_12_8_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_11_14
#define init_12_8_11_14
void FASTCALL call_12_8_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_11_15
#define init_12_8_11_15
void FASTCALL call_12_8_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_12_0
#define init_12_8_12_0
void FASTCALL call_12_8_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_12_1
#define init_12_8_12_1
void FASTCALL call_12_8_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_12_2
#define init_12_8_12_2
void FASTCALL call_12_8_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_12_3
#define init_12_8_12_3
void FASTCALL call_12_8_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_12_4
#define init_12_8_12_4
void FASTCALL call_12_8_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_12_5
#define init_12_8_12_5
void FASTCALL call_12_8_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_12_6
#define init_12_8_12_6
void FASTCALL call_12_8_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_12_7
#define init_12_8_12_7
void FASTCALL call_12_8_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_12_8
#define init_12_8_12_8
void FASTCALL call_12_8_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_12_9
#define init_12_8_12_9
void FASTCALL call_12_8_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_12_10
#define init_12_8_12_10
void FASTCALL call_12_8_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_12_11
#define init_12_8_12_11
void FASTCALL call_12_8_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_12_12
#define init_12_8_12_12
void FASTCALL call_12_8_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_12_13
#define init_12_8_12_13
void FASTCALL call_12_8_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_12_14
#define init_12_8_12_14
void FASTCALL call_12_8_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_12_15
#define init_12_8_12_15
void FASTCALL call_12_8_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_13_0
#define init_12_8_13_0
void FASTCALL call_12_8_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_13_1
#define init_12_8_13_1
void FASTCALL call_12_8_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_13_2
#define init_12_8_13_2
void FASTCALL call_12_8_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_13_3
#define init_12_8_13_3
void FASTCALL call_12_8_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_13_4
#define init_12_8_13_4
void FASTCALL call_12_8_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_13_5
#define init_12_8_13_5
void FASTCALL call_12_8_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_13_6
#define init_12_8_13_6
void FASTCALL call_12_8_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_13_7
#define init_12_8_13_7
void FASTCALL call_12_8_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_13_8
#define init_12_8_13_8
void FASTCALL call_12_8_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_13_9
#define init_12_8_13_9
void FASTCALL call_12_8_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_13_10
#define init_12_8_13_10
void FASTCALL call_12_8_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_13_11
#define init_12_8_13_11
void FASTCALL call_12_8_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_13_12
#define init_12_8_13_12
void FASTCALL call_12_8_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_13_13
#define init_12_8_13_13
void FASTCALL call_12_8_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_13_14
#define init_12_8_13_14
void FASTCALL call_12_8_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_13_15
#define init_12_8_13_15
void FASTCALL call_12_8_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_14_0
#define init_12_8_14_0
void FASTCALL call_12_8_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_14_1
#define init_12_8_14_1
void FASTCALL call_12_8_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_14_2
#define init_12_8_14_2
void FASTCALL call_12_8_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_14_3
#define init_12_8_14_3
void FASTCALL call_12_8_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_14_4
#define init_12_8_14_4
void FASTCALL call_12_8_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_14_5
#define init_12_8_14_5
void FASTCALL call_12_8_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_14_6
#define init_12_8_14_6
void FASTCALL call_12_8_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_14_7
#define init_12_8_14_7
void FASTCALL call_12_8_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_14_8
#define init_12_8_14_8
void FASTCALL call_12_8_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_14_9
#define init_12_8_14_9
void FASTCALL call_12_8_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_14_10
#define init_12_8_14_10
void FASTCALL call_12_8_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_14_11
#define init_12_8_14_11
void FASTCALL call_12_8_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_14_12
#define init_12_8_14_12
void FASTCALL call_12_8_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_14_13
#define init_12_8_14_13
void FASTCALL call_12_8_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_14_14
#define init_12_8_14_14
void FASTCALL call_12_8_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_14_15
#define init_12_8_14_15
void FASTCALL call_12_8_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_15_0
#define init_12_8_15_0
void FASTCALL call_12_8_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_15_1
#define init_12_8_15_1
void FASTCALL call_12_8_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_15_2
#define init_12_8_15_2
void FASTCALL call_12_8_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_15_3
#define init_12_8_15_3
void FASTCALL call_12_8_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_15_4
#define init_12_8_15_4
void FASTCALL call_12_8_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_15_5
#define init_12_8_15_5
void FASTCALL call_12_8_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_15_6
#define init_12_8_15_6
void FASTCALL call_12_8_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_15_7
#define init_12_8_15_7
void FASTCALL call_12_8_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_15_8
#define init_12_8_15_8
void FASTCALL call_12_8_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_15_9
#define init_12_8_15_9
void FASTCALL call_12_8_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_15_10
#define init_12_8_15_10
void FASTCALL call_12_8_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_15_11
#define init_12_8_15_11
void FASTCALL call_12_8_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_15_12
#define init_12_8_15_12
void FASTCALL call_12_8_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_15_13
#define init_12_8_15_13
void FASTCALL call_12_8_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_15_14
#define init_12_8_15_14
void FASTCALL call_12_8_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_8_15_15
#define init_12_8_15_15
void FASTCALL call_12_8_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_0_0
#define init_12_9_0_0
void FASTCALL call_12_9_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_0_1
#define init_12_9_0_1
void FASTCALL call_12_9_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_0_2
#define init_12_9_0_2
void FASTCALL call_12_9_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_0_3
#define init_12_9_0_3
void FASTCALL call_12_9_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_0_4
#define init_12_9_0_4
void FASTCALL call_12_9_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_0_5
#define init_12_9_0_5
void FASTCALL call_12_9_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_0_6
#define init_12_9_0_6
void FASTCALL call_12_9_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_0_7
#define init_12_9_0_7
void FASTCALL call_12_9_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_0_8
#define init_12_9_0_8
void FASTCALL call_12_9_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_0_9
#define init_12_9_0_9
void FASTCALL call_12_9_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_0_10
#define init_12_9_0_10
void FASTCALL call_12_9_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_0_11
#define init_12_9_0_11
void FASTCALL call_12_9_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_0_12
#define init_12_9_0_12
void FASTCALL call_12_9_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_0_13
#define init_12_9_0_13
void FASTCALL call_12_9_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_0_14
#define init_12_9_0_14
void FASTCALL call_12_9_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_0_15
#define init_12_9_0_15
void FASTCALL call_12_9_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_1_0
#define init_12_9_1_0
void FASTCALL call_12_9_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_1_1
#define init_12_9_1_1
void FASTCALL call_12_9_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_1_2
#define init_12_9_1_2
void FASTCALL call_12_9_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_1_3
#define init_12_9_1_3
void FASTCALL call_12_9_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_1_4
#define init_12_9_1_4
void FASTCALL call_12_9_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_1_5
#define init_12_9_1_5
void FASTCALL call_12_9_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_1_6
#define init_12_9_1_6
void FASTCALL call_12_9_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_1_7
#define init_12_9_1_7
void FASTCALL call_12_9_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_1_8
#define init_12_9_1_8
void FASTCALL call_12_9_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_1_9
#define init_12_9_1_9
void FASTCALL call_12_9_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_1_10
#define init_12_9_1_10
void FASTCALL call_12_9_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_1_11
#define init_12_9_1_11
void FASTCALL call_12_9_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_1_12
#define init_12_9_1_12
void FASTCALL call_12_9_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_1_13
#define init_12_9_1_13
void FASTCALL call_12_9_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_1_14
#define init_12_9_1_14
void FASTCALL call_12_9_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_1_15
#define init_12_9_1_15
void FASTCALL call_12_9_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_2_0
#define init_12_9_2_0
void FASTCALL call_12_9_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_2_1
#define init_12_9_2_1
void FASTCALL call_12_9_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_2_2
#define init_12_9_2_2
void FASTCALL call_12_9_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_2_3
#define init_12_9_2_3
void FASTCALL call_12_9_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_2_4
#define init_12_9_2_4
void FASTCALL call_12_9_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_2_5
#define init_12_9_2_5
void FASTCALL call_12_9_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_2_6
#define init_12_9_2_6
void FASTCALL call_12_9_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_2_7
#define init_12_9_2_7
void FASTCALL call_12_9_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_2_8
#define init_12_9_2_8
void FASTCALL call_12_9_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_2_9
#define init_12_9_2_9
void FASTCALL call_12_9_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_2_10
#define init_12_9_2_10
void FASTCALL call_12_9_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_2_11
#define init_12_9_2_11
void FASTCALL call_12_9_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_2_12
#define init_12_9_2_12
void FASTCALL call_12_9_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_2_13
#define init_12_9_2_13
void FASTCALL call_12_9_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_2_14
#define init_12_9_2_14
void FASTCALL call_12_9_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_2_15
#define init_12_9_2_15
void FASTCALL call_12_9_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_3_0
#define init_12_9_3_0
void FASTCALL call_12_9_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_3_1
#define init_12_9_3_1
void FASTCALL call_12_9_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_3_2
#define init_12_9_3_2
void FASTCALL call_12_9_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_3_3
#define init_12_9_3_3
void FASTCALL call_12_9_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_3_4
#define init_12_9_3_4
void FASTCALL call_12_9_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_3_5
#define init_12_9_3_5
void FASTCALL call_12_9_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_3_6
#define init_12_9_3_6
void FASTCALL call_12_9_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_3_7
#define init_12_9_3_7
void FASTCALL call_12_9_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_3_8
#define init_12_9_3_8
void FASTCALL call_12_9_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_3_9
#define init_12_9_3_9
void FASTCALL call_12_9_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_3_10
#define init_12_9_3_10
void FASTCALL call_12_9_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_3_11
#define init_12_9_3_11
void FASTCALL call_12_9_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_3_12
#define init_12_9_3_12
void FASTCALL call_12_9_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_3_13
#define init_12_9_3_13
void FASTCALL call_12_9_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_3_14
#define init_12_9_3_14
void FASTCALL call_12_9_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_3_15
#define init_12_9_3_15
void FASTCALL call_12_9_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_4_0
#define init_12_9_4_0
void FASTCALL call_12_9_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_4_1
#define init_12_9_4_1
void FASTCALL call_12_9_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_4_2
#define init_12_9_4_2
void FASTCALL call_12_9_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_4_3
#define init_12_9_4_3
void FASTCALL call_12_9_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_4_4
#define init_12_9_4_4
void FASTCALL call_12_9_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_4_5
#define init_12_9_4_5
void FASTCALL call_12_9_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_4_6
#define init_12_9_4_6
void FASTCALL call_12_9_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_4_7
#define init_12_9_4_7
void FASTCALL call_12_9_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_4_8
#define init_12_9_4_8
void FASTCALL call_12_9_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_4_9
#define init_12_9_4_9
void FASTCALL call_12_9_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_4_10
#define init_12_9_4_10
void FASTCALL call_12_9_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_4_11
#define init_12_9_4_11
void FASTCALL call_12_9_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_4_12
#define init_12_9_4_12
void FASTCALL call_12_9_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_4_13
#define init_12_9_4_13
void FASTCALL call_12_9_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_4_14
#define init_12_9_4_14
void FASTCALL call_12_9_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_4_15
#define init_12_9_4_15
void FASTCALL call_12_9_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_5_0
#define init_12_9_5_0
void FASTCALL call_12_9_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_5_1
#define init_12_9_5_1
void FASTCALL call_12_9_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_5_2
#define init_12_9_5_2
void FASTCALL call_12_9_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_5_3
#define init_12_9_5_3
void FASTCALL call_12_9_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_5_4
#define init_12_9_5_4
void FASTCALL call_12_9_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_5_5
#define init_12_9_5_5
void FASTCALL call_12_9_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_5_6
#define init_12_9_5_6
void FASTCALL call_12_9_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_5_7
#define init_12_9_5_7
void FASTCALL call_12_9_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_5_8
#define init_12_9_5_8
void FASTCALL call_12_9_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_5_9
#define init_12_9_5_9
void FASTCALL call_12_9_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_5_10
#define init_12_9_5_10
void FASTCALL call_12_9_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_5_11
#define init_12_9_5_11
void FASTCALL call_12_9_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_5_12
#define init_12_9_5_12
void FASTCALL call_12_9_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_5_13
#define init_12_9_5_13
void FASTCALL call_12_9_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_5_14
#define init_12_9_5_14
void FASTCALL call_12_9_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_5_15
#define init_12_9_5_15
void FASTCALL call_12_9_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_6_0
#define init_12_9_6_0
void FASTCALL call_12_9_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_6_1
#define init_12_9_6_1
void FASTCALL call_12_9_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_6_2
#define init_12_9_6_2
void FASTCALL call_12_9_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_6_3
#define init_12_9_6_3
void FASTCALL call_12_9_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_6_4
#define init_12_9_6_4
void FASTCALL call_12_9_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_6_5
#define init_12_9_6_5
void FASTCALL call_12_9_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_6_6
#define init_12_9_6_6
void FASTCALL call_12_9_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_6_7
#define init_12_9_6_7
void FASTCALL call_12_9_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_6_8
#define init_12_9_6_8
void FASTCALL call_12_9_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_6_9
#define init_12_9_6_9
void FASTCALL call_12_9_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_6_10
#define init_12_9_6_10
void FASTCALL call_12_9_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_6_11
#define init_12_9_6_11
void FASTCALL call_12_9_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_6_12
#define init_12_9_6_12
void FASTCALL call_12_9_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_6_13
#define init_12_9_6_13
void FASTCALL call_12_9_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_6_14
#define init_12_9_6_14
void FASTCALL call_12_9_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_6_15
#define init_12_9_6_15
void FASTCALL call_12_9_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_7_0
#define init_12_9_7_0
void FASTCALL call_12_9_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_7_1
#define init_12_9_7_1
void FASTCALL call_12_9_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_7_2
#define init_12_9_7_2
void FASTCALL call_12_9_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_7_3
#define init_12_9_7_3
void FASTCALL call_12_9_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_7_4
#define init_12_9_7_4
void FASTCALL call_12_9_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_7_5
#define init_12_9_7_5
void FASTCALL call_12_9_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_7_6
#define init_12_9_7_6
void FASTCALL call_12_9_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_7_7
#define init_12_9_7_7
void FASTCALL call_12_9_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_7_8
#define init_12_9_7_8
void FASTCALL call_12_9_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_7_9
#define init_12_9_7_9
void FASTCALL call_12_9_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_7_10
#define init_12_9_7_10
void FASTCALL call_12_9_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_7_11
#define init_12_9_7_11
void FASTCALL call_12_9_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_7_12
#define init_12_9_7_12
void FASTCALL call_12_9_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_7_13
#define init_12_9_7_13
void FASTCALL call_12_9_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_7_14
#define init_12_9_7_14
void FASTCALL call_12_9_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_7_15
#define init_12_9_7_15
void FASTCALL call_12_9_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_8_0
#define init_12_9_8_0
void FASTCALL call_12_9_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_8_1
#define init_12_9_8_1
void FASTCALL call_12_9_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_8_2
#define init_12_9_8_2
void FASTCALL call_12_9_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_8_3
#define init_12_9_8_3
void FASTCALL call_12_9_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_8_4
#define init_12_9_8_4
void FASTCALL call_12_9_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_8_5
#define init_12_9_8_5
void FASTCALL call_12_9_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_8_6
#define init_12_9_8_6
void FASTCALL call_12_9_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_8_7
#define init_12_9_8_7
void FASTCALL call_12_9_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_8_8
#define init_12_9_8_8
void FASTCALL call_12_9_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_8_9
#define init_12_9_8_9
void FASTCALL call_12_9_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_8_10
#define init_12_9_8_10
void FASTCALL call_12_9_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_8_11
#define init_12_9_8_11
void FASTCALL call_12_9_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_8_12
#define init_12_9_8_12
void FASTCALL call_12_9_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_8_13
#define init_12_9_8_13
void FASTCALL call_12_9_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_8_14
#define init_12_9_8_14
void FASTCALL call_12_9_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_8_15
#define init_12_9_8_15
void FASTCALL call_12_9_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_9_0
#define init_12_9_9_0
void FASTCALL call_12_9_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_9_1
#define init_12_9_9_1
void FASTCALL call_12_9_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_9_2
#define init_12_9_9_2
void FASTCALL call_12_9_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_9_3
#define init_12_9_9_3
void FASTCALL call_12_9_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_9_4
#define init_12_9_9_4
void FASTCALL call_12_9_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_9_5
#define init_12_9_9_5
void FASTCALL call_12_9_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_9_6
#define init_12_9_9_6
void FASTCALL call_12_9_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_9_7
#define init_12_9_9_7
void FASTCALL call_12_9_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_9_8
#define init_12_9_9_8
void FASTCALL call_12_9_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_9_9
#define init_12_9_9_9
void FASTCALL call_12_9_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_9_10
#define init_12_9_9_10
void FASTCALL call_12_9_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_9_11
#define init_12_9_9_11
void FASTCALL call_12_9_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_9_12
#define init_12_9_9_12
void FASTCALL call_12_9_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_9_13
#define init_12_9_9_13
void FASTCALL call_12_9_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_9_14
#define init_12_9_9_14
void FASTCALL call_12_9_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_9_15
#define init_12_9_9_15
void FASTCALL call_12_9_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_10_0
#define init_12_9_10_0
void FASTCALL call_12_9_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_10_1
#define init_12_9_10_1
void FASTCALL call_12_9_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_10_2
#define init_12_9_10_2
void FASTCALL call_12_9_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_10_3
#define init_12_9_10_3
void FASTCALL call_12_9_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_10_4
#define init_12_9_10_4
void FASTCALL call_12_9_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_10_5
#define init_12_9_10_5
void FASTCALL call_12_9_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_10_6
#define init_12_9_10_6
void FASTCALL call_12_9_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_10_7
#define init_12_9_10_7
void FASTCALL call_12_9_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_10_8
#define init_12_9_10_8
void FASTCALL call_12_9_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_10_9
#define init_12_9_10_9
void FASTCALL call_12_9_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_10_10
#define init_12_9_10_10
void FASTCALL call_12_9_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_10_11
#define init_12_9_10_11
void FASTCALL call_12_9_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_10_12
#define init_12_9_10_12
void FASTCALL call_12_9_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_10_13
#define init_12_9_10_13
void FASTCALL call_12_9_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_10_14
#define init_12_9_10_14
void FASTCALL call_12_9_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_10_15
#define init_12_9_10_15
void FASTCALL call_12_9_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_11_0
#define init_12_9_11_0
void FASTCALL call_12_9_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_11_1
#define init_12_9_11_1
void FASTCALL call_12_9_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_11_2
#define init_12_9_11_2
void FASTCALL call_12_9_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_11_3
#define init_12_9_11_3
void FASTCALL call_12_9_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_11_4
#define init_12_9_11_4
void FASTCALL call_12_9_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_11_5
#define init_12_9_11_5
void FASTCALL call_12_9_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_11_6
#define init_12_9_11_6
void FASTCALL call_12_9_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_11_7
#define init_12_9_11_7
void FASTCALL call_12_9_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_11_8
#define init_12_9_11_8
void FASTCALL call_12_9_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_11_9
#define init_12_9_11_9
void FASTCALL call_12_9_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_11_10
#define init_12_9_11_10
void FASTCALL call_12_9_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_11_11
#define init_12_9_11_11
void FASTCALL call_12_9_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_11_12
#define init_12_9_11_12
void FASTCALL call_12_9_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_11_13
#define init_12_9_11_13
void FASTCALL call_12_9_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_11_14
#define init_12_9_11_14
void FASTCALL call_12_9_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_11_15
#define init_12_9_11_15
void FASTCALL call_12_9_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_12_0
#define init_12_9_12_0
void FASTCALL call_12_9_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_12_1
#define init_12_9_12_1
void FASTCALL call_12_9_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_12_2
#define init_12_9_12_2
void FASTCALL call_12_9_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_12_3
#define init_12_9_12_3
void FASTCALL call_12_9_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_12_4
#define init_12_9_12_4
void FASTCALL call_12_9_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_12_5
#define init_12_9_12_5
void FASTCALL call_12_9_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_12_6
#define init_12_9_12_6
void FASTCALL call_12_9_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_12_7
#define init_12_9_12_7
void FASTCALL call_12_9_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_12_8
#define init_12_9_12_8
void FASTCALL call_12_9_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_12_9
#define init_12_9_12_9
void FASTCALL call_12_9_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_12_10
#define init_12_9_12_10
void FASTCALL call_12_9_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_12_11
#define init_12_9_12_11
void FASTCALL call_12_9_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_12_12
#define init_12_9_12_12
void FASTCALL call_12_9_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_12_13
#define init_12_9_12_13
void FASTCALL call_12_9_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_12_14
#define init_12_9_12_14
void FASTCALL call_12_9_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_12_15
#define init_12_9_12_15
void FASTCALL call_12_9_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_13_0
#define init_12_9_13_0
void FASTCALL call_12_9_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_13_1
#define init_12_9_13_1
void FASTCALL call_12_9_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_13_2
#define init_12_9_13_2
void FASTCALL call_12_9_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_13_3
#define init_12_9_13_3
void FASTCALL call_12_9_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_13_4
#define init_12_9_13_4
void FASTCALL call_12_9_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_13_5
#define init_12_9_13_5
void FASTCALL call_12_9_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_13_6
#define init_12_9_13_6
void FASTCALL call_12_9_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_13_7
#define init_12_9_13_7
void FASTCALL call_12_9_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_13_8
#define init_12_9_13_8
void FASTCALL call_12_9_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_13_9
#define init_12_9_13_9
void FASTCALL call_12_9_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_13_10
#define init_12_9_13_10
void FASTCALL call_12_9_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_13_11
#define init_12_9_13_11
void FASTCALL call_12_9_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_13_12
#define init_12_9_13_12
void FASTCALL call_12_9_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_13_13
#define init_12_9_13_13
void FASTCALL call_12_9_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_13_14
#define init_12_9_13_14
void FASTCALL call_12_9_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_13_15
#define init_12_9_13_15
void FASTCALL call_12_9_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_14_0
#define init_12_9_14_0
void FASTCALL call_12_9_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_14_1
#define init_12_9_14_1
void FASTCALL call_12_9_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_14_2
#define init_12_9_14_2
void FASTCALL call_12_9_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_14_3
#define init_12_9_14_3
void FASTCALL call_12_9_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_14_4
#define init_12_9_14_4
void FASTCALL call_12_9_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_14_5
#define init_12_9_14_5
void FASTCALL call_12_9_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_14_6
#define init_12_9_14_6
void FASTCALL call_12_9_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_14_7
#define init_12_9_14_7
void FASTCALL call_12_9_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_14_8
#define init_12_9_14_8
void FASTCALL call_12_9_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_14_9
#define init_12_9_14_9
void FASTCALL call_12_9_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_14_10
#define init_12_9_14_10
void FASTCALL call_12_9_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_14_11
#define init_12_9_14_11
void FASTCALL call_12_9_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_14_12
#define init_12_9_14_12
void FASTCALL call_12_9_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_14_13
#define init_12_9_14_13
void FASTCALL call_12_9_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_14_14
#define init_12_9_14_14
void FASTCALL call_12_9_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_14_15
#define init_12_9_14_15
void FASTCALL call_12_9_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_15_0
#define init_12_9_15_0
void FASTCALL call_12_9_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_15_1
#define init_12_9_15_1
void FASTCALL call_12_9_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_15_2
#define init_12_9_15_2
void FASTCALL call_12_9_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_15_3
#define init_12_9_15_3
void FASTCALL call_12_9_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_15_4
#define init_12_9_15_4
void FASTCALL call_12_9_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_15_5
#define init_12_9_15_5
void FASTCALL call_12_9_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_15_6
#define init_12_9_15_6
void FASTCALL call_12_9_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_15_7
#define init_12_9_15_7
void FASTCALL call_12_9_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_15_8
#define init_12_9_15_8
void FASTCALL call_12_9_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_15_9
#define init_12_9_15_9
void FASTCALL call_12_9_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_15_10
#define init_12_9_15_10
void FASTCALL call_12_9_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_15_11
#define init_12_9_15_11
void FASTCALL call_12_9_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_15_12
#define init_12_9_15_12
void FASTCALL call_12_9_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_15_13
#define init_12_9_15_13
void FASTCALL call_12_9_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_15_14
#define init_12_9_15_14
void FASTCALL call_12_9_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_9_15_15
#define init_12_9_15_15
void FASTCALL call_12_9_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_0_0
#define init_12_10_0_0
void FASTCALL call_12_10_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_0_1
#define init_12_10_0_1
void FASTCALL call_12_10_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_0_2
#define init_12_10_0_2
void FASTCALL call_12_10_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_0_3
#define init_12_10_0_3
void FASTCALL call_12_10_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_0_4
#define init_12_10_0_4
void FASTCALL call_12_10_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_0_5
#define init_12_10_0_5
void FASTCALL call_12_10_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_0_6
#define init_12_10_0_6
void FASTCALL call_12_10_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_0_7
#define init_12_10_0_7
void FASTCALL call_12_10_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_0_8
#define init_12_10_0_8
void FASTCALL call_12_10_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_0_9
#define init_12_10_0_9
void FASTCALL call_12_10_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_0_10
#define init_12_10_0_10
void FASTCALL call_12_10_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_0_11
#define init_12_10_0_11
void FASTCALL call_12_10_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_0_12
#define init_12_10_0_12
void FASTCALL call_12_10_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_0_13
#define init_12_10_0_13
void FASTCALL call_12_10_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_0_14
#define init_12_10_0_14
void FASTCALL call_12_10_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_0_15
#define init_12_10_0_15
void FASTCALL call_12_10_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_1_0
#define init_12_10_1_0
void FASTCALL call_12_10_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_1_1
#define init_12_10_1_1
void FASTCALL call_12_10_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_1_2
#define init_12_10_1_2
void FASTCALL call_12_10_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_1_3
#define init_12_10_1_3
void FASTCALL call_12_10_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_1_4
#define init_12_10_1_4
void FASTCALL call_12_10_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_1_5
#define init_12_10_1_5
void FASTCALL call_12_10_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_1_6
#define init_12_10_1_6
void FASTCALL call_12_10_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_1_7
#define init_12_10_1_7
void FASTCALL call_12_10_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_1_8
#define init_12_10_1_8
void FASTCALL call_12_10_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_1_9
#define init_12_10_1_9
void FASTCALL call_12_10_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_1_10
#define init_12_10_1_10
void FASTCALL call_12_10_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_1_11
#define init_12_10_1_11
void FASTCALL call_12_10_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_1_12
#define init_12_10_1_12
void FASTCALL call_12_10_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_1_13
#define init_12_10_1_13
void FASTCALL call_12_10_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_1_14
#define init_12_10_1_14
void FASTCALL call_12_10_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_1_15
#define init_12_10_1_15
void FASTCALL call_12_10_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_2_0
#define init_12_10_2_0
void FASTCALL call_12_10_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_2_1
#define init_12_10_2_1
void FASTCALL call_12_10_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_2_2
#define init_12_10_2_2
void FASTCALL call_12_10_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_2_3
#define init_12_10_2_3
void FASTCALL call_12_10_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_2_4
#define init_12_10_2_4
void FASTCALL call_12_10_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_2_5
#define init_12_10_2_5
void FASTCALL call_12_10_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_2_6
#define init_12_10_2_6
void FASTCALL call_12_10_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_2_7
#define init_12_10_2_7
void FASTCALL call_12_10_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_2_8
#define init_12_10_2_8
void FASTCALL call_12_10_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_2_9
#define init_12_10_2_9
void FASTCALL call_12_10_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_2_10
#define init_12_10_2_10
void FASTCALL call_12_10_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_2_11
#define init_12_10_2_11
void FASTCALL call_12_10_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_2_12
#define init_12_10_2_12
void FASTCALL call_12_10_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_2_13
#define init_12_10_2_13
void FASTCALL call_12_10_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_2_14
#define init_12_10_2_14
void FASTCALL call_12_10_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_2_15
#define init_12_10_2_15
void FASTCALL call_12_10_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_3_0
#define init_12_10_3_0
void FASTCALL call_12_10_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_3_1
#define init_12_10_3_1
void FASTCALL call_12_10_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_3_2
#define init_12_10_3_2
void FASTCALL call_12_10_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_3_3
#define init_12_10_3_3
void FASTCALL call_12_10_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_3_4
#define init_12_10_3_4
void FASTCALL call_12_10_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_3_5
#define init_12_10_3_5
void FASTCALL call_12_10_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_3_6
#define init_12_10_3_6
void FASTCALL call_12_10_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_3_7
#define init_12_10_3_7
void FASTCALL call_12_10_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_3_8
#define init_12_10_3_8
void FASTCALL call_12_10_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_3_9
#define init_12_10_3_9
void FASTCALL call_12_10_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_3_10
#define init_12_10_3_10
void FASTCALL call_12_10_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_3_11
#define init_12_10_3_11
void FASTCALL call_12_10_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_3_12
#define init_12_10_3_12
void FASTCALL call_12_10_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_3_13
#define init_12_10_3_13
void FASTCALL call_12_10_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_3_14
#define init_12_10_3_14
void FASTCALL call_12_10_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_3_15
#define init_12_10_3_15
void FASTCALL call_12_10_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_4_0
#define init_12_10_4_0
void FASTCALL call_12_10_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_4_1
#define init_12_10_4_1
void FASTCALL call_12_10_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_4_2
#define init_12_10_4_2
void FASTCALL call_12_10_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_4_3
#define init_12_10_4_3
void FASTCALL call_12_10_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_4_4
#define init_12_10_4_4
void FASTCALL call_12_10_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_4_5
#define init_12_10_4_5
void FASTCALL call_12_10_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_4_6
#define init_12_10_4_6
void FASTCALL call_12_10_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_4_7
#define init_12_10_4_7
void FASTCALL call_12_10_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_4_8
#define init_12_10_4_8
void FASTCALL call_12_10_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_4_9
#define init_12_10_4_9
void FASTCALL call_12_10_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_4_10
#define init_12_10_4_10
void FASTCALL call_12_10_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_4_11
#define init_12_10_4_11
void FASTCALL call_12_10_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_4_12
#define init_12_10_4_12
void FASTCALL call_12_10_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_4_13
#define init_12_10_4_13
void FASTCALL call_12_10_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_4_14
#define init_12_10_4_14
void FASTCALL call_12_10_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_4_15
#define init_12_10_4_15
void FASTCALL call_12_10_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_5_0
#define init_12_10_5_0
void FASTCALL call_12_10_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_5_1
#define init_12_10_5_1
void FASTCALL call_12_10_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_5_2
#define init_12_10_5_2
void FASTCALL call_12_10_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_5_3
#define init_12_10_5_3
void FASTCALL call_12_10_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_5_4
#define init_12_10_5_4
void FASTCALL call_12_10_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_5_5
#define init_12_10_5_5
void FASTCALL call_12_10_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_5_6
#define init_12_10_5_6
void FASTCALL call_12_10_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_5_7
#define init_12_10_5_7
void FASTCALL call_12_10_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_5_8
#define init_12_10_5_8
void FASTCALL call_12_10_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_5_9
#define init_12_10_5_9
void FASTCALL call_12_10_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_5_10
#define init_12_10_5_10
void FASTCALL call_12_10_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_5_11
#define init_12_10_5_11
void FASTCALL call_12_10_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_5_12
#define init_12_10_5_12
void FASTCALL call_12_10_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_5_13
#define init_12_10_5_13
void FASTCALL call_12_10_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_5_14
#define init_12_10_5_14
void FASTCALL call_12_10_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_5_15
#define init_12_10_5_15
void FASTCALL call_12_10_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_6_0
#define init_12_10_6_0
void FASTCALL call_12_10_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_6_1
#define init_12_10_6_1
void FASTCALL call_12_10_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_6_2
#define init_12_10_6_2
void FASTCALL call_12_10_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_6_3
#define init_12_10_6_3
void FASTCALL call_12_10_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_6_4
#define init_12_10_6_4
void FASTCALL call_12_10_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_6_5
#define init_12_10_6_5
void FASTCALL call_12_10_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_6_6
#define init_12_10_6_6
void FASTCALL call_12_10_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_6_7
#define init_12_10_6_7
void FASTCALL call_12_10_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_6_8
#define init_12_10_6_8
void FASTCALL call_12_10_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_6_9
#define init_12_10_6_9
void FASTCALL call_12_10_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_6_10
#define init_12_10_6_10
void FASTCALL call_12_10_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_6_11
#define init_12_10_6_11
void FASTCALL call_12_10_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_6_12
#define init_12_10_6_12
void FASTCALL call_12_10_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_6_13
#define init_12_10_6_13
void FASTCALL call_12_10_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_6_14
#define init_12_10_6_14
void FASTCALL call_12_10_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_6_15
#define init_12_10_6_15
void FASTCALL call_12_10_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_7_0
#define init_12_10_7_0
void FASTCALL call_12_10_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_7_1
#define init_12_10_7_1
void FASTCALL call_12_10_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_7_2
#define init_12_10_7_2
void FASTCALL call_12_10_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_7_3
#define init_12_10_7_3
void FASTCALL call_12_10_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_7_4
#define init_12_10_7_4
void FASTCALL call_12_10_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_7_5
#define init_12_10_7_5
void FASTCALL call_12_10_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_7_6
#define init_12_10_7_6
void FASTCALL call_12_10_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_7_7
#define init_12_10_7_7
void FASTCALL call_12_10_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_7_8
#define init_12_10_7_8
void FASTCALL call_12_10_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_7_9
#define init_12_10_7_9
void FASTCALL call_12_10_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_7_10
#define init_12_10_7_10
void FASTCALL call_12_10_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_7_11
#define init_12_10_7_11
void FASTCALL call_12_10_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_7_12
#define init_12_10_7_12
void FASTCALL call_12_10_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_7_13
#define init_12_10_7_13
void FASTCALL call_12_10_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_7_14
#define init_12_10_7_14
void FASTCALL call_12_10_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_7_15
#define init_12_10_7_15
void FASTCALL call_12_10_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_8_0
#define init_12_10_8_0
void FASTCALL call_12_10_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_8_1
#define init_12_10_8_1
void FASTCALL call_12_10_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_8_2
#define init_12_10_8_2
void FASTCALL call_12_10_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_8_3
#define init_12_10_8_3
void FASTCALL call_12_10_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_8_4
#define init_12_10_8_4
void FASTCALL call_12_10_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_8_5
#define init_12_10_8_5
void FASTCALL call_12_10_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_8_6
#define init_12_10_8_6
void FASTCALL call_12_10_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_8_7
#define init_12_10_8_7
void FASTCALL call_12_10_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_8_8
#define init_12_10_8_8
void FASTCALL call_12_10_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_8_9
#define init_12_10_8_9
void FASTCALL call_12_10_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_8_10
#define init_12_10_8_10
void FASTCALL call_12_10_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_8_11
#define init_12_10_8_11
void FASTCALL call_12_10_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_8_12
#define init_12_10_8_12
void FASTCALL call_12_10_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_8_13
#define init_12_10_8_13
void FASTCALL call_12_10_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_8_14
#define init_12_10_8_14
void FASTCALL call_12_10_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_8_15
#define init_12_10_8_15
void FASTCALL call_12_10_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_9_0
#define init_12_10_9_0
void FASTCALL call_12_10_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_9_1
#define init_12_10_9_1
void FASTCALL call_12_10_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_9_2
#define init_12_10_9_2
void FASTCALL call_12_10_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_9_3
#define init_12_10_9_3
void FASTCALL call_12_10_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_9_4
#define init_12_10_9_4
void FASTCALL call_12_10_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_9_5
#define init_12_10_9_5
void FASTCALL call_12_10_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_9_6
#define init_12_10_9_6
void FASTCALL call_12_10_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_9_7
#define init_12_10_9_7
void FASTCALL call_12_10_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_9_8
#define init_12_10_9_8
void FASTCALL call_12_10_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_9_9
#define init_12_10_9_9
void FASTCALL call_12_10_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_9_10
#define init_12_10_9_10
void FASTCALL call_12_10_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_9_11
#define init_12_10_9_11
void FASTCALL call_12_10_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_9_12
#define init_12_10_9_12
void FASTCALL call_12_10_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_9_13
#define init_12_10_9_13
void FASTCALL call_12_10_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_9_14
#define init_12_10_9_14
void FASTCALL call_12_10_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_9_15
#define init_12_10_9_15
void FASTCALL call_12_10_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_10_0
#define init_12_10_10_0
void FASTCALL call_12_10_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_10_1
#define init_12_10_10_1
void FASTCALL call_12_10_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_10_2
#define init_12_10_10_2
void FASTCALL call_12_10_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_10_3
#define init_12_10_10_3
void FASTCALL call_12_10_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_10_4
#define init_12_10_10_4
void FASTCALL call_12_10_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_10_5
#define init_12_10_10_5
void FASTCALL call_12_10_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_10_6
#define init_12_10_10_6
void FASTCALL call_12_10_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_10_7
#define init_12_10_10_7
void FASTCALL call_12_10_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_10_8
#define init_12_10_10_8
void FASTCALL call_12_10_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_10_9
#define init_12_10_10_9
void FASTCALL call_12_10_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_10_10
#define init_12_10_10_10
void FASTCALL call_12_10_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_10_11
#define init_12_10_10_11
void FASTCALL call_12_10_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_10_12
#define init_12_10_10_12
void FASTCALL call_12_10_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_10_13
#define init_12_10_10_13
void FASTCALL call_12_10_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_10_14
#define init_12_10_10_14
void FASTCALL call_12_10_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_10_15
#define init_12_10_10_15
void FASTCALL call_12_10_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_11_0
#define init_12_10_11_0
void FASTCALL call_12_10_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_11_1
#define init_12_10_11_1
void FASTCALL call_12_10_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_11_2
#define init_12_10_11_2
void FASTCALL call_12_10_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_11_3
#define init_12_10_11_3
void FASTCALL call_12_10_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_11_4
#define init_12_10_11_4
void FASTCALL call_12_10_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_11_5
#define init_12_10_11_5
void FASTCALL call_12_10_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_11_6
#define init_12_10_11_6
void FASTCALL call_12_10_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_11_7
#define init_12_10_11_7
void FASTCALL call_12_10_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_11_8
#define init_12_10_11_8
void FASTCALL call_12_10_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_11_9
#define init_12_10_11_9
void FASTCALL call_12_10_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_11_10
#define init_12_10_11_10
void FASTCALL call_12_10_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_11_11
#define init_12_10_11_11
void FASTCALL call_12_10_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_11_12
#define init_12_10_11_12
void FASTCALL call_12_10_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_11_13
#define init_12_10_11_13
void FASTCALL call_12_10_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_11_14
#define init_12_10_11_14
void FASTCALL call_12_10_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_11_15
#define init_12_10_11_15
void FASTCALL call_12_10_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_12_0
#define init_12_10_12_0
void FASTCALL call_12_10_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_12_1
#define init_12_10_12_1
void FASTCALL call_12_10_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_12_2
#define init_12_10_12_2
void FASTCALL call_12_10_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_12_3
#define init_12_10_12_3
void FASTCALL call_12_10_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_12_4
#define init_12_10_12_4
void FASTCALL call_12_10_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_12_5
#define init_12_10_12_5
void FASTCALL call_12_10_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_12_6
#define init_12_10_12_6
void FASTCALL call_12_10_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_12_7
#define init_12_10_12_7
void FASTCALL call_12_10_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_12_8
#define init_12_10_12_8
void FASTCALL call_12_10_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_12_9
#define init_12_10_12_9
void FASTCALL call_12_10_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_12_10
#define init_12_10_12_10
void FASTCALL call_12_10_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_12_11
#define init_12_10_12_11
void FASTCALL call_12_10_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_12_12
#define init_12_10_12_12
void FASTCALL call_12_10_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_12_13
#define init_12_10_12_13
void FASTCALL call_12_10_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_12_14
#define init_12_10_12_14
void FASTCALL call_12_10_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_12_15
#define init_12_10_12_15
void FASTCALL call_12_10_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_13_0
#define init_12_10_13_0
void FASTCALL call_12_10_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_13_1
#define init_12_10_13_1
void FASTCALL call_12_10_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_13_2
#define init_12_10_13_2
void FASTCALL call_12_10_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_13_3
#define init_12_10_13_3
void FASTCALL call_12_10_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_13_4
#define init_12_10_13_4
void FASTCALL call_12_10_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_13_5
#define init_12_10_13_5
void FASTCALL call_12_10_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_13_6
#define init_12_10_13_6
void FASTCALL call_12_10_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_13_7
#define init_12_10_13_7
void FASTCALL call_12_10_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_13_8
#define init_12_10_13_8
void FASTCALL call_12_10_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_13_9
#define init_12_10_13_9
void FASTCALL call_12_10_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_13_10
#define init_12_10_13_10
void FASTCALL call_12_10_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_13_11
#define init_12_10_13_11
void FASTCALL call_12_10_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_13_12
#define init_12_10_13_12
void FASTCALL call_12_10_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_13_13
#define init_12_10_13_13
void FASTCALL call_12_10_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_13_14
#define init_12_10_13_14
void FASTCALL call_12_10_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_13_15
#define init_12_10_13_15
void FASTCALL call_12_10_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_14_0
#define init_12_10_14_0
void FASTCALL call_12_10_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_14_1
#define init_12_10_14_1
void FASTCALL call_12_10_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_14_2
#define init_12_10_14_2
void FASTCALL call_12_10_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_14_3
#define init_12_10_14_3
void FASTCALL call_12_10_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_14_4
#define init_12_10_14_4
void FASTCALL call_12_10_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_14_5
#define init_12_10_14_5
void FASTCALL call_12_10_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_14_6
#define init_12_10_14_6
void FASTCALL call_12_10_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_14_7
#define init_12_10_14_7
void FASTCALL call_12_10_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_14_8
#define init_12_10_14_8
void FASTCALL call_12_10_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_14_9
#define init_12_10_14_9
void FASTCALL call_12_10_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_14_10
#define init_12_10_14_10
void FASTCALL call_12_10_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_14_11
#define init_12_10_14_11
void FASTCALL call_12_10_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_14_12
#define init_12_10_14_12
void FASTCALL call_12_10_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_14_13
#define init_12_10_14_13
void FASTCALL call_12_10_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_14_14
#define init_12_10_14_14
void FASTCALL call_12_10_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_14_15
#define init_12_10_14_15
void FASTCALL call_12_10_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_15_0
#define init_12_10_15_0
void FASTCALL call_12_10_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_15_1
#define init_12_10_15_1
void FASTCALL call_12_10_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_15_2
#define init_12_10_15_2
void FASTCALL call_12_10_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_15_3
#define init_12_10_15_3
void FASTCALL call_12_10_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_15_4
#define init_12_10_15_4
void FASTCALL call_12_10_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_15_5
#define init_12_10_15_5
void FASTCALL call_12_10_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_15_6
#define init_12_10_15_6
void FASTCALL call_12_10_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_15_7
#define init_12_10_15_7
void FASTCALL call_12_10_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_15_8
#define init_12_10_15_8
void FASTCALL call_12_10_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_15_9
#define init_12_10_15_9
void FASTCALL call_12_10_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_15_10
#define init_12_10_15_10
void FASTCALL call_12_10_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_15_11
#define init_12_10_15_11
void FASTCALL call_12_10_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_15_12
#define init_12_10_15_12
void FASTCALL call_12_10_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_15_13
#define init_12_10_15_13
void FASTCALL call_12_10_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_15_14
#define init_12_10_15_14
void FASTCALL call_12_10_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_10_15_15
#define init_12_10_15_15
void FASTCALL call_12_10_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_0_0
#define init_12_11_0_0
void FASTCALL call_12_11_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_0_1
#define init_12_11_0_1
void FASTCALL call_12_11_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_0_2
#define init_12_11_0_2
void FASTCALL call_12_11_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_0_3
#define init_12_11_0_3
void FASTCALL call_12_11_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_0_4
#define init_12_11_0_4
void FASTCALL call_12_11_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_0_5
#define init_12_11_0_5
void FASTCALL call_12_11_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_0_6
#define init_12_11_0_6
void FASTCALL call_12_11_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_0_7
#define init_12_11_0_7
void FASTCALL call_12_11_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_0_8
#define init_12_11_0_8
void FASTCALL call_12_11_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_0_9
#define init_12_11_0_9
void FASTCALL call_12_11_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_0_10
#define init_12_11_0_10
void FASTCALL call_12_11_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_0_11
#define init_12_11_0_11
void FASTCALL call_12_11_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_0_12
#define init_12_11_0_12
void FASTCALL call_12_11_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_0_13
#define init_12_11_0_13
void FASTCALL call_12_11_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_0_14
#define init_12_11_0_14
void FASTCALL call_12_11_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_0_15
#define init_12_11_0_15
void FASTCALL call_12_11_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_1_0
#define init_12_11_1_0
void FASTCALL call_12_11_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_1_1
#define init_12_11_1_1
void FASTCALL call_12_11_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_1_2
#define init_12_11_1_2
void FASTCALL call_12_11_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_1_3
#define init_12_11_1_3
void FASTCALL call_12_11_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_1_4
#define init_12_11_1_4
void FASTCALL call_12_11_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_1_5
#define init_12_11_1_5
void FASTCALL call_12_11_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_1_6
#define init_12_11_1_6
void FASTCALL call_12_11_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_1_7
#define init_12_11_1_7
void FASTCALL call_12_11_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_1_8
#define init_12_11_1_8
void FASTCALL call_12_11_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_1_9
#define init_12_11_1_9
void FASTCALL call_12_11_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_1_10
#define init_12_11_1_10
void FASTCALL call_12_11_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_1_11
#define init_12_11_1_11
void FASTCALL call_12_11_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_1_12
#define init_12_11_1_12
void FASTCALL call_12_11_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_1_13
#define init_12_11_1_13
void FASTCALL call_12_11_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_1_14
#define init_12_11_1_14
void FASTCALL call_12_11_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_1_15
#define init_12_11_1_15
void FASTCALL call_12_11_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_2_0
#define init_12_11_2_0
void FASTCALL call_12_11_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_2_1
#define init_12_11_2_1
void FASTCALL call_12_11_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_2_2
#define init_12_11_2_2
void FASTCALL call_12_11_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_2_3
#define init_12_11_2_3
void FASTCALL call_12_11_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_2_4
#define init_12_11_2_4
void FASTCALL call_12_11_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_2_5
#define init_12_11_2_5
void FASTCALL call_12_11_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_2_6
#define init_12_11_2_6
void FASTCALL call_12_11_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_2_7
#define init_12_11_2_7
void FASTCALL call_12_11_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_2_8
#define init_12_11_2_8
void FASTCALL call_12_11_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_2_9
#define init_12_11_2_9
void FASTCALL call_12_11_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_2_10
#define init_12_11_2_10
void FASTCALL call_12_11_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_2_11
#define init_12_11_2_11
void FASTCALL call_12_11_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_2_12
#define init_12_11_2_12
void FASTCALL call_12_11_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_2_13
#define init_12_11_2_13
void FASTCALL call_12_11_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_2_14
#define init_12_11_2_14
void FASTCALL call_12_11_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_2_15
#define init_12_11_2_15
void FASTCALL call_12_11_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_3_0
#define init_12_11_3_0
void FASTCALL call_12_11_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_3_1
#define init_12_11_3_1
void FASTCALL call_12_11_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_3_2
#define init_12_11_3_2
void FASTCALL call_12_11_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_3_3
#define init_12_11_3_3
void FASTCALL call_12_11_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_3_4
#define init_12_11_3_4
void FASTCALL call_12_11_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_3_5
#define init_12_11_3_5
void FASTCALL call_12_11_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_3_6
#define init_12_11_3_6
void FASTCALL call_12_11_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_3_7
#define init_12_11_3_7
void FASTCALL call_12_11_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_3_8
#define init_12_11_3_8
void FASTCALL call_12_11_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_3_9
#define init_12_11_3_9
void FASTCALL call_12_11_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_3_10
#define init_12_11_3_10
void FASTCALL call_12_11_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_3_11
#define init_12_11_3_11
void FASTCALL call_12_11_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_3_12
#define init_12_11_3_12
void FASTCALL call_12_11_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_3_13
#define init_12_11_3_13
void FASTCALL call_12_11_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_3_14
#define init_12_11_3_14
void FASTCALL call_12_11_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_3_15
#define init_12_11_3_15
void FASTCALL call_12_11_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_4_0
#define init_12_11_4_0
void FASTCALL call_12_11_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_4_1
#define init_12_11_4_1
void FASTCALL call_12_11_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_4_2
#define init_12_11_4_2
void FASTCALL call_12_11_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_4_3
#define init_12_11_4_3
void FASTCALL call_12_11_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_4_4
#define init_12_11_4_4
void FASTCALL call_12_11_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_4_5
#define init_12_11_4_5
void FASTCALL call_12_11_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_4_6
#define init_12_11_4_6
void FASTCALL call_12_11_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_4_7
#define init_12_11_4_7
void FASTCALL call_12_11_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_4_8
#define init_12_11_4_8
void FASTCALL call_12_11_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_4_9
#define init_12_11_4_9
void FASTCALL call_12_11_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_4_10
#define init_12_11_4_10
void FASTCALL call_12_11_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_4_11
#define init_12_11_4_11
void FASTCALL call_12_11_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_4_12
#define init_12_11_4_12
void FASTCALL call_12_11_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_4_13
#define init_12_11_4_13
void FASTCALL call_12_11_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_4_14
#define init_12_11_4_14
void FASTCALL call_12_11_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_4_15
#define init_12_11_4_15
void FASTCALL call_12_11_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_5_0
#define init_12_11_5_0
void FASTCALL call_12_11_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_5_1
#define init_12_11_5_1
void FASTCALL call_12_11_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_5_2
#define init_12_11_5_2
void FASTCALL call_12_11_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_5_3
#define init_12_11_5_3
void FASTCALL call_12_11_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_5_4
#define init_12_11_5_4
void FASTCALL call_12_11_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_5_5
#define init_12_11_5_5
void FASTCALL call_12_11_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_5_6
#define init_12_11_5_6
void FASTCALL call_12_11_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_5_7
#define init_12_11_5_7
void FASTCALL call_12_11_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_5_8
#define init_12_11_5_8
void FASTCALL call_12_11_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_5_9
#define init_12_11_5_9
void FASTCALL call_12_11_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_5_10
#define init_12_11_5_10
void FASTCALL call_12_11_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_5_11
#define init_12_11_5_11
void FASTCALL call_12_11_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_5_12
#define init_12_11_5_12
void FASTCALL call_12_11_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_5_13
#define init_12_11_5_13
void FASTCALL call_12_11_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_5_14
#define init_12_11_5_14
void FASTCALL call_12_11_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_5_15
#define init_12_11_5_15
void FASTCALL call_12_11_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_6_0
#define init_12_11_6_0
void FASTCALL call_12_11_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_6_1
#define init_12_11_6_1
void FASTCALL call_12_11_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_6_2
#define init_12_11_6_2
void FASTCALL call_12_11_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_6_3
#define init_12_11_6_3
void FASTCALL call_12_11_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_6_4
#define init_12_11_6_4
void FASTCALL call_12_11_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_6_5
#define init_12_11_6_5
void FASTCALL call_12_11_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_6_6
#define init_12_11_6_6
void FASTCALL call_12_11_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_6_7
#define init_12_11_6_7
void FASTCALL call_12_11_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_6_8
#define init_12_11_6_8
void FASTCALL call_12_11_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_6_9
#define init_12_11_6_9
void FASTCALL call_12_11_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_6_10
#define init_12_11_6_10
void FASTCALL call_12_11_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_6_11
#define init_12_11_6_11
void FASTCALL call_12_11_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_6_12
#define init_12_11_6_12
void FASTCALL call_12_11_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_6_13
#define init_12_11_6_13
void FASTCALL call_12_11_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_6_14
#define init_12_11_6_14
void FASTCALL call_12_11_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_6_15
#define init_12_11_6_15
void FASTCALL call_12_11_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_7_0
#define init_12_11_7_0
void FASTCALL call_12_11_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_7_1
#define init_12_11_7_1
void FASTCALL call_12_11_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_7_2
#define init_12_11_7_2
void FASTCALL call_12_11_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_7_3
#define init_12_11_7_3
void FASTCALL call_12_11_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_7_4
#define init_12_11_7_4
void FASTCALL call_12_11_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_7_5
#define init_12_11_7_5
void FASTCALL call_12_11_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_7_6
#define init_12_11_7_6
void FASTCALL call_12_11_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_7_7
#define init_12_11_7_7
void FASTCALL call_12_11_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_7_8
#define init_12_11_7_8
void FASTCALL call_12_11_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_7_9
#define init_12_11_7_9
void FASTCALL call_12_11_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_7_10
#define init_12_11_7_10
void FASTCALL call_12_11_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_7_11
#define init_12_11_7_11
void FASTCALL call_12_11_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_7_12
#define init_12_11_7_12
void FASTCALL call_12_11_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_7_13
#define init_12_11_7_13
void FASTCALL call_12_11_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_7_14
#define init_12_11_7_14
void FASTCALL call_12_11_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_7_15
#define init_12_11_7_15
void FASTCALL call_12_11_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_8_0
#define init_12_11_8_0
void FASTCALL call_12_11_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_8_1
#define init_12_11_8_1
void FASTCALL call_12_11_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_8_2
#define init_12_11_8_2
void FASTCALL call_12_11_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_8_3
#define init_12_11_8_3
void FASTCALL call_12_11_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_8_4
#define init_12_11_8_4
void FASTCALL call_12_11_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_8_5
#define init_12_11_8_5
void FASTCALL call_12_11_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_8_6
#define init_12_11_8_6
void FASTCALL call_12_11_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_8_7
#define init_12_11_8_7
void FASTCALL call_12_11_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_8_8
#define init_12_11_8_8
void FASTCALL call_12_11_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_8_9
#define init_12_11_8_9
void FASTCALL call_12_11_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_8_10
#define init_12_11_8_10
void FASTCALL call_12_11_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_8_11
#define init_12_11_8_11
void FASTCALL call_12_11_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_8_12
#define init_12_11_8_12
void FASTCALL call_12_11_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_8_13
#define init_12_11_8_13
void FASTCALL call_12_11_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_8_14
#define init_12_11_8_14
void FASTCALL call_12_11_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_8_15
#define init_12_11_8_15
void FASTCALL call_12_11_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_9_0
#define init_12_11_9_0
void FASTCALL call_12_11_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_9_1
#define init_12_11_9_1
void FASTCALL call_12_11_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_9_2
#define init_12_11_9_2
void FASTCALL call_12_11_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_9_3
#define init_12_11_9_3
void FASTCALL call_12_11_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_9_4
#define init_12_11_9_4
void FASTCALL call_12_11_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_9_5
#define init_12_11_9_5
void FASTCALL call_12_11_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_9_6
#define init_12_11_9_6
void FASTCALL call_12_11_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_9_7
#define init_12_11_9_7
void FASTCALL call_12_11_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_9_8
#define init_12_11_9_8
void FASTCALL call_12_11_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_9_9
#define init_12_11_9_9
void FASTCALL call_12_11_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_9_10
#define init_12_11_9_10
void FASTCALL call_12_11_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_9_11
#define init_12_11_9_11
void FASTCALL call_12_11_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_9_12
#define init_12_11_9_12
void FASTCALL call_12_11_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_9_13
#define init_12_11_9_13
void FASTCALL call_12_11_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_9_14
#define init_12_11_9_14
void FASTCALL call_12_11_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_9_15
#define init_12_11_9_15
void FASTCALL call_12_11_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_10_0
#define init_12_11_10_0
void FASTCALL call_12_11_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_10_1
#define init_12_11_10_1
void FASTCALL call_12_11_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_10_2
#define init_12_11_10_2
void FASTCALL call_12_11_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_10_3
#define init_12_11_10_3
void FASTCALL call_12_11_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_10_4
#define init_12_11_10_4
void FASTCALL call_12_11_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_10_5
#define init_12_11_10_5
void FASTCALL call_12_11_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_10_6
#define init_12_11_10_6
void FASTCALL call_12_11_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_10_7
#define init_12_11_10_7
void FASTCALL call_12_11_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_10_8
#define init_12_11_10_8
void FASTCALL call_12_11_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_10_9
#define init_12_11_10_9
void FASTCALL call_12_11_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_10_10
#define init_12_11_10_10
void FASTCALL call_12_11_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_10_11
#define init_12_11_10_11
void FASTCALL call_12_11_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_10_12
#define init_12_11_10_12
void FASTCALL call_12_11_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_10_13
#define init_12_11_10_13
void FASTCALL call_12_11_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_10_14
#define init_12_11_10_14
void FASTCALL call_12_11_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_10_15
#define init_12_11_10_15
void FASTCALL call_12_11_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_11_0
#define init_12_11_11_0
void FASTCALL call_12_11_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_11_1
#define init_12_11_11_1
void FASTCALL call_12_11_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_11_2
#define init_12_11_11_2
void FASTCALL call_12_11_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_11_3
#define init_12_11_11_3
void FASTCALL call_12_11_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_11_4
#define init_12_11_11_4
void FASTCALL call_12_11_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_11_5
#define init_12_11_11_5
void FASTCALL call_12_11_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_11_6
#define init_12_11_11_6
void FASTCALL call_12_11_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_11_7
#define init_12_11_11_7
void FASTCALL call_12_11_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_11_8
#define init_12_11_11_8
void FASTCALL call_12_11_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_11_9
#define init_12_11_11_9
void FASTCALL call_12_11_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_11_10
#define init_12_11_11_10
void FASTCALL call_12_11_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_11_11
#define init_12_11_11_11
void FASTCALL call_12_11_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_11_12
#define init_12_11_11_12
void FASTCALL call_12_11_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_11_13
#define init_12_11_11_13
void FASTCALL call_12_11_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_11_14
#define init_12_11_11_14
void FASTCALL call_12_11_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_11_15
#define init_12_11_11_15
void FASTCALL call_12_11_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_12_0
#define init_12_11_12_0
void FASTCALL call_12_11_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_12_1
#define init_12_11_12_1
void FASTCALL call_12_11_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_12_2
#define init_12_11_12_2
void FASTCALL call_12_11_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_12_3
#define init_12_11_12_3
void FASTCALL call_12_11_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_12_4
#define init_12_11_12_4
void FASTCALL call_12_11_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_12_5
#define init_12_11_12_5
void FASTCALL call_12_11_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_12_6
#define init_12_11_12_6
void FASTCALL call_12_11_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_12_7
#define init_12_11_12_7
void FASTCALL call_12_11_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_12_8
#define init_12_11_12_8
void FASTCALL call_12_11_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_12_9
#define init_12_11_12_9
void FASTCALL call_12_11_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_12_10
#define init_12_11_12_10
void FASTCALL call_12_11_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_12_11
#define init_12_11_12_11
void FASTCALL call_12_11_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_12_12
#define init_12_11_12_12
void FASTCALL call_12_11_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_12_13
#define init_12_11_12_13
void FASTCALL call_12_11_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_12_14
#define init_12_11_12_14
void FASTCALL call_12_11_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_12_15
#define init_12_11_12_15
void FASTCALL call_12_11_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_13_0
#define init_12_11_13_0
void FASTCALL call_12_11_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_13_1
#define init_12_11_13_1
void FASTCALL call_12_11_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_13_2
#define init_12_11_13_2
void FASTCALL call_12_11_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_13_3
#define init_12_11_13_3
void FASTCALL call_12_11_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_13_4
#define init_12_11_13_4
void FASTCALL call_12_11_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_13_5
#define init_12_11_13_5
void FASTCALL call_12_11_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_13_6
#define init_12_11_13_6
void FASTCALL call_12_11_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_13_7
#define init_12_11_13_7
void FASTCALL call_12_11_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_13_8
#define init_12_11_13_8
void FASTCALL call_12_11_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_13_9
#define init_12_11_13_9
void FASTCALL call_12_11_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_13_10
#define init_12_11_13_10
void FASTCALL call_12_11_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_13_11
#define init_12_11_13_11
void FASTCALL call_12_11_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_13_12
#define init_12_11_13_12
void FASTCALL call_12_11_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_13_13
#define init_12_11_13_13
void FASTCALL call_12_11_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_13_14
#define init_12_11_13_14
void FASTCALL call_12_11_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_13_15
#define init_12_11_13_15
void FASTCALL call_12_11_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_14_0
#define init_12_11_14_0
void FASTCALL call_12_11_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_14_1
#define init_12_11_14_1
void FASTCALL call_12_11_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_14_2
#define init_12_11_14_2
void FASTCALL call_12_11_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_14_3
#define init_12_11_14_3
void FASTCALL call_12_11_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_14_4
#define init_12_11_14_4
void FASTCALL call_12_11_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_14_5
#define init_12_11_14_5
void FASTCALL call_12_11_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_14_6
#define init_12_11_14_6
void FASTCALL call_12_11_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_14_7
#define init_12_11_14_7
void FASTCALL call_12_11_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_14_8
#define init_12_11_14_8
void FASTCALL call_12_11_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_14_9
#define init_12_11_14_9
void FASTCALL call_12_11_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_14_10
#define init_12_11_14_10
void FASTCALL call_12_11_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_14_11
#define init_12_11_14_11
void FASTCALL call_12_11_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_14_12
#define init_12_11_14_12
void FASTCALL call_12_11_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_14_13
#define init_12_11_14_13
void FASTCALL call_12_11_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_14_14
#define init_12_11_14_14
void FASTCALL call_12_11_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_14_15
#define init_12_11_14_15
void FASTCALL call_12_11_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_15_0
#define init_12_11_15_0
void FASTCALL call_12_11_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_15_1
#define init_12_11_15_1
void FASTCALL call_12_11_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_15_2
#define init_12_11_15_2
void FASTCALL call_12_11_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_15_3
#define init_12_11_15_3
void FASTCALL call_12_11_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_15_4
#define init_12_11_15_4
void FASTCALL call_12_11_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_15_5
#define init_12_11_15_5
void FASTCALL call_12_11_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_15_6
#define init_12_11_15_6
void FASTCALL call_12_11_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_15_7
#define init_12_11_15_7
void FASTCALL call_12_11_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_15_8
#define init_12_11_15_8
void FASTCALL call_12_11_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_15_9
#define init_12_11_15_9
void FASTCALL call_12_11_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_15_10
#define init_12_11_15_10
void FASTCALL call_12_11_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_15_11
#define init_12_11_15_11
void FASTCALL call_12_11_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_15_12
#define init_12_11_15_12
void FASTCALL call_12_11_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_15_13
#define init_12_11_15_13
void FASTCALL call_12_11_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_15_14
#define init_12_11_15_14
void FASTCALL call_12_11_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_11_15_15
#define init_12_11_15_15
void FASTCALL call_12_11_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_0_0
#define init_12_12_0_0
void FASTCALL call_12_12_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_0_1
#define init_12_12_0_1
void FASTCALL call_12_12_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_0_2
#define init_12_12_0_2
void FASTCALL call_12_12_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_0_3
#define init_12_12_0_3
void FASTCALL call_12_12_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_0_4
#define init_12_12_0_4
void FASTCALL call_12_12_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_0_5
#define init_12_12_0_5
void FASTCALL call_12_12_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_0_6
#define init_12_12_0_6
void FASTCALL call_12_12_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_0_7
#define init_12_12_0_7
void FASTCALL call_12_12_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_0_8
#define init_12_12_0_8
void FASTCALL call_12_12_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_0_9
#define init_12_12_0_9
void FASTCALL call_12_12_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_0_10
#define init_12_12_0_10
void FASTCALL call_12_12_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_0_11
#define init_12_12_0_11
void FASTCALL call_12_12_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_0_12
#define init_12_12_0_12
void FASTCALL call_12_12_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_0_13
#define init_12_12_0_13
void FASTCALL call_12_12_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_0_14
#define init_12_12_0_14
void FASTCALL call_12_12_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_0_15
#define init_12_12_0_15
void FASTCALL call_12_12_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_1_0
#define init_12_12_1_0
void FASTCALL call_12_12_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_1_1
#define init_12_12_1_1
void FASTCALL call_12_12_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_1_2
#define init_12_12_1_2
void FASTCALL call_12_12_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_1_3
#define init_12_12_1_3
void FASTCALL call_12_12_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_1_4
#define init_12_12_1_4
void FASTCALL call_12_12_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_1_5
#define init_12_12_1_5
void FASTCALL call_12_12_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_1_6
#define init_12_12_1_6
void FASTCALL call_12_12_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_1_7
#define init_12_12_1_7
void FASTCALL call_12_12_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_1_8
#define init_12_12_1_8
void FASTCALL call_12_12_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_1_9
#define init_12_12_1_9
void FASTCALL call_12_12_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_1_10
#define init_12_12_1_10
void FASTCALL call_12_12_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_1_11
#define init_12_12_1_11
void FASTCALL call_12_12_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_1_12
#define init_12_12_1_12
void FASTCALL call_12_12_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_1_13
#define init_12_12_1_13
void FASTCALL call_12_12_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_1_14
#define init_12_12_1_14
void FASTCALL call_12_12_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_1_15
#define init_12_12_1_15
void FASTCALL call_12_12_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_2_0
#define init_12_12_2_0
void FASTCALL call_12_12_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_2_1
#define init_12_12_2_1
void FASTCALL call_12_12_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_2_2
#define init_12_12_2_2
void FASTCALL call_12_12_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_2_3
#define init_12_12_2_3
void FASTCALL call_12_12_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_2_4
#define init_12_12_2_4
void FASTCALL call_12_12_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_2_5
#define init_12_12_2_5
void FASTCALL call_12_12_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_2_6
#define init_12_12_2_6
void FASTCALL call_12_12_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_2_7
#define init_12_12_2_7
void FASTCALL call_12_12_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_2_8
#define init_12_12_2_8
void FASTCALL call_12_12_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_2_9
#define init_12_12_2_9
void FASTCALL call_12_12_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_2_10
#define init_12_12_2_10
void FASTCALL call_12_12_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_2_11
#define init_12_12_2_11
void FASTCALL call_12_12_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_2_12
#define init_12_12_2_12
void FASTCALL call_12_12_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_2_13
#define init_12_12_2_13
void FASTCALL call_12_12_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_2_14
#define init_12_12_2_14
void FASTCALL call_12_12_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_2_15
#define init_12_12_2_15
void FASTCALL call_12_12_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_3_0
#define init_12_12_3_0
void FASTCALL call_12_12_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_3_1
#define init_12_12_3_1
void FASTCALL call_12_12_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_3_2
#define init_12_12_3_2
void FASTCALL call_12_12_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_3_3
#define init_12_12_3_3
void FASTCALL call_12_12_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_3_4
#define init_12_12_3_4
void FASTCALL call_12_12_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_3_5
#define init_12_12_3_5
void FASTCALL call_12_12_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_3_6
#define init_12_12_3_6
void FASTCALL call_12_12_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_3_7
#define init_12_12_3_7
void FASTCALL call_12_12_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_3_8
#define init_12_12_3_8
void FASTCALL call_12_12_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_3_9
#define init_12_12_3_9
void FASTCALL call_12_12_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_3_10
#define init_12_12_3_10
void FASTCALL call_12_12_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_3_11
#define init_12_12_3_11
void FASTCALL call_12_12_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_3_12
#define init_12_12_3_12
void FASTCALL call_12_12_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_3_13
#define init_12_12_3_13
void FASTCALL call_12_12_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_3_14
#define init_12_12_3_14
void FASTCALL call_12_12_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_3_15
#define init_12_12_3_15
void FASTCALL call_12_12_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_4_0
#define init_12_12_4_0
void FASTCALL call_12_12_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_4_1
#define init_12_12_4_1
void FASTCALL call_12_12_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_4_2
#define init_12_12_4_2
void FASTCALL call_12_12_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_4_3
#define init_12_12_4_3
void FASTCALL call_12_12_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_4_4
#define init_12_12_4_4
void FASTCALL call_12_12_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_4_5
#define init_12_12_4_5
void FASTCALL call_12_12_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_4_6
#define init_12_12_4_6
void FASTCALL call_12_12_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_4_7
#define init_12_12_4_7
void FASTCALL call_12_12_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_4_8
#define init_12_12_4_8
void FASTCALL call_12_12_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_4_9
#define init_12_12_4_9
void FASTCALL call_12_12_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_4_10
#define init_12_12_4_10
void FASTCALL call_12_12_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_4_11
#define init_12_12_4_11
void FASTCALL call_12_12_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_4_12
#define init_12_12_4_12
void FASTCALL call_12_12_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_4_13
#define init_12_12_4_13
void FASTCALL call_12_12_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_4_14
#define init_12_12_4_14
void FASTCALL call_12_12_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_4_15
#define init_12_12_4_15
void FASTCALL call_12_12_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_5_0
#define init_12_12_5_0
void FASTCALL call_12_12_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_5_1
#define init_12_12_5_1
void FASTCALL call_12_12_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_5_2
#define init_12_12_5_2
void FASTCALL call_12_12_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_5_3
#define init_12_12_5_3
void FASTCALL call_12_12_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_5_4
#define init_12_12_5_4
void FASTCALL call_12_12_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_5_5
#define init_12_12_5_5
void FASTCALL call_12_12_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_5_6
#define init_12_12_5_6
void FASTCALL call_12_12_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_5_7
#define init_12_12_5_7
void FASTCALL call_12_12_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_5_8
#define init_12_12_5_8
void FASTCALL call_12_12_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_5_9
#define init_12_12_5_9
void FASTCALL call_12_12_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_5_10
#define init_12_12_5_10
void FASTCALL call_12_12_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_5_11
#define init_12_12_5_11
void FASTCALL call_12_12_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_5_12
#define init_12_12_5_12
void FASTCALL call_12_12_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_5_13
#define init_12_12_5_13
void FASTCALL call_12_12_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_5_14
#define init_12_12_5_14
void FASTCALL call_12_12_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_5_15
#define init_12_12_5_15
void FASTCALL call_12_12_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_6_0
#define init_12_12_6_0
void FASTCALL call_12_12_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_6_1
#define init_12_12_6_1
void FASTCALL call_12_12_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_6_2
#define init_12_12_6_2
void FASTCALL call_12_12_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_6_3
#define init_12_12_6_3
void FASTCALL call_12_12_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_6_4
#define init_12_12_6_4
void FASTCALL call_12_12_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_6_5
#define init_12_12_6_5
void FASTCALL call_12_12_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_6_6
#define init_12_12_6_6
void FASTCALL call_12_12_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_6_7
#define init_12_12_6_7
void FASTCALL call_12_12_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_6_8
#define init_12_12_6_8
void FASTCALL call_12_12_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_6_9
#define init_12_12_6_9
void FASTCALL call_12_12_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_6_10
#define init_12_12_6_10
void FASTCALL call_12_12_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_6_11
#define init_12_12_6_11
void FASTCALL call_12_12_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_6_12
#define init_12_12_6_12
void FASTCALL call_12_12_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_6_13
#define init_12_12_6_13
void FASTCALL call_12_12_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_6_14
#define init_12_12_6_14
void FASTCALL call_12_12_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_6_15
#define init_12_12_6_15
void FASTCALL call_12_12_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_7_0
#define init_12_12_7_0
void FASTCALL call_12_12_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_7_1
#define init_12_12_7_1
void FASTCALL call_12_12_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_7_2
#define init_12_12_7_2
void FASTCALL call_12_12_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_7_3
#define init_12_12_7_3
void FASTCALL call_12_12_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_7_4
#define init_12_12_7_4
void FASTCALL call_12_12_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_7_5
#define init_12_12_7_5
void FASTCALL call_12_12_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_7_6
#define init_12_12_7_6
void FASTCALL call_12_12_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_7_7
#define init_12_12_7_7
void FASTCALL call_12_12_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_7_8
#define init_12_12_7_8
void FASTCALL call_12_12_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_7_9
#define init_12_12_7_9
void FASTCALL call_12_12_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_7_10
#define init_12_12_7_10
void FASTCALL call_12_12_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_7_11
#define init_12_12_7_11
void FASTCALL call_12_12_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_7_12
#define init_12_12_7_12
void FASTCALL call_12_12_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_7_13
#define init_12_12_7_13
void FASTCALL call_12_12_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_7_14
#define init_12_12_7_14
void FASTCALL call_12_12_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_7_15
#define init_12_12_7_15
void FASTCALL call_12_12_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_8_0
#define init_12_12_8_0
void FASTCALL call_12_12_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_8_1
#define init_12_12_8_1
void FASTCALL call_12_12_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_8_2
#define init_12_12_8_2
void FASTCALL call_12_12_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_8_3
#define init_12_12_8_3
void FASTCALL call_12_12_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_8_4
#define init_12_12_8_4
void FASTCALL call_12_12_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_8_5
#define init_12_12_8_5
void FASTCALL call_12_12_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_8_6
#define init_12_12_8_6
void FASTCALL call_12_12_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_8_7
#define init_12_12_8_7
void FASTCALL call_12_12_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_8_8
#define init_12_12_8_8
void FASTCALL call_12_12_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_8_9
#define init_12_12_8_9
void FASTCALL call_12_12_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_8_10
#define init_12_12_8_10
void FASTCALL call_12_12_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_8_11
#define init_12_12_8_11
void FASTCALL call_12_12_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_8_12
#define init_12_12_8_12
void FASTCALL call_12_12_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_8_13
#define init_12_12_8_13
void FASTCALL call_12_12_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_8_14
#define init_12_12_8_14
void FASTCALL call_12_12_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_8_15
#define init_12_12_8_15
void FASTCALL call_12_12_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_9_0
#define init_12_12_9_0
void FASTCALL call_12_12_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_9_1
#define init_12_12_9_1
void FASTCALL call_12_12_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_9_2
#define init_12_12_9_2
void FASTCALL call_12_12_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_9_3
#define init_12_12_9_3
void FASTCALL call_12_12_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_9_4
#define init_12_12_9_4
void FASTCALL call_12_12_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_9_5
#define init_12_12_9_5
void FASTCALL call_12_12_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_9_6
#define init_12_12_9_6
void FASTCALL call_12_12_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_9_7
#define init_12_12_9_7
void FASTCALL call_12_12_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_9_8
#define init_12_12_9_8
void FASTCALL call_12_12_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_9_9
#define init_12_12_9_9
void FASTCALL call_12_12_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_9_10
#define init_12_12_9_10
void FASTCALL call_12_12_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_9_11
#define init_12_12_9_11
void FASTCALL call_12_12_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_9_12
#define init_12_12_9_12
void FASTCALL call_12_12_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_9_13
#define init_12_12_9_13
void FASTCALL call_12_12_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_9_14
#define init_12_12_9_14
void FASTCALL call_12_12_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_9_15
#define init_12_12_9_15
void FASTCALL call_12_12_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_10_0
#define init_12_12_10_0
void FASTCALL call_12_12_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_10_1
#define init_12_12_10_1
void FASTCALL call_12_12_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_10_2
#define init_12_12_10_2
void FASTCALL call_12_12_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_10_3
#define init_12_12_10_3
void FASTCALL call_12_12_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_10_4
#define init_12_12_10_4
void FASTCALL call_12_12_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_10_5
#define init_12_12_10_5
void FASTCALL call_12_12_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_10_6
#define init_12_12_10_6
void FASTCALL call_12_12_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_10_7
#define init_12_12_10_7
void FASTCALL call_12_12_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_10_8
#define init_12_12_10_8
void FASTCALL call_12_12_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_10_9
#define init_12_12_10_9
void FASTCALL call_12_12_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_10_10
#define init_12_12_10_10
void FASTCALL call_12_12_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_10_11
#define init_12_12_10_11
void FASTCALL call_12_12_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_10_12
#define init_12_12_10_12
void FASTCALL call_12_12_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_10_13
#define init_12_12_10_13
void FASTCALL call_12_12_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_10_14
#define init_12_12_10_14
void FASTCALL call_12_12_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_10_15
#define init_12_12_10_15
void FASTCALL call_12_12_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_11_0
#define init_12_12_11_0
void FASTCALL call_12_12_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_11_1
#define init_12_12_11_1
void FASTCALL call_12_12_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_11_2
#define init_12_12_11_2
void FASTCALL call_12_12_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_11_3
#define init_12_12_11_3
void FASTCALL call_12_12_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_11_4
#define init_12_12_11_4
void FASTCALL call_12_12_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_11_5
#define init_12_12_11_5
void FASTCALL call_12_12_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_11_6
#define init_12_12_11_6
void FASTCALL call_12_12_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_11_7
#define init_12_12_11_7
void FASTCALL call_12_12_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_11_8
#define init_12_12_11_8
void FASTCALL call_12_12_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_11_9
#define init_12_12_11_9
void FASTCALL call_12_12_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_11_10
#define init_12_12_11_10
void FASTCALL call_12_12_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_11_11
#define init_12_12_11_11
void FASTCALL call_12_12_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_11_12
#define init_12_12_11_12
void FASTCALL call_12_12_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_11_13
#define init_12_12_11_13
void FASTCALL call_12_12_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_11_14
#define init_12_12_11_14
void FASTCALL call_12_12_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_11_15
#define init_12_12_11_15
void FASTCALL call_12_12_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_12_0
#define init_12_12_12_0
void FASTCALL call_12_12_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_12_1
#define init_12_12_12_1
void FASTCALL call_12_12_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_12_2
#define init_12_12_12_2
void FASTCALL call_12_12_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_12_3
#define init_12_12_12_3
void FASTCALL call_12_12_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_12_4
#define init_12_12_12_4
void FASTCALL call_12_12_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_12_5
#define init_12_12_12_5
void FASTCALL call_12_12_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_12_6
#define init_12_12_12_6
void FASTCALL call_12_12_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_12_7
#define init_12_12_12_7
void FASTCALL call_12_12_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_12_8
#define init_12_12_12_8
void FASTCALL call_12_12_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_12_9
#define init_12_12_12_9
void FASTCALL call_12_12_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_12_10
#define init_12_12_12_10
void FASTCALL call_12_12_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_12_11
#define init_12_12_12_11
void FASTCALL call_12_12_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_12_12
#define init_12_12_12_12
void FASTCALL call_12_12_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_12_13
#define init_12_12_12_13
void FASTCALL call_12_12_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_12_14
#define init_12_12_12_14
void FASTCALL call_12_12_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_12_15
#define init_12_12_12_15
void FASTCALL call_12_12_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_13_0
#define init_12_12_13_0
void FASTCALL call_12_12_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_13_1
#define init_12_12_13_1
void FASTCALL call_12_12_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_13_2
#define init_12_12_13_2
void FASTCALL call_12_12_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_13_3
#define init_12_12_13_3
void FASTCALL call_12_12_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_13_4
#define init_12_12_13_4
void FASTCALL call_12_12_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_13_5
#define init_12_12_13_5
void FASTCALL call_12_12_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_13_6
#define init_12_12_13_6
void FASTCALL call_12_12_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_13_7
#define init_12_12_13_7
void FASTCALL call_12_12_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_13_8
#define init_12_12_13_8
void FASTCALL call_12_12_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_13_9
#define init_12_12_13_9
void FASTCALL call_12_12_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_13_10
#define init_12_12_13_10
void FASTCALL call_12_12_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_13_11
#define init_12_12_13_11
void FASTCALL call_12_12_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_13_12
#define init_12_12_13_12
void FASTCALL call_12_12_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_13_13
#define init_12_12_13_13
void FASTCALL call_12_12_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_13_14
#define init_12_12_13_14
void FASTCALL call_12_12_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_13_15
#define init_12_12_13_15
void FASTCALL call_12_12_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_14_0
#define init_12_12_14_0
void FASTCALL call_12_12_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_14_1
#define init_12_12_14_1
void FASTCALL call_12_12_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_14_2
#define init_12_12_14_2
void FASTCALL call_12_12_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_14_3
#define init_12_12_14_3
void FASTCALL call_12_12_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_14_4
#define init_12_12_14_4
void FASTCALL call_12_12_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_14_5
#define init_12_12_14_5
void FASTCALL call_12_12_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_14_6
#define init_12_12_14_6
void FASTCALL call_12_12_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_14_7
#define init_12_12_14_7
void FASTCALL call_12_12_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_14_8
#define init_12_12_14_8
void FASTCALL call_12_12_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_14_9
#define init_12_12_14_9
void FASTCALL call_12_12_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_14_10
#define init_12_12_14_10
void FASTCALL call_12_12_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_14_11
#define init_12_12_14_11
void FASTCALL call_12_12_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_14_12
#define init_12_12_14_12
void FASTCALL call_12_12_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_14_13
#define init_12_12_14_13
void FASTCALL call_12_12_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_14_14
#define init_12_12_14_14
void FASTCALL call_12_12_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_14_15
#define init_12_12_14_15
void FASTCALL call_12_12_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_15_0
#define init_12_12_15_0
void FASTCALL call_12_12_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_15_1
#define init_12_12_15_1
void FASTCALL call_12_12_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_15_2
#define init_12_12_15_2
void FASTCALL call_12_12_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_15_3
#define init_12_12_15_3
void FASTCALL call_12_12_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_15_4
#define init_12_12_15_4
void FASTCALL call_12_12_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_15_5
#define init_12_12_15_5
void FASTCALL call_12_12_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_15_6
#define init_12_12_15_6
void FASTCALL call_12_12_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_15_7
#define init_12_12_15_7
void FASTCALL call_12_12_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_15_8
#define init_12_12_15_8
void FASTCALL call_12_12_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_15_9
#define init_12_12_15_9
void FASTCALL call_12_12_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_15_10
#define init_12_12_15_10
void FASTCALL call_12_12_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_15_11
#define init_12_12_15_11
void FASTCALL call_12_12_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_15_12
#define init_12_12_15_12
void FASTCALL call_12_12_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_15_13
#define init_12_12_15_13
void FASTCALL call_12_12_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_15_14
#define init_12_12_15_14
void FASTCALL call_12_12_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_12_15_15
#define init_12_12_15_15
void FASTCALL call_12_12_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_0_0
#define init_12_13_0_0
void FASTCALL call_12_13_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_0_1
#define init_12_13_0_1
void FASTCALL call_12_13_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_0_2
#define init_12_13_0_2
void FASTCALL call_12_13_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_0_3
#define init_12_13_0_3
void FASTCALL call_12_13_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_0_4
#define init_12_13_0_4
void FASTCALL call_12_13_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_0_5
#define init_12_13_0_5
void FASTCALL call_12_13_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_0_6
#define init_12_13_0_6
void FASTCALL call_12_13_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_0_7
#define init_12_13_0_7
void FASTCALL call_12_13_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_0_8
#define init_12_13_0_8
void FASTCALL call_12_13_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_0_9
#define init_12_13_0_9
void FASTCALL call_12_13_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_0_10
#define init_12_13_0_10
void FASTCALL call_12_13_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_0_11
#define init_12_13_0_11
void FASTCALL call_12_13_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_0_12
#define init_12_13_0_12
void FASTCALL call_12_13_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_0_13
#define init_12_13_0_13
void FASTCALL call_12_13_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_0_14
#define init_12_13_0_14
void FASTCALL call_12_13_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_0_15
#define init_12_13_0_15
void FASTCALL call_12_13_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_1_0
#define init_12_13_1_0
void FASTCALL call_12_13_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_1_1
#define init_12_13_1_1
void FASTCALL call_12_13_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_1_2
#define init_12_13_1_2
void FASTCALL call_12_13_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_1_3
#define init_12_13_1_3
void FASTCALL call_12_13_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_1_4
#define init_12_13_1_4
void FASTCALL call_12_13_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_1_5
#define init_12_13_1_5
void FASTCALL call_12_13_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_1_6
#define init_12_13_1_6
void FASTCALL call_12_13_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_1_7
#define init_12_13_1_7
void FASTCALL call_12_13_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_1_8
#define init_12_13_1_8
void FASTCALL call_12_13_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_1_9
#define init_12_13_1_9
void FASTCALL call_12_13_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_1_10
#define init_12_13_1_10
void FASTCALL call_12_13_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_1_11
#define init_12_13_1_11
void FASTCALL call_12_13_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_1_12
#define init_12_13_1_12
void FASTCALL call_12_13_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_1_13
#define init_12_13_1_13
void FASTCALL call_12_13_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_1_14
#define init_12_13_1_14
void FASTCALL call_12_13_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_1_15
#define init_12_13_1_15
void FASTCALL call_12_13_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_2_0
#define init_12_13_2_0
void FASTCALL call_12_13_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_2_1
#define init_12_13_2_1
void FASTCALL call_12_13_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_2_2
#define init_12_13_2_2
void FASTCALL call_12_13_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_2_3
#define init_12_13_2_3
void FASTCALL call_12_13_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_2_4
#define init_12_13_2_4
void FASTCALL call_12_13_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_2_5
#define init_12_13_2_5
void FASTCALL call_12_13_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_2_6
#define init_12_13_2_6
void FASTCALL call_12_13_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_2_7
#define init_12_13_2_7
void FASTCALL call_12_13_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_2_8
#define init_12_13_2_8
void FASTCALL call_12_13_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_2_9
#define init_12_13_2_9
void FASTCALL call_12_13_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_2_10
#define init_12_13_2_10
void FASTCALL call_12_13_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_2_11
#define init_12_13_2_11
void FASTCALL call_12_13_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_2_12
#define init_12_13_2_12
void FASTCALL call_12_13_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_2_13
#define init_12_13_2_13
void FASTCALL call_12_13_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_2_14
#define init_12_13_2_14
void FASTCALL call_12_13_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_2_15
#define init_12_13_2_15
void FASTCALL call_12_13_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_3_0
#define init_12_13_3_0
void FASTCALL call_12_13_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_3_1
#define init_12_13_3_1
void FASTCALL call_12_13_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_3_2
#define init_12_13_3_2
void FASTCALL call_12_13_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_3_3
#define init_12_13_3_3
void FASTCALL call_12_13_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_3_4
#define init_12_13_3_4
void FASTCALL call_12_13_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_3_5
#define init_12_13_3_5
void FASTCALL call_12_13_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_3_6
#define init_12_13_3_6
void FASTCALL call_12_13_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_3_7
#define init_12_13_3_7
void FASTCALL call_12_13_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_3_8
#define init_12_13_3_8
void FASTCALL call_12_13_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_3_9
#define init_12_13_3_9
void FASTCALL call_12_13_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_3_10
#define init_12_13_3_10
void FASTCALL call_12_13_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_3_11
#define init_12_13_3_11
void FASTCALL call_12_13_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_3_12
#define init_12_13_3_12
void FASTCALL call_12_13_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_3_13
#define init_12_13_3_13
void FASTCALL call_12_13_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_3_14
#define init_12_13_3_14
void FASTCALL call_12_13_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_3_15
#define init_12_13_3_15
void FASTCALL call_12_13_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_4_0
#define init_12_13_4_0
void FASTCALL call_12_13_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_4_1
#define init_12_13_4_1
void FASTCALL call_12_13_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_4_2
#define init_12_13_4_2
void FASTCALL call_12_13_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_4_3
#define init_12_13_4_3
void FASTCALL call_12_13_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_4_4
#define init_12_13_4_4
void FASTCALL call_12_13_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_4_5
#define init_12_13_4_5
void FASTCALL call_12_13_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_4_6
#define init_12_13_4_6
void FASTCALL call_12_13_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_4_7
#define init_12_13_4_7
void FASTCALL call_12_13_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_4_8
#define init_12_13_4_8
void FASTCALL call_12_13_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_4_9
#define init_12_13_4_9
void FASTCALL call_12_13_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_4_10
#define init_12_13_4_10
void FASTCALL call_12_13_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_4_11
#define init_12_13_4_11
void FASTCALL call_12_13_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_4_12
#define init_12_13_4_12
void FASTCALL call_12_13_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_4_13
#define init_12_13_4_13
void FASTCALL call_12_13_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_4_14
#define init_12_13_4_14
void FASTCALL call_12_13_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_4_15
#define init_12_13_4_15
void FASTCALL call_12_13_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_5_0
#define init_12_13_5_0
void FASTCALL call_12_13_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_5_1
#define init_12_13_5_1
void FASTCALL call_12_13_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_5_2
#define init_12_13_5_2
void FASTCALL call_12_13_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_5_3
#define init_12_13_5_3
void FASTCALL call_12_13_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_5_4
#define init_12_13_5_4
void FASTCALL call_12_13_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_5_5
#define init_12_13_5_5
void FASTCALL call_12_13_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_5_6
#define init_12_13_5_6
void FASTCALL call_12_13_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_5_7
#define init_12_13_5_7
void FASTCALL call_12_13_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_5_8
#define init_12_13_5_8
void FASTCALL call_12_13_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_5_9
#define init_12_13_5_9
void FASTCALL call_12_13_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_5_10
#define init_12_13_5_10
void FASTCALL call_12_13_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_5_11
#define init_12_13_5_11
void FASTCALL call_12_13_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_5_12
#define init_12_13_5_12
void FASTCALL call_12_13_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_5_13
#define init_12_13_5_13
void FASTCALL call_12_13_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_5_14
#define init_12_13_5_14
void FASTCALL call_12_13_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_5_15
#define init_12_13_5_15
void FASTCALL call_12_13_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_6_0
#define init_12_13_6_0
void FASTCALL call_12_13_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_6_1
#define init_12_13_6_1
void FASTCALL call_12_13_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_6_2
#define init_12_13_6_2
void FASTCALL call_12_13_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_6_3
#define init_12_13_6_3
void FASTCALL call_12_13_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_6_4
#define init_12_13_6_4
void FASTCALL call_12_13_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_6_5
#define init_12_13_6_5
void FASTCALL call_12_13_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_6_6
#define init_12_13_6_6
void FASTCALL call_12_13_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_6_7
#define init_12_13_6_7
void FASTCALL call_12_13_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_6_8
#define init_12_13_6_8
void FASTCALL call_12_13_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_6_9
#define init_12_13_6_9
void FASTCALL call_12_13_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_6_10
#define init_12_13_6_10
void FASTCALL call_12_13_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_6_11
#define init_12_13_6_11
void FASTCALL call_12_13_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_6_12
#define init_12_13_6_12
void FASTCALL call_12_13_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_6_13
#define init_12_13_6_13
void FASTCALL call_12_13_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_6_14
#define init_12_13_6_14
void FASTCALL call_12_13_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_6_15
#define init_12_13_6_15
void FASTCALL call_12_13_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_7_0
#define init_12_13_7_0
void FASTCALL call_12_13_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_7_1
#define init_12_13_7_1
void FASTCALL call_12_13_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_7_2
#define init_12_13_7_2
void FASTCALL call_12_13_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_7_3
#define init_12_13_7_3
void FASTCALL call_12_13_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_7_4
#define init_12_13_7_4
void FASTCALL call_12_13_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_7_5
#define init_12_13_7_5
void FASTCALL call_12_13_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_7_6
#define init_12_13_7_6
void FASTCALL call_12_13_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_7_7
#define init_12_13_7_7
void FASTCALL call_12_13_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_7_8
#define init_12_13_7_8
void FASTCALL call_12_13_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_7_9
#define init_12_13_7_9
void FASTCALL call_12_13_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_7_10
#define init_12_13_7_10
void FASTCALL call_12_13_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_7_11
#define init_12_13_7_11
void FASTCALL call_12_13_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_7_12
#define init_12_13_7_12
void FASTCALL call_12_13_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_7_13
#define init_12_13_7_13
void FASTCALL call_12_13_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_7_14
#define init_12_13_7_14
void FASTCALL call_12_13_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_7_15
#define init_12_13_7_15
void FASTCALL call_12_13_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_8_0
#define init_12_13_8_0
void FASTCALL call_12_13_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_8_1
#define init_12_13_8_1
void FASTCALL call_12_13_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_8_2
#define init_12_13_8_2
void FASTCALL call_12_13_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_8_3
#define init_12_13_8_3
void FASTCALL call_12_13_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_8_4
#define init_12_13_8_4
void FASTCALL call_12_13_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_8_5
#define init_12_13_8_5
void FASTCALL call_12_13_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_8_6
#define init_12_13_8_6
void FASTCALL call_12_13_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_8_7
#define init_12_13_8_7
void FASTCALL call_12_13_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_8_8
#define init_12_13_8_8
void FASTCALL call_12_13_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_8_9
#define init_12_13_8_9
void FASTCALL call_12_13_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_8_10
#define init_12_13_8_10
void FASTCALL call_12_13_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_8_11
#define init_12_13_8_11
void FASTCALL call_12_13_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_8_12
#define init_12_13_8_12
void FASTCALL call_12_13_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_8_13
#define init_12_13_8_13
void FASTCALL call_12_13_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_8_14
#define init_12_13_8_14
void FASTCALL call_12_13_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_8_15
#define init_12_13_8_15
void FASTCALL call_12_13_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_9_0
#define init_12_13_9_0
void FASTCALL call_12_13_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_9_1
#define init_12_13_9_1
void FASTCALL call_12_13_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_9_2
#define init_12_13_9_2
void FASTCALL call_12_13_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_9_3
#define init_12_13_9_3
void FASTCALL call_12_13_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_9_4
#define init_12_13_9_4
void FASTCALL call_12_13_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_9_5
#define init_12_13_9_5
void FASTCALL call_12_13_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_9_6
#define init_12_13_9_6
void FASTCALL call_12_13_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_9_7
#define init_12_13_9_7
void FASTCALL call_12_13_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_9_8
#define init_12_13_9_8
void FASTCALL call_12_13_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_9_9
#define init_12_13_9_9
void FASTCALL call_12_13_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_9_10
#define init_12_13_9_10
void FASTCALL call_12_13_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_9_11
#define init_12_13_9_11
void FASTCALL call_12_13_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_9_12
#define init_12_13_9_12
void FASTCALL call_12_13_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_9_13
#define init_12_13_9_13
void FASTCALL call_12_13_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_9_14
#define init_12_13_9_14
void FASTCALL call_12_13_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_9_15
#define init_12_13_9_15
void FASTCALL call_12_13_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_10_0
#define init_12_13_10_0
void FASTCALL call_12_13_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_10_1
#define init_12_13_10_1
void FASTCALL call_12_13_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_10_2
#define init_12_13_10_2
void FASTCALL call_12_13_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_10_3
#define init_12_13_10_3
void FASTCALL call_12_13_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_10_4
#define init_12_13_10_4
void FASTCALL call_12_13_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_10_5
#define init_12_13_10_5
void FASTCALL call_12_13_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_10_6
#define init_12_13_10_6
void FASTCALL call_12_13_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_10_7
#define init_12_13_10_7
void FASTCALL call_12_13_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_10_8
#define init_12_13_10_8
void FASTCALL call_12_13_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_10_9
#define init_12_13_10_9
void FASTCALL call_12_13_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_10_10
#define init_12_13_10_10
void FASTCALL call_12_13_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_10_11
#define init_12_13_10_11
void FASTCALL call_12_13_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_10_12
#define init_12_13_10_12
void FASTCALL call_12_13_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_10_13
#define init_12_13_10_13
void FASTCALL call_12_13_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_10_14
#define init_12_13_10_14
void FASTCALL call_12_13_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_10_15
#define init_12_13_10_15
void FASTCALL call_12_13_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_11_0
#define init_12_13_11_0
void FASTCALL call_12_13_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_11_1
#define init_12_13_11_1
void FASTCALL call_12_13_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_11_2
#define init_12_13_11_2
void FASTCALL call_12_13_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_11_3
#define init_12_13_11_3
void FASTCALL call_12_13_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_11_4
#define init_12_13_11_4
void FASTCALL call_12_13_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_11_5
#define init_12_13_11_5
void FASTCALL call_12_13_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_11_6
#define init_12_13_11_6
void FASTCALL call_12_13_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_11_7
#define init_12_13_11_7
void FASTCALL call_12_13_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_11_8
#define init_12_13_11_8
void FASTCALL call_12_13_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_11_9
#define init_12_13_11_9
void FASTCALL call_12_13_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_11_10
#define init_12_13_11_10
void FASTCALL call_12_13_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_11_11
#define init_12_13_11_11
void FASTCALL call_12_13_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_11_12
#define init_12_13_11_12
void FASTCALL call_12_13_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_11_13
#define init_12_13_11_13
void FASTCALL call_12_13_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_11_14
#define init_12_13_11_14
void FASTCALL call_12_13_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_11_15
#define init_12_13_11_15
void FASTCALL call_12_13_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_12_0
#define init_12_13_12_0
void FASTCALL call_12_13_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_12_1
#define init_12_13_12_1
void FASTCALL call_12_13_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_12_2
#define init_12_13_12_2
void FASTCALL call_12_13_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_12_3
#define init_12_13_12_3
void FASTCALL call_12_13_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_12_4
#define init_12_13_12_4
void FASTCALL call_12_13_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_12_5
#define init_12_13_12_5
void FASTCALL call_12_13_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_12_6
#define init_12_13_12_6
void FASTCALL call_12_13_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_12_7
#define init_12_13_12_7
void FASTCALL call_12_13_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_12_8
#define init_12_13_12_8
void FASTCALL call_12_13_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_12_9
#define init_12_13_12_9
void FASTCALL call_12_13_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_12_10
#define init_12_13_12_10
void FASTCALL call_12_13_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_12_11
#define init_12_13_12_11
void FASTCALL call_12_13_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_12_12
#define init_12_13_12_12
void FASTCALL call_12_13_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_12_13
#define init_12_13_12_13
void FASTCALL call_12_13_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_12_14
#define init_12_13_12_14
void FASTCALL call_12_13_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_12_15
#define init_12_13_12_15
void FASTCALL call_12_13_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_13_0
#define init_12_13_13_0
void FASTCALL call_12_13_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_13_1
#define init_12_13_13_1
void FASTCALL call_12_13_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_13_2
#define init_12_13_13_2
void FASTCALL call_12_13_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_13_3
#define init_12_13_13_3
void FASTCALL call_12_13_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_13_4
#define init_12_13_13_4
void FASTCALL call_12_13_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_13_5
#define init_12_13_13_5
void FASTCALL call_12_13_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_13_6
#define init_12_13_13_6
void FASTCALL call_12_13_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_13_7
#define init_12_13_13_7
void FASTCALL call_12_13_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_13_8
#define init_12_13_13_8
void FASTCALL call_12_13_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_13_9
#define init_12_13_13_9
void FASTCALL call_12_13_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_13_10
#define init_12_13_13_10
void FASTCALL call_12_13_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_13_11
#define init_12_13_13_11
void FASTCALL call_12_13_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_13_12
#define init_12_13_13_12
void FASTCALL call_12_13_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_13_13
#define init_12_13_13_13
void FASTCALL call_12_13_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_13_14
#define init_12_13_13_14
void FASTCALL call_12_13_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_13_15
#define init_12_13_13_15
void FASTCALL call_12_13_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_14_0
#define init_12_13_14_0
void FASTCALL call_12_13_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_14_1
#define init_12_13_14_1
void FASTCALL call_12_13_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_14_2
#define init_12_13_14_2
void FASTCALL call_12_13_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_14_3
#define init_12_13_14_3
void FASTCALL call_12_13_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_14_4
#define init_12_13_14_4
void FASTCALL call_12_13_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_14_5
#define init_12_13_14_5
void FASTCALL call_12_13_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_14_6
#define init_12_13_14_6
void FASTCALL call_12_13_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_14_7
#define init_12_13_14_7
void FASTCALL call_12_13_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_14_8
#define init_12_13_14_8
void FASTCALL call_12_13_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_14_9
#define init_12_13_14_9
void FASTCALL call_12_13_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_14_10
#define init_12_13_14_10
void FASTCALL call_12_13_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_14_11
#define init_12_13_14_11
void FASTCALL call_12_13_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_14_12
#define init_12_13_14_12
void FASTCALL call_12_13_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_14_13
#define init_12_13_14_13
void FASTCALL call_12_13_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_14_14
#define init_12_13_14_14
void FASTCALL call_12_13_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_14_15
#define init_12_13_14_15
void FASTCALL call_12_13_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_15_0
#define init_12_13_15_0
void FASTCALL call_12_13_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_15_1
#define init_12_13_15_1
void FASTCALL call_12_13_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_15_2
#define init_12_13_15_2
void FASTCALL call_12_13_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_15_3
#define init_12_13_15_3
void FASTCALL call_12_13_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_15_4
#define init_12_13_15_4
void FASTCALL call_12_13_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_15_5
#define init_12_13_15_5
void FASTCALL call_12_13_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_15_6
#define init_12_13_15_6
void FASTCALL call_12_13_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_15_7
#define init_12_13_15_7
void FASTCALL call_12_13_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_15_8
#define init_12_13_15_8
void FASTCALL call_12_13_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_15_9
#define init_12_13_15_9
void FASTCALL call_12_13_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_15_10
#define init_12_13_15_10
void FASTCALL call_12_13_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_15_11
#define init_12_13_15_11
void FASTCALL call_12_13_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_15_12
#define init_12_13_15_12
void FASTCALL call_12_13_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_15_13
#define init_12_13_15_13
void FASTCALL call_12_13_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_15_14
#define init_12_13_15_14
void FASTCALL call_12_13_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_13_15_15
#define init_12_13_15_15
void FASTCALL call_12_13_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_0_0
#define init_12_14_0_0
void FASTCALL call_12_14_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_0_1
#define init_12_14_0_1
void FASTCALL call_12_14_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_0_2
#define init_12_14_0_2
void FASTCALL call_12_14_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_0_3
#define init_12_14_0_3
void FASTCALL call_12_14_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_0_4
#define init_12_14_0_4
void FASTCALL call_12_14_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_0_5
#define init_12_14_0_5
void FASTCALL call_12_14_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_0_6
#define init_12_14_0_6
void FASTCALL call_12_14_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_0_7
#define init_12_14_0_7
void FASTCALL call_12_14_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_0_8
#define init_12_14_0_8
void FASTCALL call_12_14_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_0_9
#define init_12_14_0_9
void FASTCALL call_12_14_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_0_10
#define init_12_14_0_10
void FASTCALL call_12_14_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_0_11
#define init_12_14_0_11
void FASTCALL call_12_14_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_0_12
#define init_12_14_0_12
void FASTCALL call_12_14_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_0_13
#define init_12_14_0_13
void FASTCALL call_12_14_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_0_14
#define init_12_14_0_14
void FASTCALL call_12_14_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_0_15
#define init_12_14_0_15
void FASTCALL call_12_14_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_1_0
#define init_12_14_1_0
void FASTCALL call_12_14_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_1_1
#define init_12_14_1_1
void FASTCALL call_12_14_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_1_2
#define init_12_14_1_2
void FASTCALL call_12_14_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_1_3
#define init_12_14_1_3
void FASTCALL call_12_14_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_1_4
#define init_12_14_1_4
void FASTCALL call_12_14_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_1_5
#define init_12_14_1_5
void FASTCALL call_12_14_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_1_6
#define init_12_14_1_6
void FASTCALL call_12_14_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_1_7
#define init_12_14_1_7
void FASTCALL call_12_14_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_1_8
#define init_12_14_1_8
void FASTCALL call_12_14_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_1_9
#define init_12_14_1_9
void FASTCALL call_12_14_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_1_10
#define init_12_14_1_10
void FASTCALL call_12_14_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_1_11
#define init_12_14_1_11
void FASTCALL call_12_14_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_1_12
#define init_12_14_1_12
void FASTCALL call_12_14_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_1_13
#define init_12_14_1_13
void FASTCALL call_12_14_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_1_14
#define init_12_14_1_14
void FASTCALL call_12_14_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_1_15
#define init_12_14_1_15
void FASTCALL call_12_14_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_2_0
#define init_12_14_2_0
void FASTCALL call_12_14_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_2_1
#define init_12_14_2_1
void FASTCALL call_12_14_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_2_2
#define init_12_14_2_2
void FASTCALL call_12_14_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_2_3
#define init_12_14_2_3
void FASTCALL call_12_14_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_2_4
#define init_12_14_2_4
void FASTCALL call_12_14_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_2_5
#define init_12_14_2_5
void FASTCALL call_12_14_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_2_6
#define init_12_14_2_6
void FASTCALL call_12_14_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_2_7
#define init_12_14_2_7
void FASTCALL call_12_14_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_2_8
#define init_12_14_2_8
void FASTCALL call_12_14_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_2_9
#define init_12_14_2_9
void FASTCALL call_12_14_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_2_10
#define init_12_14_2_10
void FASTCALL call_12_14_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_2_11
#define init_12_14_2_11
void FASTCALL call_12_14_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_2_12
#define init_12_14_2_12
void FASTCALL call_12_14_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_2_13
#define init_12_14_2_13
void FASTCALL call_12_14_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_2_14
#define init_12_14_2_14
void FASTCALL call_12_14_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_2_15
#define init_12_14_2_15
void FASTCALL call_12_14_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_3_0
#define init_12_14_3_0
void FASTCALL call_12_14_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_3_1
#define init_12_14_3_1
void FASTCALL call_12_14_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_3_2
#define init_12_14_3_2
void FASTCALL call_12_14_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_3_3
#define init_12_14_3_3
void FASTCALL call_12_14_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_3_4
#define init_12_14_3_4
void FASTCALL call_12_14_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_3_5
#define init_12_14_3_5
void FASTCALL call_12_14_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_3_6
#define init_12_14_3_6
void FASTCALL call_12_14_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_3_7
#define init_12_14_3_7
void FASTCALL call_12_14_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_3_8
#define init_12_14_3_8
void FASTCALL call_12_14_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_3_9
#define init_12_14_3_9
void FASTCALL call_12_14_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_3_10
#define init_12_14_3_10
void FASTCALL call_12_14_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_3_11
#define init_12_14_3_11
void FASTCALL call_12_14_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_3_12
#define init_12_14_3_12
void FASTCALL call_12_14_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_3_13
#define init_12_14_3_13
void FASTCALL call_12_14_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_3_14
#define init_12_14_3_14
void FASTCALL call_12_14_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_3_15
#define init_12_14_3_15
void FASTCALL call_12_14_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_4_0
#define init_12_14_4_0
void FASTCALL call_12_14_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_4_1
#define init_12_14_4_1
void FASTCALL call_12_14_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_4_2
#define init_12_14_4_2
void FASTCALL call_12_14_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_4_3
#define init_12_14_4_3
void FASTCALL call_12_14_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_4_4
#define init_12_14_4_4
void FASTCALL call_12_14_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_4_5
#define init_12_14_4_5
void FASTCALL call_12_14_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_4_6
#define init_12_14_4_6
void FASTCALL call_12_14_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_4_7
#define init_12_14_4_7
void FASTCALL call_12_14_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_4_8
#define init_12_14_4_8
void FASTCALL call_12_14_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_4_9
#define init_12_14_4_9
void FASTCALL call_12_14_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_4_10
#define init_12_14_4_10
void FASTCALL call_12_14_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_4_11
#define init_12_14_4_11
void FASTCALL call_12_14_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_4_12
#define init_12_14_4_12
void FASTCALL call_12_14_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_4_13
#define init_12_14_4_13
void FASTCALL call_12_14_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_4_14
#define init_12_14_4_14
void FASTCALL call_12_14_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_4_15
#define init_12_14_4_15
void FASTCALL call_12_14_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_5_0
#define init_12_14_5_0
void FASTCALL call_12_14_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_5_1
#define init_12_14_5_1
void FASTCALL call_12_14_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_5_2
#define init_12_14_5_2
void FASTCALL call_12_14_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_5_3
#define init_12_14_5_3
void FASTCALL call_12_14_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_5_4
#define init_12_14_5_4
void FASTCALL call_12_14_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_5_5
#define init_12_14_5_5
void FASTCALL call_12_14_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_5_6
#define init_12_14_5_6
void FASTCALL call_12_14_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_5_7
#define init_12_14_5_7
void FASTCALL call_12_14_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_5_8
#define init_12_14_5_8
void FASTCALL call_12_14_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_5_9
#define init_12_14_5_9
void FASTCALL call_12_14_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_5_10
#define init_12_14_5_10
void FASTCALL call_12_14_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_5_11
#define init_12_14_5_11
void FASTCALL call_12_14_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_5_12
#define init_12_14_5_12
void FASTCALL call_12_14_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_5_13
#define init_12_14_5_13
void FASTCALL call_12_14_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_5_14
#define init_12_14_5_14
void FASTCALL call_12_14_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_5_15
#define init_12_14_5_15
void FASTCALL call_12_14_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_6_0
#define init_12_14_6_0
void FASTCALL call_12_14_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_6_1
#define init_12_14_6_1
void FASTCALL call_12_14_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_6_2
#define init_12_14_6_2
void FASTCALL call_12_14_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_6_3
#define init_12_14_6_3
void FASTCALL call_12_14_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_6_4
#define init_12_14_6_4
void FASTCALL call_12_14_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_6_5
#define init_12_14_6_5
void FASTCALL call_12_14_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_6_6
#define init_12_14_6_6
void FASTCALL call_12_14_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_6_7
#define init_12_14_6_7
void FASTCALL call_12_14_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_6_8
#define init_12_14_6_8
void FASTCALL call_12_14_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_6_9
#define init_12_14_6_9
void FASTCALL call_12_14_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_6_10
#define init_12_14_6_10
void FASTCALL call_12_14_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_6_11
#define init_12_14_6_11
void FASTCALL call_12_14_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_6_12
#define init_12_14_6_12
void FASTCALL call_12_14_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_6_13
#define init_12_14_6_13
void FASTCALL call_12_14_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_6_14
#define init_12_14_6_14
void FASTCALL call_12_14_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_6_15
#define init_12_14_6_15
void FASTCALL call_12_14_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_7_0
#define init_12_14_7_0
void FASTCALL call_12_14_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_7_1
#define init_12_14_7_1
void FASTCALL call_12_14_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_7_2
#define init_12_14_7_2
void FASTCALL call_12_14_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_7_3
#define init_12_14_7_3
void FASTCALL call_12_14_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_7_4
#define init_12_14_7_4
void FASTCALL call_12_14_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_7_5
#define init_12_14_7_5
void FASTCALL call_12_14_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_7_6
#define init_12_14_7_6
void FASTCALL call_12_14_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_7_7
#define init_12_14_7_7
void FASTCALL call_12_14_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_7_8
#define init_12_14_7_8
void FASTCALL call_12_14_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_7_9
#define init_12_14_7_9
void FASTCALL call_12_14_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_7_10
#define init_12_14_7_10
void FASTCALL call_12_14_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_7_11
#define init_12_14_7_11
void FASTCALL call_12_14_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_7_12
#define init_12_14_7_12
void FASTCALL call_12_14_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_7_13
#define init_12_14_7_13
void FASTCALL call_12_14_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_7_14
#define init_12_14_7_14
void FASTCALL call_12_14_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_7_15
#define init_12_14_7_15
void FASTCALL call_12_14_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_8_0
#define init_12_14_8_0
void FASTCALL call_12_14_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_8_1
#define init_12_14_8_1
void FASTCALL call_12_14_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_8_2
#define init_12_14_8_2
void FASTCALL call_12_14_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_8_3
#define init_12_14_8_3
void FASTCALL call_12_14_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_8_4
#define init_12_14_8_4
void FASTCALL call_12_14_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_8_5
#define init_12_14_8_5
void FASTCALL call_12_14_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_8_6
#define init_12_14_8_6
void FASTCALL call_12_14_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_8_7
#define init_12_14_8_7
void FASTCALL call_12_14_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_8_8
#define init_12_14_8_8
void FASTCALL call_12_14_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_8_9
#define init_12_14_8_9
void FASTCALL call_12_14_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_8_10
#define init_12_14_8_10
void FASTCALL call_12_14_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_8_11
#define init_12_14_8_11
void FASTCALL call_12_14_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_8_12
#define init_12_14_8_12
void FASTCALL call_12_14_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_8_13
#define init_12_14_8_13
void FASTCALL call_12_14_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_8_14
#define init_12_14_8_14
void FASTCALL call_12_14_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_8_15
#define init_12_14_8_15
void FASTCALL call_12_14_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_9_0
#define init_12_14_9_0
void FASTCALL call_12_14_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_9_1
#define init_12_14_9_1
void FASTCALL call_12_14_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_9_2
#define init_12_14_9_2
void FASTCALL call_12_14_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_9_3
#define init_12_14_9_3
void FASTCALL call_12_14_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_9_4
#define init_12_14_9_4
void FASTCALL call_12_14_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_9_5
#define init_12_14_9_5
void FASTCALL call_12_14_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_9_6
#define init_12_14_9_6
void FASTCALL call_12_14_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_9_7
#define init_12_14_9_7
void FASTCALL call_12_14_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_9_8
#define init_12_14_9_8
void FASTCALL call_12_14_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_9_9
#define init_12_14_9_9
void FASTCALL call_12_14_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_9_10
#define init_12_14_9_10
void FASTCALL call_12_14_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_9_11
#define init_12_14_9_11
void FASTCALL call_12_14_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_9_12
#define init_12_14_9_12
void FASTCALL call_12_14_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_9_13
#define init_12_14_9_13
void FASTCALL call_12_14_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_9_14
#define init_12_14_9_14
void FASTCALL call_12_14_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_9_15
#define init_12_14_9_15
void FASTCALL call_12_14_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_10_0
#define init_12_14_10_0
void FASTCALL call_12_14_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_10_1
#define init_12_14_10_1
void FASTCALL call_12_14_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_10_2
#define init_12_14_10_2
void FASTCALL call_12_14_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_10_3
#define init_12_14_10_3
void FASTCALL call_12_14_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_10_4
#define init_12_14_10_4
void FASTCALL call_12_14_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_10_5
#define init_12_14_10_5
void FASTCALL call_12_14_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_10_6
#define init_12_14_10_6
void FASTCALL call_12_14_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_10_7
#define init_12_14_10_7
void FASTCALL call_12_14_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_10_8
#define init_12_14_10_8
void FASTCALL call_12_14_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_10_9
#define init_12_14_10_9
void FASTCALL call_12_14_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_10_10
#define init_12_14_10_10
void FASTCALL call_12_14_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_10_11
#define init_12_14_10_11
void FASTCALL call_12_14_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_10_12
#define init_12_14_10_12
void FASTCALL call_12_14_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_10_13
#define init_12_14_10_13
void FASTCALL call_12_14_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_10_14
#define init_12_14_10_14
void FASTCALL call_12_14_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_10_15
#define init_12_14_10_15
void FASTCALL call_12_14_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_11_0
#define init_12_14_11_0
void FASTCALL call_12_14_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_11_1
#define init_12_14_11_1
void FASTCALL call_12_14_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_11_2
#define init_12_14_11_2
void FASTCALL call_12_14_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_11_3
#define init_12_14_11_3
void FASTCALL call_12_14_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_11_4
#define init_12_14_11_4
void FASTCALL call_12_14_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_11_5
#define init_12_14_11_5
void FASTCALL call_12_14_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_11_6
#define init_12_14_11_6
void FASTCALL call_12_14_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_11_7
#define init_12_14_11_7
void FASTCALL call_12_14_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_11_8
#define init_12_14_11_8
void FASTCALL call_12_14_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_11_9
#define init_12_14_11_9
void FASTCALL call_12_14_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_11_10
#define init_12_14_11_10
void FASTCALL call_12_14_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_11_11
#define init_12_14_11_11
void FASTCALL call_12_14_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_11_12
#define init_12_14_11_12
void FASTCALL call_12_14_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_11_13
#define init_12_14_11_13
void FASTCALL call_12_14_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_11_14
#define init_12_14_11_14
void FASTCALL call_12_14_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_11_15
#define init_12_14_11_15
void FASTCALL call_12_14_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_12_0
#define init_12_14_12_0
void FASTCALL call_12_14_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_12_1
#define init_12_14_12_1
void FASTCALL call_12_14_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_12_2
#define init_12_14_12_2
void FASTCALL call_12_14_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_12_3
#define init_12_14_12_3
void FASTCALL call_12_14_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_12_4
#define init_12_14_12_4
void FASTCALL call_12_14_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_12_5
#define init_12_14_12_5
void FASTCALL call_12_14_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_12_6
#define init_12_14_12_6
void FASTCALL call_12_14_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_12_7
#define init_12_14_12_7
void FASTCALL call_12_14_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_12_8
#define init_12_14_12_8
void FASTCALL call_12_14_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_12_9
#define init_12_14_12_9
void FASTCALL call_12_14_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_12_10
#define init_12_14_12_10
void FASTCALL call_12_14_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_12_11
#define init_12_14_12_11
void FASTCALL call_12_14_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_12_12
#define init_12_14_12_12
void FASTCALL call_12_14_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_12_13
#define init_12_14_12_13
void FASTCALL call_12_14_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_12_14
#define init_12_14_12_14
void FASTCALL call_12_14_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_12_15
#define init_12_14_12_15
void FASTCALL call_12_14_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_13_0
#define init_12_14_13_0
void FASTCALL call_12_14_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_13_1
#define init_12_14_13_1
void FASTCALL call_12_14_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_13_2
#define init_12_14_13_2
void FASTCALL call_12_14_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_13_3
#define init_12_14_13_3
void FASTCALL call_12_14_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_13_4
#define init_12_14_13_4
void FASTCALL call_12_14_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_13_5
#define init_12_14_13_5
void FASTCALL call_12_14_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_13_6
#define init_12_14_13_6
void FASTCALL call_12_14_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_13_7
#define init_12_14_13_7
void FASTCALL call_12_14_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_13_8
#define init_12_14_13_8
void FASTCALL call_12_14_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_13_9
#define init_12_14_13_9
void FASTCALL call_12_14_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_13_10
#define init_12_14_13_10
void FASTCALL call_12_14_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_13_11
#define init_12_14_13_11
void FASTCALL call_12_14_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_13_12
#define init_12_14_13_12
void FASTCALL call_12_14_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_13_13
#define init_12_14_13_13
void FASTCALL call_12_14_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_13_14
#define init_12_14_13_14
void FASTCALL call_12_14_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_13_15
#define init_12_14_13_15
void FASTCALL call_12_14_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_14_0
#define init_12_14_14_0
void FASTCALL call_12_14_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_14_1
#define init_12_14_14_1
void FASTCALL call_12_14_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_14_2
#define init_12_14_14_2
void FASTCALL call_12_14_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_14_3
#define init_12_14_14_3
void FASTCALL call_12_14_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_14_4
#define init_12_14_14_4
void FASTCALL call_12_14_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_14_5
#define init_12_14_14_5
void FASTCALL call_12_14_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_14_6
#define init_12_14_14_6
void FASTCALL call_12_14_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_14_7
#define init_12_14_14_7
void FASTCALL call_12_14_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_14_8
#define init_12_14_14_8
void FASTCALL call_12_14_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_14_9
#define init_12_14_14_9
void FASTCALL call_12_14_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_14_10
#define init_12_14_14_10
void FASTCALL call_12_14_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_14_11
#define init_12_14_14_11
void FASTCALL call_12_14_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_14_12
#define init_12_14_14_12
void FASTCALL call_12_14_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_14_13
#define init_12_14_14_13
void FASTCALL call_12_14_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_14_14
#define init_12_14_14_14
void FASTCALL call_12_14_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_14_15
#define init_12_14_14_15
void FASTCALL call_12_14_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_15_0
#define init_12_14_15_0
void FASTCALL call_12_14_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_15_1
#define init_12_14_15_1
void FASTCALL call_12_14_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_15_2
#define init_12_14_15_2
void FASTCALL call_12_14_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_15_3
#define init_12_14_15_3
void FASTCALL call_12_14_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_15_4
#define init_12_14_15_4
void FASTCALL call_12_14_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_15_5
#define init_12_14_15_5
void FASTCALL call_12_14_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_15_6
#define init_12_14_15_6
void FASTCALL call_12_14_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_15_7
#define init_12_14_15_7
void FASTCALL call_12_14_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_15_8
#define init_12_14_15_8
void FASTCALL call_12_14_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_15_9
#define init_12_14_15_9
void FASTCALL call_12_14_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_15_10
#define init_12_14_15_10
void FASTCALL call_12_14_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_15_11
#define init_12_14_15_11
void FASTCALL call_12_14_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_15_12
#define init_12_14_15_12
void FASTCALL call_12_14_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_15_13
#define init_12_14_15_13
void FASTCALL call_12_14_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_15_14
#define init_12_14_15_14
void FASTCALL call_12_14_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_14_15_15
#define init_12_14_15_15
void FASTCALL call_12_14_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_0_0
#define init_12_15_0_0
void FASTCALL call_12_15_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_0_1
#define init_12_15_0_1
void FASTCALL call_12_15_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_0_2
#define init_12_15_0_2
void FASTCALL call_12_15_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_0_3
#define init_12_15_0_3
void FASTCALL call_12_15_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_0_4
#define init_12_15_0_4
void FASTCALL call_12_15_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_0_5
#define init_12_15_0_5
void FASTCALL call_12_15_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_0_6
#define init_12_15_0_6
void FASTCALL call_12_15_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_0_7
#define init_12_15_0_7
void FASTCALL call_12_15_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_0_8
#define init_12_15_0_8
void FASTCALL call_12_15_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_0_9
#define init_12_15_0_9
void FASTCALL call_12_15_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_0_10
#define init_12_15_0_10
void FASTCALL call_12_15_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_0_11
#define init_12_15_0_11
void FASTCALL call_12_15_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_0_12
#define init_12_15_0_12
void FASTCALL call_12_15_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_0_13
#define init_12_15_0_13
void FASTCALL call_12_15_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_0_14
#define init_12_15_0_14
void FASTCALL call_12_15_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_0_15
#define init_12_15_0_15
void FASTCALL call_12_15_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_1_0
#define init_12_15_1_0
void FASTCALL call_12_15_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_1_1
#define init_12_15_1_1
void FASTCALL call_12_15_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_1_2
#define init_12_15_1_2
void FASTCALL call_12_15_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_1_3
#define init_12_15_1_3
void FASTCALL call_12_15_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_1_4
#define init_12_15_1_4
void FASTCALL call_12_15_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_1_5
#define init_12_15_1_5
void FASTCALL call_12_15_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_1_6
#define init_12_15_1_6
void FASTCALL call_12_15_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_1_7
#define init_12_15_1_7
void FASTCALL call_12_15_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_1_8
#define init_12_15_1_8
void FASTCALL call_12_15_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_1_9
#define init_12_15_1_9
void FASTCALL call_12_15_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_1_10
#define init_12_15_1_10
void FASTCALL call_12_15_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_1_11
#define init_12_15_1_11
void FASTCALL call_12_15_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_1_12
#define init_12_15_1_12
void FASTCALL call_12_15_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_1_13
#define init_12_15_1_13
void FASTCALL call_12_15_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_1_14
#define init_12_15_1_14
void FASTCALL call_12_15_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_1_15
#define init_12_15_1_15
void FASTCALL call_12_15_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_2_0
#define init_12_15_2_0
void FASTCALL call_12_15_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_2_1
#define init_12_15_2_1
void FASTCALL call_12_15_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_2_2
#define init_12_15_2_2
void FASTCALL call_12_15_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_2_3
#define init_12_15_2_3
void FASTCALL call_12_15_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_2_4
#define init_12_15_2_4
void FASTCALL call_12_15_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_2_5
#define init_12_15_2_5
void FASTCALL call_12_15_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_2_6
#define init_12_15_2_6
void FASTCALL call_12_15_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_2_7
#define init_12_15_2_7
void FASTCALL call_12_15_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_2_8
#define init_12_15_2_8
void FASTCALL call_12_15_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_2_9
#define init_12_15_2_9
void FASTCALL call_12_15_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_2_10
#define init_12_15_2_10
void FASTCALL call_12_15_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_2_11
#define init_12_15_2_11
void FASTCALL call_12_15_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_2_12
#define init_12_15_2_12
void FASTCALL call_12_15_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_2_13
#define init_12_15_2_13
void FASTCALL call_12_15_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_2_14
#define init_12_15_2_14
void FASTCALL call_12_15_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_2_15
#define init_12_15_2_15
void FASTCALL call_12_15_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_3_0
#define init_12_15_3_0
void FASTCALL call_12_15_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_3_1
#define init_12_15_3_1
void FASTCALL call_12_15_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_3_2
#define init_12_15_3_2
void FASTCALL call_12_15_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_3_3
#define init_12_15_3_3
void FASTCALL call_12_15_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_3_4
#define init_12_15_3_4
void FASTCALL call_12_15_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_3_5
#define init_12_15_3_5
void FASTCALL call_12_15_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_3_6
#define init_12_15_3_6
void FASTCALL call_12_15_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_3_7
#define init_12_15_3_7
void FASTCALL call_12_15_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_3_8
#define init_12_15_3_8
void FASTCALL call_12_15_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_3_9
#define init_12_15_3_9
void FASTCALL call_12_15_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_3_10
#define init_12_15_3_10
void FASTCALL call_12_15_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_3_11
#define init_12_15_3_11
void FASTCALL call_12_15_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_3_12
#define init_12_15_3_12
void FASTCALL call_12_15_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_3_13
#define init_12_15_3_13
void FASTCALL call_12_15_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_3_14
#define init_12_15_3_14
void FASTCALL call_12_15_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_3_15
#define init_12_15_3_15
void FASTCALL call_12_15_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_4_0
#define init_12_15_4_0
void FASTCALL call_12_15_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_4_1
#define init_12_15_4_1
void FASTCALL call_12_15_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_4_2
#define init_12_15_4_2
void FASTCALL call_12_15_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_4_3
#define init_12_15_4_3
void FASTCALL call_12_15_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_4_4
#define init_12_15_4_4
void FASTCALL call_12_15_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_4_5
#define init_12_15_4_5
void FASTCALL call_12_15_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_4_6
#define init_12_15_4_6
void FASTCALL call_12_15_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_4_7
#define init_12_15_4_7
void FASTCALL call_12_15_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_4_8
#define init_12_15_4_8
void FASTCALL call_12_15_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_4_9
#define init_12_15_4_9
void FASTCALL call_12_15_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_4_10
#define init_12_15_4_10
void FASTCALL call_12_15_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_4_11
#define init_12_15_4_11
void FASTCALL call_12_15_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_4_12
#define init_12_15_4_12
void FASTCALL call_12_15_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_4_13
#define init_12_15_4_13
void FASTCALL call_12_15_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_4_14
#define init_12_15_4_14
void FASTCALL call_12_15_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_4_15
#define init_12_15_4_15
void FASTCALL call_12_15_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_5_0
#define init_12_15_5_0
void FASTCALL call_12_15_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_5_1
#define init_12_15_5_1
void FASTCALL call_12_15_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_5_2
#define init_12_15_5_2
void FASTCALL call_12_15_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_5_3
#define init_12_15_5_3
void FASTCALL call_12_15_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_5_4
#define init_12_15_5_4
void FASTCALL call_12_15_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_5_5
#define init_12_15_5_5
void FASTCALL call_12_15_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_5_6
#define init_12_15_5_6
void FASTCALL call_12_15_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_5_7
#define init_12_15_5_7
void FASTCALL call_12_15_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_5_8
#define init_12_15_5_8
void FASTCALL call_12_15_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_5_9
#define init_12_15_5_9
void FASTCALL call_12_15_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_5_10
#define init_12_15_5_10
void FASTCALL call_12_15_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_5_11
#define init_12_15_5_11
void FASTCALL call_12_15_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_5_12
#define init_12_15_5_12
void FASTCALL call_12_15_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_5_13
#define init_12_15_5_13
void FASTCALL call_12_15_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_5_14
#define init_12_15_5_14
void FASTCALL call_12_15_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_5_15
#define init_12_15_5_15
void FASTCALL call_12_15_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_6_0
#define init_12_15_6_0
void FASTCALL call_12_15_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_6_1
#define init_12_15_6_1
void FASTCALL call_12_15_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_6_2
#define init_12_15_6_2
void FASTCALL call_12_15_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_6_3
#define init_12_15_6_3
void FASTCALL call_12_15_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_6_4
#define init_12_15_6_4
void FASTCALL call_12_15_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_6_5
#define init_12_15_6_5
void FASTCALL call_12_15_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_6_6
#define init_12_15_6_6
void FASTCALL call_12_15_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_6_7
#define init_12_15_6_7
void FASTCALL call_12_15_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_6_8
#define init_12_15_6_8
void FASTCALL call_12_15_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_6_9
#define init_12_15_6_9
void FASTCALL call_12_15_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_6_10
#define init_12_15_6_10
void FASTCALL call_12_15_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_6_11
#define init_12_15_6_11
void FASTCALL call_12_15_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_6_12
#define init_12_15_6_12
void FASTCALL call_12_15_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_6_13
#define init_12_15_6_13
void FASTCALL call_12_15_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_6_14
#define init_12_15_6_14
void FASTCALL call_12_15_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_6_15
#define init_12_15_6_15
void FASTCALL call_12_15_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_7_0
#define init_12_15_7_0
void FASTCALL call_12_15_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_7_1
#define init_12_15_7_1
void FASTCALL call_12_15_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_7_2
#define init_12_15_7_2
void FASTCALL call_12_15_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_7_3
#define init_12_15_7_3
void FASTCALL call_12_15_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_7_4
#define init_12_15_7_4
void FASTCALL call_12_15_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_7_5
#define init_12_15_7_5
void FASTCALL call_12_15_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_7_6
#define init_12_15_7_6
void FASTCALL call_12_15_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_7_7
#define init_12_15_7_7
void FASTCALL call_12_15_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_7_8
#define init_12_15_7_8
void FASTCALL call_12_15_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_7_9
#define init_12_15_7_9
void FASTCALL call_12_15_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_7_10
#define init_12_15_7_10
void FASTCALL call_12_15_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_7_11
#define init_12_15_7_11
void FASTCALL call_12_15_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_7_12
#define init_12_15_7_12
void FASTCALL call_12_15_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_7_13
#define init_12_15_7_13
void FASTCALL call_12_15_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_7_14
#define init_12_15_7_14
void FASTCALL call_12_15_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_7_15
#define init_12_15_7_15
void FASTCALL call_12_15_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_8_0
#define init_12_15_8_0
void FASTCALL call_12_15_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_8_1
#define init_12_15_8_1
void FASTCALL call_12_15_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_8_2
#define init_12_15_8_2
void FASTCALL call_12_15_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_8_3
#define init_12_15_8_3
void FASTCALL call_12_15_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_8_4
#define init_12_15_8_4
void FASTCALL call_12_15_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_8_5
#define init_12_15_8_5
void FASTCALL call_12_15_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_8_6
#define init_12_15_8_6
void FASTCALL call_12_15_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_8_7
#define init_12_15_8_7
void FASTCALL call_12_15_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_8_8
#define init_12_15_8_8
void FASTCALL call_12_15_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_8_9
#define init_12_15_8_9
void FASTCALL call_12_15_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_8_10
#define init_12_15_8_10
void FASTCALL call_12_15_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_8_11
#define init_12_15_8_11
void FASTCALL call_12_15_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_8_12
#define init_12_15_8_12
void FASTCALL call_12_15_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_8_13
#define init_12_15_8_13
void FASTCALL call_12_15_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_8_14
#define init_12_15_8_14
void FASTCALL call_12_15_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_8_15
#define init_12_15_8_15
void FASTCALL call_12_15_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_9_0
#define init_12_15_9_0
void FASTCALL call_12_15_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_9_1
#define init_12_15_9_1
void FASTCALL call_12_15_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_9_2
#define init_12_15_9_2
void FASTCALL call_12_15_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_9_3
#define init_12_15_9_3
void FASTCALL call_12_15_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_9_4
#define init_12_15_9_4
void FASTCALL call_12_15_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_9_5
#define init_12_15_9_5
void FASTCALL call_12_15_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_9_6
#define init_12_15_9_6
void FASTCALL call_12_15_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_9_7
#define init_12_15_9_7
void FASTCALL call_12_15_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_9_8
#define init_12_15_9_8
void FASTCALL call_12_15_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_9_9
#define init_12_15_9_9
void FASTCALL call_12_15_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_9_10
#define init_12_15_9_10
void FASTCALL call_12_15_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_9_11
#define init_12_15_9_11
void FASTCALL call_12_15_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_9_12
#define init_12_15_9_12
void FASTCALL call_12_15_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_9_13
#define init_12_15_9_13
void FASTCALL call_12_15_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_9_14
#define init_12_15_9_14
void FASTCALL call_12_15_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_9_15
#define init_12_15_9_15
void FASTCALL call_12_15_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_10_0
#define init_12_15_10_0
void FASTCALL call_12_15_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_10_1
#define init_12_15_10_1
void FASTCALL call_12_15_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_10_2
#define init_12_15_10_2
void FASTCALL call_12_15_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_10_3
#define init_12_15_10_3
void FASTCALL call_12_15_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_10_4
#define init_12_15_10_4
void FASTCALL call_12_15_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_10_5
#define init_12_15_10_5
void FASTCALL call_12_15_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_10_6
#define init_12_15_10_6
void FASTCALL call_12_15_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_10_7
#define init_12_15_10_7
void FASTCALL call_12_15_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_10_8
#define init_12_15_10_8
void FASTCALL call_12_15_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_10_9
#define init_12_15_10_9
void FASTCALL call_12_15_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_10_10
#define init_12_15_10_10
void FASTCALL call_12_15_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_10_11
#define init_12_15_10_11
void FASTCALL call_12_15_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_10_12
#define init_12_15_10_12
void FASTCALL call_12_15_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_10_13
#define init_12_15_10_13
void FASTCALL call_12_15_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_10_14
#define init_12_15_10_14
void FASTCALL call_12_15_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_10_15
#define init_12_15_10_15
void FASTCALL call_12_15_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_11_0
#define init_12_15_11_0
void FASTCALL call_12_15_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_11_1
#define init_12_15_11_1
void FASTCALL call_12_15_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_11_2
#define init_12_15_11_2
void FASTCALL call_12_15_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_11_3
#define init_12_15_11_3
void FASTCALL call_12_15_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_11_4
#define init_12_15_11_4
void FASTCALL call_12_15_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_11_5
#define init_12_15_11_5
void FASTCALL call_12_15_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_11_6
#define init_12_15_11_6
void FASTCALL call_12_15_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_11_7
#define init_12_15_11_7
void FASTCALL call_12_15_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_11_8
#define init_12_15_11_8
void FASTCALL call_12_15_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_11_9
#define init_12_15_11_9
void FASTCALL call_12_15_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_11_10
#define init_12_15_11_10
void FASTCALL call_12_15_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_11_11
#define init_12_15_11_11
void FASTCALL call_12_15_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_11_12
#define init_12_15_11_12
void FASTCALL call_12_15_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_11_13
#define init_12_15_11_13
void FASTCALL call_12_15_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_11_14
#define init_12_15_11_14
void FASTCALL call_12_15_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_11_15
#define init_12_15_11_15
void FASTCALL call_12_15_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_12_0
#define init_12_15_12_0
void FASTCALL call_12_15_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_12_1
#define init_12_15_12_1
void FASTCALL call_12_15_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_12_2
#define init_12_15_12_2
void FASTCALL call_12_15_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_12_3
#define init_12_15_12_3
void FASTCALL call_12_15_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_12_4
#define init_12_15_12_4
void FASTCALL call_12_15_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_12_5
#define init_12_15_12_5
void FASTCALL call_12_15_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_12_6
#define init_12_15_12_6
void FASTCALL call_12_15_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_12_7
#define init_12_15_12_7
void FASTCALL call_12_15_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_12_8
#define init_12_15_12_8
void FASTCALL call_12_15_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_12_9
#define init_12_15_12_9
void FASTCALL call_12_15_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_12_10
#define init_12_15_12_10
void FASTCALL call_12_15_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_12_11
#define init_12_15_12_11
void FASTCALL call_12_15_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_12_12
#define init_12_15_12_12
void FASTCALL call_12_15_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_12_13
#define init_12_15_12_13
void FASTCALL call_12_15_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_12_14
#define init_12_15_12_14
void FASTCALL call_12_15_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_12_15
#define init_12_15_12_15
void FASTCALL call_12_15_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_13_0
#define init_12_15_13_0
void FASTCALL call_12_15_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_13_1
#define init_12_15_13_1
void FASTCALL call_12_15_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_13_2
#define init_12_15_13_2
void FASTCALL call_12_15_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_13_3
#define init_12_15_13_3
void FASTCALL call_12_15_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_13_4
#define init_12_15_13_4
void FASTCALL call_12_15_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_13_5
#define init_12_15_13_5
void FASTCALL call_12_15_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_13_6
#define init_12_15_13_6
void FASTCALL call_12_15_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_13_7
#define init_12_15_13_7
void FASTCALL call_12_15_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_13_8
#define init_12_15_13_8
void FASTCALL call_12_15_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_13_9
#define init_12_15_13_9
void FASTCALL call_12_15_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_13_10
#define init_12_15_13_10
void FASTCALL call_12_15_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_13_11
#define init_12_15_13_11
void FASTCALL call_12_15_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_13_12
#define init_12_15_13_12
void FASTCALL call_12_15_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_13_13
#define init_12_15_13_13
void FASTCALL call_12_15_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_13_14
#define init_12_15_13_14
void FASTCALL call_12_15_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_13_15
#define init_12_15_13_15
void FASTCALL call_12_15_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_14_0
#define init_12_15_14_0
void FASTCALL call_12_15_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_14_1
#define init_12_15_14_1
void FASTCALL call_12_15_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_14_2
#define init_12_15_14_2
void FASTCALL call_12_15_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_14_3
#define init_12_15_14_3
void FASTCALL call_12_15_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_14_4
#define init_12_15_14_4
void FASTCALL call_12_15_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_14_5
#define init_12_15_14_5
void FASTCALL call_12_15_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_14_6
#define init_12_15_14_6
void FASTCALL call_12_15_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_14_7
#define init_12_15_14_7
void FASTCALL call_12_15_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_14_8
#define init_12_15_14_8
void FASTCALL call_12_15_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_14_9
#define init_12_15_14_9
void FASTCALL call_12_15_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_14_10
#define init_12_15_14_10
void FASTCALL call_12_15_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_14_11
#define init_12_15_14_11
void FASTCALL call_12_15_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_14_12
#define init_12_15_14_12
void FASTCALL call_12_15_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_14_13
#define init_12_15_14_13
void FASTCALL call_12_15_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_14_14
#define init_12_15_14_14
void FASTCALL call_12_15_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_14_15
#define init_12_15_14_15
void FASTCALL call_12_15_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_15_0
#define init_12_15_15_0
void FASTCALL call_12_15_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_15_1
#define init_12_15_15_1
void FASTCALL call_12_15_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_15_2
#define init_12_15_15_2
void FASTCALL call_12_15_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_15_3
#define init_12_15_15_3
void FASTCALL call_12_15_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_15_4
#define init_12_15_15_4
void FASTCALL call_12_15_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_15_5
#define init_12_15_15_5
void FASTCALL call_12_15_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_15_6
#define init_12_15_15_6
void FASTCALL call_12_15_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_15_7
#define init_12_15_15_7
void FASTCALL call_12_15_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_15_8
#define init_12_15_15_8
void FASTCALL call_12_15_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_15_9
#define init_12_15_15_9
void FASTCALL call_12_15_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_15_10
#define init_12_15_15_10
void FASTCALL call_12_15_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_15_11
#define init_12_15_15_11
void FASTCALL call_12_15_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_15_12
#define init_12_15_15_12
void FASTCALL call_12_15_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_15_13
#define init_12_15_15_13
void FASTCALL call_12_15_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_15_14
#define init_12_15_15_14
void FASTCALL call_12_15_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_12_15_15_15
#define init_12_15_15_15
void FASTCALL call_12_15_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_0_0
#define init_13_0_0_0
void FASTCALL call_13_0_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_0_1
#define init_13_0_0_1
void FASTCALL call_13_0_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_0_2
#define init_13_0_0_2
void FASTCALL call_13_0_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_0_3
#define init_13_0_0_3
void FASTCALL call_13_0_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_0_4
#define init_13_0_0_4
void FASTCALL call_13_0_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_0_5
#define init_13_0_0_5
void FASTCALL call_13_0_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_0_6
#define init_13_0_0_6
void FASTCALL call_13_0_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_0_7
#define init_13_0_0_7
void FASTCALL call_13_0_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_0_8
#define init_13_0_0_8
void FASTCALL call_13_0_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_0_9
#define init_13_0_0_9
void FASTCALL call_13_0_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_0_10
#define init_13_0_0_10
void FASTCALL call_13_0_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_0_11
#define init_13_0_0_11
void FASTCALL call_13_0_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_0_12
#define init_13_0_0_12
void FASTCALL call_13_0_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_0_13
#define init_13_0_0_13
void FASTCALL call_13_0_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_0_14
#define init_13_0_0_14
void FASTCALL call_13_0_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_0_15
#define init_13_0_0_15
void FASTCALL call_13_0_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_1_0
#define init_13_0_1_0
void FASTCALL call_13_0_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_1_1
#define init_13_0_1_1
void FASTCALL call_13_0_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_1_2
#define init_13_0_1_2
void FASTCALL call_13_0_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_1_3
#define init_13_0_1_3
void FASTCALL call_13_0_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_1_4
#define init_13_0_1_4
void FASTCALL call_13_0_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_1_5
#define init_13_0_1_5
void FASTCALL call_13_0_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_1_6
#define init_13_0_1_6
void FASTCALL call_13_0_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_1_7
#define init_13_0_1_7
void FASTCALL call_13_0_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_1_8
#define init_13_0_1_8
void FASTCALL call_13_0_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_1_9
#define init_13_0_1_9
void FASTCALL call_13_0_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_1_10
#define init_13_0_1_10
void FASTCALL call_13_0_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_1_11
#define init_13_0_1_11
void FASTCALL call_13_0_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_1_12
#define init_13_0_1_12
void FASTCALL call_13_0_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_1_13
#define init_13_0_1_13
void FASTCALL call_13_0_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_1_14
#define init_13_0_1_14
void FASTCALL call_13_0_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_1_15
#define init_13_0_1_15
void FASTCALL call_13_0_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_2_0
#define init_13_0_2_0
void FASTCALL call_13_0_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_2_1
#define init_13_0_2_1
void FASTCALL call_13_0_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_2_2
#define init_13_0_2_2
void FASTCALL call_13_0_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_2_3
#define init_13_0_2_3
void FASTCALL call_13_0_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_2_4
#define init_13_0_2_4
void FASTCALL call_13_0_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_2_5
#define init_13_0_2_5
void FASTCALL call_13_0_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_2_6
#define init_13_0_2_6
void FASTCALL call_13_0_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_2_7
#define init_13_0_2_7
void FASTCALL call_13_0_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_2_8
#define init_13_0_2_8
void FASTCALL call_13_0_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_2_9
#define init_13_0_2_9
void FASTCALL call_13_0_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_2_10
#define init_13_0_2_10
void FASTCALL call_13_0_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_2_11
#define init_13_0_2_11
void FASTCALL call_13_0_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_2_12
#define init_13_0_2_12
void FASTCALL call_13_0_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_2_13
#define init_13_0_2_13
void FASTCALL call_13_0_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_2_14
#define init_13_0_2_14
void FASTCALL call_13_0_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_2_15
#define init_13_0_2_15
void FASTCALL call_13_0_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_3_0
#define init_13_0_3_0
void FASTCALL call_13_0_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_3_1
#define init_13_0_3_1
void FASTCALL call_13_0_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_3_2
#define init_13_0_3_2
void FASTCALL call_13_0_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_3_3
#define init_13_0_3_3
void FASTCALL call_13_0_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_3_4
#define init_13_0_3_4
void FASTCALL call_13_0_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_3_5
#define init_13_0_3_5
void FASTCALL call_13_0_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_3_6
#define init_13_0_3_6
void FASTCALL call_13_0_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_3_7
#define init_13_0_3_7
void FASTCALL call_13_0_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_3_8
#define init_13_0_3_8
void FASTCALL call_13_0_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_3_9
#define init_13_0_3_9
void FASTCALL call_13_0_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_3_10
#define init_13_0_3_10
void FASTCALL call_13_0_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_3_11
#define init_13_0_3_11
void FASTCALL call_13_0_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_3_12
#define init_13_0_3_12
void FASTCALL call_13_0_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_3_13
#define init_13_0_3_13
void FASTCALL call_13_0_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_3_14
#define init_13_0_3_14
void FASTCALL call_13_0_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_3_15
#define init_13_0_3_15
void FASTCALL call_13_0_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_4_0
#define init_13_0_4_0
void FASTCALL call_13_0_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_4_1
#define init_13_0_4_1
void FASTCALL call_13_0_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_4_2
#define init_13_0_4_2
void FASTCALL call_13_0_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_4_3
#define init_13_0_4_3
void FASTCALL call_13_0_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_4_4
#define init_13_0_4_4
void FASTCALL call_13_0_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_4_5
#define init_13_0_4_5
void FASTCALL call_13_0_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_4_6
#define init_13_0_4_6
void FASTCALL call_13_0_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_4_7
#define init_13_0_4_7
void FASTCALL call_13_0_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_4_8
#define init_13_0_4_8
void FASTCALL call_13_0_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_4_9
#define init_13_0_4_9
void FASTCALL call_13_0_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_4_10
#define init_13_0_4_10
void FASTCALL call_13_0_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_4_11
#define init_13_0_4_11
void FASTCALL call_13_0_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_4_12
#define init_13_0_4_12
void FASTCALL call_13_0_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_4_13
#define init_13_0_4_13
void FASTCALL call_13_0_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_4_14
#define init_13_0_4_14
void FASTCALL call_13_0_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_4_15
#define init_13_0_4_15
void FASTCALL call_13_0_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_5_0
#define init_13_0_5_0
void FASTCALL call_13_0_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_5_1
#define init_13_0_5_1
void FASTCALL call_13_0_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_5_2
#define init_13_0_5_2
void FASTCALL call_13_0_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_5_3
#define init_13_0_5_3
void FASTCALL call_13_0_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_5_4
#define init_13_0_5_4
void FASTCALL call_13_0_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_5_5
#define init_13_0_5_5
void FASTCALL call_13_0_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_5_6
#define init_13_0_5_6
void FASTCALL call_13_0_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_5_7
#define init_13_0_5_7
void FASTCALL call_13_0_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_5_8
#define init_13_0_5_8
void FASTCALL call_13_0_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_5_9
#define init_13_0_5_9
void FASTCALL call_13_0_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_5_10
#define init_13_0_5_10
void FASTCALL call_13_0_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_5_11
#define init_13_0_5_11
void FASTCALL call_13_0_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_5_12
#define init_13_0_5_12
void FASTCALL call_13_0_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_5_13
#define init_13_0_5_13
void FASTCALL call_13_0_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_5_14
#define init_13_0_5_14
void FASTCALL call_13_0_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_5_15
#define init_13_0_5_15
void FASTCALL call_13_0_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_6_0
#define init_13_0_6_0
void FASTCALL call_13_0_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_6_1
#define init_13_0_6_1
void FASTCALL call_13_0_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_6_2
#define init_13_0_6_2
void FASTCALL call_13_0_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_6_3
#define init_13_0_6_3
void FASTCALL call_13_0_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_6_4
#define init_13_0_6_4
void FASTCALL call_13_0_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_6_5
#define init_13_0_6_5
void FASTCALL call_13_0_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_6_6
#define init_13_0_6_6
void FASTCALL call_13_0_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_6_7
#define init_13_0_6_7
void FASTCALL call_13_0_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_6_8
#define init_13_0_6_8
void FASTCALL call_13_0_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_6_9
#define init_13_0_6_9
void FASTCALL call_13_0_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_6_10
#define init_13_0_6_10
void FASTCALL call_13_0_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_6_11
#define init_13_0_6_11
void FASTCALL call_13_0_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_6_12
#define init_13_0_6_12
void FASTCALL call_13_0_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_6_13
#define init_13_0_6_13
void FASTCALL call_13_0_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_6_14
#define init_13_0_6_14
void FASTCALL call_13_0_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_6_15
#define init_13_0_6_15
void FASTCALL call_13_0_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_7_0
#define init_13_0_7_0
void FASTCALL call_13_0_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_7_1
#define init_13_0_7_1
void FASTCALL call_13_0_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_7_2
#define init_13_0_7_2
void FASTCALL call_13_0_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_7_3
#define init_13_0_7_3
void FASTCALL call_13_0_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_7_4
#define init_13_0_7_4
void FASTCALL call_13_0_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_7_5
#define init_13_0_7_5
void FASTCALL call_13_0_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_7_6
#define init_13_0_7_6
void FASTCALL call_13_0_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_7_7
#define init_13_0_7_7
void FASTCALL call_13_0_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_7_8
#define init_13_0_7_8
void FASTCALL call_13_0_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_7_9
#define init_13_0_7_9
void FASTCALL call_13_0_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_7_10
#define init_13_0_7_10
void FASTCALL call_13_0_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_7_11
#define init_13_0_7_11
void FASTCALL call_13_0_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_7_12
#define init_13_0_7_12
void FASTCALL call_13_0_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_7_13
#define init_13_0_7_13
void FASTCALL call_13_0_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_7_14
#define init_13_0_7_14
void FASTCALL call_13_0_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_7_15
#define init_13_0_7_15
void FASTCALL call_13_0_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_8_0
#define init_13_0_8_0
void FASTCALL call_13_0_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_8_1
#define init_13_0_8_1
void FASTCALL call_13_0_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_8_2
#define init_13_0_8_2
void FASTCALL call_13_0_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_8_3
#define init_13_0_8_3
void FASTCALL call_13_0_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_8_4
#define init_13_0_8_4
void FASTCALL call_13_0_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_8_5
#define init_13_0_8_5
void FASTCALL call_13_0_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_8_6
#define init_13_0_8_6
void FASTCALL call_13_0_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_8_7
#define init_13_0_8_7
void FASTCALL call_13_0_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_8_8
#define init_13_0_8_8
void FASTCALL call_13_0_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_8_9
#define init_13_0_8_9
void FASTCALL call_13_0_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_8_10
#define init_13_0_8_10
void FASTCALL call_13_0_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_8_11
#define init_13_0_8_11
void FASTCALL call_13_0_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_8_12
#define init_13_0_8_12
void FASTCALL call_13_0_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_8_13
#define init_13_0_8_13
void FASTCALL call_13_0_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_8_14
#define init_13_0_8_14
void FASTCALL call_13_0_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_8_15
#define init_13_0_8_15
void FASTCALL call_13_0_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_9_0
#define init_13_0_9_0
void FASTCALL call_13_0_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_9_1
#define init_13_0_9_1
void FASTCALL call_13_0_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_9_2
#define init_13_0_9_2
void FASTCALL call_13_0_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_9_3
#define init_13_0_9_3
void FASTCALL call_13_0_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_9_4
#define init_13_0_9_4
void FASTCALL call_13_0_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_9_5
#define init_13_0_9_5
void FASTCALL call_13_0_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_9_6
#define init_13_0_9_6
void FASTCALL call_13_0_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_9_7
#define init_13_0_9_7
void FASTCALL call_13_0_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_9_8
#define init_13_0_9_8
void FASTCALL call_13_0_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_9_9
#define init_13_0_9_9
void FASTCALL call_13_0_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_9_10
#define init_13_0_9_10
void FASTCALL call_13_0_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_9_11
#define init_13_0_9_11
void FASTCALL call_13_0_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_9_12
#define init_13_0_9_12
void FASTCALL call_13_0_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_9_13
#define init_13_0_9_13
void FASTCALL call_13_0_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_9_14
#define init_13_0_9_14
void FASTCALL call_13_0_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_9_15
#define init_13_0_9_15
void FASTCALL call_13_0_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_10_0
#define init_13_0_10_0
void FASTCALL call_13_0_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_10_1
#define init_13_0_10_1
void FASTCALL call_13_0_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_10_2
#define init_13_0_10_2
void FASTCALL call_13_0_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_10_3
#define init_13_0_10_3
void FASTCALL call_13_0_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_10_4
#define init_13_0_10_4
void FASTCALL call_13_0_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_10_5
#define init_13_0_10_5
void FASTCALL call_13_0_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_10_6
#define init_13_0_10_6
void FASTCALL call_13_0_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_10_7
#define init_13_0_10_7
void FASTCALL call_13_0_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_10_8
#define init_13_0_10_8
void FASTCALL call_13_0_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_10_9
#define init_13_0_10_9
void FASTCALL call_13_0_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_10_10
#define init_13_0_10_10
void FASTCALL call_13_0_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_10_11
#define init_13_0_10_11
void FASTCALL call_13_0_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_10_12
#define init_13_0_10_12
void FASTCALL call_13_0_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_10_13
#define init_13_0_10_13
void FASTCALL call_13_0_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_10_14
#define init_13_0_10_14
void FASTCALL call_13_0_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_10_15
#define init_13_0_10_15
void FASTCALL call_13_0_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_11_0
#define init_13_0_11_0
void FASTCALL call_13_0_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_11_1
#define init_13_0_11_1
void FASTCALL call_13_0_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_11_2
#define init_13_0_11_2
void FASTCALL call_13_0_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_11_3
#define init_13_0_11_3
void FASTCALL call_13_0_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_11_4
#define init_13_0_11_4
void FASTCALL call_13_0_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_11_5
#define init_13_0_11_5
void FASTCALL call_13_0_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_11_6
#define init_13_0_11_6
void FASTCALL call_13_0_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_11_7
#define init_13_0_11_7
void FASTCALL call_13_0_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_11_8
#define init_13_0_11_8
void FASTCALL call_13_0_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_11_9
#define init_13_0_11_9
void FASTCALL call_13_0_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_11_10
#define init_13_0_11_10
void FASTCALL call_13_0_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_11_11
#define init_13_0_11_11
void FASTCALL call_13_0_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_11_12
#define init_13_0_11_12
void FASTCALL call_13_0_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_11_13
#define init_13_0_11_13
void FASTCALL call_13_0_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_11_14
#define init_13_0_11_14
void FASTCALL call_13_0_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_11_15
#define init_13_0_11_15
void FASTCALL call_13_0_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_12_0
#define init_13_0_12_0
void FASTCALL call_13_0_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_12_1
#define init_13_0_12_1
void FASTCALL call_13_0_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_12_2
#define init_13_0_12_2
void FASTCALL call_13_0_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_12_3
#define init_13_0_12_3
void FASTCALL call_13_0_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_12_4
#define init_13_0_12_4
void FASTCALL call_13_0_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_12_5
#define init_13_0_12_5
void FASTCALL call_13_0_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_12_6
#define init_13_0_12_6
void FASTCALL call_13_0_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_12_7
#define init_13_0_12_7
void FASTCALL call_13_0_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_12_8
#define init_13_0_12_8
void FASTCALL call_13_0_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_12_9
#define init_13_0_12_9
void FASTCALL call_13_0_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_12_10
#define init_13_0_12_10
void FASTCALL call_13_0_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_12_11
#define init_13_0_12_11
void FASTCALL call_13_0_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_12_12
#define init_13_0_12_12
void FASTCALL call_13_0_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_12_13
#define init_13_0_12_13
void FASTCALL call_13_0_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_12_14
#define init_13_0_12_14
void FASTCALL call_13_0_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_12_15
#define init_13_0_12_15
void FASTCALL call_13_0_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_13_0
#define init_13_0_13_0
void FASTCALL call_13_0_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_13_1
#define init_13_0_13_1
void FASTCALL call_13_0_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_13_2
#define init_13_0_13_2
void FASTCALL call_13_0_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_13_3
#define init_13_0_13_3
void FASTCALL call_13_0_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_13_4
#define init_13_0_13_4
void FASTCALL call_13_0_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_13_5
#define init_13_0_13_5
void FASTCALL call_13_0_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_13_6
#define init_13_0_13_6
void FASTCALL call_13_0_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_13_7
#define init_13_0_13_7
void FASTCALL call_13_0_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_13_8
#define init_13_0_13_8
void FASTCALL call_13_0_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_13_9
#define init_13_0_13_9
void FASTCALL call_13_0_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_13_10
#define init_13_0_13_10
void FASTCALL call_13_0_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_13_11
#define init_13_0_13_11
void FASTCALL call_13_0_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_13_12
#define init_13_0_13_12
void FASTCALL call_13_0_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_13_13
#define init_13_0_13_13
void FASTCALL call_13_0_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_13_14
#define init_13_0_13_14
void FASTCALL call_13_0_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_13_15
#define init_13_0_13_15
void FASTCALL call_13_0_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_14_0
#define init_13_0_14_0
void FASTCALL call_13_0_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_14_1
#define init_13_0_14_1
void FASTCALL call_13_0_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_14_2
#define init_13_0_14_2
void FASTCALL call_13_0_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_14_3
#define init_13_0_14_3
void FASTCALL call_13_0_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_14_4
#define init_13_0_14_4
void FASTCALL call_13_0_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_14_5
#define init_13_0_14_5
void FASTCALL call_13_0_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_14_6
#define init_13_0_14_6
void FASTCALL call_13_0_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_14_7
#define init_13_0_14_7
void FASTCALL call_13_0_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_14_8
#define init_13_0_14_8
void FASTCALL call_13_0_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_14_9
#define init_13_0_14_9
void FASTCALL call_13_0_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_14_10
#define init_13_0_14_10
void FASTCALL call_13_0_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_14_11
#define init_13_0_14_11
void FASTCALL call_13_0_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_14_12
#define init_13_0_14_12
void FASTCALL call_13_0_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_14_13
#define init_13_0_14_13
void FASTCALL call_13_0_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_14_14
#define init_13_0_14_14
void FASTCALL call_13_0_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_14_15
#define init_13_0_14_15
void FASTCALL call_13_0_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_15_0
#define init_13_0_15_0
void FASTCALL call_13_0_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_15_1
#define init_13_0_15_1
void FASTCALL call_13_0_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_15_2
#define init_13_0_15_2
void FASTCALL call_13_0_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_15_3
#define init_13_0_15_3
void FASTCALL call_13_0_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_15_4
#define init_13_0_15_4
void FASTCALL call_13_0_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_15_5
#define init_13_0_15_5
void FASTCALL call_13_0_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_15_6
#define init_13_0_15_6
void FASTCALL call_13_0_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_15_7
#define init_13_0_15_7
void FASTCALL call_13_0_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_15_8
#define init_13_0_15_8
void FASTCALL call_13_0_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_15_9
#define init_13_0_15_9
void FASTCALL call_13_0_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_15_10
#define init_13_0_15_10
void FASTCALL call_13_0_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_15_11
#define init_13_0_15_11
void FASTCALL call_13_0_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_15_12
#define init_13_0_15_12
void FASTCALL call_13_0_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_15_13
#define init_13_0_15_13
void FASTCALL call_13_0_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_15_14
#define init_13_0_15_14
void FASTCALL call_13_0_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_0_15_15
#define init_13_0_15_15
void FASTCALL call_13_0_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_0_0
#define init_13_1_0_0
void FASTCALL call_13_1_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_0_1
#define init_13_1_0_1
void FASTCALL call_13_1_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_0_2
#define init_13_1_0_2
void FASTCALL call_13_1_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_0_3
#define init_13_1_0_3
void FASTCALL call_13_1_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_0_4
#define init_13_1_0_4
void FASTCALL call_13_1_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_0_5
#define init_13_1_0_5
void FASTCALL call_13_1_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_0_6
#define init_13_1_0_6
void FASTCALL call_13_1_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_0_7
#define init_13_1_0_7
void FASTCALL call_13_1_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_0_8
#define init_13_1_0_8
void FASTCALL call_13_1_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_0_9
#define init_13_1_0_9
void FASTCALL call_13_1_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_0_10
#define init_13_1_0_10
void FASTCALL call_13_1_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_0_11
#define init_13_1_0_11
void FASTCALL call_13_1_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_0_12
#define init_13_1_0_12
void FASTCALL call_13_1_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_0_13
#define init_13_1_0_13
void FASTCALL call_13_1_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_0_14
#define init_13_1_0_14
void FASTCALL call_13_1_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_0_15
#define init_13_1_0_15
void FASTCALL call_13_1_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_1_0
#define init_13_1_1_0
void FASTCALL call_13_1_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_1_1
#define init_13_1_1_1
void FASTCALL call_13_1_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_1_2
#define init_13_1_1_2
void FASTCALL call_13_1_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_1_3
#define init_13_1_1_3
void FASTCALL call_13_1_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_1_4
#define init_13_1_1_4
void FASTCALL call_13_1_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_1_5
#define init_13_1_1_5
void FASTCALL call_13_1_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_1_6
#define init_13_1_1_6
void FASTCALL call_13_1_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_1_7
#define init_13_1_1_7
void FASTCALL call_13_1_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_1_8
#define init_13_1_1_8
void FASTCALL call_13_1_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_1_9
#define init_13_1_1_9
void FASTCALL call_13_1_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_1_10
#define init_13_1_1_10
void FASTCALL call_13_1_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_1_11
#define init_13_1_1_11
void FASTCALL call_13_1_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_1_12
#define init_13_1_1_12
void FASTCALL call_13_1_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_1_13
#define init_13_1_1_13
void FASTCALL call_13_1_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_1_14
#define init_13_1_1_14
void FASTCALL call_13_1_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_1_15
#define init_13_1_1_15
void FASTCALL call_13_1_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_2_0
#define init_13_1_2_0
void FASTCALL call_13_1_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_2_1
#define init_13_1_2_1
void FASTCALL call_13_1_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_2_2
#define init_13_1_2_2
void FASTCALL call_13_1_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_2_3
#define init_13_1_2_3
void FASTCALL call_13_1_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_2_4
#define init_13_1_2_4
void FASTCALL call_13_1_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_2_5
#define init_13_1_2_5
void FASTCALL call_13_1_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_2_6
#define init_13_1_2_6
void FASTCALL call_13_1_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_2_7
#define init_13_1_2_7
void FASTCALL call_13_1_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_2_8
#define init_13_1_2_8
void FASTCALL call_13_1_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_2_9
#define init_13_1_2_9
void FASTCALL call_13_1_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_2_10
#define init_13_1_2_10
void FASTCALL call_13_1_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_2_11
#define init_13_1_2_11
void FASTCALL call_13_1_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_2_12
#define init_13_1_2_12
void FASTCALL call_13_1_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_2_13
#define init_13_1_2_13
void FASTCALL call_13_1_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_2_14
#define init_13_1_2_14
void FASTCALL call_13_1_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_2_15
#define init_13_1_2_15
void FASTCALL call_13_1_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_3_0
#define init_13_1_3_0
void FASTCALL call_13_1_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_3_1
#define init_13_1_3_1
void FASTCALL call_13_1_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_3_2
#define init_13_1_3_2
void FASTCALL call_13_1_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_3_3
#define init_13_1_3_3
void FASTCALL call_13_1_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_3_4
#define init_13_1_3_4
void FASTCALL call_13_1_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_3_5
#define init_13_1_3_5
void FASTCALL call_13_1_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_3_6
#define init_13_1_3_6
void FASTCALL call_13_1_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_3_7
#define init_13_1_3_7
void FASTCALL call_13_1_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_3_8
#define init_13_1_3_8
void FASTCALL call_13_1_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_3_9
#define init_13_1_3_9
void FASTCALL call_13_1_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_3_10
#define init_13_1_3_10
void FASTCALL call_13_1_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_3_11
#define init_13_1_3_11
void FASTCALL call_13_1_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_3_12
#define init_13_1_3_12
void FASTCALL call_13_1_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_3_13
#define init_13_1_3_13
void FASTCALL call_13_1_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_3_14
#define init_13_1_3_14
void FASTCALL call_13_1_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_3_15
#define init_13_1_3_15
void FASTCALL call_13_1_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_4_0
#define init_13_1_4_0
void FASTCALL call_13_1_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_4_1
#define init_13_1_4_1
void FASTCALL call_13_1_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_4_2
#define init_13_1_4_2
void FASTCALL call_13_1_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_4_3
#define init_13_1_4_3
void FASTCALL call_13_1_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_4_4
#define init_13_1_4_4
void FASTCALL call_13_1_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_4_5
#define init_13_1_4_5
void FASTCALL call_13_1_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_4_6
#define init_13_1_4_6
void FASTCALL call_13_1_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_4_7
#define init_13_1_4_7
void FASTCALL call_13_1_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_4_8
#define init_13_1_4_8
void FASTCALL call_13_1_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_4_9
#define init_13_1_4_9
void FASTCALL call_13_1_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_4_10
#define init_13_1_4_10
void FASTCALL call_13_1_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_4_11
#define init_13_1_4_11
void FASTCALL call_13_1_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_4_12
#define init_13_1_4_12
void FASTCALL call_13_1_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_4_13
#define init_13_1_4_13
void FASTCALL call_13_1_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_4_14
#define init_13_1_4_14
void FASTCALL call_13_1_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_4_15
#define init_13_1_4_15
void FASTCALL call_13_1_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_5_0
#define init_13_1_5_0
void FASTCALL call_13_1_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_5_1
#define init_13_1_5_1
void FASTCALL call_13_1_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_5_2
#define init_13_1_5_2
void FASTCALL call_13_1_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_5_3
#define init_13_1_5_3
void FASTCALL call_13_1_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_5_4
#define init_13_1_5_4
void FASTCALL call_13_1_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_5_5
#define init_13_1_5_5
void FASTCALL call_13_1_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_5_6
#define init_13_1_5_6
void FASTCALL call_13_1_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_5_7
#define init_13_1_5_7
void FASTCALL call_13_1_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_5_8
#define init_13_1_5_8
void FASTCALL call_13_1_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_5_9
#define init_13_1_5_9
void FASTCALL call_13_1_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_5_10
#define init_13_1_5_10
void FASTCALL call_13_1_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_5_11
#define init_13_1_5_11
void FASTCALL call_13_1_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_5_12
#define init_13_1_5_12
void FASTCALL call_13_1_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_5_13
#define init_13_1_5_13
void FASTCALL call_13_1_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_5_14
#define init_13_1_5_14
void FASTCALL call_13_1_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_5_15
#define init_13_1_5_15
void FASTCALL call_13_1_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_6_0
#define init_13_1_6_0
void FASTCALL call_13_1_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_6_1
#define init_13_1_6_1
void FASTCALL call_13_1_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_6_2
#define init_13_1_6_2
void FASTCALL call_13_1_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_6_3
#define init_13_1_6_3
void FASTCALL call_13_1_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_6_4
#define init_13_1_6_4
void FASTCALL call_13_1_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_6_5
#define init_13_1_6_5
void FASTCALL call_13_1_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_6_6
#define init_13_1_6_6
void FASTCALL call_13_1_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_6_7
#define init_13_1_6_7
void FASTCALL call_13_1_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_6_8
#define init_13_1_6_8
void FASTCALL call_13_1_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_6_9
#define init_13_1_6_9
void FASTCALL call_13_1_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_6_10
#define init_13_1_6_10
void FASTCALL call_13_1_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_6_11
#define init_13_1_6_11
void FASTCALL call_13_1_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_6_12
#define init_13_1_6_12
void FASTCALL call_13_1_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_6_13
#define init_13_1_6_13
void FASTCALL call_13_1_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_6_14
#define init_13_1_6_14
void FASTCALL call_13_1_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_6_15
#define init_13_1_6_15
void FASTCALL call_13_1_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_7_0
#define init_13_1_7_0
void FASTCALL call_13_1_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_7_1
#define init_13_1_7_1
void FASTCALL call_13_1_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_7_2
#define init_13_1_7_2
void FASTCALL call_13_1_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_7_3
#define init_13_1_7_3
void FASTCALL call_13_1_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_7_4
#define init_13_1_7_4
void FASTCALL call_13_1_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_7_5
#define init_13_1_7_5
void FASTCALL call_13_1_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_7_6
#define init_13_1_7_6
void FASTCALL call_13_1_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_7_7
#define init_13_1_7_7
void FASTCALL call_13_1_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_7_8
#define init_13_1_7_8
void FASTCALL call_13_1_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_7_9
#define init_13_1_7_9
void FASTCALL call_13_1_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_7_10
#define init_13_1_7_10
void FASTCALL call_13_1_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_7_11
#define init_13_1_7_11
void FASTCALL call_13_1_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_7_12
#define init_13_1_7_12
void FASTCALL call_13_1_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_7_13
#define init_13_1_7_13
void FASTCALL call_13_1_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_7_14
#define init_13_1_7_14
void FASTCALL call_13_1_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_7_15
#define init_13_1_7_15
void FASTCALL call_13_1_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_8_0
#define init_13_1_8_0
void FASTCALL call_13_1_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_8_1
#define init_13_1_8_1
void FASTCALL call_13_1_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_8_2
#define init_13_1_8_2
void FASTCALL call_13_1_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_8_3
#define init_13_1_8_3
void FASTCALL call_13_1_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_8_4
#define init_13_1_8_4
void FASTCALL call_13_1_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_8_5
#define init_13_1_8_5
void FASTCALL call_13_1_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_8_6
#define init_13_1_8_6
void FASTCALL call_13_1_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_8_7
#define init_13_1_8_7
void FASTCALL call_13_1_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_8_8
#define init_13_1_8_8
void FASTCALL call_13_1_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_8_9
#define init_13_1_8_9
void FASTCALL call_13_1_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_8_10
#define init_13_1_8_10
void FASTCALL call_13_1_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_8_11
#define init_13_1_8_11
void FASTCALL call_13_1_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_8_12
#define init_13_1_8_12
void FASTCALL call_13_1_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_8_13
#define init_13_1_8_13
void FASTCALL call_13_1_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_8_14
#define init_13_1_8_14
void FASTCALL call_13_1_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_8_15
#define init_13_1_8_15
void FASTCALL call_13_1_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_9_0
#define init_13_1_9_0
void FASTCALL call_13_1_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_9_1
#define init_13_1_9_1
void FASTCALL call_13_1_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_9_2
#define init_13_1_9_2
void FASTCALL call_13_1_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_9_3
#define init_13_1_9_3
void FASTCALL call_13_1_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_9_4
#define init_13_1_9_4
void FASTCALL call_13_1_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_9_5
#define init_13_1_9_5
void FASTCALL call_13_1_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_9_6
#define init_13_1_9_6
void FASTCALL call_13_1_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_9_7
#define init_13_1_9_7
void FASTCALL call_13_1_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_9_8
#define init_13_1_9_8
void FASTCALL call_13_1_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_9_9
#define init_13_1_9_9
void FASTCALL call_13_1_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_9_10
#define init_13_1_9_10
void FASTCALL call_13_1_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_9_11
#define init_13_1_9_11
void FASTCALL call_13_1_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_9_12
#define init_13_1_9_12
void FASTCALL call_13_1_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_9_13
#define init_13_1_9_13
void FASTCALL call_13_1_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_9_14
#define init_13_1_9_14
void FASTCALL call_13_1_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_9_15
#define init_13_1_9_15
void FASTCALL call_13_1_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_10_0
#define init_13_1_10_0
void FASTCALL call_13_1_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_10_1
#define init_13_1_10_1
void FASTCALL call_13_1_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_10_2
#define init_13_1_10_2
void FASTCALL call_13_1_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_10_3
#define init_13_1_10_3
void FASTCALL call_13_1_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_10_4
#define init_13_1_10_4
void FASTCALL call_13_1_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_10_5
#define init_13_1_10_5
void FASTCALL call_13_1_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_10_6
#define init_13_1_10_6
void FASTCALL call_13_1_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_10_7
#define init_13_1_10_7
void FASTCALL call_13_1_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_10_8
#define init_13_1_10_8
void FASTCALL call_13_1_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_10_9
#define init_13_1_10_9
void FASTCALL call_13_1_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_10_10
#define init_13_1_10_10
void FASTCALL call_13_1_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_10_11
#define init_13_1_10_11
void FASTCALL call_13_1_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_10_12
#define init_13_1_10_12
void FASTCALL call_13_1_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_10_13
#define init_13_1_10_13
void FASTCALL call_13_1_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_10_14
#define init_13_1_10_14
void FASTCALL call_13_1_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_10_15
#define init_13_1_10_15
void FASTCALL call_13_1_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_11_0
#define init_13_1_11_0
void FASTCALL call_13_1_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_11_1
#define init_13_1_11_1
void FASTCALL call_13_1_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_11_2
#define init_13_1_11_2
void FASTCALL call_13_1_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_11_3
#define init_13_1_11_3
void FASTCALL call_13_1_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_11_4
#define init_13_1_11_4
void FASTCALL call_13_1_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_11_5
#define init_13_1_11_5
void FASTCALL call_13_1_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_11_6
#define init_13_1_11_6
void FASTCALL call_13_1_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_11_7
#define init_13_1_11_7
void FASTCALL call_13_1_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_11_8
#define init_13_1_11_8
void FASTCALL call_13_1_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_11_9
#define init_13_1_11_9
void FASTCALL call_13_1_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_11_10
#define init_13_1_11_10
void FASTCALL call_13_1_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_11_11
#define init_13_1_11_11
void FASTCALL call_13_1_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_11_12
#define init_13_1_11_12
void FASTCALL call_13_1_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_11_13
#define init_13_1_11_13
void FASTCALL call_13_1_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_11_14
#define init_13_1_11_14
void FASTCALL call_13_1_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_11_15
#define init_13_1_11_15
void FASTCALL call_13_1_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_12_0
#define init_13_1_12_0
void FASTCALL call_13_1_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_12_1
#define init_13_1_12_1
void FASTCALL call_13_1_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_12_2
#define init_13_1_12_2
void FASTCALL call_13_1_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_12_3
#define init_13_1_12_3
void FASTCALL call_13_1_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_12_4
#define init_13_1_12_4
void FASTCALL call_13_1_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_12_5
#define init_13_1_12_5
void FASTCALL call_13_1_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_12_6
#define init_13_1_12_6
void FASTCALL call_13_1_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_12_7
#define init_13_1_12_7
void FASTCALL call_13_1_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_12_8
#define init_13_1_12_8
void FASTCALL call_13_1_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_12_9
#define init_13_1_12_9
void FASTCALL call_13_1_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_12_10
#define init_13_1_12_10
void FASTCALL call_13_1_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_12_11
#define init_13_1_12_11
void FASTCALL call_13_1_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_12_12
#define init_13_1_12_12
void FASTCALL call_13_1_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_12_13
#define init_13_1_12_13
void FASTCALL call_13_1_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_12_14
#define init_13_1_12_14
void FASTCALL call_13_1_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_12_15
#define init_13_1_12_15
void FASTCALL call_13_1_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_13_0
#define init_13_1_13_0
void FASTCALL call_13_1_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_13_1
#define init_13_1_13_1
void FASTCALL call_13_1_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_13_2
#define init_13_1_13_2
void FASTCALL call_13_1_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_13_3
#define init_13_1_13_3
void FASTCALL call_13_1_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_13_4
#define init_13_1_13_4
void FASTCALL call_13_1_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_13_5
#define init_13_1_13_5
void FASTCALL call_13_1_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_13_6
#define init_13_1_13_6
void FASTCALL call_13_1_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_13_7
#define init_13_1_13_7
void FASTCALL call_13_1_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_13_8
#define init_13_1_13_8
void FASTCALL call_13_1_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_13_9
#define init_13_1_13_9
void FASTCALL call_13_1_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_13_10
#define init_13_1_13_10
void FASTCALL call_13_1_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_13_11
#define init_13_1_13_11
void FASTCALL call_13_1_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_13_12
#define init_13_1_13_12
void FASTCALL call_13_1_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_13_13
#define init_13_1_13_13
void FASTCALL call_13_1_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_13_14
#define init_13_1_13_14
void FASTCALL call_13_1_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_13_15
#define init_13_1_13_15
void FASTCALL call_13_1_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_14_0
#define init_13_1_14_0
void FASTCALL call_13_1_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_14_1
#define init_13_1_14_1
void FASTCALL call_13_1_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_14_2
#define init_13_1_14_2
void FASTCALL call_13_1_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_14_3
#define init_13_1_14_3
void FASTCALL call_13_1_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_14_4
#define init_13_1_14_4
void FASTCALL call_13_1_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_14_5
#define init_13_1_14_5
void FASTCALL call_13_1_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_14_6
#define init_13_1_14_6
void FASTCALL call_13_1_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_14_7
#define init_13_1_14_7
void FASTCALL call_13_1_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_14_8
#define init_13_1_14_8
void FASTCALL call_13_1_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_14_9
#define init_13_1_14_9
void FASTCALL call_13_1_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_14_10
#define init_13_1_14_10
void FASTCALL call_13_1_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_14_11
#define init_13_1_14_11
void FASTCALL call_13_1_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_14_12
#define init_13_1_14_12
void FASTCALL call_13_1_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_14_13
#define init_13_1_14_13
void FASTCALL call_13_1_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_14_14
#define init_13_1_14_14
void FASTCALL call_13_1_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_14_15
#define init_13_1_14_15
void FASTCALL call_13_1_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_15_0
#define init_13_1_15_0
void FASTCALL call_13_1_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_15_1
#define init_13_1_15_1
void FASTCALL call_13_1_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_15_2
#define init_13_1_15_2
void FASTCALL call_13_1_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_15_3
#define init_13_1_15_3
void FASTCALL call_13_1_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_15_4
#define init_13_1_15_4
void FASTCALL call_13_1_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_15_5
#define init_13_1_15_5
void FASTCALL call_13_1_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_15_6
#define init_13_1_15_6
void FASTCALL call_13_1_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_15_7
#define init_13_1_15_7
void FASTCALL call_13_1_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_15_8
#define init_13_1_15_8
void FASTCALL call_13_1_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_15_9
#define init_13_1_15_9
void FASTCALL call_13_1_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_15_10
#define init_13_1_15_10
void FASTCALL call_13_1_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_15_11
#define init_13_1_15_11
void FASTCALL call_13_1_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_15_12
#define init_13_1_15_12
void FASTCALL call_13_1_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_15_13
#define init_13_1_15_13
void FASTCALL call_13_1_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_15_14
#define init_13_1_15_14
void FASTCALL call_13_1_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_1_15_15
#define init_13_1_15_15
void FASTCALL call_13_1_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_0_0
#define init_13_2_0_0
void FASTCALL call_13_2_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_0_1
#define init_13_2_0_1
void FASTCALL call_13_2_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_0_2
#define init_13_2_0_2
void FASTCALL call_13_2_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_0_3
#define init_13_2_0_3
void FASTCALL call_13_2_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_0_4
#define init_13_2_0_4
void FASTCALL call_13_2_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_0_5
#define init_13_2_0_5
void FASTCALL call_13_2_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_0_6
#define init_13_2_0_6
void FASTCALL call_13_2_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_0_7
#define init_13_2_0_7
void FASTCALL call_13_2_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_0_8
#define init_13_2_0_8
void FASTCALL call_13_2_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_0_9
#define init_13_2_0_9
void FASTCALL call_13_2_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_0_10
#define init_13_2_0_10
void FASTCALL call_13_2_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_0_11
#define init_13_2_0_11
void FASTCALL call_13_2_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_0_12
#define init_13_2_0_12
void FASTCALL call_13_2_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_0_13
#define init_13_2_0_13
void FASTCALL call_13_2_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_0_14
#define init_13_2_0_14
void FASTCALL call_13_2_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_0_15
#define init_13_2_0_15
void FASTCALL call_13_2_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_1_0
#define init_13_2_1_0
void FASTCALL call_13_2_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_1_1
#define init_13_2_1_1
void FASTCALL call_13_2_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_1_2
#define init_13_2_1_2
void FASTCALL call_13_2_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_1_3
#define init_13_2_1_3
void FASTCALL call_13_2_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_1_4
#define init_13_2_1_4
void FASTCALL call_13_2_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_1_5
#define init_13_2_1_5
void FASTCALL call_13_2_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_1_6
#define init_13_2_1_6
void FASTCALL call_13_2_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_1_7
#define init_13_2_1_7
void FASTCALL call_13_2_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_1_8
#define init_13_2_1_8
void FASTCALL call_13_2_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_1_9
#define init_13_2_1_9
void FASTCALL call_13_2_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_1_10
#define init_13_2_1_10
void FASTCALL call_13_2_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_1_11
#define init_13_2_1_11
void FASTCALL call_13_2_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_1_12
#define init_13_2_1_12
void FASTCALL call_13_2_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_1_13
#define init_13_2_1_13
void FASTCALL call_13_2_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_1_14
#define init_13_2_1_14
void FASTCALL call_13_2_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_1_15
#define init_13_2_1_15
void FASTCALL call_13_2_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_2_0
#define init_13_2_2_0
void FASTCALL call_13_2_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_2_1
#define init_13_2_2_1
void FASTCALL call_13_2_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_2_2
#define init_13_2_2_2
void FASTCALL call_13_2_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_2_3
#define init_13_2_2_3
void FASTCALL call_13_2_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_2_4
#define init_13_2_2_4
void FASTCALL call_13_2_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_2_5
#define init_13_2_2_5
void FASTCALL call_13_2_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_2_6
#define init_13_2_2_6
void FASTCALL call_13_2_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_2_7
#define init_13_2_2_7
void FASTCALL call_13_2_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_2_8
#define init_13_2_2_8
void FASTCALL call_13_2_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_2_9
#define init_13_2_2_9
void FASTCALL call_13_2_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_2_10
#define init_13_2_2_10
void FASTCALL call_13_2_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_2_11
#define init_13_2_2_11
void FASTCALL call_13_2_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_2_12
#define init_13_2_2_12
void FASTCALL call_13_2_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_2_13
#define init_13_2_2_13
void FASTCALL call_13_2_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_2_14
#define init_13_2_2_14
void FASTCALL call_13_2_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_2_15
#define init_13_2_2_15
void FASTCALL call_13_2_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_3_0
#define init_13_2_3_0
void FASTCALL call_13_2_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_3_1
#define init_13_2_3_1
void FASTCALL call_13_2_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_3_2
#define init_13_2_3_2
void FASTCALL call_13_2_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_3_3
#define init_13_2_3_3
void FASTCALL call_13_2_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_3_4
#define init_13_2_3_4
void FASTCALL call_13_2_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_3_5
#define init_13_2_3_5
void FASTCALL call_13_2_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_3_6
#define init_13_2_3_6
void FASTCALL call_13_2_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_3_7
#define init_13_2_3_7
void FASTCALL call_13_2_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_3_8
#define init_13_2_3_8
void FASTCALL call_13_2_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_3_9
#define init_13_2_3_9
void FASTCALL call_13_2_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_3_10
#define init_13_2_3_10
void FASTCALL call_13_2_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_3_11
#define init_13_2_3_11
void FASTCALL call_13_2_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_3_12
#define init_13_2_3_12
void FASTCALL call_13_2_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_3_13
#define init_13_2_3_13
void FASTCALL call_13_2_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_3_14
#define init_13_2_3_14
void FASTCALL call_13_2_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_3_15
#define init_13_2_3_15
void FASTCALL call_13_2_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_4_0
#define init_13_2_4_0
void FASTCALL call_13_2_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_4_1
#define init_13_2_4_1
void FASTCALL call_13_2_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_4_2
#define init_13_2_4_2
void FASTCALL call_13_2_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_4_3
#define init_13_2_4_3
void FASTCALL call_13_2_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_4_4
#define init_13_2_4_4
void FASTCALL call_13_2_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_4_5
#define init_13_2_4_5
void FASTCALL call_13_2_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_4_6
#define init_13_2_4_6
void FASTCALL call_13_2_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_4_7
#define init_13_2_4_7
void FASTCALL call_13_2_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_4_8
#define init_13_2_4_8
void FASTCALL call_13_2_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_4_9
#define init_13_2_4_9
void FASTCALL call_13_2_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_4_10
#define init_13_2_4_10
void FASTCALL call_13_2_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_4_11
#define init_13_2_4_11
void FASTCALL call_13_2_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_4_12
#define init_13_2_4_12
void FASTCALL call_13_2_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_4_13
#define init_13_2_4_13
void FASTCALL call_13_2_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_4_14
#define init_13_2_4_14
void FASTCALL call_13_2_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_4_15
#define init_13_2_4_15
void FASTCALL call_13_2_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_5_0
#define init_13_2_5_0
void FASTCALL call_13_2_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_5_1
#define init_13_2_5_1
void FASTCALL call_13_2_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_5_2
#define init_13_2_5_2
void FASTCALL call_13_2_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_5_3
#define init_13_2_5_3
void FASTCALL call_13_2_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_5_4
#define init_13_2_5_4
void FASTCALL call_13_2_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_5_5
#define init_13_2_5_5
void FASTCALL call_13_2_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_5_6
#define init_13_2_5_6
void FASTCALL call_13_2_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_5_7
#define init_13_2_5_7
void FASTCALL call_13_2_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_5_8
#define init_13_2_5_8
void FASTCALL call_13_2_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_5_9
#define init_13_2_5_9
void FASTCALL call_13_2_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_5_10
#define init_13_2_5_10
void FASTCALL call_13_2_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_5_11
#define init_13_2_5_11
void FASTCALL call_13_2_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_5_12
#define init_13_2_5_12
void FASTCALL call_13_2_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_5_13
#define init_13_2_5_13
void FASTCALL call_13_2_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_5_14
#define init_13_2_5_14
void FASTCALL call_13_2_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_5_15
#define init_13_2_5_15
void FASTCALL call_13_2_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_6_0
#define init_13_2_6_0
void FASTCALL call_13_2_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_6_1
#define init_13_2_6_1
void FASTCALL call_13_2_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_6_2
#define init_13_2_6_2
void FASTCALL call_13_2_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_6_3
#define init_13_2_6_3
void FASTCALL call_13_2_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_6_4
#define init_13_2_6_4
void FASTCALL call_13_2_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_6_5
#define init_13_2_6_5
void FASTCALL call_13_2_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_6_6
#define init_13_2_6_6
void FASTCALL call_13_2_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_6_7
#define init_13_2_6_7
void FASTCALL call_13_2_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_6_8
#define init_13_2_6_8
void FASTCALL call_13_2_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_6_9
#define init_13_2_6_9
void FASTCALL call_13_2_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_6_10
#define init_13_2_6_10
void FASTCALL call_13_2_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_6_11
#define init_13_2_6_11
void FASTCALL call_13_2_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_6_12
#define init_13_2_6_12
void FASTCALL call_13_2_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_6_13
#define init_13_2_6_13
void FASTCALL call_13_2_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_6_14
#define init_13_2_6_14
void FASTCALL call_13_2_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_6_15
#define init_13_2_6_15
void FASTCALL call_13_2_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_7_0
#define init_13_2_7_0
void FASTCALL call_13_2_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_7_1
#define init_13_2_7_1
void FASTCALL call_13_2_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_7_2
#define init_13_2_7_2
void FASTCALL call_13_2_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_7_3
#define init_13_2_7_3
void FASTCALL call_13_2_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_7_4
#define init_13_2_7_4
void FASTCALL call_13_2_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_7_5
#define init_13_2_7_5
void FASTCALL call_13_2_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_7_6
#define init_13_2_7_6
void FASTCALL call_13_2_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_7_7
#define init_13_2_7_7
void FASTCALL call_13_2_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_7_8
#define init_13_2_7_8
void FASTCALL call_13_2_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_7_9
#define init_13_2_7_9
void FASTCALL call_13_2_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_7_10
#define init_13_2_7_10
void FASTCALL call_13_2_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_7_11
#define init_13_2_7_11
void FASTCALL call_13_2_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_7_12
#define init_13_2_7_12
void FASTCALL call_13_2_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_7_13
#define init_13_2_7_13
void FASTCALL call_13_2_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_7_14
#define init_13_2_7_14
void FASTCALL call_13_2_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_7_15
#define init_13_2_7_15
void FASTCALL call_13_2_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_8_0
#define init_13_2_8_0
void FASTCALL call_13_2_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_8_1
#define init_13_2_8_1
void FASTCALL call_13_2_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_8_2
#define init_13_2_8_2
void FASTCALL call_13_2_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_8_3
#define init_13_2_8_3
void FASTCALL call_13_2_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_8_4
#define init_13_2_8_4
void FASTCALL call_13_2_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_8_5
#define init_13_2_8_5
void FASTCALL call_13_2_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_8_6
#define init_13_2_8_6
void FASTCALL call_13_2_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_8_7
#define init_13_2_8_7
void FASTCALL call_13_2_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_8_8
#define init_13_2_8_8
void FASTCALL call_13_2_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_8_9
#define init_13_2_8_9
void FASTCALL call_13_2_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_8_10
#define init_13_2_8_10
void FASTCALL call_13_2_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_8_11
#define init_13_2_8_11
void FASTCALL call_13_2_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_8_12
#define init_13_2_8_12
void FASTCALL call_13_2_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_8_13
#define init_13_2_8_13
void FASTCALL call_13_2_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_8_14
#define init_13_2_8_14
void FASTCALL call_13_2_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_8_15
#define init_13_2_8_15
void FASTCALL call_13_2_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_9_0
#define init_13_2_9_0
void FASTCALL call_13_2_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_9_1
#define init_13_2_9_1
void FASTCALL call_13_2_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_9_2
#define init_13_2_9_2
void FASTCALL call_13_2_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_9_3
#define init_13_2_9_3
void FASTCALL call_13_2_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_9_4
#define init_13_2_9_4
void FASTCALL call_13_2_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_9_5
#define init_13_2_9_5
void FASTCALL call_13_2_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_9_6
#define init_13_2_9_6
void FASTCALL call_13_2_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_9_7
#define init_13_2_9_7
void FASTCALL call_13_2_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_9_8
#define init_13_2_9_8
void FASTCALL call_13_2_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_9_9
#define init_13_2_9_9
void FASTCALL call_13_2_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_9_10
#define init_13_2_9_10
void FASTCALL call_13_2_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_9_11
#define init_13_2_9_11
void FASTCALL call_13_2_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_9_12
#define init_13_2_9_12
void FASTCALL call_13_2_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_9_13
#define init_13_2_9_13
void FASTCALL call_13_2_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_9_14
#define init_13_2_9_14
void FASTCALL call_13_2_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_9_15
#define init_13_2_9_15
void FASTCALL call_13_2_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_10_0
#define init_13_2_10_0
void FASTCALL call_13_2_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_10_1
#define init_13_2_10_1
void FASTCALL call_13_2_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_10_2
#define init_13_2_10_2
void FASTCALL call_13_2_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_10_3
#define init_13_2_10_3
void FASTCALL call_13_2_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_10_4
#define init_13_2_10_4
void FASTCALL call_13_2_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_10_5
#define init_13_2_10_5
void FASTCALL call_13_2_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_10_6
#define init_13_2_10_6
void FASTCALL call_13_2_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_10_7
#define init_13_2_10_7
void FASTCALL call_13_2_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_10_8
#define init_13_2_10_8
void FASTCALL call_13_2_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_10_9
#define init_13_2_10_9
void FASTCALL call_13_2_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_10_10
#define init_13_2_10_10
void FASTCALL call_13_2_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_10_11
#define init_13_2_10_11
void FASTCALL call_13_2_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_10_12
#define init_13_2_10_12
void FASTCALL call_13_2_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_10_13
#define init_13_2_10_13
void FASTCALL call_13_2_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_10_14
#define init_13_2_10_14
void FASTCALL call_13_2_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_10_15
#define init_13_2_10_15
void FASTCALL call_13_2_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_11_0
#define init_13_2_11_0
void FASTCALL call_13_2_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_11_1
#define init_13_2_11_1
void FASTCALL call_13_2_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_11_2
#define init_13_2_11_2
void FASTCALL call_13_2_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_11_3
#define init_13_2_11_3
void FASTCALL call_13_2_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_11_4
#define init_13_2_11_4
void FASTCALL call_13_2_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_11_5
#define init_13_2_11_5
void FASTCALL call_13_2_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_11_6
#define init_13_2_11_6
void FASTCALL call_13_2_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_11_7
#define init_13_2_11_7
void FASTCALL call_13_2_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_11_8
#define init_13_2_11_8
void FASTCALL call_13_2_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_11_9
#define init_13_2_11_9
void FASTCALL call_13_2_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_11_10
#define init_13_2_11_10
void FASTCALL call_13_2_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_11_11
#define init_13_2_11_11
void FASTCALL call_13_2_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_11_12
#define init_13_2_11_12
void FASTCALL call_13_2_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_11_13
#define init_13_2_11_13
void FASTCALL call_13_2_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_11_14
#define init_13_2_11_14
void FASTCALL call_13_2_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_11_15
#define init_13_2_11_15
void FASTCALL call_13_2_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_12_0
#define init_13_2_12_0
void FASTCALL call_13_2_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_12_1
#define init_13_2_12_1
void FASTCALL call_13_2_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_12_2
#define init_13_2_12_2
void FASTCALL call_13_2_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_12_3
#define init_13_2_12_3
void FASTCALL call_13_2_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_12_4
#define init_13_2_12_4
void FASTCALL call_13_2_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_12_5
#define init_13_2_12_5
void FASTCALL call_13_2_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_12_6
#define init_13_2_12_6
void FASTCALL call_13_2_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_12_7
#define init_13_2_12_7
void FASTCALL call_13_2_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_12_8
#define init_13_2_12_8
void FASTCALL call_13_2_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_12_9
#define init_13_2_12_9
void FASTCALL call_13_2_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_12_10
#define init_13_2_12_10
void FASTCALL call_13_2_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_12_11
#define init_13_2_12_11
void FASTCALL call_13_2_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_12_12
#define init_13_2_12_12
void FASTCALL call_13_2_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_12_13
#define init_13_2_12_13
void FASTCALL call_13_2_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_12_14
#define init_13_2_12_14
void FASTCALL call_13_2_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_12_15
#define init_13_2_12_15
void FASTCALL call_13_2_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_13_0
#define init_13_2_13_0
void FASTCALL call_13_2_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_13_1
#define init_13_2_13_1
void FASTCALL call_13_2_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_13_2
#define init_13_2_13_2
void FASTCALL call_13_2_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_13_3
#define init_13_2_13_3
void FASTCALL call_13_2_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_13_4
#define init_13_2_13_4
void FASTCALL call_13_2_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_13_5
#define init_13_2_13_5
void FASTCALL call_13_2_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_13_6
#define init_13_2_13_6
void FASTCALL call_13_2_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_13_7
#define init_13_2_13_7
void FASTCALL call_13_2_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_13_8
#define init_13_2_13_8
void FASTCALL call_13_2_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_13_9
#define init_13_2_13_9
void FASTCALL call_13_2_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_13_10
#define init_13_2_13_10
void FASTCALL call_13_2_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_13_11
#define init_13_2_13_11
void FASTCALL call_13_2_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_13_12
#define init_13_2_13_12
void FASTCALL call_13_2_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_13_13
#define init_13_2_13_13
void FASTCALL call_13_2_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_13_14
#define init_13_2_13_14
void FASTCALL call_13_2_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_13_15
#define init_13_2_13_15
void FASTCALL call_13_2_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_14_0
#define init_13_2_14_0
void FASTCALL call_13_2_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_14_1
#define init_13_2_14_1
void FASTCALL call_13_2_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_14_2
#define init_13_2_14_2
void FASTCALL call_13_2_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_14_3
#define init_13_2_14_3
void FASTCALL call_13_2_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_14_4
#define init_13_2_14_4
void FASTCALL call_13_2_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_14_5
#define init_13_2_14_5
void FASTCALL call_13_2_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_14_6
#define init_13_2_14_6
void FASTCALL call_13_2_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_14_7
#define init_13_2_14_7
void FASTCALL call_13_2_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_14_8
#define init_13_2_14_8
void FASTCALL call_13_2_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_14_9
#define init_13_2_14_9
void FASTCALL call_13_2_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_14_10
#define init_13_2_14_10
void FASTCALL call_13_2_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_14_11
#define init_13_2_14_11
void FASTCALL call_13_2_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_14_12
#define init_13_2_14_12
void FASTCALL call_13_2_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_14_13
#define init_13_2_14_13
void FASTCALL call_13_2_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_14_14
#define init_13_2_14_14
void FASTCALL call_13_2_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_14_15
#define init_13_2_14_15
void FASTCALL call_13_2_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_15_0
#define init_13_2_15_0
void FASTCALL call_13_2_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_15_1
#define init_13_2_15_1
void FASTCALL call_13_2_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_15_2
#define init_13_2_15_2
void FASTCALL call_13_2_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_15_3
#define init_13_2_15_3
void FASTCALL call_13_2_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_15_4
#define init_13_2_15_4
void FASTCALL call_13_2_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_15_5
#define init_13_2_15_5
void FASTCALL call_13_2_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_15_6
#define init_13_2_15_6
void FASTCALL call_13_2_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_15_7
#define init_13_2_15_7
void FASTCALL call_13_2_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_15_8
#define init_13_2_15_8
void FASTCALL call_13_2_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_15_9
#define init_13_2_15_9
void FASTCALL call_13_2_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_15_10
#define init_13_2_15_10
void FASTCALL call_13_2_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_15_11
#define init_13_2_15_11
void FASTCALL call_13_2_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_15_12
#define init_13_2_15_12
void FASTCALL call_13_2_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_15_13
#define init_13_2_15_13
void FASTCALL call_13_2_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_15_14
#define init_13_2_15_14
void FASTCALL call_13_2_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_2_15_15
#define init_13_2_15_15
void FASTCALL call_13_2_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_0_0
#define init_13_3_0_0
void FASTCALL call_13_3_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_0_1
#define init_13_3_0_1
void FASTCALL call_13_3_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_0_2
#define init_13_3_0_2
void FASTCALL call_13_3_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_0_3
#define init_13_3_0_3
void FASTCALL call_13_3_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_0_4
#define init_13_3_0_4
void FASTCALL call_13_3_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_0_5
#define init_13_3_0_5
void FASTCALL call_13_3_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_0_6
#define init_13_3_0_6
void FASTCALL call_13_3_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_0_7
#define init_13_3_0_7
void FASTCALL call_13_3_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_0_8
#define init_13_3_0_8
void FASTCALL call_13_3_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_0_9
#define init_13_3_0_9
void FASTCALL call_13_3_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_0_10
#define init_13_3_0_10
void FASTCALL call_13_3_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_0_11
#define init_13_3_0_11
void FASTCALL call_13_3_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_0_12
#define init_13_3_0_12
void FASTCALL call_13_3_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_0_13
#define init_13_3_0_13
void FASTCALL call_13_3_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_0_14
#define init_13_3_0_14
void FASTCALL call_13_3_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_0_15
#define init_13_3_0_15
void FASTCALL call_13_3_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_1_0
#define init_13_3_1_0
void FASTCALL call_13_3_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_1_1
#define init_13_3_1_1
void FASTCALL call_13_3_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_1_2
#define init_13_3_1_2
void FASTCALL call_13_3_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_1_3
#define init_13_3_1_3
void FASTCALL call_13_3_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_1_4
#define init_13_3_1_4
void FASTCALL call_13_3_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_1_5
#define init_13_3_1_5
void FASTCALL call_13_3_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_1_6
#define init_13_3_1_6
void FASTCALL call_13_3_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_1_7
#define init_13_3_1_7
void FASTCALL call_13_3_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_1_8
#define init_13_3_1_8
void FASTCALL call_13_3_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_1_9
#define init_13_3_1_9
void FASTCALL call_13_3_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_1_10
#define init_13_3_1_10
void FASTCALL call_13_3_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_1_11
#define init_13_3_1_11
void FASTCALL call_13_3_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_1_12
#define init_13_3_1_12
void FASTCALL call_13_3_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_1_13
#define init_13_3_1_13
void FASTCALL call_13_3_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_1_14
#define init_13_3_1_14
void FASTCALL call_13_3_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_1_15
#define init_13_3_1_15
void FASTCALL call_13_3_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_2_0
#define init_13_3_2_0
void FASTCALL call_13_3_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_2_1
#define init_13_3_2_1
void FASTCALL call_13_3_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_2_2
#define init_13_3_2_2
void FASTCALL call_13_3_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_2_3
#define init_13_3_2_3
void FASTCALL call_13_3_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_2_4
#define init_13_3_2_4
void FASTCALL call_13_3_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_2_5
#define init_13_3_2_5
void FASTCALL call_13_3_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_2_6
#define init_13_3_2_6
void FASTCALL call_13_3_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_2_7
#define init_13_3_2_7
void FASTCALL call_13_3_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_2_8
#define init_13_3_2_8
void FASTCALL call_13_3_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_2_9
#define init_13_3_2_9
void FASTCALL call_13_3_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_2_10
#define init_13_3_2_10
void FASTCALL call_13_3_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_2_11
#define init_13_3_2_11
void FASTCALL call_13_3_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_2_12
#define init_13_3_2_12
void FASTCALL call_13_3_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_2_13
#define init_13_3_2_13
void FASTCALL call_13_3_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_2_14
#define init_13_3_2_14
void FASTCALL call_13_3_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_2_15
#define init_13_3_2_15
void FASTCALL call_13_3_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_3_0
#define init_13_3_3_0
void FASTCALL call_13_3_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_3_1
#define init_13_3_3_1
void FASTCALL call_13_3_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_3_2
#define init_13_3_3_2
void FASTCALL call_13_3_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_3_3
#define init_13_3_3_3
void FASTCALL call_13_3_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_3_4
#define init_13_3_3_4
void FASTCALL call_13_3_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_3_5
#define init_13_3_3_5
void FASTCALL call_13_3_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_3_6
#define init_13_3_3_6
void FASTCALL call_13_3_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_3_7
#define init_13_3_3_7
void FASTCALL call_13_3_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_3_8
#define init_13_3_3_8
void FASTCALL call_13_3_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_3_9
#define init_13_3_3_9
void FASTCALL call_13_3_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_3_10
#define init_13_3_3_10
void FASTCALL call_13_3_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_3_11
#define init_13_3_3_11
void FASTCALL call_13_3_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_3_12
#define init_13_3_3_12
void FASTCALL call_13_3_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_3_13
#define init_13_3_3_13
void FASTCALL call_13_3_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_3_14
#define init_13_3_3_14
void FASTCALL call_13_3_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_3_15
#define init_13_3_3_15
void FASTCALL call_13_3_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_4_0
#define init_13_3_4_0
void FASTCALL call_13_3_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_4_1
#define init_13_3_4_1
void FASTCALL call_13_3_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_4_2
#define init_13_3_4_2
void FASTCALL call_13_3_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_4_3
#define init_13_3_4_3
void FASTCALL call_13_3_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_4_4
#define init_13_3_4_4
void FASTCALL call_13_3_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_4_5
#define init_13_3_4_5
void FASTCALL call_13_3_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_4_6
#define init_13_3_4_6
void FASTCALL call_13_3_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_4_7
#define init_13_3_4_7
void FASTCALL call_13_3_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_4_8
#define init_13_3_4_8
void FASTCALL call_13_3_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_4_9
#define init_13_3_4_9
void FASTCALL call_13_3_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_4_10
#define init_13_3_4_10
void FASTCALL call_13_3_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_4_11
#define init_13_3_4_11
void FASTCALL call_13_3_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_4_12
#define init_13_3_4_12
void FASTCALL call_13_3_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_4_13
#define init_13_3_4_13
void FASTCALL call_13_3_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_4_14
#define init_13_3_4_14
void FASTCALL call_13_3_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_4_15
#define init_13_3_4_15
void FASTCALL call_13_3_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_5_0
#define init_13_3_5_0
void FASTCALL call_13_3_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_5_1
#define init_13_3_5_1
void FASTCALL call_13_3_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_5_2
#define init_13_3_5_2
void FASTCALL call_13_3_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_5_3
#define init_13_3_5_3
void FASTCALL call_13_3_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_5_4
#define init_13_3_5_4
void FASTCALL call_13_3_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_5_5
#define init_13_3_5_5
void FASTCALL call_13_3_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_5_6
#define init_13_3_5_6
void FASTCALL call_13_3_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_5_7
#define init_13_3_5_7
void FASTCALL call_13_3_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_5_8
#define init_13_3_5_8
void FASTCALL call_13_3_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_5_9
#define init_13_3_5_9
void FASTCALL call_13_3_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_5_10
#define init_13_3_5_10
void FASTCALL call_13_3_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_5_11
#define init_13_3_5_11
void FASTCALL call_13_3_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_5_12
#define init_13_3_5_12
void FASTCALL call_13_3_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_5_13
#define init_13_3_5_13
void FASTCALL call_13_3_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_5_14
#define init_13_3_5_14
void FASTCALL call_13_3_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_5_15
#define init_13_3_5_15
void FASTCALL call_13_3_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_6_0
#define init_13_3_6_0
void FASTCALL call_13_3_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_6_1
#define init_13_3_6_1
void FASTCALL call_13_3_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_6_2
#define init_13_3_6_2
void FASTCALL call_13_3_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_6_3
#define init_13_3_6_3
void FASTCALL call_13_3_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_6_4
#define init_13_3_6_4
void FASTCALL call_13_3_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_6_5
#define init_13_3_6_5
void FASTCALL call_13_3_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_6_6
#define init_13_3_6_6
void FASTCALL call_13_3_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_6_7
#define init_13_3_6_7
void FASTCALL call_13_3_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_6_8
#define init_13_3_6_8
void FASTCALL call_13_3_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_6_9
#define init_13_3_6_9
void FASTCALL call_13_3_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_6_10
#define init_13_3_6_10
void FASTCALL call_13_3_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_6_11
#define init_13_3_6_11
void FASTCALL call_13_3_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_6_12
#define init_13_3_6_12
void FASTCALL call_13_3_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_6_13
#define init_13_3_6_13
void FASTCALL call_13_3_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_6_14
#define init_13_3_6_14
void FASTCALL call_13_3_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_6_15
#define init_13_3_6_15
void FASTCALL call_13_3_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_7_0
#define init_13_3_7_0
void FASTCALL call_13_3_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_7_1
#define init_13_3_7_1
void FASTCALL call_13_3_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_7_2
#define init_13_3_7_2
void FASTCALL call_13_3_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_7_3
#define init_13_3_7_3
void FASTCALL call_13_3_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_7_4
#define init_13_3_7_4
void FASTCALL call_13_3_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_7_5
#define init_13_3_7_5
void FASTCALL call_13_3_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_7_6
#define init_13_3_7_6
void FASTCALL call_13_3_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_7_7
#define init_13_3_7_7
void FASTCALL call_13_3_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_7_8
#define init_13_3_7_8
void FASTCALL call_13_3_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_7_9
#define init_13_3_7_9
void FASTCALL call_13_3_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_7_10
#define init_13_3_7_10
void FASTCALL call_13_3_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_7_11
#define init_13_3_7_11
void FASTCALL call_13_3_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_7_12
#define init_13_3_7_12
void FASTCALL call_13_3_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_7_13
#define init_13_3_7_13
void FASTCALL call_13_3_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_7_14
#define init_13_3_7_14
void FASTCALL call_13_3_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_7_15
#define init_13_3_7_15
void FASTCALL call_13_3_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_8_0
#define init_13_3_8_0
void FASTCALL call_13_3_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_8_1
#define init_13_3_8_1
void FASTCALL call_13_3_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_8_2
#define init_13_3_8_2
void FASTCALL call_13_3_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_8_3
#define init_13_3_8_3
void FASTCALL call_13_3_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_8_4
#define init_13_3_8_4
void FASTCALL call_13_3_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_8_5
#define init_13_3_8_5
void FASTCALL call_13_3_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_8_6
#define init_13_3_8_6
void FASTCALL call_13_3_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_8_7
#define init_13_3_8_7
void FASTCALL call_13_3_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_8_8
#define init_13_3_8_8
void FASTCALL call_13_3_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_8_9
#define init_13_3_8_9
void FASTCALL call_13_3_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_8_10
#define init_13_3_8_10
void FASTCALL call_13_3_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_8_11
#define init_13_3_8_11
void FASTCALL call_13_3_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_8_12
#define init_13_3_8_12
void FASTCALL call_13_3_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_8_13
#define init_13_3_8_13
void FASTCALL call_13_3_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_8_14
#define init_13_3_8_14
void FASTCALL call_13_3_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_8_15
#define init_13_3_8_15
void FASTCALL call_13_3_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_9_0
#define init_13_3_9_0
void FASTCALL call_13_3_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_9_1
#define init_13_3_9_1
void FASTCALL call_13_3_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_9_2
#define init_13_3_9_2
void FASTCALL call_13_3_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_9_3
#define init_13_3_9_3
void FASTCALL call_13_3_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_9_4
#define init_13_3_9_4
void FASTCALL call_13_3_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_9_5
#define init_13_3_9_5
void FASTCALL call_13_3_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_9_6
#define init_13_3_9_6
void FASTCALL call_13_3_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_9_7
#define init_13_3_9_7
void FASTCALL call_13_3_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_9_8
#define init_13_3_9_8
void FASTCALL call_13_3_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_9_9
#define init_13_3_9_9
void FASTCALL call_13_3_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_9_10
#define init_13_3_9_10
void FASTCALL call_13_3_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_9_11
#define init_13_3_9_11
void FASTCALL call_13_3_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_9_12
#define init_13_3_9_12
void FASTCALL call_13_3_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_9_13
#define init_13_3_9_13
void FASTCALL call_13_3_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_9_14
#define init_13_3_9_14
void FASTCALL call_13_3_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_9_15
#define init_13_3_9_15
void FASTCALL call_13_3_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_10_0
#define init_13_3_10_0
void FASTCALL call_13_3_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_10_1
#define init_13_3_10_1
void FASTCALL call_13_3_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_10_2
#define init_13_3_10_2
void FASTCALL call_13_3_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_10_3
#define init_13_3_10_3
void FASTCALL call_13_3_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_10_4
#define init_13_3_10_4
void FASTCALL call_13_3_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_10_5
#define init_13_3_10_5
void FASTCALL call_13_3_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_10_6
#define init_13_3_10_6
void FASTCALL call_13_3_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_10_7
#define init_13_3_10_7
void FASTCALL call_13_3_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_10_8
#define init_13_3_10_8
void FASTCALL call_13_3_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_10_9
#define init_13_3_10_9
void FASTCALL call_13_3_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_10_10
#define init_13_3_10_10
void FASTCALL call_13_3_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_10_11
#define init_13_3_10_11
void FASTCALL call_13_3_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_10_12
#define init_13_3_10_12
void FASTCALL call_13_3_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_10_13
#define init_13_3_10_13
void FASTCALL call_13_3_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_10_14
#define init_13_3_10_14
void FASTCALL call_13_3_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_10_15
#define init_13_3_10_15
void FASTCALL call_13_3_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_11_0
#define init_13_3_11_0
void FASTCALL call_13_3_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_11_1
#define init_13_3_11_1
void FASTCALL call_13_3_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_11_2
#define init_13_3_11_2
void FASTCALL call_13_3_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_11_3
#define init_13_3_11_3
void FASTCALL call_13_3_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_11_4
#define init_13_3_11_4
void FASTCALL call_13_3_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_11_5
#define init_13_3_11_5
void FASTCALL call_13_3_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_11_6
#define init_13_3_11_6
void FASTCALL call_13_3_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_11_7
#define init_13_3_11_7
void FASTCALL call_13_3_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_11_8
#define init_13_3_11_8
void FASTCALL call_13_3_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_11_9
#define init_13_3_11_9
void FASTCALL call_13_3_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_11_10
#define init_13_3_11_10
void FASTCALL call_13_3_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_11_11
#define init_13_3_11_11
void FASTCALL call_13_3_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_11_12
#define init_13_3_11_12
void FASTCALL call_13_3_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_11_13
#define init_13_3_11_13
void FASTCALL call_13_3_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_11_14
#define init_13_3_11_14
void FASTCALL call_13_3_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_11_15
#define init_13_3_11_15
void FASTCALL call_13_3_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_12_0
#define init_13_3_12_0
void FASTCALL call_13_3_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_12_1
#define init_13_3_12_1
void FASTCALL call_13_3_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_12_2
#define init_13_3_12_2
void FASTCALL call_13_3_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_12_3
#define init_13_3_12_3
void FASTCALL call_13_3_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_12_4
#define init_13_3_12_4
void FASTCALL call_13_3_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_12_5
#define init_13_3_12_5
void FASTCALL call_13_3_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_12_6
#define init_13_3_12_6
void FASTCALL call_13_3_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_12_7
#define init_13_3_12_7
void FASTCALL call_13_3_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_12_8
#define init_13_3_12_8
void FASTCALL call_13_3_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_12_9
#define init_13_3_12_9
void FASTCALL call_13_3_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_12_10
#define init_13_3_12_10
void FASTCALL call_13_3_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_12_11
#define init_13_3_12_11
void FASTCALL call_13_3_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_12_12
#define init_13_3_12_12
void FASTCALL call_13_3_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_12_13
#define init_13_3_12_13
void FASTCALL call_13_3_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_12_14
#define init_13_3_12_14
void FASTCALL call_13_3_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_12_15
#define init_13_3_12_15
void FASTCALL call_13_3_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_13_0
#define init_13_3_13_0
void FASTCALL call_13_3_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_13_1
#define init_13_3_13_1
void FASTCALL call_13_3_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_13_2
#define init_13_3_13_2
void FASTCALL call_13_3_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_13_3
#define init_13_3_13_3
void FASTCALL call_13_3_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_13_4
#define init_13_3_13_4
void FASTCALL call_13_3_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_13_5
#define init_13_3_13_5
void FASTCALL call_13_3_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_13_6
#define init_13_3_13_6
void FASTCALL call_13_3_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_13_7
#define init_13_3_13_7
void FASTCALL call_13_3_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_13_8
#define init_13_3_13_8
void FASTCALL call_13_3_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_13_9
#define init_13_3_13_9
void FASTCALL call_13_3_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_13_10
#define init_13_3_13_10
void FASTCALL call_13_3_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_13_11
#define init_13_3_13_11
void FASTCALL call_13_3_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_13_12
#define init_13_3_13_12
void FASTCALL call_13_3_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_13_13
#define init_13_3_13_13
void FASTCALL call_13_3_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_13_14
#define init_13_3_13_14
void FASTCALL call_13_3_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_13_15
#define init_13_3_13_15
void FASTCALL call_13_3_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_14_0
#define init_13_3_14_0
void FASTCALL call_13_3_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_14_1
#define init_13_3_14_1
void FASTCALL call_13_3_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_14_2
#define init_13_3_14_2
void FASTCALL call_13_3_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_14_3
#define init_13_3_14_3
void FASTCALL call_13_3_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_14_4
#define init_13_3_14_4
void FASTCALL call_13_3_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_14_5
#define init_13_3_14_5
void FASTCALL call_13_3_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_14_6
#define init_13_3_14_6
void FASTCALL call_13_3_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_14_7
#define init_13_3_14_7
void FASTCALL call_13_3_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_14_8
#define init_13_3_14_8
void FASTCALL call_13_3_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_14_9
#define init_13_3_14_9
void FASTCALL call_13_3_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_14_10
#define init_13_3_14_10
void FASTCALL call_13_3_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_14_11
#define init_13_3_14_11
void FASTCALL call_13_3_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_14_12
#define init_13_3_14_12
void FASTCALL call_13_3_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_14_13
#define init_13_3_14_13
void FASTCALL call_13_3_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_14_14
#define init_13_3_14_14
void FASTCALL call_13_3_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_14_15
#define init_13_3_14_15
void FASTCALL call_13_3_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_15_0
#define init_13_3_15_0
void FASTCALL call_13_3_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_15_1
#define init_13_3_15_1
void FASTCALL call_13_3_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_15_2
#define init_13_3_15_2
void FASTCALL call_13_3_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_15_3
#define init_13_3_15_3
void FASTCALL call_13_3_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_15_4
#define init_13_3_15_4
void FASTCALL call_13_3_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_15_5
#define init_13_3_15_5
void FASTCALL call_13_3_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_15_6
#define init_13_3_15_6
void FASTCALL call_13_3_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_15_7
#define init_13_3_15_7
void FASTCALL call_13_3_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_15_8
#define init_13_3_15_8
void FASTCALL call_13_3_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_15_9
#define init_13_3_15_9
void FASTCALL call_13_3_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_15_10
#define init_13_3_15_10
void FASTCALL call_13_3_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_15_11
#define init_13_3_15_11
void FASTCALL call_13_3_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_15_12
#define init_13_3_15_12
void FASTCALL call_13_3_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_15_13
#define init_13_3_15_13
void FASTCALL call_13_3_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_15_14
#define init_13_3_15_14
void FASTCALL call_13_3_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_3_15_15
#define init_13_3_15_15
void FASTCALL call_13_3_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_0_0
#define init_13_4_0_0
void FASTCALL call_13_4_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_0_1
#define init_13_4_0_1
void FASTCALL call_13_4_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_0_2
#define init_13_4_0_2
void FASTCALL call_13_4_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_0_3
#define init_13_4_0_3
void FASTCALL call_13_4_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_0_4
#define init_13_4_0_4
void FASTCALL call_13_4_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_0_5
#define init_13_4_0_5
void FASTCALL call_13_4_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_0_6
#define init_13_4_0_6
void FASTCALL call_13_4_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_0_7
#define init_13_4_0_7
void FASTCALL call_13_4_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_0_8
#define init_13_4_0_8
void FASTCALL call_13_4_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_0_9
#define init_13_4_0_9
void FASTCALL call_13_4_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_0_10
#define init_13_4_0_10
void FASTCALL call_13_4_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_0_11
#define init_13_4_0_11
void FASTCALL call_13_4_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_0_12
#define init_13_4_0_12
void FASTCALL call_13_4_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_0_13
#define init_13_4_0_13
void FASTCALL call_13_4_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_0_14
#define init_13_4_0_14
void FASTCALL call_13_4_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_0_15
#define init_13_4_0_15
void FASTCALL call_13_4_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_1_0
#define init_13_4_1_0
void FASTCALL call_13_4_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_1_1
#define init_13_4_1_1
void FASTCALL call_13_4_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_1_2
#define init_13_4_1_2
void FASTCALL call_13_4_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_1_3
#define init_13_4_1_3
void FASTCALL call_13_4_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_1_4
#define init_13_4_1_4
void FASTCALL call_13_4_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_1_5
#define init_13_4_1_5
void FASTCALL call_13_4_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_1_6
#define init_13_4_1_6
void FASTCALL call_13_4_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_1_7
#define init_13_4_1_7
void FASTCALL call_13_4_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_1_8
#define init_13_4_1_8
void FASTCALL call_13_4_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_1_9
#define init_13_4_1_9
void FASTCALL call_13_4_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_1_10
#define init_13_4_1_10
void FASTCALL call_13_4_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_1_11
#define init_13_4_1_11
void FASTCALL call_13_4_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_1_12
#define init_13_4_1_12
void FASTCALL call_13_4_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_1_13
#define init_13_4_1_13
void FASTCALL call_13_4_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_1_14
#define init_13_4_1_14
void FASTCALL call_13_4_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_1_15
#define init_13_4_1_15
void FASTCALL call_13_4_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_2_0
#define init_13_4_2_0
void FASTCALL call_13_4_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_2_1
#define init_13_4_2_1
void FASTCALL call_13_4_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_2_2
#define init_13_4_2_2
void FASTCALL call_13_4_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_2_3
#define init_13_4_2_3
void FASTCALL call_13_4_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_2_4
#define init_13_4_2_4
void FASTCALL call_13_4_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_2_5
#define init_13_4_2_5
void FASTCALL call_13_4_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_2_6
#define init_13_4_2_6
void FASTCALL call_13_4_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_2_7
#define init_13_4_2_7
void FASTCALL call_13_4_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_2_8
#define init_13_4_2_8
void FASTCALL call_13_4_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_2_9
#define init_13_4_2_9
void FASTCALL call_13_4_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_2_10
#define init_13_4_2_10
void FASTCALL call_13_4_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_2_11
#define init_13_4_2_11
void FASTCALL call_13_4_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_2_12
#define init_13_4_2_12
void FASTCALL call_13_4_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_2_13
#define init_13_4_2_13
void FASTCALL call_13_4_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_2_14
#define init_13_4_2_14
void FASTCALL call_13_4_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_2_15
#define init_13_4_2_15
void FASTCALL call_13_4_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_3_0
#define init_13_4_3_0
void FASTCALL call_13_4_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_3_1
#define init_13_4_3_1
void FASTCALL call_13_4_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_3_2
#define init_13_4_3_2
void FASTCALL call_13_4_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_3_3
#define init_13_4_3_3
void FASTCALL call_13_4_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_3_4
#define init_13_4_3_4
void FASTCALL call_13_4_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_3_5
#define init_13_4_3_5
void FASTCALL call_13_4_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_3_6
#define init_13_4_3_6
void FASTCALL call_13_4_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_3_7
#define init_13_4_3_7
void FASTCALL call_13_4_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_3_8
#define init_13_4_3_8
void FASTCALL call_13_4_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_3_9
#define init_13_4_3_9
void FASTCALL call_13_4_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_3_10
#define init_13_4_3_10
void FASTCALL call_13_4_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_3_11
#define init_13_4_3_11
void FASTCALL call_13_4_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_3_12
#define init_13_4_3_12
void FASTCALL call_13_4_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_3_13
#define init_13_4_3_13
void FASTCALL call_13_4_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_3_14
#define init_13_4_3_14
void FASTCALL call_13_4_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_3_15
#define init_13_4_3_15
void FASTCALL call_13_4_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_4_0
#define init_13_4_4_0
void FASTCALL call_13_4_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_4_1
#define init_13_4_4_1
void FASTCALL call_13_4_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_4_2
#define init_13_4_4_2
void FASTCALL call_13_4_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_4_3
#define init_13_4_4_3
void FASTCALL call_13_4_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_4_4
#define init_13_4_4_4
void FASTCALL call_13_4_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_4_5
#define init_13_4_4_5
void FASTCALL call_13_4_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_4_6
#define init_13_4_4_6
void FASTCALL call_13_4_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_4_7
#define init_13_4_4_7
void FASTCALL call_13_4_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_4_8
#define init_13_4_4_8
void FASTCALL call_13_4_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_4_9
#define init_13_4_4_9
void FASTCALL call_13_4_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_4_10
#define init_13_4_4_10
void FASTCALL call_13_4_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_4_11
#define init_13_4_4_11
void FASTCALL call_13_4_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_4_12
#define init_13_4_4_12
void FASTCALL call_13_4_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_4_13
#define init_13_4_4_13
void FASTCALL call_13_4_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_4_14
#define init_13_4_4_14
void FASTCALL call_13_4_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_4_15
#define init_13_4_4_15
void FASTCALL call_13_4_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_5_0
#define init_13_4_5_0
void FASTCALL call_13_4_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_5_1
#define init_13_4_5_1
void FASTCALL call_13_4_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_5_2
#define init_13_4_5_2
void FASTCALL call_13_4_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_5_3
#define init_13_4_5_3
void FASTCALL call_13_4_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_5_4
#define init_13_4_5_4
void FASTCALL call_13_4_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_5_5
#define init_13_4_5_5
void FASTCALL call_13_4_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_5_6
#define init_13_4_5_6
void FASTCALL call_13_4_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_5_7
#define init_13_4_5_7
void FASTCALL call_13_4_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_5_8
#define init_13_4_5_8
void FASTCALL call_13_4_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_5_9
#define init_13_4_5_9
void FASTCALL call_13_4_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_5_10
#define init_13_4_5_10
void FASTCALL call_13_4_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_5_11
#define init_13_4_5_11
void FASTCALL call_13_4_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_5_12
#define init_13_4_5_12
void FASTCALL call_13_4_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_5_13
#define init_13_4_5_13
void FASTCALL call_13_4_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_5_14
#define init_13_4_5_14
void FASTCALL call_13_4_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_5_15
#define init_13_4_5_15
void FASTCALL call_13_4_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_6_0
#define init_13_4_6_0
void FASTCALL call_13_4_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_6_1
#define init_13_4_6_1
void FASTCALL call_13_4_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_6_2
#define init_13_4_6_2
void FASTCALL call_13_4_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_6_3
#define init_13_4_6_3
void FASTCALL call_13_4_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_6_4
#define init_13_4_6_4
void FASTCALL call_13_4_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_6_5
#define init_13_4_6_5
void FASTCALL call_13_4_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_6_6
#define init_13_4_6_6
void FASTCALL call_13_4_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_6_7
#define init_13_4_6_7
void FASTCALL call_13_4_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_6_8
#define init_13_4_6_8
void FASTCALL call_13_4_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_6_9
#define init_13_4_6_9
void FASTCALL call_13_4_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_6_10
#define init_13_4_6_10
void FASTCALL call_13_4_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_6_11
#define init_13_4_6_11
void FASTCALL call_13_4_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_6_12
#define init_13_4_6_12
void FASTCALL call_13_4_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_6_13
#define init_13_4_6_13
void FASTCALL call_13_4_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_6_14
#define init_13_4_6_14
void FASTCALL call_13_4_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_6_15
#define init_13_4_6_15
void FASTCALL call_13_4_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_7_0
#define init_13_4_7_0
void FASTCALL call_13_4_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_7_1
#define init_13_4_7_1
void FASTCALL call_13_4_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_7_2
#define init_13_4_7_2
void FASTCALL call_13_4_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_7_3
#define init_13_4_7_3
void FASTCALL call_13_4_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_7_4
#define init_13_4_7_4
void FASTCALL call_13_4_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_7_5
#define init_13_4_7_5
void FASTCALL call_13_4_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_7_6
#define init_13_4_7_6
void FASTCALL call_13_4_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_7_7
#define init_13_4_7_7
void FASTCALL call_13_4_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_7_8
#define init_13_4_7_8
void FASTCALL call_13_4_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_7_9
#define init_13_4_7_9
void FASTCALL call_13_4_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_7_10
#define init_13_4_7_10
void FASTCALL call_13_4_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_7_11
#define init_13_4_7_11
void FASTCALL call_13_4_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_7_12
#define init_13_4_7_12
void FASTCALL call_13_4_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_7_13
#define init_13_4_7_13
void FASTCALL call_13_4_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_7_14
#define init_13_4_7_14
void FASTCALL call_13_4_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_7_15
#define init_13_4_7_15
void FASTCALL call_13_4_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_8_0
#define init_13_4_8_0
void FASTCALL call_13_4_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_8_1
#define init_13_4_8_1
void FASTCALL call_13_4_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_8_2
#define init_13_4_8_2
void FASTCALL call_13_4_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_8_3
#define init_13_4_8_3
void FASTCALL call_13_4_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_8_4
#define init_13_4_8_4
void FASTCALL call_13_4_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_8_5
#define init_13_4_8_5
void FASTCALL call_13_4_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_8_6
#define init_13_4_8_6
void FASTCALL call_13_4_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_8_7
#define init_13_4_8_7
void FASTCALL call_13_4_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_8_8
#define init_13_4_8_8
void FASTCALL call_13_4_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_8_9
#define init_13_4_8_9
void FASTCALL call_13_4_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_8_10
#define init_13_4_8_10
void FASTCALL call_13_4_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_8_11
#define init_13_4_8_11
void FASTCALL call_13_4_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_8_12
#define init_13_4_8_12
void FASTCALL call_13_4_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_8_13
#define init_13_4_8_13
void FASTCALL call_13_4_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_8_14
#define init_13_4_8_14
void FASTCALL call_13_4_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_8_15
#define init_13_4_8_15
void FASTCALL call_13_4_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_9_0
#define init_13_4_9_0
void FASTCALL call_13_4_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_9_1
#define init_13_4_9_1
void FASTCALL call_13_4_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_9_2
#define init_13_4_9_2
void FASTCALL call_13_4_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_9_3
#define init_13_4_9_3
void FASTCALL call_13_4_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_9_4
#define init_13_4_9_4
void FASTCALL call_13_4_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_9_5
#define init_13_4_9_5
void FASTCALL call_13_4_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_9_6
#define init_13_4_9_6
void FASTCALL call_13_4_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_9_7
#define init_13_4_9_7
void FASTCALL call_13_4_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_9_8
#define init_13_4_9_8
void FASTCALL call_13_4_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_9_9
#define init_13_4_9_9
void FASTCALL call_13_4_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_9_10
#define init_13_4_9_10
void FASTCALL call_13_4_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_9_11
#define init_13_4_9_11
void FASTCALL call_13_4_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_9_12
#define init_13_4_9_12
void FASTCALL call_13_4_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_9_13
#define init_13_4_9_13
void FASTCALL call_13_4_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_9_14
#define init_13_4_9_14
void FASTCALL call_13_4_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_9_15
#define init_13_4_9_15
void FASTCALL call_13_4_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_10_0
#define init_13_4_10_0
void FASTCALL call_13_4_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_10_1
#define init_13_4_10_1
void FASTCALL call_13_4_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_10_2
#define init_13_4_10_2
void FASTCALL call_13_4_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_10_3
#define init_13_4_10_3
void FASTCALL call_13_4_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_10_4
#define init_13_4_10_4
void FASTCALL call_13_4_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_10_5
#define init_13_4_10_5
void FASTCALL call_13_4_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_10_6
#define init_13_4_10_6
void FASTCALL call_13_4_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_10_7
#define init_13_4_10_7
void FASTCALL call_13_4_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_10_8
#define init_13_4_10_8
void FASTCALL call_13_4_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_10_9
#define init_13_4_10_9
void FASTCALL call_13_4_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_10_10
#define init_13_4_10_10
void FASTCALL call_13_4_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_10_11
#define init_13_4_10_11
void FASTCALL call_13_4_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_10_12
#define init_13_4_10_12
void FASTCALL call_13_4_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_10_13
#define init_13_4_10_13
void FASTCALL call_13_4_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_10_14
#define init_13_4_10_14
void FASTCALL call_13_4_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_10_15
#define init_13_4_10_15
void FASTCALL call_13_4_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_11_0
#define init_13_4_11_0
void FASTCALL call_13_4_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_11_1
#define init_13_4_11_1
void FASTCALL call_13_4_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_11_2
#define init_13_4_11_2
void FASTCALL call_13_4_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_11_3
#define init_13_4_11_3
void FASTCALL call_13_4_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_11_4
#define init_13_4_11_4
void FASTCALL call_13_4_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_11_5
#define init_13_4_11_5
void FASTCALL call_13_4_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_11_6
#define init_13_4_11_6
void FASTCALL call_13_4_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_11_7
#define init_13_4_11_7
void FASTCALL call_13_4_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_11_8
#define init_13_4_11_8
void FASTCALL call_13_4_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_11_9
#define init_13_4_11_9
void FASTCALL call_13_4_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_11_10
#define init_13_4_11_10
void FASTCALL call_13_4_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_11_11
#define init_13_4_11_11
void FASTCALL call_13_4_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_11_12
#define init_13_4_11_12
void FASTCALL call_13_4_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_11_13
#define init_13_4_11_13
void FASTCALL call_13_4_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_11_14
#define init_13_4_11_14
void FASTCALL call_13_4_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_11_15
#define init_13_4_11_15
void FASTCALL call_13_4_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_12_0
#define init_13_4_12_0
void FASTCALL call_13_4_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_12_1
#define init_13_4_12_1
void FASTCALL call_13_4_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_12_2
#define init_13_4_12_2
void FASTCALL call_13_4_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_12_3
#define init_13_4_12_3
void FASTCALL call_13_4_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_12_4
#define init_13_4_12_4
void FASTCALL call_13_4_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_12_5
#define init_13_4_12_5
void FASTCALL call_13_4_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_12_6
#define init_13_4_12_6
void FASTCALL call_13_4_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_12_7
#define init_13_4_12_7
void FASTCALL call_13_4_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_12_8
#define init_13_4_12_8
void FASTCALL call_13_4_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_12_9
#define init_13_4_12_9
void FASTCALL call_13_4_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_12_10
#define init_13_4_12_10
void FASTCALL call_13_4_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_12_11
#define init_13_4_12_11
void FASTCALL call_13_4_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_12_12
#define init_13_4_12_12
void FASTCALL call_13_4_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_12_13
#define init_13_4_12_13
void FASTCALL call_13_4_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_12_14
#define init_13_4_12_14
void FASTCALL call_13_4_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_12_15
#define init_13_4_12_15
void FASTCALL call_13_4_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_13_0
#define init_13_4_13_0
void FASTCALL call_13_4_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_13_1
#define init_13_4_13_1
void FASTCALL call_13_4_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_13_2
#define init_13_4_13_2
void FASTCALL call_13_4_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_13_3
#define init_13_4_13_3
void FASTCALL call_13_4_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_13_4
#define init_13_4_13_4
void FASTCALL call_13_4_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_13_5
#define init_13_4_13_5
void FASTCALL call_13_4_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_13_6
#define init_13_4_13_6
void FASTCALL call_13_4_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_13_7
#define init_13_4_13_7
void FASTCALL call_13_4_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_13_8
#define init_13_4_13_8
void FASTCALL call_13_4_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_13_9
#define init_13_4_13_9
void FASTCALL call_13_4_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_13_10
#define init_13_4_13_10
void FASTCALL call_13_4_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_13_11
#define init_13_4_13_11
void FASTCALL call_13_4_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_13_12
#define init_13_4_13_12
void FASTCALL call_13_4_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_13_13
#define init_13_4_13_13
void FASTCALL call_13_4_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_13_14
#define init_13_4_13_14
void FASTCALL call_13_4_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_13_15
#define init_13_4_13_15
void FASTCALL call_13_4_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_14_0
#define init_13_4_14_0
void FASTCALL call_13_4_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_14_1
#define init_13_4_14_1
void FASTCALL call_13_4_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_14_2
#define init_13_4_14_2
void FASTCALL call_13_4_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_14_3
#define init_13_4_14_3
void FASTCALL call_13_4_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_14_4
#define init_13_4_14_4
void FASTCALL call_13_4_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_14_5
#define init_13_4_14_5
void FASTCALL call_13_4_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_14_6
#define init_13_4_14_6
void FASTCALL call_13_4_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_14_7
#define init_13_4_14_7
void FASTCALL call_13_4_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_14_8
#define init_13_4_14_8
void FASTCALL call_13_4_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_14_9
#define init_13_4_14_9
void FASTCALL call_13_4_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_14_10
#define init_13_4_14_10
void FASTCALL call_13_4_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_14_11
#define init_13_4_14_11
void FASTCALL call_13_4_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_14_12
#define init_13_4_14_12
void FASTCALL call_13_4_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_14_13
#define init_13_4_14_13
void FASTCALL call_13_4_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_14_14
#define init_13_4_14_14
void FASTCALL call_13_4_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_14_15
#define init_13_4_14_15
void FASTCALL call_13_4_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_15_0
#define init_13_4_15_0
void FASTCALL call_13_4_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_15_1
#define init_13_4_15_1
void FASTCALL call_13_4_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_15_2
#define init_13_4_15_2
void FASTCALL call_13_4_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_15_3
#define init_13_4_15_3
void FASTCALL call_13_4_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_15_4
#define init_13_4_15_4
void FASTCALL call_13_4_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_15_5
#define init_13_4_15_5
void FASTCALL call_13_4_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_15_6
#define init_13_4_15_6
void FASTCALL call_13_4_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_15_7
#define init_13_4_15_7
void FASTCALL call_13_4_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_15_8
#define init_13_4_15_8
void FASTCALL call_13_4_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_15_9
#define init_13_4_15_9
void FASTCALL call_13_4_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_15_10
#define init_13_4_15_10
void FASTCALL call_13_4_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_15_11
#define init_13_4_15_11
void FASTCALL call_13_4_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_15_12
#define init_13_4_15_12
void FASTCALL call_13_4_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_15_13
#define init_13_4_15_13
void FASTCALL call_13_4_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_15_14
#define init_13_4_15_14
void FASTCALL call_13_4_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_4_15_15
#define init_13_4_15_15
void FASTCALL call_13_4_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_0_0
#define init_13_5_0_0
void FASTCALL call_13_5_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_0_1
#define init_13_5_0_1
void FASTCALL call_13_5_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_0_2
#define init_13_5_0_2
void FASTCALL call_13_5_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_0_3
#define init_13_5_0_3
void FASTCALL call_13_5_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_0_4
#define init_13_5_0_4
void FASTCALL call_13_5_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_0_5
#define init_13_5_0_5
void FASTCALL call_13_5_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_0_6
#define init_13_5_0_6
void FASTCALL call_13_5_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_0_7
#define init_13_5_0_7
void FASTCALL call_13_5_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_0_8
#define init_13_5_0_8
void FASTCALL call_13_5_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_0_9
#define init_13_5_0_9
void FASTCALL call_13_5_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_0_10
#define init_13_5_0_10
void FASTCALL call_13_5_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_0_11
#define init_13_5_0_11
void FASTCALL call_13_5_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_0_12
#define init_13_5_0_12
void FASTCALL call_13_5_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_0_13
#define init_13_5_0_13
void FASTCALL call_13_5_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_0_14
#define init_13_5_0_14
void FASTCALL call_13_5_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_0_15
#define init_13_5_0_15
void FASTCALL call_13_5_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_1_0
#define init_13_5_1_0
void FASTCALL call_13_5_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_1_1
#define init_13_5_1_1
void FASTCALL call_13_5_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_1_2
#define init_13_5_1_2
void FASTCALL call_13_5_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_1_3
#define init_13_5_1_3
void FASTCALL call_13_5_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_1_4
#define init_13_5_1_4
void FASTCALL call_13_5_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_1_5
#define init_13_5_1_5
void FASTCALL call_13_5_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_1_6
#define init_13_5_1_6
void FASTCALL call_13_5_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_1_7
#define init_13_5_1_7
void FASTCALL call_13_5_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_1_8
#define init_13_5_1_8
void FASTCALL call_13_5_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_1_9
#define init_13_5_1_9
void FASTCALL call_13_5_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_1_10
#define init_13_5_1_10
void FASTCALL call_13_5_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_1_11
#define init_13_5_1_11
void FASTCALL call_13_5_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_1_12
#define init_13_5_1_12
void FASTCALL call_13_5_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_1_13
#define init_13_5_1_13
void FASTCALL call_13_5_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_1_14
#define init_13_5_1_14
void FASTCALL call_13_5_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_1_15
#define init_13_5_1_15
void FASTCALL call_13_5_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_2_0
#define init_13_5_2_0
void FASTCALL call_13_5_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_2_1
#define init_13_5_2_1
void FASTCALL call_13_5_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_2_2
#define init_13_5_2_2
void FASTCALL call_13_5_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_2_3
#define init_13_5_2_3
void FASTCALL call_13_5_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_2_4
#define init_13_5_2_4
void FASTCALL call_13_5_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_2_5
#define init_13_5_2_5
void FASTCALL call_13_5_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_2_6
#define init_13_5_2_6
void FASTCALL call_13_5_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_2_7
#define init_13_5_2_7
void FASTCALL call_13_5_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_2_8
#define init_13_5_2_8
void FASTCALL call_13_5_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_2_9
#define init_13_5_2_9
void FASTCALL call_13_5_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_2_10
#define init_13_5_2_10
void FASTCALL call_13_5_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_2_11
#define init_13_5_2_11
void FASTCALL call_13_5_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_2_12
#define init_13_5_2_12
void FASTCALL call_13_5_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_2_13
#define init_13_5_2_13
void FASTCALL call_13_5_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_2_14
#define init_13_5_2_14
void FASTCALL call_13_5_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_2_15
#define init_13_5_2_15
void FASTCALL call_13_5_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_3_0
#define init_13_5_3_0
void FASTCALL call_13_5_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_3_1
#define init_13_5_3_1
void FASTCALL call_13_5_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_3_2
#define init_13_5_3_2
void FASTCALL call_13_5_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_3_3
#define init_13_5_3_3
void FASTCALL call_13_5_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_3_4
#define init_13_5_3_4
void FASTCALL call_13_5_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_3_5
#define init_13_5_3_5
void FASTCALL call_13_5_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_3_6
#define init_13_5_3_6
void FASTCALL call_13_5_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_3_7
#define init_13_5_3_7
void FASTCALL call_13_5_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_3_8
#define init_13_5_3_8
void FASTCALL call_13_5_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_3_9
#define init_13_5_3_9
void FASTCALL call_13_5_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_3_10
#define init_13_5_3_10
void FASTCALL call_13_5_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_3_11
#define init_13_5_3_11
void FASTCALL call_13_5_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_3_12
#define init_13_5_3_12
void FASTCALL call_13_5_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_3_13
#define init_13_5_3_13
void FASTCALL call_13_5_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_3_14
#define init_13_5_3_14
void FASTCALL call_13_5_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_3_15
#define init_13_5_3_15
void FASTCALL call_13_5_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_4_0
#define init_13_5_4_0
void FASTCALL call_13_5_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_4_1
#define init_13_5_4_1
void FASTCALL call_13_5_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_4_2
#define init_13_5_4_2
void FASTCALL call_13_5_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_4_3
#define init_13_5_4_3
void FASTCALL call_13_5_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_4_4
#define init_13_5_4_4
void FASTCALL call_13_5_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_4_5
#define init_13_5_4_5
void FASTCALL call_13_5_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_4_6
#define init_13_5_4_6
void FASTCALL call_13_5_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_4_7
#define init_13_5_4_7
void FASTCALL call_13_5_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_4_8
#define init_13_5_4_8
void FASTCALL call_13_5_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_4_9
#define init_13_5_4_9
void FASTCALL call_13_5_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_4_10
#define init_13_5_4_10
void FASTCALL call_13_5_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_4_11
#define init_13_5_4_11
void FASTCALL call_13_5_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_4_12
#define init_13_5_4_12
void FASTCALL call_13_5_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_4_13
#define init_13_5_4_13
void FASTCALL call_13_5_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_4_14
#define init_13_5_4_14
void FASTCALL call_13_5_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_4_15
#define init_13_5_4_15
void FASTCALL call_13_5_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_5_0
#define init_13_5_5_0
void FASTCALL call_13_5_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_5_1
#define init_13_5_5_1
void FASTCALL call_13_5_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_5_2
#define init_13_5_5_2
void FASTCALL call_13_5_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_5_3
#define init_13_5_5_3
void FASTCALL call_13_5_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_5_4
#define init_13_5_5_4
void FASTCALL call_13_5_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_5_5
#define init_13_5_5_5
void FASTCALL call_13_5_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_5_6
#define init_13_5_5_6
void FASTCALL call_13_5_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_5_7
#define init_13_5_5_7
void FASTCALL call_13_5_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_5_8
#define init_13_5_5_8
void FASTCALL call_13_5_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_5_9
#define init_13_5_5_9
void FASTCALL call_13_5_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_5_10
#define init_13_5_5_10
void FASTCALL call_13_5_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_5_11
#define init_13_5_5_11
void FASTCALL call_13_5_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_5_12
#define init_13_5_5_12
void FASTCALL call_13_5_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_5_13
#define init_13_5_5_13
void FASTCALL call_13_5_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_5_14
#define init_13_5_5_14
void FASTCALL call_13_5_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_5_15
#define init_13_5_5_15
void FASTCALL call_13_5_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_6_0
#define init_13_5_6_0
void FASTCALL call_13_5_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_6_1
#define init_13_5_6_1
void FASTCALL call_13_5_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_6_2
#define init_13_5_6_2
void FASTCALL call_13_5_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_6_3
#define init_13_5_6_3
void FASTCALL call_13_5_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_6_4
#define init_13_5_6_4
void FASTCALL call_13_5_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_6_5
#define init_13_5_6_5
void FASTCALL call_13_5_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_6_6
#define init_13_5_6_6
void FASTCALL call_13_5_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_6_7
#define init_13_5_6_7
void FASTCALL call_13_5_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_6_8
#define init_13_5_6_8
void FASTCALL call_13_5_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_6_9
#define init_13_5_6_9
void FASTCALL call_13_5_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_6_10
#define init_13_5_6_10
void FASTCALL call_13_5_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_6_11
#define init_13_5_6_11
void FASTCALL call_13_5_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_6_12
#define init_13_5_6_12
void FASTCALL call_13_5_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_6_13
#define init_13_5_6_13
void FASTCALL call_13_5_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_6_14
#define init_13_5_6_14
void FASTCALL call_13_5_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_6_15
#define init_13_5_6_15
void FASTCALL call_13_5_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_7_0
#define init_13_5_7_0
void FASTCALL call_13_5_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_7_1
#define init_13_5_7_1
void FASTCALL call_13_5_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_7_2
#define init_13_5_7_2
void FASTCALL call_13_5_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_7_3
#define init_13_5_7_3
void FASTCALL call_13_5_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_7_4
#define init_13_5_7_4
void FASTCALL call_13_5_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_7_5
#define init_13_5_7_5
void FASTCALL call_13_5_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_7_6
#define init_13_5_7_6
void FASTCALL call_13_5_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_7_7
#define init_13_5_7_7
void FASTCALL call_13_5_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_7_8
#define init_13_5_7_8
void FASTCALL call_13_5_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_7_9
#define init_13_5_7_9
void FASTCALL call_13_5_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_7_10
#define init_13_5_7_10
void FASTCALL call_13_5_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_7_11
#define init_13_5_7_11
void FASTCALL call_13_5_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_7_12
#define init_13_5_7_12
void FASTCALL call_13_5_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_7_13
#define init_13_5_7_13
void FASTCALL call_13_5_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_7_14
#define init_13_5_7_14
void FASTCALL call_13_5_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_7_15
#define init_13_5_7_15
void FASTCALL call_13_5_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_8_0
#define init_13_5_8_0
void FASTCALL call_13_5_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_8_1
#define init_13_5_8_1
void FASTCALL call_13_5_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_8_2
#define init_13_5_8_2
void FASTCALL call_13_5_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_8_3
#define init_13_5_8_3
void FASTCALL call_13_5_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_8_4
#define init_13_5_8_4
void FASTCALL call_13_5_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_8_5
#define init_13_5_8_5
void FASTCALL call_13_5_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_8_6
#define init_13_5_8_6
void FASTCALL call_13_5_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_8_7
#define init_13_5_8_7
void FASTCALL call_13_5_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_8_8
#define init_13_5_8_8
void FASTCALL call_13_5_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_8_9
#define init_13_5_8_9
void FASTCALL call_13_5_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_8_10
#define init_13_5_8_10
void FASTCALL call_13_5_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_8_11
#define init_13_5_8_11
void FASTCALL call_13_5_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_8_12
#define init_13_5_8_12
void FASTCALL call_13_5_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_8_13
#define init_13_5_8_13
void FASTCALL call_13_5_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_8_14
#define init_13_5_8_14
void FASTCALL call_13_5_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_8_15
#define init_13_5_8_15
void FASTCALL call_13_5_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_9_0
#define init_13_5_9_0
void FASTCALL call_13_5_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_9_1
#define init_13_5_9_1
void FASTCALL call_13_5_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_9_2
#define init_13_5_9_2
void FASTCALL call_13_5_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_9_3
#define init_13_5_9_3
void FASTCALL call_13_5_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_9_4
#define init_13_5_9_4
void FASTCALL call_13_5_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_9_5
#define init_13_5_9_5
void FASTCALL call_13_5_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_9_6
#define init_13_5_9_6
void FASTCALL call_13_5_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_9_7
#define init_13_5_9_7
void FASTCALL call_13_5_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_9_8
#define init_13_5_9_8
void FASTCALL call_13_5_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_9_9
#define init_13_5_9_9
void FASTCALL call_13_5_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_9_10
#define init_13_5_9_10
void FASTCALL call_13_5_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_9_11
#define init_13_5_9_11
void FASTCALL call_13_5_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_9_12
#define init_13_5_9_12
void FASTCALL call_13_5_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_9_13
#define init_13_5_9_13
void FASTCALL call_13_5_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_9_14
#define init_13_5_9_14
void FASTCALL call_13_5_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_9_15
#define init_13_5_9_15
void FASTCALL call_13_5_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_10_0
#define init_13_5_10_0
void FASTCALL call_13_5_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_10_1
#define init_13_5_10_1
void FASTCALL call_13_5_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_10_2
#define init_13_5_10_2
void FASTCALL call_13_5_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_10_3
#define init_13_5_10_3
void FASTCALL call_13_5_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_10_4
#define init_13_5_10_4
void FASTCALL call_13_5_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_10_5
#define init_13_5_10_5
void FASTCALL call_13_5_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_10_6
#define init_13_5_10_6
void FASTCALL call_13_5_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_10_7
#define init_13_5_10_7
void FASTCALL call_13_5_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_10_8
#define init_13_5_10_8
void FASTCALL call_13_5_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_10_9
#define init_13_5_10_9
void FASTCALL call_13_5_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_10_10
#define init_13_5_10_10
void FASTCALL call_13_5_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_10_11
#define init_13_5_10_11
void FASTCALL call_13_5_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_10_12
#define init_13_5_10_12
void FASTCALL call_13_5_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_10_13
#define init_13_5_10_13
void FASTCALL call_13_5_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_10_14
#define init_13_5_10_14
void FASTCALL call_13_5_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_10_15
#define init_13_5_10_15
void FASTCALL call_13_5_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_11_0
#define init_13_5_11_0
void FASTCALL call_13_5_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_11_1
#define init_13_5_11_1
void FASTCALL call_13_5_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_11_2
#define init_13_5_11_2
void FASTCALL call_13_5_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_11_3
#define init_13_5_11_3
void FASTCALL call_13_5_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_11_4
#define init_13_5_11_4
void FASTCALL call_13_5_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_11_5
#define init_13_5_11_5
void FASTCALL call_13_5_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_11_6
#define init_13_5_11_6
void FASTCALL call_13_5_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_11_7
#define init_13_5_11_7
void FASTCALL call_13_5_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_11_8
#define init_13_5_11_8
void FASTCALL call_13_5_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_11_9
#define init_13_5_11_9
void FASTCALL call_13_5_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_11_10
#define init_13_5_11_10
void FASTCALL call_13_5_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_11_11
#define init_13_5_11_11
void FASTCALL call_13_5_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_11_12
#define init_13_5_11_12
void FASTCALL call_13_5_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_11_13
#define init_13_5_11_13
void FASTCALL call_13_5_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_11_14
#define init_13_5_11_14
void FASTCALL call_13_5_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_11_15
#define init_13_5_11_15
void FASTCALL call_13_5_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_12_0
#define init_13_5_12_0
void FASTCALL call_13_5_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_12_1
#define init_13_5_12_1
void FASTCALL call_13_5_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_12_2
#define init_13_5_12_2
void FASTCALL call_13_5_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_12_3
#define init_13_5_12_3
void FASTCALL call_13_5_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_12_4
#define init_13_5_12_4
void FASTCALL call_13_5_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_12_5
#define init_13_5_12_5
void FASTCALL call_13_5_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_12_6
#define init_13_5_12_6
void FASTCALL call_13_5_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_12_7
#define init_13_5_12_7
void FASTCALL call_13_5_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_12_8
#define init_13_5_12_8
void FASTCALL call_13_5_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_12_9
#define init_13_5_12_9
void FASTCALL call_13_5_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_12_10
#define init_13_5_12_10
void FASTCALL call_13_5_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_12_11
#define init_13_5_12_11
void FASTCALL call_13_5_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_12_12
#define init_13_5_12_12
void FASTCALL call_13_5_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_12_13
#define init_13_5_12_13
void FASTCALL call_13_5_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_12_14
#define init_13_5_12_14
void FASTCALL call_13_5_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_12_15
#define init_13_5_12_15
void FASTCALL call_13_5_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_13_0
#define init_13_5_13_0
void FASTCALL call_13_5_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_13_1
#define init_13_5_13_1
void FASTCALL call_13_5_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_13_2
#define init_13_5_13_2
void FASTCALL call_13_5_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_13_3
#define init_13_5_13_3
void FASTCALL call_13_5_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_13_4
#define init_13_5_13_4
void FASTCALL call_13_5_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_13_5
#define init_13_5_13_5
void FASTCALL call_13_5_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_13_6
#define init_13_5_13_6
void FASTCALL call_13_5_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_13_7
#define init_13_5_13_7
void FASTCALL call_13_5_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_13_8
#define init_13_5_13_8
void FASTCALL call_13_5_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_13_9
#define init_13_5_13_9
void FASTCALL call_13_5_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_13_10
#define init_13_5_13_10
void FASTCALL call_13_5_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_13_11
#define init_13_5_13_11
void FASTCALL call_13_5_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_13_12
#define init_13_5_13_12
void FASTCALL call_13_5_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_13_13
#define init_13_5_13_13
void FASTCALL call_13_5_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_13_14
#define init_13_5_13_14
void FASTCALL call_13_5_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_13_15
#define init_13_5_13_15
void FASTCALL call_13_5_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_14_0
#define init_13_5_14_0
void FASTCALL call_13_5_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_14_1
#define init_13_5_14_1
void FASTCALL call_13_5_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_14_2
#define init_13_5_14_2
void FASTCALL call_13_5_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_14_3
#define init_13_5_14_3
void FASTCALL call_13_5_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_14_4
#define init_13_5_14_4
void FASTCALL call_13_5_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_14_5
#define init_13_5_14_5
void FASTCALL call_13_5_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_14_6
#define init_13_5_14_6
void FASTCALL call_13_5_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_14_7
#define init_13_5_14_7
void FASTCALL call_13_5_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_14_8
#define init_13_5_14_8
void FASTCALL call_13_5_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_14_9
#define init_13_5_14_9
void FASTCALL call_13_5_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_14_10
#define init_13_5_14_10
void FASTCALL call_13_5_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_14_11
#define init_13_5_14_11
void FASTCALL call_13_5_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_14_12
#define init_13_5_14_12
void FASTCALL call_13_5_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_14_13
#define init_13_5_14_13
void FASTCALL call_13_5_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_14_14
#define init_13_5_14_14
void FASTCALL call_13_5_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_14_15
#define init_13_5_14_15
void FASTCALL call_13_5_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_15_0
#define init_13_5_15_0
void FASTCALL call_13_5_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_15_1
#define init_13_5_15_1
void FASTCALL call_13_5_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_15_2
#define init_13_5_15_2
void FASTCALL call_13_5_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_15_3
#define init_13_5_15_3
void FASTCALL call_13_5_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_15_4
#define init_13_5_15_4
void FASTCALL call_13_5_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_15_5
#define init_13_5_15_5
void FASTCALL call_13_5_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_15_6
#define init_13_5_15_6
void FASTCALL call_13_5_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_15_7
#define init_13_5_15_7
void FASTCALL call_13_5_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_15_8
#define init_13_5_15_8
void FASTCALL call_13_5_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_15_9
#define init_13_5_15_9
void FASTCALL call_13_5_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_15_10
#define init_13_5_15_10
void FASTCALL call_13_5_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_15_11
#define init_13_5_15_11
void FASTCALL call_13_5_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_15_12
#define init_13_5_15_12
void FASTCALL call_13_5_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_15_13
#define init_13_5_15_13
void FASTCALL call_13_5_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_15_14
#define init_13_5_15_14
void FASTCALL call_13_5_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_5_15_15
#define init_13_5_15_15
void FASTCALL call_13_5_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_0_0
#define init_13_6_0_0
void FASTCALL call_13_6_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_0_1
#define init_13_6_0_1
void FASTCALL call_13_6_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_0_2
#define init_13_6_0_2
void FASTCALL call_13_6_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_0_3
#define init_13_6_0_3
void FASTCALL call_13_6_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_0_4
#define init_13_6_0_4
void FASTCALL call_13_6_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_0_5
#define init_13_6_0_5
void FASTCALL call_13_6_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_0_6
#define init_13_6_0_6
void FASTCALL call_13_6_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_0_7
#define init_13_6_0_7
void FASTCALL call_13_6_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_0_8
#define init_13_6_0_8
void FASTCALL call_13_6_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_0_9
#define init_13_6_0_9
void FASTCALL call_13_6_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_0_10
#define init_13_6_0_10
void FASTCALL call_13_6_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_0_11
#define init_13_6_0_11
void FASTCALL call_13_6_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_0_12
#define init_13_6_0_12
void FASTCALL call_13_6_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_0_13
#define init_13_6_0_13
void FASTCALL call_13_6_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_0_14
#define init_13_6_0_14
void FASTCALL call_13_6_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_0_15
#define init_13_6_0_15
void FASTCALL call_13_6_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_1_0
#define init_13_6_1_0
void FASTCALL call_13_6_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_1_1
#define init_13_6_1_1
void FASTCALL call_13_6_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_1_2
#define init_13_6_1_2
void FASTCALL call_13_6_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_1_3
#define init_13_6_1_3
void FASTCALL call_13_6_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_1_4
#define init_13_6_1_4
void FASTCALL call_13_6_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_1_5
#define init_13_6_1_5
void FASTCALL call_13_6_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_1_6
#define init_13_6_1_6
void FASTCALL call_13_6_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_1_7
#define init_13_6_1_7
void FASTCALL call_13_6_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_1_8
#define init_13_6_1_8
void FASTCALL call_13_6_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_1_9
#define init_13_6_1_9
void FASTCALL call_13_6_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_1_10
#define init_13_6_1_10
void FASTCALL call_13_6_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_1_11
#define init_13_6_1_11
void FASTCALL call_13_6_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_1_12
#define init_13_6_1_12
void FASTCALL call_13_6_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_1_13
#define init_13_6_1_13
void FASTCALL call_13_6_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_1_14
#define init_13_6_1_14
void FASTCALL call_13_6_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_1_15
#define init_13_6_1_15
void FASTCALL call_13_6_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_2_0
#define init_13_6_2_0
void FASTCALL call_13_6_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_2_1
#define init_13_6_2_1
void FASTCALL call_13_6_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_2_2
#define init_13_6_2_2
void FASTCALL call_13_6_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_2_3
#define init_13_6_2_3
void FASTCALL call_13_6_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_2_4
#define init_13_6_2_4
void FASTCALL call_13_6_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_2_5
#define init_13_6_2_5
void FASTCALL call_13_6_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_2_6
#define init_13_6_2_6
void FASTCALL call_13_6_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_2_7
#define init_13_6_2_7
void FASTCALL call_13_6_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_2_8
#define init_13_6_2_8
void FASTCALL call_13_6_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_2_9
#define init_13_6_2_9
void FASTCALL call_13_6_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_2_10
#define init_13_6_2_10
void FASTCALL call_13_6_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_2_11
#define init_13_6_2_11
void FASTCALL call_13_6_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_2_12
#define init_13_6_2_12
void FASTCALL call_13_6_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_2_13
#define init_13_6_2_13
void FASTCALL call_13_6_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_2_14
#define init_13_6_2_14
void FASTCALL call_13_6_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_2_15
#define init_13_6_2_15
void FASTCALL call_13_6_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_3_0
#define init_13_6_3_0
void FASTCALL call_13_6_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_3_1
#define init_13_6_3_1
void FASTCALL call_13_6_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_3_2
#define init_13_6_3_2
void FASTCALL call_13_6_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_3_3
#define init_13_6_3_3
void FASTCALL call_13_6_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_3_4
#define init_13_6_3_4
void FASTCALL call_13_6_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_3_5
#define init_13_6_3_5
void FASTCALL call_13_6_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_3_6
#define init_13_6_3_6
void FASTCALL call_13_6_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_3_7
#define init_13_6_3_7
void FASTCALL call_13_6_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_3_8
#define init_13_6_3_8
void FASTCALL call_13_6_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_3_9
#define init_13_6_3_9
void FASTCALL call_13_6_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_3_10
#define init_13_6_3_10
void FASTCALL call_13_6_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_3_11
#define init_13_6_3_11
void FASTCALL call_13_6_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_3_12
#define init_13_6_3_12
void FASTCALL call_13_6_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_3_13
#define init_13_6_3_13
void FASTCALL call_13_6_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_3_14
#define init_13_6_3_14
void FASTCALL call_13_6_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_3_15
#define init_13_6_3_15
void FASTCALL call_13_6_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_4_0
#define init_13_6_4_0
void FASTCALL call_13_6_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_4_1
#define init_13_6_4_1
void FASTCALL call_13_6_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_4_2
#define init_13_6_4_2
void FASTCALL call_13_6_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_4_3
#define init_13_6_4_3
void FASTCALL call_13_6_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_4_4
#define init_13_6_4_4
void FASTCALL call_13_6_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_4_5
#define init_13_6_4_5
void FASTCALL call_13_6_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_4_6
#define init_13_6_4_6
void FASTCALL call_13_6_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_4_7
#define init_13_6_4_7
void FASTCALL call_13_6_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_4_8
#define init_13_6_4_8
void FASTCALL call_13_6_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_4_9
#define init_13_6_4_9
void FASTCALL call_13_6_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_4_10
#define init_13_6_4_10
void FASTCALL call_13_6_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_4_11
#define init_13_6_4_11
void FASTCALL call_13_6_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_4_12
#define init_13_6_4_12
void FASTCALL call_13_6_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_4_13
#define init_13_6_4_13
void FASTCALL call_13_6_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_4_14
#define init_13_6_4_14
void FASTCALL call_13_6_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_4_15
#define init_13_6_4_15
void FASTCALL call_13_6_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_5_0
#define init_13_6_5_0
void FASTCALL call_13_6_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_5_1
#define init_13_6_5_1
void FASTCALL call_13_6_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_5_2
#define init_13_6_5_2
void FASTCALL call_13_6_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_5_3
#define init_13_6_5_3
void FASTCALL call_13_6_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_5_4
#define init_13_6_5_4
void FASTCALL call_13_6_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_5_5
#define init_13_6_5_5
void FASTCALL call_13_6_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_5_6
#define init_13_6_5_6
void FASTCALL call_13_6_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_5_7
#define init_13_6_5_7
void FASTCALL call_13_6_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_5_8
#define init_13_6_5_8
void FASTCALL call_13_6_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_5_9
#define init_13_6_5_9
void FASTCALL call_13_6_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_5_10
#define init_13_6_5_10
void FASTCALL call_13_6_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_5_11
#define init_13_6_5_11
void FASTCALL call_13_6_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_5_12
#define init_13_6_5_12
void FASTCALL call_13_6_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_5_13
#define init_13_6_5_13
void FASTCALL call_13_6_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_5_14
#define init_13_6_5_14
void FASTCALL call_13_6_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_5_15
#define init_13_6_5_15
void FASTCALL call_13_6_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_6_0
#define init_13_6_6_0
void FASTCALL call_13_6_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_6_1
#define init_13_6_6_1
void FASTCALL call_13_6_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_6_2
#define init_13_6_6_2
void FASTCALL call_13_6_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_6_3
#define init_13_6_6_3
void FASTCALL call_13_6_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_6_4
#define init_13_6_6_4
void FASTCALL call_13_6_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_6_5
#define init_13_6_6_5
void FASTCALL call_13_6_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_6_6
#define init_13_6_6_6
void FASTCALL call_13_6_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_6_7
#define init_13_6_6_7
void FASTCALL call_13_6_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_6_8
#define init_13_6_6_8
void FASTCALL call_13_6_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_6_9
#define init_13_6_6_9
void FASTCALL call_13_6_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_6_10
#define init_13_6_6_10
void FASTCALL call_13_6_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_6_11
#define init_13_6_6_11
void FASTCALL call_13_6_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_6_12
#define init_13_6_6_12
void FASTCALL call_13_6_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_6_13
#define init_13_6_6_13
void FASTCALL call_13_6_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_6_14
#define init_13_6_6_14
void FASTCALL call_13_6_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_6_15
#define init_13_6_6_15
void FASTCALL call_13_6_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_7_0
#define init_13_6_7_0
void FASTCALL call_13_6_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_7_1
#define init_13_6_7_1
void FASTCALL call_13_6_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_7_2
#define init_13_6_7_2
void FASTCALL call_13_6_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_7_3
#define init_13_6_7_3
void FASTCALL call_13_6_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_7_4
#define init_13_6_7_4
void FASTCALL call_13_6_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_7_5
#define init_13_6_7_5
void FASTCALL call_13_6_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_7_6
#define init_13_6_7_6
void FASTCALL call_13_6_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_7_7
#define init_13_6_7_7
void FASTCALL call_13_6_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_7_8
#define init_13_6_7_8
void FASTCALL call_13_6_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_7_9
#define init_13_6_7_9
void FASTCALL call_13_6_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_7_10
#define init_13_6_7_10
void FASTCALL call_13_6_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_7_11
#define init_13_6_7_11
void FASTCALL call_13_6_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_7_12
#define init_13_6_7_12
void FASTCALL call_13_6_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_7_13
#define init_13_6_7_13
void FASTCALL call_13_6_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_7_14
#define init_13_6_7_14
void FASTCALL call_13_6_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_7_15
#define init_13_6_7_15
void FASTCALL call_13_6_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_8_0
#define init_13_6_8_0
void FASTCALL call_13_6_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_8_1
#define init_13_6_8_1
void FASTCALL call_13_6_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_8_2
#define init_13_6_8_2
void FASTCALL call_13_6_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_8_3
#define init_13_6_8_3
void FASTCALL call_13_6_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_8_4
#define init_13_6_8_4
void FASTCALL call_13_6_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_8_5
#define init_13_6_8_5
void FASTCALL call_13_6_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_8_6
#define init_13_6_8_6
void FASTCALL call_13_6_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_8_7
#define init_13_6_8_7
void FASTCALL call_13_6_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_8_8
#define init_13_6_8_8
void FASTCALL call_13_6_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_8_9
#define init_13_6_8_9
void FASTCALL call_13_6_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_8_10
#define init_13_6_8_10
void FASTCALL call_13_6_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_8_11
#define init_13_6_8_11
void FASTCALL call_13_6_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_8_12
#define init_13_6_8_12
void FASTCALL call_13_6_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_8_13
#define init_13_6_8_13
void FASTCALL call_13_6_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_8_14
#define init_13_6_8_14
void FASTCALL call_13_6_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_8_15
#define init_13_6_8_15
void FASTCALL call_13_6_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_9_0
#define init_13_6_9_0
void FASTCALL call_13_6_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_9_1
#define init_13_6_9_1
void FASTCALL call_13_6_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_9_2
#define init_13_6_9_2
void FASTCALL call_13_6_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_9_3
#define init_13_6_9_3
void FASTCALL call_13_6_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_9_4
#define init_13_6_9_4
void FASTCALL call_13_6_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_9_5
#define init_13_6_9_5
void FASTCALL call_13_6_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_9_6
#define init_13_6_9_6
void FASTCALL call_13_6_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_9_7
#define init_13_6_9_7
void FASTCALL call_13_6_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_9_8
#define init_13_6_9_8
void FASTCALL call_13_6_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_9_9
#define init_13_6_9_9
void FASTCALL call_13_6_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_9_10
#define init_13_6_9_10
void FASTCALL call_13_6_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_9_11
#define init_13_6_9_11
void FASTCALL call_13_6_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_9_12
#define init_13_6_9_12
void FASTCALL call_13_6_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_9_13
#define init_13_6_9_13
void FASTCALL call_13_6_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_9_14
#define init_13_6_9_14
void FASTCALL call_13_6_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_9_15
#define init_13_6_9_15
void FASTCALL call_13_6_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_10_0
#define init_13_6_10_0
void FASTCALL call_13_6_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_10_1
#define init_13_6_10_1
void FASTCALL call_13_6_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_10_2
#define init_13_6_10_2
void FASTCALL call_13_6_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_10_3
#define init_13_6_10_3
void FASTCALL call_13_6_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_10_4
#define init_13_6_10_4
void FASTCALL call_13_6_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_10_5
#define init_13_6_10_5
void FASTCALL call_13_6_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_10_6
#define init_13_6_10_6
void FASTCALL call_13_6_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_10_7
#define init_13_6_10_7
void FASTCALL call_13_6_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_10_8
#define init_13_6_10_8
void FASTCALL call_13_6_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_10_9
#define init_13_6_10_9
void FASTCALL call_13_6_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_10_10
#define init_13_6_10_10
void FASTCALL call_13_6_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_10_11
#define init_13_6_10_11
void FASTCALL call_13_6_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_10_12
#define init_13_6_10_12
void FASTCALL call_13_6_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_10_13
#define init_13_6_10_13
void FASTCALL call_13_6_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_10_14
#define init_13_6_10_14
void FASTCALL call_13_6_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_10_15
#define init_13_6_10_15
void FASTCALL call_13_6_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_11_0
#define init_13_6_11_0
void FASTCALL call_13_6_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_11_1
#define init_13_6_11_1
void FASTCALL call_13_6_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_11_2
#define init_13_6_11_2
void FASTCALL call_13_6_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_11_3
#define init_13_6_11_3
void FASTCALL call_13_6_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_11_4
#define init_13_6_11_4
void FASTCALL call_13_6_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_11_5
#define init_13_6_11_5
void FASTCALL call_13_6_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_11_6
#define init_13_6_11_6
void FASTCALL call_13_6_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_11_7
#define init_13_6_11_7
void FASTCALL call_13_6_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_11_8
#define init_13_6_11_8
void FASTCALL call_13_6_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_11_9
#define init_13_6_11_9
void FASTCALL call_13_6_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_11_10
#define init_13_6_11_10
void FASTCALL call_13_6_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_11_11
#define init_13_6_11_11
void FASTCALL call_13_6_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_11_12
#define init_13_6_11_12
void FASTCALL call_13_6_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_11_13
#define init_13_6_11_13
void FASTCALL call_13_6_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_11_14
#define init_13_6_11_14
void FASTCALL call_13_6_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_11_15
#define init_13_6_11_15
void FASTCALL call_13_6_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_12_0
#define init_13_6_12_0
void FASTCALL call_13_6_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_12_1
#define init_13_6_12_1
void FASTCALL call_13_6_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_12_2
#define init_13_6_12_2
void FASTCALL call_13_6_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_12_3
#define init_13_6_12_3
void FASTCALL call_13_6_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_12_4
#define init_13_6_12_4
void FASTCALL call_13_6_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_12_5
#define init_13_6_12_5
void FASTCALL call_13_6_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_12_6
#define init_13_6_12_6
void FASTCALL call_13_6_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_12_7
#define init_13_6_12_7
void FASTCALL call_13_6_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_12_8
#define init_13_6_12_8
void FASTCALL call_13_6_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_12_9
#define init_13_6_12_9
void FASTCALL call_13_6_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_12_10
#define init_13_6_12_10
void FASTCALL call_13_6_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_12_11
#define init_13_6_12_11
void FASTCALL call_13_6_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_12_12
#define init_13_6_12_12
void FASTCALL call_13_6_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_12_13
#define init_13_6_12_13
void FASTCALL call_13_6_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_12_14
#define init_13_6_12_14
void FASTCALL call_13_6_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_12_15
#define init_13_6_12_15
void FASTCALL call_13_6_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_13_0
#define init_13_6_13_0
void FASTCALL call_13_6_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_13_1
#define init_13_6_13_1
void FASTCALL call_13_6_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_13_2
#define init_13_6_13_2
void FASTCALL call_13_6_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_13_3
#define init_13_6_13_3
void FASTCALL call_13_6_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_13_4
#define init_13_6_13_4
void FASTCALL call_13_6_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_13_5
#define init_13_6_13_5
void FASTCALL call_13_6_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_13_6
#define init_13_6_13_6
void FASTCALL call_13_6_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_13_7
#define init_13_6_13_7
void FASTCALL call_13_6_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_13_8
#define init_13_6_13_8
void FASTCALL call_13_6_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_13_9
#define init_13_6_13_9
void FASTCALL call_13_6_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_13_10
#define init_13_6_13_10
void FASTCALL call_13_6_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_13_11
#define init_13_6_13_11
void FASTCALL call_13_6_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_13_12
#define init_13_6_13_12
void FASTCALL call_13_6_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_13_13
#define init_13_6_13_13
void FASTCALL call_13_6_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_13_14
#define init_13_6_13_14
void FASTCALL call_13_6_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_13_15
#define init_13_6_13_15
void FASTCALL call_13_6_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_14_0
#define init_13_6_14_0
void FASTCALL call_13_6_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_14_1
#define init_13_6_14_1
void FASTCALL call_13_6_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_14_2
#define init_13_6_14_2
void FASTCALL call_13_6_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_14_3
#define init_13_6_14_3
void FASTCALL call_13_6_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_14_4
#define init_13_6_14_4
void FASTCALL call_13_6_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_14_5
#define init_13_6_14_5
void FASTCALL call_13_6_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_14_6
#define init_13_6_14_6
void FASTCALL call_13_6_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_14_7
#define init_13_6_14_7
void FASTCALL call_13_6_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_14_8
#define init_13_6_14_8
void FASTCALL call_13_6_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_14_9
#define init_13_6_14_9
void FASTCALL call_13_6_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_14_10
#define init_13_6_14_10
void FASTCALL call_13_6_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_14_11
#define init_13_6_14_11
void FASTCALL call_13_6_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_14_12
#define init_13_6_14_12
void FASTCALL call_13_6_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_14_13
#define init_13_6_14_13
void FASTCALL call_13_6_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_14_14
#define init_13_6_14_14
void FASTCALL call_13_6_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_14_15
#define init_13_6_14_15
void FASTCALL call_13_6_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_15_0
#define init_13_6_15_0
void FASTCALL call_13_6_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_15_1
#define init_13_6_15_1
void FASTCALL call_13_6_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_15_2
#define init_13_6_15_2
void FASTCALL call_13_6_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_15_3
#define init_13_6_15_3
void FASTCALL call_13_6_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_15_4
#define init_13_6_15_4
void FASTCALL call_13_6_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_15_5
#define init_13_6_15_5
void FASTCALL call_13_6_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_15_6
#define init_13_6_15_6
void FASTCALL call_13_6_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_15_7
#define init_13_6_15_7
void FASTCALL call_13_6_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_15_8
#define init_13_6_15_8
void FASTCALL call_13_6_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_15_9
#define init_13_6_15_9
void FASTCALL call_13_6_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_15_10
#define init_13_6_15_10
void FASTCALL call_13_6_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_15_11
#define init_13_6_15_11
void FASTCALL call_13_6_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_15_12
#define init_13_6_15_12
void FASTCALL call_13_6_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_15_13
#define init_13_6_15_13
void FASTCALL call_13_6_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_15_14
#define init_13_6_15_14
void FASTCALL call_13_6_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_6_15_15
#define init_13_6_15_15
void FASTCALL call_13_6_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_0_0
#define init_13_7_0_0
void FASTCALL call_13_7_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_0_1
#define init_13_7_0_1
void FASTCALL call_13_7_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_0_2
#define init_13_7_0_2
void FASTCALL call_13_7_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_0_3
#define init_13_7_0_3
void FASTCALL call_13_7_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_0_4
#define init_13_7_0_4
void FASTCALL call_13_7_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_0_5
#define init_13_7_0_5
void FASTCALL call_13_7_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_0_6
#define init_13_7_0_6
void FASTCALL call_13_7_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_0_7
#define init_13_7_0_7
void FASTCALL call_13_7_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_0_8
#define init_13_7_0_8
void FASTCALL call_13_7_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_0_9
#define init_13_7_0_9
void FASTCALL call_13_7_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_0_10
#define init_13_7_0_10
void FASTCALL call_13_7_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_0_11
#define init_13_7_0_11
void FASTCALL call_13_7_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_0_12
#define init_13_7_0_12
void FASTCALL call_13_7_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_0_13
#define init_13_7_0_13
void FASTCALL call_13_7_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_0_14
#define init_13_7_0_14
void FASTCALL call_13_7_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_0_15
#define init_13_7_0_15
void FASTCALL call_13_7_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_1_0
#define init_13_7_1_0
void FASTCALL call_13_7_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_1_1
#define init_13_7_1_1
void FASTCALL call_13_7_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_1_2
#define init_13_7_1_2
void FASTCALL call_13_7_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_1_3
#define init_13_7_1_3
void FASTCALL call_13_7_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_1_4
#define init_13_7_1_4
void FASTCALL call_13_7_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_1_5
#define init_13_7_1_5
void FASTCALL call_13_7_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_1_6
#define init_13_7_1_6
void FASTCALL call_13_7_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_1_7
#define init_13_7_1_7
void FASTCALL call_13_7_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_1_8
#define init_13_7_1_8
void FASTCALL call_13_7_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_1_9
#define init_13_7_1_9
void FASTCALL call_13_7_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_1_10
#define init_13_7_1_10
void FASTCALL call_13_7_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_1_11
#define init_13_7_1_11
void FASTCALL call_13_7_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_1_12
#define init_13_7_1_12
void FASTCALL call_13_7_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_1_13
#define init_13_7_1_13
void FASTCALL call_13_7_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_1_14
#define init_13_7_1_14
void FASTCALL call_13_7_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_1_15
#define init_13_7_1_15
void FASTCALL call_13_7_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_2_0
#define init_13_7_2_0
void FASTCALL call_13_7_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_2_1
#define init_13_7_2_1
void FASTCALL call_13_7_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_2_2
#define init_13_7_2_2
void FASTCALL call_13_7_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_2_3
#define init_13_7_2_3
void FASTCALL call_13_7_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_2_4
#define init_13_7_2_4
void FASTCALL call_13_7_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_2_5
#define init_13_7_2_5
void FASTCALL call_13_7_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_2_6
#define init_13_7_2_6
void FASTCALL call_13_7_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_2_7
#define init_13_7_2_7
void FASTCALL call_13_7_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_2_8
#define init_13_7_2_8
void FASTCALL call_13_7_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_2_9
#define init_13_7_2_9
void FASTCALL call_13_7_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_2_10
#define init_13_7_2_10
void FASTCALL call_13_7_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_2_11
#define init_13_7_2_11
void FASTCALL call_13_7_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_2_12
#define init_13_7_2_12
void FASTCALL call_13_7_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_2_13
#define init_13_7_2_13
void FASTCALL call_13_7_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_2_14
#define init_13_7_2_14
void FASTCALL call_13_7_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_2_15
#define init_13_7_2_15
void FASTCALL call_13_7_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_3_0
#define init_13_7_3_0
void FASTCALL call_13_7_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_3_1
#define init_13_7_3_1
void FASTCALL call_13_7_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_3_2
#define init_13_7_3_2
void FASTCALL call_13_7_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_3_3
#define init_13_7_3_3
void FASTCALL call_13_7_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_3_4
#define init_13_7_3_4
void FASTCALL call_13_7_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_3_5
#define init_13_7_3_5
void FASTCALL call_13_7_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_3_6
#define init_13_7_3_6
void FASTCALL call_13_7_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_3_7
#define init_13_7_3_7
void FASTCALL call_13_7_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_3_8
#define init_13_7_3_8
void FASTCALL call_13_7_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_3_9
#define init_13_7_3_9
void FASTCALL call_13_7_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_3_10
#define init_13_7_3_10
void FASTCALL call_13_7_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_3_11
#define init_13_7_3_11
void FASTCALL call_13_7_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_3_12
#define init_13_7_3_12
void FASTCALL call_13_7_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_3_13
#define init_13_7_3_13
void FASTCALL call_13_7_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_3_14
#define init_13_7_3_14
void FASTCALL call_13_7_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_3_15
#define init_13_7_3_15
void FASTCALL call_13_7_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_4_0
#define init_13_7_4_0
void FASTCALL call_13_7_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_4_1
#define init_13_7_4_1
void FASTCALL call_13_7_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_4_2
#define init_13_7_4_2
void FASTCALL call_13_7_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_4_3
#define init_13_7_4_3
void FASTCALL call_13_7_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_4_4
#define init_13_7_4_4
void FASTCALL call_13_7_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_4_5
#define init_13_7_4_5
void FASTCALL call_13_7_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_4_6
#define init_13_7_4_6
void FASTCALL call_13_7_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_4_7
#define init_13_7_4_7
void FASTCALL call_13_7_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_4_8
#define init_13_7_4_8
void FASTCALL call_13_7_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_4_9
#define init_13_7_4_9
void FASTCALL call_13_7_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_4_10
#define init_13_7_4_10
void FASTCALL call_13_7_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_4_11
#define init_13_7_4_11
void FASTCALL call_13_7_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_4_12
#define init_13_7_4_12
void FASTCALL call_13_7_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_4_13
#define init_13_7_4_13
void FASTCALL call_13_7_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_4_14
#define init_13_7_4_14
void FASTCALL call_13_7_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_4_15
#define init_13_7_4_15
void FASTCALL call_13_7_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_5_0
#define init_13_7_5_0
void FASTCALL call_13_7_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_5_1
#define init_13_7_5_1
void FASTCALL call_13_7_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_5_2
#define init_13_7_5_2
void FASTCALL call_13_7_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_5_3
#define init_13_7_5_3
void FASTCALL call_13_7_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_5_4
#define init_13_7_5_4
void FASTCALL call_13_7_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_5_5
#define init_13_7_5_5
void FASTCALL call_13_7_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_5_6
#define init_13_7_5_6
void FASTCALL call_13_7_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_5_7
#define init_13_7_5_7
void FASTCALL call_13_7_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_5_8
#define init_13_7_5_8
void FASTCALL call_13_7_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_5_9
#define init_13_7_5_9
void FASTCALL call_13_7_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_5_10
#define init_13_7_5_10
void FASTCALL call_13_7_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_5_11
#define init_13_7_5_11
void FASTCALL call_13_7_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_5_12
#define init_13_7_5_12
void FASTCALL call_13_7_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_5_13
#define init_13_7_5_13
void FASTCALL call_13_7_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_5_14
#define init_13_7_5_14
void FASTCALL call_13_7_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_5_15
#define init_13_7_5_15
void FASTCALL call_13_7_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_6_0
#define init_13_7_6_0
void FASTCALL call_13_7_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_6_1
#define init_13_7_6_1
void FASTCALL call_13_7_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_6_2
#define init_13_7_6_2
void FASTCALL call_13_7_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_6_3
#define init_13_7_6_3
void FASTCALL call_13_7_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_6_4
#define init_13_7_6_4
void FASTCALL call_13_7_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_6_5
#define init_13_7_6_5
void FASTCALL call_13_7_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_6_6
#define init_13_7_6_6
void FASTCALL call_13_7_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_6_7
#define init_13_7_6_7
void FASTCALL call_13_7_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_6_8
#define init_13_7_6_8
void FASTCALL call_13_7_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_6_9
#define init_13_7_6_9
void FASTCALL call_13_7_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_6_10
#define init_13_7_6_10
void FASTCALL call_13_7_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_6_11
#define init_13_7_6_11
void FASTCALL call_13_7_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_6_12
#define init_13_7_6_12
void FASTCALL call_13_7_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_6_13
#define init_13_7_6_13
void FASTCALL call_13_7_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_6_14
#define init_13_7_6_14
void FASTCALL call_13_7_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_6_15
#define init_13_7_6_15
void FASTCALL call_13_7_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_7_0
#define init_13_7_7_0
void FASTCALL call_13_7_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_7_1
#define init_13_7_7_1
void FASTCALL call_13_7_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_7_2
#define init_13_7_7_2
void FASTCALL call_13_7_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_7_3
#define init_13_7_7_3
void FASTCALL call_13_7_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_7_4
#define init_13_7_7_4
void FASTCALL call_13_7_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_7_5
#define init_13_7_7_5
void FASTCALL call_13_7_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_7_6
#define init_13_7_7_6
void FASTCALL call_13_7_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_7_7
#define init_13_7_7_7
void FASTCALL call_13_7_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_7_8
#define init_13_7_7_8
void FASTCALL call_13_7_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_7_9
#define init_13_7_7_9
void FASTCALL call_13_7_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_7_10
#define init_13_7_7_10
void FASTCALL call_13_7_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_7_11
#define init_13_7_7_11
void FASTCALL call_13_7_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_7_12
#define init_13_7_7_12
void FASTCALL call_13_7_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_7_13
#define init_13_7_7_13
void FASTCALL call_13_7_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_7_14
#define init_13_7_7_14
void FASTCALL call_13_7_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_7_15
#define init_13_7_7_15
void FASTCALL call_13_7_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_8_0
#define init_13_7_8_0
void FASTCALL call_13_7_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_8_1
#define init_13_7_8_1
void FASTCALL call_13_7_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_8_2
#define init_13_7_8_2
void FASTCALL call_13_7_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_8_3
#define init_13_7_8_3
void FASTCALL call_13_7_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_8_4
#define init_13_7_8_4
void FASTCALL call_13_7_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_8_5
#define init_13_7_8_5
void FASTCALL call_13_7_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_8_6
#define init_13_7_8_6
void FASTCALL call_13_7_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_8_7
#define init_13_7_8_7
void FASTCALL call_13_7_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_8_8
#define init_13_7_8_8
void FASTCALL call_13_7_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_8_9
#define init_13_7_8_9
void FASTCALL call_13_7_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_8_10
#define init_13_7_8_10
void FASTCALL call_13_7_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_8_11
#define init_13_7_8_11
void FASTCALL call_13_7_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_8_12
#define init_13_7_8_12
void FASTCALL call_13_7_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_8_13
#define init_13_7_8_13
void FASTCALL call_13_7_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_8_14
#define init_13_7_8_14
void FASTCALL call_13_7_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_8_15
#define init_13_7_8_15
void FASTCALL call_13_7_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_9_0
#define init_13_7_9_0
void FASTCALL call_13_7_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_9_1
#define init_13_7_9_1
void FASTCALL call_13_7_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_9_2
#define init_13_7_9_2
void FASTCALL call_13_7_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_9_3
#define init_13_7_9_3
void FASTCALL call_13_7_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_9_4
#define init_13_7_9_4
void FASTCALL call_13_7_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_9_5
#define init_13_7_9_5
void FASTCALL call_13_7_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_9_6
#define init_13_7_9_6
void FASTCALL call_13_7_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_9_7
#define init_13_7_9_7
void FASTCALL call_13_7_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_9_8
#define init_13_7_9_8
void FASTCALL call_13_7_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_9_9
#define init_13_7_9_9
void FASTCALL call_13_7_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_9_10
#define init_13_7_9_10
void FASTCALL call_13_7_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_9_11
#define init_13_7_9_11
void FASTCALL call_13_7_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_9_12
#define init_13_7_9_12
void FASTCALL call_13_7_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_9_13
#define init_13_7_9_13
void FASTCALL call_13_7_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_9_14
#define init_13_7_9_14
void FASTCALL call_13_7_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_9_15
#define init_13_7_9_15
void FASTCALL call_13_7_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_10_0
#define init_13_7_10_0
void FASTCALL call_13_7_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_10_1
#define init_13_7_10_1
void FASTCALL call_13_7_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_10_2
#define init_13_7_10_2
void FASTCALL call_13_7_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_10_3
#define init_13_7_10_3
void FASTCALL call_13_7_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_10_4
#define init_13_7_10_4
void FASTCALL call_13_7_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_10_5
#define init_13_7_10_5
void FASTCALL call_13_7_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_10_6
#define init_13_7_10_6
void FASTCALL call_13_7_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_10_7
#define init_13_7_10_7
void FASTCALL call_13_7_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_10_8
#define init_13_7_10_8
void FASTCALL call_13_7_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_10_9
#define init_13_7_10_9
void FASTCALL call_13_7_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_10_10
#define init_13_7_10_10
void FASTCALL call_13_7_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_10_11
#define init_13_7_10_11
void FASTCALL call_13_7_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_10_12
#define init_13_7_10_12
void FASTCALL call_13_7_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_10_13
#define init_13_7_10_13
void FASTCALL call_13_7_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_10_14
#define init_13_7_10_14
void FASTCALL call_13_7_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_10_15
#define init_13_7_10_15
void FASTCALL call_13_7_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_11_0
#define init_13_7_11_0
void FASTCALL call_13_7_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_11_1
#define init_13_7_11_1
void FASTCALL call_13_7_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_11_2
#define init_13_7_11_2
void FASTCALL call_13_7_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_11_3
#define init_13_7_11_3
void FASTCALL call_13_7_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_11_4
#define init_13_7_11_4
void FASTCALL call_13_7_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_11_5
#define init_13_7_11_5
void FASTCALL call_13_7_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_11_6
#define init_13_7_11_6
void FASTCALL call_13_7_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_11_7
#define init_13_7_11_7
void FASTCALL call_13_7_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_11_8
#define init_13_7_11_8
void FASTCALL call_13_7_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_11_9
#define init_13_7_11_9
void FASTCALL call_13_7_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_11_10
#define init_13_7_11_10
void FASTCALL call_13_7_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_11_11
#define init_13_7_11_11
void FASTCALL call_13_7_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_11_12
#define init_13_7_11_12
void FASTCALL call_13_7_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_11_13
#define init_13_7_11_13
void FASTCALL call_13_7_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_11_14
#define init_13_7_11_14
void FASTCALL call_13_7_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_11_15
#define init_13_7_11_15
void FASTCALL call_13_7_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_12_0
#define init_13_7_12_0
void FASTCALL call_13_7_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_12_1
#define init_13_7_12_1
void FASTCALL call_13_7_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_12_2
#define init_13_7_12_2
void FASTCALL call_13_7_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_12_3
#define init_13_7_12_3
void FASTCALL call_13_7_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_12_4
#define init_13_7_12_4
void FASTCALL call_13_7_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_12_5
#define init_13_7_12_5
void FASTCALL call_13_7_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_12_6
#define init_13_7_12_6
void FASTCALL call_13_7_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_12_7
#define init_13_7_12_7
void FASTCALL call_13_7_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_12_8
#define init_13_7_12_8
void FASTCALL call_13_7_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_12_9
#define init_13_7_12_9
void FASTCALL call_13_7_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_12_10
#define init_13_7_12_10
void FASTCALL call_13_7_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_12_11
#define init_13_7_12_11
void FASTCALL call_13_7_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_12_12
#define init_13_7_12_12
void FASTCALL call_13_7_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_12_13
#define init_13_7_12_13
void FASTCALL call_13_7_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_12_14
#define init_13_7_12_14
void FASTCALL call_13_7_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_12_15
#define init_13_7_12_15
void FASTCALL call_13_7_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_13_0
#define init_13_7_13_0
void FASTCALL call_13_7_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_13_1
#define init_13_7_13_1
void FASTCALL call_13_7_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_13_2
#define init_13_7_13_2
void FASTCALL call_13_7_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_13_3
#define init_13_7_13_3
void FASTCALL call_13_7_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_13_4
#define init_13_7_13_4
void FASTCALL call_13_7_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_13_5
#define init_13_7_13_5
void FASTCALL call_13_7_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_13_6
#define init_13_7_13_6
void FASTCALL call_13_7_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_13_7
#define init_13_7_13_7
void FASTCALL call_13_7_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_13_8
#define init_13_7_13_8
void FASTCALL call_13_7_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_13_9
#define init_13_7_13_9
void FASTCALL call_13_7_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_13_10
#define init_13_7_13_10
void FASTCALL call_13_7_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_13_11
#define init_13_7_13_11
void FASTCALL call_13_7_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_13_12
#define init_13_7_13_12
void FASTCALL call_13_7_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_13_13
#define init_13_7_13_13
void FASTCALL call_13_7_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_13_14
#define init_13_7_13_14
void FASTCALL call_13_7_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_13_15
#define init_13_7_13_15
void FASTCALL call_13_7_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_14_0
#define init_13_7_14_0
void FASTCALL call_13_7_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_14_1
#define init_13_7_14_1
void FASTCALL call_13_7_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_14_2
#define init_13_7_14_2
void FASTCALL call_13_7_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_14_3
#define init_13_7_14_3
void FASTCALL call_13_7_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_14_4
#define init_13_7_14_4
void FASTCALL call_13_7_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_14_5
#define init_13_7_14_5
void FASTCALL call_13_7_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_14_6
#define init_13_7_14_6
void FASTCALL call_13_7_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_14_7
#define init_13_7_14_7
void FASTCALL call_13_7_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_14_8
#define init_13_7_14_8
void FASTCALL call_13_7_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_14_9
#define init_13_7_14_9
void FASTCALL call_13_7_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_14_10
#define init_13_7_14_10
void FASTCALL call_13_7_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_14_11
#define init_13_7_14_11
void FASTCALL call_13_7_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_14_12
#define init_13_7_14_12
void FASTCALL call_13_7_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_14_13
#define init_13_7_14_13
void FASTCALL call_13_7_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_14_14
#define init_13_7_14_14
void FASTCALL call_13_7_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_14_15
#define init_13_7_14_15
void FASTCALL call_13_7_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_15_0
#define init_13_7_15_0
void FASTCALL call_13_7_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_15_1
#define init_13_7_15_1
void FASTCALL call_13_7_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_15_2
#define init_13_7_15_2
void FASTCALL call_13_7_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_15_3
#define init_13_7_15_3
void FASTCALL call_13_7_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_15_4
#define init_13_7_15_4
void FASTCALL call_13_7_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_15_5
#define init_13_7_15_5
void FASTCALL call_13_7_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_15_6
#define init_13_7_15_6
void FASTCALL call_13_7_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_15_7
#define init_13_7_15_7
void FASTCALL call_13_7_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_15_8
#define init_13_7_15_8
void FASTCALL call_13_7_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_15_9
#define init_13_7_15_9
void FASTCALL call_13_7_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_15_10
#define init_13_7_15_10
void FASTCALL call_13_7_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_15_11
#define init_13_7_15_11
void FASTCALL call_13_7_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_15_12
#define init_13_7_15_12
void FASTCALL call_13_7_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_15_13
#define init_13_7_15_13
void FASTCALL call_13_7_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_15_14
#define init_13_7_15_14
void FASTCALL call_13_7_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_7_15_15
#define init_13_7_15_15
void FASTCALL call_13_7_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_0_0
#define init_13_8_0_0
void FASTCALL call_13_8_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_0_1
#define init_13_8_0_1
void FASTCALL call_13_8_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_0_2
#define init_13_8_0_2
void FASTCALL call_13_8_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_0_3
#define init_13_8_0_3
void FASTCALL call_13_8_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_0_4
#define init_13_8_0_4
void FASTCALL call_13_8_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_0_5
#define init_13_8_0_5
void FASTCALL call_13_8_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_0_6
#define init_13_8_0_6
void FASTCALL call_13_8_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_0_7
#define init_13_8_0_7
void FASTCALL call_13_8_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_0_8
#define init_13_8_0_8
void FASTCALL call_13_8_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_0_9
#define init_13_8_0_9
void FASTCALL call_13_8_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_0_10
#define init_13_8_0_10
void FASTCALL call_13_8_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_0_11
#define init_13_8_0_11
void FASTCALL call_13_8_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_0_12
#define init_13_8_0_12
void FASTCALL call_13_8_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_0_13
#define init_13_8_0_13
void FASTCALL call_13_8_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_0_14
#define init_13_8_0_14
void FASTCALL call_13_8_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_0_15
#define init_13_8_0_15
void FASTCALL call_13_8_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_1_0
#define init_13_8_1_0
void FASTCALL call_13_8_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_1_1
#define init_13_8_1_1
void FASTCALL call_13_8_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_1_2
#define init_13_8_1_2
void FASTCALL call_13_8_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_1_3
#define init_13_8_1_3
void FASTCALL call_13_8_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_1_4
#define init_13_8_1_4
void FASTCALL call_13_8_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_1_5
#define init_13_8_1_5
void FASTCALL call_13_8_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_1_6
#define init_13_8_1_6
void FASTCALL call_13_8_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_1_7
#define init_13_8_1_7
void FASTCALL call_13_8_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_1_8
#define init_13_8_1_8
void FASTCALL call_13_8_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_1_9
#define init_13_8_1_9
void FASTCALL call_13_8_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_1_10
#define init_13_8_1_10
void FASTCALL call_13_8_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_1_11
#define init_13_8_1_11
void FASTCALL call_13_8_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_1_12
#define init_13_8_1_12
void FASTCALL call_13_8_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_1_13
#define init_13_8_1_13
void FASTCALL call_13_8_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_1_14
#define init_13_8_1_14
void FASTCALL call_13_8_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_1_15
#define init_13_8_1_15
void FASTCALL call_13_8_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_2_0
#define init_13_8_2_0
void FASTCALL call_13_8_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_2_1
#define init_13_8_2_1
void FASTCALL call_13_8_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_2_2
#define init_13_8_2_2
void FASTCALL call_13_8_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_2_3
#define init_13_8_2_3
void FASTCALL call_13_8_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_2_4
#define init_13_8_2_4
void FASTCALL call_13_8_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_2_5
#define init_13_8_2_5
void FASTCALL call_13_8_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_2_6
#define init_13_8_2_6
void FASTCALL call_13_8_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_2_7
#define init_13_8_2_7
void FASTCALL call_13_8_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_2_8
#define init_13_8_2_8
void FASTCALL call_13_8_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_2_9
#define init_13_8_2_9
void FASTCALL call_13_8_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_2_10
#define init_13_8_2_10
void FASTCALL call_13_8_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_2_11
#define init_13_8_2_11
void FASTCALL call_13_8_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_2_12
#define init_13_8_2_12
void FASTCALL call_13_8_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_2_13
#define init_13_8_2_13
void FASTCALL call_13_8_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_2_14
#define init_13_8_2_14
void FASTCALL call_13_8_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_2_15
#define init_13_8_2_15
void FASTCALL call_13_8_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_3_0
#define init_13_8_3_0
void FASTCALL call_13_8_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_3_1
#define init_13_8_3_1
void FASTCALL call_13_8_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_3_2
#define init_13_8_3_2
void FASTCALL call_13_8_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_3_3
#define init_13_8_3_3
void FASTCALL call_13_8_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_3_4
#define init_13_8_3_4
void FASTCALL call_13_8_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_3_5
#define init_13_8_3_5
void FASTCALL call_13_8_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_3_6
#define init_13_8_3_6
void FASTCALL call_13_8_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_3_7
#define init_13_8_3_7
void FASTCALL call_13_8_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_3_8
#define init_13_8_3_8
void FASTCALL call_13_8_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_3_9
#define init_13_8_3_9
void FASTCALL call_13_8_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_3_10
#define init_13_8_3_10
void FASTCALL call_13_8_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_3_11
#define init_13_8_3_11
void FASTCALL call_13_8_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_3_12
#define init_13_8_3_12
void FASTCALL call_13_8_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_3_13
#define init_13_8_3_13
void FASTCALL call_13_8_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_3_14
#define init_13_8_3_14
void FASTCALL call_13_8_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_3_15
#define init_13_8_3_15
void FASTCALL call_13_8_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_4_0
#define init_13_8_4_0
void FASTCALL call_13_8_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_4_1
#define init_13_8_4_1
void FASTCALL call_13_8_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_4_2
#define init_13_8_4_2
void FASTCALL call_13_8_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_4_3
#define init_13_8_4_3
void FASTCALL call_13_8_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_4_4
#define init_13_8_4_4
void FASTCALL call_13_8_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_4_5
#define init_13_8_4_5
void FASTCALL call_13_8_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_4_6
#define init_13_8_4_6
void FASTCALL call_13_8_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_4_7
#define init_13_8_4_7
void FASTCALL call_13_8_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_4_8
#define init_13_8_4_8
void FASTCALL call_13_8_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_4_9
#define init_13_8_4_9
void FASTCALL call_13_8_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_4_10
#define init_13_8_4_10
void FASTCALL call_13_8_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_4_11
#define init_13_8_4_11
void FASTCALL call_13_8_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_4_12
#define init_13_8_4_12
void FASTCALL call_13_8_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_4_13
#define init_13_8_4_13
void FASTCALL call_13_8_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_4_14
#define init_13_8_4_14
void FASTCALL call_13_8_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_4_15
#define init_13_8_4_15
void FASTCALL call_13_8_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_5_0
#define init_13_8_5_0
void FASTCALL call_13_8_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_5_1
#define init_13_8_5_1
void FASTCALL call_13_8_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_5_2
#define init_13_8_5_2
void FASTCALL call_13_8_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_5_3
#define init_13_8_5_3
void FASTCALL call_13_8_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_5_4
#define init_13_8_5_4
void FASTCALL call_13_8_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_5_5
#define init_13_8_5_5
void FASTCALL call_13_8_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_5_6
#define init_13_8_5_6
void FASTCALL call_13_8_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_5_7
#define init_13_8_5_7
void FASTCALL call_13_8_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_5_8
#define init_13_8_5_8
void FASTCALL call_13_8_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_5_9
#define init_13_8_5_9
void FASTCALL call_13_8_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_5_10
#define init_13_8_5_10
void FASTCALL call_13_8_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_5_11
#define init_13_8_5_11
void FASTCALL call_13_8_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_5_12
#define init_13_8_5_12
void FASTCALL call_13_8_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_5_13
#define init_13_8_5_13
void FASTCALL call_13_8_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_5_14
#define init_13_8_5_14
void FASTCALL call_13_8_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_5_15
#define init_13_8_5_15
void FASTCALL call_13_8_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_6_0
#define init_13_8_6_0
void FASTCALL call_13_8_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_6_1
#define init_13_8_6_1
void FASTCALL call_13_8_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_6_2
#define init_13_8_6_2
void FASTCALL call_13_8_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_6_3
#define init_13_8_6_3
void FASTCALL call_13_8_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_6_4
#define init_13_8_6_4
void FASTCALL call_13_8_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_6_5
#define init_13_8_6_5
void FASTCALL call_13_8_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_6_6
#define init_13_8_6_6
void FASTCALL call_13_8_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_6_7
#define init_13_8_6_7
void FASTCALL call_13_8_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_6_8
#define init_13_8_6_8
void FASTCALL call_13_8_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_6_9
#define init_13_8_6_9
void FASTCALL call_13_8_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_6_10
#define init_13_8_6_10
void FASTCALL call_13_8_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_6_11
#define init_13_8_6_11
void FASTCALL call_13_8_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_6_12
#define init_13_8_6_12
void FASTCALL call_13_8_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_6_13
#define init_13_8_6_13
void FASTCALL call_13_8_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_6_14
#define init_13_8_6_14
void FASTCALL call_13_8_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_6_15
#define init_13_8_6_15
void FASTCALL call_13_8_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_7_0
#define init_13_8_7_0
void FASTCALL call_13_8_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_7_1
#define init_13_8_7_1
void FASTCALL call_13_8_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_7_2
#define init_13_8_7_2
void FASTCALL call_13_8_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_7_3
#define init_13_8_7_3
void FASTCALL call_13_8_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_7_4
#define init_13_8_7_4
void FASTCALL call_13_8_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_7_5
#define init_13_8_7_5
void FASTCALL call_13_8_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_7_6
#define init_13_8_7_6
void FASTCALL call_13_8_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_7_7
#define init_13_8_7_7
void FASTCALL call_13_8_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_7_8
#define init_13_8_7_8
void FASTCALL call_13_8_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_7_9
#define init_13_8_7_9
void FASTCALL call_13_8_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_7_10
#define init_13_8_7_10
void FASTCALL call_13_8_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_7_11
#define init_13_8_7_11
void FASTCALL call_13_8_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_7_12
#define init_13_8_7_12
void FASTCALL call_13_8_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_7_13
#define init_13_8_7_13
void FASTCALL call_13_8_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_7_14
#define init_13_8_7_14
void FASTCALL call_13_8_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_7_15
#define init_13_8_7_15
void FASTCALL call_13_8_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_8_0
#define init_13_8_8_0
void FASTCALL call_13_8_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_8_1
#define init_13_8_8_1
void FASTCALL call_13_8_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_8_2
#define init_13_8_8_2
void FASTCALL call_13_8_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_8_3
#define init_13_8_8_3
void FASTCALL call_13_8_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_8_4
#define init_13_8_8_4
void FASTCALL call_13_8_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_8_5
#define init_13_8_8_5
void FASTCALL call_13_8_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_8_6
#define init_13_8_8_6
void FASTCALL call_13_8_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_8_7
#define init_13_8_8_7
void FASTCALL call_13_8_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_8_8
#define init_13_8_8_8
void FASTCALL call_13_8_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_8_9
#define init_13_8_8_9
void FASTCALL call_13_8_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_8_10
#define init_13_8_8_10
void FASTCALL call_13_8_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_8_11
#define init_13_8_8_11
void FASTCALL call_13_8_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_8_12
#define init_13_8_8_12
void FASTCALL call_13_8_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_8_13
#define init_13_8_8_13
void FASTCALL call_13_8_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_8_14
#define init_13_8_8_14
void FASTCALL call_13_8_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_8_15
#define init_13_8_8_15
void FASTCALL call_13_8_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_9_0
#define init_13_8_9_0
void FASTCALL call_13_8_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_9_1
#define init_13_8_9_1
void FASTCALL call_13_8_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_9_2
#define init_13_8_9_2
void FASTCALL call_13_8_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_9_3
#define init_13_8_9_3
void FASTCALL call_13_8_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_9_4
#define init_13_8_9_4
void FASTCALL call_13_8_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_9_5
#define init_13_8_9_5
void FASTCALL call_13_8_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_9_6
#define init_13_8_9_6
void FASTCALL call_13_8_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_9_7
#define init_13_8_9_7
void FASTCALL call_13_8_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_9_8
#define init_13_8_9_8
void FASTCALL call_13_8_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_9_9
#define init_13_8_9_9
void FASTCALL call_13_8_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_9_10
#define init_13_8_9_10
void FASTCALL call_13_8_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_9_11
#define init_13_8_9_11
void FASTCALL call_13_8_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_9_12
#define init_13_8_9_12
void FASTCALL call_13_8_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_9_13
#define init_13_8_9_13
void FASTCALL call_13_8_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_9_14
#define init_13_8_9_14
void FASTCALL call_13_8_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_9_15
#define init_13_8_9_15
void FASTCALL call_13_8_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_10_0
#define init_13_8_10_0
void FASTCALL call_13_8_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_10_1
#define init_13_8_10_1
void FASTCALL call_13_8_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_10_2
#define init_13_8_10_2
void FASTCALL call_13_8_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_10_3
#define init_13_8_10_3
void FASTCALL call_13_8_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_10_4
#define init_13_8_10_4
void FASTCALL call_13_8_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_10_5
#define init_13_8_10_5
void FASTCALL call_13_8_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_10_6
#define init_13_8_10_6
void FASTCALL call_13_8_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_10_7
#define init_13_8_10_7
void FASTCALL call_13_8_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_10_8
#define init_13_8_10_8
void FASTCALL call_13_8_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_10_9
#define init_13_8_10_9
void FASTCALL call_13_8_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_10_10
#define init_13_8_10_10
void FASTCALL call_13_8_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_10_11
#define init_13_8_10_11
void FASTCALL call_13_8_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_10_12
#define init_13_8_10_12
void FASTCALL call_13_8_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_10_13
#define init_13_8_10_13
void FASTCALL call_13_8_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_10_14
#define init_13_8_10_14
void FASTCALL call_13_8_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_10_15
#define init_13_8_10_15
void FASTCALL call_13_8_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_11_0
#define init_13_8_11_0
void FASTCALL call_13_8_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_11_1
#define init_13_8_11_1
void FASTCALL call_13_8_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_11_2
#define init_13_8_11_2
void FASTCALL call_13_8_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_11_3
#define init_13_8_11_3
void FASTCALL call_13_8_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_11_4
#define init_13_8_11_4
void FASTCALL call_13_8_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_11_5
#define init_13_8_11_5
void FASTCALL call_13_8_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_11_6
#define init_13_8_11_6
void FASTCALL call_13_8_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_11_7
#define init_13_8_11_7
void FASTCALL call_13_8_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_11_8
#define init_13_8_11_8
void FASTCALL call_13_8_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_11_9
#define init_13_8_11_9
void FASTCALL call_13_8_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_11_10
#define init_13_8_11_10
void FASTCALL call_13_8_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_11_11
#define init_13_8_11_11
void FASTCALL call_13_8_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_11_12
#define init_13_8_11_12
void FASTCALL call_13_8_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_11_13
#define init_13_8_11_13
void FASTCALL call_13_8_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_11_14
#define init_13_8_11_14
void FASTCALL call_13_8_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_11_15
#define init_13_8_11_15
void FASTCALL call_13_8_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_12_0
#define init_13_8_12_0
void FASTCALL call_13_8_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_12_1
#define init_13_8_12_1
void FASTCALL call_13_8_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_12_2
#define init_13_8_12_2
void FASTCALL call_13_8_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_12_3
#define init_13_8_12_3
void FASTCALL call_13_8_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_12_4
#define init_13_8_12_4
void FASTCALL call_13_8_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_12_5
#define init_13_8_12_5
void FASTCALL call_13_8_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_12_6
#define init_13_8_12_6
void FASTCALL call_13_8_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_12_7
#define init_13_8_12_7
void FASTCALL call_13_8_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_12_8
#define init_13_8_12_8
void FASTCALL call_13_8_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_12_9
#define init_13_8_12_9
void FASTCALL call_13_8_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_12_10
#define init_13_8_12_10
void FASTCALL call_13_8_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_12_11
#define init_13_8_12_11
void FASTCALL call_13_8_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_12_12
#define init_13_8_12_12
void FASTCALL call_13_8_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_12_13
#define init_13_8_12_13
void FASTCALL call_13_8_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_12_14
#define init_13_8_12_14
void FASTCALL call_13_8_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_12_15
#define init_13_8_12_15
void FASTCALL call_13_8_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_13_0
#define init_13_8_13_0
void FASTCALL call_13_8_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_13_1
#define init_13_8_13_1
void FASTCALL call_13_8_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_13_2
#define init_13_8_13_2
void FASTCALL call_13_8_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_13_3
#define init_13_8_13_3
void FASTCALL call_13_8_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_13_4
#define init_13_8_13_4
void FASTCALL call_13_8_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_13_5
#define init_13_8_13_5
void FASTCALL call_13_8_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_13_6
#define init_13_8_13_6
void FASTCALL call_13_8_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_13_7
#define init_13_8_13_7
void FASTCALL call_13_8_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_13_8
#define init_13_8_13_8
void FASTCALL call_13_8_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_13_9
#define init_13_8_13_9
void FASTCALL call_13_8_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_13_10
#define init_13_8_13_10
void FASTCALL call_13_8_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_13_11
#define init_13_8_13_11
void FASTCALL call_13_8_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_13_12
#define init_13_8_13_12
void FASTCALL call_13_8_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_13_13
#define init_13_8_13_13
void FASTCALL call_13_8_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_13_14
#define init_13_8_13_14
void FASTCALL call_13_8_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_13_15
#define init_13_8_13_15
void FASTCALL call_13_8_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_14_0
#define init_13_8_14_0
void FASTCALL call_13_8_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_14_1
#define init_13_8_14_1
void FASTCALL call_13_8_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_14_2
#define init_13_8_14_2
void FASTCALL call_13_8_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_14_3
#define init_13_8_14_3
void FASTCALL call_13_8_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_14_4
#define init_13_8_14_4
void FASTCALL call_13_8_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_14_5
#define init_13_8_14_5
void FASTCALL call_13_8_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_14_6
#define init_13_8_14_6
void FASTCALL call_13_8_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_14_7
#define init_13_8_14_7
void FASTCALL call_13_8_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_14_8
#define init_13_8_14_8
void FASTCALL call_13_8_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_14_9
#define init_13_8_14_9
void FASTCALL call_13_8_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_14_10
#define init_13_8_14_10
void FASTCALL call_13_8_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_14_11
#define init_13_8_14_11
void FASTCALL call_13_8_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_14_12
#define init_13_8_14_12
void FASTCALL call_13_8_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_14_13
#define init_13_8_14_13
void FASTCALL call_13_8_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_14_14
#define init_13_8_14_14
void FASTCALL call_13_8_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_14_15
#define init_13_8_14_15
void FASTCALL call_13_8_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_15_0
#define init_13_8_15_0
void FASTCALL call_13_8_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_15_1
#define init_13_8_15_1
void FASTCALL call_13_8_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_15_2
#define init_13_8_15_2
void FASTCALL call_13_8_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_15_3
#define init_13_8_15_3
void FASTCALL call_13_8_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_15_4
#define init_13_8_15_4
void FASTCALL call_13_8_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_15_5
#define init_13_8_15_5
void FASTCALL call_13_8_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_15_6
#define init_13_8_15_6
void FASTCALL call_13_8_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_15_7
#define init_13_8_15_7
void FASTCALL call_13_8_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_15_8
#define init_13_8_15_8
void FASTCALL call_13_8_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_15_9
#define init_13_8_15_9
void FASTCALL call_13_8_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_15_10
#define init_13_8_15_10
void FASTCALL call_13_8_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_15_11
#define init_13_8_15_11
void FASTCALL call_13_8_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_15_12
#define init_13_8_15_12
void FASTCALL call_13_8_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_15_13
#define init_13_8_15_13
void FASTCALL call_13_8_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_15_14
#define init_13_8_15_14
void FASTCALL call_13_8_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_8_15_15
#define init_13_8_15_15
void FASTCALL call_13_8_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_0_0
#define init_13_9_0_0
void FASTCALL call_13_9_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_0_1
#define init_13_9_0_1
void FASTCALL call_13_9_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_0_2
#define init_13_9_0_2
void FASTCALL call_13_9_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_0_3
#define init_13_9_0_3
void FASTCALL call_13_9_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_0_4
#define init_13_9_0_4
void FASTCALL call_13_9_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_0_5
#define init_13_9_0_5
void FASTCALL call_13_9_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_0_6
#define init_13_9_0_6
void FASTCALL call_13_9_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_0_7
#define init_13_9_0_7
void FASTCALL call_13_9_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_0_8
#define init_13_9_0_8
void FASTCALL call_13_9_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_0_9
#define init_13_9_0_9
void FASTCALL call_13_9_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_0_10
#define init_13_9_0_10
void FASTCALL call_13_9_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_0_11
#define init_13_9_0_11
void FASTCALL call_13_9_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_0_12
#define init_13_9_0_12
void FASTCALL call_13_9_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_0_13
#define init_13_9_0_13
void FASTCALL call_13_9_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_0_14
#define init_13_9_0_14
void FASTCALL call_13_9_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_0_15
#define init_13_9_0_15
void FASTCALL call_13_9_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_1_0
#define init_13_9_1_0
void FASTCALL call_13_9_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_1_1
#define init_13_9_1_1
void FASTCALL call_13_9_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_1_2
#define init_13_9_1_2
void FASTCALL call_13_9_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_1_3
#define init_13_9_1_3
void FASTCALL call_13_9_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_1_4
#define init_13_9_1_4
void FASTCALL call_13_9_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_1_5
#define init_13_9_1_5
void FASTCALL call_13_9_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_1_6
#define init_13_9_1_6
void FASTCALL call_13_9_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_1_7
#define init_13_9_1_7
void FASTCALL call_13_9_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_1_8
#define init_13_9_1_8
void FASTCALL call_13_9_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_1_9
#define init_13_9_1_9
void FASTCALL call_13_9_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_1_10
#define init_13_9_1_10
void FASTCALL call_13_9_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_1_11
#define init_13_9_1_11
void FASTCALL call_13_9_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_1_12
#define init_13_9_1_12
void FASTCALL call_13_9_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_1_13
#define init_13_9_1_13
void FASTCALL call_13_9_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_1_14
#define init_13_9_1_14
void FASTCALL call_13_9_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_1_15
#define init_13_9_1_15
void FASTCALL call_13_9_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_2_0
#define init_13_9_2_0
void FASTCALL call_13_9_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_2_1
#define init_13_9_2_1
void FASTCALL call_13_9_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_2_2
#define init_13_9_2_2
void FASTCALL call_13_9_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_2_3
#define init_13_9_2_3
void FASTCALL call_13_9_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_2_4
#define init_13_9_2_4
void FASTCALL call_13_9_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_2_5
#define init_13_9_2_5
void FASTCALL call_13_9_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_2_6
#define init_13_9_2_6
void FASTCALL call_13_9_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_2_7
#define init_13_9_2_7
void FASTCALL call_13_9_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_2_8
#define init_13_9_2_8
void FASTCALL call_13_9_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_2_9
#define init_13_9_2_9
void FASTCALL call_13_9_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_2_10
#define init_13_9_2_10
void FASTCALL call_13_9_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_2_11
#define init_13_9_2_11
void FASTCALL call_13_9_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_2_12
#define init_13_9_2_12
void FASTCALL call_13_9_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_2_13
#define init_13_9_2_13
void FASTCALL call_13_9_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_2_14
#define init_13_9_2_14
void FASTCALL call_13_9_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_2_15
#define init_13_9_2_15
void FASTCALL call_13_9_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_3_0
#define init_13_9_3_0
void FASTCALL call_13_9_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_3_1
#define init_13_9_3_1
void FASTCALL call_13_9_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_3_2
#define init_13_9_3_2
void FASTCALL call_13_9_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_3_3
#define init_13_9_3_3
void FASTCALL call_13_9_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_3_4
#define init_13_9_3_4
void FASTCALL call_13_9_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_3_5
#define init_13_9_3_5
void FASTCALL call_13_9_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_3_6
#define init_13_9_3_6
void FASTCALL call_13_9_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_3_7
#define init_13_9_3_7
void FASTCALL call_13_9_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_3_8
#define init_13_9_3_8
void FASTCALL call_13_9_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_3_9
#define init_13_9_3_9
void FASTCALL call_13_9_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_3_10
#define init_13_9_3_10
void FASTCALL call_13_9_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_3_11
#define init_13_9_3_11
void FASTCALL call_13_9_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_3_12
#define init_13_9_3_12
void FASTCALL call_13_9_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_3_13
#define init_13_9_3_13
void FASTCALL call_13_9_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_3_14
#define init_13_9_3_14
void FASTCALL call_13_9_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_3_15
#define init_13_9_3_15
void FASTCALL call_13_9_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_4_0
#define init_13_9_4_0
void FASTCALL call_13_9_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_4_1
#define init_13_9_4_1
void FASTCALL call_13_9_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_4_2
#define init_13_9_4_2
void FASTCALL call_13_9_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_4_3
#define init_13_9_4_3
void FASTCALL call_13_9_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_4_4
#define init_13_9_4_4
void FASTCALL call_13_9_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_4_5
#define init_13_9_4_5
void FASTCALL call_13_9_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_4_6
#define init_13_9_4_6
void FASTCALL call_13_9_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_4_7
#define init_13_9_4_7
void FASTCALL call_13_9_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_4_8
#define init_13_9_4_8
void FASTCALL call_13_9_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_4_9
#define init_13_9_4_9
void FASTCALL call_13_9_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_4_10
#define init_13_9_4_10
void FASTCALL call_13_9_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_4_11
#define init_13_9_4_11
void FASTCALL call_13_9_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_4_12
#define init_13_9_4_12
void FASTCALL call_13_9_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_4_13
#define init_13_9_4_13
void FASTCALL call_13_9_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_4_14
#define init_13_9_4_14
void FASTCALL call_13_9_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_4_15
#define init_13_9_4_15
void FASTCALL call_13_9_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_5_0
#define init_13_9_5_0
void FASTCALL call_13_9_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_5_1
#define init_13_9_5_1
void FASTCALL call_13_9_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_5_2
#define init_13_9_5_2
void FASTCALL call_13_9_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_5_3
#define init_13_9_5_3
void FASTCALL call_13_9_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_5_4
#define init_13_9_5_4
void FASTCALL call_13_9_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_5_5
#define init_13_9_5_5
void FASTCALL call_13_9_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_5_6
#define init_13_9_5_6
void FASTCALL call_13_9_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_5_7
#define init_13_9_5_7
void FASTCALL call_13_9_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_5_8
#define init_13_9_5_8
void FASTCALL call_13_9_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_5_9
#define init_13_9_5_9
void FASTCALL call_13_9_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_5_10
#define init_13_9_5_10
void FASTCALL call_13_9_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_5_11
#define init_13_9_5_11
void FASTCALL call_13_9_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_5_12
#define init_13_9_5_12
void FASTCALL call_13_9_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_5_13
#define init_13_9_5_13
void FASTCALL call_13_9_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_5_14
#define init_13_9_5_14
void FASTCALL call_13_9_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_5_15
#define init_13_9_5_15
void FASTCALL call_13_9_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_6_0
#define init_13_9_6_0
void FASTCALL call_13_9_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_6_1
#define init_13_9_6_1
void FASTCALL call_13_9_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_6_2
#define init_13_9_6_2
void FASTCALL call_13_9_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_6_3
#define init_13_9_6_3
void FASTCALL call_13_9_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_6_4
#define init_13_9_6_4
void FASTCALL call_13_9_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_6_5
#define init_13_9_6_5
void FASTCALL call_13_9_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_6_6
#define init_13_9_6_6
void FASTCALL call_13_9_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_6_7
#define init_13_9_6_7
void FASTCALL call_13_9_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_6_8
#define init_13_9_6_8
void FASTCALL call_13_9_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_6_9
#define init_13_9_6_9
void FASTCALL call_13_9_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_6_10
#define init_13_9_6_10
void FASTCALL call_13_9_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_6_11
#define init_13_9_6_11
void FASTCALL call_13_9_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_6_12
#define init_13_9_6_12
void FASTCALL call_13_9_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_6_13
#define init_13_9_6_13
void FASTCALL call_13_9_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_6_14
#define init_13_9_6_14
void FASTCALL call_13_9_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_6_15
#define init_13_9_6_15
void FASTCALL call_13_9_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_7_0
#define init_13_9_7_0
void FASTCALL call_13_9_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_7_1
#define init_13_9_7_1
void FASTCALL call_13_9_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_7_2
#define init_13_9_7_2
void FASTCALL call_13_9_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_7_3
#define init_13_9_7_3
void FASTCALL call_13_9_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_7_4
#define init_13_9_7_4
void FASTCALL call_13_9_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_7_5
#define init_13_9_7_5
void FASTCALL call_13_9_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_7_6
#define init_13_9_7_6
void FASTCALL call_13_9_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_7_7
#define init_13_9_7_7
void FASTCALL call_13_9_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_7_8
#define init_13_9_7_8
void FASTCALL call_13_9_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_7_9
#define init_13_9_7_9
void FASTCALL call_13_9_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_7_10
#define init_13_9_7_10
void FASTCALL call_13_9_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_7_11
#define init_13_9_7_11
void FASTCALL call_13_9_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_7_12
#define init_13_9_7_12
void FASTCALL call_13_9_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_7_13
#define init_13_9_7_13
void FASTCALL call_13_9_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_7_14
#define init_13_9_7_14
void FASTCALL call_13_9_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_7_15
#define init_13_9_7_15
void FASTCALL call_13_9_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_8_0
#define init_13_9_8_0
void FASTCALL call_13_9_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_8_1
#define init_13_9_8_1
void FASTCALL call_13_9_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_8_2
#define init_13_9_8_2
void FASTCALL call_13_9_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_8_3
#define init_13_9_8_3
void FASTCALL call_13_9_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_8_4
#define init_13_9_8_4
void FASTCALL call_13_9_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_8_5
#define init_13_9_8_5
void FASTCALL call_13_9_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_8_6
#define init_13_9_8_6
void FASTCALL call_13_9_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_8_7
#define init_13_9_8_7
void FASTCALL call_13_9_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_8_8
#define init_13_9_8_8
void FASTCALL call_13_9_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_8_9
#define init_13_9_8_9
void FASTCALL call_13_9_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_8_10
#define init_13_9_8_10
void FASTCALL call_13_9_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_8_11
#define init_13_9_8_11
void FASTCALL call_13_9_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_8_12
#define init_13_9_8_12
void FASTCALL call_13_9_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_8_13
#define init_13_9_8_13
void FASTCALL call_13_9_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_8_14
#define init_13_9_8_14
void FASTCALL call_13_9_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_8_15
#define init_13_9_8_15
void FASTCALL call_13_9_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_9_0
#define init_13_9_9_0
void FASTCALL call_13_9_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_9_1
#define init_13_9_9_1
void FASTCALL call_13_9_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_9_2
#define init_13_9_9_2
void FASTCALL call_13_9_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_9_3
#define init_13_9_9_3
void FASTCALL call_13_9_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_9_4
#define init_13_9_9_4
void FASTCALL call_13_9_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_9_5
#define init_13_9_9_5
void FASTCALL call_13_9_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_9_6
#define init_13_9_9_6
void FASTCALL call_13_9_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_9_7
#define init_13_9_9_7
void FASTCALL call_13_9_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_9_8
#define init_13_9_9_8
void FASTCALL call_13_9_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_9_9
#define init_13_9_9_9
void FASTCALL call_13_9_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_9_10
#define init_13_9_9_10
void FASTCALL call_13_9_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_9_11
#define init_13_9_9_11
void FASTCALL call_13_9_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_9_12
#define init_13_9_9_12
void FASTCALL call_13_9_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_9_13
#define init_13_9_9_13
void FASTCALL call_13_9_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_9_14
#define init_13_9_9_14
void FASTCALL call_13_9_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_9_15
#define init_13_9_9_15
void FASTCALL call_13_9_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_10_0
#define init_13_9_10_0
void FASTCALL call_13_9_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_10_1
#define init_13_9_10_1
void FASTCALL call_13_9_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_10_2
#define init_13_9_10_2
void FASTCALL call_13_9_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_10_3
#define init_13_9_10_3
void FASTCALL call_13_9_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_10_4
#define init_13_9_10_4
void FASTCALL call_13_9_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_10_5
#define init_13_9_10_5
void FASTCALL call_13_9_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_10_6
#define init_13_9_10_6
void FASTCALL call_13_9_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_10_7
#define init_13_9_10_7
void FASTCALL call_13_9_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_10_8
#define init_13_9_10_8
void FASTCALL call_13_9_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_10_9
#define init_13_9_10_9
void FASTCALL call_13_9_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_10_10
#define init_13_9_10_10
void FASTCALL call_13_9_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_10_11
#define init_13_9_10_11
void FASTCALL call_13_9_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_10_12
#define init_13_9_10_12
void FASTCALL call_13_9_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_10_13
#define init_13_9_10_13
void FASTCALL call_13_9_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_10_14
#define init_13_9_10_14
void FASTCALL call_13_9_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_10_15
#define init_13_9_10_15
void FASTCALL call_13_9_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_11_0
#define init_13_9_11_0
void FASTCALL call_13_9_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_11_1
#define init_13_9_11_1
void FASTCALL call_13_9_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_11_2
#define init_13_9_11_2
void FASTCALL call_13_9_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_11_3
#define init_13_9_11_3
void FASTCALL call_13_9_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_11_4
#define init_13_9_11_4
void FASTCALL call_13_9_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_11_5
#define init_13_9_11_5
void FASTCALL call_13_9_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_11_6
#define init_13_9_11_6
void FASTCALL call_13_9_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_11_7
#define init_13_9_11_7
void FASTCALL call_13_9_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_11_8
#define init_13_9_11_8
void FASTCALL call_13_9_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_11_9
#define init_13_9_11_9
void FASTCALL call_13_9_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_11_10
#define init_13_9_11_10
void FASTCALL call_13_9_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_11_11
#define init_13_9_11_11
void FASTCALL call_13_9_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_11_12
#define init_13_9_11_12
void FASTCALL call_13_9_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_11_13
#define init_13_9_11_13
void FASTCALL call_13_9_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_11_14
#define init_13_9_11_14
void FASTCALL call_13_9_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_11_15
#define init_13_9_11_15
void FASTCALL call_13_9_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_12_0
#define init_13_9_12_0
void FASTCALL call_13_9_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_12_1
#define init_13_9_12_1
void FASTCALL call_13_9_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_12_2
#define init_13_9_12_2
void FASTCALL call_13_9_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_12_3
#define init_13_9_12_3
void FASTCALL call_13_9_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_12_4
#define init_13_9_12_4
void FASTCALL call_13_9_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_12_5
#define init_13_9_12_5
void FASTCALL call_13_9_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_12_6
#define init_13_9_12_6
void FASTCALL call_13_9_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_12_7
#define init_13_9_12_7
void FASTCALL call_13_9_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_12_8
#define init_13_9_12_8
void FASTCALL call_13_9_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_12_9
#define init_13_9_12_9
void FASTCALL call_13_9_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_12_10
#define init_13_9_12_10
void FASTCALL call_13_9_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_12_11
#define init_13_9_12_11
void FASTCALL call_13_9_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_12_12
#define init_13_9_12_12
void FASTCALL call_13_9_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_12_13
#define init_13_9_12_13
void FASTCALL call_13_9_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_12_14
#define init_13_9_12_14
void FASTCALL call_13_9_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_12_15
#define init_13_9_12_15
void FASTCALL call_13_9_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_13_0
#define init_13_9_13_0
void FASTCALL call_13_9_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_13_1
#define init_13_9_13_1
void FASTCALL call_13_9_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_13_2
#define init_13_9_13_2
void FASTCALL call_13_9_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_13_3
#define init_13_9_13_3
void FASTCALL call_13_9_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_13_4
#define init_13_9_13_4
void FASTCALL call_13_9_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_13_5
#define init_13_9_13_5
void FASTCALL call_13_9_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_13_6
#define init_13_9_13_6
void FASTCALL call_13_9_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_13_7
#define init_13_9_13_7
void FASTCALL call_13_9_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_13_8
#define init_13_9_13_8
void FASTCALL call_13_9_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_13_9
#define init_13_9_13_9
void FASTCALL call_13_9_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_13_10
#define init_13_9_13_10
void FASTCALL call_13_9_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_13_11
#define init_13_9_13_11
void FASTCALL call_13_9_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_13_12
#define init_13_9_13_12
void FASTCALL call_13_9_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_13_13
#define init_13_9_13_13
void FASTCALL call_13_9_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_13_14
#define init_13_9_13_14
void FASTCALL call_13_9_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_13_15
#define init_13_9_13_15
void FASTCALL call_13_9_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_14_0
#define init_13_9_14_0
void FASTCALL call_13_9_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_14_1
#define init_13_9_14_1
void FASTCALL call_13_9_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_14_2
#define init_13_9_14_2
void FASTCALL call_13_9_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_14_3
#define init_13_9_14_3
void FASTCALL call_13_9_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_14_4
#define init_13_9_14_4
void FASTCALL call_13_9_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_14_5
#define init_13_9_14_5
void FASTCALL call_13_9_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_14_6
#define init_13_9_14_6
void FASTCALL call_13_9_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_14_7
#define init_13_9_14_7
void FASTCALL call_13_9_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_14_8
#define init_13_9_14_8
void FASTCALL call_13_9_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_14_9
#define init_13_9_14_9
void FASTCALL call_13_9_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_14_10
#define init_13_9_14_10
void FASTCALL call_13_9_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_14_11
#define init_13_9_14_11
void FASTCALL call_13_9_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_14_12
#define init_13_9_14_12
void FASTCALL call_13_9_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_14_13
#define init_13_9_14_13
void FASTCALL call_13_9_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_14_14
#define init_13_9_14_14
void FASTCALL call_13_9_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_14_15
#define init_13_9_14_15
void FASTCALL call_13_9_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_15_0
#define init_13_9_15_0
void FASTCALL call_13_9_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_15_1
#define init_13_9_15_1
void FASTCALL call_13_9_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_15_2
#define init_13_9_15_2
void FASTCALL call_13_9_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_15_3
#define init_13_9_15_3
void FASTCALL call_13_9_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_15_4
#define init_13_9_15_4
void FASTCALL call_13_9_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_15_5
#define init_13_9_15_5
void FASTCALL call_13_9_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_15_6
#define init_13_9_15_6
void FASTCALL call_13_9_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_15_7
#define init_13_9_15_7
void FASTCALL call_13_9_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_15_8
#define init_13_9_15_8
void FASTCALL call_13_9_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_15_9
#define init_13_9_15_9
void FASTCALL call_13_9_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_15_10
#define init_13_9_15_10
void FASTCALL call_13_9_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_15_11
#define init_13_9_15_11
void FASTCALL call_13_9_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_15_12
#define init_13_9_15_12
void FASTCALL call_13_9_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_15_13
#define init_13_9_15_13
void FASTCALL call_13_9_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_15_14
#define init_13_9_15_14
void FASTCALL call_13_9_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_9_15_15
#define init_13_9_15_15
void FASTCALL call_13_9_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_0_0
#define init_13_10_0_0
void FASTCALL call_13_10_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_0_1
#define init_13_10_0_1
void FASTCALL call_13_10_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_0_2
#define init_13_10_0_2
void FASTCALL call_13_10_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_0_3
#define init_13_10_0_3
void FASTCALL call_13_10_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_0_4
#define init_13_10_0_4
void FASTCALL call_13_10_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_0_5
#define init_13_10_0_5
void FASTCALL call_13_10_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_0_6
#define init_13_10_0_6
void FASTCALL call_13_10_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_0_7
#define init_13_10_0_7
void FASTCALL call_13_10_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_0_8
#define init_13_10_0_8
void FASTCALL call_13_10_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_0_9
#define init_13_10_0_9
void FASTCALL call_13_10_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_0_10
#define init_13_10_0_10
void FASTCALL call_13_10_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_0_11
#define init_13_10_0_11
void FASTCALL call_13_10_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_0_12
#define init_13_10_0_12
void FASTCALL call_13_10_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_0_13
#define init_13_10_0_13
void FASTCALL call_13_10_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_0_14
#define init_13_10_0_14
void FASTCALL call_13_10_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_0_15
#define init_13_10_0_15
void FASTCALL call_13_10_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_1_0
#define init_13_10_1_0
void FASTCALL call_13_10_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_1_1
#define init_13_10_1_1
void FASTCALL call_13_10_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_1_2
#define init_13_10_1_2
void FASTCALL call_13_10_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_1_3
#define init_13_10_1_3
void FASTCALL call_13_10_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_1_4
#define init_13_10_1_4
void FASTCALL call_13_10_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_1_5
#define init_13_10_1_5
void FASTCALL call_13_10_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_1_6
#define init_13_10_1_6
void FASTCALL call_13_10_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_1_7
#define init_13_10_1_7
void FASTCALL call_13_10_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_1_8
#define init_13_10_1_8
void FASTCALL call_13_10_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_1_9
#define init_13_10_1_9
void FASTCALL call_13_10_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_1_10
#define init_13_10_1_10
void FASTCALL call_13_10_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_1_11
#define init_13_10_1_11
void FASTCALL call_13_10_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_1_12
#define init_13_10_1_12
void FASTCALL call_13_10_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_1_13
#define init_13_10_1_13
void FASTCALL call_13_10_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_1_14
#define init_13_10_1_14
void FASTCALL call_13_10_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_1_15
#define init_13_10_1_15
void FASTCALL call_13_10_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_2_0
#define init_13_10_2_0
void FASTCALL call_13_10_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_2_1
#define init_13_10_2_1
void FASTCALL call_13_10_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_2_2
#define init_13_10_2_2
void FASTCALL call_13_10_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_2_3
#define init_13_10_2_3
void FASTCALL call_13_10_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_2_4
#define init_13_10_2_4
void FASTCALL call_13_10_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_2_5
#define init_13_10_2_5
void FASTCALL call_13_10_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_2_6
#define init_13_10_2_6
void FASTCALL call_13_10_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_2_7
#define init_13_10_2_7
void FASTCALL call_13_10_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_2_8
#define init_13_10_2_8
void FASTCALL call_13_10_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_2_9
#define init_13_10_2_9
void FASTCALL call_13_10_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_2_10
#define init_13_10_2_10
void FASTCALL call_13_10_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_2_11
#define init_13_10_2_11
void FASTCALL call_13_10_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_2_12
#define init_13_10_2_12
void FASTCALL call_13_10_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_2_13
#define init_13_10_2_13
void FASTCALL call_13_10_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_2_14
#define init_13_10_2_14
void FASTCALL call_13_10_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_2_15
#define init_13_10_2_15
void FASTCALL call_13_10_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_3_0
#define init_13_10_3_0
void FASTCALL call_13_10_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_3_1
#define init_13_10_3_1
void FASTCALL call_13_10_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_3_2
#define init_13_10_3_2
void FASTCALL call_13_10_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_3_3
#define init_13_10_3_3
void FASTCALL call_13_10_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_3_4
#define init_13_10_3_4
void FASTCALL call_13_10_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_3_5
#define init_13_10_3_5
void FASTCALL call_13_10_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_3_6
#define init_13_10_3_6
void FASTCALL call_13_10_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_3_7
#define init_13_10_3_7
void FASTCALL call_13_10_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_3_8
#define init_13_10_3_8
void FASTCALL call_13_10_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_3_9
#define init_13_10_3_9
void FASTCALL call_13_10_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_3_10
#define init_13_10_3_10
void FASTCALL call_13_10_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_3_11
#define init_13_10_3_11
void FASTCALL call_13_10_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_3_12
#define init_13_10_3_12
void FASTCALL call_13_10_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_3_13
#define init_13_10_3_13
void FASTCALL call_13_10_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_3_14
#define init_13_10_3_14
void FASTCALL call_13_10_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_3_15
#define init_13_10_3_15
void FASTCALL call_13_10_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_4_0
#define init_13_10_4_0
void FASTCALL call_13_10_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_4_1
#define init_13_10_4_1
void FASTCALL call_13_10_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_4_2
#define init_13_10_4_2
void FASTCALL call_13_10_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_4_3
#define init_13_10_4_3
void FASTCALL call_13_10_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_4_4
#define init_13_10_4_4
void FASTCALL call_13_10_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_4_5
#define init_13_10_4_5
void FASTCALL call_13_10_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_4_6
#define init_13_10_4_6
void FASTCALL call_13_10_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_4_7
#define init_13_10_4_7
void FASTCALL call_13_10_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_4_8
#define init_13_10_4_8
void FASTCALL call_13_10_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_4_9
#define init_13_10_4_9
void FASTCALL call_13_10_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_4_10
#define init_13_10_4_10
void FASTCALL call_13_10_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_4_11
#define init_13_10_4_11
void FASTCALL call_13_10_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_4_12
#define init_13_10_4_12
void FASTCALL call_13_10_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_4_13
#define init_13_10_4_13
void FASTCALL call_13_10_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_4_14
#define init_13_10_4_14
void FASTCALL call_13_10_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_4_15
#define init_13_10_4_15
void FASTCALL call_13_10_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_5_0
#define init_13_10_5_0
void FASTCALL call_13_10_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_5_1
#define init_13_10_5_1
void FASTCALL call_13_10_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_5_2
#define init_13_10_5_2
void FASTCALL call_13_10_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_5_3
#define init_13_10_5_3
void FASTCALL call_13_10_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_5_4
#define init_13_10_5_4
void FASTCALL call_13_10_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_5_5
#define init_13_10_5_5
void FASTCALL call_13_10_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_5_6
#define init_13_10_5_6
void FASTCALL call_13_10_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_5_7
#define init_13_10_5_7
void FASTCALL call_13_10_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_5_8
#define init_13_10_5_8
void FASTCALL call_13_10_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_5_9
#define init_13_10_5_9
void FASTCALL call_13_10_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_5_10
#define init_13_10_5_10
void FASTCALL call_13_10_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_5_11
#define init_13_10_5_11
void FASTCALL call_13_10_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_5_12
#define init_13_10_5_12
void FASTCALL call_13_10_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_5_13
#define init_13_10_5_13
void FASTCALL call_13_10_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_5_14
#define init_13_10_5_14
void FASTCALL call_13_10_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_5_15
#define init_13_10_5_15
void FASTCALL call_13_10_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_6_0
#define init_13_10_6_0
void FASTCALL call_13_10_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_6_1
#define init_13_10_6_1
void FASTCALL call_13_10_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_6_2
#define init_13_10_6_2
void FASTCALL call_13_10_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_6_3
#define init_13_10_6_3
void FASTCALL call_13_10_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_6_4
#define init_13_10_6_4
void FASTCALL call_13_10_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_6_5
#define init_13_10_6_5
void FASTCALL call_13_10_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_6_6
#define init_13_10_6_6
void FASTCALL call_13_10_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_6_7
#define init_13_10_6_7
void FASTCALL call_13_10_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_6_8
#define init_13_10_6_8
void FASTCALL call_13_10_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_6_9
#define init_13_10_6_9
void FASTCALL call_13_10_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_6_10
#define init_13_10_6_10
void FASTCALL call_13_10_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_6_11
#define init_13_10_6_11
void FASTCALL call_13_10_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_6_12
#define init_13_10_6_12
void FASTCALL call_13_10_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_6_13
#define init_13_10_6_13
void FASTCALL call_13_10_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_6_14
#define init_13_10_6_14
void FASTCALL call_13_10_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_6_15
#define init_13_10_6_15
void FASTCALL call_13_10_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_7_0
#define init_13_10_7_0
void FASTCALL call_13_10_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_7_1
#define init_13_10_7_1
void FASTCALL call_13_10_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_7_2
#define init_13_10_7_2
void FASTCALL call_13_10_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_7_3
#define init_13_10_7_3
void FASTCALL call_13_10_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_7_4
#define init_13_10_7_4
void FASTCALL call_13_10_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_7_5
#define init_13_10_7_5
void FASTCALL call_13_10_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_7_6
#define init_13_10_7_6
void FASTCALL call_13_10_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_7_7
#define init_13_10_7_7
void FASTCALL call_13_10_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_7_8
#define init_13_10_7_8
void FASTCALL call_13_10_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_7_9
#define init_13_10_7_9
void FASTCALL call_13_10_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_7_10
#define init_13_10_7_10
void FASTCALL call_13_10_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_7_11
#define init_13_10_7_11
void FASTCALL call_13_10_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_7_12
#define init_13_10_7_12
void FASTCALL call_13_10_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_7_13
#define init_13_10_7_13
void FASTCALL call_13_10_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_7_14
#define init_13_10_7_14
void FASTCALL call_13_10_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_7_15
#define init_13_10_7_15
void FASTCALL call_13_10_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_8_0
#define init_13_10_8_0
void FASTCALL call_13_10_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_8_1
#define init_13_10_8_1
void FASTCALL call_13_10_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_8_2
#define init_13_10_8_2
void FASTCALL call_13_10_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_8_3
#define init_13_10_8_3
void FASTCALL call_13_10_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_8_4
#define init_13_10_8_4
void FASTCALL call_13_10_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_8_5
#define init_13_10_8_5
void FASTCALL call_13_10_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_8_6
#define init_13_10_8_6
void FASTCALL call_13_10_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_8_7
#define init_13_10_8_7
void FASTCALL call_13_10_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_8_8
#define init_13_10_8_8
void FASTCALL call_13_10_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_8_9
#define init_13_10_8_9
void FASTCALL call_13_10_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_8_10
#define init_13_10_8_10
void FASTCALL call_13_10_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_8_11
#define init_13_10_8_11
void FASTCALL call_13_10_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_8_12
#define init_13_10_8_12
void FASTCALL call_13_10_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_8_13
#define init_13_10_8_13
void FASTCALL call_13_10_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_8_14
#define init_13_10_8_14
void FASTCALL call_13_10_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_8_15
#define init_13_10_8_15
void FASTCALL call_13_10_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_9_0
#define init_13_10_9_0
void FASTCALL call_13_10_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_9_1
#define init_13_10_9_1
void FASTCALL call_13_10_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_9_2
#define init_13_10_9_2
void FASTCALL call_13_10_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_9_3
#define init_13_10_9_3
void FASTCALL call_13_10_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_9_4
#define init_13_10_9_4
void FASTCALL call_13_10_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_9_5
#define init_13_10_9_5
void FASTCALL call_13_10_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_9_6
#define init_13_10_9_6
void FASTCALL call_13_10_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_9_7
#define init_13_10_9_7
void FASTCALL call_13_10_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_9_8
#define init_13_10_9_8
void FASTCALL call_13_10_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_9_9
#define init_13_10_9_9
void FASTCALL call_13_10_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_9_10
#define init_13_10_9_10
void FASTCALL call_13_10_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_9_11
#define init_13_10_9_11
void FASTCALL call_13_10_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_9_12
#define init_13_10_9_12
void FASTCALL call_13_10_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_9_13
#define init_13_10_9_13
void FASTCALL call_13_10_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_9_14
#define init_13_10_9_14
void FASTCALL call_13_10_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_9_15
#define init_13_10_9_15
void FASTCALL call_13_10_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_10_0
#define init_13_10_10_0
void FASTCALL call_13_10_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_10_1
#define init_13_10_10_1
void FASTCALL call_13_10_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_10_2
#define init_13_10_10_2
void FASTCALL call_13_10_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_10_3
#define init_13_10_10_3
void FASTCALL call_13_10_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_10_4
#define init_13_10_10_4
void FASTCALL call_13_10_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_10_5
#define init_13_10_10_5
void FASTCALL call_13_10_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_10_6
#define init_13_10_10_6
void FASTCALL call_13_10_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_10_7
#define init_13_10_10_7
void FASTCALL call_13_10_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_10_8
#define init_13_10_10_8
void FASTCALL call_13_10_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_10_9
#define init_13_10_10_9
void FASTCALL call_13_10_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_10_10
#define init_13_10_10_10
void FASTCALL call_13_10_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_10_11
#define init_13_10_10_11
void FASTCALL call_13_10_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_10_12
#define init_13_10_10_12
void FASTCALL call_13_10_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_10_13
#define init_13_10_10_13
void FASTCALL call_13_10_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_10_14
#define init_13_10_10_14
void FASTCALL call_13_10_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_10_15
#define init_13_10_10_15
void FASTCALL call_13_10_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_11_0
#define init_13_10_11_0
void FASTCALL call_13_10_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_11_1
#define init_13_10_11_1
void FASTCALL call_13_10_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_11_2
#define init_13_10_11_2
void FASTCALL call_13_10_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_11_3
#define init_13_10_11_3
void FASTCALL call_13_10_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_11_4
#define init_13_10_11_4
void FASTCALL call_13_10_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_11_5
#define init_13_10_11_5
void FASTCALL call_13_10_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_11_6
#define init_13_10_11_6
void FASTCALL call_13_10_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_11_7
#define init_13_10_11_7
void FASTCALL call_13_10_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_11_8
#define init_13_10_11_8
void FASTCALL call_13_10_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_11_9
#define init_13_10_11_9
void FASTCALL call_13_10_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_11_10
#define init_13_10_11_10
void FASTCALL call_13_10_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_11_11
#define init_13_10_11_11
void FASTCALL call_13_10_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_11_12
#define init_13_10_11_12
void FASTCALL call_13_10_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_11_13
#define init_13_10_11_13
void FASTCALL call_13_10_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_11_14
#define init_13_10_11_14
void FASTCALL call_13_10_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_11_15
#define init_13_10_11_15
void FASTCALL call_13_10_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_12_0
#define init_13_10_12_0
void FASTCALL call_13_10_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_12_1
#define init_13_10_12_1
void FASTCALL call_13_10_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_12_2
#define init_13_10_12_2
void FASTCALL call_13_10_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_12_3
#define init_13_10_12_3
void FASTCALL call_13_10_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_12_4
#define init_13_10_12_4
void FASTCALL call_13_10_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_12_5
#define init_13_10_12_5
void FASTCALL call_13_10_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_12_6
#define init_13_10_12_6
void FASTCALL call_13_10_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_12_7
#define init_13_10_12_7
void FASTCALL call_13_10_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_12_8
#define init_13_10_12_8
void FASTCALL call_13_10_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_12_9
#define init_13_10_12_9
void FASTCALL call_13_10_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_12_10
#define init_13_10_12_10
void FASTCALL call_13_10_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_12_11
#define init_13_10_12_11
void FASTCALL call_13_10_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_12_12
#define init_13_10_12_12
void FASTCALL call_13_10_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_12_13
#define init_13_10_12_13
void FASTCALL call_13_10_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_12_14
#define init_13_10_12_14
void FASTCALL call_13_10_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_12_15
#define init_13_10_12_15
void FASTCALL call_13_10_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_13_0
#define init_13_10_13_0
void FASTCALL call_13_10_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_13_1
#define init_13_10_13_1
void FASTCALL call_13_10_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_13_2
#define init_13_10_13_2
void FASTCALL call_13_10_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_13_3
#define init_13_10_13_3
void FASTCALL call_13_10_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_13_4
#define init_13_10_13_4
void FASTCALL call_13_10_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_13_5
#define init_13_10_13_5
void FASTCALL call_13_10_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_13_6
#define init_13_10_13_6
void FASTCALL call_13_10_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_13_7
#define init_13_10_13_7
void FASTCALL call_13_10_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_13_8
#define init_13_10_13_8
void FASTCALL call_13_10_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_13_9
#define init_13_10_13_9
void FASTCALL call_13_10_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_13_10
#define init_13_10_13_10
void FASTCALL call_13_10_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_13_11
#define init_13_10_13_11
void FASTCALL call_13_10_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_13_12
#define init_13_10_13_12
void FASTCALL call_13_10_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_13_13
#define init_13_10_13_13
void FASTCALL call_13_10_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_13_14
#define init_13_10_13_14
void FASTCALL call_13_10_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_13_15
#define init_13_10_13_15
void FASTCALL call_13_10_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_14_0
#define init_13_10_14_0
void FASTCALL call_13_10_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_14_1
#define init_13_10_14_1
void FASTCALL call_13_10_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_14_2
#define init_13_10_14_2
void FASTCALL call_13_10_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_14_3
#define init_13_10_14_3
void FASTCALL call_13_10_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_14_4
#define init_13_10_14_4
void FASTCALL call_13_10_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_14_5
#define init_13_10_14_5
void FASTCALL call_13_10_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_14_6
#define init_13_10_14_6
void FASTCALL call_13_10_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_14_7
#define init_13_10_14_7
void FASTCALL call_13_10_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_14_8
#define init_13_10_14_8
void FASTCALL call_13_10_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_14_9
#define init_13_10_14_9
void FASTCALL call_13_10_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_14_10
#define init_13_10_14_10
void FASTCALL call_13_10_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_14_11
#define init_13_10_14_11
void FASTCALL call_13_10_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_14_12
#define init_13_10_14_12
void FASTCALL call_13_10_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_14_13
#define init_13_10_14_13
void FASTCALL call_13_10_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_14_14
#define init_13_10_14_14
void FASTCALL call_13_10_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_14_15
#define init_13_10_14_15
void FASTCALL call_13_10_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_15_0
#define init_13_10_15_0
void FASTCALL call_13_10_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_15_1
#define init_13_10_15_1
void FASTCALL call_13_10_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_15_2
#define init_13_10_15_2
void FASTCALL call_13_10_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_15_3
#define init_13_10_15_3
void FASTCALL call_13_10_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_15_4
#define init_13_10_15_4
void FASTCALL call_13_10_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_15_5
#define init_13_10_15_5
void FASTCALL call_13_10_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_15_6
#define init_13_10_15_6
void FASTCALL call_13_10_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_15_7
#define init_13_10_15_7
void FASTCALL call_13_10_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_15_8
#define init_13_10_15_8
void FASTCALL call_13_10_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_15_9
#define init_13_10_15_9
void FASTCALL call_13_10_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_15_10
#define init_13_10_15_10
void FASTCALL call_13_10_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_15_11
#define init_13_10_15_11
void FASTCALL call_13_10_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_15_12
#define init_13_10_15_12
void FASTCALL call_13_10_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_15_13
#define init_13_10_15_13
void FASTCALL call_13_10_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_15_14
#define init_13_10_15_14
void FASTCALL call_13_10_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_10_15_15
#define init_13_10_15_15
void FASTCALL call_13_10_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_0_0
#define init_13_11_0_0
void FASTCALL call_13_11_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_0_1
#define init_13_11_0_1
void FASTCALL call_13_11_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_0_2
#define init_13_11_0_2
void FASTCALL call_13_11_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_0_3
#define init_13_11_0_3
void FASTCALL call_13_11_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_0_4
#define init_13_11_0_4
void FASTCALL call_13_11_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_0_5
#define init_13_11_0_5
void FASTCALL call_13_11_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_0_6
#define init_13_11_0_6
void FASTCALL call_13_11_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_0_7
#define init_13_11_0_7
void FASTCALL call_13_11_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_0_8
#define init_13_11_0_8
void FASTCALL call_13_11_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_0_9
#define init_13_11_0_9
void FASTCALL call_13_11_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_0_10
#define init_13_11_0_10
void FASTCALL call_13_11_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_0_11
#define init_13_11_0_11
void FASTCALL call_13_11_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_0_12
#define init_13_11_0_12
void FASTCALL call_13_11_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_0_13
#define init_13_11_0_13
void FASTCALL call_13_11_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_0_14
#define init_13_11_0_14
void FASTCALL call_13_11_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_0_15
#define init_13_11_0_15
void FASTCALL call_13_11_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_1_0
#define init_13_11_1_0
void FASTCALL call_13_11_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_1_1
#define init_13_11_1_1
void FASTCALL call_13_11_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_1_2
#define init_13_11_1_2
void FASTCALL call_13_11_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_1_3
#define init_13_11_1_3
void FASTCALL call_13_11_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_1_4
#define init_13_11_1_4
void FASTCALL call_13_11_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_1_5
#define init_13_11_1_5
void FASTCALL call_13_11_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_1_6
#define init_13_11_1_6
void FASTCALL call_13_11_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_1_7
#define init_13_11_1_7
void FASTCALL call_13_11_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_1_8
#define init_13_11_1_8
void FASTCALL call_13_11_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_1_9
#define init_13_11_1_9
void FASTCALL call_13_11_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_1_10
#define init_13_11_1_10
void FASTCALL call_13_11_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_1_11
#define init_13_11_1_11
void FASTCALL call_13_11_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_1_12
#define init_13_11_1_12
void FASTCALL call_13_11_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_1_13
#define init_13_11_1_13
void FASTCALL call_13_11_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_1_14
#define init_13_11_1_14
void FASTCALL call_13_11_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_1_15
#define init_13_11_1_15
void FASTCALL call_13_11_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_2_0
#define init_13_11_2_0
void FASTCALL call_13_11_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_2_1
#define init_13_11_2_1
void FASTCALL call_13_11_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_2_2
#define init_13_11_2_2
void FASTCALL call_13_11_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_2_3
#define init_13_11_2_3
void FASTCALL call_13_11_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_2_4
#define init_13_11_2_4
void FASTCALL call_13_11_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_2_5
#define init_13_11_2_5
void FASTCALL call_13_11_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_2_6
#define init_13_11_2_6
void FASTCALL call_13_11_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_2_7
#define init_13_11_2_7
void FASTCALL call_13_11_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_2_8
#define init_13_11_2_8
void FASTCALL call_13_11_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_2_9
#define init_13_11_2_9
void FASTCALL call_13_11_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_2_10
#define init_13_11_2_10
void FASTCALL call_13_11_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_2_11
#define init_13_11_2_11
void FASTCALL call_13_11_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_2_12
#define init_13_11_2_12
void FASTCALL call_13_11_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_2_13
#define init_13_11_2_13
void FASTCALL call_13_11_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_2_14
#define init_13_11_2_14
void FASTCALL call_13_11_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_2_15
#define init_13_11_2_15
void FASTCALL call_13_11_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_3_0
#define init_13_11_3_0
void FASTCALL call_13_11_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_3_1
#define init_13_11_3_1
void FASTCALL call_13_11_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_3_2
#define init_13_11_3_2
void FASTCALL call_13_11_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_3_3
#define init_13_11_3_3
void FASTCALL call_13_11_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_3_4
#define init_13_11_3_4
void FASTCALL call_13_11_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_3_5
#define init_13_11_3_5
void FASTCALL call_13_11_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_3_6
#define init_13_11_3_6
void FASTCALL call_13_11_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_3_7
#define init_13_11_3_7
void FASTCALL call_13_11_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_3_8
#define init_13_11_3_8
void FASTCALL call_13_11_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_3_9
#define init_13_11_3_9
void FASTCALL call_13_11_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_3_10
#define init_13_11_3_10
void FASTCALL call_13_11_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_3_11
#define init_13_11_3_11
void FASTCALL call_13_11_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_3_12
#define init_13_11_3_12
void FASTCALL call_13_11_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_3_13
#define init_13_11_3_13
void FASTCALL call_13_11_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_3_14
#define init_13_11_3_14
void FASTCALL call_13_11_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_3_15
#define init_13_11_3_15
void FASTCALL call_13_11_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_4_0
#define init_13_11_4_0
void FASTCALL call_13_11_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_4_1
#define init_13_11_4_1
void FASTCALL call_13_11_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_4_2
#define init_13_11_4_2
void FASTCALL call_13_11_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_4_3
#define init_13_11_4_3
void FASTCALL call_13_11_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_4_4
#define init_13_11_4_4
void FASTCALL call_13_11_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_4_5
#define init_13_11_4_5
void FASTCALL call_13_11_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_4_6
#define init_13_11_4_6
void FASTCALL call_13_11_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_4_7
#define init_13_11_4_7
void FASTCALL call_13_11_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_4_8
#define init_13_11_4_8
void FASTCALL call_13_11_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_4_9
#define init_13_11_4_9
void FASTCALL call_13_11_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_4_10
#define init_13_11_4_10
void FASTCALL call_13_11_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_4_11
#define init_13_11_4_11
void FASTCALL call_13_11_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_4_12
#define init_13_11_4_12
void FASTCALL call_13_11_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_4_13
#define init_13_11_4_13
void FASTCALL call_13_11_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_4_14
#define init_13_11_4_14
void FASTCALL call_13_11_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_4_15
#define init_13_11_4_15
void FASTCALL call_13_11_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_5_0
#define init_13_11_5_0
void FASTCALL call_13_11_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_5_1
#define init_13_11_5_1
void FASTCALL call_13_11_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_5_2
#define init_13_11_5_2
void FASTCALL call_13_11_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_5_3
#define init_13_11_5_3
void FASTCALL call_13_11_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_5_4
#define init_13_11_5_4
void FASTCALL call_13_11_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_5_5
#define init_13_11_5_5
void FASTCALL call_13_11_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_5_6
#define init_13_11_5_6
void FASTCALL call_13_11_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_5_7
#define init_13_11_5_7
void FASTCALL call_13_11_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_5_8
#define init_13_11_5_8
void FASTCALL call_13_11_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_5_9
#define init_13_11_5_9
void FASTCALL call_13_11_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_5_10
#define init_13_11_5_10
void FASTCALL call_13_11_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_5_11
#define init_13_11_5_11
void FASTCALL call_13_11_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_5_12
#define init_13_11_5_12
void FASTCALL call_13_11_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_5_13
#define init_13_11_5_13
void FASTCALL call_13_11_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_5_14
#define init_13_11_5_14
void FASTCALL call_13_11_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_5_15
#define init_13_11_5_15
void FASTCALL call_13_11_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_6_0
#define init_13_11_6_0
void FASTCALL call_13_11_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_6_1
#define init_13_11_6_1
void FASTCALL call_13_11_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_6_2
#define init_13_11_6_2
void FASTCALL call_13_11_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_6_3
#define init_13_11_6_3
void FASTCALL call_13_11_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_6_4
#define init_13_11_6_4
void FASTCALL call_13_11_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_6_5
#define init_13_11_6_5
void FASTCALL call_13_11_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_6_6
#define init_13_11_6_6
void FASTCALL call_13_11_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_6_7
#define init_13_11_6_7
void FASTCALL call_13_11_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_6_8
#define init_13_11_6_8
void FASTCALL call_13_11_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_6_9
#define init_13_11_6_9
void FASTCALL call_13_11_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_6_10
#define init_13_11_6_10
void FASTCALL call_13_11_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_6_11
#define init_13_11_6_11
void FASTCALL call_13_11_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_6_12
#define init_13_11_6_12
void FASTCALL call_13_11_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_6_13
#define init_13_11_6_13
void FASTCALL call_13_11_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_6_14
#define init_13_11_6_14
void FASTCALL call_13_11_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_6_15
#define init_13_11_6_15
void FASTCALL call_13_11_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_7_0
#define init_13_11_7_0
void FASTCALL call_13_11_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_7_1
#define init_13_11_7_1
void FASTCALL call_13_11_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_7_2
#define init_13_11_7_2
void FASTCALL call_13_11_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_7_3
#define init_13_11_7_3
void FASTCALL call_13_11_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_7_4
#define init_13_11_7_4
void FASTCALL call_13_11_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_7_5
#define init_13_11_7_5
void FASTCALL call_13_11_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_7_6
#define init_13_11_7_6
void FASTCALL call_13_11_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_7_7
#define init_13_11_7_7
void FASTCALL call_13_11_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_7_8
#define init_13_11_7_8
void FASTCALL call_13_11_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_7_9
#define init_13_11_7_9
void FASTCALL call_13_11_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_7_10
#define init_13_11_7_10
void FASTCALL call_13_11_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_7_11
#define init_13_11_7_11
void FASTCALL call_13_11_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_7_12
#define init_13_11_7_12
void FASTCALL call_13_11_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_7_13
#define init_13_11_7_13
void FASTCALL call_13_11_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_7_14
#define init_13_11_7_14
void FASTCALL call_13_11_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_7_15
#define init_13_11_7_15
void FASTCALL call_13_11_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_8_0
#define init_13_11_8_0
void FASTCALL call_13_11_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_8_1
#define init_13_11_8_1
void FASTCALL call_13_11_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_8_2
#define init_13_11_8_2
void FASTCALL call_13_11_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_8_3
#define init_13_11_8_3
void FASTCALL call_13_11_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_8_4
#define init_13_11_8_4
void FASTCALL call_13_11_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_8_5
#define init_13_11_8_5
void FASTCALL call_13_11_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_8_6
#define init_13_11_8_6
void FASTCALL call_13_11_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_8_7
#define init_13_11_8_7
void FASTCALL call_13_11_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_8_8
#define init_13_11_8_8
void FASTCALL call_13_11_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_8_9
#define init_13_11_8_9
void FASTCALL call_13_11_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_8_10
#define init_13_11_8_10
void FASTCALL call_13_11_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_8_11
#define init_13_11_8_11
void FASTCALL call_13_11_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_8_12
#define init_13_11_8_12
void FASTCALL call_13_11_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_8_13
#define init_13_11_8_13
void FASTCALL call_13_11_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_8_14
#define init_13_11_8_14
void FASTCALL call_13_11_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_8_15
#define init_13_11_8_15
void FASTCALL call_13_11_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_9_0
#define init_13_11_9_0
void FASTCALL call_13_11_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_9_1
#define init_13_11_9_1
void FASTCALL call_13_11_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_9_2
#define init_13_11_9_2
void FASTCALL call_13_11_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_9_3
#define init_13_11_9_3
void FASTCALL call_13_11_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_9_4
#define init_13_11_9_4
void FASTCALL call_13_11_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_9_5
#define init_13_11_9_5
void FASTCALL call_13_11_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_9_6
#define init_13_11_9_6
void FASTCALL call_13_11_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_9_7
#define init_13_11_9_7
void FASTCALL call_13_11_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_9_8
#define init_13_11_9_8
void FASTCALL call_13_11_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_9_9
#define init_13_11_9_9
void FASTCALL call_13_11_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_9_10
#define init_13_11_9_10
void FASTCALL call_13_11_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_9_11
#define init_13_11_9_11
void FASTCALL call_13_11_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_9_12
#define init_13_11_9_12
void FASTCALL call_13_11_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_9_13
#define init_13_11_9_13
void FASTCALL call_13_11_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_9_14
#define init_13_11_9_14
void FASTCALL call_13_11_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_9_15
#define init_13_11_9_15
void FASTCALL call_13_11_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_10_0
#define init_13_11_10_0
void FASTCALL call_13_11_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_10_1
#define init_13_11_10_1
void FASTCALL call_13_11_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_10_2
#define init_13_11_10_2
void FASTCALL call_13_11_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_10_3
#define init_13_11_10_3
void FASTCALL call_13_11_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_10_4
#define init_13_11_10_4
void FASTCALL call_13_11_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_10_5
#define init_13_11_10_5
void FASTCALL call_13_11_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_10_6
#define init_13_11_10_6
void FASTCALL call_13_11_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_10_7
#define init_13_11_10_7
void FASTCALL call_13_11_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_10_8
#define init_13_11_10_8
void FASTCALL call_13_11_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_10_9
#define init_13_11_10_9
void FASTCALL call_13_11_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_10_10
#define init_13_11_10_10
void FASTCALL call_13_11_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_10_11
#define init_13_11_10_11
void FASTCALL call_13_11_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_10_12
#define init_13_11_10_12
void FASTCALL call_13_11_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_10_13
#define init_13_11_10_13
void FASTCALL call_13_11_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_10_14
#define init_13_11_10_14
void FASTCALL call_13_11_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_10_15
#define init_13_11_10_15
void FASTCALL call_13_11_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_11_0
#define init_13_11_11_0
void FASTCALL call_13_11_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_11_1
#define init_13_11_11_1
void FASTCALL call_13_11_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_11_2
#define init_13_11_11_2
void FASTCALL call_13_11_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_11_3
#define init_13_11_11_3
void FASTCALL call_13_11_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_11_4
#define init_13_11_11_4
void FASTCALL call_13_11_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_11_5
#define init_13_11_11_5
void FASTCALL call_13_11_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_11_6
#define init_13_11_11_6
void FASTCALL call_13_11_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_11_7
#define init_13_11_11_7
void FASTCALL call_13_11_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_11_8
#define init_13_11_11_8
void FASTCALL call_13_11_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_11_9
#define init_13_11_11_9
void FASTCALL call_13_11_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_11_10
#define init_13_11_11_10
void FASTCALL call_13_11_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_11_11
#define init_13_11_11_11
void FASTCALL call_13_11_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_11_12
#define init_13_11_11_12
void FASTCALL call_13_11_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_11_13
#define init_13_11_11_13
void FASTCALL call_13_11_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_11_14
#define init_13_11_11_14
void FASTCALL call_13_11_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_11_15
#define init_13_11_11_15
void FASTCALL call_13_11_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_12_0
#define init_13_11_12_0
void FASTCALL call_13_11_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_12_1
#define init_13_11_12_1
void FASTCALL call_13_11_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_12_2
#define init_13_11_12_2
void FASTCALL call_13_11_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_12_3
#define init_13_11_12_3
void FASTCALL call_13_11_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_12_4
#define init_13_11_12_4
void FASTCALL call_13_11_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_12_5
#define init_13_11_12_5
void FASTCALL call_13_11_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_12_6
#define init_13_11_12_6
void FASTCALL call_13_11_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_12_7
#define init_13_11_12_7
void FASTCALL call_13_11_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_12_8
#define init_13_11_12_8
void FASTCALL call_13_11_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_12_9
#define init_13_11_12_9
void FASTCALL call_13_11_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_12_10
#define init_13_11_12_10
void FASTCALL call_13_11_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_12_11
#define init_13_11_12_11
void FASTCALL call_13_11_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_12_12
#define init_13_11_12_12
void FASTCALL call_13_11_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_12_13
#define init_13_11_12_13
void FASTCALL call_13_11_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_12_14
#define init_13_11_12_14
void FASTCALL call_13_11_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_12_15
#define init_13_11_12_15
void FASTCALL call_13_11_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_13_0
#define init_13_11_13_0
void FASTCALL call_13_11_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_13_1
#define init_13_11_13_1
void FASTCALL call_13_11_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_13_2
#define init_13_11_13_2
void FASTCALL call_13_11_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_13_3
#define init_13_11_13_3
void FASTCALL call_13_11_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_13_4
#define init_13_11_13_4
void FASTCALL call_13_11_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_13_5
#define init_13_11_13_5
void FASTCALL call_13_11_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_13_6
#define init_13_11_13_6
void FASTCALL call_13_11_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_13_7
#define init_13_11_13_7
void FASTCALL call_13_11_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_13_8
#define init_13_11_13_8
void FASTCALL call_13_11_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_13_9
#define init_13_11_13_9
void FASTCALL call_13_11_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_13_10
#define init_13_11_13_10
void FASTCALL call_13_11_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_13_11
#define init_13_11_13_11
void FASTCALL call_13_11_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_13_12
#define init_13_11_13_12
void FASTCALL call_13_11_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_13_13
#define init_13_11_13_13
void FASTCALL call_13_11_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_13_14
#define init_13_11_13_14
void FASTCALL call_13_11_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_13_15
#define init_13_11_13_15
void FASTCALL call_13_11_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_14_0
#define init_13_11_14_0
void FASTCALL call_13_11_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_14_1
#define init_13_11_14_1
void FASTCALL call_13_11_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_14_2
#define init_13_11_14_2
void FASTCALL call_13_11_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_14_3
#define init_13_11_14_3
void FASTCALL call_13_11_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_14_4
#define init_13_11_14_4
void FASTCALL call_13_11_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_14_5
#define init_13_11_14_5
void FASTCALL call_13_11_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_14_6
#define init_13_11_14_6
void FASTCALL call_13_11_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_14_7
#define init_13_11_14_7
void FASTCALL call_13_11_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_14_8
#define init_13_11_14_8
void FASTCALL call_13_11_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_14_9
#define init_13_11_14_9
void FASTCALL call_13_11_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_14_10
#define init_13_11_14_10
void FASTCALL call_13_11_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_14_11
#define init_13_11_14_11
void FASTCALL call_13_11_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_14_12
#define init_13_11_14_12
void FASTCALL call_13_11_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_14_13
#define init_13_11_14_13
void FASTCALL call_13_11_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_14_14
#define init_13_11_14_14
void FASTCALL call_13_11_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_14_15
#define init_13_11_14_15
void FASTCALL call_13_11_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_15_0
#define init_13_11_15_0
void FASTCALL call_13_11_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_15_1
#define init_13_11_15_1
void FASTCALL call_13_11_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_15_2
#define init_13_11_15_2
void FASTCALL call_13_11_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_15_3
#define init_13_11_15_3
void FASTCALL call_13_11_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_15_4
#define init_13_11_15_4
void FASTCALL call_13_11_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_15_5
#define init_13_11_15_5
void FASTCALL call_13_11_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_15_6
#define init_13_11_15_6
void FASTCALL call_13_11_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_15_7
#define init_13_11_15_7
void FASTCALL call_13_11_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_15_8
#define init_13_11_15_8
void FASTCALL call_13_11_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_15_9
#define init_13_11_15_9
void FASTCALL call_13_11_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_15_10
#define init_13_11_15_10
void FASTCALL call_13_11_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_15_11
#define init_13_11_15_11
void FASTCALL call_13_11_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_15_12
#define init_13_11_15_12
void FASTCALL call_13_11_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_15_13
#define init_13_11_15_13
void FASTCALL call_13_11_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_15_14
#define init_13_11_15_14
void FASTCALL call_13_11_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_11_15_15
#define init_13_11_15_15
void FASTCALL call_13_11_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_0_0
#define init_13_12_0_0
void FASTCALL call_13_12_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_0_1
#define init_13_12_0_1
void FASTCALL call_13_12_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_0_2
#define init_13_12_0_2
void FASTCALL call_13_12_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_0_3
#define init_13_12_0_3
void FASTCALL call_13_12_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_0_4
#define init_13_12_0_4
void FASTCALL call_13_12_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_0_5
#define init_13_12_0_5
void FASTCALL call_13_12_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_0_6
#define init_13_12_0_6
void FASTCALL call_13_12_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_0_7
#define init_13_12_0_7
void FASTCALL call_13_12_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_0_8
#define init_13_12_0_8
void FASTCALL call_13_12_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_0_9
#define init_13_12_0_9
void FASTCALL call_13_12_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_0_10
#define init_13_12_0_10
void FASTCALL call_13_12_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_0_11
#define init_13_12_0_11
void FASTCALL call_13_12_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_0_12
#define init_13_12_0_12
void FASTCALL call_13_12_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_0_13
#define init_13_12_0_13
void FASTCALL call_13_12_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_0_14
#define init_13_12_0_14
void FASTCALL call_13_12_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_0_15
#define init_13_12_0_15
void FASTCALL call_13_12_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_1_0
#define init_13_12_1_0
void FASTCALL call_13_12_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_1_1
#define init_13_12_1_1
void FASTCALL call_13_12_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_1_2
#define init_13_12_1_2
void FASTCALL call_13_12_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_1_3
#define init_13_12_1_3
void FASTCALL call_13_12_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_1_4
#define init_13_12_1_4
void FASTCALL call_13_12_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_1_5
#define init_13_12_1_5
void FASTCALL call_13_12_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_1_6
#define init_13_12_1_6
void FASTCALL call_13_12_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_1_7
#define init_13_12_1_7
void FASTCALL call_13_12_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_1_8
#define init_13_12_1_8
void FASTCALL call_13_12_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_1_9
#define init_13_12_1_9
void FASTCALL call_13_12_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_1_10
#define init_13_12_1_10
void FASTCALL call_13_12_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_1_11
#define init_13_12_1_11
void FASTCALL call_13_12_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_1_12
#define init_13_12_1_12
void FASTCALL call_13_12_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_1_13
#define init_13_12_1_13
void FASTCALL call_13_12_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_1_14
#define init_13_12_1_14
void FASTCALL call_13_12_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_1_15
#define init_13_12_1_15
void FASTCALL call_13_12_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_2_0
#define init_13_12_2_0
void FASTCALL call_13_12_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_2_1
#define init_13_12_2_1
void FASTCALL call_13_12_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_2_2
#define init_13_12_2_2
void FASTCALL call_13_12_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_2_3
#define init_13_12_2_3
void FASTCALL call_13_12_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_2_4
#define init_13_12_2_4
void FASTCALL call_13_12_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_2_5
#define init_13_12_2_5
void FASTCALL call_13_12_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_2_6
#define init_13_12_2_6
void FASTCALL call_13_12_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_2_7
#define init_13_12_2_7
void FASTCALL call_13_12_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_2_8
#define init_13_12_2_8
void FASTCALL call_13_12_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_2_9
#define init_13_12_2_9
void FASTCALL call_13_12_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_2_10
#define init_13_12_2_10
void FASTCALL call_13_12_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_2_11
#define init_13_12_2_11
void FASTCALL call_13_12_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_2_12
#define init_13_12_2_12
void FASTCALL call_13_12_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_2_13
#define init_13_12_2_13
void FASTCALL call_13_12_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_2_14
#define init_13_12_2_14
void FASTCALL call_13_12_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_2_15
#define init_13_12_2_15
void FASTCALL call_13_12_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_3_0
#define init_13_12_3_0
void FASTCALL call_13_12_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_3_1
#define init_13_12_3_1
void FASTCALL call_13_12_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_3_2
#define init_13_12_3_2
void FASTCALL call_13_12_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_3_3
#define init_13_12_3_3
void FASTCALL call_13_12_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_3_4
#define init_13_12_3_4
void FASTCALL call_13_12_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_3_5
#define init_13_12_3_5
void FASTCALL call_13_12_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_3_6
#define init_13_12_3_6
void FASTCALL call_13_12_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_3_7
#define init_13_12_3_7
void FASTCALL call_13_12_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_3_8
#define init_13_12_3_8
void FASTCALL call_13_12_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_3_9
#define init_13_12_3_9
void FASTCALL call_13_12_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_3_10
#define init_13_12_3_10
void FASTCALL call_13_12_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_3_11
#define init_13_12_3_11
void FASTCALL call_13_12_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_3_12
#define init_13_12_3_12
void FASTCALL call_13_12_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_3_13
#define init_13_12_3_13
void FASTCALL call_13_12_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_3_14
#define init_13_12_3_14
void FASTCALL call_13_12_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_3_15
#define init_13_12_3_15
void FASTCALL call_13_12_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_4_0
#define init_13_12_4_0
void FASTCALL call_13_12_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_4_1
#define init_13_12_4_1
void FASTCALL call_13_12_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_4_2
#define init_13_12_4_2
void FASTCALL call_13_12_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_4_3
#define init_13_12_4_3
void FASTCALL call_13_12_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_4_4
#define init_13_12_4_4
void FASTCALL call_13_12_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_4_5
#define init_13_12_4_5
void FASTCALL call_13_12_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_4_6
#define init_13_12_4_6
void FASTCALL call_13_12_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_4_7
#define init_13_12_4_7
void FASTCALL call_13_12_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_4_8
#define init_13_12_4_8
void FASTCALL call_13_12_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_4_9
#define init_13_12_4_9
void FASTCALL call_13_12_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_4_10
#define init_13_12_4_10
void FASTCALL call_13_12_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_4_11
#define init_13_12_4_11
void FASTCALL call_13_12_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_4_12
#define init_13_12_4_12
void FASTCALL call_13_12_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_4_13
#define init_13_12_4_13
void FASTCALL call_13_12_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_4_14
#define init_13_12_4_14
void FASTCALL call_13_12_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_4_15
#define init_13_12_4_15
void FASTCALL call_13_12_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_5_0
#define init_13_12_5_0
void FASTCALL call_13_12_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_5_1
#define init_13_12_5_1
void FASTCALL call_13_12_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_5_2
#define init_13_12_5_2
void FASTCALL call_13_12_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_5_3
#define init_13_12_5_3
void FASTCALL call_13_12_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_5_4
#define init_13_12_5_4
void FASTCALL call_13_12_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_5_5
#define init_13_12_5_5
void FASTCALL call_13_12_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_5_6
#define init_13_12_5_6
void FASTCALL call_13_12_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_5_7
#define init_13_12_5_7
void FASTCALL call_13_12_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_5_8
#define init_13_12_5_8
void FASTCALL call_13_12_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_5_9
#define init_13_12_5_9
void FASTCALL call_13_12_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_5_10
#define init_13_12_5_10
void FASTCALL call_13_12_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_5_11
#define init_13_12_5_11
void FASTCALL call_13_12_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_5_12
#define init_13_12_5_12
void FASTCALL call_13_12_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_5_13
#define init_13_12_5_13
void FASTCALL call_13_12_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_5_14
#define init_13_12_5_14
void FASTCALL call_13_12_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_5_15
#define init_13_12_5_15
void FASTCALL call_13_12_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_6_0
#define init_13_12_6_0
void FASTCALL call_13_12_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_6_1
#define init_13_12_6_1
void FASTCALL call_13_12_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_6_2
#define init_13_12_6_2
void FASTCALL call_13_12_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_6_3
#define init_13_12_6_3
void FASTCALL call_13_12_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_6_4
#define init_13_12_6_4
void FASTCALL call_13_12_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_6_5
#define init_13_12_6_5
void FASTCALL call_13_12_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_6_6
#define init_13_12_6_6
void FASTCALL call_13_12_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_6_7
#define init_13_12_6_7
void FASTCALL call_13_12_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_6_8
#define init_13_12_6_8
void FASTCALL call_13_12_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_6_9
#define init_13_12_6_9
void FASTCALL call_13_12_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_6_10
#define init_13_12_6_10
void FASTCALL call_13_12_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_6_11
#define init_13_12_6_11
void FASTCALL call_13_12_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_6_12
#define init_13_12_6_12
void FASTCALL call_13_12_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_6_13
#define init_13_12_6_13
void FASTCALL call_13_12_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_6_14
#define init_13_12_6_14
void FASTCALL call_13_12_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_6_15
#define init_13_12_6_15
void FASTCALL call_13_12_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_7_0
#define init_13_12_7_0
void FASTCALL call_13_12_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_7_1
#define init_13_12_7_1
void FASTCALL call_13_12_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_7_2
#define init_13_12_7_2
void FASTCALL call_13_12_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_7_3
#define init_13_12_7_3
void FASTCALL call_13_12_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_7_4
#define init_13_12_7_4
void FASTCALL call_13_12_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_7_5
#define init_13_12_7_5
void FASTCALL call_13_12_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_7_6
#define init_13_12_7_6
void FASTCALL call_13_12_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_7_7
#define init_13_12_7_7
void FASTCALL call_13_12_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_7_8
#define init_13_12_7_8
void FASTCALL call_13_12_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_7_9
#define init_13_12_7_9
void FASTCALL call_13_12_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_7_10
#define init_13_12_7_10
void FASTCALL call_13_12_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_7_11
#define init_13_12_7_11
void FASTCALL call_13_12_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_7_12
#define init_13_12_7_12
void FASTCALL call_13_12_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_7_13
#define init_13_12_7_13
void FASTCALL call_13_12_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_7_14
#define init_13_12_7_14
void FASTCALL call_13_12_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_7_15
#define init_13_12_7_15
void FASTCALL call_13_12_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_8_0
#define init_13_12_8_0
void FASTCALL call_13_12_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_8_1
#define init_13_12_8_1
void FASTCALL call_13_12_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_8_2
#define init_13_12_8_2
void FASTCALL call_13_12_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_8_3
#define init_13_12_8_3
void FASTCALL call_13_12_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_8_4
#define init_13_12_8_4
void FASTCALL call_13_12_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_8_5
#define init_13_12_8_5
void FASTCALL call_13_12_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_8_6
#define init_13_12_8_6
void FASTCALL call_13_12_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_8_7
#define init_13_12_8_7
void FASTCALL call_13_12_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_8_8
#define init_13_12_8_8
void FASTCALL call_13_12_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_8_9
#define init_13_12_8_9
void FASTCALL call_13_12_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_8_10
#define init_13_12_8_10
void FASTCALL call_13_12_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_8_11
#define init_13_12_8_11
void FASTCALL call_13_12_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_8_12
#define init_13_12_8_12
void FASTCALL call_13_12_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_8_13
#define init_13_12_8_13
void FASTCALL call_13_12_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_8_14
#define init_13_12_8_14
void FASTCALL call_13_12_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_8_15
#define init_13_12_8_15
void FASTCALL call_13_12_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_9_0
#define init_13_12_9_0
void FASTCALL call_13_12_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_9_1
#define init_13_12_9_1
void FASTCALL call_13_12_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_9_2
#define init_13_12_9_2
void FASTCALL call_13_12_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_9_3
#define init_13_12_9_3
void FASTCALL call_13_12_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_9_4
#define init_13_12_9_4
void FASTCALL call_13_12_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_9_5
#define init_13_12_9_5
void FASTCALL call_13_12_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_9_6
#define init_13_12_9_6
void FASTCALL call_13_12_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_9_7
#define init_13_12_9_7
void FASTCALL call_13_12_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_9_8
#define init_13_12_9_8
void FASTCALL call_13_12_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_9_9
#define init_13_12_9_9
void FASTCALL call_13_12_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_9_10
#define init_13_12_9_10
void FASTCALL call_13_12_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_9_11
#define init_13_12_9_11
void FASTCALL call_13_12_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_9_12
#define init_13_12_9_12
void FASTCALL call_13_12_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_9_13
#define init_13_12_9_13
void FASTCALL call_13_12_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_9_14
#define init_13_12_9_14
void FASTCALL call_13_12_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_9_15
#define init_13_12_9_15
void FASTCALL call_13_12_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_10_0
#define init_13_12_10_0
void FASTCALL call_13_12_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_10_1
#define init_13_12_10_1
void FASTCALL call_13_12_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_10_2
#define init_13_12_10_2
void FASTCALL call_13_12_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_10_3
#define init_13_12_10_3
void FASTCALL call_13_12_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_10_4
#define init_13_12_10_4
void FASTCALL call_13_12_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_10_5
#define init_13_12_10_5
void FASTCALL call_13_12_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_10_6
#define init_13_12_10_6
void FASTCALL call_13_12_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_10_7
#define init_13_12_10_7
void FASTCALL call_13_12_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_10_8
#define init_13_12_10_8
void FASTCALL call_13_12_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_10_9
#define init_13_12_10_9
void FASTCALL call_13_12_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_10_10
#define init_13_12_10_10
void FASTCALL call_13_12_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_10_11
#define init_13_12_10_11
void FASTCALL call_13_12_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_10_12
#define init_13_12_10_12
void FASTCALL call_13_12_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_10_13
#define init_13_12_10_13
void FASTCALL call_13_12_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_10_14
#define init_13_12_10_14
void FASTCALL call_13_12_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_10_15
#define init_13_12_10_15
void FASTCALL call_13_12_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_11_0
#define init_13_12_11_0
void FASTCALL call_13_12_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_11_1
#define init_13_12_11_1
void FASTCALL call_13_12_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_11_2
#define init_13_12_11_2
void FASTCALL call_13_12_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_11_3
#define init_13_12_11_3
void FASTCALL call_13_12_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_11_4
#define init_13_12_11_4
void FASTCALL call_13_12_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_11_5
#define init_13_12_11_5
void FASTCALL call_13_12_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_11_6
#define init_13_12_11_6
void FASTCALL call_13_12_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_11_7
#define init_13_12_11_7
void FASTCALL call_13_12_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_11_8
#define init_13_12_11_8
void FASTCALL call_13_12_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_11_9
#define init_13_12_11_9
void FASTCALL call_13_12_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_11_10
#define init_13_12_11_10
void FASTCALL call_13_12_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_11_11
#define init_13_12_11_11
void FASTCALL call_13_12_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_11_12
#define init_13_12_11_12
void FASTCALL call_13_12_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_11_13
#define init_13_12_11_13
void FASTCALL call_13_12_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_11_14
#define init_13_12_11_14
void FASTCALL call_13_12_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_11_15
#define init_13_12_11_15
void FASTCALL call_13_12_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_12_0
#define init_13_12_12_0
void FASTCALL call_13_12_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_12_1
#define init_13_12_12_1
void FASTCALL call_13_12_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_12_2
#define init_13_12_12_2
void FASTCALL call_13_12_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_12_3
#define init_13_12_12_3
void FASTCALL call_13_12_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_12_4
#define init_13_12_12_4
void FASTCALL call_13_12_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_12_5
#define init_13_12_12_5
void FASTCALL call_13_12_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_12_6
#define init_13_12_12_6
void FASTCALL call_13_12_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_12_7
#define init_13_12_12_7
void FASTCALL call_13_12_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_12_8
#define init_13_12_12_8
void FASTCALL call_13_12_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_12_9
#define init_13_12_12_9
void FASTCALL call_13_12_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_12_10
#define init_13_12_12_10
void FASTCALL call_13_12_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_12_11
#define init_13_12_12_11
void FASTCALL call_13_12_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_12_12
#define init_13_12_12_12
void FASTCALL call_13_12_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_12_13
#define init_13_12_12_13
void FASTCALL call_13_12_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_12_14
#define init_13_12_12_14
void FASTCALL call_13_12_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_12_15
#define init_13_12_12_15
void FASTCALL call_13_12_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_13_0
#define init_13_12_13_0
void FASTCALL call_13_12_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_13_1
#define init_13_12_13_1
void FASTCALL call_13_12_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_13_2
#define init_13_12_13_2
void FASTCALL call_13_12_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_13_3
#define init_13_12_13_3
void FASTCALL call_13_12_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_13_4
#define init_13_12_13_4
void FASTCALL call_13_12_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_13_5
#define init_13_12_13_5
void FASTCALL call_13_12_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_13_6
#define init_13_12_13_6
void FASTCALL call_13_12_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_13_7
#define init_13_12_13_7
void FASTCALL call_13_12_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_13_8
#define init_13_12_13_8
void FASTCALL call_13_12_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_13_9
#define init_13_12_13_9
void FASTCALL call_13_12_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_13_10
#define init_13_12_13_10
void FASTCALL call_13_12_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_13_11
#define init_13_12_13_11
void FASTCALL call_13_12_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_13_12
#define init_13_12_13_12
void FASTCALL call_13_12_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_13_13
#define init_13_12_13_13
void FASTCALL call_13_12_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_13_14
#define init_13_12_13_14
void FASTCALL call_13_12_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_13_15
#define init_13_12_13_15
void FASTCALL call_13_12_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_14_0
#define init_13_12_14_0
void FASTCALL call_13_12_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_14_1
#define init_13_12_14_1
void FASTCALL call_13_12_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_14_2
#define init_13_12_14_2
void FASTCALL call_13_12_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_14_3
#define init_13_12_14_3
void FASTCALL call_13_12_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_14_4
#define init_13_12_14_4
void FASTCALL call_13_12_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_14_5
#define init_13_12_14_5
void FASTCALL call_13_12_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_14_6
#define init_13_12_14_6
void FASTCALL call_13_12_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_14_7
#define init_13_12_14_7
void FASTCALL call_13_12_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_14_8
#define init_13_12_14_8
void FASTCALL call_13_12_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_14_9
#define init_13_12_14_9
void FASTCALL call_13_12_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_14_10
#define init_13_12_14_10
void FASTCALL call_13_12_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_14_11
#define init_13_12_14_11
void FASTCALL call_13_12_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_14_12
#define init_13_12_14_12
void FASTCALL call_13_12_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_14_13
#define init_13_12_14_13
void FASTCALL call_13_12_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_14_14
#define init_13_12_14_14
void FASTCALL call_13_12_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_14_15
#define init_13_12_14_15
void FASTCALL call_13_12_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_15_0
#define init_13_12_15_0
void FASTCALL call_13_12_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_15_1
#define init_13_12_15_1
void FASTCALL call_13_12_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_15_2
#define init_13_12_15_2
void FASTCALL call_13_12_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_15_3
#define init_13_12_15_3
void FASTCALL call_13_12_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_15_4
#define init_13_12_15_4
void FASTCALL call_13_12_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_15_5
#define init_13_12_15_5
void FASTCALL call_13_12_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_15_6
#define init_13_12_15_6
void FASTCALL call_13_12_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_15_7
#define init_13_12_15_7
void FASTCALL call_13_12_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_15_8
#define init_13_12_15_8
void FASTCALL call_13_12_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_15_9
#define init_13_12_15_9
void FASTCALL call_13_12_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_15_10
#define init_13_12_15_10
void FASTCALL call_13_12_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_15_11
#define init_13_12_15_11
void FASTCALL call_13_12_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_15_12
#define init_13_12_15_12
void FASTCALL call_13_12_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_15_13
#define init_13_12_15_13
void FASTCALL call_13_12_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_15_14
#define init_13_12_15_14
void FASTCALL call_13_12_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_12_15_15
#define init_13_12_15_15
void FASTCALL call_13_12_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_0_0
#define init_13_13_0_0
void FASTCALL call_13_13_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_0_1
#define init_13_13_0_1
void FASTCALL call_13_13_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_0_2
#define init_13_13_0_2
void FASTCALL call_13_13_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_0_3
#define init_13_13_0_3
void FASTCALL call_13_13_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_0_4
#define init_13_13_0_4
void FASTCALL call_13_13_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_0_5
#define init_13_13_0_5
void FASTCALL call_13_13_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_0_6
#define init_13_13_0_6
void FASTCALL call_13_13_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_0_7
#define init_13_13_0_7
void FASTCALL call_13_13_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_0_8
#define init_13_13_0_8
void FASTCALL call_13_13_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_0_9
#define init_13_13_0_9
void FASTCALL call_13_13_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_0_10
#define init_13_13_0_10
void FASTCALL call_13_13_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_0_11
#define init_13_13_0_11
void FASTCALL call_13_13_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_0_12
#define init_13_13_0_12
void FASTCALL call_13_13_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_0_13
#define init_13_13_0_13
void FASTCALL call_13_13_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_0_14
#define init_13_13_0_14
void FASTCALL call_13_13_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_0_15
#define init_13_13_0_15
void FASTCALL call_13_13_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_1_0
#define init_13_13_1_0
void FASTCALL call_13_13_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_1_1
#define init_13_13_1_1
void FASTCALL call_13_13_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_1_2
#define init_13_13_1_2
void FASTCALL call_13_13_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_1_3
#define init_13_13_1_3
void FASTCALL call_13_13_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_1_4
#define init_13_13_1_4
void FASTCALL call_13_13_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_1_5
#define init_13_13_1_5
void FASTCALL call_13_13_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_1_6
#define init_13_13_1_6
void FASTCALL call_13_13_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_1_7
#define init_13_13_1_7
void FASTCALL call_13_13_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_1_8
#define init_13_13_1_8
void FASTCALL call_13_13_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_1_9
#define init_13_13_1_9
void FASTCALL call_13_13_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_1_10
#define init_13_13_1_10
void FASTCALL call_13_13_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_1_11
#define init_13_13_1_11
void FASTCALL call_13_13_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_1_12
#define init_13_13_1_12
void FASTCALL call_13_13_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_1_13
#define init_13_13_1_13
void FASTCALL call_13_13_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_1_14
#define init_13_13_1_14
void FASTCALL call_13_13_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_1_15
#define init_13_13_1_15
void FASTCALL call_13_13_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_2_0
#define init_13_13_2_0
void FASTCALL call_13_13_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_2_1
#define init_13_13_2_1
void FASTCALL call_13_13_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_2_2
#define init_13_13_2_2
void FASTCALL call_13_13_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_2_3
#define init_13_13_2_3
void FASTCALL call_13_13_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_2_4
#define init_13_13_2_4
void FASTCALL call_13_13_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_2_5
#define init_13_13_2_5
void FASTCALL call_13_13_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_2_6
#define init_13_13_2_6
void FASTCALL call_13_13_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_2_7
#define init_13_13_2_7
void FASTCALL call_13_13_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_2_8
#define init_13_13_2_8
void FASTCALL call_13_13_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_2_9
#define init_13_13_2_9
void FASTCALL call_13_13_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_2_10
#define init_13_13_2_10
void FASTCALL call_13_13_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_2_11
#define init_13_13_2_11
void FASTCALL call_13_13_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_2_12
#define init_13_13_2_12
void FASTCALL call_13_13_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_2_13
#define init_13_13_2_13
void FASTCALL call_13_13_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_2_14
#define init_13_13_2_14
void FASTCALL call_13_13_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_2_15
#define init_13_13_2_15
void FASTCALL call_13_13_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_3_0
#define init_13_13_3_0
void FASTCALL call_13_13_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_3_1
#define init_13_13_3_1
void FASTCALL call_13_13_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_3_2
#define init_13_13_3_2
void FASTCALL call_13_13_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_3_3
#define init_13_13_3_3
void FASTCALL call_13_13_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_3_4
#define init_13_13_3_4
void FASTCALL call_13_13_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_3_5
#define init_13_13_3_5
void FASTCALL call_13_13_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_3_6
#define init_13_13_3_6
void FASTCALL call_13_13_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_3_7
#define init_13_13_3_7
void FASTCALL call_13_13_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_3_8
#define init_13_13_3_8
void FASTCALL call_13_13_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_3_9
#define init_13_13_3_9
void FASTCALL call_13_13_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_3_10
#define init_13_13_3_10
void FASTCALL call_13_13_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_3_11
#define init_13_13_3_11
void FASTCALL call_13_13_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_3_12
#define init_13_13_3_12
void FASTCALL call_13_13_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_3_13
#define init_13_13_3_13
void FASTCALL call_13_13_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_3_14
#define init_13_13_3_14
void FASTCALL call_13_13_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_3_15
#define init_13_13_3_15
void FASTCALL call_13_13_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_4_0
#define init_13_13_4_0
void FASTCALL call_13_13_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_4_1
#define init_13_13_4_1
void FASTCALL call_13_13_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_4_2
#define init_13_13_4_2
void FASTCALL call_13_13_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_4_3
#define init_13_13_4_3
void FASTCALL call_13_13_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_4_4
#define init_13_13_4_4
void FASTCALL call_13_13_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_4_5
#define init_13_13_4_5
void FASTCALL call_13_13_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_4_6
#define init_13_13_4_6
void FASTCALL call_13_13_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_4_7
#define init_13_13_4_7
void FASTCALL call_13_13_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_4_8
#define init_13_13_4_8
void FASTCALL call_13_13_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_4_9
#define init_13_13_4_9
void FASTCALL call_13_13_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_4_10
#define init_13_13_4_10
void FASTCALL call_13_13_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_4_11
#define init_13_13_4_11
void FASTCALL call_13_13_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_4_12
#define init_13_13_4_12
void FASTCALL call_13_13_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_4_13
#define init_13_13_4_13
void FASTCALL call_13_13_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_4_14
#define init_13_13_4_14
void FASTCALL call_13_13_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_4_15
#define init_13_13_4_15
void FASTCALL call_13_13_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_5_0
#define init_13_13_5_0
void FASTCALL call_13_13_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_5_1
#define init_13_13_5_1
void FASTCALL call_13_13_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_5_2
#define init_13_13_5_2
void FASTCALL call_13_13_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_5_3
#define init_13_13_5_3
void FASTCALL call_13_13_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_5_4
#define init_13_13_5_4
void FASTCALL call_13_13_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_5_5
#define init_13_13_5_5
void FASTCALL call_13_13_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_5_6
#define init_13_13_5_6
void FASTCALL call_13_13_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_5_7
#define init_13_13_5_7
void FASTCALL call_13_13_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_5_8
#define init_13_13_5_8
void FASTCALL call_13_13_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_5_9
#define init_13_13_5_9
void FASTCALL call_13_13_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_5_10
#define init_13_13_5_10
void FASTCALL call_13_13_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_5_11
#define init_13_13_5_11
void FASTCALL call_13_13_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_5_12
#define init_13_13_5_12
void FASTCALL call_13_13_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_5_13
#define init_13_13_5_13
void FASTCALL call_13_13_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_5_14
#define init_13_13_5_14
void FASTCALL call_13_13_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_5_15
#define init_13_13_5_15
void FASTCALL call_13_13_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_6_0
#define init_13_13_6_0
void FASTCALL call_13_13_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_6_1
#define init_13_13_6_1
void FASTCALL call_13_13_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_6_2
#define init_13_13_6_2
void FASTCALL call_13_13_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_6_3
#define init_13_13_6_3
void FASTCALL call_13_13_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_6_4
#define init_13_13_6_4
void FASTCALL call_13_13_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_6_5
#define init_13_13_6_5
void FASTCALL call_13_13_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_6_6
#define init_13_13_6_6
void FASTCALL call_13_13_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_6_7
#define init_13_13_6_7
void FASTCALL call_13_13_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_6_8
#define init_13_13_6_8
void FASTCALL call_13_13_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_6_9
#define init_13_13_6_9
void FASTCALL call_13_13_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_6_10
#define init_13_13_6_10
void FASTCALL call_13_13_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_6_11
#define init_13_13_6_11
void FASTCALL call_13_13_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_6_12
#define init_13_13_6_12
void FASTCALL call_13_13_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_6_13
#define init_13_13_6_13
void FASTCALL call_13_13_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_6_14
#define init_13_13_6_14
void FASTCALL call_13_13_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_6_15
#define init_13_13_6_15
void FASTCALL call_13_13_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_7_0
#define init_13_13_7_0
void FASTCALL call_13_13_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_7_1
#define init_13_13_7_1
void FASTCALL call_13_13_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_7_2
#define init_13_13_7_2
void FASTCALL call_13_13_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_7_3
#define init_13_13_7_3
void FASTCALL call_13_13_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_7_4
#define init_13_13_7_4
void FASTCALL call_13_13_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_7_5
#define init_13_13_7_5
void FASTCALL call_13_13_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_7_6
#define init_13_13_7_6
void FASTCALL call_13_13_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_7_7
#define init_13_13_7_7
void FASTCALL call_13_13_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_7_8
#define init_13_13_7_8
void FASTCALL call_13_13_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_7_9
#define init_13_13_7_9
void FASTCALL call_13_13_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_7_10
#define init_13_13_7_10
void FASTCALL call_13_13_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_7_11
#define init_13_13_7_11
void FASTCALL call_13_13_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_7_12
#define init_13_13_7_12
void FASTCALL call_13_13_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_7_13
#define init_13_13_7_13
void FASTCALL call_13_13_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_7_14
#define init_13_13_7_14
void FASTCALL call_13_13_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_7_15
#define init_13_13_7_15
void FASTCALL call_13_13_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_8_0
#define init_13_13_8_0
void FASTCALL call_13_13_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_8_1
#define init_13_13_8_1
void FASTCALL call_13_13_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_8_2
#define init_13_13_8_2
void FASTCALL call_13_13_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_8_3
#define init_13_13_8_3
void FASTCALL call_13_13_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_8_4
#define init_13_13_8_4
void FASTCALL call_13_13_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_8_5
#define init_13_13_8_5
void FASTCALL call_13_13_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_8_6
#define init_13_13_8_6
void FASTCALL call_13_13_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_8_7
#define init_13_13_8_7
void FASTCALL call_13_13_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_8_8
#define init_13_13_8_8
void FASTCALL call_13_13_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_8_9
#define init_13_13_8_9
void FASTCALL call_13_13_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_8_10
#define init_13_13_8_10
void FASTCALL call_13_13_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_8_11
#define init_13_13_8_11
void FASTCALL call_13_13_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_8_12
#define init_13_13_8_12
void FASTCALL call_13_13_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_8_13
#define init_13_13_8_13
void FASTCALL call_13_13_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_8_14
#define init_13_13_8_14
void FASTCALL call_13_13_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_8_15
#define init_13_13_8_15
void FASTCALL call_13_13_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_9_0
#define init_13_13_9_0
void FASTCALL call_13_13_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_9_1
#define init_13_13_9_1
void FASTCALL call_13_13_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_9_2
#define init_13_13_9_2
void FASTCALL call_13_13_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_9_3
#define init_13_13_9_3
void FASTCALL call_13_13_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_9_4
#define init_13_13_9_4
void FASTCALL call_13_13_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_9_5
#define init_13_13_9_5
void FASTCALL call_13_13_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_9_6
#define init_13_13_9_6
void FASTCALL call_13_13_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_9_7
#define init_13_13_9_7
void FASTCALL call_13_13_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_9_8
#define init_13_13_9_8
void FASTCALL call_13_13_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_9_9
#define init_13_13_9_9
void FASTCALL call_13_13_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_9_10
#define init_13_13_9_10
void FASTCALL call_13_13_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_9_11
#define init_13_13_9_11
void FASTCALL call_13_13_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_9_12
#define init_13_13_9_12
void FASTCALL call_13_13_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_9_13
#define init_13_13_9_13
void FASTCALL call_13_13_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_9_14
#define init_13_13_9_14
void FASTCALL call_13_13_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_9_15
#define init_13_13_9_15
void FASTCALL call_13_13_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_10_0
#define init_13_13_10_0
void FASTCALL call_13_13_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_10_1
#define init_13_13_10_1
void FASTCALL call_13_13_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_10_2
#define init_13_13_10_2
void FASTCALL call_13_13_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_10_3
#define init_13_13_10_3
void FASTCALL call_13_13_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_10_4
#define init_13_13_10_4
void FASTCALL call_13_13_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_10_5
#define init_13_13_10_5
void FASTCALL call_13_13_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_10_6
#define init_13_13_10_6
void FASTCALL call_13_13_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_10_7
#define init_13_13_10_7
void FASTCALL call_13_13_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_10_8
#define init_13_13_10_8
void FASTCALL call_13_13_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_10_9
#define init_13_13_10_9
void FASTCALL call_13_13_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_10_10
#define init_13_13_10_10
void FASTCALL call_13_13_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_10_11
#define init_13_13_10_11
void FASTCALL call_13_13_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_10_12
#define init_13_13_10_12
void FASTCALL call_13_13_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_10_13
#define init_13_13_10_13
void FASTCALL call_13_13_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_10_14
#define init_13_13_10_14
void FASTCALL call_13_13_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_10_15
#define init_13_13_10_15
void FASTCALL call_13_13_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_11_0
#define init_13_13_11_0
void FASTCALL call_13_13_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_11_1
#define init_13_13_11_1
void FASTCALL call_13_13_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_11_2
#define init_13_13_11_2
void FASTCALL call_13_13_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_11_3
#define init_13_13_11_3
void FASTCALL call_13_13_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_11_4
#define init_13_13_11_4
void FASTCALL call_13_13_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_11_5
#define init_13_13_11_5
void FASTCALL call_13_13_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_11_6
#define init_13_13_11_6
void FASTCALL call_13_13_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_11_7
#define init_13_13_11_7
void FASTCALL call_13_13_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_11_8
#define init_13_13_11_8
void FASTCALL call_13_13_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_11_9
#define init_13_13_11_9
void FASTCALL call_13_13_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_11_10
#define init_13_13_11_10
void FASTCALL call_13_13_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_11_11
#define init_13_13_11_11
void FASTCALL call_13_13_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_11_12
#define init_13_13_11_12
void FASTCALL call_13_13_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_11_13
#define init_13_13_11_13
void FASTCALL call_13_13_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_11_14
#define init_13_13_11_14
void FASTCALL call_13_13_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_11_15
#define init_13_13_11_15
void FASTCALL call_13_13_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_12_0
#define init_13_13_12_0
void FASTCALL call_13_13_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_12_1
#define init_13_13_12_1
void FASTCALL call_13_13_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_12_2
#define init_13_13_12_2
void FASTCALL call_13_13_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_12_3
#define init_13_13_12_3
void FASTCALL call_13_13_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_12_4
#define init_13_13_12_4
void FASTCALL call_13_13_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_12_5
#define init_13_13_12_5
void FASTCALL call_13_13_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_12_6
#define init_13_13_12_6
void FASTCALL call_13_13_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_12_7
#define init_13_13_12_7
void FASTCALL call_13_13_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_12_8
#define init_13_13_12_8
void FASTCALL call_13_13_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_12_9
#define init_13_13_12_9
void FASTCALL call_13_13_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_12_10
#define init_13_13_12_10
void FASTCALL call_13_13_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_12_11
#define init_13_13_12_11
void FASTCALL call_13_13_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_12_12
#define init_13_13_12_12
void FASTCALL call_13_13_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_12_13
#define init_13_13_12_13
void FASTCALL call_13_13_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_12_14
#define init_13_13_12_14
void FASTCALL call_13_13_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_12_15
#define init_13_13_12_15
void FASTCALL call_13_13_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_13_0
#define init_13_13_13_0
void FASTCALL call_13_13_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_13_1
#define init_13_13_13_1
void FASTCALL call_13_13_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_13_2
#define init_13_13_13_2
void FASTCALL call_13_13_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_13_3
#define init_13_13_13_3
void FASTCALL call_13_13_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_13_4
#define init_13_13_13_4
void FASTCALL call_13_13_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_13_5
#define init_13_13_13_5
void FASTCALL call_13_13_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_13_6
#define init_13_13_13_6
void FASTCALL call_13_13_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_13_7
#define init_13_13_13_7
void FASTCALL call_13_13_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_13_8
#define init_13_13_13_8
void FASTCALL call_13_13_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_13_9
#define init_13_13_13_9
void FASTCALL call_13_13_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_13_10
#define init_13_13_13_10
void FASTCALL call_13_13_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_13_11
#define init_13_13_13_11
void FASTCALL call_13_13_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_13_12
#define init_13_13_13_12
void FASTCALL call_13_13_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_13_13
#define init_13_13_13_13
void FASTCALL call_13_13_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_13_14
#define init_13_13_13_14
void FASTCALL call_13_13_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_13_15
#define init_13_13_13_15
void FASTCALL call_13_13_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_14_0
#define init_13_13_14_0
void FASTCALL call_13_13_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_14_1
#define init_13_13_14_1
void FASTCALL call_13_13_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_14_2
#define init_13_13_14_2
void FASTCALL call_13_13_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_14_3
#define init_13_13_14_3
void FASTCALL call_13_13_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_14_4
#define init_13_13_14_4
void FASTCALL call_13_13_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_14_5
#define init_13_13_14_5
void FASTCALL call_13_13_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_14_6
#define init_13_13_14_6
void FASTCALL call_13_13_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_14_7
#define init_13_13_14_7
void FASTCALL call_13_13_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_14_8
#define init_13_13_14_8
void FASTCALL call_13_13_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_14_9
#define init_13_13_14_9
void FASTCALL call_13_13_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_14_10
#define init_13_13_14_10
void FASTCALL call_13_13_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_14_11
#define init_13_13_14_11
void FASTCALL call_13_13_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_14_12
#define init_13_13_14_12
void FASTCALL call_13_13_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_14_13
#define init_13_13_14_13
void FASTCALL call_13_13_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_14_14
#define init_13_13_14_14
void FASTCALL call_13_13_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_14_15
#define init_13_13_14_15
void FASTCALL call_13_13_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_15_0
#define init_13_13_15_0
void FASTCALL call_13_13_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_15_1
#define init_13_13_15_1
void FASTCALL call_13_13_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_15_2
#define init_13_13_15_2
void FASTCALL call_13_13_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_15_3
#define init_13_13_15_3
void FASTCALL call_13_13_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_15_4
#define init_13_13_15_4
void FASTCALL call_13_13_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_15_5
#define init_13_13_15_5
void FASTCALL call_13_13_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_15_6
#define init_13_13_15_6
void FASTCALL call_13_13_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_15_7
#define init_13_13_15_7
void FASTCALL call_13_13_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_15_8
#define init_13_13_15_8
void FASTCALL call_13_13_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_15_9
#define init_13_13_15_9
void FASTCALL call_13_13_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_15_10
#define init_13_13_15_10
void FASTCALL call_13_13_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_15_11
#define init_13_13_15_11
void FASTCALL call_13_13_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_15_12
#define init_13_13_15_12
void FASTCALL call_13_13_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_15_13
#define init_13_13_15_13
void FASTCALL call_13_13_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_15_14
#define init_13_13_15_14
void FASTCALL call_13_13_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_13_15_15
#define init_13_13_15_15
void FASTCALL call_13_13_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_0_0
#define init_13_14_0_0
void FASTCALL call_13_14_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_0_1
#define init_13_14_0_1
void FASTCALL call_13_14_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_0_2
#define init_13_14_0_2
void FASTCALL call_13_14_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_0_3
#define init_13_14_0_3
void FASTCALL call_13_14_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_0_4
#define init_13_14_0_4
void FASTCALL call_13_14_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_0_5
#define init_13_14_0_5
void FASTCALL call_13_14_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_0_6
#define init_13_14_0_6
void FASTCALL call_13_14_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_0_7
#define init_13_14_0_7
void FASTCALL call_13_14_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_0_8
#define init_13_14_0_8
void FASTCALL call_13_14_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_0_9
#define init_13_14_0_9
void FASTCALL call_13_14_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_0_10
#define init_13_14_0_10
void FASTCALL call_13_14_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_0_11
#define init_13_14_0_11
void FASTCALL call_13_14_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_0_12
#define init_13_14_0_12
void FASTCALL call_13_14_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_0_13
#define init_13_14_0_13
void FASTCALL call_13_14_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_0_14
#define init_13_14_0_14
void FASTCALL call_13_14_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_0_15
#define init_13_14_0_15
void FASTCALL call_13_14_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_1_0
#define init_13_14_1_0
void FASTCALL call_13_14_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_1_1
#define init_13_14_1_1
void FASTCALL call_13_14_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_1_2
#define init_13_14_1_2
void FASTCALL call_13_14_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_1_3
#define init_13_14_1_3
void FASTCALL call_13_14_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_1_4
#define init_13_14_1_4
void FASTCALL call_13_14_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_1_5
#define init_13_14_1_5
void FASTCALL call_13_14_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_1_6
#define init_13_14_1_6
void FASTCALL call_13_14_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_1_7
#define init_13_14_1_7
void FASTCALL call_13_14_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_1_8
#define init_13_14_1_8
void FASTCALL call_13_14_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_1_9
#define init_13_14_1_9
void FASTCALL call_13_14_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_1_10
#define init_13_14_1_10
void FASTCALL call_13_14_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_1_11
#define init_13_14_1_11
void FASTCALL call_13_14_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_1_12
#define init_13_14_1_12
void FASTCALL call_13_14_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_1_13
#define init_13_14_1_13
void FASTCALL call_13_14_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_1_14
#define init_13_14_1_14
void FASTCALL call_13_14_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_1_15
#define init_13_14_1_15
void FASTCALL call_13_14_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_2_0
#define init_13_14_2_0
void FASTCALL call_13_14_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_2_1
#define init_13_14_2_1
void FASTCALL call_13_14_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_2_2
#define init_13_14_2_2
void FASTCALL call_13_14_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_2_3
#define init_13_14_2_3
void FASTCALL call_13_14_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_2_4
#define init_13_14_2_4
void FASTCALL call_13_14_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_2_5
#define init_13_14_2_5
void FASTCALL call_13_14_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_2_6
#define init_13_14_2_6
void FASTCALL call_13_14_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_2_7
#define init_13_14_2_7
void FASTCALL call_13_14_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_2_8
#define init_13_14_2_8
void FASTCALL call_13_14_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_2_9
#define init_13_14_2_9
void FASTCALL call_13_14_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_2_10
#define init_13_14_2_10
void FASTCALL call_13_14_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_2_11
#define init_13_14_2_11
void FASTCALL call_13_14_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_2_12
#define init_13_14_2_12
void FASTCALL call_13_14_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_2_13
#define init_13_14_2_13
void FASTCALL call_13_14_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_2_14
#define init_13_14_2_14
void FASTCALL call_13_14_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_2_15
#define init_13_14_2_15
void FASTCALL call_13_14_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_3_0
#define init_13_14_3_0
void FASTCALL call_13_14_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_3_1
#define init_13_14_3_1
void FASTCALL call_13_14_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_3_2
#define init_13_14_3_2
void FASTCALL call_13_14_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_3_3
#define init_13_14_3_3
void FASTCALL call_13_14_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_3_4
#define init_13_14_3_4
void FASTCALL call_13_14_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_3_5
#define init_13_14_3_5
void FASTCALL call_13_14_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_3_6
#define init_13_14_3_6
void FASTCALL call_13_14_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_3_7
#define init_13_14_3_7
void FASTCALL call_13_14_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_3_8
#define init_13_14_3_8
void FASTCALL call_13_14_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_3_9
#define init_13_14_3_9
void FASTCALL call_13_14_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_3_10
#define init_13_14_3_10
void FASTCALL call_13_14_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_3_11
#define init_13_14_3_11
void FASTCALL call_13_14_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_3_12
#define init_13_14_3_12
void FASTCALL call_13_14_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_3_13
#define init_13_14_3_13
void FASTCALL call_13_14_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_3_14
#define init_13_14_3_14
void FASTCALL call_13_14_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_3_15
#define init_13_14_3_15
void FASTCALL call_13_14_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_4_0
#define init_13_14_4_0
void FASTCALL call_13_14_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_4_1
#define init_13_14_4_1
void FASTCALL call_13_14_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_4_2
#define init_13_14_4_2
void FASTCALL call_13_14_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_4_3
#define init_13_14_4_3
void FASTCALL call_13_14_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_4_4
#define init_13_14_4_4
void FASTCALL call_13_14_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_4_5
#define init_13_14_4_5
void FASTCALL call_13_14_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_4_6
#define init_13_14_4_6
void FASTCALL call_13_14_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_4_7
#define init_13_14_4_7
void FASTCALL call_13_14_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_4_8
#define init_13_14_4_8
void FASTCALL call_13_14_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_4_9
#define init_13_14_4_9
void FASTCALL call_13_14_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_4_10
#define init_13_14_4_10
void FASTCALL call_13_14_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_4_11
#define init_13_14_4_11
void FASTCALL call_13_14_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_4_12
#define init_13_14_4_12
void FASTCALL call_13_14_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_4_13
#define init_13_14_4_13
void FASTCALL call_13_14_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_4_14
#define init_13_14_4_14
void FASTCALL call_13_14_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_4_15
#define init_13_14_4_15
void FASTCALL call_13_14_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_5_0
#define init_13_14_5_0
void FASTCALL call_13_14_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_5_1
#define init_13_14_5_1
void FASTCALL call_13_14_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_5_2
#define init_13_14_5_2
void FASTCALL call_13_14_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_5_3
#define init_13_14_5_3
void FASTCALL call_13_14_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_5_4
#define init_13_14_5_4
void FASTCALL call_13_14_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_5_5
#define init_13_14_5_5
void FASTCALL call_13_14_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_5_6
#define init_13_14_5_6
void FASTCALL call_13_14_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_5_7
#define init_13_14_5_7
void FASTCALL call_13_14_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_5_8
#define init_13_14_5_8
void FASTCALL call_13_14_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_5_9
#define init_13_14_5_9
void FASTCALL call_13_14_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_5_10
#define init_13_14_5_10
void FASTCALL call_13_14_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_5_11
#define init_13_14_5_11
void FASTCALL call_13_14_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_5_12
#define init_13_14_5_12
void FASTCALL call_13_14_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_5_13
#define init_13_14_5_13
void FASTCALL call_13_14_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_5_14
#define init_13_14_5_14
void FASTCALL call_13_14_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_5_15
#define init_13_14_5_15
void FASTCALL call_13_14_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_6_0
#define init_13_14_6_0
void FASTCALL call_13_14_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_6_1
#define init_13_14_6_1
void FASTCALL call_13_14_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_6_2
#define init_13_14_6_2
void FASTCALL call_13_14_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_6_3
#define init_13_14_6_3
void FASTCALL call_13_14_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_6_4
#define init_13_14_6_4
void FASTCALL call_13_14_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_6_5
#define init_13_14_6_5
void FASTCALL call_13_14_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_6_6
#define init_13_14_6_6
void FASTCALL call_13_14_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_6_7
#define init_13_14_6_7
void FASTCALL call_13_14_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_6_8
#define init_13_14_6_8
void FASTCALL call_13_14_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_6_9
#define init_13_14_6_9
void FASTCALL call_13_14_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_6_10
#define init_13_14_6_10
void FASTCALL call_13_14_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_6_11
#define init_13_14_6_11
void FASTCALL call_13_14_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_6_12
#define init_13_14_6_12
void FASTCALL call_13_14_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_6_13
#define init_13_14_6_13
void FASTCALL call_13_14_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_6_14
#define init_13_14_6_14
void FASTCALL call_13_14_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_6_15
#define init_13_14_6_15
void FASTCALL call_13_14_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_7_0
#define init_13_14_7_0
void FASTCALL call_13_14_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_7_1
#define init_13_14_7_1
void FASTCALL call_13_14_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_7_2
#define init_13_14_7_2
void FASTCALL call_13_14_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_7_3
#define init_13_14_7_3
void FASTCALL call_13_14_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_7_4
#define init_13_14_7_4
void FASTCALL call_13_14_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_7_5
#define init_13_14_7_5
void FASTCALL call_13_14_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_7_6
#define init_13_14_7_6
void FASTCALL call_13_14_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_7_7
#define init_13_14_7_7
void FASTCALL call_13_14_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_7_8
#define init_13_14_7_8
void FASTCALL call_13_14_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_7_9
#define init_13_14_7_9
void FASTCALL call_13_14_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_7_10
#define init_13_14_7_10
void FASTCALL call_13_14_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_7_11
#define init_13_14_7_11
void FASTCALL call_13_14_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_7_12
#define init_13_14_7_12
void FASTCALL call_13_14_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_7_13
#define init_13_14_7_13
void FASTCALL call_13_14_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_7_14
#define init_13_14_7_14
void FASTCALL call_13_14_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_7_15
#define init_13_14_7_15
void FASTCALL call_13_14_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_8_0
#define init_13_14_8_0
void FASTCALL call_13_14_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_8_1
#define init_13_14_8_1
void FASTCALL call_13_14_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_8_2
#define init_13_14_8_2
void FASTCALL call_13_14_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_8_3
#define init_13_14_8_3
void FASTCALL call_13_14_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_8_4
#define init_13_14_8_4
void FASTCALL call_13_14_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_8_5
#define init_13_14_8_5
void FASTCALL call_13_14_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_8_6
#define init_13_14_8_6
void FASTCALL call_13_14_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_8_7
#define init_13_14_8_7
void FASTCALL call_13_14_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_8_8
#define init_13_14_8_8
void FASTCALL call_13_14_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_8_9
#define init_13_14_8_9
void FASTCALL call_13_14_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_8_10
#define init_13_14_8_10
void FASTCALL call_13_14_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_8_11
#define init_13_14_8_11
void FASTCALL call_13_14_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_8_12
#define init_13_14_8_12
void FASTCALL call_13_14_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_8_13
#define init_13_14_8_13
void FASTCALL call_13_14_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_8_14
#define init_13_14_8_14
void FASTCALL call_13_14_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_8_15
#define init_13_14_8_15
void FASTCALL call_13_14_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_9_0
#define init_13_14_9_0
void FASTCALL call_13_14_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_9_1
#define init_13_14_9_1
void FASTCALL call_13_14_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_9_2
#define init_13_14_9_2
void FASTCALL call_13_14_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_9_3
#define init_13_14_9_3
void FASTCALL call_13_14_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_9_4
#define init_13_14_9_4
void FASTCALL call_13_14_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_9_5
#define init_13_14_9_5
void FASTCALL call_13_14_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_9_6
#define init_13_14_9_6
void FASTCALL call_13_14_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_9_7
#define init_13_14_9_7
void FASTCALL call_13_14_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_9_8
#define init_13_14_9_8
void FASTCALL call_13_14_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_9_9
#define init_13_14_9_9
void FASTCALL call_13_14_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_9_10
#define init_13_14_9_10
void FASTCALL call_13_14_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_9_11
#define init_13_14_9_11
void FASTCALL call_13_14_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_9_12
#define init_13_14_9_12
void FASTCALL call_13_14_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_9_13
#define init_13_14_9_13
void FASTCALL call_13_14_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_9_14
#define init_13_14_9_14
void FASTCALL call_13_14_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_9_15
#define init_13_14_9_15
void FASTCALL call_13_14_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_10_0
#define init_13_14_10_0
void FASTCALL call_13_14_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_10_1
#define init_13_14_10_1
void FASTCALL call_13_14_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_10_2
#define init_13_14_10_2
void FASTCALL call_13_14_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_10_3
#define init_13_14_10_3
void FASTCALL call_13_14_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_10_4
#define init_13_14_10_4
void FASTCALL call_13_14_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_10_5
#define init_13_14_10_5
void FASTCALL call_13_14_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_10_6
#define init_13_14_10_6
void FASTCALL call_13_14_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_10_7
#define init_13_14_10_7
void FASTCALL call_13_14_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_10_8
#define init_13_14_10_8
void FASTCALL call_13_14_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_10_9
#define init_13_14_10_9
void FASTCALL call_13_14_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_10_10
#define init_13_14_10_10
void FASTCALL call_13_14_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_10_11
#define init_13_14_10_11
void FASTCALL call_13_14_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_10_12
#define init_13_14_10_12
void FASTCALL call_13_14_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_10_13
#define init_13_14_10_13
void FASTCALL call_13_14_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_10_14
#define init_13_14_10_14
void FASTCALL call_13_14_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_10_15
#define init_13_14_10_15
void FASTCALL call_13_14_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_11_0
#define init_13_14_11_0
void FASTCALL call_13_14_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_11_1
#define init_13_14_11_1
void FASTCALL call_13_14_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_11_2
#define init_13_14_11_2
void FASTCALL call_13_14_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_11_3
#define init_13_14_11_3
void FASTCALL call_13_14_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_11_4
#define init_13_14_11_4
void FASTCALL call_13_14_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_11_5
#define init_13_14_11_5
void FASTCALL call_13_14_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_11_6
#define init_13_14_11_6
void FASTCALL call_13_14_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_11_7
#define init_13_14_11_7
void FASTCALL call_13_14_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_11_8
#define init_13_14_11_8
void FASTCALL call_13_14_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_11_9
#define init_13_14_11_9
void FASTCALL call_13_14_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_11_10
#define init_13_14_11_10
void FASTCALL call_13_14_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_11_11
#define init_13_14_11_11
void FASTCALL call_13_14_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_11_12
#define init_13_14_11_12
void FASTCALL call_13_14_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_11_13
#define init_13_14_11_13
void FASTCALL call_13_14_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_11_14
#define init_13_14_11_14
void FASTCALL call_13_14_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_11_15
#define init_13_14_11_15
void FASTCALL call_13_14_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_12_0
#define init_13_14_12_0
void FASTCALL call_13_14_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_12_1
#define init_13_14_12_1
void FASTCALL call_13_14_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_12_2
#define init_13_14_12_2
void FASTCALL call_13_14_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_12_3
#define init_13_14_12_3
void FASTCALL call_13_14_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_12_4
#define init_13_14_12_4
void FASTCALL call_13_14_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_12_5
#define init_13_14_12_5
void FASTCALL call_13_14_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_12_6
#define init_13_14_12_6
void FASTCALL call_13_14_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_12_7
#define init_13_14_12_7
void FASTCALL call_13_14_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_12_8
#define init_13_14_12_8
void FASTCALL call_13_14_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_12_9
#define init_13_14_12_9
void FASTCALL call_13_14_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_12_10
#define init_13_14_12_10
void FASTCALL call_13_14_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_12_11
#define init_13_14_12_11
void FASTCALL call_13_14_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_12_12
#define init_13_14_12_12
void FASTCALL call_13_14_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_12_13
#define init_13_14_12_13
void FASTCALL call_13_14_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_12_14
#define init_13_14_12_14
void FASTCALL call_13_14_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_12_15
#define init_13_14_12_15
void FASTCALL call_13_14_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_13_0
#define init_13_14_13_0
void FASTCALL call_13_14_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_13_1
#define init_13_14_13_1
void FASTCALL call_13_14_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_13_2
#define init_13_14_13_2
void FASTCALL call_13_14_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_13_3
#define init_13_14_13_3
void FASTCALL call_13_14_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_13_4
#define init_13_14_13_4
void FASTCALL call_13_14_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_13_5
#define init_13_14_13_5
void FASTCALL call_13_14_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_13_6
#define init_13_14_13_6
void FASTCALL call_13_14_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_13_7
#define init_13_14_13_7
void FASTCALL call_13_14_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_13_8
#define init_13_14_13_8
void FASTCALL call_13_14_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_13_9
#define init_13_14_13_9
void FASTCALL call_13_14_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_13_10
#define init_13_14_13_10
void FASTCALL call_13_14_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_13_11
#define init_13_14_13_11
void FASTCALL call_13_14_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_13_12
#define init_13_14_13_12
void FASTCALL call_13_14_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_13_13
#define init_13_14_13_13
void FASTCALL call_13_14_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_13_14
#define init_13_14_13_14
void FASTCALL call_13_14_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_13_15
#define init_13_14_13_15
void FASTCALL call_13_14_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_14_0
#define init_13_14_14_0
void FASTCALL call_13_14_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_14_1
#define init_13_14_14_1
void FASTCALL call_13_14_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_14_2
#define init_13_14_14_2
void FASTCALL call_13_14_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_14_3
#define init_13_14_14_3
void FASTCALL call_13_14_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_14_4
#define init_13_14_14_4
void FASTCALL call_13_14_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_14_5
#define init_13_14_14_5
void FASTCALL call_13_14_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_14_6
#define init_13_14_14_6
void FASTCALL call_13_14_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_14_7
#define init_13_14_14_7
void FASTCALL call_13_14_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_14_8
#define init_13_14_14_8
void FASTCALL call_13_14_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_14_9
#define init_13_14_14_9
void FASTCALL call_13_14_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_14_10
#define init_13_14_14_10
void FASTCALL call_13_14_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_14_11
#define init_13_14_14_11
void FASTCALL call_13_14_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_14_12
#define init_13_14_14_12
void FASTCALL call_13_14_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_14_13
#define init_13_14_14_13
void FASTCALL call_13_14_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_14_14
#define init_13_14_14_14
void FASTCALL call_13_14_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_14_15
#define init_13_14_14_15
void FASTCALL call_13_14_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_15_0
#define init_13_14_15_0
void FASTCALL call_13_14_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_15_1
#define init_13_14_15_1
void FASTCALL call_13_14_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_15_2
#define init_13_14_15_2
void FASTCALL call_13_14_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_15_3
#define init_13_14_15_3
void FASTCALL call_13_14_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_15_4
#define init_13_14_15_4
void FASTCALL call_13_14_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_15_5
#define init_13_14_15_5
void FASTCALL call_13_14_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_15_6
#define init_13_14_15_6
void FASTCALL call_13_14_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_15_7
#define init_13_14_15_7
void FASTCALL call_13_14_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_15_8
#define init_13_14_15_8
void FASTCALL call_13_14_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_15_9
#define init_13_14_15_9
void FASTCALL call_13_14_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_15_10
#define init_13_14_15_10
void FASTCALL call_13_14_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_15_11
#define init_13_14_15_11
void FASTCALL call_13_14_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_15_12
#define init_13_14_15_12
void FASTCALL call_13_14_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_15_13
#define init_13_14_15_13
void FASTCALL call_13_14_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_15_14
#define init_13_14_15_14
void FASTCALL call_13_14_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_14_15_15
#define init_13_14_15_15
void FASTCALL call_13_14_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_0_0
#define init_13_15_0_0
void FASTCALL call_13_15_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_0_1
#define init_13_15_0_1
void FASTCALL call_13_15_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_0_2
#define init_13_15_0_2
void FASTCALL call_13_15_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_0_3
#define init_13_15_0_3
void FASTCALL call_13_15_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_0_4
#define init_13_15_0_4
void FASTCALL call_13_15_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_0_5
#define init_13_15_0_5
void FASTCALL call_13_15_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_0_6
#define init_13_15_0_6
void FASTCALL call_13_15_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_0_7
#define init_13_15_0_7
void FASTCALL call_13_15_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_0_8
#define init_13_15_0_8
void FASTCALL call_13_15_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_0_9
#define init_13_15_0_9
void FASTCALL call_13_15_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_0_10
#define init_13_15_0_10
void FASTCALL call_13_15_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_0_11
#define init_13_15_0_11
void FASTCALL call_13_15_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_0_12
#define init_13_15_0_12
void FASTCALL call_13_15_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_0_13
#define init_13_15_0_13
void FASTCALL call_13_15_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_0_14
#define init_13_15_0_14
void FASTCALL call_13_15_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_0_15
#define init_13_15_0_15
void FASTCALL call_13_15_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_1_0
#define init_13_15_1_0
void FASTCALL call_13_15_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_1_1
#define init_13_15_1_1
void FASTCALL call_13_15_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_1_2
#define init_13_15_1_2
void FASTCALL call_13_15_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_1_3
#define init_13_15_1_3
void FASTCALL call_13_15_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_1_4
#define init_13_15_1_4
void FASTCALL call_13_15_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_1_5
#define init_13_15_1_5
void FASTCALL call_13_15_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_1_6
#define init_13_15_1_6
void FASTCALL call_13_15_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_1_7
#define init_13_15_1_7
void FASTCALL call_13_15_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_1_8
#define init_13_15_1_8
void FASTCALL call_13_15_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_1_9
#define init_13_15_1_9
void FASTCALL call_13_15_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_1_10
#define init_13_15_1_10
void FASTCALL call_13_15_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_1_11
#define init_13_15_1_11
void FASTCALL call_13_15_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_1_12
#define init_13_15_1_12
void FASTCALL call_13_15_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_1_13
#define init_13_15_1_13
void FASTCALL call_13_15_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_1_14
#define init_13_15_1_14
void FASTCALL call_13_15_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_1_15
#define init_13_15_1_15
void FASTCALL call_13_15_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_2_0
#define init_13_15_2_0
void FASTCALL call_13_15_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_2_1
#define init_13_15_2_1
void FASTCALL call_13_15_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_2_2
#define init_13_15_2_2
void FASTCALL call_13_15_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_2_3
#define init_13_15_2_3
void FASTCALL call_13_15_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_2_4
#define init_13_15_2_4
void FASTCALL call_13_15_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_2_5
#define init_13_15_2_5
void FASTCALL call_13_15_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_2_6
#define init_13_15_2_6
void FASTCALL call_13_15_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_2_7
#define init_13_15_2_7
void FASTCALL call_13_15_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_2_8
#define init_13_15_2_8
void FASTCALL call_13_15_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_2_9
#define init_13_15_2_9
void FASTCALL call_13_15_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_2_10
#define init_13_15_2_10
void FASTCALL call_13_15_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_2_11
#define init_13_15_2_11
void FASTCALL call_13_15_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_2_12
#define init_13_15_2_12
void FASTCALL call_13_15_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_2_13
#define init_13_15_2_13
void FASTCALL call_13_15_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_2_14
#define init_13_15_2_14
void FASTCALL call_13_15_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_2_15
#define init_13_15_2_15
void FASTCALL call_13_15_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_3_0
#define init_13_15_3_0
void FASTCALL call_13_15_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_3_1
#define init_13_15_3_1
void FASTCALL call_13_15_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_3_2
#define init_13_15_3_2
void FASTCALL call_13_15_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_3_3
#define init_13_15_3_3
void FASTCALL call_13_15_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_3_4
#define init_13_15_3_4
void FASTCALL call_13_15_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_3_5
#define init_13_15_3_5
void FASTCALL call_13_15_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_3_6
#define init_13_15_3_6
void FASTCALL call_13_15_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_3_7
#define init_13_15_3_7
void FASTCALL call_13_15_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_3_8
#define init_13_15_3_8
void FASTCALL call_13_15_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_3_9
#define init_13_15_3_9
void FASTCALL call_13_15_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_3_10
#define init_13_15_3_10
void FASTCALL call_13_15_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_3_11
#define init_13_15_3_11
void FASTCALL call_13_15_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_3_12
#define init_13_15_3_12
void FASTCALL call_13_15_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_3_13
#define init_13_15_3_13
void FASTCALL call_13_15_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_3_14
#define init_13_15_3_14
void FASTCALL call_13_15_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_3_15
#define init_13_15_3_15
void FASTCALL call_13_15_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_4_0
#define init_13_15_4_0
void FASTCALL call_13_15_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_4_1
#define init_13_15_4_1
void FASTCALL call_13_15_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_4_2
#define init_13_15_4_2
void FASTCALL call_13_15_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_4_3
#define init_13_15_4_3
void FASTCALL call_13_15_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_4_4
#define init_13_15_4_4
void FASTCALL call_13_15_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_4_5
#define init_13_15_4_5
void FASTCALL call_13_15_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_4_6
#define init_13_15_4_6
void FASTCALL call_13_15_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_4_7
#define init_13_15_4_7
void FASTCALL call_13_15_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_4_8
#define init_13_15_4_8
void FASTCALL call_13_15_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_4_9
#define init_13_15_4_9
void FASTCALL call_13_15_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_4_10
#define init_13_15_4_10
void FASTCALL call_13_15_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_4_11
#define init_13_15_4_11
void FASTCALL call_13_15_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_4_12
#define init_13_15_4_12
void FASTCALL call_13_15_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_4_13
#define init_13_15_4_13
void FASTCALL call_13_15_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_4_14
#define init_13_15_4_14
void FASTCALL call_13_15_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_4_15
#define init_13_15_4_15
void FASTCALL call_13_15_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_5_0
#define init_13_15_5_0
void FASTCALL call_13_15_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_5_1
#define init_13_15_5_1
void FASTCALL call_13_15_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_5_2
#define init_13_15_5_2
void FASTCALL call_13_15_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_5_3
#define init_13_15_5_3
void FASTCALL call_13_15_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_5_4
#define init_13_15_5_4
void FASTCALL call_13_15_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_5_5
#define init_13_15_5_5
void FASTCALL call_13_15_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_5_6
#define init_13_15_5_6
void FASTCALL call_13_15_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_5_7
#define init_13_15_5_7
void FASTCALL call_13_15_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_5_8
#define init_13_15_5_8
void FASTCALL call_13_15_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_5_9
#define init_13_15_5_9
void FASTCALL call_13_15_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_5_10
#define init_13_15_5_10
void FASTCALL call_13_15_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_5_11
#define init_13_15_5_11
void FASTCALL call_13_15_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_5_12
#define init_13_15_5_12
void FASTCALL call_13_15_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_5_13
#define init_13_15_5_13
void FASTCALL call_13_15_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_5_14
#define init_13_15_5_14
void FASTCALL call_13_15_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_5_15
#define init_13_15_5_15
void FASTCALL call_13_15_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_6_0
#define init_13_15_6_0
void FASTCALL call_13_15_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_6_1
#define init_13_15_6_1
void FASTCALL call_13_15_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_6_2
#define init_13_15_6_2
void FASTCALL call_13_15_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_6_3
#define init_13_15_6_3
void FASTCALL call_13_15_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_6_4
#define init_13_15_6_4
void FASTCALL call_13_15_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_6_5
#define init_13_15_6_5
void FASTCALL call_13_15_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_6_6
#define init_13_15_6_6
void FASTCALL call_13_15_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_6_7
#define init_13_15_6_7
void FASTCALL call_13_15_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_6_8
#define init_13_15_6_8
void FASTCALL call_13_15_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_6_9
#define init_13_15_6_9
void FASTCALL call_13_15_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_6_10
#define init_13_15_6_10
void FASTCALL call_13_15_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_6_11
#define init_13_15_6_11
void FASTCALL call_13_15_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_6_12
#define init_13_15_6_12
void FASTCALL call_13_15_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_6_13
#define init_13_15_6_13
void FASTCALL call_13_15_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_6_14
#define init_13_15_6_14
void FASTCALL call_13_15_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_6_15
#define init_13_15_6_15
void FASTCALL call_13_15_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_7_0
#define init_13_15_7_0
void FASTCALL call_13_15_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_7_1
#define init_13_15_7_1
void FASTCALL call_13_15_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_7_2
#define init_13_15_7_2
void FASTCALL call_13_15_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_7_3
#define init_13_15_7_3
void FASTCALL call_13_15_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_7_4
#define init_13_15_7_4
void FASTCALL call_13_15_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_7_5
#define init_13_15_7_5
void FASTCALL call_13_15_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_7_6
#define init_13_15_7_6
void FASTCALL call_13_15_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_7_7
#define init_13_15_7_7
void FASTCALL call_13_15_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_7_8
#define init_13_15_7_8
void FASTCALL call_13_15_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_7_9
#define init_13_15_7_9
void FASTCALL call_13_15_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_7_10
#define init_13_15_7_10
void FASTCALL call_13_15_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_7_11
#define init_13_15_7_11
void FASTCALL call_13_15_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_7_12
#define init_13_15_7_12
void FASTCALL call_13_15_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_7_13
#define init_13_15_7_13
void FASTCALL call_13_15_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_7_14
#define init_13_15_7_14
void FASTCALL call_13_15_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_7_15
#define init_13_15_7_15
void FASTCALL call_13_15_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_8_0
#define init_13_15_8_0
void FASTCALL call_13_15_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_8_1
#define init_13_15_8_1
void FASTCALL call_13_15_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_8_2
#define init_13_15_8_2
void FASTCALL call_13_15_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_8_3
#define init_13_15_8_3
void FASTCALL call_13_15_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_8_4
#define init_13_15_8_4
void FASTCALL call_13_15_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_8_5
#define init_13_15_8_5
void FASTCALL call_13_15_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_8_6
#define init_13_15_8_6
void FASTCALL call_13_15_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_8_7
#define init_13_15_8_7
void FASTCALL call_13_15_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_8_8
#define init_13_15_8_8
void FASTCALL call_13_15_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_8_9
#define init_13_15_8_9
void FASTCALL call_13_15_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_8_10
#define init_13_15_8_10
void FASTCALL call_13_15_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_8_11
#define init_13_15_8_11
void FASTCALL call_13_15_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_8_12
#define init_13_15_8_12
void FASTCALL call_13_15_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_8_13
#define init_13_15_8_13
void FASTCALL call_13_15_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_8_14
#define init_13_15_8_14
void FASTCALL call_13_15_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_8_15
#define init_13_15_8_15
void FASTCALL call_13_15_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_9_0
#define init_13_15_9_0
void FASTCALL call_13_15_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_9_1
#define init_13_15_9_1
void FASTCALL call_13_15_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_9_2
#define init_13_15_9_2
void FASTCALL call_13_15_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_9_3
#define init_13_15_9_3
void FASTCALL call_13_15_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_9_4
#define init_13_15_9_4
void FASTCALL call_13_15_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_9_5
#define init_13_15_9_5
void FASTCALL call_13_15_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_9_6
#define init_13_15_9_6
void FASTCALL call_13_15_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_9_7
#define init_13_15_9_7
void FASTCALL call_13_15_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_9_8
#define init_13_15_9_8
void FASTCALL call_13_15_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_9_9
#define init_13_15_9_9
void FASTCALL call_13_15_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_9_10
#define init_13_15_9_10
void FASTCALL call_13_15_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_9_11
#define init_13_15_9_11
void FASTCALL call_13_15_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_9_12
#define init_13_15_9_12
void FASTCALL call_13_15_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_9_13
#define init_13_15_9_13
void FASTCALL call_13_15_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_9_14
#define init_13_15_9_14
void FASTCALL call_13_15_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_9_15
#define init_13_15_9_15
void FASTCALL call_13_15_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_10_0
#define init_13_15_10_0
void FASTCALL call_13_15_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_10_1
#define init_13_15_10_1
void FASTCALL call_13_15_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_10_2
#define init_13_15_10_2
void FASTCALL call_13_15_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_10_3
#define init_13_15_10_3
void FASTCALL call_13_15_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_10_4
#define init_13_15_10_4
void FASTCALL call_13_15_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_10_5
#define init_13_15_10_5
void FASTCALL call_13_15_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_10_6
#define init_13_15_10_6
void FASTCALL call_13_15_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_10_7
#define init_13_15_10_7
void FASTCALL call_13_15_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_10_8
#define init_13_15_10_8
void FASTCALL call_13_15_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_10_9
#define init_13_15_10_9
void FASTCALL call_13_15_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_10_10
#define init_13_15_10_10
void FASTCALL call_13_15_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_10_11
#define init_13_15_10_11
void FASTCALL call_13_15_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_10_12
#define init_13_15_10_12
void FASTCALL call_13_15_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_10_13
#define init_13_15_10_13
void FASTCALL call_13_15_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_10_14
#define init_13_15_10_14
void FASTCALL call_13_15_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_10_15
#define init_13_15_10_15
void FASTCALL call_13_15_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_11_0
#define init_13_15_11_0
void FASTCALL call_13_15_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_11_1
#define init_13_15_11_1
void FASTCALL call_13_15_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_11_2
#define init_13_15_11_2
void FASTCALL call_13_15_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_11_3
#define init_13_15_11_3
void FASTCALL call_13_15_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_11_4
#define init_13_15_11_4
void FASTCALL call_13_15_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_11_5
#define init_13_15_11_5
void FASTCALL call_13_15_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_11_6
#define init_13_15_11_6
void FASTCALL call_13_15_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_11_7
#define init_13_15_11_7
void FASTCALL call_13_15_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_11_8
#define init_13_15_11_8
void FASTCALL call_13_15_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_11_9
#define init_13_15_11_9
void FASTCALL call_13_15_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_11_10
#define init_13_15_11_10
void FASTCALL call_13_15_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_11_11
#define init_13_15_11_11
void FASTCALL call_13_15_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_11_12
#define init_13_15_11_12
void FASTCALL call_13_15_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_11_13
#define init_13_15_11_13
void FASTCALL call_13_15_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_11_14
#define init_13_15_11_14
void FASTCALL call_13_15_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_11_15
#define init_13_15_11_15
void FASTCALL call_13_15_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_12_0
#define init_13_15_12_0
void FASTCALL call_13_15_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_12_1
#define init_13_15_12_1
void FASTCALL call_13_15_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_12_2
#define init_13_15_12_2
void FASTCALL call_13_15_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_12_3
#define init_13_15_12_3
void FASTCALL call_13_15_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_12_4
#define init_13_15_12_4
void FASTCALL call_13_15_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_12_5
#define init_13_15_12_5
void FASTCALL call_13_15_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_12_6
#define init_13_15_12_6
void FASTCALL call_13_15_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_12_7
#define init_13_15_12_7
void FASTCALL call_13_15_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_12_8
#define init_13_15_12_8
void FASTCALL call_13_15_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_12_9
#define init_13_15_12_9
void FASTCALL call_13_15_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_12_10
#define init_13_15_12_10
void FASTCALL call_13_15_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_12_11
#define init_13_15_12_11
void FASTCALL call_13_15_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_12_12
#define init_13_15_12_12
void FASTCALL call_13_15_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_12_13
#define init_13_15_12_13
void FASTCALL call_13_15_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_12_14
#define init_13_15_12_14
void FASTCALL call_13_15_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_12_15
#define init_13_15_12_15
void FASTCALL call_13_15_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_13_0
#define init_13_15_13_0
void FASTCALL call_13_15_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_13_1
#define init_13_15_13_1
void FASTCALL call_13_15_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_13_2
#define init_13_15_13_2
void FASTCALL call_13_15_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_13_3
#define init_13_15_13_3
void FASTCALL call_13_15_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_13_4
#define init_13_15_13_4
void FASTCALL call_13_15_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_13_5
#define init_13_15_13_5
void FASTCALL call_13_15_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_13_6
#define init_13_15_13_6
void FASTCALL call_13_15_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_13_7
#define init_13_15_13_7
void FASTCALL call_13_15_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_13_8
#define init_13_15_13_8
void FASTCALL call_13_15_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_13_9
#define init_13_15_13_9
void FASTCALL call_13_15_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_13_10
#define init_13_15_13_10
void FASTCALL call_13_15_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_13_11
#define init_13_15_13_11
void FASTCALL call_13_15_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_13_12
#define init_13_15_13_12
void FASTCALL call_13_15_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_13_13
#define init_13_15_13_13
void FASTCALL call_13_15_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_13_14
#define init_13_15_13_14
void FASTCALL call_13_15_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_13_15
#define init_13_15_13_15
void FASTCALL call_13_15_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_14_0
#define init_13_15_14_0
void FASTCALL call_13_15_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_14_1
#define init_13_15_14_1
void FASTCALL call_13_15_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_14_2
#define init_13_15_14_2
void FASTCALL call_13_15_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_14_3
#define init_13_15_14_3
void FASTCALL call_13_15_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_14_4
#define init_13_15_14_4
void FASTCALL call_13_15_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_14_5
#define init_13_15_14_5
void FASTCALL call_13_15_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_14_6
#define init_13_15_14_6
void FASTCALL call_13_15_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_14_7
#define init_13_15_14_7
void FASTCALL call_13_15_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_14_8
#define init_13_15_14_8
void FASTCALL call_13_15_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_14_9
#define init_13_15_14_9
void FASTCALL call_13_15_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_14_10
#define init_13_15_14_10
void FASTCALL call_13_15_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_14_11
#define init_13_15_14_11
void FASTCALL call_13_15_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_14_12
#define init_13_15_14_12
void FASTCALL call_13_15_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_14_13
#define init_13_15_14_13
void FASTCALL call_13_15_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_14_14
#define init_13_15_14_14
void FASTCALL call_13_15_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_14_15
#define init_13_15_14_15
void FASTCALL call_13_15_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_15_0
#define init_13_15_15_0
void FASTCALL call_13_15_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_15_1
#define init_13_15_15_1
void FASTCALL call_13_15_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_15_2
#define init_13_15_15_2
void FASTCALL call_13_15_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_15_3
#define init_13_15_15_3
void FASTCALL call_13_15_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_15_4
#define init_13_15_15_4
void FASTCALL call_13_15_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_15_5
#define init_13_15_15_5
void FASTCALL call_13_15_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_15_6
#define init_13_15_15_6
void FASTCALL call_13_15_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_15_7
#define init_13_15_15_7
void FASTCALL call_13_15_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_15_8
#define init_13_15_15_8
void FASTCALL call_13_15_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_15_9
#define init_13_15_15_9
void FASTCALL call_13_15_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_15_10
#define init_13_15_15_10
void FASTCALL call_13_15_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_15_11
#define init_13_15_15_11
void FASTCALL call_13_15_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_15_12
#define init_13_15_15_12
void FASTCALL call_13_15_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_15_13
#define init_13_15_15_13
void FASTCALL call_13_15_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_15_14
#define init_13_15_15_14
void FASTCALL call_13_15_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_13_15_15_15
#define init_13_15_15_15
void FASTCALL call_13_15_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_0_0
#define init_14_0_0_0
void FASTCALL call_14_0_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_0_1
#define init_14_0_0_1
void FASTCALL call_14_0_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_0_2
#define init_14_0_0_2
void FASTCALL call_14_0_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_0_3
#define init_14_0_0_3
void FASTCALL call_14_0_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_0_4
#define init_14_0_0_4
void FASTCALL call_14_0_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_0_5
#define init_14_0_0_5
void FASTCALL call_14_0_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_0_6
#define init_14_0_0_6
void FASTCALL call_14_0_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_0_7
#define init_14_0_0_7
void FASTCALL call_14_0_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_0_8
#define init_14_0_0_8
void FASTCALL call_14_0_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_0_9
#define init_14_0_0_9
void FASTCALL call_14_0_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_0_10
#define init_14_0_0_10
void FASTCALL call_14_0_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_0_11
#define init_14_0_0_11
void FASTCALL call_14_0_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_0_12
#define init_14_0_0_12
void FASTCALL call_14_0_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_0_13
#define init_14_0_0_13
void FASTCALL call_14_0_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_0_14
#define init_14_0_0_14
void FASTCALL call_14_0_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_0_15
#define init_14_0_0_15
void FASTCALL call_14_0_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_1_0
#define init_14_0_1_0
void FASTCALL call_14_0_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_1_1
#define init_14_0_1_1
void FASTCALL call_14_0_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_1_2
#define init_14_0_1_2
void FASTCALL call_14_0_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_1_3
#define init_14_0_1_3
void FASTCALL call_14_0_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_1_4
#define init_14_0_1_4
void FASTCALL call_14_0_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_1_5
#define init_14_0_1_5
void FASTCALL call_14_0_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_1_6
#define init_14_0_1_6
void FASTCALL call_14_0_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_1_7
#define init_14_0_1_7
void FASTCALL call_14_0_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_1_8
#define init_14_0_1_8
void FASTCALL call_14_0_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_1_9
#define init_14_0_1_9
void FASTCALL call_14_0_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_1_10
#define init_14_0_1_10
void FASTCALL call_14_0_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_1_11
#define init_14_0_1_11
void FASTCALL call_14_0_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_1_12
#define init_14_0_1_12
void FASTCALL call_14_0_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_1_13
#define init_14_0_1_13
void FASTCALL call_14_0_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_1_14
#define init_14_0_1_14
void FASTCALL call_14_0_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_1_15
#define init_14_0_1_15
void FASTCALL call_14_0_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_2_0
#define init_14_0_2_0
void FASTCALL call_14_0_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_2_1
#define init_14_0_2_1
void FASTCALL call_14_0_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_2_2
#define init_14_0_2_2
void FASTCALL call_14_0_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_2_3
#define init_14_0_2_3
void FASTCALL call_14_0_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_2_4
#define init_14_0_2_4
void FASTCALL call_14_0_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_2_5
#define init_14_0_2_5
void FASTCALL call_14_0_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_2_6
#define init_14_0_2_6
void FASTCALL call_14_0_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_2_7
#define init_14_0_2_7
void FASTCALL call_14_0_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_2_8
#define init_14_0_2_8
void FASTCALL call_14_0_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_2_9
#define init_14_0_2_9
void FASTCALL call_14_0_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_2_10
#define init_14_0_2_10
void FASTCALL call_14_0_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_2_11
#define init_14_0_2_11
void FASTCALL call_14_0_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_2_12
#define init_14_0_2_12
void FASTCALL call_14_0_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_2_13
#define init_14_0_2_13
void FASTCALL call_14_0_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_2_14
#define init_14_0_2_14
void FASTCALL call_14_0_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_2_15
#define init_14_0_2_15
void FASTCALL call_14_0_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_3_0
#define init_14_0_3_0
void FASTCALL call_14_0_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_3_1
#define init_14_0_3_1
void FASTCALL call_14_0_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_3_2
#define init_14_0_3_2
void FASTCALL call_14_0_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_3_3
#define init_14_0_3_3
void FASTCALL call_14_0_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_3_4
#define init_14_0_3_4
void FASTCALL call_14_0_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_3_5
#define init_14_0_3_5
void FASTCALL call_14_0_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_3_6
#define init_14_0_3_6
void FASTCALL call_14_0_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_3_7
#define init_14_0_3_7
void FASTCALL call_14_0_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_3_8
#define init_14_0_3_8
void FASTCALL call_14_0_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_3_9
#define init_14_0_3_9
void FASTCALL call_14_0_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_3_10
#define init_14_0_3_10
void FASTCALL call_14_0_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_3_11
#define init_14_0_3_11
void FASTCALL call_14_0_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_3_12
#define init_14_0_3_12
void FASTCALL call_14_0_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_3_13
#define init_14_0_3_13
void FASTCALL call_14_0_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_3_14
#define init_14_0_3_14
void FASTCALL call_14_0_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_3_15
#define init_14_0_3_15
void FASTCALL call_14_0_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_4_0
#define init_14_0_4_0
void FASTCALL call_14_0_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_4_1
#define init_14_0_4_1
void FASTCALL call_14_0_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_4_2
#define init_14_0_4_2
void FASTCALL call_14_0_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_4_3
#define init_14_0_4_3
void FASTCALL call_14_0_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_4_4
#define init_14_0_4_4
void FASTCALL call_14_0_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_4_5
#define init_14_0_4_5
void FASTCALL call_14_0_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_4_6
#define init_14_0_4_6
void FASTCALL call_14_0_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_4_7
#define init_14_0_4_7
void FASTCALL call_14_0_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_4_8
#define init_14_0_4_8
void FASTCALL call_14_0_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_4_9
#define init_14_0_4_9
void FASTCALL call_14_0_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_4_10
#define init_14_0_4_10
void FASTCALL call_14_0_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_4_11
#define init_14_0_4_11
void FASTCALL call_14_0_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_4_12
#define init_14_0_4_12
void FASTCALL call_14_0_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_4_13
#define init_14_0_4_13
void FASTCALL call_14_0_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_4_14
#define init_14_0_4_14
void FASTCALL call_14_0_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_4_15
#define init_14_0_4_15
void FASTCALL call_14_0_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_5_0
#define init_14_0_5_0
void FASTCALL call_14_0_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_5_1
#define init_14_0_5_1
void FASTCALL call_14_0_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_5_2
#define init_14_0_5_2
void FASTCALL call_14_0_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_5_3
#define init_14_0_5_3
void FASTCALL call_14_0_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_5_4
#define init_14_0_5_4
void FASTCALL call_14_0_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_5_5
#define init_14_0_5_5
void FASTCALL call_14_0_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_5_6
#define init_14_0_5_6
void FASTCALL call_14_0_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_5_7
#define init_14_0_5_7
void FASTCALL call_14_0_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_5_8
#define init_14_0_5_8
void FASTCALL call_14_0_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_5_9
#define init_14_0_5_9
void FASTCALL call_14_0_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_5_10
#define init_14_0_5_10
void FASTCALL call_14_0_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_5_11
#define init_14_0_5_11
void FASTCALL call_14_0_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_5_12
#define init_14_0_5_12
void FASTCALL call_14_0_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_5_13
#define init_14_0_5_13
void FASTCALL call_14_0_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_5_14
#define init_14_0_5_14
void FASTCALL call_14_0_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_5_15
#define init_14_0_5_15
void FASTCALL call_14_0_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_6_0
#define init_14_0_6_0
void FASTCALL call_14_0_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_6_1
#define init_14_0_6_1
void FASTCALL call_14_0_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_6_2
#define init_14_0_6_2
void FASTCALL call_14_0_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_6_3
#define init_14_0_6_3
void FASTCALL call_14_0_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_6_4
#define init_14_0_6_4
void FASTCALL call_14_0_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_6_5
#define init_14_0_6_5
void FASTCALL call_14_0_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_6_6
#define init_14_0_6_6
void FASTCALL call_14_0_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_6_7
#define init_14_0_6_7
void FASTCALL call_14_0_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_6_8
#define init_14_0_6_8
void FASTCALL call_14_0_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_6_9
#define init_14_0_6_9
void FASTCALL call_14_0_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_6_10
#define init_14_0_6_10
void FASTCALL call_14_0_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_6_11
#define init_14_0_6_11
void FASTCALL call_14_0_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_6_12
#define init_14_0_6_12
void FASTCALL call_14_0_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_6_13
#define init_14_0_6_13
void FASTCALL call_14_0_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_6_14
#define init_14_0_6_14
void FASTCALL call_14_0_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_6_15
#define init_14_0_6_15
void FASTCALL call_14_0_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_7_0
#define init_14_0_7_0
void FASTCALL call_14_0_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_7_1
#define init_14_0_7_1
void FASTCALL call_14_0_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_7_2
#define init_14_0_7_2
void FASTCALL call_14_0_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_7_3
#define init_14_0_7_3
void FASTCALL call_14_0_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_7_4
#define init_14_0_7_4
void FASTCALL call_14_0_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_7_5
#define init_14_0_7_5
void FASTCALL call_14_0_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_7_6
#define init_14_0_7_6
void FASTCALL call_14_0_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_7_7
#define init_14_0_7_7
void FASTCALL call_14_0_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_7_8
#define init_14_0_7_8
void FASTCALL call_14_0_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_7_9
#define init_14_0_7_9
void FASTCALL call_14_0_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_7_10
#define init_14_0_7_10
void FASTCALL call_14_0_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_7_11
#define init_14_0_7_11
void FASTCALL call_14_0_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_7_12
#define init_14_0_7_12
void FASTCALL call_14_0_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_7_13
#define init_14_0_7_13
void FASTCALL call_14_0_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_7_14
#define init_14_0_7_14
void FASTCALL call_14_0_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_7_15
#define init_14_0_7_15
void FASTCALL call_14_0_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_8_0
#define init_14_0_8_0
void FASTCALL call_14_0_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_8_1
#define init_14_0_8_1
void FASTCALL call_14_0_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_8_2
#define init_14_0_8_2
void FASTCALL call_14_0_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_8_3
#define init_14_0_8_3
void FASTCALL call_14_0_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_8_4
#define init_14_0_8_4
void FASTCALL call_14_0_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_8_5
#define init_14_0_8_5
void FASTCALL call_14_0_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_8_6
#define init_14_0_8_6
void FASTCALL call_14_0_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_8_7
#define init_14_0_8_7
void FASTCALL call_14_0_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_8_8
#define init_14_0_8_8
void FASTCALL call_14_0_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_8_9
#define init_14_0_8_9
void FASTCALL call_14_0_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_8_10
#define init_14_0_8_10
void FASTCALL call_14_0_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_8_11
#define init_14_0_8_11
void FASTCALL call_14_0_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_8_12
#define init_14_0_8_12
void FASTCALL call_14_0_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_8_13
#define init_14_0_8_13
void FASTCALL call_14_0_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_8_14
#define init_14_0_8_14
void FASTCALL call_14_0_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_8_15
#define init_14_0_8_15
void FASTCALL call_14_0_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_9_0
#define init_14_0_9_0
void FASTCALL call_14_0_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_9_1
#define init_14_0_9_1
void FASTCALL call_14_0_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_9_2
#define init_14_0_9_2
void FASTCALL call_14_0_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_9_3
#define init_14_0_9_3
void FASTCALL call_14_0_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_9_4
#define init_14_0_9_4
void FASTCALL call_14_0_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_9_5
#define init_14_0_9_5
void FASTCALL call_14_0_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_9_6
#define init_14_0_9_6
void FASTCALL call_14_0_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_9_7
#define init_14_0_9_7
void FASTCALL call_14_0_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_9_8
#define init_14_0_9_8
void FASTCALL call_14_0_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_9_9
#define init_14_0_9_9
void FASTCALL call_14_0_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_9_10
#define init_14_0_9_10
void FASTCALL call_14_0_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_9_11
#define init_14_0_9_11
void FASTCALL call_14_0_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_9_12
#define init_14_0_9_12
void FASTCALL call_14_0_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_9_13
#define init_14_0_9_13
void FASTCALL call_14_0_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_9_14
#define init_14_0_9_14
void FASTCALL call_14_0_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_9_15
#define init_14_0_9_15
void FASTCALL call_14_0_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_10_0
#define init_14_0_10_0
void FASTCALL call_14_0_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_10_1
#define init_14_0_10_1
void FASTCALL call_14_0_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_10_2
#define init_14_0_10_2
void FASTCALL call_14_0_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_10_3
#define init_14_0_10_3
void FASTCALL call_14_0_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_10_4
#define init_14_0_10_4
void FASTCALL call_14_0_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_10_5
#define init_14_0_10_5
void FASTCALL call_14_0_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_10_6
#define init_14_0_10_6
void FASTCALL call_14_0_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_10_7
#define init_14_0_10_7
void FASTCALL call_14_0_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_10_8
#define init_14_0_10_8
void FASTCALL call_14_0_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_10_9
#define init_14_0_10_9
void FASTCALL call_14_0_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_10_10
#define init_14_0_10_10
void FASTCALL call_14_0_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_10_11
#define init_14_0_10_11
void FASTCALL call_14_0_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_10_12
#define init_14_0_10_12
void FASTCALL call_14_0_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_10_13
#define init_14_0_10_13
void FASTCALL call_14_0_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_10_14
#define init_14_0_10_14
void FASTCALL call_14_0_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_10_15
#define init_14_0_10_15
void FASTCALL call_14_0_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_11_0
#define init_14_0_11_0
void FASTCALL call_14_0_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_11_1
#define init_14_0_11_1
void FASTCALL call_14_0_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_11_2
#define init_14_0_11_2
void FASTCALL call_14_0_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_11_3
#define init_14_0_11_3
void FASTCALL call_14_0_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_11_4
#define init_14_0_11_4
void FASTCALL call_14_0_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_11_5
#define init_14_0_11_5
void FASTCALL call_14_0_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_11_6
#define init_14_0_11_6
void FASTCALL call_14_0_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_11_7
#define init_14_0_11_7
void FASTCALL call_14_0_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_11_8
#define init_14_0_11_8
void FASTCALL call_14_0_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_11_9
#define init_14_0_11_9
void FASTCALL call_14_0_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_11_10
#define init_14_0_11_10
void FASTCALL call_14_0_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_11_11
#define init_14_0_11_11
void FASTCALL call_14_0_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_11_12
#define init_14_0_11_12
void FASTCALL call_14_0_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_11_13
#define init_14_0_11_13
void FASTCALL call_14_0_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_11_14
#define init_14_0_11_14
void FASTCALL call_14_0_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_11_15
#define init_14_0_11_15
void FASTCALL call_14_0_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_12_0
#define init_14_0_12_0
void FASTCALL call_14_0_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_12_1
#define init_14_0_12_1
void FASTCALL call_14_0_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_12_2
#define init_14_0_12_2
void FASTCALL call_14_0_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_12_3
#define init_14_0_12_3
void FASTCALL call_14_0_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_12_4
#define init_14_0_12_4
void FASTCALL call_14_0_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_12_5
#define init_14_0_12_5
void FASTCALL call_14_0_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_12_6
#define init_14_0_12_6
void FASTCALL call_14_0_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_12_7
#define init_14_0_12_7
void FASTCALL call_14_0_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_12_8
#define init_14_0_12_8
void FASTCALL call_14_0_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_12_9
#define init_14_0_12_9
void FASTCALL call_14_0_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_12_10
#define init_14_0_12_10
void FASTCALL call_14_0_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_12_11
#define init_14_0_12_11
void FASTCALL call_14_0_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_12_12
#define init_14_0_12_12
void FASTCALL call_14_0_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_12_13
#define init_14_0_12_13
void FASTCALL call_14_0_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_12_14
#define init_14_0_12_14
void FASTCALL call_14_0_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_12_15
#define init_14_0_12_15
void FASTCALL call_14_0_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_13_0
#define init_14_0_13_0
void FASTCALL call_14_0_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_13_1
#define init_14_0_13_1
void FASTCALL call_14_0_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_13_2
#define init_14_0_13_2
void FASTCALL call_14_0_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_13_3
#define init_14_0_13_3
void FASTCALL call_14_0_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_13_4
#define init_14_0_13_4
void FASTCALL call_14_0_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_13_5
#define init_14_0_13_5
void FASTCALL call_14_0_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_13_6
#define init_14_0_13_6
void FASTCALL call_14_0_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_13_7
#define init_14_0_13_7
void FASTCALL call_14_0_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_13_8
#define init_14_0_13_8
void FASTCALL call_14_0_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_13_9
#define init_14_0_13_9
void FASTCALL call_14_0_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_13_10
#define init_14_0_13_10
void FASTCALL call_14_0_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_13_11
#define init_14_0_13_11
void FASTCALL call_14_0_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_13_12
#define init_14_0_13_12
void FASTCALL call_14_0_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_13_13
#define init_14_0_13_13
void FASTCALL call_14_0_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_13_14
#define init_14_0_13_14
void FASTCALL call_14_0_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_13_15
#define init_14_0_13_15
void FASTCALL call_14_0_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_14_0
#define init_14_0_14_0
void FASTCALL call_14_0_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_14_1
#define init_14_0_14_1
void FASTCALL call_14_0_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_14_2
#define init_14_0_14_2
void FASTCALL call_14_0_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_14_3
#define init_14_0_14_3
void FASTCALL call_14_0_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_14_4
#define init_14_0_14_4
void FASTCALL call_14_0_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_14_5
#define init_14_0_14_5
void FASTCALL call_14_0_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_14_6
#define init_14_0_14_6
void FASTCALL call_14_0_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_14_7
#define init_14_0_14_7
void FASTCALL call_14_0_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_14_8
#define init_14_0_14_8
void FASTCALL call_14_0_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_14_9
#define init_14_0_14_9
void FASTCALL call_14_0_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_14_10
#define init_14_0_14_10
void FASTCALL call_14_0_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_14_11
#define init_14_0_14_11
void FASTCALL call_14_0_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_14_12
#define init_14_0_14_12
void FASTCALL call_14_0_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_14_13
#define init_14_0_14_13
void FASTCALL call_14_0_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_14_14
#define init_14_0_14_14
void FASTCALL call_14_0_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_14_15
#define init_14_0_14_15
void FASTCALL call_14_0_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_15_0
#define init_14_0_15_0
void FASTCALL call_14_0_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_15_1
#define init_14_0_15_1
void FASTCALL call_14_0_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_15_2
#define init_14_0_15_2
void FASTCALL call_14_0_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_15_3
#define init_14_0_15_3
void FASTCALL call_14_0_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_15_4
#define init_14_0_15_4
void FASTCALL call_14_0_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_15_5
#define init_14_0_15_5
void FASTCALL call_14_0_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_15_6
#define init_14_0_15_6
void FASTCALL call_14_0_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_15_7
#define init_14_0_15_7
void FASTCALL call_14_0_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_15_8
#define init_14_0_15_8
void FASTCALL call_14_0_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_15_9
#define init_14_0_15_9
void FASTCALL call_14_0_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_15_10
#define init_14_0_15_10
void FASTCALL call_14_0_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_15_11
#define init_14_0_15_11
void FASTCALL call_14_0_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_15_12
#define init_14_0_15_12
void FASTCALL call_14_0_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_15_13
#define init_14_0_15_13
void FASTCALL call_14_0_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_15_14
#define init_14_0_15_14
void FASTCALL call_14_0_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_0_15_15
#define init_14_0_15_15
void FASTCALL call_14_0_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_0_0
#define init_14_1_0_0
void FASTCALL call_14_1_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_0_1
#define init_14_1_0_1
void FASTCALL call_14_1_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_0_2
#define init_14_1_0_2
void FASTCALL call_14_1_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_0_3
#define init_14_1_0_3
void FASTCALL call_14_1_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_0_4
#define init_14_1_0_4
void FASTCALL call_14_1_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_0_5
#define init_14_1_0_5
void FASTCALL call_14_1_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_0_6
#define init_14_1_0_6
void FASTCALL call_14_1_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_0_7
#define init_14_1_0_7
void FASTCALL call_14_1_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_0_8
#define init_14_1_0_8
void FASTCALL call_14_1_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_0_9
#define init_14_1_0_9
void FASTCALL call_14_1_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_0_10
#define init_14_1_0_10
void FASTCALL call_14_1_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_0_11
#define init_14_1_0_11
void FASTCALL call_14_1_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_0_12
#define init_14_1_0_12
void FASTCALL call_14_1_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_0_13
#define init_14_1_0_13
void FASTCALL call_14_1_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_0_14
#define init_14_1_0_14
void FASTCALL call_14_1_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_0_15
#define init_14_1_0_15
void FASTCALL call_14_1_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_1_0
#define init_14_1_1_0
void FASTCALL call_14_1_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_1_1
#define init_14_1_1_1
void FASTCALL call_14_1_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_1_2
#define init_14_1_1_2
void FASTCALL call_14_1_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_1_3
#define init_14_1_1_3
void FASTCALL call_14_1_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_1_4
#define init_14_1_1_4
void FASTCALL call_14_1_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_1_5
#define init_14_1_1_5
void FASTCALL call_14_1_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_1_6
#define init_14_1_1_6
void FASTCALL call_14_1_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_1_7
#define init_14_1_1_7
void FASTCALL call_14_1_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_1_8
#define init_14_1_1_8
void FASTCALL call_14_1_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_1_9
#define init_14_1_1_9
void FASTCALL call_14_1_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_1_10
#define init_14_1_1_10
void FASTCALL call_14_1_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_1_11
#define init_14_1_1_11
void FASTCALL call_14_1_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_1_12
#define init_14_1_1_12
void FASTCALL call_14_1_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_1_13
#define init_14_1_1_13
void FASTCALL call_14_1_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_1_14
#define init_14_1_1_14
void FASTCALL call_14_1_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_1_15
#define init_14_1_1_15
void FASTCALL call_14_1_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_2_0
#define init_14_1_2_0
void FASTCALL call_14_1_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_2_1
#define init_14_1_2_1
void FASTCALL call_14_1_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_2_2
#define init_14_1_2_2
void FASTCALL call_14_1_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_2_3
#define init_14_1_2_3
void FASTCALL call_14_1_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_2_4
#define init_14_1_2_4
void FASTCALL call_14_1_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_2_5
#define init_14_1_2_5
void FASTCALL call_14_1_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_2_6
#define init_14_1_2_6
void FASTCALL call_14_1_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_2_7
#define init_14_1_2_7
void FASTCALL call_14_1_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_2_8
#define init_14_1_2_8
void FASTCALL call_14_1_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_2_9
#define init_14_1_2_9
void FASTCALL call_14_1_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_2_10
#define init_14_1_2_10
void FASTCALL call_14_1_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_2_11
#define init_14_1_2_11
void FASTCALL call_14_1_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_2_12
#define init_14_1_2_12
void FASTCALL call_14_1_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_2_13
#define init_14_1_2_13
void FASTCALL call_14_1_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_2_14
#define init_14_1_2_14
void FASTCALL call_14_1_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_2_15
#define init_14_1_2_15
void FASTCALL call_14_1_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_3_0
#define init_14_1_3_0
void FASTCALL call_14_1_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_3_1
#define init_14_1_3_1
void FASTCALL call_14_1_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_3_2
#define init_14_1_3_2
void FASTCALL call_14_1_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_3_3
#define init_14_1_3_3
void FASTCALL call_14_1_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_3_4
#define init_14_1_3_4
void FASTCALL call_14_1_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_3_5
#define init_14_1_3_5
void FASTCALL call_14_1_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_3_6
#define init_14_1_3_6
void FASTCALL call_14_1_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_3_7
#define init_14_1_3_7
void FASTCALL call_14_1_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_3_8
#define init_14_1_3_8
void FASTCALL call_14_1_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_3_9
#define init_14_1_3_9
void FASTCALL call_14_1_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_3_10
#define init_14_1_3_10
void FASTCALL call_14_1_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_3_11
#define init_14_1_3_11
void FASTCALL call_14_1_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_3_12
#define init_14_1_3_12
void FASTCALL call_14_1_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_3_13
#define init_14_1_3_13
void FASTCALL call_14_1_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_3_14
#define init_14_1_3_14
void FASTCALL call_14_1_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_3_15
#define init_14_1_3_15
void FASTCALL call_14_1_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_4_0
#define init_14_1_4_0
void FASTCALL call_14_1_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_4_1
#define init_14_1_4_1
void FASTCALL call_14_1_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_4_2
#define init_14_1_4_2
void FASTCALL call_14_1_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_4_3
#define init_14_1_4_3
void FASTCALL call_14_1_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_4_4
#define init_14_1_4_4
void FASTCALL call_14_1_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_4_5
#define init_14_1_4_5
void FASTCALL call_14_1_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_4_6
#define init_14_1_4_6
void FASTCALL call_14_1_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_4_7
#define init_14_1_4_7
void FASTCALL call_14_1_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_4_8
#define init_14_1_4_8
void FASTCALL call_14_1_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_4_9
#define init_14_1_4_9
void FASTCALL call_14_1_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_4_10
#define init_14_1_4_10
void FASTCALL call_14_1_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_4_11
#define init_14_1_4_11
void FASTCALL call_14_1_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_4_12
#define init_14_1_4_12
void FASTCALL call_14_1_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_4_13
#define init_14_1_4_13
void FASTCALL call_14_1_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_4_14
#define init_14_1_4_14
void FASTCALL call_14_1_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_4_15
#define init_14_1_4_15
void FASTCALL call_14_1_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_5_0
#define init_14_1_5_0
void FASTCALL call_14_1_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_5_1
#define init_14_1_5_1
void FASTCALL call_14_1_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_5_2
#define init_14_1_5_2
void FASTCALL call_14_1_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_5_3
#define init_14_1_5_3
void FASTCALL call_14_1_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_5_4
#define init_14_1_5_4
void FASTCALL call_14_1_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_5_5
#define init_14_1_5_5
void FASTCALL call_14_1_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_5_6
#define init_14_1_5_6
void FASTCALL call_14_1_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_5_7
#define init_14_1_5_7
void FASTCALL call_14_1_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_5_8
#define init_14_1_5_8
void FASTCALL call_14_1_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_5_9
#define init_14_1_5_9
void FASTCALL call_14_1_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_5_10
#define init_14_1_5_10
void FASTCALL call_14_1_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_5_11
#define init_14_1_5_11
void FASTCALL call_14_1_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_5_12
#define init_14_1_5_12
void FASTCALL call_14_1_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_5_13
#define init_14_1_5_13
void FASTCALL call_14_1_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_5_14
#define init_14_1_5_14
void FASTCALL call_14_1_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_5_15
#define init_14_1_5_15
void FASTCALL call_14_1_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_6_0
#define init_14_1_6_0
void FASTCALL call_14_1_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_6_1
#define init_14_1_6_1
void FASTCALL call_14_1_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_6_2
#define init_14_1_6_2
void FASTCALL call_14_1_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_6_3
#define init_14_1_6_3
void FASTCALL call_14_1_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_6_4
#define init_14_1_6_4
void FASTCALL call_14_1_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_6_5
#define init_14_1_6_5
void FASTCALL call_14_1_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_6_6
#define init_14_1_6_6
void FASTCALL call_14_1_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_6_7
#define init_14_1_6_7
void FASTCALL call_14_1_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_6_8
#define init_14_1_6_8
void FASTCALL call_14_1_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_6_9
#define init_14_1_6_9
void FASTCALL call_14_1_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_6_10
#define init_14_1_6_10
void FASTCALL call_14_1_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_6_11
#define init_14_1_6_11
void FASTCALL call_14_1_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_6_12
#define init_14_1_6_12
void FASTCALL call_14_1_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_6_13
#define init_14_1_6_13
void FASTCALL call_14_1_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_6_14
#define init_14_1_6_14
void FASTCALL call_14_1_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_6_15
#define init_14_1_6_15
void FASTCALL call_14_1_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_7_0
#define init_14_1_7_0
void FASTCALL call_14_1_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_7_1
#define init_14_1_7_1
void FASTCALL call_14_1_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_7_2
#define init_14_1_7_2
void FASTCALL call_14_1_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_7_3
#define init_14_1_7_3
void FASTCALL call_14_1_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_7_4
#define init_14_1_7_4
void FASTCALL call_14_1_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_7_5
#define init_14_1_7_5
void FASTCALL call_14_1_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_7_6
#define init_14_1_7_6
void FASTCALL call_14_1_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_7_7
#define init_14_1_7_7
void FASTCALL call_14_1_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_7_8
#define init_14_1_7_8
void FASTCALL call_14_1_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_7_9
#define init_14_1_7_9
void FASTCALL call_14_1_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_7_10
#define init_14_1_7_10
void FASTCALL call_14_1_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_7_11
#define init_14_1_7_11
void FASTCALL call_14_1_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_7_12
#define init_14_1_7_12
void FASTCALL call_14_1_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_7_13
#define init_14_1_7_13
void FASTCALL call_14_1_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_7_14
#define init_14_1_7_14
void FASTCALL call_14_1_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_7_15
#define init_14_1_7_15
void FASTCALL call_14_1_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_8_0
#define init_14_1_8_0
void FASTCALL call_14_1_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_8_1
#define init_14_1_8_1
void FASTCALL call_14_1_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_8_2
#define init_14_1_8_2
void FASTCALL call_14_1_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_8_3
#define init_14_1_8_3
void FASTCALL call_14_1_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_8_4
#define init_14_1_8_4
void FASTCALL call_14_1_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_8_5
#define init_14_1_8_5
void FASTCALL call_14_1_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_8_6
#define init_14_1_8_6
void FASTCALL call_14_1_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_8_7
#define init_14_1_8_7
void FASTCALL call_14_1_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_8_8
#define init_14_1_8_8
void FASTCALL call_14_1_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_8_9
#define init_14_1_8_9
void FASTCALL call_14_1_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_8_10
#define init_14_1_8_10
void FASTCALL call_14_1_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_8_11
#define init_14_1_8_11
void FASTCALL call_14_1_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_8_12
#define init_14_1_8_12
void FASTCALL call_14_1_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_8_13
#define init_14_1_8_13
void FASTCALL call_14_1_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_8_14
#define init_14_1_8_14
void FASTCALL call_14_1_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_8_15
#define init_14_1_8_15
void FASTCALL call_14_1_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_9_0
#define init_14_1_9_0
void FASTCALL call_14_1_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_9_1
#define init_14_1_9_1
void FASTCALL call_14_1_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_9_2
#define init_14_1_9_2
void FASTCALL call_14_1_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_9_3
#define init_14_1_9_3
void FASTCALL call_14_1_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_9_4
#define init_14_1_9_4
void FASTCALL call_14_1_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_9_5
#define init_14_1_9_5
void FASTCALL call_14_1_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_9_6
#define init_14_1_9_6
void FASTCALL call_14_1_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_9_7
#define init_14_1_9_7
void FASTCALL call_14_1_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_9_8
#define init_14_1_9_8
void FASTCALL call_14_1_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_9_9
#define init_14_1_9_9
void FASTCALL call_14_1_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_9_10
#define init_14_1_9_10
void FASTCALL call_14_1_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_9_11
#define init_14_1_9_11
void FASTCALL call_14_1_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_9_12
#define init_14_1_9_12
void FASTCALL call_14_1_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_9_13
#define init_14_1_9_13
void FASTCALL call_14_1_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_9_14
#define init_14_1_9_14
void FASTCALL call_14_1_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_9_15
#define init_14_1_9_15
void FASTCALL call_14_1_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_10_0
#define init_14_1_10_0
void FASTCALL call_14_1_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_10_1
#define init_14_1_10_1
void FASTCALL call_14_1_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_10_2
#define init_14_1_10_2
void FASTCALL call_14_1_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_10_3
#define init_14_1_10_3
void FASTCALL call_14_1_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_10_4
#define init_14_1_10_4
void FASTCALL call_14_1_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_10_5
#define init_14_1_10_5
void FASTCALL call_14_1_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_10_6
#define init_14_1_10_6
void FASTCALL call_14_1_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_10_7
#define init_14_1_10_7
void FASTCALL call_14_1_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_10_8
#define init_14_1_10_8
void FASTCALL call_14_1_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_10_9
#define init_14_1_10_9
void FASTCALL call_14_1_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_10_10
#define init_14_1_10_10
void FASTCALL call_14_1_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_10_11
#define init_14_1_10_11
void FASTCALL call_14_1_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_10_12
#define init_14_1_10_12
void FASTCALL call_14_1_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_10_13
#define init_14_1_10_13
void FASTCALL call_14_1_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_10_14
#define init_14_1_10_14
void FASTCALL call_14_1_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_10_15
#define init_14_1_10_15
void FASTCALL call_14_1_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_11_0
#define init_14_1_11_0
void FASTCALL call_14_1_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_11_1
#define init_14_1_11_1
void FASTCALL call_14_1_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_11_2
#define init_14_1_11_2
void FASTCALL call_14_1_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_11_3
#define init_14_1_11_3
void FASTCALL call_14_1_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_11_4
#define init_14_1_11_4
void FASTCALL call_14_1_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_11_5
#define init_14_1_11_5
void FASTCALL call_14_1_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_11_6
#define init_14_1_11_6
void FASTCALL call_14_1_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_11_7
#define init_14_1_11_7
void FASTCALL call_14_1_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_11_8
#define init_14_1_11_8
void FASTCALL call_14_1_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_11_9
#define init_14_1_11_9
void FASTCALL call_14_1_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_11_10
#define init_14_1_11_10
void FASTCALL call_14_1_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_11_11
#define init_14_1_11_11
void FASTCALL call_14_1_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_11_12
#define init_14_1_11_12
void FASTCALL call_14_1_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_11_13
#define init_14_1_11_13
void FASTCALL call_14_1_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_11_14
#define init_14_1_11_14
void FASTCALL call_14_1_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_11_15
#define init_14_1_11_15
void FASTCALL call_14_1_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_12_0
#define init_14_1_12_0
void FASTCALL call_14_1_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_12_1
#define init_14_1_12_1
void FASTCALL call_14_1_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_12_2
#define init_14_1_12_2
void FASTCALL call_14_1_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_12_3
#define init_14_1_12_3
void FASTCALL call_14_1_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_12_4
#define init_14_1_12_4
void FASTCALL call_14_1_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_12_5
#define init_14_1_12_5
void FASTCALL call_14_1_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_12_6
#define init_14_1_12_6
void FASTCALL call_14_1_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_12_7
#define init_14_1_12_7
void FASTCALL call_14_1_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_12_8
#define init_14_1_12_8
void FASTCALL call_14_1_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_12_9
#define init_14_1_12_9
void FASTCALL call_14_1_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_12_10
#define init_14_1_12_10
void FASTCALL call_14_1_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_12_11
#define init_14_1_12_11
void FASTCALL call_14_1_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_12_12
#define init_14_1_12_12
void FASTCALL call_14_1_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_12_13
#define init_14_1_12_13
void FASTCALL call_14_1_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_12_14
#define init_14_1_12_14
void FASTCALL call_14_1_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_12_15
#define init_14_1_12_15
void FASTCALL call_14_1_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_13_0
#define init_14_1_13_0
void FASTCALL call_14_1_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_13_1
#define init_14_1_13_1
void FASTCALL call_14_1_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_13_2
#define init_14_1_13_2
void FASTCALL call_14_1_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_13_3
#define init_14_1_13_3
void FASTCALL call_14_1_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_13_4
#define init_14_1_13_4
void FASTCALL call_14_1_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_13_5
#define init_14_1_13_5
void FASTCALL call_14_1_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_13_6
#define init_14_1_13_6
void FASTCALL call_14_1_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_13_7
#define init_14_1_13_7
void FASTCALL call_14_1_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_13_8
#define init_14_1_13_8
void FASTCALL call_14_1_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_13_9
#define init_14_1_13_9
void FASTCALL call_14_1_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_13_10
#define init_14_1_13_10
void FASTCALL call_14_1_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_13_11
#define init_14_1_13_11
void FASTCALL call_14_1_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_13_12
#define init_14_1_13_12
void FASTCALL call_14_1_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_13_13
#define init_14_1_13_13
void FASTCALL call_14_1_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_13_14
#define init_14_1_13_14
void FASTCALL call_14_1_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_13_15
#define init_14_1_13_15
void FASTCALL call_14_1_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_14_0
#define init_14_1_14_0
void FASTCALL call_14_1_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_14_1
#define init_14_1_14_1
void FASTCALL call_14_1_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_14_2
#define init_14_1_14_2
void FASTCALL call_14_1_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_14_3
#define init_14_1_14_3
void FASTCALL call_14_1_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_14_4
#define init_14_1_14_4
void FASTCALL call_14_1_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_14_5
#define init_14_1_14_5
void FASTCALL call_14_1_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_14_6
#define init_14_1_14_6
void FASTCALL call_14_1_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_14_7
#define init_14_1_14_7
void FASTCALL call_14_1_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_14_8
#define init_14_1_14_8
void FASTCALL call_14_1_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_14_9
#define init_14_1_14_9
void FASTCALL call_14_1_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_14_10
#define init_14_1_14_10
void FASTCALL call_14_1_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_14_11
#define init_14_1_14_11
void FASTCALL call_14_1_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_14_12
#define init_14_1_14_12
void FASTCALL call_14_1_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_14_13
#define init_14_1_14_13
void FASTCALL call_14_1_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_14_14
#define init_14_1_14_14
void FASTCALL call_14_1_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_14_15
#define init_14_1_14_15
void FASTCALL call_14_1_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_15_0
#define init_14_1_15_0
void FASTCALL call_14_1_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_15_1
#define init_14_1_15_1
void FASTCALL call_14_1_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_15_2
#define init_14_1_15_2
void FASTCALL call_14_1_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_15_3
#define init_14_1_15_3
void FASTCALL call_14_1_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_15_4
#define init_14_1_15_4
void FASTCALL call_14_1_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_15_5
#define init_14_1_15_5
void FASTCALL call_14_1_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_15_6
#define init_14_1_15_6
void FASTCALL call_14_1_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_15_7
#define init_14_1_15_7
void FASTCALL call_14_1_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_15_8
#define init_14_1_15_8
void FASTCALL call_14_1_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_15_9
#define init_14_1_15_9
void FASTCALL call_14_1_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_15_10
#define init_14_1_15_10
void FASTCALL call_14_1_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_15_11
#define init_14_1_15_11
void FASTCALL call_14_1_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_15_12
#define init_14_1_15_12
void FASTCALL call_14_1_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_15_13
#define init_14_1_15_13
void FASTCALL call_14_1_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_15_14
#define init_14_1_15_14
void FASTCALL call_14_1_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_1_15_15
#define init_14_1_15_15
void FASTCALL call_14_1_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_0_0
#define init_14_2_0_0
void FASTCALL call_14_2_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_0_1
#define init_14_2_0_1
void FASTCALL call_14_2_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_0_2
#define init_14_2_0_2
void FASTCALL call_14_2_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_0_3
#define init_14_2_0_3
void FASTCALL call_14_2_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_0_4
#define init_14_2_0_4
void FASTCALL call_14_2_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_0_5
#define init_14_2_0_5
void FASTCALL call_14_2_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_0_6
#define init_14_2_0_6
void FASTCALL call_14_2_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_0_7
#define init_14_2_0_7
void FASTCALL call_14_2_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_0_8
#define init_14_2_0_8
void FASTCALL call_14_2_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_0_9
#define init_14_2_0_9
void FASTCALL call_14_2_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_0_10
#define init_14_2_0_10
void FASTCALL call_14_2_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_0_11
#define init_14_2_0_11
void FASTCALL call_14_2_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_0_12
#define init_14_2_0_12
void FASTCALL call_14_2_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_0_13
#define init_14_2_0_13
void FASTCALL call_14_2_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_0_14
#define init_14_2_0_14
void FASTCALL call_14_2_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_0_15
#define init_14_2_0_15
void FASTCALL call_14_2_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_1_0
#define init_14_2_1_0
void FASTCALL call_14_2_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_1_1
#define init_14_2_1_1
void FASTCALL call_14_2_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_1_2
#define init_14_2_1_2
void FASTCALL call_14_2_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_1_3
#define init_14_2_1_3
void FASTCALL call_14_2_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_1_4
#define init_14_2_1_4
void FASTCALL call_14_2_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_1_5
#define init_14_2_1_5
void FASTCALL call_14_2_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_1_6
#define init_14_2_1_6
void FASTCALL call_14_2_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_1_7
#define init_14_2_1_7
void FASTCALL call_14_2_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_1_8
#define init_14_2_1_8
void FASTCALL call_14_2_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_1_9
#define init_14_2_1_9
void FASTCALL call_14_2_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_1_10
#define init_14_2_1_10
void FASTCALL call_14_2_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_1_11
#define init_14_2_1_11
void FASTCALL call_14_2_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_1_12
#define init_14_2_1_12
void FASTCALL call_14_2_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_1_13
#define init_14_2_1_13
void FASTCALL call_14_2_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_1_14
#define init_14_2_1_14
void FASTCALL call_14_2_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_1_15
#define init_14_2_1_15
void FASTCALL call_14_2_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_2_0
#define init_14_2_2_0
void FASTCALL call_14_2_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_2_1
#define init_14_2_2_1
void FASTCALL call_14_2_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_2_2
#define init_14_2_2_2
void FASTCALL call_14_2_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_2_3
#define init_14_2_2_3
void FASTCALL call_14_2_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_2_4
#define init_14_2_2_4
void FASTCALL call_14_2_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_2_5
#define init_14_2_2_5
void FASTCALL call_14_2_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_2_6
#define init_14_2_2_6
void FASTCALL call_14_2_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_2_7
#define init_14_2_2_7
void FASTCALL call_14_2_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_2_8
#define init_14_2_2_8
void FASTCALL call_14_2_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_2_9
#define init_14_2_2_9
void FASTCALL call_14_2_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_2_10
#define init_14_2_2_10
void FASTCALL call_14_2_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_2_11
#define init_14_2_2_11
void FASTCALL call_14_2_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_2_12
#define init_14_2_2_12
void FASTCALL call_14_2_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_2_13
#define init_14_2_2_13
void FASTCALL call_14_2_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_2_14
#define init_14_2_2_14
void FASTCALL call_14_2_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_2_15
#define init_14_2_2_15
void FASTCALL call_14_2_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_3_0
#define init_14_2_3_0
void FASTCALL call_14_2_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_3_1
#define init_14_2_3_1
void FASTCALL call_14_2_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_3_2
#define init_14_2_3_2
void FASTCALL call_14_2_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_3_3
#define init_14_2_3_3
void FASTCALL call_14_2_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_3_4
#define init_14_2_3_4
void FASTCALL call_14_2_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_3_5
#define init_14_2_3_5
void FASTCALL call_14_2_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_3_6
#define init_14_2_3_6
void FASTCALL call_14_2_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_3_7
#define init_14_2_3_7
void FASTCALL call_14_2_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_3_8
#define init_14_2_3_8
void FASTCALL call_14_2_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_3_9
#define init_14_2_3_9
void FASTCALL call_14_2_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_3_10
#define init_14_2_3_10
void FASTCALL call_14_2_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_3_11
#define init_14_2_3_11
void FASTCALL call_14_2_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_3_12
#define init_14_2_3_12
void FASTCALL call_14_2_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_3_13
#define init_14_2_3_13
void FASTCALL call_14_2_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_3_14
#define init_14_2_3_14
void FASTCALL call_14_2_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_3_15
#define init_14_2_3_15
void FASTCALL call_14_2_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_4_0
#define init_14_2_4_0
void FASTCALL call_14_2_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_4_1
#define init_14_2_4_1
void FASTCALL call_14_2_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_4_2
#define init_14_2_4_2
void FASTCALL call_14_2_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_4_3
#define init_14_2_4_3
void FASTCALL call_14_2_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_4_4
#define init_14_2_4_4
void FASTCALL call_14_2_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_4_5
#define init_14_2_4_5
void FASTCALL call_14_2_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_4_6
#define init_14_2_4_6
void FASTCALL call_14_2_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_4_7
#define init_14_2_4_7
void FASTCALL call_14_2_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_4_8
#define init_14_2_4_8
void FASTCALL call_14_2_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_4_9
#define init_14_2_4_9
void FASTCALL call_14_2_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_4_10
#define init_14_2_4_10
void FASTCALL call_14_2_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_4_11
#define init_14_2_4_11
void FASTCALL call_14_2_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_4_12
#define init_14_2_4_12
void FASTCALL call_14_2_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_4_13
#define init_14_2_4_13
void FASTCALL call_14_2_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_4_14
#define init_14_2_4_14
void FASTCALL call_14_2_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_4_15
#define init_14_2_4_15
void FASTCALL call_14_2_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_5_0
#define init_14_2_5_0
void FASTCALL call_14_2_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_5_1
#define init_14_2_5_1
void FASTCALL call_14_2_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_5_2
#define init_14_2_5_2
void FASTCALL call_14_2_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_5_3
#define init_14_2_5_3
void FASTCALL call_14_2_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_5_4
#define init_14_2_5_4
void FASTCALL call_14_2_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_5_5
#define init_14_2_5_5
void FASTCALL call_14_2_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_5_6
#define init_14_2_5_6
void FASTCALL call_14_2_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_5_7
#define init_14_2_5_7
void FASTCALL call_14_2_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_5_8
#define init_14_2_5_8
void FASTCALL call_14_2_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_5_9
#define init_14_2_5_9
void FASTCALL call_14_2_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_5_10
#define init_14_2_5_10
void FASTCALL call_14_2_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_5_11
#define init_14_2_5_11
void FASTCALL call_14_2_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_5_12
#define init_14_2_5_12
void FASTCALL call_14_2_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_5_13
#define init_14_2_5_13
void FASTCALL call_14_2_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_5_14
#define init_14_2_5_14
void FASTCALL call_14_2_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_5_15
#define init_14_2_5_15
void FASTCALL call_14_2_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_6_0
#define init_14_2_6_0
void FASTCALL call_14_2_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_6_1
#define init_14_2_6_1
void FASTCALL call_14_2_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_6_2
#define init_14_2_6_2
void FASTCALL call_14_2_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_6_3
#define init_14_2_6_3
void FASTCALL call_14_2_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_6_4
#define init_14_2_6_4
void FASTCALL call_14_2_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_6_5
#define init_14_2_6_5
void FASTCALL call_14_2_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_6_6
#define init_14_2_6_6
void FASTCALL call_14_2_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_6_7
#define init_14_2_6_7
void FASTCALL call_14_2_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_6_8
#define init_14_2_6_8
void FASTCALL call_14_2_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_6_9
#define init_14_2_6_9
void FASTCALL call_14_2_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_6_10
#define init_14_2_6_10
void FASTCALL call_14_2_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_6_11
#define init_14_2_6_11
void FASTCALL call_14_2_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_6_12
#define init_14_2_6_12
void FASTCALL call_14_2_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_6_13
#define init_14_2_6_13
void FASTCALL call_14_2_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_6_14
#define init_14_2_6_14
void FASTCALL call_14_2_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_6_15
#define init_14_2_6_15
void FASTCALL call_14_2_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_7_0
#define init_14_2_7_0
void FASTCALL call_14_2_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_7_1
#define init_14_2_7_1
void FASTCALL call_14_2_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_7_2
#define init_14_2_7_2
void FASTCALL call_14_2_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_7_3
#define init_14_2_7_3
void FASTCALL call_14_2_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_7_4
#define init_14_2_7_4
void FASTCALL call_14_2_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_7_5
#define init_14_2_7_5
void FASTCALL call_14_2_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_7_6
#define init_14_2_7_6
void FASTCALL call_14_2_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_7_7
#define init_14_2_7_7
void FASTCALL call_14_2_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_7_8
#define init_14_2_7_8
void FASTCALL call_14_2_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_7_9
#define init_14_2_7_9
void FASTCALL call_14_2_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_7_10
#define init_14_2_7_10
void FASTCALL call_14_2_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_7_11
#define init_14_2_7_11
void FASTCALL call_14_2_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_7_12
#define init_14_2_7_12
void FASTCALL call_14_2_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_7_13
#define init_14_2_7_13
void FASTCALL call_14_2_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_7_14
#define init_14_2_7_14
void FASTCALL call_14_2_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_7_15
#define init_14_2_7_15
void FASTCALL call_14_2_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_8_0
#define init_14_2_8_0
void FASTCALL call_14_2_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_8_1
#define init_14_2_8_1
void FASTCALL call_14_2_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_8_2
#define init_14_2_8_2
void FASTCALL call_14_2_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_8_3
#define init_14_2_8_3
void FASTCALL call_14_2_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_8_4
#define init_14_2_8_4
void FASTCALL call_14_2_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_8_5
#define init_14_2_8_5
void FASTCALL call_14_2_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_8_6
#define init_14_2_8_6
void FASTCALL call_14_2_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_8_7
#define init_14_2_8_7
void FASTCALL call_14_2_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_8_8
#define init_14_2_8_8
void FASTCALL call_14_2_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_8_9
#define init_14_2_8_9
void FASTCALL call_14_2_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_8_10
#define init_14_2_8_10
void FASTCALL call_14_2_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_8_11
#define init_14_2_8_11
void FASTCALL call_14_2_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_8_12
#define init_14_2_8_12
void FASTCALL call_14_2_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_8_13
#define init_14_2_8_13
void FASTCALL call_14_2_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_8_14
#define init_14_2_8_14
void FASTCALL call_14_2_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_8_15
#define init_14_2_8_15
void FASTCALL call_14_2_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_9_0
#define init_14_2_9_0
void FASTCALL call_14_2_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_9_1
#define init_14_2_9_1
void FASTCALL call_14_2_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_9_2
#define init_14_2_9_2
void FASTCALL call_14_2_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_9_3
#define init_14_2_9_3
void FASTCALL call_14_2_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_9_4
#define init_14_2_9_4
void FASTCALL call_14_2_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_9_5
#define init_14_2_9_5
void FASTCALL call_14_2_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_9_6
#define init_14_2_9_6
void FASTCALL call_14_2_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_9_7
#define init_14_2_9_7
void FASTCALL call_14_2_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_9_8
#define init_14_2_9_8
void FASTCALL call_14_2_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_9_9
#define init_14_2_9_9
void FASTCALL call_14_2_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_9_10
#define init_14_2_9_10
void FASTCALL call_14_2_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_9_11
#define init_14_2_9_11
void FASTCALL call_14_2_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_9_12
#define init_14_2_9_12
void FASTCALL call_14_2_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_9_13
#define init_14_2_9_13
void FASTCALL call_14_2_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_9_14
#define init_14_2_9_14
void FASTCALL call_14_2_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_9_15
#define init_14_2_9_15
void FASTCALL call_14_2_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_10_0
#define init_14_2_10_0
void FASTCALL call_14_2_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_10_1
#define init_14_2_10_1
void FASTCALL call_14_2_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_10_2
#define init_14_2_10_2
void FASTCALL call_14_2_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_10_3
#define init_14_2_10_3
void FASTCALL call_14_2_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_10_4
#define init_14_2_10_4
void FASTCALL call_14_2_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_10_5
#define init_14_2_10_5
void FASTCALL call_14_2_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_10_6
#define init_14_2_10_6
void FASTCALL call_14_2_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_10_7
#define init_14_2_10_7
void FASTCALL call_14_2_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_10_8
#define init_14_2_10_8
void FASTCALL call_14_2_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_10_9
#define init_14_2_10_9
void FASTCALL call_14_2_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_10_10
#define init_14_2_10_10
void FASTCALL call_14_2_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_10_11
#define init_14_2_10_11
void FASTCALL call_14_2_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_10_12
#define init_14_2_10_12
void FASTCALL call_14_2_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_10_13
#define init_14_2_10_13
void FASTCALL call_14_2_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_10_14
#define init_14_2_10_14
void FASTCALL call_14_2_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_10_15
#define init_14_2_10_15
void FASTCALL call_14_2_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_11_0
#define init_14_2_11_0
void FASTCALL call_14_2_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_11_1
#define init_14_2_11_1
void FASTCALL call_14_2_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_11_2
#define init_14_2_11_2
void FASTCALL call_14_2_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_11_3
#define init_14_2_11_3
void FASTCALL call_14_2_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_11_4
#define init_14_2_11_4
void FASTCALL call_14_2_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_11_5
#define init_14_2_11_5
void FASTCALL call_14_2_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_11_6
#define init_14_2_11_6
void FASTCALL call_14_2_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_11_7
#define init_14_2_11_7
void FASTCALL call_14_2_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_11_8
#define init_14_2_11_8
void FASTCALL call_14_2_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_11_9
#define init_14_2_11_9
void FASTCALL call_14_2_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_11_10
#define init_14_2_11_10
void FASTCALL call_14_2_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_11_11
#define init_14_2_11_11
void FASTCALL call_14_2_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_11_12
#define init_14_2_11_12
void FASTCALL call_14_2_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_11_13
#define init_14_2_11_13
void FASTCALL call_14_2_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_11_14
#define init_14_2_11_14
void FASTCALL call_14_2_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_11_15
#define init_14_2_11_15
void FASTCALL call_14_2_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_12_0
#define init_14_2_12_0
void FASTCALL call_14_2_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_12_1
#define init_14_2_12_1
void FASTCALL call_14_2_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_12_2
#define init_14_2_12_2
void FASTCALL call_14_2_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_12_3
#define init_14_2_12_3
void FASTCALL call_14_2_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_12_4
#define init_14_2_12_4
void FASTCALL call_14_2_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_12_5
#define init_14_2_12_5
void FASTCALL call_14_2_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_12_6
#define init_14_2_12_6
void FASTCALL call_14_2_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_12_7
#define init_14_2_12_7
void FASTCALL call_14_2_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_12_8
#define init_14_2_12_8
void FASTCALL call_14_2_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_12_9
#define init_14_2_12_9
void FASTCALL call_14_2_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_12_10
#define init_14_2_12_10
void FASTCALL call_14_2_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_12_11
#define init_14_2_12_11
void FASTCALL call_14_2_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_12_12
#define init_14_2_12_12
void FASTCALL call_14_2_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_12_13
#define init_14_2_12_13
void FASTCALL call_14_2_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_12_14
#define init_14_2_12_14
void FASTCALL call_14_2_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_12_15
#define init_14_2_12_15
void FASTCALL call_14_2_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_13_0
#define init_14_2_13_0
void FASTCALL call_14_2_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_13_1
#define init_14_2_13_1
void FASTCALL call_14_2_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_13_2
#define init_14_2_13_2
void FASTCALL call_14_2_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_13_3
#define init_14_2_13_3
void FASTCALL call_14_2_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_13_4
#define init_14_2_13_4
void FASTCALL call_14_2_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_13_5
#define init_14_2_13_5
void FASTCALL call_14_2_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_13_6
#define init_14_2_13_6
void FASTCALL call_14_2_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_13_7
#define init_14_2_13_7
void FASTCALL call_14_2_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_13_8
#define init_14_2_13_8
void FASTCALL call_14_2_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_13_9
#define init_14_2_13_9
void FASTCALL call_14_2_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_13_10
#define init_14_2_13_10
void FASTCALL call_14_2_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_13_11
#define init_14_2_13_11
void FASTCALL call_14_2_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_13_12
#define init_14_2_13_12
void FASTCALL call_14_2_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_13_13
#define init_14_2_13_13
void FASTCALL call_14_2_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_13_14
#define init_14_2_13_14
void FASTCALL call_14_2_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_13_15
#define init_14_2_13_15
void FASTCALL call_14_2_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_14_0
#define init_14_2_14_0
void FASTCALL call_14_2_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_14_1
#define init_14_2_14_1
void FASTCALL call_14_2_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_14_2
#define init_14_2_14_2
void FASTCALL call_14_2_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_14_3
#define init_14_2_14_3
void FASTCALL call_14_2_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_14_4
#define init_14_2_14_4
void FASTCALL call_14_2_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_14_5
#define init_14_2_14_5
void FASTCALL call_14_2_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_14_6
#define init_14_2_14_6
void FASTCALL call_14_2_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_14_7
#define init_14_2_14_7
void FASTCALL call_14_2_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_14_8
#define init_14_2_14_8
void FASTCALL call_14_2_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_14_9
#define init_14_2_14_9
void FASTCALL call_14_2_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_14_10
#define init_14_2_14_10
void FASTCALL call_14_2_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_14_11
#define init_14_2_14_11
void FASTCALL call_14_2_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_14_12
#define init_14_2_14_12
void FASTCALL call_14_2_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_14_13
#define init_14_2_14_13
void FASTCALL call_14_2_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_14_14
#define init_14_2_14_14
void FASTCALL call_14_2_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_14_15
#define init_14_2_14_15
void FASTCALL call_14_2_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_15_0
#define init_14_2_15_0
void FASTCALL call_14_2_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_15_1
#define init_14_2_15_1
void FASTCALL call_14_2_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_15_2
#define init_14_2_15_2
void FASTCALL call_14_2_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_15_3
#define init_14_2_15_3
void FASTCALL call_14_2_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_15_4
#define init_14_2_15_4
void FASTCALL call_14_2_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_15_5
#define init_14_2_15_5
void FASTCALL call_14_2_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_15_6
#define init_14_2_15_6
void FASTCALL call_14_2_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_15_7
#define init_14_2_15_7
void FASTCALL call_14_2_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_15_8
#define init_14_2_15_8
void FASTCALL call_14_2_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_15_9
#define init_14_2_15_9
void FASTCALL call_14_2_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_15_10
#define init_14_2_15_10
void FASTCALL call_14_2_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_15_11
#define init_14_2_15_11
void FASTCALL call_14_2_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_15_12
#define init_14_2_15_12
void FASTCALL call_14_2_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_15_13
#define init_14_2_15_13
void FASTCALL call_14_2_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_15_14
#define init_14_2_15_14
void FASTCALL call_14_2_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_2_15_15
#define init_14_2_15_15
void FASTCALL call_14_2_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_0_0
#define init_14_3_0_0
void FASTCALL call_14_3_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_0_1
#define init_14_3_0_1
void FASTCALL call_14_3_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_0_2
#define init_14_3_0_2
void FASTCALL call_14_3_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_0_3
#define init_14_3_0_3
void FASTCALL call_14_3_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_0_4
#define init_14_3_0_4
void FASTCALL call_14_3_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_0_5
#define init_14_3_0_5
void FASTCALL call_14_3_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_0_6
#define init_14_3_0_6
void FASTCALL call_14_3_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_0_7
#define init_14_3_0_7
void FASTCALL call_14_3_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_0_8
#define init_14_3_0_8
void FASTCALL call_14_3_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_0_9
#define init_14_3_0_9
void FASTCALL call_14_3_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_0_10
#define init_14_3_0_10
void FASTCALL call_14_3_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_0_11
#define init_14_3_0_11
void FASTCALL call_14_3_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_0_12
#define init_14_3_0_12
void FASTCALL call_14_3_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_0_13
#define init_14_3_0_13
void FASTCALL call_14_3_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_0_14
#define init_14_3_0_14
void FASTCALL call_14_3_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_0_15
#define init_14_3_0_15
void FASTCALL call_14_3_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_1_0
#define init_14_3_1_0
void FASTCALL call_14_3_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_1_1
#define init_14_3_1_1
void FASTCALL call_14_3_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_1_2
#define init_14_3_1_2
void FASTCALL call_14_3_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_1_3
#define init_14_3_1_3
void FASTCALL call_14_3_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_1_4
#define init_14_3_1_4
void FASTCALL call_14_3_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_1_5
#define init_14_3_1_5
void FASTCALL call_14_3_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_1_6
#define init_14_3_1_6
void FASTCALL call_14_3_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_1_7
#define init_14_3_1_7
void FASTCALL call_14_3_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_1_8
#define init_14_3_1_8
void FASTCALL call_14_3_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_1_9
#define init_14_3_1_9
void FASTCALL call_14_3_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_1_10
#define init_14_3_1_10
void FASTCALL call_14_3_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_1_11
#define init_14_3_1_11
void FASTCALL call_14_3_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_1_12
#define init_14_3_1_12
void FASTCALL call_14_3_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_1_13
#define init_14_3_1_13
void FASTCALL call_14_3_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_1_14
#define init_14_3_1_14
void FASTCALL call_14_3_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_1_15
#define init_14_3_1_15
void FASTCALL call_14_3_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_2_0
#define init_14_3_2_0
void FASTCALL call_14_3_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_2_1
#define init_14_3_2_1
void FASTCALL call_14_3_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_2_2
#define init_14_3_2_2
void FASTCALL call_14_3_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_2_3
#define init_14_3_2_3
void FASTCALL call_14_3_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_2_4
#define init_14_3_2_4
void FASTCALL call_14_3_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_2_5
#define init_14_3_2_5
void FASTCALL call_14_3_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_2_6
#define init_14_3_2_6
void FASTCALL call_14_3_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_2_7
#define init_14_3_2_7
void FASTCALL call_14_3_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_2_8
#define init_14_3_2_8
void FASTCALL call_14_3_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_2_9
#define init_14_3_2_9
void FASTCALL call_14_3_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_2_10
#define init_14_3_2_10
void FASTCALL call_14_3_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_2_11
#define init_14_3_2_11
void FASTCALL call_14_3_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_2_12
#define init_14_3_2_12
void FASTCALL call_14_3_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_2_13
#define init_14_3_2_13
void FASTCALL call_14_3_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_2_14
#define init_14_3_2_14
void FASTCALL call_14_3_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_2_15
#define init_14_3_2_15
void FASTCALL call_14_3_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_3_0
#define init_14_3_3_0
void FASTCALL call_14_3_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_3_1
#define init_14_3_3_1
void FASTCALL call_14_3_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_3_2
#define init_14_3_3_2
void FASTCALL call_14_3_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_3_3
#define init_14_3_3_3
void FASTCALL call_14_3_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_3_4
#define init_14_3_3_4
void FASTCALL call_14_3_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_3_5
#define init_14_3_3_5
void FASTCALL call_14_3_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_3_6
#define init_14_3_3_6
void FASTCALL call_14_3_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_3_7
#define init_14_3_3_7
void FASTCALL call_14_3_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_3_8
#define init_14_3_3_8
void FASTCALL call_14_3_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_3_9
#define init_14_3_3_9
void FASTCALL call_14_3_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_3_10
#define init_14_3_3_10
void FASTCALL call_14_3_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_3_11
#define init_14_3_3_11
void FASTCALL call_14_3_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_3_12
#define init_14_3_3_12
void FASTCALL call_14_3_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_3_13
#define init_14_3_3_13
void FASTCALL call_14_3_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_3_14
#define init_14_3_3_14
void FASTCALL call_14_3_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_3_15
#define init_14_3_3_15
void FASTCALL call_14_3_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_4_0
#define init_14_3_4_0
void FASTCALL call_14_3_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_4_1
#define init_14_3_4_1
void FASTCALL call_14_3_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_4_2
#define init_14_3_4_2
void FASTCALL call_14_3_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_4_3
#define init_14_3_4_3
void FASTCALL call_14_3_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_4_4
#define init_14_3_4_4
void FASTCALL call_14_3_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_4_5
#define init_14_3_4_5
void FASTCALL call_14_3_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_4_6
#define init_14_3_4_6
void FASTCALL call_14_3_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_4_7
#define init_14_3_4_7
void FASTCALL call_14_3_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_4_8
#define init_14_3_4_8
void FASTCALL call_14_3_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_4_9
#define init_14_3_4_9
void FASTCALL call_14_3_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_4_10
#define init_14_3_4_10
void FASTCALL call_14_3_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_4_11
#define init_14_3_4_11
void FASTCALL call_14_3_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_4_12
#define init_14_3_4_12
void FASTCALL call_14_3_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_4_13
#define init_14_3_4_13
void FASTCALL call_14_3_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_4_14
#define init_14_3_4_14
void FASTCALL call_14_3_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_4_15
#define init_14_3_4_15
void FASTCALL call_14_3_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_5_0
#define init_14_3_5_0
void FASTCALL call_14_3_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_5_1
#define init_14_3_5_1
void FASTCALL call_14_3_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_5_2
#define init_14_3_5_2
void FASTCALL call_14_3_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_5_3
#define init_14_3_5_3
void FASTCALL call_14_3_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_5_4
#define init_14_3_5_4
void FASTCALL call_14_3_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_5_5
#define init_14_3_5_5
void FASTCALL call_14_3_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_5_6
#define init_14_3_5_6
void FASTCALL call_14_3_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_5_7
#define init_14_3_5_7
void FASTCALL call_14_3_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_5_8
#define init_14_3_5_8
void FASTCALL call_14_3_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_5_9
#define init_14_3_5_9
void FASTCALL call_14_3_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_5_10
#define init_14_3_5_10
void FASTCALL call_14_3_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_5_11
#define init_14_3_5_11
void FASTCALL call_14_3_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_5_12
#define init_14_3_5_12
void FASTCALL call_14_3_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_5_13
#define init_14_3_5_13
void FASTCALL call_14_3_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_5_14
#define init_14_3_5_14
void FASTCALL call_14_3_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_5_15
#define init_14_3_5_15
void FASTCALL call_14_3_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_6_0
#define init_14_3_6_0
void FASTCALL call_14_3_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_6_1
#define init_14_3_6_1
void FASTCALL call_14_3_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_6_2
#define init_14_3_6_2
void FASTCALL call_14_3_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_6_3
#define init_14_3_6_3
void FASTCALL call_14_3_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_6_4
#define init_14_3_6_4
void FASTCALL call_14_3_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_6_5
#define init_14_3_6_5
void FASTCALL call_14_3_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_6_6
#define init_14_3_6_6
void FASTCALL call_14_3_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_6_7
#define init_14_3_6_7
void FASTCALL call_14_3_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_6_8
#define init_14_3_6_8
void FASTCALL call_14_3_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_6_9
#define init_14_3_6_9
void FASTCALL call_14_3_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_6_10
#define init_14_3_6_10
void FASTCALL call_14_3_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_6_11
#define init_14_3_6_11
void FASTCALL call_14_3_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_6_12
#define init_14_3_6_12
void FASTCALL call_14_3_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_6_13
#define init_14_3_6_13
void FASTCALL call_14_3_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_6_14
#define init_14_3_6_14
void FASTCALL call_14_3_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_6_15
#define init_14_3_6_15
void FASTCALL call_14_3_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_7_0
#define init_14_3_7_0
void FASTCALL call_14_3_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_7_1
#define init_14_3_7_1
void FASTCALL call_14_3_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_7_2
#define init_14_3_7_2
void FASTCALL call_14_3_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_7_3
#define init_14_3_7_3
void FASTCALL call_14_3_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_7_4
#define init_14_3_7_4
void FASTCALL call_14_3_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_7_5
#define init_14_3_7_5
void FASTCALL call_14_3_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_7_6
#define init_14_3_7_6
void FASTCALL call_14_3_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_7_7
#define init_14_3_7_7
void FASTCALL call_14_3_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_7_8
#define init_14_3_7_8
void FASTCALL call_14_3_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_7_9
#define init_14_3_7_9
void FASTCALL call_14_3_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_7_10
#define init_14_3_7_10
void FASTCALL call_14_3_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_7_11
#define init_14_3_7_11
void FASTCALL call_14_3_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_7_12
#define init_14_3_7_12
void FASTCALL call_14_3_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_7_13
#define init_14_3_7_13
void FASTCALL call_14_3_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_7_14
#define init_14_3_7_14
void FASTCALL call_14_3_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_7_15
#define init_14_3_7_15
void FASTCALL call_14_3_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_8_0
#define init_14_3_8_0
void FASTCALL call_14_3_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_8_1
#define init_14_3_8_1
void FASTCALL call_14_3_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_8_2
#define init_14_3_8_2
void FASTCALL call_14_3_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_8_3
#define init_14_3_8_3
void FASTCALL call_14_3_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_8_4
#define init_14_3_8_4
void FASTCALL call_14_3_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_8_5
#define init_14_3_8_5
void FASTCALL call_14_3_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_8_6
#define init_14_3_8_6
void FASTCALL call_14_3_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_8_7
#define init_14_3_8_7
void FASTCALL call_14_3_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_8_8
#define init_14_3_8_8
void FASTCALL call_14_3_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_8_9
#define init_14_3_8_9
void FASTCALL call_14_3_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_8_10
#define init_14_3_8_10
void FASTCALL call_14_3_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_8_11
#define init_14_3_8_11
void FASTCALL call_14_3_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_8_12
#define init_14_3_8_12
void FASTCALL call_14_3_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_8_13
#define init_14_3_8_13
void FASTCALL call_14_3_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_8_14
#define init_14_3_8_14
void FASTCALL call_14_3_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_8_15
#define init_14_3_8_15
void FASTCALL call_14_3_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_9_0
#define init_14_3_9_0
void FASTCALL call_14_3_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_9_1
#define init_14_3_9_1
void FASTCALL call_14_3_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_9_2
#define init_14_3_9_2
void FASTCALL call_14_3_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_9_3
#define init_14_3_9_3
void FASTCALL call_14_3_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_9_4
#define init_14_3_9_4
void FASTCALL call_14_3_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_9_5
#define init_14_3_9_5
void FASTCALL call_14_3_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_9_6
#define init_14_3_9_6
void FASTCALL call_14_3_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_9_7
#define init_14_3_9_7
void FASTCALL call_14_3_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_9_8
#define init_14_3_9_8
void FASTCALL call_14_3_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_9_9
#define init_14_3_9_9
void FASTCALL call_14_3_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_9_10
#define init_14_3_9_10
void FASTCALL call_14_3_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_9_11
#define init_14_3_9_11
void FASTCALL call_14_3_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_9_12
#define init_14_3_9_12
void FASTCALL call_14_3_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_9_13
#define init_14_3_9_13
void FASTCALL call_14_3_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_9_14
#define init_14_3_9_14
void FASTCALL call_14_3_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_9_15
#define init_14_3_9_15
void FASTCALL call_14_3_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_10_0
#define init_14_3_10_0
void FASTCALL call_14_3_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_10_1
#define init_14_3_10_1
void FASTCALL call_14_3_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_10_2
#define init_14_3_10_2
void FASTCALL call_14_3_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_10_3
#define init_14_3_10_3
void FASTCALL call_14_3_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_10_4
#define init_14_3_10_4
void FASTCALL call_14_3_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_10_5
#define init_14_3_10_5
void FASTCALL call_14_3_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_10_6
#define init_14_3_10_6
void FASTCALL call_14_3_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_10_7
#define init_14_3_10_7
void FASTCALL call_14_3_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_10_8
#define init_14_3_10_8
void FASTCALL call_14_3_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_10_9
#define init_14_3_10_9
void FASTCALL call_14_3_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_10_10
#define init_14_3_10_10
void FASTCALL call_14_3_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_10_11
#define init_14_3_10_11
void FASTCALL call_14_3_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_10_12
#define init_14_3_10_12
void FASTCALL call_14_3_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_10_13
#define init_14_3_10_13
void FASTCALL call_14_3_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_10_14
#define init_14_3_10_14
void FASTCALL call_14_3_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_10_15
#define init_14_3_10_15
void FASTCALL call_14_3_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_11_0
#define init_14_3_11_0
void FASTCALL call_14_3_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_11_1
#define init_14_3_11_1
void FASTCALL call_14_3_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_11_2
#define init_14_3_11_2
void FASTCALL call_14_3_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_11_3
#define init_14_3_11_3
void FASTCALL call_14_3_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_11_4
#define init_14_3_11_4
void FASTCALL call_14_3_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_11_5
#define init_14_3_11_5
void FASTCALL call_14_3_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_11_6
#define init_14_3_11_6
void FASTCALL call_14_3_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_11_7
#define init_14_3_11_7
void FASTCALL call_14_3_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_11_8
#define init_14_3_11_8
void FASTCALL call_14_3_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_11_9
#define init_14_3_11_9
void FASTCALL call_14_3_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_11_10
#define init_14_3_11_10
void FASTCALL call_14_3_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_11_11
#define init_14_3_11_11
void FASTCALL call_14_3_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_11_12
#define init_14_3_11_12
void FASTCALL call_14_3_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_11_13
#define init_14_3_11_13
void FASTCALL call_14_3_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_11_14
#define init_14_3_11_14
void FASTCALL call_14_3_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_11_15
#define init_14_3_11_15
void FASTCALL call_14_3_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_12_0
#define init_14_3_12_0
void FASTCALL call_14_3_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_12_1
#define init_14_3_12_1
void FASTCALL call_14_3_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_12_2
#define init_14_3_12_2
void FASTCALL call_14_3_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_12_3
#define init_14_3_12_3
void FASTCALL call_14_3_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_12_4
#define init_14_3_12_4
void FASTCALL call_14_3_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_12_5
#define init_14_3_12_5
void FASTCALL call_14_3_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_12_6
#define init_14_3_12_6
void FASTCALL call_14_3_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_12_7
#define init_14_3_12_7
void FASTCALL call_14_3_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_12_8
#define init_14_3_12_8
void FASTCALL call_14_3_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_12_9
#define init_14_3_12_9
void FASTCALL call_14_3_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_12_10
#define init_14_3_12_10
void FASTCALL call_14_3_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_12_11
#define init_14_3_12_11
void FASTCALL call_14_3_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_12_12
#define init_14_3_12_12
void FASTCALL call_14_3_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_12_13
#define init_14_3_12_13
void FASTCALL call_14_3_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_12_14
#define init_14_3_12_14
void FASTCALL call_14_3_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_12_15
#define init_14_3_12_15
void FASTCALL call_14_3_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_13_0
#define init_14_3_13_0
void FASTCALL call_14_3_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_13_1
#define init_14_3_13_1
void FASTCALL call_14_3_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_13_2
#define init_14_3_13_2
void FASTCALL call_14_3_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_13_3
#define init_14_3_13_3
void FASTCALL call_14_3_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_13_4
#define init_14_3_13_4
void FASTCALL call_14_3_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_13_5
#define init_14_3_13_5
void FASTCALL call_14_3_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_13_6
#define init_14_3_13_6
void FASTCALL call_14_3_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_13_7
#define init_14_3_13_7
void FASTCALL call_14_3_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_13_8
#define init_14_3_13_8
void FASTCALL call_14_3_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_13_9
#define init_14_3_13_9
void FASTCALL call_14_3_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_13_10
#define init_14_3_13_10
void FASTCALL call_14_3_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_13_11
#define init_14_3_13_11
void FASTCALL call_14_3_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_13_12
#define init_14_3_13_12
void FASTCALL call_14_3_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_13_13
#define init_14_3_13_13
void FASTCALL call_14_3_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_13_14
#define init_14_3_13_14
void FASTCALL call_14_3_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_13_15
#define init_14_3_13_15
void FASTCALL call_14_3_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_14_0
#define init_14_3_14_0
void FASTCALL call_14_3_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_14_1
#define init_14_3_14_1
void FASTCALL call_14_3_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_14_2
#define init_14_3_14_2
void FASTCALL call_14_3_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_14_3
#define init_14_3_14_3
void FASTCALL call_14_3_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_14_4
#define init_14_3_14_4
void FASTCALL call_14_3_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_14_5
#define init_14_3_14_5
void FASTCALL call_14_3_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_14_6
#define init_14_3_14_6
void FASTCALL call_14_3_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_14_7
#define init_14_3_14_7
void FASTCALL call_14_3_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_14_8
#define init_14_3_14_8
void FASTCALL call_14_3_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_14_9
#define init_14_3_14_9
void FASTCALL call_14_3_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_14_10
#define init_14_3_14_10
void FASTCALL call_14_3_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_14_11
#define init_14_3_14_11
void FASTCALL call_14_3_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_14_12
#define init_14_3_14_12
void FASTCALL call_14_3_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_14_13
#define init_14_3_14_13
void FASTCALL call_14_3_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_14_14
#define init_14_3_14_14
void FASTCALL call_14_3_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_14_15
#define init_14_3_14_15
void FASTCALL call_14_3_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_15_0
#define init_14_3_15_0
void FASTCALL call_14_3_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_15_1
#define init_14_3_15_1
void FASTCALL call_14_3_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_15_2
#define init_14_3_15_2
void FASTCALL call_14_3_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_15_3
#define init_14_3_15_3
void FASTCALL call_14_3_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_15_4
#define init_14_3_15_4
void FASTCALL call_14_3_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_15_5
#define init_14_3_15_5
void FASTCALL call_14_3_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_15_6
#define init_14_3_15_6
void FASTCALL call_14_3_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_15_7
#define init_14_3_15_7
void FASTCALL call_14_3_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_15_8
#define init_14_3_15_8
void FASTCALL call_14_3_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_15_9
#define init_14_3_15_9
void FASTCALL call_14_3_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_15_10
#define init_14_3_15_10
void FASTCALL call_14_3_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_15_11
#define init_14_3_15_11
void FASTCALL call_14_3_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_15_12
#define init_14_3_15_12
void FASTCALL call_14_3_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_15_13
#define init_14_3_15_13
void FASTCALL call_14_3_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_15_14
#define init_14_3_15_14
void FASTCALL call_14_3_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_3_15_15
#define init_14_3_15_15
void FASTCALL call_14_3_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_0_0
#define init_14_4_0_0
void FASTCALL call_14_4_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_0_1
#define init_14_4_0_1
void FASTCALL call_14_4_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_0_2
#define init_14_4_0_2
void FASTCALL call_14_4_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_0_3
#define init_14_4_0_3
void FASTCALL call_14_4_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_0_4
#define init_14_4_0_4
void FASTCALL call_14_4_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_0_5
#define init_14_4_0_5
void FASTCALL call_14_4_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_0_6
#define init_14_4_0_6
void FASTCALL call_14_4_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_0_7
#define init_14_4_0_7
void FASTCALL call_14_4_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_0_8
#define init_14_4_0_8
void FASTCALL call_14_4_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_0_9
#define init_14_4_0_9
void FASTCALL call_14_4_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_0_10
#define init_14_4_0_10
void FASTCALL call_14_4_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_0_11
#define init_14_4_0_11
void FASTCALL call_14_4_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_0_12
#define init_14_4_0_12
void FASTCALL call_14_4_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_0_13
#define init_14_4_0_13
void FASTCALL call_14_4_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_0_14
#define init_14_4_0_14
void FASTCALL call_14_4_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_0_15
#define init_14_4_0_15
void FASTCALL call_14_4_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_1_0
#define init_14_4_1_0
void FASTCALL call_14_4_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_1_1
#define init_14_4_1_1
void FASTCALL call_14_4_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_1_2
#define init_14_4_1_2
void FASTCALL call_14_4_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_1_3
#define init_14_4_1_3
void FASTCALL call_14_4_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_1_4
#define init_14_4_1_4
void FASTCALL call_14_4_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_1_5
#define init_14_4_1_5
void FASTCALL call_14_4_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_1_6
#define init_14_4_1_6
void FASTCALL call_14_4_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_1_7
#define init_14_4_1_7
void FASTCALL call_14_4_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_1_8
#define init_14_4_1_8
void FASTCALL call_14_4_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_1_9
#define init_14_4_1_9
void FASTCALL call_14_4_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_1_10
#define init_14_4_1_10
void FASTCALL call_14_4_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_1_11
#define init_14_4_1_11
void FASTCALL call_14_4_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_1_12
#define init_14_4_1_12
void FASTCALL call_14_4_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_1_13
#define init_14_4_1_13
void FASTCALL call_14_4_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_1_14
#define init_14_4_1_14
void FASTCALL call_14_4_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_1_15
#define init_14_4_1_15
void FASTCALL call_14_4_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_2_0
#define init_14_4_2_0
void FASTCALL call_14_4_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_2_1
#define init_14_4_2_1
void FASTCALL call_14_4_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_2_2
#define init_14_4_2_2
void FASTCALL call_14_4_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_2_3
#define init_14_4_2_3
void FASTCALL call_14_4_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_2_4
#define init_14_4_2_4
void FASTCALL call_14_4_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_2_5
#define init_14_4_2_5
void FASTCALL call_14_4_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_2_6
#define init_14_4_2_6
void FASTCALL call_14_4_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_2_7
#define init_14_4_2_7
void FASTCALL call_14_4_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_2_8
#define init_14_4_2_8
void FASTCALL call_14_4_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_2_9
#define init_14_4_2_9
void FASTCALL call_14_4_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_2_10
#define init_14_4_2_10
void FASTCALL call_14_4_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_2_11
#define init_14_4_2_11
void FASTCALL call_14_4_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_2_12
#define init_14_4_2_12
void FASTCALL call_14_4_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_2_13
#define init_14_4_2_13
void FASTCALL call_14_4_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_2_14
#define init_14_4_2_14
void FASTCALL call_14_4_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_2_15
#define init_14_4_2_15
void FASTCALL call_14_4_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_3_0
#define init_14_4_3_0
void FASTCALL call_14_4_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_3_1
#define init_14_4_3_1
void FASTCALL call_14_4_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_3_2
#define init_14_4_3_2
void FASTCALL call_14_4_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_3_3
#define init_14_4_3_3
void FASTCALL call_14_4_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_3_4
#define init_14_4_3_4
void FASTCALL call_14_4_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_3_5
#define init_14_4_3_5
void FASTCALL call_14_4_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_3_6
#define init_14_4_3_6
void FASTCALL call_14_4_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_3_7
#define init_14_4_3_7
void FASTCALL call_14_4_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_3_8
#define init_14_4_3_8
void FASTCALL call_14_4_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_3_9
#define init_14_4_3_9
void FASTCALL call_14_4_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_3_10
#define init_14_4_3_10
void FASTCALL call_14_4_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_3_11
#define init_14_4_3_11
void FASTCALL call_14_4_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_3_12
#define init_14_4_3_12
void FASTCALL call_14_4_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_3_13
#define init_14_4_3_13
void FASTCALL call_14_4_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_3_14
#define init_14_4_3_14
void FASTCALL call_14_4_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_3_15
#define init_14_4_3_15
void FASTCALL call_14_4_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_4_0
#define init_14_4_4_0
void FASTCALL call_14_4_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_4_1
#define init_14_4_4_1
void FASTCALL call_14_4_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_4_2
#define init_14_4_4_2
void FASTCALL call_14_4_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_4_3
#define init_14_4_4_3
void FASTCALL call_14_4_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_4_4
#define init_14_4_4_4
void FASTCALL call_14_4_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_4_5
#define init_14_4_4_5
void FASTCALL call_14_4_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_4_6
#define init_14_4_4_6
void FASTCALL call_14_4_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_4_7
#define init_14_4_4_7
void FASTCALL call_14_4_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_4_8
#define init_14_4_4_8
void FASTCALL call_14_4_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_4_9
#define init_14_4_4_9
void FASTCALL call_14_4_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_4_10
#define init_14_4_4_10
void FASTCALL call_14_4_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_4_11
#define init_14_4_4_11
void FASTCALL call_14_4_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_4_12
#define init_14_4_4_12
void FASTCALL call_14_4_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_4_13
#define init_14_4_4_13
void FASTCALL call_14_4_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_4_14
#define init_14_4_4_14
void FASTCALL call_14_4_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_4_15
#define init_14_4_4_15
void FASTCALL call_14_4_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_5_0
#define init_14_4_5_0
void FASTCALL call_14_4_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_5_1
#define init_14_4_5_1
void FASTCALL call_14_4_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_5_2
#define init_14_4_5_2
void FASTCALL call_14_4_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_5_3
#define init_14_4_5_3
void FASTCALL call_14_4_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_5_4
#define init_14_4_5_4
void FASTCALL call_14_4_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_5_5
#define init_14_4_5_5
void FASTCALL call_14_4_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_5_6
#define init_14_4_5_6
void FASTCALL call_14_4_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_5_7
#define init_14_4_5_7
void FASTCALL call_14_4_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_5_8
#define init_14_4_5_8
void FASTCALL call_14_4_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_5_9
#define init_14_4_5_9
void FASTCALL call_14_4_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_5_10
#define init_14_4_5_10
void FASTCALL call_14_4_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_5_11
#define init_14_4_5_11
void FASTCALL call_14_4_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_5_12
#define init_14_4_5_12
void FASTCALL call_14_4_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_5_13
#define init_14_4_5_13
void FASTCALL call_14_4_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_5_14
#define init_14_4_5_14
void FASTCALL call_14_4_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_5_15
#define init_14_4_5_15
void FASTCALL call_14_4_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_6_0
#define init_14_4_6_0
void FASTCALL call_14_4_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_6_1
#define init_14_4_6_1
void FASTCALL call_14_4_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_6_2
#define init_14_4_6_2
void FASTCALL call_14_4_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_6_3
#define init_14_4_6_3
void FASTCALL call_14_4_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_6_4
#define init_14_4_6_4
void FASTCALL call_14_4_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_6_5
#define init_14_4_6_5
void FASTCALL call_14_4_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_6_6
#define init_14_4_6_6
void FASTCALL call_14_4_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_6_7
#define init_14_4_6_7
void FASTCALL call_14_4_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_6_8
#define init_14_4_6_8
void FASTCALL call_14_4_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_6_9
#define init_14_4_6_9
void FASTCALL call_14_4_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_6_10
#define init_14_4_6_10
void FASTCALL call_14_4_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_6_11
#define init_14_4_6_11
void FASTCALL call_14_4_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_6_12
#define init_14_4_6_12
void FASTCALL call_14_4_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_6_13
#define init_14_4_6_13
void FASTCALL call_14_4_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_6_14
#define init_14_4_6_14
void FASTCALL call_14_4_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_6_15
#define init_14_4_6_15
void FASTCALL call_14_4_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_7_0
#define init_14_4_7_0
void FASTCALL call_14_4_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_7_1
#define init_14_4_7_1
void FASTCALL call_14_4_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_7_2
#define init_14_4_7_2
void FASTCALL call_14_4_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_7_3
#define init_14_4_7_3
void FASTCALL call_14_4_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_7_4
#define init_14_4_7_4
void FASTCALL call_14_4_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_7_5
#define init_14_4_7_5
void FASTCALL call_14_4_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_7_6
#define init_14_4_7_6
void FASTCALL call_14_4_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_7_7
#define init_14_4_7_7
void FASTCALL call_14_4_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_7_8
#define init_14_4_7_8
void FASTCALL call_14_4_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_7_9
#define init_14_4_7_9
void FASTCALL call_14_4_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_7_10
#define init_14_4_7_10
void FASTCALL call_14_4_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_7_11
#define init_14_4_7_11
void FASTCALL call_14_4_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_7_12
#define init_14_4_7_12
void FASTCALL call_14_4_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_7_13
#define init_14_4_7_13
void FASTCALL call_14_4_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_7_14
#define init_14_4_7_14
void FASTCALL call_14_4_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_7_15
#define init_14_4_7_15
void FASTCALL call_14_4_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_8_0
#define init_14_4_8_0
void FASTCALL call_14_4_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_8_1
#define init_14_4_8_1
void FASTCALL call_14_4_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_8_2
#define init_14_4_8_2
void FASTCALL call_14_4_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_8_3
#define init_14_4_8_3
void FASTCALL call_14_4_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_8_4
#define init_14_4_8_4
void FASTCALL call_14_4_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_8_5
#define init_14_4_8_5
void FASTCALL call_14_4_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_8_6
#define init_14_4_8_6
void FASTCALL call_14_4_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_8_7
#define init_14_4_8_7
void FASTCALL call_14_4_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_8_8
#define init_14_4_8_8
void FASTCALL call_14_4_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_8_9
#define init_14_4_8_9
void FASTCALL call_14_4_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_8_10
#define init_14_4_8_10
void FASTCALL call_14_4_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_8_11
#define init_14_4_8_11
void FASTCALL call_14_4_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_8_12
#define init_14_4_8_12
void FASTCALL call_14_4_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_8_13
#define init_14_4_8_13
void FASTCALL call_14_4_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_8_14
#define init_14_4_8_14
void FASTCALL call_14_4_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_8_15
#define init_14_4_8_15
void FASTCALL call_14_4_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_9_0
#define init_14_4_9_0
void FASTCALL call_14_4_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_9_1
#define init_14_4_9_1
void FASTCALL call_14_4_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_9_2
#define init_14_4_9_2
void FASTCALL call_14_4_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_9_3
#define init_14_4_9_3
void FASTCALL call_14_4_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_9_4
#define init_14_4_9_4
void FASTCALL call_14_4_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_9_5
#define init_14_4_9_5
void FASTCALL call_14_4_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_9_6
#define init_14_4_9_6
void FASTCALL call_14_4_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_9_7
#define init_14_4_9_7
void FASTCALL call_14_4_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_9_8
#define init_14_4_9_8
void FASTCALL call_14_4_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_9_9
#define init_14_4_9_9
void FASTCALL call_14_4_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_9_10
#define init_14_4_9_10
void FASTCALL call_14_4_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_9_11
#define init_14_4_9_11
void FASTCALL call_14_4_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_9_12
#define init_14_4_9_12
void FASTCALL call_14_4_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_9_13
#define init_14_4_9_13
void FASTCALL call_14_4_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_9_14
#define init_14_4_9_14
void FASTCALL call_14_4_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_9_15
#define init_14_4_9_15
void FASTCALL call_14_4_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_10_0
#define init_14_4_10_0
void FASTCALL call_14_4_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_10_1
#define init_14_4_10_1
void FASTCALL call_14_4_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_10_2
#define init_14_4_10_2
void FASTCALL call_14_4_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_10_3
#define init_14_4_10_3
void FASTCALL call_14_4_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_10_4
#define init_14_4_10_4
void FASTCALL call_14_4_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_10_5
#define init_14_4_10_5
void FASTCALL call_14_4_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_10_6
#define init_14_4_10_6
void FASTCALL call_14_4_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_10_7
#define init_14_4_10_7
void FASTCALL call_14_4_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_10_8
#define init_14_4_10_8
void FASTCALL call_14_4_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_10_9
#define init_14_4_10_9
void FASTCALL call_14_4_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_10_10
#define init_14_4_10_10
void FASTCALL call_14_4_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_10_11
#define init_14_4_10_11
void FASTCALL call_14_4_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_10_12
#define init_14_4_10_12
void FASTCALL call_14_4_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_10_13
#define init_14_4_10_13
void FASTCALL call_14_4_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_10_14
#define init_14_4_10_14
void FASTCALL call_14_4_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_10_15
#define init_14_4_10_15
void FASTCALL call_14_4_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_11_0
#define init_14_4_11_0
void FASTCALL call_14_4_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_11_1
#define init_14_4_11_1
void FASTCALL call_14_4_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_11_2
#define init_14_4_11_2
void FASTCALL call_14_4_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_11_3
#define init_14_4_11_3
void FASTCALL call_14_4_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_11_4
#define init_14_4_11_4
void FASTCALL call_14_4_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_11_5
#define init_14_4_11_5
void FASTCALL call_14_4_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_11_6
#define init_14_4_11_6
void FASTCALL call_14_4_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_11_7
#define init_14_4_11_7
void FASTCALL call_14_4_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_11_8
#define init_14_4_11_8
void FASTCALL call_14_4_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_11_9
#define init_14_4_11_9
void FASTCALL call_14_4_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_11_10
#define init_14_4_11_10
void FASTCALL call_14_4_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_11_11
#define init_14_4_11_11
void FASTCALL call_14_4_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_11_12
#define init_14_4_11_12
void FASTCALL call_14_4_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_11_13
#define init_14_4_11_13
void FASTCALL call_14_4_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_11_14
#define init_14_4_11_14
void FASTCALL call_14_4_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_11_15
#define init_14_4_11_15
void FASTCALL call_14_4_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_12_0
#define init_14_4_12_0
void FASTCALL call_14_4_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_12_1
#define init_14_4_12_1
void FASTCALL call_14_4_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_12_2
#define init_14_4_12_2
void FASTCALL call_14_4_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_12_3
#define init_14_4_12_3
void FASTCALL call_14_4_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_12_4
#define init_14_4_12_4
void FASTCALL call_14_4_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_12_5
#define init_14_4_12_5
void FASTCALL call_14_4_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_12_6
#define init_14_4_12_6
void FASTCALL call_14_4_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_12_7
#define init_14_4_12_7
void FASTCALL call_14_4_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_12_8
#define init_14_4_12_8
void FASTCALL call_14_4_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_12_9
#define init_14_4_12_9
void FASTCALL call_14_4_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_12_10
#define init_14_4_12_10
void FASTCALL call_14_4_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_12_11
#define init_14_4_12_11
void FASTCALL call_14_4_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_12_12
#define init_14_4_12_12
void FASTCALL call_14_4_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_12_13
#define init_14_4_12_13
void FASTCALL call_14_4_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_12_14
#define init_14_4_12_14
void FASTCALL call_14_4_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_12_15
#define init_14_4_12_15
void FASTCALL call_14_4_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_13_0
#define init_14_4_13_0
void FASTCALL call_14_4_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_13_1
#define init_14_4_13_1
void FASTCALL call_14_4_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_13_2
#define init_14_4_13_2
void FASTCALL call_14_4_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_13_3
#define init_14_4_13_3
void FASTCALL call_14_4_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_13_4
#define init_14_4_13_4
void FASTCALL call_14_4_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_13_5
#define init_14_4_13_5
void FASTCALL call_14_4_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_13_6
#define init_14_4_13_6
void FASTCALL call_14_4_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_13_7
#define init_14_4_13_7
void FASTCALL call_14_4_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_13_8
#define init_14_4_13_8
void FASTCALL call_14_4_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_13_9
#define init_14_4_13_9
void FASTCALL call_14_4_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_13_10
#define init_14_4_13_10
void FASTCALL call_14_4_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_13_11
#define init_14_4_13_11
void FASTCALL call_14_4_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_13_12
#define init_14_4_13_12
void FASTCALL call_14_4_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_13_13
#define init_14_4_13_13
void FASTCALL call_14_4_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_13_14
#define init_14_4_13_14
void FASTCALL call_14_4_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_13_15
#define init_14_4_13_15
void FASTCALL call_14_4_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_14_0
#define init_14_4_14_0
void FASTCALL call_14_4_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_14_1
#define init_14_4_14_1
void FASTCALL call_14_4_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_14_2
#define init_14_4_14_2
void FASTCALL call_14_4_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_14_3
#define init_14_4_14_3
void FASTCALL call_14_4_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_14_4
#define init_14_4_14_4
void FASTCALL call_14_4_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_14_5
#define init_14_4_14_5
void FASTCALL call_14_4_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_14_6
#define init_14_4_14_6
void FASTCALL call_14_4_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_14_7
#define init_14_4_14_7
void FASTCALL call_14_4_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_14_8
#define init_14_4_14_8
void FASTCALL call_14_4_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_14_9
#define init_14_4_14_9
void FASTCALL call_14_4_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_14_10
#define init_14_4_14_10
void FASTCALL call_14_4_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_14_11
#define init_14_4_14_11
void FASTCALL call_14_4_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_14_12
#define init_14_4_14_12
void FASTCALL call_14_4_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_14_13
#define init_14_4_14_13
void FASTCALL call_14_4_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_14_14
#define init_14_4_14_14
void FASTCALL call_14_4_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_14_15
#define init_14_4_14_15
void FASTCALL call_14_4_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_15_0
#define init_14_4_15_0
void FASTCALL call_14_4_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_15_1
#define init_14_4_15_1
void FASTCALL call_14_4_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_15_2
#define init_14_4_15_2
void FASTCALL call_14_4_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_15_3
#define init_14_4_15_3
void FASTCALL call_14_4_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_15_4
#define init_14_4_15_4
void FASTCALL call_14_4_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_15_5
#define init_14_4_15_5
void FASTCALL call_14_4_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_15_6
#define init_14_4_15_6
void FASTCALL call_14_4_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_15_7
#define init_14_4_15_7
void FASTCALL call_14_4_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_15_8
#define init_14_4_15_8
void FASTCALL call_14_4_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_15_9
#define init_14_4_15_9
void FASTCALL call_14_4_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_15_10
#define init_14_4_15_10
void FASTCALL call_14_4_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_15_11
#define init_14_4_15_11
void FASTCALL call_14_4_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_15_12
#define init_14_4_15_12
void FASTCALL call_14_4_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_15_13
#define init_14_4_15_13
void FASTCALL call_14_4_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_15_14
#define init_14_4_15_14
void FASTCALL call_14_4_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_4_15_15
#define init_14_4_15_15
void FASTCALL call_14_4_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_0_0
#define init_14_5_0_0
void FASTCALL call_14_5_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_0_1
#define init_14_5_0_1
void FASTCALL call_14_5_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_0_2
#define init_14_5_0_2
void FASTCALL call_14_5_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_0_3
#define init_14_5_0_3
void FASTCALL call_14_5_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_0_4
#define init_14_5_0_4
void FASTCALL call_14_5_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_0_5
#define init_14_5_0_5
void FASTCALL call_14_5_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_0_6
#define init_14_5_0_6
void FASTCALL call_14_5_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_0_7
#define init_14_5_0_7
void FASTCALL call_14_5_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_0_8
#define init_14_5_0_8
void FASTCALL call_14_5_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_0_9
#define init_14_5_0_9
void FASTCALL call_14_5_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_0_10
#define init_14_5_0_10
void FASTCALL call_14_5_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_0_11
#define init_14_5_0_11
void FASTCALL call_14_5_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_0_12
#define init_14_5_0_12
void FASTCALL call_14_5_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_0_13
#define init_14_5_0_13
void FASTCALL call_14_5_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_0_14
#define init_14_5_0_14
void FASTCALL call_14_5_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_0_15
#define init_14_5_0_15
void FASTCALL call_14_5_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_1_0
#define init_14_5_1_0
void FASTCALL call_14_5_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_1_1
#define init_14_5_1_1
void FASTCALL call_14_5_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_1_2
#define init_14_5_1_2
void FASTCALL call_14_5_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_1_3
#define init_14_5_1_3
void FASTCALL call_14_5_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_1_4
#define init_14_5_1_4
void FASTCALL call_14_5_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_1_5
#define init_14_5_1_5
void FASTCALL call_14_5_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_1_6
#define init_14_5_1_6
void FASTCALL call_14_5_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_1_7
#define init_14_5_1_7
void FASTCALL call_14_5_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_1_8
#define init_14_5_1_8
void FASTCALL call_14_5_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_1_9
#define init_14_5_1_9
void FASTCALL call_14_5_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_1_10
#define init_14_5_1_10
void FASTCALL call_14_5_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_1_11
#define init_14_5_1_11
void FASTCALL call_14_5_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_1_12
#define init_14_5_1_12
void FASTCALL call_14_5_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_1_13
#define init_14_5_1_13
void FASTCALL call_14_5_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_1_14
#define init_14_5_1_14
void FASTCALL call_14_5_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_1_15
#define init_14_5_1_15
void FASTCALL call_14_5_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_2_0
#define init_14_5_2_0
void FASTCALL call_14_5_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_2_1
#define init_14_5_2_1
void FASTCALL call_14_5_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_2_2
#define init_14_5_2_2
void FASTCALL call_14_5_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_2_3
#define init_14_5_2_3
void FASTCALL call_14_5_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_2_4
#define init_14_5_2_4
void FASTCALL call_14_5_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_2_5
#define init_14_5_2_5
void FASTCALL call_14_5_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_2_6
#define init_14_5_2_6
void FASTCALL call_14_5_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_2_7
#define init_14_5_2_7
void FASTCALL call_14_5_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_2_8
#define init_14_5_2_8
void FASTCALL call_14_5_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_2_9
#define init_14_5_2_9
void FASTCALL call_14_5_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_2_10
#define init_14_5_2_10
void FASTCALL call_14_5_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_2_11
#define init_14_5_2_11
void FASTCALL call_14_5_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_2_12
#define init_14_5_2_12
void FASTCALL call_14_5_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_2_13
#define init_14_5_2_13
void FASTCALL call_14_5_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_2_14
#define init_14_5_2_14
void FASTCALL call_14_5_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_2_15
#define init_14_5_2_15
void FASTCALL call_14_5_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_3_0
#define init_14_5_3_0
void FASTCALL call_14_5_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_3_1
#define init_14_5_3_1
void FASTCALL call_14_5_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_3_2
#define init_14_5_3_2
void FASTCALL call_14_5_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_3_3
#define init_14_5_3_3
void FASTCALL call_14_5_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_3_4
#define init_14_5_3_4
void FASTCALL call_14_5_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_3_5
#define init_14_5_3_5
void FASTCALL call_14_5_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_3_6
#define init_14_5_3_6
void FASTCALL call_14_5_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_3_7
#define init_14_5_3_7
void FASTCALL call_14_5_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_3_8
#define init_14_5_3_8
void FASTCALL call_14_5_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_3_9
#define init_14_5_3_9
void FASTCALL call_14_5_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_3_10
#define init_14_5_3_10
void FASTCALL call_14_5_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_3_11
#define init_14_5_3_11
void FASTCALL call_14_5_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_3_12
#define init_14_5_3_12
void FASTCALL call_14_5_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_3_13
#define init_14_5_3_13
void FASTCALL call_14_5_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_3_14
#define init_14_5_3_14
void FASTCALL call_14_5_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_3_15
#define init_14_5_3_15
void FASTCALL call_14_5_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_4_0
#define init_14_5_4_0
void FASTCALL call_14_5_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_4_1
#define init_14_5_4_1
void FASTCALL call_14_5_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_4_2
#define init_14_5_4_2
void FASTCALL call_14_5_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_4_3
#define init_14_5_4_3
void FASTCALL call_14_5_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_4_4
#define init_14_5_4_4
void FASTCALL call_14_5_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_4_5
#define init_14_5_4_5
void FASTCALL call_14_5_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_4_6
#define init_14_5_4_6
void FASTCALL call_14_5_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_4_7
#define init_14_5_4_7
void FASTCALL call_14_5_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_4_8
#define init_14_5_4_8
void FASTCALL call_14_5_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_4_9
#define init_14_5_4_9
void FASTCALL call_14_5_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_4_10
#define init_14_5_4_10
void FASTCALL call_14_5_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_4_11
#define init_14_5_4_11
void FASTCALL call_14_5_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_4_12
#define init_14_5_4_12
void FASTCALL call_14_5_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_4_13
#define init_14_5_4_13
void FASTCALL call_14_5_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_4_14
#define init_14_5_4_14
void FASTCALL call_14_5_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_4_15
#define init_14_5_4_15
void FASTCALL call_14_5_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_5_0
#define init_14_5_5_0
void FASTCALL call_14_5_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_5_1
#define init_14_5_5_1
void FASTCALL call_14_5_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_5_2
#define init_14_5_5_2
void FASTCALL call_14_5_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_5_3
#define init_14_5_5_3
void FASTCALL call_14_5_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_5_4
#define init_14_5_5_4
void FASTCALL call_14_5_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_5_5
#define init_14_5_5_5
void FASTCALL call_14_5_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_5_6
#define init_14_5_5_6
void FASTCALL call_14_5_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_5_7
#define init_14_5_5_7
void FASTCALL call_14_5_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_5_8
#define init_14_5_5_8
void FASTCALL call_14_5_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_5_9
#define init_14_5_5_9
void FASTCALL call_14_5_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_5_10
#define init_14_5_5_10
void FASTCALL call_14_5_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_5_11
#define init_14_5_5_11
void FASTCALL call_14_5_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_5_12
#define init_14_5_5_12
void FASTCALL call_14_5_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_5_13
#define init_14_5_5_13
void FASTCALL call_14_5_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_5_14
#define init_14_5_5_14
void FASTCALL call_14_5_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_5_15
#define init_14_5_5_15
void FASTCALL call_14_5_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_6_0
#define init_14_5_6_0
void FASTCALL call_14_5_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_6_1
#define init_14_5_6_1
void FASTCALL call_14_5_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_6_2
#define init_14_5_6_2
void FASTCALL call_14_5_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_6_3
#define init_14_5_6_3
void FASTCALL call_14_5_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_6_4
#define init_14_5_6_4
void FASTCALL call_14_5_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_6_5
#define init_14_5_6_5
void FASTCALL call_14_5_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_6_6
#define init_14_5_6_6
void FASTCALL call_14_5_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_6_7
#define init_14_5_6_7
void FASTCALL call_14_5_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_6_8
#define init_14_5_6_8
void FASTCALL call_14_5_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_6_9
#define init_14_5_6_9
void FASTCALL call_14_5_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_6_10
#define init_14_5_6_10
void FASTCALL call_14_5_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_6_11
#define init_14_5_6_11
void FASTCALL call_14_5_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_6_12
#define init_14_5_6_12
void FASTCALL call_14_5_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_6_13
#define init_14_5_6_13
void FASTCALL call_14_5_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_6_14
#define init_14_5_6_14
void FASTCALL call_14_5_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_6_15
#define init_14_5_6_15
void FASTCALL call_14_5_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_7_0
#define init_14_5_7_0
void FASTCALL call_14_5_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_7_1
#define init_14_5_7_1
void FASTCALL call_14_5_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_7_2
#define init_14_5_7_2
void FASTCALL call_14_5_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_7_3
#define init_14_5_7_3
void FASTCALL call_14_5_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_7_4
#define init_14_5_7_4
void FASTCALL call_14_5_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_7_5
#define init_14_5_7_5
void FASTCALL call_14_5_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_7_6
#define init_14_5_7_6
void FASTCALL call_14_5_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_7_7
#define init_14_5_7_7
void FASTCALL call_14_5_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_7_8
#define init_14_5_7_8
void FASTCALL call_14_5_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_7_9
#define init_14_5_7_9
void FASTCALL call_14_5_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_7_10
#define init_14_5_7_10
void FASTCALL call_14_5_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_7_11
#define init_14_5_7_11
void FASTCALL call_14_5_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_7_12
#define init_14_5_7_12
void FASTCALL call_14_5_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_7_13
#define init_14_5_7_13
void FASTCALL call_14_5_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_7_14
#define init_14_5_7_14
void FASTCALL call_14_5_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_7_15
#define init_14_5_7_15
void FASTCALL call_14_5_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_8_0
#define init_14_5_8_0
void FASTCALL call_14_5_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_8_1
#define init_14_5_8_1
void FASTCALL call_14_5_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_8_2
#define init_14_5_8_2
void FASTCALL call_14_5_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_8_3
#define init_14_5_8_3
void FASTCALL call_14_5_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_8_4
#define init_14_5_8_4
void FASTCALL call_14_5_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_8_5
#define init_14_5_8_5
void FASTCALL call_14_5_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_8_6
#define init_14_5_8_6
void FASTCALL call_14_5_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_8_7
#define init_14_5_8_7
void FASTCALL call_14_5_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_8_8
#define init_14_5_8_8
void FASTCALL call_14_5_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_8_9
#define init_14_5_8_9
void FASTCALL call_14_5_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_8_10
#define init_14_5_8_10
void FASTCALL call_14_5_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_8_11
#define init_14_5_8_11
void FASTCALL call_14_5_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_8_12
#define init_14_5_8_12
void FASTCALL call_14_5_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_8_13
#define init_14_5_8_13
void FASTCALL call_14_5_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_8_14
#define init_14_5_8_14
void FASTCALL call_14_5_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_8_15
#define init_14_5_8_15
void FASTCALL call_14_5_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_9_0
#define init_14_5_9_0
void FASTCALL call_14_5_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_9_1
#define init_14_5_9_1
void FASTCALL call_14_5_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_9_2
#define init_14_5_9_2
void FASTCALL call_14_5_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_9_3
#define init_14_5_9_3
void FASTCALL call_14_5_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_9_4
#define init_14_5_9_4
void FASTCALL call_14_5_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_9_5
#define init_14_5_9_5
void FASTCALL call_14_5_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_9_6
#define init_14_5_9_6
void FASTCALL call_14_5_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_9_7
#define init_14_5_9_7
void FASTCALL call_14_5_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_9_8
#define init_14_5_9_8
void FASTCALL call_14_5_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_9_9
#define init_14_5_9_9
void FASTCALL call_14_5_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_9_10
#define init_14_5_9_10
void FASTCALL call_14_5_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_9_11
#define init_14_5_9_11
void FASTCALL call_14_5_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_9_12
#define init_14_5_9_12
void FASTCALL call_14_5_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_9_13
#define init_14_5_9_13
void FASTCALL call_14_5_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_9_14
#define init_14_5_9_14
void FASTCALL call_14_5_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_9_15
#define init_14_5_9_15
void FASTCALL call_14_5_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_10_0
#define init_14_5_10_0
void FASTCALL call_14_5_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_10_1
#define init_14_5_10_1
void FASTCALL call_14_5_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_10_2
#define init_14_5_10_2
void FASTCALL call_14_5_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_10_3
#define init_14_5_10_3
void FASTCALL call_14_5_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_10_4
#define init_14_5_10_4
void FASTCALL call_14_5_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_10_5
#define init_14_5_10_5
void FASTCALL call_14_5_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_10_6
#define init_14_5_10_6
void FASTCALL call_14_5_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_10_7
#define init_14_5_10_7
void FASTCALL call_14_5_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_10_8
#define init_14_5_10_8
void FASTCALL call_14_5_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_10_9
#define init_14_5_10_9
void FASTCALL call_14_5_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_10_10
#define init_14_5_10_10
void FASTCALL call_14_5_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_10_11
#define init_14_5_10_11
void FASTCALL call_14_5_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_10_12
#define init_14_5_10_12
void FASTCALL call_14_5_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_10_13
#define init_14_5_10_13
void FASTCALL call_14_5_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_10_14
#define init_14_5_10_14
void FASTCALL call_14_5_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_10_15
#define init_14_5_10_15
void FASTCALL call_14_5_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_11_0
#define init_14_5_11_0
void FASTCALL call_14_5_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_11_1
#define init_14_5_11_1
void FASTCALL call_14_5_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_11_2
#define init_14_5_11_2
void FASTCALL call_14_5_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_11_3
#define init_14_5_11_3
void FASTCALL call_14_5_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_11_4
#define init_14_5_11_4
void FASTCALL call_14_5_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_11_5
#define init_14_5_11_5
void FASTCALL call_14_5_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_11_6
#define init_14_5_11_6
void FASTCALL call_14_5_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_11_7
#define init_14_5_11_7
void FASTCALL call_14_5_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_11_8
#define init_14_5_11_8
void FASTCALL call_14_5_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_11_9
#define init_14_5_11_9
void FASTCALL call_14_5_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_11_10
#define init_14_5_11_10
void FASTCALL call_14_5_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_11_11
#define init_14_5_11_11
void FASTCALL call_14_5_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_11_12
#define init_14_5_11_12
void FASTCALL call_14_5_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_11_13
#define init_14_5_11_13
void FASTCALL call_14_5_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_11_14
#define init_14_5_11_14
void FASTCALL call_14_5_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_11_15
#define init_14_5_11_15
void FASTCALL call_14_5_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_12_0
#define init_14_5_12_0
void FASTCALL call_14_5_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_12_1
#define init_14_5_12_1
void FASTCALL call_14_5_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_12_2
#define init_14_5_12_2
void FASTCALL call_14_5_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_12_3
#define init_14_5_12_3
void FASTCALL call_14_5_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_12_4
#define init_14_5_12_4
void FASTCALL call_14_5_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_12_5
#define init_14_5_12_5
void FASTCALL call_14_5_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_12_6
#define init_14_5_12_6
void FASTCALL call_14_5_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_12_7
#define init_14_5_12_7
void FASTCALL call_14_5_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_12_8
#define init_14_5_12_8
void FASTCALL call_14_5_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_12_9
#define init_14_5_12_9
void FASTCALL call_14_5_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_12_10
#define init_14_5_12_10
void FASTCALL call_14_5_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_12_11
#define init_14_5_12_11
void FASTCALL call_14_5_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_12_12
#define init_14_5_12_12
void FASTCALL call_14_5_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_12_13
#define init_14_5_12_13
void FASTCALL call_14_5_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_12_14
#define init_14_5_12_14
void FASTCALL call_14_5_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_12_15
#define init_14_5_12_15
void FASTCALL call_14_5_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_13_0
#define init_14_5_13_0
void FASTCALL call_14_5_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_13_1
#define init_14_5_13_1
void FASTCALL call_14_5_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_13_2
#define init_14_5_13_2
void FASTCALL call_14_5_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_13_3
#define init_14_5_13_3
void FASTCALL call_14_5_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_13_4
#define init_14_5_13_4
void FASTCALL call_14_5_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_13_5
#define init_14_5_13_5
void FASTCALL call_14_5_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_13_6
#define init_14_5_13_6
void FASTCALL call_14_5_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_13_7
#define init_14_5_13_7
void FASTCALL call_14_5_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_13_8
#define init_14_5_13_8
void FASTCALL call_14_5_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_13_9
#define init_14_5_13_9
void FASTCALL call_14_5_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_13_10
#define init_14_5_13_10
void FASTCALL call_14_5_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_13_11
#define init_14_5_13_11
void FASTCALL call_14_5_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_13_12
#define init_14_5_13_12
void FASTCALL call_14_5_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_13_13
#define init_14_5_13_13
void FASTCALL call_14_5_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_13_14
#define init_14_5_13_14
void FASTCALL call_14_5_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_13_15
#define init_14_5_13_15
void FASTCALL call_14_5_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_14_0
#define init_14_5_14_0
void FASTCALL call_14_5_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_14_1
#define init_14_5_14_1
void FASTCALL call_14_5_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_14_2
#define init_14_5_14_2
void FASTCALL call_14_5_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_14_3
#define init_14_5_14_3
void FASTCALL call_14_5_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_14_4
#define init_14_5_14_4
void FASTCALL call_14_5_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_14_5
#define init_14_5_14_5
void FASTCALL call_14_5_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_14_6
#define init_14_5_14_6
void FASTCALL call_14_5_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_14_7
#define init_14_5_14_7
void FASTCALL call_14_5_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_14_8
#define init_14_5_14_8
void FASTCALL call_14_5_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_14_9
#define init_14_5_14_9
void FASTCALL call_14_5_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_14_10
#define init_14_5_14_10
void FASTCALL call_14_5_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_14_11
#define init_14_5_14_11
void FASTCALL call_14_5_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_14_12
#define init_14_5_14_12
void FASTCALL call_14_5_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_14_13
#define init_14_5_14_13
void FASTCALL call_14_5_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_14_14
#define init_14_5_14_14
void FASTCALL call_14_5_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_14_15
#define init_14_5_14_15
void FASTCALL call_14_5_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_15_0
#define init_14_5_15_0
void FASTCALL call_14_5_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_15_1
#define init_14_5_15_1
void FASTCALL call_14_5_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_15_2
#define init_14_5_15_2
void FASTCALL call_14_5_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_15_3
#define init_14_5_15_3
void FASTCALL call_14_5_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_15_4
#define init_14_5_15_4
void FASTCALL call_14_5_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_15_5
#define init_14_5_15_5
void FASTCALL call_14_5_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_15_6
#define init_14_5_15_6
void FASTCALL call_14_5_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_15_7
#define init_14_5_15_7
void FASTCALL call_14_5_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_15_8
#define init_14_5_15_8
void FASTCALL call_14_5_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_15_9
#define init_14_5_15_9
void FASTCALL call_14_5_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_15_10
#define init_14_5_15_10
void FASTCALL call_14_5_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_15_11
#define init_14_5_15_11
void FASTCALL call_14_5_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_15_12
#define init_14_5_15_12
void FASTCALL call_14_5_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_15_13
#define init_14_5_15_13
void FASTCALL call_14_5_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_15_14
#define init_14_5_15_14
void FASTCALL call_14_5_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_5_15_15
#define init_14_5_15_15
void FASTCALL call_14_5_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_0_0
#define init_14_6_0_0
void FASTCALL call_14_6_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_0_1
#define init_14_6_0_1
void FASTCALL call_14_6_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_0_2
#define init_14_6_0_2
void FASTCALL call_14_6_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_0_3
#define init_14_6_0_3
void FASTCALL call_14_6_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_0_4
#define init_14_6_0_4
void FASTCALL call_14_6_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_0_5
#define init_14_6_0_5
void FASTCALL call_14_6_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_0_6
#define init_14_6_0_6
void FASTCALL call_14_6_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_0_7
#define init_14_6_0_7
void FASTCALL call_14_6_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_0_8
#define init_14_6_0_8
void FASTCALL call_14_6_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_0_9
#define init_14_6_0_9
void FASTCALL call_14_6_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_0_10
#define init_14_6_0_10
void FASTCALL call_14_6_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_0_11
#define init_14_6_0_11
void FASTCALL call_14_6_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_0_12
#define init_14_6_0_12
void FASTCALL call_14_6_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_0_13
#define init_14_6_0_13
void FASTCALL call_14_6_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_0_14
#define init_14_6_0_14
void FASTCALL call_14_6_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_0_15
#define init_14_6_0_15
void FASTCALL call_14_6_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_1_0
#define init_14_6_1_0
void FASTCALL call_14_6_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_1_1
#define init_14_6_1_1
void FASTCALL call_14_6_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_1_2
#define init_14_6_1_2
void FASTCALL call_14_6_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_1_3
#define init_14_6_1_3
void FASTCALL call_14_6_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_1_4
#define init_14_6_1_4
void FASTCALL call_14_6_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_1_5
#define init_14_6_1_5
void FASTCALL call_14_6_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_1_6
#define init_14_6_1_6
void FASTCALL call_14_6_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_1_7
#define init_14_6_1_7
void FASTCALL call_14_6_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_1_8
#define init_14_6_1_8
void FASTCALL call_14_6_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_1_9
#define init_14_6_1_9
void FASTCALL call_14_6_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_1_10
#define init_14_6_1_10
void FASTCALL call_14_6_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_1_11
#define init_14_6_1_11
void FASTCALL call_14_6_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_1_12
#define init_14_6_1_12
void FASTCALL call_14_6_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_1_13
#define init_14_6_1_13
void FASTCALL call_14_6_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_1_14
#define init_14_6_1_14
void FASTCALL call_14_6_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_1_15
#define init_14_6_1_15
void FASTCALL call_14_6_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_2_0
#define init_14_6_2_0
void FASTCALL call_14_6_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_2_1
#define init_14_6_2_1
void FASTCALL call_14_6_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_2_2
#define init_14_6_2_2
void FASTCALL call_14_6_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_2_3
#define init_14_6_2_3
void FASTCALL call_14_6_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_2_4
#define init_14_6_2_4
void FASTCALL call_14_6_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_2_5
#define init_14_6_2_5
void FASTCALL call_14_6_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_2_6
#define init_14_6_2_6
void FASTCALL call_14_6_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_2_7
#define init_14_6_2_7
void FASTCALL call_14_6_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_2_8
#define init_14_6_2_8
void FASTCALL call_14_6_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_2_9
#define init_14_6_2_9
void FASTCALL call_14_6_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_2_10
#define init_14_6_2_10
void FASTCALL call_14_6_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_2_11
#define init_14_6_2_11
void FASTCALL call_14_6_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_2_12
#define init_14_6_2_12
void FASTCALL call_14_6_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_2_13
#define init_14_6_2_13
void FASTCALL call_14_6_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_2_14
#define init_14_6_2_14
void FASTCALL call_14_6_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_2_15
#define init_14_6_2_15
void FASTCALL call_14_6_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_3_0
#define init_14_6_3_0
void FASTCALL call_14_6_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_3_1
#define init_14_6_3_1
void FASTCALL call_14_6_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_3_2
#define init_14_6_3_2
void FASTCALL call_14_6_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_3_3
#define init_14_6_3_3
void FASTCALL call_14_6_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_3_4
#define init_14_6_3_4
void FASTCALL call_14_6_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_3_5
#define init_14_6_3_5
void FASTCALL call_14_6_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_3_6
#define init_14_6_3_6
void FASTCALL call_14_6_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_3_7
#define init_14_6_3_7
void FASTCALL call_14_6_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_3_8
#define init_14_6_3_8
void FASTCALL call_14_6_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_3_9
#define init_14_6_3_9
void FASTCALL call_14_6_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_3_10
#define init_14_6_3_10
void FASTCALL call_14_6_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_3_11
#define init_14_6_3_11
void FASTCALL call_14_6_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_3_12
#define init_14_6_3_12
void FASTCALL call_14_6_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_3_13
#define init_14_6_3_13
void FASTCALL call_14_6_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_3_14
#define init_14_6_3_14
void FASTCALL call_14_6_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_3_15
#define init_14_6_3_15
void FASTCALL call_14_6_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_4_0
#define init_14_6_4_0
void FASTCALL call_14_6_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_4_1
#define init_14_6_4_1
void FASTCALL call_14_6_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_4_2
#define init_14_6_4_2
void FASTCALL call_14_6_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_4_3
#define init_14_6_4_3
void FASTCALL call_14_6_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_4_4
#define init_14_6_4_4
void FASTCALL call_14_6_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_4_5
#define init_14_6_4_5
void FASTCALL call_14_6_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_4_6
#define init_14_6_4_6
void FASTCALL call_14_6_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_4_7
#define init_14_6_4_7
void FASTCALL call_14_6_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_4_8
#define init_14_6_4_8
void FASTCALL call_14_6_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_4_9
#define init_14_6_4_9
void FASTCALL call_14_6_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_4_10
#define init_14_6_4_10
void FASTCALL call_14_6_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_4_11
#define init_14_6_4_11
void FASTCALL call_14_6_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_4_12
#define init_14_6_4_12
void FASTCALL call_14_6_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_4_13
#define init_14_6_4_13
void FASTCALL call_14_6_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_4_14
#define init_14_6_4_14
void FASTCALL call_14_6_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_4_15
#define init_14_6_4_15
void FASTCALL call_14_6_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_5_0
#define init_14_6_5_0
void FASTCALL call_14_6_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_5_1
#define init_14_6_5_1
void FASTCALL call_14_6_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_5_2
#define init_14_6_5_2
void FASTCALL call_14_6_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_5_3
#define init_14_6_5_3
void FASTCALL call_14_6_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_5_4
#define init_14_6_5_4
void FASTCALL call_14_6_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_5_5
#define init_14_6_5_5
void FASTCALL call_14_6_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_5_6
#define init_14_6_5_6
void FASTCALL call_14_6_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_5_7
#define init_14_6_5_7
void FASTCALL call_14_6_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_5_8
#define init_14_6_5_8
void FASTCALL call_14_6_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_5_9
#define init_14_6_5_9
void FASTCALL call_14_6_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_5_10
#define init_14_6_5_10
void FASTCALL call_14_6_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_5_11
#define init_14_6_5_11
void FASTCALL call_14_6_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_5_12
#define init_14_6_5_12
void FASTCALL call_14_6_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_5_13
#define init_14_6_5_13
void FASTCALL call_14_6_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_5_14
#define init_14_6_5_14
void FASTCALL call_14_6_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_5_15
#define init_14_6_5_15
void FASTCALL call_14_6_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_6_0
#define init_14_6_6_0
void FASTCALL call_14_6_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_6_1
#define init_14_6_6_1
void FASTCALL call_14_6_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_6_2
#define init_14_6_6_2
void FASTCALL call_14_6_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_6_3
#define init_14_6_6_3
void FASTCALL call_14_6_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_6_4
#define init_14_6_6_4
void FASTCALL call_14_6_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_6_5
#define init_14_6_6_5
void FASTCALL call_14_6_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_6_6
#define init_14_6_6_6
void FASTCALL call_14_6_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_6_7
#define init_14_6_6_7
void FASTCALL call_14_6_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_6_8
#define init_14_6_6_8
void FASTCALL call_14_6_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_6_9
#define init_14_6_6_9
void FASTCALL call_14_6_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_6_10
#define init_14_6_6_10
void FASTCALL call_14_6_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_6_11
#define init_14_6_6_11
void FASTCALL call_14_6_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_6_12
#define init_14_6_6_12
void FASTCALL call_14_6_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_6_13
#define init_14_6_6_13
void FASTCALL call_14_6_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_6_14
#define init_14_6_6_14
void FASTCALL call_14_6_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_6_15
#define init_14_6_6_15
void FASTCALL call_14_6_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_7_0
#define init_14_6_7_0
void FASTCALL call_14_6_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_7_1
#define init_14_6_7_1
void FASTCALL call_14_6_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_7_2
#define init_14_6_7_2
void FASTCALL call_14_6_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_7_3
#define init_14_6_7_3
void FASTCALL call_14_6_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_7_4
#define init_14_6_7_4
void FASTCALL call_14_6_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_7_5
#define init_14_6_7_5
void FASTCALL call_14_6_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_7_6
#define init_14_6_7_6
void FASTCALL call_14_6_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_7_7
#define init_14_6_7_7
void FASTCALL call_14_6_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_7_8
#define init_14_6_7_8
void FASTCALL call_14_6_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_7_9
#define init_14_6_7_9
void FASTCALL call_14_6_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_7_10
#define init_14_6_7_10
void FASTCALL call_14_6_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_7_11
#define init_14_6_7_11
void FASTCALL call_14_6_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_7_12
#define init_14_6_7_12
void FASTCALL call_14_6_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_7_13
#define init_14_6_7_13
void FASTCALL call_14_6_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_7_14
#define init_14_6_7_14
void FASTCALL call_14_6_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_7_15
#define init_14_6_7_15
void FASTCALL call_14_6_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_8_0
#define init_14_6_8_0
void FASTCALL call_14_6_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_8_1
#define init_14_6_8_1
void FASTCALL call_14_6_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_8_2
#define init_14_6_8_2
void FASTCALL call_14_6_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_8_3
#define init_14_6_8_3
void FASTCALL call_14_6_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_8_4
#define init_14_6_8_4
void FASTCALL call_14_6_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_8_5
#define init_14_6_8_5
void FASTCALL call_14_6_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_8_6
#define init_14_6_8_6
void FASTCALL call_14_6_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_8_7
#define init_14_6_8_7
void FASTCALL call_14_6_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_8_8
#define init_14_6_8_8
void FASTCALL call_14_6_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_8_9
#define init_14_6_8_9
void FASTCALL call_14_6_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_8_10
#define init_14_6_8_10
void FASTCALL call_14_6_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_8_11
#define init_14_6_8_11
void FASTCALL call_14_6_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_8_12
#define init_14_6_8_12
void FASTCALL call_14_6_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_8_13
#define init_14_6_8_13
void FASTCALL call_14_6_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_8_14
#define init_14_6_8_14
void FASTCALL call_14_6_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_8_15
#define init_14_6_8_15
void FASTCALL call_14_6_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_9_0
#define init_14_6_9_0
void FASTCALL call_14_6_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_9_1
#define init_14_6_9_1
void FASTCALL call_14_6_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_9_2
#define init_14_6_9_2
void FASTCALL call_14_6_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_9_3
#define init_14_6_9_3
void FASTCALL call_14_6_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_9_4
#define init_14_6_9_4
void FASTCALL call_14_6_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_9_5
#define init_14_6_9_5
void FASTCALL call_14_6_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_9_6
#define init_14_6_9_6
void FASTCALL call_14_6_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_9_7
#define init_14_6_9_7
void FASTCALL call_14_6_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_9_8
#define init_14_6_9_8
void FASTCALL call_14_6_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_9_9
#define init_14_6_9_9
void FASTCALL call_14_6_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_9_10
#define init_14_6_9_10
void FASTCALL call_14_6_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_9_11
#define init_14_6_9_11
void FASTCALL call_14_6_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_9_12
#define init_14_6_9_12
void FASTCALL call_14_6_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_9_13
#define init_14_6_9_13
void FASTCALL call_14_6_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_9_14
#define init_14_6_9_14
void FASTCALL call_14_6_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_9_15
#define init_14_6_9_15
void FASTCALL call_14_6_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_10_0
#define init_14_6_10_0
void FASTCALL call_14_6_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_10_1
#define init_14_6_10_1
void FASTCALL call_14_6_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_10_2
#define init_14_6_10_2
void FASTCALL call_14_6_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_10_3
#define init_14_6_10_3
void FASTCALL call_14_6_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_10_4
#define init_14_6_10_4
void FASTCALL call_14_6_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_10_5
#define init_14_6_10_5
void FASTCALL call_14_6_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_10_6
#define init_14_6_10_6
void FASTCALL call_14_6_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_10_7
#define init_14_6_10_7
void FASTCALL call_14_6_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_10_8
#define init_14_6_10_8
void FASTCALL call_14_6_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_10_9
#define init_14_6_10_9
void FASTCALL call_14_6_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_10_10
#define init_14_6_10_10
void FASTCALL call_14_6_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_10_11
#define init_14_6_10_11
void FASTCALL call_14_6_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_10_12
#define init_14_6_10_12
void FASTCALL call_14_6_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_10_13
#define init_14_6_10_13
void FASTCALL call_14_6_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_10_14
#define init_14_6_10_14
void FASTCALL call_14_6_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_10_15
#define init_14_6_10_15
void FASTCALL call_14_6_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_11_0
#define init_14_6_11_0
void FASTCALL call_14_6_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_11_1
#define init_14_6_11_1
void FASTCALL call_14_6_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_11_2
#define init_14_6_11_2
void FASTCALL call_14_6_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_11_3
#define init_14_6_11_3
void FASTCALL call_14_6_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_11_4
#define init_14_6_11_4
void FASTCALL call_14_6_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_11_5
#define init_14_6_11_5
void FASTCALL call_14_6_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_11_6
#define init_14_6_11_6
void FASTCALL call_14_6_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_11_7
#define init_14_6_11_7
void FASTCALL call_14_6_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_11_8
#define init_14_6_11_8
void FASTCALL call_14_6_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_11_9
#define init_14_6_11_9
void FASTCALL call_14_6_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_11_10
#define init_14_6_11_10
void FASTCALL call_14_6_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_11_11
#define init_14_6_11_11
void FASTCALL call_14_6_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_11_12
#define init_14_6_11_12
void FASTCALL call_14_6_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_11_13
#define init_14_6_11_13
void FASTCALL call_14_6_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_11_14
#define init_14_6_11_14
void FASTCALL call_14_6_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_11_15
#define init_14_6_11_15
void FASTCALL call_14_6_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_12_0
#define init_14_6_12_0
void FASTCALL call_14_6_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_12_1
#define init_14_6_12_1
void FASTCALL call_14_6_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_12_2
#define init_14_6_12_2
void FASTCALL call_14_6_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_12_3
#define init_14_6_12_3
void FASTCALL call_14_6_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_12_4
#define init_14_6_12_4
void FASTCALL call_14_6_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_12_5
#define init_14_6_12_5
void FASTCALL call_14_6_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_12_6
#define init_14_6_12_6
void FASTCALL call_14_6_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_12_7
#define init_14_6_12_7
void FASTCALL call_14_6_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_12_8
#define init_14_6_12_8
void FASTCALL call_14_6_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_12_9
#define init_14_6_12_9
void FASTCALL call_14_6_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_12_10
#define init_14_6_12_10
void FASTCALL call_14_6_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_12_11
#define init_14_6_12_11
void FASTCALL call_14_6_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_12_12
#define init_14_6_12_12
void FASTCALL call_14_6_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_12_13
#define init_14_6_12_13
void FASTCALL call_14_6_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_12_14
#define init_14_6_12_14
void FASTCALL call_14_6_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_12_15
#define init_14_6_12_15
void FASTCALL call_14_6_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_13_0
#define init_14_6_13_0
void FASTCALL call_14_6_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_13_1
#define init_14_6_13_1
void FASTCALL call_14_6_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_13_2
#define init_14_6_13_2
void FASTCALL call_14_6_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_13_3
#define init_14_6_13_3
void FASTCALL call_14_6_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_13_4
#define init_14_6_13_4
void FASTCALL call_14_6_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_13_5
#define init_14_6_13_5
void FASTCALL call_14_6_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_13_6
#define init_14_6_13_6
void FASTCALL call_14_6_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_13_7
#define init_14_6_13_7
void FASTCALL call_14_6_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_13_8
#define init_14_6_13_8
void FASTCALL call_14_6_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_13_9
#define init_14_6_13_9
void FASTCALL call_14_6_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_13_10
#define init_14_6_13_10
void FASTCALL call_14_6_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_13_11
#define init_14_6_13_11
void FASTCALL call_14_6_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_13_12
#define init_14_6_13_12
void FASTCALL call_14_6_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_13_13
#define init_14_6_13_13
void FASTCALL call_14_6_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_13_14
#define init_14_6_13_14
void FASTCALL call_14_6_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_13_15
#define init_14_6_13_15
void FASTCALL call_14_6_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_14_0
#define init_14_6_14_0
void FASTCALL call_14_6_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_14_1
#define init_14_6_14_1
void FASTCALL call_14_6_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_14_2
#define init_14_6_14_2
void FASTCALL call_14_6_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_14_3
#define init_14_6_14_3
void FASTCALL call_14_6_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_14_4
#define init_14_6_14_4
void FASTCALL call_14_6_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_14_5
#define init_14_6_14_5
void FASTCALL call_14_6_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_14_6
#define init_14_6_14_6
void FASTCALL call_14_6_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_14_7
#define init_14_6_14_7
void FASTCALL call_14_6_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_14_8
#define init_14_6_14_8
void FASTCALL call_14_6_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_14_9
#define init_14_6_14_9
void FASTCALL call_14_6_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_14_10
#define init_14_6_14_10
void FASTCALL call_14_6_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_14_11
#define init_14_6_14_11
void FASTCALL call_14_6_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_14_12
#define init_14_6_14_12
void FASTCALL call_14_6_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_14_13
#define init_14_6_14_13
void FASTCALL call_14_6_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_14_14
#define init_14_6_14_14
void FASTCALL call_14_6_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_14_15
#define init_14_6_14_15
void FASTCALL call_14_6_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_15_0
#define init_14_6_15_0
void FASTCALL call_14_6_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_15_1
#define init_14_6_15_1
void FASTCALL call_14_6_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_15_2
#define init_14_6_15_2
void FASTCALL call_14_6_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_15_3
#define init_14_6_15_3
void FASTCALL call_14_6_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_15_4
#define init_14_6_15_4
void FASTCALL call_14_6_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_15_5
#define init_14_6_15_5
void FASTCALL call_14_6_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_15_6
#define init_14_6_15_6
void FASTCALL call_14_6_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_15_7
#define init_14_6_15_7
void FASTCALL call_14_6_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_15_8
#define init_14_6_15_8
void FASTCALL call_14_6_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_15_9
#define init_14_6_15_9
void FASTCALL call_14_6_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_15_10
#define init_14_6_15_10
void FASTCALL call_14_6_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_15_11
#define init_14_6_15_11
void FASTCALL call_14_6_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_15_12
#define init_14_6_15_12
void FASTCALL call_14_6_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_15_13
#define init_14_6_15_13
void FASTCALL call_14_6_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_15_14
#define init_14_6_15_14
void FASTCALL call_14_6_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_6_15_15
#define init_14_6_15_15
void FASTCALL call_14_6_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_0_0
#define init_14_7_0_0
void FASTCALL call_14_7_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_0_1
#define init_14_7_0_1
void FASTCALL call_14_7_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_0_2
#define init_14_7_0_2
void FASTCALL call_14_7_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_0_3
#define init_14_7_0_3
void FASTCALL call_14_7_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_0_4
#define init_14_7_0_4
void FASTCALL call_14_7_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_0_5
#define init_14_7_0_5
void FASTCALL call_14_7_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_0_6
#define init_14_7_0_6
void FASTCALL call_14_7_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_0_7
#define init_14_7_0_7
void FASTCALL call_14_7_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_0_8
#define init_14_7_0_8
void FASTCALL call_14_7_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_0_9
#define init_14_7_0_9
void FASTCALL call_14_7_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_0_10
#define init_14_7_0_10
void FASTCALL call_14_7_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_0_11
#define init_14_7_0_11
void FASTCALL call_14_7_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_0_12
#define init_14_7_0_12
void FASTCALL call_14_7_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_0_13
#define init_14_7_0_13
void FASTCALL call_14_7_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_0_14
#define init_14_7_0_14
void FASTCALL call_14_7_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_0_15
#define init_14_7_0_15
void FASTCALL call_14_7_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_1_0
#define init_14_7_1_0
void FASTCALL call_14_7_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_1_1
#define init_14_7_1_1
void FASTCALL call_14_7_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_1_2
#define init_14_7_1_2
void FASTCALL call_14_7_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_1_3
#define init_14_7_1_3
void FASTCALL call_14_7_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_1_4
#define init_14_7_1_4
void FASTCALL call_14_7_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_1_5
#define init_14_7_1_5
void FASTCALL call_14_7_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_1_6
#define init_14_7_1_6
void FASTCALL call_14_7_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_1_7
#define init_14_7_1_7
void FASTCALL call_14_7_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_1_8
#define init_14_7_1_8
void FASTCALL call_14_7_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_1_9
#define init_14_7_1_9
void FASTCALL call_14_7_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_1_10
#define init_14_7_1_10
void FASTCALL call_14_7_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_1_11
#define init_14_7_1_11
void FASTCALL call_14_7_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_1_12
#define init_14_7_1_12
void FASTCALL call_14_7_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_1_13
#define init_14_7_1_13
void FASTCALL call_14_7_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_1_14
#define init_14_7_1_14
void FASTCALL call_14_7_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_1_15
#define init_14_7_1_15
void FASTCALL call_14_7_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_2_0
#define init_14_7_2_0
void FASTCALL call_14_7_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_2_1
#define init_14_7_2_1
void FASTCALL call_14_7_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_2_2
#define init_14_7_2_2
void FASTCALL call_14_7_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_2_3
#define init_14_7_2_3
void FASTCALL call_14_7_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_2_4
#define init_14_7_2_4
void FASTCALL call_14_7_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_2_5
#define init_14_7_2_5
void FASTCALL call_14_7_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_2_6
#define init_14_7_2_6
void FASTCALL call_14_7_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_2_7
#define init_14_7_2_7
void FASTCALL call_14_7_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_2_8
#define init_14_7_2_8
void FASTCALL call_14_7_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_2_9
#define init_14_7_2_9
void FASTCALL call_14_7_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_2_10
#define init_14_7_2_10
void FASTCALL call_14_7_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_2_11
#define init_14_7_2_11
void FASTCALL call_14_7_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_2_12
#define init_14_7_2_12
void FASTCALL call_14_7_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_2_13
#define init_14_7_2_13
void FASTCALL call_14_7_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_2_14
#define init_14_7_2_14
void FASTCALL call_14_7_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_2_15
#define init_14_7_2_15
void FASTCALL call_14_7_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_3_0
#define init_14_7_3_0
void FASTCALL call_14_7_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_3_1
#define init_14_7_3_1
void FASTCALL call_14_7_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_3_2
#define init_14_7_3_2
void FASTCALL call_14_7_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_3_3
#define init_14_7_3_3
void FASTCALL call_14_7_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_3_4
#define init_14_7_3_4
void FASTCALL call_14_7_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_3_5
#define init_14_7_3_5
void FASTCALL call_14_7_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_3_6
#define init_14_7_3_6
void FASTCALL call_14_7_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_3_7
#define init_14_7_3_7
void FASTCALL call_14_7_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_3_8
#define init_14_7_3_8
void FASTCALL call_14_7_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_3_9
#define init_14_7_3_9
void FASTCALL call_14_7_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_3_10
#define init_14_7_3_10
void FASTCALL call_14_7_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_3_11
#define init_14_7_3_11
void FASTCALL call_14_7_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_3_12
#define init_14_7_3_12
void FASTCALL call_14_7_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_3_13
#define init_14_7_3_13
void FASTCALL call_14_7_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_3_14
#define init_14_7_3_14
void FASTCALL call_14_7_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_3_15
#define init_14_7_3_15
void FASTCALL call_14_7_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_4_0
#define init_14_7_4_0
void FASTCALL call_14_7_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_4_1
#define init_14_7_4_1
void FASTCALL call_14_7_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_4_2
#define init_14_7_4_2
void FASTCALL call_14_7_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_4_3
#define init_14_7_4_3
void FASTCALL call_14_7_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_4_4
#define init_14_7_4_4
void FASTCALL call_14_7_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_4_5
#define init_14_7_4_5
void FASTCALL call_14_7_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_4_6
#define init_14_7_4_6
void FASTCALL call_14_7_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_4_7
#define init_14_7_4_7
void FASTCALL call_14_7_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_4_8
#define init_14_7_4_8
void FASTCALL call_14_7_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_4_9
#define init_14_7_4_9
void FASTCALL call_14_7_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_4_10
#define init_14_7_4_10
void FASTCALL call_14_7_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_4_11
#define init_14_7_4_11
void FASTCALL call_14_7_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_4_12
#define init_14_7_4_12
void FASTCALL call_14_7_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_4_13
#define init_14_7_4_13
void FASTCALL call_14_7_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_4_14
#define init_14_7_4_14
void FASTCALL call_14_7_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_4_15
#define init_14_7_4_15
void FASTCALL call_14_7_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_5_0
#define init_14_7_5_0
void FASTCALL call_14_7_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_5_1
#define init_14_7_5_1
void FASTCALL call_14_7_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_5_2
#define init_14_7_5_2
void FASTCALL call_14_7_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_5_3
#define init_14_7_5_3
void FASTCALL call_14_7_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_5_4
#define init_14_7_5_4
void FASTCALL call_14_7_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_5_5
#define init_14_7_5_5
void FASTCALL call_14_7_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_5_6
#define init_14_7_5_6
void FASTCALL call_14_7_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_5_7
#define init_14_7_5_7
void FASTCALL call_14_7_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_5_8
#define init_14_7_5_8
void FASTCALL call_14_7_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_5_9
#define init_14_7_5_9
void FASTCALL call_14_7_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_5_10
#define init_14_7_5_10
void FASTCALL call_14_7_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_5_11
#define init_14_7_5_11
void FASTCALL call_14_7_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_5_12
#define init_14_7_5_12
void FASTCALL call_14_7_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_5_13
#define init_14_7_5_13
void FASTCALL call_14_7_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_5_14
#define init_14_7_5_14
void FASTCALL call_14_7_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_5_15
#define init_14_7_5_15
void FASTCALL call_14_7_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_6_0
#define init_14_7_6_0
void FASTCALL call_14_7_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_6_1
#define init_14_7_6_1
void FASTCALL call_14_7_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_6_2
#define init_14_7_6_2
void FASTCALL call_14_7_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_6_3
#define init_14_7_6_3
void FASTCALL call_14_7_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_6_4
#define init_14_7_6_4
void FASTCALL call_14_7_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_6_5
#define init_14_7_6_5
void FASTCALL call_14_7_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_6_6
#define init_14_7_6_6
void FASTCALL call_14_7_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_6_7
#define init_14_7_6_7
void FASTCALL call_14_7_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_6_8
#define init_14_7_6_8
void FASTCALL call_14_7_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_6_9
#define init_14_7_6_9
void FASTCALL call_14_7_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_6_10
#define init_14_7_6_10
void FASTCALL call_14_7_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_6_11
#define init_14_7_6_11
void FASTCALL call_14_7_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_6_12
#define init_14_7_6_12
void FASTCALL call_14_7_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_6_13
#define init_14_7_6_13
void FASTCALL call_14_7_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_6_14
#define init_14_7_6_14
void FASTCALL call_14_7_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_6_15
#define init_14_7_6_15
void FASTCALL call_14_7_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_7_0
#define init_14_7_7_0
void FASTCALL call_14_7_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_7_1
#define init_14_7_7_1
void FASTCALL call_14_7_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_7_2
#define init_14_7_7_2
void FASTCALL call_14_7_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_7_3
#define init_14_7_7_3
void FASTCALL call_14_7_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_7_4
#define init_14_7_7_4
void FASTCALL call_14_7_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_7_5
#define init_14_7_7_5
void FASTCALL call_14_7_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_7_6
#define init_14_7_7_6
void FASTCALL call_14_7_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_7_7
#define init_14_7_7_7
void FASTCALL call_14_7_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_7_8
#define init_14_7_7_8
void FASTCALL call_14_7_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_7_9
#define init_14_7_7_9
void FASTCALL call_14_7_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_7_10
#define init_14_7_7_10
void FASTCALL call_14_7_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_7_11
#define init_14_7_7_11
void FASTCALL call_14_7_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_7_12
#define init_14_7_7_12
void FASTCALL call_14_7_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_7_13
#define init_14_7_7_13
void FASTCALL call_14_7_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_7_14
#define init_14_7_7_14
void FASTCALL call_14_7_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_7_15
#define init_14_7_7_15
void FASTCALL call_14_7_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_8_0
#define init_14_7_8_0
void FASTCALL call_14_7_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_8_1
#define init_14_7_8_1
void FASTCALL call_14_7_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_8_2
#define init_14_7_8_2
void FASTCALL call_14_7_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_8_3
#define init_14_7_8_3
void FASTCALL call_14_7_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_8_4
#define init_14_7_8_4
void FASTCALL call_14_7_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_8_5
#define init_14_7_8_5
void FASTCALL call_14_7_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_8_6
#define init_14_7_8_6
void FASTCALL call_14_7_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_8_7
#define init_14_7_8_7
void FASTCALL call_14_7_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_8_8
#define init_14_7_8_8
void FASTCALL call_14_7_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_8_9
#define init_14_7_8_9
void FASTCALL call_14_7_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_8_10
#define init_14_7_8_10
void FASTCALL call_14_7_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_8_11
#define init_14_7_8_11
void FASTCALL call_14_7_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_8_12
#define init_14_7_8_12
void FASTCALL call_14_7_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_8_13
#define init_14_7_8_13
void FASTCALL call_14_7_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_8_14
#define init_14_7_8_14
void FASTCALL call_14_7_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_8_15
#define init_14_7_8_15
void FASTCALL call_14_7_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_9_0
#define init_14_7_9_0
void FASTCALL call_14_7_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_9_1
#define init_14_7_9_1
void FASTCALL call_14_7_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_9_2
#define init_14_7_9_2
void FASTCALL call_14_7_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_9_3
#define init_14_7_9_3
void FASTCALL call_14_7_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_9_4
#define init_14_7_9_4
void FASTCALL call_14_7_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_9_5
#define init_14_7_9_5
void FASTCALL call_14_7_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_9_6
#define init_14_7_9_6
void FASTCALL call_14_7_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_9_7
#define init_14_7_9_7
void FASTCALL call_14_7_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_9_8
#define init_14_7_9_8
void FASTCALL call_14_7_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_9_9
#define init_14_7_9_9
void FASTCALL call_14_7_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_9_10
#define init_14_7_9_10
void FASTCALL call_14_7_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_9_11
#define init_14_7_9_11
void FASTCALL call_14_7_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_9_12
#define init_14_7_9_12
void FASTCALL call_14_7_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_9_13
#define init_14_7_9_13
void FASTCALL call_14_7_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_9_14
#define init_14_7_9_14
void FASTCALL call_14_7_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_9_15
#define init_14_7_9_15
void FASTCALL call_14_7_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_10_0
#define init_14_7_10_0
void FASTCALL call_14_7_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_10_1
#define init_14_7_10_1
void FASTCALL call_14_7_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_10_2
#define init_14_7_10_2
void FASTCALL call_14_7_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_10_3
#define init_14_7_10_3
void FASTCALL call_14_7_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_10_4
#define init_14_7_10_4
void FASTCALL call_14_7_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_10_5
#define init_14_7_10_5
void FASTCALL call_14_7_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_10_6
#define init_14_7_10_6
void FASTCALL call_14_7_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_10_7
#define init_14_7_10_7
void FASTCALL call_14_7_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_10_8
#define init_14_7_10_8
void FASTCALL call_14_7_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_10_9
#define init_14_7_10_9
void FASTCALL call_14_7_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_10_10
#define init_14_7_10_10
void FASTCALL call_14_7_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_10_11
#define init_14_7_10_11
void FASTCALL call_14_7_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_10_12
#define init_14_7_10_12
void FASTCALL call_14_7_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_10_13
#define init_14_7_10_13
void FASTCALL call_14_7_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_10_14
#define init_14_7_10_14
void FASTCALL call_14_7_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_10_15
#define init_14_7_10_15
void FASTCALL call_14_7_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_11_0
#define init_14_7_11_0
void FASTCALL call_14_7_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_11_1
#define init_14_7_11_1
void FASTCALL call_14_7_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_11_2
#define init_14_7_11_2
void FASTCALL call_14_7_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_11_3
#define init_14_7_11_3
void FASTCALL call_14_7_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_11_4
#define init_14_7_11_4
void FASTCALL call_14_7_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_11_5
#define init_14_7_11_5
void FASTCALL call_14_7_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_11_6
#define init_14_7_11_6
void FASTCALL call_14_7_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_11_7
#define init_14_7_11_7
void FASTCALL call_14_7_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_11_8
#define init_14_7_11_8
void FASTCALL call_14_7_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_11_9
#define init_14_7_11_9
void FASTCALL call_14_7_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_11_10
#define init_14_7_11_10
void FASTCALL call_14_7_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_11_11
#define init_14_7_11_11
void FASTCALL call_14_7_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_11_12
#define init_14_7_11_12
void FASTCALL call_14_7_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_11_13
#define init_14_7_11_13
void FASTCALL call_14_7_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_11_14
#define init_14_7_11_14
void FASTCALL call_14_7_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_11_15
#define init_14_7_11_15
void FASTCALL call_14_7_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_12_0
#define init_14_7_12_0
void FASTCALL call_14_7_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_12_1
#define init_14_7_12_1
void FASTCALL call_14_7_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_12_2
#define init_14_7_12_2
void FASTCALL call_14_7_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_12_3
#define init_14_7_12_3
void FASTCALL call_14_7_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_12_4
#define init_14_7_12_4
void FASTCALL call_14_7_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_12_5
#define init_14_7_12_5
void FASTCALL call_14_7_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_12_6
#define init_14_7_12_6
void FASTCALL call_14_7_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_12_7
#define init_14_7_12_7
void FASTCALL call_14_7_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_12_8
#define init_14_7_12_8
void FASTCALL call_14_7_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_12_9
#define init_14_7_12_9
void FASTCALL call_14_7_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_12_10
#define init_14_7_12_10
void FASTCALL call_14_7_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_12_11
#define init_14_7_12_11
void FASTCALL call_14_7_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_12_12
#define init_14_7_12_12
void FASTCALL call_14_7_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_12_13
#define init_14_7_12_13
void FASTCALL call_14_7_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_12_14
#define init_14_7_12_14
void FASTCALL call_14_7_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_12_15
#define init_14_7_12_15
void FASTCALL call_14_7_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_13_0
#define init_14_7_13_0
void FASTCALL call_14_7_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_13_1
#define init_14_7_13_1
void FASTCALL call_14_7_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_13_2
#define init_14_7_13_2
void FASTCALL call_14_7_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_13_3
#define init_14_7_13_3
void FASTCALL call_14_7_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_13_4
#define init_14_7_13_4
void FASTCALL call_14_7_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_13_5
#define init_14_7_13_5
void FASTCALL call_14_7_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_13_6
#define init_14_7_13_6
void FASTCALL call_14_7_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_13_7
#define init_14_7_13_7
void FASTCALL call_14_7_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_13_8
#define init_14_7_13_8
void FASTCALL call_14_7_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_13_9
#define init_14_7_13_9
void FASTCALL call_14_7_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_13_10
#define init_14_7_13_10
void FASTCALL call_14_7_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_13_11
#define init_14_7_13_11
void FASTCALL call_14_7_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_13_12
#define init_14_7_13_12
void FASTCALL call_14_7_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_13_13
#define init_14_7_13_13
void FASTCALL call_14_7_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_13_14
#define init_14_7_13_14
void FASTCALL call_14_7_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_13_15
#define init_14_7_13_15
void FASTCALL call_14_7_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_14_0
#define init_14_7_14_0
void FASTCALL call_14_7_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_14_1
#define init_14_7_14_1
void FASTCALL call_14_7_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_14_2
#define init_14_7_14_2
void FASTCALL call_14_7_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_14_3
#define init_14_7_14_3
void FASTCALL call_14_7_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_14_4
#define init_14_7_14_4
void FASTCALL call_14_7_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_14_5
#define init_14_7_14_5
void FASTCALL call_14_7_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_14_6
#define init_14_7_14_6
void FASTCALL call_14_7_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_14_7
#define init_14_7_14_7
void FASTCALL call_14_7_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_14_8
#define init_14_7_14_8
void FASTCALL call_14_7_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_14_9
#define init_14_7_14_9
void FASTCALL call_14_7_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_14_10
#define init_14_7_14_10
void FASTCALL call_14_7_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_14_11
#define init_14_7_14_11
void FASTCALL call_14_7_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_14_12
#define init_14_7_14_12
void FASTCALL call_14_7_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_14_13
#define init_14_7_14_13
void FASTCALL call_14_7_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_14_14
#define init_14_7_14_14
void FASTCALL call_14_7_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_14_15
#define init_14_7_14_15
void FASTCALL call_14_7_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_15_0
#define init_14_7_15_0
void FASTCALL call_14_7_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_15_1
#define init_14_7_15_1
void FASTCALL call_14_7_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_15_2
#define init_14_7_15_2
void FASTCALL call_14_7_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_15_3
#define init_14_7_15_3
void FASTCALL call_14_7_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_15_4
#define init_14_7_15_4
void FASTCALL call_14_7_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_15_5
#define init_14_7_15_5
void FASTCALL call_14_7_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_15_6
#define init_14_7_15_6
void FASTCALL call_14_7_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_15_7
#define init_14_7_15_7
void FASTCALL call_14_7_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_15_8
#define init_14_7_15_8
void FASTCALL call_14_7_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_15_9
#define init_14_7_15_9
void FASTCALL call_14_7_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_15_10
#define init_14_7_15_10
void FASTCALL call_14_7_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_15_11
#define init_14_7_15_11
void FASTCALL call_14_7_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_15_12
#define init_14_7_15_12
void FASTCALL call_14_7_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_15_13
#define init_14_7_15_13
void FASTCALL call_14_7_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_15_14
#define init_14_7_15_14
void FASTCALL call_14_7_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_7_15_15
#define init_14_7_15_15
void FASTCALL call_14_7_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_0_0
#define init_14_8_0_0
void FASTCALL call_14_8_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_0_1
#define init_14_8_0_1
void FASTCALL call_14_8_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_0_2
#define init_14_8_0_2
void FASTCALL call_14_8_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_0_3
#define init_14_8_0_3
void FASTCALL call_14_8_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_0_4
#define init_14_8_0_4
void FASTCALL call_14_8_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_0_5
#define init_14_8_0_5
void FASTCALL call_14_8_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_0_6
#define init_14_8_0_6
void FASTCALL call_14_8_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_0_7
#define init_14_8_0_7
void FASTCALL call_14_8_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_0_8
#define init_14_8_0_8
void FASTCALL call_14_8_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_0_9
#define init_14_8_0_9
void FASTCALL call_14_8_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_0_10
#define init_14_8_0_10
void FASTCALL call_14_8_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_0_11
#define init_14_8_0_11
void FASTCALL call_14_8_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_0_12
#define init_14_8_0_12
void FASTCALL call_14_8_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_0_13
#define init_14_8_0_13
void FASTCALL call_14_8_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_0_14
#define init_14_8_0_14
void FASTCALL call_14_8_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_0_15
#define init_14_8_0_15
void FASTCALL call_14_8_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_1_0
#define init_14_8_1_0
void FASTCALL call_14_8_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_1_1
#define init_14_8_1_1
void FASTCALL call_14_8_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_1_2
#define init_14_8_1_2
void FASTCALL call_14_8_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_1_3
#define init_14_8_1_3
void FASTCALL call_14_8_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_1_4
#define init_14_8_1_4
void FASTCALL call_14_8_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_1_5
#define init_14_8_1_5
void FASTCALL call_14_8_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_1_6
#define init_14_8_1_6
void FASTCALL call_14_8_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_1_7
#define init_14_8_1_7
void FASTCALL call_14_8_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_1_8
#define init_14_8_1_8
void FASTCALL call_14_8_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_1_9
#define init_14_8_1_9
void FASTCALL call_14_8_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_1_10
#define init_14_8_1_10
void FASTCALL call_14_8_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_1_11
#define init_14_8_1_11
void FASTCALL call_14_8_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_1_12
#define init_14_8_1_12
void FASTCALL call_14_8_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_1_13
#define init_14_8_1_13
void FASTCALL call_14_8_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_1_14
#define init_14_8_1_14
void FASTCALL call_14_8_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_1_15
#define init_14_8_1_15
void FASTCALL call_14_8_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_2_0
#define init_14_8_2_0
void FASTCALL call_14_8_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_2_1
#define init_14_8_2_1
void FASTCALL call_14_8_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_2_2
#define init_14_8_2_2
void FASTCALL call_14_8_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_2_3
#define init_14_8_2_3
void FASTCALL call_14_8_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_2_4
#define init_14_8_2_4
void FASTCALL call_14_8_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_2_5
#define init_14_8_2_5
void FASTCALL call_14_8_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_2_6
#define init_14_8_2_6
void FASTCALL call_14_8_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_2_7
#define init_14_8_2_7
void FASTCALL call_14_8_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_2_8
#define init_14_8_2_8
void FASTCALL call_14_8_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_2_9
#define init_14_8_2_9
void FASTCALL call_14_8_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_2_10
#define init_14_8_2_10
void FASTCALL call_14_8_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_2_11
#define init_14_8_2_11
void FASTCALL call_14_8_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_2_12
#define init_14_8_2_12
void FASTCALL call_14_8_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_2_13
#define init_14_8_2_13
void FASTCALL call_14_8_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_2_14
#define init_14_8_2_14
void FASTCALL call_14_8_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_2_15
#define init_14_8_2_15
void FASTCALL call_14_8_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_3_0
#define init_14_8_3_0
void FASTCALL call_14_8_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_3_1
#define init_14_8_3_1
void FASTCALL call_14_8_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_3_2
#define init_14_8_3_2
void FASTCALL call_14_8_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_3_3
#define init_14_8_3_3
void FASTCALL call_14_8_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_3_4
#define init_14_8_3_4
void FASTCALL call_14_8_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_3_5
#define init_14_8_3_5
void FASTCALL call_14_8_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_3_6
#define init_14_8_3_6
void FASTCALL call_14_8_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_3_7
#define init_14_8_3_7
void FASTCALL call_14_8_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_3_8
#define init_14_8_3_8
void FASTCALL call_14_8_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_3_9
#define init_14_8_3_9
void FASTCALL call_14_8_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_3_10
#define init_14_8_3_10
void FASTCALL call_14_8_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_3_11
#define init_14_8_3_11
void FASTCALL call_14_8_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_3_12
#define init_14_8_3_12
void FASTCALL call_14_8_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_3_13
#define init_14_8_3_13
void FASTCALL call_14_8_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_3_14
#define init_14_8_3_14
void FASTCALL call_14_8_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_3_15
#define init_14_8_3_15
void FASTCALL call_14_8_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_4_0
#define init_14_8_4_0
void FASTCALL call_14_8_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_4_1
#define init_14_8_4_1
void FASTCALL call_14_8_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_4_2
#define init_14_8_4_2
void FASTCALL call_14_8_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_4_3
#define init_14_8_4_3
void FASTCALL call_14_8_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_4_4
#define init_14_8_4_4
void FASTCALL call_14_8_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_4_5
#define init_14_8_4_5
void FASTCALL call_14_8_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_4_6
#define init_14_8_4_6
void FASTCALL call_14_8_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_4_7
#define init_14_8_4_7
void FASTCALL call_14_8_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_4_8
#define init_14_8_4_8
void FASTCALL call_14_8_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_4_9
#define init_14_8_4_9
void FASTCALL call_14_8_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_4_10
#define init_14_8_4_10
void FASTCALL call_14_8_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_4_11
#define init_14_8_4_11
void FASTCALL call_14_8_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_4_12
#define init_14_8_4_12
void FASTCALL call_14_8_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_4_13
#define init_14_8_4_13
void FASTCALL call_14_8_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_4_14
#define init_14_8_4_14
void FASTCALL call_14_8_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_4_15
#define init_14_8_4_15
void FASTCALL call_14_8_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_5_0
#define init_14_8_5_0
void FASTCALL call_14_8_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_5_1
#define init_14_8_5_1
void FASTCALL call_14_8_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_5_2
#define init_14_8_5_2
void FASTCALL call_14_8_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_5_3
#define init_14_8_5_3
void FASTCALL call_14_8_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_5_4
#define init_14_8_5_4
void FASTCALL call_14_8_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_5_5
#define init_14_8_5_5
void FASTCALL call_14_8_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_5_6
#define init_14_8_5_6
void FASTCALL call_14_8_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_5_7
#define init_14_8_5_7
void FASTCALL call_14_8_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_5_8
#define init_14_8_5_8
void FASTCALL call_14_8_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_5_9
#define init_14_8_5_9
void FASTCALL call_14_8_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_5_10
#define init_14_8_5_10
void FASTCALL call_14_8_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_5_11
#define init_14_8_5_11
void FASTCALL call_14_8_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_5_12
#define init_14_8_5_12
void FASTCALL call_14_8_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_5_13
#define init_14_8_5_13
void FASTCALL call_14_8_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_5_14
#define init_14_8_5_14
void FASTCALL call_14_8_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_5_15
#define init_14_8_5_15
void FASTCALL call_14_8_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_6_0
#define init_14_8_6_0
void FASTCALL call_14_8_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_6_1
#define init_14_8_6_1
void FASTCALL call_14_8_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_6_2
#define init_14_8_6_2
void FASTCALL call_14_8_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_6_3
#define init_14_8_6_3
void FASTCALL call_14_8_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_6_4
#define init_14_8_6_4
void FASTCALL call_14_8_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_6_5
#define init_14_8_6_5
void FASTCALL call_14_8_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_6_6
#define init_14_8_6_6
void FASTCALL call_14_8_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_6_7
#define init_14_8_6_7
void FASTCALL call_14_8_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_6_8
#define init_14_8_6_8
void FASTCALL call_14_8_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_6_9
#define init_14_8_6_9
void FASTCALL call_14_8_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_6_10
#define init_14_8_6_10
void FASTCALL call_14_8_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_6_11
#define init_14_8_6_11
void FASTCALL call_14_8_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_6_12
#define init_14_8_6_12
void FASTCALL call_14_8_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_6_13
#define init_14_8_6_13
void FASTCALL call_14_8_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_6_14
#define init_14_8_6_14
void FASTCALL call_14_8_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_6_15
#define init_14_8_6_15
void FASTCALL call_14_8_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_7_0
#define init_14_8_7_0
void FASTCALL call_14_8_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_7_1
#define init_14_8_7_1
void FASTCALL call_14_8_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_7_2
#define init_14_8_7_2
void FASTCALL call_14_8_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_7_3
#define init_14_8_7_3
void FASTCALL call_14_8_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_7_4
#define init_14_8_7_4
void FASTCALL call_14_8_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_7_5
#define init_14_8_7_5
void FASTCALL call_14_8_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_7_6
#define init_14_8_7_6
void FASTCALL call_14_8_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_7_7
#define init_14_8_7_7
void FASTCALL call_14_8_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_7_8
#define init_14_8_7_8
void FASTCALL call_14_8_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_7_9
#define init_14_8_7_9
void FASTCALL call_14_8_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_7_10
#define init_14_8_7_10
void FASTCALL call_14_8_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_7_11
#define init_14_8_7_11
void FASTCALL call_14_8_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_7_12
#define init_14_8_7_12
void FASTCALL call_14_8_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_7_13
#define init_14_8_7_13
void FASTCALL call_14_8_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_7_14
#define init_14_8_7_14
void FASTCALL call_14_8_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_7_15
#define init_14_8_7_15
void FASTCALL call_14_8_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_8_0
#define init_14_8_8_0
void FASTCALL call_14_8_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_8_1
#define init_14_8_8_1
void FASTCALL call_14_8_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_8_2
#define init_14_8_8_2
void FASTCALL call_14_8_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_8_3
#define init_14_8_8_3
void FASTCALL call_14_8_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_8_4
#define init_14_8_8_4
void FASTCALL call_14_8_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_8_5
#define init_14_8_8_5
void FASTCALL call_14_8_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_8_6
#define init_14_8_8_6
void FASTCALL call_14_8_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_8_7
#define init_14_8_8_7
void FASTCALL call_14_8_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_8_8
#define init_14_8_8_8
void FASTCALL call_14_8_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_8_9
#define init_14_8_8_9
void FASTCALL call_14_8_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_8_10
#define init_14_8_8_10
void FASTCALL call_14_8_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_8_11
#define init_14_8_8_11
void FASTCALL call_14_8_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_8_12
#define init_14_8_8_12
void FASTCALL call_14_8_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_8_13
#define init_14_8_8_13
void FASTCALL call_14_8_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_8_14
#define init_14_8_8_14
void FASTCALL call_14_8_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_8_15
#define init_14_8_8_15
void FASTCALL call_14_8_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_9_0
#define init_14_8_9_0
void FASTCALL call_14_8_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_9_1
#define init_14_8_9_1
void FASTCALL call_14_8_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_9_2
#define init_14_8_9_2
void FASTCALL call_14_8_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_9_3
#define init_14_8_9_3
void FASTCALL call_14_8_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_9_4
#define init_14_8_9_4
void FASTCALL call_14_8_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_9_5
#define init_14_8_9_5
void FASTCALL call_14_8_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_9_6
#define init_14_8_9_6
void FASTCALL call_14_8_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_9_7
#define init_14_8_9_7
void FASTCALL call_14_8_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_9_8
#define init_14_8_9_8
void FASTCALL call_14_8_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_9_9
#define init_14_8_9_9
void FASTCALL call_14_8_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_9_10
#define init_14_8_9_10
void FASTCALL call_14_8_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_9_11
#define init_14_8_9_11
void FASTCALL call_14_8_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_9_12
#define init_14_8_9_12
void FASTCALL call_14_8_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_9_13
#define init_14_8_9_13
void FASTCALL call_14_8_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_9_14
#define init_14_8_9_14
void FASTCALL call_14_8_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_9_15
#define init_14_8_9_15
void FASTCALL call_14_8_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_10_0
#define init_14_8_10_0
void FASTCALL call_14_8_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_10_1
#define init_14_8_10_1
void FASTCALL call_14_8_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_10_2
#define init_14_8_10_2
void FASTCALL call_14_8_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_10_3
#define init_14_8_10_3
void FASTCALL call_14_8_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_10_4
#define init_14_8_10_4
void FASTCALL call_14_8_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_10_5
#define init_14_8_10_5
void FASTCALL call_14_8_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_10_6
#define init_14_8_10_6
void FASTCALL call_14_8_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_10_7
#define init_14_8_10_7
void FASTCALL call_14_8_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_10_8
#define init_14_8_10_8
void FASTCALL call_14_8_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_10_9
#define init_14_8_10_9
void FASTCALL call_14_8_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_10_10
#define init_14_8_10_10
void FASTCALL call_14_8_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_10_11
#define init_14_8_10_11
void FASTCALL call_14_8_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_10_12
#define init_14_8_10_12
void FASTCALL call_14_8_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_10_13
#define init_14_8_10_13
void FASTCALL call_14_8_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_10_14
#define init_14_8_10_14
void FASTCALL call_14_8_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_10_15
#define init_14_8_10_15
void FASTCALL call_14_8_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_11_0
#define init_14_8_11_0
void FASTCALL call_14_8_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_11_1
#define init_14_8_11_1
void FASTCALL call_14_8_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_11_2
#define init_14_8_11_2
void FASTCALL call_14_8_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_11_3
#define init_14_8_11_3
void FASTCALL call_14_8_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_11_4
#define init_14_8_11_4
void FASTCALL call_14_8_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_11_5
#define init_14_8_11_5
void FASTCALL call_14_8_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_11_6
#define init_14_8_11_6
void FASTCALL call_14_8_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_11_7
#define init_14_8_11_7
void FASTCALL call_14_8_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_11_8
#define init_14_8_11_8
void FASTCALL call_14_8_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_11_9
#define init_14_8_11_9
void FASTCALL call_14_8_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_11_10
#define init_14_8_11_10
void FASTCALL call_14_8_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_11_11
#define init_14_8_11_11
void FASTCALL call_14_8_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_11_12
#define init_14_8_11_12
void FASTCALL call_14_8_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_11_13
#define init_14_8_11_13
void FASTCALL call_14_8_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_11_14
#define init_14_8_11_14
void FASTCALL call_14_8_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_11_15
#define init_14_8_11_15
void FASTCALL call_14_8_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_12_0
#define init_14_8_12_0
void FASTCALL call_14_8_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_12_1
#define init_14_8_12_1
void FASTCALL call_14_8_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_12_2
#define init_14_8_12_2
void FASTCALL call_14_8_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_12_3
#define init_14_8_12_3
void FASTCALL call_14_8_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_12_4
#define init_14_8_12_4
void FASTCALL call_14_8_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_12_5
#define init_14_8_12_5
void FASTCALL call_14_8_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_12_6
#define init_14_8_12_6
void FASTCALL call_14_8_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_12_7
#define init_14_8_12_7
void FASTCALL call_14_8_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_12_8
#define init_14_8_12_8
void FASTCALL call_14_8_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_12_9
#define init_14_8_12_9
void FASTCALL call_14_8_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_12_10
#define init_14_8_12_10
void FASTCALL call_14_8_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_12_11
#define init_14_8_12_11
void FASTCALL call_14_8_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_12_12
#define init_14_8_12_12
void FASTCALL call_14_8_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_12_13
#define init_14_8_12_13
void FASTCALL call_14_8_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_12_14
#define init_14_8_12_14
void FASTCALL call_14_8_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_12_15
#define init_14_8_12_15
void FASTCALL call_14_8_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_13_0
#define init_14_8_13_0
void FASTCALL call_14_8_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_13_1
#define init_14_8_13_1
void FASTCALL call_14_8_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_13_2
#define init_14_8_13_2
void FASTCALL call_14_8_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_13_3
#define init_14_8_13_3
void FASTCALL call_14_8_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_13_4
#define init_14_8_13_4
void FASTCALL call_14_8_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_13_5
#define init_14_8_13_5
void FASTCALL call_14_8_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_13_6
#define init_14_8_13_6
void FASTCALL call_14_8_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_13_7
#define init_14_8_13_7
void FASTCALL call_14_8_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_13_8
#define init_14_8_13_8
void FASTCALL call_14_8_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_13_9
#define init_14_8_13_9
void FASTCALL call_14_8_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_13_10
#define init_14_8_13_10
void FASTCALL call_14_8_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_13_11
#define init_14_8_13_11
void FASTCALL call_14_8_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_13_12
#define init_14_8_13_12
void FASTCALL call_14_8_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_13_13
#define init_14_8_13_13
void FASTCALL call_14_8_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_13_14
#define init_14_8_13_14
void FASTCALL call_14_8_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_13_15
#define init_14_8_13_15
void FASTCALL call_14_8_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_14_0
#define init_14_8_14_0
void FASTCALL call_14_8_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_14_1
#define init_14_8_14_1
void FASTCALL call_14_8_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_14_2
#define init_14_8_14_2
void FASTCALL call_14_8_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_14_3
#define init_14_8_14_3
void FASTCALL call_14_8_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_14_4
#define init_14_8_14_4
void FASTCALL call_14_8_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_14_5
#define init_14_8_14_5
void FASTCALL call_14_8_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_14_6
#define init_14_8_14_6
void FASTCALL call_14_8_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_14_7
#define init_14_8_14_7
void FASTCALL call_14_8_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_14_8
#define init_14_8_14_8
void FASTCALL call_14_8_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_14_9
#define init_14_8_14_9
void FASTCALL call_14_8_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_14_10
#define init_14_8_14_10
void FASTCALL call_14_8_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_14_11
#define init_14_8_14_11
void FASTCALL call_14_8_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_14_12
#define init_14_8_14_12
void FASTCALL call_14_8_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_14_13
#define init_14_8_14_13
void FASTCALL call_14_8_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_14_14
#define init_14_8_14_14
void FASTCALL call_14_8_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_14_15
#define init_14_8_14_15
void FASTCALL call_14_8_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_15_0
#define init_14_8_15_0
void FASTCALL call_14_8_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_15_1
#define init_14_8_15_1
void FASTCALL call_14_8_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_15_2
#define init_14_8_15_2
void FASTCALL call_14_8_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_15_3
#define init_14_8_15_3
void FASTCALL call_14_8_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_15_4
#define init_14_8_15_4
void FASTCALL call_14_8_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_15_5
#define init_14_8_15_5
void FASTCALL call_14_8_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_15_6
#define init_14_8_15_6
void FASTCALL call_14_8_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_15_7
#define init_14_8_15_7
void FASTCALL call_14_8_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_15_8
#define init_14_8_15_8
void FASTCALL call_14_8_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_15_9
#define init_14_8_15_9
void FASTCALL call_14_8_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_15_10
#define init_14_8_15_10
void FASTCALL call_14_8_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_15_11
#define init_14_8_15_11
void FASTCALL call_14_8_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_15_12
#define init_14_8_15_12
void FASTCALL call_14_8_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_15_13
#define init_14_8_15_13
void FASTCALL call_14_8_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_15_14
#define init_14_8_15_14
void FASTCALL call_14_8_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_8_15_15
#define init_14_8_15_15
void FASTCALL call_14_8_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_0_0
#define init_14_9_0_0
void FASTCALL call_14_9_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_0_1
#define init_14_9_0_1
void FASTCALL call_14_9_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_0_2
#define init_14_9_0_2
void FASTCALL call_14_9_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_0_3
#define init_14_9_0_3
void FASTCALL call_14_9_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_0_4
#define init_14_9_0_4
void FASTCALL call_14_9_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_0_5
#define init_14_9_0_5
void FASTCALL call_14_9_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_0_6
#define init_14_9_0_6
void FASTCALL call_14_9_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_0_7
#define init_14_9_0_7
void FASTCALL call_14_9_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_0_8
#define init_14_9_0_8
void FASTCALL call_14_9_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_0_9
#define init_14_9_0_9
void FASTCALL call_14_9_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_0_10
#define init_14_9_0_10
void FASTCALL call_14_9_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_0_11
#define init_14_9_0_11
void FASTCALL call_14_9_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_0_12
#define init_14_9_0_12
void FASTCALL call_14_9_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_0_13
#define init_14_9_0_13
void FASTCALL call_14_9_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_0_14
#define init_14_9_0_14
void FASTCALL call_14_9_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_0_15
#define init_14_9_0_15
void FASTCALL call_14_9_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_1_0
#define init_14_9_1_0
void FASTCALL call_14_9_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_1_1
#define init_14_9_1_1
void FASTCALL call_14_9_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_1_2
#define init_14_9_1_2
void FASTCALL call_14_9_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_1_3
#define init_14_9_1_3
void FASTCALL call_14_9_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_1_4
#define init_14_9_1_4
void FASTCALL call_14_9_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_1_5
#define init_14_9_1_5
void FASTCALL call_14_9_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_1_6
#define init_14_9_1_6
void FASTCALL call_14_9_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_1_7
#define init_14_9_1_7
void FASTCALL call_14_9_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_1_8
#define init_14_9_1_8
void FASTCALL call_14_9_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_1_9
#define init_14_9_1_9
void FASTCALL call_14_9_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_1_10
#define init_14_9_1_10
void FASTCALL call_14_9_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_1_11
#define init_14_9_1_11
void FASTCALL call_14_9_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_1_12
#define init_14_9_1_12
void FASTCALL call_14_9_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_1_13
#define init_14_9_1_13
void FASTCALL call_14_9_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_1_14
#define init_14_9_1_14
void FASTCALL call_14_9_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_1_15
#define init_14_9_1_15
void FASTCALL call_14_9_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_2_0
#define init_14_9_2_0
void FASTCALL call_14_9_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_2_1
#define init_14_9_2_1
void FASTCALL call_14_9_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_2_2
#define init_14_9_2_2
void FASTCALL call_14_9_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_2_3
#define init_14_9_2_3
void FASTCALL call_14_9_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_2_4
#define init_14_9_2_4
void FASTCALL call_14_9_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_2_5
#define init_14_9_2_5
void FASTCALL call_14_9_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_2_6
#define init_14_9_2_6
void FASTCALL call_14_9_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_2_7
#define init_14_9_2_7
void FASTCALL call_14_9_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_2_8
#define init_14_9_2_8
void FASTCALL call_14_9_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_2_9
#define init_14_9_2_9
void FASTCALL call_14_9_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_2_10
#define init_14_9_2_10
void FASTCALL call_14_9_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_2_11
#define init_14_9_2_11
void FASTCALL call_14_9_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_2_12
#define init_14_9_2_12
void FASTCALL call_14_9_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_2_13
#define init_14_9_2_13
void FASTCALL call_14_9_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_2_14
#define init_14_9_2_14
void FASTCALL call_14_9_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_2_15
#define init_14_9_2_15
void FASTCALL call_14_9_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_3_0
#define init_14_9_3_0
void FASTCALL call_14_9_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_3_1
#define init_14_9_3_1
void FASTCALL call_14_9_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_3_2
#define init_14_9_3_2
void FASTCALL call_14_9_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_3_3
#define init_14_9_3_3
void FASTCALL call_14_9_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_3_4
#define init_14_9_3_4
void FASTCALL call_14_9_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_3_5
#define init_14_9_3_5
void FASTCALL call_14_9_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_3_6
#define init_14_9_3_6
void FASTCALL call_14_9_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_3_7
#define init_14_9_3_7
void FASTCALL call_14_9_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_3_8
#define init_14_9_3_8
void FASTCALL call_14_9_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_3_9
#define init_14_9_3_9
void FASTCALL call_14_9_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_3_10
#define init_14_9_3_10
void FASTCALL call_14_9_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_3_11
#define init_14_9_3_11
void FASTCALL call_14_9_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_3_12
#define init_14_9_3_12
void FASTCALL call_14_9_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_3_13
#define init_14_9_3_13
void FASTCALL call_14_9_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_3_14
#define init_14_9_3_14
void FASTCALL call_14_9_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_3_15
#define init_14_9_3_15
void FASTCALL call_14_9_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_4_0
#define init_14_9_4_0
void FASTCALL call_14_9_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_4_1
#define init_14_9_4_1
void FASTCALL call_14_9_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_4_2
#define init_14_9_4_2
void FASTCALL call_14_9_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_4_3
#define init_14_9_4_3
void FASTCALL call_14_9_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_4_4
#define init_14_9_4_4
void FASTCALL call_14_9_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_4_5
#define init_14_9_4_5
void FASTCALL call_14_9_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_4_6
#define init_14_9_4_6
void FASTCALL call_14_9_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_4_7
#define init_14_9_4_7
void FASTCALL call_14_9_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_4_8
#define init_14_9_4_8
void FASTCALL call_14_9_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_4_9
#define init_14_9_4_9
void FASTCALL call_14_9_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_4_10
#define init_14_9_4_10
void FASTCALL call_14_9_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_4_11
#define init_14_9_4_11
void FASTCALL call_14_9_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_4_12
#define init_14_9_4_12
void FASTCALL call_14_9_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_4_13
#define init_14_9_4_13
void FASTCALL call_14_9_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_4_14
#define init_14_9_4_14
void FASTCALL call_14_9_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_4_15
#define init_14_9_4_15
void FASTCALL call_14_9_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_5_0
#define init_14_9_5_0
void FASTCALL call_14_9_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_5_1
#define init_14_9_5_1
void FASTCALL call_14_9_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_5_2
#define init_14_9_5_2
void FASTCALL call_14_9_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_5_3
#define init_14_9_5_3
void FASTCALL call_14_9_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_5_4
#define init_14_9_5_4
void FASTCALL call_14_9_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_5_5
#define init_14_9_5_5
void FASTCALL call_14_9_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_5_6
#define init_14_9_5_6
void FASTCALL call_14_9_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_5_7
#define init_14_9_5_7
void FASTCALL call_14_9_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_5_8
#define init_14_9_5_8
void FASTCALL call_14_9_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_5_9
#define init_14_9_5_9
void FASTCALL call_14_9_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_5_10
#define init_14_9_5_10
void FASTCALL call_14_9_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_5_11
#define init_14_9_5_11
void FASTCALL call_14_9_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_5_12
#define init_14_9_5_12
void FASTCALL call_14_9_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_5_13
#define init_14_9_5_13
void FASTCALL call_14_9_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_5_14
#define init_14_9_5_14
void FASTCALL call_14_9_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_5_15
#define init_14_9_5_15
void FASTCALL call_14_9_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_6_0
#define init_14_9_6_0
void FASTCALL call_14_9_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_6_1
#define init_14_9_6_1
void FASTCALL call_14_9_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_6_2
#define init_14_9_6_2
void FASTCALL call_14_9_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_6_3
#define init_14_9_6_3
void FASTCALL call_14_9_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_6_4
#define init_14_9_6_4
void FASTCALL call_14_9_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_6_5
#define init_14_9_6_5
void FASTCALL call_14_9_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_6_6
#define init_14_9_6_6
void FASTCALL call_14_9_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_6_7
#define init_14_9_6_7
void FASTCALL call_14_9_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_6_8
#define init_14_9_6_8
void FASTCALL call_14_9_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_6_9
#define init_14_9_6_9
void FASTCALL call_14_9_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_6_10
#define init_14_9_6_10
void FASTCALL call_14_9_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_6_11
#define init_14_9_6_11
void FASTCALL call_14_9_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_6_12
#define init_14_9_6_12
void FASTCALL call_14_9_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_6_13
#define init_14_9_6_13
void FASTCALL call_14_9_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_6_14
#define init_14_9_6_14
void FASTCALL call_14_9_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_6_15
#define init_14_9_6_15
void FASTCALL call_14_9_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_7_0
#define init_14_9_7_0
void FASTCALL call_14_9_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_7_1
#define init_14_9_7_1
void FASTCALL call_14_9_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_7_2
#define init_14_9_7_2
void FASTCALL call_14_9_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_7_3
#define init_14_9_7_3
void FASTCALL call_14_9_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_7_4
#define init_14_9_7_4
void FASTCALL call_14_9_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_7_5
#define init_14_9_7_5
void FASTCALL call_14_9_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_7_6
#define init_14_9_7_6
void FASTCALL call_14_9_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_7_7
#define init_14_9_7_7
void FASTCALL call_14_9_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_7_8
#define init_14_9_7_8
void FASTCALL call_14_9_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_7_9
#define init_14_9_7_9
void FASTCALL call_14_9_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_7_10
#define init_14_9_7_10
void FASTCALL call_14_9_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_7_11
#define init_14_9_7_11
void FASTCALL call_14_9_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_7_12
#define init_14_9_7_12
void FASTCALL call_14_9_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_7_13
#define init_14_9_7_13
void FASTCALL call_14_9_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_7_14
#define init_14_9_7_14
void FASTCALL call_14_9_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_7_15
#define init_14_9_7_15
void FASTCALL call_14_9_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_8_0
#define init_14_9_8_0
void FASTCALL call_14_9_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_8_1
#define init_14_9_8_1
void FASTCALL call_14_9_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_8_2
#define init_14_9_8_2
void FASTCALL call_14_9_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_8_3
#define init_14_9_8_3
void FASTCALL call_14_9_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_8_4
#define init_14_9_8_4
void FASTCALL call_14_9_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_8_5
#define init_14_9_8_5
void FASTCALL call_14_9_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_8_6
#define init_14_9_8_6
void FASTCALL call_14_9_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_8_7
#define init_14_9_8_7
void FASTCALL call_14_9_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_8_8
#define init_14_9_8_8
void FASTCALL call_14_9_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_8_9
#define init_14_9_8_9
void FASTCALL call_14_9_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_8_10
#define init_14_9_8_10
void FASTCALL call_14_9_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_8_11
#define init_14_9_8_11
void FASTCALL call_14_9_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_8_12
#define init_14_9_8_12
void FASTCALL call_14_9_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_8_13
#define init_14_9_8_13
void FASTCALL call_14_9_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_8_14
#define init_14_9_8_14
void FASTCALL call_14_9_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_8_15
#define init_14_9_8_15
void FASTCALL call_14_9_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_9_0
#define init_14_9_9_0
void FASTCALL call_14_9_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_9_1
#define init_14_9_9_1
void FASTCALL call_14_9_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_9_2
#define init_14_9_9_2
void FASTCALL call_14_9_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_9_3
#define init_14_9_9_3
void FASTCALL call_14_9_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_9_4
#define init_14_9_9_4
void FASTCALL call_14_9_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_9_5
#define init_14_9_9_5
void FASTCALL call_14_9_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_9_6
#define init_14_9_9_6
void FASTCALL call_14_9_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_9_7
#define init_14_9_9_7
void FASTCALL call_14_9_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_9_8
#define init_14_9_9_8
void FASTCALL call_14_9_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_9_9
#define init_14_9_9_9
void FASTCALL call_14_9_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_9_10
#define init_14_9_9_10
void FASTCALL call_14_9_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_9_11
#define init_14_9_9_11
void FASTCALL call_14_9_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_9_12
#define init_14_9_9_12
void FASTCALL call_14_9_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_9_13
#define init_14_9_9_13
void FASTCALL call_14_9_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_9_14
#define init_14_9_9_14
void FASTCALL call_14_9_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_9_15
#define init_14_9_9_15
void FASTCALL call_14_9_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_10_0
#define init_14_9_10_0
void FASTCALL call_14_9_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_10_1
#define init_14_9_10_1
void FASTCALL call_14_9_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_10_2
#define init_14_9_10_2
void FASTCALL call_14_9_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_10_3
#define init_14_9_10_3
void FASTCALL call_14_9_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_10_4
#define init_14_9_10_4
void FASTCALL call_14_9_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_10_5
#define init_14_9_10_5
void FASTCALL call_14_9_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_10_6
#define init_14_9_10_6
void FASTCALL call_14_9_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_10_7
#define init_14_9_10_7
void FASTCALL call_14_9_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_10_8
#define init_14_9_10_8
void FASTCALL call_14_9_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_10_9
#define init_14_9_10_9
void FASTCALL call_14_9_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_10_10
#define init_14_9_10_10
void FASTCALL call_14_9_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_10_11
#define init_14_9_10_11
void FASTCALL call_14_9_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_10_12
#define init_14_9_10_12
void FASTCALL call_14_9_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_10_13
#define init_14_9_10_13
void FASTCALL call_14_9_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_10_14
#define init_14_9_10_14
void FASTCALL call_14_9_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_10_15
#define init_14_9_10_15
void FASTCALL call_14_9_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_11_0
#define init_14_9_11_0
void FASTCALL call_14_9_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_11_1
#define init_14_9_11_1
void FASTCALL call_14_9_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_11_2
#define init_14_9_11_2
void FASTCALL call_14_9_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_11_3
#define init_14_9_11_3
void FASTCALL call_14_9_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_11_4
#define init_14_9_11_4
void FASTCALL call_14_9_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_11_5
#define init_14_9_11_5
void FASTCALL call_14_9_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_11_6
#define init_14_9_11_6
void FASTCALL call_14_9_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_11_7
#define init_14_9_11_7
void FASTCALL call_14_9_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_11_8
#define init_14_9_11_8
void FASTCALL call_14_9_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_11_9
#define init_14_9_11_9
void FASTCALL call_14_9_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_11_10
#define init_14_9_11_10
void FASTCALL call_14_9_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_11_11
#define init_14_9_11_11
void FASTCALL call_14_9_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_11_12
#define init_14_9_11_12
void FASTCALL call_14_9_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_11_13
#define init_14_9_11_13
void FASTCALL call_14_9_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_11_14
#define init_14_9_11_14
void FASTCALL call_14_9_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_11_15
#define init_14_9_11_15
void FASTCALL call_14_9_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_12_0
#define init_14_9_12_0
void FASTCALL call_14_9_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_12_1
#define init_14_9_12_1
void FASTCALL call_14_9_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_12_2
#define init_14_9_12_2
void FASTCALL call_14_9_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_12_3
#define init_14_9_12_3
void FASTCALL call_14_9_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_12_4
#define init_14_9_12_4
void FASTCALL call_14_9_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_12_5
#define init_14_9_12_5
void FASTCALL call_14_9_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_12_6
#define init_14_9_12_6
void FASTCALL call_14_9_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_12_7
#define init_14_9_12_7
void FASTCALL call_14_9_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_12_8
#define init_14_9_12_8
void FASTCALL call_14_9_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_12_9
#define init_14_9_12_9
void FASTCALL call_14_9_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_12_10
#define init_14_9_12_10
void FASTCALL call_14_9_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_12_11
#define init_14_9_12_11
void FASTCALL call_14_9_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_12_12
#define init_14_9_12_12
void FASTCALL call_14_9_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_12_13
#define init_14_9_12_13
void FASTCALL call_14_9_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_12_14
#define init_14_9_12_14
void FASTCALL call_14_9_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_12_15
#define init_14_9_12_15
void FASTCALL call_14_9_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_13_0
#define init_14_9_13_0
void FASTCALL call_14_9_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_13_1
#define init_14_9_13_1
void FASTCALL call_14_9_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_13_2
#define init_14_9_13_2
void FASTCALL call_14_9_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_13_3
#define init_14_9_13_3
void FASTCALL call_14_9_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_13_4
#define init_14_9_13_4
void FASTCALL call_14_9_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_13_5
#define init_14_9_13_5
void FASTCALL call_14_9_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_13_6
#define init_14_9_13_6
void FASTCALL call_14_9_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_13_7
#define init_14_9_13_7
void FASTCALL call_14_9_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_13_8
#define init_14_9_13_8
void FASTCALL call_14_9_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_13_9
#define init_14_9_13_9
void FASTCALL call_14_9_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_13_10
#define init_14_9_13_10
void FASTCALL call_14_9_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_13_11
#define init_14_9_13_11
void FASTCALL call_14_9_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_13_12
#define init_14_9_13_12
void FASTCALL call_14_9_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_13_13
#define init_14_9_13_13
void FASTCALL call_14_9_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_13_14
#define init_14_9_13_14
void FASTCALL call_14_9_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_13_15
#define init_14_9_13_15
void FASTCALL call_14_9_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_14_0
#define init_14_9_14_0
void FASTCALL call_14_9_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_14_1
#define init_14_9_14_1
void FASTCALL call_14_9_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_14_2
#define init_14_9_14_2
void FASTCALL call_14_9_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_14_3
#define init_14_9_14_3
void FASTCALL call_14_9_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_14_4
#define init_14_9_14_4
void FASTCALL call_14_9_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_14_5
#define init_14_9_14_5
void FASTCALL call_14_9_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_14_6
#define init_14_9_14_6
void FASTCALL call_14_9_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_14_7
#define init_14_9_14_7
void FASTCALL call_14_9_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_14_8
#define init_14_9_14_8
void FASTCALL call_14_9_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_14_9
#define init_14_9_14_9
void FASTCALL call_14_9_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_14_10
#define init_14_9_14_10
void FASTCALL call_14_9_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_14_11
#define init_14_9_14_11
void FASTCALL call_14_9_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_14_12
#define init_14_9_14_12
void FASTCALL call_14_9_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_14_13
#define init_14_9_14_13
void FASTCALL call_14_9_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_14_14
#define init_14_9_14_14
void FASTCALL call_14_9_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_14_15
#define init_14_9_14_15
void FASTCALL call_14_9_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_15_0
#define init_14_9_15_0
void FASTCALL call_14_9_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_15_1
#define init_14_9_15_1
void FASTCALL call_14_9_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_15_2
#define init_14_9_15_2
void FASTCALL call_14_9_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_15_3
#define init_14_9_15_3
void FASTCALL call_14_9_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_15_4
#define init_14_9_15_4
void FASTCALL call_14_9_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_15_5
#define init_14_9_15_5
void FASTCALL call_14_9_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_15_6
#define init_14_9_15_6
void FASTCALL call_14_9_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_15_7
#define init_14_9_15_7
void FASTCALL call_14_9_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_15_8
#define init_14_9_15_8
void FASTCALL call_14_9_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_15_9
#define init_14_9_15_9
void FASTCALL call_14_9_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_15_10
#define init_14_9_15_10
void FASTCALL call_14_9_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_15_11
#define init_14_9_15_11
void FASTCALL call_14_9_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_15_12
#define init_14_9_15_12
void FASTCALL call_14_9_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_15_13
#define init_14_9_15_13
void FASTCALL call_14_9_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_15_14
#define init_14_9_15_14
void FASTCALL call_14_9_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_9_15_15
#define init_14_9_15_15
void FASTCALL call_14_9_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_0_0
#define init_14_10_0_0
void FASTCALL call_14_10_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_0_1
#define init_14_10_0_1
void FASTCALL call_14_10_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_0_2
#define init_14_10_0_2
void FASTCALL call_14_10_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_0_3
#define init_14_10_0_3
void FASTCALL call_14_10_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_0_4
#define init_14_10_0_4
void FASTCALL call_14_10_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_0_5
#define init_14_10_0_5
void FASTCALL call_14_10_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_0_6
#define init_14_10_0_6
void FASTCALL call_14_10_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_0_7
#define init_14_10_0_7
void FASTCALL call_14_10_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_0_8
#define init_14_10_0_8
void FASTCALL call_14_10_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_0_9
#define init_14_10_0_9
void FASTCALL call_14_10_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_0_10
#define init_14_10_0_10
void FASTCALL call_14_10_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_0_11
#define init_14_10_0_11
void FASTCALL call_14_10_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_0_12
#define init_14_10_0_12
void FASTCALL call_14_10_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_0_13
#define init_14_10_0_13
void FASTCALL call_14_10_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_0_14
#define init_14_10_0_14
void FASTCALL call_14_10_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_0_15
#define init_14_10_0_15
void FASTCALL call_14_10_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_1_0
#define init_14_10_1_0
void FASTCALL call_14_10_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_1_1
#define init_14_10_1_1
void FASTCALL call_14_10_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_1_2
#define init_14_10_1_2
void FASTCALL call_14_10_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_1_3
#define init_14_10_1_3
void FASTCALL call_14_10_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_1_4
#define init_14_10_1_4
void FASTCALL call_14_10_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_1_5
#define init_14_10_1_5
void FASTCALL call_14_10_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_1_6
#define init_14_10_1_6
void FASTCALL call_14_10_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_1_7
#define init_14_10_1_7
void FASTCALL call_14_10_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_1_8
#define init_14_10_1_8
void FASTCALL call_14_10_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_1_9
#define init_14_10_1_9
void FASTCALL call_14_10_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_1_10
#define init_14_10_1_10
void FASTCALL call_14_10_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_1_11
#define init_14_10_1_11
void FASTCALL call_14_10_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_1_12
#define init_14_10_1_12
void FASTCALL call_14_10_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_1_13
#define init_14_10_1_13
void FASTCALL call_14_10_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_1_14
#define init_14_10_1_14
void FASTCALL call_14_10_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_1_15
#define init_14_10_1_15
void FASTCALL call_14_10_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_2_0
#define init_14_10_2_0
void FASTCALL call_14_10_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_2_1
#define init_14_10_2_1
void FASTCALL call_14_10_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_2_2
#define init_14_10_2_2
void FASTCALL call_14_10_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_2_3
#define init_14_10_2_3
void FASTCALL call_14_10_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_2_4
#define init_14_10_2_4
void FASTCALL call_14_10_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_2_5
#define init_14_10_2_5
void FASTCALL call_14_10_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_2_6
#define init_14_10_2_6
void FASTCALL call_14_10_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_2_7
#define init_14_10_2_7
void FASTCALL call_14_10_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_2_8
#define init_14_10_2_8
void FASTCALL call_14_10_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_2_9
#define init_14_10_2_9
void FASTCALL call_14_10_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_2_10
#define init_14_10_2_10
void FASTCALL call_14_10_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_2_11
#define init_14_10_2_11
void FASTCALL call_14_10_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_2_12
#define init_14_10_2_12
void FASTCALL call_14_10_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_2_13
#define init_14_10_2_13
void FASTCALL call_14_10_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_2_14
#define init_14_10_2_14
void FASTCALL call_14_10_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_2_15
#define init_14_10_2_15
void FASTCALL call_14_10_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_3_0
#define init_14_10_3_0
void FASTCALL call_14_10_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_3_1
#define init_14_10_3_1
void FASTCALL call_14_10_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_3_2
#define init_14_10_3_2
void FASTCALL call_14_10_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_3_3
#define init_14_10_3_3
void FASTCALL call_14_10_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_3_4
#define init_14_10_3_4
void FASTCALL call_14_10_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_3_5
#define init_14_10_3_5
void FASTCALL call_14_10_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_3_6
#define init_14_10_3_6
void FASTCALL call_14_10_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_3_7
#define init_14_10_3_7
void FASTCALL call_14_10_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_3_8
#define init_14_10_3_8
void FASTCALL call_14_10_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_3_9
#define init_14_10_3_9
void FASTCALL call_14_10_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_3_10
#define init_14_10_3_10
void FASTCALL call_14_10_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_3_11
#define init_14_10_3_11
void FASTCALL call_14_10_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_3_12
#define init_14_10_3_12
void FASTCALL call_14_10_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_3_13
#define init_14_10_3_13
void FASTCALL call_14_10_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_3_14
#define init_14_10_3_14
void FASTCALL call_14_10_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_3_15
#define init_14_10_3_15
void FASTCALL call_14_10_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_4_0
#define init_14_10_4_0
void FASTCALL call_14_10_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_4_1
#define init_14_10_4_1
void FASTCALL call_14_10_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_4_2
#define init_14_10_4_2
void FASTCALL call_14_10_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_4_3
#define init_14_10_4_3
void FASTCALL call_14_10_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_4_4
#define init_14_10_4_4
void FASTCALL call_14_10_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_4_5
#define init_14_10_4_5
void FASTCALL call_14_10_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_4_6
#define init_14_10_4_6
void FASTCALL call_14_10_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_4_7
#define init_14_10_4_7
void FASTCALL call_14_10_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_4_8
#define init_14_10_4_8
void FASTCALL call_14_10_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_4_9
#define init_14_10_4_9
void FASTCALL call_14_10_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_4_10
#define init_14_10_4_10
void FASTCALL call_14_10_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_4_11
#define init_14_10_4_11
void FASTCALL call_14_10_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_4_12
#define init_14_10_4_12
void FASTCALL call_14_10_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_4_13
#define init_14_10_4_13
void FASTCALL call_14_10_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_4_14
#define init_14_10_4_14
void FASTCALL call_14_10_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_4_15
#define init_14_10_4_15
void FASTCALL call_14_10_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_5_0
#define init_14_10_5_0
void FASTCALL call_14_10_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_5_1
#define init_14_10_5_1
void FASTCALL call_14_10_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_5_2
#define init_14_10_5_2
void FASTCALL call_14_10_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_5_3
#define init_14_10_5_3
void FASTCALL call_14_10_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_5_4
#define init_14_10_5_4
void FASTCALL call_14_10_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_5_5
#define init_14_10_5_5
void FASTCALL call_14_10_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_5_6
#define init_14_10_5_6
void FASTCALL call_14_10_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_5_7
#define init_14_10_5_7
void FASTCALL call_14_10_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_5_8
#define init_14_10_5_8
void FASTCALL call_14_10_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_5_9
#define init_14_10_5_9
void FASTCALL call_14_10_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_5_10
#define init_14_10_5_10
void FASTCALL call_14_10_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_5_11
#define init_14_10_5_11
void FASTCALL call_14_10_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_5_12
#define init_14_10_5_12
void FASTCALL call_14_10_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_5_13
#define init_14_10_5_13
void FASTCALL call_14_10_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_5_14
#define init_14_10_5_14
void FASTCALL call_14_10_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_5_15
#define init_14_10_5_15
void FASTCALL call_14_10_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_6_0
#define init_14_10_6_0
void FASTCALL call_14_10_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_6_1
#define init_14_10_6_1
void FASTCALL call_14_10_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_6_2
#define init_14_10_6_2
void FASTCALL call_14_10_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_6_3
#define init_14_10_6_3
void FASTCALL call_14_10_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_6_4
#define init_14_10_6_4
void FASTCALL call_14_10_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_6_5
#define init_14_10_6_5
void FASTCALL call_14_10_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_6_6
#define init_14_10_6_6
void FASTCALL call_14_10_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_6_7
#define init_14_10_6_7
void FASTCALL call_14_10_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_6_8
#define init_14_10_6_8
void FASTCALL call_14_10_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_6_9
#define init_14_10_6_9
void FASTCALL call_14_10_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_6_10
#define init_14_10_6_10
void FASTCALL call_14_10_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_6_11
#define init_14_10_6_11
void FASTCALL call_14_10_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_6_12
#define init_14_10_6_12
void FASTCALL call_14_10_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_6_13
#define init_14_10_6_13
void FASTCALL call_14_10_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_6_14
#define init_14_10_6_14
void FASTCALL call_14_10_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_6_15
#define init_14_10_6_15
void FASTCALL call_14_10_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_7_0
#define init_14_10_7_0
void FASTCALL call_14_10_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_7_1
#define init_14_10_7_1
void FASTCALL call_14_10_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_7_2
#define init_14_10_7_2
void FASTCALL call_14_10_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_7_3
#define init_14_10_7_3
void FASTCALL call_14_10_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_7_4
#define init_14_10_7_4
void FASTCALL call_14_10_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_7_5
#define init_14_10_7_5
void FASTCALL call_14_10_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_7_6
#define init_14_10_7_6
void FASTCALL call_14_10_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_7_7
#define init_14_10_7_7
void FASTCALL call_14_10_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_7_8
#define init_14_10_7_8
void FASTCALL call_14_10_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_7_9
#define init_14_10_7_9
void FASTCALL call_14_10_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_7_10
#define init_14_10_7_10
void FASTCALL call_14_10_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_7_11
#define init_14_10_7_11
void FASTCALL call_14_10_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_7_12
#define init_14_10_7_12
void FASTCALL call_14_10_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_7_13
#define init_14_10_7_13
void FASTCALL call_14_10_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_7_14
#define init_14_10_7_14
void FASTCALL call_14_10_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_7_15
#define init_14_10_7_15
void FASTCALL call_14_10_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_8_0
#define init_14_10_8_0
void FASTCALL call_14_10_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_8_1
#define init_14_10_8_1
void FASTCALL call_14_10_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_8_2
#define init_14_10_8_2
void FASTCALL call_14_10_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_8_3
#define init_14_10_8_3
void FASTCALL call_14_10_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_8_4
#define init_14_10_8_4
void FASTCALL call_14_10_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_8_5
#define init_14_10_8_5
void FASTCALL call_14_10_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_8_6
#define init_14_10_8_6
void FASTCALL call_14_10_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_8_7
#define init_14_10_8_7
void FASTCALL call_14_10_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_8_8
#define init_14_10_8_8
void FASTCALL call_14_10_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_8_9
#define init_14_10_8_9
void FASTCALL call_14_10_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_8_10
#define init_14_10_8_10
void FASTCALL call_14_10_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_8_11
#define init_14_10_8_11
void FASTCALL call_14_10_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_8_12
#define init_14_10_8_12
void FASTCALL call_14_10_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_8_13
#define init_14_10_8_13
void FASTCALL call_14_10_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_8_14
#define init_14_10_8_14
void FASTCALL call_14_10_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_8_15
#define init_14_10_8_15
void FASTCALL call_14_10_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_9_0
#define init_14_10_9_0
void FASTCALL call_14_10_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_9_1
#define init_14_10_9_1
void FASTCALL call_14_10_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_9_2
#define init_14_10_9_2
void FASTCALL call_14_10_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_9_3
#define init_14_10_9_3
void FASTCALL call_14_10_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_9_4
#define init_14_10_9_4
void FASTCALL call_14_10_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_9_5
#define init_14_10_9_5
void FASTCALL call_14_10_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_9_6
#define init_14_10_9_6
void FASTCALL call_14_10_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_9_7
#define init_14_10_9_7
void FASTCALL call_14_10_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_9_8
#define init_14_10_9_8
void FASTCALL call_14_10_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_9_9
#define init_14_10_9_9
void FASTCALL call_14_10_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_9_10
#define init_14_10_9_10
void FASTCALL call_14_10_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_9_11
#define init_14_10_9_11
void FASTCALL call_14_10_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_9_12
#define init_14_10_9_12
void FASTCALL call_14_10_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_9_13
#define init_14_10_9_13
void FASTCALL call_14_10_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_9_14
#define init_14_10_9_14
void FASTCALL call_14_10_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_9_15
#define init_14_10_9_15
void FASTCALL call_14_10_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_10_0
#define init_14_10_10_0
void FASTCALL call_14_10_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_10_1
#define init_14_10_10_1
void FASTCALL call_14_10_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_10_2
#define init_14_10_10_2
void FASTCALL call_14_10_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_10_3
#define init_14_10_10_3
void FASTCALL call_14_10_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_10_4
#define init_14_10_10_4
void FASTCALL call_14_10_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_10_5
#define init_14_10_10_5
void FASTCALL call_14_10_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_10_6
#define init_14_10_10_6
void FASTCALL call_14_10_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_10_7
#define init_14_10_10_7
void FASTCALL call_14_10_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_10_8
#define init_14_10_10_8
void FASTCALL call_14_10_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_10_9
#define init_14_10_10_9
void FASTCALL call_14_10_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_10_10
#define init_14_10_10_10
void FASTCALL call_14_10_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_10_11
#define init_14_10_10_11
void FASTCALL call_14_10_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_10_12
#define init_14_10_10_12
void FASTCALL call_14_10_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_10_13
#define init_14_10_10_13
void FASTCALL call_14_10_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_10_14
#define init_14_10_10_14
void FASTCALL call_14_10_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_10_15
#define init_14_10_10_15
void FASTCALL call_14_10_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_11_0
#define init_14_10_11_0
void FASTCALL call_14_10_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_11_1
#define init_14_10_11_1
void FASTCALL call_14_10_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_11_2
#define init_14_10_11_2
void FASTCALL call_14_10_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_11_3
#define init_14_10_11_3
void FASTCALL call_14_10_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_11_4
#define init_14_10_11_4
void FASTCALL call_14_10_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_11_5
#define init_14_10_11_5
void FASTCALL call_14_10_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_11_6
#define init_14_10_11_6
void FASTCALL call_14_10_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_11_7
#define init_14_10_11_7
void FASTCALL call_14_10_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_11_8
#define init_14_10_11_8
void FASTCALL call_14_10_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_11_9
#define init_14_10_11_9
void FASTCALL call_14_10_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_11_10
#define init_14_10_11_10
void FASTCALL call_14_10_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_11_11
#define init_14_10_11_11
void FASTCALL call_14_10_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_11_12
#define init_14_10_11_12
void FASTCALL call_14_10_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_11_13
#define init_14_10_11_13
void FASTCALL call_14_10_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_11_14
#define init_14_10_11_14
void FASTCALL call_14_10_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_11_15
#define init_14_10_11_15
void FASTCALL call_14_10_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_12_0
#define init_14_10_12_0
void FASTCALL call_14_10_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_12_1
#define init_14_10_12_1
void FASTCALL call_14_10_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_12_2
#define init_14_10_12_2
void FASTCALL call_14_10_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_12_3
#define init_14_10_12_3
void FASTCALL call_14_10_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_12_4
#define init_14_10_12_4
void FASTCALL call_14_10_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_12_5
#define init_14_10_12_5
void FASTCALL call_14_10_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_12_6
#define init_14_10_12_6
void FASTCALL call_14_10_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_12_7
#define init_14_10_12_7
void FASTCALL call_14_10_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_12_8
#define init_14_10_12_8
void FASTCALL call_14_10_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_12_9
#define init_14_10_12_9
void FASTCALL call_14_10_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_12_10
#define init_14_10_12_10
void FASTCALL call_14_10_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_12_11
#define init_14_10_12_11
void FASTCALL call_14_10_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_12_12
#define init_14_10_12_12
void FASTCALL call_14_10_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_12_13
#define init_14_10_12_13
void FASTCALL call_14_10_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_12_14
#define init_14_10_12_14
void FASTCALL call_14_10_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_12_15
#define init_14_10_12_15
void FASTCALL call_14_10_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_13_0
#define init_14_10_13_0
void FASTCALL call_14_10_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_13_1
#define init_14_10_13_1
void FASTCALL call_14_10_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_13_2
#define init_14_10_13_2
void FASTCALL call_14_10_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_13_3
#define init_14_10_13_3
void FASTCALL call_14_10_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_13_4
#define init_14_10_13_4
void FASTCALL call_14_10_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_13_5
#define init_14_10_13_5
void FASTCALL call_14_10_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_13_6
#define init_14_10_13_6
void FASTCALL call_14_10_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_13_7
#define init_14_10_13_7
void FASTCALL call_14_10_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_13_8
#define init_14_10_13_8
void FASTCALL call_14_10_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_13_9
#define init_14_10_13_9
void FASTCALL call_14_10_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_13_10
#define init_14_10_13_10
void FASTCALL call_14_10_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_13_11
#define init_14_10_13_11
void FASTCALL call_14_10_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_13_12
#define init_14_10_13_12
void FASTCALL call_14_10_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_13_13
#define init_14_10_13_13
void FASTCALL call_14_10_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_13_14
#define init_14_10_13_14
void FASTCALL call_14_10_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_13_15
#define init_14_10_13_15
void FASTCALL call_14_10_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_14_0
#define init_14_10_14_0
void FASTCALL call_14_10_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_14_1
#define init_14_10_14_1
void FASTCALL call_14_10_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_14_2
#define init_14_10_14_2
void FASTCALL call_14_10_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_14_3
#define init_14_10_14_3
void FASTCALL call_14_10_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_14_4
#define init_14_10_14_4
void FASTCALL call_14_10_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_14_5
#define init_14_10_14_5
void FASTCALL call_14_10_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_14_6
#define init_14_10_14_6
void FASTCALL call_14_10_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_14_7
#define init_14_10_14_7
void FASTCALL call_14_10_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_14_8
#define init_14_10_14_8
void FASTCALL call_14_10_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_14_9
#define init_14_10_14_9
void FASTCALL call_14_10_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_14_10
#define init_14_10_14_10
void FASTCALL call_14_10_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_14_11
#define init_14_10_14_11
void FASTCALL call_14_10_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_14_12
#define init_14_10_14_12
void FASTCALL call_14_10_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_14_13
#define init_14_10_14_13
void FASTCALL call_14_10_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_14_14
#define init_14_10_14_14
void FASTCALL call_14_10_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_14_15
#define init_14_10_14_15
void FASTCALL call_14_10_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_15_0
#define init_14_10_15_0
void FASTCALL call_14_10_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_15_1
#define init_14_10_15_1
void FASTCALL call_14_10_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_15_2
#define init_14_10_15_2
void FASTCALL call_14_10_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_15_3
#define init_14_10_15_3
void FASTCALL call_14_10_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_15_4
#define init_14_10_15_4
void FASTCALL call_14_10_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_15_5
#define init_14_10_15_5
void FASTCALL call_14_10_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_15_6
#define init_14_10_15_6
void FASTCALL call_14_10_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_15_7
#define init_14_10_15_7
void FASTCALL call_14_10_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_15_8
#define init_14_10_15_8
void FASTCALL call_14_10_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_15_9
#define init_14_10_15_9
void FASTCALL call_14_10_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_15_10
#define init_14_10_15_10
void FASTCALL call_14_10_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_15_11
#define init_14_10_15_11
void FASTCALL call_14_10_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_15_12
#define init_14_10_15_12
void FASTCALL call_14_10_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_15_13
#define init_14_10_15_13
void FASTCALL call_14_10_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_15_14
#define init_14_10_15_14
void FASTCALL call_14_10_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_10_15_15
#define init_14_10_15_15
void FASTCALL call_14_10_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_0_0
#define init_14_11_0_0
void FASTCALL call_14_11_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_0_1
#define init_14_11_0_1
void FASTCALL call_14_11_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_0_2
#define init_14_11_0_2
void FASTCALL call_14_11_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_0_3
#define init_14_11_0_3
void FASTCALL call_14_11_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_0_4
#define init_14_11_0_4
void FASTCALL call_14_11_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_0_5
#define init_14_11_0_5
void FASTCALL call_14_11_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_0_6
#define init_14_11_0_6
void FASTCALL call_14_11_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_0_7
#define init_14_11_0_7
void FASTCALL call_14_11_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_0_8
#define init_14_11_0_8
void FASTCALL call_14_11_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_0_9
#define init_14_11_0_9
void FASTCALL call_14_11_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_0_10
#define init_14_11_0_10
void FASTCALL call_14_11_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_0_11
#define init_14_11_0_11
void FASTCALL call_14_11_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_0_12
#define init_14_11_0_12
void FASTCALL call_14_11_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_0_13
#define init_14_11_0_13
void FASTCALL call_14_11_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_0_14
#define init_14_11_0_14
void FASTCALL call_14_11_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_0_15
#define init_14_11_0_15
void FASTCALL call_14_11_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_1_0
#define init_14_11_1_0
void FASTCALL call_14_11_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_1_1
#define init_14_11_1_1
void FASTCALL call_14_11_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_1_2
#define init_14_11_1_2
void FASTCALL call_14_11_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_1_3
#define init_14_11_1_3
void FASTCALL call_14_11_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_1_4
#define init_14_11_1_4
void FASTCALL call_14_11_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_1_5
#define init_14_11_1_5
void FASTCALL call_14_11_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_1_6
#define init_14_11_1_6
void FASTCALL call_14_11_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_1_7
#define init_14_11_1_7
void FASTCALL call_14_11_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_1_8
#define init_14_11_1_8
void FASTCALL call_14_11_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_1_9
#define init_14_11_1_9
void FASTCALL call_14_11_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_1_10
#define init_14_11_1_10
void FASTCALL call_14_11_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_1_11
#define init_14_11_1_11
void FASTCALL call_14_11_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_1_12
#define init_14_11_1_12
void FASTCALL call_14_11_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_1_13
#define init_14_11_1_13
void FASTCALL call_14_11_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_1_14
#define init_14_11_1_14
void FASTCALL call_14_11_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_1_15
#define init_14_11_1_15
void FASTCALL call_14_11_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_2_0
#define init_14_11_2_0
void FASTCALL call_14_11_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_2_1
#define init_14_11_2_1
void FASTCALL call_14_11_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_2_2
#define init_14_11_2_2
void FASTCALL call_14_11_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_2_3
#define init_14_11_2_3
void FASTCALL call_14_11_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_2_4
#define init_14_11_2_4
void FASTCALL call_14_11_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_2_5
#define init_14_11_2_5
void FASTCALL call_14_11_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_2_6
#define init_14_11_2_6
void FASTCALL call_14_11_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_2_7
#define init_14_11_2_7
void FASTCALL call_14_11_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_2_8
#define init_14_11_2_8
void FASTCALL call_14_11_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_2_9
#define init_14_11_2_9
void FASTCALL call_14_11_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_2_10
#define init_14_11_2_10
void FASTCALL call_14_11_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_2_11
#define init_14_11_2_11
void FASTCALL call_14_11_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_2_12
#define init_14_11_2_12
void FASTCALL call_14_11_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_2_13
#define init_14_11_2_13
void FASTCALL call_14_11_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_2_14
#define init_14_11_2_14
void FASTCALL call_14_11_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_2_15
#define init_14_11_2_15
void FASTCALL call_14_11_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_3_0
#define init_14_11_3_0
void FASTCALL call_14_11_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_3_1
#define init_14_11_3_1
void FASTCALL call_14_11_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_3_2
#define init_14_11_3_2
void FASTCALL call_14_11_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_3_3
#define init_14_11_3_3
void FASTCALL call_14_11_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_3_4
#define init_14_11_3_4
void FASTCALL call_14_11_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_3_5
#define init_14_11_3_5
void FASTCALL call_14_11_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_3_6
#define init_14_11_3_6
void FASTCALL call_14_11_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_3_7
#define init_14_11_3_7
void FASTCALL call_14_11_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_3_8
#define init_14_11_3_8
void FASTCALL call_14_11_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_3_9
#define init_14_11_3_9
void FASTCALL call_14_11_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_3_10
#define init_14_11_3_10
void FASTCALL call_14_11_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_3_11
#define init_14_11_3_11
void FASTCALL call_14_11_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_3_12
#define init_14_11_3_12
void FASTCALL call_14_11_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_3_13
#define init_14_11_3_13
void FASTCALL call_14_11_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_3_14
#define init_14_11_3_14
void FASTCALL call_14_11_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_3_15
#define init_14_11_3_15
void FASTCALL call_14_11_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_4_0
#define init_14_11_4_0
void FASTCALL call_14_11_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_4_1
#define init_14_11_4_1
void FASTCALL call_14_11_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_4_2
#define init_14_11_4_2
void FASTCALL call_14_11_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_4_3
#define init_14_11_4_3
void FASTCALL call_14_11_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_4_4
#define init_14_11_4_4
void FASTCALL call_14_11_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_4_5
#define init_14_11_4_5
void FASTCALL call_14_11_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_4_6
#define init_14_11_4_6
void FASTCALL call_14_11_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_4_7
#define init_14_11_4_7
void FASTCALL call_14_11_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_4_8
#define init_14_11_4_8
void FASTCALL call_14_11_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_4_9
#define init_14_11_4_9
void FASTCALL call_14_11_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_4_10
#define init_14_11_4_10
void FASTCALL call_14_11_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_4_11
#define init_14_11_4_11
void FASTCALL call_14_11_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_4_12
#define init_14_11_4_12
void FASTCALL call_14_11_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_4_13
#define init_14_11_4_13
void FASTCALL call_14_11_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_4_14
#define init_14_11_4_14
void FASTCALL call_14_11_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_4_15
#define init_14_11_4_15
void FASTCALL call_14_11_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_5_0
#define init_14_11_5_0
void FASTCALL call_14_11_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_5_1
#define init_14_11_5_1
void FASTCALL call_14_11_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_5_2
#define init_14_11_5_2
void FASTCALL call_14_11_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_5_3
#define init_14_11_5_3
void FASTCALL call_14_11_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_5_4
#define init_14_11_5_4
void FASTCALL call_14_11_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_5_5
#define init_14_11_5_5
void FASTCALL call_14_11_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_5_6
#define init_14_11_5_6
void FASTCALL call_14_11_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_5_7
#define init_14_11_5_7
void FASTCALL call_14_11_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_5_8
#define init_14_11_5_8
void FASTCALL call_14_11_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_5_9
#define init_14_11_5_9
void FASTCALL call_14_11_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_5_10
#define init_14_11_5_10
void FASTCALL call_14_11_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_5_11
#define init_14_11_5_11
void FASTCALL call_14_11_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_5_12
#define init_14_11_5_12
void FASTCALL call_14_11_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_5_13
#define init_14_11_5_13
void FASTCALL call_14_11_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_5_14
#define init_14_11_5_14
void FASTCALL call_14_11_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_5_15
#define init_14_11_5_15
void FASTCALL call_14_11_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_6_0
#define init_14_11_6_0
void FASTCALL call_14_11_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_6_1
#define init_14_11_6_1
void FASTCALL call_14_11_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_6_2
#define init_14_11_6_2
void FASTCALL call_14_11_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_6_3
#define init_14_11_6_3
void FASTCALL call_14_11_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_6_4
#define init_14_11_6_4
void FASTCALL call_14_11_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_6_5
#define init_14_11_6_5
void FASTCALL call_14_11_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_6_6
#define init_14_11_6_6
void FASTCALL call_14_11_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_6_7
#define init_14_11_6_7
void FASTCALL call_14_11_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_6_8
#define init_14_11_6_8
void FASTCALL call_14_11_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_6_9
#define init_14_11_6_9
void FASTCALL call_14_11_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_6_10
#define init_14_11_6_10
void FASTCALL call_14_11_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_6_11
#define init_14_11_6_11
void FASTCALL call_14_11_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_6_12
#define init_14_11_6_12
void FASTCALL call_14_11_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_6_13
#define init_14_11_6_13
void FASTCALL call_14_11_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_6_14
#define init_14_11_6_14
void FASTCALL call_14_11_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_6_15
#define init_14_11_6_15
void FASTCALL call_14_11_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_7_0
#define init_14_11_7_0
void FASTCALL call_14_11_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_7_1
#define init_14_11_7_1
void FASTCALL call_14_11_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_7_2
#define init_14_11_7_2
void FASTCALL call_14_11_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_7_3
#define init_14_11_7_3
void FASTCALL call_14_11_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_7_4
#define init_14_11_7_4
void FASTCALL call_14_11_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_7_5
#define init_14_11_7_5
void FASTCALL call_14_11_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_7_6
#define init_14_11_7_6
void FASTCALL call_14_11_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_7_7
#define init_14_11_7_7
void FASTCALL call_14_11_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_7_8
#define init_14_11_7_8
void FASTCALL call_14_11_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_7_9
#define init_14_11_7_9
void FASTCALL call_14_11_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_7_10
#define init_14_11_7_10
void FASTCALL call_14_11_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_7_11
#define init_14_11_7_11
void FASTCALL call_14_11_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_7_12
#define init_14_11_7_12
void FASTCALL call_14_11_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_7_13
#define init_14_11_7_13
void FASTCALL call_14_11_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_7_14
#define init_14_11_7_14
void FASTCALL call_14_11_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_7_15
#define init_14_11_7_15
void FASTCALL call_14_11_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_8_0
#define init_14_11_8_0
void FASTCALL call_14_11_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_8_1
#define init_14_11_8_1
void FASTCALL call_14_11_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_8_2
#define init_14_11_8_2
void FASTCALL call_14_11_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_8_3
#define init_14_11_8_3
void FASTCALL call_14_11_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_8_4
#define init_14_11_8_4
void FASTCALL call_14_11_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_8_5
#define init_14_11_8_5
void FASTCALL call_14_11_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_8_6
#define init_14_11_8_6
void FASTCALL call_14_11_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_8_7
#define init_14_11_8_7
void FASTCALL call_14_11_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_8_8
#define init_14_11_8_8
void FASTCALL call_14_11_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_8_9
#define init_14_11_8_9
void FASTCALL call_14_11_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_8_10
#define init_14_11_8_10
void FASTCALL call_14_11_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_8_11
#define init_14_11_8_11
void FASTCALL call_14_11_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_8_12
#define init_14_11_8_12
void FASTCALL call_14_11_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_8_13
#define init_14_11_8_13
void FASTCALL call_14_11_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_8_14
#define init_14_11_8_14
void FASTCALL call_14_11_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_8_15
#define init_14_11_8_15
void FASTCALL call_14_11_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_9_0
#define init_14_11_9_0
void FASTCALL call_14_11_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_9_1
#define init_14_11_9_1
void FASTCALL call_14_11_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_9_2
#define init_14_11_9_2
void FASTCALL call_14_11_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_9_3
#define init_14_11_9_3
void FASTCALL call_14_11_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_9_4
#define init_14_11_9_4
void FASTCALL call_14_11_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_9_5
#define init_14_11_9_5
void FASTCALL call_14_11_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_9_6
#define init_14_11_9_6
void FASTCALL call_14_11_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_9_7
#define init_14_11_9_7
void FASTCALL call_14_11_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_9_8
#define init_14_11_9_8
void FASTCALL call_14_11_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_9_9
#define init_14_11_9_9
void FASTCALL call_14_11_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_9_10
#define init_14_11_9_10
void FASTCALL call_14_11_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_9_11
#define init_14_11_9_11
void FASTCALL call_14_11_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_9_12
#define init_14_11_9_12
void FASTCALL call_14_11_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_9_13
#define init_14_11_9_13
void FASTCALL call_14_11_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_9_14
#define init_14_11_9_14
void FASTCALL call_14_11_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_9_15
#define init_14_11_9_15
void FASTCALL call_14_11_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_10_0
#define init_14_11_10_0
void FASTCALL call_14_11_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_10_1
#define init_14_11_10_1
void FASTCALL call_14_11_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_10_2
#define init_14_11_10_2
void FASTCALL call_14_11_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_10_3
#define init_14_11_10_3
void FASTCALL call_14_11_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_10_4
#define init_14_11_10_4
void FASTCALL call_14_11_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_10_5
#define init_14_11_10_5
void FASTCALL call_14_11_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_10_6
#define init_14_11_10_6
void FASTCALL call_14_11_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_10_7
#define init_14_11_10_7
void FASTCALL call_14_11_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_10_8
#define init_14_11_10_8
void FASTCALL call_14_11_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_10_9
#define init_14_11_10_9
void FASTCALL call_14_11_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_10_10
#define init_14_11_10_10
void FASTCALL call_14_11_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_10_11
#define init_14_11_10_11
void FASTCALL call_14_11_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_10_12
#define init_14_11_10_12
void FASTCALL call_14_11_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_10_13
#define init_14_11_10_13
void FASTCALL call_14_11_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_10_14
#define init_14_11_10_14
void FASTCALL call_14_11_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_10_15
#define init_14_11_10_15
void FASTCALL call_14_11_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_11_0
#define init_14_11_11_0
void FASTCALL call_14_11_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_11_1
#define init_14_11_11_1
void FASTCALL call_14_11_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_11_2
#define init_14_11_11_2
void FASTCALL call_14_11_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_11_3
#define init_14_11_11_3
void FASTCALL call_14_11_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_11_4
#define init_14_11_11_4
void FASTCALL call_14_11_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_11_5
#define init_14_11_11_5
void FASTCALL call_14_11_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_11_6
#define init_14_11_11_6
void FASTCALL call_14_11_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_11_7
#define init_14_11_11_7
void FASTCALL call_14_11_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_11_8
#define init_14_11_11_8
void FASTCALL call_14_11_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_11_9
#define init_14_11_11_9
void FASTCALL call_14_11_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_11_10
#define init_14_11_11_10
void FASTCALL call_14_11_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_11_11
#define init_14_11_11_11
void FASTCALL call_14_11_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_11_12
#define init_14_11_11_12
void FASTCALL call_14_11_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_11_13
#define init_14_11_11_13
void FASTCALL call_14_11_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_11_14
#define init_14_11_11_14
void FASTCALL call_14_11_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_11_15
#define init_14_11_11_15
void FASTCALL call_14_11_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_12_0
#define init_14_11_12_0
void FASTCALL call_14_11_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_12_1
#define init_14_11_12_1
void FASTCALL call_14_11_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_12_2
#define init_14_11_12_2
void FASTCALL call_14_11_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_12_3
#define init_14_11_12_3
void FASTCALL call_14_11_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_12_4
#define init_14_11_12_4
void FASTCALL call_14_11_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_12_5
#define init_14_11_12_5
void FASTCALL call_14_11_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_12_6
#define init_14_11_12_6
void FASTCALL call_14_11_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_12_7
#define init_14_11_12_7
void FASTCALL call_14_11_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_12_8
#define init_14_11_12_8
void FASTCALL call_14_11_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_12_9
#define init_14_11_12_9
void FASTCALL call_14_11_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_12_10
#define init_14_11_12_10
void FASTCALL call_14_11_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_12_11
#define init_14_11_12_11
void FASTCALL call_14_11_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_12_12
#define init_14_11_12_12
void FASTCALL call_14_11_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_12_13
#define init_14_11_12_13
void FASTCALL call_14_11_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_12_14
#define init_14_11_12_14
void FASTCALL call_14_11_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_12_15
#define init_14_11_12_15
void FASTCALL call_14_11_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_13_0
#define init_14_11_13_0
void FASTCALL call_14_11_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_13_1
#define init_14_11_13_1
void FASTCALL call_14_11_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_13_2
#define init_14_11_13_2
void FASTCALL call_14_11_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_13_3
#define init_14_11_13_3
void FASTCALL call_14_11_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_13_4
#define init_14_11_13_4
void FASTCALL call_14_11_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_13_5
#define init_14_11_13_5
void FASTCALL call_14_11_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_13_6
#define init_14_11_13_6
void FASTCALL call_14_11_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_13_7
#define init_14_11_13_7
void FASTCALL call_14_11_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_13_8
#define init_14_11_13_8
void FASTCALL call_14_11_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_13_9
#define init_14_11_13_9
void FASTCALL call_14_11_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_13_10
#define init_14_11_13_10
void FASTCALL call_14_11_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_13_11
#define init_14_11_13_11
void FASTCALL call_14_11_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_13_12
#define init_14_11_13_12
void FASTCALL call_14_11_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_13_13
#define init_14_11_13_13
void FASTCALL call_14_11_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_13_14
#define init_14_11_13_14
void FASTCALL call_14_11_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_13_15
#define init_14_11_13_15
void FASTCALL call_14_11_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_14_0
#define init_14_11_14_0
void FASTCALL call_14_11_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_14_1
#define init_14_11_14_1
void FASTCALL call_14_11_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_14_2
#define init_14_11_14_2
void FASTCALL call_14_11_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_14_3
#define init_14_11_14_3
void FASTCALL call_14_11_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_14_4
#define init_14_11_14_4
void FASTCALL call_14_11_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_14_5
#define init_14_11_14_5
void FASTCALL call_14_11_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_14_6
#define init_14_11_14_6
void FASTCALL call_14_11_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_14_7
#define init_14_11_14_7
void FASTCALL call_14_11_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_14_8
#define init_14_11_14_8
void FASTCALL call_14_11_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_14_9
#define init_14_11_14_9
void FASTCALL call_14_11_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_14_10
#define init_14_11_14_10
void FASTCALL call_14_11_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_14_11
#define init_14_11_14_11
void FASTCALL call_14_11_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_14_12
#define init_14_11_14_12
void FASTCALL call_14_11_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_14_13
#define init_14_11_14_13
void FASTCALL call_14_11_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_14_14
#define init_14_11_14_14
void FASTCALL call_14_11_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_14_15
#define init_14_11_14_15
void FASTCALL call_14_11_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_15_0
#define init_14_11_15_0
void FASTCALL call_14_11_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_15_1
#define init_14_11_15_1
void FASTCALL call_14_11_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_15_2
#define init_14_11_15_2
void FASTCALL call_14_11_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_15_3
#define init_14_11_15_3
void FASTCALL call_14_11_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_15_4
#define init_14_11_15_4
void FASTCALL call_14_11_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_15_5
#define init_14_11_15_5
void FASTCALL call_14_11_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_15_6
#define init_14_11_15_6
void FASTCALL call_14_11_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_15_7
#define init_14_11_15_7
void FASTCALL call_14_11_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_15_8
#define init_14_11_15_8
void FASTCALL call_14_11_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_15_9
#define init_14_11_15_9
void FASTCALL call_14_11_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_15_10
#define init_14_11_15_10
void FASTCALL call_14_11_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_15_11
#define init_14_11_15_11
void FASTCALL call_14_11_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_15_12
#define init_14_11_15_12
void FASTCALL call_14_11_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_15_13
#define init_14_11_15_13
void FASTCALL call_14_11_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_15_14
#define init_14_11_15_14
void FASTCALL call_14_11_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_11_15_15
#define init_14_11_15_15
void FASTCALL call_14_11_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_0_0
#define init_14_12_0_0
void FASTCALL call_14_12_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_0_1
#define init_14_12_0_1
void FASTCALL call_14_12_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_0_2
#define init_14_12_0_2
void FASTCALL call_14_12_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_0_3
#define init_14_12_0_3
void FASTCALL call_14_12_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_0_4
#define init_14_12_0_4
void FASTCALL call_14_12_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_0_5
#define init_14_12_0_5
void FASTCALL call_14_12_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_0_6
#define init_14_12_0_6
void FASTCALL call_14_12_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_0_7
#define init_14_12_0_7
void FASTCALL call_14_12_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_0_8
#define init_14_12_0_8
void FASTCALL call_14_12_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_0_9
#define init_14_12_0_9
void FASTCALL call_14_12_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_0_10
#define init_14_12_0_10
void FASTCALL call_14_12_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_0_11
#define init_14_12_0_11
void FASTCALL call_14_12_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_0_12
#define init_14_12_0_12
void FASTCALL call_14_12_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_0_13
#define init_14_12_0_13
void FASTCALL call_14_12_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_0_14
#define init_14_12_0_14
void FASTCALL call_14_12_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_0_15
#define init_14_12_0_15
void FASTCALL call_14_12_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_1_0
#define init_14_12_1_0
void FASTCALL call_14_12_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_1_1
#define init_14_12_1_1
void FASTCALL call_14_12_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_1_2
#define init_14_12_1_2
void FASTCALL call_14_12_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_1_3
#define init_14_12_1_3
void FASTCALL call_14_12_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_1_4
#define init_14_12_1_4
void FASTCALL call_14_12_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_1_5
#define init_14_12_1_5
void FASTCALL call_14_12_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_1_6
#define init_14_12_1_6
void FASTCALL call_14_12_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_1_7
#define init_14_12_1_7
void FASTCALL call_14_12_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_1_8
#define init_14_12_1_8
void FASTCALL call_14_12_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_1_9
#define init_14_12_1_9
void FASTCALL call_14_12_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_1_10
#define init_14_12_1_10
void FASTCALL call_14_12_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_1_11
#define init_14_12_1_11
void FASTCALL call_14_12_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_1_12
#define init_14_12_1_12
void FASTCALL call_14_12_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_1_13
#define init_14_12_1_13
void FASTCALL call_14_12_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_1_14
#define init_14_12_1_14
void FASTCALL call_14_12_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_1_15
#define init_14_12_1_15
void FASTCALL call_14_12_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_2_0
#define init_14_12_2_0
void FASTCALL call_14_12_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_2_1
#define init_14_12_2_1
void FASTCALL call_14_12_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_2_2
#define init_14_12_2_2
void FASTCALL call_14_12_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_2_3
#define init_14_12_2_3
void FASTCALL call_14_12_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_2_4
#define init_14_12_2_4
void FASTCALL call_14_12_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_2_5
#define init_14_12_2_5
void FASTCALL call_14_12_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_2_6
#define init_14_12_2_6
void FASTCALL call_14_12_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_2_7
#define init_14_12_2_7
void FASTCALL call_14_12_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_2_8
#define init_14_12_2_8
void FASTCALL call_14_12_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_2_9
#define init_14_12_2_9
void FASTCALL call_14_12_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_2_10
#define init_14_12_2_10
void FASTCALL call_14_12_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_2_11
#define init_14_12_2_11
void FASTCALL call_14_12_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_2_12
#define init_14_12_2_12
void FASTCALL call_14_12_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_2_13
#define init_14_12_2_13
void FASTCALL call_14_12_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_2_14
#define init_14_12_2_14
void FASTCALL call_14_12_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_2_15
#define init_14_12_2_15
void FASTCALL call_14_12_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_3_0
#define init_14_12_3_0
void FASTCALL call_14_12_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_3_1
#define init_14_12_3_1
void FASTCALL call_14_12_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_3_2
#define init_14_12_3_2
void FASTCALL call_14_12_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_3_3
#define init_14_12_3_3
void FASTCALL call_14_12_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_3_4
#define init_14_12_3_4
void FASTCALL call_14_12_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_3_5
#define init_14_12_3_5
void FASTCALL call_14_12_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_3_6
#define init_14_12_3_6
void FASTCALL call_14_12_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_3_7
#define init_14_12_3_7
void FASTCALL call_14_12_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_3_8
#define init_14_12_3_8
void FASTCALL call_14_12_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_3_9
#define init_14_12_3_9
void FASTCALL call_14_12_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_3_10
#define init_14_12_3_10
void FASTCALL call_14_12_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_3_11
#define init_14_12_3_11
void FASTCALL call_14_12_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_3_12
#define init_14_12_3_12
void FASTCALL call_14_12_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_3_13
#define init_14_12_3_13
void FASTCALL call_14_12_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_3_14
#define init_14_12_3_14
void FASTCALL call_14_12_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_3_15
#define init_14_12_3_15
void FASTCALL call_14_12_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_4_0
#define init_14_12_4_0
void FASTCALL call_14_12_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_4_1
#define init_14_12_4_1
void FASTCALL call_14_12_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_4_2
#define init_14_12_4_2
void FASTCALL call_14_12_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_4_3
#define init_14_12_4_3
void FASTCALL call_14_12_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_4_4
#define init_14_12_4_4
void FASTCALL call_14_12_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_4_5
#define init_14_12_4_5
void FASTCALL call_14_12_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_4_6
#define init_14_12_4_6
void FASTCALL call_14_12_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_4_7
#define init_14_12_4_7
void FASTCALL call_14_12_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_4_8
#define init_14_12_4_8
void FASTCALL call_14_12_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_4_9
#define init_14_12_4_9
void FASTCALL call_14_12_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_4_10
#define init_14_12_4_10
void FASTCALL call_14_12_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_4_11
#define init_14_12_4_11
void FASTCALL call_14_12_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_4_12
#define init_14_12_4_12
void FASTCALL call_14_12_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_4_13
#define init_14_12_4_13
void FASTCALL call_14_12_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_4_14
#define init_14_12_4_14
void FASTCALL call_14_12_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_4_15
#define init_14_12_4_15
void FASTCALL call_14_12_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_5_0
#define init_14_12_5_0
void FASTCALL call_14_12_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_5_1
#define init_14_12_5_1
void FASTCALL call_14_12_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_5_2
#define init_14_12_5_2
void FASTCALL call_14_12_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_5_3
#define init_14_12_5_3
void FASTCALL call_14_12_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_5_4
#define init_14_12_5_4
void FASTCALL call_14_12_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_5_5
#define init_14_12_5_5
void FASTCALL call_14_12_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_5_6
#define init_14_12_5_6
void FASTCALL call_14_12_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_5_7
#define init_14_12_5_7
void FASTCALL call_14_12_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_5_8
#define init_14_12_5_8
void FASTCALL call_14_12_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_5_9
#define init_14_12_5_9
void FASTCALL call_14_12_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_5_10
#define init_14_12_5_10
void FASTCALL call_14_12_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_5_11
#define init_14_12_5_11
void FASTCALL call_14_12_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_5_12
#define init_14_12_5_12
void FASTCALL call_14_12_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_5_13
#define init_14_12_5_13
void FASTCALL call_14_12_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_5_14
#define init_14_12_5_14
void FASTCALL call_14_12_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_5_15
#define init_14_12_5_15
void FASTCALL call_14_12_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_6_0
#define init_14_12_6_0
void FASTCALL call_14_12_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_6_1
#define init_14_12_6_1
void FASTCALL call_14_12_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_6_2
#define init_14_12_6_2
void FASTCALL call_14_12_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_6_3
#define init_14_12_6_3
void FASTCALL call_14_12_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_6_4
#define init_14_12_6_4
void FASTCALL call_14_12_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_6_5
#define init_14_12_6_5
void FASTCALL call_14_12_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_6_6
#define init_14_12_6_6
void FASTCALL call_14_12_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_6_7
#define init_14_12_6_7
void FASTCALL call_14_12_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_6_8
#define init_14_12_6_8
void FASTCALL call_14_12_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_6_9
#define init_14_12_6_9
void FASTCALL call_14_12_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_6_10
#define init_14_12_6_10
void FASTCALL call_14_12_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_6_11
#define init_14_12_6_11
void FASTCALL call_14_12_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_6_12
#define init_14_12_6_12
void FASTCALL call_14_12_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_6_13
#define init_14_12_6_13
void FASTCALL call_14_12_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_6_14
#define init_14_12_6_14
void FASTCALL call_14_12_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_6_15
#define init_14_12_6_15
void FASTCALL call_14_12_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_7_0
#define init_14_12_7_0
void FASTCALL call_14_12_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_7_1
#define init_14_12_7_1
void FASTCALL call_14_12_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_7_2
#define init_14_12_7_2
void FASTCALL call_14_12_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_7_3
#define init_14_12_7_3
void FASTCALL call_14_12_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_7_4
#define init_14_12_7_4
void FASTCALL call_14_12_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_7_5
#define init_14_12_7_5
void FASTCALL call_14_12_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_7_6
#define init_14_12_7_6
void FASTCALL call_14_12_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_7_7
#define init_14_12_7_7
void FASTCALL call_14_12_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_7_8
#define init_14_12_7_8
void FASTCALL call_14_12_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_7_9
#define init_14_12_7_9
void FASTCALL call_14_12_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_7_10
#define init_14_12_7_10
void FASTCALL call_14_12_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_7_11
#define init_14_12_7_11
void FASTCALL call_14_12_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_7_12
#define init_14_12_7_12
void FASTCALL call_14_12_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_7_13
#define init_14_12_7_13
void FASTCALL call_14_12_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_7_14
#define init_14_12_7_14
void FASTCALL call_14_12_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_7_15
#define init_14_12_7_15
void FASTCALL call_14_12_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_8_0
#define init_14_12_8_0
void FASTCALL call_14_12_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_8_1
#define init_14_12_8_1
void FASTCALL call_14_12_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_8_2
#define init_14_12_8_2
void FASTCALL call_14_12_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_8_3
#define init_14_12_8_3
void FASTCALL call_14_12_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_8_4
#define init_14_12_8_4
void FASTCALL call_14_12_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_8_5
#define init_14_12_8_5
void FASTCALL call_14_12_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_8_6
#define init_14_12_8_6
void FASTCALL call_14_12_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_8_7
#define init_14_12_8_7
void FASTCALL call_14_12_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_8_8
#define init_14_12_8_8
void FASTCALL call_14_12_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_8_9
#define init_14_12_8_9
void FASTCALL call_14_12_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_8_10
#define init_14_12_8_10
void FASTCALL call_14_12_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_8_11
#define init_14_12_8_11
void FASTCALL call_14_12_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_8_12
#define init_14_12_8_12
void FASTCALL call_14_12_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_8_13
#define init_14_12_8_13
void FASTCALL call_14_12_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_8_14
#define init_14_12_8_14
void FASTCALL call_14_12_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_8_15
#define init_14_12_8_15
void FASTCALL call_14_12_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_9_0
#define init_14_12_9_0
void FASTCALL call_14_12_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_9_1
#define init_14_12_9_1
void FASTCALL call_14_12_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_9_2
#define init_14_12_9_2
void FASTCALL call_14_12_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_9_3
#define init_14_12_9_3
void FASTCALL call_14_12_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_9_4
#define init_14_12_9_4
void FASTCALL call_14_12_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_9_5
#define init_14_12_9_5
void FASTCALL call_14_12_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_9_6
#define init_14_12_9_6
void FASTCALL call_14_12_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_9_7
#define init_14_12_9_7
void FASTCALL call_14_12_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_9_8
#define init_14_12_9_8
void FASTCALL call_14_12_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_9_9
#define init_14_12_9_9
void FASTCALL call_14_12_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_9_10
#define init_14_12_9_10
void FASTCALL call_14_12_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_9_11
#define init_14_12_9_11
void FASTCALL call_14_12_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_9_12
#define init_14_12_9_12
void FASTCALL call_14_12_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_9_13
#define init_14_12_9_13
void FASTCALL call_14_12_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_9_14
#define init_14_12_9_14
void FASTCALL call_14_12_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_9_15
#define init_14_12_9_15
void FASTCALL call_14_12_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_10_0
#define init_14_12_10_0
void FASTCALL call_14_12_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_10_1
#define init_14_12_10_1
void FASTCALL call_14_12_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_10_2
#define init_14_12_10_2
void FASTCALL call_14_12_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_10_3
#define init_14_12_10_3
void FASTCALL call_14_12_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_10_4
#define init_14_12_10_4
void FASTCALL call_14_12_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_10_5
#define init_14_12_10_5
void FASTCALL call_14_12_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_10_6
#define init_14_12_10_6
void FASTCALL call_14_12_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_10_7
#define init_14_12_10_7
void FASTCALL call_14_12_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_10_8
#define init_14_12_10_8
void FASTCALL call_14_12_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_10_9
#define init_14_12_10_9
void FASTCALL call_14_12_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_10_10
#define init_14_12_10_10
void FASTCALL call_14_12_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_10_11
#define init_14_12_10_11
void FASTCALL call_14_12_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_10_12
#define init_14_12_10_12
void FASTCALL call_14_12_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_10_13
#define init_14_12_10_13
void FASTCALL call_14_12_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_10_14
#define init_14_12_10_14
void FASTCALL call_14_12_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_10_15
#define init_14_12_10_15
void FASTCALL call_14_12_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_11_0
#define init_14_12_11_0
void FASTCALL call_14_12_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_11_1
#define init_14_12_11_1
void FASTCALL call_14_12_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_11_2
#define init_14_12_11_2
void FASTCALL call_14_12_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_11_3
#define init_14_12_11_3
void FASTCALL call_14_12_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_11_4
#define init_14_12_11_4
void FASTCALL call_14_12_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_11_5
#define init_14_12_11_5
void FASTCALL call_14_12_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_11_6
#define init_14_12_11_6
void FASTCALL call_14_12_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_11_7
#define init_14_12_11_7
void FASTCALL call_14_12_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_11_8
#define init_14_12_11_8
void FASTCALL call_14_12_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_11_9
#define init_14_12_11_9
void FASTCALL call_14_12_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_11_10
#define init_14_12_11_10
void FASTCALL call_14_12_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_11_11
#define init_14_12_11_11
void FASTCALL call_14_12_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_11_12
#define init_14_12_11_12
void FASTCALL call_14_12_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_11_13
#define init_14_12_11_13
void FASTCALL call_14_12_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_11_14
#define init_14_12_11_14
void FASTCALL call_14_12_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_11_15
#define init_14_12_11_15
void FASTCALL call_14_12_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_12_0
#define init_14_12_12_0
void FASTCALL call_14_12_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_12_1
#define init_14_12_12_1
void FASTCALL call_14_12_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_12_2
#define init_14_12_12_2
void FASTCALL call_14_12_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_12_3
#define init_14_12_12_3
void FASTCALL call_14_12_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_12_4
#define init_14_12_12_4
void FASTCALL call_14_12_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_12_5
#define init_14_12_12_5
void FASTCALL call_14_12_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_12_6
#define init_14_12_12_6
void FASTCALL call_14_12_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_12_7
#define init_14_12_12_7
void FASTCALL call_14_12_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_12_8
#define init_14_12_12_8
void FASTCALL call_14_12_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_12_9
#define init_14_12_12_9
void FASTCALL call_14_12_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_12_10
#define init_14_12_12_10
void FASTCALL call_14_12_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_12_11
#define init_14_12_12_11
void FASTCALL call_14_12_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_12_12
#define init_14_12_12_12
void FASTCALL call_14_12_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_12_13
#define init_14_12_12_13
void FASTCALL call_14_12_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_12_14
#define init_14_12_12_14
void FASTCALL call_14_12_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_12_15
#define init_14_12_12_15
void FASTCALL call_14_12_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_13_0
#define init_14_12_13_0
void FASTCALL call_14_12_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_13_1
#define init_14_12_13_1
void FASTCALL call_14_12_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_13_2
#define init_14_12_13_2
void FASTCALL call_14_12_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_13_3
#define init_14_12_13_3
void FASTCALL call_14_12_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_13_4
#define init_14_12_13_4
void FASTCALL call_14_12_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_13_5
#define init_14_12_13_5
void FASTCALL call_14_12_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_13_6
#define init_14_12_13_6
void FASTCALL call_14_12_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_13_7
#define init_14_12_13_7
void FASTCALL call_14_12_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_13_8
#define init_14_12_13_8
void FASTCALL call_14_12_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_13_9
#define init_14_12_13_9
void FASTCALL call_14_12_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_13_10
#define init_14_12_13_10
void FASTCALL call_14_12_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_13_11
#define init_14_12_13_11
void FASTCALL call_14_12_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_13_12
#define init_14_12_13_12
void FASTCALL call_14_12_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_13_13
#define init_14_12_13_13
void FASTCALL call_14_12_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_13_14
#define init_14_12_13_14
void FASTCALL call_14_12_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_13_15
#define init_14_12_13_15
void FASTCALL call_14_12_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_14_0
#define init_14_12_14_0
void FASTCALL call_14_12_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_14_1
#define init_14_12_14_1
void FASTCALL call_14_12_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_14_2
#define init_14_12_14_2
void FASTCALL call_14_12_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_14_3
#define init_14_12_14_3
void FASTCALL call_14_12_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_14_4
#define init_14_12_14_4
void FASTCALL call_14_12_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_14_5
#define init_14_12_14_5
void FASTCALL call_14_12_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_14_6
#define init_14_12_14_6
void FASTCALL call_14_12_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_14_7
#define init_14_12_14_7
void FASTCALL call_14_12_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_14_8
#define init_14_12_14_8
void FASTCALL call_14_12_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_14_9
#define init_14_12_14_9
void FASTCALL call_14_12_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_14_10
#define init_14_12_14_10
void FASTCALL call_14_12_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_14_11
#define init_14_12_14_11
void FASTCALL call_14_12_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_14_12
#define init_14_12_14_12
void FASTCALL call_14_12_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_14_13
#define init_14_12_14_13
void FASTCALL call_14_12_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_14_14
#define init_14_12_14_14
void FASTCALL call_14_12_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_14_15
#define init_14_12_14_15
void FASTCALL call_14_12_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_15_0
#define init_14_12_15_0
void FASTCALL call_14_12_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_15_1
#define init_14_12_15_1
void FASTCALL call_14_12_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_15_2
#define init_14_12_15_2
void FASTCALL call_14_12_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_15_3
#define init_14_12_15_3
void FASTCALL call_14_12_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_15_4
#define init_14_12_15_4
void FASTCALL call_14_12_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_15_5
#define init_14_12_15_5
void FASTCALL call_14_12_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_15_6
#define init_14_12_15_6
void FASTCALL call_14_12_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_15_7
#define init_14_12_15_7
void FASTCALL call_14_12_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_15_8
#define init_14_12_15_8
void FASTCALL call_14_12_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_15_9
#define init_14_12_15_9
void FASTCALL call_14_12_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_15_10
#define init_14_12_15_10
void FASTCALL call_14_12_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_15_11
#define init_14_12_15_11
void FASTCALL call_14_12_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_15_12
#define init_14_12_15_12
void FASTCALL call_14_12_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_15_13
#define init_14_12_15_13
void FASTCALL call_14_12_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_15_14
#define init_14_12_15_14
void FASTCALL call_14_12_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_12_15_15
#define init_14_12_15_15
void FASTCALL call_14_12_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_0_0
#define init_14_13_0_0
void FASTCALL call_14_13_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_0_1
#define init_14_13_0_1
void FASTCALL call_14_13_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_0_2
#define init_14_13_0_2
void FASTCALL call_14_13_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_0_3
#define init_14_13_0_3
void FASTCALL call_14_13_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_0_4
#define init_14_13_0_4
void FASTCALL call_14_13_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_0_5
#define init_14_13_0_5
void FASTCALL call_14_13_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_0_6
#define init_14_13_0_6
void FASTCALL call_14_13_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_0_7
#define init_14_13_0_7
void FASTCALL call_14_13_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_0_8
#define init_14_13_0_8
void FASTCALL call_14_13_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_0_9
#define init_14_13_0_9
void FASTCALL call_14_13_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_0_10
#define init_14_13_0_10
void FASTCALL call_14_13_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_0_11
#define init_14_13_0_11
void FASTCALL call_14_13_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_0_12
#define init_14_13_0_12
void FASTCALL call_14_13_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_0_13
#define init_14_13_0_13
void FASTCALL call_14_13_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_0_14
#define init_14_13_0_14
void FASTCALL call_14_13_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_0_15
#define init_14_13_0_15
void FASTCALL call_14_13_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_1_0
#define init_14_13_1_0
void FASTCALL call_14_13_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_1_1
#define init_14_13_1_1
void FASTCALL call_14_13_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_1_2
#define init_14_13_1_2
void FASTCALL call_14_13_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_1_3
#define init_14_13_1_3
void FASTCALL call_14_13_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_1_4
#define init_14_13_1_4
void FASTCALL call_14_13_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_1_5
#define init_14_13_1_5
void FASTCALL call_14_13_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_1_6
#define init_14_13_1_6
void FASTCALL call_14_13_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_1_7
#define init_14_13_1_7
void FASTCALL call_14_13_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_1_8
#define init_14_13_1_8
void FASTCALL call_14_13_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_1_9
#define init_14_13_1_9
void FASTCALL call_14_13_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_1_10
#define init_14_13_1_10
void FASTCALL call_14_13_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_1_11
#define init_14_13_1_11
void FASTCALL call_14_13_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_1_12
#define init_14_13_1_12
void FASTCALL call_14_13_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_1_13
#define init_14_13_1_13
void FASTCALL call_14_13_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_1_14
#define init_14_13_1_14
void FASTCALL call_14_13_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_1_15
#define init_14_13_1_15
void FASTCALL call_14_13_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_2_0
#define init_14_13_2_0
void FASTCALL call_14_13_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_2_1
#define init_14_13_2_1
void FASTCALL call_14_13_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_2_2
#define init_14_13_2_2
void FASTCALL call_14_13_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_2_3
#define init_14_13_2_3
void FASTCALL call_14_13_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_2_4
#define init_14_13_2_4
void FASTCALL call_14_13_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_2_5
#define init_14_13_2_5
void FASTCALL call_14_13_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_2_6
#define init_14_13_2_6
void FASTCALL call_14_13_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_2_7
#define init_14_13_2_7
void FASTCALL call_14_13_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_2_8
#define init_14_13_2_8
void FASTCALL call_14_13_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_2_9
#define init_14_13_2_9
void FASTCALL call_14_13_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_2_10
#define init_14_13_2_10
void FASTCALL call_14_13_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_2_11
#define init_14_13_2_11
void FASTCALL call_14_13_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_2_12
#define init_14_13_2_12
void FASTCALL call_14_13_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_2_13
#define init_14_13_2_13
void FASTCALL call_14_13_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_2_14
#define init_14_13_2_14
void FASTCALL call_14_13_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_2_15
#define init_14_13_2_15
void FASTCALL call_14_13_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_3_0
#define init_14_13_3_0
void FASTCALL call_14_13_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_3_1
#define init_14_13_3_1
void FASTCALL call_14_13_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_3_2
#define init_14_13_3_2
void FASTCALL call_14_13_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_3_3
#define init_14_13_3_3
void FASTCALL call_14_13_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_3_4
#define init_14_13_3_4
void FASTCALL call_14_13_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_3_5
#define init_14_13_3_5
void FASTCALL call_14_13_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_3_6
#define init_14_13_3_6
void FASTCALL call_14_13_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_3_7
#define init_14_13_3_7
void FASTCALL call_14_13_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_3_8
#define init_14_13_3_8
void FASTCALL call_14_13_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_3_9
#define init_14_13_3_9
void FASTCALL call_14_13_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_3_10
#define init_14_13_3_10
void FASTCALL call_14_13_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_3_11
#define init_14_13_3_11
void FASTCALL call_14_13_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_3_12
#define init_14_13_3_12
void FASTCALL call_14_13_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_3_13
#define init_14_13_3_13
void FASTCALL call_14_13_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_3_14
#define init_14_13_3_14
void FASTCALL call_14_13_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_3_15
#define init_14_13_3_15
void FASTCALL call_14_13_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_4_0
#define init_14_13_4_0
void FASTCALL call_14_13_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_4_1
#define init_14_13_4_1
void FASTCALL call_14_13_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_4_2
#define init_14_13_4_2
void FASTCALL call_14_13_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_4_3
#define init_14_13_4_3
void FASTCALL call_14_13_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_4_4
#define init_14_13_4_4
void FASTCALL call_14_13_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_4_5
#define init_14_13_4_5
void FASTCALL call_14_13_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_4_6
#define init_14_13_4_6
void FASTCALL call_14_13_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_4_7
#define init_14_13_4_7
void FASTCALL call_14_13_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_4_8
#define init_14_13_4_8
void FASTCALL call_14_13_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_4_9
#define init_14_13_4_9
void FASTCALL call_14_13_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_4_10
#define init_14_13_4_10
void FASTCALL call_14_13_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_4_11
#define init_14_13_4_11
void FASTCALL call_14_13_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_4_12
#define init_14_13_4_12
void FASTCALL call_14_13_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_4_13
#define init_14_13_4_13
void FASTCALL call_14_13_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_4_14
#define init_14_13_4_14
void FASTCALL call_14_13_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_4_15
#define init_14_13_4_15
void FASTCALL call_14_13_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_5_0
#define init_14_13_5_0
void FASTCALL call_14_13_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_5_1
#define init_14_13_5_1
void FASTCALL call_14_13_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_5_2
#define init_14_13_5_2
void FASTCALL call_14_13_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_5_3
#define init_14_13_5_3
void FASTCALL call_14_13_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_5_4
#define init_14_13_5_4
void FASTCALL call_14_13_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_5_5
#define init_14_13_5_5
void FASTCALL call_14_13_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_5_6
#define init_14_13_5_6
void FASTCALL call_14_13_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_5_7
#define init_14_13_5_7
void FASTCALL call_14_13_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_5_8
#define init_14_13_5_8
void FASTCALL call_14_13_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_5_9
#define init_14_13_5_9
void FASTCALL call_14_13_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_5_10
#define init_14_13_5_10
void FASTCALL call_14_13_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_5_11
#define init_14_13_5_11
void FASTCALL call_14_13_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_5_12
#define init_14_13_5_12
void FASTCALL call_14_13_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_5_13
#define init_14_13_5_13
void FASTCALL call_14_13_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_5_14
#define init_14_13_5_14
void FASTCALL call_14_13_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_5_15
#define init_14_13_5_15
void FASTCALL call_14_13_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_6_0
#define init_14_13_6_0
void FASTCALL call_14_13_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_6_1
#define init_14_13_6_1
void FASTCALL call_14_13_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_6_2
#define init_14_13_6_2
void FASTCALL call_14_13_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_6_3
#define init_14_13_6_3
void FASTCALL call_14_13_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_6_4
#define init_14_13_6_4
void FASTCALL call_14_13_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_6_5
#define init_14_13_6_5
void FASTCALL call_14_13_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_6_6
#define init_14_13_6_6
void FASTCALL call_14_13_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_6_7
#define init_14_13_6_7
void FASTCALL call_14_13_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_6_8
#define init_14_13_6_8
void FASTCALL call_14_13_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_6_9
#define init_14_13_6_9
void FASTCALL call_14_13_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_6_10
#define init_14_13_6_10
void FASTCALL call_14_13_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_6_11
#define init_14_13_6_11
void FASTCALL call_14_13_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_6_12
#define init_14_13_6_12
void FASTCALL call_14_13_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_6_13
#define init_14_13_6_13
void FASTCALL call_14_13_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_6_14
#define init_14_13_6_14
void FASTCALL call_14_13_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_6_15
#define init_14_13_6_15
void FASTCALL call_14_13_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_7_0
#define init_14_13_7_0
void FASTCALL call_14_13_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_7_1
#define init_14_13_7_1
void FASTCALL call_14_13_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_7_2
#define init_14_13_7_2
void FASTCALL call_14_13_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_7_3
#define init_14_13_7_3
void FASTCALL call_14_13_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_7_4
#define init_14_13_7_4
void FASTCALL call_14_13_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_7_5
#define init_14_13_7_5
void FASTCALL call_14_13_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_7_6
#define init_14_13_7_6
void FASTCALL call_14_13_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_7_7
#define init_14_13_7_7
void FASTCALL call_14_13_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_7_8
#define init_14_13_7_8
void FASTCALL call_14_13_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_7_9
#define init_14_13_7_9
void FASTCALL call_14_13_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_7_10
#define init_14_13_7_10
void FASTCALL call_14_13_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_7_11
#define init_14_13_7_11
void FASTCALL call_14_13_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_7_12
#define init_14_13_7_12
void FASTCALL call_14_13_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_7_13
#define init_14_13_7_13
void FASTCALL call_14_13_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_7_14
#define init_14_13_7_14
void FASTCALL call_14_13_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_7_15
#define init_14_13_7_15
void FASTCALL call_14_13_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_8_0
#define init_14_13_8_0
void FASTCALL call_14_13_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_8_1
#define init_14_13_8_1
void FASTCALL call_14_13_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_8_2
#define init_14_13_8_2
void FASTCALL call_14_13_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_8_3
#define init_14_13_8_3
void FASTCALL call_14_13_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_8_4
#define init_14_13_8_4
void FASTCALL call_14_13_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_8_5
#define init_14_13_8_5
void FASTCALL call_14_13_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_8_6
#define init_14_13_8_6
void FASTCALL call_14_13_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_8_7
#define init_14_13_8_7
void FASTCALL call_14_13_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_8_8
#define init_14_13_8_8
void FASTCALL call_14_13_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_8_9
#define init_14_13_8_9
void FASTCALL call_14_13_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_8_10
#define init_14_13_8_10
void FASTCALL call_14_13_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_8_11
#define init_14_13_8_11
void FASTCALL call_14_13_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_8_12
#define init_14_13_8_12
void FASTCALL call_14_13_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_8_13
#define init_14_13_8_13
void FASTCALL call_14_13_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_8_14
#define init_14_13_8_14
void FASTCALL call_14_13_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_8_15
#define init_14_13_8_15
void FASTCALL call_14_13_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_9_0
#define init_14_13_9_0
void FASTCALL call_14_13_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_9_1
#define init_14_13_9_1
void FASTCALL call_14_13_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_9_2
#define init_14_13_9_2
void FASTCALL call_14_13_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_9_3
#define init_14_13_9_3
void FASTCALL call_14_13_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_9_4
#define init_14_13_9_4
void FASTCALL call_14_13_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_9_5
#define init_14_13_9_5
void FASTCALL call_14_13_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_9_6
#define init_14_13_9_6
void FASTCALL call_14_13_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_9_7
#define init_14_13_9_7
void FASTCALL call_14_13_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_9_8
#define init_14_13_9_8
void FASTCALL call_14_13_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_9_9
#define init_14_13_9_9
void FASTCALL call_14_13_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_9_10
#define init_14_13_9_10
void FASTCALL call_14_13_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_9_11
#define init_14_13_9_11
void FASTCALL call_14_13_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_9_12
#define init_14_13_9_12
void FASTCALL call_14_13_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_9_13
#define init_14_13_9_13
void FASTCALL call_14_13_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_9_14
#define init_14_13_9_14
void FASTCALL call_14_13_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_9_15
#define init_14_13_9_15
void FASTCALL call_14_13_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_10_0
#define init_14_13_10_0
void FASTCALL call_14_13_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_10_1
#define init_14_13_10_1
void FASTCALL call_14_13_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_10_2
#define init_14_13_10_2
void FASTCALL call_14_13_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_10_3
#define init_14_13_10_3
void FASTCALL call_14_13_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_10_4
#define init_14_13_10_4
void FASTCALL call_14_13_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_10_5
#define init_14_13_10_5
void FASTCALL call_14_13_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_10_6
#define init_14_13_10_6
void FASTCALL call_14_13_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_10_7
#define init_14_13_10_7
void FASTCALL call_14_13_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_10_8
#define init_14_13_10_8
void FASTCALL call_14_13_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_10_9
#define init_14_13_10_9
void FASTCALL call_14_13_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_10_10
#define init_14_13_10_10
void FASTCALL call_14_13_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_10_11
#define init_14_13_10_11
void FASTCALL call_14_13_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_10_12
#define init_14_13_10_12
void FASTCALL call_14_13_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_10_13
#define init_14_13_10_13
void FASTCALL call_14_13_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_10_14
#define init_14_13_10_14
void FASTCALL call_14_13_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_10_15
#define init_14_13_10_15
void FASTCALL call_14_13_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_11_0
#define init_14_13_11_0
void FASTCALL call_14_13_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_11_1
#define init_14_13_11_1
void FASTCALL call_14_13_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_11_2
#define init_14_13_11_2
void FASTCALL call_14_13_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_11_3
#define init_14_13_11_3
void FASTCALL call_14_13_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_11_4
#define init_14_13_11_4
void FASTCALL call_14_13_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_11_5
#define init_14_13_11_5
void FASTCALL call_14_13_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_11_6
#define init_14_13_11_6
void FASTCALL call_14_13_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_11_7
#define init_14_13_11_7
void FASTCALL call_14_13_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_11_8
#define init_14_13_11_8
void FASTCALL call_14_13_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_11_9
#define init_14_13_11_9
void FASTCALL call_14_13_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_11_10
#define init_14_13_11_10
void FASTCALL call_14_13_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_11_11
#define init_14_13_11_11
void FASTCALL call_14_13_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_11_12
#define init_14_13_11_12
void FASTCALL call_14_13_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_11_13
#define init_14_13_11_13
void FASTCALL call_14_13_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_11_14
#define init_14_13_11_14
void FASTCALL call_14_13_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_11_15
#define init_14_13_11_15
void FASTCALL call_14_13_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_12_0
#define init_14_13_12_0
void FASTCALL call_14_13_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_12_1
#define init_14_13_12_1
void FASTCALL call_14_13_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_12_2
#define init_14_13_12_2
void FASTCALL call_14_13_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_12_3
#define init_14_13_12_3
void FASTCALL call_14_13_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_12_4
#define init_14_13_12_4
void FASTCALL call_14_13_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_12_5
#define init_14_13_12_5
void FASTCALL call_14_13_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_12_6
#define init_14_13_12_6
void FASTCALL call_14_13_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_12_7
#define init_14_13_12_7
void FASTCALL call_14_13_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_12_8
#define init_14_13_12_8
void FASTCALL call_14_13_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_12_9
#define init_14_13_12_9
void FASTCALL call_14_13_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_12_10
#define init_14_13_12_10
void FASTCALL call_14_13_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_12_11
#define init_14_13_12_11
void FASTCALL call_14_13_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_12_12
#define init_14_13_12_12
void FASTCALL call_14_13_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_12_13
#define init_14_13_12_13
void FASTCALL call_14_13_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_12_14
#define init_14_13_12_14
void FASTCALL call_14_13_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_12_15
#define init_14_13_12_15
void FASTCALL call_14_13_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_13_0
#define init_14_13_13_0
void FASTCALL call_14_13_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_13_1
#define init_14_13_13_1
void FASTCALL call_14_13_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_13_2
#define init_14_13_13_2
void FASTCALL call_14_13_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_13_3
#define init_14_13_13_3
void FASTCALL call_14_13_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_13_4
#define init_14_13_13_4
void FASTCALL call_14_13_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_13_5
#define init_14_13_13_5
void FASTCALL call_14_13_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_13_6
#define init_14_13_13_6
void FASTCALL call_14_13_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_13_7
#define init_14_13_13_7
void FASTCALL call_14_13_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_13_8
#define init_14_13_13_8
void FASTCALL call_14_13_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_13_9
#define init_14_13_13_9
void FASTCALL call_14_13_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_13_10
#define init_14_13_13_10
void FASTCALL call_14_13_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_13_11
#define init_14_13_13_11
void FASTCALL call_14_13_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_13_12
#define init_14_13_13_12
void FASTCALL call_14_13_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_13_13
#define init_14_13_13_13
void FASTCALL call_14_13_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_13_14
#define init_14_13_13_14
void FASTCALL call_14_13_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_13_15
#define init_14_13_13_15
void FASTCALL call_14_13_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_14_0
#define init_14_13_14_0
void FASTCALL call_14_13_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_14_1
#define init_14_13_14_1
void FASTCALL call_14_13_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_14_2
#define init_14_13_14_2
void FASTCALL call_14_13_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_14_3
#define init_14_13_14_3
void FASTCALL call_14_13_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_14_4
#define init_14_13_14_4
void FASTCALL call_14_13_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_14_5
#define init_14_13_14_5
void FASTCALL call_14_13_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_14_6
#define init_14_13_14_6
void FASTCALL call_14_13_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_14_7
#define init_14_13_14_7
void FASTCALL call_14_13_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_14_8
#define init_14_13_14_8
void FASTCALL call_14_13_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_14_9
#define init_14_13_14_9
void FASTCALL call_14_13_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_14_10
#define init_14_13_14_10
void FASTCALL call_14_13_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_14_11
#define init_14_13_14_11
void FASTCALL call_14_13_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_14_12
#define init_14_13_14_12
void FASTCALL call_14_13_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_14_13
#define init_14_13_14_13
void FASTCALL call_14_13_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_14_14
#define init_14_13_14_14
void FASTCALL call_14_13_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_14_15
#define init_14_13_14_15
void FASTCALL call_14_13_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_15_0
#define init_14_13_15_0
void FASTCALL call_14_13_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_15_1
#define init_14_13_15_1
void FASTCALL call_14_13_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_15_2
#define init_14_13_15_2
void FASTCALL call_14_13_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_15_3
#define init_14_13_15_3
void FASTCALL call_14_13_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_15_4
#define init_14_13_15_4
void FASTCALL call_14_13_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_15_5
#define init_14_13_15_5
void FASTCALL call_14_13_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_15_6
#define init_14_13_15_6
void FASTCALL call_14_13_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_15_7
#define init_14_13_15_7
void FASTCALL call_14_13_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_15_8
#define init_14_13_15_8
void FASTCALL call_14_13_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_15_9
#define init_14_13_15_9
void FASTCALL call_14_13_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_15_10
#define init_14_13_15_10
void FASTCALL call_14_13_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_15_11
#define init_14_13_15_11
void FASTCALL call_14_13_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_15_12
#define init_14_13_15_12
void FASTCALL call_14_13_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_15_13
#define init_14_13_15_13
void FASTCALL call_14_13_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_15_14
#define init_14_13_15_14
void FASTCALL call_14_13_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_13_15_15
#define init_14_13_15_15
void FASTCALL call_14_13_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_0_0
#define init_14_14_0_0
void FASTCALL call_14_14_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_0_1
#define init_14_14_0_1
void FASTCALL call_14_14_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_0_2
#define init_14_14_0_2
void FASTCALL call_14_14_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_0_3
#define init_14_14_0_3
void FASTCALL call_14_14_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_0_4
#define init_14_14_0_4
void FASTCALL call_14_14_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_0_5
#define init_14_14_0_5
void FASTCALL call_14_14_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_0_6
#define init_14_14_0_6
void FASTCALL call_14_14_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_0_7
#define init_14_14_0_7
void FASTCALL call_14_14_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_0_8
#define init_14_14_0_8
void FASTCALL call_14_14_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_0_9
#define init_14_14_0_9
void FASTCALL call_14_14_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_0_10
#define init_14_14_0_10
void FASTCALL call_14_14_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_0_11
#define init_14_14_0_11
void FASTCALL call_14_14_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_0_12
#define init_14_14_0_12
void FASTCALL call_14_14_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_0_13
#define init_14_14_0_13
void FASTCALL call_14_14_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_0_14
#define init_14_14_0_14
void FASTCALL call_14_14_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_0_15
#define init_14_14_0_15
void FASTCALL call_14_14_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_1_0
#define init_14_14_1_0
void FASTCALL call_14_14_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_1_1
#define init_14_14_1_1
void FASTCALL call_14_14_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_1_2
#define init_14_14_1_2
void FASTCALL call_14_14_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_1_3
#define init_14_14_1_3
void FASTCALL call_14_14_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_1_4
#define init_14_14_1_4
void FASTCALL call_14_14_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_1_5
#define init_14_14_1_5
void FASTCALL call_14_14_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_1_6
#define init_14_14_1_6
void FASTCALL call_14_14_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_1_7
#define init_14_14_1_7
void FASTCALL call_14_14_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_1_8
#define init_14_14_1_8
void FASTCALL call_14_14_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_1_9
#define init_14_14_1_9
void FASTCALL call_14_14_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_1_10
#define init_14_14_1_10
void FASTCALL call_14_14_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_1_11
#define init_14_14_1_11
void FASTCALL call_14_14_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_1_12
#define init_14_14_1_12
void FASTCALL call_14_14_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_1_13
#define init_14_14_1_13
void FASTCALL call_14_14_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_1_14
#define init_14_14_1_14
void FASTCALL call_14_14_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_1_15
#define init_14_14_1_15
void FASTCALL call_14_14_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_2_0
#define init_14_14_2_0
void FASTCALL call_14_14_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_2_1
#define init_14_14_2_1
void FASTCALL call_14_14_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_2_2
#define init_14_14_2_2
void FASTCALL call_14_14_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_2_3
#define init_14_14_2_3
void FASTCALL call_14_14_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_2_4
#define init_14_14_2_4
void FASTCALL call_14_14_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_2_5
#define init_14_14_2_5
void FASTCALL call_14_14_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_2_6
#define init_14_14_2_6
void FASTCALL call_14_14_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_2_7
#define init_14_14_2_7
void FASTCALL call_14_14_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_2_8
#define init_14_14_2_8
void FASTCALL call_14_14_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_2_9
#define init_14_14_2_9
void FASTCALL call_14_14_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_2_10
#define init_14_14_2_10
void FASTCALL call_14_14_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_2_11
#define init_14_14_2_11
void FASTCALL call_14_14_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_2_12
#define init_14_14_2_12
void FASTCALL call_14_14_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_2_13
#define init_14_14_2_13
void FASTCALL call_14_14_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_2_14
#define init_14_14_2_14
void FASTCALL call_14_14_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_2_15
#define init_14_14_2_15
void FASTCALL call_14_14_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_3_0
#define init_14_14_3_0
void FASTCALL call_14_14_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_3_1
#define init_14_14_3_1
void FASTCALL call_14_14_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_3_2
#define init_14_14_3_2
void FASTCALL call_14_14_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_3_3
#define init_14_14_3_3
void FASTCALL call_14_14_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_3_4
#define init_14_14_3_4
void FASTCALL call_14_14_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_3_5
#define init_14_14_3_5
void FASTCALL call_14_14_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_3_6
#define init_14_14_3_6
void FASTCALL call_14_14_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_3_7
#define init_14_14_3_7
void FASTCALL call_14_14_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_3_8
#define init_14_14_3_8
void FASTCALL call_14_14_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_3_9
#define init_14_14_3_9
void FASTCALL call_14_14_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_3_10
#define init_14_14_3_10
void FASTCALL call_14_14_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_3_11
#define init_14_14_3_11
void FASTCALL call_14_14_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_3_12
#define init_14_14_3_12
void FASTCALL call_14_14_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_3_13
#define init_14_14_3_13
void FASTCALL call_14_14_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_3_14
#define init_14_14_3_14
void FASTCALL call_14_14_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_3_15
#define init_14_14_3_15
void FASTCALL call_14_14_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_4_0
#define init_14_14_4_0
void FASTCALL call_14_14_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_4_1
#define init_14_14_4_1
void FASTCALL call_14_14_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_4_2
#define init_14_14_4_2
void FASTCALL call_14_14_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_4_3
#define init_14_14_4_3
void FASTCALL call_14_14_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_4_4
#define init_14_14_4_4
void FASTCALL call_14_14_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_4_5
#define init_14_14_4_5
void FASTCALL call_14_14_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_4_6
#define init_14_14_4_6
void FASTCALL call_14_14_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_4_7
#define init_14_14_4_7
void FASTCALL call_14_14_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_4_8
#define init_14_14_4_8
void FASTCALL call_14_14_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_4_9
#define init_14_14_4_9
void FASTCALL call_14_14_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_4_10
#define init_14_14_4_10
void FASTCALL call_14_14_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_4_11
#define init_14_14_4_11
void FASTCALL call_14_14_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_4_12
#define init_14_14_4_12
void FASTCALL call_14_14_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_4_13
#define init_14_14_4_13
void FASTCALL call_14_14_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_4_14
#define init_14_14_4_14
void FASTCALL call_14_14_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_4_15
#define init_14_14_4_15
void FASTCALL call_14_14_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_5_0
#define init_14_14_5_0
void FASTCALL call_14_14_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_5_1
#define init_14_14_5_1
void FASTCALL call_14_14_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_5_2
#define init_14_14_5_2
void FASTCALL call_14_14_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_5_3
#define init_14_14_5_3
void FASTCALL call_14_14_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_5_4
#define init_14_14_5_4
void FASTCALL call_14_14_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_5_5
#define init_14_14_5_5
void FASTCALL call_14_14_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_5_6
#define init_14_14_5_6
void FASTCALL call_14_14_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_5_7
#define init_14_14_5_7
void FASTCALL call_14_14_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_5_8
#define init_14_14_5_8
void FASTCALL call_14_14_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_5_9
#define init_14_14_5_9
void FASTCALL call_14_14_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_5_10
#define init_14_14_5_10
void FASTCALL call_14_14_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_5_11
#define init_14_14_5_11
void FASTCALL call_14_14_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_5_12
#define init_14_14_5_12
void FASTCALL call_14_14_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_5_13
#define init_14_14_5_13
void FASTCALL call_14_14_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_5_14
#define init_14_14_5_14
void FASTCALL call_14_14_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_5_15
#define init_14_14_5_15
void FASTCALL call_14_14_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_6_0
#define init_14_14_6_0
void FASTCALL call_14_14_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_6_1
#define init_14_14_6_1
void FASTCALL call_14_14_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_6_2
#define init_14_14_6_2
void FASTCALL call_14_14_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_6_3
#define init_14_14_6_3
void FASTCALL call_14_14_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_6_4
#define init_14_14_6_4
void FASTCALL call_14_14_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_6_5
#define init_14_14_6_5
void FASTCALL call_14_14_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_6_6
#define init_14_14_6_6
void FASTCALL call_14_14_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_6_7
#define init_14_14_6_7
void FASTCALL call_14_14_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_6_8
#define init_14_14_6_8
void FASTCALL call_14_14_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_6_9
#define init_14_14_6_9
void FASTCALL call_14_14_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_6_10
#define init_14_14_6_10
void FASTCALL call_14_14_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_6_11
#define init_14_14_6_11
void FASTCALL call_14_14_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_6_12
#define init_14_14_6_12
void FASTCALL call_14_14_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_6_13
#define init_14_14_6_13
void FASTCALL call_14_14_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_6_14
#define init_14_14_6_14
void FASTCALL call_14_14_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_6_15
#define init_14_14_6_15
void FASTCALL call_14_14_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_7_0
#define init_14_14_7_0
void FASTCALL call_14_14_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_7_1
#define init_14_14_7_1
void FASTCALL call_14_14_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_7_2
#define init_14_14_7_2
void FASTCALL call_14_14_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_7_3
#define init_14_14_7_3
void FASTCALL call_14_14_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_7_4
#define init_14_14_7_4
void FASTCALL call_14_14_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_7_5
#define init_14_14_7_5
void FASTCALL call_14_14_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_7_6
#define init_14_14_7_6
void FASTCALL call_14_14_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_7_7
#define init_14_14_7_7
void FASTCALL call_14_14_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_7_8
#define init_14_14_7_8
void FASTCALL call_14_14_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_7_9
#define init_14_14_7_9
void FASTCALL call_14_14_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_7_10
#define init_14_14_7_10
void FASTCALL call_14_14_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_7_11
#define init_14_14_7_11
void FASTCALL call_14_14_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_7_12
#define init_14_14_7_12
void FASTCALL call_14_14_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_7_13
#define init_14_14_7_13
void FASTCALL call_14_14_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_7_14
#define init_14_14_7_14
void FASTCALL call_14_14_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_7_15
#define init_14_14_7_15
void FASTCALL call_14_14_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_8_0
#define init_14_14_8_0
void FASTCALL call_14_14_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_8_1
#define init_14_14_8_1
void FASTCALL call_14_14_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_8_2
#define init_14_14_8_2
void FASTCALL call_14_14_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_8_3
#define init_14_14_8_3
void FASTCALL call_14_14_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_8_4
#define init_14_14_8_4
void FASTCALL call_14_14_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_8_5
#define init_14_14_8_5
void FASTCALL call_14_14_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_8_6
#define init_14_14_8_6
void FASTCALL call_14_14_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_8_7
#define init_14_14_8_7
void FASTCALL call_14_14_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_8_8
#define init_14_14_8_8
void FASTCALL call_14_14_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_8_9
#define init_14_14_8_9
void FASTCALL call_14_14_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_8_10
#define init_14_14_8_10
void FASTCALL call_14_14_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_8_11
#define init_14_14_8_11
void FASTCALL call_14_14_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_8_12
#define init_14_14_8_12
void FASTCALL call_14_14_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_8_13
#define init_14_14_8_13
void FASTCALL call_14_14_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_8_14
#define init_14_14_8_14
void FASTCALL call_14_14_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_8_15
#define init_14_14_8_15
void FASTCALL call_14_14_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_9_0
#define init_14_14_9_0
void FASTCALL call_14_14_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_9_1
#define init_14_14_9_1
void FASTCALL call_14_14_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_9_2
#define init_14_14_9_2
void FASTCALL call_14_14_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_9_3
#define init_14_14_9_3
void FASTCALL call_14_14_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_9_4
#define init_14_14_9_4
void FASTCALL call_14_14_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_9_5
#define init_14_14_9_5
void FASTCALL call_14_14_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_9_6
#define init_14_14_9_6
void FASTCALL call_14_14_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_9_7
#define init_14_14_9_7
void FASTCALL call_14_14_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_9_8
#define init_14_14_9_8
void FASTCALL call_14_14_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_9_9
#define init_14_14_9_9
void FASTCALL call_14_14_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_9_10
#define init_14_14_9_10
void FASTCALL call_14_14_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_9_11
#define init_14_14_9_11
void FASTCALL call_14_14_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_9_12
#define init_14_14_9_12
void FASTCALL call_14_14_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_9_13
#define init_14_14_9_13
void FASTCALL call_14_14_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_9_14
#define init_14_14_9_14
void FASTCALL call_14_14_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_9_15
#define init_14_14_9_15
void FASTCALL call_14_14_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_10_0
#define init_14_14_10_0
void FASTCALL call_14_14_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_10_1
#define init_14_14_10_1
void FASTCALL call_14_14_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_10_2
#define init_14_14_10_2
void FASTCALL call_14_14_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_10_3
#define init_14_14_10_3
void FASTCALL call_14_14_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_10_4
#define init_14_14_10_4
void FASTCALL call_14_14_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_10_5
#define init_14_14_10_5
void FASTCALL call_14_14_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_10_6
#define init_14_14_10_6
void FASTCALL call_14_14_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_10_7
#define init_14_14_10_7
void FASTCALL call_14_14_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_10_8
#define init_14_14_10_8
void FASTCALL call_14_14_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_10_9
#define init_14_14_10_9
void FASTCALL call_14_14_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_10_10
#define init_14_14_10_10
void FASTCALL call_14_14_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_10_11
#define init_14_14_10_11
void FASTCALL call_14_14_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_10_12
#define init_14_14_10_12
void FASTCALL call_14_14_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_10_13
#define init_14_14_10_13
void FASTCALL call_14_14_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_10_14
#define init_14_14_10_14
void FASTCALL call_14_14_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_10_15
#define init_14_14_10_15
void FASTCALL call_14_14_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_11_0
#define init_14_14_11_0
void FASTCALL call_14_14_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_11_1
#define init_14_14_11_1
void FASTCALL call_14_14_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_11_2
#define init_14_14_11_2
void FASTCALL call_14_14_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_11_3
#define init_14_14_11_3
void FASTCALL call_14_14_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_11_4
#define init_14_14_11_4
void FASTCALL call_14_14_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_11_5
#define init_14_14_11_5
void FASTCALL call_14_14_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_11_6
#define init_14_14_11_6
void FASTCALL call_14_14_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_11_7
#define init_14_14_11_7
void FASTCALL call_14_14_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_11_8
#define init_14_14_11_8
void FASTCALL call_14_14_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_11_9
#define init_14_14_11_9
void FASTCALL call_14_14_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_11_10
#define init_14_14_11_10
void FASTCALL call_14_14_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_11_11
#define init_14_14_11_11
void FASTCALL call_14_14_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_11_12
#define init_14_14_11_12
void FASTCALL call_14_14_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_11_13
#define init_14_14_11_13
void FASTCALL call_14_14_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_11_14
#define init_14_14_11_14
void FASTCALL call_14_14_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_11_15
#define init_14_14_11_15
void FASTCALL call_14_14_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_12_0
#define init_14_14_12_0
void FASTCALL call_14_14_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_12_1
#define init_14_14_12_1
void FASTCALL call_14_14_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_12_2
#define init_14_14_12_2
void FASTCALL call_14_14_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_12_3
#define init_14_14_12_3
void FASTCALL call_14_14_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_12_4
#define init_14_14_12_4
void FASTCALL call_14_14_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_12_5
#define init_14_14_12_5
void FASTCALL call_14_14_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_12_6
#define init_14_14_12_6
void FASTCALL call_14_14_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_12_7
#define init_14_14_12_7
void FASTCALL call_14_14_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_12_8
#define init_14_14_12_8
void FASTCALL call_14_14_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_12_9
#define init_14_14_12_9
void FASTCALL call_14_14_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_12_10
#define init_14_14_12_10
void FASTCALL call_14_14_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_12_11
#define init_14_14_12_11
void FASTCALL call_14_14_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_12_12
#define init_14_14_12_12
void FASTCALL call_14_14_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_12_13
#define init_14_14_12_13
void FASTCALL call_14_14_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_12_14
#define init_14_14_12_14
void FASTCALL call_14_14_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_12_15
#define init_14_14_12_15
void FASTCALL call_14_14_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_13_0
#define init_14_14_13_0
void FASTCALL call_14_14_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_13_1
#define init_14_14_13_1
void FASTCALL call_14_14_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_13_2
#define init_14_14_13_2
void FASTCALL call_14_14_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_13_3
#define init_14_14_13_3
void FASTCALL call_14_14_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_13_4
#define init_14_14_13_4
void FASTCALL call_14_14_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_13_5
#define init_14_14_13_5
void FASTCALL call_14_14_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_13_6
#define init_14_14_13_6
void FASTCALL call_14_14_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_13_7
#define init_14_14_13_7
void FASTCALL call_14_14_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_13_8
#define init_14_14_13_8
void FASTCALL call_14_14_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_13_9
#define init_14_14_13_9
void FASTCALL call_14_14_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_13_10
#define init_14_14_13_10
void FASTCALL call_14_14_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_13_11
#define init_14_14_13_11
void FASTCALL call_14_14_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_13_12
#define init_14_14_13_12
void FASTCALL call_14_14_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_13_13
#define init_14_14_13_13
void FASTCALL call_14_14_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_13_14
#define init_14_14_13_14
void FASTCALL call_14_14_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_13_15
#define init_14_14_13_15
void FASTCALL call_14_14_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_14_0
#define init_14_14_14_0
void FASTCALL call_14_14_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_14_1
#define init_14_14_14_1
void FASTCALL call_14_14_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_14_2
#define init_14_14_14_2
void FASTCALL call_14_14_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_14_3
#define init_14_14_14_3
void FASTCALL call_14_14_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_14_4
#define init_14_14_14_4
void FASTCALL call_14_14_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_14_5
#define init_14_14_14_5
void FASTCALL call_14_14_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_14_6
#define init_14_14_14_6
void FASTCALL call_14_14_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_14_7
#define init_14_14_14_7
void FASTCALL call_14_14_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_14_8
#define init_14_14_14_8
void FASTCALL call_14_14_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_14_9
#define init_14_14_14_9
void FASTCALL call_14_14_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_14_10
#define init_14_14_14_10
void FASTCALL call_14_14_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_14_11
#define init_14_14_14_11
void FASTCALL call_14_14_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_14_12
#define init_14_14_14_12
void FASTCALL call_14_14_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_14_13
#define init_14_14_14_13
void FASTCALL call_14_14_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_14_14
#define init_14_14_14_14
void FASTCALL call_14_14_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_14_15
#define init_14_14_14_15
void FASTCALL call_14_14_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_15_0
#define init_14_14_15_0
void FASTCALL call_14_14_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_15_1
#define init_14_14_15_1
void FASTCALL call_14_14_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_15_2
#define init_14_14_15_2
void FASTCALL call_14_14_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_15_3
#define init_14_14_15_3
void FASTCALL call_14_14_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_15_4
#define init_14_14_15_4
void FASTCALL call_14_14_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_15_5
#define init_14_14_15_5
void FASTCALL call_14_14_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_15_6
#define init_14_14_15_6
void FASTCALL call_14_14_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_15_7
#define init_14_14_15_7
void FASTCALL call_14_14_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_15_8
#define init_14_14_15_8
void FASTCALL call_14_14_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_15_9
#define init_14_14_15_9
void FASTCALL call_14_14_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_15_10
#define init_14_14_15_10
void FASTCALL call_14_14_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_15_11
#define init_14_14_15_11
void FASTCALL call_14_14_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_15_12
#define init_14_14_15_12
void FASTCALL call_14_14_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_15_13
#define init_14_14_15_13
void FASTCALL call_14_14_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_15_14
#define init_14_14_15_14
void FASTCALL call_14_14_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_14_15_15
#define init_14_14_15_15
void FASTCALL call_14_14_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_0_0
#define init_14_15_0_0
void FASTCALL call_14_15_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_0_1
#define init_14_15_0_1
void FASTCALL call_14_15_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_0_2
#define init_14_15_0_2
void FASTCALL call_14_15_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_0_3
#define init_14_15_0_3
void FASTCALL call_14_15_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_0_4
#define init_14_15_0_4
void FASTCALL call_14_15_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_0_5
#define init_14_15_0_5
void FASTCALL call_14_15_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_0_6
#define init_14_15_0_6
void FASTCALL call_14_15_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_0_7
#define init_14_15_0_7
void FASTCALL call_14_15_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_0_8
#define init_14_15_0_8
void FASTCALL call_14_15_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_0_9
#define init_14_15_0_9
void FASTCALL call_14_15_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_0_10
#define init_14_15_0_10
void FASTCALL call_14_15_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_0_11
#define init_14_15_0_11
void FASTCALL call_14_15_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_0_12
#define init_14_15_0_12
void FASTCALL call_14_15_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_0_13
#define init_14_15_0_13
void FASTCALL call_14_15_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_0_14
#define init_14_15_0_14
void FASTCALL call_14_15_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_0_15
#define init_14_15_0_15
void FASTCALL call_14_15_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_1_0
#define init_14_15_1_0
void FASTCALL call_14_15_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_1_1
#define init_14_15_1_1
void FASTCALL call_14_15_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_1_2
#define init_14_15_1_2
void FASTCALL call_14_15_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_1_3
#define init_14_15_1_3
void FASTCALL call_14_15_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_1_4
#define init_14_15_1_4
void FASTCALL call_14_15_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_1_5
#define init_14_15_1_5
void FASTCALL call_14_15_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_1_6
#define init_14_15_1_6
void FASTCALL call_14_15_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_1_7
#define init_14_15_1_7
void FASTCALL call_14_15_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_1_8
#define init_14_15_1_8
void FASTCALL call_14_15_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_1_9
#define init_14_15_1_9
void FASTCALL call_14_15_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_1_10
#define init_14_15_1_10
void FASTCALL call_14_15_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_1_11
#define init_14_15_1_11
void FASTCALL call_14_15_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_1_12
#define init_14_15_1_12
void FASTCALL call_14_15_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_1_13
#define init_14_15_1_13
void FASTCALL call_14_15_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_1_14
#define init_14_15_1_14
void FASTCALL call_14_15_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_1_15
#define init_14_15_1_15
void FASTCALL call_14_15_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_2_0
#define init_14_15_2_0
void FASTCALL call_14_15_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_2_1
#define init_14_15_2_1
void FASTCALL call_14_15_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_2_2
#define init_14_15_2_2
void FASTCALL call_14_15_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_2_3
#define init_14_15_2_3
void FASTCALL call_14_15_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_2_4
#define init_14_15_2_4
void FASTCALL call_14_15_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_2_5
#define init_14_15_2_5
void FASTCALL call_14_15_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_2_6
#define init_14_15_2_6
void FASTCALL call_14_15_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_2_7
#define init_14_15_2_7
void FASTCALL call_14_15_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_2_8
#define init_14_15_2_8
void FASTCALL call_14_15_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_2_9
#define init_14_15_2_9
void FASTCALL call_14_15_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_2_10
#define init_14_15_2_10
void FASTCALL call_14_15_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_2_11
#define init_14_15_2_11
void FASTCALL call_14_15_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_2_12
#define init_14_15_2_12
void FASTCALL call_14_15_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_2_13
#define init_14_15_2_13
void FASTCALL call_14_15_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_2_14
#define init_14_15_2_14
void FASTCALL call_14_15_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_2_15
#define init_14_15_2_15
void FASTCALL call_14_15_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_3_0
#define init_14_15_3_0
void FASTCALL call_14_15_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_3_1
#define init_14_15_3_1
void FASTCALL call_14_15_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_3_2
#define init_14_15_3_2
void FASTCALL call_14_15_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_3_3
#define init_14_15_3_3
void FASTCALL call_14_15_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_3_4
#define init_14_15_3_4
void FASTCALL call_14_15_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_3_5
#define init_14_15_3_5
void FASTCALL call_14_15_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_3_6
#define init_14_15_3_6
void FASTCALL call_14_15_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_3_7
#define init_14_15_3_7
void FASTCALL call_14_15_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_3_8
#define init_14_15_3_8
void FASTCALL call_14_15_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_3_9
#define init_14_15_3_9
void FASTCALL call_14_15_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_3_10
#define init_14_15_3_10
void FASTCALL call_14_15_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_3_11
#define init_14_15_3_11
void FASTCALL call_14_15_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_3_12
#define init_14_15_3_12
void FASTCALL call_14_15_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_3_13
#define init_14_15_3_13
void FASTCALL call_14_15_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_3_14
#define init_14_15_3_14
void FASTCALL call_14_15_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_3_15
#define init_14_15_3_15
void FASTCALL call_14_15_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_4_0
#define init_14_15_4_0
void FASTCALL call_14_15_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_4_1
#define init_14_15_4_1
void FASTCALL call_14_15_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_4_2
#define init_14_15_4_2
void FASTCALL call_14_15_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_4_3
#define init_14_15_4_3
void FASTCALL call_14_15_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_4_4
#define init_14_15_4_4
void FASTCALL call_14_15_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_4_5
#define init_14_15_4_5
void FASTCALL call_14_15_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_4_6
#define init_14_15_4_6
void FASTCALL call_14_15_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_4_7
#define init_14_15_4_7
void FASTCALL call_14_15_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_4_8
#define init_14_15_4_8
void FASTCALL call_14_15_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_4_9
#define init_14_15_4_9
void FASTCALL call_14_15_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_4_10
#define init_14_15_4_10
void FASTCALL call_14_15_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_4_11
#define init_14_15_4_11
void FASTCALL call_14_15_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_4_12
#define init_14_15_4_12
void FASTCALL call_14_15_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_4_13
#define init_14_15_4_13
void FASTCALL call_14_15_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_4_14
#define init_14_15_4_14
void FASTCALL call_14_15_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_4_15
#define init_14_15_4_15
void FASTCALL call_14_15_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_5_0
#define init_14_15_5_0
void FASTCALL call_14_15_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_5_1
#define init_14_15_5_1
void FASTCALL call_14_15_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_5_2
#define init_14_15_5_2
void FASTCALL call_14_15_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_5_3
#define init_14_15_5_3
void FASTCALL call_14_15_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_5_4
#define init_14_15_5_4
void FASTCALL call_14_15_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_5_5
#define init_14_15_5_5
void FASTCALL call_14_15_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_5_6
#define init_14_15_5_6
void FASTCALL call_14_15_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_5_7
#define init_14_15_5_7
void FASTCALL call_14_15_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_5_8
#define init_14_15_5_8
void FASTCALL call_14_15_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_5_9
#define init_14_15_5_9
void FASTCALL call_14_15_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_5_10
#define init_14_15_5_10
void FASTCALL call_14_15_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_5_11
#define init_14_15_5_11
void FASTCALL call_14_15_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_5_12
#define init_14_15_5_12
void FASTCALL call_14_15_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_5_13
#define init_14_15_5_13
void FASTCALL call_14_15_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_5_14
#define init_14_15_5_14
void FASTCALL call_14_15_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_5_15
#define init_14_15_5_15
void FASTCALL call_14_15_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_6_0
#define init_14_15_6_0
void FASTCALL call_14_15_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_6_1
#define init_14_15_6_1
void FASTCALL call_14_15_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_6_2
#define init_14_15_6_2
void FASTCALL call_14_15_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_6_3
#define init_14_15_6_3
void FASTCALL call_14_15_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_6_4
#define init_14_15_6_4
void FASTCALL call_14_15_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_6_5
#define init_14_15_6_5
void FASTCALL call_14_15_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_6_6
#define init_14_15_6_6
void FASTCALL call_14_15_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_6_7
#define init_14_15_6_7
void FASTCALL call_14_15_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_6_8
#define init_14_15_6_8
void FASTCALL call_14_15_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_6_9
#define init_14_15_6_9
void FASTCALL call_14_15_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_6_10
#define init_14_15_6_10
void FASTCALL call_14_15_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_6_11
#define init_14_15_6_11
void FASTCALL call_14_15_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_6_12
#define init_14_15_6_12
void FASTCALL call_14_15_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_6_13
#define init_14_15_6_13
void FASTCALL call_14_15_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_6_14
#define init_14_15_6_14
void FASTCALL call_14_15_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_6_15
#define init_14_15_6_15
void FASTCALL call_14_15_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_7_0
#define init_14_15_7_0
void FASTCALL call_14_15_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_7_1
#define init_14_15_7_1
void FASTCALL call_14_15_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_7_2
#define init_14_15_7_2
void FASTCALL call_14_15_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_7_3
#define init_14_15_7_3
void FASTCALL call_14_15_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_7_4
#define init_14_15_7_4
void FASTCALL call_14_15_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_7_5
#define init_14_15_7_5
void FASTCALL call_14_15_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_7_6
#define init_14_15_7_6
void FASTCALL call_14_15_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_7_7
#define init_14_15_7_7
void FASTCALL call_14_15_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_7_8
#define init_14_15_7_8
void FASTCALL call_14_15_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_7_9
#define init_14_15_7_9
void FASTCALL call_14_15_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_7_10
#define init_14_15_7_10
void FASTCALL call_14_15_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_7_11
#define init_14_15_7_11
void FASTCALL call_14_15_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_7_12
#define init_14_15_7_12
void FASTCALL call_14_15_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_7_13
#define init_14_15_7_13
void FASTCALL call_14_15_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_7_14
#define init_14_15_7_14
void FASTCALL call_14_15_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_7_15
#define init_14_15_7_15
void FASTCALL call_14_15_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_8_0
#define init_14_15_8_0
void FASTCALL call_14_15_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_8_1
#define init_14_15_8_1
void FASTCALL call_14_15_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_8_2
#define init_14_15_8_2
void FASTCALL call_14_15_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_8_3
#define init_14_15_8_3
void FASTCALL call_14_15_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_8_4
#define init_14_15_8_4
void FASTCALL call_14_15_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_8_5
#define init_14_15_8_5
void FASTCALL call_14_15_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_8_6
#define init_14_15_8_6
void FASTCALL call_14_15_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_8_7
#define init_14_15_8_7
void FASTCALL call_14_15_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_8_8
#define init_14_15_8_8
void FASTCALL call_14_15_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_8_9
#define init_14_15_8_9
void FASTCALL call_14_15_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_8_10
#define init_14_15_8_10
void FASTCALL call_14_15_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_8_11
#define init_14_15_8_11
void FASTCALL call_14_15_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_8_12
#define init_14_15_8_12
void FASTCALL call_14_15_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_8_13
#define init_14_15_8_13
void FASTCALL call_14_15_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_8_14
#define init_14_15_8_14
void FASTCALL call_14_15_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_8_15
#define init_14_15_8_15
void FASTCALL call_14_15_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_9_0
#define init_14_15_9_0
void FASTCALL call_14_15_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_9_1
#define init_14_15_9_1
void FASTCALL call_14_15_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_9_2
#define init_14_15_9_2
void FASTCALL call_14_15_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_9_3
#define init_14_15_9_3
void FASTCALL call_14_15_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_9_4
#define init_14_15_9_4
void FASTCALL call_14_15_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_9_5
#define init_14_15_9_5
void FASTCALL call_14_15_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_9_6
#define init_14_15_9_6
void FASTCALL call_14_15_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_9_7
#define init_14_15_9_7
void FASTCALL call_14_15_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_9_8
#define init_14_15_9_8
void FASTCALL call_14_15_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_9_9
#define init_14_15_9_9
void FASTCALL call_14_15_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_9_10
#define init_14_15_9_10
void FASTCALL call_14_15_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_9_11
#define init_14_15_9_11
void FASTCALL call_14_15_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_9_12
#define init_14_15_9_12
void FASTCALL call_14_15_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_9_13
#define init_14_15_9_13
void FASTCALL call_14_15_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_9_14
#define init_14_15_9_14
void FASTCALL call_14_15_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_9_15
#define init_14_15_9_15
void FASTCALL call_14_15_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_10_0
#define init_14_15_10_0
void FASTCALL call_14_15_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_10_1
#define init_14_15_10_1
void FASTCALL call_14_15_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_10_2
#define init_14_15_10_2
void FASTCALL call_14_15_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_10_3
#define init_14_15_10_3
void FASTCALL call_14_15_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_10_4
#define init_14_15_10_4
void FASTCALL call_14_15_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_10_5
#define init_14_15_10_5
void FASTCALL call_14_15_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_10_6
#define init_14_15_10_6
void FASTCALL call_14_15_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_10_7
#define init_14_15_10_7
void FASTCALL call_14_15_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_10_8
#define init_14_15_10_8
void FASTCALL call_14_15_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_10_9
#define init_14_15_10_9
void FASTCALL call_14_15_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_10_10
#define init_14_15_10_10
void FASTCALL call_14_15_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_10_11
#define init_14_15_10_11
void FASTCALL call_14_15_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_10_12
#define init_14_15_10_12
void FASTCALL call_14_15_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_10_13
#define init_14_15_10_13
void FASTCALL call_14_15_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_10_14
#define init_14_15_10_14
void FASTCALL call_14_15_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_10_15
#define init_14_15_10_15
void FASTCALL call_14_15_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_11_0
#define init_14_15_11_0
void FASTCALL call_14_15_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_11_1
#define init_14_15_11_1
void FASTCALL call_14_15_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_11_2
#define init_14_15_11_2
void FASTCALL call_14_15_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_11_3
#define init_14_15_11_3
void FASTCALL call_14_15_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_11_4
#define init_14_15_11_4
void FASTCALL call_14_15_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_11_5
#define init_14_15_11_5
void FASTCALL call_14_15_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_11_6
#define init_14_15_11_6
void FASTCALL call_14_15_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_11_7
#define init_14_15_11_7
void FASTCALL call_14_15_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_11_8
#define init_14_15_11_8
void FASTCALL call_14_15_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_11_9
#define init_14_15_11_9
void FASTCALL call_14_15_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_11_10
#define init_14_15_11_10
void FASTCALL call_14_15_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_11_11
#define init_14_15_11_11
void FASTCALL call_14_15_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_11_12
#define init_14_15_11_12
void FASTCALL call_14_15_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_11_13
#define init_14_15_11_13
void FASTCALL call_14_15_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_11_14
#define init_14_15_11_14
void FASTCALL call_14_15_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_11_15
#define init_14_15_11_15
void FASTCALL call_14_15_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_12_0
#define init_14_15_12_0
void FASTCALL call_14_15_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_12_1
#define init_14_15_12_1
void FASTCALL call_14_15_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_12_2
#define init_14_15_12_2
void FASTCALL call_14_15_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_12_3
#define init_14_15_12_3
void FASTCALL call_14_15_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_12_4
#define init_14_15_12_4
void FASTCALL call_14_15_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_12_5
#define init_14_15_12_5
void FASTCALL call_14_15_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_12_6
#define init_14_15_12_6
void FASTCALL call_14_15_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_12_7
#define init_14_15_12_7
void FASTCALL call_14_15_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_12_8
#define init_14_15_12_8
void FASTCALL call_14_15_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_12_9
#define init_14_15_12_9
void FASTCALL call_14_15_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_12_10
#define init_14_15_12_10
void FASTCALL call_14_15_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_12_11
#define init_14_15_12_11
void FASTCALL call_14_15_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_12_12
#define init_14_15_12_12
void FASTCALL call_14_15_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_12_13
#define init_14_15_12_13
void FASTCALL call_14_15_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_12_14
#define init_14_15_12_14
void FASTCALL call_14_15_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_12_15
#define init_14_15_12_15
void FASTCALL call_14_15_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_13_0
#define init_14_15_13_0
void FASTCALL call_14_15_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_13_1
#define init_14_15_13_1
void FASTCALL call_14_15_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_13_2
#define init_14_15_13_2
void FASTCALL call_14_15_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_13_3
#define init_14_15_13_3
void FASTCALL call_14_15_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_13_4
#define init_14_15_13_4
void FASTCALL call_14_15_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_13_5
#define init_14_15_13_5
void FASTCALL call_14_15_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_13_6
#define init_14_15_13_6
void FASTCALL call_14_15_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_13_7
#define init_14_15_13_7
void FASTCALL call_14_15_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_13_8
#define init_14_15_13_8
void FASTCALL call_14_15_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_13_9
#define init_14_15_13_9
void FASTCALL call_14_15_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_13_10
#define init_14_15_13_10
void FASTCALL call_14_15_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_13_11
#define init_14_15_13_11
void FASTCALL call_14_15_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_13_12
#define init_14_15_13_12
void FASTCALL call_14_15_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_13_13
#define init_14_15_13_13
void FASTCALL call_14_15_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_13_14
#define init_14_15_13_14
void FASTCALL call_14_15_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_13_15
#define init_14_15_13_15
void FASTCALL call_14_15_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_14_0
#define init_14_15_14_0
void FASTCALL call_14_15_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_14_1
#define init_14_15_14_1
void FASTCALL call_14_15_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_14_2
#define init_14_15_14_2
void FASTCALL call_14_15_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_14_3
#define init_14_15_14_3
void FASTCALL call_14_15_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_14_4
#define init_14_15_14_4
void FASTCALL call_14_15_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_14_5
#define init_14_15_14_5
void FASTCALL call_14_15_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_14_6
#define init_14_15_14_6
void FASTCALL call_14_15_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_14_7
#define init_14_15_14_7
void FASTCALL call_14_15_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_14_8
#define init_14_15_14_8
void FASTCALL call_14_15_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_14_9
#define init_14_15_14_9
void FASTCALL call_14_15_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_14_10
#define init_14_15_14_10
void FASTCALL call_14_15_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_14_11
#define init_14_15_14_11
void FASTCALL call_14_15_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_14_12
#define init_14_15_14_12
void FASTCALL call_14_15_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_14_13
#define init_14_15_14_13
void FASTCALL call_14_15_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_14_14
#define init_14_15_14_14
void FASTCALL call_14_15_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_14_15
#define init_14_15_14_15
void FASTCALL call_14_15_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_15_0
#define init_14_15_15_0
void FASTCALL call_14_15_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_15_1
#define init_14_15_15_1
void FASTCALL call_14_15_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_15_2
#define init_14_15_15_2
void FASTCALL call_14_15_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_15_3
#define init_14_15_15_3
void FASTCALL call_14_15_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_15_4
#define init_14_15_15_4
void FASTCALL call_14_15_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_15_5
#define init_14_15_15_5
void FASTCALL call_14_15_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_15_6
#define init_14_15_15_6
void FASTCALL call_14_15_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_15_7
#define init_14_15_15_7
void FASTCALL call_14_15_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_15_8
#define init_14_15_15_8
void FASTCALL call_14_15_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_15_9
#define init_14_15_15_9
void FASTCALL call_14_15_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_15_10
#define init_14_15_15_10
void FASTCALL call_14_15_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_15_11
#define init_14_15_15_11
void FASTCALL call_14_15_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_15_12
#define init_14_15_15_12
void FASTCALL call_14_15_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_15_13
#define init_14_15_15_13
void FASTCALL call_14_15_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_15_14
#define init_14_15_15_14
void FASTCALL call_14_15_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_14_15_15_15
#define init_14_15_15_15
void FASTCALL call_14_15_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_0_0
#define init_15_0_0_0
void FASTCALL call_15_0_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_0_1
#define init_15_0_0_1
void FASTCALL call_15_0_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_0_2
#define init_15_0_0_2
void FASTCALL call_15_0_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_0_3
#define init_15_0_0_3
void FASTCALL call_15_0_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_0_4
#define init_15_0_0_4
void FASTCALL call_15_0_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_0_5
#define init_15_0_0_5
void FASTCALL call_15_0_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_0_6
#define init_15_0_0_6
void FASTCALL call_15_0_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_0_7
#define init_15_0_0_7
void FASTCALL call_15_0_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_0_8
#define init_15_0_0_8
void FASTCALL call_15_0_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_0_9
#define init_15_0_0_9
void FASTCALL call_15_0_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_0_10
#define init_15_0_0_10
void FASTCALL call_15_0_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_0_11
#define init_15_0_0_11
void FASTCALL call_15_0_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_0_12
#define init_15_0_0_12
void FASTCALL call_15_0_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_0_13
#define init_15_0_0_13
void FASTCALL call_15_0_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_0_14
#define init_15_0_0_14
void FASTCALL call_15_0_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_0_15
#define init_15_0_0_15
void FASTCALL call_15_0_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_1_0
#define init_15_0_1_0
void FASTCALL call_15_0_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_1_1
#define init_15_0_1_1
void FASTCALL call_15_0_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_1_2
#define init_15_0_1_2
void FASTCALL call_15_0_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_1_3
#define init_15_0_1_3
void FASTCALL call_15_0_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_1_4
#define init_15_0_1_4
void FASTCALL call_15_0_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_1_5
#define init_15_0_1_5
void FASTCALL call_15_0_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_1_6
#define init_15_0_1_6
void FASTCALL call_15_0_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_1_7
#define init_15_0_1_7
void FASTCALL call_15_0_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_1_8
#define init_15_0_1_8
void FASTCALL call_15_0_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_1_9
#define init_15_0_1_9
void FASTCALL call_15_0_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_1_10
#define init_15_0_1_10
void FASTCALL call_15_0_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_1_11
#define init_15_0_1_11
void FASTCALL call_15_0_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_1_12
#define init_15_0_1_12
void FASTCALL call_15_0_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_1_13
#define init_15_0_1_13
void FASTCALL call_15_0_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_1_14
#define init_15_0_1_14
void FASTCALL call_15_0_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_1_15
#define init_15_0_1_15
void FASTCALL call_15_0_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_2_0
#define init_15_0_2_0
void FASTCALL call_15_0_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_2_1
#define init_15_0_2_1
void FASTCALL call_15_0_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_2_2
#define init_15_0_2_2
void FASTCALL call_15_0_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_2_3
#define init_15_0_2_3
void FASTCALL call_15_0_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_2_4
#define init_15_0_2_4
void FASTCALL call_15_0_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_2_5
#define init_15_0_2_5
void FASTCALL call_15_0_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_2_6
#define init_15_0_2_6
void FASTCALL call_15_0_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_2_7
#define init_15_0_2_7
void FASTCALL call_15_0_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_2_8
#define init_15_0_2_8
void FASTCALL call_15_0_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_2_9
#define init_15_0_2_9
void FASTCALL call_15_0_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_2_10
#define init_15_0_2_10
void FASTCALL call_15_0_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_2_11
#define init_15_0_2_11
void FASTCALL call_15_0_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_2_12
#define init_15_0_2_12
void FASTCALL call_15_0_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_2_13
#define init_15_0_2_13
void FASTCALL call_15_0_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_2_14
#define init_15_0_2_14
void FASTCALL call_15_0_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_2_15
#define init_15_0_2_15
void FASTCALL call_15_0_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_3_0
#define init_15_0_3_0
void FASTCALL call_15_0_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_3_1
#define init_15_0_3_1
void FASTCALL call_15_0_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_3_2
#define init_15_0_3_2
void FASTCALL call_15_0_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_3_3
#define init_15_0_3_3
void FASTCALL call_15_0_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_3_4
#define init_15_0_3_4
void FASTCALL call_15_0_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_3_5
#define init_15_0_3_5
void FASTCALL call_15_0_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_3_6
#define init_15_0_3_6
void FASTCALL call_15_0_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_3_7
#define init_15_0_3_7
void FASTCALL call_15_0_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_3_8
#define init_15_0_3_8
void FASTCALL call_15_0_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_3_9
#define init_15_0_3_9
void FASTCALL call_15_0_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_3_10
#define init_15_0_3_10
void FASTCALL call_15_0_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_3_11
#define init_15_0_3_11
void FASTCALL call_15_0_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_3_12
#define init_15_0_3_12
void FASTCALL call_15_0_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_3_13
#define init_15_0_3_13
void FASTCALL call_15_0_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_3_14
#define init_15_0_3_14
void FASTCALL call_15_0_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_3_15
#define init_15_0_3_15
void FASTCALL call_15_0_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_4_0
#define init_15_0_4_0
void FASTCALL call_15_0_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_4_1
#define init_15_0_4_1
void FASTCALL call_15_0_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_4_2
#define init_15_0_4_2
void FASTCALL call_15_0_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_4_3
#define init_15_0_4_3
void FASTCALL call_15_0_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_4_4
#define init_15_0_4_4
void FASTCALL call_15_0_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_4_5
#define init_15_0_4_5
void FASTCALL call_15_0_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_4_6
#define init_15_0_4_6
void FASTCALL call_15_0_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_4_7
#define init_15_0_4_7
void FASTCALL call_15_0_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_4_8
#define init_15_0_4_8
void FASTCALL call_15_0_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_4_9
#define init_15_0_4_9
void FASTCALL call_15_0_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_4_10
#define init_15_0_4_10
void FASTCALL call_15_0_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_4_11
#define init_15_0_4_11
void FASTCALL call_15_0_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_4_12
#define init_15_0_4_12
void FASTCALL call_15_0_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_4_13
#define init_15_0_4_13
void FASTCALL call_15_0_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_4_14
#define init_15_0_4_14
void FASTCALL call_15_0_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_4_15
#define init_15_0_4_15
void FASTCALL call_15_0_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_5_0
#define init_15_0_5_0
void FASTCALL call_15_0_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_5_1
#define init_15_0_5_1
void FASTCALL call_15_0_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_5_2
#define init_15_0_5_2
void FASTCALL call_15_0_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_5_3
#define init_15_0_5_3
void FASTCALL call_15_0_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_5_4
#define init_15_0_5_4
void FASTCALL call_15_0_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_5_5
#define init_15_0_5_5
void FASTCALL call_15_0_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_5_6
#define init_15_0_5_6
void FASTCALL call_15_0_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_5_7
#define init_15_0_5_7
void FASTCALL call_15_0_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_5_8
#define init_15_0_5_8
void FASTCALL call_15_0_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_5_9
#define init_15_0_5_9
void FASTCALL call_15_0_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_5_10
#define init_15_0_5_10
void FASTCALL call_15_0_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_5_11
#define init_15_0_5_11
void FASTCALL call_15_0_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_5_12
#define init_15_0_5_12
void FASTCALL call_15_0_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_5_13
#define init_15_0_5_13
void FASTCALL call_15_0_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_5_14
#define init_15_0_5_14
void FASTCALL call_15_0_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_5_15
#define init_15_0_5_15
void FASTCALL call_15_0_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_6_0
#define init_15_0_6_0
void FASTCALL call_15_0_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_6_1
#define init_15_0_6_1
void FASTCALL call_15_0_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_6_2
#define init_15_0_6_2
void FASTCALL call_15_0_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_6_3
#define init_15_0_6_3
void FASTCALL call_15_0_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_6_4
#define init_15_0_6_4
void FASTCALL call_15_0_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_6_5
#define init_15_0_6_5
void FASTCALL call_15_0_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_6_6
#define init_15_0_6_6
void FASTCALL call_15_0_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_6_7
#define init_15_0_6_7
void FASTCALL call_15_0_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_6_8
#define init_15_0_6_8
void FASTCALL call_15_0_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_6_9
#define init_15_0_6_9
void FASTCALL call_15_0_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_6_10
#define init_15_0_6_10
void FASTCALL call_15_0_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_6_11
#define init_15_0_6_11
void FASTCALL call_15_0_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_6_12
#define init_15_0_6_12
void FASTCALL call_15_0_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_6_13
#define init_15_0_6_13
void FASTCALL call_15_0_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_6_14
#define init_15_0_6_14
void FASTCALL call_15_0_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_6_15
#define init_15_0_6_15
void FASTCALL call_15_0_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_7_0
#define init_15_0_7_0
void FASTCALL call_15_0_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_7_1
#define init_15_0_7_1
void FASTCALL call_15_0_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_7_2
#define init_15_0_7_2
void FASTCALL call_15_0_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_7_3
#define init_15_0_7_3
void FASTCALL call_15_0_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_7_4
#define init_15_0_7_4
void FASTCALL call_15_0_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_7_5
#define init_15_0_7_5
void FASTCALL call_15_0_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_7_6
#define init_15_0_7_6
void FASTCALL call_15_0_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_7_7
#define init_15_0_7_7
void FASTCALL call_15_0_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_7_8
#define init_15_0_7_8
void FASTCALL call_15_0_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_7_9
#define init_15_0_7_9
void FASTCALL call_15_0_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_7_10
#define init_15_0_7_10
void FASTCALL call_15_0_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_7_11
#define init_15_0_7_11
void FASTCALL call_15_0_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_7_12
#define init_15_0_7_12
void FASTCALL call_15_0_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_7_13
#define init_15_0_7_13
void FASTCALL call_15_0_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_7_14
#define init_15_0_7_14
void FASTCALL call_15_0_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_7_15
#define init_15_0_7_15
void FASTCALL call_15_0_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_8_0
#define init_15_0_8_0
void FASTCALL call_15_0_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_8_1
#define init_15_0_8_1
void FASTCALL call_15_0_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_8_2
#define init_15_0_8_2
void FASTCALL call_15_0_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_8_3
#define init_15_0_8_3
void FASTCALL call_15_0_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_8_4
#define init_15_0_8_4
void FASTCALL call_15_0_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_8_5
#define init_15_0_8_5
void FASTCALL call_15_0_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_8_6
#define init_15_0_8_6
void FASTCALL call_15_0_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_8_7
#define init_15_0_8_7
void FASTCALL call_15_0_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_8_8
#define init_15_0_8_8
void FASTCALL call_15_0_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_8_9
#define init_15_0_8_9
void FASTCALL call_15_0_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_8_10
#define init_15_0_8_10
void FASTCALL call_15_0_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_8_11
#define init_15_0_8_11
void FASTCALL call_15_0_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_8_12
#define init_15_0_8_12
void FASTCALL call_15_0_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_8_13
#define init_15_0_8_13
void FASTCALL call_15_0_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_8_14
#define init_15_0_8_14
void FASTCALL call_15_0_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_8_15
#define init_15_0_8_15
void FASTCALL call_15_0_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_9_0
#define init_15_0_9_0
void FASTCALL call_15_0_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_9_1
#define init_15_0_9_1
void FASTCALL call_15_0_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_9_2
#define init_15_0_9_2
void FASTCALL call_15_0_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_9_3
#define init_15_0_9_3
void FASTCALL call_15_0_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_9_4
#define init_15_0_9_4
void FASTCALL call_15_0_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_9_5
#define init_15_0_9_5
void FASTCALL call_15_0_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_9_6
#define init_15_0_9_6
void FASTCALL call_15_0_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_9_7
#define init_15_0_9_7
void FASTCALL call_15_0_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_9_8
#define init_15_0_9_8
void FASTCALL call_15_0_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_9_9
#define init_15_0_9_9
void FASTCALL call_15_0_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_9_10
#define init_15_0_9_10
void FASTCALL call_15_0_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_9_11
#define init_15_0_9_11
void FASTCALL call_15_0_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_9_12
#define init_15_0_9_12
void FASTCALL call_15_0_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_9_13
#define init_15_0_9_13
void FASTCALL call_15_0_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_9_14
#define init_15_0_9_14
void FASTCALL call_15_0_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_9_15
#define init_15_0_9_15
void FASTCALL call_15_0_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_10_0
#define init_15_0_10_0
void FASTCALL call_15_0_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_10_1
#define init_15_0_10_1
void FASTCALL call_15_0_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_10_2
#define init_15_0_10_2
void FASTCALL call_15_0_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_10_3
#define init_15_0_10_3
void FASTCALL call_15_0_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_10_4
#define init_15_0_10_4
void FASTCALL call_15_0_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_10_5
#define init_15_0_10_5
void FASTCALL call_15_0_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_10_6
#define init_15_0_10_6
void FASTCALL call_15_0_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_10_7
#define init_15_0_10_7
void FASTCALL call_15_0_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_10_8
#define init_15_0_10_8
void FASTCALL call_15_0_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_10_9
#define init_15_0_10_9
void FASTCALL call_15_0_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_10_10
#define init_15_0_10_10
void FASTCALL call_15_0_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_10_11
#define init_15_0_10_11
void FASTCALL call_15_0_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_10_12
#define init_15_0_10_12
void FASTCALL call_15_0_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_10_13
#define init_15_0_10_13
void FASTCALL call_15_0_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_10_14
#define init_15_0_10_14
void FASTCALL call_15_0_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_10_15
#define init_15_0_10_15
void FASTCALL call_15_0_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_11_0
#define init_15_0_11_0
void FASTCALL call_15_0_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_11_1
#define init_15_0_11_1
void FASTCALL call_15_0_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_11_2
#define init_15_0_11_2
void FASTCALL call_15_0_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_11_3
#define init_15_0_11_3
void FASTCALL call_15_0_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_11_4
#define init_15_0_11_4
void FASTCALL call_15_0_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_11_5
#define init_15_0_11_5
void FASTCALL call_15_0_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_11_6
#define init_15_0_11_6
void FASTCALL call_15_0_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_11_7
#define init_15_0_11_7
void FASTCALL call_15_0_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_11_8
#define init_15_0_11_8
void FASTCALL call_15_0_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_11_9
#define init_15_0_11_9
void FASTCALL call_15_0_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_11_10
#define init_15_0_11_10
void FASTCALL call_15_0_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_11_11
#define init_15_0_11_11
void FASTCALL call_15_0_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_11_12
#define init_15_0_11_12
void FASTCALL call_15_0_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_11_13
#define init_15_0_11_13
void FASTCALL call_15_0_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_11_14
#define init_15_0_11_14
void FASTCALL call_15_0_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_11_15
#define init_15_0_11_15
void FASTCALL call_15_0_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_12_0
#define init_15_0_12_0
void FASTCALL call_15_0_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_12_1
#define init_15_0_12_1
void FASTCALL call_15_0_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_12_2
#define init_15_0_12_2
void FASTCALL call_15_0_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_12_3
#define init_15_0_12_3
void FASTCALL call_15_0_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_12_4
#define init_15_0_12_4
void FASTCALL call_15_0_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_12_5
#define init_15_0_12_5
void FASTCALL call_15_0_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_12_6
#define init_15_0_12_6
void FASTCALL call_15_0_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_12_7
#define init_15_0_12_7
void FASTCALL call_15_0_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_12_8
#define init_15_0_12_8
void FASTCALL call_15_0_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_12_9
#define init_15_0_12_9
void FASTCALL call_15_0_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_12_10
#define init_15_0_12_10
void FASTCALL call_15_0_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_12_11
#define init_15_0_12_11
void FASTCALL call_15_0_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_12_12
#define init_15_0_12_12
void FASTCALL call_15_0_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_12_13
#define init_15_0_12_13
void FASTCALL call_15_0_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_12_14
#define init_15_0_12_14
void FASTCALL call_15_0_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_12_15
#define init_15_0_12_15
void FASTCALL call_15_0_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_13_0
#define init_15_0_13_0
void FASTCALL call_15_0_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_13_1
#define init_15_0_13_1
void FASTCALL call_15_0_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_13_2
#define init_15_0_13_2
void FASTCALL call_15_0_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_13_3
#define init_15_0_13_3
void FASTCALL call_15_0_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_13_4
#define init_15_0_13_4
void FASTCALL call_15_0_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_13_5
#define init_15_0_13_5
void FASTCALL call_15_0_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_13_6
#define init_15_0_13_6
void FASTCALL call_15_0_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_13_7
#define init_15_0_13_7
void FASTCALL call_15_0_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_13_8
#define init_15_0_13_8
void FASTCALL call_15_0_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_13_9
#define init_15_0_13_9
void FASTCALL call_15_0_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_13_10
#define init_15_0_13_10
void FASTCALL call_15_0_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_13_11
#define init_15_0_13_11
void FASTCALL call_15_0_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_13_12
#define init_15_0_13_12
void FASTCALL call_15_0_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_13_13
#define init_15_0_13_13
void FASTCALL call_15_0_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_13_14
#define init_15_0_13_14
void FASTCALL call_15_0_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_13_15
#define init_15_0_13_15
void FASTCALL call_15_0_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_14_0
#define init_15_0_14_0
void FASTCALL call_15_0_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_14_1
#define init_15_0_14_1
void FASTCALL call_15_0_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_14_2
#define init_15_0_14_2
void FASTCALL call_15_0_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_14_3
#define init_15_0_14_3
void FASTCALL call_15_0_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_14_4
#define init_15_0_14_4
void FASTCALL call_15_0_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_14_5
#define init_15_0_14_5
void FASTCALL call_15_0_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_14_6
#define init_15_0_14_6
void FASTCALL call_15_0_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_14_7
#define init_15_0_14_7
void FASTCALL call_15_0_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_14_8
#define init_15_0_14_8
void FASTCALL call_15_0_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_14_9
#define init_15_0_14_9
void FASTCALL call_15_0_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_14_10
#define init_15_0_14_10
void FASTCALL call_15_0_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_14_11
#define init_15_0_14_11
void FASTCALL call_15_0_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_14_12
#define init_15_0_14_12
void FASTCALL call_15_0_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_14_13
#define init_15_0_14_13
void FASTCALL call_15_0_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_14_14
#define init_15_0_14_14
void FASTCALL call_15_0_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_14_15
#define init_15_0_14_15
void FASTCALL call_15_0_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_15_0
#define init_15_0_15_0
void FASTCALL call_15_0_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_15_1
#define init_15_0_15_1
void FASTCALL call_15_0_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_15_2
#define init_15_0_15_2
void FASTCALL call_15_0_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_15_3
#define init_15_0_15_3
void FASTCALL call_15_0_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_15_4
#define init_15_0_15_4
void FASTCALL call_15_0_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_15_5
#define init_15_0_15_5
void FASTCALL call_15_0_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_15_6
#define init_15_0_15_6
void FASTCALL call_15_0_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_15_7
#define init_15_0_15_7
void FASTCALL call_15_0_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_15_8
#define init_15_0_15_8
void FASTCALL call_15_0_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_15_9
#define init_15_0_15_9
void FASTCALL call_15_0_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_15_10
#define init_15_0_15_10
void FASTCALL call_15_0_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_15_11
#define init_15_0_15_11
void FASTCALL call_15_0_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_15_12
#define init_15_0_15_12
void FASTCALL call_15_0_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_15_13
#define init_15_0_15_13
void FASTCALL call_15_0_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_15_14
#define init_15_0_15_14
void FASTCALL call_15_0_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_0_15_15
#define init_15_0_15_15
void FASTCALL call_15_0_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_0_0
#define init_15_1_0_0
void FASTCALL call_15_1_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_0_1
#define init_15_1_0_1
void FASTCALL call_15_1_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_0_2
#define init_15_1_0_2
void FASTCALL call_15_1_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_0_3
#define init_15_1_0_3
void FASTCALL call_15_1_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_0_4
#define init_15_1_0_4
void FASTCALL call_15_1_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_0_5
#define init_15_1_0_5
void FASTCALL call_15_1_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_0_6
#define init_15_1_0_6
void FASTCALL call_15_1_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_0_7
#define init_15_1_0_7
void FASTCALL call_15_1_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_0_8
#define init_15_1_0_8
void FASTCALL call_15_1_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_0_9
#define init_15_1_0_9
void FASTCALL call_15_1_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_0_10
#define init_15_1_0_10
void FASTCALL call_15_1_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_0_11
#define init_15_1_0_11
void FASTCALL call_15_1_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_0_12
#define init_15_1_0_12
void FASTCALL call_15_1_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_0_13
#define init_15_1_0_13
void FASTCALL call_15_1_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_0_14
#define init_15_1_0_14
void FASTCALL call_15_1_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_0_15
#define init_15_1_0_15
void FASTCALL call_15_1_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_1_0
#define init_15_1_1_0
void FASTCALL call_15_1_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_1_1
#define init_15_1_1_1
void FASTCALL call_15_1_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_1_2
#define init_15_1_1_2
void FASTCALL call_15_1_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_1_3
#define init_15_1_1_3
void FASTCALL call_15_1_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_1_4
#define init_15_1_1_4
void FASTCALL call_15_1_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_1_5
#define init_15_1_1_5
void FASTCALL call_15_1_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_1_6
#define init_15_1_1_6
void FASTCALL call_15_1_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_1_7
#define init_15_1_1_7
void FASTCALL call_15_1_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_1_8
#define init_15_1_1_8
void FASTCALL call_15_1_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_1_9
#define init_15_1_1_9
void FASTCALL call_15_1_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_1_10
#define init_15_1_1_10
void FASTCALL call_15_1_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_1_11
#define init_15_1_1_11
void FASTCALL call_15_1_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_1_12
#define init_15_1_1_12
void FASTCALL call_15_1_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_1_13
#define init_15_1_1_13
void FASTCALL call_15_1_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_1_14
#define init_15_1_1_14
void FASTCALL call_15_1_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_1_15
#define init_15_1_1_15
void FASTCALL call_15_1_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_2_0
#define init_15_1_2_0
void FASTCALL call_15_1_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_2_1
#define init_15_1_2_1
void FASTCALL call_15_1_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_2_2
#define init_15_1_2_2
void FASTCALL call_15_1_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_2_3
#define init_15_1_2_3
void FASTCALL call_15_1_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_2_4
#define init_15_1_2_4
void FASTCALL call_15_1_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_2_5
#define init_15_1_2_5
void FASTCALL call_15_1_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_2_6
#define init_15_1_2_6
void FASTCALL call_15_1_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_2_7
#define init_15_1_2_7
void FASTCALL call_15_1_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_2_8
#define init_15_1_2_8
void FASTCALL call_15_1_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_2_9
#define init_15_1_2_9
void FASTCALL call_15_1_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_2_10
#define init_15_1_2_10
void FASTCALL call_15_1_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_2_11
#define init_15_1_2_11
void FASTCALL call_15_1_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_2_12
#define init_15_1_2_12
void FASTCALL call_15_1_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_2_13
#define init_15_1_2_13
void FASTCALL call_15_1_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_2_14
#define init_15_1_2_14
void FASTCALL call_15_1_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_2_15
#define init_15_1_2_15
void FASTCALL call_15_1_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_3_0
#define init_15_1_3_0
void FASTCALL call_15_1_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_3_1
#define init_15_1_3_1
void FASTCALL call_15_1_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_3_2
#define init_15_1_3_2
void FASTCALL call_15_1_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_3_3
#define init_15_1_3_3
void FASTCALL call_15_1_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_3_4
#define init_15_1_3_4
void FASTCALL call_15_1_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_3_5
#define init_15_1_3_5
void FASTCALL call_15_1_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_3_6
#define init_15_1_3_6
void FASTCALL call_15_1_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_3_7
#define init_15_1_3_7
void FASTCALL call_15_1_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_3_8
#define init_15_1_3_8
void FASTCALL call_15_1_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_3_9
#define init_15_1_3_9
void FASTCALL call_15_1_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_3_10
#define init_15_1_3_10
void FASTCALL call_15_1_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_3_11
#define init_15_1_3_11
void FASTCALL call_15_1_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_3_12
#define init_15_1_3_12
void FASTCALL call_15_1_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_3_13
#define init_15_1_3_13
void FASTCALL call_15_1_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_3_14
#define init_15_1_3_14
void FASTCALL call_15_1_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_3_15
#define init_15_1_3_15
void FASTCALL call_15_1_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_4_0
#define init_15_1_4_0
void FASTCALL call_15_1_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_4_1
#define init_15_1_4_1
void FASTCALL call_15_1_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_4_2
#define init_15_1_4_2
void FASTCALL call_15_1_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_4_3
#define init_15_1_4_3
void FASTCALL call_15_1_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_4_4
#define init_15_1_4_4
void FASTCALL call_15_1_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_4_5
#define init_15_1_4_5
void FASTCALL call_15_1_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_4_6
#define init_15_1_4_6
void FASTCALL call_15_1_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_4_7
#define init_15_1_4_7
void FASTCALL call_15_1_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_4_8
#define init_15_1_4_8
void FASTCALL call_15_1_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_4_9
#define init_15_1_4_9
void FASTCALL call_15_1_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_4_10
#define init_15_1_4_10
void FASTCALL call_15_1_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_4_11
#define init_15_1_4_11
void FASTCALL call_15_1_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_4_12
#define init_15_1_4_12
void FASTCALL call_15_1_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_4_13
#define init_15_1_4_13
void FASTCALL call_15_1_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_4_14
#define init_15_1_4_14
void FASTCALL call_15_1_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_4_15
#define init_15_1_4_15
void FASTCALL call_15_1_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_5_0
#define init_15_1_5_0
void FASTCALL call_15_1_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_5_1
#define init_15_1_5_1
void FASTCALL call_15_1_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_5_2
#define init_15_1_5_2
void FASTCALL call_15_1_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_5_3
#define init_15_1_5_3
void FASTCALL call_15_1_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_5_4
#define init_15_1_5_4
void FASTCALL call_15_1_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_5_5
#define init_15_1_5_5
void FASTCALL call_15_1_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_5_6
#define init_15_1_5_6
void FASTCALL call_15_1_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_5_7
#define init_15_1_5_7
void FASTCALL call_15_1_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_5_8
#define init_15_1_5_8
void FASTCALL call_15_1_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_5_9
#define init_15_1_5_9
void FASTCALL call_15_1_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_5_10
#define init_15_1_5_10
void FASTCALL call_15_1_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_5_11
#define init_15_1_5_11
void FASTCALL call_15_1_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_5_12
#define init_15_1_5_12
void FASTCALL call_15_1_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_5_13
#define init_15_1_5_13
void FASTCALL call_15_1_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_5_14
#define init_15_1_5_14
void FASTCALL call_15_1_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_5_15
#define init_15_1_5_15
void FASTCALL call_15_1_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_6_0
#define init_15_1_6_0
void FASTCALL call_15_1_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_6_1
#define init_15_1_6_1
void FASTCALL call_15_1_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_6_2
#define init_15_1_6_2
void FASTCALL call_15_1_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_6_3
#define init_15_1_6_3
void FASTCALL call_15_1_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_6_4
#define init_15_1_6_4
void FASTCALL call_15_1_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_6_5
#define init_15_1_6_5
void FASTCALL call_15_1_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_6_6
#define init_15_1_6_6
void FASTCALL call_15_1_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_6_7
#define init_15_1_6_7
void FASTCALL call_15_1_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_6_8
#define init_15_1_6_8
void FASTCALL call_15_1_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_6_9
#define init_15_1_6_9
void FASTCALL call_15_1_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_6_10
#define init_15_1_6_10
void FASTCALL call_15_1_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_6_11
#define init_15_1_6_11
void FASTCALL call_15_1_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_6_12
#define init_15_1_6_12
void FASTCALL call_15_1_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_6_13
#define init_15_1_6_13
void FASTCALL call_15_1_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_6_14
#define init_15_1_6_14
void FASTCALL call_15_1_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_6_15
#define init_15_1_6_15
void FASTCALL call_15_1_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_7_0
#define init_15_1_7_0
void FASTCALL call_15_1_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_7_1
#define init_15_1_7_1
void FASTCALL call_15_1_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_7_2
#define init_15_1_7_2
void FASTCALL call_15_1_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_7_3
#define init_15_1_7_3
void FASTCALL call_15_1_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_7_4
#define init_15_1_7_4
void FASTCALL call_15_1_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_7_5
#define init_15_1_7_5
void FASTCALL call_15_1_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_7_6
#define init_15_1_7_6
void FASTCALL call_15_1_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_7_7
#define init_15_1_7_7
void FASTCALL call_15_1_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_7_8
#define init_15_1_7_8
void FASTCALL call_15_1_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_7_9
#define init_15_1_7_9
void FASTCALL call_15_1_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_7_10
#define init_15_1_7_10
void FASTCALL call_15_1_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_7_11
#define init_15_1_7_11
void FASTCALL call_15_1_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_7_12
#define init_15_1_7_12
void FASTCALL call_15_1_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_7_13
#define init_15_1_7_13
void FASTCALL call_15_1_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_7_14
#define init_15_1_7_14
void FASTCALL call_15_1_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_7_15
#define init_15_1_7_15
void FASTCALL call_15_1_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_8_0
#define init_15_1_8_0
void FASTCALL call_15_1_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_8_1
#define init_15_1_8_1
void FASTCALL call_15_1_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_8_2
#define init_15_1_8_2
void FASTCALL call_15_1_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_8_3
#define init_15_1_8_3
void FASTCALL call_15_1_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_8_4
#define init_15_1_8_4
void FASTCALL call_15_1_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_8_5
#define init_15_1_8_5
void FASTCALL call_15_1_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_8_6
#define init_15_1_8_6
void FASTCALL call_15_1_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_8_7
#define init_15_1_8_7
void FASTCALL call_15_1_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_8_8
#define init_15_1_8_8
void FASTCALL call_15_1_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_8_9
#define init_15_1_8_9
void FASTCALL call_15_1_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_8_10
#define init_15_1_8_10
void FASTCALL call_15_1_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_8_11
#define init_15_1_8_11
void FASTCALL call_15_1_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_8_12
#define init_15_1_8_12
void FASTCALL call_15_1_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_8_13
#define init_15_1_8_13
void FASTCALL call_15_1_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_8_14
#define init_15_1_8_14
void FASTCALL call_15_1_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_8_15
#define init_15_1_8_15
void FASTCALL call_15_1_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_9_0
#define init_15_1_9_0
void FASTCALL call_15_1_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_9_1
#define init_15_1_9_1
void FASTCALL call_15_1_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_9_2
#define init_15_1_9_2
void FASTCALL call_15_1_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_9_3
#define init_15_1_9_3
void FASTCALL call_15_1_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_9_4
#define init_15_1_9_4
void FASTCALL call_15_1_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_9_5
#define init_15_1_9_5
void FASTCALL call_15_1_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_9_6
#define init_15_1_9_6
void FASTCALL call_15_1_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_9_7
#define init_15_1_9_7
void FASTCALL call_15_1_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_9_8
#define init_15_1_9_8
void FASTCALL call_15_1_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_9_9
#define init_15_1_9_9
void FASTCALL call_15_1_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_9_10
#define init_15_1_9_10
void FASTCALL call_15_1_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_9_11
#define init_15_1_9_11
void FASTCALL call_15_1_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_9_12
#define init_15_1_9_12
void FASTCALL call_15_1_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_9_13
#define init_15_1_9_13
void FASTCALL call_15_1_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_9_14
#define init_15_1_9_14
void FASTCALL call_15_1_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_9_15
#define init_15_1_9_15
void FASTCALL call_15_1_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_10_0
#define init_15_1_10_0
void FASTCALL call_15_1_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_10_1
#define init_15_1_10_1
void FASTCALL call_15_1_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_10_2
#define init_15_1_10_2
void FASTCALL call_15_1_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_10_3
#define init_15_1_10_3
void FASTCALL call_15_1_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_10_4
#define init_15_1_10_4
void FASTCALL call_15_1_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_10_5
#define init_15_1_10_5
void FASTCALL call_15_1_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_10_6
#define init_15_1_10_6
void FASTCALL call_15_1_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_10_7
#define init_15_1_10_7
void FASTCALL call_15_1_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_10_8
#define init_15_1_10_8
void FASTCALL call_15_1_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_10_9
#define init_15_1_10_9
void FASTCALL call_15_1_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_10_10
#define init_15_1_10_10
void FASTCALL call_15_1_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_10_11
#define init_15_1_10_11
void FASTCALL call_15_1_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_10_12
#define init_15_1_10_12
void FASTCALL call_15_1_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_10_13
#define init_15_1_10_13
void FASTCALL call_15_1_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_10_14
#define init_15_1_10_14
void FASTCALL call_15_1_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_10_15
#define init_15_1_10_15
void FASTCALL call_15_1_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_11_0
#define init_15_1_11_0
void FASTCALL call_15_1_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_11_1
#define init_15_1_11_1
void FASTCALL call_15_1_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_11_2
#define init_15_1_11_2
void FASTCALL call_15_1_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_11_3
#define init_15_1_11_3
void FASTCALL call_15_1_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_11_4
#define init_15_1_11_4
void FASTCALL call_15_1_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_11_5
#define init_15_1_11_5
void FASTCALL call_15_1_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_11_6
#define init_15_1_11_6
void FASTCALL call_15_1_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_11_7
#define init_15_1_11_7
void FASTCALL call_15_1_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_11_8
#define init_15_1_11_8
void FASTCALL call_15_1_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_11_9
#define init_15_1_11_9
void FASTCALL call_15_1_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_11_10
#define init_15_1_11_10
void FASTCALL call_15_1_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_11_11
#define init_15_1_11_11
void FASTCALL call_15_1_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_11_12
#define init_15_1_11_12
void FASTCALL call_15_1_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_11_13
#define init_15_1_11_13
void FASTCALL call_15_1_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_11_14
#define init_15_1_11_14
void FASTCALL call_15_1_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_11_15
#define init_15_1_11_15
void FASTCALL call_15_1_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_12_0
#define init_15_1_12_0
void FASTCALL call_15_1_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_12_1
#define init_15_1_12_1
void FASTCALL call_15_1_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_12_2
#define init_15_1_12_2
void FASTCALL call_15_1_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_12_3
#define init_15_1_12_3
void FASTCALL call_15_1_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_12_4
#define init_15_1_12_4
void FASTCALL call_15_1_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_12_5
#define init_15_1_12_5
void FASTCALL call_15_1_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_12_6
#define init_15_1_12_6
void FASTCALL call_15_1_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_12_7
#define init_15_1_12_7
void FASTCALL call_15_1_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_12_8
#define init_15_1_12_8
void FASTCALL call_15_1_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_12_9
#define init_15_1_12_9
void FASTCALL call_15_1_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_12_10
#define init_15_1_12_10
void FASTCALL call_15_1_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_12_11
#define init_15_1_12_11
void FASTCALL call_15_1_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_12_12
#define init_15_1_12_12
void FASTCALL call_15_1_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_12_13
#define init_15_1_12_13
void FASTCALL call_15_1_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_12_14
#define init_15_1_12_14
void FASTCALL call_15_1_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_12_15
#define init_15_1_12_15
void FASTCALL call_15_1_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_13_0
#define init_15_1_13_0
void FASTCALL call_15_1_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_13_1
#define init_15_1_13_1
void FASTCALL call_15_1_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_13_2
#define init_15_1_13_2
void FASTCALL call_15_1_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_13_3
#define init_15_1_13_3
void FASTCALL call_15_1_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_13_4
#define init_15_1_13_4
void FASTCALL call_15_1_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_13_5
#define init_15_1_13_5
void FASTCALL call_15_1_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_13_6
#define init_15_1_13_6
void FASTCALL call_15_1_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_13_7
#define init_15_1_13_7
void FASTCALL call_15_1_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_13_8
#define init_15_1_13_8
void FASTCALL call_15_1_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_13_9
#define init_15_1_13_9
void FASTCALL call_15_1_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_13_10
#define init_15_1_13_10
void FASTCALL call_15_1_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_13_11
#define init_15_1_13_11
void FASTCALL call_15_1_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_13_12
#define init_15_1_13_12
void FASTCALL call_15_1_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_13_13
#define init_15_1_13_13
void FASTCALL call_15_1_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_13_14
#define init_15_1_13_14
void FASTCALL call_15_1_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_13_15
#define init_15_1_13_15
void FASTCALL call_15_1_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_14_0
#define init_15_1_14_0
void FASTCALL call_15_1_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_14_1
#define init_15_1_14_1
void FASTCALL call_15_1_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_14_2
#define init_15_1_14_2
void FASTCALL call_15_1_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_14_3
#define init_15_1_14_3
void FASTCALL call_15_1_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_14_4
#define init_15_1_14_4
void FASTCALL call_15_1_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_14_5
#define init_15_1_14_5
void FASTCALL call_15_1_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_14_6
#define init_15_1_14_6
void FASTCALL call_15_1_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_14_7
#define init_15_1_14_7
void FASTCALL call_15_1_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_14_8
#define init_15_1_14_8
void FASTCALL call_15_1_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_14_9
#define init_15_1_14_9
void FASTCALL call_15_1_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_14_10
#define init_15_1_14_10
void FASTCALL call_15_1_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_14_11
#define init_15_1_14_11
void FASTCALL call_15_1_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_14_12
#define init_15_1_14_12
void FASTCALL call_15_1_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_14_13
#define init_15_1_14_13
void FASTCALL call_15_1_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_14_14
#define init_15_1_14_14
void FASTCALL call_15_1_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_14_15
#define init_15_1_14_15
void FASTCALL call_15_1_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_15_0
#define init_15_1_15_0
void FASTCALL call_15_1_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_15_1
#define init_15_1_15_1
void FASTCALL call_15_1_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_15_2
#define init_15_1_15_2
void FASTCALL call_15_1_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_15_3
#define init_15_1_15_3
void FASTCALL call_15_1_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_15_4
#define init_15_1_15_4
void FASTCALL call_15_1_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_15_5
#define init_15_1_15_5
void FASTCALL call_15_1_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_15_6
#define init_15_1_15_6
void FASTCALL call_15_1_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_15_7
#define init_15_1_15_7
void FASTCALL call_15_1_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_15_8
#define init_15_1_15_8
void FASTCALL call_15_1_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_15_9
#define init_15_1_15_9
void FASTCALL call_15_1_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_15_10
#define init_15_1_15_10
void FASTCALL call_15_1_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_15_11
#define init_15_1_15_11
void FASTCALL call_15_1_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_15_12
#define init_15_1_15_12
void FASTCALL call_15_1_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_15_13
#define init_15_1_15_13
void FASTCALL call_15_1_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_15_14
#define init_15_1_15_14
void FASTCALL call_15_1_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_1_15_15
#define init_15_1_15_15
void FASTCALL call_15_1_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_0_0
#define init_15_2_0_0
void FASTCALL call_15_2_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_0_1
#define init_15_2_0_1
void FASTCALL call_15_2_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_0_2
#define init_15_2_0_2
void FASTCALL call_15_2_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_0_3
#define init_15_2_0_3
void FASTCALL call_15_2_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_0_4
#define init_15_2_0_4
void FASTCALL call_15_2_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_0_5
#define init_15_2_0_5
void FASTCALL call_15_2_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_0_6
#define init_15_2_0_6
void FASTCALL call_15_2_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_0_7
#define init_15_2_0_7
void FASTCALL call_15_2_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_0_8
#define init_15_2_0_8
void FASTCALL call_15_2_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_0_9
#define init_15_2_0_9
void FASTCALL call_15_2_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_0_10
#define init_15_2_0_10
void FASTCALL call_15_2_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_0_11
#define init_15_2_0_11
void FASTCALL call_15_2_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_0_12
#define init_15_2_0_12
void FASTCALL call_15_2_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_0_13
#define init_15_2_0_13
void FASTCALL call_15_2_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_0_14
#define init_15_2_0_14
void FASTCALL call_15_2_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_0_15
#define init_15_2_0_15
void FASTCALL call_15_2_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_1_0
#define init_15_2_1_0
void FASTCALL call_15_2_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_1_1
#define init_15_2_1_1
void FASTCALL call_15_2_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_1_2
#define init_15_2_1_2
void FASTCALL call_15_2_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_1_3
#define init_15_2_1_3
void FASTCALL call_15_2_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_1_4
#define init_15_2_1_4
void FASTCALL call_15_2_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_1_5
#define init_15_2_1_5
void FASTCALL call_15_2_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_1_6
#define init_15_2_1_6
void FASTCALL call_15_2_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_1_7
#define init_15_2_1_7
void FASTCALL call_15_2_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_1_8
#define init_15_2_1_8
void FASTCALL call_15_2_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_1_9
#define init_15_2_1_9
void FASTCALL call_15_2_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_1_10
#define init_15_2_1_10
void FASTCALL call_15_2_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_1_11
#define init_15_2_1_11
void FASTCALL call_15_2_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_1_12
#define init_15_2_1_12
void FASTCALL call_15_2_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_1_13
#define init_15_2_1_13
void FASTCALL call_15_2_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_1_14
#define init_15_2_1_14
void FASTCALL call_15_2_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_1_15
#define init_15_2_1_15
void FASTCALL call_15_2_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_2_0
#define init_15_2_2_0
void FASTCALL call_15_2_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_2_1
#define init_15_2_2_1
void FASTCALL call_15_2_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_2_2
#define init_15_2_2_2
void FASTCALL call_15_2_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_2_3
#define init_15_2_2_3
void FASTCALL call_15_2_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_2_4
#define init_15_2_2_4
void FASTCALL call_15_2_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_2_5
#define init_15_2_2_5
void FASTCALL call_15_2_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_2_6
#define init_15_2_2_6
void FASTCALL call_15_2_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_2_7
#define init_15_2_2_7
void FASTCALL call_15_2_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_2_8
#define init_15_2_2_8
void FASTCALL call_15_2_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_2_9
#define init_15_2_2_9
void FASTCALL call_15_2_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_2_10
#define init_15_2_2_10
void FASTCALL call_15_2_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_2_11
#define init_15_2_2_11
void FASTCALL call_15_2_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_2_12
#define init_15_2_2_12
void FASTCALL call_15_2_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_2_13
#define init_15_2_2_13
void FASTCALL call_15_2_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_2_14
#define init_15_2_2_14
void FASTCALL call_15_2_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_2_15
#define init_15_2_2_15
void FASTCALL call_15_2_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_3_0
#define init_15_2_3_0
void FASTCALL call_15_2_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_3_1
#define init_15_2_3_1
void FASTCALL call_15_2_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_3_2
#define init_15_2_3_2
void FASTCALL call_15_2_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_3_3
#define init_15_2_3_3
void FASTCALL call_15_2_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_3_4
#define init_15_2_3_4
void FASTCALL call_15_2_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_3_5
#define init_15_2_3_5
void FASTCALL call_15_2_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_3_6
#define init_15_2_3_6
void FASTCALL call_15_2_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_3_7
#define init_15_2_3_7
void FASTCALL call_15_2_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_3_8
#define init_15_2_3_8
void FASTCALL call_15_2_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_3_9
#define init_15_2_3_9
void FASTCALL call_15_2_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_3_10
#define init_15_2_3_10
void FASTCALL call_15_2_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_3_11
#define init_15_2_3_11
void FASTCALL call_15_2_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_3_12
#define init_15_2_3_12
void FASTCALL call_15_2_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_3_13
#define init_15_2_3_13
void FASTCALL call_15_2_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_3_14
#define init_15_2_3_14
void FASTCALL call_15_2_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_3_15
#define init_15_2_3_15
void FASTCALL call_15_2_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_4_0
#define init_15_2_4_0
void FASTCALL call_15_2_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_4_1
#define init_15_2_4_1
void FASTCALL call_15_2_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_4_2
#define init_15_2_4_2
void FASTCALL call_15_2_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_4_3
#define init_15_2_4_3
void FASTCALL call_15_2_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_4_4
#define init_15_2_4_4
void FASTCALL call_15_2_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_4_5
#define init_15_2_4_5
void FASTCALL call_15_2_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_4_6
#define init_15_2_4_6
void FASTCALL call_15_2_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_4_7
#define init_15_2_4_7
void FASTCALL call_15_2_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_4_8
#define init_15_2_4_8
void FASTCALL call_15_2_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_4_9
#define init_15_2_4_9
void FASTCALL call_15_2_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_4_10
#define init_15_2_4_10
void FASTCALL call_15_2_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_4_11
#define init_15_2_4_11
void FASTCALL call_15_2_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_4_12
#define init_15_2_4_12
void FASTCALL call_15_2_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_4_13
#define init_15_2_4_13
void FASTCALL call_15_2_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_4_14
#define init_15_2_4_14
void FASTCALL call_15_2_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_4_15
#define init_15_2_4_15
void FASTCALL call_15_2_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_5_0
#define init_15_2_5_0
void FASTCALL call_15_2_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_5_1
#define init_15_2_5_1
void FASTCALL call_15_2_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_5_2
#define init_15_2_5_2
void FASTCALL call_15_2_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_5_3
#define init_15_2_5_3
void FASTCALL call_15_2_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_5_4
#define init_15_2_5_4
void FASTCALL call_15_2_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_5_5
#define init_15_2_5_5
void FASTCALL call_15_2_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_5_6
#define init_15_2_5_6
void FASTCALL call_15_2_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_5_7
#define init_15_2_5_7
void FASTCALL call_15_2_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_5_8
#define init_15_2_5_8
void FASTCALL call_15_2_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_5_9
#define init_15_2_5_9
void FASTCALL call_15_2_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_5_10
#define init_15_2_5_10
void FASTCALL call_15_2_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_5_11
#define init_15_2_5_11
void FASTCALL call_15_2_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_5_12
#define init_15_2_5_12
void FASTCALL call_15_2_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_5_13
#define init_15_2_5_13
void FASTCALL call_15_2_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_5_14
#define init_15_2_5_14
void FASTCALL call_15_2_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_5_15
#define init_15_2_5_15
void FASTCALL call_15_2_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_6_0
#define init_15_2_6_0
void FASTCALL call_15_2_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_6_1
#define init_15_2_6_1
void FASTCALL call_15_2_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_6_2
#define init_15_2_6_2
void FASTCALL call_15_2_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_6_3
#define init_15_2_6_3
void FASTCALL call_15_2_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_6_4
#define init_15_2_6_4
void FASTCALL call_15_2_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_6_5
#define init_15_2_6_5
void FASTCALL call_15_2_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_6_6
#define init_15_2_6_6
void FASTCALL call_15_2_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_6_7
#define init_15_2_6_7
void FASTCALL call_15_2_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_6_8
#define init_15_2_6_8
void FASTCALL call_15_2_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_6_9
#define init_15_2_6_9
void FASTCALL call_15_2_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_6_10
#define init_15_2_6_10
void FASTCALL call_15_2_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_6_11
#define init_15_2_6_11
void FASTCALL call_15_2_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_6_12
#define init_15_2_6_12
void FASTCALL call_15_2_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_6_13
#define init_15_2_6_13
void FASTCALL call_15_2_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_6_14
#define init_15_2_6_14
void FASTCALL call_15_2_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_6_15
#define init_15_2_6_15
void FASTCALL call_15_2_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_7_0
#define init_15_2_7_0
void FASTCALL call_15_2_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_7_1
#define init_15_2_7_1
void FASTCALL call_15_2_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_7_2
#define init_15_2_7_2
void FASTCALL call_15_2_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_7_3
#define init_15_2_7_3
void FASTCALL call_15_2_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_7_4
#define init_15_2_7_4
void FASTCALL call_15_2_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_7_5
#define init_15_2_7_5
void FASTCALL call_15_2_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_7_6
#define init_15_2_7_6
void FASTCALL call_15_2_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_7_7
#define init_15_2_7_7
void FASTCALL call_15_2_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_7_8
#define init_15_2_7_8
void FASTCALL call_15_2_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_7_9
#define init_15_2_7_9
void FASTCALL call_15_2_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_7_10
#define init_15_2_7_10
void FASTCALL call_15_2_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_7_11
#define init_15_2_7_11
void FASTCALL call_15_2_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_7_12
#define init_15_2_7_12
void FASTCALL call_15_2_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_7_13
#define init_15_2_7_13
void FASTCALL call_15_2_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_7_14
#define init_15_2_7_14
void FASTCALL call_15_2_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_7_15
#define init_15_2_7_15
void FASTCALL call_15_2_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_8_0
#define init_15_2_8_0
void FASTCALL call_15_2_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_8_1
#define init_15_2_8_1
void FASTCALL call_15_2_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_8_2
#define init_15_2_8_2
void FASTCALL call_15_2_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_8_3
#define init_15_2_8_3
void FASTCALL call_15_2_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_8_4
#define init_15_2_8_4
void FASTCALL call_15_2_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_8_5
#define init_15_2_8_5
void FASTCALL call_15_2_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_8_6
#define init_15_2_8_6
void FASTCALL call_15_2_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_8_7
#define init_15_2_8_7
void FASTCALL call_15_2_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_8_8
#define init_15_2_8_8
void FASTCALL call_15_2_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_8_9
#define init_15_2_8_9
void FASTCALL call_15_2_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_8_10
#define init_15_2_8_10
void FASTCALL call_15_2_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_8_11
#define init_15_2_8_11
void FASTCALL call_15_2_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_8_12
#define init_15_2_8_12
void FASTCALL call_15_2_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_8_13
#define init_15_2_8_13
void FASTCALL call_15_2_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_8_14
#define init_15_2_8_14
void FASTCALL call_15_2_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_8_15
#define init_15_2_8_15
void FASTCALL call_15_2_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_9_0
#define init_15_2_9_0
void FASTCALL call_15_2_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_9_1
#define init_15_2_9_1
void FASTCALL call_15_2_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_9_2
#define init_15_2_9_2
void FASTCALL call_15_2_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_9_3
#define init_15_2_9_3
void FASTCALL call_15_2_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_9_4
#define init_15_2_9_4
void FASTCALL call_15_2_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_9_5
#define init_15_2_9_5
void FASTCALL call_15_2_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_9_6
#define init_15_2_9_6
void FASTCALL call_15_2_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_9_7
#define init_15_2_9_7
void FASTCALL call_15_2_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_9_8
#define init_15_2_9_8
void FASTCALL call_15_2_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_9_9
#define init_15_2_9_9
void FASTCALL call_15_2_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_9_10
#define init_15_2_9_10
void FASTCALL call_15_2_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_9_11
#define init_15_2_9_11
void FASTCALL call_15_2_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_9_12
#define init_15_2_9_12
void FASTCALL call_15_2_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_9_13
#define init_15_2_9_13
void FASTCALL call_15_2_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_9_14
#define init_15_2_9_14
void FASTCALL call_15_2_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_9_15
#define init_15_2_9_15
void FASTCALL call_15_2_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_10_0
#define init_15_2_10_0
void FASTCALL call_15_2_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_10_1
#define init_15_2_10_1
void FASTCALL call_15_2_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_10_2
#define init_15_2_10_2
void FASTCALL call_15_2_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_10_3
#define init_15_2_10_3
void FASTCALL call_15_2_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_10_4
#define init_15_2_10_4
void FASTCALL call_15_2_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_10_5
#define init_15_2_10_5
void FASTCALL call_15_2_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_10_6
#define init_15_2_10_6
void FASTCALL call_15_2_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_10_7
#define init_15_2_10_7
void FASTCALL call_15_2_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_10_8
#define init_15_2_10_8
void FASTCALL call_15_2_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_10_9
#define init_15_2_10_9
void FASTCALL call_15_2_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_10_10
#define init_15_2_10_10
void FASTCALL call_15_2_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_10_11
#define init_15_2_10_11
void FASTCALL call_15_2_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_10_12
#define init_15_2_10_12
void FASTCALL call_15_2_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_10_13
#define init_15_2_10_13
void FASTCALL call_15_2_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_10_14
#define init_15_2_10_14
void FASTCALL call_15_2_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_10_15
#define init_15_2_10_15
void FASTCALL call_15_2_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_11_0
#define init_15_2_11_0
void FASTCALL call_15_2_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_11_1
#define init_15_2_11_1
void FASTCALL call_15_2_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_11_2
#define init_15_2_11_2
void FASTCALL call_15_2_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_11_3
#define init_15_2_11_3
void FASTCALL call_15_2_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_11_4
#define init_15_2_11_4
void FASTCALL call_15_2_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_11_5
#define init_15_2_11_5
void FASTCALL call_15_2_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_11_6
#define init_15_2_11_6
void FASTCALL call_15_2_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_11_7
#define init_15_2_11_7
void FASTCALL call_15_2_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_11_8
#define init_15_2_11_8
void FASTCALL call_15_2_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_11_9
#define init_15_2_11_9
void FASTCALL call_15_2_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_11_10
#define init_15_2_11_10
void FASTCALL call_15_2_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_11_11
#define init_15_2_11_11
void FASTCALL call_15_2_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_11_12
#define init_15_2_11_12
void FASTCALL call_15_2_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_11_13
#define init_15_2_11_13
void FASTCALL call_15_2_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_11_14
#define init_15_2_11_14
void FASTCALL call_15_2_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_11_15
#define init_15_2_11_15
void FASTCALL call_15_2_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_12_0
#define init_15_2_12_0
void FASTCALL call_15_2_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_12_1
#define init_15_2_12_1
void FASTCALL call_15_2_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_12_2
#define init_15_2_12_2
void FASTCALL call_15_2_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_12_3
#define init_15_2_12_3
void FASTCALL call_15_2_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_12_4
#define init_15_2_12_4
void FASTCALL call_15_2_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_12_5
#define init_15_2_12_5
void FASTCALL call_15_2_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_12_6
#define init_15_2_12_6
void FASTCALL call_15_2_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_12_7
#define init_15_2_12_7
void FASTCALL call_15_2_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_12_8
#define init_15_2_12_8
void FASTCALL call_15_2_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_12_9
#define init_15_2_12_9
void FASTCALL call_15_2_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_12_10
#define init_15_2_12_10
void FASTCALL call_15_2_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_12_11
#define init_15_2_12_11
void FASTCALL call_15_2_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_12_12
#define init_15_2_12_12
void FASTCALL call_15_2_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_12_13
#define init_15_2_12_13
void FASTCALL call_15_2_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_12_14
#define init_15_2_12_14
void FASTCALL call_15_2_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_12_15
#define init_15_2_12_15
void FASTCALL call_15_2_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_13_0
#define init_15_2_13_0
void FASTCALL call_15_2_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_13_1
#define init_15_2_13_1
void FASTCALL call_15_2_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_13_2
#define init_15_2_13_2
void FASTCALL call_15_2_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_13_3
#define init_15_2_13_3
void FASTCALL call_15_2_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_13_4
#define init_15_2_13_4
void FASTCALL call_15_2_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_13_5
#define init_15_2_13_5
void FASTCALL call_15_2_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_13_6
#define init_15_2_13_6
void FASTCALL call_15_2_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_13_7
#define init_15_2_13_7
void FASTCALL call_15_2_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_13_8
#define init_15_2_13_8
void FASTCALL call_15_2_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_13_9
#define init_15_2_13_9
void FASTCALL call_15_2_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_13_10
#define init_15_2_13_10
void FASTCALL call_15_2_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_13_11
#define init_15_2_13_11
void FASTCALL call_15_2_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_13_12
#define init_15_2_13_12
void FASTCALL call_15_2_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_13_13
#define init_15_2_13_13
void FASTCALL call_15_2_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_13_14
#define init_15_2_13_14
void FASTCALL call_15_2_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_13_15
#define init_15_2_13_15
void FASTCALL call_15_2_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_14_0
#define init_15_2_14_0
void FASTCALL call_15_2_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_14_1
#define init_15_2_14_1
void FASTCALL call_15_2_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_14_2
#define init_15_2_14_2
void FASTCALL call_15_2_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_14_3
#define init_15_2_14_3
void FASTCALL call_15_2_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_14_4
#define init_15_2_14_4
void FASTCALL call_15_2_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_14_5
#define init_15_2_14_5
void FASTCALL call_15_2_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_14_6
#define init_15_2_14_6
void FASTCALL call_15_2_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_14_7
#define init_15_2_14_7
void FASTCALL call_15_2_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_14_8
#define init_15_2_14_8
void FASTCALL call_15_2_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_14_9
#define init_15_2_14_9
void FASTCALL call_15_2_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_14_10
#define init_15_2_14_10
void FASTCALL call_15_2_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_14_11
#define init_15_2_14_11
void FASTCALL call_15_2_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_14_12
#define init_15_2_14_12
void FASTCALL call_15_2_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_14_13
#define init_15_2_14_13
void FASTCALL call_15_2_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_14_14
#define init_15_2_14_14
void FASTCALL call_15_2_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_14_15
#define init_15_2_14_15
void FASTCALL call_15_2_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_15_0
#define init_15_2_15_0
void FASTCALL call_15_2_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_15_1
#define init_15_2_15_1
void FASTCALL call_15_2_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_15_2
#define init_15_2_15_2
void FASTCALL call_15_2_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_15_3
#define init_15_2_15_3
void FASTCALL call_15_2_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_15_4
#define init_15_2_15_4
void FASTCALL call_15_2_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_15_5
#define init_15_2_15_5
void FASTCALL call_15_2_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_15_6
#define init_15_2_15_6
void FASTCALL call_15_2_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_15_7
#define init_15_2_15_7
void FASTCALL call_15_2_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_15_8
#define init_15_2_15_8
void FASTCALL call_15_2_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_15_9
#define init_15_2_15_9
void FASTCALL call_15_2_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_15_10
#define init_15_2_15_10
void FASTCALL call_15_2_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_15_11
#define init_15_2_15_11
void FASTCALL call_15_2_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_15_12
#define init_15_2_15_12
void FASTCALL call_15_2_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_15_13
#define init_15_2_15_13
void FASTCALL call_15_2_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_15_14
#define init_15_2_15_14
void FASTCALL call_15_2_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_2_15_15
#define init_15_2_15_15
void FASTCALL call_15_2_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_0_0
#define init_15_3_0_0
void FASTCALL call_15_3_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_0_1
#define init_15_3_0_1
void FASTCALL call_15_3_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_0_2
#define init_15_3_0_2
void FASTCALL call_15_3_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_0_3
#define init_15_3_0_3
void FASTCALL call_15_3_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_0_4
#define init_15_3_0_4
void FASTCALL call_15_3_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_0_5
#define init_15_3_0_5
void FASTCALL call_15_3_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_0_6
#define init_15_3_0_6
void FASTCALL call_15_3_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_0_7
#define init_15_3_0_7
void FASTCALL call_15_3_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_0_8
#define init_15_3_0_8
void FASTCALL call_15_3_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_0_9
#define init_15_3_0_9
void FASTCALL call_15_3_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_0_10
#define init_15_3_0_10
void FASTCALL call_15_3_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_0_11
#define init_15_3_0_11
void FASTCALL call_15_3_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_0_12
#define init_15_3_0_12
void FASTCALL call_15_3_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_0_13
#define init_15_3_0_13
void FASTCALL call_15_3_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_0_14
#define init_15_3_0_14
void FASTCALL call_15_3_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_0_15
#define init_15_3_0_15
void FASTCALL call_15_3_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_1_0
#define init_15_3_1_0
void FASTCALL call_15_3_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_1_1
#define init_15_3_1_1
void FASTCALL call_15_3_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_1_2
#define init_15_3_1_2
void FASTCALL call_15_3_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_1_3
#define init_15_3_1_3
void FASTCALL call_15_3_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_1_4
#define init_15_3_1_4
void FASTCALL call_15_3_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_1_5
#define init_15_3_1_5
void FASTCALL call_15_3_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_1_6
#define init_15_3_1_6
void FASTCALL call_15_3_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_1_7
#define init_15_3_1_7
void FASTCALL call_15_3_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_1_8
#define init_15_3_1_8
void FASTCALL call_15_3_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_1_9
#define init_15_3_1_9
void FASTCALL call_15_3_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_1_10
#define init_15_3_1_10
void FASTCALL call_15_3_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_1_11
#define init_15_3_1_11
void FASTCALL call_15_3_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_1_12
#define init_15_3_1_12
void FASTCALL call_15_3_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_1_13
#define init_15_3_1_13
void FASTCALL call_15_3_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_1_14
#define init_15_3_1_14
void FASTCALL call_15_3_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_1_15
#define init_15_3_1_15
void FASTCALL call_15_3_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_2_0
#define init_15_3_2_0
void FASTCALL call_15_3_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_2_1
#define init_15_3_2_1
void FASTCALL call_15_3_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_2_2
#define init_15_3_2_2
void FASTCALL call_15_3_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_2_3
#define init_15_3_2_3
void FASTCALL call_15_3_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_2_4
#define init_15_3_2_4
void FASTCALL call_15_3_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_2_5
#define init_15_3_2_5
void FASTCALL call_15_3_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_2_6
#define init_15_3_2_6
void FASTCALL call_15_3_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_2_7
#define init_15_3_2_7
void FASTCALL call_15_3_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_2_8
#define init_15_3_2_8
void FASTCALL call_15_3_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_2_9
#define init_15_3_2_9
void FASTCALL call_15_3_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_2_10
#define init_15_3_2_10
void FASTCALL call_15_3_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_2_11
#define init_15_3_2_11
void FASTCALL call_15_3_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_2_12
#define init_15_3_2_12
void FASTCALL call_15_3_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_2_13
#define init_15_3_2_13
void FASTCALL call_15_3_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_2_14
#define init_15_3_2_14
void FASTCALL call_15_3_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_2_15
#define init_15_3_2_15
void FASTCALL call_15_3_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_3_0
#define init_15_3_3_0
void FASTCALL call_15_3_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_3_1
#define init_15_3_3_1
void FASTCALL call_15_3_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_3_2
#define init_15_3_3_2
void FASTCALL call_15_3_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_3_3
#define init_15_3_3_3
void FASTCALL call_15_3_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_3_4
#define init_15_3_3_4
void FASTCALL call_15_3_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_3_5
#define init_15_3_3_5
void FASTCALL call_15_3_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_3_6
#define init_15_3_3_6
void FASTCALL call_15_3_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_3_7
#define init_15_3_3_7
void FASTCALL call_15_3_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_3_8
#define init_15_3_3_8
void FASTCALL call_15_3_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_3_9
#define init_15_3_3_9
void FASTCALL call_15_3_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_3_10
#define init_15_3_3_10
void FASTCALL call_15_3_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_3_11
#define init_15_3_3_11
void FASTCALL call_15_3_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_3_12
#define init_15_3_3_12
void FASTCALL call_15_3_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_3_13
#define init_15_3_3_13
void FASTCALL call_15_3_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_3_14
#define init_15_3_3_14
void FASTCALL call_15_3_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_3_15
#define init_15_3_3_15
void FASTCALL call_15_3_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_4_0
#define init_15_3_4_0
void FASTCALL call_15_3_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_4_1
#define init_15_3_4_1
void FASTCALL call_15_3_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_4_2
#define init_15_3_4_2
void FASTCALL call_15_3_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_4_3
#define init_15_3_4_3
void FASTCALL call_15_3_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_4_4
#define init_15_3_4_4
void FASTCALL call_15_3_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_4_5
#define init_15_3_4_5
void FASTCALL call_15_3_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_4_6
#define init_15_3_4_6
void FASTCALL call_15_3_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_4_7
#define init_15_3_4_7
void FASTCALL call_15_3_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_4_8
#define init_15_3_4_8
void FASTCALL call_15_3_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_4_9
#define init_15_3_4_9
void FASTCALL call_15_3_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_4_10
#define init_15_3_4_10
void FASTCALL call_15_3_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_4_11
#define init_15_3_4_11
void FASTCALL call_15_3_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_4_12
#define init_15_3_4_12
void FASTCALL call_15_3_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_4_13
#define init_15_3_4_13
void FASTCALL call_15_3_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_4_14
#define init_15_3_4_14
void FASTCALL call_15_3_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_4_15
#define init_15_3_4_15
void FASTCALL call_15_3_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_5_0
#define init_15_3_5_0
void FASTCALL call_15_3_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_5_1
#define init_15_3_5_1
void FASTCALL call_15_3_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_5_2
#define init_15_3_5_2
void FASTCALL call_15_3_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_5_3
#define init_15_3_5_3
void FASTCALL call_15_3_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_5_4
#define init_15_3_5_4
void FASTCALL call_15_3_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_5_5
#define init_15_3_5_5
void FASTCALL call_15_3_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_5_6
#define init_15_3_5_6
void FASTCALL call_15_3_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_5_7
#define init_15_3_5_7
void FASTCALL call_15_3_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_5_8
#define init_15_3_5_8
void FASTCALL call_15_3_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_5_9
#define init_15_3_5_9
void FASTCALL call_15_3_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_5_10
#define init_15_3_5_10
void FASTCALL call_15_3_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_5_11
#define init_15_3_5_11
void FASTCALL call_15_3_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_5_12
#define init_15_3_5_12
void FASTCALL call_15_3_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_5_13
#define init_15_3_5_13
void FASTCALL call_15_3_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_5_14
#define init_15_3_5_14
void FASTCALL call_15_3_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_5_15
#define init_15_3_5_15
void FASTCALL call_15_3_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_6_0
#define init_15_3_6_0
void FASTCALL call_15_3_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_6_1
#define init_15_3_6_1
void FASTCALL call_15_3_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_6_2
#define init_15_3_6_2
void FASTCALL call_15_3_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_6_3
#define init_15_3_6_3
void FASTCALL call_15_3_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_6_4
#define init_15_3_6_4
void FASTCALL call_15_3_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_6_5
#define init_15_3_6_5
void FASTCALL call_15_3_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_6_6
#define init_15_3_6_6
void FASTCALL call_15_3_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_6_7
#define init_15_3_6_7
void FASTCALL call_15_3_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_6_8
#define init_15_3_6_8
void FASTCALL call_15_3_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_6_9
#define init_15_3_6_9
void FASTCALL call_15_3_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_6_10
#define init_15_3_6_10
void FASTCALL call_15_3_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_6_11
#define init_15_3_6_11
void FASTCALL call_15_3_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_6_12
#define init_15_3_6_12
void FASTCALL call_15_3_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_6_13
#define init_15_3_6_13
void FASTCALL call_15_3_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_6_14
#define init_15_3_6_14
void FASTCALL call_15_3_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_6_15
#define init_15_3_6_15
void FASTCALL call_15_3_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_7_0
#define init_15_3_7_0
void FASTCALL call_15_3_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_7_1
#define init_15_3_7_1
void FASTCALL call_15_3_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_7_2
#define init_15_3_7_2
void FASTCALL call_15_3_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_7_3
#define init_15_3_7_3
void FASTCALL call_15_3_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_7_4
#define init_15_3_7_4
void FASTCALL call_15_3_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_7_5
#define init_15_3_7_5
void FASTCALL call_15_3_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_7_6
#define init_15_3_7_6
void FASTCALL call_15_3_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_7_7
#define init_15_3_7_7
void FASTCALL call_15_3_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_7_8
#define init_15_3_7_8
void FASTCALL call_15_3_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_7_9
#define init_15_3_7_9
void FASTCALL call_15_3_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_7_10
#define init_15_3_7_10
void FASTCALL call_15_3_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_7_11
#define init_15_3_7_11
void FASTCALL call_15_3_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_7_12
#define init_15_3_7_12
void FASTCALL call_15_3_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_7_13
#define init_15_3_7_13
void FASTCALL call_15_3_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_7_14
#define init_15_3_7_14
void FASTCALL call_15_3_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_7_15
#define init_15_3_7_15
void FASTCALL call_15_3_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_8_0
#define init_15_3_8_0
void FASTCALL call_15_3_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_8_1
#define init_15_3_8_1
void FASTCALL call_15_3_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_8_2
#define init_15_3_8_2
void FASTCALL call_15_3_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_8_3
#define init_15_3_8_3
void FASTCALL call_15_3_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_8_4
#define init_15_3_8_4
void FASTCALL call_15_3_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_8_5
#define init_15_3_8_5
void FASTCALL call_15_3_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_8_6
#define init_15_3_8_6
void FASTCALL call_15_3_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_8_7
#define init_15_3_8_7
void FASTCALL call_15_3_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_8_8
#define init_15_3_8_8
void FASTCALL call_15_3_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_8_9
#define init_15_3_8_9
void FASTCALL call_15_3_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_8_10
#define init_15_3_8_10
void FASTCALL call_15_3_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_8_11
#define init_15_3_8_11
void FASTCALL call_15_3_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_8_12
#define init_15_3_8_12
void FASTCALL call_15_3_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_8_13
#define init_15_3_8_13
void FASTCALL call_15_3_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_8_14
#define init_15_3_8_14
void FASTCALL call_15_3_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_8_15
#define init_15_3_8_15
void FASTCALL call_15_3_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_9_0
#define init_15_3_9_0
void FASTCALL call_15_3_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_9_1
#define init_15_3_9_1
void FASTCALL call_15_3_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_9_2
#define init_15_3_9_2
void FASTCALL call_15_3_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_9_3
#define init_15_3_9_3
void FASTCALL call_15_3_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_9_4
#define init_15_3_9_4
void FASTCALL call_15_3_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_9_5
#define init_15_3_9_5
void FASTCALL call_15_3_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_9_6
#define init_15_3_9_6
void FASTCALL call_15_3_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_9_7
#define init_15_3_9_7
void FASTCALL call_15_3_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_9_8
#define init_15_3_9_8
void FASTCALL call_15_3_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_9_9
#define init_15_3_9_9
void FASTCALL call_15_3_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_9_10
#define init_15_3_9_10
void FASTCALL call_15_3_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_9_11
#define init_15_3_9_11
void FASTCALL call_15_3_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_9_12
#define init_15_3_9_12
void FASTCALL call_15_3_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_9_13
#define init_15_3_9_13
void FASTCALL call_15_3_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_9_14
#define init_15_3_9_14
void FASTCALL call_15_3_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_9_15
#define init_15_3_9_15
void FASTCALL call_15_3_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_10_0
#define init_15_3_10_0
void FASTCALL call_15_3_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_10_1
#define init_15_3_10_1
void FASTCALL call_15_3_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_10_2
#define init_15_3_10_2
void FASTCALL call_15_3_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_10_3
#define init_15_3_10_3
void FASTCALL call_15_3_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_10_4
#define init_15_3_10_4
void FASTCALL call_15_3_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_10_5
#define init_15_3_10_5
void FASTCALL call_15_3_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_10_6
#define init_15_3_10_6
void FASTCALL call_15_3_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_10_7
#define init_15_3_10_7
void FASTCALL call_15_3_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_10_8
#define init_15_3_10_8
void FASTCALL call_15_3_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_10_9
#define init_15_3_10_9
void FASTCALL call_15_3_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_10_10
#define init_15_3_10_10
void FASTCALL call_15_3_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_10_11
#define init_15_3_10_11
void FASTCALL call_15_3_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_10_12
#define init_15_3_10_12
void FASTCALL call_15_3_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_10_13
#define init_15_3_10_13
void FASTCALL call_15_3_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_10_14
#define init_15_3_10_14
void FASTCALL call_15_3_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_10_15
#define init_15_3_10_15
void FASTCALL call_15_3_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_11_0
#define init_15_3_11_0
void FASTCALL call_15_3_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_11_1
#define init_15_3_11_1
void FASTCALL call_15_3_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_11_2
#define init_15_3_11_2
void FASTCALL call_15_3_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_11_3
#define init_15_3_11_3
void FASTCALL call_15_3_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_11_4
#define init_15_3_11_4
void FASTCALL call_15_3_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_11_5
#define init_15_3_11_5
void FASTCALL call_15_3_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_11_6
#define init_15_3_11_6
void FASTCALL call_15_3_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_11_7
#define init_15_3_11_7
void FASTCALL call_15_3_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_11_8
#define init_15_3_11_8
void FASTCALL call_15_3_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_11_9
#define init_15_3_11_9
void FASTCALL call_15_3_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_11_10
#define init_15_3_11_10
void FASTCALL call_15_3_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_11_11
#define init_15_3_11_11
void FASTCALL call_15_3_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_11_12
#define init_15_3_11_12
void FASTCALL call_15_3_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_11_13
#define init_15_3_11_13
void FASTCALL call_15_3_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_11_14
#define init_15_3_11_14
void FASTCALL call_15_3_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_11_15
#define init_15_3_11_15
void FASTCALL call_15_3_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_12_0
#define init_15_3_12_0
void FASTCALL call_15_3_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_12_1
#define init_15_3_12_1
void FASTCALL call_15_3_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_12_2
#define init_15_3_12_2
void FASTCALL call_15_3_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_12_3
#define init_15_3_12_3
void FASTCALL call_15_3_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_12_4
#define init_15_3_12_4
void FASTCALL call_15_3_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_12_5
#define init_15_3_12_5
void FASTCALL call_15_3_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_12_6
#define init_15_3_12_6
void FASTCALL call_15_3_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_12_7
#define init_15_3_12_7
void FASTCALL call_15_3_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_12_8
#define init_15_3_12_8
void FASTCALL call_15_3_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_12_9
#define init_15_3_12_9
void FASTCALL call_15_3_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_12_10
#define init_15_3_12_10
void FASTCALL call_15_3_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_12_11
#define init_15_3_12_11
void FASTCALL call_15_3_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_12_12
#define init_15_3_12_12
void FASTCALL call_15_3_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_12_13
#define init_15_3_12_13
void FASTCALL call_15_3_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_12_14
#define init_15_3_12_14
void FASTCALL call_15_3_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_12_15
#define init_15_3_12_15
void FASTCALL call_15_3_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_13_0
#define init_15_3_13_0
void FASTCALL call_15_3_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_13_1
#define init_15_3_13_1
void FASTCALL call_15_3_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_13_2
#define init_15_3_13_2
void FASTCALL call_15_3_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_13_3
#define init_15_3_13_3
void FASTCALL call_15_3_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_13_4
#define init_15_3_13_4
void FASTCALL call_15_3_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_13_5
#define init_15_3_13_5
void FASTCALL call_15_3_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_13_6
#define init_15_3_13_6
void FASTCALL call_15_3_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_13_7
#define init_15_3_13_7
void FASTCALL call_15_3_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_13_8
#define init_15_3_13_8
void FASTCALL call_15_3_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_13_9
#define init_15_3_13_9
void FASTCALL call_15_3_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_13_10
#define init_15_3_13_10
void FASTCALL call_15_3_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_13_11
#define init_15_3_13_11
void FASTCALL call_15_3_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_13_12
#define init_15_3_13_12
void FASTCALL call_15_3_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_13_13
#define init_15_3_13_13
void FASTCALL call_15_3_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_13_14
#define init_15_3_13_14
void FASTCALL call_15_3_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_13_15
#define init_15_3_13_15
void FASTCALL call_15_3_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_14_0
#define init_15_3_14_0
void FASTCALL call_15_3_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_14_1
#define init_15_3_14_1
void FASTCALL call_15_3_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_14_2
#define init_15_3_14_2
void FASTCALL call_15_3_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_14_3
#define init_15_3_14_3
void FASTCALL call_15_3_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_14_4
#define init_15_3_14_4
void FASTCALL call_15_3_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_14_5
#define init_15_3_14_5
void FASTCALL call_15_3_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_14_6
#define init_15_3_14_6
void FASTCALL call_15_3_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_14_7
#define init_15_3_14_7
void FASTCALL call_15_3_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_14_8
#define init_15_3_14_8
void FASTCALL call_15_3_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_14_9
#define init_15_3_14_9
void FASTCALL call_15_3_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_14_10
#define init_15_3_14_10
void FASTCALL call_15_3_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_14_11
#define init_15_3_14_11
void FASTCALL call_15_3_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_14_12
#define init_15_3_14_12
void FASTCALL call_15_3_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_14_13
#define init_15_3_14_13
void FASTCALL call_15_3_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_14_14
#define init_15_3_14_14
void FASTCALL call_15_3_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_14_15
#define init_15_3_14_15
void FASTCALL call_15_3_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_15_0
#define init_15_3_15_0
void FASTCALL call_15_3_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_15_1
#define init_15_3_15_1
void FASTCALL call_15_3_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_15_2
#define init_15_3_15_2
void FASTCALL call_15_3_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_15_3
#define init_15_3_15_3
void FASTCALL call_15_3_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_15_4
#define init_15_3_15_4
void FASTCALL call_15_3_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_15_5
#define init_15_3_15_5
void FASTCALL call_15_3_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_15_6
#define init_15_3_15_6
void FASTCALL call_15_3_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_15_7
#define init_15_3_15_7
void FASTCALL call_15_3_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_15_8
#define init_15_3_15_8
void FASTCALL call_15_3_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_15_9
#define init_15_3_15_9
void FASTCALL call_15_3_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_15_10
#define init_15_3_15_10
void FASTCALL call_15_3_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_15_11
#define init_15_3_15_11
void FASTCALL call_15_3_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_15_12
#define init_15_3_15_12
void FASTCALL call_15_3_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_15_13
#define init_15_3_15_13
void FASTCALL call_15_3_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_15_14
#define init_15_3_15_14
void FASTCALL call_15_3_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_3_15_15
#define init_15_3_15_15
void FASTCALL call_15_3_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_0_0
#define init_15_4_0_0
void FASTCALL call_15_4_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_0_1
#define init_15_4_0_1
void FASTCALL call_15_4_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_0_2
#define init_15_4_0_2
void FASTCALL call_15_4_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_0_3
#define init_15_4_0_3
void FASTCALL call_15_4_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_0_4
#define init_15_4_0_4
void FASTCALL call_15_4_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_0_5
#define init_15_4_0_5
void FASTCALL call_15_4_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_0_6
#define init_15_4_0_6
void FASTCALL call_15_4_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_0_7
#define init_15_4_0_7
void FASTCALL call_15_4_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_0_8
#define init_15_4_0_8
void FASTCALL call_15_4_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_0_9
#define init_15_4_0_9
void FASTCALL call_15_4_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_0_10
#define init_15_4_0_10
void FASTCALL call_15_4_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_0_11
#define init_15_4_0_11
void FASTCALL call_15_4_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_0_12
#define init_15_4_0_12
void FASTCALL call_15_4_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_0_13
#define init_15_4_0_13
void FASTCALL call_15_4_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_0_14
#define init_15_4_0_14
void FASTCALL call_15_4_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_0_15
#define init_15_4_0_15
void FASTCALL call_15_4_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_1_0
#define init_15_4_1_0
void FASTCALL call_15_4_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_1_1
#define init_15_4_1_1
void FASTCALL call_15_4_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_1_2
#define init_15_4_1_2
void FASTCALL call_15_4_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_1_3
#define init_15_4_1_3
void FASTCALL call_15_4_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_1_4
#define init_15_4_1_4
void FASTCALL call_15_4_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_1_5
#define init_15_4_1_5
void FASTCALL call_15_4_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_1_6
#define init_15_4_1_6
void FASTCALL call_15_4_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_1_7
#define init_15_4_1_7
void FASTCALL call_15_4_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_1_8
#define init_15_4_1_8
void FASTCALL call_15_4_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_1_9
#define init_15_4_1_9
void FASTCALL call_15_4_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_1_10
#define init_15_4_1_10
void FASTCALL call_15_4_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_1_11
#define init_15_4_1_11
void FASTCALL call_15_4_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_1_12
#define init_15_4_1_12
void FASTCALL call_15_4_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_1_13
#define init_15_4_1_13
void FASTCALL call_15_4_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_1_14
#define init_15_4_1_14
void FASTCALL call_15_4_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_1_15
#define init_15_4_1_15
void FASTCALL call_15_4_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_2_0
#define init_15_4_2_0
void FASTCALL call_15_4_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_2_1
#define init_15_4_2_1
void FASTCALL call_15_4_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_2_2
#define init_15_4_2_2
void FASTCALL call_15_4_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_2_3
#define init_15_4_2_3
void FASTCALL call_15_4_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_2_4
#define init_15_4_2_4
void FASTCALL call_15_4_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_2_5
#define init_15_4_2_5
void FASTCALL call_15_4_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_2_6
#define init_15_4_2_6
void FASTCALL call_15_4_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_2_7
#define init_15_4_2_7
void FASTCALL call_15_4_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_2_8
#define init_15_4_2_8
void FASTCALL call_15_4_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_2_9
#define init_15_4_2_9
void FASTCALL call_15_4_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_2_10
#define init_15_4_2_10
void FASTCALL call_15_4_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_2_11
#define init_15_4_2_11
void FASTCALL call_15_4_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_2_12
#define init_15_4_2_12
void FASTCALL call_15_4_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_2_13
#define init_15_4_2_13
void FASTCALL call_15_4_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_2_14
#define init_15_4_2_14
void FASTCALL call_15_4_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_2_15
#define init_15_4_2_15
void FASTCALL call_15_4_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_3_0
#define init_15_4_3_0
void FASTCALL call_15_4_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_3_1
#define init_15_4_3_1
void FASTCALL call_15_4_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_3_2
#define init_15_4_3_2
void FASTCALL call_15_4_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_3_3
#define init_15_4_3_3
void FASTCALL call_15_4_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_3_4
#define init_15_4_3_4
void FASTCALL call_15_4_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_3_5
#define init_15_4_3_5
void FASTCALL call_15_4_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_3_6
#define init_15_4_3_6
void FASTCALL call_15_4_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_3_7
#define init_15_4_3_7
void FASTCALL call_15_4_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_3_8
#define init_15_4_3_8
void FASTCALL call_15_4_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_3_9
#define init_15_4_3_9
void FASTCALL call_15_4_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_3_10
#define init_15_4_3_10
void FASTCALL call_15_4_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_3_11
#define init_15_4_3_11
void FASTCALL call_15_4_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_3_12
#define init_15_4_3_12
void FASTCALL call_15_4_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_3_13
#define init_15_4_3_13
void FASTCALL call_15_4_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_3_14
#define init_15_4_3_14
void FASTCALL call_15_4_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_3_15
#define init_15_4_3_15
void FASTCALL call_15_4_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_4_0
#define init_15_4_4_0
void FASTCALL call_15_4_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_4_1
#define init_15_4_4_1
void FASTCALL call_15_4_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_4_2
#define init_15_4_4_2
void FASTCALL call_15_4_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_4_3
#define init_15_4_4_3
void FASTCALL call_15_4_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_4_4
#define init_15_4_4_4
void FASTCALL call_15_4_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_4_5
#define init_15_4_4_5
void FASTCALL call_15_4_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_4_6
#define init_15_4_4_6
void FASTCALL call_15_4_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_4_7
#define init_15_4_4_7
void FASTCALL call_15_4_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_4_8
#define init_15_4_4_8
void FASTCALL call_15_4_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_4_9
#define init_15_4_4_9
void FASTCALL call_15_4_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_4_10
#define init_15_4_4_10
void FASTCALL call_15_4_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_4_11
#define init_15_4_4_11
void FASTCALL call_15_4_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_4_12
#define init_15_4_4_12
void FASTCALL call_15_4_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_4_13
#define init_15_4_4_13
void FASTCALL call_15_4_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_4_14
#define init_15_4_4_14
void FASTCALL call_15_4_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_4_15
#define init_15_4_4_15
void FASTCALL call_15_4_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_5_0
#define init_15_4_5_0
void FASTCALL call_15_4_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_5_1
#define init_15_4_5_1
void FASTCALL call_15_4_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_5_2
#define init_15_4_5_2
void FASTCALL call_15_4_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_5_3
#define init_15_4_5_3
void FASTCALL call_15_4_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_5_4
#define init_15_4_5_4
void FASTCALL call_15_4_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_5_5
#define init_15_4_5_5
void FASTCALL call_15_4_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_5_6
#define init_15_4_5_6
void FASTCALL call_15_4_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_5_7
#define init_15_4_5_7
void FASTCALL call_15_4_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_5_8
#define init_15_4_5_8
void FASTCALL call_15_4_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_5_9
#define init_15_4_5_9
void FASTCALL call_15_4_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_5_10
#define init_15_4_5_10
void FASTCALL call_15_4_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_5_11
#define init_15_4_5_11
void FASTCALL call_15_4_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_5_12
#define init_15_4_5_12
void FASTCALL call_15_4_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_5_13
#define init_15_4_5_13
void FASTCALL call_15_4_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_5_14
#define init_15_4_5_14
void FASTCALL call_15_4_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_5_15
#define init_15_4_5_15
void FASTCALL call_15_4_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_6_0
#define init_15_4_6_0
void FASTCALL call_15_4_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_6_1
#define init_15_4_6_1
void FASTCALL call_15_4_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_6_2
#define init_15_4_6_2
void FASTCALL call_15_4_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_6_3
#define init_15_4_6_3
void FASTCALL call_15_4_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_6_4
#define init_15_4_6_4
void FASTCALL call_15_4_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_6_5
#define init_15_4_6_5
void FASTCALL call_15_4_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_6_6
#define init_15_4_6_6
void FASTCALL call_15_4_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_6_7
#define init_15_4_6_7
void FASTCALL call_15_4_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_6_8
#define init_15_4_6_8
void FASTCALL call_15_4_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_6_9
#define init_15_4_6_9
void FASTCALL call_15_4_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_6_10
#define init_15_4_6_10
void FASTCALL call_15_4_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_6_11
#define init_15_4_6_11
void FASTCALL call_15_4_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_6_12
#define init_15_4_6_12
void FASTCALL call_15_4_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_6_13
#define init_15_4_6_13
void FASTCALL call_15_4_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_6_14
#define init_15_4_6_14
void FASTCALL call_15_4_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_6_15
#define init_15_4_6_15
void FASTCALL call_15_4_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_7_0
#define init_15_4_7_0
void FASTCALL call_15_4_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_7_1
#define init_15_4_7_1
void FASTCALL call_15_4_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_7_2
#define init_15_4_7_2
void FASTCALL call_15_4_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_7_3
#define init_15_4_7_3
void FASTCALL call_15_4_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_7_4
#define init_15_4_7_4
void FASTCALL call_15_4_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_7_5
#define init_15_4_7_5
void FASTCALL call_15_4_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_7_6
#define init_15_4_7_6
void FASTCALL call_15_4_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_7_7
#define init_15_4_7_7
void FASTCALL call_15_4_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_7_8
#define init_15_4_7_8
void FASTCALL call_15_4_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_7_9
#define init_15_4_7_9
void FASTCALL call_15_4_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_7_10
#define init_15_4_7_10
void FASTCALL call_15_4_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_7_11
#define init_15_4_7_11
void FASTCALL call_15_4_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_7_12
#define init_15_4_7_12
void FASTCALL call_15_4_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_7_13
#define init_15_4_7_13
void FASTCALL call_15_4_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_7_14
#define init_15_4_7_14
void FASTCALL call_15_4_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_7_15
#define init_15_4_7_15
void FASTCALL call_15_4_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_8_0
#define init_15_4_8_0
void FASTCALL call_15_4_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_8_1
#define init_15_4_8_1
void FASTCALL call_15_4_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_8_2
#define init_15_4_8_2
void FASTCALL call_15_4_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_8_3
#define init_15_4_8_3
void FASTCALL call_15_4_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_8_4
#define init_15_4_8_4
void FASTCALL call_15_4_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_8_5
#define init_15_4_8_5
void FASTCALL call_15_4_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_8_6
#define init_15_4_8_6
void FASTCALL call_15_4_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_8_7
#define init_15_4_8_7
void FASTCALL call_15_4_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_8_8
#define init_15_4_8_8
void FASTCALL call_15_4_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_8_9
#define init_15_4_8_9
void FASTCALL call_15_4_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_8_10
#define init_15_4_8_10
void FASTCALL call_15_4_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_8_11
#define init_15_4_8_11
void FASTCALL call_15_4_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_8_12
#define init_15_4_8_12
void FASTCALL call_15_4_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_8_13
#define init_15_4_8_13
void FASTCALL call_15_4_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_8_14
#define init_15_4_8_14
void FASTCALL call_15_4_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_8_15
#define init_15_4_8_15
void FASTCALL call_15_4_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_9_0
#define init_15_4_9_0
void FASTCALL call_15_4_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_9_1
#define init_15_4_9_1
void FASTCALL call_15_4_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_9_2
#define init_15_4_9_2
void FASTCALL call_15_4_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_9_3
#define init_15_4_9_3
void FASTCALL call_15_4_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_9_4
#define init_15_4_9_4
void FASTCALL call_15_4_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_9_5
#define init_15_4_9_5
void FASTCALL call_15_4_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_9_6
#define init_15_4_9_6
void FASTCALL call_15_4_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_9_7
#define init_15_4_9_7
void FASTCALL call_15_4_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_9_8
#define init_15_4_9_8
void FASTCALL call_15_4_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_9_9
#define init_15_4_9_9
void FASTCALL call_15_4_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_9_10
#define init_15_4_9_10
void FASTCALL call_15_4_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_9_11
#define init_15_4_9_11
void FASTCALL call_15_4_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_9_12
#define init_15_4_9_12
void FASTCALL call_15_4_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_9_13
#define init_15_4_9_13
void FASTCALL call_15_4_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_9_14
#define init_15_4_9_14
void FASTCALL call_15_4_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_9_15
#define init_15_4_9_15
void FASTCALL call_15_4_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_10_0
#define init_15_4_10_0
void FASTCALL call_15_4_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_10_1
#define init_15_4_10_1
void FASTCALL call_15_4_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_10_2
#define init_15_4_10_2
void FASTCALL call_15_4_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_10_3
#define init_15_4_10_3
void FASTCALL call_15_4_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_10_4
#define init_15_4_10_4
void FASTCALL call_15_4_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_10_5
#define init_15_4_10_5
void FASTCALL call_15_4_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_10_6
#define init_15_4_10_6
void FASTCALL call_15_4_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_10_7
#define init_15_4_10_7
void FASTCALL call_15_4_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_10_8
#define init_15_4_10_8
void FASTCALL call_15_4_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_10_9
#define init_15_4_10_9
void FASTCALL call_15_4_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_10_10
#define init_15_4_10_10
void FASTCALL call_15_4_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_10_11
#define init_15_4_10_11
void FASTCALL call_15_4_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_10_12
#define init_15_4_10_12
void FASTCALL call_15_4_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_10_13
#define init_15_4_10_13
void FASTCALL call_15_4_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_10_14
#define init_15_4_10_14
void FASTCALL call_15_4_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_10_15
#define init_15_4_10_15
void FASTCALL call_15_4_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_11_0
#define init_15_4_11_0
void FASTCALL call_15_4_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_11_1
#define init_15_4_11_1
void FASTCALL call_15_4_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_11_2
#define init_15_4_11_2
void FASTCALL call_15_4_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_11_3
#define init_15_4_11_3
void FASTCALL call_15_4_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_11_4
#define init_15_4_11_4
void FASTCALL call_15_4_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_11_5
#define init_15_4_11_5
void FASTCALL call_15_4_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_11_6
#define init_15_4_11_6
void FASTCALL call_15_4_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_11_7
#define init_15_4_11_7
void FASTCALL call_15_4_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_11_8
#define init_15_4_11_8
void FASTCALL call_15_4_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_11_9
#define init_15_4_11_9
void FASTCALL call_15_4_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_11_10
#define init_15_4_11_10
void FASTCALL call_15_4_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_11_11
#define init_15_4_11_11
void FASTCALL call_15_4_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_11_12
#define init_15_4_11_12
void FASTCALL call_15_4_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_11_13
#define init_15_4_11_13
void FASTCALL call_15_4_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_11_14
#define init_15_4_11_14
void FASTCALL call_15_4_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_11_15
#define init_15_4_11_15
void FASTCALL call_15_4_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_12_0
#define init_15_4_12_0
void FASTCALL call_15_4_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_12_1
#define init_15_4_12_1
void FASTCALL call_15_4_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_12_2
#define init_15_4_12_2
void FASTCALL call_15_4_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_12_3
#define init_15_4_12_3
void FASTCALL call_15_4_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_12_4
#define init_15_4_12_4
void FASTCALL call_15_4_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_12_5
#define init_15_4_12_5
void FASTCALL call_15_4_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_12_6
#define init_15_4_12_6
void FASTCALL call_15_4_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_12_7
#define init_15_4_12_7
void FASTCALL call_15_4_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_12_8
#define init_15_4_12_8
void FASTCALL call_15_4_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_12_9
#define init_15_4_12_9
void FASTCALL call_15_4_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_12_10
#define init_15_4_12_10
void FASTCALL call_15_4_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_12_11
#define init_15_4_12_11
void FASTCALL call_15_4_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_12_12
#define init_15_4_12_12
void FASTCALL call_15_4_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_12_13
#define init_15_4_12_13
void FASTCALL call_15_4_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_12_14
#define init_15_4_12_14
void FASTCALL call_15_4_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_12_15
#define init_15_4_12_15
void FASTCALL call_15_4_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_13_0
#define init_15_4_13_0
void FASTCALL call_15_4_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_13_1
#define init_15_4_13_1
void FASTCALL call_15_4_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_13_2
#define init_15_4_13_2
void FASTCALL call_15_4_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_13_3
#define init_15_4_13_3
void FASTCALL call_15_4_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_13_4
#define init_15_4_13_4
void FASTCALL call_15_4_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_13_5
#define init_15_4_13_5
void FASTCALL call_15_4_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_13_6
#define init_15_4_13_6
void FASTCALL call_15_4_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_13_7
#define init_15_4_13_7
void FASTCALL call_15_4_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_13_8
#define init_15_4_13_8
void FASTCALL call_15_4_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_13_9
#define init_15_4_13_9
void FASTCALL call_15_4_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_13_10
#define init_15_4_13_10
void FASTCALL call_15_4_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_13_11
#define init_15_4_13_11
void FASTCALL call_15_4_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_13_12
#define init_15_4_13_12
void FASTCALL call_15_4_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_13_13
#define init_15_4_13_13
void FASTCALL call_15_4_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_13_14
#define init_15_4_13_14
void FASTCALL call_15_4_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_13_15
#define init_15_4_13_15
void FASTCALL call_15_4_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_14_0
#define init_15_4_14_0
void FASTCALL call_15_4_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_14_1
#define init_15_4_14_1
void FASTCALL call_15_4_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_14_2
#define init_15_4_14_2
void FASTCALL call_15_4_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_14_3
#define init_15_4_14_3
void FASTCALL call_15_4_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_14_4
#define init_15_4_14_4
void FASTCALL call_15_4_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_14_5
#define init_15_4_14_5
void FASTCALL call_15_4_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_14_6
#define init_15_4_14_6
void FASTCALL call_15_4_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_14_7
#define init_15_4_14_7
void FASTCALL call_15_4_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_14_8
#define init_15_4_14_8
void FASTCALL call_15_4_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_14_9
#define init_15_4_14_9
void FASTCALL call_15_4_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_14_10
#define init_15_4_14_10
void FASTCALL call_15_4_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_14_11
#define init_15_4_14_11
void FASTCALL call_15_4_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_14_12
#define init_15_4_14_12
void FASTCALL call_15_4_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_14_13
#define init_15_4_14_13
void FASTCALL call_15_4_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_14_14
#define init_15_4_14_14
void FASTCALL call_15_4_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_14_15
#define init_15_4_14_15
void FASTCALL call_15_4_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_15_0
#define init_15_4_15_0
void FASTCALL call_15_4_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_15_1
#define init_15_4_15_1
void FASTCALL call_15_4_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_15_2
#define init_15_4_15_2
void FASTCALL call_15_4_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_15_3
#define init_15_4_15_3
void FASTCALL call_15_4_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_15_4
#define init_15_4_15_4
void FASTCALL call_15_4_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_15_5
#define init_15_4_15_5
void FASTCALL call_15_4_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_15_6
#define init_15_4_15_6
void FASTCALL call_15_4_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_15_7
#define init_15_4_15_7
void FASTCALL call_15_4_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_15_8
#define init_15_4_15_8
void FASTCALL call_15_4_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_15_9
#define init_15_4_15_9
void FASTCALL call_15_4_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_15_10
#define init_15_4_15_10
void FASTCALL call_15_4_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_15_11
#define init_15_4_15_11
void FASTCALL call_15_4_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_15_12
#define init_15_4_15_12
void FASTCALL call_15_4_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_15_13
#define init_15_4_15_13
void FASTCALL call_15_4_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_15_14
#define init_15_4_15_14
void FASTCALL call_15_4_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_4_15_15
#define init_15_4_15_15
void FASTCALL call_15_4_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_0_0
#define init_15_5_0_0
void FASTCALL call_15_5_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_0_1
#define init_15_5_0_1
void FASTCALL call_15_5_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_0_2
#define init_15_5_0_2
void FASTCALL call_15_5_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_0_3
#define init_15_5_0_3
void FASTCALL call_15_5_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_0_4
#define init_15_5_0_4
void FASTCALL call_15_5_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_0_5
#define init_15_5_0_5
void FASTCALL call_15_5_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_0_6
#define init_15_5_0_6
void FASTCALL call_15_5_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_0_7
#define init_15_5_0_7
void FASTCALL call_15_5_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_0_8
#define init_15_5_0_8
void FASTCALL call_15_5_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_0_9
#define init_15_5_0_9
void FASTCALL call_15_5_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_0_10
#define init_15_5_0_10
void FASTCALL call_15_5_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_0_11
#define init_15_5_0_11
void FASTCALL call_15_5_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_0_12
#define init_15_5_0_12
void FASTCALL call_15_5_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_0_13
#define init_15_5_0_13
void FASTCALL call_15_5_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_0_14
#define init_15_5_0_14
void FASTCALL call_15_5_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_0_15
#define init_15_5_0_15
void FASTCALL call_15_5_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_1_0
#define init_15_5_1_0
void FASTCALL call_15_5_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_1_1
#define init_15_5_1_1
void FASTCALL call_15_5_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_1_2
#define init_15_5_1_2
void FASTCALL call_15_5_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_1_3
#define init_15_5_1_3
void FASTCALL call_15_5_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_1_4
#define init_15_5_1_4
void FASTCALL call_15_5_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_1_5
#define init_15_5_1_5
void FASTCALL call_15_5_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_1_6
#define init_15_5_1_6
void FASTCALL call_15_5_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_1_7
#define init_15_5_1_7
void FASTCALL call_15_5_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_1_8
#define init_15_5_1_8
void FASTCALL call_15_5_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_1_9
#define init_15_5_1_9
void FASTCALL call_15_5_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_1_10
#define init_15_5_1_10
void FASTCALL call_15_5_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_1_11
#define init_15_5_1_11
void FASTCALL call_15_5_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_1_12
#define init_15_5_1_12
void FASTCALL call_15_5_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_1_13
#define init_15_5_1_13
void FASTCALL call_15_5_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_1_14
#define init_15_5_1_14
void FASTCALL call_15_5_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_1_15
#define init_15_5_1_15
void FASTCALL call_15_5_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_2_0
#define init_15_5_2_0
void FASTCALL call_15_5_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_2_1
#define init_15_5_2_1
void FASTCALL call_15_5_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_2_2
#define init_15_5_2_2
void FASTCALL call_15_5_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_2_3
#define init_15_5_2_3
void FASTCALL call_15_5_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_2_4
#define init_15_5_2_4
void FASTCALL call_15_5_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_2_5
#define init_15_5_2_5
void FASTCALL call_15_5_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_2_6
#define init_15_5_2_6
void FASTCALL call_15_5_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_2_7
#define init_15_5_2_7
void FASTCALL call_15_5_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_2_8
#define init_15_5_2_8
void FASTCALL call_15_5_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_2_9
#define init_15_5_2_9
void FASTCALL call_15_5_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_2_10
#define init_15_5_2_10
void FASTCALL call_15_5_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_2_11
#define init_15_5_2_11
void FASTCALL call_15_5_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_2_12
#define init_15_5_2_12
void FASTCALL call_15_5_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_2_13
#define init_15_5_2_13
void FASTCALL call_15_5_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_2_14
#define init_15_5_2_14
void FASTCALL call_15_5_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_2_15
#define init_15_5_2_15
void FASTCALL call_15_5_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_3_0
#define init_15_5_3_0
void FASTCALL call_15_5_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_3_1
#define init_15_5_3_1
void FASTCALL call_15_5_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_3_2
#define init_15_5_3_2
void FASTCALL call_15_5_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_3_3
#define init_15_5_3_3
void FASTCALL call_15_5_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_3_4
#define init_15_5_3_4
void FASTCALL call_15_5_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_3_5
#define init_15_5_3_5
void FASTCALL call_15_5_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_3_6
#define init_15_5_3_6
void FASTCALL call_15_5_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_3_7
#define init_15_5_3_7
void FASTCALL call_15_5_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_3_8
#define init_15_5_3_8
void FASTCALL call_15_5_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_3_9
#define init_15_5_3_9
void FASTCALL call_15_5_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_3_10
#define init_15_5_3_10
void FASTCALL call_15_5_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_3_11
#define init_15_5_3_11
void FASTCALL call_15_5_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_3_12
#define init_15_5_3_12
void FASTCALL call_15_5_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_3_13
#define init_15_5_3_13
void FASTCALL call_15_5_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_3_14
#define init_15_5_3_14
void FASTCALL call_15_5_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_3_15
#define init_15_5_3_15
void FASTCALL call_15_5_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_4_0
#define init_15_5_4_0
void FASTCALL call_15_5_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_4_1
#define init_15_5_4_1
void FASTCALL call_15_5_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_4_2
#define init_15_5_4_2
void FASTCALL call_15_5_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_4_3
#define init_15_5_4_3
void FASTCALL call_15_5_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_4_4
#define init_15_5_4_4
void FASTCALL call_15_5_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_4_5
#define init_15_5_4_5
void FASTCALL call_15_5_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_4_6
#define init_15_5_4_6
void FASTCALL call_15_5_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_4_7
#define init_15_5_4_7
void FASTCALL call_15_5_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_4_8
#define init_15_5_4_8
void FASTCALL call_15_5_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_4_9
#define init_15_5_4_9
void FASTCALL call_15_5_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_4_10
#define init_15_5_4_10
void FASTCALL call_15_5_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_4_11
#define init_15_5_4_11
void FASTCALL call_15_5_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_4_12
#define init_15_5_4_12
void FASTCALL call_15_5_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_4_13
#define init_15_5_4_13
void FASTCALL call_15_5_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_4_14
#define init_15_5_4_14
void FASTCALL call_15_5_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_4_15
#define init_15_5_4_15
void FASTCALL call_15_5_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_5_0
#define init_15_5_5_0
void FASTCALL call_15_5_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_5_1
#define init_15_5_5_1
void FASTCALL call_15_5_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_5_2
#define init_15_5_5_2
void FASTCALL call_15_5_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_5_3
#define init_15_5_5_3
void FASTCALL call_15_5_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_5_4
#define init_15_5_5_4
void FASTCALL call_15_5_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_5_5
#define init_15_5_5_5
void FASTCALL call_15_5_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_5_6
#define init_15_5_5_6
void FASTCALL call_15_5_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_5_7
#define init_15_5_5_7
void FASTCALL call_15_5_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_5_8
#define init_15_5_5_8
void FASTCALL call_15_5_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_5_9
#define init_15_5_5_9
void FASTCALL call_15_5_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_5_10
#define init_15_5_5_10
void FASTCALL call_15_5_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_5_11
#define init_15_5_5_11
void FASTCALL call_15_5_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_5_12
#define init_15_5_5_12
void FASTCALL call_15_5_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_5_13
#define init_15_5_5_13
void FASTCALL call_15_5_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_5_14
#define init_15_5_5_14
void FASTCALL call_15_5_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_5_15
#define init_15_5_5_15
void FASTCALL call_15_5_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_6_0
#define init_15_5_6_0
void FASTCALL call_15_5_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_6_1
#define init_15_5_6_1
void FASTCALL call_15_5_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_6_2
#define init_15_5_6_2
void FASTCALL call_15_5_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_6_3
#define init_15_5_6_3
void FASTCALL call_15_5_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_6_4
#define init_15_5_6_4
void FASTCALL call_15_5_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_6_5
#define init_15_5_6_5
void FASTCALL call_15_5_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_6_6
#define init_15_5_6_6
void FASTCALL call_15_5_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_6_7
#define init_15_5_6_7
void FASTCALL call_15_5_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_6_8
#define init_15_5_6_8
void FASTCALL call_15_5_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_6_9
#define init_15_5_6_9
void FASTCALL call_15_5_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_6_10
#define init_15_5_6_10
void FASTCALL call_15_5_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_6_11
#define init_15_5_6_11
void FASTCALL call_15_5_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_6_12
#define init_15_5_6_12
void FASTCALL call_15_5_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_6_13
#define init_15_5_6_13
void FASTCALL call_15_5_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_6_14
#define init_15_5_6_14
void FASTCALL call_15_5_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_6_15
#define init_15_5_6_15
void FASTCALL call_15_5_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_7_0
#define init_15_5_7_0
void FASTCALL call_15_5_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_7_1
#define init_15_5_7_1
void FASTCALL call_15_5_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_7_2
#define init_15_5_7_2
void FASTCALL call_15_5_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_7_3
#define init_15_5_7_3
void FASTCALL call_15_5_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_7_4
#define init_15_5_7_4
void FASTCALL call_15_5_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_7_5
#define init_15_5_7_5
void FASTCALL call_15_5_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_7_6
#define init_15_5_7_6
void FASTCALL call_15_5_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_7_7
#define init_15_5_7_7
void FASTCALL call_15_5_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_7_8
#define init_15_5_7_8
void FASTCALL call_15_5_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_7_9
#define init_15_5_7_9
void FASTCALL call_15_5_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_7_10
#define init_15_5_7_10
void FASTCALL call_15_5_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_7_11
#define init_15_5_7_11
void FASTCALL call_15_5_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_7_12
#define init_15_5_7_12
void FASTCALL call_15_5_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_7_13
#define init_15_5_7_13
void FASTCALL call_15_5_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_7_14
#define init_15_5_7_14
void FASTCALL call_15_5_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_7_15
#define init_15_5_7_15
void FASTCALL call_15_5_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_8_0
#define init_15_5_8_0
void FASTCALL call_15_5_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_8_1
#define init_15_5_8_1
void FASTCALL call_15_5_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_8_2
#define init_15_5_8_2
void FASTCALL call_15_5_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_8_3
#define init_15_5_8_3
void FASTCALL call_15_5_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_8_4
#define init_15_5_8_4
void FASTCALL call_15_5_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_8_5
#define init_15_5_8_5
void FASTCALL call_15_5_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_8_6
#define init_15_5_8_6
void FASTCALL call_15_5_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_8_7
#define init_15_5_8_7
void FASTCALL call_15_5_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_8_8
#define init_15_5_8_8
void FASTCALL call_15_5_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_8_9
#define init_15_5_8_9
void FASTCALL call_15_5_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_8_10
#define init_15_5_8_10
void FASTCALL call_15_5_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_8_11
#define init_15_5_8_11
void FASTCALL call_15_5_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_8_12
#define init_15_5_8_12
void FASTCALL call_15_5_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_8_13
#define init_15_5_8_13
void FASTCALL call_15_5_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_8_14
#define init_15_5_8_14
void FASTCALL call_15_5_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_8_15
#define init_15_5_8_15
void FASTCALL call_15_5_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_9_0
#define init_15_5_9_0
void FASTCALL call_15_5_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_9_1
#define init_15_5_9_1
void FASTCALL call_15_5_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_9_2
#define init_15_5_9_2
void FASTCALL call_15_5_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_9_3
#define init_15_5_9_3
void FASTCALL call_15_5_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_9_4
#define init_15_5_9_4
void FASTCALL call_15_5_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_9_5
#define init_15_5_9_5
void FASTCALL call_15_5_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_9_6
#define init_15_5_9_6
void FASTCALL call_15_5_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_9_7
#define init_15_5_9_7
void FASTCALL call_15_5_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_9_8
#define init_15_5_9_8
void FASTCALL call_15_5_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_9_9
#define init_15_5_9_9
void FASTCALL call_15_5_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_9_10
#define init_15_5_9_10
void FASTCALL call_15_5_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_9_11
#define init_15_5_9_11
void FASTCALL call_15_5_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_9_12
#define init_15_5_9_12
void FASTCALL call_15_5_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_9_13
#define init_15_5_9_13
void FASTCALL call_15_5_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_9_14
#define init_15_5_9_14
void FASTCALL call_15_5_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_9_15
#define init_15_5_9_15
void FASTCALL call_15_5_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_10_0
#define init_15_5_10_0
void FASTCALL call_15_5_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_10_1
#define init_15_5_10_1
void FASTCALL call_15_5_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_10_2
#define init_15_5_10_2
void FASTCALL call_15_5_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_10_3
#define init_15_5_10_3
void FASTCALL call_15_5_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_10_4
#define init_15_5_10_4
void FASTCALL call_15_5_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_10_5
#define init_15_5_10_5
void FASTCALL call_15_5_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_10_6
#define init_15_5_10_6
void FASTCALL call_15_5_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_10_7
#define init_15_5_10_7
void FASTCALL call_15_5_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_10_8
#define init_15_5_10_8
void FASTCALL call_15_5_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_10_9
#define init_15_5_10_9
void FASTCALL call_15_5_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_10_10
#define init_15_5_10_10
void FASTCALL call_15_5_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_10_11
#define init_15_5_10_11
void FASTCALL call_15_5_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_10_12
#define init_15_5_10_12
void FASTCALL call_15_5_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_10_13
#define init_15_5_10_13
void FASTCALL call_15_5_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_10_14
#define init_15_5_10_14
void FASTCALL call_15_5_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_10_15
#define init_15_5_10_15
void FASTCALL call_15_5_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_11_0
#define init_15_5_11_0
void FASTCALL call_15_5_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_11_1
#define init_15_5_11_1
void FASTCALL call_15_5_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_11_2
#define init_15_5_11_2
void FASTCALL call_15_5_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_11_3
#define init_15_5_11_3
void FASTCALL call_15_5_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_11_4
#define init_15_5_11_4
void FASTCALL call_15_5_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_11_5
#define init_15_5_11_5
void FASTCALL call_15_5_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_11_6
#define init_15_5_11_6
void FASTCALL call_15_5_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_11_7
#define init_15_5_11_7
void FASTCALL call_15_5_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_11_8
#define init_15_5_11_8
void FASTCALL call_15_5_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_11_9
#define init_15_5_11_9
void FASTCALL call_15_5_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_11_10
#define init_15_5_11_10
void FASTCALL call_15_5_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_11_11
#define init_15_5_11_11
void FASTCALL call_15_5_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_11_12
#define init_15_5_11_12
void FASTCALL call_15_5_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_11_13
#define init_15_5_11_13
void FASTCALL call_15_5_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_11_14
#define init_15_5_11_14
void FASTCALL call_15_5_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_11_15
#define init_15_5_11_15
void FASTCALL call_15_5_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_12_0
#define init_15_5_12_0
void FASTCALL call_15_5_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_12_1
#define init_15_5_12_1
void FASTCALL call_15_5_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_12_2
#define init_15_5_12_2
void FASTCALL call_15_5_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_12_3
#define init_15_5_12_3
void FASTCALL call_15_5_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_12_4
#define init_15_5_12_4
void FASTCALL call_15_5_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_12_5
#define init_15_5_12_5
void FASTCALL call_15_5_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_12_6
#define init_15_5_12_6
void FASTCALL call_15_5_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_12_7
#define init_15_5_12_7
void FASTCALL call_15_5_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_12_8
#define init_15_5_12_8
void FASTCALL call_15_5_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_12_9
#define init_15_5_12_9
void FASTCALL call_15_5_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_12_10
#define init_15_5_12_10
void FASTCALL call_15_5_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_12_11
#define init_15_5_12_11
void FASTCALL call_15_5_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_12_12
#define init_15_5_12_12
void FASTCALL call_15_5_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_12_13
#define init_15_5_12_13
void FASTCALL call_15_5_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_12_14
#define init_15_5_12_14
void FASTCALL call_15_5_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_12_15
#define init_15_5_12_15
void FASTCALL call_15_5_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_13_0
#define init_15_5_13_0
void FASTCALL call_15_5_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_13_1
#define init_15_5_13_1
void FASTCALL call_15_5_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_13_2
#define init_15_5_13_2
void FASTCALL call_15_5_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_13_3
#define init_15_5_13_3
void FASTCALL call_15_5_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_13_4
#define init_15_5_13_4
void FASTCALL call_15_5_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_13_5
#define init_15_5_13_5
void FASTCALL call_15_5_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_13_6
#define init_15_5_13_6
void FASTCALL call_15_5_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_13_7
#define init_15_5_13_7
void FASTCALL call_15_5_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_13_8
#define init_15_5_13_8
void FASTCALL call_15_5_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_13_9
#define init_15_5_13_9
void FASTCALL call_15_5_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_13_10
#define init_15_5_13_10
void FASTCALL call_15_5_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_13_11
#define init_15_5_13_11
void FASTCALL call_15_5_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_13_12
#define init_15_5_13_12
void FASTCALL call_15_5_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_13_13
#define init_15_5_13_13
void FASTCALL call_15_5_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_13_14
#define init_15_5_13_14
void FASTCALL call_15_5_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_13_15
#define init_15_5_13_15
void FASTCALL call_15_5_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_14_0
#define init_15_5_14_0
void FASTCALL call_15_5_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_14_1
#define init_15_5_14_1
void FASTCALL call_15_5_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_14_2
#define init_15_5_14_2
void FASTCALL call_15_5_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_14_3
#define init_15_5_14_3
void FASTCALL call_15_5_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_14_4
#define init_15_5_14_4
void FASTCALL call_15_5_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_14_5
#define init_15_5_14_5
void FASTCALL call_15_5_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_14_6
#define init_15_5_14_6
void FASTCALL call_15_5_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_14_7
#define init_15_5_14_7
void FASTCALL call_15_5_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_14_8
#define init_15_5_14_8
void FASTCALL call_15_5_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_14_9
#define init_15_5_14_9
void FASTCALL call_15_5_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_14_10
#define init_15_5_14_10
void FASTCALL call_15_5_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_14_11
#define init_15_5_14_11
void FASTCALL call_15_5_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_14_12
#define init_15_5_14_12
void FASTCALL call_15_5_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_14_13
#define init_15_5_14_13
void FASTCALL call_15_5_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_14_14
#define init_15_5_14_14
void FASTCALL call_15_5_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_14_15
#define init_15_5_14_15
void FASTCALL call_15_5_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_15_0
#define init_15_5_15_0
void FASTCALL call_15_5_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_15_1
#define init_15_5_15_1
void FASTCALL call_15_5_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_15_2
#define init_15_5_15_2
void FASTCALL call_15_5_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_15_3
#define init_15_5_15_3
void FASTCALL call_15_5_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_15_4
#define init_15_5_15_4
void FASTCALL call_15_5_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_15_5
#define init_15_5_15_5
void FASTCALL call_15_5_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_15_6
#define init_15_5_15_6
void FASTCALL call_15_5_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_15_7
#define init_15_5_15_7
void FASTCALL call_15_5_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_15_8
#define init_15_5_15_8
void FASTCALL call_15_5_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_15_9
#define init_15_5_15_9
void FASTCALL call_15_5_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_15_10
#define init_15_5_15_10
void FASTCALL call_15_5_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_15_11
#define init_15_5_15_11
void FASTCALL call_15_5_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_15_12
#define init_15_5_15_12
void FASTCALL call_15_5_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_15_13
#define init_15_5_15_13
void FASTCALL call_15_5_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_15_14
#define init_15_5_15_14
void FASTCALL call_15_5_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_5_15_15
#define init_15_5_15_15
void FASTCALL call_15_5_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_0_0
#define init_15_6_0_0
void FASTCALL call_15_6_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_0_1
#define init_15_6_0_1
void FASTCALL call_15_6_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_0_2
#define init_15_6_0_2
void FASTCALL call_15_6_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_0_3
#define init_15_6_0_3
void FASTCALL call_15_6_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_0_4
#define init_15_6_0_4
void FASTCALL call_15_6_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_0_5
#define init_15_6_0_5
void FASTCALL call_15_6_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_0_6
#define init_15_6_0_6
void FASTCALL call_15_6_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_0_7
#define init_15_6_0_7
void FASTCALL call_15_6_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_0_8
#define init_15_6_0_8
void FASTCALL call_15_6_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_0_9
#define init_15_6_0_9
void FASTCALL call_15_6_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_0_10
#define init_15_6_0_10
void FASTCALL call_15_6_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_0_11
#define init_15_6_0_11
void FASTCALL call_15_6_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_0_12
#define init_15_6_0_12
void FASTCALL call_15_6_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_0_13
#define init_15_6_0_13
void FASTCALL call_15_6_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_0_14
#define init_15_6_0_14
void FASTCALL call_15_6_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_0_15
#define init_15_6_0_15
void FASTCALL call_15_6_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_1_0
#define init_15_6_1_0
void FASTCALL call_15_6_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_1_1
#define init_15_6_1_1
void FASTCALL call_15_6_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_1_2
#define init_15_6_1_2
void FASTCALL call_15_6_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_1_3
#define init_15_6_1_3
void FASTCALL call_15_6_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_1_4
#define init_15_6_1_4
void FASTCALL call_15_6_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_1_5
#define init_15_6_1_5
void FASTCALL call_15_6_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_1_6
#define init_15_6_1_6
void FASTCALL call_15_6_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_1_7
#define init_15_6_1_7
void FASTCALL call_15_6_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_1_8
#define init_15_6_1_8
void FASTCALL call_15_6_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_1_9
#define init_15_6_1_9
void FASTCALL call_15_6_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_1_10
#define init_15_6_1_10
void FASTCALL call_15_6_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_1_11
#define init_15_6_1_11
void FASTCALL call_15_6_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_1_12
#define init_15_6_1_12
void FASTCALL call_15_6_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_1_13
#define init_15_6_1_13
void FASTCALL call_15_6_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_1_14
#define init_15_6_1_14
void FASTCALL call_15_6_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_1_15
#define init_15_6_1_15
void FASTCALL call_15_6_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_2_0
#define init_15_6_2_0
void FASTCALL call_15_6_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_2_1
#define init_15_6_2_1
void FASTCALL call_15_6_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_2_2
#define init_15_6_2_2
void FASTCALL call_15_6_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_2_3
#define init_15_6_2_3
void FASTCALL call_15_6_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_2_4
#define init_15_6_2_4
void FASTCALL call_15_6_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_2_5
#define init_15_6_2_5
void FASTCALL call_15_6_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_2_6
#define init_15_6_2_6
void FASTCALL call_15_6_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_2_7
#define init_15_6_2_7
void FASTCALL call_15_6_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_2_8
#define init_15_6_2_8
void FASTCALL call_15_6_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_2_9
#define init_15_6_2_9
void FASTCALL call_15_6_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_2_10
#define init_15_6_2_10
void FASTCALL call_15_6_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_2_11
#define init_15_6_2_11
void FASTCALL call_15_6_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_2_12
#define init_15_6_2_12
void FASTCALL call_15_6_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_2_13
#define init_15_6_2_13
void FASTCALL call_15_6_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_2_14
#define init_15_6_2_14
void FASTCALL call_15_6_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_2_15
#define init_15_6_2_15
void FASTCALL call_15_6_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_3_0
#define init_15_6_3_0
void FASTCALL call_15_6_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_3_1
#define init_15_6_3_1
void FASTCALL call_15_6_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_3_2
#define init_15_6_3_2
void FASTCALL call_15_6_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_3_3
#define init_15_6_3_3
void FASTCALL call_15_6_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_3_4
#define init_15_6_3_4
void FASTCALL call_15_6_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_3_5
#define init_15_6_3_5
void FASTCALL call_15_6_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_3_6
#define init_15_6_3_6
void FASTCALL call_15_6_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_3_7
#define init_15_6_3_7
void FASTCALL call_15_6_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_3_8
#define init_15_6_3_8
void FASTCALL call_15_6_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_3_9
#define init_15_6_3_9
void FASTCALL call_15_6_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_3_10
#define init_15_6_3_10
void FASTCALL call_15_6_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_3_11
#define init_15_6_3_11
void FASTCALL call_15_6_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_3_12
#define init_15_6_3_12
void FASTCALL call_15_6_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_3_13
#define init_15_6_3_13
void FASTCALL call_15_6_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_3_14
#define init_15_6_3_14
void FASTCALL call_15_6_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_3_15
#define init_15_6_3_15
void FASTCALL call_15_6_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_4_0
#define init_15_6_4_0
void FASTCALL call_15_6_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_4_1
#define init_15_6_4_1
void FASTCALL call_15_6_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_4_2
#define init_15_6_4_2
void FASTCALL call_15_6_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_4_3
#define init_15_6_4_3
void FASTCALL call_15_6_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_4_4
#define init_15_6_4_4
void FASTCALL call_15_6_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_4_5
#define init_15_6_4_5
void FASTCALL call_15_6_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_4_6
#define init_15_6_4_6
void FASTCALL call_15_6_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_4_7
#define init_15_6_4_7
void FASTCALL call_15_6_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_4_8
#define init_15_6_4_8
void FASTCALL call_15_6_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_4_9
#define init_15_6_4_9
void FASTCALL call_15_6_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_4_10
#define init_15_6_4_10
void FASTCALL call_15_6_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_4_11
#define init_15_6_4_11
void FASTCALL call_15_6_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_4_12
#define init_15_6_4_12
void FASTCALL call_15_6_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_4_13
#define init_15_6_4_13
void FASTCALL call_15_6_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_4_14
#define init_15_6_4_14
void FASTCALL call_15_6_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_4_15
#define init_15_6_4_15
void FASTCALL call_15_6_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_5_0
#define init_15_6_5_0
void FASTCALL call_15_6_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_5_1
#define init_15_6_5_1
void FASTCALL call_15_6_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_5_2
#define init_15_6_5_2
void FASTCALL call_15_6_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_5_3
#define init_15_6_5_3
void FASTCALL call_15_6_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_5_4
#define init_15_6_5_4
void FASTCALL call_15_6_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_5_5
#define init_15_6_5_5
void FASTCALL call_15_6_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_5_6
#define init_15_6_5_6
void FASTCALL call_15_6_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_5_7
#define init_15_6_5_7
void FASTCALL call_15_6_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_5_8
#define init_15_6_5_8
void FASTCALL call_15_6_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_5_9
#define init_15_6_5_9
void FASTCALL call_15_6_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_5_10
#define init_15_6_5_10
void FASTCALL call_15_6_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_5_11
#define init_15_6_5_11
void FASTCALL call_15_6_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_5_12
#define init_15_6_5_12
void FASTCALL call_15_6_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_5_13
#define init_15_6_5_13
void FASTCALL call_15_6_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_5_14
#define init_15_6_5_14
void FASTCALL call_15_6_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_5_15
#define init_15_6_5_15
void FASTCALL call_15_6_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_6_0
#define init_15_6_6_0
void FASTCALL call_15_6_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_6_1
#define init_15_6_6_1
void FASTCALL call_15_6_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_6_2
#define init_15_6_6_2
void FASTCALL call_15_6_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_6_3
#define init_15_6_6_3
void FASTCALL call_15_6_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_6_4
#define init_15_6_6_4
void FASTCALL call_15_6_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_6_5
#define init_15_6_6_5
void FASTCALL call_15_6_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_6_6
#define init_15_6_6_6
void FASTCALL call_15_6_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_6_7
#define init_15_6_6_7
void FASTCALL call_15_6_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_6_8
#define init_15_6_6_8
void FASTCALL call_15_6_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_6_9
#define init_15_6_6_9
void FASTCALL call_15_6_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_6_10
#define init_15_6_6_10
void FASTCALL call_15_6_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_6_11
#define init_15_6_6_11
void FASTCALL call_15_6_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_6_12
#define init_15_6_6_12
void FASTCALL call_15_6_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_6_13
#define init_15_6_6_13
void FASTCALL call_15_6_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_6_14
#define init_15_6_6_14
void FASTCALL call_15_6_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_6_15
#define init_15_6_6_15
void FASTCALL call_15_6_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_7_0
#define init_15_6_7_0
void FASTCALL call_15_6_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_7_1
#define init_15_6_7_1
void FASTCALL call_15_6_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_7_2
#define init_15_6_7_2
void FASTCALL call_15_6_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_7_3
#define init_15_6_7_3
void FASTCALL call_15_6_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_7_4
#define init_15_6_7_4
void FASTCALL call_15_6_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_7_5
#define init_15_6_7_5
void FASTCALL call_15_6_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_7_6
#define init_15_6_7_6
void FASTCALL call_15_6_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_7_7
#define init_15_6_7_7
void FASTCALL call_15_6_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_7_8
#define init_15_6_7_8
void FASTCALL call_15_6_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_7_9
#define init_15_6_7_9
void FASTCALL call_15_6_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_7_10
#define init_15_6_7_10
void FASTCALL call_15_6_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_7_11
#define init_15_6_7_11
void FASTCALL call_15_6_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_7_12
#define init_15_6_7_12
void FASTCALL call_15_6_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_7_13
#define init_15_6_7_13
void FASTCALL call_15_6_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_7_14
#define init_15_6_7_14
void FASTCALL call_15_6_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_7_15
#define init_15_6_7_15
void FASTCALL call_15_6_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_8_0
#define init_15_6_8_0
void FASTCALL call_15_6_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_8_1
#define init_15_6_8_1
void FASTCALL call_15_6_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_8_2
#define init_15_6_8_2
void FASTCALL call_15_6_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_8_3
#define init_15_6_8_3
void FASTCALL call_15_6_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_8_4
#define init_15_6_8_4
void FASTCALL call_15_6_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_8_5
#define init_15_6_8_5
void FASTCALL call_15_6_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_8_6
#define init_15_6_8_6
void FASTCALL call_15_6_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_8_7
#define init_15_6_8_7
void FASTCALL call_15_6_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_8_8
#define init_15_6_8_8
void FASTCALL call_15_6_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_8_9
#define init_15_6_8_9
void FASTCALL call_15_6_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_8_10
#define init_15_6_8_10
void FASTCALL call_15_6_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_8_11
#define init_15_6_8_11
void FASTCALL call_15_6_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_8_12
#define init_15_6_8_12
void FASTCALL call_15_6_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_8_13
#define init_15_6_8_13
void FASTCALL call_15_6_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_8_14
#define init_15_6_8_14
void FASTCALL call_15_6_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_8_15
#define init_15_6_8_15
void FASTCALL call_15_6_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_9_0
#define init_15_6_9_0
void FASTCALL call_15_6_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_9_1
#define init_15_6_9_1
void FASTCALL call_15_6_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_9_2
#define init_15_6_9_2
void FASTCALL call_15_6_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_9_3
#define init_15_6_9_3
void FASTCALL call_15_6_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_9_4
#define init_15_6_9_4
void FASTCALL call_15_6_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_9_5
#define init_15_6_9_5
void FASTCALL call_15_6_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_9_6
#define init_15_6_9_6
void FASTCALL call_15_6_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_9_7
#define init_15_6_9_7
void FASTCALL call_15_6_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_9_8
#define init_15_6_9_8
void FASTCALL call_15_6_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_9_9
#define init_15_6_9_9
void FASTCALL call_15_6_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_9_10
#define init_15_6_9_10
void FASTCALL call_15_6_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_9_11
#define init_15_6_9_11
void FASTCALL call_15_6_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_9_12
#define init_15_6_9_12
void FASTCALL call_15_6_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_9_13
#define init_15_6_9_13
void FASTCALL call_15_6_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_9_14
#define init_15_6_9_14
void FASTCALL call_15_6_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_9_15
#define init_15_6_9_15
void FASTCALL call_15_6_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_10_0
#define init_15_6_10_0
void FASTCALL call_15_6_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_10_1
#define init_15_6_10_1
void FASTCALL call_15_6_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_10_2
#define init_15_6_10_2
void FASTCALL call_15_6_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_10_3
#define init_15_6_10_3
void FASTCALL call_15_6_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_10_4
#define init_15_6_10_4
void FASTCALL call_15_6_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_10_5
#define init_15_6_10_5
void FASTCALL call_15_6_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_10_6
#define init_15_6_10_6
void FASTCALL call_15_6_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_10_7
#define init_15_6_10_7
void FASTCALL call_15_6_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_10_8
#define init_15_6_10_8
void FASTCALL call_15_6_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_10_9
#define init_15_6_10_9
void FASTCALL call_15_6_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_10_10
#define init_15_6_10_10
void FASTCALL call_15_6_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_10_11
#define init_15_6_10_11
void FASTCALL call_15_6_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_10_12
#define init_15_6_10_12
void FASTCALL call_15_6_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_10_13
#define init_15_6_10_13
void FASTCALL call_15_6_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_10_14
#define init_15_6_10_14
void FASTCALL call_15_6_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_10_15
#define init_15_6_10_15
void FASTCALL call_15_6_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_11_0
#define init_15_6_11_0
void FASTCALL call_15_6_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_11_1
#define init_15_6_11_1
void FASTCALL call_15_6_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_11_2
#define init_15_6_11_2
void FASTCALL call_15_6_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_11_3
#define init_15_6_11_3
void FASTCALL call_15_6_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_11_4
#define init_15_6_11_4
void FASTCALL call_15_6_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_11_5
#define init_15_6_11_5
void FASTCALL call_15_6_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_11_6
#define init_15_6_11_6
void FASTCALL call_15_6_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_11_7
#define init_15_6_11_7
void FASTCALL call_15_6_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_11_8
#define init_15_6_11_8
void FASTCALL call_15_6_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_11_9
#define init_15_6_11_9
void FASTCALL call_15_6_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_11_10
#define init_15_6_11_10
void FASTCALL call_15_6_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_11_11
#define init_15_6_11_11
void FASTCALL call_15_6_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_11_12
#define init_15_6_11_12
void FASTCALL call_15_6_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_11_13
#define init_15_6_11_13
void FASTCALL call_15_6_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_11_14
#define init_15_6_11_14
void FASTCALL call_15_6_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_11_15
#define init_15_6_11_15
void FASTCALL call_15_6_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_12_0
#define init_15_6_12_0
void FASTCALL call_15_6_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_12_1
#define init_15_6_12_1
void FASTCALL call_15_6_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_12_2
#define init_15_6_12_2
void FASTCALL call_15_6_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_12_3
#define init_15_6_12_3
void FASTCALL call_15_6_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_12_4
#define init_15_6_12_4
void FASTCALL call_15_6_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_12_5
#define init_15_6_12_5
void FASTCALL call_15_6_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_12_6
#define init_15_6_12_6
void FASTCALL call_15_6_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_12_7
#define init_15_6_12_7
void FASTCALL call_15_6_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_12_8
#define init_15_6_12_8
void FASTCALL call_15_6_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_12_9
#define init_15_6_12_9
void FASTCALL call_15_6_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_12_10
#define init_15_6_12_10
void FASTCALL call_15_6_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_12_11
#define init_15_6_12_11
void FASTCALL call_15_6_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_12_12
#define init_15_6_12_12
void FASTCALL call_15_6_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_12_13
#define init_15_6_12_13
void FASTCALL call_15_6_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_12_14
#define init_15_6_12_14
void FASTCALL call_15_6_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_12_15
#define init_15_6_12_15
void FASTCALL call_15_6_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_13_0
#define init_15_6_13_0
void FASTCALL call_15_6_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_13_1
#define init_15_6_13_1
void FASTCALL call_15_6_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_13_2
#define init_15_6_13_2
void FASTCALL call_15_6_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_13_3
#define init_15_6_13_3
void FASTCALL call_15_6_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_13_4
#define init_15_6_13_4
void FASTCALL call_15_6_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_13_5
#define init_15_6_13_5
void FASTCALL call_15_6_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_13_6
#define init_15_6_13_6
void FASTCALL call_15_6_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_13_7
#define init_15_6_13_7
void FASTCALL call_15_6_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_13_8
#define init_15_6_13_8
void FASTCALL call_15_6_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_13_9
#define init_15_6_13_9
void FASTCALL call_15_6_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_13_10
#define init_15_6_13_10
void FASTCALL call_15_6_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_13_11
#define init_15_6_13_11
void FASTCALL call_15_6_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_13_12
#define init_15_6_13_12
void FASTCALL call_15_6_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_13_13
#define init_15_6_13_13
void FASTCALL call_15_6_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_13_14
#define init_15_6_13_14
void FASTCALL call_15_6_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_13_15
#define init_15_6_13_15
void FASTCALL call_15_6_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_14_0
#define init_15_6_14_0
void FASTCALL call_15_6_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_14_1
#define init_15_6_14_1
void FASTCALL call_15_6_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_14_2
#define init_15_6_14_2
void FASTCALL call_15_6_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_14_3
#define init_15_6_14_3
void FASTCALL call_15_6_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_14_4
#define init_15_6_14_4
void FASTCALL call_15_6_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_14_5
#define init_15_6_14_5
void FASTCALL call_15_6_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_14_6
#define init_15_6_14_6
void FASTCALL call_15_6_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_14_7
#define init_15_6_14_7
void FASTCALL call_15_6_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_14_8
#define init_15_6_14_8
void FASTCALL call_15_6_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_14_9
#define init_15_6_14_9
void FASTCALL call_15_6_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_14_10
#define init_15_6_14_10
void FASTCALL call_15_6_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_14_11
#define init_15_6_14_11
void FASTCALL call_15_6_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_14_12
#define init_15_6_14_12
void FASTCALL call_15_6_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_14_13
#define init_15_6_14_13
void FASTCALL call_15_6_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_14_14
#define init_15_6_14_14
void FASTCALL call_15_6_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_14_15
#define init_15_6_14_15
void FASTCALL call_15_6_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_15_0
#define init_15_6_15_0
void FASTCALL call_15_6_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_15_1
#define init_15_6_15_1
void FASTCALL call_15_6_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_15_2
#define init_15_6_15_2
void FASTCALL call_15_6_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_15_3
#define init_15_6_15_3
void FASTCALL call_15_6_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_15_4
#define init_15_6_15_4
void FASTCALL call_15_6_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_15_5
#define init_15_6_15_5
void FASTCALL call_15_6_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_15_6
#define init_15_6_15_6
void FASTCALL call_15_6_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_15_7
#define init_15_6_15_7
void FASTCALL call_15_6_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_15_8
#define init_15_6_15_8
void FASTCALL call_15_6_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_15_9
#define init_15_6_15_9
void FASTCALL call_15_6_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_15_10
#define init_15_6_15_10
void FASTCALL call_15_6_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_15_11
#define init_15_6_15_11
void FASTCALL call_15_6_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_15_12
#define init_15_6_15_12
void FASTCALL call_15_6_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_15_13
#define init_15_6_15_13
void FASTCALL call_15_6_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_15_14
#define init_15_6_15_14
void FASTCALL call_15_6_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_6_15_15
#define init_15_6_15_15
void FASTCALL call_15_6_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_0_0
#define init_15_7_0_0
void FASTCALL call_15_7_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_0_1
#define init_15_7_0_1
void FASTCALL call_15_7_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_0_2
#define init_15_7_0_2
void FASTCALL call_15_7_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_0_3
#define init_15_7_0_3
void FASTCALL call_15_7_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_0_4
#define init_15_7_0_4
void FASTCALL call_15_7_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_0_5
#define init_15_7_0_5
void FASTCALL call_15_7_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_0_6
#define init_15_7_0_6
void FASTCALL call_15_7_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_0_7
#define init_15_7_0_7
void FASTCALL call_15_7_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_0_8
#define init_15_7_0_8
void FASTCALL call_15_7_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_0_9
#define init_15_7_0_9
void FASTCALL call_15_7_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_0_10
#define init_15_7_0_10
void FASTCALL call_15_7_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_0_11
#define init_15_7_0_11
void FASTCALL call_15_7_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_0_12
#define init_15_7_0_12
void FASTCALL call_15_7_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_0_13
#define init_15_7_0_13
void FASTCALL call_15_7_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_0_14
#define init_15_7_0_14
void FASTCALL call_15_7_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_0_15
#define init_15_7_0_15
void FASTCALL call_15_7_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_1_0
#define init_15_7_1_0
void FASTCALL call_15_7_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_1_1
#define init_15_7_1_1
void FASTCALL call_15_7_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_1_2
#define init_15_7_1_2
void FASTCALL call_15_7_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_1_3
#define init_15_7_1_3
void FASTCALL call_15_7_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_1_4
#define init_15_7_1_4
void FASTCALL call_15_7_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_1_5
#define init_15_7_1_5
void FASTCALL call_15_7_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_1_6
#define init_15_7_1_6
void FASTCALL call_15_7_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_1_7
#define init_15_7_1_7
void FASTCALL call_15_7_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_1_8
#define init_15_7_1_8
void FASTCALL call_15_7_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_1_9
#define init_15_7_1_9
void FASTCALL call_15_7_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_1_10
#define init_15_7_1_10
void FASTCALL call_15_7_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_1_11
#define init_15_7_1_11
void FASTCALL call_15_7_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_1_12
#define init_15_7_1_12
void FASTCALL call_15_7_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_1_13
#define init_15_7_1_13
void FASTCALL call_15_7_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_1_14
#define init_15_7_1_14
void FASTCALL call_15_7_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_1_15
#define init_15_7_1_15
void FASTCALL call_15_7_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_2_0
#define init_15_7_2_0
void FASTCALL call_15_7_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_2_1
#define init_15_7_2_1
void FASTCALL call_15_7_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_2_2
#define init_15_7_2_2
void FASTCALL call_15_7_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_2_3
#define init_15_7_2_3
void FASTCALL call_15_7_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_2_4
#define init_15_7_2_4
void FASTCALL call_15_7_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_2_5
#define init_15_7_2_5
void FASTCALL call_15_7_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_2_6
#define init_15_7_2_6
void FASTCALL call_15_7_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_2_7
#define init_15_7_2_7
void FASTCALL call_15_7_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_2_8
#define init_15_7_2_8
void FASTCALL call_15_7_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_2_9
#define init_15_7_2_9
void FASTCALL call_15_7_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_2_10
#define init_15_7_2_10
void FASTCALL call_15_7_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_2_11
#define init_15_7_2_11
void FASTCALL call_15_7_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_2_12
#define init_15_7_2_12
void FASTCALL call_15_7_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_2_13
#define init_15_7_2_13
void FASTCALL call_15_7_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_2_14
#define init_15_7_2_14
void FASTCALL call_15_7_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_2_15
#define init_15_7_2_15
void FASTCALL call_15_7_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_3_0
#define init_15_7_3_0
void FASTCALL call_15_7_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_3_1
#define init_15_7_3_1
void FASTCALL call_15_7_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_3_2
#define init_15_7_3_2
void FASTCALL call_15_7_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_3_3
#define init_15_7_3_3
void FASTCALL call_15_7_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_3_4
#define init_15_7_3_4
void FASTCALL call_15_7_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_3_5
#define init_15_7_3_5
void FASTCALL call_15_7_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_3_6
#define init_15_7_3_6
void FASTCALL call_15_7_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_3_7
#define init_15_7_3_7
void FASTCALL call_15_7_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_3_8
#define init_15_7_3_8
void FASTCALL call_15_7_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_3_9
#define init_15_7_3_9
void FASTCALL call_15_7_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_3_10
#define init_15_7_3_10
void FASTCALL call_15_7_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_3_11
#define init_15_7_3_11
void FASTCALL call_15_7_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_3_12
#define init_15_7_3_12
void FASTCALL call_15_7_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_3_13
#define init_15_7_3_13
void FASTCALL call_15_7_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_3_14
#define init_15_7_3_14
void FASTCALL call_15_7_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_3_15
#define init_15_7_3_15
void FASTCALL call_15_7_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_4_0
#define init_15_7_4_0
void FASTCALL call_15_7_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_4_1
#define init_15_7_4_1
void FASTCALL call_15_7_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_4_2
#define init_15_7_4_2
void FASTCALL call_15_7_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_4_3
#define init_15_7_4_3
void FASTCALL call_15_7_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_4_4
#define init_15_7_4_4
void FASTCALL call_15_7_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_4_5
#define init_15_7_4_5
void FASTCALL call_15_7_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_4_6
#define init_15_7_4_6
void FASTCALL call_15_7_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_4_7
#define init_15_7_4_7
void FASTCALL call_15_7_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_4_8
#define init_15_7_4_8
void FASTCALL call_15_7_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_4_9
#define init_15_7_4_9
void FASTCALL call_15_7_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_4_10
#define init_15_7_4_10
void FASTCALL call_15_7_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_4_11
#define init_15_7_4_11
void FASTCALL call_15_7_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_4_12
#define init_15_7_4_12
void FASTCALL call_15_7_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_4_13
#define init_15_7_4_13
void FASTCALL call_15_7_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_4_14
#define init_15_7_4_14
void FASTCALL call_15_7_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_4_15
#define init_15_7_4_15
void FASTCALL call_15_7_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_5_0
#define init_15_7_5_0
void FASTCALL call_15_7_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_5_1
#define init_15_7_5_1
void FASTCALL call_15_7_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_5_2
#define init_15_7_5_2
void FASTCALL call_15_7_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_5_3
#define init_15_7_5_3
void FASTCALL call_15_7_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_5_4
#define init_15_7_5_4
void FASTCALL call_15_7_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_5_5
#define init_15_7_5_5
void FASTCALL call_15_7_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_5_6
#define init_15_7_5_6
void FASTCALL call_15_7_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_5_7
#define init_15_7_5_7
void FASTCALL call_15_7_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_5_8
#define init_15_7_5_8
void FASTCALL call_15_7_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_5_9
#define init_15_7_5_9
void FASTCALL call_15_7_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_5_10
#define init_15_7_5_10
void FASTCALL call_15_7_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_5_11
#define init_15_7_5_11
void FASTCALL call_15_7_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_5_12
#define init_15_7_5_12
void FASTCALL call_15_7_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_5_13
#define init_15_7_5_13
void FASTCALL call_15_7_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_5_14
#define init_15_7_5_14
void FASTCALL call_15_7_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_5_15
#define init_15_7_5_15
void FASTCALL call_15_7_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_6_0
#define init_15_7_6_0
void FASTCALL call_15_7_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_6_1
#define init_15_7_6_1
void FASTCALL call_15_7_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_6_2
#define init_15_7_6_2
void FASTCALL call_15_7_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_6_3
#define init_15_7_6_3
void FASTCALL call_15_7_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_6_4
#define init_15_7_6_4
void FASTCALL call_15_7_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_6_5
#define init_15_7_6_5
void FASTCALL call_15_7_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_6_6
#define init_15_7_6_6
void FASTCALL call_15_7_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_6_7
#define init_15_7_6_7
void FASTCALL call_15_7_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_6_8
#define init_15_7_6_8
void FASTCALL call_15_7_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_6_9
#define init_15_7_6_9
void FASTCALL call_15_7_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_6_10
#define init_15_7_6_10
void FASTCALL call_15_7_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_6_11
#define init_15_7_6_11
void FASTCALL call_15_7_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_6_12
#define init_15_7_6_12
void FASTCALL call_15_7_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_6_13
#define init_15_7_6_13
void FASTCALL call_15_7_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_6_14
#define init_15_7_6_14
void FASTCALL call_15_7_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_6_15
#define init_15_7_6_15
void FASTCALL call_15_7_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_7_0
#define init_15_7_7_0
void FASTCALL call_15_7_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_7_1
#define init_15_7_7_1
void FASTCALL call_15_7_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_7_2
#define init_15_7_7_2
void FASTCALL call_15_7_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_7_3
#define init_15_7_7_3
void FASTCALL call_15_7_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_7_4
#define init_15_7_7_4
void FASTCALL call_15_7_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_7_5
#define init_15_7_7_5
void FASTCALL call_15_7_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_7_6
#define init_15_7_7_6
void FASTCALL call_15_7_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_7_7
#define init_15_7_7_7
void FASTCALL call_15_7_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_7_8
#define init_15_7_7_8
void FASTCALL call_15_7_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_7_9
#define init_15_7_7_9
void FASTCALL call_15_7_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_7_10
#define init_15_7_7_10
void FASTCALL call_15_7_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_7_11
#define init_15_7_7_11
void FASTCALL call_15_7_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_7_12
#define init_15_7_7_12
void FASTCALL call_15_7_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_7_13
#define init_15_7_7_13
void FASTCALL call_15_7_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_7_14
#define init_15_7_7_14
void FASTCALL call_15_7_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_7_15
#define init_15_7_7_15
void FASTCALL call_15_7_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_8_0
#define init_15_7_8_0
void FASTCALL call_15_7_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_8_1
#define init_15_7_8_1
void FASTCALL call_15_7_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_8_2
#define init_15_7_8_2
void FASTCALL call_15_7_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_8_3
#define init_15_7_8_3
void FASTCALL call_15_7_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_8_4
#define init_15_7_8_4
void FASTCALL call_15_7_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_8_5
#define init_15_7_8_5
void FASTCALL call_15_7_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_8_6
#define init_15_7_8_6
void FASTCALL call_15_7_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_8_7
#define init_15_7_8_7
void FASTCALL call_15_7_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_8_8
#define init_15_7_8_8
void FASTCALL call_15_7_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_8_9
#define init_15_7_8_9
void FASTCALL call_15_7_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_8_10
#define init_15_7_8_10
void FASTCALL call_15_7_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_8_11
#define init_15_7_8_11
void FASTCALL call_15_7_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_8_12
#define init_15_7_8_12
void FASTCALL call_15_7_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_8_13
#define init_15_7_8_13
void FASTCALL call_15_7_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_8_14
#define init_15_7_8_14
void FASTCALL call_15_7_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_8_15
#define init_15_7_8_15
void FASTCALL call_15_7_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_9_0
#define init_15_7_9_0
void FASTCALL call_15_7_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_9_1
#define init_15_7_9_1
void FASTCALL call_15_7_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_9_2
#define init_15_7_9_2
void FASTCALL call_15_7_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_9_3
#define init_15_7_9_3
void FASTCALL call_15_7_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_9_4
#define init_15_7_9_4
void FASTCALL call_15_7_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_9_5
#define init_15_7_9_5
void FASTCALL call_15_7_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_9_6
#define init_15_7_9_6
void FASTCALL call_15_7_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_9_7
#define init_15_7_9_7
void FASTCALL call_15_7_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_9_8
#define init_15_7_9_8
void FASTCALL call_15_7_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_9_9
#define init_15_7_9_9
void FASTCALL call_15_7_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_9_10
#define init_15_7_9_10
void FASTCALL call_15_7_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_9_11
#define init_15_7_9_11
void FASTCALL call_15_7_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_9_12
#define init_15_7_9_12
void FASTCALL call_15_7_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_9_13
#define init_15_7_9_13
void FASTCALL call_15_7_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_9_14
#define init_15_7_9_14
void FASTCALL call_15_7_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_9_15
#define init_15_7_9_15
void FASTCALL call_15_7_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_10_0
#define init_15_7_10_0
void FASTCALL call_15_7_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_10_1
#define init_15_7_10_1
void FASTCALL call_15_7_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_10_2
#define init_15_7_10_2
void FASTCALL call_15_7_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_10_3
#define init_15_7_10_3
void FASTCALL call_15_7_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_10_4
#define init_15_7_10_4
void FASTCALL call_15_7_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_10_5
#define init_15_7_10_5
void FASTCALL call_15_7_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_10_6
#define init_15_7_10_6
void FASTCALL call_15_7_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_10_7
#define init_15_7_10_7
void FASTCALL call_15_7_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_10_8
#define init_15_7_10_8
void FASTCALL call_15_7_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_10_9
#define init_15_7_10_9
void FASTCALL call_15_7_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_10_10
#define init_15_7_10_10
void FASTCALL call_15_7_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_10_11
#define init_15_7_10_11
void FASTCALL call_15_7_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_10_12
#define init_15_7_10_12
void FASTCALL call_15_7_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_10_13
#define init_15_7_10_13
void FASTCALL call_15_7_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_10_14
#define init_15_7_10_14
void FASTCALL call_15_7_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_10_15
#define init_15_7_10_15
void FASTCALL call_15_7_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_11_0
#define init_15_7_11_0
void FASTCALL call_15_7_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_11_1
#define init_15_7_11_1
void FASTCALL call_15_7_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_11_2
#define init_15_7_11_2
void FASTCALL call_15_7_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_11_3
#define init_15_7_11_3
void FASTCALL call_15_7_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_11_4
#define init_15_7_11_4
void FASTCALL call_15_7_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_11_5
#define init_15_7_11_5
void FASTCALL call_15_7_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_11_6
#define init_15_7_11_6
void FASTCALL call_15_7_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_11_7
#define init_15_7_11_7
void FASTCALL call_15_7_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_11_8
#define init_15_7_11_8
void FASTCALL call_15_7_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_11_9
#define init_15_7_11_9
void FASTCALL call_15_7_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_11_10
#define init_15_7_11_10
void FASTCALL call_15_7_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_11_11
#define init_15_7_11_11
void FASTCALL call_15_7_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_11_12
#define init_15_7_11_12
void FASTCALL call_15_7_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_11_13
#define init_15_7_11_13
void FASTCALL call_15_7_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_11_14
#define init_15_7_11_14
void FASTCALL call_15_7_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_11_15
#define init_15_7_11_15
void FASTCALL call_15_7_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_12_0
#define init_15_7_12_0
void FASTCALL call_15_7_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_12_1
#define init_15_7_12_1
void FASTCALL call_15_7_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_12_2
#define init_15_7_12_2
void FASTCALL call_15_7_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_12_3
#define init_15_7_12_3
void FASTCALL call_15_7_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_12_4
#define init_15_7_12_4
void FASTCALL call_15_7_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_12_5
#define init_15_7_12_5
void FASTCALL call_15_7_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_12_6
#define init_15_7_12_6
void FASTCALL call_15_7_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_12_7
#define init_15_7_12_7
void FASTCALL call_15_7_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_12_8
#define init_15_7_12_8
void FASTCALL call_15_7_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_12_9
#define init_15_7_12_9
void FASTCALL call_15_7_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_12_10
#define init_15_7_12_10
void FASTCALL call_15_7_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_12_11
#define init_15_7_12_11
void FASTCALL call_15_7_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_12_12
#define init_15_7_12_12
void FASTCALL call_15_7_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_12_13
#define init_15_7_12_13
void FASTCALL call_15_7_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_12_14
#define init_15_7_12_14
void FASTCALL call_15_7_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_12_15
#define init_15_7_12_15
void FASTCALL call_15_7_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_13_0
#define init_15_7_13_0
void FASTCALL call_15_7_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_13_1
#define init_15_7_13_1
void FASTCALL call_15_7_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_13_2
#define init_15_7_13_2
void FASTCALL call_15_7_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_13_3
#define init_15_7_13_3
void FASTCALL call_15_7_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_13_4
#define init_15_7_13_4
void FASTCALL call_15_7_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_13_5
#define init_15_7_13_5
void FASTCALL call_15_7_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_13_6
#define init_15_7_13_6
void FASTCALL call_15_7_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_13_7
#define init_15_7_13_7
void FASTCALL call_15_7_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_13_8
#define init_15_7_13_8
void FASTCALL call_15_7_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_13_9
#define init_15_7_13_9
void FASTCALL call_15_7_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_13_10
#define init_15_7_13_10
void FASTCALL call_15_7_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_13_11
#define init_15_7_13_11
void FASTCALL call_15_7_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_13_12
#define init_15_7_13_12
void FASTCALL call_15_7_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_13_13
#define init_15_7_13_13
void FASTCALL call_15_7_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_13_14
#define init_15_7_13_14
void FASTCALL call_15_7_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_13_15
#define init_15_7_13_15
void FASTCALL call_15_7_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_14_0
#define init_15_7_14_0
void FASTCALL call_15_7_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_14_1
#define init_15_7_14_1
void FASTCALL call_15_7_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_14_2
#define init_15_7_14_2
void FASTCALL call_15_7_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_14_3
#define init_15_7_14_3
void FASTCALL call_15_7_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_14_4
#define init_15_7_14_4
void FASTCALL call_15_7_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_14_5
#define init_15_7_14_5
void FASTCALL call_15_7_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_14_6
#define init_15_7_14_6
void FASTCALL call_15_7_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_14_7
#define init_15_7_14_7
void FASTCALL call_15_7_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_14_8
#define init_15_7_14_8
void FASTCALL call_15_7_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_14_9
#define init_15_7_14_9
void FASTCALL call_15_7_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_14_10
#define init_15_7_14_10
void FASTCALL call_15_7_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_14_11
#define init_15_7_14_11
void FASTCALL call_15_7_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_14_12
#define init_15_7_14_12
void FASTCALL call_15_7_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_14_13
#define init_15_7_14_13
void FASTCALL call_15_7_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_14_14
#define init_15_7_14_14
void FASTCALL call_15_7_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_14_15
#define init_15_7_14_15
void FASTCALL call_15_7_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_15_0
#define init_15_7_15_0
void FASTCALL call_15_7_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_15_1
#define init_15_7_15_1
void FASTCALL call_15_7_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_15_2
#define init_15_7_15_2
void FASTCALL call_15_7_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_15_3
#define init_15_7_15_3
void FASTCALL call_15_7_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_15_4
#define init_15_7_15_4
void FASTCALL call_15_7_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_15_5
#define init_15_7_15_5
void FASTCALL call_15_7_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_15_6
#define init_15_7_15_6
void FASTCALL call_15_7_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_15_7
#define init_15_7_15_7
void FASTCALL call_15_7_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_15_8
#define init_15_7_15_8
void FASTCALL call_15_7_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_15_9
#define init_15_7_15_9
void FASTCALL call_15_7_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_15_10
#define init_15_7_15_10
void FASTCALL call_15_7_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_15_11
#define init_15_7_15_11
void FASTCALL call_15_7_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_15_12
#define init_15_7_15_12
void FASTCALL call_15_7_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_15_13
#define init_15_7_15_13
void FASTCALL call_15_7_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_15_14
#define init_15_7_15_14
void FASTCALL call_15_7_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_7_15_15
#define init_15_7_15_15
void FASTCALL call_15_7_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_0_0
#define init_15_8_0_0
void FASTCALL call_15_8_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_0_1
#define init_15_8_0_1
void FASTCALL call_15_8_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_0_2
#define init_15_8_0_2
void FASTCALL call_15_8_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_0_3
#define init_15_8_0_3
void FASTCALL call_15_8_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_0_4
#define init_15_8_0_4
void FASTCALL call_15_8_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_0_5
#define init_15_8_0_5
void FASTCALL call_15_8_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_0_6
#define init_15_8_0_6
void FASTCALL call_15_8_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_0_7
#define init_15_8_0_7
void FASTCALL call_15_8_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_0_8
#define init_15_8_0_8
void FASTCALL call_15_8_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_0_9
#define init_15_8_0_9
void FASTCALL call_15_8_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_0_10
#define init_15_8_0_10
void FASTCALL call_15_8_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_0_11
#define init_15_8_0_11
void FASTCALL call_15_8_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_0_12
#define init_15_8_0_12
void FASTCALL call_15_8_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_0_13
#define init_15_8_0_13
void FASTCALL call_15_8_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_0_14
#define init_15_8_0_14
void FASTCALL call_15_8_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_0_15
#define init_15_8_0_15
void FASTCALL call_15_8_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_1_0
#define init_15_8_1_0
void FASTCALL call_15_8_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_1_1
#define init_15_8_1_1
void FASTCALL call_15_8_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_1_2
#define init_15_8_1_2
void FASTCALL call_15_8_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_1_3
#define init_15_8_1_3
void FASTCALL call_15_8_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_1_4
#define init_15_8_1_4
void FASTCALL call_15_8_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_1_5
#define init_15_8_1_5
void FASTCALL call_15_8_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_1_6
#define init_15_8_1_6
void FASTCALL call_15_8_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_1_7
#define init_15_8_1_7
void FASTCALL call_15_8_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_1_8
#define init_15_8_1_8
void FASTCALL call_15_8_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_1_9
#define init_15_8_1_9
void FASTCALL call_15_8_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_1_10
#define init_15_8_1_10
void FASTCALL call_15_8_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_1_11
#define init_15_8_1_11
void FASTCALL call_15_8_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_1_12
#define init_15_8_1_12
void FASTCALL call_15_8_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_1_13
#define init_15_8_1_13
void FASTCALL call_15_8_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_1_14
#define init_15_8_1_14
void FASTCALL call_15_8_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_1_15
#define init_15_8_1_15
void FASTCALL call_15_8_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_2_0
#define init_15_8_2_0
void FASTCALL call_15_8_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_2_1
#define init_15_8_2_1
void FASTCALL call_15_8_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_2_2
#define init_15_8_2_2
void FASTCALL call_15_8_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_2_3
#define init_15_8_2_3
void FASTCALL call_15_8_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_2_4
#define init_15_8_2_4
void FASTCALL call_15_8_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_2_5
#define init_15_8_2_5
void FASTCALL call_15_8_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_2_6
#define init_15_8_2_6
void FASTCALL call_15_8_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_2_7
#define init_15_8_2_7
void FASTCALL call_15_8_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_2_8
#define init_15_8_2_8
void FASTCALL call_15_8_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_2_9
#define init_15_8_2_9
void FASTCALL call_15_8_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_2_10
#define init_15_8_2_10
void FASTCALL call_15_8_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_2_11
#define init_15_8_2_11
void FASTCALL call_15_8_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_2_12
#define init_15_8_2_12
void FASTCALL call_15_8_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_2_13
#define init_15_8_2_13
void FASTCALL call_15_8_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_2_14
#define init_15_8_2_14
void FASTCALL call_15_8_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_2_15
#define init_15_8_2_15
void FASTCALL call_15_8_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_3_0
#define init_15_8_3_0
void FASTCALL call_15_8_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_3_1
#define init_15_8_3_1
void FASTCALL call_15_8_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_3_2
#define init_15_8_3_2
void FASTCALL call_15_8_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_3_3
#define init_15_8_3_3
void FASTCALL call_15_8_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_3_4
#define init_15_8_3_4
void FASTCALL call_15_8_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_3_5
#define init_15_8_3_5
void FASTCALL call_15_8_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_3_6
#define init_15_8_3_6
void FASTCALL call_15_8_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_3_7
#define init_15_8_3_7
void FASTCALL call_15_8_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_3_8
#define init_15_8_3_8
void FASTCALL call_15_8_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_3_9
#define init_15_8_3_9
void FASTCALL call_15_8_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_3_10
#define init_15_8_3_10
void FASTCALL call_15_8_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_3_11
#define init_15_8_3_11
void FASTCALL call_15_8_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_3_12
#define init_15_8_3_12
void FASTCALL call_15_8_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_3_13
#define init_15_8_3_13
void FASTCALL call_15_8_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_3_14
#define init_15_8_3_14
void FASTCALL call_15_8_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_3_15
#define init_15_8_3_15
void FASTCALL call_15_8_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_4_0
#define init_15_8_4_0
void FASTCALL call_15_8_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_4_1
#define init_15_8_4_1
void FASTCALL call_15_8_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_4_2
#define init_15_8_4_2
void FASTCALL call_15_8_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_4_3
#define init_15_8_4_3
void FASTCALL call_15_8_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_4_4
#define init_15_8_4_4
void FASTCALL call_15_8_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_4_5
#define init_15_8_4_5
void FASTCALL call_15_8_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_4_6
#define init_15_8_4_6
void FASTCALL call_15_8_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_4_7
#define init_15_8_4_7
void FASTCALL call_15_8_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_4_8
#define init_15_8_4_8
void FASTCALL call_15_8_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_4_9
#define init_15_8_4_9
void FASTCALL call_15_8_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_4_10
#define init_15_8_4_10
void FASTCALL call_15_8_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_4_11
#define init_15_8_4_11
void FASTCALL call_15_8_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_4_12
#define init_15_8_4_12
void FASTCALL call_15_8_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_4_13
#define init_15_8_4_13
void FASTCALL call_15_8_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_4_14
#define init_15_8_4_14
void FASTCALL call_15_8_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_4_15
#define init_15_8_4_15
void FASTCALL call_15_8_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_5_0
#define init_15_8_5_0
void FASTCALL call_15_8_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_5_1
#define init_15_8_5_1
void FASTCALL call_15_8_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_5_2
#define init_15_8_5_2
void FASTCALL call_15_8_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_5_3
#define init_15_8_5_3
void FASTCALL call_15_8_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_5_4
#define init_15_8_5_4
void FASTCALL call_15_8_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_5_5
#define init_15_8_5_5
void FASTCALL call_15_8_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_5_6
#define init_15_8_5_6
void FASTCALL call_15_8_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_5_7
#define init_15_8_5_7
void FASTCALL call_15_8_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_5_8
#define init_15_8_5_8
void FASTCALL call_15_8_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_5_9
#define init_15_8_5_9
void FASTCALL call_15_8_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_5_10
#define init_15_8_5_10
void FASTCALL call_15_8_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_5_11
#define init_15_8_5_11
void FASTCALL call_15_8_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_5_12
#define init_15_8_5_12
void FASTCALL call_15_8_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_5_13
#define init_15_8_5_13
void FASTCALL call_15_8_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_5_14
#define init_15_8_5_14
void FASTCALL call_15_8_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_5_15
#define init_15_8_5_15
void FASTCALL call_15_8_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_6_0
#define init_15_8_6_0
void FASTCALL call_15_8_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_6_1
#define init_15_8_6_1
void FASTCALL call_15_8_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_6_2
#define init_15_8_6_2
void FASTCALL call_15_8_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_6_3
#define init_15_8_6_3
void FASTCALL call_15_8_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_6_4
#define init_15_8_6_4
void FASTCALL call_15_8_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_6_5
#define init_15_8_6_5
void FASTCALL call_15_8_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_6_6
#define init_15_8_6_6
void FASTCALL call_15_8_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_6_7
#define init_15_8_6_7
void FASTCALL call_15_8_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_6_8
#define init_15_8_6_8
void FASTCALL call_15_8_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_6_9
#define init_15_8_6_9
void FASTCALL call_15_8_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_6_10
#define init_15_8_6_10
void FASTCALL call_15_8_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_6_11
#define init_15_8_6_11
void FASTCALL call_15_8_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_6_12
#define init_15_8_6_12
void FASTCALL call_15_8_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_6_13
#define init_15_8_6_13
void FASTCALL call_15_8_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_6_14
#define init_15_8_6_14
void FASTCALL call_15_8_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_6_15
#define init_15_8_6_15
void FASTCALL call_15_8_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_7_0
#define init_15_8_7_0
void FASTCALL call_15_8_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_7_1
#define init_15_8_7_1
void FASTCALL call_15_8_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_7_2
#define init_15_8_7_2
void FASTCALL call_15_8_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_7_3
#define init_15_8_7_3
void FASTCALL call_15_8_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_7_4
#define init_15_8_7_4
void FASTCALL call_15_8_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_7_5
#define init_15_8_7_5
void FASTCALL call_15_8_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_7_6
#define init_15_8_7_6
void FASTCALL call_15_8_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_7_7
#define init_15_8_7_7
void FASTCALL call_15_8_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_7_8
#define init_15_8_7_8
void FASTCALL call_15_8_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_7_9
#define init_15_8_7_9
void FASTCALL call_15_8_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_7_10
#define init_15_8_7_10
void FASTCALL call_15_8_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_7_11
#define init_15_8_7_11
void FASTCALL call_15_8_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_7_12
#define init_15_8_7_12
void FASTCALL call_15_8_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_7_13
#define init_15_8_7_13
void FASTCALL call_15_8_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_7_14
#define init_15_8_7_14
void FASTCALL call_15_8_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_7_15
#define init_15_8_7_15
void FASTCALL call_15_8_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_8_0
#define init_15_8_8_0
void FASTCALL call_15_8_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_8_1
#define init_15_8_8_1
void FASTCALL call_15_8_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_8_2
#define init_15_8_8_2
void FASTCALL call_15_8_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_8_3
#define init_15_8_8_3
void FASTCALL call_15_8_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_8_4
#define init_15_8_8_4
void FASTCALL call_15_8_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_8_5
#define init_15_8_8_5
void FASTCALL call_15_8_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_8_6
#define init_15_8_8_6
void FASTCALL call_15_8_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_8_7
#define init_15_8_8_7
void FASTCALL call_15_8_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_8_8
#define init_15_8_8_8
void FASTCALL call_15_8_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_8_9
#define init_15_8_8_9
void FASTCALL call_15_8_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_8_10
#define init_15_8_8_10
void FASTCALL call_15_8_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_8_11
#define init_15_8_8_11
void FASTCALL call_15_8_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_8_12
#define init_15_8_8_12
void FASTCALL call_15_8_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_8_13
#define init_15_8_8_13
void FASTCALL call_15_8_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_8_14
#define init_15_8_8_14
void FASTCALL call_15_8_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_8_15
#define init_15_8_8_15
void FASTCALL call_15_8_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_9_0
#define init_15_8_9_0
void FASTCALL call_15_8_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_9_1
#define init_15_8_9_1
void FASTCALL call_15_8_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_9_2
#define init_15_8_9_2
void FASTCALL call_15_8_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_9_3
#define init_15_8_9_3
void FASTCALL call_15_8_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_9_4
#define init_15_8_9_4
void FASTCALL call_15_8_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_9_5
#define init_15_8_9_5
void FASTCALL call_15_8_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_9_6
#define init_15_8_9_6
void FASTCALL call_15_8_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_9_7
#define init_15_8_9_7
void FASTCALL call_15_8_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_9_8
#define init_15_8_9_8
void FASTCALL call_15_8_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_9_9
#define init_15_8_9_9
void FASTCALL call_15_8_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_9_10
#define init_15_8_9_10
void FASTCALL call_15_8_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_9_11
#define init_15_8_9_11
void FASTCALL call_15_8_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_9_12
#define init_15_8_9_12
void FASTCALL call_15_8_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_9_13
#define init_15_8_9_13
void FASTCALL call_15_8_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_9_14
#define init_15_8_9_14
void FASTCALL call_15_8_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_9_15
#define init_15_8_9_15
void FASTCALL call_15_8_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_10_0
#define init_15_8_10_0
void FASTCALL call_15_8_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_10_1
#define init_15_8_10_1
void FASTCALL call_15_8_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_10_2
#define init_15_8_10_2
void FASTCALL call_15_8_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_10_3
#define init_15_8_10_3
void FASTCALL call_15_8_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_10_4
#define init_15_8_10_4
void FASTCALL call_15_8_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_10_5
#define init_15_8_10_5
void FASTCALL call_15_8_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_10_6
#define init_15_8_10_6
void FASTCALL call_15_8_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_10_7
#define init_15_8_10_7
void FASTCALL call_15_8_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_10_8
#define init_15_8_10_8
void FASTCALL call_15_8_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_10_9
#define init_15_8_10_9
void FASTCALL call_15_8_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_10_10
#define init_15_8_10_10
void FASTCALL call_15_8_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_10_11
#define init_15_8_10_11
void FASTCALL call_15_8_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_10_12
#define init_15_8_10_12
void FASTCALL call_15_8_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_10_13
#define init_15_8_10_13
void FASTCALL call_15_8_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_10_14
#define init_15_8_10_14
void FASTCALL call_15_8_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_10_15
#define init_15_8_10_15
void FASTCALL call_15_8_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_11_0
#define init_15_8_11_0
void FASTCALL call_15_8_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_11_1
#define init_15_8_11_1
void FASTCALL call_15_8_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_11_2
#define init_15_8_11_2
void FASTCALL call_15_8_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_11_3
#define init_15_8_11_3
void FASTCALL call_15_8_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_11_4
#define init_15_8_11_4
void FASTCALL call_15_8_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_11_5
#define init_15_8_11_5
void FASTCALL call_15_8_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_11_6
#define init_15_8_11_6
void FASTCALL call_15_8_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_11_7
#define init_15_8_11_7
void FASTCALL call_15_8_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_11_8
#define init_15_8_11_8
void FASTCALL call_15_8_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_11_9
#define init_15_8_11_9
void FASTCALL call_15_8_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_11_10
#define init_15_8_11_10
void FASTCALL call_15_8_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_11_11
#define init_15_8_11_11
void FASTCALL call_15_8_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_11_12
#define init_15_8_11_12
void FASTCALL call_15_8_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_11_13
#define init_15_8_11_13
void FASTCALL call_15_8_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_11_14
#define init_15_8_11_14
void FASTCALL call_15_8_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_11_15
#define init_15_8_11_15
void FASTCALL call_15_8_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_12_0
#define init_15_8_12_0
void FASTCALL call_15_8_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_12_1
#define init_15_8_12_1
void FASTCALL call_15_8_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_12_2
#define init_15_8_12_2
void FASTCALL call_15_8_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_12_3
#define init_15_8_12_3
void FASTCALL call_15_8_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_12_4
#define init_15_8_12_4
void FASTCALL call_15_8_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_12_5
#define init_15_8_12_5
void FASTCALL call_15_8_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_12_6
#define init_15_8_12_6
void FASTCALL call_15_8_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_12_7
#define init_15_8_12_7
void FASTCALL call_15_8_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_12_8
#define init_15_8_12_8
void FASTCALL call_15_8_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_12_9
#define init_15_8_12_9
void FASTCALL call_15_8_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_12_10
#define init_15_8_12_10
void FASTCALL call_15_8_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_12_11
#define init_15_8_12_11
void FASTCALL call_15_8_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_12_12
#define init_15_8_12_12
void FASTCALL call_15_8_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_12_13
#define init_15_8_12_13
void FASTCALL call_15_8_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_12_14
#define init_15_8_12_14
void FASTCALL call_15_8_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_12_15
#define init_15_8_12_15
void FASTCALL call_15_8_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_13_0
#define init_15_8_13_0
void FASTCALL call_15_8_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_13_1
#define init_15_8_13_1
void FASTCALL call_15_8_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_13_2
#define init_15_8_13_2
void FASTCALL call_15_8_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_13_3
#define init_15_8_13_3
void FASTCALL call_15_8_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_13_4
#define init_15_8_13_4
void FASTCALL call_15_8_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_13_5
#define init_15_8_13_5
void FASTCALL call_15_8_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_13_6
#define init_15_8_13_6
void FASTCALL call_15_8_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_13_7
#define init_15_8_13_7
void FASTCALL call_15_8_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_13_8
#define init_15_8_13_8
void FASTCALL call_15_8_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_13_9
#define init_15_8_13_9
void FASTCALL call_15_8_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_13_10
#define init_15_8_13_10
void FASTCALL call_15_8_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_13_11
#define init_15_8_13_11
void FASTCALL call_15_8_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_13_12
#define init_15_8_13_12
void FASTCALL call_15_8_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_13_13
#define init_15_8_13_13
void FASTCALL call_15_8_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_13_14
#define init_15_8_13_14
void FASTCALL call_15_8_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_13_15
#define init_15_8_13_15
void FASTCALL call_15_8_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_14_0
#define init_15_8_14_0
void FASTCALL call_15_8_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_14_1
#define init_15_8_14_1
void FASTCALL call_15_8_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_14_2
#define init_15_8_14_2
void FASTCALL call_15_8_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_14_3
#define init_15_8_14_3
void FASTCALL call_15_8_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_14_4
#define init_15_8_14_4
void FASTCALL call_15_8_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_14_5
#define init_15_8_14_5
void FASTCALL call_15_8_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_14_6
#define init_15_8_14_6
void FASTCALL call_15_8_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_14_7
#define init_15_8_14_7
void FASTCALL call_15_8_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_14_8
#define init_15_8_14_8
void FASTCALL call_15_8_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_14_9
#define init_15_8_14_9
void FASTCALL call_15_8_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_14_10
#define init_15_8_14_10
void FASTCALL call_15_8_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_14_11
#define init_15_8_14_11
void FASTCALL call_15_8_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_14_12
#define init_15_8_14_12
void FASTCALL call_15_8_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_14_13
#define init_15_8_14_13
void FASTCALL call_15_8_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_14_14
#define init_15_8_14_14
void FASTCALL call_15_8_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_14_15
#define init_15_8_14_15
void FASTCALL call_15_8_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_15_0
#define init_15_8_15_0
void FASTCALL call_15_8_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_15_1
#define init_15_8_15_1
void FASTCALL call_15_8_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_15_2
#define init_15_8_15_2
void FASTCALL call_15_8_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_15_3
#define init_15_8_15_3
void FASTCALL call_15_8_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_15_4
#define init_15_8_15_4
void FASTCALL call_15_8_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_15_5
#define init_15_8_15_5
void FASTCALL call_15_8_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_15_6
#define init_15_8_15_6
void FASTCALL call_15_8_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_15_7
#define init_15_8_15_7
void FASTCALL call_15_8_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_15_8
#define init_15_8_15_8
void FASTCALL call_15_8_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_15_9
#define init_15_8_15_9
void FASTCALL call_15_8_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_15_10
#define init_15_8_15_10
void FASTCALL call_15_8_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_15_11
#define init_15_8_15_11
void FASTCALL call_15_8_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_15_12
#define init_15_8_15_12
void FASTCALL call_15_8_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_15_13
#define init_15_8_15_13
void FASTCALL call_15_8_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_15_14
#define init_15_8_15_14
void FASTCALL call_15_8_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_8_15_15
#define init_15_8_15_15
void FASTCALL call_15_8_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_0_0
#define init_15_9_0_0
void FASTCALL call_15_9_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_0_1
#define init_15_9_0_1
void FASTCALL call_15_9_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_0_2
#define init_15_9_0_2
void FASTCALL call_15_9_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_0_3
#define init_15_9_0_3
void FASTCALL call_15_9_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_0_4
#define init_15_9_0_4
void FASTCALL call_15_9_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_0_5
#define init_15_9_0_5
void FASTCALL call_15_9_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_0_6
#define init_15_9_0_6
void FASTCALL call_15_9_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_0_7
#define init_15_9_0_7
void FASTCALL call_15_9_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_0_8
#define init_15_9_0_8
void FASTCALL call_15_9_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_0_9
#define init_15_9_0_9
void FASTCALL call_15_9_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_0_10
#define init_15_9_0_10
void FASTCALL call_15_9_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_0_11
#define init_15_9_0_11
void FASTCALL call_15_9_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_0_12
#define init_15_9_0_12
void FASTCALL call_15_9_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_0_13
#define init_15_9_0_13
void FASTCALL call_15_9_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_0_14
#define init_15_9_0_14
void FASTCALL call_15_9_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_0_15
#define init_15_9_0_15
void FASTCALL call_15_9_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_1_0
#define init_15_9_1_0
void FASTCALL call_15_9_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_1_1
#define init_15_9_1_1
void FASTCALL call_15_9_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_1_2
#define init_15_9_1_2
void FASTCALL call_15_9_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_1_3
#define init_15_9_1_3
void FASTCALL call_15_9_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_1_4
#define init_15_9_1_4
void FASTCALL call_15_9_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_1_5
#define init_15_9_1_5
void FASTCALL call_15_9_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_1_6
#define init_15_9_1_6
void FASTCALL call_15_9_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_1_7
#define init_15_9_1_7
void FASTCALL call_15_9_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_1_8
#define init_15_9_1_8
void FASTCALL call_15_9_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_1_9
#define init_15_9_1_9
void FASTCALL call_15_9_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_1_10
#define init_15_9_1_10
void FASTCALL call_15_9_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_1_11
#define init_15_9_1_11
void FASTCALL call_15_9_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_1_12
#define init_15_9_1_12
void FASTCALL call_15_9_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_1_13
#define init_15_9_1_13
void FASTCALL call_15_9_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_1_14
#define init_15_9_1_14
void FASTCALL call_15_9_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_1_15
#define init_15_9_1_15
void FASTCALL call_15_9_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_2_0
#define init_15_9_2_0
void FASTCALL call_15_9_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_2_1
#define init_15_9_2_1
void FASTCALL call_15_9_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_2_2
#define init_15_9_2_2
void FASTCALL call_15_9_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_2_3
#define init_15_9_2_3
void FASTCALL call_15_9_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_2_4
#define init_15_9_2_4
void FASTCALL call_15_9_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_2_5
#define init_15_9_2_5
void FASTCALL call_15_9_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_2_6
#define init_15_9_2_6
void FASTCALL call_15_9_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_2_7
#define init_15_9_2_7
void FASTCALL call_15_9_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_2_8
#define init_15_9_2_8
void FASTCALL call_15_9_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_2_9
#define init_15_9_2_9
void FASTCALL call_15_9_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_2_10
#define init_15_9_2_10
void FASTCALL call_15_9_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_2_11
#define init_15_9_2_11
void FASTCALL call_15_9_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_2_12
#define init_15_9_2_12
void FASTCALL call_15_9_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_2_13
#define init_15_9_2_13
void FASTCALL call_15_9_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_2_14
#define init_15_9_2_14
void FASTCALL call_15_9_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_2_15
#define init_15_9_2_15
void FASTCALL call_15_9_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_3_0
#define init_15_9_3_0
void FASTCALL call_15_9_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_3_1
#define init_15_9_3_1
void FASTCALL call_15_9_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_3_2
#define init_15_9_3_2
void FASTCALL call_15_9_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_3_3
#define init_15_9_3_3
void FASTCALL call_15_9_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_3_4
#define init_15_9_3_4
void FASTCALL call_15_9_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_3_5
#define init_15_9_3_5
void FASTCALL call_15_9_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_3_6
#define init_15_9_3_6
void FASTCALL call_15_9_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_3_7
#define init_15_9_3_7
void FASTCALL call_15_9_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_3_8
#define init_15_9_3_8
void FASTCALL call_15_9_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_3_9
#define init_15_9_3_9
void FASTCALL call_15_9_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_3_10
#define init_15_9_3_10
void FASTCALL call_15_9_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_3_11
#define init_15_9_3_11
void FASTCALL call_15_9_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_3_12
#define init_15_9_3_12
void FASTCALL call_15_9_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_3_13
#define init_15_9_3_13
void FASTCALL call_15_9_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_3_14
#define init_15_9_3_14
void FASTCALL call_15_9_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_3_15
#define init_15_9_3_15
void FASTCALL call_15_9_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_4_0
#define init_15_9_4_0
void FASTCALL call_15_9_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_4_1
#define init_15_9_4_1
void FASTCALL call_15_9_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_4_2
#define init_15_9_4_2
void FASTCALL call_15_9_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_4_3
#define init_15_9_4_3
void FASTCALL call_15_9_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_4_4
#define init_15_9_4_4
void FASTCALL call_15_9_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_4_5
#define init_15_9_4_5
void FASTCALL call_15_9_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_4_6
#define init_15_9_4_6
void FASTCALL call_15_9_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_4_7
#define init_15_9_4_7
void FASTCALL call_15_9_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_4_8
#define init_15_9_4_8
void FASTCALL call_15_9_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_4_9
#define init_15_9_4_9
void FASTCALL call_15_9_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_4_10
#define init_15_9_4_10
void FASTCALL call_15_9_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_4_11
#define init_15_9_4_11
void FASTCALL call_15_9_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_4_12
#define init_15_9_4_12
void FASTCALL call_15_9_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_4_13
#define init_15_9_4_13
void FASTCALL call_15_9_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_4_14
#define init_15_9_4_14
void FASTCALL call_15_9_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_4_15
#define init_15_9_4_15
void FASTCALL call_15_9_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_5_0
#define init_15_9_5_0
void FASTCALL call_15_9_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_5_1
#define init_15_9_5_1
void FASTCALL call_15_9_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_5_2
#define init_15_9_5_2
void FASTCALL call_15_9_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_5_3
#define init_15_9_5_3
void FASTCALL call_15_9_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_5_4
#define init_15_9_5_4
void FASTCALL call_15_9_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_5_5
#define init_15_9_5_5
void FASTCALL call_15_9_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_5_6
#define init_15_9_5_6
void FASTCALL call_15_9_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_5_7
#define init_15_9_5_7
void FASTCALL call_15_9_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_5_8
#define init_15_9_5_8
void FASTCALL call_15_9_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_5_9
#define init_15_9_5_9
void FASTCALL call_15_9_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_5_10
#define init_15_9_5_10
void FASTCALL call_15_9_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_5_11
#define init_15_9_5_11
void FASTCALL call_15_9_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_5_12
#define init_15_9_5_12
void FASTCALL call_15_9_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_5_13
#define init_15_9_5_13
void FASTCALL call_15_9_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_5_14
#define init_15_9_5_14
void FASTCALL call_15_9_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_5_15
#define init_15_9_5_15
void FASTCALL call_15_9_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_6_0
#define init_15_9_6_0
void FASTCALL call_15_9_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_6_1
#define init_15_9_6_1
void FASTCALL call_15_9_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_6_2
#define init_15_9_6_2
void FASTCALL call_15_9_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_6_3
#define init_15_9_6_3
void FASTCALL call_15_9_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_6_4
#define init_15_9_6_4
void FASTCALL call_15_9_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_6_5
#define init_15_9_6_5
void FASTCALL call_15_9_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_6_6
#define init_15_9_6_6
void FASTCALL call_15_9_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_6_7
#define init_15_9_6_7
void FASTCALL call_15_9_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_6_8
#define init_15_9_6_8
void FASTCALL call_15_9_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_6_9
#define init_15_9_6_9
void FASTCALL call_15_9_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_6_10
#define init_15_9_6_10
void FASTCALL call_15_9_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_6_11
#define init_15_9_6_11
void FASTCALL call_15_9_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_6_12
#define init_15_9_6_12
void FASTCALL call_15_9_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_6_13
#define init_15_9_6_13
void FASTCALL call_15_9_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_6_14
#define init_15_9_6_14
void FASTCALL call_15_9_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_6_15
#define init_15_9_6_15
void FASTCALL call_15_9_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_7_0
#define init_15_9_7_0
void FASTCALL call_15_9_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_7_1
#define init_15_9_7_1
void FASTCALL call_15_9_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_7_2
#define init_15_9_7_2
void FASTCALL call_15_9_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_7_3
#define init_15_9_7_3
void FASTCALL call_15_9_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_7_4
#define init_15_9_7_4
void FASTCALL call_15_9_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_7_5
#define init_15_9_7_5
void FASTCALL call_15_9_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_7_6
#define init_15_9_7_6
void FASTCALL call_15_9_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_7_7
#define init_15_9_7_7
void FASTCALL call_15_9_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_7_8
#define init_15_9_7_8
void FASTCALL call_15_9_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_7_9
#define init_15_9_7_9
void FASTCALL call_15_9_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_7_10
#define init_15_9_7_10
void FASTCALL call_15_9_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_7_11
#define init_15_9_7_11
void FASTCALL call_15_9_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_7_12
#define init_15_9_7_12
void FASTCALL call_15_9_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_7_13
#define init_15_9_7_13
void FASTCALL call_15_9_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_7_14
#define init_15_9_7_14
void FASTCALL call_15_9_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_7_15
#define init_15_9_7_15
void FASTCALL call_15_9_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_8_0
#define init_15_9_8_0
void FASTCALL call_15_9_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_8_1
#define init_15_9_8_1
void FASTCALL call_15_9_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_8_2
#define init_15_9_8_2
void FASTCALL call_15_9_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_8_3
#define init_15_9_8_3
void FASTCALL call_15_9_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_8_4
#define init_15_9_8_4
void FASTCALL call_15_9_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_8_5
#define init_15_9_8_5
void FASTCALL call_15_9_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_8_6
#define init_15_9_8_6
void FASTCALL call_15_9_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_8_7
#define init_15_9_8_7
void FASTCALL call_15_9_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_8_8
#define init_15_9_8_8
void FASTCALL call_15_9_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_8_9
#define init_15_9_8_9
void FASTCALL call_15_9_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_8_10
#define init_15_9_8_10
void FASTCALL call_15_9_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_8_11
#define init_15_9_8_11
void FASTCALL call_15_9_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_8_12
#define init_15_9_8_12
void FASTCALL call_15_9_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_8_13
#define init_15_9_8_13
void FASTCALL call_15_9_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_8_14
#define init_15_9_8_14
void FASTCALL call_15_9_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_8_15
#define init_15_9_8_15
void FASTCALL call_15_9_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_9_0
#define init_15_9_9_0
void FASTCALL call_15_9_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_9_1
#define init_15_9_9_1
void FASTCALL call_15_9_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_9_2
#define init_15_9_9_2
void FASTCALL call_15_9_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_9_3
#define init_15_9_9_3
void FASTCALL call_15_9_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_9_4
#define init_15_9_9_4
void FASTCALL call_15_9_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_9_5
#define init_15_9_9_5
void FASTCALL call_15_9_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_9_6
#define init_15_9_9_6
void FASTCALL call_15_9_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_9_7
#define init_15_9_9_7
void FASTCALL call_15_9_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_9_8
#define init_15_9_9_8
void FASTCALL call_15_9_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_9_9
#define init_15_9_9_9
void FASTCALL call_15_9_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_9_10
#define init_15_9_9_10
void FASTCALL call_15_9_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_9_11
#define init_15_9_9_11
void FASTCALL call_15_9_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_9_12
#define init_15_9_9_12
void FASTCALL call_15_9_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_9_13
#define init_15_9_9_13
void FASTCALL call_15_9_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_9_14
#define init_15_9_9_14
void FASTCALL call_15_9_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_9_15
#define init_15_9_9_15
void FASTCALL call_15_9_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_10_0
#define init_15_9_10_0
void FASTCALL call_15_9_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_10_1
#define init_15_9_10_1
void FASTCALL call_15_9_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_10_2
#define init_15_9_10_2
void FASTCALL call_15_9_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_10_3
#define init_15_9_10_3
void FASTCALL call_15_9_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_10_4
#define init_15_9_10_4
void FASTCALL call_15_9_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_10_5
#define init_15_9_10_5
void FASTCALL call_15_9_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_10_6
#define init_15_9_10_6
void FASTCALL call_15_9_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_10_7
#define init_15_9_10_7
void FASTCALL call_15_9_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_10_8
#define init_15_9_10_8
void FASTCALL call_15_9_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_10_9
#define init_15_9_10_9
void FASTCALL call_15_9_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_10_10
#define init_15_9_10_10
void FASTCALL call_15_9_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_10_11
#define init_15_9_10_11
void FASTCALL call_15_9_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_10_12
#define init_15_9_10_12
void FASTCALL call_15_9_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_10_13
#define init_15_9_10_13
void FASTCALL call_15_9_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_10_14
#define init_15_9_10_14
void FASTCALL call_15_9_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_10_15
#define init_15_9_10_15
void FASTCALL call_15_9_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_11_0
#define init_15_9_11_0
void FASTCALL call_15_9_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_11_1
#define init_15_9_11_1
void FASTCALL call_15_9_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_11_2
#define init_15_9_11_2
void FASTCALL call_15_9_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_11_3
#define init_15_9_11_3
void FASTCALL call_15_9_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_11_4
#define init_15_9_11_4
void FASTCALL call_15_9_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_11_5
#define init_15_9_11_5
void FASTCALL call_15_9_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_11_6
#define init_15_9_11_6
void FASTCALL call_15_9_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_11_7
#define init_15_9_11_7
void FASTCALL call_15_9_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_11_8
#define init_15_9_11_8
void FASTCALL call_15_9_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_11_9
#define init_15_9_11_9
void FASTCALL call_15_9_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_11_10
#define init_15_9_11_10
void FASTCALL call_15_9_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_11_11
#define init_15_9_11_11
void FASTCALL call_15_9_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_11_12
#define init_15_9_11_12
void FASTCALL call_15_9_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_11_13
#define init_15_9_11_13
void FASTCALL call_15_9_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_11_14
#define init_15_9_11_14
void FASTCALL call_15_9_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_11_15
#define init_15_9_11_15
void FASTCALL call_15_9_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_12_0
#define init_15_9_12_0
void FASTCALL call_15_9_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_12_1
#define init_15_9_12_1
void FASTCALL call_15_9_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_12_2
#define init_15_9_12_2
void FASTCALL call_15_9_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_12_3
#define init_15_9_12_3
void FASTCALL call_15_9_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_12_4
#define init_15_9_12_4
void FASTCALL call_15_9_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_12_5
#define init_15_9_12_5
void FASTCALL call_15_9_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_12_6
#define init_15_9_12_6
void FASTCALL call_15_9_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_12_7
#define init_15_9_12_7
void FASTCALL call_15_9_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_12_8
#define init_15_9_12_8
void FASTCALL call_15_9_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_12_9
#define init_15_9_12_9
void FASTCALL call_15_9_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_12_10
#define init_15_9_12_10
void FASTCALL call_15_9_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_12_11
#define init_15_9_12_11
void FASTCALL call_15_9_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_12_12
#define init_15_9_12_12
void FASTCALL call_15_9_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_12_13
#define init_15_9_12_13
void FASTCALL call_15_9_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_12_14
#define init_15_9_12_14
void FASTCALL call_15_9_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_12_15
#define init_15_9_12_15
void FASTCALL call_15_9_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_13_0
#define init_15_9_13_0
void FASTCALL call_15_9_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_13_1
#define init_15_9_13_1
void FASTCALL call_15_9_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_13_2
#define init_15_9_13_2
void FASTCALL call_15_9_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_13_3
#define init_15_9_13_3
void FASTCALL call_15_9_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_13_4
#define init_15_9_13_4
void FASTCALL call_15_9_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_13_5
#define init_15_9_13_5
void FASTCALL call_15_9_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_13_6
#define init_15_9_13_6
void FASTCALL call_15_9_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_13_7
#define init_15_9_13_7
void FASTCALL call_15_9_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_13_8
#define init_15_9_13_8
void FASTCALL call_15_9_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_13_9
#define init_15_9_13_9
void FASTCALL call_15_9_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_13_10
#define init_15_9_13_10
void FASTCALL call_15_9_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_13_11
#define init_15_9_13_11
void FASTCALL call_15_9_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_13_12
#define init_15_9_13_12
void FASTCALL call_15_9_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_13_13
#define init_15_9_13_13
void FASTCALL call_15_9_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_13_14
#define init_15_9_13_14
void FASTCALL call_15_9_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_13_15
#define init_15_9_13_15
void FASTCALL call_15_9_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_14_0
#define init_15_9_14_0
void FASTCALL call_15_9_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_14_1
#define init_15_9_14_1
void FASTCALL call_15_9_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_14_2
#define init_15_9_14_2
void FASTCALL call_15_9_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_14_3
#define init_15_9_14_3
void FASTCALL call_15_9_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_14_4
#define init_15_9_14_4
void FASTCALL call_15_9_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_14_5
#define init_15_9_14_5
void FASTCALL call_15_9_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_14_6
#define init_15_9_14_6
void FASTCALL call_15_9_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_14_7
#define init_15_9_14_7
void FASTCALL call_15_9_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_14_8
#define init_15_9_14_8
void FASTCALL call_15_9_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_14_9
#define init_15_9_14_9
void FASTCALL call_15_9_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_14_10
#define init_15_9_14_10
void FASTCALL call_15_9_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_14_11
#define init_15_9_14_11
void FASTCALL call_15_9_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_14_12
#define init_15_9_14_12
void FASTCALL call_15_9_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_14_13
#define init_15_9_14_13
void FASTCALL call_15_9_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_14_14
#define init_15_9_14_14
void FASTCALL call_15_9_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_14_15
#define init_15_9_14_15
void FASTCALL call_15_9_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_15_0
#define init_15_9_15_0
void FASTCALL call_15_9_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_15_1
#define init_15_9_15_1
void FASTCALL call_15_9_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_15_2
#define init_15_9_15_2
void FASTCALL call_15_9_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_15_3
#define init_15_9_15_3
void FASTCALL call_15_9_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_15_4
#define init_15_9_15_4
void FASTCALL call_15_9_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_15_5
#define init_15_9_15_5
void FASTCALL call_15_9_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_15_6
#define init_15_9_15_6
void FASTCALL call_15_9_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_15_7
#define init_15_9_15_7
void FASTCALL call_15_9_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_15_8
#define init_15_9_15_8
void FASTCALL call_15_9_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_15_9
#define init_15_9_15_9
void FASTCALL call_15_9_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_15_10
#define init_15_9_15_10
void FASTCALL call_15_9_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_15_11
#define init_15_9_15_11
void FASTCALL call_15_9_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_15_12
#define init_15_9_15_12
void FASTCALL call_15_9_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_15_13
#define init_15_9_15_13
void FASTCALL call_15_9_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_15_14
#define init_15_9_15_14
void FASTCALL call_15_9_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_9_15_15
#define init_15_9_15_15
void FASTCALL call_15_9_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_0_0
#define init_15_10_0_0
void FASTCALL call_15_10_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_0_1
#define init_15_10_0_1
void FASTCALL call_15_10_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_0_2
#define init_15_10_0_2
void FASTCALL call_15_10_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_0_3
#define init_15_10_0_3
void FASTCALL call_15_10_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_0_4
#define init_15_10_0_4
void FASTCALL call_15_10_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_0_5
#define init_15_10_0_5
void FASTCALL call_15_10_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_0_6
#define init_15_10_0_6
void FASTCALL call_15_10_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_0_7
#define init_15_10_0_7
void FASTCALL call_15_10_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_0_8
#define init_15_10_0_8
void FASTCALL call_15_10_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_0_9
#define init_15_10_0_9
void FASTCALL call_15_10_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_0_10
#define init_15_10_0_10
void FASTCALL call_15_10_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_0_11
#define init_15_10_0_11
void FASTCALL call_15_10_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_0_12
#define init_15_10_0_12
void FASTCALL call_15_10_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_0_13
#define init_15_10_0_13
void FASTCALL call_15_10_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_0_14
#define init_15_10_0_14
void FASTCALL call_15_10_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_0_15
#define init_15_10_0_15
void FASTCALL call_15_10_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_1_0
#define init_15_10_1_0
void FASTCALL call_15_10_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_1_1
#define init_15_10_1_1
void FASTCALL call_15_10_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_1_2
#define init_15_10_1_2
void FASTCALL call_15_10_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_1_3
#define init_15_10_1_3
void FASTCALL call_15_10_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_1_4
#define init_15_10_1_4
void FASTCALL call_15_10_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_1_5
#define init_15_10_1_5
void FASTCALL call_15_10_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_1_6
#define init_15_10_1_6
void FASTCALL call_15_10_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_1_7
#define init_15_10_1_7
void FASTCALL call_15_10_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_1_8
#define init_15_10_1_8
void FASTCALL call_15_10_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_1_9
#define init_15_10_1_9
void FASTCALL call_15_10_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_1_10
#define init_15_10_1_10
void FASTCALL call_15_10_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_1_11
#define init_15_10_1_11
void FASTCALL call_15_10_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_1_12
#define init_15_10_1_12
void FASTCALL call_15_10_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_1_13
#define init_15_10_1_13
void FASTCALL call_15_10_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_1_14
#define init_15_10_1_14
void FASTCALL call_15_10_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_1_15
#define init_15_10_1_15
void FASTCALL call_15_10_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_2_0
#define init_15_10_2_0
void FASTCALL call_15_10_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_2_1
#define init_15_10_2_1
void FASTCALL call_15_10_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_2_2
#define init_15_10_2_2
void FASTCALL call_15_10_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_2_3
#define init_15_10_2_3
void FASTCALL call_15_10_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_2_4
#define init_15_10_2_4
void FASTCALL call_15_10_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_2_5
#define init_15_10_2_5
void FASTCALL call_15_10_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_2_6
#define init_15_10_2_6
void FASTCALL call_15_10_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_2_7
#define init_15_10_2_7
void FASTCALL call_15_10_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_2_8
#define init_15_10_2_8
void FASTCALL call_15_10_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_2_9
#define init_15_10_2_9
void FASTCALL call_15_10_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_2_10
#define init_15_10_2_10
void FASTCALL call_15_10_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_2_11
#define init_15_10_2_11
void FASTCALL call_15_10_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_2_12
#define init_15_10_2_12
void FASTCALL call_15_10_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_2_13
#define init_15_10_2_13
void FASTCALL call_15_10_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_2_14
#define init_15_10_2_14
void FASTCALL call_15_10_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_2_15
#define init_15_10_2_15
void FASTCALL call_15_10_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_3_0
#define init_15_10_3_0
void FASTCALL call_15_10_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_3_1
#define init_15_10_3_1
void FASTCALL call_15_10_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_3_2
#define init_15_10_3_2
void FASTCALL call_15_10_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_3_3
#define init_15_10_3_3
void FASTCALL call_15_10_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_3_4
#define init_15_10_3_4
void FASTCALL call_15_10_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_3_5
#define init_15_10_3_5
void FASTCALL call_15_10_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_3_6
#define init_15_10_3_6
void FASTCALL call_15_10_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_3_7
#define init_15_10_3_7
void FASTCALL call_15_10_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_3_8
#define init_15_10_3_8
void FASTCALL call_15_10_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_3_9
#define init_15_10_3_9
void FASTCALL call_15_10_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_3_10
#define init_15_10_3_10
void FASTCALL call_15_10_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_3_11
#define init_15_10_3_11
void FASTCALL call_15_10_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_3_12
#define init_15_10_3_12
void FASTCALL call_15_10_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_3_13
#define init_15_10_3_13
void FASTCALL call_15_10_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_3_14
#define init_15_10_3_14
void FASTCALL call_15_10_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_3_15
#define init_15_10_3_15
void FASTCALL call_15_10_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_4_0
#define init_15_10_4_0
void FASTCALL call_15_10_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_4_1
#define init_15_10_4_1
void FASTCALL call_15_10_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_4_2
#define init_15_10_4_2
void FASTCALL call_15_10_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_4_3
#define init_15_10_4_3
void FASTCALL call_15_10_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_4_4
#define init_15_10_4_4
void FASTCALL call_15_10_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_4_5
#define init_15_10_4_5
void FASTCALL call_15_10_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_4_6
#define init_15_10_4_6
void FASTCALL call_15_10_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_4_7
#define init_15_10_4_7
void FASTCALL call_15_10_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_4_8
#define init_15_10_4_8
void FASTCALL call_15_10_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_4_9
#define init_15_10_4_9
void FASTCALL call_15_10_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_4_10
#define init_15_10_4_10
void FASTCALL call_15_10_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_4_11
#define init_15_10_4_11
void FASTCALL call_15_10_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_4_12
#define init_15_10_4_12
void FASTCALL call_15_10_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_4_13
#define init_15_10_4_13
void FASTCALL call_15_10_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_4_14
#define init_15_10_4_14
void FASTCALL call_15_10_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_4_15
#define init_15_10_4_15
void FASTCALL call_15_10_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_5_0
#define init_15_10_5_0
void FASTCALL call_15_10_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_5_1
#define init_15_10_5_1
void FASTCALL call_15_10_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_5_2
#define init_15_10_5_2
void FASTCALL call_15_10_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_5_3
#define init_15_10_5_3
void FASTCALL call_15_10_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_5_4
#define init_15_10_5_4
void FASTCALL call_15_10_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_5_5
#define init_15_10_5_5
void FASTCALL call_15_10_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_5_6
#define init_15_10_5_6
void FASTCALL call_15_10_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_5_7
#define init_15_10_5_7
void FASTCALL call_15_10_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_5_8
#define init_15_10_5_8
void FASTCALL call_15_10_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_5_9
#define init_15_10_5_9
void FASTCALL call_15_10_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_5_10
#define init_15_10_5_10
void FASTCALL call_15_10_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_5_11
#define init_15_10_5_11
void FASTCALL call_15_10_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_5_12
#define init_15_10_5_12
void FASTCALL call_15_10_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_5_13
#define init_15_10_5_13
void FASTCALL call_15_10_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_5_14
#define init_15_10_5_14
void FASTCALL call_15_10_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_5_15
#define init_15_10_5_15
void FASTCALL call_15_10_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_6_0
#define init_15_10_6_0
void FASTCALL call_15_10_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_6_1
#define init_15_10_6_1
void FASTCALL call_15_10_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_6_2
#define init_15_10_6_2
void FASTCALL call_15_10_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_6_3
#define init_15_10_6_3
void FASTCALL call_15_10_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_6_4
#define init_15_10_6_4
void FASTCALL call_15_10_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_6_5
#define init_15_10_6_5
void FASTCALL call_15_10_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_6_6
#define init_15_10_6_6
void FASTCALL call_15_10_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_6_7
#define init_15_10_6_7
void FASTCALL call_15_10_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_6_8
#define init_15_10_6_8
void FASTCALL call_15_10_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_6_9
#define init_15_10_6_9
void FASTCALL call_15_10_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_6_10
#define init_15_10_6_10
void FASTCALL call_15_10_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_6_11
#define init_15_10_6_11
void FASTCALL call_15_10_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_6_12
#define init_15_10_6_12
void FASTCALL call_15_10_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_6_13
#define init_15_10_6_13
void FASTCALL call_15_10_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_6_14
#define init_15_10_6_14
void FASTCALL call_15_10_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_6_15
#define init_15_10_6_15
void FASTCALL call_15_10_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_7_0
#define init_15_10_7_0
void FASTCALL call_15_10_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_7_1
#define init_15_10_7_1
void FASTCALL call_15_10_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_7_2
#define init_15_10_7_2
void FASTCALL call_15_10_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_7_3
#define init_15_10_7_3
void FASTCALL call_15_10_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_7_4
#define init_15_10_7_4
void FASTCALL call_15_10_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_7_5
#define init_15_10_7_5
void FASTCALL call_15_10_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_7_6
#define init_15_10_7_6
void FASTCALL call_15_10_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_7_7
#define init_15_10_7_7
void FASTCALL call_15_10_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_7_8
#define init_15_10_7_8
void FASTCALL call_15_10_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_7_9
#define init_15_10_7_9
void FASTCALL call_15_10_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_7_10
#define init_15_10_7_10
void FASTCALL call_15_10_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_7_11
#define init_15_10_7_11
void FASTCALL call_15_10_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_7_12
#define init_15_10_7_12
void FASTCALL call_15_10_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_7_13
#define init_15_10_7_13
void FASTCALL call_15_10_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_7_14
#define init_15_10_7_14
void FASTCALL call_15_10_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_7_15
#define init_15_10_7_15
void FASTCALL call_15_10_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_8_0
#define init_15_10_8_0
void FASTCALL call_15_10_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_8_1
#define init_15_10_8_1
void FASTCALL call_15_10_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_8_2
#define init_15_10_8_2
void FASTCALL call_15_10_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_8_3
#define init_15_10_8_3
void FASTCALL call_15_10_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_8_4
#define init_15_10_8_4
void FASTCALL call_15_10_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_8_5
#define init_15_10_8_5
void FASTCALL call_15_10_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_8_6
#define init_15_10_8_6
void FASTCALL call_15_10_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_8_7
#define init_15_10_8_7
void FASTCALL call_15_10_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_8_8
#define init_15_10_8_8
void FASTCALL call_15_10_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_8_9
#define init_15_10_8_9
void FASTCALL call_15_10_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_8_10
#define init_15_10_8_10
void FASTCALL call_15_10_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_8_11
#define init_15_10_8_11
void FASTCALL call_15_10_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_8_12
#define init_15_10_8_12
void FASTCALL call_15_10_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_8_13
#define init_15_10_8_13
void FASTCALL call_15_10_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_8_14
#define init_15_10_8_14
void FASTCALL call_15_10_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_8_15
#define init_15_10_8_15
void FASTCALL call_15_10_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_9_0
#define init_15_10_9_0
void FASTCALL call_15_10_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_9_1
#define init_15_10_9_1
void FASTCALL call_15_10_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_9_2
#define init_15_10_9_2
void FASTCALL call_15_10_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_9_3
#define init_15_10_9_3
void FASTCALL call_15_10_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_9_4
#define init_15_10_9_4
void FASTCALL call_15_10_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_9_5
#define init_15_10_9_5
void FASTCALL call_15_10_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_9_6
#define init_15_10_9_6
void FASTCALL call_15_10_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_9_7
#define init_15_10_9_7
void FASTCALL call_15_10_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_9_8
#define init_15_10_9_8
void FASTCALL call_15_10_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_9_9
#define init_15_10_9_9
void FASTCALL call_15_10_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_9_10
#define init_15_10_9_10
void FASTCALL call_15_10_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_9_11
#define init_15_10_9_11
void FASTCALL call_15_10_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_9_12
#define init_15_10_9_12
void FASTCALL call_15_10_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_9_13
#define init_15_10_9_13
void FASTCALL call_15_10_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_9_14
#define init_15_10_9_14
void FASTCALL call_15_10_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_9_15
#define init_15_10_9_15
void FASTCALL call_15_10_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_10_0
#define init_15_10_10_0
void FASTCALL call_15_10_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_10_1
#define init_15_10_10_1
void FASTCALL call_15_10_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_10_2
#define init_15_10_10_2
void FASTCALL call_15_10_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_10_3
#define init_15_10_10_3
void FASTCALL call_15_10_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_10_4
#define init_15_10_10_4
void FASTCALL call_15_10_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_10_5
#define init_15_10_10_5
void FASTCALL call_15_10_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_10_6
#define init_15_10_10_6
void FASTCALL call_15_10_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_10_7
#define init_15_10_10_7
void FASTCALL call_15_10_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_10_8
#define init_15_10_10_8
void FASTCALL call_15_10_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_10_9
#define init_15_10_10_9
void FASTCALL call_15_10_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_10_10
#define init_15_10_10_10
void FASTCALL call_15_10_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_10_11
#define init_15_10_10_11
void FASTCALL call_15_10_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_10_12
#define init_15_10_10_12
void FASTCALL call_15_10_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_10_13
#define init_15_10_10_13
void FASTCALL call_15_10_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_10_14
#define init_15_10_10_14
void FASTCALL call_15_10_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_10_15
#define init_15_10_10_15
void FASTCALL call_15_10_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_11_0
#define init_15_10_11_0
void FASTCALL call_15_10_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_11_1
#define init_15_10_11_1
void FASTCALL call_15_10_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_11_2
#define init_15_10_11_2
void FASTCALL call_15_10_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_11_3
#define init_15_10_11_3
void FASTCALL call_15_10_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_11_4
#define init_15_10_11_4
void FASTCALL call_15_10_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_11_5
#define init_15_10_11_5
void FASTCALL call_15_10_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_11_6
#define init_15_10_11_6
void FASTCALL call_15_10_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_11_7
#define init_15_10_11_7
void FASTCALL call_15_10_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_11_8
#define init_15_10_11_8
void FASTCALL call_15_10_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_11_9
#define init_15_10_11_9
void FASTCALL call_15_10_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_11_10
#define init_15_10_11_10
void FASTCALL call_15_10_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_11_11
#define init_15_10_11_11
void FASTCALL call_15_10_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_11_12
#define init_15_10_11_12
void FASTCALL call_15_10_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_11_13
#define init_15_10_11_13
void FASTCALL call_15_10_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_11_14
#define init_15_10_11_14
void FASTCALL call_15_10_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_11_15
#define init_15_10_11_15
void FASTCALL call_15_10_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_12_0
#define init_15_10_12_0
void FASTCALL call_15_10_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_12_1
#define init_15_10_12_1
void FASTCALL call_15_10_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_12_2
#define init_15_10_12_2
void FASTCALL call_15_10_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_12_3
#define init_15_10_12_3
void FASTCALL call_15_10_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_12_4
#define init_15_10_12_4
void FASTCALL call_15_10_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_12_5
#define init_15_10_12_5
void FASTCALL call_15_10_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_12_6
#define init_15_10_12_6
void FASTCALL call_15_10_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_12_7
#define init_15_10_12_7
void FASTCALL call_15_10_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_12_8
#define init_15_10_12_8
void FASTCALL call_15_10_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_12_9
#define init_15_10_12_9
void FASTCALL call_15_10_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_12_10
#define init_15_10_12_10
void FASTCALL call_15_10_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_12_11
#define init_15_10_12_11
void FASTCALL call_15_10_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_12_12
#define init_15_10_12_12
void FASTCALL call_15_10_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_12_13
#define init_15_10_12_13
void FASTCALL call_15_10_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_12_14
#define init_15_10_12_14
void FASTCALL call_15_10_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_12_15
#define init_15_10_12_15
void FASTCALL call_15_10_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_13_0
#define init_15_10_13_0
void FASTCALL call_15_10_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_13_1
#define init_15_10_13_1
void FASTCALL call_15_10_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_13_2
#define init_15_10_13_2
void FASTCALL call_15_10_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_13_3
#define init_15_10_13_3
void FASTCALL call_15_10_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_13_4
#define init_15_10_13_4
void FASTCALL call_15_10_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_13_5
#define init_15_10_13_5
void FASTCALL call_15_10_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_13_6
#define init_15_10_13_6
void FASTCALL call_15_10_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_13_7
#define init_15_10_13_7
void FASTCALL call_15_10_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_13_8
#define init_15_10_13_8
void FASTCALL call_15_10_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_13_9
#define init_15_10_13_9
void FASTCALL call_15_10_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_13_10
#define init_15_10_13_10
void FASTCALL call_15_10_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_13_11
#define init_15_10_13_11
void FASTCALL call_15_10_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_13_12
#define init_15_10_13_12
void FASTCALL call_15_10_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_13_13
#define init_15_10_13_13
void FASTCALL call_15_10_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_13_14
#define init_15_10_13_14
void FASTCALL call_15_10_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_13_15
#define init_15_10_13_15
void FASTCALL call_15_10_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_14_0
#define init_15_10_14_0
void FASTCALL call_15_10_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_14_1
#define init_15_10_14_1
void FASTCALL call_15_10_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_14_2
#define init_15_10_14_2
void FASTCALL call_15_10_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_14_3
#define init_15_10_14_3
void FASTCALL call_15_10_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_14_4
#define init_15_10_14_4
void FASTCALL call_15_10_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_14_5
#define init_15_10_14_5
void FASTCALL call_15_10_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_14_6
#define init_15_10_14_6
void FASTCALL call_15_10_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_14_7
#define init_15_10_14_7
void FASTCALL call_15_10_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_14_8
#define init_15_10_14_8
void FASTCALL call_15_10_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_14_9
#define init_15_10_14_9
void FASTCALL call_15_10_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_14_10
#define init_15_10_14_10
void FASTCALL call_15_10_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_14_11
#define init_15_10_14_11
void FASTCALL call_15_10_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_14_12
#define init_15_10_14_12
void FASTCALL call_15_10_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_14_13
#define init_15_10_14_13
void FASTCALL call_15_10_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_14_14
#define init_15_10_14_14
void FASTCALL call_15_10_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_14_15
#define init_15_10_14_15
void FASTCALL call_15_10_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_15_0
#define init_15_10_15_0
void FASTCALL call_15_10_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_15_1
#define init_15_10_15_1
void FASTCALL call_15_10_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_15_2
#define init_15_10_15_2
void FASTCALL call_15_10_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_15_3
#define init_15_10_15_3
void FASTCALL call_15_10_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_15_4
#define init_15_10_15_4
void FASTCALL call_15_10_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_15_5
#define init_15_10_15_5
void FASTCALL call_15_10_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_15_6
#define init_15_10_15_6
void FASTCALL call_15_10_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_15_7
#define init_15_10_15_7
void FASTCALL call_15_10_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_15_8
#define init_15_10_15_8
void FASTCALL call_15_10_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_15_9
#define init_15_10_15_9
void FASTCALL call_15_10_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_15_10
#define init_15_10_15_10
void FASTCALL call_15_10_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_15_11
#define init_15_10_15_11
void FASTCALL call_15_10_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_15_12
#define init_15_10_15_12
void FASTCALL call_15_10_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_15_13
#define init_15_10_15_13
void FASTCALL call_15_10_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_15_14
#define init_15_10_15_14
void FASTCALL call_15_10_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_10_15_15
#define init_15_10_15_15
void FASTCALL call_15_10_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_0_0
#define init_15_11_0_0
void FASTCALL call_15_11_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_0_1
#define init_15_11_0_1
void FASTCALL call_15_11_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_0_2
#define init_15_11_0_2
void FASTCALL call_15_11_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_0_3
#define init_15_11_0_3
void FASTCALL call_15_11_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_0_4
#define init_15_11_0_4
void FASTCALL call_15_11_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_0_5
#define init_15_11_0_5
void FASTCALL call_15_11_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_0_6
#define init_15_11_0_6
void FASTCALL call_15_11_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_0_7
#define init_15_11_0_7
void FASTCALL call_15_11_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_0_8
#define init_15_11_0_8
void FASTCALL call_15_11_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_0_9
#define init_15_11_0_9
void FASTCALL call_15_11_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_0_10
#define init_15_11_0_10
void FASTCALL call_15_11_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_0_11
#define init_15_11_0_11
void FASTCALL call_15_11_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_0_12
#define init_15_11_0_12
void FASTCALL call_15_11_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_0_13
#define init_15_11_0_13
void FASTCALL call_15_11_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_0_14
#define init_15_11_0_14
void FASTCALL call_15_11_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_0_15
#define init_15_11_0_15
void FASTCALL call_15_11_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_1_0
#define init_15_11_1_0
void FASTCALL call_15_11_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_1_1
#define init_15_11_1_1
void FASTCALL call_15_11_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_1_2
#define init_15_11_1_2
void FASTCALL call_15_11_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_1_3
#define init_15_11_1_3
void FASTCALL call_15_11_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_1_4
#define init_15_11_1_4
void FASTCALL call_15_11_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_1_5
#define init_15_11_1_5
void FASTCALL call_15_11_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_1_6
#define init_15_11_1_6
void FASTCALL call_15_11_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_1_7
#define init_15_11_1_7
void FASTCALL call_15_11_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_1_8
#define init_15_11_1_8
void FASTCALL call_15_11_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_1_9
#define init_15_11_1_9
void FASTCALL call_15_11_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_1_10
#define init_15_11_1_10
void FASTCALL call_15_11_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_1_11
#define init_15_11_1_11
void FASTCALL call_15_11_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_1_12
#define init_15_11_1_12
void FASTCALL call_15_11_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_1_13
#define init_15_11_1_13
void FASTCALL call_15_11_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_1_14
#define init_15_11_1_14
void FASTCALL call_15_11_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_1_15
#define init_15_11_1_15
void FASTCALL call_15_11_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_2_0
#define init_15_11_2_0
void FASTCALL call_15_11_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_2_1
#define init_15_11_2_1
void FASTCALL call_15_11_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_2_2
#define init_15_11_2_2
void FASTCALL call_15_11_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_2_3
#define init_15_11_2_3
void FASTCALL call_15_11_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_2_4
#define init_15_11_2_4
void FASTCALL call_15_11_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_2_5
#define init_15_11_2_5
void FASTCALL call_15_11_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_2_6
#define init_15_11_2_6
void FASTCALL call_15_11_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_2_7
#define init_15_11_2_7
void FASTCALL call_15_11_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_2_8
#define init_15_11_2_8
void FASTCALL call_15_11_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_2_9
#define init_15_11_2_9
void FASTCALL call_15_11_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_2_10
#define init_15_11_2_10
void FASTCALL call_15_11_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_2_11
#define init_15_11_2_11
void FASTCALL call_15_11_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_2_12
#define init_15_11_2_12
void FASTCALL call_15_11_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_2_13
#define init_15_11_2_13
void FASTCALL call_15_11_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_2_14
#define init_15_11_2_14
void FASTCALL call_15_11_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_2_15
#define init_15_11_2_15
void FASTCALL call_15_11_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_3_0
#define init_15_11_3_0
void FASTCALL call_15_11_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_3_1
#define init_15_11_3_1
void FASTCALL call_15_11_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_3_2
#define init_15_11_3_2
void FASTCALL call_15_11_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_3_3
#define init_15_11_3_3
void FASTCALL call_15_11_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_3_4
#define init_15_11_3_4
void FASTCALL call_15_11_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_3_5
#define init_15_11_3_5
void FASTCALL call_15_11_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_3_6
#define init_15_11_3_6
void FASTCALL call_15_11_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_3_7
#define init_15_11_3_7
void FASTCALL call_15_11_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_3_8
#define init_15_11_3_8
void FASTCALL call_15_11_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_3_9
#define init_15_11_3_9
void FASTCALL call_15_11_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_3_10
#define init_15_11_3_10
void FASTCALL call_15_11_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_3_11
#define init_15_11_3_11
void FASTCALL call_15_11_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_3_12
#define init_15_11_3_12
void FASTCALL call_15_11_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_3_13
#define init_15_11_3_13
void FASTCALL call_15_11_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_3_14
#define init_15_11_3_14
void FASTCALL call_15_11_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_3_15
#define init_15_11_3_15
void FASTCALL call_15_11_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_4_0
#define init_15_11_4_0
void FASTCALL call_15_11_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_4_1
#define init_15_11_4_1
void FASTCALL call_15_11_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_4_2
#define init_15_11_4_2
void FASTCALL call_15_11_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_4_3
#define init_15_11_4_3
void FASTCALL call_15_11_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_4_4
#define init_15_11_4_4
void FASTCALL call_15_11_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_4_5
#define init_15_11_4_5
void FASTCALL call_15_11_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_4_6
#define init_15_11_4_6
void FASTCALL call_15_11_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_4_7
#define init_15_11_4_7
void FASTCALL call_15_11_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_4_8
#define init_15_11_4_8
void FASTCALL call_15_11_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_4_9
#define init_15_11_4_9
void FASTCALL call_15_11_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_4_10
#define init_15_11_4_10
void FASTCALL call_15_11_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_4_11
#define init_15_11_4_11
void FASTCALL call_15_11_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_4_12
#define init_15_11_4_12
void FASTCALL call_15_11_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_4_13
#define init_15_11_4_13
void FASTCALL call_15_11_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_4_14
#define init_15_11_4_14
void FASTCALL call_15_11_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_4_15
#define init_15_11_4_15
void FASTCALL call_15_11_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_5_0
#define init_15_11_5_0
void FASTCALL call_15_11_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_5_1
#define init_15_11_5_1
void FASTCALL call_15_11_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_5_2
#define init_15_11_5_2
void FASTCALL call_15_11_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_5_3
#define init_15_11_5_3
void FASTCALL call_15_11_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_5_4
#define init_15_11_5_4
void FASTCALL call_15_11_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_5_5
#define init_15_11_5_5
void FASTCALL call_15_11_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_5_6
#define init_15_11_5_6
void FASTCALL call_15_11_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_5_7
#define init_15_11_5_7
void FASTCALL call_15_11_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_5_8
#define init_15_11_5_8
void FASTCALL call_15_11_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_5_9
#define init_15_11_5_9
void FASTCALL call_15_11_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_5_10
#define init_15_11_5_10
void FASTCALL call_15_11_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_5_11
#define init_15_11_5_11
void FASTCALL call_15_11_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_5_12
#define init_15_11_5_12
void FASTCALL call_15_11_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_5_13
#define init_15_11_5_13
void FASTCALL call_15_11_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_5_14
#define init_15_11_5_14
void FASTCALL call_15_11_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_5_15
#define init_15_11_5_15
void FASTCALL call_15_11_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_6_0
#define init_15_11_6_0
void FASTCALL call_15_11_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_6_1
#define init_15_11_6_1
void FASTCALL call_15_11_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_6_2
#define init_15_11_6_2
void FASTCALL call_15_11_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_6_3
#define init_15_11_6_3
void FASTCALL call_15_11_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_6_4
#define init_15_11_6_4
void FASTCALL call_15_11_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_6_5
#define init_15_11_6_5
void FASTCALL call_15_11_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_6_6
#define init_15_11_6_6
void FASTCALL call_15_11_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_6_7
#define init_15_11_6_7
void FASTCALL call_15_11_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_6_8
#define init_15_11_6_8
void FASTCALL call_15_11_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_6_9
#define init_15_11_6_9
void FASTCALL call_15_11_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_6_10
#define init_15_11_6_10
void FASTCALL call_15_11_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_6_11
#define init_15_11_6_11
void FASTCALL call_15_11_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_6_12
#define init_15_11_6_12
void FASTCALL call_15_11_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_6_13
#define init_15_11_6_13
void FASTCALL call_15_11_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_6_14
#define init_15_11_6_14
void FASTCALL call_15_11_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_6_15
#define init_15_11_6_15
void FASTCALL call_15_11_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_7_0
#define init_15_11_7_0
void FASTCALL call_15_11_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_7_1
#define init_15_11_7_1
void FASTCALL call_15_11_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_7_2
#define init_15_11_7_2
void FASTCALL call_15_11_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_7_3
#define init_15_11_7_3
void FASTCALL call_15_11_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_7_4
#define init_15_11_7_4
void FASTCALL call_15_11_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_7_5
#define init_15_11_7_5
void FASTCALL call_15_11_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_7_6
#define init_15_11_7_6
void FASTCALL call_15_11_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_7_7
#define init_15_11_7_7
void FASTCALL call_15_11_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_7_8
#define init_15_11_7_8
void FASTCALL call_15_11_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_7_9
#define init_15_11_7_9
void FASTCALL call_15_11_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_7_10
#define init_15_11_7_10
void FASTCALL call_15_11_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_7_11
#define init_15_11_7_11
void FASTCALL call_15_11_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_7_12
#define init_15_11_7_12
void FASTCALL call_15_11_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_7_13
#define init_15_11_7_13
void FASTCALL call_15_11_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_7_14
#define init_15_11_7_14
void FASTCALL call_15_11_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_7_15
#define init_15_11_7_15
void FASTCALL call_15_11_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_8_0
#define init_15_11_8_0
void FASTCALL call_15_11_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_8_1
#define init_15_11_8_1
void FASTCALL call_15_11_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_8_2
#define init_15_11_8_2
void FASTCALL call_15_11_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_8_3
#define init_15_11_8_3
void FASTCALL call_15_11_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_8_4
#define init_15_11_8_4
void FASTCALL call_15_11_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_8_5
#define init_15_11_8_5
void FASTCALL call_15_11_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_8_6
#define init_15_11_8_6
void FASTCALL call_15_11_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_8_7
#define init_15_11_8_7
void FASTCALL call_15_11_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_8_8
#define init_15_11_8_8
void FASTCALL call_15_11_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_8_9
#define init_15_11_8_9
void FASTCALL call_15_11_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_8_10
#define init_15_11_8_10
void FASTCALL call_15_11_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_8_11
#define init_15_11_8_11
void FASTCALL call_15_11_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_8_12
#define init_15_11_8_12
void FASTCALL call_15_11_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_8_13
#define init_15_11_8_13
void FASTCALL call_15_11_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_8_14
#define init_15_11_8_14
void FASTCALL call_15_11_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_8_15
#define init_15_11_8_15
void FASTCALL call_15_11_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_9_0
#define init_15_11_9_0
void FASTCALL call_15_11_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_9_1
#define init_15_11_9_1
void FASTCALL call_15_11_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_9_2
#define init_15_11_9_2
void FASTCALL call_15_11_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_9_3
#define init_15_11_9_3
void FASTCALL call_15_11_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_9_4
#define init_15_11_9_4
void FASTCALL call_15_11_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_9_5
#define init_15_11_9_5
void FASTCALL call_15_11_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_9_6
#define init_15_11_9_6
void FASTCALL call_15_11_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_9_7
#define init_15_11_9_7
void FASTCALL call_15_11_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_9_8
#define init_15_11_9_8
void FASTCALL call_15_11_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_9_9
#define init_15_11_9_9
void FASTCALL call_15_11_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_9_10
#define init_15_11_9_10
void FASTCALL call_15_11_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_9_11
#define init_15_11_9_11
void FASTCALL call_15_11_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_9_12
#define init_15_11_9_12
void FASTCALL call_15_11_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_9_13
#define init_15_11_9_13
void FASTCALL call_15_11_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_9_14
#define init_15_11_9_14
void FASTCALL call_15_11_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_9_15
#define init_15_11_9_15
void FASTCALL call_15_11_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_10_0
#define init_15_11_10_0
void FASTCALL call_15_11_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_10_1
#define init_15_11_10_1
void FASTCALL call_15_11_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_10_2
#define init_15_11_10_2
void FASTCALL call_15_11_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_10_3
#define init_15_11_10_3
void FASTCALL call_15_11_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_10_4
#define init_15_11_10_4
void FASTCALL call_15_11_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_10_5
#define init_15_11_10_5
void FASTCALL call_15_11_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_10_6
#define init_15_11_10_6
void FASTCALL call_15_11_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_10_7
#define init_15_11_10_7
void FASTCALL call_15_11_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_10_8
#define init_15_11_10_8
void FASTCALL call_15_11_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_10_9
#define init_15_11_10_9
void FASTCALL call_15_11_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_10_10
#define init_15_11_10_10
void FASTCALL call_15_11_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_10_11
#define init_15_11_10_11
void FASTCALL call_15_11_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_10_12
#define init_15_11_10_12
void FASTCALL call_15_11_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_10_13
#define init_15_11_10_13
void FASTCALL call_15_11_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_10_14
#define init_15_11_10_14
void FASTCALL call_15_11_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_10_15
#define init_15_11_10_15
void FASTCALL call_15_11_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_11_0
#define init_15_11_11_0
void FASTCALL call_15_11_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_11_1
#define init_15_11_11_1
void FASTCALL call_15_11_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_11_2
#define init_15_11_11_2
void FASTCALL call_15_11_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_11_3
#define init_15_11_11_3
void FASTCALL call_15_11_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_11_4
#define init_15_11_11_4
void FASTCALL call_15_11_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_11_5
#define init_15_11_11_5
void FASTCALL call_15_11_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_11_6
#define init_15_11_11_6
void FASTCALL call_15_11_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_11_7
#define init_15_11_11_7
void FASTCALL call_15_11_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_11_8
#define init_15_11_11_8
void FASTCALL call_15_11_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_11_9
#define init_15_11_11_9
void FASTCALL call_15_11_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_11_10
#define init_15_11_11_10
void FASTCALL call_15_11_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_11_11
#define init_15_11_11_11
void FASTCALL call_15_11_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_11_12
#define init_15_11_11_12
void FASTCALL call_15_11_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_11_13
#define init_15_11_11_13
void FASTCALL call_15_11_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_11_14
#define init_15_11_11_14
void FASTCALL call_15_11_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_11_15
#define init_15_11_11_15
void FASTCALL call_15_11_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_12_0
#define init_15_11_12_0
void FASTCALL call_15_11_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_12_1
#define init_15_11_12_1
void FASTCALL call_15_11_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_12_2
#define init_15_11_12_2
void FASTCALL call_15_11_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_12_3
#define init_15_11_12_3
void FASTCALL call_15_11_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_12_4
#define init_15_11_12_4
void FASTCALL call_15_11_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_12_5
#define init_15_11_12_5
void FASTCALL call_15_11_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_12_6
#define init_15_11_12_6
void FASTCALL call_15_11_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_12_7
#define init_15_11_12_7
void FASTCALL call_15_11_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_12_8
#define init_15_11_12_8
void FASTCALL call_15_11_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_12_9
#define init_15_11_12_9
void FASTCALL call_15_11_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_12_10
#define init_15_11_12_10
void FASTCALL call_15_11_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_12_11
#define init_15_11_12_11
void FASTCALL call_15_11_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_12_12
#define init_15_11_12_12
void FASTCALL call_15_11_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_12_13
#define init_15_11_12_13
void FASTCALL call_15_11_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_12_14
#define init_15_11_12_14
void FASTCALL call_15_11_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_12_15
#define init_15_11_12_15
void FASTCALL call_15_11_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_13_0
#define init_15_11_13_0
void FASTCALL call_15_11_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_13_1
#define init_15_11_13_1
void FASTCALL call_15_11_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_13_2
#define init_15_11_13_2
void FASTCALL call_15_11_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_13_3
#define init_15_11_13_3
void FASTCALL call_15_11_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_13_4
#define init_15_11_13_4
void FASTCALL call_15_11_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_13_5
#define init_15_11_13_5
void FASTCALL call_15_11_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_13_6
#define init_15_11_13_6
void FASTCALL call_15_11_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_13_7
#define init_15_11_13_7
void FASTCALL call_15_11_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_13_8
#define init_15_11_13_8
void FASTCALL call_15_11_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_13_9
#define init_15_11_13_9
void FASTCALL call_15_11_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_13_10
#define init_15_11_13_10
void FASTCALL call_15_11_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_13_11
#define init_15_11_13_11
void FASTCALL call_15_11_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_13_12
#define init_15_11_13_12
void FASTCALL call_15_11_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_13_13
#define init_15_11_13_13
void FASTCALL call_15_11_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_13_14
#define init_15_11_13_14
void FASTCALL call_15_11_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_13_15
#define init_15_11_13_15
void FASTCALL call_15_11_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_14_0
#define init_15_11_14_0
void FASTCALL call_15_11_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_14_1
#define init_15_11_14_1
void FASTCALL call_15_11_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_14_2
#define init_15_11_14_2
void FASTCALL call_15_11_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_14_3
#define init_15_11_14_3
void FASTCALL call_15_11_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_14_4
#define init_15_11_14_4
void FASTCALL call_15_11_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_14_5
#define init_15_11_14_5
void FASTCALL call_15_11_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_14_6
#define init_15_11_14_6
void FASTCALL call_15_11_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_14_7
#define init_15_11_14_7
void FASTCALL call_15_11_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_14_8
#define init_15_11_14_8
void FASTCALL call_15_11_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_14_9
#define init_15_11_14_9
void FASTCALL call_15_11_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_14_10
#define init_15_11_14_10
void FASTCALL call_15_11_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_14_11
#define init_15_11_14_11
void FASTCALL call_15_11_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_14_12
#define init_15_11_14_12
void FASTCALL call_15_11_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_14_13
#define init_15_11_14_13
void FASTCALL call_15_11_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_14_14
#define init_15_11_14_14
void FASTCALL call_15_11_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_14_15
#define init_15_11_14_15
void FASTCALL call_15_11_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_15_0
#define init_15_11_15_0
void FASTCALL call_15_11_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_15_1
#define init_15_11_15_1
void FASTCALL call_15_11_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_15_2
#define init_15_11_15_2
void FASTCALL call_15_11_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_15_3
#define init_15_11_15_3
void FASTCALL call_15_11_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_15_4
#define init_15_11_15_4
void FASTCALL call_15_11_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_15_5
#define init_15_11_15_5
void FASTCALL call_15_11_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_15_6
#define init_15_11_15_6
void FASTCALL call_15_11_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_15_7
#define init_15_11_15_7
void FASTCALL call_15_11_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_15_8
#define init_15_11_15_8
void FASTCALL call_15_11_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_15_9
#define init_15_11_15_9
void FASTCALL call_15_11_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_15_10
#define init_15_11_15_10
void FASTCALL call_15_11_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_15_11
#define init_15_11_15_11
void FASTCALL call_15_11_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_15_12
#define init_15_11_15_12
void FASTCALL call_15_11_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_15_13
#define init_15_11_15_13
void FASTCALL call_15_11_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_15_14
#define init_15_11_15_14
void FASTCALL call_15_11_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_11_15_15
#define init_15_11_15_15
void FASTCALL call_15_11_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_0_0
#define init_15_12_0_0
void FASTCALL call_15_12_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_0_1
#define init_15_12_0_1
void FASTCALL call_15_12_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_0_2
#define init_15_12_0_2
void FASTCALL call_15_12_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_0_3
#define init_15_12_0_3
void FASTCALL call_15_12_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_0_4
#define init_15_12_0_4
void FASTCALL call_15_12_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_0_5
#define init_15_12_0_5
void FASTCALL call_15_12_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_0_6
#define init_15_12_0_6
void FASTCALL call_15_12_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_0_7
#define init_15_12_0_7
void FASTCALL call_15_12_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_0_8
#define init_15_12_0_8
void FASTCALL call_15_12_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_0_9
#define init_15_12_0_9
void FASTCALL call_15_12_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_0_10
#define init_15_12_0_10
void FASTCALL call_15_12_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_0_11
#define init_15_12_0_11
void FASTCALL call_15_12_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_0_12
#define init_15_12_0_12
void FASTCALL call_15_12_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_0_13
#define init_15_12_0_13
void FASTCALL call_15_12_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_0_14
#define init_15_12_0_14
void FASTCALL call_15_12_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_0_15
#define init_15_12_0_15
void FASTCALL call_15_12_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_1_0
#define init_15_12_1_0
void FASTCALL call_15_12_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_1_1
#define init_15_12_1_1
void FASTCALL call_15_12_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_1_2
#define init_15_12_1_2
void FASTCALL call_15_12_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_1_3
#define init_15_12_1_3
void FASTCALL call_15_12_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_1_4
#define init_15_12_1_4
void FASTCALL call_15_12_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_1_5
#define init_15_12_1_5
void FASTCALL call_15_12_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_1_6
#define init_15_12_1_6
void FASTCALL call_15_12_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_1_7
#define init_15_12_1_7
void FASTCALL call_15_12_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_1_8
#define init_15_12_1_8
void FASTCALL call_15_12_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_1_9
#define init_15_12_1_9
void FASTCALL call_15_12_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_1_10
#define init_15_12_1_10
void FASTCALL call_15_12_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_1_11
#define init_15_12_1_11
void FASTCALL call_15_12_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_1_12
#define init_15_12_1_12
void FASTCALL call_15_12_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_1_13
#define init_15_12_1_13
void FASTCALL call_15_12_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_1_14
#define init_15_12_1_14
void FASTCALL call_15_12_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_1_15
#define init_15_12_1_15
void FASTCALL call_15_12_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_2_0
#define init_15_12_2_0
void FASTCALL call_15_12_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_2_1
#define init_15_12_2_1
void FASTCALL call_15_12_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_2_2
#define init_15_12_2_2
void FASTCALL call_15_12_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_2_3
#define init_15_12_2_3
void FASTCALL call_15_12_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_2_4
#define init_15_12_2_4
void FASTCALL call_15_12_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_2_5
#define init_15_12_2_5
void FASTCALL call_15_12_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_2_6
#define init_15_12_2_6
void FASTCALL call_15_12_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_2_7
#define init_15_12_2_7
void FASTCALL call_15_12_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_2_8
#define init_15_12_2_8
void FASTCALL call_15_12_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_2_9
#define init_15_12_2_9
void FASTCALL call_15_12_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_2_10
#define init_15_12_2_10
void FASTCALL call_15_12_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_2_11
#define init_15_12_2_11
void FASTCALL call_15_12_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_2_12
#define init_15_12_2_12
void FASTCALL call_15_12_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_2_13
#define init_15_12_2_13
void FASTCALL call_15_12_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_2_14
#define init_15_12_2_14
void FASTCALL call_15_12_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_2_15
#define init_15_12_2_15
void FASTCALL call_15_12_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_3_0
#define init_15_12_3_0
void FASTCALL call_15_12_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_3_1
#define init_15_12_3_1
void FASTCALL call_15_12_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_3_2
#define init_15_12_3_2
void FASTCALL call_15_12_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_3_3
#define init_15_12_3_3
void FASTCALL call_15_12_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_3_4
#define init_15_12_3_4
void FASTCALL call_15_12_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_3_5
#define init_15_12_3_5
void FASTCALL call_15_12_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_3_6
#define init_15_12_3_6
void FASTCALL call_15_12_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_3_7
#define init_15_12_3_7
void FASTCALL call_15_12_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_3_8
#define init_15_12_3_8
void FASTCALL call_15_12_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_3_9
#define init_15_12_3_9
void FASTCALL call_15_12_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_3_10
#define init_15_12_3_10
void FASTCALL call_15_12_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_3_11
#define init_15_12_3_11
void FASTCALL call_15_12_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_3_12
#define init_15_12_3_12
void FASTCALL call_15_12_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_3_13
#define init_15_12_3_13
void FASTCALL call_15_12_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_3_14
#define init_15_12_3_14
void FASTCALL call_15_12_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_3_15
#define init_15_12_3_15
void FASTCALL call_15_12_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_4_0
#define init_15_12_4_0
void FASTCALL call_15_12_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_4_1
#define init_15_12_4_1
void FASTCALL call_15_12_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_4_2
#define init_15_12_4_2
void FASTCALL call_15_12_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_4_3
#define init_15_12_4_3
void FASTCALL call_15_12_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_4_4
#define init_15_12_4_4
void FASTCALL call_15_12_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_4_5
#define init_15_12_4_5
void FASTCALL call_15_12_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_4_6
#define init_15_12_4_6
void FASTCALL call_15_12_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_4_7
#define init_15_12_4_7
void FASTCALL call_15_12_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_4_8
#define init_15_12_4_8
void FASTCALL call_15_12_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_4_9
#define init_15_12_4_9
void FASTCALL call_15_12_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_4_10
#define init_15_12_4_10
void FASTCALL call_15_12_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_4_11
#define init_15_12_4_11
void FASTCALL call_15_12_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_4_12
#define init_15_12_4_12
void FASTCALL call_15_12_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_4_13
#define init_15_12_4_13
void FASTCALL call_15_12_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_4_14
#define init_15_12_4_14
void FASTCALL call_15_12_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_4_15
#define init_15_12_4_15
void FASTCALL call_15_12_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_5_0
#define init_15_12_5_0
void FASTCALL call_15_12_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_5_1
#define init_15_12_5_1
void FASTCALL call_15_12_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_5_2
#define init_15_12_5_2
void FASTCALL call_15_12_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_5_3
#define init_15_12_5_3
void FASTCALL call_15_12_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_5_4
#define init_15_12_5_4
void FASTCALL call_15_12_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_5_5
#define init_15_12_5_5
void FASTCALL call_15_12_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_5_6
#define init_15_12_5_6
void FASTCALL call_15_12_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_5_7
#define init_15_12_5_7
void FASTCALL call_15_12_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_5_8
#define init_15_12_5_8
void FASTCALL call_15_12_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_5_9
#define init_15_12_5_9
void FASTCALL call_15_12_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_5_10
#define init_15_12_5_10
void FASTCALL call_15_12_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_5_11
#define init_15_12_5_11
void FASTCALL call_15_12_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_5_12
#define init_15_12_5_12
void FASTCALL call_15_12_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_5_13
#define init_15_12_5_13
void FASTCALL call_15_12_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_5_14
#define init_15_12_5_14
void FASTCALL call_15_12_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_5_15
#define init_15_12_5_15
void FASTCALL call_15_12_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_6_0
#define init_15_12_6_0
void FASTCALL call_15_12_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_6_1
#define init_15_12_6_1
void FASTCALL call_15_12_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_6_2
#define init_15_12_6_2
void FASTCALL call_15_12_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_6_3
#define init_15_12_6_3
void FASTCALL call_15_12_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_6_4
#define init_15_12_6_4
void FASTCALL call_15_12_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_6_5
#define init_15_12_6_5
void FASTCALL call_15_12_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_6_6
#define init_15_12_6_6
void FASTCALL call_15_12_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_6_7
#define init_15_12_6_7
void FASTCALL call_15_12_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_6_8
#define init_15_12_6_8
void FASTCALL call_15_12_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_6_9
#define init_15_12_6_9
void FASTCALL call_15_12_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_6_10
#define init_15_12_6_10
void FASTCALL call_15_12_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_6_11
#define init_15_12_6_11
void FASTCALL call_15_12_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_6_12
#define init_15_12_6_12
void FASTCALL call_15_12_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_6_13
#define init_15_12_6_13
void FASTCALL call_15_12_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_6_14
#define init_15_12_6_14
void FASTCALL call_15_12_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_6_15
#define init_15_12_6_15
void FASTCALL call_15_12_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_7_0
#define init_15_12_7_0
void FASTCALL call_15_12_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_7_1
#define init_15_12_7_1
void FASTCALL call_15_12_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_7_2
#define init_15_12_7_2
void FASTCALL call_15_12_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_7_3
#define init_15_12_7_3
void FASTCALL call_15_12_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_7_4
#define init_15_12_7_4
void FASTCALL call_15_12_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_7_5
#define init_15_12_7_5
void FASTCALL call_15_12_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_7_6
#define init_15_12_7_6
void FASTCALL call_15_12_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_7_7
#define init_15_12_7_7
void FASTCALL call_15_12_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_7_8
#define init_15_12_7_8
void FASTCALL call_15_12_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_7_9
#define init_15_12_7_9
void FASTCALL call_15_12_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_7_10
#define init_15_12_7_10
void FASTCALL call_15_12_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_7_11
#define init_15_12_7_11
void FASTCALL call_15_12_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_7_12
#define init_15_12_7_12
void FASTCALL call_15_12_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_7_13
#define init_15_12_7_13
void FASTCALL call_15_12_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_7_14
#define init_15_12_7_14
void FASTCALL call_15_12_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_7_15
#define init_15_12_7_15
void FASTCALL call_15_12_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_8_0
#define init_15_12_8_0
void FASTCALL call_15_12_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_8_1
#define init_15_12_8_1
void FASTCALL call_15_12_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_8_2
#define init_15_12_8_2
void FASTCALL call_15_12_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_8_3
#define init_15_12_8_3
void FASTCALL call_15_12_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_8_4
#define init_15_12_8_4
void FASTCALL call_15_12_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_8_5
#define init_15_12_8_5
void FASTCALL call_15_12_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_8_6
#define init_15_12_8_6
void FASTCALL call_15_12_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_8_7
#define init_15_12_8_7
void FASTCALL call_15_12_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_8_8
#define init_15_12_8_8
void FASTCALL call_15_12_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_8_9
#define init_15_12_8_9
void FASTCALL call_15_12_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_8_10
#define init_15_12_8_10
void FASTCALL call_15_12_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_8_11
#define init_15_12_8_11
void FASTCALL call_15_12_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_8_12
#define init_15_12_8_12
void FASTCALL call_15_12_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_8_13
#define init_15_12_8_13
void FASTCALL call_15_12_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_8_14
#define init_15_12_8_14
void FASTCALL call_15_12_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_8_15
#define init_15_12_8_15
void FASTCALL call_15_12_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_9_0
#define init_15_12_9_0
void FASTCALL call_15_12_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_9_1
#define init_15_12_9_1
void FASTCALL call_15_12_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_9_2
#define init_15_12_9_2
void FASTCALL call_15_12_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_9_3
#define init_15_12_9_3
void FASTCALL call_15_12_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_9_4
#define init_15_12_9_4
void FASTCALL call_15_12_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_9_5
#define init_15_12_9_5
void FASTCALL call_15_12_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_9_6
#define init_15_12_9_6
void FASTCALL call_15_12_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_9_7
#define init_15_12_9_7
void FASTCALL call_15_12_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_9_8
#define init_15_12_9_8
void FASTCALL call_15_12_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_9_9
#define init_15_12_9_9
void FASTCALL call_15_12_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_9_10
#define init_15_12_9_10
void FASTCALL call_15_12_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_9_11
#define init_15_12_9_11
void FASTCALL call_15_12_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_9_12
#define init_15_12_9_12
void FASTCALL call_15_12_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_9_13
#define init_15_12_9_13
void FASTCALL call_15_12_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_9_14
#define init_15_12_9_14
void FASTCALL call_15_12_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_9_15
#define init_15_12_9_15
void FASTCALL call_15_12_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_10_0
#define init_15_12_10_0
void FASTCALL call_15_12_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_10_1
#define init_15_12_10_1
void FASTCALL call_15_12_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_10_2
#define init_15_12_10_2
void FASTCALL call_15_12_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_10_3
#define init_15_12_10_3
void FASTCALL call_15_12_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_10_4
#define init_15_12_10_4
void FASTCALL call_15_12_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_10_5
#define init_15_12_10_5
void FASTCALL call_15_12_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_10_6
#define init_15_12_10_6
void FASTCALL call_15_12_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_10_7
#define init_15_12_10_7
void FASTCALL call_15_12_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_10_8
#define init_15_12_10_8
void FASTCALL call_15_12_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_10_9
#define init_15_12_10_9
void FASTCALL call_15_12_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_10_10
#define init_15_12_10_10
void FASTCALL call_15_12_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_10_11
#define init_15_12_10_11
void FASTCALL call_15_12_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_10_12
#define init_15_12_10_12
void FASTCALL call_15_12_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_10_13
#define init_15_12_10_13
void FASTCALL call_15_12_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_10_14
#define init_15_12_10_14
void FASTCALL call_15_12_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_10_15
#define init_15_12_10_15
void FASTCALL call_15_12_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_11_0
#define init_15_12_11_0
void FASTCALL call_15_12_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_11_1
#define init_15_12_11_1
void FASTCALL call_15_12_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_11_2
#define init_15_12_11_2
void FASTCALL call_15_12_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_11_3
#define init_15_12_11_3
void FASTCALL call_15_12_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_11_4
#define init_15_12_11_4
void FASTCALL call_15_12_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_11_5
#define init_15_12_11_5
void FASTCALL call_15_12_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_11_6
#define init_15_12_11_6
void FASTCALL call_15_12_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_11_7
#define init_15_12_11_7
void FASTCALL call_15_12_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_11_8
#define init_15_12_11_8
void FASTCALL call_15_12_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_11_9
#define init_15_12_11_9
void FASTCALL call_15_12_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_11_10
#define init_15_12_11_10
void FASTCALL call_15_12_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_11_11
#define init_15_12_11_11
void FASTCALL call_15_12_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_11_12
#define init_15_12_11_12
void FASTCALL call_15_12_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_11_13
#define init_15_12_11_13
void FASTCALL call_15_12_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_11_14
#define init_15_12_11_14
void FASTCALL call_15_12_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_11_15
#define init_15_12_11_15
void FASTCALL call_15_12_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_12_0
#define init_15_12_12_0
void FASTCALL call_15_12_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_12_1
#define init_15_12_12_1
void FASTCALL call_15_12_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_12_2
#define init_15_12_12_2
void FASTCALL call_15_12_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_12_3
#define init_15_12_12_3
void FASTCALL call_15_12_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_12_4
#define init_15_12_12_4
void FASTCALL call_15_12_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_12_5
#define init_15_12_12_5
void FASTCALL call_15_12_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_12_6
#define init_15_12_12_6
void FASTCALL call_15_12_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_12_7
#define init_15_12_12_7
void FASTCALL call_15_12_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_12_8
#define init_15_12_12_8
void FASTCALL call_15_12_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_12_9
#define init_15_12_12_9
void FASTCALL call_15_12_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_12_10
#define init_15_12_12_10
void FASTCALL call_15_12_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_12_11
#define init_15_12_12_11
void FASTCALL call_15_12_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_12_12
#define init_15_12_12_12
void FASTCALL call_15_12_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_12_13
#define init_15_12_12_13
void FASTCALL call_15_12_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_12_14
#define init_15_12_12_14
void FASTCALL call_15_12_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_12_15
#define init_15_12_12_15
void FASTCALL call_15_12_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_13_0
#define init_15_12_13_0
void FASTCALL call_15_12_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_13_1
#define init_15_12_13_1
void FASTCALL call_15_12_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_13_2
#define init_15_12_13_2
void FASTCALL call_15_12_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_13_3
#define init_15_12_13_3
void FASTCALL call_15_12_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_13_4
#define init_15_12_13_4
void FASTCALL call_15_12_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_13_5
#define init_15_12_13_5
void FASTCALL call_15_12_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_13_6
#define init_15_12_13_6
void FASTCALL call_15_12_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_13_7
#define init_15_12_13_7
void FASTCALL call_15_12_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_13_8
#define init_15_12_13_8
void FASTCALL call_15_12_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_13_9
#define init_15_12_13_9
void FASTCALL call_15_12_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_13_10
#define init_15_12_13_10
void FASTCALL call_15_12_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_13_11
#define init_15_12_13_11
void FASTCALL call_15_12_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_13_12
#define init_15_12_13_12
void FASTCALL call_15_12_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_13_13
#define init_15_12_13_13
void FASTCALL call_15_12_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_13_14
#define init_15_12_13_14
void FASTCALL call_15_12_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_13_15
#define init_15_12_13_15
void FASTCALL call_15_12_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_14_0
#define init_15_12_14_0
void FASTCALL call_15_12_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_14_1
#define init_15_12_14_1
void FASTCALL call_15_12_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_14_2
#define init_15_12_14_2
void FASTCALL call_15_12_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_14_3
#define init_15_12_14_3
void FASTCALL call_15_12_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_14_4
#define init_15_12_14_4
void FASTCALL call_15_12_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_14_5
#define init_15_12_14_5
void FASTCALL call_15_12_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_14_6
#define init_15_12_14_6
void FASTCALL call_15_12_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_14_7
#define init_15_12_14_7
void FASTCALL call_15_12_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_14_8
#define init_15_12_14_8
void FASTCALL call_15_12_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_14_9
#define init_15_12_14_9
void FASTCALL call_15_12_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_14_10
#define init_15_12_14_10
void FASTCALL call_15_12_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_14_11
#define init_15_12_14_11
void FASTCALL call_15_12_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_14_12
#define init_15_12_14_12
void FASTCALL call_15_12_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_14_13
#define init_15_12_14_13
void FASTCALL call_15_12_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_14_14
#define init_15_12_14_14
void FASTCALL call_15_12_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_14_15
#define init_15_12_14_15
void FASTCALL call_15_12_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_15_0
#define init_15_12_15_0
void FASTCALL call_15_12_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_15_1
#define init_15_12_15_1
void FASTCALL call_15_12_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_15_2
#define init_15_12_15_2
void FASTCALL call_15_12_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_15_3
#define init_15_12_15_3
void FASTCALL call_15_12_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_15_4
#define init_15_12_15_4
void FASTCALL call_15_12_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_15_5
#define init_15_12_15_5
void FASTCALL call_15_12_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_15_6
#define init_15_12_15_6
void FASTCALL call_15_12_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_15_7
#define init_15_12_15_7
void FASTCALL call_15_12_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_15_8
#define init_15_12_15_8
void FASTCALL call_15_12_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_15_9
#define init_15_12_15_9
void FASTCALL call_15_12_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_15_10
#define init_15_12_15_10
void FASTCALL call_15_12_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_15_11
#define init_15_12_15_11
void FASTCALL call_15_12_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_15_12
#define init_15_12_15_12
void FASTCALL call_15_12_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_15_13
#define init_15_12_15_13
void FASTCALL call_15_12_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_15_14
#define init_15_12_15_14
void FASTCALL call_15_12_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_12_15_15
#define init_15_12_15_15
void FASTCALL call_15_12_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_0_0
#define init_15_13_0_0
void FASTCALL call_15_13_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_0_1
#define init_15_13_0_1
void FASTCALL call_15_13_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_0_2
#define init_15_13_0_2
void FASTCALL call_15_13_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_0_3
#define init_15_13_0_3
void FASTCALL call_15_13_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_0_4
#define init_15_13_0_4
void FASTCALL call_15_13_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_0_5
#define init_15_13_0_5
void FASTCALL call_15_13_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_0_6
#define init_15_13_0_6
void FASTCALL call_15_13_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_0_7
#define init_15_13_0_7
void FASTCALL call_15_13_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_0_8
#define init_15_13_0_8
void FASTCALL call_15_13_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_0_9
#define init_15_13_0_9
void FASTCALL call_15_13_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_0_10
#define init_15_13_0_10
void FASTCALL call_15_13_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_0_11
#define init_15_13_0_11
void FASTCALL call_15_13_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_0_12
#define init_15_13_0_12
void FASTCALL call_15_13_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_0_13
#define init_15_13_0_13
void FASTCALL call_15_13_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_0_14
#define init_15_13_0_14
void FASTCALL call_15_13_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_0_15
#define init_15_13_0_15
void FASTCALL call_15_13_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_1_0
#define init_15_13_1_0
void FASTCALL call_15_13_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_1_1
#define init_15_13_1_1
void FASTCALL call_15_13_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_1_2
#define init_15_13_1_2
void FASTCALL call_15_13_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_1_3
#define init_15_13_1_3
void FASTCALL call_15_13_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_1_4
#define init_15_13_1_4
void FASTCALL call_15_13_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_1_5
#define init_15_13_1_5
void FASTCALL call_15_13_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_1_6
#define init_15_13_1_6
void FASTCALL call_15_13_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_1_7
#define init_15_13_1_7
void FASTCALL call_15_13_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_1_8
#define init_15_13_1_8
void FASTCALL call_15_13_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_1_9
#define init_15_13_1_9
void FASTCALL call_15_13_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_1_10
#define init_15_13_1_10
void FASTCALL call_15_13_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_1_11
#define init_15_13_1_11
void FASTCALL call_15_13_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_1_12
#define init_15_13_1_12
void FASTCALL call_15_13_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_1_13
#define init_15_13_1_13
void FASTCALL call_15_13_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_1_14
#define init_15_13_1_14
void FASTCALL call_15_13_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_1_15
#define init_15_13_1_15
void FASTCALL call_15_13_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_2_0
#define init_15_13_2_0
void FASTCALL call_15_13_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_2_1
#define init_15_13_2_1
void FASTCALL call_15_13_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_2_2
#define init_15_13_2_2
void FASTCALL call_15_13_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_2_3
#define init_15_13_2_3
void FASTCALL call_15_13_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_2_4
#define init_15_13_2_4
void FASTCALL call_15_13_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_2_5
#define init_15_13_2_5
void FASTCALL call_15_13_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_2_6
#define init_15_13_2_6
void FASTCALL call_15_13_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_2_7
#define init_15_13_2_7
void FASTCALL call_15_13_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_2_8
#define init_15_13_2_8
void FASTCALL call_15_13_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_2_9
#define init_15_13_2_9
void FASTCALL call_15_13_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_2_10
#define init_15_13_2_10
void FASTCALL call_15_13_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_2_11
#define init_15_13_2_11
void FASTCALL call_15_13_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_2_12
#define init_15_13_2_12
void FASTCALL call_15_13_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_2_13
#define init_15_13_2_13
void FASTCALL call_15_13_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_2_14
#define init_15_13_2_14
void FASTCALL call_15_13_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_2_15
#define init_15_13_2_15
void FASTCALL call_15_13_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_3_0
#define init_15_13_3_0
void FASTCALL call_15_13_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_3_1
#define init_15_13_3_1
void FASTCALL call_15_13_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_3_2
#define init_15_13_3_2
void FASTCALL call_15_13_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_3_3
#define init_15_13_3_3
void FASTCALL call_15_13_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_3_4
#define init_15_13_3_4
void FASTCALL call_15_13_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_3_5
#define init_15_13_3_5
void FASTCALL call_15_13_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_3_6
#define init_15_13_3_6
void FASTCALL call_15_13_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_3_7
#define init_15_13_3_7
void FASTCALL call_15_13_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_3_8
#define init_15_13_3_8
void FASTCALL call_15_13_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_3_9
#define init_15_13_3_9
void FASTCALL call_15_13_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_3_10
#define init_15_13_3_10
void FASTCALL call_15_13_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_3_11
#define init_15_13_3_11
void FASTCALL call_15_13_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_3_12
#define init_15_13_3_12
void FASTCALL call_15_13_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_3_13
#define init_15_13_3_13
void FASTCALL call_15_13_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_3_14
#define init_15_13_3_14
void FASTCALL call_15_13_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_3_15
#define init_15_13_3_15
void FASTCALL call_15_13_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_4_0
#define init_15_13_4_0
void FASTCALL call_15_13_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_4_1
#define init_15_13_4_1
void FASTCALL call_15_13_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_4_2
#define init_15_13_4_2
void FASTCALL call_15_13_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_4_3
#define init_15_13_4_3
void FASTCALL call_15_13_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_4_4
#define init_15_13_4_4
void FASTCALL call_15_13_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_4_5
#define init_15_13_4_5
void FASTCALL call_15_13_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_4_6
#define init_15_13_4_6
void FASTCALL call_15_13_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_4_7
#define init_15_13_4_7
void FASTCALL call_15_13_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_4_8
#define init_15_13_4_8
void FASTCALL call_15_13_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_4_9
#define init_15_13_4_9
void FASTCALL call_15_13_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_4_10
#define init_15_13_4_10
void FASTCALL call_15_13_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_4_11
#define init_15_13_4_11
void FASTCALL call_15_13_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_4_12
#define init_15_13_4_12
void FASTCALL call_15_13_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_4_13
#define init_15_13_4_13
void FASTCALL call_15_13_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_4_14
#define init_15_13_4_14
void FASTCALL call_15_13_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_4_15
#define init_15_13_4_15
void FASTCALL call_15_13_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_5_0
#define init_15_13_5_0
void FASTCALL call_15_13_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_5_1
#define init_15_13_5_1
void FASTCALL call_15_13_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_5_2
#define init_15_13_5_2
void FASTCALL call_15_13_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_5_3
#define init_15_13_5_3
void FASTCALL call_15_13_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_5_4
#define init_15_13_5_4
void FASTCALL call_15_13_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_5_5
#define init_15_13_5_5
void FASTCALL call_15_13_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_5_6
#define init_15_13_5_6
void FASTCALL call_15_13_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_5_7
#define init_15_13_5_7
void FASTCALL call_15_13_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_5_8
#define init_15_13_5_8
void FASTCALL call_15_13_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_5_9
#define init_15_13_5_9
void FASTCALL call_15_13_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_5_10
#define init_15_13_5_10
void FASTCALL call_15_13_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_5_11
#define init_15_13_5_11
void FASTCALL call_15_13_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_5_12
#define init_15_13_5_12
void FASTCALL call_15_13_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_5_13
#define init_15_13_5_13
void FASTCALL call_15_13_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_5_14
#define init_15_13_5_14
void FASTCALL call_15_13_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_5_15
#define init_15_13_5_15
void FASTCALL call_15_13_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_6_0
#define init_15_13_6_0
void FASTCALL call_15_13_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_6_1
#define init_15_13_6_1
void FASTCALL call_15_13_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_6_2
#define init_15_13_6_2
void FASTCALL call_15_13_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_6_3
#define init_15_13_6_3
void FASTCALL call_15_13_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_6_4
#define init_15_13_6_4
void FASTCALL call_15_13_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_6_5
#define init_15_13_6_5
void FASTCALL call_15_13_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_6_6
#define init_15_13_6_6
void FASTCALL call_15_13_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_6_7
#define init_15_13_6_7
void FASTCALL call_15_13_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_6_8
#define init_15_13_6_8
void FASTCALL call_15_13_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_6_9
#define init_15_13_6_9
void FASTCALL call_15_13_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_6_10
#define init_15_13_6_10
void FASTCALL call_15_13_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_6_11
#define init_15_13_6_11
void FASTCALL call_15_13_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_6_12
#define init_15_13_6_12
void FASTCALL call_15_13_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_6_13
#define init_15_13_6_13
void FASTCALL call_15_13_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_6_14
#define init_15_13_6_14
void FASTCALL call_15_13_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_6_15
#define init_15_13_6_15
void FASTCALL call_15_13_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_7_0
#define init_15_13_7_0
void FASTCALL call_15_13_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_7_1
#define init_15_13_7_1
void FASTCALL call_15_13_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_7_2
#define init_15_13_7_2
void FASTCALL call_15_13_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_7_3
#define init_15_13_7_3
void FASTCALL call_15_13_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_7_4
#define init_15_13_7_4
void FASTCALL call_15_13_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_7_5
#define init_15_13_7_5
void FASTCALL call_15_13_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_7_6
#define init_15_13_7_6
void FASTCALL call_15_13_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_7_7
#define init_15_13_7_7
void FASTCALL call_15_13_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_7_8
#define init_15_13_7_8
void FASTCALL call_15_13_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_7_9
#define init_15_13_7_9
void FASTCALL call_15_13_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_7_10
#define init_15_13_7_10
void FASTCALL call_15_13_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_7_11
#define init_15_13_7_11
void FASTCALL call_15_13_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_7_12
#define init_15_13_7_12
void FASTCALL call_15_13_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_7_13
#define init_15_13_7_13
void FASTCALL call_15_13_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_7_14
#define init_15_13_7_14
void FASTCALL call_15_13_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_7_15
#define init_15_13_7_15
void FASTCALL call_15_13_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_8_0
#define init_15_13_8_0
void FASTCALL call_15_13_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_8_1
#define init_15_13_8_1
void FASTCALL call_15_13_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_8_2
#define init_15_13_8_2
void FASTCALL call_15_13_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_8_3
#define init_15_13_8_3
void FASTCALL call_15_13_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_8_4
#define init_15_13_8_4
void FASTCALL call_15_13_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_8_5
#define init_15_13_8_5
void FASTCALL call_15_13_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_8_6
#define init_15_13_8_6
void FASTCALL call_15_13_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_8_7
#define init_15_13_8_7
void FASTCALL call_15_13_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_8_8
#define init_15_13_8_8
void FASTCALL call_15_13_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_8_9
#define init_15_13_8_9
void FASTCALL call_15_13_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_8_10
#define init_15_13_8_10
void FASTCALL call_15_13_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_8_11
#define init_15_13_8_11
void FASTCALL call_15_13_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_8_12
#define init_15_13_8_12
void FASTCALL call_15_13_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_8_13
#define init_15_13_8_13
void FASTCALL call_15_13_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_8_14
#define init_15_13_8_14
void FASTCALL call_15_13_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_8_15
#define init_15_13_8_15
void FASTCALL call_15_13_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_9_0
#define init_15_13_9_0
void FASTCALL call_15_13_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_9_1
#define init_15_13_9_1
void FASTCALL call_15_13_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_9_2
#define init_15_13_9_2
void FASTCALL call_15_13_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_9_3
#define init_15_13_9_3
void FASTCALL call_15_13_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_9_4
#define init_15_13_9_4
void FASTCALL call_15_13_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_9_5
#define init_15_13_9_5
void FASTCALL call_15_13_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_9_6
#define init_15_13_9_6
void FASTCALL call_15_13_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_9_7
#define init_15_13_9_7
void FASTCALL call_15_13_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_9_8
#define init_15_13_9_8
void FASTCALL call_15_13_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_9_9
#define init_15_13_9_9
void FASTCALL call_15_13_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_9_10
#define init_15_13_9_10
void FASTCALL call_15_13_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_9_11
#define init_15_13_9_11
void FASTCALL call_15_13_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_9_12
#define init_15_13_9_12
void FASTCALL call_15_13_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_9_13
#define init_15_13_9_13
void FASTCALL call_15_13_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_9_14
#define init_15_13_9_14
void FASTCALL call_15_13_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_9_15
#define init_15_13_9_15
void FASTCALL call_15_13_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_10_0
#define init_15_13_10_0
void FASTCALL call_15_13_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_10_1
#define init_15_13_10_1
void FASTCALL call_15_13_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_10_2
#define init_15_13_10_2
void FASTCALL call_15_13_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_10_3
#define init_15_13_10_3
void FASTCALL call_15_13_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_10_4
#define init_15_13_10_4
void FASTCALL call_15_13_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_10_5
#define init_15_13_10_5
void FASTCALL call_15_13_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_10_6
#define init_15_13_10_6
void FASTCALL call_15_13_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_10_7
#define init_15_13_10_7
void FASTCALL call_15_13_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_10_8
#define init_15_13_10_8
void FASTCALL call_15_13_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_10_9
#define init_15_13_10_9
void FASTCALL call_15_13_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_10_10
#define init_15_13_10_10
void FASTCALL call_15_13_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_10_11
#define init_15_13_10_11
void FASTCALL call_15_13_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_10_12
#define init_15_13_10_12
void FASTCALL call_15_13_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_10_13
#define init_15_13_10_13
void FASTCALL call_15_13_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_10_14
#define init_15_13_10_14
void FASTCALL call_15_13_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_10_15
#define init_15_13_10_15
void FASTCALL call_15_13_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_11_0
#define init_15_13_11_0
void FASTCALL call_15_13_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_11_1
#define init_15_13_11_1
void FASTCALL call_15_13_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_11_2
#define init_15_13_11_2
void FASTCALL call_15_13_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_11_3
#define init_15_13_11_3
void FASTCALL call_15_13_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_11_4
#define init_15_13_11_4
void FASTCALL call_15_13_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_11_5
#define init_15_13_11_5
void FASTCALL call_15_13_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_11_6
#define init_15_13_11_6
void FASTCALL call_15_13_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_11_7
#define init_15_13_11_7
void FASTCALL call_15_13_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_11_8
#define init_15_13_11_8
void FASTCALL call_15_13_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_11_9
#define init_15_13_11_9
void FASTCALL call_15_13_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_11_10
#define init_15_13_11_10
void FASTCALL call_15_13_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_11_11
#define init_15_13_11_11
void FASTCALL call_15_13_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_11_12
#define init_15_13_11_12
void FASTCALL call_15_13_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_11_13
#define init_15_13_11_13
void FASTCALL call_15_13_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_11_14
#define init_15_13_11_14
void FASTCALL call_15_13_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_11_15
#define init_15_13_11_15
void FASTCALL call_15_13_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_12_0
#define init_15_13_12_0
void FASTCALL call_15_13_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_12_1
#define init_15_13_12_1
void FASTCALL call_15_13_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_12_2
#define init_15_13_12_2
void FASTCALL call_15_13_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_12_3
#define init_15_13_12_3
void FASTCALL call_15_13_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_12_4
#define init_15_13_12_4
void FASTCALL call_15_13_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_12_5
#define init_15_13_12_5
void FASTCALL call_15_13_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_12_6
#define init_15_13_12_6
void FASTCALL call_15_13_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_12_7
#define init_15_13_12_7
void FASTCALL call_15_13_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_12_8
#define init_15_13_12_8
void FASTCALL call_15_13_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_12_9
#define init_15_13_12_9
void FASTCALL call_15_13_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_12_10
#define init_15_13_12_10
void FASTCALL call_15_13_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_12_11
#define init_15_13_12_11
void FASTCALL call_15_13_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_12_12
#define init_15_13_12_12
void FASTCALL call_15_13_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_12_13
#define init_15_13_12_13
void FASTCALL call_15_13_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_12_14
#define init_15_13_12_14
void FASTCALL call_15_13_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_12_15
#define init_15_13_12_15
void FASTCALL call_15_13_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_13_0
#define init_15_13_13_0
void FASTCALL call_15_13_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_13_1
#define init_15_13_13_1
void FASTCALL call_15_13_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_13_2
#define init_15_13_13_2
void FASTCALL call_15_13_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_13_3
#define init_15_13_13_3
void FASTCALL call_15_13_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_13_4
#define init_15_13_13_4
void FASTCALL call_15_13_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_13_5
#define init_15_13_13_5
void FASTCALL call_15_13_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_13_6
#define init_15_13_13_6
void FASTCALL call_15_13_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_13_7
#define init_15_13_13_7
void FASTCALL call_15_13_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_13_8
#define init_15_13_13_8
void FASTCALL call_15_13_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_13_9
#define init_15_13_13_9
void FASTCALL call_15_13_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_13_10
#define init_15_13_13_10
void FASTCALL call_15_13_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_13_11
#define init_15_13_13_11
void FASTCALL call_15_13_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_13_12
#define init_15_13_13_12
void FASTCALL call_15_13_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_13_13
#define init_15_13_13_13
void FASTCALL call_15_13_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_13_14
#define init_15_13_13_14
void FASTCALL call_15_13_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_13_15
#define init_15_13_13_15
void FASTCALL call_15_13_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_14_0
#define init_15_13_14_0
void FASTCALL call_15_13_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_14_1
#define init_15_13_14_1
void FASTCALL call_15_13_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_14_2
#define init_15_13_14_2
void FASTCALL call_15_13_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_14_3
#define init_15_13_14_3
void FASTCALL call_15_13_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_14_4
#define init_15_13_14_4
void FASTCALL call_15_13_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_14_5
#define init_15_13_14_5
void FASTCALL call_15_13_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_14_6
#define init_15_13_14_6
void FASTCALL call_15_13_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_14_7
#define init_15_13_14_7
void FASTCALL call_15_13_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_14_8
#define init_15_13_14_8
void FASTCALL call_15_13_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_14_9
#define init_15_13_14_9
void FASTCALL call_15_13_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_14_10
#define init_15_13_14_10
void FASTCALL call_15_13_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_14_11
#define init_15_13_14_11
void FASTCALL call_15_13_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_14_12
#define init_15_13_14_12
void FASTCALL call_15_13_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_14_13
#define init_15_13_14_13
void FASTCALL call_15_13_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_14_14
#define init_15_13_14_14
void FASTCALL call_15_13_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_14_15
#define init_15_13_14_15
void FASTCALL call_15_13_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_15_0
#define init_15_13_15_0
void FASTCALL call_15_13_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_15_1
#define init_15_13_15_1
void FASTCALL call_15_13_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_15_2
#define init_15_13_15_2
void FASTCALL call_15_13_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_15_3
#define init_15_13_15_3
void FASTCALL call_15_13_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_15_4
#define init_15_13_15_4
void FASTCALL call_15_13_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_15_5
#define init_15_13_15_5
void FASTCALL call_15_13_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_15_6
#define init_15_13_15_6
void FASTCALL call_15_13_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_15_7
#define init_15_13_15_7
void FASTCALL call_15_13_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_15_8
#define init_15_13_15_8
void FASTCALL call_15_13_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_15_9
#define init_15_13_15_9
void FASTCALL call_15_13_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_15_10
#define init_15_13_15_10
void FASTCALL call_15_13_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_15_11
#define init_15_13_15_11
void FASTCALL call_15_13_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_15_12
#define init_15_13_15_12
void FASTCALL call_15_13_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_15_13
#define init_15_13_15_13
void FASTCALL call_15_13_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_15_14
#define init_15_13_15_14
void FASTCALL call_15_13_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_13_15_15
#define init_15_13_15_15
void FASTCALL call_15_13_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_0_0
#define init_15_14_0_0
void FASTCALL call_15_14_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_0_1
#define init_15_14_0_1
void FASTCALL call_15_14_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_0_2
#define init_15_14_0_2
void FASTCALL call_15_14_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_0_3
#define init_15_14_0_3
void FASTCALL call_15_14_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_0_4
#define init_15_14_0_4
void FASTCALL call_15_14_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_0_5
#define init_15_14_0_5
void FASTCALL call_15_14_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_0_6
#define init_15_14_0_6
void FASTCALL call_15_14_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_0_7
#define init_15_14_0_7
void FASTCALL call_15_14_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_0_8
#define init_15_14_0_8
void FASTCALL call_15_14_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_0_9
#define init_15_14_0_9
void FASTCALL call_15_14_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_0_10
#define init_15_14_0_10
void FASTCALL call_15_14_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_0_11
#define init_15_14_0_11
void FASTCALL call_15_14_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_0_12
#define init_15_14_0_12
void FASTCALL call_15_14_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_0_13
#define init_15_14_0_13
void FASTCALL call_15_14_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_0_14
#define init_15_14_0_14
void FASTCALL call_15_14_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_0_15
#define init_15_14_0_15
void FASTCALL call_15_14_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_1_0
#define init_15_14_1_0
void FASTCALL call_15_14_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_1_1
#define init_15_14_1_1
void FASTCALL call_15_14_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_1_2
#define init_15_14_1_2
void FASTCALL call_15_14_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_1_3
#define init_15_14_1_3
void FASTCALL call_15_14_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_1_4
#define init_15_14_1_4
void FASTCALL call_15_14_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_1_5
#define init_15_14_1_5
void FASTCALL call_15_14_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_1_6
#define init_15_14_1_6
void FASTCALL call_15_14_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_1_7
#define init_15_14_1_7
void FASTCALL call_15_14_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_1_8
#define init_15_14_1_8
void FASTCALL call_15_14_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_1_9
#define init_15_14_1_9
void FASTCALL call_15_14_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_1_10
#define init_15_14_1_10
void FASTCALL call_15_14_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_1_11
#define init_15_14_1_11
void FASTCALL call_15_14_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_1_12
#define init_15_14_1_12
void FASTCALL call_15_14_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_1_13
#define init_15_14_1_13
void FASTCALL call_15_14_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_1_14
#define init_15_14_1_14
void FASTCALL call_15_14_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_1_15
#define init_15_14_1_15
void FASTCALL call_15_14_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_2_0
#define init_15_14_2_0
void FASTCALL call_15_14_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_2_1
#define init_15_14_2_1
void FASTCALL call_15_14_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_2_2
#define init_15_14_2_2
void FASTCALL call_15_14_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_2_3
#define init_15_14_2_3
void FASTCALL call_15_14_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_2_4
#define init_15_14_2_4
void FASTCALL call_15_14_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_2_5
#define init_15_14_2_5
void FASTCALL call_15_14_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_2_6
#define init_15_14_2_6
void FASTCALL call_15_14_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_2_7
#define init_15_14_2_7
void FASTCALL call_15_14_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_2_8
#define init_15_14_2_8
void FASTCALL call_15_14_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_2_9
#define init_15_14_2_9
void FASTCALL call_15_14_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_2_10
#define init_15_14_2_10
void FASTCALL call_15_14_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_2_11
#define init_15_14_2_11
void FASTCALL call_15_14_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_2_12
#define init_15_14_2_12
void FASTCALL call_15_14_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_2_13
#define init_15_14_2_13
void FASTCALL call_15_14_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_2_14
#define init_15_14_2_14
void FASTCALL call_15_14_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_2_15
#define init_15_14_2_15
void FASTCALL call_15_14_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_3_0
#define init_15_14_3_0
void FASTCALL call_15_14_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_3_1
#define init_15_14_3_1
void FASTCALL call_15_14_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_3_2
#define init_15_14_3_2
void FASTCALL call_15_14_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_3_3
#define init_15_14_3_3
void FASTCALL call_15_14_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_3_4
#define init_15_14_3_4
void FASTCALL call_15_14_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_3_5
#define init_15_14_3_5
void FASTCALL call_15_14_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_3_6
#define init_15_14_3_6
void FASTCALL call_15_14_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_3_7
#define init_15_14_3_7
void FASTCALL call_15_14_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_3_8
#define init_15_14_3_8
void FASTCALL call_15_14_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_3_9
#define init_15_14_3_9
void FASTCALL call_15_14_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_3_10
#define init_15_14_3_10
void FASTCALL call_15_14_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_3_11
#define init_15_14_3_11
void FASTCALL call_15_14_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_3_12
#define init_15_14_3_12
void FASTCALL call_15_14_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_3_13
#define init_15_14_3_13
void FASTCALL call_15_14_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_3_14
#define init_15_14_3_14
void FASTCALL call_15_14_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_3_15
#define init_15_14_3_15
void FASTCALL call_15_14_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_4_0
#define init_15_14_4_0
void FASTCALL call_15_14_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_4_1
#define init_15_14_4_1
void FASTCALL call_15_14_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_4_2
#define init_15_14_4_2
void FASTCALL call_15_14_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_4_3
#define init_15_14_4_3
void FASTCALL call_15_14_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_4_4
#define init_15_14_4_4
void FASTCALL call_15_14_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_4_5
#define init_15_14_4_5
void FASTCALL call_15_14_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_4_6
#define init_15_14_4_6
void FASTCALL call_15_14_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_4_7
#define init_15_14_4_7
void FASTCALL call_15_14_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_4_8
#define init_15_14_4_8
void FASTCALL call_15_14_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_4_9
#define init_15_14_4_9
void FASTCALL call_15_14_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_4_10
#define init_15_14_4_10
void FASTCALL call_15_14_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_4_11
#define init_15_14_4_11
void FASTCALL call_15_14_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_4_12
#define init_15_14_4_12
void FASTCALL call_15_14_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_4_13
#define init_15_14_4_13
void FASTCALL call_15_14_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_4_14
#define init_15_14_4_14
void FASTCALL call_15_14_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_4_15
#define init_15_14_4_15
void FASTCALL call_15_14_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_5_0
#define init_15_14_5_0
void FASTCALL call_15_14_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_5_1
#define init_15_14_5_1
void FASTCALL call_15_14_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_5_2
#define init_15_14_5_2
void FASTCALL call_15_14_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_5_3
#define init_15_14_5_3
void FASTCALL call_15_14_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_5_4
#define init_15_14_5_4
void FASTCALL call_15_14_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_5_5
#define init_15_14_5_5
void FASTCALL call_15_14_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_5_6
#define init_15_14_5_6
void FASTCALL call_15_14_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_5_7
#define init_15_14_5_7
void FASTCALL call_15_14_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_5_8
#define init_15_14_5_8
void FASTCALL call_15_14_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_5_9
#define init_15_14_5_9
void FASTCALL call_15_14_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_5_10
#define init_15_14_5_10
void FASTCALL call_15_14_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_5_11
#define init_15_14_5_11
void FASTCALL call_15_14_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_5_12
#define init_15_14_5_12
void FASTCALL call_15_14_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_5_13
#define init_15_14_5_13
void FASTCALL call_15_14_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_5_14
#define init_15_14_5_14
void FASTCALL call_15_14_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_5_15
#define init_15_14_5_15
void FASTCALL call_15_14_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_6_0
#define init_15_14_6_0
void FASTCALL call_15_14_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_6_1
#define init_15_14_6_1
void FASTCALL call_15_14_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_6_2
#define init_15_14_6_2
void FASTCALL call_15_14_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_6_3
#define init_15_14_6_3
void FASTCALL call_15_14_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_6_4
#define init_15_14_6_4
void FASTCALL call_15_14_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_6_5
#define init_15_14_6_5
void FASTCALL call_15_14_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_6_6
#define init_15_14_6_6
void FASTCALL call_15_14_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_6_7
#define init_15_14_6_7
void FASTCALL call_15_14_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_6_8
#define init_15_14_6_8
void FASTCALL call_15_14_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_6_9
#define init_15_14_6_9
void FASTCALL call_15_14_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_6_10
#define init_15_14_6_10
void FASTCALL call_15_14_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_6_11
#define init_15_14_6_11
void FASTCALL call_15_14_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_6_12
#define init_15_14_6_12
void FASTCALL call_15_14_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_6_13
#define init_15_14_6_13
void FASTCALL call_15_14_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_6_14
#define init_15_14_6_14
void FASTCALL call_15_14_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_6_15
#define init_15_14_6_15
void FASTCALL call_15_14_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_7_0
#define init_15_14_7_0
void FASTCALL call_15_14_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_7_1
#define init_15_14_7_1
void FASTCALL call_15_14_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_7_2
#define init_15_14_7_2
void FASTCALL call_15_14_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_7_3
#define init_15_14_7_3
void FASTCALL call_15_14_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_7_4
#define init_15_14_7_4
void FASTCALL call_15_14_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_7_5
#define init_15_14_7_5
void FASTCALL call_15_14_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_7_6
#define init_15_14_7_6
void FASTCALL call_15_14_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_7_7
#define init_15_14_7_7
void FASTCALL call_15_14_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_7_8
#define init_15_14_7_8
void FASTCALL call_15_14_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_7_9
#define init_15_14_7_9
void FASTCALL call_15_14_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_7_10
#define init_15_14_7_10
void FASTCALL call_15_14_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_7_11
#define init_15_14_7_11
void FASTCALL call_15_14_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_7_12
#define init_15_14_7_12
void FASTCALL call_15_14_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_7_13
#define init_15_14_7_13
void FASTCALL call_15_14_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_7_14
#define init_15_14_7_14
void FASTCALL call_15_14_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_7_15
#define init_15_14_7_15
void FASTCALL call_15_14_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_8_0
#define init_15_14_8_0
void FASTCALL call_15_14_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_8_1
#define init_15_14_8_1
void FASTCALL call_15_14_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_8_2
#define init_15_14_8_2
void FASTCALL call_15_14_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_8_3
#define init_15_14_8_3
void FASTCALL call_15_14_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_8_4
#define init_15_14_8_4
void FASTCALL call_15_14_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_8_5
#define init_15_14_8_5
void FASTCALL call_15_14_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_8_6
#define init_15_14_8_6
void FASTCALL call_15_14_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_8_7
#define init_15_14_8_7
void FASTCALL call_15_14_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_8_8
#define init_15_14_8_8
void FASTCALL call_15_14_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_8_9
#define init_15_14_8_9
void FASTCALL call_15_14_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_8_10
#define init_15_14_8_10
void FASTCALL call_15_14_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_8_11
#define init_15_14_8_11
void FASTCALL call_15_14_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_8_12
#define init_15_14_8_12
void FASTCALL call_15_14_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_8_13
#define init_15_14_8_13
void FASTCALL call_15_14_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_8_14
#define init_15_14_8_14
void FASTCALL call_15_14_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_8_15
#define init_15_14_8_15
void FASTCALL call_15_14_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_9_0
#define init_15_14_9_0
void FASTCALL call_15_14_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_9_1
#define init_15_14_9_1
void FASTCALL call_15_14_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_9_2
#define init_15_14_9_2
void FASTCALL call_15_14_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_9_3
#define init_15_14_9_3
void FASTCALL call_15_14_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_9_4
#define init_15_14_9_4
void FASTCALL call_15_14_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_9_5
#define init_15_14_9_5
void FASTCALL call_15_14_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_9_6
#define init_15_14_9_6
void FASTCALL call_15_14_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_9_7
#define init_15_14_9_7
void FASTCALL call_15_14_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_9_8
#define init_15_14_9_8
void FASTCALL call_15_14_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_9_9
#define init_15_14_9_9
void FASTCALL call_15_14_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_9_10
#define init_15_14_9_10
void FASTCALL call_15_14_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_9_11
#define init_15_14_9_11
void FASTCALL call_15_14_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_9_12
#define init_15_14_9_12
void FASTCALL call_15_14_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_9_13
#define init_15_14_9_13
void FASTCALL call_15_14_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_9_14
#define init_15_14_9_14
void FASTCALL call_15_14_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_9_15
#define init_15_14_9_15
void FASTCALL call_15_14_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_10_0
#define init_15_14_10_0
void FASTCALL call_15_14_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_10_1
#define init_15_14_10_1
void FASTCALL call_15_14_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_10_2
#define init_15_14_10_2
void FASTCALL call_15_14_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_10_3
#define init_15_14_10_3
void FASTCALL call_15_14_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_10_4
#define init_15_14_10_4
void FASTCALL call_15_14_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_10_5
#define init_15_14_10_5
void FASTCALL call_15_14_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_10_6
#define init_15_14_10_6
void FASTCALL call_15_14_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_10_7
#define init_15_14_10_7
void FASTCALL call_15_14_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_10_8
#define init_15_14_10_8
void FASTCALL call_15_14_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_10_9
#define init_15_14_10_9
void FASTCALL call_15_14_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_10_10
#define init_15_14_10_10
void FASTCALL call_15_14_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_10_11
#define init_15_14_10_11
void FASTCALL call_15_14_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_10_12
#define init_15_14_10_12
void FASTCALL call_15_14_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_10_13
#define init_15_14_10_13
void FASTCALL call_15_14_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_10_14
#define init_15_14_10_14
void FASTCALL call_15_14_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_10_15
#define init_15_14_10_15
void FASTCALL call_15_14_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_11_0
#define init_15_14_11_0
void FASTCALL call_15_14_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_11_1
#define init_15_14_11_1
void FASTCALL call_15_14_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_11_2
#define init_15_14_11_2
void FASTCALL call_15_14_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_11_3
#define init_15_14_11_3
void FASTCALL call_15_14_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_11_4
#define init_15_14_11_4
void FASTCALL call_15_14_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_11_5
#define init_15_14_11_5
void FASTCALL call_15_14_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_11_6
#define init_15_14_11_6
void FASTCALL call_15_14_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_11_7
#define init_15_14_11_7
void FASTCALL call_15_14_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_11_8
#define init_15_14_11_8
void FASTCALL call_15_14_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_11_9
#define init_15_14_11_9
void FASTCALL call_15_14_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_11_10
#define init_15_14_11_10
void FASTCALL call_15_14_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_11_11
#define init_15_14_11_11
void FASTCALL call_15_14_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_11_12
#define init_15_14_11_12
void FASTCALL call_15_14_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_11_13
#define init_15_14_11_13
void FASTCALL call_15_14_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_11_14
#define init_15_14_11_14
void FASTCALL call_15_14_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_11_15
#define init_15_14_11_15
void FASTCALL call_15_14_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_12_0
#define init_15_14_12_0
void FASTCALL call_15_14_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_12_1
#define init_15_14_12_1
void FASTCALL call_15_14_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_12_2
#define init_15_14_12_2
void FASTCALL call_15_14_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_12_3
#define init_15_14_12_3
void FASTCALL call_15_14_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_12_4
#define init_15_14_12_4
void FASTCALL call_15_14_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_12_5
#define init_15_14_12_5
void FASTCALL call_15_14_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_12_6
#define init_15_14_12_6
void FASTCALL call_15_14_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_12_7
#define init_15_14_12_7
void FASTCALL call_15_14_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_12_8
#define init_15_14_12_8
void FASTCALL call_15_14_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_12_9
#define init_15_14_12_9
void FASTCALL call_15_14_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_12_10
#define init_15_14_12_10
void FASTCALL call_15_14_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_12_11
#define init_15_14_12_11
void FASTCALL call_15_14_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_12_12
#define init_15_14_12_12
void FASTCALL call_15_14_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_12_13
#define init_15_14_12_13
void FASTCALL call_15_14_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_12_14
#define init_15_14_12_14
void FASTCALL call_15_14_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_12_15
#define init_15_14_12_15
void FASTCALL call_15_14_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_13_0
#define init_15_14_13_0
void FASTCALL call_15_14_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_13_1
#define init_15_14_13_1
void FASTCALL call_15_14_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_13_2
#define init_15_14_13_2
void FASTCALL call_15_14_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_13_3
#define init_15_14_13_3
void FASTCALL call_15_14_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_13_4
#define init_15_14_13_4
void FASTCALL call_15_14_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_13_5
#define init_15_14_13_5
void FASTCALL call_15_14_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_13_6
#define init_15_14_13_6
void FASTCALL call_15_14_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_13_7
#define init_15_14_13_7
void FASTCALL call_15_14_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_13_8
#define init_15_14_13_8
void FASTCALL call_15_14_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_13_9
#define init_15_14_13_9
void FASTCALL call_15_14_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_13_10
#define init_15_14_13_10
void FASTCALL call_15_14_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_13_11
#define init_15_14_13_11
void FASTCALL call_15_14_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_13_12
#define init_15_14_13_12
void FASTCALL call_15_14_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_13_13
#define init_15_14_13_13
void FASTCALL call_15_14_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_13_14
#define init_15_14_13_14
void FASTCALL call_15_14_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_13_15
#define init_15_14_13_15
void FASTCALL call_15_14_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_14_0
#define init_15_14_14_0
void FASTCALL call_15_14_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_14_1
#define init_15_14_14_1
void FASTCALL call_15_14_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_14_2
#define init_15_14_14_2
void FASTCALL call_15_14_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_14_3
#define init_15_14_14_3
void FASTCALL call_15_14_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_14_4
#define init_15_14_14_4
void FASTCALL call_15_14_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_14_5
#define init_15_14_14_5
void FASTCALL call_15_14_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_14_6
#define init_15_14_14_6
void FASTCALL call_15_14_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_14_7
#define init_15_14_14_7
void FASTCALL call_15_14_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_14_8
#define init_15_14_14_8
void FASTCALL call_15_14_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_14_9
#define init_15_14_14_9
void FASTCALL call_15_14_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_14_10
#define init_15_14_14_10
void FASTCALL call_15_14_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_14_11
#define init_15_14_14_11
void FASTCALL call_15_14_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_14_12
#define init_15_14_14_12
void FASTCALL call_15_14_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_14_13
#define init_15_14_14_13
void FASTCALL call_15_14_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_14_14
#define init_15_14_14_14
void FASTCALL call_15_14_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_14_15
#define init_15_14_14_15
void FASTCALL call_15_14_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_15_0
#define init_15_14_15_0
void FASTCALL call_15_14_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_15_1
#define init_15_14_15_1
void FASTCALL call_15_14_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_15_2
#define init_15_14_15_2
void FASTCALL call_15_14_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_15_3
#define init_15_14_15_3
void FASTCALL call_15_14_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_15_4
#define init_15_14_15_4
void FASTCALL call_15_14_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_15_5
#define init_15_14_15_5
void FASTCALL call_15_14_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_15_6
#define init_15_14_15_6
void FASTCALL call_15_14_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_15_7
#define init_15_14_15_7
void FASTCALL call_15_14_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_15_8
#define init_15_14_15_8
void FASTCALL call_15_14_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_15_9
#define init_15_14_15_9
void FASTCALL call_15_14_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_15_10
#define init_15_14_15_10
void FASTCALL call_15_14_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_15_11
#define init_15_14_15_11
void FASTCALL call_15_14_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_15_12
#define init_15_14_15_12
void FASTCALL call_15_14_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_15_13
#define init_15_14_15_13
void FASTCALL call_15_14_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_15_14
#define init_15_14_15_14
void FASTCALL call_15_14_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_14_15_15
#define init_15_14_15_15
void FASTCALL call_15_14_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_0_0
#define init_15_15_0_0
void FASTCALL call_15_15_0_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_0_1
#define init_15_15_0_1
void FASTCALL call_15_15_0_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_0_2
#define init_15_15_0_2
void FASTCALL call_15_15_0_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_0_3
#define init_15_15_0_3
void FASTCALL call_15_15_0_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_0_4
#define init_15_15_0_4
void FASTCALL call_15_15_0_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_0_5
#define init_15_15_0_5
void FASTCALL call_15_15_0_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_0_6
#define init_15_15_0_6
void FASTCALL call_15_15_0_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_0_7
#define init_15_15_0_7
void FASTCALL call_15_15_0_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_0_8
#define init_15_15_0_8
void FASTCALL call_15_15_0_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_0_9
#define init_15_15_0_9
void FASTCALL call_15_15_0_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_0_10
#define init_15_15_0_10
void FASTCALL call_15_15_0_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_0_11
#define init_15_15_0_11
void FASTCALL call_15_15_0_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_0_12
#define init_15_15_0_12
void FASTCALL call_15_15_0_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_0_13
#define init_15_15_0_13
void FASTCALL call_15_15_0_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_0_14
#define init_15_15_0_14
void FASTCALL call_15_15_0_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_0_15
#define init_15_15_0_15
void FASTCALL call_15_15_0_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_1_0
#define init_15_15_1_0
void FASTCALL call_15_15_1_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_1_1
#define init_15_15_1_1
void FASTCALL call_15_15_1_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_1_2
#define init_15_15_1_2
void FASTCALL call_15_15_1_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_1_3
#define init_15_15_1_3
void FASTCALL call_15_15_1_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_1_4
#define init_15_15_1_4
void FASTCALL call_15_15_1_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_1_5
#define init_15_15_1_5
void FASTCALL call_15_15_1_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_1_6
#define init_15_15_1_6
void FASTCALL call_15_15_1_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_1_7
#define init_15_15_1_7
void FASTCALL call_15_15_1_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_1_8
#define init_15_15_1_8
void FASTCALL call_15_15_1_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_1_9
#define init_15_15_1_9
void FASTCALL call_15_15_1_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_1_10
#define init_15_15_1_10
void FASTCALL call_15_15_1_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_1_11
#define init_15_15_1_11
void FASTCALL call_15_15_1_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_1_12
#define init_15_15_1_12
void FASTCALL call_15_15_1_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_1_13
#define init_15_15_1_13
void FASTCALL call_15_15_1_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_1_14
#define init_15_15_1_14
void FASTCALL call_15_15_1_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_1_15
#define init_15_15_1_15
void FASTCALL call_15_15_1_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_2_0
#define init_15_15_2_0
void FASTCALL call_15_15_2_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_2_1
#define init_15_15_2_1
void FASTCALL call_15_15_2_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_2_2
#define init_15_15_2_2
void FASTCALL call_15_15_2_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_2_3
#define init_15_15_2_3
void FASTCALL call_15_15_2_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_2_4
#define init_15_15_2_4
void FASTCALL call_15_15_2_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_2_5
#define init_15_15_2_5
void FASTCALL call_15_15_2_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_2_6
#define init_15_15_2_6
void FASTCALL call_15_15_2_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_2_7
#define init_15_15_2_7
void FASTCALL call_15_15_2_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_2_8
#define init_15_15_2_8
void FASTCALL call_15_15_2_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_2_9
#define init_15_15_2_9
void FASTCALL call_15_15_2_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_2_10
#define init_15_15_2_10
void FASTCALL call_15_15_2_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_2_11
#define init_15_15_2_11
void FASTCALL call_15_15_2_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_2_12
#define init_15_15_2_12
void FASTCALL call_15_15_2_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_2_13
#define init_15_15_2_13
void FASTCALL call_15_15_2_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_2_14
#define init_15_15_2_14
void FASTCALL call_15_15_2_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_2_15
#define init_15_15_2_15
void FASTCALL call_15_15_2_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_3_0
#define init_15_15_3_0
void FASTCALL call_15_15_3_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_3_1
#define init_15_15_3_1
void FASTCALL call_15_15_3_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_3_2
#define init_15_15_3_2
void FASTCALL call_15_15_3_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_3_3
#define init_15_15_3_3
void FASTCALL call_15_15_3_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_3_4
#define init_15_15_3_4
void FASTCALL call_15_15_3_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_3_5
#define init_15_15_3_5
void FASTCALL call_15_15_3_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_3_6
#define init_15_15_3_6
void FASTCALL call_15_15_3_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_3_7
#define init_15_15_3_7
void FASTCALL call_15_15_3_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_3_8
#define init_15_15_3_8
void FASTCALL call_15_15_3_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_3_9
#define init_15_15_3_9
void FASTCALL call_15_15_3_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_3_10
#define init_15_15_3_10
void FASTCALL call_15_15_3_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_3_11
#define init_15_15_3_11
void FASTCALL call_15_15_3_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_3_12
#define init_15_15_3_12
void FASTCALL call_15_15_3_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_3_13
#define init_15_15_3_13
void FASTCALL call_15_15_3_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_3_14
#define init_15_15_3_14
void FASTCALL call_15_15_3_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_3_15
#define init_15_15_3_15
void FASTCALL call_15_15_3_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_4_0
#define init_15_15_4_0
void FASTCALL call_15_15_4_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_4_1
#define init_15_15_4_1
void FASTCALL call_15_15_4_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_4_2
#define init_15_15_4_2
void FASTCALL call_15_15_4_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_4_3
#define init_15_15_4_3
void FASTCALL call_15_15_4_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_4_4
#define init_15_15_4_4
void FASTCALL call_15_15_4_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_4_5
#define init_15_15_4_5
void FASTCALL call_15_15_4_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_4_6
#define init_15_15_4_6
void FASTCALL call_15_15_4_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_4_7
#define init_15_15_4_7
void FASTCALL call_15_15_4_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_4_8
#define init_15_15_4_8
void FASTCALL call_15_15_4_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_4_9
#define init_15_15_4_9
void FASTCALL call_15_15_4_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_4_10
#define init_15_15_4_10
void FASTCALL call_15_15_4_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_4_11
#define init_15_15_4_11
void FASTCALL call_15_15_4_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_4_12
#define init_15_15_4_12
void FASTCALL call_15_15_4_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_4_13
#define init_15_15_4_13
void FASTCALL call_15_15_4_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_4_14
#define init_15_15_4_14
void FASTCALL call_15_15_4_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_4_15
#define init_15_15_4_15
void FASTCALL call_15_15_4_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_5_0
#define init_15_15_5_0
void FASTCALL call_15_15_5_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_5_1
#define init_15_15_5_1
void FASTCALL call_15_15_5_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_5_2
#define init_15_15_5_2
void FASTCALL call_15_15_5_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_5_3
#define init_15_15_5_3
void FASTCALL call_15_15_5_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_5_4
#define init_15_15_5_4
void FASTCALL call_15_15_5_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_5_5
#define init_15_15_5_5
void FASTCALL call_15_15_5_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_5_6
#define init_15_15_5_6
void FASTCALL call_15_15_5_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_5_7
#define init_15_15_5_7
void FASTCALL call_15_15_5_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_5_8
#define init_15_15_5_8
void FASTCALL call_15_15_5_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_5_9
#define init_15_15_5_9
void FASTCALL call_15_15_5_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_5_10
#define init_15_15_5_10
void FASTCALL call_15_15_5_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_5_11
#define init_15_15_5_11
void FASTCALL call_15_15_5_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_5_12
#define init_15_15_5_12
void FASTCALL call_15_15_5_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_5_13
#define init_15_15_5_13
void FASTCALL call_15_15_5_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_5_14
#define init_15_15_5_14
void FASTCALL call_15_15_5_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_5_15
#define init_15_15_5_15
void FASTCALL call_15_15_5_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_6_0
#define init_15_15_6_0
void FASTCALL call_15_15_6_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_6_1
#define init_15_15_6_1
void FASTCALL call_15_15_6_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_6_2
#define init_15_15_6_2
void FASTCALL call_15_15_6_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_6_3
#define init_15_15_6_3
void FASTCALL call_15_15_6_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_6_4
#define init_15_15_6_4
void FASTCALL call_15_15_6_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_6_5
#define init_15_15_6_5
void FASTCALL call_15_15_6_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_6_6
#define init_15_15_6_6
void FASTCALL call_15_15_6_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_6_7
#define init_15_15_6_7
void FASTCALL call_15_15_6_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_6_8
#define init_15_15_6_8
void FASTCALL call_15_15_6_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_6_9
#define init_15_15_6_9
void FASTCALL call_15_15_6_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_6_10
#define init_15_15_6_10
void FASTCALL call_15_15_6_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_6_11
#define init_15_15_6_11
void FASTCALL call_15_15_6_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_6_12
#define init_15_15_6_12
void FASTCALL call_15_15_6_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_6_13
#define init_15_15_6_13
void FASTCALL call_15_15_6_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_6_14
#define init_15_15_6_14
void FASTCALL call_15_15_6_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_6_15
#define init_15_15_6_15
void FASTCALL call_15_15_6_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_7_0
#define init_15_15_7_0
void FASTCALL call_15_15_7_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_7_1
#define init_15_15_7_1
void FASTCALL call_15_15_7_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_7_2
#define init_15_15_7_2
void FASTCALL call_15_15_7_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_7_3
#define init_15_15_7_3
void FASTCALL call_15_15_7_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_7_4
#define init_15_15_7_4
void FASTCALL call_15_15_7_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_7_5
#define init_15_15_7_5
void FASTCALL call_15_15_7_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_7_6
#define init_15_15_7_6
void FASTCALL call_15_15_7_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_7_7
#define init_15_15_7_7
void FASTCALL call_15_15_7_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_7_8
#define init_15_15_7_8
void FASTCALL call_15_15_7_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_7_9
#define init_15_15_7_9
void FASTCALL call_15_15_7_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_7_10
#define init_15_15_7_10
void FASTCALL call_15_15_7_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_7_11
#define init_15_15_7_11
void FASTCALL call_15_15_7_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_7_12
#define init_15_15_7_12
void FASTCALL call_15_15_7_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_7_13
#define init_15_15_7_13
void FASTCALL call_15_15_7_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_7_14
#define init_15_15_7_14
void FASTCALL call_15_15_7_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_7_15
#define init_15_15_7_15
void FASTCALL call_15_15_7_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_8_0
#define init_15_15_8_0
void FASTCALL call_15_15_8_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_8_1
#define init_15_15_8_1
void FASTCALL call_15_15_8_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_8_2
#define init_15_15_8_2
void FASTCALL call_15_15_8_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_8_3
#define init_15_15_8_3
void FASTCALL call_15_15_8_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_8_4
#define init_15_15_8_4
void FASTCALL call_15_15_8_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_8_5
#define init_15_15_8_5
void FASTCALL call_15_15_8_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_8_6
#define init_15_15_8_6
void FASTCALL call_15_15_8_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_8_7
#define init_15_15_8_7
void FASTCALL call_15_15_8_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_8_8
#define init_15_15_8_8
void FASTCALL call_15_15_8_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_8_9
#define init_15_15_8_9
void FASTCALL call_15_15_8_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_8_10
#define init_15_15_8_10
void FASTCALL call_15_15_8_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_8_11
#define init_15_15_8_11
void FASTCALL call_15_15_8_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_8_12
#define init_15_15_8_12
void FASTCALL call_15_15_8_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_8_13
#define init_15_15_8_13
void FASTCALL call_15_15_8_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_8_14
#define init_15_15_8_14
void FASTCALL call_15_15_8_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_8_15
#define init_15_15_8_15
void FASTCALL call_15_15_8_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_9_0
#define init_15_15_9_0
void FASTCALL call_15_15_9_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_9_1
#define init_15_15_9_1
void FASTCALL call_15_15_9_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_9_2
#define init_15_15_9_2
void FASTCALL call_15_15_9_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_9_3
#define init_15_15_9_3
void FASTCALL call_15_15_9_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_9_4
#define init_15_15_9_4
void FASTCALL call_15_15_9_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_9_5
#define init_15_15_9_5
void FASTCALL call_15_15_9_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_9_6
#define init_15_15_9_6
void FASTCALL call_15_15_9_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_9_7
#define init_15_15_9_7
void FASTCALL call_15_15_9_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_9_8
#define init_15_15_9_8
void FASTCALL call_15_15_9_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_9_9
#define init_15_15_9_9
void FASTCALL call_15_15_9_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_9_10
#define init_15_15_9_10
void FASTCALL call_15_15_9_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_9_11
#define init_15_15_9_11
void FASTCALL call_15_15_9_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_9_12
#define init_15_15_9_12
void FASTCALL call_15_15_9_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_9_13
#define init_15_15_9_13
void FASTCALL call_15_15_9_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_9_14
#define init_15_15_9_14
void FASTCALL call_15_15_9_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_9_15
#define init_15_15_9_15
void FASTCALL call_15_15_9_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_10_0
#define init_15_15_10_0
void FASTCALL call_15_15_10_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_10_1
#define init_15_15_10_1
void FASTCALL call_15_15_10_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_10_2
#define init_15_15_10_2
void FASTCALL call_15_15_10_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_10_3
#define init_15_15_10_3
void FASTCALL call_15_15_10_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_10_4
#define init_15_15_10_4
void FASTCALL call_15_15_10_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_10_5
#define init_15_15_10_5
void FASTCALL call_15_15_10_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_10_6
#define init_15_15_10_6
void FASTCALL call_15_15_10_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_10_7
#define init_15_15_10_7
void FASTCALL call_15_15_10_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_10_8
#define init_15_15_10_8
void FASTCALL call_15_15_10_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_10_9
#define init_15_15_10_9
void FASTCALL call_15_15_10_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_10_10
#define init_15_15_10_10
void FASTCALL call_15_15_10_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_10_11
#define init_15_15_10_11
void FASTCALL call_15_15_10_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_10_12
#define init_15_15_10_12
void FASTCALL call_15_15_10_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_10_13
#define init_15_15_10_13
void FASTCALL call_15_15_10_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_10_14
#define init_15_15_10_14
void FASTCALL call_15_15_10_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_10_15
#define init_15_15_10_15
void FASTCALL call_15_15_10_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_11_0
#define init_15_15_11_0
void FASTCALL call_15_15_11_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_11_1
#define init_15_15_11_1
void FASTCALL call_15_15_11_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_11_2
#define init_15_15_11_2
void FASTCALL call_15_15_11_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_11_3
#define init_15_15_11_3
void FASTCALL call_15_15_11_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_11_4
#define init_15_15_11_4
void FASTCALL call_15_15_11_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_11_5
#define init_15_15_11_5
void FASTCALL call_15_15_11_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_11_6
#define init_15_15_11_6
void FASTCALL call_15_15_11_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_11_7
#define init_15_15_11_7
void FASTCALL call_15_15_11_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_11_8
#define init_15_15_11_8
void FASTCALL call_15_15_11_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_11_9
#define init_15_15_11_9
void FASTCALL call_15_15_11_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_11_10
#define init_15_15_11_10
void FASTCALL call_15_15_11_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_11_11
#define init_15_15_11_11
void FASTCALL call_15_15_11_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_11_12
#define init_15_15_11_12
void FASTCALL call_15_15_11_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_11_13
#define init_15_15_11_13
void FASTCALL call_15_15_11_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_11_14
#define init_15_15_11_14
void FASTCALL call_15_15_11_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_11_15
#define init_15_15_11_15
void FASTCALL call_15_15_11_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_12_0
#define init_15_15_12_0
void FASTCALL call_15_15_12_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_12_1
#define init_15_15_12_1
void FASTCALL call_15_15_12_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_12_2
#define init_15_15_12_2
void FASTCALL call_15_15_12_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_12_3
#define init_15_15_12_3
void FASTCALL call_15_15_12_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_12_4
#define init_15_15_12_4
void FASTCALL call_15_15_12_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_12_5
#define init_15_15_12_5
void FASTCALL call_15_15_12_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_12_6
#define init_15_15_12_6
void FASTCALL call_15_15_12_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_12_7
#define init_15_15_12_7
void FASTCALL call_15_15_12_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_12_8
#define init_15_15_12_8
void FASTCALL call_15_15_12_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_12_9
#define init_15_15_12_9
void FASTCALL call_15_15_12_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_12_10
#define init_15_15_12_10
void FASTCALL call_15_15_12_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_12_11
#define init_15_15_12_11
void FASTCALL call_15_15_12_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_12_12
#define init_15_15_12_12
void FASTCALL call_15_15_12_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_12_13
#define init_15_15_12_13
void FASTCALL call_15_15_12_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_12_14
#define init_15_15_12_14
void FASTCALL call_15_15_12_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_12_15
#define init_15_15_12_15
void FASTCALL call_15_15_12_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_13_0
#define init_15_15_13_0
void FASTCALL call_15_15_13_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_13_1
#define init_15_15_13_1
void FASTCALL call_15_15_13_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_13_2
#define init_15_15_13_2
void FASTCALL call_15_15_13_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_13_3
#define init_15_15_13_3
void FASTCALL call_15_15_13_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_13_4
#define init_15_15_13_4
void FASTCALL call_15_15_13_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_13_5
#define init_15_15_13_5
void FASTCALL call_15_15_13_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_13_6
#define init_15_15_13_6
void FASTCALL call_15_15_13_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_13_7
#define init_15_15_13_7
void FASTCALL call_15_15_13_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_13_8
#define init_15_15_13_8
void FASTCALL call_15_15_13_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_13_9
#define init_15_15_13_9
void FASTCALL call_15_15_13_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_13_10
#define init_15_15_13_10
void FASTCALL call_15_15_13_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_13_11
#define init_15_15_13_11
void FASTCALL call_15_15_13_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_13_12
#define init_15_15_13_12
void FASTCALL call_15_15_13_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_13_13
#define init_15_15_13_13
void FASTCALL call_15_15_13_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_13_14
#define init_15_15_13_14
void FASTCALL call_15_15_13_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_13_15
#define init_15_15_13_15
void FASTCALL call_15_15_13_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_14_0
#define init_15_15_14_0
void FASTCALL call_15_15_14_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_14_1
#define init_15_15_14_1
void FASTCALL call_15_15_14_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_14_2
#define init_15_15_14_2
void FASTCALL call_15_15_14_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_14_3
#define init_15_15_14_3
void FASTCALL call_15_15_14_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_14_4
#define init_15_15_14_4
void FASTCALL call_15_15_14_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_14_5
#define init_15_15_14_5
void FASTCALL call_15_15_14_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_14_6
#define init_15_15_14_6
void FASTCALL call_15_15_14_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_14_7
#define init_15_15_14_7
void FASTCALL call_15_15_14_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_14_8
#define init_15_15_14_8
void FASTCALL call_15_15_14_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_14_9
#define init_15_15_14_9
void FASTCALL call_15_15_14_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_14_10
#define init_15_15_14_10
void FASTCALL call_15_15_14_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_14_11
#define init_15_15_14_11
void FASTCALL call_15_15_14_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_14_12
#define init_15_15_14_12
void FASTCALL call_15_15_14_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_14_13
#define init_15_15_14_13
void FASTCALL call_15_15_14_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_14_14
#define init_15_15_14_14
void FASTCALL call_15_15_14_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_14_15
#define init_15_15_14_15
void FASTCALL call_15_15_14_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_15_0
#define init_15_15_15_0
void FASTCALL call_15_15_15_0(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_15_1
#define init_15_15_15_1
void FASTCALL call_15_15_15_1(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_15_2
#define init_15_15_15_2
void FASTCALL call_15_15_15_2(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_15_3
#define init_15_15_15_3
void FASTCALL call_15_15_15_3(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_15_4
#define init_15_15_15_4
void FASTCALL call_15_15_15_4(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_15_5
#define init_15_15_15_5
void FASTCALL call_15_15_15_5(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_15_6
#define init_15_15_15_6
void FASTCALL call_15_15_15_6(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_15_7
#define init_15_15_15_7
void FASTCALL call_15_15_15_7(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_15_8
#define init_15_15_15_8
void FASTCALL call_15_15_15_8(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_15_9
#define init_15_15_15_9
void FASTCALL call_15_15_15_9(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_15_10
#define init_15_15_15_10
void FASTCALL call_15_15_15_10(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_15_11
#define init_15_15_15_11
void FASTCALL call_15_15_15_11(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_15_12
#define init_15_15_15_12
void FASTCALL call_15_15_15_12(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_15_13
#define init_15_15_15_13
void FASTCALL call_15_15_15_13(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_15_14
#define init_15_15_15_14
void FASTCALL call_15_15_15_14(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
#ifndef init_15_15_15_15
#define init_15_15_15_15
void FASTCALL call_15_15_15_15(SH2_struct* sh)
{
  SH2undecoded(sh);
}
#endif
opcode_func opcodeTable[0x10000] = { \
call_0_0_0_0, \
call_0_0_0_1, \
call_0_0_0_2, \
call_0_0_0_3, \
call_0_0_0_4, \
call_0_0_0_5, \
call_0_0_0_6, \
call_0_0_0_7, \
call_0_0_0_8, \
call_0_0_0_9, \
call_0_0_0_10, \
call_0_0_0_11, \
call_0_0_0_12, \
call_0_0_0_13, \
call_0_0_0_14, \
call_0_0_0_15, \
call_0_0_1_0, \
call_0_0_1_1, \
call_0_0_1_2, \
call_0_0_1_3, \
call_0_0_1_4, \
call_0_0_1_5, \
call_0_0_1_6, \
call_0_0_1_7, \
call_0_0_1_8, \
call_0_0_1_9, \
call_0_0_1_10, \
call_0_0_1_11, \
call_0_0_1_12, \
call_0_0_1_13, \
call_0_0_1_14, \
call_0_0_1_15, \
call_0_0_2_0, \
call_0_0_2_1, \
call_0_0_2_2, \
call_0_0_2_3, \
call_0_0_2_4, \
call_0_0_2_5, \
call_0_0_2_6, \
call_0_0_2_7, \
call_0_0_2_8, \
call_0_0_2_9, \
call_0_0_2_10, \
call_0_0_2_11, \
call_0_0_2_12, \
call_0_0_2_13, \
call_0_0_2_14, \
call_0_0_2_15, \
call_0_0_3_0, \
call_0_0_3_1, \
call_0_0_3_2, \
call_0_0_3_3, \
call_0_0_3_4, \
call_0_0_3_5, \
call_0_0_3_6, \
call_0_0_3_7, \
call_0_0_3_8, \
call_0_0_3_9, \
call_0_0_3_10, \
call_0_0_3_11, \
call_0_0_3_12, \
call_0_0_3_13, \
call_0_0_3_14, \
call_0_0_3_15, \
call_0_0_4_0, \
call_0_0_4_1, \
call_0_0_4_2, \
call_0_0_4_3, \
call_0_0_4_4, \
call_0_0_4_5, \
call_0_0_4_6, \
call_0_0_4_7, \
call_0_0_4_8, \
call_0_0_4_9, \
call_0_0_4_10, \
call_0_0_4_11, \
call_0_0_4_12, \
call_0_0_4_13, \
call_0_0_4_14, \
call_0_0_4_15, \
call_0_0_5_0, \
call_0_0_5_1, \
call_0_0_5_2, \
call_0_0_5_3, \
call_0_0_5_4, \
call_0_0_5_5, \
call_0_0_5_6, \
call_0_0_5_7, \
call_0_0_5_8, \
call_0_0_5_9, \
call_0_0_5_10, \
call_0_0_5_11, \
call_0_0_5_12, \
call_0_0_5_13, \
call_0_0_5_14, \
call_0_0_5_15, \
call_0_0_6_0, \
call_0_0_6_1, \
call_0_0_6_2, \
call_0_0_6_3, \
call_0_0_6_4, \
call_0_0_6_5, \
call_0_0_6_6, \
call_0_0_6_7, \
call_0_0_6_8, \
call_0_0_6_9, \
call_0_0_6_10, \
call_0_0_6_11, \
call_0_0_6_12, \
call_0_0_6_13, \
call_0_0_6_14, \
call_0_0_6_15, \
call_0_0_7_0, \
call_0_0_7_1, \
call_0_0_7_2, \
call_0_0_7_3, \
call_0_0_7_4, \
call_0_0_7_5, \
call_0_0_7_6, \
call_0_0_7_7, \
call_0_0_7_8, \
call_0_0_7_9, \
call_0_0_7_10, \
call_0_0_7_11, \
call_0_0_7_12, \
call_0_0_7_13, \
call_0_0_7_14, \
call_0_0_7_15, \
call_0_0_8_0, \
call_0_0_8_1, \
call_0_0_8_2, \
call_0_0_8_3, \
call_0_0_8_4, \
call_0_0_8_5, \
call_0_0_8_6, \
call_0_0_8_7, \
call_0_0_8_8, \
call_0_0_8_9, \
call_0_0_8_10, \
call_0_0_8_11, \
call_0_0_8_12, \
call_0_0_8_13, \
call_0_0_8_14, \
call_0_0_8_15, \
call_0_0_9_0, \
call_0_0_9_1, \
call_0_0_9_2, \
call_0_0_9_3, \
call_0_0_9_4, \
call_0_0_9_5, \
call_0_0_9_6, \
call_0_0_9_7, \
call_0_0_9_8, \
call_0_0_9_9, \
call_0_0_9_10, \
call_0_0_9_11, \
call_0_0_9_12, \
call_0_0_9_13, \
call_0_0_9_14, \
call_0_0_9_15, \
call_0_0_10_0, \
call_0_0_10_1, \
call_0_0_10_2, \
call_0_0_10_3, \
call_0_0_10_4, \
call_0_0_10_5, \
call_0_0_10_6, \
call_0_0_10_7, \
call_0_0_10_8, \
call_0_0_10_9, \
call_0_0_10_10, \
call_0_0_10_11, \
call_0_0_10_12, \
call_0_0_10_13, \
call_0_0_10_14, \
call_0_0_10_15, \
call_0_0_11_0, \
call_0_0_11_1, \
call_0_0_11_2, \
call_0_0_11_3, \
call_0_0_11_4, \
call_0_0_11_5, \
call_0_0_11_6, \
call_0_0_11_7, \
call_0_0_11_8, \
call_0_0_11_9, \
call_0_0_11_10, \
call_0_0_11_11, \
call_0_0_11_12, \
call_0_0_11_13, \
call_0_0_11_14, \
call_0_0_11_15, \
call_0_0_12_0, \
call_0_0_12_1, \
call_0_0_12_2, \
call_0_0_12_3, \
call_0_0_12_4, \
call_0_0_12_5, \
call_0_0_12_6, \
call_0_0_12_7, \
call_0_0_12_8, \
call_0_0_12_9, \
call_0_0_12_10, \
call_0_0_12_11, \
call_0_0_12_12, \
call_0_0_12_13, \
call_0_0_12_14, \
call_0_0_12_15, \
call_0_0_13_0, \
call_0_0_13_1, \
call_0_0_13_2, \
call_0_0_13_3, \
call_0_0_13_4, \
call_0_0_13_5, \
call_0_0_13_6, \
call_0_0_13_7, \
call_0_0_13_8, \
call_0_0_13_9, \
call_0_0_13_10, \
call_0_0_13_11, \
call_0_0_13_12, \
call_0_0_13_13, \
call_0_0_13_14, \
call_0_0_13_15, \
call_0_0_14_0, \
call_0_0_14_1, \
call_0_0_14_2, \
call_0_0_14_3, \
call_0_0_14_4, \
call_0_0_14_5, \
call_0_0_14_6, \
call_0_0_14_7, \
call_0_0_14_8, \
call_0_0_14_9, \
call_0_0_14_10, \
call_0_0_14_11, \
call_0_0_14_12, \
call_0_0_14_13, \
call_0_0_14_14, \
call_0_0_14_15, \
call_0_0_15_0, \
call_0_0_15_1, \
call_0_0_15_2, \
call_0_0_15_3, \
call_0_0_15_4, \
call_0_0_15_5, \
call_0_0_15_6, \
call_0_0_15_7, \
call_0_0_15_8, \
call_0_0_15_9, \
call_0_0_15_10, \
call_0_0_15_11, \
call_0_0_15_12, \
call_0_0_15_13, \
call_0_0_15_14, \
call_0_0_15_15, \
call_0_1_0_0, \
call_0_1_0_1, \
call_0_1_0_2, \
call_0_1_0_3, \
call_0_1_0_4, \
call_0_1_0_5, \
call_0_1_0_6, \
call_0_1_0_7, \
call_0_1_0_8, \
call_0_1_0_9, \
call_0_1_0_10, \
call_0_1_0_11, \
call_0_1_0_12, \
call_0_1_0_13, \
call_0_1_0_14, \
call_0_1_0_15, \
call_0_1_1_0, \
call_0_1_1_1, \
call_0_1_1_2, \
call_0_1_1_3, \
call_0_1_1_4, \
call_0_1_1_5, \
call_0_1_1_6, \
call_0_1_1_7, \
call_0_1_1_8, \
call_0_1_1_9, \
call_0_1_1_10, \
call_0_1_1_11, \
call_0_1_1_12, \
call_0_1_1_13, \
call_0_1_1_14, \
call_0_1_1_15, \
call_0_1_2_0, \
call_0_1_2_1, \
call_0_1_2_2, \
call_0_1_2_3, \
call_0_1_2_4, \
call_0_1_2_5, \
call_0_1_2_6, \
call_0_1_2_7, \
call_0_1_2_8, \
call_0_1_2_9, \
call_0_1_2_10, \
call_0_1_2_11, \
call_0_1_2_12, \
call_0_1_2_13, \
call_0_1_2_14, \
call_0_1_2_15, \
call_0_1_3_0, \
call_0_1_3_1, \
call_0_1_3_2, \
call_0_1_3_3, \
call_0_1_3_4, \
call_0_1_3_5, \
call_0_1_3_6, \
call_0_1_3_7, \
call_0_1_3_8, \
call_0_1_3_9, \
call_0_1_3_10, \
call_0_1_3_11, \
call_0_1_3_12, \
call_0_1_3_13, \
call_0_1_3_14, \
call_0_1_3_15, \
call_0_1_4_0, \
call_0_1_4_1, \
call_0_1_4_2, \
call_0_1_4_3, \
call_0_1_4_4, \
call_0_1_4_5, \
call_0_1_4_6, \
call_0_1_4_7, \
call_0_1_4_8, \
call_0_1_4_9, \
call_0_1_4_10, \
call_0_1_4_11, \
call_0_1_4_12, \
call_0_1_4_13, \
call_0_1_4_14, \
call_0_1_4_15, \
call_0_1_5_0, \
call_0_1_5_1, \
call_0_1_5_2, \
call_0_1_5_3, \
call_0_1_5_4, \
call_0_1_5_5, \
call_0_1_5_6, \
call_0_1_5_7, \
call_0_1_5_8, \
call_0_1_5_9, \
call_0_1_5_10, \
call_0_1_5_11, \
call_0_1_5_12, \
call_0_1_5_13, \
call_0_1_5_14, \
call_0_1_5_15, \
call_0_1_6_0, \
call_0_1_6_1, \
call_0_1_6_2, \
call_0_1_6_3, \
call_0_1_6_4, \
call_0_1_6_5, \
call_0_1_6_6, \
call_0_1_6_7, \
call_0_1_6_8, \
call_0_1_6_9, \
call_0_1_6_10, \
call_0_1_6_11, \
call_0_1_6_12, \
call_0_1_6_13, \
call_0_1_6_14, \
call_0_1_6_15, \
call_0_1_7_0, \
call_0_1_7_1, \
call_0_1_7_2, \
call_0_1_7_3, \
call_0_1_7_4, \
call_0_1_7_5, \
call_0_1_7_6, \
call_0_1_7_7, \
call_0_1_7_8, \
call_0_1_7_9, \
call_0_1_7_10, \
call_0_1_7_11, \
call_0_1_7_12, \
call_0_1_7_13, \
call_0_1_7_14, \
call_0_1_7_15, \
call_0_1_8_0, \
call_0_1_8_1, \
call_0_1_8_2, \
call_0_1_8_3, \
call_0_1_8_4, \
call_0_1_8_5, \
call_0_1_8_6, \
call_0_1_8_7, \
call_0_1_8_8, \
call_0_1_8_9, \
call_0_1_8_10, \
call_0_1_8_11, \
call_0_1_8_12, \
call_0_1_8_13, \
call_0_1_8_14, \
call_0_1_8_15, \
call_0_1_9_0, \
call_0_1_9_1, \
call_0_1_9_2, \
call_0_1_9_3, \
call_0_1_9_4, \
call_0_1_9_5, \
call_0_1_9_6, \
call_0_1_9_7, \
call_0_1_9_8, \
call_0_1_9_9, \
call_0_1_9_10, \
call_0_1_9_11, \
call_0_1_9_12, \
call_0_1_9_13, \
call_0_1_9_14, \
call_0_1_9_15, \
call_0_1_10_0, \
call_0_1_10_1, \
call_0_1_10_2, \
call_0_1_10_3, \
call_0_1_10_4, \
call_0_1_10_5, \
call_0_1_10_6, \
call_0_1_10_7, \
call_0_1_10_8, \
call_0_1_10_9, \
call_0_1_10_10, \
call_0_1_10_11, \
call_0_1_10_12, \
call_0_1_10_13, \
call_0_1_10_14, \
call_0_1_10_15, \
call_0_1_11_0, \
call_0_1_11_1, \
call_0_1_11_2, \
call_0_1_11_3, \
call_0_1_11_4, \
call_0_1_11_5, \
call_0_1_11_6, \
call_0_1_11_7, \
call_0_1_11_8, \
call_0_1_11_9, \
call_0_1_11_10, \
call_0_1_11_11, \
call_0_1_11_12, \
call_0_1_11_13, \
call_0_1_11_14, \
call_0_1_11_15, \
call_0_1_12_0, \
call_0_1_12_1, \
call_0_1_12_2, \
call_0_1_12_3, \
call_0_1_12_4, \
call_0_1_12_5, \
call_0_1_12_6, \
call_0_1_12_7, \
call_0_1_12_8, \
call_0_1_12_9, \
call_0_1_12_10, \
call_0_1_12_11, \
call_0_1_12_12, \
call_0_1_12_13, \
call_0_1_12_14, \
call_0_1_12_15, \
call_0_1_13_0, \
call_0_1_13_1, \
call_0_1_13_2, \
call_0_1_13_3, \
call_0_1_13_4, \
call_0_1_13_5, \
call_0_1_13_6, \
call_0_1_13_7, \
call_0_1_13_8, \
call_0_1_13_9, \
call_0_1_13_10, \
call_0_1_13_11, \
call_0_1_13_12, \
call_0_1_13_13, \
call_0_1_13_14, \
call_0_1_13_15, \
call_0_1_14_0, \
call_0_1_14_1, \
call_0_1_14_2, \
call_0_1_14_3, \
call_0_1_14_4, \
call_0_1_14_5, \
call_0_1_14_6, \
call_0_1_14_7, \
call_0_1_14_8, \
call_0_1_14_9, \
call_0_1_14_10, \
call_0_1_14_11, \
call_0_1_14_12, \
call_0_1_14_13, \
call_0_1_14_14, \
call_0_1_14_15, \
call_0_1_15_0, \
call_0_1_15_1, \
call_0_1_15_2, \
call_0_1_15_3, \
call_0_1_15_4, \
call_0_1_15_5, \
call_0_1_15_6, \
call_0_1_15_7, \
call_0_1_15_8, \
call_0_1_15_9, \
call_0_1_15_10, \
call_0_1_15_11, \
call_0_1_15_12, \
call_0_1_15_13, \
call_0_1_15_14, \
call_0_1_15_15, \
call_0_2_0_0, \
call_0_2_0_1, \
call_0_2_0_2, \
call_0_2_0_3, \
call_0_2_0_4, \
call_0_2_0_5, \
call_0_2_0_6, \
call_0_2_0_7, \
call_0_2_0_8, \
call_0_2_0_9, \
call_0_2_0_10, \
call_0_2_0_11, \
call_0_2_0_12, \
call_0_2_0_13, \
call_0_2_0_14, \
call_0_2_0_15, \
call_0_2_1_0, \
call_0_2_1_1, \
call_0_2_1_2, \
call_0_2_1_3, \
call_0_2_1_4, \
call_0_2_1_5, \
call_0_2_1_6, \
call_0_2_1_7, \
call_0_2_1_8, \
call_0_2_1_9, \
call_0_2_1_10, \
call_0_2_1_11, \
call_0_2_1_12, \
call_0_2_1_13, \
call_0_2_1_14, \
call_0_2_1_15, \
call_0_2_2_0, \
call_0_2_2_1, \
call_0_2_2_2, \
call_0_2_2_3, \
call_0_2_2_4, \
call_0_2_2_5, \
call_0_2_2_6, \
call_0_2_2_7, \
call_0_2_2_8, \
call_0_2_2_9, \
call_0_2_2_10, \
call_0_2_2_11, \
call_0_2_2_12, \
call_0_2_2_13, \
call_0_2_2_14, \
call_0_2_2_15, \
call_0_2_3_0, \
call_0_2_3_1, \
call_0_2_3_2, \
call_0_2_3_3, \
call_0_2_3_4, \
call_0_2_3_5, \
call_0_2_3_6, \
call_0_2_3_7, \
call_0_2_3_8, \
call_0_2_3_9, \
call_0_2_3_10, \
call_0_2_3_11, \
call_0_2_3_12, \
call_0_2_3_13, \
call_0_2_3_14, \
call_0_2_3_15, \
call_0_2_4_0, \
call_0_2_4_1, \
call_0_2_4_2, \
call_0_2_4_3, \
call_0_2_4_4, \
call_0_2_4_5, \
call_0_2_4_6, \
call_0_2_4_7, \
call_0_2_4_8, \
call_0_2_4_9, \
call_0_2_4_10, \
call_0_2_4_11, \
call_0_2_4_12, \
call_0_2_4_13, \
call_0_2_4_14, \
call_0_2_4_15, \
call_0_2_5_0, \
call_0_2_5_1, \
call_0_2_5_2, \
call_0_2_5_3, \
call_0_2_5_4, \
call_0_2_5_5, \
call_0_2_5_6, \
call_0_2_5_7, \
call_0_2_5_8, \
call_0_2_5_9, \
call_0_2_5_10, \
call_0_2_5_11, \
call_0_2_5_12, \
call_0_2_5_13, \
call_0_2_5_14, \
call_0_2_5_15, \
call_0_2_6_0, \
call_0_2_6_1, \
call_0_2_6_2, \
call_0_2_6_3, \
call_0_2_6_4, \
call_0_2_6_5, \
call_0_2_6_6, \
call_0_2_6_7, \
call_0_2_6_8, \
call_0_2_6_9, \
call_0_2_6_10, \
call_0_2_6_11, \
call_0_2_6_12, \
call_0_2_6_13, \
call_0_2_6_14, \
call_0_2_6_15, \
call_0_2_7_0, \
call_0_2_7_1, \
call_0_2_7_2, \
call_0_2_7_3, \
call_0_2_7_4, \
call_0_2_7_5, \
call_0_2_7_6, \
call_0_2_7_7, \
call_0_2_7_8, \
call_0_2_7_9, \
call_0_2_7_10, \
call_0_2_7_11, \
call_0_2_7_12, \
call_0_2_7_13, \
call_0_2_7_14, \
call_0_2_7_15, \
call_0_2_8_0, \
call_0_2_8_1, \
call_0_2_8_2, \
call_0_2_8_3, \
call_0_2_8_4, \
call_0_2_8_5, \
call_0_2_8_6, \
call_0_2_8_7, \
call_0_2_8_8, \
call_0_2_8_9, \
call_0_2_8_10, \
call_0_2_8_11, \
call_0_2_8_12, \
call_0_2_8_13, \
call_0_2_8_14, \
call_0_2_8_15, \
call_0_2_9_0, \
call_0_2_9_1, \
call_0_2_9_2, \
call_0_2_9_3, \
call_0_2_9_4, \
call_0_2_9_5, \
call_0_2_9_6, \
call_0_2_9_7, \
call_0_2_9_8, \
call_0_2_9_9, \
call_0_2_9_10, \
call_0_2_9_11, \
call_0_2_9_12, \
call_0_2_9_13, \
call_0_2_9_14, \
call_0_2_9_15, \
call_0_2_10_0, \
call_0_2_10_1, \
call_0_2_10_2, \
call_0_2_10_3, \
call_0_2_10_4, \
call_0_2_10_5, \
call_0_2_10_6, \
call_0_2_10_7, \
call_0_2_10_8, \
call_0_2_10_9, \
call_0_2_10_10, \
call_0_2_10_11, \
call_0_2_10_12, \
call_0_2_10_13, \
call_0_2_10_14, \
call_0_2_10_15, \
call_0_2_11_0, \
call_0_2_11_1, \
call_0_2_11_2, \
call_0_2_11_3, \
call_0_2_11_4, \
call_0_2_11_5, \
call_0_2_11_6, \
call_0_2_11_7, \
call_0_2_11_8, \
call_0_2_11_9, \
call_0_2_11_10, \
call_0_2_11_11, \
call_0_2_11_12, \
call_0_2_11_13, \
call_0_2_11_14, \
call_0_2_11_15, \
call_0_2_12_0, \
call_0_2_12_1, \
call_0_2_12_2, \
call_0_2_12_3, \
call_0_2_12_4, \
call_0_2_12_5, \
call_0_2_12_6, \
call_0_2_12_7, \
call_0_2_12_8, \
call_0_2_12_9, \
call_0_2_12_10, \
call_0_2_12_11, \
call_0_2_12_12, \
call_0_2_12_13, \
call_0_2_12_14, \
call_0_2_12_15, \
call_0_2_13_0, \
call_0_2_13_1, \
call_0_2_13_2, \
call_0_2_13_3, \
call_0_2_13_4, \
call_0_2_13_5, \
call_0_2_13_6, \
call_0_2_13_7, \
call_0_2_13_8, \
call_0_2_13_9, \
call_0_2_13_10, \
call_0_2_13_11, \
call_0_2_13_12, \
call_0_2_13_13, \
call_0_2_13_14, \
call_0_2_13_15, \
call_0_2_14_0, \
call_0_2_14_1, \
call_0_2_14_2, \
call_0_2_14_3, \
call_0_2_14_4, \
call_0_2_14_5, \
call_0_2_14_6, \
call_0_2_14_7, \
call_0_2_14_8, \
call_0_2_14_9, \
call_0_2_14_10, \
call_0_2_14_11, \
call_0_2_14_12, \
call_0_2_14_13, \
call_0_2_14_14, \
call_0_2_14_15, \
call_0_2_15_0, \
call_0_2_15_1, \
call_0_2_15_2, \
call_0_2_15_3, \
call_0_2_15_4, \
call_0_2_15_5, \
call_0_2_15_6, \
call_0_2_15_7, \
call_0_2_15_8, \
call_0_2_15_9, \
call_0_2_15_10, \
call_0_2_15_11, \
call_0_2_15_12, \
call_0_2_15_13, \
call_0_2_15_14, \
call_0_2_15_15, \
call_0_3_0_0, \
call_0_3_0_1, \
call_0_3_0_2, \
call_0_3_0_3, \
call_0_3_0_4, \
call_0_3_0_5, \
call_0_3_0_6, \
call_0_3_0_7, \
call_0_3_0_8, \
call_0_3_0_9, \
call_0_3_0_10, \
call_0_3_0_11, \
call_0_3_0_12, \
call_0_3_0_13, \
call_0_3_0_14, \
call_0_3_0_15, \
call_0_3_1_0, \
call_0_3_1_1, \
call_0_3_1_2, \
call_0_3_1_3, \
call_0_3_1_4, \
call_0_3_1_5, \
call_0_3_1_6, \
call_0_3_1_7, \
call_0_3_1_8, \
call_0_3_1_9, \
call_0_3_1_10, \
call_0_3_1_11, \
call_0_3_1_12, \
call_0_3_1_13, \
call_0_3_1_14, \
call_0_3_1_15, \
call_0_3_2_0, \
call_0_3_2_1, \
call_0_3_2_2, \
call_0_3_2_3, \
call_0_3_2_4, \
call_0_3_2_5, \
call_0_3_2_6, \
call_0_3_2_7, \
call_0_3_2_8, \
call_0_3_2_9, \
call_0_3_2_10, \
call_0_3_2_11, \
call_0_3_2_12, \
call_0_3_2_13, \
call_0_3_2_14, \
call_0_3_2_15, \
call_0_3_3_0, \
call_0_3_3_1, \
call_0_3_3_2, \
call_0_3_3_3, \
call_0_3_3_4, \
call_0_3_3_5, \
call_0_3_3_6, \
call_0_3_3_7, \
call_0_3_3_8, \
call_0_3_3_9, \
call_0_3_3_10, \
call_0_3_3_11, \
call_0_3_3_12, \
call_0_3_3_13, \
call_0_3_3_14, \
call_0_3_3_15, \
call_0_3_4_0, \
call_0_3_4_1, \
call_0_3_4_2, \
call_0_3_4_3, \
call_0_3_4_4, \
call_0_3_4_5, \
call_0_3_4_6, \
call_0_3_4_7, \
call_0_3_4_8, \
call_0_3_4_9, \
call_0_3_4_10, \
call_0_3_4_11, \
call_0_3_4_12, \
call_0_3_4_13, \
call_0_3_4_14, \
call_0_3_4_15, \
call_0_3_5_0, \
call_0_3_5_1, \
call_0_3_5_2, \
call_0_3_5_3, \
call_0_3_5_4, \
call_0_3_5_5, \
call_0_3_5_6, \
call_0_3_5_7, \
call_0_3_5_8, \
call_0_3_5_9, \
call_0_3_5_10, \
call_0_3_5_11, \
call_0_3_5_12, \
call_0_3_5_13, \
call_0_3_5_14, \
call_0_3_5_15, \
call_0_3_6_0, \
call_0_3_6_1, \
call_0_3_6_2, \
call_0_3_6_3, \
call_0_3_6_4, \
call_0_3_6_5, \
call_0_3_6_6, \
call_0_3_6_7, \
call_0_3_6_8, \
call_0_3_6_9, \
call_0_3_6_10, \
call_0_3_6_11, \
call_0_3_6_12, \
call_0_3_6_13, \
call_0_3_6_14, \
call_0_3_6_15, \
call_0_3_7_0, \
call_0_3_7_1, \
call_0_3_7_2, \
call_0_3_7_3, \
call_0_3_7_4, \
call_0_3_7_5, \
call_0_3_7_6, \
call_0_3_7_7, \
call_0_3_7_8, \
call_0_3_7_9, \
call_0_3_7_10, \
call_0_3_7_11, \
call_0_3_7_12, \
call_0_3_7_13, \
call_0_3_7_14, \
call_0_3_7_15, \
call_0_3_8_0, \
call_0_3_8_1, \
call_0_3_8_2, \
call_0_3_8_3, \
call_0_3_8_4, \
call_0_3_8_5, \
call_0_3_8_6, \
call_0_3_8_7, \
call_0_3_8_8, \
call_0_3_8_9, \
call_0_3_8_10, \
call_0_3_8_11, \
call_0_3_8_12, \
call_0_3_8_13, \
call_0_3_8_14, \
call_0_3_8_15, \
call_0_3_9_0, \
call_0_3_9_1, \
call_0_3_9_2, \
call_0_3_9_3, \
call_0_3_9_4, \
call_0_3_9_5, \
call_0_3_9_6, \
call_0_3_9_7, \
call_0_3_9_8, \
call_0_3_9_9, \
call_0_3_9_10, \
call_0_3_9_11, \
call_0_3_9_12, \
call_0_3_9_13, \
call_0_3_9_14, \
call_0_3_9_15, \
call_0_3_10_0, \
call_0_3_10_1, \
call_0_3_10_2, \
call_0_3_10_3, \
call_0_3_10_4, \
call_0_3_10_5, \
call_0_3_10_6, \
call_0_3_10_7, \
call_0_3_10_8, \
call_0_3_10_9, \
call_0_3_10_10, \
call_0_3_10_11, \
call_0_3_10_12, \
call_0_3_10_13, \
call_0_3_10_14, \
call_0_3_10_15, \
call_0_3_11_0, \
call_0_3_11_1, \
call_0_3_11_2, \
call_0_3_11_3, \
call_0_3_11_4, \
call_0_3_11_5, \
call_0_3_11_6, \
call_0_3_11_7, \
call_0_3_11_8, \
call_0_3_11_9, \
call_0_3_11_10, \
call_0_3_11_11, \
call_0_3_11_12, \
call_0_3_11_13, \
call_0_3_11_14, \
call_0_3_11_15, \
call_0_3_12_0, \
call_0_3_12_1, \
call_0_3_12_2, \
call_0_3_12_3, \
call_0_3_12_4, \
call_0_3_12_5, \
call_0_3_12_6, \
call_0_3_12_7, \
call_0_3_12_8, \
call_0_3_12_9, \
call_0_3_12_10, \
call_0_3_12_11, \
call_0_3_12_12, \
call_0_3_12_13, \
call_0_3_12_14, \
call_0_3_12_15, \
call_0_3_13_0, \
call_0_3_13_1, \
call_0_3_13_2, \
call_0_3_13_3, \
call_0_3_13_4, \
call_0_3_13_5, \
call_0_3_13_6, \
call_0_3_13_7, \
call_0_3_13_8, \
call_0_3_13_9, \
call_0_3_13_10, \
call_0_3_13_11, \
call_0_3_13_12, \
call_0_3_13_13, \
call_0_3_13_14, \
call_0_3_13_15, \
call_0_3_14_0, \
call_0_3_14_1, \
call_0_3_14_2, \
call_0_3_14_3, \
call_0_3_14_4, \
call_0_3_14_5, \
call_0_3_14_6, \
call_0_3_14_7, \
call_0_3_14_8, \
call_0_3_14_9, \
call_0_3_14_10, \
call_0_3_14_11, \
call_0_3_14_12, \
call_0_3_14_13, \
call_0_3_14_14, \
call_0_3_14_15, \
call_0_3_15_0, \
call_0_3_15_1, \
call_0_3_15_2, \
call_0_3_15_3, \
call_0_3_15_4, \
call_0_3_15_5, \
call_0_3_15_6, \
call_0_3_15_7, \
call_0_3_15_8, \
call_0_3_15_9, \
call_0_3_15_10, \
call_0_3_15_11, \
call_0_3_15_12, \
call_0_3_15_13, \
call_0_3_15_14, \
call_0_3_15_15, \
call_0_4_0_0, \
call_0_4_0_1, \
call_0_4_0_2, \
call_0_4_0_3, \
call_0_4_0_4, \
call_0_4_0_5, \
call_0_4_0_6, \
call_0_4_0_7, \
call_0_4_0_8, \
call_0_4_0_9, \
call_0_4_0_10, \
call_0_4_0_11, \
call_0_4_0_12, \
call_0_4_0_13, \
call_0_4_0_14, \
call_0_4_0_15, \
call_0_4_1_0, \
call_0_4_1_1, \
call_0_4_1_2, \
call_0_4_1_3, \
call_0_4_1_4, \
call_0_4_1_5, \
call_0_4_1_6, \
call_0_4_1_7, \
call_0_4_1_8, \
call_0_4_1_9, \
call_0_4_1_10, \
call_0_4_1_11, \
call_0_4_1_12, \
call_0_4_1_13, \
call_0_4_1_14, \
call_0_4_1_15, \
call_0_4_2_0, \
call_0_4_2_1, \
call_0_4_2_2, \
call_0_4_2_3, \
call_0_4_2_4, \
call_0_4_2_5, \
call_0_4_2_6, \
call_0_4_2_7, \
call_0_4_2_8, \
call_0_4_2_9, \
call_0_4_2_10, \
call_0_4_2_11, \
call_0_4_2_12, \
call_0_4_2_13, \
call_0_4_2_14, \
call_0_4_2_15, \
call_0_4_3_0, \
call_0_4_3_1, \
call_0_4_3_2, \
call_0_4_3_3, \
call_0_4_3_4, \
call_0_4_3_5, \
call_0_4_3_6, \
call_0_4_3_7, \
call_0_4_3_8, \
call_0_4_3_9, \
call_0_4_3_10, \
call_0_4_3_11, \
call_0_4_3_12, \
call_0_4_3_13, \
call_0_4_3_14, \
call_0_4_3_15, \
call_0_4_4_0, \
call_0_4_4_1, \
call_0_4_4_2, \
call_0_4_4_3, \
call_0_4_4_4, \
call_0_4_4_5, \
call_0_4_4_6, \
call_0_4_4_7, \
call_0_4_4_8, \
call_0_4_4_9, \
call_0_4_4_10, \
call_0_4_4_11, \
call_0_4_4_12, \
call_0_4_4_13, \
call_0_4_4_14, \
call_0_4_4_15, \
call_0_4_5_0, \
call_0_4_5_1, \
call_0_4_5_2, \
call_0_4_5_3, \
call_0_4_5_4, \
call_0_4_5_5, \
call_0_4_5_6, \
call_0_4_5_7, \
call_0_4_5_8, \
call_0_4_5_9, \
call_0_4_5_10, \
call_0_4_5_11, \
call_0_4_5_12, \
call_0_4_5_13, \
call_0_4_5_14, \
call_0_4_5_15, \
call_0_4_6_0, \
call_0_4_6_1, \
call_0_4_6_2, \
call_0_4_6_3, \
call_0_4_6_4, \
call_0_4_6_5, \
call_0_4_6_6, \
call_0_4_6_7, \
call_0_4_6_8, \
call_0_4_6_9, \
call_0_4_6_10, \
call_0_4_6_11, \
call_0_4_6_12, \
call_0_4_6_13, \
call_0_4_6_14, \
call_0_4_6_15, \
call_0_4_7_0, \
call_0_4_7_1, \
call_0_4_7_2, \
call_0_4_7_3, \
call_0_4_7_4, \
call_0_4_7_5, \
call_0_4_7_6, \
call_0_4_7_7, \
call_0_4_7_8, \
call_0_4_7_9, \
call_0_4_7_10, \
call_0_4_7_11, \
call_0_4_7_12, \
call_0_4_7_13, \
call_0_4_7_14, \
call_0_4_7_15, \
call_0_4_8_0, \
call_0_4_8_1, \
call_0_4_8_2, \
call_0_4_8_3, \
call_0_4_8_4, \
call_0_4_8_5, \
call_0_4_8_6, \
call_0_4_8_7, \
call_0_4_8_8, \
call_0_4_8_9, \
call_0_4_8_10, \
call_0_4_8_11, \
call_0_4_8_12, \
call_0_4_8_13, \
call_0_4_8_14, \
call_0_4_8_15, \
call_0_4_9_0, \
call_0_4_9_1, \
call_0_4_9_2, \
call_0_4_9_3, \
call_0_4_9_4, \
call_0_4_9_5, \
call_0_4_9_6, \
call_0_4_9_7, \
call_0_4_9_8, \
call_0_4_9_9, \
call_0_4_9_10, \
call_0_4_9_11, \
call_0_4_9_12, \
call_0_4_9_13, \
call_0_4_9_14, \
call_0_4_9_15, \
call_0_4_10_0, \
call_0_4_10_1, \
call_0_4_10_2, \
call_0_4_10_3, \
call_0_4_10_4, \
call_0_4_10_5, \
call_0_4_10_6, \
call_0_4_10_7, \
call_0_4_10_8, \
call_0_4_10_9, \
call_0_4_10_10, \
call_0_4_10_11, \
call_0_4_10_12, \
call_0_4_10_13, \
call_0_4_10_14, \
call_0_4_10_15, \
call_0_4_11_0, \
call_0_4_11_1, \
call_0_4_11_2, \
call_0_4_11_3, \
call_0_4_11_4, \
call_0_4_11_5, \
call_0_4_11_6, \
call_0_4_11_7, \
call_0_4_11_8, \
call_0_4_11_9, \
call_0_4_11_10, \
call_0_4_11_11, \
call_0_4_11_12, \
call_0_4_11_13, \
call_0_4_11_14, \
call_0_4_11_15, \
call_0_4_12_0, \
call_0_4_12_1, \
call_0_4_12_2, \
call_0_4_12_3, \
call_0_4_12_4, \
call_0_4_12_5, \
call_0_4_12_6, \
call_0_4_12_7, \
call_0_4_12_8, \
call_0_4_12_9, \
call_0_4_12_10, \
call_0_4_12_11, \
call_0_4_12_12, \
call_0_4_12_13, \
call_0_4_12_14, \
call_0_4_12_15, \
call_0_4_13_0, \
call_0_4_13_1, \
call_0_4_13_2, \
call_0_4_13_3, \
call_0_4_13_4, \
call_0_4_13_5, \
call_0_4_13_6, \
call_0_4_13_7, \
call_0_4_13_8, \
call_0_4_13_9, \
call_0_4_13_10, \
call_0_4_13_11, \
call_0_4_13_12, \
call_0_4_13_13, \
call_0_4_13_14, \
call_0_4_13_15, \
call_0_4_14_0, \
call_0_4_14_1, \
call_0_4_14_2, \
call_0_4_14_3, \
call_0_4_14_4, \
call_0_4_14_5, \
call_0_4_14_6, \
call_0_4_14_7, \
call_0_4_14_8, \
call_0_4_14_9, \
call_0_4_14_10, \
call_0_4_14_11, \
call_0_4_14_12, \
call_0_4_14_13, \
call_0_4_14_14, \
call_0_4_14_15, \
call_0_4_15_0, \
call_0_4_15_1, \
call_0_4_15_2, \
call_0_4_15_3, \
call_0_4_15_4, \
call_0_4_15_5, \
call_0_4_15_6, \
call_0_4_15_7, \
call_0_4_15_8, \
call_0_4_15_9, \
call_0_4_15_10, \
call_0_4_15_11, \
call_0_4_15_12, \
call_0_4_15_13, \
call_0_4_15_14, \
call_0_4_15_15, \
call_0_5_0_0, \
call_0_5_0_1, \
call_0_5_0_2, \
call_0_5_0_3, \
call_0_5_0_4, \
call_0_5_0_5, \
call_0_5_0_6, \
call_0_5_0_7, \
call_0_5_0_8, \
call_0_5_0_9, \
call_0_5_0_10, \
call_0_5_0_11, \
call_0_5_0_12, \
call_0_5_0_13, \
call_0_5_0_14, \
call_0_5_0_15, \
call_0_5_1_0, \
call_0_5_1_1, \
call_0_5_1_2, \
call_0_5_1_3, \
call_0_5_1_4, \
call_0_5_1_5, \
call_0_5_1_6, \
call_0_5_1_7, \
call_0_5_1_8, \
call_0_5_1_9, \
call_0_5_1_10, \
call_0_5_1_11, \
call_0_5_1_12, \
call_0_5_1_13, \
call_0_5_1_14, \
call_0_5_1_15, \
call_0_5_2_0, \
call_0_5_2_1, \
call_0_5_2_2, \
call_0_5_2_3, \
call_0_5_2_4, \
call_0_5_2_5, \
call_0_5_2_6, \
call_0_5_2_7, \
call_0_5_2_8, \
call_0_5_2_9, \
call_0_5_2_10, \
call_0_5_2_11, \
call_0_5_2_12, \
call_0_5_2_13, \
call_0_5_2_14, \
call_0_5_2_15, \
call_0_5_3_0, \
call_0_5_3_1, \
call_0_5_3_2, \
call_0_5_3_3, \
call_0_5_3_4, \
call_0_5_3_5, \
call_0_5_3_6, \
call_0_5_3_7, \
call_0_5_3_8, \
call_0_5_3_9, \
call_0_5_3_10, \
call_0_5_3_11, \
call_0_5_3_12, \
call_0_5_3_13, \
call_0_5_3_14, \
call_0_5_3_15, \
call_0_5_4_0, \
call_0_5_4_1, \
call_0_5_4_2, \
call_0_5_4_3, \
call_0_5_4_4, \
call_0_5_4_5, \
call_0_5_4_6, \
call_0_5_4_7, \
call_0_5_4_8, \
call_0_5_4_9, \
call_0_5_4_10, \
call_0_5_4_11, \
call_0_5_4_12, \
call_0_5_4_13, \
call_0_5_4_14, \
call_0_5_4_15, \
call_0_5_5_0, \
call_0_5_5_1, \
call_0_5_5_2, \
call_0_5_5_3, \
call_0_5_5_4, \
call_0_5_5_5, \
call_0_5_5_6, \
call_0_5_5_7, \
call_0_5_5_8, \
call_0_5_5_9, \
call_0_5_5_10, \
call_0_5_5_11, \
call_0_5_5_12, \
call_0_5_5_13, \
call_0_5_5_14, \
call_0_5_5_15, \
call_0_5_6_0, \
call_0_5_6_1, \
call_0_5_6_2, \
call_0_5_6_3, \
call_0_5_6_4, \
call_0_5_6_5, \
call_0_5_6_6, \
call_0_5_6_7, \
call_0_5_6_8, \
call_0_5_6_9, \
call_0_5_6_10, \
call_0_5_6_11, \
call_0_5_6_12, \
call_0_5_6_13, \
call_0_5_6_14, \
call_0_5_6_15, \
call_0_5_7_0, \
call_0_5_7_1, \
call_0_5_7_2, \
call_0_5_7_3, \
call_0_5_7_4, \
call_0_5_7_5, \
call_0_5_7_6, \
call_0_5_7_7, \
call_0_5_7_8, \
call_0_5_7_9, \
call_0_5_7_10, \
call_0_5_7_11, \
call_0_5_7_12, \
call_0_5_7_13, \
call_0_5_7_14, \
call_0_5_7_15, \
call_0_5_8_0, \
call_0_5_8_1, \
call_0_5_8_2, \
call_0_5_8_3, \
call_0_5_8_4, \
call_0_5_8_5, \
call_0_5_8_6, \
call_0_5_8_7, \
call_0_5_8_8, \
call_0_5_8_9, \
call_0_5_8_10, \
call_0_5_8_11, \
call_0_5_8_12, \
call_0_5_8_13, \
call_0_5_8_14, \
call_0_5_8_15, \
call_0_5_9_0, \
call_0_5_9_1, \
call_0_5_9_2, \
call_0_5_9_3, \
call_0_5_9_4, \
call_0_5_9_5, \
call_0_5_9_6, \
call_0_5_9_7, \
call_0_5_9_8, \
call_0_5_9_9, \
call_0_5_9_10, \
call_0_5_9_11, \
call_0_5_9_12, \
call_0_5_9_13, \
call_0_5_9_14, \
call_0_5_9_15, \
call_0_5_10_0, \
call_0_5_10_1, \
call_0_5_10_2, \
call_0_5_10_3, \
call_0_5_10_4, \
call_0_5_10_5, \
call_0_5_10_6, \
call_0_5_10_7, \
call_0_5_10_8, \
call_0_5_10_9, \
call_0_5_10_10, \
call_0_5_10_11, \
call_0_5_10_12, \
call_0_5_10_13, \
call_0_5_10_14, \
call_0_5_10_15, \
call_0_5_11_0, \
call_0_5_11_1, \
call_0_5_11_2, \
call_0_5_11_3, \
call_0_5_11_4, \
call_0_5_11_5, \
call_0_5_11_6, \
call_0_5_11_7, \
call_0_5_11_8, \
call_0_5_11_9, \
call_0_5_11_10, \
call_0_5_11_11, \
call_0_5_11_12, \
call_0_5_11_13, \
call_0_5_11_14, \
call_0_5_11_15, \
call_0_5_12_0, \
call_0_5_12_1, \
call_0_5_12_2, \
call_0_5_12_3, \
call_0_5_12_4, \
call_0_5_12_5, \
call_0_5_12_6, \
call_0_5_12_7, \
call_0_5_12_8, \
call_0_5_12_9, \
call_0_5_12_10, \
call_0_5_12_11, \
call_0_5_12_12, \
call_0_5_12_13, \
call_0_5_12_14, \
call_0_5_12_15, \
call_0_5_13_0, \
call_0_5_13_1, \
call_0_5_13_2, \
call_0_5_13_3, \
call_0_5_13_4, \
call_0_5_13_5, \
call_0_5_13_6, \
call_0_5_13_7, \
call_0_5_13_8, \
call_0_5_13_9, \
call_0_5_13_10, \
call_0_5_13_11, \
call_0_5_13_12, \
call_0_5_13_13, \
call_0_5_13_14, \
call_0_5_13_15, \
call_0_5_14_0, \
call_0_5_14_1, \
call_0_5_14_2, \
call_0_5_14_3, \
call_0_5_14_4, \
call_0_5_14_5, \
call_0_5_14_6, \
call_0_5_14_7, \
call_0_5_14_8, \
call_0_5_14_9, \
call_0_5_14_10, \
call_0_5_14_11, \
call_0_5_14_12, \
call_0_5_14_13, \
call_0_5_14_14, \
call_0_5_14_15, \
call_0_5_15_0, \
call_0_5_15_1, \
call_0_5_15_2, \
call_0_5_15_3, \
call_0_5_15_4, \
call_0_5_15_5, \
call_0_5_15_6, \
call_0_5_15_7, \
call_0_5_15_8, \
call_0_5_15_9, \
call_0_5_15_10, \
call_0_5_15_11, \
call_0_5_15_12, \
call_0_5_15_13, \
call_0_5_15_14, \
call_0_5_15_15, \
call_0_6_0_0, \
call_0_6_0_1, \
call_0_6_0_2, \
call_0_6_0_3, \
call_0_6_0_4, \
call_0_6_0_5, \
call_0_6_0_6, \
call_0_6_0_7, \
call_0_6_0_8, \
call_0_6_0_9, \
call_0_6_0_10, \
call_0_6_0_11, \
call_0_6_0_12, \
call_0_6_0_13, \
call_0_6_0_14, \
call_0_6_0_15, \
call_0_6_1_0, \
call_0_6_1_1, \
call_0_6_1_2, \
call_0_6_1_3, \
call_0_6_1_4, \
call_0_6_1_5, \
call_0_6_1_6, \
call_0_6_1_7, \
call_0_6_1_8, \
call_0_6_1_9, \
call_0_6_1_10, \
call_0_6_1_11, \
call_0_6_1_12, \
call_0_6_1_13, \
call_0_6_1_14, \
call_0_6_1_15, \
call_0_6_2_0, \
call_0_6_2_1, \
call_0_6_2_2, \
call_0_6_2_3, \
call_0_6_2_4, \
call_0_6_2_5, \
call_0_6_2_6, \
call_0_6_2_7, \
call_0_6_2_8, \
call_0_6_2_9, \
call_0_6_2_10, \
call_0_6_2_11, \
call_0_6_2_12, \
call_0_6_2_13, \
call_0_6_2_14, \
call_0_6_2_15, \
call_0_6_3_0, \
call_0_6_3_1, \
call_0_6_3_2, \
call_0_6_3_3, \
call_0_6_3_4, \
call_0_6_3_5, \
call_0_6_3_6, \
call_0_6_3_7, \
call_0_6_3_8, \
call_0_6_3_9, \
call_0_6_3_10, \
call_0_6_3_11, \
call_0_6_3_12, \
call_0_6_3_13, \
call_0_6_3_14, \
call_0_6_3_15, \
call_0_6_4_0, \
call_0_6_4_1, \
call_0_6_4_2, \
call_0_6_4_3, \
call_0_6_4_4, \
call_0_6_4_5, \
call_0_6_4_6, \
call_0_6_4_7, \
call_0_6_4_8, \
call_0_6_4_9, \
call_0_6_4_10, \
call_0_6_4_11, \
call_0_6_4_12, \
call_0_6_4_13, \
call_0_6_4_14, \
call_0_6_4_15, \
call_0_6_5_0, \
call_0_6_5_1, \
call_0_6_5_2, \
call_0_6_5_3, \
call_0_6_5_4, \
call_0_6_5_5, \
call_0_6_5_6, \
call_0_6_5_7, \
call_0_6_5_8, \
call_0_6_5_9, \
call_0_6_5_10, \
call_0_6_5_11, \
call_0_6_5_12, \
call_0_6_5_13, \
call_0_6_5_14, \
call_0_6_5_15, \
call_0_6_6_0, \
call_0_6_6_1, \
call_0_6_6_2, \
call_0_6_6_3, \
call_0_6_6_4, \
call_0_6_6_5, \
call_0_6_6_6, \
call_0_6_6_7, \
call_0_6_6_8, \
call_0_6_6_9, \
call_0_6_6_10, \
call_0_6_6_11, \
call_0_6_6_12, \
call_0_6_6_13, \
call_0_6_6_14, \
call_0_6_6_15, \
call_0_6_7_0, \
call_0_6_7_1, \
call_0_6_7_2, \
call_0_6_7_3, \
call_0_6_7_4, \
call_0_6_7_5, \
call_0_6_7_6, \
call_0_6_7_7, \
call_0_6_7_8, \
call_0_6_7_9, \
call_0_6_7_10, \
call_0_6_7_11, \
call_0_6_7_12, \
call_0_6_7_13, \
call_0_6_7_14, \
call_0_6_7_15, \
call_0_6_8_0, \
call_0_6_8_1, \
call_0_6_8_2, \
call_0_6_8_3, \
call_0_6_8_4, \
call_0_6_8_5, \
call_0_6_8_6, \
call_0_6_8_7, \
call_0_6_8_8, \
call_0_6_8_9, \
call_0_6_8_10, \
call_0_6_8_11, \
call_0_6_8_12, \
call_0_6_8_13, \
call_0_6_8_14, \
call_0_6_8_15, \
call_0_6_9_0, \
call_0_6_9_1, \
call_0_6_9_2, \
call_0_6_9_3, \
call_0_6_9_4, \
call_0_6_9_5, \
call_0_6_9_6, \
call_0_6_9_7, \
call_0_6_9_8, \
call_0_6_9_9, \
call_0_6_9_10, \
call_0_6_9_11, \
call_0_6_9_12, \
call_0_6_9_13, \
call_0_6_9_14, \
call_0_6_9_15, \
call_0_6_10_0, \
call_0_6_10_1, \
call_0_6_10_2, \
call_0_6_10_3, \
call_0_6_10_4, \
call_0_6_10_5, \
call_0_6_10_6, \
call_0_6_10_7, \
call_0_6_10_8, \
call_0_6_10_9, \
call_0_6_10_10, \
call_0_6_10_11, \
call_0_6_10_12, \
call_0_6_10_13, \
call_0_6_10_14, \
call_0_6_10_15, \
call_0_6_11_0, \
call_0_6_11_1, \
call_0_6_11_2, \
call_0_6_11_3, \
call_0_6_11_4, \
call_0_6_11_5, \
call_0_6_11_6, \
call_0_6_11_7, \
call_0_6_11_8, \
call_0_6_11_9, \
call_0_6_11_10, \
call_0_6_11_11, \
call_0_6_11_12, \
call_0_6_11_13, \
call_0_6_11_14, \
call_0_6_11_15, \
call_0_6_12_0, \
call_0_6_12_1, \
call_0_6_12_2, \
call_0_6_12_3, \
call_0_6_12_4, \
call_0_6_12_5, \
call_0_6_12_6, \
call_0_6_12_7, \
call_0_6_12_8, \
call_0_6_12_9, \
call_0_6_12_10, \
call_0_6_12_11, \
call_0_6_12_12, \
call_0_6_12_13, \
call_0_6_12_14, \
call_0_6_12_15, \
call_0_6_13_0, \
call_0_6_13_1, \
call_0_6_13_2, \
call_0_6_13_3, \
call_0_6_13_4, \
call_0_6_13_5, \
call_0_6_13_6, \
call_0_6_13_7, \
call_0_6_13_8, \
call_0_6_13_9, \
call_0_6_13_10, \
call_0_6_13_11, \
call_0_6_13_12, \
call_0_6_13_13, \
call_0_6_13_14, \
call_0_6_13_15, \
call_0_6_14_0, \
call_0_6_14_1, \
call_0_6_14_2, \
call_0_6_14_3, \
call_0_6_14_4, \
call_0_6_14_5, \
call_0_6_14_6, \
call_0_6_14_7, \
call_0_6_14_8, \
call_0_6_14_9, \
call_0_6_14_10, \
call_0_6_14_11, \
call_0_6_14_12, \
call_0_6_14_13, \
call_0_6_14_14, \
call_0_6_14_15, \
call_0_6_15_0, \
call_0_6_15_1, \
call_0_6_15_2, \
call_0_6_15_3, \
call_0_6_15_4, \
call_0_6_15_5, \
call_0_6_15_6, \
call_0_6_15_7, \
call_0_6_15_8, \
call_0_6_15_9, \
call_0_6_15_10, \
call_0_6_15_11, \
call_0_6_15_12, \
call_0_6_15_13, \
call_0_6_15_14, \
call_0_6_15_15, \
call_0_7_0_0, \
call_0_7_0_1, \
call_0_7_0_2, \
call_0_7_0_3, \
call_0_7_0_4, \
call_0_7_0_5, \
call_0_7_0_6, \
call_0_7_0_7, \
call_0_7_0_8, \
call_0_7_0_9, \
call_0_7_0_10, \
call_0_7_0_11, \
call_0_7_0_12, \
call_0_7_0_13, \
call_0_7_0_14, \
call_0_7_0_15, \
call_0_7_1_0, \
call_0_7_1_1, \
call_0_7_1_2, \
call_0_7_1_3, \
call_0_7_1_4, \
call_0_7_1_5, \
call_0_7_1_6, \
call_0_7_1_7, \
call_0_7_1_8, \
call_0_7_1_9, \
call_0_7_1_10, \
call_0_7_1_11, \
call_0_7_1_12, \
call_0_7_1_13, \
call_0_7_1_14, \
call_0_7_1_15, \
call_0_7_2_0, \
call_0_7_2_1, \
call_0_7_2_2, \
call_0_7_2_3, \
call_0_7_2_4, \
call_0_7_2_5, \
call_0_7_2_6, \
call_0_7_2_7, \
call_0_7_2_8, \
call_0_7_2_9, \
call_0_7_2_10, \
call_0_7_2_11, \
call_0_7_2_12, \
call_0_7_2_13, \
call_0_7_2_14, \
call_0_7_2_15, \
call_0_7_3_0, \
call_0_7_3_1, \
call_0_7_3_2, \
call_0_7_3_3, \
call_0_7_3_4, \
call_0_7_3_5, \
call_0_7_3_6, \
call_0_7_3_7, \
call_0_7_3_8, \
call_0_7_3_9, \
call_0_7_3_10, \
call_0_7_3_11, \
call_0_7_3_12, \
call_0_7_3_13, \
call_0_7_3_14, \
call_0_7_3_15, \
call_0_7_4_0, \
call_0_7_4_1, \
call_0_7_4_2, \
call_0_7_4_3, \
call_0_7_4_4, \
call_0_7_4_5, \
call_0_7_4_6, \
call_0_7_4_7, \
call_0_7_4_8, \
call_0_7_4_9, \
call_0_7_4_10, \
call_0_7_4_11, \
call_0_7_4_12, \
call_0_7_4_13, \
call_0_7_4_14, \
call_0_7_4_15, \
call_0_7_5_0, \
call_0_7_5_1, \
call_0_7_5_2, \
call_0_7_5_3, \
call_0_7_5_4, \
call_0_7_5_5, \
call_0_7_5_6, \
call_0_7_5_7, \
call_0_7_5_8, \
call_0_7_5_9, \
call_0_7_5_10, \
call_0_7_5_11, \
call_0_7_5_12, \
call_0_7_5_13, \
call_0_7_5_14, \
call_0_7_5_15, \
call_0_7_6_0, \
call_0_7_6_1, \
call_0_7_6_2, \
call_0_7_6_3, \
call_0_7_6_4, \
call_0_7_6_5, \
call_0_7_6_6, \
call_0_7_6_7, \
call_0_7_6_8, \
call_0_7_6_9, \
call_0_7_6_10, \
call_0_7_6_11, \
call_0_7_6_12, \
call_0_7_6_13, \
call_0_7_6_14, \
call_0_7_6_15, \
call_0_7_7_0, \
call_0_7_7_1, \
call_0_7_7_2, \
call_0_7_7_3, \
call_0_7_7_4, \
call_0_7_7_5, \
call_0_7_7_6, \
call_0_7_7_7, \
call_0_7_7_8, \
call_0_7_7_9, \
call_0_7_7_10, \
call_0_7_7_11, \
call_0_7_7_12, \
call_0_7_7_13, \
call_0_7_7_14, \
call_0_7_7_15, \
call_0_7_8_0, \
call_0_7_8_1, \
call_0_7_8_2, \
call_0_7_8_3, \
call_0_7_8_4, \
call_0_7_8_5, \
call_0_7_8_6, \
call_0_7_8_7, \
call_0_7_8_8, \
call_0_7_8_9, \
call_0_7_8_10, \
call_0_7_8_11, \
call_0_7_8_12, \
call_0_7_8_13, \
call_0_7_8_14, \
call_0_7_8_15, \
call_0_7_9_0, \
call_0_7_9_1, \
call_0_7_9_2, \
call_0_7_9_3, \
call_0_7_9_4, \
call_0_7_9_5, \
call_0_7_9_6, \
call_0_7_9_7, \
call_0_7_9_8, \
call_0_7_9_9, \
call_0_7_9_10, \
call_0_7_9_11, \
call_0_7_9_12, \
call_0_7_9_13, \
call_0_7_9_14, \
call_0_7_9_15, \
call_0_7_10_0, \
call_0_7_10_1, \
call_0_7_10_2, \
call_0_7_10_3, \
call_0_7_10_4, \
call_0_7_10_5, \
call_0_7_10_6, \
call_0_7_10_7, \
call_0_7_10_8, \
call_0_7_10_9, \
call_0_7_10_10, \
call_0_7_10_11, \
call_0_7_10_12, \
call_0_7_10_13, \
call_0_7_10_14, \
call_0_7_10_15, \
call_0_7_11_0, \
call_0_7_11_1, \
call_0_7_11_2, \
call_0_7_11_3, \
call_0_7_11_4, \
call_0_7_11_5, \
call_0_7_11_6, \
call_0_7_11_7, \
call_0_7_11_8, \
call_0_7_11_9, \
call_0_7_11_10, \
call_0_7_11_11, \
call_0_7_11_12, \
call_0_7_11_13, \
call_0_7_11_14, \
call_0_7_11_15, \
call_0_7_12_0, \
call_0_7_12_1, \
call_0_7_12_2, \
call_0_7_12_3, \
call_0_7_12_4, \
call_0_7_12_5, \
call_0_7_12_6, \
call_0_7_12_7, \
call_0_7_12_8, \
call_0_7_12_9, \
call_0_7_12_10, \
call_0_7_12_11, \
call_0_7_12_12, \
call_0_7_12_13, \
call_0_7_12_14, \
call_0_7_12_15, \
call_0_7_13_0, \
call_0_7_13_1, \
call_0_7_13_2, \
call_0_7_13_3, \
call_0_7_13_4, \
call_0_7_13_5, \
call_0_7_13_6, \
call_0_7_13_7, \
call_0_7_13_8, \
call_0_7_13_9, \
call_0_7_13_10, \
call_0_7_13_11, \
call_0_7_13_12, \
call_0_7_13_13, \
call_0_7_13_14, \
call_0_7_13_15, \
call_0_7_14_0, \
call_0_7_14_1, \
call_0_7_14_2, \
call_0_7_14_3, \
call_0_7_14_4, \
call_0_7_14_5, \
call_0_7_14_6, \
call_0_7_14_7, \
call_0_7_14_8, \
call_0_7_14_9, \
call_0_7_14_10, \
call_0_7_14_11, \
call_0_7_14_12, \
call_0_7_14_13, \
call_0_7_14_14, \
call_0_7_14_15, \
call_0_7_15_0, \
call_0_7_15_1, \
call_0_7_15_2, \
call_0_7_15_3, \
call_0_7_15_4, \
call_0_7_15_5, \
call_0_7_15_6, \
call_0_7_15_7, \
call_0_7_15_8, \
call_0_7_15_9, \
call_0_7_15_10, \
call_0_7_15_11, \
call_0_7_15_12, \
call_0_7_15_13, \
call_0_7_15_14, \
call_0_7_15_15, \
call_0_8_0_0, \
call_0_8_0_1, \
call_0_8_0_2, \
call_0_8_0_3, \
call_0_8_0_4, \
call_0_8_0_5, \
call_0_8_0_6, \
call_0_8_0_7, \
call_0_8_0_8, \
call_0_8_0_9, \
call_0_8_0_10, \
call_0_8_0_11, \
call_0_8_0_12, \
call_0_8_0_13, \
call_0_8_0_14, \
call_0_8_0_15, \
call_0_8_1_0, \
call_0_8_1_1, \
call_0_8_1_2, \
call_0_8_1_3, \
call_0_8_1_4, \
call_0_8_1_5, \
call_0_8_1_6, \
call_0_8_1_7, \
call_0_8_1_8, \
call_0_8_1_9, \
call_0_8_1_10, \
call_0_8_1_11, \
call_0_8_1_12, \
call_0_8_1_13, \
call_0_8_1_14, \
call_0_8_1_15, \
call_0_8_2_0, \
call_0_8_2_1, \
call_0_8_2_2, \
call_0_8_2_3, \
call_0_8_2_4, \
call_0_8_2_5, \
call_0_8_2_6, \
call_0_8_2_7, \
call_0_8_2_8, \
call_0_8_2_9, \
call_0_8_2_10, \
call_0_8_2_11, \
call_0_8_2_12, \
call_0_8_2_13, \
call_0_8_2_14, \
call_0_8_2_15, \
call_0_8_3_0, \
call_0_8_3_1, \
call_0_8_3_2, \
call_0_8_3_3, \
call_0_8_3_4, \
call_0_8_3_5, \
call_0_8_3_6, \
call_0_8_3_7, \
call_0_8_3_8, \
call_0_8_3_9, \
call_0_8_3_10, \
call_0_8_3_11, \
call_0_8_3_12, \
call_0_8_3_13, \
call_0_8_3_14, \
call_0_8_3_15, \
call_0_8_4_0, \
call_0_8_4_1, \
call_0_8_4_2, \
call_0_8_4_3, \
call_0_8_4_4, \
call_0_8_4_5, \
call_0_8_4_6, \
call_0_8_4_7, \
call_0_8_4_8, \
call_0_8_4_9, \
call_0_8_4_10, \
call_0_8_4_11, \
call_0_8_4_12, \
call_0_8_4_13, \
call_0_8_4_14, \
call_0_8_4_15, \
call_0_8_5_0, \
call_0_8_5_1, \
call_0_8_5_2, \
call_0_8_5_3, \
call_0_8_5_4, \
call_0_8_5_5, \
call_0_8_5_6, \
call_0_8_5_7, \
call_0_8_5_8, \
call_0_8_5_9, \
call_0_8_5_10, \
call_0_8_5_11, \
call_0_8_5_12, \
call_0_8_5_13, \
call_0_8_5_14, \
call_0_8_5_15, \
call_0_8_6_0, \
call_0_8_6_1, \
call_0_8_6_2, \
call_0_8_6_3, \
call_0_8_6_4, \
call_0_8_6_5, \
call_0_8_6_6, \
call_0_8_6_7, \
call_0_8_6_8, \
call_0_8_6_9, \
call_0_8_6_10, \
call_0_8_6_11, \
call_0_8_6_12, \
call_0_8_6_13, \
call_0_8_6_14, \
call_0_8_6_15, \
call_0_8_7_0, \
call_0_8_7_1, \
call_0_8_7_2, \
call_0_8_7_3, \
call_0_8_7_4, \
call_0_8_7_5, \
call_0_8_7_6, \
call_0_8_7_7, \
call_0_8_7_8, \
call_0_8_7_9, \
call_0_8_7_10, \
call_0_8_7_11, \
call_0_8_7_12, \
call_0_8_7_13, \
call_0_8_7_14, \
call_0_8_7_15, \
call_0_8_8_0, \
call_0_8_8_1, \
call_0_8_8_2, \
call_0_8_8_3, \
call_0_8_8_4, \
call_0_8_8_5, \
call_0_8_8_6, \
call_0_8_8_7, \
call_0_8_8_8, \
call_0_8_8_9, \
call_0_8_8_10, \
call_0_8_8_11, \
call_0_8_8_12, \
call_0_8_8_13, \
call_0_8_8_14, \
call_0_8_8_15, \
call_0_8_9_0, \
call_0_8_9_1, \
call_0_8_9_2, \
call_0_8_9_3, \
call_0_8_9_4, \
call_0_8_9_5, \
call_0_8_9_6, \
call_0_8_9_7, \
call_0_8_9_8, \
call_0_8_9_9, \
call_0_8_9_10, \
call_0_8_9_11, \
call_0_8_9_12, \
call_0_8_9_13, \
call_0_8_9_14, \
call_0_8_9_15, \
call_0_8_10_0, \
call_0_8_10_1, \
call_0_8_10_2, \
call_0_8_10_3, \
call_0_8_10_4, \
call_0_8_10_5, \
call_0_8_10_6, \
call_0_8_10_7, \
call_0_8_10_8, \
call_0_8_10_9, \
call_0_8_10_10, \
call_0_8_10_11, \
call_0_8_10_12, \
call_0_8_10_13, \
call_0_8_10_14, \
call_0_8_10_15, \
call_0_8_11_0, \
call_0_8_11_1, \
call_0_8_11_2, \
call_0_8_11_3, \
call_0_8_11_4, \
call_0_8_11_5, \
call_0_8_11_6, \
call_0_8_11_7, \
call_0_8_11_8, \
call_0_8_11_9, \
call_0_8_11_10, \
call_0_8_11_11, \
call_0_8_11_12, \
call_0_8_11_13, \
call_0_8_11_14, \
call_0_8_11_15, \
call_0_8_12_0, \
call_0_8_12_1, \
call_0_8_12_2, \
call_0_8_12_3, \
call_0_8_12_4, \
call_0_8_12_5, \
call_0_8_12_6, \
call_0_8_12_7, \
call_0_8_12_8, \
call_0_8_12_9, \
call_0_8_12_10, \
call_0_8_12_11, \
call_0_8_12_12, \
call_0_8_12_13, \
call_0_8_12_14, \
call_0_8_12_15, \
call_0_8_13_0, \
call_0_8_13_1, \
call_0_8_13_2, \
call_0_8_13_3, \
call_0_8_13_4, \
call_0_8_13_5, \
call_0_8_13_6, \
call_0_8_13_7, \
call_0_8_13_8, \
call_0_8_13_9, \
call_0_8_13_10, \
call_0_8_13_11, \
call_0_8_13_12, \
call_0_8_13_13, \
call_0_8_13_14, \
call_0_8_13_15, \
call_0_8_14_0, \
call_0_8_14_1, \
call_0_8_14_2, \
call_0_8_14_3, \
call_0_8_14_4, \
call_0_8_14_5, \
call_0_8_14_6, \
call_0_8_14_7, \
call_0_8_14_8, \
call_0_8_14_9, \
call_0_8_14_10, \
call_0_8_14_11, \
call_0_8_14_12, \
call_0_8_14_13, \
call_0_8_14_14, \
call_0_8_14_15, \
call_0_8_15_0, \
call_0_8_15_1, \
call_0_8_15_2, \
call_0_8_15_3, \
call_0_8_15_4, \
call_0_8_15_5, \
call_0_8_15_6, \
call_0_8_15_7, \
call_0_8_15_8, \
call_0_8_15_9, \
call_0_8_15_10, \
call_0_8_15_11, \
call_0_8_15_12, \
call_0_8_15_13, \
call_0_8_15_14, \
call_0_8_15_15, \
call_0_9_0_0, \
call_0_9_0_1, \
call_0_9_0_2, \
call_0_9_0_3, \
call_0_9_0_4, \
call_0_9_0_5, \
call_0_9_0_6, \
call_0_9_0_7, \
call_0_9_0_8, \
call_0_9_0_9, \
call_0_9_0_10, \
call_0_9_0_11, \
call_0_9_0_12, \
call_0_9_0_13, \
call_0_9_0_14, \
call_0_9_0_15, \
call_0_9_1_0, \
call_0_9_1_1, \
call_0_9_1_2, \
call_0_9_1_3, \
call_0_9_1_4, \
call_0_9_1_5, \
call_0_9_1_6, \
call_0_9_1_7, \
call_0_9_1_8, \
call_0_9_1_9, \
call_0_9_1_10, \
call_0_9_1_11, \
call_0_9_1_12, \
call_0_9_1_13, \
call_0_9_1_14, \
call_0_9_1_15, \
call_0_9_2_0, \
call_0_9_2_1, \
call_0_9_2_2, \
call_0_9_2_3, \
call_0_9_2_4, \
call_0_9_2_5, \
call_0_9_2_6, \
call_0_9_2_7, \
call_0_9_2_8, \
call_0_9_2_9, \
call_0_9_2_10, \
call_0_9_2_11, \
call_0_9_2_12, \
call_0_9_2_13, \
call_0_9_2_14, \
call_0_9_2_15, \
call_0_9_3_0, \
call_0_9_3_1, \
call_0_9_3_2, \
call_0_9_3_3, \
call_0_9_3_4, \
call_0_9_3_5, \
call_0_9_3_6, \
call_0_9_3_7, \
call_0_9_3_8, \
call_0_9_3_9, \
call_0_9_3_10, \
call_0_9_3_11, \
call_0_9_3_12, \
call_0_9_3_13, \
call_0_9_3_14, \
call_0_9_3_15, \
call_0_9_4_0, \
call_0_9_4_1, \
call_0_9_4_2, \
call_0_9_4_3, \
call_0_9_4_4, \
call_0_9_4_5, \
call_0_9_4_6, \
call_0_9_4_7, \
call_0_9_4_8, \
call_0_9_4_9, \
call_0_9_4_10, \
call_0_9_4_11, \
call_0_9_4_12, \
call_0_9_4_13, \
call_0_9_4_14, \
call_0_9_4_15, \
call_0_9_5_0, \
call_0_9_5_1, \
call_0_9_5_2, \
call_0_9_5_3, \
call_0_9_5_4, \
call_0_9_5_5, \
call_0_9_5_6, \
call_0_9_5_7, \
call_0_9_5_8, \
call_0_9_5_9, \
call_0_9_5_10, \
call_0_9_5_11, \
call_0_9_5_12, \
call_0_9_5_13, \
call_0_9_5_14, \
call_0_9_5_15, \
call_0_9_6_0, \
call_0_9_6_1, \
call_0_9_6_2, \
call_0_9_6_3, \
call_0_9_6_4, \
call_0_9_6_5, \
call_0_9_6_6, \
call_0_9_6_7, \
call_0_9_6_8, \
call_0_9_6_9, \
call_0_9_6_10, \
call_0_9_6_11, \
call_0_9_6_12, \
call_0_9_6_13, \
call_0_9_6_14, \
call_0_9_6_15, \
call_0_9_7_0, \
call_0_9_7_1, \
call_0_9_7_2, \
call_0_9_7_3, \
call_0_9_7_4, \
call_0_9_7_5, \
call_0_9_7_6, \
call_0_9_7_7, \
call_0_9_7_8, \
call_0_9_7_9, \
call_0_9_7_10, \
call_0_9_7_11, \
call_0_9_7_12, \
call_0_9_7_13, \
call_0_9_7_14, \
call_0_9_7_15, \
call_0_9_8_0, \
call_0_9_8_1, \
call_0_9_8_2, \
call_0_9_8_3, \
call_0_9_8_4, \
call_0_9_8_5, \
call_0_9_8_6, \
call_0_9_8_7, \
call_0_9_8_8, \
call_0_9_8_9, \
call_0_9_8_10, \
call_0_9_8_11, \
call_0_9_8_12, \
call_0_9_8_13, \
call_0_9_8_14, \
call_0_9_8_15, \
call_0_9_9_0, \
call_0_9_9_1, \
call_0_9_9_2, \
call_0_9_9_3, \
call_0_9_9_4, \
call_0_9_9_5, \
call_0_9_9_6, \
call_0_9_9_7, \
call_0_9_9_8, \
call_0_9_9_9, \
call_0_9_9_10, \
call_0_9_9_11, \
call_0_9_9_12, \
call_0_9_9_13, \
call_0_9_9_14, \
call_0_9_9_15, \
call_0_9_10_0, \
call_0_9_10_1, \
call_0_9_10_2, \
call_0_9_10_3, \
call_0_9_10_4, \
call_0_9_10_5, \
call_0_9_10_6, \
call_0_9_10_7, \
call_0_9_10_8, \
call_0_9_10_9, \
call_0_9_10_10, \
call_0_9_10_11, \
call_0_9_10_12, \
call_0_9_10_13, \
call_0_9_10_14, \
call_0_9_10_15, \
call_0_9_11_0, \
call_0_9_11_1, \
call_0_9_11_2, \
call_0_9_11_3, \
call_0_9_11_4, \
call_0_9_11_5, \
call_0_9_11_6, \
call_0_9_11_7, \
call_0_9_11_8, \
call_0_9_11_9, \
call_0_9_11_10, \
call_0_9_11_11, \
call_0_9_11_12, \
call_0_9_11_13, \
call_0_9_11_14, \
call_0_9_11_15, \
call_0_9_12_0, \
call_0_9_12_1, \
call_0_9_12_2, \
call_0_9_12_3, \
call_0_9_12_4, \
call_0_9_12_5, \
call_0_9_12_6, \
call_0_9_12_7, \
call_0_9_12_8, \
call_0_9_12_9, \
call_0_9_12_10, \
call_0_9_12_11, \
call_0_9_12_12, \
call_0_9_12_13, \
call_0_9_12_14, \
call_0_9_12_15, \
call_0_9_13_0, \
call_0_9_13_1, \
call_0_9_13_2, \
call_0_9_13_3, \
call_0_9_13_4, \
call_0_9_13_5, \
call_0_9_13_6, \
call_0_9_13_7, \
call_0_9_13_8, \
call_0_9_13_9, \
call_0_9_13_10, \
call_0_9_13_11, \
call_0_9_13_12, \
call_0_9_13_13, \
call_0_9_13_14, \
call_0_9_13_15, \
call_0_9_14_0, \
call_0_9_14_1, \
call_0_9_14_2, \
call_0_9_14_3, \
call_0_9_14_4, \
call_0_9_14_5, \
call_0_9_14_6, \
call_0_9_14_7, \
call_0_9_14_8, \
call_0_9_14_9, \
call_0_9_14_10, \
call_0_9_14_11, \
call_0_9_14_12, \
call_0_9_14_13, \
call_0_9_14_14, \
call_0_9_14_15, \
call_0_9_15_0, \
call_0_9_15_1, \
call_0_9_15_2, \
call_0_9_15_3, \
call_0_9_15_4, \
call_0_9_15_5, \
call_0_9_15_6, \
call_0_9_15_7, \
call_0_9_15_8, \
call_0_9_15_9, \
call_0_9_15_10, \
call_0_9_15_11, \
call_0_9_15_12, \
call_0_9_15_13, \
call_0_9_15_14, \
call_0_9_15_15, \
call_0_10_0_0, \
call_0_10_0_1, \
call_0_10_0_2, \
call_0_10_0_3, \
call_0_10_0_4, \
call_0_10_0_5, \
call_0_10_0_6, \
call_0_10_0_7, \
call_0_10_0_8, \
call_0_10_0_9, \
call_0_10_0_10, \
call_0_10_0_11, \
call_0_10_0_12, \
call_0_10_0_13, \
call_0_10_0_14, \
call_0_10_0_15, \
call_0_10_1_0, \
call_0_10_1_1, \
call_0_10_1_2, \
call_0_10_1_3, \
call_0_10_1_4, \
call_0_10_1_5, \
call_0_10_1_6, \
call_0_10_1_7, \
call_0_10_1_8, \
call_0_10_1_9, \
call_0_10_1_10, \
call_0_10_1_11, \
call_0_10_1_12, \
call_0_10_1_13, \
call_0_10_1_14, \
call_0_10_1_15, \
call_0_10_2_0, \
call_0_10_2_1, \
call_0_10_2_2, \
call_0_10_2_3, \
call_0_10_2_4, \
call_0_10_2_5, \
call_0_10_2_6, \
call_0_10_2_7, \
call_0_10_2_8, \
call_0_10_2_9, \
call_0_10_2_10, \
call_0_10_2_11, \
call_0_10_2_12, \
call_0_10_2_13, \
call_0_10_2_14, \
call_0_10_2_15, \
call_0_10_3_0, \
call_0_10_3_1, \
call_0_10_3_2, \
call_0_10_3_3, \
call_0_10_3_4, \
call_0_10_3_5, \
call_0_10_3_6, \
call_0_10_3_7, \
call_0_10_3_8, \
call_0_10_3_9, \
call_0_10_3_10, \
call_0_10_3_11, \
call_0_10_3_12, \
call_0_10_3_13, \
call_0_10_3_14, \
call_0_10_3_15, \
call_0_10_4_0, \
call_0_10_4_1, \
call_0_10_4_2, \
call_0_10_4_3, \
call_0_10_4_4, \
call_0_10_4_5, \
call_0_10_4_6, \
call_0_10_4_7, \
call_0_10_4_8, \
call_0_10_4_9, \
call_0_10_4_10, \
call_0_10_4_11, \
call_0_10_4_12, \
call_0_10_4_13, \
call_0_10_4_14, \
call_0_10_4_15, \
call_0_10_5_0, \
call_0_10_5_1, \
call_0_10_5_2, \
call_0_10_5_3, \
call_0_10_5_4, \
call_0_10_5_5, \
call_0_10_5_6, \
call_0_10_5_7, \
call_0_10_5_8, \
call_0_10_5_9, \
call_0_10_5_10, \
call_0_10_5_11, \
call_0_10_5_12, \
call_0_10_5_13, \
call_0_10_5_14, \
call_0_10_5_15, \
call_0_10_6_0, \
call_0_10_6_1, \
call_0_10_6_2, \
call_0_10_6_3, \
call_0_10_6_4, \
call_0_10_6_5, \
call_0_10_6_6, \
call_0_10_6_7, \
call_0_10_6_8, \
call_0_10_6_9, \
call_0_10_6_10, \
call_0_10_6_11, \
call_0_10_6_12, \
call_0_10_6_13, \
call_0_10_6_14, \
call_0_10_6_15, \
call_0_10_7_0, \
call_0_10_7_1, \
call_0_10_7_2, \
call_0_10_7_3, \
call_0_10_7_4, \
call_0_10_7_5, \
call_0_10_7_6, \
call_0_10_7_7, \
call_0_10_7_8, \
call_0_10_7_9, \
call_0_10_7_10, \
call_0_10_7_11, \
call_0_10_7_12, \
call_0_10_7_13, \
call_0_10_7_14, \
call_0_10_7_15, \
call_0_10_8_0, \
call_0_10_8_1, \
call_0_10_8_2, \
call_0_10_8_3, \
call_0_10_8_4, \
call_0_10_8_5, \
call_0_10_8_6, \
call_0_10_8_7, \
call_0_10_8_8, \
call_0_10_8_9, \
call_0_10_8_10, \
call_0_10_8_11, \
call_0_10_8_12, \
call_0_10_8_13, \
call_0_10_8_14, \
call_0_10_8_15, \
call_0_10_9_0, \
call_0_10_9_1, \
call_0_10_9_2, \
call_0_10_9_3, \
call_0_10_9_4, \
call_0_10_9_5, \
call_0_10_9_6, \
call_0_10_9_7, \
call_0_10_9_8, \
call_0_10_9_9, \
call_0_10_9_10, \
call_0_10_9_11, \
call_0_10_9_12, \
call_0_10_9_13, \
call_0_10_9_14, \
call_0_10_9_15, \
call_0_10_10_0, \
call_0_10_10_1, \
call_0_10_10_2, \
call_0_10_10_3, \
call_0_10_10_4, \
call_0_10_10_5, \
call_0_10_10_6, \
call_0_10_10_7, \
call_0_10_10_8, \
call_0_10_10_9, \
call_0_10_10_10, \
call_0_10_10_11, \
call_0_10_10_12, \
call_0_10_10_13, \
call_0_10_10_14, \
call_0_10_10_15, \
call_0_10_11_0, \
call_0_10_11_1, \
call_0_10_11_2, \
call_0_10_11_3, \
call_0_10_11_4, \
call_0_10_11_5, \
call_0_10_11_6, \
call_0_10_11_7, \
call_0_10_11_8, \
call_0_10_11_9, \
call_0_10_11_10, \
call_0_10_11_11, \
call_0_10_11_12, \
call_0_10_11_13, \
call_0_10_11_14, \
call_0_10_11_15, \
call_0_10_12_0, \
call_0_10_12_1, \
call_0_10_12_2, \
call_0_10_12_3, \
call_0_10_12_4, \
call_0_10_12_5, \
call_0_10_12_6, \
call_0_10_12_7, \
call_0_10_12_8, \
call_0_10_12_9, \
call_0_10_12_10, \
call_0_10_12_11, \
call_0_10_12_12, \
call_0_10_12_13, \
call_0_10_12_14, \
call_0_10_12_15, \
call_0_10_13_0, \
call_0_10_13_1, \
call_0_10_13_2, \
call_0_10_13_3, \
call_0_10_13_4, \
call_0_10_13_5, \
call_0_10_13_6, \
call_0_10_13_7, \
call_0_10_13_8, \
call_0_10_13_9, \
call_0_10_13_10, \
call_0_10_13_11, \
call_0_10_13_12, \
call_0_10_13_13, \
call_0_10_13_14, \
call_0_10_13_15, \
call_0_10_14_0, \
call_0_10_14_1, \
call_0_10_14_2, \
call_0_10_14_3, \
call_0_10_14_4, \
call_0_10_14_5, \
call_0_10_14_6, \
call_0_10_14_7, \
call_0_10_14_8, \
call_0_10_14_9, \
call_0_10_14_10, \
call_0_10_14_11, \
call_0_10_14_12, \
call_0_10_14_13, \
call_0_10_14_14, \
call_0_10_14_15, \
call_0_10_15_0, \
call_0_10_15_1, \
call_0_10_15_2, \
call_0_10_15_3, \
call_0_10_15_4, \
call_0_10_15_5, \
call_0_10_15_6, \
call_0_10_15_7, \
call_0_10_15_8, \
call_0_10_15_9, \
call_0_10_15_10, \
call_0_10_15_11, \
call_0_10_15_12, \
call_0_10_15_13, \
call_0_10_15_14, \
call_0_10_15_15, \
call_0_11_0_0, \
call_0_11_0_1, \
call_0_11_0_2, \
call_0_11_0_3, \
call_0_11_0_4, \
call_0_11_0_5, \
call_0_11_0_6, \
call_0_11_0_7, \
call_0_11_0_8, \
call_0_11_0_9, \
call_0_11_0_10, \
call_0_11_0_11, \
call_0_11_0_12, \
call_0_11_0_13, \
call_0_11_0_14, \
call_0_11_0_15, \
call_0_11_1_0, \
call_0_11_1_1, \
call_0_11_1_2, \
call_0_11_1_3, \
call_0_11_1_4, \
call_0_11_1_5, \
call_0_11_1_6, \
call_0_11_1_7, \
call_0_11_1_8, \
call_0_11_1_9, \
call_0_11_1_10, \
call_0_11_1_11, \
call_0_11_1_12, \
call_0_11_1_13, \
call_0_11_1_14, \
call_0_11_1_15, \
call_0_11_2_0, \
call_0_11_2_1, \
call_0_11_2_2, \
call_0_11_2_3, \
call_0_11_2_4, \
call_0_11_2_5, \
call_0_11_2_6, \
call_0_11_2_7, \
call_0_11_2_8, \
call_0_11_2_9, \
call_0_11_2_10, \
call_0_11_2_11, \
call_0_11_2_12, \
call_0_11_2_13, \
call_0_11_2_14, \
call_0_11_2_15, \
call_0_11_3_0, \
call_0_11_3_1, \
call_0_11_3_2, \
call_0_11_3_3, \
call_0_11_3_4, \
call_0_11_3_5, \
call_0_11_3_6, \
call_0_11_3_7, \
call_0_11_3_8, \
call_0_11_3_9, \
call_0_11_3_10, \
call_0_11_3_11, \
call_0_11_3_12, \
call_0_11_3_13, \
call_0_11_3_14, \
call_0_11_3_15, \
call_0_11_4_0, \
call_0_11_4_1, \
call_0_11_4_2, \
call_0_11_4_3, \
call_0_11_4_4, \
call_0_11_4_5, \
call_0_11_4_6, \
call_0_11_4_7, \
call_0_11_4_8, \
call_0_11_4_9, \
call_0_11_4_10, \
call_0_11_4_11, \
call_0_11_4_12, \
call_0_11_4_13, \
call_0_11_4_14, \
call_0_11_4_15, \
call_0_11_5_0, \
call_0_11_5_1, \
call_0_11_5_2, \
call_0_11_5_3, \
call_0_11_5_4, \
call_0_11_5_5, \
call_0_11_5_6, \
call_0_11_5_7, \
call_0_11_5_8, \
call_0_11_5_9, \
call_0_11_5_10, \
call_0_11_5_11, \
call_0_11_5_12, \
call_0_11_5_13, \
call_0_11_5_14, \
call_0_11_5_15, \
call_0_11_6_0, \
call_0_11_6_1, \
call_0_11_6_2, \
call_0_11_6_3, \
call_0_11_6_4, \
call_0_11_6_5, \
call_0_11_6_6, \
call_0_11_6_7, \
call_0_11_6_8, \
call_0_11_6_9, \
call_0_11_6_10, \
call_0_11_6_11, \
call_0_11_6_12, \
call_0_11_6_13, \
call_0_11_6_14, \
call_0_11_6_15, \
call_0_11_7_0, \
call_0_11_7_1, \
call_0_11_7_2, \
call_0_11_7_3, \
call_0_11_7_4, \
call_0_11_7_5, \
call_0_11_7_6, \
call_0_11_7_7, \
call_0_11_7_8, \
call_0_11_7_9, \
call_0_11_7_10, \
call_0_11_7_11, \
call_0_11_7_12, \
call_0_11_7_13, \
call_0_11_7_14, \
call_0_11_7_15, \
call_0_11_8_0, \
call_0_11_8_1, \
call_0_11_8_2, \
call_0_11_8_3, \
call_0_11_8_4, \
call_0_11_8_5, \
call_0_11_8_6, \
call_0_11_8_7, \
call_0_11_8_8, \
call_0_11_8_9, \
call_0_11_8_10, \
call_0_11_8_11, \
call_0_11_8_12, \
call_0_11_8_13, \
call_0_11_8_14, \
call_0_11_8_15, \
call_0_11_9_0, \
call_0_11_9_1, \
call_0_11_9_2, \
call_0_11_9_3, \
call_0_11_9_4, \
call_0_11_9_5, \
call_0_11_9_6, \
call_0_11_9_7, \
call_0_11_9_8, \
call_0_11_9_9, \
call_0_11_9_10, \
call_0_11_9_11, \
call_0_11_9_12, \
call_0_11_9_13, \
call_0_11_9_14, \
call_0_11_9_15, \
call_0_11_10_0, \
call_0_11_10_1, \
call_0_11_10_2, \
call_0_11_10_3, \
call_0_11_10_4, \
call_0_11_10_5, \
call_0_11_10_6, \
call_0_11_10_7, \
call_0_11_10_8, \
call_0_11_10_9, \
call_0_11_10_10, \
call_0_11_10_11, \
call_0_11_10_12, \
call_0_11_10_13, \
call_0_11_10_14, \
call_0_11_10_15, \
call_0_11_11_0, \
call_0_11_11_1, \
call_0_11_11_2, \
call_0_11_11_3, \
call_0_11_11_4, \
call_0_11_11_5, \
call_0_11_11_6, \
call_0_11_11_7, \
call_0_11_11_8, \
call_0_11_11_9, \
call_0_11_11_10, \
call_0_11_11_11, \
call_0_11_11_12, \
call_0_11_11_13, \
call_0_11_11_14, \
call_0_11_11_15, \
call_0_11_12_0, \
call_0_11_12_1, \
call_0_11_12_2, \
call_0_11_12_3, \
call_0_11_12_4, \
call_0_11_12_5, \
call_0_11_12_6, \
call_0_11_12_7, \
call_0_11_12_8, \
call_0_11_12_9, \
call_0_11_12_10, \
call_0_11_12_11, \
call_0_11_12_12, \
call_0_11_12_13, \
call_0_11_12_14, \
call_0_11_12_15, \
call_0_11_13_0, \
call_0_11_13_1, \
call_0_11_13_2, \
call_0_11_13_3, \
call_0_11_13_4, \
call_0_11_13_5, \
call_0_11_13_6, \
call_0_11_13_7, \
call_0_11_13_8, \
call_0_11_13_9, \
call_0_11_13_10, \
call_0_11_13_11, \
call_0_11_13_12, \
call_0_11_13_13, \
call_0_11_13_14, \
call_0_11_13_15, \
call_0_11_14_0, \
call_0_11_14_1, \
call_0_11_14_2, \
call_0_11_14_3, \
call_0_11_14_4, \
call_0_11_14_5, \
call_0_11_14_6, \
call_0_11_14_7, \
call_0_11_14_8, \
call_0_11_14_9, \
call_0_11_14_10, \
call_0_11_14_11, \
call_0_11_14_12, \
call_0_11_14_13, \
call_0_11_14_14, \
call_0_11_14_15, \
call_0_11_15_0, \
call_0_11_15_1, \
call_0_11_15_2, \
call_0_11_15_3, \
call_0_11_15_4, \
call_0_11_15_5, \
call_0_11_15_6, \
call_0_11_15_7, \
call_0_11_15_8, \
call_0_11_15_9, \
call_0_11_15_10, \
call_0_11_15_11, \
call_0_11_15_12, \
call_0_11_15_13, \
call_0_11_15_14, \
call_0_11_15_15, \
call_0_12_0_0, \
call_0_12_0_1, \
call_0_12_0_2, \
call_0_12_0_3, \
call_0_12_0_4, \
call_0_12_0_5, \
call_0_12_0_6, \
call_0_12_0_7, \
call_0_12_0_8, \
call_0_12_0_9, \
call_0_12_0_10, \
call_0_12_0_11, \
call_0_12_0_12, \
call_0_12_0_13, \
call_0_12_0_14, \
call_0_12_0_15, \
call_0_12_1_0, \
call_0_12_1_1, \
call_0_12_1_2, \
call_0_12_1_3, \
call_0_12_1_4, \
call_0_12_1_5, \
call_0_12_1_6, \
call_0_12_1_7, \
call_0_12_1_8, \
call_0_12_1_9, \
call_0_12_1_10, \
call_0_12_1_11, \
call_0_12_1_12, \
call_0_12_1_13, \
call_0_12_1_14, \
call_0_12_1_15, \
call_0_12_2_0, \
call_0_12_2_1, \
call_0_12_2_2, \
call_0_12_2_3, \
call_0_12_2_4, \
call_0_12_2_5, \
call_0_12_2_6, \
call_0_12_2_7, \
call_0_12_2_8, \
call_0_12_2_9, \
call_0_12_2_10, \
call_0_12_2_11, \
call_0_12_2_12, \
call_0_12_2_13, \
call_0_12_2_14, \
call_0_12_2_15, \
call_0_12_3_0, \
call_0_12_3_1, \
call_0_12_3_2, \
call_0_12_3_3, \
call_0_12_3_4, \
call_0_12_3_5, \
call_0_12_3_6, \
call_0_12_3_7, \
call_0_12_3_8, \
call_0_12_3_9, \
call_0_12_3_10, \
call_0_12_3_11, \
call_0_12_3_12, \
call_0_12_3_13, \
call_0_12_3_14, \
call_0_12_3_15, \
call_0_12_4_0, \
call_0_12_4_1, \
call_0_12_4_2, \
call_0_12_4_3, \
call_0_12_4_4, \
call_0_12_4_5, \
call_0_12_4_6, \
call_0_12_4_7, \
call_0_12_4_8, \
call_0_12_4_9, \
call_0_12_4_10, \
call_0_12_4_11, \
call_0_12_4_12, \
call_0_12_4_13, \
call_0_12_4_14, \
call_0_12_4_15, \
call_0_12_5_0, \
call_0_12_5_1, \
call_0_12_5_2, \
call_0_12_5_3, \
call_0_12_5_4, \
call_0_12_5_5, \
call_0_12_5_6, \
call_0_12_5_7, \
call_0_12_5_8, \
call_0_12_5_9, \
call_0_12_5_10, \
call_0_12_5_11, \
call_0_12_5_12, \
call_0_12_5_13, \
call_0_12_5_14, \
call_0_12_5_15, \
call_0_12_6_0, \
call_0_12_6_1, \
call_0_12_6_2, \
call_0_12_6_3, \
call_0_12_6_4, \
call_0_12_6_5, \
call_0_12_6_6, \
call_0_12_6_7, \
call_0_12_6_8, \
call_0_12_6_9, \
call_0_12_6_10, \
call_0_12_6_11, \
call_0_12_6_12, \
call_0_12_6_13, \
call_0_12_6_14, \
call_0_12_6_15, \
call_0_12_7_0, \
call_0_12_7_1, \
call_0_12_7_2, \
call_0_12_7_3, \
call_0_12_7_4, \
call_0_12_7_5, \
call_0_12_7_6, \
call_0_12_7_7, \
call_0_12_7_8, \
call_0_12_7_9, \
call_0_12_7_10, \
call_0_12_7_11, \
call_0_12_7_12, \
call_0_12_7_13, \
call_0_12_7_14, \
call_0_12_7_15, \
call_0_12_8_0, \
call_0_12_8_1, \
call_0_12_8_2, \
call_0_12_8_3, \
call_0_12_8_4, \
call_0_12_8_5, \
call_0_12_8_6, \
call_0_12_8_7, \
call_0_12_8_8, \
call_0_12_8_9, \
call_0_12_8_10, \
call_0_12_8_11, \
call_0_12_8_12, \
call_0_12_8_13, \
call_0_12_8_14, \
call_0_12_8_15, \
call_0_12_9_0, \
call_0_12_9_1, \
call_0_12_9_2, \
call_0_12_9_3, \
call_0_12_9_4, \
call_0_12_9_5, \
call_0_12_9_6, \
call_0_12_9_7, \
call_0_12_9_8, \
call_0_12_9_9, \
call_0_12_9_10, \
call_0_12_9_11, \
call_0_12_9_12, \
call_0_12_9_13, \
call_0_12_9_14, \
call_0_12_9_15, \
call_0_12_10_0, \
call_0_12_10_1, \
call_0_12_10_2, \
call_0_12_10_3, \
call_0_12_10_4, \
call_0_12_10_5, \
call_0_12_10_6, \
call_0_12_10_7, \
call_0_12_10_8, \
call_0_12_10_9, \
call_0_12_10_10, \
call_0_12_10_11, \
call_0_12_10_12, \
call_0_12_10_13, \
call_0_12_10_14, \
call_0_12_10_15, \
call_0_12_11_0, \
call_0_12_11_1, \
call_0_12_11_2, \
call_0_12_11_3, \
call_0_12_11_4, \
call_0_12_11_5, \
call_0_12_11_6, \
call_0_12_11_7, \
call_0_12_11_8, \
call_0_12_11_9, \
call_0_12_11_10, \
call_0_12_11_11, \
call_0_12_11_12, \
call_0_12_11_13, \
call_0_12_11_14, \
call_0_12_11_15, \
call_0_12_12_0, \
call_0_12_12_1, \
call_0_12_12_2, \
call_0_12_12_3, \
call_0_12_12_4, \
call_0_12_12_5, \
call_0_12_12_6, \
call_0_12_12_7, \
call_0_12_12_8, \
call_0_12_12_9, \
call_0_12_12_10, \
call_0_12_12_11, \
call_0_12_12_12, \
call_0_12_12_13, \
call_0_12_12_14, \
call_0_12_12_15, \
call_0_12_13_0, \
call_0_12_13_1, \
call_0_12_13_2, \
call_0_12_13_3, \
call_0_12_13_4, \
call_0_12_13_5, \
call_0_12_13_6, \
call_0_12_13_7, \
call_0_12_13_8, \
call_0_12_13_9, \
call_0_12_13_10, \
call_0_12_13_11, \
call_0_12_13_12, \
call_0_12_13_13, \
call_0_12_13_14, \
call_0_12_13_15, \
call_0_12_14_0, \
call_0_12_14_1, \
call_0_12_14_2, \
call_0_12_14_3, \
call_0_12_14_4, \
call_0_12_14_5, \
call_0_12_14_6, \
call_0_12_14_7, \
call_0_12_14_8, \
call_0_12_14_9, \
call_0_12_14_10, \
call_0_12_14_11, \
call_0_12_14_12, \
call_0_12_14_13, \
call_0_12_14_14, \
call_0_12_14_15, \
call_0_12_15_0, \
call_0_12_15_1, \
call_0_12_15_2, \
call_0_12_15_3, \
call_0_12_15_4, \
call_0_12_15_5, \
call_0_12_15_6, \
call_0_12_15_7, \
call_0_12_15_8, \
call_0_12_15_9, \
call_0_12_15_10, \
call_0_12_15_11, \
call_0_12_15_12, \
call_0_12_15_13, \
call_0_12_15_14, \
call_0_12_15_15, \
call_0_13_0_0, \
call_0_13_0_1, \
call_0_13_0_2, \
call_0_13_0_3, \
call_0_13_0_4, \
call_0_13_0_5, \
call_0_13_0_6, \
call_0_13_0_7, \
call_0_13_0_8, \
call_0_13_0_9, \
call_0_13_0_10, \
call_0_13_0_11, \
call_0_13_0_12, \
call_0_13_0_13, \
call_0_13_0_14, \
call_0_13_0_15, \
call_0_13_1_0, \
call_0_13_1_1, \
call_0_13_1_2, \
call_0_13_1_3, \
call_0_13_1_4, \
call_0_13_1_5, \
call_0_13_1_6, \
call_0_13_1_7, \
call_0_13_1_8, \
call_0_13_1_9, \
call_0_13_1_10, \
call_0_13_1_11, \
call_0_13_1_12, \
call_0_13_1_13, \
call_0_13_1_14, \
call_0_13_1_15, \
call_0_13_2_0, \
call_0_13_2_1, \
call_0_13_2_2, \
call_0_13_2_3, \
call_0_13_2_4, \
call_0_13_2_5, \
call_0_13_2_6, \
call_0_13_2_7, \
call_0_13_2_8, \
call_0_13_2_9, \
call_0_13_2_10, \
call_0_13_2_11, \
call_0_13_2_12, \
call_0_13_2_13, \
call_0_13_2_14, \
call_0_13_2_15, \
call_0_13_3_0, \
call_0_13_3_1, \
call_0_13_3_2, \
call_0_13_3_3, \
call_0_13_3_4, \
call_0_13_3_5, \
call_0_13_3_6, \
call_0_13_3_7, \
call_0_13_3_8, \
call_0_13_3_9, \
call_0_13_3_10, \
call_0_13_3_11, \
call_0_13_3_12, \
call_0_13_3_13, \
call_0_13_3_14, \
call_0_13_3_15, \
call_0_13_4_0, \
call_0_13_4_1, \
call_0_13_4_2, \
call_0_13_4_3, \
call_0_13_4_4, \
call_0_13_4_5, \
call_0_13_4_6, \
call_0_13_4_7, \
call_0_13_4_8, \
call_0_13_4_9, \
call_0_13_4_10, \
call_0_13_4_11, \
call_0_13_4_12, \
call_0_13_4_13, \
call_0_13_4_14, \
call_0_13_4_15, \
call_0_13_5_0, \
call_0_13_5_1, \
call_0_13_5_2, \
call_0_13_5_3, \
call_0_13_5_4, \
call_0_13_5_5, \
call_0_13_5_6, \
call_0_13_5_7, \
call_0_13_5_8, \
call_0_13_5_9, \
call_0_13_5_10, \
call_0_13_5_11, \
call_0_13_5_12, \
call_0_13_5_13, \
call_0_13_5_14, \
call_0_13_5_15, \
call_0_13_6_0, \
call_0_13_6_1, \
call_0_13_6_2, \
call_0_13_6_3, \
call_0_13_6_4, \
call_0_13_6_5, \
call_0_13_6_6, \
call_0_13_6_7, \
call_0_13_6_8, \
call_0_13_6_9, \
call_0_13_6_10, \
call_0_13_6_11, \
call_0_13_6_12, \
call_0_13_6_13, \
call_0_13_6_14, \
call_0_13_6_15, \
call_0_13_7_0, \
call_0_13_7_1, \
call_0_13_7_2, \
call_0_13_7_3, \
call_0_13_7_4, \
call_0_13_7_5, \
call_0_13_7_6, \
call_0_13_7_7, \
call_0_13_7_8, \
call_0_13_7_9, \
call_0_13_7_10, \
call_0_13_7_11, \
call_0_13_7_12, \
call_0_13_7_13, \
call_0_13_7_14, \
call_0_13_7_15, \
call_0_13_8_0, \
call_0_13_8_1, \
call_0_13_8_2, \
call_0_13_8_3, \
call_0_13_8_4, \
call_0_13_8_5, \
call_0_13_8_6, \
call_0_13_8_7, \
call_0_13_8_8, \
call_0_13_8_9, \
call_0_13_8_10, \
call_0_13_8_11, \
call_0_13_8_12, \
call_0_13_8_13, \
call_0_13_8_14, \
call_0_13_8_15, \
call_0_13_9_0, \
call_0_13_9_1, \
call_0_13_9_2, \
call_0_13_9_3, \
call_0_13_9_4, \
call_0_13_9_5, \
call_0_13_9_6, \
call_0_13_9_7, \
call_0_13_9_8, \
call_0_13_9_9, \
call_0_13_9_10, \
call_0_13_9_11, \
call_0_13_9_12, \
call_0_13_9_13, \
call_0_13_9_14, \
call_0_13_9_15, \
call_0_13_10_0, \
call_0_13_10_1, \
call_0_13_10_2, \
call_0_13_10_3, \
call_0_13_10_4, \
call_0_13_10_5, \
call_0_13_10_6, \
call_0_13_10_7, \
call_0_13_10_8, \
call_0_13_10_9, \
call_0_13_10_10, \
call_0_13_10_11, \
call_0_13_10_12, \
call_0_13_10_13, \
call_0_13_10_14, \
call_0_13_10_15, \
call_0_13_11_0, \
call_0_13_11_1, \
call_0_13_11_2, \
call_0_13_11_3, \
call_0_13_11_4, \
call_0_13_11_5, \
call_0_13_11_6, \
call_0_13_11_7, \
call_0_13_11_8, \
call_0_13_11_9, \
call_0_13_11_10, \
call_0_13_11_11, \
call_0_13_11_12, \
call_0_13_11_13, \
call_0_13_11_14, \
call_0_13_11_15, \
call_0_13_12_0, \
call_0_13_12_1, \
call_0_13_12_2, \
call_0_13_12_3, \
call_0_13_12_4, \
call_0_13_12_5, \
call_0_13_12_6, \
call_0_13_12_7, \
call_0_13_12_8, \
call_0_13_12_9, \
call_0_13_12_10, \
call_0_13_12_11, \
call_0_13_12_12, \
call_0_13_12_13, \
call_0_13_12_14, \
call_0_13_12_15, \
call_0_13_13_0, \
call_0_13_13_1, \
call_0_13_13_2, \
call_0_13_13_3, \
call_0_13_13_4, \
call_0_13_13_5, \
call_0_13_13_6, \
call_0_13_13_7, \
call_0_13_13_8, \
call_0_13_13_9, \
call_0_13_13_10, \
call_0_13_13_11, \
call_0_13_13_12, \
call_0_13_13_13, \
call_0_13_13_14, \
call_0_13_13_15, \
call_0_13_14_0, \
call_0_13_14_1, \
call_0_13_14_2, \
call_0_13_14_3, \
call_0_13_14_4, \
call_0_13_14_5, \
call_0_13_14_6, \
call_0_13_14_7, \
call_0_13_14_8, \
call_0_13_14_9, \
call_0_13_14_10, \
call_0_13_14_11, \
call_0_13_14_12, \
call_0_13_14_13, \
call_0_13_14_14, \
call_0_13_14_15, \
call_0_13_15_0, \
call_0_13_15_1, \
call_0_13_15_2, \
call_0_13_15_3, \
call_0_13_15_4, \
call_0_13_15_5, \
call_0_13_15_6, \
call_0_13_15_7, \
call_0_13_15_8, \
call_0_13_15_9, \
call_0_13_15_10, \
call_0_13_15_11, \
call_0_13_15_12, \
call_0_13_15_13, \
call_0_13_15_14, \
call_0_13_15_15, \
call_0_14_0_0, \
call_0_14_0_1, \
call_0_14_0_2, \
call_0_14_0_3, \
call_0_14_0_4, \
call_0_14_0_5, \
call_0_14_0_6, \
call_0_14_0_7, \
call_0_14_0_8, \
call_0_14_0_9, \
call_0_14_0_10, \
call_0_14_0_11, \
call_0_14_0_12, \
call_0_14_0_13, \
call_0_14_0_14, \
call_0_14_0_15, \
call_0_14_1_0, \
call_0_14_1_1, \
call_0_14_1_2, \
call_0_14_1_3, \
call_0_14_1_4, \
call_0_14_1_5, \
call_0_14_1_6, \
call_0_14_1_7, \
call_0_14_1_8, \
call_0_14_1_9, \
call_0_14_1_10, \
call_0_14_1_11, \
call_0_14_1_12, \
call_0_14_1_13, \
call_0_14_1_14, \
call_0_14_1_15, \
call_0_14_2_0, \
call_0_14_2_1, \
call_0_14_2_2, \
call_0_14_2_3, \
call_0_14_2_4, \
call_0_14_2_5, \
call_0_14_2_6, \
call_0_14_2_7, \
call_0_14_2_8, \
call_0_14_2_9, \
call_0_14_2_10, \
call_0_14_2_11, \
call_0_14_2_12, \
call_0_14_2_13, \
call_0_14_2_14, \
call_0_14_2_15, \
call_0_14_3_0, \
call_0_14_3_1, \
call_0_14_3_2, \
call_0_14_3_3, \
call_0_14_3_4, \
call_0_14_3_5, \
call_0_14_3_6, \
call_0_14_3_7, \
call_0_14_3_8, \
call_0_14_3_9, \
call_0_14_3_10, \
call_0_14_3_11, \
call_0_14_3_12, \
call_0_14_3_13, \
call_0_14_3_14, \
call_0_14_3_15, \
call_0_14_4_0, \
call_0_14_4_1, \
call_0_14_4_2, \
call_0_14_4_3, \
call_0_14_4_4, \
call_0_14_4_5, \
call_0_14_4_6, \
call_0_14_4_7, \
call_0_14_4_8, \
call_0_14_4_9, \
call_0_14_4_10, \
call_0_14_4_11, \
call_0_14_4_12, \
call_0_14_4_13, \
call_0_14_4_14, \
call_0_14_4_15, \
call_0_14_5_0, \
call_0_14_5_1, \
call_0_14_5_2, \
call_0_14_5_3, \
call_0_14_5_4, \
call_0_14_5_5, \
call_0_14_5_6, \
call_0_14_5_7, \
call_0_14_5_8, \
call_0_14_5_9, \
call_0_14_5_10, \
call_0_14_5_11, \
call_0_14_5_12, \
call_0_14_5_13, \
call_0_14_5_14, \
call_0_14_5_15, \
call_0_14_6_0, \
call_0_14_6_1, \
call_0_14_6_2, \
call_0_14_6_3, \
call_0_14_6_4, \
call_0_14_6_5, \
call_0_14_6_6, \
call_0_14_6_7, \
call_0_14_6_8, \
call_0_14_6_9, \
call_0_14_6_10, \
call_0_14_6_11, \
call_0_14_6_12, \
call_0_14_6_13, \
call_0_14_6_14, \
call_0_14_6_15, \
call_0_14_7_0, \
call_0_14_7_1, \
call_0_14_7_2, \
call_0_14_7_3, \
call_0_14_7_4, \
call_0_14_7_5, \
call_0_14_7_6, \
call_0_14_7_7, \
call_0_14_7_8, \
call_0_14_7_9, \
call_0_14_7_10, \
call_0_14_7_11, \
call_0_14_7_12, \
call_0_14_7_13, \
call_0_14_7_14, \
call_0_14_7_15, \
call_0_14_8_0, \
call_0_14_8_1, \
call_0_14_8_2, \
call_0_14_8_3, \
call_0_14_8_4, \
call_0_14_8_5, \
call_0_14_8_6, \
call_0_14_8_7, \
call_0_14_8_8, \
call_0_14_8_9, \
call_0_14_8_10, \
call_0_14_8_11, \
call_0_14_8_12, \
call_0_14_8_13, \
call_0_14_8_14, \
call_0_14_8_15, \
call_0_14_9_0, \
call_0_14_9_1, \
call_0_14_9_2, \
call_0_14_9_3, \
call_0_14_9_4, \
call_0_14_9_5, \
call_0_14_9_6, \
call_0_14_9_7, \
call_0_14_9_8, \
call_0_14_9_9, \
call_0_14_9_10, \
call_0_14_9_11, \
call_0_14_9_12, \
call_0_14_9_13, \
call_0_14_9_14, \
call_0_14_9_15, \
call_0_14_10_0, \
call_0_14_10_1, \
call_0_14_10_2, \
call_0_14_10_3, \
call_0_14_10_4, \
call_0_14_10_5, \
call_0_14_10_6, \
call_0_14_10_7, \
call_0_14_10_8, \
call_0_14_10_9, \
call_0_14_10_10, \
call_0_14_10_11, \
call_0_14_10_12, \
call_0_14_10_13, \
call_0_14_10_14, \
call_0_14_10_15, \
call_0_14_11_0, \
call_0_14_11_1, \
call_0_14_11_2, \
call_0_14_11_3, \
call_0_14_11_4, \
call_0_14_11_5, \
call_0_14_11_6, \
call_0_14_11_7, \
call_0_14_11_8, \
call_0_14_11_9, \
call_0_14_11_10, \
call_0_14_11_11, \
call_0_14_11_12, \
call_0_14_11_13, \
call_0_14_11_14, \
call_0_14_11_15, \
call_0_14_12_0, \
call_0_14_12_1, \
call_0_14_12_2, \
call_0_14_12_3, \
call_0_14_12_4, \
call_0_14_12_5, \
call_0_14_12_6, \
call_0_14_12_7, \
call_0_14_12_8, \
call_0_14_12_9, \
call_0_14_12_10, \
call_0_14_12_11, \
call_0_14_12_12, \
call_0_14_12_13, \
call_0_14_12_14, \
call_0_14_12_15, \
call_0_14_13_0, \
call_0_14_13_1, \
call_0_14_13_2, \
call_0_14_13_3, \
call_0_14_13_4, \
call_0_14_13_5, \
call_0_14_13_6, \
call_0_14_13_7, \
call_0_14_13_8, \
call_0_14_13_9, \
call_0_14_13_10, \
call_0_14_13_11, \
call_0_14_13_12, \
call_0_14_13_13, \
call_0_14_13_14, \
call_0_14_13_15, \
call_0_14_14_0, \
call_0_14_14_1, \
call_0_14_14_2, \
call_0_14_14_3, \
call_0_14_14_4, \
call_0_14_14_5, \
call_0_14_14_6, \
call_0_14_14_7, \
call_0_14_14_8, \
call_0_14_14_9, \
call_0_14_14_10, \
call_0_14_14_11, \
call_0_14_14_12, \
call_0_14_14_13, \
call_0_14_14_14, \
call_0_14_14_15, \
call_0_14_15_0, \
call_0_14_15_1, \
call_0_14_15_2, \
call_0_14_15_3, \
call_0_14_15_4, \
call_0_14_15_5, \
call_0_14_15_6, \
call_0_14_15_7, \
call_0_14_15_8, \
call_0_14_15_9, \
call_0_14_15_10, \
call_0_14_15_11, \
call_0_14_15_12, \
call_0_14_15_13, \
call_0_14_15_14, \
call_0_14_15_15, \
call_0_15_0_0, \
call_0_15_0_1, \
call_0_15_0_2, \
call_0_15_0_3, \
call_0_15_0_4, \
call_0_15_0_5, \
call_0_15_0_6, \
call_0_15_0_7, \
call_0_15_0_8, \
call_0_15_0_9, \
call_0_15_0_10, \
call_0_15_0_11, \
call_0_15_0_12, \
call_0_15_0_13, \
call_0_15_0_14, \
call_0_15_0_15, \
call_0_15_1_0, \
call_0_15_1_1, \
call_0_15_1_2, \
call_0_15_1_3, \
call_0_15_1_4, \
call_0_15_1_5, \
call_0_15_1_6, \
call_0_15_1_7, \
call_0_15_1_8, \
call_0_15_1_9, \
call_0_15_1_10, \
call_0_15_1_11, \
call_0_15_1_12, \
call_0_15_1_13, \
call_0_15_1_14, \
call_0_15_1_15, \
call_0_15_2_0, \
call_0_15_2_1, \
call_0_15_2_2, \
call_0_15_2_3, \
call_0_15_2_4, \
call_0_15_2_5, \
call_0_15_2_6, \
call_0_15_2_7, \
call_0_15_2_8, \
call_0_15_2_9, \
call_0_15_2_10, \
call_0_15_2_11, \
call_0_15_2_12, \
call_0_15_2_13, \
call_0_15_2_14, \
call_0_15_2_15, \
call_0_15_3_0, \
call_0_15_3_1, \
call_0_15_3_2, \
call_0_15_3_3, \
call_0_15_3_4, \
call_0_15_3_5, \
call_0_15_3_6, \
call_0_15_3_7, \
call_0_15_3_8, \
call_0_15_3_9, \
call_0_15_3_10, \
call_0_15_3_11, \
call_0_15_3_12, \
call_0_15_3_13, \
call_0_15_3_14, \
call_0_15_3_15, \
call_0_15_4_0, \
call_0_15_4_1, \
call_0_15_4_2, \
call_0_15_4_3, \
call_0_15_4_4, \
call_0_15_4_5, \
call_0_15_4_6, \
call_0_15_4_7, \
call_0_15_4_8, \
call_0_15_4_9, \
call_0_15_4_10, \
call_0_15_4_11, \
call_0_15_4_12, \
call_0_15_4_13, \
call_0_15_4_14, \
call_0_15_4_15, \
call_0_15_5_0, \
call_0_15_5_1, \
call_0_15_5_2, \
call_0_15_5_3, \
call_0_15_5_4, \
call_0_15_5_5, \
call_0_15_5_6, \
call_0_15_5_7, \
call_0_15_5_8, \
call_0_15_5_9, \
call_0_15_5_10, \
call_0_15_5_11, \
call_0_15_5_12, \
call_0_15_5_13, \
call_0_15_5_14, \
call_0_15_5_15, \
call_0_15_6_0, \
call_0_15_6_1, \
call_0_15_6_2, \
call_0_15_6_3, \
call_0_15_6_4, \
call_0_15_6_5, \
call_0_15_6_6, \
call_0_15_6_7, \
call_0_15_6_8, \
call_0_15_6_9, \
call_0_15_6_10, \
call_0_15_6_11, \
call_0_15_6_12, \
call_0_15_6_13, \
call_0_15_6_14, \
call_0_15_6_15, \
call_0_15_7_0, \
call_0_15_7_1, \
call_0_15_7_2, \
call_0_15_7_3, \
call_0_15_7_4, \
call_0_15_7_5, \
call_0_15_7_6, \
call_0_15_7_7, \
call_0_15_7_8, \
call_0_15_7_9, \
call_0_15_7_10, \
call_0_15_7_11, \
call_0_15_7_12, \
call_0_15_7_13, \
call_0_15_7_14, \
call_0_15_7_15, \
call_0_15_8_0, \
call_0_15_8_1, \
call_0_15_8_2, \
call_0_15_8_3, \
call_0_15_8_4, \
call_0_15_8_5, \
call_0_15_8_6, \
call_0_15_8_7, \
call_0_15_8_8, \
call_0_15_8_9, \
call_0_15_8_10, \
call_0_15_8_11, \
call_0_15_8_12, \
call_0_15_8_13, \
call_0_15_8_14, \
call_0_15_8_15, \
call_0_15_9_0, \
call_0_15_9_1, \
call_0_15_9_2, \
call_0_15_9_3, \
call_0_15_9_4, \
call_0_15_9_5, \
call_0_15_9_6, \
call_0_15_9_7, \
call_0_15_9_8, \
call_0_15_9_9, \
call_0_15_9_10, \
call_0_15_9_11, \
call_0_15_9_12, \
call_0_15_9_13, \
call_0_15_9_14, \
call_0_15_9_15, \
call_0_15_10_0, \
call_0_15_10_1, \
call_0_15_10_2, \
call_0_15_10_3, \
call_0_15_10_4, \
call_0_15_10_5, \
call_0_15_10_6, \
call_0_15_10_7, \
call_0_15_10_8, \
call_0_15_10_9, \
call_0_15_10_10, \
call_0_15_10_11, \
call_0_15_10_12, \
call_0_15_10_13, \
call_0_15_10_14, \
call_0_15_10_15, \
call_0_15_11_0, \
call_0_15_11_1, \
call_0_15_11_2, \
call_0_15_11_3, \
call_0_15_11_4, \
call_0_15_11_5, \
call_0_15_11_6, \
call_0_15_11_7, \
call_0_15_11_8, \
call_0_15_11_9, \
call_0_15_11_10, \
call_0_15_11_11, \
call_0_15_11_12, \
call_0_15_11_13, \
call_0_15_11_14, \
call_0_15_11_15, \
call_0_15_12_0, \
call_0_15_12_1, \
call_0_15_12_2, \
call_0_15_12_3, \
call_0_15_12_4, \
call_0_15_12_5, \
call_0_15_12_6, \
call_0_15_12_7, \
call_0_15_12_8, \
call_0_15_12_9, \
call_0_15_12_10, \
call_0_15_12_11, \
call_0_15_12_12, \
call_0_15_12_13, \
call_0_15_12_14, \
call_0_15_12_15, \
call_0_15_13_0, \
call_0_15_13_1, \
call_0_15_13_2, \
call_0_15_13_3, \
call_0_15_13_4, \
call_0_15_13_5, \
call_0_15_13_6, \
call_0_15_13_7, \
call_0_15_13_8, \
call_0_15_13_9, \
call_0_15_13_10, \
call_0_15_13_11, \
call_0_15_13_12, \
call_0_15_13_13, \
call_0_15_13_14, \
call_0_15_13_15, \
call_0_15_14_0, \
call_0_15_14_1, \
call_0_15_14_2, \
call_0_15_14_3, \
call_0_15_14_4, \
call_0_15_14_5, \
call_0_15_14_6, \
call_0_15_14_7, \
call_0_15_14_8, \
call_0_15_14_9, \
call_0_15_14_10, \
call_0_15_14_11, \
call_0_15_14_12, \
call_0_15_14_13, \
call_0_15_14_14, \
call_0_15_14_15, \
call_0_15_15_0, \
call_0_15_15_1, \
call_0_15_15_2, \
call_0_15_15_3, \
call_0_15_15_4, \
call_0_15_15_5, \
call_0_15_15_6, \
call_0_15_15_7, \
call_0_15_15_8, \
call_0_15_15_9, \
call_0_15_15_10, \
call_0_15_15_11, \
call_0_15_15_12, \
call_0_15_15_13, \
call_0_15_15_14, \
call_0_15_15_15, \
call_1_0_0_0, \
call_1_0_0_1, \
call_1_0_0_2, \
call_1_0_0_3, \
call_1_0_0_4, \
call_1_0_0_5, \
call_1_0_0_6, \
call_1_0_0_7, \
call_1_0_0_8, \
call_1_0_0_9, \
call_1_0_0_10, \
call_1_0_0_11, \
call_1_0_0_12, \
call_1_0_0_13, \
call_1_0_0_14, \
call_1_0_0_15, \
call_1_0_1_0, \
call_1_0_1_1, \
call_1_0_1_2, \
call_1_0_1_3, \
call_1_0_1_4, \
call_1_0_1_5, \
call_1_0_1_6, \
call_1_0_1_7, \
call_1_0_1_8, \
call_1_0_1_9, \
call_1_0_1_10, \
call_1_0_1_11, \
call_1_0_1_12, \
call_1_0_1_13, \
call_1_0_1_14, \
call_1_0_1_15, \
call_1_0_2_0, \
call_1_0_2_1, \
call_1_0_2_2, \
call_1_0_2_3, \
call_1_0_2_4, \
call_1_0_2_5, \
call_1_0_2_6, \
call_1_0_2_7, \
call_1_0_2_8, \
call_1_0_2_9, \
call_1_0_2_10, \
call_1_0_2_11, \
call_1_0_2_12, \
call_1_0_2_13, \
call_1_0_2_14, \
call_1_0_2_15, \
call_1_0_3_0, \
call_1_0_3_1, \
call_1_0_3_2, \
call_1_0_3_3, \
call_1_0_3_4, \
call_1_0_3_5, \
call_1_0_3_6, \
call_1_0_3_7, \
call_1_0_3_8, \
call_1_0_3_9, \
call_1_0_3_10, \
call_1_0_3_11, \
call_1_0_3_12, \
call_1_0_3_13, \
call_1_0_3_14, \
call_1_0_3_15, \
call_1_0_4_0, \
call_1_0_4_1, \
call_1_0_4_2, \
call_1_0_4_3, \
call_1_0_4_4, \
call_1_0_4_5, \
call_1_0_4_6, \
call_1_0_4_7, \
call_1_0_4_8, \
call_1_0_4_9, \
call_1_0_4_10, \
call_1_0_4_11, \
call_1_0_4_12, \
call_1_0_4_13, \
call_1_0_4_14, \
call_1_0_4_15, \
call_1_0_5_0, \
call_1_0_5_1, \
call_1_0_5_2, \
call_1_0_5_3, \
call_1_0_5_4, \
call_1_0_5_5, \
call_1_0_5_6, \
call_1_0_5_7, \
call_1_0_5_8, \
call_1_0_5_9, \
call_1_0_5_10, \
call_1_0_5_11, \
call_1_0_5_12, \
call_1_0_5_13, \
call_1_0_5_14, \
call_1_0_5_15, \
call_1_0_6_0, \
call_1_0_6_1, \
call_1_0_6_2, \
call_1_0_6_3, \
call_1_0_6_4, \
call_1_0_6_5, \
call_1_0_6_6, \
call_1_0_6_7, \
call_1_0_6_8, \
call_1_0_6_9, \
call_1_0_6_10, \
call_1_0_6_11, \
call_1_0_6_12, \
call_1_0_6_13, \
call_1_0_6_14, \
call_1_0_6_15, \
call_1_0_7_0, \
call_1_0_7_1, \
call_1_0_7_2, \
call_1_0_7_3, \
call_1_0_7_4, \
call_1_0_7_5, \
call_1_0_7_6, \
call_1_0_7_7, \
call_1_0_7_8, \
call_1_0_7_9, \
call_1_0_7_10, \
call_1_0_7_11, \
call_1_0_7_12, \
call_1_0_7_13, \
call_1_0_7_14, \
call_1_0_7_15, \
call_1_0_8_0, \
call_1_0_8_1, \
call_1_0_8_2, \
call_1_0_8_3, \
call_1_0_8_4, \
call_1_0_8_5, \
call_1_0_8_6, \
call_1_0_8_7, \
call_1_0_8_8, \
call_1_0_8_9, \
call_1_0_8_10, \
call_1_0_8_11, \
call_1_0_8_12, \
call_1_0_8_13, \
call_1_0_8_14, \
call_1_0_8_15, \
call_1_0_9_0, \
call_1_0_9_1, \
call_1_0_9_2, \
call_1_0_9_3, \
call_1_0_9_4, \
call_1_0_9_5, \
call_1_0_9_6, \
call_1_0_9_7, \
call_1_0_9_8, \
call_1_0_9_9, \
call_1_0_9_10, \
call_1_0_9_11, \
call_1_0_9_12, \
call_1_0_9_13, \
call_1_0_9_14, \
call_1_0_9_15, \
call_1_0_10_0, \
call_1_0_10_1, \
call_1_0_10_2, \
call_1_0_10_3, \
call_1_0_10_4, \
call_1_0_10_5, \
call_1_0_10_6, \
call_1_0_10_7, \
call_1_0_10_8, \
call_1_0_10_9, \
call_1_0_10_10, \
call_1_0_10_11, \
call_1_0_10_12, \
call_1_0_10_13, \
call_1_0_10_14, \
call_1_0_10_15, \
call_1_0_11_0, \
call_1_0_11_1, \
call_1_0_11_2, \
call_1_0_11_3, \
call_1_0_11_4, \
call_1_0_11_5, \
call_1_0_11_6, \
call_1_0_11_7, \
call_1_0_11_8, \
call_1_0_11_9, \
call_1_0_11_10, \
call_1_0_11_11, \
call_1_0_11_12, \
call_1_0_11_13, \
call_1_0_11_14, \
call_1_0_11_15, \
call_1_0_12_0, \
call_1_0_12_1, \
call_1_0_12_2, \
call_1_0_12_3, \
call_1_0_12_4, \
call_1_0_12_5, \
call_1_0_12_6, \
call_1_0_12_7, \
call_1_0_12_8, \
call_1_0_12_9, \
call_1_0_12_10, \
call_1_0_12_11, \
call_1_0_12_12, \
call_1_0_12_13, \
call_1_0_12_14, \
call_1_0_12_15, \
call_1_0_13_0, \
call_1_0_13_1, \
call_1_0_13_2, \
call_1_0_13_3, \
call_1_0_13_4, \
call_1_0_13_5, \
call_1_0_13_6, \
call_1_0_13_7, \
call_1_0_13_8, \
call_1_0_13_9, \
call_1_0_13_10, \
call_1_0_13_11, \
call_1_0_13_12, \
call_1_0_13_13, \
call_1_0_13_14, \
call_1_0_13_15, \
call_1_0_14_0, \
call_1_0_14_1, \
call_1_0_14_2, \
call_1_0_14_3, \
call_1_0_14_4, \
call_1_0_14_5, \
call_1_0_14_6, \
call_1_0_14_7, \
call_1_0_14_8, \
call_1_0_14_9, \
call_1_0_14_10, \
call_1_0_14_11, \
call_1_0_14_12, \
call_1_0_14_13, \
call_1_0_14_14, \
call_1_0_14_15, \
call_1_0_15_0, \
call_1_0_15_1, \
call_1_0_15_2, \
call_1_0_15_3, \
call_1_0_15_4, \
call_1_0_15_5, \
call_1_0_15_6, \
call_1_0_15_7, \
call_1_0_15_8, \
call_1_0_15_9, \
call_1_0_15_10, \
call_1_0_15_11, \
call_1_0_15_12, \
call_1_0_15_13, \
call_1_0_15_14, \
call_1_0_15_15, \
call_1_1_0_0, \
call_1_1_0_1, \
call_1_1_0_2, \
call_1_1_0_3, \
call_1_1_0_4, \
call_1_1_0_5, \
call_1_1_0_6, \
call_1_1_0_7, \
call_1_1_0_8, \
call_1_1_0_9, \
call_1_1_0_10, \
call_1_1_0_11, \
call_1_1_0_12, \
call_1_1_0_13, \
call_1_1_0_14, \
call_1_1_0_15, \
call_1_1_1_0, \
call_1_1_1_1, \
call_1_1_1_2, \
call_1_1_1_3, \
call_1_1_1_4, \
call_1_1_1_5, \
call_1_1_1_6, \
call_1_1_1_7, \
call_1_1_1_8, \
call_1_1_1_9, \
call_1_1_1_10, \
call_1_1_1_11, \
call_1_1_1_12, \
call_1_1_1_13, \
call_1_1_1_14, \
call_1_1_1_15, \
call_1_1_2_0, \
call_1_1_2_1, \
call_1_1_2_2, \
call_1_1_2_3, \
call_1_1_2_4, \
call_1_1_2_5, \
call_1_1_2_6, \
call_1_1_2_7, \
call_1_1_2_8, \
call_1_1_2_9, \
call_1_1_2_10, \
call_1_1_2_11, \
call_1_1_2_12, \
call_1_1_2_13, \
call_1_1_2_14, \
call_1_1_2_15, \
call_1_1_3_0, \
call_1_1_3_1, \
call_1_1_3_2, \
call_1_1_3_3, \
call_1_1_3_4, \
call_1_1_3_5, \
call_1_1_3_6, \
call_1_1_3_7, \
call_1_1_3_8, \
call_1_1_3_9, \
call_1_1_3_10, \
call_1_1_3_11, \
call_1_1_3_12, \
call_1_1_3_13, \
call_1_1_3_14, \
call_1_1_3_15, \
call_1_1_4_0, \
call_1_1_4_1, \
call_1_1_4_2, \
call_1_1_4_3, \
call_1_1_4_4, \
call_1_1_4_5, \
call_1_1_4_6, \
call_1_1_4_7, \
call_1_1_4_8, \
call_1_1_4_9, \
call_1_1_4_10, \
call_1_1_4_11, \
call_1_1_4_12, \
call_1_1_4_13, \
call_1_1_4_14, \
call_1_1_4_15, \
call_1_1_5_0, \
call_1_1_5_1, \
call_1_1_5_2, \
call_1_1_5_3, \
call_1_1_5_4, \
call_1_1_5_5, \
call_1_1_5_6, \
call_1_1_5_7, \
call_1_1_5_8, \
call_1_1_5_9, \
call_1_1_5_10, \
call_1_1_5_11, \
call_1_1_5_12, \
call_1_1_5_13, \
call_1_1_5_14, \
call_1_1_5_15, \
call_1_1_6_0, \
call_1_1_6_1, \
call_1_1_6_2, \
call_1_1_6_3, \
call_1_1_6_4, \
call_1_1_6_5, \
call_1_1_6_6, \
call_1_1_6_7, \
call_1_1_6_8, \
call_1_1_6_9, \
call_1_1_6_10, \
call_1_1_6_11, \
call_1_1_6_12, \
call_1_1_6_13, \
call_1_1_6_14, \
call_1_1_6_15, \
call_1_1_7_0, \
call_1_1_7_1, \
call_1_1_7_2, \
call_1_1_7_3, \
call_1_1_7_4, \
call_1_1_7_5, \
call_1_1_7_6, \
call_1_1_7_7, \
call_1_1_7_8, \
call_1_1_7_9, \
call_1_1_7_10, \
call_1_1_7_11, \
call_1_1_7_12, \
call_1_1_7_13, \
call_1_1_7_14, \
call_1_1_7_15, \
call_1_1_8_0, \
call_1_1_8_1, \
call_1_1_8_2, \
call_1_1_8_3, \
call_1_1_8_4, \
call_1_1_8_5, \
call_1_1_8_6, \
call_1_1_8_7, \
call_1_1_8_8, \
call_1_1_8_9, \
call_1_1_8_10, \
call_1_1_8_11, \
call_1_1_8_12, \
call_1_1_8_13, \
call_1_1_8_14, \
call_1_1_8_15, \
call_1_1_9_0, \
call_1_1_9_1, \
call_1_1_9_2, \
call_1_1_9_3, \
call_1_1_9_4, \
call_1_1_9_5, \
call_1_1_9_6, \
call_1_1_9_7, \
call_1_1_9_8, \
call_1_1_9_9, \
call_1_1_9_10, \
call_1_1_9_11, \
call_1_1_9_12, \
call_1_1_9_13, \
call_1_1_9_14, \
call_1_1_9_15, \
call_1_1_10_0, \
call_1_1_10_1, \
call_1_1_10_2, \
call_1_1_10_3, \
call_1_1_10_4, \
call_1_1_10_5, \
call_1_1_10_6, \
call_1_1_10_7, \
call_1_1_10_8, \
call_1_1_10_9, \
call_1_1_10_10, \
call_1_1_10_11, \
call_1_1_10_12, \
call_1_1_10_13, \
call_1_1_10_14, \
call_1_1_10_15, \
call_1_1_11_0, \
call_1_1_11_1, \
call_1_1_11_2, \
call_1_1_11_3, \
call_1_1_11_4, \
call_1_1_11_5, \
call_1_1_11_6, \
call_1_1_11_7, \
call_1_1_11_8, \
call_1_1_11_9, \
call_1_1_11_10, \
call_1_1_11_11, \
call_1_1_11_12, \
call_1_1_11_13, \
call_1_1_11_14, \
call_1_1_11_15, \
call_1_1_12_0, \
call_1_1_12_1, \
call_1_1_12_2, \
call_1_1_12_3, \
call_1_1_12_4, \
call_1_1_12_5, \
call_1_1_12_6, \
call_1_1_12_7, \
call_1_1_12_8, \
call_1_1_12_9, \
call_1_1_12_10, \
call_1_1_12_11, \
call_1_1_12_12, \
call_1_1_12_13, \
call_1_1_12_14, \
call_1_1_12_15, \
call_1_1_13_0, \
call_1_1_13_1, \
call_1_1_13_2, \
call_1_1_13_3, \
call_1_1_13_4, \
call_1_1_13_5, \
call_1_1_13_6, \
call_1_1_13_7, \
call_1_1_13_8, \
call_1_1_13_9, \
call_1_1_13_10, \
call_1_1_13_11, \
call_1_1_13_12, \
call_1_1_13_13, \
call_1_1_13_14, \
call_1_1_13_15, \
call_1_1_14_0, \
call_1_1_14_1, \
call_1_1_14_2, \
call_1_1_14_3, \
call_1_1_14_4, \
call_1_1_14_5, \
call_1_1_14_6, \
call_1_1_14_7, \
call_1_1_14_8, \
call_1_1_14_9, \
call_1_1_14_10, \
call_1_1_14_11, \
call_1_1_14_12, \
call_1_1_14_13, \
call_1_1_14_14, \
call_1_1_14_15, \
call_1_1_15_0, \
call_1_1_15_1, \
call_1_1_15_2, \
call_1_1_15_3, \
call_1_1_15_4, \
call_1_1_15_5, \
call_1_1_15_6, \
call_1_1_15_7, \
call_1_1_15_8, \
call_1_1_15_9, \
call_1_1_15_10, \
call_1_1_15_11, \
call_1_1_15_12, \
call_1_1_15_13, \
call_1_1_15_14, \
call_1_1_15_15, \
call_1_2_0_0, \
call_1_2_0_1, \
call_1_2_0_2, \
call_1_2_0_3, \
call_1_2_0_4, \
call_1_2_0_5, \
call_1_2_0_6, \
call_1_2_0_7, \
call_1_2_0_8, \
call_1_2_0_9, \
call_1_2_0_10, \
call_1_2_0_11, \
call_1_2_0_12, \
call_1_2_0_13, \
call_1_2_0_14, \
call_1_2_0_15, \
call_1_2_1_0, \
call_1_2_1_1, \
call_1_2_1_2, \
call_1_2_1_3, \
call_1_2_1_4, \
call_1_2_1_5, \
call_1_2_1_6, \
call_1_2_1_7, \
call_1_2_1_8, \
call_1_2_1_9, \
call_1_2_1_10, \
call_1_2_1_11, \
call_1_2_1_12, \
call_1_2_1_13, \
call_1_2_1_14, \
call_1_2_1_15, \
call_1_2_2_0, \
call_1_2_2_1, \
call_1_2_2_2, \
call_1_2_2_3, \
call_1_2_2_4, \
call_1_2_2_5, \
call_1_2_2_6, \
call_1_2_2_7, \
call_1_2_2_8, \
call_1_2_2_9, \
call_1_2_2_10, \
call_1_2_2_11, \
call_1_2_2_12, \
call_1_2_2_13, \
call_1_2_2_14, \
call_1_2_2_15, \
call_1_2_3_0, \
call_1_2_3_1, \
call_1_2_3_2, \
call_1_2_3_3, \
call_1_2_3_4, \
call_1_2_3_5, \
call_1_2_3_6, \
call_1_2_3_7, \
call_1_2_3_8, \
call_1_2_3_9, \
call_1_2_3_10, \
call_1_2_3_11, \
call_1_2_3_12, \
call_1_2_3_13, \
call_1_2_3_14, \
call_1_2_3_15, \
call_1_2_4_0, \
call_1_2_4_1, \
call_1_2_4_2, \
call_1_2_4_3, \
call_1_2_4_4, \
call_1_2_4_5, \
call_1_2_4_6, \
call_1_2_4_7, \
call_1_2_4_8, \
call_1_2_4_9, \
call_1_2_4_10, \
call_1_2_4_11, \
call_1_2_4_12, \
call_1_2_4_13, \
call_1_2_4_14, \
call_1_2_4_15, \
call_1_2_5_0, \
call_1_2_5_1, \
call_1_2_5_2, \
call_1_2_5_3, \
call_1_2_5_4, \
call_1_2_5_5, \
call_1_2_5_6, \
call_1_2_5_7, \
call_1_2_5_8, \
call_1_2_5_9, \
call_1_2_5_10, \
call_1_2_5_11, \
call_1_2_5_12, \
call_1_2_5_13, \
call_1_2_5_14, \
call_1_2_5_15, \
call_1_2_6_0, \
call_1_2_6_1, \
call_1_2_6_2, \
call_1_2_6_3, \
call_1_2_6_4, \
call_1_2_6_5, \
call_1_2_6_6, \
call_1_2_6_7, \
call_1_2_6_8, \
call_1_2_6_9, \
call_1_2_6_10, \
call_1_2_6_11, \
call_1_2_6_12, \
call_1_2_6_13, \
call_1_2_6_14, \
call_1_2_6_15, \
call_1_2_7_0, \
call_1_2_7_1, \
call_1_2_7_2, \
call_1_2_7_3, \
call_1_2_7_4, \
call_1_2_7_5, \
call_1_2_7_6, \
call_1_2_7_7, \
call_1_2_7_8, \
call_1_2_7_9, \
call_1_2_7_10, \
call_1_2_7_11, \
call_1_2_7_12, \
call_1_2_7_13, \
call_1_2_7_14, \
call_1_2_7_15, \
call_1_2_8_0, \
call_1_2_8_1, \
call_1_2_8_2, \
call_1_2_8_3, \
call_1_2_8_4, \
call_1_2_8_5, \
call_1_2_8_6, \
call_1_2_8_7, \
call_1_2_8_8, \
call_1_2_8_9, \
call_1_2_8_10, \
call_1_2_8_11, \
call_1_2_8_12, \
call_1_2_8_13, \
call_1_2_8_14, \
call_1_2_8_15, \
call_1_2_9_0, \
call_1_2_9_1, \
call_1_2_9_2, \
call_1_2_9_3, \
call_1_2_9_4, \
call_1_2_9_5, \
call_1_2_9_6, \
call_1_2_9_7, \
call_1_2_9_8, \
call_1_2_9_9, \
call_1_2_9_10, \
call_1_2_9_11, \
call_1_2_9_12, \
call_1_2_9_13, \
call_1_2_9_14, \
call_1_2_9_15, \
call_1_2_10_0, \
call_1_2_10_1, \
call_1_2_10_2, \
call_1_2_10_3, \
call_1_2_10_4, \
call_1_2_10_5, \
call_1_2_10_6, \
call_1_2_10_7, \
call_1_2_10_8, \
call_1_2_10_9, \
call_1_2_10_10, \
call_1_2_10_11, \
call_1_2_10_12, \
call_1_2_10_13, \
call_1_2_10_14, \
call_1_2_10_15, \
call_1_2_11_0, \
call_1_2_11_1, \
call_1_2_11_2, \
call_1_2_11_3, \
call_1_2_11_4, \
call_1_2_11_5, \
call_1_2_11_6, \
call_1_2_11_7, \
call_1_2_11_8, \
call_1_2_11_9, \
call_1_2_11_10, \
call_1_2_11_11, \
call_1_2_11_12, \
call_1_2_11_13, \
call_1_2_11_14, \
call_1_2_11_15, \
call_1_2_12_0, \
call_1_2_12_1, \
call_1_2_12_2, \
call_1_2_12_3, \
call_1_2_12_4, \
call_1_2_12_5, \
call_1_2_12_6, \
call_1_2_12_7, \
call_1_2_12_8, \
call_1_2_12_9, \
call_1_2_12_10, \
call_1_2_12_11, \
call_1_2_12_12, \
call_1_2_12_13, \
call_1_2_12_14, \
call_1_2_12_15, \
call_1_2_13_0, \
call_1_2_13_1, \
call_1_2_13_2, \
call_1_2_13_3, \
call_1_2_13_4, \
call_1_2_13_5, \
call_1_2_13_6, \
call_1_2_13_7, \
call_1_2_13_8, \
call_1_2_13_9, \
call_1_2_13_10, \
call_1_2_13_11, \
call_1_2_13_12, \
call_1_2_13_13, \
call_1_2_13_14, \
call_1_2_13_15, \
call_1_2_14_0, \
call_1_2_14_1, \
call_1_2_14_2, \
call_1_2_14_3, \
call_1_2_14_4, \
call_1_2_14_5, \
call_1_2_14_6, \
call_1_2_14_7, \
call_1_2_14_8, \
call_1_2_14_9, \
call_1_2_14_10, \
call_1_2_14_11, \
call_1_2_14_12, \
call_1_2_14_13, \
call_1_2_14_14, \
call_1_2_14_15, \
call_1_2_15_0, \
call_1_2_15_1, \
call_1_2_15_2, \
call_1_2_15_3, \
call_1_2_15_4, \
call_1_2_15_5, \
call_1_2_15_6, \
call_1_2_15_7, \
call_1_2_15_8, \
call_1_2_15_9, \
call_1_2_15_10, \
call_1_2_15_11, \
call_1_2_15_12, \
call_1_2_15_13, \
call_1_2_15_14, \
call_1_2_15_15, \
call_1_3_0_0, \
call_1_3_0_1, \
call_1_3_0_2, \
call_1_3_0_3, \
call_1_3_0_4, \
call_1_3_0_5, \
call_1_3_0_6, \
call_1_3_0_7, \
call_1_3_0_8, \
call_1_3_0_9, \
call_1_3_0_10, \
call_1_3_0_11, \
call_1_3_0_12, \
call_1_3_0_13, \
call_1_3_0_14, \
call_1_3_0_15, \
call_1_3_1_0, \
call_1_3_1_1, \
call_1_3_1_2, \
call_1_3_1_3, \
call_1_3_1_4, \
call_1_3_1_5, \
call_1_3_1_6, \
call_1_3_1_7, \
call_1_3_1_8, \
call_1_3_1_9, \
call_1_3_1_10, \
call_1_3_1_11, \
call_1_3_1_12, \
call_1_3_1_13, \
call_1_3_1_14, \
call_1_3_1_15, \
call_1_3_2_0, \
call_1_3_2_1, \
call_1_3_2_2, \
call_1_3_2_3, \
call_1_3_2_4, \
call_1_3_2_5, \
call_1_3_2_6, \
call_1_3_2_7, \
call_1_3_2_8, \
call_1_3_2_9, \
call_1_3_2_10, \
call_1_3_2_11, \
call_1_3_2_12, \
call_1_3_2_13, \
call_1_3_2_14, \
call_1_3_2_15, \
call_1_3_3_0, \
call_1_3_3_1, \
call_1_3_3_2, \
call_1_3_3_3, \
call_1_3_3_4, \
call_1_3_3_5, \
call_1_3_3_6, \
call_1_3_3_7, \
call_1_3_3_8, \
call_1_3_3_9, \
call_1_3_3_10, \
call_1_3_3_11, \
call_1_3_3_12, \
call_1_3_3_13, \
call_1_3_3_14, \
call_1_3_3_15, \
call_1_3_4_0, \
call_1_3_4_1, \
call_1_3_4_2, \
call_1_3_4_3, \
call_1_3_4_4, \
call_1_3_4_5, \
call_1_3_4_6, \
call_1_3_4_7, \
call_1_3_4_8, \
call_1_3_4_9, \
call_1_3_4_10, \
call_1_3_4_11, \
call_1_3_4_12, \
call_1_3_4_13, \
call_1_3_4_14, \
call_1_3_4_15, \
call_1_3_5_0, \
call_1_3_5_1, \
call_1_3_5_2, \
call_1_3_5_3, \
call_1_3_5_4, \
call_1_3_5_5, \
call_1_3_5_6, \
call_1_3_5_7, \
call_1_3_5_8, \
call_1_3_5_9, \
call_1_3_5_10, \
call_1_3_5_11, \
call_1_3_5_12, \
call_1_3_5_13, \
call_1_3_5_14, \
call_1_3_5_15, \
call_1_3_6_0, \
call_1_3_6_1, \
call_1_3_6_2, \
call_1_3_6_3, \
call_1_3_6_4, \
call_1_3_6_5, \
call_1_3_6_6, \
call_1_3_6_7, \
call_1_3_6_8, \
call_1_3_6_9, \
call_1_3_6_10, \
call_1_3_6_11, \
call_1_3_6_12, \
call_1_3_6_13, \
call_1_3_6_14, \
call_1_3_6_15, \
call_1_3_7_0, \
call_1_3_7_1, \
call_1_3_7_2, \
call_1_3_7_3, \
call_1_3_7_4, \
call_1_3_7_5, \
call_1_3_7_6, \
call_1_3_7_7, \
call_1_3_7_8, \
call_1_3_7_9, \
call_1_3_7_10, \
call_1_3_7_11, \
call_1_3_7_12, \
call_1_3_7_13, \
call_1_3_7_14, \
call_1_3_7_15, \
call_1_3_8_0, \
call_1_3_8_1, \
call_1_3_8_2, \
call_1_3_8_3, \
call_1_3_8_4, \
call_1_3_8_5, \
call_1_3_8_6, \
call_1_3_8_7, \
call_1_3_8_8, \
call_1_3_8_9, \
call_1_3_8_10, \
call_1_3_8_11, \
call_1_3_8_12, \
call_1_3_8_13, \
call_1_3_8_14, \
call_1_3_8_15, \
call_1_3_9_0, \
call_1_3_9_1, \
call_1_3_9_2, \
call_1_3_9_3, \
call_1_3_9_4, \
call_1_3_9_5, \
call_1_3_9_6, \
call_1_3_9_7, \
call_1_3_9_8, \
call_1_3_9_9, \
call_1_3_9_10, \
call_1_3_9_11, \
call_1_3_9_12, \
call_1_3_9_13, \
call_1_3_9_14, \
call_1_3_9_15, \
call_1_3_10_0, \
call_1_3_10_1, \
call_1_3_10_2, \
call_1_3_10_3, \
call_1_3_10_4, \
call_1_3_10_5, \
call_1_3_10_6, \
call_1_3_10_7, \
call_1_3_10_8, \
call_1_3_10_9, \
call_1_3_10_10, \
call_1_3_10_11, \
call_1_3_10_12, \
call_1_3_10_13, \
call_1_3_10_14, \
call_1_3_10_15, \
call_1_3_11_0, \
call_1_3_11_1, \
call_1_3_11_2, \
call_1_3_11_3, \
call_1_3_11_4, \
call_1_3_11_5, \
call_1_3_11_6, \
call_1_3_11_7, \
call_1_3_11_8, \
call_1_3_11_9, \
call_1_3_11_10, \
call_1_3_11_11, \
call_1_3_11_12, \
call_1_3_11_13, \
call_1_3_11_14, \
call_1_3_11_15, \
call_1_3_12_0, \
call_1_3_12_1, \
call_1_3_12_2, \
call_1_3_12_3, \
call_1_3_12_4, \
call_1_3_12_5, \
call_1_3_12_6, \
call_1_3_12_7, \
call_1_3_12_8, \
call_1_3_12_9, \
call_1_3_12_10, \
call_1_3_12_11, \
call_1_3_12_12, \
call_1_3_12_13, \
call_1_3_12_14, \
call_1_3_12_15, \
call_1_3_13_0, \
call_1_3_13_1, \
call_1_3_13_2, \
call_1_3_13_3, \
call_1_3_13_4, \
call_1_3_13_5, \
call_1_3_13_6, \
call_1_3_13_7, \
call_1_3_13_8, \
call_1_3_13_9, \
call_1_3_13_10, \
call_1_3_13_11, \
call_1_3_13_12, \
call_1_3_13_13, \
call_1_3_13_14, \
call_1_3_13_15, \
call_1_3_14_0, \
call_1_3_14_1, \
call_1_3_14_2, \
call_1_3_14_3, \
call_1_3_14_4, \
call_1_3_14_5, \
call_1_3_14_6, \
call_1_3_14_7, \
call_1_3_14_8, \
call_1_3_14_9, \
call_1_3_14_10, \
call_1_3_14_11, \
call_1_3_14_12, \
call_1_3_14_13, \
call_1_3_14_14, \
call_1_3_14_15, \
call_1_3_15_0, \
call_1_3_15_1, \
call_1_3_15_2, \
call_1_3_15_3, \
call_1_3_15_4, \
call_1_3_15_5, \
call_1_3_15_6, \
call_1_3_15_7, \
call_1_3_15_8, \
call_1_3_15_9, \
call_1_3_15_10, \
call_1_3_15_11, \
call_1_3_15_12, \
call_1_3_15_13, \
call_1_3_15_14, \
call_1_3_15_15, \
call_1_4_0_0, \
call_1_4_0_1, \
call_1_4_0_2, \
call_1_4_0_3, \
call_1_4_0_4, \
call_1_4_0_5, \
call_1_4_0_6, \
call_1_4_0_7, \
call_1_4_0_8, \
call_1_4_0_9, \
call_1_4_0_10, \
call_1_4_0_11, \
call_1_4_0_12, \
call_1_4_0_13, \
call_1_4_0_14, \
call_1_4_0_15, \
call_1_4_1_0, \
call_1_4_1_1, \
call_1_4_1_2, \
call_1_4_1_3, \
call_1_4_1_4, \
call_1_4_1_5, \
call_1_4_1_6, \
call_1_4_1_7, \
call_1_4_1_8, \
call_1_4_1_9, \
call_1_4_1_10, \
call_1_4_1_11, \
call_1_4_1_12, \
call_1_4_1_13, \
call_1_4_1_14, \
call_1_4_1_15, \
call_1_4_2_0, \
call_1_4_2_1, \
call_1_4_2_2, \
call_1_4_2_3, \
call_1_4_2_4, \
call_1_4_2_5, \
call_1_4_2_6, \
call_1_4_2_7, \
call_1_4_2_8, \
call_1_4_2_9, \
call_1_4_2_10, \
call_1_4_2_11, \
call_1_4_2_12, \
call_1_4_2_13, \
call_1_4_2_14, \
call_1_4_2_15, \
call_1_4_3_0, \
call_1_4_3_1, \
call_1_4_3_2, \
call_1_4_3_3, \
call_1_4_3_4, \
call_1_4_3_5, \
call_1_4_3_6, \
call_1_4_3_7, \
call_1_4_3_8, \
call_1_4_3_9, \
call_1_4_3_10, \
call_1_4_3_11, \
call_1_4_3_12, \
call_1_4_3_13, \
call_1_4_3_14, \
call_1_4_3_15, \
call_1_4_4_0, \
call_1_4_4_1, \
call_1_4_4_2, \
call_1_4_4_3, \
call_1_4_4_4, \
call_1_4_4_5, \
call_1_4_4_6, \
call_1_4_4_7, \
call_1_4_4_8, \
call_1_4_4_9, \
call_1_4_4_10, \
call_1_4_4_11, \
call_1_4_4_12, \
call_1_4_4_13, \
call_1_4_4_14, \
call_1_4_4_15, \
call_1_4_5_0, \
call_1_4_5_1, \
call_1_4_5_2, \
call_1_4_5_3, \
call_1_4_5_4, \
call_1_4_5_5, \
call_1_4_5_6, \
call_1_4_5_7, \
call_1_4_5_8, \
call_1_4_5_9, \
call_1_4_5_10, \
call_1_4_5_11, \
call_1_4_5_12, \
call_1_4_5_13, \
call_1_4_5_14, \
call_1_4_5_15, \
call_1_4_6_0, \
call_1_4_6_1, \
call_1_4_6_2, \
call_1_4_6_3, \
call_1_4_6_4, \
call_1_4_6_5, \
call_1_4_6_6, \
call_1_4_6_7, \
call_1_4_6_8, \
call_1_4_6_9, \
call_1_4_6_10, \
call_1_4_6_11, \
call_1_4_6_12, \
call_1_4_6_13, \
call_1_4_6_14, \
call_1_4_6_15, \
call_1_4_7_0, \
call_1_4_7_1, \
call_1_4_7_2, \
call_1_4_7_3, \
call_1_4_7_4, \
call_1_4_7_5, \
call_1_4_7_6, \
call_1_4_7_7, \
call_1_4_7_8, \
call_1_4_7_9, \
call_1_4_7_10, \
call_1_4_7_11, \
call_1_4_7_12, \
call_1_4_7_13, \
call_1_4_7_14, \
call_1_4_7_15, \
call_1_4_8_0, \
call_1_4_8_1, \
call_1_4_8_2, \
call_1_4_8_3, \
call_1_4_8_4, \
call_1_4_8_5, \
call_1_4_8_6, \
call_1_4_8_7, \
call_1_4_8_8, \
call_1_4_8_9, \
call_1_4_8_10, \
call_1_4_8_11, \
call_1_4_8_12, \
call_1_4_8_13, \
call_1_4_8_14, \
call_1_4_8_15, \
call_1_4_9_0, \
call_1_4_9_1, \
call_1_4_9_2, \
call_1_4_9_3, \
call_1_4_9_4, \
call_1_4_9_5, \
call_1_4_9_6, \
call_1_4_9_7, \
call_1_4_9_8, \
call_1_4_9_9, \
call_1_4_9_10, \
call_1_4_9_11, \
call_1_4_9_12, \
call_1_4_9_13, \
call_1_4_9_14, \
call_1_4_9_15, \
call_1_4_10_0, \
call_1_4_10_1, \
call_1_4_10_2, \
call_1_4_10_3, \
call_1_4_10_4, \
call_1_4_10_5, \
call_1_4_10_6, \
call_1_4_10_7, \
call_1_4_10_8, \
call_1_4_10_9, \
call_1_4_10_10, \
call_1_4_10_11, \
call_1_4_10_12, \
call_1_4_10_13, \
call_1_4_10_14, \
call_1_4_10_15, \
call_1_4_11_0, \
call_1_4_11_1, \
call_1_4_11_2, \
call_1_4_11_3, \
call_1_4_11_4, \
call_1_4_11_5, \
call_1_4_11_6, \
call_1_4_11_7, \
call_1_4_11_8, \
call_1_4_11_9, \
call_1_4_11_10, \
call_1_4_11_11, \
call_1_4_11_12, \
call_1_4_11_13, \
call_1_4_11_14, \
call_1_4_11_15, \
call_1_4_12_0, \
call_1_4_12_1, \
call_1_4_12_2, \
call_1_4_12_3, \
call_1_4_12_4, \
call_1_4_12_5, \
call_1_4_12_6, \
call_1_4_12_7, \
call_1_4_12_8, \
call_1_4_12_9, \
call_1_4_12_10, \
call_1_4_12_11, \
call_1_4_12_12, \
call_1_4_12_13, \
call_1_4_12_14, \
call_1_4_12_15, \
call_1_4_13_0, \
call_1_4_13_1, \
call_1_4_13_2, \
call_1_4_13_3, \
call_1_4_13_4, \
call_1_4_13_5, \
call_1_4_13_6, \
call_1_4_13_7, \
call_1_4_13_8, \
call_1_4_13_9, \
call_1_4_13_10, \
call_1_4_13_11, \
call_1_4_13_12, \
call_1_4_13_13, \
call_1_4_13_14, \
call_1_4_13_15, \
call_1_4_14_0, \
call_1_4_14_1, \
call_1_4_14_2, \
call_1_4_14_3, \
call_1_4_14_4, \
call_1_4_14_5, \
call_1_4_14_6, \
call_1_4_14_7, \
call_1_4_14_8, \
call_1_4_14_9, \
call_1_4_14_10, \
call_1_4_14_11, \
call_1_4_14_12, \
call_1_4_14_13, \
call_1_4_14_14, \
call_1_4_14_15, \
call_1_4_15_0, \
call_1_4_15_1, \
call_1_4_15_2, \
call_1_4_15_3, \
call_1_4_15_4, \
call_1_4_15_5, \
call_1_4_15_6, \
call_1_4_15_7, \
call_1_4_15_8, \
call_1_4_15_9, \
call_1_4_15_10, \
call_1_4_15_11, \
call_1_4_15_12, \
call_1_4_15_13, \
call_1_4_15_14, \
call_1_4_15_15, \
call_1_5_0_0, \
call_1_5_0_1, \
call_1_5_0_2, \
call_1_5_0_3, \
call_1_5_0_4, \
call_1_5_0_5, \
call_1_5_0_6, \
call_1_5_0_7, \
call_1_5_0_8, \
call_1_5_0_9, \
call_1_5_0_10, \
call_1_5_0_11, \
call_1_5_0_12, \
call_1_5_0_13, \
call_1_5_0_14, \
call_1_5_0_15, \
call_1_5_1_0, \
call_1_5_1_1, \
call_1_5_1_2, \
call_1_5_1_3, \
call_1_5_1_4, \
call_1_5_1_5, \
call_1_5_1_6, \
call_1_5_1_7, \
call_1_5_1_8, \
call_1_5_1_9, \
call_1_5_1_10, \
call_1_5_1_11, \
call_1_5_1_12, \
call_1_5_1_13, \
call_1_5_1_14, \
call_1_5_1_15, \
call_1_5_2_0, \
call_1_5_2_1, \
call_1_5_2_2, \
call_1_5_2_3, \
call_1_5_2_4, \
call_1_5_2_5, \
call_1_5_2_6, \
call_1_5_2_7, \
call_1_5_2_8, \
call_1_5_2_9, \
call_1_5_2_10, \
call_1_5_2_11, \
call_1_5_2_12, \
call_1_5_2_13, \
call_1_5_2_14, \
call_1_5_2_15, \
call_1_5_3_0, \
call_1_5_3_1, \
call_1_5_3_2, \
call_1_5_3_3, \
call_1_5_3_4, \
call_1_5_3_5, \
call_1_5_3_6, \
call_1_5_3_7, \
call_1_5_3_8, \
call_1_5_3_9, \
call_1_5_3_10, \
call_1_5_3_11, \
call_1_5_3_12, \
call_1_5_3_13, \
call_1_5_3_14, \
call_1_5_3_15, \
call_1_5_4_0, \
call_1_5_4_1, \
call_1_5_4_2, \
call_1_5_4_3, \
call_1_5_4_4, \
call_1_5_4_5, \
call_1_5_4_6, \
call_1_5_4_7, \
call_1_5_4_8, \
call_1_5_4_9, \
call_1_5_4_10, \
call_1_5_4_11, \
call_1_5_4_12, \
call_1_5_4_13, \
call_1_5_4_14, \
call_1_5_4_15, \
call_1_5_5_0, \
call_1_5_5_1, \
call_1_5_5_2, \
call_1_5_5_3, \
call_1_5_5_4, \
call_1_5_5_5, \
call_1_5_5_6, \
call_1_5_5_7, \
call_1_5_5_8, \
call_1_5_5_9, \
call_1_5_5_10, \
call_1_5_5_11, \
call_1_5_5_12, \
call_1_5_5_13, \
call_1_5_5_14, \
call_1_5_5_15, \
call_1_5_6_0, \
call_1_5_6_1, \
call_1_5_6_2, \
call_1_5_6_3, \
call_1_5_6_4, \
call_1_5_6_5, \
call_1_5_6_6, \
call_1_5_6_7, \
call_1_5_6_8, \
call_1_5_6_9, \
call_1_5_6_10, \
call_1_5_6_11, \
call_1_5_6_12, \
call_1_5_6_13, \
call_1_5_6_14, \
call_1_5_6_15, \
call_1_5_7_0, \
call_1_5_7_1, \
call_1_5_7_2, \
call_1_5_7_3, \
call_1_5_7_4, \
call_1_5_7_5, \
call_1_5_7_6, \
call_1_5_7_7, \
call_1_5_7_8, \
call_1_5_7_9, \
call_1_5_7_10, \
call_1_5_7_11, \
call_1_5_7_12, \
call_1_5_7_13, \
call_1_5_7_14, \
call_1_5_7_15, \
call_1_5_8_0, \
call_1_5_8_1, \
call_1_5_8_2, \
call_1_5_8_3, \
call_1_5_8_4, \
call_1_5_8_5, \
call_1_5_8_6, \
call_1_5_8_7, \
call_1_5_8_8, \
call_1_5_8_9, \
call_1_5_8_10, \
call_1_5_8_11, \
call_1_5_8_12, \
call_1_5_8_13, \
call_1_5_8_14, \
call_1_5_8_15, \
call_1_5_9_0, \
call_1_5_9_1, \
call_1_5_9_2, \
call_1_5_9_3, \
call_1_5_9_4, \
call_1_5_9_5, \
call_1_5_9_6, \
call_1_5_9_7, \
call_1_5_9_8, \
call_1_5_9_9, \
call_1_5_9_10, \
call_1_5_9_11, \
call_1_5_9_12, \
call_1_5_9_13, \
call_1_5_9_14, \
call_1_5_9_15, \
call_1_5_10_0, \
call_1_5_10_1, \
call_1_5_10_2, \
call_1_5_10_3, \
call_1_5_10_4, \
call_1_5_10_5, \
call_1_5_10_6, \
call_1_5_10_7, \
call_1_5_10_8, \
call_1_5_10_9, \
call_1_5_10_10, \
call_1_5_10_11, \
call_1_5_10_12, \
call_1_5_10_13, \
call_1_5_10_14, \
call_1_5_10_15, \
call_1_5_11_0, \
call_1_5_11_1, \
call_1_5_11_2, \
call_1_5_11_3, \
call_1_5_11_4, \
call_1_5_11_5, \
call_1_5_11_6, \
call_1_5_11_7, \
call_1_5_11_8, \
call_1_5_11_9, \
call_1_5_11_10, \
call_1_5_11_11, \
call_1_5_11_12, \
call_1_5_11_13, \
call_1_5_11_14, \
call_1_5_11_15, \
call_1_5_12_0, \
call_1_5_12_1, \
call_1_5_12_2, \
call_1_5_12_3, \
call_1_5_12_4, \
call_1_5_12_5, \
call_1_5_12_6, \
call_1_5_12_7, \
call_1_5_12_8, \
call_1_5_12_9, \
call_1_5_12_10, \
call_1_5_12_11, \
call_1_5_12_12, \
call_1_5_12_13, \
call_1_5_12_14, \
call_1_5_12_15, \
call_1_5_13_0, \
call_1_5_13_1, \
call_1_5_13_2, \
call_1_5_13_3, \
call_1_5_13_4, \
call_1_5_13_5, \
call_1_5_13_6, \
call_1_5_13_7, \
call_1_5_13_8, \
call_1_5_13_9, \
call_1_5_13_10, \
call_1_5_13_11, \
call_1_5_13_12, \
call_1_5_13_13, \
call_1_5_13_14, \
call_1_5_13_15, \
call_1_5_14_0, \
call_1_5_14_1, \
call_1_5_14_2, \
call_1_5_14_3, \
call_1_5_14_4, \
call_1_5_14_5, \
call_1_5_14_6, \
call_1_5_14_7, \
call_1_5_14_8, \
call_1_5_14_9, \
call_1_5_14_10, \
call_1_5_14_11, \
call_1_5_14_12, \
call_1_5_14_13, \
call_1_5_14_14, \
call_1_5_14_15, \
call_1_5_15_0, \
call_1_5_15_1, \
call_1_5_15_2, \
call_1_5_15_3, \
call_1_5_15_4, \
call_1_5_15_5, \
call_1_5_15_6, \
call_1_5_15_7, \
call_1_5_15_8, \
call_1_5_15_9, \
call_1_5_15_10, \
call_1_5_15_11, \
call_1_5_15_12, \
call_1_5_15_13, \
call_1_5_15_14, \
call_1_5_15_15, \
call_1_6_0_0, \
call_1_6_0_1, \
call_1_6_0_2, \
call_1_6_0_3, \
call_1_6_0_4, \
call_1_6_0_5, \
call_1_6_0_6, \
call_1_6_0_7, \
call_1_6_0_8, \
call_1_6_0_9, \
call_1_6_0_10, \
call_1_6_0_11, \
call_1_6_0_12, \
call_1_6_0_13, \
call_1_6_0_14, \
call_1_6_0_15, \
call_1_6_1_0, \
call_1_6_1_1, \
call_1_6_1_2, \
call_1_6_1_3, \
call_1_6_1_4, \
call_1_6_1_5, \
call_1_6_1_6, \
call_1_6_1_7, \
call_1_6_1_8, \
call_1_6_1_9, \
call_1_6_1_10, \
call_1_6_1_11, \
call_1_6_1_12, \
call_1_6_1_13, \
call_1_6_1_14, \
call_1_6_1_15, \
call_1_6_2_0, \
call_1_6_2_1, \
call_1_6_2_2, \
call_1_6_2_3, \
call_1_6_2_4, \
call_1_6_2_5, \
call_1_6_2_6, \
call_1_6_2_7, \
call_1_6_2_8, \
call_1_6_2_9, \
call_1_6_2_10, \
call_1_6_2_11, \
call_1_6_2_12, \
call_1_6_2_13, \
call_1_6_2_14, \
call_1_6_2_15, \
call_1_6_3_0, \
call_1_6_3_1, \
call_1_6_3_2, \
call_1_6_3_3, \
call_1_6_3_4, \
call_1_6_3_5, \
call_1_6_3_6, \
call_1_6_3_7, \
call_1_6_3_8, \
call_1_6_3_9, \
call_1_6_3_10, \
call_1_6_3_11, \
call_1_6_3_12, \
call_1_6_3_13, \
call_1_6_3_14, \
call_1_6_3_15, \
call_1_6_4_0, \
call_1_6_4_1, \
call_1_6_4_2, \
call_1_6_4_3, \
call_1_6_4_4, \
call_1_6_4_5, \
call_1_6_4_6, \
call_1_6_4_7, \
call_1_6_4_8, \
call_1_6_4_9, \
call_1_6_4_10, \
call_1_6_4_11, \
call_1_6_4_12, \
call_1_6_4_13, \
call_1_6_4_14, \
call_1_6_4_15, \
call_1_6_5_0, \
call_1_6_5_1, \
call_1_6_5_2, \
call_1_6_5_3, \
call_1_6_5_4, \
call_1_6_5_5, \
call_1_6_5_6, \
call_1_6_5_7, \
call_1_6_5_8, \
call_1_6_5_9, \
call_1_6_5_10, \
call_1_6_5_11, \
call_1_6_5_12, \
call_1_6_5_13, \
call_1_6_5_14, \
call_1_6_5_15, \
call_1_6_6_0, \
call_1_6_6_1, \
call_1_6_6_2, \
call_1_6_6_3, \
call_1_6_6_4, \
call_1_6_6_5, \
call_1_6_6_6, \
call_1_6_6_7, \
call_1_6_6_8, \
call_1_6_6_9, \
call_1_6_6_10, \
call_1_6_6_11, \
call_1_6_6_12, \
call_1_6_6_13, \
call_1_6_6_14, \
call_1_6_6_15, \
call_1_6_7_0, \
call_1_6_7_1, \
call_1_6_7_2, \
call_1_6_7_3, \
call_1_6_7_4, \
call_1_6_7_5, \
call_1_6_7_6, \
call_1_6_7_7, \
call_1_6_7_8, \
call_1_6_7_9, \
call_1_6_7_10, \
call_1_6_7_11, \
call_1_6_7_12, \
call_1_6_7_13, \
call_1_6_7_14, \
call_1_6_7_15, \
call_1_6_8_0, \
call_1_6_8_1, \
call_1_6_8_2, \
call_1_6_8_3, \
call_1_6_8_4, \
call_1_6_8_5, \
call_1_6_8_6, \
call_1_6_8_7, \
call_1_6_8_8, \
call_1_6_8_9, \
call_1_6_8_10, \
call_1_6_8_11, \
call_1_6_8_12, \
call_1_6_8_13, \
call_1_6_8_14, \
call_1_6_8_15, \
call_1_6_9_0, \
call_1_6_9_1, \
call_1_6_9_2, \
call_1_6_9_3, \
call_1_6_9_4, \
call_1_6_9_5, \
call_1_6_9_6, \
call_1_6_9_7, \
call_1_6_9_8, \
call_1_6_9_9, \
call_1_6_9_10, \
call_1_6_9_11, \
call_1_6_9_12, \
call_1_6_9_13, \
call_1_6_9_14, \
call_1_6_9_15, \
call_1_6_10_0, \
call_1_6_10_1, \
call_1_6_10_2, \
call_1_6_10_3, \
call_1_6_10_4, \
call_1_6_10_5, \
call_1_6_10_6, \
call_1_6_10_7, \
call_1_6_10_8, \
call_1_6_10_9, \
call_1_6_10_10, \
call_1_6_10_11, \
call_1_6_10_12, \
call_1_6_10_13, \
call_1_6_10_14, \
call_1_6_10_15, \
call_1_6_11_0, \
call_1_6_11_1, \
call_1_6_11_2, \
call_1_6_11_3, \
call_1_6_11_4, \
call_1_6_11_5, \
call_1_6_11_6, \
call_1_6_11_7, \
call_1_6_11_8, \
call_1_6_11_9, \
call_1_6_11_10, \
call_1_6_11_11, \
call_1_6_11_12, \
call_1_6_11_13, \
call_1_6_11_14, \
call_1_6_11_15, \
call_1_6_12_0, \
call_1_6_12_1, \
call_1_6_12_2, \
call_1_6_12_3, \
call_1_6_12_4, \
call_1_6_12_5, \
call_1_6_12_6, \
call_1_6_12_7, \
call_1_6_12_8, \
call_1_6_12_9, \
call_1_6_12_10, \
call_1_6_12_11, \
call_1_6_12_12, \
call_1_6_12_13, \
call_1_6_12_14, \
call_1_6_12_15, \
call_1_6_13_0, \
call_1_6_13_1, \
call_1_6_13_2, \
call_1_6_13_3, \
call_1_6_13_4, \
call_1_6_13_5, \
call_1_6_13_6, \
call_1_6_13_7, \
call_1_6_13_8, \
call_1_6_13_9, \
call_1_6_13_10, \
call_1_6_13_11, \
call_1_6_13_12, \
call_1_6_13_13, \
call_1_6_13_14, \
call_1_6_13_15, \
call_1_6_14_0, \
call_1_6_14_1, \
call_1_6_14_2, \
call_1_6_14_3, \
call_1_6_14_4, \
call_1_6_14_5, \
call_1_6_14_6, \
call_1_6_14_7, \
call_1_6_14_8, \
call_1_6_14_9, \
call_1_6_14_10, \
call_1_6_14_11, \
call_1_6_14_12, \
call_1_6_14_13, \
call_1_6_14_14, \
call_1_6_14_15, \
call_1_6_15_0, \
call_1_6_15_1, \
call_1_6_15_2, \
call_1_6_15_3, \
call_1_6_15_4, \
call_1_6_15_5, \
call_1_6_15_6, \
call_1_6_15_7, \
call_1_6_15_8, \
call_1_6_15_9, \
call_1_6_15_10, \
call_1_6_15_11, \
call_1_6_15_12, \
call_1_6_15_13, \
call_1_6_15_14, \
call_1_6_15_15, \
call_1_7_0_0, \
call_1_7_0_1, \
call_1_7_0_2, \
call_1_7_0_3, \
call_1_7_0_4, \
call_1_7_0_5, \
call_1_7_0_6, \
call_1_7_0_7, \
call_1_7_0_8, \
call_1_7_0_9, \
call_1_7_0_10, \
call_1_7_0_11, \
call_1_7_0_12, \
call_1_7_0_13, \
call_1_7_0_14, \
call_1_7_0_15, \
call_1_7_1_0, \
call_1_7_1_1, \
call_1_7_1_2, \
call_1_7_1_3, \
call_1_7_1_4, \
call_1_7_1_5, \
call_1_7_1_6, \
call_1_7_1_7, \
call_1_7_1_8, \
call_1_7_1_9, \
call_1_7_1_10, \
call_1_7_1_11, \
call_1_7_1_12, \
call_1_7_1_13, \
call_1_7_1_14, \
call_1_7_1_15, \
call_1_7_2_0, \
call_1_7_2_1, \
call_1_7_2_2, \
call_1_7_2_3, \
call_1_7_2_4, \
call_1_7_2_5, \
call_1_7_2_6, \
call_1_7_2_7, \
call_1_7_2_8, \
call_1_7_2_9, \
call_1_7_2_10, \
call_1_7_2_11, \
call_1_7_2_12, \
call_1_7_2_13, \
call_1_7_2_14, \
call_1_7_2_15, \
call_1_7_3_0, \
call_1_7_3_1, \
call_1_7_3_2, \
call_1_7_3_3, \
call_1_7_3_4, \
call_1_7_3_5, \
call_1_7_3_6, \
call_1_7_3_7, \
call_1_7_3_8, \
call_1_7_3_9, \
call_1_7_3_10, \
call_1_7_3_11, \
call_1_7_3_12, \
call_1_7_3_13, \
call_1_7_3_14, \
call_1_7_3_15, \
call_1_7_4_0, \
call_1_7_4_1, \
call_1_7_4_2, \
call_1_7_4_3, \
call_1_7_4_4, \
call_1_7_4_5, \
call_1_7_4_6, \
call_1_7_4_7, \
call_1_7_4_8, \
call_1_7_4_9, \
call_1_7_4_10, \
call_1_7_4_11, \
call_1_7_4_12, \
call_1_7_4_13, \
call_1_7_4_14, \
call_1_7_4_15, \
call_1_7_5_0, \
call_1_7_5_1, \
call_1_7_5_2, \
call_1_7_5_3, \
call_1_7_5_4, \
call_1_7_5_5, \
call_1_7_5_6, \
call_1_7_5_7, \
call_1_7_5_8, \
call_1_7_5_9, \
call_1_7_5_10, \
call_1_7_5_11, \
call_1_7_5_12, \
call_1_7_5_13, \
call_1_7_5_14, \
call_1_7_5_15, \
call_1_7_6_0, \
call_1_7_6_1, \
call_1_7_6_2, \
call_1_7_6_3, \
call_1_7_6_4, \
call_1_7_6_5, \
call_1_7_6_6, \
call_1_7_6_7, \
call_1_7_6_8, \
call_1_7_6_9, \
call_1_7_6_10, \
call_1_7_6_11, \
call_1_7_6_12, \
call_1_7_6_13, \
call_1_7_6_14, \
call_1_7_6_15, \
call_1_7_7_0, \
call_1_7_7_1, \
call_1_7_7_2, \
call_1_7_7_3, \
call_1_7_7_4, \
call_1_7_7_5, \
call_1_7_7_6, \
call_1_7_7_7, \
call_1_7_7_8, \
call_1_7_7_9, \
call_1_7_7_10, \
call_1_7_7_11, \
call_1_7_7_12, \
call_1_7_7_13, \
call_1_7_7_14, \
call_1_7_7_15, \
call_1_7_8_0, \
call_1_7_8_1, \
call_1_7_8_2, \
call_1_7_8_3, \
call_1_7_8_4, \
call_1_7_8_5, \
call_1_7_8_6, \
call_1_7_8_7, \
call_1_7_8_8, \
call_1_7_8_9, \
call_1_7_8_10, \
call_1_7_8_11, \
call_1_7_8_12, \
call_1_7_8_13, \
call_1_7_8_14, \
call_1_7_8_15, \
call_1_7_9_0, \
call_1_7_9_1, \
call_1_7_9_2, \
call_1_7_9_3, \
call_1_7_9_4, \
call_1_7_9_5, \
call_1_7_9_6, \
call_1_7_9_7, \
call_1_7_9_8, \
call_1_7_9_9, \
call_1_7_9_10, \
call_1_7_9_11, \
call_1_7_9_12, \
call_1_7_9_13, \
call_1_7_9_14, \
call_1_7_9_15, \
call_1_7_10_0, \
call_1_7_10_1, \
call_1_7_10_2, \
call_1_7_10_3, \
call_1_7_10_4, \
call_1_7_10_5, \
call_1_7_10_6, \
call_1_7_10_7, \
call_1_7_10_8, \
call_1_7_10_9, \
call_1_7_10_10, \
call_1_7_10_11, \
call_1_7_10_12, \
call_1_7_10_13, \
call_1_7_10_14, \
call_1_7_10_15, \
call_1_7_11_0, \
call_1_7_11_1, \
call_1_7_11_2, \
call_1_7_11_3, \
call_1_7_11_4, \
call_1_7_11_5, \
call_1_7_11_6, \
call_1_7_11_7, \
call_1_7_11_8, \
call_1_7_11_9, \
call_1_7_11_10, \
call_1_7_11_11, \
call_1_7_11_12, \
call_1_7_11_13, \
call_1_7_11_14, \
call_1_7_11_15, \
call_1_7_12_0, \
call_1_7_12_1, \
call_1_7_12_2, \
call_1_7_12_3, \
call_1_7_12_4, \
call_1_7_12_5, \
call_1_7_12_6, \
call_1_7_12_7, \
call_1_7_12_8, \
call_1_7_12_9, \
call_1_7_12_10, \
call_1_7_12_11, \
call_1_7_12_12, \
call_1_7_12_13, \
call_1_7_12_14, \
call_1_7_12_15, \
call_1_7_13_0, \
call_1_7_13_1, \
call_1_7_13_2, \
call_1_7_13_3, \
call_1_7_13_4, \
call_1_7_13_5, \
call_1_7_13_6, \
call_1_7_13_7, \
call_1_7_13_8, \
call_1_7_13_9, \
call_1_7_13_10, \
call_1_7_13_11, \
call_1_7_13_12, \
call_1_7_13_13, \
call_1_7_13_14, \
call_1_7_13_15, \
call_1_7_14_0, \
call_1_7_14_1, \
call_1_7_14_2, \
call_1_7_14_3, \
call_1_7_14_4, \
call_1_7_14_5, \
call_1_7_14_6, \
call_1_7_14_7, \
call_1_7_14_8, \
call_1_7_14_9, \
call_1_7_14_10, \
call_1_7_14_11, \
call_1_7_14_12, \
call_1_7_14_13, \
call_1_7_14_14, \
call_1_7_14_15, \
call_1_7_15_0, \
call_1_7_15_1, \
call_1_7_15_2, \
call_1_7_15_3, \
call_1_7_15_4, \
call_1_7_15_5, \
call_1_7_15_6, \
call_1_7_15_7, \
call_1_7_15_8, \
call_1_7_15_9, \
call_1_7_15_10, \
call_1_7_15_11, \
call_1_7_15_12, \
call_1_7_15_13, \
call_1_7_15_14, \
call_1_7_15_15, \
call_1_8_0_0, \
call_1_8_0_1, \
call_1_8_0_2, \
call_1_8_0_3, \
call_1_8_0_4, \
call_1_8_0_5, \
call_1_8_0_6, \
call_1_8_0_7, \
call_1_8_0_8, \
call_1_8_0_9, \
call_1_8_0_10, \
call_1_8_0_11, \
call_1_8_0_12, \
call_1_8_0_13, \
call_1_8_0_14, \
call_1_8_0_15, \
call_1_8_1_0, \
call_1_8_1_1, \
call_1_8_1_2, \
call_1_8_1_3, \
call_1_8_1_4, \
call_1_8_1_5, \
call_1_8_1_6, \
call_1_8_1_7, \
call_1_8_1_8, \
call_1_8_1_9, \
call_1_8_1_10, \
call_1_8_1_11, \
call_1_8_1_12, \
call_1_8_1_13, \
call_1_8_1_14, \
call_1_8_1_15, \
call_1_8_2_0, \
call_1_8_2_1, \
call_1_8_2_2, \
call_1_8_2_3, \
call_1_8_2_4, \
call_1_8_2_5, \
call_1_8_2_6, \
call_1_8_2_7, \
call_1_8_2_8, \
call_1_8_2_9, \
call_1_8_2_10, \
call_1_8_2_11, \
call_1_8_2_12, \
call_1_8_2_13, \
call_1_8_2_14, \
call_1_8_2_15, \
call_1_8_3_0, \
call_1_8_3_1, \
call_1_8_3_2, \
call_1_8_3_3, \
call_1_8_3_4, \
call_1_8_3_5, \
call_1_8_3_6, \
call_1_8_3_7, \
call_1_8_3_8, \
call_1_8_3_9, \
call_1_8_3_10, \
call_1_8_3_11, \
call_1_8_3_12, \
call_1_8_3_13, \
call_1_8_3_14, \
call_1_8_3_15, \
call_1_8_4_0, \
call_1_8_4_1, \
call_1_8_4_2, \
call_1_8_4_3, \
call_1_8_4_4, \
call_1_8_4_5, \
call_1_8_4_6, \
call_1_8_4_7, \
call_1_8_4_8, \
call_1_8_4_9, \
call_1_8_4_10, \
call_1_8_4_11, \
call_1_8_4_12, \
call_1_8_4_13, \
call_1_8_4_14, \
call_1_8_4_15, \
call_1_8_5_0, \
call_1_8_5_1, \
call_1_8_5_2, \
call_1_8_5_3, \
call_1_8_5_4, \
call_1_8_5_5, \
call_1_8_5_6, \
call_1_8_5_7, \
call_1_8_5_8, \
call_1_8_5_9, \
call_1_8_5_10, \
call_1_8_5_11, \
call_1_8_5_12, \
call_1_8_5_13, \
call_1_8_5_14, \
call_1_8_5_15, \
call_1_8_6_0, \
call_1_8_6_1, \
call_1_8_6_2, \
call_1_8_6_3, \
call_1_8_6_4, \
call_1_8_6_5, \
call_1_8_6_6, \
call_1_8_6_7, \
call_1_8_6_8, \
call_1_8_6_9, \
call_1_8_6_10, \
call_1_8_6_11, \
call_1_8_6_12, \
call_1_8_6_13, \
call_1_8_6_14, \
call_1_8_6_15, \
call_1_8_7_0, \
call_1_8_7_1, \
call_1_8_7_2, \
call_1_8_7_3, \
call_1_8_7_4, \
call_1_8_7_5, \
call_1_8_7_6, \
call_1_8_7_7, \
call_1_8_7_8, \
call_1_8_7_9, \
call_1_8_7_10, \
call_1_8_7_11, \
call_1_8_7_12, \
call_1_8_7_13, \
call_1_8_7_14, \
call_1_8_7_15, \
call_1_8_8_0, \
call_1_8_8_1, \
call_1_8_8_2, \
call_1_8_8_3, \
call_1_8_8_4, \
call_1_8_8_5, \
call_1_8_8_6, \
call_1_8_8_7, \
call_1_8_8_8, \
call_1_8_8_9, \
call_1_8_8_10, \
call_1_8_8_11, \
call_1_8_8_12, \
call_1_8_8_13, \
call_1_8_8_14, \
call_1_8_8_15, \
call_1_8_9_0, \
call_1_8_9_1, \
call_1_8_9_2, \
call_1_8_9_3, \
call_1_8_9_4, \
call_1_8_9_5, \
call_1_8_9_6, \
call_1_8_9_7, \
call_1_8_9_8, \
call_1_8_9_9, \
call_1_8_9_10, \
call_1_8_9_11, \
call_1_8_9_12, \
call_1_8_9_13, \
call_1_8_9_14, \
call_1_8_9_15, \
call_1_8_10_0, \
call_1_8_10_1, \
call_1_8_10_2, \
call_1_8_10_3, \
call_1_8_10_4, \
call_1_8_10_5, \
call_1_8_10_6, \
call_1_8_10_7, \
call_1_8_10_8, \
call_1_8_10_9, \
call_1_8_10_10, \
call_1_8_10_11, \
call_1_8_10_12, \
call_1_8_10_13, \
call_1_8_10_14, \
call_1_8_10_15, \
call_1_8_11_0, \
call_1_8_11_1, \
call_1_8_11_2, \
call_1_8_11_3, \
call_1_8_11_4, \
call_1_8_11_5, \
call_1_8_11_6, \
call_1_8_11_7, \
call_1_8_11_8, \
call_1_8_11_9, \
call_1_8_11_10, \
call_1_8_11_11, \
call_1_8_11_12, \
call_1_8_11_13, \
call_1_8_11_14, \
call_1_8_11_15, \
call_1_8_12_0, \
call_1_8_12_1, \
call_1_8_12_2, \
call_1_8_12_3, \
call_1_8_12_4, \
call_1_8_12_5, \
call_1_8_12_6, \
call_1_8_12_7, \
call_1_8_12_8, \
call_1_8_12_9, \
call_1_8_12_10, \
call_1_8_12_11, \
call_1_8_12_12, \
call_1_8_12_13, \
call_1_8_12_14, \
call_1_8_12_15, \
call_1_8_13_0, \
call_1_8_13_1, \
call_1_8_13_2, \
call_1_8_13_3, \
call_1_8_13_4, \
call_1_8_13_5, \
call_1_8_13_6, \
call_1_8_13_7, \
call_1_8_13_8, \
call_1_8_13_9, \
call_1_8_13_10, \
call_1_8_13_11, \
call_1_8_13_12, \
call_1_8_13_13, \
call_1_8_13_14, \
call_1_8_13_15, \
call_1_8_14_0, \
call_1_8_14_1, \
call_1_8_14_2, \
call_1_8_14_3, \
call_1_8_14_4, \
call_1_8_14_5, \
call_1_8_14_6, \
call_1_8_14_7, \
call_1_8_14_8, \
call_1_8_14_9, \
call_1_8_14_10, \
call_1_8_14_11, \
call_1_8_14_12, \
call_1_8_14_13, \
call_1_8_14_14, \
call_1_8_14_15, \
call_1_8_15_0, \
call_1_8_15_1, \
call_1_8_15_2, \
call_1_8_15_3, \
call_1_8_15_4, \
call_1_8_15_5, \
call_1_8_15_6, \
call_1_8_15_7, \
call_1_8_15_8, \
call_1_8_15_9, \
call_1_8_15_10, \
call_1_8_15_11, \
call_1_8_15_12, \
call_1_8_15_13, \
call_1_8_15_14, \
call_1_8_15_15, \
call_1_9_0_0, \
call_1_9_0_1, \
call_1_9_0_2, \
call_1_9_0_3, \
call_1_9_0_4, \
call_1_9_0_5, \
call_1_9_0_6, \
call_1_9_0_7, \
call_1_9_0_8, \
call_1_9_0_9, \
call_1_9_0_10, \
call_1_9_0_11, \
call_1_9_0_12, \
call_1_9_0_13, \
call_1_9_0_14, \
call_1_9_0_15, \
call_1_9_1_0, \
call_1_9_1_1, \
call_1_9_1_2, \
call_1_9_1_3, \
call_1_9_1_4, \
call_1_9_1_5, \
call_1_9_1_6, \
call_1_9_1_7, \
call_1_9_1_8, \
call_1_9_1_9, \
call_1_9_1_10, \
call_1_9_1_11, \
call_1_9_1_12, \
call_1_9_1_13, \
call_1_9_1_14, \
call_1_9_1_15, \
call_1_9_2_0, \
call_1_9_2_1, \
call_1_9_2_2, \
call_1_9_2_3, \
call_1_9_2_4, \
call_1_9_2_5, \
call_1_9_2_6, \
call_1_9_2_7, \
call_1_9_2_8, \
call_1_9_2_9, \
call_1_9_2_10, \
call_1_9_2_11, \
call_1_9_2_12, \
call_1_9_2_13, \
call_1_9_2_14, \
call_1_9_2_15, \
call_1_9_3_0, \
call_1_9_3_1, \
call_1_9_3_2, \
call_1_9_3_3, \
call_1_9_3_4, \
call_1_9_3_5, \
call_1_9_3_6, \
call_1_9_3_7, \
call_1_9_3_8, \
call_1_9_3_9, \
call_1_9_3_10, \
call_1_9_3_11, \
call_1_9_3_12, \
call_1_9_3_13, \
call_1_9_3_14, \
call_1_9_3_15, \
call_1_9_4_0, \
call_1_9_4_1, \
call_1_9_4_2, \
call_1_9_4_3, \
call_1_9_4_4, \
call_1_9_4_5, \
call_1_9_4_6, \
call_1_9_4_7, \
call_1_9_4_8, \
call_1_9_4_9, \
call_1_9_4_10, \
call_1_9_4_11, \
call_1_9_4_12, \
call_1_9_4_13, \
call_1_9_4_14, \
call_1_9_4_15, \
call_1_9_5_0, \
call_1_9_5_1, \
call_1_9_5_2, \
call_1_9_5_3, \
call_1_9_5_4, \
call_1_9_5_5, \
call_1_9_5_6, \
call_1_9_5_7, \
call_1_9_5_8, \
call_1_9_5_9, \
call_1_9_5_10, \
call_1_9_5_11, \
call_1_9_5_12, \
call_1_9_5_13, \
call_1_9_5_14, \
call_1_9_5_15, \
call_1_9_6_0, \
call_1_9_6_1, \
call_1_9_6_2, \
call_1_9_6_3, \
call_1_9_6_4, \
call_1_9_6_5, \
call_1_9_6_6, \
call_1_9_6_7, \
call_1_9_6_8, \
call_1_9_6_9, \
call_1_9_6_10, \
call_1_9_6_11, \
call_1_9_6_12, \
call_1_9_6_13, \
call_1_9_6_14, \
call_1_9_6_15, \
call_1_9_7_0, \
call_1_9_7_1, \
call_1_9_7_2, \
call_1_9_7_3, \
call_1_9_7_4, \
call_1_9_7_5, \
call_1_9_7_6, \
call_1_9_7_7, \
call_1_9_7_8, \
call_1_9_7_9, \
call_1_9_7_10, \
call_1_9_7_11, \
call_1_9_7_12, \
call_1_9_7_13, \
call_1_9_7_14, \
call_1_9_7_15, \
call_1_9_8_0, \
call_1_9_8_1, \
call_1_9_8_2, \
call_1_9_8_3, \
call_1_9_8_4, \
call_1_9_8_5, \
call_1_9_8_6, \
call_1_9_8_7, \
call_1_9_8_8, \
call_1_9_8_9, \
call_1_9_8_10, \
call_1_9_8_11, \
call_1_9_8_12, \
call_1_9_8_13, \
call_1_9_8_14, \
call_1_9_8_15, \
call_1_9_9_0, \
call_1_9_9_1, \
call_1_9_9_2, \
call_1_9_9_3, \
call_1_9_9_4, \
call_1_9_9_5, \
call_1_9_9_6, \
call_1_9_9_7, \
call_1_9_9_8, \
call_1_9_9_9, \
call_1_9_9_10, \
call_1_9_9_11, \
call_1_9_9_12, \
call_1_9_9_13, \
call_1_9_9_14, \
call_1_9_9_15, \
call_1_9_10_0, \
call_1_9_10_1, \
call_1_9_10_2, \
call_1_9_10_3, \
call_1_9_10_4, \
call_1_9_10_5, \
call_1_9_10_6, \
call_1_9_10_7, \
call_1_9_10_8, \
call_1_9_10_9, \
call_1_9_10_10, \
call_1_9_10_11, \
call_1_9_10_12, \
call_1_9_10_13, \
call_1_9_10_14, \
call_1_9_10_15, \
call_1_9_11_0, \
call_1_9_11_1, \
call_1_9_11_2, \
call_1_9_11_3, \
call_1_9_11_4, \
call_1_9_11_5, \
call_1_9_11_6, \
call_1_9_11_7, \
call_1_9_11_8, \
call_1_9_11_9, \
call_1_9_11_10, \
call_1_9_11_11, \
call_1_9_11_12, \
call_1_9_11_13, \
call_1_9_11_14, \
call_1_9_11_15, \
call_1_9_12_0, \
call_1_9_12_1, \
call_1_9_12_2, \
call_1_9_12_3, \
call_1_9_12_4, \
call_1_9_12_5, \
call_1_9_12_6, \
call_1_9_12_7, \
call_1_9_12_8, \
call_1_9_12_9, \
call_1_9_12_10, \
call_1_9_12_11, \
call_1_9_12_12, \
call_1_9_12_13, \
call_1_9_12_14, \
call_1_9_12_15, \
call_1_9_13_0, \
call_1_9_13_1, \
call_1_9_13_2, \
call_1_9_13_3, \
call_1_9_13_4, \
call_1_9_13_5, \
call_1_9_13_6, \
call_1_9_13_7, \
call_1_9_13_8, \
call_1_9_13_9, \
call_1_9_13_10, \
call_1_9_13_11, \
call_1_9_13_12, \
call_1_9_13_13, \
call_1_9_13_14, \
call_1_9_13_15, \
call_1_9_14_0, \
call_1_9_14_1, \
call_1_9_14_2, \
call_1_9_14_3, \
call_1_9_14_4, \
call_1_9_14_5, \
call_1_9_14_6, \
call_1_9_14_7, \
call_1_9_14_8, \
call_1_9_14_9, \
call_1_9_14_10, \
call_1_9_14_11, \
call_1_9_14_12, \
call_1_9_14_13, \
call_1_9_14_14, \
call_1_9_14_15, \
call_1_9_15_0, \
call_1_9_15_1, \
call_1_9_15_2, \
call_1_9_15_3, \
call_1_9_15_4, \
call_1_9_15_5, \
call_1_9_15_6, \
call_1_9_15_7, \
call_1_9_15_8, \
call_1_9_15_9, \
call_1_9_15_10, \
call_1_9_15_11, \
call_1_9_15_12, \
call_1_9_15_13, \
call_1_9_15_14, \
call_1_9_15_15, \
call_1_10_0_0, \
call_1_10_0_1, \
call_1_10_0_2, \
call_1_10_0_3, \
call_1_10_0_4, \
call_1_10_0_5, \
call_1_10_0_6, \
call_1_10_0_7, \
call_1_10_0_8, \
call_1_10_0_9, \
call_1_10_0_10, \
call_1_10_0_11, \
call_1_10_0_12, \
call_1_10_0_13, \
call_1_10_0_14, \
call_1_10_0_15, \
call_1_10_1_0, \
call_1_10_1_1, \
call_1_10_1_2, \
call_1_10_1_3, \
call_1_10_1_4, \
call_1_10_1_5, \
call_1_10_1_6, \
call_1_10_1_7, \
call_1_10_1_8, \
call_1_10_1_9, \
call_1_10_1_10, \
call_1_10_1_11, \
call_1_10_1_12, \
call_1_10_1_13, \
call_1_10_1_14, \
call_1_10_1_15, \
call_1_10_2_0, \
call_1_10_2_1, \
call_1_10_2_2, \
call_1_10_2_3, \
call_1_10_2_4, \
call_1_10_2_5, \
call_1_10_2_6, \
call_1_10_2_7, \
call_1_10_2_8, \
call_1_10_2_9, \
call_1_10_2_10, \
call_1_10_2_11, \
call_1_10_2_12, \
call_1_10_2_13, \
call_1_10_2_14, \
call_1_10_2_15, \
call_1_10_3_0, \
call_1_10_3_1, \
call_1_10_3_2, \
call_1_10_3_3, \
call_1_10_3_4, \
call_1_10_3_5, \
call_1_10_3_6, \
call_1_10_3_7, \
call_1_10_3_8, \
call_1_10_3_9, \
call_1_10_3_10, \
call_1_10_3_11, \
call_1_10_3_12, \
call_1_10_3_13, \
call_1_10_3_14, \
call_1_10_3_15, \
call_1_10_4_0, \
call_1_10_4_1, \
call_1_10_4_2, \
call_1_10_4_3, \
call_1_10_4_4, \
call_1_10_4_5, \
call_1_10_4_6, \
call_1_10_4_7, \
call_1_10_4_8, \
call_1_10_4_9, \
call_1_10_4_10, \
call_1_10_4_11, \
call_1_10_4_12, \
call_1_10_4_13, \
call_1_10_4_14, \
call_1_10_4_15, \
call_1_10_5_0, \
call_1_10_5_1, \
call_1_10_5_2, \
call_1_10_5_3, \
call_1_10_5_4, \
call_1_10_5_5, \
call_1_10_5_6, \
call_1_10_5_7, \
call_1_10_5_8, \
call_1_10_5_9, \
call_1_10_5_10, \
call_1_10_5_11, \
call_1_10_5_12, \
call_1_10_5_13, \
call_1_10_5_14, \
call_1_10_5_15, \
call_1_10_6_0, \
call_1_10_6_1, \
call_1_10_6_2, \
call_1_10_6_3, \
call_1_10_6_4, \
call_1_10_6_5, \
call_1_10_6_6, \
call_1_10_6_7, \
call_1_10_6_8, \
call_1_10_6_9, \
call_1_10_6_10, \
call_1_10_6_11, \
call_1_10_6_12, \
call_1_10_6_13, \
call_1_10_6_14, \
call_1_10_6_15, \
call_1_10_7_0, \
call_1_10_7_1, \
call_1_10_7_2, \
call_1_10_7_3, \
call_1_10_7_4, \
call_1_10_7_5, \
call_1_10_7_6, \
call_1_10_7_7, \
call_1_10_7_8, \
call_1_10_7_9, \
call_1_10_7_10, \
call_1_10_7_11, \
call_1_10_7_12, \
call_1_10_7_13, \
call_1_10_7_14, \
call_1_10_7_15, \
call_1_10_8_0, \
call_1_10_8_1, \
call_1_10_8_2, \
call_1_10_8_3, \
call_1_10_8_4, \
call_1_10_8_5, \
call_1_10_8_6, \
call_1_10_8_7, \
call_1_10_8_8, \
call_1_10_8_9, \
call_1_10_8_10, \
call_1_10_8_11, \
call_1_10_8_12, \
call_1_10_8_13, \
call_1_10_8_14, \
call_1_10_8_15, \
call_1_10_9_0, \
call_1_10_9_1, \
call_1_10_9_2, \
call_1_10_9_3, \
call_1_10_9_4, \
call_1_10_9_5, \
call_1_10_9_6, \
call_1_10_9_7, \
call_1_10_9_8, \
call_1_10_9_9, \
call_1_10_9_10, \
call_1_10_9_11, \
call_1_10_9_12, \
call_1_10_9_13, \
call_1_10_9_14, \
call_1_10_9_15, \
call_1_10_10_0, \
call_1_10_10_1, \
call_1_10_10_2, \
call_1_10_10_3, \
call_1_10_10_4, \
call_1_10_10_5, \
call_1_10_10_6, \
call_1_10_10_7, \
call_1_10_10_8, \
call_1_10_10_9, \
call_1_10_10_10, \
call_1_10_10_11, \
call_1_10_10_12, \
call_1_10_10_13, \
call_1_10_10_14, \
call_1_10_10_15, \
call_1_10_11_0, \
call_1_10_11_1, \
call_1_10_11_2, \
call_1_10_11_3, \
call_1_10_11_4, \
call_1_10_11_5, \
call_1_10_11_6, \
call_1_10_11_7, \
call_1_10_11_8, \
call_1_10_11_9, \
call_1_10_11_10, \
call_1_10_11_11, \
call_1_10_11_12, \
call_1_10_11_13, \
call_1_10_11_14, \
call_1_10_11_15, \
call_1_10_12_0, \
call_1_10_12_1, \
call_1_10_12_2, \
call_1_10_12_3, \
call_1_10_12_4, \
call_1_10_12_5, \
call_1_10_12_6, \
call_1_10_12_7, \
call_1_10_12_8, \
call_1_10_12_9, \
call_1_10_12_10, \
call_1_10_12_11, \
call_1_10_12_12, \
call_1_10_12_13, \
call_1_10_12_14, \
call_1_10_12_15, \
call_1_10_13_0, \
call_1_10_13_1, \
call_1_10_13_2, \
call_1_10_13_3, \
call_1_10_13_4, \
call_1_10_13_5, \
call_1_10_13_6, \
call_1_10_13_7, \
call_1_10_13_8, \
call_1_10_13_9, \
call_1_10_13_10, \
call_1_10_13_11, \
call_1_10_13_12, \
call_1_10_13_13, \
call_1_10_13_14, \
call_1_10_13_15, \
call_1_10_14_0, \
call_1_10_14_1, \
call_1_10_14_2, \
call_1_10_14_3, \
call_1_10_14_4, \
call_1_10_14_5, \
call_1_10_14_6, \
call_1_10_14_7, \
call_1_10_14_8, \
call_1_10_14_9, \
call_1_10_14_10, \
call_1_10_14_11, \
call_1_10_14_12, \
call_1_10_14_13, \
call_1_10_14_14, \
call_1_10_14_15, \
call_1_10_15_0, \
call_1_10_15_1, \
call_1_10_15_2, \
call_1_10_15_3, \
call_1_10_15_4, \
call_1_10_15_5, \
call_1_10_15_6, \
call_1_10_15_7, \
call_1_10_15_8, \
call_1_10_15_9, \
call_1_10_15_10, \
call_1_10_15_11, \
call_1_10_15_12, \
call_1_10_15_13, \
call_1_10_15_14, \
call_1_10_15_15, \
call_1_11_0_0, \
call_1_11_0_1, \
call_1_11_0_2, \
call_1_11_0_3, \
call_1_11_0_4, \
call_1_11_0_5, \
call_1_11_0_6, \
call_1_11_0_7, \
call_1_11_0_8, \
call_1_11_0_9, \
call_1_11_0_10, \
call_1_11_0_11, \
call_1_11_0_12, \
call_1_11_0_13, \
call_1_11_0_14, \
call_1_11_0_15, \
call_1_11_1_0, \
call_1_11_1_1, \
call_1_11_1_2, \
call_1_11_1_3, \
call_1_11_1_4, \
call_1_11_1_5, \
call_1_11_1_6, \
call_1_11_1_7, \
call_1_11_1_8, \
call_1_11_1_9, \
call_1_11_1_10, \
call_1_11_1_11, \
call_1_11_1_12, \
call_1_11_1_13, \
call_1_11_1_14, \
call_1_11_1_15, \
call_1_11_2_0, \
call_1_11_2_1, \
call_1_11_2_2, \
call_1_11_2_3, \
call_1_11_2_4, \
call_1_11_2_5, \
call_1_11_2_6, \
call_1_11_2_7, \
call_1_11_2_8, \
call_1_11_2_9, \
call_1_11_2_10, \
call_1_11_2_11, \
call_1_11_2_12, \
call_1_11_2_13, \
call_1_11_2_14, \
call_1_11_2_15, \
call_1_11_3_0, \
call_1_11_3_1, \
call_1_11_3_2, \
call_1_11_3_3, \
call_1_11_3_4, \
call_1_11_3_5, \
call_1_11_3_6, \
call_1_11_3_7, \
call_1_11_3_8, \
call_1_11_3_9, \
call_1_11_3_10, \
call_1_11_3_11, \
call_1_11_3_12, \
call_1_11_3_13, \
call_1_11_3_14, \
call_1_11_3_15, \
call_1_11_4_0, \
call_1_11_4_1, \
call_1_11_4_2, \
call_1_11_4_3, \
call_1_11_4_4, \
call_1_11_4_5, \
call_1_11_4_6, \
call_1_11_4_7, \
call_1_11_4_8, \
call_1_11_4_9, \
call_1_11_4_10, \
call_1_11_4_11, \
call_1_11_4_12, \
call_1_11_4_13, \
call_1_11_4_14, \
call_1_11_4_15, \
call_1_11_5_0, \
call_1_11_5_1, \
call_1_11_5_2, \
call_1_11_5_3, \
call_1_11_5_4, \
call_1_11_5_5, \
call_1_11_5_6, \
call_1_11_5_7, \
call_1_11_5_8, \
call_1_11_5_9, \
call_1_11_5_10, \
call_1_11_5_11, \
call_1_11_5_12, \
call_1_11_5_13, \
call_1_11_5_14, \
call_1_11_5_15, \
call_1_11_6_0, \
call_1_11_6_1, \
call_1_11_6_2, \
call_1_11_6_3, \
call_1_11_6_4, \
call_1_11_6_5, \
call_1_11_6_6, \
call_1_11_6_7, \
call_1_11_6_8, \
call_1_11_6_9, \
call_1_11_6_10, \
call_1_11_6_11, \
call_1_11_6_12, \
call_1_11_6_13, \
call_1_11_6_14, \
call_1_11_6_15, \
call_1_11_7_0, \
call_1_11_7_1, \
call_1_11_7_2, \
call_1_11_7_3, \
call_1_11_7_4, \
call_1_11_7_5, \
call_1_11_7_6, \
call_1_11_7_7, \
call_1_11_7_8, \
call_1_11_7_9, \
call_1_11_7_10, \
call_1_11_7_11, \
call_1_11_7_12, \
call_1_11_7_13, \
call_1_11_7_14, \
call_1_11_7_15, \
call_1_11_8_0, \
call_1_11_8_1, \
call_1_11_8_2, \
call_1_11_8_3, \
call_1_11_8_4, \
call_1_11_8_5, \
call_1_11_8_6, \
call_1_11_8_7, \
call_1_11_8_8, \
call_1_11_8_9, \
call_1_11_8_10, \
call_1_11_8_11, \
call_1_11_8_12, \
call_1_11_8_13, \
call_1_11_8_14, \
call_1_11_8_15, \
call_1_11_9_0, \
call_1_11_9_1, \
call_1_11_9_2, \
call_1_11_9_3, \
call_1_11_9_4, \
call_1_11_9_5, \
call_1_11_9_6, \
call_1_11_9_7, \
call_1_11_9_8, \
call_1_11_9_9, \
call_1_11_9_10, \
call_1_11_9_11, \
call_1_11_9_12, \
call_1_11_9_13, \
call_1_11_9_14, \
call_1_11_9_15, \
call_1_11_10_0, \
call_1_11_10_1, \
call_1_11_10_2, \
call_1_11_10_3, \
call_1_11_10_4, \
call_1_11_10_5, \
call_1_11_10_6, \
call_1_11_10_7, \
call_1_11_10_8, \
call_1_11_10_9, \
call_1_11_10_10, \
call_1_11_10_11, \
call_1_11_10_12, \
call_1_11_10_13, \
call_1_11_10_14, \
call_1_11_10_15, \
call_1_11_11_0, \
call_1_11_11_1, \
call_1_11_11_2, \
call_1_11_11_3, \
call_1_11_11_4, \
call_1_11_11_5, \
call_1_11_11_6, \
call_1_11_11_7, \
call_1_11_11_8, \
call_1_11_11_9, \
call_1_11_11_10, \
call_1_11_11_11, \
call_1_11_11_12, \
call_1_11_11_13, \
call_1_11_11_14, \
call_1_11_11_15, \
call_1_11_12_0, \
call_1_11_12_1, \
call_1_11_12_2, \
call_1_11_12_3, \
call_1_11_12_4, \
call_1_11_12_5, \
call_1_11_12_6, \
call_1_11_12_7, \
call_1_11_12_8, \
call_1_11_12_9, \
call_1_11_12_10, \
call_1_11_12_11, \
call_1_11_12_12, \
call_1_11_12_13, \
call_1_11_12_14, \
call_1_11_12_15, \
call_1_11_13_0, \
call_1_11_13_1, \
call_1_11_13_2, \
call_1_11_13_3, \
call_1_11_13_4, \
call_1_11_13_5, \
call_1_11_13_6, \
call_1_11_13_7, \
call_1_11_13_8, \
call_1_11_13_9, \
call_1_11_13_10, \
call_1_11_13_11, \
call_1_11_13_12, \
call_1_11_13_13, \
call_1_11_13_14, \
call_1_11_13_15, \
call_1_11_14_0, \
call_1_11_14_1, \
call_1_11_14_2, \
call_1_11_14_3, \
call_1_11_14_4, \
call_1_11_14_5, \
call_1_11_14_6, \
call_1_11_14_7, \
call_1_11_14_8, \
call_1_11_14_9, \
call_1_11_14_10, \
call_1_11_14_11, \
call_1_11_14_12, \
call_1_11_14_13, \
call_1_11_14_14, \
call_1_11_14_15, \
call_1_11_15_0, \
call_1_11_15_1, \
call_1_11_15_2, \
call_1_11_15_3, \
call_1_11_15_4, \
call_1_11_15_5, \
call_1_11_15_6, \
call_1_11_15_7, \
call_1_11_15_8, \
call_1_11_15_9, \
call_1_11_15_10, \
call_1_11_15_11, \
call_1_11_15_12, \
call_1_11_15_13, \
call_1_11_15_14, \
call_1_11_15_15, \
call_1_12_0_0, \
call_1_12_0_1, \
call_1_12_0_2, \
call_1_12_0_3, \
call_1_12_0_4, \
call_1_12_0_5, \
call_1_12_0_6, \
call_1_12_0_7, \
call_1_12_0_8, \
call_1_12_0_9, \
call_1_12_0_10, \
call_1_12_0_11, \
call_1_12_0_12, \
call_1_12_0_13, \
call_1_12_0_14, \
call_1_12_0_15, \
call_1_12_1_0, \
call_1_12_1_1, \
call_1_12_1_2, \
call_1_12_1_3, \
call_1_12_1_4, \
call_1_12_1_5, \
call_1_12_1_6, \
call_1_12_1_7, \
call_1_12_1_8, \
call_1_12_1_9, \
call_1_12_1_10, \
call_1_12_1_11, \
call_1_12_1_12, \
call_1_12_1_13, \
call_1_12_1_14, \
call_1_12_1_15, \
call_1_12_2_0, \
call_1_12_2_1, \
call_1_12_2_2, \
call_1_12_2_3, \
call_1_12_2_4, \
call_1_12_2_5, \
call_1_12_2_6, \
call_1_12_2_7, \
call_1_12_2_8, \
call_1_12_2_9, \
call_1_12_2_10, \
call_1_12_2_11, \
call_1_12_2_12, \
call_1_12_2_13, \
call_1_12_2_14, \
call_1_12_2_15, \
call_1_12_3_0, \
call_1_12_3_1, \
call_1_12_3_2, \
call_1_12_3_3, \
call_1_12_3_4, \
call_1_12_3_5, \
call_1_12_3_6, \
call_1_12_3_7, \
call_1_12_3_8, \
call_1_12_3_9, \
call_1_12_3_10, \
call_1_12_3_11, \
call_1_12_3_12, \
call_1_12_3_13, \
call_1_12_3_14, \
call_1_12_3_15, \
call_1_12_4_0, \
call_1_12_4_1, \
call_1_12_4_2, \
call_1_12_4_3, \
call_1_12_4_4, \
call_1_12_4_5, \
call_1_12_4_6, \
call_1_12_4_7, \
call_1_12_4_8, \
call_1_12_4_9, \
call_1_12_4_10, \
call_1_12_4_11, \
call_1_12_4_12, \
call_1_12_4_13, \
call_1_12_4_14, \
call_1_12_4_15, \
call_1_12_5_0, \
call_1_12_5_1, \
call_1_12_5_2, \
call_1_12_5_3, \
call_1_12_5_4, \
call_1_12_5_5, \
call_1_12_5_6, \
call_1_12_5_7, \
call_1_12_5_8, \
call_1_12_5_9, \
call_1_12_5_10, \
call_1_12_5_11, \
call_1_12_5_12, \
call_1_12_5_13, \
call_1_12_5_14, \
call_1_12_5_15, \
call_1_12_6_0, \
call_1_12_6_1, \
call_1_12_6_2, \
call_1_12_6_3, \
call_1_12_6_4, \
call_1_12_6_5, \
call_1_12_6_6, \
call_1_12_6_7, \
call_1_12_6_8, \
call_1_12_6_9, \
call_1_12_6_10, \
call_1_12_6_11, \
call_1_12_6_12, \
call_1_12_6_13, \
call_1_12_6_14, \
call_1_12_6_15, \
call_1_12_7_0, \
call_1_12_7_1, \
call_1_12_7_2, \
call_1_12_7_3, \
call_1_12_7_4, \
call_1_12_7_5, \
call_1_12_7_6, \
call_1_12_7_7, \
call_1_12_7_8, \
call_1_12_7_9, \
call_1_12_7_10, \
call_1_12_7_11, \
call_1_12_7_12, \
call_1_12_7_13, \
call_1_12_7_14, \
call_1_12_7_15, \
call_1_12_8_0, \
call_1_12_8_1, \
call_1_12_8_2, \
call_1_12_8_3, \
call_1_12_8_4, \
call_1_12_8_5, \
call_1_12_8_6, \
call_1_12_8_7, \
call_1_12_8_8, \
call_1_12_8_9, \
call_1_12_8_10, \
call_1_12_8_11, \
call_1_12_8_12, \
call_1_12_8_13, \
call_1_12_8_14, \
call_1_12_8_15, \
call_1_12_9_0, \
call_1_12_9_1, \
call_1_12_9_2, \
call_1_12_9_3, \
call_1_12_9_4, \
call_1_12_9_5, \
call_1_12_9_6, \
call_1_12_9_7, \
call_1_12_9_8, \
call_1_12_9_9, \
call_1_12_9_10, \
call_1_12_9_11, \
call_1_12_9_12, \
call_1_12_9_13, \
call_1_12_9_14, \
call_1_12_9_15, \
call_1_12_10_0, \
call_1_12_10_1, \
call_1_12_10_2, \
call_1_12_10_3, \
call_1_12_10_4, \
call_1_12_10_5, \
call_1_12_10_6, \
call_1_12_10_7, \
call_1_12_10_8, \
call_1_12_10_9, \
call_1_12_10_10, \
call_1_12_10_11, \
call_1_12_10_12, \
call_1_12_10_13, \
call_1_12_10_14, \
call_1_12_10_15, \
call_1_12_11_0, \
call_1_12_11_1, \
call_1_12_11_2, \
call_1_12_11_3, \
call_1_12_11_4, \
call_1_12_11_5, \
call_1_12_11_6, \
call_1_12_11_7, \
call_1_12_11_8, \
call_1_12_11_9, \
call_1_12_11_10, \
call_1_12_11_11, \
call_1_12_11_12, \
call_1_12_11_13, \
call_1_12_11_14, \
call_1_12_11_15, \
call_1_12_12_0, \
call_1_12_12_1, \
call_1_12_12_2, \
call_1_12_12_3, \
call_1_12_12_4, \
call_1_12_12_5, \
call_1_12_12_6, \
call_1_12_12_7, \
call_1_12_12_8, \
call_1_12_12_9, \
call_1_12_12_10, \
call_1_12_12_11, \
call_1_12_12_12, \
call_1_12_12_13, \
call_1_12_12_14, \
call_1_12_12_15, \
call_1_12_13_0, \
call_1_12_13_1, \
call_1_12_13_2, \
call_1_12_13_3, \
call_1_12_13_4, \
call_1_12_13_5, \
call_1_12_13_6, \
call_1_12_13_7, \
call_1_12_13_8, \
call_1_12_13_9, \
call_1_12_13_10, \
call_1_12_13_11, \
call_1_12_13_12, \
call_1_12_13_13, \
call_1_12_13_14, \
call_1_12_13_15, \
call_1_12_14_0, \
call_1_12_14_1, \
call_1_12_14_2, \
call_1_12_14_3, \
call_1_12_14_4, \
call_1_12_14_5, \
call_1_12_14_6, \
call_1_12_14_7, \
call_1_12_14_8, \
call_1_12_14_9, \
call_1_12_14_10, \
call_1_12_14_11, \
call_1_12_14_12, \
call_1_12_14_13, \
call_1_12_14_14, \
call_1_12_14_15, \
call_1_12_15_0, \
call_1_12_15_1, \
call_1_12_15_2, \
call_1_12_15_3, \
call_1_12_15_4, \
call_1_12_15_5, \
call_1_12_15_6, \
call_1_12_15_7, \
call_1_12_15_8, \
call_1_12_15_9, \
call_1_12_15_10, \
call_1_12_15_11, \
call_1_12_15_12, \
call_1_12_15_13, \
call_1_12_15_14, \
call_1_12_15_15, \
call_1_13_0_0, \
call_1_13_0_1, \
call_1_13_0_2, \
call_1_13_0_3, \
call_1_13_0_4, \
call_1_13_0_5, \
call_1_13_0_6, \
call_1_13_0_7, \
call_1_13_0_8, \
call_1_13_0_9, \
call_1_13_0_10, \
call_1_13_0_11, \
call_1_13_0_12, \
call_1_13_0_13, \
call_1_13_0_14, \
call_1_13_0_15, \
call_1_13_1_0, \
call_1_13_1_1, \
call_1_13_1_2, \
call_1_13_1_3, \
call_1_13_1_4, \
call_1_13_1_5, \
call_1_13_1_6, \
call_1_13_1_7, \
call_1_13_1_8, \
call_1_13_1_9, \
call_1_13_1_10, \
call_1_13_1_11, \
call_1_13_1_12, \
call_1_13_1_13, \
call_1_13_1_14, \
call_1_13_1_15, \
call_1_13_2_0, \
call_1_13_2_1, \
call_1_13_2_2, \
call_1_13_2_3, \
call_1_13_2_4, \
call_1_13_2_5, \
call_1_13_2_6, \
call_1_13_2_7, \
call_1_13_2_8, \
call_1_13_2_9, \
call_1_13_2_10, \
call_1_13_2_11, \
call_1_13_2_12, \
call_1_13_2_13, \
call_1_13_2_14, \
call_1_13_2_15, \
call_1_13_3_0, \
call_1_13_3_1, \
call_1_13_3_2, \
call_1_13_3_3, \
call_1_13_3_4, \
call_1_13_3_5, \
call_1_13_3_6, \
call_1_13_3_7, \
call_1_13_3_8, \
call_1_13_3_9, \
call_1_13_3_10, \
call_1_13_3_11, \
call_1_13_3_12, \
call_1_13_3_13, \
call_1_13_3_14, \
call_1_13_3_15, \
call_1_13_4_0, \
call_1_13_4_1, \
call_1_13_4_2, \
call_1_13_4_3, \
call_1_13_4_4, \
call_1_13_4_5, \
call_1_13_4_6, \
call_1_13_4_7, \
call_1_13_4_8, \
call_1_13_4_9, \
call_1_13_4_10, \
call_1_13_4_11, \
call_1_13_4_12, \
call_1_13_4_13, \
call_1_13_4_14, \
call_1_13_4_15, \
call_1_13_5_0, \
call_1_13_5_1, \
call_1_13_5_2, \
call_1_13_5_3, \
call_1_13_5_4, \
call_1_13_5_5, \
call_1_13_5_6, \
call_1_13_5_7, \
call_1_13_5_8, \
call_1_13_5_9, \
call_1_13_5_10, \
call_1_13_5_11, \
call_1_13_5_12, \
call_1_13_5_13, \
call_1_13_5_14, \
call_1_13_5_15, \
call_1_13_6_0, \
call_1_13_6_1, \
call_1_13_6_2, \
call_1_13_6_3, \
call_1_13_6_4, \
call_1_13_6_5, \
call_1_13_6_6, \
call_1_13_6_7, \
call_1_13_6_8, \
call_1_13_6_9, \
call_1_13_6_10, \
call_1_13_6_11, \
call_1_13_6_12, \
call_1_13_6_13, \
call_1_13_6_14, \
call_1_13_6_15, \
call_1_13_7_0, \
call_1_13_7_1, \
call_1_13_7_2, \
call_1_13_7_3, \
call_1_13_7_4, \
call_1_13_7_5, \
call_1_13_7_6, \
call_1_13_7_7, \
call_1_13_7_8, \
call_1_13_7_9, \
call_1_13_7_10, \
call_1_13_7_11, \
call_1_13_7_12, \
call_1_13_7_13, \
call_1_13_7_14, \
call_1_13_7_15, \
call_1_13_8_0, \
call_1_13_8_1, \
call_1_13_8_2, \
call_1_13_8_3, \
call_1_13_8_4, \
call_1_13_8_5, \
call_1_13_8_6, \
call_1_13_8_7, \
call_1_13_8_8, \
call_1_13_8_9, \
call_1_13_8_10, \
call_1_13_8_11, \
call_1_13_8_12, \
call_1_13_8_13, \
call_1_13_8_14, \
call_1_13_8_15, \
call_1_13_9_0, \
call_1_13_9_1, \
call_1_13_9_2, \
call_1_13_9_3, \
call_1_13_9_4, \
call_1_13_9_5, \
call_1_13_9_6, \
call_1_13_9_7, \
call_1_13_9_8, \
call_1_13_9_9, \
call_1_13_9_10, \
call_1_13_9_11, \
call_1_13_9_12, \
call_1_13_9_13, \
call_1_13_9_14, \
call_1_13_9_15, \
call_1_13_10_0, \
call_1_13_10_1, \
call_1_13_10_2, \
call_1_13_10_3, \
call_1_13_10_4, \
call_1_13_10_5, \
call_1_13_10_6, \
call_1_13_10_7, \
call_1_13_10_8, \
call_1_13_10_9, \
call_1_13_10_10, \
call_1_13_10_11, \
call_1_13_10_12, \
call_1_13_10_13, \
call_1_13_10_14, \
call_1_13_10_15, \
call_1_13_11_0, \
call_1_13_11_1, \
call_1_13_11_2, \
call_1_13_11_3, \
call_1_13_11_4, \
call_1_13_11_5, \
call_1_13_11_6, \
call_1_13_11_7, \
call_1_13_11_8, \
call_1_13_11_9, \
call_1_13_11_10, \
call_1_13_11_11, \
call_1_13_11_12, \
call_1_13_11_13, \
call_1_13_11_14, \
call_1_13_11_15, \
call_1_13_12_0, \
call_1_13_12_1, \
call_1_13_12_2, \
call_1_13_12_3, \
call_1_13_12_4, \
call_1_13_12_5, \
call_1_13_12_6, \
call_1_13_12_7, \
call_1_13_12_8, \
call_1_13_12_9, \
call_1_13_12_10, \
call_1_13_12_11, \
call_1_13_12_12, \
call_1_13_12_13, \
call_1_13_12_14, \
call_1_13_12_15, \
call_1_13_13_0, \
call_1_13_13_1, \
call_1_13_13_2, \
call_1_13_13_3, \
call_1_13_13_4, \
call_1_13_13_5, \
call_1_13_13_6, \
call_1_13_13_7, \
call_1_13_13_8, \
call_1_13_13_9, \
call_1_13_13_10, \
call_1_13_13_11, \
call_1_13_13_12, \
call_1_13_13_13, \
call_1_13_13_14, \
call_1_13_13_15, \
call_1_13_14_0, \
call_1_13_14_1, \
call_1_13_14_2, \
call_1_13_14_3, \
call_1_13_14_4, \
call_1_13_14_5, \
call_1_13_14_6, \
call_1_13_14_7, \
call_1_13_14_8, \
call_1_13_14_9, \
call_1_13_14_10, \
call_1_13_14_11, \
call_1_13_14_12, \
call_1_13_14_13, \
call_1_13_14_14, \
call_1_13_14_15, \
call_1_13_15_0, \
call_1_13_15_1, \
call_1_13_15_2, \
call_1_13_15_3, \
call_1_13_15_4, \
call_1_13_15_5, \
call_1_13_15_6, \
call_1_13_15_7, \
call_1_13_15_8, \
call_1_13_15_9, \
call_1_13_15_10, \
call_1_13_15_11, \
call_1_13_15_12, \
call_1_13_15_13, \
call_1_13_15_14, \
call_1_13_15_15, \
call_1_14_0_0, \
call_1_14_0_1, \
call_1_14_0_2, \
call_1_14_0_3, \
call_1_14_0_4, \
call_1_14_0_5, \
call_1_14_0_6, \
call_1_14_0_7, \
call_1_14_0_8, \
call_1_14_0_9, \
call_1_14_0_10, \
call_1_14_0_11, \
call_1_14_0_12, \
call_1_14_0_13, \
call_1_14_0_14, \
call_1_14_0_15, \
call_1_14_1_0, \
call_1_14_1_1, \
call_1_14_1_2, \
call_1_14_1_3, \
call_1_14_1_4, \
call_1_14_1_5, \
call_1_14_1_6, \
call_1_14_1_7, \
call_1_14_1_8, \
call_1_14_1_9, \
call_1_14_1_10, \
call_1_14_1_11, \
call_1_14_1_12, \
call_1_14_1_13, \
call_1_14_1_14, \
call_1_14_1_15, \
call_1_14_2_0, \
call_1_14_2_1, \
call_1_14_2_2, \
call_1_14_2_3, \
call_1_14_2_4, \
call_1_14_2_5, \
call_1_14_2_6, \
call_1_14_2_7, \
call_1_14_2_8, \
call_1_14_2_9, \
call_1_14_2_10, \
call_1_14_2_11, \
call_1_14_2_12, \
call_1_14_2_13, \
call_1_14_2_14, \
call_1_14_2_15, \
call_1_14_3_0, \
call_1_14_3_1, \
call_1_14_3_2, \
call_1_14_3_3, \
call_1_14_3_4, \
call_1_14_3_5, \
call_1_14_3_6, \
call_1_14_3_7, \
call_1_14_3_8, \
call_1_14_3_9, \
call_1_14_3_10, \
call_1_14_3_11, \
call_1_14_3_12, \
call_1_14_3_13, \
call_1_14_3_14, \
call_1_14_3_15, \
call_1_14_4_0, \
call_1_14_4_1, \
call_1_14_4_2, \
call_1_14_4_3, \
call_1_14_4_4, \
call_1_14_4_5, \
call_1_14_4_6, \
call_1_14_4_7, \
call_1_14_4_8, \
call_1_14_4_9, \
call_1_14_4_10, \
call_1_14_4_11, \
call_1_14_4_12, \
call_1_14_4_13, \
call_1_14_4_14, \
call_1_14_4_15, \
call_1_14_5_0, \
call_1_14_5_1, \
call_1_14_5_2, \
call_1_14_5_3, \
call_1_14_5_4, \
call_1_14_5_5, \
call_1_14_5_6, \
call_1_14_5_7, \
call_1_14_5_8, \
call_1_14_5_9, \
call_1_14_5_10, \
call_1_14_5_11, \
call_1_14_5_12, \
call_1_14_5_13, \
call_1_14_5_14, \
call_1_14_5_15, \
call_1_14_6_0, \
call_1_14_6_1, \
call_1_14_6_2, \
call_1_14_6_3, \
call_1_14_6_4, \
call_1_14_6_5, \
call_1_14_6_6, \
call_1_14_6_7, \
call_1_14_6_8, \
call_1_14_6_9, \
call_1_14_6_10, \
call_1_14_6_11, \
call_1_14_6_12, \
call_1_14_6_13, \
call_1_14_6_14, \
call_1_14_6_15, \
call_1_14_7_0, \
call_1_14_7_1, \
call_1_14_7_2, \
call_1_14_7_3, \
call_1_14_7_4, \
call_1_14_7_5, \
call_1_14_7_6, \
call_1_14_7_7, \
call_1_14_7_8, \
call_1_14_7_9, \
call_1_14_7_10, \
call_1_14_7_11, \
call_1_14_7_12, \
call_1_14_7_13, \
call_1_14_7_14, \
call_1_14_7_15, \
call_1_14_8_0, \
call_1_14_8_1, \
call_1_14_8_2, \
call_1_14_8_3, \
call_1_14_8_4, \
call_1_14_8_5, \
call_1_14_8_6, \
call_1_14_8_7, \
call_1_14_8_8, \
call_1_14_8_9, \
call_1_14_8_10, \
call_1_14_8_11, \
call_1_14_8_12, \
call_1_14_8_13, \
call_1_14_8_14, \
call_1_14_8_15, \
call_1_14_9_0, \
call_1_14_9_1, \
call_1_14_9_2, \
call_1_14_9_3, \
call_1_14_9_4, \
call_1_14_9_5, \
call_1_14_9_6, \
call_1_14_9_7, \
call_1_14_9_8, \
call_1_14_9_9, \
call_1_14_9_10, \
call_1_14_9_11, \
call_1_14_9_12, \
call_1_14_9_13, \
call_1_14_9_14, \
call_1_14_9_15, \
call_1_14_10_0, \
call_1_14_10_1, \
call_1_14_10_2, \
call_1_14_10_3, \
call_1_14_10_4, \
call_1_14_10_5, \
call_1_14_10_6, \
call_1_14_10_7, \
call_1_14_10_8, \
call_1_14_10_9, \
call_1_14_10_10, \
call_1_14_10_11, \
call_1_14_10_12, \
call_1_14_10_13, \
call_1_14_10_14, \
call_1_14_10_15, \
call_1_14_11_0, \
call_1_14_11_1, \
call_1_14_11_2, \
call_1_14_11_3, \
call_1_14_11_4, \
call_1_14_11_5, \
call_1_14_11_6, \
call_1_14_11_7, \
call_1_14_11_8, \
call_1_14_11_9, \
call_1_14_11_10, \
call_1_14_11_11, \
call_1_14_11_12, \
call_1_14_11_13, \
call_1_14_11_14, \
call_1_14_11_15, \
call_1_14_12_0, \
call_1_14_12_1, \
call_1_14_12_2, \
call_1_14_12_3, \
call_1_14_12_4, \
call_1_14_12_5, \
call_1_14_12_6, \
call_1_14_12_7, \
call_1_14_12_8, \
call_1_14_12_9, \
call_1_14_12_10, \
call_1_14_12_11, \
call_1_14_12_12, \
call_1_14_12_13, \
call_1_14_12_14, \
call_1_14_12_15, \
call_1_14_13_0, \
call_1_14_13_1, \
call_1_14_13_2, \
call_1_14_13_3, \
call_1_14_13_4, \
call_1_14_13_5, \
call_1_14_13_6, \
call_1_14_13_7, \
call_1_14_13_8, \
call_1_14_13_9, \
call_1_14_13_10, \
call_1_14_13_11, \
call_1_14_13_12, \
call_1_14_13_13, \
call_1_14_13_14, \
call_1_14_13_15, \
call_1_14_14_0, \
call_1_14_14_1, \
call_1_14_14_2, \
call_1_14_14_3, \
call_1_14_14_4, \
call_1_14_14_5, \
call_1_14_14_6, \
call_1_14_14_7, \
call_1_14_14_8, \
call_1_14_14_9, \
call_1_14_14_10, \
call_1_14_14_11, \
call_1_14_14_12, \
call_1_14_14_13, \
call_1_14_14_14, \
call_1_14_14_15, \
call_1_14_15_0, \
call_1_14_15_1, \
call_1_14_15_2, \
call_1_14_15_3, \
call_1_14_15_4, \
call_1_14_15_5, \
call_1_14_15_6, \
call_1_14_15_7, \
call_1_14_15_8, \
call_1_14_15_9, \
call_1_14_15_10, \
call_1_14_15_11, \
call_1_14_15_12, \
call_1_14_15_13, \
call_1_14_15_14, \
call_1_14_15_15, \
call_1_15_0_0, \
call_1_15_0_1, \
call_1_15_0_2, \
call_1_15_0_3, \
call_1_15_0_4, \
call_1_15_0_5, \
call_1_15_0_6, \
call_1_15_0_7, \
call_1_15_0_8, \
call_1_15_0_9, \
call_1_15_0_10, \
call_1_15_0_11, \
call_1_15_0_12, \
call_1_15_0_13, \
call_1_15_0_14, \
call_1_15_0_15, \
call_1_15_1_0, \
call_1_15_1_1, \
call_1_15_1_2, \
call_1_15_1_3, \
call_1_15_1_4, \
call_1_15_1_5, \
call_1_15_1_6, \
call_1_15_1_7, \
call_1_15_1_8, \
call_1_15_1_9, \
call_1_15_1_10, \
call_1_15_1_11, \
call_1_15_1_12, \
call_1_15_1_13, \
call_1_15_1_14, \
call_1_15_1_15, \
call_1_15_2_0, \
call_1_15_2_1, \
call_1_15_2_2, \
call_1_15_2_3, \
call_1_15_2_4, \
call_1_15_2_5, \
call_1_15_2_6, \
call_1_15_2_7, \
call_1_15_2_8, \
call_1_15_2_9, \
call_1_15_2_10, \
call_1_15_2_11, \
call_1_15_2_12, \
call_1_15_2_13, \
call_1_15_2_14, \
call_1_15_2_15, \
call_1_15_3_0, \
call_1_15_3_1, \
call_1_15_3_2, \
call_1_15_3_3, \
call_1_15_3_4, \
call_1_15_3_5, \
call_1_15_3_6, \
call_1_15_3_7, \
call_1_15_3_8, \
call_1_15_3_9, \
call_1_15_3_10, \
call_1_15_3_11, \
call_1_15_3_12, \
call_1_15_3_13, \
call_1_15_3_14, \
call_1_15_3_15, \
call_1_15_4_0, \
call_1_15_4_1, \
call_1_15_4_2, \
call_1_15_4_3, \
call_1_15_4_4, \
call_1_15_4_5, \
call_1_15_4_6, \
call_1_15_4_7, \
call_1_15_4_8, \
call_1_15_4_9, \
call_1_15_4_10, \
call_1_15_4_11, \
call_1_15_4_12, \
call_1_15_4_13, \
call_1_15_4_14, \
call_1_15_4_15, \
call_1_15_5_0, \
call_1_15_5_1, \
call_1_15_5_2, \
call_1_15_5_3, \
call_1_15_5_4, \
call_1_15_5_5, \
call_1_15_5_6, \
call_1_15_5_7, \
call_1_15_5_8, \
call_1_15_5_9, \
call_1_15_5_10, \
call_1_15_5_11, \
call_1_15_5_12, \
call_1_15_5_13, \
call_1_15_5_14, \
call_1_15_5_15, \
call_1_15_6_0, \
call_1_15_6_1, \
call_1_15_6_2, \
call_1_15_6_3, \
call_1_15_6_4, \
call_1_15_6_5, \
call_1_15_6_6, \
call_1_15_6_7, \
call_1_15_6_8, \
call_1_15_6_9, \
call_1_15_6_10, \
call_1_15_6_11, \
call_1_15_6_12, \
call_1_15_6_13, \
call_1_15_6_14, \
call_1_15_6_15, \
call_1_15_7_0, \
call_1_15_7_1, \
call_1_15_7_2, \
call_1_15_7_3, \
call_1_15_7_4, \
call_1_15_7_5, \
call_1_15_7_6, \
call_1_15_7_7, \
call_1_15_7_8, \
call_1_15_7_9, \
call_1_15_7_10, \
call_1_15_7_11, \
call_1_15_7_12, \
call_1_15_7_13, \
call_1_15_7_14, \
call_1_15_7_15, \
call_1_15_8_0, \
call_1_15_8_1, \
call_1_15_8_2, \
call_1_15_8_3, \
call_1_15_8_4, \
call_1_15_8_5, \
call_1_15_8_6, \
call_1_15_8_7, \
call_1_15_8_8, \
call_1_15_8_9, \
call_1_15_8_10, \
call_1_15_8_11, \
call_1_15_8_12, \
call_1_15_8_13, \
call_1_15_8_14, \
call_1_15_8_15, \
call_1_15_9_0, \
call_1_15_9_1, \
call_1_15_9_2, \
call_1_15_9_3, \
call_1_15_9_4, \
call_1_15_9_5, \
call_1_15_9_6, \
call_1_15_9_7, \
call_1_15_9_8, \
call_1_15_9_9, \
call_1_15_9_10, \
call_1_15_9_11, \
call_1_15_9_12, \
call_1_15_9_13, \
call_1_15_9_14, \
call_1_15_9_15, \
call_1_15_10_0, \
call_1_15_10_1, \
call_1_15_10_2, \
call_1_15_10_3, \
call_1_15_10_4, \
call_1_15_10_5, \
call_1_15_10_6, \
call_1_15_10_7, \
call_1_15_10_8, \
call_1_15_10_9, \
call_1_15_10_10, \
call_1_15_10_11, \
call_1_15_10_12, \
call_1_15_10_13, \
call_1_15_10_14, \
call_1_15_10_15, \
call_1_15_11_0, \
call_1_15_11_1, \
call_1_15_11_2, \
call_1_15_11_3, \
call_1_15_11_4, \
call_1_15_11_5, \
call_1_15_11_6, \
call_1_15_11_7, \
call_1_15_11_8, \
call_1_15_11_9, \
call_1_15_11_10, \
call_1_15_11_11, \
call_1_15_11_12, \
call_1_15_11_13, \
call_1_15_11_14, \
call_1_15_11_15, \
call_1_15_12_0, \
call_1_15_12_1, \
call_1_15_12_2, \
call_1_15_12_3, \
call_1_15_12_4, \
call_1_15_12_5, \
call_1_15_12_6, \
call_1_15_12_7, \
call_1_15_12_8, \
call_1_15_12_9, \
call_1_15_12_10, \
call_1_15_12_11, \
call_1_15_12_12, \
call_1_15_12_13, \
call_1_15_12_14, \
call_1_15_12_15, \
call_1_15_13_0, \
call_1_15_13_1, \
call_1_15_13_2, \
call_1_15_13_3, \
call_1_15_13_4, \
call_1_15_13_5, \
call_1_15_13_6, \
call_1_15_13_7, \
call_1_15_13_8, \
call_1_15_13_9, \
call_1_15_13_10, \
call_1_15_13_11, \
call_1_15_13_12, \
call_1_15_13_13, \
call_1_15_13_14, \
call_1_15_13_15, \
call_1_15_14_0, \
call_1_15_14_1, \
call_1_15_14_2, \
call_1_15_14_3, \
call_1_15_14_4, \
call_1_15_14_5, \
call_1_15_14_6, \
call_1_15_14_7, \
call_1_15_14_8, \
call_1_15_14_9, \
call_1_15_14_10, \
call_1_15_14_11, \
call_1_15_14_12, \
call_1_15_14_13, \
call_1_15_14_14, \
call_1_15_14_15, \
call_1_15_15_0, \
call_1_15_15_1, \
call_1_15_15_2, \
call_1_15_15_3, \
call_1_15_15_4, \
call_1_15_15_5, \
call_1_15_15_6, \
call_1_15_15_7, \
call_1_15_15_8, \
call_1_15_15_9, \
call_1_15_15_10, \
call_1_15_15_11, \
call_1_15_15_12, \
call_1_15_15_13, \
call_1_15_15_14, \
call_1_15_15_15, \
call_2_0_0_0, \
call_2_0_0_1, \
call_2_0_0_2, \
call_2_0_0_3, \
call_2_0_0_4, \
call_2_0_0_5, \
call_2_0_0_6, \
call_2_0_0_7, \
call_2_0_0_8, \
call_2_0_0_9, \
call_2_0_0_10, \
call_2_0_0_11, \
call_2_0_0_12, \
call_2_0_0_13, \
call_2_0_0_14, \
call_2_0_0_15, \
call_2_0_1_0, \
call_2_0_1_1, \
call_2_0_1_2, \
call_2_0_1_3, \
call_2_0_1_4, \
call_2_0_1_5, \
call_2_0_1_6, \
call_2_0_1_7, \
call_2_0_1_8, \
call_2_0_1_9, \
call_2_0_1_10, \
call_2_0_1_11, \
call_2_0_1_12, \
call_2_0_1_13, \
call_2_0_1_14, \
call_2_0_1_15, \
call_2_0_2_0, \
call_2_0_2_1, \
call_2_0_2_2, \
call_2_0_2_3, \
call_2_0_2_4, \
call_2_0_2_5, \
call_2_0_2_6, \
call_2_0_2_7, \
call_2_0_2_8, \
call_2_0_2_9, \
call_2_0_2_10, \
call_2_0_2_11, \
call_2_0_2_12, \
call_2_0_2_13, \
call_2_0_2_14, \
call_2_0_2_15, \
call_2_0_3_0, \
call_2_0_3_1, \
call_2_0_3_2, \
call_2_0_3_3, \
call_2_0_3_4, \
call_2_0_3_5, \
call_2_0_3_6, \
call_2_0_3_7, \
call_2_0_3_8, \
call_2_0_3_9, \
call_2_0_3_10, \
call_2_0_3_11, \
call_2_0_3_12, \
call_2_0_3_13, \
call_2_0_3_14, \
call_2_0_3_15, \
call_2_0_4_0, \
call_2_0_4_1, \
call_2_0_4_2, \
call_2_0_4_3, \
call_2_0_4_4, \
call_2_0_4_5, \
call_2_0_4_6, \
call_2_0_4_7, \
call_2_0_4_8, \
call_2_0_4_9, \
call_2_0_4_10, \
call_2_0_4_11, \
call_2_0_4_12, \
call_2_0_4_13, \
call_2_0_4_14, \
call_2_0_4_15, \
call_2_0_5_0, \
call_2_0_5_1, \
call_2_0_5_2, \
call_2_0_5_3, \
call_2_0_5_4, \
call_2_0_5_5, \
call_2_0_5_6, \
call_2_0_5_7, \
call_2_0_5_8, \
call_2_0_5_9, \
call_2_0_5_10, \
call_2_0_5_11, \
call_2_0_5_12, \
call_2_0_5_13, \
call_2_0_5_14, \
call_2_0_5_15, \
call_2_0_6_0, \
call_2_0_6_1, \
call_2_0_6_2, \
call_2_0_6_3, \
call_2_0_6_4, \
call_2_0_6_5, \
call_2_0_6_6, \
call_2_0_6_7, \
call_2_0_6_8, \
call_2_0_6_9, \
call_2_0_6_10, \
call_2_0_6_11, \
call_2_0_6_12, \
call_2_0_6_13, \
call_2_0_6_14, \
call_2_0_6_15, \
call_2_0_7_0, \
call_2_0_7_1, \
call_2_0_7_2, \
call_2_0_7_3, \
call_2_0_7_4, \
call_2_0_7_5, \
call_2_0_7_6, \
call_2_0_7_7, \
call_2_0_7_8, \
call_2_0_7_9, \
call_2_0_7_10, \
call_2_0_7_11, \
call_2_0_7_12, \
call_2_0_7_13, \
call_2_0_7_14, \
call_2_0_7_15, \
call_2_0_8_0, \
call_2_0_8_1, \
call_2_0_8_2, \
call_2_0_8_3, \
call_2_0_8_4, \
call_2_0_8_5, \
call_2_0_8_6, \
call_2_0_8_7, \
call_2_0_8_8, \
call_2_0_8_9, \
call_2_0_8_10, \
call_2_0_8_11, \
call_2_0_8_12, \
call_2_0_8_13, \
call_2_0_8_14, \
call_2_0_8_15, \
call_2_0_9_0, \
call_2_0_9_1, \
call_2_0_9_2, \
call_2_0_9_3, \
call_2_0_9_4, \
call_2_0_9_5, \
call_2_0_9_6, \
call_2_0_9_7, \
call_2_0_9_8, \
call_2_0_9_9, \
call_2_0_9_10, \
call_2_0_9_11, \
call_2_0_9_12, \
call_2_0_9_13, \
call_2_0_9_14, \
call_2_0_9_15, \
call_2_0_10_0, \
call_2_0_10_1, \
call_2_0_10_2, \
call_2_0_10_3, \
call_2_0_10_4, \
call_2_0_10_5, \
call_2_0_10_6, \
call_2_0_10_7, \
call_2_0_10_8, \
call_2_0_10_9, \
call_2_0_10_10, \
call_2_0_10_11, \
call_2_0_10_12, \
call_2_0_10_13, \
call_2_0_10_14, \
call_2_0_10_15, \
call_2_0_11_0, \
call_2_0_11_1, \
call_2_0_11_2, \
call_2_0_11_3, \
call_2_0_11_4, \
call_2_0_11_5, \
call_2_0_11_6, \
call_2_0_11_7, \
call_2_0_11_8, \
call_2_0_11_9, \
call_2_0_11_10, \
call_2_0_11_11, \
call_2_0_11_12, \
call_2_0_11_13, \
call_2_0_11_14, \
call_2_0_11_15, \
call_2_0_12_0, \
call_2_0_12_1, \
call_2_0_12_2, \
call_2_0_12_3, \
call_2_0_12_4, \
call_2_0_12_5, \
call_2_0_12_6, \
call_2_0_12_7, \
call_2_0_12_8, \
call_2_0_12_9, \
call_2_0_12_10, \
call_2_0_12_11, \
call_2_0_12_12, \
call_2_0_12_13, \
call_2_0_12_14, \
call_2_0_12_15, \
call_2_0_13_0, \
call_2_0_13_1, \
call_2_0_13_2, \
call_2_0_13_3, \
call_2_0_13_4, \
call_2_0_13_5, \
call_2_0_13_6, \
call_2_0_13_7, \
call_2_0_13_8, \
call_2_0_13_9, \
call_2_0_13_10, \
call_2_0_13_11, \
call_2_0_13_12, \
call_2_0_13_13, \
call_2_0_13_14, \
call_2_0_13_15, \
call_2_0_14_0, \
call_2_0_14_1, \
call_2_0_14_2, \
call_2_0_14_3, \
call_2_0_14_4, \
call_2_0_14_5, \
call_2_0_14_6, \
call_2_0_14_7, \
call_2_0_14_8, \
call_2_0_14_9, \
call_2_0_14_10, \
call_2_0_14_11, \
call_2_0_14_12, \
call_2_0_14_13, \
call_2_0_14_14, \
call_2_0_14_15, \
call_2_0_15_0, \
call_2_0_15_1, \
call_2_0_15_2, \
call_2_0_15_3, \
call_2_0_15_4, \
call_2_0_15_5, \
call_2_0_15_6, \
call_2_0_15_7, \
call_2_0_15_8, \
call_2_0_15_9, \
call_2_0_15_10, \
call_2_0_15_11, \
call_2_0_15_12, \
call_2_0_15_13, \
call_2_0_15_14, \
call_2_0_15_15, \
call_2_1_0_0, \
call_2_1_0_1, \
call_2_1_0_2, \
call_2_1_0_3, \
call_2_1_0_4, \
call_2_1_0_5, \
call_2_1_0_6, \
call_2_1_0_7, \
call_2_1_0_8, \
call_2_1_0_9, \
call_2_1_0_10, \
call_2_1_0_11, \
call_2_1_0_12, \
call_2_1_0_13, \
call_2_1_0_14, \
call_2_1_0_15, \
call_2_1_1_0, \
call_2_1_1_1, \
call_2_1_1_2, \
call_2_1_1_3, \
call_2_1_1_4, \
call_2_1_1_5, \
call_2_1_1_6, \
call_2_1_1_7, \
call_2_1_1_8, \
call_2_1_1_9, \
call_2_1_1_10, \
call_2_1_1_11, \
call_2_1_1_12, \
call_2_1_1_13, \
call_2_1_1_14, \
call_2_1_1_15, \
call_2_1_2_0, \
call_2_1_2_1, \
call_2_1_2_2, \
call_2_1_2_3, \
call_2_1_2_4, \
call_2_1_2_5, \
call_2_1_2_6, \
call_2_1_2_7, \
call_2_1_2_8, \
call_2_1_2_9, \
call_2_1_2_10, \
call_2_1_2_11, \
call_2_1_2_12, \
call_2_1_2_13, \
call_2_1_2_14, \
call_2_1_2_15, \
call_2_1_3_0, \
call_2_1_3_1, \
call_2_1_3_2, \
call_2_1_3_3, \
call_2_1_3_4, \
call_2_1_3_5, \
call_2_1_3_6, \
call_2_1_3_7, \
call_2_1_3_8, \
call_2_1_3_9, \
call_2_1_3_10, \
call_2_1_3_11, \
call_2_1_3_12, \
call_2_1_3_13, \
call_2_1_3_14, \
call_2_1_3_15, \
call_2_1_4_0, \
call_2_1_4_1, \
call_2_1_4_2, \
call_2_1_4_3, \
call_2_1_4_4, \
call_2_1_4_5, \
call_2_1_4_6, \
call_2_1_4_7, \
call_2_1_4_8, \
call_2_1_4_9, \
call_2_1_4_10, \
call_2_1_4_11, \
call_2_1_4_12, \
call_2_1_4_13, \
call_2_1_4_14, \
call_2_1_4_15, \
call_2_1_5_0, \
call_2_1_5_1, \
call_2_1_5_2, \
call_2_1_5_3, \
call_2_1_5_4, \
call_2_1_5_5, \
call_2_1_5_6, \
call_2_1_5_7, \
call_2_1_5_8, \
call_2_1_5_9, \
call_2_1_5_10, \
call_2_1_5_11, \
call_2_1_5_12, \
call_2_1_5_13, \
call_2_1_5_14, \
call_2_1_5_15, \
call_2_1_6_0, \
call_2_1_6_1, \
call_2_1_6_2, \
call_2_1_6_3, \
call_2_1_6_4, \
call_2_1_6_5, \
call_2_1_6_6, \
call_2_1_6_7, \
call_2_1_6_8, \
call_2_1_6_9, \
call_2_1_6_10, \
call_2_1_6_11, \
call_2_1_6_12, \
call_2_1_6_13, \
call_2_1_6_14, \
call_2_1_6_15, \
call_2_1_7_0, \
call_2_1_7_1, \
call_2_1_7_2, \
call_2_1_7_3, \
call_2_1_7_4, \
call_2_1_7_5, \
call_2_1_7_6, \
call_2_1_7_7, \
call_2_1_7_8, \
call_2_1_7_9, \
call_2_1_7_10, \
call_2_1_7_11, \
call_2_1_7_12, \
call_2_1_7_13, \
call_2_1_7_14, \
call_2_1_7_15, \
call_2_1_8_0, \
call_2_1_8_1, \
call_2_1_8_2, \
call_2_1_8_3, \
call_2_1_8_4, \
call_2_1_8_5, \
call_2_1_8_6, \
call_2_1_8_7, \
call_2_1_8_8, \
call_2_1_8_9, \
call_2_1_8_10, \
call_2_1_8_11, \
call_2_1_8_12, \
call_2_1_8_13, \
call_2_1_8_14, \
call_2_1_8_15, \
call_2_1_9_0, \
call_2_1_9_1, \
call_2_1_9_2, \
call_2_1_9_3, \
call_2_1_9_4, \
call_2_1_9_5, \
call_2_1_9_6, \
call_2_1_9_7, \
call_2_1_9_8, \
call_2_1_9_9, \
call_2_1_9_10, \
call_2_1_9_11, \
call_2_1_9_12, \
call_2_1_9_13, \
call_2_1_9_14, \
call_2_1_9_15, \
call_2_1_10_0, \
call_2_1_10_1, \
call_2_1_10_2, \
call_2_1_10_3, \
call_2_1_10_4, \
call_2_1_10_5, \
call_2_1_10_6, \
call_2_1_10_7, \
call_2_1_10_8, \
call_2_1_10_9, \
call_2_1_10_10, \
call_2_1_10_11, \
call_2_1_10_12, \
call_2_1_10_13, \
call_2_1_10_14, \
call_2_1_10_15, \
call_2_1_11_0, \
call_2_1_11_1, \
call_2_1_11_2, \
call_2_1_11_3, \
call_2_1_11_4, \
call_2_1_11_5, \
call_2_1_11_6, \
call_2_1_11_7, \
call_2_1_11_8, \
call_2_1_11_9, \
call_2_1_11_10, \
call_2_1_11_11, \
call_2_1_11_12, \
call_2_1_11_13, \
call_2_1_11_14, \
call_2_1_11_15, \
call_2_1_12_0, \
call_2_1_12_1, \
call_2_1_12_2, \
call_2_1_12_3, \
call_2_1_12_4, \
call_2_1_12_5, \
call_2_1_12_6, \
call_2_1_12_7, \
call_2_1_12_8, \
call_2_1_12_9, \
call_2_1_12_10, \
call_2_1_12_11, \
call_2_1_12_12, \
call_2_1_12_13, \
call_2_1_12_14, \
call_2_1_12_15, \
call_2_1_13_0, \
call_2_1_13_1, \
call_2_1_13_2, \
call_2_1_13_3, \
call_2_1_13_4, \
call_2_1_13_5, \
call_2_1_13_6, \
call_2_1_13_7, \
call_2_1_13_8, \
call_2_1_13_9, \
call_2_1_13_10, \
call_2_1_13_11, \
call_2_1_13_12, \
call_2_1_13_13, \
call_2_1_13_14, \
call_2_1_13_15, \
call_2_1_14_0, \
call_2_1_14_1, \
call_2_1_14_2, \
call_2_1_14_3, \
call_2_1_14_4, \
call_2_1_14_5, \
call_2_1_14_6, \
call_2_1_14_7, \
call_2_1_14_8, \
call_2_1_14_9, \
call_2_1_14_10, \
call_2_1_14_11, \
call_2_1_14_12, \
call_2_1_14_13, \
call_2_1_14_14, \
call_2_1_14_15, \
call_2_1_15_0, \
call_2_1_15_1, \
call_2_1_15_2, \
call_2_1_15_3, \
call_2_1_15_4, \
call_2_1_15_5, \
call_2_1_15_6, \
call_2_1_15_7, \
call_2_1_15_8, \
call_2_1_15_9, \
call_2_1_15_10, \
call_2_1_15_11, \
call_2_1_15_12, \
call_2_1_15_13, \
call_2_1_15_14, \
call_2_1_15_15, \
call_2_2_0_0, \
call_2_2_0_1, \
call_2_2_0_2, \
call_2_2_0_3, \
call_2_2_0_4, \
call_2_2_0_5, \
call_2_2_0_6, \
call_2_2_0_7, \
call_2_2_0_8, \
call_2_2_0_9, \
call_2_2_0_10, \
call_2_2_0_11, \
call_2_2_0_12, \
call_2_2_0_13, \
call_2_2_0_14, \
call_2_2_0_15, \
call_2_2_1_0, \
call_2_2_1_1, \
call_2_2_1_2, \
call_2_2_1_3, \
call_2_2_1_4, \
call_2_2_1_5, \
call_2_2_1_6, \
call_2_2_1_7, \
call_2_2_1_8, \
call_2_2_1_9, \
call_2_2_1_10, \
call_2_2_1_11, \
call_2_2_1_12, \
call_2_2_1_13, \
call_2_2_1_14, \
call_2_2_1_15, \
call_2_2_2_0, \
call_2_2_2_1, \
call_2_2_2_2, \
call_2_2_2_3, \
call_2_2_2_4, \
call_2_2_2_5, \
call_2_2_2_6, \
call_2_2_2_7, \
call_2_2_2_8, \
call_2_2_2_9, \
call_2_2_2_10, \
call_2_2_2_11, \
call_2_2_2_12, \
call_2_2_2_13, \
call_2_2_2_14, \
call_2_2_2_15, \
call_2_2_3_0, \
call_2_2_3_1, \
call_2_2_3_2, \
call_2_2_3_3, \
call_2_2_3_4, \
call_2_2_3_5, \
call_2_2_3_6, \
call_2_2_3_7, \
call_2_2_3_8, \
call_2_2_3_9, \
call_2_2_3_10, \
call_2_2_3_11, \
call_2_2_3_12, \
call_2_2_3_13, \
call_2_2_3_14, \
call_2_2_3_15, \
call_2_2_4_0, \
call_2_2_4_1, \
call_2_2_4_2, \
call_2_2_4_3, \
call_2_2_4_4, \
call_2_2_4_5, \
call_2_2_4_6, \
call_2_2_4_7, \
call_2_2_4_8, \
call_2_2_4_9, \
call_2_2_4_10, \
call_2_2_4_11, \
call_2_2_4_12, \
call_2_2_4_13, \
call_2_2_4_14, \
call_2_2_4_15, \
call_2_2_5_0, \
call_2_2_5_1, \
call_2_2_5_2, \
call_2_2_5_3, \
call_2_2_5_4, \
call_2_2_5_5, \
call_2_2_5_6, \
call_2_2_5_7, \
call_2_2_5_8, \
call_2_2_5_9, \
call_2_2_5_10, \
call_2_2_5_11, \
call_2_2_5_12, \
call_2_2_5_13, \
call_2_2_5_14, \
call_2_2_5_15, \
call_2_2_6_0, \
call_2_2_6_1, \
call_2_2_6_2, \
call_2_2_6_3, \
call_2_2_6_4, \
call_2_2_6_5, \
call_2_2_6_6, \
call_2_2_6_7, \
call_2_2_6_8, \
call_2_2_6_9, \
call_2_2_6_10, \
call_2_2_6_11, \
call_2_2_6_12, \
call_2_2_6_13, \
call_2_2_6_14, \
call_2_2_6_15, \
call_2_2_7_0, \
call_2_2_7_1, \
call_2_2_7_2, \
call_2_2_7_3, \
call_2_2_7_4, \
call_2_2_7_5, \
call_2_2_7_6, \
call_2_2_7_7, \
call_2_2_7_8, \
call_2_2_7_9, \
call_2_2_7_10, \
call_2_2_7_11, \
call_2_2_7_12, \
call_2_2_7_13, \
call_2_2_7_14, \
call_2_2_7_15, \
call_2_2_8_0, \
call_2_2_8_1, \
call_2_2_8_2, \
call_2_2_8_3, \
call_2_2_8_4, \
call_2_2_8_5, \
call_2_2_8_6, \
call_2_2_8_7, \
call_2_2_8_8, \
call_2_2_8_9, \
call_2_2_8_10, \
call_2_2_8_11, \
call_2_2_8_12, \
call_2_2_8_13, \
call_2_2_8_14, \
call_2_2_8_15, \
call_2_2_9_0, \
call_2_2_9_1, \
call_2_2_9_2, \
call_2_2_9_3, \
call_2_2_9_4, \
call_2_2_9_5, \
call_2_2_9_6, \
call_2_2_9_7, \
call_2_2_9_8, \
call_2_2_9_9, \
call_2_2_9_10, \
call_2_2_9_11, \
call_2_2_9_12, \
call_2_2_9_13, \
call_2_2_9_14, \
call_2_2_9_15, \
call_2_2_10_0, \
call_2_2_10_1, \
call_2_2_10_2, \
call_2_2_10_3, \
call_2_2_10_4, \
call_2_2_10_5, \
call_2_2_10_6, \
call_2_2_10_7, \
call_2_2_10_8, \
call_2_2_10_9, \
call_2_2_10_10, \
call_2_2_10_11, \
call_2_2_10_12, \
call_2_2_10_13, \
call_2_2_10_14, \
call_2_2_10_15, \
call_2_2_11_0, \
call_2_2_11_1, \
call_2_2_11_2, \
call_2_2_11_3, \
call_2_2_11_4, \
call_2_2_11_5, \
call_2_2_11_6, \
call_2_2_11_7, \
call_2_2_11_8, \
call_2_2_11_9, \
call_2_2_11_10, \
call_2_2_11_11, \
call_2_2_11_12, \
call_2_2_11_13, \
call_2_2_11_14, \
call_2_2_11_15, \
call_2_2_12_0, \
call_2_2_12_1, \
call_2_2_12_2, \
call_2_2_12_3, \
call_2_2_12_4, \
call_2_2_12_5, \
call_2_2_12_6, \
call_2_2_12_7, \
call_2_2_12_8, \
call_2_2_12_9, \
call_2_2_12_10, \
call_2_2_12_11, \
call_2_2_12_12, \
call_2_2_12_13, \
call_2_2_12_14, \
call_2_2_12_15, \
call_2_2_13_0, \
call_2_2_13_1, \
call_2_2_13_2, \
call_2_2_13_3, \
call_2_2_13_4, \
call_2_2_13_5, \
call_2_2_13_6, \
call_2_2_13_7, \
call_2_2_13_8, \
call_2_2_13_9, \
call_2_2_13_10, \
call_2_2_13_11, \
call_2_2_13_12, \
call_2_2_13_13, \
call_2_2_13_14, \
call_2_2_13_15, \
call_2_2_14_0, \
call_2_2_14_1, \
call_2_2_14_2, \
call_2_2_14_3, \
call_2_2_14_4, \
call_2_2_14_5, \
call_2_2_14_6, \
call_2_2_14_7, \
call_2_2_14_8, \
call_2_2_14_9, \
call_2_2_14_10, \
call_2_2_14_11, \
call_2_2_14_12, \
call_2_2_14_13, \
call_2_2_14_14, \
call_2_2_14_15, \
call_2_2_15_0, \
call_2_2_15_1, \
call_2_2_15_2, \
call_2_2_15_3, \
call_2_2_15_4, \
call_2_2_15_5, \
call_2_2_15_6, \
call_2_2_15_7, \
call_2_2_15_8, \
call_2_2_15_9, \
call_2_2_15_10, \
call_2_2_15_11, \
call_2_2_15_12, \
call_2_2_15_13, \
call_2_2_15_14, \
call_2_2_15_15, \
call_2_3_0_0, \
call_2_3_0_1, \
call_2_3_0_2, \
call_2_3_0_3, \
call_2_3_0_4, \
call_2_3_0_5, \
call_2_3_0_6, \
call_2_3_0_7, \
call_2_3_0_8, \
call_2_3_0_9, \
call_2_3_0_10, \
call_2_3_0_11, \
call_2_3_0_12, \
call_2_3_0_13, \
call_2_3_0_14, \
call_2_3_0_15, \
call_2_3_1_0, \
call_2_3_1_1, \
call_2_3_1_2, \
call_2_3_1_3, \
call_2_3_1_4, \
call_2_3_1_5, \
call_2_3_1_6, \
call_2_3_1_7, \
call_2_3_1_8, \
call_2_3_1_9, \
call_2_3_1_10, \
call_2_3_1_11, \
call_2_3_1_12, \
call_2_3_1_13, \
call_2_3_1_14, \
call_2_3_1_15, \
call_2_3_2_0, \
call_2_3_2_1, \
call_2_3_2_2, \
call_2_3_2_3, \
call_2_3_2_4, \
call_2_3_2_5, \
call_2_3_2_6, \
call_2_3_2_7, \
call_2_3_2_8, \
call_2_3_2_9, \
call_2_3_2_10, \
call_2_3_2_11, \
call_2_3_2_12, \
call_2_3_2_13, \
call_2_3_2_14, \
call_2_3_2_15, \
call_2_3_3_0, \
call_2_3_3_1, \
call_2_3_3_2, \
call_2_3_3_3, \
call_2_3_3_4, \
call_2_3_3_5, \
call_2_3_3_6, \
call_2_3_3_7, \
call_2_3_3_8, \
call_2_3_3_9, \
call_2_3_3_10, \
call_2_3_3_11, \
call_2_3_3_12, \
call_2_3_3_13, \
call_2_3_3_14, \
call_2_3_3_15, \
call_2_3_4_0, \
call_2_3_4_1, \
call_2_3_4_2, \
call_2_3_4_3, \
call_2_3_4_4, \
call_2_3_4_5, \
call_2_3_4_6, \
call_2_3_4_7, \
call_2_3_4_8, \
call_2_3_4_9, \
call_2_3_4_10, \
call_2_3_4_11, \
call_2_3_4_12, \
call_2_3_4_13, \
call_2_3_4_14, \
call_2_3_4_15, \
call_2_3_5_0, \
call_2_3_5_1, \
call_2_3_5_2, \
call_2_3_5_3, \
call_2_3_5_4, \
call_2_3_5_5, \
call_2_3_5_6, \
call_2_3_5_7, \
call_2_3_5_8, \
call_2_3_5_9, \
call_2_3_5_10, \
call_2_3_5_11, \
call_2_3_5_12, \
call_2_3_5_13, \
call_2_3_5_14, \
call_2_3_5_15, \
call_2_3_6_0, \
call_2_3_6_1, \
call_2_3_6_2, \
call_2_3_6_3, \
call_2_3_6_4, \
call_2_3_6_5, \
call_2_3_6_6, \
call_2_3_6_7, \
call_2_3_6_8, \
call_2_3_6_9, \
call_2_3_6_10, \
call_2_3_6_11, \
call_2_3_6_12, \
call_2_3_6_13, \
call_2_3_6_14, \
call_2_3_6_15, \
call_2_3_7_0, \
call_2_3_7_1, \
call_2_3_7_2, \
call_2_3_7_3, \
call_2_3_7_4, \
call_2_3_7_5, \
call_2_3_7_6, \
call_2_3_7_7, \
call_2_3_7_8, \
call_2_3_7_9, \
call_2_3_7_10, \
call_2_3_7_11, \
call_2_3_7_12, \
call_2_3_7_13, \
call_2_3_7_14, \
call_2_3_7_15, \
call_2_3_8_0, \
call_2_3_8_1, \
call_2_3_8_2, \
call_2_3_8_3, \
call_2_3_8_4, \
call_2_3_8_5, \
call_2_3_8_6, \
call_2_3_8_7, \
call_2_3_8_8, \
call_2_3_8_9, \
call_2_3_8_10, \
call_2_3_8_11, \
call_2_3_8_12, \
call_2_3_8_13, \
call_2_3_8_14, \
call_2_3_8_15, \
call_2_3_9_0, \
call_2_3_9_1, \
call_2_3_9_2, \
call_2_3_9_3, \
call_2_3_9_4, \
call_2_3_9_5, \
call_2_3_9_6, \
call_2_3_9_7, \
call_2_3_9_8, \
call_2_3_9_9, \
call_2_3_9_10, \
call_2_3_9_11, \
call_2_3_9_12, \
call_2_3_9_13, \
call_2_3_9_14, \
call_2_3_9_15, \
call_2_3_10_0, \
call_2_3_10_1, \
call_2_3_10_2, \
call_2_3_10_3, \
call_2_3_10_4, \
call_2_3_10_5, \
call_2_3_10_6, \
call_2_3_10_7, \
call_2_3_10_8, \
call_2_3_10_9, \
call_2_3_10_10, \
call_2_3_10_11, \
call_2_3_10_12, \
call_2_3_10_13, \
call_2_3_10_14, \
call_2_3_10_15, \
call_2_3_11_0, \
call_2_3_11_1, \
call_2_3_11_2, \
call_2_3_11_3, \
call_2_3_11_4, \
call_2_3_11_5, \
call_2_3_11_6, \
call_2_3_11_7, \
call_2_3_11_8, \
call_2_3_11_9, \
call_2_3_11_10, \
call_2_3_11_11, \
call_2_3_11_12, \
call_2_3_11_13, \
call_2_3_11_14, \
call_2_3_11_15, \
call_2_3_12_0, \
call_2_3_12_1, \
call_2_3_12_2, \
call_2_3_12_3, \
call_2_3_12_4, \
call_2_3_12_5, \
call_2_3_12_6, \
call_2_3_12_7, \
call_2_3_12_8, \
call_2_3_12_9, \
call_2_3_12_10, \
call_2_3_12_11, \
call_2_3_12_12, \
call_2_3_12_13, \
call_2_3_12_14, \
call_2_3_12_15, \
call_2_3_13_0, \
call_2_3_13_1, \
call_2_3_13_2, \
call_2_3_13_3, \
call_2_3_13_4, \
call_2_3_13_5, \
call_2_3_13_6, \
call_2_3_13_7, \
call_2_3_13_8, \
call_2_3_13_9, \
call_2_3_13_10, \
call_2_3_13_11, \
call_2_3_13_12, \
call_2_3_13_13, \
call_2_3_13_14, \
call_2_3_13_15, \
call_2_3_14_0, \
call_2_3_14_1, \
call_2_3_14_2, \
call_2_3_14_3, \
call_2_3_14_4, \
call_2_3_14_5, \
call_2_3_14_6, \
call_2_3_14_7, \
call_2_3_14_8, \
call_2_3_14_9, \
call_2_3_14_10, \
call_2_3_14_11, \
call_2_3_14_12, \
call_2_3_14_13, \
call_2_3_14_14, \
call_2_3_14_15, \
call_2_3_15_0, \
call_2_3_15_1, \
call_2_3_15_2, \
call_2_3_15_3, \
call_2_3_15_4, \
call_2_3_15_5, \
call_2_3_15_6, \
call_2_3_15_7, \
call_2_3_15_8, \
call_2_3_15_9, \
call_2_3_15_10, \
call_2_3_15_11, \
call_2_3_15_12, \
call_2_3_15_13, \
call_2_3_15_14, \
call_2_3_15_15, \
call_2_4_0_0, \
call_2_4_0_1, \
call_2_4_0_2, \
call_2_4_0_3, \
call_2_4_0_4, \
call_2_4_0_5, \
call_2_4_0_6, \
call_2_4_0_7, \
call_2_4_0_8, \
call_2_4_0_9, \
call_2_4_0_10, \
call_2_4_0_11, \
call_2_4_0_12, \
call_2_4_0_13, \
call_2_4_0_14, \
call_2_4_0_15, \
call_2_4_1_0, \
call_2_4_1_1, \
call_2_4_1_2, \
call_2_4_1_3, \
call_2_4_1_4, \
call_2_4_1_5, \
call_2_4_1_6, \
call_2_4_1_7, \
call_2_4_1_8, \
call_2_4_1_9, \
call_2_4_1_10, \
call_2_4_1_11, \
call_2_4_1_12, \
call_2_4_1_13, \
call_2_4_1_14, \
call_2_4_1_15, \
call_2_4_2_0, \
call_2_4_2_1, \
call_2_4_2_2, \
call_2_4_2_3, \
call_2_4_2_4, \
call_2_4_2_5, \
call_2_4_2_6, \
call_2_4_2_7, \
call_2_4_2_8, \
call_2_4_2_9, \
call_2_4_2_10, \
call_2_4_2_11, \
call_2_4_2_12, \
call_2_4_2_13, \
call_2_4_2_14, \
call_2_4_2_15, \
call_2_4_3_0, \
call_2_4_3_1, \
call_2_4_3_2, \
call_2_4_3_3, \
call_2_4_3_4, \
call_2_4_3_5, \
call_2_4_3_6, \
call_2_4_3_7, \
call_2_4_3_8, \
call_2_4_3_9, \
call_2_4_3_10, \
call_2_4_3_11, \
call_2_4_3_12, \
call_2_4_3_13, \
call_2_4_3_14, \
call_2_4_3_15, \
call_2_4_4_0, \
call_2_4_4_1, \
call_2_4_4_2, \
call_2_4_4_3, \
call_2_4_4_4, \
call_2_4_4_5, \
call_2_4_4_6, \
call_2_4_4_7, \
call_2_4_4_8, \
call_2_4_4_9, \
call_2_4_4_10, \
call_2_4_4_11, \
call_2_4_4_12, \
call_2_4_4_13, \
call_2_4_4_14, \
call_2_4_4_15, \
call_2_4_5_0, \
call_2_4_5_1, \
call_2_4_5_2, \
call_2_4_5_3, \
call_2_4_5_4, \
call_2_4_5_5, \
call_2_4_5_6, \
call_2_4_5_7, \
call_2_4_5_8, \
call_2_4_5_9, \
call_2_4_5_10, \
call_2_4_5_11, \
call_2_4_5_12, \
call_2_4_5_13, \
call_2_4_5_14, \
call_2_4_5_15, \
call_2_4_6_0, \
call_2_4_6_1, \
call_2_4_6_2, \
call_2_4_6_3, \
call_2_4_6_4, \
call_2_4_6_5, \
call_2_4_6_6, \
call_2_4_6_7, \
call_2_4_6_8, \
call_2_4_6_9, \
call_2_4_6_10, \
call_2_4_6_11, \
call_2_4_6_12, \
call_2_4_6_13, \
call_2_4_6_14, \
call_2_4_6_15, \
call_2_4_7_0, \
call_2_4_7_1, \
call_2_4_7_2, \
call_2_4_7_3, \
call_2_4_7_4, \
call_2_4_7_5, \
call_2_4_7_6, \
call_2_4_7_7, \
call_2_4_7_8, \
call_2_4_7_9, \
call_2_4_7_10, \
call_2_4_7_11, \
call_2_4_7_12, \
call_2_4_7_13, \
call_2_4_7_14, \
call_2_4_7_15, \
call_2_4_8_0, \
call_2_4_8_1, \
call_2_4_8_2, \
call_2_4_8_3, \
call_2_4_8_4, \
call_2_4_8_5, \
call_2_4_8_6, \
call_2_4_8_7, \
call_2_4_8_8, \
call_2_4_8_9, \
call_2_4_8_10, \
call_2_4_8_11, \
call_2_4_8_12, \
call_2_4_8_13, \
call_2_4_8_14, \
call_2_4_8_15, \
call_2_4_9_0, \
call_2_4_9_1, \
call_2_4_9_2, \
call_2_4_9_3, \
call_2_4_9_4, \
call_2_4_9_5, \
call_2_4_9_6, \
call_2_4_9_7, \
call_2_4_9_8, \
call_2_4_9_9, \
call_2_4_9_10, \
call_2_4_9_11, \
call_2_4_9_12, \
call_2_4_9_13, \
call_2_4_9_14, \
call_2_4_9_15, \
call_2_4_10_0, \
call_2_4_10_1, \
call_2_4_10_2, \
call_2_4_10_3, \
call_2_4_10_4, \
call_2_4_10_5, \
call_2_4_10_6, \
call_2_4_10_7, \
call_2_4_10_8, \
call_2_4_10_9, \
call_2_4_10_10, \
call_2_4_10_11, \
call_2_4_10_12, \
call_2_4_10_13, \
call_2_4_10_14, \
call_2_4_10_15, \
call_2_4_11_0, \
call_2_4_11_1, \
call_2_4_11_2, \
call_2_4_11_3, \
call_2_4_11_4, \
call_2_4_11_5, \
call_2_4_11_6, \
call_2_4_11_7, \
call_2_4_11_8, \
call_2_4_11_9, \
call_2_4_11_10, \
call_2_4_11_11, \
call_2_4_11_12, \
call_2_4_11_13, \
call_2_4_11_14, \
call_2_4_11_15, \
call_2_4_12_0, \
call_2_4_12_1, \
call_2_4_12_2, \
call_2_4_12_3, \
call_2_4_12_4, \
call_2_4_12_5, \
call_2_4_12_6, \
call_2_4_12_7, \
call_2_4_12_8, \
call_2_4_12_9, \
call_2_4_12_10, \
call_2_4_12_11, \
call_2_4_12_12, \
call_2_4_12_13, \
call_2_4_12_14, \
call_2_4_12_15, \
call_2_4_13_0, \
call_2_4_13_1, \
call_2_4_13_2, \
call_2_4_13_3, \
call_2_4_13_4, \
call_2_4_13_5, \
call_2_4_13_6, \
call_2_4_13_7, \
call_2_4_13_8, \
call_2_4_13_9, \
call_2_4_13_10, \
call_2_4_13_11, \
call_2_4_13_12, \
call_2_4_13_13, \
call_2_4_13_14, \
call_2_4_13_15, \
call_2_4_14_0, \
call_2_4_14_1, \
call_2_4_14_2, \
call_2_4_14_3, \
call_2_4_14_4, \
call_2_4_14_5, \
call_2_4_14_6, \
call_2_4_14_7, \
call_2_4_14_8, \
call_2_4_14_9, \
call_2_4_14_10, \
call_2_4_14_11, \
call_2_4_14_12, \
call_2_4_14_13, \
call_2_4_14_14, \
call_2_4_14_15, \
call_2_4_15_0, \
call_2_4_15_1, \
call_2_4_15_2, \
call_2_4_15_3, \
call_2_4_15_4, \
call_2_4_15_5, \
call_2_4_15_6, \
call_2_4_15_7, \
call_2_4_15_8, \
call_2_4_15_9, \
call_2_4_15_10, \
call_2_4_15_11, \
call_2_4_15_12, \
call_2_4_15_13, \
call_2_4_15_14, \
call_2_4_15_15, \
call_2_5_0_0, \
call_2_5_0_1, \
call_2_5_0_2, \
call_2_5_0_3, \
call_2_5_0_4, \
call_2_5_0_5, \
call_2_5_0_6, \
call_2_5_0_7, \
call_2_5_0_8, \
call_2_5_0_9, \
call_2_5_0_10, \
call_2_5_0_11, \
call_2_5_0_12, \
call_2_5_0_13, \
call_2_5_0_14, \
call_2_5_0_15, \
call_2_5_1_0, \
call_2_5_1_1, \
call_2_5_1_2, \
call_2_5_1_3, \
call_2_5_1_4, \
call_2_5_1_5, \
call_2_5_1_6, \
call_2_5_1_7, \
call_2_5_1_8, \
call_2_5_1_9, \
call_2_5_1_10, \
call_2_5_1_11, \
call_2_5_1_12, \
call_2_5_1_13, \
call_2_5_1_14, \
call_2_5_1_15, \
call_2_5_2_0, \
call_2_5_2_1, \
call_2_5_2_2, \
call_2_5_2_3, \
call_2_5_2_4, \
call_2_5_2_5, \
call_2_5_2_6, \
call_2_5_2_7, \
call_2_5_2_8, \
call_2_5_2_9, \
call_2_5_2_10, \
call_2_5_2_11, \
call_2_5_2_12, \
call_2_5_2_13, \
call_2_5_2_14, \
call_2_5_2_15, \
call_2_5_3_0, \
call_2_5_3_1, \
call_2_5_3_2, \
call_2_5_3_3, \
call_2_5_3_4, \
call_2_5_3_5, \
call_2_5_3_6, \
call_2_5_3_7, \
call_2_5_3_8, \
call_2_5_3_9, \
call_2_5_3_10, \
call_2_5_3_11, \
call_2_5_3_12, \
call_2_5_3_13, \
call_2_5_3_14, \
call_2_5_3_15, \
call_2_5_4_0, \
call_2_5_4_1, \
call_2_5_4_2, \
call_2_5_4_3, \
call_2_5_4_4, \
call_2_5_4_5, \
call_2_5_4_6, \
call_2_5_4_7, \
call_2_5_4_8, \
call_2_5_4_9, \
call_2_5_4_10, \
call_2_5_4_11, \
call_2_5_4_12, \
call_2_5_4_13, \
call_2_5_4_14, \
call_2_5_4_15, \
call_2_5_5_0, \
call_2_5_5_1, \
call_2_5_5_2, \
call_2_5_5_3, \
call_2_5_5_4, \
call_2_5_5_5, \
call_2_5_5_6, \
call_2_5_5_7, \
call_2_5_5_8, \
call_2_5_5_9, \
call_2_5_5_10, \
call_2_5_5_11, \
call_2_5_5_12, \
call_2_5_5_13, \
call_2_5_5_14, \
call_2_5_5_15, \
call_2_5_6_0, \
call_2_5_6_1, \
call_2_5_6_2, \
call_2_5_6_3, \
call_2_5_6_4, \
call_2_5_6_5, \
call_2_5_6_6, \
call_2_5_6_7, \
call_2_5_6_8, \
call_2_5_6_9, \
call_2_5_6_10, \
call_2_5_6_11, \
call_2_5_6_12, \
call_2_5_6_13, \
call_2_5_6_14, \
call_2_5_6_15, \
call_2_5_7_0, \
call_2_5_7_1, \
call_2_5_7_2, \
call_2_5_7_3, \
call_2_5_7_4, \
call_2_5_7_5, \
call_2_5_7_6, \
call_2_5_7_7, \
call_2_5_7_8, \
call_2_5_7_9, \
call_2_5_7_10, \
call_2_5_7_11, \
call_2_5_7_12, \
call_2_5_7_13, \
call_2_5_7_14, \
call_2_5_7_15, \
call_2_5_8_0, \
call_2_5_8_1, \
call_2_5_8_2, \
call_2_5_8_3, \
call_2_5_8_4, \
call_2_5_8_5, \
call_2_5_8_6, \
call_2_5_8_7, \
call_2_5_8_8, \
call_2_5_8_9, \
call_2_5_8_10, \
call_2_5_8_11, \
call_2_5_8_12, \
call_2_5_8_13, \
call_2_5_8_14, \
call_2_5_8_15, \
call_2_5_9_0, \
call_2_5_9_1, \
call_2_5_9_2, \
call_2_5_9_3, \
call_2_5_9_4, \
call_2_5_9_5, \
call_2_5_9_6, \
call_2_5_9_7, \
call_2_5_9_8, \
call_2_5_9_9, \
call_2_5_9_10, \
call_2_5_9_11, \
call_2_5_9_12, \
call_2_5_9_13, \
call_2_5_9_14, \
call_2_5_9_15, \
call_2_5_10_0, \
call_2_5_10_1, \
call_2_5_10_2, \
call_2_5_10_3, \
call_2_5_10_4, \
call_2_5_10_5, \
call_2_5_10_6, \
call_2_5_10_7, \
call_2_5_10_8, \
call_2_5_10_9, \
call_2_5_10_10, \
call_2_5_10_11, \
call_2_5_10_12, \
call_2_5_10_13, \
call_2_5_10_14, \
call_2_5_10_15, \
call_2_5_11_0, \
call_2_5_11_1, \
call_2_5_11_2, \
call_2_5_11_3, \
call_2_5_11_4, \
call_2_5_11_5, \
call_2_5_11_6, \
call_2_5_11_7, \
call_2_5_11_8, \
call_2_5_11_9, \
call_2_5_11_10, \
call_2_5_11_11, \
call_2_5_11_12, \
call_2_5_11_13, \
call_2_5_11_14, \
call_2_5_11_15, \
call_2_5_12_0, \
call_2_5_12_1, \
call_2_5_12_2, \
call_2_5_12_3, \
call_2_5_12_4, \
call_2_5_12_5, \
call_2_5_12_6, \
call_2_5_12_7, \
call_2_5_12_8, \
call_2_5_12_9, \
call_2_5_12_10, \
call_2_5_12_11, \
call_2_5_12_12, \
call_2_5_12_13, \
call_2_5_12_14, \
call_2_5_12_15, \
call_2_5_13_0, \
call_2_5_13_1, \
call_2_5_13_2, \
call_2_5_13_3, \
call_2_5_13_4, \
call_2_5_13_5, \
call_2_5_13_6, \
call_2_5_13_7, \
call_2_5_13_8, \
call_2_5_13_9, \
call_2_5_13_10, \
call_2_5_13_11, \
call_2_5_13_12, \
call_2_5_13_13, \
call_2_5_13_14, \
call_2_5_13_15, \
call_2_5_14_0, \
call_2_5_14_1, \
call_2_5_14_2, \
call_2_5_14_3, \
call_2_5_14_4, \
call_2_5_14_5, \
call_2_5_14_6, \
call_2_5_14_7, \
call_2_5_14_8, \
call_2_5_14_9, \
call_2_5_14_10, \
call_2_5_14_11, \
call_2_5_14_12, \
call_2_5_14_13, \
call_2_5_14_14, \
call_2_5_14_15, \
call_2_5_15_0, \
call_2_5_15_1, \
call_2_5_15_2, \
call_2_5_15_3, \
call_2_5_15_4, \
call_2_5_15_5, \
call_2_5_15_6, \
call_2_5_15_7, \
call_2_5_15_8, \
call_2_5_15_9, \
call_2_5_15_10, \
call_2_5_15_11, \
call_2_5_15_12, \
call_2_5_15_13, \
call_2_5_15_14, \
call_2_5_15_15, \
call_2_6_0_0, \
call_2_6_0_1, \
call_2_6_0_2, \
call_2_6_0_3, \
call_2_6_0_4, \
call_2_6_0_5, \
call_2_6_0_6, \
call_2_6_0_7, \
call_2_6_0_8, \
call_2_6_0_9, \
call_2_6_0_10, \
call_2_6_0_11, \
call_2_6_0_12, \
call_2_6_0_13, \
call_2_6_0_14, \
call_2_6_0_15, \
call_2_6_1_0, \
call_2_6_1_1, \
call_2_6_1_2, \
call_2_6_1_3, \
call_2_6_1_4, \
call_2_6_1_5, \
call_2_6_1_6, \
call_2_6_1_7, \
call_2_6_1_8, \
call_2_6_1_9, \
call_2_6_1_10, \
call_2_6_1_11, \
call_2_6_1_12, \
call_2_6_1_13, \
call_2_6_1_14, \
call_2_6_1_15, \
call_2_6_2_0, \
call_2_6_2_1, \
call_2_6_2_2, \
call_2_6_2_3, \
call_2_6_2_4, \
call_2_6_2_5, \
call_2_6_2_6, \
call_2_6_2_7, \
call_2_6_2_8, \
call_2_6_2_9, \
call_2_6_2_10, \
call_2_6_2_11, \
call_2_6_2_12, \
call_2_6_2_13, \
call_2_6_2_14, \
call_2_6_2_15, \
call_2_6_3_0, \
call_2_6_3_1, \
call_2_6_3_2, \
call_2_6_3_3, \
call_2_6_3_4, \
call_2_6_3_5, \
call_2_6_3_6, \
call_2_6_3_7, \
call_2_6_3_8, \
call_2_6_3_9, \
call_2_6_3_10, \
call_2_6_3_11, \
call_2_6_3_12, \
call_2_6_3_13, \
call_2_6_3_14, \
call_2_6_3_15, \
call_2_6_4_0, \
call_2_6_4_1, \
call_2_6_4_2, \
call_2_6_4_3, \
call_2_6_4_4, \
call_2_6_4_5, \
call_2_6_4_6, \
call_2_6_4_7, \
call_2_6_4_8, \
call_2_6_4_9, \
call_2_6_4_10, \
call_2_6_4_11, \
call_2_6_4_12, \
call_2_6_4_13, \
call_2_6_4_14, \
call_2_6_4_15, \
call_2_6_5_0, \
call_2_6_5_1, \
call_2_6_5_2, \
call_2_6_5_3, \
call_2_6_5_4, \
call_2_6_5_5, \
call_2_6_5_6, \
call_2_6_5_7, \
call_2_6_5_8, \
call_2_6_5_9, \
call_2_6_5_10, \
call_2_6_5_11, \
call_2_6_5_12, \
call_2_6_5_13, \
call_2_6_5_14, \
call_2_6_5_15, \
call_2_6_6_0, \
call_2_6_6_1, \
call_2_6_6_2, \
call_2_6_6_3, \
call_2_6_6_4, \
call_2_6_6_5, \
call_2_6_6_6, \
call_2_6_6_7, \
call_2_6_6_8, \
call_2_6_6_9, \
call_2_6_6_10, \
call_2_6_6_11, \
call_2_6_6_12, \
call_2_6_6_13, \
call_2_6_6_14, \
call_2_6_6_15, \
call_2_6_7_0, \
call_2_6_7_1, \
call_2_6_7_2, \
call_2_6_7_3, \
call_2_6_7_4, \
call_2_6_7_5, \
call_2_6_7_6, \
call_2_6_7_7, \
call_2_6_7_8, \
call_2_6_7_9, \
call_2_6_7_10, \
call_2_6_7_11, \
call_2_6_7_12, \
call_2_6_7_13, \
call_2_6_7_14, \
call_2_6_7_15, \
call_2_6_8_0, \
call_2_6_8_1, \
call_2_6_8_2, \
call_2_6_8_3, \
call_2_6_8_4, \
call_2_6_8_5, \
call_2_6_8_6, \
call_2_6_8_7, \
call_2_6_8_8, \
call_2_6_8_9, \
call_2_6_8_10, \
call_2_6_8_11, \
call_2_6_8_12, \
call_2_6_8_13, \
call_2_6_8_14, \
call_2_6_8_15, \
call_2_6_9_0, \
call_2_6_9_1, \
call_2_6_9_2, \
call_2_6_9_3, \
call_2_6_9_4, \
call_2_6_9_5, \
call_2_6_9_6, \
call_2_6_9_7, \
call_2_6_9_8, \
call_2_6_9_9, \
call_2_6_9_10, \
call_2_6_9_11, \
call_2_6_9_12, \
call_2_6_9_13, \
call_2_6_9_14, \
call_2_6_9_15, \
call_2_6_10_0, \
call_2_6_10_1, \
call_2_6_10_2, \
call_2_6_10_3, \
call_2_6_10_4, \
call_2_6_10_5, \
call_2_6_10_6, \
call_2_6_10_7, \
call_2_6_10_8, \
call_2_6_10_9, \
call_2_6_10_10, \
call_2_6_10_11, \
call_2_6_10_12, \
call_2_6_10_13, \
call_2_6_10_14, \
call_2_6_10_15, \
call_2_6_11_0, \
call_2_6_11_1, \
call_2_6_11_2, \
call_2_6_11_3, \
call_2_6_11_4, \
call_2_6_11_5, \
call_2_6_11_6, \
call_2_6_11_7, \
call_2_6_11_8, \
call_2_6_11_9, \
call_2_6_11_10, \
call_2_6_11_11, \
call_2_6_11_12, \
call_2_6_11_13, \
call_2_6_11_14, \
call_2_6_11_15, \
call_2_6_12_0, \
call_2_6_12_1, \
call_2_6_12_2, \
call_2_6_12_3, \
call_2_6_12_4, \
call_2_6_12_5, \
call_2_6_12_6, \
call_2_6_12_7, \
call_2_6_12_8, \
call_2_6_12_9, \
call_2_6_12_10, \
call_2_6_12_11, \
call_2_6_12_12, \
call_2_6_12_13, \
call_2_6_12_14, \
call_2_6_12_15, \
call_2_6_13_0, \
call_2_6_13_1, \
call_2_6_13_2, \
call_2_6_13_3, \
call_2_6_13_4, \
call_2_6_13_5, \
call_2_6_13_6, \
call_2_6_13_7, \
call_2_6_13_8, \
call_2_6_13_9, \
call_2_6_13_10, \
call_2_6_13_11, \
call_2_6_13_12, \
call_2_6_13_13, \
call_2_6_13_14, \
call_2_6_13_15, \
call_2_6_14_0, \
call_2_6_14_1, \
call_2_6_14_2, \
call_2_6_14_3, \
call_2_6_14_4, \
call_2_6_14_5, \
call_2_6_14_6, \
call_2_6_14_7, \
call_2_6_14_8, \
call_2_6_14_9, \
call_2_6_14_10, \
call_2_6_14_11, \
call_2_6_14_12, \
call_2_6_14_13, \
call_2_6_14_14, \
call_2_6_14_15, \
call_2_6_15_0, \
call_2_6_15_1, \
call_2_6_15_2, \
call_2_6_15_3, \
call_2_6_15_4, \
call_2_6_15_5, \
call_2_6_15_6, \
call_2_6_15_7, \
call_2_6_15_8, \
call_2_6_15_9, \
call_2_6_15_10, \
call_2_6_15_11, \
call_2_6_15_12, \
call_2_6_15_13, \
call_2_6_15_14, \
call_2_6_15_15, \
call_2_7_0_0, \
call_2_7_0_1, \
call_2_7_0_2, \
call_2_7_0_3, \
call_2_7_0_4, \
call_2_7_0_5, \
call_2_7_0_6, \
call_2_7_0_7, \
call_2_7_0_8, \
call_2_7_0_9, \
call_2_7_0_10, \
call_2_7_0_11, \
call_2_7_0_12, \
call_2_7_0_13, \
call_2_7_0_14, \
call_2_7_0_15, \
call_2_7_1_0, \
call_2_7_1_1, \
call_2_7_1_2, \
call_2_7_1_3, \
call_2_7_1_4, \
call_2_7_1_5, \
call_2_7_1_6, \
call_2_7_1_7, \
call_2_7_1_8, \
call_2_7_1_9, \
call_2_7_1_10, \
call_2_7_1_11, \
call_2_7_1_12, \
call_2_7_1_13, \
call_2_7_1_14, \
call_2_7_1_15, \
call_2_7_2_0, \
call_2_7_2_1, \
call_2_7_2_2, \
call_2_7_2_3, \
call_2_7_2_4, \
call_2_7_2_5, \
call_2_7_2_6, \
call_2_7_2_7, \
call_2_7_2_8, \
call_2_7_2_9, \
call_2_7_2_10, \
call_2_7_2_11, \
call_2_7_2_12, \
call_2_7_2_13, \
call_2_7_2_14, \
call_2_7_2_15, \
call_2_7_3_0, \
call_2_7_3_1, \
call_2_7_3_2, \
call_2_7_3_3, \
call_2_7_3_4, \
call_2_7_3_5, \
call_2_7_3_6, \
call_2_7_3_7, \
call_2_7_3_8, \
call_2_7_3_9, \
call_2_7_3_10, \
call_2_7_3_11, \
call_2_7_3_12, \
call_2_7_3_13, \
call_2_7_3_14, \
call_2_7_3_15, \
call_2_7_4_0, \
call_2_7_4_1, \
call_2_7_4_2, \
call_2_7_4_3, \
call_2_7_4_4, \
call_2_7_4_5, \
call_2_7_4_6, \
call_2_7_4_7, \
call_2_7_4_8, \
call_2_7_4_9, \
call_2_7_4_10, \
call_2_7_4_11, \
call_2_7_4_12, \
call_2_7_4_13, \
call_2_7_4_14, \
call_2_7_4_15, \
call_2_7_5_0, \
call_2_7_5_1, \
call_2_7_5_2, \
call_2_7_5_3, \
call_2_7_5_4, \
call_2_7_5_5, \
call_2_7_5_6, \
call_2_7_5_7, \
call_2_7_5_8, \
call_2_7_5_9, \
call_2_7_5_10, \
call_2_7_5_11, \
call_2_7_5_12, \
call_2_7_5_13, \
call_2_7_5_14, \
call_2_7_5_15, \
call_2_7_6_0, \
call_2_7_6_1, \
call_2_7_6_2, \
call_2_7_6_3, \
call_2_7_6_4, \
call_2_7_6_5, \
call_2_7_6_6, \
call_2_7_6_7, \
call_2_7_6_8, \
call_2_7_6_9, \
call_2_7_6_10, \
call_2_7_6_11, \
call_2_7_6_12, \
call_2_7_6_13, \
call_2_7_6_14, \
call_2_7_6_15, \
call_2_7_7_0, \
call_2_7_7_1, \
call_2_7_7_2, \
call_2_7_7_3, \
call_2_7_7_4, \
call_2_7_7_5, \
call_2_7_7_6, \
call_2_7_7_7, \
call_2_7_7_8, \
call_2_7_7_9, \
call_2_7_7_10, \
call_2_7_7_11, \
call_2_7_7_12, \
call_2_7_7_13, \
call_2_7_7_14, \
call_2_7_7_15, \
call_2_7_8_0, \
call_2_7_8_1, \
call_2_7_8_2, \
call_2_7_8_3, \
call_2_7_8_4, \
call_2_7_8_5, \
call_2_7_8_6, \
call_2_7_8_7, \
call_2_7_8_8, \
call_2_7_8_9, \
call_2_7_8_10, \
call_2_7_8_11, \
call_2_7_8_12, \
call_2_7_8_13, \
call_2_7_8_14, \
call_2_7_8_15, \
call_2_7_9_0, \
call_2_7_9_1, \
call_2_7_9_2, \
call_2_7_9_3, \
call_2_7_9_4, \
call_2_7_9_5, \
call_2_7_9_6, \
call_2_7_9_7, \
call_2_7_9_8, \
call_2_7_9_9, \
call_2_7_9_10, \
call_2_7_9_11, \
call_2_7_9_12, \
call_2_7_9_13, \
call_2_7_9_14, \
call_2_7_9_15, \
call_2_7_10_0, \
call_2_7_10_1, \
call_2_7_10_2, \
call_2_7_10_3, \
call_2_7_10_4, \
call_2_7_10_5, \
call_2_7_10_6, \
call_2_7_10_7, \
call_2_7_10_8, \
call_2_7_10_9, \
call_2_7_10_10, \
call_2_7_10_11, \
call_2_7_10_12, \
call_2_7_10_13, \
call_2_7_10_14, \
call_2_7_10_15, \
call_2_7_11_0, \
call_2_7_11_1, \
call_2_7_11_2, \
call_2_7_11_3, \
call_2_7_11_4, \
call_2_7_11_5, \
call_2_7_11_6, \
call_2_7_11_7, \
call_2_7_11_8, \
call_2_7_11_9, \
call_2_7_11_10, \
call_2_7_11_11, \
call_2_7_11_12, \
call_2_7_11_13, \
call_2_7_11_14, \
call_2_7_11_15, \
call_2_7_12_0, \
call_2_7_12_1, \
call_2_7_12_2, \
call_2_7_12_3, \
call_2_7_12_4, \
call_2_7_12_5, \
call_2_7_12_6, \
call_2_7_12_7, \
call_2_7_12_8, \
call_2_7_12_9, \
call_2_7_12_10, \
call_2_7_12_11, \
call_2_7_12_12, \
call_2_7_12_13, \
call_2_7_12_14, \
call_2_7_12_15, \
call_2_7_13_0, \
call_2_7_13_1, \
call_2_7_13_2, \
call_2_7_13_3, \
call_2_7_13_4, \
call_2_7_13_5, \
call_2_7_13_6, \
call_2_7_13_7, \
call_2_7_13_8, \
call_2_7_13_9, \
call_2_7_13_10, \
call_2_7_13_11, \
call_2_7_13_12, \
call_2_7_13_13, \
call_2_7_13_14, \
call_2_7_13_15, \
call_2_7_14_0, \
call_2_7_14_1, \
call_2_7_14_2, \
call_2_7_14_3, \
call_2_7_14_4, \
call_2_7_14_5, \
call_2_7_14_6, \
call_2_7_14_7, \
call_2_7_14_8, \
call_2_7_14_9, \
call_2_7_14_10, \
call_2_7_14_11, \
call_2_7_14_12, \
call_2_7_14_13, \
call_2_7_14_14, \
call_2_7_14_15, \
call_2_7_15_0, \
call_2_7_15_1, \
call_2_7_15_2, \
call_2_7_15_3, \
call_2_7_15_4, \
call_2_7_15_5, \
call_2_7_15_6, \
call_2_7_15_7, \
call_2_7_15_8, \
call_2_7_15_9, \
call_2_7_15_10, \
call_2_7_15_11, \
call_2_7_15_12, \
call_2_7_15_13, \
call_2_7_15_14, \
call_2_7_15_15, \
call_2_8_0_0, \
call_2_8_0_1, \
call_2_8_0_2, \
call_2_8_0_3, \
call_2_8_0_4, \
call_2_8_0_5, \
call_2_8_0_6, \
call_2_8_0_7, \
call_2_8_0_8, \
call_2_8_0_9, \
call_2_8_0_10, \
call_2_8_0_11, \
call_2_8_0_12, \
call_2_8_0_13, \
call_2_8_0_14, \
call_2_8_0_15, \
call_2_8_1_0, \
call_2_8_1_1, \
call_2_8_1_2, \
call_2_8_1_3, \
call_2_8_1_4, \
call_2_8_1_5, \
call_2_8_1_6, \
call_2_8_1_7, \
call_2_8_1_8, \
call_2_8_1_9, \
call_2_8_1_10, \
call_2_8_1_11, \
call_2_8_1_12, \
call_2_8_1_13, \
call_2_8_1_14, \
call_2_8_1_15, \
call_2_8_2_0, \
call_2_8_2_1, \
call_2_8_2_2, \
call_2_8_2_3, \
call_2_8_2_4, \
call_2_8_2_5, \
call_2_8_2_6, \
call_2_8_2_7, \
call_2_8_2_8, \
call_2_8_2_9, \
call_2_8_2_10, \
call_2_8_2_11, \
call_2_8_2_12, \
call_2_8_2_13, \
call_2_8_2_14, \
call_2_8_2_15, \
call_2_8_3_0, \
call_2_8_3_1, \
call_2_8_3_2, \
call_2_8_3_3, \
call_2_8_3_4, \
call_2_8_3_5, \
call_2_8_3_6, \
call_2_8_3_7, \
call_2_8_3_8, \
call_2_8_3_9, \
call_2_8_3_10, \
call_2_8_3_11, \
call_2_8_3_12, \
call_2_8_3_13, \
call_2_8_3_14, \
call_2_8_3_15, \
call_2_8_4_0, \
call_2_8_4_1, \
call_2_8_4_2, \
call_2_8_4_3, \
call_2_8_4_4, \
call_2_8_4_5, \
call_2_8_4_6, \
call_2_8_4_7, \
call_2_8_4_8, \
call_2_8_4_9, \
call_2_8_4_10, \
call_2_8_4_11, \
call_2_8_4_12, \
call_2_8_4_13, \
call_2_8_4_14, \
call_2_8_4_15, \
call_2_8_5_0, \
call_2_8_5_1, \
call_2_8_5_2, \
call_2_8_5_3, \
call_2_8_5_4, \
call_2_8_5_5, \
call_2_8_5_6, \
call_2_8_5_7, \
call_2_8_5_8, \
call_2_8_5_9, \
call_2_8_5_10, \
call_2_8_5_11, \
call_2_8_5_12, \
call_2_8_5_13, \
call_2_8_5_14, \
call_2_8_5_15, \
call_2_8_6_0, \
call_2_8_6_1, \
call_2_8_6_2, \
call_2_8_6_3, \
call_2_8_6_4, \
call_2_8_6_5, \
call_2_8_6_6, \
call_2_8_6_7, \
call_2_8_6_8, \
call_2_8_6_9, \
call_2_8_6_10, \
call_2_8_6_11, \
call_2_8_6_12, \
call_2_8_6_13, \
call_2_8_6_14, \
call_2_8_6_15, \
call_2_8_7_0, \
call_2_8_7_1, \
call_2_8_7_2, \
call_2_8_7_3, \
call_2_8_7_4, \
call_2_8_7_5, \
call_2_8_7_6, \
call_2_8_7_7, \
call_2_8_7_8, \
call_2_8_7_9, \
call_2_8_7_10, \
call_2_8_7_11, \
call_2_8_7_12, \
call_2_8_7_13, \
call_2_8_7_14, \
call_2_8_7_15, \
call_2_8_8_0, \
call_2_8_8_1, \
call_2_8_8_2, \
call_2_8_8_3, \
call_2_8_8_4, \
call_2_8_8_5, \
call_2_8_8_6, \
call_2_8_8_7, \
call_2_8_8_8, \
call_2_8_8_9, \
call_2_8_8_10, \
call_2_8_8_11, \
call_2_8_8_12, \
call_2_8_8_13, \
call_2_8_8_14, \
call_2_8_8_15, \
call_2_8_9_0, \
call_2_8_9_1, \
call_2_8_9_2, \
call_2_8_9_3, \
call_2_8_9_4, \
call_2_8_9_5, \
call_2_8_9_6, \
call_2_8_9_7, \
call_2_8_9_8, \
call_2_8_9_9, \
call_2_8_9_10, \
call_2_8_9_11, \
call_2_8_9_12, \
call_2_8_9_13, \
call_2_8_9_14, \
call_2_8_9_15, \
call_2_8_10_0, \
call_2_8_10_1, \
call_2_8_10_2, \
call_2_8_10_3, \
call_2_8_10_4, \
call_2_8_10_5, \
call_2_8_10_6, \
call_2_8_10_7, \
call_2_8_10_8, \
call_2_8_10_9, \
call_2_8_10_10, \
call_2_8_10_11, \
call_2_8_10_12, \
call_2_8_10_13, \
call_2_8_10_14, \
call_2_8_10_15, \
call_2_8_11_0, \
call_2_8_11_1, \
call_2_8_11_2, \
call_2_8_11_3, \
call_2_8_11_4, \
call_2_8_11_5, \
call_2_8_11_6, \
call_2_8_11_7, \
call_2_8_11_8, \
call_2_8_11_9, \
call_2_8_11_10, \
call_2_8_11_11, \
call_2_8_11_12, \
call_2_8_11_13, \
call_2_8_11_14, \
call_2_8_11_15, \
call_2_8_12_0, \
call_2_8_12_1, \
call_2_8_12_2, \
call_2_8_12_3, \
call_2_8_12_4, \
call_2_8_12_5, \
call_2_8_12_6, \
call_2_8_12_7, \
call_2_8_12_8, \
call_2_8_12_9, \
call_2_8_12_10, \
call_2_8_12_11, \
call_2_8_12_12, \
call_2_8_12_13, \
call_2_8_12_14, \
call_2_8_12_15, \
call_2_8_13_0, \
call_2_8_13_1, \
call_2_8_13_2, \
call_2_8_13_3, \
call_2_8_13_4, \
call_2_8_13_5, \
call_2_8_13_6, \
call_2_8_13_7, \
call_2_8_13_8, \
call_2_8_13_9, \
call_2_8_13_10, \
call_2_8_13_11, \
call_2_8_13_12, \
call_2_8_13_13, \
call_2_8_13_14, \
call_2_8_13_15, \
call_2_8_14_0, \
call_2_8_14_1, \
call_2_8_14_2, \
call_2_8_14_3, \
call_2_8_14_4, \
call_2_8_14_5, \
call_2_8_14_6, \
call_2_8_14_7, \
call_2_8_14_8, \
call_2_8_14_9, \
call_2_8_14_10, \
call_2_8_14_11, \
call_2_8_14_12, \
call_2_8_14_13, \
call_2_8_14_14, \
call_2_8_14_15, \
call_2_8_15_0, \
call_2_8_15_1, \
call_2_8_15_2, \
call_2_8_15_3, \
call_2_8_15_4, \
call_2_8_15_5, \
call_2_8_15_6, \
call_2_8_15_7, \
call_2_8_15_8, \
call_2_8_15_9, \
call_2_8_15_10, \
call_2_8_15_11, \
call_2_8_15_12, \
call_2_8_15_13, \
call_2_8_15_14, \
call_2_8_15_15, \
call_2_9_0_0, \
call_2_9_0_1, \
call_2_9_0_2, \
call_2_9_0_3, \
call_2_9_0_4, \
call_2_9_0_5, \
call_2_9_0_6, \
call_2_9_0_7, \
call_2_9_0_8, \
call_2_9_0_9, \
call_2_9_0_10, \
call_2_9_0_11, \
call_2_9_0_12, \
call_2_9_0_13, \
call_2_9_0_14, \
call_2_9_0_15, \
call_2_9_1_0, \
call_2_9_1_1, \
call_2_9_1_2, \
call_2_9_1_3, \
call_2_9_1_4, \
call_2_9_1_5, \
call_2_9_1_6, \
call_2_9_1_7, \
call_2_9_1_8, \
call_2_9_1_9, \
call_2_9_1_10, \
call_2_9_1_11, \
call_2_9_1_12, \
call_2_9_1_13, \
call_2_9_1_14, \
call_2_9_1_15, \
call_2_9_2_0, \
call_2_9_2_1, \
call_2_9_2_2, \
call_2_9_2_3, \
call_2_9_2_4, \
call_2_9_2_5, \
call_2_9_2_6, \
call_2_9_2_7, \
call_2_9_2_8, \
call_2_9_2_9, \
call_2_9_2_10, \
call_2_9_2_11, \
call_2_9_2_12, \
call_2_9_2_13, \
call_2_9_2_14, \
call_2_9_2_15, \
call_2_9_3_0, \
call_2_9_3_1, \
call_2_9_3_2, \
call_2_9_3_3, \
call_2_9_3_4, \
call_2_9_3_5, \
call_2_9_3_6, \
call_2_9_3_7, \
call_2_9_3_8, \
call_2_9_3_9, \
call_2_9_3_10, \
call_2_9_3_11, \
call_2_9_3_12, \
call_2_9_3_13, \
call_2_9_3_14, \
call_2_9_3_15, \
call_2_9_4_0, \
call_2_9_4_1, \
call_2_9_4_2, \
call_2_9_4_3, \
call_2_9_4_4, \
call_2_9_4_5, \
call_2_9_4_6, \
call_2_9_4_7, \
call_2_9_4_8, \
call_2_9_4_9, \
call_2_9_4_10, \
call_2_9_4_11, \
call_2_9_4_12, \
call_2_9_4_13, \
call_2_9_4_14, \
call_2_9_4_15, \
call_2_9_5_0, \
call_2_9_5_1, \
call_2_9_5_2, \
call_2_9_5_3, \
call_2_9_5_4, \
call_2_9_5_5, \
call_2_9_5_6, \
call_2_9_5_7, \
call_2_9_5_8, \
call_2_9_5_9, \
call_2_9_5_10, \
call_2_9_5_11, \
call_2_9_5_12, \
call_2_9_5_13, \
call_2_9_5_14, \
call_2_9_5_15, \
call_2_9_6_0, \
call_2_9_6_1, \
call_2_9_6_2, \
call_2_9_6_3, \
call_2_9_6_4, \
call_2_9_6_5, \
call_2_9_6_6, \
call_2_9_6_7, \
call_2_9_6_8, \
call_2_9_6_9, \
call_2_9_6_10, \
call_2_9_6_11, \
call_2_9_6_12, \
call_2_9_6_13, \
call_2_9_6_14, \
call_2_9_6_15, \
call_2_9_7_0, \
call_2_9_7_1, \
call_2_9_7_2, \
call_2_9_7_3, \
call_2_9_7_4, \
call_2_9_7_5, \
call_2_9_7_6, \
call_2_9_7_7, \
call_2_9_7_8, \
call_2_9_7_9, \
call_2_9_7_10, \
call_2_9_7_11, \
call_2_9_7_12, \
call_2_9_7_13, \
call_2_9_7_14, \
call_2_9_7_15, \
call_2_9_8_0, \
call_2_9_8_1, \
call_2_9_8_2, \
call_2_9_8_3, \
call_2_9_8_4, \
call_2_9_8_5, \
call_2_9_8_6, \
call_2_9_8_7, \
call_2_9_8_8, \
call_2_9_8_9, \
call_2_9_8_10, \
call_2_9_8_11, \
call_2_9_8_12, \
call_2_9_8_13, \
call_2_9_8_14, \
call_2_9_8_15, \
call_2_9_9_0, \
call_2_9_9_1, \
call_2_9_9_2, \
call_2_9_9_3, \
call_2_9_9_4, \
call_2_9_9_5, \
call_2_9_9_6, \
call_2_9_9_7, \
call_2_9_9_8, \
call_2_9_9_9, \
call_2_9_9_10, \
call_2_9_9_11, \
call_2_9_9_12, \
call_2_9_9_13, \
call_2_9_9_14, \
call_2_9_9_15, \
call_2_9_10_0, \
call_2_9_10_1, \
call_2_9_10_2, \
call_2_9_10_3, \
call_2_9_10_4, \
call_2_9_10_5, \
call_2_9_10_6, \
call_2_9_10_7, \
call_2_9_10_8, \
call_2_9_10_9, \
call_2_9_10_10, \
call_2_9_10_11, \
call_2_9_10_12, \
call_2_9_10_13, \
call_2_9_10_14, \
call_2_9_10_15, \
call_2_9_11_0, \
call_2_9_11_1, \
call_2_9_11_2, \
call_2_9_11_3, \
call_2_9_11_4, \
call_2_9_11_5, \
call_2_9_11_6, \
call_2_9_11_7, \
call_2_9_11_8, \
call_2_9_11_9, \
call_2_9_11_10, \
call_2_9_11_11, \
call_2_9_11_12, \
call_2_9_11_13, \
call_2_9_11_14, \
call_2_9_11_15, \
call_2_9_12_0, \
call_2_9_12_1, \
call_2_9_12_2, \
call_2_9_12_3, \
call_2_9_12_4, \
call_2_9_12_5, \
call_2_9_12_6, \
call_2_9_12_7, \
call_2_9_12_8, \
call_2_9_12_9, \
call_2_9_12_10, \
call_2_9_12_11, \
call_2_9_12_12, \
call_2_9_12_13, \
call_2_9_12_14, \
call_2_9_12_15, \
call_2_9_13_0, \
call_2_9_13_1, \
call_2_9_13_2, \
call_2_9_13_3, \
call_2_9_13_4, \
call_2_9_13_5, \
call_2_9_13_6, \
call_2_9_13_7, \
call_2_9_13_8, \
call_2_9_13_9, \
call_2_9_13_10, \
call_2_9_13_11, \
call_2_9_13_12, \
call_2_9_13_13, \
call_2_9_13_14, \
call_2_9_13_15, \
call_2_9_14_0, \
call_2_9_14_1, \
call_2_9_14_2, \
call_2_9_14_3, \
call_2_9_14_4, \
call_2_9_14_5, \
call_2_9_14_6, \
call_2_9_14_7, \
call_2_9_14_8, \
call_2_9_14_9, \
call_2_9_14_10, \
call_2_9_14_11, \
call_2_9_14_12, \
call_2_9_14_13, \
call_2_9_14_14, \
call_2_9_14_15, \
call_2_9_15_0, \
call_2_9_15_1, \
call_2_9_15_2, \
call_2_9_15_3, \
call_2_9_15_4, \
call_2_9_15_5, \
call_2_9_15_6, \
call_2_9_15_7, \
call_2_9_15_8, \
call_2_9_15_9, \
call_2_9_15_10, \
call_2_9_15_11, \
call_2_9_15_12, \
call_2_9_15_13, \
call_2_9_15_14, \
call_2_9_15_15, \
call_2_10_0_0, \
call_2_10_0_1, \
call_2_10_0_2, \
call_2_10_0_3, \
call_2_10_0_4, \
call_2_10_0_5, \
call_2_10_0_6, \
call_2_10_0_7, \
call_2_10_0_8, \
call_2_10_0_9, \
call_2_10_0_10, \
call_2_10_0_11, \
call_2_10_0_12, \
call_2_10_0_13, \
call_2_10_0_14, \
call_2_10_0_15, \
call_2_10_1_0, \
call_2_10_1_1, \
call_2_10_1_2, \
call_2_10_1_3, \
call_2_10_1_4, \
call_2_10_1_5, \
call_2_10_1_6, \
call_2_10_1_7, \
call_2_10_1_8, \
call_2_10_1_9, \
call_2_10_1_10, \
call_2_10_1_11, \
call_2_10_1_12, \
call_2_10_1_13, \
call_2_10_1_14, \
call_2_10_1_15, \
call_2_10_2_0, \
call_2_10_2_1, \
call_2_10_2_2, \
call_2_10_2_3, \
call_2_10_2_4, \
call_2_10_2_5, \
call_2_10_2_6, \
call_2_10_2_7, \
call_2_10_2_8, \
call_2_10_2_9, \
call_2_10_2_10, \
call_2_10_2_11, \
call_2_10_2_12, \
call_2_10_2_13, \
call_2_10_2_14, \
call_2_10_2_15, \
call_2_10_3_0, \
call_2_10_3_1, \
call_2_10_3_2, \
call_2_10_3_3, \
call_2_10_3_4, \
call_2_10_3_5, \
call_2_10_3_6, \
call_2_10_3_7, \
call_2_10_3_8, \
call_2_10_3_9, \
call_2_10_3_10, \
call_2_10_3_11, \
call_2_10_3_12, \
call_2_10_3_13, \
call_2_10_3_14, \
call_2_10_3_15, \
call_2_10_4_0, \
call_2_10_4_1, \
call_2_10_4_2, \
call_2_10_4_3, \
call_2_10_4_4, \
call_2_10_4_5, \
call_2_10_4_6, \
call_2_10_4_7, \
call_2_10_4_8, \
call_2_10_4_9, \
call_2_10_4_10, \
call_2_10_4_11, \
call_2_10_4_12, \
call_2_10_4_13, \
call_2_10_4_14, \
call_2_10_4_15, \
call_2_10_5_0, \
call_2_10_5_1, \
call_2_10_5_2, \
call_2_10_5_3, \
call_2_10_5_4, \
call_2_10_5_5, \
call_2_10_5_6, \
call_2_10_5_7, \
call_2_10_5_8, \
call_2_10_5_9, \
call_2_10_5_10, \
call_2_10_5_11, \
call_2_10_5_12, \
call_2_10_5_13, \
call_2_10_5_14, \
call_2_10_5_15, \
call_2_10_6_0, \
call_2_10_6_1, \
call_2_10_6_2, \
call_2_10_6_3, \
call_2_10_6_4, \
call_2_10_6_5, \
call_2_10_6_6, \
call_2_10_6_7, \
call_2_10_6_8, \
call_2_10_6_9, \
call_2_10_6_10, \
call_2_10_6_11, \
call_2_10_6_12, \
call_2_10_6_13, \
call_2_10_6_14, \
call_2_10_6_15, \
call_2_10_7_0, \
call_2_10_7_1, \
call_2_10_7_2, \
call_2_10_7_3, \
call_2_10_7_4, \
call_2_10_7_5, \
call_2_10_7_6, \
call_2_10_7_7, \
call_2_10_7_8, \
call_2_10_7_9, \
call_2_10_7_10, \
call_2_10_7_11, \
call_2_10_7_12, \
call_2_10_7_13, \
call_2_10_7_14, \
call_2_10_7_15, \
call_2_10_8_0, \
call_2_10_8_1, \
call_2_10_8_2, \
call_2_10_8_3, \
call_2_10_8_4, \
call_2_10_8_5, \
call_2_10_8_6, \
call_2_10_8_7, \
call_2_10_8_8, \
call_2_10_8_9, \
call_2_10_8_10, \
call_2_10_8_11, \
call_2_10_8_12, \
call_2_10_8_13, \
call_2_10_8_14, \
call_2_10_8_15, \
call_2_10_9_0, \
call_2_10_9_1, \
call_2_10_9_2, \
call_2_10_9_3, \
call_2_10_9_4, \
call_2_10_9_5, \
call_2_10_9_6, \
call_2_10_9_7, \
call_2_10_9_8, \
call_2_10_9_9, \
call_2_10_9_10, \
call_2_10_9_11, \
call_2_10_9_12, \
call_2_10_9_13, \
call_2_10_9_14, \
call_2_10_9_15, \
call_2_10_10_0, \
call_2_10_10_1, \
call_2_10_10_2, \
call_2_10_10_3, \
call_2_10_10_4, \
call_2_10_10_5, \
call_2_10_10_6, \
call_2_10_10_7, \
call_2_10_10_8, \
call_2_10_10_9, \
call_2_10_10_10, \
call_2_10_10_11, \
call_2_10_10_12, \
call_2_10_10_13, \
call_2_10_10_14, \
call_2_10_10_15, \
call_2_10_11_0, \
call_2_10_11_1, \
call_2_10_11_2, \
call_2_10_11_3, \
call_2_10_11_4, \
call_2_10_11_5, \
call_2_10_11_6, \
call_2_10_11_7, \
call_2_10_11_8, \
call_2_10_11_9, \
call_2_10_11_10, \
call_2_10_11_11, \
call_2_10_11_12, \
call_2_10_11_13, \
call_2_10_11_14, \
call_2_10_11_15, \
call_2_10_12_0, \
call_2_10_12_1, \
call_2_10_12_2, \
call_2_10_12_3, \
call_2_10_12_4, \
call_2_10_12_5, \
call_2_10_12_6, \
call_2_10_12_7, \
call_2_10_12_8, \
call_2_10_12_9, \
call_2_10_12_10, \
call_2_10_12_11, \
call_2_10_12_12, \
call_2_10_12_13, \
call_2_10_12_14, \
call_2_10_12_15, \
call_2_10_13_0, \
call_2_10_13_1, \
call_2_10_13_2, \
call_2_10_13_3, \
call_2_10_13_4, \
call_2_10_13_5, \
call_2_10_13_6, \
call_2_10_13_7, \
call_2_10_13_8, \
call_2_10_13_9, \
call_2_10_13_10, \
call_2_10_13_11, \
call_2_10_13_12, \
call_2_10_13_13, \
call_2_10_13_14, \
call_2_10_13_15, \
call_2_10_14_0, \
call_2_10_14_1, \
call_2_10_14_2, \
call_2_10_14_3, \
call_2_10_14_4, \
call_2_10_14_5, \
call_2_10_14_6, \
call_2_10_14_7, \
call_2_10_14_8, \
call_2_10_14_9, \
call_2_10_14_10, \
call_2_10_14_11, \
call_2_10_14_12, \
call_2_10_14_13, \
call_2_10_14_14, \
call_2_10_14_15, \
call_2_10_15_0, \
call_2_10_15_1, \
call_2_10_15_2, \
call_2_10_15_3, \
call_2_10_15_4, \
call_2_10_15_5, \
call_2_10_15_6, \
call_2_10_15_7, \
call_2_10_15_8, \
call_2_10_15_9, \
call_2_10_15_10, \
call_2_10_15_11, \
call_2_10_15_12, \
call_2_10_15_13, \
call_2_10_15_14, \
call_2_10_15_15, \
call_2_11_0_0, \
call_2_11_0_1, \
call_2_11_0_2, \
call_2_11_0_3, \
call_2_11_0_4, \
call_2_11_0_5, \
call_2_11_0_6, \
call_2_11_0_7, \
call_2_11_0_8, \
call_2_11_0_9, \
call_2_11_0_10, \
call_2_11_0_11, \
call_2_11_0_12, \
call_2_11_0_13, \
call_2_11_0_14, \
call_2_11_0_15, \
call_2_11_1_0, \
call_2_11_1_1, \
call_2_11_1_2, \
call_2_11_1_3, \
call_2_11_1_4, \
call_2_11_1_5, \
call_2_11_1_6, \
call_2_11_1_7, \
call_2_11_1_8, \
call_2_11_1_9, \
call_2_11_1_10, \
call_2_11_1_11, \
call_2_11_1_12, \
call_2_11_1_13, \
call_2_11_1_14, \
call_2_11_1_15, \
call_2_11_2_0, \
call_2_11_2_1, \
call_2_11_2_2, \
call_2_11_2_3, \
call_2_11_2_4, \
call_2_11_2_5, \
call_2_11_2_6, \
call_2_11_2_7, \
call_2_11_2_8, \
call_2_11_2_9, \
call_2_11_2_10, \
call_2_11_2_11, \
call_2_11_2_12, \
call_2_11_2_13, \
call_2_11_2_14, \
call_2_11_2_15, \
call_2_11_3_0, \
call_2_11_3_1, \
call_2_11_3_2, \
call_2_11_3_3, \
call_2_11_3_4, \
call_2_11_3_5, \
call_2_11_3_6, \
call_2_11_3_7, \
call_2_11_3_8, \
call_2_11_3_9, \
call_2_11_3_10, \
call_2_11_3_11, \
call_2_11_3_12, \
call_2_11_3_13, \
call_2_11_3_14, \
call_2_11_3_15, \
call_2_11_4_0, \
call_2_11_4_1, \
call_2_11_4_2, \
call_2_11_4_3, \
call_2_11_4_4, \
call_2_11_4_5, \
call_2_11_4_6, \
call_2_11_4_7, \
call_2_11_4_8, \
call_2_11_4_9, \
call_2_11_4_10, \
call_2_11_4_11, \
call_2_11_4_12, \
call_2_11_4_13, \
call_2_11_4_14, \
call_2_11_4_15, \
call_2_11_5_0, \
call_2_11_5_1, \
call_2_11_5_2, \
call_2_11_5_3, \
call_2_11_5_4, \
call_2_11_5_5, \
call_2_11_5_6, \
call_2_11_5_7, \
call_2_11_5_8, \
call_2_11_5_9, \
call_2_11_5_10, \
call_2_11_5_11, \
call_2_11_5_12, \
call_2_11_5_13, \
call_2_11_5_14, \
call_2_11_5_15, \
call_2_11_6_0, \
call_2_11_6_1, \
call_2_11_6_2, \
call_2_11_6_3, \
call_2_11_6_4, \
call_2_11_6_5, \
call_2_11_6_6, \
call_2_11_6_7, \
call_2_11_6_8, \
call_2_11_6_9, \
call_2_11_6_10, \
call_2_11_6_11, \
call_2_11_6_12, \
call_2_11_6_13, \
call_2_11_6_14, \
call_2_11_6_15, \
call_2_11_7_0, \
call_2_11_7_1, \
call_2_11_7_2, \
call_2_11_7_3, \
call_2_11_7_4, \
call_2_11_7_5, \
call_2_11_7_6, \
call_2_11_7_7, \
call_2_11_7_8, \
call_2_11_7_9, \
call_2_11_7_10, \
call_2_11_7_11, \
call_2_11_7_12, \
call_2_11_7_13, \
call_2_11_7_14, \
call_2_11_7_15, \
call_2_11_8_0, \
call_2_11_8_1, \
call_2_11_8_2, \
call_2_11_8_3, \
call_2_11_8_4, \
call_2_11_8_5, \
call_2_11_8_6, \
call_2_11_8_7, \
call_2_11_8_8, \
call_2_11_8_9, \
call_2_11_8_10, \
call_2_11_8_11, \
call_2_11_8_12, \
call_2_11_8_13, \
call_2_11_8_14, \
call_2_11_8_15, \
call_2_11_9_0, \
call_2_11_9_1, \
call_2_11_9_2, \
call_2_11_9_3, \
call_2_11_9_4, \
call_2_11_9_5, \
call_2_11_9_6, \
call_2_11_9_7, \
call_2_11_9_8, \
call_2_11_9_9, \
call_2_11_9_10, \
call_2_11_9_11, \
call_2_11_9_12, \
call_2_11_9_13, \
call_2_11_9_14, \
call_2_11_9_15, \
call_2_11_10_0, \
call_2_11_10_1, \
call_2_11_10_2, \
call_2_11_10_3, \
call_2_11_10_4, \
call_2_11_10_5, \
call_2_11_10_6, \
call_2_11_10_7, \
call_2_11_10_8, \
call_2_11_10_9, \
call_2_11_10_10, \
call_2_11_10_11, \
call_2_11_10_12, \
call_2_11_10_13, \
call_2_11_10_14, \
call_2_11_10_15, \
call_2_11_11_0, \
call_2_11_11_1, \
call_2_11_11_2, \
call_2_11_11_3, \
call_2_11_11_4, \
call_2_11_11_5, \
call_2_11_11_6, \
call_2_11_11_7, \
call_2_11_11_8, \
call_2_11_11_9, \
call_2_11_11_10, \
call_2_11_11_11, \
call_2_11_11_12, \
call_2_11_11_13, \
call_2_11_11_14, \
call_2_11_11_15, \
call_2_11_12_0, \
call_2_11_12_1, \
call_2_11_12_2, \
call_2_11_12_3, \
call_2_11_12_4, \
call_2_11_12_5, \
call_2_11_12_6, \
call_2_11_12_7, \
call_2_11_12_8, \
call_2_11_12_9, \
call_2_11_12_10, \
call_2_11_12_11, \
call_2_11_12_12, \
call_2_11_12_13, \
call_2_11_12_14, \
call_2_11_12_15, \
call_2_11_13_0, \
call_2_11_13_1, \
call_2_11_13_2, \
call_2_11_13_3, \
call_2_11_13_4, \
call_2_11_13_5, \
call_2_11_13_6, \
call_2_11_13_7, \
call_2_11_13_8, \
call_2_11_13_9, \
call_2_11_13_10, \
call_2_11_13_11, \
call_2_11_13_12, \
call_2_11_13_13, \
call_2_11_13_14, \
call_2_11_13_15, \
call_2_11_14_0, \
call_2_11_14_1, \
call_2_11_14_2, \
call_2_11_14_3, \
call_2_11_14_4, \
call_2_11_14_5, \
call_2_11_14_6, \
call_2_11_14_7, \
call_2_11_14_8, \
call_2_11_14_9, \
call_2_11_14_10, \
call_2_11_14_11, \
call_2_11_14_12, \
call_2_11_14_13, \
call_2_11_14_14, \
call_2_11_14_15, \
call_2_11_15_0, \
call_2_11_15_1, \
call_2_11_15_2, \
call_2_11_15_3, \
call_2_11_15_4, \
call_2_11_15_5, \
call_2_11_15_6, \
call_2_11_15_7, \
call_2_11_15_8, \
call_2_11_15_9, \
call_2_11_15_10, \
call_2_11_15_11, \
call_2_11_15_12, \
call_2_11_15_13, \
call_2_11_15_14, \
call_2_11_15_15, \
call_2_12_0_0, \
call_2_12_0_1, \
call_2_12_0_2, \
call_2_12_0_3, \
call_2_12_0_4, \
call_2_12_0_5, \
call_2_12_0_6, \
call_2_12_0_7, \
call_2_12_0_8, \
call_2_12_0_9, \
call_2_12_0_10, \
call_2_12_0_11, \
call_2_12_0_12, \
call_2_12_0_13, \
call_2_12_0_14, \
call_2_12_0_15, \
call_2_12_1_0, \
call_2_12_1_1, \
call_2_12_1_2, \
call_2_12_1_3, \
call_2_12_1_4, \
call_2_12_1_5, \
call_2_12_1_6, \
call_2_12_1_7, \
call_2_12_1_8, \
call_2_12_1_9, \
call_2_12_1_10, \
call_2_12_1_11, \
call_2_12_1_12, \
call_2_12_1_13, \
call_2_12_1_14, \
call_2_12_1_15, \
call_2_12_2_0, \
call_2_12_2_1, \
call_2_12_2_2, \
call_2_12_2_3, \
call_2_12_2_4, \
call_2_12_2_5, \
call_2_12_2_6, \
call_2_12_2_7, \
call_2_12_2_8, \
call_2_12_2_9, \
call_2_12_2_10, \
call_2_12_2_11, \
call_2_12_2_12, \
call_2_12_2_13, \
call_2_12_2_14, \
call_2_12_2_15, \
call_2_12_3_0, \
call_2_12_3_1, \
call_2_12_3_2, \
call_2_12_3_3, \
call_2_12_3_4, \
call_2_12_3_5, \
call_2_12_3_6, \
call_2_12_3_7, \
call_2_12_3_8, \
call_2_12_3_9, \
call_2_12_3_10, \
call_2_12_3_11, \
call_2_12_3_12, \
call_2_12_3_13, \
call_2_12_3_14, \
call_2_12_3_15, \
call_2_12_4_0, \
call_2_12_4_1, \
call_2_12_4_2, \
call_2_12_4_3, \
call_2_12_4_4, \
call_2_12_4_5, \
call_2_12_4_6, \
call_2_12_4_7, \
call_2_12_4_8, \
call_2_12_4_9, \
call_2_12_4_10, \
call_2_12_4_11, \
call_2_12_4_12, \
call_2_12_4_13, \
call_2_12_4_14, \
call_2_12_4_15, \
call_2_12_5_0, \
call_2_12_5_1, \
call_2_12_5_2, \
call_2_12_5_3, \
call_2_12_5_4, \
call_2_12_5_5, \
call_2_12_5_6, \
call_2_12_5_7, \
call_2_12_5_8, \
call_2_12_5_9, \
call_2_12_5_10, \
call_2_12_5_11, \
call_2_12_5_12, \
call_2_12_5_13, \
call_2_12_5_14, \
call_2_12_5_15, \
call_2_12_6_0, \
call_2_12_6_1, \
call_2_12_6_2, \
call_2_12_6_3, \
call_2_12_6_4, \
call_2_12_6_5, \
call_2_12_6_6, \
call_2_12_6_7, \
call_2_12_6_8, \
call_2_12_6_9, \
call_2_12_6_10, \
call_2_12_6_11, \
call_2_12_6_12, \
call_2_12_6_13, \
call_2_12_6_14, \
call_2_12_6_15, \
call_2_12_7_0, \
call_2_12_7_1, \
call_2_12_7_2, \
call_2_12_7_3, \
call_2_12_7_4, \
call_2_12_7_5, \
call_2_12_7_6, \
call_2_12_7_7, \
call_2_12_7_8, \
call_2_12_7_9, \
call_2_12_7_10, \
call_2_12_7_11, \
call_2_12_7_12, \
call_2_12_7_13, \
call_2_12_7_14, \
call_2_12_7_15, \
call_2_12_8_0, \
call_2_12_8_1, \
call_2_12_8_2, \
call_2_12_8_3, \
call_2_12_8_4, \
call_2_12_8_5, \
call_2_12_8_6, \
call_2_12_8_7, \
call_2_12_8_8, \
call_2_12_8_9, \
call_2_12_8_10, \
call_2_12_8_11, \
call_2_12_8_12, \
call_2_12_8_13, \
call_2_12_8_14, \
call_2_12_8_15, \
call_2_12_9_0, \
call_2_12_9_1, \
call_2_12_9_2, \
call_2_12_9_3, \
call_2_12_9_4, \
call_2_12_9_5, \
call_2_12_9_6, \
call_2_12_9_7, \
call_2_12_9_8, \
call_2_12_9_9, \
call_2_12_9_10, \
call_2_12_9_11, \
call_2_12_9_12, \
call_2_12_9_13, \
call_2_12_9_14, \
call_2_12_9_15, \
call_2_12_10_0, \
call_2_12_10_1, \
call_2_12_10_2, \
call_2_12_10_3, \
call_2_12_10_4, \
call_2_12_10_5, \
call_2_12_10_6, \
call_2_12_10_7, \
call_2_12_10_8, \
call_2_12_10_9, \
call_2_12_10_10, \
call_2_12_10_11, \
call_2_12_10_12, \
call_2_12_10_13, \
call_2_12_10_14, \
call_2_12_10_15, \
call_2_12_11_0, \
call_2_12_11_1, \
call_2_12_11_2, \
call_2_12_11_3, \
call_2_12_11_4, \
call_2_12_11_5, \
call_2_12_11_6, \
call_2_12_11_7, \
call_2_12_11_8, \
call_2_12_11_9, \
call_2_12_11_10, \
call_2_12_11_11, \
call_2_12_11_12, \
call_2_12_11_13, \
call_2_12_11_14, \
call_2_12_11_15, \
call_2_12_12_0, \
call_2_12_12_1, \
call_2_12_12_2, \
call_2_12_12_3, \
call_2_12_12_4, \
call_2_12_12_5, \
call_2_12_12_6, \
call_2_12_12_7, \
call_2_12_12_8, \
call_2_12_12_9, \
call_2_12_12_10, \
call_2_12_12_11, \
call_2_12_12_12, \
call_2_12_12_13, \
call_2_12_12_14, \
call_2_12_12_15, \
call_2_12_13_0, \
call_2_12_13_1, \
call_2_12_13_2, \
call_2_12_13_3, \
call_2_12_13_4, \
call_2_12_13_5, \
call_2_12_13_6, \
call_2_12_13_7, \
call_2_12_13_8, \
call_2_12_13_9, \
call_2_12_13_10, \
call_2_12_13_11, \
call_2_12_13_12, \
call_2_12_13_13, \
call_2_12_13_14, \
call_2_12_13_15, \
call_2_12_14_0, \
call_2_12_14_1, \
call_2_12_14_2, \
call_2_12_14_3, \
call_2_12_14_4, \
call_2_12_14_5, \
call_2_12_14_6, \
call_2_12_14_7, \
call_2_12_14_8, \
call_2_12_14_9, \
call_2_12_14_10, \
call_2_12_14_11, \
call_2_12_14_12, \
call_2_12_14_13, \
call_2_12_14_14, \
call_2_12_14_15, \
call_2_12_15_0, \
call_2_12_15_1, \
call_2_12_15_2, \
call_2_12_15_3, \
call_2_12_15_4, \
call_2_12_15_5, \
call_2_12_15_6, \
call_2_12_15_7, \
call_2_12_15_8, \
call_2_12_15_9, \
call_2_12_15_10, \
call_2_12_15_11, \
call_2_12_15_12, \
call_2_12_15_13, \
call_2_12_15_14, \
call_2_12_15_15, \
call_2_13_0_0, \
call_2_13_0_1, \
call_2_13_0_2, \
call_2_13_0_3, \
call_2_13_0_4, \
call_2_13_0_5, \
call_2_13_0_6, \
call_2_13_0_7, \
call_2_13_0_8, \
call_2_13_0_9, \
call_2_13_0_10, \
call_2_13_0_11, \
call_2_13_0_12, \
call_2_13_0_13, \
call_2_13_0_14, \
call_2_13_0_15, \
call_2_13_1_0, \
call_2_13_1_1, \
call_2_13_1_2, \
call_2_13_1_3, \
call_2_13_1_4, \
call_2_13_1_5, \
call_2_13_1_6, \
call_2_13_1_7, \
call_2_13_1_8, \
call_2_13_1_9, \
call_2_13_1_10, \
call_2_13_1_11, \
call_2_13_1_12, \
call_2_13_1_13, \
call_2_13_1_14, \
call_2_13_1_15, \
call_2_13_2_0, \
call_2_13_2_1, \
call_2_13_2_2, \
call_2_13_2_3, \
call_2_13_2_4, \
call_2_13_2_5, \
call_2_13_2_6, \
call_2_13_2_7, \
call_2_13_2_8, \
call_2_13_2_9, \
call_2_13_2_10, \
call_2_13_2_11, \
call_2_13_2_12, \
call_2_13_2_13, \
call_2_13_2_14, \
call_2_13_2_15, \
call_2_13_3_0, \
call_2_13_3_1, \
call_2_13_3_2, \
call_2_13_3_3, \
call_2_13_3_4, \
call_2_13_3_5, \
call_2_13_3_6, \
call_2_13_3_7, \
call_2_13_3_8, \
call_2_13_3_9, \
call_2_13_3_10, \
call_2_13_3_11, \
call_2_13_3_12, \
call_2_13_3_13, \
call_2_13_3_14, \
call_2_13_3_15, \
call_2_13_4_0, \
call_2_13_4_1, \
call_2_13_4_2, \
call_2_13_4_3, \
call_2_13_4_4, \
call_2_13_4_5, \
call_2_13_4_6, \
call_2_13_4_7, \
call_2_13_4_8, \
call_2_13_4_9, \
call_2_13_4_10, \
call_2_13_4_11, \
call_2_13_4_12, \
call_2_13_4_13, \
call_2_13_4_14, \
call_2_13_4_15, \
call_2_13_5_0, \
call_2_13_5_1, \
call_2_13_5_2, \
call_2_13_5_3, \
call_2_13_5_4, \
call_2_13_5_5, \
call_2_13_5_6, \
call_2_13_5_7, \
call_2_13_5_8, \
call_2_13_5_9, \
call_2_13_5_10, \
call_2_13_5_11, \
call_2_13_5_12, \
call_2_13_5_13, \
call_2_13_5_14, \
call_2_13_5_15, \
call_2_13_6_0, \
call_2_13_6_1, \
call_2_13_6_2, \
call_2_13_6_3, \
call_2_13_6_4, \
call_2_13_6_5, \
call_2_13_6_6, \
call_2_13_6_7, \
call_2_13_6_8, \
call_2_13_6_9, \
call_2_13_6_10, \
call_2_13_6_11, \
call_2_13_6_12, \
call_2_13_6_13, \
call_2_13_6_14, \
call_2_13_6_15, \
call_2_13_7_0, \
call_2_13_7_1, \
call_2_13_7_2, \
call_2_13_7_3, \
call_2_13_7_4, \
call_2_13_7_5, \
call_2_13_7_6, \
call_2_13_7_7, \
call_2_13_7_8, \
call_2_13_7_9, \
call_2_13_7_10, \
call_2_13_7_11, \
call_2_13_7_12, \
call_2_13_7_13, \
call_2_13_7_14, \
call_2_13_7_15, \
call_2_13_8_0, \
call_2_13_8_1, \
call_2_13_8_2, \
call_2_13_8_3, \
call_2_13_8_4, \
call_2_13_8_5, \
call_2_13_8_6, \
call_2_13_8_7, \
call_2_13_8_8, \
call_2_13_8_9, \
call_2_13_8_10, \
call_2_13_8_11, \
call_2_13_8_12, \
call_2_13_8_13, \
call_2_13_8_14, \
call_2_13_8_15, \
call_2_13_9_0, \
call_2_13_9_1, \
call_2_13_9_2, \
call_2_13_9_3, \
call_2_13_9_4, \
call_2_13_9_5, \
call_2_13_9_6, \
call_2_13_9_7, \
call_2_13_9_8, \
call_2_13_9_9, \
call_2_13_9_10, \
call_2_13_9_11, \
call_2_13_9_12, \
call_2_13_9_13, \
call_2_13_9_14, \
call_2_13_9_15, \
call_2_13_10_0, \
call_2_13_10_1, \
call_2_13_10_2, \
call_2_13_10_3, \
call_2_13_10_4, \
call_2_13_10_5, \
call_2_13_10_6, \
call_2_13_10_7, \
call_2_13_10_8, \
call_2_13_10_9, \
call_2_13_10_10, \
call_2_13_10_11, \
call_2_13_10_12, \
call_2_13_10_13, \
call_2_13_10_14, \
call_2_13_10_15, \
call_2_13_11_0, \
call_2_13_11_1, \
call_2_13_11_2, \
call_2_13_11_3, \
call_2_13_11_4, \
call_2_13_11_5, \
call_2_13_11_6, \
call_2_13_11_7, \
call_2_13_11_8, \
call_2_13_11_9, \
call_2_13_11_10, \
call_2_13_11_11, \
call_2_13_11_12, \
call_2_13_11_13, \
call_2_13_11_14, \
call_2_13_11_15, \
call_2_13_12_0, \
call_2_13_12_1, \
call_2_13_12_2, \
call_2_13_12_3, \
call_2_13_12_4, \
call_2_13_12_5, \
call_2_13_12_6, \
call_2_13_12_7, \
call_2_13_12_8, \
call_2_13_12_9, \
call_2_13_12_10, \
call_2_13_12_11, \
call_2_13_12_12, \
call_2_13_12_13, \
call_2_13_12_14, \
call_2_13_12_15, \
call_2_13_13_0, \
call_2_13_13_1, \
call_2_13_13_2, \
call_2_13_13_3, \
call_2_13_13_4, \
call_2_13_13_5, \
call_2_13_13_6, \
call_2_13_13_7, \
call_2_13_13_8, \
call_2_13_13_9, \
call_2_13_13_10, \
call_2_13_13_11, \
call_2_13_13_12, \
call_2_13_13_13, \
call_2_13_13_14, \
call_2_13_13_15, \
call_2_13_14_0, \
call_2_13_14_1, \
call_2_13_14_2, \
call_2_13_14_3, \
call_2_13_14_4, \
call_2_13_14_5, \
call_2_13_14_6, \
call_2_13_14_7, \
call_2_13_14_8, \
call_2_13_14_9, \
call_2_13_14_10, \
call_2_13_14_11, \
call_2_13_14_12, \
call_2_13_14_13, \
call_2_13_14_14, \
call_2_13_14_15, \
call_2_13_15_0, \
call_2_13_15_1, \
call_2_13_15_2, \
call_2_13_15_3, \
call_2_13_15_4, \
call_2_13_15_5, \
call_2_13_15_6, \
call_2_13_15_7, \
call_2_13_15_8, \
call_2_13_15_9, \
call_2_13_15_10, \
call_2_13_15_11, \
call_2_13_15_12, \
call_2_13_15_13, \
call_2_13_15_14, \
call_2_13_15_15, \
call_2_14_0_0, \
call_2_14_0_1, \
call_2_14_0_2, \
call_2_14_0_3, \
call_2_14_0_4, \
call_2_14_0_5, \
call_2_14_0_6, \
call_2_14_0_7, \
call_2_14_0_8, \
call_2_14_0_9, \
call_2_14_0_10, \
call_2_14_0_11, \
call_2_14_0_12, \
call_2_14_0_13, \
call_2_14_0_14, \
call_2_14_0_15, \
call_2_14_1_0, \
call_2_14_1_1, \
call_2_14_1_2, \
call_2_14_1_3, \
call_2_14_1_4, \
call_2_14_1_5, \
call_2_14_1_6, \
call_2_14_1_7, \
call_2_14_1_8, \
call_2_14_1_9, \
call_2_14_1_10, \
call_2_14_1_11, \
call_2_14_1_12, \
call_2_14_1_13, \
call_2_14_1_14, \
call_2_14_1_15, \
call_2_14_2_0, \
call_2_14_2_1, \
call_2_14_2_2, \
call_2_14_2_3, \
call_2_14_2_4, \
call_2_14_2_5, \
call_2_14_2_6, \
call_2_14_2_7, \
call_2_14_2_8, \
call_2_14_2_9, \
call_2_14_2_10, \
call_2_14_2_11, \
call_2_14_2_12, \
call_2_14_2_13, \
call_2_14_2_14, \
call_2_14_2_15, \
call_2_14_3_0, \
call_2_14_3_1, \
call_2_14_3_2, \
call_2_14_3_3, \
call_2_14_3_4, \
call_2_14_3_5, \
call_2_14_3_6, \
call_2_14_3_7, \
call_2_14_3_8, \
call_2_14_3_9, \
call_2_14_3_10, \
call_2_14_3_11, \
call_2_14_3_12, \
call_2_14_3_13, \
call_2_14_3_14, \
call_2_14_3_15, \
call_2_14_4_0, \
call_2_14_4_1, \
call_2_14_4_2, \
call_2_14_4_3, \
call_2_14_4_4, \
call_2_14_4_5, \
call_2_14_4_6, \
call_2_14_4_7, \
call_2_14_4_8, \
call_2_14_4_9, \
call_2_14_4_10, \
call_2_14_4_11, \
call_2_14_4_12, \
call_2_14_4_13, \
call_2_14_4_14, \
call_2_14_4_15, \
call_2_14_5_0, \
call_2_14_5_1, \
call_2_14_5_2, \
call_2_14_5_3, \
call_2_14_5_4, \
call_2_14_5_5, \
call_2_14_5_6, \
call_2_14_5_7, \
call_2_14_5_8, \
call_2_14_5_9, \
call_2_14_5_10, \
call_2_14_5_11, \
call_2_14_5_12, \
call_2_14_5_13, \
call_2_14_5_14, \
call_2_14_5_15, \
call_2_14_6_0, \
call_2_14_6_1, \
call_2_14_6_2, \
call_2_14_6_3, \
call_2_14_6_4, \
call_2_14_6_5, \
call_2_14_6_6, \
call_2_14_6_7, \
call_2_14_6_8, \
call_2_14_6_9, \
call_2_14_6_10, \
call_2_14_6_11, \
call_2_14_6_12, \
call_2_14_6_13, \
call_2_14_6_14, \
call_2_14_6_15, \
call_2_14_7_0, \
call_2_14_7_1, \
call_2_14_7_2, \
call_2_14_7_3, \
call_2_14_7_4, \
call_2_14_7_5, \
call_2_14_7_6, \
call_2_14_7_7, \
call_2_14_7_8, \
call_2_14_7_9, \
call_2_14_7_10, \
call_2_14_7_11, \
call_2_14_7_12, \
call_2_14_7_13, \
call_2_14_7_14, \
call_2_14_7_15, \
call_2_14_8_0, \
call_2_14_8_1, \
call_2_14_8_2, \
call_2_14_8_3, \
call_2_14_8_4, \
call_2_14_8_5, \
call_2_14_8_6, \
call_2_14_8_7, \
call_2_14_8_8, \
call_2_14_8_9, \
call_2_14_8_10, \
call_2_14_8_11, \
call_2_14_8_12, \
call_2_14_8_13, \
call_2_14_8_14, \
call_2_14_8_15, \
call_2_14_9_0, \
call_2_14_9_1, \
call_2_14_9_2, \
call_2_14_9_3, \
call_2_14_9_4, \
call_2_14_9_5, \
call_2_14_9_6, \
call_2_14_9_7, \
call_2_14_9_8, \
call_2_14_9_9, \
call_2_14_9_10, \
call_2_14_9_11, \
call_2_14_9_12, \
call_2_14_9_13, \
call_2_14_9_14, \
call_2_14_9_15, \
call_2_14_10_0, \
call_2_14_10_1, \
call_2_14_10_2, \
call_2_14_10_3, \
call_2_14_10_4, \
call_2_14_10_5, \
call_2_14_10_6, \
call_2_14_10_7, \
call_2_14_10_8, \
call_2_14_10_9, \
call_2_14_10_10, \
call_2_14_10_11, \
call_2_14_10_12, \
call_2_14_10_13, \
call_2_14_10_14, \
call_2_14_10_15, \
call_2_14_11_0, \
call_2_14_11_1, \
call_2_14_11_2, \
call_2_14_11_3, \
call_2_14_11_4, \
call_2_14_11_5, \
call_2_14_11_6, \
call_2_14_11_7, \
call_2_14_11_8, \
call_2_14_11_9, \
call_2_14_11_10, \
call_2_14_11_11, \
call_2_14_11_12, \
call_2_14_11_13, \
call_2_14_11_14, \
call_2_14_11_15, \
call_2_14_12_0, \
call_2_14_12_1, \
call_2_14_12_2, \
call_2_14_12_3, \
call_2_14_12_4, \
call_2_14_12_5, \
call_2_14_12_6, \
call_2_14_12_7, \
call_2_14_12_8, \
call_2_14_12_9, \
call_2_14_12_10, \
call_2_14_12_11, \
call_2_14_12_12, \
call_2_14_12_13, \
call_2_14_12_14, \
call_2_14_12_15, \
call_2_14_13_0, \
call_2_14_13_1, \
call_2_14_13_2, \
call_2_14_13_3, \
call_2_14_13_4, \
call_2_14_13_5, \
call_2_14_13_6, \
call_2_14_13_7, \
call_2_14_13_8, \
call_2_14_13_9, \
call_2_14_13_10, \
call_2_14_13_11, \
call_2_14_13_12, \
call_2_14_13_13, \
call_2_14_13_14, \
call_2_14_13_15, \
call_2_14_14_0, \
call_2_14_14_1, \
call_2_14_14_2, \
call_2_14_14_3, \
call_2_14_14_4, \
call_2_14_14_5, \
call_2_14_14_6, \
call_2_14_14_7, \
call_2_14_14_8, \
call_2_14_14_9, \
call_2_14_14_10, \
call_2_14_14_11, \
call_2_14_14_12, \
call_2_14_14_13, \
call_2_14_14_14, \
call_2_14_14_15, \
call_2_14_15_0, \
call_2_14_15_1, \
call_2_14_15_2, \
call_2_14_15_3, \
call_2_14_15_4, \
call_2_14_15_5, \
call_2_14_15_6, \
call_2_14_15_7, \
call_2_14_15_8, \
call_2_14_15_9, \
call_2_14_15_10, \
call_2_14_15_11, \
call_2_14_15_12, \
call_2_14_15_13, \
call_2_14_15_14, \
call_2_14_15_15, \
call_2_15_0_0, \
call_2_15_0_1, \
call_2_15_0_2, \
call_2_15_0_3, \
call_2_15_0_4, \
call_2_15_0_5, \
call_2_15_0_6, \
call_2_15_0_7, \
call_2_15_0_8, \
call_2_15_0_9, \
call_2_15_0_10, \
call_2_15_0_11, \
call_2_15_0_12, \
call_2_15_0_13, \
call_2_15_0_14, \
call_2_15_0_15, \
call_2_15_1_0, \
call_2_15_1_1, \
call_2_15_1_2, \
call_2_15_1_3, \
call_2_15_1_4, \
call_2_15_1_5, \
call_2_15_1_6, \
call_2_15_1_7, \
call_2_15_1_8, \
call_2_15_1_9, \
call_2_15_1_10, \
call_2_15_1_11, \
call_2_15_1_12, \
call_2_15_1_13, \
call_2_15_1_14, \
call_2_15_1_15, \
call_2_15_2_0, \
call_2_15_2_1, \
call_2_15_2_2, \
call_2_15_2_3, \
call_2_15_2_4, \
call_2_15_2_5, \
call_2_15_2_6, \
call_2_15_2_7, \
call_2_15_2_8, \
call_2_15_2_9, \
call_2_15_2_10, \
call_2_15_2_11, \
call_2_15_2_12, \
call_2_15_2_13, \
call_2_15_2_14, \
call_2_15_2_15, \
call_2_15_3_0, \
call_2_15_3_1, \
call_2_15_3_2, \
call_2_15_3_3, \
call_2_15_3_4, \
call_2_15_3_5, \
call_2_15_3_6, \
call_2_15_3_7, \
call_2_15_3_8, \
call_2_15_3_9, \
call_2_15_3_10, \
call_2_15_3_11, \
call_2_15_3_12, \
call_2_15_3_13, \
call_2_15_3_14, \
call_2_15_3_15, \
call_2_15_4_0, \
call_2_15_4_1, \
call_2_15_4_2, \
call_2_15_4_3, \
call_2_15_4_4, \
call_2_15_4_5, \
call_2_15_4_6, \
call_2_15_4_7, \
call_2_15_4_8, \
call_2_15_4_9, \
call_2_15_4_10, \
call_2_15_4_11, \
call_2_15_4_12, \
call_2_15_4_13, \
call_2_15_4_14, \
call_2_15_4_15, \
call_2_15_5_0, \
call_2_15_5_1, \
call_2_15_5_2, \
call_2_15_5_3, \
call_2_15_5_4, \
call_2_15_5_5, \
call_2_15_5_6, \
call_2_15_5_7, \
call_2_15_5_8, \
call_2_15_5_9, \
call_2_15_5_10, \
call_2_15_5_11, \
call_2_15_5_12, \
call_2_15_5_13, \
call_2_15_5_14, \
call_2_15_5_15, \
call_2_15_6_0, \
call_2_15_6_1, \
call_2_15_6_2, \
call_2_15_6_3, \
call_2_15_6_4, \
call_2_15_6_5, \
call_2_15_6_6, \
call_2_15_6_7, \
call_2_15_6_8, \
call_2_15_6_9, \
call_2_15_6_10, \
call_2_15_6_11, \
call_2_15_6_12, \
call_2_15_6_13, \
call_2_15_6_14, \
call_2_15_6_15, \
call_2_15_7_0, \
call_2_15_7_1, \
call_2_15_7_2, \
call_2_15_7_3, \
call_2_15_7_4, \
call_2_15_7_5, \
call_2_15_7_6, \
call_2_15_7_7, \
call_2_15_7_8, \
call_2_15_7_9, \
call_2_15_7_10, \
call_2_15_7_11, \
call_2_15_7_12, \
call_2_15_7_13, \
call_2_15_7_14, \
call_2_15_7_15, \
call_2_15_8_0, \
call_2_15_8_1, \
call_2_15_8_2, \
call_2_15_8_3, \
call_2_15_8_4, \
call_2_15_8_5, \
call_2_15_8_6, \
call_2_15_8_7, \
call_2_15_8_8, \
call_2_15_8_9, \
call_2_15_8_10, \
call_2_15_8_11, \
call_2_15_8_12, \
call_2_15_8_13, \
call_2_15_8_14, \
call_2_15_8_15, \
call_2_15_9_0, \
call_2_15_9_1, \
call_2_15_9_2, \
call_2_15_9_3, \
call_2_15_9_4, \
call_2_15_9_5, \
call_2_15_9_6, \
call_2_15_9_7, \
call_2_15_9_8, \
call_2_15_9_9, \
call_2_15_9_10, \
call_2_15_9_11, \
call_2_15_9_12, \
call_2_15_9_13, \
call_2_15_9_14, \
call_2_15_9_15, \
call_2_15_10_0, \
call_2_15_10_1, \
call_2_15_10_2, \
call_2_15_10_3, \
call_2_15_10_4, \
call_2_15_10_5, \
call_2_15_10_6, \
call_2_15_10_7, \
call_2_15_10_8, \
call_2_15_10_9, \
call_2_15_10_10, \
call_2_15_10_11, \
call_2_15_10_12, \
call_2_15_10_13, \
call_2_15_10_14, \
call_2_15_10_15, \
call_2_15_11_0, \
call_2_15_11_1, \
call_2_15_11_2, \
call_2_15_11_3, \
call_2_15_11_4, \
call_2_15_11_5, \
call_2_15_11_6, \
call_2_15_11_7, \
call_2_15_11_8, \
call_2_15_11_9, \
call_2_15_11_10, \
call_2_15_11_11, \
call_2_15_11_12, \
call_2_15_11_13, \
call_2_15_11_14, \
call_2_15_11_15, \
call_2_15_12_0, \
call_2_15_12_1, \
call_2_15_12_2, \
call_2_15_12_3, \
call_2_15_12_4, \
call_2_15_12_5, \
call_2_15_12_6, \
call_2_15_12_7, \
call_2_15_12_8, \
call_2_15_12_9, \
call_2_15_12_10, \
call_2_15_12_11, \
call_2_15_12_12, \
call_2_15_12_13, \
call_2_15_12_14, \
call_2_15_12_15, \
call_2_15_13_0, \
call_2_15_13_1, \
call_2_15_13_2, \
call_2_15_13_3, \
call_2_15_13_4, \
call_2_15_13_5, \
call_2_15_13_6, \
call_2_15_13_7, \
call_2_15_13_8, \
call_2_15_13_9, \
call_2_15_13_10, \
call_2_15_13_11, \
call_2_15_13_12, \
call_2_15_13_13, \
call_2_15_13_14, \
call_2_15_13_15, \
call_2_15_14_0, \
call_2_15_14_1, \
call_2_15_14_2, \
call_2_15_14_3, \
call_2_15_14_4, \
call_2_15_14_5, \
call_2_15_14_6, \
call_2_15_14_7, \
call_2_15_14_8, \
call_2_15_14_9, \
call_2_15_14_10, \
call_2_15_14_11, \
call_2_15_14_12, \
call_2_15_14_13, \
call_2_15_14_14, \
call_2_15_14_15, \
call_2_15_15_0, \
call_2_15_15_1, \
call_2_15_15_2, \
call_2_15_15_3, \
call_2_15_15_4, \
call_2_15_15_5, \
call_2_15_15_6, \
call_2_15_15_7, \
call_2_15_15_8, \
call_2_15_15_9, \
call_2_15_15_10, \
call_2_15_15_11, \
call_2_15_15_12, \
call_2_15_15_13, \
call_2_15_15_14, \
call_2_15_15_15, \
call_3_0_0_0, \
call_3_0_0_1, \
call_3_0_0_2, \
call_3_0_0_3, \
call_3_0_0_4, \
call_3_0_0_5, \
call_3_0_0_6, \
call_3_0_0_7, \
call_3_0_0_8, \
call_3_0_0_9, \
call_3_0_0_10, \
call_3_0_0_11, \
call_3_0_0_12, \
call_3_0_0_13, \
call_3_0_0_14, \
call_3_0_0_15, \
call_3_0_1_0, \
call_3_0_1_1, \
call_3_0_1_2, \
call_3_0_1_3, \
call_3_0_1_4, \
call_3_0_1_5, \
call_3_0_1_6, \
call_3_0_1_7, \
call_3_0_1_8, \
call_3_0_1_9, \
call_3_0_1_10, \
call_3_0_1_11, \
call_3_0_1_12, \
call_3_0_1_13, \
call_3_0_1_14, \
call_3_0_1_15, \
call_3_0_2_0, \
call_3_0_2_1, \
call_3_0_2_2, \
call_3_0_2_3, \
call_3_0_2_4, \
call_3_0_2_5, \
call_3_0_2_6, \
call_3_0_2_7, \
call_3_0_2_8, \
call_3_0_2_9, \
call_3_0_2_10, \
call_3_0_2_11, \
call_3_0_2_12, \
call_3_0_2_13, \
call_3_0_2_14, \
call_3_0_2_15, \
call_3_0_3_0, \
call_3_0_3_1, \
call_3_0_3_2, \
call_3_0_3_3, \
call_3_0_3_4, \
call_3_0_3_5, \
call_3_0_3_6, \
call_3_0_3_7, \
call_3_0_3_8, \
call_3_0_3_9, \
call_3_0_3_10, \
call_3_0_3_11, \
call_3_0_3_12, \
call_3_0_3_13, \
call_3_0_3_14, \
call_3_0_3_15, \
call_3_0_4_0, \
call_3_0_4_1, \
call_3_0_4_2, \
call_3_0_4_3, \
call_3_0_4_4, \
call_3_0_4_5, \
call_3_0_4_6, \
call_3_0_4_7, \
call_3_0_4_8, \
call_3_0_4_9, \
call_3_0_4_10, \
call_3_0_4_11, \
call_3_0_4_12, \
call_3_0_4_13, \
call_3_0_4_14, \
call_3_0_4_15, \
call_3_0_5_0, \
call_3_0_5_1, \
call_3_0_5_2, \
call_3_0_5_3, \
call_3_0_5_4, \
call_3_0_5_5, \
call_3_0_5_6, \
call_3_0_5_7, \
call_3_0_5_8, \
call_3_0_5_9, \
call_3_0_5_10, \
call_3_0_5_11, \
call_3_0_5_12, \
call_3_0_5_13, \
call_3_0_5_14, \
call_3_0_5_15, \
call_3_0_6_0, \
call_3_0_6_1, \
call_3_0_6_2, \
call_3_0_6_3, \
call_3_0_6_4, \
call_3_0_6_5, \
call_3_0_6_6, \
call_3_0_6_7, \
call_3_0_6_8, \
call_3_0_6_9, \
call_3_0_6_10, \
call_3_0_6_11, \
call_3_0_6_12, \
call_3_0_6_13, \
call_3_0_6_14, \
call_3_0_6_15, \
call_3_0_7_0, \
call_3_0_7_1, \
call_3_0_7_2, \
call_3_0_7_3, \
call_3_0_7_4, \
call_3_0_7_5, \
call_3_0_7_6, \
call_3_0_7_7, \
call_3_0_7_8, \
call_3_0_7_9, \
call_3_0_7_10, \
call_3_0_7_11, \
call_3_0_7_12, \
call_3_0_7_13, \
call_3_0_7_14, \
call_3_0_7_15, \
call_3_0_8_0, \
call_3_0_8_1, \
call_3_0_8_2, \
call_3_0_8_3, \
call_3_0_8_4, \
call_3_0_8_5, \
call_3_0_8_6, \
call_3_0_8_7, \
call_3_0_8_8, \
call_3_0_8_9, \
call_3_0_8_10, \
call_3_0_8_11, \
call_3_0_8_12, \
call_3_0_8_13, \
call_3_0_8_14, \
call_3_0_8_15, \
call_3_0_9_0, \
call_3_0_9_1, \
call_3_0_9_2, \
call_3_0_9_3, \
call_3_0_9_4, \
call_3_0_9_5, \
call_3_0_9_6, \
call_3_0_9_7, \
call_3_0_9_8, \
call_3_0_9_9, \
call_3_0_9_10, \
call_3_0_9_11, \
call_3_0_9_12, \
call_3_0_9_13, \
call_3_0_9_14, \
call_3_0_9_15, \
call_3_0_10_0, \
call_3_0_10_1, \
call_3_0_10_2, \
call_3_0_10_3, \
call_3_0_10_4, \
call_3_0_10_5, \
call_3_0_10_6, \
call_3_0_10_7, \
call_3_0_10_8, \
call_3_0_10_9, \
call_3_0_10_10, \
call_3_0_10_11, \
call_3_0_10_12, \
call_3_0_10_13, \
call_3_0_10_14, \
call_3_0_10_15, \
call_3_0_11_0, \
call_3_0_11_1, \
call_3_0_11_2, \
call_3_0_11_3, \
call_3_0_11_4, \
call_3_0_11_5, \
call_3_0_11_6, \
call_3_0_11_7, \
call_3_0_11_8, \
call_3_0_11_9, \
call_3_0_11_10, \
call_3_0_11_11, \
call_3_0_11_12, \
call_3_0_11_13, \
call_3_0_11_14, \
call_3_0_11_15, \
call_3_0_12_0, \
call_3_0_12_1, \
call_3_0_12_2, \
call_3_0_12_3, \
call_3_0_12_4, \
call_3_0_12_5, \
call_3_0_12_6, \
call_3_0_12_7, \
call_3_0_12_8, \
call_3_0_12_9, \
call_3_0_12_10, \
call_3_0_12_11, \
call_3_0_12_12, \
call_3_0_12_13, \
call_3_0_12_14, \
call_3_0_12_15, \
call_3_0_13_0, \
call_3_0_13_1, \
call_3_0_13_2, \
call_3_0_13_3, \
call_3_0_13_4, \
call_3_0_13_5, \
call_3_0_13_6, \
call_3_0_13_7, \
call_3_0_13_8, \
call_3_0_13_9, \
call_3_0_13_10, \
call_3_0_13_11, \
call_3_0_13_12, \
call_3_0_13_13, \
call_3_0_13_14, \
call_3_0_13_15, \
call_3_0_14_0, \
call_3_0_14_1, \
call_3_0_14_2, \
call_3_0_14_3, \
call_3_0_14_4, \
call_3_0_14_5, \
call_3_0_14_6, \
call_3_0_14_7, \
call_3_0_14_8, \
call_3_0_14_9, \
call_3_0_14_10, \
call_3_0_14_11, \
call_3_0_14_12, \
call_3_0_14_13, \
call_3_0_14_14, \
call_3_0_14_15, \
call_3_0_15_0, \
call_3_0_15_1, \
call_3_0_15_2, \
call_3_0_15_3, \
call_3_0_15_4, \
call_3_0_15_5, \
call_3_0_15_6, \
call_3_0_15_7, \
call_3_0_15_8, \
call_3_0_15_9, \
call_3_0_15_10, \
call_3_0_15_11, \
call_3_0_15_12, \
call_3_0_15_13, \
call_3_0_15_14, \
call_3_0_15_15, \
call_3_1_0_0, \
call_3_1_0_1, \
call_3_1_0_2, \
call_3_1_0_3, \
call_3_1_0_4, \
call_3_1_0_5, \
call_3_1_0_6, \
call_3_1_0_7, \
call_3_1_0_8, \
call_3_1_0_9, \
call_3_1_0_10, \
call_3_1_0_11, \
call_3_1_0_12, \
call_3_1_0_13, \
call_3_1_0_14, \
call_3_1_0_15, \
call_3_1_1_0, \
call_3_1_1_1, \
call_3_1_1_2, \
call_3_1_1_3, \
call_3_1_1_4, \
call_3_1_1_5, \
call_3_1_1_6, \
call_3_1_1_7, \
call_3_1_1_8, \
call_3_1_1_9, \
call_3_1_1_10, \
call_3_1_1_11, \
call_3_1_1_12, \
call_3_1_1_13, \
call_3_1_1_14, \
call_3_1_1_15, \
call_3_1_2_0, \
call_3_1_2_1, \
call_3_1_2_2, \
call_3_1_2_3, \
call_3_1_2_4, \
call_3_1_2_5, \
call_3_1_2_6, \
call_3_1_2_7, \
call_3_1_2_8, \
call_3_1_2_9, \
call_3_1_2_10, \
call_3_1_2_11, \
call_3_1_2_12, \
call_3_1_2_13, \
call_3_1_2_14, \
call_3_1_2_15, \
call_3_1_3_0, \
call_3_1_3_1, \
call_3_1_3_2, \
call_3_1_3_3, \
call_3_1_3_4, \
call_3_1_3_5, \
call_3_1_3_6, \
call_3_1_3_7, \
call_3_1_3_8, \
call_3_1_3_9, \
call_3_1_3_10, \
call_3_1_3_11, \
call_3_1_3_12, \
call_3_1_3_13, \
call_3_1_3_14, \
call_3_1_3_15, \
call_3_1_4_0, \
call_3_1_4_1, \
call_3_1_4_2, \
call_3_1_4_3, \
call_3_1_4_4, \
call_3_1_4_5, \
call_3_1_4_6, \
call_3_1_4_7, \
call_3_1_4_8, \
call_3_1_4_9, \
call_3_1_4_10, \
call_3_1_4_11, \
call_3_1_4_12, \
call_3_1_4_13, \
call_3_1_4_14, \
call_3_1_4_15, \
call_3_1_5_0, \
call_3_1_5_1, \
call_3_1_5_2, \
call_3_1_5_3, \
call_3_1_5_4, \
call_3_1_5_5, \
call_3_1_5_6, \
call_3_1_5_7, \
call_3_1_5_8, \
call_3_1_5_9, \
call_3_1_5_10, \
call_3_1_5_11, \
call_3_1_5_12, \
call_3_1_5_13, \
call_3_1_5_14, \
call_3_1_5_15, \
call_3_1_6_0, \
call_3_1_6_1, \
call_3_1_6_2, \
call_3_1_6_3, \
call_3_1_6_4, \
call_3_1_6_5, \
call_3_1_6_6, \
call_3_1_6_7, \
call_3_1_6_8, \
call_3_1_6_9, \
call_3_1_6_10, \
call_3_1_6_11, \
call_3_1_6_12, \
call_3_1_6_13, \
call_3_1_6_14, \
call_3_1_6_15, \
call_3_1_7_0, \
call_3_1_7_1, \
call_3_1_7_2, \
call_3_1_7_3, \
call_3_1_7_4, \
call_3_1_7_5, \
call_3_1_7_6, \
call_3_1_7_7, \
call_3_1_7_8, \
call_3_1_7_9, \
call_3_1_7_10, \
call_3_1_7_11, \
call_3_1_7_12, \
call_3_1_7_13, \
call_3_1_7_14, \
call_3_1_7_15, \
call_3_1_8_0, \
call_3_1_8_1, \
call_3_1_8_2, \
call_3_1_8_3, \
call_3_1_8_4, \
call_3_1_8_5, \
call_3_1_8_6, \
call_3_1_8_7, \
call_3_1_8_8, \
call_3_1_8_9, \
call_3_1_8_10, \
call_3_1_8_11, \
call_3_1_8_12, \
call_3_1_8_13, \
call_3_1_8_14, \
call_3_1_8_15, \
call_3_1_9_0, \
call_3_1_9_1, \
call_3_1_9_2, \
call_3_1_9_3, \
call_3_1_9_4, \
call_3_1_9_5, \
call_3_1_9_6, \
call_3_1_9_7, \
call_3_1_9_8, \
call_3_1_9_9, \
call_3_1_9_10, \
call_3_1_9_11, \
call_3_1_9_12, \
call_3_1_9_13, \
call_3_1_9_14, \
call_3_1_9_15, \
call_3_1_10_0, \
call_3_1_10_1, \
call_3_1_10_2, \
call_3_1_10_3, \
call_3_1_10_4, \
call_3_1_10_5, \
call_3_1_10_6, \
call_3_1_10_7, \
call_3_1_10_8, \
call_3_1_10_9, \
call_3_1_10_10, \
call_3_1_10_11, \
call_3_1_10_12, \
call_3_1_10_13, \
call_3_1_10_14, \
call_3_1_10_15, \
call_3_1_11_0, \
call_3_1_11_1, \
call_3_1_11_2, \
call_3_1_11_3, \
call_3_1_11_4, \
call_3_1_11_5, \
call_3_1_11_6, \
call_3_1_11_7, \
call_3_1_11_8, \
call_3_1_11_9, \
call_3_1_11_10, \
call_3_1_11_11, \
call_3_1_11_12, \
call_3_1_11_13, \
call_3_1_11_14, \
call_3_1_11_15, \
call_3_1_12_0, \
call_3_1_12_1, \
call_3_1_12_2, \
call_3_1_12_3, \
call_3_1_12_4, \
call_3_1_12_5, \
call_3_1_12_6, \
call_3_1_12_7, \
call_3_1_12_8, \
call_3_1_12_9, \
call_3_1_12_10, \
call_3_1_12_11, \
call_3_1_12_12, \
call_3_1_12_13, \
call_3_1_12_14, \
call_3_1_12_15, \
call_3_1_13_0, \
call_3_1_13_1, \
call_3_1_13_2, \
call_3_1_13_3, \
call_3_1_13_4, \
call_3_1_13_5, \
call_3_1_13_6, \
call_3_1_13_7, \
call_3_1_13_8, \
call_3_1_13_9, \
call_3_1_13_10, \
call_3_1_13_11, \
call_3_1_13_12, \
call_3_1_13_13, \
call_3_1_13_14, \
call_3_1_13_15, \
call_3_1_14_0, \
call_3_1_14_1, \
call_3_1_14_2, \
call_3_1_14_3, \
call_3_1_14_4, \
call_3_1_14_5, \
call_3_1_14_6, \
call_3_1_14_7, \
call_3_1_14_8, \
call_3_1_14_9, \
call_3_1_14_10, \
call_3_1_14_11, \
call_3_1_14_12, \
call_3_1_14_13, \
call_3_1_14_14, \
call_3_1_14_15, \
call_3_1_15_0, \
call_3_1_15_1, \
call_3_1_15_2, \
call_3_1_15_3, \
call_3_1_15_4, \
call_3_1_15_5, \
call_3_1_15_6, \
call_3_1_15_7, \
call_3_1_15_8, \
call_3_1_15_9, \
call_3_1_15_10, \
call_3_1_15_11, \
call_3_1_15_12, \
call_3_1_15_13, \
call_3_1_15_14, \
call_3_1_15_15, \
call_3_2_0_0, \
call_3_2_0_1, \
call_3_2_0_2, \
call_3_2_0_3, \
call_3_2_0_4, \
call_3_2_0_5, \
call_3_2_0_6, \
call_3_2_0_7, \
call_3_2_0_8, \
call_3_2_0_9, \
call_3_2_0_10, \
call_3_2_0_11, \
call_3_2_0_12, \
call_3_2_0_13, \
call_3_2_0_14, \
call_3_2_0_15, \
call_3_2_1_0, \
call_3_2_1_1, \
call_3_2_1_2, \
call_3_2_1_3, \
call_3_2_1_4, \
call_3_2_1_5, \
call_3_2_1_6, \
call_3_2_1_7, \
call_3_2_1_8, \
call_3_2_1_9, \
call_3_2_1_10, \
call_3_2_1_11, \
call_3_2_1_12, \
call_3_2_1_13, \
call_3_2_1_14, \
call_3_2_1_15, \
call_3_2_2_0, \
call_3_2_2_1, \
call_3_2_2_2, \
call_3_2_2_3, \
call_3_2_2_4, \
call_3_2_2_5, \
call_3_2_2_6, \
call_3_2_2_7, \
call_3_2_2_8, \
call_3_2_2_9, \
call_3_2_2_10, \
call_3_2_2_11, \
call_3_2_2_12, \
call_3_2_2_13, \
call_3_2_2_14, \
call_3_2_2_15, \
call_3_2_3_0, \
call_3_2_3_1, \
call_3_2_3_2, \
call_3_2_3_3, \
call_3_2_3_4, \
call_3_2_3_5, \
call_3_2_3_6, \
call_3_2_3_7, \
call_3_2_3_8, \
call_3_2_3_9, \
call_3_2_3_10, \
call_3_2_3_11, \
call_3_2_3_12, \
call_3_2_3_13, \
call_3_2_3_14, \
call_3_2_3_15, \
call_3_2_4_0, \
call_3_2_4_1, \
call_3_2_4_2, \
call_3_2_4_3, \
call_3_2_4_4, \
call_3_2_4_5, \
call_3_2_4_6, \
call_3_2_4_7, \
call_3_2_4_8, \
call_3_2_4_9, \
call_3_2_4_10, \
call_3_2_4_11, \
call_3_2_4_12, \
call_3_2_4_13, \
call_3_2_4_14, \
call_3_2_4_15, \
call_3_2_5_0, \
call_3_2_5_1, \
call_3_2_5_2, \
call_3_2_5_3, \
call_3_2_5_4, \
call_3_2_5_5, \
call_3_2_5_6, \
call_3_2_5_7, \
call_3_2_5_8, \
call_3_2_5_9, \
call_3_2_5_10, \
call_3_2_5_11, \
call_3_2_5_12, \
call_3_2_5_13, \
call_3_2_5_14, \
call_3_2_5_15, \
call_3_2_6_0, \
call_3_2_6_1, \
call_3_2_6_2, \
call_3_2_6_3, \
call_3_2_6_4, \
call_3_2_6_5, \
call_3_2_6_6, \
call_3_2_6_7, \
call_3_2_6_8, \
call_3_2_6_9, \
call_3_2_6_10, \
call_3_2_6_11, \
call_3_2_6_12, \
call_3_2_6_13, \
call_3_2_6_14, \
call_3_2_6_15, \
call_3_2_7_0, \
call_3_2_7_1, \
call_3_2_7_2, \
call_3_2_7_3, \
call_3_2_7_4, \
call_3_2_7_5, \
call_3_2_7_6, \
call_3_2_7_7, \
call_3_2_7_8, \
call_3_2_7_9, \
call_3_2_7_10, \
call_3_2_7_11, \
call_3_2_7_12, \
call_3_2_7_13, \
call_3_2_7_14, \
call_3_2_7_15, \
call_3_2_8_0, \
call_3_2_8_1, \
call_3_2_8_2, \
call_3_2_8_3, \
call_3_2_8_4, \
call_3_2_8_5, \
call_3_2_8_6, \
call_3_2_8_7, \
call_3_2_8_8, \
call_3_2_8_9, \
call_3_2_8_10, \
call_3_2_8_11, \
call_3_2_8_12, \
call_3_2_8_13, \
call_3_2_8_14, \
call_3_2_8_15, \
call_3_2_9_0, \
call_3_2_9_1, \
call_3_2_9_2, \
call_3_2_9_3, \
call_3_2_9_4, \
call_3_2_9_5, \
call_3_2_9_6, \
call_3_2_9_7, \
call_3_2_9_8, \
call_3_2_9_9, \
call_3_2_9_10, \
call_3_2_9_11, \
call_3_2_9_12, \
call_3_2_9_13, \
call_3_2_9_14, \
call_3_2_9_15, \
call_3_2_10_0, \
call_3_2_10_1, \
call_3_2_10_2, \
call_3_2_10_3, \
call_3_2_10_4, \
call_3_2_10_5, \
call_3_2_10_6, \
call_3_2_10_7, \
call_3_2_10_8, \
call_3_2_10_9, \
call_3_2_10_10, \
call_3_2_10_11, \
call_3_2_10_12, \
call_3_2_10_13, \
call_3_2_10_14, \
call_3_2_10_15, \
call_3_2_11_0, \
call_3_2_11_1, \
call_3_2_11_2, \
call_3_2_11_3, \
call_3_2_11_4, \
call_3_2_11_5, \
call_3_2_11_6, \
call_3_2_11_7, \
call_3_2_11_8, \
call_3_2_11_9, \
call_3_2_11_10, \
call_3_2_11_11, \
call_3_2_11_12, \
call_3_2_11_13, \
call_3_2_11_14, \
call_3_2_11_15, \
call_3_2_12_0, \
call_3_2_12_1, \
call_3_2_12_2, \
call_3_2_12_3, \
call_3_2_12_4, \
call_3_2_12_5, \
call_3_2_12_6, \
call_3_2_12_7, \
call_3_2_12_8, \
call_3_2_12_9, \
call_3_2_12_10, \
call_3_2_12_11, \
call_3_2_12_12, \
call_3_2_12_13, \
call_3_2_12_14, \
call_3_2_12_15, \
call_3_2_13_0, \
call_3_2_13_1, \
call_3_2_13_2, \
call_3_2_13_3, \
call_3_2_13_4, \
call_3_2_13_5, \
call_3_2_13_6, \
call_3_2_13_7, \
call_3_2_13_8, \
call_3_2_13_9, \
call_3_2_13_10, \
call_3_2_13_11, \
call_3_2_13_12, \
call_3_2_13_13, \
call_3_2_13_14, \
call_3_2_13_15, \
call_3_2_14_0, \
call_3_2_14_1, \
call_3_2_14_2, \
call_3_2_14_3, \
call_3_2_14_4, \
call_3_2_14_5, \
call_3_2_14_6, \
call_3_2_14_7, \
call_3_2_14_8, \
call_3_2_14_9, \
call_3_2_14_10, \
call_3_2_14_11, \
call_3_2_14_12, \
call_3_2_14_13, \
call_3_2_14_14, \
call_3_2_14_15, \
call_3_2_15_0, \
call_3_2_15_1, \
call_3_2_15_2, \
call_3_2_15_3, \
call_3_2_15_4, \
call_3_2_15_5, \
call_3_2_15_6, \
call_3_2_15_7, \
call_3_2_15_8, \
call_3_2_15_9, \
call_3_2_15_10, \
call_3_2_15_11, \
call_3_2_15_12, \
call_3_2_15_13, \
call_3_2_15_14, \
call_3_2_15_15, \
call_3_3_0_0, \
call_3_3_0_1, \
call_3_3_0_2, \
call_3_3_0_3, \
call_3_3_0_4, \
call_3_3_0_5, \
call_3_3_0_6, \
call_3_3_0_7, \
call_3_3_0_8, \
call_3_3_0_9, \
call_3_3_0_10, \
call_3_3_0_11, \
call_3_3_0_12, \
call_3_3_0_13, \
call_3_3_0_14, \
call_3_3_0_15, \
call_3_3_1_0, \
call_3_3_1_1, \
call_3_3_1_2, \
call_3_3_1_3, \
call_3_3_1_4, \
call_3_3_1_5, \
call_3_3_1_6, \
call_3_3_1_7, \
call_3_3_1_8, \
call_3_3_1_9, \
call_3_3_1_10, \
call_3_3_1_11, \
call_3_3_1_12, \
call_3_3_1_13, \
call_3_3_1_14, \
call_3_3_1_15, \
call_3_3_2_0, \
call_3_3_2_1, \
call_3_3_2_2, \
call_3_3_2_3, \
call_3_3_2_4, \
call_3_3_2_5, \
call_3_3_2_6, \
call_3_3_2_7, \
call_3_3_2_8, \
call_3_3_2_9, \
call_3_3_2_10, \
call_3_3_2_11, \
call_3_3_2_12, \
call_3_3_2_13, \
call_3_3_2_14, \
call_3_3_2_15, \
call_3_3_3_0, \
call_3_3_3_1, \
call_3_3_3_2, \
call_3_3_3_3, \
call_3_3_3_4, \
call_3_3_3_5, \
call_3_3_3_6, \
call_3_3_3_7, \
call_3_3_3_8, \
call_3_3_3_9, \
call_3_3_3_10, \
call_3_3_3_11, \
call_3_3_3_12, \
call_3_3_3_13, \
call_3_3_3_14, \
call_3_3_3_15, \
call_3_3_4_0, \
call_3_3_4_1, \
call_3_3_4_2, \
call_3_3_4_3, \
call_3_3_4_4, \
call_3_3_4_5, \
call_3_3_4_6, \
call_3_3_4_7, \
call_3_3_4_8, \
call_3_3_4_9, \
call_3_3_4_10, \
call_3_3_4_11, \
call_3_3_4_12, \
call_3_3_4_13, \
call_3_3_4_14, \
call_3_3_4_15, \
call_3_3_5_0, \
call_3_3_5_1, \
call_3_3_5_2, \
call_3_3_5_3, \
call_3_3_5_4, \
call_3_3_5_5, \
call_3_3_5_6, \
call_3_3_5_7, \
call_3_3_5_8, \
call_3_3_5_9, \
call_3_3_5_10, \
call_3_3_5_11, \
call_3_3_5_12, \
call_3_3_5_13, \
call_3_3_5_14, \
call_3_3_5_15, \
call_3_3_6_0, \
call_3_3_6_1, \
call_3_3_6_2, \
call_3_3_6_3, \
call_3_3_6_4, \
call_3_3_6_5, \
call_3_3_6_6, \
call_3_3_6_7, \
call_3_3_6_8, \
call_3_3_6_9, \
call_3_3_6_10, \
call_3_3_6_11, \
call_3_3_6_12, \
call_3_3_6_13, \
call_3_3_6_14, \
call_3_3_6_15, \
call_3_3_7_0, \
call_3_3_7_1, \
call_3_3_7_2, \
call_3_3_7_3, \
call_3_3_7_4, \
call_3_3_7_5, \
call_3_3_7_6, \
call_3_3_7_7, \
call_3_3_7_8, \
call_3_3_7_9, \
call_3_3_7_10, \
call_3_3_7_11, \
call_3_3_7_12, \
call_3_3_7_13, \
call_3_3_7_14, \
call_3_3_7_15, \
call_3_3_8_0, \
call_3_3_8_1, \
call_3_3_8_2, \
call_3_3_8_3, \
call_3_3_8_4, \
call_3_3_8_5, \
call_3_3_8_6, \
call_3_3_8_7, \
call_3_3_8_8, \
call_3_3_8_9, \
call_3_3_8_10, \
call_3_3_8_11, \
call_3_3_8_12, \
call_3_3_8_13, \
call_3_3_8_14, \
call_3_3_8_15, \
call_3_3_9_0, \
call_3_3_9_1, \
call_3_3_9_2, \
call_3_3_9_3, \
call_3_3_9_4, \
call_3_3_9_5, \
call_3_3_9_6, \
call_3_3_9_7, \
call_3_3_9_8, \
call_3_3_9_9, \
call_3_3_9_10, \
call_3_3_9_11, \
call_3_3_9_12, \
call_3_3_9_13, \
call_3_3_9_14, \
call_3_3_9_15, \
call_3_3_10_0, \
call_3_3_10_1, \
call_3_3_10_2, \
call_3_3_10_3, \
call_3_3_10_4, \
call_3_3_10_5, \
call_3_3_10_6, \
call_3_3_10_7, \
call_3_3_10_8, \
call_3_3_10_9, \
call_3_3_10_10, \
call_3_3_10_11, \
call_3_3_10_12, \
call_3_3_10_13, \
call_3_3_10_14, \
call_3_3_10_15, \
call_3_3_11_0, \
call_3_3_11_1, \
call_3_3_11_2, \
call_3_3_11_3, \
call_3_3_11_4, \
call_3_3_11_5, \
call_3_3_11_6, \
call_3_3_11_7, \
call_3_3_11_8, \
call_3_3_11_9, \
call_3_3_11_10, \
call_3_3_11_11, \
call_3_3_11_12, \
call_3_3_11_13, \
call_3_3_11_14, \
call_3_3_11_15, \
call_3_3_12_0, \
call_3_3_12_1, \
call_3_3_12_2, \
call_3_3_12_3, \
call_3_3_12_4, \
call_3_3_12_5, \
call_3_3_12_6, \
call_3_3_12_7, \
call_3_3_12_8, \
call_3_3_12_9, \
call_3_3_12_10, \
call_3_3_12_11, \
call_3_3_12_12, \
call_3_3_12_13, \
call_3_3_12_14, \
call_3_3_12_15, \
call_3_3_13_0, \
call_3_3_13_1, \
call_3_3_13_2, \
call_3_3_13_3, \
call_3_3_13_4, \
call_3_3_13_5, \
call_3_3_13_6, \
call_3_3_13_7, \
call_3_3_13_8, \
call_3_3_13_9, \
call_3_3_13_10, \
call_3_3_13_11, \
call_3_3_13_12, \
call_3_3_13_13, \
call_3_3_13_14, \
call_3_3_13_15, \
call_3_3_14_0, \
call_3_3_14_1, \
call_3_3_14_2, \
call_3_3_14_3, \
call_3_3_14_4, \
call_3_3_14_5, \
call_3_3_14_6, \
call_3_3_14_7, \
call_3_3_14_8, \
call_3_3_14_9, \
call_3_3_14_10, \
call_3_3_14_11, \
call_3_3_14_12, \
call_3_3_14_13, \
call_3_3_14_14, \
call_3_3_14_15, \
call_3_3_15_0, \
call_3_3_15_1, \
call_3_3_15_2, \
call_3_3_15_3, \
call_3_3_15_4, \
call_3_3_15_5, \
call_3_3_15_6, \
call_3_3_15_7, \
call_3_3_15_8, \
call_3_3_15_9, \
call_3_3_15_10, \
call_3_3_15_11, \
call_3_3_15_12, \
call_3_3_15_13, \
call_3_3_15_14, \
call_3_3_15_15, \
call_3_4_0_0, \
call_3_4_0_1, \
call_3_4_0_2, \
call_3_4_0_3, \
call_3_4_0_4, \
call_3_4_0_5, \
call_3_4_0_6, \
call_3_4_0_7, \
call_3_4_0_8, \
call_3_4_0_9, \
call_3_4_0_10, \
call_3_4_0_11, \
call_3_4_0_12, \
call_3_4_0_13, \
call_3_4_0_14, \
call_3_4_0_15, \
call_3_4_1_0, \
call_3_4_1_1, \
call_3_4_1_2, \
call_3_4_1_3, \
call_3_4_1_4, \
call_3_4_1_5, \
call_3_4_1_6, \
call_3_4_1_7, \
call_3_4_1_8, \
call_3_4_1_9, \
call_3_4_1_10, \
call_3_4_1_11, \
call_3_4_1_12, \
call_3_4_1_13, \
call_3_4_1_14, \
call_3_4_1_15, \
call_3_4_2_0, \
call_3_4_2_1, \
call_3_4_2_2, \
call_3_4_2_3, \
call_3_4_2_4, \
call_3_4_2_5, \
call_3_4_2_6, \
call_3_4_2_7, \
call_3_4_2_8, \
call_3_4_2_9, \
call_3_4_2_10, \
call_3_4_2_11, \
call_3_4_2_12, \
call_3_4_2_13, \
call_3_4_2_14, \
call_3_4_2_15, \
call_3_4_3_0, \
call_3_4_3_1, \
call_3_4_3_2, \
call_3_4_3_3, \
call_3_4_3_4, \
call_3_4_3_5, \
call_3_4_3_6, \
call_3_4_3_7, \
call_3_4_3_8, \
call_3_4_3_9, \
call_3_4_3_10, \
call_3_4_3_11, \
call_3_4_3_12, \
call_3_4_3_13, \
call_3_4_3_14, \
call_3_4_3_15, \
call_3_4_4_0, \
call_3_4_4_1, \
call_3_4_4_2, \
call_3_4_4_3, \
call_3_4_4_4, \
call_3_4_4_5, \
call_3_4_4_6, \
call_3_4_4_7, \
call_3_4_4_8, \
call_3_4_4_9, \
call_3_4_4_10, \
call_3_4_4_11, \
call_3_4_4_12, \
call_3_4_4_13, \
call_3_4_4_14, \
call_3_4_4_15, \
call_3_4_5_0, \
call_3_4_5_1, \
call_3_4_5_2, \
call_3_4_5_3, \
call_3_4_5_4, \
call_3_4_5_5, \
call_3_4_5_6, \
call_3_4_5_7, \
call_3_4_5_8, \
call_3_4_5_9, \
call_3_4_5_10, \
call_3_4_5_11, \
call_3_4_5_12, \
call_3_4_5_13, \
call_3_4_5_14, \
call_3_4_5_15, \
call_3_4_6_0, \
call_3_4_6_1, \
call_3_4_6_2, \
call_3_4_6_3, \
call_3_4_6_4, \
call_3_4_6_5, \
call_3_4_6_6, \
call_3_4_6_7, \
call_3_4_6_8, \
call_3_4_6_9, \
call_3_4_6_10, \
call_3_4_6_11, \
call_3_4_6_12, \
call_3_4_6_13, \
call_3_4_6_14, \
call_3_4_6_15, \
call_3_4_7_0, \
call_3_4_7_1, \
call_3_4_7_2, \
call_3_4_7_3, \
call_3_4_7_4, \
call_3_4_7_5, \
call_3_4_7_6, \
call_3_4_7_7, \
call_3_4_7_8, \
call_3_4_7_9, \
call_3_4_7_10, \
call_3_4_7_11, \
call_3_4_7_12, \
call_3_4_7_13, \
call_3_4_7_14, \
call_3_4_7_15, \
call_3_4_8_0, \
call_3_4_8_1, \
call_3_4_8_2, \
call_3_4_8_3, \
call_3_4_8_4, \
call_3_4_8_5, \
call_3_4_8_6, \
call_3_4_8_7, \
call_3_4_8_8, \
call_3_4_8_9, \
call_3_4_8_10, \
call_3_4_8_11, \
call_3_4_8_12, \
call_3_4_8_13, \
call_3_4_8_14, \
call_3_4_8_15, \
call_3_4_9_0, \
call_3_4_9_1, \
call_3_4_9_2, \
call_3_4_9_3, \
call_3_4_9_4, \
call_3_4_9_5, \
call_3_4_9_6, \
call_3_4_9_7, \
call_3_4_9_8, \
call_3_4_9_9, \
call_3_4_9_10, \
call_3_4_9_11, \
call_3_4_9_12, \
call_3_4_9_13, \
call_3_4_9_14, \
call_3_4_9_15, \
call_3_4_10_0, \
call_3_4_10_1, \
call_3_4_10_2, \
call_3_4_10_3, \
call_3_4_10_4, \
call_3_4_10_5, \
call_3_4_10_6, \
call_3_4_10_7, \
call_3_4_10_8, \
call_3_4_10_9, \
call_3_4_10_10, \
call_3_4_10_11, \
call_3_4_10_12, \
call_3_4_10_13, \
call_3_4_10_14, \
call_3_4_10_15, \
call_3_4_11_0, \
call_3_4_11_1, \
call_3_4_11_2, \
call_3_4_11_3, \
call_3_4_11_4, \
call_3_4_11_5, \
call_3_4_11_6, \
call_3_4_11_7, \
call_3_4_11_8, \
call_3_4_11_9, \
call_3_4_11_10, \
call_3_4_11_11, \
call_3_4_11_12, \
call_3_4_11_13, \
call_3_4_11_14, \
call_3_4_11_15, \
call_3_4_12_0, \
call_3_4_12_1, \
call_3_4_12_2, \
call_3_4_12_3, \
call_3_4_12_4, \
call_3_4_12_5, \
call_3_4_12_6, \
call_3_4_12_7, \
call_3_4_12_8, \
call_3_4_12_9, \
call_3_4_12_10, \
call_3_4_12_11, \
call_3_4_12_12, \
call_3_4_12_13, \
call_3_4_12_14, \
call_3_4_12_15, \
call_3_4_13_0, \
call_3_4_13_1, \
call_3_4_13_2, \
call_3_4_13_3, \
call_3_4_13_4, \
call_3_4_13_5, \
call_3_4_13_6, \
call_3_4_13_7, \
call_3_4_13_8, \
call_3_4_13_9, \
call_3_4_13_10, \
call_3_4_13_11, \
call_3_4_13_12, \
call_3_4_13_13, \
call_3_4_13_14, \
call_3_4_13_15, \
call_3_4_14_0, \
call_3_4_14_1, \
call_3_4_14_2, \
call_3_4_14_3, \
call_3_4_14_4, \
call_3_4_14_5, \
call_3_4_14_6, \
call_3_4_14_7, \
call_3_4_14_8, \
call_3_4_14_9, \
call_3_4_14_10, \
call_3_4_14_11, \
call_3_4_14_12, \
call_3_4_14_13, \
call_3_4_14_14, \
call_3_4_14_15, \
call_3_4_15_0, \
call_3_4_15_1, \
call_3_4_15_2, \
call_3_4_15_3, \
call_3_4_15_4, \
call_3_4_15_5, \
call_3_4_15_6, \
call_3_4_15_7, \
call_3_4_15_8, \
call_3_4_15_9, \
call_3_4_15_10, \
call_3_4_15_11, \
call_3_4_15_12, \
call_3_4_15_13, \
call_3_4_15_14, \
call_3_4_15_15, \
call_3_5_0_0, \
call_3_5_0_1, \
call_3_5_0_2, \
call_3_5_0_3, \
call_3_5_0_4, \
call_3_5_0_5, \
call_3_5_0_6, \
call_3_5_0_7, \
call_3_5_0_8, \
call_3_5_0_9, \
call_3_5_0_10, \
call_3_5_0_11, \
call_3_5_0_12, \
call_3_5_0_13, \
call_3_5_0_14, \
call_3_5_0_15, \
call_3_5_1_0, \
call_3_5_1_1, \
call_3_5_1_2, \
call_3_5_1_3, \
call_3_5_1_4, \
call_3_5_1_5, \
call_3_5_1_6, \
call_3_5_1_7, \
call_3_5_1_8, \
call_3_5_1_9, \
call_3_5_1_10, \
call_3_5_1_11, \
call_3_5_1_12, \
call_3_5_1_13, \
call_3_5_1_14, \
call_3_5_1_15, \
call_3_5_2_0, \
call_3_5_2_1, \
call_3_5_2_2, \
call_3_5_2_3, \
call_3_5_2_4, \
call_3_5_2_5, \
call_3_5_2_6, \
call_3_5_2_7, \
call_3_5_2_8, \
call_3_5_2_9, \
call_3_5_2_10, \
call_3_5_2_11, \
call_3_5_2_12, \
call_3_5_2_13, \
call_3_5_2_14, \
call_3_5_2_15, \
call_3_5_3_0, \
call_3_5_3_1, \
call_3_5_3_2, \
call_3_5_3_3, \
call_3_5_3_4, \
call_3_5_3_5, \
call_3_5_3_6, \
call_3_5_3_7, \
call_3_5_3_8, \
call_3_5_3_9, \
call_3_5_3_10, \
call_3_5_3_11, \
call_3_5_3_12, \
call_3_5_3_13, \
call_3_5_3_14, \
call_3_5_3_15, \
call_3_5_4_0, \
call_3_5_4_1, \
call_3_5_4_2, \
call_3_5_4_3, \
call_3_5_4_4, \
call_3_5_4_5, \
call_3_5_4_6, \
call_3_5_4_7, \
call_3_5_4_8, \
call_3_5_4_9, \
call_3_5_4_10, \
call_3_5_4_11, \
call_3_5_4_12, \
call_3_5_4_13, \
call_3_5_4_14, \
call_3_5_4_15, \
call_3_5_5_0, \
call_3_5_5_1, \
call_3_5_5_2, \
call_3_5_5_3, \
call_3_5_5_4, \
call_3_5_5_5, \
call_3_5_5_6, \
call_3_5_5_7, \
call_3_5_5_8, \
call_3_5_5_9, \
call_3_5_5_10, \
call_3_5_5_11, \
call_3_5_5_12, \
call_3_5_5_13, \
call_3_5_5_14, \
call_3_5_5_15, \
call_3_5_6_0, \
call_3_5_6_1, \
call_3_5_6_2, \
call_3_5_6_3, \
call_3_5_6_4, \
call_3_5_6_5, \
call_3_5_6_6, \
call_3_5_6_7, \
call_3_5_6_8, \
call_3_5_6_9, \
call_3_5_6_10, \
call_3_5_6_11, \
call_3_5_6_12, \
call_3_5_6_13, \
call_3_5_6_14, \
call_3_5_6_15, \
call_3_5_7_0, \
call_3_5_7_1, \
call_3_5_7_2, \
call_3_5_7_3, \
call_3_5_7_4, \
call_3_5_7_5, \
call_3_5_7_6, \
call_3_5_7_7, \
call_3_5_7_8, \
call_3_5_7_9, \
call_3_5_7_10, \
call_3_5_7_11, \
call_3_5_7_12, \
call_3_5_7_13, \
call_3_5_7_14, \
call_3_5_7_15, \
call_3_5_8_0, \
call_3_5_8_1, \
call_3_5_8_2, \
call_3_5_8_3, \
call_3_5_8_4, \
call_3_5_8_5, \
call_3_5_8_6, \
call_3_5_8_7, \
call_3_5_8_8, \
call_3_5_8_9, \
call_3_5_8_10, \
call_3_5_8_11, \
call_3_5_8_12, \
call_3_5_8_13, \
call_3_5_8_14, \
call_3_5_8_15, \
call_3_5_9_0, \
call_3_5_9_1, \
call_3_5_9_2, \
call_3_5_9_3, \
call_3_5_9_4, \
call_3_5_9_5, \
call_3_5_9_6, \
call_3_5_9_7, \
call_3_5_9_8, \
call_3_5_9_9, \
call_3_5_9_10, \
call_3_5_9_11, \
call_3_5_9_12, \
call_3_5_9_13, \
call_3_5_9_14, \
call_3_5_9_15, \
call_3_5_10_0, \
call_3_5_10_1, \
call_3_5_10_2, \
call_3_5_10_3, \
call_3_5_10_4, \
call_3_5_10_5, \
call_3_5_10_6, \
call_3_5_10_7, \
call_3_5_10_8, \
call_3_5_10_9, \
call_3_5_10_10, \
call_3_5_10_11, \
call_3_5_10_12, \
call_3_5_10_13, \
call_3_5_10_14, \
call_3_5_10_15, \
call_3_5_11_0, \
call_3_5_11_1, \
call_3_5_11_2, \
call_3_5_11_3, \
call_3_5_11_4, \
call_3_5_11_5, \
call_3_5_11_6, \
call_3_5_11_7, \
call_3_5_11_8, \
call_3_5_11_9, \
call_3_5_11_10, \
call_3_5_11_11, \
call_3_5_11_12, \
call_3_5_11_13, \
call_3_5_11_14, \
call_3_5_11_15, \
call_3_5_12_0, \
call_3_5_12_1, \
call_3_5_12_2, \
call_3_5_12_3, \
call_3_5_12_4, \
call_3_5_12_5, \
call_3_5_12_6, \
call_3_5_12_7, \
call_3_5_12_8, \
call_3_5_12_9, \
call_3_5_12_10, \
call_3_5_12_11, \
call_3_5_12_12, \
call_3_5_12_13, \
call_3_5_12_14, \
call_3_5_12_15, \
call_3_5_13_0, \
call_3_5_13_1, \
call_3_5_13_2, \
call_3_5_13_3, \
call_3_5_13_4, \
call_3_5_13_5, \
call_3_5_13_6, \
call_3_5_13_7, \
call_3_5_13_8, \
call_3_5_13_9, \
call_3_5_13_10, \
call_3_5_13_11, \
call_3_5_13_12, \
call_3_5_13_13, \
call_3_5_13_14, \
call_3_5_13_15, \
call_3_5_14_0, \
call_3_5_14_1, \
call_3_5_14_2, \
call_3_5_14_3, \
call_3_5_14_4, \
call_3_5_14_5, \
call_3_5_14_6, \
call_3_5_14_7, \
call_3_5_14_8, \
call_3_5_14_9, \
call_3_5_14_10, \
call_3_5_14_11, \
call_3_5_14_12, \
call_3_5_14_13, \
call_3_5_14_14, \
call_3_5_14_15, \
call_3_5_15_0, \
call_3_5_15_1, \
call_3_5_15_2, \
call_3_5_15_3, \
call_3_5_15_4, \
call_3_5_15_5, \
call_3_5_15_6, \
call_3_5_15_7, \
call_3_5_15_8, \
call_3_5_15_9, \
call_3_5_15_10, \
call_3_5_15_11, \
call_3_5_15_12, \
call_3_5_15_13, \
call_3_5_15_14, \
call_3_5_15_15, \
call_3_6_0_0, \
call_3_6_0_1, \
call_3_6_0_2, \
call_3_6_0_3, \
call_3_6_0_4, \
call_3_6_0_5, \
call_3_6_0_6, \
call_3_6_0_7, \
call_3_6_0_8, \
call_3_6_0_9, \
call_3_6_0_10, \
call_3_6_0_11, \
call_3_6_0_12, \
call_3_6_0_13, \
call_3_6_0_14, \
call_3_6_0_15, \
call_3_6_1_0, \
call_3_6_1_1, \
call_3_6_1_2, \
call_3_6_1_3, \
call_3_6_1_4, \
call_3_6_1_5, \
call_3_6_1_6, \
call_3_6_1_7, \
call_3_6_1_8, \
call_3_6_1_9, \
call_3_6_1_10, \
call_3_6_1_11, \
call_3_6_1_12, \
call_3_6_1_13, \
call_3_6_1_14, \
call_3_6_1_15, \
call_3_6_2_0, \
call_3_6_2_1, \
call_3_6_2_2, \
call_3_6_2_3, \
call_3_6_2_4, \
call_3_6_2_5, \
call_3_6_2_6, \
call_3_6_2_7, \
call_3_6_2_8, \
call_3_6_2_9, \
call_3_6_2_10, \
call_3_6_2_11, \
call_3_6_2_12, \
call_3_6_2_13, \
call_3_6_2_14, \
call_3_6_2_15, \
call_3_6_3_0, \
call_3_6_3_1, \
call_3_6_3_2, \
call_3_6_3_3, \
call_3_6_3_4, \
call_3_6_3_5, \
call_3_6_3_6, \
call_3_6_3_7, \
call_3_6_3_8, \
call_3_6_3_9, \
call_3_6_3_10, \
call_3_6_3_11, \
call_3_6_3_12, \
call_3_6_3_13, \
call_3_6_3_14, \
call_3_6_3_15, \
call_3_6_4_0, \
call_3_6_4_1, \
call_3_6_4_2, \
call_3_6_4_3, \
call_3_6_4_4, \
call_3_6_4_5, \
call_3_6_4_6, \
call_3_6_4_7, \
call_3_6_4_8, \
call_3_6_4_9, \
call_3_6_4_10, \
call_3_6_4_11, \
call_3_6_4_12, \
call_3_6_4_13, \
call_3_6_4_14, \
call_3_6_4_15, \
call_3_6_5_0, \
call_3_6_5_1, \
call_3_6_5_2, \
call_3_6_5_3, \
call_3_6_5_4, \
call_3_6_5_5, \
call_3_6_5_6, \
call_3_6_5_7, \
call_3_6_5_8, \
call_3_6_5_9, \
call_3_6_5_10, \
call_3_6_5_11, \
call_3_6_5_12, \
call_3_6_5_13, \
call_3_6_5_14, \
call_3_6_5_15, \
call_3_6_6_0, \
call_3_6_6_1, \
call_3_6_6_2, \
call_3_6_6_3, \
call_3_6_6_4, \
call_3_6_6_5, \
call_3_6_6_6, \
call_3_6_6_7, \
call_3_6_6_8, \
call_3_6_6_9, \
call_3_6_6_10, \
call_3_6_6_11, \
call_3_6_6_12, \
call_3_6_6_13, \
call_3_6_6_14, \
call_3_6_6_15, \
call_3_6_7_0, \
call_3_6_7_1, \
call_3_6_7_2, \
call_3_6_7_3, \
call_3_6_7_4, \
call_3_6_7_5, \
call_3_6_7_6, \
call_3_6_7_7, \
call_3_6_7_8, \
call_3_6_7_9, \
call_3_6_7_10, \
call_3_6_7_11, \
call_3_6_7_12, \
call_3_6_7_13, \
call_3_6_7_14, \
call_3_6_7_15, \
call_3_6_8_0, \
call_3_6_8_1, \
call_3_6_8_2, \
call_3_6_8_3, \
call_3_6_8_4, \
call_3_6_8_5, \
call_3_6_8_6, \
call_3_6_8_7, \
call_3_6_8_8, \
call_3_6_8_9, \
call_3_6_8_10, \
call_3_6_8_11, \
call_3_6_8_12, \
call_3_6_8_13, \
call_3_6_8_14, \
call_3_6_8_15, \
call_3_6_9_0, \
call_3_6_9_1, \
call_3_6_9_2, \
call_3_6_9_3, \
call_3_6_9_4, \
call_3_6_9_5, \
call_3_6_9_6, \
call_3_6_9_7, \
call_3_6_9_8, \
call_3_6_9_9, \
call_3_6_9_10, \
call_3_6_9_11, \
call_3_6_9_12, \
call_3_6_9_13, \
call_3_6_9_14, \
call_3_6_9_15, \
call_3_6_10_0, \
call_3_6_10_1, \
call_3_6_10_2, \
call_3_6_10_3, \
call_3_6_10_4, \
call_3_6_10_5, \
call_3_6_10_6, \
call_3_6_10_7, \
call_3_6_10_8, \
call_3_6_10_9, \
call_3_6_10_10, \
call_3_6_10_11, \
call_3_6_10_12, \
call_3_6_10_13, \
call_3_6_10_14, \
call_3_6_10_15, \
call_3_6_11_0, \
call_3_6_11_1, \
call_3_6_11_2, \
call_3_6_11_3, \
call_3_6_11_4, \
call_3_6_11_5, \
call_3_6_11_6, \
call_3_6_11_7, \
call_3_6_11_8, \
call_3_6_11_9, \
call_3_6_11_10, \
call_3_6_11_11, \
call_3_6_11_12, \
call_3_6_11_13, \
call_3_6_11_14, \
call_3_6_11_15, \
call_3_6_12_0, \
call_3_6_12_1, \
call_3_6_12_2, \
call_3_6_12_3, \
call_3_6_12_4, \
call_3_6_12_5, \
call_3_6_12_6, \
call_3_6_12_7, \
call_3_6_12_8, \
call_3_6_12_9, \
call_3_6_12_10, \
call_3_6_12_11, \
call_3_6_12_12, \
call_3_6_12_13, \
call_3_6_12_14, \
call_3_6_12_15, \
call_3_6_13_0, \
call_3_6_13_1, \
call_3_6_13_2, \
call_3_6_13_3, \
call_3_6_13_4, \
call_3_6_13_5, \
call_3_6_13_6, \
call_3_6_13_7, \
call_3_6_13_8, \
call_3_6_13_9, \
call_3_6_13_10, \
call_3_6_13_11, \
call_3_6_13_12, \
call_3_6_13_13, \
call_3_6_13_14, \
call_3_6_13_15, \
call_3_6_14_0, \
call_3_6_14_1, \
call_3_6_14_2, \
call_3_6_14_3, \
call_3_6_14_4, \
call_3_6_14_5, \
call_3_6_14_6, \
call_3_6_14_7, \
call_3_6_14_8, \
call_3_6_14_9, \
call_3_6_14_10, \
call_3_6_14_11, \
call_3_6_14_12, \
call_3_6_14_13, \
call_3_6_14_14, \
call_3_6_14_15, \
call_3_6_15_0, \
call_3_6_15_1, \
call_3_6_15_2, \
call_3_6_15_3, \
call_3_6_15_4, \
call_3_6_15_5, \
call_3_6_15_6, \
call_3_6_15_7, \
call_3_6_15_8, \
call_3_6_15_9, \
call_3_6_15_10, \
call_3_6_15_11, \
call_3_6_15_12, \
call_3_6_15_13, \
call_3_6_15_14, \
call_3_6_15_15, \
call_3_7_0_0, \
call_3_7_0_1, \
call_3_7_0_2, \
call_3_7_0_3, \
call_3_7_0_4, \
call_3_7_0_5, \
call_3_7_0_6, \
call_3_7_0_7, \
call_3_7_0_8, \
call_3_7_0_9, \
call_3_7_0_10, \
call_3_7_0_11, \
call_3_7_0_12, \
call_3_7_0_13, \
call_3_7_0_14, \
call_3_7_0_15, \
call_3_7_1_0, \
call_3_7_1_1, \
call_3_7_1_2, \
call_3_7_1_3, \
call_3_7_1_4, \
call_3_7_1_5, \
call_3_7_1_6, \
call_3_7_1_7, \
call_3_7_1_8, \
call_3_7_1_9, \
call_3_7_1_10, \
call_3_7_1_11, \
call_3_7_1_12, \
call_3_7_1_13, \
call_3_7_1_14, \
call_3_7_1_15, \
call_3_7_2_0, \
call_3_7_2_1, \
call_3_7_2_2, \
call_3_7_2_3, \
call_3_7_2_4, \
call_3_7_2_5, \
call_3_7_2_6, \
call_3_7_2_7, \
call_3_7_2_8, \
call_3_7_2_9, \
call_3_7_2_10, \
call_3_7_2_11, \
call_3_7_2_12, \
call_3_7_2_13, \
call_3_7_2_14, \
call_3_7_2_15, \
call_3_7_3_0, \
call_3_7_3_1, \
call_3_7_3_2, \
call_3_7_3_3, \
call_3_7_3_4, \
call_3_7_3_5, \
call_3_7_3_6, \
call_3_7_3_7, \
call_3_7_3_8, \
call_3_7_3_9, \
call_3_7_3_10, \
call_3_7_3_11, \
call_3_7_3_12, \
call_3_7_3_13, \
call_3_7_3_14, \
call_3_7_3_15, \
call_3_7_4_0, \
call_3_7_4_1, \
call_3_7_4_2, \
call_3_7_4_3, \
call_3_7_4_4, \
call_3_7_4_5, \
call_3_7_4_6, \
call_3_7_4_7, \
call_3_7_4_8, \
call_3_7_4_9, \
call_3_7_4_10, \
call_3_7_4_11, \
call_3_7_4_12, \
call_3_7_4_13, \
call_3_7_4_14, \
call_3_7_4_15, \
call_3_7_5_0, \
call_3_7_5_1, \
call_3_7_5_2, \
call_3_7_5_3, \
call_3_7_5_4, \
call_3_7_5_5, \
call_3_7_5_6, \
call_3_7_5_7, \
call_3_7_5_8, \
call_3_7_5_9, \
call_3_7_5_10, \
call_3_7_5_11, \
call_3_7_5_12, \
call_3_7_5_13, \
call_3_7_5_14, \
call_3_7_5_15, \
call_3_7_6_0, \
call_3_7_6_1, \
call_3_7_6_2, \
call_3_7_6_3, \
call_3_7_6_4, \
call_3_7_6_5, \
call_3_7_6_6, \
call_3_7_6_7, \
call_3_7_6_8, \
call_3_7_6_9, \
call_3_7_6_10, \
call_3_7_6_11, \
call_3_7_6_12, \
call_3_7_6_13, \
call_3_7_6_14, \
call_3_7_6_15, \
call_3_7_7_0, \
call_3_7_7_1, \
call_3_7_7_2, \
call_3_7_7_3, \
call_3_7_7_4, \
call_3_7_7_5, \
call_3_7_7_6, \
call_3_7_7_7, \
call_3_7_7_8, \
call_3_7_7_9, \
call_3_7_7_10, \
call_3_7_7_11, \
call_3_7_7_12, \
call_3_7_7_13, \
call_3_7_7_14, \
call_3_7_7_15, \
call_3_7_8_0, \
call_3_7_8_1, \
call_3_7_8_2, \
call_3_7_8_3, \
call_3_7_8_4, \
call_3_7_8_5, \
call_3_7_8_6, \
call_3_7_8_7, \
call_3_7_8_8, \
call_3_7_8_9, \
call_3_7_8_10, \
call_3_7_8_11, \
call_3_7_8_12, \
call_3_7_8_13, \
call_3_7_8_14, \
call_3_7_8_15, \
call_3_7_9_0, \
call_3_7_9_1, \
call_3_7_9_2, \
call_3_7_9_3, \
call_3_7_9_4, \
call_3_7_9_5, \
call_3_7_9_6, \
call_3_7_9_7, \
call_3_7_9_8, \
call_3_7_9_9, \
call_3_7_9_10, \
call_3_7_9_11, \
call_3_7_9_12, \
call_3_7_9_13, \
call_3_7_9_14, \
call_3_7_9_15, \
call_3_7_10_0, \
call_3_7_10_1, \
call_3_7_10_2, \
call_3_7_10_3, \
call_3_7_10_4, \
call_3_7_10_5, \
call_3_7_10_6, \
call_3_7_10_7, \
call_3_7_10_8, \
call_3_7_10_9, \
call_3_7_10_10, \
call_3_7_10_11, \
call_3_7_10_12, \
call_3_7_10_13, \
call_3_7_10_14, \
call_3_7_10_15, \
call_3_7_11_0, \
call_3_7_11_1, \
call_3_7_11_2, \
call_3_7_11_3, \
call_3_7_11_4, \
call_3_7_11_5, \
call_3_7_11_6, \
call_3_7_11_7, \
call_3_7_11_8, \
call_3_7_11_9, \
call_3_7_11_10, \
call_3_7_11_11, \
call_3_7_11_12, \
call_3_7_11_13, \
call_3_7_11_14, \
call_3_7_11_15, \
call_3_7_12_0, \
call_3_7_12_1, \
call_3_7_12_2, \
call_3_7_12_3, \
call_3_7_12_4, \
call_3_7_12_5, \
call_3_7_12_6, \
call_3_7_12_7, \
call_3_7_12_8, \
call_3_7_12_9, \
call_3_7_12_10, \
call_3_7_12_11, \
call_3_7_12_12, \
call_3_7_12_13, \
call_3_7_12_14, \
call_3_7_12_15, \
call_3_7_13_0, \
call_3_7_13_1, \
call_3_7_13_2, \
call_3_7_13_3, \
call_3_7_13_4, \
call_3_7_13_5, \
call_3_7_13_6, \
call_3_7_13_7, \
call_3_7_13_8, \
call_3_7_13_9, \
call_3_7_13_10, \
call_3_7_13_11, \
call_3_7_13_12, \
call_3_7_13_13, \
call_3_7_13_14, \
call_3_7_13_15, \
call_3_7_14_0, \
call_3_7_14_1, \
call_3_7_14_2, \
call_3_7_14_3, \
call_3_7_14_4, \
call_3_7_14_5, \
call_3_7_14_6, \
call_3_7_14_7, \
call_3_7_14_8, \
call_3_7_14_9, \
call_3_7_14_10, \
call_3_7_14_11, \
call_3_7_14_12, \
call_3_7_14_13, \
call_3_7_14_14, \
call_3_7_14_15, \
call_3_7_15_0, \
call_3_7_15_1, \
call_3_7_15_2, \
call_3_7_15_3, \
call_3_7_15_4, \
call_3_7_15_5, \
call_3_7_15_6, \
call_3_7_15_7, \
call_3_7_15_8, \
call_3_7_15_9, \
call_3_7_15_10, \
call_3_7_15_11, \
call_3_7_15_12, \
call_3_7_15_13, \
call_3_7_15_14, \
call_3_7_15_15, \
call_3_8_0_0, \
call_3_8_0_1, \
call_3_8_0_2, \
call_3_8_0_3, \
call_3_8_0_4, \
call_3_8_0_5, \
call_3_8_0_6, \
call_3_8_0_7, \
call_3_8_0_8, \
call_3_8_0_9, \
call_3_8_0_10, \
call_3_8_0_11, \
call_3_8_0_12, \
call_3_8_0_13, \
call_3_8_0_14, \
call_3_8_0_15, \
call_3_8_1_0, \
call_3_8_1_1, \
call_3_8_1_2, \
call_3_8_1_3, \
call_3_8_1_4, \
call_3_8_1_5, \
call_3_8_1_6, \
call_3_8_1_7, \
call_3_8_1_8, \
call_3_8_1_9, \
call_3_8_1_10, \
call_3_8_1_11, \
call_3_8_1_12, \
call_3_8_1_13, \
call_3_8_1_14, \
call_3_8_1_15, \
call_3_8_2_0, \
call_3_8_2_1, \
call_3_8_2_2, \
call_3_8_2_3, \
call_3_8_2_4, \
call_3_8_2_5, \
call_3_8_2_6, \
call_3_8_2_7, \
call_3_8_2_8, \
call_3_8_2_9, \
call_3_8_2_10, \
call_3_8_2_11, \
call_3_8_2_12, \
call_3_8_2_13, \
call_3_8_2_14, \
call_3_8_2_15, \
call_3_8_3_0, \
call_3_8_3_1, \
call_3_8_3_2, \
call_3_8_3_3, \
call_3_8_3_4, \
call_3_8_3_5, \
call_3_8_3_6, \
call_3_8_3_7, \
call_3_8_3_8, \
call_3_8_3_9, \
call_3_8_3_10, \
call_3_8_3_11, \
call_3_8_3_12, \
call_3_8_3_13, \
call_3_8_3_14, \
call_3_8_3_15, \
call_3_8_4_0, \
call_3_8_4_1, \
call_3_8_4_2, \
call_3_8_4_3, \
call_3_8_4_4, \
call_3_8_4_5, \
call_3_8_4_6, \
call_3_8_4_7, \
call_3_8_4_8, \
call_3_8_4_9, \
call_3_8_4_10, \
call_3_8_4_11, \
call_3_8_4_12, \
call_3_8_4_13, \
call_3_8_4_14, \
call_3_8_4_15, \
call_3_8_5_0, \
call_3_8_5_1, \
call_3_8_5_2, \
call_3_8_5_3, \
call_3_8_5_4, \
call_3_8_5_5, \
call_3_8_5_6, \
call_3_8_5_7, \
call_3_8_5_8, \
call_3_8_5_9, \
call_3_8_5_10, \
call_3_8_5_11, \
call_3_8_5_12, \
call_3_8_5_13, \
call_3_8_5_14, \
call_3_8_5_15, \
call_3_8_6_0, \
call_3_8_6_1, \
call_3_8_6_2, \
call_3_8_6_3, \
call_3_8_6_4, \
call_3_8_6_5, \
call_3_8_6_6, \
call_3_8_6_7, \
call_3_8_6_8, \
call_3_8_6_9, \
call_3_8_6_10, \
call_3_8_6_11, \
call_3_8_6_12, \
call_3_8_6_13, \
call_3_8_6_14, \
call_3_8_6_15, \
call_3_8_7_0, \
call_3_8_7_1, \
call_3_8_7_2, \
call_3_8_7_3, \
call_3_8_7_4, \
call_3_8_7_5, \
call_3_8_7_6, \
call_3_8_7_7, \
call_3_8_7_8, \
call_3_8_7_9, \
call_3_8_7_10, \
call_3_8_7_11, \
call_3_8_7_12, \
call_3_8_7_13, \
call_3_8_7_14, \
call_3_8_7_15, \
call_3_8_8_0, \
call_3_8_8_1, \
call_3_8_8_2, \
call_3_8_8_3, \
call_3_8_8_4, \
call_3_8_8_5, \
call_3_8_8_6, \
call_3_8_8_7, \
call_3_8_8_8, \
call_3_8_8_9, \
call_3_8_8_10, \
call_3_8_8_11, \
call_3_8_8_12, \
call_3_8_8_13, \
call_3_8_8_14, \
call_3_8_8_15, \
call_3_8_9_0, \
call_3_8_9_1, \
call_3_8_9_2, \
call_3_8_9_3, \
call_3_8_9_4, \
call_3_8_9_5, \
call_3_8_9_6, \
call_3_8_9_7, \
call_3_8_9_8, \
call_3_8_9_9, \
call_3_8_9_10, \
call_3_8_9_11, \
call_3_8_9_12, \
call_3_8_9_13, \
call_3_8_9_14, \
call_3_8_9_15, \
call_3_8_10_0, \
call_3_8_10_1, \
call_3_8_10_2, \
call_3_8_10_3, \
call_3_8_10_4, \
call_3_8_10_5, \
call_3_8_10_6, \
call_3_8_10_7, \
call_3_8_10_8, \
call_3_8_10_9, \
call_3_8_10_10, \
call_3_8_10_11, \
call_3_8_10_12, \
call_3_8_10_13, \
call_3_8_10_14, \
call_3_8_10_15, \
call_3_8_11_0, \
call_3_8_11_1, \
call_3_8_11_2, \
call_3_8_11_3, \
call_3_8_11_4, \
call_3_8_11_5, \
call_3_8_11_6, \
call_3_8_11_7, \
call_3_8_11_8, \
call_3_8_11_9, \
call_3_8_11_10, \
call_3_8_11_11, \
call_3_8_11_12, \
call_3_8_11_13, \
call_3_8_11_14, \
call_3_8_11_15, \
call_3_8_12_0, \
call_3_8_12_1, \
call_3_8_12_2, \
call_3_8_12_3, \
call_3_8_12_4, \
call_3_8_12_5, \
call_3_8_12_6, \
call_3_8_12_7, \
call_3_8_12_8, \
call_3_8_12_9, \
call_3_8_12_10, \
call_3_8_12_11, \
call_3_8_12_12, \
call_3_8_12_13, \
call_3_8_12_14, \
call_3_8_12_15, \
call_3_8_13_0, \
call_3_8_13_1, \
call_3_8_13_2, \
call_3_8_13_3, \
call_3_8_13_4, \
call_3_8_13_5, \
call_3_8_13_6, \
call_3_8_13_7, \
call_3_8_13_8, \
call_3_8_13_9, \
call_3_8_13_10, \
call_3_8_13_11, \
call_3_8_13_12, \
call_3_8_13_13, \
call_3_8_13_14, \
call_3_8_13_15, \
call_3_8_14_0, \
call_3_8_14_1, \
call_3_8_14_2, \
call_3_8_14_3, \
call_3_8_14_4, \
call_3_8_14_5, \
call_3_8_14_6, \
call_3_8_14_7, \
call_3_8_14_8, \
call_3_8_14_9, \
call_3_8_14_10, \
call_3_8_14_11, \
call_3_8_14_12, \
call_3_8_14_13, \
call_3_8_14_14, \
call_3_8_14_15, \
call_3_8_15_0, \
call_3_8_15_1, \
call_3_8_15_2, \
call_3_8_15_3, \
call_3_8_15_4, \
call_3_8_15_5, \
call_3_8_15_6, \
call_3_8_15_7, \
call_3_8_15_8, \
call_3_8_15_9, \
call_3_8_15_10, \
call_3_8_15_11, \
call_3_8_15_12, \
call_3_8_15_13, \
call_3_8_15_14, \
call_3_8_15_15, \
call_3_9_0_0, \
call_3_9_0_1, \
call_3_9_0_2, \
call_3_9_0_3, \
call_3_9_0_4, \
call_3_9_0_5, \
call_3_9_0_6, \
call_3_9_0_7, \
call_3_9_0_8, \
call_3_9_0_9, \
call_3_9_0_10, \
call_3_9_0_11, \
call_3_9_0_12, \
call_3_9_0_13, \
call_3_9_0_14, \
call_3_9_0_15, \
call_3_9_1_0, \
call_3_9_1_1, \
call_3_9_1_2, \
call_3_9_1_3, \
call_3_9_1_4, \
call_3_9_1_5, \
call_3_9_1_6, \
call_3_9_1_7, \
call_3_9_1_8, \
call_3_9_1_9, \
call_3_9_1_10, \
call_3_9_1_11, \
call_3_9_1_12, \
call_3_9_1_13, \
call_3_9_1_14, \
call_3_9_1_15, \
call_3_9_2_0, \
call_3_9_2_1, \
call_3_9_2_2, \
call_3_9_2_3, \
call_3_9_2_4, \
call_3_9_2_5, \
call_3_9_2_6, \
call_3_9_2_7, \
call_3_9_2_8, \
call_3_9_2_9, \
call_3_9_2_10, \
call_3_9_2_11, \
call_3_9_2_12, \
call_3_9_2_13, \
call_3_9_2_14, \
call_3_9_2_15, \
call_3_9_3_0, \
call_3_9_3_1, \
call_3_9_3_2, \
call_3_9_3_3, \
call_3_9_3_4, \
call_3_9_3_5, \
call_3_9_3_6, \
call_3_9_3_7, \
call_3_9_3_8, \
call_3_9_3_9, \
call_3_9_3_10, \
call_3_9_3_11, \
call_3_9_3_12, \
call_3_9_3_13, \
call_3_9_3_14, \
call_3_9_3_15, \
call_3_9_4_0, \
call_3_9_4_1, \
call_3_9_4_2, \
call_3_9_4_3, \
call_3_9_4_4, \
call_3_9_4_5, \
call_3_9_4_6, \
call_3_9_4_7, \
call_3_9_4_8, \
call_3_9_4_9, \
call_3_9_4_10, \
call_3_9_4_11, \
call_3_9_4_12, \
call_3_9_4_13, \
call_3_9_4_14, \
call_3_9_4_15, \
call_3_9_5_0, \
call_3_9_5_1, \
call_3_9_5_2, \
call_3_9_5_3, \
call_3_9_5_4, \
call_3_9_5_5, \
call_3_9_5_6, \
call_3_9_5_7, \
call_3_9_5_8, \
call_3_9_5_9, \
call_3_9_5_10, \
call_3_9_5_11, \
call_3_9_5_12, \
call_3_9_5_13, \
call_3_9_5_14, \
call_3_9_5_15, \
call_3_9_6_0, \
call_3_9_6_1, \
call_3_9_6_2, \
call_3_9_6_3, \
call_3_9_6_4, \
call_3_9_6_5, \
call_3_9_6_6, \
call_3_9_6_7, \
call_3_9_6_8, \
call_3_9_6_9, \
call_3_9_6_10, \
call_3_9_6_11, \
call_3_9_6_12, \
call_3_9_6_13, \
call_3_9_6_14, \
call_3_9_6_15, \
call_3_9_7_0, \
call_3_9_7_1, \
call_3_9_7_2, \
call_3_9_7_3, \
call_3_9_7_4, \
call_3_9_7_5, \
call_3_9_7_6, \
call_3_9_7_7, \
call_3_9_7_8, \
call_3_9_7_9, \
call_3_9_7_10, \
call_3_9_7_11, \
call_3_9_7_12, \
call_3_9_7_13, \
call_3_9_7_14, \
call_3_9_7_15, \
call_3_9_8_0, \
call_3_9_8_1, \
call_3_9_8_2, \
call_3_9_8_3, \
call_3_9_8_4, \
call_3_9_8_5, \
call_3_9_8_6, \
call_3_9_8_7, \
call_3_9_8_8, \
call_3_9_8_9, \
call_3_9_8_10, \
call_3_9_8_11, \
call_3_9_8_12, \
call_3_9_8_13, \
call_3_9_8_14, \
call_3_9_8_15, \
call_3_9_9_0, \
call_3_9_9_1, \
call_3_9_9_2, \
call_3_9_9_3, \
call_3_9_9_4, \
call_3_9_9_5, \
call_3_9_9_6, \
call_3_9_9_7, \
call_3_9_9_8, \
call_3_9_9_9, \
call_3_9_9_10, \
call_3_9_9_11, \
call_3_9_9_12, \
call_3_9_9_13, \
call_3_9_9_14, \
call_3_9_9_15, \
call_3_9_10_0, \
call_3_9_10_1, \
call_3_9_10_2, \
call_3_9_10_3, \
call_3_9_10_4, \
call_3_9_10_5, \
call_3_9_10_6, \
call_3_9_10_7, \
call_3_9_10_8, \
call_3_9_10_9, \
call_3_9_10_10, \
call_3_9_10_11, \
call_3_9_10_12, \
call_3_9_10_13, \
call_3_9_10_14, \
call_3_9_10_15, \
call_3_9_11_0, \
call_3_9_11_1, \
call_3_9_11_2, \
call_3_9_11_3, \
call_3_9_11_4, \
call_3_9_11_5, \
call_3_9_11_6, \
call_3_9_11_7, \
call_3_9_11_8, \
call_3_9_11_9, \
call_3_9_11_10, \
call_3_9_11_11, \
call_3_9_11_12, \
call_3_9_11_13, \
call_3_9_11_14, \
call_3_9_11_15, \
call_3_9_12_0, \
call_3_9_12_1, \
call_3_9_12_2, \
call_3_9_12_3, \
call_3_9_12_4, \
call_3_9_12_5, \
call_3_9_12_6, \
call_3_9_12_7, \
call_3_9_12_8, \
call_3_9_12_9, \
call_3_9_12_10, \
call_3_9_12_11, \
call_3_9_12_12, \
call_3_9_12_13, \
call_3_9_12_14, \
call_3_9_12_15, \
call_3_9_13_0, \
call_3_9_13_1, \
call_3_9_13_2, \
call_3_9_13_3, \
call_3_9_13_4, \
call_3_9_13_5, \
call_3_9_13_6, \
call_3_9_13_7, \
call_3_9_13_8, \
call_3_9_13_9, \
call_3_9_13_10, \
call_3_9_13_11, \
call_3_9_13_12, \
call_3_9_13_13, \
call_3_9_13_14, \
call_3_9_13_15, \
call_3_9_14_0, \
call_3_9_14_1, \
call_3_9_14_2, \
call_3_9_14_3, \
call_3_9_14_4, \
call_3_9_14_5, \
call_3_9_14_6, \
call_3_9_14_7, \
call_3_9_14_8, \
call_3_9_14_9, \
call_3_9_14_10, \
call_3_9_14_11, \
call_3_9_14_12, \
call_3_9_14_13, \
call_3_9_14_14, \
call_3_9_14_15, \
call_3_9_15_0, \
call_3_9_15_1, \
call_3_9_15_2, \
call_3_9_15_3, \
call_3_9_15_4, \
call_3_9_15_5, \
call_3_9_15_6, \
call_3_9_15_7, \
call_3_9_15_8, \
call_3_9_15_9, \
call_3_9_15_10, \
call_3_9_15_11, \
call_3_9_15_12, \
call_3_9_15_13, \
call_3_9_15_14, \
call_3_9_15_15, \
call_3_10_0_0, \
call_3_10_0_1, \
call_3_10_0_2, \
call_3_10_0_3, \
call_3_10_0_4, \
call_3_10_0_5, \
call_3_10_0_6, \
call_3_10_0_7, \
call_3_10_0_8, \
call_3_10_0_9, \
call_3_10_0_10, \
call_3_10_0_11, \
call_3_10_0_12, \
call_3_10_0_13, \
call_3_10_0_14, \
call_3_10_0_15, \
call_3_10_1_0, \
call_3_10_1_1, \
call_3_10_1_2, \
call_3_10_1_3, \
call_3_10_1_4, \
call_3_10_1_5, \
call_3_10_1_6, \
call_3_10_1_7, \
call_3_10_1_8, \
call_3_10_1_9, \
call_3_10_1_10, \
call_3_10_1_11, \
call_3_10_1_12, \
call_3_10_1_13, \
call_3_10_1_14, \
call_3_10_1_15, \
call_3_10_2_0, \
call_3_10_2_1, \
call_3_10_2_2, \
call_3_10_2_3, \
call_3_10_2_4, \
call_3_10_2_5, \
call_3_10_2_6, \
call_3_10_2_7, \
call_3_10_2_8, \
call_3_10_2_9, \
call_3_10_2_10, \
call_3_10_2_11, \
call_3_10_2_12, \
call_3_10_2_13, \
call_3_10_2_14, \
call_3_10_2_15, \
call_3_10_3_0, \
call_3_10_3_1, \
call_3_10_3_2, \
call_3_10_3_3, \
call_3_10_3_4, \
call_3_10_3_5, \
call_3_10_3_6, \
call_3_10_3_7, \
call_3_10_3_8, \
call_3_10_3_9, \
call_3_10_3_10, \
call_3_10_3_11, \
call_3_10_3_12, \
call_3_10_3_13, \
call_3_10_3_14, \
call_3_10_3_15, \
call_3_10_4_0, \
call_3_10_4_1, \
call_3_10_4_2, \
call_3_10_4_3, \
call_3_10_4_4, \
call_3_10_4_5, \
call_3_10_4_6, \
call_3_10_4_7, \
call_3_10_4_8, \
call_3_10_4_9, \
call_3_10_4_10, \
call_3_10_4_11, \
call_3_10_4_12, \
call_3_10_4_13, \
call_3_10_4_14, \
call_3_10_4_15, \
call_3_10_5_0, \
call_3_10_5_1, \
call_3_10_5_2, \
call_3_10_5_3, \
call_3_10_5_4, \
call_3_10_5_5, \
call_3_10_5_6, \
call_3_10_5_7, \
call_3_10_5_8, \
call_3_10_5_9, \
call_3_10_5_10, \
call_3_10_5_11, \
call_3_10_5_12, \
call_3_10_5_13, \
call_3_10_5_14, \
call_3_10_5_15, \
call_3_10_6_0, \
call_3_10_6_1, \
call_3_10_6_2, \
call_3_10_6_3, \
call_3_10_6_4, \
call_3_10_6_5, \
call_3_10_6_6, \
call_3_10_6_7, \
call_3_10_6_8, \
call_3_10_6_9, \
call_3_10_6_10, \
call_3_10_6_11, \
call_3_10_6_12, \
call_3_10_6_13, \
call_3_10_6_14, \
call_3_10_6_15, \
call_3_10_7_0, \
call_3_10_7_1, \
call_3_10_7_2, \
call_3_10_7_3, \
call_3_10_7_4, \
call_3_10_7_5, \
call_3_10_7_6, \
call_3_10_7_7, \
call_3_10_7_8, \
call_3_10_7_9, \
call_3_10_7_10, \
call_3_10_7_11, \
call_3_10_7_12, \
call_3_10_7_13, \
call_3_10_7_14, \
call_3_10_7_15, \
call_3_10_8_0, \
call_3_10_8_1, \
call_3_10_8_2, \
call_3_10_8_3, \
call_3_10_8_4, \
call_3_10_8_5, \
call_3_10_8_6, \
call_3_10_8_7, \
call_3_10_8_8, \
call_3_10_8_9, \
call_3_10_8_10, \
call_3_10_8_11, \
call_3_10_8_12, \
call_3_10_8_13, \
call_3_10_8_14, \
call_3_10_8_15, \
call_3_10_9_0, \
call_3_10_9_1, \
call_3_10_9_2, \
call_3_10_9_3, \
call_3_10_9_4, \
call_3_10_9_5, \
call_3_10_9_6, \
call_3_10_9_7, \
call_3_10_9_8, \
call_3_10_9_9, \
call_3_10_9_10, \
call_3_10_9_11, \
call_3_10_9_12, \
call_3_10_9_13, \
call_3_10_9_14, \
call_3_10_9_15, \
call_3_10_10_0, \
call_3_10_10_1, \
call_3_10_10_2, \
call_3_10_10_3, \
call_3_10_10_4, \
call_3_10_10_5, \
call_3_10_10_6, \
call_3_10_10_7, \
call_3_10_10_8, \
call_3_10_10_9, \
call_3_10_10_10, \
call_3_10_10_11, \
call_3_10_10_12, \
call_3_10_10_13, \
call_3_10_10_14, \
call_3_10_10_15, \
call_3_10_11_0, \
call_3_10_11_1, \
call_3_10_11_2, \
call_3_10_11_3, \
call_3_10_11_4, \
call_3_10_11_5, \
call_3_10_11_6, \
call_3_10_11_7, \
call_3_10_11_8, \
call_3_10_11_9, \
call_3_10_11_10, \
call_3_10_11_11, \
call_3_10_11_12, \
call_3_10_11_13, \
call_3_10_11_14, \
call_3_10_11_15, \
call_3_10_12_0, \
call_3_10_12_1, \
call_3_10_12_2, \
call_3_10_12_3, \
call_3_10_12_4, \
call_3_10_12_5, \
call_3_10_12_6, \
call_3_10_12_7, \
call_3_10_12_8, \
call_3_10_12_9, \
call_3_10_12_10, \
call_3_10_12_11, \
call_3_10_12_12, \
call_3_10_12_13, \
call_3_10_12_14, \
call_3_10_12_15, \
call_3_10_13_0, \
call_3_10_13_1, \
call_3_10_13_2, \
call_3_10_13_3, \
call_3_10_13_4, \
call_3_10_13_5, \
call_3_10_13_6, \
call_3_10_13_7, \
call_3_10_13_8, \
call_3_10_13_9, \
call_3_10_13_10, \
call_3_10_13_11, \
call_3_10_13_12, \
call_3_10_13_13, \
call_3_10_13_14, \
call_3_10_13_15, \
call_3_10_14_0, \
call_3_10_14_1, \
call_3_10_14_2, \
call_3_10_14_3, \
call_3_10_14_4, \
call_3_10_14_5, \
call_3_10_14_6, \
call_3_10_14_7, \
call_3_10_14_8, \
call_3_10_14_9, \
call_3_10_14_10, \
call_3_10_14_11, \
call_3_10_14_12, \
call_3_10_14_13, \
call_3_10_14_14, \
call_3_10_14_15, \
call_3_10_15_0, \
call_3_10_15_1, \
call_3_10_15_2, \
call_3_10_15_3, \
call_3_10_15_4, \
call_3_10_15_5, \
call_3_10_15_6, \
call_3_10_15_7, \
call_3_10_15_8, \
call_3_10_15_9, \
call_3_10_15_10, \
call_3_10_15_11, \
call_3_10_15_12, \
call_3_10_15_13, \
call_3_10_15_14, \
call_3_10_15_15, \
call_3_11_0_0, \
call_3_11_0_1, \
call_3_11_0_2, \
call_3_11_0_3, \
call_3_11_0_4, \
call_3_11_0_5, \
call_3_11_0_6, \
call_3_11_0_7, \
call_3_11_0_8, \
call_3_11_0_9, \
call_3_11_0_10, \
call_3_11_0_11, \
call_3_11_0_12, \
call_3_11_0_13, \
call_3_11_0_14, \
call_3_11_0_15, \
call_3_11_1_0, \
call_3_11_1_1, \
call_3_11_1_2, \
call_3_11_1_3, \
call_3_11_1_4, \
call_3_11_1_5, \
call_3_11_1_6, \
call_3_11_1_7, \
call_3_11_1_8, \
call_3_11_1_9, \
call_3_11_1_10, \
call_3_11_1_11, \
call_3_11_1_12, \
call_3_11_1_13, \
call_3_11_1_14, \
call_3_11_1_15, \
call_3_11_2_0, \
call_3_11_2_1, \
call_3_11_2_2, \
call_3_11_2_3, \
call_3_11_2_4, \
call_3_11_2_5, \
call_3_11_2_6, \
call_3_11_2_7, \
call_3_11_2_8, \
call_3_11_2_9, \
call_3_11_2_10, \
call_3_11_2_11, \
call_3_11_2_12, \
call_3_11_2_13, \
call_3_11_2_14, \
call_3_11_2_15, \
call_3_11_3_0, \
call_3_11_3_1, \
call_3_11_3_2, \
call_3_11_3_3, \
call_3_11_3_4, \
call_3_11_3_5, \
call_3_11_3_6, \
call_3_11_3_7, \
call_3_11_3_8, \
call_3_11_3_9, \
call_3_11_3_10, \
call_3_11_3_11, \
call_3_11_3_12, \
call_3_11_3_13, \
call_3_11_3_14, \
call_3_11_3_15, \
call_3_11_4_0, \
call_3_11_4_1, \
call_3_11_4_2, \
call_3_11_4_3, \
call_3_11_4_4, \
call_3_11_4_5, \
call_3_11_4_6, \
call_3_11_4_7, \
call_3_11_4_8, \
call_3_11_4_9, \
call_3_11_4_10, \
call_3_11_4_11, \
call_3_11_4_12, \
call_3_11_4_13, \
call_3_11_4_14, \
call_3_11_4_15, \
call_3_11_5_0, \
call_3_11_5_1, \
call_3_11_5_2, \
call_3_11_5_3, \
call_3_11_5_4, \
call_3_11_5_5, \
call_3_11_5_6, \
call_3_11_5_7, \
call_3_11_5_8, \
call_3_11_5_9, \
call_3_11_5_10, \
call_3_11_5_11, \
call_3_11_5_12, \
call_3_11_5_13, \
call_3_11_5_14, \
call_3_11_5_15, \
call_3_11_6_0, \
call_3_11_6_1, \
call_3_11_6_2, \
call_3_11_6_3, \
call_3_11_6_4, \
call_3_11_6_5, \
call_3_11_6_6, \
call_3_11_6_7, \
call_3_11_6_8, \
call_3_11_6_9, \
call_3_11_6_10, \
call_3_11_6_11, \
call_3_11_6_12, \
call_3_11_6_13, \
call_3_11_6_14, \
call_3_11_6_15, \
call_3_11_7_0, \
call_3_11_7_1, \
call_3_11_7_2, \
call_3_11_7_3, \
call_3_11_7_4, \
call_3_11_7_5, \
call_3_11_7_6, \
call_3_11_7_7, \
call_3_11_7_8, \
call_3_11_7_9, \
call_3_11_7_10, \
call_3_11_7_11, \
call_3_11_7_12, \
call_3_11_7_13, \
call_3_11_7_14, \
call_3_11_7_15, \
call_3_11_8_0, \
call_3_11_8_1, \
call_3_11_8_2, \
call_3_11_8_3, \
call_3_11_8_4, \
call_3_11_8_5, \
call_3_11_8_6, \
call_3_11_8_7, \
call_3_11_8_8, \
call_3_11_8_9, \
call_3_11_8_10, \
call_3_11_8_11, \
call_3_11_8_12, \
call_3_11_8_13, \
call_3_11_8_14, \
call_3_11_8_15, \
call_3_11_9_0, \
call_3_11_9_1, \
call_3_11_9_2, \
call_3_11_9_3, \
call_3_11_9_4, \
call_3_11_9_5, \
call_3_11_9_6, \
call_3_11_9_7, \
call_3_11_9_8, \
call_3_11_9_9, \
call_3_11_9_10, \
call_3_11_9_11, \
call_3_11_9_12, \
call_3_11_9_13, \
call_3_11_9_14, \
call_3_11_9_15, \
call_3_11_10_0, \
call_3_11_10_1, \
call_3_11_10_2, \
call_3_11_10_3, \
call_3_11_10_4, \
call_3_11_10_5, \
call_3_11_10_6, \
call_3_11_10_7, \
call_3_11_10_8, \
call_3_11_10_9, \
call_3_11_10_10, \
call_3_11_10_11, \
call_3_11_10_12, \
call_3_11_10_13, \
call_3_11_10_14, \
call_3_11_10_15, \
call_3_11_11_0, \
call_3_11_11_1, \
call_3_11_11_2, \
call_3_11_11_3, \
call_3_11_11_4, \
call_3_11_11_5, \
call_3_11_11_6, \
call_3_11_11_7, \
call_3_11_11_8, \
call_3_11_11_9, \
call_3_11_11_10, \
call_3_11_11_11, \
call_3_11_11_12, \
call_3_11_11_13, \
call_3_11_11_14, \
call_3_11_11_15, \
call_3_11_12_0, \
call_3_11_12_1, \
call_3_11_12_2, \
call_3_11_12_3, \
call_3_11_12_4, \
call_3_11_12_5, \
call_3_11_12_6, \
call_3_11_12_7, \
call_3_11_12_8, \
call_3_11_12_9, \
call_3_11_12_10, \
call_3_11_12_11, \
call_3_11_12_12, \
call_3_11_12_13, \
call_3_11_12_14, \
call_3_11_12_15, \
call_3_11_13_0, \
call_3_11_13_1, \
call_3_11_13_2, \
call_3_11_13_3, \
call_3_11_13_4, \
call_3_11_13_5, \
call_3_11_13_6, \
call_3_11_13_7, \
call_3_11_13_8, \
call_3_11_13_9, \
call_3_11_13_10, \
call_3_11_13_11, \
call_3_11_13_12, \
call_3_11_13_13, \
call_3_11_13_14, \
call_3_11_13_15, \
call_3_11_14_0, \
call_3_11_14_1, \
call_3_11_14_2, \
call_3_11_14_3, \
call_3_11_14_4, \
call_3_11_14_5, \
call_3_11_14_6, \
call_3_11_14_7, \
call_3_11_14_8, \
call_3_11_14_9, \
call_3_11_14_10, \
call_3_11_14_11, \
call_3_11_14_12, \
call_3_11_14_13, \
call_3_11_14_14, \
call_3_11_14_15, \
call_3_11_15_0, \
call_3_11_15_1, \
call_3_11_15_2, \
call_3_11_15_3, \
call_3_11_15_4, \
call_3_11_15_5, \
call_3_11_15_6, \
call_3_11_15_7, \
call_3_11_15_8, \
call_3_11_15_9, \
call_3_11_15_10, \
call_3_11_15_11, \
call_3_11_15_12, \
call_3_11_15_13, \
call_3_11_15_14, \
call_3_11_15_15, \
call_3_12_0_0, \
call_3_12_0_1, \
call_3_12_0_2, \
call_3_12_0_3, \
call_3_12_0_4, \
call_3_12_0_5, \
call_3_12_0_6, \
call_3_12_0_7, \
call_3_12_0_8, \
call_3_12_0_9, \
call_3_12_0_10, \
call_3_12_0_11, \
call_3_12_0_12, \
call_3_12_0_13, \
call_3_12_0_14, \
call_3_12_0_15, \
call_3_12_1_0, \
call_3_12_1_1, \
call_3_12_1_2, \
call_3_12_1_3, \
call_3_12_1_4, \
call_3_12_1_5, \
call_3_12_1_6, \
call_3_12_1_7, \
call_3_12_1_8, \
call_3_12_1_9, \
call_3_12_1_10, \
call_3_12_1_11, \
call_3_12_1_12, \
call_3_12_1_13, \
call_3_12_1_14, \
call_3_12_1_15, \
call_3_12_2_0, \
call_3_12_2_1, \
call_3_12_2_2, \
call_3_12_2_3, \
call_3_12_2_4, \
call_3_12_2_5, \
call_3_12_2_6, \
call_3_12_2_7, \
call_3_12_2_8, \
call_3_12_2_9, \
call_3_12_2_10, \
call_3_12_2_11, \
call_3_12_2_12, \
call_3_12_2_13, \
call_3_12_2_14, \
call_3_12_2_15, \
call_3_12_3_0, \
call_3_12_3_1, \
call_3_12_3_2, \
call_3_12_3_3, \
call_3_12_3_4, \
call_3_12_3_5, \
call_3_12_3_6, \
call_3_12_3_7, \
call_3_12_3_8, \
call_3_12_3_9, \
call_3_12_3_10, \
call_3_12_3_11, \
call_3_12_3_12, \
call_3_12_3_13, \
call_3_12_3_14, \
call_3_12_3_15, \
call_3_12_4_0, \
call_3_12_4_1, \
call_3_12_4_2, \
call_3_12_4_3, \
call_3_12_4_4, \
call_3_12_4_5, \
call_3_12_4_6, \
call_3_12_4_7, \
call_3_12_4_8, \
call_3_12_4_9, \
call_3_12_4_10, \
call_3_12_4_11, \
call_3_12_4_12, \
call_3_12_4_13, \
call_3_12_4_14, \
call_3_12_4_15, \
call_3_12_5_0, \
call_3_12_5_1, \
call_3_12_5_2, \
call_3_12_5_3, \
call_3_12_5_4, \
call_3_12_5_5, \
call_3_12_5_6, \
call_3_12_5_7, \
call_3_12_5_8, \
call_3_12_5_9, \
call_3_12_5_10, \
call_3_12_5_11, \
call_3_12_5_12, \
call_3_12_5_13, \
call_3_12_5_14, \
call_3_12_5_15, \
call_3_12_6_0, \
call_3_12_6_1, \
call_3_12_6_2, \
call_3_12_6_3, \
call_3_12_6_4, \
call_3_12_6_5, \
call_3_12_6_6, \
call_3_12_6_7, \
call_3_12_6_8, \
call_3_12_6_9, \
call_3_12_6_10, \
call_3_12_6_11, \
call_3_12_6_12, \
call_3_12_6_13, \
call_3_12_6_14, \
call_3_12_6_15, \
call_3_12_7_0, \
call_3_12_7_1, \
call_3_12_7_2, \
call_3_12_7_3, \
call_3_12_7_4, \
call_3_12_7_5, \
call_3_12_7_6, \
call_3_12_7_7, \
call_3_12_7_8, \
call_3_12_7_9, \
call_3_12_7_10, \
call_3_12_7_11, \
call_3_12_7_12, \
call_3_12_7_13, \
call_3_12_7_14, \
call_3_12_7_15, \
call_3_12_8_0, \
call_3_12_8_1, \
call_3_12_8_2, \
call_3_12_8_3, \
call_3_12_8_4, \
call_3_12_8_5, \
call_3_12_8_6, \
call_3_12_8_7, \
call_3_12_8_8, \
call_3_12_8_9, \
call_3_12_8_10, \
call_3_12_8_11, \
call_3_12_8_12, \
call_3_12_8_13, \
call_3_12_8_14, \
call_3_12_8_15, \
call_3_12_9_0, \
call_3_12_9_1, \
call_3_12_9_2, \
call_3_12_9_3, \
call_3_12_9_4, \
call_3_12_9_5, \
call_3_12_9_6, \
call_3_12_9_7, \
call_3_12_9_8, \
call_3_12_9_9, \
call_3_12_9_10, \
call_3_12_9_11, \
call_3_12_9_12, \
call_3_12_9_13, \
call_3_12_9_14, \
call_3_12_9_15, \
call_3_12_10_0, \
call_3_12_10_1, \
call_3_12_10_2, \
call_3_12_10_3, \
call_3_12_10_4, \
call_3_12_10_5, \
call_3_12_10_6, \
call_3_12_10_7, \
call_3_12_10_8, \
call_3_12_10_9, \
call_3_12_10_10, \
call_3_12_10_11, \
call_3_12_10_12, \
call_3_12_10_13, \
call_3_12_10_14, \
call_3_12_10_15, \
call_3_12_11_0, \
call_3_12_11_1, \
call_3_12_11_2, \
call_3_12_11_3, \
call_3_12_11_4, \
call_3_12_11_5, \
call_3_12_11_6, \
call_3_12_11_7, \
call_3_12_11_8, \
call_3_12_11_9, \
call_3_12_11_10, \
call_3_12_11_11, \
call_3_12_11_12, \
call_3_12_11_13, \
call_3_12_11_14, \
call_3_12_11_15, \
call_3_12_12_0, \
call_3_12_12_1, \
call_3_12_12_2, \
call_3_12_12_3, \
call_3_12_12_4, \
call_3_12_12_5, \
call_3_12_12_6, \
call_3_12_12_7, \
call_3_12_12_8, \
call_3_12_12_9, \
call_3_12_12_10, \
call_3_12_12_11, \
call_3_12_12_12, \
call_3_12_12_13, \
call_3_12_12_14, \
call_3_12_12_15, \
call_3_12_13_0, \
call_3_12_13_1, \
call_3_12_13_2, \
call_3_12_13_3, \
call_3_12_13_4, \
call_3_12_13_5, \
call_3_12_13_6, \
call_3_12_13_7, \
call_3_12_13_8, \
call_3_12_13_9, \
call_3_12_13_10, \
call_3_12_13_11, \
call_3_12_13_12, \
call_3_12_13_13, \
call_3_12_13_14, \
call_3_12_13_15, \
call_3_12_14_0, \
call_3_12_14_1, \
call_3_12_14_2, \
call_3_12_14_3, \
call_3_12_14_4, \
call_3_12_14_5, \
call_3_12_14_6, \
call_3_12_14_7, \
call_3_12_14_8, \
call_3_12_14_9, \
call_3_12_14_10, \
call_3_12_14_11, \
call_3_12_14_12, \
call_3_12_14_13, \
call_3_12_14_14, \
call_3_12_14_15, \
call_3_12_15_0, \
call_3_12_15_1, \
call_3_12_15_2, \
call_3_12_15_3, \
call_3_12_15_4, \
call_3_12_15_5, \
call_3_12_15_6, \
call_3_12_15_7, \
call_3_12_15_8, \
call_3_12_15_9, \
call_3_12_15_10, \
call_3_12_15_11, \
call_3_12_15_12, \
call_3_12_15_13, \
call_3_12_15_14, \
call_3_12_15_15, \
call_3_13_0_0, \
call_3_13_0_1, \
call_3_13_0_2, \
call_3_13_0_3, \
call_3_13_0_4, \
call_3_13_0_5, \
call_3_13_0_6, \
call_3_13_0_7, \
call_3_13_0_8, \
call_3_13_0_9, \
call_3_13_0_10, \
call_3_13_0_11, \
call_3_13_0_12, \
call_3_13_0_13, \
call_3_13_0_14, \
call_3_13_0_15, \
call_3_13_1_0, \
call_3_13_1_1, \
call_3_13_1_2, \
call_3_13_1_3, \
call_3_13_1_4, \
call_3_13_1_5, \
call_3_13_1_6, \
call_3_13_1_7, \
call_3_13_1_8, \
call_3_13_1_9, \
call_3_13_1_10, \
call_3_13_1_11, \
call_3_13_1_12, \
call_3_13_1_13, \
call_3_13_1_14, \
call_3_13_1_15, \
call_3_13_2_0, \
call_3_13_2_1, \
call_3_13_2_2, \
call_3_13_2_3, \
call_3_13_2_4, \
call_3_13_2_5, \
call_3_13_2_6, \
call_3_13_2_7, \
call_3_13_2_8, \
call_3_13_2_9, \
call_3_13_2_10, \
call_3_13_2_11, \
call_3_13_2_12, \
call_3_13_2_13, \
call_3_13_2_14, \
call_3_13_2_15, \
call_3_13_3_0, \
call_3_13_3_1, \
call_3_13_3_2, \
call_3_13_3_3, \
call_3_13_3_4, \
call_3_13_3_5, \
call_3_13_3_6, \
call_3_13_3_7, \
call_3_13_3_8, \
call_3_13_3_9, \
call_3_13_3_10, \
call_3_13_3_11, \
call_3_13_3_12, \
call_3_13_3_13, \
call_3_13_3_14, \
call_3_13_3_15, \
call_3_13_4_0, \
call_3_13_4_1, \
call_3_13_4_2, \
call_3_13_4_3, \
call_3_13_4_4, \
call_3_13_4_5, \
call_3_13_4_6, \
call_3_13_4_7, \
call_3_13_4_8, \
call_3_13_4_9, \
call_3_13_4_10, \
call_3_13_4_11, \
call_3_13_4_12, \
call_3_13_4_13, \
call_3_13_4_14, \
call_3_13_4_15, \
call_3_13_5_0, \
call_3_13_5_1, \
call_3_13_5_2, \
call_3_13_5_3, \
call_3_13_5_4, \
call_3_13_5_5, \
call_3_13_5_6, \
call_3_13_5_7, \
call_3_13_5_8, \
call_3_13_5_9, \
call_3_13_5_10, \
call_3_13_5_11, \
call_3_13_5_12, \
call_3_13_5_13, \
call_3_13_5_14, \
call_3_13_5_15, \
call_3_13_6_0, \
call_3_13_6_1, \
call_3_13_6_2, \
call_3_13_6_3, \
call_3_13_6_4, \
call_3_13_6_5, \
call_3_13_6_6, \
call_3_13_6_7, \
call_3_13_6_8, \
call_3_13_6_9, \
call_3_13_6_10, \
call_3_13_6_11, \
call_3_13_6_12, \
call_3_13_6_13, \
call_3_13_6_14, \
call_3_13_6_15, \
call_3_13_7_0, \
call_3_13_7_1, \
call_3_13_7_2, \
call_3_13_7_3, \
call_3_13_7_4, \
call_3_13_7_5, \
call_3_13_7_6, \
call_3_13_7_7, \
call_3_13_7_8, \
call_3_13_7_9, \
call_3_13_7_10, \
call_3_13_7_11, \
call_3_13_7_12, \
call_3_13_7_13, \
call_3_13_7_14, \
call_3_13_7_15, \
call_3_13_8_0, \
call_3_13_8_1, \
call_3_13_8_2, \
call_3_13_8_3, \
call_3_13_8_4, \
call_3_13_8_5, \
call_3_13_8_6, \
call_3_13_8_7, \
call_3_13_8_8, \
call_3_13_8_9, \
call_3_13_8_10, \
call_3_13_8_11, \
call_3_13_8_12, \
call_3_13_8_13, \
call_3_13_8_14, \
call_3_13_8_15, \
call_3_13_9_0, \
call_3_13_9_1, \
call_3_13_9_2, \
call_3_13_9_3, \
call_3_13_9_4, \
call_3_13_9_5, \
call_3_13_9_6, \
call_3_13_9_7, \
call_3_13_9_8, \
call_3_13_9_9, \
call_3_13_9_10, \
call_3_13_9_11, \
call_3_13_9_12, \
call_3_13_9_13, \
call_3_13_9_14, \
call_3_13_9_15, \
call_3_13_10_0, \
call_3_13_10_1, \
call_3_13_10_2, \
call_3_13_10_3, \
call_3_13_10_4, \
call_3_13_10_5, \
call_3_13_10_6, \
call_3_13_10_7, \
call_3_13_10_8, \
call_3_13_10_9, \
call_3_13_10_10, \
call_3_13_10_11, \
call_3_13_10_12, \
call_3_13_10_13, \
call_3_13_10_14, \
call_3_13_10_15, \
call_3_13_11_0, \
call_3_13_11_1, \
call_3_13_11_2, \
call_3_13_11_3, \
call_3_13_11_4, \
call_3_13_11_5, \
call_3_13_11_6, \
call_3_13_11_7, \
call_3_13_11_8, \
call_3_13_11_9, \
call_3_13_11_10, \
call_3_13_11_11, \
call_3_13_11_12, \
call_3_13_11_13, \
call_3_13_11_14, \
call_3_13_11_15, \
call_3_13_12_0, \
call_3_13_12_1, \
call_3_13_12_2, \
call_3_13_12_3, \
call_3_13_12_4, \
call_3_13_12_5, \
call_3_13_12_6, \
call_3_13_12_7, \
call_3_13_12_8, \
call_3_13_12_9, \
call_3_13_12_10, \
call_3_13_12_11, \
call_3_13_12_12, \
call_3_13_12_13, \
call_3_13_12_14, \
call_3_13_12_15, \
call_3_13_13_0, \
call_3_13_13_1, \
call_3_13_13_2, \
call_3_13_13_3, \
call_3_13_13_4, \
call_3_13_13_5, \
call_3_13_13_6, \
call_3_13_13_7, \
call_3_13_13_8, \
call_3_13_13_9, \
call_3_13_13_10, \
call_3_13_13_11, \
call_3_13_13_12, \
call_3_13_13_13, \
call_3_13_13_14, \
call_3_13_13_15, \
call_3_13_14_0, \
call_3_13_14_1, \
call_3_13_14_2, \
call_3_13_14_3, \
call_3_13_14_4, \
call_3_13_14_5, \
call_3_13_14_6, \
call_3_13_14_7, \
call_3_13_14_8, \
call_3_13_14_9, \
call_3_13_14_10, \
call_3_13_14_11, \
call_3_13_14_12, \
call_3_13_14_13, \
call_3_13_14_14, \
call_3_13_14_15, \
call_3_13_15_0, \
call_3_13_15_1, \
call_3_13_15_2, \
call_3_13_15_3, \
call_3_13_15_4, \
call_3_13_15_5, \
call_3_13_15_6, \
call_3_13_15_7, \
call_3_13_15_8, \
call_3_13_15_9, \
call_3_13_15_10, \
call_3_13_15_11, \
call_3_13_15_12, \
call_3_13_15_13, \
call_3_13_15_14, \
call_3_13_15_15, \
call_3_14_0_0, \
call_3_14_0_1, \
call_3_14_0_2, \
call_3_14_0_3, \
call_3_14_0_4, \
call_3_14_0_5, \
call_3_14_0_6, \
call_3_14_0_7, \
call_3_14_0_8, \
call_3_14_0_9, \
call_3_14_0_10, \
call_3_14_0_11, \
call_3_14_0_12, \
call_3_14_0_13, \
call_3_14_0_14, \
call_3_14_0_15, \
call_3_14_1_0, \
call_3_14_1_1, \
call_3_14_1_2, \
call_3_14_1_3, \
call_3_14_1_4, \
call_3_14_1_5, \
call_3_14_1_6, \
call_3_14_1_7, \
call_3_14_1_8, \
call_3_14_1_9, \
call_3_14_1_10, \
call_3_14_1_11, \
call_3_14_1_12, \
call_3_14_1_13, \
call_3_14_1_14, \
call_3_14_1_15, \
call_3_14_2_0, \
call_3_14_2_1, \
call_3_14_2_2, \
call_3_14_2_3, \
call_3_14_2_4, \
call_3_14_2_5, \
call_3_14_2_6, \
call_3_14_2_7, \
call_3_14_2_8, \
call_3_14_2_9, \
call_3_14_2_10, \
call_3_14_2_11, \
call_3_14_2_12, \
call_3_14_2_13, \
call_3_14_2_14, \
call_3_14_2_15, \
call_3_14_3_0, \
call_3_14_3_1, \
call_3_14_3_2, \
call_3_14_3_3, \
call_3_14_3_4, \
call_3_14_3_5, \
call_3_14_3_6, \
call_3_14_3_7, \
call_3_14_3_8, \
call_3_14_3_9, \
call_3_14_3_10, \
call_3_14_3_11, \
call_3_14_3_12, \
call_3_14_3_13, \
call_3_14_3_14, \
call_3_14_3_15, \
call_3_14_4_0, \
call_3_14_4_1, \
call_3_14_4_2, \
call_3_14_4_3, \
call_3_14_4_4, \
call_3_14_4_5, \
call_3_14_4_6, \
call_3_14_4_7, \
call_3_14_4_8, \
call_3_14_4_9, \
call_3_14_4_10, \
call_3_14_4_11, \
call_3_14_4_12, \
call_3_14_4_13, \
call_3_14_4_14, \
call_3_14_4_15, \
call_3_14_5_0, \
call_3_14_5_1, \
call_3_14_5_2, \
call_3_14_5_3, \
call_3_14_5_4, \
call_3_14_5_5, \
call_3_14_5_6, \
call_3_14_5_7, \
call_3_14_5_8, \
call_3_14_5_9, \
call_3_14_5_10, \
call_3_14_5_11, \
call_3_14_5_12, \
call_3_14_5_13, \
call_3_14_5_14, \
call_3_14_5_15, \
call_3_14_6_0, \
call_3_14_6_1, \
call_3_14_6_2, \
call_3_14_6_3, \
call_3_14_6_4, \
call_3_14_6_5, \
call_3_14_6_6, \
call_3_14_6_7, \
call_3_14_6_8, \
call_3_14_6_9, \
call_3_14_6_10, \
call_3_14_6_11, \
call_3_14_6_12, \
call_3_14_6_13, \
call_3_14_6_14, \
call_3_14_6_15, \
call_3_14_7_0, \
call_3_14_7_1, \
call_3_14_7_2, \
call_3_14_7_3, \
call_3_14_7_4, \
call_3_14_7_5, \
call_3_14_7_6, \
call_3_14_7_7, \
call_3_14_7_8, \
call_3_14_7_9, \
call_3_14_7_10, \
call_3_14_7_11, \
call_3_14_7_12, \
call_3_14_7_13, \
call_3_14_7_14, \
call_3_14_7_15, \
call_3_14_8_0, \
call_3_14_8_1, \
call_3_14_8_2, \
call_3_14_8_3, \
call_3_14_8_4, \
call_3_14_8_5, \
call_3_14_8_6, \
call_3_14_8_7, \
call_3_14_8_8, \
call_3_14_8_9, \
call_3_14_8_10, \
call_3_14_8_11, \
call_3_14_8_12, \
call_3_14_8_13, \
call_3_14_8_14, \
call_3_14_8_15, \
call_3_14_9_0, \
call_3_14_9_1, \
call_3_14_9_2, \
call_3_14_9_3, \
call_3_14_9_4, \
call_3_14_9_5, \
call_3_14_9_6, \
call_3_14_9_7, \
call_3_14_9_8, \
call_3_14_9_9, \
call_3_14_9_10, \
call_3_14_9_11, \
call_3_14_9_12, \
call_3_14_9_13, \
call_3_14_9_14, \
call_3_14_9_15, \
call_3_14_10_0, \
call_3_14_10_1, \
call_3_14_10_2, \
call_3_14_10_3, \
call_3_14_10_4, \
call_3_14_10_5, \
call_3_14_10_6, \
call_3_14_10_7, \
call_3_14_10_8, \
call_3_14_10_9, \
call_3_14_10_10, \
call_3_14_10_11, \
call_3_14_10_12, \
call_3_14_10_13, \
call_3_14_10_14, \
call_3_14_10_15, \
call_3_14_11_0, \
call_3_14_11_1, \
call_3_14_11_2, \
call_3_14_11_3, \
call_3_14_11_4, \
call_3_14_11_5, \
call_3_14_11_6, \
call_3_14_11_7, \
call_3_14_11_8, \
call_3_14_11_9, \
call_3_14_11_10, \
call_3_14_11_11, \
call_3_14_11_12, \
call_3_14_11_13, \
call_3_14_11_14, \
call_3_14_11_15, \
call_3_14_12_0, \
call_3_14_12_1, \
call_3_14_12_2, \
call_3_14_12_3, \
call_3_14_12_4, \
call_3_14_12_5, \
call_3_14_12_6, \
call_3_14_12_7, \
call_3_14_12_8, \
call_3_14_12_9, \
call_3_14_12_10, \
call_3_14_12_11, \
call_3_14_12_12, \
call_3_14_12_13, \
call_3_14_12_14, \
call_3_14_12_15, \
call_3_14_13_0, \
call_3_14_13_1, \
call_3_14_13_2, \
call_3_14_13_3, \
call_3_14_13_4, \
call_3_14_13_5, \
call_3_14_13_6, \
call_3_14_13_7, \
call_3_14_13_8, \
call_3_14_13_9, \
call_3_14_13_10, \
call_3_14_13_11, \
call_3_14_13_12, \
call_3_14_13_13, \
call_3_14_13_14, \
call_3_14_13_15, \
call_3_14_14_0, \
call_3_14_14_1, \
call_3_14_14_2, \
call_3_14_14_3, \
call_3_14_14_4, \
call_3_14_14_5, \
call_3_14_14_6, \
call_3_14_14_7, \
call_3_14_14_8, \
call_3_14_14_9, \
call_3_14_14_10, \
call_3_14_14_11, \
call_3_14_14_12, \
call_3_14_14_13, \
call_3_14_14_14, \
call_3_14_14_15, \
call_3_14_15_0, \
call_3_14_15_1, \
call_3_14_15_2, \
call_3_14_15_3, \
call_3_14_15_4, \
call_3_14_15_5, \
call_3_14_15_6, \
call_3_14_15_7, \
call_3_14_15_8, \
call_3_14_15_9, \
call_3_14_15_10, \
call_3_14_15_11, \
call_3_14_15_12, \
call_3_14_15_13, \
call_3_14_15_14, \
call_3_14_15_15, \
call_3_15_0_0, \
call_3_15_0_1, \
call_3_15_0_2, \
call_3_15_0_3, \
call_3_15_0_4, \
call_3_15_0_5, \
call_3_15_0_6, \
call_3_15_0_7, \
call_3_15_0_8, \
call_3_15_0_9, \
call_3_15_0_10, \
call_3_15_0_11, \
call_3_15_0_12, \
call_3_15_0_13, \
call_3_15_0_14, \
call_3_15_0_15, \
call_3_15_1_0, \
call_3_15_1_1, \
call_3_15_1_2, \
call_3_15_1_3, \
call_3_15_1_4, \
call_3_15_1_5, \
call_3_15_1_6, \
call_3_15_1_7, \
call_3_15_1_8, \
call_3_15_1_9, \
call_3_15_1_10, \
call_3_15_1_11, \
call_3_15_1_12, \
call_3_15_1_13, \
call_3_15_1_14, \
call_3_15_1_15, \
call_3_15_2_0, \
call_3_15_2_1, \
call_3_15_2_2, \
call_3_15_2_3, \
call_3_15_2_4, \
call_3_15_2_5, \
call_3_15_2_6, \
call_3_15_2_7, \
call_3_15_2_8, \
call_3_15_2_9, \
call_3_15_2_10, \
call_3_15_2_11, \
call_3_15_2_12, \
call_3_15_2_13, \
call_3_15_2_14, \
call_3_15_2_15, \
call_3_15_3_0, \
call_3_15_3_1, \
call_3_15_3_2, \
call_3_15_3_3, \
call_3_15_3_4, \
call_3_15_3_5, \
call_3_15_3_6, \
call_3_15_3_7, \
call_3_15_3_8, \
call_3_15_3_9, \
call_3_15_3_10, \
call_3_15_3_11, \
call_3_15_3_12, \
call_3_15_3_13, \
call_3_15_3_14, \
call_3_15_3_15, \
call_3_15_4_0, \
call_3_15_4_1, \
call_3_15_4_2, \
call_3_15_4_3, \
call_3_15_4_4, \
call_3_15_4_5, \
call_3_15_4_6, \
call_3_15_4_7, \
call_3_15_4_8, \
call_3_15_4_9, \
call_3_15_4_10, \
call_3_15_4_11, \
call_3_15_4_12, \
call_3_15_4_13, \
call_3_15_4_14, \
call_3_15_4_15, \
call_3_15_5_0, \
call_3_15_5_1, \
call_3_15_5_2, \
call_3_15_5_3, \
call_3_15_5_4, \
call_3_15_5_5, \
call_3_15_5_6, \
call_3_15_5_7, \
call_3_15_5_8, \
call_3_15_5_9, \
call_3_15_5_10, \
call_3_15_5_11, \
call_3_15_5_12, \
call_3_15_5_13, \
call_3_15_5_14, \
call_3_15_5_15, \
call_3_15_6_0, \
call_3_15_6_1, \
call_3_15_6_2, \
call_3_15_6_3, \
call_3_15_6_4, \
call_3_15_6_5, \
call_3_15_6_6, \
call_3_15_6_7, \
call_3_15_6_8, \
call_3_15_6_9, \
call_3_15_6_10, \
call_3_15_6_11, \
call_3_15_6_12, \
call_3_15_6_13, \
call_3_15_6_14, \
call_3_15_6_15, \
call_3_15_7_0, \
call_3_15_7_1, \
call_3_15_7_2, \
call_3_15_7_3, \
call_3_15_7_4, \
call_3_15_7_5, \
call_3_15_7_6, \
call_3_15_7_7, \
call_3_15_7_8, \
call_3_15_7_9, \
call_3_15_7_10, \
call_3_15_7_11, \
call_3_15_7_12, \
call_3_15_7_13, \
call_3_15_7_14, \
call_3_15_7_15, \
call_3_15_8_0, \
call_3_15_8_1, \
call_3_15_8_2, \
call_3_15_8_3, \
call_3_15_8_4, \
call_3_15_8_5, \
call_3_15_8_6, \
call_3_15_8_7, \
call_3_15_8_8, \
call_3_15_8_9, \
call_3_15_8_10, \
call_3_15_8_11, \
call_3_15_8_12, \
call_3_15_8_13, \
call_3_15_8_14, \
call_3_15_8_15, \
call_3_15_9_0, \
call_3_15_9_1, \
call_3_15_9_2, \
call_3_15_9_3, \
call_3_15_9_4, \
call_3_15_9_5, \
call_3_15_9_6, \
call_3_15_9_7, \
call_3_15_9_8, \
call_3_15_9_9, \
call_3_15_9_10, \
call_3_15_9_11, \
call_3_15_9_12, \
call_3_15_9_13, \
call_3_15_9_14, \
call_3_15_9_15, \
call_3_15_10_0, \
call_3_15_10_1, \
call_3_15_10_2, \
call_3_15_10_3, \
call_3_15_10_4, \
call_3_15_10_5, \
call_3_15_10_6, \
call_3_15_10_7, \
call_3_15_10_8, \
call_3_15_10_9, \
call_3_15_10_10, \
call_3_15_10_11, \
call_3_15_10_12, \
call_3_15_10_13, \
call_3_15_10_14, \
call_3_15_10_15, \
call_3_15_11_0, \
call_3_15_11_1, \
call_3_15_11_2, \
call_3_15_11_3, \
call_3_15_11_4, \
call_3_15_11_5, \
call_3_15_11_6, \
call_3_15_11_7, \
call_3_15_11_8, \
call_3_15_11_9, \
call_3_15_11_10, \
call_3_15_11_11, \
call_3_15_11_12, \
call_3_15_11_13, \
call_3_15_11_14, \
call_3_15_11_15, \
call_3_15_12_0, \
call_3_15_12_1, \
call_3_15_12_2, \
call_3_15_12_3, \
call_3_15_12_4, \
call_3_15_12_5, \
call_3_15_12_6, \
call_3_15_12_7, \
call_3_15_12_8, \
call_3_15_12_9, \
call_3_15_12_10, \
call_3_15_12_11, \
call_3_15_12_12, \
call_3_15_12_13, \
call_3_15_12_14, \
call_3_15_12_15, \
call_3_15_13_0, \
call_3_15_13_1, \
call_3_15_13_2, \
call_3_15_13_3, \
call_3_15_13_4, \
call_3_15_13_5, \
call_3_15_13_6, \
call_3_15_13_7, \
call_3_15_13_8, \
call_3_15_13_9, \
call_3_15_13_10, \
call_3_15_13_11, \
call_3_15_13_12, \
call_3_15_13_13, \
call_3_15_13_14, \
call_3_15_13_15, \
call_3_15_14_0, \
call_3_15_14_1, \
call_3_15_14_2, \
call_3_15_14_3, \
call_3_15_14_4, \
call_3_15_14_5, \
call_3_15_14_6, \
call_3_15_14_7, \
call_3_15_14_8, \
call_3_15_14_9, \
call_3_15_14_10, \
call_3_15_14_11, \
call_3_15_14_12, \
call_3_15_14_13, \
call_3_15_14_14, \
call_3_15_14_15, \
call_3_15_15_0, \
call_3_15_15_1, \
call_3_15_15_2, \
call_3_15_15_3, \
call_3_15_15_4, \
call_3_15_15_5, \
call_3_15_15_6, \
call_3_15_15_7, \
call_3_15_15_8, \
call_3_15_15_9, \
call_3_15_15_10, \
call_3_15_15_11, \
call_3_15_15_12, \
call_3_15_15_13, \
call_3_15_15_14, \
call_3_15_15_15, \
call_4_0_0_0, \
call_4_0_0_1, \
call_4_0_0_2, \
call_4_0_0_3, \
call_4_0_0_4, \
call_4_0_0_5, \
call_4_0_0_6, \
call_4_0_0_7, \
call_4_0_0_8, \
call_4_0_0_9, \
call_4_0_0_10, \
call_4_0_0_11, \
call_4_0_0_12, \
call_4_0_0_13, \
call_4_0_0_14, \
call_4_0_0_15, \
call_4_0_1_0, \
call_4_0_1_1, \
call_4_0_1_2, \
call_4_0_1_3, \
call_4_0_1_4, \
call_4_0_1_5, \
call_4_0_1_6, \
call_4_0_1_7, \
call_4_0_1_8, \
call_4_0_1_9, \
call_4_0_1_10, \
call_4_0_1_11, \
call_4_0_1_12, \
call_4_0_1_13, \
call_4_0_1_14, \
call_4_0_1_15, \
call_4_0_2_0, \
call_4_0_2_1, \
call_4_0_2_2, \
call_4_0_2_3, \
call_4_0_2_4, \
call_4_0_2_5, \
call_4_0_2_6, \
call_4_0_2_7, \
call_4_0_2_8, \
call_4_0_2_9, \
call_4_0_2_10, \
call_4_0_2_11, \
call_4_0_2_12, \
call_4_0_2_13, \
call_4_0_2_14, \
call_4_0_2_15, \
call_4_0_3_0, \
call_4_0_3_1, \
call_4_0_3_2, \
call_4_0_3_3, \
call_4_0_3_4, \
call_4_0_3_5, \
call_4_0_3_6, \
call_4_0_3_7, \
call_4_0_3_8, \
call_4_0_3_9, \
call_4_0_3_10, \
call_4_0_3_11, \
call_4_0_3_12, \
call_4_0_3_13, \
call_4_0_3_14, \
call_4_0_3_15, \
call_4_0_4_0, \
call_4_0_4_1, \
call_4_0_4_2, \
call_4_0_4_3, \
call_4_0_4_4, \
call_4_0_4_5, \
call_4_0_4_6, \
call_4_0_4_7, \
call_4_0_4_8, \
call_4_0_4_9, \
call_4_0_4_10, \
call_4_0_4_11, \
call_4_0_4_12, \
call_4_0_4_13, \
call_4_0_4_14, \
call_4_0_4_15, \
call_4_0_5_0, \
call_4_0_5_1, \
call_4_0_5_2, \
call_4_0_5_3, \
call_4_0_5_4, \
call_4_0_5_5, \
call_4_0_5_6, \
call_4_0_5_7, \
call_4_0_5_8, \
call_4_0_5_9, \
call_4_0_5_10, \
call_4_0_5_11, \
call_4_0_5_12, \
call_4_0_5_13, \
call_4_0_5_14, \
call_4_0_5_15, \
call_4_0_6_0, \
call_4_0_6_1, \
call_4_0_6_2, \
call_4_0_6_3, \
call_4_0_6_4, \
call_4_0_6_5, \
call_4_0_6_6, \
call_4_0_6_7, \
call_4_0_6_8, \
call_4_0_6_9, \
call_4_0_6_10, \
call_4_0_6_11, \
call_4_0_6_12, \
call_4_0_6_13, \
call_4_0_6_14, \
call_4_0_6_15, \
call_4_0_7_0, \
call_4_0_7_1, \
call_4_0_7_2, \
call_4_0_7_3, \
call_4_0_7_4, \
call_4_0_7_5, \
call_4_0_7_6, \
call_4_0_7_7, \
call_4_0_7_8, \
call_4_0_7_9, \
call_4_0_7_10, \
call_4_0_7_11, \
call_4_0_7_12, \
call_4_0_7_13, \
call_4_0_7_14, \
call_4_0_7_15, \
call_4_0_8_0, \
call_4_0_8_1, \
call_4_0_8_2, \
call_4_0_8_3, \
call_4_0_8_4, \
call_4_0_8_5, \
call_4_0_8_6, \
call_4_0_8_7, \
call_4_0_8_8, \
call_4_0_8_9, \
call_4_0_8_10, \
call_4_0_8_11, \
call_4_0_8_12, \
call_4_0_8_13, \
call_4_0_8_14, \
call_4_0_8_15, \
call_4_0_9_0, \
call_4_0_9_1, \
call_4_0_9_2, \
call_4_0_9_3, \
call_4_0_9_4, \
call_4_0_9_5, \
call_4_0_9_6, \
call_4_0_9_7, \
call_4_0_9_8, \
call_4_0_9_9, \
call_4_0_9_10, \
call_4_0_9_11, \
call_4_0_9_12, \
call_4_0_9_13, \
call_4_0_9_14, \
call_4_0_9_15, \
call_4_0_10_0, \
call_4_0_10_1, \
call_4_0_10_2, \
call_4_0_10_3, \
call_4_0_10_4, \
call_4_0_10_5, \
call_4_0_10_6, \
call_4_0_10_7, \
call_4_0_10_8, \
call_4_0_10_9, \
call_4_0_10_10, \
call_4_0_10_11, \
call_4_0_10_12, \
call_4_0_10_13, \
call_4_0_10_14, \
call_4_0_10_15, \
call_4_0_11_0, \
call_4_0_11_1, \
call_4_0_11_2, \
call_4_0_11_3, \
call_4_0_11_4, \
call_4_0_11_5, \
call_4_0_11_6, \
call_4_0_11_7, \
call_4_0_11_8, \
call_4_0_11_9, \
call_4_0_11_10, \
call_4_0_11_11, \
call_4_0_11_12, \
call_4_0_11_13, \
call_4_0_11_14, \
call_4_0_11_15, \
call_4_0_12_0, \
call_4_0_12_1, \
call_4_0_12_2, \
call_4_0_12_3, \
call_4_0_12_4, \
call_4_0_12_5, \
call_4_0_12_6, \
call_4_0_12_7, \
call_4_0_12_8, \
call_4_0_12_9, \
call_4_0_12_10, \
call_4_0_12_11, \
call_4_0_12_12, \
call_4_0_12_13, \
call_4_0_12_14, \
call_4_0_12_15, \
call_4_0_13_0, \
call_4_0_13_1, \
call_4_0_13_2, \
call_4_0_13_3, \
call_4_0_13_4, \
call_4_0_13_5, \
call_4_0_13_6, \
call_4_0_13_7, \
call_4_0_13_8, \
call_4_0_13_9, \
call_4_0_13_10, \
call_4_0_13_11, \
call_4_0_13_12, \
call_4_0_13_13, \
call_4_0_13_14, \
call_4_0_13_15, \
call_4_0_14_0, \
call_4_0_14_1, \
call_4_0_14_2, \
call_4_0_14_3, \
call_4_0_14_4, \
call_4_0_14_5, \
call_4_0_14_6, \
call_4_0_14_7, \
call_4_0_14_8, \
call_4_0_14_9, \
call_4_0_14_10, \
call_4_0_14_11, \
call_4_0_14_12, \
call_4_0_14_13, \
call_4_0_14_14, \
call_4_0_14_15, \
call_4_0_15_0, \
call_4_0_15_1, \
call_4_0_15_2, \
call_4_0_15_3, \
call_4_0_15_4, \
call_4_0_15_5, \
call_4_0_15_6, \
call_4_0_15_7, \
call_4_0_15_8, \
call_4_0_15_9, \
call_4_0_15_10, \
call_4_0_15_11, \
call_4_0_15_12, \
call_4_0_15_13, \
call_4_0_15_14, \
call_4_0_15_15, \
call_4_1_0_0, \
call_4_1_0_1, \
call_4_1_0_2, \
call_4_1_0_3, \
call_4_1_0_4, \
call_4_1_0_5, \
call_4_1_0_6, \
call_4_1_0_7, \
call_4_1_0_8, \
call_4_1_0_9, \
call_4_1_0_10, \
call_4_1_0_11, \
call_4_1_0_12, \
call_4_1_0_13, \
call_4_1_0_14, \
call_4_1_0_15, \
call_4_1_1_0, \
call_4_1_1_1, \
call_4_1_1_2, \
call_4_1_1_3, \
call_4_1_1_4, \
call_4_1_1_5, \
call_4_1_1_6, \
call_4_1_1_7, \
call_4_1_1_8, \
call_4_1_1_9, \
call_4_1_1_10, \
call_4_1_1_11, \
call_4_1_1_12, \
call_4_1_1_13, \
call_4_1_1_14, \
call_4_1_1_15, \
call_4_1_2_0, \
call_4_1_2_1, \
call_4_1_2_2, \
call_4_1_2_3, \
call_4_1_2_4, \
call_4_1_2_5, \
call_4_1_2_6, \
call_4_1_2_7, \
call_4_1_2_8, \
call_4_1_2_9, \
call_4_1_2_10, \
call_4_1_2_11, \
call_4_1_2_12, \
call_4_1_2_13, \
call_4_1_2_14, \
call_4_1_2_15, \
call_4_1_3_0, \
call_4_1_3_1, \
call_4_1_3_2, \
call_4_1_3_3, \
call_4_1_3_4, \
call_4_1_3_5, \
call_4_1_3_6, \
call_4_1_3_7, \
call_4_1_3_8, \
call_4_1_3_9, \
call_4_1_3_10, \
call_4_1_3_11, \
call_4_1_3_12, \
call_4_1_3_13, \
call_4_1_3_14, \
call_4_1_3_15, \
call_4_1_4_0, \
call_4_1_4_1, \
call_4_1_4_2, \
call_4_1_4_3, \
call_4_1_4_4, \
call_4_1_4_5, \
call_4_1_4_6, \
call_4_1_4_7, \
call_4_1_4_8, \
call_4_1_4_9, \
call_4_1_4_10, \
call_4_1_4_11, \
call_4_1_4_12, \
call_4_1_4_13, \
call_4_1_4_14, \
call_4_1_4_15, \
call_4_1_5_0, \
call_4_1_5_1, \
call_4_1_5_2, \
call_4_1_5_3, \
call_4_1_5_4, \
call_4_1_5_5, \
call_4_1_5_6, \
call_4_1_5_7, \
call_4_1_5_8, \
call_4_1_5_9, \
call_4_1_5_10, \
call_4_1_5_11, \
call_4_1_5_12, \
call_4_1_5_13, \
call_4_1_5_14, \
call_4_1_5_15, \
call_4_1_6_0, \
call_4_1_6_1, \
call_4_1_6_2, \
call_4_1_6_3, \
call_4_1_6_4, \
call_4_1_6_5, \
call_4_1_6_6, \
call_4_1_6_7, \
call_4_1_6_8, \
call_4_1_6_9, \
call_4_1_6_10, \
call_4_1_6_11, \
call_4_1_6_12, \
call_4_1_6_13, \
call_4_1_6_14, \
call_4_1_6_15, \
call_4_1_7_0, \
call_4_1_7_1, \
call_4_1_7_2, \
call_4_1_7_3, \
call_4_1_7_4, \
call_4_1_7_5, \
call_4_1_7_6, \
call_4_1_7_7, \
call_4_1_7_8, \
call_4_1_7_9, \
call_4_1_7_10, \
call_4_1_7_11, \
call_4_1_7_12, \
call_4_1_7_13, \
call_4_1_7_14, \
call_4_1_7_15, \
call_4_1_8_0, \
call_4_1_8_1, \
call_4_1_8_2, \
call_4_1_8_3, \
call_4_1_8_4, \
call_4_1_8_5, \
call_4_1_8_6, \
call_4_1_8_7, \
call_4_1_8_8, \
call_4_1_8_9, \
call_4_1_8_10, \
call_4_1_8_11, \
call_4_1_8_12, \
call_4_1_8_13, \
call_4_1_8_14, \
call_4_1_8_15, \
call_4_1_9_0, \
call_4_1_9_1, \
call_4_1_9_2, \
call_4_1_9_3, \
call_4_1_9_4, \
call_4_1_9_5, \
call_4_1_9_6, \
call_4_1_9_7, \
call_4_1_9_8, \
call_4_1_9_9, \
call_4_1_9_10, \
call_4_1_9_11, \
call_4_1_9_12, \
call_4_1_9_13, \
call_4_1_9_14, \
call_4_1_9_15, \
call_4_1_10_0, \
call_4_1_10_1, \
call_4_1_10_2, \
call_4_1_10_3, \
call_4_1_10_4, \
call_4_1_10_5, \
call_4_1_10_6, \
call_4_1_10_7, \
call_4_1_10_8, \
call_4_1_10_9, \
call_4_1_10_10, \
call_4_1_10_11, \
call_4_1_10_12, \
call_4_1_10_13, \
call_4_1_10_14, \
call_4_1_10_15, \
call_4_1_11_0, \
call_4_1_11_1, \
call_4_1_11_2, \
call_4_1_11_3, \
call_4_1_11_4, \
call_4_1_11_5, \
call_4_1_11_6, \
call_4_1_11_7, \
call_4_1_11_8, \
call_4_1_11_9, \
call_4_1_11_10, \
call_4_1_11_11, \
call_4_1_11_12, \
call_4_1_11_13, \
call_4_1_11_14, \
call_4_1_11_15, \
call_4_1_12_0, \
call_4_1_12_1, \
call_4_1_12_2, \
call_4_1_12_3, \
call_4_1_12_4, \
call_4_1_12_5, \
call_4_1_12_6, \
call_4_1_12_7, \
call_4_1_12_8, \
call_4_1_12_9, \
call_4_1_12_10, \
call_4_1_12_11, \
call_4_1_12_12, \
call_4_1_12_13, \
call_4_1_12_14, \
call_4_1_12_15, \
call_4_1_13_0, \
call_4_1_13_1, \
call_4_1_13_2, \
call_4_1_13_3, \
call_4_1_13_4, \
call_4_1_13_5, \
call_4_1_13_6, \
call_4_1_13_7, \
call_4_1_13_8, \
call_4_1_13_9, \
call_4_1_13_10, \
call_4_1_13_11, \
call_4_1_13_12, \
call_4_1_13_13, \
call_4_1_13_14, \
call_4_1_13_15, \
call_4_1_14_0, \
call_4_1_14_1, \
call_4_1_14_2, \
call_4_1_14_3, \
call_4_1_14_4, \
call_4_1_14_5, \
call_4_1_14_6, \
call_4_1_14_7, \
call_4_1_14_8, \
call_4_1_14_9, \
call_4_1_14_10, \
call_4_1_14_11, \
call_4_1_14_12, \
call_4_1_14_13, \
call_4_1_14_14, \
call_4_1_14_15, \
call_4_1_15_0, \
call_4_1_15_1, \
call_4_1_15_2, \
call_4_1_15_3, \
call_4_1_15_4, \
call_4_1_15_5, \
call_4_1_15_6, \
call_4_1_15_7, \
call_4_1_15_8, \
call_4_1_15_9, \
call_4_1_15_10, \
call_4_1_15_11, \
call_4_1_15_12, \
call_4_1_15_13, \
call_4_1_15_14, \
call_4_1_15_15, \
call_4_2_0_0, \
call_4_2_0_1, \
call_4_2_0_2, \
call_4_2_0_3, \
call_4_2_0_4, \
call_4_2_0_5, \
call_4_2_0_6, \
call_4_2_0_7, \
call_4_2_0_8, \
call_4_2_0_9, \
call_4_2_0_10, \
call_4_2_0_11, \
call_4_2_0_12, \
call_4_2_0_13, \
call_4_2_0_14, \
call_4_2_0_15, \
call_4_2_1_0, \
call_4_2_1_1, \
call_4_2_1_2, \
call_4_2_1_3, \
call_4_2_1_4, \
call_4_2_1_5, \
call_4_2_1_6, \
call_4_2_1_7, \
call_4_2_1_8, \
call_4_2_1_9, \
call_4_2_1_10, \
call_4_2_1_11, \
call_4_2_1_12, \
call_4_2_1_13, \
call_4_2_1_14, \
call_4_2_1_15, \
call_4_2_2_0, \
call_4_2_2_1, \
call_4_2_2_2, \
call_4_2_2_3, \
call_4_2_2_4, \
call_4_2_2_5, \
call_4_2_2_6, \
call_4_2_2_7, \
call_4_2_2_8, \
call_4_2_2_9, \
call_4_2_2_10, \
call_4_2_2_11, \
call_4_2_2_12, \
call_4_2_2_13, \
call_4_2_2_14, \
call_4_2_2_15, \
call_4_2_3_0, \
call_4_2_3_1, \
call_4_2_3_2, \
call_4_2_3_3, \
call_4_2_3_4, \
call_4_2_3_5, \
call_4_2_3_6, \
call_4_2_3_7, \
call_4_2_3_8, \
call_4_2_3_9, \
call_4_2_3_10, \
call_4_2_3_11, \
call_4_2_3_12, \
call_4_2_3_13, \
call_4_2_3_14, \
call_4_2_3_15, \
call_4_2_4_0, \
call_4_2_4_1, \
call_4_2_4_2, \
call_4_2_4_3, \
call_4_2_4_4, \
call_4_2_4_5, \
call_4_2_4_6, \
call_4_2_4_7, \
call_4_2_4_8, \
call_4_2_4_9, \
call_4_2_4_10, \
call_4_2_4_11, \
call_4_2_4_12, \
call_4_2_4_13, \
call_4_2_4_14, \
call_4_2_4_15, \
call_4_2_5_0, \
call_4_2_5_1, \
call_4_2_5_2, \
call_4_2_5_3, \
call_4_2_5_4, \
call_4_2_5_5, \
call_4_2_5_6, \
call_4_2_5_7, \
call_4_2_5_8, \
call_4_2_5_9, \
call_4_2_5_10, \
call_4_2_5_11, \
call_4_2_5_12, \
call_4_2_5_13, \
call_4_2_5_14, \
call_4_2_5_15, \
call_4_2_6_0, \
call_4_2_6_1, \
call_4_2_6_2, \
call_4_2_6_3, \
call_4_2_6_4, \
call_4_2_6_5, \
call_4_2_6_6, \
call_4_2_6_7, \
call_4_2_6_8, \
call_4_2_6_9, \
call_4_2_6_10, \
call_4_2_6_11, \
call_4_2_6_12, \
call_4_2_6_13, \
call_4_2_6_14, \
call_4_2_6_15, \
call_4_2_7_0, \
call_4_2_7_1, \
call_4_2_7_2, \
call_4_2_7_3, \
call_4_2_7_4, \
call_4_2_7_5, \
call_4_2_7_6, \
call_4_2_7_7, \
call_4_2_7_8, \
call_4_2_7_9, \
call_4_2_7_10, \
call_4_2_7_11, \
call_4_2_7_12, \
call_4_2_7_13, \
call_4_2_7_14, \
call_4_2_7_15, \
call_4_2_8_0, \
call_4_2_8_1, \
call_4_2_8_2, \
call_4_2_8_3, \
call_4_2_8_4, \
call_4_2_8_5, \
call_4_2_8_6, \
call_4_2_8_7, \
call_4_2_8_8, \
call_4_2_8_9, \
call_4_2_8_10, \
call_4_2_8_11, \
call_4_2_8_12, \
call_4_2_8_13, \
call_4_2_8_14, \
call_4_2_8_15, \
call_4_2_9_0, \
call_4_2_9_1, \
call_4_2_9_2, \
call_4_2_9_3, \
call_4_2_9_4, \
call_4_2_9_5, \
call_4_2_9_6, \
call_4_2_9_7, \
call_4_2_9_8, \
call_4_2_9_9, \
call_4_2_9_10, \
call_4_2_9_11, \
call_4_2_9_12, \
call_4_2_9_13, \
call_4_2_9_14, \
call_4_2_9_15, \
call_4_2_10_0, \
call_4_2_10_1, \
call_4_2_10_2, \
call_4_2_10_3, \
call_4_2_10_4, \
call_4_2_10_5, \
call_4_2_10_6, \
call_4_2_10_7, \
call_4_2_10_8, \
call_4_2_10_9, \
call_4_2_10_10, \
call_4_2_10_11, \
call_4_2_10_12, \
call_4_2_10_13, \
call_4_2_10_14, \
call_4_2_10_15, \
call_4_2_11_0, \
call_4_2_11_1, \
call_4_2_11_2, \
call_4_2_11_3, \
call_4_2_11_4, \
call_4_2_11_5, \
call_4_2_11_6, \
call_4_2_11_7, \
call_4_2_11_8, \
call_4_2_11_9, \
call_4_2_11_10, \
call_4_2_11_11, \
call_4_2_11_12, \
call_4_2_11_13, \
call_4_2_11_14, \
call_4_2_11_15, \
call_4_2_12_0, \
call_4_2_12_1, \
call_4_2_12_2, \
call_4_2_12_3, \
call_4_2_12_4, \
call_4_2_12_5, \
call_4_2_12_6, \
call_4_2_12_7, \
call_4_2_12_8, \
call_4_2_12_9, \
call_4_2_12_10, \
call_4_2_12_11, \
call_4_2_12_12, \
call_4_2_12_13, \
call_4_2_12_14, \
call_4_2_12_15, \
call_4_2_13_0, \
call_4_2_13_1, \
call_4_2_13_2, \
call_4_2_13_3, \
call_4_2_13_4, \
call_4_2_13_5, \
call_4_2_13_6, \
call_4_2_13_7, \
call_4_2_13_8, \
call_4_2_13_9, \
call_4_2_13_10, \
call_4_2_13_11, \
call_4_2_13_12, \
call_4_2_13_13, \
call_4_2_13_14, \
call_4_2_13_15, \
call_4_2_14_0, \
call_4_2_14_1, \
call_4_2_14_2, \
call_4_2_14_3, \
call_4_2_14_4, \
call_4_2_14_5, \
call_4_2_14_6, \
call_4_2_14_7, \
call_4_2_14_8, \
call_4_2_14_9, \
call_4_2_14_10, \
call_4_2_14_11, \
call_4_2_14_12, \
call_4_2_14_13, \
call_4_2_14_14, \
call_4_2_14_15, \
call_4_2_15_0, \
call_4_2_15_1, \
call_4_2_15_2, \
call_4_2_15_3, \
call_4_2_15_4, \
call_4_2_15_5, \
call_4_2_15_6, \
call_4_2_15_7, \
call_4_2_15_8, \
call_4_2_15_9, \
call_4_2_15_10, \
call_4_2_15_11, \
call_4_2_15_12, \
call_4_2_15_13, \
call_4_2_15_14, \
call_4_2_15_15, \
call_4_3_0_0, \
call_4_3_0_1, \
call_4_3_0_2, \
call_4_3_0_3, \
call_4_3_0_4, \
call_4_3_0_5, \
call_4_3_0_6, \
call_4_3_0_7, \
call_4_3_0_8, \
call_4_3_0_9, \
call_4_3_0_10, \
call_4_3_0_11, \
call_4_3_0_12, \
call_4_3_0_13, \
call_4_3_0_14, \
call_4_3_0_15, \
call_4_3_1_0, \
call_4_3_1_1, \
call_4_3_1_2, \
call_4_3_1_3, \
call_4_3_1_4, \
call_4_3_1_5, \
call_4_3_1_6, \
call_4_3_1_7, \
call_4_3_1_8, \
call_4_3_1_9, \
call_4_3_1_10, \
call_4_3_1_11, \
call_4_3_1_12, \
call_4_3_1_13, \
call_4_3_1_14, \
call_4_3_1_15, \
call_4_3_2_0, \
call_4_3_2_1, \
call_4_3_2_2, \
call_4_3_2_3, \
call_4_3_2_4, \
call_4_3_2_5, \
call_4_3_2_6, \
call_4_3_2_7, \
call_4_3_2_8, \
call_4_3_2_9, \
call_4_3_2_10, \
call_4_3_2_11, \
call_4_3_2_12, \
call_4_3_2_13, \
call_4_3_2_14, \
call_4_3_2_15, \
call_4_3_3_0, \
call_4_3_3_1, \
call_4_3_3_2, \
call_4_3_3_3, \
call_4_3_3_4, \
call_4_3_3_5, \
call_4_3_3_6, \
call_4_3_3_7, \
call_4_3_3_8, \
call_4_3_3_9, \
call_4_3_3_10, \
call_4_3_3_11, \
call_4_3_3_12, \
call_4_3_3_13, \
call_4_3_3_14, \
call_4_3_3_15, \
call_4_3_4_0, \
call_4_3_4_1, \
call_4_3_4_2, \
call_4_3_4_3, \
call_4_3_4_4, \
call_4_3_4_5, \
call_4_3_4_6, \
call_4_3_4_7, \
call_4_3_4_8, \
call_4_3_4_9, \
call_4_3_4_10, \
call_4_3_4_11, \
call_4_3_4_12, \
call_4_3_4_13, \
call_4_3_4_14, \
call_4_3_4_15, \
call_4_3_5_0, \
call_4_3_5_1, \
call_4_3_5_2, \
call_4_3_5_3, \
call_4_3_5_4, \
call_4_3_5_5, \
call_4_3_5_6, \
call_4_3_5_7, \
call_4_3_5_8, \
call_4_3_5_9, \
call_4_3_5_10, \
call_4_3_5_11, \
call_4_3_5_12, \
call_4_3_5_13, \
call_4_3_5_14, \
call_4_3_5_15, \
call_4_3_6_0, \
call_4_3_6_1, \
call_4_3_6_2, \
call_4_3_6_3, \
call_4_3_6_4, \
call_4_3_6_5, \
call_4_3_6_6, \
call_4_3_6_7, \
call_4_3_6_8, \
call_4_3_6_9, \
call_4_3_6_10, \
call_4_3_6_11, \
call_4_3_6_12, \
call_4_3_6_13, \
call_4_3_6_14, \
call_4_3_6_15, \
call_4_3_7_0, \
call_4_3_7_1, \
call_4_3_7_2, \
call_4_3_7_3, \
call_4_3_7_4, \
call_4_3_7_5, \
call_4_3_7_6, \
call_4_3_7_7, \
call_4_3_7_8, \
call_4_3_7_9, \
call_4_3_7_10, \
call_4_3_7_11, \
call_4_3_7_12, \
call_4_3_7_13, \
call_4_3_7_14, \
call_4_3_7_15, \
call_4_3_8_0, \
call_4_3_8_1, \
call_4_3_8_2, \
call_4_3_8_3, \
call_4_3_8_4, \
call_4_3_8_5, \
call_4_3_8_6, \
call_4_3_8_7, \
call_4_3_8_8, \
call_4_3_8_9, \
call_4_3_8_10, \
call_4_3_8_11, \
call_4_3_8_12, \
call_4_3_8_13, \
call_4_3_8_14, \
call_4_3_8_15, \
call_4_3_9_0, \
call_4_3_9_1, \
call_4_3_9_2, \
call_4_3_9_3, \
call_4_3_9_4, \
call_4_3_9_5, \
call_4_3_9_6, \
call_4_3_9_7, \
call_4_3_9_8, \
call_4_3_9_9, \
call_4_3_9_10, \
call_4_3_9_11, \
call_4_3_9_12, \
call_4_3_9_13, \
call_4_3_9_14, \
call_4_3_9_15, \
call_4_3_10_0, \
call_4_3_10_1, \
call_4_3_10_2, \
call_4_3_10_3, \
call_4_3_10_4, \
call_4_3_10_5, \
call_4_3_10_6, \
call_4_3_10_7, \
call_4_3_10_8, \
call_4_3_10_9, \
call_4_3_10_10, \
call_4_3_10_11, \
call_4_3_10_12, \
call_4_3_10_13, \
call_4_3_10_14, \
call_4_3_10_15, \
call_4_3_11_0, \
call_4_3_11_1, \
call_4_3_11_2, \
call_4_3_11_3, \
call_4_3_11_4, \
call_4_3_11_5, \
call_4_3_11_6, \
call_4_3_11_7, \
call_4_3_11_8, \
call_4_3_11_9, \
call_4_3_11_10, \
call_4_3_11_11, \
call_4_3_11_12, \
call_4_3_11_13, \
call_4_3_11_14, \
call_4_3_11_15, \
call_4_3_12_0, \
call_4_3_12_1, \
call_4_3_12_2, \
call_4_3_12_3, \
call_4_3_12_4, \
call_4_3_12_5, \
call_4_3_12_6, \
call_4_3_12_7, \
call_4_3_12_8, \
call_4_3_12_9, \
call_4_3_12_10, \
call_4_3_12_11, \
call_4_3_12_12, \
call_4_3_12_13, \
call_4_3_12_14, \
call_4_3_12_15, \
call_4_3_13_0, \
call_4_3_13_1, \
call_4_3_13_2, \
call_4_3_13_3, \
call_4_3_13_4, \
call_4_3_13_5, \
call_4_3_13_6, \
call_4_3_13_7, \
call_4_3_13_8, \
call_4_3_13_9, \
call_4_3_13_10, \
call_4_3_13_11, \
call_4_3_13_12, \
call_4_3_13_13, \
call_4_3_13_14, \
call_4_3_13_15, \
call_4_3_14_0, \
call_4_3_14_1, \
call_4_3_14_2, \
call_4_3_14_3, \
call_4_3_14_4, \
call_4_3_14_5, \
call_4_3_14_6, \
call_4_3_14_7, \
call_4_3_14_8, \
call_4_3_14_9, \
call_4_3_14_10, \
call_4_3_14_11, \
call_4_3_14_12, \
call_4_3_14_13, \
call_4_3_14_14, \
call_4_3_14_15, \
call_4_3_15_0, \
call_4_3_15_1, \
call_4_3_15_2, \
call_4_3_15_3, \
call_4_3_15_4, \
call_4_3_15_5, \
call_4_3_15_6, \
call_4_3_15_7, \
call_4_3_15_8, \
call_4_3_15_9, \
call_4_3_15_10, \
call_4_3_15_11, \
call_4_3_15_12, \
call_4_3_15_13, \
call_4_3_15_14, \
call_4_3_15_15, \
call_4_4_0_0, \
call_4_4_0_1, \
call_4_4_0_2, \
call_4_4_0_3, \
call_4_4_0_4, \
call_4_4_0_5, \
call_4_4_0_6, \
call_4_4_0_7, \
call_4_4_0_8, \
call_4_4_0_9, \
call_4_4_0_10, \
call_4_4_0_11, \
call_4_4_0_12, \
call_4_4_0_13, \
call_4_4_0_14, \
call_4_4_0_15, \
call_4_4_1_0, \
call_4_4_1_1, \
call_4_4_1_2, \
call_4_4_1_3, \
call_4_4_1_4, \
call_4_4_1_5, \
call_4_4_1_6, \
call_4_4_1_7, \
call_4_4_1_8, \
call_4_4_1_9, \
call_4_4_1_10, \
call_4_4_1_11, \
call_4_4_1_12, \
call_4_4_1_13, \
call_4_4_1_14, \
call_4_4_1_15, \
call_4_4_2_0, \
call_4_4_2_1, \
call_4_4_2_2, \
call_4_4_2_3, \
call_4_4_2_4, \
call_4_4_2_5, \
call_4_4_2_6, \
call_4_4_2_7, \
call_4_4_2_8, \
call_4_4_2_9, \
call_4_4_2_10, \
call_4_4_2_11, \
call_4_4_2_12, \
call_4_4_2_13, \
call_4_4_2_14, \
call_4_4_2_15, \
call_4_4_3_0, \
call_4_4_3_1, \
call_4_4_3_2, \
call_4_4_3_3, \
call_4_4_3_4, \
call_4_4_3_5, \
call_4_4_3_6, \
call_4_4_3_7, \
call_4_4_3_8, \
call_4_4_3_9, \
call_4_4_3_10, \
call_4_4_3_11, \
call_4_4_3_12, \
call_4_4_3_13, \
call_4_4_3_14, \
call_4_4_3_15, \
call_4_4_4_0, \
call_4_4_4_1, \
call_4_4_4_2, \
call_4_4_4_3, \
call_4_4_4_4, \
call_4_4_4_5, \
call_4_4_4_6, \
call_4_4_4_7, \
call_4_4_4_8, \
call_4_4_4_9, \
call_4_4_4_10, \
call_4_4_4_11, \
call_4_4_4_12, \
call_4_4_4_13, \
call_4_4_4_14, \
call_4_4_4_15, \
call_4_4_5_0, \
call_4_4_5_1, \
call_4_4_5_2, \
call_4_4_5_3, \
call_4_4_5_4, \
call_4_4_5_5, \
call_4_4_5_6, \
call_4_4_5_7, \
call_4_4_5_8, \
call_4_4_5_9, \
call_4_4_5_10, \
call_4_4_5_11, \
call_4_4_5_12, \
call_4_4_5_13, \
call_4_4_5_14, \
call_4_4_5_15, \
call_4_4_6_0, \
call_4_4_6_1, \
call_4_4_6_2, \
call_4_4_6_3, \
call_4_4_6_4, \
call_4_4_6_5, \
call_4_4_6_6, \
call_4_4_6_7, \
call_4_4_6_8, \
call_4_4_6_9, \
call_4_4_6_10, \
call_4_4_6_11, \
call_4_4_6_12, \
call_4_4_6_13, \
call_4_4_6_14, \
call_4_4_6_15, \
call_4_4_7_0, \
call_4_4_7_1, \
call_4_4_7_2, \
call_4_4_7_3, \
call_4_4_7_4, \
call_4_4_7_5, \
call_4_4_7_6, \
call_4_4_7_7, \
call_4_4_7_8, \
call_4_4_7_9, \
call_4_4_7_10, \
call_4_4_7_11, \
call_4_4_7_12, \
call_4_4_7_13, \
call_4_4_7_14, \
call_4_4_7_15, \
call_4_4_8_0, \
call_4_4_8_1, \
call_4_4_8_2, \
call_4_4_8_3, \
call_4_4_8_4, \
call_4_4_8_5, \
call_4_4_8_6, \
call_4_4_8_7, \
call_4_4_8_8, \
call_4_4_8_9, \
call_4_4_8_10, \
call_4_4_8_11, \
call_4_4_8_12, \
call_4_4_8_13, \
call_4_4_8_14, \
call_4_4_8_15, \
call_4_4_9_0, \
call_4_4_9_1, \
call_4_4_9_2, \
call_4_4_9_3, \
call_4_4_9_4, \
call_4_4_9_5, \
call_4_4_9_6, \
call_4_4_9_7, \
call_4_4_9_8, \
call_4_4_9_9, \
call_4_4_9_10, \
call_4_4_9_11, \
call_4_4_9_12, \
call_4_4_9_13, \
call_4_4_9_14, \
call_4_4_9_15, \
call_4_4_10_0, \
call_4_4_10_1, \
call_4_4_10_2, \
call_4_4_10_3, \
call_4_4_10_4, \
call_4_4_10_5, \
call_4_4_10_6, \
call_4_4_10_7, \
call_4_4_10_8, \
call_4_4_10_9, \
call_4_4_10_10, \
call_4_4_10_11, \
call_4_4_10_12, \
call_4_4_10_13, \
call_4_4_10_14, \
call_4_4_10_15, \
call_4_4_11_0, \
call_4_4_11_1, \
call_4_4_11_2, \
call_4_4_11_3, \
call_4_4_11_4, \
call_4_4_11_5, \
call_4_4_11_6, \
call_4_4_11_7, \
call_4_4_11_8, \
call_4_4_11_9, \
call_4_4_11_10, \
call_4_4_11_11, \
call_4_4_11_12, \
call_4_4_11_13, \
call_4_4_11_14, \
call_4_4_11_15, \
call_4_4_12_0, \
call_4_4_12_1, \
call_4_4_12_2, \
call_4_4_12_3, \
call_4_4_12_4, \
call_4_4_12_5, \
call_4_4_12_6, \
call_4_4_12_7, \
call_4_4_12_8, \
call_4_4_12_9, \
call_4_4_12_10, \
call_4_4_12_11, \
call_4_4_12_12, \
call_4_4_12_13, \
call_4_4_12_14, \
call_4_4_12_15, \
call_4_4_13_0, \
call_4_4_13_1, \
call_4_4_13_2, \
call_4_4_13_3, \
call_4_4_13_4, \
call_4_4_13_5, \
call_4_4_13_6, \
call_4_4_13_7, \
call_4_4_13_8, \
call_4_4_13_9, \
call_4_4_13_10, \
call_4_4_13_11, \
call_4_4_13_12, \
call_4_4_13_13, \
call_4_4_13_14, \
call_4_4_13_15, \
call_4_4_14_0, \
call_4_4_14_1, \
call_4_4_14_2, \
call_4_4_14_3, \
call_4_4_14_4, \
call_4_4_14_5, \
call_4_4_14_6, \
call_4_4_14_7, \
call_4_4_14_8, \
call_4_4_14_9, \
call_4_4_14_10, \
call_4_4_14_11, \
call_4_4_14_12, \
call_4_4_14_13, \
call_4_4_14_14, \
call_4_4_14_15, \
call_4_4_15_0, \
call_4_4_15_1, \
call_4_4_15_2, \
call_4_4_15_3, \
call_4_4_15_4, \
call_4_4_15_5, \
call_4_4_15_6, \
call_4_4_15_7, \
call_4_4_15_8, \
call_4_4_15_9, \
call_4_4_15_10, \
call_4_4_15_11, \
call_4_4_15_12, \
call_4_4_15_13, \
call_4_4_15_14, \
call_4_4_15_15, \
call_4_5_0_0, \
call_4_5_0_1, \
call_4_5_0_2, \
call_4_5_0_3, \
call_4_5_0_4, \
call_4_5_0_5, \
call_4_5_0_6, \
call_4_5_0_7, \
call_4_5_0_8, \
call_4_5_0_9, \
call_4_5_0_10, \
call_4_5_0_11, \
call_4_5_0_12, \
call_4_5_0_13, \
call_4_5_0_14, \
call_4_5_0_15, \
call_4_5_1_0, \
call_4_5_1_1, \
call_4_5_1_2, \
call_4_5_1_3, \
call_4_5_1_4, \
call_4_5_1_5, \
call_4_5_1_6, \
call_4_5_1_7, \
call_4_5_1_8, \
call_4_5_1_9, \
call_4_5_1_10, \
call_4_5_1_11, \
call_4_5_1_12, \
call_4_5_1_13, \
call_4_5_1_14, \
call_4_5_1_15, \
call_4_5_2_0, \
call_4_5_2_1, \
call_4_5_2_2, \
call_4_5_2_3, \
call_4_5_2_4, \
call_4_5_2_5, \
call_4_5_2_6, \
call_4_5_2_7, \
call_4_5_2_8, \
call_4_5_2_9, \
call_4_5_2_10, \
call_4_5_2_11, \
call_4_5_2_12, \
call_4_5_2_13, \
call_4_5_2_14, \
call_4_5_2_15, \
call_4_5_3_0, \
call_4_5_3_1, \
call_4_5_3_2, \
call_4_5_3_3, \
call_4_5_3_4, \
call_4_5_3_5, \
call_4_5_3_6, \
call_4_5_3_7, \
call_4_5_3_8, \
call_4_5_3_9, \
call_4_5_3_10, \
call_4_5_3_11, \
call_4_5_3_12, \
call_4_5_3_13, \
call_4_5_3_14, \
call_4_5_3_15, \
call_4_5_4_0, \
call_4_5_4_1, \
call_4_5_4_2, \
call_4_5_4_3, \
call_4_5_4_4, \
call_4_5_4_5, \
call_4_5_4_6, \
call_4_5_4_7, \
call_4_5_4_8, \
call_4_5_4_9, \
call_4_5_4_10, \
call_4_5_4_11, \
call_4_5_4_12, \
call_4_5_4_13, \
call_4_5_4_14, \
call_4_5_4_15, \
call_4_5_5_0, \
call_4_5_5_1, \
call_4_5_5_2, \
call_4_5_5_3, \
call_4_5_5_4, \
call_4_5_5_5, \
call_4_5_5_6, \
call_4_5_5_7, \
call_4_5_5_8, \
call_4_5_5_9, \
call_4_5_5_10, \
call_4_5_5_11, \
call_4_5_5_12, \
call_4_5_5_13, \
call_4_5_5_14, \
call_4_5_5_15, \
call_4_5_6_0, \
call_4_5_6_1, \
call_4_5_6_2, \
call_4_5_6_3, \
call_4_5_6_4, \
call_4_5_6_5, \
call_4_5_6_6, \
call_4_5_6_7, \
call_4_5_6_8, \
call_4_5_6_9, \
call_4_5_6_10, \
call_4_5_6_11, \
call_4_5_6_12, \
call_4_5_6_13, \
call_4_5_6_14, \
call_4_5_6_15, \
call_4_5_7_0, \
call_4_5_7_1, \
call_4_5_7_2, \
call_4_5_7_3, \
call_4_5_7_4, \
call_4_5_7_5, \
call_4_5_7_6, \
call_4_5_7_7, \
call_4_5_7_8, \
call_4_5_7_9, \
call_4_5_7_10, \
call_4_5_7_11, \
call_4_5_7_12, \
call_4_5_7_13, \
call_4_5_7_14, \
call_4_5_7_15, \
call_4_5_8_0, \
call_4_5_8_1, \
call_4_5_8_2, \
call_4_5_8_3, \
call_4_5_8_4, \
call_4_5_8_5, \
call_4_5_8_6, \
call_4_5_8_7, \
call_4_5_8_8, \
call_4_5_8_9, \
call_4_5_8_10, \
call_4_5_8_11, \
call_4_5_8_12, \
call_4_5_8_13, \
call_4_5_8_14, \
call_4_5_8_15, \
call_4_5_9_0, \
call_4_5_9_1, \
call_4_5_9_2, \
call_4_5_9_3, \
call_4_5_9_4, \
call_4_5_9_5, \
call_4_5_9_6, \
call_4_5_9_7, \
call_4_5_9_8, \
call_4_5_9_9, \
call_4_5_9_10, \
call_4_5_9_11, \
call_4_5_9_12, \
call_4_5_9_13, \
call_4_5_9_14, \
call_4_5_9_15, \
call_4_5_10_0, \
call_4_5_10_1, \
call_4_5_10_2, \
call_4_5_10_3, \
call_4_5_10_4, \
call_4_5_10_5, \
call_4_5_10_6, \
call_4_5_10_7, \
call_4_5_10_8, \
call_4_5_10_9, \
call_4_5_10_10, \
call_4_5_10_11, \
call_4_5_10_12, \
call_4_5_10_13, \
call_4_5_10_14, \
call_4_5_10_15, \
call_4_5_11_0, \
call_4_5_11_1, \
call_4_5_11_2, \
call_4_5_11_3, \
call_4_5_11_4, \
call_4_5_11_5, \
call_4_5_11_6, \
call_4_5_11_7, \
call_4_5_11_8, \
call_4_5_11_9, \
call_4_5_11_10, \
call_4_5_11_11, \
call_4_5_11_12, \
call_4_5_11_13, \
call_4_5_11_14, \
call_4_5_11_15, \
call_4_5_12_0, \
call_4_5_12_1, \
call_4_5_12_2, \
call_4_5_12_3, \
call_4_5_12_4, \
call_4_5_12_5, \
call_4_5_12_6, \
call_4_5_12_7, \
call_4_5_12_8, \
call_4_5_12_9, \
call_4_5_12_10, \
call_4_5_12_11, \
call_4_5_12_12, \
call_4_5_12_13, \
call_4_5_12_14, \
call_4_5_12_15, \
call_4_5_13_0, \
call_4_5_13_1, \
call_4_5_13_2, \
call_4_5_13_3, \
call_4_5_13_4, \
call_4_5_13_5, \
call_4_5_13_6, \
call_4_5_13_7, \
call_4_5_13_8, \
call_4_5_13_9, \
call_4_5_13_10, \
call_4_5_13_11, \
call_4_5_13_12, \
call_4_5_13_13, \
call_4_5_13_14, \
call_4_5_13_15, \
call_4_5_14_0, \
call_4_5_14_1, \
call_4_5_14_2, \
call_4_5_14_3, \
call_4_5_14_4, \
call_4_5_14_5, \
call_4_5_14_6, \
call_4_5_14_7, \
call_4_5_14_8, \
call_4_5_14_9, \
call_4_5_14_10, \
call_4_5_14_11, \
call_4_5_14_12, \
call_4_5_14_13, \
call_4_5_14_14, \
call_4_5_14_15, \
call_4_5_15_0, \
call_4_5_15_1, \
call_4_5_15_2, \
call_4_5_15_3, \
call_4_5_15_4, \
call_4_5_15_5, \
call_4_5_15_6, \
call_4_5_15_7, \
call_4_5_15_8, \
call_4_5_15_9, \
call_4_5_15_10, \
call_4_5_15_11, \
call_4_5_15_12, \
call_4_5_15_13, \
call_4_5_15_14, \
call_4_5_15_15, \
call_4_6_0_0, \
call_4_6_0_1, \
call_4_6_0_2, \
call_4_6_0_3, \
call_4_6_0_4, \
call_4_6_0_5, \
call_4_6_0_6, \
call_4_6_0_7, \
call_4_6_0_8, \
call_4_6_0_9, \
call_4_6_0_10, \
call_4_6_0_11, \
call_4_6_0_12, \
call_4_6_0_13, \
call_4_6_0_14, \
call_4_6_0_15, \
call_4_6_1_0, \
call_4_6_1_1, \
call_4_6_1_2, \
call_4_6_1_3, \
call_4_6_1_4, \
call_4_6_1_5, \
call_4_6_1_6, \
call_4_6_1_7, \
call_4_6_1_8, \
call_4_6_1_9, \
call_4_6_1_10, \
call_4_6_1_11, \
call_4_6_1_12, \
call_4_6_1_13, \
call_4_6_1_14, \
call_4_6_1_15, \
call_4_6_2_0, \
call_4_6_2_1, \
call_4_6_2_2, \
call_4_6_2_3, \
call_4_6_2_4, \
call_4_6_2_5, \
call_4_6_2_6, \
call_4_6_2_7, \
call_4_6_2_8, \
call_4_6_2_9, \
call_4_6_2_10, \
call_4_6_2_11, \
call_4_6_2_12, \
call_4_6_2_13, \
call_4_6_2_14, \
call_4_6_2_15, \
call_4_6_3_0, \
call_4_6_3_1, \
call_4_6_3_2, \
call_4_6_3_3, \
call_4_6_3_4, \
call_4_6_3_5, \
call_4_6_3_6, \
call_4_6_3_7, \
call_4_6_3_8, \
call_4_6_3_9, \
call_4_6_3_10, \
call_4_6_3_11, \
call_4_6_3_12, \
call_4_6_3_13, \
call_4_6_3_14, \
call_4_6_3_15, \
call_4_6_4_0, \
call_4_6_4_1, \
call_4_6_4_2, \
call_4_6_4_3, \
call_4_6_4_4, \
call_4_6_4_5, \
call_4_6_4_6, \
call_4_6_4_7, \
call_4_6_4_8, \
call_4_6_4_9, \
call_4_6_4_10, \
call_4_6_4_11, \
call_4_6_4_12, \
call_4_6_4_13, \
call_4_6_4_14, \
call_4_6_4_15, \
call_4_6_5_0, \
call_4_6_5_1, \
call_4_6_5_2, \
call_4_6_5_3, \
call_4_6_5_4, \
call_4_6_5_5, \
call_4_6_5_6, \
call_4_6_5_7, \
call_4_6_5_8, \
call_4_6_5_9, \
call_4_6_5_10, \
call_4_6_5_11, \
call_4_6_5_12, \
call_4_6_5_13, \
call_4_6_5_14, \
call_4_6_5_15, \
call_4_6_6_0, \
call_4_6_6_1, \
call_4_6_6_2, \
call_4_6_6_3, \
call_4_6_6_4, \
call_4_6_6_5, \
call_4_6_6_6, \
call_4_6_6_7, \
call_4_6_6_8, \
call_4_6_6_9, \
call_4_6_6_10, \
call_4_6_6_11, \
call_4_6_6_12, \
call_4_6_6_13, \
call_4_6_6_14, \
call_4_6_6_15, \
call_4_6_7_0, \
call_4_6_7_1, \
call_4_6_7_2, \
call_4_6_7_3, \
call_4_6_7_4, \
call_4_6_7_5, \
call_4_6_7_6, \
call_4_6_7_7, \
call_4_6_7_8, \
call_4_6_7_9, \
call_4_6_7_10, \
call_4_6_7_11, \
call_4_6_7_12, \
call_4_6_7_13, \
call_4_6_7_14, \
call_4_6_7_15, \
call_4_6_8_0, \
call_4_6_8_1, \
call_4_6_8_2, \
call_4_6_8_3, \
call_4_6_8_4, \
call_4_6_8_5, \
call_4_6_8_6, \
call_4_6_8_7, \
call_4_6_8_8, \
call_4_6_8_9, \
call_4_6_8_10, \
call_4_6_8_11, \
call_4_6_8_12, \
call_4_6_8_13, \
call_4_6_8_14, \
call_4_6_8_15, \
call_4_6_9_0, \
call_4_6_9_1, \
call_4_6_9_2, \
call_4_6_9_3, \
call_4_6_9_4, \
call_4_6_9_5, \
call_4_6_9_6, \
call_4_6_9_7, \
call_4_6_9_8, \
call_4_6_9_9, \
call_4_6_9_10, \
call_4_6_9_11, \
call_4_6_9_12, \
call_4_6_9_13, \
call_4_6_9_14, \
call_4_6_9_15, \
call_4_6_10_0, \
call_4_6_10_1, \
call_4_6_10_2, \
call_4_6_10_3, \
call_4_6_10_4, \
call_4_6_10_5, \
call_4_6_10_6, \
call_4_6_10_7, \
call_4_6_10_8, \
call_4_6_10_9, \
call_4_6_10_10, \
call_4_6_10_11, \
call_4_6_10_12, \
call_4_6_10_13, \
call_4_6_10_14, \
call_4_6_10_15, \
call_4_6_11_0, \
call_4_6_11_1, \
call_4_6_11_2, \
call_4_6_11_3, \
call_4_6_11_4, \
call_4_6_11_5, \
call_4_6_11_6, \
call_4_6_11_7, \
call_4_6_11_8, \
call_4_6_11_9, \
call_4_6_11_10, \
call_4_6_11_11, \
call_4_6_11_12, \
call_4_6_11_13, \
call_4_6_11_14, \
call_4_6_11_15, \
call_4_6_12_0, \
call_4_6_12_1, \
call_4_6_12_2, \
call_4_6_12_3, \
call_4_6_12_4, \
call_4_6_12_5, \
call_4_6_12_6, \
call_4_6_12_7, \
call_4_6_12_8, \
call_4_6_12_9, \
call_4_6_12_10, \
call_4_6_12_11, \
call_4_6_12_12, \
call_4_6_12_13, \
call_4_6_12_14, \
call_4_6_12_15, \
call_4_6_13_0, \
call_4_6_13_1, \
call_4_6_13_2, \
call_4_6_13_3, \
call_4_6_13_4, \
call_4_6_13_5, \
call_4_6_13_6, \
call_4_6_13_7, \
call_4_6_13_8, \
call_4_6_13_9, \
call_4_6_13_10, \
call_4_6_13_11, \
call_4_6_13_12, \
call_4_6_13_13, \
call_4_6_13_14, \
call_4_6_13_15, \
call_4_6_14_0, \
call_4_6_14_1, \
call_4_6_14_2, \
call_4_6_14_3, \
call_4_6_14_4, \
call_4_6_14_5, \
call_4_6_14_6, \
call_4_6_14_7, \
call_4_6_14_8, \
call_4_6_14_9, \
call_4_6_14_10, \
call_4_6_14_11, \
call_4_6_14_12, \
call_4_6_14_13, \
call_4_6_14_14, \
call_4_6_14_15, \
call_4_6_15_0, \
call_4_6_15_1, \
call_4_6_15_2, \
call_4_6_15_3, \
call_4_6_15_4, \
call_4_6_15_5, \
call_4_6_15_6, \
call_4_6_15_7, \
call_4_6_15_8, \
call_4_6_15_9, \
call_4_6_15_10, \
call_4_6_15_11, \
call_4_6_15_12, \
call_4_6_15_13, \
call_4_6_15_14, \
call_4_6_15_15, \
call_4_7_0_0, \
call_4_7_0_1, \
call_4_7_0_2, \
call_4_7_0_3, \
call_4_7_0_4, \
call_4_7_0_5, \
call_4_7_0_6, \
call_4_7_0_7, \
call_4_7_0_8, \
call_4_7_0_9, \
call_4_7_0_10, \
call_4_7_0_11, \
call_4_7_0_12, \
call_4_7_0_13, \
call_4_7_0_14, \
call_4_7_0_15, \
call_4_7_1_0, \
call_4_7_1_1, \
call_4_7_1_2, \
call_4_7_1_3, \
call_4_7_1_4, \
call_4_7_1_5, \
call_4_7_1_6, \
call_4_7_1_7, \
call_4_7_1_8, \
call_4_7_1_9, \
call_4_7_1_10, \
call_4_7_1_11, \
call_4_7_1_12, \
call_4_7_1_13, \
call_4_7_1_14, \
call_4_7_1_15, \
call_4_7_2_0, \
call_4_7_2_1, \
call_4_7_2_2, \
call_4_7_2_3, \
call_4_7_2_4, \
call_4_7_2_5, \
call_4_7_2_6, \
call_4_7_2_7, \
call_4_7_2_8, \
call_4_7_2_9, \
call_4_7_2_10, \
call_4_7_2_11, \
call_4_7_2_12, \
call_4_7_2_13, \
call_4_7_2_14, \
call_4_7_2_15, \
call_4_7_3_0, \
call_4_7_3_1, \
call_4_7_3_2, \
call_4_7_3_3, \
call_4_7_3_4, \
call_4_7_3_5, \
call_4_7_3_6, \
call_4_7_3_7, \
call_4_7_3_8, \
call_4_7_3_9, \
call_4_7_3_10, \
call_4_7_3_11, \
call_4_7_3_12, \
call_4_7_3_13, \
call_4_7_3_14, \
call_4_7_3_15, \
call_4_7_4_0, \
call_4_7_4_1, \
call_4_7_4_2, \
call_4_7_4_3, \
call_4_7_4_4, \
call_4_7_4_5, \
call_4_7_4_6, \
call_4_7_4_7, \
call_4_7_4_8, \
call_4_7_4_9, \
call_4_7_4_10, \
call_4_7_4_11, \
call_4_7_4_12, \
call_4_7_4_13, \
call_4_7_4_14, \
call_4_7_4_15, \
call_4_7_5_0, \
call_4_7_5_1, \
call_4_7_5_2, \
call_4_7_5_3, \
call_4_7_5_4, \
call_4_7_5_5, \
call_4_7_5_6, \
call_4_7_5_7, \
call_4_7_5_8, \
call_4_7_5_9, \
call_4_7_5_10, \
call_4_7_5_11, \
call_4_7_5_12, \
call_4_7_5_13, \
call_4_7_5_14, \
call_4_7_5_15, \
call_4_7_6_0, \
call_4_7_6_1, \
call_4_7_6_2, \
call_4_7_6_3, \
call_4_7_6_4, \
call_4_7_6_5, \
call_4_7_6_6, \
call_4_7_6_7, \
call_4_7_6_8, \
call_4_7_6_9, \
call_4_7_6_10, \
call_4_7_6_11, \
call_4_7_6_12, \
call_4_7_6_13, \
call_4_7_6_14, \
call_4_7_6_15, \
call_4_7_7_0, \
call_4_7_7_1, \
call_4_7_7_2, \
call_4_7_7_3, \
call_4_7_7_4, \
call_4_7_7_5, \
call_4_7_7_6, \
call_4_7_7_7, \
call_4_7_7_8, \
call_4_7_7_9, \
call_4_7_7_10, \
call_4_7_7_11, \
call_4_7_7_12, \
call_4_7_7_13, \
call_4_7_7_14, \
call_4_7_7_15, \
call_4_7_8_0, \
call_4_7_8_1, \
call_4_7_8_2, \
call_4_7_8_3, \
call_4_7_8_4, \
call_4_7_8_5, \
call_4_7_8_6, \
call_4_7_8_7, \
call_4_7_8_8, \
call_4_7_8_9, \
call_4_7_8_10, \
call_4_7_8_11, \
call_4_7_8_12, \
call_4_7_8_13, \
call_4_7_8_14, \
call_4_7_8_15, \
call_4_7_9_0, \
call_4_7_9_1, \
call_4_7_9_2, \
call_4_7_9_3, \
call_4_7_9_4, \
call_4_7_9_5, \
call_4_7_9_6, \
call_4_7_9_7, \
call_4_7_9_8, \
call_4_7_9_9, \
call_4_7_9_10, \
call_4_7_9_11, \
call_4_7_9_12, \
call_4_7_9_13, \
call_4_7_9_14, \
call_4_7_9_15, \
call_4_7_10_0, \
call_4_7_10_1, \
call_4_7_10_2, \
call_4_7_10_3, \
call_4_7_10_4, \
call_4_7_10_5, \
call_4_7_10_6, \
call_4_7_10_7, \
call_4_7_10_8, \
call_4_7_10_9, \
call_4_7_10_10, \
call_4_7_10_11, \
call_4_7_10_12, \
call_4_7_10_13, \
call_4_7_10_14, \
call_4_7_10_15, \
call_4_7_11_0, \
call_4_7_11_1, \
call_4_7_11_2, \
call_4_7_11_3, \
call_4_7_11_4, \
call_4_7_11_5, \
call_4_7_11_6, \
call_4_7_11_7, \
call_4_7_11_8, \
call_4_7_11_9, \
call_4_7_11_10, \
call_4_7_11_11, \
call_4_7_11_12, \
call_4_7_11_13, \
call_4_7_11_14, \
call_4_7_11_15, \
call_4_7_12_0, \
call_4_7_12_1, \
call_4_7_12_2, \
call_4_7_12_3, \
call_4_7_12_4, \
call_4_7_12_5, \
call_4_7_12_6, \
call_4_7_12_7, \
call_4_7_12_8, \
call_4_7_12_9, \
call_4_7_12_10, \
call_4_7_12_11, \
call_4_7_12_12, \
call_4_7_12_13, \
call_4_7_12_14, \
call_4_7_12_15, \
call_4_7_13_0, \
call_4_7_13_1, \
call_4_7_13_2, \
call_4_7_13_3, \
call_4_7_13_4, \
call_4_7_13_5, \
call_4_7_13_6, \
call_4_7_13_7, \
call_4_7_13_8, \
call_4_7_13_9, \
call_4_7_13_10, \
call_4_7_13_11, \
call_4_7_13_12, \
call_4_7_13_13, \
call_4_7_13_14, \
call_4_7_13_15, \
call_4_7_14_0, \
call_4_7_14_1, \
call_4_7_14_2, \
call_4_7_14_3, \
call_4_7_14_4, \
call_4_7_14_5, \
call_4_7_14_6, \
call_4_7_14_7, \
call_4_7_14_8, \
call_4_7_14_9, \
call_4_7_14_10, \
call_4_7_14_11, \
call_4_7_14_12, \
call_4_7_14_13, \
call_4_7_14_14, \
call_4_7_14_15, \
call_4_7_15_0, \
call_4_7_15_1, \
call_4_7_15_2, \
call_4_7_15_3, \
call_4_7_15_4, \
call_4_7_15_5, \
call_4_7_15_6, \
call_4_7_15_7, \
call_4_7_15_8, \
call_4_7_15_9, \
call_4_7_15_10, \
call_4_7_15_11, \
call_4_7_15_12, \
call_4_7_15_13, \
call_4_7_15_14, \
call_4_7_15_15, \
call_4_8_0_0, \
call_4_8_0_1, \
call_4_8_0_2, \
call_4_8_0_3, \
call_4_8_0_4, \
call_4_8_0_5, \
call_4_8_0_6, \
call_4_8_0_7, \
call_4_8_0_8, \
call_4_8_0_9, \
call_4_8_0_10, \
call_4_8_0_11, \
call_4_8_0_12, \
call_4_8_0_13, \
call_4_8_0_14, \
call_4_8_0_15, \
call_4_8_1_0, \
call_4_8_1_1, \
call_4_8_1_2, \
call_4_8_1_3, \
call_4_8_1_4, \
call_4_8_1_5, \
call_4_8_1_6, \
call_4_8_1_7, \
call_4_8_1_8, \
call_4_8_1_9, \
call_4_8_1_10, \
call_4_8_1_11, \
call_4_8_1_12, \
call_4_8_1_13, \
call_4_8_1_14, \
call_4_8_1_15, \
call_4_8_2_0, \
call_4_8_2_1, \
call_4_8_2_2, \
call_4_8_2_3, \
call_4_8_2_4, \
call_4_8_2_5, \
call_4_8_2_6, \
call_4_8_2_7, \
call_4_8_2_8, \
call_4_8_2_9, \
call_4_8_2_10, \
call_4_8_2_11, \
call_4_8_2_12, \
call_4_8_2_13, \
call_4_8_2_14, \
call_4_8_2_15, \
call_4_8_3_0, \
call_4_8_3_1, \
call_4_8_3_2, \
call_4_8_3_3, \
call_4_8_3_4, \
call_4_8_3_5, \
call_4_8_3_6, \
call_4_8_3_7, \
call_4_8_3_8, \
call_4_8_3_9, \
call_4_8_3_10, \
call_4_8_3_11, \
call_4_8_3_12, \
call_4_8_3_13, \
call_4_8_3_14, \
call_4_8_3_15, \
call_4_8_4_0, \
call_4_8_4_1, \
call_4_8_4_2, \
call_4_8_4_3, \
call_4_8_4_4, \
call_4_8_4_5, \
call_4_8_4_6, \
call_4_8_4_7, \
call_4_8_4_8, \
call_4_8_4_9, \
call_4_8_4_10, \
call_4_8_4_11, \
call_4_8_4_12, \
call_4_8_4_13, \
call_4_8_4_14, \
call_4_8_4_15, \
call_4_8_5_0, \
call_4_8_5_1, \
call_4_8_5_2, \
call_4_8_5_3, \
call_4_8_5_4, \
call_4_8_5_5, \
call_4_8_5_6, \
call_4_8_5_7, \
call_4_8_5_8, \
call_4_8_5_9, \
call_4_8_5_10, \
call_4_8_5_11, \
call_4_8_5_12, \
call_4_8_5_13, \
call_4_8_5_14, \
call_4_8_5_15, \
call_4_8_6_0, \
call_4_8_6_1, \
call_4_8_6_2, \
call_4_8_6_3, \
call_4_8_6_4, \
call_4_8_6_5, \
call_4_8_6_6, \
call_4_8_6_7, \
call_4_8_6_8, \
call_4_8_6_9, \
call_4_8_6_10, \
call_4_8_6_11, \
call_4_8_6_12, \
call_4_8_6_13, \
call_4_8_6_14, \
call_4_8_6_15, \
call_4_8_7_0, \
call_4_8_7_1, \
call_4_8_7_2, \
call_4_8_7_3, \
call_4_8_7_4, \
call_4_8_7_5, \
call_4_8_7_6, \
call_4_8_7_7, \
call_4_8_7_8, \
call_4_8_7_9, \
call_4_8_7_10, \
call_4_8_7_11, \
call_4_8_7_12, \
call_4_8_7_13, \
call_4_8_7_14, \
call_4_8_7_15, \
call_4_8_8_0, \
call_4_8_8_1, \
call_4_8_8_2, \
call_4_8_8_3, \
call_4_8_8_4, \
call_4_8_8_5, \
call_4_8_8_6, \
call_4_8_8_7, \
call_4_8_8_8, \
call_4_8_8_9, \
call_4_8_8_10, \
call_4_8_8_11, \
call_4_8_8_12, \
call_4_8_8_13, \
call_4_8_8_14, \
call_4_8_8_15, \
call_4_8_9_0, \
call_4_8_9_1, \
call_4_8_9_2, \
call_4_8_9_3, \
call_4_8_9_4, \
call_4_8_9_5, \
call_4_8_9_6, \
call_4_8_9_7, \
call_4_8_9_8, \
call_4_8_9_9, \
call_4_8_9_10, \
call_4_8_9_11, \
call_4_8_9_12, \
call_4_8_9_13, \
call_4_8_9_14, \
call_4_8_9_15, \
call_4_8_10_0, \
call_4_8_10_1, \
call_4_8_10_2, \
call_4_8_10_3, \
call_4_8_10_4, \
call_4_8_10_5, \
call_4_8_10_6, \
call_4_8_10_7, \
call_4_8_10_8, \
call_4_8_10_9, \
call_4_8_10_10, \
call_4_8_10_11, \
call_4_8_10_12, \
call_4_8_10_13, \
call_4_8_10_14, \
call_4_8_10_15, \
call_4_8_11_0, \
call_4_8_11_1, \
call_4_8_11_2, \
call_4_8_11_3, \
call_4_8_11_4, \
call_4_8_11_5, \
call_4_8_11_6, \
call_4_8_11_7, \
call_4_8_11_8, \
call_4_8_11_9, \
call_4_8_11_10, \
call_4_8_11_11, \
call_4_8_11_12, \
call_4_8_11_13, \
call_4_8_11_14, \
call_4_8_11_15, \
call_4_8_12_0, \
call_4_8_12_1, \
call_4_8_12_2, \
call_4_8_12_3, \
call_4_8_12_4, \
call_4_8_12_5, \
call_4_8_12_6, \
call_4_8_12_7, \
call_4_8_12_8, \
call_4_8_12_9, \
call_4_8_12_10, \
call_4_8_12_11, \
call_4_8_12_12, \
call_4_8_12_13, \
call_4_8_12_14, \
call_4_8_12_15, \
call_4_8_13_0, \
call_4_8_13_1, \
call_4_8_13_2, \
call_4_8_13_3, \
call_4_8_13_4, \
call_4_8_13_5, \
call_4_8_13_6, \
call_4_8_13_7, \
call_4_8_13_8, \
call_4_8_13_9, \
call_4_8_13_10, \
call_4_8_13_11, \
call_4_8_13_12, \
call_4_8_13_13, \
call_4_8_13_14, \
call_4_8_13_15, \
call_4_8_14_0, \
call_4_8_14_1, \
call_4_8_14_2, \
call_4_8_14_3, \
call_4_8_14_4, \
call_4_8_14_5, \
call_4_8_14_6, \
call_4_8_14_7, \
call_4_8_14_8, \
call_4_8_14_9, \
call_4_8_14_10, \
call_4_8_14_11, \
call_4_8_14_12, \
call_4_8_14_13, \
call_4_8_14_14, \
call_4_8_14_15, \
call_4_8_15_0, \
call_4_8_15_1, \
call_4_8_15_2, \
call_4_8_15_3, \
call_4_8_15_4, \
call_4_8_15_5, \
call_4_8_15_6, \
call_4_8_15_7, \
call_4_8_15_8, \
call_4_8_15_9, \
call_4_8_15_10, \
call_4_8_15_11, \
call_4_8_15_12, \
call_4_8_15_13, \
call_4_8_15_14, \
call_4_8_15_15, \
call_4_9_0_0, \
call_4_9_0_1, \
call_4_9_0_2, \
call_4_9_0_3, \
call_4_9_0_4, \
call_4_9_0_5, \
call_4_9_0_6, \
call_4_9_0_7, \
call_4_9_0_8, \
call_4_9_0_9, \
call_4_9_0_10, \
call_4_9_0_11, \
call_4_9_0_12, \
call_4_9_0_13, \
call_4_9_0_14, \
call_4_9_0_15, \
call_4_9_1_0, \
call_4_9_1_1, \
call_4_9_1_2, \
call_4_9_1_3, \
call_4_9_1_4, \
call_4_9_1_5, \
call_4_9_1_6, \
call_4_9_1_7, \
call_4_9_1_8, \
call_4_9_1_9, \
call_4_9_1_10, \
call_4_9_1_11, \
call_4_9_1_12, \
call_4_9_1_13, \
call_4_9_1_14, \
call_4_9_1_15, \
call_4_9_2_0, \
call_4_9_2_1, \
call_4_9_2_2, \
call_4_9_2_3, \
call_4_9_2_4, \
call_4_9_2_5, \
call_4_9_2_6, \
call_4_9_2_7, \
call_4_9_2_8, \
call_4_9_2_9, \
call_4_9_2_10, \
call_4_9_2_11, \
call_4_9_2_12, \
call_4_9_2_13, \
call_4_9_2_14, \
call_4_9_2_15, \
call_4_9_3_0, \
call_4_9_3_1, \
call_4_9_3_2, \
call_4_9_3_3, \
call_4_9_3_4, \
call_4_9_3_5, \
call_4_9_3_6, \
call_4_9_3_7, \
call_4_9_3_8, \
call_4_9_3_9, \
call_4_9_3_10, \
call_4_9_3_11, \
call_4_9_3_12, \
call_4_9_3_13, \
call_4_9_3_14, \
call_4_9_3_15, \
call_4_9_4_0, \
call_4_9_4_1, \
call_4_9_4_2, \
call_4_9_4_3, \
call_4_9_4_4, \
call_4_9_4_5, \
call_4_9_4_6, \
call_4_9_4_7, \
call_4_9_4_8, \
call_4_9_4_9, \
call_4_9_4_10, \
call_4_9_4_11, \
call_4_9_4_12, \
call_4_9_4_13, \
call_4_9_4_14, \
call_4_9_4_15, \
call_4_9_5_0, \
call_4_9_5_1, \
call_4_9_5_2, \
call_4_9_5_3, \
call_4_9_5_4, \
call_4_9_5_5, \
call_4_9_5_6, \
call_4_9_5_7, \
call_4_9_5_8, \
call_4_9_5_9, \
call_4_9_5_10, \
call_4_9_5_11, \
call_4_9_5_12, \
call_4_9_5_13, \
call_4_9_5_14, \
call_4_9_5_15, \
call_4_9_6_0, \
call_4_9_6_1, \
call_4_9_6_2, \
call_4_9_6_3, \
call_4_9_6_4, \
call_4_9_6_5, \
call_4_9_6_6, \
call_4_9_6_7, \
call_4_9_6_8, \
call_4_9_6_9, \
call_4_9_6_10, \
call_4_9_6_11, \
call_4_9_6_12, \
call_4_9_6_13, \
call_4_9_6_14, \
call_4_9_6_15, \
call_4_9_7_0, \
call_4_9_7_1, \
call_4_9_7_2, \
call_4_9_7_3, \
call_4_9_7_4, \
call_4_9_7_5, \
call_4_9_7_6, \
call_4_9_7_7, \
call_4_9_7_8, \
call_4_9_7_9, \
call_4_9_7_10, \
call_4_9_7_11, \
call_4_9_7_12, \
call_4_9_7_13, \
call_4_9_7_14, \
call_4_9_7_15, \
call_4_9_8_0, \
call_4_9_8_1, \
call_4_9_8_2, \
call_4_9_8_3, \
call_4_9_8_4, \
call_4_9_8_5, \
call_4_9_8_6, \
call_4_9_8_7, \
call_4_9_8_8, \
call_4_9_8_9, \
call_4_9_8_10, \
call_4_9_8_11, \
call_4_9_8_12, \
call_4_9_8_13, \
call_4_9_8_14, \
call_4_9_8_15, \
call_4_9_9_0, \
call_4_9_9_1, \
call_4_9_9_2, \
call_4_9_9_3, \
call_4_9_9_4, \
call_4_9_9_5, \
call_4_9_9_6, \
call_4_9_9_7, \
call_4_9_9_8, \
call_4_9_9_9, \
call_4_9_9_10, \
call_4_9_9_11, \
call_4_9_9_12, \
call_4_9_9_13, \
call_4_9_9_14, \
call_4_9_9_15, \
call_4_9_10_0, \
call_4_9_10_1, \
call_4_9_10_2, \
call_4_9_10_3, \
call_4_9_10_4, \
call_4_9_10_5, \
call_4_9_10_6, \
call_4_9_10_7, \
call_4_9_10_8, \
call_4_9_10_9, \
call_4_9_10_10, \
call_4_9_10_11, \
call_4_9_10_12, \
call_4_9_10_13, \
call_4_9_10_14, \
call_4_9_10_15, \
call_4_9_11_0, \
call_4_9_11_1, \
call_4_9_11_2, \
call_4_9_11_3, \
call_4_9_11_4, \
call_4_9_11_5, \
call_4_9_11_6, \
call_4_9_11_7, \
call_4_9_11_8, \
call_4_9_11_9, \
call_4_9_11_10, \
call_4_9_11_11, \
call_4_9_11_12, \
call_4_9_11_13, \
call_4_9_11_14, \
call_4_9_11_15, \
call_4_9_12_0, \
call_4_9_12_1, \
call_4_9_12_2, \
call_4_9_12_3, \
call_4_9_12_4, \
call_4_9_12_5, \
call_4_9_12_6, \
call_4_9_12_7, \
call_4_9_12_8, \
call_4_9_12_9, \
call_4_9_12_10, \
call_4_9_12_11, \
call_4_9_12_12, \
call_4_9_12_13, \
call_4_9_12_14, \
call_4_9_12_15, \
call_4_9_13_0, \
call_4_9_13_1, \
call_4_9_13_2, \
call_4_9_13_3, \
call_4_9_13_4, \
call_4_9_13_5, \
call_4_9_13_6, \
call_4_9_13_7, \
call_4_9_13_8, \
call_4_9_13_9, \
call_4_9_13_10, \
call_4_9_13_11, \
call_4_9_13_12, \
call_4_9_13_13, \
call_4_9_13_14, \
call_4_9_13_15, \
call_4_9_14_0, \
call_4_9_14_1, \
call_4_9_14_2, \
call_4_9_14_3, \
call_4_9_14_4, \
call_4_9_14_5, \
call_4_9_14_6, \
call_4_9_14_7, \
call_4_9_14_8, \
call_4_9_14_9, \
call_4_9_14_10, \
call_4_9_14_11, \
call_4_9_14_12, \
call_4_9_14_13, \
call_4_9_14_14, \
call_4_9_14_15, \
call_4_9_15_0, \
call_4_9_15_1, \
call_4_9_15_2, \
call_4_9_15_3, \
call_4_9_15_4, \
call_4_9_15_5, \
call_4_9_15_6, \
call_4_9_15_7, \
call_4_9_15_8, \
call_4_9_15_9, \
call_4_9_15_10, \
call_4_9_15_11, \
call_4_9_15_12, \
call_4_9_15_13, \
call_4_9_15_14, \
call_4_9_15_15, \
call_4_10_0_0, \
call_4_10_0_1, \
call_4_10_0_2, \
call_4_10_0_3, \
call_4_10_0_4, \
call_4_10_0_5, \
call_4_10_0_6, \
call_4_10_0_7, \
call_4_10_0_8, \
call_4_10_0_9, \
call_4_10_0_10, \
call_4_10_0_11, \
call_4_10_0_12, \
call_4_10_0_13, \
call_4_10_0_14, \
call_4_10_0_15, \
call_4_10_1_0, \
call_4_10_1_1, \
call_4_10_1_2, \
call_4_10_1_3, \
call_4_10_1_4, \
call_4_10_1_5, \
call_4_10_1_6, \
call_4_10_1_7, \
call_4_10_1_8, \
call_4_10_1_9, \
call_4_10_1_10, \
call_4_10_1_11, \
call_4_10_1_12, \
call_4_10_1_13, \
call_4_10_1_14, \
call_4_10_1_15, \
call_4_10_2_0, \
call_4_10_2_1, \
call_4_10_2_2, \
call_4_10_2_3, \
call_4_10_2_4, \
call_4_10_2_5, \
call_4_10_2_6, \
call_4_10_2_7, \
call_4_10_2_8, \
call_4_10_2_9, \
call_4_10_2_10, \
call_4_10_2_11, \
call_4_10_2_12, \
call_4_10_2_13, \
call_4_10_2_14, \
call_4_10_2_15, \
call_4_10_3_0, \
call_4_10_3_1, \
call_4_10_3_2, \
call_4_10_3_3, \
call_4_10_3_4, \
call_4_10_3_5, \
call_4_10_3_6, \
call_4_10_3_7, \
call_4_10_3_8, \
call_4_10_3_9, \
call_4_10_3_10, \
call_4_10_3_11, \
call_4_10_3_12, \
call_4_10_3_13, \
call_4_10_3_14, \
call_4_10_3_15, \
call_4_10_4_0, \
call_4_10_4_1, \
call_4_10_4_2, \
call_4_10_4_3, \
call_4_10_4_4, \
call_4_10_4_5, \
call_4_10_4_6, \
call_4_10_4_7, \
call_4_10_4_8, \
call_4_10_4_9, \
call_4_10_4_10, \
call_4_10_4_11, \
call_4_10_4_12, \
call_4_10_4_13, \
call_4_10_4_14, \
call_4_10_4_15, \
call_4_10_5_0, \
call_4_10_5_1, \
call_4_10_5_2, \
call_4_10_5_3, \
call_4_10_5_4, \
call_4_10_5_5, \
call_4_10_5_6, \
call_4_10_5_7, \
call_4_10_5_8, \
call_4_10_5_9, \
call_4_10_5_10, \
call_4_10_5_11, \
call_4_10_5_12, \
call_4_10_5_13, \
call_4_10_5_14, \
call_4_10_5_15, \
call_4_10_6_0, \
call_4_10_6_1, \
call_4_10_6_2, \
call_4_10_6_3, \
call_4_10_6_4, \
call_4_10_6_5, \
call_4_10_6_6, \
call_4_10_6_7, \
call_4_10_6_8, \
call_4_10_6_9, \
call_4_10_6_10, \
call_4_10_6_11, \
call_4_10_6_12, \
call_4_10_6_13, \
call_4_10_6_14, \
call_4_10_6_15, \
call_4_10_7_0, \
call_4_10_7_1, \
call_4_10_7_2, \
call_4_10_7_3, \
call_4_10_7_4, \
call_4_10_7_5, \
call_4_10_7_6, \
call_4_10_7_7, \
call_4_10_7_8, \
call_4_10_7_9, \
call_4_10_7_10, \
call_4_10_7_11, \
call_4_10_7_12, \
call_4_10_7_13, \
call_4_10_7_14, \
call_4_10_7_15, \
call_4_10_8_0, \
call_4_10_8_1, \
call_4_10_8_2, \
call_4_10_8_3, \
call_4_10_8_4, \
call_4_10_8_5, \
call_4_10_8_6, \
call_4_10_8_7, \
call_4_10_8_8, \
call_4_10_8_9, \
call_4_10_8_10, \
call_4_10_8_11, \
call_4_10_8_12, \
call_4_10_8_13, \
call_4_10_8_14, \
call_4_10_8_15, \
call_4_10_9_0, \
call_4_10_9_1, \
call_4_10_9_2, \
call_4_10_9_3, \
call_4_10_9_4, \
call_4_10_9_5, \
call_4_10_9_6, \
call_4_10_9_7, \
call_4_10_9_8, \
call_4_10_9_9, \
call_4_10_9_10, \
call_4_10_9_11, \
call_4_10_9_12, \
call_4_10_9_13, \
call_4_10_9_14, \
call_4_10_9_15, \
call_4_10_10_0, \
call_4_10_10_1, \
call_4_10_10_2, \
call_4_10_10_3, \
call_4_10_10_4, \
call_4_10_10_5, \
call_4_10_10_6, \
call_4_10_10_7, \
call_4_10_10_8, \
call_4_10_10_9, \
call_4_10_10_10, \
call_4_10_10_11, \
call_4_10_10_12, \
call_4_10_10_13, \
call_4_10_10_14, \
call_4_10_10_15, \
call_4_10_11_0, \
call_4_10_11_1, \
call_4_10_11_2, \
call_4_10_11_3, \
call_4_10_11_4, \
call_4_10_11_5, \
call_4_10_11_6, \
call_4_10_11_7, \
call_4_10_11_8, \
call_4_10_11_9, \
call_4_10_11_10, \
call_4_10_11_11, \
call_4_10_11_12, \
call_4_10_11_13, \
call_4_10_11_14, \
call_4_10_11_15, \
call_4_10_12_0, \
call_4_10_12_1, \
call_4_10_12_2, \
call_4_10_12_3, \
call_4_10_12_4, \
call_4_10_12_5, \
call_4_10_12_6, \
call_4_10_12_7, \
call_4_10_12_8, \
call_4_10_12_9, \
call_4_10_12_10, \
call_4_10_12_11, \
call_4_10_12_12, \
call_4_10_12_13, \
call_4_10_12_14, \
call_4_10_12_15, \
call_4_10_13_0, \
call_4_10_13_1, \
call_4_10_13_2, \
call_4_10_13_3, \
call_4_10_13_4, \
call_4_10_13_5, \
call_4_10_13_6, \
call_4_10_13_7, \
call_4_10_13_8, \
call_4_10_13_9, \
call_4_10_13_10, \
call_4_10_13_11, \
call_4_10_13_12, \
call_4_10_13_13, \
call_4_10_13_14, \
call_4_10_13_15, \
call_4_10_14_0, \
call_4_10_14_1, \
call_4_10_14_2, \
call_4_10_14_3, \
call_4_10_14_4, \
call_4_10_14_5, \
call_4_10_14_6, \
call_4_10_14_7, \
call_4_10_14_8, \
call_4_10_14_9, \
call_4_10_14_10, \
call_4_10_14_11, \
call_4_10_14_12, \
call_4_10_14_13, \
call_4_10_14_14, \
call_4_10_14_15, \
call_4_10_15_0, \
call_4_10_15_1, \
call_4_10_15_2, \
call_4_10_15_3, \
call_4_10_15_4, \
call_4_10_15_5, \
call_4_10_15_6, \
call_4_10_15_7, \
call_4_10_15_8, \
call_4_10_15_9, \
call_4_10_15_10, \
call_4_10_15_11, \
call_4_10_15_12, \
call_4_10_15_13, \
call_4_10_15_14, \
call_4_10_15_15, \
call_4_11_0_0, \
call_4_11_0_1, \
call_4_11_0_2, \
call_4_11_0_3, \
call_4_11_0_4, \
call_4_11_0_5, \
call_4_11_0_6, \
call_4_11_0_7, \
call_4_11_0_8, \
call_4_11_0_9, \
call_4_11_0_10, \
call_4_11_0_11, \
call_4_11_0_12, \
call_4_11_0_13, \
call_4_11_0_14, \
call_4_11_0_15, \
call_4_11_1_0, \
call_4_11_1_1, \
call_4_11_1_2, \
call_4_11_1_3, \
call_4_11_1_4, \
call_4_11_1_5, \
call_4_11_1_6, \
call_4_11_1_7, \
call_4_11_1_8, \
call_4_11_1_9, \
call_4_11_1_10, \
call_4_11_1_11, \
call_4_11_1_12, \
call_4_11_1_13, \
call_4_11_1_14, \
call_4_11_1_15, \
call_4_11_2_0, \
call_4_11_2_1, \
call_4_11_2_2, \
call_4_11_2_3, \
call_4_11_2_4, \
call_4_11_2_5, \
call_4_11_2_6, \
call_4_11_2_7, \
call_4_11_2_8, \
call_4_11_2_9, \
call_4_11_2_10, \
call_4_11_2_11, \
call_4_11_2_12, \
call_4_11_2_13, \
call_4_11_2_14, \
call_4_11_2_15, \
call_4_11_3_0, \
call_4_11_3_1, \
call_4_11_3_2, \
call_4_11_3_3, \
call_4_11_3_4, \
call_4_11_3_5, \
call_4_11_3_6, \
call_4_11_3_7, \
call_4_11_3_8, \
call_4_11_3_9, \
call_4_11_3_10, \
call_4_11_3_11, \
call_4_11_3_12, \
call_4_11_3_13, \
call_4_11_3_14, \
call_4_11_3_15, \
call_4_11_4_0, \
call_4_11_4_1, \
call_4_11_4_2, \
call_4_11_4_3, \
call_4_11_4_4, \
call_4_11_4_5, \
call_4_11_4_6, \
call_4_11_4_7, \
call_4_11_4_8, \
call_4_11_4_9, \
call_4_11_4_10, \
call_4_11_4_11, \
call_4_11_4_12, \
call_4_11_4_13, \
call_4_11_4_14, \
call_4_11_4_15, \
call_4_11_5_0, \
call_4_11_5_1, \
call_4_11_5_2, \
call_4_11_5_3, \
call_4_11_5_4, \
call_4_11_5_5, \
call_4_11_5_6, \
call_4_11_5_7, \
call_4_11_5_8, \
call_4_11_5_9, \
call_4_11_5_10, \
call_4_11_5_11, \
call_4_11_5_12, \
call_4_11_5_13, \
call_4_11_5_14, \
call_4_11_5_15, \
call_4_11_6_0, \
call_4_11_6_1, \
call_4_11_6_2, \
call_4_11_6_3, \
call_4_11_6_4, \
call_4_11_6_5, \
call_4_11_6_6, \
call_4_11_6_7, \
call_4_11_6_8, \
call_4_11_6_9, \
call_4_11_6_10, \
call_4_11_6_11, \
call_4_11_6_12, \
call_4_11_6_13, \
call_4_11_6_14, \
call_4_11_6_15, \
call_4_11_7_0, \
call_4_11_7_1, \
call_4_11_7_2, \
call_4_11_7_3, \
call_4_11_7_4, \
call_4_11_7_5, \
call_4_11_7_6, \
call_4_11_7_7, \
call_4_11_7_8, \
call_4_11_7_9, \
call_4_11_7_10, \
call_4_11_7_11, \
call_4_11_7_12, \
call_4_11_7_13, \
call_4_11_7_14, \
call_4_11_7_15, \
call_4_11_8_0, \
call_4_11_8_1, \
call_4_11_8_2, \
call_4_11_8_3, \
call_4_11_8_4, \
call_4_11_8_5, \
call_4_11_8_6, \
call_4_11_8_7, \
call_4_11_8_8, \
call_4_11_8_9, \
call_4_11_8_10, \
call_4_11_8_11, \
call_4_11_8_12, \
call_4_11_8_13, \
call_4_11_8_14, \
call_4_11_8_15, \
call_4_11_9_0, \
call_4_11_9_1, \
call_4_11_9_2, \
call_4_11_9_3, \
call_4_11_9_4, \
call_4_11_9_5, \
call_4_11_9_6, \
call_4_11_9_7, \
call_4_11_9_8, \
call_4_11_9_9, \
call_4_11_9_10, \
call_4_11_9_11, \
call_4_11_9_12, \
call_4_11_9_13, \
call_4_11_9_14, \
call_4_11_9_15, \
call_4_11_10_0, \
call_4_11_10_1, \
call_4_11_10_2, \
call_4_11_10_3, \
call_4_11_10_4, \
call_4_11_10_5, \
call_4_11_10_6, \
call_4_11_10_7, \
call_4_11_10_8, \
call_4_11_10_9, \
call_4_11_10_10, \
call_4_11_10_11, \
call_4_11_10_12, \
call_4_11_10_13, \
call_4_11_10_14, \
call_4_11_10_15, \
call_4_11_11_0, \
call_4_11_11_1, \
call_4_11_11_2, \
call_4_11_11_3, \
call_4_11_11_4, \
call_4_11_11_5, \
call_4_11_11_6, \
call_4_11_11_7, \
call_4_11_11_8, \
call_4_11_11_9, \
call_4_11_11_10, \
call_4_11_11_11, \
call_4_11_11_12, \
call_4_11_11_13, \
call_4_11_11_14, \
call_4_11_11_15, \
call_4_11_12_0, \
call_4_11_12_1, \
call_4_11_12_2, \
call_4_11_12_3, \
call_4_11_12_4, \
call_4_11_12_5, \
call_4_11_12_6, \
call_4_11_12_7, \
call_4_11_12_8, \
call_4_11_12_9, \
call_4_11_12_10, \
call_4_11_12_11, \
call_4_11_12_12, \
call_4_11_12_13, \
call_4_11_12_14, \
call_4_11_12_15, \
call_4_11_13_0, \
call_4_11_13_1, \
call_4_11_13_2, \
call_4_11_13_3, \
call_4_11_13_4, \
call_4_11_13_5, \
call_4_11_13_6, \
call_4_11_13_7, \
call_4_11_13_8, \
call_4_11_13_9, \
call_4_11_13_10, \
call_4_11_13_11, \
call_4_11_13_12, \
call_4_11_13_13, \
call_4_11_13_14, \
call_4_11_13_15, \
call_4_11_14_0, \
call_4_11_14_1, \
call_4_11_14_2, \
call_4_11_14_3, \
call_4_11_14_4, \
call_4_11_14_5, \
call_4_11_14_6, \
call_4_11_14_7, \
call_4_11_14_8, \
call_4_11_14_9, \
call_4_11_14_10, \
call_4_11_14_11, \
call_4_11_14_12, \
call_4_11_14_13, \
call_4_11_14_14, \
call_4_11_14_15, \
call_4_11_15_0, \
call_4_11_15_1, \
call_4_11_15_2, \
call_4_11_15_3, \
call_4_11_15_4, \
call_4_11_15_5, \
call_4_11_15_6, \
call_4_11_15_7, \
call_4_11_15_8, \
call_4_11_15_9, \
call_4_11_15_10, \
call_4_11_15_11, \
call_4_11_15_12, \
call_4_11_15_13, \
call_4_11_15_14, \
call_4_11_15_15, \
call_4_12_0_0, \
call_4_12_0_1, \
call_4_12_0_2, \
call_4_12_0_3, \
call_4_12_0_4, \
call_4_12_0_5, \
call_4_12_0_6, \
call_4_12_0_7, \
call_4_12_0_8, \
call_4_12_0_9, \
call_4_12_0_10, \
call_4_12_0_11, \
call_4_12_0_12, \
call_4_12_0_13, \
call_4_12_0_14, \
call_4_12_0_15, \
call_4_12_1_0, \
call_4_12_1_1, \
call_4_12_1_2, \
call_4_12_1_3, \
call_4_12_1_4, \
call_4_12_1_5, \
call_4_12_1_6, \
call_4_12_1_7, \
call_4_12_1_8, \
call_4_12_1_9, \
call_4_12_1_10, \
call_4_12_1_11, \
call_4_12_1_12, \
call_4_12_1_13, \
call_4_12_1_14, \
call_4_12_1_15, \
call_4_12_2_0, \
call_4_12_2_1, \
call_4_12_2_2, \
call_4_12_2_3, \
call_4_12_2_4, \
call_4_12_2_5, \
call_4_12_2_6, \
call_4_12_2_7, \
call_4_12_2_8, \
call_4_12_2_9, \
call_4_12_2_10, \
call_4_12_2_11, \
call_4_12_2_12, \
call_4_12_2_13, \
call_4_12_2_14, \
call_4_12_2_15, \
call_4_12_3_0, \
call_4_12_3_1, \
call_4_12_3_2, \
call_4_12_3_3, \
call_4_12_3_4, \
call_4_12_3_5, \
call_4_12_3_6, \
call_4_12_3_7, \
call_4_12_3_8, \
call_4_12_3_9, \
call_4_12_3_10, \
call_4_12_3_11, \
call_4_12_3_12, \
call_4_12_3_13, \
call_4_12_3_14, \
call_4_12_3_15, \
call_4_12_4_0, \
call_4_12_4_1, \
call_4_12_4_2, \
call_4_12_4_3, \
call_4_12_4_4, \
call_4_12_4_5, \
call_4_12_4_6, \
call_4_12_4_7, \
call_4_12_4_8, \
call_4_12_4_9, \
call_4_12_4_10, \
call_4_12_4_11, \
call_4_12_4_12, \
call_4_12_4_13, \
call_4_12_4_14, \
call_4_12_4_15, \
call_4_12_5_0, \
call_4_12_5_1, \
call_4_12_5_2, \
call_4_12_5_3, \
call_4_12_5_4, \
call_4_12_5_5, \
call_4_12_5_6, \
call_4_12_5_7, \
call_4_12_5_8, \
call_4_12_5_9, \
call_4_12_5_10, \
call_4_12_5_11, \
call_4_12_5_12, \
call_4_12_5_13, \
call_4_12_5_14, \
call_4_12_5_15, \
call_4_12_6_0, \
call_4_12_6_1, \
call_4_12_6_2, \
call_4_12_6_3, \
call_4_12_6_4, \
call_4_12_6_5, \
call_4_12_6_6, \
call_4_12_6_7, \
call_4_12_6_8, \
call_4_12_6_9, \
call_4_12_6_10, \
call_4_12_6_11, \
call_4_12_6_12, \
call_4_12_6_13, \
call_4_12_6_14, \
call_4_12_6_15, \
call_4_12_7_0, \
call_4_12_7_1, \
call_4_12_7_2, \
call_4_12_7_3, \
call_4_12_7_4, \
call_4_12_7_5, \
call_4_12_7_6, \
call_4_12_7_7, \
call_4_12_7_8, \
call_4_12_7_9, \
call_4_12_7_10, \
call_4_12_7_11, \
call_4_12_7_12, \
call_4_12_7_13, \
call_4_12_7_14, \
call_4_12_7_15, \
call_4_12_8_0, \
call_4_12_8_1, \
call_4_12_8_2, \
call_4_12_8_3, \
call_4_12_8_4, \
call_4_12_8_5, \
call_4_12_8_6, \
call_4_12_8_7, \
call_4_12_8_8, \
call_4_12_8_9, \
call_4_12_8_10, \
call_4_12_8_11, \
call_4_12_8_12, \
call_4_12_8_13, \
call_4_12_8_14, \
call_4_12_8_15, \
call_4_12_9_0, \
call_4_12_9_1, \
call_4_12_9_2, \
call_4_12_9_3, \
call_4_12_9_4, \
call_4_12_9_5, \
call_4_12_9_6, \
call_4_12_9_7, \
call_4_12_9_8, \
call_4_12_9_9, \
call_4_12_9_10, \
call_4_12_9_11, \
call_4_12_9_12, \
call_4_12_9_13, \
call_4_12_9_14, \
call_4_12_9_15, \
call_4_12_10_0, \
call_4_12_10_1, \
call_4_12_10_2, \
call_4_12_10_3, \
call_4_12_10_4, \
call_4_12_10_5, \
call_4_12_10_6, \
call_4_12_10_7, \
call_4_12_10_8, \
call_4_12_10_9, \
call_4_12_10_10, \
call_4_12_10_11, \
call_4_12_10_12, \
call_4_12_10_13, \
call_4_12_10_14, \
call_4_12_10_15, \
call_4_12_11_0, \
call_4_12_11_1, \
call_4_12_11_2, \
call_4_12_11_3, \
call_4_12_11_4, \
call_4_12_11_5, \
call_4_12_11_6, \
call_4_12_11_7, \
call_4_12_11_8, \
call_4_12_11_9, \
call_4_12_11_10, \
call_4_12_11_11, \
call_4_12_11_12, \
call_4_12_11_13, \
call_4_12_11_14, \
call_4_12_11_15, \
call_4_12_12_0, \
call_4_12_12_1, \
call_4_12_12_2, \
call_4_12_12_3, \
call_4_12_12_4, \
call_4_12_12_5, \
call_4_12_12_6, \
call_4_12_12_7, \
call_4_12_12_8, \
call_4_12_12_9, \
call_4_12_12_10, \
call_4_12_12_11, \
call_4_12_12_12, \
call_4_12_12_13, \
call_4_12_12_14, \
call_4_12_12_15, \
call_4_12_13_0, \
call_4_12_13_1, \
call_4_12_13_2, \
call_4_12_13_3, \
call_4_12_13_4, \
call_4_12_13_5, \
call_4_12_13_6, \
call_4_12_13_7, \
call_4_12_13_8, \
call_4_12_13_9, \
call_4_12_13_10, \
call_4_12_13_11, \
call_4_12_13_12, \
call_4_12_13_13, \
call_4_12_13_14, \
call_4_12_13_15, \
call_4_12_14_0, \
call_4_12_14_1, \
call_4_12_14_2, \
call_4_12_14_3, \
call_4_12_14_4, \
call_4_12_14_5, \
call_4_12_14_6, \
call_4_12_14_7, \
call_4_12_14_8, \
call_4_12_14_9, \
call_4_12_14_10, \
call_4_12_14_11, \
call_4_12_14_12, \
call_4_12_14_13, \
call_4_12_14_14, \
call_4_12_14_15, \
call_4_12_15_0, \
call_4_12_15_1, \
call_4_12_15_2, \
call_4_12_15_3, \
call_4_12_15_4, \
call_4_12_15_5, \
call_4_12_15_6, \
call_4_12_15_7, \
call_4_12_15_8, \
call_4_12_15_9, \
call_4_12_15_10, \
call_4_12_15_11, \
call_4_12_15_12, \
call_4_12_15_13, \
call_4_12_15_14, \
call_4_12_15_15, \
call_4_13_0_0, \
call_4_13_0_1, \
call_4_13_0_2, \
call_4_13_0_3, \
call_4_13_0_4, \
call_4_13_0_5, \
call_4_13_0_6, \
call_4_13_0_7, \
call_4_13_0_8, \
call_4_13_0_9, \
call_4_13_0_10, \
call_4_13_0_11, \
call_4_13_0_12, \
call_4_13_0_13, \
call_4_13_0_14, \
call_4_13_0_15, \
call_4_13_1_0, \
call_4_13_1_1, \
call_4_13_1_2, \
call_4_13_1_3, \
call_4_13_1_4, \
call_4_13_1_5, \
call_4_13_1_6, \
call_4_13_1_7, \
call_4_13_1_8, \
call_4_13_1_9, \
call_4_13_1_10, \
call_4_13_1_11, \
call_4_13_1_12, \
call_4_13_1_13, \
call_4_13_1_14, \
call_4_13_1_15, \
call_4_13_2_0, \
call_4_13_2_1, \
call_4_13_2_2, \
call_4_13_2_3, \
call_4_13_2_4, \
call_4_13_2_5, \
call_4_13_2_6, \
call_4_13_2_7, \
call_4_13_2_8, \
call_4_13_2_9, \
call_4_13_2_10, \
call_4_13_2_11, \
call_4_13_2_12, \
call_4_13_2_13, \
call_4_13_2_14, \
call_4_13_2_15, \
call_4_13_3_0, \
call_4_13_3_1, \
call_4_13_3_2, \
call_4_13_3_3, \
call_4_13_3_4, \
call_4_13_3_5, \
call_4_13_3_6, \
call_4_13_3_7, \
call_4_13_3_8, \
call_4_13_3_9, \
call_4_13_3_10, \
call_4_13_3_11, \
call_4_13_3_12, \
call_4_13_3_13, \
call_4_13_3_14, \
call_4_13_3_15, \
call_4_13_4_0, \
call_4_13_4_1, \
call_4_13_4_2, \
call_4_13_4_3, \
call_4_13_4_4, \
call_4_13_4_5, \
call_4_13_4_6, \
call_4_13_4_7, \
call_4_13_4_8, \
call_4_13_4_9, \
call_4_13_4_10, \
call_4_13_4_11, \
call_4_13_4_12, \
call_4_13_4_13, \
call_4_13_4_14, \
call_4_13_4_15, \
call_4_13_5_0, \
call_4_13_5_1, \
call_4_13_5_2, \
call_4_13_5_3, \
call_4_13_5_4, \
call_4_13_5_5, \
call_4_13_5_6, \
call_4_13_5_7, \
call_4_13_5_8, \
call_4_13_5_9, \
call_4_13_5_10, \
call_4_13_5_11, \
call_4_13_5_12, \
call_4_13_5_13, \
call_4_13_5_14, \
call_4_13_5_15, \
call_4_13_6_0, \
call_4_13_6_1, \
call_4_13_6_2, \
call_4_13_6_3, \
call_4_13_6_4, \
call_4_13_6_5, \
call_4_13_6_6, \
call_4_13_6_7, \
call_4_13_6_8, \
call_4_13_6_9, \
call_4_13_6_10, \
call_4_13_6_11, \
call_4_13_6_12, \
call_4_13_6_13, \
call_4_13_6_14, \
call_4_13_6_15, \
call_4_13_7_0, \
call_4_13_7_1, \
call_4_13_7_2, \
call_4_13_7_3, \
call_4_13_7_4, \
call_4_13_7_5, \
call_4_13_7_6, \
call_4_13_7_7, \
call_4_13_7_8, \
call_4_13_7_9, \
call_4_13_7_10, \
call_4_13_7_11, \
call_4_13_7_12, \
call_4_13_7_13, \
call_4_13_7_14, \
call_4_13_7_15, \
call_4_13_8_0, \
call_4_13_8_1, \
call_4_13_8_2, \
call_4_13_8_3, \
call_4_13_8_4, \
call_4_13_8_5, \
call_4_13_8_6, \
call_4_13_8_7, \
call_4_13_8_8, \
call_4_13_8_9, \
call_4_13_8_10, \
call_4_13_8_11, \
call_4_13_8_12, \
call_4_13_8_13, \
call_4_13_8_14, \
call_4_13_8_15, \
call_4_13_9_0, \
call_4_13_9_1, \
call_4_13_9_2, \
call_4_13_9_3, \
call_4_13_9_4, \
call_4_13_9_5, \
call_4_13_9_6, \
call_4_13_9_7, \
call_4_13_9_8, \
call_4_13_9_9, \
call_4_13_9_10, \
call_4_13_9_11, \
call_4_13_9_12, \
call_4_13_9_13, \
call_4_13_9_14, \
call_4_13_9_15, \
call_4_13_10_0, \
call_4_13_10_1, \
call_4_13_10_2, \
call_4_13_10_3, \
call_4_13_10_4, \
call_4_13_10_5, \
call_4_13_10_6, \
call_4_13_10_7, \
call_4_13_10_8, \
call_4_13_10_9, \
call_4_13_10_10, \
call_4_13_10_11, \
call_4_13_10_12, \
call_4_13_10_13, \
call_4_13_10_14, \
call_4_13_10_15, \
call_4_13_11_0, \
call_4_13_11_1, \
call_4_13_11_2, \
call_4_13_11_3, \
call_4_13_11_4, \
call_4_13_11_5, \
call_4_13_11_6, \
call_4_13_11_7, \
call_4_13_11_8, \
call_4_13_11_9, \
call_4_13_11_10, \
call_4_13_11_11, \
call_4_13_11_12, \
call_4_13_11_13, \
call_4_13_11_14, \
call_4_13_11_15, \
call_4_13_12_0, \
call_4_13_12_1, \
call_4_13_12_2, \
call_4_13_12_3, \
call_4_13_12_4, \
call_4_13_12_5, \
call_4_13_12_6, \
call_4_13_12_7, \
call_4_13_12_8, \
call_4_13_12_9, \
call_4_13_12_10, \
call_4_13_12_11, \
call_4_13_12_12, \
call_4_13_12_13, \
call_4_13_12_14, \
call_4_13_12_15, \
call_4_13_13_0, \
call_4_13_13_1, \
call_4_13_13_2, \
call_4_13_13_3, \
call_4_13_13_4, \
call_4_13_13_5, \
call_4_13_13_6, \
call_4_13_13_7, \
call_4_13_13_8, \
call_4_13_13_9, \
call_4_13_13_10, \
call_4_13_13_11, \
call_4_13_13_12, \
call_4_13_13_13, \
call_4_13_13_14, \
call_4_13_13_15, \
call_4_13_14_0, \
call_4_13_14_1, \
call_4_13_14_2, \
call_4_13_14_3, \
call_4_13_14_4, \
call_4_13_14_5, \
call_4_13_14_6, \
call_4_13_14_7, \
call_4_13_14_8, \
call_4_13_14_9, \
call_4_13_14_10, \
call_4_13_14_11, \
call_4_13_14_12, \
call_4_13_14_13, \
call_4_13_14_14, \
call_4_13_14_15, \
call_4_13_15_0, \
call_4_13_15_1, \
call_4_13_15_2, \
call_4_13_15_3, \
call_4_13_15_4, \
call_4_13_15_5, \
call_4_13_15_6, \
call_4_13_15_7, \
call_4_13_15_8, \
call_4_13_15_9, \
call_4_13_15_10, \
call_4_13_15_11, \
call_4_13_15_12, \
call_4_13_15_13, \
call_4_13_15_14, \
call_4_13_15_15, \
call_4_14_0_0, \
call_4_14_0_1, \
call_4_14_0_2, \
call_4_14_0_3, \
call_4_14_0_4, \
call_4_14_0_5, \
call_4_14_0_6, \
call_4_14_0_7, \
call_4_14_0_8, \
call_4_14_0_9, \
call_4_14_0_10, \
call_4_14_0_11, \
call_4_14_0_12, \
call_4_14_0_13, \
call_4_14_0_14, \
call_4_14_0_15, \
call_4_14_1_0, \
call_4_14_1_1, \
call_4_14_1_2, \
call_4_14_1_3, \
call_4_14_1_4, \
call_4_14_1_5, \
call_4_14_1_6, \
call_4_14_1_7, \
call_4_14_1_8, \
call_4_14_1_9, \
call_4_14_1_10, \
call_4_14_1_11, \
call_4_14_1_12, \
call_4_14_1_13, \
call_4_14_1_14, \
call_4_14_1_15, \
call_4_14_2_0, \
call_4_14_2_1, \
call_4_14_2_2, \
call_4_14_2_3, \
call_4_14_2_4, \
call_4_14_2_5, \
call_4_14_2_6, \
call_4_14_2_7, \
call_4_14_2_8, \
call_4_14_2_9, \
call_4_14_2_10, \
call_4_14_2_11, \
call_4_14_2_12, \
call_4_14_2_13, \
call_4_14_2_14, \
call_4_14_2_15, \
call_4_14_3_0, \
call_4_14_3_1, \
call_4_14_3_2, \
call_4_14_3_3, \
call_4_14_3_4, \
call_4_14_3_5, \
call_4_14_3_6, \
call_4_14_3_7, \
call_4_14_3_8, \
call_4_14_3_9, \
call_4_14_3_10, \
call_4_14_3_11, \
call_4_14_3_12, \
call_4_14_3_13, \
call_4_14_3_14, \
call_4_14_3_15, \
call_4_14_4_0, \
call_4_14_4_1, \
call_4_14_4_2, \
call_4_14_4_3, \
call_4_14_4_4, \
call_4_14_4_5, \
call_4_14_4_6, \
call_4_14_4_7, \
call_4_14_4_8, \
call_4_14_4_9, \
call_4_14_4_10, \
call_4_14_4_11, \
call_4_14_4_12, \
call_4_14_4_13, \
call_4_14_4_14, \
call_4_14_4_15, \
call_4_14_5_0, \
call_4_14_5_1, \
call_4_14_5_2, \
call_4_14_5_3, \
call_4_14_5_4, \
call_4_14_5_5, \
call_4_14_5_6, \
call_4_14_5_7, \
call_4_14_5_8, \
call_4_14_5_9, \
call_4_14_5_10, \
call_4_14_5_11, \
call_4_14_5_12, \
call_4_14_5_13, \
call_4_14_5_14, \
call_4_14_5_15, \
call_4_14_6_0, \
call_4_14_6_1, \
call_4_14_6_2, \
call_4_14_6_3, \
call_4_14_6_4, \
call_4_14_6_5, \
call_4_14_6_6, \
call_4_14_6_7, \
call_4_14_6_8, \
call_4_14_6_9, \
call_4_14_6_10, \
call_4_14_6_11, \
call_4_14_6_12, \
call_4_14_6_13, \
call_4_14_6_14, \
call_4_14_6_15, \
call_4_14_7_0, \
call_4_14_7_1, \
call_4_14_7_2, \
call_4_14_7_3, \
call_4_14_7_4, \
call_4_14_7_5, \
call_4_14_7_6, \
call_4_14_7_7, \
call_4_14_7_8, \
call_4_14_7_9, \
call_4_14_7_10, \
call_4_14_7_11, \
call_4_14_7_12, \
call_4_14_7_13, \
call_4_14_7_14, \
call_4_14_7_15, \
call_4_14_8_0, \
call_4_14_8_1, \
call_4_14_8_2, \
call_4_14_8_3, \
call_4_14_8_4, \
call_4_14_8_5, \
call_4_14_8_6, \
call_4_14_8_7, \
call_4_14_8_8, \
call_4_14_8_9, \
call_4_14_8_10, \
call_4_14_8_11, \
call_4_14_8_12, \
call_4_14_8_13, \
call_4_14_8_14, \
call_4_14_8_15, \
call_4_14_9_0, \
call_4_14_9_1, \
call_4_14_9_2, \
call_4_14_9_3, \
call_4_14_9_4, \
call_4_14_9_5, \
call_4_14_9_6, \
call_4_14_9_7, \
call_4_14_9_8, \
call_4_14_9_9, \
call_4_14_9_10, \
call_4_14_9_11, \
call_4_14_9_12, \
call_4_14_9_13, \
call_4_14_9_14, \
call_4_14_9_15, \
call_4_14_10_0, \
call_4_14_10_1, \
call_4_14_10_2, \
call_4_14_10_3, \
call_4_14_10_4, \
call_4_14_10_5, \
call_4_14_10_6, \
call_4_14_10_7, \
call_4_14_10_8, \
call_4_14_10_9, \
call_4_14_10_10, \
call_4_14_10_11, \
call_4_14_10_12, \
call_4_14_10_13, \
call_4_14_10_14, \
call_4_14_10_15, \
call_4_14_11_0, \
call_4_14_11_1, \
call_4_14_11_2, \
call_4_14_11_3, \
call_4_14_11_4, \
call_4_14_11_5, \
call_4_14_11_6, \
call_4_14_11_7, \
call_4_14_11_8, \
call_4_14_11_9, \
call_4_14_11_10, \
call_4_14_11_11, \
call_4_14_11_12, \
call_4_14_11_13, \
call_4_14_11_14, \
call_4_14_11_15, \
call_4_14_12_0, \
call_4_14_12_1, \
call_4_14_12_2, \
call_4_14_12_3, \
call_4_14_12_4, \
call_4_14_12_5, \
call_4_14_12_6, \
call_4_14_12_7, \
call_4_14_12_8, \
call_4_14_12_9, \
call_4_14_12_10, \
call_4_14_12_11, \
call_4_14_12_12, \
call_4_14_12_13, \
call_4_14_12_14, \
call_4_14_12_15, \
call_4_14_13_0, \
call_4_14_13_1, \
call_4_14_13_2, \
call_4_14_13_3, \
call_4_14_13_4, \
call_4_14_13_5, \
call_4_14_13_6, \
call_4_14_13_7, \
call_4_14_13_8, \
call_4_14_13_9, \
call_4_14_13_10, \
call_4_14_13_11, \
call_4_14_13_12, \
call_4_14_13_13, \
call_4_14_13_14, \
call_4_14_13_15, \
call_4_14_14_0, \
call_4_14_14_1, \
call_4_14_14_2, \
call_4_14_14_3, \
call_4_14_14_4, \
call_4_14_14_5, \
call_4_14_14_6, \
call_4_14_14_7, \
call_4_14_14_8, \
call_4_14_14_9, \
call_4_14_14_10, \
call_4_14_14_11, \
call_4_14_14_12, \
call_4_14_14_13, \
call_4_14_14_14, \
call_4_14_14_15, \
call_4_14_15_0, \
call_4_14_15_1, \
call_4_14_15_2, \
call_4_14_15_3, \
call_4_14_15_4, \
call_4_14_15_5, \
call_4_14_15_6, \
call_4_14_15_7, \
call_4_14_15_8, \
call_4_14_15_9, \
call_4_14_15_10, \
call_4_14_15_11, \
call_4_14_15_12, \
call_4_14_15_13, \
call_4_14_15_14, \
call_4_14_15_15, \
call_4_15_0_0, \
call_4_15_0_1, \
call_4_15_0_2, \
call_4_15_0_3, \
call_4_15_0_4, \
call_4_15_0_5, \
call_4_15_0_6, \
call_4_15_0_7, \
call_4_15_0_8, \
call_4_15_0_9, \
call_4_15_0_10, \
call_4_15_0_11, \
call_4_15_0_12, \
call_4_15_0_13, \
call_4_15_0_14, \
call_4_15_0_15, \
call_4_15_1_0, \
call_4_15_1_1, \
call_4_15_1_2, \
call_4_15_1_3, \
call_4_15_1_4, \
call_4_15_1_5, \
call_4_15_1_6, \
call_4_15_1_7, \
call_4_15_1_8, \
call_4_15_1_9, \
call_4_15_1_10, \
call_4_15_1_11, \
call_4_15_1_12, \
call_4_15_1_13, \
call_4_15_1_14, \
call_4_15_1_15, \
call_4_15_2_0, \
call_4_15_2_1, \
call_4_15_2_2, \
call_4_15_2_3, \
call_4_15_2_4, \
call_4_15_2_5, \
call_4_15_2_6, \
call_4_15_2_7, \
call_4_15_2_8, \
call_4_15_2_9, \
call_4_15_2_10, \
call_4_15_2_11, \
call_4_15_2_12, \
call_4_15_2_13, \
call_4_15_2_14, \
call_4_15_2_15, \
call_4_15_3_0, \
call_4_15_3_1, \
call_4_15_3_2, \
call_4_15_3_3, \
call_4_15_3_4, \
call_4_15_3_5, \
call_4_15_3_6, \
call_4_15_3_7, \
call_4_15_3_8, \
call_4_15_3_9, \
call_4_15_3_10, \
call_4_15_3_11, \
call_4_15_3_12, \
call_4_15_3_13, \
call_4_15_3_14, \
call_4_15_3_15, \
call_4_15_4_0, \
call_4_15_4_1, \
call_4_15_4_2, \
call_4_15_4_3, \
call_4_15_4_4, \
call_4_15_4_5, \
call_4_15_4_6, \
call_4_15_4_7, \
call_4_15_4_8, \
call_4_15_4_9, \
call_4_15_4_10, \
call_4_15_4_11, \
call_4_15_4_12, \
call_4_15_4_13, \
call_4_15_4_14, \
call_4_15_4_15, \
call_4_15_5_0, \
call_4_15_5_1, \
call_4_15_5_2, \
call_4_15_5_3, \
call_4_15_5_4, \
call_4_15_5_5, \
call_4_15_5_6, \
call_4_15_5_7, \
call_4_15_5_8, \
call_4_15_5_9, \
call_4_15_5_10, \
call_4_15_5_11, \
call_4_15_5_12, \
call_4_15_5_13, \
call_4_15_5_14, \
call_4_15_5_15, \
call_4_15_6_0, \
call_4_15_6_1, \
call_4_15_6_2, \
call_4_15_6_3, \
call_4_15_6_4, \
call_4_15_6_5, \
call_4_15_6_6, \
call_4_15_6_7, \
call_4_15_6_8, \
call_4_15_6_9, \
call_4_15_6_10, \
call_4_15_6_11, \
call_4_15_6_12, \
call_4_15_6_13, \
call_4_15_6_14, \
call_4_15_6_15, \
call_4_15_7_0, \
call_4_15_7_1, \
call_4_15_7_2, \
call_4_15_7_3, \
call_4_15_7_4, \
call_4_15_7_5, \
call_4_15_7_6, \
call_4_15_7_7, \
call_4_15_7_8, \
call_4_15_7_9, \
call_4_15_7_10, \
call_4_15_7_11, \
call_4_15_7_12, \
call_4_15_7_13, \
call_4_15_7_14, \
call_4_15_7_15, \
call_4_15_8_0, \
call_4_15_8_1, \
call_4_15_8_2, \
call_4_15_8_3, \
call_4_15_8_4, \
call_4_15_8_5, \
call_4_15_8_6, \
call_4_15_8_7, \
call_4_15_8_8, \
call_4_15_8_9, \
call_4_15_8_10, \
call_4_15_8_11, \
call_4_15_8_12, \
call_4_15_8_13, \
call_4_15_8_14, \
call_4_15_8_15, \
call_4_15_9_0, \
call_4_15_9_1, \
call_4_15_9_2, \
call_4_15_9_3, \
call_4_15_9_4, \
call_4_15_9_5, \
call_4_15_9_6, \
call_4_15_9_7, \
call_4_15_9_8, \
call_4_15_9_9, \
call_4_15_9_10, \
call_4_15_9_11, \
call_4_15_9_12, \
call_4_15_9_13, \
call_4_15_9_14, \
call_4_15_9_15, \
call_4_15_10_0, \
call_4_15_10_1, \
call_4_15_10_2, \
call_4_15_10_3, \
call_4_15_10_4, \
call_4_15_10_5, \
call_4_15_10_6, \
call_4_15_10_7, \
call_4_15_10_8, \
call_4_15_10_9, \
call_4_15_10_10, \
call_4_15_10_11, \
call_4_15_10_12, \
call_4_15_10_13, \
call_4_15_10_14, \
call_4_15_10_15, \
call_4_15_11_0, \
call_4_15_11_1, \
call_4_15_11_2, \
call_4_15_11_3, \
call_4_15_11_4, \
call_4_15_11_5, \
call_4_15_11_6, \
call_4_15_11_7, \
call_4_15_11_8, \
call_4_15_11_9, \
call_4_15_11_10, \
call_4_15_11_11, \
call_4_15_11_12, \
call_4_15_11_13, \
call_4_15_11_14, \
call_4_15_11_15, \
call_4_15_12_0, \
call_4_15_12_1, \
call_4_15_12_2, \
call_4_15_12_3, \
call_4_15_12_4, \
call_4_15_12_5, \
call_4_15_12_6, \
call_4_15_12_7, \
call_4_15_12_8, \
call_4_15_12_9, \
call_4_15_12_10, \
call_4_15_12_11, \
call_4_15_12_12, \
call_4_15_12_13, \
call_4_15_12_14, \
call_4_15_12_15, \
call_4_15_13_0, \
call_4_15_13_1, \
call_4_15_13_2, \
call_4_15_13_3, \
call_4_15_13_4, \
call_4_15_13_5, \
call_4_15_13_6, \
call_4_15_13_7, \
call_4_15_13_8, \
call_4_15_13_9, \
call_4_15_13_10, \
call_4_15_13_11, \
call_4_15_13_12, \
call_4_15_13_13, \
call_4_15_13_14, \
call_4_15_13_15, \
call_4_15_14_0, \
call_4_15_14_1, \
call_4_15_14_2, \
call_4_15_14_3, \
call_4_15_14_4, \
call_4_15_14_5, \
call_4_15_14_6, \
call_4_15_14_7, \
call_4_15_14_8, \
call_4_15_14_9, \
call_4_15_14_10, \
call_4_15_14_11, \
call_4_15_14_12, \
call_4_15_14_13, \
call_4_15_14_14, \
call_4_15_14_15, \
call_4_15_15_0, \
call_4_15_15_1, \
call_4_15_15_2, \
call_4_15_15_3, \
call_4_15_15_4, \
call_4_15_15_5, \
call_4_15_15_6, \
call_4_15_15_7, \
call_4_15_15_8, \
call_4_15_15_9, \
call_4_15_15_10, \
call_4_15_15_11, \
call_4_15_15_12, \
call_4_15_15_13, \
call_4_15_15_14, \
call_4_15_15_15, \
call_5_0_0_0, \
call_5_0_0_1, \
call_5_0_0_2, \
call_5_0_0_3, \
call_5_0_0_4, \
call_5_0_0_5, \
call_5_0_0_6, \
call_5_0_0_7, \
call_5_0_0_8, \
call_5_0_0_9, \
call_5_0_0_10, \
call_5_0_0_11, \
call_5_0_0_12, \
call_5_0_0_13, \
call_5_0_0_14, \
call_5_0_0_15, \
call_5_0_1_0, \
call_5_0_1_1, \
call_5_0_1_2, \
call_5_0_1_3, \
call_5_0_1_4, \
call_5_0_1_5, \
call_5_0_1_6, \
call_5_0_1_7, \
call_5_0_1_8, \
call_5_0_1_9, \
call_5_0_1_10, \
call_5_0_1_11, \
call_5_0_1_12, \
call_5_0_1_13, \
call_5_0_1_14, \
call_5_0_1_15, \
call_5_0_2_0, \
call_5_0_2_1, \
call_5_0_2_2, \
call_5_0_2_3, \
call_5_0_2_4, \
call_5_0_2_5, \
call_5_0_2_6, \
call_5_0_2_7, \
call_5_0_2_8, \
call_5_0_2_9, \
call_5_0_2_10, \
call_5_0_2_11, \
call_5_0_2_12, \
call_5_0_2_13, \
call_5_0_2_14, \
call_5_0_2_15, \
call_5_0_3_0, \
call_5_0_3_1, \
call_5_0_3_2, \
call_5_0_3_3, \
call_5_0_3_4, \
call_5_0_3_5, \
call_5_0_3_6, \
call_5_0_3_7, \
call_5_0_3_8, \
call_5_0_3_9, \
call_5_0_3_10, \
call_5_0_3_11, \
call_5_0_3_12, \
call_5_0_3_13, \
call_5_0_3_14, \
call_5_0_3_15, \
call_5_0_4_0, \
call_5_0_4_1, \
call_5_0_4_2, \
call_5_0_4_3, \
call_5_0_4_4, \
call_5_0_4_5, \
call_5_0_4_6, \
call_5_0_4_7, \
call_5_0_4_8, \
call_5_0_4_9, \
call_5_0_4_10, \
call_5_0_4_11, \
call_5_0_4_12, \
call_5_0_4_13, \
call_5_0_4_14, \
call_5_0_4_15, \
call_5_0_5_0, \
call_5_0_5_1, \
call_5_0_5_2, \
call_5_0_5_3, \
call_5_0_5_4, \
call_5_0_5_5, \
call_5_0_5_6, \
call_5_0_5_7, \
call_5_0_5_8, \
call_5_0_5_9, \
call_5_0_5_10, \
call_5_0_5_11, \
call_5_0_5_12, \
call_5_0_5_13, \
call_5_0_5_14, \
call_5_0_5_15, \
call_5_0_6_0, \
call_5_0_6_1, \
call_5_0_6_2, \
call_5_0_6_3, \
call_5_0_6_4, \
call_5_0_6_5, \
call_5_0_6_6, \
call_5_0_6_7, \
call_5_0_6_8, \
call_5_0_6_9, \
call_5_0_6_10, \
call_5_0_6_11, \
call_5_0_6_12, \
call_5_0_6_13, \
call_5_0_6_14, \
call_5_0_6_15, \
call_5_0_7_0, \
call_5_0_7_1, \
call_5_0_7_2, \
call_5_0_7_3, \
call_5_0_7_4, \
call_5_0_7_5, \
call_5_0_7_6, \
call_5_0_7_7, \
call_5_0_7_8, \
call_5_0_7_9, \
call_5_0_7_10, \
call_5_0_7_11, \
call_5_0_7_12, \
call_5_0_7_13, \
call_5_0_7_14, \
call_5_0_7_15, \
call_5_0_8_0, \
call_5_0_8_1, \
call_5_0_8_2, \
call_5_0_8_3, \
call_5_0_8_4, \
call_5_0_8_5, \
call_5_0_8_6, \
call_5_0_8_7, \
call_5_0_8_8, \
call_5_0_8_9, \
call_5_0_8_10, \
call_5_0_8_11, \
call_5_0_8_12, \
call_5_0_8_13, \
call_5_0_8_14, \
call_5_0_8_15, \
call_5_0_9_0, \
call_5_0_9_1, \
call_5_0_9_2, \
call_5_0_9_3, \
call_5_0_9_4, \
call_5_0_9_5, \
call_5_0_9_6, \
call_5_0_9_7, \
call_5_0_9_8, \
call_5_0_9_9, \
call_5_0_9_10, \
call_5_0_9_11, \
call_5_0_9_12, \
call_5_0_9_13, \
call_5_0_9_14, \
call_5_0_9_15, \
call_5_0_10_0, \
call_5_0_10_1, \
call_5_0_10_2, \
call_5_0_10_3, \
call_5_0_10_4, \
call_5_0_10_5, \
call_5_0_10_6, \
call_5_0_10_7, \
call_5_0_10_8, \
call_5_0_10_9, \
call_5_0_10_10, \
call_5_0_10_11, \
call_5_0_10_12, \
call_5_0_10_13, \
call_5_0_10_14, \
call_5_0_10_15, \
call_5_0_11_0, \
call_5_0_11_1, \
call_5_0_11_2, \
call_5_0_11_3, \
call_5_0_11_4, \
call_5_0_11_5, \
call_5_0_11_6, \
call_5_0_11_7, \
call_5_0_11_8, \
call_5_0_11_9, \
call_5_0_11_10, \
call_5_0_11_11, \
call_5_0_11_12, \
call_5_0_11_13, \
call_5_0_11_14, \
call_5_0_11_15, \
call_5_0_12_0, \
call_5_0_12_1, \
call_5_0_12_2, \
call_5_0_12_3, \
call_5_0_12_4, \
call_5_0_12_5, \
call_5_0_12_6, \
call_5_0_12_7, \
call_5_0_12_8, \
call_5_0_12_9, \
call_5_0_12_10, \
call_5_0_12_11, \
call_5_0_12_12, \
call_5_0_12_13, \
call_5_0_12_14, \
call_5_0_12_15, \
call_5_0_13_0, \
call_5_0_13_1, \
call_5_0_13_2, \
call_5_0_13_3, \
call_5_0_13_4, \
call_5_0_13_5, \
call_5_0_13_6, \
call_5_0_13_7, \
call_5_0_13_8, \
call_5_0_13_9, \
call_5_0_13_10, \
call_5_0_13_11, \
call_5_0_13_12, \
call_5_0_13_13, \
call_5_0_13_14, \
call_5_0_13_15, \
call_5_0_14_0, \
call_5_0_14_1, \
call_5_0_14_2, \
call_5_0_14_3, \
call_5_0_14_4, \
call_5_0_14_5, \
call_5_0_14_6, \
call_5_0_14_7, \
call_5_0_14_8, \
call_5_0_14_9, \
call_5_0_14_10, \
call_5_0_14_11, \
call_5_0_14_12, \
call_5_0_14_13, \
call_5_0_14_14, \
call_5_0_14_15, \
call_5_0_15_0, \
call_5_0_15_1, \
call_5_0_15_2, \
call_5_0_15_3, \
call_5_0_15_4, \
call_5_0_15_5, \
call_5_0_15_6, \
call_5_0_15_7, \
call_5_0_15_8, \
call_5_0_15_9, \
call_5_0_15_10, \
call_5_0_15_11, \
call_5_0_15_12, \
call_5_0_15_13, \
call_5_0_15_14, \
call_5_0_15_15, \
call_5_1_0_0, \
call_5_1_0_1, \
call_5_1_0_2, \
call_5_1_0_3, \
call_5_1_0_4, \
call_5_1_0_5, \
call_5_1_0_6, \
call_5_1_0_7, \
call_5_1_0_8, \
call_5_1_0_9, \
call_5_1_0_10, \
call_5_1_0_11, \
call_5_1_0_12, \
call_5_1_0_13, \
call_5_1_0_14, \
call_5_1_0_15, \
call_5_1_1_0, \
call_5_1_1_1, \
call_5_1_1_2, \
call_5_1_1_3, \
call_5_1_1_4, \
call_5_1_1_5, \
call_5_1_1_6, \
call_5_1_1_7, \
call_5_1_1_8, \
call_5_1_1_9, \
call_5_1_1_10, \
call_5_1_1_11, \
call_5_1_1_12, \
call_5_1_1_13, \
call_5_1_1_14, \
call_5_1_1_15, \
call_5_1_2_0, \
call_5_1_2_1, \
call_5_1_2_2, \
call_5_1_2_3, \
call_5_1_2_4, \
call_5_1_2_5, \
call_5_1_2_6, \
call_5_1_2_7, \
call_5_1_2_8, \
call_5_1_2_9, \
call_5_1_2_10, \
call_5_1_2_11, \
call_5_1_2_12, \
call_5_1_2_13, \
call_5_1_2_14, \
call_5_1_2_15, \
call_5_1_3_0, \
call_5_1_3_1, \
call_5_1_3_2, \
call_5_1_3_3, \
call_5_1_3_4, \
call_5_1_3_5, \
call_5_1_3_6, \
call_5_1_3_7, \
call_5_1_3_8, \
call_5_1_3_9, \
call_5_1_3_10, \
call_5_1_3_11, \
call_5_1_3_12, \
call_5_1_3_13, \
call_5_1_3_14, \
call_5_1_3_15, \
call_5_1_4_0, \
call_5_1_4_1, \
call_5_1_4_2, \
call_5_1_4_3, \
call_5_1_4_4, \
call_5_1_4_5, \
call_5_1_4_6, \
call_5_1_4_7, \
call_5_1_4_8, \
call_5_1_4_9, \
call_5_1_4_10, \
call_5_1_4_11, \
call_5_1_4_12, \
call_5_1_4_13, \
call_5_1_4_14, \
call_5_1_4_15, \
call_5_1_5_0, \
call_5_1_5_1, \
call_5_1_5_2, \
call_5_1_5_3, \
call_5_1_5_4, \
call_5_1_5_5, \
call_5_1_5_6, \
call_5_1_5_7, \
call_5_1_5_8, \
call_5_1_5_9, \
call_5_1_5_10, \
call_5_1_5_11, \
call_5_1_5_12, \
call_5_1_5_13, \
call_5_1_5_14, \
call_5_1_5_15, \
call_5_1_6_0, \
call_5_1_6_1, \
call_5_1_6_2, \
call_5_1_6_3, \
call_5_1_6_4, \
call_5_1_6_5, \
call_5_1_6_6, \
call_5_1_6_7, \
call_5_1_6_8, \
call_5_1_6_9, \
call_5_1_6_10, \
call_5_1_6_11, \
call_5_1_6_12, \
call_5_1_6_13, \
call_5_1_6_14, \
call_5_1_6_15, \
call_5_1_7_0, \
call_5_1_7_1, \
call_5_1_7_2, \
call_5_1_7_3, \
call_5_1_7_4, \
call_5_1_7_5, \
call_5_1_7_6, \
call_5_1_7_7, \
call_5_1_7_8, \
call_5_1_7_9, \
call_5_1_7_10, \
call_5_1_7_11, \
call_5_1_7_12, \
call_5_1_7_13, \
call_5_1_7_14, \
call_5_1_7_15, \
call_5_1_8_0, \
call_5_1_8_1, \
call_5_1_8_2, \
call_5_1_8_3, \
call_5_1_8_4, \
call_5_1_8_5, \
call_5_1_8_6, \
call_5_1_8_7, \
call_5_1_8_8, \
call_5_1_8_9, \
call_5_1_8_10, \
call_5_1_8_11, \
call_5_1_8_12, \
call_5_1_8_13, \
call_5_1_8_14, \
call_5_1_8_15, \
call_5_1_9_0, \
call_5_1_9_1, \
call_5_1_9_2, \
call_5_1_9_3, \
call_5_1_9_4, \
call_5_1_9_5, \
call_5_1_9_6, \
call_5_1_9_7, \
call_5_1_9_8, \
call_5_1_9_9, \
call_5_1_9_10, \
call_5_1_9_11, \
call_5_1_9_12, \
call_5_1_9_13, \
call_5_1_9_14, \
call_5_1_9_15, \
call_5_1_10_0, \
call_5_1_10_1, \
call_5_1_10_2, \
call_5_1_10_3, \
call_5_1_10_4, \
call_5_1_10_5, \
call_5_1_10_6, \
call_5_1_10_7, \
call_5_1_10_8, \
call_5_1_10_9, \
call_5_1_10_10, \
call_5_1_10_11, \
call_5_1_10_12, \
call_5_1_10_13, \
call_5_1_10_14, \
call_5_1_10_15, \
call_5_1_11_0, \
call_5_1_11_1, \
call_5_1_11_2, \
call_5_1_11_3, \
call_5_1_11_4, \
call_5_1_11_5, \
call_5_1_11_6, \
call_5_1_11_7, \
call_5_1_11_8, \
call_5_1_11_9, \
call_5_1_11_10, \
call_5_1_11_11, \
call_5_1_11_12, \
call_5_1_11_13, \
call_5_1_11_14, \
call_5_1_11_15, \
call_5_1_12_0, \
call_5_1_12_1, \
call_5_1_12_2, \
call_5_1_12_3, \
call_5_1_12_4, \
call_5_1_12_5, \
call_5_1_12_6, \
call_5_1_12_7, \
call_5_1_12_8, \
call_5_1_12_9, \
call_5_1_12_10, \
call_5_1_12_11, \
call_5_1_12_12, \
call_5_1_12_13, \
call_5_1_12_14, \
call_5_1_12_15, \
call_5_1_13_0, \
call_5_1_13_1, \
call_5_1_13_2, \
call_5_1_13_3, \
call_5_1_13_4, \
call_5_1_13_5, \
call_5_1_13_6, \
call_5_1_13_7, \
call_5_1_13_8, \
call_5_1_13_9, \
call_5_1_13_10, \
call_5_1_13_11, \
call_5_1_13_12, \
call_5_1_13_13, \
call_5_1_13_14, \
call_5_1_13_15, \
call_5_1_14_0, \
call_5_1_14_1, \
call_5_1_14_2, \
call_5_1_14_3, \
call_5_1_14_4, \
call_5_1_14_5, \
call_5_1_14_6, \
call_5_1_14_7, \
call_5_1_14_8, \
call_5_1_14_9, \
call_5_1_14_10, \
call_5_1_14_11, \
call_5_1_14_12, \
call_5_1_14_13, \
call_5_1_14_14, \
call_5_1_14_15, \
call_5_1_15_0, \
call_5_1_15_1, \
call_5_1_15_2, \
call_5_1_15_3, \
call_5_1_15_4, \
call_5_1_15_5, \
call_5_1_15_6, \
call_5_1_15_7, \
call_5_1_15_8, \
call_5_1_15_9, \
call_5_1_15_10, \
call_5_1_15_11, \
call_5_1_15_12, \
call_5_1_15_13, \
call_5_1_15_14, \
call_5_1_15_15, \
call_5_2_0_0, \
call_5_2_0_1, \
call_5_2_0_2, \
call_5_2_0_3, \
call_5_2_0_4, \
call_5_2_0_5, \
call_5_2_0_6, \
call_5_2_0_7, \
call_5_2_0_8, \
call_5_2_0_9, \
call_5_2_0_10, \
call_5_2_0_11, \
call_5_2_0_12, \
call_5_2_0_13, \
call_5_2_0_14, \
call_5_2_0_15, \
call_5_2_1_0, \
call_5_2_1_1, \
call_5_2_1_2, \
call_5_2_1_3, \
call_5_2_1_4, \
call_5_2_1_5, \
call_5_2_1_6, \
call_5_2_1_7, \
call_5_2_1_8, \
call_5_2_1_9, \
call_5_2_1_10, \
call_5_2_1_11, \
call_5_2_1_12, \
call_5_2_1_13, \
call_5_2_1_14, \
call_5_2_1_15, \
call_5_2_2_0, \
call_5_2_2_1, \
call_5_2_2_2, \
call_5_2_2_3, \
call_5_2_2_4, \
call_5_2_2_5, \
call_5_2_2_6, \
call_5_2_2_7, \
call_5_2_2_8, \
call_5_2_2_9, \
call_5_2_2_10, \
call_5_2_2_11, \
call_5_2_2_12, \
call_5_2_2_13, \
call_5_2_2_14, \
call_5_2_2_15, \
call_5_2_3_0, \
call_5_2_3_1, \
call_5_2_3_2, \
call_5_2_3_3, \
call_5_2_3_4, \
call_5_2_3_5, \
call_5_2_3_6, \
call_5_2_3_7, \
call_5_2_3_8, \
call_5_2_3_9, \
call_5_2_3_10, \
call_5_2_3_11, \
call_5_2_3_12, \
call_5_2_3_13, \
call_5_2_3_14, \
call_5_2_3_15, \
call_5_2_4_0, \
call_5_2_4_1, \
call_5_2_4_2, \
call_5_2_4_3, \
call_5_2_4_4, \
call_5_2_4_5, \
call_5_2_4_6, \
call_5_2_4_7, \
call_5_2_4_8, \
call_5_2_4_9, \
call_5_2_4_10, \
call_5_2_4_11, \
call_5_2_4_12, \
call_5_2_4_13, \
call_5_2_4_14, \
call_5_2_4_15, \
call_5_2_5_0, \
call_5_2_5_1, \
call_5_2_5_2, \
call_5_2_5_3, \
call_5_2_5_4, \
call_5_2_5_5, \
call_5_2_5_6, \
call_5_2_5_7, \
call_5_2_5_8, \
call_5_2_5_9, \
call_5_2_5_10, \
call_5_2_5_11, \
call_5_2_5_12, \
call_5_2_5_13, \
call_5_2_5_14, \
call_5_2_5_15, \
call_5_2_6_0, \
call_5_2_6_1, \
call_5_2_6_2, \
call_5_2_6_3, \
call_5_2_6_4, \
call_5_2_6_5, \
call_5_2_6_6, \
call_5_2_6_7, \
call_5_2_6_8, \
call_5_2_6_9, \
call_5_2_6_10, \
call_5_2_6_11, \
call_5_2_6_12, \
call_5_2_6_13, \
call_5_2_6_14, \
call_5_2_6_15, \
call_5_2_7_0, \
call_5_2_7_1, \
call_5_2_7_2, \
call_5_2_7_3, \
call_5_2_7_4, \
call_5_2_7_5, \
call_5_2_7_6, \
call_5_2_7_7, \
call_5_2_7_8, \
call_5_2_7_9, \
call_5_2_7_10, \
call_5_2_7_11, \
call_5_2_7_12, \
call_5_2_7_13, \
call_5_2_7_14, \
call_5_2_7_15, \
call_5_2_8_0, \
call_5_2_8_1, \
call_5_2_8_2, \
call_5_2_8_3, \
call_5_2_8_4, \
call_5_2_8_5, \
call_5_2_8_6, \
call_5_2_8_7, \
call_5_2_8_8, \
call_5_2_8_9, \
call_5_2_8_10, \
call_5_2_8_11, \
call_5_2_8_12, \
call_5_2_8_13, \
call_5_2_8_14, \
call_5_2_8_15, \
call_5_2_9_0, \
call_5_2_9_1, \
call_5_2_9_2, \
call_5_2_9_3, \
call_5_2_9_4, \
call_5_2_9_5, \
call_5_2_9_6, \
call_5_2_9_7, \
call_5_2_9_8, \
call_5_2_9_9, \
call_5_2_9_10, \
call_5_2_9_11, \
call_5_2_9_12, \
call_5_2_9_13, \
call_5_2_9_14, \
call_5_2_9_15, \
call_5_2_10_0, \
call_5_2_10_1, \
call_5_2_10_2, \
call_5_2_10_3, \
call_5_2_10_4, \
call_5_2_10_5, \
call_5_2_10_6, \
call_5_2_10_7, \
call_5_2_10_8, \
call_5_2_10_9, \
call_5_2_10_10, \
call_5_2_10_11, \
call_5_2_10_12, \
call_5_2_10_13, \
call_5_2_10_14, \
call_5_2_10_15, \
call_5_2_11_0, \
call_5_2_11_1, \
call_5_2_11_2, \
call_5_2_11_3, \
call_5_2_11_4, \
call_5_2_11_5, \
call_5_2_11_6, \
call_5_2_11_7, \
call_5_2_11_8, \
call_5_2_11_9, \
call_5_2_11_10, \
call_5_2_11_11, \
call_5_2_11_12, \
call_5_2_11_13, \
call_5_2_11_14, \
call_5_2_11_15, \
call_5_2_12_0, \
call_5_2_12_1, \
call_5_2_12_2, \
call_5_2_12_3, \
call_5_2_12_4, \
call_5_2_12_5, \
call_5_2_12_6, \
call_5_2_12_7, \
call_5_2_12_8, \
call_5_2_12_9, \
call_5_2_12_10, \
call_5_2_12_11, \
call_5_2_12_12, \
call_5_2_12_13, \
call_5_2_12_14, \
call_5_2_12_15, \
call_5_2_13_0, \
call_5_2_13_1, \
call_5_2_13_2, \
call_5_2_13_3, \
call_5_2_13_4, \
call_5_2_13_5, \
call_5_2_13_6, \
call_5_2_13_7, \
call_5_2_13_8, \
call_5_2_13_9, \
call_5_2_13_10, \
call_5_2_13_11, \
call_5_2_13_12, \
call_5_2_13_13, \
call_5_2_13_14, \
call_5_2_13_15, \
call_5_2_14_0, \
call_5_2_14_1, \
call_5_2_14_2, \
call_5_2_14_3, \
call_5_2_14_4, \
call_5_2_14_5, \
call_5_2_14_6, \
call_5_2_14_7, \
call_5_2_14_8, \
call_5_2_14_9, \
call_5_2_14_10, \
call_5_2_14_11, \
call_5_2_14_12, \
call_5_2_14_13, \
call_5_2_14_14, \
call_5_2_14_15, \
call_5_2_15_0, \
call_5_2_15_1, \
call_5_2_15_2, \
call_5_2_15_3, \
call_5_2_15_4, \
call_5_2_15_5, \
call_5_2_15_6, \
call_5_2_15_7, \
call_5_2_15_8, \
call_5_2_15_9, \
call_5_2_15_10, \
call_5_2_15_11, \
call_5_2_15_12, \
call_5_2_15_13, \
call_5_2_15_14, \
call_5_2_15_15, \
call_5_3_0_0, \
call_5_3_0_1, \
call_5_3_0_2, \
call_5_3_0_3, \
call_5_3_0_4, \
call_5_3_0_5, \
call_5_3_0_6, \
call_5_3_0_7, \
call_5_3_0_8, \
call_5_3_0_9, \
call_5_3_0_10, \
call_5_3_0_11, \
call_5_3_0_12, \
call_5_3_0_13, \
call_5_3_0_14, \
call_5_3_0_15, \
call_5_3_1_0, \
call_5_3_1_1, \
call_5_3_1_2, \
call_5_3_1_3, \
call_5_3_1_4, \
call_5_3_1_5, \
call_5_3_1_6, \
call_5_3_1_7, \
call_5_3_1_8, \
call_5_3_1_9, \
call_5_3_1_10, \
call_5_3_1_11, \
call_5_3_1_12, \
call_5_3_1_13, \
call_5_3_1_14, \
call_5_3_1_15, \
call_5_3_2_0, \
call_5_3_2_1, \
call_5_3_2_2, \
call_5_3_2_3, \
call_5_3_2_4, \
call_5_3_2_5, \
call_5_3_2_6, \
call_5_3_2_7, \
call_5_3_2_8, \
call_5_3_2_9, \
call_5_3_2_10, \
call_5_3_2_11, \
call_5_3_2_12, \
call_5_3_2_13, \
call_5_3_2_14, \
call_5_3_2_15, \
call_5_3_3_0, \
call_5_3_3_1, \
call_5_3_3_2, \
call_5_3_3_3, \
call_5_3_3_4, \
call_5_3_3_5, \
call_5_3_3_6, \
call_5_3_3_7, \
call_5_3_3_8, \
call_5_3_3_9, \
call_5_3_3_10, \
call_5_3_3_11, \
call_5_3_3_12, \
call_5_3_3_13, \
call_5_3_3_14, \
call_5_3_3_15, \
call_5_3_4_0, \
call_5_3_4_1, \
call_5_3_4_2, \
call_5_3_4_3, \
call_5_3_4_4, \
call_5_3_4_5, \
call_5_3_4_6, \
call_5_3_4_7, \
call_5_3_4_8, \
call_5_3_4_9, \
call_5_3_4_10, \
call_5_3_4_11, \
call_5_3_4_12, \
call_5_3_4_13, \
call_5_3_4_14, \
call_5_3_4_15, \
call_5_3_5_0, \
call_5_3_5_1, \
call_5_3_5_2, \
call_5_3_5_3, \
call_5_3_5_4, \
call_5_3_5_5, \
call_5_3_5_6, \
call_5_3_5_7, \
call_5_3_5_8, \
call_5_3_5_9, \
call_5_3_5_10, \
call_5_3_5_11, \
call_5_3_5_12, \
call_5_3_5_13, \
call_5_3_5_14, \
call_5_3_5_15, \
call_5_3_6_0, \
call_5_3_6_1, \
call_5_3_6_2, \
call_5_3_6_3, \
call_5_3_6_4, \
call_5_3_6_5, \
call_5_3_6_6, \
call_5_3_6_7, \
call_5_3_6_8, \
call_5_3_6_9, \
call_5_3_6_10, \
call_5_3_6_11, \
call_5_3_6_12, \
call_5_3_6_13, \
call_5_3_6_14, \
call_5_3_6_15, \
call_5_3_7_0, \
call_5_3_7_1, \
call_5_3_7_2, \
call_5_3_7_3, \
call_5_3_7_4, \
call_5_3_7_5, \
call_5_3_7_6, \
call_5_3_7_7, \
call_5_3_7_8, \
call_5_3_7_9, \
call_5_3_7_10, \
call_5_3_7_11, \
call_5_3_7_12, \
call_5_3_7_13, \
call_5_3_7_14, \
call_5_3_7_15, \
call_5_3_8_0, \
call_5_3_8_1, \
call_5_3_8_2, \
call_5_3_8_3, \
call_5_3_8_4, \
call_5_3_8_5, \
call_5_3_8_6, \
call_5_3_8_7, \
call_5_3_8_8, \
call_5_3_8_9, \
call_5_3_8_10, \
call_5_3_8_11, \
call_5_3_8_12, \
call_5_3_8_13, \
call_5_3_8_14, \
call_5_3_8_15, \
call_5_3_9_0, \
call_5_3_9_1, \
call_5_3_9_2, \
call_5_3_9_3, \
call_5_3_9_4, \
call_5_3_9_5, \
call_5_3_9_6, \
call_5_3_9_7, \
call_5_3_9_8, \
call_5_3_9_9, \
call_5_3_9_10, \
call_5_3_9_11, \
call_5_3_9_12, \
call_5_3_9_13, \
call_5_3_9_14, \
call_5_3_9_15, \
call_5_3_10_0, \
call_5_3_10_1, \
call_5_3_10_2, \
call_5_3_10_3, \
call_5_3_10_4, \
call_5_3_10_5, \
call_5_3_10_6, \
call_5_3_10_7, \
call_5_3_10_8, \
call_5_3_10_9, \
call_5_3_10_10, \
call_5_3_10_11, \
call_5_3_10_12, \
call_5_3_10_13, \
call_5_3_10_14, \
call_5_3_10_15, \
call_5_3_11_0, \
call_5_3_11_1, \
call_5_3_11_2, \
call_5_3_11_3, \
call_5_3_11_4, \
call_5_3_11_5, \
call_5_3_11_6, \
call_5_3_11_7, \
call_5_3_11_8, \
call_5_3_11_9, \
call_5_3_11_10, \
call_5_3_11_11, \
call_5_3_11_12, \
call_5_3_11_13, \
call_5_3_11_14, \
call_5_3_11_15, \
call_5_3_12_0, \
call_5_3_12_1, \
call_5_3_12_2, \
call_5_3_12_3, \
call_5_3_12_4, \
call_5_3_12_5, \
call_5_3_12_6, \
call_5_3_12_7, \
call_5_3_12_8, \
call_5_3_12_9, \
call_5_3_12_10, \
call_5_3_12_11, \
call_5_3_12_12, \
call_5_3_12_13, \
call_5_3_12_14, \
call_5_3_12_15, \
call_5_3_13_0, \
call_5_3_13_1, \
call_5_3_13_2, \
call_5_3_13_3, \
call_5_3_13_4, \
call_5_3_13_5, \
call_5_3_13_6, \
call_5_3_13_7, \
call_5_3_13_8, \
call_5_3_13_9, \
call_5_3_13_10, \
call_5_3_13_11, \
call_5_3_13_12, \
call_5_3_13_13, \
call_5_3_13_14, \
call_5_3_13_15, \
call_5_3_14_0, \
call_5_3_14_1, \
call_5_3_14_2, \
call_5_3_14_3, \
call_5_3_14_4, \
call_5_3_14_5, \
call_5_3_14_6, \
call_5_3_14_7, \
call_5_3_14_8, \
call_5_3_14_9, \
call_5_3_14_10, \
call_5_3_14_11, \
call_5_3_14_12, \
call_5_3_14_13, \
call_5_3_14_14, \
call_5_3_14_15, \
call_5_3_15_0, \
call_5_3_15_1, \
call_5_3_15_2, \
call_5_3_15_3, \
call_5_3_15_4, \
call_5_3_15_5, \
call_5_3_15_6, \
call_5_3_15_7, \
call_5_3_15_8, \
call_5_3_15_9, \
call_5_3_15_10, \
call_5_3_15_11, \
call_5_3_15_12, \
call_5_3_15_13, \
call_5_3_15_14, \
call_5_3_15_15, \
call_5_4_0_0, \
call_5_4_0_1, \
call_5_4_0_2, \
call_5_4_0_3, \
call_5_4_0_4, \
call_5_4_0_5, \
call_5_4_0_6, \
call_5_4_0_7, \
call_5_4_0_8, \
call_5_4_0_9, \
call_5_4_0_10, \
call_5_4_0_11, \
call_5_4_0_12, \
call_5_4_0_13, \
call_5_4_0_14, \
call_5_4_0_15, \
call_5_4_1_0, \
call_5_4_1_1, \
call_5_4_1_2, \
call_5_4_1_3, \
call_5_4_1_4, \
call_5_4_1_5, \
call_5_4_1_6, \
call_5_4_1_7, \
call_5_4_1_8, \
call_5_4_1_9, \
call_5_4_1_10, \
call_5_4_1_11, \
call_5_4_1_12, \
call_5_4_1_13, \
call_5_4_1_14, \
call_5_4_1_15, \
call_5_4_2_0, \
call_5_4_2_1, \
call_5_4_2_2, \
call_5_4_2_3, \
call_5_4_2_4, \
call_5_4_2_5, \
call_5_4_2_6, \
call_5_4_2_7, \
call_5_4_2_8, \
call_5_4_2_9, \
call_5_4_2_10, \
call_5_4_2_11, \
call_5_4_2_12, \
call_5_4_2_13, \
call_5_4_2_14, \
call_5_4_2_15, \
call_5_4_3_0, \
call_5_4_3_1, \
call_5_4_3_2, \
call_5_4_3_3, \
call_5_4_3_4, \
call_5_4_3_5, \
call_5_4_3_6, \
call_5_4_3_7, \
call_5_4_3_8, \
call_5_4_3_9, \
call_5_4_3_10, \
call_5_4_3_11, \
call_5_4_3_12, \
call_5_4_3_13, \
call_5_4_3_14, \
call_5_4_3_15, \
call_5_4_4_0, \
call_5_4_4_1, \
call_5_4_4_2, \
call_5_4_4_3, \
call_5_4_4_4, \
call_5_4_4_5, \
call_5_4_4_6, \
call_5_4_4_7, \
call_5_4_4_8, \
call_5_4_4_9, \
call_5_4_4_10, \
call_5_4_4_11, \
call_5_4_4_12, \
call_5_4_4_13, \
call_5_4_4_14, \
call_5_4_4_15, \
call_5_4_5_0, \
call_5_4_5_1, \
call_5_4_5_2, \
call_5_4_5_3, \
call_5_4_5_4, \
call_5_4_5_5, \
call_5_4_5_6, \
call_5_4_5_7, \
call_5_4_5_8, \
call_5_4_5_9, \
call_5_4_5_10, \
call_5_4_5_11, \
call_5_4_5_12, \
call_5_4_5_13, \
call_5_4_5_14, \
call_5_4_5_15, \
call_5_4_6_0, \
call_5_4_6_1, \
call_5_4_6_2, \
call_5_4_6_3, \
call_5_4_6_4, \
call_5_4_6_5, \
call_5_4_6_6, \
call_5_4_6_7, \
call_5_4_6_8, \
call_5_4_6_9, \
call_5_4_6_10, \
call_5_4_6_11, \
call_5_4_6_12, \
call_5_4_6_13, \
call_5_4_6_14, \
call_5_4_6_15, \
call_5_4_7_0, \
call_5_4_7_1, \
call_5_4_7_2, \
call_5_4_7_3, \
call_5_4_7_4, \
call_5_4_7_5, \
call_5_4_7_6, \
call_5_4_7_7, \
call_5_4_7_8, \
call_5_4_7_9, \
call_5_4_7_10, \
call_5_4_7_11, \
call_5_4_7_12, \
call_5_4_7_13, \
call_5_4_7_14, \
call_5_4_7_15, \
call_5_4_8_0, \
call_5_4_8_1, \
call_5_4_8_2, \
call_5_4_8_3, \
call_5_4_8_4, \
call_5_4_8_5, \
call_5_4_8_6, \
call_5_4_8_7, \
call_5_4_8_8, \
call_5_4_8_9, \
call_5_4_8_10, \
call_5_4_8_11, \
call_5_4_8_12, \
call_5_4_8_13, \
call_5_4_8_14, \
call_5_4_8_15, \
call_5_4_9_0, \
call_5_4_9_1, \
call_5_4_9_2, \
call_5_4_9_3, \
call_5_4_9_4, \
call_5_4_9_5, \
call_5_4_9_6, \
call_5_4_9_7, \
call_5_4_9_8, \
call_5_4_9_9, \
call_5_4_9_10, \
call_5_4_9_11, \
call_5_4_9_12, \
call_5_4_9_13, \
call_5_4_9_14, \
call_5_4_9_15, \
call_5_4_10_0, \
call_5_4_10_1, \
call_5_4_10_2, \
call_5_4_10_3, \
call_5_4_10_4, \
call_5_4_10_5, \
call_5_4_10_6, \
call_5_4_10_7, \
call_5_4_10_8, \
call_5_4_10_9, \
call_5_4_10_10, \
call_5_4_10_11, \
call_5_4_10_12, \
call_5_4_10_13, \
call_5_4_10_14, \
call_5_4_10_15, \
call_5_4_11_0, \
call_5_4_11_1, \
call_5_4_11_2, \
call_5_4_11_3, \
call_5_4_11_4, \
call_5_4_11_5, \
call_5_4_11_6, \
call_5_4_11_7, \
call_5_4_11_8, \
call_5_4_11_9, \
call_5_4_11_10, \
call_5_4_11_11, \
call_5_4_11_12, \
call_5_4_11_13, \
call_5_4_11_14, \
call_5_4_11_15, \
call_5_4_12_0, \
call_5_4_12_1, \
call_5_4_12_2, \
call_5_4_12_3, \
call_5_4_12_4, \
call_5_4_12_5, \
call_5_4_12_6, \
call_5_4_12_7, \
call_5_4_12_8, \
call_5_4_12_9, \
call_5_4_12_10, \
call_5_4_12_11, \
call_5_4_12_12, \
call_5_4_12_13, \
call_5_4_12_14, \
call_5_4_12_15, \
call_5_4_13_0, \
call_5_4_13_1, \
call_5_4_13_2, \
call_5_4_13_3, \
call_5_4_13_4, \
call_5_4_13_5, \
call_5_4_13_6, \
call_5_4_13_7, \
call_5_4_13_8, \
call_5_4_13_9, \
call_5_4_13_10, \
call_5_4_13_11, \
call_5_4_13_12, \
call_5_4_13_13, \
call_5_4_13_14, \
call_5_4_13_15, \
call_5_4_14_0, \
call_5_4_14_1, \
call_5_4_14_2, \
call_5_4_14_3, \
call_5_4_14_4, \
call_5_4_14_5, \
call_5_4_14_6, \
call_5_4_14_7, \
call_5_4_14_8, \
call_5_4_14_9, \
call_5_4_14_10, \
call_5_4_14_11, \
call_5_4_14_12, \
call_5_4_14_13, \
call_5_4_14_14, \
call_5_4_14_15, \
call_5_4_15_0, \
call_5_4_15_1, \
call_5_4_15_2, \
call_5_4_15_3, \
call_5_4_15_4, \
call_5_4_15_5, \
call_5_4_15_6, \
call_5_4_15_7, \
call_5_4_15_8, \
call_5_4_15_9, \
call_5_4_15_10, \
call_5_4_15_11, \
call_5_4_15_12, \
call_5_4_15_13, \
call_5_4_15_14, \
call_5_4_15_15, \
call_5_5_0_0, \
call_5_5_0_1, \
call_5_5_0_2, \
call_5_5_0_3, \
call_5_5_0_4, \
call_5_5_0_5, \
call_5_5_0_6, \
call_5_5_0_7, \
call_5_5_0_8, \
call_5_5_0_9, \
call_5_5_0_10, \
call_5_5_0_11, \
call_5_5_0_12, \
call_5_5_0_13, \
call_5_5_0_14, \
call_5_5_0_15, \
call_5_5_1_0, \
call_5_5_1_1, \
call_5_5_1_2, \
call_5_5_1_3, \
call_5_5_1_4, \
call_5_5_1_5, \
call_5_5_1_6, \
call_5_5_1_7, \
call_5_5_1_8, \
call_5_5_1_9, \
call_5_5_1_10, \
call_5_5_1_11, \
call_5_5_1_12, \
call_5_5_1_13, \
call_5_5_1_14, \
call_5_5_1_15, \
call_5_5_2_0, \
call_5_5_2_1, \
call_5_5_2_2, \
call_5_5_2_3, \
call_5_5_2_4, \
call_5_5_2_5, \
call_5_5_2_6, \
call_5_5_2_7, \
call_5_5_2_8, \
call_5_5_2_9, \
call_5_5_2_10, \
call_5_5_2_11, \
call_5_5_2_12, \
call_5_5_2_13, \
call_5_5_2_14, \
call_5_5_2_15, \
call_5_5_3_0, \
call_5_5_3_1, \
call_5_5_3_2, \
call_5_5_3_3, \
call_5_5_3_4, \
call_5_5_3_5, \
call_5_5_3_6, \
call_5_5_3_7, \
call_5_5_3_8, \
call_5_5_3_9, \
call_5_5_3_10, \
call_5_5_3_11, \
call_5_5_3_12, \
call_5_5_3_13, \
call_5_5_3_14, \
call_5_5_3_15, \
call_5_5_4_0, \
call_5_5_4_1, \
call_5_5_4_2, \
call_5_5_4_3, \
call_5_5_4_4, \
call_5_5_4_5, \
call_5_5_4_6, \
call_5_5_4_7, \
call_5_5_4_8, \
call_5_5_4_9, \
call_5_5_4_10, \
call_5_5_4_11, \
call_5_5_4_12, \
call_5_5_4_13, \
call_5_5_4_14, \
call_5_5_4_15, \
call_5_5_5_0, \
call_5_5_5_1, \
call_5_5_5_2, \
call_5_5_5_3, \
call_5_5_5_4, \
call_5_5_5_5, \
call_5_5_5_6, \
call_5_5_5_7, \
call_5_5_5_8, \
call_5_5_5_9, \
call_5_5_5_10, \
call_5_5_5_11, \
call_5_5_5_12, \
call_5_5_5_13, \
call_5_5_5_14, \
call_5_5_5_15, \
call_5_5_6_0, \
call_5_5_6_1, \
call_5_5_6_2, \
call_5_5_6_3, \
call_5_5_6_4, \
call_5_5_6_5, \
call_5_5_6_6, \
call_5_5_6_7, \
call_5_5_6_8, \
call_5_5_6_9, \
call_5_5_6_10, \
call_5_5_6_11, \
call_5_5_6_12, \
call_5_5_6_13, \
call_5_5_6_14, \
call_5_5_6_15, \
call_5_5_7_0, \
call_5_5_7_1, \
call_5_5_7_2, \
call_5_5_7_3, \
call_5_5_7_4, \
call_5_5_7_5, \
call_5_5_7_6, \
call_5_5_7_7, \
call_5_5_7_8, \
call_5_5_7_9, \
call_5_5_7_10, \
call_5_5_7_11, \
call_5_5_7_12, \
call_5_5_7_13, \
call_5_5_7_14, \
call_5_5_7_15, \
call_5_5_8_0, \
call_5_5_8_1, \
call_5_5_8_2, \
call_5_5_8_3, \
call_5_5_8_4, \
call_5_5_8_5, \
call_5_5_8_6, \
call_5_5_8_7, \
call_5_5_8_8, \
call_5_5_8_9, \
call_5_5_8_10, \
call_5_5_8_11, \
call_5_5_8_12, \
call_5_5_8_13, \
call_5_5_8_14, \
call_5_5_8_15, \
call_5_5_9_0, \
call_5_5_9_1, \
call_5_5_9_2, \
call_5_5_9_3, \
call_5_5_9_4, \
call_5_5_9_5, \
call_5_5_9_6, \
call_5_5_9_7, \
call_5_5_9_8, \
call_5_5_9_9, \
call_5_5_9_10, \
call_5_5_9_11, \
call_5_5_9_12, \
call_5_5_9_13, \
call_5_5_9_14, \
call_5_5_9_15, \
call_5_5_10_0, \
call_5_5_10_1, \
call_5_5_10_2, \
call_5_5_10_3, \
call_5_5_10_4, \
call_5_5_10_5, \
call_5_5_10_6, \
call_5_5_10_7, \
call_5_5_10_8, \
call_5_5_10_9, \
call_5_5_10_10, \
call_5_5_10_11, \
call_5_5_10_12, \
call_5_5_10_13, \
call_5_5_10_14, \
call_5_5_10_15, \
call_5_5_11_0, \
call_5_5_11_1, \
call_5_5_11_2, \
call_5_5_11_3, \
call_5_5_11_4, \
call_5_5_11_5, \
call_5_5_11_6, \
call_5_5_11_7, \
call_5_5_11_8, \
call_5_5_11_9, \
call_5_5_11_10, \
call_5_5_11_11, \
call_5_5_11_12, \
call_5_5_11_13, \
call_5_5_11_14, \
call_5_5_11_15, \
call_5_5_12_0, \
call_5_5_12_1, \
call_5_5_12_2, \
call_5_5_12_3, \
call_5_5_12_4, \
call_5_5_12_5, \
call_5_5_12_6, \
call_5_5_12_7, \
call_5_5_12_8, \
call_5_5_12_9, \
call_5_5_12_10, \
call_5_5_12_11, \
call_5_5_12_12, \
call_5_5_12_13, \
call_5_5_12_14, \
call_5_5_12_15, \
call_5_5_13_0, \
call_5_5_13_1, \
call_5_5_13_2, \
call_5_5_13_3, \
call_5_5_13_4, \
call_5_5_13_5, \
call_5_5_13_6, \
call_5_5_13_7, \
call_5_5_13_8, \
call_5_5_13_9, \
call_5_5_13_10, \
call_5_5_13_11, \
call_5_5_13_12, \
call_5_5_13_13, \
call_5_5_13_14, \
call_5_5_13_15, \
call_5_5_14_0, \
call_5_5_14_1, \
call_5_5_14_2, \
call_5_5_14_3, \
call_5_5_14_4, \
call_5_5_14_5, \
call_5_5_14_6, \
call_5_5_14_7, \
call_5_5_14_8, \
call_5_5_14_9, \
call_5_5_14_10, \
call_5_5_14_11, \
call_5_5_14_12, \
call_5_5_14_13, \
call_5_5_14_14, \
call_5_5_14_15, \
call_5_5_15_0, \
call_5_5_15_1, \
call_5_5_15_2, \
call_5_5_15_3, \
call_5_5_15_4, \
call_5_5_15_5, \
call_5_5_15_6, \
call_5_5_15_7, \
call_5_5_15_8, \
call_5_5_15_9, \
call_5_5_15_10, \
call_5_5_15_11, \
call_5_5_15_12, \
call_5_5_15_13, \
call_5_5_15_14, \
call_5_5_15_15, \
call_5_6_0_0, \
call_5_6_0_1, \
call_5_6_0_2, \
call_5_6_0_3, \
call_5_6_0_4, \
call_5_6_0_5, \
call_5_6_0_6, \
call_5_6_0_7, \
call_5_6_0_8, \
call_5_6_0_9, \
call_5_6_0_10, \
call_5_6_0_11, \
call_5_6_0_12, \
call_5_6_0_13, \
call_5_6_0_14, \
call_5_6_0_15, \
call_5_6_1_0, \
call_5_6_1_1, \
call_5_6_1_2, \
call_5_6_1_3, \
call_5_6_1_4, \
call_5_6_1_5, \
call_5_6_1_6, \
call_5_6_1_7, \
call_5_6_1_8, \
call_5_6_1_9, \
call_5_6_1_10, \
call_5_6_1_11, \
call_5_6_1_12, \
call_5_6_1_13, \
call_5_6_1_14, \
call_5_6_1_15, \
call_5_6_2_0, \
call_5_6_2_1, \
call_5_6_2_2, \
call_5_6_2_3, \
call_5_6_2_4, \
call_5_6_2_5, \
call_5_6_2_6, \
call_5_6_2_7, \
call_5_6_2_8, \
call_5_6_2_9, \
call_5_6_2_10, \
call_5_6_2_11, \
call_5_6_2_12, \
call_5_6_2_13, \
call_5_6_2_14, \
call_5_6_2_15, \
call_5_6_3_0, \
call_5_6_3_1, \
call_5_6_3_2, \
call_5_6_3_3, \
call_5_6_3_4, \
call_5_6_3_5, \
call_5_6_3_6, \
call_5_6_3_7, \
call_5_6_3_8, \
call_5_6_3_9, \
call_5_6_3_10, \
call_5_6_3_11, \
call_5_6_3_12, \
call_5_6_3_13, \
call_5_6_3_14, \
call_5_6_3_15, \
call_5_6_4_0, \
call_5_6_4_1, \
call_5_6_4_2, \
call_5_6_4_3, \
call_5_6_4_4, \
call_5_6_4_5, \
call_5_6_4_6, \
call_5_6_4_7, \
call_5_6_4_8, \
call_5_6_4_9, \
call_5_6_4_10, \
call_5_6_4_11, \
call_5_6_4_12, \
call_5_6_4_13, \
call_5_6_4_14, \
call_5_6_4_15, \
call_5_6_5_0, \
call_5_6_5_1, \
call_5_6_5_2, \
call_5_6_5_3, \
call_5_6_5_4, \
call_5_6_5_5, \
call_5_6_5_6, \
call_5_6_5_7, \
call_5_6_5_8, \
call_5_6_5_9, \
call_5_6_5_10, \
call_5_6_5_11, \
call_5_6_5_12, \
call_5_6_5_13, \
call_5_6_5_14, \
call_5_6_5_15, \
call_5_6_6_0, \
call_5_6_6_1, \
call_5_6_6_2, \
call_5_6_6_3, \
call_5_6_6_4, \
call_5_6_6_5, \
call_5_6_6_6, \
call_5_6_6_7, \
call_5_6_6_8, \
call_5_6_6_9, \
call_5_6_6_10, \
call_5_6_6_11, \
call_5_6_6_12, \
call_5_6_6_13, \
call_5_6_6_14, \
call_5_6_6_15, \
call_5_6_7_0, \
call_5_6_7_1, \
call_5_6_7_2, \
call_5_6_7_3, \
call_5_6_7_4, \
call_5_6_7_5, \
call_5_6_7_6, \
call_5_6_7_7, \
call_5_6_7_8, \
call_5_6_7_9, \
call_5_6_7_10, \
call_5_6_7_11, \
call_5_6_7_12, \
call_5_6_7_13, \
call_5_6_7_14, \
call_5_6_7_15, \
call_5_6_8_0, \
call_5_6_8_1, \
call_5_6_8_2, \
call_5_6_8_3, \
call_5_6_8_4, \
call_5_6_8_5, \
call_5_6_8_6, \
call_5_6_8_7, \
call_5_6_8_8, \
call_5_6_8_9, \
call_5_6_8_10, \
call_5_6_8_11, \
call_5_6_8_12, \
call_5_6_8_13, \
call_5_6_8_14, \
call_5_6_8_15, \
call_5_6_9_0, \
call_5_6_9_1, \
call_5_6_9_2, \
call_5_6_9_3, \
call_5_6_9_4, \
call_5_6_9_5, \
call_5_6_9_6, \
call_5_6_9_7, \
call_5_6_9_8, \
call_5_6_9_9, \
call_5_6_9_10, \
call_5_6_9_11, \
call_5_6_9_12, \
call_5_6_9_13, \
call_5_6_9_14, \
call_5_6_9_15, \
call_5_6_10_0, \
call_5_6_10_1, \
call_5_6_10_2, \
call_5_6_10_3, \
call_5_6_10_4, \
call_5_6_10_5, \
call_5_6_10_6, \
call_5_6_10_7, \
call_5_6_10_8, \
call_5_6_10_9, \
call_5_6_10_10, \
call_5_6_10_11, \
call_5_6_10_12, \
call_5_6_10_13, \
call_5_6_10_14, \
call_5_6_10_15, \
call_5_6_11_0, \
call_5_6_11_1, \
call_5_6_11_2, \
call_5_6_11_3, \
call_5_6_11_4, \
call_5_6_11_5, \
call_5_6_11_6, \
call_5_6_11_7, \
call_5_6_11_8, \
call_5_6_11_9, \
call_5_6_11_10, \
call_5_6_11_11, \
call_5_6_11_12, \
call_5_6_11_13, \
call_5_6_11_14, \
call_5_6_11_15, \
call_5_6_12_0, \
call_5_6_12_1, \
call_5_6_12_2, \
call_5_6_12_3, \
call_5_6_12_4, \
call_5_6_12_5, \
call_5_6_12_6, \
call_5_6_12_7, \
call_5_6_12_8, \
call_5_6_12_9, \
call_5_6_12_10, \
call_5_6_12_11, \
call_5_6_12_12, \
call_5_6_12_13, \
call_5_6_12_14, \
call_5_6_12_15, \
call_5_6_13_0, \
call_5_6_13_1, \
call_5_6_13_2, \
call_5_6_13_3, \
call_5_6_13_4, \
call_5_6_13_5, \
call_5_6_13_6, \
call_5_6_13_7, \
call_5_6_13_8, \
call_5_6_13_9, \
call_5_6_13_10, \
call_5_6_13_11, \
call_5_6_13_12, \
call_5_6_13_13, \
call_5_6_13_14, \
call_5_6_13_15, \
call_5_6_14_0, \
call_5_6_14_1, \
call_5_6_14_2, \
call_5_6_14_3, \
call_5_6_14_4, \
call_5_6_14_5, \
call_5_6_14_6, \
call_5_6_14_7, \
call_5_6_14_8, \
call_5_6_14_9, \
call_5_6_14_10, \
call_5_6_14_11, \
call_5_6_14_12, \
call_5_6_14_13, \
call_5_6_14_14, \
call_5_6_14_15, \
call_5_6_15_0, \
call_5_6_15_1, \
call_5_6_15_2, \
call_5_6_15_3, \
call_5_6_15_4, \
call_5_6_15_5, \
call_5_6_15_6, \
call_5_6_15_7, \
call_5_6_15_8, \
call_5_6_15_9, \
call_5_6_15_10, \
call_5_6_15_11, \
call_5_6_15_12, \
call_5_6_15_13, \
call_5_6_15_14, \
call_5_6_15_15, \
call_5_7_0_0, \
call_5_7_0_1, \
call_5_7_0_2, \
call_5_7_0_3, \
call_5_7_0_4, \
call_5_7_0_5, \
call_5_7_0_6, \
call_5_7_0_7, \
call_5_7_0_8, \
call_5_7_0_9, \
call_5_7_0_10, \
call_5_7_0_11, \
call_5_7_0_12, \
call_5_7_0_13, \
call_5_7_0_14, \
call_5_7_0_15, \
call_5_7_1_0, \
call_5_7_1_1, \
call_5_7_1_2, \
call_5_7_1_3, \
call_5_7_1_4, \
call_5_7_1_5, \
call_5_7_1_6, \
call_5_7_1_7, \
call_5_7_1_8, \
call_5_7_1_9, \
call_5_7_1_10, \
call_5_7_1_11, \
call_5_7_1_12, \
call_5_7_1_13, \
call_5_7_1_14, \
call_5_7_1_15, \
call_5_7_2_0, \
call_5_7_2_1, \
call_5_7_2_2, \
call_5_7_2_3, \
call_5_7_2_4, \
call_5_7_2_5, \
call_5_7_2_6, \
call_5_7_2_7, \
call_5_7_2_8, \
call_5_7_2_9, \
call_5_7_2_10, \
call_5_7_2_11, \
call_5_7_2_12, \
call_5_7_2_13, \
call_5_7_2_14, \
call_5_7_2_15, \
call_5_7_3_0, \
call_5_7_3_1, \
call_5_7_3_2, \
call_5_7_3_3, \
call_5_7_3_4, \
call_5_7_3_5, \
call_5_7_3_6, \
call_5_7_3_7, \
call_5_7_3_8, \
call_5_7_3_9, \
call_5_7_3_10, \
call_5_7_3_11, \
call_5_7_3_12, \
call_5_7_3_13, \
call_5_7_3_14, \
call_5_7_3_15, \
call_5_7_4_0, \
call_5_7_4_1, \
call_5_7_4_2, \
call_5_7_4_3, \
call_5_7_4_4, \
call_5_7_4_5, \
call_5_7_4_6, \
call_5_7_4_7, \
call_5_7_4_8, \
call_5_7_4_9, \
call_5_7_4_10, \
call_5_7_4_11, \
call_5_7_4_12, \
call_5_7_4_13, \
call_5_7_4_14, \
call_5_7_4_15, \
call_5_7_5_0, \
call_5_7_5_1, \
call_5_7_5_2, \
call_5_7_5_3, \
call_5_7_5_4, \
call_5_7_5_5, \
call_5_7_5_6, \
call_5_7_5_7, \
call_5_7_5_8, \
call_5_7_5_9, \
call_5_7_5_10, \
call_5_7_5_11, \
call_5_7_5_12, \
call_5_7_5_13, \
call_5_7_5_14, \
call_5_7_5_15, \
call_5_7_6_0, \
call_5_7_6_1, \
call_5_7_6_2, \
call_5_7_6_3, \
call_5_7_6_4, \
call_5_7_6_5, \
call_5_7_6_6, \
call_5_7_6_7, \
call_5_7_6_8, \
call_5_7_6_9, \
call_5_7_6_10, \
call_5_7_6_11, \
call_5_7_6_12, \
call_5_7_6_13, \
call_5_7_6_14, \
call_5_7_6_15, \
call_5_7_7_0, \
call_5_7_7_1, \
call_5_7_7_2, \
call_5_7_7_3, \
call_5_7_7_4, \
call_5_7_7_5, \
call_5_7_7_6, \
call_5_7_7_7, \
call_5_7_7_8, \
call_5_7_7_9, \
call_5_7_7_10, \
call_5_7_7_11, \
call_5_7_7_12, \
call_5_7_7_13, \
call_5_7_7_14, \
call_5_7_7_15, \
call_5_7_8_0, \
call_5_7_8_1, \
call_5_7_8_2, \
call_5_7_8_3, \
call_5_7_8_4, \
call_5_7_8_5, \
call_5_7_8_6, \
call_5_7_8_7, \
call_5_7_8_8, \
call_5_7_8_9, \
call_5_7_8_10, \
call_5_7_8_11, \
call_5_7_8_12, \
call_5_7_8_13, \
call_5_7_8_14, \
call_5_7_8_15, \
call_5_7_9_0, \
call_5_7_9_1, \
call_5_7_9_2, \
call_5_7_9_3, \
call_5_7_9_4, \
call_5_7_9_5, \
call_5_7_9_6, \
call_5_7_9_7, \
call_5_7_9_8, \
call_5_7_9_9, \
call_5_7_9_10, \
call_5_7_9_11, \
call_5_7_9_12, \
call_5_7_9_13, \
call_5_7_9_14, \
call_5_7_9_15, \
call_5_7_10_0, \
call_5_7_10_1, \
call_5_7_10_2, \
call_5_7_10_3, \
call_5_7_10_4, \
call_5_7_10_5, \
call_5_7_10_6, \
call_5_7_10_7, \
call_5_7_10_8, \
call_5_7_10_9, \
call_5_7_10_10, \
call_5_7_10_11, \
call_5_7_10_12, \
call_5_7_10_13, \
call_5_7_10_14, \
call_5_7_10_15, \
call_5_7_11_0, \
call_5_7_11_1, \
call_5_7_11_2, \
call_5_7_11_3, \
call_5_7_11_4, \
call_5_7_11_5, \
call_5_7_11_6, \
call_5_7_11_7, \
call_5_7_11_8, \
call_5_7_11_9, \
call_5_7_11_10, \
call_5_7_11_11, \
call_5_7_11_12, \
call_5_7_11_13, \
call_5_7_11_14, \
call_5_7_11_15, \
call_5_7_12_0, \
call_5_7_12_1, \
call_5_7_12_2, \
call_5_7_12_3, \
call_5_7_12_4, \
call_5_7_12_5, \
call_5_7_12_6, \
call_5_7_12_7, \
call_5_7_12_8, \
call_5_7_12_9, \
call_5_7_12_10, \
call_5_7_12_11, \
call_5_7_12_12, \
call_5_7_12_13, \
call_5_7_12_14, \
call_5_7_12_15, \
call_5_7_13_0, \
call_5_7_13_1, \
call_5_7_13_2, \
call_5_7_13_3, \
call_5_7_13_4, \
call_5_7_13_5, \
call_5_7_13_6, \
call_5_7_13_7, \
call_5_7_13_8, \
call_5_7_13_9, \
call_5_7_13_10, \
call_5_7_13_11, \
call_5_7_13_12, \
call_5_7_13_13, \
call_5_7_13_14, \
call_5_7_13_15, \
call_5_7_14_0, \
call_5_7_14_1, \
call_5_7_14_2, \
call_5_7_14_3, \
call_5_7_14_4, \
call_5_7_14_5, \
call_5_7_14_6, \
call_5_7_14_7, \
call_5_7_14_8, \
call_5_7_14_9, \
call_5_7_14_10, \
call_5_7_14_11, \
call_5_7_14_12, \
call_5_7_14_13, \
call_5_7_14_14, \
call_5_7_14_15, \
call_5_7_15_0, \
call_5_7_15_1, \
call_5_7_15_2, \
call_5_7_15_3, \
call_5_7_15_4, \
call_5_7_15_5, \
call_5_7_15_6, \
call_5_7_15_7, \
call_5_7_15_8, \
call_5_7_15_9, \
call_5_7_15_10, \
call_5_7_15_11, \
call_5_7_15_12, \
call_5_7_15_13, \
call_5_7_15_14, \
call_5_7_15_15, \
call_5_8_0_0, \
call_5_8_0_1, \
call_5_8_0_2, \
call_5_8_0_3, \
call_5_8_0_4, \
call_5_8_0_5, \
call_5_8_0_6, \
call_5_8_0_7, \
call_5_8_0_8, \
call_5_8_0_9, \
call_5_8_0_10, \
call_5_8_0_11, \
call_5_8_0_12, \
call_5_8_0_13, \
call_5_8_0_14, \
call_5_8_0_15, \
call_5_8_1_0, \
call_5_8_1_1, \
call_5_8_1_2, \
call_5_8_1_3, \
call_5_8_1_4, \
call_5_8_1_5, \
call_5_8_1_6, \
call_5_8_1_7, \
call_5_8_1_8, \
call_5_8_1_9, \
call_5_8_1_10, \
call_5_8_1_11, \
call_5_8_1_12, \
call_5_8_1_13, \
call_5_8_1_14, \
call_5_8_1_15, \
call_5_8_2_0, \
call_5_8_2_1, \
call_5_8_2_2, \
call_5_8_2_3, \
call_5_8_2_4, \
call_5_8_2_5, \
call_5_8_2_6, \
call_5_8_2_7, \
call_5_8_2_8, \
call_5_8_2_9, \
call_5_8_2_10, \
call_5_8_2_11, \
call_5_8_2_12, \
call_5_8_2_13, \
call_5_8_2_14, \
call_5_8_2_15, \
call_5_8_3_0, \
call_5_8_3_1, \
call_5_8_3_2, \
call_5_8_3_3, \
call_5_8_3_4, \
call_5_8_3_5, \
call_5_8_3_6, \
call_5_8_3_7, \
call_5_8_3_8, \
call_5_8_3_9, \
call_5_8_3_10, \
call_5_8_3_11, \
call_5_8_3_12, \
call_5_8_3_13, \
call_5_8_3_14, \
call_5_8_3_15, \
call_5_8_4_0, \
call_5_8_4_1, \
call_5_8_4_2, \
call_5_8_4_3, \
call_5_8_4_4, \
call_5_8_4_5, \
call_5_8_4_6, \
call_5_8_4_7, \
call_5_8_4_8, \
call_5_8_4_9, \
call_5_8_4_10, \
call_5_8_4_11, \
call_5_8_4_12, \
call_5_8_4_13, \
call_5_8_4_14, \
call_5_8_4_15, \
call_5_8_5_0, \
call_5_8_5_1, \
call_5_8_5_2, \
call_5_8_5_3, \
call_5_8_5_4, \
call_5_8_5_5, \
call_5_8_5_6, \
call_5_8_5_7, \
call_5_8_5_8, \
call_5_8_5_9, \
call_5_8_5_10, \
call_5_8_5_11, \
call_5_8_5_12, \
call_5_8_5_13, \
call_5_8_5_14, \
call_5_8_5_15, \
call_5_8_6_0, \
call_5_8_6_1, \
call_5_8_6_2, \
call_5_8_6_3, \
call_5_8_6_4, \
call_5_8_6_5, \
call_5_8_6_6, \
call_5_8_6_7, \
call_5_8_6_8, \
call_5_8_6_9, \
call_5_8_6_10, \
call_5_8_6_11, \
call_5_8_6_12, \
call_5_8_6_13, \
call_5_8_6_14, \
call_5_8_6_15, \
call_5_8_7_0, \
call_5_8_7_1, \
call_5_8_7_2, \
call_5_8_7_3, \
call_5_8_7_4, \
call_5_8_7_5, \
call_5_8_7_6, \
call_5_8_7_7, \
call_5_8_7_8, \
call_5_8_7_9, \
call_5_8_7_10, \
call_5_8_7_11, \
call_5_8_7_12, \
call_5_8_7_13, \
call_5_8_7_14, \
call_5_8_7_15, \
call_5_8_8_0, \
call_5_8_8_1, \
call_5_8_8_2, \
call_5_8_8_3, \
call_5_8_8_4, \
call_5_8_8_5, \
call_5_8_8_6, \
call_5_8_8_7, \
call_5_8_8_8, \
call_5_8_8_9, \
call_5_8_8_10, \
call_5_8_8_11, \
call_5_8_8_12, \
call_5_8_8_13, \
call_5_8_8_14, \
call_5_8_8_15, \
call_5_8_9_0, \
call_5_8_9_1, \
call_5_8_9_2, \
call_5_8_9_3, \
call_5_8_9_4, \
call_5_8_9_5, \
call_5_8_9_6, \
call_5_8_9_7, \
call_5_8_9_8, \
call_5_8_9_9, \
call_5_8_9_10, \
call_5_8_9_11, \
call_5_8_9_12, \
call_5_8_9_13, \
call_5_8_9_14, \
call_5_8_9_15, \
call_5_8_10_0, \
call_5_8_10_1, \
call_5_8_10_2, \
call_5_8_10_3, \
call_5_8_10_4, \
call_5_8_10_5, \
call_5_8_10_6, \
call_5_8_10_7, \
call_5_8_10_8, \
call_5_8_10_9, \
call_5_8_10_10, \
call_5_8_10_11, \
call_5_8_10_12, \
call_5_8_10_13, \
call_5_8_10_14, \
call_5_8_10_15, \
call_5_8_11_0, \
call_5_8_11_1, \
call_5_8_11_2, \
call_5_8_11_3, \
call_5_8_11_4, \
call_5_8_11_5, \
call_5_8_11_6, \
call_5_8_11_7, \
call_5_8_11_8, \
call_5_8_11_9, \
call_5_8_11_10, \
call_5_8_11_11, \
call_5_8_11_12, \
call_5_8_11_13, \
call_5_8_11_14, \
call_5_8_11_15, \
call_5_8_12_0, \
call_5_8_12_1, \
call_5_8_12_2, \
call_5_8_12_3, \
call_5_8_12_4, \
call_5_8_12_5, \
call_5_8_12_6, \
call_5_8_12_7, \
call_5_8_12_8, \
call_5_8_12_9, \
call_5_8_12_10, \
call_5_8_12_11, \
call_5_8_12_12, \
call_5_8_12_13, \
call_5_8_12_14, \
call_5_8_12_15, \
call_5_8_13_0, \
call_5_8_13_1, \
call_5_8_13_2, \
call_5_8_13_3, \
call_5_8_13_4, \
call_5_8_13_5, \
call_5_8_13_6, \
call_5_8_13_7, \
call_5_8_13_8, \
call_5_8_13_9, \
call_5_8_13_10, \
call_5_8_13_11, \
call_5_8_13_12, \
call_5_8_13_13, \
call_5_8_13_14, \
call_5_8_13_15, \
call_5_8_14_0, \
call_5_8_14_1, \
call_5_8_14_2, \
call_5_8_14_3, \
call_5_8_14_4, \
call_5_8_14_5, \
call_5_8_14_6, \
call_5_8_14_7, \
call_5_8_14_8, \
call_5_8_14_9, \
call_5_8_14_10, \
call_5_8_14_11, \
call_5_8_14_12, \
call_5_8_14_13, \
call_5_8_14_14, \
call_5_8_14_15, \
call_5_8_15_0, \
call_5_8_15_1, \
call_5_8_15_2, \
call_5_8_15_3, \
call_5_8_15_4, \
call_5_8_15_5, \
call_5_8_15_6, \
call_5_8_15_7, \
call_5_8_15_8, \
call_5_8_15_9, \
call_5_8_15_10, \
call_5_8_15_11, \
call_5_8_15_12, \
call_5_8_15_13, \
call_5_8_15_14, \
call_5_8_15_15, \
call_5_9_0_0, \
call_5_9_0_1, \
call_5_9_0_2, \
call_5_9_0_3, \
call_5_9_0_4, \
call_5_9_0_5, \
call_5_9_0_6, \
call_5_9_0_7, \
call_5_9_0_8, \
call_5_9_0_9, \
call_5_9_0_10, \
call_5_9_0_11, \
call_5_9_0_12, \
call_5_9_0_13, \
call_5_9_0_14, \
call_5_9_0_15, \
call_5_9_1_0, \
call_5_9_1_1, \
call_5_9_1_2, \
call_5_9_1_3, \
call_5_9_1_4, \
call_5_9_1_5, \
call_5_9_1_6, \
call_5_9_1_7, \
call_5_9_1_8, \
call_5_9_1_9, \
call_5_9_1_10, \
call_5_9_1_11, \
call_5_9_1_12, \
call_5_9_1_13, \
call_5_9_1_14, \
call_5_9_1_15, \
call_5_9_2_0, \
call_5_9_2_1, \
call_5_9_2_2, \
call_5_9_2_3, \
call_5_9_2_4, \
call_5_9_2_5, \
call_5_9_2_6, \
call_5_9_2_7, \
call_5_9_2_8, \
call_5_9_2_9, \
call_5_9_2_10, \
call_5_9_2_11, \
call_5_9_2_12, \
call_5_9_2_13, \
call_5_9_2_14, \
call_5_9_2_15, \
call_5_9_3_0, \
call_5_9_3_1, \
call_5_9_3_2, \
call_5_9_3_3, \
call_5_9_3_4, \
call_5_9_3_5, \
call_5_9_3_6, \
call_5_9_3_7, \
call_5_9_3_8, \
call_5_9_3_9, \
call_5_9_3_10, \
call_5_9_3_11, \
call_5_9_3_12, \
call_5_9_3_13, \
call_5_9_3_14, \
call_5_9_3_15, \
call_5_9_4_0, \
call_5_9_4_1, \
call_5_9_4_2, \
call_5_9_4_3, \
call_5_9_4_4, \
call_5_9_4_5, \
call_5_9_4_6, \
call_5_9_4_7, \
call_5_9_4_8, \
call_5_9_4_9, \
call_5_9_4_10, \
call_5_9_4_11, \
call_5_9_4_12, \
call_5_9_4_13, \
call_5_9_4_14, \
call_5_9_4_15, \
call_5_9_5_0, \
call_5_9_5_1, \
call_5_9_5_2, \
call_5_9_5_3, \
call_5_9_5_4, \
call_5_9_5_5, \
call_5_9_5_6, \
call_5_9_5_7, \
call_5_9_5_8, \
call_5_9_5_9, \
call_5_9_5_10, \
call_5_9_5_11, \
call_5_9_5_12, \
call_5_9_5_13, \
call_5_9_5_14, \
call_5_9_5_15, \
call_5_9_6_0, \
call_5_9_6_1, \
call_5_9_6_2, \
call_5_9_6_3, \
call_5_9_6_4, \
call_5_9_6_5, \
call_5_9_6_6, \
call_5_9_6_7, \
call_5_9_6_8, \
call_5_9_6_9, \
call_5_9_6_10, \
call_5_9_6_11, \
call_5_9_6_12, \
call_5_9_6_13, \
call_5_9_6_14, \
call_5_9_6_15, \
call_5_9_7_0, \
call_5_9_7_1, \
call_5_9_7_2, \
call_5_9_7_3, \
call_5_9_7_4, \
call_5_9_7_5, \
call_5_9_7_6, \
call_5_9_7_7, \
call_5_9_7_8, \
call_5_9_7_9, \
call_5_9_7_10, \
call_5_9_7_11, \
call_5_9_7_12, \
call_5_9_7_13, \
call_5_9_7_14, \
call_5_9_7_15, \
call_5_9_8_0, \
call_5_9_8_1, \
call_5_9_8_2, \
call_5_9_8_3, \
call_5_9_8_4, \
call_5_9_8_5, \
call_5_9_8_6, \
call_5_9_8_7, \
call_5_9_8_8, \
call_5_9_8_9, \
call_5_9_8_10, \
call_5_9_8_11, \
call_5_9_8_12, \
call_5_9_8_13, \
call_5_9_8_14, \
call_5_9_8_15, \
call_5_9_9_0, \
call_5_9_9_1, \
call_5_9_9_2, \
call_5_9_9_3, \
call_5_9_9_4, \
call_5_9_9_5, \
call_5_9_9_6, \
call_5_9_9_7, \
call_5_9_9_8, \
call_5_9_9_9, \
call_5_9_9_10, \
call_5_9_9_11, \
call_5_9_9_12, \
call_5_9_9_13, \
call_5_9_9_14, \
call_5_9_9_15, \
call_5_9_10_0, \
call_5_9_10_1, \
call_5_9_10_2, \
call_5_9_10_3, \
call_5_9_10_4, \
call_5_9_10_5, \
call_5_9_10_6, \
call_5_9_10_7, \
call_5_9_10_8, \
call_5_9_10_9, \
call_5_9_10_10, \
call_5_9_10_11, \
call_5_9_10_12, \
call_5_9_10_13, \
call_5_9_10_14, \
call_5_9_10_15, \
call_5_9_11_0, \
call_5_9_11_1, \
call_5_9_11_2, \
call_5_9_11_3, \
call_5_9_11_4, \
call_5_9_11_5, \
call_5_9_11_6, \
call_5_9_11_7, \
call_5_9_11_8, \
call_5_9_11_9, \
call_5_9_11_10, \
call_5_9_11_11, \
call_5_9_11_12, \
call_5_9_11_13, \
call_5_9_11_14, \
call_5_9_11_15, \
call_5_9_12_0, \
call_5_9_12_1, \
call_5_9_12_2, \
call_5_9_12_3, \
call_5_9_12_4, \
call_5_9_12_5, \
call_5_9_12_6, \
call_5_9_12_7, \
call_5_9_12_8, \
call_5_9_12_9, \
call_5_9_12_10, \
call_5_9_12_11, \
call_5_9_12_12, \
call_5_9_12_13, \
call_5_9_12_14, \
call_5_9_12_15, \
call_5_9_13_0, \
call_5_9_13_1, \
call_5_9_13_2, \
call_5_9_13_3, \
call_5_9_13_4, \
call_5_9_13_5, \
call_5_9_13_6, \
call_5_9_13_7, \
call_5_9_13_8, \
call_5_9_13_9, \
call_5_9_13_10, \
call_5_9_13_11, \
call_5_9_13_12, \
call_5_9_13_13, \
call_5_9_13_14, \
call_5_9_13_15, \
call_5_9_14_0, \
call_5_9_14_1, \
call_5_9_14_2, \
call_5_9_14_3, \
call_5_9_14_4, \
call_5_9_14_5, \
call_5_9_14_6, \
call_5_9_14_7, \
call_5_9_14_8, \
call_5_9_14_9, \
call_5_9_14_10, \
call_5_9_14_11, \
call_5_9_14_12, \
call_5_9_14_13, \
call_5_9_14_14, \
call_5_9_14_15, \
call_5_9_15_0, \
call_5_9_15_1, \
call_5_9_15_2, \
call_5_9_15_3, \
call_5_9_15_4, \
call_5_9_15_5, \
call_5_9_15_6, \
call_5_9_15_7, \
call_5_9_15_8, \
call_5_9_15_9, \
call_5_9_15_10, \
call_5_9_15_11, \
call_5_9_15_12, \
call_5_9_15_13, \
call_5_9_15_14, \
call_5_9_15_15, \
call_5_10_0_0, \
call_5_10_0_1, \
call_5_10_0_2, \
call_5_10_0_3, \
call_5_10_0_4, \
call_5_10_0_5, \
call_5_10_0_6, \
call_5_10_0_7, \
call_5_10_0_8, \
call_5_10_0_9, \
call_5_10_0_10, \
call_5_10_0_11, \
call_5_10_0_12, \
call_5_10_0_13, \
call_5_10_0_14, \
call_5_10_0_15, \
call_5_10_1_0, \
call_5_10_1_1, \
call_5_10_1_2, \
call_5_10_1_3, \
call_5_10_1_4, \
call_5_10_1_5, \
call_5_10_1_6, \
call_5_10_1_7, \
call_5_10_1_8, \
call_5_10_1_9, \
call_5_10_1_10, \
call_5_10_1_11, \
call_5_10_1_12, \
call_5_10_1_13, \
call_5_10_1_14, \
call_5_10_1_15, \
call_5_10_2_0, \
call_5_10_2_1, \
call_5_10_2_2, \
call_5_10_2_3, \
call_5_10_2_4, \
call_5_10_2_5, \
call_5_10_2_6, \
call_5_10_2_7, \
call_5_10_2_8, \
call_5_10_2_9, \
call_5_10_2_10, \
call_5_10_2_11, \
call_5_10_2_12, \
call_5_10_2_13, \
call_5_10_2_14, \
call_5_10_2_15, \
call_5_10_3_0, \
call_5_10_3_1, \
call_5_10_3_2, \
call_5_10_3_3, \
call_5_10_3_4, \
call_5_10_3_5, \
call_5_10_3_6, \
call_5_10_3_7, \
call_5_10_3_8, \
call_5_10_3_9, \
call_5_10_3_10, \
call_5_10_3_11, \
call_5_10_3_12, \
call_5_10_3_13, \
call_5_10_3_14, \
call_5_10_3_15, \
call_5_10_4_0, \
call_5_10_4_1, \
call_5_10_4_2, \
call_5_10_4_3, \
call_5_10_4_4, \
call_5_10_4_5, \
call_5_10_4_6, \
call_5_10_4_7, \
call_5_10_4_8, \
call_5_10_4_9, \
call_5_10_4_10, \
call_5_10_4_11, \
call_5_10_4_12, \
call_5_10_4_13, \
call_5_10_4_14, \
call_5_10_4_15, \
call_5_10_5_0, \
call_5_10_5_1, \
call_5_10_5_2, \
call_5_10_5_3, \
call_5_10_5_4, \
call_5_10_5_5, \
call_5_10_5_6, \
call_5_10_5_7, \
call_5_10_5_8, \
call_5_10_5_9, \
call_5_10_5_10, \
call_5_10_5_11, \
call_5_10_5_12, \
call_5_10_5_13, \
call_5_10_5_14, \
call_5_10_5_15, \
call_5_10_6_0, \
call_5_10_6_1, \
call_5_10_6_2, \
call_5_10_6_3, \
call_5_10_6_4, \
call_5_10_6_5, \
call_5_10_6_6, \
call_5_10_6_7, \
call_5_10_6_8, \
call_5_10_6_9, \
call_5_10_6_10, \
call_5_10_6_11, \
call_5_10_6_12, \
call_5_10_6_13, \
call_5_10_6_14, \
call_5_10_6_15, \
call_5_10_7_0, \
call_5_10_7_1, \
call_5_10_7_2, \
call_5_10_7_3, \
call_5_10_7_4, \
call_5_10_7_5, \
call_5_10_7_6, \
call_5_10_7_7, \
call_5_10_7_8, \
call_5_10_7_9, \
call_5_10_7_10, \
call_5_10_7_11, \
call_5_10_7_12, \
call_5_10_7_13, \
call_5_10_7_14, \
call_5_10_7_15, \
call_5_10_8_0, \
call_5_10_8_1, \
call_5_10_8_2, \
call_5_10_8_3, \
call_5_10_8_4, \
call_5_10_8_5, \
call_5_10_8_6, \
call_5_10_8_7, \
call_5_10_8_8, \
call_5_10_8_9, \
call_5_10_8_10, \
call_5_10_8_11, \
call_5_10_8_12, \
call_5_10_8_13, \
call_5_10_8_14, \
call_5_10_8_15, \
call_5_10_9_0, \
call_5_10_9_1, \
call_5_10_9_2, \
call_5_10_9_3, \
call_5_10_9_4, \
call_5_10_9_5, \
call_5_10_9_6, \
call_5_10_9_7, \
call_5_10_9_8, \
call_5_10_9_9, \
call_5_10_9_10, \
call_5_10_9_11, \
call_5_10_9_12, \
call_5_10_9_13, \
call_5_10_9_14, \
call_5_10_9_15, \
call_5_10_10_0, \
call_5_10_10_1, \
call_5_10_10_2, \
call_5_10_10_3, \
call_5_10_10_4, \
call_5_10_10_5, \
call_5_10_10_6, \
call_5_10_10_7, \
call_5_10_10_8, \
call_5_10_10_9, \
call_5_10_10_10, \
call_5_10_10_11, \
call_5_10_10_12, \
call_5_10_10_13, \
call_5_10_10_14, \
call_5_10_10_15, \
call_5_10_11_0, \
call_5_10_11_1, \
call_5_10_11_2, \
call_5_10_11_3, \
call_5_10_11_4, \
call_5_10_11_5, \
call_5_10_11_6, \
call_5_10_11_7, \
call_5_10_11_8, \
call_5_10_11_9, \
call_5_10_11_10, \
call_5_10_11_11, \
call_5_10_11_12, \
call_5_10_11_13, \
call_5_10_11_14, \
call_5_10_11_15, \
call_5_10_12_0, \
call_5_10_12_1, \
call_5_10_12_2, \
call_5_10_12_3, \
call_5_10_12_4, \
call_5_10_12_5, \
call_5_10_12_6, \
call_5_10_12_7, \
call_5_10_12_8, \
call_5_10_12_9, \
call_5_10_12_10, \
call_5_10_12_11, \
call_5_10_12_12, \
call_5_10_12_13, \
call_5_10_12_14, \
call_5_10_12_15, \
call_5_10_13_0, \
call_5_10_13_1, \
call_5_10_13_2, \
call_5_10_13_3, \
call_5_10_13_4, \
call_5_10_13_5, \
call_5_10_13_6, \
call_5_10_13_7, \
call_5_10_13_8, \
call_5_10_13_9, \
call_5_10_13_10, \
call_5_10_13_11, \
call_5_10_13_12, \
call_5_10_13_13, \
call_5_10_13_14, \
call_5_10_13_15, \
call_5_10_14_0, \
call_5_10_14_1, \
call_5_10_14_2, \
call_5_10_14_3, \
call_5_10_14_4, \
call_5_10_14_5, \
call_5_10_14_6, \
call_5_10_14_7, \
call_5_10_14_8, \
call_5_10_14_9, \
call_5_10_14_10, \
call_5_10_14_11, \
call_5_10_14_12, \
call_5_10_14_13, \
call_5_10_14_14, \
call_5_10_14_15, \
call_5_10_15_0, \
call_5_10_15_1, \
call_5_10_15_2, \
call_5_10_15_3, \
call_5_10_15_4, \
call_5_10_15_5, \
call_5_10_15_6, \
call_5_10_15_7, \
call_5_10_15_8, \
call_5_10_15_9, \
call_5_10_15_10, \
call_5_10_15_11, \
call_5_10_15_12, \
call_5_10_15_13, \
call_5_10_15_14, \
call_5_10_15_15, \
call_5_11_0_0, \
call_5_11_0_1, \
call_5_11_0_2, \
call_5_11_0_3, \
call_5_11_0_4, \
call_5_11_0_5, \
call_5_11_0_6, \
call_5_11_0_7, \
call_5_11_0_8, \
call_5_11_0_9, \
call_5_11_0_10, \
call_5_11_0_11, \
call_5_11_0_12, \
call_5_11_0_13, \
call_5_11_0_14, \
call_5_11_0_15, \
call_5_11_1_0, \
call_5_11_1_1, \
call_5_11_1_2, \
call_5_11_1_3, \
call_5_11_1_4, \
call_5_11_1_5, \
call_5_11_1_6, \
call_5_11_1_7, \
call_5_11_1_8, \
call_5_11_1_9, \
call_5_11_1_10, \
call_5_11_1_11, \
call_5_11_1_12, \
call_5_11_1_13, \
call_5_11_1_14, \
call_5_11_1_15, \
call_5_11_2_0, \
call_5_11_2_1, \
call_5_11_2_2, \
call_5_11_2_3, \
call_5_11_2_4, \
call_5_11_2_5, \
call_5_11_2_6, \
call_5_11_2_7, \
call_5_11_2_8, \
call_5_11_2_9, \
call_5_11_2_10, \
call_5_11_2_11, \
call_5_11_2_12, \
call_5_11_2_13, \
call_5_11_2_14, \
call_5_11_2_15, \
call_5_11_3_0, \
call_5_11_3_1, \
call_5_11_3_2, \
call_5_11_3_3, \
call_5_11_3_4, \
call_5_11_3_5, \
call_5_11_3_6, \
call_5_11_3_7, \
call_5_11_3_8, \
call_5_11_3_9, \
call_5_11_3_10, \
call_5_11_3_11, \
call_5_11_3_12, \
call_5_11_3_13, \
call_5_11_3_14, \
call_5_11_3_15, \
call_5_11_4_0, \
call_5_11_4_1, \
call_5_11_4_2, \
call_5_11_4_3, \
call_5_11_4_4, \
call_5_11_4_5, \
call_5_11_4_6, \
call_5_11_4_7, \
call_5_11_4_8, \
call_5_11_4_9, \
call_5_11_4_10, \
call_5_11_4_11, \
call_5_11_4_12, \
call_5_11_4_13, \
call_5_11_4_14, \
call_5_11_4_15, \
call_5_11_5_0, \
call_5_11_5_1, \
call_5_11_5_2, \
call_5_11_5_3, \
call_5_11_5_4, \
call_5_11_5_5, \
call_5_11_5_6, \
call_5_11_5_7, \
call_5_11_5_8, \
call_5_11_5_9, \
call_5_11_5_10, \
call_5_11_5_11, \
call_5_11_5_12, \
call_5_11_5_13, \
call_5_11_5_14, \
call_5_11_5_15, \
call_5_11_6_0, \
call_5_11_6_1, \
call_5_11_6_2, \
call_5_11_6_3, \
call_5_11_6_4, \
call_5_11_6_5, \
call_5_11_6_6, \
call_5_11_6_7, \
call_5_11_6_8, \
call_5_11_6_9, \
call_5_11_6_10, \
call_5_11_6_11, \
call_5_11_6_12, \
call_5_11_6_13, \
call_5_11_6_14, \
call_5_11_6_15, \
call_5_11_7_0, \
call_5_11_7_1, \
call_5_11_7_2, \
call_5_11_7_3, \
call_5_11_7_4, \
call_5_11_7_5, \
call_5_11_7_6, \
call_5_11_7_7, \
call_5_11_7_8, \
call_5_11_7_9, \
call_5_11_7_10, \
call_5_11_7_11, \
call_5_11_7_12, \
call_5_11_7_13, \
call_5_11_7_14, \
call_5_11_7_15, \
call_5_11_8_0, \
call_5_11_8_1, \
call_5_11_8_2, \
call_5_11_8_3, \
call_5_11_8_4, \
call_5_11_8_5, \
call_5_11_8_6, \
call_5_11_8_7, \
call_5_11_8_8, \
call_5_11_8_9, \
call_5_11_8_10, \
call_5_11_8_11, \
call_5_11_8_12, \
call_5_11_8_13, \
call_5_11_8_14, \
call_5_11_8_15, \
call_5_11_9_0, \
call_5_11_9_1, \
call_5_11_9_2, \
call_5_11_9_3, \
call_5_11_9_4, \
call_5_11_9_5, \
call_5_11_9_6, \
call_5_11_9_7, \
call_5_11_9_8, \
call_5_11_9_9, \
call_5_11_9_10, \
call_5_11_9_11, \
call_5_11_9_12, \
call_5_11_9_13, \
call_5_11_9_14, \
call_5_11_9_15, \
call_5_11_10_0, \
call_5_11_10_1, \
call_5_11_10_2, \
call_5_11_10_3, \
call_5_11_10_4, \
call_5_11_10_5, \
call_5_11_10_6, \
call_5_11_10_7, \
call_5_11_10_8, \
call_5_11_10_9, \
call_5_11_10_10, \
call_5_11_10_11, \
call_5_11_10_12, \
call_5_11_10_13, \
call_5_11_10_14, \
call_5_11_10_15, \
call_5_11_11_0, \
call_5_11_11_1, \
call_5_11_11_2, \
call_5_11_11_3, \
call_5_11_11_4, \
call_5_11_11_5, \
call_5_11_11_6, \
call_5_11_11_7, \
call_5_11_11_8, \
call_5_11_11_9, \
call_5_11_11_10, \
call_5_11_11_11, \
call_5_11_11_12, \
call_5_11_11_13, \
call_5_11_11_14, \
call_5_11_11_15, \
call_5_11_12_0, \
call_5_11_12_1, \
call_5_11_12_2, \
call_5_11_12_3, \
call_5_11_12_4, \
call_5_11_12_5, \
call_5_11_12_6, \
call_5_11_12_7, \
call_5_11_12_8, \
call_5_11_12_9, \
call_5_11_12_10, \
call_5_11_12_11, \
call_5_11_12_12, \
call_5_11_12_13, \
call_5_11_12_14, \
call_5_11_12_15, \
call_5_11_13_0, \
call_5_11_13_1, \
call_5_11_13_2, \
call_5_11_13_3, \
call_5_11_13_4, \
call_5_11_13_5, \
call_5_11_13_6, \
call_5_11_13_7, \
call_5_11_13_8, \
call_5_11_13_9, \
call_5_11_13_10, \
call_5_11_13_11, \
call_5_11_13_12, \
call_5_11_13_13, \
call_5_11_13_14, \
call_5_11_13_15, \
call_5_11_14_0, \
call_5_11_14_1, \
call_5_11_14_2, \
call_5_11_14_3, \
call_5_11_14_4, \
call_5_11_14_5, \
call_5_11_14_6, \
call_5_11_14_7, \
call_5_11_14_8, \
call_5_11_14_9, \
call_5_11_14_10, \
call_5_11_14_11, \
call_5_11_14_12, \
call_5_11_14_13, \
call_5_11_14_14, \
call_5_11_14_15, \
call_5_11_15_0, \
call_5_11_15_1, \
call_5_11_15_2, \
call_5_11_15_3, \
call_5_11_15_4, \
call_5_11_15_5, \
call_5_11_15_6, \
call_5_11_15_7, \
call_5_11_15_8, \
call_5_11_15_9, \
call_5_11_15_10, \
call_5_11_15_11, \
call_5_11_15_12, \
call_5_11_15_13, \
call_5_11_15_14, \
call_5_11_15_15, \
call_5_12_0_0, \
call_5_12_0_1, \
call_5_12_0_2, \
call_5_12_0_3, \
call_5_12_0_4, \
call_5_12_0_5, \
call_5_12_0_6, \
call_5_12_0_7, \
call_5_12_0_8, \
call_5_12_0_9, \
call_5_12_0_10, \
call_5_12_0_11, \
call_5_12_0_12, \
call_5_12_0_13, \
call_5_12_0_14, \
call_5_12_0_15, \
call_5_12_1_0, \
call_5_12_1_1, \
call_5_12_1_2, \
call_5_12_1_3, \
call_5_12_1_4, \
call_5_12_1_5, \
call_5_12_1_6, \
call_5_12_1_7, \
call_5_12_1_8, \
call_5_12_1_9, \
call_5_12_1_10, \
call_5_12_1_11, \
call_5_12_1_12, \
call_5_12_1_13, \
call_5_12_1_14, \
call_5_12_1_15, \
call_5_12_2_0, \
call_5_12_2_1, \
call_5_12_2_2, \
call_5_12_2_3, \
call_5_12_2_4, \
call_5_12_2_5, \
call_5_12_2_6, \
call_5_12_2_7, \
call_5_12_2_8, \
call_5_12_2_9, \
call_5_12_2_10, \
call_5_12_2_11, \
call_5_12_2_12, \
call_5_12_2_13, \
call_5_12_2_14, \
call_5_12_2_15, \
call_5_12_3_0, \
call_5_12_3_1, \
call_5_12_3_2, \
call_5_12_3_3, \
call_5_12_3_4, \
call_5_12_3_5, \
call_5_12_3_6, \
call_5_12_3_7, \
call_5_12_3_8, \
call_5_12_3_9, \
call_5_12_3_10, \
call_5_12_3_11, \
call_5_12_3_12, \
call_5_12_3_13, \
call_5_12_3_14, \
call_5_12_3_15, \
call_5_12_4_0, \
call_5_12_4_1, \
call_5_12_4_2, \
call_5_12_4_3, \
call_5_12_4_4, \
call_5_12_4_5, \
call_5_12_4_6, \
call_5_12_4_7, \
call_5_12_4_8, \
call_5_12_4_9, \
call_5_12_4_10, \
call_5_12_4_11, \
call_5_12_4_12, \
call_5_12_4_13, \
call_5_12_4_14, \
call_5_12_4_15, \
call_5_12_5_0, \
call_5_12_5_1, \
call_5_12_5_2, \
call_5_12_5_3, \
call_5_12_5_4, \
call_5_12_5_5, \
call_5_12_5_6, \
call_5_12_5_7, \
call_5_12_5_8, \
call_5_12_5_9, \
call_5_12_5_10, \
call_5_12_5_11, \
call_5_12_5_12, \
call_5_12_5_13, \
call_5_12_5_14, \
call_5_12_5_15, \
call_5_12_6_0, \
call_5_12_6_1, \
call_5_12_6_2, \
call_5_12_6_3, \
call_5_12_6_4, \
call_5_12_6_5, \
call_5_12_6_6, \
call_5_12_6_7, \
call_5_12_6_8, \
call_5_12_6_9, \
call_5_12_6_10, \
call_5_12_6_11, \
call_5_12_6_12, \
call_5_12_6_13, \
call_5_12_6_14, \
call_5_12_6_15, \
call_5_12_7_0, \
call_5_12_7_1, \
call_5_12_7_2, \
call_5_12_7_3, \
call_5_12_7_4, \
call_5_12_7_5, \
call_5_12_7_6, \
call_5_12_7_7, \
call_5_12_7_8, \
call_5_12_7_9, \
call_5_12_7_10, \
call_5_12_7_11, \
call_5_12_7_12, \
call_5_12_7_13, \
call_5_12_7_14, \
call_5_12_7_15, \
call_5_12_8_0, \
call_5_12_8_1, \
call_5_12_8_2, \
call_5_12_8_3, \
call_5_12_8_4, \
call_5_12_8_5, \
call_5_12_8_6, \
call_5_12_8_7, \
call_5_12_8_8, \
call_5_12_8_9, \
call_5_12_8_10, \
call_5_12_8_11, \
call_5_12_8_12, \
call_5_12_8_13, \
call_5_12_8_14, \
call_5_12_8_15, \
call_5_12_9_0, \
call_5_12_9_1, \
call_5_12_9_2, \
call_5_12_9_3, \
call_5_12_9_4, \
call_5_12_9_5, \
call_5_12_9_6, \
call_5_12_9_7, \
call_5_12_9_8, \
call_5_12_9_9, \
call_5_12_9_10, \
call_5_12_9_11, \
call_5_12_9_12, \
call_5_12_9_13, \
call_5_12_9_14, \
call_5_12_9_15, \
call_5_12_10_0, \
call_5_12_10_1, \
call_5_12_10_2, \
call_5_12_10_3, \
call_5_12_10_4, \
call_5_12_10_5, \
call_5_12_10_6, \
call_5_12_10_7, \
call_5_12_10_8, \
call_5_12_10_9, \
call_5_12_10_10, \
call_5_12_10_11, \
call_5_12_10_12, \
call_5_12_10_13, \
call_5_12_10_14, \
call_5_12_10_15, \
call_5_12_11_0, \
call_5_12_11_1, \
call_5_12_11_2, \
call_5_12_11_3, \
call_5_12_11_4, \
call_5_12_11_5, \
call_5_12_11_6, \
call_5_12_11_7, \
call_5_12_11_8, \
call_5_12_11_9, \
call_5_12_11_10, \
call_5_12_11_11, \
call_5_12_11_12, \
call_5_12_11_13, \
call_5_12_11_14, \
call_5_12_11_15, \
call_5_12_12_0, \
call_5_12_12_1, \
call_5_12_12_2, \
call_5_12_12_3, \
call_5_12_12_4, \
call_5_12_12_5, \
call_5_12_12_6, \
call_5_12_12_7, \
call_5_12_12_8, \
call_5_12_12_9, \
call_5_12_12_10, \
call_5_12_12_11, \
call_5_12_12_12, \
call_5_12_12_13, \
call_5_12_12_14, \
call_5_12_12_15, \
call_5_12_13_0, \
call_5_12_13_1, \
call_5_12_13_2, \
call_5_12_13_3, \
call_5_12_13_4, \
call_5_12_13_5, \
call_5_12_13_6, \
call_5_12_13_7, \
call_5_12_13_8, \
call_5_12_13_9, \
call_5_12_13_10, \
call_5_12_13_11, \
call_5_12_13_12, \
call_5_12_13_13, \
call_5_12_13_14, \
call_5_12_13_15, \
call_5_12_14_0, \
call_5_12_14_1, \
call_5_12_14_2, \
call_5_12_14_3, \
call_5_12_14_4, \
call_5_12_14_5, \
call_5_12_14_6, \
call_5_12_14_7, \
call_5_12_14_8, \
call_5_12_14_9, \
call_5_12_14_10, \
call_5_12_14_11, \
call_5_12_14_12, \
call_5_12_14_13, \
call_5_12_14_14, \
call_5_12_14_15, \
call_5_12_15_0, \
call_5_12_15_1, \
call_5_12_15_2, \
call_5_12_15_3, \
call_5_12_15_4, \
call_5_12_15_5, \
call_5_12_15_6, \
call_5_12_15_7, \
call_5_12_15_8, \
call_5_12_15_9, \
call_5_12_15_10, \
call_5_12_15_11, \
call_5_12_15_12, \
call_5_12_15_13, \
call_5_12_15_14, \
call_5_12_15_15, \
call_5_13_0_0, \
call_5_13_0_1, \
call_5_13_0_2, \
call_5_13_0_3, \
call_5_13_0_4, \
call_5_13_0_5, \
call_5_13_0_6, \
call_5_13_0_7, \
call_5_13_0_8, \
call_5_13_0_9, \
call_5_13_0_10, \
call_5_13_0_11, \
call_5_13_0_12, \
call_5_13_0_13, \
call_5_13_0_14, \
call_5_13_0_15, \
call_5_13_1_0, \
call_5_13_1_1, \
call_5_13_1_2, \
call_5_13_1_3, \
call_5_13_1_4, \
call_5_13_1_5, \
call_5_13_1_6, \
call_5_13_1_7, \
call_5_13_1_8, \
call_5_13_1_9, \
call_5_13_1_10, \
call_5_13_1_11, \
call_5_13_1_12, \
call_5_13_1_13, \
call_5_13_1_14, \
call_5_13_1_15, \
call_5_13_2_0, \
call_5_13_2_1, \
call_5_13_2_2, \
call_5_13_2_3, \
call_5_13_2_4, \
call_5_13_2_5, \
call_5_13_2_6, \
call_5_13_2_7, \
call_5_13_2_8, \
call_5_13_2_9, \
call_5_13_2_10, \
call_5_13_2_11, \
call_5_13_2_12, \
call_5_13_2_13, \
call_5_13_2_14, \
call_5_13_2_15, \
call_5_13_3_0, \
call_5_13_3_1, \
call_5_13_3_2, \
call_5_13_3_3, \
call_5_13_3_4, \
call_5_13_3_5, \
call_5_13_3_6, \
call_5_13_3_7, \
call_5_13_3_8, \
call_5_13_3_9, \
call_5_13_3_10, \
call_5_13_3_11, \
call_5_13_3_12, \
call_5_13_3_13, \
call_5_13_3_14, \
call_5_13_3_15, \
call_5_13_4_0, \
call_5_13_4_1, \
call_5_13_4_2, \
call_5_13_4_3, \
call_5_13_4_4, \
call_5_13_4_5, \
call_5_13_4_6, \
call_5_13_4_7, \
call_5_13_4_8, \
call_5_13_4_9, \
call_5_13_4_10, \
call_5_13_4_11, \
call_5_13_4_12, \
call_5_13_4_13, \
call_5_13_4_14, \
call_5_13_4_15, \
call_5_13_5_0, \
call_5_13_5_1, \
call_5_13_5_2, \
call_5_13_5_3, \
call_5_13_5_4, \
call_5_13_5_5, \
call_5_13_5_6, \
call_5_13_5_7, \
call_5_13_5_8, \
call_5_13_5_9, \
call_5_13_5_10, \
call_5_13_5_11, \
call_5_13_5_12, \
call_5_13_5_13, \
call_5_13_5_14, \
call_5_13_5_15, \
call_5_13_6_0, \
call_5_13_6_1, \
call_5_13_6_2, \
call_5_13_6_3, \
call_5_13_6_4, \
call_5_13_6_5, \
call_5_13_6_6, \
call_5_13_6_7, \
call_5_13_6_8, \
call_5_13_6_9, \
call_5_13_6_10, \
call_5_13_6_11, \
call_5_13_6_12, \
call_5_13_6_13, \
call_5_13_6_14, \
call_5_13_6_15, \
call_5_13_7_0, \
call_5_13_7_1, \
call_5_13_7_2, \
call_5_13_7_3, \
call_5_13_7_4, \
call_5_13_7_5, \
call_5_13_7_6, \
call_5_13_7_7, \
call_5_13_7_8, \
call_5_13_7_9, \
call_5_13_7_10, \
call_5_13_7_11, \
call_5_13_7_12, \
call_5_13_7_13, \
call_5_13_7_14, \
call_5_13_7_15, \
call_5_13_8_0, \
call_5_13_8_1, \
call_5_13_8_2, \
call_5_13_8_3, \
call_5_13_8_4, \
call_5_13_8_5, \
call_5_13_8_6, \
call_5_13_8_7, \
call_5_13_8_8, \
call_5_13_8_9, \
call_5_13_8_10, \
call_5_13_8_11, \
call_5_13_8_12, \
call_5_13_8_13, \
call_5_13_8_14, \
call_5_13_8_15, \
call_5_13_9_0, \
call_5_13_9_1, \
call_5_13_9_2, \
call_5_13_9_3, \
call_5_13_9_4, \
call_5_13_9_5, \
call_5_13_9_6, \
call_5_13_9_7, \
call_5_13_9_8, \
call_5_13_9_9, \
call_5_13_9_10, \
call_5_13_9_11, \
call_5_13_9_12, \
call_5_13_9_13, \
call_5_13_9_14, \
call_5_13_9_15, \
call_5_13_10_0, \
call_5_13_10_1, \
call_5_13_10_2, \
call_5_13_10_3, \
call_5_13_10_4, \
call_5_13_10_5, \
call_5_13_10_6, \
call_5_13_10_7, \
call_5_13_10_8, \
call_5_13_10_9, \
call_5_13_10_10, \
call_5_13_10_11, \
call_5_13_10_12, \
call_5_13_10_13, \
call_5_13_10_14, \
call_5_13_10_15, \
call_5_13_11_0, \
call_5_13_11_1, \
call_5_13_11_2, \
call_5_13_11_3, \
call_5_13_11_4, \
call_5_13_11_5, \
call_5_13_11_6, \
call_5_13_11_7, \
call_5_13_11_8, \
call_5_13_11_9, \
call_5_13_11_10, \
call_5_13_11_11, \
call_5_13_11_12, \
call_5_13_11_13, \
call_5_13_11_14, \
call_5_13_11_15, \
call_5_13_12_0, \
call_5_13_12_1, \
call_5_13_12_2, \
call_5_13_12_3, \
call_5_13_12_4, \
call_5_13_12_5, \
call_5_13_12_6, \
call_5_13_12_7, \
call_5_13_12_8, \
call_5_13_12_9, \
call_5_13_12_10, \
call_5_13_12_11, \
call_5_13_12_12, \
call_5_13_12_13, \
call_5_13_12_14, \
call_5_13_12_15, \
call_5_13_13_0, \
call_5_13_13_1, \
call_5_13_13_2, \
call_5_13_13_3, \
call_5_13_13_4, \
call_5_13_13_5, \
call_5_13_13_6, \
call_5_13_13_7, \
call_5_13_13_8, \
call_5_13_13_9, \
call_5_13_13_10, \
call_5_13_13_11, \
call_5_13_13_12, \
call_5_13_13_13, \
call_5_13_13_14, \
call_5_13_13_15, \
call_5_13_14_0, \
call_5_13_14_1, \
call_5_13_14_2, \
call_5_13_14_3, \
call_5_13_14_4, \
call_5_13_14_5, \
call_5_13_14_6, \
call_5_13_14_7, \
call_5_13_14_8, \
call_5_13_14_9, \
call_5_13_14_10, \
call_5_13_14_11, \
call_5_13_14_12, \
call_5_13_14_13, \
call_5_13_14_14, \
call_5_13_14_15, \
call_5_13_15_0, \
call_5_13_15_1, \
call_5_13_15_2, \
call_5_13_15_3, \
call_5_13_15_4, \
call_5_13_15_5, \
call_5_13_15_6, \
call_5_13_15_7, \
call_5_13_15_8, \
call_5_13_15_9, \
call_5_13_15_10, \
call_5_13_15_11, \
call_5_13_15_12, \
call_5_13_15_13, \
call_5_13_15_14, \
call_5_13_15_15, \
call_5_14_0_0, \
call_5_14_0_1, \
call_5_14_0_2, \
call_5_14_0_3, \
call_5_14_0_4, \
call_5_14_0_5, \
call_5_14_0_6, \
call_5_14_0_7, \
call_5_14_0_8, \
call_5_14_0_9, \
call_5_14_0_10, \
call_5_14_0_11, \
call_5_14_0_12, \
call_5_14_0_13, \
call_5_14_0_14, \
call_5_14_0_15, \
call_5_14_1_0, \
call_5_14_1_1, \
call_5_14_1_2, \
call_5_14_1_3, \
call_5_14_1_4, \
call_5_14_1_5, \
call_5_14_1_6, \
call_5_14_1_7, \
call_5_14_1_8, \
call_5_14_1_9, \
call_5_14_1_10, \
call_5_14_1_11, \
call_5_14_1_12, \
call_5_14_1_13, \
call_5_14_1_14, \
call_5_14_1_15, \
call_5_14_2_0, \
call_5_14_2_1, \
call_5_14_2_2, \
call_5_14_2_3, \
call_5_14_2_4, \
call_5_14_2_5, \
call_5_14_2_6, \
call_5_14_2_7, \
call_5_14_2_8, \
call_5_14_2_9, \
call_5_14_2_10, \
call_5_14_2_11, \
call_5_14_2_12, \
call_5_14_2_13, \
call_5_14_2_14, \
call_5_14_2_15, \
call_5_14_3_0, \
call_5_14_3_1, \
call_5_14_3_2, \
call_5_14_3_3, \
call_5_14_3_4, \
call_5_14_3_5, \
call_5_14_3_6, \
call_5_14_3_7, \
call_5_14_3_8, \
call_5_14_3_9, \
call_5_14_3_10, \
call_5_14_3_11, \
call_5_14_3_12, \
call_5_14_3_13, \
call_5_14_3_14, \
call_5_14_3_15, \
call_5_14_4_0, \
call_5_14_4_1, \
call_5_14_4_2, \
call_5_14_4_3, \
call_5_14_4_4, \
call_5_14_4_5, \
call_5_14_4_6, \
call_5_14_4_7, \
call_5_14_4_8, \
call_5_14_4_9, \
call_5_14_4_10, \
call_5_14_4_11, \
call_5_14_4_12, \
call_5_14_4_13, \
call_5_14_4_14, \
call_5_14_4_15, \
call_5_14_5_0, \
call_5_14_5_1, \
call_5_14_5_2, \
call_5_14_5_3, \
call_5_14_5_4, \
call_5_14_5_5, \
call_5_14_5_6, \
call_5_14_5_7, \
call_5_14_5_8, \
call_5_14_5_9, \
call_5_14_5_10, \
call_5_14_5_11, \
call_5_14_5_12, \
call_5_14_5_13, \
call_5_14_5_14, \
call_5_14_5_15, \
call_5_14_6_0, \
call_5_14_6_1, \
call_5_14_6_2, \
call_5_14_6_3, \
call_5_14_6_4, \
call_5_14_6_5, \
call_5_14_6_6, \
call_5_14_6_7, \
call_5_14_6_8, \
call_5_14_6_9, \
call_5_14_6_10, \
call_5_14_6_11, \
call_5_14_6_12, \
call_5_14_6_13, \
call_5_14_6_14, \
call_5_14_6_15, \
call_5_14_7_0, \
call_5_14_7_1, \
call_5_14_7_2, \
call_5_14_7_3, \
call_5_14_7_4, \
call_5_14_7_5, \
call_5_14_7_6, \
call_5_14_7_7, \
call_5_14_7_8, \
call_5_14_7_9, \
call_5_14_7_10, \
call_5_14_7_11, \
call_5_14_7_12, \
call_5_14_7_13, \
call_5_14_7_14, \
call_5_14_7_15, \
call_5_14_8_0, \
call_5_14_8_1, \
call_5_14_8_2, \
call_5_14_8_3, \
call_5_14_8_4, \
call_5_14_8_5, \
call_5_14_8_6, \
call_5_14_8_7, \
call_5_14_8_8, \
call_5_14_8_9, \
call_5_14_8_10, \
call_5_14_8_11, \
call_5_14_8_12, \
call_5_14_8_13, \
call_5_14_8_14, \
call_5_14_8_15, \
call_5_14_9_0, \
call_5_14_9_1, \
call_5_14_9_2, \
call_5_14_9_3, \
call_5_14_9_4, \
call_5_14_9_5, \
call_5_14_9_6, \
call_5_14_9_7, \
call_5_14_9_8, \
call_5_14_9_9, \
call_5_14_9_10, \
call_5_14_9_11, \
call_5_14_9_12, \
call_5_14_9_13, \
call_5_14_9_14, \
call_5_14_9_15, \
call_5_14_10_0, \
call_5_14_10_1, \
call_5_14_10_2, \
call_5_14_10_3, \
call_5_14_10_4, \
call_5_14_10_5, \
call_5_14_10_6, \
call_5_14_10_7, \
call_5_14_10_8, \
call_5_14_10_9, \
call_5_14_10_10, \
call_5_14_10_11, \
call_5_14_10_12, \
call_5_14_10_13, \
call_5_14_10_14, \
call_5_14_10_15, \
call_5_14_11_0, \
call_5_14_11_1, \
call_5_14_11_2, \
call_5_14_11_3, \
call_5_14_11_4, \
call_5_14_11_5, \
call_5_14_11_6, \
call_5_14_11_7, \
call_5_14_11_8, \
call_5_14_11_9, \
call_5_14_11_10, \
call_5_14_11_11, \
call_5_14_11_12, \
call_5_14_11_13, \
call_5_14_11_14, \
call_5_14_11_15, \
call_5_14_12_0, \
call_5_14_12_1, \
call_5_14_12_2, \
call_5_14_12_3, \
call_5_14_12_4, \
call_5_14_12_5, \
call_5_14_12_6, \
call_5_14_12_7, \
call_5_14_12_8, \
call_5_14_12_9, \
call_5_14_12_10, \
call_5_14_12_11, \
call_5_14_12_12, \
call_5_14_12_13, \
call_5_14_12_14, \
call_5_14_12_15, \
call_5_14_13_0, \
call_5_14_13_1, \
call_5_14_13_2, \
call_5_14_13_3, \
call_5_14_13_4, \
call_5_14_13_5, \
call_5_14_13_6, \
call_5_14_13_7, \
call_5_14_13_8, \
call_5_14_13_9, \
call_5_14_13_10, \
call_5_14_13_11, \
call_5_14_13_12, \
call_5_14_13_13, \
call_5_14_13_14, \
call_5_14_13_15, \
call_5_14_14_0, \
call_5_14_14_1, \
call_5_14_14_2, \
call_5_14_14_3, \
call_5_14_14_4, \
call_5_14_14_5, \
call_5_14_14_6, \
call_5_14_14_7, \
call_5_14_14_8, \
call_5_14_14_9, \
call_5_14_14_10, \
call_5_14_14_11, \
call_5_14_14_12, \
call_5_14_14_13, \
call_5_14_14_14, \
call_5_14_14_15, \
call_5_14_15_0, \
call_5_14_15_1, \
call_5_14_15_2, \
call_5_14_15_3, \
call_5_14_15_4, \
call_5_14_15_5, \
call_5_14_15_6, \
call_5_14_15_7, \
call_5_14_15_8, \
call_5_14_15_9, \
call_5_14_15_10, \
call_5_14_15_11, \
call_5_14_15_12, \
call_5_14_15_13, \
call_5_14_15_14, \
call_5_14_15_15, \
call_5_15_0_0, \
call_5_15_0_1, \
call_5_15_0_2, \
call_5_15_0_3, \
call_5_15_0_4, \
call_5_15_0_5, \
call_5_15_0_6, \
call_5_15_0_7, \
call_5_15_0_8, \
call_5_15_0_9, \
call_5_15_0_10, \
call_5_15_0_11, \
call_5_15_0_12, \
call_5_15_0_13, \
call_5_15_0_14, \
call_5_15_0_15, \
call_5_15_1_0, \
call_5_15_1_1, \
call_5_15_1_2, \
call_5_15_1_3, \
call_5_15_1_4, \
call_5_15_1_5, \
call_5_15_1_6, \
call_5_15_1_7, \
call_5_15_1_8, \
call_5_15_1_9, \
call_5_15_1_10, \
call_5_15_1_11, \
call_5_15_1_12, \
call_5_15_1_13, \
call_5_15_1_14, \
call_5_15_1_15, \
call_5_15_2_0, \
call_5_15_2_1, \
call_5_15_2_2, \
call_5_15_2_3, \
call_5_15_2_4, \
call_5_15_2_5, \
call_5_15_2_6, \
call_5_15_2_7, \
call_5_15_2_8, \
call_5_15_2_9, \
call_5_15_2_10, \
call_5_15_2_11, \
call_5_15_2_12, \
call_5_15_2_13, \
call_5_15_2_14, \
call_5_15_2_15, \
call_5_15_3_0, \
call_5_15_3_1, \
call_5_15_3_2, \
call_5_15_3_3, \
call_5_15_3_4, \
call_5_15_3_5, \
call_5_15_3_6, \
call_5_15_3_7, \
call_5_15_3_8, \
call_5_15_3_9, \
call_5_15_3_10, \
call_5_15_3_11, \
call_5_15_3_12, \
call_5_15_3_13, \
call_5_15_3_14, \
call_5_15_3_15, \
call_5_15_4_0, \
call_5_15_4_1, \
call_5_15_4_2, \
call_5_15_4_3, \
call_5_15_4_4, \
call_5_15_4_5, \
call_5_15_4_6, \
call_5_15_4_7, \
call_5_15_4_8, \
call_5_15_4_9, \
call_5_15_4_10, \
call_5_15_4_11, \
call_5_15_4_12, \
call_5_15_4_13, \
call_5_15_4_14, \
call_5_15_4_15, \
call_5_15_5_0, \
call_5_15_5_1, \
call_5_15_5_2, \
call_5_15_5_3, \
call_5_15_5_4, \
call_5_15_5_5, \
call_5_15_5_6, \
call_5_15_5_7, \
call_5_15_5_8, \
call_5_15_5_9, \
call_5_15_5_10, \
call_5_15_5_11, \
call_5_15_5_12, \
call_5_15_5_13, \
call_5_15_5_14, \
call_5_15_5_15, \
call_5_15_6_0, \
call_5_15_6_1, \
call_5_15_6_2, \
call_5_15_6_3, \
call_5_15_6_4, \
call_5_15_6_5, \
call_5_15_6_6, \
call_5_15_6_7, \
call_5_15_6_8, \
call_5_15_6_9, \
call_5_15_6_10, \
call_5_15_6_11, \
call_5_15_6_12, \
call_5_15_6_13, \
call_5_15_6_14, \
call_5_15_6_15, \
call_5_15_7_0, \
call_5_15_7_1, \
call_5_15_7_2, \
call_5_15_7_3, \
call_5_15_7_4, \
call_5_15_7_5, \
call_5_15_7_6, \
call_5_15_7_7, \
call_5_15_7_8, \
call_5_15_7_9, \
call_5_15_7_10, \
call_5_15_7_11, \
call_5_15_7_12, \
call_5_15_7_13, \
call_5_15_7_14, \
call_5_15_7_15, \
call_5_15_8_0, \
call_5_15_8_1, \
call_5_15_8_2, \
call_5_15_8_3, \
call_5_15_8_4, \
call_5_15_8_5, \
call_5_15_8_6, \
call_5_15_8_7, \
call_5_15_8_8, \
call_5_15_8_9, \
call_5_15_8_10, \
call_5_15_8_11, \
call_5_15_8_12, \
call_5_15_8_13, \
call_5_15_8_14, \
call_5_15_8_15, \
call_5_15_9_0, \
call_5_15_9_1, \
call_5_15_9_2, \
call_5_15_9_3, \
call_5_15_9_4, \
call_5_15_9_5, \
call_5_15_9_6, \
call_5_15_9_7, \
call_5_15_9_8, \
call_5_15_9_9, \
call_5_15_9_10, \
call_5_15_9_11, \
call_5_15_9_12, \
call_5_15_9_13, \
call_5_15_9_14, \
call_5_15_9_15, \
call_5_15_10_0, \
call_5_15_10_1, \
call_5_15_10_2, \
call_5_15_10_3, \
call_5_15_10_4, \
call_5_15_10_5, \
call_5_15_10_6, \
call_5_15_10_7, \
call_5_15_10_8, \
call_5_15_10_9, \
call_5_15_10_10, \
call_5_15_10_11, \
call_5_15_10_12, \
call_5_15_10_13, \
call_5_15_10_14, \
call_5_15_10_15, \
call_5_15_11_0, \
call_5_15_11_1, \
call_5_15_11_2, \
call_5_15_11_3, \
call_5_15_11_4, \
call_5_15_11_5, \
call_5_15_11_6, \
call_5_15_11_7, \
call_5_15_11_8, \
call_5_15_11_9, \
call_5_15_11_10, \
call_5_15_11_11, \
call_5_15_11_12, \
call_5_15_11_13, \
call_5_15_11_14, \
call_5_15_11_15, \
call_5_15_12_0, \
call_5_15_12_1, \
call_5_15_12_2, \
call_5_15_12_3, \
call_5_15_12_4, \
call_5_15_12_5, \
call_5_15_12_6, \
call_5_15_12_7, \
call_5_15_12_8, \
call_5_15_12_9, \
call_5_15_12_10, \
call_5_15_12_11, \
call_5_15_12_12, \
call_5_15_12_13, \
call_5_15_12_14, \
call_5_15_12_15, \
call_5_15_13_0, \
call_5_15_13_1, \
call_5_15_13_2, \
call_5_15_13_3, \
call_5_15_13_4, \
call_5_15_13_5, \
call_5_15_13_6, \
call_5_15_13_7, \
call_5_15_13_8, \
call_5_15_13_9, \
call_5_15_13_10, \
call_5_15_13_11, \
call_5_15_13_12, \
call_5_15_13_13, \
call_5_15_13_14, \
call_5_15_13_15, \
call_5_15_14_0, \
call_5_15_14_1, \
call_5_15_14_2, \
call_5_15_14_3, \
call_5_15_14_4, \
call_5_15_14_5, \
call_5_15_14_6, \
call_5_15_14_7, \
call_5_15_14_8, \
call_5_15_14_9, \
call_5_15_14_10, \
call_5_15_14_11, \
call_5_15_14_12, \
call_5_15_14_13, \
call_5_15_14_14, \
call_5_15_14_15, \
call_5_15_15_0, \
call_5_15_15_1, \
call_5_15_15_2, \
call_5_15_15_3, \
call_5_15_15_4, \
call_5_15_15_5, \
call_5_15_15_6, \
call_5_15_15_7, \
call_5_15_15_8, \
call_5_15_15_9, \
call_5_15_15_10, \
call_5_15_15_11, \
call_5_15_15_12, \
call_5_15_15_13, \
call_5_15_15_14, \
call_5_15_15_15, \
call_6_0_0_0, \
call_6_0_0_1, \
call_6_0_0_2, \
call_6_0_0_3, \
call_6_0_0_4, \
call_6_0_0_5, \
call_6_0_0_6, \
call_6_0_0_7, \
call_6_0_0_8, \
call_6_0_0_9, \
call_6_0_0_10, \
call_6_0_0_11, \
call_6_0_0_12, \
call_6_0_0_13, \
call_6_0_0_14, \
call_6_0_0_15, \
call_6_0_1_0, \
call_6_0_1_1, \
call_6_0_1_2, \
call_6_0_1_3, \
call_6_0_1_4, \
call_6_0_1_5, \
call_6_0_1_6, \
call_6_0_1_7, \
call_6_0_1_8, \
call_6_0_1_9, \
call_6_0_1_10, \
call_6_0_1_11, \
call_6_0_1_12, \
call_6_0_1_13, \
call_6_0_1_14, \
call_6_0_1_15, \
call_6_0_2_0, \
call_6_0_2_1, \
call_6_0_2_2, \
call_6_0_2_3, \
call_6_0_2_4, \
call_6_0_2_5, \
call_6_0_2_6, \
call_6_0_2_7, \
call_6_0_2_8, \
call_6_0_2_9, \
call_6_0_2_10, \
call_6_0_2_11, \
call_6_0_2_12, \
call_6_0_2_13, \
call_6_0_2_14, \
call_6_0_2_15, \
call_6_0_3_0, \
call_6_0_3_1, \
call_6_0_3_2, \
call_6_0_3_3, \
call_6_0_3_4, \
call_6_0_3_5, \
call_6_0_3_6, \
call_6_0_3_7, \
call_6_0_3_8, \
call_6_0_3_9, \
call_6_0_3_10, \
call_6_0_3_11, \
call_6_0_3_12, \
call_6_0_3_13, \
call_6_0_3_14, \
call_6_0_3_15, \
call_6_0_4_0, \
call_6_0_4_1, \
call_6_0_4_2, \
call_6_0_4_3, \
call_6_0_4_4, \
call_6_0_4_5, \
call_6_0_4_6, \
call_6_0_4_7, \
call_6_0_4_8, \
call_6_0_4_9, \
call_6_0_4_10, \
call_6_0_4_11, \
call_6_0_4_12, \
call_6_0_4_13, \
call_6_0_4_14, \
call_6_0_4_15, \
call_6_0_5_0, \
call_6_0_5_1, \
call_6_0_5_2, \
call_6_0_5_3, \
call_6_0_5_4, \
call_6_0_5_5, \
call_6_0_5_6, \
call_6_0_5_7, \
call_6_0_5_8, \
call_6_0_5_9, \
call_6_0_5_10, \
call_6_0_5_11, \
call_6_0_5_12, \
call_6_0_5_13, \
call_6_0_5_14, \
call_6_0_5_15, \
call_6_0_6_0, \
call_6_0_6_1, \
call_6_0_6_2, \
call_6_0_6_3, \
call_6_0_6_4, \
call_6_0_6_5, \
call_6_0_6_6, \
call_6_0_6_7, \
call_6_0_6_8, \
call_6_0_6_9, \
call_6_0_6_10, \
call_6_0_6_11, \
call_6_0_6_12, \
call_6_0_6_13, \
call_6_0_6_14, \
call_6_0_6_15, \
call_6_0_7_0, \
call_6_0_7_1, \
call_6_0_7_2, \
call_6_0_7_3, \
call_6_0_7_4, \
call_6_0_7_5, \
call_6_0_7_6, \
call_6_0_7_7, \
call_6_0_7_8, \
call_6_0_7_9, \
call_6_0_7_10, \
call_6_0_7_11, \
call_6_0_7_12, \
call_6_0_7_13, \
call_6_0_7_14, \
call_6_0_7_15, \
call_6_0_8_0, \
call_6_0_8_1, \
call_6_0_8_2, \
call_6_0_8_3, \
call_6_0_8_4, \
call_6_0_8_5, \
call_6_0_8_6, \
call_6_0_8_7, \
call_6_0_8_8, \
call_6_0_8_9, \
call_6_0_8_10, \
call_6_0_8_11, \
call_6_0_8_12, \
call_6_0_8_13, \
call_6_0_8_14, \
call_6_0_8_15, \
call_6_0_9_0, \
call_6_0_9_1, \
call_6_0_9_2, \
call_6_0_9_3, \
call_6_0_9_4, \
call_6_0_9_5, \
call_6_0_9_6, \
call_6_0_9_7, \
call_6_0_9_8, \
call_6_0_9_9, \
call_6_0_9_10, \
call_6_0_9_11, \
call_6_0_9_12, \
call_6_0_9_13, \
call_6_0_9_14, \
call_6_0_9_15, \
call_6_0_10_0, \
call_6_0_10_1, \
call_6_0_10_2, \
call_6_0_10_3, \
call_6_0_10_4, \
call_6_0_10_5, \
call_6_0_10_6, \
call_6_0_10_7, \
call_6_0_10_8, \
call_6_0_10_9, \
call_6_0_10_10, \
call_6_0_10_11, \
call_6_0_10_12, \
call_6_0_10_13, \
call_6_0_10_14, \
call_6_0_10_15, \
call_6_0_11_0, \
call_6_0_11_1, \
call_6_0_11_2, \
call_6_0_11_3, \
call_6_0_11_4, \
call_6_0_11_5, \
call_6_0_11_6, \
call_6_0_11_7, \
call_6_0_11_8, \
call_6_0_11_9, \
call_6_0_11_10, \
call_6_0_11_11, \
call_6_0_11_12, \
call_6_0_11_13, \
call_6_0_11_14, \
call_6_0_11_15, \
call_6_0_12_0, \
call_6_0_12_1, \
call_6_0_12_2, \
call_6_0_12_3, \
call_6_0_12_4, \
call_6_0_12_5, \
call_6_0_12_6, \
call_6_0_12_7, \
call_6_0_12_8, \
call_6_0_12_9, \
call_6_0_12_10, \
call_6_0_12_11, \
call_6_0_12_12, \
call_6_0_12_13, \
call_6_0_12_14, \
call_6_0_12_15, \
call_6_0_13_0, \
call_6_0_13_1, \
call_6_0_13_2, \
call_6_0_13_3, \
call_6_0_13_4, \
call_6_0_13_5, \
call_6_0_13_6, \
call_6_0_13_7, \
call_6_0_13_8, \
call_6_0_13_9, \
call_6_0_13_10, \
call_6_0_13_11, \
call_6_0_13_12, \
call_6_0_13_13, \
call_6_0_13_14, \
call_6_0_13_15, \
call_6_0_14_0, \
call_6_0_14_1, \
call_6_0_14_2, \
call_6_0_14_3, \
call_6_0_14_4, \
call_6_0_14_5, \
call_6_0_14_6, \
call_6_0_14_7, \
call_6_0_14_8, \
call_6_0_14_9, \
call_6_0_14_10, \
call_6_0_14_11, \
call_6_0_14_12, \
call_6_0_14_13, \
call_6_0_14_14, \
call_6_0_14_15, \
call_6_0_15_0, \
call_6_0_15_1, \
call_6_0_15_2, \
call_6_0_15_3, \
call_6_0_15_4, \
call_6_0_15_5, \
call_6_0_15_6, \
call_6_0_15_7, \
call_6_0_15_8, \
call_6_0_15_9, \
call_6_0_15_10, \
call_6_0_15_11, \
call_6_0_15_12, \
call_6_0_15_13, \
call_6_0_15_14, \
call_6_0_15_15, \
call_6_1_0_0, \
call_6_1_0_1, \
call_6_1_0_2, \
call_6_1_0_3, \
call_6_1_0_4, \
call_6_1_0_5, \
call_6_1_0_6, \
call_6_1_0_7, \
call_6_1_0_8, \
call_6_1_0_9, \
call_6_1_0_10, \
call_6_1_0_11, \
call_6_1_0_12, \
call_6_1_0_13, \
call_6_1_0_14, \
call_6_1_0_15, \
call_6_1_1_0, \
call_6_1_1_1, \
call_6_1_1_2, \
call_6_1_1_3, \
call_6_1_1_4, \
call_6_1_1_5, \
call_6_1_1_6, \
call_6_1_1_7, \
call_6_1_1_8, \
call_6_1_1_9, \
call_6_1_1_10, \
call_6_1_1_11, \
call_6_1_1_12, \
call_6_1_1_13, \
call_6_1_1_14, \
call_6_1_1_15, \
call_6_1_2_0, \
call_6_1_2_1, \
call_6_1_2_2, \
call_6_1_2_3, \
call_6_1_2_4, \
call_6_1_2_5, \
call_6_1_2_6, \
call_6_1_2_7, \
call_6_1_2_8, \
call_6_1_2_9, \
call_6_1_2_10, \
call_6_1_2_11, \
call_6_1_2_12, \
call_6_1_2_13, \
call_6_1_2_14, \
call_6_1_2_15, \
call_6_1_3_0, \
call_6_1_3_1, \
call_6_1_3_2, \
call_6_1_3_3, \
call_6_1_3_4, \
call_6_1_3_5, \
call_6_1_3_6, \
call_6_1_3_7, \
call_6_1_3_8, \
call_6_1_3_9, \
call_6_1_3_10, \
call_6_1_3_11, \
call_6_1_3_12, \
call_6_1_3_13, \
call_6_1_3_14, \
call_6_1_3_15, \
call_6_1_4_0, \
call_6_1_4_1, \
call_6_1_4_2, \
call_6_1_4_3, \
call_6_1_4_4, \
call_6_1_4_5, \
call_6_1_4_6, \
call_6_1_4_7, \
call_6_1_4_8, \
call_6_1_4_9, \
call_6_1_4_10, \
call_6_1_4_11, \
call_6_1_4_12, \
call_6_1_4_13, \
call_6_1_4_14, \
call_6_1_4_15, \
call_6_1_5_0, \
call_6_1_5_1, \
call_6_1_5_2, \
call_6_1_5_3, \
call_6_1_5_4, \
call_6_1_5_5, \
call_6_1_5_6, \
call_6_1_5_7, \
call_6_1_5_8, \
call_6_1_5_9, \
call_6_1_5_10, \
call_6_1_5_11, \
call_6_1_5_12, \
call_6_1_5_13, \
call_6_1_5_14, \
call_6_1_5_15, \
call_6_1_6_0, \
call_6_1_6_1, \
call_6_1_6_2, \
call_6_1_6_3, \
call_6_1_6_4, \
call_6_1_6_5, \
call_6_1_6_6, \
call_6_1_6_7, \
call_6_1_6_8, \
call_6_1_6_9, \
call_6_1_6_10, \
call_6_1_6_11, \
call_6_1_6_12, \
call_6_1_6_13, \
call_6_1_6_14, \
call_6_1_6_15, \
call_6_1_7_0, \
call_6_1_7_1, \
call_6_1_7_2, \
call_6_1_7_3, \
call_6_1_7_4, \
call_6_1_7_5, \
call_6_1_7_6, \
call_6_1_7_7, \
call_6_1_7_8, \
call_6_1_7_9, \
call_6_1_7_10, \
call_6_1_7_11, \
call_6_1_7_12, \
call_6_1_7_13, \
call_6_1_7_14, \
call_6_1_7_15, \
call_6_1_8_0, \
call_6_1_8_1, \
call_6_1_8_2, \
call_6_1_8_3, \
call_6_1_8_4, \
call_6_1_8_5, \
call_6_1_8_6, \
call_6_1_8_7, \
call_6_1_8_8, \
call_6_1_8_9, \
call_6_1_8_10, \
call_6_1_8_11, \
call_6_1_8_12, \
call_6_1_8_13, \
call_6_1_8_14, \
call_6_1_8_15, \
call_6_1_9_0, \
call_6_1_9_1, \
call_6_1_9_2, \
call_6_1_9_3, \
call_6_1_9_4, \
call_6_1_9_5, \
call_6_1_9_6, \
call_6_1_9_7, \
call_6_1_9_8, \
call_6_1_9_9, \
call_6_1_9_10, \
call_6_1_9_11, \
call_6_1_9_12, \
call_6_1_9_13, \
call_6_1_9_14, \
call_6_1_9_15, \
call_6_1_10_0, \
call_6_1_10_1, \
call_6_1_10_2, \
call_6_1_10_3, \
call_6_1_10_4, \
call_6_1_10_5, \
call_6_1_10_6, \
call_6_1_10_7, \
call_6_1_10_8, \
call_6_1_10_9, \
call_6_1_10_10, \
call_6_1_10_11, \
call_6_1_10_12, \
call_6_1_10_13, \
call_6_1_10_14, \
call_6_1_10_15, \
call_6_1_11_0, \
call_6_1_11_1, \
call_6_1_11_2, \
call_6_1_11_3, \
call_6_1_11_4, \
call_6_1_11_5, \
call_6_1_11_6, \
call_6_1_11_7, \
call_6_1_11_8, \
call_6_1_11_9, \
call_6_1_11_10, \
call_6_1_11_11, \
call_6_1_11_12, \
call_6_1_11_13, \
call_6_1_11_14, \
call_6_1_11_15, \
call_6_1_12_0, \
call_6_1_12_1, \
call_6_1_12_2, \
call_6_1_12_3, \
call_6_1_12_4, \
call_6_1_12_5, \
call_6_1_12_6, \
call_6_1_12_7, \
call_6_1_12_8, \
call_6_1_12_9, \
call_6_1_12_10, \
call_6_1_12_11, \
call_6_1_12_12, \
call_6_1_12_13, \
call_6_1_12_14, \
call_6_1_12_15, \
call_6_1_13_0, \
call_6_1_13_1, \
call_6_1_13_2, \
call_6_1_13_3, \
call_6_1_13_4, \
call_6_1_13_5, \
call_6_1_13_6, \
call_6_1_13_7, \
call_6_1_13_8, \
call_6_1_13_9, \
call_6_1_13_10, \
call_6_1_13_11, \
call_6_1_13_12, \
call_6_1_13_13, \
call_6_1_13_14, \
call_6_1_13_15, \
call_6_1_14_0, \
call_6_1_14_1, \
call_6_1_14_2, \
call_6_1_14_3, \
call_6_1_14_4, \
call_6_1_14_5, \
call_6_1_14_6, \
call_6_1_14_7, \
call_6_1_14_8, \
call_6_1_14_9, \
call_6_1_14_10, \
call_6_1_14_11, \
call_6_1_14_12, \
call_6_1_14_13, \
call_6_1_14_14, \
call_6_1_14_15, \
call_6_1_15_0, \
call_6_1_15_1, \
call_6_1_15_2, \
call_6_1_15_3, \
call_6_1_15_4, \
call_6_1_15_5, \
call_6_1_15_6, \
call_6_1_15_7, \
call_6_1_15_8, \
call_6_1_15_9, \
call_6_1_15_10, \
call_6_1_15_11, \
call_6_1_15_12, \
call_6_1_15_13, \
call_6_1_15_14, \
call_6_1_15_15, \
call_6_2_0_0, \
call_6_2_0_1, \
call_6_2_0_2, \
call_6_2_0_3, \
call_6_2_0_4, \
call_6_2_0_5, \
call_6_2_0_6, \
call_6_2_0_7, \
call_6_2_0_8, \
call_6_2_0_9, \
call_6_2_0_10, \
call_6_2_0_11, \
call_6_2_0_12, \
call_6_2_0_13, \
call_6_2_0_14, \
call_6_2_0_15, \
call_6_2_1_0, \
call_6_2_1_1, \
call_6_2_1_2, \
call_6_2_1_3, \
call_6_2_1_4, \
call_6_2_1_5, \
call_6_2_1_6, \
call_6_2_1_7, \
call_6_2_1_8, \
call_6_2_1_9, \
call_6_2_1_10, \
call_6_2_1_11, \
call_6_2_1_12, \
call_6_2_1_13, \
call_6_2_1_14, \
call_6_2_1_15, \
call_6_2_2_0, \
call_6_2_2_1, \
call_6_2_2_2, \
call_6_2_2_3, \
call_6_2_2_4, \
call_6_2_2_5, \
call_6_2_2_6, \
call_6_2_2_7, \
call_6_2_2_8, \
call_6_2_2_9, \
call_6_2_2_10, \
call_6_2_2_11, \
call_6_2_2_12, \
call_6_2_2_13, \
call_6_2_2_14, \
call_6_2_2_15, \
call_6_2_3_0, \
call_6_2_3_1, \
call_6_2_3_2, \
call_6_2_3_3, \
call_6_2_3_4, \
call_6_2_3_5, \
call_6_2_3_6, \
call_6_2_3_7, \
call_6_2_3_8, \
call_6_2_3_9, \
call_6_2_3_10, \
call_6_2_3_11, \
call_6_2_3_12, \
call_6_2_3_13, \
call_6_2_3_14, \
call_6_2_3_15, \
call_6_2_4_0, \
call_6_2_4_1, \
call_6_2_4_2, \
call_6_2_4_3, \
call_6_2_4_4, \
call_6_2_4_5, \
call_6_2_4_6, \
call_6_2_4_7, \
call_6_2_4_8, \
call_6_2_4_9, \
call_6_2_4_10, \
call_6_2_4_11, \
call_6_2_4_12, \
call_6_2_4_13, \
call_6_2_4_14, \
call_6_2_4_15, \
call_6_2_5_0, \
call_6_2_5_1, \
call_6_2_5_2, \
call_6_2_5_3, \
call_6_2_5_4, \
call_6_2_5_5, \
call_6_2_5_6, \
call_6_2_5_7, \
call_6_2_5_8, \
call_6_2_5_9, \
call_6_2_5_10, \
call_6_2_5_11, \
call_6_2_5_12, \
call_6_2_5_13, \
call_6_2_5_14, \
call_6_2_5_15, \
call_6_2_6_0, \
call_6_2_6_1, \
call_6_2_6_2, \
call_6_2_6_3, \
call_6_2_6_4, \
call_6_2_6_5, \
call_6_2_6_6, \
call_6_2_6_7, \
call_6_2_6_8, \
call_6_2_6_9, \
call_6_2_6_10, \
call_6_2_6_11, \
call_6_2_6_12, \
call_6_2_6_13, \
call_6_2_6_14, \
call_6_2_6_15, \
call_6_2_7_0, \
call_6_2_7_1, \
call_6_2_7_2, \
call_6_2_7_3, \
call_6_2_7_4, \
call_6_2_7_5, \
call_6_2_7_6, \
call_6_2_7_7, \
call_6_2_7_8, \
call_6_2_7_9, \
call_6_2_7_10, \
call_6_2_7_11, \
call_6_2_7_12, \
call_6_2_7_13, \
call_6_2_7_14, \
call_6_2_7_15, \
call_6_2_8_0, \
call_6_2_8_1, \
call_6_2_8_2, \
call_6_2_8_3, \
call_6_2_8_4, \
call_6_2_8_5, \
call_6_2_8_6, \
call_6_2_8_7, \
call_6_2_8_8, \
call_6_2_8_9, \
call_6_2_8_10, \
call_6_2_8_11, \
call_6_2_8_12, \
call_6_2_8_13, \
call_6_2_8_14, \
call_6_2_8_15, \
call_6_2_9_0, \
call_6_2_9_1, \
call_6_2_9_2, \
call_6_2_9_3, \
call_6_2_9_4, \
call_6_2_9_5, \
call_6_2_9_6, \
call_6_2_9_7, \
call_6_2_9_8, \
call_6_2_9_9, \
call_6_2_9_10, \
call_6_2_9_11, \
call_6_2_9_12, \
call_6_2_9_13, \
call_6_2_9_14, \
call_6_2_9_15, \
call_6_2_10_0, \
call_6_2_10_1, \
call_6_2_10_2, \
call_6_2_10_3, \
call_6_2_10_4, \
call_6_2_10_5, \
call_6_2_10_6, \
call_6_2_10_7, \
call_6_2_10_8, \
call_6_2_10_9, \
call_6_2_10_10, \
call_6_2_10_11, \
call_6_2_10_12, \
call_6_2_10_13, \
call_6_2_10_14, \
call_6_2_10_15, \
call_6_2_11_0, \
call_6_2_11_1, \
call_6_2_11_2, \
call_6_2_11_3, \
call_6_2_11_4, \
call_6_2_11_5, \
call_6_2_11_6, \
call_6_2_11_7, \
call_6_2_11_8, \
call_6_2_11_9, \
call_6_2_11_10, \
call_6_2_11_11, \
call_6_2_11_12, \
call_6_2_11_13, \
call_6_2_11_14, \
call_6_2_11_15, \
call_6_2_12_0, \
call_6_2_12_1, \
call_6_2_12_2, \
call_6_2_12_3, \
call_6_2_12_4, \
call_6_2_12_5, \
call_6_2_12_6, \
call_6_2_12_7, \
call_6_2_12_8, \
call_6_2_12_9, \
call_6_2_12_10, \
call_6_2_12_11, \
call_6_2_12_12, \
call_6_2_12_13, \
call_6_2_12_14, \
call_6_2_12_15, \
call_6_2_13_0, \
call_6_2_13_1, \
call_6_2_13_2, \
call_6_2_13_3, \
call_6_2_13_4, \
call_6_2_13_5, \
call_6_2_13_6, \
call_6_2_13_7, \
call_6_2_13_8, \
call_6_2_13_9, \
call_6_2_13_10, \
call_6_2_13_11, \
call_6_2_13_12, \
call_6_2_13_13, \
call_6_2_13_14, \
call_6_2_13_15, \
call_6_2_14_0, \
call_6_2_14_1, \
call_6_2_14_2, \
call_6_2_14_3, \
call_6_2_14_4, \
call_6_2_14_5, \
call_6_2_14_6, \
call_6_2_14_7, \
call_6_2_14_8, \
call_6_2_14_9, \
call_6_2_14_10, \
call_6_2_14_11, \
call_6_2_14_12, \
call_6_2_14_13, \
call_6_2_14_14, \
call_6_2_14_15, \
call_6_2_15_0, \
call_6_2_15_1, \
call_6_2_15_2, \
call_6_2_15_3, \
call_6_2_15_4, \
call_6_2_15_5, \
call_6_2_15_6, \
call_6_2_15_7, \
call_6_2_15_8, \
call_6_2_15_9, \
call_6_2_15_10, \
call_6_2_15_11, \
call_6_2_15_12, \
call_6_2_15_13, \
call_6_2_15_14, \
call_6_2_15_15, \
call_6_3_0_0, \
call_6_3_0_1, \
call_6_3_0_2, \
call_6_3_0_3, \
call_6_3_0_4, \
call_6_3_0_5, \
call_6_3_0_6, \
call_6_3_0_7, \
call_6_3_0_8, \
call_6_3_0_9, \
call_6_3_0_10, \
call_6_3_0_11, \
call_6_3_0_12, \
call_6_3_0_13, \
call_6_3_0_14, \
call_6_3_0_15, \
call_6_3_1_0, \
call_6_3_1_1, \
call_6_3_1_2, \
call_6_3_1_3, \
call_6_3_1_4, \
call_6_3_1_5, \
call_6_3_1_6, \
call_6_3_1_7, \
call_6_3_1_8, \
call_6_3_1_9, \
call_6_3_1_10, \
call_6_3_1_11, \
call_6_3_1_12, \
call_6_3_1_13, \
call_6_3_1_14, \
call_6_3_1_15, \
call_6_3_2_0, \
call_6_3_2_1, \
call_6_3_2_2, \
call_6_3_2_3, \
call_6_3_2_4, \
call_6_3_2_5, \
call_6_3_2_6, \
call_6_3_2_7, \
call_6_3_2_8, \
call_6_3_2_9, \
call_6_3_2_10, \
call_6_3_2_11, \
call_6_3_2_12, \
call_6_3_2_13, \
call_6_3_2_14, \
call_6_3_2_15, \
call_6_3_3_0, \
call_6_3_3_1, \
call_6_3_3_2, \
call_6_3_3_3, \
call_6_3_3_4, \
call_6_3_3_5, \
call_6_3_3_6, \
call_6_3_3_7, \
call_6_3_3_8, \
call_6_3_3_9, \
call_6_3_3_10, \
call_6_3_3_11, \
call_6_3_3_12, \
call_6_3_3_13, \
call_6_3_3_14, \
call_6_3_3_15, \
call_6_3_4_0, \
call_6_3_4_1, \
call_6_3_4_2, \
call_6_3_4_3, \
call_6_3_4_4, \
call_6_3_4_5, \
call_6_3_4_6, \
call_6_3_4_7, \
call_6_3_4_8, \
call_6_3_4_9, \
call_6_3_4_10, \
call_6_3_4_11, \
call_6_3_4_12, \
call_6_3_4_13, \
call_6_3_4_14, \
call_6_3_4_15, \
call_6_3_5_0, \
call_6_3_5_1, \
call_6_3_5_2, \
call_6_3_5_3, \
call_6_3_5_4, \
call_6_3_5_5, \
call_6_3_5_6, \
call_6_3_5_7, \
call_6_3_5_8, \
call_6_3_5_9, \
call_6_3_5_10, \
call_6_3_5_11, \
call_6_3_5_12, \
call_6_3_5_13, \
call_6_3_5_14, \
call_6_3_5_15, \
call_6_3_6_0, \
call_6_3_6_1, \
call_6_3_6_2, \
call_6_3_6_3, \
call_6_3_6_4, \
call_6_3_6_5, \
call_6_3_6_6, \
call_6_3_6_7, \
call_6_3_6_8, \
call_6_3_6_9, \
call_6_3_6_10, \
call_6_3_6_11, \
call_6_3_6_12, \
call_6_3_6_13, \
call_6_3_6_14, \
call_6_3_6_15, \
call_6_3_7_0, \
call_6_3_7_1, \
call_6_3_7_2, \
call_6_3_7_3, \
call_6_3_7_4, \
call_6_3_7_5, \
call_6_3_7_6, \
call_6_3_7_7, \
call_6_3_7_8, \
call_6_3_7_9, \
call_6_3_7_10, \
call_6_3_7_11, \
call_6_3_7_12, \
call_6_3_7_13, \
call_6_3_7_14, \
call_6_3_7_15, \
call_6_3_8_0, \
call_6_3_8_1, \
call_6_3_8_2, \
call_6_3_8_3, \
call_6_3_8_4, \
call_6_3_8_5, \
call_6_3_8_6, \
call_6_3_8_7, \
call_6_3_8_8, \
call_6_3_8_9, \
call_6_3_8_10, \
call_6_3_8_11, \
call_6_3_8_12, \
call_6_3_8_13, \
call_6_3_8_14, \
call_6_3_8_15, \
call_6_3_9_0, \
call_6_3_9_1, \
call_6_3_9_2, \
call_6_3_9_3, \
call_6_3_9_4, \
call_6_3_9_5, \
call_6_3_9_6, \
call_6_3_9_7, \
call_6_3_9_8, \
call_6_3_9_9, \
call_6_3_9_10, \
call_6_3_9_11, \
call_6_3_9_12, \
call_6_3_9_13, \
call_6_3_9_14, \
call_6_3_9_15, \
call_6_3_10_0, \
call_6_3_10_1, \
call_6_3_10_2, \
call_6_3_10_3, \
call_6_3_10_4, \
call_6_3_10_5, \
call_6_3_10_6, \
call_6_3_10_7, \
call_6_3_10_8, \
call_6_3_10_9, \
call_6_3_10_10, \
call_6_3_10_11, \
call_6_3_10_12, \
call_6_3_10_13, \
call_6_3_10_14, \
call_6_3_10_15, \
call_6_3_11_0, \
call_6_3_11_1, \
call_6_3_11_2, \
call_6_3_11_3, \
call_6_3_11_4, \
call_6_3_11_5, \
call_6_3_11_6, \
call_6_3_11_7, \
call_6_3_11_8, \
call_6_3_11_9, \
call_6_3_11_10, \
call_6_3_11_11, \
call_6_3_11_12, \
call_6_3_11_13, \
call_6_3_11_14, \
call_6_3_11_15, \
call_6_3_12_0, \
call_6_3_12_1, \
call_6_3_12_2, \
call_6_3_12_3, \
call_6_3_12_4, \
call_6_3_12_5, \
call_6_3_12_6, \
call_6_3_12_7, \
call_6_3_12_8, \
call_6_3_12_9, \
call_6_3_12_10, \
call_6_3_12_11, \
call_6_3_12_12, \
call_6_3_12_13, \
call_6_3_12_14, \
call_6_3_12_15, \
call_6_3_13_0, \
call_6_3_13_1, \
call_6_3_13_2, \
call_6_3_13_3, \
call_6_3_13_4, \
call_6_3_13_5, \
call_6_3_13_6, \
call_6_3_13_7, \
call_6_3_13_8, \
call_6_3_13_9, \
call_6_3_13_10, \
call_6_3_13_11, \
call_6_3_13_12, \
call_6_3_13_13, \
call_6_3_13_14, \
call_6_3_13_15, \
call_6_3_14_0, \
call_6_3_14_1, \
call_6_3_14_2, \
call_6_3_14_3, \
call_6_3_14_4, \
call_6_3_14_5, \
call_6_3_14_6, \
call_6_3_14_7, \
call_6_3_14_8, \
call_6_3_14_9, \
call_6_3_14_10, \
call_6_3_14_11, \
call_6_3_14_12, \
call_6_3_14_13, \
call_6_3_14_14, \
call_6_3_14_15, \
call_6_3_15_0, \
call_6_3_15_1, \
call_6_3_15_2, \
call_6_3_15_3, \
call_6_3_15_4, \
call_6_3_15_5, \
call_6_3_15_6, \
call_6_3_15_7, \
call_6_3_15_8, \
call_6_3_15_9, \
call_6_3_15_10, \
call_6_3_15_11, \
call_6_3_15_12, \
call_6_3_15_13, \
call_6_3_15_14, \
call_6_3_15_15, \
call_6_4_0_0, \
call_6_4_0_1, \
call_6_4_0_2, \
call_6_4_0_3, \
call_6_4_0_4, \
call_6_4_0_5, \
call_6_4_0_6, \
call_6_4_0_7, \
call_6_4_0_8, \
call_6_4_0_9, \
call_6_4_0_10, \
call_6_4_0_11, \
call_6_4_0_12, \
call_6_4_0_13, \
call_6_4_0_14, \
call_6_4_0_15, \
call_6_4_1_0, \
call_6_4_1_1, \
call_6_4_1_2, \
call_6_4_1_3, \
call_6_4_1_4, \
call_6_4_1_5, \
call_6_4_1_6, \
call_6_4_1_7, \
call_6_4_1_8, \
call_6_4_1_9, \
call_6_4_1_10, \
call_6_4_1_11, \
call_6_4_1_12, \
call_6_4_1_13, \
call_6_4_1_14, \
call_6_4_1_15, \
call_6_4_2_0, \
call_6_4_2_1, \
call_6_4_2_2, \
call_6_4_2_3, \
call_6_4_2_4, \
call_6_4_2_5, \
call_6_4_2_6, \
call_6_4_2_7, \
call_6_4_2_8, \
call_6_4_2_9, \
call_6_4_2_10, \
call_6_4_2_11, \
call_6_4_2_12, \
call_6_4_2_13, \
call_6_4_2_14, \
call_6_4_2_15, \
call_6_4_3_0, \
call_6_4_3_1, \
call_6_4_3_2, \
call_6_4_3_3, \
call_6_4_3_4, \
call_6_4_3_5, \
call_6_4_3_6, \
call_6_4_3_7, \
call_6_4_3_8, \
call_6_4_3_9, \
call_6_4_3_10, \
call_6_4_3_11, \
call_6_4_3_12, \
call_6_4_3_13, \
call_6_4_3_14, \
call_6_4_3_15, \
call_6_4_4_0, \
call_6_4_4_1, \
call_6_4_4_2, \
call_6_4_4_3, \
call_6_4_4_4, \
call_6_4_4_5, \
call_6_4_4_6, \
call_6_4_4_7, \
call_6_4_4_8, \
call_6_4_4_9, \
call_6_4_4_10, \
call_6_4_4_11, \
call_6_4_4_12, \
call_6_4_4_13, \
call_6_4_4_14, \
call_6_4_4_15, \
call_6_4_5_0, \
call_6_4_5_1, \
call_6_4_5_2, \
call_6_4_5_3, \
call_6_4_5_4, \
call_6_4_5_5, \
call_6_4_5_6, \
call_6_4_5_7, \
call_6_4_5_8, \
call_6_4_5_9, \
call_6_4_5_10, \
call_6_4_5_11, \
call_6_4_5_12, \
call_6_4_5_13, \
call_6_4_5_14, \
call_6_4_5_15, \
call_6_4_6_0, \
call_6_4_6_1, \
call_6_4_6_2, \
call_6_4_6_3, \
call_6_4_6_4, \
call_6_4_6_5, \
call_6_4_6_6, \
call_6_4_6_7, \
call_6_4_6_8, \
call_6_4_6_9, \
call_6_4_6_10, \
call_6_4_6_11, \
call_6_4_6_12, \
call_6_4_6_13, \
call_6_4_6_14, \
call_6_4_6_15, \
call_6_4_7_0, \
call_6_4_7_1, \
call_6_4_7_2, \
call_6_4_7_3, \
call_6_4_7_4, \
call_6_4_7_5, \
call_6_4_7_6, \
call_6_4_7_7, \
call_6_4_7_8, \
call_6_4_7_9, \
call_6_4_7_10, \
call_6_4_7_11, \
call_6_4_7_12, \
call_6_4_7_13, \
call_6_4_7_14, \
call_6_4_7_15, \
call_6_4_8_0, \
call_6_4_8_1, \
call_6_4_8_2, \
call_6_4_8_3, \
call_6_4_8_4, \
call_6_4_8_5, \
call_6_4_8_6, \
call_6_4_8_7, \
call_6_4_8_8, \
call_6_4_8_9, \
call_6_4_8_10, \
call_6_4_8_11, \
call_6_4_8_12, \
call_6_4_8_13, \
call_6_4_8_14, \
call_6_4_8_15, \
call_6_4_9_0, \
call_6_4_9_1, \
call_6_4_9_2, \
call_6_4_9_3, \
call_6_4_9_4, \
call_6_4_9_5, \
call_6_4_9_6, \
call_6_4_9_7, \
call_6_4_9_8, \
call_6_4_9_9, \
call_6_4_9_10, \
call_6_4_9_11, \
call_6_4_9_12, \
call_6_4_9_13, \
call_6_4_9_14, \
call_6_4_9_15, \
call_6_4_10_0, \
call_6_4_10_1, \
call_6_4_10_2, \
call_6_4_10_3, \
call_6_4_10_4, \
call_6_4_10_5, \
call_6_4_10_6, \
call_6_4_10_7, \
call_6_4_10_8, \
call_6_4_10_9, \
call_6_4_10_10, \
call_6_4_10_11, \
call_6_4_10_12, \
call_6_4_10_13, \
call_6_4_10_14, \
call_6_4_10_15, \
call_6_4_11_0, \
call_6_4_11_1, \
call_6_4_11_2, \
call_6_4_11_3, \
call_6_4_11_4, \
call_6_4_11_5, \
call_6_4_11_6, \
call_6_4_11_7, \
call_6_4_11_8, \
call_6_4_11_9, \
call_6_4_11_10, \
call_6_4_11_11, \
call_6_4_11_12, \
call_6_4_11_13, \
call_6_4_11_14, \
call_6_4_11_15, \
call_6_4_12_0, \
call_6_4_12_1, \
call_6_4_12_2, \
call_6_4_12_3, \
call_6_4_12_4, \
call_6_4_12_5, \
call_6_4_12_6, \
call_6_4_12_7, \
call_6_4_12_8, \
call_6_4_12_9, \
call_6_4_12_10, \
call_6_4_12_11, \
call_6_4_12_12, \
call_6_4_12_13, \
call_6_4_12_14, \
call_6_4_12_15, \
call_6_4_13_0, \
call_6_4_13_1, \
call_6_4_13_2, \
call_6_4_13_3, \
call_6_4_13_4, \
call_6_4_13_5, \
call_6_4_13_6, \
call_6_4_13_7, \
call_6_4_13_8, \
call_6_4_13_9, \
call_6_4_13_10, \
call_6_4_13_11, \
call_6_4_13_12, \
call_6_4_13_13, \
call_6_4_13_14, \
call_6_4_13_15, \
call_6_4_14_0, \
call_6_4_14_1, \
call_6_4_14_2, \
call_6_4_14_3, \
call_6_4_14_4, \
call_6_4_14_5, \
call_6_4_14_6, \
call_6_4_14_7, \
call_6_4_14_8, \
call_6_4_14_9, \
call_6_4_14_10, \
call_6_4_14_11, \
call_6_4_14_12, \
call_6_4_14_13, \
call_6_4_14_14, \
call_6_4_14_15, \
call_6_4_15_0, \
call_6_4_15_1, \
call_6_4_15_2, \
call_6_4_15_3, \
call_6_4_15_4, \
call_6_4_15_5, \
call_6_4_15_6, \
call_6_4_15_7, \
call_6_4_15_8, \
call_6_4_15_9, \
call_6_4_15_10, \
call_6_4_15_11, \
call_6_4_15_12, \
call_6_4_15_13, \
call_6_4_15_14, \
call_6_4_15_15, \
call_6_5_0_0, \
call_6_5_0_1, \
call_6_5_0_2, \
call_6_5_0_3, \
call_6_5_0_4, \
call_6_5_0_5, \
call_6_5_0_6, \
call_6_5_0_7, \
call_6_5_0_8, \
call_6_5_0_9, \
call_6_5_0_10, \
call_6_5_0_11, \
call_6_5_0_12, \
call_6_5_0_13, \
call_6_5_0_14, \
call_6_5_0_15, \
call_6_5_1_0, \
call_6_5_1_1, \
call_6_5_1_2, \
call_6_5_1_3, \
call_6_5_1_4, \
call_6_5_1_5, \
call_6_5_1_6, \
call_6_5_1_7, \
call_6_5_1_8, \
call_6_5_1_9, \
call_6_5_1_10, \
call_6_5_1_11, \
call_6_5_1_12, \
call_6_5_1_13, \
call_6_5_1_14, \
call_6_5_1_15, \
call_6_5_2_0, \
call_6_5_2_1, \
call_6_5_2_2, \
call_6_5_2_3, \
call_6_5_2_4, \
call_6_5_2_5, \
call_6_5_2_6, \
call_6_5_2_7, \
call_6_5_2_8, \
call_6_5_2_9, \
call_6_5_2_10, \
call_6_5_2_11, \
call_6_5_2_12, \
call_6_5_2_13, \
call_6_5_2_14, \
call_6_5_2_15, \
call_6_5_3_0, \
call_6_5_3_1, \
call_6_5_3_2, \
call_6_5_3_3, \
call_6_5_3_4, \
call_6_5_3_5, \
call_6_5_3_6, \
call_6_5_3_7, \
call_6_5_3_8, \
call_6_5_3_9, \
call_6_5_3_10, \
call_6_5_3_11, \
call_6_5_3_12, \
call_6_5_3_13, \
call_6_5_3_14, \
call_6_5_3_15, \
call_6_5_4_0, \
call_6_5_4_1, \
call_6_5_4_2, \
call_6_5_4_3, \
call_6_5_4_4, \
call_6_5_4_5, \
call_6_5_4_6, \
call_6_5_4_7, \
call_6_5_4_8, \
call_6_5_4_9, \
call_6_5_4_10, \
call_6_5_4_11, \
call_6_5_4_12, \
call_6_5_4_13, \
call_6_5_4_14, \
call_6_5_4_15, \
call_6_5_5_0, \
call_6_5_5_1, \
call_6_5_5_2, \
call_6_5_5_3, \
call_6_5_5_4, \
call_6_5_5_5, \
call_6_5_5_6, \
call_6_5_5_7, \
call_6_5_5_8, \
call_6_5_5_9, \
call_6_5_5_10, \
call_6_5_5_11, \
call_6_5_5_12, \
call_6_5_5_13, \
call_6_5_5_14, \
call_6_5_5_15, \
call_6_5_6_0, \
call_6_5_6_1, \
call_6_5_6_2, \
call_6_5_6_3, \
call_6_5_6_4, \
call_6_5_6_5, \
call_6_5_6_6, \
call_6_5_6_7, \
call_6_5_6_8, \
call_6_5_6_9, \
call_6_5_6_10, \
call_6_5_6_11, \
call_6_5_6_12, \
call_6_5_6_13, \
call_6_5_6_14, \
call_6_5_6_15, \
call_6_5_7_0, \
call_6_5_7_1, \
call_6_5_7_2, \
call_6_5_7_3, \
call_6_5_7_4, \
call_6_5_7_5, \
call_6_5_7_6, \
call_6_5_7_7, \
call_6_5_7_8, \
call_6_5_7_9, \
call_6_5_7_10, \
call_6_5_7_11, \
call_6_5_7_12, \
call_6_5_7_13, \
call_6_5_7_14, \
call_6_5_7_15, \
call_6_5_8_0, \
call_6_5_8_1, \
call_6_5_8_2, \
call_6_5_8_3, \
call_6_5_8_4, \
call_6_5_8_5, \
call_6_5_8_6, \
call_6_5_8_7, \
call_6_5_8_8, \
call_6_5_8_9, \
call_6_5_8_10, \
call_6_5_8_11, \
call_6_5_8_12, \
call_6_5_8_13, \
call_6_5_8_14, \
call_6_5_8_15, \
call_6_5_9_0, \
call_6_5_9_1, \
call_6_5_9_2, \
call_6_5_9_3, \
call_6_5_9_4, \
call_6_5_9_5, \
call_6_5_9_6, \
call_6_5_9_7, \
call_6_5_9_8, \
call_6_5_9_9, \
call_6_5_9_10, \
call_6_5_9_11, \
call_6_5_9_12, \
call_6_5_9_13, \
call_6_5_9_14, \
call_6_5_9_15, \
call_6_5_10_0, \
call_6_5_10_1, \
call_6_5_10_2, \
call_6_5_10_3, \
call_6_5_10_4, \
call_6_5_10_5, \
call_6_5_10_6, \
call_6_5_10_7, \
call_6_5_10_8, \
call_6_5_10_9, \
call_6_5_10_10, \
call_6_5_10_11, \
call_6_5_10_12, \
call_6_5_10_13, \
call_6_5_10_14, \
call_6_5_10_15, \
call_6_5_11_0, \
call_6_5_11_1, \
call_6_5_11_2, \
call_6_5_11_3, \
call_6_5_11_4, \
call_6_5_11_5, \
call_6_5_11_6, \
call_6_5_11_7, \
call_6_5_11_8, \
call_6_5_11_9, \
call_6_5_11_10, \
call_6_5_11_11, \
call_6_5_11_12, \
call_6_5_11_13, \
call_6_5_11_14, \
call_6_5_11_15, \
call_6_5_12_0, \
call_6_5_12_1, \
call_6_5_12_2, \
call_6_5_12_3, \
call_6_5_12_4, \
call_6_5_12_5, \
call_6_5_12_6, \
call_6_5_12_7, \
call_6_5_12_8, \
call_6_5_12_9, \
call_6_5_12_10, \
call_6_5_12_11, \
call_6_5_12_12, \
call_6_5_12_13, \
call_6_5_12_14, \
call_6_5_12_15, \
call_6_5_13_0, \
call_6_5_13_1, \
call_6_5_13_2, \
call_6_5_13_3, \
call_6_5_13_4, \
call_6_5_13_5, \
call_6_5_13_6, \
call_6_5_13_7, \
call_6_5_13_8, \
call_6_5_13_9, \
call_6_5_13_10, \
call_6_5_13_11, \
call_6_5_13_12, \
call_6_5_13_13, \
call_6_5_13_14, \
call_6_5_13_15, \
call_6_5_14_0, \
call_6_5_14_1, \
call_6_5_14_2, \
call_6_5_14_3, \
call_6_5_14_4, \
call_6_5_14_5, \
call_6_5_14_6, \
call_6_5_14_7, \
call_6_5_14_8, \
call_6_5_14_9, \
call_6_5_14_10, \
call_6_5_14_11, \
call_6_5_14_12, \
call_6_5_14_13, \
call_6_5_14_14, \
call_6_5_14_15, \
call_6_5_15_0, \
call_6_5_15_1, \
call_6_5_15_2, \
call_6_5_15_3, \
call_6_5_15_4, \
call_6_5_15_5, \
call_6_5_15_6, \
call_6_5_15_7, \
call_6_5_15_8, \
call_6_5_15_9, \
call_6_5_15_10, \
call_6_5_15_11, \
call_6_5_15_12, \
call_6_5_15_13, \
call_6_5_15_14, \
call_6_5_15_15, \
call_6_6_0_0, \
call_6_6_0_1, \
call_6_6_0_2, \
call_6_6_0_3, \
call_6_6_0_4, \
call_6_6_0_5, \
call_6_6_0_6, \
call_6_6_0_7, \
call_6_6_0_8, \
call_6_6_0_9, \
call_6_6_0_10, \
call_6_6_0_11, \
call_6_6_0_12, \
call_6_6_0_13, \
call_6_6_0_14, \
call_6_6_0_15, \
call_6_6_1_0, \
call_6_6_1_1, \
call_6_6_1_2, \
call_6_6_1_3, \
call_6_6_1_4, \
call_6_6_1_5, \
call_6_6_1_6, \
call_6_6_1_7, \
call_6_6_1_8, \
call_6_6_1_9, \
call_6_6_1_10, \
call_6_6_1_11, \
call_6_6_1_12, \
call_6_6_1_13, \
call_6_6_1_14, \
call_6_6_1_15, \
call_6_6_2_0, \
call_6_6_2_1, \
call_6_6_2_2, \
call_6_6_2_3, \
call_6_6_2_4, \
call_6_6_2_5, \
call_6_6_2_6, \
call_6_6_2_7, \
call_6_6_2_8, \
call_6_6_2_9, \
call_6_6_2_10, \
call_6_6_2_11, \
call_6_6_2_12, \
call_6_6_2_13, \
call_6_6_2_14, \
call_6_6_2_15, \
call_6_6_3_0, \
call_6_6_3_1, \
call_6_6_3_2, \
call_6_6_3_3, \
call_6_6_3_4, \
call_6_6_3_5, \
call_6_6_3_6, \
call_6_6_3_7, \
call_6_6_3_8, \
call_6_6_3_9, \
call_6_6_3_10, \
call_6_6_3_11, \
call_6_6_3_12, \
call_6_6_3_13, \
call_6_6_3_14, \
call_6_6_3_15, \
call_6_6_4_0, \
call_6_6_4_1, \
call_6_6_4_2, \
call_6_6_4_3, \
call_6_6_4_4, \
call_6_6_4_5, \
call_6_6_4_6, \
call_6_6_4_7, \
call_6_6_4_8, \
call_6_6_4_9, \
call_6_6_4_10, \
call_6_6_4_11, \
call_6_6_4_12, \
call_6_6_4_13, \
call_6_6_4_14, \
call_6_6_4_15, \
call_6_6_5_0, \
call_6_6_5_1, \
call_6_6_5_2, \
call_6_6_5_3, \
call_6_6_5_4, \
call_6_6_5_5, \
call_6_6_5_6, \
call_6_6_5_7, \
call_6_6_5_8, \
call_6_6_5_9, \
call_6_6_5_10, \
call_6_6_5_11, \
call_6_6_5_12, \
call_6_6_5_13, \
call_6_6_5_14, \
call_6_6_5_15, \
call_6_6_6_0, \
call_6_6_6_1, \
call_6_6_6_2, \
call_6_6_6_3, \
call_6_6_6_4, \
call_6_6_6_5, \
call_6_6_6_6, \
call_6_6_6_7, \
call_6_6_6_8, \
call_6_6_6_9, \
call_6_6_6_10, \
call_6_6_6_11, \
call_6_6_6_12, \
call_6_6_6_13, \
call_6_6_6_14, \
call_6_6_6_15, \
call_6_6_7_0, \
call_6_6_7_1, \
call_6_6_7_2, \
call_6_6_7_3, \
call_6_6_7_4, \
call_6_6_7_5, \
call_6_6_7_6, \
call_6_6_7_7, \
call_6_6_7_8, \
call_6_6_7_9, \
call_6_6_7_10, \
call_6_6_7_11, \
call_6_6_7_12, \
call_6_6_7_13, \
call_6_6_7_14, \
call_6_6_7_15, \
call_6_6_8_0, \
call_6_6_8_1, \
call_6_6_8_2, \
call_6_6_8_3, \
call_6_6_8_4, \
call_6_6_8_5, \
call_6_6_8_6, \
call_6_6_8_7, \
call_6_6_8_8, \
call_6_6_8_9, \
call_6_6_8_10, \
call_6_6_8_11, \
call_6_6_8_12, \
call_6_6_8_13, \
call_6_6_8_14, \
call_6_6_8_15, \
call_6_6_9_0, \
call_6_6_9_1, \
call_6_6_9_2, \
call_6_6_9_3, \
call_6_6_9_4, \
call_6_6_9_5, \
call_6_6_9_6, \
call_6_6_9_7, \
call_6_6_9_8, \
call_6_6_9_9, \
call_6_6_9_10, \
call_6_6_9_11, \
call_6_6_9_12, \
call_6_6_9_13, \
call_6_6_9_14, \
call_6_6_9_15, \
call_6_6_10_0, \
call_6_6_10_1, \
call_6_6_10_2, \
call_6_6_10_3, \
call_6_6_10_4, \
call_6_6_10_5, \
call_6_6_10_6, \
call_6_6_10_7, \
call_6_6_10_8, \
call_6_6_10_9, \
call_6_6_10_10, \
call_6_6_10_11, \
call_6_6_10_12, \
call_6_6_10_13, \
call_6_6_10_14, \
call_6_6_10_15, \
call_6_6_11_0, \
call_6_6_11_1, \
call_6_6_11_2, \
call_6_6_11_3, \
call_6_6_11_4, \
call_6_6_11_5, \
call_6_6_11_6, \
call_6_6_11_7, \
call_6_6_11_8, \
call_6_6_11_9, \
call_6_6_11_10, \
call_6_6_11_11, \
call_6_6_11_12, \
call_6_6_11_13, \
call_6_6_11_14, \
call_6_6_11_15, \
call_6_6_12_0, \
call_6_6_12_1, \
call_6_6_12_2, \
call_6_6_12_3, \
call_6_6_12_4, \
call_6_6_12_5, \
call_6_6_12_6, \
call_6_6_12_7, \
call_6_6_12_8, \
call_6_6_12_9, \
call_6_6_12_10, \
call_6_6_12_11, \
call_6_6_12_12, \
call_6_6_12_13, \
call_6_6_12_14, \
call_6_6_12_15, \
call_6_6_13_0, \
call_6_6_13_1, \
call_6_6_13_2, \
call_6_6_13_3, \
call_6_6_13_4, \
call_6_6_13_5, \
call_6_6_13_6, \
call_6_6_13_7, \
call_6_6_13_8, \
call_6_6_13_9, \
call_6_6_13_10, \
call_6_6_13_11, \
call_6_6_13_12, \
call_6_6_13_13, \
call_6_6_13_14, \
call_6_6_13_15, \
call_6_6_14_0, \
call_6_6_14_1, \
call_6_6_14_2, \
call_6_6_14_3, \
call_6_6_14_4, \
call_6_6_14_5, \
call_6_6_14_6, \
call_6_6_14_7, \
call_6_6_14_8, \
call_6_6_14_9, \
call_6_6_14_10, \
call_6_6_14_11, \
call_6_6_14_12, \
call_6_6_14_13, \
call_6_6_14_14, \
call_6_6_14_15, \
call_6_6_15_0, \
call_6_6_15_1, \
call_6_6_15_2, \
call_6_6_15_3, \
call_6_6_15_4, \
call_6_6_15_5, \
call_6_6_15_6, \
call_6_6_15_7, \
call_6_6_15_8, \
call_6_6_15_9, \
call_6_6_15_10, \
call_6_6_15_11, \
call_6_6_15_12, \
call_6_6_15_13, \
call_6_6_15_14, \
call_6_6_15_15, \
call_6_7_0_0, \
call_6_7_0_1, \
call_6_7_0_2, \
call_6_7_0_3, \
call_6_7_0_4, \
call_6_7_0_5, \
call_6_7_0_6, \
call_6_7_0_7, \
call_6_7_0_8, \
call_6_7_0_9, \
call_6_7_0_10, \
call_6_7_0_11, \
call_6_7_0_12, \
call_6_7_0_13, \
call_6_7_0_14, \
call_6_7_0_15, \
call_6_7_1_0, \
call_6_7_1_1, \
call_6_7_1_2, \
call_6_7_1_3, \
call_6_7_1_4, \
call_6_7_1_5, \
call_6_7_1_6, \
call_6_7_1_7, \
call_6_7_1_8, \
call_6_7_1_9, \
call_6_7_1_10, \
call_6_7_1_11, \
call_6_7_1_12, \
call_6_7_1_13, \
call_6_7_1_14, \
call_6_7_1_15, \
call_6_7_2_0, \
call_6_7_2_1, \
call_6_7_2_2, \
call_6_7_2_3, \
call_6_7_2_4, \
call_6_7_2_5, \
call_6_7_2_6, \
call_6_7_2_7, \
call_6_7_2_8, \
call_6_7_2_9, \
call_6_7_2_10, \
call_6_7_2_11, \
call_6_7_2_12, \
call_6_7_2_13, \
call_6_7_2_14, \
call_6_7_2_15, \
call_6_7_3_0, \
call_6_7_3_1, \
call_6_7_3_2, \
call_6_7_3_3, \
call_6_7_3_4, \
call_6_7_3_5, \
call_6_7_3_6, \
call_6_7_3_7, \
call_6_7_3_8, \
call_6_7_3_9, \
call_6_7_3_10, \
call_6_7_3_11, \
call_6_7_3_12, \
call_6_7_3_13, \
call_6_7_3_14, \
call_6_7_3_15, \
call_6_7_4_0, \
call_6_7_4_1, \
call_6_7_4_2, \
call_6_7_4_3, \
call_6_7_4_4, \
call_6_7_4_5, \
call_6_7_4_6, \
call_6_7_4_7, \
call_6_7_4_8, \
call_6_7_4_9, \
call_6_7_4_10, \
call_6_7_4_11, \
call_6_7_4_12, \
call_6_7_4_13, \
call_6_7_4_14, \
call_6_7_4_15, \
call_6_7_5_0, \
call_6_7_5_1, \
call_6_7_5_2, \
call_6_7_5_3, \
call_6_7_5_4, \
call_6_7_5_5, \
call_6_7_5_6, \
call_6_7_5_7, \
call_6_7_5_8, \
call_6_7_5_9, \
call_6_7_5_10, \
call_6_7_5_11, \
call_6_7_5_12, \
call_6_7_5_13, \
call_6_7_5_14, \
call_6_7_5_15, \
call_6_7_6_0, \
call_6_7_6_1, \
call_6_7_6_2, \
call_6_7_6_3, \
call_6_7_6_4, \
call_6_7_6_5, \
call_6_7_6_6, \
call_6_7_6_7, \
call_6_7_6_8, \
call_6_7_6_9, \
call_6_7_6_10, \
call_6_7_6_11, \
call_6_7_6_12, \
call_6_7_6_13, \
call_6_7_6_14, \
call_6_7_6_15, \
call_6_7_7_0, \
call_6_7_7_1, \
call_6_7_7_2, \
call_6_7_7_3, \
call_6_7_7_4, \
call_6_7_7_5, \
call_6_7_7_6, \
call_6_7_7_7, \
call_6_7_7_8, \
call_6_7_7_9, \
call_6_7_7_10, \
call_6_7_7_11, \
call_6_7_7_12, \
call_6_7_7_13, \
call_6_7_7_14, \
call_6_7_7_15, \
call_6_7_8_0, \
call_6_7_8_1, \
call_6_7_8_2, \
call_6_7_8_3, \
call_6_7_8_4, \
call_6_7_8_5, \
call_6_7_8_6, \
call_6_7_8_7, \
call_6_7_8_8, \
call_6_7_8_9, \
call_6_7_8_10, \
call_6_7_8_11, \
call_6_7_8_12, \
call_6_7_8_13, \
call_6_7_8_14, \
call_6_7_8_15, \
call_6_7_9_0, \
call_6_7_9_1, \
call_6_7_9_2, \
call_6_7_9_3, \
call_6_7_9_4, \
call_6_7_9_5, \
call_6_7_9_6, \
call_6_7_9_7, \
call_6_7_9_8, \
call_6_7_9_9, \
call_6_7_9_10, \
call_6_7_9_11, \
call_6_7_9_12, \
call_6_7_9_13, \
call_6_7_9_14, \
call_6_7_9_15, \
call_6_7_10_0, \
call_6_7_10_1, \
call_6_7_10_2, \
call_6_7_10_3, \
call_6_7_10_4, \
call_6_7_10_5, \
call_6_7_10_6, \
call_6_7_10_7, \
call_6_7_10_8, \
call_6_7_10_9, \
call_6_7_10_10, \
call_6_7_10_11, \
call_6_7_10_12, \
call_6_7_10_13, \
call_6_7_10_14, \
call_6_7_10_15, \
call_6_7_11_0, \
call_6_7_11_1, \
call_6_7_11_2, \
call_6_7_11_3, \
call_6_7_11_4, \
call_6_7_11_5, \
call_6_7_11_6, \
call_6_7_11_7, \
call_6_7_11_8, \
call_6_7_11_9, \
call_6_7_11_10, \
call_6_7_11_11, \
call_6_7_11_12, \
call_6_7_11_13, \
call_6_7_11_14, \
call_6_7_11_15, \
call_6_7_12_0, \
call_6_7_12_1, \
call_6_7_12_2, \
call_6_7_12_3, \
call_6_7_12_4, \
call_6_7_12_5, \
call_6_7_12_6, \
call_6_7_12_7, \
call_6_7_12_8, \
call_6_7_12_9, \
call_6_7_12_10, \
call_6_7_12_11, \
call_6_7_12_12, \
call_6_7_12_13, \
call_6_7_12_14, \
call_6_7_12_15, \
call_6_7_13_0, \
call_6_7_13_1, \
call_6_7_13_2, \
call_6_7_13_3, \
call_6_7_13_4, \
call_6_7_13_5, \
call_6_7_13_6, \
call_6_7_13_7, \
call_6_7_13_8, \
call_6_7_13_9, \
call_6_7_13_10, \
call_6_7_13_11, \
call_6_7_13_12, \
call_6_7_13_13, \
call_6_7_13_14, \
call_6_7_13_15, \
call_6_7_14_0, \
call_6_7_14_1, \
call_6_7_14_2, \
call_6_7_14_3, \
call_6_7_14_4, \
call_6_7_14_5, \
call_6_7_14_6, \
call_6_7_14_7, \
call_6_7_14_8, \
call_6_7_14_9, \
call_6_7_14_10, \
call_6_7_14_11, \
call_6_7_14_12, \
call_6_7_14_13, \
call_6_7_14_14, \
call_6_7_14_15, \
call_6_7_15_0, \
call_6_7_15_1, \
call_6_7_15_2, \
call_6_7_15_3, \
call_6_7_15_4, \
call_6_7_15_5, \
call_6_7_15_6, \
call_6_7_15_7, \
call_6_7_15_8, \
call_6_7_15_9, \
call_6_7_15_10, \
call_6_7_15_11, \
call_6_7_15_12, \
call_6_7_15_13, \
call_6_7_15_14, \
call_6_7_15_15, \
call_6_8_0_0, \
call_6_8_0_1, \
call_6_8_0_2, \
call_6_8_0_3, \
call_6_8_0_4, \
call_6_8_0_5, \
call_6_8_0_6, \
call_6_8_0_7, \
call_6_8_0_8, \
call_6_8_0_9, \
call_6_8_0_10, \
call_6_8_0_11, \
call_6_8_0_12, \
call_6_8_0_13, \
call_6_8_0_14, \
call_6_8_0_15, \
call_6_8_1_0, \
call_6_8_1_1, \
call_6_8_1_2, \
call_6_8_1_3, \
call_6_8_1_4, \
call_6_8_1_5, \
call_6_8_1_6, \
call_6_8_1_7, \
call_6_8_1_8, \
call_6_8_1_9, \
call_6_8_1_10, \
call_6_8_1_11, \
call_6_8_1_12, \
call_6_8_1_13, \
call_6_8_1_14, \
call_6_8_1_15, \
call_6_8_2_0, \
call_6_8_2_1, \
call_6_8_2_2, \
call_6_8_2_3, \
call_6_8_2_4, \
call_6_8_2_5, \
call_6_8_2_6, \
call_6_8_2_7, \
call_6_8_2_8, \
call_6_8_2_9, \
call_6_8_2_10, \
call_6_8_2_11, \
call_6_8_2_12, \
call_6_8_2_13, \
call_6_8_2_14, \
call_6_8_2_15, \
call_6_8_3_0, \
call_6_8_3_1, \
call_6_8_3_2, \
call_6_8_3_3, \
call_6_8_3_4, \
call_6_8_3_5, \
call_6_8_3_6, \
call_6_8_3_7, \
call_6_8_3_8, \
call_6_8_3_9, \
call_6_8_3_10, \
call_6_8_3_11, \
call_6_8_3_12, \
call_6_8_3_13, \
call_6_8_3_14, \
call_6_8_3_15, \
call_6_8_4_0, \
call_6_8_4_1, \
call_6_8_4_2, \
call_6_8_4_3, \
call_6_8_4_4, \
call_6_8_4_5, \
call_6_8_4_6, \
call_6_8_4_7, \
call_6_8_4_8, \
call_6_8_4_9, \
call_6_8_4_10, \
call_6_8_4_11, \
call_6_8_4_12, \
call_6_8_4_13, \
call_6_8_4_14, \
call_6_8_4_15, \
call_6_8_5_0, \
call_6_8_5_1, \
call_6_8_5_2, \
call_6_8_5_3, \
call_6_8_5_4, \
call_6_8_5_5, \
call_6_8_5_6, \
call_6_8_5_7, \
call_6_8_5_8, \
call_6_8_5_9, \
call_6_8_5_10, \
call_6_8_5_11, \
call_6_8_5_12, \
call_6_8_5_13, \
call_6_8_5_14, \
call_6_8_5_15, \
call_6_8_6_0, \
call_6_8_6_1, \
call_6_8_6_2, \
call_6_8_6_3, \
call_6_8_6_4, \
call_6_8_6_5, \
call_6_8_6_6, \
call_6_8_6_7, \
call_6_8_6_8, \
call_6_8_6_9, \
call_6_8_6_10, \
call_6_8_6_11, \
call_6_8_6_12, \
call_6_8_6_13, \
call_6_8_6_14, \
call_6_8_6_15, \
call_6_8_7_0, \
call_6_8_7_1, \
call_6_8_7_2, \
call_6_8_7_3, \
call_6_8_7_4, \
call_6_8_7_5, \
call_6_8_7_6, \
call_6_8_7_7, \
call_6_8_7_8, \
call_6_8_7_9, \
call_6_8_7_10, \
call_6_8_7_11, \
call_6_8_7_12, \
call_6_8_7_13, \
call_6_8_7_14, \
call_6_8_7_15, \
call_6_8_8_0, \
call_6_8_8_1, \
call_6_8_8_2, \
call_6_8_8_3, \
call_6_8_8_4, \
call_6_8_8_5, \
call_6_8_8_6, \
call_6_8_8_7, \
call_6_8_8_8, \
call_6_8_8_9, \
call_6_8_8_10, \
call_6_8_8_11, \
call_6_8_8_12, \
call_6_8_8_13, \
call_6_8_8_14, \
call_6_8_8_15, \
call_6_8_9_0, \
call_6_8_9_1, \
call_6_8_9_2, \
call_6_8_9_3, \
call_6_8_9_4, \
call_6_8_9_5, \
call_6_8_9_6, \
call_6_8_9_7, \
call_6_8_9_8, \
call_6_8_9_9, \
call_6_8_9_10, \
call_6_8_9_11, \
call_6_8_9_12, \
call_6_8_9_13, \
call_6_8_9_14, \
call_6_8_9_15, \
call_6_8_10_0, \
call_6_8_10_1, \
call_6_8_10_2, \
call_6_8_10_3, \
call_6_8_10_4, \
call_6_8_10_5, \
call_6_8_10_6, \
call_6_8_10_7, \
call_6_8_10_8, \
call_6_8_10_9, \
call_6_8_10_10, \
call_6_8_10_11, \
call_6_8_10_12, \
call_6_8_10_13, \
call_6_8_10_14, \
call_6_8_10_15, \
call_6_8_11_0, \
call_6_8_11_1, \
call_6_8_11_2, \
call_6_8_11_3, \
call_6_8_11_4, \
call_6_8_11_5, \
call_6_8_11_6, \
call_6_8_11_7, \
call_6_8_11_8, \
call_6_8_11_9, \
call_6_8_11_10, \
call_6_8_11_11, \
call_6_8_11_12, \
call_6_8_11_13, \
call_6_8_11_14, \
call_6_8_11_15, \
call_6_8_12_0, \
call_6_8_12_1, \
call_6_8_12_2, \
call_6_8_12_3, \
call_6_8_12_4, \
call_6_8_12_5, \
call_6_8_12_6, \
call_6_8_12_7, \
call_6_8_12_8, \
call_6_8_12_9, \
call_6_8_12_10, \
call_6_8_12_11, \
call_6_8_12_12, \
call_6_8_12_13, \
call_6_8_12_14, \
call_6_8_12_15, \
call_6_8_13_0, \
call_6_8_13_1, \
call_6_8_13_2, \
call_6_8_13_3, \
call_6_8_13_4, \
call_6_8_13_5, \
call_6_8_13_6, \
call_6_8_13_7, \
call_6_8_13_8, \
call_6_8_13_9, \
call_6_8_13_10, \
call_6_8_13_11, \
call_6_8_13_12, \
call_6_8_13_13, \
call_6_8_13_14, \
call_6_8_13_15, \
call_6_8_14_0, \
call_6_8_14_1, \
call_6_8_14_2, \
call_6_8_14_3, \
call_6_8_14_4, \
call_6_8_14_5, \
call_6_8_14_6, \
call_6_8_14_7, \
call_6_8_14_8, \
call_6_8_14_9, \
call_6_8_14_10, \
call_6_8_14_11, \
call_6_8_14_12, \
call_6_8_14_13, \
call_6_8_14_14, \
call_6_8_14_15, \
call_6_8_15_0, \
call_6_8_15_1, \
call_6_8_15_2, \
call_6_8_15_3, \
call_6_8_15_4, \
call_6_8_15_5, \
call_6_8_15_6, \
call_6_8_15_7, \
call_6_8_15_8, \
call_6_8_15_9, \
call_6_8_15_10, \
call_6_8_15_11, \
call_6_8_15_12, \
call_6_8_15_13, \
call_6_8_15_14, \
call_6_8_15_15, \
call_6_9_0_0, \
call_6_9_0_1, \
call_6_9_0_2, \
call_6_9_0_3, \
call_6_9_0_4, \
call_6_9_0_5, \
call_6_9_0_6, \
call_6_9_0_7, \
call_6_9_0_8, \
call_6_9_0_9, \
call_6_9_0_10, \
call_6_9_0_11, \
call_6_9_0_12, \
call_6_9_0_13, \
call_6_9_0_14, \
call_6_9_0_15, \
call_6_9_1_0, \
call_6_9_1_1, \
call_6_9_1_2, \
call_6_9_1_3, \
call_6_9_1_4, \
call_6_9_1_5, \
call_6_9_1_6, \
call_6_9_1_7, \
call_6_9_1_8, \
call_6_9_1_9, \
call_6_9_1_10, \
call_6_9_1_11, \
call_6_9_1_12, \
call_6_9_1_13, \
call_6_9_1_14, \
call_6_9_1_15, \
call_6_9_2_0, \
call_6_9_2_1, \
call_6_9_2_2, \
call_6_9_2_3, \
call_6_9_2_4, \
call_6_9_2_5, \
call_6_9_2_6, \
call_6_9_2_7, \
call_6_9_2_8, \
call_6_9_2_9, \
call_6_9_2_10, \
call_6_9_2_11, \
call_6_9_2_12, \
call_6_9_2_13, \
call_6_9_2_14, \
call_6_9_2_15, \
call_6_9_3_0, \
call_6_9_3_1, \
call_6_9_3_2, \
call_6_9_3_3, \
call_6_9_3_4, \
call_6_9_3_5, \
call_6_9_3_6, \
call_6_9_3_7, \
call_6_9_3_8, \
call_6_9_3_9, \
call_6_9_3_10, \
call_6_9_3_11, \
call_6_9_3_12, \
call_6_9_3_13, \
call_6_9_3_14, \
call_6_9_3_15, \
call_6_9_4_0, \
call_6_9_4_1, \
call_6_9_4_2, \
call_6_9_4_3, \
call_6_9_4_4, \
call_6_9_4_5, \
call_6_9_4_6, \
call_6_9_4_7, \
call_6_9_4_8, \
call_6_9_4_9, \
call_6_9_4_10, \
call_6_9_4_11, \
call_6_9_4_12, \
call_6_9_4_13, \
call_6_9_4_14, \
call_6_9_4_15, \
call_6_9_5_0, \
call_6_9_5_1, \
call_6_9_5_2, \
call_6_9_5_3, \
call_6_9_5_4, \
call_6_9_5_5, \
call_6_9_5_6, \
call_6_9_5_7, \
call_6_9_5_8, \
call_6_9_5_9, \
call_6_9_5_10, \
call_6_9_5_11, \
call_6_9_5_12, \
call_6_9_5_13, \
call_6_9_5_14, \
call_6_9_5_15, \
call_6_9_6_0, \
call_6_9_6_1, \
call_6_9_6_2, \
call_6_9_6_3, \
call_6_9_6_4, \
call_6_9_6_5, \
call_6_9_6_6, \
call_6_9_6_7, \
call_6_9_6_8, \
call_6_9_6_9, \
call_6_9_6_10, \
call_6_9_6_11, \
call_6_9_6_12, \
call_6_9_6_13, \
call_6_9_6_14, \
call_6_9_6_15, \
call_6_9_7_0, \
call_6_9_7_1, \
call_6_9_7_2, \
call_6_9_7_3, \
call_6_9_7_4, \
call_6_9_7_5, \
call_6_9_7_6, \
call_6_9_7_7, \
call_6_9_7_8, \
call_6_9_7_9, \
call_6_9_7_10, \
call_6_9_7_11, \
call_6_9_7_12, \
call_6_9_7_13, \
call_6_9_7_14, \
call_6_9_7_15, \
call_6_9_8_0, \
call_6_9_8_1, \
call_6_9_8_2, \
call_6_9_8_3, \
call_6_9_8_4, \
call_6_9_8_5, \
call_6_9_8_6, \
call_6_9_8_7, \
call_6_9_8_8, \
call_6_9_8_9, \
call_6_9_8_10, \
call_6_9_8_11, \
call_6_9_8_12, \
call_6_9_8_13, \
call_6_9_8_14, \
call_6_9_8_15, \
call_6_9_9_0, \
call_6_9_9_1, \
call_6_9_9_2, \
call_6_9_9_3, \
call_6_9_9_4, \
call_6_9_9_5, \
call_6_9_9_6, \
call_6_9_9_7, \
call_6_9_9_8, \
call_6_9_9_9, \
call_6_9_9_10, \
call_6_9_9_11, \
call_6_9_9_12, \
call_6_9_9_13, \
call_6_9_9_14, \
call_6_9_9_15, \
call_6_9_10_0, \
call_6_9_10_1, \
call_6_9_10_2, \
call_6_9_10_3, \
call_6_9_10_4, \
call_6_9_10_5, \
call_6_9_10_6, \
call_6_9_10_7, \
call_6_9_10_8, \
call_6_9_10_9, \
call_6_9_10_10, \
call_6_9_10_11, \
call_6_9_10_12, \
call_6_9_10_13, \
call_6_9_10_14, \
call_6_9_10_15, \
call_6_9_11_0, \
call_6_9_11_1, \
call_6_9_11_2, \
call_6_9_11_3, \
call_6_9_11_4, \
call_6_9_11_5, \
call_6_9_11_6, \
call_6_9_11_7, \
call_6_9_11_8, \
call_6_9_11_9, \
call_6_9_11_10, \
call_6_9_11_11, \
call_6_9_11_12, \
call_6_9_11_13, \
call_6_9_11_14, \
call_6_9_11_15, \
call_6_9_12_0, \
call_6_9_12_1, \
call_6_9_12_2, \
call_6_9_12_3, \
call_6_9_12_4, \
call_6_9_12_5, \
call_6_9_12_6, \
call_6_9_12_7, \
call_6_9_12_8, \
call_6_9_12_9, \
call_6_9_12_10, \
call_6_9_12_11, \
call_6_9_12_12, \
call_6_9_12_13, \
call_6_9_12_14, \
call_6_9_12_15, \
call_6_9_13_0, \
call_6_9_13_1, \
call_6_9_13_2, \
call_6_9_13_3, \
call_6_9_13_4, \
call_6_9_13_5, \
call_6_9_13_6, \
call_6_9_13_7, \
call_6_9_13_8, \
call_6_9_13_9, \
call_6_9_13_10, \
call_6_9_13_11, \
call_6_9_13_12, \
call_6_9_13_13, \
call_6_9_13_14, \
call_6_9_13_15, \
call_6_9_14_0, \
call_6_9_14_1, \
call_6_9_14_2, \
call_6_9_14_3, \
call_6_9_14_4, \
call_6_9_14_5, \
call_6_9_14_6, \
call_6_9_14_7, \
call_6_9_14_8, \
call_6_9_14_9, \
call_6_9_14_10, \
call_6_9_14_11, \
call_6_9_14_12, \
call_6_9_14_13, \
call_6_9_14_14, \
call_6_9_14_15, \
call_6_9_15_0, \
call_6_9_15_1, \
call_6_9_15_2, \
call_6_9_15_3, \
call_6_9_15_4, \
call_6_9_15_5, \
call_6_9_15_6, \
call_6_9_15_7, \
call_6_9_15_8, \
call_6_9_15_9, \
call_6_9_15_10, \
call_6_9_15_11, \
call_6_9_15_12, \
call_6_9_15_13, \
call_6_9_15_14, \
call_6_9_15_15, \
call_6_10_0_0, \
call_6_10_0_1, \
call_6_10_0_2, \
call_6_10_0_3, \
call_6_10_0_4, \
call_6_10_0_5, \
call_6_10_0_6, \
call_6_10_0_7, \
call_6_10_0_8, \
call_6_10_0_9, \
call_6_10_0_10, \
call_6_10_0_11, \
call_6_10_0_12, \
call_6_10_0_13, \
call_6_10_0_14, \
call_6_10_0_15, \
call_6_10_1_0, \
call_6_10_1_1, \
call_6_10_1_2, \
call_6_10_1_3, \
call_6_10_1_4, \
call_6_10_1_5, \
call_6_10_1_6, \
call_6_10_1_7, \
call_6_10_1_8, \
call_6_10_1_9, \
call_6_10_1_10, \
call_6_10_1_11, \
call_6_10_1_12, \
call_6_10_1_13, \
call_6_10_1_14, \
call_6_10_1_15, \
call_6_10_2_0, \
call_6_10_2_1, \
call_6_10_2_2, \
call_6_10_2_3, \
call_6_10_2_4, \
call_6_10_2_5, \
call_6_10_2_6, \
call_6_10_2_7, \
call_6_10_2_8, \
call_6_10_2_9, \
call_6_10_2_10, \
call_6_10_2_11, \
call_6_10_2_12, \
call_6_10_2_13, \
call_6_10_2_14, \
call_6_10_2_15, \
call_6_10_3_0, \
call_6_10_3_1, \
call_6_10_3_2, \
call_6_10_3_3, \
call_6_10_3_4, \
call_6_10_3_5, \
call_6_10_3_6, \
call_6_10_3_7, \
call_6_10_3_8, \
call_6_10_3_9, \
call_6_10_3_10, \
call_6_10_3_11, \
call_6_10_3_12, \
call_6_10_3_13, \
call_6_10_3_14, \
call_6_10_3_15, \
call_6_10_4_0, \
call_6_10_4_1, \
call_6_10_4_2, \
call_6_10_4_3, \
call_6_10_4_4, \
call_6_10_4_5, \
call_6_10_4_6, \
call_6_10_4_7, \
call_6_10_4_8, \
call_6_10_4_9, \
call_6_10_4_10, \
call_6_10_4_11, \
call_6_10_4_12, \
call_6_10_4_13, \
call_6_10_4_14, \
call_6_10_4_15, \
call_6_10_5_0, \
call_6_10_5_1, \
call_6_10_5_2, \
call_6_10_5_3, \
call_6_10_5_4, \
call_6_10_5_5, \
call_6_10_5_6, \
call_6_10_5_7, \
call_6_10_5_8, \
call_6_10_5_9, \
call_6_10_5_10, \
call_6_10_5_11, \
call_6_10_5_12, \
call_6_10_5_13, \
call_6_10_5_14, \
call_6_10_5_15, \
call_6_10_6_0, \
call_6_10_6_1, \
call_6_10_6_2, \
call_6_10_6_3, \
call_6_10_6_4, \
call_6_10_6_5, \
call_6_10_6_6, \
call_6_10_6_7, \
call_6_10_6_8, \
call_6_10_6_9, \
call_6_10_6_10, \
call_6_10_6_11, \
call_6_10_6_12, \
call_6_10_6_13, \
call_6_10_6_14, \
call_6_10_6_15, \
call_6_10_7_0, \
call_6_10_7_1, \
call_6_10_7_2, \
call_6_10_7_3, \
call_6_10_7_4, \
call_6_10_7_5, \
call_6_10_7_6, \
call_6_10_7_7, \
call_6_10_7_8, \
call_6_10_7_9, \
call_6_10_7_10, \
call_6_10_7_11, \
call_6_10_7_12, \
call_6_10_7_13, \
call_6_10_7_14, \
call_6_10_7_15, \
call_6_10_8_0, \
call_6_10_8_1, \
call_6_10_8_2, \
call_6_10_8_3, \
call_6_10_8_4, \
call_6_10_8_5, \
call_6_10_8_6, \
call_6_10_8_7, \
call_6_10_8_8, \
call_6_10_8_9, \
call_6_10_8_10, \
call_6_10_8_11, \
call_6_10_8_12, \
call_6_10_8_13, \
call_6_10_8_14, \
call_6_10_8_15, \
call_6_10_9_0, \
call_6_10_9_1, \
call_6_10_9_2, \
call_6_10_9_3, \
call_6_10_9_4, \
call_6_10_9_5, \
call_6_10_9_6, \
call_6_10_9_7, \
call_6_10_9_8, \
call_6_10_9_9, \
call_6_10_9_10, \
call_6_10_9_11, \
call_6_10_9_12, \
call_6_10_9_13, \
call_6_10_9_14, \
call_6_10_9_15, \
call_6_10_10_0, \
call_6_10_10_1, \
call_6_10_10_2, \
call_6_10_10_3, \
call_6_10_10_4, \
call_6_10_10_5, \
call_6_10_10_6, \
call_6_10_10_7, \
call_6_10_10_8, \
call_6_10_10_9, \
call_6_10_10_10, \
call_6_10_10_11, \
call_6_10_10_12, \
call_6_10_10_13, \
call_6_10_10_14, \
call_6_10_10_15, \
call_6_10_11_0, \
call_6_10_11_1, \
call_6_10_11_2, \
call_6_10_11_3, \
call_6_10_11_4, \
call_6_10_11_5, \
call_6_10_11_6, \
call_6_10_11_7, \
call_6_10_11_8, \
call_6_10_11_9, \
call_6_10_11_10, \
call_6_10_11_11, \
call_6_10_11_12, \
call_6_10_11_13, \
call_6_10_11_14, \
call_6_10_11_15, \
call_6_10_12_0, \
call_6_10_12_1, \
call_6_10_12_2, \
call_6_10_12_3, \
call_6_10_12_4, \
call_6_10_12_5, \
call_6_10_12_6, \
call_6_10_12_7, \
call_6_10_12_8, \
call_6_10_12_9, \
call_6_10_12_10, \
call_6_10_12_11, \
call_6_10_12_12, \
call_6_10_12_13, \
call_6_10_12_14, \
call_6_10_12_15, \
call_6_10_13_0, \
call_6_10_13_1, \
call_6_10_13_2, \
call_6_10_13_3, \
call_6_10_13_4, \
call_6_10_13_5, \
call_6_10_13_6, \
call_6_10_13_7, \
call_6_10_13_8, \
call_6_10_13_9, \
call_6_10_13_10, \
call_6_10_13_11, \
call_6_10_13_12, \
call_6_10_13_13, \
call_6_10_13_14, \
call_6_10_13_15, \
call_6_10_14_0, \
call_6_10_14_1, \
call_6_10_14_2, \
call_6_10_14_3, \
call_6_10_14_4, \
call_6_10_14_5, \
call_6_10_14_6, \
call_6_10_14_7, \
call_6_10_14_8, \
call_6_10_14_9, \
call_6_10_14_10, \
call_6_10_14_11, \
call_6_10_14_12, \
call_6_10_14_13, \
call_6_10_14_14, \
call_6_10_14_15, \
call_6_10_15_0, \
call_6_10_15_1, \
call_6_10_15_2, \
call_6_10_15_3, \
call_6_10_15_4, \
call_6_10_15_5, \
call_6_10_15_6, \
call_6_10_15_7, \
call_6_10_15_8, \
call_6_10_15_9, \
call_6_10_15_10, \
call_6_10_15_11, \
call_6_10_15_12, \
call_6_10_15_13, \
call_6_10_15_14, \
call_6_10_15_15, \
call_6_11_0_0, \
call_6_11_0_1, \
call_6_11_0_2, \
call_6_11_0_3, \
call_6_11_0_4, \
call_6_11_0_5, \
call_6_11_0_6, \
call_6_11_0_7, \
call_6_11_0_8, \
call_6_11_0_9, \
call_6_11_0_10, \
call_6_11_0_11, \
call_6_11_0_12, \
call_6_11_0_13, \
call_6_11_0_14, \
call_6_11_0_15, \
call_6_11_1_0, \
call_6_11_1_1, \
call_6_11_1_2, \
call_6_11_1_3, \
call_6_11_1_4, \
call_6_11_1_5, \
call_6_11_1_6, \
call_6_11_1_7, \
call_6_11_1_8, \
call_6_11_1_9, \
call_6_11_1_10, \
call_6_11_1_11, \
call_6_11_1_12, \
call_6_11_1_13, \
call_6_11_1_14, \
call_6_11_1_15, \
call_6_11_2_0, \
call_6_11_2_1, \
call_6_11_2_2, \
call_6_11_2_3, \
call_6_11_2_4, \
call_6_11_2_5, \
call_6_11_2_6, \
call_6_11_2_7, \
call_6_11_2_8, \
call_6_11_2_9, \
call_6_11_2_10, \
call_6_11_2_11, \
call_6_11_2_12, \
call_6_11_2_13, \
call_6_11_2_14, \
call_6_11_2_15, \
call_6_11_3_0, \
call_6_11_3_1, \
call_6_11_3_2, \
call_6_11_3_3, \
call_6_11_3_4, \
call_6_11_3_5, \
call_6_11_3_6, \
call_6_11_3_7, \
call_6_11_3_8, \
call_6_11_3_9, \
call_6_11_3_10, \
call_6_11_3_11, \
call_6_11_3_12, \
call_6_11_3_13, \
call_6_11_3_14, \
call_6_11_3_15, \
call_6_11_4_0, \
call_6_11_4_1, \
call_6_11_4_2, \
call_6_11_4_3, \
call_6_11_4_4, \
call_6_11_4_5, \
call_6_11_4_6, \
call_6_11_4_7, \
call_6_11_4_8, \
call_6_11_4_9, \
call_6_11_4_10, \
call_6_11_4_11, \
call_6_11_4_12, \
call_6_11_4_13, \
call_6_11_4_14, \
call_6_11_4_15, \
call_6_11_5_0, \
call_6_11_5_1, \
call_6_11_5_2, \
call_6_11_5_3, \
call_6_11_5_4, \
call_6_11_5_5, \
call_6_11_5_6, \
call_6_11_5_7, \
call_6_11_5_8, \
call_6_11_5_9, \
call_6_11_5_10, \
call_6_11_5_11, \
call_6_11_5_12, \
call_6_11_5_13, \
call_6_11_5_14, \
call_6_11_5_15, \
call_6_11_6_0, \
call_6_11_6_1, \
call_6_11_6_2, \
call_6_11_6_3, \
call_6_11_6_4, \
call_6_11_6_5, \
call_6_11_6_6, \
call_6_11_6_7, \
call_6_11_6_8, \
call_6_11_6_9, \
call_6_11_6_10, \
call_6_11_6_11, \
call_6_11_6_12, \
call_6_11_6_13, \
call_6_11_6_14, \
call_6_11_6_15, \
call_6_11_7_0, \
call_6_11_7_1, \
call_6_11_7_2, \
call_6_11_7_3, \
call_6_11_7_4, \
call_6_11_7_5, \
call_6_11_7_6, \
call_6_11_7_7, \
call_6_11_7_8, \
call_6_11_7_9, \
call_6_11_7_10, \
call_6_11_7_11, \
call_6_11_7_12, \
call_6_11_7_13, \
call_6_11_7_14, \
call_6_11_7_15, \
call_6_11_8_0, \
call_6_11_8_1, \
call_6_11_8_2, \
call_6_11_8_3, \
call_6_11_8_4, \
call_6_11_8_5, \
call_6_11_8_6, \
call_6_11_8_7, \
call_6_11_8_8, \
call_6_11_8_9, \
call_6_11_8_10, \
call_6_11_8_11, \
call_6_11_8_12, \
call_6_11_8_13, \
call_6_11_8_14, \
call_6_11_8_15, \
call_6_11_9_0, \
call_6_11_9_1, \
call_6_11_9_2, \
call_6_11_9_3, \
call_6_11_9_4, \
call_6_11_9_5, \
call_6_11_9_6, \
call_6_11_9_7, \
call_6_11_9_8, \
call_6_11_9_9, \
call_6_11_9_10, \
call_6_11_9_11, \
call_6_11_9_12, \
call_6_11_9_13, \
call_6_11_9_14, \
call_6_11_9_15, \
call_6_11_10_0, \
call_6_11_10_1, \
call_6_11_10_2, \
call_6_11_10_3, \
call_6_11_10_4, \
call_6_11_10_5, \
call_6_11_10_6, \
call_6_11_10_7, \
call_6_11_10_8, \
call_6_11_10_9, \
call_6_11_10_10, \
call_6_11_10_11, \
call_6_11_10_12, \
call_6_11_10_13, \
call_6_11_10_14, \
call_6_11_10_15, \
call_6_11_11_0, \
call_6_11_11_1, \
call_6_11_11_2, \
call_6_11_11_3, \
call_6_11_11_4, \
call_6_11_11_5, \
call_6_11_11_6, \
call_6_11_11_7, \
call_6_11_11_8, \
call_6_11_11_9, \
call_6_11_11_10, \
call_6_11_11_11, \
call_6_11_11_12, \
call_6_11_11_13, \
call_6_11_11_14, \
call_6_11_11_15, \
call_6_11_12_0, \
call_6_11_12_1, \
call_6_11_12_2, \
call_6_11_12_3, \
call_6_11_12_4, \
call_6_11_12_5, \
call_6_11_12_6, \
call_6_11_12_7, \
call_6_11_12_8, \
call_6_11_12_9, \
call_6_11_12_10, \
call_6_11_12_11, \
call_6_11_12_12, \
call_6_11_12_13, \
call_6_11_12_14, \
call_6_11_12_15, \
call_6_11_13_0, \
call_6_11_13_1, \
call_6_11_13_2, \
call_6_11_13_3, \
call_6_11_13_4, \
call_6_11_13_5, \
call_6_11_13_6, \
call_6_11_13_7, \
call_6_11_13_8, \
call_6_11_13_9, \
call_6_11_13_10, \
call_6_11_13_11, \
call_6_11_13_12, \
call_6_11_13_13, \
call_6_11_13_14, \
call_6_11_13_15, \
call_6_11_14_0, \
call_6_11_14_1, \
call_6_11_14_2, \
call_6_11_14_3, \
call_6_11_14_4, \
call_6_11_14_5, \
call_6_11_14_6, \
call_6_11_14_7, \
call_6_11_14_8, \
call_6_11_14_9, \
call_6_11_14_10, \
call_6_11_14_11, \
call_6_11_14_12, \
call_6_11_14_13, \
call_6_11_14_14, \
call_6_11_14_15, \
call_6_11_15_0, \
call_6_11_15_1, \
call_6_11_15_2, \
call_6_11_15_3, \
call_6_11_15_4, \
call_6_11_15_5, \
call_6_11_15_6, \
call_6_11_15_7, \
call_6_11_15_8, \
call_6_11_15_9, \
call_6_11_15_10, \
call_6_11_15_11, \
call_6_11_15_12, \
call_6_11_15_13, \
call_6_11_15_14, \
call_6_11_15_15, \
call_6_12_0_0, \
call_6_12_0_1, \
call_6_12_0_2, \
call_6_12_0_3, \
call_6_12_0_4, \
call_6_12_0_5, \
call_6_12_0_6, \
call_6_12_0_7, \
call_6_12_0_8, \
call_6_12_0_9, \
call_6_12_0_10, \
call_6_12_0_11, \
call_6_12_0_12, \
call_6_12_0_13, \
call_6_12_0_14, \
call_6_12_0_15, \
call_6_12_1_0, \
call_6_12_1_1, \
call_6_12_1_2, \
call_6_12_1_3, \
call_6_12_1_4, \
call_6_12_1_5, \
call_6_12_1_6, \
call_6_12_1_7, \
call_6_12_1_8, \
call_6_12_1_9, \
call_6_12_1_10, \
call_6_12_1_11, \
call_6_12_1_12, \
call_6_12_1_13, \
call_6_12_1_14, \
call_6_12_1_15, \
call_6_12_2_0, \
call_6_12_2_1, \
call_6_12_2_2, \
call_6_12_2_3, \
call_6_12_2_4, \
call_6_12_2_5, \
call_6_12_2_6, \
call_6_12_2_7, \
call_6_12_2_8, \
call_6_12_2_9, \
call_6_12_2_10, \
call_6_12_2_11, \
call_6_12_2_12, \
call_6_12_2_13, \
call_6_12_2_14, \
call_6_12_2_15, \
call_6_12_3_0, \
call_6_12_3_1, \
call_6_12_3_2, \
call_6_12_3_3, \
call_6_12_3_4, \
call_6_12_3_5, \
call_6_12_3_6, \
call_6_12_3_7, \
call_6_12_3_8, \
call_6_12_3_9, \
call_6_12_3_10, \
call_6_12_3_11, \
call_6_12_3_12, \
call_6_12_3_13, \
call_6_12_3_14, \
call_6_12_3_15, \
call_6_12_4_0, \
call_6_12_4_1, \
call_6_12_4_2, \
call_6_12_4_3, \
call_6_12_4_4, \
call_6_12_4_5, \
call_6_12_4_6, \
call_6_12_4_7, \
call_6_12_4_8, \
call_6_12_4_9, \
call_6_12_4_10, \
call_6_12_4_11, \
call_6_12_4_12, \
call_6_12_4_13, \
call_6_12_4_14, \
call_6_12_4_15, \
call_6_12_5_0, \
call_6_12_5_1, \
call_6_12_5_2, \
call_6_12_5_3, \
call_6_12_5_4, \
call_6_12_5_5, \
call_6_12_5_6, \
call_6_12_5_7, \
call_6_12_5_8, \
call_6_12_5_9, \
call_6_12_5_10, \
call_6_12_5_11, \
call_6_12_5_12, \
call_6_12_5_13, \
call_6_12_5_14, \
call_6_12_5_15, \
call_6_12_6_0, \
call_6_12_6_1, \
call_6_12_6_2, \
call_6_12_6_3, \
call_6_12_6_4, \
call_6_12_6_5, \
call_6_12_6_6, \
call_6_12_6_7, \
call_6_12_6_8, \
call_6_12_6_9, \
call_6_12_6_10, \
call_6_12_6_11, \
call_6_12_6_12, \
call_6_12_6_13, \
call_6_12_6_14, \
call_6_12_6_15, \
call_6_12_7_0, \
call_6_12_7_1, \
call_6_12_7_2, \
call_6_12_7_3, \
call_6_12_7_4, \
call_6_12_7_5, \
call_6_12_7_6, \
call_6_12_7_7, \
call_6_12_7_8, \
call_6_12_7_9, \
call_6_12_7_10, \
call_6_12_7_11, \
call_6_12_7_12, \
call_6_12_7_13, \
call_6_12_7_14, \
call_6_12_7_15, \
call_6_12_8_0, \
call_6_12_8_1, \
call_6_12_8_2, \
call_6_12_8_3, \
call_6_12_8_4, \
call_6_12_8_5, \
call_6_12_8_6, \
call_6_12_8_7, \
call_6_12_8_8, \
call_6_12_8_9, \
call_6_12_8_10, \
call_6_12_8_11, \
call_6_12_8_12, \
call_6_12_8_13, \
call_6_12_8_14, \
call_6_12_8_15, \
call_6_12_9_0, \
call_6_12_9_1, \
call_6_12_9_2, \
call_6_12_9_3, \
call_6_12_9_4, \
call_6_12_9_5, \
call_6_12_9_6, \
call_6_12_9_7, \
call_6_12_9_8, \
call_6_12_9_9, \
call_6_12_9_10, \
call_6_12_9_11, \
call_6_12_9_12, \
call_6_12_9_13, \
call_6_12_9_14, \
call_6_12_9_15, \
call_6_12_10_0, \
call_6_12_10_1, \
call_6_12_10_2, \
call_6_12_10_3, \
call_6_12_10_4, \
call_6_12_10_5, \
call_6_12_10_6, \
call_6_12_10_7, \
call_6_12_10_8, \
call_6_12_10_9, \
call_6_12_10_10, \
call_6_12_10_11, \
call_6_12_10_12, \
call_6_12_10_13, \
call_6_12_10_14, \
call_6_12_10_15, \
call_6_12_11_0, \
call_6_12_11_1, \
call_6_12_11_2, \
call_6_12_11_3, \
call_6_12_11_4, \
call_6_12_11_5, \
call_6_12_11_6, \
call_6_12_11_7, \
call_6_12_11_8, \
call_6_12_11_9, \
call_6_12_11_10, \
call_6_12_11_11, \
call_6_12_11_12, \
call_6_12_11_13, \
call_6_12_11_14, \
call_6_12_11_15, \
call_6_12_12_0, \
call_6_12_12_1, \
call_6_12_12_2, \
call_6_12_12_3, \
call_6_12_12_4, \
call_6_12_12_5, \
call_6_12_12_6, \
call_6_12_12_7, \
call_6_12_12_8, \
call_6_12_12_9, \
call_6_12_12_10, \
call_6_12_12_11, \
call_6_12_12_12, \
call_6_12_12_13, \
call_6_12_12_14, \
call_6_12_12_15, \
call_6_12_13_0, \
call_6_12_13_1, \
call_6_12_13_2, \
call_6_12_13_3, \
call_6_12_13_4, \
call_6_12_13_5, \
call_6_12_13_6, \
call_6_12_13_7, \
call_6_12_13_8, \
call_6_12_13_9, \
call_6_12_13_10, \
call_6_12_13_11, \
call_6_12_13_12, \
call_6_12_13_13, \
call_6_12_13_14, \
call_6_12_13_15, \
call_6_12_14_0, \
call_6_12_14_1, \
call_6_12_14_2, \
call_6_12_14_3, \
call_6_12_14_4, \
call_6_12_14_5, \
call_6_12_14_6, \
call_6_12_14_7, \
call_6_12_14_8, \
call_6_12_14_9, \
call_6_12_14_10, \
call_6_12_14_11, \
call_6_12_14_12, \
call_6_12_14_13, \
call_6_12_14_14, \
call_6_12_14_15, \
call_6_12_15_0, \
call_6_12_15_1, \
call_6_12_15_2, \
call_6_12_15_3, \
call_6_12_15_4, \
call_6_12_15_5, \
call_6_12_15_6, \
call_6_12_15_7, \
call_6_12_15_8, \
call_6_12_15_9, \
call_6_12_15_10, \
call_6_12_15_11, \
call_6_12_15_12, \
call_6_12_15_13, \
call_6_12_15_14, \
call_6_12_15_15, \
call_6_13_0_0, \
call_6_13_0_1, \
call_6_13_0_2, \
call_6_13_0_3, \
call_6_13_0_4, \
call_6_13_0_5, \
call_6_13_0_6, \
call_6_13_0_7, \
call_6_13_0_8, \
call_6_13_0_9, \
call_6_13_0_10, \
call_6_13_0_11, \
call_6_13_0_12, \
call_6_13_0_13, \
call_6_13_0_14, \
call_6_13_0_15, \
call_6_13_1_0, \
call_6_13_1_1, \
call_6_13_1_2, \
call_6_13_1_3, \
call_6_13_1_4, \
call_6_13_1_5, \
call_6_13_1_6, \
call_6_13_1_7, \
call_6_13_1_8, \
call_6_13_1_9, \
call_6_13_1_10, \
call_6_13_1_11, \
call_6_13_1_12, \
call_6_13_1_13, \
call_6_13_1_14, \
call_6_13_1_15, \
call_6_13_2_0, \
call_6_13_2_1, \
call_6_13_2_2, \
call_6_13_2_3, \
call_6_13_2_4, \
call_6_13_2_5, \
call_6_13_2_6, \
call_6_13_2_7, \
call_6_13_2_8, \
call_6_13_2_9, \
call_6_13_2_10, \
call_6_13_2_11, \
call_6_13_2_12, \
call_6_13_2_13, \
call_6_13_2_14, \
call_6_13_2_15, \
call_6_13_3_0, \
call_6_13_3_1, \
call_6_13_3_2, \
call_6_13_3_3, \
call_6_13_3_4, \
call_6_13_3_5, \
call_6_13_3_6, \
call_6_13_3_7, \
call_6_13_3_8, \
call_6_13_3_9, \
call_6_13_3_10, \
call_6_13_3_11, \
call_6_13_3_12, \
call_6_13_3_13, \
call_6_13_3_14, \
call_6_13_3_15, \
call_6_13_4_0, \
call_6_13_4_1, \
call_6_13_4_2, \
call_6_13_4_3, \
call_6_13_4_4, \
call_6_13_4_5, \
call_6_13_4_6, \
call_6_13_4_7, \
call_6_13_4_8, \
call_6_13_4_9, \
call_6_13_4_10, \
call_6_13_4_11, \
call_6_13_4_12, \
call_6_13_4_13, \
call_6_13_4_14, \
call_6_13_4_15, \
call_6_13_5_0, \
call_6_13_5_1, \
call_6_13_5_2, \
call_6_13_5_3, \
call_6_13_5_4, \
call_6_13_5_5, \
call_6_13_5_6, \
call_6_13_5_7, \
call_6_13_5_8, \
call_6_13_5_9, \
call_6_13_5_10, \
call_6_13_5_11, \
call_6_13_5_12, \
call_6_13_5_13, \
call_6_13_5_14, \
call_6_13_5_15, \
call_6_13_6_0, \
call_6_13_6_1, \
call_6_13_6_2, \
call_6_13_6_3, \
call_6_13_6_4, \
call_6_13_6_5, \
call_6_13_6_6, \
call_6_13_6_7, \
call_6_13_6_8, \
call_6_13_6_9, \
call_6_13_6_10, \
call_6_13_6_11, \
call_6_13_6_12, \
call_6_13_6_13, \
call_6_13_6_14, \
call_6_13_6_15, \
call_6_13_7_0, \
call_6_13_7_1, \
call_6_13_7_2, \
call_6_13_7_3, \
call_6_13_7_4, \
call_6_13_7_5, \
call_6_13_7_6, \
call_6_13_7_7, \
call_6_13_7_8, \
call_6_13_7_9, \
call_6_13_7_10, \
call_6_13_7_11, \
call_6_13_7_12, \
call_6_13_7_13, \
call_6_13_7_14, \
call_6_13_7_15, \
call_6_13_8_0, \
call_6_13_8_1, \
call_6_13_8_2, \
call_6_13_8_3, \
call_6_13_8_4, \
call_6_13_8_5, \
call_6_13_8_6, \
call_6_13_8_7, \
call_6_13_8_8, \
call_6_13_8_9, \
call_6_13_8_10, \
call_6_13_8_11, \
call_6_13_8_12, \
call_6_13_8_13, \
call_6_13_8_14, \
call_6_13_8_15, \
call_6_13_9_0, \
call_6_13_9_1, \
call_6_13_9_2, \
call_6_13_9_3, \
call_6_13_9_4, \
call_6_13_9_5, \
call_6_13_9_6, \
call_6_13_9_7, \
call_6_13_9_8, \
call_6_13_9_9, \
call_6_13_9_10, \
call_6_13_9_11, \
call_6_13_9_12, \
call_6_13_9_13, \
call_6_13_9_14, \
call_6_13_9_15, \
call_6_13_10_0, \
call_6_13_10_1, \
call_6_13_10_2, \
call_6_13_10_3, \
call_6_13_10_4, \
call_6_13_10_5, \
call_6_13_10_6, \
call_6_13_10_7, \
call_6_13_10_8, \
call_6_13_10_9, \
call_6_13_10_10, \
call_6_13_10_11, \
call_6_13_10_12, \
call_6_13_10_13, \
call_6_13_10_14, \
call_6_13_10_15, \
call_6_13_11_0, \
call_6_13_11_1, \
call_6_13_11_2, \
call_6_13_11_3, \
call_6_13_11_4, \
call_6_13_11_5, \
call_6_13_11_6, \
call_6_13_11_7, \
call_6_13_11_8, \
call_6_13_11_9, \
call_6_13_11_10, \
call_6_13_11_11, \
call_6_13_11_12, \
call_6_13_11_13, \
call_6_13_11_14, \
call_6_13_11_15, \
call_6_13_12_0, \
call_6_13_12_1, \
call_6_13_12_2, \
call_6_13_12_3, \
call_6_13_12_4, \
call_6_13_12_5, \
call_6_13_12_6, \
call_6_13_12_7, \
call_6_13_12_8, \
call_6_13_12_9, \
call_6_13_12_10, \
call_6_13_12_11, \
call_6_13_12_12, \
call_6_13_12_13, \
call_6_13_12_14, \
call_6_13_12_15, \
call_6_13_13_0, \
call_6_13_13_1, \
call_6_13_13_2, \
call_6_13_13_3, \
call_6_13_13_4, \
call_6_13_13_5, \
call_6_13_13_6, \
call_6_13_13_7, \
call_6_13_13_8, \
call_6_13_13_9, \
call_6_13_13_10, \
call_6_13_13_11, \
call_6_13_13_12, \
call_6_13_13_13, \
call_6_13_13_14, \
call_6_13_13_15, \
call_6_13_14_0, \
call_6_13_14_1, \
call_6_13_14_2, \
call_6_13_14_3, \
call_6_13_14_4, \
call_6_13_14_5, \
call_6_13_14_6, \
call_6_13_14_7, \
call_6_13_14_8, \
call_6_13_14_9, \
call_6_13_14_10, \
call_6_13_14_11, \
call_6_13_14_12, \
call_6_13_14_13, \
call_6_13_14_14, \
call_6_13_14_15, \
call_6_13_15_0, \
call_6_13_15_1, \
call_6_13_15_2, \
call_6_13_15_3, \
call_6_13_15_4, \
call_6_13_15_5, \
call_6_13_15_6, \
call_6_13_15_7, \
call_6_13_15_8, \
call_6_13_15_9, \
call_6_13_15_10, \
call_6_13_15_11, \
call_6_13_15_12, \
call_6_13_15_13, \
call_6_13_15_14, \
call_6_13_15_15, \
call_6_14_0_0, \
call_6_14_0_1, \
call_6_14_0_2, \
call_6_14_0_3, \
call_6_14_0_4, \
call_6_14_0_5, \
call_6_14_0_6, \
call_6_14_0_7, \
call_6_14_0_8, \
call_6_14_0_9, \
call_6_14_0_10, \
call_6_14_0_11, \
call_6_14_0_12, \
call_6_14_0_13, \
call_6_14_0_14, \
call_6_14_0_15, \
call_6_14_1_0, \
call_6_14_1_1, \
call_6_14_1_2, \
call_6_14_1_3, \
call_6_14_1_4, \
call_6_14_1_5, \
call_6_14_1_6, \
call_6_14_1_7, \
call_6_14_1_8, \
call_6_14_1_9, \
call_6_14_1_10, \
call_6_14_1_11, \
call_6_14_1_12, \
call_6_14_1_13, \
call_6_14_1_14, \
call_6_14_1_15, \
call_6_14_2_0, \
call_6_14_2_1, \
call_6_14_2_2, \
call_6_14_2_3, \
call_6_14_2_4, \
call_6_14_2_5, \
call_6_14_2_6, \
call_6_14_2_7, \
call_6_14_2_8, \
call_6_14_2_9, \
call_6_14_2_10, \
call_6_14_2_11, \
call_6_14_2_12, \
call_6_14_2_13, \
call_6_14_2_14, \
call_6_14_2_15, \
call_6_14_3_0, \
call_6_14_3_1, \
call_6_14_3_2, \
call_6_14_3_3, \
call_6_14_3_4, \
call_6_14_3_5, \
call_6_14_3_6, \
call_6_14_3_7, \
call_6_14_3_8, \
call_6_14_3_9, \
call_6_14_3_10, \
call_6_14_3_11, \
call_6_14_3_12, \
call_6_14_3_13, \
call_6_14_3_14, \
call_6_14_3_15, \
call_6_14_4_0, \
call_6_14_4_1, \
call_6_14_4_2, \
call_6_14_4_3, \
call_6_14_4_4, \
call_6_14_4_5, \
call_6_14_4_6, \
call_6_14_4_7, \
call_6_14_4_8, \
call_6_14_4_9, \
call_6_14_4_10, \
call_6_14_4_11, \
call_6_14_4_12, \
call_6_14_4_13, \
call_6_14_4_14, \
call_6_14_4_15, \
call_6_14_5_0, \
call_6_14_5_1, \
call_6_14_5_2, \
call_6_14_5_3, \
call_6_14_5_4, \
call_6_14_5_5, \
call_6_14_5_6, \
call_6_14_5_7, \
call_6_14_5_8, \
call_6_14_5_9, \
call_6_14_5_10, \
call_6_14_5_11, \
call_6_14_5_12, \
call_6_14_5_13, \
call_6_14_5_14, \
call_6_14_5_15, \
call_6_14_6_0, \
call_6_14_6_1, \
call_6_14_6_2, \
call_6_14_6_3, \
call_6_14_6_4, \
call_6_14_6_5, \
call_6_14_6_6, \
call_6_14_6_7, \
call_6_14_6_8, \
call_6_14_6_9, \
call_6_14_6_10, \
call_6_14_6_11, \
call_6_14_6_12, \
call_6_14_6_13, \
call_6_14_6_14, \
call_6_14_6_15, \
call_6_14_7_0, \
call_6_14_7_1, \
call_6_14_7_2, \
call_6_14_7_3, \
call_6_14_7_4, \
call_6_14_7_5, \
call_6_14_7_6, \
call_6_14_7_7, \
call_6_14_7_8, \
call_6_14_7_9, \
call_6_14_7_10, \
call_6_14_7_11, \
call_6_14_7_12, \
call_6_14_7_13, \
call_6_14_7_14, \
call_6_14_7_15, \
call_6_14_8_0, \
call_6_14_8_1, \
call_6_14_8_2, \
call_6_14_8_3, \
call_6_14_8_4, \
call_6_14_8_5, \
call_6_14_8_6, \
call_6_14_8_7, \
call_6_14_8_8, \
call_6_14_8_9, \
call_6_14_8_10, \
call_6_14_8_11, \
call_6_14_8_12, \
call_6_14_8_13, \
call_6_14_8_14, \
call_6_14_8_15, \
call_6_14_9_0, \
call_6_14_9_1, \
call_6_14_9_2, \
call_6_14_9_3, \
call_6_14_9_4, \
call_6_14_9_5, \
call_6_14_9_6, \
call_6_14_9_7, \
call_6_14_9_8, \
call_6_14_9_9, \
call_6_14_9_10, \
call_6_14_9_11, \
call_6_14_9_12, \
call_6_14_9_13, \
call_6_14_9_14, \
call_6_14_9_15, \
call_6_14_10_0, \
call_6_14_10_1, \
call_6_14_10_2, \
call_6_14_10_3, \
call_6_14_10_4, \
call_6_14_10_5, \
call_6_14_10_6, \
call_6_14_10_7, \
call_6_14_10_8, \
call_6_14_10_9, \
call_6_14_10_10, \
call_6_14_10_11, \
call_6_14_10_12, \
call_6_14_10_13, \
call_6_14_10_14, \
call_6_14_10_15, \
call_6_14_11_0, \
call_6_14_11_1, \
call_6_14_11_2, \
call_6_14_11_3, \
call_6_14_11_4, \
call_6_14_11_5, \
call_6_14_11_6, \
call_6_14_11_7, \
call_6_14_11_8, \
call_6_14_11_9, \
call_6_14_11_10, \
call_6_14_11_11, \
call_6_14_11_12, \
call_6_14_11_13, \
call_6_14_11_14, \
call_6_14_11_15, \
call_6_14_12_0, \
call_6_14_12_1, \
call_6_14_12_2, \
call_6_14_12_3, \
call_6_14_12_4, \
call_6_14_12_5, \
call_6_14_12_6, \
call_6_14_12_7, \
call_6_14_12_8, \
call_6_14_12_9, \
call_6_14_12_10, \
call_6_14_12_11, \
call_6_14_12_12, \
call_6_14_12_13, \
call_6_14_12_14, \
call_6_14_12_15, \
call_6_14_13_0, \
call_6_14_13_1, \
call_6_14_13_2, \
call_6_14_13_3, \
call_6_14_13_4, \
call_6_14_13_5, \
call_6_14_13_6, \
call_6_14_13_7, \
call_6_14_13_8, \
call_6_14_13_9, \
call_6_14_13_10, \
call_6_14_13_11, \
call_6_14_13_12, \
call_6_14_13_13, \
call_6_14_13_14, \
call_6_14_13_15, \
call_6_14_14_0, \
call_6_14_14_1, \
call_6_14_14_2, \
call_6_14_14_3, \
call_6_14_14_4, \
call_6_14_14_5, \
call_6_14_14_6, \
call_6_14_14_7, \
call_6_14_14_8, \
call_6_14_14_9, \
call_6_14_14_10, \
call_6_14_14_11, \
call_6_14_14_12, \
call_6_14_14_13, \
call_6_14_14_14, \
call_6_14_14_15, \
call_6_14_15_0, \
call_6_14_15_1, \
call_6_14_15_2, \
call_6_14_15_3, \
call_6_14_15_4, \
call_6_14_15_5, \
call_6_14_15_6, \
call_6_14_15_7, \
call_6_14_15_8, \
call_6_14_15_9, \
call_6_14_15_10, \
call_6_14_15_11, \
call_6_14_15_12, \
call_6_14_15_13, \
call_6_14_15_14, \
call_6_14_15_15, \
call_6_15_0_0, \
call_6_15_0_1, \
call_6_15_0_2, \
call_6_15_0_3, \
call_6_15_0_4, \
call_6_15_0_5, \
call_6_15_0_6, \
call_6_15_0_7, \
call_6_15_0_8, \
call_6_15_0_9, \
call_6_15_0_10, \
call_6_15_0_11, \
call_6_15_0_12, \
call_6_15_0_13, \
call_6_15_0_14, \
call_6_15_0_15, \
call_6_15_1_0, \
call_6_15_1_1, \
call_6_15_1_2, \
call_6_15_1_3, \
call_6_15_1_4, \
call_6_15_1_5, \
call_6_15_1_6, \
call_6_15_1_7, \
call_6_15_1_8, \
call_6_15_1_9, \
call_6_15_1_10, \
call_6_15_1_11, \
call_6_15_1_12, \
call_6_15_1_13, \
call_6_15_1_14, \
call_6_15_1_15, \
call_6_15_2_0, \
call_6_15_2_1, \
call_6_15_2_2, \
call_6_15_2_3, \
call_6_15_2_4, \
call_6_15_2_5, \
call_6_15_2_6, \
call_6_15_2_7, \
call_6_15_2_8, \
call_6_15_2_9, \
call_6_15_2_10, \
call_6_15_2_11, \
call_6_15_2_12, \
call_6_15_2_13, \
call_6_15_2_14, \
call_6_15_2_15, \
call_6_15_3_0, \
call_6_15_3_1, \
call_6_15_3_2, \
call_6_15_3_3, \
call_6_15_3_4, \
call_6_15_3_5, \
call_6_15_3_6, \
call_6_15_3_7, \
call_6_15_3_8, \
call_6_15_3_9, \
call_6_15_3_10, \
call_6_15_3_11, \
call_6_15_3_12, \
call_6_15_3_13, \
call_6_15_3_14, \
call_6_15_3_15, \
call_6_15_4_0, \
call_6_15_4_1, \
call_6_15_4_2, \
call_6_15_4_3, \
call_6_15_4_4, \
call_6_15_4_5, \
call_6_15_4_6, \
call_6_15_4_7, \
call_6_15_4_8, \
call_6_15_4_9, \
call_6_15_4_10, \
call_6_15_4_11, \
call_6_15_4_12, \
call_6_15_4_13, \
call_6_15_4_14, \
call_6_15_4_15, \
call_6_15_5_0, \
call_6_15_5_1, \
call_6_15_5_2, \
call_6_15_5_3, \
call_6_15_5_4, \
call_6_15_5_5, \
call_6_15_5_6, \
call_6_15_5_7, \
call_6_15_5_8, \
call_6_15_5_9, \
call_6_15_5_10, \
call_6_15_5_11, \
call_6_15_5_12, \
call_6_15_5_13, \
call_6_15_5_14, \
call_6_15_5_15, \
call_6_15_6_0, \
call_6_15_6_1, \
call_6_15_6_2, \
call_6_15_6_3, \
call_6_15_6_4, \
call_6_15_6_5, \
call_6_15_6_6, \
call_6_15_6_7, \
call_6_15_6_8, \
call_6_15_6_9, \
call_6_15_6_10, \
call_6_15_6_11, \
call_6_15_6_12, \
call_6_15_6_13, \
call_6_15_6_14, \
call_6_15_6_15, \
call_6_15_7_0, \
call_6_15_7_1, \
call_6_15_7_2, \
call_6_15_7_3, \
call_6_15_7_4, \
call_6_15_7_5, \
call_6_15_7_6, \
call_6_15_7_7, \
call_6_15_7_8, \
call_6_15_7_9, \
call_6_15_7_10, \
call_6_15_7_11, \
call_6_15_7_12, \
call_6_15_7_13, \
call_6_15_7_14, \
call_6_15_7_15, \
call_6_15_8_0, \
call_6_15_8_1, \
call_6_15_8_2, \
call_6_15_8_3, \
call_6_15_8_4, \
call_6_15_8_5, \
call_6_15_8_6, \
call_6_15_8_7, \
call_6_15_8_8, \
call_6_15_8_9, \
call_6_15_8_10, \
call_6_15_8_11, \
call_6_15_8_12, \
call_6_15_8_13, \
call_6_15_8_14, \
call_6_15_8_15, \
call_6_15_9_0, \
call_6_15_9_1, \
call_6_15_9_2, \
call_6_15_9_3, \
call_6_15_9_4, \
call_6_15_9_5, \
call_6_15_9_6, \
call_6_15_9_7, \
call_6_15_9_8, \
call_6_15_9_9, \
call_6_15_9_10, \
call_6_15_9_11, \
call_6_15_9_12, \
call_6_15_9_13, \
call_6_15_9_14, \
call_6_15_9_15, \
call_6_15_10_0, \
call_6_15_10_1, \
call_6_15_10_2, \
call_6_15_10_3, \
call_6_15_10_4, \
call_6_15_10_5, \
call_6_15_10_6, \
call_6_15_10_7, \
call_6_15_10_8, \
call_6_15_10_9, \
call_6_15_10_10, \
call_6_15_10_11, \
call_6_15_10_12, \
call_6_15_10_13, \
call_6_15_10_14, \
call_6_15_10_15, \
call_6_15_11_0, \
call_6_15_11_1, \
call_6_15_11_2, \
call_6_15_11_3, \
call_6_15_11_4, \
call_6_15_11_5, \
call_6_15_11_6, \
call_6_15_11_7, \
call_6_15_11_8, \
call_6_15_11_9, \
call_6_15_11_10, \
call_6_15_11_11, \
call_6_15_11_12, \
call_6_15_11_13, \
call_6_15_11_14, \
call_6_15_11_15, \
call_6_15_12_0, \
call_6_15_12_1, \
call_6_15_12_2, \
call_6_15_12_3, \
call_6_15_12_4, \
call_6_15_12_5, \
call_6_15_12_6, \
call_6_15_12_7, \
call_6_15_12_8, \
call_6_15_12_9, \
call_6_15_12_10, \
call_6_15_12_11, \
call_6_15_12_12, \
call_6_15_12_13, \
call_6_15_12_14, \
call_6_15_12_15, \
call_6_15_13_0, \
call_6_15_13_1, \
call_6_15_13_2, \
call_6_15_13_3, \
call_6_15_13_4, \
call_6_15_13_5, \
call_6_15_13_6, \
call_6_15_13_7, \
call_6_15_13_8, \
call_6_15_13_9, \
call_6_15_13_10, \
call_6_15_13_11, \
call_6_15_13_12, \
call_6_15_13_13, \
call_6_15_13_14, \
call_6_15_13_15, \
call_6_15_14_0, \
call_6_15_14_1, \
call_6_15_14_2, \
call_6_15_14_3, \
call_6_15_14_4, \
call_6_15_14_5, \
call_6_15_14_6, \
call_6_15_14_7, \
call_6_15_14_8, \
call_6_15_14_9, \
call_6_15_14_10, \
call_6_15_14_11, \
call_6_15_14_12, \
call_6_15_14_13, \
call_6_15_14_14, \
call_6_15_14_15, \
call_6_15_15_0, \
call_6_15_15_1, \
call_6_15_15_2, \
call_6_15_15_3, \
call_6_15_15_4, \
call_6_15_15_5, \
call_6_15_15_6, \
call_6_15_15_7, \
call_6_15_15_8, \
call_6_15_15_9, \
call_6_15_15_10, \
call_6_15_15_11, \
call_6_15_15_12, \
call_6_15_15_13, \
call_6_15_15_14, \
call_6_15_15_15, \
call_7_0_0_0, \
call_7_0_0_1, \
call_7_0_0_2, \
call_7_0_0_3, \
call_7_0_0_4, \
call_7_0_0_5, \
call_7_0_0_6, \
call_7_0_0_7, \
call_7_0_0_8, \
call_7_0_0_9, \
call_7_0_0_10, \
call_7_0_0_11, \
call_7_0_0_12, \
call_7_0_0_13, \
call_7_0_0_14, \
call_7_0_0_15, \
call_7_0_1_0, \
call_7_0_1_1, \
call_7_0_1_2, \
call_7_0_1_3, \
call_7_0_1_4, \
call_7_0_1_5, \
call_7_0_1_6, \
call_7_0_1_7, \
call_7_0_1_8, \
call_7_0_1_9, \
call_7_0_1_10, \
call_7_0_1_11, \
call_7_0_1_12, \
call_7_0_1_13, \
call_7_0_1_14, \
call_7_0_1_15, \
call_7_0_2_0, \
call_7_0_2_1, \
call_7_0_2_2, \
call_7_0_2_3, \
call_7_0_2_4, \
call_7_0_2_5, \
call_7_0_2_6, \
call_7_0_2_7, \
call_7_0_2_8, \
call_7_0_2_9, \
call_7_0_2_10, \
call_7_0_2_11, \
call_7_0_2_12, \
call_7_0_2_13, \
call_7_0_2_14, \
call_7_0_2_15, \
call_7_0_3_0, \
call_7_0_3_1, \
call_7_0_3_2, \
call_7_0_3_3, \
call_7_0_3_4, \
call_7_0_3_5, \
call_7_0_3_6, \
call_7_0_3_7, \
call_7_0_3_8, \
call_7_0_3_9, \
call_7_0_3_10, \
call_7_0_3_11, \
call_7_0_3_12, \
call_7_0_3_13, \
call_7_0_3_14, \
call_7_0_3_15, \
call_7_0_4_0, \
call_7_0_4_1, \
call_7_0_4_2, \
call_7_0_4_3, \
call_7_0_4_4, \
call_7_0_4_5, \
call_7_0_4_6, \
call_7_0_4_7, \
call_7_0_4_8, \
call_7_0_4_9, \
call_7_0_4_10, \
call_7_0_4_11, \
call_7_0_4_12, \
call_7_0_4_13, \
call_7_0_4_14, \
call_7_0_4_15, \
call_7_0_5_0, \
call_7_0_5_1, \
call_7_0_5_2, \
call_7_0_5_3, \
call_7_0_5_4, \
call_7_0_5_5, \
call_7_0_5_6, \
call_7_0_5_7, \
call_7_0_5_8, \
call_7_0_5_9, \
call_7_0_5_10, \
call_7_0_5_11, \
call_7_0_5_12, \
call_7_0_5_13, \
call_7_0_5_14, \
call_7_0_5_15, \
call_7_0_6_0, \
call_7_0_6_1, \
call_7_0_6_2, \
call_7_0_6_3, \
call_7_0_6_4, \
call_7_0_6_5, \
call_7_0_6_6, \
call_7_0_6_7, \
call_7_0_6_8, \
call_7_0_6_9, \
call_7_0_6_10, \
call_7_0_6_11, \
call_7_0_6_12, \
call_7_0_6_13, \
call_7_0_6_14, \
call_7_0_6_15, \
call_7_0_7_0, \
call_7_0_7_1, \
call_7_0_7_2, \
call_7_0_7_3, \
call_7_0_7_4, \
call_7_0_7_5, \
call_7_0_7_6, \
call_7_0_7_7, \
call_7_0_7_8, \
call_7_0_7_9, \
call_7_0_7_10, \
call_7_0_7_11, \
call_7_0_7_12, \
call_7_0_7_13, \
call_7_0_7_14, \
call_7_0_7_15, \
call_7_0_8_0, \
call_7_0_8_1, \
call_7_0_8_2, \
call_7_0_8_3, \
call_7_0_8_4, \
call_7_0_8_5, \
call_7_0_8_6, \
call_7_0_8_7, \
call_7_0_8_8, \
call_7_0_8_9, \
call_7_0_8_10, \
call_7_0_8_11, \
call_7_0_8_12, \
call_7_0_8_13, \
call_7_0_8_14, \
call_7_0_8_15, \
call_7_0_9_0, \
call_7_0_9_1, \
call_7_0_9_2, \
call_7_0_9_3, \
call_7_0_9_4, \
call_7_0_9_5, \
call_7_0_9_6, \
call_7_0_9_7, \
call_7_0_9_8, \
call_7_0_9_9, \
call_7_0_9_10, \
call_7_0_9_11, \
call_7_0_9_12, \
call_7_0_9_13, \
call_7_0_9_14, \
call_7_0_9_15, \
call_7_0_10_0, \
call_7_0_10_1, \
call_7_0_10_2, \
call_7_0_10_3, \
call_7_0_10_4, \
call_7_0_10_5, \
call_7_0_10_6, \
call_7_0_10_7, \
call_7_0_10_8, \
call_7_0_10_9, \
call_7_0_10_10, \
call_7_0_10_11, \
call_7_0_10_12, \
call_7_0_10_13, \
call_7_0_10_14, \
call_7_0_10_15, \
call_7_0_11_0, \
call_7_0_11_1, \
call_7_0_11_2, \
call_7_0_11_3, \
call_7_0_11_4, \
call_7_0_11_5, \
call_7_0_11_6, \
call_7_0_11_7, \
call_7_0_11_8, \
call_7_0_11_9, \
call_7_0_11_10, \
call_7_0_11_11, \
call_7_0_11_12, \
call_7_0_11_13, \
call_7_0_11_14, \
call_7_0_11_15, \
call_7_0_12_0, \
call_7_0_12_1, \
call_7_0_12_2, \
call_7_0_12_3, \
call_7_0_12_4, \
call_7_0_12_5, \
call_7_0_12_6, \
call_7_0_12_7, \
call_7_0_12_8, \
call_7_0_12_9, \
call_7_0_12_10, \
call_7_0_12_11, \
call_7_0_12_12, \
call_7_0_12_13, \
call_7_0_12_14, \
call_7_0_12_15, \
call_7_0_13_0, \
call_7_0_13_1, \
call_7_0_13_2, \
call_7_0_13_3, \
call_7_0_13_4, \
call_7_0_13_5, \
call_7_0_13_6, \
call_7_0_13_7, \
call_7_0_13_8, \
call_7_0_13_9, \
call_7_0_13_10, \
call_7_0_13_11, \
call_7_0_13_12, \
call_7_0_13_13, \
call_7_0_13_14, \
call_7_0_13_15, \
call_7_0_14_0, \
call_7_0_14_1, \
call_7_0_14_2, \
call_7_0_14_3, \
call_7_0_14_4, \
call_7_0_14_5, \
call_7_0_14_6, \
call_7_0_14_7, \
call_7_0_14_8, \
call_7_0_14_9, \
call_7_0_14_10, \
call_7_0_14_11, \
call_7_0_14_12, \
call_7_0_14_13, \
call_7_0_14_14, \
call_7_0_14_15, \
call_7_0_15_0, \
call_7_0_15_1, \
call_7_0_15_2, \
call_7_0_15_3, \
call_7_0_15_4, \
call_7_0_15_5, \
call_7_0_15_6, \
call_7_0_15_7, \
call_7_0_15_8, \
call_7_0_15_9, \
call_7_0_15_10, \
call_7_0_15_11, \
call_7_0_15_12, \
call_7_0_15_13, \
call_7_0_15_14, \
call_7_0_15_15, \
call_7_1_0_0, \
call_7_1_0_1, \
call_7_1_0_2, \
call_7_1_0_3, \
call_7_1_0_4, \
call_7_1_0_5, \
call_7_1_0_6, \
call_7_1_0_7, \
call_7_1_0_8, \
call_7_1_0_9, \
call_7_1_0_10, \
call_7_1_0_11, \
call_7_1_0_12, \
call_7_1_0_13, \
call_7_1_0_14, \
call_7_1_0_15, \
call_7_1_1_0, \
call_7_1_1_1, \
call_7_1_1_2, \
call_7_1_1_3, \
call_7_1_1_4, \
call_7_1_1_5, \
call_7_1_1_6, \
call_7_1_1_7, \
call_7_1_1_8, \
call_7_1_1_9, \
call_7_1_1_10, \
call_7_1_1_11, \
call_7_1_1_12, \
call_7_1_1_13, \
call_7_1_1_14, \
call_7_1_1_15, \
call_7_1_2_0, \
call_7_1_2_1, \
call_7_1_2_2, \
call_7_1_2_3, \
call_7_1_2_4, \
call_7_1_2_5, \
call_7_1_2_6, \
call_7_1_2_7, \
call_7_1_2_8, \
call_7_1_2_9, \
call_7_1_2_10, \
call_7_1_2_11, \
call_7_1_2_12, \
call_7_1_2_13, \
call_7_1_2_14, \
call_7_1_2_15, \
call_7_1_3_0, \
call_7_1_3_1, \
call_7_1_3_2, \
call_7_1_3_3, \
call_7_1_3_4, \
call_7_1_3_5, \
call_7_1_3_6, \
call_7_1_3_7, \
call_7_1_3_8, \
call_7_1_3_9, \
call_7_1_3_10, \
call_7_1_3_11, \
call_7_1_3_12, \
call_7_1_3_13, \
call_7_1_3_14, \
call_7_1_3_15, \
call_7_1_4_0, \
call_7_1_4_1, \
call_7_1_4_2, \
call_7_1_4_3, \
call_7_1_4_4, \
call_7_1_4_5, \
call_7_1_4_6, \
call_7_1_4_7, \
call_7_1_4_8, \
call_7_1_4_9, \
call_7_1_4_10, \
call_7_1_4_11, \
call_7_1_4_12, \
call_7_1_4_13, \
call_7_1_4_14, \
call_7_1_4_15, \
call_7_1_5_0, \
call_7_1_5_1, \
call_7_1_5_2, \
call_7_1_5_3, \
call_7_1_5_4, \
call_7_1_5_5, \
call_7_1_5_6, \
call_7_1_5_7, \
call_7_1_5_8, \
call_7_1_5_9, \
call_7_1_5_10, \
call_7_1_5_11, \
call_7_1_5_12, \
call_7_1_5_13, \
call_7_1_5_14, \
call_7_1_5_15, \
call_7_1_6_0, \
call_7_1_6_1, \
call_7_1_6_2, \
call_7_1_6_3, \
call_7_1_6_4, \
call_7_1_6_5, \
call_7_1_6_6, \
call_7_1_6_7, \
call_7_1_6_8, \
call_7_1_6_9, \
call_7_1_6_10, \
call_7_1_6_11, \
call_7_1_6_12, \
call_7_1_6_13, \
call_7_1_6_14, \
call_7_1_6_15, \
call_7_1_7_0, \
call_7_1_7_1, \
call_7_1_7_2, \
call_7_1_7_3, \
call_7_1_7_4, \
call_7_1_7_5, \
call_7_1_7_6, \
call_7_1_7_7, \
call_7_1_7_8, \
call_7_1_7_9, \
call_7_1_7_10, \
call_7_1_7_11, \
call_7_1_7_12, \
call_7_1_7_13, \
call_7_1_7_14, \
call_7_1_7_15, \
call_7_1_8_0, \
call_7_1_8_1, \
call_7_1_8_2, \
call_7_1_8_3, \
call_7_1_8_4, \
call_7_1_8_5, \
call_7_1_8_6, \
call_7_1_8_7, \
call_7_1_8_8, \
call_7_1_8_9, \
call_7_1_8_10, \
call_7_1_8_11, \
call_7_1_8_12, \
call_7_1_8_13, \
call_7_1_8_14, \
call_7_1_8_15, \
call_7_1_9_0, \
call_7_1_9_1, \
call_7_1_9_2, \
call_7_1_9_3, \
call_7_1_9_4, \
call_7_1_9_5, \
call_7_1_9_6, \
call_7_1_9_7, \
call_7_1_9_8, \
call_7_1_9_9, \
call_7_1_9_10, \
call_7_1_9_11, \
call_7_1_9_12, \
call_7_1_9_13, \
call_7_1_9_14, \
call_7_1_9_15, \
call_7_1_10_0, \
call_7_1_10_1, \
call_7_1_10_2, \
call_7_1_10_3, \
call_7_1_10_4, \
call_7_1_10_5, \
call_7_1_10_6, \
call_7_1_10_7, \
call_7_1_10_8, \
call_7_1_10_9, \
call_7_1_10_10, \
call_7_1_10_11, \
call_7_1_10_12, \
call_7_1_10_13, \
call_7_1_10_14, \
call_7_1_10_15, \
call_7_1_11_0, \
call_7_1_11_1, \
call_7_1_11_2, \
call_7_1_11_3, \
call_7_1_11_4, \
call_7_1_11_5, \
call_7_1_11_6, \
call_7_1_11_7, \
call_7_1_11_8, \
call_7_1_11_9, \
call_7_1_11_10, \
call_7_1_11_11, \
call_7_1_11_12, \
call_7_1_11_13, \
call_7_1_11_14, \
call_7_1_11_15, \
call_7_1_12_0, \
call_7_1_12_1, \
call_7_1_12_2, \
call_7_1_12_3, \
call_7_1_12_4, \
call_7_1_12_5, \
call_7_1_12_6, \
call_7_1_12_7, \
call_7_1_12_8, \
call_7_1_12_9, \
call_7_1_12_10, \
call_7_1_12_11, \
call_7_1_12_12, \
call_7_1_12_13, \
call_7_1_12_14, \
call_7_1_12_15, \
call_7_1_13_0, \
call_7_1_13_1, \
call_7_1_13_2, \
call_7_1_13_3, \
call_7_1_13_4, \
call_7_1_13_5, \
call_7_1_13_6, \
call_7_1_13_7, \
call_7_1_13_8, \
call_7_1_13_9, \
call_7_1_13_10, \
call_7_1_13_11, \
call_7_1_13_12, \
call_7_1_13_13, \
call_7_1_13_14, \
call_7_1_13_15, \
call_7_1_14_0, \
call_7_1_14_1, \
call_7_1_14_2, \
call_7_1_14_3, \
call_7_1_14_4, \
call_7_1_14_5, \
call_7_1_14_6, \
call_7_1_14_7, \
call_7_1_14_8, \
call_7_1_14_9, \
call_7_1_14_10, \
call_7_1_14_11, \
call_7_1_14_12, \
call_7_1_14_13, \
call_7_1_14_14, \
call_7_1_14_15, \
call_7_1_15_0, \
call_7_1_15_1, \
call_7_1_15_2, \
call_7_1_15_3, \
call_7_1_15_4, \
call_7_1_15_5, \
call_7_1_15_6, \
call_7_1_15_7, \
call_7_1_15_8, \
call_7_1_15_9, \
call_7_1_15_10, \
call_7_1_15_11, \
call_7_1_15_12, \
call_7_1_15_13, \
call_7_1_15_14, \
call_7_1_15_15, \
call_7_2_0_0, \
call_7_2_0_1, \
call_7_2_0_2, \
call_7_2_0_3, \
call_7_2_0_4, \
call_7_2_0_5, \
call_7_2_0_6, \
call_7_2_0_7, \
call_7_2_0_8, \
call_7_2_0_9, \
call_7_2_0_10, \
call_7_2_0_11, \
call_7_2_0_12, \
call_7_2_0_13, \
call_7_2_0_14, \
call_7_2_0_15, \
call_7_2_1_0, \
call_7_2_1_1, \
call_7_2_1_2, \
call_7_2_1_3, \
call_7_2_1_4, \
call_7_2_1_5, \
call_7_2_1_6, \
call_7_2_1_7, \
call_7_2_1_8, \
call_7_2_1_9, \
call_7_2_1_10, \
call_7_2_1_11, \
call_7_2_1_12, \
call_7_2_1_13, \
call_7_2_1_14, \
call_7_2_1_15, \
call_7_2_2_0, \
call_7_2_2_1, \
call_7_2_2_2, \
call_7_2_2_3, \
call_7_2_2_4, \
call_7_2_2_5, \
call_7_2_2_6, \
call_7_2_2_7, \
call_7_2_2_8, \
call_7_2_2_9, \
call_7_2_2_10, \
call_7_2_2_11, \
call_7_2_2_12, \
call_7_2_2_13, \
call_7_2_2_14, \
call_7_2_2_15, \
call_7_2_3_0, \
call_7_2_3_1, \
call_7_2_3_2, \
call_7_2_3_3, \
call_7_2_3_4, \
call_7_2_3_5, \
call_7_2_3_6, \
call_7_2_3_7, \
call_7_2_3_8, \
call_7_2_3_9, \
call_7_2_3_10, \
call_7_2_3_11, \
call_7_2_3_12, \
call_7_2_3_13, \
call_7_2_3_14, \
call_7_2_3_15, \
call_7_2_4_0, \
call_7_2_4_1, \
call_7_2_4_2, \
call_7_2_4_3, \
call_7_2_4_4, \
call_7_2_4_5, \
call_7_2_4_6, \
call_7_2_4_7, \
call_7_2_4_8, \
call_7_2_4_9, \
call_7_2_4_10, \
call_7_2_4_11, \
call_7_2_4_12, \
call_7_2_4_13, \
call_7_2_4_14, \
call_7_2_4_15, \
call_7_2_5_0, \
call_7_2_5_1, \
call_7_2_5_2, \
call_7_2_5_3, \
call_7_2_5_4, \
call_7_2_5_5, \
call_7_2_5_6, \
call_7_2_5_7, \
call_7_2_5_8, \
call_7_2_5_9, \
call_7_2_5_10, \
call_7_2_5_11, \
call_7_2_5_12, \
call_7_2_5_13, \
call_7_2_5_14, \
call_7_2_5_15, \
call_7_2_6_0, \
call_7_2_6_1, \
call_7_2_6_2, \
call_7_2_6_3, \
call_7_2_6_4, \
call_7_2_6_5, \
call_7_2_6_6, \
call_7_2_6_7, \
call_7_2_6_8, \
call_7_2_6_9, \
call_7_2_6_10, \
call_7_2_6_11, \
call_7_2_6_12, \
call_7_2_6_13, \
call_7_2_6_14, \
call_7_2_6_15, \
call_7_2_7_0, \
call_7_2_7_1, \
call_7_2_7_2, \
call_7_2_7_3, \
call_7_2_7_4, \
call_7_2_7_5, \
call_7_2_7_6, \
call_7_2_7_7, \
call_7_2_7_8, \
call_7_2_7_9, \
call_7_2_7_10, \
call_7_2_7_11, \
call_7_2_7_12, \
call_7_2_7_13, \
call_7_2_7_14, \
call_7_2_7_15, \
call_7_2_8_0, \
call_7_2_8_1, \
call_7_2_8_2, \
call_7_2_8_3, \
call_7_2_8_4, \
call_7_2_8_5, \
call_7_2_8_6, \
call_7_2_8_7, \
call_7_2_8_8, \
call_7_2_8_9, \
call_7_2_8_10, \
call_7_2_8_11, \
call_7_2_8_12, \
call_7_2_8_13, \
call_7_2_8_14, \
call_7_2_8_15, \
call_7_2_9_0, \
call_7_2_9_1, \
call_7_2_9_2, \
call_7_2_9_3, \
call_7_2_9_4, \
call_7_2_9_5, \
call_7_2_9_6, \
call_7_2_9_7, \
call_7_2_9_8, \
call_7_2_9_9, \
call_7_2_9_10, \
call_7_2_9_11, \
call_7_2_9_12, \
call_7_2_9_13, \
call_7_2_9_14, \
call_7_2_9_15, \
call_7_2_10_0, \
call_7_2_10_1, \
call_7_2_10_2, \
call_7_2_10_3, \
call_7_2_10_4, \
call_7_2_10_5, \
call_7_2_10_6, \
call_7_2_10_7, \
call_7_2_10_8, \
call_7_2_10_9, \
call_7_2_10_10, \
call_7_2_10_11, \
call_7_2_10_12, \
call_7_2_10_13, \
call_7_2_10_14, \
call_7_2_10_15, \
call_7_2_11_0, \
call_7_2_11_1, \
call_7_2_11_2, \
call_7_2_11_3, \
call_7_2_11_4, \
call_7_2_11_5, \
call_7_2_11_6, \
call_7_2_11_7, \
call_7_2_11_8, \
call_7_2_11_9, \
call_7_2_11_10, \
call_7_2_11_11, \
call_7_2_11_12, \
call_7_2_11_13, \
call_7_2_11_14, \
call_7_2_11_15, \
call_7_2_12_0, \
call_7_2_12_1, \
call_7_2_12_2, \
call_7_2_12_3, \
call_7_2_12_4, \
call_7_2_12_5, \
call_7_2_12_6, \
call_7_2_12_7, \
call_7_2_12_8, \
call_7_2_12_9, \
call_7_2_12_10, \
call_7_2_12_11, \
call_7_2_12_12, \
call_7_2_12_13, \
call_7_2_12_14, \
call_7_2_12_15, \
call_7_2_13_0, \
call_7_2_13_1, \
call_7_2_13_2, \
call_7_2_13_3, \
call_7_2_13_4, \
call_7_2_13_5, \
call_7_2_13_6, \
call_7_2_13_7, \
call_7_2_13_8, \
call_7_2_13_9, \
call_7_2_13_10, \
call_7_2_13_11, \
call_7_2_13_12, \
call_7_2_13_13, \
call_7_2_13_14, \
call_7_2_13_15, \
call_7_2_14_0, \
call_7_2_14_1, \
call_7_2_14_2, \
call_7_2_14_3, \
call_7_2_14_4, \
call_7_2_14_5, \
call_7_2_14_6, \
call_7_2_14_7, \
call_7_2_14_8, \
call_7_2_14_9, \
call_7_2_14_10, \
call_7_2_14_11, \
call_7_2_14_12, \
call_7_2_14_13, \
call_7_2_14_14, \
call_7_2_14_15, \
call_7_2_15_0, \
call_7_2_15_1, \
call_7_2_15_2, \
call_7_2_15_3, \
call_7_2_15_4, \
call_7_2_15_5, \
call_7_2_15_6, \
call_7_2_15_7, \
call_7_2_15_8, \
call_7_2_15_9, \
call_7_2_15_10, \
call_7_2_15_11, \
call_7_2_15_12, \
call_7_2_15_13, \
call_7_2_15_14, \
call_7_2_15_15, \
call_7_3_0_0, \
call_7_3_0_1, \
call_7_3_0_2, \
call_7_3_0_3, \
call_7_3_0_4, \
call_7_3_0_5, \
call_7_3_0_6, \
call_7_3_0_7, \
call_7_3_0_8, \
call_7_3_0_9, \
call_7_3_0_10, \
call_7_3_0_11, \
call_7_3_0_12, \
call_7_3_0_13, \
call_7_3_0_14, \
call_7_3_0_15, \
call_7_3_1_0, \
call_7_3_1_1, \
call_7_3_1_2, \
call_7_3_1_3, \
call_7_3_1_4, \
call_7_3_1_5, \
call_7_3_1_6, \
call_7_3_1_7, \
call_7_3_1_8, \
call_7_3_1_9, \
call_7_3_1_10, \
call_7_3_1_11, \
call_7_3_1_12, \
call_7_3_1_13, \
call_7_3_1_14, \
call_7_3_1_15, \
call_7_3_2_0, \
call_7_3_2_1, \
call_7_3_2_2, \
call_7_3_2_3, \
call_7_3_2_4, \
call_7_3_2_5, \
call_7_3_2_6, \
call_7_3_2_7, \
call_7_3_2_8, \
call_7_3_2_9, \
call_7_3_2_10, \
call_7_3_2_11, \
call_7_3_2_12, \
call_7_3_2_13, \
call_7_3_2_14, \
call_7_3_2_15, \
call_7_3_3_0, \
call_7_3_3_1, \
call_7_3_3_2, \
call_7_3_3_3, \
call_7_3_3_4, \
call_7_3_3_5, \
call_7_3_3_6, \
call_7_3_3_7, \
call_7_3_3_8, \
call_7_3_3_9, \
call_7_3_3_10, \
call_7_3_3_11, \
call_7_3_3_12, \
call_7_3_3_13, \
call_7_3_3_14, \
call_7_3_3_15, \
call_7_3_4_0, \
call_7_3_4_1, \
call_7_3_4_2, \
call_7_3_4_3, \
call_7_3_4_4, \
call_7_3_4_5, \
call_7_3_4_6, \
call_7_3_4_7, \
call_7_3_4_8, \
call_7_3_4_9, \
call_7_3_4_10, \
call_7_3_4_11, \
call_7_3_4_12, \
call_7_3_4_13, \
call_7_3_4_14, \
call_7_3_4_15, \
call_7_3_5_0, \
call_7_3_5_1, \
call_7_3_5_2, \
call_7_3_5_3, \
call_7_3_5_4, \
call_7_3_5_5, \
call_7_3_5_6, \
call_7_3_5_7, \
call_7_3_5_8, \
call_7_3_5_9, \
call_7_3_5_10, \
call_7_3_5_11, \
call_7_3_5_12, \
call_7_3_5_13, \
call_7_3_5_14, \
call_7_3_5_15, \
call_7_3_6_0, \
call_7_3_6_1, \
call_7_3_6_2, \
call_7_3_6_3, \
call_7_3_6_4, \
call_7_3_6_5, \
call_7_3_6_6, \
call_7_3_6_7, \
call_7_3_6_8, \
call_7_3_6_9, \
call_7_3_6_10, \
call_7_3_6_11, \
call_7_3_6_12, \
call_7_3_6_13, \
call_7_3_6_14, \
call_7_3_6_15, \
call_7_3_7_0, \
call_7_3_7_1, \
call_7_3_7_2, \
call_7_3_7_3, \
call_7_3_7_4, \
call_7_3_7_5, \
call_7_3_7_6, \
call_7_3_7_7, \
call_7_3_7_8, \
call_7_3_7_9, \
call_7_3_7_10, \
call_7_3_7_11, \
call_7_3_7_12, \
call_7_3_7_13, \
call_7_3_7_14, \
call_7_3_7_15, \
call_7_3_8_0, \
call_7_3_8_1, \
call_7_3_8_2, \
call_7_3_8_3, \
call_7_3_8_4, \
call_7_3_8_5, \
call_7_3_8_6, \
call_7_3_8_7, \
call_7_3_8_8, \
call_7_3_8_9, \
call_7_3_8_10, \
call_7_3_8_11, \
call_7_3_8_12, \
call_7_3_8_13, \
call_7_3_8_14, \
call_7_3_8_15, \
call_7_3_9_0, \
call_7_3_9_1, \
call_7_3_9_2, \
call_7_3_9_3, \
call_7_3_9_4, \
call_7_3_9_5, \
call_7_3_9_6, \
call_7_3_9_7, \
call_7_3_9_8, \
call_7_3_9_9, \
call_7_3_9_10, \
call_7_3_9_11, \
call_7_3_9_12, \
call_7_3_9_13, \
call_7_3_9_14, \
call_7_3_9_15, \
call_7_3_10_0, \
call_7_3_10_1, \
call_7_3_10_2, \
call_7_3_10_3, \
call_7_3_10_4, \
call_7_3_10_5, \
call_7_3_10_6, \
call_7_3_10_7, \
call_7_3_10_8, \
call_7_3_10_9, \
call_7_3_10_10, \
call_7_3_10_11, \
call_7_3_10_12, \
call_7_3_10_13, \
call_7_3_10_14, \
call_7_3_10_15, \
call_7_3_11_0, \
call_7_3_11_1, \
call_7_3_11_2, \
call_7_3_11_3, \
call_7_3_11_4, \
call_7_3_11_5, \
call_7_3_11_6, \
call_7_3_11_7, \
call_7_3_11_8, \
call_7_3_11_9, \
call_7_3_11_10, \
call_7_3_11_11, \
call_7_3_11_12, \
call_7_3_11_13, \
call_7_3_11_14, \
call_7_3_11_15, \
call_7_3_12_0, \
call_7_3_12_1, \
call_7_3_12_2, \
call_7_3_12_3, \
call_7_3_12_4, \
call_7_3_12_5, \
call_7_3_12_6, \
call_7_3_12_7, \
call_7_3_12_8, \
call_7_3_12_9, \
call_7_3_12_10, \
call_7_3_12_11, \
call_7_3_12_12, \
call_7_3_12_13, \
call_7_3_12_14, \
call_7_3_12_15, \
call_7_3_13_0, \
call_7_3_13_1, \
call_7_3_13_2, \
call_7_3_13_3, \
call_7_3_13_4, \
call_7_3_13_5, \
call_7_3_13_6, \
call_7_3_13_7, \
call_7_3_13_8, \
call_7_3_13_9, \
call_7_3_13_10, \
call_7_3_13_11, \
call_7_3_13_12, \
call_7_3_13_13, \
call_7_3_13_14, \
call_7_3_13_15, \
call_7_3_14_0, \
call_7_3_14_1, \
call_7_3_14_2, \
call_7_3_14_3, \
call_7_3_14_4, \
call_7_3_14_5, \
call_7_3_14_6, \
call_7_3_14_7, \
call_7_3_14_8, \
call_7_3_14_9, \
call_7_3_14_10, \
call_7_3_14_11, \
call_7_3_14_12, \
call_7_3_14_13, \
call_7_3_14_14, \
call_7_3_14_15, \
call_7_3_15_0, \
call_7_3_15_1, \
call_7_3_15_2, \
call_7_3_15_3, \
call_7_3_15_4, \
call_7_3_15_5, \
call_7_3_15_6, \
call_7_3_15_7, \
call_7_3_15_8, \
call_7_3_15_9, \
call_7_3_15_10, \
call_7_3_15_11, \
call_7_3_15_12, \
call_7_3_15_13, \
call_7_3_15_14, \
call_7_3_15_15, \
call_7_4_0_0, \
call_7_4_0_1, \
call_7_4_0_2, \
call_7_4_0_3, \
call_7_4_0_4, \
call_7_4_0_5, \
call_7_4_0_6, \
call_7_4_0_7, \
call_7_4_0_8, \
call_7_4_0_9, \
call_7_4_0_10, \
call_7_4_0_11, \
call_7_4_0_12, \
call_7_4_0_13, \
call_7_4_0_14, \
call_7_4_0_15, \
call_7_4_1_0, \
call_7_4_1_1, \
call_7_4_1_2, \
call_7_4_1_3, \
call_7_4_1_4, \
call_7_4_1_5, \
call_7_4_1_6, \
call_7_4_1_7, \
call_7_4_1_8, \
call_7_4_1_9, \
call_7_4_1_10, \
call_7_4_1_11, \
call_7_4_1_12, \
call_7_4_1_13, \
call_7_4_1_14, \
call_7_4_1_15, \
call_7_4_2_0, \
call_7_4_2_1, \
call_7_4_2_2, \
call_7_4_2_3, \
call_7_4_2_4, \
call_7_4_2_5, \
call_7_4_2_6, \
call_7_4_2_7, \
call_7_4_2_8, \
call_7_4_2_9, \
call_7_4_2_10, \
call_7_4_2_11, \
call_7_4_2_12, \
call_7_4_2_13, \
call_7_4_2_14, \
call_7_4_2_15, \
call_7_4_3_0, \
call_7_4_3_1, \
call_7_4_3_2, \
call_7_4_3_3, \
call_7_4_3_4, \
call_7_4_3_5, \
call_7_4_3_6, \
call_7_4_3_7, \
call_7_4_3_8, \
call_7_4_3_9, \
call_7_4_3_10, \
call_7_4_3_11, \
call_7_4_3_12, \
call_7_4_3_13, \
call_7_4_3_14, \
call_7_4_3_15, \
call_7_4_4_0, \
call_7_4_4_1, \
call_7_4_4_2, \
call_7_4_4_3, \
call_7_4_4_4, \
call_7_4_4_5, \
call_7_4_4_6, \
call_7_4_4_7, \
call_7_4_4_8, \
call_7_4_4_9, \
call_7_4_4_10, \
call_7_4_4_11, \
call_7_4_4_12, \
call_7_4_4_13, \
call_7_4_4_14, \
call_7_4_4_15, \
call_7_4_5_0, \
call_7_4_5_1, \
call_7_4_5_2, \
call_7_4_5_3, \
call_7_4_5_4, \
call_7_4_5_5, \
call_7_4_5_6, \
call_7_4_5_7, \
call_7_4_5_8, \
call_7_4_5_9, \
call_7_4_5_10, \
call_7_4_5_11, \
call_7_4_5_12, \
call_7_4_5_13, \
call_7_4_5_14, \
call_7_4_5_15, \
call_7_4_6_0, \
call_7_4_6_1, \
call_7_4_6_2, \
call_7_4_6_3, \
call_7_4_6_4, \
call_7_4_6_5, \
call_7_4_6_6, \
call_7_4_6_7, \
call_7_4_6_8, \
call_7_4_6_9, \
call_7_4_6_10, \
call_7_4_6_11, \
call_7_4_6_12, \
call_7_4_6_13, \
call_7_4_6_14, \
call_7_4_6_15, \
call_7_4_7_0, \
call_7_4_7_1, \
call_7_4_7_2, \
call_7_4_7_3, \
call_7_4_7_4, \
call_7_4_7_5, \
call_7_4_7_6, \
call_7_4_7_7, \
call_7_4_7_8, \
call_7_4_7_9, \
call_7_4_7_10, \
call_7_4_7_11, \
call_7_4_7_12, \
call_7_4_7_13, \
call_7_4_7_14, \
call_7_4_7_15, \
call_7_4_8_0, \
call_7_4_8_1, \
call_7_4_8_2, \
call_7_4_8_3, \
call_7_4_8_4, \
call_7_4_8_5, \
call_7_4_8_6, \
call_7_4_8_7, \
call_7_4_8_8, \
call_7_4_8_9, \
call_7_4_8_10, \
call_7_4_8_11, \
call_7_4_8_12, \
call_7_4_8_13, \
call_7_4_8_14, \
call_7_4_8_15, \
call_7_4_9_0, \
call_7_4_9_1, \
call_7_4_9_2, \
call_7_4_9_3, \
call_7_4_9_4, \
call_7_4_9_5, \
call_7_4_9_6, \
call_7_4_9_7, \
call_7_4_9_8, \
call_7_4_9_9, \
call_7_4_9_10, \
call_7_4_9_11, \
call_7_4_9_12, \
call_7_4_9_13, \
call_7_4_9_14, \
call_7_4_9_15, \
call_7_4_10_0, \
call_7_4_10_1, \
call_7_4_10_2, \
call_7_4_10_3, \
call_7_4_10_4, \
call_7_4_10_5, \
call_7_4_10_6, \
call_7_4_10_7, \
call_7_4_10_8, \
call_7_4_10_9, \
call_7_4_10_10, \
call_7_4_10_11, \
call_7_4_10_12, \
call_7_4_10_13, \
call_7_4_10_14, \
call_7_4_10_15, \
call_7_4_11_0, \
call_7_4_11_1, \
call_7_4_11_2, \
call_7_4_11_3, \
call_7_4_11_4, \
call_7_4_11_5, \
call_7_4_11_6, \
call_7_4_11_7, \
call_7_4_11_8, \
call_7_4_11_9, \
call_7_4_11_10, \
call_7_4_11_11, \
call_7_4_11_12, \
call_7_4_11_13, \
call_7_4_11_14, \
call_7_4_11_15, \
call_7_4_12_0, \
call_7_4_12_1, \
call_7_4_12_2, \
call_7_4_12_3, \
call_7_4_12_4, \
call_7_4_12_5, \
call_7_4_12_6, \
call_7_4_12_7, \
call_7_4_12_8, \
call_7_4_12_9, \
call_7_4_12_10, \
call_7_4_12_11, \
call_7_4_12_12, \
call_7_4_12_13, \
call_7_4_12_14, \
call_7_4_12_15, \
call_7_4_13_0, \
call_7_4_13_1, \
call_7_4_13_2, \
call_7_4_13_3, \
call_7_4_13_4, \
call_7_4_13_5, \
call_7_4_13_6, \
call_7_4_13_7, \
call_7_4_13_8, \
call_7_4_13_9, \
call_7_4_13_10, \
call_7_4_13_11, \
call_7_4_13_12, \
call_7_4_13_13, \
call_7_4_13_14, \
call_7_4_13_15, \
call_7_4_14_0, \
call_7_4_14_1, \
call_7_4_14_2, \
call_7_4_14_3, \
call_7_4_14_4, \
call_7_4_14_5, \
call_7_4_14_6, \
call_7_4_14_7, \
call_7_4_14_8, \
call_7_4_14_9, \
call_7_4_14_10, \
call_7_4_14_11, \
call_7_4_14_12, \
call_7_4_14_13, \
call_7_4_14_14, \
call_7_4_14_15, \
call_7_4_15_0, \
call_7_4_15_1, \
call_7_4_15_2, \
call_7_4_15_3, \
call_7_4_15_4, \
call_7_4_15_5, \
call_7_4_15_6, \
call_7_4_15_7, \
call_7_4_15_8, \
call_7_4_15_9, \
call_7_4_15_10, \
call_7_4_15_11, \
call_7_4_15_12, \
call_7_4_15_13, \
call_7_4_15_14, \
call_7_4_15_15, \
call_7_5_0_0, \
call_7_5_0_1, \
call_7_5_0_2, \
call_7_5_0_3, \
call_7_5_0_4, \
call_7_5_0_5, \
call_7_5_0_6, \
call_7_5_0_7, \
call_7_5_0_8, \
call_7_5_0_9, \
call_7_5_0_10, \
call_7_5_0_11, \
call_7_5_0_12, \
call_7_5_0_13, \
call_7_5_0_14, \
call_7_5_0_15, \
call_7_5_1_0, \
call_7_5_1_1, \
call_7_5_1_2, \
call_7_5_1_3, \
call_7_5_1_4, \
call_7_5_1_5, \
call_7_5_1_6, \
call_7_5_1_7, \
call_7_5_1_8, \
call_7_5_1_9, \
call_7_5_1_10, \
call_7_5_1_11, \
call_7_5_1_12, \
call_7_5_1_13, \
call_7_5_1_14, \
call_7_5_1_15, \
call_7_5_2_0, \
call_7_5_2_1, \
call_7_5_2_2, \
call_7_5_2_3, \
call_7_5_2_4, \
call_7_5_2_5, \
call_7_5_2_6, \
call_7_5_2_7, \
call_7_5_2_8, \
call_7_5_2_9, \
call_7_5_2_10, \
call_7_5_2_11, \
call_7_5_2_12, \
call_7_5_2_13, \
call_7_5_2_14, \
call_7_5_2_15, \
call_7_5_3_0, \
call_7_5_3_1, \
call_7_5_3_2, \
call_7_5_3_3, \
call_7_5_3_4, \
call_7_5_3_5, \
call_7_5_3_6, \
call_7_5_3_7, \
call_7_5_3_8, \
call_7_5_3_9, \
call_7_5_3_10, \
call_7_5_3_11, \
call_7_5_3_12, \
call_7_5_3_13, \
call_7_5_3_14, \
call_7_5_3_15, \
call_7_5_4_0, \
call_7_5_4_1, \
call_7_5_4_2, \
call_7_5_4_3, \
call_7_5_4_4, \
call_7_5_4_5, \
call_7_5_4_6, \
call_7_5_4_7, \
call_7_5_4_8, \
call_7_5_4_9, \
call_7_5_4_10, \
call_7_5_4_11, \
call_7_5_4_12, \
call_7_5_4_13, \
call_7_5_4_14, \
call_7_5_4_15, \
call_7_5_5_0, \
call_7_5_5_1, \
call_7_5_5_2, \
call_7_5_5_3, \
call_7_5_5_4, \
call_7_5_5_5, \
call_7_5_5_6, \
call_7_5_5_7, \
call_7_5_5_8, \
call_7_5_5_9, \
call_7_5_5_10, \
call_7_5_5_11, \
call_7_5_5_12, \
call_7_5_5_13, \
call_7_5_5_14, \
call_7_5_5_15, \
call_7_5_6_0, \
call_7_5_6_1, \
call_7_5_6_2, \
call_7_5_6_3, \
call_7_5_6_4, \
call_7_5_6_5, \
call_7_5_6_6, \
call_7_5_6_7, \
call_7_5_6_8, \
call_7_5_6_9, \
call_7_5_6_10, \
call_7_5_6_11, \
call_7_5_6_12, \
call_7_5_6_13, \
call_7_5_6_14, \
call_7_5_6_15, \
call_7_5_7_0, \
call_7_5_7_1, \
call_7_5_7_2, \
call_7_5_7_3, \
call_7_5_7_4, \
call_7_5_7_5, \
call_7_5_7_6, \
call_7_5_7_7, \
call_7_5_7_8, \
call_7_5_7_9, \
call_7_5_7_10, \
call_7_5_7_11, \
call_7_5_7_12, \
call_7_5_7_13, \
call_7_5_7_14, \
call_7_5_7_15, \
call_7_5_8_0, \
call_7_5_8_1, \
call_7_5_8_2, \
call_7_5_8_3, \
call_7_5_8_4, \
call_7_5_8_5, \
call_7_5_8_6, \
call_7_5_8_7, \
call_7_5_8_8, \
call_7_5_8_9, \
call_7_5_8_10, \
call_7_5_8_11, \
call_7_5_8_12, \
call_7_5_8_13, \
call_7_5_8_14, \
call_7_5_8_15, \
call_7_5_9_0, \
call_7_5_9_1, \
call_7_5_9_2, \
call_7_5_9_3, \
call_7_5_9_4, \
call_7_5_9_5, \
call_7_5_9_6, \
call_7_5_9_7, \
call_7_5_9_8, \
call_7_5_9_9, \
call_7_5_9_10, \
call_7_5_9_11, \
call_7_5_9_12, \
call_7_5_9_13, \
call_7_5_9_14, \
call_7_5_9_15, \
call_7_5_10_0, \
call_7_5_10_1, \
call_7_5_10_2, \
call_7_5_10_3, \
call_7_5_10_4, \
call_7_5_10_5, \
call_7_5_10_6, \
call_7_5_10_7, \
call_7_5_10_8, \
call_7_5_10_9, \
call_7_5_10_10, \
call_7_5_10_11, \
call_7_5_10_12, \
call_7_5_10_13, \
call_7_5_10_14, \
call_7_5_10_15, \
call_7_5_11_0, \
call_7_5_11_1, \
call_7_5_11_2, \
call_7_5_11_3, \
call_7_5_11_4, \
call_7_5_11_5, \
call_7_5_11_6, \
call_7_5_11_7, \
call_7_5_11_8, \
call_7_5_11_9, \
call_7_5_11_10, \
call_7_5_11_11, \
call_7_5_11_12, \
call_7_5_11_13, \
call_7_5_11_14, \
call_7_5_11_15, \
call_7_5_12_0, \
call_7_5_12_1, \
call_7_5_12_2, \
call_7_5_12_3, \
call_7_5_12_4, \
call_7_5_12_5, \
call_7_5_12_6, \
call_7_5_12_7, \
call_7_5_12_8, \
call_7_5_12_9, \
call_7_5_12_10, \
call_7_5_12_11, \
call_7_5_12_12, \
call_7_5_12_13, \
call_7_5_12_14, \
call_7_5_12_15, \
call_7_5_13_0, \
call_7_5_13_1, \
call_7_5_13_2, \
call_7_5_13_3, \
call_7_5_13_4, \
call_7_5_13_5, \
call_7_5_13_6, \
call_7_5_13_7, \
call_7_5_13_8, \
call_7_5_13_9, \
call_7_5_13_10, \
call_7_5_13_11, \
call_7_5_13_12, \
call_7_5_13_13, \
call_7_5_13_14, \
call_7_5_13_15, \
call_7_5_14_0, \
call_7_5_14_1, \
call_7_5_14_2, \
call_7_5_14_3, \
call_7_5_14_4, \
call_7_5_14_5, \
call_7_5_14_6, \
call_7_5_14_7, \
call_7_5_14_8, \
call_7_5_14_9, \
call_7_5_14_10, \
call_7_5_14_11, \
call_7_5_14_12, \
call_7_5_14_13, \
call_7_5_14_14, \
call_7_5_14_15, \
call_7_5_15_0, \
call_7_5_15_1, \
call_7_5_15_2, \
call_7_5_15_3, \
call_7_5_15_4, \
call_7_5_15_5, \
call_7_5_15_6, \
call_7_5_15_7, \
call_7_5_15_8, \
call_7_5_15_9, \
call_7_5_15_10, \
call_7_5_15_11, \
call_7_5_15_12, \
call_7_5_15_13, \
call_7_5_15_14, \
call_7_5_15_15, \
call_7_6_0_0, \
call_7_6_0_1, \
call_7_6_0_2, \
call_7_6_0_3, \
call_7_6_0_4, \
call_7_6_0_5, \
call_7_6_0_6, \
call_7_6_0_7, \
call_7_6_0_8, \
call_7_6_0_9, \
call_7_6_0_10, \
call_7_6_0_11, \
call_7_6_0_12, \
call_7_6_0_13, \
call_7_6_0_14, \
call_7_6_0_15, \
call_7_6_1_0, \
call_7_6_1_1, \
call_7_6_1_2, \
call_7_6_1_3, \
call_7_6_1_4, \
call_7_6_1_5, \
call_7_6_1_6, \
call_7_6_1_7, \
call_7_6_1_8, \
call_7_6_1_9, \
call_7_6_1_10, \
call_7_6_1_11, \
call_7_6_1_12, \
call_7_6_1_13, \
call_7_6_1_14, \
call_7_6_1_15, \
call_7_6_2_0, \
call_7_6_2_1, \
call_7_6_2_2, \
call_7_6_2_3, \
call_7_6_2_4, \
call_7_6_2_5, \
call_7_6_2_6, \
call_7_6_2_7, \
call_7_6_2_8, \
call_7_6_2_9, \
call_7_6_2_10, \
call_7_6_2_11, \
call_7_6_2_12, \
call_7_6_2_13, \
call_7_6_2_14, \
call_7_6_2_15, \
call_7_6_3_0, \
call_7_6_3_1, \
call_7_6_3_2, \
call_7_6_3_3, \
call_7_6_3_4, \
call_7_6_3_5, \
call_7_6_3_6, \
call_7_6_3_7, \
call_7_6_3_8, \
call_7_6_3_9, \
call_7_6_3_10, \
call_7_6_3_11, \
call_7_6_3_12, \
call_7_6_3_13, \
call_7_6_3_14, \
call_7_6_3_15, \
call_7_6_4_0, \
call_7_6_4_1, \
call_7_6_4_2, \
call_7_6_4_3, \
call_7_6_4_4, \
call_7_6_4_5, \
call_7_6_4_6, \
call_7_6_4_7, \
call_7_6_4_8, \
call_7_6_4_9, \
call_7_6_4_10, \
call_7_6_4_11, \
call_7_6_4_12, \
call_7_6_4_13, \
call_7_6_4_14, \
call_7_6_4_15, \
call_7_6_5_0, \
call_7_6_5_1, \
call_7_6_5_2, \
call_7_6_5_3, \
call_7_6_5_4, \
call_7_6_5_5, \
call_7_6_5_6, \
call_7_6_5_7, \
call_7_6_5_8, \
call_7_6_5_9, \
call_7_6_5_10, \
call_7_6_5_11, \
call_7_6_5_12, \
call_7_6_5_13, \
call_7_6_5_14, \
call_7_6_5_15, \
call_7_6_6_0, \
call_7_6_6_1, \
call_7_6_6_2, \
call_7_6_6_3, \
call_7_6_6_4, \
call_7_6_6_5, \
call_7_6_6_6, \
call_7_6_6_7, \
call_7_6_6_8, \
call_7_6_6_9, \
call_7_6_6_10, \
call_7_6_6_11, \
call_7_6_6_12, \
call_7_6_6_13, \
call_7_6_6_14, \
call_7_6_6_15, \
call_7_6_7_0, \
call_7_6_7_1, \
call_7_6_7_2, \
call_7_6_7_3, \
call_7_6_7_4, \
call_7_6_7_5, \
call_7_6_7_6, \
call_7_6_7_7, \
call_7_6_7_8, \
call_7_6_7_9, \
call_7_6_7_10, \
call_7_6_7_11, \
call_7_6_7_12, \
call_7_6_7_13, \
call_7_6_7_14, \
call_7_6_7_15, \
call_7_6_8_0, \
call_7_6_8_1, \
call_7_6_8_2, \
call_7_6_8_3, \
call_7_6_8_4, \
call_7_6_8_5, \
call_7_6_8_6, \
call_7_6_8_7, \
call_7_6_8_8, \
call_7_6_8_9, \
call_7_6_8_10, \
call_7_6_8_11, \
call_7_6_8_12, \
call_7_6_8_13, \
call_7_6_8_14, \
call_7_6_8_15, \
call_7_6_9_0, \
call_7_6_9_1, \
call_7_6_9_2, \
call_7_6_9_3, \
call_7_6_9_4, \
call_7_6_9_5, \
call_7_6_9_6, \
call_7_6_9_7, \
call_7_6_9_8, \
call_7_6_9_9, \
call_7_6_9_10, \
call_7_6_9_11, \
call_7_6_9_12, \
call_7_6_9_13, \
call_7_6_9_14, \
call_7_6_9_15, \
call_7_6_10_0, \
call_7_6_10_1, \
call_7_6_10_2, \
call_7_6_10_3, \
call_7_6_10_4, \
call_7_6_10_5, \
call_7_6_10_6, \
call_7_6_10_7, \
call_7_6_10_8, \
call_7_6_10_9, \
call_7_6_10_10, \
call_7_6_10_11, \
call_7_6_10_12, \
call_7_6_10_13, \
call_7_6_10_14, \
call_7_6_10_15, \
call_7_6_11_0, \
call_7_6_11_1, \
call_7_6_11_2, \
call_7_6_11_3, \
call_7_6_11_4, \
call_7_6_11_5, \
call_7_6_11_6, \
call_7_6_11_7, \
call_7_6_11_8, \
call_7_6_11_9, \
call_7_6_11_10, \
call_7_6_11_11, \
call_7_6_11_12, \
call_7_6_11_13, \
call_7_6_11_14, \
call_7_6_11_15, \
call_7_6_12_0, \
call_7_6_12_1, \
call_7_6_12_2, \
call_7_6_12_3, \
call_7_6_12_4, \
call_7_6_12_5, \
call_7_6_12_6, \
call_7_6_12_7, \
call_7_6_12_8, \
call_7_6_12_9, \
call_7_6_12_10, \
call_7_6_12_11, \
call_7_6_12_12, \
call_7_6_12_13, \
call_7_6_12_14, \
call_7_6_12_15, \
call_7_6_13_0, \
call_7_6_13_1, \
call_7_6_13_2, \
call_7_6_13_3, \
call_7_6_13_4, \
call_7_6_13_5, \
call_7_6_13_6, \
call_7_6_13_7, \
call_7_6_13_8, \
call_7_6_13_9, \
call_7_6_13_10, \
call_7_6_13_11, \
call_7_6_13_12, \
call_7_6_13_13, \
call_7_6_13_14, \
call_7_6_13_15, \
call_7_6_14_0, \
call_7_6_14_1, \
call_7_6_14_2, \
call_7_6_14_3, \
call_7_6_14_4, \
call_7_6_14_5, \
call_7_6_14_6, \
call_7_6_14_7, \
call_7_6_14_8, \
call_7_6_14_9, \
call_7_6_14_10, \
call_7_6_14_11, \
call_7_6_14_12, \
call_7_6_14_13, \
call_7_6_14_14, \
call_7_6_14_15, \
call_7_6_15_0, \
call_7_6_15_1, \
call_7_6_15_2, \
call_7_6_15_3, \
call_7_6_15_4, \
call_7_6_15_5, \
call_7_6_15_6, \
call_7_6_15_7, \
call_7_6_15_8, \
call_7_6_15_9, \
call_7_6_15_10, \
call_7_6_15_11, \
call_7_6_15_12, \
call_7_6_15_13, \
call_7_6_15_14, \
call_7_6_15_15, \
call_7_7_0_0, \
call_7_7_0_1, \
call_7_7_0_2, \
call_7_7_0_3, \
call_7_7_0_4, \
call_7_7_0_5, \
call_7_7_0_6, \
call_7_7_0_7, \
call_7_7_0_8, \
call_7_7_0_9, \
call_7_7_0_10, \
call_7_7_0_11, \
call_7_7_0_12, \
call_7_7_0_13, \
call_7_7_0_14, \
call_7_7_0_15, \
call_7_7_1_0, \
call_7_7_1_1, \
call_7_7_1_2, \
call_7_7_1_3, \
call_7_7_1_4, \
call_7_7_1_5, \
call_7_7_1_6, \
call_7_7_1_7, \
call_7_7_1_8, \
call_7_7_1_9, \
call_7_7_1_10, \
call_7_7_1_11, \
call_7_7_1_12, \
call_7_7_1_13, \
call_7_7_1_14, \
call_7_7_1_15, \
call_7_7_2_0, \
call_7_7_2_1, \
call_7_7_2_2, \
call_7_7_2_3, \
call_7_7_2_4, \
call_7_7_2_5, \
call_7_7_2_6, \
call_7_7_2_7, \
call_7_7_2_8, \
call_7_7_2_9, \
call_7_7_2_10, \
call_7_7_2_11, \
call_7_7_2_12, \
call_7_7_2_13, \
call_7_7_2_14, \
call_7_7_2_15, \
call_7_7_3_0, \
call_7_7_3_1, \
call_7_7_3_2, \
call_7_7_3_3, \
call_7_7_3_4, \
call_7_7_3_5, \
call_7_7_3_6, \
call_7_7_3_7, \
call_7_7_3_8, \
call_7_7_3_9, \
call_7_7_3_10, \
call_7_7_3_11, \
call_7_7_3_12, \
call_7_7_3_13, \
call_7_7_3_14, \
call_7_7_3_15, \
call_7_7_4_0, \
call_7_7_4_1, \
call_7_7_4_2, \
call_7_7_4_3, \
call_7_7_4_4, \
call_7_7_4_5, \
call_7_7_4_6, \
call_7_7_4_7, \
call_7_7_4_8, \
call_7_7_4_9, \
call_7_7_4_10, \
call_7_7_4_11, \
call_7_7_4_12, \
call_7_7_4_13, \
call_7_7_4_14, \
call_7_7_4_15, \
call_7_7_5_0, \
call_7_7_5_1, \
call_7_7_5_2, \
call_7_7_5_3, \
call_7_7_5_4, \
call_7_7_5_5, \
call_7_7_5_6, \
call_7_7_5_7, \
call_7_7_5_8, \
call_7_7_5_9, \
call_7_7_5_10, \
call_7_7_5_11, \
call_7_7_5_12, \
call_7_7_5_13, \
call_7_7_5_14, \
call_7_7_5_15, \
call_7_7_6_0, \
call_7_7_6_1, \
call_7_7_6_2, \
call_7_7_6_3, \
call_7_7_6_4, \
call_7_7_6_5, \
call_7_7_6_6, \
call_7_7_6_7, \
call_7_7_6_8, \
call_7_7_6_9, \
call_7_7_6_10, \
call_7_7_6_11, \
call_7_7_6_12, \
call_7_7_6_13, \
call_7_7_6_14, \
call_7_7_6_15, \
call_7_7_7_0, \
call_7_7_7_1, \
call_7_7_7_2, \
call_7_7_7_3, \
call_7_7_7_4, \
call_7_7_7_5, \
call_7_7_7_6, \
call_7_7_7_7, \
call_7_7_7_8, \
call_7_7_7_9, \
call_7_7_7_10, \
call_7_7_7_11, \
call_7_7_7_12, \
call_7_7_7_13, \
call_7_7_7_14, \
call_7_7_7_15, \
call_7_7_8_0, \
call_7_7_8_1, \
call_7_7_8_2, \
call_7_7_8_3, \
call_7_7_8_4, \
call_7_7_8_5, \
call_7_7_8_6, \
call_7_7_8_7, \
call_7_7_8_8, \
call_7_7_8_9, \
call_7_7_8_10, \
call_7_7_8_11, \
call_7_7_8_12, \
call_7_7_8_13, \
call_7_7_8_14, \
call_7_7_8_15, \
call_7_7_9_0, \
call_7_7_9_1, \
call_7_7_9_2, \
call_7_7_9_3, \
call_7_7_9_4, \
call_7_7_9_5, \
call_7_7_9_6, \
call_7_7_9_7, \
call_7_7_9_8, \
call_7_7_9_9, \
call_7_7_9_10, \
call_7_7_9_11, \
call_7_7_9_12, \
call_7_7_9_13, \
call_7_7_9_14, \
call_7_7_9_15, \
call_7_7_10_0, \
call_7_7_10_1, \
call_7_7_10_2, \
call_7_7_10_3, \
call_7_7_10_4, \
call_7_7_10_5, \
call_7_7_10_6, \
call_7_7_10_7, \
call_7_7_10_8, \
call_7_7_10_9, \
call_7_7_10_10, \
call_7_7_10_11, \
call_7_7_10_12, \
call_7_7_10_13, \
call_7_7_10_14, \
call_7_7_10_15, \
call_7_7_11_0, \
call_7_7_11_1, \
call_7_7_11_2, \
call_7_7_11_3, \
call_7_7_11_4, \
call_7_7_11_5, \
call_7_7_11_6, \
call_7_7_11_7, \
call_7_7_11_8, \
call_7_7_11_9, \
call_7_7_11_10, \
call_7_7_11_11, \
call_7_7_11_12, \
call_7_7_11_13, \
call_7_7_11_14, \
call_7_7_11_15, \
call_7_7_12_0, \
call_7_7_12_1, \
call_7_7_12_2, \
call_7_7_12_3, \
call_7_7_12_4, \
call_7_7_12_5, \
call_7_7_12_6, \
call_7_7_12_7, \
call_7_7_12_8, \
call_7_7_12_9, \
call_7_7_12_10, \
call_7_7_12_11, \
call_7_7_12_12, \
call_7_7_12_13, \
call_7_7_12_14, \
call_7_7_12_15, \
call_7_7_13_0, \
call_7_7_13_1, \
call_7_7_13_2, \
call_7_7_13_3, \
call_7_7_13_4, \
call_7_7_13_5, \
call_7_7_13_6, \
call_7_7_13_7, \
call_7_7_13_8, \
call_7_7_13_9, \
call_7_7_13_10, \
call_7_7_13_11, \
call_7_7_13_12, \
call_7_7_13_13, \
call_7_7_13_14, \
call_7_7_13_15, \
call_7_7_14_0, \
call_7_7_14_1, \
call_7_7_14_2, \
call_7_7_14_3, \
call_7_7_14_4, \
call_7_7_14_5, \
call_7_7_14_6, \
call_7_7_14_7, \
call_7_7_14_8, \
call_7_7_14_9, \
call_7_7_14_10, \
call_7_7_14_11, \
call_7_7_14_12, \
call_7_7_14_13, \
call_7_7_14_14, \
call_7_7_14_15, \
call_7_7_15_0, \
call_7_7_15_1, \
call_7_7_15_2, \
call_7_7_15_3, \
call_7_7_15_4, \
call_7_7_15_5, \
call_7_7_15_6, \
call_7_7_15_7, \
call_7_7_15_8, \
call_7_7_15_9, \
call_7_7_15_10, \
call_7_7_15_11, \
call_7_7_15_12, \
call_7_7_15_13, \
call_7_7_15_14, \
call_7_7_15_15, \
call_7_8_0_0, \
call_7_8_0_1, \
call_7_8_0_2, \
call_7_8_0_3, \
call_7_8_0_4, \
call_7_8_0_5, \
call_7_8_0_6, \
call_7_8_0_7, \
call_7_8_0_8, \
call_7_8_0_9, \
call_7_8_0_10, \
call_7_8_0_11, \
call_7_8_0_12, \
call_7_8_0_13, \
call_7_8_0_14, \
call_7_8_0_15, \
call_7_8_1_0, \
call_7_8_1_1, \
call_7_8_1_2, \
call_7_8_1_3, \
call_7_8_1_4, \
call_7_8_1_5, \
call_7_8_1_6, \
call_7_8_1_7, \
call_7_8_1_8, \
call_7_8_1_9, \
call_7_8_1_10, \
call_7_8_1_11, \
call_7_8_1_12, \
call_7_8_1_13, \
call_7_8_1_14, \
call_7_8_1_15, \
call_7_8_2_0, \
call_7_8_2_1, \
call_7_8_2_2, \
call_7_8_2_3, \
call_7_8_2_4, \
call_7_8_2_5, \
call_7_8_2_6, \
call_7_8_2_7, \
call_7_8_2_8, \
call_7_8_2_9, \
call_7_8_2_10, \
call_7_8_2_11, \
call_7_8_2_12, \
call_7_8_2_13, \
call_7_8_2_14, \
call_7_8_2_15, \
call_7_8_3_0, \
call_7_8_3_1, \
call_7_8_3_2, \
call_7_8_3_3, \
call_7_8_3_4, \
call_7_8_3_5, \
call_7_8_3_6, \
call_7_8_3_7, \
call_7_8_3_8, \
call_7_8_3_9, \
call_7_8_3_10, \
call_7_8_3_11, \
call_7_8_3_12, \
call_7_8_3_13, \
call_7_8_3_14, \
call_7_8_3_15, \
call_7_8_4_0, \
call_7_8_4_1, \
call_7_8_4_2, \
call_7_8_4_3, \
call_7_8_4_4, \
call_7_8_4_5, \
call_7_8_4_6, \
call_7_8_4_7, \
call_7_8_4_8, \
call_7_8_4_9, \
call_7_8_4_10, \
call_7_8_4_11, \
call_7_8_4_12, \
call_7_8_4_13, \
call_7_8_4_14, \
call_7_8_4_15, \
call_7_8_5_0, \
call_7_8_5_1, \
call_7_8_5_2, \
call_7_8_5_3, \
call_7_8_5_4, \
call_7_8_5_5, \
call_7_8_5_6, \
call_7_8_5_7, \
call_7_8_5_8, \
call_7_8_5_9, \
call_7_8_5_10, \
call_7_8_5_11, \
call_7_8_5_12, \
call_7_8_5_13, \
call_7_8_5_14, \
call_7_8_5_15, \
call_7_8_6_0, \
call_7_8_6_1, \
call_7_8_6_2, \
call_7_8_6_3, \
call_7_8_6_4, \
call_7_8_6_5, \
call_7_8_6_6, \
call_7_8_6_7, \
call_7_8_6_8, \
call_7_8_6_9, \
call_7_8_6_10, \
call_7_8_6_11, \
call_7_8_6_12, \
call_7_8_6_13, \
call_7_8_6_14, \
call_7_8_6_15, \
call_7_8_7_0, \
call_7_8_7_1, \
call_7_8_7_2, \
call_7_8_7_3, \
call_7_8_7_4, \
call_7_8_7_5, \
call_7_8_7_6, \
call_7_8_7_7, \
call_7_8_7_8, \
call_7_8_7_9, \
call_7_8_7_10, \
call_7_8_7_11, \
call_7_8_7_12, \
call_7_8_7_13, \
call_7_8_7_14, \
call_7_8_7_15, \
call_7_8_8_0, \
call_7_8_8_1, \
call_7_8_8_2, \
call_7_8_8_3, \
call_7_8_8_4, \
call_7_8_8_5, \
call_7_8_8_6, \
call_7_8_8_7, \
call_7_8_8_8, \
call_7_8_8_9, \
call_7_8_8_10, \
call_7_8_8_11, \
call_7_8_8_12, \
call_7_8_8_13, \
call_7_8_8_14, \
call_7_8_8_15, \
call_7_8_9_0, \
call_7_8_9_1, \
call_7_8_9_2, \
call_7_8_9_3, \
call_7_8_9_4, \
call_7_8_9_5, \
call_7_8_9_6, \
call_7_8_9_7, \
call_7_8_9_8, \
call_7_8_9_9, \
call_7_8_9_10, \
call_7_8_9_11, \
call_7_8_9_12, \
call_7_8_9_13, \
call_7_8_9_14, \
call_7_8_9_15, \
call_7_8_10_0, \
call_7_8_10_1, \
call_7_8_10_2, \
call_7_8_10_3, \
call_7_8_10_4, \
call_7_8_10_5, \
call_7_8_10_6, \
call_7_8_10_7, \
call_7_8_10_8, \
call_7_8_10_9, \
call_7_8_10_10, \
call_7_8_10_11, \
call_7_8_10_12, \
call_7_8_10_13, \
call_7_8_10_14, \
call_7_8_10_15, \
call_7_8_11_0, \
call_7_8_11_1, \
call_7_8_11_2, \
call_7_8_11_3, \
call_7_8_11_4, \
call_7_8_11_5, \
call_7_8_11_6, \
call_7_8_11_7, \
call_7_8_11_8, \
call_7_8_11_9, \
call_7_8_11_10, \
call_7_8_11_11, \
call_7_8_11_12, \
call_7_8_11_13, \
call_7_8_11_14, \
call_7_8_11_15, \
call_7_8_12_0, \
call_7_8_12_1, \
call_7_8_12_2, \
call_7_8_12_3, \
call_7_8_12_4, \
call_7_8_12_5, \
call_7_8_12_6, \
call_7_8_12_7, \
call_7_8_12_8, \
call_7_8_12_9, \
call_7_8_12_10, \
call_7_8_12_11, \
call_7_8_12_12, \
call_7_8_12_13, \
call_7_8_12_14, \
call_7_8_12_15, \
call_7_8_13_0, \
call_7_8_13_1, \
call_7_8_13_2, \
call_7_8_13_3, \
call_7_8_13_4, \
call_7_8_13_5, \
call_7_8_13_6, \
call_7_8_13_7, \
call_7_8_13_8, \
call_7_8_13_9, \
call_7_8_13_10, \
call_7_8_13_11, \
call_7_8_13_12, \
call_7_8_13_13, \
call_7_8_13_14, \
call_7_8_13_15, \
call_7_8_14_0, \
call_7_8_14_1, \
call_7_8_14_2, \
call_7_8_14_3, \
call_7_8_14_4, \
call_7_8_14_5, \
call_7_8_14_6, \
call_7_8_14_7, \
call_7_8_14_8, \
call_7_8_14_9, \
call_7_8_14_10, \
call_7_8_14_11, \
call_7_8_14_12, \
call_7_8_14_13, \
call_7_8_14_14, \
call_7_8_14_15, \
call_7_8_15_0, \
call_7_8_15_1, \
call_7_8_15_2, \
call_7_8_15_3, \
call_7_8_15_4, \
call_7_8_15_5, \
call_7_8_15_6, \
call_7_8_15_7, \
call_7_8_15_8, \
call_7_8_15_9, \
call_7_8_15_10, \
call_7_8_15_11, \
call_7_8_15_12, \
call_7_8_15_13, \
call_7_8_15_14, \
call_7_8_15_15, \
call_7_9_0_0, \
call_7_9_0_1, \
call_7_9_0_2, \
call_7_9_0_3, \
call_7_9_0_4, \
call_7_9_0_5, \
call_7_9_0_6, \
call_7_9_0_7, \
call_7_9_0_8, \
call_7_9_0_9, \
call_7_9_0_10, \
call_7_9_0_11, \
call_7_9_0_12, \
call_7_9_0_13, \
call_7_9_0_14, \
call_7_9_0_15, \
call_7_9_1_0, \
call_7_9_1_1, \
call_7_9_1_2, \
call_7_9_1_3, \
call_7_9_1_4, \
call_7_9_1_5, \
call_7_9_1_6, \
call_7_9_1_7, \
call_7_9_1_8, \
call_7_9_1_9, \
call_7_9_1_10, \
call_7_9_1_11, \
call_7_9_1_12, \
call_7_9_1_13, \
call_7_9_1_14, \
call_7_9_1_15, \
call_7_9_2_0, \
call_7_9_2_1, \
call_7_9_2_2, \
call_7_9_2_3, \
call_7_9_2_4, \
call_7_9_2_5, \
call_7_9_2_6, \
call_7_9_2_7, \
call_7_9_2_8, \
call_7_9_2_9, \
call_7_9_2_10, \
call_7_9_2_11, \
call_7_9_2_12, \
call_7_9_2_13, \
call_7_9_2_14, \
call_7_9_2_15, \
call_7_9_3_0, \
call_7_9_3_1, \
call_7_9_3_2, \
call_7_9_3_3, \
call_7_9_3_4, \
call_7_9_3_5, \
call_7_9_3_6, \
call_7_9_3_7, \
call_7_9_3_8, \
call_7_9_3_9, \
call_7_9_3_10, \
call_7_9_3_11, \
call_7_9_3_12, \
call_7_9_3_13, \
call_7_9_3_14, \
call_7_9_3_15, \
call_7_9_4_0, \
call_7_9_4_1, \
call_7_9_4_2, \
call_7_9_4_3, \
call_7_9_4_4, \
call_7_9_4_5, \
call_7_9_4_6, \
call_7_9_4_7, \
call_7_9_4_8, \
call_7_9_4_9, \
call_7_9_4_10, \
call_7_9_4_11, \
call_7_9_4_12, \
call_7_9_4_13, \
call_7_9_4_14, \
call_7_9_4_15, \
call_7_9_5_0, \
call_7_9_5_1, \
call_7_9_5_2, \
call_7_9_5_3, \
call_7_9_5_4, \
call_7_9_5_5, \
call_7_9_5_6, \
call_7_9_5_7, \
call_7_9_5_8, \
call_7_9_5_9, \
call_7_9_5_10, \
call_7_9_5_11, \
call_7_9_5_12, \
call_7_9_5_13, \
call_7_9_5_14, \
call_7_9_5_15, \
call_7_9_6_0, \
call_7_9_6_1, \
call_7_9_6_2, \
call_7_9_6_3, \
call_7_9_6_4, \
call_7_9_6_5, \
call_7_9_6_6, \
call_7_9_6_7, \
call_7_9_6_8, \
call_7_9_6_9, \
call_7_9_6_10, \
call_7_9_6_11, \
call_7_9_6_12, \
call_7_9_6_13, \
call_7_9_6_14, \
call_7_9_6_15, \
call_7_9_7_0, \
call_7_9_7_1, \
call_7_9_7_2, \
call_7_9_7_3, \
call_7_9_7_4, \
call_7_9_7_5, \
call_7_9_7_6, \
call_7_9_7_7, \
call_7_9_7_8, \
call_7_9_7_9, \
call_7_9_7_10, \
call_7_9_7_11, \
call_7_9_7_12, \
call_7_9_7_13, \
call_7_9_7_14, \
call_7_9_7_15, \
call_7_9_8_0, \
call_7_9_8_1, \
call_7_9_8_2, \
call_7_9_8_3, \
call_7_9_8_4, \
call_7_9_8_5, \
call_7_9_8_6, \
call_7_9_8_7, \
call_7_9_8_8, \
call_7_9_8_9, \
call_7_9_8_10, \
call_7_9_8_11, \
call_7_9_8_12, \
call_7_9_8_13, \
call_7_9_8_14, \
call_7_9_8_15, \
call_7_9_9_0, \
call_7_9_9_1, \
call_7_9_9_2, \
call_7_9_9_3, \
call_7_9_9_4, \
call_7_9_9_5, \
call_7_9_9_6, \
call_7_9_9_7, \
call_7_9_9_8, \
call_7_9_9_9, \
call_7_9_9_10, \
call_7_9_9_11, \
call_7_9_9_12, \
call_7_9_9_13, \
call_7_9_9_14, \
call_7_9_9_15, \
call_7_9_10_0, \
call_7_9_10_1, \
call_7_9_10_2, \
call_7_9_10_3, \
call_7_9_10_4, \
call_7_9_10_5, \
call_7_9_10_6, \
call_7_9_10_7, \
call_7_9_10_8, \
call_7_9_10_9, \
call_7_9_10_10, \
call_7_9_10_11, \
call_7_9_10_12, \
call_7_9_10_13, \
call_7_9_10_14, \
call_7_9_10_15, \
call_7_9_11_0, \
call_7_9_11_1, \
call_7_9_11_2, \
call_7_9_11_3, \
call_7_9_11_4, \
call_7_9_11_5, \
call_7_9_11_6, \
call_7_9_11_7, \
call_7_9_11_8, \
call_7_9_11_9, \
call_7_9_11_10, \
call_7_9_11_11, \
call_7_9_11_12, \
call_7_9_11_13, \
call_7_9_11_14, \
call_7_9_11_15, \
call_7_9_12_0, \
call_7_9_12_1, \
call_7_9_12_2, \
call_7_9_12_3, \
call_7_9_12_4, \
call_7_9_12_5, \
call_7_9_12_6, \
call_7_9_12_7, \
call_7_9_12_8, \
call_7_9_12_9, \
call_7_9_12_10, \
call_7_9_12_11, \
call_7_9_12_12, \
call_7_9_12_13, \
call_7_9_12_14, \
call_7_9_12_15, \
call_7_9_13_0, \
call_7_9_13_1, \
call_7_9_13_2, \
call_7_9_13_3, \
call_7_9_13_4, \
call_7_9_13_5, \
call_7_9_13_6, \
call_7_9_13_7, \
call_7_9_13_8, \
call_7_9_13_9, \
call_7_9_13_10, \
call_7_9_13_11, \
call_7_9_13_12, \
call_7_9_13_13, \
call_7_9_13_14, \
call_7_9_13_15, \
call_7_9_14_0, \
call_7_9_14_1, \
call_7_9_14_2, \
call_7_9_14_3, \
call_7_9_14_4, \
call_7_9_14_5, \
call_7_9_14_6, \
call_7_9_14_7, \
call_7_9_14_8, \
call_7_9_14_9, \
call_7_9_14_10, \
call_7_9_14_11, \
call_7_9_14_12, \
call_7_9_14_13, \
call_7_9_14_14, \
call_7_9_14_15, \
call_7_9_15_0, \
call_7_9_15_1, \
call_7_9_15_2, \
call_7_9_15_3, \
call_7_9_15_4, \
call_7_9_15_5, \
call_7_9_15_6, \
call_7_9_15_7, \
call_7_9_15_8, \
call_7_9_15_9, \
call_7_9_15_10, \
call_7_9_15_11, \
call_7_9_15_12, \
call_7_9_15_13, \
call_7_9_15_14, \
call_7_9_15_15, \
call_7_10_0_0, \
call_7_10_0_1, \
call_7_10_0_2, \
call_7_10_0_3, \
call_7_10_0_4, \
call_7_10_0_5, \
call_7_10_0_6, \
call_7_10_0_7, \
call_7_10_0_8, \
call_7_10_0_9, \
call_7_10_0_10, \
call_7_10_0_11, \
call_7_10_0_12, \
call_7_10_0_13, \
call_7_10_0_14, \
call_7_10_0_15, \
call_7_10_1_0, \
call_7_10_1_1, \
call_7_10_1_2, \
call_7_10_1_3, \
call_7_10_1_4, \
call_7_10_1_5, \
call_7_10_1_6, \
call_7_10_1_7, \
call_7_10_1_8, \
call_7_10_1_9, \
call_7_10_1_10, \
call_7_10_1_11, \
call_7_10_1_12, \
call_7_10_1_13, \
call_7_10_1_14, \
call_7_10_1_15, \
call_7_10_2_0, \
call_7_10_2_1, \
call_7_10_2_2, \
call_7_10_2_3, \
call_7_10_2_4, \
call_7_10_2_5, \
call_7_10_2_6, \
call_7_10_2_7, \
call_7_10_2_8, \
call_7_10_2_9, \
call_7_10_2_10, \
call_7_10_2_11, \
call_7_10_2_12, \
call_7_10_2_13, \
call_7_10_2_14, \
call_7_10_2_15, \
call_7_10_3_0, \
call_7_10_3_1, \
call_7_10_3_2, \
call_7_10_3_3, \
call_7_10_3_4, \
call_7_10_3_5, \
call_7_10_3_6, \
call_7_10_3_7, \
call_7_10_3_8, \
call_7_10_3_9, \
call_7_10_3_10, \
call_7_10_3_11, \
call_7_10_3_12, \
call_7_10_3_13, \
call_7_10_3_14, \
call_7_10_3_15, \
call_7_10_4_0, \
call_7_10_4_1, \
call_7_10_4_2, \
call_7_10_4_3, \
call_7_10_4_4, \
call_7_10_4_5, \
call_7_10_4_6, \
call_7_10_4_7, \
call_7_10_4_8, \
call_7_10_4_9, \
call_7_10_4_10, \
call_7_10_4_11, \
call_7_10_4_12, \
call_7_10_4_13, \
call_7_10_4_14, \
call_7_10_4_15, \
call_7_10_5_0, \
call_7_10_5_1, \
call_7_10_5_2, \
call_7_10_5_3, \
call_7_10_5_4, \
call_7_10_5_5, \
call_7_10_5_6, \
call_7_10_5_7, \
call_7_10_5_8, \
call_7_10_5_9, \
call_7_10_5_10, \
call_7_10_5_11, \
call_7_10_5_12, \
call_7_10_5_13, \
call_7_10_5_14, \
call_7_10_5_15, \
call_7_10_6_0, \
call_7_10_6_1, \
call_7_10_6_2, \
call_7_10_6_3, \
call_7_10_6_4, \
call_7_10_6_5, \
call_7_10_6_6, \
call_7_10_6_7, \
call_7_10_6_8, \
call_7_10_6_9, \
call_7_10_6_10, \
call_7_10_6_11, \
call_7_10_6_12, \
call_7_10_6_13, \
call_7_10_6_14, \
call_7_10_6_15, \
call_7_10_7_0, \
call_7_10_7_1, \
call_7_10_7_2, \
call_7_10_7_3, \
call_7_10_7_4, \
call_7_10_7_5, \
call_7_10_7_6, \
call_7_10_7_7, \
call_7_10_7_8, \
call_7_10_7_9, \
call_7_10_7_10, \
call_7_10_7_11, \
call_7_10_7_12, \
call_7_10_7_13, \
call_7_10_7_14, \
call_7_10_7_15, \
call_7_10_8_0, \
call_7_10_8_1, \
call_7_10_8_2, \
call_7_10_8_3, \
call_7_10_8_4, \
call_7_10_8_5, \
call_7_10_8_6, \
call_7_10_8_7, \
call_7_10_8_8, \
call_7_10_8_9, \
call_7_10_8_10, \
call_7_10_8_11, \
call_7_10_8_12, \
call_7_10_8_13, \
call_7_10_8_14, \
call_7_10_8_15, \
call_7_10_9_0, \
call_7_10_9_1, \
call_7_10_9_2, \
call_7_10_9_3, \
call_7_10_9_4, \
call_7_10_9_5, \
call_7_10_9_6, \
call_7_10_9_7, \
call_7_10_9_8, \
call_7_10_9_9, \
call_7_10_9_10, \
call_7_10_9_11, \
call_7_10_9_12, \
call_7_10_9_13, \
call_7_10_9_14, \
call_7_10_9_15, \
call_7_10_10_0, \
call_7_10_10_1, \
call_7_10_10_2, \
call_7_10_10_3, \
call_7_10_10_4, \
call_7_10_10_5, \
call_7_10_10_6, \
call_7_10_10_7, \
call_7_10_10_8, \
call_7_10_10_9, \
call_7_10_10_10, \
call_7_10_10_11, \
call_7_10_10_12, \
call_7_10_10_13, \
call_7_10_10_14, \
call_7_10_10_15, \
call_7_10_11_0, \
call_7_10_11_1, \
call_7_10_11_2, \
call_7_10_11_3, \
call_7_10_11_4, \
call_7_10_11_5, \
call_7_10_11_6, \
call_7_10_11_7, \
call_7_10_11_8, \
call_7_10_11_9, \
call_7_10_11_10, \
call_7_10_11_11, \
call_7_10_11_12, \
call_7_10_11_13, \
call_7_10_11_14, \
call_7_10_11_15, \
call_7_10_12_0, \
call_7_10_12_1, \
call_7_10_12_2, \
call_7_10_12_3, \
call_7_10_12_4, \
call_7_10_12_5, \
call_7_10_12_6, \
call_7_10_12_7, \
call_7_10_12_8, \
call_7_10_12_9, \
call_7_10_12_10, \
call_7_10_12_11, \
call_7_10_12_12, \
call_7_10_12_13, \
call_7_10_12_14, \
call_7_10_12_15, \
call_7_10_13_0, \
call_7_10_13_1, \
call_7_10_13_2, \
call_7_10_13_3, \
call_7_10_13_4, \
call_7_10_13_5, \
call_7_10_13_6, \
call_7_10_13_7, \
call_7_10_13_8, \
call_7_10_13_9, \
call_7_10_13_10, \
call_7_10_13_11, \
call_7_10_13_12, \
call_7_10_13_13, \
call_7_10_13_14, \
call_7_10_13_15, \
call_7_10_14_0, \
call_7_10_14_1, \
call_7_10_14_2, \
call_7_10_14_3, \
call_7_10_14_4, \
call_7_10_14_5, \
call_7_10_14_6, \
call_7_10_14_7, \
call_7_10_14_8, \
call_7_10_14_9, \
call_7_10_14_10, \
call_7_10_14_11, \
call_7_10_14_12, \
call_7_10_14_13, \
call_7_10_14_14, \
call_7_10_14_15, \
call_7_10_15_0, \
call_7_10_15_1, \
call_7_10_15_2, \
call_7_10_15_3, \
call_7_10_15_4, \
call_7_10_15_5, \
call_7_10_15_6, \
call_7_10_15_7, \
call_7_10_15_8, \
call_7_10_15_9, \
call_7_10_15_10, \
call_7_10_15_11, \
call_7_10_15_12, \
call_7_10_15_13, \
call_7_10_15_14, \
call_7_10_15_15, \
call_7_11_0_0, \
call_7_11_0_1, \
call_7_11_0_2, \
call_7_11_0_3, \
call_7_11_0_4, \
call_7_11_0_5, \
call_7_11_0_6, \
call_7_11_0_7, \
call_7_11_0_8, \
call_7_11_0_9, \
call_7_11_0_10, \
call_7_11_0_11, \
call_7_11_0_12, \
call_7_11_0_13, \
call_7_11_0_14, \
call_7_11_0_15, \
call_7_11_1_0, \
call_7_11_1_1, \
call_7_11_1_2, \
call_7_11_1_3, \
call_7_11_1_4, \
call_7_11_1_5, \
call_7_11_1_6, \
call_7_11_1_7, \
call_7_11_1_8, \
call_7_11_1_9, \
call_7_11_1_10, \
call_7_11_1_11, \
call_7_11_1_12, \
call_7_11_1_13, \
call_7_11_1_14, \
call_7_11_1_15, \
call_7_11_2_0, \
call_7_11_2_1, \
call_7_11_2_2, \
call_7_11_2_3, \
call_7_11_2_4, \
call_7_11_2_5, \
call_7_11_2_6, \
call_7_11_2_7, \
call_7_11_2_8, \
call_7_11_2_9, \
call_7_11_2_10, \
call_7_11_2_11, \
call_7_11_2_12, \
call_7_11_2_13, \
call_7_11_2_14, \
call_7_11_2_15, \
call_7_11_3_0, \
call_7_11_3_1, \
call_7_11_3_2, \
call_7_11_3_3, \
call_7_11_3_4, \
call_7_11_3_5, \
call_7_11_3_6, \
call_7_11_3_7, \
call_7_11_3_8, \
call_7_11_3_9, \
call_7_11_3_10, \
call_7_11_3_11, \
call_7_11_3_12, \
call_7_11_3_13, \
call_7_11_3_14, \
call_7_11_3_15, \
call_7_11_4_0, \
call_7_11_4_1, \
call_7_11_4_2, \
call_7_11_4_3, \
call_7_11_4_4, \
call_7_11_4_5, \
call_7_11_4_6, \
call_7_11_4_7, \
call_7_11_4_8, \
call_7_11_4_9, \
call_7_11_4_10, \
call_7_11_4_11, \
call_7_11_4_12, \
call_7_11_4_13, \
call_7_11_4_14, \
call_7_11_4_15, \
call_7_11_5_0, \
call_7_11_5_1, \
call_7_11_5_2, \
call_7_11_5_3, \
call_7_11_5_4, \
call_7_11_5_5, \
call_7_11_5_6, \
call_7_11_5_7, \
call_7_11_5_8, \
call_7_11_5_9, \
call_7_11_5_10, \
call_7_11_5_11, \
call_7_11_5_12, \
call_7_11_5_13, \
call_7_11_5_14, \
call_7_11_5_15, \
call_7_11_6_0, \
call_7_11_6_1, \
call_7_11_6_2, \
call_7_11_6_3, \
call_7_11_6_4, \
call_7_11_6_5, \
call_7_11_6_6, \
call_7_11_6_7, \
call_7_11_6_8, \
call_7_11_6_9, \
call_7_11_6_10, \
call_7_11_6_11, \
call_7_11_6_12, \
call_7_11_6_13, \
call_7_11_6_14, \
call_7_11_6_15, \
call_7_11_7_0, \
call_7_11_7_1, \
call_7_11_7_2, \
call_7_11_7_3, \
call_7_11_7_4, \
call_7_11_7_5, \
call_7_11_7_6, \
call_7_11_7_7, \
call_7_11_7_8, \
call_7_11_7_9, \
call_7_11_7_10, \
call_7_11_7_11, \
call_7_11_7_12, \
call_7_11_7_13, \
call_7_11_7_14, \
call_7_11_7_15, \
call_7_11_8_0, \
call_7_11_8_1, \
call_7_11_8_2, \
call_7_11_8_3, \
call_7_11_8_4, \
call_7_11_8_5, \
call_7_11_8_6, \
call_7_11_8_7, \
call_7_11_8_8, \
call_7_11_8_9, \
call_7_11_8_10, \
call_7_11_8_11, \
call_7_11_8_12, \
call_7_11_8_13, \
call_7_11_8_14, \
call_7_11_8_15, \
call_7_11_9_0, \
call_7_11_9_1, \
call_7_11_9_2, \
call_7_11_9_3, \
call_7_11_9_4, \
call_7_11_9_5, \
call_7_11_9_6, \
call_7_11_9_7, \
call_7_11_9_8, \
call_7_11_9_9, \
call_7_11_9_10, \
call_7_11_9_11, \
call_7_11_9_12, \
call_7_11_9_13, \
call_7_11_9_14, \
call_7_11_9_15, \
call_7_11_10_0, \
call_7_11_10_1, \
call_7_11_10_2, \
call_7_11_10_3, \
call_7_11_10_4, \
call_7_11_10_5, \
call_7_11_10_6, \
call_7_11_10_7, \
call_7_11_10_8, \
call_7_11_10_9, \
call_7_11_10_10, \
call_7_11_10_11, \
call_7_11_10_12, \
call_7_11_10_13, \
call_7_11_10_14, \
call_7_11_10_15, \
call_7_11_11_0, \
call_7_11_11_1, \
call_7_11_11_2, \
call_7_11_11_3, \
call_7_11_11_4, \
call_7_11_11_5, \
call_7_11_11_6, \
call_7_11_11_7, \
call_7_11_11_8, \
call_7_11_11_9, \
call_7_11_11_10, \
call_7_11_11_11, \
call_7_11_11_12, \
call_7_11_11_13, \
call_7_11_11_14, \
call_7_11_11_15, \
call_7_11_12_0, \
call_7_11_12_1, \
call_7_11_12_2, \
call_7_11_12_3, \
call_7_11_12_4, \
call_7_11_12_5, \
call_7_11_12_6, \
call_7_11_12_7, \
call_7_11_12_8, \
call_7_11_12_9, \
call_7_11_12_10, \
call_7_11_12_11, \
call_7_11_12_12, \
call_7_11_12_13, \
call_7_11_12_14, \
call_7_11_12_15, \
call_7_11_13_0, \
call_7_11_13_1, \
call_7_11_13_2, \
call_7_11_13_3, \
call_7_11_13_4, \
call_7_11_13_5, \
call_7_11_13_6, \
call_7_11_13_7, \
call_7_11_13_8, \
call_7_11_13_9, \
call_7_11_13_10, \
call_7_11_13_11, \
call_7_11_13_12, \
call_7_11_13_13, \
call_7_11_13_14, \
call_7_11_13_15, \
call_7_11_14_0, \
call_7_11_14_1, \
call_7_11_14_2, \
call_7_11_14_3, \
call_7_11_14_4, \
call_7_11_14_5, \
call_7_11_14_6, \
call_7_11_14_7, \
call_7_11_14_8, \
call_7_11_14_9, \
call_7_11_14_10, \
call_7_11_14_11, \
call_7_11_14_12, \
call_7_11_14_13, \
call_7_11_14_14, \
call_7_11_14_15, \
call_7_11_15_0, \
call_7_11_15_1, \
call_7_11_15_2, \
call_7_11_15_3, \
call_7_11_15_4, \
call_7_11_15_5, \
call_7_11_15_6, \
call_7_11_15_7, \
call_7_11_15_8, \
call_7_11_15_9, \
call_7_11_15_10, \
call_7_11_15_11, \
call_7_11_15_12, \
call_7_11_15_13, \
call_7_11_15_14, \
call_7_11_15_15, \
call_7_12_0_0, \
call_7_12_0_1, \
call_7_12_0_2, \
call_7_12_0_3, \
call_7_12_0_4, \
call_7_12_0_5, \
call_7_12_0_6, \
call_7_12_0_7, \
call_7_12_0_8, \
call_7_12_0_9, \
call_7_12_0_10, \
call_7_12_0_11, \
call_7_12_0_12, \
call_7_12_0_13, \
call_7_12_0_14, \
call_7_12_0_15, \
call_7_12_1_0, \
call_7_12_1_1, \
call_7_12_1_2, \
call_7_12_1_3, \
call_7_12_1_4, \
call_7_12_1_5, \
call_7_12_1_6, \
call_7_12_1_7, \
call_7_12_1_8, \
call_7_12_1_9, \
call_7_12_1_10, \
call_7_12_1_11, \
call_7_12_1_12, \
call_7_12_1_13, \
call_7_12_1_14, \
call_7_12_1_15, \
call_7_12_2_0, \
call_7_12_2_1, \
call_7_12_2_2, \
call_7_12_2_3, \
call_7_12_2_4, \
call_7_12_2_5, \
call_7_12_2_6, \
call_7_12_2_7, \
call_7_12_2_8, \
call_7_12_2_9, \
call_7_12_2_10, \
call_7_12_2_11, \
call_7_12_2_12, \
call_7_12_2_13, \
call_7_12_2_14, \
call_7_12_2_15, \
call_7_12_3_0, \
call_7_12_3_1, \
call_7_12_3_2, \
call_7_12_3_3, \
call_7_12_3_4, \
call_7_12_3_5, \
call_7_12_3_6, \
call_7_12_3_7, \
call_7_12_3_8, \
call_7_12_3_9, \
call_7_12_3_10, \
call_7_12_3_11, \
call_7_12_3_12, \
call_7_12_3_13, \
call_7_12_3_14, \
call_7_12_3_15, \
call_7_12_4_0, \
call_7_12_4_1, \
call_7_12_4_2, \
call_7_12_4_3, \
call_7_12_4_4, \
call_7_12_4_5, \
call_7_12_4_6, \
call_7_12_4_7, \
call_7_12_4_8, \
call_7_12_4_9, \
call_7_12_4_10, \
call_7_12_4_11, \
call_7_12_4_12, \
call_7_12_4_13, \
call_7_12_4_14, \
call_7_12_4_15, \
call_7_12_5_0, \
call_7_12_5_1, \
call_7_12_5_2, \
call_7_12_5_3, \
call_7_12_5_4, \
call_7_12_5_5, \
call_7_12_5_6, \
call_7_12_5_7, \
call_7_12_5_8, \
call_7_12_5_9, \
call_7_12_5_10, \
call_7_12_5_11, \
call_7_12_5_12, \
call_7_12_5_13, \
call_7_12_5_14, \
call_7_12_5_15, \
call_7_12_6_0, \
call_7_12_6_1, \
call_7_12_6_2, \
call_7_12_6_3, \
call_7_12_6_4, \
call_7_12_6_5, \
call_7_12_6_6, \
call_7_12_6_7, \
call_7_12_6_8, \
call_7_12_6_9, \
call_7_12_6_10, \
call_7_12_6_11, \
call_7_12_6_12, \
call_7_12_6_13, \
call_7_12_6_14, \
call_7_12_6_15, \
call_7_12_7_0, \
call_7_12_7_1, \
call_7_12_7_2, \
call_7_12_7_3, \
call_7_12_7_4, \
call_7_12_7_5, \
call_7_12_7_6, \
call_7_12_7_7, \
call_7_12_7_8, \
call_7_12_7_9, \
call_7_12_7_10, \
call_7_12_7_11, \
call_7_12_7_12, \
call_7_12_7_13, \
call_7_12_7_14, \
call_7_12_7_15, \
call_7_12_8_0, \
call_7_12_8_1, \
call_7_12_8_2, \
call_7_12_8_3, \
call_7_12_8_4, \
call_7_12_8_5, \
call_7_12_8_6, \
call_7_12_8_7, \
call_7_12_8_8, \
call_7_12_8_9, \
call_7_12_8_10, \
call_7_12_8_11, \
call_7_12_8_12, \
call_7_12_8_13, \
call_7_12_8_14, \
call_7_12_8_15, \
call_7_12_9_0, \
call_7_12_9_1, \
call_7_12_9_2, \
call_7_12_9_3, \
call_7_12_9_4, \
call_7_12_9_5, \
call_7_12_9_6, \
call_7_12_9_7, \
call_7_12_9_8, \
call_7_12_9_9, \
call_7_12_9_10, \
call_7_12_9_11, \
call_7_12_9_12, \
call_7_12_9_13, \
call_7_12_9_14, \
call_7_12_9_15, \
call_7_12_10_0, \
call_7_12_10_1, \
call_7_12_10_2, \
call_7_12_10_3, \
call_7_12_10_4, \
call_7_12_10_5, \
call_7_12_10_6, \
call_7_12_10_7, \
call_7_12_10_8, \
call_7_12_10_9, \
call_7_12_10_10, \
call_7_12_10_11, \
call_7_12_10_12, \
call_7_12_10_13, \
call_7_12_10_14, \
call_7_12_10_15, \
call_7_12_11_0, \
call_7_12_11_1, \
call_7_12_11_2, \
call_7_12_11_3, \
call_7_12_11_4, \
call_7_12_11_5, \
call_7_12_11_6, \
call_7_12_11_7, \
call_7_12_11_8, \
call_7_12_11_9, \
call_7_12_11_10, \
call_7_12_11_11, \
call_7_12_11_12, \
call_7_12_11_13, \
call_7_12_11_14, \
call_7_12_11_15, \
call_7_12_12_0, \
call_7_12_12_1, \
call_7_12_12_2, \
call_7_12_12_3, \
call_7_12_12_4, \
call_7_12_12_5, \
call_7_12_12_6, \
call_7_12_12_7, \
call_7_12_12_8, \
call_7_12_12_9, \
call_7_12_12_10, \
call_7_12_12_11, \
call_7_12_12_12, \
call_7_12_12_13, \
call_7_12_12_14, \
call_7_12_12_15, \
call_7_12_13_0, \
call_7_12_13_1, \
call_7_12_13_2, \
call_7_12_13_3, \
call_7_12_13_4, \
call_7_12_13_5, \
call_7_12_13_6, \
call_7_12_13_7, \
call_7_12_13_8, \
call_7_12_13_9, \
call_7_12_13_10, \
call_7_12_13_11, \
call_7_12_13_12, \
call_7_12_13_13, \
call_7_12_13_14, \
call_7_12_13_15, \
call_7_12_14_0, \
call_7_12_14_1, \
call_7_12_14_2, \
call_7_12_14_3, \
call_7_12_14_4, \
call_7_12_14_5, \
call_7_12_14_6, \
call_7_12_14_7, \
call_7_12_14_8, \
call_7_12_14_9, \
call_7_12_14_10, \
call_7_12_14_11, \
call_7_12_14_12, \
call_7_12_14_13, \
call_7_12_14_14, \
call_7_12_14_15, \
call_7_12_15_0, \
call_7_12_15_1, \
call_7_12_15_2, \
call_7_12_15_3, \
call_7_12_15_4, \
call_7_12_15_5, \
call_7_12_15_6, \
call_7_12_15_7, \
call_7_12_15_8, \
call_7_12_15_9, \
call_7_12_15_10, \
call_7_12_15_11, \
call_7_12_15_12, \
call_7_12_15_13, \
call_7_12_15_14, \
call_7_12_15_15, \
call_7_13_0_0, \
call_7_13_0_1, \
call_7_13_0_2, \
call_7_13_0_3, \
call_7_13_0_4, \
call_7_13_0_5, \
call_7_13_0_6, \
call_7_13_0_7, \
call_7_13_0_8, \
call_7_13_0_9, \
call_7_13_0_10, \
call_7_13_0_11, \
call_7_13_0_12, \
call_7_13_0_13, \
call_7_13_0_14, \
call_7_13_0_15, \
call_7_13_1_0, \
call_7_13_1_1, \
call_7_13_1_2, \
call_7_13_1_3, \
call_7_13_1_4, \
call_7_13_1_5, \
call_7_13_1_6, \
call_7_13_1_7, \
call_7_13_1_8, \
call_7_13_1_9, \
call_7_13_1_10, \
call_7_13_1_11, \
call_7_13_1_12, \
call_7_13_1_13, \
call_7_13_1_14, \
call_7_13_1_15, \
call_7_13_2_0, \
call_7_13_2_1, \
call_7_13_2_2, \
call_7_13_2_3, \
call_7_13_2_4, \
call_7_13_2_5, \
call_7_13_2_6, \
call_7_13_2_7, \
call_7_13_2_8, \
call_7_13_2_9, \
call_7_13_2_10, \
call_7_13_2_11, \
call_7_13_2_12, \
call_7_13_2_13, \
call_7_13_2_14, \
call_7_13_2_15, \
call_7_13_3_0, \
call_7_13_3_1, \
call_7_13_3_2, \
call_7_13_3_3, \
call_7_13_3_4, \
call_7_13_3_5, \
call_7_13_3_6, \
call_7_13_3_7, \
call_7_13_3_8, \
call_7_13_3_9, \
call_7_13_3_10, \
call_7_13_3_11, \
call_7_13_3_12, \
call_7_13_3_13, \
call_7_13_3_14, \
call_7_13_3_15, \
call_7_13_4_0, \
call_7_13_4_1, \
call_7_13_4_2, \
call_7_13_4_3, \
call_7_13_4_4, \
call_7_13_4_5, \
call_7_13_4_6, \
call_7_13_4_7, \
call_7_13_4_8, \
call_7_13_4_9, \
call_7_13_4_10, \
call_7_13_4_11, \
call_7_13_4_12, \
call_7_13_4_13, \
call_7_13_4_14, \
call_7_13_4_15, \
call_7_13_5_0, \
call_7_13_5_1, \
call_7_13_5_2, \
call_7_13_5_3, \
call_7_13_5_4, \
call_7_13_5_5, \
call_7_13_5_6, \
call_7_13_5_7, \
call_7_13_5_8, \
call_7_13_5_9, \
call_7_13_5_10, \
call_7_13_5_11, \
call_7_13_5_12, \
call_7_13_5_13, \
call_7_13_5_14, \
call_7_13_5_15, \
call_7_13_6_0, \
call_7_13_6_1, \
call_7_13_6_2, \
call_7_13_6_3, \
call_7_13_6_4, \
call_7_13_6_5, \
call_7_13_6_6, \
call_7_13_6_7, \
call_7_13_6_8, \
call_7_13_6_9, \
call_7_13_6_10, \
call_7_13_6_11, \
call_7_13_6_12, \
call_7_13_6_13, \
call_7_13_6_14, \
call_7_13_6_15, \
call_7_13_7_0, \
call_7_13_7_1, \
call_7_13_7_2, \
call_7_13_7_3, \
call_7_13_7_4, \
call_7_13_7_5, \
call_7_13_7_6, \
call_7_13_7_7, \
call_7_13_7_8, \
call_7_13_7_9, \
call_7_13_7_10, \
call_7_13_7_11, \
call_7_13_7_12, \
call_7_13_7_13, \
call_7_13_7_14, \
call_7_13_7_15, \
call_7_13_8_0, \
call_7_13_8_1, \
call_7_13_8_2, \
call_7_13_8_3, \
call_7_13_8_4, \
call_7_13_8_5, \
call_7_13_8_6, \
call_7_13_8_7, \
call_7_13_8_8, \
call_7_13_8_9, \
call_7_13_8_10, \
call_7_13_8_11, \
call_7_13_8_12, \
call_7_13_8_13, \
call_7_13_8_14, \
call_7_13_8_15, \
call_7_13_9_0, \
call_7_13_9_1, \
call_7_13_9_2, \
call_7_13_9_3, \
call_7_13_9_4, \
call_7_13_9_5, \
call_7_13_9_6, \
call_7_13_9_7, \
call_7_13_9_8, \
call_7_13_9_9, \
call_7_13_9_10, \
call_7_13_9_11, \
call_7_13_9_12, \
call_7_13_9_13, \
call_7_13_9_14, \
call_7_13_9_15, \
call_7_13_10_0, \
call_7_13_10_1, \
call_7_13_10_2, \
call_7_13_10_3, \
call_7_13_10_4, \
call_7_13_10_5, \
call_7_13_10_6, \
call_7_13_10_7, \
call_7_13_10_8, \
call_7_13_10_9, \
call_7_13_10_10, \
call_7_13_10_11, \
call_7_13_10_12, \
call_7_13_10_13, \
call_7_13_10_14, \
call_7_13_10_15, \
call_7_13_11_0, \
call_7_13_11_1, \
call_7_13_11_2, \
call_7_13_11_3, \
call_7_13_11_4, \
call_7_13_11_5, \
call_7_13_11_6, \
call_7_13_11_7, \
call_7_13_11_8, \
call_7_13_11_9, \
call_7_13_11_10, \
call_7_13_11_11, \
call_7_13_11_12, \
call_7_13_11_13, \
call_7_13_11_14, \
call_7_13_11_15, \
call_7_13_12_0, \
call_7_13_12_1, \
call_7_13_12_2, \
call_7_13_12_3, \
call_7_13_12_4, \
call_7_13_12_5, \
call_7_13_12_6, \
call_7_13_12_7, \
call_7_13_12_8, \
call_7_13_12_9, \
call_7_13_12_10, \
call_7_13_12_11, \
call_7_13_12_12, \
call_7_13_12_13, \
call_7_13_12_14, \
call_7_13_12_15, \
call_7_13_13_0, \
call_7_13_13_1, \
call_7_13_13_2, \
call_7_13_13_3, \
call_7_13_13_4, \
call_7_13_13_5, \
call_7_13_13_6, \
call_7_13_13_7, \
call_7_13_13_8, \
call_7_13_13_9, \
call_7_13_13_10, \
call_7_13_13_11, \
call_7_13_13_12, \
call_7_13_13_13, \
call_7_13_13_14, \
call_7_13_13_15, \
call_7_13_14_0, \
call_7_13_14_1, \
call_7_13_14_2, \
call_7_13_14_3, \
call_7_13_14_4, \
call_7_13_14_5, \
call_7_13_14_6, \
call_7_13_14_7, \
call_7_13_14_8, \
call_7_13_14_9, \
call_7_13_14_10, \
call_7_13_14_11, \
call_7_13_14_12, \
call_7_13_14_13, \
call_7_13_14_14, \
call_7_13_14_15, \
call_7_13_15_0, \
call_7_13_15_1, \
call_7_13_15_2, \
call_7_13_15_3, \
call_7_13_15_4, \
call_7_13_15_5, \
call_7_13_15_6, \
call_7_13_15_7, \
call_7_13_15_8, \
call_7_13_15_9, \
call_7_13_15_10, \
call_7_13_15_11, \
call_7_13_15_12, \
call_7_13_15_13, \
call_7_13_15_14, \
call_7_13_15_15, \
call_7_14_0_0, \
call_7_14_0_1, \
call_7_14_0_2, \
call_7_14_0_3, \
call_7_14_0_4, \
call_7_14_0_5, \
call_7_14_0_6, \
call_7_14_0_7, \
call_7_14_0_8, \
call_7_14_0_9, \
call_7_14_0_10, \
call_7_14_0_11, \
call_7_14_0_12, \
call_7_14_0_13, \
call_7_14_0_14, \
call_7_14_0_15, \
call_7_14_1_0, \
call_7_14_1_1, \
call_7_14_1_2, \
call_7_14_1_3, \
call_7_14_1_4, \
call_7_14_1_5, \
call_7_14_1_6, \
call_7_14_1_7, \
call_7_14_1_8, \
call_7_14_1_9, \
call_7_14_1_10, \
call_7_14_1_11, \
call_7_14_1_12, \
call_7_14_1_13, \
call_7_14_1_14, \
call_7_14_1_15, \
call_7_14_2_0, \
call_7_14_2_1, \
call_7_14_2_2, \
call_7_14_2_3, \
call_7_14_2_4, \
call_7_14_2_5, \
call_7_14_2_6, \
call_7_14_2_7, \
call_7_14_2_8, \
call_7_14_2_9, \
call_7_14_2_10, \
call_7_14_2_11, \
call_7_14_2_12, \
call_7_14_2_13, \
call_7_14_2_14, \
call_7_14_2_15, \
call_7_14_3_0, \
call_7_14_3_1, \
call_7_14_3_2, \
call_7_14_3_3, \
call_7_14_3_4, \
call_7_14_3_5, \
call_7_14_3_6, \
call_7_14_3_7, \
call_7_14_3_8, \
call_7_14_3_9, \
call_7_14_3_10, \
call_7_14_3_11, \
call_7_14_3_12, \
call_7_14_3_13, \
call_7_14_3_14, \
call_7_14_3_15, \
call_7_14_4_0, \
call_7_14_4_1, \
call_7_14_4_2, \
call_7_14_4_3, \
call_7_14_4_4, \
call_7_14_4_5, \
call_7_14_4_6, \
call_7_14_4_7, \
call_7_14_4_8, \
call_7_14_4_9, \
call_7_14_4_10, \
call_7_14_4_11, \
call_7_14_4_12, \
call_7_14_4_13, \
call_7_14_4_14, \
call_7_14_4_15, \
call_7_14_5_0, \
call_7_14_5_1, \
call_7_14_5_2, \
call_7_14_5_3, \
call_7_14_5_4, \
call_7_14_5_5, \
call_7_14_5_6, \
call_7_14_5_7, \
call_7_14_5_8, \
call_7_14_5_9, \
call_7_14_5_10, \
call_7_14_5_11, \
call_7_14_5_12, \
call_7_14_5_13, \
call_7_14_5_14, \
call_7_14_5_15, \
call_7_14_6_0, \
call_7_14_6_1, \
call_7_14_6_2, \
call_7_14_6_3, \
call_7_14_6_4, \
call_7_14_6_5, \
call_7_14_6_6, \
call_7_14_6_7, \
call_7_14_6_8, \
call_7_14_6_9, \
call_7_14_6_10, \
call_7_14_6_11, \
call_7_14_6_12, \
call_7_14_6_13, \
call_7_14_6_14, \
call_7_14_6_15, \
call_7_14_7_0, \
call_7_14_7_1, \
call_7_14_7_2, \
call_7_14_7_3, \
call_7_14_7_4, \
call_7_14_7_5, \
call_7_14_7_6, \
call_7_14_7_7, \
call_7_14_7_8, \
call_7_14_7_9, \
call_7_14_7_10, \
call_7_14_7_11, \
call_7_14_7_12, \
call_7_14_7_13, \
call_7_14_7_14, \
call_7_14_7_15, \
call_7_14_8_0, \
call_7_14_8_1, \
call_7_14_8_2, \
call_7_14_8_3, \
call_7_14_8_4, \
call_7_14_8_5, \
call_7_14_8_6, \
call_7_14_8_7, \
call_7_14_8_8, \
call_7_14_8_9, \
call_7_14_8_10, \
call_7_14_8_11, \
call_7_14_8_12, \
call_7_14_8_13, \
call_7_14_8_14, \
call_7_14_8_15, \
call_7_14_9_0, \
call_7_14_9_1, \
call_7_14_9_2, \
call_7_14_9_3, \
call_7_14_9_4, \
call_7_14_9_5, \
call_7_14_9_6, \
call_7_14_9_7, \
call_7_14_9_8, \
call_7_14_9_9, \
call_7_14_9_10, \
call_7_14_9_11, \
call_7_14_9_12, \
call_7_14_9_13, \
call_7_14_9_14, \
call_7_14_9_15, \
call_7_14_10_0, \
call_7_14_10_1, \
call_7_14_10_2, \
call_7_14_10_3, \
call_7_14_10_4, \
call_7_14_10_5, \
call_7_14_10_6, \
call_7_14_10_7, \
call_7_14_10_8, \
call_7_14_10_9, \
call_7_14_10_10, \
call_7_14_10_11, \
call_7_14_10_12, \
call_7_14_10_13, \
call_7_14_10_14, \
call_7_14_10_15, \
call_7_14_11_0, \
call_7_14_11_1, \
call_7_14_11_2, \
call_7_14_11_3, \
call_7_14_11_4, \
call_7_14_11_5, \
call_7_14_11_6, \
call_7_14_11_7, \
call_7_14_11_8, \
call_7_14_11_9, \
call_7_14_11_10, \
call_7_14_11_11, \
call_7_14_11_12, \
call_7_14_11_13, \
call_7_14_11_14, \
call_7_14_11_15, \
call_7_14_12_0, \
call_7_14_12_1, \
call_7_14_12_2, \
call_7_14_12_3, \
call_7_14_12_4, \
call_7_14_12_5, \
call_7_14_12_6, \
call_7_14_12_7, \
call_7_14_12_8, \
call_7_14_12_9, \
call_7_14_12_10, \
call_7_14_12_11, \
call_7_14_12_12, \
call_7_14_12_13, \
call_7_14_12_14, \
call_7_14_12_15, \
call_7_14_13_0, \
call_7_14_13_1, \
call_7_14_13_2, \
call_7_14_13_3, \
call_7_14_13_4, \
call_7_14_13_5, \
call_7_14_13_6, \
call_7_14_13_7, \
call_7_14_13_8, \
call_7_14_13_9, \
call_7_14_13_10, \
call_7_14_13_11, \
call_7_14_13_12, \
call_7_14_13_13, \
call_7_14_13_14, \
call_7_14_13_15, \
call_7_14_14_0, \
call_7_14_14_1, \
call_7_14_14_2, \
call_7_14_14_3, \
call_7_14_14_4, \
call_7_14_14_5, \
call_7_14_14_6, \
call_7_14_14_7, \
call_7_14_14_8, \
call_7_14_14_9, \
call_7_14_14_10, \
call_7_14_14_11, \
call_7_14_14_12, \
call_7_14_14_13, \
call_7_14_14_14, \
call_7_14_14_15, \
call_7_14_15_0, \
call_7_14_15_1, \
call_7_14_15_2, \
call_7_14_15_3, \
call_7_14_15_4, \
call_7_14_15_5, \
call_7_14_15_6, \
call_7_14_15_7, \
call_7_14_15_8, \
call_7_14_15_9, \
call_7_14_15_10, \
call_7_14_15_11, \
call_7_14_15_12, \
call_7_14_15_13, \
call_7_14_15_14, \
call_7_14_15_15, \
call_7_15_0_0, \
call_7_15_0_1, \
call_7_15_0_2, \
call_7_15_0_3, \
call_7_15_0_4, \
call_7_15_0_5, \
call_7_15_0_6, \
call_7_15_0_7, \
call_7_15_0_8, \
call_7_15_0_9, \
call_7_15_0_10, \
call_7_15_0_11, \
call_7_15_0_12, \
call_7_15_0_13, \
call_7_15_0_14, \
call_7_15_0_15, \
call_7_15_1_0, \
call_7_15_1_1, \
call_7_15_1_2, \
call_7_15_1_3, \
call_7_15_1_4, \
call_7_15_1_5, \
call_7_15_1_6, \
call_7_15_1_7, \
call_7_15_1_8, \
call_7_15_1_9, \
call_7_15_1_10, \
call_7_15_1_11, \
call_7_15_1_12, \
call_7_15_1_13, \
call_7_15_1_14, \
call_7_15_1_15, \
call_7_15_2_0, \
call_7_15_2_1, \
call_7_15_2_2, \
call_7_15_2_3, \
call_7_15_2_4, \
call_7_15_2_5, \
call_7_15_2_6, \
call_7_15_2_7, \
call_7_15_2_8, \
call_7_15_2_9, \
call_7_15_2_10, \
call_7_15_2_11, \
call_7_15_2_12, \
call_7_15_2_13, \
call_7_15_2_14, \
call_7_15_2_15, \
call_7_15_3_0, \
call_7_15_3_1, \
call_7_15_3_2, \
call_7_15_3_3, \
call_7_15_3_4, \
call_7_15_3_5, \
call_7_15_3_6, \
call_7_15_3_7, \
call_7_15_3_8, \
call_7_15_3_9, \
call_7_15_3_10, \
call_7_15_3_11, \
call_7_15_3_12, \
call_7_15_3_13, \
call_7_15_3_14, \
call_7_15_3_15, \
call_7_15_4_0, \
call_7_15_4_1, \
call_7_15_4_2, \
call_7_15_4_3, \
call_7_15_4_4, \
call_7_15_4_5, \
call_7_15_4_6, \
call_7_15_4_7, \
call_7_15_4_8, \
call_7_15_4_9, \
call_7_15_4_10, \
call_7_15_4_11, \
call_7_15_4_12, \
call_7_15_4_13, \
call_7_15_4_14, \
call_7_15_4_15, \
call_7_15_5_0, \
call_7_15_5_1, \
call_7_15_5_2, \
call_7_15_5_3, \
call_7_15_5_4, \
call_7_15_5_5, \
call_7_15_5_6, \
call_7_15_5_7, \
call_7_15_5_8, \
call_7_15_5_9, \
call_7_15_5_10, \
call_7_15_5_11, \
call_7_15_5_12, \
call_7_15_5_13, \
call_7_15_5_14, \
call_7_15_5_15, \
call_7_15_6_0, \
call_7_15_6_1, \
call_7_15_6_2, \
call_7_15_6_3, \
call_7_15_6_4, \
call_7_15_6_5, \
call_7_15_6_6, \
call_7_15_6_7, \
call_7_15_6_8, \
call_7_15_6_9, \
call_7_15_6_10, \
call_7_15_6_11, \
call_7_15_6_12, \
call_7_15_6_13, \
call_7_15_6_14, \
call_7_15_6_15, \
call_7_15_7_0, \
call_7_15_7_1, \
call_7_15_7_2, \
call_7_15_7_3, \
call_7_15_7_4, \
call_7_15_7_5, \
call_7_15_7_6, \
call_7_15_7_7, \
call_7_15_7_8, \
call_7_15_7_9, \
call_7_15_7_10, \
call_7_15_7_11, \
call_7_15_7_12, \
call_7_15_7_13, \
call_7_15_7_14, \
call_7_15_7_15, \
call_7_15_8_0, \
call_7_15_8_1, \
call_7_15_8_2, \
call_7_15_8_3, \
call_7_15_8_4, \
call_7_15_8_5, \
call_7_15_8_6, \
call_7_15_8_7, \
call_7_15_8_8, \
call_7_15_8_9, \
call_7_15_8_10, \
call_7_15_8_11, \
call_7_15_8_12, \
call_7_15_8_13, \
call_7_15_8_14, \
call_7_15_8_15, \
call_7_15_9_0, \
call_7_15_9_1, \
call_7_15_9_2, \
call_7_15_9_3, \
call_7_15_9_4, \
call_7_15_9_5, \
call_7_15_9_6, \
call_7_15_9_7, \
call_7_15_9_8, \
call_7_15_9_9, \
call_7_15_9_10, \
call_7_15_9_11, \
call_7_15_9_12, \
call_7_15_9_13, \
call_7_15_9_14, \
call_7_15_9_15, \
call_7_15_10_0, \
call_7_15_10_1, \
call_7_15_10_2, \
call_7_15_10_3, \
call_7_15_10_4, \
call_7_15_10_5, \
call_7_15_10_6, \
call_7_15_10_7, \
call_7_15_10_8, \
call_7_15_10_9, \
call_7_15_10_10, \
call_7_15_10_11, \
call_7_15_10_12, \
call_7_15_10_13, \
call_7_15_10_14, \
call_7_15_10_15, \
call_7_15_11_0, \
call_7_15_11_1, \
call_7_15_11_2, \
call_7_15_11_3, \
call_7_15_11_4, \
call_7_15_11_5, \
call_7_15_11_6, \
call_7_15_11_7, \
call_7_15_11_8, \
call_7_15_11_9, \
call_7_15_11_10, \
call_7_15_11_11, \
call_7_15_11_12, \
call_7_15_11_13, \
call_7_15_11_14, \
call_7_15_11_15, \
call_7_15_12_0, \
call_7_15_12_1, \
call_7_15_12_2, \
call_7_15_12_3, \
call_7_15_12_4, \
call_7_15_12_5, \
call_7_15_12_6, \
call_7_15_12_7, \
call_7_15_12_8, \
call_7_15_12_9, \
call_7_15_12_10, \
call_7_15_12_11, \
call_7_15_12_12, \
call_7_15_12_13, \
call_7_15_12_14, \
call_7_15_12_15, \
call_7_15_13_0, \
call_7_15_13_1, \
call_7_15_13_2, \
call_7_15_13_3, \
call_7_15_13_4, \
call_7_15_13_5, \
call_7_15_13_6, \
call_7_15_13_7, \
call_7_15_13_8, \
call_7_15_13_9, \
call_7_15_13_10, \
call_7_15_13_11, \
call_7_15_13_12, \
call_7_15_13_13, \
call_7_15_13_14, \
call_7_15_13_15, \
call_7_15_14_0, \
call_7_15_14_1, \
call_7_15_14_2, \
call_7_15_14_3, \
call_7_15_14_4, \
call_7_15_14_5, \
call_7_15_14_6, \
call_7_15_14_7, \
call_7_15_14_8, \
call_7_15_14_9, \
call_7_15_14_10, \
call_7_15_14_11, \
call_7_15_14_12, \
call_7_15_14_13, \
call_7_15_14_14, \
call_7_15_14_15, \
call_7_15_15_0, \
call_7_15_15_1, \
call_7_15_15_2, \
call_7_15_15_3, \
call_7_15_15_4, \
call_7_15_15_5, \
call_7_15_15_6, \
call_7_15_15_7, \
call_7_15_15_8, \
call_7_15_15_9, \
call_7_15_15_10, \
call_7_15_15_11, \
call_7_15_15_12, \
call_7_15_15_13, \
call_7_15_15_14, \
call_7_15_15_15, \
call_8_0_0_0, \
call_8_0_0_1, \
call_8_0_0_2, \
call_8_0_0_3, \
call_8_0_0_4, \
call_8_0_0_5, \
call_8_0_0_6, \
call_8_0_0_7, \
call_8_0_0_8, \
call_8_0_0_9, \
call_8_0_0_10, \
call_8_0_0_11, \
call_8_0_0_12, \
call_8_0_0_13, \
call_8_0_0_14, \
call_8_0_0_15, \
call_8_0_1_0, \
call_8_0_1_1, \
call_8_0_1_2, \
call_8_0_1_3, \
call_8_0_1_4, \
call_8_0_1_5, \
call_8_0_1_6, \
call_8_0_1_7, \
call_8_0_1_8, \
call_8_0_1_9, \
call_8_0_1_10, \
call_8_0_1_11, \
call_8_0_1_12, \
call_8_0_1_13, \
call_8_0_1_14, \
call_8_0_1_15, \
call_8_0_2_0, \
call_8_0_2_1, \
call_8_0_2_2, \
call_8_0_2_3, \
call_8_0_2_4, \
call_8_0_2_5, \
call_8_0_2_6, \
call_8_0_2_7, \
call_8_0_2_8, \
call_8_0_2_9, \
call_8_0_2_10, \
call_8_0_2_11, \
call_8_0_2_12, \
call_8_0_2_13, \
call_8_0_2_14, \
call_8_0_2_15, \
call_8_0_3_0, \
call_8_0_3_1, \
call_8_0_3_2, \
call_8_0_3_3, \
call_8_0_3_4, \
call_8_0_3_5, \
call_8_0_3_6, \
call_8_0_3_7, \
call_8_0_3_8, \
call_8_0_3_9, \
call_8_0_3_10, \
call_8_0_3_11, \
call_8_0_3_12, \
call_8_0_3_13, \
call_8_0_3_14, \
call_8_0_3_15, \
call_8_0_4_0, \
call_8_0_4_1, \
call_8_0_4_2, \
call_8_0_4_3, \
call_8_0_4_4, \
call_8_0_4_5, \
call_8_0_4_6, \
call_8_0_4_7, \
call_8_0_4_8, \
call_8_0_4_9, \
call_8_0_4_10, \
call_8_0_4_11, \
call_8_0_4_12, \
call_8_0_4_13, \
call_8_0_4_14, \
call_8_0_4_15, \
call_8_0_5_0, \
call_8_0_5_1, \
call_8_0_5_2, \
call_8_0_5_3, \
call_8_0_5_4, \
call_8_0_5_5, \
call_8_0_5_6, \
call_8_0_5_7, \
call_8_0_5_8, \
call_8_0_5_9, \
call_8_0_5_10, \
call_8_0_5_11, \
call_8_0_5_12, \
call_8_0_5_13, \
call_8_0_5_14, \
call_8_0_5_15, \
call_8_0_6_0, \
call_8_0_6_1, \
call_8_0_6_2, \
call_8_0_6_3, \
call_8_0_6_4, \
call_8_0_6_5, \
call_8_0_6_6, \
call_8_0_6_7, \
call_8_0_6_8, \
call_8_0_6_9, \
call_8_0_6_10, \
call_8_0_6_11, \
call_8_0_6_12, \
call_8_0_6_13, \
call_8_0_6_14, \
call_8_0_6_15, \
call_8_0_7_0, \
call_8_0_7_1, \
call_8_0_7_2, \
call_8_0_7_3, \
call_8_0_7_4, \
call_8_0_7_5, \
call_8_0_7_6, \
call_8_0_7_7, \
call_8_0_7_8, \
call_8_0_7_9, \
call_8_0_7_10, \
call_8_0_7_11, \
call_8_0_7_12, \
call_8_0_7_13, \
call_8_0_7_14, \
call_8_0_7_15, \
call_8_0_8_0, \
call_8_0_8_1, \
call_8_0_8_2, \
call_8_0_8_3, \
call_8_0_8_4, \
call_8_0_8_5, \
call_8_0_8_6, \
call_8_0_8_7, \
call_8_0_8_8, \
call_8_0_8_9, \
call_8_0_8_10, \
call_8_0_8_11, \
call_8_0_8_12, \
call_8_0_8_13, \
call_8_0_8_14, \
call_8_0_8_15, \
call_8_0_9_0, \
call_8_0_9_1, \
call_8_0_9_2, \
call_8_0_9_3, \
call_8_0_9_4, \
call_8_0_9_5, \
call_8_0_9_6, \
call_8_0_9_7, \
call_8_0_9_8, \
call_8_0_9_9, \
call_8_0_9_10, \
call_8_0_9_11, \
call_8_0_9_12, \
call_8_0_9_13, \
call_8_0_9_14, \
call_8_0_9_15, \
call_8_0_10_0, \
call_8_0_10_1, \
call_8_0_10_2, \
call_8_0_10_3, \
call_8_0_10_4, \
call_8_0_10_5, \
call_8_0_10_6, \
call_8_0_10_7, \
call_8_0_10_8, \
call_8_0_10_9, \
call_8_0_10_10, \
call_8_0_10_11, \
call_8_0_10_12, \
call_8_0_10_13, \
call_8_0_10_14, \
call_8_0_10_15, \
call_8_0_11_0, \
call_8_0_11_1, \
call_8_0_11_2, \
call_8_0_11_3, \
call_8_0_11_4, \
call_8_0_11_5, \
call_8_0_11_6, \
call_8_0_11_7, \
call_8_0_11_8, \
call_8_0_11_9, \
call_8_0_11_10, \
call_8_0_11_11, \
call_8_0_11_12, \
call_8_0_11_13, \
call_8_0_11_14, \
call_8_0_11_15, \
call_8_0_12_0, \
call_8_0_12_1, \
call_8_0_12_2, \
call_8_0_12_3, \
call_8_0_12_4, \
call_8_0_12_5, \
call_8_0_12_6, \
call_8_0_12_7, \
call_8_0_12_8, \
call_8_0_12_9, \
call_8_0_12_10, \
call_8_0_12_11, \
call_8_0_12_12, \
call_8_0_12_13, \
call_8_0_12_14, \
call_8_0_12_15, \
call_8_0_13_0, \
call_8_0_13_1, \
call_8_0_13_2, \
call_8_0_13_3, \
call_8_0_13_4, \
call_8_0_13_5, \
call_8_0_13_6, \
call_8_0_13_7, \
call_8_0_13_8, \
call_8_0_13_9, \
call_8_0_13_10, \
call_8_0_13_11, \
call_8_0_13_12, \
call_8_0_13_13, \
call_8_0_13_14, \
call_8_0_13_15, \
call_8_0_14_0, \
call_8_0_14_1, \
call_8_0_14_2, \
call_8_0_14_3, \
call_8_0_14_4, \
call_8_0_14_5, \
call_8_0_14_6, \
call_8_0_14_7, \
call_8_0_14_8, \
call_8_0_14_9, \
call_8_0_14_10, \
call_8_0_14_11, \
call_8_0_14_12, \
call_8_0_14_13, \
call_8_0_14_14, \
call_8_0_14_15, \
call_8_0_15_0, \
call_8_0_15_1, \
call_8_0_15_2, \
call_8_0_15_3, \
call_8_0_15_4, \
call_8_0_15_5, \
call_8_0_15_6, \
call_8_0_15_7, \
call_8_0_15_8, \
call_8_0_15_9, \
call_8_0_15_10, \
call_8_0_15_11, \
call_8_0_15_12, \
call_8_0_15_13, \
call_8_0_15_14, \
call_8_0_15_15, \
call_8_1_0_0, \
call_8_1_0_1, \
call_8_1_0_2, \
call_8_1_0_3, \
call_8_1_0_4, \
call_8_1_0_5, \
call_8_1_0_6, \
call_8_1_0_7, \
call_8_1_0_8, \
call_8_1_0_9, \
call_8_1_0_10, \
call_8_1_0_11, \
call_8_1_0_12, \
call_8_1_0_13, \
call_8_1_0_14, \
call_8_1_0_15, \
call_8_1_1_0, \
call_8_1_1_1, \
call_8_1_1_2, \
call_8_1_1_3, \
call_8_1_1_4, \
call_8_1_1_5, \
call_8_1_1_6, \
call_8_1_1_7, \
call_8_1_1_8, \
call_8_1_1_9, \
call_8_1_1_10, \
call_8_1_1_11, \
call_8_1_1_12, \
call_8_1_1_13, \
call_8_1_1_14, \
call_8_1_1_15, \
call_8_1_2_0, \
call_8_1_2_1, \
call_8_1_2_2, \
call_8_1_2_3, \
call_8_1_2_4, \
call_8_1_2_5, \
call_8_1_2_6, \
call_8_1_2_7, \
call_8_1_2_8, \
call_8_1_2_9, \
call_8_1_2_10, \
call_8_1_2_11, \
call_8_1_2_12, \
call_8_1_2_13, \
call_8_1_2_14, \
call_8_1_2_15, \
call_8_1_3_0, \
call_8_1_3_1, \
call_8_1_3_2, \
call_8_1_3_3, \
call_8_1_3_4, \
call_8_1_3_5, \
call_8_1_3_6, \
call_8_1_3_7, \
call_8_1_3_8, \
call_8_1_3_9, \
call_8_1_3_10, \
call_8_1_3_11, \
call_8_1_3_12, \
call_8_1_3_13, \
call_8_1_3_14, \
call_8_1_3_15, \
call_8_1_4_0, \
call_8_1_4_1, \
call_8_1_4_2, \
call_8_1_4_3, \
call_8_1_4_4, \
call_8_1_4_5, \
call_8_1_4_6, \
call_8_1_4_7, \
call_8_1_4_8, \
call_8_1_4_9, \
call_8_1_4_10, \
call_8_1_4_11, \
call_8_1_4_12, \
call_8_1_4_13, \
call_8_1_4_14, \
call_8_1_4_15, \
call_8_1_5_0, \
call_8_1_5_1, \
call_8_1_5_2, \
call_8_1_5_3, \
call_8_1_5_4, \
call_8_1_5_5, \
call_8_1_5_6, \
call_8_1_5_7, \
call_8_1_5_8, \
call_8_1_5_9, \
call_8_1_5_10, \
call_8_1_5_11, \
call_8_1_5_12, \
call_8_1_5_13, \
call_8_1_5_14, \
call_8_1_5_15, \
call_8_1_6_0, \
call_8_1_6_1, \
call_8_1_6_2, \
call_8_1_6_3, \
call_8_1_6_4, \
call_8_1_6_5, \
call_8_1_6_6, \
call_8_1_6_7, \
call_8_1_6_8, \
call_8_1_6_9, \
call_8_1_6_10, \
call_8_1_6_11, \
call_8_1_6_12, \
call_8_1_6_13, \
call_8_1_6_14, \
call_8_1_6_15, \
call_8_1_7_0, \
call_8_1_7_1, \
call_8_1_7_2, \
call_8_1_7_3, \
call_8_1_7_4, \
call_8_1_7_5, \
call_8_1_7_6, \
call_8_1_7_7, \
call_8_1_7_8, \
call_8_1_7_9, \
call_8_1_7_10, \
call_8_1_7_11, \
call_8_1_7_12, \
call_8_1_7_13, \
call_8_1_7_14, \
call_8_1_7_15, \
call_8_1_8_0, \
call_8_1_8_1, \
call_8_1_8_2, \
call_8_1_8_3, \
call_8_1_8_4, \
call_8_1_8_5, \
call_8_1_8_6, \
call_8_1_8_7, \
call_8_1_8_8, \
call_8_1_8_9, \
call_8_1_8_10, \
call_8_1_8_11, \
call_8_1_8_12, \
call_8_1_8_13, \
call_8_1_8_14, \
call_8_1_8_15, \
call_8_1_9_0, \
call_8_1_9_1, \
call_8_1_9_2, \
call_8_1_9_3, \
call_8_1_9_4, \
call_8_1_9_5, \
call_8_1_9_6, \
call_8_1_9_7, \
call_8_1_9_8, \
call_8_1_9_9, \
call_8_1_9_10, \
call_8_1_9_11, \
call_8_1_9_12, \
call_8_1_9_13, \
call_8_1_9_14, \
call_8_1_9_15, \
call_8_1_10_0, \
call_8_1_10_1, \
call_8_1_10_2, \
call_8_1_10_3, \
call_8_1_10_4, \
call_8_1_10_5, \
call_8_1_10_6, \
call_8_1_10_7, \
call_8_1_10_8, \
call_8_1_10_9, \
call_8_1_10_10, \
call_8_1_10_11, \
call_8_1_10_12, \
call_8_1_10_13, \
call_8_1_10_14, \
call_8_1_10_15, \
call_8_1_11_0, \
call_8_1_11_1, \
call_8_1_11_2, \
call_8_1_11_3, \
call_8_1_11_4, \
call_8_1_11_5, \
call_8_1_11_6, \
call_8_1_11_7, \
call_8_1_11_8, \
call_8_1_11_9, \
call_8_1_11_10, \
call_8_1_11_11, \
call_8_1_11_12, \
call_8_1_11_13, \
call_8_1_11_14, \
call_8_1_11_15, \
call_8_1_12_0, \
call_8_1_12_1, \
call_8_1_12_2, \
call_8_1_12_3, \
call_8_1_12_4, \
call_8_1_12_5, \
call_8_1_12_6, \
call_8_1_12_7, \
call_8_1_12_8, \
call_8_1_12_9, \
call_8_1_12_10, \
call_8_1_12_11, \
call_8_1_12_12, \
call_8_1_12_13, \
call_8_1_12_14, \
call_8_1_12_15, \
call_8_1_13_0, \
call_8_1_13_1, \
call_8_1_13_2, \
call_8_1_13_3, \
call_8_1_13_4, \
call_8_1_13_5, \
call_8_1_13_6, \
call_8_1_13_7, \
call_8_1_13_8, \
call_8_1_13_9, \
call_8_1_13_10, \
call_8_1_13_11, \
call_8_1_13_12, \
call_8_1_13_13, \
call_8_1_13_14, \
call_8_1_13_15, \
call_8_1_14_0, \
call_8_1_14_1, \
call_8_1_14_2, \
call_8_1_14_3, \
call_8_1_14_4, \
call_8_1_14_5, \
call_8_1_14_6, \
call_8_1_14_7, \
call_8_1_14_8, \
call_8_1_14_9, \
call_8_1_14_10, \
call_8_1_14_11, \
call_8_1_14_12, \
call_8_1_14_13, \
call_8_1_14_14, \
call_8_1_14_15, \
call_8_1_15_0, \
call_8_1_15_1, \
call_8_1_15_2, \
call_8_1_15_3, \
call_8_1_15_4, \
call_8_1_15_5, \
call_8_1_15_6, \
call_8_1_15_7, \
call_8_1_15_8, \
call_8_1_15_9, \
call_8_1_15_10, \
call_8_1_15_11, \
call_8_1_15_12, \
call_8_1_15_13, \
call_8_1_15_14, \
call_8_1_15_15, \
call_8_2_0_0, \
call_8_2_0_1, \
call_8_2_0_2, \
call_8_2_0_3, \
call_8_2_0_4, \
call_8_2_0_5, \
call_8_2_0_6, \
call_8_2_0_7, \
call_8_2_0_8, \
call_8_2_0_9, \
call_8_2_0_10, \
call_8_2_0_11, \
call_8_2_0_12, \
call_8_2_0_13, \
call_8_2_0_14, \
call_8_2_0_15, \
call_8_2_1_0, \
call_8_2_1_1, \
call_8_2_1_2, \
call_8_2_1_3, \
call_8_2_1_4, \
call_8_2_1_5, \
call_8_2_1_6, \
call_8_2_1_7, \
call_8_2_1_8, \
call_8_2_1_9, \
call_8_2_1_10, \
call_8_2_1_11, \
call_8_2_1_12, \
call_8_2_1_13, \
call_8_2_1_14, \
call_8_2_1_15, \
call_8_2_2_0, \
call_8_2_2_1, \
call_8_2_2_2, \
call_8_2_2_3, \
call_8_2_2_4, \
call_8_2_2_5, \
call_8_2_2_6, \
call_8_2_2_7, \
call_8_2_2_8, \
call_8_2_2_9, \
call_8_2_2_10, \
call_8_2_2_11, \
call_8_2_2_12, \
call_8_2_2_13, \
call_8_2_2_14, \
call_8_2_2_15, \
call_8_2_3_0, \
call_8_2_3_1, \
call_8_2_3_2, \
call_8_2_3_3, \
call_8_2_3_4, \
call_8_2_3_5, \
call_8_2_3_6, \
call_8_2_3_7, \
call_8_2_3_8, \
call_8_2_3_9, \
call_8_2_3_10, \
call_8_2_3_11, \
call_8_2_3_12, \
call_8_2_3_13, \
call_8_2_3_14, \
call_8_2_3_15, \
call_8_2_4_0, \
call_8_2_4_1, \
call_8_2_4_2, \
call_8_2_4_3, \
call_8_2_4_4, \
call_8_2_4_5, \
call_8_2_4_6, \
call_8_2_4_7, \
call_8_2_4_8, \
call_8_2_4_9, \
call_8_2_4_10, \
call_8_2_4_11, \
call_8_2_4_12, \
call_8_2_4_13, \
call_8_2_4_14, \
call_8_2_4_15, \
call_8_2_5_0, \
call_8_2_5_1, \
call_8_2_5_2, \
call_8_2_5_3, \
call_8_2_5_4, \
call_8_2_5_5, \
call_8_2_5_6, \
call_8_2_5_7, \
call_8_2_5_8, \
call_8_2_5_9, \
call_8_2_5_10, \
call_8_2_5_11, \
call_8_2_5_12, \
call_8_2_5_13, \
call_8_2_5_14, \
call_8_2_5_15, \
call_8_2_6_0, \
call_8_2_6_1, \
call_8_2_6_2, \
call_8_2_6_3, \
call_8_2_6_4, \
call_8_2_6_5, \
call_8_2_6_6, \
call_8_2_6_7, \
call_8_2_6_8, \
call_8_2_6_9, \
call_8_2_6_10, \
call_8_2_6_11, \
call_8_2_6_12, \
call_8_2_6_13, \
call_8_2_6_14, \
call_8_2_6_15, \
call_8_2_7_0, \
call_8_2_7_1, \
call_8_2_7_2, \
call_8_2_7_3, \
call_8_2_7_4, \
call_8_2_7_5, \
call_8_2_7_6, \
call_8_2_7_7, \
call_8_2_7_8, \
call_8_2_7_9, \
call_8_2_7_10, \
call_8_2_7_11, \
call_8_2_7_12, \
call_8_2_7_13, \
call_8_2_7_14, \
call_8_2_7_15, \
call_8_2_8_0, \
call_8_2_8_1, \
call_8_2_8_2, \
call_8_2_8_3, \
call_8_2_8_4, \
call_8_2_8_5, \
call_8_2_8_6, \
call_8_2_8_7, \
call_8_2_8_8, \
call_8_2_8_9, \
call_8_2_8_10, \
call_8_2_8_11, \
call_8_2_8_12, \
call_8_2_8_13, \
call_8_2_8_14, \
call_8_2_8_15, \
call_8_2_9_0, \
call_8_2_9_1, \
call_8_2_9_2, \
call_8_2_9_3, \
call_8_2_9_4, \
call_8_2_9_5, \
call_8_2_9_6, \
call_8_2_9_7, \
call_8_2_9_8, \
call_8_2_9_9, \
call_8_2_9_10, \
call_8_2_9_11, \
call_8_2_9_12, \
call_8_2_9_13, \
call_8_2_9_14, \
call_8_2_9_15, \
call_8_2_10_0, \
call_8_2_10_1, \
call_8_2_10_2, \
call_8_2_10_3, \
call_8_2_10_4, \
call_8_2_10_5, \
call_8_2_10_6, \
call_8_2_10_7, \
call_8_2_10_8, \
call_8_2_10_9, \
call_8_2_10_10, \
call_8_2_10_11, \
call_8_2_10_12, \
call_8_2_10_13, \
call_8_2_10_14, \
call_8_2_10_15, \
call_8_2_11_0, \
call_8_2_11_1, \
call_8_2_11_2, \
call_8_2_11_3, \
call_8_2_11_4, \
call_8_2_11_5, \
call_8_2_11_6, \
call_8_2_11_7, \
call_8_2_11_8, \
call_8_2_11_9, \
call_8_2_11_10, \
call_8_2_11_11, \
call_8_2_11_12, \
call_8_2_11_13, \
call_8_2_11_14, \
call_8_2_11_15, \
call_8_2_12_0, \
call_8_2_12_1, \
call_8_2_12_2, \
call_8_2_12_3, \
call_8_2_12_4, \
call_8_2_12_5, \
call_8_2_12_6, \
call_8_2_12_7, \
call_8_2_12_8, \
call_8_2_12_9, \
call_8_2_12_10, \
call_8_2_12_11, \
call_8_2_12_12, \
call_8_2_12_13, \
call_8_2_12_14, \
call_8_2_12_15, \
call_8_2_13_0, \
call_8_2_13_1, \
call_8_2_13_2, \
call_8_2_13_3, \
call_8_2_13_4, \
call_8_2_13_5, \
call_8_2_13_6, \
call_8_2_13_7, \
call_8_2_13_8, \
call_8_2_13_9, \
call_8_2_13_10, \
call_8_2_13_11, \
call_8_2_13_12, \
call_8_2_13_13, \
call_8_2_13_14, \
call_8_2_13_15, \
call_8_2_14_0, \
call_8_2_14_1, \
call_8_2_14_2, \
call_8_2_14_3, \
call_8_2_14_4, \
call_8_2_14_5, \
call_8_2_14_6, \
call_8_2_14_7, \
call_8_2_14_8, \
call_8_2_14_9, \
call_8_2_14_10, \
call_8_2_14_11, \
call_8_2_14_12, \
call_8_2_14_13, \
call_8_2_14_14, \
call_8_2_14_15, \
call_8_2_15_0, \
call_8_2_15_1, \
call_8_2_15_2, \
call_8_2_15_3, \
call_8_2_15_4, \
call_8_2_15_5, \
call_8_2_15_6, \
call_8_2_15_7, \
call_8_2_15_8, \
call_8_2_15_9, \
call_8_2_15_10, \
call_8_2_15_11, \
call_8_2_15_12, \
call_8_2_15_13, \
call_8_2_15_14, \
call_8_2_15_15, \
call_8_3_0_0, \
call_8_3_0_1, \
call_8_3_0_2, \
call_8_3_0_3, \
call_8_3_0_4, \
call_8_3_0_5, \
call_8_3_0_6, \
call_8_3_0_7, \
call_8_3_0_8, \
call_8_3_0_9, \
call_8_3_0_10, \
call_8_3_0_11, \
call_8_3_0_12, \
call_8_3_0_13, \
call_8_3_0_14, \
call_8_3_0_15, \
call_8_3_1_0, \
call_8_3_1_1, \
call_8_3_1_2, \
call_8_3_1_3, \
call_8_3_1_4, \
call_8_3_1_5, \
call_8_3_1_6, \
call_8_3_1_7, \
call_8_3_1_8, \
call_8_3_1_9, \
call_8_3_1_10, \
call_8_3_1_11, \
call_8_3_1_12, \
call_8_3_1_13, \
call_8_3_1_14, \
call_8_3_1_15, \
call_8_3_2_0, \
call_8_3_2_1, \
call_8_3_2_2, \
call_8_3_2_3, \
call_8_3_2_4, \
call_8_3_2_5, \
call_8_3_2_6, \
call_8_3_2_7, \
call_8_3_2_8, \
call_8_3_2_9, \
call_8_3_2_10, \
call_8_3_2_11, \
call_8_3_2_12, \
call_8_3_2_13, \
call_8_3_2_14, \
call_8_3_2_15, \
call_8_3_3_0, \
call_8_3_3_1, \
call_8_3_3_2, \
call_8_3_3_3, \
call_8_3_3_4, \
call_8_3_3_5, \
call_8_3_3_6, \
call_8_3_3_7, \
call_8_3_3_8, \
call_8_3_3_9, \
call_8_3_3_10, \
call_8_3_3_11, \
call_8_3_3_12, \
call_8_3_3_13, \
call_8_3_3_14, \
call_8_3_3_15, \
call_8_3_4_0, \
call_8_3_4_1, \
call_8_3_4_2, \
call_8_3_4_3, \
call_8_3_4_4, \
call_8_3_4_5, \
call_8_3_4_6, \
call_8_3_4_7, \
call_8_3_4_8, \
call_8_3_4_9, \
call_8_3_4_10, \
call_8_3_4_11, \
call_8_3_4_12, \
call_8_3_4_13, \
call_8_3_4_14, \
call_8_3_4_15, \
call_8_3_5_0, \
call_8_3_5_1, \
call_8_3_5_2, \
call_8_3_5_3, \
call_8_3_5_4, \
call_8_3_5_5, \
call_8_3_5_6, \
call_8_3_5_7, \
call_8_3_5_8, \
call_8_3_5_9, \
call_8_3_5_10, \
call_8_3_5_11, \
call_8_3_5_12, \
call_8_3_5_13, \
call_8_3_5_14, \
call_8_3_5_15, \
call_8_3_6_0, \
call_8_3_6_1, \
call_8_3_6_2, \
call_8_3_6_3, \
call_8_3_6_4, \
call_8_3_6_5, \
call_8_3_6_6, \
call_8_3_6_7, \
call_8_3_6_8, \
call_8_3_6_9, \
call_8_3_6_10, \
call_8_3_6_11, \
call_8_3_6_12, \
call_8_3_6_13, \
call_8_3_6_14, \
call_8_3_6_15, \
call_8_3_7_0, \
call_8_3_7_1, \
call_8_3_7_2, \
call_8_3_7_3, \
call_8_3_7_4, \
call_8_3_7_5, \
call_8_3_7_6, \
call_8_3_7_7, \
call_8_3_7_8, \
call_8_3_7_9, \
call_8_3_7_10, \
call_8_3_7_11, \
call_8_3_7_12, \
call_8_3_7_13, \
call_8_3_7_14, \
call_8_3_7_15, \
call_8_3_8_0, \
call_8_3_8_1, \
call_8_3_8_2, \
call_8_3_8_3, \
call_8_3_8_4, \
call_8_3_8_5, \
call_8_3_8_6, \
call_8_3_8_7, \
call_8_3_8_8, \
call_8_3_8_9, \
call_8_3_8_10, \
call_8_3_8_11, \
call_8_3_8_12, \
call_8_3_8_13, \
call_8_3_8_14, \
call_8_3_8_15, \
call_8_3_9_0, \
call_8_3_9_1, \
call_8_3_9_2, \
call_8_3_9_3, \
call_8_3_9_4, \
call_8_3_9_5, \
call_8_3_9_6, \
call_8_3_9_7, \
call_8_3_9_8, \
call_8_3_9_9, \
call_8_3_9_10, \
call_8_3_9_11, \
call_8_3_9_12, \
call_8_3_9_13, \
call_8_3_9_14, \
call_8_3_9_15, \
call_8_3_10_0, \
call_8_3_10_1, \
call_8_3_10_2, \
call_8_3_10_3, \
call_8_3_10_4, \
call_8_3_10_5, \
call_8_3_10_6, \
call_8_3_10_7, \
call_8_3_10_8, \
call_8_3_10_9, \
call_8_3_10_10, \
call_8_3_10_11, \
call_8_3_10_12, \
call_8_3_10_13, \
call_8_3_10_14, \
call_8_3_10_15, \
call_8_3_11_0, \
call_8_3_11_1, \
call_8_3_11_2, \
call_8_3_11_3, \
call_8_3_11_4, \
call_8_3_11_5, \
call_8_3_11_6, \
call_8_3_11_7, \
call_8_3_11_8, \
call_8_3_11_9, \
call_8_3_11_10, \
call_8_3_11_11, \
call_8_3_11_12, \
call_8_3_11_13, \
call_8_3_11_14, \
call_8_3_11_15, \
call_8_3_12_0, \
call_8_3_12_1, \
call_8_3_12_2, \
call_8_3_12_3, \
call_8_3_12_4, \
call_8_3_12_5, \
call_8_3_12_6, \
call_8_3_12_7, \
call_8_3_12_8, \
call_8_3_12_9, \
call_8_3_12_10, \
call_8_3_12_11, \
call_8_3_12_12, \
call_8_3_12_13, \
call_8_3_12_14, \
call_8_3_12_15, \
call_8_3_13_0, \
call_8_3_13_1, \
call_8_3_13_2, \
call_8_3_13_3, \
call_8_3_13_4, \
call_8_3_13_5, \
call_8_3_13_6, \
call_8_3_13_7, \
call_8_3_13_8, \
call_8_3_13_9, \
call_8_3_13_10, \
call_8_3_13_11, \
call_8_3_13_12, \
call_8_3_13_13, \
call_8_3_13_14, \
call_8_3_13_15, \
call_8_3_14_0, \
call_8_3_14_1, \
call_8_3_14_2, \
call_8_3_14_3, \
call_8_3_14_4, \
call_8_3_14_5, \
call_8_3_14_6, \
call_8_3_14_7, \
call_8_3_14_8, \
call_8_3_14_9, \
call_8_3_14_10, \
call_8_3_14_11, \
call_8_3_14_12, \
call_8_3_14_13, \
call_8_3_14_14, \
call_8_3_14_15, \
call_8_3_15_0, \
call_8_3_15_1, \
call_8_3_15_2, \
call_8_3_15_3, \
call_8_3_15_4, \
call_8_3_15_5, \
call_8_3_15_6, \
call_8_3_15_7, \
call_8_3_15_8, \
call_8_3_15_9, \
call_8_3_15_10, \
call_8_3_15_11, \
call_8_3_15_12, \
call_8_3_15_13, \
call_8_3_15_14, \
call_8_3_15_15, \
call_8_4_0_0, \
call_8_4_0_1, \
call_8_4_0_2, \
call_8_4_0_3, \
call_8_4_0_4, \
call_8_4_0_5, \
call_8_4_0_6, \
call_8_4_0_7, \
call_8_4_0_8, \
call_8_4_0_9, \
call_8_4_0_10, \
call_8_4_0_11, \
call_8_4_0_12, \
call_8_4_0_13, \
call_8_4_0_14, \
call_8_4_0_15, \
call_8_4_1_0, \
call_8_4_1_1, \
call_8_4_1_2, \
call_8_4_1_3, \
call_8_4_1_4, \
call_8_4_1_5, \
call_8_4_1_6, \
call_8_4_1_7, \
call_8_4_1_8, \
call_8_4_1_9, \
call_8_4_1_10, \
call_8_4_1_11, \
call_8_4_1_12, \
call_8_4_1_13, \
call_8_4_1_14, \
call_8_4_1_15, \
call_8_4_2_0, \
call_8_4_2_1, \
call_8_4_2_2, \
call_8_4_2_3, \
call_8_4_2_4, \
call_8_4_2_5, \
call_8_4_2_6, \
call_8_4_2_7, \
call_8_4_2_8, \
call_8_4_2_9, \
call_8_4_2_10, \
call_8_4_2_11, \
call_8_4_2_12, \
call_8_4_2_13, \
call_8_4_2_14, \
call_8_4_2_15, \
call_8_4_3_0, \
call_8_4_3_1, \
call_8_4_3_2, \
call_8_4_3_3, \
call_8_4_3_4, \
call_8_4_3_5, \
call_8_4_3_6, \
call_8_4_3_7, \
call_8_4_3_8, \
call_8_4_3_9, \
call_8_4_3_10, \
call_8_4_3_11, \
call_8_4_3_12, \
call_8_4_3_13, \
call_8_4_3_14, \
call_8_4_3_15, \
call_8_4_4_0, \
call_8_4_4_1, \
call_8_4_4_2, \
call_8_4_4_3, \
call_8_4_4_4, \
call_8_4_4_5, \
call_8_4_4_6, \
call_8_4_4_7, \
call_8_4_4_8, \
call_8_4_4_9, \
call_8_4_4_10, \
call_8_4_4_11, \
call_8_4_4_12, \
call_8_4_4_13, \
call_8_4_4_14, \
call_8_4_4_15, \
call_8_4_5_0, \
call_8_4_5_1, \
call_8_4_5_2, \
call_8_4_5_3, \
call_8_4_5_4, \
call_8_4_5_5, \
call_8_4_5_6, \
call_8_4_5_7, \
call_8_4_5_8, \
call_8_4_5_9, \
call_8_4_5_10, \
call_8_4_5_11, \
call_8_4_5_12, \
call_8_4_5_13, \
call_8_4_5_14, \
call_8_4_5_15, \
call_8_4_6_0, \
call_8_4_6_1, \
call_8_4_6_2, \
call_8_4_6_3, \
call_8_4_6_4, \
call_8_4_6_5, \
call_8_4_6_6, \
call_8_4_6_7, \
call_8_4_6_8, \
call_8_4_6_9, \
call_8_4_6_10, \
call_8_4_6_11, \
call_8_4_6_12, \
call_8_4_6_13, \
call_8_4_6_14, \
call_8_4_6_15, \
call_8_4_7_0, \
call_8_4_7_1, \
call_8_4_7_2, \
call_8_4_7_3, \
call_8_4_7_4, \
call_8_4_7_5, \
call_8_4_7_6, \
call_8_4_7_7, \
call_8_4_7_8, \
call_8_4_7_9, \
call_8_4_7_10, \
call_8_4_7_11, \
call_8_4_7_12, \
call_8_4_7_13, \
call_8_4_7_14, \
call_8_4_7_15, \
call_8_4_8_0, \
call_8_4_8_1, \
call_8_4_8_2, \
call_8_4_8_3, \
call_8_4_8_4, \
call_8_4_8_5, \
call_8_4_8_6, \
call_8_4_8_7, \
call_8_4_8_8, \
call_8_4_8_9, \
call_8_4_8_10, \
call_8_4_8_11, \
call_8_4_8_12, \
call_8_4_8_13, \
call_8_4_8_14, \
call_8_4_8_15, \
call_8_4_9_0, \
call_8_4_9_1, \
call_8_4_9_2, \
call_8_4_9_3, \
call_8_4_9_4, \
call_8_4_9_5, \
call_8_4_9_6, \
call_8_4_9_7, \
call_8_4_9_8, \
call_8_4_9_9, \
call_8_4_9_10, \
call_8_4_9_11, \
call_8_4_9_12, \
call_8_4_9_13, \
call_8_4_9_14, \
call_8_4_9_15, \
call_8_4_10_0, \
call_8_4_10_1, \
call_8_4_10_2, \
call_8_4_10_3, \
call_8_4_10_4, \
call_8_4_10_5, \
call_8_4_10_6, \
call_8_4_10_7, \
call_8_4_10_8, \
call_8_4_10_9, \
call_8_4_10_10, \
call_8_4_10_11, \
call_8_4_10_12, \
call_8_4_10_13, \
call_8_4_10_14, \
call_8_4_10_15, \
call_8_4_11_0, \
call_8_4_11_1, \
call_8_4_11_2, \
call_8_4_11_3, \
call_8_4_11_4, \
call_8_4_11_5, \
call_8_4_11_6, \
call_8_4_11_7, \
call_8_4_11_8, \
call_8_4_11_9, \
call_8_4_11_10, \
call_8_4_11_11, \
call_8_4_11_12, \
call_8_4_11_13, \
call_8_4_11_14, \
call_8_4_11_15, \
call_8_4_12_0, \
call_8_4_12_1, \
call_8_4_12_2, \
call_8_4_12_3, \
call_8_4_12_4, \
call_8_4_12_5, \
call_8_4_12_6, \
call_8_4_12_7, \
call_8_4_12_8, \
call_8_4_12_9, \
call_8_4_12_10, \
call_8_4_12_11, \
call_8_4_12_12, \
call_8_4_12_13, \
call_8_4_12_14, \
call_8_4_12_15, \
call_8_4_13_0, \
call_8_4_13_1, \
call_8_4_13_2, \
call_8_4_13_3, \
call_8_4_13_4, \
call_8_4_13_5, \
call_8_4_13_6, \
call_8_4_13_7, \
call_8_4_13_8, \
call_8_4_13_9, \
call_8_4_13_10, \
call_8_4_13_11, \
call_8_4_13_12, \
call_8_4_13_13, \
call_8_4_13_14, \
call_8_4_13_15, \
call_8_4_14_0, \
call_8_4_14_1, \
call_8_4_14_2, \
call_8_4_14_3, \
call_8_4_14_4, \
call_8_4_14_5, \
call_8_4_14_6, \
call_8_4_14_7, \
call_8_4_14_8, \
call_8_4_14_9, \
call_8_4_14_10, \
call_8_4_14_11, \
call_8_4_14_12, \
call_8_4_14_13, \
call_8_4_14_14, \
call_8_4_14_15, \
call_8_4_15_0, \
call_8_4_15_1, \
call_8_4_15_2, \
call_8_4_15_3, \
call_8_4_15_4, \
call_8_4_15_5, \
call_8_4_15_6, \
call_8_4_15_7, \
call_8_4_15_8, \
call_8_4_15_9, \
call_8_4_15_10, \
call_8_4_15_11, \
call_8_4_15_12, \
call_8_4_15_13, \
call_8_4_15_14, \
call_8_4_15_15, \
call_8_5_0_0, \
call_8_5_0_1, \
call_8_5_0_2, \
call_8_5_0_3, \
call_8_5_0_4, \
call_8_5_0_5, \
call_8_5_0_6, \
call_8_5_0_7, \
call_8_5_0_8, \
call_8_5_0_9, \
call_8_5_0_10, \
call_8_5_0_11, \
call_8_5_0_12, \
call_8_5_0_13, \
call_8_5_0_14, \
call_8_5_0_15, \
call_8_5_1_0, \
call_8_5_1_1, \
call_8_5_1_2, \
call_8_5_1_3, \
call_8_5_1_4, \
call_8_5_1_5, \
call_8_5_1_6, \
call_8_5_1_7, \
call_8_5_1_8, \
call_8_5_1_9, \
call_8_5_1_10, \
call_8_5_1_11, \
call_8_5_1_12, \
call_8_5_1_13, \
call_8_5_1_14, \
call_8_5_1_15, \
call_8_5_2_0, \
call_8_5_2_1, \
call_8_5_2_2, \
call_8_5_2_3, \
call_8_5_2_4, \
call_8_5_2_5, \
call_8_5_2_6, \
call_8_5_2_7, \
call_8_5_2_8, \
call_8_5_2_9, \
call_8_5_2_10, \
call_8_5_2_11, \
call_8_5_2_12, \
call_8_5_2_13, \
call_8_5_2_14, \
call_8_5_2_15, \
call_8_5_3_0, \
call_8_5_3_1, \
call_8_5_3_2, \
call_8_5_3_3, \
call_8_5_3_4, \
call_8_5_3_5, \
call_8_5_3_6, \
call_8_5_3_7, \
call_8_5_3_8, \
call_8_5_3_9, \
call_8_5_3_10, \
call_8_5_3_11, \
call_8_5_3_12, \
call_8_5_3_13, \
call_8_5_3_14, \
call_8_5_3_15, \
call_8_5_4_0, \
call_8_5_4_1, \
call_8_5_4_2, \
call_8_5_4_3, \
call_8_5_4_4, \
call_8_5_4_5, \
call_8_5_4_6, \
call_8_5_4_7, \
call_8_5_4_8, \
call_8_5_4_9, \
call_8_5_4_10, \
call_8_5_4_11, \
call_8_5_4_12, \
call_8_5_4_13, \
call_8_5_4_14, \
call_8_5_4_15, \
call_8_5_5_0, \
call_8_5_5_1, \
call_8_5_5_2, \
call_8_5_5_3, \
call_8_5_5_4, \
call_8_5_5_5, \
call_8_5_5_6, \
call_8_5_5_7, \
call_8_5_5_8, \
call_8_5_5_9, \
call_8_5_5_10, \
call_8_5_5_11, \
call_8_5_5_12, \
call_8_5_5_13, \
call_8_5_5_14, \
call_8_5_5_15, \
call_8_5_6_0, \
call_8_5_6_1, \
call_8_5_6_2, \
call_8_5_6_3, \
call_8_5_6_4, \
call_8_5_6_5, \
call_8_5_6_6, \
call_8_5_6_7, \
call_8_5_6_8, \
call_8_5_6_9, \
call_8_5_6_10, \
call_8_5_6_11, \
call_8_5_6_12, \
call_8_5_6_13, \
call_8_5_6_14, \
call_8_5_6_15, \
call_8_5_7_0, \
call_8_5_7_1, \
call_8_5_7_2, \
call_8_5_7_3, \
call_8_5_7_4, \
call_8_5_7_5, \
call_8_5_7_6, \
call_8_5_7_7, \
call_8_5_7_8, \
call_8_5_7_9, \
call_8_5_7_10, \
call_8_5_7_11, \
call_8_5_7_12, \
call_8_5_7_13, \
call_8_5_7_14, \
call_8_5_7_15, \
call_8_5_8_0, \
call_8_5_8_1, \
call_8_5_8_2, \
call_8_5_8_3, \
call_8_5_8_4, \
call_8_5_8_5, \
call_8_5_8_6, \
call_8_5_8_7, \
call_8_5_8_8, \
call_8_5_8_9, \
call_8_5_8_10, \
call_8_5_8_11, \
call_8_5_8_12, \
call_8_5_8_13, \
call_8_5_8_14, \
call_8_5_8_15, \
call_8_5_9_0, \
call_8_5_9_1, \
call_8_5_9_2, \
call_8_5_9_3, \
call_8_5_9_4, \
call_8_5_9_5, \
call_8_5_9_6, \
call_8_5_9_7, \
call_8_5_9_8, \
call_8_5_9_9, \
call_8_5_9_10, \
call_8_5_9_11, \
call_8_5_9_12, \
call_8_5_9_13, \
call_8_5_9_14, \
call_8_5_9_15, \
call_8_5_10_0, \
call_8_5_10_1, \
call_8_5_10_2, \
call_8_5_10_3, \
call_8_5_10_4, \
call_8_5_10_5, \
call_8_5_10_6, \
call_8_5_10_7, \
call_8_5_10_8, \
call_8_5_10_9, \
call_8_5_10_10, \
call_8_5_10_11, \
call_8_5_10_12, \
call_8_5_10_13, \
call_8_5_10_14, \
call_8_5_10_15, \
call_8_5_11_0, \
call_8_5_11_1, \
call_8_5_11_2, \
call_8_5_11_3, \
call_8_5_11_4, \
call_8_5_11_5, \
call_8_5_11_6, \
call_8_5_11_7, \
call_8_5_11_8, \
call_8_5_11_9, \
call_8_5_11_10, \
call_8_5_11_11, \
call_8_5_11_12, \
call_8_5_11_13, \
call_8_5_11_14, \
call_8_5_11_15, \
call_8_5_12_0, \
call_8_5_12_1, \
call_8_5_12_2, \
call_8_5_12_3, \
call_8_5_12_4, \
call_8_5_12_5, \
call_8_5_12_6, \
call_8_5_12_7, \
call_8_5_12_8, \
call_8_5_12_9, \
call_8_5_12_10, \
call_8_5_12_11, \
call_8_5_12_12, \
call_8_5_12_13, \
call_8_5_12_14, \
call_8_5_12_15, \
call_8_5_13_0, \
call_8_5_13_1, \
call_8_5_13_2, \
call_8_5_13_3, \
call_8_5_13_4, \
call_8_5_13_5, \
call_8_5_13_6, \
call_8_5_13_7, \
call_8_5_13_8, \
call_8_5_13_9, \
call_8_5_13_10, \
call_8_5_13_11, \
call_8_5_13_12, \
call_8_5_13_13, \
call_8_5_13_14, \
call_8_5_13_15, \
call_8_5_14_0, \
call_8_5_14_1, \
call_8_5_14_2, \
call_8_5_14_3, \
call_8_5_14_4, \
call_8_5_14_5, \
call_8_5_14_6, \
call_8_5_14_7, \
call_8_5_14_8, \
call_8_5_14_9, \
call_8_5_14_10, \
call_8_5_14_11, \
call_8_5_14_12, \
call_8_5_14_13, \
call_8_5_14_14, \
call_8_5_14_15, \
call_8_5_15_0, \
call_8_5_15_1, \
call_8_5_15_2, \
call_8_5_15_3, \
call_8_5_15_4, \
call_8_5_15_5, \
call_8_5_15_6, \
call_8_5_15_7, \
call_8_5_15_8, \
call_8_5_15_9, \
call_8_5_15_10, \
call_8_5_15_11, \
call_8_5_15_12, \
call_8_5_15_13, \
call_8_5_15_14, \
call_8_5_15_15, \
call_8_6_0_0, \
call_8_6_0_1, \
call_8_6_0_2, \
call_8_6_0_3, \
call_8_6_0_4, \
call_8_6_0_5, \
call_8_6_0_6, \
call_8_6_0_7, \
call_8_6_0_8, \
call_8_6_0_9, \
call_8_6_0_10, \
call_8_6_0_11, \
call_8_6_0_12, \
call_8_6_0_13, \
call_8_6_0_14, \
call_8_6_0_15, \
call_8_6_1_0, \
call_8_6_1_1, \
call_8_6_1_2, \
call_8_6_1_3, \
call_8_6_1_4, \
call_8_6_1_5, \
call_8_6_1_6, \
call_8_6_1_7, \
call_8_6_1_8, \
call_8_6_1_9, \
call_8_6_1_10, \
call_8_6_1_11, \
call_8_6_1_12, \
call_8_6_1_13, \
call_8_6_1_14, \
call_8_6_1_15, \
call_8_6_2_0, \
call_8_6_2_1, \
call_8_6_2_2, \
call_8_6_2_3, \
call_8_6_2_4, \
call_8_6_2_5, \
call_8_6_2_6, \
call_8_6_2_7, \
call_8_6_2_8, \
call_8_6_2_9, \
call_8_6_2_10, \
call_8_6_2_11, \
call_8_6_2_12, \
call_8_6_2_13, \
call_8_6_2_14, \
call_8_6_2_15, \
call_8_6_3_0, \
call_8_6_3_1, \
call_8_6_3_2, \
call_8_6_3_3, \
call_8_6_3_4, \
call_8_6_3_5, \
call_8_6_3_6, \
call_8_6_3_7, \
call_8_6_3_8, \
call_8_6_3_9, \
call_8_6_3_10, \
call_8_6_3_11, \
call_8_6_3_12, \
call_8_6_3_13, \
call_8_6_3_14, \
call_8_6_3_15, \
call_8_6_4_0, \
call_8_6_4_1, \
call_8_6_4_2, \
call_8_6_4_3, \
call_8_6_4_4, \
call_8_6_4_5, \
call_8_6_4_6, \
call_8_6_4_7, \
call_8_6_4_8, \
call_8_6_4_9, \
call_8_6_4_10, \
call_8_6_4_11, \
call_8_6_4_12, \
call_8_6_4_13, \
call_8_6_4_14, \
call_8_6_4_15, \
call_8_6_5_0, \
call_8_6_5_1, \
call_8_6_5_2, \
call_8_6_5_3, \
call_8_6_5_4, \
call_8_6_5_5, \
call_8_6_5_6, \
call_8_6_5_7, \
call_8_6_5_8, \
call_8_6_5_9, \
call_8_6_5_10, \
call_8_6_5_11, \
call_8_6_5_12, \
call_8_6_5_13, \
call_8_6_5_14, \
call_8_6_5_15, \
call_8_6_6_0, \
call_8_6_6_1, \
call_8_6_6_2, \
call_8_6_6_3, \
call_8_6_6_4, \
call_8_6_6_5, \
call_8_6_6_6, \
call_8_6_6_7, \
call_8_6_6_8, \
call_8_6_6_9, \
call_8_6_6_10, \
call_8_6_6_11, \
call_8_6_6_12, \
call_8_6_6_13, \
call_8_6_6_14, \
call_8_6_6_15, \
call_8_6_7_0, \
call_8_6_7_1, \
call_8_6_7_2, \
call_8_6_7_3, \
call_8_6_7_4, \
call_8_6_7_5, \
call_8_6_7_6, \
call_8_6_7_7, \
call_8_6_7_8, \
call_8_6_7_9, \
call_8_6_7_10, \
call_8_6_7_11, \
call_8_6_7_12, \
call_8_6_7_13, \
call_8_6_7_14, \
call_8_6_7_15, \
call_8_6_8_0, \
call_8_6_8_1, \
call_8_6_8_2, \
call_8_6_8_3, \
call_8_6_8_4, \
call_8_6_8_5, \
call_8_6_8_6, \
call_8_6_8_7, \
call_8_6_8_8, \
call_8_6_8_9, \
call_8_6_8_10, \
call_8_6_8_11, \
call_8_6_8_12, \
call_8_6_8_13, \
call_8_6_8_14, \
call_8_6_8_15, \
call_8_6_9_0, \
call_8_6_9_1, \
call_8_6_9_2, \
call_8_6_9_3, \
call_8_6_9_4, \
call_8_6_9_5, \
call_8_6_9_6, \
call_8_6_9_7, \
call_8_6_9_8, \
call_8_6_9_9, \
call_8_6_9_10, \
call_8_6_9_11, \
call_8_6_9_12, \
call_8_6_9_13, \
call_8_6_9_14, \
call_8_6_9_15, \
call_8_6_10_0, \
call_8_6_10_1, \
call_8_6_10_2, \
call_8_6_10_3, \
call_8_6_10_4, \
call_8_6_10_5, \
call_8_6_10_6, \
call_8_6_10_7, \
call_8_6_10_8, \
call_8_6_10_9, \
call_8_6_10_10, \
call_8_6_10_11, \
call_8_6_10_12, \
call_8_6_10_13, \
call_8_6_10_14, \
call_8_6_10_15, \
call_8_6_11_0, \
call_8_6_11_1, \
call_8_6_11_2, \
call_8_6_11_3, \
call_8_6_11_4, \
call_8_6_11_5, \
call_8_6_11_6, \
call_8_6_11_7, \
call_8_6_11_8, \
call_8_6_11_9, \
call_8_6_11_10, \
call_8_6_11_11, \
call_8_6_11_12, \
call_8_6_11_13, \
call_8_6_11_14, \
call_8_6_11_15, \
call_8_6_12_0, \
call_8_6_12_1, \
call_8_6_12_2, \
call_8_6_12_3, \
call_8_6_12_4, \
call_8_6_12_5, \
call_8_6_12_6, \
call_8_6_12_7, \
call_8_6_12_8, \
call_8_6_12_9, \
call_8_6_12_10, \
call_8_6_12_11, \
call_8_6_12_12, \
call_8_6_12_13, \
call_8_6_12_14, \
call_8_6_12_15, \
call_8_6_13_0, \
call_8_6_13_1, \
call_8_6_13_2, \
call_8_6_13_3, \
call_8_6_13_4, \
call_8_6_13_5, \
call_8_6_13_6, \
call_8_6_13_7, \
call_8_6_13_8, \
call_8_6_13_9, \
call_8_6_13_10, \
call_8_6_13_11, \
call_8_6_13_12, \
call_8_6_13_13, \
call_8_6_13_14, \
call_8_6_13_15, \
call_8_6_14_0, \
call_8_6_14_1, \
call_8_6_14_2, \
call_8_6_14_3, \
call_8_6_14_4, \
call_8_6_14_5, \
call_8_6_14_6, \
call_8_6_14_7, \
call_8_6_14_8, \
call_8_6_14_9, \
call_8_6_14_10, \
call_8_6_14_11, \
call_8_6_14_12, \
call_8_6_14_13, \
call_8_6_14_14, \
call_8_6_14_15, \
call_8_6_15_0, \
call_8_6_15_1, \
call_8_6_15_2, \
call_8_6_15_3, \
call_8_6_15_4, \
call_8_6_15_5, \
call_8_6_15_6, \
call_8_6_15_7, \
call_8_6_15_8, \
call_8_6_15_9, \
call_8_6_15_10, \
call_8_6_15_11, \
call_8_6_15_12, \
call_8_6_15_13, \
call_8_6_15_14, \
call_8_6_15_15, \
call_8_7_0_0, \
call_8_7_0_1, \
call_8_7_0_2, \
call_8_7_0_3, \
call_8_7_0_4, \
call_8_7_0_5, \
call_8_7_0_6, \
call_8_7_0_7, \
call_8_7_0_8, \
call_8_7_0_9, \
call_8_7_0_10, \
call_8_7_0_11, \
call_8_7_0_12, \
call_8_7_0_13, \
call_8_7_0_14, \
call_8_7_0_15, \
call_8_7_1_0, \
call_8_7_1_1, \
call_8_7_1_2, \
call_8_7_1_3, \
call_8_7_1_4, \
call_8_7_1_5, \
call_8_7_1_6, \
call_8_7_1_7, \
call_8_7_1_8, \
call_8_7_1_9, \
call_8_7_1_10, \
call_8_7_1_11, \
call_8_7_1_12, \
call_8_7_1_13, \
call_8_7_1_14, \
call_8_7_1_15, \
call_8_7_2_0, \
call_8_7_2_1, \
call_8_7_2_2, \
call_8_7_2_3, \
call_8_7_2_4, \
call_8_7_2_5, \
call_8_7_2_6, \
call_8_7_2_7, \
call_8_7_2_8, \
call_8_7_2_9, \
call_8_7_2_10, \
call_8_7_2_11, \
call_8_7_2_12, \
call_8_7_2_13, \
call_8_7_2_14, \
call_8_7_2_15, \
call_8_7_3_0, \
call_8_7_3_1, \
call_8_7_3_2, \
call_8_7_3_3, \
call_8_7_3_4, \
call_8_7_3_5, \
call_8_7_3_6, \
call_8_7_3_7, \
call_8_7_3_8, \
call_8_7_3_9, \
call_8_7_3_10, \
call_8_7_3_11, \
call_8_7_3_12, \
call_8_7_3_13, \
call_8_7_3_14, \
call_8_7_3_15, \
call_8_7_4_0, \
call_8_7_4_1, \
call_8_7_4_2, \
call_8_7_4_3, \
call_8_7_4_4, \
call_8_7_4_5, \
call_8_7_4_6, \
call_8_7_4_7, \
call_8_7_4_8, \
call_8_7_4_9, \
call_8_7_4_10, \
call_8_7_4_11, \
call_8_7_4_12, \
call_8_7_4_13, \
call_8_7_4_14, \
call_8_7_4_15, \
call_8_7_5_0, \
call_8_7_5_1, \
call_8_7_5_2, \
call_8_7_5_3, \
call_8_7_5_4, \
call_8_7_5_5, \
call_8_7_5_6, \
call_8_7_5_7, \
call_8_7_5_8, \
call_8_7_5_9, \
call_8_7_5_10, \
call_8_7_5_11, \
call_8_7_5_12, \
call_8_7_5_13, \
call_8_7_5_14, \
call_8_7_5_15, \
call_8_7_6_0, \
call_8_7_6_1, \
call_8_7_6_2, \
call_8_7_6_3, \
call_8_7_6_4, \
call_8_7_6_5, \
call_8_7_6_6, \
call_8_7_6_7, \
call_8_7_6_8, \
call_8_7_6_9, \
call_8_7_6_10, \
call_8_7_6_11, \
call_8_7_6_12, \
call_8_7_6_13, \
call_8_7_6_14, \
call_8_7_6_15, \
call_8_7_7_0, \
call_8_7_7_1, \
call_8_7_7_2, \
call_8_7_7_3, \
call_8_7_7_4, \
call_8_7_7_5, \
call_8_7_7_6, \
call_8_7_7_7, \
call_8_7_7_8, \
call_8_7_7_9, \
call_8_7_7_10, \
call_8_7_7_11, \
call_8_7_7_12, \
call_8_7_7_13, \
call_8_7_7_14, \
call_8_7_7_15, \
call_8_7_8_0, \
call_8_7_8_1, \
call_8_7_8_2, \
call_8_7_8_3, \
call_8_7_8_4, \
call_8_7_8_5, \
call_8_7_8_6, \
call_8_7_8_7, \
call_8_7_8_8, \
call_8_7_8_9, \
call_8_7_8_10, \
call_8_7_8_11, \
call_8_7_8_12, \
call_8_7_8_13, \
call_8_7_8_14, \
call_8_7_8_15, \
call_8_7_9_0, \
call_8_7_9_1, \
call_8_7_9_2, \
call_8_7_9_3, \
call_8_7_9_4, \
call_8_7_9_5, \
call_8_7_9_6, \
call_8_7_9_7, \
call_8_7_9_8, \
call_8_7_9_9, \
call_8_7_9_10, \
call_8_7_9_11, \
call_8_7_9_12, \
call_8_7_9_13, \
call_8_7_9_14, \
call_8_7_9_15, \
call_8_7_10_0, \
call_8_7_10_1, \
call_8_7_10_2, \
call_8_7_10_3, \
call_8_7_10_4, \
call_8_7_10_5, \
call_8_7_10_6, \
call_8_7_10_7, \
call_8_7_10_8, \
call_8_7_10_9, \
call_8_7_10_10, \
call_8_7_10_11, \
call_8_7_10_12, \
call_8_7_10_13, \
call_8_7_10_14, \
call_8_7_10_15, \
call_8_7_11_0, \
call_8_7_11_1, \
call_8_7_11_2, \
call_8_7_11_3, \
call_8_7_11_4, \
call_8_7_11_5, \
call_8_7_11_6, \
call_8_7_11_7, \
call_8_7_11_8, \
call_8_7_11_9, \
call_8_7_11_10, \
call_8_7_11_11, \
call_8_7_11_12, \
call_8_7_11_13, \
call_8_7_11_14, \
call_8_7_11_15, \
call_8_7_12_0, \
call_8_7_12_1, \
call_8_7_12_2, \
call_8_7_12_3, \
call_8_7_12_4, \
call_8_7_12_5, \
call_8_7_12_6, \
call_8_7_12_7, \
call_8_7_12_8, \
call_8_7_12_9, \
call_8_7_12_10, \
call_8_7_12_11, \
call_8_7_12_12, \
call_8_7_12_13, \
call_8_7_12_14, \
call_8_7_12_15, \
call_8_7_13_0, \
call_8_7_13_1, \
call_8_7_13_2, \
call_8_7_13_3, \
call_8_7_13_4, \
call_8_7_13_5, \
call_8_7_13_6, \
call_8_7_13_7, \
call_8_7_13_8, \
call_8_7_13_9, \
call_8_7_13_10, \
call_8_7_13_11, \
call_8_7_13_12, \
call_8_7_13_13, \
call_8_7_13_14, \
call_8_7_13_15, \
call_8_7_14_0, \
call_8_7_14_1, \
call_8_7_14_2, \
call_8_7_14_3, \
call_8_7_14_4, \
call_8_7_14_5, \
call_8_7_14_6, \
call_8_7_14_7, \
call_8_7_14_8, \
call_8_7_14_9, \
call_8_7_14_10, \
call_8_7_14_11, \
call_8_7_14_12, \
call_8_7_14_13, \
call_8_7_14_14, \
call_8_7_14_15, \
call_8_7_15_0, \
call_8_7_15_1, \
call_8_7_15_2, \
call_8_7_15_3, \
call_8_7_15_4, \
call_8_7_15_5, \
call_8_7_15_6, \
call_8_7_15_7, \
call_8_7_15_8, \
call_8_7_15_9, \
call_8_7_15_10, \
call_8_7_15_11, \
call_8_7_15_12, \
call_8_7_15_13, \
call_8_7_15_14, \
call_8_7_15_15, \
call_8_8_0_0, \
call_8_8_0_1, \
call_8_8_0_2, \
call_8_8_0_3, \
call_8_8_0_4, \
call_8_8_0_5, \
call_8_8_0_6, \
call_8_8_0_7, \
call_8_8_0_8, \
call_8_8_0_9, \
call_8_8_0_10, \
call_8_8_0_11, \
call_8_8_0_12, \
call_8_8_0_13, \
call_8_8_0_14, \
call_8_8_0_15, \
call_8_8_1_0, \
call_8_8_1_1, \
call_8_8_1_2, \
call_8_8_1_3, \
call_8_8_1_4, \
call_8_8_1_5, \
call_8_8_1_6, \
call_8_8_1_7, \
call_8_8_1_8, \
call_8_8_1_9, \
call_8_8_1_10, \
call_8_8_1_11, \
call_8_8_1_12, \
call_8_8_1_13, \
call_8_8_1_14, \
call_8_8_1_15, \
call_8_8_2_0, \
call_8_8_2_1, \
call_8_8_2_2, \
call_8_8_2_3, \
call_8_8_2_4, \
call_8_8_2_5, \
call_8_8_2_6, \
call_8_8_2_7, \
call_8_8_2_8, \
call_8_8_2_9, \
call_8_8_2_10, \
call_8_8_2_11, \
call_8_8_2_12, \
call_8_8_2_13, \
call_8_8_2_14, \
call_8_8_2_15, \
call_8_8_3_0, \
call_8_8_3_1, \
call_8_8_3_2, \
call_8_8_3_3, \
call_8_8_3_4, \
call_8_8_3_5, \
call_8_8_3_6, \
call_8_8_3_7, \
call_8_8_3_8, \
call_8_8_3_9, \
call_8_8_3_10, \
call_8_8_3_11, \
call_8_8_3_12, \
call_8_8_3_13, \
call_8_8_3_14, \
call_8_8_3_15, \
call_8_8_4_0, \
call_8_8_4_1, \
call_8_8_4_2, \
call_8_8_4_3, \
call_8_8_4_4, \
call_8_8_4_5, \
call_8_8_4_6, \
call_8_8_4_7, \
call_8_8_4_8, \
call_8_8_4_9, \
call_8_8_4_10, \
call_8_8_4_11, \
call_8_8_4_12, \
call_8_8_4_13, \
call_8_8_4_14, \
call_8_8_4_15, \
call_8_8_5_0, \
call_8_8_5_1, \
call_8_8_5_2, \
call_8_8_5_3, \
call_8_8_5_4, \
call_8_8_5_5, \
call_8_8_5_6, \
call_8_8_5_7, \
call_8_8_5_8, \
call_8_8_5_9, \
call_8_8_5_10, \
call_8_8_5_11, \
call_8_8_5_12, \
call_8_8_5_13, \
call_8_8_5_14, \
call_8_8_5_15, \
call_8_8_6_0, \
call_8_8_6_1, \
call_8_8_6_2, \
call_8_8_6_3, \
call_8_8_6_4, \
call_8_8_6_5, \
call_8_8_6_6, \
call_8_8_6_7, \
call_8_8_6_8, \
call_8_8_6_9, \
call_8_8_6_10, \
call_8_8_6_11, \
call_8_8_6_12, \
call_8_8_6_13, \
call_8_8_6_14, \
call_8_8_6_15, \
call_8_8_7_0, \
call_8_8_7_1, \
call_8_8_7_2, \
call_8_8_7_3, \
call_8_8_7_4, \
call_8_8_7_5, \
call_8_8_7_6, \
call_8_8_7_7, \
call_8_8_7_8, \
call_8_8_7_9, \
call_8_8_7_10, \
call_8_8_7_11, \
call_8_8_7_12, \
call_8_8_7_13, \
call_8_8_7_14, \
call_8_8_7_15, \
call_8_8_8_0, \
call_8_8_8_1, \
call_8_8_8_2, \
call_8_8_8_3, \
call_8_8_8_4, \
call_8_8_8_5, \
call_8_8_8_6, \
call_8_8_8_7, \
call_8_8_8_8, \
call_8_8_8_9, \
call_8_8_8_10, \
call_8_8_8_11, \
call_8_8_8_12, \
call_8_8_8_13, \
call_8_8_8_14, \
call_8_8_8_15, \
call_8_8_9_0, \
call_8_8_9_1, \
call_8_8_9_2, \
call_8_8_9_3, \
call_8_8_9_4, \
call_8_8_9_5, \
call_8_8_9_6, \
call_8_8_9_7, \
call_8_8_9_8, \
call_8_8_9_9, \
call_8_8_9_10, \
call_8_8_9_11, \
call_8_8_9_12, \
call_8_8_9_13, \
call_8_8_9_14, \
call_8_8_9_15, \
call_8_8_10_0, \
call_8_8_10_1, \
call_8_8_10_2, \
call_8_8_10_3, \
call_8_8_10_4, \
call_8_8_10_5, \
call_8_8_10_6, \
call_8_8_10_7, \
call_8_8_10_8, \
call_8_8_10_9, \
call_8_8_10_10, \
call_8_8_10_11, \
call_8_8_10_12, \
call_8_8_10_13, \
call_8_8_10_14, \
call_8_8_10_15, \
call_8_8_11_0, \
call_8_8_11_1, \
call_8_8_11_2, \
call_8_8_11_3, \
call_8_8_11_4, \
call_8_8_11_5, \
call_8_8_11_6, \
call_8_8_11_7, \
call_8_8_11_8, \
call_8_8_11_9, \
call_8_8_11_10, \
call_8_8_11_11, \
call_8_8_11_12, \
call_8_8_11_13, \
call_8_8_11_14, \
call_8_8_11_15, \
call_8_8_12_0, \
call_8_8_12_1, \
call_8_8_12_2, \
call_8_8_12_3, \
call_8_8_12_4, \
call_8_8_12_5, \
call_8_8_12_6, \
call_8_8_12_7, \
call_8_8_12_8, \
call_8_8_12_9, \
call_8_8_12_10, \
call_8_8_12_11, \
call_8_8_12_12, \
call_8_8_12_13, \
call_8_8_12_14, \
call_8_8_12_15, \
call_8_8_13_0, \
call_8_8_13_1, \
call_8_8_13_2, \
call_8_8_13_3, \
call_8_8_13_4, \
call_8_8_13_5, \
call_8_8_13_6, \
call_8_8_13_7, \
call_8_8_13_8, \
call_8_8_13_9, \
call_8_8_13_10, \
call_8_8_13_11, \
call_8_8_13_12, \
call_8_8_13_13, \
call_8_8_13_14, \
call_8_8_13_15, \
call_8_8_14_0, \
call_8_8_14_1, \
call_8_8_14_2, \
call_8_8_14_3, \
call_8_8_14_4, \
call_8_8_14_5, \
call_8_8_14_6, \
call_8_8_14_7, \
call_8_8_14_8, \
call_8_8_14_9, \
call_8_8_14_10, \
call_8_8_14_11, \
call_8_8_14_12, \
call_8_8_14_13, \
call_8_8_14_14, \
call_8_8_14_15, \
call_8_8_15_0, \
call_8_8_15_1, \
call_8_8_15_2, \
call_8_8_15_3, \
call_8_8_15_4, \
call_8_8_15_5, \
call_8_8_15_6, \
call_8_8_15_7, \
call_8_8_15_8, \
call_8_8_15_9, \
call_8_8_15_10, \
call_8_8_15_11, \
call_8_8_15_12, \
call_8_8_15_13, \
call_8_8_15_14, \
call_8_8_15_15, \
call_8_9_0_0, \
call_8_9_0_1, \
call_8_9_0_2, \
call_8_9_0_3, \
call_8_9_0_4, \
call_8_9_0_5, \
call_8_9_0_6, \
call_8_9_0_7, \
call_8_9_0_8, \
call_8_9_0_9, \
call_8_9_0_10, \
call_8_9_0_11, \
call_8_9_0_12, \
call_8_9_0_13, \
call_8_9_0_14, \
call_8_9_0_15, \
call_8_9_1_0, \
call_8_9_1_1, \
call_8_9_1_2, \
call_8_9_1_3, \
call_8_9_1_4, \
call_8_9_1_5, \
call_8_9_1_6, \
call_8_9_1_7, \
call_8_9_1_8, \
call_8_9_1_9, \
call_8_9_1_10, \
call_8_9_1_11, \
call_8_9_1_12, \
call_8_9_1_13, \
call_8_9_1_14, \
call_8_9_1_15, \
call_8_9_2_0, \
call_8_9_2_1, \
call_8_9_2_2, \
call_8_9_2_3, \
call_8_9_2_4, \
call_8_9_2_5, \
call_8_9_2_6, \
call_8_9_2_7, \
call_8_9_2_8, \
call_8_9_2_9, \
call_8_9_2_10, \
call_8_9_2_11, \
call_8_9_2_12, \
call_8_9_2_13, \
call_8_9_2_14, \
call_8_9_2_15, \
call_8_9_3_0, \
call_8_9_3_1, \
call_8_9_3_2, \
call_8_9_3_3, \
call_8_9_3_4, \
call_8_9_3_5, \
call_8_9_3_6, \
call_8_9_3_7, \
call_8_9_3_8, \
call_8_9_3_9, \
call_8_9_3_10, \
call_8_9_3_11, \
call_8_9_3_12, \
call_8_9_3_13, \
call_8_9_3_14, \
call_8_9_3_15, \
call_8_9_4_0, \
call_8_9_4_1, \
call_8_9_4_2, \
call_8_9_4_3, \
call_8_9_4_4, \
call_8_9_4_5, \
call_8_9_4_6, \
call_8_9_4_7, \
call_8_9_4_8, \
call_8_9_4_9, \
call_8_9_4_10, \
call_8_9_4_11, \
call_8_9_4_12, \
call_8_9_4_13, \
call_8_9_4_14, \
call_8_9_4_15, \
call_8_9_5_0, \
call_8_9_5_1, \
call_8_9_5_2, \
call_8_9_5_3, \
call_8_9_5_4, \
call_8_9_5_5, \
call_8_9_5_6, \
call_8_9_5_7, \
call_8_9_5_8, \
call_8_9_5_9, \
call_8_9_5_10, \
call_8_9_5_11, \
call_8_9_5_12, \
call_8_9_5_13, \
call_8_9_5_14, \
call_8_9_5_15, \
call_8_9_6_0, \
call_8_9_6_1, \
call_8_9_6_2, \
call_8_9_6_3, \
call_8_9_6_4, \
call_8_9_6_5, \
call_8_9_6_6, \
call_8_9_6_7, \
call_8_9_6_8, \
call_8_9_6_9, \
call_8_9_6_10, \
call_8_9_6_11, \
call_8_9_6_12, \
call_8_9_6_13, \
call_8_9_6_14, \
call_8_9_6_15, \
call_8_9_7_0, \
call_8_9_7_1, \
call_8_9_7_2, \
call_8_9_7_3, \
call_8_9_7_4, \
call_8_9_7_5, \
call_8_9_7_6, \
call_8_9_7_7, \
call_8_9_7_8, \
call_8_9_7_9, \
call_8_9_7_10, \
call_8_9_7_11, \
call_8_9_7_12, \
call_8_9_7_13, \
call_8_9_7_14, \
call_8_9_7_15, \
call_8_9_8_0, \
call_8_9_8_1, \
call_8_9_8_2, \
call_8_9_8_3, \
call_8_9_8_4, \
call_8_9_8_5, \
call_8_9_8_6, \
call_8_9_8_7, \
call_8_9_8_8, \
call_8_9_8_9, \
call_8_9_8_10, \
call_8_9_8_11, \
call_8_9_8_12, \
call_8_9_8_13, \
call_8_9_8_14, \
call_8_9_8_15, \
call_8_9_9_0, \
call_8_9_9_1, \
call_8_9_9_2, \
call_8_9_9_3, \
call_8_9_9_4, \
call_8_9_9_5, \
call_8_9_9_6, \
call_8_9_9_7, \
call_8_9_9_8, \
call_8_9_9_9, \
call_8_9_9_10, \
call_8_9_9_11, \
call_8_9_9_12, \
call_8_9_9_13, \
call_8_9_9_14, \
call_8_9_9_15, \
call_8_9_10_0, \
call_8_9_10_1, \
call_8_9_10_2, \
call_8_9_10_3, \
call_8_9_10_4, \
call_8_9_10_5, \
call_8_9_10_6, \
call_8_9_10_7, \
call_8_9_10_8, \
call_8_9_10_9, \
call_8_9_10_10, \
call_8_9_10_11, \
call_8_9_10_12, \
call_8_9_10_13, \
call_8_9_10_14, \
call_8_9_10_15, \
call_8_9_11_0, \
call_8_9_11_1, \
call_8_9_11_2, \
call_8_9_11_3, \
call_8_9_11_4, \
call_8_9_11_5, \
call_8_9_11_6, \
call_8_9_11_7, \
call_8_9_11_8, \
call_8_9_11_9, \
call_8_9_11_10, \
call_8_9_11_11, \
call_8_9_11_12, \
call_8_9_11_13, \
call_8_9_11_14, \
call_8_9_11_15, \
call_8_9_12_0, \
call_8_9_12_1, \
call_8_9_12_2, \
call_8_9_12_3, \
call_8_9_12_4, \
call_8_9_12_5, \
call_8_9_12_6, \
call_8_9_12_7, \
call_8_9_12_8, \
call_8_9_12_9, \
call_8_9_12_10, \
call_8_9_12_11, \
call_8_9_12_12, \
call_8_9_12_13, \
call_8_9_12_14, \
call_8_9_12_15, \
call_8_9_13_0, \
call_8_9_13_1, \
call_8_9_13_2, \
call_8_9_13_3, \
call_8_9_13_4, \
call_8_9_13_5, \
call_8_9_13_6, \
call_8_9_13_7, \
call_8_9_13_8, \
call_8_9_13_9, \
call_8_9_13_10, \
call_8_9_13_11, \
call_8_9_13_12, \
call_8_9_13_13, \
call_8_9_13_14, \
call_8_9_13_15, \
call_8_9_14_0, \
call_8_9_14_1, \
call_8_9_14_2, \
call_8_9_14_3, \
call_8_9_14_4, \
call_8_9_14_5, \
call_8_9_14_6, \
call_8_9_14_7, \
call_8_9_14_8, \
call_8_9_14_9, \
call_8_9_14_10, \
call_8_9_14_11, \
call_8_9_14_12, \
call_8_9_14_13, \
call_8_9_14_14, \
call_8_9_14_15, \
call_8_9_15_0, \
call_8_9_15_1, \
call_8_9_15_2, \
call_8_9_15_3, \
call_8_9_15_4, \
call_8_9_15_5, \
call_8_9_15_6, \
call_8_9_15_7, \
call_8_9_15_8, \
call_8_9_15_9, \
call_8_9_15_10, \
call_8_9_15_11, \
call_8_9_15_12, \
call_8_9_15_13, \
call_8_9_15_14, \
call_8_9_15_15, \
call_8_10_0_0, \
call_8_10_0_1, \
call_8_10_0_2, \
call_8_10_0_3, \
call_8_10_0_4, \
call_8_10_0_5, \
call_8_10_0_6, \
call_8_10_0_7, \
call_8_10_0_8, \
call_8_10_0_9, \
call_8_10_0_10, \
call_8_10_0_11, \
call_8_10_0_12, \
call_8_10_0_13, \
call_8_10_0_14, \
call_8_10_0_15, \
call_8_10_1_0, \
call_8_10_1_1, \
call_8_10_1_2, \
call_8_10_1_3, \
call_8_10_1_4, \
call_8_10_1_5, \
call_8_10_1_6, \
call_8_10_1_7, \
call_8_10_1_8, \
call_8_10_1_9, \
call_8_10_1_10, \
call_8_10_1_11, \
call_8_10_1_12, \
call_8_10_1_13, \
call_8_10_1_14, \
call_8_10_1_15, \
call_8_10_2_0, \
call_8_10_2_1, \
call_8_10_2_2, \
call_8_10_2_3, \
call_8_10_2_4, \
call_8_10_2_5, \
call_8_10_2_6, \
call_8_10_2_7, \
call_8_10_2_8, \
call_8_10_2_9, \
call_8_10_2_10, \
call_8_10_2_11, \
call_8_10_2_12, \
call_8_10_2_13, \
call_8_10_2_14, \
call_8_10_2_15, \
call_8_10_3_0, \
call_8_10_3_1, \
call_8_10_3_2, \
call_8_10_3_3, \
call_8_10_3_4, \
call_8_10_3_5, \
call_8_10_3_6, \
call_8_10_3_7, \
call_8_10_3_8, \
call_8_10_3_9, \
call_8_10_3_10, \
call_8_10_3_11, \
call_8_10_3_12, \
call_8_10_3_13, \
call_8_10_3_14, \
call_8_10_3_15, \
call_8_10_4_0, \
call_8_10_4_1, \
call_8_10_4_2, \
call_8_10_4_3, \
call_8_10_4_4, \
call_8_10_4_5, \
call_8_10_4_6, \
call_8_10_4_7, \
call_8_10_4_8, \
call_8_10_4_9, \
call_8_10_4_10, \
call_8_10_4_11, \
call_8_10_4_12, \
call_8_10_4_13, \
call_8_10_4_14, \
call_8_10_4_15, \
call_8_10_5_0, \
call_8_10_5_1, \
call_8_10_5_2, \
call_8_10_5_3, \
call_8_10_5_4, \
call_8_10_5_5, \
call_8_10_5_6, \
call_8_10_5_7, \
call_8_10_5_8, \
call_8_10_5_9, \
call_8_10_5_10, \
call_8_10_5_11, \
call_8_10_5_12, \
call_8_10_5_13, \
call_8_10_5_14, \
call_8_10_5_15, \
call_8_10_6_0, \
call_8_10_6_1, \
call_8_10_6_2, \
call_8_10_6_3, \
call_8_10_6_4, \
call_8_10_6_5, \
call_8_10_6_6, \
call_8_10_6_7, \
call_8_10_6_8, \
call_8_10_6_9, \
call_8_10_6_10, \
call_8_10_6_11, \
call_8_10_6_12, \
call_8_10_6_13, \
call_8_10_6_14, \
call_8_10_6_15, \
call_8_10_7_0, \
call_8_10_7_1, \
call_8_10_7_2, \
call_8_10_7_3, \
call_8_10_7_4, \
call_8_10_7_5, \
call_8_10_7_6, \
call_8_10_7_7, \
call_8_10_7_8, \
call_8_10_7_9, \
call_8_10_7_10, \
call_8_10_7_11, \
call_8_10_7_12, \
call_8_10_7_13, \
call_8_10_7_14, \
call_8_10_7_15, \
call_8_10_8_0, \
call_8_10_8_1, \
call_8_10_8_2, \
call_8_10_8_3, \
call_8_10_8_4, \
call_8_10_8_5, \
call_8_10_8_6, \
call_8_10_8_7, \
call_8_10_8_8, \
call_8_10_8_9, \
call_8_10_8_10, \
call_8_10_8_11, \
call_8_10_8_12, \
call_8_10_8_13, \
call_8_10_8_14, \
call_8_10_8_15, \
call_8_10_9_0, \
call_8_10_9_1, \
call_8_10_9_2, \
call_8_10_9_3, \
call_8_10_9_4, \
call_8_10_9_5, \
call_8_10_9_6, \
call_8_10_9_7, \
call_8_10_9_8, \
call_8_10_9_9, \
call_8_10_9_10, \
call_8_10_9_11, \
call_8_10_9_12, \
call_8_10_9_13, \
call_8_10_9_14, \
call_8_10_9_15, \
call_8_10_10_0, \
call_8_10_10_1, \
call_8_10_10_2, \
call_8_10_10_3, \
call_8_10_10_4, \
call_8_10_10_5, \
call_8_10_10_6, \
call_8_10_10_7, \
call_8_10_10_8, \
call_8_10_10_9, \
call_8_10_10_10, \
call_8_10_10_11, \
call_8_10_10_12, \
call_8_10_10_13, \
call_8_10_10_14, \
call_8_10_10_15, \
call_8_10_11_0, \
call_8_10_11_1, \
call_8_10_11_2, \
call_8_10_11_3, \
call_8_10_11_4, \
call_8_10_11_5, \
call_8_10_11_6, \
call_8_10_11_7, \
call_8_10_11_8, \
call_8_10_11_9, \
call_8_10_11_10, \
call_8_10_11_11, \
call_8_10_11_12, \
call_8_10_11_13, \
call_8_10_11_14, \
call_8_10_11_15, \
call_8_10_12_0, \
call_8_10_12_1, \
call_8_10_12_2, \
call_8_10_12_3, \
call_8_10_12_4, \
call_8_10_12_5, \
call_8_10_12_6, \
call_8_10_12_7, \
call_8_10_12_8, \
call_8_10_12_9, \
call_8_10_12_10, \
call_8_10_12_11, \
call_8_10_12_12, \
call_8_10_12_13, \
call_8_10_12_14, \
call_8_10_12_15, \
call_8_10_13_0, \
call_8_10_13_1, \
call_8_10_13_2, \
call_8_10_13_3, \
call_8_10_13_4, \
call_8_10_13_5, \
call_8_10_13_6, \
call_8_10_13_7, \
call_8_10_13_8, \
call_8_10_13_9, \
call_8_10_13_10, \
call_8_10_13_11, \
call_8_10_13_12, \
call_8_10_13_13, \
call_8_10_13_14, \
call_8_10_13_15, \
call_8_10_14_0, \
call_8_10_14_1, \
call_8_10_14_2, \
call_8_10_14_3, \
call_8_10_14_4, \
call_8_10_14_5, \
call_8_10_14_6, \
call_8_10_14_7, \
call_8_10_14_8, \
call_8_10_14_9, \
call_8_10_14_10, \
call_8_10_14_11, \
call_8_10_14_12, \
call_8_10_14_13, \
call_8_10_14_14, \
call_8_10_14_15, \
call_8_10_15_0, \
call_8_10_15_1, \
call_8_10_15_2, \
call_8_10_15_3, \
call_8_10_15_4, \
call_8_10_15_5, \
call_8_10_15_6, \
call_8_10_15_7, \
call_8_10_15_8, \
call_8_10_15_9, \
call_8_10_15_10, \
call_8_10_15_11, \
call_8_10_15_12, \
call_8_10_15_13, \
call_8_10_15_14, \
call_8_10_15_15, \
call_8_11_0_0, \
call_8_11_0_1, \
call_8_11_0_2, \
call_8_11_0_3, \
call_8_11_0_4, \
call_8_11_0_5, \
call_8_11_0_6, \
call_8_11_0_7, \
call_8_11_0_8, \
call_8_11_0_9, \
call_8_11_0_10, \
call_8_11_0_11, \
call_8_11_0_12, \
call_8_11_0_13, \
call_8_11_0_14, \
call_8_11_0_15, \
call_8_11_1_0, \
call_8_11_1_1, \
call_8_11_1_2, \
call_8_11_1_3, \
call_8_11_1_4, \
call_8_11_1_5, \
call_8_11_1_6, \
call_8_11_1_7, \
call_8_11_1_8, \
call_8_11_1_9, \
call_8_11_1_10, \
call_8_11_1_11, \
call_8_11_1_12, \
call_8_11_1_13, \
call_8_11_1_14, \
call_8_11_1_15, \
call_8_11_2_0, \
call_8_11_2_1, \
call_8_11_2_2, \
call_8_11_2_3, \
call_8_11_2_4, \
call_8_11_2_5, \
call_8_11_2_6, \
call_8_11_2_7, \
call_8_11_2_8, \
call_8_11_2_9, \
call_8_11_2_10, \
call_8_11_2_11, \
call_8_11_2_12, \
call_8_11_2_13, \
call_8_11_2_14, \
call_8_11_2_15, \
call_8_11_3_0, \
call_8_11_3_1, \
call_8_11_3_2, \
call_8_11_3_3, \
call_8_11_3_4, \
call_8_11_3_5, \
call_8_11_3_6, \
call_8_11_3_7, \
call_8_11_3_8, \
call_8_11_3_9, \
call_8_11_3_10, \
call_8_11_3_11, \
call_8_11_3_12, \
call_8_11_3_13, \
call_8_11_3_14, \
call_8_11_3_15, \
call_8_11_4_0, \
call_8_11_4_1, \
call_8_11_4_2, \
call_8_11_4_3, \
call_8_11_4_4, \
call_8_11_4_5, \
call_8_11_4_6, \
call_8_11_4_7, \
call_8_11_4_8, \
call_8_11_4_9, \
call_8_11_4_10, \
call_8_11_4_11, \
call_8_11_4_12, \
call_8_11_4_13, \
call_8_11_4_14, \
call_8_11_4_15, \
call_8_11_5_0, \
call_8_11_5_1, \
call_8_11_5_2, \
call_8_11_5_3, \
call_8_11_5_4, \
call_8_11_5_5, \
call_8_11_5_6, \
call_8_11_5_7, \
call_8_11_5_8, \
call_8_11_5_9, \
call_8_11_5_10, \
call_8_11_5_11, \
call_8_11_5_12, \
call_8_11_5_13, \
call_8_11_5_14, \
call_8_11_5_15, \
call_8_11_6_0, \
call_8_11_6_1, \
call_8_11_6_2, \
call_8_11_6_3, \
call_8_11_6_4, \
call_8_11_6_5, \
call_8_11_6_6, \
call_8_11_6_7, \
call_8_11_6_8, \
call_8_11_6_9, \
call_8_11_6_10, \
call_8_11_6_11, \
call_8_11_6_12, \
call_8_11_6_13, \
call_8_11_6_14, \
call_8_11_6_15, \
call_8_11_7_0, \
call_8_11_7_1, \
call_8_11_7_2, \
call_8_11_7_3, \
call_8_11_7_4, \
call_8_11_7_5, \
call_8_11_7_6, \
call_8_11_7_7, \
call_8_11_7_8, \
call_8_11_7_9, \
call_8_11_7_10, \
call_8_11_7_11, \
call_8_11_7_12, \
call_8_11_7_13, \
call_8_11_7_14, \
call_8_11_7_15, \
call_8_11_8_0, \
call_8_11_8_1, \
call_8_11_8_2, \
call_8_11_8_3, \
call_8_11_8_4, \
call_8_11_8_5, \
call_8_11_8_6, \
call_8_11_8_7, \
call_8_11_8_8, \
call_8_11_8_9, \
call_8_11_8_10, \
call_8_11_8_11, \
call_8_11_8_12, \
call_8_11_8_13, \
call_8_11_8_14, \
call_8_11_8_15, \
call_8_11_9_0, \
call_8_11_9_1, \
call_8_11_9_2, \
call_8_11_9_3, \
call_8_11_9_4, \
call_8_11_9_5, \
call_8_11_9_6, \
call_8_11_9_7, \
call_8_11_9_8, \
call_8_11_9_9, \
call_8_11_9_10, \
call_8_11_9_11, \
call_8_11_9_12, \
call_8_11_9_13, \
call_8_11_9_14, \
call_8_11_9_15, \
call_8_11_10_0, \
call_8_11_10_1, \
call_8_11_10_2, \
call_8_11_10_3, \
call_8_11_10_4, \
call_8_11_10_5, \
call_8_11_10_6, \
call_8_11_10_7, \
call_8_11_10_8, \
call_8_11_10_9, \
call_8_11_10_10, \
call_8_11_10_11, \
call_8_11_10_12, \
call_8_11_10_13, \
call_8_11_10_14, \
call_8_11_10_15, \
call_8_11_11_0, \
call_8_11_11_1, \
call_8_11_11_2, \
call_8_11_11_3, \
call_8_11_11_4, \
call_8_11_11_5, \
call_8_11_11_6, \
call_8_11_11_7, \
call_8_11_11_8, \
call_8_11_11_9, \
call_8_11_11_10, \
call_8_11_11_11, \
call_8_11_11_12, \
call_8_11_11_13, \
call_8_11_11_14, \
call_8_11_11_15, \
call_8_11_12_0, \
call_8_11_12_1, \
call_8_11_12_2, \
call_8_11_12_3, \
call_8_11_12_4, \
call_8_11_12_5, \
call_8_11_12_6, \
call_8_11_12_7, \
call_8_11_12_8, \
call_8_11_12_9, \
call_8_11_12_10, \
call_8_11_12_11, \
call_8_11_12_12, \
call_8_11_12_13, \
call_8_11_12_14, \
call_8_11_12_15, \
call_8_11_13_0, \
call_8_11_13_1, \
call_8_11_13_2, \
call_8_11_13_3, \
call_8_11_13_4, \
call_8_11_13_5, \
call_8_11_13_6, \
call_8_11_13_7, \
call_8_11_13_8, \
call_8_11_13_9, \
call_8_11_13_10, \
call_8_11_13_11, \
call_8_11_13_12, \
call_8_11_13_13, \
call_8_11_13_14, \
call_8_11_13_15, \
call_8_11_14_0, \
call_8_11_14_1, \
call_8_11_14_2, \
call_8_11_14_3, \
call_8_11_14_4, \
call_8_11_14_5, \
call_8_11_14_6, \
call_8_11_14_7, \
call_8_11_14_8, \
call_8_11_14_9, \
call_8_11_14_10, \
call_8_11_14_11, \
call_8_11_14_12, \
call_8_11_14_13, \
call_8_11_14_14, \
call_8_11_14_15, \
call_8_11_15_0, \
call_8_11_15_1, \
call_8_11_15_2, \
call_8_11_15_3, \
call_8_11_15_4, \
call_8_11_15_5, \
call_8_11_15_6, \
call_8_11_15_7, \
call_8_11_15_8, \
call_8_11_15_9, \
call_8_11_15_10, \
call_8_11_15_11, \
call_8_11_15_12, \
call_8_11_15_13, \
call_8_11_15_14, \
call_8_11_15_15, \
call_8_12_0_0, \
call_8_12_0_1, \
call_8_12_0_2, \
call_8_12_0_3, \
call_8_12_0_4, \
call_8_12_0_5, \
call_8_12_0_6, \
call_8_12_0_7, \
call_8_12_0_8, \
call_8_12_0_9, \
call_8_12_0_10, \
call_8_12_0_11, \
call_8_12_0_12, \
call_8_12_0_13, \
call_8_12_0_14, \
call_8_12_0_15, \
call_8_12_1_0, \
call_8_12_1_1, \
call_8_12_1_2, \
call_8_12_1_3, \
call_8_12_1_4, \
call_8_12_1_5, \
call_8_12_1_6, \
call_8_12_1_7, \
call_8_12_1_8, \
call_8_12_1_9, \
call_8_12_1_10, \
call_8_12_1_11, \
call_8_12_1_12, \
call_8_12_1_13, \
call_8_12_1_14, \
call_8_12_1_15, \
call_8_12_2_0, \
call_8_12_2_1, \
call_8_12_2_2, \
call_8_12_2_3, \
call_8_12_2_4, \
call_8_12_2_5, \
call_8_12_2_6, \
call_8_12_2_7, \
call_8_12_2_8, \
call_8_12_2_9, \
call_8_12_2_10, \
call_8_12_2_11, \
call_8_12_2_12, \
call_8_12_2_13, \
call_8_12_2_14, \
call_8_12_2_15, \
call_8_12_3_0, \
call_8_12_3_1, \
call_8_12_3_2, \
call_8_12_3_3, \
call_8_12_3_4, \
call_8_12_3_5, \
call_8_12_3_6, \
call_8_12_3_7, \
call_8_12_3_8, \
call_8_12_3_9, \
call_8_12_3_10, \
call_8_12_3_11, \
call_8_12_3_12, \
call_8_12_3_13, \
call_8_12_3_14, \
call_8_12_3_15, \
call_8_12_4_0, \
call_8_12_4_1, \
call_8_12_4_2, \
call_8_12_4_3, \
call_8_12_4_4, \
call_8_12_4_5, \
call_8_12_4_6, \
call_8_12_4_7, \
call_8_12_4_8, \
call_8_12_4_9, \
call_8_12_4_10, \
call_8_12_4_11, \
call_8_12_4_12, \
call_8_12_4_13, \
call_8_12_4_14, \
call_8_12_4_15, \
call_8_12_5_0, \
call_8_12_5_1, \
call_8_12_5_2, \
call_8_12_5_3, \
call_8_12_5_4, \
call_8_12_5_5, \
call_8_12_5_6, \
call_8_12_5_7, \
call_8_12_5_8, \
call_8_12_5_9, \
call_8_12_5_10, \
call_8_12_5_11, \
call_8_12_5_12, \
call_8_12_5_13, \
call_8_12_5_14, \
call_8_12_5_15, \
call_8_12_6_0, \
call_8_12_6_1, \
call_8_12_6_2, \
call_8_12_6_3, \
call_8_12_6_4, \
call_8_12_6_5, \
call_8_12_6_6, \
call_8_12_6_7, \
call_8_12_6_8, \
call_8_12_6_9, \
call_8_12_6_10, \
call_8_12_6_11, \
call_8_12_6_12, \
call_8_12_6_13, \
call_8_12_6_14, \
call_8_12_6_15, \
call_8_12_7_0, \
call_8_12_7_1, \
call_8_12_7_2, \
call_8_12_7_3, \
call_8_12_7_4, \
call_8_12_7_5, \
call_8_12_7_6, \
call_8_12_7_7, \
call_8_12_7_8, \
call_8_12_7_9, \
call_8_12_7_10, \
call_8_12_7_11, \
call_8_12_7_12, \
call_8_12_7_13, \
call_8_12_7_14, \
call_8_12_7_15, \
call_8_12_8_0, \
call_8_12_8_1, \
call_8_12_8_2, \
call_8_12_8_3, \
call_8_12_8_4, \
call_8_12_8_5, \
call_8_12_8_6, \
call_8_12_8_7, \
call_8_12_8_8, \
call_8_12_8_9, \
call_8_12_8_10, \
call_8_12_8_11, \
call_8_12_8_12, \
call_8_12_8_13, \
call_8_12_8_14, \
call_8_12_8_15, \
call_8_12_9_0, \
call_8_12_9_1, \
call_8_12_9_2, \
call_8_12_9_3, \
call_8_12_9_4, \
call_8_12_9_5, \
call_8_12_9_6, \
call_8_12_9_7, \
call_8_12_9_8, \
call_8_12_9_9, \
call_8_12_9_10, \
call_8_12_9_11, \
call_8_12_9_12, \
call_8_12_9_13, \
call_8_12_9_14, \
call_8_12_9_15, \
call_8_12_10_0, \
call_8_12_10_1, \
call_8_12_10_2, \
call_8_12_10_3, \
call_8_12_10_4, \
call_8_12_10_5, \
call_8_12_10_6, \
call_8_12_10_7, \
call_8_12_10_8, \
call_8_12_10_9, \
call_8_12_10_10, \
call_8_12_10_11, \
call_8_12_10_12, \
call_8_12_10_13, \
call_8_12_10_14, \
call_8_12_10_15, \
call_8_12_11_0, \
call_8_12_11_1, \
call_8_12_11_2, \
call_8_12_11_3, \
call_8_12_11_4, \
call_8_12_11_5, \
call_8_12_11_6, \
call_8_12_11_7, \
call_8_12_11_8, \
call_8_12_11_9, \
call_8_12_11_10, \
call_8_12_11_11, \
call_8_12_11_12, \
call_8_12_11_13, \
call_8_12_11_14, \
call_8_12_11_15, \
call_8_12_12_0, \
call_8_12_12_1, \
call_8_12_12_2, \
call_8_12_12_3, \
call_8_12_12_4, \
call_8_12_12_5, \
call_8_12_12_6, \
call_8_12_12_7, \
call_8_12_12_8, \
call_8_12_12_9, \
call_8_12_12_10, \
call_8_12_12_11, \
call_8_12_12_12, \
call_8_12_12_13, \
call_8_12_12_14, \
call_8_12_12_15, \
call_8_12_13_0, \
call_8_12_13_1, \
call_8_12_13_2, \
call_8_12_13_3, \
call_8_12_13_4, \
call_8_12_13_5, \
call_8_12_13_6, \
call_8_12_13_7, \
call_8_12_13_8, \
call_8_12_13_9, \
call_8_12_13_10, \
call_8_12_13_11, \
call_8_12_13_12, \
call_8_12_13_13, \
call_8_12_13_14, \
call_8_12_13_15, \
call_8_12_14_0, \
call_8_12_14_1, \
call_8_12_14_2, \
call_8_12_14_3, \
call_8_12_14_4, \
call_8_12_14_5, \
call_8_12_14_6, \
call_8_12_14_7, \
call_8_12_14_8, \
call_8_12_14_9, \
call_8_12_14_10, \
call_8_12_14_11, \
call_8_12_14_12, \
call_8_12_14_13, \
call_8_12_14_14, \
call_8_12_14_15, \
call_8_12_15_0, \
call_8_12_15_1, \
call_8_12_15_2, \
call_8_12_15_3, \
call_8_12_15_4, \
call_8_12_15_5, \
call_8_12_15_6, \
call_8_12_15_7, \
call_8_12_15_8, \
call_8_12_15_9, \
call_8_12_15_10, \
call_8_12_15_11, \
call_8_12_15_12, \
call_8_12_15_13, \
call_8_12_15_14, \
call_8_12_15_15, \
call_8_13_0_0, \
call_8_13_0_1, \
call_8_13_0_2, \
call_8_13_0_3, \
call_8_13_0_4, \
call_8_13_0_5, \
call_8_13_0_6, \
call_8_13_0_7, \
call_8_13_0_8, \
call_8_13_0_9, \
call_8_13_0_10, \
call_8_13_0_11, \
call_8_13_0_12, \
call_8_13_0_13, \
call_8_13_0_14, \
call_8_13_0_15, \
call_8_13_1_0, \
call_8_13_1_1, \
call_8_13_1_2, \
call_8_13_1_3, \
call_8_13_1_4, \
call_8_13_1_5, \
call_8_13_1_6, \
call_8_13_1_7, \
call_8_13_1_8, \
call_8_13_1_9, \
call_8_13_1_10, \
call_8_13_1_11, \
call_8_13_1_12, \
call_8_13_1_13, \
call_8_13_1_14, \
call_8_13_1_15, \
call_8_13_2_0, \
call_8_13_2_1, \
call_8_13_2_2, \
call_8_13_2_3, \
call_8_13_2_4, \
call_8_13_2_5, \
call_8_13_2_6, \
call_8_13_2_7, \
call_8_13_2_8, \
call_8_13_2_9, \
call_8_13_2_10, \
call_8_13_2_11, \
call_8_13_2_12, \
call_8_13_2_13, \
call_8_13_2_14, \
call_8_13_2_15, \
call_8_13_3_0, \
call_8_13_3_1, \
call_8_13_3_2, \
call_8_13_3_3, \
call_8_13_3_4, \
call_8_13_3_5, \
call_8_13_3_6, \
call_8_13_3_7, \
call_8_13_3_8, \
call_8_13_3_9, \
call_8_13_3_10, \
call_8_13_3_11, \
call_8_13_3_12, \
call_8_13_3_13, \
call_8_13_3_14, \
call_8_13_3_15, \
call_8_13_4_0, \
call_8_13_4_1, \
call_8_13_4_2, \
call_8_13_4_3, \
call_8_13_4_4, \
call_8_13_4_5, \
call_8_13_4_6, \
call_8_13_4_7, \
call_8_13_4_8, \
call_8_13_4_9, \
call_8_13_4_10, \
call_8_13_4_11, \
call_8_13_4_12, \
call_8_13_4_13, \
call_8_13_4_14, \
call_8_13_4_15, \
call_8_13_5_0, \
call_8_13_5_1, \
call_8_13_5_2, \
call_8_13_5_3, \
call_8_13_5_4, \
call_8_13_5_5, \
call_8_13_5_6, \
call_8_13_5_7, \
call_8_13_5_8, \
call_8_13_5_9, \
call_8_13_5_10, \
call_8_13_5_11, \
call_8_13_5_12, \
call_8_13_5_13, \
call_8_13_5_14, \
call_8_13_5_15, \
call_8_13_6_0, \
call_8_13_6_1, \
call_8_13_6_2, \
call_8_13_6_3, \
call_8_13_6_4, \
call_8_13_6_5, \
call_8_13_6_6, \
call_8_13_6_7, \
call_8_13_6_8, \
call_8_13_6_9, \
call_8_13_6_10, \
call_8_13_6_11, \
call_8_13_6_12, \
call_8_13_6_13, \
call_8_13_6_14, \
call_8_13_6_15, \
call_8_13_7_0, \
call_8_13_7_1, \
call_8_13_7_2, \
call_8_13_7_3, \
call_8_13_7_4, \
call_8_13_7_5, \
call_8_13_7_6, \
call_8_13_7_7, \
call_8_13_7_8, \
call_8_13_7_9, \
call_8_13_7_10, \
call_8_13_7_11, \
call_8_13_7_12, \
call_8_13_7_13, \
call_8_13_7_14, \
call_8_13_7_15, \
call_8_13_8_0, \
call_8_13_8_1, \
call_8_13_8_2, \
call_8_13_8_3, \
call_8_13_8_4, \
call_8_13_8_5, \
call_8_13_8_6, \
call_8_13_8_7, \
call_8_13_8_8, \
call_8_13_8_9, \
call_8_13_8_10, \
call_8_13_8_11, \
call_8_13_8_12, \
call_8_13_8_13, \
call_8_13_8_14, \
call_8_13_8_15, \
call_8_13_9_0, \
call_8_13_9_1, \
call_8_13_9_2, \
call_8_13_9_3, \
call_8_13_9_4, \
call_8_13_9_5, \
call_8_13_9_6, \
call_8_13_9_7, \
call_8_13_9_8, \
call_8_13_9_9, \
call_8_13_9_10, \
call_8_13_9_11, \
call_8_13_9_12, \
call_8_13_9_13, \
call_8_13_9_14, \
call_8_13_9_15, \
call_8_13_10_0, \
call_8_13_10_1, \
call_8_13_10_2, \
call_8_13_10_3, \
call_8_13_10_4, \
call_8_13_10_5, \
call_8_13_10_6, \
call_8_13_10_7, \
call_8_13_10_8, \
call_8_13_10_9, \
call_8_13_10_10, \
call_8_13_10_11, \
call_8_13_10_12, \
call_8_13_10_13, \
call_8_13_10_14, \
call_8_13_10_15, \
call_8_13_11_0, \
call_8_13_11_1, \
call_8_13_11_2, \
call_8_13_11_3, \
call_8_13_11_4, \
call_8_13_11_5, \
call_8_13_11_6, \
call_8_13_11_7, \
call_8_13_11_8, \
call_8_13_11_9, \
call_8_13_11_10, \
call_8_13_11_11, \
call_8_13_11_12, \
call_8_13_11_13, \
call_8_13_11_14, \
call_8_13_11_15, \
call_8_13_12_0, \
call_8_13_12_1, \
call_8_13_12_2, \
call_8_13_12_3, \
call_8_13_12_4, \
call_8_13_12_5, \
call_8_13_12_6, \
call_8_13_12_7, \
call_8_13_12_8, \
call_8_13_12_9, \
call_8_13_12_10, \
call_8_13_12_11, \
call_8_13_12_12, \
call_8_13_12_13, \
call_8_13_12_14, \
call_8_13_12_15, \
call_8_13_13_0, \
call_8_13_13_1, \
call_8_13_13_2, \
call_8_13_13_3, \
call_8_13_13_4, \
call_8_13_13_5, \
call_8_13_13_6, \
call_8_13_13_7, \
call_8_13_13_8, \
call_8_13_13_9, \
call_8_13_13_10, \
call_8_13_13_11, \
call_8_13_13_12, \
call_8_13_13_13, \
call_8_13_13_14, \
call_8_13_13_15, \
call_8_13_14_0, \
call_8_13_14_1, \
call_8_13_14_2, \
call_8_13_14_3, \
call_8_13_14_4, \
call_8_13_14_5, \
call_8_13_14_6, \
call_8_13_14_7, \
call_8_13_14_8, \
call_8_13_14_9, \
call_8_13_14_10, \
call_8_13_14_11, \
call_8_13_14_12, \
call_8_13_14_13, \
call_8_13_14_14, \
call_8_13_14_15, \
call_8_13_15_0, \
call_8_13_15_1, \
call_8_13_15_2, \
call_8_13_15_3, \
call_8_13_15_4, \
call_8_13_15_5, \
call_8_13_15_6, \
call_8_13_15_7, \
call_8_13_15_8, \
call_8_13_15_9, \
call_8_13_15_10, \
call_8_13_15_11, \
call_8_13_15_12, \
call_8_13_15_13, \
call_8_13_15_14, \
call_8_13_15_15, \
call_8_14_0_0, \
call_8_14_0_1, \
call_8_14_0_2, \
call_8_14_0_3, \
call_8_14_0_4, \
call_8_14_0_5, \
call_8_14_0_6, \
call_8_14_0_7, \
call_8_14_0_8, \
call_8_14_0_9, \
call_8_14_0_10, \
call_8_14_0_11, \
call_8_14_0_12, \
call_8_14_0_13, \
call_8_14_0_14, \
call_8_14_0_15, \
call_8_14_1_0, \
call_8_14_1_1, \
call_8_14_1_2, \
call_8_14_1_3, \
call_8_14_1_4, \
call_8_14_1_5, \
call_8_14_1_6, \
call_8_14_1_7, \
call_8_14_1_8, \
call_8_14_1_9, \
call_8_14_1_10, \
call_8_14_1_11, \
call_8_14_1_12, \
call_8_14_1_13, \
call_8_14_1_14, \
call_8_14_1_15, \
call_8_14_2_0, \
call_8_14_2_1, \
call_8_14_2_2, \
call_8_14_2_3, \
call_8_14_2_4, \
call_8_14_2_5, \
call_8_14_2_6, \
call_8_14_2_7, \
call_8_14_2_8, \
call_8_14_2_9, \
call_8_14_2_10, \
call_8_14_2_11, \
call_8_14_2_12, \
call_8_14_2_13, \
call_8_14_2_14, \
call_8_14_2_15, \
call_8_14_3_0, \
call_8_14_3_1, \
call_8_14_3_2, \
call_8_14_3_3, \
call_8_14_3_4, \
call_8_14_3_5, \
call_8_14_3_6, \
call_8_14_3_7, \
call_8_14_3_8, \
call_8_14_3_9, \
call_8_14_3_10, \
call_8_14_3_11, \
call_8_14_3_12, \
call_8_14_3_13, \
call_8_14_3_14, \
call_8_14_3_15, \
call_8_14_4_0, \
call_8_14_4_1, \
call_8_14_4_2, \
call_8_14_4_3, \
call_8_14_4_4, \
call_8_14_4_5, \
call_8_14_4_6, \
call_8_14_4_7, \
call_8_14_4_8, \
call_8_14_4_9, \
call_8_14_4_10, \
call_8_14_4_11, \
call_8_14_4_12, \
call_8_14_4_13, \
call_8_14_4_14, \
call_8_14_4_15, \
call_8_14_5_0, \
call_8_14_5_1, \
call_8_14_5_2, \
call_8_14_5_3, \
call_8_14_5_4, \
call_8_14_5_5, \
call_8_14_5_6, \
call_8_14_5_7, \
call_8_14_5_8, \
call_8_14_5_9, \
call_8_14_5_10, \
call_8_14_5_11, \
call_8_14_5_12, \
call_8_14_5_13, \
call_8_14_5_14, \
call_8_14_5_15, \
call_8_14_6_0, \
call_8_14_6_1, \
call_8_14_6_2, \
call_8_14_6_3, \
call_8_14_6_4, \
call_8_14_6_5, \
call_8_14_6_6, \
call_8_14_6_7, \
call_8_14_6_8, \
call_8_14_6_9, \
call_8_14_6_10, \
call_8_14_6_11, \
call_8_14_6_12, \
call_8_14_6_13, \
call_8_14_6_14, \
call_8_14_6_15, \
call_8_14_7_0, \
call_8_14_7_1, \
call_8_14_7_2, \
call_8_14_7_3, \
call_8_14_7_4, \
call_8_14_7_5, \
call_8_14_7_6, \
call_8_14_7_7, \
call_8_14_7_8, \
call_8_14_7_9, \
call_8_14_7_10, \
call_8_14_7_11, \
call_8_14_7_12, \
call_8_14_7_13, \
call_8_14_7_14, \
call_8_14_7_15, \
call_8_14_8_0, \
call_8_14_8_1, \
call_8_14_8_2, \
call_8_14_8_3, \
call_8_14_8_4, \
call_8_14_8_5, \
call_8_14_8_6, \
call_8_14_8_7, \
call_8_14_8_8, \
call_8_14_8_9, \
call_8_14_8_10, \
call_8_14_8_11, \
call_8_14_8_12, \
call_8_14_8_13, \
call_8_14_8_14, \
call_8_14_8_15, \
call_8_14_9_0, \
call_8_14_9_1, \
call_8_14_9_2, \
call_8_14_9_3, \
call_8_14_9_4, \
call_8_14_9_5, \
call_8_14_9_6, \
call_8_14_9_7, \
call_8_14_9_8, \
call_8_14_9_9, \
call_8_14_9_10, \
call_8_14_9_11, \
call_8_14_9_12, \
call_8_14_9_13, \
call_8_14_9_14, \
call_8_14_9_15, \
call_8_14_10_0, \
call_8_14_10_1, \
call_8_14_10_2, \
call_8_14_10_3, \
call_8_14_10_4, \
call_8_14_10_5, \
call_8_14_10_6, \
call_8_14_10_7, \
call_8_14_10_8, \
call_8_14_10_9, \
call_8_14_10_10, \
call_8_14_10_11, \
call_8_14_10_12, \
call_8_14_10_13, \
call_8_14_10_14, \
call_8_14_10_15, \
call_8_14_11_0, \
call_8_14_11_1, \
call_8_14_11_2, \
call_8_14_11_3, \
call_8_14_11_4, \
call_8_14_11_5, \
call_8_14_11_6, \
call_8_14_11_7, \
call_8_14_11_8, \
call_8_14_11_9, \
call_8_14_11_10, \
call_8_14_11_11, \
call_8_14_11_12, \
call_8_14_11_13, \
call_8_14_11_14, \
call_8_14_11_15, \
call_8_14_12_0, \
call_8_14_12_1, \
call_8_14_12_2, \
call_8_14_12_3, \
call_8_14_12_4, \
call_8_14_12_5, \
call_8_14_12_6, \
call_8_14_12_7, \
call_8_14_12_8, \
call_8_14_12_9, \
call_8_14_12_10, \
call_8_14_12_11, \
call_8_14_12_12, \
call_8_14_12_13, \
call_8_14_12_14, \
call_8_14_12_15, \
call_8_14_13_0, \
call_8_14_13_1, \
call_8_14_13_2, \
call_8_14_13_3, \
call_8_14_13_4, \
call_8_14_13_5, \
call_8_14_13_6, \
call_8_14_13_7, \
call_8_14_13_8, \
call_8_14_13_9, \
call_8_14_13_10, \
call_8_14_13_11, \
call_8_14_13_12, \
call_8_14_13_13, \
call_8_14_13_14, \
call_8_14_13_15, \
call_8_14_14_0, \
call_8_14_14_1, \
call_8_14_14_2, \
call_8_14_14_3, \
call_8_14_14_4, \
call_8_14_14_5, \
call_8_14_14_6, \
call_8_14_14_7, \
call_8_14_14_8, \
call_8_14_14_9, \
call_8_14_14_10, \
call_8_14_14_11, \
call_8_14_14_12, \
call_8_14_14_13, \
call_8_14_14_14, \
call_8_14_14_15, \
call_8_14_15_0, \
call_8_14_15_1, \
call_8_14_15_2, \
call_8_14_15_3, \
call_8_14_15_4, \
call_8_14_15_5, \
call_8_14_15_6, \
call_8_14_15_7, \
call_8_14_15_8, \
call_8_14_15_9, \
call_8_14_15_10, \
call_8_14_15_11, \
call_8_14_15_12, \
call_8_14_15_13, \
call_8_14_15_14, \
call_8_14_15_15, \
call_8_15_0_0, \
call_8_15_0_1, \
call_8_15_0_2, \
call_8_15_0_3, \
call_8_15_0_4, \
call_8_15_0_5, \
call_8_15_0_6, \
call_8_15_0_7, \
call_8_15_0_8, \
call_8_15_0_9, \
call_8_15_0_10, \
call_8_15_0_11, \
call_8_15_0_12, \
call_8_15_0_13, \
call_8_15_0_14, \
call_8_15_0_15, \
call_8_15_1_0, \
call_8_15_1_1, \
call_8_15_1_2, \
call_8_15_1_3, \
call_8_15_1_4, \
call_8_15_1_5, \
call_8_15_1_6, \
call_8_15_1_7, \
call_8_15_1_8, \
call_8_15_1_9, \
call_8_15_1_10, \
call_8_15_1_11, \
call_8_15_1_12, \
call_8_15_1_13, \
call_8_15_1_14, \
call_8_15_1_15, \
call_8_15_2_0, \
call_8_15_2_1, \
call_8_15_2_2, \
call_8_15_2_3, \
call_8_15_2_4, \
call_8_15_2_5, \
call_8_15_2_6, \
call_8_15_2_7, \
call_8_15_2_8, \
call_8_15_2_9, \
call_8_15_2_10, \
call_8_15_2_11, \
call_8_15_2_12, \
call_8_15_2_13, \
call_8_15_2_14, \
call_8_15_2_15, \
call_8_15_3_0, \
call_8_15_3_1, \
call_8_15_3_2, \
call_8_15_3_3, \
call_8_15_3_4, \
call_8_15_3_5, \
call_8_15_3_6, \
call_8_15_3_7, \
call_8_15_3_8, \
call_8_15_3_9, \
call_8_15_3_10, \
call_8_15_3_11, \
call_8_15_3_12, \
call_8_15_3_13, \
call_8_15_3_14, \
call_8_15_3_15, \
call_8_15_4_0, \
call_8_15_4_1, \
call_8_15_4_2, \
call_8_15_4_3, \
call_8_15_4_4, \
call_8_15_4_5, \
call_8_15_4_6, \
call_8_15_4_7, \
call_8_15_4_8, \
call_8_15_4_9, \
call_8_15_4_10, \
call_8_15_4_11, \
call_8_15_4_12, \
call_8_15_4_13, \
call_8_15_4_14, \
call_8_15_4_15, \
call_8_15_5_0, \
call_8_15_5_1, \
call_8_15_5_2, \
call_8_15_5_3, \
call_8_15_5_4, \
call_8_15_5_5, \
call_8_15_5_6, \
call_8_15_5_7, \
call_8_15_5_8, \
call_8_15_5_9, \
call_8_15_5_10, \
call_8_15_5_11, \
call_8_15_5_12, \
call_8_15_5_13, \
call_8_15_5_14, \
call_8_15_5_15, \
call_8_15_6_0, \
call_8_15_6_1, \
call_8_15_6_2, \
call_8_15_6_3, \
call_8_15_6_4, \
call_8_15_6_5, \
call_8_15_6_6, \
call_8_15_6_7, \
call_8_15_6_8, \
call_8_15_6_9, \
call_8_15_6_10, \
call_8_15_6_11, \
call_8_15_6_12, \
call_8_15_6_13, \
call_8_15_6_14, \
call_8_15_6_15, \
call_8_15_7_0, \
call_8_15_7_1, \
call_8_15_7_2, \
call_8_15_7_3, \
call_8_15_7_4, \
call_8_15_7_5, \
call_8_15_7_6, \
call_8_15_7_7, \
call_8_15_7_8, \
call_8_15_7_9, \
call_8_15_7_10, \
call_8_15_7_11, \
call_8_15_7_12, \
call_8_15_7_13, \
call_8_15_7_14, \
call_8_15_7_15, \
call_8_15_8_0, \
call_8_15_8_1, \
call_8_15_8_2, \
call_8_15_8_3, \
call_8_15_8_4, \
call_8_15_8_5, \
call_8_15_8_6, \
call_8_15_8_7, \
call_8_15_8_8, \
call_8_15_8_9, \
call_8_15_8_10, \
call_8_15_8_11, \
call_8_15_8_12, \
call_8_15_8_13, \
call_8_15_8_14, \
call_8_15_8_15, \
call_8_15_9_0, \
call_8_15_9_1, \
call_8_15_9_2, \
call_8_15_9_3, \
call_8_15_9_4, \
call_8_15_9_5, \
call_8_15_9_6, \
call_8_15_9_7, \
call_8_15_9_8, \
call_8_15_9_9, \
call_8_15_9_10, \
call_8_15_9_11, \
call_8_15_9_12, \
call_8_15_9_13, \
call_8_15_9_14, \
call_8_15_9_15, \
call_8_15_10_0, \
call_8_15_10_1, \
call_8_15_10_2, \
call_8_15_10_3, \
call_8_15_10_4, \
call_8_15_10_5, \
call_8_15_10_6, \
call_8_15_10_7, \
call_8_15_10_8, \
call_8_15_10_9, \
call_8_15_10_10, \
call_8_15_10_11, \
call_8_15_10_12, \
call_8_15_10_13, \
call_8_15_10_14, \
call_8_15_10_15, \
call_8_15_11_0, \
call_8_15_11_1, \
call_8_15_11_2, \
call_8_15_11_3, \
call_8_15_11_4, \
call_8_15_11_5, \
call_8_15_11_6, \
call_8_15_11_7, \
call_8_15_11_8, \
call_8_15_11_9, \
call_8_15_11_10, \
call_8_15_11_11, \
call_8_15_11_12, \
call_8_15_11_13, \
call_8_15_11_14, \
call_8_15_11_15, \
call_8_15_12_0, \
call_8_15_12_1, \
call_8_15_12_2, \
call_8_15_12_3, \
call_8_15_12_4, \
call_8_15_12_5, \
call_8_15_12_6, \
call_8_15_12_7, \
call_8_15_12_8, \
call_8_15_12_9, \
call_8_15_12_10, \
call_8_15_12_11, \
call_8_15_12_12, \
call_8_15_12_13, \
call_8_15_12_14, \
call_8_15_12_15, \
call_8_15_13_0, \
call_8_15_13_1, \
call_8_15_13_2, \
call_8_15_13_3, \
call_8_15_13_4, \
call_8_15_13_5, \
call_8_15_13_6, \
call_8_15_13_7, \
call_8_15_13_8, \
call_8_15_13_9, \
call_8_15_13_10, \
call_8_15_13_11, \
call_8_15_13_12, \
call_8_15_13_13, \
call_8_15_13_14, \
call_8_15_13_15, \
call_8_15_14_0, \
call_8_15_14_1, \
call_8_15_14_2, \
call_8_15_14_3, \
call_8_15_14_4, \
call_8_15_14_5, \
call_8_15_14_6, \
call_8_15_14_7, \
call_8_15_14_8, \
call_8_15_14_9, \
call_8_15_14_10, \
call_8_15_14_11, \
call_8_15_14_12, \
call_8_15_14_13, \
call_8_15_14_14, \
call_8_15_14_15, \
call_8_15_15_0, \
call_8_15_15_1, \
call_8_15_15_2, \
call_8_15_15_3, \
call_8_15_15_4, \
call_8_15_15_5, \
call_8_15_15_6, \
call_8_15_15_7, \
call_8_15_15_8, \
call_8_15_15_9, \
call_8_15_15_10, \
call_8_15_15_11, \
call_8_15_15_12, \
call_8_15_15_13, \
call_8_15_15_14, \
call_8_15_15_15, \
call_9_0_0_0, \
call_9_0_0_1, \
call_9_0_0_2, \
call_9_0_0_3, \
call_9_0_0_4, \
call_9_0_0_5, \
call_9_0_0_6, \
call_9_0_0_7, \
call_9_0_0_8, \
call_9_0_0_9, \
call_9_0_0_10, \
call_9_0_0_11, \
call_9_0_0_12, \
call_9_0_0_13, \
call_9_0_0_14, \
call_9_0_0_15, \
call_9_0_1_0, \
call_9_0_1_1, \
call_9_0_1_2, \
call_9_0_1_3, \
call_9_0_1_4, \
call_9_0_1_5, \
call_9_0_1_6, \
call_9_0_1_7, \
call_9_0_1_8, \
call_9_0_1_9, \
call_9_0_1_10, \
call_9_0_1_11, \
call_9_0_1_12, \
call_9_0_1_13, \
call_9_0_1_14, \
call_9_0_1_15, \
call_9_0_2_0, \
call_9_0_2_1, \
call_9_0_2_2, \
call_9_0_2_3, \
call_9_0_2_4, \
call_9_0_2_5, \
call_9_0_2_6, \
call_9_0_2_7, \
call_9_0_2_8, \
call_9_0_2_9, \
call_9_0_2_10, \
call_9_0_2_11, \
call_9_0_2_12, \
call_9_0_2_13, \
call_9_0_2_14, \
call_9_0_2_15, \
call_9_0_3_0, \
call_9_0_3_1, \
call_9_0_3_2, \
call_9_0_3_3, \
call_9_0_3_4, \
call_9_0_3_5, \
call_9_0_3_6, \
call_9_0_3_7, \
call_9_0_3_8, \
call_9_0_3_9, \
call_9_0_3_10, \
call_9_0_3_11, \
call_9_0_3_12, \
call_9_0_3_13, \
call_9_0_3_14, \
call_9_0_3_15, \
call_9_0_4_0, \
call_9_0_4_1, \
call_9_0_4_2, \
call_9_0_4_3, \
call_9_0_4_4, \
call_9_0_4_5, \
call_9_0_4_6, \
call_9_0_4_7, \
call_9_0_4_8, \
call_9_0_4_9, \
call_9_0_4_10, \
call_9_0_4_11, \
call_9_0_4_12, \
call_9_0_4_13, \
call_9_0_4_14, \
call_9_0_4_15, \
call_9_0_5_0, \
call_9_0_5_1, \
call_9_0_5_2, \
call_9_0_5_3, \
call_9_0_5_4, \
call_9_0_5_5, \
call_9_0_5_6, \
call_9_0_5_7, \
call_9_0_5_8, \
call_9_0_5_9, \
call_9_0_5_10, \
call_9_0_5_11, \
call_9_0_5_12, \
call_9_0_5_13, \
call_9_0_5_14, \
call_9_0_5_15, \
call_9_0_6_0, \
call_9_0_6_1, \
call_9_0_6_2, \
call_9_0_6_3, \
call_9_0_6_4, \
call_9_0_6_5, \
call_9_0_6_6, \
call_9_0_6_7, \
call_9_0_6_8, \
call_9_0_6_9, \
call_9_0_6_10, \
call_9_0_6_11, \
call_9_0_6_12, \
call_9_0_6_13, \
call_9_0_6_14, \
call_9_0_6_15, \
call_9_0_7_0, \
call_9_0_7_1, \
call_9_0_7_2, \
call_9_0_7_3, \
call_9_0_7_4, \
call_9_0_7_5, \
call_9_0_7_6, \
call_9_0_7_7, \
call_9_0_7_8, \
call_9_0_7_9, \
call_9_0_7_10, \
call_9_0_7_11, \
call_9_0_7_12, \
call_9_0_7_13, \
call_9_0_7_14, \
call_9_0_7_15, \
call_9_0_8_0, \
call_9_0_8_1, \
call_9_0_8_2, \
call_9_0_8_3, \
call_9_0_8_4, \
call_9_0_8_5, \
call_9_0_8_6, \
call_9_0_8_7, \
call_9_0_8_8, \
call_9_0_8_9, \
call_9_0_8_10, \
call_9_0_8_11, \
call_9_0_8_12, \
call_9_0_8_13, \
call_9_0_8_14, \
call_9_0_8_15, \
call_9_0_9_0, \
call_9_0_9_1, \
call_9_0_9_2, \
call_9_0_9_3, \
call_9_0_9_4, \
call_9_0_9_5, \
call_9_0_9_6, \
call_9_0_9_7, \
call_9_0_9_8, \
call_9_0_9_9, \
call_9_0_9_10, \
call_9_0_9_11, \
call_9_0_9_12, \
call_9_0_9_13, \
call_9_0_9_14, \
call_9_0_9_15, \
call_9_0_10_0, \
call_9_0_10_1, \
call_9_0_10_2, \
call_9_0_10_3, \
call_9_0_10_4, \
call_9_0_10_5, \
call_9_0_10_6, \
call_9_0_10_7, \
call_9_0_10_8, \
call_9_0_10_9, \
call_9_0_10_10, \
call_9_0_10_11, \
call_9_0_10_12, \
call_9_0_10_13, \
call_9_0_10_14, \
call_9_0_10_15, \
call_9_0_11_0, \
call_9_0_11_1, \
call_9_0_11_2, \
call_9_0_11_3, \
call_9_0_11_4, \
call_9_0_11_5, \
call_9_0_11_6, \
call_9_0_11_7, \
call_9_0_11_8, \
call_9_0_11_9, \
call_9_0_11_10, \
call_9_0_11_11, \
call_9_0_11_12, \
call_9_0_11_13, \
call_9_0_11_14, \
call_9_0_11_15, \
call_9_0_12_0, \
call_9_0_12_1, \
call_9_0_12_2, \
call_9_0_12_3, \
call_9_0_12_4, \
call_9_0_12_5, \
call_9_0_12_6, \
call_9_0_12_7, \
call_9_0_12_8, \
call_9_0_12_9, \
call_9_0_12_10, \
call_9_0_12_11, \
call_9_0_12_12, \
call_9_0_12_13, \
call_9_0_12_14, \
call_9_0_12_15, \
call_9_0_13_0, \
call_9_0_13_1, \
call_9_0_13_2, \
call_9_0_13_3, \
call_9_0_13_4, \
call_9_0_13_5, \
call_9_0_13_6, \
call_9_0_13_7, \
call_9_0_13_8, \
call_9_0_13_9, \
call_9_0_13_10, \
call_9_0_13_11, \
call_9_0_13_12, \
call_9_0_13_13, \
call_9_0_13_14, \
call_9_0_13_15, \
call_9_0_14_0, \
call_9_0_14_1, \
call_9_0_14_2, \
call_9_0_14_3, \
call_9_0_14_4, \
call_9_0_14_5, \
call_9_0_14_6, \
call_9_0_14_7, \
call_9_0_14_8, \
call_9_0_14_9, \
call_9_0_14_10, \
call_9_0_14_11, \
call_9_0_14_12, \
call_9_0_14_13, \
call_9_0_14_14, \
call_9_0_14_15, \
call_9_0_15_0, \
call_9_0_15_1, \
call_9_0_15_2, \
call_9_0_15_3, \
call_9_0_15_4, \
call_9_0_15_5, \
call_9_0_15_6, \
call_9_0_15_7, \
call_9_0_15_8, \
call_9_0_15_9, \
call_9_0_15_10, \
call_9_0_15_11, \
call_9_0_15_12, \
call_9_0_15_13, \
call_9_0_15_14, \
call_9_0_15_15, \
call_9_1_0_0, \
call_9_1_0_1, \
call_9_1_0_2, \
call_9_1_0_3, \
call_9_1_0_4, \
call_9_1_0_5, \
call_9_1_0_6, \
call_9_1_0_7, \
call_9_1_0_8, \
call_9_1_0_9, \
call_9_1_0_10, \
call_9_1_0_11, \
call_9_1_0_12, \
call_9_1_0_13, \
call_9_1_0_14, \
call_9_1_0_15, \
call_9_1_1_0, \
call_9_1_1_1, \
call_9_1_1_2, \
call_9_1_1_3, \
call_9_1_1_4, \
call_9_1_1_5, \
call_9_1_1_6, \
call_9_1_1_7, \
call_9_1_1_8, \
call_9_1_1_9, \
call_9_1_1_10, \
call_9_1_1_11, \
call_9_1_1_12, \
call_9_1_1_13, \
call_9_1_1_14, \
call_9_1_1_15, \
call_9_1_2_0, \
call_9_1_2_1, \
call_9_1_2_2, \
call_9_1_2_3, \
call_9_1_2_4, \
call_9_1_2_5, \
call_9_1_2_6, \
call_9_1_2_7, \
call_9_1_2_8, \
call_9_1_2_9, \
call_9_1_2_10, \
call_9_1_2_11, \
call_9_1_2_12, \
call_9_1_2_13, \
call_9_1_2_14, \
call_9_1_2_15, \
call_9_1_3_0, \
call_9_1_3_1, \
call_9_1_3_2, \
call_9_1_3_3, \
call_9_1_3_4, \
call_9_1_3_5, \
call_9_1_3_6, \
call_9_1_3_7, \
call_9_1_3_8, \
call_9_1_3_9, \
call_9_1_3_10, \
call_9_1_3_11, \
call_9_1_3_12, \
call_9_1_3_13, \
call_9_1_3_14, \
call_9_1_3_15, \
call_9_1_4_0, \
call_9_1_4_1, \
call_9_1_4_2, \
call_9_1_4_3, \
call_9_1_4_4, \
call_9_1_4_5, \
call_9_1_4_6, \
call_9_1_4_7, \
call_9_1_4_8, \
call_9_1_4_9, \
call_9_1_4_10, \
call_9_1_4_11, \
call_9_1_4_12, \
call_9_1_4_13, \
call_9_1_4_14, \
call_9_1_4_15, \
call_9_1_5_0, \
call_9_1_5_1, \
call_9_1_5_2, \
call_9_1_5_3, \
call_9_1_5_4, \
call_9_1_5_5, \
call_9_1_5_6, \
call_9_1_5_7, \
call_9_1_5_8, \
call_9_1_5_9, \
call_9_1_5_10, \
call_9_1_5_11, \
call_9_1_5_12, \
call_9_1_5_13, \
call_9_1_5_14, \
call_9_1_5_15, \
call_9_1_6_0, \
call_9_1_6_1, \
call_9_1_6_2, \
call_9_1_6_3, \
call_9_1_6_4, \
call_9_1_6_5, \
call_9_1_6_6, \
call_9_1_6_7, \
call_9_1_6_8, \
call_9_1_6_9, \
call_9_1_6_10, \
call_9_1_6_11, \
call_9_1_6_12, \
call_9_1_6_13, \
call_9_1_6_14, \
call_9_1_6_15, \
call_9_1_7_0, \
call_9_1_7_1, \
call_9_1_7_2, \
call_9_1_7_3, \
call_9_1_7_4, \
call_9_1_7_5, \
call_9_1_7_6, \
call_9_1_7_7, \
call_9_1_7_8, \
call_9_1_7_9, \
call_9_1_7_10, \
call_9_1_7_11, \
call_9_1_7_12, \
call_9_1_7_13, \
call_9_1_7_14, \
call_9_1_7_15, \
call_9_1_8_0, \
call_9_1_8_1, \
call_9_1_8_2, \
call_9_1_8_3, \
call_9_1_8_4, \
call_9_1_8_5, \
call_9_1_8_6, \
call_9_1_8_7, \
call_9_1_8_8, \
call_9_1_8_9, \
call_9_1_8_10, \
call_9_1_8_11, \
call_9_1_8_12, \
call_9_1_8_13, \
call_9_1_8_14, \
call_9_1_8_15, \
call_9_1_9_0, \
call_9_1_9_1, \
call_9_1_9_2, \
call_9_1_9_3, \
call_9_1_9_4, \
call_9_1_9_5, \
call_9_1_9_6, \
call_9_1_9_7, \
call_9_1_9_8, \
call_9_1_9_9, \
call_9_1_9_10, \
call_9_1_9_11, \
call_9_1_9_12, \
call_9_1_9_13, \
call_9_1_9_14, \
call_9_1_9_15, \
call_9_1_10_0, \
call_9_1_10_1, \
call_9_1_10_2, \
call_9_1_10_3, \
call_9_1_10_4, \
call_9_1_10_5, \
call_9_1_10_6, \
call_9_1_10_7, \
call_9_1_10_8, \
call_9_1_10_9, \
call_9_1_10_10, \
call_9_1_10_11, \
call_9_1_10_12, \
call_9_1_10_13, \
call_9_1_10_14, \
call_9_1_10_15, \
call_9_1_11_0, \
call_9_1_11_1, \
call_9_1_11_2, \
call_9_1_11_3, \
call_9_1_11_4, \
call_9_1_11_5, \
call_9_1_11_6, \
call_9_1_11_7, \
call_9_1_11_8, \
call_9_1_11_9, \
call_9_1_11_10, \
call_9_1_11_11, \
call_9_1_11_12, \
call_9_1_11_13, \
call_9_1_11_14, \
call_9_1_11_15, \
call_9_1_12_0, \
call_9_1_12_1, \
call_9_1_12_2, \
call_9_1_12_3, \
call_9_1_12_4, \
call_9_1_12_5, \
call_9_1_12_6, \
call_9_1_12_7, \
call_9_1_12_8, \
call_9_1_12_9, \
call_9_1_12_10, \
call_9_1_12_11, \
call_9_1_12_12, \
call_9_1_12_13, \
call_9_1_12_14, \
call_9_1_12_15, \
call_9_1_13_0, \
call_9_1_13_1, \
call_9_1_13_2, \
call_9_1_13_3, \
call_9_1_13_4, \
call_9_1_13_5, \
call_9_1_13_6, \
call_9_1_13_7, \
call_9_1_13_8, \
call_9_1_13_9, \
call_9_1_13_10, \
call_9_1_13_11, \
call_9_1_13_12, \
call_9_1_13_13, \
call_9_1_13_14, \
call_9_1_13_15, \
call_9_1_14_0, \
call_9_1_14_1, \
call_9_1_14_2, \
call_9_1_14_3, \
call_9_1_14_4, \
call_9_1_14_5, \
call_9_1_14_6, \
call_9_1_14_7, \
call_9_1_14_8, \
call_9_1_14_9, \
call_9_1_14_10, \
call_9_1_14_11, \
call_9_1_14_12, \
call_9_1_14_13, \
call_9_1_14_14, \
call_9_1_14_15, \
call_9_1_15_0, \
call_9_1_15_1, \
call_9_1_15_2, \
call_9_1_15_3, \
call_9_1_15_4, \
call_9_1_15_5, \
call_9_1_15_6, \
call_9_1_15_7, \
call_9_1_15_8, \
call_9_1_15_9, \
call_9_1_15_10, \
call_9_1_15_11, \
call_9_1_15_12, \
call_9_1_15_13, \
call_9_1_15_14, \
call_9_1_15_15, \
call_9_2_0_0, \
call_9_2_0_1, \
call_9_2_0_2, \
call_9_2_0_3, \
call_9_2_0_4, \
call_9_2_0_5, \
call_9_2_0_6, \
call_9_2_0_7, \
call_9_2_0_8, \
call_9_2_0_9, \
call_9_2_0_10, \
call_9_2_0_11, \
call_9_2_0_12, \
call_9_2_0_13, \
call_9_2_0_14, \
call_9_2_0_15, \
call_9_2_1_0, \
call_9_2_1_1, \
call_9_2_1_2, \
call_9_2_1_3, \
call_9_2_1_4, \
call_9_2_1_5, \
call_9_2_1_6, \
call_9_2_1_7, \
call_9_2_1_8, \
call_9_2_1_9, \
call_9_2_1_10, \
call_9_2_1_11, \
call_9_2_1_12, \
call_9_2_1_13, \
call_9_2_1_14, \
call_9_2_1_15, \
call_9_2_2_0, \
call_9_2_2_1, \
call_9_2_2_2, \
call_9_2_2_3, \
call_9_2_2_4, \
call_9_2_2_5, \
call_9_2_2_6, \
call_9_2_2_7, \
call_9_2_2_8, \
call_9_2_2_9, \
call_9_2_2_10, \
call_9_2_2_11, \
call_9_2_2_12, \
call_9_2_2_13, \
call_9_2_2_14, \
call_9_2_2_15, \
call_9_2_3_0, \
call_9_2_3_1, \
call_9_2_3_2, \
call_9_2_3_3, \
call_9_2_3_4, \
call_9_2_3_5, \
call_9_2_3_6, \
call_9_2_3_7, \
call_9_2_3_8, \
call_9_2_3_9, \
call_9_2_3_10, \
call_9_2_3_11, \
call_9_2_3_12, \
call_9_2_3_13, \
call_9_2_3_14, \
call_9_2_3_15, \
call_9_2_4_0, \
call_9_2_4_1, \
call_9_2_4_2, \
call_9_2_4_3, \
call_9_2_4_4, \
call_9_2_4_5, \
call_9_2_4_6, \
call_9_2_4_7, \
call_9_2_4_8, \
call_9_2_4_9, \
call_9_2_4_10, \
call_9_2_4_11, \
call_9_2_4_12, \
call_9_2_4_13, \
call_9_2_4_14, \
call_9_2_4_15, \
call_9_2_5_0, \
call_9_2_5_1, \
call_9_2_5_2, \
call_9_2_5_3, \
call_9_2_5_4, \
call_9_2_5_5, \
call_9_2_5_6, \
call_9_2_5_7, \
call_9_2_5_8, \
call_9_2_5_9, \
call_9_2_5_10, \
call_9_2_5_11, \
call_9_2_5_12, \
call_9_2_5_13, \
call_9_2_5_14, \
call_9_2_5_15, \
call_9_2_6_0, \
call_9_2_6_1, \
call_9_2_6_2, \
call_9_2_6_3, \
call_9_2_6_4, \
call_9_2_6_5, \
call_9_2_6_6, \
call_9_2_6_7, \
call_9_2_6_8, \
call_9_2_6_9, \
call_9_2_6_10, \
call_9_2_6_11, \
call_9_2_6_12, \
call_9_2_6_13, \
call_9_2_6_14, \
call_9_2_6_15, \
call_9_2_7_0, \
call_9_2_7_1, \
call_9_2_7_2, \
call_9_2_7_3, \
call_9_2_7_4, \
call_9_2_7_5, \
call_9_2_7_6, \
call_9_2_7_7, \
call_9_2_7_8, \
call_9_2_7_9, \
call_9_2_7_10, \
call_9_2_7_11, \
call_9_2_7_12, \
call_9_2_7_13, \
call_9_2_7_14, \
call_9_2_7_15, \
call_9_2_8_0, \
call_9_2_8_1, \
call_9_2_8_2, \
call_9_2_8_3, \
call_9_2_8_4, \
call_9_2_8_5, \
call_9_2_8_6, \
call_9_2_8_7, \
call_9_2_8_8, \
call_9_2_8_9, \
call_9_2_8_10, \
call_9_2_8_11, \
call_9_2_8_12, \
call_9_2_8_13, \
call_9_2_8_14, \
call_9_2_8_15, \
call_9_2_9_0, \
call_9_2_9_1, \
call_9_2_9_2, \
call_9_2_9_3, \
call_9_2_9_4, \
call_9_2_9_5, \
call_9_2_9_6, \
call_9_2_9_7, \
call_9_2_9_8, \
call_9_2_9_9, \
call_9_2_9_10, \
call_9_2_9_11, \
call_9_2_9_12, \
call_9_2_9_13, \
call_9_2_9_14, \
call_9_2_9_15, \
call_9_2_10_0, \
call_9_2_10_1, \
call_9_2_10_2, \
call_9_2_10_3, \
call_9_2_10_4, \
call_9_2_10_5, \
call_9_2_10_6, \
call_9_2_10_7, \
call_9_2_10_8, \
call_9_2_10_9, \
call_9_2_10_10, \
call_9_2_10_11, \
call_9_2_10_12, \
call_9_2_10_13, \
call_9_2_10_14, \
call_9_2_10_15, \
call_9_2_11_0, \
call_9_2_11_1, \
call_9_2_11_2, \
call_9_2_11_3, \
call_9_2_11_4, \
call_9_2_11_5, \
call_9_2_11_6, \
call_9_2_11_7, \
call_9_2_11_8, \
call_9_2_11_9, \
call_9_2_11_10, \
call_9_2_11_11, \
call_9_2_11_12, \
call_9_2_11_13, \
call_9_2_11_14, \
call_9_2_11_15, \
call_9_2_12_0, \
call_9_2_12_1, \
call_9_2_12_2, \
call_9_2_12_3, \
call_9_2_12_4, \
call_9_2_12_5, \
call_9_2_12_6, \
call_9_2_12_7, \
call_9_2_12_8, \
call_9_2_12_9, \
call_9_2_12_10, \
call_9_2_12_11, \
call_9_2_12_12, \
call_9_2_12_13, \
call_9_2_12_14, \
call_9_2_12_15, \
call_9_2_13_0, \
call_9_2_13_1, \
call_9_2_13_2, \
call_9_2_13_3, \
call_9_2_13_4, \
call_9_2_13_5, \
call_9_2_13_6, \
call_9_2_13_7, \
call_9_2_13_8, \
call_9_2_13_9, \
call_9_2_13_10, \
call_9_2_13_11, \
call_9_2_13_12, \
call_9_2_13_13, \
call_9_2_13_14, \
call_9_2_13_15, \
call_9_2_14_0, \
call_9_2_14_1, \
call_9_2_14_2, \
call_9_2_14_3, \
call_9_2_14_4, \
call_9_2_14_5, \
call_9_2_14_6, \
call_9_2_14_7, \
call_9_2_14_8, \
call_9_2_14_9, \
call_9_2_14_10, \
call_9_2_14_11, \
call_9_2_14_12, \
call_9_2_14_13, \
call_9_2_14_14, \
call_9_2_14_15, \
call_9_2_15_0, \
call_9_2_15_1, \
call_9_2_15_2, \
call_9_2_15_3, \
call_9_2_15_4, \
call_9_2_15_5, \
call_9_2_15_6, \
call_9_2_15_7, \
call_9_2_15_8, \
call_9_2_15_9, \
call_9_2_15_10, \
call_9_2_15_11, \
call_9_2_15_12, \
call_9_2_15_13, \
call_9_2_15_14, \
call_9_2_15_15, \
call_9_3_0_0, \
call_9_3_0_1, \
call_9_3_0_2, \
call_9_3_0_3, \
call_9_3_0_4, \
call_9_3_0_5, \
call_9_3_0_6, \
call_9_3_0_7, \
call_9_3_0_8, \
call_9_3_0_9, \
call_9_3_0_10, \
call_9_3_0_11, \
call_9_3_0_12, \
call_9_3_0_13, \
call_9_3_0_14, \
call_9_3_0_15, \
call_9_3_1_0, \
call_9_3_1_1, \
call_9_3_1_2, \
call_9_3_1_3, \
call_9_3_1_4, \
call_9_3_1_5, \
call_9_3_1_6, \
call_9_3_1_7, \
call_9_3_1_8, \
call_9_3_1_9, \
call_9_3_1_10, \
call_9_3_1_11, \
call_9_3_1_12, \
call_9_3_1_13, \
call_9_3_1_14, \
call_9_3_1_15, \
call_9_3_2_0, \
call_9_3_2_1, \
call_9_3_2_2, \
call_9_3_2_3, \
call_9_3_2_4, \
call_9_3_2_5, \
call_9_3_2_6, \
call_9_3_2_7, \
call_9_3_2_8, \
call_9_3_2_9, \
call_9_3_2_10, \
call_9_3_2_11, \
call_9_3_2_12, \
call_9_3_2_13, \
call_9_3_2_14, \
call_9_3_2_15, \
call_9_3_3_0, \
call_9_3_3_1, \
call_9_3_3_2, \
call_9_3_3_3, \
call_9_3_3_4, \
call_9_3_3_5, \
call_9_3_3_6, \
call_9_3_3_7, \
call_9_3_3_8, \
call_9_3_3_9, \
call_9_3_3_10, \
call_9_3_3_11, \
call_9_3_3_12, \
call_9_3_3_13, \
call_9_3_3_14, \
call_9_3_3_15, \
call_9_3_4_0, \
call_9_3_4_1, \
call_9_3_4_2, \
call_9_3_4_3, \
call_9_3_4_4, \
call_9_3_4_5, \
call_9_3_4_6, \
call_9_3_4_7, \
call_9_3_4_8, \
call_9_3_4_9, \
call_9_3_4_10, \
call_9_3_4_11, \
call_9_3_4_12, \
call_9_3_4_13, \
call_9_3_4_14, \
call_9_3_4_15, \
call_9_3_5_0, \
call_9_3_5_1, \
call_9_3_5_2, \
call_9_3_5_3, \
call_9_3_5_4, \
call_9_3_5_5, \
call_9_3_5_6, \
call_9_3_5_7, \
call_9_3_5_8, \
call_9_3_5_9, \
call_9_3_5_10, \
call_9_3_5_11, \
call_9_3_5_12, \
call_9_3_5_13, \
call_9_3_5_14, \
call_9_3_5_15, \
call_9_3_6_0, \
call_9_3_6_1, \
call_9_3_6_2, \
call_9_3_6_3, \
call_9_3_6_4, \
call_9_3_6_5, \
call_9_3_6_6, \
call_9_3_6_7, \
call_9_3_6_8, \
call_9_3_6_9, \
call_9_3_6_10, \
call_9_3_6_11, \
call_9_3_6_12, \
call_9_3_6_13, \
call_9_3_6_14, \
call_9_3_6_15, \
call_9_3_7_0, \
call_9_3_7_1, \
call_9_3_7_2, \
call_9_3_7_3, \
call_9_3_7_4, \
call_9_3_7_5, \
call_9_3_7_6, \
call_9_3_7_7, \
call_9_3_7_8, \
call_9_3_7_9, \
call_9_3_7_10, \
call_9_3_7_11, \
call_9_3_7_12, \
call_9_3_7_13, \
call_9_3_7_14, \
call_9_3_7_15, \
call_9_3_8_0, \
call_9_3_8_1, \
call_9_3_8_2, \
call_9_3_8_3, \
call_9_3_8_4, \
call_9_3_8_5, \
call_9_3_8_6, \
call_9_3_8_7, \
call_9_3_8_8, \
call_9_3_8_9, \
call_9_3_8_10, \
call_9_3_8_11, \
call_9_3_8_12, \
call_9_3_8_13, \
call_9_3_8_14, \
call_9_3_8_15, \
call_9_3_9_0, \
call_9_3_9_1, \
call_9_3_9_2, \
call_9_3_9_3, \
call_9_3_9_4, \
call_9_3_9_5, \
call_9_3_9_6, \
call_9_3_9_7, \
call_9_3_9_8, \
call_9_3_9_9, \
call_9_3_9_10, \
call_9_3_9_11, \
call_9_3_9_12, \
call_9_3_9_13, \
call_9_3_9_14, \
call_9_3_9_15, \
call_9_3_10_0, \
call_9_3_10_1, \
call_9_3_10_2, \
call_9_3_10_3, \
call_9_3_10_4, \
call_9_3_10_5, \
call_9_3_10_6, \
call_9_3_10_7, \
call_9_3_10_8, \
call_9_3_10_9, \
call_9_3_10_10, \
call_9_3_10_11, \
call_9_3_10_12, \
call_9_3_10_13, \
call_9_3_10_14, \
call_9_3_10_15, \
call_9_3_11_0, \
call_9_3_11_1, \
call_9_3_11_2, \
call_9_3_11_3, \
call_9_3_11_4, \
call_9_3_11_5, \
call_9_3_11_6, \
call_9_3_11_7, \
call_9_3_11_8, \
call_9_3_11_9, \
call_9_3_11_10, \
call_9_3_11_11, \
call_9_3_11_12, \
call_9_3_11_13, \
call_9_3_11_14, \
call_9_3_11_15, \
call_9_3_12_0, \
call_9_3_12_1, \
call_9_3_12_2, \
call_9_3_12_3, \
call_9_3_12_4, \
call_9_3_12_5, \
call_9_3_12_6, \
call_9_3_12_7, \
call_9_3_12_8, \
call_9_3_12_9, \
call_9_3_12_10, \
call_9_3_12_11, \
call_9_3_12_12, \
call_9_3_12_13, \
call_9_3_12_14, \
call_9_3_12_15, \
call_9_3_13_0, \
call_9_3_13_1, \
call_9_3_13_2, \
call_9_3_13_3, \
call_9_3_13_4, \
call_9_3_13_5, \
call_9_3_13_6, \
call_9_3_13_7, \
call_9_3_13_8, \
call_9_3_13_9, \
call_9_3_13_10, \
call_9_3_13_11, \
call_9_3_13_12, \
call_9_3_13_13, \
call_9_3_13_14, \
call_9_3_13_15, \
call_9_3_14_0, \
call_9_3_14_1, \
call_9_3_14_2, \
call_9_3_14_3, \
call_9_3_14_4, \
call_9_3_14_5, \
call_9_3_14_6, \
call_9_3_14_7, \
call_9_3_14_8, \
call_9_3_14_9, \
call_9_3_14_10, \
call_9_3_14_11, \
call_9_3_14_12, \
call_9_3_14_13, \
call_9_3_14_14, \
call_9_3_14_15, \
call_9_3_15_0, \
call_9_3_15_1, \
call_9_3_15_2, \
call_9_3_15_3, \
call_9_3_15_4, \
call_9_3_15_5, \
call_9_3_15_6, \
call_9_3_15_7, \
call_9_3_15_8, \
call_9_3_15_9, \
call_9_3_15_10, \
call_9_3_15_11, \
call_9_3_15_12, \
call_9_3_15_13, \
call_9_3_15_14, \
call_9_3_15_15, \
call_9_4_0_0, \
call_9_4_0_1, \
call_9_4_0_2, \
call_9_4_0_3, \
call_9_4_0_4, \
call_9_4_0_5, \
call_9_4_0_6, \
call_9_4_0_7, \
call_9_4_0_8, \
call_9_4_0_9, \
call_9_4_0_10, \
call_9_4_0_11, \
call_9_4_0_12, \
call_9_4_0_13, \
call_9_4_0_14, \
call_9_4_0_15, \
call_9_4_1_0, \
call_9_4_1_1, \
call_9_4_1_2, \
call_9_4_1_3, \
call_9_4_1_4, \
call_9_4_1_5, \
call_9_4_1_6, \
call_9_4_1_7, \
call_9_4_1_8, \
call_9_4_1_9, \
call_9_4_1_10, \
call_9_4_1_11, \
call_9_4_1_12, \
call_9_4_1_13, \
call_9_4_1_14, \
call_9_4_1_15, \
call_9_4_2_0, \
call_9_4_2_1, \
call_9_4_2_2, \
call_9_4_2_3, \
call_9_4_2_4, \
call_9_4_2_5, \
call_9_4_2_6, \
call_9_4_2_7, \
call_9_4_2_8, \
call_9_4_2_9, \
call_9_4_2_10, \
call_9_4_2_11, \
call_9_4_2_12, \
call_9_4_2_13, \
call_9_4_2_14, \
call_9_4_2_15, \
call_9_4_3_0, \
call_9_4_3_1, \
call_9_4_3_2, \
call_9_4_3_3, \
call_9_4_3_4, \
call_9_4_3_5, \
call_9_4_3_6, \
call_9_4_3_7, \
call_9_4_3_8, \
call_9_4_3_9, \
call_9_4_3_10, \
call_9_4_3_11, \
call_9_4_3_12, \
call_9_4_3_13, \
call_9_4_3_14, \
call_9_4_3_15, \
call_9_4_4_0, \
call_9_4_4_1, \
call_9_4_4_2, \
call_9_4_4_3, \
call_9_4_4_4, \
call_9_4_4_5, \
call_9_4_4_6, \
call_9_4_4_7, \
call_9_4_4_8, \
call_9_4_4_9, \
call_9_4_4_10, \
call_9_4_4_11, \
call_9_4_4_12, \
call_9_4_4_13, \
call_9_4_4_14, \
call_9_4_4_15, \
call_9_4_5_0, \
call_9_4_5_1, \
call_9_4_5_2, \
call_9_4_5_3, \
call_9_4_5_4, \
call_9_4_5_5, \
call_9_4_5_6, \
call_9_4_5_7, \
call_9_4_5_8, \
call_9_4_5_9, \
call_9_4_5_10, \
call_9_4_5_11, \
call_9_4_5_12, \
call_9_4_5_13, \
call_9_4_5_14, \
call_9_4_5_15, \
call_9_4_6_0, \
call_9_4_6_1, \
call_9_4_6_2, \
call_9_4_6_3, \
call_9_4_6_4, \
call_9_4_6_5, \
call_9_4_6_6, \
call_9_4_6_7, \
call_9_4_6_8, \
call_9_4_6_9, \
call_9_4_6_10, \
call_9_4_6_11, \
call_9_4_6_12, \
call_9_4_6_13, \
call_9_4_6_14, \
call_9_4_6_15, \
call_9_4_7_0, \
call_9_4_7_1, \
call_9_4_7_2, \
call_9_4_7_3, \
call_9_4_7_4, \
call_9_4_7_5, \
call_9_4_7_6, \
call_9_4_7_7, \
call_9_4_7_8, \
call_9_4_7_9, \
call_9_4_7_10, \
call_9_4_7_11, \
call_9_4_7_12, \
call_9_4_7_13, \
call_9_4_7_14, \
call_9_4_7_15, \
call_9_4_8_0, \
call_9_4_8_1, \
call_9_4_8_2, \
call_9_4_8_3, \
call_9_4_8_4, \
call_9_4_8_5, \
call_9_4_8_6, \
call_9_4_8_7, \
call_9_4_8_8, \
call_9_4_8_9, \
call_9_4_8_10, \
call_9_4_8_11, \
call_9_4_8_12, \
call_9_4_8_13, \
call_9_4_8_14, \
call_9_4_8_15, \
call_9_4_9_0, \
call_9_4_9_1, \
call_9_4_9_2, \
call_9_4_9_3, \
call_9_4_9_4, \
call_9_4_9_5, \
call_9_4_9_6, \
call_9_4_9_7, \
call_9_4_9_8, \
call_9_4_9_9, \
call_9_4_9_10, \
call_9_4_9_11, \
call_9_4_9_12, \
call_9_4_9_13, \
call_9_4_9_14, \
call_9_4_9_15, \
call_9_4_10_0, \
call_9_4_10_1, \
call_9_4_10_2, \
call_9_4_10_3, \
call_9_4_10_4, \
call_9_4_10_5, \
call_9_4_10_6, \
call_9_4_10_7, \
call_9_4_10_8, \
call_9_4_10_9, \
call_9_4_10_10, \
call_9_4_10_11, \
call_9_4_10_12, \
call_9_4_10_13, \
call_9_4_10_14, \
call_9_4_10_15, \
call_9_4_11_0, \
call_9_4_11_1, \
call_9_4_11_2, \
call_9_4_11_3, \
call_9_4_11_4, \
call_9_4_11_5, \
call_9_4_11_6, \
call_9_4_11_7, \
call_9_4_11_8, \
call_9_4_11_9, \
call_9_4_11_10, \
call_9_4_11_11, \
call_9_4_11_12, \
call_9_4_11_13, \
call_9_4_11_14, \
call_9_4_11_15, \
call_9_4_12_0, \
call_9_4_12_1, \
call_9_4_12_2, \
call_9_4_12_3, \
call_9_4_12_4, \
call_9_4_12_5, \
call_9_4_12_6, \
call_9_4_12_7, \
call_9_4_12_8, \
call_9_4_12_9, \
call_9_4_12_10, \
call_9_4_12_11, \
call_9_4_12_12, \
call_9_4_12_13, \
call_9_4_12_14, \
call_9_4_12_15, \
call_9_4_13_0, \
call_9_4_13_1, \
call_9_4_13_2, \
call_9_4_13_3, \
call_9_4_13_4, \
call_9_4_13_5, \
call_9_4_13_6, \
call_9_4_13_7, \
call_9_4_13_8, \
call_9_4_13_9, \
call_9_4_13_10, \
call_9_4_13_11, \
call_9_4_13_12, \
call_9_4_13_13, \
call_9_4_13_14, \
call_9_4_13_15, \
call_9_4_14_0, \
call_9_4_14_1, \
call_9_4_14_2, \
call_9_4_14_3, \
call_9_4_14_4, \
call_9_4_14_5, \
call_9_4_14_6, \
call_9_4_14_7, \
call_9_4_14_8, \
call_9_4_14_9, \
call_9_4_14_10, \
call_9_4_14_11, \
call_9_4_14_12, \
call_9_4_14_13, \
call_9_4_14_14, \
call_9_4_14_15, \
call_9_4_15_0, \
call_9_4_15_1, \
call_9_4_15_2, \
call_9_4_15_3, \
call_9_4_15_4, \
call_9_4_15_5, \
call_9_4_15_6, \
call_9_4_15_7, \
call_9_4_15_8, \
call_9_4_15_9, \
call_9_4_15_10, \
call_9_4_15_11, \
call_9_4_15_12, \
call_9_4_15_13, \
call_9_4_15_14, \
call_9_4_15_15, \
call_9_5_0_0, \
call_9_5_0_1, \
call_9_5_0_2, \
call_9_5_0_3, \
call_9_5_0_4, \
call_9_5_0_5, \
call_9_5_0_6, \
call_9_5_0_7, \
call_9_5_0_8, \
call_9_5_0_9, \
call_9_5_0_10, \
call_9_5_0_11, \
call_9_5_0_12, \
call_9_5_0_13, \
call_9_5_0_14, \
call_9_5_0_15, \
call_9_5_1_0, \
call_9_5_1_1, \
call_9_5_1_2, \
call_9_5_1_3, \
call_9_5_1_4, \
call_9_5_1_5, \
call_9_5_1_6, \
call_9_5_1_7, \
call_9_5_1_8, \
call_9_5_1_9, \
call_9_5_1_10, \
call_9_5_1_11, \
call_9_5_1_12, \
call_9_5_1_13, \
call_9_5_1_14, \
call_9_5_1_15, \
call_9_5_2_0, \
call_9_5_2_1, \
call_9_5_2_2, \
call_9_5_2_3, \
call_9_5_2_4, \
call_9_5_2_5, \
call_9_5_2_6, \
call_9_5_2_7, \
call_9_5_2_8, \
call_9_5_2_9, \
call_9_5_2_10, \
call_9_5_2_11, \
call_9_5_2_12, \
call_9_5_2_13, \
call_9_5_2_14, \
call_9_5_2_15, \
call_9_5_3_0, \
call_9_5_3_1, \
call_9_5_3_2, \
call_9_5_3_3, \
call_9_5_3_4, \
call_9_5_3_5, \
call_9_5_3_6, \
call_9_5_3_7, \
call_9_5_3_8, \
call_9_5_3_9, \
call_9_5_3_10, \
call_9_5_3_11, \
call_9_5_3_12, \
call_9_5_3_13, \
call_9_5_3_14, \
call_9_5_3_15, \
call_9_5_4_0, \
call_9_5_4_1, \
call_9_5_4_2, \
call_9_5_4_3, \
call_9_5_4_4, \
call_9_5_4_5, \
call_9_5_4_6, \
call_9_5_4_7, \
call_9_5_4_8, \
call_9_5_4_9, \
call_9_5_4_10, \
call_9_5_4_11, \
call_9_5_4_12, \
call_9_5_4_13, \
call_9_5_4_14, \
call_9_5_4_15, \
call_9_5_5_0, \
call_9_5_5_1, \
call_9_5_5_2, \
call_9_5_5_3, \
call_9_5_5_4, \
call_9_5_5_5, \
call_9_5_5_6, \
call_9_5_5_7, \
call_9_5_5_8, \
call_9_5_5_9, \
call_9_5_5_10, \
call_9_5_5_11, \
call_9_5_5_12, \
call_9_5_5_13, \
call_9_5_5_14, \
call_9_5_5_15, \
call_9_5_6_0, \
call_9_5_6_1, \
call_9_5_6_2, \
call_9_5_6_3, \
call_9_5_6_4, \
call_9_5_6_5, \
call_9_5_6_6, \
call_9_5_6_7, \
call_9_5_6_8, \
call_9_5_6_9, \
call_9_5_6_10, \
call_9_5_6_11, \
call_9_5_6_12, \
call_9_5_6_13, \
call_9_5_6_14, \
call_9_5_6_15, \
call_9_5_7_0, \
call_9_5_7_1, \
call_9_5_7_2, \
call_9_5_7_3, \
call_9_5_7_4, \
call_9_5_7_5, \
call_9_5_7_6, \
call_9_5_7_7, \
call_9_5_7_8, \
call_9_5_7_9, \
call_9_5_7_10, \
call_9_5_7_11, \
call_9_5_7_12, \
call_9_5_7_13, \
call_9_5_7_14, \
call_9_5_7_15, \
call_9_5_8_0, \
call_9_5_8_1, \
call_9_5_8_2, \
call_9_5_8_3, \
call_9_5_8_4, \
call_9_5_8_5, \
call_9_5_8_6, \
call_9_5_8_7, \
call_9_5_8_8, \
call_9_5_8_9, \
call_9_5_8_10, \
call_9_5_8_11, \
call_9_5_8_12, \
call_9_5_8_13, \
call_9_5_8_14, \
call_9_5_8_15, \
call_9_5_9_0, \
call_9_5_9_1, \
call_9_5_9_2, \
call_9_5_9_3, \
call_9_5_9_4, \
call_9_5_9_5, \
call_9_5_9_6, \
call_9_5_9_7, \
call_9_5_9_8, \
call_9_5_9_9, \
call_9_5_9_10, \
call_9_5_9_11, \
call_9_5_9_12, \
call_9_5_9_13, \
call_9_5_9_14, \
call_9_5_9_15, \
call_9_5_10_0, \
call_9_5_10_1, \
call_9_5_10_2, \
call_9_5_10_3, \
call_9_5_10_4, \
call_9_5_10_5, \
call_9_5_10_6, \
call_9_5_10_7, \
call_9_5_10_8, \
call_9_5_10_9, \
call_9_5_10_10, \
call_9_5_10_11, \
call_9_5_10_12, \
call_9_5_10_13, \
call_9_5_10_14, \
call_9_5_10_15, \
call_9_5_11_0, \
call_9_5_11_1, \
call_9_5_11_2, \
call_9_5_11_3, \
call_9_5_11_4, \
call_9_5_11_5, \
call_9_5_11_6, \
call_9_5_11_7, \
call_9_5_11_8, \
call_9_5_11_9, \
call_9_5_11_10, \
call_9_5_11_11, \
call_9_5_11_12, \
call_9_5_11_13, \
call_9_5_11_14, \
call_9_5_11_15, \
call_9_5_12_0, \
call_9_5_12_1, \
call_9_5_12_2, \
call_9_5_12_3, \
call_9_5_12_4, \
call_9_5_12_5, \
call_9_5_12_6, \
call_9_5_12_7, \
call_9_5_12_8, \
call_9_5_12_9, \
call_9_5_12_10, \
call_9_5_12_11, \
call_9_5_12_12, \
call_9_5_12_13, \
call_9_5_12_14, \
call_9_5_12_15, \
call_9_5_13_0, \
call_9_5_13_1, \
call_9_5_13_2, \
call_9_5_13_3, \
call_9_5_13_4, \
call_9_5_13_5, \
call_9_5_13_6, \
call_9_5_13_7, \
call_9_5_13_8, \
call_9_5_13_9, \
call_9_5_13_10, \
call_9_5_13_11, \
call_9_5_13_12, \
call_9_5_13_13, \
call_9_5_13_14, \
call_9_5_13_15, \
call_9_5_14_0, \
call_9_5_14_1, \
call_9_5_14_2, \
call_9_5_14_3, \
call_9_5_14_4, \
call_9_5_14_5, \
call_9_5_14_6, \
call_9_5_14_7, \
call_9_5_14_8, \
call_9_5_14_9, \
call_9_5_14_10, \
call_9_5_14_11, \
call_9_5_14_12, \
call_9_5_14_13, \
call_9_5_14_14, \
call_9_5_14_15, \
call_9_5_15_0, \
call_9_5_15_1, \
call_9_5_15_2, \
call_9_5_15_3, \
call_9_5_15_4, \
call_9_5_15_5, \
call_9_5_15_6, \
call_9_5_15_7, \
call_9_5_15_8, \
call_9_5_15_9, \
call_9_5_15_10, \
call_9_5_15_11, \
call_9_5_15_12, \
call_9_5_15_13, \
call_9_5_15_14, \
call_9_5_15_15, \
call_9_6_0_0, \
call_9_6_0_1, \
call_9_6_0_2, \
call_9_6_0_3, \
call_9_6_0_4, \
call_9_6_0_5, \
call_9_6_0_6, \
call_9_6_0_7, \
call_9_6_0_8, \
call_9_6_0_9, \
call_9_6_0_10, \
call_9_6_0_11, \
call_9_6_0_12, \
call_9_6_0_13, \
call_9_6_0_14, \
call_9_6_0_15, \
call_9_6_1_0, \
call_9_6_1_1, \
call_9_6_1_2, \
call_9_6_1_3, \
call_9_6_1_4, \
call_9_6_1_5, \
call_9_6_1_6, \
call_9_6_1_7, \
call_9_6_1_8, \
call_9_6_1_9, \
call_9_6_1_10, \
call_9_6_1_11, \
call_9_6_1_12, \
call_9_6_1_13, \
call_9_6_1_14, \
call_9_6_1_15, \
call_9_6_2_0, \
call_9_6_2_1, \
call_9_6_2_2, \
call_9_6_2_3, \
call_9_6_2_4, \
call_9_6_2_5, \
call_9_6_2_6, \
call_9_6_2_7, \
call_9_6_2_8, \
call_9_6_2_9, \
call_9_6_2_10, \
call_9_6_2_11, \
call_9_6_2_12, \
call_9_6_2_13, \
call_9_6_2_14, \
call_9_6_2_15, \
call_9_6_3_0, \
call_9_6_3_1, \
call_9_6_3_2, \
call_9_6_3_3, \
call_9_6_3_4, \
call_9_6_3_5, \
call_9_6_3_6, \
call_9_6_3_7, \
call_9_6_3_8, \
call_9_6_3_9, \
call_9_6_3_10, \
call_9_6_3_11, \
call_9_6_3_12, \
call_9_6_3_13, \
call_9_6_3_14, \
call_9_6_3_15, \
call_9_6_4_0, \
call_9_6_4_1, \
call_9_6_4_2, \
call_9_6_4_3, \
call_9_6_4_4, \
call_9_6_4_5, \
call_9_6_4_6, \
call_9_6_4_7, \
call_9_6_4_8, \
call_9_6_4_9, \
call_9_6_4_10, \
call_9_6_4_11, \
call_9_6_4_12, \
call_9_6_4_13, \
call_9_6_4_14, \
call_9_6_4_15, \
call_9_6_5_0, \
call_9_6_5_1, \
call_9_6_5_2, \
call_9_6_5_3, \
call_9_6_5_4, \
call_9_6_5_5, \
call_9_6_5_6, \
call_9_6_5_7, \
call_9_6_5_8, \
call_9_6_5_9, \
call_9_6_5_10, \
call_9_6_5_11, \
call_9_6_5_12, \
call_9_6_5_13, \
call_9_6_5_14, \
call_9_6_5_15, \
call_9_6_6_0, \
call_9_6_6_1, \
call_9_6_6_2, \
call_9_6_6_3, \
call_9_6_6_4, \
call_9_6_6_5, \
call_9_6_6_6, \
call_9_6_6_7, \
call_9_6_6_8, \
call_9_6_6_9, \
call_9_6_6_10, \
call_9_6_6_11, \
call_9_6_6_12, \
call_9_6_6_13, \
call_9_6_6_14, \
call_9_6_6_15, \
call_9_6_7_0, \
call_9_6_7_1, \
call_9_6_7_2, \
call_9_6_7_3, \
call_9_6_7_4, \
call_9_6_7_5, \
call_9_6_7_6, \
call_9_6_7_7, \
call_9_6_7_8, \
call_9_6_7_9, \
call_9_6_7_10, \
call_9_6_7_11, \
call_9_6_7_12, \
call_9_6_7_13, \
call_9_6_7_14, \
call_9_6_7_15, \
call_9_6_8_0, \
call_9_6_8_1, \
call_9_6_8_2, \
call_9_6_8_3, \
call_9_6_8_4, \
call_9_6_8_5, \
call_9_6_8_6, \
call_9_6_8_7, \
call_9_6_8_8, \
call_9_6_8_9, \
call_9_6_8_10, \
call_9_6_8_11, \
call_9_6_8_12, \
call_9_6_8_13, \
call_9_6_8_14, \
call_9_6_8_15, \
call_9_6_9_0, \
call_9_6_9_1, \
call_9_6_9_2, \
call_9_6_9_3, \
call_9_6_9_4, \
call_9_6_9_5, \
call_9_6_9_6, \
call_9_6_9_7, \
call_9_6_9_8, \
call_9_6_9_9, \
call_9_6_9_10, \
call_9_6_9_11, \
call_9_6_9_12, \
call_9_6_9_13, \
call_9_6_9_14, \
call_9_6_9_15, \
call_9_6_10_0, \
call_9_6_10_1, \
call_9_6_10_2, \
call_9_6_10_3, \
call_9_6_10_4, \
call_9_6_10_5, \
call_9_6_10_6, \
call_9_6_10_7, \
call_9_6_10_8, \
call_9_6_10_9, \
call_9_6_10_10, \
call_9_6_10_11, \
call_9_6_10_12, \
call_9_6_10_13, \
call_9_6_10_14, \
call_9_6_10_15, \
call_9_6_11_0, \
call_9_6_11_1, \
call_9_6_11_2, \
call_9_6_11_3, \
call_9_6_11_4, \
call_9_6_11_5, \
call_9_6_11_6, \
call_9_6_11_7, \
call_9_6_11_8, \
call_9_6_11_9, \
call_9_6_11_10, \
call_9_6_11_11, \
call_9_6_11_12, \
call_9_6_11_13, \
call_9_6_11_14, \
call_9_6_11_15, \
call_9_6_12_0, \
call_9_6_12_1, \
call_9_6_12_2, \
call_9_6_12_3, \
call_9_6_12_4, \
call_9_6_12_5, \
call_9_6_12_6, \
call_9_6_12_7, \
call_9_6_12_8, \
call_9_6_12_9, \
call_9_6_12_10, \
call_9_6_12_11, \
call_9_6_12_12, \
call_9_6_12_13, \
call_9_6_12_14, \
call_9_6_12_15, \
call_9_6_13_0, \
call_9_6_13_1, \
call_9_6_13_2, \
call_9_6_13_3, \
call_9_6_13_4, \
call_9_6_13_5, \
call_9_6_13_6, \
call_9_6_13_7, \
call_9_6_13_8, \
call_9_6_13_9, \
call_9_6_13_10, \
call_9_6_13_11, \
call_9_6_13_12, \
call_9_6_13_13, \
call_9_6_13_14, \
call_9_6_13_15, \
call_9_6_14_0, \
call_9_6_14_1, \
call_9_6_14_2, \
call_9_6_14_3, \
call_9_6_14_4, \
call_9_6_14_5, \
call_9_6_14_6, \
call_9_6_14_7, \
call_9_6_14_8, \
call_9_6_14_9, \
call_9_6_14_10, \
call_9_6_14_11, \
call_9_6_14_12, \
call_9_6_14_13, \
call_9_6_14_14, \
call_9_6_14_15, \
call_9_6_15_0, \
call_9_6_15_1, \
call_9_6_15_2, \
call_9_6_15_3, \
call_9_6_15_4, \
call_9_6_15_5, \
call_9_6_15_6, \
call_9_6_15_7, \
call_9_6_15_8, \
call_9_6_15_9, \
call_9_6_15_10, \
call_9_6_15_11, \
call_9_6_15_12, \
call_9_6_15_13, \
call_9_6_15_14, \
call_9_6_15_15, \
call_9_7_0_0, \
call_9_7_0_1, \
call_9_7_0_2, \
call_9_7_0_3, \
call_9_7_0_4, \
call_9_7_0_5, \
call_9_7_0_6, \
call_9_7_0_7, \
call_9_7_0_8, \
call_9_7_0_9, \
call_9_7_0_10, \
call_9_7_0_11, \
call_9_7_0_12, \
call_9_7_0_13, \
call_9_7_0_14, \
call_9_7_0_15, \
call_9_7_1_0, \
call_9_7_1_1, \
call_9_7_1_2, \
call_9_7_1_3, \
call_9_7_1_4, \
call_9_7_1_5, \
call_9_7_1_6, \
call_9_7_1_7, \
call_9_7_1_8, \
call_9_7_1_9, \
call_9_7_1_10, \
call_9_7_1_11, \
call_9_7_1_12, \
call_9_7_1_13, \
call_9_7_1_14, \
call_9_7_1_15, \
call_9_7_2_0, \
call_9_7_2_1, \
call_9_7_2_2, \
call_9_7_2_3, \
call_9_7_2_4, \
call_9_7_2_5, \
call_9_7_2_6, \
call_9_7_2_7, \
call_9_7_2_8, \
call_9_7_2_9, \
call_9_7_2_10, \
call_9_7_2_11, \
call_9_7_2_12, \
call_9_7_2_13, \
call_9_7_2_14, \
call_9_7_2_15, \
call_9_7_3_0, \
call_9_7_3_1, \
call_9_7_3_2, \
call_9_7_3_3, \
call_9_7_3_4, \
call_9_7_3_5, \
call_9_7_3_6, \
call_9_7_3_7, \
call_9_7_3_8, \
call_9_7_3_9, \
call_9_7_3_10, \
call_9_7_3_11, \
call_9_7_3_12, \
call_9_7_3_13, \
call_9_7_3_14, \
call_9_7_3_15, \
call_9_7_4_0, \
call_9_7_4_1, \
call_9_7_4_2, \
call_9_7_4_3, \
call_9_7_4_4, \
call_9_7_4_5, \
call_9_7_4_6, \
call_9_7_4_7, \
call_9_7_4_8, \
call_9_7_4_9, \
call_9_7_4_10, \
call_9_7_4_11, \
call_9_7_4_12, \
call_9_7_4_13, \
call_9_7_4_14, \
call_9_7_4_15, \
call_9_7_5_0, \
call_9_7_5_1, \
call_9_7_5_2, \
call_9_7_5_3, \
call_9_7_5_4, \
call_9_7_5_5, \
call_9_7_5_6, \
call_9_7_5_7, \
call_9_7_5_8, \
call_9_7_5_9, \
call_9_7_5_10, \
call_9_7_5_11, \
call_9_7_5_12, \
call_9_7_5_13, \
call_9_7_5_14, \
call_9_7_5_15, \
call_9_7_6_0, \
call_9_7_6_1, \
call_9_7_6_2, \
call_9_7_6_3, \
call_9_7_6_4, \
call_9_7_6_5, \
call_9_7_6_6, \
call_9_7_6_7, \
call_9_7_6_8, \
call_9_7_6_9, \
call_9_7_6_10, \
call_9_7_6_11, \
call_9_7_6_12, \
call_9_7_6_13, \
call_9_7_6_14, \
call_9_7_6_15, \
call_9_7_7_0, \
call_9_7_7_1, \
call_9_7_7_2, \
call_9_7_7_3, \
call_9_7_7_4, \
call_9_7_7_5, \
call_9_7_7_6, \
call_9_7_7_7, \
call_9_7_7_8, \
call_9_7_7_9, \
call_9_7_7_10, \
call_9_7_7_11, \
call_9_7_7_12, \
call_9_7_7_13, \
call_9_7_7_14, \
call_9_7_7_15, \
call_9_7_8_0, \
call_9_7_8_1, \
call_9_7_8_2, \
call_9_7_8_3, \
call_9_7_8_4, \
call_9_7_8_5, \
call_9_7_8_6, \
call_9_7_8_7, \
call_9_7_8_8, \
call_9_7_8_9, \
call_9_7_8_10, \
call_9_7_8_11, \
call_9_7_8_12, \
call_9_7_8_13, \
call_9_7_8_14, \
call_9_7_8_15, \
call_9_7_9_0, \
call_9_7_9_1, \
call_9_7_9_2, \
call_9_7_9_3, \
call_9_7_9_4, \
call_9_7_9_5, \
call_9_7_9_6, \
call_9_7_9_7, \
call_9_7_9_8, \
call_9_7_9_9, \
call_9_7_9_10, \
call_9_7_9_11, \
call_9_7_9_12, \
call_9_7_9_13, \
call_9_7_9_14, \
call_9_7_9_15, \
call_9_7_10_0, \
call_9_7_10_1, \
call_9_7_10_2, \
call_9_7_10_3, \
call_9_7_10_4, \
call_9_7_10_5, \
call_9_7_10_6, \
call_9_7_10_7, \
call_9_7_10_8, \
call_9_7_10_9, \
call_9_7_10_10, \
call_9_7_10_11, \
call_9_7_10_12, \
call_9_7_10_13, \
call_9_7_10_14, \
call_9_7_10_15, \
call_9_7_11_0, \
call_9_7_11_1, \
call_9_7_11_2, \
call_9_7_11_3, \
call_9_7_11_4, \
call_9_7_11_5, \
call_9_7_11_6, \
call_9_7_11_7, \
call_9_7_11_8, \
call_9_7_11_9, \
call_9_7_11_10, \
call_9_7_11_11, \
call_9_7_11_12, \
call_9_7_11_13, \
call_9_7_11_14, \
call_9_7_11_15, \
call_9_7_12_0, \
call_9_7_12_1, \
call_9_7_12_2, \
call_9_7_12_3, \
call_9_7_12_4, \
call_9_7_12_5, \
call_9_7_12_6, \
call_9_7_12_7, \
call_9_7_12_8, \
call_9_7_12_9, \
call_9_7_12_10, \
call_9_7_12_11, \
call_9_7_12_12, \
call_9_7_12_13, \
call_9_7_12_14, \
call_9_7_12_15, \
call_9_7_13_0, \
call_9_7_13_1, \
call_9_7_13_2, \
call_9_7_13_3, \
call_9_7_13_4, \
call_9_7_13_5, \
call_9_7_13_6, \
call_9_7_13_7, \
call_9_7_13_8, \
call_9_7_13_9, \
call_9_7_13_10, \
call_9_7_13_11, \
call_9_7_13_12, \
call_9_7_13_13, \
call_9_7_13_14, \
call_9_7_13_15, \
call_9_7_14_0, \
call_9_7_14_1, \
call_9_7_14_2, \
call_9_7_14_3, \
call_9_7_14_4, \
call_9_7_14_5, \
call_9_7_14_6, \
call_9_7_14_7, \
call_9_7_14_8, \
call_9_7_14_9, \
call_9_7_14_10, \
call_9_7_14_11, \
call_9_7_14_12, \
call_9_7_14_13, \
call_9_7_14_14, \
call_9_7_14_15, \
call_9_7_15_0, \
call_9_7_15_1, \
call_9_7_15_2, \
call_9_7_15_3, \
call_9_7_15_4, \
call_9_7_15_5, \
call_9_7_15_6, \
call_9_7_15_7, \
call_9_7_15_8, \
call_9_7_15_9, \
call_9_7_15_10, \
call_9_7_15_11, \
call_9_7_15_12, \
call_9_7_15_13, \
call_9_7_15_14, \
call_9_7_15_15, \
call_9_8_0_0, \
call_9_8_0_1, \
call_9_8_0_2, \
call_9_8_0_3, \
call_9_8_0_4, \
call_9_8_0_5, \
call_9_8_0_6, \
call_9_8_0_7, \
call_9_8_0_8, \
call_9_8_0_9, \
call_9_8_0_10, \
call_9_8_0_11, \
call_9_8_0_12, \
call_9_8_0_13, \
call_9_8_0_14, \
call_9_8_0_15, \
call_9_8_1_0, \
call_9_8_1_1, \
call_9_8_1_2, \
call_9_8_1_3, \
call_9_8_1_4, \
call_9_8_1_5, \
call_9_8_1_6, \
call_9_8_1_7, \
call_9_8_1_8, \
call_9_8_1_9, \
call_9_8_1_10, \
call_9_8_1_11, \
call_9_8_1_12, \
call_9_8_1_13, \
call_9_8_1_14, \
call_9_8_1_15, \
call_9_8_2_0, \
call_9_8_2_1, \
call_9_8_2_2, \
call_9_8_2_3, \
call_9_8_2_4, \
call_9_8_2_5, \
call_9_8_2_6, \
call_9_8_2_7, \
call_9_8_2_8, \
call_9_8_2_9, \
call_9_8_2_10, \
call_9_8_2_11, \
call_9_8_2_12, \
call_9_8_2_13, \
call_9_8_2_14, \
call_9_8_2_15, \
call_9_8_3_0, \
call_9_8_3_1, \
call_9_8_3_2, \
call_9_8_3_3, \
call_9_8_3_4, \
call_9_8_3_5, \
call_9_8_3_6, \
call_9_8_3_7, \
call_9_8_3_8, \
call_9_8_3_9, \
call_9_8_3_10, \
call_9_8_3_11, \
call_9_8_3_12, \
call_9_8_3_13, \
call_9_8_3_14, \
call_9_8_3_15, \
call_9_8_4_0, \
call_9_8_4_1, \
call_9_8_4_2, \
call_9_8_4_3, \
call_9_8_4_4, \
call_9_8_4_5, \
call_9_8_4_6, \
call_9_8_4_7, \
call_9_8_4_8, \
call_9_8_4_9, \
call_9_8_4_10, \
call_9_8_4_11, \
call_9_8_4_12, \
call_9_8_4_13, \
call_9_8_4_14, \
call_9_8_4_15, \
call_9_8_5_0, \
call_9_8_5_1, \
call_9_8_5_2, \
call_9_8_5_3, \
call_9_8_5_4, \
call_9_8_5_5, \
call_9_8_5_6, \
call_9_8_5_7, \
call_9_8_5_8, \
call_9_8_5_9, \
call_9_8_5_10, \
call_9_8_5_11, \
call_9_8_5_12, \
call_9_8_5_13, \
call_9_8_5_14, \
call_9_8_5_15, \
call_9_8_6_0, \
call_9_8_6_1, \
call_9_8_6_2, \
call_9_8_6_3, \
call_9_8_6_4, \
call_9_8_6_5, \
call_9_8_6_6, \
call_9_8_6_7, \
call_9_8_6_8, \
call_9_8_6_9, \
call_9_8_6_10, \
call_9_8_6_11, \
call_9_8_6_12, \
call_9_8_6_13, \
call_9_8_6_14, \
call_9_8_6_15, \
call_9_8_7_0, \
call_9_8_7_1, \
call_9_8_7_2, \
call_9_8_7_3, \
call_9_8_7_4, \
call_9_8_7_5, \
call_9_8_7_6, \
call_9_8_7_7, \
call_9_8_7_8, \
call_9_8_7_9, \
call_9_8_7_10, \
call_9_8_7_11, \
call_9_8_7_12, \
call_9_8_7_13, \
call_9_8_7_14, \
call_9_8_7_15, \
call_9_8_8_0, \
call_9_8_8_1, \
call_9_8_8_2, \
call_9_8_8_3, \
call_9_8_8_4, \
call_9_8_8_5, \
call_9_8_8_6, \
call_9_8_8_7, \
call_9_8_8_8, \
call_9_8_8_9, \
call_9_8_8_10, \
call_9_8_8_11, \
call_9_8_8_12, \
call_9_8_8_13, \
call_9_8_8_14, \
call_9_8_8_15, \
call_9_8_9_0, \
call_9_8_9_1, \
call_9_8_9_2, \
call_9_8_9_3, \
call_9_8_9_4, \
call_9_8_9_5, \
call_9_8_9_6, \
call_9_8_9_7, \
call_9_8_9_8, \
call_9_8_9_9, \
call_9_8_9_10, \
call_9_8_9_11, \
call_9_8_9_12, \
call_9_8_9_13, \
call_9_8_9_14, \
call_9_8_9_15, \
call_9_8_10_0, \
call_9_8_10_1, \
call_9_8_10_2, \
call_9_8_10_3, \
call_9_8_10_4, \
call_9_8_10_5, \
call_9_8_10_6, \
call_9_8_10_7, \
call_9_8_10_8, \
call_9_8_10_9, \
call_9_8_10_10, \
call_9_8_10_11, \
call_9_8_10_12, \
call_9_8_10_13, \
call_9_8_10_14, \
call_9_8_10_15, \
call_9_8_11_0, \
call_9_8_11_1, \
call_9_8_11_2, \
call_9_8_11_3, \
call_9_8_11_4, \
call_9_8_11_5, \
call_9_8_11_6, \
call_9_8_11_7, \
call_9_8_11_8, \
call_9_8_11_9, \
call_9_8_11_10, \
call_9_8_11_11, \
call_9_8_11_12, \
call_9_8_11_13, \
call_9_8_11_14, \
call_9_8_11_15, \
call_9_8_12_0, \
call_9_8_12_1, \
call_9_8_12_2, \
call_9_8_12_3, \
call_9_8_12_4, \
call_9_8_12_5, \
call_9_8_12_6, \
call_9_8_12_7, \
call_9_8_12_8, \
call_9_8_12_9, \
call_9_8_12_10, \
call_9_8_12_11, \
call_9_8_12_12, \
call_9_8_12_13, \
call_9_8_12_14, \
call_9_8_12_15, \
call_9_8_13_0, \
call_9_8_13_1, \
call_9_8_13_2, \
call_9_8_13_3, \
call_9_8_13_4, \
call_9_8_13_5, \
call_9_8_13_6, \
call_9_8_13_7, \
call_9_8_13_8, \
call_9_8_13_9, \
call_9_8_13_10, \
call_9_8_13_11, \
call_9_8_13_12, \
call_9_8_13_13, \
call_9_8_13_14, \
call_9_8_13_15, \
call_9_8_14_0, \
call_9_8_14_1, \
call_9_8_14_2, \
call_9_8_14_3, \
call_9_8_14_4, \
call_9_8_14_5, \
call_9_8_14_6, \
call_9_8_14_7, \
call_9_8_14_8, \
call_9_8_14_9, \
call_9_8_14_10, \
call_9_8_14_11, \
call_9_8_14_12, \
call_9_8_14_13, \
call_9_8_14_14, \
call_9_8_14_15, \
call_9_8_15_0, \
call_9_8_15_1, \
call_9_8_15_2, \
call_9_8_15_3, \
call_9_8_15_4, \
call_9_8_15_5, \
call_9_8_15_6, \
call_9_8_15_7, \
call_9_8_15_8, \
call_9_8_15_9, \
call_9_8_15_10, \
call_9_8_15_11, \
call_9_8_15_12, \
call_9_8_15_13, \
call_9_8_15_14, \
call_9_8_15_15, \
call_9_9_0_0, \
call_9_9_0_1, \
call_9_9_0_2, \
call_9_9_0_3, \
call_9_9_0_4, \
call_9_9_0_5, \
call_9_9_0_6, \
call_9_9_0_7, \
call_9_9_0_8, \
call_9_9_0_9, \
call_9_9_0_10, \
call_9_9_0_11, \
call_9_9_0_12, \
call_9_9_0_13, \
call_9_9_0_14, \
call_9_9_0_15, \
call_9_9_1_0, \
call_9_9_1_1, \
call_9_9_1_2, \
call_9_9_1_3, \
call_9_9_1_4, \
call_9_9_1_5, \
call_9_9_1_6, \
call_9_9_1_7, \
call_9_9_1_8, \
call_9_9_1_9, \
call_9_9_1_10, \
call_9_9_1_11, \
call_9_9_1_12, \
call_9_9_1_13, \
call_9_9_1_14, \
call_9_9_1_15, \
call_9_9_2_0, \
call_9_9_2_1, \
call_9_9_2_2, \
call_9_9_2_3, \
call_9_9_2_4, \
call_9_9_2_5, \
call_9_9_2_6, \
call_9_9_2_7, \
call_9_9_2_8, \
call_9_9_2_9, \
call_9_9_2_10, \
call_9_9_2_11, \
call_9_9_2_12, \
call_9_9_2_13, \
call_9_9_2_14, \
call_9_9_2_15, \
call_9_9_3_0, \
call_9_9_3_1, \
call_9_9_3_2, \
call_9_9_3_3, \
call_9_9_3_4, \
call_9_9_3_5, \
call_9_9_3_6, \
call_9_9_3_7, \
call_9_9_3_8, \
call_9_9_3_9, \
call_9_9_3_10, \
call_9_9_3_11, \
call_9_9_3_12, \
call_9_9_3_13, \
call_9_9_3_14, \
call_9_9_3_15, \
call_9_9_4_0, \
call_9_9_4_1, \
call_9_9_4_2, \
call_9_9_4_3, \
call_9_9_4_4, \
call_9_9_4_5, \
call_9_9_4_6, \
call_9_9_4_7, \
call_9_9_4_8, \
call_9_9_4_9, \
call_9_9_4_10, \
call_9_9_4_11, \
call_9_9_4_12, \
call_9_9_4_13, \
call_9_9_4_14, \
call_9_9_4_15, \
call_9_9_5_0, \
call_9_9_5_1, \
call_9_9_5_2, \
call_9_9_5_3, \
call_9_9_5_4, \
call_9_9_5_5, \
call_9_9_5_6, \
call_9_9_5_7, \
call_9_9_5_8, \
call_9_9_5_9, \
call_9_9_5_10, \
call_9_9_5_11, \
call_9_9_5_12, \
call_9_9_5_13, \
call_9_9_5_14, \
call_9_9_5_15, \
call_9_9_6_0, \
call_9_9_6_1, \
call_9_9_6_2, \
call_9_9_6_3, \
call_9_9_6_4, \
call_9_9_6_5, \
call_9_9_6_6, \
call_9_9_6_7, \
call_9_9_6_8, \
call_9_9_6_9, \
call_9_9_6_10, \
call_9_9_6_11, \
call_9_9_6_12, \
call_9_9_6_13, \
call_9_9_6_14, \
call_9_9_6_15, \
call_9_9_7_0, \
call_9_9_7_1, \
call_9_9_7_2, \
call_9_9_7_3, \
call_9_9_7_4, \
call_9_9_7_5, \
call_9_9_7_6, \
call_9_9_7_7, \
call_9_9_7_8, \
call_9_9_7_9, \
call_9_9_7_10, \
call_9_9_7_11, \
call_9_9_7_12, \
call_9_9_7_13, \
call_9_9_7_14, \
call_9_9_7_15, \
call_9_9_8_0, \
call_9_9_8_1, \
call_9_9_8_2, \
call_9_9_8_3, \
call_9_9_8_4, \
call_9_9_8_5, \
call_9_9_8_6, \
call_9_9_8_7, \
call_9_9_8_8, \
call_9_9_8_9, \
call_9_9_8_10, \
call_9_9_8_11, \
call_9_9_8_12, \
call_9_9_8_13, \
call_9_9_8_14, \
call_9_9_8_15, \
call_9_9_9_0, \
call_9_9_9_1, \
call_9_9_9_2, \
call_9_9_9_3, \
call_9_9_9_4, \
call_9_9_9_5, \
call_9_9_9_6, \
call_9_9_9_7, \
call_9_9_9_8, \
call_9_9_9_9, \
call_9_9_9_10, \
call_9_9_9_11, \
call_9_9_9_12, \
call_9_9_9_13, \
call_9_9_9_14, \
call_9_9_9_15, \
call_9_9_10_0, \
call_9_9_10_1, \
call_9_9_10_2, \
call_9_9_10_3, \
call_9_9_10_4, \
call_9_9_10_5, \
call_9_9_10_6, \
call_9_9_10_7, \
call_9_9_10_8, \
call_9_9_10_9, \
call_9_9_10_10, \
call_9_9_10_11, \
call_9_9_10_12, \
call_9_9_10_13, \
call_9_9_10_14, \
call_9_9_10_15, \
call_9_9_11_0, \
call_9_9_11_1, \
call_9_9_11_2, \
call_9_9_11_3, \
call_9_9_11_4, \
call_9_9_11_5, \
call_9_9_11_6, \
call_9_9_11_7, \
call_9_9_11_8, \
call_9_9_11_9, \
call_9_9_11_10, \
call_9_9_11_11, \
call_9_9_11_12, \
call_9_9_11_13, \
call_9_9_11_14, \
call_9_9_11_15, \
call_9_9_12_0, \
call_9_9_12_1, \
call_9_9_12_2, \
call_9_9_12_3, \
call_9_9_12_4, \
call_9_9_12_5, \
call_9_9_12_6, \
call_9_9_12_7, \
call_9_9_12_8, \
call_9_9_12_9, \
call_9_9_12_10, \
call_9_9_12_11, \
call_9_9_12_12, \
call_9_9_12_13, \
call_9_9_12_14, \
call_9_9_12_15, \
call_9_9_13_0, \
call_9_9_13_1, \
call_9_9_13_2, \
call_9_9_13_3, \
call_9_9_13_4, \
call_9_9_13_5, \
call_9_9_13_6, \
call_9_9_13_7, \
call_9_9_13_8, \
call_9_9_13_9, \
call_9_9_13_10, \
call_9_9_13_11, \
call_9_9_13_12, \
call_9_9_13_13, \
call_9_9_13_14, \
call_9_9_13_15, \
call_9_9_14_0, \
call_9_9_14_1, \
call_9_9_14_2, \
call_9_9_14_3, \
call_9_9_14_4, \
call_9_9_14_5, \
call_9_9_14_6, \
call_9_9_14_7, \
call_9_9_14_8, \
call_9_9_14_9, \
call_9_9_14_10, \
call_9_9_14_11, \
call_9_9_14_12, \
call_9_9_14_13, \
call_9_9_14_14, \
call_9_9_14_15, \
call_9_9_15_0, \
call_9_9_15_1, \
call_9_9_15_2, \
call_9_9_15_3, \
call_9_9_15_4, \
call_9_9_15_5, \
call_9_9_15_6, \
call_9_9_15_7, \
call_9_9_15_8, \
call_9_9_15_9, \
call_9_9_15_10, \
call_9_9_15_11, \
call_9_9_15_12, \
call_9_9_15_13, \
call_9_9_15_14, \
call_9_9_15_15, \
call_9_10_0_0, \
call_9_10_0_1, \
call_9_10_0_2, \
call_9_10_0_3, \
call_9_10_0_4, \
call_9_10_0_5, \
call_9_10_0_6, \
call_9_10_0_7, \
call_9_10_0_8, \
call_9_10_0_9, \
call_9_10_0_10, \
call_9_10_0_11, \
call_9_10_0_12, \
call_9_10_0_13, \
call_9_10_0_14, \
call_9_10_0_15, \
call_9_10_1_0, \
call_9_10_1_1, \
call_9_10_1_2, \
call_9_10_1_3, \
call_9_10_1_4, \
call_9_10_1_5, \
call_9_10_1_6, \
call_9_10_1_7, \
call_9_10_1_8, \
call_9_10_1_9, \
call_9_10_1_10, \
call_9_10_1_11, \
call_9_10_1_12, \
call_9_10_1_13, \
call_9_10_1_14, \
call_9_10_1_15, \
call_9_10_2_0, \
call_9_10_2_1, \
call_9_10_2_2, \
call_9_10_2_3, \
call_9_10_2_4, \
call_9_10_2_5, \
call_9_10_2_6, \
call_9_10_2_7, \
call_9_10_2_8, \
call_9_10_2_9, \
call_9_10_2_10, \
call_9_10_2_11, \
call_9_10_2_12, \
call_9_10_2_13, \
call_9_10_2_14, \
call_9_10_2_15, \
call_9_10_3_0, \
call_9_10_3_1, \
call_9_10_3_2, \
call_9_10_3_3, \
call_9_10_3_4, \
call_9_10_3_5, \
call_9_10_3_6, \
call_9_10_3_7, \
call_9_10_3_8, \
call_9_10_3_9, \
call_9_10_3_10, \
call_9_10_3_11, \
call_9_10_3_12, \
call_9_10_3_13, \
call_9_10_3_14, \
call_9_10_3_15, \
call_9_10_4_0, \
call_9_10_4_1, \
call_9_10_4_2, \
call_9_10_4_3, \
call_9_10_4_4, \
call_9_10_4_5, \
call_9_10_4_6, \
call_9_10_4_7, \
call_9_10_4_8, \
call_9_10_4_9, \
call_9_10_4_10, \
call_9_10_4_11, \
call_9_10_4_12, \
call_9_10_4_13, \
call_9_10_4_14, \
call_9_10_4_15, \
call_9_10_5_0, \
call_9_10_5_1, \
call_9_10_5_2, \
call_9_10_5_3, \
call_9_10_5_4, \
call_9_10_5_5, \
call_9_10_5_6, \
call_9_10_5_7, \
call_9_10_5_8, \
call_9_10_5_9, \
call_9_10_5_10, \
call_9_10_5_11, \
call_9_10_5_12, \
call_9_10_5_13, \
call_9_10_5_14, \
call_9_10_5_15, \
call_9_10_6_0, \
call_9_10_6_1, \
call_9_10_6_2, \
call_9_10_6_3, \
call_9_10_6_4, \
call_9_10_6_5, \
call_9_10_6_6, \
call_9_10_6_7, \
call_9_10_6_8, \
call_9_10_6_9, \
call_9_10_6_10, \
call_9_10_6_11, \
call_9_10_6_12, \
call_9_10_6_13, \
call_9_10_6_14, \
call_9_10_6_15, \
call_9_10_7_0, \
call_9_10_7_1, \
call_9_10_7_2, \
call_9_10_7_3, \
call_9_10_7_4, \
call_9_10_7_5, \
call_9_10_7_6, \
call_9_10_7_7, \
call_9_10_7_8, \
call_9_10_7_9, \
call_9_10_7_10, \
call_9_10_7_11, \
call_9_10_7_12, \
call_9_10_7_13, \
call_9_10_7_14, \
call_9_10_7_15, \
call_9_10_8_0, \
call_9_10_8_1, \
call_9_10_8_2, \
call_9_10_8_3, \
call_9_10_8_4, \
call_9_10_8_5, \
call_9_10_8_6, \
call_9_10_8_7, \
call_9_10_8_8, \
call_9_10_8_9, \
call_9_10_8_10, \
call_9_10_8_11, \
call_9_10_8_12, \
call_9_10_8_13, \
call_9_10_8_14, \
call_9_10_8_15, \
call_9_10_9_0, \
call_9_10_9_1, \
call_9_10_9_2, \
call_9_10_9_3, \
call_9_10_9_4, \
call_9_10_9_5, \
call_9_10_9_6, \
call_9_10_9_7, \
call_9_10_9_8, \
call_9_10_9_9, \
call_9_10_9_10, \
call_9_10_9_11, \
call_9_10_9_12, \
call_9_10_9_13, \
call_9_10_9_14, \
call_9_10_9_15, \
call_9_10_10_0, \
call_9_10_10_1, \
call_9_10_10_2, \
call_9_10_10_3, \
call_9_10_10_4, \
call_9_10_10_5, \
call_9_10_10_6, \
call_9_10_10_7, \
call_9_10_10_8, \
call_9_10_10_9, \
call_9_10_10_10, \
call_9_10_10_11, \
call_9_10_10_12, \
call_9_10_10_13, \
call_9_10_10_14, \
call_9_10_10_15, \
call_9_10_11_0, \
call_9_10_11_1, \
call_9_10_11_2, \
call_9_10_11_3, \
call_9_10_11_4, \
call_9_10_11_5, \
call_9_10_11_6, \
call_9_10_11_7, \
call_9_10_11_8, \
call_9_10_11_9, \
call_9_10_11_10, \
call_9_10_11_11, \
call_9_10_11_12, \
call_9_10_11_13, \
call_9_10_11_14, \
call_9_10_11_15, \
call_9_10_12_0, \
call_9_10_12_1, \
call_9_10_12_2, \
call_9_10_12_3, \
call_9_10_12_4, \
call_9_10_12_5, \
call_9_10_12_6, \
call_9_10_12_7, \
call_9_10_12_8, \
call_9_10_12_9, \
call_9_10_12_10, \
call_9_10_12_11, \
call_9_10_12_12, \
call_9_10_12_13, \
call_9_10_12_14, \
call_9_10_12_15, \
call_9_10_13_0, \
call_9_10_13_1, \
call_9_10_13_2, \
call_9_10_13_3, \
call_9_10_13_4, \
call_9_10_13_5, \
call_9_10_13_6, \
call_9_10_13_7, \
call_9_10_13_8, \
call_9_10_13_9, \
call_9_10_13_10, \
call_9_10_13_11, \
call_9_10_13_12, \
call_9_10_13_13, \
call_9_10_13_14, \
call_9_10_13_15, \
call_9_10_14_0, \
call_9_10_14_1, \
call_9_10_14_2, \
call_9_10_14_3, \
call_9_10_14_4, \
call_9_10_14_5, \
call_9_10_14_6, \
call_9_10_14_7, \
call_9_10_14_8, \
call_9_10_14_9, \
call_9_10_14_10, \
call_9_10_14_11, \
call_9_10_14_12, \
call_9_10_14_13, \
call_9_10_14_14, \
call_9_10_14_15, \
call_9_10_15_0, \
call_9_10_15_1, \
call_9_10_15_2, \
call_9_10_15_3, \
call_9_10_15_4, \
call_9_10_15_5, \
call_9_10_15_6, \
call_9_10_15_7, \
call_9_10_15_8, \
call_9_10_15_9, \
call_9_10_15_10, \
call_9_10_15_11, \
call_9_10_15_12, \
call_9_10_15_13, \
call_9_10_15_14, \
call_9_10_15_15, \
call_9_11_0_0, \
call_9_11_0_1, \
call_9_11_0_2, \
call_9_11_0_3, \
call_9_11_0_4, \
call_9_11_0_5, \
call_9_11_0_6, \
call_9_11_0_7, \
call_9_11_0_8, \
call_9_11_0_9, \
call_9_11_0_10, \
call_9_11_0_11, \
call_9_11_0_12, \
call_9_11_0_13, \
call_9_11_0_14, \
call_9_11_0_15, \
call_9_11_1_0, \
call_9_11_1_1, \
call_9_11_1_2, \
call_9_11_1_3, \
call_9_11_1_4, \
call_9_11_1_5, \
call_9_11_1_6, \
call_9_11_1_7, \
call_9_11_1_8, \
call_9_11_1_9, \
call_9_11_1_10, \
call_9_11_1_11, \
call_9_11_1_12, \
call_9_11_1_13, \
call_9_11_1_14, \
call_9_11_1_15, \
call_9_11_2_0, \
call_9_11_2_1, \
call_9_11_2_2, \
call_9_11_2_3, \
call_9_11_2_4, \
call_9_11_2_5, \
call_9_11_2_6, \
call_9_11_2_7, \
call_9_11_2_8, \
call_9_11_2_9, \
call_9_11_2_10, \
call_9_11_2_11, \
call_9_11_2_12, \
call_9_11_2_13, \
call_9_11_2_14, \
call_9_11_2_15, \
call_9_11_3_0, \
call_9_11_3_1, \
call_9_11_3_2, \
call_9_11_3_3, \
call_9_11_3_4, \
call_9_11_3_5, \
call_9_11_3_6, \
call_9_11_3_7, \
call_9_11_3_8, \
call_9_11_3_9, \
call_9_11_3_10, \
call_9_11_3_11, \
call_9_11_3_12, \
call_9_11_3_13, \
call_9_11_3_14, \
call_9_11_3_15, \
call_9_11_4_0, \
call_9_11_4_1, \
call_9_11_4_2, \
call_9_11_4_3, \
call_9_11_4_4, \
call_9_11_4_5, \
call_9_11_4_6, \
call_9_11_4_7, \
call_9_11_4_8, \
call_9_11_4_9, \
call_9_11_4_10, \
call_9_11_4_11, \
call_9_11_4_12, \
call_9_11_4_13, \
call_9_11_4_14, \
call_9_11_4_15, \
call_9_11_5_0, \
call_9_11_5_1, \
call_9_11_5_2, \
call_9_11_5_3, \
call_9_11_5_4, \
call_9_11_5_5, \
call_9_11_5_6, \
call_9_11_5_7, \
call_9_11_5_8, \
call_9_11_5_9, \
call_9_11_5_10, \
call_9_11_5_11, \
call_9_11_5_12, \
call_9_11_5_13, \
call_9_11_5_14, \
call_9_11_5_15, \
call_9_11_6_0, \
call_9_11_6_1, \
call_9_11_6_2, \
call_9_11_6_3, \
call_9_11_6_4, \
call_9_11_6_5, \
call_9_11_6_6, \
call_9_11_6_7, \
call_9_11_6_8, \
call_9_11_6_9, \
call_9_11_6_10, \
call_9_11_6_11, \
call_9_11_6_12, \
call_9_11_6_13, \
call_9_11_6_14, \
call_9_11_6_15, \
call_9_11_7_0, \
call_9_11_7_1, \
call_9_11_7_2, \
call_9_11_7_3, \
call_9_11_7_4, \
call_9_11_7_5, \
call_9_11_7_6, \
call_9_11_7_7, \
call_9_11_7_8, \
call_9_11_7_9, \
call_9_11_7_10, \
call_9_11_7_11, \
call_9_11_7_12, \
call_9_11_7_13, \
call_9_11_7_14, \
call_9_11_7_15, \
call_9_11_8_0, \
call_9_11_8_1, \
call_9_11_8_2, \
call_9_11_8_3, \
call_9_11_8_4, \
call_9_11_8_5, \
call_9_11_8_6, \
call_9_11_8_7, \
call_9_11_8_8, \
call_9_11_8_9, \
call_9_11_8_10, \
call_9_11_8_11, \
call_9_11_8_12, \
call_9_11_8_13, \
call_9_11_8_14, \
call_9_11_8_15, \
call_9_11_9_0, \
call_9_11_9_1, \
call_9_11_9_2, \
call_9_11_9_3, \
call_9_11_9_4, \
call_9_11_9_5, \
call_9_11_9_6, \
call_9_11_9_7, \
call_9_11_9_8, \
call_9_11_9_9, \
call_9_11_9_10, \
call_9_11_9_11, \
call_9_11_9_12, \
call_9_11_9_13, \
call_9_11_9_14, \
call_9_11_9_15, \
call_9_11_10_0, \
call_9_11_10_1, \
call_9_11_10_2, \
call_9_11_10_3, \
call_9_11_10_4, \
call_9_11_10_5, \
call_9_11_10_6, \
call_9_11_10_7, \
call_9_11_10_8, \
call_9_11_10_9, \
call_9_11_10_10, \
call_9_11_10_11, \
call_9_11_10_12, \
call_9_11_10_13, \
call_9_11_10_14, \
call_9_11_10_15, \
call_9_11_11_0, \
call_9_11_11_1, \
call_9_11_11_2, \
call_9_11_11_3, \
call_9_11_11_4, \
call_9_11_11_5, \
call_9_11_11_6, \
call_9_11_11_7, \
call_9_11_11_8, \
call_9_11_11_9, \
call_9_11_11_10, \
call_9_11_11_11, \
call_9_11_11_12, \
call_9_11_11_13, \
call_9_11_11_14, \
call_9_11_11_15, \
call_9_11_12_0, \
call_9_11_12_1, \
call_9_11_12_2, \
call_9_11_12_3, \
call_9_11_12_4, \
call_9_11_12_5, \
call_9_11_12_6, \
call_9_11_12_7, \
call_9_11_12_8, \
call_9_11_12_9, \
call_9_11_12_10, \
call_9_11_12_11, \
call_9_11_12_12, \
call_9_11_12_13, \
call_9_11_12_14, \
call_9_11_12_15, \
call_9_11_13_0, \
call_9_11_13_1, \
call_9_11_13_2, \
call_9_11_13_3, \
call_9_11_13_4, \
call_9_11_13_5, \
call_9_11_13_6, \
call_9_11_13_7, \
call_9_11_13_8, \
call_9_11_13_9, \
call_9_11_13_10, \
call_9_11_13_11, \
call_9_11_13_12, \
call_9_11_13_13, \
call_9_11_13_14, \
call_9_11_13_15, \
call_9_11_14_0, \
call_9_11_14_1, \
call_9_11_14_2, \
call_9_11_14_3, \
call_9_11_14_4, \
call_9_11_14_5, \
call_9_11_14_6, \
call_9_11_14_7, \
call_9_11_14_8, \
call_9_11_14_9, \
call_9_11_14_10, \
call_9_11_14_11, \
call_9_11_14_12, \
call_9_11_14_13, \
call_9_11_14_14, \
call_9_11_14_15, \
call_9_11_15_0, \
call_9_11_15_1, \
call_9_11_15_2, \
call_9_11_15_3, \
call_9_11_15_4, \
call_9_11_15_5, \
call_9_11_15_6, \
call_9_11_15_7, \
call_9_11_15_8, \
call_9_11_15_9, \
call_9_11_15_10, \
call_9_11_15_11, \
call_9_11_15_12, \
call_9_11_15_13, \
call_9_11_15_14, \
call_9_11_15_15, \
call_9_12_0_0, \
call_9_12_0_1, \
call_9_12_0_2, \
call_9_12_0_3, \
call_9_12_0_4, \
call_9_12_0_5, \
call_9_12_0_6, \
call_9_12_0_7, \
call_9_12_0_8, \
call_9_12_0_9, \
call_9_12_0_10, \
call_9_12_0_11, \
call_9_12_0_12, \
call_9_12_0_13, \
call_9_12_0_14, \
call_9_12_0_15, \
call_9_12_1_0, \
call_9_12_1_1, \
call_9_12_1_2, \
call_9_12_1_3, \
call_9_12_1_4, \
call_9_12_1_5, \
call_9_12_1_6, \
call_9_12_1_7, \
call_9_12_1_8, \
call_9_12_1_9, \
call_9_12_1_10, \
call_9_12_1_11, \
call_9_12_1_12, \
call_9_12_1_13, \
call_9_12_1_14, \
call_9_12_1_15, \
call_9_12_2_0, \
call_9_12_2_1, \
call_9_12_2_2, \
call_9_12_2_3, \
call_9_12_2_4, \
call_9_12_2_5, \
call_9_12_2_6, \
call_9_12_2_7, \
call_9_12_2_8, \
call_9_12_2_9, \
call_9_12_2_10, \
call_9_12_2_11, \
call_9_12_2_12, \
call_9_12_2_13, \
call_9_12_2_14, \
call_9_12_2_15, \
call_9_12_3_0, \
call_9_12_3_1, \
call_9_12_3_2, \
call_9_12_3_3, \
call_9_12_3_4, \
call_9_12_3_5, \
call_9_12_3_6, \
call_9_12_3_7, \
call_9_12_3_8, \
call_9_12_3_9, \
call_9_12_3_10, \
call_9_12_3_11, \
call_9_12_3_12, \
call_9_12_3_13, \
call_9_12_3_14, \
call_9_12_3_15, \
call_9_12_4_0, \
call_9_12_4_1, \
call_9_12_4_2, \
call_9_12_4_3, \
call_9_12_4_4, \
call_9_12_4_5, \
call_9_12_4_6, \
call_9_12_4_7, \
call_9_12_4_8, \
call_9_12_4_9, \
call_9_12_4_10, \
call_9_12_4_11, \
call_9_12_4_12, \
call_9_12_4_13, \
call_9_12_4_14, \
call_9_12_4_15, \
call_9_12_5_0, \
call_9_12_5_1, \
call_9_12_5_2, \
call_9_12_5_3, \
call_9_12_5_4, \
call_9_12_5_5, \
call_9_12_5_6, \
call_9_12_5_7, \
call_9_12_5_8, \
call_9_12_5_9, \
call_9_12_5_10, \
call_9_12_5_11, \
call_9_12_5_12, \
call_9_12_5_13, \
call_9_12_5_14, \
call_9_12_5_15, \
call_9_12_6_0, \
call_9_12_6_1, \
call_9_12_6_2, \
call_9_12_6_3, \
call_9_12_6_4, \
call_9_12_6_5, \
call_9_12_6_6, \
call_9_12_6_7, \
call_9_12_6_8, \
call_9_12_6_9, \
call_9_12_6_10, \
call_9_12_6_11, \
call_9_12_6_12, \
call_9_12_6_13, \
call_9_12_6_14, \
call_9_12_6_15, \
call_9_12_7_0, \
call_9_12_7_1, \
call_9_12_7_2, \
call_9_12_7_3, \
call_9_12_7_4, \
call_9_12_7_5, \
call_9_12_7_6, \
call_9_12_7_7, \
call_9_12_7_8, \
call_9_12_7_9, \
call_9_12_7_10, \
call_9_12_7_11, \
call_9_12_7_12, \
call_9_12_7_13, \
call_9_12_7_14, \
call_9_12_7_15, \
call_9_12_8_0, \
call_9_12_8_1, \
call_9_12_8_2, \
call_9_12_8_3, \
call_9_12_8_4, \
call_9_12_8_5, \
call_9_12_8_6, \
call_9_12_8_7, \
call_9_12_8_8, \
call_9_12_8_9, \
call_9_12_8_10, \
call_9_12_8_11, \
call_9_12_8_12, \
call_9_12_8_13, \
call_9_12_8_14, \
call_9_12_8_15, \
call_9_12_9_0, \
call_9_12_9_1, \
call_9_12_9_2, \
call_9_12_9_3, \
call_9_12_9_4, \
call_9_12_9_5, \
call_9_12_9_6, \
call_9_12_9_7, \
call_9_12_9_8, \
call_9_12_9_9, \
call_9_12_9_10, \
call_9_12_9_11, \
call_9_12_9_12, \
call_9_12_9_13, \
call_9_12_9_14, \
call_9_12_9_15, \
call_9_12_10_0, \
call_9_12_10_1, \
call_9_12_10_2, \
call_9_12_10_3, \
call_9_12_10_4, \
call_9_12_10_5, \
call_9_12_10_6, \
call_9_12_10_7, \
call_9_12_10_8, \
call_9_12_10_9, \
call_9_12_10_10, \
call_9_12_10_11, \
call_9_12_10_12, \
call_9_12_10_13, \
call_9_12_10_14, \
call_9_12_10_15, \
call_9_12_11_0, \
call_9_12_11_1, \
call_9_12_11_2, \
call_9_12_11_3, \
call_9_12_11_4, \
call_9_12_11_5, \
call_9_12_11_6, \
call_9_12_11_7, \
call_9_12_11_8, \
call_9_12_11_9, \
call_9_12_11_10, \
call_9_12_11_11, \
call_9_12_11_12, \
call_9_12_11_13, \
call_9_12_11_14, \
call_9_12_11_15, \
call_9_12_12_0, \
call_9_12_12_1, \
call_9_12_12_2, \
call_9_12_12_3, \
call_9_12_12_4, \
call_9_12_12_5, \
call_9_12_12_6, \
call_9_12_12_7, \
call_9_12_12_8, \
call_9_12_12_9, \
call_9_12_12_10, \
call_9_12_12_11, \
call_9_12_12_12, \
call_9_12_12_13, \
call_9_12_12_14, \
call_9_12_12_15, \
call_9_12_13_0, \
call_9_12_13_1, \
call_9_12_13_2, \
call_9_12_13_3, \
call_9_12_13_4, \
call_9_12_13_5, \
call_9_12_13_6, \
call_9_12_13_7, \
call_9_12_13_8, \
call_9_12_13_9, \
call_9_12_13_10, \
call_9_12_13_11, \
call_9_12_13_12, \
call_9_12_13_13, \
call_9_12_13_14, \
call_9_12_13_15, \
call_9_12_14_0, \
call_9_12_14_1, \
call_9_12_14_2, \
call_9_12_14_3, \
call_9_12_14_4, \
call_9_12_14_5, \
call_9_12_14_6, \
call_9_12_14_7, \
call_9_12_14_8, \
call_9_12_14_9, \
call_9_12_14_10, \
call_9_12_14_11, \
call_9_12_14_12, \
call_9_12_14_13, \
call_9_12_14_14, \
call_9_12_14_15, \
call_9_12_15_0, \
call_9_12_15_1, \
call_9_12_15_2, \
call_9_12_15_3, \
call_9_12_15_4, \
call_9_12_15_5, \
call_9_12_15_6, \
call_9_12_15_7, \
call_9_12_15_8, \
call_9_12_15_9, \
call_9_12_15_10, \
call_9_12_15_11, \
call_9_12_15_12, \
call_9_12_15_13, \
call_9_12_15_14, \
call_9_12_15_15, \
call_9_13_0_0, \
call_9_13_0_1, \
call_9_13_0_2, \
call_9_13_0_3, \
call_9_13_0_4, \
call_9_13_0_5, \
call_9_13_0_6, \
call_9_13_0_7, \
call_9_13_0_8, \
call_9_13_0_9, \
call_9_13_0_10, \
call_9_13_0_11, \
call_9_13_0_12, \
call_9_13_0_13, \
call_9_13_0_14, \
call_9_13_0_15, \
call_9_13_1_0, \
call_9_13_1_1, \
call_9_13_1_2, \
call_9_13_1_3, \
call_9_13_1_4, \
call_9_13_1_5, \
call_9_13_1_6, \
call_9_13_1_7, \
call_9_13_1_8, \
call_9_13_1_9, \
call_9_13_1_10, \
call_9_13_1_11, \
call_9_13_1_12, \
call_9_13_1_13, \
call_9_13_1_14, \
call_9_13_1_15, \
call_9_13_2_0, \
call_9_13_2_1, \
call_9_13_2_2, \
call_9_13_2_3, \
call_9_13_2_4, \
call_9_13_2_5, \
call_9_13_2_6, \
call_9_13_2_7, \
call_9_13_2_8, \
call_9_13_2_9, \
call_9_13_2_10, \
call_9_13_2_11, \
call_9_13_2_12, \
call_9_13_2_13, \
call_9_13_2_14, \
call_9_13_2_15, \
call_9_13_3_0, \
call_9_13_3_1, \
call_9_13_3_2, \
call_9_13_3_3, \
call_9_13_3_4, \
call_9_13_3_5, \
call_9_13_3_6, \
call_9_13_3_7, \
call_9_13_3_8, \
call_9_13_3_9, \
call_9_13_3_10, \
call_9_13_3_11, \
call_9_13_3_12, \
call_9_13_3_13, \
call_9_13_3_14, \
call_9_13_3_15, \
call_9_13_4_0, \
call_9_13_4_1, \
call_9_13_4_2, \
call_9_13_4_3, \
call_9_13_4_4, \
call_9_13_4_5, \
call_9_13_4_6, \
call_9_13_4_7, \
call_9_13_4_8, \
call_9_13_4_9, \
call_9_13_4_10, \
call_9_13_4_11, \
call_9_13_4_12, \
call_9_13_4_13, \
call_9_13_4_14, \
call_9_13_4_15, \
call_9_13_5_0, \
call_9_13_5_1, \
call_9_13_5_2, \
call_9_13_5_3, \
call_9_13_5_4, \
call_9_13_5_5, \
call_9_13_5_6, \
call_9_13_5_7, \
call_9_13_5_8, \
call_9_13_5_9, \
call_9_13_5_10, \
call_9_13_5_11, \
call_9_13_5_12, \
call_9_13_5_13, \
call_9_13_5_14, \
call_9_13_5_15, \
call_9_13_6_0, \
call_9_13_6_1, \
call_9_13_6_2, \
call_9_13_6_3, \
call_9_13_6_4, \
call_9_13_6_5, \
call_9_13_6_6, \
call_9_13_6_7, \
call_9_13_6_8, \
call_9_13_6_9, \
call_9_13_6_10, \
call_9_13_6_11, \
call_9_13_6_12, \
call_9_13_6_13, \
call_9_13_6_14, \
call_9_13_6_15, \
call_9_13_7_0, \
call_9_13_7_1, \
call_9_13_7_2, \
call_9_13_7_3, \
call_9_13_7_4, \
call_9_13_7_5, \
call_9_13_7_6, \
call_9_13_7_7, \
call_9_13_7_8, \
call_9_13_7_9, \
call_9_13_7_10, \
call_9_13_7_11, \
call_9_13_7_12, \
call_9_13_7_13, \
call_9_13_7_14, \
call_9_13_7_15, \
call_9_13_8_0, \
call_9_13_8_1, \
call_9_13_8_2, \
call_9_13_8_3, \
call_9_13_8_4, \
call_9_13_8_5, \
call_9_13_8_6, \
call_9_13_8_7, \
call_9_13_8_8, \
call_9_13_8_9, \
call_9_13_8_10, \
call_9_13_8_11, \
call_9_13_8_12, \
call_9_13_8_13, \
call_9_13_8_14, \
call_9_13_8_15, \
call_9_13_9_0, \
call_9_13_9_1, \
call_9_13_9_2, \
call_9_13_9_3, \
call_9_13_9_4, \
call_9_13_9_5, \
call_9_13_9_6, \
call_9_13_9_7, \
call_9_13_9_8, \
call_9_13_9_9, \
call_9_13_9_10, \
call_9_13_9_11, \
call_9_13_9_12, \
call_9_13_9_13, \
call_9_13_9_14, \
call_9_13_9_15, \
call_9_13_10_0, \
call_9_13_10_1, \
call_9_13_10_2, \
call_9_13_10_3, \
call_9_13_10_4, \
call_9_13_10_5, \
call_9_13_10_6, \
call_9_13_10_7, \
call_9_13_10_8, \
call_9_13_10_9, \
call_9_13_10_10, \
call_9_13_10_11, \
call_9_13_10_12, \
call_9_13_10_13, \
call_9_13_10_14, \
call_9_13_10_15, \
call_9_13_11_0, \
call_9_13_11_1, \
call_9_13_11_2, \
call_9_13_11_3, \
call_9_13_11_4, \
call_9_13_11_5, \
call_9_13_11_6, \
call_9_13_11_7, \
call_9_13_11_8, \
call_9_13_11_9, \
call_9_13_11_10, \
call_9_13_11_11, \
call_9_13_11_12, \
call_9_13_11_13, \
call_9_13_11_14, \
call_9_13_11_15, \
call_9_13_12_0, \
call_9_13_12_1, \
call_9_13_12_2, \
call_9_13_12_3, \
call_9_13_12_4, \
call_9_13_12_5, \
call_9_13_12_6, \
call_9_13_12_7, \
call_9_13_12_8, \
call_9_13_12_9, \
call_9_13_12_10, \
call_9_13_12_11, \
call_9_13_12_12, \
call_9_13_12_13, \
call_9_13_12_14, \
call_9_13_12_15, \
call_9_13_13_0, \
call_9_13_13_1, \
call_9_13_13_2, \
call_9_13_13_3, \
call_9_13_13_4, \
call_9_13_13_5, \
call_9_13_13_6, \
call_9_13_13_7, \
call_9_13_13_8, \
call_9_13_13_9, \
call_9_13_13_10, \
call_9_13_13_11, \
call_9_13_13_12, \
call_9_13_13_13, \
call_9_13_13_14, \
call_9_13_13_15, \
call_9_13_14_0, \
call_9_13_14_1, \
call_9_13_14_2, \
call_9_13_14_3, \
call_9_13_14_4, \
call_9_13_14_5, \
call_9_13_14_6, \
call_9_13_14_7, \
call_9_13_14_8, \
call_9_13_14_9, \
call_9_13_14_10, \
call_9_13_14_11, \
call_9_13_14_12, \
call_9_13_14_13, \
call_9_13_14_14, \
call_9_13_14_15, \
call_9_13_15_0, \
call_9_13_15_1, \
call_9_13_15_2, \
call_9_13_15_3, \
call_9_13_15_4, \
call_9_13_15_5, \
call_9_13_15_6, \
call_9_13_15_7, \
call_9_13_15_8, \
call_9_13_15_9, \
call_9_13_15_10, \
call_9_13_15_11, \
call_9_13_15_12, \
call_9_13_15_13, \
call_9_13_15_14, \
call_9_13_15_15, \
call_9_14_0_0, \
call_9_14_0_1, \
call_9_14_0_2, \
call_9_14_0_3, \
call_9_14_0_4, \
call_9_14_0_5, \
call_9_14_0_6, \
call_9_14_0_7, \
call_9_14_0_8, \
call_9_14_0_9, \
call_9_14_0_10, \
call_9_14_0_11, \
call_9_14_0_12, \
call_9_14_0_13, \
call_9_14_0_14, \
call_9_14_0_15, \
call_9_14_1_0, \
call_9_14_1_1, \
call_9_14_1_2, \
call_9_14_1_3, \
call_9_14_1_4, \
call_9_14_1_5, \
call_9_14_1_6, \
call_9_14_1_7, \
call_9_14_1_8, \
call_9_14_1_9, \
call_9_14_1_10, \
call_9_14_1_11, \
call_9_14_1_12, \
call_9_14_1_13, \
call_9_14_1_14, \
call_9_14_1_15, \
call_9_14_2_0, \
call_9_14_2_1, \
call_9_14_2_2, \
call_9_14_2_3, \
call_9_14_2_4, \
call_9_14_2_5, \
call_9_14_2_6, \
call_9_14_2_7, \
call_9_14_2_8, \
call_9_14_2_9, \
call_9_14_2_10, \
call_9_14_2_11, \
call_9_14_2_12, \
call_9_14_2_13, \
call_9_14_2_14, \
call_9_14_2_15, \
call_9_14_3_0, \
call_9_14_3_1, \
call_9_14_3_2, \
call_9_14_3_3, \
call_9_14_3_4, \
call_9_14_3_5, \
call_9_14_3_6, \
call_9_14_3_7, \
call_9_14_3_8, \
call_9_14_3_9, \
call_9_14_3_10, \
call_9_14_3_11, \
call_9_14_3_12, \
call_9_14_3_13, \
call_9_14_3_14, \
call_9_14_3_15, \
call_9_14_4_0, \
call_9_14_4_1, \
call_9_14_4_2, \
call_9_14_4_3, \
call_9_14_4_4, \
call_9_14_4_5, \
call_9_14_4_6, \
call_9_14_4_7, \
call_9_14_4_8, \
call_9_14_4_9, \
call_9_14_4_10, \
call_9_14_4_11, \
call_9_14_4_12, \
call_9_14_4_13, \
call_9_14_4_14, \
call_9_14_4_15, \
call_9_14_5_0, \
call_9_14_5_1, \
call_9_14_5_2, \
call_9_14_5_3, \
call_9_14_5_4, \
call_9_14_5_5, \
call_9_14_5_6, \
call_9_14_5_7, \
call_9_14_5_8, \
call_9_14_5_9, \
call_9_14_5_10, \
call_9_14_5_11, \
call_9_14_5_12, \
call_9_14_5_13, \
call_9_14_5_14, \
call_9_14_5_15, \
call_9_14_6_0, \
call_9_14_6_1, \
call_9_14_6_2, \
call_9_14_6_3, \
call_9_14_6_4, \
call_9_14_6_5, \
call_9_14_6_6, \
call_9_14_6_7, \
call_9_14_6_8, \
call_9_14_6_9, \
call_9_14_6_10, \
call_9_14_6_11, \
call_9_14_6_12, \
call_9_14_6_13, \
call_9_14_6_14, \
call_9_14_6_15, \
call_9_14_7_0, \
call_9_14_7_1, \
call_9_14_7_2, \
call_9_14_7_3, \
call_9_14_7_4, \
call_9_14_7_5, \
call_9_14_7_6, \
call_9_14_7_7, \
call_9_14_7_8, \
call_9_14_7_9, \
call_9_14_7_10, \
call_9_14_7_11, \
call_9_14_7_12, \
call_9_14_7_13, \
call_9_14_7_14, \
call_9_14_7_15, \
call_9_14_8_0, \
call_9_14_8_1, \
call_9_14_8_2, \
call_9_14_8_3, \
call_9_14_8_4, \
call_9_14_8_5, \
call_9_14_8_6, \
call_9_14_8_7, \
call_9_14_8_8, \
call_9_14_8_9, \
call_9_14_8_10, \
call_9_14_8_11, \
call_9_14_8_12, \
call_9_14_8_13, \
call_9_14_8_14, \
call_9_14_8_15, \
call_9_14_9_0, \
call_9_14_9_1, \
call_9_14_9_2, \
call_9_14_9_3, \
call_9_14_9_4, \
call_9_14_9_5, \
call_9_14_9_6, \
call_9_14_9_7, \
call_9_14_9_8, \
call_9_14_9_9, \
call_9_14_9_10, \
call_9_14_9_11, \
call_9_14_9_12, \
call_9_14_9_13, \
call_9_14_9_14, \
call_9_14_9_15, \
call_9_14_10_0, \
call_9_14_10_1, \
call_9_14_10_2, \
call_9_14_10_3, \
call_9_14_10_4, \
call_9_14_10_5, \
call_9_14_10_6, \
call_9_14_10_7, \
call_9_14_10_8, \
call_9_14_10_9, \
call_9_14_10_10, \
call_9_14_10_11, \
call_9_14_10_12, \
call_9_14_10_13, \
call_9_14_10_14, \
call_9_14_10_15, \
call_9_14_11_0, \
call_9_14_11_1, \
call_9_14_11_2, \
call_9_14_11_3, \
call_9_14_11_4, \
call_9_14_11_5, \
call_9_14_11_6, \
call_9_14_11_7, \
call_9_14_11_8, \
call_9_14_11_9, \
call_9_14_11_10, \
call_9_14_11_11, \
call_9_14_11_12, \
call_9_14_11_13, \
call_9_14_11_14, \
call_9_14_11_15, \
call_9_14_12_0, \
call_9_14_12_1, \
call_9_14_12_2, \
call_9_14_12_3, \
call_9_14_12_4, \
call_9_14_12_5, \
call_9_14_12_6, \
call_9_14_12_7, \
call_9_14_12_8, \
call_9_14_12_9, \
call_9_14_12_10, \
call_9_14_12_11, \
call_9_14_12_12, \
call_9_14_12_13, \
call_9_14_12_14, \
call_9_14_12_15, \
call_9_14_13_0, \
call_9_14_13_1, \
call_9_14_13_2, \
call_9_14_13_3, \
call_9_14_13_4, \
call_9_14_13_5, \
call_9_14_13_6, \
call_9_14_13_7, \
call_9_14_13_8, \
call_9_14_13_9, \
call_9_14_13_10, \
call_9_14_13_11, \
call_9_14_13_12, \
call_9_14_13_13, \
call_9_14_13_14, \
call_9_14_13_15, \
call_9_14_14_0, \
call_9_14_14_1, \
call_9_14_14_2, \
call_9_14_14_3, \
call_9_14_14_4, \
call_9_14_14_5, \
call_9_14_14_6, \
call_9_14_14_7, \
call_9_14_14_8, \
call_9_14_14_9, \
call_9_14_14_10, \
call_9_14_14_11, \
call_9_14_14_12, \
call_9_14_14_13, \
call_9_14_14_14, \
call_9_14_14_15, \
call_9_14_15_0, \
call_9_14_15_1, \
call_9_14_15_2, \
call_9_14_15_3, \
call_9_14_15_4, \
call_9_14_15_5, \
call_9_14_15_6, \
call_9_14_15_7, \
call_9_14_15_8, \
call_9_14_15_9, \
call_9_14_15_10, \
call_9_14_15_11, \
call_9_14_15_12, \
call_9_14_15_13, \
call_9_14_15_14, \
call_9_14_15_15, \
call_9_15_0_0, \
call_9_15_0_1, \
call_9_15_0_2, \
call_9_15_0_3, \
call_9_15_0_4, \
call_9_15_0_5, \
call_9_15_0_6, \
call_9_15_0_7, \
call_9_15_0_8, \
call_9_15_0_9, \
call_9_15_0_10, \
call_9_15_0_11, \
call_9_15_0_12, \
call_9_15_0_13, \
call_9_15_0_14, \
call_9_15_0_15, \
call_9_15_1_0, \
call_9_15_1_1, \
call_9_15_1_2, \
call_9_15_1_3, \
call_9_15_1_4, \
call_9_15_1_5, \
call_9_15_1_6, \
call_9_15_1_7, \
call_9_15_1_8, \
call_9_15_1_9, \
call_9_15_1_10, \
call_9_15_1_11, \
call_9_15_1_12, \
call_9_15_1_13, \
call_9_15_1_14, \
call_9_15_1_15, \
call_9_15_2_0, \
call_9_15_2_1, \
call_9_15_2_2, \
call_9_15_2_3, \
call_9_15_2_4, \
call_9_15_2_5, \
call_9_15_2_6, \
call_9_15_2_7, \
call_9_15_2_8, \
call_9_15_2_9, \
call_9_15_2_10, \
call_9_15_2_11, \
call_9_15_2_12, \
call_9_15_2_13, \
call_9_15_2_14, \
call_9_15_2_15, \
call_9_15_3_0, \
call_9_15_3_1, \
call_9_15_3_2, \
call_9_15_3_3, \
call_9_15_3_4, \
call_9_15_3_5, \
call_9_15_3_6, \
call_9_15_3_7, \
call_9_15_3_8, \
call_9_15_3_9, \
call_9_15_3_10, \
call_9_15_3_11, \
call_9_15_3_12, \
call_9_15_3_13, \
call_9_15_3_14, \
call_9_15_3_15, \
call_9_15_4_0, \
call_9_15_4_1, \
call_9_15_4_2, \
call_9_15_4_3, \
call_9_15_4_4, \
call_9_15_4_5, \
call_9_15_4_6, \
call_9_15_4_7, \
call_9_15_4_8, \
call_9_15_4_9, \
call_9_15_4_10, \
call_9_15_4_11, \
call_9_15_4_12, \
call_9_15_4_13, \
call_9_15_4_14, \
call_9_15_4_15, \
call_9_15_5_0, \
call_9_15_5_1, \
call_9_15_5_2, \
call_9_15_5_3, \
call_9_15_5_4, \
call_9_15_5_5, \
call_9_15_5_6, \
call_9_15_5_7, \
call_9_15_5_8, \
call_9_15_5_9, \
call_9_15_5_10, \
call_9_15_5_11, \
call_9_15_5_12, \
call_9_15_5_13, \
call_9_15_5_14, \
call_9_15_5_15, \
call_9_15_6_0, \
call_9_15_6_1, \
call_9_15_6_2, \
call_9_15_6_3, \
call_9_15_6_4, \
call_9_15_6_5, \
call_9_15_6_6, \
call_9_15_6_7, \
call_9_15_6_8, \
call_9_15_6_9, \
call_9_15_6_10, \
call_9_15_6_11, \
call_9_15_6_12, \
call_9_15_6_13, \
call_9_15_6_14, \
call_9_15_6_15, \
call_9_15_7_0, \
call_9_15_7_1, \
call_9_15_7_2, \
call_9_15_7_3, \
call_9_15_7_4, \
call_9_15_7_5, \
call_9_15_7_6, \
call_9_15_7_7, \
call_9_15_7_8, \
call_9_15_7_9, \
call_9_15_7_10, \
call_9_15_7_11, \
call_9_15_7_12, \
call_9_15_7_13, \
call_9_15_7_14, \
call_9_15_7_15, \
call_9_15_8_0, \
call_9_15_8_1, \
call_9_15_8_2, \
call_9_15_8_3, \
call_9_15_8_4, \
call_9_15_8_5, \
call_9_15_8_6, \
call_9_15_8_7, \
call_9_15_8_8, \
call_9_15_8_9, \
call_9_15_8_10, \
call_9_15_8_11, \
call_9_15_8_12, \
call_9_15_8_13, \
call_9_15_8_14, \
call_9_15_8_15, \
call_9_15_9_0, \
call_9_15_9_1, \
call_9_15_9_2, \
call_9_15_9_3, \
call_9_15_9_4, \
call_9_15_9_5, \
call_9_15_9_6, \
call_9_15_9_7, \
call_9_15_9_8, \
call_9_15_9_9, \
call_9_15_9_10, \
call_9_15_9_11, \
call_9_15_9_12, \
call_9_15_9_13, \
call_9_15_9_14, \
call_9_15_9_15, \
call_9_15_10_0, \
call_9_15_10_1, \
call_9_15_10_2, \
call_9_15_10_3, \
call_9_15_10_4, \
call_9_15_10_5, \
call_9_15_10_6, \
call_9_15_10_7, \
call_9_15_10_8, \
call_9_15_10_9, \
call_9_15_10_10, \
call_9_15_10_11, \
call_9_15_10_12, \
call_9_15_10_13, \
call_9_15_10_14, \
call_9_15_10_15, \
call_9_15_11_0, \
call_9_15_11_1, \
call_9_15_11_2, \
call_9_15_11_3, \
call_9_15_11_4, \
call_9_15_11_5, \
call_9_15_11_6, \
call_9_15_11_7, \
call_9_15_11_8, \
call_9_15_11_9, \
call_9_15_11_10, \
call_9_15_11_11, \
call_9_15_11_12, \
call_9_15_11_13, \
call_9_15_11_14, \
call_9_15_11_15, \
call_9_15_12_0, \
call_9_15_12_1, \
call_9_15_12_2, \
call_9_15_12_3, \
call_9_15_12_4, \
call_9_15_12_5, \
call_9_15_12_6, \
call_9_15_12_7, \
call_9_15_12_8, \
call_9_15_12_9, \
call_9_15_12_10, \
call_9_15_12_11, \
call_9_15_12_12, \
call_9_15_12_13, \
call_9_15_12_14, \
call_9_15_12_15, \
call_9_15_13_0, \
call_9_15_13_1, \
call_9_15_13_2, \
call_9_15_13_3, \
call_9_15_13_4, \
call_9_15_13_5, \
call_9_15_13_6, \
call_9_15_13_7, \
call_9_15_13_8, \
call_9_15_13_9, \
call_9_15_13_10, \
call_9_15_13_11, \
call_9_15_13_12, \
call_9_15_13_13, \
call_9_15_13_14, \
call_9_15_13_15, \
call_9_15_14_0, \
call_9_15_14_1, \
call_9_15_14_2, \
call_9_15_14_3, \
call_9_15_14_4, \
call_9_15_14_5, \
call_9_15_14_6, \
call_9_15_14_7, \
call_9_15_14_8, \
call_9_15_14_9, \
call_9_15_14_10, \
call_9_15_14_11, \
call_9_15_14_12, \
call_9_15_14_13, \
call_9_15_14_14, \
call_9_15_14_15, \
call_9_15_15_0, \
call_9_15_15_1, \
call_9_15_15_2, \
call_9_15_15_3, \
call_9_15_15_4, \
call_9_15_15_5, \
call_9_15_15_6, \
call_9_15_15_7, \
call_9_15_15_8, \
call_9_15_15_9, \
call_9_15_15_10, \
call_9_15_15_11, \
call_9_15_15_12, \
call_9_15_15_13, \
call_9_15_15_14, \
call_9_15_15_15, \
call_10_0_0_0, \
call_10_0_0_1, \
call_10_0_0_2, \
call_10_0_0_3, \
call_10_0_0_4, \
call_10_0_0_5, \
call_10_0_0_6, \
call_10_0_0_7, \
call_10_0_0_8, \
call_10_0_0_9, \
call_10_0_0_10, \
call_10_0_0_11, \
call_10_0_0_12, \
call_10_0_0_13, \
call_10_0_0_14, \
call_10_0_0_15, \
call_10_0_1_0, \
call_10_0_1_1, \
call_10_0_1_2, \
call_10_0_1_3, \
call_10_0_1_4, \
call_10_0_1_5, \
call_10_0_1_6, \
call_10_0_1_7, \
call_10_0_1_8, \
call_10_0_1_9, \
call_10_0_1_10, \
call_10_0_1_11, \
call_10_0_1_12, \
call_10_0_1_13, \
call_10_0_1_14, \
call_10_0_1_15, \
call_10_0_2_0, \
call_10_0_2_1, \
call_10_0_2_2, \
call_10_0_2_3, \
call_10_0_2_4, \
call_10_0_2_5, \
call_10_0_2_6, \
call_10_0_2_7, \
call_10_0_2_8, \
call_10_0_2_9, \
call_10_0_2_10, \
call_10_0_2_11, \
call_10_0_2_12, \
call_10_0_2_13, \
call_10_0_2_14, \
call_10_0_2_15, \
call_10_0_3_0, \
call_10_0_3_1, \
call_10_0_3_2, \
call_10_0_3_3, \
call_10_0_3_4, \
call_10_0_3_5, \
call_10_0_3_6, \
call_10_0_3_7, \
call_10_0_3_8, \
call_10_0_3_9, \
call_10_0_3_10, \
call_10_0_3_11, \
call_10_0_3_12, \
call_10_0_3_13, \
call_10_0_3_14, \
call_10_0_3_15, \
call_10_0_4_0, \
call_10_0_4_1, \
call_10_0_4_2, \
call_10_0_4_3, \
call_10_0_4_4, \
call_10_0_4_5, \
call_10_0_4_6, \
call_10_0_4_7, \
call_10_0_4_8, \
call_10_0_4_9, \
call_10_0_4_10, \
call_10_0_4_11, \
call_10_0_4_12, \
call_10_0_4_13, \
call_10_0_4_14, \
call_10_0_4_15, \
call_10_0_5_0, \
call_10_0_5_1, \
call_10_0_5_2, \
call_10_0_5_3, \
call_10_0_5_4, \
call_10_0_5_5, \
call_10_0_5_6, \
call_10_0_5_7, \
call_10_0_5_8, \
call_10_0_5_9, \
call_10_0_5_10, \
call_10_0_5_11, \
call_10_0_5_12, \
call_10_0_5_13, \
call_10_0_5_14, \
call_10_0_5_15, \
call_10_0_6_0, \
call_10_0_6_1, \
call_10_0_6_2, \
call_10_0_6_3, \
call_10_0_6_4, \
call_10_0_6_5, \
call_10_0_6_6, \
call_10_0_6_7, \
call_10_0_6_8, \
call_10_0_6_9, \
call_10_0_6_10, \
call_10_0_6_11, \
call_10_0_6_12, \
call_10_0_6_13, \
call_10_0_6_14, \
call_10_0_6_15, \
call_10_0_7_0, \
call_10_0_7_1, \
call_10_0_7_2, \
call_10_0_7_3, \
call_10_0_7_4, \
call_10_0_7_5, \
call_10_0_7_6, \
call_10_0_7_7, \
call_10_0_7_8, \
call_10_0_7_9, \
call_10_0_7_10, \
call_10_0_7_11, \
call_10_0_7_12, \
call_10_0_7_13, \
call_10_0_7_14, \
call_10_0_7_15, \
call_10_0_8_0, \
call_10_0_8_1, \
call_10_0_8_2, \
call_10_0_8_3, \
call_10_0_8_4, \
call_10_0_8_5, \
call_10_0_8_6, \
call_10_0_8_7, \
call_10_0_8_8, \
call_10_0_8_9, \
call_10_0_8_10, \
call_10_0_8_11, \
call_10_0_8_12, \
call_10_0_8_13, \
call_10_0_8_14, \
call_10_0_8_15, \
call_10_0_9_0, \
call_10_0_9_1, \
call_10_0_9_2, \
call_10_0_9_3, \
call_10_0_9_4, \
call_10_0_9_5, \
call_10_0_9_6, \
call_10_0_9_7, \
call_10_0_9_8, \
call_10_0_9_9, \
call_10_0_9_10, \
call_10_0_9_11, \
call_10_0_9_12, \
call_10_0_9_13, \
call_10_0_9_14, \
call_10_0_9_15, \
call_10_0_10_0, \
call_10_0_10_1, \
call_10_0_10_2, \
call_10_0_10_3, \
call_10_0_10_4, \
call_10_0_10_5, \
call_10_0_10_6, \
call_10_0_10_7, \
call_10_0_10_8, \
call_10_0_10_9, \
call_10_0_10_10, \
call_10_0_10_11, \
call_10_0_10_12, \
call_10_0_10_13, \
call_10_0_10_14, \
call_10_0_10_15, \
call_10_0_11_0, \
call_10_0_11_1, \
call_10_0_11_2, \
call_10_0_11_3, \
call_10_0_11_4, \
call_10_0_11_5, \
call_10_0_11_6, \
call_10_0_11_7, \
call_10_0_11_8, \
call_10_0_11_9, \
call_10_0_11_10, \
call_10_0_11_11, \
call_10_0_11_12, \
call_10_0_11_13, \
call_10_0_11_14, \
call_10_0_11_15, \
call_10_0_12_0, \
call_10_0_12_1, \
call_10_0_12_2, \
call_10_0_12_3, \
call_10_0_12_4, \
call_10_0_12_5, \
call_10_0_12_6, \
call_10_0_12_7, \
call_10_0_12_8, \
call_10_0_12_9, \
call_10_0_12_10, \
call_10_0_12_11, \
call_10_0_12_12, \
call_10_0_12_13, \
call_10_0_12_14, \
call_10_0_12_15, \
call_10_0_13_0, \
call_10_0_13_1, \
call_10_0_13_2, \
call_10_0_13_3, \
call_10_0_13_4, \
call_10_0_13_5, \
call_10_0_13_6, \
call_10_0_13_7, \
call_10_0_13_8, \
call_10_0_13_9, \
call_10_0_13_10, \
call_10_0_13_11, \
call_10_0_13_12, \
call_10_0_13_13, \
call_10_0_13_14, \
call_10_0_13_15, \
call_10_0_14_0, \
call_10_0_14_1, \
call_10_0_14_2, \
call_10_0_14_3, \
call_10_0_14_4, \
call_10_0_14_5, \
call_10_0_14_6, \
call_10_0_14_7, \
call_10_0_14_8, \
call_10_0_14_9, \
call_10_0_14_10, \
call_10_0_14_11, \
call_10_0_14_12, \
call_10_0_14_13, \
call_10_0_14_14, \
call_10_0_14_15, \
call_10_0_15_0, \
call_10_0_15_1, \
call_10_0_15_2, \
call_10_0_15_3, \
call_10_0_15_4, \
call_10_0_15_5, \
call_10_0_15_6, \
call_10_0_15_7, \
call_10_0_15_8, \
call_10_0_15_9, \
call_10_0_15_10, \
call_10_0_15_11, \
call_10_0_15_12, \
call_10_0_15_13, \
call_10_0_15_14, \
call_10_0_15_15, \
call_10_1_0_0, \
call_10_1_0_1, \
call_10_1_0_2, \
call_10_1_0_3, \
call_10_1_0_4, \
call_10_1_0_5, \
call_10_1_0_6, \
call_10_1_0_7, \
call_10_1_0_8, \
call_10_1_0_9, \
call_10_1_0_10, \
call_10_1_0_11, \
call_10_1_0_12, \
call_10_1_0_13, \
call_10_1_0_14, \
call_10_1_0_15, \
call_10_1_1_0, \
call_10_1_1_1, \
call_10_1_1_2, \
call_10_1_1_3, \
call_10_1_1_4, \
call_10_1_1_5, \
call_10_1_1_6, \
call_10_1_1_7, \
call_10_1_1_8, \
call_10_1_1_9, \
call_10_1_1_10, \
call_10_1_1_11, \
call_10_1_1_12, \
call_10_1_1_13, \
call_10_1_1_14, \
call_10_1_1_15, \
call_10_1_2_0, \
call_10_1_2_1, \
call_10_1_2_2, \
call_10_1_2_3, \
call_10_1_2_4, \
call_10_1_2_5, \
call_10_1_2_6, \
call_10_1_2_7, \
call_10_1_2_8, \
call_10_1_2_9, \
call_10_1_2_10, \
call_10_1_2_11, \
call_10_1_2_12, \
call_10_1_2_13, \
call_10_1_2_14, \
call_10_1_2_15, \
call_10_1_3_0, \
call_10_1_3_1, \
call_10_1_3_2, \
call_10_1_3_3, \
call_10_1_3_4, \
call_10_1_3_5, \
call_10_1_3_6, \
call_10_1_3_7, \
call_10_1_3_8, \
call_10_1_3_9, \
call_10_1_3_10, \
call_10_1_3_11, \
call_10_1_3_12, \
call_10_1_3_13, \
call_10_1_3_14, \
call_10_1_3_15, \
call_10_1_4_0, \
call_10_1_4_1, \
call_10_1_4_2, \
call_10_1_4_3, \
call_10_1_4_4, \
call_10_1_4_5, \
call_10_1_4_6, \
call_10_1_4_7, \
call_10_1_4_8, \
call_10_1_4_9, \
call_10_1_4_10, \
call_10_1_4_11, \
call_10_1_4_12, \
call_10_1_4_13, \
call_10_1_4_14, \
call_10_1_4_15, \
call_10_1_5_0, \
call_10_1_5_1, \
call_10_1_5_2, \
call_10_1_5_3, \
call_10_1_5_4, \
call_10_1_5_5, \
call_10_1_5_6, \
call_10_1_5_7, \
call_10_1_5_8, \
call_10_1_5_9, \
call_10_1_5_10, \
call_10_1_5_11, \
call_10_1_5_12, \
call_10_1_5_13, \
call_10_1_5_14, \
call_10_1_5_15, \
call_10_1_6_0, \
call_10_1_6_1, \
call_10_1_6_2, \
call_10_1_6_3, \
call_10_1_6_4, \
call_10_1_6_5, \
call_10_1_6_6, \
call_10_1_6_7, \
call_10_1_6_8, \
call_10_1_6_9, \
call_10_1_6_10, \
call_10_1_6_11, \
call_10_1_6_12, \
call_10_1_6_13, \
call_10_1_6_14, \
call_10_1_6_15, \
call_10_1_7_0, \
call_10_1_7_1, \
call_10_1_7_2, \
call_10_1_7_3, \
call_10_1_7_4, \
call_10_1_7_5, \
call_10_1_7_6, \
call_10_1_7_7, \
call_10_1_7_8, \
call_10_1_7_9, \
call_10_1_7_10, \
call_10_1_7_11, \
call_10_1_7_12, \
call_10_1_7_13, \
call_10_1_7_14, \
call_10_1_7_15, \
call_10_1_8_0, \
call_10_1_8_1, \
call_10_1_8_2, \
call_10_1_8_3, \
call_10_1_8_4, \
call_10_1_8_5, \
call_10_1_8_6, \
call_10_1_8_7, \
call_10_1_8_8, \
call_10_1_8_9, \
call_10_1_8_10, \
call_10_1_8_11, \
call_10_1_8_12, \
call_10_1_8_13, \
call_10_1_8_14, \
call_10_1_8_15, \
call_10_1_9_0, \
call_10_1_9_1, \
call_10_1_9_2, \
call_10_1_9_3, \
call_10_1_9_4, \
call_10_1_9_5, \
call_10_1_9_6, \
call_10_1_9_7, \
call_10_1_9_8, \
call_10_1_9_9, \
call_10_1_9_10, \
call_10_1_9_11, \
call_10_1_9_12, \
call_10_1_9_13, \
call_10_1_9_14, \
call_10_1_9_15, \
call_10_1_10_0, \
call_10_1_10_1, \
call_10_1_10_2, \
call_10_1_10_3, \
call_10_1_10_4, \
call_10_1_10_5, \
call_10_1_10_6, \
call_10_1_10_7, \
call_10_1_10_8, \
call_10_1_10_9, \
call_10_1_10_10, \
call_10_1_10_11, \
call_10_1_10_12, \
call_10_1_10_13, \
call_10_1_10_14, \
call_10_1_10_15, \
call_10_1_11_0, \
call_10_1_11_1, \
call_10_1_11_2, \
call_10_1_11_3, \
call_10_1_11_4, \
call_10_1_11_5, \
call_10_1_11_6, \
call_10_1_11_7, \
call_10_1_11_8, \
call_10_1_11_9, \
call_10_1_11_10, \
call_10_1_11_11, \
call_10_1_11_12, \
call_10_1_11_13, \
call_10_1_11_14, \
call_10_1_11_15, \
call_10_1_12_0, \
call_10_1_12_1, \
call_10_1_12_2, \
call_10_1_12_3, \
call_10_1_12_4, \
call_10_1_12_5, \
call_10_1_12_6, \
call_10_1_12_7, \
call_10_1_12_8, \
call_10_1_12_9, \
call_10_1_12_10, \
call_10_1_12_11, \
call_10_1_12_12, \
call_10_1_12_13, \
call_10_1_12_14, \
call_10_1_12_15, \
call_10_1_13_0, \
call_10_1_13_1, \
call_10_1_13_2, \
call_10_1_13_3, \
call_10_1_13_4, \
call_10_1_13_5, \
call_10_1_13_6, \
call_10_1_13_7, \
call_10_1_13_8, \
call_10_1_13_9, \
call_10_1_13_10, \
call_10_1_13_11, \
call_10_1_13_12, \
call_10_1_13_13, \
call_10_1_13_14, \
call_10_1_13_15, \
call_10_1_14_0, \
call_10_1_14_1, \
call_10_1_14_2, \
call_10_1_14_3, \
call_10_1_14_4, \
call_10_1_14_5, \
call_10_1_14_6, \
call_10_1_14_7, \
call_10_1_14_8, \
call_10_1_14_9, \
call_10_1_14_10, \
call_10_1_14_11, \
call_10_1_14_12, \
call_10_1_14_13, \
call_10_1_14_14, \
call_10_1_14_15, \
call_10_1_15_0, \
call_10_1_15_1, \
call_10_1_15_2, \
call_10_1_15_3, \
call_10_1_15_4, \
call_10_1_15_5, \
call_10_1_15_6, \
call_10_1_15_7, \
call_10_1_15_8, \
call_10_1_15_9, \
call_10_1_15_10, \
call_10_1_15_11, \
call_10_1_15_12, \
call_10_1_15_13, \
call_10_1_15_14, \
call_10_1_15_15, \
call_10_2_0_0, \
call_10_2_0_1, \
call_10_2_0_2, \
call_10_2_0_3, \
call_10_2_0_4, \
call_10_2_0_5, \
call_10_2_0_6, \
call_10_2_0_7, \
call_10_2_0_8, \
call_10_2_0_9, \
call_10_2_0_10, \
call_10_2_0_11, \
call_10_2_0_12, \
call_10_2_0_13, \
call_10_2_0_14, \
call_10_2_0_15, \
call_10_2_1_0, \
call_10_2_1_1, \
call_10_2_1_2, \
call_10_2_1_3, \
call_10_2_1_4, \
call_10_2_1_5, \
call_10_2_1_6, \
call_10_2_1_7, \
call_10_2_1_8, \
call_10_2_1_9, \
call_10_2_1_10, \
call_10_2_1_11, \
call_10_2_1_12, \
call_10_2_1_13, \
call_10_2_1_14, \
call_10_2_1_15, \
call_10_2_2_0, \
call_10_2_2_1, \
call_10_2_2_2, \
call_10_2_2_3, \
call_10_2_2_4, \
call_10_2_2_5, \
call_10_2_2_6, \
call_10_2_2_7, \
call_10_2_2_8, \
call_10_2_2_9, \
call_10_2_2_10, \
call_10_2_2_11, \
call_10_2_2_12, \
call_10_2_2_13, \
call_10_2_2_14, \
call_10_2_2_15, \
call_10_2_3_0, \
call_10_2_3_1, \
call_10_2_3_2, \
call_10_2_3_3, \
call_10_2_3_4, \
call_10_2_3_5, \
call_10_2_3_6, \
call_10_2_3_7, \
call_10_2_3_8, \
call_10_2_3_9, \
call_10_2_3_10, \
call_10_2_3_11, \
call_10_2_3_12, \
call_10_2_3_13, \
call_10_2_3_14, \
call_10_2_3_15, \
call_10_2_4_0, \
call_10_2_4_1, \
call_10_2_4_2, \
call_10_2_4_3, \
call_10_2_4_4, \
call_10_2_4_5, \
call_10_2_4_6, \
call_10_2_4_7, \
call_10_2_4_8, \
call_10_2_4_9, \
call_10_2_4_10, \
call_10_2_4_11, \
call_10_2_4_12, \
call_10_2_4_13, \
call_10_2_4_14, \
call_10_2_4_15, \
call_10_2_5_0, \
call_10_2_5_1, \
call_10_2_5_2, \
call_10_2_5_3, \
call_10_2_5_4, \
call_10_2_5_5, \
call_10_2_5_6, \
call_10_2_5_7, \
call_10_2_5_8, \
call_10_2_5_9, \
call_10_2_5_10, \
call_10_2_5_11, \
call_10_2_5_12, \
call_10_2_5_13, \
call_10_2_5_14, \
call_10_2_5_15, \
call_10_2_6_0, \
call_10_2_6_1, \
call_10_2_6_2, \
call_10_2_6_3, \
call_10_2_6_4, \
call_10_2_6_5, \
call_10_2_6_6, \
call_10_2_6_7, \
call_10_2_6_8, \
call_10_2_6_9, \
call_10_2_6_10, \
call_10_2_6_11, \
call_10_2_6_12, \
call_10_2_6_13, \
call_10_2_6_14, \
call_10_2_6_15, \
call_10_2_7_0, \
call_10_2_7_1, \
call_10_2_7_2, \
call_10_2_7_3, \
call_10_2_7_4, \
call_10_2_7_5, \
call_10_2_7_6, \
call_10_2_7_7, \
call_10_2_7_8, \
call_10_2_7_9, \
call_10_2_7_10, \
call_10_2_7_11, \
call_10_2_7_12, \
call_10_2_7_13, \
call_10_2_7_14, \
call_10_2_7_15, \
call_10_2_8_0, \
call_10_2_8_1, \
call_10_2_8_2, \
call_10_2_8_3, \
call_10_2_8_4, \
call_10_2_8_5, \
call_10_2_8_6, \
call_10_2_8_7, \
call_10_2_8_8, \
call_10_2_8_9, \
call_10_2_8_10, \
call_10_2_8_11, \
call_10_2_8_12, \
call_10_2_8_13, \
call_10_2_8_14, \
call_10_2_8_15, \
call_10_2_9_0, \
call_10_2_9_1, \
call_10_2_9_2, \
call_10_2_9_3, \
call_10_2_9_4, \
call_10_2_9_5, \
call_10_2_9_6, \
call_10_2_9_7, \
call_10_2_9_8, \
call_10_2_9_9, \
call_10_2_9_10, \
call_10_2_9_11, \
call_10_2_9_12, \
call_10_2_9_13, \
call_10_2_9_14, \
call_10_2_9_15, \
call_10_2_10_0, \
call_10_2_10_1, \
call_10_2_10_2, \
call_10_2_10_3, \
call_10_2_10_4, \
call_10_2_10_5, \
call_10_2_10_6, \
call_10_2_10_7, \
call_10_2_10_8, \
call_10_2_10_9, \
call_10_2_10_10, \
call_10_2_10_11, \
call_10_2_10_12, \
call_10_2_10_13, \
call_10_2_10_14, \
call_10_2_10_15, \
call_10_2_11_0, \
call_10_2_11_1, \
call_10_2_11_2, \
call_10_2_11_3, \
call_10_2_11_4, \
call_10_2_11_5, \
call_10_2_11_6, \
call_10_2_11_7, \
call_10_2_11_8, \
call_10_2_11_9, \
call_10_2_11_10, \
call_10_2_11_11, \
call_10_2_11_12, \
call_10_2_11_13, \
call_10_2_11_14, \
call_10_2_11_15, \
call_10_2_12_0, \
call_10_2_12_1, \
call_10_2_12_2, \
call_10_2_12_3, \
call_10_2_12_4, \
call_10_2_12_5, \
call_10_2_12_6, \
call_10_2_12_7, \
call_10_2_12_8, \
call_10_2_12_9, \
call_10_2_12_10, \
call_10_2_12_11, \
call_10_2_12_12, \
call_10_2_12_13, \
call_10_2_12_14, \
call_10_2_12_15, \
call_10_2_13_0, \
call_10_2_13_1, \
call_10_2_13_2, \
call_10_2_13_3, \
call_10_2_13_4, \
call_10_2_13_5, \
call_10_2_13_6, \
call_10_2_13_7, \
call_10_2_13_8, \
call_10_2_13_9, \
call_10_2_13_10, \
call_10_2_13_11, \
call_10_2_13_12, \
call_10_2_13_13, \
call_10_2_13_14, \
call_10_2_13_15, \
call_10_2_14_0, \
call_10_2_14_1, \
call_10_2_14_2, \
call_10_2_14_3, \
call_10_2_14_4, \
call_10_2_14_5, \
call_10_2_14_6, \
call_10_2_14_7, \
call_10_2_14_8, \
call_10_2_14_9, \
call_10_2_14_10, \
call_10_2_14_11, \
call_10_2_14_12, \
call_10_2_14_13, \
call_10_2_14_14, \
call_10_2_14_15, \
call_10_2_15_0, \
call_10_2_15_1, \
call_10_2_15_2, \
call_10_2_15_3, \
call_10_2_15_4, \
call_10_2_15_5, \
call_10_2_15_6, \
call_10_2_15_7, \
call_10_2_15_8, \
call_10_2_15_9, \
call_10_2_15_10, \
call_10_2_15_11, \
call_10_2_15_12, \
call_10_2_15_13, \
call_10_2_15_14, \
call_10_2_15_15, \
call_10_3_0_0, \
call_10_3_0_1, \
call_10_3_0_2, \
call_10_3_0_3, \
call_10_3_0_4, \
call_10_3_0_5, \
call_10_3_0_6, \
call_10_3_0_7, \
call_10_3_0_8, \
call_10_3_0_9, \
call_10_3_0_10, \
call_10_3_0_11, \
call_10_3_0_12, \
call_10_3_0_13, \
call_10_3_0_14, \
call_10_3_0_15, \
call_10_3_1_0, \
call_10_3_1_1, \
call_10_3_1_2, \
call_10_3_1_3, \
call_10_3_1_4, \
call_10_3_1_5, \
call_10_3_1_6, \
call_10_3_1_7, \
call_10_3_1_8, \
call_10_3_1_9, \
call_10_3_1_10, \
call_10_3_1_11, \
call_10_3_1_12, \
call_10_3_1_13, \
call_10_3_1_14, \
call_10_3_1_15, \
call_10_3_2_0, \
call_10_3_2_1, \
call_10_3_2_2, \
call_10_3_2_3, \
call_10_3_2_4, \
call_10_3_2_5, \
call_10_3_2_6, \
call_10_3_2_7, \
call_10_3_2_8, \
call_10_3_2_9, \
call_10_3_2_10, \
call_10_3_2_11, \
call_10_3_2_12, \
call_10_3_2_13, \
call_10_3_2_14, \
call_10_3_2_15, \
call_10_3_3_0, \
call_10_3_3_1, \
call_10_3_3_2, \
call_10_3_3_3, \
call_10_3_3_4, \
call_10_3_3_5, \
call_10_3_3_6, \
call_10_3_3_7, \
call_10_3_3_8, \
call_10_3_3_9, \
call_10_3_3_10, \
call_10_3_3_11, \
call_10_3_3_12, \
call_10_3_3_13, \
call_10_3_3_14, \
call_10_3_3_15, \
call_10_3_4_0, \
call_10_3_4_1, \
call_10_3_4_2, \
call_10_3_4_3, \
call_10_3_4_4, \
call_10_3_4_5, \
call_10_3_4_6, \
call_10_3_4_7, \
call_10_3_4_8, \
call_10_3_4_9, \
call_10_3_4_10, \
call_10_3_4_11, \
call_10_3_4_12, \
call_10_3_4_13, \
call_10_3_4_14, \
call_10_3_4_15, \
call_10_3_5_0, \
call_10_3_5_1, \
call_10_3_5_2, \
call_10_3_5_3, \
call_10_3_5_4, \
call_10_3_5_5, \
call_10_3_5_6, \
call_10_3_5_7, \
call_10_3_5_8, \
call_10_3_5_9, \
call_10_3_5_10, \
call_10_3_5_11, \
call_10_3_5_12, \
call_10_3_5_13, \
call_10_3_5_14, \
call_10_3_5_15, \
call_10_3_6_0, \
call_10_3_6_1, \
call_10_3_6_2, \
call_10_3_6_3, \
call_10_3_6_4, \
call_10_3_6_5, \
call_10_3_6_6, \
call_10_3_6_7, \
call_10_3_6_8, \
call_10_3_6_9, \
call_10_3_6_10, \
call_10_3_6_11, \
call_10_3_6_12, \
call_10_3_6_13, \
call_10_3_6_14, \
call_10_3_6_15, \
call_10_3_7_0, \
call_10_3_7_1, \
call_10_3_7_2, \
call_10_3_7_3, \
call_10_3_7_4, \
call_10_3_7_5, \
call_10_3_7_6, \
call_10_3_7_7, \
call_10_3_7_8, \
call_10_3_7_9, \
call_10_3_7_10, \
call_10_3_7_11, \
call_10_3_7_12, \
call_10_3_7_13, \
call_10_3_7_14, \
call_10_3_7_15, \
call_10_3_8_0, \
call_10_3_8_1, \
call_10_3_8_2, \
call_10_3_8_3, \
call_10_3_8_4, \
call_10_3_8_5, \
call_10_3_8_6, \
call_10_3_8_7, \
call_10_3_8_8, \
call_10_3_8_9, \
call_10_3_8_10, \
call_10_3_8_11, \
call_10_3_8_12, \
call_10_3_8_13, \
call_10_3_8_14, \
call_10_3_8_15, \
call_10_3_9_0, \
call_10_3_9_1, \
call_10_3_9_2, \
call_10_3_9_3, \
call_10_3_9_4, \
call_10_3_9_5, \
call_10_3_9_6, \
call_10_3_9_7, \
call_10_3_9_8, \
call_10_3_9_9, \
call_10_3_9_10, \
call_10_3_9_11, \
call_10_3_9_12, \
call_10_3_9_13, \
call_10_3_9_14, \
call_10_3_9_15, \
call_10_3_10_0, \
call_10_3_10_1, \
call_10_3_10_2, \
call_10_3_10_3, \
call_10_3_10_4, \
call_10_3_10_5, \
call_10_3_10_6, \
call_10_3_10_7, \
call_10_3_10_8, \
call_10_3_10_9, \
call_10_3_10_10, \
call_10_3_10_11, \
call_10_3_10_12, \
call_10_3_10_13, \
call_10_3_10_14, \
call_10_3_10_15, \
call_10_3_11_0, \
call_10_3_11_1, \
call_10_3_11_2, \
call_10_3_11_3, \
call_10_3_11_4, \
call_10_3_11_5, \
call_10_3_11_6, \
call_10_3_11_7, \
call_10_3_11_8, \
call_10_3_11_9, \
call_10_3_11_10, \
call_10_3_11_11, \
call_10_3_11_12, \
call_10_3_11_13, \
call_10_3_11_14, \
call_10_3_11_15, \
call_10_3_12_0, \
call_10_3_12_1, \
call_10_3_12_2, \
call_10_3_12_3, \
call_10_3_12_4, \
call_10_3_12_5, \
call_10_3_12_6, \
call_10_3_12_7, \
call_10_3_12_8, \
call_10_3_12_9, \
call_10_3_12_10, \
call_10_3_12_11, \
call_10_3_12_12, \
call_10_3_12_13, \
call_10_3_12_14, \
call_10_3_12_15, \
call_10_3_13_0, \
call_10_3_13_1, \
call_10_3_13_2, \
call_10_3_13_3, \
call_10_3_13_4, \
call_10_3_13_5, \
call_10_3_13_6, \
call_10_3_13_7, \
call_10_3_13_8, \
call_10_3_13_9, \
call_10_3_13_10, \
call_10_3_13_11, \
call_10_3_13_12, \
call_10_3_13_13, \
call_10_3_13_14, \
call_10_3_13_15, \
call_10_3_14_0, \
call_10_3_14_1, \
call_10_3_14_2, \
call_10_3_14_3, \
call_10_3_14_4, \
call_10_3_14_5, \
call_10_3_14_6, \
call_10_3_14_7, \
call_10_3_14_8, \
call_10_3_14_9, \
call_10_3_14_10, \
call_10_3_14_11, \
call_10_3_14_12, \
call_10_3_14_13, \
call_10_3_14_14, \
call_10_3_14_15, \
call_10_3_15_0, \
call_10_3_15_1, \
call_10_3_15_2, \
call_10_3_15_3, \
call_10_3_15_4, \
call_10_3_15_5, \
call_10_3_15_6, \
call_10_3_15_7, \
call_10_3_15_8, \
call_10_3_15_9, \
call_10_3_15_10, \
call_10_3_15_11, \
call_10_3_15_12, \
call_10_3_15_13, \
call_10_3_15_14, \
call_10_3_15_15, \
call_10_4_0_0, \
call_10_4_0_1, \
call_10_4_0_2, \
call_10_4_0_3, \
call_10_4_0_4, \
call_10_4_0_5, \
call_10_4_0_6, \
call_10_4_0_7, \
call_10_4_0_8, \
call_10_4_0_9, \
call_10_4_0_10, \
call_10_4_0_11, \
call_10_4_0_12, \
call_10_4_0_13, \
call_10_4_0_14, \
call_10_4_0_15, \
call_10_4_1_0, \
call_10_4_1_1, \
call_10_4_1_2, \
call_10_4_1_3, \
call_10_4_1_4, \
call_10_4_1_5, \
call_10_4_1_6, \
call_10_4_1_7, \
call_10_4_1_8, \
call_10_4_1_9, \
call_10_4_1_10, \
call_10_4_1_11, \
call_10_4_1_12, \
call_10_4_1_13, \
call_10_4_1_14, \
call_10_4_1_15, \
call_10_4_2_0, \
call_10_4_2_1, \
call_10_4_2_2, \
call_10_4_2_3, \
call_10_4_2_4, \
call_10_4_2_5, \
call_10_4_2_6, \
call_10_4_2_7, \
call_10_4_2_8, \
call_10_4_2_9, \
call_10_4_2_10, \
call_10_4_2_11, \
call_10_4_2_12, \
call_10_4_2_13, \
call_10_4_2_14, \
call_10_4_2_15, \
call_10_4_3_0, \
call_10_4_3_1, \
call_10_4_3_2, \
call_10_4_3_3, \
call_10_4_3_4, \
call_10_4_3_5, \
call_10_4_3_6, \
call_10_4_3_7, \
call_10_4_3_8, \
call_10_4_3_9, \
call_10_4_3_10, \
call_10_4_3_11, \
call_10_4_3_12, \
call_10_4_3_13, \
call_10_4_3_14, \
call_10_4_3_15, \
call_10_4_4_0, \
call_10_4_4_1, \
call_10_4_4_2, \
call_10_4_4_3, \
call_10_4_4_4, \
call_10_4_4_5, \
call_10_4_4_6, \
call_10_4_4_7, \
call_10_4_4_8, \
call_10_4_4_9, \
call_10_4_4_10, \
call_10_4_4_11, \
call_10_4_4_12, \
call_10_4_4_13, \
call_10_4_4_14, \
call_10_4_4_15, \
call_10_4_5_0, \
call_10_4_5_1, \
call_10_4_5_2, \
call_10_4_5_3, \
call_10_4_5_4, \
call_10_4_5_5, \
call_10_4_5_6, \
call_10_4_5_7, \
call_10_4_5_8, \
call_10_4_5_9, \
call_10_4_5_10, \
call_10_4_5_11, \
call_10_4_5_12, \
call_10_4_5_13, \
call_10_4_5_14, \
call_10_4_5_15, \
call_10_4_6_0, \
call_10_4_6_1, \
call_10_4_6_2, \
call_10_4_6_3, \
call_10_4_6_4, \
call_10_4_6_5, \
call_10_4_6_6, \
call_10_4_6_7, \
call_10_4_6_8, \
call_10_4_6_9, \
call_10_4_6_10, \
call_10_4_6_11, \
call_10_4_6_12, \
call_10_4_6_13, \
call_10_4_6_14, \
call_10_4_6_15, \
call_10_4_7_0, \
call_10_4_7_1, \
call_10_4_7_2, \
call_10_4_7_3, \
call_10_4_7_4, \
call_10_4_7_5, \
call_10_4_7_6, \
call_10_4_7_7, \
call_10_4_7_8, \
call_10_4_7_9, \
call_10_4_7_10, \
call_10_4_7_11, \
call_10_4_7_12, \
call_10_4_7_13, \
call_10_4_7_14, \
call_10_4_7_15, \
call_10_4_8_0, \
call_10_4_8_1, \
call_10_4_8_2, \
call_10_4_8_3, \
call_10_4_8_4, \
call_10_4_8_5, \
call_10_4_8_6, \
call_10_4_8_7, \
call_10_4_8_8, \
call_10_4_8_9, \
call_10_4_8_10, \
call_10_4_8_11, \
call_10_4_8_12, \
call_10_4_8_13, \
call_10_4_8_14, \
call_10_4_8_15, \
call_10_4_9_0, \
call_10_4_9_1, \
call_10_4_9_2, \
call_10_4_9_3, \
call_10_4_9_4, \
call_10_4_9_5, \
call_10_4_9_6, \
call_10_4_9_7, \
call_10_4_9_8, \
call_10_4_9_9, \
call_10_4_9_10, \
call_10_4_9_11, \
call_10_4_9_12, \
call_10_4_9_13, \
call_10_4_9_14, \
call_10_4_9_15, \
call_10_4_10_0, \
call_10_4_10_1, \
call_10_4_10_2, \
call_10_4_10_3, \
call_10_4_10_4, \
call_10_4_10_5, \
call_10_4_10_6, \
call_10_4_10_7, \
call_10_4_10_8, \
call_10_4_10_9, \
call_10_4_10_10, \
call_10_4_10_11, \
call_10_4_10_12, \
call_10_4_10_13, \
call_10_4_10_14, \
call_10_4_10_15, \
call_10_4_11_0, \
call_10_4_11_1, \
call_10_4_11_2, \
call_10_4_11_3, \
call_10_4_11_4, \
call_10_4_11_5, \
call_10_4_11_6, \
call_10_4_11_7, \
call_10_4_11_8, \
call_10_4_11_9, \
call_10_4_11_10, \
call_10_4_11_11, \
call_10_4_11_12, \
call_10_4_11_13, \
call_10_4_11_14, \
call_10_4_11_15, \
call_10_4_12_0, \
call_10_4_12_1, \
call_10_4_12_2, \
call_10_4_12_3, \
call_10_4_12_4, \
call_10_4_12_5, \
call_10_4_12_6, \
call_10_4_12_7, \
call_10_4_12_8, \
call_10_4_12_9, \
call_10_4_12_10, \
call_10_4_12_11, \
call_10_4_12_12, \
call_10_4_12_13, \
call_10_4_12_14, \
call_10_4_12_15, \
call_10_4_13_0, \
call_10_4_13_1, \
call_10_4_13_2, \
call_10_4_13_3, \
call_10_4_13_4, \
call_10_4_13_5, \
call_10_4_13_6, \
call_10_4_13_7, \
call_10_4_13_8, \
call_10_4_13_9, \
call_10_4_13_10, \
call_10_4_13_11, \
call_10_4_13_12, \
call_10_4_13_13, \
call_10_4_13_14, \
call_10_4_13_15, \
call_10_4_14_0, \
call_10_4_14_1, \
call_10_4_14_2, \
call_10_4_14_3, \
call_10_4_14_4, \
call_10_4_14_5, \
call_10_4_14_6, \
call_10_4_14_7, \
call_10_4_14_8, \
call_10_4_14_9, \
call_10_4_14_10, \
call_10_4_14_11, \
call_10_4_14_12, \
call_10_4_14_13, \
call_10_4_14_14, \
call_10_4_14_15, \
call_10_4_15_0, \
call_10_4_15_1, \
call_10_4_15_2, \
call_10_4_15_3, \
call_10_4_15_4, \
call_10_4_15_5, \
call_10_4_15_6, \
call_10_4_15_7, \
call_10_4_15_8, \
call_10_4_15_9, \
call_10_4_15_10, \
call_10_4_15_11, \
call_10_4_15_12, \
call_10_4_15_13, \
call_10_4_15_14, \
call_10_4_15_15, \
call_10_5_0_0, \
call_10_5_0_1, \
call_10_5_0_2, \
call_10_5_0_3, \
call_10_5_0_4, \
call_10_5_0_5, \
call_10_5_0_6, \
call_10_5_0_7, \
call_10_5_0_8, \
call_10_5_0_9, \
call_10_5_0_10, \
call_10_5_0_11, \
call_10_5_0_12, \
call_10_5_0_13, \
call_10_5_0_14, \
call_10_5_0_15, \
call_10_5_1_0, \
call_10_5_1_1, \
call_10_5_1_2, \
call_10_5_1_3, \
call_10_5_1_4, \
call_10_5_1_5, \
call_10_5_1_6, \
call_10_5_1_7, \
call_10_5_1_8, \
call_10_5_1_9, \
call_10_5_1_10, \
call_10_5_1_11, \
call_10_5_1_12, \
call_10_5_1_13, \
call_10_5_1_14, \
call_10_5_1_15, \
call_10_5_2_0, \
call_10_5_2_1, \
call_10_5_2_2, \
call_10_5_2_3, \
call_10_5_2_4, \
call_10_5_2_5, \
call_10_5_2_6, \
call_10_5_2_7, \
call_10_5_2_8, \
call_10_5_2_9, \
call_10_5_2_10, \
call_10_5_2_11, \
call_10_5_2_12, \
call_10_5_2_13, \
call_10_5_2_14, \
call_10_5_2_15, \
call_10_5_3_0, \
call_10_5_3_1, \
call_10_5_3_2, \
call_10_5_3_3, \
call_10_5_3_4, \
call_10_5_3_5, \
call_10_5_3_6, \
call_10_5_3_7, \
call_10_5_3_8, \
call_10_5_3_9, \
call_10_5_3_10, \
call_10_5_3_11, \
call_10_5_3_12, \
call_10_5_3_13, \
call_10_5_3_14, \
call_10_5_3_15, \
call_10_5_4_0, \
call_10_5_4_1, \
call_10_5_4_2, \
call_10_5_4_3, \
call_10_5_4_4, \
call_10_5_4_5, \
call_10_5_4_6, \
call_10_5_4_7, \
call_10_5_4_8, \
call_10_5_4_9, \
call_10_5_4_10, \
call_10_5_4_11, \
call_10_5_4_12, \
call_10_5_4_13, \
call_10_5_4_14, \
call_10_5_4_15, \
call_10_5_5_0, \
call_10_5_5_1, \
call_10_5_5_2, \
call_10_5_5_3, \
call_10_5_5_4, \
call_10_5_5_5, \
call_10_5_5_6, \
call_10_5_5_7, \
call_10_5_5_8, \
call_10_5_5_9, \
call_10_5_5_10, \
call_10_5_5_11, \
call_10_5_5_12, \
call_10_5_5_13, \
call_10_5_5_14, \
call_10_5_5_15, \
call_10_5_6_0, \
call_10_5_6_1, \
call_10_5_6_2, \
call_10_5_6_3, \
call_10_5_6_4, \
call_10_5_6_5, \
call_10_5_6_6, \
call_10_5_6_7, \
call_10_5_6_8, \
call_10_5_6_9, \
call_10_5_6_10, \
call_10_5_6_11, \
call_10_5_6_12, \
call_10_5_6_13, \
call_10_5_6_14, \
call_10_5_6_15, \
call_10_5_7_0, \
call_10_5_7_1, \
call_10_5_7_2, \
call_10_5_7_3, \
call_10_5_7_4, \
call_10_5_7_5, \
call_10_5_7_6, \
call_10_5_7_7, \
call_10_5_7_8, \
call_10_5_7_9, \
call_10_5_7_10, \
call_10_5_7_11, \
call_10_5_7_12, \
call_10_5_7_13, \
call_10_5_7_14, \
call_10_5_7_15, \
call_10_5_8_0, \
call_10_5_8_1, \
call_10_5_8_2, \
call_10_5_8_3, \
call_10_5_8_4, \
call_10_5_8_5, \
call_10_5_8_6, \
call_10_5_8_7, \
call_10_5_8_8, \
call_10_5_8_9, \
call_10_5_8_10, \
call_10_5_8_11, \
call_10_5_8_12, \
call_10_5_8_13, \
call_10_5_8_14, \
call_10_5_8_15, \
call_10_5_9_0, \
call_10_5_9_1, \
call_10_5_9_2, \
call_10_5_9_3, \
call_10_5_9_4, \
call_10_5_9_5, \
call_10_5_9_6, \
call_10_5_9_7, \
call_10_5_9_8, \
call_10_5_9_9, \
call_10_5_9_10, \
call_10_5_9_11, \
call_10_5_9_12, \
call_10_5_9_13, \
call_10_5_9_14, \
call_10_5_9_15, \
call_10_5_10_0, \
call_10_5_10_1, \
call_10_5_10_2, \
call_10_5_10_3, \
call_10_5_10_4, \
call_10_5_10_5, \
call_10_5_10_6, \
call_10_5_10_7, \
call_10_5_10_8, \
call_10_5_10_9, \
call_10_5_10_10, \
call_10_5_10_11, \
call_10_5_10_12, \
call_10_5_10_13, \
call_10_5_10_14, \
call_10_5_10_15, \
call_10_5_11_0, \
call_10_5_11_1, \
call_10_5_11_2, \
call_10_5_11_3, \
call_10_5_11_4, \
call_10_5_11_5, \
call_10_5_11_6, \
call_10_5_11_7, \
call_10_5_11_8, \
call_10_5_11_9, \
call_10_5_11_10, \
call_10_5_11_11, \
call_10_5_11_12, \
call_10_5_11_13, \
call_10_5_11_14, \
call_10_5_11_15, \
call_10_5_12_0, \
call_10_5_12_1, \
call_10_5_12_2, \
call_10_5_12_3, \
call_10_5_12_4, \
call_10_5_12_5, \
call_10_5_12_6, \
call_10_5_12_7, \
call_10_5_12_8, \
call_10_5_12_9, \
call_10_5_12_10, \
call_10_5_12_11, \
call_10_5_12_12, \
call_10_5_12_13, \
call_10_5_12_14, \
call_10_5_12_15, \
call_10_5_13_0, \
call_10_5_13_1, \
call_10_5_13_2, \
call_10_5_13_3, \
call_10_5_13_4, \
call_10_5_13_5, \
call_10_5_13_6, \
call_10_5_13_7, \
call_10_5_13_8, \
call_10_5_13_9, \
call_10_5_13_10, \
call_10_5_13_11, \
call_10_5_13_12, \
call_10_5_13_13, \
call_10_5_13_14, \
call_10_5_13_15, \
call_10_5_14_0, \
call_10_5_14_1, \
call_10_5_14_2, \
call_10_5_14_3, \
call_10_5_14_4, \
call_10_5_14_5, \
call_10_5_14_6, \
call_10_5_14_7, \
call_10_5_14_8, \
call_10_5_14_9, \
call_10_5_14_10, \
call_10_5_14_11, \
call_10_5_14_12, \
call_10_5_14_13, \
call_10_5_14_14, \
call_10_5_14_15, \
call_10_5_15_0, \
call_10_5_15_1, \
call_10_5_15_2, \
call_10_5_15_3, \
call_10_5_15_4, \
call_10_5_15_5, \
call_10_5_15_6, \
call_10_5_15_7, \
call_10_5_15_8, \
call_10_5_15_9, \
call_10_5_15_10, \
call_10_5_15_11, \
call_10_5_15_12, \
call_10_5_15_13, \
call_10_5_15_14, \
call_10_5_15_15, \
call_10_6_0_0, \
call_10_6_0_1, \
call_10_6_0_2, \
call_10_6_0_3, \
call_10_6_0_4, \
call_10_6_0_5, \
call_10_6_0_6, \
call_10_6_0_7, \
call_10_6_0_8, \
call_10_6_0_9, \
call_10_6_0_10, \
call_10_6_0_11, \
call_10_6_0_12, \
call_10_6_0_13, \
call_10_6_0_14, \
call_10_6_0_15, \
call_10_6_1_0, \
call_10_6_1_1, \
call_10_6_1_2, \
call_10_6_1_3, \
call_10_6_1_4, \
call_10_6_1_5, \
call_10_6_1_6, \
call_10_6_1_7, \
call_10_6_1_8, \
call_10_6_1_9, \
call_10_6_1_10, \
call_10_6_1_11, \
call_10_6_1_12, \
call_10_6_1_13, \
call_10_6_1_14, \
call_10_6_1_15, \
call_10_6_2_0, \
call_10_6_2_1, \
call_10_6_2_2, \
call_10_6_2_3, \
call_10_6_2_4, \
call_10_6_2_5, \
call_10_6_2_6, \
call_10_6_2_7, \
call_10_6_2_8, \
call_10_6_2_9, \
call_10_6_2_10, \
call_10_6_2_11, \
call_10_6_2_12, \
call_10_6_2_13, \
call_10_6_2_14, \
call_10_6_2_15, \
call_10_6_3_0, \
call_10_6_3_1, \
call_10_6_3_2, \
call_10_6_3_3, \
call_10_6_3_4, \
call_10_6_3_5, \
call_10_6_3_6, \
call_10_6_3_7, \
call_10_6_3_8, \
call_10_6_3_9, \
call_10_6_3_10, \
call_10_6_3_11, \
call_10_6_3_12, \
call_10_6_3_13, \
call_10_6_3_14, \
call_10_6_3_15, \
call_10_6_4_0, \
call_10_6_4_1, \
call_10_6_4_2, \
call_10_6_4_3, \
call_10_6_4_4, \
call_10_6_4_5, \
call_10_6_4_6, \
call_10_6_4_7, \
call_10_6_4_8, \
call_10_6_4_9, \
call_10_6_4_10, \
call_10_6_4_11, \
call_10_6_4_12, \
call_10_6_4_13, \
call_10_6_4_14, \
call_10_6_4_15, \
call_10_6_5_0, \
call_10_6_5_1, \
call_10_6_5_2, \
call_10_6_5_3, \
call_10_6_5_4, \
call_10_6_5_5, \
call_10_6_5_6, \
call_10_6_5_7, \
call_10_6_5_8, \
call_10_6_5_9, \
call_10_6_5_10, \
call_10_6_5_11, \
call_10_6_5_12, \
call_10_6_5_13, \
call_10_6_5_14, \
call_10_6_5_15, \
call_10_6_6_0, \
call_10_6_6_1, \
call_10_6_6_2, \
call_10_6_6_3, \
call_10_6_6_4, \
call_10_6_6_5, \
call_10_6_6_6, \
call_10_6_6_7, \
call_10_6_6_8, \
call_10_6_6_9, \
call_10_6_6_10, \
call_10_6_6_11, \
call_10_6_6_12, \
call_10_6_6_13, \
call_10_6_6_14, \
call_10_6_6_15, \
call_10_6_7_0, \
call_10_6_7_1, \
call_10_6_7_2, \
call_10_6_7_3, \
call_10_6_7_4, \
call_10_6_7_5, \
call_10_6_7_6, \
call_10_6_7_7, \
call_10_6_7_8, \
call_10_6_7_9, \
call_10_6_7_10, \
call_10_6_7_11, \
call_10_6_7_12, \
call_10_6_7_13, \
call_10_6_7_14, \
call_10_6_7_15, \
call_10_6_8_0, \
call_10_6_8_1, \
call_10_6_8_2, \
call_10_6_8_3, \
call_10_6_8_4, \
call_10_6_8_5, \
call_10_6_8_6, \
call_10_6_8_7, \
call_10_6_8_8, \
call_10_6_8_9, \
call_10_6_8_10, \
call_10_6_8_11, \
call_10_6_8_12, \
call_10_6_8_13, \
call_10_6_8_14, \
call_10_6_8_15, \
call_10_6_9_0, \
call_10_6_9_1, \
call_10_6_9_2, \
call_10_6_9_3, \
call_10_6_9_4, \
call_10_6_9_5, \
call_10_6_9_6, \
call_10_6_9_7, \
call_10_6_9_8, \
call_10_6_9_9, \
call_10_6_9_10, \
call_10_6_9_11, \
call_10_6_9_12, \
call_10_6_9_13, \
call_10_6_9_14, \
call_10_6_9_15, \
call_10_6_10_0, \
call_10_6_10_1, \
call_10_6_10_2, \
call_10_6_10_3, \
call_10_6_10_4, \
call_10_6_10_5, \
call_10_6_10_6, \
call_10_6_10_7, \
call_10_6_10_8, \
call_10_6_10_9, \
call_10_6_10_10, \
call_10_6_10_11, \
call_10_6_10_12, \
call_10_6_10_13, \
call_10_6_10_14, \
call_10_6_10_15, \
call_10_6_11_0, \
call_10_6_11_1, \
call_10_6_11_2, \
call_10_6_11_3, \
call_10_6_11_4, \
call_10_6_11_5, \
call_10_6_11_6, \
call_10_6_11_7, \
call_10_6_11_8, \
call_10_6_11_9, \
call_10_6_11_10, \
call_10_6_11_11, \
call_10_6_11_12, \
call_10_6_11_13, \
call_10_6_11_14, \
call_10_6_11_15, \
call_10_6_12_0, \
call_10_6_12_1, \
call_10_6_12_2, \
call_10_6_12_3, \
call_10_6_12_4, \
call_10_6_12_5, \
call_10_6_12_6, \
call_10_6_12_7, \
call_10_6_12_8, \
call_10_6_12_9, \
call_10_6_12_10, \
call_10_6_12_11, \
call_10_6_12_12, \
call_10_6_12_13, \
call_10_6_12_14, \
call_10_6_12_15, \
call_10_6_13_0, \
call_10_6_13_1, \
call_10_6_13_2, \
call_10_6_13_3, \
call_10_6_13_4, \
call_10_6_13_5, \
call_10_6_13_6, \
call_10_6_13_7, \
call_10_6_13_8, \
call_10_6_13_9, \
call_10_6_13_10, \
call_10_6_13_11, \
call_10_6_13_12, \
call_10_6_13_13, \
call_10_6_13_14, \
call_10_6_13_15, \
call_10_6_14_0, \
call_10_6_14_1, \
call_10_6_14_2, \
call_10_6_14_3, \
call_10_6_14_4, \
call_10_6_14_5, \
call_10_6_14_6, \
call_10_6_14_7, \
call_10_6_14_8, \
call_10_6_14_9, \
call_10_6_14_10, \
call_10_6_14_11, \
call_10_6_14_12, \
call_10_6_14_13, \
call_10_6_14_14, \
call_10_6_14_15, \
call_10_6_15_0, \
call_10_6_15_1, \
call_10_6_15_2, \
call_10_6_15_3, \
call_10_6_15_4, \
call_10_6_15_5, \
call_10_6_15_6, \
call_10_6_15_7, \
call_10_6_15_8, \
call_10_6_15_9, \
call_10_6_15_10, \
call_10_6_15_11, \
call_10_6_15_12, \
call_10_6_15_13, \
call_10_6_15_14, \
call_10_6_15_15, \
call_10_7_0_0, \
call_10_7_0_1, \
call_10_7_0_2, \
call_10_7_0_3, \
call_10_7_0_4, \
call_10_7_0_5, \
call_10_7_0_6, \
call_10_7_0_7, \
call_10_7_0_8, \
call_10_7_0_9, \
call_10_7_0_10, \
call_10_7_0_11, \
call_10_7_0_12, \
call_10_7_0_13, \
call_10_7_0_14, \
call_10_7_0_15, \
call_10_7_1_0, \
call_10_7_1_1, \
call_10_7_1_2, \
call_10_7_1_3, \
call_10_7_1_4, \
call_10_7_1_5, \
call_10_7_1_6, \
call_10_7_1_7, \
call_10_7_1_8, \
call_10_7_1_9, \
call_10_7_1_10, \
call_10_7_1_11, \
call_10_7_1_12, \
call_10_7_1_13, \
call_10_7_1_14, \
call_10_7_1_15, \
call_10_7_2_0, \
call_10_7_2_1, \
call_10_7_2_2, \
call_10_7_2_3, \
call_10_7_2_4, \
call_10_7_2_5, \
call_10_7_2_6, \
call_10_7_2_7, \
call_10_7_2_8, \
call_10_7_2_9, \
call_10_7_2_10, \
call_10_7_2_11, \
call_10_7_2_12, \
call_10_7_2_13, \
call_10_7_2_14, \
call_10_7_2_15, \
call_10_7_3_0, \
call_10_7_3_1, \
call_10_7_3_2, \
call_10_7_3_3, \
call_10_7_3_4, \
call_10_7_3_5, \
call_10_7_3_6, \
call_10_7_3_7, \
call_10_7_3_8, \
call_10_7_3_9, \
call_10_7_3_10, \
call_10_7_3_11, \
call_10_7_3_12, \
call_10_7_3_13, \
call_10_7_3_14, \
call_10_7_3_15, \
call_10_7_4_0, \
call_10_7_4_1, \
call_10_7_4_2, \
call_10_7_4_3, \
call_10_7_4_4, \
call_10_7_4_5, \
call_10_7_4_6, \
call_10_7_4_7, \
call_10_7_4_8, \
call_10_7_4_9, \
call_10_7_4_10, \
call_10_7_4_11, \
call_10_7_4_12, \
call_10_7_4_13, \
call_10_7_4_14, \
call_10_7_4_15, \
call_10_7_5_0, \
call_10_7_5_1, \
call_10_7_5_2, \
call_10_7_5_3, \
call_10_7_5_4, \
call_10_7_5_5, \
call_10_7_5_6, \
call_10_7_5_7, \
call_10_7_5_8, \
call_10_7_5_9, \
call_10_7_5_10, \
call_10_7_5_11, \
call_10_7_5_12, \
call_10_7_5_13, \
call_10_7_5_14, \
call_10_7_5_15, \
call_10_7_6_0, \
call_10_7_6_1, \
call_10_7_6_2, \
call_10_7_6_3, \
call_10_7_6_4, \
call_10_7_6_5, \
call_10_7_6_6, \
call_10_7_6_7, \
call_10_7_6_8, \
call_10_7_6_9, \
call_10_7_6_10, \
call_10_7_6_11, \
call_10_7_6_12, \
call_10_7_6_13, \
call_10_7_6_14, \
call_10_7_6_15, \
call_10_7_7_0, \
call_10_7_7_1, \
call_10_7_7_2, \
call_10_7_7_3, \
call_10_7_7_4, \
call_10_7_7_5, \
call_10_7_7_6, \
call_10_7_7_7, \
call_10_7_7_8, \
call_10_7_7_9, \
call_10_7_7_10, \
call_10_7_7_11, \
call_10_7_7_12, \
call_10_7_7_13, \
call_10_7_7_14, \
call_10_7_7_15, \
call_10_7_8_0, \
call_10_7_8_1, \
call_10_7_8_2, \
call_10_7_8_3, \
call_10_7_8_4, \
call_10_7_8_5, \
call_10_7_8_6, \
call_10_7_8_7, \
call_10_7_8_8, \
call_10_7_8_9, \
call_10_7_8_10, \
call_10_7_8_11, \
call_10_7_8_12, \
call_10_7_8_13, \
call_10_7_8_14, \
call_10_7_8_15, \
call_10_7_9_0, \
call_10_7_9_1, \
call_10_7_9_2, \
call_10_7_9_3, \
call_10_7_9_4, \
call_10_7_9_5, \
call_10_7_9_6, \
call_10_7_9_7, \
call_10_7_9_8, \
call_10_7_9_9, \
call_10_7_9_10, \
call_10_7_9_11, \
call_10_7_9_12, \
call_10_7_9_13, \
call_10_7_9_14, \
call_10_7_9_15, \
call_10_7_10_0, \
call_10_7_10_1, \
call_10_7_10_2, \
call_10_7_10_3, \
call_10_7_10_4, \
call_10_7_10_5, \
call_10_7_10_6, \
call_10_7_10_7, \
call_10_7_10_8, \
call_10_7_10_9, \
call_10_7_10_10, \
call_10_7_10_11, \
call_10_7_10_12, \
call_10_7_10_13, \
call_10_7_10_14, \
call_10_7_10_15, \
call_10_7_11_0, \
call_10_7_11_1, \
call_10_7_11_2, \
call_10_7_11_3, \
call_10_7_11_4, \
call_10_7_11_5, \
call_10_7_11_6, \
call_10_7_11_7, \
call_10_7_11_8, \
call_10_7_11_9, \
call_10_7_11_10, \
call_10_7_11_11, \
call_10_7_11_12, \
call_10_7_11_13, \
call_10_7_11_14, \
call_10_7_11_15, \
call_10_7_12_0, \
call_10_7_12_1, \
call_10_7_12_2, \
call_10_7_12_3, \
call_10_7_12_4, \
call_10_7_12_5, \
call_10_7_12_6, \
call_10_7_12_7, \
call_10_7_12_8, \
call_10_7_12_9, \
call_10_7_12_10, \
call_10_7_12_11, \
call_10_7_12_12, \
call_10_7_12_13, \
call_10_7_12_14, \
call_10_7_12_15, \
call_10_7_13_0, \
call_10_7_13_1, \
call_10_7_13_2, \
call_10_7_13_3, \
call_10_7_13_4, \
call_10_7_13_5, \
call_10_7_13_6, \
call_10_7_13_7, \
call_10_7_13_8, \
call_10_7_13_9, \
call_10_7_13_10, \
call_10_7_13_11, \
call_10_7_13_12, \
call_10_7_13_13, \
call_10_7_13_14, \
call_10_7_13_15, \
call_10_7_14_0, \
call_10_7_14_1, \
call_10_7_14_2, \
call_10_7_14_3, \
call_10_7_14_4, \
call_10_7_14_5, \
call_10_7_14_6, \
call_10_7_14_7, \
call_10_7_14_8, \
call_10_7_14_9, \
call_10_7_14_10, \
call_10_7_14_11, \
call_10_7_14_12, \
call_10_7_14_13, \
call_10_7_14_14, \
call_10_7_14_15, \
call_10_7_15_0, \
call_10_7_15_1, \
call_10_7_15_2, \
call_10_7_15_3, \
call_10_7_15_4, \
call_10_7_15_5, \
call_10_7_15_6, \
call_10_7_15_7, \
call_10_7_15_8, \
call_10_7_15_9, \
call_10_7_15_10, \
call_10_7_15_11, \
call_10_7_15_12, \
call_10_7_15_13, \
call_10_7_15_14, \
call_10_7_15_15, \
call_10_8_0_0, \
call_10_8_0_1, \
call_10_8_0_2, \
call_10_8_0_3, \
call_10_8_0_4, \
call_10_8_0_5, \
call_10_8_0_6, \
call_10_8_0_7, \
call_10_8_0_8, \
call_10_8_0_9, \
call_10_8_0_10, \
call_10_8_0_11, \
call_10_8_0_12, \
call_10_8_0_13, \
call_10_8_0_14, \
call_10_8_0_15, \
call_10_8_1_0, \
call_10_8_1_1, \
call_10_8_1_2, \
call_10_8_1_3, \
call_10_8_1_4, \
call_10_8_1_5, \
call_10_8_1_6, \
call_10_8_1_7, \
call_10_8_1_8, \
call_10_8_1_9, \
call_10_8_1_10, \
call_10_8_1_11, \
call_10_8_1_12, \
call_10_8_1_13, \
call_10_8_1_14, \
call_10_8_1_15, \
call_10_8_2_0, \
call_10_8_2_1, \
call_10_8_2_2, \
call_10_8_2_3, \
call_10_8_2_4, \
call_10_8_2_5, \
call_10_8_2_6, \
call_10_8_2_7, \
call_10_8_2_8, \
call_10_8_2_9, \
call_10_8_2_10, \
call_10_8_2_11, \
call_10_8_2_12, \
call_10_8_2_13, \
call_10_8_2_14, \
call_10_8_2_15, \
call_10_8_3_0, \
call_10_8_3_1, \
call_10_8_3_2, \
call_10_8_3_3, \
call_10_8_3_4, \
call_10_8_3_5, \
call_10_8_3_6, \
call_10_8_3_7, \
call_10_8_3_8, \
call_10_8_3_9, \
call_10_8_3_10, \
call_10_8_3_11, \
call_10_8_3_12, \
call_10_8_3_13, \
call_10_8_3_14, \
call_10_8_3_15, \
call_10_8_4_0, \
call_10_8_4_1, \
call_10_8_4_2, \
call_10_8_4_3, \
call_10_8_4_4, \
call_10_8_4_5, \
call_10_8_4_6, \
call_10_8_4_7, \
call_10_8_4_8, \
call_10_8_4_9, \
call_10_8_4_10, \
call_10_8_4_11, \
call_10_8_4_12, \
call_10_8_4_13, \
call_10_8_4_14, \
call_10_8_4_15, \
call_10_8_5_0, \
call_10_8_5_1, \
call_10_8_5_2, \
call_10_8_5_3, \
call_10_8_5_4, \
call_10_8_5_5, \
call_10_8_5_6, \
call_10_8_5_7, \
call_10_8_5_8, \
call_10_8_5_9, \
call_10_8_5_10, \
call_10_8_5_11, \
call_10_8_5_12, \
call_10_8_5_13, \
call_10_8_5_14, \
call_10_8_5_15, \
call_10_8_6_0, \
call_10_8_6_1, \
call_10_8_6_2, \
call_10_8_6_3, \
call_10_8_6_4, \
call_10_8_6_5, \
call_10_8_6_6, \
call_10_8_6_7, \
call_10_8_6_8, \
call_10_8_6_9, \
call_10_8_6_10, \
call_10_8_6_11, \
call_10_8_6_12, \
call_10_8_6_13, \
call_10_8_6_14, \
call_10_8_6_15, \
call_10_8_7_0, \
call_10_8_7_1, \
call_10_8_7_2, \
call_10_8_7_3, \
call_10_8_7_4, \
call_10_8_7_5, \
call_10_8_7_6, \
call_10_8_7_7, \
call_10_8_7_8, \
call_10_8_7_9, \
call_10_8_7_10, \
call_10_8_7_11, \
call_10_8_7_12, \
call_10_8_7_13, \
call_10_8_7_14, \
call_10_8_7_15, \
call_10_8_8_0, \
call_10_8_8_1, \
call_10_8_8_2, \
call_10_8_8_3, \
call_10_8_8_4, \
call_10_8_8_5, \
call_10_8_8_6, \
call_10_8_8_7, \
call_10_8_8_8, \
call_10_8_8_9, \
call_10_8_8_10, \
call_10_8_8_11, \
call_10_8_8_12, \
call_10_8_8_13, \
call_10_8_8_14, \
call_10_8_8_15, \
call_10_8_9_0, \
call_10_8_9_1, \
call_10_8_9_2, \
call_10_8_9_3, \
call_10_8_9_4, \
call_10_8_9_5, \
call_10_8_9_6, \
call_10_8_9_7, \
call_10_8_9_8, \
call_10_8_9_9, \
call_10_8_9_10, \
call_10_8_9_11, \
call_10_8_9_12, \
call_10_8_9_13, \
call_10_8_9_14, \
call_10_8_9_15, \
call_10_8_10_0, \
call_10_8_10_1, \
call_10_8_10_2, \
call_10_8_10_3, \
call_10_8_10_4, \
call_10_8_10_5, \
call_10_8_10_6, \
call_10_8_10_7, \
call_10_8_10_8, \
call_10_8_10_9, \
call_10_8_10_10, \
call_10_8_10_11, \
call_10_8_10_12, \
call_10_8_10_13, \
call_10_8_10_14, \
call_10_8_10_15, \
call_10_8_11_0, \
call_10_8_11_1, \
call_10_8_11_2, \
call_10_8_11_3, \
call_10_8_11_4, \
call_10_8_11_5, \
call_10_8_11_6, \
call_10_8_11_7, \
call_10_8_11_8, \
call_10_8_11_9, \
call_10_8_11_10, \
call_10_8_11_11, \
call_10_8_11_12, \
call_10_8_11_13, \
call_10_8_11_14, \
call_10_8_11_15, \
call_10_8_12_0, \
call_10_8_12_1, \
call_10_8_12_2, \
call_10_8_12_3, \
call_10_8_12_4, \
call_10_8_12_5, \
call_10_8_12_6, \
call_10_8_12_7, \
call_10_8_12_8, \
call_10_8_12_9, \
call_10_8_12_10, \
call_10_8_12_11, \
call_10_8_12_12, \
call_10_8_12_13, \
call_10_8_12_14, \
call_10_8_12_15, \
call_10_8_13_0, \
call_10_8_13_1, \
call_10_8_13_2, \
call_10_8_13_3, \
call_10_8_13_4, \
call_10_8_13_5, \
call_10_8_13_6, \
call_10_8_13_7, \
call_10_8_13_8, \
call_10_8_13_9, \
call_10_8_13_10, \
call_10_8_13_11, \
call_10_8_13_12, \
call_10_8_13_13, \
call_10_8_13_14, \
call_10_8_13_15, \
call_10_8_14_0, \
call_10_8_14_1, \
call_10_8_14_2, \
call_10_8_14_3, \
call_10_8_14_4, \
call_10_8_14_5, \
call_10_8_14_6, \
call_10_8_14_7, \
call_10_8_14_8, \
call_10_8_14_9, \
call_10_8_14_10, \
call_10_8_14_11, \
call_10_8_14_12, \
call_10_8_14_13, \
call_10_8_14_14, \
call_10_8_14_15, \
call_10_8_15_0, \
call_10_8_15_1, \
call_10_8_15_2, \
call_10_8_15_3, \
call_10_8_15_4, \
call_10_8_15_5, \
call_10_8_15_6, \
call_10_8_15_7, \
call_10_8_15_8, \
call_10_8_15_9, \
call_10_8_15_10, \
call_10_8_15_11, \
call_10_8_15_12, \
call_10_8_15_13, \
call_10_8_15_14, \
call_10_8_15_15, \
call_10_9_0_0, \
call_10_9_0_1, \
call_10_9_0_2, \
call_10_9_0_3, \
call_10_9_0_4, \
call_10_9_0_5, \
call_10_9_0_6, \
call_10_9_0_7, \
call_10_9_0_8, \
call_10_9_0_9, \
call_10_9_0_10, \
call_10_9_0_11, \
call_10_9_0_12, \
call_10_9_0_13, \
call_10_9_0_14, \
call_10_9_0_15, \
call_10_9_1_0, \
call_10_9_1_1, \
call_10_9_1_2, \
call_10_9_1_3, \
call_10_9_1_4, \
call_10_9_1_5, \
call_10_9_1_6, \
call_10_9_1_7, \
call_10_9_1_8, \
call_10_9_1_9, \
call_10_9_1_10, \
call_10_9_1_11, \
call_10_9_1_12, \
call_10_9_1_13, \
call_10_9_1_14, \
call_10_9_1_15, \
call_10_9_2_0, \
call_10_9_2_1, \
call_10_9_2_2, \
call_10_9_2_3, \
call_10_9_2_4, \
call_10_9_2_5, \
call_10_9_2_6, \
call_10_9_2_7, \
call_10_9_2_8, \
call_10_9_2_9, \
call_10_9_2_10, \
call_10_9_2_11, \
call_10_9_2_12, \
call_10_9_2_13, \
call_10_9_2_14, \
call_10_9_2_15, \
call_10_9_3_0, \
call_10_9_3_1, \
call_10_9_3_2, \
call_10_9_3_3, \
call_10_9_3_4, \
call_10_9_3_5, \
call_10_9_3_6, \
call_10_9_3_7, \
call_10_9_3_8, \
call_10_9_3_9, \
call_10_9_3_10, \
call_10_9_3_11, \
call_10_9_3_12, \
call_10_9_3_13, \
call_10_9_3_14, \
call_10_9_3_15, \
call_10_9_4_0, \
call_10_9_4_1, \
call_10_9_4_2, \
call_10_9_4_3, \
call_10_9_4_4, \
call_10_9_4_5, \
call_10_9_4_6, \
call_10_9_4_7, \
call_10_9_4_8, \
call_10_9_4_9, \
call_10_9_4_10, \
call_10_9_4_11, \
call_10_9_4_12, \
call_10_9_4_13, \
call_10_9_4_14, \
call_10_9_4_15, \
call_10_9_5_0, \
call_10_9_5_1, \
call_10_9_5_2, \
call_10_9_5_3, \
call_10_9_5_4, \
call_10_9_5_5, \
call_10_9_5_6, \
call_10_9_5_7, \
call_10_9_5_8, \
call_10_9_5_9, \
call_10_9_5_10, \
call_10_9_5_11, \
call_10_9_5_12, \
call_10_9_5_13, \
call_10_9_5_14, \
call_10_9_5_15, \
call_10_9_6_0, \
call_10_9_6_1, \
call_10_9_6_2, \
call_10_9_6_3, \
call_10_9_6_4, \
call_10_9_6_5, \
call_10_9_6_6, \
call_10_9_6_7, \
call_10_9_6_8, \
call_10_9_6_9, \
call_10_9_6_10, \
call_10_9_6_11, \
call_10_9_6_12, \
call_10_9_6_13, \
call_10_9_6_14, \
call_10_9_6_15, \
call_10_9_7_0, \
call_10_9_7_1, \
call_10_9_7_2, \
call_10_9_7_3, \
call_10_9_7_4, \
call_10_9_7_5, \
call_10_9_7_6, \
call_10_9_7_7, \
call_10_9_7_8, \
call_10_9_7_9, \
call_10_9_7_10, \
call_10_9_7_11, \
call_10_9_7_12, \
call_10_9_7_13, \
call_10_9_7_14, \
call_10_9_7_15, \
call_10_9_8_0, \
call_10_9_8_1, \
call_10_9_8_2, \
call_10_9_8_3, \
call_10_9_8_4, \
call_10_9_8_5, \
call_10_9_8_6, \
call_10_9_8_7, \
call_10_9_8_8, \
call_10_9_8_9, \
call_10_9_8_10, \
call_10_9_8_11, \
call_10_9_8_12, \
call_10_9_8_13, \
call_10_9_8_14, \
call_10_9_8_15, \
call_10_9_9_0, \
call_10_9_9_1, \
call_10_9_9_2, \
call_10_9_9_3, \
call_10_9_9_4, \
call_10_9_9_5, \
call_10_9_9_6, \
call_10_9_9_7, \
call_10_9_9_8, \
call_10_9_9_9, \
call_10_9_9_10, \
call_10_9_9_11, \
call_10_9_9_12, \
call_10_9_9_13, \
call_10_9_9_14, \
call_10_9_9_15, \
call_10_9_10_0, \
call_10_9_10_1, \
call_10_9_10_2, \
call_10_9_10_3, \
call_10_9_10_4, \
call_10_9_10_5, \
call_10_9_10_6, \
call_10_9_10_7, \
call_10_9_10_8, \
call_10_9_10_9, \
call_10_9_10_10, \
call_10_9_10_11, \
call_10_9_10_12, \
call_10_9_10_13, \
call_10_9_10_14, \
call_10_9_10_15, \
call_10_9_11_0, \
call_10_9_11_1, \
call_10_9_11_2, \
call_10_9_11_3, \
call_10_9_11_4, \
call_10_9_11_5, \
call_10_9_11_6, \
call_10_9_11_7, \
call_10_9_11_8, \
call_10_9_11_9, \
call_10_9_11_10, \
call_10_9_11_11, \
call_10_9_11_12, \
call_10_9_11_13, \
call_10_9_11_14, \
call_10_9_11_15, \
call_10_9_12_0, \
call_10_9_12_1, \
call_10_9_12_2, \
call_10_9_12_3, \
call_10_9_12_4, \
call_10_9_12_5, \
call_10_9_12_6, \
call_10_9_12_7, \
call_10_9_12_8, \
call_10_9_12_9, \
call_10_9_12_10, \
call_10_9_12_11, \
call_10_9_12_12, \
call_10_9_12_13, \
call_10_9_12_14, \
call_10_9_12_15, \
call_10_9_13_0, \
call_10_9_13_1, \
call_10_9_13_2, \
call_10_9_13_3, \
call_10_9_13_4, \
call_10_9_13_5, \
call_10_9_13_6, \
call_10_9_13_7, \
call_10_9_13_8, \
call_10_9_13_9, \
call_10_9_13_10, \
call_10_9_13_11, \
call_10_9_13_12, \
call_10_9_13_13, \
call_10_9_13_14, \
call_10_9_13_15, \
call_10_9_14_0, \
call_10_9_14_1, \
call_10_9_14_2, \
call_10_9_14_3, \
call_10_9_14_4, \
call_10_9_14_5, \
call_10_9_14_6, \
call_10_9_14_7, \
call_10_9_14_8, \
call_10_9_14_9, \
call_10_9_14_10, \
call_10_9_14_11, \
call_10_9_14_12, \
call_10_9_14_13, \
call_10_9_14_14, \
call_10_9_14_15, \
call_10_9_15_0, \
call_10_9_15_1, \
call_10_9_15_2, \
call_10_9_15_3, \
call_10_9_15_4, \
call_10_9_15_5, \
call_10_9_15_6, \
call_10_9_15_7, \
call_10_9_15_8, \
call_10_9_15_9, \
call_10_9_15_10, \
call_10_9_15_11, \
call_10_9_15_12, \
call_10_9_15_13, \
call_10_9_15_14, \
call_10_9_15_15, \
call_10_10_0_0, \
call_10_10_0_1, \
call_10_10_0_2, \
call_10_10_0_3, \
call_10_10_0_4, \
call_10_10_0_5, \
call_10_10_0_6, \
call_10_10_0_7, \
call_10_10_0_8, \
call_10_10_0_9, \
call_10_10_0_10, \
call_10_10_0_11, \
call_10_10_0_12, \
call_10_10_0_13, \
call_10_10_0_14, \
call_10_10_0_15, \
call_10_10_1_0, \
call_10_10_1_1, \
call_10_10_1_2, \
call_10_10_1_3, \
call_10_10_1_4, \
call_10_10_1_5, \
call_10_10_1_6, \
call_10_10_1_7, \
call_10_10_1_8, \
call_10_10_1_9, \
call_10_10_1_10, \
call_10_10_1_11, \
call_10_10_1_12, \
call_10_10_1_13, \
call_10_10_1_14, \
call_10_10_1_15, \
call_10_10_2_0, \
call_10_10_2_1, \
call_10_10_2_2, \
call_10_10_2_3, \
call_10_10_2_4, \
call_10_10_2_5, \
call_10_10_2_6, \
call_10_10_2_7, \
call_10_10_2_8, \
call_10_10_2_9, \
call_10_10_2_10, \
call_10_10_2_11, \
call_10_10_2_12, \
call_10_10_2_13, \
call_10_10_2_14, \
call_10_10_2_15, \
call_10_10_3_0, \
call_10_10_3_1, \
call_10_10_3_2, \
call_10_10_3_3, \
call_10_10_3_4, \
call_10_10_3_5, \
call_10_10_3_6, \
call_10_10_3_7, \
call_10_10_3_8, \
call_10_10_3_9, \
call_10_10_3_10, \
call_10_10_3_11, \
call_10_10_3_12, \
call_10_10_3_13, \
call_10_10_3_14, \
call_10_10_3_15, \
call_10_10_4_0, \
call_10_10_4_1, \
call_10_10_4_2, \
call_10_10_4_3, \
call_10_10_4_4, \
call_10_10_4_5, \
call_10_10_4_6, \
call_10_10_4_7, \
call_10_10_4_8, \
call_10_10_4_9, \
call_10_10_4_10, \
call_10_10_4_11, \
call_10_10_4_12, \
call_10_10_4_13, \
call_10_10_4_14, \
call_10_10_4_15, \
call_10_10_5_0, \
call_10_10_5_1, \
call_10_10_5_2, \
call_10_10_5_3, \
call_10_10_5_4, \
call_10_10_5_5, \
call_10_10_5_6, \
call_10_10_5_7, \
call_10_10_5_8, \
call_10_10_5_9, \
call_10_10_5_10, \
call_10_10_5_11, \
call_10_10_5_12, \
call_10_10_5_13, \
call_10_10_5_14, \
call_10_10_5_15, \
call_10_10_6_0, \
call_10_10_6_1, \
call_10_10_6_2, \
call_10_10_6_3, \
call_10_10_6_4, \
call_10_10_6_5, \
call_10_10_6_6, \
call_10_10_6_7, \
call_10_10_6_8, \
call_10_10_6_9, \
call_10_10_6_10, \
call_10_10_6_11, \
call_10_10_6_12, \
call_10_10_6_13, \
call_10_10_6_14, \
call_10_10_6_15, \
call_10_10_7_0, \
call_10_10_7_1, \
call_10_10_7_2, \
call_10_10_7_3, \
call_10_10_7_4, \
call_10_10_7_5, \
call_10_10_7_6, \
call_10_10_7_7, \
call_10_10_7_8, \
call_10_10_7_9, \
call_10_10_7_10, \
call_10_10_7_11, \
call_10_10_7_12, \
call_10_10_7_13, \
call_10_10_7_14, \
call_10_10_7_15, \
call_10_10_8_0, \
call_10_10_8_1, \
call_10_10_8_2, \
call_10_10_8_3, \
call_10_10_8_4, \
call_10_10_8_5, \
call_10_10_8_6, \
call_10_10_8_7, \
call_10_10_8_8, \
call_10_10_8_9, \
call_10_10_8_10, \
call_10_10_8_11, \
call_10_10_8_12, \
call_10_10_8_13, \
call_10_10_8_14, \
call_10_10_8_15, \
call_10_10_9_0, \
call_10_10_9_1, \
call_10_10_9_2, \
call_10_10_9_3, \
call_10_10_9_4, \
call_10_10_9_5, \
call_10_10_9_6, \
call_10_10_9_7, \
call_10_10_9_8, \
call_10_10_9_9, \
call_10_10_9_10, \
call_10_10_9_11, \
call_10_10_9_12, \
call_10_10_9_13, \
call_10_10_9_14, \
call_10_10_9_15, \
call_10_10_10_0, \
call_10_10_10_1, \
call_10_10_10_2, \
call_10_10_10_3, \
call_10_10_10_4, \
call_10_10_10_5, \
call_10_10_10_6, \
call_10_10_10_7, \
call_10_10_10_8, \
call_10_10_10_9, \
call_10_10_10_10, \
call_10_10_10_11, \
call_10_10_10_12, \
call_10_10_10_13, \
call_10_10_10_14, \
call_10_10_10_15, \
call_10_10_11_0, \
call_10_10_11_1, \
call_10_10_11_2, \
call_10_10_11_3, \
call_10_10_11_4, \
call_10_10_11_5, \
call_10_10_11_6, \
call_10_10_11_7, \
call_10_10_11_8, \
call_10_10_11_9, \
call_10_10_11_10, \
call_10_10_11_11, \
call_10_10_11_12, \
call_10_10_11_13, \
call_10_10_11_14, \
call_10_10_11_15, \
call_10_10_12_0, \
call_10_10_12_1, \
call_10_10_12_2, \
call_10_10_12_3, \
call_10_10_12_4, \
call_10_10_12_5, \
call_10_10_12_6, \
call_10_10_12_7, \
call_10_10_12_8, \
call_10_10_12_9, \
call_10_10_12_10, \
call_10_10_12_11, \
call_10_10_12_12, \
call_10_10_12_13, \
call_10_10_12_14, \
call_10_10_12_15, \
call_10_10_13_0, \
call_10_10_13_1, \
call_10_10_13_2, \
call_10_10_13_3, \
call_10_10_13_4, \
call_10_10_13_5, \
call_10_10_13_6, \
call_10_10_13_7, \
call_10_10_13_8, \
call_10_10_13_9, \
call_10_10_13_10, \
call_10_10_13_11, \
call_10_10_13_12, \
call_10_10_13_13, \
call_10_10_13_14, \
call_10_10_13_15, \
call_10_10_14_0, \
call_10_10_14_1, \
call_10_10_14_2, \
call_10_10_14_3, \
call_10_10_14_4, \
call_10_10_14_5, \
call_10_10_14_6, \
call_10_10_14_7, \
call_10_10_14_8, \
call_10_10_14_9, \
call_10_10_14_10, \
call_10_10_14_11, \
call_10_10_14_12, \
call_10_10_14_13, \
call_10_10_14_14, \
call_10_10_14_15, \
call_10_10_15_0, \
call_10_10_15_1, \
call_10_10_15_2, \
call_10_10_15_3, \
call_10_10_15_4, \
call_10_10_15_5, \
call_10_10_15_6, \
call_10_10_15_7, \
call_10_10_15_8, \
call_10_10_15_9, \
call_10_10_15_10, \
call_10_10_15_11, \
call_10_10_15_12, \
call_10_10_15_13, \
call_10_10_15_14, \
call_10_10_15_15, \
call_10_11_0_0, \
call_10_11_0_1, \
call_10_11_0_2, \
call_10_11_0_3, \
call_10_11_0_4, \
call_10_11_0_5, \
call_10_11_0_6, \
call_10_11_0_7, \
call_10_11_0_8, \
call_10_11_0_9, \
call_10_11_0_10, \
call_10_11_0_11, \
call_10_11_0_12, \
call_10_11_0_13, \
call_10_11_0_14, \
call_10_11_0_15, \
call_10_11_1_0, \
call_10_11_1_1, \
call_10_11_1_2, \
call_10_11_1_3, \
call_10_11_1_4, \
call_10_11_1_5, \
call_10_11_1_6, \
call_10_11_1_7, \
call_10_11_1_8, \
call_10_11_1_9, \
call_10_11_1_10, \
call_10_11_1_11, \
call_10_11_1_12, \
call_10_11_1_13, \
call_10_11_1_14, \
call_10_11_1_15, \
call_10_11_2_0, \
call_10_11_2_1, \
call_10_11_2_2, \
call_10_11_2_3, \
call_10_11_2_4, \
call_10_11_2_5, \
call_10_11_2_6, \
call_10_11_2_7, \
call_10_11_2_8, \
call_10_11_2_9, \
call_10_11_2_10, \
call_10_11_2_11, \
call_10_11_2_12, \
call_10_11_2_13, \
call_10_11_2_14, \
call_10_11_2_15, \
call_10_11_3_0, \
call_10_11_3_1, \
call_10_11_3_2, \
call_10_11_3_3, \
call_10_11_3_4, \
call_10_11_3_5, \
call_10_11_3_6, \
call_10_11_3_7, \
call_10_11_3_8, \
call_10_11_3_9, \
call_10_11_3_10, \
call_10_11_3_11, \
call_10_11_3_12, \
call_10_11_3_13, \
call_10_11_3_14, \
call_10_11_3_15, \
call_10_11_4_0, \
call_10_11_4_1, \
call_10_11_4_2, \
call_10_11_4_3, \
call_10_11_4_4, \
call_10_11_4_5, \
call_10_11_4_6, \
call_10_11_4_7, \
call_10_11_4_8, \
call_10_11_4_9, \
call_10_11_4_10, \
call_10_11_4_11, \
call_10_11_4_12, \
call_10_11_4_13, \
call_10_11_4_14, \
call_10_11_4_15, \
call_10_11_5_0, \
call_10_11_5_1, \
call_10_11_5_2, \
call_10_11_5_3, \
call_10_11_5_4, \
call_10_11_5_5, \
call_10_11_5_6, \
call_10_11_5_7, \
call_10_11_5_8, \
call_10_11_5_9, \
call_10_11_5_10, \
call_10_11_5_11, \
call_10_11_5_12, \
call_10_11_5_13, \
call_10_11_5_14, \
call_10_11_5_15, \
call_10_11_6_0, \
call_10_11_6_1, \
call_10_11_6_2, \
call_10_11_6_3, \
call_10_11_6_4, \
call_10_11_6_5, \
call_10_11_6_6, \
call_10_11_6_7, \
call_10_11_6_8, \
call_10_11_6_9, \
call_10_11_6_10, \
call_10_11_6_11, \
call_10_11_6_12, \
call_10_11_6_13, \
call_10_11_6_14, \
call_10_11_6_15, \
call_10_11_7_0, \
call_10_11_7_1, \
call_10_11_7_2, \
call_10_11_7_3, \
call_10_11_7_4, \
call_10_11_7_5, \
call_10_11_7_6, \
call_10_11_7_7, \
call_10_11_7_8, \
call_10_11_7_9, \
call_10_11_7_10, \
call_10_11_7_11, \
call_10_11_7_12, \
call_10_11_7_13, \
call_10_11_7_14, \
call_10_11_7_15, \
call_10_11_8_0, \
call_10_11_8_1, \
call_10_11_8_2, \
call_10_11_8_3, \
call_10_11_8_4, \
call_10_11_8_5, \
call_10_11_8_6, \
call_10_11_8_7, \
call_10_11_8_8, \
call_10_11_8_9, \
call_10_11_8_10, \
call_10_11_8_11, \
call_10_11_8_12, \
call_10_11_8_13, \
call_10_11_8_14, \
call_10_11_8_15, \
call_10_11_9_0, \
call_10_11_9_1, \
call_10_11_9_2, \
call_10_11_9_3, \
call_10_11_9_4, \
call_10_11_9_5, \
call_10_11_9_6, \
call_10_11_9_7, \
call_10_11_9_8, \
call_10_11_9_9, \
call_10_11_9_10, \
call_10_11_9_11, \
call_10_11_9_12, \
call_10_11_9_13, \
call_10_11_9_14, \
call_10_11_9_15, \
call_10_11_10_0, \
call_10_11_10_1, \
call_10_11_10_2, \
call_10_11_10_3, \
call_10_11_10_4, \
call_10_11_10_5, \
call_10_11_10_6, \
call_10_11_10_7, \
call_10_11_10_8, \
call_10_11_10_9, \
call_10_11_10_10, \
call_10_11_10_11, \
call_10_11_10_12, \
call_10_11_10_13, \
call_10_11_10_14, \
call_10_11_10_15, \
call_10_11_11_0, \
call_10_11_11_1, \
call_10_11_11_2, \
call_10_11_11_3, \
call_10_11_11_4, \
call_10_11_11_5, \
call_10_11_11_6, \
call_10_11_11_7, \
call_10_11_11_8, \
call_10_11_11_9, \
call_10_11_11_10, \
call_10_11_11_11, \
call_10_11_11_12, \
call_10_11_11_13, \
call_10_11_11_14, \
call_10_11_11_15, \
call_10_11_12_0, \
call_10_11_12_1, \
call_10_11_12_2, \
call_10_11_12_3, \
call_10_11_12_4, \
call_10_11_12_5, \
call_10_11_12_6, \
call_10_11_12_7, \
call_10_11_12_8, \
call_10_11_12_9, \
call_10_11_12_10, \
call_10_11_12_11, \
call_10_11_12_12, \
call_10_11_12_13, \
call_10_11_12_14, \
call_10_11_12_15, \
call_10_11_13_0, \
call_10_11_13_1, \
call_10_11_13_2, \
call_10_11_13_3, \
call_10_11_13_4, \
call_10_11_13_5, \
call_10_11_13_6, \
call_10_11_13_7, \
call_10_11_13_8, \
call_10_11_13_9, \
call_10_11_13_10, \
call_10_11_13_11, \
call_10_11_13_12, \
call_10_11_13_13, \
call_10_11_13_14, \
call_10_11_13_15, \
call_10_11_14_0, \
call_10_11_14_1, \
call_10_11_14_2, \
call_10_11_14_3, \
call_10_11_14_4, \
call_10_11_14_5, \
call_10_11_14_6, \
call_10_11_14_7, \
call_10_11_14_8, \
call_10_11_14_9, \
call_10_11_14_10, \
call_10_11_14_11, \
call_10_11_14_12, \
call_10_11_14_13, \
call_10_11_14_14, \
call_10_11_14_15, \
call_10_11_15_0, \
call_10_11_15_1, \
call_10_11_15_2, \
call_10_11_15_3, \
call_10_11_15_4, \
call_10_11_15_5, \
call_10_11_15_6, \
call_10_11_15_7, \
call_10_11_15_8, \
call_10_11_15_9, \
call_10_11_15_10, \
call_10_11_15_11, \
call_10_11_15_12, \
call_10_11_15_13, \
call_10_11_15_14, \
call_10_11_15_15, \
call_10_12_0_0, \
call_10_12_0_1, \
call_10_12_0_2, \
call_10_12_0_3, \
call_10_12_0_4, \
call_10_12_0_5, \
call_10_12_0_6, \
call_10_12_0_7, \
call_10_12_0_8, \
call_10_12_0_9, \
call_10_12_0_10, \
call_10_12_0_11, \
call_10_12_0_12, \
call_10_12_0_13, \
call_10_12_0_14, \
call_10_12_0_15, \
call_10_12_1_0, \
call_10_12_1_1, \
call_10_12_1_2, \
call_10_12_1_3, \
call_10_12_1_4, \
call_10_12_1_5, \
call_10_12_1_6, \
call_10_12_1_7, \
call_10_12_1_8, \
call_10_12_1_9, \
call_10_12_1_10, \
call_10_12_1_11, \
call_10_12_1_12, \
call_10_12_1_13, \
call_10_12_1_14, \
call_10_12_1_15, \
call_10_12_2_0, \
call_10_12_2_1, \
call_10_12_2_2, \
call_10_12_2_3, \
call_10_12_2_4, \
call_10_12_2_5, \
call_10_12_2_6, \
call_10_12_2_7, \
call_10_12_2_8, \
call_10_12_2_9, \
call_10_12_2_10, \
call_10_12_2_11, \
call_10_12_2_12, \
call_10_12_2_13, \
call_10_12_2_14, \
call_10_12_2_15, \
call_10_12_3_0, \
call_10_12_3_1, \
call_10_12_3_2, \
call_10_12_3_3, \
call_10_12_3_4, \
call_10_12_3_5, \
call_10_12_3_6, \
call_10_12_3_7, \
call_10_12_3_8, \
call_10_12_3_9, \
call_10_12_3_10, \
call_10_12_3_11, \
call_10_12_3_12, \
call_10_12_3_13, \
call_10_12_3_14, \
call_10_12_3_15, \
call_10_12_4_0, \
call_10_12_4_1, \
call_10_12_4_2, \
call_10_12_4_3, \
call_10_12_4_4, \
call_10_12_4_5, \
call_10_12_4_6, \
call_10_12_4_7, \
call_10_12_4_8, \
call_10_12_4_9, \
call_10_12_4_10, \
call_10_12_4_11, \
call_10_12_4_12, \
call_10_12_4_13, \
call_10_12_4_14, \
call_10_12_4_15, \
call_10_12_5_0, \
call_10_12_5_1, \
call_10_12_5_2, \
call_10_12_5_3, \
call_10_12_5_4, \
call_10_12_5_5, \
call_10_12_5_6, \
call_10_12_5_7, \
call_10_12_5_8, \
call_10_12_5_9, \
call_10_12_5_10, \
call_10_12_5_11, \
call_10_12_5_12, \
call_10_12_5_13, \
call_10_12_5_14, \
call_10_12_5_15, \
call_10_12_6_0, \
call_10_12_6_1, \
call_10_12_6_2, \
call_10_12_6_3, \
call_10_12_6_4, \
call_10_12_6_5, \
call_10_12_6_6, \
call_10_12_6_7, \
call_10_12_6_8, \
call_10_12_6_9, \
call_10_12_6_10, \
call_10_12_6_11, \
call_10_12_6_12, \
call_10_12_6_13, \
call_10_12_6_14, \
call_10_12_6_15, \
call_10_12_7_0, \
call_10_12_7_1, \
call_10_12_7_2, \
call_10_12_7_3, \
call_10_12_7_4, \
call_10_12_7_5, \
call_10_12_7_6, \
call_10_12_7_7, \
call_10_12_7_8, \
call_10_12_7_9, \
call_10_12_7_10, \
call_10_12_7_11, \
call_10_12_7_12, \
call_10_12_7_13, \
call_10_12_7_14, \
call_10_12_7_15, \
call_10_12_8_0, \
call_10_12_8_1, \
call_10_12_8_2, \
call_10_12_8_3, \
call_10_12_8_4, \
call_10_12_8_5, \
call_10_12_8_6, \
call_10_12_8_7, \
call_10_12_8_8, \
call_10_12_8_9, \
call_10_12_8_10, \
call_10_12_8_11, \
call_10_12_8_12, \
call_10_12_8_13, \
call_10_12_8_14, \
call_10_12_8_15, \
call_10_12_9_0, \
call_10_12_9_1, \
call_10_12_9_2, \
call_10_12_9_3, \
call_10_12_9_4, \
call_10_12_9_5, \
call_10_12_9_6, \
call_10_12_9_7, \
call_10_12_9_8, \
call_10_12_9_9, \
call_10_12_9_10, \
call_10_12_9_11, \
call_10_12_9_12, \
call_10_12_9_13, \
call_10_12_9_14, \
call_10_12_9_15, \
call_10_12_10_0, \
call_10_12_10_1, \
call_10_12_10_2, \
call_10_12_10_3, \
call_10_12_10_4, \
call_10_12_10_5, \
call_10_12_10_6, \
call_10_12_10_7, \
call_10_12_10_8, \
call_10_12_10_9, \
call_10_12_10_10, \
call_10_12_10_11, \
call_10_12_10_12, \
call_10_12_10_13, \
call_10_12_10_14, \
call_10_12_10_15, \
call_10_12_11_0, \
call_10_12_11_1, \
call_10_12_11_2, \
call_10_12_11_3, \
call_10_12_11_4, \
call_10_12_11_5, \
call_10_12_11_6, \
call_10_12_11_7, \
call_10_12_11_8, \
call_10_12_11_9, \
call_10_12_11_10, \
call_10_12_11_11, \
call_10_12_11_12, \
call_10_12_11_13, \
call_10_12_11_14, \
call_10_12_11_15, \
call_10_12_12_0, \
call_10_12_12_1, \
call_10_12_12_2, \
call_10_12_12_3, \
call_10_12_12_4, \
call_10_12_12_5, \
call_10_12_12_6, \
call_10_12_12_7, \
call_10_12_12_8, \
call_10_12_12_9, \
call_10_12_12_10, \
call_10_12_12_11, \
call_10_12_12_12, \
call_10_12_12_13, \
call_10_12_12_14, \
call_10_12_12_15, \
call_10_12_13_0, \
call_10_12_13_1, \
call_10_12_13_2, \
call_10_12_13_3, \
call_10_12_13_4, \
call_10_12_13_5, \
call_10_12_13_6, \
call_10_12_13_7, \
call_10_12_13_8, \
call_10_12_13_9, \
call_10_12_13_10, \
call_10_12_13_11, \
call_10_12_13_12, \
call_10_12_13_13, \
call_10_12_13_14, \
call_10_12_13_15, \
call_10_12_14_0, \
call_10_12_14_1, \
call_10_12_14_2, \
call_10_12_14_3, \
call_10_12_14_4, \
call_10_12_14_5, \
call_10_12_14_6, \
call_10_12_14_7, \
call_10_12_14_8, \
call_10_12_14_9, \
call_10_12_14_10, \
call_10_12_14_11, \
call_10_12_14_12, \
call_10_12_14_13, \
call_10_12_14_14, \
call_10_12_14_15, \
call_10_12_15_0, \
call_10_12_15_1, \
call_10_12_15_2, \
call_10_12_15_3, \
call_10_12_15_4, \
call_10_12_15_5, \
call_10_12_15_6, \
call_10_12_15_7, \
call_10_12_15_8, \
call_10_12_15_9, \
call_10_12_15_10, \
call_10_12_15_11, \
call_10_12_15_12, \
call_10_12_15_13, \
call_10_12_15_14, \
call_10_12_15_15, \
call_10_13_0_0, \
call_10_13_0_1, \
call_10_13_0_2, \
call_10_13_0_3, \
call_10_13_0_4, \
call_10_13_0_5, \
call_10_13_0_6, \
call_10_13_0_7, \
call_10_13_0_8, \
call_10_13_0_9, \
call_10_13_0_10, \
call_10_13_0_11, \
call_10_13_0_12, \
call_10_13_0_13, \
call_10_13_0_14, \
call_10_13_0_15, \
call_10_13_1_0, \
call_10_13_1_1, \
call_10_13_1_2, \
call_10_13_1_3, \
call_10_13_1_4, \
call_10_13_1_5, \
call_10_13_1_6, \
call_10_13_1_7, \
call_10_13_1_8, \
call_10_13_1_9, \
call_10_13_1_10, \
call_10_13_1_11, \
call_10_13_1_12, \
call_10_13_1_13, \
call_10_13_1_14, \
call_10_13_1_15, \
call_10_13_2_0, \
call_10_13_2_1, \
call_10_13_2_2, \
call_10_13_2_3, \
call_10_13_2_4, \
call_10_13_2_5, \
call_10_13_2_6, \
call_10_13_2_7, \
call_10_13_2_8, \
call_10_13_2_9, \
call_10_13_2_10, \
call_10_13_2_11, \
call_10_13_2_12, \
call_10_13_2_13, \
call_10_13_2_14, \
call_10_13_2_15, \
call_10_13_3_0, \
call_10_13_3_1, \
call_10_13_3_2, \
call_10_13_3_3, \
call_10_13_3_4, \
call_10_13_3_5, \
call_10_13_3_6, \
call_10_13_3_7, \
call_10_13_3_8, \
call_10_13_3_9, \
call_10_13_3_10, \
call_10_13_3_11, \
call_10_13_3_12, \
call_10_13_3_13, \
call_10_13_3_14, \
call_10_13_3_15, \
call_10_13_4_0, \
call_10_13_4_1, \
call_10_13_4_2, \
call_10_13_4_3, \
call_10_13_4_4, \
call_10_13_4_5, \
call_10_13_4_6, \
call_10_13_4_7, \
call_10_13_4_8, \
call_10_13_4_9, \
call_10_13_4_10, \
call_10_13_4_11, \
call_10_13_4_12, \
call_10_13_4_13, \
call_10_13_4_14, \
call_10_13_4_15, \
call_10_13_5_0, \
call_10_13_5_1, \
call_10_13_5_2, \
call_10_13_5_3, \
call_10_13_5_4, \
call_10_13_5_5, \
call_10_13_5_6, \
call_10_13_5_7, \
call_10_13_5_8, \
call_10_13_5_9, \
call_10_13_5_10, \
call_10_13_5_11, \
call_10_13_5_12, \
call_10_13_5_13, \
call_10_13_5_14, \
call_10_13_5_15, \
call_10_13_6_0, \
call_10_13_6_1, \
call_10_13_6_2, \
call_10_13_6_3, \
call_10_13_6_4, \
call_10_13_6_5, \
call_10_13_6_6, \
call_10_13_6_7, \
call_10_13_6_8, \
call_10_13_6_9, \
call_10_13_6_10, \
call_10_13_6_11, \
call_10_13_6_12, \
call_10_13_6_13, \
call_10_13_6_14, \
call_10_13_6_15, \
call_10_13_7_0, \
call_10_13_7_1, \
call_10_13_7_2, \
call_10_13_7_3, \
call_10_13_7_4, \
call_10_13_7_5, \
call_10_13_7_6, \
call_10_13_7_7, \
call_10_13_7_8, \
call_10_13_7_9, \
call_10_13_7_10, \
call_10_13_7_11, \
call_10_13_7_12, \
call_10_13_7_13, \
call_10_13_7_14, \
call_10_13_7_15, \
call_10_13_8_0, \
call_10_13_8_1, \
call_10_13_8_2, \
call_10_13_8_3, \
call_10_13_8_4, \
call_10_13_8_5, \
call_10_13_8_6, \
call_10_13_8_7, \
call_10_13_8_8, \
call_10_13_8_9, \
call_10_13_8_10, \
call_10_13_8_11, \
call_10_13_8_12, \
call_10_13_8_13, \
call_10_13_8_14, \
call_10_13_8_15, \
call_10_13_9_0, \
call_10_13_9_1, \
call_10_13_9_2, \
call_10_13_9_3, \
call_10_13_9_4, \
call_10_13_9_5, \
call_10_13_9_6, \
call_10_13_9_7, \
call_10_13_9_8, \
call_10_13_9_9, \
call_10_13_9_10, \
call_10_13_9_11, \
call_10_13_9_12, \
call_10_13_9_13, \
call_10_13_9_14, \
call_10_13_9_15, \
call_10_13_10_0, \
call_10_13_10_1, \
call_10_13_10_2, \
call_10_13_10_3, \
call_10_13_10_4, \
call_10_13_10_5, \
call_10_13_10_6, \
call_10_13_10_7, \
call_10_13_10_8, \
call_10_13_10_9, \
call_10_13_10_10, \
call_10_13_10_11, \
call_10_13_10_12, \
call_10_13_10_13, \
call_10_13_10_14, \
call_10_13_10_15, \
call_10_13_11_0, \
call_10_13_11_1, \
call_10_13_11_2, \
call_10_13_11_3, \
call_10_13_11_4, \
call_10_13_11_5, \
call_10_13_11_6, \
call_10_13_11_7, \
call_10_13_11_8, \
call_10_13_11_9, \
call_10_13_11_10, \
call_10_13_11_11, \
call_10_13_11_12, \
call_10_13_11_13, \
call_10_13_11_14, \
call_10_13_11_15, \
call_10_13_12_0, \
call_10_13_12_1, \
call_10_13_12_2, \
call_10_13_12_3, \
call_10_13_12_4, \
call_10_13_12_5, \
call_10_13_12_6, \
call_10_13_12_7, \
call_10_13_12_8, \
call_10_13_12_9, \
call_10_13_12_10, \
call_10_13_12_11, \
call_10_13_12_12, \
call_10_13_12_13, \
call_10_13_12_14, \
call_10_13_12_15, \
call_10_13_13_0, \
call_10_13_13_1, \
call_10_13_13_2, \
call_10_13_13_3, \
call_10_13_13_4, \
call_10_13_13_5, \
call_10_13_13_6, \
call_10_13_13_7, \
call_10_13_13_8, \
call_10_13_13_9, \
call_10_13_13_10, \
call_10_13_13_11, \
call_10_13_13_12, \
call_10_13_13_13, \
call_10_13_13_14, \
call_10_13_13_15, \
call_10_13_14_0, \
call_10_13_14_1, \
call_10_13_14_2, \
call_10_13_14_3, \
call_10_13_14_4, \
call_10_13_14_5, \
call_10_13_14_6, \
call_10_13_14_7, \
call_10_13_14_8, \
call_10_13_14_9, \
call_10_13_14_10, \
call_10_13_14_11, \
call_10_13_14_12, \
call_10_13_14_13, \
call_10_13_14_14, \
call_10_13_14_15, \
call_10_13_15_0, \
call_10_13_15_1, \
call_10_13_15_2, \
call_10_13_15_3, \
call_10_13_15_4, \
call_10_13_15_5, \
call_10_13_15_6, \
call_10_13_15_7, \
call_10_13_15_8, \
call_10_13_15_9, \
call_10_13_15_10, \
call_10_13_15_11, \
call_10_13_15_12, \
call_10_13_15_13, \
call_10_13_15_14, \
call_10_13_15_15, \
call_10_14_0_0, \
call_10_14_0_1, \
call_10_14_0_2, \
call_10_14_0_3, \
call_10_14_0_4, \
call_10_14_0_5, \
call_10_14_0_6, \
call_10_14_0_7, \
call_10_14_0_8, \
call_10_14_0_9, \
call_10_14_0_10, \
call_10_14_0_11, \
call_10_14_0_12, \
call_10_14_0_13, \
call_10_14_0_14, \
call_10_14_0_15, \
call_10_14_1_0, \
call_10_14_1_1, \
call_10_14_1_2, \
call_10_14_1_3, \
call_10_14_1_4, \
call_10_14_1_5, \
call_10_14_1_6, \
call_10_14_1_7, \
call_10_14_1_8, \
call_10_14_1_9, \
call_10_14_1_10, \
call_10_14_1_11, \
call_10_14_1_12, \
call_10_14_1_13, \
call_10_14_1_14, \
call_10_14_1_15, \
call_10_14_2_0, \
call_10_14_2_1, \
call_10_14_2_2, \
call_10_14_2_3, \
call_10_14_2_4, \
call_10_14_2_5, \
call_10_14_2_6, \
call_10_14_2_7, \
call_10_14_2_8, \
call_10_14_2_9, \
call_10_14_2_10, \
call_10_14_2_11, \
call_10_14_2_12, \
call_10_14_2_13, \
call_10_14_2_14, \
call_10_14_2_15, \
call_10_14_3_0, \
call_10_14_3_1, \
call_10_14_3_2, \
call_10_14_3_3, \
call_10_14_3_4, \
call_10_14_3_5, \
call_10_14_3_6, \
call_10_14_3_7, \
call_10_14_3_8, \
call_10_14_3_9, \
call_10_14_3_10, \
call_10_14_3_11, \
call_10_14_3_12, \
call_10_14_3_13, \
call_10_14_3_14, \
call_10_14_3_15, \
call_10_14_4_0, \
call_10_14_4_1, \
call_10_14_4_2, \
call_10_14_4_3, \
call_10_14_4_4, \
call_10_14_4_5, \
call_10_14_4_6, \
call_10_14_4_7, \
call_10_14_4_8, \
call_10_14_4_9, \
call_10_14_4_10, \
call_10_14_4_11, \
call_10_14_4_12, \
call_10_14_4_13, \
call_10_14_4_14, \
call_10_14_4_15, \
call_10_14_5_0, \
call_10_14_5_1, \
call_10_14_5_2, \
call_10_14_5_3, \
call_10_14_5_4, \
call_10_14_5_5, \
call_10_14_5_6, \
call_10_14_5_7, \
call_10_14_5_8, \
call_10_14_5_9, \
call_10_14_5_10, \
call_10_14_5_11, \
call_10_14_5_12, \
call_10_14_5_13, \
call_10_14_5_14, \
call_10_14_5_15, \
call_10_14_6_0, \
call_10_14_6_1, \
call_10_14_6_2, \
call_10_14_6_3, \
call_10_14_6_4, \
call_10_14_6_5, \
call_10_14_6_6, \
call_10_14_6_7, \
call_10_14_6_8, \
call_10_14_6_9, \
call_10_14_6_10, \
call_10_14_6_11, \
call_10_14_6_12, \
call_10_14_6_13, \
call_10_14_6_14, \
call_10_14_6_15, \
call_10_14_7_0, \
call_10_14_7_1, \
call_10_14_7_2, \
call_10_14_7_3, \
call_10_14_7_4, \
call_10_14_7_5, \
call_10_14_7_6, \
call_10_14_7_7, \
call_10_14_7_8, \
call_10_14_7_9, \
call_10_14_7_10, \
call_10_14_7_11, \
call_10_14_7_12, \
call_10_14_7_13, \
call_10_14_7_14, \
call_10_14_7_15, \
call_10_14_8_0, \
call_10_14_8_1, \
call_10_14_8_2, \
call_10_14_8_3, \
call_10_14_8_4, \
call_10_14_8_5, \
call_10_14_8_6, \
call_10_14_8_7, \
call_10_14_8_8, \
call_10_14_8_9, \
call_10_14_8_10, \
call_10_14_8_11, \
call_10_14_8_12, \
call_10_14_8_13, \
call_10_14_8_14, \
call_10_14_8_15, \
call_10_14_9_0, \
call_10_14_9_1, \
call_10_14_9_2, \
call_10_14_9_3, \
call_10_14_9_4, \
call_10_14_9_5, \
call_10_14_9_6, \
call_10_14_9_7, \
call_10_14_9_8, \
call_10_14_9_9, \
call_10_14_9_10, \
call_10_14_9_11, \
call_10_14_9_12, \
call_10_14_9_13, \
call_10_14_9_14, \
call_10_14_9_15, \
call_10_14_10_0, \
call_10_14_10_1, \
call_10_14_10_2, \
call_10_14_10_3, \
call_10_14_10_4, \
call_10_14_10_5, \
call_10_14_10_6, \
call_10_14_10_7, \
call_10_14_10_8, \
call_10_14_10_9, \
call_10_14_10_10, \
call_10_14_10_11, \
call_10_14_10_12, \
call_10_14_10_13, \
call_10_14_10_14, \
call_10_14_10_15, \
call_10_14_11_0, \
call_10_14_11_1, \
call_10_14_11_2, \
call_10_14_11_3, \
call_10_14_11_4, \
call_10_14_11_5, \
call_10_14_11_6, \
call_10_14_11_7, \
call_10_14_11_8, \
call_10_14_11_9, \
call_10_14_11_10, \
call_10_14_11_11, \
call_10_14_11_12, \
call_10_14_11_13, \
call_10_14_11_14, \
call_10_14_11_15, \
call_10_14_12_0, \
call_10_14_12_1, \
call_10_14_12_2, \
call_10_14_12_3, \
call_10_14_12_4, \
call_10_14_12_5, \
call_10_14_12_6, \
call_10_14_12_7, \
call_10_14_12_8, \
call_10_14_12_9, \
call_10_14_12_10, \
call_10_14_12_11, \
call_10_14_12_12, \
call_10_14_12_13, \
call_10_14_12_14, \
call_10_14_12_15, \
call_10_14_13_0, \
call_10_14_13_1, \
call_10_14_13_2, \
call_10_14_13_3, \
call_10_14_13_4, \
call_10_14_13_5, \
call_10_14_13_6, \
call_10_14_13_7, \
call_10_14_13_8, \
call_10_14_13_9, \
call_10_14_13_10, \
call_10_14_13_11, \
call_10_14_13_12, \
call_10_14_13_13, \
call_10_14_13_14, \
call_10_14_13_15, \
call_10_14_14_0, \
call_10_14_14_1, \
call_10_14_14_2, \
call_10_14_14_3, \
call_10_14_14_4, \
call_10_14_14_5, \
call_10_14_14_6, \
call_10_14_14_7, \
call_10_14_14_8, \
call_10_14_14_9, \
call_10_14_14_10, \
call_10_14_14_11, \
call_10_14_14_12, \
call_10_14_14_13, \
call_10_14_14_14, \
call_10_14_14_15, \
call_10_14_15_0, \
call_10_14_15_1, \
call_10_14_15_2, \
call_10_14_15_3, \
call_10_14_15_4, \
call_10_14_15_5, \
call_10_14_15_6, \
call_10_14_15_7, \
call_10_14_15_8, \
call_10_14_15_9, \
call_10_14_15_10, \
call_10_14_15_11, \
call_10_14_15_12, \
call_10_14_15_13, \
call_10_14_15_14, \
call_10_14_15_15, \
call_10_15_0_0, \
call_10_15_0_1, \
call_10_15_0_2, \
call_10_15_0_3, \
call_10_15_0_4, \
call_10_15_0_5, \
call_10_15_0_6, \
call_10_15_0_7, \
call_10_15_0_8, \
call_10_15_0_9, \
call_10_15_0_10, \
call_10_15_0_11, \
call_10_15_0_12, \
call_10_15_0_13, \
call_10_15_0_14, \
call_10_15_0_15, \
call_10_15_1_0, \
call_10_15_1_1, \
call_10_15_1_2, \
call_10_15_1_3, \
call_10_15_1_4, \
call_10_15_1_5, \
call_10_15_1_6, \
call_10_15_1_7, \
call_10_15_1_8, \
call_10_15_1_9, \
call_10_15_1_10, \
call_10_15_1_11, \
call_10_15_1_12, \
call_10_15_1_13, \
call_10_15_1_14, \
call_10_15_1_15, \
call_10_15_2_0, \
call_10_15_2_1, \
call_10_15_2_2, \
call_10_15_2_3, \
call_10_15_2_4, \
call_10_15_2_5, \
call_10_15_2_6, \
call_10_15_2_7, \
call_10_15_2_8, \
call_10_15_2_9, \
call_10_15_2_10, \
call_10_15_2_11, \
call_10_15_2_12, \
call_10_15_2_13, \
call_10_15_2_14, \
call_10_15_2_15, \
call_10_15_3_0, \
call_10_15_3_1, \
call_10_15_3_2, \
call_10_15_3_3, \
call_10_15_3_4, \
call_10_15_3_5, \
call_10_15_3_6, \
call_10_15_3_7, \
call_10_15_3_8, \
call_10_15_3_9, \
call_10_15_3_10, \
call_10_15_3_11, \
call_10_15_3_12, \
call_10_15_3_13, \
call_10_15_3_14, \
call_10_15_3_15, \
call_10_15_4_0, \
call_10_15_4_1, \
call_10_15_4_2, \
call_10_15_4_3, \
call_10_15_4_4, \
call_10_15_4_5, \
call_10_15_4_6, \
call_10_15_4_7, \
call_10_15_4_8, \
call_10_15_4_9, \
call_10_15_4_10, \
call_10_15_4_11, \
call_10_15_4_12, \
call_10_15_4_13, \
call_10_15_4_14, \
call_10_15_4_15, \
call_10_15_5_0, \
call_10_15_5_1, \
call_10_15_5_2, \
call_10_15_5_3, \
call_10_15_5_4, \
call_10_15_5_5, \
call_10_15_5_6, \
call_10_15_5_7, \
call_10_15_5_8, \
call_10_15_5_9, \
call_10_15_5_10, \
call_10_15_5_11, \
call_10_15_5_12, \
call_10_15_5_13, \
call_10_15_5_14, \
call_10_15_5_15, \
call_10_15_6_0, \
call_10_15_6_1, \
call_10_15_6_2, \
call_10_15_6_3, \
call_10_15_6_4, \
call_10_15_6_5, \
call_10_15_6_6, \
call_10_15_6_7, \
call_10_15_6_8, \
call_10_15_6_9, \
call_10_15_6_10, \
call_10_15_6_11, \
call_10_15_6_12, \
call_10_15_6_13, \
call_10_15_6_14, \
call_10_15_6_15, \
call_10_15_7_0, \
call_10_15_7_1, \
call_10_15_7_2, \
call_10_15_7_3, \
call_10_15_7_4, \
call_10_15_7_5, \
call_10_15_7_6, \
call_10_15_7_7, \
call_10_15_7_8, \
call_10_15_7_9, \
call_10_15_7_10, \
call_10_15_7_11, \
call_10_15_7_12, \
call_10_15_7_13, \
call_10_15_7_14, \
call_10_15_7_15, \
call_10_15_8_0, \
call_10_15_8_1, \
call_10_15_8_2, \
call_10_15_8_3, \
call_10_15_8_4, \
call_10_15_8_5, \
call_10_15_8_6, \
call_10_15_8_7, \
call_10_15_8_8, \
call_10_15_8_9, \
call_10_15_8_10, \
call_10_15_8_11, \
call_10_15_8_12, \
call_10_15_8_13, \
call_10_15_8_14, \
call_10_15_8_15, \
call_10_15_9_0, \
call_10_15_9_1, \
call_10_15_9_2, \
call_10_15_9_3, \
call_10_15_9_4, \
call_10_15_9_5, \
call_10_15_9_6, \
call_10_15_9_7, \
call_10_15_9_8, \
call_10_15_9_9, \
call_10_15_9_10, \
call_10_15_9_11, \
call_10_15_9_12, \
call_10_15_9_13, \
call_10_15_9_14, \
call_10_15_9_15, \
call_10_15_10_0, \
call_10_15_10_1, \
call_10_15_10_2, \
call_10_15_10_3, \
call_10_15_10_4, \
call_10_15_10_5, \
call_10_15_10_6, \
call_10_15_10_7, \
call_10_15_10_8, \
call_10_15_10_9, \
call_10_15_10_10, \
call_10_15_10_11, \
call_10_15_10_12, \
call_10_15_10_13, \
call_10_15_10_14, \
call_10_15_10_15, \
call_10_15_11_0, \
call_10_15_11_1, \
call_10_15_11_2, \
call_10_15_11_3, \
call_10_15_11_4, \
call_10_15_11_5, \
call_10_15_11_6, \
call_10_15_11_7, \
call_10_15_11_8, \
call_10_15_11_9, \
call_10_15_11_10, \
call_10_15_11_11, \
call_10_15_11_12, \
call_10_15_11_13, \
call_10_15_11_14, \
call_10_15_11_15, \
call_10_15_12_0, \
call_10_15_12_1, \
call_10_15_12_2, \
call_10_15_12_3, \
call_10_15_12_4, \
call_10_15_12_5, \
call_10_15_12_6, \
call_10_15_12_7, \
call_10_15_12_8, \
call_10_15_12_9, \
call_10_15_12_10, \
call_10_15_12_11, \
call_10_15_12_12, \
call_10_15_12_13, \
call_10_15_12_14, \
call_10_15_12_15, \
call_10_15_13_0, \
call_10_15_13_1, \
call_10_15_13_2, \
call_10_15_13_3, \
call_10_15_13_4, \
call_10_15_13_5, \
call_10_15_13_6, \
call_10_15_13_7, \
call_10_15_13_8, \
call_10_15_13_9, \
call_10_15_13_10, \
call_10_15_13_11, \
call_10_15_13_12, \
call_10_15_13_13, \
call_10_15_13_14, \
call_10_15_13_15, \
call_10_15_14_0, \
call_10_15_14_1, \
call_10_15_14_2, \
call_10_15_14_3, \
call_10_15_14_4, \
call_10_15_14_5, \
call_10_15_14_6, \
call_10_15_14_7, \
call_10_15_14_8, \
call_10_15_14_9, \
call_10_15_14_10, \
call_10_15_14_11, \
call_10_15_14_12, \
call_10_15_14_13, \
call_10_15_14_14, \
call_10_15_14_15, \
call_10_15_15_0, \
call_10_15_15_1, \
call_10_15_15_2, \
call_10_15_15_3, \
call_10_15_15_4, \
call_10_15_15_5, \
call_10_15_15_6, \
call_10_15_15_7, \
call_10_15_15_8, \
call_10_15_15_9, \
call_10_15_15_10, \
call_10_15_15_11, \
call_10_15_15_12, \
call_10_15_15_13, \
call_10_15_15_14, \
call_10_15_15_15, \
call_11_0_0_0, \
call_11_0_0_1, \
call_11_0_0_2, \
call_11_0_0_3, \
call_11_0_0_4, \
call_11_0_0_5, \
call_11_0_0_6, \
call_11_0_0_7, \
call_11_0_0_8, \
call_11_0_0_9, \
call_11_0_0_10, \
call_11_0_0_11, \
call_11_0_0_12, \
call_11_0_0_13, \
call_11_0_0_14, \
call_11_0_0_15, \
call_11_0_1_0, \
call_11_0_1_1, \
call_11_0_1_2, \
call_11_0_1_3, \
call_11_0_1_4, \
call_11_0_1_5, \
call_11_0_1_6, \
call_11_0_1_7, \
call_11_0_1_8, \
call_11_0_1_9, \
call_11_0_1_10, \
call_11_0_1_11, \
call_11_0_1_12, \
call_11_0_1_13, \
call_11_0_1_14, \
call_11_0_1_15, \
call_11_0_2_0, \
call_11_0_2_1, \
call_11_0_2_2, \
call_11_0_2_3, \
call_11_0_2_4, \
call_11_0_2_5, \
call_11_0_2_6, \
call_11_0_2_7, \
call_11_0_2_8, \
call_11_0_2_9, \
call_11_0_2_10, \
call_11_0_2_11, \
call_11_0_2_12, \
call_11_0_2_13, \
call_11_0_2_14, \
call_11_0_2_15, \
call_11_0_3_0, \
call_11_0_3_1, \
call_11_0_3_2, \
call_11_0_3_3, \
call_11_0_3_4, \
call_11_0_3_5, \
call_11_0_3_6, \
call_11_0_3_7, \
call_11_0_3_8, \
call_11_0_3_9, \
call_11_0_3_10, \
call_11_0_3_11, \
call_11_0_3_12, \
call_11_0_3_13, \
call_11_0_3_14, \
call_11_0_3_15, \
call_11_0_4_0, \
call_11_0_4_1, \
call_11_0_4_2, \
call_11_0_4_3, \
call_11_0_4_4, \
call_11_0_4_5, \
call_11_0_4_6, \
call_11_0_4_7, \
call_11_0_4_8, \
call_11_0_4_9, \
call_11_0_4_10, \
call_11_0_4_11, \
call_11_0_4_12, \
call_11_0_4_13, \
call_11_0_4_14, \
call_11_0_4_15, \
call_11_0_5_0, \
call_11_0_5_1, \
call_11_0_5_2, \
call_11_0_5_3, \
call_11_0_5_4, \
call_11_0_5_5, \
call_11_0_5_6, \
call_11_0_5_7, \
call_11_0_5_8, \
call_11_0_5_9, \
call_11_0_5_10, \
call_11_0_5_11, \
call_11_0_5_12, \
call_11_0_5_13, \
call_11_0_5_14, \
call_11_0_5_15, \
call_11_0_6_0, \
call_11_0_6_1, \
call_11_0_6_2, \
call_11_0_6_3, \
call_11_0_6_4, \
call_11_0_6_5, \
call_11_0_6_6, \
call_11_0_6_7, \
call_11_0_6_8, \
call_11_0_6_9, \
call_11_0_6_10, \
call_11_0_6_11, \
call_11_0_6_12, \
call_11_0_6_13, \
call_11_0_6_14, \
call_11_0_6_15, \
call_11_0_7_0, \
call_11_0_7_1, \
call_11_0_7_2, \
call_11_0_7_3, \
call_11_0_7_4, \
call_11_0_7_5, \
call_11_0_7_6, \
call_11_0_7_7, \
call_11_0_7_8, \
call_11_0_7_9, \
call_11_0_7_10, \
call_11_0_7_11, \
call_11_0_7_12, \
call_11_0_7_13, \
call_11_0_7_14, \
call_11_0_7_15, \
call_11_0_8_0, \
call_11_0_8_1, \
call_11_0_8_2, \
call_11_0_8_3, \
call_11_0_8_4, \
call_11_0_8_5, \
call_11_0_8_6, \
call_11_0_8_7, \
call_11_0_8_8, \
call_11_0_8_9, \
call_11_0_8_10, \
call_11_0_8_11, \
call_11_0_8_12, \
call_11_0_8_13, \
call_11_0_8_14, \
call_11_0_8_15, \
call_11_0_9_0, \
call_11_0_9_1, \
call_11_0_9_2, \
call_11_0_9_3, \
call_11_0_9_4, \
call_11_0_9_5, \
call_11_0_9_6, \
call_11_0_9_7, \
call_11_0_9_8, \
call_11_0_9_9, \
call_11_0_9_10, \
call_11_0_9_11, \
call_11_0_9_12, \
call_11_0_9_13, \
call_11_0_9_14, \
call_11_0_9_15, \
call_11_0_10_0, \
call_11_0_10_1, \
call_11_0_10_2, \
call_11_0_10_3, \
call_11_0_10_4, \
call_11_0_10_5, \
call_11_0_10_6, \
call_11_0_10_7, \
call_11_0_10_8, \
call_11_0_10_9, \
call_11_0_10_10, \
call_11_0_10_11, \
call_11_0_10_12, \
call_11_0_10_13, \
call_11_0_10_14, \
call_11_0_10_15, \
call_11_0_11_0, \
call_11_0_11_1, \
call_11_0_11_2, \
call_11_0_11_3, \
call_11_0_11_4, \
call_11_0_11_5, \
call_11_0_11_6, \
call_11_0_11_7, \
call_11_0_11_8, \
call_11_0_11_9, \
call_11_0_11_10, \
call_11_0_11_11, \
call_11_0_11_12, \
call_11_0_11_13, \
call_11_0_11_14, \
call_11_0_11_15, \
call_11_0_12_0, \
call_11_0_12_1, \
call_11_0_12_2, \
call_11_0_12_3, \
call_11_0_12_4, \
call_11_0_12_5, \
call_11_0_12_6, \
call_11_0_12_7, \
call_11_0_12_8, \
call_11_0_12_9, \
call_11_0_12_10, \
call_11_0_12_11, \
call_11_0_12_12, \
call_11_0_12_13, \
call_11_0_12_14, \
call_11_0_12_15, \
call_11_0_13_0, \
call_11_0_13_1, \
call_11_0_13_2, \
call_11_0_13_3, \
call_11_0_13_4, \
call_11_0_13_5, \
call_11_0_13_6, \
call_11_0_13_7, \
call_11_0_13_8, \
call_11_0_13_9, \
call_11_0_13_10, \
call_11_0_13_11, \
call_11_0_13_12, \
call_11_0_13_13, \
call_11_0_13_14, \
call_11_0_13_15, \
call_11_0_14_0, \
call_11_0_14_1, \
call_11_0_14_2, \
call_11_0_14_3, \
call_11_0_14_4, \
call_11_0_14_5, \
call_11_0_14_6, \
call_11_0_14_7, \
call_11_0_14_8, \
call_11_0_14_9, \
call_11_0_14_10, \
call_11_0_14_11, \
call_11_0_14_12, \
call_11_0_14_13, \
call_11_0_14_14, \
call_11_0_14_15, \
call_11_0_15_0, \
call_11_0_15_1, \
call_11_0_15_2, \
call_11_0_15_3, \
call_11_0_15_4, \
call_11_0_15_5, \
call_11_0_15_6, \
call_11_0_15_7, \
call_11_0_15_8, \
call_11_0_15_9, \
call_11_0_15_10, \
call_11_0_15_11, \
call_11_0_15_12, \
call_11_0_15_13, \
call_11_0_15_14, \
call_11_0_15_15, \
call_11_1_0_0, \
call_11_1_0_1, \
call_11_1_0_2, \
call_11_1_0_3, \
call_11_1_0_4, \
call_11_1_0_5, \
call_11_1_0_6, \
call_11_1_0_7, \
call_11_1_0_8, \
call_11_1_0_9, \
call_11_1_0_10, \
call_11_1_0_11, \
call_11_1_0_12, \
call_11_1_0_13, \
call_11_1_0_14, \
call_11_1_0_15, \
call_11_1_1_0, \
call_11_1_1_1, \
call_11_1_1_2, \
call_11_1_1_3, \
call_11_1_1_4, \
call_11_1_1_5, \
call_11_1_1_6, \
call_11_1_1_7, \
call_11_1_1_8, \
call_11_1_1_9, \
call_11_1_1_10, \
call_11_1_1_11, \
call_11_1_1_12, \
call_11_1_1_13, \
call_11_1_1_14, \
call_11_1_1_15, \
call_11_1_2_0, \
call_11_1_2_1, \
call_11_1_2_2, \
call_11_1_2_3, \
call_11_1_2_4, \
call_11_1_2_5, \
call_11_1_2_6, \
call_11_1_2_7, \
call_11_1_2_8, \
call_11_1_2_9, \
call_11_1_2_10, \
call_11_1_2_11, \
call_11_1_2_12, \
call_11_1_2_13, \
call_11_1_2_14, \
call_11_1_2_15, \
call_11_1_3_0, \
call_11_1_3_1, \
call_11_1_3_2, \
call_11_1_3_3, \
call_11_1_3_4, \
call_11_1_3_5, \
call_11_1_3_6, \
call_11_1_3_7, \
call_11_1_3_8, \
call_11_1_3_9, \
call_11_1_3_10, \
call_11_1_3_11, \
call_11_1_3_12, \
call_11_1_3_13, \
call_11_1_3_14, \
call_11_1_3_15, \
call_11_1_4_0, \
call_11_1_4_1, \
call_11_1_4_2, \
call_11_1_4_3, \
call_11_1_4_4, \
call_11_1_4_5, \
call_11_1_4_6, \
call_11_1_4_7, \
call_11_1_4_8, \
call_11_1_4_9, \
call_11_1_4_10, \
call_11_1_4_11, \
call_11_1_4_12, \
call_11_1_4_13, \
call_11_1_4_14, \
call_11_1_4_15, \
call_11_1_5_0, \
call_11_1_5_1, \
call_11_1_5_2, \
call_11_1_5_3, \
call_11_1_5_4, \
call_11_1_5_5, \
call_11_1_5_6, \
call_11_1_5_7, \
call_11_1_5_8, \
call_11_1_5_9, \
call_11_1_5_10, \
call_11_1_5_11, \
call_11_1_5_12, \
call_11_1_5_13, \
call_11_1_5_14, \
call_11_1_5_15, \
call_11_1_6_0, \
call_11_1_6_1, \
call_11_1_6_2, \
call_11_1_6_3, \
call_11_1_6_4, \
call_11_1_6_5, \
call_11_1_6_6, \
call_11_1_6_7, \
call_11_1_6_8, \
call_11_1_6_9, \
call_11_1_6_10, \
call_11_1_6_11, \
call_11_1_6_12, \
call_11_1_6_13, \
call_11_1_6_14, \
call_11_1_6_15, \
call_11_1_7_0, \
call_11_1_7_1, \
call_11_1_7_2, \
call_11_1_7_3, \
call_11_1_7_4, \
call_11_1_7_5, \
call_11_1_7_6, \
call_11_1_7_7, \
call_11_1_7_8, \
call_11_1_7_9, \
call_11_1_7_10, \
call_11_1_7_11, \
call_11_1_7_12, \
call_11_1_7_13, \
call_11_1_7_14, \
call_11_1_7_15, \
call_11_1_8_0, \
call_11_1_8_1, \
call_11_1_8_2, \
call_11_1_8_3, \
call_11_1_8_4, \
call_11_1_8_5, \
call_11_1_8_6, \
call_11_1_8_7, \
call_11_1_8_8, \
call_11_1_8_9, \
call_11_1_8_10, \
call_11_1_8_11, \
call_11_1_8_12, \
call_11_1_8_13, \
call_11_1_8_14, \
call_11_1_8_15, \
call_11_1_9_0, \
call_11_1_9_1, \
call_11_1_9_2, \
call_11_1_9_3, \
call_11_1_9_4, \
call_11_1_9_5, \
call_11_1_9_6, \
call_11_1_9_7, \
call_11_1_9_8, \
call_11_1_9_9, \
call_11_1_9_10, \
call_11_1_9_11, \
call_11_1_9_12, \
call_11_1_9_13, \
call_11_1_9_14, \
call_11_1_9_15, \
call_11_1_10_0, \
call_11_1_10_1, \
call_11_1_10_2, \
call_11_1_10_3, \
call_11_1_10_4, \
call_11_1_10_5, \
call_11_1_10_6, \
call_11_1_10_7, \
call_11_1_10_8, \
call_11_1_10_9, \
call_11_1_10_10, \
call_11_1_10_11, \
call_11_1_10_12, \
call_11_1_10_13, \
call_11_1_10_14, \
call_11_1_10_15, \
call_11_1_11_0, \
call_11_1_11_1, \
call_11_1_11_2, \
call_11_1_11_3, \
call_11_1_11_4, \
call_11_1_11_5, \
call_11_1_11_6, \
call_11_1_11_7, \
call_11_1_11_8, \
call_11_1_11_9, \
call_11_1_11_10, \
call_11_1_11_11, \
call_11_1_11_12, \
call_11_1_11_13, \
call_11_1_11_14, \
call_11_1_11_15, \
call_11_1_12_0, \
call_11_1_12_1, \
call_11_1_12_2, \
call_11_1_12_3, \
call_11_1_12_4, \
call_11_1_12_5, \
call_11_1_12_6, \
call_11_1_12_7, \
call_11_1_12_8, \
call_11_1_12_9, \
call_11_1_12_10, \
call_11_1_12_11, \
call_11_1_12_12, \
call_11_1_12_13, \
call_11_1_12_14, \
call_11_1_12_15, \
call_11_1_13_0, \
call_11_1_13_1, \
call_11_1_13_2, \
call_11_1_13_3, \
call_11_1_13_4, \
call_11_1_13_5, \
call_11_1_13_6, \
call_11_1_13_7, \
call_11_1_13_8, \
call_11_1_13_9, \
call_11_1_13_10, \
call_11_1_13_11, \
call_11_1_13_12, \
call_11_1_13_13, \
call_11_1_13_14, \
call_11_1_13_15, \
call_11_1_14_0, \
call_11_1_14_1, \
call_11_1_14_2, \
call_11_1_14_3, \
call_11_1_14_4, \
call_11_1_14_5, \
call_11_1_14_6, \
call_11_1_14_7, \
call_11_1_14_8, \
call_11_1_14_9, \
call_11_1_14_10, \
call_11_1_14_11, \
call_11_1_14_12, \
call_11_1_14_13, \
call_11_1_14_14, \
call_11_1_14_15, \
call_11_1_15_0, \
call_11_1_15_1, \
call_11_1_15_2, \
call_11_1_15_3, \
call_11_1_15_4, \
call_11_1_15_5, \
call_11_1_15_6, \
call_11_1_15_7, \
call_11_1_15_8, \
call_11_1_15_9, \
call_11_1_15_10, \
call_11_1_15_11, \
call_11_1_15_12, \
call_11_1_15_13, \
call_11_1_15_14, \
call_11_1_15_15, \
call_11_2_0_0, \
call_11_2_0_1, \
call_11_2_0_2, \
call_11_2_0_3, \
call_11_2_0_4, \
call_11_2_0_5, \
call_11_2_0_6, \
call_11_2_0_7, \
call_11_2_0_8, \
call_11_2_0_9, \
call_11_2_0_10, \
call_11_2_0_11, \
call_11_2_0_12, \
call_11_2_0_13, \
call_11_2_0_14, \
call_11_2_0_15, \
call_11_2_1_0, \
call_11_2_1_1, \
call_11_2_1_2, \
call_11_2_1_3, \
call_11_2_1_4, \
call_11_2_1_5, \
call_11_2_1_6, \
call_11_2_1_7, \
call_11_2_1_8, \
call_11_2_1_9, \
call_11_2_1_10, \
call_11_2_1_11, \
call_11_2_1_12, \
call_11_2_1_13, \
call_11_2_1_14, \
call_11_2_1_15, \
call_11_2_2_0, \
call_11_2_2_1, \
call_11_2_2_2, \
call_11_2_2_3, \
call_11_2_2_4, \
call_11_2_2_5, \
call_11_2_2_6, \
call_11_2_2_7, \
call_11_2_2_8, \
call_11_2_2_9, \
call_11_2_2_10, \
call_11_2_2_11, \
call_11_2_2_12, \
call_11_2_2_13, \
call_11_2_2_14, \
call_11_2_2_15, \
call_11_2_3_0, \
call_11_2_3_1, \
call_11_2_3_2, \
call_11_2_3_3, \
call_11_2_3_4, \
call_11_2_3_5, \
call_11_2_3_6, \
call_11_2_3_7, \
call_11_2_3_8, \
call_11_2_3_9, \
call_11_2_3_10, \
call_11_2_3_11, \
call_11_2_3_12, \
call_11_2_3_13, \
call_11_2_3_14, \
call_11_2_3_15, \
call_11_2_4_0, \
call_11_2_4_1, \
call_11_2_4_2, \
call_11_2_4_3, \
call_11_2_4_4, \
call_11_2_4_5, \
call_11_2_4_6, \
call_11_2_4_7, \
call_11_2_4_8, \
call_11_2_4_9, \
call_11_2_4_10, \
call_11_2_4_11, \
call_11_2_4_12, \
call_11_2_4_13, \
call_11_2_4_14, \
call_11_2_4_15, \
call_11_2_5_0, \
call_11_2_5_1, \
call_11_2_5_2, \
call_11_2_5_3, \
call_11_2_5_4, \
call_11_2_5_5, \
call_11_2_5_6, \
call_11_2_5_7, \
call_11_2_5_8, \
call_11_2_5_9, \
call_11_2_5_10, \
call_11_2_5_11, \
call_11_2_5_12, \
call_11_2_5_13, \
call_11_2_5_14, \
call_11_2_5_15, \
call_11_2_6_0, \
call_11_2_6_1, \
call_11_2_6_2, \
call_11_2_6_3, \
call_11_2_6_4, \
call_11_2_6_5, \
call_11_2_6_6, \
call_11_2_6_7, \
call_11_2_6_8, \
call_11_2_6_9, \
call_11_2_6_10, \
call_11_2_6_11, \
call_11_2_6_12, \
call_11_2_6_13, \
call_11_2_6_14, \
call_11_2_6_15, \
call_11_2_7_0, \
call_11_2_7_1, \
call_11_2_7_2, \
call_11_2_7_3, \
call_11_2_7_4, \
call_11_2_7_5, \
call_11_2_7_6, \
call_11_2_7_7, \
call_11_2_7_8, \
call_11_2_7_9, \
call_11_2_7_10, \
call_11_2_7_11, \
call_11_2_7_12, \
call_11_2_7_13, \
call_11_2_7_14, \
call_11_2_7_15, \
call_11_2_8_0, \
call_11_2_8_1, \
call_11_2_8_2, \
call_11_2_8_3, \
call_11_2_8_4, \
call_11_2_8_5, \
call_11_2_8_6, \
call_11_2_8_7, \
call_11_2_8_8, \
call_11_2_8_9, \
call_11_2_8_10, \
call_11_2_8_11, \
call_11_2_8_12, \
call_11_2_8_13, \
call_11_2_8_14, \
call_11_2_8_15, \
call_11_2_9_0, \
call_11_2_9_1, \
call_11_2_9_2, \
call_11_2_9_3, \
call_11_2_9_4, \
call_11_2_9_5, \
call_11_2_9_6, \
call_11_2_9_7, \
call_11_2_9_8, \
call_11_2_9_9, \
call_11_2_9_10, \
call_11_2_9_11, \
call_11_2_9_12, \
call_11_2_9_13, \
call_11_2_9_14, \
call_11_2_9_15, \
call_11_2_10_0, \
call_11_2_10_1, \
call_11_2_10_2, \
call_11_2_10_3, \
call_11_2_10_4, \
call_11_2_10_5, \
call_11_2_10_6, \
call_11_2_10_7, \
call_11_2_10_8, \
call_11_2_10_9, \
call_11_2_10_10, \
call_11_2_10_11, \
call_11_2_10_12, \
call_11_2_10_13, \
call_11_2_10_14, \
call_11_2_10_15, \
call_11_2_11_0, \
call_11_2_11_1, \
call_11_2_11_2, \
call_11_2_11_3, \
call_11_2_11_4, \
call_11_2_11_5, \
call_11_2_11_6, \
call_11_2_11_7, \
call_11_2_11_8, \
call_11_2_11_9, \
call_11_2_11_10, \
call_11_2_11_11, \
call_11_2_11_12, \
call_11_2_11_13, \
call_11_2_11_14, \
call_11_2_11_15, \
call_11_2_12_0, \
call_11_2_12_1, \
call_11_2_12_2, \
call_11_2_12_3, \
call_11_2_12_4, \
call_11_2_12_5, \
call_11_2_12_6, \
call_11_2_12_7, \
call_11_2_12_8, \
call_11_2_12_9, \
call_11_2_12_10, \
call_11_2_12_11, \
call_11_2_12_12, \
call_11_2_12_13, \
call_11_2_12_14, \
call_11_2_12_15, \
call_11_2_13_0, \
call_11_2_13_1, \
call_11_2_13_2, \
call_11_2_13_3, \
call_11_2_13_4, \
call_11_2_13_5, \
call_11_2_13_6, \
call_11_2_13_7, \
call_11_2_13_8, \
call_11_2_13_9, \
call_11_2_13_10, \
call_11_2_13_11, \
call_11_2_13_12, \
call_11_2_13_13, \
call_11_2_13_14, \
call_11_2_13_15, \
call_11_2_14_0, \
call_11_2_14_1, \
call_11_2_14_2, \
call_11_2_14_3, \
call_11_2_14_4, \
call_11_2_14_5, \
call_11_2_14_6, \
call_11_2_14_7, \
call_11_2_14_8, \
call_11_2_14_9, \
call_11_2_14_10, \
call_11_2_14_11, \
call_11_2_14_12, \
call_11_2_14_13, \
call_11_2_14_14, \
call_11_2_14_15, \
call_11_2_15_0, \
call_11_2_15_1, \
call_11_2_15_2, \
call_11_2_15_3, \
call_11_2_15_4, \
call_11_2_15_5, \
call_11_2_15_6, \
call_11_2_15_7, \
call_11_2_15_8, \
call_11_2_15_9, \
call_11_2_15_10, \
call_11_2_15_11, \
call_11_2_15_12, \
call_11_2_15_13, \
call_11_2_15_14, \
call_11_2_15_15, \
call_11_3_0_0, \
call_11_3_0_1, \
call_11_3_0_2, \
call_11_3_0_3, \
call_11_3_0_4, \
call_11_3_0_5, \
call_11_3_0_6, \
call_11_3_0_7, \
call_11_3_0_8, \
call_11_3_0_9, \
call_11_3_0_10, \
call_11_3_0_11, \
call_11_3_0_12, \
call_11_3_0_13, \
call_11_3_0_14, \
call_11_3_0_15, \
call_11_3_1_0, \
call_11_3_1_1, \
call_11_3_1_2, \
call_11_3_1_3, \
call_11_3_1_4, \
call_11_3_1_5, \
call_11_3_1_6, \
call_11_3_1_7, \
call_11_3_1_8, \
call_11_3_1_9, \
call_11_3_1_10, \
call_11_3_1_11, \
call_11_3_1_12, \
call_11_3_1_13, \
call_11_3_1_14, \
call_11_3_1_15, \
call_11_3_2_0, \
call_11_3_2_1, \
call_11_3_2_2, \
call_11_3_2_3, \
call_11_3_2_4, \
call_11_3_2_5, \
call_11_3_2_6, \
call_11_3_2_7, \
call_11_3_2_8, \
call_11_3_2_9, \
call_11_3_2_10, \
call_11_3_2_11, \
call_11_3_2_12, \
call_11_3_2_13, \
call_11_3_2_14, \
call_11_3_2_15, \
call_11_3_3_0, \
call_11_3_3_1, \
call_11_3_3_2, \
call_11_3_3_3, \
call_11_3_3_4, \
call_11_3_3_5, \
call_11_3_3_6, \
call_11_3_3_7, \
call_11_3_3_8, \
call_11_3_3_9, \
call_11_3_3_10, \
call_11_3_3_11, \
call_11_3_3_12, \
call_11_3_3_13, \
call_11_3_3_14, \
call_11_3_3_15, \
call_11_3_4_0, \
call_11_3_4_1, \
call_11_3_4_2, \
call_11_3_4_3, \
call_11_3_4_4, \
call_11_3_4_5, \
call_11_3_4_6, \
call_11_3_4_7, \
call_11_3_4_8, \
call_11_3_4_9, \
call_11_3_4_10, \
call_11_3_4_11, \
call_11_3_4_12, \
call_11_3_4_13, \
call_11_3_4_14, \
call_11_3_4_15, \
call_11_3_5_0, \
call_11_3_5_1, \
call_11_3_5_2, \
call_11_3_5_3, \
call_11_3_5_4, \
call_11_3_5_5, \
call_11_3_5_6, \
call_11_3_5_7, \
call_11_3_5_8, \
call_11_3_5_9, \
call_11_3_5_10, \
call_11_3_5_11, \
call_11_3_5_12, \
call_11_3_5_13, \
call_11_3_5_14, \
call_11_3_5_15, \
call_11_3_6_0, \
call_11_3_6_1, \
call_11_3_6_2, \
call_11_3_6_3, \
call_11_3_6_4, \
call_11_3_6_5, \
call_11_3_6_6, \
call_11_3_6_7, \
call_11_3_6_8, \
call_11_3_6_9, \
call_11_3_6_10, \
call_11_3_6_11, \
call_11_3_6_12, \
call_11_3_6_13, \
call_11_3_6_14, \
call_11_3_6_15, \
call_11_3_7_0, \
call_11_3_7_1, \
call_11_3_7_2, \
call_11_3_7_3, \
call_11_3_7_4, \
call_11_3_7_5, \
call_11_3_7_6, \
call_11_3_7_7, \
call_11_3_7_8, \
call_11_3_7_9, \
call_11_3_7_10, \
call_11_3_7_11, \
call_11_3_7_12, \
call_11_3_7_13, \
call_11_3_7_14, \
call_11_3_7_15, \
call_11_3_8_0, \
call_11_3_8_1, \
call_11_3_8_2, \
call_11_3_8_3, \
call_11_3_8_4, \
call_11_3_8_5, \
call_11_3_8_6, \
call_11_3_8_7, \
call_11_3_8_8, \
call_11_3_8_9, \
call_11_3_8_10, \
call_11_3_8_11, \
call_11_3_8_12, \
call_11_3_8_13, \
call_11_3_8_14, \
call_11_3_8_15, \
call_11_3_9_0, \
call_11_3_9_1, \
call_11_3_9_2, \
call_11_3_9_3, \
call_11_3_9_4, \
call_11_3_9_5, \
call_11_3_9_6, \
call_11_3_9_7, \
call_11_3_9_8, \
call_11_3_9_9, \
call_11_3_9_10, \
call_11_3_9_11, \
call_11_3_9_12, \
call_11_3_9_13, \
call_11_3_9_14, \
call_11_3_9_15, \
call_11_3_10_0, \
call_11_3_10_1, \
call_11_3_10_2, \
call_11_3_10_3, \
call_11_3_10_4, \
call_11_3_10_5, \
call_11_3_10_6, \
call_11_3_10_7, \
call_11_3_10_8, \
call_11_3_10_9, \
call_11_3_10_10, \
call_11_3_10_11, \
call_11_3_10_12, \
call_11_3_10_13, \
call_11_3_10_14, \
call_11_3_10_15, \
call_11_3_11_0, \
call_11_3_11_1, \
call_11_3_11_2, \
call_11_3_11_3, \
call_11_3_11_4, \
call_11_3_11_5, \
call_11_3_11_6, \
call_11_3_11_7, \
call_11_3_11_8, \
call_11_3_11_9, \
call_11_3_11_10, \
call_11_3_11_11, \
call_11_3_11_12, \
call_11_3_11_13, \
call_11_3_11_14, \
call_11_3_11_15, \
call_11_3_12_0, \
call_11_3_12_1, \
call_11_3_12_2, \
call_11_3_12_3, \
call_11_3_12_4, \
call_11_3_12_5, \
call_11_3_12_6, \
call_11_3_12_7, \
call_11_3_12_8, \
call_11_3_12_9, \
call_11_3_12_10, \
call_11_3_12_11, \
call_11_3_12_12, \
call_11_3_12_13, \
call_11_3_12_14, \
call_11_3_12_15, \
call_11_3_13_0, \
call_11_3_13_1, \
call_11_3_13_2, \
call_11_3_13_3, \
call_11_3_13_4, \
call_11_3_13_5, \
call_11_3_13_6, \
call_11_3_13_7, \
call_11_3_13_8, \
call_11_3_13_9, \
call_11_3_13_10, \
call_11_3_13_11, \
call_11_3_13_12, \
call_11_3_13_13, \
call_11_3_13_14, \
call_11_3_13_15, \
call_11_3_14_0, \
call_11_3_14_1, \
call_11_3_14_2, \
call_11_3_14_3, \
call_11_3_14_4, \
call_11_3_14_5, \
call_11_3_14_6, \
call_11_3_14_7, \
call_11_3_14_8, \
call_11_3_14_9, \
call_11_3_14_10, \
call_11_3_14_11, \
call_11_3_14_12, \
call_11_3_14_13, \
call_11_3_14_14, \
call_11_3_14_15, \
call_11_3_15_0, \
call_11_3_15_1, \
call_11_3_15_2, \
call_11_3_15_3, \
call_11_3_15_4, \
call_11_3_15_5, \
call_11_3_15_6, \
call_11_3_15_7, \
call_11_3_15_8, \
call_11_3_15_9, \
call_11_3_15_10, \
call_11_3_15_11, \
call_11_3_15_12, \
call_11_3_15_13, \
call_11_3_15_14, \
call_11_3_15_15, \
call_11_4_0_0, \
call_11_4_0_1, \
call_11_4_0_2, \
call_11_4_0_3, \
call_11_4_0_4, \
call_11_4_0_5, \
call_11_4_0_6, \
call_11_4_0_7, \
call_11_4_0_8, \
call_11_4_0_9, \
call_11_4_0_10, \
call_11_4_0_11, \
call_11_4_0_12, \
call_11_4_0_13, \
call_11_4_0_14, \
call_11_4_0_15, \
call_11_4_1_0, \
call_11_4_1_1, \
call_11_4_1_2, \
call_11_4_1_3, \
call_11_4_1_4, \
call_11_4_1_5, \
call_11_4_1_6, \
call_11_4_1_7, \
call_11_4_1_8, \
call_11_4_1_9, \
call_11_4_1_10, \
call_11_4_1_11, \
call_11_4_1_12, \
call_11_4_1_13, \
call_11_4_1_14, \
call_11_4_1_15, \
call_11_4_2_0, \
call_11_4_2_1, \
call_11_4_2_2, \
call_11_4_2_3, \
call_11_4_2_4, \
call_11_4_2_5, \
call_11_4_2_6, \
call_11_4_2_7, \
call_11_4_2_8, \
call_11_4_2_9, \
call_11_4_2_10, \
call_11_4_2_11, \
call_11_4_2_12, \
call_11_4_2_13, \
call_11_4_2_14, \
call_11_4_2_15, \
call_11_4_3_0, \
call_11_4_3_1, \
call_11_4_3_2, \
call_11_4_3_3, \
call_11_4_3_4, \
call_11_4_3_5, \
call_11_4_3_6, \
call_11_4_3_7, \
call_11_4_3_8, \
call_11_4_3_9, \
call_11_4_3_10, \
call_11_4_3_11, \
call_11_4_3_12, \
call_11_4_3_13, \
call_11_4_3_14, \
call_11_4_3_15, \
call_11_4_4_0, \
call_11_4_4_1, \
call_11_4_4_2, \
call_11_4_4_3, \
call_11_4_4_4, \
call_11_4_4_5, \
call_11_4_4_6, \
call_11_4_4_7, \
call_11_4_4_8, \
call_11_4_4_9, \
call_11_4_4_10, \
call_11_4_4_11, \
call_11_4_4_12, \
call_11_4_4_13, \
call_11_4_4_14, \
call_11_4_4_15, \
call_11_4_5_0, \
call_11_4_5_1, \
call_11_4_5_2, \
call_11_4_5_3, \
call_11_4_5_4, \
call_11_4_5_5, \
call_11_4_5_6, \
call_11_4_5_7, \
call_11_4_5_8, \
call_11_4_5_9, \
call_11_4_5_10, \
call_11_4_5_11, \
call_11_4_5_12, \
call_11_4_5_13, \
call_11_4_5_14, \
call_11_4_5_15, \
call_11_4_6_0, \
call_11_4_6_1, \
call_11_4_6_2, \
call_11_4_6_3, \
call_11_4_6_4, \
call_11_4_6_5, \
call_11_4_6_6, \
call_11_4_6_7, \
call_11_4_6_8, \
call_11_4_6_9, \
call_11_4_6_10, \
call_11_4_6_11, \
call_11_4_6_12, \
call_11_4_6_13, \
call_11_4_6_14, \
call_11_4_6_15, \
call_11_4_7_0, \
call_11_4_7_1, \
call_11_4_7_2, \
call_11_4_7_3, \
call_11_4_7_4, \
call_11_4_7_5, \
call_11_4_7_6, \
call_11_4_7_7, \
call_11_4_7_8, \
call_11_4_7_9, \
call_11_4_7_10, \
call_11_4_7_11, \
call_11_4_7_12, \
call_11_4_7_13, \
call_11_4_7_14, \
call_11_4_7_15, \
call_11_4_8_0, \
call_11_4_8_1, \
call_11_4_8_2, \
call_11_4_8_3, \
call_11_4_8_4, \
call_11_4_8_5, \
call_11_4_8_6, \
call_11_4_8_7, \
call_11_4_8_8, \
call_11_4_8_9, \
call_11_4_8_10, \
call_11_4_8_11, \
call_11_4_8_12, \
call_11_4_8_13, \
call_11_4_8_14, \
call_11_4_8_15, \
call_11_4_9_0, \
call_11_4_9_1, \
call_11_4_9_2, \
call_11_4_9_3, \
call_11_4_9_4, \
call_11_4_9_5, \
call_11_4_9_6, \
call_11_4_9_7, \
call_11_4_9_8, \
call_11_4_9_9, \
call_11_4_9_10, \
call_11_4_9_11, \
call_11_4_9_12, \
call_11_4_9_13, \
call_11_4_9_14, \
call_11_4_9_15, \
call_11_4_10_0, \
call_11_4_10_1, \
call_11_4_10_2, \
call_11_4_10_3, \
call_11_4_10_4, \
call_11_4_10_5, \
call_11_4_10_6, \
call_11_4_10_7, \
call_11_4_10_8, \
call_11_4_10_9, \
call_11_4_10_10, \
call_11_4_10_11, \
call_11_4_10_12, \
call_11_4_10_13, \
call_11_4_10_14, \
call_11_4_10_15, \
call_11_4_11_0, \
call_11_4_11_1, \
call_11_4_11_2, \
call_11_4_11_3, \
call_11_4_11_4, \
call_11_4_11_5, \
call_11_4_11_6, \
call_11_4_11_7, \
call_11_4_11_8, \
call_11_4_11_9, \
call_11_4_11_10, \
call_11_4_11_11, \
call_11_4_11_12, \
call_11_4_11_13, \
call_11_4_11_14, \
call_11_4_11_15, \
call_11_4_12_0, \
call_11_4_12_1, \
call_11_4_12_2, \
call_11_4_12_3, \
call_11_4_12_4, \
call_11_4_12_5, \
call_11_4_12_6, \
call_11_4_12_7, \
call_11_4_12_8, \
call_11_4_12_9, \
call_11_4_12_10, \
call_11_4_12_11, \
call_11_4_12_12, \
call_11_4_12_13, \
call_11_4_12_14, \
call_11_4_12_15, \
call_11_4_13_0, \
call_11_4_13_1, \
call_11_4_13_2, \
call_11_4_13_3, \
call_11_4_13_4, \
call_11_4_13_5, \
call_11_4_13_6, \
call_11_4_13_7, \
call_11_4_13_8, \
call_11_4_13_9, \
call_11_4_13_10, \
call_11_4_13_11, \
call_11_4_13_12, \
call_11_4_13_13, \
call_11_4_13_14, \
call_11_4_13_15, \
call_11_4_14_0, \
call_11_4_14_1, \
call_11_4_14_2, \
call_11_4_14_3, \
call_11_4_14_4, \
call_11_4_14_5, \
call_11_4_14_6, \
call_11_4_14_7, \
call_11_4_14_8, \
call_11_4_14_9, \
call_11_4_14_10, \
call_11_4_14_11, \
call_11_4_14_12, \
call_11_4_14_13, \
call_11_4_14_14, \
call_11_4_14_15, \
call_11_4_15_0, \
call_11_4_15_1, \
call_11_4_15_2, \
call_11_4_15_3, \
call_11_4_15_4, \
call_11_4_15_5, \
call_11_4_15_6, \
call_11_4_15_7, \
call_11_4_15_8, \
call_11_4_15_9, \
call_11_4_15_10, \
call_11_4_15_11, \
call_11_4_15_12, \
call_11_4_15_13, \
call_11_4_15_14, \
call_11_4_15_15, \
call_11_5_0_0, \
call_11_5_0_1, \
call_11_5_0_2, \
call_11_5_0_3, \
call_11_5_0_4, \
call_11_5_0_5, \
call_11_5_0_6, \
call_11_5_0_7, \
call_11_5_0_8, \
call_11_5_0_9, \
call_11_5_0_10, \
call_11_5_0_11, \
call_11_5_0_12, \
call_11_5_0_13, \
call_11_5_0_14, \
call_11_5_0_15, \
call_11_5_1_0, \
call_11_5_1_1, \
call_11_5_1_2, \
call_11_5_1_3, \
call_11_5_1_4, \
call_11_5_1_5, \
call_11_5_1_6, \
call_11_5_1_7, \
call_11_5_1_8, \
call_11_5_1_9, \
call_11_5_1_10, \
call_11_5_1_11, \
call_11_5_1_12, \
call_11_5_1_13, \
call_11_5_1_14, \
call_11_5_1_15, \
call_11_5_2_0, \
call_11_5_2_1, \
call_11_5_2_2, \
call_11_5_2_3, \
call_11_5_2_4, \
call_11_5_2_5, \
call_11_5_2_6, \
call_11_5_2_7, \
call_11_5_2_8, \
call_11_5_2_9, \
call_11_5_2_10, \
call_11_5_2_11, \
call_11_5_2_12, \
call_11_5_2_13, \
call_11_5_2_14, \
call_11_5_2_15, \
call_11_5_3_0, \
call_11_5_3_1, \
call_11_5_3_2, \
call_11_5_3_3, \
call_11_5_3_4, \
call_11_5_3_5, \
call_11_5_3_6, \
call_11_5_3_7, \
call_11_5_3_8, \
call_11_5_3_9, \
call_11_5_3_10, \
call_11_5_3_11, \
call_11_5_3_12, \
call_11_5_3_13, \
call_11_5_3_14, \
call_11_5_3_15, \
call_11_5_4_0, \
call_11_5_4_1, \
call_11_5_4_2, \
call_11_5_4_3, \
call_11_5_4_4, \
call_11_5_4_5, \
call_11_5_4_6, \
call_11_5_4_7, \
call_11_5_4_8, \
call_11_5_4_9, \
call_11_5_4_10, \
call_11_5_4_11, \
call_11_5_4_12, \
call_11_5_4_13, \
call_11_5_4_14, \
call_11_5_4_15, \
call_11_5_5_0, \
call_11_5_5_1, \
call_11_5_5_2, \
call_11_5_5_3, \
call_11_5_5_4, \
call_11_5_5_5, \
call_11_5_5_6, \
call_11_5_5_7, \
call_11_5_5_8, \
call_11_5_5_9, \
call_11_5_5_10, \
call_11_5_5_11, \
call_11_5_5_12, \
call_11_5_5_13, \
call_11_5_5_14, \
call_11_5_5_15, \
call_11_5_6_0, \
call_11_5_6_1, \
call_11_5_6_2, \
call_11_5_6_3, \
call_11_5_6_4, \
call_11_5_6_5, \
call_11_5_6_6, \
call_11_5_6_7, \
call_11_5_6_8, \
call_11_5_6_9, \
call_11_5_6_10, \
call_11_5_6_11, \
call_11_5_6_12, \
call_11_5_6_13, \
call_11_5_6_14, \
call_11_5_6_15, \
call_11_5_7_0, \
call_11_5_7_1, \
call_11_5_7_2, \
call_11_5_7_3, \
call_11_5_7_4, \
call_11_5_7_5, \
call_11_5_7_6, \
call_11_5_7_7, \
call_11_5_7_8, \
call_11_5_7_9, \
call_11_5_7_10, \
call_11_5_7_11, \
call_11_5_7_12, \
call_11_5_7_13, \
call_11_5_7_14, \
call_11_5_7_15, \
call_11_5_8_0, \
call_11_5_8_1, \
call_11_5_8_2, \
call_11_5_8_3, \
call_11_5_8_4, \
call_11_5_8_5, \
call_11_5_8_6, \
call_11_5_8_7, \
call_11_5_8_8, \
call_11_5_8_9, \
call_11_5_8_10, \
call_11_5_8_11, \
call_11_5_8_12, \
call_11_5_8_13, \
call_11_5_8_14, \
call_11_5_8_15, \
call_11_5_9_0, \
call_11_5_9_1, \
call_11_5_9_2, \
call_11_5_9_3, \
call_11_5_9_4, \
call_11_5_9_5, \
call_11_5_9_6, \
call_11_5_9_7, \
call_11_5_9_8, \
call_11_5_9_9, \
call_11_5_9_10, \
call_11_5_9_11, \
call_11_5_9_12, \
call_11_5_9_13, \
call_11_5_9_14, \
call_11_5_9_15, \
call_11_5_10_0, \
call_11_5_10_1, \
call_11_5_10_2, \
call_11_5_10_3, \
call_11_5_10_4, \
call_11_5_10_5, \
call_11_5_10_6, \
call_11_5_10_7, \
call_11_5_10_8, \
call_11_5_10_9, \
call_11_5_10_10, \
call_11_5_10_11, \
call_11_5_10_12, \
call_11_5_10_13, \
call_11_5_10_14, \
call_11_5_10_15, \
call_11_5_11_0, \
call_11_5_11_1, \
call_11_5_11_2, \
call_11_5_11_3, \
call_11_5_11_4, \
call_11_5_11_5, \
call_11_5_11_6, \
call_11_5_11_7, \
call_11_5_11_8, \
call_11_5_11_9, \
call_11_5_11_10, \
call_11_5_11_11, \
call_11_5_11_12, \
call_11_5_11_13, \
call_11_5_11_14, \
call_11_5_11_15, \
call_11_5_12_0, \
call_11_5_12_1, \
call_11_5_12_2, \
call_11_5_12_3, \
call_11_5_12_4, \
call_11_5_12_5, \
call_11_5_12_6, \
call_11_5_12_7, \
call_11_5_12_8, \
call_11_5_12_9, \
call_11_5_12_10, \
call_11_5_12_11, \
call_11_5_12_12, \
call_11_5_12_13, \
call_11_5_12_14, \
call_11_5_12_15, \
call_11_5_13_0, \
call_11_5_13_1, \
call_11_5_13_2, \
call_11_5_13_3, \
call_11_5_13_4, \
call_11_5_13_5, \
call_11_5_13_6, \
call_11_5_13_7, \
call_11_5_13_8, \
call_11_5_13_9, \
call_11_5_13_10, \
call_11_5_13_11, \
call_11_5_13_12, \
call_11_5_13_13, \
call_11_5_13_14, \
call_11_5_13_15, \
call_11_5_14_0, \
call_11_5_14_1, \
call_11_5_14_2, \
call_11_5_14_3, \
call_11_5_14_4, \
call_11_5_14_5, \
call_11_5_14_6, \
call_11_5_14_7, \
call_11_5_14_8, \
call_11_5_14_9, \
call_11_5_14_10, \
call_11_5_14_11, \
call_11_5_14_12, \
call_11_5_14_13, \
call_11_5_14_14, \
call_11_5_14_15, \
call_11_5_15_0, \
call_11_5_15_1, \
call_11_5_15_2, \
call_11_5_15_3, \
call_11_5_15_4, \
call_11_5_15_5, \
call_11_5_15_6, \
call_11_5_15_7, \
call_11_5_15_8, \
call_11_5_15_9, \
call_11_5_15_10, \
call_11_5_15_11, \
call_11_5_15_12, \
call_11_5_15_13, \
call_11_5_15_14, \
call_11_5_15_15, \
call_11_6_0_0, \
call_11_6_0_1, \
call_11_6_0_2, \
call_11_6_0_3, \
call_11_6_0_4, \
call_11_6_0_5, \
call_11_6_0_6, \
call_11_6_0_7, \
call_11_6_0_8, \
call_11_6_0_9, \
call_11_6_0_10, \
call_11_6_0_11, \
call_11_6_0_12, \
call_11_6_0_13, \
call_11_6_0_14, \
call_11_6_0_15, \
call_11_6_1_0, \
call_11_6_1_1, \
call_11_6_1_2, \
call_11_6_1_3, \
call_11_6_1_4, \
call_11_6_1_5, \
call_11_6_1_6, \
call_11_6_1_7, \
call_11_6_1_8, \
call_11_6_1_9, \
call_11_6_1_10, \
call_11_6_1_11, \
call_11_6_1_12, \
call_11_6_1_13, \
call_11_6_1_14, \
call_11_6_1_15, \
call_11_6_2_0, \
call_11_6_2_1, \
call_11_6_2_2, \
call_11_6_2_3, \
call_11_6_2_4, \
call_11_6_2_5, \
call_11_6_2_6, \
call_11_6_2_7, \
call_11_6_2_8, \
call_11_6_2_9, \
call_11_6_2_10, \
call_11_6_2_11, \
call_11_6_2_12, \
call_11_6_2_13, \
call_11_6_2_14, \
call_11_6_2_15, \
call_11_6_3_0, \
call_11_6_3_1, \
call_11_6_3_2, \
call_11_6_3_3, \
call_11_6_3_4, \
call_11_6_3_5, \
call_11_6_3_6, \
call_11_6_3_7, \
call_11_6_3_8, \
call_11_6_3_9, \
call_11_6_3_10, \
call_11_6_3_11, \
call_11_6_3_12, \
call_11_6_3_13, \
call_11_6_3_14, \
call_11_6_3_15, \
call_11_6_4_0, \
call_11_6_4_1, \
call_11_6_4_2, \
call_11_6_4_3, \
call_11_6_4_4, \
call_11_6_4_5, \
call_11_6_4_6, \
call_11_6_4_7, \
call_11_6_4_8, \
call_11_6_4_9, \
call_11_6_4_10, \
call_11_6_4_11, \
call_11_6_4_12, \
call_11_6_4_13, \
call_11_6_4_14, \
call_11_6_4_15, \
call_11_6_5_0, \
call_11_6_5_1, \
call_11_6_5_2, \
call_11_6_5_3, \
call_11_6_5_4, \
call_11_6_5_5, \
call_11_6_5_6, \
call_11_6_5_7, \
call_11_6_5_8, \
call_11_6_5_9, \
call_11_6_5_10, \
call_11_6_5_11, \
call_11_6_5_12, \
call_11_6_5_13, \
call_11_6_5_14, \
call_11_6_5_15, \
call_11_6_6_0, \
call_11_6_6_1, \
call_11_6_6_2, \
call_11_6_6_3, \
call_11_6_6_4, \
call_11_6_6_5, \
call_11_6_6_6, \
call_11_6_6_7, \
call_11_6_6_8, \
call_11_6_6_9, \
call_11_6_6_10, \
call_11_6_6_11, \
call_11_6_6_12, \
call_11_6_6_13, \
call_11_6_6_14, \
call_11_6_6_15, \
call_11_6_7_0, \
call_11_6_7_1, \
call_11_6_7_2, \
call_11_6_7_3, \
call_11_6_7_4, \
call_11_6_7_5, \
call_11_6_7_6, \
call_11_6_7_7, \
call_11_6_7_8, \
call_11_6_7_9, \
call_11_6_7_10, \
call_11_6_7_11, \
call_11_6_7_12, \
call_11_6_7_13, \
call_11_6_7_14, \
call_11_6_7_15, \
call_11_6_8_0, \
call_11_6_8_1, \
call_11_6_8_2, \
call_11_6_8_3, \
call_11_6_8_4, \
call_11_6_8_5, \
call_11_6_8_6, \
call_11_6_8_7, \
call_11_6_8_8, \
call_11_6_8_9, \
call_11_6_8_10, \
call_11_6_8_11, \
call_11_6_8_12, \
call_11_6_8_13, \
call_11_6_8_14, \
call_11_6_8_15, \
call_11_6_9_0, \
call_11_6_9_1, \
call_11_6_9_2, \
call_11_6_9_3, \
call_11_6_9_4, \
call_11_6_9_5, \
call_11_6_9_6, \
call_11_6_9_7, \
call_11_6_9_8, \
call_11_6_9_9, \
call_11_6_9_10, \
call_11_6_9_11, \
call_11_6_9_12, \
call_11_6_9_13, \
call_11_6_9_14, \
call_11_6_9_15, \
call_11_6_10_0, \
call_11_6_10_1, \
call_11_6_10_2, \
call_11_6_10_3, \
call_11_6_10_4, \
call_11_6_10_5, \
call_11_6_10_6, \
call_11_6_10_7, \
call_11_6_10_8, \
call_11_6_10_9, \
call_11_6_10_10, \
call_11_6_10_11, \
call_11_6_10_12, \
call_11_6_10_13, \
call_11_6_10_14, \
call_11_6_10_15, \
call_11_6_11_0, \
call_11_6_11_1, \
call_11_6_11_2, \
call_11_6_11_3, \
call_11_6_11_4, \
call_11_6_11_5, \
call_11_6_11_6, \
call_11_6_11_7, \
call_11_6_11_8, \
call_11_6_11_9, \
call_11_6_11_10, \
call_11_6_11_11, \
call_11_6_11_12, \
call_11_6_11_13, \
call_11_6_11_14, \
call_11_6_11_15, \
call_11_6_12_0, \
call_11_6_12_1, \
call_11_6_12_2, \
call_11_6_12_3, \
call_11_6_12_4, \
call_11_6_12_5, \
call_11_6_12_6, \
call_11_6_12_7, \
call_11_6_12_8, \
call_11_6_12_9, \
call_11_6_12_10, \
call_11_6_12_11, \
call_11_6_12_12, \
call_11_6_12_13, \
call_11_6_12_14, \
call_11_6_12_15, \
call_11_6_13_0, \
call_11_6_13_1, \
call_11_6_13_2, \
call_11_6_13_3, \
call_11_6_13_4, \
call_11_6_13_5, \
call_11_6_13_6, \
call_11_6_13_7, \
call_11_6_13_8, \
call_11_6_13_9, \
call_11_6_13_10, \
call_11_6_13_11, \
call_11_6_13_12, \
call_11_6_13_13, \
call_11_6_13_14, \
call_11_6_13_15, \
call_11_6_14_0, \
call_11_6_14_1, \
call_11_6_14_2, \
call_11_6_14_3, \
call_11_6_14_4, \
call_11_6_14_5, \
call_11_6_14_6, \
call_11_6_14_7, \
call_11_6_14_8, \
call_11_6_14_9, \
call_11_6_14_10, \
call_11_6_14_11, \
call_11_6_14_12, \
call_11_6_14_13, \
call_11_6_14_14, \
call_11_6_14_15, \
call_11_6_15_0, \
call_11_6_15_1, \
call_11_6_15_2, \
call_11_6_15_3, \
call_11_6_15_4, \
call_11_6_15_5, \
call_11_6_15_6, \
call_11_6_15_7, \
call_11_6_15_8, \
call_11_6_15_9, \
call_11_6_15_10, \
call_11_6_15_11, \
call_11_6_15_12, \
call_11_6_15_13, \
call_11_6_15_14, \
call_11_6_15_15, \
call_11_7_0_0, \
call_11_7_0_1, \
call_11_7_0_2, \
call_11_7_0_3, \
call_11_7_0_4, \
call_11_7_0_5, \
call_11_7_0_6, \
call_11_7_0_7, \
call_11_7_0_8, \
call_11_7_0_9, \
call_11_7_0_10, \
call_11_7_0_11, \
call_11_7_0_12, \
call_11_7_0_13, \
call_11_7_0_14, \
call_11_7_0_15, \
call_11_7_1_0, \
call_11_7_1_1, \
call_11_7_1_2, \
call_11_7_1_3, \
call_11_7_1_4, \
call_11_7_1_5, \
call_11_7_1_6, \
call_11_7_1_7, \
call_11_7_1_8, \
call_11_7_1_9, \
call_11_7_1_10, \
call_11_7_1_11, \
call_11_7_1_12, \
call_11_7_1_13, \
call_11_7_1_14, \
call_11_7_1_15, \
call_11_7_2_0, \
call_11_7_2_1, \
call_11_7_2_2, \
call_11_7_2_3, \
call_11_7_2_4, \
call_11_7_2_5, \
call_11_7_2_6, \
call_11_7_2_7, \
call_11_7_2_8, \
call_11_7_2_9, \
call_11_7_2_10, \
call_11_7_2_11, \
call_11_7_2_12, \
call_11_7_2_13, \
call_11_7_2_14, \
call_11_7_2_15, \
call_11_7_3_0, \
call_11_7_3_1, \
call_11_7_3_2, \
call_11_7_3_3, \
call_11_7_3_4, \
call_11_7_3_5, \
call_11_7_3_6, \
call_11_7_3_7, \
call_11_7_3_8, \
call_11_7_3_9, \
call_11_7_3_10, \
call_11_7_3_11, \
call_11_7_3_12, \
call_11_7_3_13, \
call_11_7_3_14, \
call_11_7_3_15, \
call_11_7_4_0, \
call_11_7_4_1, \
call_11_7_4_2, \
call_11_7_4_3, \
call_11_7_4_4, \
call_11_7_4_5, \
call_11_7_4_6, \
call_11_7_4_7, \
call_11_7_4_8, \
call_11_7_4_9, \
call_11_7_4_10, \
call_11_7_4_11, \
call_11_7_4_12, \
call_11_7_4_13, \
call_11_7_4_14, \
call_11_7_4_15, \
call_11_7_5_0, \
call_11_7_5_1, \
call_11_7_5_2, \
call_11_7_5_3, \
call_11_7_5_4, \
call_11_7_5_5, \
call_11_7_5_6, \
call_11_7_5_7, \
call_11_7_5_8, \
call_11_7_5_9, \
call_11_7_5_10, \
call_11_7_5_11, \
call_11_7_5_12, \
call_11_7_5_13, \
call_11_7_5_14, \
call_11_7_5_15, \
call_11_7_6_0, \
call_11_7_6_1, \
call_11_7_6_2, \
call_11_7_6_3, \
call_11_7_6_4, \
call_11_7_6_5, \
call_11_7_6_6, \
call_11_7_6_7, \
call_11_7_6_8, \
call_11_7_6_9, \
call_11_7_6_10, \
call_11_7_6_11, \
call_11_7_6_12, \
call_11_7_6_13, \
call_11_7_6_14, \
call_11_7_6_15, \
call_11_7_7_0, \
call_11_7_7_1, \
call_11_7_7_2, \
call_11_7_7_3, \
call_11_7_7_4, \
call_11_7_7_5, \
call_11_7_7_6, \
call_11_7_7_7, \
call_11_7_7_8, \
call_11_7_7_9, \
call_11_7_7_10, \
call_11_7_7_11, \
call_11_7_7_12, \
call_11_7_7_13, \
call_11_7_7_14, \
call_11_7_7_15, \
call_11_7_8_0, \
call_11_7_8_1, \
call_11_7_8_2, \
call_11_7_8_3, \
call_11_7_8_4, \
call_11_7_8_5, \
call_11_7_8_6, \
call_11_7_8_7, \
call_11_7_8_8, \
call_11_7_8_9, \
call_11_7_8_10, \
call_11_7_8_11, \
call_11_7_8_12, \
call_11_7_8_13, \
call_11_7_8_14, \
call_11_7_8_15, \
call_11_7_9_0, \
call_11_7_9_1, \
call_11_7_9_2, \
call_11_7_9_3, \
call_11_7_9_4, \
call_11_7_9_5, \
call_11_7_9_6, \
call_11_7_9_7, \
call_11_7_9_8, \
call_11_7_9_9, \
call_11_7_9_10, \
call_11_7_9_11, \
call_11_7_9_12, \
call_11_7_9_13, \
call_11_7_9_14, \
call_11_7_9_15, \
call_11_7_10_0, \
call_11_7_10_1, \
call_11_7_10_2, \
call_11_7_10_3, \
call_11_7_10_4, \
call_11_7_10_5, \
call_11_7_10_6, \
call_11_7_10_7, \
call_11_7_10_8, \
call_11_7_10_9, \
call_11_7_10_10, \
call_11_7_10_11, \
call_11_7_10_12, \
call_11_7_10_13, \
call_11_7_10_14, \
call_11_7_10_15, \
call_11_7_11_0, \
call_11_7_11_1, \
call_11_7_11_2, \
call_11_7_11_3, \
call_11_7_11_4, \
call_11_7_11_5, \
call_11_7_11_6, \
call_11_7_11_7, \
call_11_7_11_8, \
call_11_7_11_9, \
call_11_7_11_10, \
call_11_7_11_11, \
call_11_7_11_12, \
call_11_7_11_13, \
call_11_7_11_14, \
call_11_7_11_15, \
call_11_7_12_0, \
call_11_7_12_1, \
call_11_7_12_2, \
call_11_7_12_3, \
call_11_7_12_4, \
call_11_7_12_5, \
call_11_7_12_6, \
call_11_7_12_7, \
call_11_7_12_8, \
call_11_7_12_9, \
call_11_7_12_10, \
call_11_7_12_11, \
call_11_7_12_12, \
call_11_7_12_13, \
call_11_7_12_14, \
call_11_7_12_15, \
call_11_7_13_0, \
call_11_7_13_1, \
call_11_7_13_2, \
call_11_7_13_3, \
call_11_7_13_4, \
call_11_7_13_5, \
call_11_7_13_6, \
call_11_7_13_7, \
call_11_7_13_8, \
call_11_7_13_9, \
call_11_7_13_10, \
call_11_7_13_11, \
call_11_7_13_12, \
call_11_7_13_13, \
call_11_7_13_14, \
call_11_7_13_15, \
call_11_7_14_0, \
call_11_7_14_1, \
call_11_7_14_2, \
call_11_7_14_3, \
call_11_7_14_4, \
call_11_7_14_5, \
call_11_7_14_6, \
call_11_7_14_7, \
call_11_7_14_8, \
call_11_7_14_9, \
call_11_7_14_10, \
call_11_7_14_11, \
call_11_7_14_12, \
call_11_7_14_13, \
call_11_7_14_14, \
call_11_7_14_15, \
call_11_7_15_0, \
call_11_7_15_1, \
call_11_7_15_2, \
call_11_7_15_3, \
call_11_7_15_4, \
call_11_7_15_5, \
call_11_7_15_6, \
call_11_7_15_7, \
call_11_7_15_8, \
call_11_7_15_9, \
call_11_7_15_10, \
call_11_7_15_11, \
call_11_7_15_12, \
call_11_7_15_13, \
call_11_7_15_14, \
call_11_7_15_15, \
call_11_8_0_0, \
call_11_8_0_1, \
call_11_8_0_2, \
call_11_8_0_3, \
call_11_8_0_4, \
call_11_8_0_5, \
call_11_8_0_6, \
call_11_8_0_7, \
call_11_8_0_8, \
call_11_8_0_9, \
call_11_8_0_10, \
call_11_8_0_11, \
call_11_8_0_12, \
call_11_8_0_13, \
call_11_8_0_14, \
call_11_8_0_15, \
call_11_8_1_0, \
call_11_8_1_1, \
call_11_8_1_2, \
call_11_8_1_3, \
call_11_8_1_4, \
call_11_8_1_5, \
call_11_8_1_6, \
call_11_8_1_7, \
call_11_8_1_8, \
call_11_8_1_9, \
call_11_8_1_10, \
call_11_8_1_11, \
call_11_8_1_12, \
call_11_8_1_13, \
call_11_8_1_14, \
call_11_8_1_15, \
call_11_8_2_0, \
call_11_8_2_1, \
call_11_8_2_2, \
call_11_8_2_3, \
call_11_8_2_4, \
call_11_8_2_5, \
call_11_8_2_6, \
call_11_8_2_7, \
call_11_8_2_8, \
call_11_8_2_9, \
call_11_8_2_10, \
call_11_8_2_11, \
call_11_8_2_12, \
call_11_8_2_13, \
call_11_8_2_14, \
call_11_8_2_15, \
call_11_8_3_0, \
call_11_8_3_1, \
call_11_8_3_2, \
call_11_8_3_3, \
call_11_8_3_4, \
call_11_8_3_5, \
call_11_8_3_6, \
call_11_8_3_7, \
call_11_8_3_8, \
call_11_8_3_9, \
call_11_8_3_10, \
call_11_8_3_11, \
call_11_8_3_12, \
call_11_8_3_13, \
call_11_8_3_14, \
call_11_8_3_15, \
call_11_8_4_0, \
call_11_8_4_1, \
call_11_8_4_2, \
call_11_8_4_3, \
call_11_8_4_4, \
call_11_8_4_5, \
call_11_8_4_6, \
call_11_8_4_7, \
call_11_8_4_8, \
call_11_8_4_9, \
call_11_8_4_10, \
call_11_8_4_11, \
call_11_8_4_12, \
call_11_8_4_13, \
call_11_8_4_14, \
call_11_8_4_15, \
call_11_8_5_0, \
call_11_8_5_1, \
call_11_8_5_2, \
call_11_8_5_3, \
call_11_8_5_4, \
call_11_8_5_5, \
call_11_8_5_6, \
call_11_8_5_7, \
call_11_8_5_8, \
call_11_8_5_9, \
call_11_8_5_10, \
call_11_8_5_11, \
call_11_8_5_12, \
call_11_8_5_13, \
call_11_8_5_14, \
call_11_8_5_15, \
call_11_8_6_0, \
call_11_8_6_1, \
call_11_8_6_2, \
call_11_8_6_3, \
call_11_8_6_4, \
call_11_8_6_5, \
call_11_8_6_6, \
call_11_8_6_7, \
call_11_8_6_8, \
call_11_8_6_9, \
call_11_8_6_10, \
call_11_8_6_11, \
call_11_8_6_12, \
call_11_8_6_13, \
call_11_8_6_14, \
call_11_8_6_15, \
call_11_8_7_0, \
call_11_8_7_1, \
call_11_8_7_2, \
call_11_8_7_3, \
call_11_8_7_4, \
call_11_8_7_5, \
call_11_8_7_6, \
call_11_8_7_7, \
call_11_8_7_8, \
call_11_8_7_9, \
call_11_8_7_10, \
call_11_8_7_11, \
call_11_8_7_12, \
call_11_8_7_13, \
call_11_8_7_14, \
call_11_8_7_15, \
call_11_8_8_0, \
call_11_8_8_1, \
call_11_8_8_2, \
call_11_8_8_3, \
call_11_8_8_4, \
call_11_8_8_5, \
call_11_8_8_6, \
call_11_8_8_7, \
call_11_8_8_8, \
call_11_8_8_9, \
call_11_8_8_10, \
call_11_8_8_11, \
call_11_8_8_12, \
call_11_8_8_13, \
call_11_8_8_14, \
call_11_8_8_15, \
call_11_8_9_0, \
call_11_8_9_1, \
call_11_8_9_2, \
call_11_8_9_3, \
call_11_8_9_4, \
call_11_8_9_5, \
call_11_8_9_6, \
call_11_8_9_7, \
call_11_8_9_8, \
call_11_8_9_9, \
call_11_8_9_10, \
call_11_8_9_11, \
call_11_8_9_12, \
call_11_8_9_13, \
call_11_8_9_14, \
call_11_8_9_15, \
call_11_8_10_0, \
call_11_8_10_1, \
call_11_8_10_2, \
call_11_8_10_3, \
call_11_8_10_4, \
call_11_8_10_5, \
call_11_8_10_6, \
call_11_8_10_7, \
call_11_8_10_8, \
call_11_8_10_9, \
call_11_8_10_10, \
call_11_8_10_11, \
call_11_8_10_12, \
call_11_8_10_13, \
call_11_8_10_14, \
call_11_8_10_15, \
call_11_8_11_0, \
call_11_8_11_1, \
call_11_8_11_2, \
call_11_8_11_3, \
call_11_8_11_4, \
call_11_8_11_5, \
call_11_8_11_6, \
call_11_8_11_7, \
call_11_8_11_8, \
call_11_8_11_9, \
call_11_8_11_10, \
call_11_8_11_11, \
call_11_8_11_12, \
call_11_8_11_13, \
call_11_8_11_14, \
call_11_8_11_15, \
call_11_8_12_0, \
call_11_8_12_1, \
call_11_8_12_2, \
call_11_8_12_3, \
call_11_8_12_4, \
call_11_8_12_5, \
call_11_8_12_6, \
call_11_8_12_7, \
call_11_8_12_8, \
call_11_8_12_9, \
call_11_8_12_10, \
call_11_8_12_11, \
call_11_8_12_12, \
call_11_8_12_13, \
call_11_8_12_14, \
call_11_8_12_15, \
call_11_8_13_0, \
call_11_8_13_1, \
call_11_8_13_2, \
call_11_8_13_3, \
call_11_8_13_4, \
call_11_8_13_5, \
call_11_8_13_6, \
call_11_8_13_7, \
call_11_8_13_8, \
call_11_8_13_9, \
call_11_8_13_10, \
call_11_8_13_11, \
call_11_8_13_12, \
call_11_8_13_13, \
call_11_8_13_14, \
call_11_8_13_15, \
call_11_8_14_0, \
call_11_8_14_1, \
call_11_8_14_2, \
call_11_8_14_3, \
call_11_8_14_4, \
call_11_8_14_5, \
call_11_8_14_6, \
call_11_8_14_7, \
call_11_8_14_8, \
call_11_8_14_9, \
call_11_8_14_10, \
call_11_8_14_11, \
call_11_8_14_12, \
call_11_8_14_13, \
call_11_8_14_14, \
call_11_8_14_15, \
call_11_8_15_0, \
call_11_8_15_1, \
call_11_8_15_2, \
call_11_8_15_3, \
call_11_8_15_4, \
call_11_8_15_5, \
call_11_8_15_6, \
call_11_8_15_7, \
call_11_8_15_8, \
call_11_8_15_9, \
call_11_8_15_10, \
call_11_8_15_11, \
call_11_8_15_12, \
call_11_8_15_13, \
call_11_8_15_14, \
call_11_8_15_15, \
call_11_9_0_0, \
call_11_9_0_1, \
call_11_9_0_2, \
call_11_9_0_3, \
call_11_9_0_4, \
call_11_9_0_5, \
call_11_9_0_6, \
call_11_9_0_7, \
call_11_9_0_8, \
call_11_9_0_9, \
call_11_9_0_10, \
call_11_9_0_11, \
call_11_9_0_12, \
call_11_9_0_13, \
call_11_9_0_14, \
call_11_9_0_15, \
call_11_9_1_0, \
call_11_9_1_1, \
call_11_9_1_2, \
call_11_9_1_3, \
call_11_9_1_4, \
call_11_9_1_5, \
call_11_9_1_6, \
call_11_9_1_7, \
call_11_9_1_8, \
call_11_9_1_9, \
call_11_9_1_10, \
call_11_9_1_11, \
call_11_9_1_12, \
call_11_9_1_13, \
call_11_9_1_14, \
call_11_9_1_15, \
call_11_9_2_0, \
call_11_9_2_1, \
call_11_9_2_2, \
call_11_9_2_3, \
call_11_9_2_4, \
call_11_9_2_5, \
call_11_9_2_6, \
call_11_9_2_7, \
call_11_9_2_8, \
call_11_9_2_9, \
call_11_9_2_10, \
call_11_9_2_11, \
call_11_9_2_12, \
call_11_9_2_13, \
call_11_9_2_14, \
call_11_9_2_15, \
call_11_9_3_0, \
call_11_9_3_1, \
call_11_9_3_2, \
call_11_9_3_3, \
call_11_9_3_4, \
call_11_9_3_5, \
call_11_9_3_6, \
call_11_9_3_7, \
call_11_9_3_8, \
call_11_9_3_9, \
call_11_9_3_10, \
call_11_9_3_11, \
call_11_9_3_12, \
call_11_9_3_13, \
call_11_9_3_14, \
call_11_9_3_15, \
call_11_9_4_0, \
call_11_9_4_1, \
call_11_9_4_2, \
call_11_9_4_3, \
call_11_9_4_4, \
call_11_9_4_5, \
call_11_9_4_6, \
call_11_9_4_7, \
call_11_9_4_8, \
call_11_9_4_9, \
call_11_9_4_10, \
call_11_9_4_11, \
call_11_9_4_12, \
call_11_9_4_13, \
call_11_9_4_14, \
call_11_9_4_15, \
call_11_9_5_0, \
call_11_9_5_1, \
call_11_9_5_2, \
call_11_9_5_3, \
call_11_9_5_4, \
call_11_9_5_5, \
call_11_9_5_6, \
call_11_9_5_7, \
call_11_9_5_8, \
call_11_9_5_9, \
call_11_9_5_10, \
call_11_9_5_11, \
call_11_9_5_12, \
call_11_9_5_13, \
call_11_9_5_14, \
call_11_9_5_15, \
call_11_9_6_0, \
call_11_9_6_1, \
call_11_9_6_2, \
call_11_9_6_3, \
call_11_9_6_4, \
call_11_9_6_5, \
call_11_9_6_6, \
call_11_9_6_7, \
call_11_9_6_8, \
call_11_9_6_9, \
call_11_9_6_10, \
call_11_9_6_11, \
call_11_9_6_12, \
call_11_9_6_13, \
call_11_9_6_14, \
call_11_9_6_15, \
call_11_9_7_0, \
call_11_9_7_1, \
call_11_9_7_2, \
call_11_9_7_3, \
call_11_9_7_4, \
call_11_9_7_5, \
call_11_9_7_6, \
call_11_9_7_7, \
call_11_9_7_8, \
call_11_9_7_9, \
call_11_9_7_10, \
call_11_9_7_11, \
call_11_9_7_12, \
call_11_9_7_13, \
call_11_9_7_14, \
call_11_9_7_15, \
call_11_9_8_0, \
call_11_9_8_1, \
call_11_9_8_2, \
call_11_9_8_3, \
call_11_9_8_4, \
call_11_9_8_5, \
call_11_9_8_6, \
call_11_9_8_7, \
call_11_9_8_8, \
call_11_9_8_9, \
call_11_9_8_10, \
call_11_9_8_11, \
call_11_9_8_12, \
call_11_9_8_13, \
call_11_9_8_14, \
call_11_9_8_15, \
call_11_9_9_0, \
call_11_9_9_1, \
call_11_9_9_2, \
call_11_9_9_3, \
call_11_9_9_4, \
call_11_9_9_5, \
call_11_9_9_6, \
call_11_9_9_7, \
call_11_9_9_8, \
call_11_9_9_9, \
call_11_9_9_10, \
call_11_9_9_11, \
call_11_9_9_12, \
call_11_9_9_13, \
call_11_9_9_14, \
call_11_9_9_15, \
call_11_9_10_0, \
call_11_9_10_1, \
call_11_9_10_2, \
call_11_9_10_3, \
call_11_9_10_4, \
call_11_9_10_5, \
call_11_9_10_6, \
call_11_9_10_7, \
call_11_9_10_8, \
call_11_9_10_9, \
call_11_9_10_10, \
call_11_9_10_11, \
call_11_9_10_12, \
call_11_9_10_13, \
call_11_9_10_14, \
call_11_9_10_15, \
call_11_9_11_0, \
call_11_9_11_1, \
call_11_9_11_2, \
call_11_9_11_3, \
call_11_9_11_4, \
call_11_9_11_5, \
call_11_9_11_6, \
call_11_9_11_7, \
call_11_9_11_8, \
call_11_9_11_9, \
call_11_9_11_10, \
call_11_9_11_11, \
call_11_9_11_12, \
call_11_9_11_13, \
call_11_9_11_14, \
call_11_9_11_15, \
call_11_9_12_0, \
call_11_9_12_1, \
call_11_9_12_2, \
call_11_9_12_3, \
call_11_9_12_4, \
call_11_9_12_5, \
call_11_9_12_6, \
call_11_9_12_7, \
call_11_9_12_8, \
call_11_9_12_9, \
call_11_9_12_10, \
call_11_9_12_11, \
call_11_9_12_12, \
call_11_9_12_13, \
call_11_9_12_14, \
call_11_9_12_15, \
call_11_9_13_0, \
call_11_9_13_1, \
call_11_9_13_2, \
call_11_9_13_3, \
call_11_9_13_4, \
call_11_9_13_5, \
call_11_9_13_6, \
call_11_9_13_7, \
call_11_9_13_8, \
call_11_9_13_9, \
call_11_9_13_10, \
call_11_9_13_11, \
call_11_9_13_12, \
call_11_9_13_13, \
call_11_9_13_14, \
call_11_9_13_15, \
call_11_9_14_0, \
call_11_9_14_1, \
call_11_9_14_2, \
call_11_9_14_3, \
call_11_9_14_4, \
call_11_9_14_5, \
call_11_9_14_6, \
call_11_9_14_7, \
call_11_9_14_8, \
call_11_9_14_9, \
call_11_9_14_10, \
call_11_9_14_11, \
call_11_9_14_12, \
call_11_9_14_13, \
call_11_9_14_14, \
call_11_9_14_15, \
call_11_9_15_0, \
call_11_9_15_1, \
call_11_9_15_2, \
call_11_9_15_3, \
call_11_9_15_4, \
call_11_9_15_5, \
call_11_9_15_6, \
call_11_9_15_7, \
call_11_9_15_8, \
call_11_9_15_9, \
call_11_9_15_10, \
call_11_9_15_11, \
call_11_9_15_12, \
call_11_9_15_13, \
call_11_9_15_14, \
call_11_9_15_15, \
call_11_10_0_0, \
call_11_10_0_1, \
call_11_10_0_2, \
call_11_10_0_3, \
call_11_10_0_4, \
call_11_10_0_5, \
call_11_10_0_6, \
call_11_10_0_7, \
call_11_10_0_8, \
call_11_10_0_9, \
call_11_10_0_10, \
call_11_10_0_11, \
call_11_10_0_12, \
call_11_10_0_13, \
call_11_10_0_14, \
call_11_10_0_15, \
call_11_10_1_0, \
call_11_10_1_1, \
call_11_10_1_2, \
call_11_10_1_3, \
call_11_10_1_4, \
call_11_10_1_5, \
call_11_10_1_6, \
call_11_10_1_7, \
call_11_10_1_8, \
call_11_10_1_9, \
call_11_10_1_10, \
call_11_10_1_11, \
call_11_10_1_12, \
call_11_10_1_13, \
call_11_10_1_14, \
call_11_10_1_15, \
call_11_10_2_0, \
call_11_10_2_1, \
call_11_10_2_2, \
call_11_10_2_3, \
call_11_10_2_4, \
call_11_10_2_5, \
call_11_10_2_6, \
call_11_10_2_7, \
call_11_10_2_8, \
call_11_10_2_9, \
call_11_10_2_10, \
call_11_10_2_11, \
call_11_10_2_12, \
call_11_10_2_13, \
call_11_10_2_14, \
call_11_10_2_15, \
call_11_10_3_0, \
call_11_10_3_1, \
call_11_10_3_2, \
call_11_10_3_3, \
call_11_10_3_4, \
call_11_10_3_5, \
call_11_10_3_6, \
call_11_10_3_7, \
call_11_10_3_8, \
call_11_10_3_9, \
call_11_10_3_10, \
call_11_10_3_11, \
call_11_10_3_12, \
call_11_10_3_13, \
call_11_10_3_14, \
call_11_10_3_15, \
call_11_10_4_0, \
call_11_10_4_1, \
call_11_10_4_2, \
call_11_10_4_3, \
call_11_10_4_4, \
call_11_10_4_5, \
call_11_10_4_6, \
call_11_10_4_7, \
call_11_10_4_8, \
call_11_10_4_9, \
call_11_10_4_10, \
call_11_10_4_11, \
call_11_10_4_12, \
call_11_10_4_13, \
call_11_10_4_14, \
call_11_10_4_15, \
call_11_10_5_0, \
call_11_10_5_1, \
call_11_10_5_2, \
call_11_10_5_3, \
call_11_10_5_4, \
call_11_10_5_5, \
call_11_10_5_6, \
call_11_10_5_7, \
call_11_10_5_8, \
call_11_10_5_9, \
call_11_10_5_10, \
call_11_10_5_11, \
call_11_10_5_12, \
call_11_10_5_13, \
call_11_10_5_14, \
call_11_10_5_15, \
call_11_10_6_0, \
call_11_10_6_1, \
call_11_10_6_2, \
call_11_10_6_3, \
call_11_10_6_4, \
call_11_10_6_5, \
call_11_10_6_6, \
call_11_10_6_7, \
call_11_10_6_8, \
call_11_10_6_9, \
call_11_10_6_10, \
call_11_10_6_11, \
call_11_10_6_12, \
call_11_10_6_13, \
call_11_10_6_14, \
call_11_10_6_15, \
call_11_10_7_0, \
call_11_10_7_1, \
call_11_10_7_2, \
call_11_10_7_3, \
call_11_10_7_4, \
call_11_10_7_5, \
call_11_10_7_6, \
call_11_10_7_7, \
call_11_10_7_8, \
call_11_10_7_9, \
call_11_10_7_10, \
call_11_10_7_11, \
call_11_10_7_12, \
call_11_10_7_13, \
call_11_10_7_14, \
call_11_10_7_15, \
call_11_10_8_0, \
call_11_10_8_1, \
call_11_10_8_2, \
call_11_10_8_3, \
call_11_10_8_4, \
call_11_10_8_5, \
call_11_10_8_6, \
call_11_10_8_7, \
call_11_10_8_8, \
call_11_10_8_9, \
call_11_10_8_10, \
call_11_10_8_11, \
call_11_10_8_12, \
call_11_10_8_13, \
call_11_10_8_14, \
call_11_10_8_15, \
call_11_10_9_0, \
call_11_10_9_1, \
call_11_10_9_2, \
call_11_10_9_3, \
call_11_10_9_4, \
call_11_10_9_5, \
call_11_10_9_6, \
call_11_10_9_7, \
call_11_10_9_8, \
call_11_10_9_9, \
call_11_10_9_10, \
call_11_10_9_11, \
call_11_10_9_12, \
call_11_10_9_13, \
call_11_10_9_14, \
call_11_10_9_15, \
call_11_10_10_0, \
call_11_10_10_1, \
call_11_10_10_2, \
call_11_10_10_3, \
call_11_10_10_4, \
call_11_10_10_5, \
call_11_10_10_6, \
call_11_10_10_7, \
call_11_10_10_8, \
call_11_10_10_9, \
call_11_10_10_10, \
call_11_10_10_11, \
call_11_10_10_12, \
call_11_10_10_13, \
call_11_10_10_14, \
call_11_10_10_15, \
call_11_10_11_0, \
call_11_10_11_1, \
call_11_10_11_2, \
call_11_10_11_3, \
call_11_10_11_4, \
call_11_10_11_5, \
call_11_10_11_6, \
call_11_10_11_7, \
call_11_10_11_8, \
call_11_10_11_9, \
call_11_10_11_10, \
call_11_10_11_11, \
call_11_10_11_12, \
call_11_10_11_13, \
call_11_10_11_14, \
call_11_10_11_15, \
call_11_10_12_0, \
call_11_10_12_1, \
call_11_10_12_2, \
call_11_10_12_3, \
call_11_10_12_4, \
call_11_10_12_5, \
call_11_10_12_6, \
call_11_10_12_7, \
call_11_10_12_8, \
call_11_10_12_9, \
call_11_10_12_10, \
call_11_10_12_11, \
call_11_10_12_12, \
call_11_10_12_13, \
call_11_10_12_14, \
call_11_10_12_15, \
call_11_10_13_0, \
call_11_10_13_1, \
call_11_10_13_2, \
call_11_10_13_3, \
call_11_10_13_4, \
call_11_10_13_5, \
call_11_10_13_6, \
call_11_10_13_7, \
call_11_10_13_8, \
call_11_10_13_9, \
call_11_10_13_10, \
call_11_10_13_11, \
call_11_10_13_12, \
call_11_10_13_13, \
call_11_10_13_14, \
call_11_10_13_15, \
call_11_10_14_0, \
call_11_10_14_1, \
call_11_10_14_2, \
call_11_10_14_3, \
call_11_10_14_4, \
call_11_10_14_5, \
call_11_10_14_6, \
call_11_10_14_7, \
call_11_10_14_8, \
call_11_10_14_9, \
call_11_10_14_10, \
call_11_10_14_11, \
call_11_10_14_12, \
call_11_10_14_13, \
call_11_10_14_14, \
call_11_10_14_15, \
call_11_10_15_0, \
call_11_10_15_1, \
call_11_10_15_2, \
call_11_10_15_3, \
call_11_10_15_4, \
call_11_10_15_5, \
call_11_10_15_6, \
call_11_10_15_7, \
call_11_10_15_8, \
call_11_10_15_9, \
call_11_10_15_10, \
call_11_10_15_11, \
call_11_10_15_12, \
call_11_10_15_13, \
call_11_10_15_14, \
call_11_10_15_15, \
call_11_11_0_0, \
call_11_11_0_1, \
call_11_11_0_2, \
call_11_11_0_3, \
call_11_11_0_4, \
call_11_11_0_5, \
call_11_11_0_6, \
call_11_11_0_7, \
call_11_11_0_8, \
call_11_11_0_9, \
call_11_11_0_10, \
call_11_11_0_11, \
call_11_11_0_12, \
call_11_11_0_13, \
call_11_11_0_14, \
call_11_11_0_15, \
call_11_11_1_0, \
call_11_11_1_1, \
call_11_11_1_2, \
call_11_11_1_3, \
call_11_11_1_4, \
call_11_11_1_5, \
call_11_11_1_6, \
call_11_11_1_7, \
call_11_11_1_8, \
call_11_11_1_9, \
call_11_11_1_10, \
call_11_11_1_11, \
call_11_11_1_12, \
call_11_11_1_13, \
call_11_11_1_14, \
call_11_11_1_15, \
call_11_11_2_0, \
call_11_11_2_1, \
call_11_11_2_2, \
call_11_11_2_3, \
call_11_11_2_4, \
call_11_11_2_5, \
call_11_11_2_6, \
call_11_11_2_7, \
call_11_11_2_8, \
call_11_11_2_9, \
call_11_11_2_10, \
call_11_11_2_11, \
call_11_11_2_12, \
call_11_11_2_13, \
call_11_11_2_14, \
call_11_11_2_15, \
call_11_11_3_0, \
call_11_11_3_1, \
call_11_11_3_2, \
call_11_11_3_3, \
call_11_11_3_4, \
call_11_11_3_5, \
call_11_11_3_6, \
call_11_11_3_7, \
call_11_11_3_8, \
call_11_11_3_9, \
call_11_11_3_10, \
call_11_11_3_11, \
call_11_11_3_12, \
call_11_11_3_13, \
call_11_11_3_14, \
call_11_11_3_15, \
call_11_11_4_0, \
call_11_11_4_1, \
call_11_11_4_2, \
call_11_11_4_3, \
call_11_11_4_4, \
call_11_11_4_5, \
call_11_11_4_6, \
call_11_11_4_7, \
call_11_11_4_8, \
call_11_11_4_9, \
call_11_11_4_10, \
call_11_11_4_11, \
call_11_11_4_12, \
call_11_11_4_13, \
call_11_11_4_14, \
call_11_11_4_15, \
call_11_11_5_0, \
call_11_11_5_1, \
call_11_11_5_2, \
call_11_11_5_3, \
call_11_11_5_4, \
call_11_11_5_5, \
call_11_11_5_6, \
call_11_11_5_7, \
call_11_11_5_8, \
call_11_11_5_9, \
call_11_11_5_10, \
call_11_11_5_11, \
call_11_11_5_12, \
call_11_11_5_13, \
call_11_11_5_14, \
call_11_11_5_15, \
call_11_11_6_0, \
call_11_11_6_1, \
call_11_11_6_2, \
call_11_11_6_3, \
call_11_11_6_4, \
call_11_11_6_5, \
call_11_11_6_6, \
call_11_11_6_7, \
call_11_11_6_8, \
call_11_11_6_9, \
call_11_11_6_10, \
call_11_11_6_11, \
call_11_11_6_12, \
call_11_11_6_13, \
call_11_11_6_14, \
call_11_11_6_15, \
call_11_11_7_0, \
call_11_11_7_1, \
call_11_11_7_2, \
call_11_11_7_3, \
call_11_11_7_4, \
call_11_11_7_5, \
call_11_11_7_6, \
call_11_11_7_7, \
call_11_11_7_8, \
call_11_11_7_9, \
call_11_11_7_10, \
call_11_11_7_11, \
call_11_11_7_12, \
call_11_11_7_13, \
call_11_11_7_14, \
call_11_11_7_15, \
call_11_11_8_0, \
call_11_11_8_1, \
call_11_11_8_2, \
call_11_11_8_3, \
call_11_11_8_4, \
call_11_11_8_5, \
call_11_11_8_6, \
call_11_11_8_7, \
call_11_11_8_8, \
call_11_11_8_9, \
call_11_11_8_10, \
call_11_11_8_11, \
call_11_11_8_12, \
call_11_11_8_13, \
call_11_11_8_14, \
call_11_11_8_15, \
call_11_11_9_0, \
call_11_11_9_1, \
call_11_11_9_2, \
call_11_11_9_3, \
call_11_11_9_4, \
call_11_11_9_5, \
call_11_11_9_6, \
call_11_11_9_7, \
call_11_11_9_8, \
call_11_11_9_9, \
call_11_11_9_10, \
call_11_11_9_11, \
call_11_11_9_12, \
call_11_11_9_13, \
call_11_11_9_14, \
call_11_11_9_15, \
call_11_11_10_0, \
call_11_11_10_1, \
call_11_11_10_2, \
call_11_11_10_3, \
call_11_11_10_4, \
call_11_11_10_5, \
call_11_11_10_6, \
call_11_11_10_7, \
call_11_11_10_8, \
call_11_11_10_9, \
call_11_11_10_10, \
call_11_11_10_11, \
call_11_11_10_12, \
call_11_11_10_13, \
call_11_11_10_14, \
call_11_11_10_15, \
call_11_11_11_0, \
call_11_11_11_1, \
call_11_11_11_2, \
call_11_11_11_3, \
call_11_11_11_4, \
call_11_11_11_5, \
call_11_11_11_6, \
call_11_11_11_7, \
call_11_11_11_8, \
call_11_11_11_9, \
call_11_11_11_10, \
call_11_11_11_11, \
call_11_11_11_12, \
call_11_11_11_13, \
call_11_11_11_14, \
call_11_11_11_15, \
call_11_11_12_0, \
call_11_11_12_1, \
call_11_11_12_2, \
call_11_11_12_3, \
call_11_11_12_4, \
call_11_11_12_5, \
call_11_11_12_6, \
call_11_11_12_7, \
call_11_11_12_8, \
call_11_11_12_9, \
call_11_11_12_10, \
call_11_11_12_11, \
call_11_11_12_12, \
call_11_11_12_13, \
call_11_11_12_14, \
call_11_11_12_15, \
call_11_11_13_0, \
call_11_11_13_1, \
call_11_11_13_2, \
call_11_11_13_3, \
call_11_11_13_4, \
call_11_11_13_5, \
call_11_11_13_6, \
call_11_11_13_7, \
call_11_11_13_8, \
call_11_11_13_9, \
call_11_11_13_10, \
call_11_11_13_11, \
call_11_11_13_12, \
call_11_11_13_13, \
call_11_11_13_14, \
call_11_11_13_15, \
call_11_11_14_0, \
call_11_11_14_1, \
call_11_11_14_2, \
call_11_11_14_3, \
call_11_11_14_4, \
call_11_11_14_5, \
call_11_11_14_6, \
call_11_11_14_7, \
call_11_11_14_8, \
call_11_11_14_9, \
call_11_11_14_10, \
call_11_11_14_11, \
call_11_11_14_12, \
call_11_11_14_13, \
call_11_11_14_14, \
call_11_11_14_15, \
call_11_11_15_0, \
call_11_11_15_1, \
call_11_11_15_2, \
call_11_11_15_3, \
call_11_11_15_4, \
call_11_11_15_5, \
call_11_11_15_6, \
call_11_11_15_7, \
call_11_11_15_8, \
call_11_11_15_9, \
call_11_11_15_10, \
call_11_11_15_11, \
call_11_11_15_12, \
call_11_11_15_13, \
call_11_11_15_14, \
call_11_11_15_15, \
call_11_12_0_0, \
call_11_12_0_1, \
call_11_12_0_2, \
call_11_12_0_3, \
call_11_12_0_4, \
call_11_12_0_5, \
call_11_12_0_6, \
call_11_12_0_7, \
call_11_12_0_8, \
call_11_12_0_9, \
call_11_12_0_10, \
call_11_12_0_11, \
call_11_12_0_12, \
call_11_12_0_13, \
call_11_12_0_14, \
call_11_12_0_15, \
call_11_12_1_0, \
call_11_12_1_1, \
call_11_12_1_2, \
call_11_12_1_3, \
call_11_12_1_4, \
call_11_12_1_5, \
call_11_12_1_6, \
call_11_12_1_7, \
call_11_12_1_8, \
call_11_12_1_9, \
call_11_12_1_10, \
call_11_12_1_11, \
call_11_12_1_12, \
call_11_12_1_13, \
call_11_12_1_14, \
call_11_12_1_15, \
call_11_12_2_0, \
call_11_12_2_1, \
call_11_12_2_2, \
call_11_12_2_3, \
call_11_12_2_4, \
call_11_12_2_5, \
call_11_12_2_6, \
call_11_12_2_7, \
call_11_12_2_8, \
call_11_12_2_9, \
call_11_12_2_10, \
call_11_12_2_11, \
call_11_12_2_12, \
call_11_12_2_13, \
call_11_12_2_14, \
call_11_12_2_15, \
call_11_12_3_0, \
call_11_12_3_1, \
call_11_12_3_2, \
call_11_12_3_3, \
call_11_12_3_4, \
call_11_12_3_5, \
call_11_12_3_6, \
call_11_12_3_7, \
call_11_12_3_8, \
call_11_12_3_9, \
call_11_12_3_10, \
call_11_12_3_11, \
call_11_12_3_12, \
call_11_12_3_13, \
call_11_12_3_14, \
call_11_12_3_15, \
call_11_12_4_0, \
call_11_12_4_1, \
call_11_12_4_2, \
call_11_12_4_3, \
call_11_12_4_4, \
call_11_12_4_5, \
call_11_12_4_6, \
call_11_12_4_7, \
call_11_12_4_8, \
call_11_12_4_9, \
call_11_12_4_10, \
call_11_12_4_11, \
call_11_12_4_12, \
call_11_12_4_13, \
call_11_12_4_14, \
call_11_12_4_15, \
call_11_12_5_0, \
call_11_12_5_1, \
call_11_12_5_2, \
call_11_12_5_3, \
call_11_12_5_4, \
call_11_12_5_5, \
call_11_12_5_6, \
call_11_12_5_7, \
call_11_12_5_8, \
call_11_12_5_9, \
call_11_12_5_10, \
call_11_12_5_11, \
call_11_12_5_12, \
call_11_12_5_13, \
call_11_12_5_14, \
call_11_12_5_15, \
call_11_12_6_0, \
call_11_12_6_1, \
call_11_12_6_2, \
call_11_12_6_3, \
call_11_12_6_4, \
call_11_12_6_5, \
call_11_12_6_6, \
call_11_12_6_7, \
call_11_12_6_8, \
call_11_12_6_9, \
call_11_12_6_10, \
call_11_12_6_11, \
call_11_12_6_12, \
call_11_12_6_13, \
call_11_12_6_14, \
call_11_12_6_15, \
call_11_12_7_0, \
call_11_12_7_1, \
call_11_12_7_2, \
call_11_12_7_3, \
call_11_12_7_4, \
call_11_12_7_5, \
call_11_12_7_6, \
call_11_12_7_7, \
call_11_12_7_8, \
call_11_12_7_9, \
call_11_12_7_10, \
call_11_12_7_11, \
call_11_12_7_12, \
call_11_12_7_13, \
call_11_12_7_14, \
call_11_12_7_15, \
call_11_12_8_0, \
call_11_12_8_1, \
call_11_12_8_2, \
call_11_12_8_3, \
call_11_12_8_4, \
call_11_12_8_5, \
call_11_12_8_6, \
call_11_12_8_7, \
call_11_12_8_8, \
call_11_12_8_9, \
call_11_12_8_10, \
call_11_12_8_11, \
call_11_12_8_12, \
call_11_12_8_13, \
call_11_12_8_14, \
call_11_12_8_15, \
call_11_12_9_0, \
call_11_12_9_1, \
call_11_12_9_2, \
call_11_12_9_3, \
call_11_12_9_4, \
call_11_12_9_5, \
call_11_12_9_6, \
call_11_12_9_7, \
call_11_12_9_8, \
call_11_12_9_9, \
call_11_12_9_10, \
call_11_12_9_11, \
call_11_12_9_12, \
call_11_12_9_13, \
call_11_12_9_14, \
call_11_12_9_15, \
call_11_12_10_0, \
call_11_12_10_1, \
call_11_12_10_2, \
call_11_12_10_3, \
call_11_12_10_4, \
call_11_12_10_5, \
call_11_12_10_6, \
call_11_12_10_7, \
call_11_12_10_8, \
call_11_12_10_9, \
call_11_12_10_10, \
call_11_12_10_11, \
call_11_12_10_12, \
call_11_12_10_13, \
call_11_12_10_14, \
call_11_12_10_15, \
call_11_12_11_0, \
call_11_12_11_1, \
call_11_12_11_2, \
call_11_12_11_3, \
call_11_12_11_4, \
call_11_12_11_5, \
call_11_12_11_6, \
call_11_12_11_7, \
call_11_12_11_8, \
call_11_12_11_9, \
call_11_12_11_10, \
call_11_12_11_11, \
call_11_12_11_12, \
call_11_12_11_13, \
call_11_12_11_14, \
call_11_12_11_15, \
call_11_12_12_0, \
call_11_12_12_1, \
call_11_12_12_2, \
call_11_12_12_3, \
call_11_12_12_4, \
call_11_12_12_5, \
call_11_12_12_6, \
call_11_12_12_7, \
call_11_12_12_8, \
call_11_12_12_9, \
call_11_12_12_10, \
call_11_12_12_11, \
call_11_12_12_12, \
call_11_12_12_13, \
call_11_12_12_14, \
call_11_12_12_15, \
call_11_12_13_0, \
call_11_12_13_1, \
call_11_12_13_2, \
call_11_12_13_3, \
call_11_12_13_4, \
call_11_12_13_5, \
call_11_12_13_6, \
call_11_12_13_7, \
call_11_12_13_8, \
call_11_12_13_9, \
call_11_12_13_10, \
call_11_12_13_11, \
call_11_12_13_12, \
call_11_12_13_13, \
call_11_12_13_14, \
call_11_12_13_15, \
call_11_12_14_0, \
call_11_12_14_1, \
call_11_12_14_2, \
call_11_12_14_3, \
call_11_12_14_4, \
call_11_12_14_5, \
call_11_12_14_6, \
call_11_12_14_7, \
call_11_12_14_8, \
call_11_12_14_9, \
call_11_12_14_10, \
call_11_12_14_11, \
call_11_12_14_12, \
call_11_12_14_13, \
call_11_12_14_14, \
call_11_12_14_15, \
call_11_12_15_0, \
call_11_12_15_1, \
call_11_12_15_2, \
call_11_12_15_3, \
call_11_12_15_4, \
call_11_12_15_5, \
call_11_12_15_6, \
call_11_12_15_7, \
call_11_12_15_8, \
call_11_12_15_9, \
call_11_12_15_10, \
call_11_12_15_11, \
call_11_12_15_12, \
call_11_12_15_13, \
call_11_12_15_14, \
call_11_12_15_15, \
call_11_13_0_0, \
call_11_13_0_1, \
call_11_13_0_2, \
call_11_13_0_3, \
call_11_13_0_4, \
call_11_13_0_5, \
call_11_13_0_6, \
call_11_13_0_7, \
call_11_13_0_8, \
call_11_13_0_9, \
call_11_13_0_10, \
call_11_13_0_11, \
call_11_13_0_12, \
call_11_13_0_13, \
call_11_13_0_14, \
call_11_13_0_15, \
call_11_13_1_0, \
call_11_13_1_1, \
call_11_13_1_2, \
call_11_13_1_3, \
call_11_13_1_4, \
call_11_13_1_5, \
call_11_13_1_6, \
call_11_13_1_7, \
call_11_13_1_8, \
call_11_13_1_9, \
call_11_13_1_10, \
call_11_13_1_11, \
call_11_13_1_12, \
call_11_13_1_13, \
call_11_13_1_14, \
call_11_13_1_15, \
call_11_13_2_0, \
call_11_13_2_1, \
call_11_13_2_2, \
call_11_13_2_3, \
call_11_13_2_4, \
call_11_13_2_5, \
call_11_13_2_6, \
call_11_13_2_7, \
call_11_13_2_8, \
call_11_13_2_9, \
call_11_13_2_10, \
call_11_13_2_11, \
call_11_13_2_12, \
call_11_13_2_13, \
call_11_13_2_14, \
call_11_13_2_15, \
call_11_13_3_0, \
call_11_13_3_1, \
call_11_13_3_2, \
call_11_13_3_3, \
call_11_13_3_4, \
call_11_13_3_5, \
call_11_13_3_6, \
call_11_13_3_7, \
call_11_13_3_8, \
call_11_13_3_9, \
call_11_13_3_10, \
call_11_13_3_11, \
call_11_13_3_12, \
call_11_13_3_13, \
call_11_13_3_14, \
call_11_13_3_15, \
call_11_13_4_0, \
call_11_13_4_1, \
call_11_13_4_2, \
call_11_13_4_3, \
call_11_13_4_4, \
call_11_13_4_5, \
call_11_13_4_6, \
call_11_13_4_7, \
call_11_13_4_8, \
call_11_13_4_9, \
call_11_13_4_10, \
call_11_13_4_11, \
call_11_13_4_12, \
call_11_13_4_13, \
call_11_13_4_14, \
call_11_13_4_15, \
call_11_13_5_0, \
call_11_13_5_1, \
call_11_13_5_2, \
call_11_13_5_3, \
call_11_13_5_4, \
call_11_13_5_5, \
call_11_13_5_6, \
call_11_13_5_7, \
call_11_13_5_8, \
call_11_13_5_9, \
call_11_13_5_10, \
call_11_13_5_11, \
call_11_13_5_12, \
call_11_13_5_13, \
call_11_13_5_14, \
call_11_13_5_15, \
call_11_13_6_0, \
call_11_13_6_1, \
call_11_13_6_2, \
call_11_13_6_3, \
call_11_13_6_4, \
call_11_13_6_5, \
call_11_13_6_6, \
call_11_13_6_7, \
call_11_13_6_8, \
call_11_13_6_9, \
call_11_13_6_10, \
call_11_13_6_11, \
call_11_13_6_12, \
call_11_13_6_13, \
call_11_13_6_14, \
call_11_13_6_15, \
call_11_13_7_0, \
call_11_13_7_1, \
call_11_13_7_2, \
call_11_13_7_3, \
call_11_13_7_4, \
call_11_13_7_5, \
call_11_13_7_6, \
call_11_13_7_7, \
call_11_13_7_8, \
call_11_13_7_9, \
call_11_13_7_10, \
call_11_13_7_11, \
call_11_13_7_12, \
call_11_13_7_13, \
call_11_13_7_14, \
call_11_13_7_15, \
call_11_13_8_0, \
call_11_13_8_1, \
call_11_13_8_2, \
call_11_13_8_3, \
call_11_13_8_4, \
call_11_13_8_5, \
call_11_13_8_6, \
call_11_13_8_7, \
call_11_13_8_8, \
call_11_13_8_9, \
call_11_13_8_10, \
call_11_13_8_11, \
call_11_13_8_12, \
call_11_13_8_13, \
call_11_13_8_14, \
call_11_13_8_15, \
call_11_13_9_0, \
call_11_13_9_1, \
call_11_13_9_2, \
call_11_13_9_3, \
call_11_13_9_4, \
call_11_13_9_5, \
call_11_13_9_6, \
call_11_13_9_7, \
call_11_13_9_8, \
call_11_13_9_9, \
call_11_13_9_10, \
call_11_13_9_11, \
call_11_13_9_12, \
call_11_13_9_13, \
call_11_13_9_14, \
call_11_13_9_15, \
call_11_13_10_0, \
call_11_13_10_1, \
call_11_13_10_2, \
call_11_13_10_3, \
call_11_13_10_4, \
call_11_13_10_5, \
call_11_13_10_6, \
call_11_13_10_7, \
call_11_13_10_8, \
call_11_13_10_9, \
call_11_13_10_10, \
call_11_13_10_11, \
call_11_13_10_12, \
call_11_13_10_13, \
call_11_13_10_14, \
call_11_13_10_15, \
call_11_13_11_0, \
call_11_13_11_1, \
call_11_13_11_2, \
call_11_13_11_3, \
call_11_13_11_4, \
call_11_13_11_5, \
call_11_13_11_6, \
call_11_13_11_7, \
call_11_13_11_8, \
call_11_13_11_9, \
call_11_13_11_10, \
call_11_13_11_11, \
call_11_13_11_12, \
call_11_13_11_13, \
call_11_13_11_14, \
call_11_13_11_15, \
call_11_13_12_0, \
call_11_13_12_1, \
call_11_13_12_2, \
call_11_13_12_3, \
call_11_13_12_4, \
call_11_13_12_5, \
call_11_13_12_6, \
call_11_13_12_7, \
call_11_13_12_8, \
call_11_13_12_9, \
call_11_13_12_10, \
call_11_13_12_11, \
call_11_13_12_12, \
call_11_13_12_13, \
call_11_13_12_14, \
call_11_13_12_15, \
call_11_13_13_0, \
call_11_13_13_1, \
call_11_13_13_2, \
call_11_13_13_3, \
call_11_13_13_4, \
call_11_13_13_5, \
call_11_13_13_6, \
call_11_13_13_7, \
call_11_13_13_8, \
call_11_13_13_9, \
call_11_13_13_10, \
call_11_13_13_11, \
call_11_13_13_12, \
call_11_13_13_13, \
call_11_13_13_14, \
call_11_13_13_15, \
call_11_13_14_0, \
call_11_13_14_1, \
call_11_13_14_2, \
call_11_13_14_3, \
call_11_13_14_4, \
call_11_13_14_5, \
call_11_13_14_6, \
call_11_13_14_7, \
call_11_13_14_8, \
call_11_13_14_9, \
call_11_13_14_10, \
call_11_13_14_11, \
call_11_13_14_12, \
call_11_13_14_13, \
call_11_13_14_14, \
call_11_13_14_15, \
call_11_13_15_0, \
call_11_13_15_1, \
call_11_13_15_2, \
call_11_13_15_3, \
call_11_13_15_4, \
call_11_13_15_5, \
call_11_13_15_6, \
call_11_13_15_7, \
call_11_13_15_8, \
call_11_13_15_9, \
call_11_13_15_10, \
call_11_13_15_11, \
call_11_13_15_12, \
call_11_13_15_13, \
call_11_13_15_14, \
call_11_13_15_15, \
call_11_14_0_0, \
call_11_14_0_1, \
call_11_14_0_2, \
call_11_14_0_3, \
call_11_14_0_4, \
call_11_14_0_5, \
call_11_14_0_6, \
call_11_14_0_7, \
call_11_14_0_8, \
call_11_14_0_9, \
call_11_14_0_10, \
call_11_14_0_11, \
call_11_14_0_12, \
call_11_14_0_13, \
call_11_14_0_14, \
call_11_14_0_15, \
call_11_14_1_0, \
call_11_14_1_1, \
call_11_14_1_2, \
call_11_14_1_3, \
call_11_14_1_4, \
call_11_14_1_5, \
call_11_14_1_6, \
call_11_14_1_7, \
call_11_14_1_8, \
call_11_14_1_9, \
call_11_14_1_10, \
call_11_14_1_11, \
call_11_14_1_12, \
call_11_14_1_13, \
call_11_14_1_14, \
call_11_14_1_15, \
call_11_14_2_0, \
call_11_14_2_1, \
call_11_14_2_2, \
call_11_14_2_3, \
call_11_14_2_4, \
call_11_14_2_5, \
call_11_14_2_6, \
call_11_14_2_7, \
call_11_14_2_8, \
call_11_14_2_9, \
call_11_14_2_10, \
call_11_14_2_11, \
call_11_14_2_12, \
call_11_14_2_13, \
call_11_14_2_14, \
call_11_14_2_15, \
call_11_14_3_0, \
call_11_14_3_1, \
call_11_14_3_2, \
call_11_14_3_3, \
call_11_14_3_4, \
call_11_14_3_5, \
call_11_14_3_6, \
call_11_14_3_7, \
call_11_14_3_8, \
call_11_14_3_9, \
call_11_14_3_10, \
call_11_14_3_11, \
call_11_14_3_12, \
call_11_14_3_13, \
call_11_14_3_14, \
call_11_14_3_15, \
call_11_14_4_0, \
call_11_14_4_1, \
call_11_14_4_2, \
call_11_14_4_3, \
call_11_14_4_4, \
call_11_14_4_5, \
call_11_14_4_6, \
call_11_14_4_7, \
call_11_14_4_8, \
call_11_14_4_9, \
call_11_14_4_10, \
call_11_14_4_11, \
call_11_14_4_12, \
call_11_14_4_13, \
call_11_14_4_14, \
call_11_14_4_15, \
call_11_14_5_0, \
call_11_14_5_1, \
call_11_14_5_2, \
call_11_14_5_3, \
call_11_14_5_4, \
call_11_14_5_5, \
call_11_14_5_6, \
call_11_14_5_7, \
call_11_14_5_8, \
call_11_14_5_9, \
call_11_14_5_10, \
call_11_14_5_11, \
call_11_14_5_12, \
call_11_14_5_13, \
call_11_14_5_14, \
call_11_14_5_15, \
call_11_14_6_0, \
call_11_14_6_1, \
call_11_14_6_2, \
call_11_14_6_3, \
call_11_14_6_4, \
call_11_14_6_5, \
call_11_14_6_6, \
call_11_14_6_7, \
call_11_14_6_8, \
call_11_14_6_9, \
call_11_14_6_10, \
call_11_14_6_11, \
call_11_14_6_12, \
call_11_14_6_13, \
call_11_14_6_14, \
call_11_14_6_15, \
call_11_14_7_0, \
call_11_14_7_1, \
call_11_14_7_2, \
call_11_14_7_3, \
call_11_14_7_4, \
call_11_14_7_5, \
call_11_14_7_6, \
call_11_14_7_7, \
call_11_14_7_8, \
call_11_14_7_9, \
call_11_14_7_10, \
call_11_14_7_11, \
call_11_14_7_12, \
call_11_14_7_13, \
call_11_14_7_14, \
call_11_14_7_15, \
call_11_14_8_0, \
call_11_14_8_1, \
call_11_14_8_2, \
call_11_14_8_3, \
call_11_14_8_4, \
call_11_14_8_5, \
call_11_14_8_6, \
call_11_14_8_7, \
call_11_14_8_8, \
call_11_14_8_9, \
call_11_14_8_10, \
call_11_14_8_11, \
call_11_14_8_12, \
call_11_14_8_13, \
call_11_14_8_14, \
call_11_14_8_15, \
call_11_14_9_0, \
call_11_14_9_1, \
call_11_14_9_2, \
call_11_14_9_3, \
call_11_14_9_4, \
call_11_14_9_5, \
call_11_14_9_6, \
call_11_14_9_7, \
call_11_14_9_8, \
call_11_14_9_9, \
call_11_14_9_10, \
call_11_14_9_11, \
call_11_14_9_12, \
call_11_14_9_13, \
call_11_14_9_14, \
call_11_14_9_15, \
call_11_14_10_0, \
call_11_14_10_1, \
call_11_14_10_2, \
call_11_14_10_3, \
call_11_14_10_4, \
call_11_14_10_5, \
call_11_14_10_6, \
call_11_14_10_7, \
call_11_14_10_8, \
call_11_14_10_9, \
call_11_14_10_10, \
call_11_14_10_11, \
call_11_14_10_12, \
call_11_14_10_13, \
call_11_14_10_14, \
call_11_14_10_15, \
call_11_14_11_0, \
call_11_14_11_1, \
call_11_14_11_2, \
call_11_14_11_3, \
call_11_14_11_4, \
call_11_14_11_5, \
call_11_14_11_6, \
call_11_14_11_7, \
call_11_14_11_8, \
call_11_14_11_9, \
call_11_14_11_10, \
call_11_14_11_11, \
call_11_14_11_12, \
call_11_14_11_13, \
call_11_14_11_14, \
call_11_14_11_15, \
call_11_14_12_0, \
call_11_14_12_1, \
call_11_14_12_2, \
call_11_14_12_3, \
call_11_14_12_4, \
call_11_14_12_5, \
call_11_14_12_6, \
call_11_14_12_7, \
call_11_14_12_8, \
call_11_14_12_9, \
call_11_14_12_10, \
call_11_14_12_11, \
call_11_14_12_12, \
call_11_14_12_13, \
call_11_14_12_14, \
call_11_14_12_15, \
call_11_14_13_0, \
call_11_14_13_1, \
call_11_14_13_2, \
call_11_14_13_3, \
call_11_14_13_4, \
call_11_14_13_5, \
call_11_14_13_6, \
call_11_14_13_7, \
call_11_14_13_8, \
call_11_14_13_9, \
call_11_14_13_10, \
call_11_14_13_11, \
call_11_14_13_12, \
call_11_14_13_13, \
call_11_14_13_14, \
call_11_14_13_15, \
call_11_14_14_0, \
call_11_14_14_1, \
call_11_14_14_2, \
call_11_14_14_3, \
call_11_14_14_4, \
call_11_14_14_5, \
call_11_14_14_6, \
call_11_14_14_7, \
call_11_14_14_8, \
call_11_14_14_9, \
call_11_14_14_10, \
call_11_14_14_11, \
call_11_14_14_12, \
call_11_14_14_13, \
call_11_14_14_14, \
call_11_14_14_15, \
call_11_14_15_0, \
call_11_14_15_1, \
call_11_14_15_2, \
call_11_14_15_3, \
call_11_14_15_4, \
call_11_14_15_5, \
call_11_14_15_6, \
call_11_14_15_7, \
call_11_14_15_8, \
call_11_14_15_9, \
call_11_14_15_10, \
call_11_14_15_11, \
call_11_14_15_12, \
call_11_14_15_13, \
call_11_14_15_14, \
call_11_14_15_15, \
call_11_15_0_0, \
call_11_15_0_1, \
call_11_15_0_2, \
call_11_15_0_3, \
call_11_15_0_4, \
call_11_15_0_5, \
call_11_15_0_6, \
call_11_15_0_7, \
call_11_15_0_8, \
call_11_15_0_9, \
call_11_15_0_10, \
call_11_15_0_11, \
call_11_15_0_12, \
call_11_15_0_13, \
call_11_15_0_14, \
call_11_15_0_15, \
call_11_15_1_0, \
call_11_15_1_1, \
call_11_15_1_2, \
call_11_15_1_3, \
call_11_15_1_4, \
call_11_15_1_5, \
call_11_15_1_6, \
call_11_15_1_7, \
call_11_15_1_8, \
call_11_15_1_9, \
call_11_15_1_10, \
call_11_15_1_11, \
call_11_15_1_12, \
call_11_15_1_13, \
call_11_15_1_14, \
call_11_15_1_15, \
call_11_15_2_0, \
call_11_15_2_1, \
call_11_15_2_2, \
call_11_15_2_3, \
call_11_15_2_4, \
call_11_15_2_5, \
call_11_15_2_6, \
call_11_15_2_7, \
call_11_15_2_8, \
call_11_15_2_9, \
call_11_15_2_10, \
call_11_15_2_11, \
call_11_15_2_12, \
call_11_15_2_13, \
call_11_15_2_14, \
call_11_15_2_15, \
call_11_15_3_0, \
call_11_15_3_1, \
call_11_15_3_2, \
call_11_15_3_3, \
call_11_15_3_4, \
call_11_15_3_5, \
call_11_15_3_6, \
call_11_15_3_7, \
call_11_15_3_8, \
call_11_15_3_9, \
call_11_15_3_10, \
call_11_15_3_11, \
call_11_15_3_12, \
call_11_15_3_13, \
call_11_15_3_14, \
call_11_15_3_15, \
call_11_15_4_0, \
call_11_15_4_1, \
call_11_15_4_2, \
call_11_15_4_3, \
call_11_15_4_4, \
call_11_15_4_5, \
call_11_15_4_6, \
call_11_15_4_7, \
call_11_15_4_8, \
call_11_15_4_9, \
call_11_15_4_10, \
call_11_15_4_11, \
call_11_15_4_12, \
call_11_15_4_13, \
call_11_15_4_14, \
call_11_15_4_15, \
call_11_15_5_0, \
call_11_15_5_1, \
call_11_15_5_2, \
call_11_15_5_3, \
call_11_15_5_4, \
call_11_15_5_5, \
call_11_15_5_6, \
call_11_15_5_7, \
call_11_15_5_8, \
call_11_15_5_9, \
call_11_15_5_10, \
call_11_15_5_11, \
call_11_15_5_12, \
call_11_15_5_13, \
call_11_15_5_14, \
call_11_15_5_15, \
call_11_15_6_0, \
call_11_15_6_1, \
call_11_15_6_2, \
call_11_15_6_3, \
call_11_15_6_4, \
call_11_15_6_5, \
call_11_15_6_6, \
call_11_15_6_7, \
call_11_15_6_8, \
call_11_15_6_9, \
call_11_15_6_10, \
call_11_15_6_11, \
call_11_15_6_12, \
call_11_15_6_13, \
call_11_15_6_14, \
call_11_15_6_15, \
call_11_15_7_0, \
call_11_15_7_1, \
call_11_15_7_2, \
call_11_15_7_3, \
call_11_15_7_4, \
call_11_15_7_5, \
call_11_15_7_6, \
call_11_15_7_7, \
call_11_15_7_8, \
call_11_15_7_9, \
call_11_15_7_10, \
call_11_15_7_11, \
call_11_15_7_12, \
call_11_15_7_13, \
call_11_15_7_14, \
call_11_15_7_15, \
call_11_15_8_0, \
call_11_15_8_1, \
call_11_15_8_2, \
call_11_15_8_3, \
call_11_15_8_4, \
call_11_15_8_5, \
call_11_15_8_6, \
call_11_15_8_7, \
call_11_15_8_8, \
call_11_15_8_9, \
call_11_15_8_10, \
call_11_15_8_11, \
call_11_15_8_12, \
call_11_15_8_13, \
call_11_15_8_14, \
call_11_15_8_15, \
call_11_15_9_0, \
call_11_15_9_1, \
call_11_15_9_2, \
call_11_15_9_3, \
call_11_15_9_4, \
call_11_15_9_5, \
call_11_15_9_6, \
call_11_15_9_7, \
call_11_15_9_8, \
call_11_15_9_9, \
call_11_15_9_10, \
call_11_15_9_11, \
call_11_15_9_12, \
call_11_15_9_13, \
call_11_15_9_14, \
call_11_15_9_15, \
call_11_15_10_0, \
call_11_15_10_1, \
call_11_15_10_2, \
call_11_15_10_3, \
call_11_15_10_4, \
call_11_15_10_5, \
call_11_15_10_6, \
call_11_15_10_7, \
call_11_15_10_8, \
call_11_15_10_9, \
call_11_15_10_10, \
call_11_15_10_11, \
call_11_15_10_12, \
call_11_15_10_13, \
call_11_15_10_14, \
call_11_15_10_15, \
call_11_15_11_0, \
call_11_15_11_1, \
call_11_15_11_2, \
call_11_15_11_3, \
call_11_15_11_4, \
call_11_15_11_5, \
call_11_15_11_6, \
call_11_15_11_7, \
call_11_15_11_8, \
call_11_15_11_9, \
call_11_15_11_10, \
call_11_15_11_11, \
call_11_15_11_12, \
call_11_15_11_13, \
call_11_15_11_14, \
call_11_15_11_15, \
call_11_15_12_0, \
call_11_15_12_1, \
call_11_15_12_2, \
call_11_15_12_3, \
call_11_15_12_4, \
call_11_15_12_5, \
call_11_15_12_6, \
call_11_15_12_7, \
call_11_15_12_8, \
call_11_15_12_9, \
call_11_15_12_10, \
call_11_15_12_11, \
call_11_15_12_12, \
call_11_15_12_13, \
call_11_15_12_14, \
call_11_15_12_15, \
call_11_15_13_0, \
call_11_15_13_1, \
call_11_15_13_2, \
call_11_15_13_3, \
call_11_15_13_4, \
call_11_15_13_5, \
call_11_15_13_6, \
call_11_15_13_7, \
call_11_15_13_8, \
call_11_15_13_9, \
call_11_15_13_10, \
call_11_15_13_11, \
call_11_15_13_12, \
call_11_15_13_13, \
call_11_15_13_14, \
call_11_15_13_15, \
call_11_15_14_0, \
call_11_15_14_1, \
call_11_15_14_2, \
call_11_15_14_3, \
call_11_15_14_4, \
call_11_15_14_5, \
call_11_15_14_6, \
call_11_15_14_7, \
call_11_15_14_8, \
call_11_15_14_9, \
call_11_15_14_10, \
call_11_15_14_11, \
call_11_15_14_12, \
call_11_15_14_13, \
call_11_15_14_14, \
call_11_15_14_15, \
call_11_15_15_0, \
call_11_15_15_1, \
call_11_15_15_2, \
call_11_15_15_3, \
call_11_15_15_4, \
call_11_15_15_5, \
call_11_15_15_6, \
call_11_15_15_7, \
call_11_15_15_8, \
call_11_15_15_9, \
call_11_15_15_10, \
call_11_15_15_11, \
call_11_15_15_12, \
call_11_15_15_13, \
call_11_15_15_14, \
call_11_15_15_15, \
call_12_0_0_0, \
call_12_0_0_1, \
call_12_0_0_2, \
call_12_0_0_3, \
call_12_0_0_4, \
call_12_0_0_5, \
call_12_0_0_6, \
call_12_0_0_7, \
call_12_0_0_8, \
call_12_0_0_9, \
call_12_0_0_10, \
call_12_0_0_11, \
call_12_0_0_12, \
call_12_0_0_13, \
call_12_0_0_14, \
call_12_0_0_15, \
call_12_0_1_0, \
call_12_0_1_1, \
call_12_0_1_2, \
call_12_0_1_3, \
call_12_0_1_4, \
call_12_0_1_5, \
call_12_0_1_6, \
call_12_0_1_7, \
call_12_0_1_8, \
call_12_0_1_9, \
call_12_0_1_10, \
call_12_0_1_11, \
call_12_0_1_12, \
call_12_0_1_13, \
call_12_0_1_14, \
call_12_0_1_15, \
call_12_0_2_0, \
call_12_0_2_1, \
call_12_0_2_2, \
call_12_0_2_3, \
call_12_0_2_4, \
call_12_0_2_5, \
call_12_0_2_6, \
call_12_0_2_7, \
call_12_0_2_8, \
call_12_0_2_9, \
call_12_0_2_10, \
call_12_0_2_11, \
call_12_0_2_12, \
call_12_0_2_13, \
call_12_0_2_14, \
call_12_0_2_15, \
call_12_0_3_0, \
call_12_0_3_1, \
call_12_0_3_2, \
call_12_0_3_3, \
call_12_0_3_4, \
call_12_0_3_5, \
call_12_0_3_6, \
call_12_0_3_7, \
call_12_0_3_8, \
call_12_0_3_9, \
call_12_0_3_10, \
call_12_0_3_11, \
call_12_0_3_12, \
call_12_0_3_13, \
call_12_0_3_14, \
call_12_0_3_15, \
call_12_0_4_0, \
call_12_0_4_1, \
call_12_0_4_2, \
call_12_0_4_3, \
call_12_0_4_4, \
call_12_0_4_5, \
call_12_0_4_6, \
call_12_0_4_7, \
call_12_0_4_8, \
call_12_0_4_9, \
call_12_0_4_10, \
call_12_0_4_11, \
call_12_0_4_12, \
call_12_0_4_13, \
call_12_0_4_14, \
call_12_0_4_15, \
call_12_0_5_0, \
call_12_0_5_1, \
call_12_0_5_2, \
call_12_0_5_3, \
call_12_0_5_4, \
call_12_0_5_5, \
call_12_0_5_6, \
call_12_0_5_7, \
call_12_0_5_8, \
call_12_0_5_9, \
call_12_0_5_10, \
call_12_0_5_11, \
call_12_0_5_12, \
call_12_0_5_13, \
call_12_0_5_14, \
call_12_0_5_15, \
call_12_0_6_0, \
call_12_0_6_1, \
call_12_0_6_2, \
call_12_0_6_3, \
call_12_0_6_4, \
call_12_0_6_5, \
call_12_0_6_6, \
call_12_0_6_7, \
call_12_0_6_8, \
call_12_0_6_9, \
call_12_0_6_10, \
call_12_0_6_11, \
call_12_0_6_12, \
call_12_0_6_13, \
call_12_0_6_14, \
call_12_0_6_15, \
call_12_0_7_0, \
call_12_0_7_1, \
call_12_0_7_2, \
call_12_0_7_3, \
call_12_0_7_4, \
call_12_0_7_5, \
call_12_0_7_6, \
call_12_0_7_7, \
call_12_0_7_8, \
call_12_0_7_9, \
call_12_0_7_10, \
call_12_0_7_11, \
call_12_0_7_12, \
call_12_0_7_13, \
call_12_0_7_14, \
call_12_0_7_15, \
call_12_0_8_0, \
call_12_0_8_1, \
call_12_0_8_2, \
call_12_0_8_3, \
call_12_0_8_4, \
call_12_0_8_5, \
call_12_0_8_6, \
call_12_0_8_7, \
call_12_0_8_8, \
call_12_0_8_9, \
call_12_0_8_10, \
call_12_0_8_11, \
call_12_0_8_12, \
call_12_0_8_13, \
call_12_0_8_14, \
call_12_0_8_15, \
call_12_0_9_0, \
call_12_0_9_1, \
call_12_0_9_2, \
call_12_0_9_3, \
call_12_0_9_4, \
call_12_0_9_5, \
call_12_0_9_6, \
call_12_0_9_7, \
call_12_0_9_8, \
call_12_0_9_9, \
call_12_0_9_10, \
call_12_0_9_11, \
call_12_0_9_12, \
call_12_0_9_13, \
call_12_0_9_14, \
call_12_0_9_15, \
call_12_0_10_0, \
call_12_0_10_1, \
call_12_0_10_2, \
call_12_0_10_3, \
call_12_0_10_4, \
call_12_0_10_5, \
call_12_0_10_6, \
call_12_0_10_7, \
call_12_0_10_8, \
call_12_0_10_9, \
call_12_0_10_10, \
call_12_0_10_11, \
call_12_0_10_12, \
call_12_0_10_13, \
call_12_0_10_14, \
call_12_0_10_15, \
call_12_0_11_0, \
call_12_0_11_1, \
call_12_0_11_2, \
call_12_0_11_3, \
call_12_0_11_4, \
call_12_0_11_5, \
call_12_0_11_6, \
call_12_0_11_7, \
call_12_0_11_8, \
call_12_0_11_9, \
call_12_0_11_10, \
call_12_0_11_11, \
call_12_0_11_12, \
call_12_0_11_13, \
call_12_0_11_14, \
call_12_0_11_15, \
call_12_0_12_0, \
call_12_0_12_1, \
call_12_0_12_2, \
call_12_0_12_3, \
call_12_0_12_4, \
call_12_0_12_5, \
call_12_0_12_6, \
call_12_0_12_7, \
call_12_0_12_8, \
call_12_0_12_9, \
call_12_0_12_10, \
call_12_0_12_11, \
call_12_0_12_12, \
call_12_0_12_13, \
call_12_0_12_14, \
call_12_0_12_15, \
call_12_0_13_0, \
call_12_0_13_1, \
call_12_0_13_2, \
call_12_0_13_3, \
call_12_0_13_4, \
call_12_0_13_5, \
call_12_0_13_6, \
call_12_0_13_7, \
call_12_0_13_8, \
call_12_0_13_9, \
call_12_0_13_10, \
call_12_0_13_11, \
call_12_0_13_12, \
call_12_0_13_13, \
call_12_0_13_14, \
call_12_0_13_15, \
call_12_0_14_0, \
call_12_0_14_1, \
call_12_0_14_2, \
call_12_0_14_3, \
call_12_0_14_4, \
call_12_0_14_5, \
call_12_0_14_6, \
call_12_0_14_7, \
call_12_0_14_8, \
call_12_0_14_9, \
call_12_0_14_10, \
call_12_0_14_11, \
call_12_0_14_12, \
call_12_0_14_13, \
call_12_0_14_14, \
call_12_0_14_15, \
call_12_0_15_0, \
call_12_0_15_1, \
call_12_0_15_2, \
call_12_0_15_3, \
call_12_0_15_4, \
call_12_0_15_5, \
call_12_0_15_6, \
call_12_0_15_7, \
call_12_0_15_8, \
call_12_0_15_9, \
call_12_0_15_10, \
call_12_0_15_11, \
call_12_0_15_12, \
call_12_0_15_13, \
call_12_0_15_14, \
call_12_0_15_15, \
call_12_1_0_0, \
call_12_1_0_1, \
call_12_1_0_2, \
call_12_1_0_3, \
call_12_1_0_4, \
call_12_1_0_5, \
call_12_1_0_6, \
call_12_1_0_7, \
call_12_1_0_8, \
call_12_1_0_9, \
call_12_1_0_10, \
call_12_1_0_11, \
call_12_1_0_12, \
call_12_1_0_13, \
call_12_1_0_14, \
call_12_1_0_15, \
call_12_1_1_0, \
call_12_1_1_1, \
call_12_1_1_2, \
call_12_1_1_3, \
call_12_1_1_4, \
call_12_1_1_5, \
call_12_1_1_6, \
call_12_1_1_7, \
call_12_1_1_8, \
call_12_1_1_9, \
call_12_1_1_10, \
call_12_1_1_11, \
call_12_1_1_12, \
call_12_1_1_13, \
call_12_1_1_14, \
call_12_1_1_15, \
call_12_1_2_0, \
call_12_1_2_1, \
call_12_1_2_2, \
call_12_1_2_3, \
call_12_1_2_4, \
call_12_1_2_5, \
call_12_1_2_6, \
call_12_1_2_7, \
call_12_1_2_8, \
call_12_1_2_9, \
call_12_1_2_10, \
call_12_1_2_11, \
call_12_1_2_12, \
call_12_1_2_13, \
call_12_1_2_14, \
call_12_1_2_15, \
call_12_1_3_0, \
call_12_1_3_1, \
call_12_1_3_2, \
call_12_1_3_3, \
call_12_1_3_4, \
call_12_1_3_5, \
call_12_1_3_6, \
call_12_1_3_7, \
call_12_1_3_8, \
call_12_1_3_9, \
call_12_1_3_10, \
call_12_1_3_11, \
call_12_1_3_12, \
call_12_1_3_13, \
call_12_1_3_14, \
call_12_1_3_15, \
call_12_1_4_0, \
call_12_1_4_1, \
call_12_1_4_2, \
call_12_1_4_3, \
call_12_1_4_4, \
call_12_1_4_5, \
call_12_1_4_6, \
call_12_1_4_7, \
call_12_1_4_8, \
call_12_1_4_9, \
call_12_1_4_10, \
call_12_1_4_11, \
call_12_1_4_12, \
call_12_1_4_13, \
call_12_1_4_14, \
call_12_1_4_15, \
call_12_1_5_0, \
call_12_1_5_1, \
call_12_1_5_2, \
call_12_1_5_3, \
call_12_1_5_4, \
call_12_1_5_5, \
call_12_1_5_6, \
call_12_1_5_7, \
call_12_1_5_8, \
call_12_1_5_9, \
call_12_1_5_10, \
call_12_1_5_11, \
call_12_1_5_12, \
call_12_1_5_13, \
call_12_1_5_14, \
call_12_1_5_15, \
call_12_1_6_0, \
call_12_1_6_1, \
call_12_1_6_2, \
call_12_1_6_3, \
call_12_1_6_4, \
call_12_1_6_5, \
call_12_1_6_6, \
call_12_1_6_7, \
call_12_1_6_8, \
call_12_1_6_9, \
call_12_1_6_10, \
call_12_1_6_11, \
call_12_1_6_12, \
call_12_1_6_13, \
call_12_1_6_14, \
call_12_1_6_15, \
call_12_1_7_0, \
call_12_1_7_1, \
call_12_1_7_2, \
call_12_1_7_3, \
call_12_1_7_4, \
call_12_1_7_5, \
call_12_1_7_6, \
call_12_1_7_7, \
call_12_1_7_8, \
call_12_1_7_9, \
call_12_1_7_10, \
call_12_1_7_11, \
call_12_1_7_12, \
call_12_1_7_13, \
call_12_1_7_14, \
call_12_1_7_15, \
call_12_1_8_0, \
call_12_1_8_1, \
call_12_1_8_2, \
call_12_1_8_3, \
call_12_1_8_4, \
call_12_1_8_5, \
call_12_1_8_6, \
call_12_1_8_7, \
call_12_1_8_8, \
call_12_1_8_9, \
call_12_1_8_10, \
call_12_1_8_11, \
call_12_1_8_12, \
call_12_1_8_13, \
call_12_1_8_14, \
call_12_1_8_15, \
call_12_1_9_0, \
call_12_1_9_1, \
call_12_1_9_2, \
call_12_1_9_3, \
call_12_1_9_4, \
call_12_1_9_5, \
call_12_1_9_6, \
call_12_1_9_7, \
call_12_1_9_8, \
call_12_1_9_9, \
call_12_1_9_10, \
call_12_1_9_11, \
call_12_1_9_12, \
call_12_1_9_13, \
call_12_1_9_14, \
call_12_1_9_15, \
call_12_1_10_0, \
call_12_1_10_1, \
call_12_1_10_2, \
call_12_1_10_3, \
call_12_1_10_4, \
call_12_1_10_5, \
call_12_1_10_6, \
call_12_1_10_7, \
call_12_1_10_8, \
call_12_1_10_9, \
call_12_1_10_10, \
call_12_1_10_11, \
call_12_1_10_12, \
call_12_1_10_13, \
call_12_1_10_14, \
call_12_1_10_15, \
call_12_1_11_0, \
call_12_1_11_1, \
call_12_1_11_2, \
call_12_1_11_3, \
call_12_1_11_4, \
call_12_1_11_5, \
call_12_1_11_6, \
call_12_1_11_7, \
call_12_1_11_8, \
call_12_1_11_9, \
call_12_1_11_10, \
call_12_1_11_11, \
call_12_1_11_12, \
call_12_1_11_13, \
call_12_1_11_14, \
call_12_1_11_15, \
call_12_1_12_0, \
call_12_1_12_1, \
call_12_1_12_2, \
call_12_1_12_3, \
call_12_1_12_4, \
call_12_1_12_5, \
call_12_1_12_6, \
call_12_1_12_7, \
call_12_1_12_8, \
call_12_1_12_9, \
call_12_1_12_10, \
call_12_1_12_11, \
call_12_1_12_12, \
call_12_1_12_13, \
call_12_1_12_14, \
call_12_1_12_15, \
call_12_1_13_0, \
call_12_1_13_1, \
call_12_1_13_2, \
call_12_1_13_3, \
call_12_1_13_4, \
call_12_1_13_5, \
call_12_1_13_6, \
call_12_1_13_7, \
call_12_1_13_8, \
call_12_1_13_9, \
call_12_1_13_10, \
call_12_1_13_11, \
call_12_1_13_12, \
call_12_1_13_13, \
call_12_1_13_14, \
call_12_1_13_15, \
call_12_1_14_0, \
call_12_1_14_1, \
call_12_1_14_2, \
call_12_1_14_3, \
call_12_1_14_4, \
call_12_1_14_5, \
call_12_1_14_6, \
call_12_1_14_7, \
call_12_1_14_8, \
call_12_1_14_9, \
call_12_1_14_10, \
call_12_1_14_11, \
call_12_1_14_12, \
call_12_1_14_13, \
call_12_1_14_14, \
call_12_1_14_15, \
call_12_1_15_0, \
call_12_1_15_1, \
call_12_1_15_2, \
call_12_1_15_3, \
call_12_1_15_4, \
call_12_1_15_5, \
call_12_1_15_6, \
call_12_1_15_7, \
call_12_1_15_8, \
call_12_1_15_9, \
call_12_1_15_10, \
call_12_1_15_11, \
call_12_1_15_12, \
call_12_1_15_13, \
call_12_1_15_14, \
call_12_1_15_15, \
call_12_2_0_0, \
call_12_2_0_1, \
call_12_2_0_2, \
call_12_2_0_3, \
call_12_2_0_4, \
call_12_2_0_5, \
call_12_2_0_6, \
call_12_2_0_7, \
call_12_2_0_8, \
call_12_2_0_9, \
call_12_2_0_10, \
call_12_2_0_11, \
call_12_2_0_12, \
call_12_2_0_13, \
call_12_2_0_14, \
call_12_2_0_15, \
call_12_2_1_0, \
call_12_2_1_1, \
call_12_2_1_2, \
call_12_2_1_3, \
call_12_2_1_4, \
call_12_2_1_5, \
call_12_2_1_6, \
call_12_2_1_7, \
call_12_2_1_8, \
call_12_2_1_9, \
call_12_2_1_10, \
call_12_2_1_11, \
call_12_2_1_12, \
call_12_2_1_13, \
call_12_2_1_14, \
call_12_2_1_15, \
call_12_2_2_0, \
call_12_2_2_1, \
call_12_2_2_2, \
call_12_2_2_3, \
call_12_2_2_4, \
call_12_2_2_5, \
call_12_2_2_6, \
call_12_2_2_7, \
call_12_2_2_8, \
call_12_2_2_9, \
call_12_2_2_10, \
call_12_2_2_11, \
call_12_2_2_12, \
call_12_2_2_13, \
call_12_2_2_14, \
call_12_2_2_15, \
call_12_2_3_0, \
call_12_2_3_1, \
call_12_2_3_2, \
call_12_2_3_3, \
call_12_2_3_4, \
call_12_2_3_5, \
call_12_2_3_6, \
call_12_2_3_7, \
call_12_2_3_8, \
call_12_2_3_9, \
call_12_2_3_10, \
call_12_2_3_11, \
call_12_2_3_12, \
call_12_2_3_13, \
call_12_2_3_14, \
call_12_2_3_15, \
call_12_2_4_0, \
call_12_2_4_1, \
call_12_2_4_2, \
call_12_2_4_3, \
call_12_2_4_4, \
call_12_2_4_5, \
call_12_2_4_6, \
call_12_2_4_7, \
call_12_2_4_8, \
call_12_2_4_9, \
call_12_2_4_10, \
call_12_2_4_11, \
call_12_2_4_12, \
call_12_2_4_13, \
call_12_2_4_14, \
call_12_2_4_15, \
call_12_2_5_0, \
call_12_2_5_1, \
call_12_2_5_2, \
call_12_2_5_3, \
call_12_2_5_4, \
call_12_2_5_5, \
call_12_2_5_6, \
call_12_2_5_7, \
call_12_2_5_8, \
call_12_2_5_9, \
call_12_2_5_10, \
call_12_2_5_11, \
call_12_2_5_12, \
call_12_2_5_13, \
call_12_2_5_14, \
call_12_2_5_15, \
call_12_2_6_0, \
call_12_2_6_1, \
call_12_2_6_2, \
call_12_2_6_3, \
call_12_2_6_4, \
call_12_2_6_5, \
call_12_2_6_6, \
call_12_2_6_7, \
call_12_2_6_8, \
call_12_2_6_9, \
call_12_2_6_10, \
call_12_2_6_11, \
call_12_2_6_12, \
call_12_2_6_13, \
call_12_2_6_14, \
call_12_2_6_15, \
call_12_2_7_0, \
call_12_2_7_1, \
call_12_2_7_2, \
call_12_2_7_3, \
call_12_2_7_4, \
call_12_2_7_5, \
call_12_2_7_6, \
call_12_2_7_7, \
call_12_2_7_8, \
call_12_2_7_9, \
call_12_2_7_10, \
call_12_2_7_11, \
call_12_2_7_12, \
call_12_2_7_13, \
call_12_2_7_14, \
call_12_2_7_15, \
call_12_2_8_0, \
call_12_2_8_1, \
call_12_2_8_2, \
call_12_2_8_3, \
call_12_2_8_4, \
call_12_2_8_5, \
call_12_2_8_6, \
call_12_2_8_7, \
call_12_2_8_8, \
call_12_2_8_9, \
call_12_2_8_10, \
call_12_2_8_11, \
call_12_2_8_12, \
call_12_2_8_13, \
call_12_2_8_14, \
call_12_2_8_15, \
call_12_2_9_0, \
call_12_2_9_1, \
call_12_2_9_2, \
call_12_2_9_3, \
call_12_2_9_4, \
call_12_2_9_5, \
call_12_2_9_6, \
call_12_2_9_7, \
call_12_2_9_8, \
call_12_2_9_9, \
call_12_2_9_10, \
call_12_2_9_11, \
call_12_2_9_12, \
call_12_2_9_13, \
call_12_2_9_14, \
call_12_2_9_15, \
call_12_2_10_0, \
call_12_2_10_1, \
call_12_2_10_2, \
call_12_2_10_3, \
call_12_2_10_4, \
call_12_2_10_5, \
call_12_2_10_6, \
call_12_2_10_7, \
call_12_2_10_8, \
call_12_2_10_9, \
call_12_2_10_10, \
call_12_2_10_11, \
call_12_2_10_12, \
call_12_2_10_13, \
call_12_2_10_14, \
call_12_2_10_15, \
call_12_2_11_0, \
call_12_2_11_1, \
call_12_2_11_2, \
call_12_2_11_3, \
call_12_2_11_4, \
call_12_2_11_5, \
call_12_2_11_6, \
call_12_2_11_7, \
call_12_2_11_8, \
call_12_2_11_9, \
call_12_2_11_10, \
call_12_2_11_11, \
call_12_2_11_12, \
call_12_2_11_13, \
call_12_2_11_14, \
call_12_2_11_15, \
call_12_2_12_0, \
call_12_2_12_1, \
call_12_2_12_2, \
call_12_2_12_3, \
call_12_2_12_4, \
call_12_2_12_5, \
call_12_2_12_6, \
call_12_2_12_7, \
call_12_2_12_8, \
call_12_2_12_9, \
call_12_2_12_10, \
call_12_2_12_11, \
call_12_2_12_12, \
call_12_2_12_13, \
call_12_2_12_14, \
call_12_2_12_15, \
call_12_2_13_0, \
call_12_2_13_1, \
call_12_2_13_2, \
call_12_2_13_3, \
call_12_2_13_4, \
call_12_2_13_5, \
call_12_2_13_6, \
call_12_2_13_7, \
call_12_2_13_8, \
call_12_2_13_9, \
call_12_2_13_10, \
call_12_2_13_11, \
call_12_2_13_12, \
call_12_2_13_13, \
call_12_2_13_14, \
call_12_2_13_15, \
call_12_2_14_0, \
call_12_2_14_1, \
call_12_2_14_2, \
call_12_2_14_3, \
call_12_2_14_4, \
call_12_2_14_5, \
call_12_2_14_6, \
call_12_2_14_7, \
call_12_2_14_8, \
call_12_2_14_9, \
call_12_2_14_10, \
call_12_2_14_11, \
call_12_2_14_12, \
call_12_2_14_13, \
call_12_2_14_14, \
call_12_2_14_15, \
call_12_2_15_0, \
call_12_2_15_1, \
call_12_2_15_2, \
call_12_2_15_3, \
call_12_2_15_4, \
call_12_2_15_5, \
call_12_2_15_6, \
call_12_2_15_7, \
call_12_2_15_8, \
call_12_2_15_9, \
call_12_2_15_10, \
call_12_2_15_11, \
call_12_2_15_12, \
call_12_2_15_13, \
call_12_2_15_14, \
call_12_2_15_15, \
call_12_3_0_0, \
call_12_3_0_1, \
call_12_3_0_2, \
call_12_3_0_3, \
call_12_3_0_4, \
call_12_3_0_5, \
call_12_3_0_6, \
call_12_3_0_7, \
call_12_3_0_8, \
call_12_3_0_9, \
call_12_3_0_10, \
call_12_3_0_11, \
call_12_3_0_12, \
call_12_3_0_13, \
call_12_3_0_14, \
call_12_3_0_15, \
call_12_3_1_0, \
call_12_3_1_1, \
call_12_3_1_2, \
call_12_3_1_3, \
call_12_3_1_4, \
call_12_3_1_5, \
call_12_3_1_6, \
call_12_3_1_7, \
call_12_3_1_8, \
call_12_3_1_9, \
call_12_3_1_10, \
call_12_3_1_11, \
call_12_3_1_12, \
call_12_3_1_13, \
call_12_3_1_14, \
call_12_3_1_15, \
call_12_3_2_0, \
call_12_3_2_1, \
call_12_3_2_2, \
call_12_3_2_3, \
call_12_3_2_4, \
call_12_3_2_5, \
call_12_3_2_6, \
call_12_3_2_7, \
call_12_3_2_8, \
call_12_3_2_9, \
call_12_3_2_10, \
call_12_3_2_11, \
call_12_3_2_12, \
call_12_3_2_13, \
call_12_3_2_14, \
call_12_3_2_15, \
call_12_3_3_0, \
call_12_3_3_1, \
call_12_3_3_2, \
call_12_3_3_3, \
call_12_3_3_4, \
call_12_3_3_5, \
call_12_3_3_6, \
call_12_3_3_7, \
call_12_3_3_8, \
call_12_3_3_9, \
call_12_3_3_10, \
call_12_3_3_11, \
call_12_3_3_12, \
call_12_3_3_13, \
call_12_3_3_14, \
call_12_3_3_15, \
call_12_3_4_0, \
call_12_3_4_1, \
call_12_3_4_2, \
call_12_3_4_3, \
call_12_3_4_4, \
call_12_3_4_5, \
call_12_3_4_6, \
call_12_3_4_7, \
call_12_3_4_8, \
call_12_3_4_9, \
call_12_3_4_10, \
call_12_3_4_11, \
call_12_3_4_12, \
call_12_3_4_13, \
call_12_3_4_14, \
call_12_3_4_15, \
call_12_3_5_0, \
call_12_3_5_1, \
call_12_3_5_2, \
call_12_3_5_3, \
call_12_3_5_4, \
call_12_3_5_5, \
call_12_3_5_6, \
call_12_3_5_7, \
call_12_3_5_8, \
call_12_3_5_9, \
call_12_3_5_10, \
call_12_3_5_11, \
call_12_3_5_12, \
call_12_3_5_13, \
call_12_3_5_14, \
call_12_3_5_15, \
call_12_3_6_0, \
call_12_3_6_1, \
call_12_3_6_2, \
call_12_3_6_3, \
call_12_3_6_4, \
call_12_3_6_5, \
call_12_3_6_6, \
call_12_3_6_7, \
call_12_3_6_8, \
call_12_3_6_9, \
call_12_3_6_10, \
call_12_3_6_11, \
call_12_3_6_12, \
call_12_3_6_13, \
call_12_3_6_14, \
call_12_3_6_15, \
call_12_3_7_0, \
call_12_3_7_1, \
call_12_3_7_2, \
call_12_3_7_3, \
call_12_3_7_4, \
call_12_3_7_5, \
call_12_3_7_6, \
call_12_3_7_7, \
call_12_3_7_8, \
call_12_3_7_9, \
call_12_3_7_10, \
call_12_3_7_11, \
call_12_3_7_12, \
call_12_3_7_13, \
call_12_3_7_14, \
call_12_3_7_15, \
call_12_3_8_0, \
call_12_3_8_1, \
call_12_3_8_2, \
call_12_3_8_3, \
call_12_3_8_4, \
call_12_3_8_5, \
call_12_3_8_6, \
call_12_3_8_7, \
call_12_3_8_8, \
call_12_3_8_9, \
call_12_3_8_10, \
call_12_3_8_11, \
call_12_3_8_12, \
call_12_3_8_13, \
call_12_3_8_14, \
call_12_3_8_15, \
call_12_3_9_0, \
call_12_3_9_1, \
call_12_3_9_2, \
call_12_3_9_3, \
call_12_3_9_4, \
call_12_3_9_5, \
call_12_3_9_6, \
call_12_3_9_7, \
call_12_3_9_8, \
call_12_3_9_9, \
call_12_3_9_10, \
call_12_3_9_11, \
call_12_3_9_12, \
call_12_3_9_13, \
call_12_3_9_14, \
call_12_3_9_15, \
call_12_3_10_0, \
call_12_3_10_1, \
call_12_3_10_2, \
call_12_3_10_3, \
call_12_3_10_4, \
call_12_3_10_5, \
call_12_3_10_6, \
call_12_3_10_7, \
call_12_3_10_8, \
call_12_3_10_9, \
call_12_3_10_10, \
call_12_3_10_11, \
call_12_3_10_12, \
call_12_3_10_13, \
call_12_3_10_14, \
call_12_3_10_15, \
call_12_3_11_0, \
call_12_3_11_1, \
call_12_3_11_2, \
call_12_3_11_3, \
call_12_3_11_4, \
call_12_3_11_5, \
call_12_3_11_6, \
call_12_3_11_7, \
call_12_3_11_8, \
call_12_3_11_9, \
call_12_3_11_10, \
call_12_3_11_11, \
call_12_3_11_12, \
call_12_3_11_13, \
call_12_3_11_14, \
call_12_3_11_15, \
call_12_3_12_0, \
call_12_3_12_1, \
call_12_3_12_2, \
call_12_3_12_3, \
call_12_3_12_4, \
call_12_3_12_5, \
call_12_3_12_6, \
call_12_3_12_7, \
call_12_3_12_8, \
call_12_3_12_9, \
call_12_3_12_10, \
call_12_3_12_11, \
call_12_3_12_12, \
call_12_3_12_13, \
call_12_3_12_14, \
call_12_3_12_15, \
call_12_3_13_0, \
call_12_3_13_1, \
call_12_3_13_2, \
call_12_3_13_3, \
call_12_3_13_4, \
call_12_3_13_5, \
call_12_3_13_6, \
call_12_3_13_7, \
call_12_3_13_8, \
call_12_3_13_9, \
call_12_3_13_10, \
call_12_3_13_11, \
call_12_3_13_12, \
call_12_3_13_13, \
call_12_3_13_14, \
call_12_3_13_15, \
call_12_3_14_0, \
call_12_3_14_1, \
call_12_3_14_2, \
call_12_3_14_3, \
call_12_3_14_4, \
call_12_3_14_5, \
call_12_3_14_6, \
call_12_3_14_7, \
call_12_3_14_8, \
call_12_3_14_9, \
call_12_3_14_10, \
call_12_3_14_11, \
call_12_3_14_12, \
call_12_3_14_13, \
call_12_3_14_14, \
call_12_3_14_15, \
call_12_3_15_0, \
call_12_3_15_1, \
call_12_3_15_2, \
call_12_3_15_3, \
call_12_3_15_4, \
call_12_3_15_5, \
call_12_3_15_6, \
call_12_3_15_7, \
call_12_3_15_8, \
call_12_3_15_9, \
call_12_3_15_10, \
call_12_3_15_11, \
call_12_3_15_12, \
call_12_3_15_13, \
call_12_3_15_14, \
call_12_3_15_15, \
call_12_4_0_0, \
call_12_4_0_1, \
call_12_4_0_2, \
call_12_4_0_3, \
call_12_4_0_4, \
call_12_4_0_5, \
call_12_4_0_6, \
call_12_4_0_7, \
call_12_4_0_8, \
call_12_4_0_9, \
call_12_4_0_10, \
call_12_4_0_11, \
call_12_4_0_12, \
call_12_4_0_13, \
call_12_4_0_14, \
call_12_4_0_15, \
call_12_4_1_0, \
call_12_4_1_1, \
call_12_4_1_2, \
call_12_4_1_3, \
call_12_4_1_4, \
call_12_4_1_5, \
call_12_4_1_6, \
call_12_4_1_7, \
call_12_4_1_8, \
call_12_4_1_9, \
call_12_4_1_10, \
call_12_4_1_11, \
call_12_4_1_12, \
call_12_4_1_13, \
call_12_4_1_14, \
call_12_4_1_15, \
call_12_4_2_0, \
call_12_4_2_1, \
call_12_4_2_2, \
call_12_4_2_3, \
call_12_4_2_4, \
call_12_4_2_5, \
call_12_4_2_6, \
call_12_4_2_7, \
call_12_4_2_8, \
call_12_4_2_9, \
call_12_4_2_10, \
call_12_4_2_11, \
call_12_4_2_12, \
call_12_4_2_13, \
call_12_4_2_14, \
call_12_4_2_15, \
call_12_4_3_0, \
call_12_4_3_1, \
call_12_4_3_2, \
call_12_4_3_3, \
call_12_4_3_4, \
call_12_4_3_5, \
call_12_4_3_6, \
call_12_4_3_7, \
call_12_4_3_8, \
call_12_4_3_9, \
call_12_4_3_10, \
call_12_4_3_11, \
call_12_4_3_12, \
call_12_4_3_13, \
call_12_4_3_14, \
call_12_4_3_15, \
call_12_4_4_0, \
call_12_4_4_1, \
call_12_4_4_2, \
call_12_4_4_3, \
call_12_4_4_4, \
call_12_4_4_5, \
call_12_4_4_6, \
call_12_4_4_7, \
call_12_4_4_8, \
call_12_4_4_9, \
call_12_4_4_10, \
call_12_4_4_11, \
call_12_4_4_12, \
call_12_4_4_13, \
call_12_4_4_14, \
call_12_4_4_15, \
call_12_4_5_0, \
call_12_4_5_1, \
call_12_4_5_2, \
call_12_4_5_3, \
call_12_4_5_4, \
call_12_4_5_5, \
call_12_4_5_6, \
call_12_4_5_7, \
call_12_4_5_8, \
call_12_4_5_9, \
call_12_4_5_10, \
call_12_4_5_11, \
call_12_4_5_12, \
call_12_4_5_13, \
call_12_4_5_14, \
call_12_4_5_15, \
call_12_4_6_0, \
call_12_4_6_1, \
call_12_4_6_2, \
call_12_4_6_3, \
call_12_4_6_4, \
call_12_4_6_5, \
call_12_4_6_6, \
call_12_4_6_7, \
call_12_4_6_8, \
call_12_4_6_9, \
call_12_4_6_10, \
call_12_4_6_11, \
call_12_4_6_12, \
call_12_4_6_13, \
call_12_4_6_14, \
call_12_4_6_15, \
call_12_4_7_0, \
call_12_4_7_1, \
call_12_4_7_2, \
call_12_4_7_3, \
call_12_4_7_4, \
call_12_4_7_5, \
call_12_4_7_6, \
call_12_4_7_7, \
call_12_4_7_8, \
call_12_4_7_9, \
call_12_4_7_10, \
call_12_4_7_11, \
call_12_4_7_12, \
call_12_4_7_13, \
call_12_4_7_14, \
call_12_4_7_15, \
call_12_4_8_0, \
call_12_4_8_1, \
call_12_4_8_2, \
call_12_4_8_3, \
call_12_4_8_4, \
call_12_4_8_5, \
call_12_4_8_6, \
call_12_4_8_7, \
call_12_4_8_8, \
call_12_4_8_9, \
call_12_4_8_10, \
call_12_4_8_11, \
call_12_4_8_12, \
call_12_4_8_13, \
call_12_4_8_14, \
call_12_4_8_15, \
call_12_4_9_0, \
call_12_4_9_1, \
call_12_4_9_2, \
call_12_4_9_3, \
call_12_4_9_4, \
call_12_4_9_5, \
call_12_4_9_6, \
call_12_4_9_7, \
call_12_4_9_8, \
call_12_4_9_9, \
call_12_4_9_10, \
call_12_4_9_11, \
call_12_4_9_12, \
call_12_4_9_13, \
call_12_4_9_14, \
call_12_4_9_15, \
call_12_4_10_0, \
call_12_4_10_1, \
call_12_4_10_2, \
call_12_4_10_3, \
call_12_4_10_4, \
call_12_4_10_5, \
call_12_4_10_6, \
call_12_4_10_7, \
call_12_4_10_8, \
call_12_4_10_9, \
call_12_4_10_10, \
call_12_4_10_11, \
call_12_4_10_12, \
call_12_4_10_13, \
call_12_4_10_14, \
call_12_4_10_15, \
call_12_4_11_0, \
call_12_4_11_1, \
call_12_4_11_2, \
call_12_4_11_3, \
call_12_4_11_4, \
call_12_4_11_5, \
call_12_4_11_6, \
call_12_4_11_7, \
call_12_4_11_8, \
call_12_4_11_9, \
call_12_4_11_10, \
call_12_4_11_11, \
call_12_4_11_12, \
call_12_4_11_13, \
call_12_4_11_14, \
call_12_4_11_15, \
call_12_4_12_0, \
call_12_4_12_1, \
call_12_4_12_2, \
call_12_4_12_3, \
call_12_4_12_4, \
call_12_4_12_5, \
call_12_4_12_6, \
call_12_4_12_7, \
call_12_4_12_8, \
call_12_4_12_9, \
call_12_4_12_10, \
call_12_4_12_11, \
call_12_4_12_12, \
call_12_4_12_13, \
call_12_4_12_14, \
call_12_4_12_15, \
call_12_4_13_0, \
call_12_4_13_1, \
call_12_4_13_2, \
call_12_4_13_3, \
call_12_4_13_4, \
call_12_4_13_5, \
call_12_4_13_6, \
call_12_4_13_7, \
call_12_4_13_8, \
call_12_4_13_9, \
call_12_4_13_10, \
call_12_4_13_11, \
call_12_4_13_12, \
call_12_4_13_13, \
call_12_4_13_14, \
call_12_4_13_15, \
call_12_4_14_0, \
call_12_4_14_1, \
call_12_4_14_2, \
call_12_4_14_3, \
call_12_4_14_4, \
call_12_4_14_5, \
call_12_4_14_6, \
call_12_4_14_7, \
call_12_4_14_8, \
call_12_4_14_9, \
call_12_4_14_10, \
call_12_4_14_11, \
call_12_4_14_12, \
call_12_4_14_13, \
call_12_4_14_14, \
call_12_4_14_15, \
call_12_4_15_0, \
call_12_4_15_1, \
call_12_4_15_2, \
call_12_4_15_3, \
call_12_4_15_4, \
call_12_4_15_5, \
call_12_4_15_6, \
call_12_4_15_7, \
call_12_4_15_8, \
call_12_4_15_9, \
call_12_4_15_10, \
call_12_4_15_11, \
call_12_4_15_12, \
call_12_4_15_13, \
call_12_4_15_14, \
call_12_4_15_15, \
call_12_5_0_0, \
call_12_5_0_1, \
call_12_5_0_2, \
call_12_5_0_3, \
call_12_5_0_4, \
call_12_5_0_5, \
call_12_5_0_6, \
call_12_5_0_7, \
call_12_5_0_8, \
call_12_5_0_9, \
call_12_5_0_10, \
call_12_5_0_11, \
call_12_5_0_12, \
call_12_5_0_13, \
call_12_5_0_14, \
call_12_5_0_15, \
call_12_5_1_0, \
call_12_5_1_1, \
call_12_5_1_2, \
call_12_5_1_3, \
call_12_5_1_4, \
call_12_5_1_5, \
call_12_5_1_6, \
call_12_5_1_7, \
call_12_5_1_8, \
call_12_5_1_9, \
call_12_5_1_10, \
call_12_5_1_11, \
call_12_5_1_12, \
call_12_5_1_13, \
call_12_5_1_14, \
call_12_5_1_15, \
call_12_5_2_0, \
call_12_5_2_1, \
call_12_5_2_2, \
call_12_5_2_3, \
call_12_5_2_4, \
call_12_5_2_5, \
call_12_5_2_6, \
call_12_5_2_7, \
call_12_5_2_8, \
call_12_5_2_9, \
call_12_5_2_10, \
call_12_5_2_11, \
call_12_5_2_12, \
call_12_5_2_13, \
call_12_5_2_14, \
call_12_5_2_15, \
call_12_5_3_0, \
call_12_5_3_1, \
call_12_5_3_2, \
call_12_5_3_3, \
call_12_5_3_4, \
call_12_5_3_5, \
call_12_5_3_6, \
call_12_5_3_7, \
call_12_5_3_8, \
call_12_5_3_9, \
call_12_5_3_10, \
call_12_5_3_11, \
call_12_5_3_12, \
call_12_5_3_13, \
call_12_5_3_14, \
call_12_5_3_15, \
call_12_5_4_0, \
call_12_5_4_1, \
call_12_5_4_2, \
call_12_5_4_3, \
call_12_5_4_4, \
call_12_5_4_5, \
call_12_5_4_6, \
call_12_5_4_7, \
call_12_5_4_8, \
call_12_5_4_9, \
call_12_5_4_10, \
call_12_5_4_11, \
call_12_5_4_12, \
call_12_5_4_13, \
call_12_5_4_14, \
call_12_5_4_15, \
call_12_5_5_0, \
call_12_5_5_1, \
call_12_5_5_2, \
call_12_5_5_3, \
call_12_5_5_4, \
call_12_5_5_5, \
call_12_5_5_6, \
call_12_5_5_7, \
call_12_5_5_8, \
call_12_5_5_9, \
call_12_5_5_10, \
call_12_5_5_11, \
call_12_5_5_12, \
call_12_5_5_13, \
call_12_5_5_14, \
call_12_5_5_15, \
call_12_5_6_0, \
call_12_5_6_1, \
call_12_5_6_2, \
call_12_5_6_3, \
call_12_5_6_4, \
call_12_5_6_5, \
call_12_5_6_6, \
call_12_5_6_7, \
call_12_5_6_8, \
call_12_5_6_9, \
call_12_5_6_10, \
call_12_5_6_11, \
call_12_5_6_12, \
call_12_5_6_13, \
call_12_5_6_14, \
call_12_5_6_15, \
call_12_5_7_0, \
call_12_5_7_1, \
call_12_5_7_2, \
call_12_5_7_3, \
call_12_5_7_4, \
call_12_5_7_5, \
call_12_5_7_6, \
call_12_5_7_7, \
call_12_5_7_8, \
call_12_5_7_9, \
call_12_5_7_10, \
call_12_5_7_11, \
call_12_5_7_12, \
call_12_5_7_13, \
call_12_5_7_14, \
call_12_5_7_15, \
call_12_5_8_0, \
call_12_5_8_1, \
call_12_5_8_2, \
call_12_5_8_3, \
call_12_5_8_4, \
call_12_5_8_5, \
call_12_5_8_6, \
call_12_5_8_7, \
call_12_5_8_8, \
call_12_5_8_9, \
call_12_5_8_10, \
call_12_5_8_11, \
call_12_5_8_12, \
call_12_5_8_13, \
call_12_5_8_14, \
call_12_5_8_15, \
call_12_5_9_0, \
call_12_5_9_1, \
call_12_5_9_2, \
call_12_5_9_3, \
call_12_5_9_4, \
call_12_5_9_5, \
call_12_5_9_6, \
call_12_5_9_7, \
call_12_5_9_8, \
call_12_5_9_9, \
call_12_5_9_10, \
call_12_5_9_11, \
call_12_5_9_12, \
call_12_5_9_13, \
call_12_5_9_14, \
call_12_5_9_15, \
call_12_5_10_0, \
call_12_5_10_1, \
call_12_5_10_2, \
call_12_5_10_3, \
call_12_5_10_4, \
call_12_5_10_5, \
call_12_5_10_6, \
call_12_5_10_7, \
call_12_5_10_8, \
call_12_5_10_9, \
call_12_5_10_10, \
call_12_5_10_11, \
call_12_5_10_12, \
call_12_5_10_13, \
call_12_5_10_14, \
call_12_5_10_15, \
call_12_5_11_0, \
call_12_5_11_1, \
call_12_5_11_2, \
call_12_5_11_3, \
call_12_5_11_4, \
call_12_5_11_5, \
call_12_5_11_6, \
call_12_5_11_7, \
call_12_5_11_8, \
call_12_5_11_9, \
call_12_5_11_10, \
call_12_5_11_11, \
call_12_5_11_12, \
call_12_5_11_13, \
call_12_5_11_14, \
call_12_5_11_15, \
call_12_5_12_0, \
call_12_5_12_1, \
call_12_5_12_2, \
call_12_5_12_3, \
call_12_5_12_4, \
call_12_5_12_5, \
call_12_5_12_6, \
call_12_5_12_7, \
call_12_5_12_8, \
call_12_5_12_9, \
call_12_5_12_10, \
call_12_5_12_11, \
call_12_5_12_12, \
call_12_5_12_13, \
call_12_5_12_14, \
call_12_5_12_15, \
call_12_5_13_0, \
call_12_5_13_1, \
call_12_5_13_2, \
call_12_5_13_3, \
call_12_5_13_4, \
call_12_5_13_5, \
call_12_5_13_6, \
call_12_5_13_7, \
call_12_5_13_8, \
call_12_5_13_9, \
call_12_5_13_10, \
call_12_5_13_11, \
call_12_5_13_12, \
call_12_5_13_13, \
call_12_5_13_14, \
call_12_5_13_15, \
call_12_5_14_0, \
call_12_5_14_1, \
call_12_5_14_2, \
call_12_5_14_3, \
call_12_5_14_4, \
call_12_5_14_5, \
call_12_5_14_6, \
call_12_5_14_7, \
call_12_5_14_8, \
call_12_5_14_9, \
call_12_5_14_10, \
call_12_5_14_11, \
call_12_5_14_12, \
call_12_5_14_13, \
call_12_5_14_14, \
call_12_5_14_15, \
call_12_5_15_0, \
call_12_5_15_1, \
call_12_5_15_2, \
call_12_5_15_3, \
call_12_5_15_4, \
call_12_5_15_5, \
call_12_5_15_6, \
call_12_5_15_7, \
call_12_5_15_8, \
call_12_5_15_9, \
call_12_5_15_10, \
call_12_5_15_11, \
call_12_5_15_12, \
call_12_5_15_13, \
call_12_5_15_14, \
call_12_5_15_15, \
call_12_6_0_0, \
call_12_6_0_1, \
call_12_6_0_2, \
call_12_6_0_3, \
call_12_6_0_4, \
call_12_6_0_5, \
call_12_6_0_6, \
call_12_6_0_7, \
call_12_6_0_8, \
call_12_6_0_9, \
call_12_6_0_10, \
call_12_6_0_11, \
call_12_6_0_12, \
call_12_6_0_13, \
call_12_6_0_14, \
call_12_6_0_15, \
call_12_6_1_0, \
call_12_6_1_1, \
call_12_6_1_2, \
call_12_6_1_3, \
call_12_6_1_4, \
call_12_6_1_5, \
call_12_6_1_6, \
call_12_6_1_7, \
call_12_6_1_8, \
call_12_6_1_9, \
call_12_6_1_10, \
call_12_6_1_11, \
call_12_6_1_12, \
call_12_6_1_13, \
call_12_6_1_14, \
call_12_6_1_15, \
call_12_6_2_0, \
call_12_6_2_1, \
call_12_6_2_2, \
call_12_6_2_3, \
call_12_6_2_4, \
call_12_6_2_5, \
call_12_6_2_6, \
call_12_6_2_7, \
call_12_6_2_8, \
call_12_6_2_9, \
call_12_6_2_10, \
call_12_6_2_11, \
call_12_6_2_12, \
call_12_6_2_13, \
call_12_6_2_14, \
call_12_6_2_15, \
call_12_6_3_0, \
call_12_6_3_1, \
call_12_6_3_2, \
call_12_6_3_3, \
call_12_6_3_4, \
call_12_6_3_5, \
call_12_6_3_6, \
call_12_6_3_7, \
call_12_6_3_8, \
call_12_6_3_9, \
call_12_6_3_10, \
call_12_6_3_11, \
call_12_6_3_12, \
call_12_6_3_13, \
call_12_6_3_14, \
call_12_6_3_15, \
call_12_6_4_0, \
call_12_6_4_1, \
call_12_6_4_2, \
call_12_6_4_3, \
call_12_6_4_4, \
call_12_6_4_5, \
call_12_6_4_6, \
call_12_6_4_7, \
call_12_6_4_8, \
call_12_6_4_9, \
call_12_6_4_10, \
call_12_6_4_11, \
call_12_6_4_12, \
call_12_6_4_13, \
call_12_6_4_14, \
call_12_6_4_15, \
call_12_6_5_0, \
call_12_6_5_1, \
call_12_6_5_2, \
call_12_6_5_3, \
call_12_6_5_4, \
call_12_6_5_5, \
call_12_6_5_6, \
call_12_6_5_7, \
call_12_6_5_8, \
call_12_6_5_9, \
call_12_6_5_10, \
call_12_6_5_11, \
call_12_6_5_12, \
call_12_6_5_13, \
call_12_6_5_14, \
call_12_6_5_15, \
call_12_6_6_0, \
call_12_6_6_1, \
call_12_6_6_2, \
call_12_6_6_3, \
call_12_6_6_4, \
call_12_6_6_5, \
call_12_6_6_6, \
call_12_6_6_7, \
call_12_6_6_8, \
call_12_6_6_9, \
call_12_6_6_10, \
call_12_6_6_11, \
call_12_6_6_12, \
call_12_6_6_13, \
call_12_6_6_14, \
call_12_6_6_15, \
call_12_6_7_0, \
call_12_6_7_1, \
call_12_6_7_2, \
call_12_6_7_3, \
call_12_6_7_4, \
call_12_6_7_5, \
call_12_6_7_6, \
call_12_6_7_7, \
call_12_6_7_8, \
call_12_6_7_9, \
call_12_6_7_10, \
call_12_6_7_11, \
call_12_6_7_12, \
call_12_6_7_13, \
call_12_6_7_14, \
call_12_6_7_15, \
call_12_6_8_0, \
call_12_6_8_1, \
call_12_6_8_2, \
call_12_6_8_3, \
call_12_6_8_4, \
call_12_6_8_5, \
call_12_6_8_6, \
call_12_6_8_7, \
call_12_6_8_8, \
call_12_6_8_9, \
call_12_6_8_10, \
call_12_6_8_11, \
call_12_6_8_12, \
call_12_6_8_13, \
call_12_6_8_14, \
call_12_6_8_15, \
call_12_6_9_0, \
call_12_6_9_1, \
call_12_6_9_2, \
call_12_6_9_3, \
call_12_6_9_4, \
call_12_6_9_5, \
call_12_6_9_6, \
call_12_6_9_7, \
call_12_6_9_8, \
call_12_6_9_9, \
call_12_6_9_10, \
call_12_6_9_11, \
call_12_6_9_12, \
call_12_6_9_13, \
call_12_6_9_14, \
call_12_6_9_15, \
call_12_6_10_0, \
call_12_6_10_1, \
call_12_6_10_2, \
call_12_6_10_3, \
call_12_6_10_4, \
call_12_6_10_5, \
call_12_6_10_6, \
call_12_6_10_7, \
call_12_6_10_8, \
call_12_6_10_9, \
call_12_6_10_10, \
call_12_6_10_11, \
call_12_6_10_12, \
call_12_6_10_13, \
call_12_6_10_14, \
call_12_6_10_15, \
call_12_6_11_0, \
call_12_6_11_1, \
call_12_6_11_2, \
call_12_6_11_3, \
call_12_6_11_4, \
call_12_6_11_5, \
call_12_6_11_6, \
call_12_6_11_7, \
call_12_6_11_8, \
call_12_6_11_9, \
call_12_6_11_10, \
call_12_6_11_11, \
call_12_6_11_12, \
call_12_6_11_13, \
call_12_6_11_14, \
call_12_6_11_15, \
call_12_6_12_0, \
call_12_6_12_1, \
call_12_6_12_2, \
call_12_6_12_3, \
call_12_6_12_4, \
call_12_6_12_5, \
call_12_6_12_6, \
call_12_6_12_7, \
call_12_6_12_8, \
call_12_6_12_9, \
call_12_6_12_10, \
call_12_6_12_11, \
call_12_6_12_12, \
call_12_6_12_13, \
call_12_6_12_14, \
call_12_6_12_15, \
call_12_6_13_0, \
call_12_6_13_1, \
call_12_6_13_2, \
call_12_6_13_3, \
call_12_6_13_4, \
call_12_6_13_5, \
call_12_6_13_6, \
call_12_6_13_7, \
call_12_6_13_8, \
call_12_6_13_9, \
call_12_6_13_10, \
call_12_6_13_11, \
call_12_6_13_12, \
call_12_6_13_13, \
call_12_6_13_14, \
call_12_6_13_15, \
call_12_6_14_0, \
call_12_6_14_1, \
call_12_6_14_2, \
call_12_6_14_3, \
call_12_6_14_4, \
call_12_6_14_5, \
call_12_6_14_6, \
call_12_6_14_7, \
call_12_6_14_8, \
call_12_6_14_9, \
call_12_6_14_10, \
call_12_6_14_11, \
call_12_6_14_12, \
call_12_6_14_13, \
call_12_6_14_14, \
call_12_6_14_15, \
call_12_6_15_0, \
call_12_6_15_1, \
call_12_6_15_2, \
call_12_6_15_3, \
call_12_6_15_4, \
call_12_6_15_5, \
call_12_6_15_6, \
call_12_6_15_7, \
call_12_6_15_8, \
call_12_6_15_9, \
call_12_6_15_10, \
call_12_6_15_11, \
call_12_6_15_12, \
call_12_6_15_13, \
call_12_6_15_14, \
call_12_6_15_15, \
call_12_7_0_0, \
call_12_7_0_1, \
call_12_7_0_2, \
call_12_7_0_3, \
call_12_7_0_4, \
call_12_7_0_5, \
call_12_7_0_6, \
call_12_7_0_7, \
call_12_7_0_8, \
call_12_7_0_9, \
call_12_7_0_10, \
call_12_7_0_11, \
call_12_7_0_12, \
call_12_7_0_13, \
call_12_7_0_14, \
call_12_7_0_15, \
call_12_7_1_0, \
call_12_7_1_1, \
call_12_7_1_2, \
call_12_7_1_3, \
call_12_7_1_4, \
call_12_7_1_5, \
call_12_7_1_6, \
call_12_7_1_7, \
call_12_7_1_8, \
call_12_7_1_9, \
call_12_7_1_10, \
call_12_7_1_11, \
call_12_7_1_12, \
call_12_7_1_13, \
call_12_7_1_14, \
call_12_7_1_15, \
call_12_7_2_0, \
call_12_7_2_1, \
call_12_7_2_2, \
call_12_7_2_3, \
call_12_7_2_4, \
call_12_7_2_5, \
call_12_7_2_6, \
call_12_7_2_7, \
call_12_7_2_8, \
call_12_7_2_9, \
call_12_7_2_10, \
call_12_7_2_11, \
call_12_7_2_12, \
call_12_7_2_13, \
call_12_7_2_14, \
call_12_7_2_15, \
call_12_7_3_0, \
call_12_7_3_1, \
call_12_7_3_2, \
call_12_7_3_3, \
call_12_7_3_4, \
call_12_7_3_5, \
call_12_7_3_6, \
call_12_7_3_7, \
call_12_7_3_8, \
call_12_7_3_9, \
call_12_7_3_10, \
call_12_7_3_11, \
call_12_7_3_12, \
call_12_7_3_13, \
call_12_7_3_14, \
call_12_7_3_15, \
call_12_7_4_0, \
call_12_7_4_1, \
call_12_7_4_2, \
call_12_7_4_3, \
call_12_7_4_4, \
call_12_7_4_5, \
call_12_7_4_6, \
call_12_7_4_7, \
call_12_7_4_8, \
call_12_7_4_9, \
call_12_7_4_10, \
call_12_7_4_11, \
call_12_7_4_12, \
call_12_7_4_13, \
call_12_7_4_14, \
call_12_7_4_15, \
call_12_7_5_0, \
call_12_7_5_1, \
call_12_7_5_2, \
call_12_7_5_3, \
call_12_7_5_4, \
call_12_7_5_5, \
call_12_7_5_6, \
call_12_7_5_7, \
call_12_7_5_8, \
call_12_7_5_9, \
call_12_7_5_10, \
call_12_7_5_11, \
call_12_7_5_12, \
call_12_7_5_13, \
call_12_7_5_14, \
call_12_7_5_15, \
call_12_7_6_0, \
call_12_7_6_1, \
call_12_7_6_2, \
call_12_7_6_3, \
call_12_7_6_4, \
call_12_7_6_5, \
call_12_7_6_6, \
call_12_7_6_7, \
call_12_7_6_8, \
call_12_7_6_9, \
call_12_7_6_10, \
call_12_7_6_11, \
call_12_7_6_12, \
call_12_7_6_13, \
call_12_7_6_14, \
call_12_7_6_15, \
call_12_7_7_0, \
call_12_7_7_1, \
call_12_7_7_2, \
call_12_7_7_3, \
call_12_7_7_4, \
call_12_7_7_5, \
call_12_7_7_6, \
call_12_7_7_7, \
call_12_7_7_8, \
call_12_7_7_9, \
call_12_7_7_10, \
call_12_7_7_11, \
call_12_7_7_12, \
call_12_7_7_13, \
call_12_7_7_14, \
call_12_7_7_15, \
call_12_7_8_0, \
call_12_7_8_1, \
call_12_7_8_2, \
call_12_7_8_3, \
call_12_7_8_4, \
call_12_7_8_5, \
call_12_7_8_6, \
call_12_7_8_7, \
call_12_7_8_8, \
call_12_7_8_9, \
call_12_7_8_10, \
call_12_7_8_11, \
call_12_7_8_12, \
call_12_7_8_13, \
call_12_7_8_14, \
call_12_7_8_15, \
call_12_7_9_0, \
call_12_7_9_1, \
call_12_7_9_2, \
call_12_7_9_3, \
call_12_7_9_4, \
call_12_7_9_5, \
call_12_7_9_6, \
call_12_7_9_7, \
call_12_7_9_8, \
call_12_7_9_9, \
call_12_7_9_10, \
call_12_7_9_11, \
call_12_7_9_12, \
call_12_7_9_13, \
call_12_7_9_14, \
call_12_7_9_15, \
call_12_7_10_0, \
call_12_7_10_1, \
call_12_7_10_2, \
call_12_7_10_3, \
call_12_7_10_4, \
call_12_7_10_5, \
call_12_7_10_6, \
call_12_7_10_7, \
call_12_7_10_8, \
call_12_7_10_9, \
call_12_7_10_10, \
call_12_7_10_11, \
call_12_7_10_12, \
call_12_7_10_13, \
call_12_7_10_14, \
call_12_7_10_15, \
call_12_7_11_0, \
call_12_7_11_1, \
call_12_7_11_2, \
call_12_7_11_3, \
call_12_7_11_4, \
call_12_7_11_5, \
call_12_7_11_6, \
call_12_7_11_7, \
call_12_7_11_8, \
call_12_7_11_9, \
call_12_7_11_10, \
call_12_7_11_11, \
call_12_7_11_12, \
call_12_7_11_13, \
call_12_7_11_14, \
call_12_7_11_15, \
call_12_7_12_0, \
call_12_7_12_1, \
call_12_7_12_2, \
call_12_7_12_3, \
call_12_7_12_4, \
call_12_7_12_5, \
call_12_7_12_6, \
call_12_7_12_7, \
call_12_7_12_8, \
call_12_7_12_9, \
call_12_7_12_10, \
call_12_7_12_11, \
call_12_7_12_12, \
call_12_7_12_13, \
call_12_7_12_14, \
call_12_7_12_15, \
call_12_7_13_0, \
call_12_7_13_1, \
call_12_7_13_2, \
call_12_7_13_3, \
call_12_7_13_4, \
call_12_7_13_5, \
call_12_7_13_6, \
call_12_7_13_7, \
call_12_7_13_8, \
call_12_7_13_9, \
call_12_7_13_10, \
call_12_7_13_11, \
call_12_7_13_12, \
call_12_7_13_13, \
call_12_7_13_14, \
call_12_7_13_15, \
call_12_7_14_0, \
call_12_7_14_1, \
call_12_7_14_2, \
call_12_7_14_3, \
call_12_7_14_4, \
call_12_7_14_5, \
call_12_7_14_6, \
call_12_7_14_7, \
call_12_7_14_8, \
call_12_7_14_9, \
call_12_7_14_10, \
call_12_7_14_11, \
call_12_7_14_12, \
call_12_7_14_13, \
call_12_7_14_14, \
call_12_7_14_15, \
call_12_7_15_0, \
call_12_7_15_1, \
call_12_7_15_2, \
call_12_7_15_3, \
call_12_7_15_4, \
call_12_7_15_5, \
call_12_7_15_6, \
call_12_7_15_7, \
call_12_7_15_8, \
call_12_7_15_9, \
call_12_7_15_10, \
call_12_7_15_11, \
call_12_7_15_12, \
call_12_7_15_13, \
call_12_7_15_14, \
call_12_7_15_15, \
call_12_8_0_0, \
call_12_8_0_1, \
call_12_8_0_2, \
call_12_8_0_3, \
call_12_8_0_4, \
call_12_8_0_5, \
call_12_8_0_6, \
call_12_8_0_7, \
call_12_8_0_8, \
call_12_8_0_9, \
call_12_8_0_10, \
call_12_8_0_11, \
call_12_8_0_12, \
call_12_8_0_13, \
call_12_8_0_14, \
call_12_8_0_15, \
call_12_8_1_0, \
call_12_8_1_1, \
call_12_8_1_2, \
call_12_8_1_3, \
call_12_8_1_4, \
call_12_8_1_5, \
call_12_8_1_6, \
call_12_8_1_7, \
call_12_8_1_8, \
call_12_8_1_9, \
call_12_8_1_10, \
call_12_8_1_11, \
call_12_8_1_12, \
call_12_8_1_13, \
call_12_8_1_14, \
call_12_8_1_15, \
call_12_8_2_0, \
call_12_8_2_1, \
call_12_8_2_2, \
call_12_8_2_3, \
call_12_8_2_4, \
call_12_8_2_5, \
call_12_8_2_6, \
call_12_8_2_7, \
call_12_8_2_8, \
call_12_8_2_9, \
call_12_8_2_10, \
call_12_8_2_11, \
call_12_8_2_12, \
call_12_8_2_13, \
call_12_8_2_14, \
call_12_8_2_15, \
call_12_8_3_0, \
call_12_8_3_1, \
call_12_8_3_2, \
call_12_8_3_3, \
call_12_8_3_4, \
call_12_8_3_5, \
call_12_8_3_6, \
call_12_8_3_7, \
call_12_8_3_8, \
call_12_8_3_9, \
call_12_8_3_10, \
call_12_8_3_11, \
call_12_8_3_12, \
call_12_8_3_13, \
call_12_8_3_14, \
call_12_8_3_15, \
call_12_8_4_0, \
call_12_8_4_1, \
call_12_8_4_2, \
call_12_8_4_3, \
call_12_8_4_4, \
call_12_8_4_5, \
call_12_8_4_6, \
call_12_8_4_7, \
call_12_8_4_8, \
call_12_8_4_9, \
call_12_8_4_10, \
call_12_8_4_11, \
call_12_8_4_12, \
call_12_8_4_13, \
call_12_8_4_14, \
call_12_8_4_15, \
call_12_8_5_0, \
call_12_8_5_1, \
call_12_8_5_2, \
call_12_8_5_3, \
call_12_8_5_4, \
call_12_8_5_5, \
call_12_8_5_6, \
call_12_8_5_7, \
call_12_8_5_8, \
call_12_8_5_9, \
call_12_8_5_10, \
call_12_8_5_11, \
call_12_8_5_12, \
call_12_8_5_13, \
call_12_8_5_14, \
call_12_8_5_15, \
call_12_8_6_0, \
call_12_8_6_1, \
call_12_8_6_2, \
call_12_8_6_3, \
call_12_8_6_4, \
call_12_8_6_5, \
call_12_8_6_6, \
call_12_8_6_7, \
call_12_8_6_8, \
call_12_8_6_9, \
call_12_8_6_10, \
call_12_8_6_11, \
call_12_8_6_12, \
call_12_8_6_13, \
call_12_8_6_14, \
call_12_8_6_15, \
call_12_8_7_0, \
call_12_8_7_1, \
call_12_8_7_2, \
call_12_8_7_3, \
call_12_8_7_4, \
call_12_8_7_5, \
call_12_8_7_6, \
call_12_8_7_7, \
call_12_8_7_8, \
call_12_8_7_9, \
call_12_8_7_10, \
call_12_8_7_11, \
call_12_8_7_12, \
call_12_8_7_13, \
call_12_8_7_14, \
call_12_8_7_15, \
call_12_8_8_0, \
call_12_8_8_1, \
call_12_8_8_2, \
call_12_8_8_3, \
call_12_8_8_4, \
call_12_8_8_5, \
call_12_8_8_6, \
call_12_8_8_7, \
call_12_8_8_8, \
call_12_8_8_9, \
call_12_8_8_10, \
call_12_8_8_11, \
call_12_8_8_12, \
call_12_8_8_13, \
call_12_8_8_14, \
call_12_8_8_15, \
call_12_8_9_0, \
call_12_8_9_1, \
call_12_8_9_2, \
call_12_8_9_3, \
call_12_8_9_4, \
call_12_8_9_5, \
call_12_8_9_6, \
call_12_8_9_7, \
call_12_8_9_8, \
call_12_8_9_9, \
call_12_8_9_10, \
call_12_8_9_11, \
call_12_8_9_12, \
call_12_8_9_13, \
call_12_8_9_14, \
call_12_8_9_15, \
call_12_8_10_0, \
call_12_8_10_1, \
call_12_8_10_2, \
call_12_8_10_3, \
call_12_8_10_4, \
call_12_8_10_5, \
call_12_8_10_6, \
call_12_8_10_7, \
call_12_8_10_8, \
call_12_8_10_9, \
call_12_8_10_10, \
call_12_8_10_11, \
call_12_8_10_12, \
call_12_8_10_13, \
call_12_8_10_14, \
call_12_8_10_15, \
call_12_8_11_0, \
call_12_8_11_1, \
call_12_8_11_2, \
call_12_8_11_3, \
call_12_8_11_4, \
call_12_8_11_5, \
call_12_8_11_6, \
call_12_8_11_7, \
call_12_8_11_8, \
call_12_8_11_9, \
call_12_8_11_10, \
call_12_8_11_11, \
call_12_8_11_12, \
call_12_8_11_13, \
call_12_8_11_14, \
call_12_8_11_15, \
call_12_8_12_0, \
call_12_8_12_1, \
call_12_8_12_2, \
call_12_8_12_3, \
call_12_8_12_4, \
call_12_8_12_5, \
call_12_8_12_6, \
call_12_8_12_7, \
call_12_8_12_8, \
call_12_8_12_9, \
call_12_8_12_10, \
call_12_8_12_11, \
call_12_8_12_12, \
call_12_8_12_13, \
call_12_8_12_14, \
call_12_8_12_15, \
call_12_8_13_0, \
call_12_8_13_1, \
call_12_8_13_2, \
call_12_8_13_3, \
call_12_8_13_4, \
call_12_8_13_5, \
call_12_8_13_6, \
call_12_8_13_7, \
call_12_8_13_8, \
call_12_8_13_9, \
call_12_8_13_10, \
call_12_8_13_11, \
call_12_8_13_12, \
call_12_8_13_13, \
call_12_8_13_14, \
call_12_8_13_15, \
call_12_8_14_0, \
call_12_8_14_1, \
call_12_8_14_2, \
call_12_8_14_3, \
call_12_8_14_4, \
call_12_8_14_5, \
call_12_8_14_6, \
call_12_8_14_7, \
call_12_8_14_8, \
call_12_8_14_9, \
call_12_8_14_10, \
call_12_8_14_11, \
call_12_8_14_12, \
call_12_8_14_13, \
call_12_8_14_14, \
call_12_8_14_15, \
call_12_8_15_0, \
call_12_8_15_1, \
call_12_8_15_2, \
call_12_8_15_3, \
call_12_8_15_4, \
call_12_8_15_5, \
call_12_8_15_6, \
call_12_8_15_7, \
call_12_8_15_8, \
call_12_8_15_9, \
call_12_8_15_10, \
call_12_8_15_11, \
call_12_8_15_12, \
call_12_8_15_13, \
call_12_8_15_14, \
call_12_8_15_15, \
call_12_9_0_0, \
call_12_9_0_1, \
call_12_9_0_2, \
call_12_9_0_3, \
call_12_9_0_4, \
call_12_9_0_5, \
call_12_9_0_6, \
call_12_9_0_7, \
call_12_9_0_8, \
call_12_9_0_9, \
call_12_9_0_10, \
call_12_9_0_11, \
call_12_9_0_12, \
call_12_9_0_13, \
call_12_9_0_14, \
call_12_9_0_15, \
call_12_9_1_0, \
call_12_9_1_1, \
call_12_9_1_2, \
call_12_9_1_3, \
call_12_9_1_4, \
call_12_9_1_5, \
call_12_9_1_6, \
call_12_9_1_7, \
call_12_9_1_8, \
call_12_9_1_9, \
call_12_9_1_10, \
call_12_9_1_11, \
call_12_9_1_12, \
call_12_9_1_13, \
call_12_9_1_14, \
call_12_9_1_15, \
call_12_9_2_0, \
call_12_9_2_1, \
call_12_9_2_2, \
call_12_9_2_3, \
call_12_9_2_4, \
call_12_9_2_5, \
call_12_9_2_6, \
call_12_9_2_7, \
call_12_9_2_8, \
call_12_9_2_9, \
call_12_9_2_10, \
call_12_9_2_11, \
call_12_9_2_12, \
call_12_9_2_13, \
call_12_9_2_14, \
call_12_9_2_15, \
call_12_9_3_0, \
call_12_9_3_1, \
call_12_9_3_2, \
call_12_9_3_3, \
call_12_9_3_4, \
call_12_9_3_5, \
call_12_9_3_6, \
call_12_9_3_7, \
call_12_9_3_8, \
call_12_9_3_9, \
call_12_9_3_10, \
call_12_9_3_11, \
call_12_9_3_12, \
call_12_9_3_13, \
call_12_9_3_14, \
call_12_9_3_15, \
call_12_9_4_0, \
call_12_9_4_1, \
call_12_9_4_2, \
call_12_9_4_3, \
call_12_9_4_4, \
call_12_9_4_5, \
call_12_9_4_6, \
call_12_9_4_7, \
call_12_9_4_8, \
call_12_9_4_9, \
call_12_9_4_10, \
call_12_9_4_11, \
call_12_9_4_12, \
call_12_9_4_13, \
call_12_9_4_14, \
call_12_9_4_15, \
call_12_9_5_0, \
call_12_9_5_1, \
call_12_9_5_2, \
call_12_9_5_3, \
call_12_9_5_4, \
call_12_9_5_5, \
call_12_9_5_6, \
call_12_9_5_7, \
call_12_9_5_8, \
call_12_9_5_9, \
call_12_9_5_10, \
call_12_9_5_11, \
call_12_9_5_12, \
call_12_9_5_13, \
call_12_9_5_14, \
call_12_9_5_15, \
call_12_9_6_0, \
call_12_9_6_1, \
call_12_9_6_2, \
call_12_9_6_3, \
call_12_9_6_4, \
call_12_9_6_5, \
call_12_9_6_6, \
call_12_9_6_7, \
call_12_9_6_8, \
call_12_9_6_9, \
call_12_9_6_10, \
call_12_9_6_11, \
call_12_9_6_12, \
call_12_9_6_13, \
call_12_9_6_14, \
call_12_9_6_15, \
call_12_9_7_0, \
call_12_9_7_1, \
call_12_9_7_2, \
call_12_9_7_3, \
call_12_9_7_4, \
call_12_9_7_5, \
call_12_9_7_6, \
call_12_9_7_7, \
call_12_9_7_8, \
call_12_9_7_9, \
call_12_9_7_10, \
call_12_9_7_11, \
call_12_9_7_12, \
call_12_9_7_13, \
call_12_9_7_14, \
call_12_9_7_15, \
call_12_9_8_0, \
call_12_9_8_1, \
call_12_9_8_2, \
call_12_9_8_3, \
call_12_9_8_4, \
call_12_9_8_5, \
call_12_9_8_6, \
call_12_9_8_7, \
call_12_9_8_8, \
call_12_9_8_9, \
call_12_9_8_10, \
call_12_9_8_11, \
call_12_9_8_12, \
call_12_9_8_13, \
call_12_9_8_14, \
call_12_9_8_15, \
call_12_9_9_0, \
call_12_9_9_1, \
call_12_9_9_2, \
call_12_9_9_3, \
call_12_9_9_4, \
call_12_9_9_5, \
call_12_9_9_6, \
call_12_9_9_7, \
call_12_9_9_8, \
call_12_9_9_9, \
call_12_9_9_10, \
call_12_9_9_11, \
call_12_9_9_12, \
call_12_9_9_13, \
call_12_9_9_14, \
call_12_9_9_15, \
call_12_9_10_0, \
call_12_9_10_1, \
call_12_9_10_2, \
call_12_9_10_3, \
call_12_9_10_4, \
call_12_9_10_5, \
call_12_9_10_6, \
call_12_9_10_7, \
call_12_9_10_8, \
call_12_9_10_9, \
call_12_9_10_10, \
call_12_9_10_11, \
call_12_9_10_12, \
call_12_9_10_13, \
call_12_9_10_14, \
call_12_9_10_15, \
call_12_9_11_0, \
call_12_9_11_1, \
call_12_9_11_2, \
call_12_9_11_3, \
call_12_9_11_4, \
call_12_9_11_5, \
call_12_9_11_6, \
call_12_9_11_7, \
call_12_9_11_8, \
call_12_9_11_9, \
call_12_9_11_10, \
call_12_9_11_11, \
call_12_9_11_12, \
call_12_9_11_13, \
call_12_9_11_14, \
call_12_9_11_15, \
call_12_9_12_0, \
call_12_9_12_1, \
call_12_9_12_2, \
call_12_9_12_3, \
call_12_9_12_4, \
call_12_9_12_5, \
call_12_9_12_6, \
call_12_9_12_7, \
call_12_9_12_8, \
call_12_9_12_9, \
call_12_9_12_10, \
call_12_9_12_11, \
call_12_9_12_12, \
call_12_9_12_13, \
call_12_9_12_14, \
call_12_9_12_15, \
call_12_9_13_0, \
call_12_9_13_1, \
call_12_9_13_2, \
call_12_9_13_3, \
call_12_9_13_4, \
call_12_9_13_5, \
call_12_9_13_6, \
call_12_9_13_7, \
call_12_9_13_8, \
call_12_9_13_9, \
call_12_9_13_10, \
call_12_9_13_11, \
call_12_9_13_12, \
call_12_9_13_13, \
call_12_9_13_14, \
call_12_9_13_15, \
call_12_9_14_0, \
call_12_9_14_1, \
call_12_9_14_2, \
call_12_9_14_3, \
call_12_9_14_4, \
call_12_9_14_5, \
call_12_9_14_6, \
call_12_9_14_7, \
call_12_9_14_8, \
call_12_9_14_9, \
call_12_9_14_10, \
call_12_9_14_11, \
call_12_9_14_12, \
call_12_9_14_13, \
call_12_9_14_14, \
call_12_9_14_15, \
call_12_9_15_0, \
call_12_9_15_1, \
call_12_9_15_2, \
call_12_9_15_3, \
call_12_9_15_4, \
call_12_9_15_5, \
call_12_9_15_6, \
call_12_9_15_7, \
call_12_9_15_8, \
call_12_9_15_9, \
call_12_9_15_10, \
call_12_9_15_11, \
call_12_9_15_12, \
call_12_9_15_13, \
call_12_9_15_14, \
call_12_9_15_15, \
call_12_10_0_0, \
call_12_10_0_1, \
call_12_10_0_2, \
call_12_10_0_3, \
call_12_10_0_4, \
call_12_10_0_5, \
call_12_10_0_6, \
call_12_10_0_7, \
call_12_10_0_8, \
call_12_10_0_9, \
call_12_10_0_10, \
call_12_10_0_11, \
call_12_10_0_12, \
call_12_10_0_13, \
call_12_10_0_14, \
call_12_10_0_15, \
call_12_10_1_0, \
call_12_10_1_1, \
call_12_10_1_2, \
call_12_10_1_3, \
call_12_10_1_4, \
call_12_10_1_5, \
call_12_10_1_6, \
call_12_10_1_7, \
call_12_10_1_8, \
call_12_10_1_9, \
call_12_10_1_10, \
call_12_10_1_11, \
call_12_10_1_12, \
call_12_10_1_13, \
call_12_10_1_14, \
call_12_10_1_15, \
call_12_10_2_0, \
call_12_10_2_1, \
call_12_10_2_2, \
call_12_10_2_3, \
call_12_10_2_4, \
call_12_10_2_5, \
call_12_10_2_6, \
call_12_10_2_7, \
call_12_10_2_8, \
call_12_10_2_9, \
call_12_10_2_10, \
call_12_10_2_11, \
call_12_10_2_12, \
call_12_10_2_13, \
call_12_10_2_14, \
call_12_10_2_15, \
call_12_10_3_0, \
call_12_10_3_1, \
call_12_10_3_2, \
call_12_10_3_3, \
call_12_10_3_4, \
call_12_10_3_5, \
call_12_10_3_6, \
call_12_10_3_7, \
call_12_10_3_8, \
call_12_10_3_9, \
call_12_10_3_10, \
call_12_10_3_11, \
call_12_10_3_12, \
call_12_10_3_13, \
call_12_10_3_14, \
call_12_10_3_15, \
call_12_10_4_0, \
call_12_10_4_1, \
call_12_10_4_2, \
call_12_10_4_3, \
call_12_10_4_4, \
call_12_10_4_5, \
call_12_10_4_6, \
call_12_10_4_7, \
call_12_10_4_8, \
call_12_10_4_9, \
call_12_10_4_10, \
call_12_10_4_11, \
call_12_10_4_12, \
call_12_10_4_13, \
call_12_10_4_14, \
call_12_10_4_15, \
call_12_10_5_0, \
call_12_10_5_1, \
call_12_10_5_2, \
call_12_10_5_3, \
call_12_10_5_4, \
call_12_10_5_5, \
call_12_10_5_6, \
call_12_10_5_7, \
call_12_10_5_8, \
call_12_10_5_9, \
call_12_10_5_10, \
call_12_10_5_11, \
call_12_10_5_12, \
call_12_10_5_13, \
call_12_10_5_14, \
call_12_10_5_15, \
call_12_10_6_0, \
call_12_10_6_1, \
call_12_10_6_2, \
call_12_10_6_3, \
call_12_10_6_4, \
call_12_10_6_5, \
call_12_10_6_6, \
call_12_10_6_7, \
call_12_10_6_8, \
call_12_10_6_9, \
call_12_10_6_10, \
call_12_10_6_11, \
call_12_10_6_12, \
call_12_10_6_13, \
call_12_10_6_14, \
call_12_10_6_15, \
call_12_10_7_0, \
call_12_10_7_1, \
call_12_10_7_2, \
call_12_10_7_3, \
call_12_10_7_4, \
call_12_10_7_5, \
call_12_10_7_6, \
call_12_10_7_7, \
call_12_10_7_8, \
call_12_10_7_9, \
call_12_10_7_10, \
call_12_10_7_11, \
call_12_10_7_12, \
call_12_10_7_13, \
call_12_10_7_14, \
call_12_10_7_15, \
call_12_10_8_0, \
call_12_10_8_1, \
call_12_10_8_2, \
call_12_10_8_3, \
call_12_10_8_4, \
call_12_10_8_5, \
call_12_10_8_6, \
call_12_10_8_7, \
call_12_10_8_8, \
call_12_10_8_9, \
call_12_10_8_10, \
call_12_10_8_11, \
call_12_10_8_12, \
call_12_10_8_13, \
call_12_10_8_14, \
call_12_10_8_15, \
call_12_10_9_0, \
call_12_10_9_1, \
call_12_10_9_2, \
call_12_10_9_3, \
call_12_10_9_4, \
call_12_10_9_5, \
call_12_10_9_6, \
call_12_10_9_7, \
call_12_10_9_8, \
call_12_10_9_9, \
call_12_10_9_10, \
call_12_10_9_11, \
call_12_10_9_12, \
call_12_10_9_13, \
call_12_10_9_14, \
call_12_10_9_15, \
call_12_10_10_0, \
call_12_10_10_1, \
call_12_10_10_2, \
call_12_10_10_3, \
call_12_10_10_4, \
call_12_10_10_5, \
call_12_10_10_6, \
call_12_10_10_7, \
call_12_10_10_8, \
call_12_10_10_9, \
call_12_10_10_10, \
call_12_10_10_11, \
call_12_10_10_12, \
call_12_10_10_13, \
call_12_10_10_14, \
call_12_10_10_15, \
call_12_10_11_0, \
call_12_10_11_1, \
call_12_10_11_2, \
call_12_10_11_3, \
call_12_10_11_4, \
call_12_10_11_5, \
call_12_10_11_6, \
call_12_10_11_7, \
call_12_10_11_8, \
call_12_10_11_9, \
call_12_10_11_10, \
call_12_10_11_11, \
call_12_10_11_12, \
call_12_10_11_13, \
call_12_10_11_14, \
call_12_10_11_15, \
call_12_10_12_0, \
call_12_10_12_1, \
call_12_10_12_2, \
call_12_10_12_3, \
call_12_10_12_4, \
call_12_10_12_5, \
call_12_10_12_6, \
call_12_10_12_7, \
call_12_10_12_8, \
call_12_10_12_9, \
call_12_10_12_10, \
call_12_10_12_11, \
call_12_10_12_12, \
call_12_10_12_13, \
call_12_10_12_14, \
call_12_10_12_15, \
call_12_10_13_0, \
call_12_10_13_1, \
call_12_10_13_2, \
call_12_10_13_3, \
call_12_10_13_4, \
call_12_10_13_5, \
call_12_10_13_6, \
call_12_10_13_7, \
call_12_10_13_8, \
call_12_10_13_9, \
call_12_10_13_10, \
call_12_10_13_11, \
call_12_10_13_12, \
call_12_10_13_13, \
call_12_10_13_14, \
call_12_10_13_15, \
call_12_10_14_0, \
call_12_10_14_1, \
call_12_10_14_2, \
call_12_10_14_3, \
call_12_10_14_4, \
call_12_10_14_5, \
call_12_10_14_6, \
call_12_10_14_7, \
call_12_10_14_8, \
call_12_10_14_9, \
call_12_10_14_10, \
call_12_10_14_11, \
call_12_10_14_12, \
call_12_10_14_13, \
call_12_10_14_14, \
call_12_10_14_15, \
call_12_10_15_0, \
call_12_10_15_1, \
call_12_10_15_2, \
call_12_10_15_3, \
call_12_10_15_4, \
call_12_10_15_5, \
call_12_10_15_6, \
call_12_10_15_7, \
call_12_10_15_8, \
call_12_10_15_9, \
call_12_10_15_10, \
call_12_10_15_11, \
call_12_10_15_12, \
call_12_10_15_13, \
call_12_10_15_14, \
call_12_10_15_15, \
call_12_11_0_0, \
call_12_11_0_1, \
call_12_11_0_2, \
call_12_11_0_3, \
call_12_11_0_4, \
call_12_11_0_5, \
call_12_11_0_6, \
call_12_11_0_7, \
call_12_11_0_8, \
call_12_11_0_9, \
call_12_11_0_10, \
call_12_11_0_11, \
call_12_11_0_12, \
call_12_11_0_13, \
call_12_11_0_14, \
call_12_11_0_15, \
call_12_11_1_0, \
call_12_11_1_1, \
call_12_11_1_2, \
call_12_11_1_3, \
call_12_11_1_4, \
call_12_11_1_5, \
call_12_11_1_6, \
call_12_11_1_7, \
call_12_11_1_8, \
call_12_11_1_9, \
call_12_11_1_10, \
call_12_11_1_11, \
call_12_11_1_12, \
call_12_11_1_13, \
call_12_11_1_14, \
call_12_11_1_15, \
call_12_11_2_0, \
call_12_11_2_1, \
call_12_11_2_2, \
call_12_11_2_3, \
call_12_11_2_4, \
call_12_11_2_5, \
call_12_11_2_6, \
call_12_11_2_7, \
call_12_11_2_8, \
call_12_11_2_9, \
call_12_11_2_10, \
call_12_11_2_11, \
call_12_11_2_12, \
call_12_11_2_13, \
call_12_11_2_14, \
call_12_11_2_15, \
call_12_11_3_0, \
call_12_11_3_1, \
call_12_11_3_2, \
call_12_11_3_3, \
call_12_11_3_4, \
call_12_11_3_5, \
call_12_11_3_6, \
call_12_11_3_7, \
call_12_11_3_8, \
call_12_11_3_9, \
call_12_11_3_10, \
call_12_11_3_11, \
call_12_11_3_12, \
call_12_11_3_13, \
call_12_11_3_14, \
call_12_11_3_15, \
call_12_11_4_0, \
call_12_11_4_1, \
call_12_11_4_2, \
call_12_11_4_3, \
call_12_11_4_4, \
call_12_11_4_5, \
call_12_11_4_6, \
call_12_11_4_7, \
call_12_11_4_8, \
call_12_11_4_9, \
call_12_11_4_10, \
call_12_11_4_11, \
call_12_11_4_12, \
call_12_11_4_13, \
call_12_11_4_14, \
call_12_11_4_15, \
call_12_11_5_0, \
call_12_11_5_1, \
call_12_11_5_2, \
call_12_11_5_3, \
call_12_11_5_4, \
call_12_11_5_5, \
call_12_11_5_6, \
call_12_11_5_7, \
call_12_11_5_8, \
call_12_11_5_9, \
call_12_11_5_10, \
call_12_11_5_11, \
call_12_11_5_12, \
call_12_11_5_13, \
call_12_11_5_14, \
call_12_11_5_15, \
call_12_11_6_0, \
call_12_11_6_1, \
call_12_11_6_2, \
call_12_11_6_3, \
call_12_11_6_4, \
call_12_11_6_5, \
call_12_11_6_6, \
call_12_11_6_7, \
call_12_11_6_8, \
call_12_11_6_9, \
call_12_11_6_10, \
call_12_11_6_11, \
call_12_11_6_12, \
call_12_11_6_13, \
call_12_11_6_14, \
call_12_11_6_15, \
call_12_11_7_0, \
call_12_11_7_1, \
call_12_11_7_2, \
call_12_11_7_3, \
call_12_11_7_4, \
call_12_11_7_5, \
call_12_11_7_6, \
call_12_11_7_7, \
call_12_11_7_8, \
call_12_11_7_9, \
call_12_11_7_10, \
call_12_11_7_11, \
call_12_11_7_12, \
call_12_11_7_13, \
call_12_11_7_14, \
call_12_11_7_15, \
call_12_11_8_0, \
call_12_11_8_1, \
call_12_11_8_2, \
call_12_11_8_3, \
call_12_11_8_4, \
call_12_11_8_5, \
call_12_11_8_6, \
call_12_11_8_7, \
call_12_11_8_8, \
call_12_11_8_9, \
call_12_11_8_10, \
call_12_11_8_11, \
call_12_11_8_12, \
call_12_11_8_13, \
call_12_11_8_14, \
call_12_11_8_15, \
call_12_11_9_0, \
call_12_11_9_1, \
call_12_11_9_2, \
call_12_11_9_3, \
call_12_11_9_4, \
call_12_11_9_5, \
call_12_11_9_6, \
call_12_11_9_7, \
call_12_11_9_8, \
call_12_11_9_9, \
call_12_11_9_10, \
call_12_11_9_11, \
call_12_11_9_12, \
call_12_11_9_13, \
call_12_11_9_14, \
call_12_11_9_15, \
call_12_11_10_0, \
call_12_11_10_1, \
call_12_11_10_2, \
call_12_11_10_3, \
call_12_11_10_4, \
call_12_11_10_5, \
call_12_11_10_6, \
call_12_11_10_7, \
call_12_11_10_8, \
call_12_11_10_9, \
call_12_11_10_10, \
call_12_11_10_11, \
call_12_11_10_12, \
call_12_11_10_13, \
call_12_11_10_14, \
call_12_11_10_15, \
call_12_11_11_0, \
call_12_11_11_1, \
call_12_11_11_2, \
call_12_11_11_3, \
call_12_11_11_4, \
call_12_11_11_5, \
call_12_11_11_6, \
call_12_11_11_7, \
call_12_11_11_8, \
call_12_11_11_9, \
call_12_11_11_10, \
call_12_11_11_11, \
call_12_11_11_12, \
call_12_11_11_13, \
call_12_11_11_14, \
call_12_11_11_15, \
call_12_11_12_0, \
call_12_11_12_1, \
call_12_11_12_2, \
call_12_11_12_3, \
call_12_11_12_4, \
call_12_11_12_5, \
call_12_11_12_6, \
call_12_11_12_7, \
call_12_11_12_8, \
call_12_11_12_9, \
call_12_11_12_10, \
call_12_11_12_11, \
call_12_11_12_12, \
call_12_11_12_13, \
call_12_11_12_14, \
call_12_11_12_15, \
call_12_11_13_0, \
call_12_11_13_1, \
call_12_11_13_2, \
call_12_11_13_3, \
call_12_11_13_4, \
call_12_11_13_5, \
call_12_11_13_6, \
call_12_11_13_7, \
call_12_11_13_8, \
call_12_11_13_9, \
call_12_11_13_10, \
call_12_11_13_11, \
call_12_11_13_12, \
call_12_11_13_13, \
call_12_11_13_14, \
call_12_11_13_15, \
call_12_11_14_0, \
call_12_11_14_1, \
call_12_11_14_2, \
call_12_11_14_3, \
call_12_11_14_4, \
call_12_11_14_5, \
call_12_11_14_6, \
call_12_11_14_7, \
call_12_11_14_8, \
call_12_11_14_9, \
call_12_11_14_10, \
call_12_11_14_11, \
call_12_11_14_12, \
call_12_11_14_13, \
call_12_11_14_14, \
call_12_11_14_15, \
call_12_11_15_0, \
call_12_11_15_1, \
call_12_11_15_2, \
call_12_11_15_3, \
call_12_11_15_4, \
call_12_11_15_5, \
call_12_11_15_6, \
call_12_11_15_7, \
call_12_11_15_8, \
call_12_11_15_9, \
call_12_11_15_10, \
call_12_11_15_11, \
call_12_11_15_12, \
call_12_11_15_13, \
call_12_11_15_14, \
call_12_11_15_15, \
call_12_12_0_0, \
call_12_12_0_1, \
call_12_12_0_2, \
call_12_12_0_3, \
call_12_12_0_4, \
call_12_12_0_5, \
call_12_12_0_6, \
call_12_12_0_7, \
call_12_12_0_8, \
call_12_12_0_9, \
call_12_12_0_10, \
call_12_12_0_11, \
call_12_12_0_12, \
call_12_12_0_13, \
call_12_12_0_14, \
call_12_12_0_15, \
call_12_12_1_0, \
call_12_12_1_1, \
call_12_12_1_2, \
call_12_12_1_3, \
call_12_12_1_4, \
call_12_12_1_5, \
call_12_12_1_6, \
call_12_12_1_7, \
call_12_12_1_8, \
call_12_12_1_9, \
call_12_12_1_10, \
call_12_12_1_11, \
call_12_12_1_12, \
call_12_12_1_13, \
call_12_12_1_14, \
call_12_12_1_15, \
call_12_12_2_0, \
call_12_12_2_1, \
call_12_12_2_2, \
call_12_12_2_3, \
call_12_12_2_4, \
call_12_12_2_5, \
call_12_12_2_6, \
call_12_12_2_7, \
call_12_12_2_8, \
call_12_12_2_9, \
call_12_12_2_10, \
call_12_12_2_11, \
call_12_12_2_12, \
call_12_12_2_13, \
call_12_12_2_14, \
call_12_12_2_15, \
call_12_12_3_0, \
call_12_12_3_1, \
call_12_12_3_2, \
call_12_12_3_3, \
call_12_12_3_4, \
call_12_12_3_5, \
call_12_12_3_6, \
call_12_12_3_7, \
call_12_12_3_8, \
call_12_12_3_9, \
call_12_12_3_10, \
call_12_12_3_11, \
call_12_12_3_12, \
call_12_12_3_13, \
call_12_12_3_14, \
call_12_12_3_15, \
call_12_12_4_0, \
call_12_12_4_1, \
call_12_12_4_2, \
call_12_12_4_3, \
call_12_12_4_4, \
call_12_12_4_5, \
call_12_12_4_6, \
call_12_12_4_7, \
call_12_12_4_8, \
call_12_12_4_9, \
call_12_12_4_10, \
call_12_12_4_11, \
call_12_12_4_12, \
call_12_12_4_13, \
call_12_12_4_14, \
call_12_12_4_15, \
call_12_12_5_0, \
call_12_12_5_1, \
call_12_12_5_2, \
call_12_12_5_3, \
call_12_12_5_4, \
call_12_12_5_5, \
call_12_12_5_6, \
call_12_12_5_7, \
call_12_12_5_8, \
call_12_12_5_9, \
call_12_12_5_10, \
call_12_12_5_11, \
call_12_12_5_12, \
call_12_12_5_13, \
call_12_12_5_14, \
call_12_12_5_15, \
call_12_12_6_0, \
call_12_12_6_1, \
call_12_12_6_2, \
call_12_12_6_3, \
call_12_12_6_4, \
call_12_12_6_5, \
call_12_12_6_6, \
call_12_12_6_7, \
call_12_12_6_8, \
call_12_12_6_9, \
call_12_12_6_10, \
call_12_12_6_11, \
call_12_12_6_12, \
call_12_12_6_13, \
call_12_12_6_14, \
call_12_12_6_15, \
call_12_12_7_0, \
call_12_12_7_1, \
call_12_12_7_2, \
call_12_12_7_3, \
call_12_12_7_4, \
call_12_12_7_5, \
call_12_12_7_6, \
call_12_12_7_7, \
call_12_12_7_8, \
call_12_12_7_9, \
call_12_12_7_10, \
call_12_12_7_11, \
call_12_12_7_12, \
call_12_12_7_13, \
call_12_12_7_14, \
call_12_12_7_15, \
call_12_12_8_0, \
call_12_12_8_1, \
call_12_12_8_2, \
call_12_12_8_3, \
call_12_12_8_4, \
call_12_12_8_5, \
call_12_12_8_6, \
call_12_12_8_7, \
call_12_12_8_8, \
call_12_12_8_9, \
call_12_12_8_10, \
call_12_12_8_11, \
call_12_12_8_12, \
call_12_12_8_13, \
call_12_12_8_14, \
call_12_12_8_15, \
call_12_12_9_0, \
call_12_12_9_1, \
call_12_12_9_2, \
call_12_12_9_3, \
call_12_12_9_4, \
call_12_12_9_5, \
call_12_12_9_6, \
call_12_12_9_7, \
call_12_12_9_8, \
call_12_12_9_9, \
call_12_12_9_10, \
call_12_12_9_11, \
call_12_12_9_12, \
call_12_12_9_13, \
call_12_12_9_14, \
call_12_12_9_15, \
call_12_12_10_0, \
call_12_12_10_1, \
call_12_12_10_2, \
call_12_12_10_3, \
call_12_12_10_4, \
call_12_12_10_5, \
call_12_12_10_6, \
call_12_12_10_7, \
call_12_12_10_8, \
call_12_12_10_9, \
call_12_12_10_10, \
call_12_12_10_11, \
call_12_12_10_12, \
call_12_12_10_13, \
call_12_12_10_14, \
call_12_12_10_15, \
call_12_12_11_0, \
call_12_12_11_1, \
call_12_12_11_2, \
call_12_12_11_3, \
call_12_12_11_4, \
call_12_12_11_5, \
call_12_12_11_6, \
call_12_12_11_7, \
call_12_12_11_8, \
call_12_12_11_9, \
call_12_12_11_10, \
call_12_12_11_11, \
call_12_12_11_12, \
call_12_12_11_13, \
call_12_12_11_14, \
call_12_12_11_15, \
call_12_12_12_0, \
call_12_12_12_1, \
call_12_12_12_2, \
call_12_12_12_3, \
call_12_12_12_4, \
call_12_12_12_5, \
call_12_12_12_6, \
call_12_12_12_7, \
call_12_12_12_8, \
call_12_12_12_9, \
call_12_12_12_10, \
call_12_12_12_11, \
call_12_12_12_12, \
call_12_12_12_13, \
call_12_12_12_14, \
call_12_12_12_15, \
call_12_12_13_0, \
call_12_12_13_1, \
call_12_12_13_2, \
call_12_12_13_3, \
call_12_12_13_4, \
call_12_12_13_5, \
call_12_12_13_6, \
call_12_12_13_7, \
call_12_12_13_8, \
call_12_12_13_9, \
call_12_12_13_10, \
call_12_12_13_11, \
call_12_12_13_12, \
call_12_12_13_13, \
call_12_12_13_14, \
call_12_12_13_15, \
call_12_12_14_0, \
call_12_12_14_1, \
call_12_12_14_2, \
call_12_12_14_3, \
call_12_12_14_4, \
call_12_12_14_5, \
call_12_12_14_6, \
call_12_12_14_7, \
call_12_12_14_8, \
call_12_12_14_9, \
call_12_12_14_10, \
call_12_12_14_11, \
call_12_12_14_12, \
call_12_12_14_13, \
call_12_12_14_14, \
call_12_12_14_15, \
call_12_12_15_0, \
call_12_12_15_1, \
call_12_12_15_2, \
call_12_12_15_3, \
call_12_12_15_4, \
call_12_12_15_5, \
call_12_12_15_6, \
call_12_12_15_7, \
call_12_12_15_8, \
call_12_12_15_9, \
call_12_12_15_10, \
call_12_12_15_11, \
call_12_12_15_12, \
call_12_12_15_13, \
call_12_12_15_14, \
call_12_12_15_15, \
call_12_13_0_0, \
call_12_13_0_1, \
call_12_13_0_2, \
call_12_13_0_3, \
call_12_13_0_4, \
call_12_13_0_5, \
call_12_13_0_6, \
call_12_13_0_7, \
call_12_13_0_8, \
call_12_13_0_9, \
call_12_13_0_10, \
call_12_13_0_11, \
call_12_13_0_12, \
call_12_13_0_13, \
call_12_13_0_14, \
call_12_13_0_15, \
call_12_13_1_0, \
call_12_13_1_1, \
call_12_13_1_2, \
call_12_13_1_3, \
call_12_13_1_4, \
call_12_13_1_5, \
call_12_13_1_6, \
call_12_13_1_7, \
call_12_13_1_8, \
call_12_13_1_9, \
call_12_13_1_10, \
call_12_13_1_11, \
call_12_13_1_12, \
call_12_13_1_13, \
call_12_13_1_14, \
call_12_13_1_15, \
call_12_13_2_0, \
call_12_13_2_1, \
call_12_13_2_2, \
call_12_13_2_3, \
call_12_13_2_4, \
call_12_13_2_5, \
call_12_13_2_6, \
call_12_13_2_7, \
call_12_13_2_8, \
call_12_13_2_9, \
call_12_13_2_10, \
call_12_13_2_11, \
call_12_13_2_12, \
call_12_13_2_13, \
call_12_13_2_14, \
call_12_13_2_15, \
call_12_13_3_0, \
call_12_13_3_1, \
call_12_13_3_2, \
call_12_13_3_3, \
call_12_13_3_4, \
call_12_13_3_5, \
call_12_13_3_6, \
call_12_13_3_7, \
call_12_13_3_8, \
call_12_13_3_9, \
call_12_13_3_10, \
call_12_13_3_11, \
call_12_13_3_12, \
call_12_13_3_13, \
call_12_13_3_14, \
call_12_13_3_15, \
call_12_13_4_0, \
call_12_13_4_1, \
call_12_13_4_2, \
call_12_13_4_3, \
call_12_13_4_4, \
call_12_13_4_5, \
call_12_13_4_6, \
call_12_13_4_7, \
call_12_13_4_8, \
call_12_13_4_9, \
call_12_13_4_10, \
call_12_13_4_11, \
call_12_13_4_12, \
call_12_13_4_13, \
call_12_13_4_14, \
call_12_13_4_15, \
call_12_13_5_0, \
call_12_13_5_1, \
call_12_13_5_2, \
call_12_13_5_3, \
call_12_13_5_4, \
call_12_13_5_5, \
call_12_13_5_6, \
call_12_13_5_7, \
call_12_13_5_8, \
call_12_13_5_9, \
call_12_13_5_10, \
call_12_13_5_11, \
call_12_13_5_12, \
call_12_13_5_13, \
call_12_13_5_14, \
call_12_13_5_15, \
call_12_13_6_0, \
call_12_13_6_1, \
call_12_13_6_2, \
call_12_13_6_3, \
call_12_13_6_4, \
call_12_13_6_5, \
call_12_13_6_6, \
call_12_13_6_7, \
call_12_13_6_8, \
call_12_13_6_9, \
call_12_13_6_10, \
call_12_13_6_11, \
call_12_13_6_12, \
call_12_13_6_13, \
call_12_13_6_14, \
call_12_13_6_15, \
call_12_13_7_0, \
call_12_13_7_1, \
call_12_13_7_2, \
call_12_13_7_3, \
call_12_13_7_4, \
call_12_13_7_5, \
call_12_13_7_6, \
call_12_13_7_7, \
call_12_13_7_8, \
call_12_13_7_9, \
call_12_13_7_10, \
call_12_13_7_11, \
call_12_13_7_12, \
call_12_13_7_13, \
call_12_13_7_14, \
call_12_13_7_15, \
call_12_13_8_0, \
call_12_13_8_1, \
call_12_13_8_2, \
call_12_13_8_3, \
call_12_13_8_4, \
call_12_13_8_5, \
call_12_13_8_6, \
call_12_13_8_7, \
call_12_13_8_8, \
call_12_13_8_9, \
call_12_13_8_10, \
call_12_13_8_11, \
call_12_13_8_12, \
call_12_13_8_13, \
call_12_13_8_14, \
call_12_13_8_15, \
call_12_13_9_0, \
call_12_13_9_1, \
call_12_13_9_2, \
call_12_13_9_3, \
call_12_13_9_4, \
call_12_13_9_5, \
call_12_13_9_6, \
call_12_13_9_7, \
call_12_13_9_8, \
call_12_13_9_9, \
call_12_13_9_10, \
call_12_13_9_11, \
call_12_13_9_12, \
call_12_13_9_13, \
call_12_13_9_14, \
call_12_13_9_15, \
call_12_13_10_0, \
call_12_13_10_1, \
call_12_13_10_2, \
call_12_13_10_3, \
call_12_13_10_4, \
call_12_13_10_5, \
call_12_13_10_6, \
call_12_13_10_7, \
call_12_13_10_8, \
call_12_13_10_9, \
call_12_13_10_10, \
call_12_13_10_11, \
call_12_13_10_12, \
call_12_13_10_13, \
call_12_13_10_14, \
call_12_13_10_15, \
call_12_13_11_0, \
call_12_13_11_1, \
call_12_13_11_2, \
call_12_13_11_3, \
call_12_13_11_4, \
call_12_13_11_5, \
call_12_13_11_6, \
call_12_13_11_7, \
call_12_13_11_8, \
call_12_13_11_9, \
call_12_13_11_10, \
call_12_13_11_11, \
call_12_13_11_12, \
call_12_13_11_13, \
call_12_13_11_14, \
call_12_13_11_15, \
call_12_13_12_0, \
call_12_13_12_1, \
call_12_13_12_2, \
call_12_13_12_3, \
call_12_13_12_4, \
call_12_13_12_5, \
call_12_13_12_6, \
call_12_13_12_7, \
call_12_13_12_8, \
call_12_13_12_9, \
call_12_13_12_10, \
call_12_13_12_11, \
call_12_13_12_12, \
call_12_13_12_13, \
call_12_13_12_14, \
call_12_13_12_15, \
call_12_13_13_0, \
call_12_13_13_1, \
call_12_13_13_2, \
call_12_13_13_3, \
call_12_13_13_4, \
call_12_13_13_5, \
call_12_13_13_6, \
call_12_13_13_7, \
call_12_13_13_8, \
call_12_13_13_9, \
call_12_13_13_10, \
call_12_13_13_11, \
call_12_13_13_12, \
call_12_13_13_13, \
call_12_13_13_14, \
call_12_13_13_15, \
call_12_13_14_0, \
call_12_13_14_1, \
call_12_13_14_2, \
call_12_13_14_3, \
call_12_13_14_4, \
call_12_13_14_5, \
call_12_13_14_6, \
call_12_13_14_7, \
call_12_13_14_8, \
call_12_13_14_9, \
call_12_13_14_10, \
call_12_13_14_11, \
call_12_13_14_12, \
call_12_13_14_13, \
call_12_13_14_14, \
call_12_13_14_15, \
call_12_13_15_0, \
call_12_13_15_1, \
call_12_13_15_2, \
call_12_13_15_3, \
call_12_13_15_4, \
call_12_13_15_5, \
call_12_13_15_6, \
call_12_13_15_7, \
call_12_13_15_8, \
call_12_13_15_9, \
call_12_13_15_10, \
call_12_13_15_11, \
call_12_13_15_12, \
call_12_13_15_13, \
call_12_13_15_14, \
call_12_13_15_15, \
call_12_14_0_0, \
call_12_14_0_1, \
call_12_14_0_2, \
call_12_14_0_3, \
call_12_14_0_4, \
call_12_14_0_5, \
call_12_14_0_6, \
call_12_14_0_7, \
call_12_14_0_8, \
call_12_14_0_9, \
call_12_14_0_10, \
call_12_14_0_11, \
call_12_14_0_12, \
call_12_14_0_13, \
call_12_14_0_14, \
call_12_14_0_15, \
call_12_14_1_0, \
call_12_14_1_1, \
call_12_14_1_2, \
call_12_14_1_3, \
call_12_14_1_4, \
call_12_14_1_5, \
call_12_14_1_6, \
call_12_14_1_7, \
call_12_14_1_8, \
call_12_14_1_9, \
call_12_14_1_10, \
call_12_14_1_11, \
call_12_14_1_12, \
call_12_14_1_13, \
call_12_14_1_14, \
call_12_14_1_15, \
call_12_14_2_0, \
call_12_14_2_1, \
call_12_14_2_2, \
call_12_14_2_3, \
call_12_14_2_4, \
call_12_14_2_5, \
call_12_14_2_6, \
call_12_14_2_7, \
call_12_14_2_8, \
call_12_14_2_9, \
call_12_14_2_10, \
call_12_14_2_11, \
call_12_14_2_12, \
call_12_14_2_13, \
call_12_14_2_14, \
call_12_14_2_15, \
call_12_14_3_0, \
call_12_14_3_1, \
call_12_14_3_2, \
call_12_14_3_3, \
call_12_14_3_4, \
call_12_14_3_5, \
call_12_14_3_6, \
call_12_14_3_7, \
call_12_14_3_8, \
call_12_14_3_9, \
call_12_14_3_10, \
call_12_14_3_11, \
call_12_14_3_12, \
call_12_14_3_13, \
call_12_14_3_14, \
call_12_14_3_15, \
call_12_14_4_0, \
call_12_14_4_1, \
call_12_14_4_2, \
call_12_14_4_3, \
call_12_14_4_4, \
call_12_14_4_5, \
call_12_14_4_6, \
call_12_14_4_7, \
call_12_14_4_8, \
call_12_14_4_9, \
call_12_14_4_10, \
call_12_14_4_11, \
call_12_14_4_12, \
call_12_14_4_13, \
call_12_14_4_14, \
call_12_14_4_15, \
call_12_14_5_0, \
call_12_14_5_1, \
call_12_14_5_2, \
call_12_14_5_3, \
call_12_14_5_4, \
call_12_14_5_5, \
call_12_14_5_6, \
call_12_14_5_7, \
call_12_14_5_8, \
call_12_14_5_9, \
call_12_14_5_10, \
call_12_14_5_11, \
call_12_14_5_12, \
call_12_14_5_13, \
call_12_14_5_14, \
call_12_14_5_15, \
call_12_14_6_0, \
call_12_14_6_1, \
call_12_14_6_2, \
call_12_14_6_3, \
call_12_14_6_4, \
call_12_14_6_5, \
call_12_14_6_6, \
call_12_14_6_7, \
call_12_14_6_8, \
call_12_14_6_9, \
call_12_14_6_10, \
call_12_14_6_11, \
call_12_14_6_12, \
call_12_14_6_13, \
call_12_14_6_14, \
call_12_14_6_15, \
call_12_14_7_0, \
call_12_14_7_1, \
call_12_14_7_2, \
call_12_14_7_3, \
call_12_14_7_4, \
call_12_14_7_5, \
call_12_14_7_6, \
call_12_14_7_7, \
call_12_14_7_8, \
call_12_14_7_9, \
call_12_14_7_10, \
call_12_14_7_11, \
call_12_14_7_12, \
call_12_14_7_13, \
call_12_14_7_14, \
call_12_14_7_15, \
call_12_14_8_0, \
call_12_14_8_1, \
call_12_14_8_2, \
call_12_14_8_3, \
call_12_14_8_4, \
call_12_14_8_5, \
call_12_14_8_6, \
call_12_14_8_7, \
call_12_14_8_8, \
call_12_14_8_9, \
call_12_14_8_10, \
call_12_14_8_11, \
call_12_14_8_12, \
call_12_14_8_13, \
call_12_14_8_14, \
call_12_14_8_15, \
call_12_14_9_0, \
call_12_14_9_1, \
call_12_14_9_2, \
call_12_14_9_3, \
call_12_14_9_4, \
call_12_14_9_5, \
call_12_14_9_6, \
call_12_14_9_7, \
call_12_14_9_8, \
call_12_14_9_9, \
call_12_14_9_10, \
call_12_14_9_11, \
call_12_14_9_12, \
call_12_14_9_13, \
call_12_14_9_14, \
call_12_14_9_15, \
call_12_14_10_0, \
call_12_14_10_1, \
call_12_14_10_2, \
call_12_14_10_3, \
call_12_14_10_4, \
call_12_14_10_5, \
call_12_14_10_6, \
call_12_14_10_7, \
call_12_14_10_8, \
call_12_14_10_9, \
call_12_14_10_10, \
call_12_14_10_11, \
call_12_14_10_12, \
call_12_14_10_13, \
call_12_14_10_14, \
call_12_14_10_15, \
call_12_14_11_0, \
call_12_14_11_1, \
call_12_14_11_2, \
call_12_14_11_3, \
call_12_14_11_4, \
call_12_14_11_5, \
call_12_14_11_6, \
call_12_14_11_7, \
call_12_14_11_8, \
call_12_14_11_9, \
call_12_14_11_10, \
call_12_14_11_11, \
call_12_14_11_12, \
call_12_14_11_13, \
call_12_14_11_14, \
call_12_14_11_15, \
call_12_14_12_0, \
call_12_14_12_1, \
call_12_14_12_2, \
call_12_14_12_3, \
call_12_14_12_4, \
call_12_14_12_5, \
call_12_14_12_6, \
call_12_14_12_7, \
call_12_14_12_8, \
call_12_14_12_9, \
call_12_14_12_10, \
call_12_14_12_11, \
call_12_14_12_12, \
call_12_14_12_13, \
call_12_14_12_14, \
call_12_14_12_15, \
call_12_14_13_0, \
call_12_14_13_1, \
call_12_14_13_2, \
call_12_14_13_3, \
call_12_14_13_4, \
call_12_14_13_5, \
call_12_14_13_6, \
call_12_14_13_7, \
call_12_14_13_8, \
call_12_14_13_9, \
call_12_14_13_10, \
call_12_14_13_11, \
call_12_14_13_12, \
call_12_14_13_13, \
call_12_14_13_14, \
call_12_14_13_15, \
call_12_14_14_0, \
call_12_14_14_1, \
call_12_14_14_2, \
call_12_14_14_3, \
call_12_14_14_4, \
call_12_14_14_5, \
call_12_14_14_6, \
call_12_14_14_7, \
call_12_14_14_8, \
call_12_14_14_9, \
call_12_14_14_10, \
call_12_14_14_11, \
call_12_14_14_12, \
call_12_14_14_13, \
call_12_14_14_14, \
call_12_14_14_15, \
call_12_14_15_0, \
call_12_14_15_1, \
call_12_14_15_2, \
call_12_14_15_3, \
call_12_14_15_4, \
call_12_14_15_5, \
call_12_14_15_6, \
call_12_14_15_7, \
call_12_14_15_8, \
call_12_14_15_9, \
call_12_14_15_10, \
call_12_14_15_11, \
call_12_14_15_12, \
call_12_14_15_13, \
call_12_14_15_14, \
call_12_14_15_15, \
call_12_15_0_0, \
call_12_15_0_1, \
call_12_15_0_2, \
call_12_15_0_3, \
call_12_15_0_4, \
call_12_15_0_5, \
call_12_15_0_6, \
call_12_15_0_7, \
call_12_15_0_8, \
call_12_15_0_9, \
call_12_15_0_10, \
call_12_15_0_11, \
call_12_15_0_12, \
call_12_15_0_13, \
call_12_15_0_14, \
call_12_15_0_15, \
call_12_15_1_0, \
call_12_15_1_1, \
call_12_15_1_2, \
call_12_15_1_3, \
call_12_15_1_4, \
call_12_15_1_5, \
call_12_15_1_6, \
call_12_15_1_7, \
call_12_15_1_8, \
call_12_15_1_9, \
call_12_15_1_10, \
call_12_15_1_11, \
call_12_15_1_12, \
call_12_15_1_13, \
call_12_15_1_14, \
call_12_15_1_15, \
call_12_15_2_0, \
call_12_15_2_1, \
call_12_15_2_2, \
call_12_15_2_3, \
call_12_15_2_4, \
call_12_15_2_5, \
call_12_15_2_6, \
call_12_15_2_7, \
call_12_15_2_8, \
call_12_15_2_9, \
call_12_15_2_10, \
call_12_15_2_11, \
call_12_15_2_12, \
call_12_15_2_13, \
call_12_15_2_14, \
call_12_15_2_15, \
call_12_15_3_0, \
call_12_15_3_1, \
call_12_15_3_2, \
call_12_15_3_3, \
call_12_15_3_4, \
call_12_15_3_5, \
call_12_15_3_6, \
call_12_15_3_7, \
call_12_15_3_8, \
call_12_15_3_9, \
call_12_15_3_10, \
call_12_15_3_11, \
call_12_15_3_12, \
call_12_15_3_13, \
call_12_15_3_14, \
call_12_15_3_15, \
call_12_15_4_0, \
call_12_15_4_1, \
call_12_15_4_2, \
call_12_15_4_3, \
call_12_15_4_4, \
call_12_15_4_5, \
call_12_15_4_6, \
call_12_15_4_7, \
call_12_15_4_8, \
call_12_15_4_9, \
call_12_15_4_10, \
call_12_15_4_11, \
call_12_15_4_12, \
call_12_15_4_13, \
call_12_15_4_14, \
call_12_15_4_15, \
call_12_15_5_0, \
call_12_15_5_1, \
call_12_15_5_2, \
call_12_15_5_3, \
call_12_15_5_4, \
call_12_15_5_5, \
call_12_15_5_6, \
call_12_15_5_7, \
call_12_15_5_8, \
call_12_15_5_9, \
call_12_15_5_10, \
call_12_15_5_11, \
call_12_15_5_12, \
call_12_15_5_13, \
call_12_15_5_14, \
call_12_15_5_15, \
call_12_15_6_0, \
call_12_15_6_1, \
call_12_15_6_2, \
call_12_15_6_3, \
call_12_15_6_4, \
call_12_15_6_5, \
call_12_15_6_6, \
call_12_15_6_7, \
call_12_15_6_8, \
call_12_15_6_9, \
call_12_15_6_10, \
call_12_15_6_11, \
call_12_15_6_12, \
call_12_15_6_13, \
call_12_15_6_14, \
call_12_15_6_15, \
call_12_15_7_0, \
call_12_15_7_1, \
call_12_15_7_2, \
call_12_15_7_3, \
call_12_15_7_4, \
call_12_15_7_5, \
call_12_15_7_6, \
call_12_15_7_7, \
call_12_15_7_8, \
call_12_15_7_9, \
call_12_15_7_10, \
call_12_15_7_11, \
call_12_15_7_12, \
call_12_15_7_13, \
call_12_15_7_14, \
call_12_15_7_15, \
call_12_15_8_0, \
call_12_15_8_1, \
call_12_15_8_2, \
call_12_15_8_3, \
call_12_15_8_4, \
call_12_15_8_5, \
call_12_15_8_6, \
call_12_15_8_7, \
call_12_15_8_8, \
call_12_15_8_9, \
call_12_15_8_10, \
call_12_15_8_11, \
call_12_15_8_12, \
call_12_15_8_13, \
call_12_15_8_14, \
call_12_15_8_15, \
call_12_15_9_0, \
call_12_15_9_1, \
call_12_15_9_2, \
call_12_15_9_3, \
call_12_15_9_4, \
call_12_15_9_5, \
call_12_15_9_6, \
call_12_15_9_7, \
call_12_15_9_8, \
call_12_15_9_9, \
call_12_15_9_10, \
call_12_15_9_11, \
call_12_15_9_12, \
call_12_15_9_13, \
call_12_15_9_14, \
call_12_15_9_15, \
call_12_15_10_0, \
call_12_15_10_1, \
call_12_15_10_2, \
call_12_15_10_3, \
call_12_15_10_4, \
call_12_15_10_5, \
call_12_15_10_6, \
call_12_15_10_7, \
call_12_15_10_8, \
call_12_15_10_9, \
call_12_15_10_10, \
call_12_15_10_11, \
call_12_15_10_12, \
call_12_15_10_13, \
call_12_15_10_14, \
call_12_15_10_15, \
call_12_15_11_0, \
call_12_15_11_1, \
call_12_15_11_2, \
call_12_15_11_3, \
call_12_15_11_4, \
call_12_15_11_5, \
call_12_15_11_6, \
call_12_15_11_7, \
call_12_15_11_8, \
call_12_15_11_9, \
call_12_15_11_10, \
call_12_15_11_11, \
call_12_15_11_12, \
call_12_15_11_13, \
call_12_15_11_14, \
call_12_15_11_15, \
call_12_15_12_0, \
call_12_15_12_1, \
call_12_15_12_2, \
call_12_15_12_3, \
call_12_15_12_4, \
call_12_15_12_5, \
call_12_15_12_6, \
call_12_15_12_7, \
call_12_15_12_8, \
call_12_15_12_9, \
call_12_15_12_10, \
call_12_15_12_11, \
call_12_15_12_12, \
call_12_15_12_13, \
call_12_15_12_14, \
call_12_15_12_15, \
call_12_15_13_0, \
call_12_15_13_1, \
call_12_15_13_2, \
call_12_15_13_3, \
call_12_15_13_4, \
call_12_15_13_5, \
call_12_15_13_6, \
call_12_15_13_7, \
call_12_15_13_8, \
call_12_15_13_9, \
call_12_15_13_10, \
call_12_15_13_11, \
call_12_15_13_12, \
call_12_15_13_13, \
call_12_15_13_14, \
call_12_15_13_15, \
call_12_15_14_0, \
call_12_15_14_1, \
call_12_15_14_2, \
call_12_15_14_3, \
call_12_15_14_4, \
call_12_15_14_5, \
call_12_15_14_6, \
call_12_15_14_7, \
call_12_15_14_8, \
call_12_15_14_9, \
call_12_15_14_10, \
call_12_15_14_11, \
call_12_15_14_12, \
call_12_15_14_13, \
call_12_15_14_14, \
call_12_15_14_15, \
call_12_15_15_0, \
call_12_15_15_1, \
call_12_15_15_2, \
call_12_15_15_3, \
call_12_15_15_4, \
call_12_15_15_5, \
call_12_15_15_6, \
call_12_15_15_7, \
call_12_15_15_8, \
call_12_15_15_9, \
call_12_15_15_10, \
call_12_15_15_11, \
call_12_15_15_12, \
call_12_15_15_13, \
call_12_15_15_14, \
call_12_15_15_15, \
call_13_0_0_0, \
call_13_0_0_1, \
call_13_0_0_2, \
call_13_0_0_3, \
call_13_0_0_4, \
call_13_0_0_5, \
call_13_0_0_6, \
call_13_0_0_7, \
call_13_0_0_8, \
call_13_0_0_9, \
call_13_0_0_10, \
call_13_0_0_11, \
call_13_0_0_12, \
call_13_0_0_13, \
call_13_0_0_14, \
call_13_0_0_15, \
call_13_0_1_0, \
call_13_0_1_1, \
call_13_0_1_2, \
call_13_0_1_3, \
call_13_0_1_4, \
call_13_0_1_5, \
call_13_0_1_6, \
call_13_0_1_7, \
call_13_0_1_8, \
call_13_0_1_9, \
call_13_0_1_10, \
call_13_0_1_11, \
call_13_0_1_12, \
call_13_0_1_13, \
call_13_0_1_14, \
call_13_0_1_15, \
call_13_0_2_0, \
call_13_0_2_1, \
call_13_0_2_2, \
call_13_0_2_3, \
call_13_0_2_4, \
call_13_0_2_5, \
call_13_0_2_6, \
call_13_0_2_7, \
call_13_0_2_8, \
call_13_0_2_9, \
call_13_0_2_10, \
call_13_0_2_11, \
call_13_0_2_12, \
call_13_0_2_13, \
call_13_0_2_14, \
call_13_0_2_15, \
call_13_0_3_0, \
call_13_0_3_1, \
call_13_0_3_2, \
call_13_0_3_3, \
call_13_0_3_4, \
call_13_0_3_5, \
call_13_0_3_6, \
call_13_0_3_7, \
call_13_0_3_8, \
call_13_0_3_9, \
call_13_0_3_10, \
call_13_0_3_11, \
call_13_0_3_12, \
call_13_0_3_13, \
call_13_0_3_14, \
call_13_0_3_15, \
call_13_0_4_0, \
call_13_0_4_1, \
call_13_0_4_2, \
call_13_0_4_3, \
call_13_0_4_4, \
call_13_0_4_5, \
call_13_0_4_6, \
call_13_0_4_7, \
call_13_0_4_8, \
call_13_0_4_9, \
call_13_0_4_10, \
call_13_0_4_11, \
call_13_0_4_12, \
call_13_0_4_13, \
call_13_0_4_14, \
call_13_0_4_15, \
call_13_0_5_0, \
call_13_0_5_1, \
call_13_0_5_2, \
call_13_0_5_3, \
call_13_0_5_4, \
call_13_0_5_5, \
call_13_0_5_6, \
call_13_0_5_7, \
call_13_0_5_8, \
call_13_0_5_9, \
call_13_0_5_10, \
call_13_0_5_11, \
call_13_0_5_12, \
call_13_0_5_13, \
call_13_0_5_14, \
call_13_0_5_15, \
call_13_0_6_0, \
call_13_0_6_1, \
call_13_0_6_2, \
call_13_0_6_3, \
call_13_0_6_4, \
call_13_0_6_5, \
call_13_0_6_6, \
call_13_0_6_7, \
call_13_0_6_8, \
call_13_0_6_9, \
call_13_0_6_10, \
call_13_0_6_11, \
call_13_0_6_12, \
call_13_0_6_13, \
call_13_0_6_14, \
call_13_0_6_15, \
call_13_0_7_0, \
call_13_0_7_1, \
call_13_0_7_2, \
call_13_0_7_3, \
call_13_0_7_4, \
call_13_0_7_5, \
call_13_0_7_6, \
call_13_0_7_7, \
call_13_0_7_8, \
call_13_0_7_9, \
call_13_0_7_10, \
call_13_0_7_11, \
call_13_0_7_12, \
call_13_0_7_13, \
call_13_0_7_14, \
call_13_0_7_15, \
call_13_0_8_0, \
call_13_0_8_1, \
call_13_0_8_2, \
call_13_0_8_3, \
call_13_0_8_4, \
call_13_0_8_5, \
call_13_0_8_6, \
call_13_0_8_7, \
call_13_0_8_8, \
call_13_0_8_9, \
call_13_0_8_10, \
call_13_0_8_11, \
call_13_0_8_12, \
call_13_0_8_13, \
call_13_0_8_14, \
call_13_0_8_15, \
call_13_0_9_0, \
call_13_0_9_1, \
call_13_0_9_2, \
call_13_0_9_3, \
call_13_0_9_4, \
call_13_0_9_5, \
call_13_0_9_6, \
call_13_0_9_7, \
call_13_0_9_8, \
call_13_0_9_9, \
call_13_0_9_10, \
call_13_0_9_11, \
call_13_0_9_12, \
call_13_0_9_13, \
call_13_0_9_14, \
call_13_0_9_15, \
call_13_0_10_0, \
call_13_0_10_1, \
call_13_0_10_2, \
call_13_0_10_3, \
call_13_0_10_4, \
call_13_0_10_5, \
call_13_0_10_6, \
call_13_0_10_7, \
call_13_0_10_8, \
call_13_0_10_9, \
call_13_0_10_10, \
call_13_0_10_11, \
call_13_0_10_12, \
call_13_0_10_13, \
call_13_0_10_14, \
call_13_0_10_15, \
call_13_0_11_0, \
call_13_0_11_1, \
call_13_0_11_2, \
call_13_0_11_3, \
call_13_0_11_4, \
call_13_0_11_5, \
call_13_0_11_6, \
call_13_0_11_7, \
call_13_0_11_8, \
call_13_0_11_9, \
call_13_0_11_10, \
call_13_0_11_11, \
call_13_0_11_12, \
call_13_0_11_13, \
call_13_0_11_14, \
call_13_0_11_15, \
call_13_0_12_0, \
call_13_0_12_1, \
call_13_0_12_2, \
call_13_0_12_3, \
call_13_0_12_4, \
call_13_0_12_5, \
call_13_0_12_6, \
call_13_0_12_7, \
call_13_0_12_8, \
call_13_0_12_9, \
call_13_0_12_10, \
call_13_0_12_11, \
call_13_0_12_12, \
call_13_0_12_13, \
call_13_0_12_14, \
call_13_0_12_15, \
call_13_0_13_0, \
call_13_0_13_1, \
call_13_0_13_2, \
call_13_0_13_3, \
call_13_0_13_4, \
call_13_0_13_5, \
call_13_0_13_6, \
call_13_0_13_7, \
call_13_0_13_8, \
call_13_0_13_9, \
call_13_0_13_10, \
call_13_0_13_11, \
call_13_0_13_12, \
call_13_0_13_13, \
call_13_0_13_14, \
call_13_0_13_15, \
call_13_0_14_0, \
call_13_0_14_1, \
call_13_0_14_2, \
call_13_0_14_3, \
call_13_0_14_4, \
call_13_0_14_5, \
call_13_0_14_6, \
call_13_0_14_7, \
call_13_0_14_8, \
call_13_0_14_9, \
call_13_0_14_10, \
call_13_0_14_11, \
call_13_0_14_12, \
call_13_0_14_13, \
call_13_0_14_14, \
call_13_0_14_15, \
call_13_0_15_0, \
call_13_0_15_1, \
call_13_0_15_2, \
call_13_0_15_3, \
call_13_0_15_4, \
call_13_0_15_5, \
call_13_0_15_6, \
call_13_0_15_7, \
call_13_0_15_8, \
call_13_0_15_9, \
call_13_0_15_10, \
call_13_0_15_11, \
call_13_0_15_12, \
call_13_0_15_13, \
call_13_0_15_14, \
call_13_0_15_15, \
call_13_1_0_0, \
call_13_1_0_1, \
call_13_1_0_2, \
call_13_1_0_3, \
call_13_1_0_4, \
call_13_1_0_5, \
call_13_1_0_6, \
call_13_1_0_7, \
call_13_1_0_8, \
call_13_1_0_9, \
call_13_1_0_10, \
call_13_1_0_11, \
call_13_1_0_12, \
call_13_1_0_13, \
call_13_1_0_14, \
call_13_1_0_15, \
call_13_1_1_0, \
call_13_1_1_1, \
call_13_1_1_2, \
call_13_1_1_3, \
call_13_1_1_4, \
call_13_1_1_5, \
call_13_1_1_6, \
call_13_1_1_7, \
call_13_1_1_8, \
call_13_1_1_9, \
call_13_1_1_10, \
call_13_1_1_11, \
call_13_1_1_12, \
call_13_1_1_13, \
call_13_1_1_14, \
call_13_1_1_15, \
call_13_1_2_0, \
call_13_1_2_1, \
call_13_1_2_2, \
call_13_1_2_3, \
call_13_1_2_4, \
call_13_1_2_5, \
call_13_1_2_6, \
call_13_1_2_7, \
call_13_1_2_8, \
call_13_1_2_9, \
call_13_1_2_10, \
call_13_1_2_11, \
call_13_1_2_12, \
call_13_1_2_13, \
call_13_1_2_14, \
call_13_1_2_15, \
call_13_1_3_0, \
call_13_1_3_1, \
call_13_1_3_2, \
call_13_1_3_3, \
call_13_1_3_4, \
call_13_1_3_5, \
call_13_1_3_6, \
call_13_1_3_7, \
call_13_1_3_8, \
call_13_1_3_9, \
call_13_1_3_10, \
call_13_1_3_11, \
call_13_1_3_12, \
call_13_1_3_13, \
call_13_1_3_14, \
call_13_1_3_15, \
call_13_1_4_0, \
call_13_1_4_1, \
call_13_1_4_2, \
call_13_1_4_3, \
call_13_1_4_4, \
call_13_1_4_5, \
call_13_1_4_6, \
call_13_1_4_7, \
call_13_1_4_8, \
call_13_1_4_9, \
call_13_1_4_10, \
call_13_1_4_11, \
call_13_1_4_12, \
call_13_1_4_13, \
call_13_1_4_14, \
call_13_1_4_15, \
call_13_1_5_0, \
call_13_1_5_1, \
call_13_1_5_2, \
call_13_1_5_3, \
call_13_1_5_4, \
call_13_1_5_5, \
call_13_1_5_6, \
call_13_1_5_7, \
call_13_1_5_8, \
call_13_1_5_9, \
call_13_1_5_10, \
call_13_1_5_11, \
call_13_1_5_12, \
call_13_1_5_13, \
call_13_1_5_14, \
call_13_1_5_15, \
call_13_1_6_0, \
call_13_1_6_1, \
call_13_1_6_2, \
call_13_1_6_3, \
call_13_1_6_4, \
call_13_1_6_5, \
call_13_1_6_6, \
call_13_1_6_7, \
call_13_1_6_8, \
call_13_1_6_9, \
call_13_1_6_10, \
call_13_1_6_11, \
call_13_1_6_12, \
call_13_1_6_13, \
call_13_1_6_14, \
call_13_1_6_15, \
call_13_1_7_0, \
call_13_1_7_1, \
call_13_1_7_2, \
call_13_1_7_3, \
call_13_1_7_4, \
call_13_1_7_5, \
call_13_1_7_6, \
call_13_1_7_7, \
call_13_1_7_8, \
call_13_1_7_9, \
call_13_1_7_10, \
call_13_1_7_11, \
call_13_1_7_12, \
call_13_1_7_13, \
call_13_1_7_14, \
call_13_1_7_15, \
call_13_1_8_0, \
call_13_1_8_1, \
call_13_1_8_2, \
call_13_1_8_3, \
call_13_1_8_4, \
call_13_1_8_5, \
call_13_1_8_6, \
call_13_1_8_7, \
call_13_1_8_8, \
call_13_1_8_9, \
call_13_1_8_10, \
call_13_1_8_11, \
call_13_1_8_12, \
call_13_1_8_13, \
call_13_1_8_14, \
call_13_1_8_15, \
call_13_1_9_0, \
call_13_1_9_1, \
call_13_1_9_2, \
call_13_1_9_3, \
call_13_1_9_4, \
call_13_1_9_5, \
call_13_1_9_6, \
call_13_1_9_7, \
call_13_1_9_8, \
call_13_1_9_9, \
call_13_1_9_10, \
call_13_1_9_11, \
call_13_1_9_12, \
call_13_1_9_13, \
call_13_1_9_14, \
call_13_1_9_15, \
call_13_1_10_0, \
call_13_1_10_1, \
call_13_1_10_2, \
call_13_1_10_3, \
call_13_1_10_4, \
call_13_1_10_5, \
call_13_1_10_6, \
call_13_1_10_7, \
call_13_1_10_8, \
call_13_1_10_9, \
call_13_1_10_10, \
call_13_1_10_11, \
call_13_1_10_12, \
call_13_1_10_13, \
call_13_1_10_14, \
call_13_1_10_15, \
call_13_1_11_0, \
call_13_1_11_1, \
call_13_1_11_2, \
call_13_1_11_3, \
call_13_1_11_4, \
call_13_1_11_5, \
call_13_1_11_6, \
call_13_1_11_7, \
call_13_1_11_8, \
call_13_1_11_9, \
call_13_1_11_10, \
call_13_1_11_11, \
call_13_1_11_12, \
call_13_1_11_13, \
call_13_1_11_14, \
call_13_1_11_15, \
call_13_1_12_0, \
call_13_1_12_1, \
call_13_1_12_2, \
call_13_1_12_3, \
call_13_1_12_4, \
call_13_1_12_5, \
call_13_1_12_6, \
call_13_1_12_7, \
call_13_1_12_8, \
call_13_1_12_9, \
call_13_1_12_10, \
call_13_1_12_11, \
call_13_1_12_12, \
call_13_1_12_13, \
call_13_1_12_14, \
call_13_1_12_15, \
call_13_1_13_0, \
call_13_1_13_1, \
call_13_1_13_2, \
call_13_1_13_3, \
call_13_1_13_4, \
call_13_1_13_5, \
call_13_1_13_6, \
call_13_1_13_7, \
call_13_1_13_8, \
call_13_1_13_9, \
call_13_1_13_10, \
call_13_1_13_11, \
call_13_1_13_12, \
call_13_1_13_13, \
call_13_1_13_14, \
call_13_1_13_15, \
call_13_1_14_0, \
call_13_1_14_1, \
call_13_1_14_2, \
call_13_1_14_3, \
call_13_1_14_4, \
call_13_1_14_5, \
call_13_1_14_6, \
call_13_1_14_7, \
call_13_1_14_8, \
call_13_1_14_9, \
call_13_1_14_10, \
call_13_1_14_11, \
call_13_1_14_12, \
call_13_1_14_13, \
call_13_1_14_14, \
call_13_1_14_15, \
call_13_1_15_0, \
call_13_1_15_1, \
call_13_1_15_2, \
call_13_1_15_3, \
call_13_1_15_4, \
call_13_1_15_5, \
call_13_1_15_6, \
call_13_1_15_7, \
call_13_1_15_8, \
call_13_1_15_9, \
call_13_1_15_10, \
call_13_1_15_11, \
call_13_1_15_12, \
call_13_1_15_13, \
call_13_1_15_14, \
call_13_1_15_15, \
call_13_2_0_0, \
call_13_2_0_1, \
call_13_2_0_2, \
call_13_2_0_3, \
call_13_2_0_4, \
call_13_2_0_5, \
call_13_2_0_6, \
call_13_2_0_7, \
call_13_2_0_8, \
call_13_2_0_9, \
call_13_2_0_10, \
call_13_2_0_11, \
call_13_2_0_12, \
call_13_2_0_13, \
call_13_2_0_14, \
call_13_2_0_15, \
call_13_2_1_0, \
call_13_2_1_1, \
call_13_2_1_2, \
call_13_2_1_3, \
call_13_2_1_4, \
call_13_2_1_5, \
call_13_2_1_6, \
call_13_2_1_7, \
call_13_2_1_8, \
call_13_2_1_9, \
call_13_2_1_10, \
call_13_2_1_11, \
call_13_2_1_12, \
call_13_2_1_13, \
call_13_2_1_14, \
call_13_2_1_15, \
call_13_2_2_0, \
call_13_2_2_1, \
call_13_2_2_2, \
call_13_2_2_3, \
call_13_2_2_4, \
call_13_2_2_5, \
call_13_2_2_6, \
call_13_2_2_7, \
call_13_2_2_8, \
call_13_2_2_9, \
call_13_2_2_10, \
call_13_2_2_11, \
call_13_2_2_12, \
call_13_2_2_13, \
call_13_2_2_14, \
call_13_2_2_15, \
call_13_2_3_0, \
call_13_2_3_1, \
call_13_2_3_2, \
call_13_2_3_3, \
call_13_2_3_4, \
call_13_2_3_5, \
call_13_2_3_6, \
call_13_2_3_7, \
call_13_2_3_8, \
call_13_2_3_9, \
call_13_2_3_10, \
call_13_2_3_11, \
call_13_2_3_12, \
call_13_2_3_13, \
call_13_2_3_14, \
call_13_2_3_15, \
call_13_2_4_0, \
call_13_2_4_1, \
call_13_2_4_2, \
call_13_2_4_3, \
call_13_2_4_4, \
call_13_2_4_5, \
call_13_2_4_6, \
call_13_2_4_7, \
call_13_2_4_8, \
call_13_2_4_9, \
call_13_2_4_10, \
call_13_2_4_11, \
call_13_2_4_12, \
call_13_2_4_13, \
call_13_2_4_14, \
call_13_2_4_15, \
call_13_2_5_0, \
call_13_2_5_1, \
call_13_2_5_2, \
call_13_2_5_3, \
call_13_2_5_4, \
call_13_2_5_5, \
call_13_2_5_6, \
call_13_2_5_7, \
call_13_2_5_8, \
call_13_2_5_9, \
call_13_2_5_10, \
call_13_2_5_11, \
call_13_2_5_12, \
call_13_2_5_13, \
call_13_2_5_14, \
call_13_2_5_15, \
call_13_2_6_0, \
call_13_2_6_1, \
call_13_2_6_2, \
call_13_2_6_3, \
call_13_2_6_4, \
call_13_2_6_5, \
call_13_2_6_6, \
call_13_2_6_7, \
call_13_2_6_8, \
call_13_2_6_9, \
call_13_2_6_10, \
call_13_2_6_11, \
call_13_2_6_12, \
call_13_2_6_13, \
call_13_2_6_14, \
call_13_2_6_15, \
call_13_2_7_0, \
call_13_2_7_1, \
call_13_2_7_2, \
call_13_2_7_3, \
call_13_2_7_4, \
call_13_2_7_5, \
call_13_2_7_6, \
call_13_2_7_7, \
call_13_2_7_8, \
call_13_2_7_9, \
call_13_2_7_10, \
call_13_2_7_11, \
call_13_2_7_12, \
call_13_2_7_13, \
call_13_2_7_14, \
call_13_2_7_15, \
call_13_2_8_0, \
call_13_2_8_1, \
call_13_2_8_2, \
call_13_2_8_3, \
call_13_2_8_4, \
call_13_2_8_5, \
call_13_2_8_6, \
call_13_2_8_7, \
call_13_2_8_8, \
call_13_2_8_9, \
call_13_2_8_10, \
call_13_2_8_11, \
call_13_2_8_12, \
call_13_2_8_13, \
call_13_2_8_14, \
call_13_2_8_15, \
call_13_2_9_0, \
call_13_2_9_1, \
call_13_2_9_2, \
call_13_2_9_3, \
call_13_2_9_4, \
call_13_2_9_5, \
call_13_2_9_6, \
call_13_2_9_7, \
call_13_2_9_8, \
call_13_2_9_9, \
call_13_2_9_10, \
call_13_2_9_11, \
call_13_2_9_12, \
call_13_2_9_13, \
call_13_2_9_14, \
call_13_2_9_15, \
call_13_2_10_0, \
call_13_2_10_1, \
call_13_2_10_2, \
call_13_2_10_3, \
call_13_2_10_4, \
call_13_2_10_5, \
call_13_2_10_6, \
call_13_2_10_7, \
call_13_2_10_8, \
call_13_2_10_9, \
call_13_2_10_10, \
call_13_2_10_11, \
call_13_2_10_12, \
call_13_2_10_13, \
call_13_2_10_14, \
call_13_2_10_15, \
call_13_2_11_0, \
call_13_2_11_1, \
call_13_2_11_2, \
call_13_2_11_3, \
call_13_2_11_4, \
call_13_2_11_5, \
call_13_2_11_6, \
call_13_2_11_7, \
call_13_2_11_8, \
call_13_2_11_9, \
call_13_2_11_10, \
call_13_2_11_11, \
call_13_2_11_12, \
call_13_2_11_13, \
call_13_2_11_14, \
call_13_2_11_15, \
call_13_2_12_0, \
call_13_2_12_1, \
call_13_2_12_2, \
call_13_2_12_3, \
call_13_2_12_4, \
call_13_2_12_5, \
call_13_2_12_6, \
call_13_2_12_7, \
call_13_2_12_8, \
call_13_2_12_9, \
call_13_2_12_10, \
call_13_2_12_11, \
call_13_2_12_12, \
call_13_2_12_13, \
call_13_2_12_14, \
call_13_2_12_15, \
call_13_2_13_0, \
call_13_2_13_1, \
call_13_2_13_2, \
call_13_2_13_3, \
call_13_2_13_4, \
call_13_2_13_5, \
call_13_2_13_6, \
call_13_2_13_7, \
call_13_2_13_8, \
call_13_2_13_9, \
call_13_2_13_10, \
call_13_2_13_11, \
call_13_2_13_12, \
call_13_2_13_13, \
call_13_2_13_14, \
call_13_2_13_15, \
call_13_2_14_0, \
call_13_2_14_1, \
call_13_2_14_2, \
call_13_2_14_3, \
call_13_2_14_4, \
call_13_2_14_5, \
call_13_2_14_6, \
call_13_2_14_7, \
call_13_2_14_8, \
call_13_2_14_9, \
call_13_2_14_10, \
call_13_2_14_11, \
call_13_2_14_12, \
call_13_2_14_13, \
call_13_2_14_14, \
call_13_2_14_15, \
call_13_2_15_0, \
call_13_2_15_1, \
call_13_2_15_2, \
call_13_2_15_3, \
call_13_2_15_4, \
call_13_2_15_5, \
call_13_2_15_6, \
call_13_2_15_7, \
call_13_2_15_8, \
call_13_2_15_9, \
call_13_2_15_10, \
call_13_2_15_11, \
call_13_2_15_12, \
call_13_2_15_13, \
call_13_2_15_14, \
call_13_2_15_15, \
call_13_3_0_0, \
call_13_3_0_1, \
call_13_3_0_2, \
call_13_3_0_3, \
call_13_3_0_4, \
call_13_3_0_5, \
call_13_3_0_6, \
call_13_3_0_7, \
call_13_3_0_8, \
call_13_3_0_9, \
call_13_3_0_10, \
call_13_3_0_11, \
call_13_3_0_12, \
call_13_3_0_13, \
call_13_3_0_14, \
call_13_3_0_15, \
call_13_3_1_0, \
call_13_3_1_1, \
call_13_3_1_2, \
call_13_3_1_3, \
call_13_3_1_4, \
call_13_3_1_5, \
call_13_3_1_6, \
call_13_3_1_7, \
call_13_3_1_8, \
call_13_3_1_9, \
call_13_3_1_10, \
call_13_3_1_11, \
call_13_3_1_12, \
call_13_3_1_13, \
call_13_3_1_14, \
call_13_3_1_15, \
call_13_3_2_0, \
call_13_3_2_1, \
call_13_3_2_2, \
call_13_3_2_3, \
call_13_3_2_4, \
call_13_3_2_5, \
call_13_3_2_6, \
call_13_3_2_7, \
call_13_3_2_8, \
call_13_3_2_9, \
call_13_3_2_10, \
call_13_3_2_11, \
call_13_3_2_12, \
call_13_3_2_13, \
call_13_3_2_14, \
call_13_3_2_15, \
call_13_3_3_0, \
call_13_3_3_1, \
call_13_3_3_2, \
call_13_3_3_3, \
call_13_3_3_4, \
call_13_3_3_5, \
call_13_3_3_6, \
call_13_3_3_7, \
call_13_3_3_8, \
call_13_3_3_9, \
call_13_3_3_10, \
call_13_3_3_11, \
call_13_3_3_12, \
call_13_3_3_13, \
call_13_3_3_14, \
call_13_3_3_15, \
call_13_3_4_0, \
call_13_3_4_1, \
call_13_3_4_2, \
call_13_3_4_3, \
call_13_3_4_4, \
call_13_3_4_5, \
call_13_3_4_6, \
call_13_3_4_7, \
call_13_3_4_8, \
call_13_3_4_9, \
call_13_3_4_10, \
call_13_3_4_11, \
call_13_3_4_12, \
call_13_3_4_13, \
call_13_3_4_14, \
call_13_3_4_15, \
call_13_3_5_0, \
call_13_3_5_1, \
call_13_3_5_2, \
call_13_3_5_3, \
call_13_3_5_4, \
call_13_3_5_5, \
call_13_3_5_6, \
call_13_3_5_7, \
call_13_3_5_8, \
call_13_3_5_9, \
call_13_3_5_10, \
call_13_3_5_11, \
call_13_3_5_12, \
call_13_3_5_13, \
call_13_3_5_14, \
call_13_3_5_15, \
call_13_3_6_0, \
call_13_3_6_1, \
call_13_3_6_2, \
call_13_3_6_3, \
call_13_3_6_4, \
call_13_3_6_5, \
call_13_3_6_6, \
call_13_3_6_7, \
call_13_3_6_8, \
call_13_3_6_9, \
call_13_3_6_10, \
call_13_3_6_11, \
call_13_3_6_12, \
call_13_3_6_13, \
call_13_3_6_14, \
call_13_3_6_15, \
call_13_3_7_0, \
call_13_3_7_1, \
call_13_3_7_2, \
call_13_3_7_3, \
call_13_3_7_4, \
call_13_3_7_5, \
call_13_3_7_6, \
call_13_3_7_7, \
call_13_3_7_8, \
call_13_3_7_9, \
call_13_3_7_10, \
call_13_3_7_11, \
call_13_3_7_12, \
call_13_3_7_13, \
call_13_3_7_14, \
call_13_3_7_15, \
call_13_3_8_0, \
call_13_3_8_1, \
call_13_3_8_2, \
call_13_3_8_3, \
call_13_3_8_4, \
call_13_3_8_5, \
call_13_3_8_6, \
call_13_3_8_7, \
call_13_3_8_8, \
call_13_3_8_9, \
call_13_3_8_10, \
call_13_3_8_11, \
call_13_3_8_12, \
call_13_3_8_13, \
call_13_3_8_14, \
call_13_3_8_15, \
call_13_3_9_0, \
call_13_3_9_1, \
call_13_3_9_2, \
call_13_3_9_3, \
call_13_3_9_4, \
call_13_3_9_5, \
call_13_3_9_6, \
call_13_3_9_7, \
call_13_3_9_8, \
call_13_3_9_9, \
call_13_3_9_10, \
call_13_3_9_11, \
call_13_3_9_12, \
call_13_3_9_13, \
call_13_3_9_14, \
call_13_3_9_15, \
call_13_3_10_0, \
call_13_3_10_1, \
call_13_3_10_2, \
call_13_3_10_3, \
call_13_3_10_4, \
call_13_3_10_5, \
call_13_3_10_6, \
call_13_3_10_7, \
call_13_3_10_8, \
call_13_3_10_9, \
call_13_3_10_10, \
call_13_3_10_11, \
call_13_3_10_12, \
call_13_3_10_13, \
call_13_3_10_14, \
call_13_3_10_15, \
call_13_3_11_0, \
call_13_3_11_1, \
call_13_3_11_2, \
call_13_3_11_3, \
call_13_3_11_4, \
call_13_3_11_5, \
call_13_3_11_6, \
call_13_3_11_7, \
call_13_3_11_8, \
call_13_3_11_9, \
call_13_3_11_10, \
call_13_3_11_11, \
call_13_3_11_12, \
call_13_3_11_13, \
call_13_3_11_14, \
call_13_3_11_15, \
call_13_3_12_0, \
call_13_3_12_1, \
call_13_3_12_2, \
call_13_3_12_3, \
call_13_3_12_4, \
call_13_3_12_5, \
call_13_3_12_6, \
call_13_3_12_7, \
call_13_3_12_8, \
call_13_3_12_9, \
call_13_3_12_10, \
call_13_3_12_11, \
call_13_3_12_12, \
call_13_3_12_13, \
call_13_3_12_14, \
call_13_3_12_15, \
call_13_3_13_0, \
call_13_3_13_1, \
call_13_3_13_2, \
call_13_3_13_3, \
call_13_3_13_4, \
call_13_3_13_5, \
call_13_3_13_6, \
call_13_3_13_7, \
call_13_3_13_8, \
call_13_3_13_9, \
call_13_3_13_10, \
call_13_3_13_11, \
call_13_3_13_12, \
call_13_3_13_13, \
call_13_3_13_14, \
call_13_3_13_15, \
call_13_3_14_0, \
call_13_3_14_1, \
call_13_3_14_2, \
call_13_3_14_3, \
call_13_3_14_4, \
call_13_3_14_5, \
call_13_3_14_6, \
call_13_3_14_7, \
call_13_3_14_8, \
call_13_3_14_9, \
call_13_3_14_10, \
call_13_3_14_11, \
call_13_3_14_12, \
call_13_3_14_13, \
call_13_3_14_14, \
call_13_3_14_15, \
call_13_3_15_0, \
call_13_3_15_1, \
call_13_3_15_2, \
call_13_3_15_3, \
call_13_3_15_4, \
call_13_3_15_5, \
call_13_3_15_6, \
call_13_3_15_7, \
call_13_3_15_8, \
call_13_3_15_9, \
call_13_3_15_10, \
call_13_3_15_11, \
call_13_3_15_12, \
call_13_3_15_13, \
call_13_3_15_14, \
call_13_3_15_15, \
call_13_4_0_0, \
call_13_4_0_1, \
call_13_4_0_2, \
call_13_4_0_3, \
call_13_4_0_4, \
call_13_4_0_5, \
call_13_4_0_6, \
call_13_4_0_7, \
call_13_4_0_8, \
call_13_4_0_9, \
call_13_4_0_10, \
call_13_4_0_11, \
call_13_4_0_12, \
call_13_4_0_13, \
call_13_4_0_14, \
call_13_4_0_15, \
call_13_4_1_0, \
call_13_4_1_1, \
call_13_4_1_2, \
call_13_4_1_3, \
call_13_4_1_4, \
call_13_4_1_5, \
call_13_4_1_6, \
call_13_4_1_7, \
call_13_4_1_8, \
call_13_4_1_9, \
call_13_4_1_10, \
call_13_4_1_11, \
call_13_4_1_12, \
call_13_4_1_13, \
call_13_4_1_14, \
call_13_4_1_15, \
call_13_4_2_0, \
call_13_4_2_1, \
call_13_4_2_2, \
call_13_4_2_3, \
call_13_4_2_4, \
call_13_4_2_5, \
call_13_4_2_6, \
call_13_4_2_7, \
call_13_4_2_8, \
call_13_4_2_9, \
call_13_4_2_10, \
call_13_4_2_11, \
call_13_4_2_12, \
call_13_4_2_13, \
call_13_4_2_14, \
call_13_4_2_15, \
call_13_4_3_0, \
call_13_4_3_1, \
call_13_4_3_2, \
call_13_4_3_3, \
call_13_4_3_4, \
call_13_4_3_5, \
call_13_4_3_6, \
call_13_4_3_7, \
call_13_4_3_8, \
call_13_4_3_9, \
call_13_4_3_10, \
call_13_4_3_11, \
call_13_4_3_12, \
call_13_4_3_13, \
call_13_4_3_14, \
call_13_4_3_15, \
call_13_4_4_0, \
call_13_4_4_1, \
call_13_4_4_2, \
call_13_4_4_3, \
call_13_4_4_4, \
call_13_4_4_5, \
call_13_4_4_6, \
call_13_4_4_7, \
call_13_4_4_8, \
call_13_4_4_9, \
call_13_4_4_10, \
call_13_4_4_11, \
call_13_4_4_12, \
call_13_4_4_13, \
call_13_4_4_14, \
call_13_4_4_15, \
call_13_4_5_0, \
call_13_4_5_1, \
call_13_4_5_2, \
call_13_4_5_3, \
call_13_4_5_4, \
call_13_4_5_5, \
call_13_4_5_6, \
call_13_4_5_7, \
call_13_4_5_8, \
call_13_4_5_9, \
call_13_4_5_10, \
call_13_4_5_11, \
call_13_4_5_12, \
call_13_4_5_13, \
call_13_4_5_14, \
call_13_4_5_15, \
call_13_4_6_0, \
call_13_4_6_1, \
call_13_4_6_2, \
call_13_4_6_3, \
call_13_4_6_4, \
call_13_4_6_5, \
call_13_4_6_6, \
call_13_4_6_7, \
call_13_4_6_8, \
call_13_4_6_9, \
call_13_4_6_10, \
call_13_4_6_11, \
call_13_4_6_12, \
call_13_4_6_13, \
call_13_4_6_14, \
call_13_4_6_15, \
call_13_4_7_0, \
call_13_4_7_1, \
call_13_4_7_2, \
call_13_4_7_3, \
call_13_4_7_4, \
call_13_4_7_5, \
call_13_4_7_6, \
call_13_4_7_7, \
call_13_4_7_8, \
call_13_4_7_9, \
call_13_4_7_10, \
call_13_4_7_11, \
call_13_4_7_12, \
call_13_4_7_13, \
call_13_4_7_14, \
call_13_4_7_15, \
call_13_4_8_0, \
call_13_4_8_1, \
call_13_4_8_2, \
call_13_4_8_3, \
call_13_4_8_4, \
call_13_4_8_5, \
call_13_4_8_6, \
call_13_4_8_7, \
call_13_4_8_8, \
call_13_4_8_9, \
call_13_4_8_10, \
call_13_4_8_11, \
call_13_4_8_12, \
call_13_4_8_13, \
call_13_4_8_14, \
call_13_4_8_15, \
call_13_4_9_0, \
call_13_4_9_1, \
call_13_4_9_2, \
call_13_4_9_3, \
call_13_4_9_4, \
call_13_4_9_5, \
call_13_4_9_6, \
call_13_4_9_7, \
call_13_4_9_8, \
call_13_4_9_9, \
call_13_4_9_10, \
call_13_4_9_11, \
call_13_4_9_12, \
call_13_4_9_13, \
call_13_4_9_14, \
call_13_4_9_15, \
call_13_4_10_0, \
call_13_4_10_1, \
call_13_4_10_2, \
call_13_4_10_3, \
call_13_4_10_4, \
call_13_4_10_5, \
call_13_4_10_6, \
call_13_4_10_7, \
call_13_4_10_8, \
call_13_4_10_9, \
call_13_4_10_10, \
call_13_4_10_11, \
call_13_4_10_12, \
call_13_4_10_13, \
call_13_4_10_14, \
call_13_4_10_15, \
call_13_4_11_0, \
call_13_4_11_1, \
call_13_4_11_2, \
call_13_4_11_3, \
call_13_4_11_4, \
call_13_4_11_5, \
call_13_4_11_6, \
call_13_4_11_7, \
call_13_4_11_8, \
call_13_4_11_9, \
call_13_4_11_10, \
call_13_4_11_11, \
call_13_4_11_12, \
call_13_4_11_13, \
call_13_4_11_14, \
call_13_4_11_15, \
call_13_4_12_0, \
call_13_4_12_1, \
call_13_4_12_2, \
call_13_4_12_3, \
call_13_4_12_4, \
call_13_4_12_5, \
call_13_4_12_6, \
call_13_4_12_7, \
call_13_4_12_8, \
call_13_4_12_9, \
call_13_4_12_10, \
call_13_4_12_11, \
call_13_4_12_12, \
call_13_4_12_13, \
call_13_4_12_14, \
call_13_4_12_15, \
call_13_4_13_0, \
call_13_4_13_1, \
call_13_4_13_2, \
call_13_4_13_3, \
call_13_4_13_4, \
call_13_4_13_5, \
call_13_4_13_6, \
call_13_4_13_7, \
call_13_4_13_8, \
call_13_4_13_9, \
call_13_4_13_10, \
call_13_4_13_11, \
call_13_4_13_12, \
call_13_4_13_13, \
call_13_4_13_14, \
call_13_4_13_15, \
call_13_4_14_0, \
call_13_4_14_1, \
call_13_4_14_2, \
call_13_4_14_3, \
call_13_4_14_4, \
call_13_4_14_5, \
call_13_4_14_6, \
call_13_4_14_7, \
call_13_4_14_8, \
call_13_4_14_9, \
call_13_4_14_10, \
call_13_4_14_11, \
call_13_4_14_12, \
call_13_4_14_13, \
call_13_4_14_14, \
call_13_4_14_15, \
call_13_4_15_0, \
call_13_4_15_1, \
call_13_4_15_2, \
call_13_4_15_3, \
call_13_4_15_4, \
call_13_4_15_5, \
call_13_4_15_6, \
call_13_4_15_7, \
call_13_4_15_8, \
call_13_4_15_9, \
call_13_4_15_10, \
call_13_4_15_11, \
call_13_4_15_12, \
call_13_4_15_13, \
call_13_4_15_14, \
call_13_4_15_15, \
call_13_5_0_0, \
call_13_5_0_1, \
call_13_5_0_2, \
call_13_5_0_3, \
call_13_5_0_4, \
call_13_5_0_5, \
call_13_5_0_6, \
call_13_5_0_7, \
call_13_5_0_8, \
call_13_5_0_9, \
call_13_5_0_10, \
call_13_5_0_11, \
call_13_5_0_12, \
call_13_5_0_13, \
call_13_5_0_14, \
call_13_5_0_15, \
call_13_5_1_0, \
call_13_5_1_1, \
call_13_5_1_2, \
call_13_5_1_3, \
call_13_5_1_4, \
call_13_5_1_5, \
call_13_5_1_6, \
call_13_5_1_7, \
call_13_5_1_8, \
call_13_5_1_9, \
call_13_5_1_10, \
call_13_5_1_11, \
call_13_5_1_12, \
call_13_5_1_13, \
call_13_5_1_14, \
call_13_5_1_15, \
call_13_5_2_0, \
call_13_5_2_1, \
call_13_5_2_2, \
call_13_5_2_3, \
call_13_5_2_4, \
call_13_5_2_5, \
call_13_5_2_6, \
call_13_5_2_7, \
call_13_5_2_8, \
call_13_5_2_9, \
call_13_5_2_10, \
call_13_5_2_11, \
call_13_5_2_12, \
call_13_5_2_13, \
call_13_5_2_14, \
call_13_5_2_15, \
call_13_5_3_0, \
call_13_5_3_1, \
call_13_5_3_2, \
call_13_5_3_3, \
call_13_5_3_4, \
call_13_5_3_5, \
call_13_5_3_6, \
call_13_5_3_7, \
call_13_5_3_8, \
call_13_5_3_9, \
call_13_5_3_10, \
call_13_5_3_11, \
call_13_5_3_12, \
call_13_5_3_13, \
call_13_5_3_14, \
call_13_5_3_15, \
call_13_5_4_0, \
call_13_5_4_1, \
call_13_5_4_2, \
call_13_5_4_3, \
call_13_5_4_4, \
call_13_5_4_5, \
call_13_5_4_6, \
call_13_5_4_7, \
call_13_5_4_8, \
call_13_5_4_9, \
call_13_5_4_10, \
call_13_5_4_11, \
call_13_5_4_12, \
call_13_5_4_13, \
call_13_5_4_14, \
call_13_5_4_15, \
call_13_5_5_0, \
call_13_5_5_1, \
call_13_5_5_2, \
call_13_5_5_3, \
call_13_5_5_4, \
call_13_5_5_5, \
call_13_5_5_6, \
call_13_5_5_7, \
call_13_5_5_8, \
call_13_5_5_9, \
call_13_5_5_10, \
call_13_5_5_11, \
call_13_5_5_12, \
call_13_5_5_13, \
call_13_5_5_14, \
call_13_5_5_15, \
call_13_5_6_0, \
call_13_5_6_1, \
call_13_5_6_2, \
call_13_5_6_3, \
call_13_5_6_4, \
call_13_5_6_5, \
call_13_5_6_6, \
call_13_5_6_7, \
call_13_5_6_8, \
call_13_5_6_9, \
call_13_5_6_10, \
call_13_5_6_11, \
call_13_5_6_12, \
call_13_5_6_13, \
call_13_5_6_14, \
call_13_5_6_15, \
call_13_5_7_0, \
call_13_5_7_1, \
call_13_5_7_2, \
call_13_5_7_3, \
call_13_5_7_4, \
call_13_5_7_5, \
call_13_5_7_6, \
call_13_5_7_7, \
call_13_5_7_8, \
call_13_5_7_9, \
call_13_5_7_10, \
call_13_5_7_11, \
call_13_5_7_12, \
call_13_5_7_13, \
call_13_5_7_14, \
call_13_5_7_15, \
call_13_5_8_0, \
call_13_5_8_1, \
call_13_5_8_2, \
call_13_5_8_3, \
call_13_5_8_4, \
call_13_5_8_5, \
call_13_5_8_6, \
call_13_5_8_7, \
call_13_5_8_8, \
call_13_5_8_9, \
call_13_5_8_10, \
call_13_5_8_11, \
call_13_5_8_12, \
call_13_5_8_13, \
call_13_5_8_14, \
call_13_5_8_15, \
call_13_5_9_0, \
call_13_5_9_1, \
call_13_5_9_2, \
call_13_5_9_3, \
call_13_5_9_4, \
call_13_5_9_5, \
call_13_5_9_6, \
call_13_5_9_7, \
call_13_5_9_8, \
call_13_5_9_9, \
call_13_5_9_10, \
call_13_5_9_11, \
call_13_5_9_12, \
call_13_5_9_13, \
call_13_5_9_14, \
call_13_5_9_15, \
call_13_5_10_0, \
call_13_5_10_1, \
call_13_5_10_2, \
call_13_5_10_3, \
call_13_5_10_4, \
call_13_5_10_5, \
call_13_5_10_6, \
call_13_5_10_7, \
call_13_5_10_8, \
call_13_5_10_9, \
call_13_5_10_10, \
call_13_5_10_11, \
call_13_5_10_12, \
call_13_5_10_13, \
call_13_5_10_14, \
call_13_5_10_15, \
call_13_5_11_0, \
call_13_5_11_1, \
call_13_5_11_2, \
call_13_5_11_3, \
call_13_5_11_4, \
call_13_5_11_5, \
call_13_5_11_6, \
call_13_5_11_7, \
call_13_5_11_8, \
call_13_5_11_9, \
call_13_5_11_10, \
call_13_5_11_11, \
call_13_5_11_12, \
call_13_5_11_13, \
call_13_5_11_14, \
call_13_5_11_15, \
call_13_5_12_0, \
call_13_5_12_1, \
call_13_5_12_2, \
call_13_5_12_3, \
call_13_5_12_4, \
call_13_5_12_5, \
call_13_5_12_6, \
call_13_5_12_7, \
call_13_5_12_8, \
call_13_5_12_9, \
call_13_5_12_10, \
call_13_5_12_11, \
call_13_5_12_12, \
call_13_5_12_13, \
call_13_5_12_14, \
call_13_5_12_15, \
call_13_5_13_0, \
call_13_5_13_1, \
call_13_5_13_2, \
call_13_5_13_3, \
call_13_5_13_4, \
call_13_5_13_5, \
call_13_5_13_6, \
call_13_5_13_7, \
call_13_5_13_8, \
call_13_5_13_9, \
call_13_5_13_10, \
call_13_5_13_11, \
call_13_5_13_12, \
call_13_5_13_13, \
call_13_5_13_14, \
call_13_5_13_15, \
call_13_5_14_0, \
call_13_5_14_1, \
call_13_5_14_2, \
call_13_5_14_3, \
call_13_5_14_4, \
call_13_5_14_5, \
call_13_5_14_6, \
call_13_5_14_7, \
call_13_5_14_8, \
call_13_5_14_9, \
call_13_5_14_10, \
call_13_5_14_11, \
call_13_5_14_12, \
call_13_5_14_13, \
call_13_5_14_14, \
call_13_5_14_15, \
call_13_5_15_0, \
call_13_5_15_1, \
call_13_5_15_2, \
call_13_5_15_3, \
call_13_5_15_4, \
call_13_5_15_5, \
call_13_5_15_6, \
call_13_5_15_7, \
call_13_5_15_8, \
call_13_5_15_9, \
call_13_5_15_10, \
call_13_5_15_11, \
call_13_5_15_12, \
call_13_5_15_13, \
call_13_5_15_14, \
call_13_5_15_15, \
call_13_6_0_0, \
call_13_6_0_1, \
call_13_6_0_2, \
call_13_6_0_3, \
call_13_6_0_4, \
call_13_6_0_5, \
call_13_6_0_6, \
call_13_6_0_7, \
call_13_6_0_8, \
call_13_6_0_9, \
call_13_6_0_10, \
call_13_6_0_11, \
call_13_6_0_12, \
call_13_6_0_13, \
call_13_6_0_14, \
call_13_6_0_15, \
call_13_6_1_0, \
call_13_6_1_1, \
call_13_6_1_2, \
call_13_6_1_3, \
call_13_6_1_4, \
call_13_6_1_5, \
call_13_6_1_6, \
call_13_6_1_7, \
call_13_6_1_8, \
call_13_6_1_9, \
call_13_6_1_10, \
call_13_6_1_11, \
call_13_6_1_12, \
call_13_6_1_13, \
call_13_6_1_14, \
call_13_6_1_15, \
call_13_6_2_0, \
call_13_6_2_1, \
call_13_6_2_2, \
call_13_6_2_3, \
call_13_6_2_4, \
call_13_6_2_5, \
call_13_6_2_6, \
call_13_6_2_7, \
call_13_6_2_8, \
call_13_6_2_9, \
call_13_6_2_10, \
call_13_6_2_11, \
call_13_6_2_12, \
call_13_6_2_13, \
call_13_6_2_14, \
call_13_6_2_15, \
call_13_6_3_0, \
call_13_6_3_1, \
call_13_6_3_2, \
call_13_6_3_3, \
call_13_6_3_4, \
call_13_6_3_5, \
call_13_6_3_6, \
call_13_6_3_7, \
call_13_6_3_8, \
call_13_6_3_9, \
call_13_6_3_10, \
call_13_6_3_11, \
call_13_6_3_12, \
call_13_6_3_13, \
call_13_6_3_14, \
call_13_6_3_15, \
call_13_6_4_0, \
call_13_6_4_1, \
call_13_6_4_2, \
call_13_6_4_3, \
call_13_6_4_4, \
call_13_6_4_5, \
call_13_6_4_6, \
call_13_6_4_7, \
call_13_6_4_8, \
call_13_6_4_9, \
call_13_6_4_10, \
call_13_6_4_11, \
call_13_6_4_12, \
call_13_6_4_13, \
call_13_6_4_14, \
call_13_6_4_15, \
call_13_6_5_0, \
call_13_6_5_1, \
call_13_6_5_2, \
call_13_6_5_3, \
call_13_6_5_4, \
call_13_6_5_5, \
call_13_6_5_6, \
call_13_6_5_7, \
call_13_6_5_8, \
call_13_6_5_9, \
call_13_6_5_10, \
call_13_6_5_11, \
call_13_6_5_12, \
call_13_6_5_13, \
call_13_6_5_14, \
call_13_6_5_15, \
call_13_6_6_0, \
call_13_6_6_1, \
call_13_6_6_2, \
call_13_6_6_3, \
call_13_6_6_4, \
call_13_6_6_5, \
call_13_6_6_6, \
call_13_6_6_7, \
call_13_6_6_8, \
call_13_6_6_9, \
call_13_6_6_10, \
call_13_6_6_11, \
call_13_6_6_12, \
call_13_6_6_13, \
call_13_6_6_14, \
call_13_6_6_15, \
call_13_6_7_0, \
call_13_6_7_1, \
call_13_6_7_2, \
call_13_6_7_3, \
call_13_6_7_4, \
call_13_6_7_5, \
call_13_6_7_6, \
call_13_6_7_7, \
call_13_6_7_8, \
call_13_6_7_9, \
call_13_6_7_10, \
call_13_6_7_11, \
call_13_6_7_12, \
call_13_6_7_13, \
call_13_6_7_14, \
call_13_6_7_15, \
call_13_6_8_0, \
call_13_6_8_1, \
call_13_6_8_2, \
call_13_6_8_3, \
call_13_6_8_4, \
call_13_6_8_5, \
call_13_6_8_6, \
call_13_6_8_7, \
call_13_6_8_8, \
call_13_6_8_9, \
call_13_6_8_10, \
call_13_6_8_11, \
call_13_6_8_12, \
call_13_6_8_13, \
call_13_6_8_14, \
call_13_6_8_15, \
call_13_6_9_0, \
call_13_6_9_1, \
call_13_6_9_2, \
call_13_6_9_3, \
call_13_6_9_4, \
call_13_6_9_5, \
call_13_6_9_6, \
call_13_6_9_7, \
call_13_6_9_8, \
call_13_6_9_9, \
call_13_6_9_10, \
call_13_6_9_11, \
call_13_6_9_12, \
call_13_6_9_13, \
call_13_6_9_14, \
call_13_6_9_15, \
call_13_6_10_0, \
call_13_6_10_1, \
call_13_6_10_2, \
call_13_6_10_3, \
call_13_6_10_4, \
call_13_6_10_5, \
call_13_6_10_6, \
call_13_6_10_7, \
call_13_6_10_8, \
call_13_6_10_9, \
call_13_6_10_10, \
call_13_6_10_11, \
call_13_6_10_12, \
call_13_6_10_13, \
call_13_6_10_14, \
call_13_6_10_15, \
call_13_6_11_0, \
call_13_6_11_1, \
call_13_6_11_2, \
call_13_6_11_3, \
call_13_6_11_4, \
call_13_6_11_5, \
call_13_6_11_6, \
call_13_6_11_7, \
call_13_6_11_8, \
call_13_6_11_9, \
call_13_6_11_10, \
call_13_6_11_11, \
call_13_6_11_12, \
call_13_6_11_13, \
call_13_6_11_14, \
call_13_6_11_15, \
call_13_6_12_0, \
call_13_6_12_1, \
call_13_6_12_2, \
call_13_6_12_3, \
call_13_6_12_4, \
call_13_6_12_5, \
call_13_6_12_6, \
call_13_6_12_7, \
call_13_6_12_8, \
call_13_6_12_9, \
call_13_6_12_10, \
call_13_6_12_11, \
call_13_6_12_12, \
call_13_6_12_13, \
call_13_6_12_14, \
call_13_6_12_15, \
call_13_6_13_0, \
call_13_6_13_1, \
call_13_6_13_2, \
call_13_6_13_3, \
call_13_6_13_4, \
call_13_6_13_5, \
call_13_6_13_6, \
call_13_6_13_7, \
call_13_6_13_8, \
call_13_6_13_9, \
call_13_6_13_10, \
call_13_6_13_11, \
call_13_6_13_12, \
call_13_6_13_13, \
call_13_6_13_14, \
call_13_6_13_15, \
call_13_6_14_0, \
call_13_6_14_1, \
call_13_6_14_2, \
call_13_6_14_3, \
call_13_6_14_4, \
call_13_6_14_5, \
call_13_6_14_6, \
call_13_6_14_7, \
call_13_6_14_8, \
call_13_6_14_9, \
call_13_6_14_10, \
call_13_6_14_11, \
call_13_6_14_12, \
call_13_6_14_13, \
call_13_6_14_14, \
call_13_6_14_15, \
call_13_6_15_0, \
call_13_6_15_1, \
call_13_6_15_2, \
call_13_6_15_3, \
call_13_6_15_4, \
call_13_6_15_5, \
call_13_6_15_6, \
call_13_6_15_7, \
call_13_6_15_8, \
call_13_6_15_9, \
call_13_6_15_10, \
call_13_6_15_11, \
call_13_6_15_12, \
call_13_6_15_13, \
call_13_6_15_14, \
call_13_6_15_15, \
call_13_7_0_0, \
call_13_7_0_1, \
call_13_7_0_2, \
call_13_7_0_3, \
call_13_7_0_4, \
call_13_7_0_5, \
call_13_7_0_6, \
call_13_7_0_7, \
call_13_7_0_8, \
call_13_7_0_9, \
call_13_7_0_10, \
call_13_7_0_11, \
call_13_7_0_12, \
call_13_7_0_13, \
call_13_7_0_14, \
call_13_7_0_15, \
call_13_7_1_0, \
call_13_7_1_1, \
call_13_7_1_2, \
call_13_7_1_3, \
call_13_7_1_4, \
call_13_7_1_5, \
call_13_7_1_6, \
call_13_7_1_7, \
call_13_7_1_8, \
call_13_7_1_9, \
call_13_7_1_10, \
call_13_7_1_11, \
call_13_7_1_12, \
call_13_7_1_13, \
call_13_7_1_14, \
call_13_7_1_15, \
call_13_7_2_0, \
call_13_7_2_1, \
call_13_7_2_2, \
call_13_7_2_3, \
call_13_7_2_4, \
call_13_7_2_5, \
call_13_7_2_6, \
call_13_7_2_7, \
call_13_7_2_8, \
call_13_7_2_9, \
call_13_7_2_10, \
call_13_7_2_11, \
call_13_7_2_12, \
call_13_7_2_13, \
call_13_7_2_14, \
call_13_7_2_15, \
call_13_7_3_0, \
call_13_7_3_1, \
call_13_7_3_2, \
call_13_7_3_3, \
call_13_7_3_4, \
call_13_7_3_5, \
call_13_7_3_6, \
call_13_7_3_7, \
call_13_7_3_8, \
call_13_7_3_9, \
call_13_7_3_10, \
call_13_7_3_11, \
call_13_7_3_12, \
call_13_7_3_13, \
call_13_7_3_14, \
call_13_7_3_15, \
call_13_7_4_0, \
call_13_7_4_1, \
call_13_7_4_2, \
call_13_7_4_3, \
call_13_7_4_4, \
call_13_7_4_5, \
call_13_7_4_6, \
call_13_7_4_7, \
call_13_7_4_8, \
call_13_7_4_9, \
call_13_7_4_10, \
call_13_7_4_11, \
call_13_7_4_12, \
call_13_7_4_13, \
call_13_7_4_14, \
call_13_7_4_15, \
call_13_7_5_0, \
call_13_7_5_1, \
call_13_7_5_2, \
call_13_7_5_3, \
call_13_7_5_4, \
call_13_7_5_5, \
call_13_7_5_6, \
call_13_7_5_7, \
call_13_7_5_8, \
call_13_7_5_9, \
call_13_7_5_10, \
call_13_7_5_11, \
call_13_7_5_12, \
call_13_7_5_13, \
call_13_7_5_14, \
call_13_7_5_15, \
call_13_7_6_0, \
call_13_7_6_1, \
call_13_7_6_2, \
call_13_7_6_3, \
call_13_7_6_4, \
call_13_7_6_5, \
call_13_7_6_6, \
call_13_7_6_7, \
call_13_7_6_8, \
call_13_7_6_9, \
call_13_7_6_10, \
call_13_7_6_11, \
call_13_7_6_12, \
call_13_7_6_13, \
call_13_7_6_14, \
call_13_7_6_15, \
call_13_7_7_0, \
call_13_7_7_1, \
call_13_7_7_2, \
call_13_7_7_3, \
call_13_7_7_4, \
call_13_7_7_5, \
call_13_7_7_6, \
call_13_7_7_7, \
call_13_7_7_8, \
call_13_7_7_9, \
call_13_7_7_10, \
call_13_7_7_11, \
call_13_7_7_12, \
call_13_7_7_13, \
call_13_7_7_14, \
call_13_7_7_15, \
call_13_7_8_0, \
call_13_7_8_1, \
call_13_7_8_2, \
call_13_7_8_3, \
call_13_7_8_4, \
call_13_7_8_5, \
call_13_7_8_6, \
call_13_7_8_7, \
call_13_7_8_8, \
call_13_7_8_9, \
call_13_7_8_10, \
call_13_7_8_11, \
call_13_7_8_12, \
call_13_7_8_13, \
call_13_7_8_14, \
call_13_7_8_15, \
call_13_7_9_0, \
call_13_7_9_1, \
call_13_7_9_2, \
call_13_7_9_3, \
call_13_7_9_4, \
call_13_7_9_5, \
call_13_7_9_6, \
call_13_7_9_7, \
call_13_7_9_8, \
call_13_7_9_9, \
call_13_7_9_10, \
call_13_7_9_11, \
call_13_7_9_12, \
call_13_7_9_13, \
call_13_7_9_14, \
call_13_7_9_15, \
call_13_7_10_0, \
call_13_7_10_1, \
call_13_7_10_2, \
call_13_7_10_3, \
call_13_7_10_4, \
call_13_7_10_5, \
call_13_7_10_6, \
call_13_7_10_7, \
call_13_7_10_8, \
call_13_7_10_9, \
call_13_7_10_10, \
call_13_7_10_11, \
call_13_7_10_12, \
call_13_7_10_13, \
call_13_7_10_14, \
call_13_7_10_15, \
call_13_7_11_0, \
call_13_7_11_1, \
call_13_7_11_2, \
call_13_7_11_3, \
call_13_7_11_4, \
call_13_7_11_5, \
call_13_7_11_6, \
call_13_7_11_7, \
call_13_7_11_8, \
call_13_7_11_9, \
call_13_7_11_10, \
call_13_7_11_11, \
call_13_7_11_12, \
call_13_7_11_13, \
call_13_7_11_14, \
call_13_7_11_15, \
call_13_7_12_0, \
call_13_7_12_1, \
call_13_7_12_2, \
call_13_7_12_3, \
call_13_7_12_4, \
call_13_7_12_5, \
call_13_7_12_6, \
call_13_7_12_7, \
call_13_7_12_8, \
call_13_7_12_9, \
call_13_7_12_10, \
call_13_7_12_11, \
call_13_7_12_12, \
call_13_7_12_13, \
call_13_7_12_14, \
call_13_7_12_15, \
call_13_7_13_0, \
call_13_7_13_1, \
call_13_7_13_2, \
call_13_7_13_3, \
call_13_7_13_4, \
call_13_7_13_5, \
call_13_7_13_6, \
call_13_7_13_7, \
call_13_7_13_8, \
call_13_7_13_9, \
call_13_7_13_10, \
call_13_7_13_11, \
call_13_7_13_12, \
call_13_7_13_13, \
call_13_7_13_14, \
call_13_7_13_15, \
call_13_7_14_0, \
call_13_7_14_1, \
call_13_7_14_2, \
call_13_7_14_3, \
call_13_7_14_4, \
call_13_7_14_5, \
call_13_7_14_6, \
call_13_7_14_7, \
call_13_7_14_8, \
call_13_7_14_9, \
call_13_7_14_10, \
call_13_7_14_11, \
call_13_7_14_12, \
call_13_7_14_13, \
call_13_7_14_14, \
call_13_7_14_15, \
call_13_7_15_0, \
call_13_7_15_1, \
call_13_7_15_2, \
call_13_7_15_3, \
call_13_7_15_4, \
call_13_7_15_5, \
call_13_7_15_6, \
call_13_7_15_7, \
call_13_7_15_8, \
call_13_7_15_9, \
call_13_7_15_10, \
call_13_7_15_11, \
call_13_7_15_12, \
call_13_7_15_13, \
call_13_7_15_14, \
call_13_7_15_15, \
call_13_8_0_0, \
call_13_8_0_1, \
call_13_8_0_2, \
call_13_8_0_3, \
call_13_8_0_4, \
call_13_8_0_5, \
call_13_8_0_6, \
call_13_8_0_7, \
call_13_8_0_8, \
call_13_8_0_9, \
call_13_8_0_10, \
call_13_8_0_11, \
call_13_8_0_12, \
call_13_8_0_13, \
call_13_8_0_14, \
call_13_8_0_15, \
call_13_8_1_0, \
call_13_8_1_1, \
call_13_8_1_2, \
call_13_8_1_3, \
call_13_8_1_4, \
call_13_8_1_5, \
call_13_8_1_6, \
call_13_8_1_7, \
call_13_8_1_8, \
call_13_8_1_9, \
call_13_8_1_10, \
call_13_8_1_11, \
call_13_8_1_12, \
call_13_8_1_13, \
call_13_8_1_14, \
call_13_8_1_15, \
call_13_8_2_0, \
call_13_8_2_1, \
call_13_8_2_2, \
call_13_8_2_3, \
call_13_8_2_4, \
call_13_8_2_5, \
call_13_8_2_6, \
call_13_8_2_7, \
call_13_8_2_8, \
call_13_8_2_9, \
call_13_8_2_10, \
call_13_8_2_11, \
call_13_8_2_12, \
call_13_8_2_13, \
call_13_8_2_14, \
call_13_8_2_15, \
call_13_8_3_0, \
call_13_8_3_1, \
call_13_8_3_2, \
call_13_8_3_3, \
call_13_8_3_4, \
call_13_8_3_5, \
call_13_8_3_6, \
call_13_8_3_7, \
call_13_8_3_8, \
call_13_8_3_9, \
call_13_8_3_10, \
call_13_8_3_11, \
call_13_8_3_12, \
call_13_8_3_13, \
call_13_8_3_14, \
call_13_8_3_15, \
call_13_8_4_0, \
call_13_8_4_1, \
call_13_8_4_2, \
call_13_8_4_3, \
call_13_8_4_4, \
call_13_8_4_5, \
call_13_8_4_6, \
call_13_8_4_7, \
call_13_8_4_8, \
call_13_8_4_9, \
call_13_8_4_10, \
call_13_8_4_11, \
call_13_8_4_12, \
call_13_8_4_13, \
call_13_8_4_14, \
call_13_8_4_15, \
call_13_8_5_0, \
call_13_8_5_1, \
call_13_8_5_2, \
call_13_8_5_3, \
call_13_8_5_4, \
call_13_8_5_5, \
call_13_8_5_6, \
call_13_8_5_7, \
call_13_8_5_8, \
call_13_8_5_9, \
call_13_8_5_10, \
call_13_8_5_11, \
call_13_8_5_12, \
call_13_8_5_13, \
call_13_8_5_14, \
call_13_8_5_15, \
call_13_8_6_0, \
call_13_8_6_1, \
call_13_8_6_2, \
call_13_8_6_3, \
call_13_8_6_4, \
call_13_8_6_5, \
call_13_8_6_6, \
call_13_8_6_7, \
call_13_8_6_8, \
call_13_8_6_9, \
call_13_8_6_10, \
call_13_8_6_11, \
call_13_8_6_12, \
call_13_8_6_13, \
call_13_8_6_14, \
call_13_8_6_15, \
call_13_8_7_0, \
call_13_8_7_1, \
call_13_8_7_2, \
call_13_8_7_3, \
call_13_8_7_4, \
call_13_8_7_5, \
call_13_8_7_6, \
call_13_8_7_7, \
call_13_8_7_8, \
call_13_8_7_9, \
call_13_8_7_10, \
call_13_8_7_11, \
call_13_8_7_12, \
call_13_8_7_13, \
call_13_8_7_14, \
call_13_8_7_15, \
call_13_8_8_0, \
call_13_8_8_1, \
call_13_8_8_2, \
call_13_8_8_3, \
call_13_8_8_4, \
call_13_8_8_5, \
call_13_8_8_6, \
call_13_8_8_7, \
call_13_8_8_8, \
call_13_8_8_9, \
call_13_8_8_10, \
call_13_8_8_11, \
call_13_8_8_12, \
call_13_8_8_13, \
call_13_8_8_14, \
call_13_8_8_15, \
call_13_8_9_0, \
call_13_8_9_1, \
call_13_8_9_2, \
call_13_8_9_3, \
call_13_8_9_4, \
call_13_8_9_5, \
call_13_8_9_6, \
call_13_8_9_7, \
call_13_8_9_8, \
call_13_8_9_9, \
call_13_8_9_10, \
call_13_8_9_11, \
call_13_8_9_12, \
call_13_8_9_13, \
call_13_8_9_14, \
call_13_8_9_15, \
call_13_8_10_0, \
call_13_8_10_1, \
call_13_8_10_2, \
call_13_8_10_3, \
call_13_8_10_4, \
call_13_8_10_5, \
call_13_8_10_6, \
call_13_8_10_7, \
call_13_8_10_8, \
call_13_8_10_9, \
call_13_8_10_10, \
call_13_8_10_11, \
call_13_8_10_12, \
call_13_8_10_13, \
call_13_8_10_14, \
call_13_8_10_15, \
call_13_8_11_0, \
call_13_8_11_1, \
call_13_8_11_2, \
call_13_8_11_3, \
call_13_8_11_4, \
call_13_8_11_5, \
call_13_8_11_6, \
call_13_8_11_7, \
call_13_8_11_8, \
call_13_8_11_9, \
call_13_8_11_10, \
call_13_8_11_11, \
call_13_8_11_12, \
call_13_8_11_13, \
call_13_8_11_14, \
call_13_8_11_15, \
call_13_8_12_0, \
call_13_8_12_1, \
call_13_8_12_2, \
call_13_8_12_3, \
call_13_8_12_4, \
call_13_8_12_5, \
call_13_8_12_6, \
call_13_8_12_7, \
call_13_8_12_8, \
call_13_8_12_9, \
call_13_8_12_10, \
call_13_8_12_11, \
call_13_8_12_12, \
call_13_8_12_13, \
call_13_8_12_14, \
call_13_8_12_15, \
call_13_8_13_0, \
call_13_8_13_1, \
call_13_8_13_2, \
call_13_8_13_3, \
call_13_8_13_4, \
call_13_8_13_5, \
call_13_8_13_6, \
call_13_8_13_7, \
call_13_8_13_8, \
call_13_8_13_9, \
call_13_8_13_10, \
call_13_8_13_11, \
call_13_8_13_12, \
call_13_8_13_13, \
call_13_8_13_14, \
call_13_8_13_15, \
call_13_8_14_0, \
call_13_8_14_1, \
call_13_8_14_2, \
call_13_8_14_3, \
call_13_8_14_4, \
call_13_8_14_5, \
call_13_8_14_6, \
call_13_8_14_7, \
call_13_8_14_8, \
call_13_8_14_9, \
call_13_8_14_10, \
call_13_8_14_11, \
call_13_8_14_12, \
call_13_8_14_13, \
call_13_8_14_14, \
call_13_8_14_15, \
call_13_8_15_0, \
call_13_8_15_1, \
call_13_8_15_2, \
call_13_8_15_3, \
call_13_8_15_4, \
call_13_8_15_5, \
call_13_8_15_6, \
call_13_8_15_7, \
call_13_8_15_8, \
call_13_8_15_9, \
call_13_8_15_10, \
call_13_8_15_11, \
call_13_8_15_12, \
call_13_8_15_13, \
call_13_8_15_14, \
call_13_8_15_15, \
call_13_9_0_0, \
call_13_9_0_1, \
call_13_9_0_2, \
call_13_9_0_3, \
call_13_9_0_4, \
call_13_9_0_5, \
call_13_9_0_6, \
call_13_9_0_7, \
call_13_9_0_8, \
call_13_9_0_9, \
call_13_9_0_10, \
call_13_9_0_11, \
call_13_9_0_12, \
call_13_9_0_13, \
call_13_9_0_14, \
call_13_9_0_15, \
call_13_9_1_0, \
call_13_9_1_1, \
call_13_9_1_2, \
call_13_9_1_3, \
call_13_9_1_4, \
call_13_9_1_5, \
call_13_9_1_6, \
call_13_9_1_7, \
call_13_9_1_8, \
call_13_9_1_9, \
call_13_9_1_10, \
call_13_9_1_11, \
call_13_9_1_12, \
call_13_9_1_13, \
call_13_9_1_14, \
call_13_9_1_15, \
call_13_9_2_0, \
call_13_9_2_1, \
call_13_9_2_2, \
call_13_9_2_3, \
call_13_9_2_4, \
call_13_9_2_5, \
call_13_9_2_6, \
call_13_9_2_7, \
call_13_9_2_8, \
call_13_9_2_9, \
call_13_9_2_10, \
call_13_9_2_11, \
call_13_9_2_12, \
call_13_9_2_13, \
call_13_9_2_14, \
call_13_9_2_15, \
call_13_9_3_0, \
call_13_9_3_1, \
call_13_9_3_2, \
call_13_9_3_3, \
call_13_9_3_4, \
call_13_9_3_5, \
call_13_9_3_6, \
call_13_9_3_7, \
call_13_9_3_8, \
call_13_9_3_9, \
call_13_9_3_10, \
call_13_9_3_11, \
call_13_9_3_12, \
call_13_9_3_13, \
call_13_9_3_14, \
call_13_9_3_15, \
call_13_9_4_0, \
call_13_9_4_1, \
call_13_9_4_2, \
call_13_9_4_3, \
call_13_9_4_4, \
call_13_9_4_5, \
call_13_9_4_6, \
call_13_9_4_7, \
call_13_9_4_8, \
call_13_9_4_9, \
call_13_9_4_10, \
call_13_9_4_11, \
call_13_9_4_12, \
call_13_9_4_13, \
call_13_9_4_14, \
call_13_9_4_15, \
call_13_9_5_0, \
call_13_9_5_1, \
call_13_9_5_2, \
call_13_9_5_3, \
call_13_9_5_4, \
call_13_9_5_5, \
call_13_9_5_6, \
call_13_9_5_7, \
call_13_9_5_8, \
call_13_9_5_9, \
call_13_9_5_10, \
call_13_9_5_11, \
call_13_9_5_12, \
call_13_9_5_13, \
call_13_9_5_14, \
call_13_9_5_15, \
call_13_9_6_0, \
call_13_9_6_1, \
call_13_9_6_2, \
call_13_9_6_3, \
call_13_9_6_4, \
call_13_9_6_5, \
call_13_9_6_6, \
call_13_9_6_7, \
call_13_9_6_8, \
call_13_9_6_9, \
call_13_9_6_10, \
call_13_9_6_11, \
call_13_9_6_12, \
call_13_9_6_13, \
call_13_9_6_14, \
call_13_9_6_15, \
call_13_9_7_0, \
call_13_9_7_1, \
call_13_9_7_2, \
call_13_9_7_3, \
call_13_9_7_4, \
call_13_9_7_5, \
call_13_9_7_6, \
call_13_9_7_7, \
call_13_9_7_8, \
call_13_9_7_9, \
call_13_9_7_10, \
call_13_9_7_11, \
call_13_9_7_12, \
call_13_9_7_13, \
call_13_9_7_14, \
call_13_9_7_15, \
call_13_9_8_0, \
call_13_9_8_1, \
call_13_9_8_2, \
call_13_9_8_3, \
call_13_9_8_4, \
call_13_9_8_5, \
call_13_9_8_6, \
call_13_9_8_7, \
call_13_9_8_8, \
call_13_9_8_9, \
call_13_9_8_10, \
call_13_9_8_11, \
call_13_9_8_12, \
call_13_9_8_13, \
call_13_9_8_14, \
call_13_9_8_15, \
call_13_9_9_0, \
call_13_9_9_1, \
call_13_9_9_2, \
call_13_9_9_3, \
call_13_9_9_4, \
call_13_9_9_5, \
call_13_9_9_6, \
call_13_9_9_7, \
call_13_9_9_8, \
call_13_9_9_9, \
call_13_9_9_10, \
call_13_9_9_11, \
call_13_9_9_12, \
call_13_9_9_13, \
call_13_9_9_14, \
call_13_9_9_15, \
call_13_9_10_0, \
call_13_9_10_1, \
call_13_9_10_2, \
call_13_9_10_3, \
call_13_9_10_4, \
call_13_9_10_5, \
call_13_9_10_6, \
call_13_9_10_7, \
call_13_9_10_8, \
call_13_9_10_9, \
call_13_9_10_10, \
call_13_9_10_11, \
call_13_9_10_12, \
call_13_9_10_13, \
call_13_9_10_14, \
call_13_9_10_15, \
call_13_9_11_0, \
call_13_9_11_1, \
call_13_9_11_2, \
call_13_9_11_3, \
call_13_9_11_4, \
call_13_9_11_5, \
call_13_9_11_6, \
call_13_9_11_7, \
call_13_9_11_8, \
call_13_9_11_9, \
call_13_9_11_10, \
call_13_9_11_11, \
call_13_9_11_12, \
call_13_9_11_13, \
call_13_9_11_14, \
call_13_9_11_15, \
call_13_9_12_0, \
call_13_9_12_1, \
call_13_9_12_2, \
call_13_9_12_3, \
call_13_9_12_4, \
call_13_9_12_5, \
call_13_9_12_6, \
call_13_9_12_7, \
call_13_9_12_8, \
call_13_9_12_9, \
call_13_9_12_10, \
call_13_9_12_11, \
call_13_9_12_12, \
call_13_9_12_13, \
call_13_9_12_14, \
call_13_9_12_15, \
call_13_9_13_0, \
call_13_9_13_1, \
call_13_9_13_2, \
call_13_9_13_3, \
call_13_9_13_4, \
call_13_9_13_5, \
call_13_9_13_6, \
call_13_9_13_7, \
call_13_9_13_8, \
call_13_9_13_9, \
call_13_9_13_10, \
call_13_9_13_11, \
call_13_9_13_12, \
call_13_9_13_13, \
call_13_9_13_14, \
call_13_9_13_15, \
call_13_9_14_0, \
call_13_9_14_1, \
call_13_9_14_2, \
call_13_9_14_3, \
call_13_9_14_4, \
call_13_9_14_5, \
call_13_9_14_6, \
call_13_9_14_7, \
call_13_9_14_8, \
call_13_9_14_9, \
call_13_9_14_10, \
call_13_9_14_11, \
call_13_9_14_12, \
call_13_9_14_13, \
call_13_9_14_14, \
call_13_9_14_15, \
call_13_9_15_0, \
call_13_9_15_1, \
call_13_9_15_2, \
call_13_9_15_3, \
call_13_9_15_4, \
call_13_9_15_5, \
call_13_9_15_6, \
call_13_9_15_7, \
call_13_9_15_8, \
call_13_9_15_9, \
call_13_9_15_10, \
call_13_9_15_11, \
call_13_9_15_12, \
call_13_9_15_13, \
call_13_9_15_14, \
call_13_9_15_15, \
call_13_10_0_0, \
call_13_10_0_1, \
call_13_10_0_2, \
call_13_10_0_3, \
call_13_10_0_4, \
call_13_10_0_5, \
call_13_10_0_6, \
call_13_10_0_7, \
call_13_10_0_8, \
call_13_10_0_9, \
call_13_10_0_10, \
call_13_10_0_11, \
call_13_10_0_12, \
call_13_10_0_13, \
call_13_10_0_14, \
call_13_10_0_15, \
call_13_10_1_0, \
call_13_10_1_1, \
call_13_10_1_2, \
call_13_10_1_3, \
call_13_10_1_4, \
call_13_10_1_5, \
call_13_10_1_6, \
call_13_10_1_7, \
call_13_10_1_8, \
call_13_10_1_9, \
call_13_10_1_10, \
call_13_10_1_11, \
call_13_10_1_12, \
call_13_10_1_13, \
call_13_10_1_14, \
call_13_10_1_15, \
call_13_10_2_0, \
call_13_10_2_1, \
call_13_10_2_2, \
call_13_10_2_3, \
call_13_10_2_4, \
call_13_10_2_5, \
call_13_10_2_6, \
call_13_10_2_7, \
call_13_10_2_8, \
call_13_10_2_9, \
call_13_10_2_10, \
call_13_10_2_11, \
call_13_10_2_12, \
call_13_10_2_13, \
call_13_10_2_14, \
call_13_10_2_15, \
call_13_10_3_0, \
call_13_10_3_1, \
call_13_10_3_2, \
call_13_10_3_3, \
call_13_10_3_4, \
call_13_10_3_5, \
call_13_10_3_6, \
call_13_10_3_7, \
call_13_10_3_8, \
call_13_10_3_9, \
call_13_10_3_10, \
call_13_10_3_11, \
call_13_10_3_12, \
call_13_10_3_13, \
call_13_10_3_14, \
call_13_10_3_15, \
call_13_10_4_0, \
call_13_10_4_1, \
call_13_10_4_2, \
call_13_10_4_3, \
call_13_10_4_4, \
call_13_10_4_5, \
call_13_10_4_6, \
call_13_10_4_7, \
call_13_10_4_8, \
call_13_10_4_9, \
call_13_10_4_10, \
call_13_10_4_11, \
call_13_10_4_12, \
call_13_10_4_13, \
call_13_10_4_14, \
call_13_10_4_15, \
call_13_10_5_0, \
call_13_10_5_1, \
call_13_10_5_2, \
call_13_10_5_3, \
call_13_10_5_4, \
call_13_10_5_5, \
call_13_10_5_6, \
call_13_10_5_7, \
call_13_10_5_8, \
call_13_10_5_9, \
call_13_10_5_10, \
call_13_10_5_11, \
call_13_10_5_12, \
call_13_10_5_13, \
call_13_10_5_14, \
call_13_10_5_15, \
call_13_10_6_0, \
call_13_10_6_1, \
call_13_10_6_2, \
call_13_10_6_3, \
call_13_10_6_4, \
call_13_10_6_5, \
call_13_10_6_6, \
call_13_10_6_7, \
call_13_10_6_8, \
call_13_10_6_9, \
call_13_10_6_10, \
call_13_10_6_11, \
call_13_10_6_12, \
call_13_10_6_13, \
call_13_10_6_14, \
call_13_10_6_15, \
call_13_10_7_0, \
call_13_10_7_1, \
call_13_10_7_2, \
call_13_10_7_3, \
call_13_10_7_4, \
call_13_10_7_5, \
call_13_10_7_6, \
call_13_10_7_7, \
call_13_10_7_8, \
call_13_10_7_9, \
call_13_10_7_10, \
call_13_10_7_11, \
call_13_10_7_12, \
call_13_10_7_13, \
call_13_10_7_14, \
call_13_10_7_15, \
call_13_10_8_0, \
call_13_10_8_1, \
call_13_10_8_2, \
call_13_10_8_3, \
call_13_10_8_4, \
call_13_10_8_5, \
call_13_10_8_6, \
call_13_10_8_7, \
call_13_10_8_8, \
call_13_10_8_9, \
call_13_10_8_10, \
call_13_10_8_11, \
call_13_10_8_12, \
call_13_10_8_13, \
call_13_10_8_14, \
call_13_10_8_15, \
call_13_10_9_0, \
call_13_10_9_1, \
call_13_10_9_2, \
call_13_10_9_3, \
call_13_10_9_4, \
call_13_10_9_5, \
call_13_10_9_6, \
call_13_10_9_7, \
call_13_10_9_8, \
call_13_10_9_9, \
call_13_10_9_10, \
call_13_10_9_11, \
call_13_10_9_12, \
call_13_10_9_13, \
call_13_10_9_14, \
call_13_10_9_15, \
call_13_10_10_0, \
call_13_10_10_1, \
call_13_10_10_2, \
call_13_10_10_3, \
call_13_10_10_4, \
call_13_10_10_5, \
call_13_10_10_6, \
call_13_10_10_7, \
call_13_10_10_8, \
call_13_10_10_9, \
call_13_10_10_10, \
call_13_10_10_11, \
call_13_10_10_12, \
call_13_10_10_13, \
call_13_10_10_14, \
call_13_10_10_15, \
call_13_10_11_0, \
call_13_10_11_1, \
call_13_10_11_2, \
call_13_10_11_3, \
call_13_10_11_4, \
call_13_10_11_5, \
call_13_10_11_6, \
call_13_10_11_7, \
call_13_10_11_8, \
call_13_10_11_9, \
call_13_10_11_10, \
call_13_10_11_11, \
call_13_10_11_12, \
call_13_10_11_13, \
call_13_10_11_14, \
call_13_10_11_15, \
call_13_10_12_0, \
call_13_10_12_1, \
call_13_10_12_2, \
call_13_10_12_3, \
call_13_10_12_4, \
call_13_10_12_5, \
call_13_10_12_6, \
call_13_10_12_7, \
call_13_10_12_8, \
call_13_10_12_9, \
call_13_10_12_10, \
call_13_10_12_11, \
call_13_10_12_12, \
call_13_10_12_13, \
call_13_10_12_14, \
call_13_10_12_15, \
call_13_10_13_0, \
call_13_10_13_1, \
call_13_10_13_2, \
call_13_10_13_3, \
call_13_10_13_4, \
call_13_10_13_5, \
call_13_10_13_6, \
call_13_10_13_7, \
call_13_10_13_8, \
call_13_10_13_9, \
call_13_10_13_10, \
call_13_10_13_11, \
call_13_10_13_12, \
call_13_10_13_13, \
call_13_10_13_14, \
call_13_10_13_15, \
call_13_10_14_0, \
call_13_10_14_1, \
call_13_10_14_2, \
call_13_10_14_3, \
call_13_10_14_4, \
call_13_10_14_5, \
call_13_10_14_6, \
call_13_10_14_7, \
call_13_10_14_8, \
call_13_10_14_9, \
call_13_10_14_10, \
call_13_10_14_11, \
call_13_10_14_12, \
call_13_10_14_13, \
call_13_10_14_14, \
call_13_10_14_15, \
call_13_10_15_0, \
call_13_10_15_1, \
call_13_10_15_2, \
call_13_10_15_3, \
call_13_10_15_4, \
call_13_10_15_5, \
call_13_10_15_6, \
call_13_10_15_7, \
call_13_10_15_8, \
call_13_10_15_9, \
call_13_10_15_10, \
call_13_10_15_11, \
call_13_10_15_12, \
call_13_10_15_13, \
call_13_10_15_14, \
call_13_10_15_15, \
call_13_11_0_0, \
call_13_11_0_1, \
call_13_11_0_2, \
call_13_11_0_3, \
call_13_11_0_4, \
call_13_11_0_5, \
call_13_11_0_6, \
call_13_11_0_7, \
call_13_11_0_8, \
call_13_11_0_9, \
call_13_11_0_10, \
call_13_11_0_11, \
call_13_11_0_12, \
call_13_11_0_13, \
call_13_11_0_14, \
call_13_11_0_15, \
call_13_11_1_0, \
call_13_11_1_1, \
call_13_11_1_2, \
call_13_11_1_3, \
call_13_11_1_4, \
call_13_11_1_5, \
call_13_11_1_6, \
call_13_11_1_7, \
call_13_11_1_8, \
call_13_11_1_9, \
call_13_11_1_10, \
call_13_11_1_11, \
call_13_11_1_12, \
call_13_11_1_13, \
call_13_11_1_14, \
call_13_11_1_15, \
call_13_11_2_0, \
call_13_11_2_1, \
call_13_11_2_2, \
call_13_11_2_3, \
call_13_11_2_4, \
call_13_11_2_5, \
call_13_11_2_6, \
call_13_11_2_7, \
call_13_11_2_8, \
call_13_11_2_9, \
call_13_11_2_10, \
call_13_11_2_11, \
call_13_11_2_12, \
call_13_11_2_13, \
call_13_11_2_14, \
call_13_11_2_15, \
call_13_11_3_0, \
call_13_11_3_1, \
call_13_11_3_2, \
call_13_11_3_3, \
call_13_11_3_4, \
call_13_11_3_5, \
call_13_11_3_6, \
call_13_11_3_7, \
call_13_11_3_8, \
call_13_11_3_9, \
call_13_11_3_10, \
call_13_11_3_11, \
call_13_11_3_12, \
call_13_11_3_13, \
call_13_11_3_14, \
call_13_11_3_15, \
call_13_11_4_0, \
call_13_11_4_1, \
call_13_11_4_2, \
call_13_11_4_3, \
call_13_11_4_4, \
call_13_11_4_5, \
call_13_11_4_6, \
call_13_11_4_7, \
call_13_11_4_8, \
call_13_11_4_9, \
call_13_11_4_10, \
call_13_11_4_11, \
call_13_11_4_12, \
call_13_11_4_13, \
call_13_11_4_14, \
call_13_11_4_15, \
call_13_11_5_0, \
call_13_11_5_1, \
call_13_11_5_2, \
call_13_11_5_3, \
call_13_11_5_4, \
call_13_11_5_5, \
call_13_11_5_6, \
call_13_11_5_7, \
call_13_11_5_8, \
call_13_11_5_9, \
call_13_11_5_10, \
call_13_11_5_11, \
call_13_11_5_12, \
call_13_11_5_13, \
call_13_11_5_14, \
call_13_11_5_15, \
call_13_11_6_0, \
call_13_11_6_1, \
call_13_11_6_2, \
call_13_11_6_3, \
call_13_11_6_4, \
call_13_11_6_5, \
call_13_11_6_6, \
call_13_11_6_7, \
call_13_11_6_8, \
call_13_11_6_9, \
call_13_11_6_10, \
call_13_11_6_11, \
call_13_11_6_12, \
call_13_11_6_13, \
call_13_11_6_14, \
call_13_11_6_15, \
call_13_11_7_0, \
call_13_11_7_1, \
call_13_11_7_2, \
call_13_11_7_3, \
call_13_11_7_4, \
call_13_11_7_5, \
call_13_11_7_6, \
call_13_11_7_7, \
call_13_11_7_8, \
call_13_11_7_9, \
call_13_11_7_10, \
call_13_11_7_11, \
call_13_11_7_12, \
call_13_11_7_13, \
call_13_11_7_14, \
call_13_11_7_15, \
call_13_11_8_0, \
call_13_11_8_1, \
call_13_11_8_2, \
call_13_11_8_3, \
call_13_11_8_4, \
call_13_11_8_5, \
call_13_11_8_6, \
call_13_11_8_7, \
call_13_11_8_8, \
call_13_11_8_9, \
call_13_11_8_10, \
call_13_11_8_11, \
call_13_11_8_12, \
call_13_11_8_13, \
call_13_11_8_14, \
call_13_11_8_15, \
call_13_11_9_0, \
call_13_11_9_1, \
call_13_11_9_2, \
call_13_11_9_3, \
call_13_11_9_4, \
call_13_11_9_5, \
call_13_11_9_6, \
call_13_11_9_7, \
call_13_11_9_8, \
call_13_11_9_9, \
call_13_11_9_10, \
call_13_11_9_11, \
call_13_11_9_12, \
call_13_11_9_13, \
call_13_11_9_14, \
call_13_11_9_15, \
call_13_11_10_0, \
call_13_11_10_1, \
call_13_11_10_2, \
call_13_11_10_3, \
call_13_11_10_4, \
call_13_11_10_5, \
call_13_11_10_6, \
call_13_11_10_7, \
call_13_11_10_8, \
call_13_11_10_9, \
call_13_11_10_10, \
call_13_11_10_11, \
call_13_11_10_12, \
call_13_11_10_13, \
call_13_11_10_14, \
call_13_11_10_15, \
call_13_11_11_0, \
call_13_11_11_1, \
call_13_11_11_2, \
call_13_11_11_3, \
call_13_11_11_4, \
call_13_11_11_5, \
call_13_11_11_6, \
call_13_11_11_7, \
call_13_11_11_8, \
call_13_11_11_9, \
call_13_11_11_10, \
call_13_11_11_11, \
call_13_11_11_12, \
call_13_11_11_13, \
call_13_11_11_14, \
call_13_11_11_15, \
call_13_11_12_0, \
call_13_11_12_1, \
call_13_11_12_2, \
call_13_11_12_3, \
call_13_11_12_4, \
call_13_11_12_5, \
call_13_11_12_6, \
call_13_11_12_7, \
call_13_11_12_8, \
call_13_11_12_9, \
call_13_11_12_10, \
call_13_11_12_11, \
call_13_11_12_12, \
call_13_11_12_13, \
call_13_11_12_14, \
call_13_11_12_15, \
call_13_11_13_0, \
call_13_11_13_1, \
call_13_11_13_2, \
call_13_11_13_3, \
call_13_11_13_4, \
call_13_11_13_5, \
call_13_11_13_6, \
call_13_11_13_7, \
call_13_11_13_8, \
call_13_11_13_9, \
call_13_11_13_10, \
call_13_11_13_11, \
call_13_11_13_12, \
call_13_11_13_13, \
call_13_11_13_14, \
call_13_11_13_15, \
call_13_11_14_0, \
call_13_11_14_1, \
call_13_11_14_2, \
call_13_11_14_3, \
call_13_11_14_4, \
call_13_11_14_5, \
call_13_11_14_6, \
call_13_11_14_7, \
call_13_11_14_8, \
call_13_11_14_9, \
call_13_11_14_10, \
call_13_11_14_11, \
call_13_11_14_12, \
call_13_11_14_13, \
call_13_11_14_14, \
call_13_11_14_15, \
call_13_11_15_0, \
call_13_11_15_1, \
call_13_11_15_2, \
call_13_11_15_3, \
call_13_11_15_4, \
call_13_11_15_5, \
call_13_11_15_6, \
call_13_11_15_7, \
call_13_11_15_8, \
call_13_11_15_9, \
call_13_11_15_10, \
call_13_11_15_11, \
call_13_11_15_12, \
call_13_11_15_13, \
call_13_11_15_14, \
call_13_11_15_15, \
call_13_12_0_0, \
call_13_12_0_1, \
call_13_12_0_2, \
call_13_12_0_3, \
call_13_12_0_4, \
call_13_12_0_5, \
call_13_12_0_6, \
call_13_12_0_7, \
call_13_12_0_8, \
call_13_12_0_9, \
call_13_12_0_10, \
call_13_12_0_11, \
call_13_12_0_12, \
call_13_12_0_13, \
call_13_12_0_14, \
call_13_12_0_15, \
call_13_12_1_0, \
call_13_12_1_1, \
call_13_12_1_2, \
call_13_12_1_3, \
call_13_12_1_4, \
call_13_12_1_5, \
call_13_12_1_6, \
call_13_12_1_7, \
call_13_12_1_8, \
call_13_12_1_9, \
call_13_12_1_10, \
call_13_12_1_11, \
call_13_12_1_12, \
call_13_12_1_13, \
call_13_12_1_14, \
call_13_12_1_15, \
call_13_12_2_0, \
call_13_12_2_1, \
call_13_12_2_2, \
call_13_12_2_3, \
call_13_12_2_4, \
call_13_12_2_5, \
call_13_12_2_6, \
call_13_12_2_7, \
call_13_12_2_8, \
call_13_12_2_9, \
call_13_12_2_10, \
call_13_12_2_11, \
call_13_12_2_12, \
call_13_12_2_13, \
call_13_12_2_14, \
call_13_12_2_15, \
call_13_12_3_0, \
call_13_12_3_1, \
call_13_12_3_2, \
call_13_12_3_3, \
call_13_12_3_4, \
call_13_12_3_5, \
call_13_12_3_6, \
call_13_12_3_7, \
call_13_12_3_8, \
call_13_12_3_9, \
call_13_12_3_10, \
call_13_12_3_11, \
call_13_12_3_12, \
call_13_12_3_13, \
call_13_12_3_14, \
call_13_12_3_15, \
call_13_12_4_0, \
call_13_12_4_1, \
call_13_12_4_2, \
call_13_12_4_3, \
call_13_12_4_4, \
call_13_12_4_5, \
call_13_12_4_6, \
call_13_12_4_7, \
call_13_12_4_8, \
call_13_12_4_9, \
call_13_12_4_10, \
call_13_12_4_11, \
call_13_12_4_12, \
call_13_12_4_13, \
call_13_12_4_14, \
call_13_12_4_15, \
call_13_12_5_0, \
call_13_12_5_1, \
call_13_12_5_2, \
call_13_12_5_3, \
call_13_12_5_4, \
call_13_12_5_5, \
call_13_12_5_6, \
call_13_12_5_7, \
call_13_12_5_8, \
call_13_12_5_9, \
call_13_12_5_10, \
call_13_12_5_11, \
call_13_12_5_12, \
call_13_12_5_13, \
call_13_12_5_14, \
call_13_12_5_15, \
call_13_12_6_0, \
call_13_12_6_1, \
call_13_12_6_2, \
call_13_12_6_3, \
call_13_12_6_4, \
call_13_12_6_5, \
call_13_12_6_6, \
call_13_12_6_7, \
call_13_12_6_8, \
call_13_12_6_9, \
call_13_12_6_10, \
call_13_12_6_11, \
call_13_12_6_12, \
call_13_12_6_13, \
call_13_12_6_14, \
call_13_12_6_15, \
call_13_12_7_0, \
call_13_12_7_1, \
call_13_12_7_2, \
call_13_12_7_3, \
call_13_12_7_4, \
call_13_12_7_5, \
call_13_12_7_6, \
call_13_12_7_7, \
call_13_12_7_8, \
call_13_12_7_9, \
call_13_12_7_10, \
call_13_12_7_11, \
call_13_12_7_12, \
call_13_12_7_13, \
call_13_12_7_14, \
call_13_12_7_15, \
call_13_12_8_0, \
call_13_12_8_1, \
call_13_12_8_2, \
call_13_12_8_3, \
call_13_12_8_4, \
call_13_12_8_5, \
call_13_12_8_6, \
call_13_12_8_7, \
call_13_12_8_8, \
call_13_12_8_9, \
call_13_12_8_10, \
call_13_12_8_11, \
call_13_12_8_12, \
call_13_12_8_13, \
call_13_12_8_14, \
call_13_12_8_15, \
call_13_12_9_0, \
call_13_12_9_1, \
call_13_12_9_2, \
call_13_12_9_3, \
call_13_12_9_4, \
call_13_12_9_5, \
call_13_12_9_6, \
call_13_12_9_7, \
call_13_12_9_8, \
call_13_12_9_9, \
call_13_12_9_10, \
call_13_12_9_11, \
call_13_12_9_12, \
call_13_12_9_13, \
call_13_12_9_14, \
call_13_12_9_15, \
call_13_12_10_0, \
call_13_12_10_1, \
call_13_12_10_2, \
call_13_12_10_3, \
call_13_12_10_4, \
call_13_12_10_5, \
call_13_12_10_6, \
call_13_12_10_7, \
call_13_12_10_8, \
call_13_12_10_9, \
call_13_12_10_10, \
call_13_12_10_11, \
call_13_12_10_12, \
call_13_12_10_13, \
call_13_12_10_14, \
call_13_12_10_15, \
call_13_12_11_0, \
call_13_12_11_1, \
call_13_12_11_2, \
call_13_12_11_3, \
call_13_12_11_4, \
call_13_12_11_5, \
call_13_12_11_6, \
call_13_12_11_7, \
call_13_12_11_8, \
call_13_12_11_9, \
call_13_12_11_10, \
call_13_12_11_11, \
call_13_12_11_12, \
call_13_12_11_13, \
call_13_12_11_14, \
call_13_12_11_15, \
call_13_12_12_0, \
call_13_12_12_1, \
call_13_12_12_2, \
call_13_12_12_3, \
call_13_12_12_4, \
call_13_12_12_5, \
call_13_12_12_6, \
call_13_12_12_7, \
call_13_12_12_8, \
call_13_12_12_9, \
call_13_12_12_10, \
call_13_12_12_11, \
call_13_12_12_12, \
call_13_12_12_13, \
call_13_12_12_14, \
call_13_12_12_15, \
call_13_12_13_0, \
call_13_12_13_1, \
call_13_12_13_2, \
call_13_12_13_3, \
call_13_12_13_4, \
call_13_12_13_5, \
call_13_12_13_6, \
call_13_12_13_7, \
call_13_12_13_8, \
call_13_12_13_9, \
call_13_12_13_10, \
call_13_12_13_11, \
call_13_12_13_12, \
call_13_12_13_13, \
call_13_12_13_14, \
call_13_12_13_15, \
call_13_12_14_0, \
call_13_12_14_1, \
call_13_12_14_2, \
call_13_12_14_3, \
call_13_12_14_4, \
call_13_12_14_5, \
call_13_12_14_6, \
call_13_12_14_7, \
call_13_12_14_8, \
call_13_12_14_9, \
call_13_12_14_10, \
call_13_12_14_11, \
call_13_12_14_12, \
call_13_12_14_13, \
call_13_12_14_14, \
call_13_12_14_15, \
call_13_12_15_0, \
call_13_12_15_1, \
call_13_12_15_2, \
call_13_12_15_3, \
call_13_12_15_4, \
call_13_12_15_5, \
call_13_12_15_6, \
call_13_12_15_7, \
call_13_12_15_8, \
call_13_12_15_9, \
call_13_12_15_10, \
call_13_12_15_11, \
call_13_12_15_12, \
call_13_12_15_13, \
call_13_12_15_14, \
call_13_12_15_15, \
call_13_13_0_0, \
call_13_13_0_1, \
call_13_13_0_2, \
call_13_13_0_3, \
call_13_13_0_4, \
call_13_13_0_5, \
call_13_13_0_6, \
call_13_13_0_7, \
call_13_13_0_8, \
call_13_13_0_9, \
call_13_13_0_10, \
call_13_13_0_11, \
call_13_13_0_12, \
call_13_13_0_13, \
call_13_13_0_14, \
call_13_13_0_15, \
call_13_13_1_0, \
call_13_13_1_1, \
call_13_13_1_2, \
call_13_13_1_3, \
call_13_13_1_4, \
call_13_13_1_5, \
call_13_13_1_6, \
call_13_13_1_7, \
call_13_13_1_8, \
call_13_13_1_9, \
call_13_13_1_10, \
call_13_13_1_11, \
call_13_13_1_12, \
call_13_13_1_13, \
call_13_13_1_14, \
call_13_13_1_15, \
call_13_13_2_0, \
call_13_13_2_1, \
call_13_13_2_2, \
call_13_13_2_3, \
call_13_13_2_4, \
call_13_13_2_5, \
call_13_13_2_6, \
call_13_13_2_7, \
call_13_13_2_8, \
call_13_13_2_9, \
call_13_13_2_10, \
call_13_13_2_11, \
call_13_13_2_12, \
call_13_13_2_13, \
call_13_13_2_14, \
call_13_13_2_15, \
call_13_13_3_0, \
call_13_13_3_1, \
call_13_13_3_2, \
call_13_13_3_3, \
call_13_13_3_4, \
call_13_13_3_5, \
call_13_13_3_6, \
call_13_13_3_7, \
call_13_13_3_8, \
call_13_13_3_9, \
call_13_13_3_10, \
call_13_13_3_11, \
call_13_13_3_12, \
call_13_13_3_13, \
call_13_13_3_14, \
call_13_13_3_15, \
call_13_13_4_0, \
call_13_13_4_1, \
call_13_13_4_2, \
call_13_13_4_3, \
call_13_13_4_4, \
call_13_13_4_5, \
call_13_13_4_6, \
call_13_13_4_7, \
call_13_13_4_8, \
call_13_13_4_9, \
call_13_13_4_10, \
call_13_13_4_11, \
call_13_13_4_12, \
call_13_13_4_13, \
call_13_13_4_14, \
call_13_13_4_15, \
call_13_13_5_0, \
call_13_13_5_1, \
call_13_13_5_2, \
call_13_13_5_3, \
call_13_13_5_4, \
call_13_13_5_5, \
call_13_13_5_6, \
call_13_13_5_7, \
call_13_13_5_8, \
call_13_13_5_9, \
call_13_13_5_10, \
call_13_13_5_11, \
call_13_13_5_12, \
call_13_13_5_13, \
call_13_13_5_14, \
call_13_13_5_15, \
call_13_13_6_0, \
call_13_13_6_1, \
call_13_13_6_2, \
call_13_13_6_3, \
call_13_13_6_4, \
call_13_13_6_5, \
call_13_13_6_6, \
call_13_13_6_7, \
call_13_13_6_8, \
call_13_13_6_9, \
call_13_13_6_10, \
call_13_13_6_11, \
call_13_13_6_12, \
call_13_13_6_13, \
call_13_13_6_14, \
call_13_13_6_15, \
call_13_13_7_0, \
call_13_13_7_1, \
call_13_13_7_2, \
call_13_13_7_3, \
call_13_13_7_4, \
call_13_13_7_5, \
call_13_13_7_6, \
call_13_13_7_7, \
call_13_13_7_8, \
call_13_13_7_9, \
call_13_13_7_10, \
call_13_13_7_11, \
call_13_13_7_12, \
call_13_13_7_13, \
call_13_13_7_14, \
call_13_13_7_15, \
call_13_13_8_0, \
call_13_13_8_1, \
call_13_13_8_2, \
call_13_13_8_3, \
call_13_13_8_4, \
call_13_13_8_5, \
call_13_13_8_6, \
call_13_13_8_7, \
call_13_13_8_8, \
call_13_13_8_9, \
call_13_13_8_10, \
call_13_13_8_11, \
call_13_13_8_12, \
call_13_13_8_13, \
call_13_13_8_14, \
call_13_13_8_15, \
call_13_13_9_0, \
call_13_13_9_1, \
call_13_13_9_2, \
call_13_13_9_3, \
call_13_13_9_4, \
call_13_13_9_5, \
call_13_13_9_6, \
call_13_13_9_7, \
call_13_13_9_8, \
call_13_13_9_9, \
call_13_13_9_10, \
call_13_13_9_11, \
call_13_13_9_12, \
call_13_13_9_13, \
call_13_13_9_14, \
call_13_13_9_15, \
call_13_13_10_0, \
call_13_13_10_1, \
call_13_13_10_2, \
call_13_13_10_3, \
call_13_13_10_4, \
call_13_13_10_5, \
call_13_13_10_6, \
call_13_13_10_7, \
call_13_13_10_8, \
call_13_13_10_9, \
call_13_13_10_10, \
call_13_13_10_11, \
call_13_13_10_12, \
call_13_13_10_13, \
call_13_13_10_14, \
call_13_13_10_15, \
call_13_13_11_0, \
call_13_13_11_1, \
call_13_13_11_2, \
call_13_13_11_3, \
call_13_13_11_4, \
call_13_13_11_5, \
call_13_13_11_6, \
call_13_13_11_7, \
call_13_13_11_8, \
call_13_13_11_9, \
call_13_13_11_10, \
call_13_13_11_11, \
call_13_13_11_12, \
call_13_13_11_13, \
call_13_13_11_14, \
call_13_13_11_15, \
call_13_13_12_0, \
call_13_13_12_1, \
call_13_13_12_2, \
call_13_13_12_3, \
call_13_13_12_4, \
call_13_13_12_5, \
call_13_13_12_6, \
call_13_13_12_7, \
call_13_13_12_8, \
call_13_13_12_9, \
call_13_13_12_10, \
call_13_13_12_11, \
call_13_13_12_12, \
call_13_13_12_13, \
call_13_13_12_14, \
call_13_13_12_15, \
call_13_13_13_0, \
call_13_13_13_1, \
call_13_13_13_2, \
call_13_13_13_3, \
call_13_13_13_4, \
call_13_13_13_5, \
call_13_13_13_6, \
call_13_13_13_7, \
call_13_13_13_8, \
call_13_13_13_9, \
call_13_13_13_10, \
call_13_13_13_11, \
call_13_13_13_12, \
call_13_13_13_13, \
call_13_13_13_14, \
call_13_13_13_15, \
call_13_13_14_0, \
call_13_13_14_1, \
call_13_13_14_2, \
call_13_13_14_3, \
call_13_13_14_4, \
call_13_13_14_5, \
call_13_13_14_6, \
call_13_13_14_7, \
call_13_13_14_8, \
call_13_13_14_9, \
call_13_13_14_10, \
call_13_13_14_11, \
call_13_13_14_12, \
call_13_13_14_13, \
call_13_13_14_14, \
call_13_13_14_15, \
call_13_13_15_0, \
call_13_13_15_1, \
call_13_13_15_2, \
call_13_13_15_3, \
call_13_13_15_4, \
call_13_13_15_5, \
call_13_13_15_6, \
call_13_13_15_7, \
call_13_13_15_8, \
call_13_13_15_9, \
call_13_13_15_10, \
call_13_13_15_11, \
call_13_13_15_12, \
call_13_13_15_13, \
call_13_13_15_14, \
call_13_13_15_15, \
call_13_14_0_0, \
call_13_14_0_1, \
call_13_14_0_2, \
call_13_14_0_3, \
call_13_14_0_4, \
call_13_14_0_5, \
call_13_14_0_6, \
call_13_14_0_7, \
call_13_14_0_8, \
call_13_14_0_9, \
call_13_14_0_10, \
call_13_14_0_11, \
call_13_14_0_12, \
call_13_14_0_13, \
call_13_14_0_14, \
call_13_14_0_15, \
call_13_14_1_0, \
call_13_14_1_1, \
call_13_14_1_2, \
call_13_14_1_3, \
call_13_14_1_4, \
call_13_14_1_5, \
call_13_14_1_6, \
call_13_14_1_7, \
call_13_14_1_8, \
call_13_14_1_9, \
call_13_14_1_10, \
call_13_14_1_11, \
call_13_14_1_12, \
call_13_14_1_13, \
call_13_14_1_14, \
call_13_14_1_15, \
call_13_14_2_0, \
call_13_14_2_1, \
call_13_14_2_2, \
call_13_14_2_3, \
call_13_14_2_4, \
call_13_14_2_5, \
call_13_14_2_6, \
call_13_14_2_7, \
call_13_14_2_8, \
call_13_14_2_9, \
call_13_14_2_10, \
call_13_14_2_11, \
call_13_14_2_12, \
call_13_14_2_13, \
call_13_14_2_14, \
call_13_14_2_15, \
call_13_14_3_0, \
call_13_14_3_1, \
call_13_14_3_2, \
call_13_14_3_3, \
call_13_14_3_4, \
call_13_14_3_5, \
call_13_14_3_6, \
call_13_14_3_7, \
call_13_14_3_8, \
call_13_14_3_9, \
call_13_14_3_10, \
call_13_14_3_11, \
call_13_14_3_12, \
call_13_14_3_13, \
call_13_14_3_14, \
call_13_14_3_15, \
call_13_14_4_0, \
call_13_14_4_1, \
call_13_14_4_2, \
call_13_14_4_3, \
call_13_14_4_4, \
call_13_14_4_5, \
call_13_14_4_6, \
call_13_14_4_7, \
call_13_14_4_8, \
call_13_14_4_9, \
call_13_14_4_10, \
call_13_14_4_11, \
call_13_14_4_12, \
call_13_14_4_13, \
call_13_14_4_14, \
call_13_14_4_15, \
call_13_14_5_0, \
call_13_14_5_1, \
call_13_14_5_2, \
call_13_14_5_3, \
call_13_14_5_4, \
call_13_14_5_5, \
call_13_14_5_6, \
call_13_14_5_7, \
call_13_14_5_8, \
call_13_14_5_9, \
call_13_14_5_10, \
call_13_14_5_11, \
call_13_14_5_12, \
call_13_14_5_13, \
call_13_14_5_14, \
call_13_14_5_15, \
call_13_14_6_0, \
call_13_14_6_1, \
call_13_14_6_2, \
call_13_14_6_3, \
call_13_14_6_4, \
call_13_14_6_5, \
call_13_14_6_6, \
call_13_14_6_7, \
call_13_14_6_8, \
call_13_14_6_9, \
call_13_14_6_10, \
call_13_14_6_11, \
call_13_14_6_12, \
call_13_14_6_13, \
call_13_14_6_14, \
call_13_14_6_15, \
call_13_14_7_0, \
call_13_14_7_1, \
call_13_14_7_2, \
call_13_14_7_3, \
call_13_14_7_4, \
call_13_14_7_5, \
call_13_14_7_6, \
call_13_14_7_7, \
call_13_14_7_8, \
call_13_14_7_9, \
call_13_14_7_10, \
call_13_14_7_11, \
call_13_14_7_12, \
call_13_14_7_13, \
call_13_14_7_14, \
call_13_14_7_15, \
call_13_14_8_0, \
call_13_14_8_1, \
call_13_14_8_2, \
call_13_14_8_3, \
call_13_14_8_4, \
call_13_14_8_5, \
call_13_14_8_6, \
call_13_14_8_7, \
call_13_14_8_8, \
call_13_14_8_9, \
call_13_14_8_10, \
call_13_14_8_11, \
call_13_14_8_12, \
call_13_14_8_13, \
call_13_14_8_14, \
call_13_14_8_15, \
call_13_14_9_0, \
call_13_14_9_1, \
call_13_14_9_2, \
call_13_14_9_3, \
call_13_14_9_4, \
call_13_14_9_5, \
call_13_14_9_6, \
call_13_14_9_7, \
call_13_14_9_8, \
call_13_14_9_9, \
call_13_14_9_10, \
call_13_14_9_11, \
call_13_14_9_12, \
call_13_14_9_13, \
call_13_14_9_14, \
call_13_14_9_15, \
call_13_14_10_0, \
call_13_14_10_1, \
call_13_14_10_2, \
call_13_14_10_3, \
call_13_14_10_4, \
call_13_14_10_5, \
call_13_14_10_6, \
call_13_14_10_7, \
call_13_14_10_8, \
call_13_14_10_9, \
call_13_14_10_10, \
call_13_14_10_11, \
call_13_14_10_12, \
call_13_14_10_13, \
call_13_14_10_14, \
call_13_14_10_15, \
call_13_14_11_0, \
call_13_14_11_1, \
call_13_14_11_2, \
call_13_14_11_3, \
call_13_14_11_4, \
call_13_14_11_5, \
call_13_14_11_6, \
call_13_14_11_7, \
call_13_14_11_8, \
call_13_14_11_9, \
call_13_14_11_10, \
call_13_14_11_11, \
call_13_14_11_12, \
call_13_14_11_13, \
call_13_14_11_14, \
call_13_14_11_15, \
call_13_14_12_0, \
call_13_14_12_1, \
call_13_14_12_2, \
call_13_14_12_3, \
call_13_14_12_4, \
call_13_14_12_5, \
call_13_14_12_6, \
call_13_14_12_7, \
call_13_14_12_8, \
call_13_14_12_9, \
call_13_14_12_10, \
call_13_14_12_11, \
call_13_14_12_12, \
call_13_14_12_13, \
call_13_14_12_14, \
call_13_14_12_15, \
call_13_14_13_0, \
call_13_14_13_1, \
call_13_14_13_2, \
call_13_14_13_3, \
call_13_14_13_4, \
call_13_14_13_5, \
call_13_14_13_6, \
call_13_14_13_7, \
call_13_14_13_8, \
call_13_14_13_9, \
call_13_14_13_10, \
call_13_14_13_11, \
call_13_14_13_12, \
call_13_14_13_13, \
call_13_14_13_14, \
call_13_14_13_15, \
call_13_14_14_0, \
call_13_14_14_1, \
call_13_14_14_2, \
call_13_14_14_3, \
call_13_14_14_4, \
call_13_14_14_5, \
call_13_14_14_6, \
call_13_14_14_7, \
call_13_14_14_8, \
call_13_14_14_9, \
call_13_14_14_10, \
call_13_14_14_11, \
call_13_14_14_12, \
call_13_14_14_13, \
call_13_14_14_14, \
call_13_14_14_15, \
call_13_14_15_0, \
call_13_14_15_1, \
call_13_14_15_2, \
call_13_14_15_3, \
call_13_14_15_4, \
call_13_14_15_5, \
call_13_14_15_6, \
call_13_14_15_7, \
call_13_14_15_8, \
call_13_14_15_9, \
call_13_14_15_10, \
call_13_14_15_11, \
call_13_14_15_12, \
call_13_14_15_13, \
call_13_14_15_14, \
call_13_14_15_15, \
call_13_15_0_0, \
call_13_15_0_1, \
call_13_15_0_2, \
call_13_15_0_3, \
call_13_15_0_4, \
call_13_15_0_5, \
call_13_15_0_6, \
call_13_15_0_7, \
call_13_15_0_8, \
call_13_15_0_9, \
call_13_15_0_10, \
call_13_15_0_11, \
call_13_15_0_12, \
call_13_15_0_13, \
call_13_15_0_14, \
call_13_15_0_15, \
call_13_15_1_0, \
call_13_15_1_1, \
call_13_15_1_2, \
call_13_15_1_3, \
call_13_15_1_4, \
call_13_15_1_5, \
call_13_15_1_6, \
call_13_15_1_7, \
call_13_15_1_8, \
call_13_15_1_9, \
call_13_15_1_10, \
call_13_15_1_11, \
call_13_15_1_12, \
call_13_15_1_13, \
call_13_15_1_14, \
call_13_15_1_15, \
call_13_15_2_0, \
call_13_15_2_1, \
call_13_15_2_2, \
call_13_15_2_3, \
call_13_15_2_4, \
call_13_15_2_5, \
call_13_15_2_6, \
call_13_15_2_7, \
call_13_15_2_8, \
call_13_15_2_9, \
call_13_15_2_10, \
call_13_15_2_11, \
call_13_15_2_12, \
call_13_15_2_13, \
call_13_15_2_14, \
call_13_15_2_15, \
call_13_15_3_0, \
call_13_15_3_1, \
call_13_15_3_2, \
call_13_15_3_3, \
call_13_15_3_4, \
call_13_15_3_5, \
call_13_15_3_6, \
call_13_15_3_7, \
call_13_15_3_8, \
call_13_15_3_9, \
call_13_15_3_10, \
call_13_15_3_11, \
call_13_15_3_12, \
call_13_15_3_13, \
call_13_15_3_14, \
call_13_15_3_15, \
call_13_15_4_0, \
call_13_15_4_1, \
call_13_15_4_2, \
call_13_15_4_3, \
call_13_15_4_4, \
call_13_15_4_5, \
call_13_15_4_6, \
call_13_15_4_7, \
call_13_15_4_8, \
call_13_15_4_9, \
call_13_15_4_10, \
call_13_15_4_11, \
call_13_15_4_12, \
call_13_15_4_13, \
call_13_15_4_14, \
call_13_15_4_15, \
call_13_15_5_0, \
call_13_15_5_1, \
call_13_15_5_2, \
call_13_15_5_3, \
call_13_15_5_4, \
call_13_15_5_5, \
call_13_15_5_6, \
call_13_15_5_7, \
call_13_15_5_8, \
call_13_15_5_9, \
call_13_15_5_10, \
call_13_15_5_11, \
call_13_15_5_12, \
call_13_15_5_13, \
call_13_15_5_14, \
call_13_15_5_15, \
call_13_15_6_0, \
call_13_15_6_1, \
call_13_15_6_2, \
call_13_15_6_3, \
call_13_15_6_4, \
call_13_15_6_5, \
call_13_15_6_6, \
call_13_15_6_7, \
call_13_15_6_8, \
call_13_15_6_9, \
call_13_15_6_10, \
call_13_15_6_11, \
call_13_15_6_12, \
call_13_15_6_13, \
call_13_15_6_14, \
call_13_15_6_15, \
call_13_15_7_0, \
call_13_15_7_1, \
call_13_15_7_2, \
call_13_15_7_3, \
call_13_15_7_4, \
call_13_15_7_5, \
call_13_15_7_6, \
call_13_15_7_7, \
call_13_15_7_8, \
call_13_15_7_9, \
call_13_15_7_10, \
call_13_15_7_11, \
call_13_15_7_12, \
call_13_15_7_13, \
call_13_15_7_14, \
call_13_15_7_15, \
call_13_15_8_0, \
call_13_15_8_1, \
call_13_15_8_2, \
call_13_15_8_3, \
call_13_15_8_4, \
call_13_15_8_5, \
call_13_15_8_6, \
call_13_15_8_7, \
call_13_15_8_8, \
call_13_15_8_9, \
call_13_15_8_10, \
call_13_15_8_11, \
call_13_15_8_12, \
call_13_15_8_13, \
call_13_15_8_14, \
call_13_15_8_15, \
call_13_15_9_0, \
call_13_15_9_1, \
call_13_15_9_2, \
call_13_15_9_3, \
call_13_15_9_4, \
call_13_15_9_5, \
call_13_15_9_6, \
call_13_15_9_7, \
call_13_15_9_8, \
call_13_15_9_9, \
call_13_15_9_10, \
call_13_15_9_11, \
call_13_15_9_12, \
call_13_15_9_13, \
call_13_15_9_14, \
call_13_15_9_15, \
call_13_15_10_0, \
call_13_15_10_1, \
call_13_15_10_2, \
call_13_15_10_3, \
call_13_15_10_4, \
call_13_15_10_5, \
call_13_15_10_6, \
call_13_15_10_7, \
call_13_15_10_8, \
call_13_15_10_9, \
call_13_15_10_10, \
call_13_15_10_11, \
call_13_15_10_12, \
call_13_15_10_13, \
call_13_15_10_14, \
call_13_15_10_15, \
call_13_15_11_0, \
call_13_15_11_1, \
call_13_15_11_2, \
call_13_15_11_3, \
call_13_15_11_4, \
call_13_15_11_5, \
call_13_15_11_6, \
call_13_15_11_7, \
call_13_15_11_8, \
call_13_15_11_9, \
call_13_15_11_10, \
call_13_15_11_11, \
call_13_15_11_12, \
call_13_15_11_13, \
call_13_15_11_14, \
call_13_15_11_15, \
call_13_15_12_0, \
call_13_15_12_1, \
call_13_15_12_2, \
call_13_15_12_3, \
call_13_15_12_4, \
call_13_15_12_5, \
call_13_15_12_6, \
call_13_15_12_7, \
call_13_15_12_8, \
call_13_15_12_9, \
call_13_15_12_10, \
call_13_15_12_11, \
call_13_15_12_12, \
call_13_15_12_13, \
call_13_15_12_14, \
call_13_15_12_15, \
call_13_15_13_0, \
call_13_15_13_1, \
call_13_15_13_2, \
call_13_15_13_3, \
call_13_15_13_4, \
call_13_15_13_5, \
call_13_15_13_6, \
call_13_15_13_7, \
call_13_15_13_8, \
call_13_15_13_9, \
call_13_15_13_10, \
call_13_15_13_11, \
call_13_15_13_12, \
call_13_15_13_13, \
call_13_15_13_14, \
call_13_15_13_15, \
call_13_15_14_0, \
call_13_15_14_1, \
call_13_15_14_2, \
call_13_15_14_3, \
call_13_15_14_4, \
call_13_15_14_5, \
call_13_15_14_6, \
call_13_15_14_7, \
call_13_15_14_8, \
call_13_15_14_9, \
call_13_15_14_10, \
call_13_15_14_11, \
call_13_15_14_12, \
call_13_15_14_13, \
call_13_15_14_14, \
call_13_15_14_15, \
call_13_15_15_0, \
call_13_15_15_1, \
call_13_15_15_2, \
call_13_15_15_3, \
call_13_15_15_4, \
call_13_15_15_5, \
call_13_15_15_6, \
call_13_15_15_7, \
call_13_15_15_8, \
call_13_15_15_9, \
call_13_15_15_10, \
call_13_15_15_11, \
call_13_15_15_12, \
call_13_15_15_13, \
call_13_15_15_14, \
call_13_15_15_15, \
call_14_0_0_0, \
call_14_0_0_1, \
call_14_0_0_2, \
call_14_0_0_3, \
call_14_0_0_4, \
call_14_0_0_5, \
call_14_0_0_6, \
call_14_0_0_7, \
call_14_0_0_8, \
call_14_0_0_9, \
call_14_0_0_10, \
call_14_0_0_11, \
call_14_0_0_12, \
call_14_0_0_13, \
call_14_0_0_14, \
call_14_0_0_15, \
call_14_0_1_0, \
call_14_0_1_1, \
call_14_0_1_2, \
call_14_0_1_3, \
call_14_0_1_4, \
call_14_0_1_5, \
call_14_0_1_6, \
call_14_0_1_7, \
call_14_0_1_8, \
call_14_0_1_9, \
call_14_0_1_10, \
call_14_0_1_11, \
call_14_0_1_12, \
call_14_0_1_13, \
call_14_0_1_14, \
call_14_0_1_15, \
call_14_0_2_0, \
call_14_0_2_1, \
call_14_0_2_2, \
call_14_0_2_3, \
call_14_0_2_4, \
call_14_0_2_5, \
call_14_0_2_6, \
call_14_0_2_7, \
call_14_0_2_8, \
call_14_0_2_9, \
call_14_0_2_10, \
call_14_0_2_11, \
call_14_0_2_12, \
call_14_0_2_13, \
call_14_0_2_14, \
call_14_0_2_15, \
call_14_0_3_0, \
call_14_0_3_1, \
call_14_0_3_2, \
call_14_0_3_3, \
call_14_0_3_4, \
call_14_0_3_5, \
call_14_0_3_6, \
call_14_0_3_7, \
call_14_0_3_8, \
call_14_0_3_9, \
call_14_0_3_10, \
call_14_0_3_11, \
call_14_0_3_12, \
call_14_0_3_13, \
call_14_0_3_14, \
call_14_0_3_15, \
call_14_0_4_0, \
call_14_0_4_1, \
call_14_0_4_2, \
call_14_0_4_3, \
call_14_0_4_4, \
call_14_0_4_5, \
call_14_0_4_6, \
call_14_0_4_7, \
call_14_0_4_8, \
call_14_0_4_9, \
call_14_0_4_10, \
call_14_0_4_11, \
call_14_0_4_12, \
call_14_0_4_13, \
call_14_0_4_14, \
call_14_0_4_15, \
call_14_0_5_0, \
call_14_0_5_1, \
call_14_0_5_2, \
call_14_0_5_3, \
call_14_0_5_4, \
call_14_0_5_5, \
call_14_0_5_6, \
call_14_0_5_7, \
call_14_0_5_8, \
call_14_0_5_9, \
call_14_0_5_10, \
call_14_0_5_11, \
call_14_0_5_12, \
call_14_0_5_13, \
call_14_0_5_14, \
call_14_0_5_15, \
call_14_0_6_0, \
call_14_0_6_1, \
call_14_0_6_2, \
call_14_0_6_3, \
call_14_0_6_4, \
call_14_0_6_5, \
call_14_0_6_6, \
call_14_0_6_7, \
call_14_0_6_8, \
call_14_0_6_9, \
call_14_0_6_10, \
call_14_0_6_11, \
call_14_0_6_12, \
call_14_0_6_13, \
call_14_0_6_14, \
call_14_0_6_15, \
call_14_0_7_0, \
call_14_0_7_1, \
call_14_0_7_2, \
call_14_0_7_3, \
call_14_0_7_4, \
call_14_0_7_5, \
call_14_0_7_6, \
call_14_0_7_7, \
call_14_0_7_8, \
call_14_0_7_9, \
call_14_0_7_10, \
call_14_0_7_11, \
call_14_0_7_12, \
call_14_0_7_13, \
call_14_0_7_14, \
call_14_0_7_15, \
call_14_0_8_0, \
call_14_0_8_1, \
call_14_0_8_2, \
call_14_0_8_3, \
call_14_0_8_4, \
call_14_0_8_5, \
call_14_0_8_6, \
call_14_0_8_7, \
call_14_0_8_8, \
call_14_0_8_9, \
call_14_0_8_10, \
call_14_0_8_11, \
call_14_0_8_12, \
call_14_0_8_13, \
call_14_0_8_14, \
call_14_0_8_15, \
call_14_0_9_0, \
call_14_0_9_1, \
call_14_0_9_2, \
call_14_0_9_3, \
call_14_0_9_4, \
call_14_0_9_5, \
call_14_0_9_6, \
call_14_0_9_7, \
call_14_0_9_8, \
call_14_0_9_9, \
call_14_0_9_10, \
call_14_0_9_11, \
call_14_0_9_12, \
call_14_0_9_13, \
call_14_0_9_14, \
call_14_0_9_15, \
call_14_0_10_0, \
call_14_0_10_1, \
call_14_0_10_2, \
call_14_0_10_3, \
call_14_0_10_4, \
call_14_0_10_5, \
call_14_0_10_6, \
call_14_0_10_7, \
call_14_0_10_8, \
call_14_0_10_9, \
call_14_0_10_10, \
call_14_0_10_11, \
call_14_0_10_12, \
call_14_0_10_13, \
call_14_0_10_14, \
call_14_0_10_15, \
call_14_0_11_0, \
call_14_0_11_1, \
call_14_0_11_2, \
call_14_0_11_3, \
call_14_0_11_4, \
call_14_0_11_5, \
call_14_0_11_6, \
call_14_0_11_7, \
call_14_0_11_8, \
call_14_0_11_9, \
call_14_0_11_10, \
call_14_0_11_11, \
call_14_0_11_12, \
call_14_0_11_13, \
call_14_0_11_14, \
call_14_0_11_15, \
call_14_0_12_0, \
call_14_0_12_1, \
call_14_0_12_2, \
call_14_0_12_3, \
call_14_0_12_4, \
call_14_0_12_5, \
call_14_0_12_6, \
call_14_0_12_7, \
call_14_0_12_8, \
call_14_0_12_9, \
call_14_0_12_10, \
call_14_0_12_11, \
call_14_0_12_12, \
call_14_0_12_13, \
call_14_0_12_14, \
call_14_0_12_15, \
call_14_0_13_0, \
call_14_0_13_1, \
call_14_0_13_2, \
call_14_0_13_3, \
call_14_0_13_4, \
call_14_0_13_5, \
call_14_0_13_6, \
call_14_0_13_7, \
call_14_0_13_8, \
call_14_0_13_9, \
call_14_0_13_10, \
call_14_0_13_11, \
call_14_0_13_12, \
call_14_0_13_13, \
call_14_0_13_14, \
call_14_0_13_15, \
call_14_0_14_0, \
call_14_0_14_1, \
call_14_0_14_2, \
call_14_0_14_3, \
call_14_0_14_4, \
call_14_0_14_5, \
call_14_0_14_6, \
call_14_0_14_7, \
call_14_0_14_8, \
call_14_0_14_9, \
call_14_0_14_10, \
call_14_0_14_11, \
call_14_0_14_12, \
call_14_0_14_13, \
call_14_0_14_14, \
call_14_0_14_15, \
call_14_0_15_0, \
call_14_0_15_1, \
call_14_0_15_2, \
call_14_0_15_3, \
call_14_0_15_4, \
call_14_0_15_5, \
call_14_0_15_6, \
call_14_0_15_7, \
call_14_0_15_8, \
call_14_0_15_9, \
call_14_0_15_10, \
call_14_0_15_11, \
call_14_0_15_12, \
call_14_0_15_13, \
call_14_0_15_14, \
call_14_0_15_15, \
call_14_1_0_0, \
call_14_1_0_1, \
call_14_1_0_2, \
call_14_1_0_3, \
call_14_1_0_4, \
call_14_1_0_5, \
call_14_1_0_6, \
call_14_1_0_7, \
call_14_1_0_8, \
call_14_1_0_9, \
call_14_1_0_10, \
call_14_1_0_11, \
call_14_1_0_12, \
call_14_1_0_13, \
call_14_1_0_14, \
call_14_1_0_15, \
call_14_1_1_0, \
call_14_1_1_1, \
call_14_1_1_2, \
call_14_1_1_3, \
call_14_1_1_4, \
call_14_1_1_5, \
call_14_1_1_6, \
call_14_1_1_7, \
call_14_1_1_8, \
call_14_1_1_9, \
call_14_1_1_10, \
call_14_1_1_11, \
call_14_1_1_12, \
call_14_1_1_13, \
call_14_1_1_14, \
call_14_1_1_15, \
call_14_1_2_0, \
call_14_1_2_1, \
call_14_1_2_2, \
call_14_1_2_3, \
call_14_1_2_4, \
call_14_1_2_5, \
call_14_1_2_6, \
call_14_1_2_7, \
call_14_1_2_8, \
call_14_1_2_9, \
call_14_1_2_10, \
call_14_1_2_11, \
call_14_1_2_12, \
call_14_1_2_13, \
call_14_1_2_14, \
call_14_1_2_15, \
call_14_1_3_0, \
call_14_1_3_1, \
call_14_1_3_2, \
call_14_1_3_3, \
call_14_1_3_4, \
call_14_1_3_5, \
call_14_1_3_6, \
call_14_1_3_7, \
call_14_1_3_8, \
call_14_1_3_9, \
call_14_1_3_10, \
call_14_1_3_11, \
call_14_1_3_12, \
call_14_1_3_13, \
call_14_1_3_14, \
call_14_1_3_15, \
call_14_1_4_0, \
call_14_1_4_1, \
call_14_1_4_2, \
call_14_1_4_3, \
call_14_1_4_4, \
call_14_1_4_5, \
call_14_1_4_6, \
call_14_1_4_7, \
call_14_1_4_8, \
call_14_1_4_9, \
call_14_1_4_10, \
call_14_1_4_11, \
call_14_1_4_12, \
call_14_1_4_13, \
call_14_1_4_14, \
call_14_1_4_15, \
call_14_1_5_0, \
call_14_1_5_1, \
call_14_1_5_2, \
call_14_1_5_3, \
call_14_1_5_4, \
call_14_1_5_5, \
call_14_1_5_6, \
call_14_1_5_7, \
call_14_1_5_8, \
call_14_1_5_9, \
call_14_1_5_10, \
call_14_1_5_11, \
call_14_1_5_12, \
call_14_1_5_13, \
call_14_1_5_14, \
call_14_1_5_15, \
call_14_1_6_0, \
call_14_1_6_1, \
call_14_1_6_2, \
call_14_1_6_3, \
call_14_1_6_4, \
call_14_1_6_5, \
call_14_1_6_6, \
call_14_1_6_7, \
call_14_1_6_8, \
call_14_1_6_9, \
call_14_1_6_10, \
call_14_1_6_11, \
call_14_1_6_12, \
call_14_1_6_13, \
call_14_1_6_14, \
call_14_1_6_15, \
call_14_1_7_0, \
call_14_1_7_1, \
call_14_1_7_2, \
call_14_1_7_3, \
call_14_1_7_4, \
call_14_1_7_5, \
call_14_1_7_6, \
call_14_1_7_7, \
call_14_1_7_8, \
call_14_1_7_9, \
call_14_1_7_10, \
call_14_1_7_11, \
call_14_1_7_12, \
call_14_1_7_13, \
call_14_1_7_14, \
call_14_1_7_15, \
call_14_1_8_0, \
call_14_1_8_1, \
call_14_1_8_2, \
call_14_1_8_3, \
call_14_1_8_4, \
call_14_1_8_5, \
call_14_1_8_6, \
call_14_1_8_7, \
call_14_1_8_8, \
call_14_1_8_9, \
call_14_1_8_10, \
call_14_1_8_11, \
call_14_1_8_12, \
call_14_1_8_13, \
call_14_1_8_14, \
call_14_1_8_15, \
call_14_1_9_0, \
call_14_1_9_1, \
call_14_1_9_2, \
call_14_1_9_3, \
call_14_1_9_4, \
call_14_1_9_5, \
call_14_1_9_6, \
call_14_1_9_7, \
call_14_1_9_8, \
call_14_1_9_9, \
call_14_1_9_10, \
call_14_1_9_11, \
call_14_1_9_12, \
call_14_1_9_13, \
call_14_1_9_14, \
call_14_1_9_15, \
call_14_1_10_0, \
call_14_1_10_1, \
call_14_1_10_2, \
call_14_1_10_3, \
call_14_1_10_4, \
call_14_1_10_5, \
call_14_1_10_6, \
call_14_1_10_7, \
call_14_1_10_8, \
call_14_1_10_9, \
call_14_1_10_10, \
call_14_1_10_11, \
call_14_1_10_12, \
call_14_1_10_13, \
call_14_1_10_14, \
call_14_1_10_15, \
call_14_1_11_0, \
call_14_1_11_1, \
call_14_1_11_2, \
call_14_1_11_3, \
call_14_1_11_4, \
call_14_1_11_5, \
call_14_1_11_6, \
call_14_1_11_7, \
call_14_1_11_8, \
call_14_1_11_9, \
call_14_1_11_10, \
call_14_1_11_11, \
call_14_1_11_12, \
call_14_1_11_13, \
call_14_1_11_14, \
call_14_1_11_15, \
call_14_1_12_0, \
call_14_1_12_1, \
call_14_1_12_2, \
call_14_1_12_3, \
call_14_1_12_4, \
call_14_1_12_5, \
call_14_1_12_6, \
call_14_1_12_7, \
call_14_1_12_8, \
call_14_1_12_9, \
call_14_1_12_10, \
call_14_1_12_11, \
call_14_1_12_12, \
call_14_1_12_13, \
call_14_1_12_14, \
call_14_1_12_15, \
call_14_1_13_0, \
call_14_1_13_1, \
call_14_1_13_2, \
call_14_1_13_3, \
call_14_1_13_4, \
call_14_1_13_5, \
call_14_1_13_6, \
call_14_1_13_7, \
call_14_1_13_8, \
call_14_1_13_9, \
call_14_1_13_10, \
call_14_1_13_11, \
call_14_1_13_12, \
call_14_1_13_13, \
call_14_1_13_14, \
call_14_1_13_15, \
call_14_1_14_0, \
call_14_1_14_1, \
call_14_1_14_2, \
call_14_1_14_3, \
call_14_1_14_4, \
call_14_1_14_5, \
call_14_1_14_6, \
call_14_1_14_7, \
call_14_1_14_8, \
call_14_1_14_9, \
call_14_1_14_10, \
call_14_1_14_11, \
call_14_1_14_12, \
call_14_1_14_13, \
call_14_1_14_14, \
call_14_1_14_15, \
call_14_1_15_0, \
call_14_1_15_1, \
call_14_1_15_2, \
call_14_1_15_3, \
call_14_1_15_4, \
call_14_1_15_5, \
call_14_1_15_6, \
call_14_1_15_7, \
call_14_1_15_8, \
call_14_1_15_9, \
call_14_1_15_10, \
call_14_1_15_11, \
call_14_1_15_12, \
call_14_1_15_13, \
call_14_1_15_14, \
call_14_1_15_15, \
call_14_2_0_0, \
call_14_2_0_1, \
call_14_2_0_2, \
call_14_2_0_3, \
call_14_2_0_4, \
call_14_2_0_5, \
call_14_2_0_6, \
call_14_2_0_7, \
call_14_2_0_8, \
call_14_2_0_9, \
call_14_2_0_10, \
call_14_2_0_11, \
call_14_2_0_12, \
call_14_2_0_13, \
call_14_2_0_14, \
call_14_2_0_15, \
call_14_2_1_0, \
call_14_2_1_1, \
call_14_2_1_2, \
call_14_2_1_3, \
call_14_2_1_4, \
call_14_2_1_5, \
call_14_2_1_6, \
call_14_2_1_7, \
call_14_2_1_8, \
call_14_2_1_9, \
call_14_2_1_10, \
call_14_2_1_11, \
call_14_2_1_12, \
call_14_2_1_13, \
call_14_2_1_14, \
call_14_2_1_15, \
call_14_2_2_0, \
call_14_2_2_1, \
call_14_2_2_2, \
call_14_2_2_3, \
call_14_2_2_4, \
call_14_2_2_5, \
call_14_2_2_6, \
call_14_2_2_7, \
call_14_2_2_8, \
call_14_2_2_9, \
call_14_2_2_10, \
call_14_2_2_11, \
call_14_2_2_12, \
call_14_2_2_13, \
call_14_2_2_14, \
call_14_2_2_15, \
call_14_2_3_0, \
call_14_2_3_1, \
call_14_2_3_2, \
call_14_2_3_3, \
call_14_2_3_4, \
call_14_2_3_5, \
call_14_2_3_6, \
call_14_2_3_7, \
call_14_2_3_8, \
call_14_2_3_9, \
call_14_2_3_10, \
call_14_2_3_11, \
call_14_2_3_12, \
call_14_2_3_13, \
call_14_2_3_14, \
call_14_2_3_15, \
call_14_2_4_0, \
call_14_2_4_1, \
call_14_2_4_2, \
call_14_2_4_3, \
call_14_2_4_4, \
call_14_2_4_5, \
call_14_2_4_6, \
call_14_2_4_7, \
call_14_2_4_8, \
call_14_2_4_9, \
call_14_2_4_10, \
call_14_2_4_11, \
call_14_2_4_12, \
call_14_2_4_13, \
call_14_2_4_14, \
call_14_2_4_15, \
call_14_2_5_0, \
call_14_2_5_1, \
call_14_2_5_2, \
call_14_2_5_3, \
call_14_2_5_4, \
call_14_2_5_5, \
call_14_2_5_6, \
call_14_2_5_7, \
call_14_2_5_8, \
call_14_2_5_9, \
call_14_2_5_10, \
call_14_2_5_11, \
call_14_2_5_12, \
call_14_2_5_13, \
call_14_2_5_14, \
call_14_2_5_15, \
call_14_2_6_0, \
call_14_2_6_1, \
call_14_2_6_2, \
call_14_2_6_3, \
call_14_2_6_4, \
call_14_2_6_5, \
call_14_2_6_6, \
call_14_2_6_7, \
call_14_2_6_8, \
call_14_2_6_9, \
call_14_2_6_10, \
call_14_2_6_11, \
call_14_2_6_12, \
call_14_2_6_13, \
call_14_2_6_14, \
call_14_2_6_15, \
call_14_2_7_0, \
call_14_2_7_1, \
call_14_2_7_2, \
call_14_2_7_3, \
call_14_2_7_4, \
call_14_2_7_5, \
call_14_2_7_6, \
call_14_2_7_7, \
call_14_2_7_8, \
call_14_2_7_9, \
call_14_2_7_10, \
call_14_2_7_11, \
call_14_2_7_12, \
call_14_2_7_13, \
call_14_2_7_14, \
call_14_2_7_15, \
call_14_2_8_0, \
call_14_2_8_1, \
call_14_2_8_2, \
call_14_2_8_3, \
call_14_2_8_4, \
call_14_2_8_5, \
call_14_2_8_6, \
call_14_2_8_7, \
call_14_2_8_8, \
call_14_2_8_9, \
call_14_2_8_10, \
call_14_2_8_11, \
call_14_2_8_12, \
call_14_2_8_13, \
call_14_2_8_14, \
call_14_2_8_15, \
call_14_2_9_0, \
call_14_2_9_1, \
call_14_2_9_2, \
call_14_2_9_3, \
call_14_2_9_4, \
call_14_2_9_5, \
call_14_2_9_6, \
call_14_2_9_7, \
call_14_2_9_8, \
call_14_2_9_9, \
call_14_2_9_10, \
call_14_2_9_11, \
call_14_2_9_12, \
call_14_2_9_13, \
call_14_2_9_14, \
call_14_2_9_15, \
call_14_2_10_0, \
call_14_2_10_1, \
call_14_2_10_2, \
call_14_2_10_3, \
call_14_2_10_4, \
call_14_2_10_5, \
call_14_2_10_6, \
call_14_2_10_7, \
call_14_2_10_8, \
call_14_2_10_9, \
call_14_2_10_10, \
call_14_2_10_11, \
call_14_2_10_12, \
call_14_2_10_13, \
call_14_2_10_14, \
call_14_2_10_15, \
call_14_2_11_0, \
call_14_2_11_1, \
call_14_2_11_2, \
call_14_2_11_3, \
call_14_2_11_4, \
call_14_2_11_5, \
call_14_2_11_6, \
call_14_2_11_7, \
call_14_2_11_8, \
call_14_2_11_9, \
call_14_2_11_10, \
call_14_2_11_11, \
call_14_2_11_12, \
call_14_2_11_13, \
call_14_2_11_14, \
call_14_2_11_15, \
call_14_2_12_0, \
call_14_2_12_1, \
call_14_2_12_2, \
call_14_2_12_3, \
call_14_2_12_4, \
call_14_2_12_5, \
call_14_2_12_6, \
call_14_2_12_7, \
call_14_2_12_8, \
call_14_2_12_9, \
call_14_2_12_10, \
call_14_2_12_11, \
call_14_2_12_12, \
call_14_2_12_13, \
call_14_2_12_14, \
call_14_2_12_15, \
call_14_2_13_0, \
call_14_2_13_1, \
call_14_2_13_2, \
call_14_2_13_3, \
call_14_2_13_4, \
call_14_2_13_5, \
call_14_2_13_6, \
call_14_2_13_7, \
call_14_2_13_8, \
call_14_2_13_9, \
call_14_2_13_10, \
call_14_2_13_11, \
call_14_2_13_12, \
call_14_2_13_13, \
call_14_2_13_14, \
call_14_2_13_15, \
call_14_2_14_0, \
call_14_2_14_1, \
call_14_2_14_2, \
call_14_2_14_3, \
call_14_2_14_4, \
call_14_2_14_5, \
call_14_2_14_6, \
call_14_2_14_7, \
call_14_2_14_8, \
call_14_2_14_9, \
call_14_2_14_10, \
call_14_2_14_11, \
call_14_2_14_12, \
call_14_2_14_13, \
call_14_2_14_14, \
call_14_2_14_15, \
call_14_2_15_0, \
call_14_2_15_1, \
call_14_2_15_2, \
call_14_2_15_3, \
call_14_2_15_4, \
call_14_2_15_5, \
call_14_2_15_6, \
call_14_2_15_7, \
call_14_2_15_8, \
call_14_2_15_9, \
call_14_2_15_10, \
call_14_2_15_11, \
call_14_2_15_12, \
call_14_2_15_13, \
call_14_2_15_14, \
call_14_2_15_15, \
call_14_3_0_0, \
call_14_3_0_1, \
call_14_3_0_2, \
call_14_3_0_3, \
call_14_3_0_4, \
call_14_3_0_5, \
call_14_3_0_6, \
call_14_3_0_7, \
call_14_3_0_8, \
call_14_3_0_9, \
call_14_3_0_10, \
call_14_3_0_11, \
call_14_3_0_12, \
call_14_3_0_13, \
call_14_3_0_14, \
call_14_3_0_15, \
call_14_3_1_0, \
call_14_3_1_1, \
call_14_3_1_2, \
call_14_3_1_3, \
call_14_3_1_4, \
call_14_3_1_5, \
call_14_3_1_6, \
call_14_3_1_7, \
call_14_3_1_8, \
call_14_3_1_9, \
call_14_3_1_10, \
call_14_3_1_11, \
call_14_3_1_12, \
call_14_3_1_13, \
call_14_3_1_14, \
call_14_3_1_15, \
call_14_3_2_0, \
call_14_3_2_1, \
call_14_3_2_2, \
call_14_3_2_3, \
call_14_3_2_4, \
call_14_3_2_5, \
call_14_3_2_6, \
call_14_3_2_7, \
call_14_3_2_8, \
call_14_3_2_9, \
call_14_3_2_10, \
call_14_3_2_11, \
call_14_3_2_12, \
call_14_3_2_13, \
call_14_3_2_14, \
call_14_3_2_15, \
call_14_3_3_0, \
call_14_3_3_1, \
call_14_3_3_2, \
call_14_3_3_3, \
call_14_3_3_4, \
call_14_3_3_5, \
call_14_3_3_6, \
call_14_3_3_7, \
call_14_3_3_8, \
call_14_3_3_9, \
call_14_3_3_10, \
call_14_3_3_11, \
call_14_3_3_12, \
call_14_3_3_13, \
call_14_3_3_14, \
call_14_3_3_15, \
call_14_3_4_0, \
call_14_3_4_1, \
call_14_3_4_2, \
call_14_3_4_3, \
call_14_3_4_4, \
call_14_3_4_5, \
call_14_3_4_6, \
call_14_3_4_7, \
call_14_3_4_8, \
call_14_3_4_9, \
call_14_3_4_10, \
call_14_3_4_11, \
call_14_3_4_12, \
call_14_3_4_13, \
call_14_3_4_14, \
call_14_3_4_15, \
call_14_3_5_0, \
call_14_3_5_1, \
call_14_3_5_2, \
call_14_3_5_3, \
call_14_3_5_4, \
call_14_3_5_5, \
call_14_3_5_6, \
call_14_3_5_7, \
call_14_3_5_8, \
call_14_3_5_9, \
call_14_3_5_10, \
call_14_3_5_11, \
call_14_3_5_12, \
call_14_3_5_13, \
call_14_3_5_14, \
call_14_3_5_15, \
call_14_3_6_0, \
call_14_3_6_1, \
call_14_3_6_2, \
call_14_3_6_3, \
call_14_3_6_4, \
call_14_3_6_5, \
call_14_3_6_6, \
call_14_3_6_7, \
call_14_3_6_8, \
call_14_3_6_9, \
call_14_3_6_10, \
call_14_3_6_11, \
call_14_3_6_12, \
call_14_3_6_13, \
call_14_3_6_14, \
call_14_3_6_15, \
call_14_3_7_0, \
call_14_3_7_1, \
call_14_3_7_2, \
call_14_3_7_3, \
call_14_3_7_4, \
call_14_3_7_5, \
call_14_3_7_6, \
call_14_3_7_7, \
call_14_3_7_8, \
call_14_3_7_9, \
call_14_3_7_10, \
call_14_3_7_11, \
call_14_3_7_12, \
call_14_3_7_13, \
call_14_3_7_14, \
call_14_3_7_15, \
call_14_3_8_0, \
call_14_3_8_1, \
call_14_3_8_2, \
call_14_3_8_3, \
call_14_3_8_4, \
call_14_3_8_5, \
call_14_3_8_6, \
call_14_3_8_7, \
call_14_3_8_8, \
call_14_3_8_9, \
call_14_3_8_10, \
call_14_3_8_11, \
call_14_3_8_12, \
call_14_3_8_13, \
call_14_3_8_14, \
call_14_3_8_15, \
call_14_3_9_0, \
call_14_3_9_1, \
call_14_3_9_2, \
call_14_3_9_3, \
call_14_3_9_4, \
call_14_3_9_5, \
call_14_3_9_6, \
call_14_3_9_7, \
call_14_3_9_8, \
call_14_3_9_9, \
call_14_3_9_10, \
call_14_3_9_11, \
call_14_3_9_12, \
call_14_3_9_13, \
call_14_3_9_14, \
call_14_3_9_15, \
call_14_3_10_0, \
call_14_3_10_1, \
call_14_3_10_2, \
call_14_3_10_3, \
call_14_3_10_4, \
call_14_3_10_5, \
call_14_3_10_6, \
call_14_3_10_7, \
call_14_3_10_8, \
call_14_3_10_9, \
call_14_3_10_10, \
call_14_3_10_11, \
call_14_3_10_12, \
call_14_3_10_13, \
call_14_3_10_14, \
call_14_3_10_15, \
call_14_3_11_0, \
call_14_3_11_1, \
call_14_3_11_2, \
call_14_3_11_3, \
call_14_3_11_4, \
call_14_3_11_5, \
call_14_3_11_6, \
call_14_3_11_7, \
call_14_3_11_8, \
call_14_3_11_9, \
call_14_3_11_10, \
call_14_3_11_11, \
call_14_3_11_12, \
call_14_3_11_13, \
call_14_3_11_14, \
call_14_3_11_15, \
call_14_3_12_0, \
call_14_3_12_1, \
call_14_3_12_2, \
call_14_3_12_3, \
call_14_3_12_4, \
call_14_3_12_5, \
call_14_3_12_6, \
call_14_3_12_7, \
call_14_3_12_8, \
call_14_3_12_9, \
call_14_3_12_10, \
call_14_3_12_11, \
call_14_3_12_12, \
call_14_3_12_13, \
call_14_3_12_14, \
call_14_3_12_15, \
call_14_3_13_0, \
call_14_3_13_1, \
call_14_3_13_2, \
call_14_3_13_3, \
call_14_3_13_4, \
call_14_3_13_5, \
call_14_3_13_6, \
call_14_3_13_7, \
call_14_3_13_8, \
call_14_3_13_9, \
call_14_3_13_10, \
call_14_3_13_11, \
call_14_3_13_12, \
call_14_3_13_13, \
call_14_3_13_14, \
call_14_3_13_15, \
call_14_3_14_0, \
call_14_3_14_1, \
call_14_3_14_2, \
call_14_3_14_3, \
call_14_3_14_4, \
call_14_3_14_5, \
call_14_3_14_6, \
call_14_3_14_7, \
call_14_3_14_8, \
call_14_3_14_9, \
call_14_3_14_10, \
call_14_3_14_11, \
call_14_3_14_12, \
call_14_3_14_13, \
call_14_3_14_14, \
call_14_3_14_15, \
call_14_3_15_0, \
call_14_3_15_1, \
call_14_3_15_2, \
call_14_3_15_3, \
call_14_3_15_4, \
call_14_3_15_5, \
call_14_3_15_6, \
call_14_3_15_7, \
call_14_3_15_8, \
call_14_3_15_9, \
call_14_3_15_10, \
call_14_3_15_11, \
call_14_3_15_12, \
call_14_3_15_13, \
call_14_3_15_14, \
call_14_3_15_15, \
call_14_4_0_0, \
call_14_4_0_1, \
call_14_4_0_2, \
call_14_4_0_3, \
call_14_4_0_4, \
call_14_4_0_5, \
call_14_4_0_6, \
call_14_4_0_7, \
call_14_4_0_8, \
call_14_4_0_9, \
call_14_4_0_10, \
call_14_4_0_11, \
call_14_4_0_12, \
call_14_4_0_13, \
call_14_4_0_14, \
call_14_4_0_15, \
call_14_4_1_0, \
call_14_4_1_1, \
call_14_4_1_2, \
call_14_4_1_3, \
call_14_4_1_4, \
call_14_4_1_5, \
call_14_4_1_6, \
call_14_4_1_7, \
call_14_4_1_8, \
call_14_4_1_9, \
call_14_4_1_10, \
call_14_4_1_11, \
call_14_4_1_12, \
call_14_4_1_13, \
call_14_4_1_14, \
call_14_4_1_15, \
call_14_4_2_0, \
call_14_4_2_1, \
call_14_4_2_2, \
call_14_4_2_3, \
call_14_4_2_4, \
call_14_4_2_5, \
call_14_4_2_6, \
call_14_4_2_7, \
call_14_4_2_8, \
call_14_4_2_9, \
call_14_4_2_10, \
call_14_4_2_11, \
call_14_4_2_12, \
call_14_4_2_13, \
call_14_4_2_14, \
call_14_4_2_15, \
call_14_4_3_0, \
call_14_4_3_1, \
call_14_4_3_2, \
call_14_4_3_3, \
call_14_4_3_4, \
call_14_4_3_5, \
call_14_4_3_6, \
call_14_4_3_7, \
call_14_4_3_8, \
call_14_4_3_9, \
call_14_4_3_10, \
call_14_4_3_11, \
call_14_4_3_12, \
call_14_4_3_13, \
call_14_4_3_14, \
call_14_4_3_15, \
call_14_4_4_0, \
call_14_4_4_1, \
call_14_4_4_2, \
call_14_4_4_3, \
call_14_4_4_4, \
call_14_4_4_5, \
call_14_4_4_6, \
call_14_4_4_7, \
call_14_4_4_8, \
call_14_4_4_9, \
call_14_4_4_10, \
call_14_4_4_11, \
call_14_4_4_12, \
call_14_4_4_13, \
call_14_4_4_14, \
call_14_4_4_15, \
call_14_4_5_0, \
call_14_4_5_1, \
call_14_4_5_2, \
call_14_4_5_3, \
call_14_4_5_4, \
call_14_4_5_5, \
call_14_4_5_6, \
call_14_4_5_7, \
call_14_4_5_8, \
call_14_4_5_9, \
call_14_4_5_10, \
call_14_4_5_11, \
call_14_4_5_12, \
call_14_4_5_13, \
call_14_4_5_14, \
call_14_4_5_15, \
call_14_4_6_0, \
call_14_4_6_1, \
call_14_4_6_2, \
call_14_4_6_3, \
call_14_4_6_4, \
call_14_4_6_5, \
call_14_4_6_6, \
call_14_4_6_7, \
call_14_4_6_8, \
call_14_4_6_9, \
call_14_4_6_10, \
call_14_4_6_11, \
call_14_4_6_12, \
call_14_4_6_13, \
call_14_4_6_14, \
call_14_4_6_15, \
call_14_4_7_0, \
call_14_4_7_1, \
call_14_4_7_2, \
call_14_4_7_3, \
call_14_4_7_4, \
call_14_4_7_5, \
call_14_4_7_6, \
call_14_4_7_7, \
call_14_4_7_8, \
call_14_4_7_9, \
call_14_4_7_10, \
call_14_4_7_11, \
call_14_4_7_12, \
call_14_4_7_13, \
call_14_4_7_14, \
call_14_4_7_15, \
call_14_4_8_0, \
call_14_4_8_1, \
call_14_4_8_2, \
call_14_4_8_3, \
call_14_4_8_4, \
call_14_4_8_5, \
call_14_4_8_6, \
call_14_4_8_7, \
call_14_4_8_8, \
call_14_4_8_9, \
call_14_4_8_10, \
call_14_4_8_11, \
call_14_4_8_12, \
call_14_4_8_13, \
call_14_4_8_14, \
call_14_4_8_15, \
call_14_4_9_0, \
call_14_4_9_1, \
call_14_4_9_2, \
call_14_4_9_3, \
call_14_4_9_4, \
call_14_4_9_5, \
call_14_4_9_6, \
call_14_4_9_7, \
call_14_4_9_8, \
call_14_4_9_9, \
call_14_4_9_10, \
call_14_4_9_11, \
call_14_4_9_12, \
call_14_4_9_13, \
call_14_4_9_14, \
call_14_4_9_15, \
call_14_4_10_0, \
call_14_4_10_1, \
call_14_4_10_2, \
call_14_4_10_3, \
call_14_4_10_4, \
call_14_4_10_5, \
call_14_4_10_6, \
call_14_4_10_7, \
call_14_4_10_8, \
call_14_4_10_9, \
call_14_4_10_10, \
call_14_4_10_11, \
call_14_4_10_12, \
call_14_4_10_13, \
call_14_4_10_14, \
call_14_4_10_15, \
call_14_4_11_0, \
call_14_4_11_1, \
call_14_4_11_2, \
call_14_4_11_3, \
call_14_4_11_4, \
call_14_4_11_5, \
call_14_4_11_6, \
call_14_4_11_7, \
call_14_4_11_8, \
call_14_4_11_9, \
call_14_4_11_10, \
call_14_4_11_11, \
call_14_4_11_12, \
call_14_4_11_13, \
call_14_4_11_14, \
call_14_4_11_15, \
call_14_4_12_0, \
call_14_4_12_1, \
call_14_4_12_2, \
call_14_4_12_3, \
call_14_4_12_4, \
call_14_4_12_5, \
call_14_4_12_6, \
call_14_4_12_7, \
call_14_4_12_8, \
call_14_4_12_9, \
call_14_4_12_10, \
call_14_4_12_11, \
call_14_4_12_12, \
call_14_4_12_13, \
call_14_4_12_14, \
call_14_4_12_15, \
call_14_4_13_0, \
call_14_4_13_1, \
call_14_4_13_2, \
call_14_4_13_3, \
call_14_4_13_4, \
call_14_4_13_5, \
call_14_4_13_6, \
call_14_4_13_7, \
call_14_4_13_8, \
call_14_4_13_9, \
call_14_4_13_10, \
call_14_4_13_11, \
call_14_4_13_12, \
call_14_4_13_13, \
call_14_4_13_14, \
call_14_4_13_15, \
call_14_4_14_0, \
call_14_4_14_1, \
call_14_4_14_2, \
call_14_4_14_3, \
call_14_4_14_4, \
call_14_4_14_5, \
call_14_4_14_6, \
call_14_4_14_7, \
call_14_4_14_8, \
call_14_4_14_9, \
call_14_4_14_10, \
call_14_4_14_11, \
call_14_4_14_12, \
call_14_4_14_13, \
call_14_4_14_14, \
call_14_4_14_15, \
call_14_4_15_0, \
call_14_4_15_1, \
call_14_4_15_2, \
call_14_4_15_3, \
call_14_4_15_4, \
call_14_4_15_5, \
call_14_4_15_6, \
call_14_4_15_7, \
call_14_4_15_8, \
call_14_4_15_9, \
call_14_4_15_10, \
call_14_4_15_11, \
call_14_4_15_12, \
call_14_4_15_13, \
call_14_4_15_14, \
call_14_4_15_15, \
call_14_5_0_0, \
call_14_5_0_1, \
call_14_5_0_2, \
call_14_5_0_3, \
call_14_5_0_4, \
call_14_5_0_5, \
call_14_5_0_6, \
call_14_5_0_7, \
call_14_5_0_8, \
call_14_5_0_9, \
call_14_5_0_10, \
call_14_5_0_11, \
call_14_5_0_12, \
call_14_5_0_13, \
call_14_5_0_14, \
call_14_5_0_15, \
call_14_5_1_0, \
call_14_5_1_1, \
call_14_5_1_2, \
call_14_5_1_3, \
call_14_5_1_4, \
call_14_5_1_5, \
call_14_5_1_6, \
call_14_5_1_7, \
call_14_5_1_8, \
call_14_5_1_9, \
call_14_5_1_10, \
call_14_5_1_11, \
call_14_5_1_12, \
call_14_5_1_13, \
call_14_5_1_14, \
call_14_5_1_15, \
call_14_5_2_0, \
call_14_5_2_1, \
call_14_5_2_2, \
call_14_5_2_3, \
call_14_5_2_4, \
call_14_5_2_5, \
call_14_5_2_6, \
call_14_5_2_7, \
call_14_5_2_8, \
call_14_5_2_9, \
call_14_5_2_10, \
call_14_5_2_11, \
call_14_5_2_12, \
call_14_5_2_13, \
call_14_5_2_14, \
call_14_5_2_15, \
call_14_5_3_0, \
call_14_5_3_1, \
call_14_5_3_2, \
call_14_5_3_3, \
call_14_5_3_4, \
call_14_5_3_5, \
call_14_5_3_6, \
call_14_5_3_7, \
call_14_5_3_8, \
call_14_5_3_9, \
call_14_5_3_10, \
call_14_5_3_11, \
call_14_5_3_12, \
call_14_5_3_13, \
call_14_5_3_14, \
call_14_5_3_15, \
call_14_5_4_0, \
call_14_5_4_1, \
call_14_5_4_2, \
call_14_5_4_3, \
call_14_5_4_4, \
call_14_5_4_5, \
call_14_5_4_6, \
call_14_5_4_7, \
call_14_5_4_8, \
call_14_5_4_9, \
call_14_5_4_10, \
call_14_5_4_11, \
call_14_5_4_12, \
call_14_5_4_13, \
call_14_5_4_14, \
call_14_5_4_15, \
call_14_5_5_0, \
call_14_5_5_1, \
call_14_5_5_2, \
call_14_5_5_3, \
call_14_5_5_4, \
call_14_5_5_5, \
call_14_5_5_6, \
call_14_5_5_7, \
call_14_5_5_8, \
call_14_5_5_9, \
call_14_5_5_10, \
call_14_5_5_11, \
call_14_5_5_12, \
call_14_5_5_13, \
call_14_5_5_14, \
call_14_5_5_15, \
call_14_5_6_0, \
call_14_5_6_1, \
call_14_5_6_2, \
call_14_5_6_3, \
call_14_5_6_4, \
call_14_5_6_5, \
call_14_5_6_6, \
call_14_5_6_7, \
call_14_5_6_8, \
call_14_5_6_9, \
call_14_5_6_10, \
call_14_5_6_11, \
call_14_5_6_12, \
call_14_5_6_13, \
call_14_5_6_14, \
call_14_5_6_15, \
call_14_5_7_0, \
call_14_5_7_1, \
call_14_5_7_2, \
call_14_5_7_3, \
call_14_5_7_4, \
call_14_5_7_5, \
call_14_5_7_6, \
call_14_5_7_7, \
call_14_5_7_8, \
call_14_5_7_9, \
call_14_5_7_10, \
call_14_5_7_11, \
call_14_5_7_12, \
call_14_5_7_13, \
call_14_5_7_14, \
call_14_5_7_15, \
call_14_5_8_0, \
call_14_5_8_1, \
call_14_5_8_2, \
call_14_5_8_3, \
call_14_5_8_4, \
call_14_5_8_5, \
call_14_5_8_6, \
call_14_5_8_7, \
call_14_5_8_8, \
call_14_5_8_9, \
call_14_5_8_10, \
call_14_5_8_11, \
call_14_5_8_12, \
call_14_5_8_13, \
call_14_5_8_14, \
call_14_5_8_15, \
call_14_5_9_0, \
call_14_5_9_1, \
call_14_5_9_2, \
call_14_5_9_3, \
call_14_5_9_4, \
call_14_5_9_5, \
call_14_5_9_6, \
call_14_5_9_7, \
call_14_5_9_8, \
call_14_5_9_9, \
call_14_5_9_10, \
call_14_5_9_11, \
call_14_5_9_12, \
call_14_5_9_13, \
call_14_5_9_14, \
call_14_5_9_15, \
call_14_5_10_0, \
call_14_5_10_1, \
call_14_5_10_2, \
call_14_5_10_3, \
call_14_5_10_4, \
call_14_5_10_5, \
call_14_5_10_6, \
call_14_5_10_7, \
call_14_5_10_8, \
call_14_5_10_9, \
call_14_5_10_10, \
call_14_5_10_11, \
call_14_5_10_12, \
call_14_5_10_13, \
call_14_5_10_14, \
call_14_5_10_15, \
call_14_5_11_0, \
call_14_5_11_1, \
call_14_5_11_2, \
call_14_5_11_3, \
call_14_5_11_4, \
call_14_5_11_5, \
call_14_5_11_6, \
call_14_5_11_7, \
call_14_5_11_8, \
call_14_5_11_9, \
call_14_5_11_10, \
call_14_5_11_11, \
call_14_5_11_12, \
call_14_5_11_13, \
call_14_5_11_14, \
call_14_5_11_15, \
call_14_5_12_0, \
call_14_5_12_1, \
call_14_5_12_2, \
call_14_5_12_3, \
call_14_5_12_4, \
call_14_5_12_5, \
call_14_5_12_6, \
call_14_5_12_7, \
call_14_5_12_8, \
call_14_5_12_9, \
call_14_5_12_10, \
call_14_5_12_11, \
call_14_5_12_12, \
call_14_5_12_13, \
call_14_5_12_14, \
call_14_5_12_15, \
call_14_5_13_0, \
call_14_5_13_1, \
call_14_5_13_2, \
call_14_5_13_3, \
call_14_5_13_4, \
call_14_5_13_5, \
call_14_5_13_6, \
call_14_5_13_7, \
call_14_5_13_8, \
call_14_5_13_9, \
call_14_5_13_10, \
call_14_5_13_11, \
call_14_5_13_12, \
call_14_5_13_13, \
call_14_5_13_14, \
call_14_5_13_15, \
call_14_5_14_0, \
call_14_5_14_1, \
call_14_5_14_2, \
call_14_5_14_3, \
call_14_5_14_4, \
call_14_5_14_5, \
call_14_5_14_6, \
call_14_5_14_7, \
call_14_5_14_8, \
call_14_5_14_9, \
call_14_5_14_10, \
call_14_5_14_11, \
call_14_5_14_12, \
call_14_5_14_13, \
call_14_5_14_14, \
call_14_5_14_15, \
call_14_5_15_0, \
call_14_5_15_1, \
call_14_5_15_2, \
call_14_5_15_3, \
call_14_5_15_4, \
call_14_5_15_5, \
call_14_5_15_6, \
call_14_5_15_7, \
call_14_5_15_8, \
call_14_5_15_9, \
call_14_5_15_10, \
call_14_5_15_11, \
call_14_5_15_12, \
call_14_5_15_13, \
call_14_5_15_14, \
call_14_5_15_15, \
call_14_6_0_0, \
call_14_6_0_1, \
call_14_6_0_2, \
call_14_6_0_3, \
call_14_6_0_4, \
call_14_6_0_5, \
call_14_6_0_6, \
call_14_6_0_7, \
call_14_6_0_8, \
call_14_6_0_9, \
call_14_6_0_10, \
call_14_6_0_11, \
call_14_6_0_12, \
call_14_6_0_13, \
call_14_6_0_14, \
call_14_6_0_15, \
call_14_6_1_0, \
call_14_6_1_1, \
call_14_6_1_2, \
call_14_6_1_3, \
call_14_6_1_4, \
call_14_6_1_5, \
call_14_6_1_6, \
call_14_6_1_7, \
call_14_6_1_8, \
call_14_6_1_9, \
call_14_6_1_10, \
call_14_6_1_11, \
call_14_6_1_12, \
call_14_6_1_13, \
call_14_6_1_14, \
call_14_6_1_15, \
call_14_6_2_0, \
call_14_6_2_1, \
call_14_6_2_2, \
call_14_6_2_3, \
call_14_6_2_4, \
call_14_6_2_5, \
call_14_6_2_6, \
call_14_6_2_7, \
call_14_6_2_8, \
call_14_6_2_9, \
call_14_6_2_10, \
call_14_6_2_11, \
call_14_6_2_12, \
call_14_6_2_13, \
call_14_6_2_14, \
call_14_6_2_15, \
call_14_6_3_0, \
call_14_6_3_1, \
call_14_6_3_2, \
call_14_6_3_3, \
call_14_6_3_4, \
call_14_6_3_5, \
call_14_6_3_6, \
call_14_6_3_7, \
call_14_6_3_8, \
call_14_6_3_9, \
call_14_6_3_10, \
call_14_6_3_11, \
call_14_6_3_12, \
call_14_6_3_13, \
call_14_6_3_14, \
call_14_6_3_15, \
call_14_6_4_0, \
call_14_6_4_1, \
call_14_6_4_2, \
call_14_6_4_3, \
call_14_6_4_4, \
call_14_6_4_5, \
call_14_6_4_6, \
call_14_6_4_7, \
call_14_6_4_8, \
call_14_6_4_9, \
call_14_6_4_10, \
call_14_6_4_11, \
call_14_6_4_12, \
call_14_6_4_13, \
call_14_6_4_14, \
call_14_6_4_15, \
call_14_6_5_0, \
call_14_6_5_1, \
call_14_6_5_2, \
call_14_6_5_3, \
call_14_6_5_4, \
call_14_6_5_5, \
call_14_6_5_6, \
call_14_6_5_7, \
call_14_6_5_8, \
call_14_6_5_9, \
call_14_6_5_10, \
call_14_6_5_11, \
call_14_6_5_12, \
call_14_6_5_13, \
call_14_6_5_14, \
call_14_6_5_15, \
call_14_6_6_0, \
call_14_6_6_1, \
call_14_6_6_2, \
call_14_6_6_3, \
call_14_6_6_4, \
call_14_6_6_5, \
call_14_6_6_6, \
call_14_6_6_7, \
call_14_6_6_8, \
call_14_6_6_9, \
call_14_6_6_10, \
call_14_6_6_11, \
call_14_6_6_12, \
call_14_6_6_13, \
call_14_6_6_14, \
call_14_6_6_15, \
call_14_6_7_0, \
call_14_6_7_1, \
call_14_6_7_2, \
call_14_6_7_3, \
call_14_6_7_4, \
call_14_6_7_5, \
call_14_6_7_6, \
call_14_6_7_7, \
call_14_6_7_8, \
call_14_6_7_9, \
call_14_6_7_10, \
call_14_6_7_11, \
call_14_6_7_12, \
call_14_6_7_13, \
call_14_6_7_14, \
call_14_6_7_15, \
call_14_6_8_0, \
call_14_6_8_1, \
call_14_6_8_2, \
call_14_6_8_3, \
call_14_6_8_4, \
call_14_6_8_5, \
call_14_6_8_6, \
call_14_6_8_7, \
call_14_6_8_8, \
call_14_6_8_9, \
call_14_6_8_10, \
call_14_6_8_11, \
call_14_6_8_12, \
call_14_6_8_13, \
call_14_6_8_14, \
call_14_6_8_15, \
call_14_6_9_0, \
call_14_6_9_1, \
call_14_6_9_2, \
call_14_6_9_3, \
call_14_6_9_4, \
call_14_6_9_5, \
call_14_6_9_6, \
call_14_6_9_7, \
call_14_6_9_8, \
call_14_6_9_9, \
call_14_6_9_10, \
call_14_6_9_11, \
call_14_6_9_12, \
call_14_6_9_13, \
call_14_6_9_14, \
call_14_6_9_15, \
call_14_6_10_0, \
call_14_6_10_1, \
call_14_6_10_2, \
call_14_6_10_3, \
call_14_6_10_4, \
call_14_6_10_5, \
call_14_6_10_6, \
call_14_6_10_7, \
call_14_6_10_8, \
call_14_6_10_9, \
call_14_6_10_10, \
call_14_6_10_11, \
call_14_6_10_12, \
call_14_6_10_13, \
call_14_6_10_14, \
call_14_6_10_15, \
call_14_6_11_0, \
call_14_6_11_1, \
call_14_6_11_2, \
call_14_6_11_3, \
call_14_6_11_4, \
call_14_6_11_5, \
call_14_6_11_6, \
call_14_6_11_7, \
call_14_6_11_8, \
call_14_6_11_9, \
call_14_6_11_10, \
call_14_6_11_11, \
call_14_6_11_12, \
call_14_6_11_13, \
call_14_6_11_14, \
call_14_6_11_15, \
call_14_6_12_0, \
call_14_6_12_1, \
call_14_6_12_2, \
call_14_6_12_3, \
call_14_6_12_4, \
call_14_6_12_5, \
call_14_6_12_6, \
call_14_6_12_7, \
call_14_6_12_8, \
call_14_6_12_9, \
call_14_6_12_10, \
call_14_6_12_11, \
call_14_6_12_12, \
call_14_6_12_13, \
call_14_6_12_14, \
call_14_6_12_15, \
call_14_6_13_0, \
call_14_6_13_1, \
call_14_6_13_2, \
call_14_6_13_3, \
call_14_6_13_4, \
call_14_6_13_5, \
call_14_6_13_6, \
call_14_6_13_7, \
call_14_6_13_8, \
call_14_6_13_9, \
call_14_6_13_10, \
call_14_6_13_11, \
call_14_6_13_12, \
call_14_6_13_13, \
call_14_6_13_14, \
call_14_6_13_15, \
call_14_6_14_0, \
call_14_6_14_1, \
call_14_6_14_2, \
call_14_6_14_3, \
call_14_6_14_4, \
call_14_6_14_5, \
call_14_6_14_6, \
call_14_6_14_7, \
call_14_6_14_8, \
call_14_6_14_9, \
call_14_6_14_10, \
call_14_6_14_11, \
call_14_6_14_12, \
call_14_6_14_13, \
call_14_6_14_14, \
call_14_6_14_15, \
call_14_6_15_0, \
call_14_6_15_1, \
call_14_6_15_2, \
call_14_6_15_3, \
call_14_6_15_4, \
call_14_6_15_5, \
call_14_6_15_6, \
call_14_6_15_7, \
call_14_6_15_8, \
call_14_6_15_9, \
call_14_6_15_10, \
call_14_6_15_11, \
call_14_6_15_12, \
call_14_6_15_13, \
call_14_6_15_14, \
call_14_6_15_15, \
call_14_7_0_0, \
call_14_7_0_1, \
call_14_7_0_2, \
call_14_7_0_3, \
call_14_7_0_4, \
call_14_7_0_5, \
call_14_7_0_6, \
call_14_7_0_7, \
call_14_7_0_8, \
call_14_7_0_9, \
call_14_7_0_10, \
call_14_7_0_11, \
call_14_7_0_12, \
call_14_7_0_13, \
call_14_7_0_14, \
call_14_7_0_15, \
call_14_7_1_0, \
call_14_7_1_1, \
call_14_7_1_2, \
call_14_7_1_3, \
call_14_7_1_4, \
call_14_7_1_5, \
call_14_7_1_6, \
call_14_7_1_7, \
call_14_7_1_8, \
call_14_7_1_9, \
call_14_7_1_10, \
call_14_7_1_11, \
call_14_7_1_12, \
call_14_7_1_13, \
call_14_7_1_14, \
call_14_7_1_15, \
call_14_7_2_0, \
call_14_7_2_1, \
call_14_7_2_2, \
call_14_7_2_3, \
call_14_7_2_4, \
call_14_7_2_5, \
call_14_7_2_6, \
call_14_7_2_7, \
call_14_7_2_8, \
call_14_7_2_9, \
call_14_7_2_10, \
call_14_7_2_11, \
call_14_7_2_12, \
call_14_7_2_13, \
call_14_7_2_14, \
call_14_7_2_15, \
call_14_7_3_0, \
call_14_7_3_1, \
call_14_7_3_2, \
call_14_7_3_3, \
call_14_7_3_4, \
call_14_7_3_5, \
call_14_7_3_6, \
call_14_7_3_7, \
call_14_7_3_8, \
call_14_7_3_9, \
call_14_7_3_10, \
call_14_7_3_11, \
call_14_7_3_12, \
call_14_7_3_13, \
call_14_7_3_14, \
call_14_7_3_15, \
call_14_7_4_0, \
call_14_7_4_1, \
call_14_7_4_2, \
call_14_7_4_3, \
call_14_7_4_4, \
call_14_7_4_5, \
call_14_7_4_6, \
call_14_7_4_7, \
call_14_7_4_8, \
call_14_7_4_9, \
call_14_7_4_10, \
call_14_7_4_11, \
call_14_7_4_12, \
call_14_7_4_13, \
call_14_7_4_14, \
call_14_7_4_15, \
call_14_7_5_0, \
call_14_7_5_1, \
call_14_7_5_2, \
call_14_7_5_3, \
call_14_7_5_4, \
call_14_7_5_5, \
call_14_7_5_6, \
call_14_7_5_7, \
call_14_7_5_8, \
call_14_7_5_9, \
call_14_7_5_10, \
call_14_7_5_11, \
call_14_7_5_12, \
call_14_7_5_13, \
call_14_7_5_14, \
call_14_7_5_15, \
call_14_7_6_0, \
call_14_7_6_1, \
call_14_7_6_2, \
call_14_7_6_3, \
call_14_7_6_4, \
call_14_7_6_5, \
call_14_7_6_6, \
call_14_7_6_7, \
call_14_7_6_8, \
call_14_7_6_9, \
call_14_7_6_10, \
call_14_7_6_11, \
call_14_7_6_12, \
call_14_7_6_13, \
call_14_7_6_14, \
call_14_7_6_15, \
call_14_7_7_0, \
call_14_7_7_1, \
call_14_7_7_2, \
call_14_7_7_3, \
call_14_7_7_4, \
call_14_7_7_5, \
call_14_7_7_6, \
call_14_7_7_7, \
call_14_7_7_8, \
call_14_7_7_9, \
call_14_7_7_10, \
call_14_7_7_11, \
call_14_7_7_12, \
call_14_7_7_13, \
call_14_7_7_14, \
call_14_7_7_15, \
call_14_7_8_0, \
call_14_7_8_1, \
call_14_7_8_2, \
call_14_7_8_3, \
call_14_7_8_4, \
call_14_7_8_5, \
call_14_7_8_6, \
call_14_7_8_7, \
call_14_7_8_8, \
call_14_7_8_9, \
call_14_7_8_10, \
call_14_7_8_11, \
call_14_7_8_12, \
call_14_7_8_13, \
call_14_7_8_14, \
call_14_7_8_15, \
call_14_7_9_0, \
call_14_7_9_1, \
call_14_7_9_2, \
call_14_7_9_3, \
call_14_7_9_4, \
call_14_7_9_5, \
call_14_7_9_6, \
call_14_7_9_7, \
call_14_7_9_8, \
call_14_7_9_9, \
call_14_7_9_10, \
call_14_7_9_11, \
call_14_7_9_12, \
call_14_7_9_13, \
call_14_7_9_14, \
call_14_7_9_15, \
call_14_7_10_0, \
call_14_7_10_1, \
call_14_7_10_2, \
call_14_7_10_3, \
call_14_7_10_4, \
call_14_7_10_5, \
call_14_7_10_6, \
call_14_7_10_7, \
call_14_7_10_8, \
call_14_7_10_9, \
call_14_7_10_10, \
call_14_7_10_11, \
call_14_7_10_12, \
call_14_7_10_13, \
call_14_7_10_14, \
call_14_7_10_15, \
call_14_7_11_0, \
call_14_7_11_1, \
call_14_7_11_2, \
call_14_7_11_3, \
call_14_7_11_4, \
call_14_7_11_5, \
call_14_7_11_6, \
call_14_7_11_7, \
call_14_7_11_8, \
call_14_7_11_9, \
call_14_7_11_10, \
call_14_7_11_11, \
call_14_7_11_12, \
call_14_7_11_13, \
call_14_7_11_14, \
call_14_7_11_15, \
call_14_7_12_0, \
call_14_7_12_1, \
call_14_7_12_2, \
call_14_7_12_3, \
call_14_7_12_4, \
call_14_7_12_5, \
call_14_7_12_6, \
call_14_7_12_7, \
call_14_7_12_8, \
call_14_7_12_9, \
call_14_7_12_10, \
call_14_7_12_11, \
call_14_7_12_12, \
call_14_7_12_13, \
call_14_7_12_14, \
call_14_7_12_15, \
call_14_7_13_0, \
call_14_7_13_1, \
call_14_7_13_2, \
call_14_7_13_3, \
call_14_7_13_4, \
call_14_7_13_5, \
call_14_7_13_6, \
call_14_7_13_7, \
call_14_7_13_8, \
call_14_7_13_9, \
call_14_7_13_10, \
call_14_7_13_11, \
call_14_7_13_12, \
call_14_7_13_13, \
call_14_7_13_14, \
call_14_7_13_15, \
call_14_7_14_0, \
call_14_7_14_1, \
call_14_7_14_2, \
call_14_7_14_3, \
call_14_7_14_4, \
call_14_7_14_5, \
call_14_7_14_6, \
call_14_7_14_7, \
call_14_7_14_8, \
call_14_7_14_9, \
call_14_7_14_10, \
call_14_7_14_11, \
call_14_7_14_12, \
call_14_7_14_13, \
call_14_7_14_14, \
call_14_7_14_15, \
call_14_7_15_0, \
call_14_7_15_1, \
call_14_7_15_2, \
call_14_7_15_3, \
call_14_7_15_4, \
call_14_7_15_5, \
call_14_7_15_6, \
call_14_7_15_7, \
call_14_7_15_8, \
call_14_7_15_9, \
call_14_7_15_10, \
call_14_7_15_11, \
call_14_7_15_12, \
call_14_7_15_13, \
call_14_7_15_14, \
call_14_7_15_15, \
call_14_8_0_0, \
call_14_8_0_1, \
call_14_8_0_2, \
call_14_8_0_3, \
call_14_8_0_4, \
call_14_8_0_5, \
call_14_8_0_6, \
call_14_8_0_7, \
call_14_8_0_8, \
call_14_8_0_9, \
call_14_8_0_10, \
call_14_8_0_11, \
call_14_8_0_12, \
call_14_8_0_13, \
call_14_8_0_14, \
call_14_8_0_15, \
call_14_8_1_0, \
call_14_8_1_1, \
call_14_8_1_2, \
call_14_8_1_3, \
call_14_8_1_4, \
call_14_8_1_5, \
call_14_8_1_6, \
call_14_8_1_7, \
call_14_8_1_8, \
call_14_8_1_9, \
call_14_8_1_10, \
call_14_8_1_11, \
call_14_8_1_12, \
call_14_8_1_13, \
call_14_8_1_14, \
call_14_8_1_15, \
call_14_8_2_0, \
call_14_8_2_1, \
call_14_8_2_2, \
call_14_8_2_3, \
call_14_8_2_4, \
call_14_8_2_5, \
call_14_8_2_6, \
call_14_8_2_7, \
call_14_8_2_8, \
call_14_8_2_9, \
call_14_8_2_10, \
call_14_8_2_11, \
call_14_8_2_12, \
call_14_8_2_13, \
call_14_8_2_14, \
call_14_8_2_15, \
call_14_8_3_0, \
call_14_8_3_1, \
call_14_8_3_2, \
call_14_8_3_3, \
call_14_8_3_4, \
call_14_8_3_5, \
call_14_8_3_6, \
call_14_8_3_7, \
call_14_8_3_8, \
call_14_8_3_9, \
call_14_8_3_10, \
call_14_8_3_11, \
call_14_8_3_12, \
call_14_8_3_13, \
call_14_8_3_14, \
call_14_8_3_15, \
call_14_8_4_0, \
call_14_8_4_1, \
call_14_8_4_2, \
call_14_8_4_3, \
call_14_8_4_4, \
call_14_8_4_5, \
call_14_8_4_6, \
call_14_8_4_7, \
call_14_8_4_8, \
call_14_8_4_9, \
call_14_8_4_10, \
call_14_8_4_11, \
call_14_8_4_12, \
call_14_8_4_13, \
call_14_8_4_14, \
call_14_8_4_15, \
call_14_8_5_0, \
call_14_8_5_1, \
call_14_8_5_2, \
call_14_8_5_3, \
call_14_8_5_4, \
call_14_8_5_5, \
call_14_8_5_6, \
call_14_8_5_7, \
call_14_8_5_8, \
call_14_8_5_9, \
call_14_8_5_10, \
call_14_8_5_11, \
call_14_8_5_12, \
call_14_8_5_13, \
call_14_8_5_14, \
call_14_8_5_15, \
call_14_8_6_0, \
call_14_8_6_1, \
call_14_8_6_2, \
call_14_8_6_3, \
call_14_8_6_4, \
call_14_8_6_5, \
call_14_8_6_6, \
call_14_8_6_7, \
call_14_8_6_8, \
call_14_8_6_9, \
call_14_8_6_10, \
call_14_8_6_11, \
call_14_8_6_12, \
call_14_8_6_13, \
call_14_8_6_14, \
call_14_8_6_15, \
call_14_8_7_0, \
call_14_8_7_1, \
call_14_8_7_2, \
call_14_8_7_3, \
call_14_8_7_4, \
call_14_8_7_5, \
call_14_8_7_6, \
call_14_8_7_7, \
call_14_8_7_8, \
call_14_8_7_9, \
call_14_8_7_10, \
call_14_8_7_11, \
call_14_8_7_12, \
call_14_8_7_13, \
call_14_8_7_14, \
call_14_8_7_15, \
call_14_8_8_0, \
call_14_8_8_1, \
call_14_8_8_2, \
call_14_8_8_3, \
call_14_8_8_4, \
call_14_8_8_5, \
call_14_8_8_6, \
call_14_8_8_7, \
call_14_8_8_8, \
call_14_8_8_9, \
call_14_8_8_10, \
call_14_8_8_11, \
call_14_8_8_12, \
call_14_8_8_13, \
call_14_8_8_14, \
call_14_8_8_15, \
call_14_8_9_0, \
call_14_8_9_1, \
call_14_8_9_2, \
call_14_8_9_3, \
call_14_8_9_4, \
call_14_8_9_5, \
call_14_8_9_6, \
call_14_8_9_7, \
call_14_8_9_8, \
call_14_8_9_9, \
call_14_8_9_10, \
call_14_8_9_11, \
call_14_8_9_12, \
call_14_8_9_13, \
call_14_8_9_14, \
call_14_8_9_15, \
call_14_8_10_0, \
call_14_8_10_1, \
call_14_8_10_2, \
call_14_8_10_3, \
call_14_8_10_4, \
call_14_8_10_5, \
call_14_8_10_6, \
call_14_8_10_7, \
call_14_8_10_8, \
call_14_8_10_9, \
call_14_8_10_10, \
call_14_8_10_11, \
call_14_8_10_12, \
call_14_8_10_13, \
call_14_8_10_14, \
call_14_8_10_15, \
call_14_8_11_0, \
call_14_8_11_1, \
call_14_8_11_2, \
call_14_8_11_3, \
call_14_8_11_4, \
call_14_8_11_5, \
call_14_8_11_6, \
call_14_8_11_7, \
call_14_8_11_8, \
call_14_8_11_9, \
call_14_8_11_10, \
call_14_8_11_11, \
call_14_8_11_12, \
call_14_8_11_13, \
call_14_8_11_14, \
call_14_8_11_15, \
call_14_8_12_0, \
call_14_8_12_1, \
call_14_8_12_2, \
call_14_8_12_3, \
call_14_8_12_4, \
call_14_8_12_5, \
call_14_8_12_6, \
call_14_8_12_7, \
call_14_8_12_8, \
call_14_8_12_9, \
call_14_8_12_10, \
call_14_8_12_11, \
call_14_8_12_12, \
call_14_8_12_13, \
call_14_8_12_14, \
call_14_8_12_15, \
call_14_8_13_0, \
call_14_8_13_1, \
call_14_8_13_2, \
call_14_8_13_3, \
call_14_8_13_4, \
call_14_8_13_5, \
call_14_8_13_6, \
call_14_8_13_7, \
call_14_8_13_8, \
call_14_8_13_9, \
call_14_8_13_10, \
call_14_8_13_11, \
call_14_8_13_12, \
call_14_8_13_13, \
call_14_8_13_14, \
call_14_8_13_15, \
call_14_8_14_0, \
call_14_8_14_1, \
call_14_8_14_2, \
call_14_8_14_3, \
call_14_8_14_4, \
call_14_8_14_5, \
call_14_8_14_6, \
call_14_8_14_7, \
call_14_8_14_8, \
call_14_8_14_9, \
call_14_8_14_10, \
call_14_8_14_11, \
call_14_8_14_12, \
call_14_8_14_13, \
call_14_8_14_14, \
call_14_8_14_15, \
call_14_8_15_0, \
call_14_8_15_1, \
call_14_8_15_2, \
call_14_8_15_3, \
call_14_8_15_4, \
call_14_8_15_5, \
call_14_8_15_6, \
call_14_8_15_7, \
call_14_8_15_8, \
call_14_8_15_9, \
call_14_8_15_10, \
call_14_8_15_11, \
call_14_8_15_12, \
call_14_8_15_13, \
call_14_8_15_14, \
call_14_8_15_15, \
call_14_9_0_0, \
call_14_9_0_1, \
call_14_9_0_2, \
call_14_9_0_3, \
call_14_9_0_4, \
call_14_9_0_5, \
call_14_9_0_6, \
call_14_9_0_7, \
call_14_9_0_8, \
call_14_9_0_9, \
call_14_9_0_10, \
call_14_9_0_11, \
call_14_9_0_12, \
call_14_9_0_13, \
call_14_9_0_14, \
call_14_9_0_15, \
call_14_9_1_0, \
call_14_9_1_1, \
call_14_9_1_2, \
call_14_9_1_3, \
call_14_9_1_4, \
call_14_9_1_5, \
call_14_9_1_6, \
call_14_9_1_7, \
call_14_9_1_8, \
call_14_9_1_9, \
call_14_9_1_10, \
call_14_9_1_11, \
call_14_9_1_12, \
call_14_9_1_13, \
call_14_9_1_14, \
call_14_9_1_15, \
call_14_9_2_0, \
call_14_9_2_1, \
call_14_9_2_2, \
call_14_9_2_3, \
call_14_9_2_4, \
call_14_9_2_5, \
call_14_9_2_6, \
call_14_9_2_7, \
call_14_9_2_8, \
call_14_9_2_9, \
call_14_9_2_10, \
call_14_9_2_11, \
call_14_9_2_12, \
call_14_9_2_13, \
call_14_9_2_14, \
call_14_9_2_15, \
call_14_9_3_0, \
call_14_9_3_1, \
call_14_9_3_2, \
call_14_9_3_3, \
call_14_9_3_4, \
call_14_9_3_5, \
call_14_9_3_6, \
call_14_9_3_7, \
call_14_9_3_8, \
call_14_9_3_9, \
call_14_9_3_10, \
call_14_9_3_11, \
call_14_9_3_12, \
call_14_9_3_13, \
call_14_9_3_14, \
call_14_9_3_15, \
call_14_9_4_0, \
call_14_9_4_1, \
call_14_9_4_2, \
call_14_9_4_3, \
call_14_9_4_4, \
call_14_9_4_5, \
call_14_9_4_6, \
call_14_9_4_7, \
call_14_9_4_8, \
call_14_9_4_9, \
call_14_9_4_10, \
call_14_9_4_11, \
call_14_9_4_12, \
call_14_9_4_13, \
call_14_9_4_14, \
call_14_9_4_15, \
call_14_9_5_0, \
call_14_9_5_1, \
call_14_9_5_2, \
call_14_9_5_3, \
call_14_9_5_4, \
call_14_9_5_5, \
call_14_9_5_6, \
call_14_9_5_7, \
call_14_9_5_8, \
call_14_9_5_9, \
call_14_9_5_10, \
call_14_9_5_11, \
call_14_9_5_12, \
call_14_9_5_13, \
call_14_9_5_14, \
call_14_9_5_15, \
call_14_9_6_0, \
call_14_9_6_1, \
call_14_9_6_2, \
call_14_9_6_3, \
call_14_9_6_4, \
call_14_9_6_5, \
call_14_9_6_6, \
call_14_9_6_7, \
call_14_9_6_8, \
call_14_9_6_9, \
call_14_9_6_10, \
call_14_9_6_11, \
call_14_9_6_12, \
call_14_9_6_13, \
call_14_9_6_14, \
call_14_9_6_15, \
call_14_9_7_0, \
call_14_9_7_1, \
call_14_9_7_2, \
call_14_9_7_3, \
call_14_9_7_4, \
call_14_9_7_5, \
call_14_9_7_6, \
call_14_9_7_7, \
call_14_9_7_8, \
call_14_9_7_9, \
call_14_9_7_10, \
call_14_9_7_11, \
call_14_9_7_12, \
call_14_9_7_13, \
call_14_9_7_14, \
call_14_9_7_15, \
call_14_9_8_0, \
call_14_9_8_1, \
call_14_9_8_2, \
call_14_9_8_3, \
call_14_9_8_4, \
call_14_9_8_5, \
call_14_9_8_6, \
call_14_9_8_7, \
call_14_9_8_8, \
call_14_9_8_9, \
call_14_9_8_10, \
call_14_9_8_11, \
call_14_9_8_12, \
call_14_9_8_13, \
call_14_9_8_14, \
call_14_9_8_15, \
call_14_9_9_0, \
call_14_9_9_1, \
call_14_9_9_2, \
call_14_9_9_3, \
call_14_9_9_4, \
call_14_9_9_5, \
call_14_9_9_6, \
call_14_9_9_7, \
call_14_9_9_8, \
call_14_9_9_9, \
call_14_9_9_10, \
call_14_9_9_11, \
call_14_9_9_12, \
call_14_9_9_13, \
call_14_9_9_14, \
call_14_9_9_15, \
call_14_9_10_0, \
call_14_9_10_1, \
call_14_9_10_2, \
call_14_9_10_3, \
call_14_9_10_4, \
call_14_9_10_5, \
call_14_9_10_6, \
call_14_9_10_7, \
call_14_9_10_8, \
call_14_9_10_9, \
call_14_9_10_10, \
call_14_9_10_11, \
call_14_9_10_12, \
call_14_9_10_13, \
call_14_9_10_14, \
call_14_9_10_15, \
call_14_9_11_0, \
call_14_9_11_1, \
call_14_9_11_2, \
call_14_9_11_3, \
call_14_9_11_4, \
call_14_9_11_5, \
call_14_9_11_6, \
call_14_9_11_7, \
call_14_9_11_8, \
call_14_9_11_9, \
call_14_9_11_10, \
call_14_9_11_11, \
call_14_9_11_12, \
call_14_9_11_13, \
call_14_9_11_14, \
call_14_9_11_15, \
call_14_9_12_0, \
call_14_9_12_1, \
call_14_9_12_2, \
call_14_9_12_3, \
call_14_9_12_4, \
call_14_9_12_5, \
call_14_9_12_6, \
call_14_9_12_7, \
call_14_9_12_8, \
call_14_9_12_9, \
call_14_9_12_10, \
call_14_9_12_11, \
call_14_9_12_12, \
call_14_9_12_13, \
call_14_9_12_14, \
call_14_9_12_15, \
call_14_9_13_0, \
call_14_9_13_1, \
call_14_9_13_2, \
call_14_9_13_3, \
call_14_9_13_4, \
call_14_9_13_5, \
call_14_9_13_6, \
call_14_9_13_7, \
call_14_9_13_8, \
call_14_9_13_9, \
call_14_9_13_10, \
call_14_9_13_11, \
call_14_9_13_12, \
call_14_9_13_13, \
call_14_9_13_14, \
call_14_9_13_15, \
call_14_9_14_0, \
call_14_9_14_1, \
call_14_9_14_2, \
call_14_9_14_3, \
call_14_9_14_4, \
call_14_9_14_5, \
call_14_9_14_6, \
call_14_9_14_7, \
call_14_9_14_8, \
call_14_9_14_9, \
call_14_9_14_10, \
call_14_9_14_11, \
call_14_9_14_12, \
call_14_9_14_13, \
call_14_9_14_14, \
call_14_9_14_15, \
call_14_9_15_0, \
call_14_9_15_1, \
call_14_9_15_2, \
call_14_9_15_3, \
call_14_9_15_4, \
call_14_9_15_5, \
call_14_9_15_6, \
call_14_9_15_7, \
call_14_9_15_8, \
call_14_9_15_9, \
call_14_9_15_10, \
call_14_9_15_11, \
call_14_9_15_12, \
call_14_9_15_13, \
call_14_9_15_14, \
call_14_9_15_15, \
call_14_10_0_0, \
call_14_10_0_1, \
call_14_10_0_2, \
call_14_10_0_3, \
call_14_10_0_4, \
call_14_10_0_5, \
call_14_10_0_6, \
call_14_10_0_7, \
call_14_10_0_8, \
call_14_10_0_9, \
call_14_10_0_10, \
call_14_10_0_11, \
call_14_10_0_12, \
call_14_10_0_13, \
call_14_10_0_14, \
call_14_10_0_15, \
call_14_10_1_0, \
call_14_10_1_1, \
call_14_10_1_2, \
call_14_10_1_3, \
call_14_10_1_4, \
call_14_10_1_5, \
call_14_10_1_6, \
call_14_10_1_7, \
call_14_10_1_8, \
call_14_10_1_9, \
call_14_10_1_10, \
call_14_10_1_11, \
call_14_10_1_12, \
call_14_10_1_13, \
call_14_10_1_14, \
call_14_10_1_15, \
call_14_10_2_0, \
call_14_10_2_1, \
call_14_10_2_2, \
call_14_10_2_3, \
call_14_10_2_4, \
call_14_10_2_5, \
call_14_10_2_6, \
call_14_10_2_7, \
call_14_10_2_8, \
call_14_10_2_9, \
call_14_10_2_10, \
call_14_10_2_11, \
call_14_10_2_12, \
call_14_10_2_13, \
call_14_10_2_14, \
call_14_10_2_15, \
call_14_10_3_0, \
call_14_10_3_1, \
call_14_10_3_2, \
call_14_10_3_3, \
call_14_10_3_4, \
call_14_10_3_5, \
call_14_10_3_6, \
call_14_10_3_7, \
call_14_10_3_8, \
call_14_10_3_9, \
call_14_10_3_10, \
call_14_10_3_11, \
call_14_10_3_12, \
call_14_10_3_13, \
call_14_10_3_14, \
call_14_10_3_15, \
call_14_10_4_0, \
call_14_10_4_1, \
call_14_10_4_2, \
call_14_10_4_3, \
call_14_10_4_4, \
call_14_10_4_5, \
call_14_10_4_6, \
call_14_10_4_7, \
call_14_10_4_8, \
call_14_10_4_9, \
call_14_10_4_10, \
call_14_10_4_11, \
call_14_10_4_12, \
call_14_10_4_13, \
call_14_10_4_14, \
call_14_10_4_15, \
call_14_10_5_0, \
call_14_10_5_1, \
call_14_10_5_2, \
call_14_10_5_3, \
call_14_10_5_4, \
call_14_10_5_5, \
call_14_10_5_6, \
call_14_10_5_7, \
call_14_10_5_8, \
call_14_10_5_9, \
call_14_10_5_10, \
call_14_10_5_11, \
call_14_10_5_12, \
call_14_10_5_13, \
call_14_10_5_14, \
call_14_10_5_15, \
call_14_10_6_0, \
call_14_10_6_1, \
call_14_10_6_2, \
call_14_10_6_3, \
call_14_10_6_4, \
call_14_10_6_5, \
call_14_10_6_6, \
call_14_10_6_7, \
call_14_10_6_8, \
call_14_10_6_9, \
call_14_10_6_10, \
call_14_10_6_11, \
call_14_10_6_12, \
call_14_10_6_13, \
call_14_10_6_14, \
call_14_10_6_15, \
call_14_10_7_0, \
call_14_10_7_1, \
call_14_10_7_2, \
call_14_10_7_3, \
call_14_10_7_4, \
call_14_10_7_5, \
call_14_10_7_6, \
call_14_10_7_7, \
call_14_10_7_8, \
call_14_10_7_9, \
call_14_10_7_10, \
call_14_10_7_11, \
call_14_10_7_12, \
call_14_10_7_13, \
call_14_10_7_14, \
call_14_10_7_15, \
call_14_10_8_0, \
call_14_10_8_1, \
call_14_10_8_2, \
call_14_10_8_3, \
call_14_10_8_4, \
call_14_10_8_5, \
call_14_10_8_6, \
call_14_10_8_7, \
call_14_10_8_8, \
call_14_10_8_9, \
call_14_10_8_10, \
call_14_10_8_11, \
call_14_10_8_12, \
call_14_10_8_13, \
call_14_10_8_14, \
call_14_10_8_15, \
call_14_10_9_0, \
call_14_10_9_1, \
call_14_10_9_2, \
call_14_10_9_3, \
call_14_10_9_4, \
call_14_10_9_5, \
call_14_10_9_6, \
call_14_10_9_7, \
call_14_10_9_8, \
call_14_10_9_9, \
call_14_10_9_10, \
call_14_10_9_11, \
call_14_10_9_12, \
call_14_10_9_13, \
call_14_10_9_14, \
call_14_10_9_15, \
call_14_10_10_0, \
call_14_10_10_1, \
call_14_10_10_2, \
call_14_10_10_3, \
call_14_10_10_4, \
call_14_10_10_5, \
call_14_10_10_6, \
call_14_10_10_7, \
call_14_10_10_8, \
call_14_10_10_9, \
call_14_10_10_10, \
call_14_10_10_11, \
call_14_10_10_12, \
call_14_10_10_13, \
call_14_10_10_14, \
call_14_10_10_15, \
call_14_10_11_0, \
call_14_10_11_1, \
call_14_10_11_2, \
call_14_10_11_3, \
call_14_10_11_4, \
call_14_10_11_5, \
call_14_10_11_6, \
call_14_10_11_7, \
call_14_10_11_8, \
call_14_10_11_9, \
call_14_10_11_10, \
call_14_10_11_11, \
call_14_10_11_12, \
call_14_10_11_13, \
call_14_10_11_14, \
call_14_10_11_15, \
call_14_10_12_0, \
call_14_10_12_1, \
call_14_10_12_2, \
call_14_10_12_3, \
call_14_10_12_4, \
call_14_10_12_5, \
call_14_10_12_6, \
call_14_10_12_7, \
call_14_10_12_8, \
call_14_10_12_9, \
call_14_10_12_10, \
call_14_10_12_11, \
call_14_10_12_12, \
call_14_10_12_13, \
call_14_10_12_14, \
call_14_10_12_15, \
call_14_10_13_0, \
call_14_10_13_1, \
call_14_10_13_2, \
call_14_10_13_3, \
call_14_10_13_4, \
call_14_10_13_5, \
call_14_10_13_6, \
call_14_10_13_7, \
call_14_10_13_8, \
call_14_10_13_9, \
call_14_10_13_10, \
call_14_10_13_11, \
call_14_10_13_12, \
call_14_10_13_13, \
call_14_10_13_14, \
call_14_10_13_15, \
call_14_10_14_0, \
call_14_10_14_1, \
call_14_10_14_2, \
call_14_10_14_3, \
call_14_10_14_4, \
call_14_10_14_5, \
call_14_10_14_6, \
call_14_10_14_7, \
call_14_10_14_8, \
call_14_10_14_9, \
call_14_10_14_10, \
call_14_10_14_11, \
call_14_10_14_12, \
call_14_10_14_13, \
call_14_10_14_14, \
call_14_10_14_15, \
call_14_10_15_0, \
call_14_10_15_1, \
call_14_10_15_2, \
call_14_10_15_3, \
call_14_10_15_4, \
call_14_10_15_5, \
call_14_10_15_6, \
call_14_10_15_7, \
call_14_10_15_8, \
call_14_10_15_9, \
call_14_10_15_10, \
call_14_10_15_11, \
call_14_10_15_12, \
call_14_10_15_13, \
call_14_10_15_14, \
call_14_10_15_15, \
call_14_11_0_0, \
call_14_11_0_1, \
call_14_11_0_2, \
call_14_11_0_3, \
call_14_11_0_4, \
call_14_11_0_5, \
call_14_11_0_6, \
call_14_11_0_7, \
call_14_11_0_8, \
call_14_11_0_9, \
call_14_11_0_10, \
call_14_11_0_11, \
call_14_11_0_12, \
call_14_11_0_13, \
call_14_11_0_14, \
call_14_11_0_15, \
call_14_11_1_0, \
call_14_11_1_1, \
call_14_11_1_2, \
call_14_11_1_3, \
call_14_11_1_4, \
call_14_11_1_5, \
call_14_11_1_6, \
call_14_11_1_7, \
call_14_11_1_8, \
call_14_11_1_9, \
call_14_11_1_10, \
call_14_11_1_11, \
call_14_11_1_12, \
call_14_11_1_13, \
call_14_11_1_14, \
call_14_11_1_15, \
call_14_11_2_0, \
call_14_11_2_1, \
call_14_11_2_2, \
call_14_11_2_3, \
call_14_11_2_4, \
call_14_11_2_5, \
call_14_11_2_6, \
call_14_11_2_7, \
call_14_11_2_8, \
call_14_11_2_9, \
call_14_11_2_10, \
call_14_11_2_11, \
call_14_11_2_12, \
call_14_11_2_13, \
call_14_11_2_14, \
call_14_11_2_15, \
call_14_11_3_0, \
call_14_11_3_1, \
call_14_11_3_2, \
call_14_11_3_3, \
call_14_11_3_4, \
call_14_11_3_5, \
call_14_11_3_6, \
call_14_11_3_7, \
call_14_11_3_8, \
call_14_11_3_9, \
call_14_11_3_10, \
call_14_11_3_11, \
call_14_11_3_12, \
call_14_11_3_13, \
call_14_11_3_14, \
call_14_11_3_15, \
call_14_11_4_0, \
call_14_11_4_1, \
call_14_11_4_2, \
call_14_11_4_3, \
call_14_11_4_4, \
call_14_11_4_5, \
call_14_11_4_6, \
call_14_11_4_7, \
call_14_11_4_8, \
call_14_11_4_9, \
call_14_11_4_10, \
call_14_11_4_11, \
call_14_11_4_12, \
call_14_11_4_13, \
call_14_11_4_14, \
call_14_11_4_15, \
call_14_11_5_0, \
call_14_11_5_1, \
call_14_11_5_2, \
call_14_11_5_3, \
call_14_11_5_4, \
call_14_11_5_5, \
call_14_11_5_6, \
call_14_11_5_7, \
call_14_11_5_8, \
call_14_11_5_9, \
call_14_11_5_10, \
call_14_11_5_11, \
call_14_11_5_12, \
call_14_11_5_13, \
call_14_11_5_14, \
call_14_11_5_15, \
call_14_11_6_0, \
call_14_11_6_1, \
call_14_11_6_2, \
call_14_11_6_3, \
call_14_11_6_4, \
call_14_11_6_5, \
call_14_11_6_6, \
call_14_11_6_7, \
call_14_11_6_8, \
call_14_11_6_9, \
call_14_11_6_10, \
call_14_11_6_11, \
call_14_11_6_12, \
call_14_11_6_13, \
call_14_11_6_14, \
call_14_11_6_15, \
call_14_11_7_0, \
call_14_11_7_1, \
call_14_11_7_2, \
call_14_11_7_3, \
call_14_11_7_4, \
call_14_11_7_5, \
call_14_11_7_6, \
call_14_11_7_7, \
call_14_11_7_8, \
call_14_11_7_9, \
call_14_11_7_10, \
call_14_11_7_11, \
call_14_11_7_12, \
call_14_11_7_13, \
call_14_11_7_14, \
call_14_11_7_15, \
call_14_11_8_0, \
call_14_11_8_1, \
call_14_11_8_2, \
call_14_11_8_3, \
call_14_11_8_4, \
call_14_11_8_5, \
call_14_11_8_6, \
call_14_11_8_7, \
call_14_11_8_8, \
call_14_11_8_9, \
call_14_11_8_10, \
call_14_11_8_11, \
call_14_11_8_12, \
call_14_11_8_13, \
call_14_11_8_14, \
call_14_11_8_15, \
call_14_11_9_0, \
call_14_11_9_1, \
call_14_11_9_2, \
call_14_11_9_3, \
call_14_11_9_4, \
call_14_11_9_5, \
call_14_11_9_6, \
call_14_11_9_7, \
call_14_11_9_8, \
call_14_11_9_9, \
call_14_11_9_10, \
call_14_11_9_11, \
call_14_11_9_12, \
call_14_11_9_13, \
call_14_11_9_14, \
call_14_11_9_15, \
call_14_11_10_0, \
call_14_11_10_1, \
call_14_11_10_2, \
call_14_11_10_3, \
call_14_11_10_4, \
call_14_11_10_5, \
call_14_11_10_6, \
call_14_11_10_7, \
call_14_11_10_8, \
call_14_11_10_9, \
call_14_11_10_10, \
call_14_11_10_11, \
call_14_11_10_12, \
call_14_11_10_13, \
call_14_11_10_14, \
call_14_11_10_15, \
call_14_11_11_0, \
call_14_11_11_1, \
call_14_11_11_2, \
call_14_11_11_3, \
call_14_11_11_4, \
call_14_11_11_5, \
call_14_11_11_6, \
call_14_11_11_7, \
call_14_11_11_8, \
call_14_11_11_9, \
call_14_11_11_10, \
call_14_11_11_11, \
call_14_11_11_12, \
call_14_11_11_13, \
call_14_11_11_14, \
call_14_11_11_15, \
call_14_11_12_0, \
call_14_11_12_1, \
call_14_11_12_2, \
call_14_11_12_3, \
call_14_11_12_4, \
call_14_11_12_5, \
call_14_11_12_6, \
call_14_11_12_7, \
call_14_11_12_8, \
call_14_11_12_9, \
call_14_11_12_10, \
call_14_11_12_11, \
call_14_11_12_12, \
call_14_11_12_13, \
call_14_11_12_14, \
call_14_11_12_15, \
call_14_11_13_0, \
call_14_11_13_1, \
call_14_11_13_2, \
call_14_11_13_3, \
call_14_11_13_4, \
call_14_11_13_5, \
call_14_11_13_6, \
call_14_11_13_7, \
call_14_11_13_8, \
call_14_11_13_9, \
call_14_11_13_10, \
call_14_11_13_11, \
call_14_11_13_12, \
call_14_11_13_13, \
call_14_11_13_14, \
call_14_11_13_15, \
call_14_11_14_0, \
call_14_11_14_1, \
call_14_11_14_2, \
call_14_11_14_3, \
call_14_11_14_4, \
call_14_11_14_5, \
call_14_11_14_6, \
call_14_11_14_7, \
call_14_11_14_8, \
call_14_11_14_9, \
call_14_11_14_10, \
call_14_11_14_11, \
call_14_11_14_12, \
call_14_11_14_13, \
call_14_11_14_14, \
call_14_11_14_15, \
call_14_11_15_0, \
call_14_11_15_1, \
call_14_11_15_2, \
call_14_11_15_3, \
call_14_11_15_4, \
call_14_11_15_5, \
call_14_11_15_6, \
call_14_11_15_7, \
call_14_11_15_8, \
call_14_11_15_9, \
call_14_11_15_10, \
call_14_11_15_11, \
call_14_11_15_12, \
call_14_11_15_13, \
call_14_11_15_14, \
call_14_11_15_15, \
call_14_12_0_0, \
call_14_12_0_1, \
call_14_12_0_2, \
call_14_12_0_3, \
call_14_12_0_4, \
call_14_12_0_5, \
call_14_12_0_6, \
call_14_12_0_7, \
call_14_12_0_8, \
call_14_12_0_9, \
call_14_12_0_10, \
call_14_12_0_11, \
call_14_12_0_12, \
call_14_12_0_13, \
call_14_12_0_14, \
call_14_12_0_15, \
call_14_12_1_0, \
call_14_12_1_1, \
call_14_12_1_2, \
call_14_12_1_3, \
call_14_12_1_4, \
call_14_12_1_5, \
call_14_12_1_6, \
call_14_12_1_7, \
call_14_12_1_8, \
call_14_12_1_9, \
call_14_12_1_10, \
call_14_12_1_11, \
call_14_12_1_12, \
call_14_12_1_13, \
call_14_12_1_14, \
call_14_12_1_15, \
call_14_12_2_0, \
call_14_12_2_1, \
call_14_12_2_2, \
call_14_12_2_3, \
call_14_12_2_4, \
call_14_12_2_5, \
call_14_12_2_6, \
call_14_12_2_7, \
call_14_12_2_8, \
call_14_12_2_9, \
call_14_12_2_10, \
call_14_12_2_11, \
call_14_12_2_12, \
call_14_12_2_13, \
call_14_12_2_14, \
call_14_12_2_15, \
call_14_12_3_0, \
call_14_12_3_1, \
call_14_12_3_2, \
call_14_12_3_3, \
call_14_12_3_4, \
call_14_12_3_5, \
call_14_12_3_6, \
call_14_12_3_7, \
call_14_12_3_8, \
call_14_12_3_9, \
call_14_12_3_10, \
call_14_12_3_11, \
call_14_12_3_12, \
call_14_12_3_13, \
call_14_12_3_14, \
call_14_12_3_15, \
call_14_12_4_0, \
call_14_12_4_1, \
call_14_12_4_2, \
call_14_12_4_3, \
call_14_12_4_4, \
call_14_12_4_5, \
call_14_12_4_6, \
call_14_12_4_7, \
call_14_12_4_8, \
call_14_12_4_9, \
call_14_12_4_10, \
call_14_12_4_11, \
call_14_12_4_12, \
call_14_12_4_13, \
call_14_12_4_14, \
call_14_12_4_15, \
call_14_12_5_0, \
call_14_12_5_1, \
call_14_12_5_2, \
call_14_12_5_3, \
call_14_12_5_4, \
call_14_12_5_5, \
call_14_12_5_6, \
call_14_12_5_7, \
call_14_12_5_8, \
call_14_12_5_9, \
call_14_12_5_10, \
call_14_12_5_11, \
call_14_12_5_12, \
call_14_12_5_13, \
call_14_12_5_14, \
call_14_12_5_15, \
call_14_12_6_0, \
call_14_12_6_1, \
call_14_12_6_2, \
call_14_12_6_3, \
call_14_12_6_4, \
call_14_12_6_5, \
call_14_12_6_6, \
call_14_12_6_7, \
call_14_12_6_8, \
call_14_12_6_9, \
call_14_12_6_10, \
call_14_12_6_11, \
call_14_12_6_12, \
call_14_12_6_13, \
call_14_12_6_14, \
call_14_12_6_15, \
call_14_12_7_0, \
call_14_12_7_1, \
call_14_12_7_2, \
call_14_12_7_3, \
call_14_12_7_4, \
call_14_12_7_5, \
call_14_12_7_6, \
call_14_12_7_7, \
call_14_12_7_8, \
call_14_12_7_9, \
call_14_12_7_10, \
call_14_12_7_11, \
call_14_12_7_12, \
call_14_12_7_13, \
call_14_12_7_14, \
call_14_12_7_15, \
call_14_12_8_0, \
call_14_12_8_1, \
call_14_12_8_2, \
call_14_12_8_3, \
call_14_12_8_4, \
call_14_12_8_5, \
call_14_12_8_6, \
call_14_12_8_7, \
call_14_12_8_8, \
call_14_12_8_9, \
call_14_12_8_10, \
call_14_12_8_11, \
call_14_12_8_12, \
call_14_12_8_13, \
call_14_12_8_14, \
call_14_12_8_15, \
call_14_12_9_0, \
call_14_12_9_1, \
call_14_12_9_2, \
call_14_12_9_3, \
call_14_12_9_4, \
call_14_12_9_5, \
call_14_12_9_6, \
call_14_12_9_7, \
call_14_12_9_8, \
call_14_12_9_9, \
call_14_12_9_10, \
call_14_12_9_11, \
call_14_12_9_12, \
call_14_12_9_13, \
call_14_12_9_14, \
call_14_12_9_15, \
call_14_12_10_0, \
call_14_12_10_1, \
call_14_12_10_2, \
call_14_12_10_3, \
call_14_12_10_4, \
call_14_12_10_5, \
call_14_12_10_6, \
call_14_12_10_7, \
call_14_12_10_8, \
call_14_12_10_9, \
call_14_12_10_10, \
call_14_12_10_11, \
call_14_12_10_12, \
call_14_12_10_13, \
call_14_12_10_14, \
call_14_12_10_15, \
call_14_12_11_0, \
call_14_12_11_1, \
call_14_12_11_2, \
call_14_12_11_3, \
call_14_12_11_4, \
call_14_12_11_5, \
call_14_12_11_6, \
call_14_12_11_7, \
call_14_12_11_8, \
call_14_12_11_9, \
call_14_12_11_10, \
call_14_12_11_11, \
call_14_12_11_12, \
call_14_12_11_13, \
call_14_12_11_14, \
call_14_12_11_15, \
call_14_12_12_0, \
call_14_12_12_1, \
call_14_12_12_2, \
call_14_12_12_3, \
call_14_12_12_4, \
call_14_12_12_5, \
call_14_12_12_6, \
call_14_12_12_7, \
call_14_12_12_8, \
call_14_12_12_9, \
call_14_12_12_10, \
call_14_12_12_11, \
call_14_12_12_12, \
call_14_12_12_13, \
call_14_12_12_14, \
call_14_12_12_15, \
call_14_12_13_0, \
call_14_12_13_1, \
call_14_12_13_2, \
call_14_12_13_3, \
call_14_12_13_4, \
call_14_12_13_5, \
call_14_12_13_6, \
call_14_12_13_7, \
call_14_12_13_8, \
call_14_12_13_9, \
call_14_12_13_10, \
call_14_12_13_11, \
call_14_12_13_12, \
call_14_12_13_13, \
call_14_12_13_14, \
call_14_12_13_15, \
call_14_12_14_0, \
call_14_12_14_1, \
call_14_12_14_2, \
call_14_12_14_3, \
call_14_12_14_4, \
call_14_12_14_5, \
call_14_12_14_6, \
call_14_12_14_7, \
call_14_12_14_8, \
call_14_12_14_9, \
call_14_12_14_10, \
call_14_12_14_11, \
call_14_12_14_12, \
call_14_12_14_13, \
call_14_12_14_14, \
call_14_12_14_15, \
call_14_12_15_0, \
call_14_12_15_1, \
call_14_12_15_2, \
call_14_12_15_3, \
call_14_12_15_4, \
call_14_12_15_5, \
call_14_12_15_6, \
call_14_12_15_7, \
call_14_12_15_8, \
call_14_12_15_9, \
call_14_12_15_10, \
call_14_12_15_11, \
call_14_12_15_12, \
call_14_12_15_13, \
call_14_12_15_14, \
call_14_12_15_15, \
call_14_13_0_0, \
call_14_13_0_1, \
call_14_13_0_2, \
call_14_13_0_3, \
call_14_13_0_4, \
call_14_13_0_5, \
call_14_13_0_6, \
call_14_13_0_7, \
call_14_13_0_8, \
call_14_13_0_9, \
call_14_13_0_10, \
call_14_13_0_11, \
call_14_13_0_12, \
call_14_13_0_13, \
call_14_13_0_14, \
call_14_13_0_15, \
call_14_13_1_0, \
call_14_13_1_1, \
call_14_13_1_2, \
call_14_13_1_3, \
call_14_13_1_4, \
call_14_13_1_5, \
call_14_13_1_6, \
call_14_13_1_7, \
call_14_13_1_8, \
call_14_13_1_9, \
call_14_13_1_10, \
call_14_13_1_11, \
call_14_13_1_12, \
call_14_13_1_13, \
call_14_13_1_14, \
call_14_13_1_15, \
call_14_13_2_0, \
call_14_13_2_1, \
call_14_13_2_2, \
call_14_13_2_3, \
call_14_13_2_4, \
call_14_13_2_5, \
call_14_13_2_6, \
call_14_13_2_7, \
call_14_13_2_8, \
call_14_13_2_9, \
call_14_13_2_10, \
call_14_13_2_11, \
call_14_13_2_12, \
call_14_13_2_13, \
call_14_13_2_14, \
call_14_13_2_15, \
call_14_13_3_0, \
call_14_13_3_1, \
call_14_13_3_2, \
call_14_13_3_3, \
call_14_13_3_4, \
call_14_13_3_5, \
call_14_13_3_6, \
call_14_13_3_7, \
call_14_13_3_8, \
call_14_13_3_9, \
call_14_13_3_10, \
call_14_13_3_11, \
call_14_13_3_12, \
call_14_13_3_13, \
call_14_13_3_14, \
call_14_13_3_15, \
call_14_13_4_0, \
call_14_13_4_1, \
call_14_13_4_2, \
call_14_13_4_3, \
call_14_13_4_4, \
call_14_13_4_5, \
call_14_13_4_6, \
call_14_13_4_7, \
call_14_13_4_8, \
call_14_13_4_9, \
call_14_13_4_10, \
call_14_13_4_11, \
call_14_13_4_12, \
call_14_13_4_13, \
call_14_13_4_14, \
call_14_13_4_15, \
call_14_13_5_0, \
call_14_13_5_1, \
call_14_13_5_2, \
call_14_13_5_3, \
call_14_13_5_4, \
call_14_13_5_5, \
call_14_13_5_6, \
call_14_13_5_7, \
call_14_13_5_8, \
call_14_13_5_9, \
call_14_13_5_10, \
call_14_13_5_11, \
call_14_13_5_12, \
call_14_13_5_13, \
call_14_13_5_14, \
call_14_13_5_15, \
call_14_13_6_0, \
call_14_13_6_1, \
call_14_13_6_2, \
call_14_13_6_3, \
call_14_13_6_4, \
call_14_13_6_5, \
call_14_13_6_6, \
call_14_13_6_7, \
call_14_13_6_8, \
call_14_13_6_9, \
call_14_13_6_10, \
call_14_13_6_11, \
call_14_13_6_12, \
call_14_13_6_13, \
call_14_13_6_14, \
call_14_13_6_15, \
call_14_13_7_0, \
call_14_13_7_1, \
call_14_13_7_2, \
call_14_13_7_3, \
call_14_13_7_4, \
call_14_13_7_5, \
call_14_13_7_6, \
call_14_13_7_7, \
call_14_13_7_8, \
call_14_13_7_9, \
call_14_13_7_10, \
call_14_13_7_11, \
call_14_13_7_12, \
call_14_13_7_13, \
call_14_13_7_14, \
call_14_13_7_15, \
call_14_13_8_0, \
call_14_13_8_1, \
call_14_13_8_2, \
call_14_13_8_3, \
call_14_13_8_4, \
call_14_13_8_5, \
call_14_13_8_6, \
call_14_13_8_7, \
call_14_13_8_8, \
call_14_13_8_9, \
call_14_13_8_10, \
call_14_13_8_11, \
call_14_13_8_12, \
call_14_13_8_13, \
call_14_13_8_14, \
call_14_13_8_15, \
call_14_13_9_0, \
call_14_13_9_1, \
call_14_13_9_2, \
call_14_13_9_3, \
call_14_13_9_4, \
call_14_13_9_5, \
call_14_13_9_6, \
call_14_13_9_7, \
call_14_13_9_8, \
call_14_13_9_9, \
call_14_13_9_10, \
call_14_13_9_11, \
call_14_13_9_12, \
call_14_13_9_13, \
call_14_13_9_14, \
call_14_13_9_15, \
call_14_13_10_0, \
call_14_13_10_1, \
call_14_13_10_2, \
call_14_13_10_3, \
call_14_13_10_4, \
call_14_13_10_5, \
call_14_13_10_6, \
call_14_13_10_7, \
call_14_13_10_8, \
call_14_13_10_9, \
call_14_13_10_10, \
call_14_13_10_11, \
call_14_13_10_12, \
call_14_13_10_13, \
call_14_13_10_14, \
call_14_13_10_15, \
call_14_13_11_0, \
call_14_13_11_1, \
call_14_13_11_2, \
call_14_13_11_3, \
call_14_13_11_4, \
call_14_13_11_5, \
call_14_13_11_6, \
call_14_13_11_7, \
call_14_13_11_8, \
call_14_13_11_9, \
call_14_13_11_10, \
call_14_13_11_11, \
call_14_13_11_12, \
call_14_13_11_13, \
call_14_13_11_14, \
call_14_13_11_15, \
call_14_13_12_0, \
call_14_13_12_1, \
call_14_13_12_2, \
call_14_13_12_3, \
call_14_13_12_4, \
call_14_13_12_5, \
call_14_13_12_6, \
call_14_13_12_7, \
call_14_13_12_8, \
call_14_13_12_9, \
call_14_13_12_10, \
call_14_13_12_11, \
call_14_13_12_12, \
call_14_13_12_13, \
call_14_13_12_14, \
call_14_13_12_15, \
call_14_13_13_0, \
call_14_13_13_1, \
call_14_13_13_2, \
call_14_13_13_3, \
call_14_13_13_4, \
call_14_13_13_5, \
call_14_13_13_6, \
call_14_13_13_7, \
call_14_13_13_8, \
call_14_13_13_9, \
call_14_13_13_10, \
call_14_13_13_11, \
call_14_13_13_12, \
call_14_13_13_13, \
call_14_13_13_14, \
call_14_13_13_15, \
call_14_13_14_0, \
call_14_13_14_1, \
call_14_13_14_2, \
call_14_13_14_3, \
call_14_13_14_4, \
call_14_13_14_5, \
call_14_13_14_6, \
call_14_13_14_7, \
call_14_13_14_8, \
call_14_13_14_9, \
call_14_13_14_10, \
call_14_13_14_11, \
call_14_13_14_12, \
call_14_13_14_13, \
call_14_13_14_14, \
call_14_13_14_15, \
call_14_13_15_0, \
call_14_13_15_1, \
call_14_13_15_2, \
call_14_13_15_3, \
call_14_13_15_4, \
call_14_13_15_5, \
call_14_13_15_6, \
call_14_13_15_7, \
call_14_13_15_8, \
call_14_13_15_9, \
call_14_13_15_10, \
call_14_13_15_11, \
call_14_13_15_12, \
call_14_13_15_13, \
call_14_13_15_14, \
call_14_13_15_15, \
call_14_14_0_0, \
call_14_14_0_1, \
call_14_14_0_2, \
call_14_14_0_3, \
call_14_14_0_4, \
call_14_14_0_5, \
call_14_14_0_6, \
call_14_14_0_7, \
call_14_14_0_8, \
call_14_14_0_9, \
call_14_14_0_10, \
call_14_14_0_11, \
call_14_14_0_12, \
call_14_14_0_13, \
call_14_14_0_14, \
call_14_14_0_15, \
call_14_14_1_0, \
call_14_14_1_1, \
call_14_14_1_2, \
call_14_14_1_3, \
call_14_14_1_4, \
call_14_14_1_5, \
call_14_14_1_6, \
call_14_14_1_7, \
call_14_14_1_8, \
call_14_14_1_9, \
call_14_14_1_10, \
call_14_14_1_11, \
call_14_14_1_12, \
call_14_14_1_13, \
call_14_14_1_14, \
call_14_14_1_15, \
call_14_14_2_0, \
call_14_14_2_1, \
call_14_14_2_2, \
call_14_14_2_3, \
call_14_14_2_4, \
call_14_14_2_5, \
call_14_14_2_6, \
call_14_14_2_7, \
call_14_14_2_8, \
call_14_14_2_9, \
call_14_14_2_10, \
call_14_14_2_11, \
call_14_14_2_12, \
call_14_14_2_13, \
call_14_14_2_14, \
call_14_14_2_15, \
call_14_14_3_0, \
call_14_14_3_1, \
call_14_14_3_2, \
call_14_14_3_3, \
call_14_14_3_4, \
call_14_14_3_5, \
call_14_14_3_6, \
call_14_14_3_7, \
call_14_14_3_8, \
call_14_14_3_9, \
call_14_14_3_10, \
call_14_14_3_11, \
call_14_14_3_12, \
call_14_14_3_13, \
call_14_14_3_14, \
call_14_14_3_15, \
call_14_14_4_0, \
call_14_14_4_1, \
call_14_14_4_2, \
call_14_14_4_3, \
call_14_14_4_4, \
call_14_14_4_5, \
call_14_14_4_6, \
call_14_14_4_7, \
call_14_14_4_8, \
call_14_14_4_9, \
call_14_14_4_10, \
call_14_14_4_11, \
call_14_14_4_12, \
call_14_14_4_13, \
call_14_14_4_14, \
call_14_14_4_15, \
call_14_14_5_0, \
call_14_14_5_1, \
call_14_14_5_2, \
call_14_14_5_3, \
call_14_14_5_4, \
call_14_14_5_5, \
call_14_14_5_6, \
call_14_14_5_7, \
call_14_14_5_8, \
call_14_14_5_9, \
call_14_14_5_10, \
call_14_14_5_11, \
call_14_14_5_12, \
call_14_14_5_13, \
call_14_14_5_14, \
call_14_14_5_15, \
call_14_14_6_0, \
call_14_14_6_1, \
call_14_14_6_2, \
call_14_14_6_3, \
call_14_14_6_4, \
call_14_14_6_5, \
call_14_14_6_6, \
call_14_14_6_7, \
call_14_14_6_8, \
call_14_14_6_9, \
call_14_14_6_10, \
call_14_14_6_11, \
call_14_14_6_12, \
call_14_14_6_13, \
call_14_14_6_14, \
call_14_14_6_15, \
call_14_14_7_0, \
call_14_14_7_1, \
call_14_14_7_2, \
call_14_14_7_3, \
call_14_14_7_4, \
call_14_14_7_5, \
call_14_14_7_6, \
call_14_14_7_7, \
call_14_14_7_8, \
call_14_14_7_9, \
call_14_14_7_10, \
call_14_14_7_11, \
call_14_14_7_12, \
call_14_14_7_13, \
call_14_14_7_14, \
call_14_14_7_15, \
call_14_14_8_0, \
call_14_14_8_1, \
call_14_14_8_2, \
call_14_14_8_3, \
call_14_14_8_4, \
call_14_14_8_5, \
call_14_14_8_6, \
call_14_14_8_7, \
call_14_14_8_8, \
call_14_14_8_9, \
call_14_14_8_10, \
call_14_14_8_11, \
call_14_14_8_12, \
call_14_14_8_13, \
call_14_14_8_14, \
call_14_14_8_15, \
call_14_14_9_0, \
call_14_14_9_1, \
call_14_14_9_2, \
call_14_14_9_3, \
call_14_14_9_4, \
call_14_14_9_5, \
call_14_14_9_6, \
call_14_14_9_7, \
call_14_14_9_8, \
call_14_14_9_9, \
call_14_14_9_10, \
call_14_14_9_11, \
call_14_14_9_12, \
call_14_14_9_13, \
call_14_14_9_14, \
call_14_14_9_15, \
call_14_14_10_0, \
call_14_14_10_1, \
call_14_14_10_2, \
call_14_14_10_3, \
call_14_14_10_4, \
call_14_14_10_5, \
call_14_14_10_6, \
call_14_14_10_7, \
call_14_14_10_8, \
call_14_14_10_9, \
call_14_14_10_10, \
call_14_14_10_11, \
call_14_14_10_12, \
call_14_14_10_13, \
call_14_14_10_14, \
call_14_14_10_15, \
call_14_14_11_0, \
call_14_14_11_1, \
call_14_14_11_2, \
call_14_14_11_3, \
call_14_14_11_4, \
call_14_14_11_5, \
call_14_14_11_6, \
call_14_14_11_7, \
call_14_14_11_8, \
call_14_14_11_9, \
call_14_14_11_10, \
call_14_14_11_11, \
call_14_14_11_12, \
call_14_14_11_13, \
call_14_14_11_14, \
call_14_14_11_15, \
call_14_14_12_0, \
call_14_14_12_1, \
call_14_14_12_2, \
call_14_14_12_3, \
call_14_14_12_4, \
call_14_14_12_5, \
call_14_14_12_6, \
call_14_14_12_7, \
call_14_14_12_8, \
call_14_14_12_9, \
call_14_14_12_10, \
call_14_14_12_11, \
call_14_14_12_12, \
call_14_14_12_13, \
call_14_14_12_14, \
call_14_14_12_15, \
call_14_14_13_0, \
call_14_14_13_1, \
call_14_14_13_2, \
call_14_14_13_3, \
call_14_14_13_4, \
call_14_14_13_5, \
call_14_14_13_6, \
call_14_14_13_7, \
call_14_14_13_8, \
call_14_14_13_9, \
call_14_14_13_10, \
call_14_14_13_11, \
call_14_14_13_12, \
call_14_14_13_13, \
call_14_14_13_14, \
call_14_14_13_15, \
call_14_14_14_0, \
call_14_14_14_1, \
call_14_14_14_2, \
call_14_14_14_3, \
call_14_14_14_4, \
call_14_14_14_5, \
call_14_14_14_6, \
call_14_14_14_7, \
call_14_14_14_8, \
call_14_14_14_9, \
call_14_14_14_10, \
call_14_14_14_11, \
call_14_14_14_12, \
call_14_14_14_13, \
call_14_14_14_14, \
call_14_14_14_15, \
call_14_14_15_0, \
call_14_14_15_1, \
call_14_14_15_2, \
call_14_14_15_3, \
call_14_14_15_4, \
call_14_14_15_5, \
call_14_14_15_6, \
call_14_14_15_7, \
call_14_14_15_8, \
call_14_14_15_9, \
call_14_14_15_10, \
call_14_14_15_11, \
call_14_14_15_12, \
call_14_14_15_13, \
call_14_14_15_14, \
call_14_14_15_15, \
call_14_15_0_0, \
call_14_15_0_1, \
call_14_15_0_2, \
call_14_15_0_3, \
call_14_15_0_4, \
call_14_15_0_5, \
call_14_15_0_6, \
call_14_15_0_7, \
call_14_15_0_8, \
call_14_15_0_9, \
call_14_15_0_10, \
call_14_15_0_11, \
call_14_15_0_12, \
call_14_15_0_13, \
call_14_15_0_14, \
call_14_15_0_15, \
call_14_15_1_0, \
call_14_15_1_1, \
call_14_15_1_2, \
call_14_15_1_3, \
call_14_15_1_4, \
call_14_15_1_5, \
call_14_15_1_6, \
call_14_15_1_7, \
call_14_15_1_8, \
call_14_15_1_9, \
call_14_15_1_10, \
call_14_15_1_11, \
call_14_15_1_12, \
call_14_15_1_13, \
call_14_15_1_14, \
call_14_15_1_15, \
call_14_15_2_0, \
call_14_15_2_1, \
call_14_15_2_2, \
call_14_15_2_3, \
call_14_15_2_4, \
call_14_15_2_5, \
call_14_15_2_6, \
call_14_15_2_7, \
call_14_15_2_8, \
call_14_15_2_9, \
call_14_15_2_10, \
call_14_15_2_11, \
call_14_15_2_12, \
call_14_15_2_13, \
call_14_15_2_14, \
call_14_15_2_15, \
call_14_15_3_0, \
call_14_15_3_1, \
call_14_15_3_2, \
call_14_15_3_3, \
call_14_15_3_4, \
call_14_15_3_5, \
call_14_15_3_6, \
call_14_15_3_7, \
call_14_15_3_8, \
call_14_15_3_9, \
call_14_15_3_10, \
call_14_15_3_11, \
call_14_15_3_12, \
call_14_15_3_13, \
call_14_15_3_14, \
call_14_15_3_15, \
call_14_15_4_0, \
call_14_15_4_1, \
call_14_15_4_2, \
call_14_15_4_3, \
call_14_15_4_4, \
call_14_15_4_5, \
call_14_15_4_6, \
call_14_15_4_7, \
call_14_15_4_8, \
call_14_15_4_9, \
call_14_15_4_10, \
call_14_15_4_11, \
call_14_15_4_12, \
call_14_15_4_13, \
call_14_15_4_14, \
call_14_15_4_15, \
call_14_15_5_0, \
call_14_15_5_1, \
call_14_15_5_2, \
call_14_15_5_3, \
call_14_15_5_4, \
call_14_15_5_5, \
call_14_15_5_6, \
call_14_15_5_7, \
call_14_15_5_8, \
call_14_15_5_9, \
call_14_15_5_10, \
call_14_15_5_11, \
call_14_15_5_12, \
call_14_15_5_13, \
call_14_15_5_14, \
call_14_15_5_15, \
call_14_15_6_0, \
call_14_15_6_1, \
call_14_15_6_2, \
call_14_15_6_3, \
call_14_15_6_4, \
call_14_15_6_5, \
call_14_15_6_6, \
call_14_15_6_7, \
call_14_15_6_8, \
call_14_15_6_9, \
call_14_15_6_10, \
call_14_15_6_11, \
call_14_15_6_12, \
call_14_15_6_13, \
call_14_15_6_14, \
call_14_15_6_15, \
call_14_15_7_0, \
call_14_15_7_1, \
call_14_15_7_2, \
call_14_15_7_3, \
call_14_15_7_4, \
call_14_15_7_5, \
call_14_15_7_6, \
call_14_15_7_7, \
call_14_15_7_8, \
call_14_15_7_9, \
call_14_15_7_10, \
call_14_15_7_11, \
call_14_15_7_12, \
call_14_15_7_13, \
call_14_15_7_14, \
call_14_15_7_15, \
call_14_15_8_0, \
call_14_15_8_1, \
call_14_15_8_2, \
call_14_15_8_3, \
call_14_15_8_4, \
call_14_15_8_5, \
call_14_15_8_6, \
call_14_15_8_7, \
call_14_15_8_8, \
call_14_15_8_9, \
call_14_15_8_10, \
call_14_15_8_11, \
call_14_15_8_12, \
call_14_15_8_13, \
call_14_15_8_14, \
call_14_15_8_15, \
call_14_15_9_0, \
call_14_15_9_1, \
call_14_15_9_2, \
call_14_15_9_3, \
call_14_15_9_4, \
call_14_15_9_5, \
call_14_15_9_6, \
call_14_15_9_7, \
call_14_15_9_8, \
call_14_15_9_9, \
call_14_15_9_10, \
call_14_15_9_11, \
call_14_15_9_12, \
call_14_15_9_13, \
call_14_15_9_14, \
call_14_15_9_15, \
call_14_15_10_0, \
call_14_15_10_1, \
call_14_15_10_2, \
call_14_15_10_3, \
call_14_15_10_4, \
call_14_15_10_5, \
call_14_15_10_6, \
call_14_15_10_7, \
call_14_15_10_8, \
call_14_15_10_9, \
call_14_15_10_10, \
call_14_15_10_11, \
call_14_15_10_12, \
call_14_15_10_13, \
call_14_15_10_14, \
call_14_15_10_15, \
call_14_15_11_0, \
call_14_15_11_1, \
call_14_15_11_2, \
call_14_15_11_3, \
call_14_15_11_4, \
call_14_15_11_5, \
call_14_15_11_6, \
call_14_15_11_7, \
call_14_15_11_8, \
call_14_15_11_9, \
call_14_15_11_10, \
call_14_15_11_11, \
call_14_15_11_12, \
call_14_15_11_13, \
call_14_15_11_14, \
call_14_15_11_15, \
call_14_15_12_0, \
call_14_15_12_1, \
call_14_15_12_2, \
call_14_15_12_3, \
call_14_15_12_4, \
call_14_15_12_5, \
call_14_15_12_6, \
call_14_15_12_7, \
call_14_15_12_8, \
call_14_15_12_9, \
call_14_15_12_10, \
call_14_15_12_11, \
call_14_15_12_12, \
call_14_15_12_13, \
call_14_15_12_14, \
call_14_15_12_15, \
call_14_15_13_0, \
call_14_15_13_1, \
call_14_15_13_2, \
call_14_15_13_3, \
call_14_15_13_4, \
call_14_15_13_5, \
call_14_15_13_6, \
call_14_15_13_7, \
call_14_15_13_8, \
call_14_15_13_9, \
call_14_15_13_10, \
call_14_15_13_11, \
call_14_15_13_12, \
call_14_15_13_13, \
call_14_15_13_14, \
call_14_15_13_15, \
call_14_15_14_0, \
call_14_15_14_1, \
call_14_15_14_2, \
call_14_15_14_3, \
call_14_15_14_4, \
call_14_15_14_5, \
call_14_15_14_6, \
call_14_15_14_7, \
call_14_15_14_8, \
call_14_15_14_9, \
call_14_15_14_10, \
call_14_15_14_11, \
call_14_15_14_12, \
call_14_15_14_13, \
call_14_15_14_14, \
call_14_15_14_15, \
call_14_15_15_0, \
call_14_15_15_1, \
call_14_15_15_2, \
call_14_15_15_3, \
call_14_15_15_4, \
call_14_15_15_5, \
call_14_15_15_6, \
call_14_15_15_7, \
call_14_15_15_8, \
call_14_15_15_9, \
call_14_15_15_10, \
call_14_15_15_11, \
call_14_15_15_12, \
call_14_15_15_13, \
call_14_15_15_14, \
call_14_15_15_15, \
call_15_0_0_0, \
call_15_0_0_1, \
call_15_0_0_2, \
call_15_0_0_3, \
call_15_0_0_4, \
call_15_0_0_5, \
call_15_0_0_6, \
call_15_0_0_7, \
call_15_0_0_8, \
call_15_0_0_9, \
call_15_0_0_10, \
call_15_0_0_11, \
call_15_0_0_12, \
call_15_0_0_13, \
call_15_0_0_14, \
call_15_0_0_15, \
call_15_0_1_0, \
call_15_0_1_1, \
call_15_0_1_2, \
call_15_0_1_3, \
call_15_0_1_4, \
call_15_0_1_5, \
call_15_0_1_6, \
call_15_0_1_7, \
call_15_0_1_8, \
call_15_0_1_9, \
call_15_0_1_10, \
call_15_0_1_11, \
call_15_0_1_12, \
call_15_0_1_13, \
call_15_0_1_14, \
call_15_0_1_15, \
call_15_0_2_0, \
call_15_0_2_1, \
call_15_0_2_2, \
call_15_0_2_3, \
call_15_0_2_4, \
call_15_0_2_5, \
call_15_0_2_6, \
call_15_0_2_7, \
call_15_0_2_8, \
call_15_0_2_9, \
call_15_0_2_10, \
call_15_0_2_11, \
call_15_0_2_12, \
call_15_0_2_13, \
call_15_0_2_14, \
call_15_0_2_15, \
call_15_0_3_0, \
call_15_0_3_1, \
call_15_0_3_2, \
call_15_0_3_3, \
call_15_0_3_4, \
call_15_0_3_5, \
call_15_0_3_6, \
call_15_0_3_7, \
call_15_0_3_8, \
call_15_0_3_9, \
call_15_0_3_10, \
call_15_0_3_11, \
call_15_0_3_12, \
call_15_0_3_13, \
call_15_0_3_14, \
call_15_0_3_15, \
call_15_0_4_0, \
call_15_0_4_1, \
call_15_0_4_2, \
call_15_0_4_3, \
call_15_0_4_4, \
call_15_0_4_5, \
call_15_0_4_6, \
call_15_0_4_7, \
call_15_0_4_8, \
call_15_0_4_9, \
call_15_0_4_10, \
call_15_0_4_11, \
call_15_0_4_12, \
call_15_0_4_13, \
call_15_0_4_14, \
call_15_0_4_15, \
call_15_0_5_0, \
call_15_0_5_1, \
call_15_0_5_2, \
call_15_0_5_3, \
call_15_0_5_4, \
call_15_0_5_5, \
call_15_0_5_6, \
call_15_0_5_7, \
call_15_0_5_8, \
call_15_0_5_9, \
call_15_0_5_10, \
call_15_0_5_11, \
call_15_0_5_12, \
call_15_0_5_13, \
call_15_0_5_14, \
call_15_0_5_15, \
call_15_0_6_0, \
call_15_0_6_1, \
call_15_0_6_2, \
call_15_0_6_3, \
call_15_0_6_4, \
call_15_0_6_5, \
call_15_0_6_6, \
call_15_0_6_7, \
call_15_0_6_8, \
call_15_0_6_9, \
call_15_0_6_10, \
call_15_0_6_11, \
call_15_0_6_12, \
call_15_0_6_13, \
call_15_0_6_14, \
call_15_0_6_15, \
call_15_0_7_0, \
call_15_0_7_1, \
call_15_0_7_2, \
call_15_0_7_3, \
call_15_0_7_4, \
call_15_0_7_5, \
call_15_0_7_6, \
call_15_0_7_7, \
call_15_0_7_8, \
call_15_0_7_9, \
call_15_0_7_10, \
call_15_0_7_11, \
call_15_0_7_12, \
call_15_0_7_13, \
call_15_0_7_14, \
call_15_0_7_15, \
call_15_0_8_0, \
call_15_0_8_1, \
call_15_0_8_2, \
call_15_0_8_3, \
call_15_0_8_4, \
call_15_0_8_5, \
call_15_0_8_6, \
call_15_0_8_7, \
call_15_0_8_8, \
call_15_0_8_9, \
call_15_0_8_10, \
call_15_0_8_11, \
call_15_0_8_12, \
call_15_0_8_13, \
call_15_0_8_14, \
call_15_0_8_15, \
call_15_0_9_0, \
call_15_0_9_1, \
call_15_0_9_2, \
call_15_0_9_3, \
call_15_0_9_4, \
call_15_0_9_5, \
call_15_0_9_6, \
call_15_0_9_7, \
call_15_0_9_8, \
call_15_0_9_9, \
call_15_0_9_10, \
call_15_0_9_11, \
call_15_0_9_12, \
call_15_0_9_13, \
call_15_0_9_14, \
call_15_0_9_15, \
call_15_0_10_0, \
call_15_0_10_1, \
call_15_0_10_2, \
call_15_0_10_3, \
call_15_0_10_4, \
call_15_0_10_5, \
call_15_0_10_6, \
call_15_0_10_7, \
call_15_0_10_8, \
call_15_0_10_9, \
call_15_0_10_10, \
call_15_0_10_11, \
call_15_0_10_12, \
call_15_0_10_13, \
call_15_0_10_14, \
call_15_0_10_15, \
call_15_0_11_0, \
call_15_0_11_1, \
call_15_0_11_2, \
call_15_0_11_3, \
call_15_0_11_4, \
call_15_0_11_5, \
call_15_0_11_6, \
call_15_0_11_7, \
call_15_0_11_8, \
call_15_0_11_9, \
call_15_0_11_10, \
call_15_0_11_11, \
call_15_0_11_12, \
call_15_0_11_13, \
call_15_0_11_14, \
call_15_0_11_15, \
call_15_0_12_0, \
call_15_0_12_1, \
call_15_0_12_2, \
call_15_0_12_3, \
call_15_0_12_4, \
call_15_0_12_5, \
call_15_0_12_6, \
call_15_0_12_7, \
call_15_0_12_8, \
call_15_0_12_9, \
call_15_0_12_10, \
call_15_0_12_11, \
call_15_0_12_12, \
call_15_0_12_13, \
call_15_0_12_14, \
call_15_0_12_15, \
call_15_0_13_0, \
call_15_0_13_1, \
call_15_0_13_2, \
call_15_0_13_3, \
call_15_0_13_4, \
call_15_0_13_5, \
call_15_0_13_6, \
call_15_0_13_7, \
call_15_0_13_8, \
call_15_0_13_9, \
call_15_0_13_10, \
call_15_0_13_11, \
call_15_0_13_12, \
call_15_0_13_13, \
call_15_0_13_14, \
call_15_0_13_15, \
call_15_0_14_0, \
call_15_0_14_1, \
call_15_0_14_2, \
call_15_0_14_3, \
call_15_0_14_4, \
call_15_0_14_5, \
call_15_0_14_6, \
call_15_0_14_7, \
call_15_0_14_8, \
call_15_0_14_9, \
call_15_0_14_10, \
call_15_0_14_11, \
call_15_0_14_12, \
call_15_0_14_13, \
call_15_0_14_14, \
call_15_0_14_15, \
call_15_0_15_0, \
call_15_0_15_1, \
call_15_0_15_2, \
call_15_0_15_3, \
call_15_0_15_4, \
call_15_0_15_5, \
call_15_0_15_6, \
call_15_0_15_7, \
call_15_0_15_8, \
call_15_0_15_9, \
call_15_0_15_10, \
call_15_0_15_11, \
call_15_0_15_12, \
call_15_0_15_13, \
call_15_0_15_14, \
call_15_0_15_15, \
call_15_1_0_0, \
call_15_1_0_1, \
call_15_1_0_2, \
call_15_1_0_3, \
call_15_1_0_4, \
call_15_1_0_5, \
call_15_1_0_6, \
call_15_1_0_7, \
call_15_1_0_8, \
call_15_1_0_9, \
call_15_1_0_10, \
call_15_1_0_11, \
call_15_1_0_12, \
call_15_1_0_13, \
call_15_1_0_14, \
call_15_1_0_15, \
call_15_1_1_0, \
call_15_1_1_1, \
call_15_1_1_2, \
call_15_1_1_3, \
call_15_1_1_4, \
call_15_1_1_5, \
call_15_1_1_6, \
call_15_1_1_7, \
call_15_1_1_8, \
call_15_1_1_9, \
call_15_1_1_10, \
call_15_1_1_11, \
call_15_1_1_12, \
call_15_1_1_13, \
call_15_1_1_14, \
call_15_1_1_15, \
call_15_1_2_0, \
call_15_1_2_1, \
call_15_1_2_2, \
call_15_1_2_3, \
call_15_1_2_4, \
call_15_1_2_5, \
call_15_1_2_6, \
call_15_1_2_7, \
call_15_1_2_8, \
call_15_1_2_9, \
call_15_1_2_10, \
call_15_1_2_11, \
call_15_1_2_12, \
call_15_1_2_13, \
call_15_1_2_14, \
call_15_1_2_15, \
call_15_1_3_0, \
call_15_1_3_1, \
call_15_1_3_2, \
call_15_1_3_3, \
call_15_1_3_4, \
call_15_1_3_5, \
call_15_1_3_6, \
call_15_1_3_7, \
call_15_1_3_8, \
call_15_1_3_9, \
call_15_1_3_10, \
call_15_1_3_11, \
call_15_1_3_12, \
call_15_1_3_13, \
call_15_1_3_14, \
call_15_1_3_15, \
call_15_1_4_0, \
call_15_1_4_1, \
call_15_1_4_2, \
call_15_1_4_3, \
call_15_1_4_4, \
call_15_1_4_5, \
call_15_1_4_6, \
call_15_1_4_7, \
call_15_1_4_8, \
call_15_1_4_9, \
call_15_1_4_10, \
call_15_1_4_11, \
call_15_1_4_12, \
call_15_1_4_13, \
call_15_1_4_14, \
call_15_1_4_15, \
call_15_1_5_0, \
call_15_1_5_1, \
call_15_1_5_2, \
call_15_1_5_3, \
call_15_1_5_4, \
call_15_1_5_5, \
call_15_1_5_6, \
call_15_1_5_7, \
call_15_1_5_8, \
call_15_1_5_9, \
call_15_1_5_10, \
call_15_1_5_11, \
call_15_1_5_12, \
call_15_1_5_13, \
call_15_1_5_14, \
call_15_1_5_15, \
call_15_1_6_0, \
call_15_1_6_1, \
call_15_1_6_2, \
call_15_1_6_3, \
call_15_1_6_4, \
call_15_1_6_5, \
call_15_1_6_6, \
call_15_1_6_7, \
call_15_1_6_8, \
call_15_1_6_9, \
call_15_1_6_10, \
call_15_1_6_11, \
call_15_1_6_12, \
call_15_1_6_13, \
call_15_1_6_14, \
call_15_1_6_15, \
call_15_1_7_0, \
call_15_1_7_1, \
call_15_1_7_2, \
call_15_1_7_3, \
call_15_1_7_4, \
call_15_1_7_5, \
call_15_1_7_6, \
call_15_1_7_7, \
call_15_1_7_8, \
call_15_1_7_9, \
call_15_1_7_10, \
call_15_1_7_11, \
call_15_1_7_12, \
call_15_1_7_13, \
call_15_1_7_14, \
call_15_1_7_15, \
call_15_1_8_0, \
call_15_1_8_1, \
call_15_1_8_2, \
call_15_1_8_3, \
call_15_1_8_4, \
call_15_1_8_5, \
call_15_1_8_6, \
call_15_1_8_7, \
call_15_1_8_8, \
call_15_1_8_9, \
call_15_1_8_10, \
call_15_1_8_11, \
call_15_1_8_12, \
call_15_1_8_13, \
call_15_1_8_14, \
call_15_1_8_15, \
call_15_1_9_0, \
call_15_1_9_1, \
call_15_1_9_2, \
call_15_1_9_3, \
call_15_1_9_4, \
call_15_1_9_5, \
call_15_1_9_6, \
call_15_1_9_7, \
call_15_1_9_8, \
call_15_1_9_9, \
call_15_1_9_10, \
call_15_1_9_11, \
call_15_1_9_12, \
call_15_1_9_13, \
call_15_1_9_14, \
call_15_1_9_15, \
call_15_1_10_0, \
call_15_1_10_1, \
call_15_1_10_2, \
call_15_1_10_3, \
call_15_1_10_4, \
call_15_1_10_5, \
call_15_1_10_6, \
call_15_1_10_7, \
call_15_1_10_8, \
call_15_1_10_9, \
call_15_1_10_10, \
call_15_1_10_11, \
call_15_1_10_12, \
call_15_1_10_13, \
call_15_1_10_14, \
call_15_1_10_15, \
call_15_1_11_0, \
call_15_1_11_1, \
call_15_1_11_2, \
call_15_1_11_3, \
call_15_1_11_4, \
call_15_1_11_5, \
call_15_1_11_6, \
call_15_1_11_7, \
call_15_1_11_8, \
call_15_1_11_9, \
call_15_1_11_10, \
call_15_1_11_11, \
call_15_1_11_12, \
call_15_1_11_13, \
call_15_1_11_14, \
call_15_1_11_15, \
call_15_1_12_0, \
call_15_1_12_1, \
call_15_1_12_2, \
call_15_1_12_3, \
call_15_1_12_4, \
call_15_1_12_5, \
call_15_1_12_6, \
call_15_1_12_7, \
call_15_1_12_8, \
call_15_1_12_9, \
call_15_1_12_10, \
call_15_1_12_11, \
call_15_1_12_12, \
call_15_1_12_13, \
call_15_1_12_14, \
call_15_1_12_15, \
call_15_1_13_0, \
call_15_1_13_1, \
call_15_1_13_2, \
call_15_1_13_3, \
call_15_1_13_4, \
call_15_1_13_5, \
call_15_1_13_6, \
call_15_1_13_7, \
call_15_1_13_8, \
call_15_1_13_9, \
call_15_1_13_10, \
call_15_1_13_11, \
call_15_1_13_12, \
call_15_1_13_13, \
call_15_1_13_14, \
call_15_1_13_15, \
call_15_1_14_0, \
call_15_1_14_1, \
call_15_1_14_2, \
call_15_1_14_3, \
call_15_1_14_4, \
call_15_1_14_5, \
call_15_1_14_6, \
call_15_1_14_7, \
call_15_1_14_8, \
call_15_1_14_9, \
call_15_1_14_10, \
call_15_1_14_11, \
call_15_1_14_12, \
call_15_1_14_13, \
call_15_1_14_14, \
call_15_1_14_15, \
call_15_1_15_0, \
call_15_1_15_1, \
call_15_1_15_2, \
call_15_1_15_3, \
call_15_1_15_4, \
call_15_1_15_5, \
call_15_1_15_6, \
call_15_1_15_7, \
call_15_1_15_8, \
call_15_1_15_9, \
call_15_1_15_10, \
call_15_1_15_11, \
call_15_1_15_12, \
call_15_1_15_13, \
call_15_1_15_14, \
call_15_1_15_15, \
call_15_2_0_0, \
call_15_2_0_1, \
call_15_2_0_2, \
call_15_2_0_3, \
call_15_2_0_4, \
call_15_2_0_5, \
call_15_2_0_6, \
call_15_2_0_7, \
call_15_2_0_8, \
call_15_2_0_9, \
call_15_2_0_10, \
call_15_2_0_11, \
call_15_2_0_12, \
call_15_2_0_13, \
call_15_2_0_14, \
call_15_2_0_15, \
call_15_2_1_0, \
call_15_2_1_1, \
call_15_2_1_2, \
call_15_2_1_3, \
call_15_2_1_4, \
call_15_2_1_5, \
call_15_2_1_6, \
call_15_2_1_7, \
call_15_2_1_8, \
call_15_2_1_9, \
call_15_2_1_10, \
call_15_2_1_11, \
call_15_2_1_12, \
call_15_2_1_13, \
call_15_2_1_14, \
call_15_2_1_15, \
call_15_2_2_0, \
call_15_2_2_1, \
call_15_2_2_2, \
call_15_2_2_3, \
call_15_2_2_4, \
call_15_2_2_5, \
call_15_2_2_6, \
call_15_2_2_7, \
call_15_2_2_8, \
call_15_2_2_9, \
call_15_2_2_10, \
call_15_2_2_11, \
call_15_2_2_12, \
call_15_2_2_13, \
call_15_2_2_14, \
call_15_2_2_15, \
call_15_2_3_0, \
call_15_2_3_1, \
call_15_2_3_2, \
call_15_2_3_3, \
call_15_2_3_4, \
call_15_2_3_5, \
call_15_2_3_6, \
call_15_2_3_7, \
call_15_2_3_8, \
call_15_2_3_9, \
call_15_2_3_10, \
call_15_2_3_11, \
call_15_2_3_12, \
call_15_2_3_13, \
call_15_2_3_14, \
call_15_2_3_15, \
call_15_2_4_0, \
call_15_2_4_1, \
call_15_2_4_2, \
call_15_2_4_3, \
call_15_2_4_4, \
call_15_2_4_5, \
call_15_2_4_6, \
call_15_2_4_7, \
call_15_2_4_8, \
call_15_2_4_9, \
call_15_2_4_10, \
call_15_2_4_11, \
call_15_2_4_12, \
call_15_2_4_13, \
call_15_2_4_14, \
call_15_2_4_15, \
call_15_2_5_0, \
call_15_2_5_1, \
call_15_2_5_2, \
call_15_2_5_3, \
call_15_2_5_4, \
call_15_2_5_5, \
call_15_2_5_6, \
call_15_2_5_7, \
call_15_2_5_8, \
call_15_2_5_9, \
call_15_2_5_10, \
call_15_2_5_11, \
call_15_2_5_12, \
call_15_2_5_13, \
call_15_2_5_14, \
call_15_2_5_15, \
call_15_2_6_0, \
call_15_2_6_1, \
call_15_2_6_2, \
call_15_2_6_3, \
call_15_2_6_4, \
call_15_2_6_5, \
call_15_2_6_6, \
call_15_2_6_7, \
call_15_2_6_8, \
call_15_2_6_9, \
call_15_2_6_10, \
call_15_2_6_11, \
call_15_2_6_12, \
call_15_2_6_13, \
call_15_2_6_14, \
call_15_2_6_15, \
call_15_2_7_0, \
call_15_2_7_1, \
call_15_2_7_2, \
call_15_2_7_3, \
call_15_2_7_4, \
call_15_2_7_5, \
call_15_2_7_6, \
call_15_2_7_7, \
call_15_2_7_8, \
call_15_2_7_9, \
call_15_2_7_10, \
call_15_2_7_11, \
call_15_2_7_12, \
call_15_2_7_13, \
call_15_2_7_14, \
call_15_2_7_15, \
call_15_2_8_0, \
call_15_2_8_1, \
call_15_2_8_2, \
call_15_2_8_3, \
call_15_2_8_4, \
call_15_2_8_5, \
call_15_2_8_6, \
call_15_2_8_7, \
call_15_2_8_8, \
call_15_2_8_9, \
call_15_2_8_10, \
call_15_2_8_11, \
call_15_2_8_12, \
call_15_2_8_13, \
call_15_2_8_14, \
call_15_2_8_15, \
call_15_2_9_0, \
call_15_2_9_1, \
call_15_2_9_2, \
call_15_2_9_3, \
call_15_2_9_4, \
call_15_2_9_5, \
call_15_2_9_6, \
call_15_2_9_7, \
call_15_2_9_8, \
call_15_2_9_9, \
call_15_2_9_10, \
call_15_2_9_11, \
call_15_2_9_12, \
call_15_2_9_13, \
call_15_2_9_14, \
call_15_2_9_15, \
call_15_2_10_0, \
call_15_2_10_1, \
call_15_2_10_2, \
call_15_2_10_3, \
call_15_2_10_4, \
call_15_2_10_5, \
call_15_2_10_6, \
call_15_2_10_7, \
call_15_2_10_8, \
call_15_2_10_9, \
call_15_2_10_10, \
call_15_2_10_11, \
call_15_2_10_12, \
call_15_2_10_13, \
call_15_2_10_14, \
call_15_2_10_15, \
call_15_2_11_0, \
call_15_2_11_1, \
call_15_2_11_2, \
call_15_2_11_3, \
call_15_2_11_4, \
call_15_2_11_5, \
call_15_2_11_6, \
call_15_2_11_7, \
call_15_2_11_8, \
call_15_2_11_9, \
call_15_2_11_10, \
call_15_2_11_11, \
call_15_2_11_12, \
call_15_2_11_13, \
call_15_2_11_14, \
call_15_2_11_15, \
call_15_2_12_0, \
call_15_2_12_1, \
call_15_2_12_2, \
call_15_2_12_3, \
call_15_2_12_4, \
call_15_2_12_5, \
call_15_2_12_6, \
call_15_2_12_7, \
call_15_2_12_8, \
call_15_2_12_9, \
call_15_2_12_10, \
call_15_2_12_11, \
call_15_2_12_12, \
call_15_2_12_13, \
call_15_2_12_14, \
call_15_2_12_15, \
call_15_2_13_0, \
call_15_2_13_1, \
call_15_2_13_2, \
call_15_2_13_3, \
call_15_2_13_4, \
call_15_2_13_5, \
call_15_2_13_6, \
call_15_2_13_7, \
call_15_2_13_8, \
call_15_2_13_9, \
call_15_2_13_10, \
call_15_2_13_11, \
call_15_2_13_12, \
call_15_2_13_13, \
call_15_2_13_14, \
call_15_2_13_15, \
call_15_2_14_0, \
call_15_2_14_1, \
call_15_2_14_2, \
call_15_2_14_3, \
call_15_2_14_4, \
call_15_2_14_5, \
call_15_2_14_6, \
call_15_2_14_7, \
call_15_2_14_8, \
call_15_2_14_9, \
call_15_2_14_10, \
call_15_2_14_11, \
call_15_2_14_12, \
call_15_2_14_13, \
call_15_2_14_14, \
call_15_2_14_15, \
call_15_2_15_0, \
call_15_2_15_1, \
call_15_2_15_2, \
call_15_2_15_3, \
call_15_2_15_4, \
call_15_2_15_5, \
call_15_2_15_6, \
call_15_2_15_7, \
call_15_2_15_8, \
call_15_2_15_9, \
call_15_2_15_10, \
call_15_2_15_11, \
call_15_2_15_12, \
call_15_2_15_13, \
call_15_2_15_14, \
call_15_2_15_15, \
call_15_3_0_0, \
call_15_3_0_1, \
call_15_3_0_2, \
call_15_3_0_3, \
call_15_3_0_4, \
call_15_3_0_5, \
call_15_3_0_6, \
call_15_3_0_7, \
call_15_3_0_8, \
call_15_3_0_9, \
call_15_3_0_10, \
call_15_3_0_11, \
call_15_3_0_12, \
call_15_3_0_13, \
call_15_3_0_14, \
call_15_3_0_15, \
call_15_3_1_0, \
call_15_3_1_1, \
call_15_3_1_2, \
call_15_3_1_3, \
call_15_3_1_4, \
call_15_3_1_5, \
call_15_3_1_6, \
call_15_3_1_7, \
call_15_3_1_8, \
call_15_3_1_9, \
call_15_3_1_10, \
call_15_3_1_11, \
call_15_3_1_12, \
call_15_3_1_13, \
call_15_3_1_14, \
call_15_3_1_15, \
call_15_3_2_0, \
call_15_3_2_1, \
call_15_3_2_2, \
call_15_3_2_3, \
call_15_3_2_4, \
call_15_3_2_5, \
call_15_3_2_6, \
call_15_3_2_7, \
call_15_3_2_8, \
call_15_3_2_9, \
call_15_3_2_10, \
call_15_3_2_11, \
call_15_3_2_12, \
call_15_3_2_13, \
call_15_3_2_14, \
call_15_3_2_15, \
call_15_3_3_0, \
call_15_3_3_1, \
call_15_3_3_2, \
call_15_3_3_3, \
call_15_3_3_4, \
call_15_3_3_5, \
call_15_3_3_6, \
call_15_3_3_7, \
call_15_3_3_8, \
call_15_3_3_9, \
call_15_3_3_10, \
call_15_3_3_11, \
call_15_3_3_12, \
call_15_3_3_13, \
call_15_3_3_14, \
call_15_3_3_15, \
call_15_3_4_0, \
call_15_3_4_1, \
call_15_3_4_2, \
call_15_3_4_3, \
call_15_3_4_4, \
call_15_3_4_5, \
call_15_3_4_6, \
call_15_3_4_7, \
call_15_3_4_8, \
call_15_3_4_9, \
call_15_3_4_10, \
call_15_3_4_11, \
call_15_3_4_12, \
call_15_3_4_13, \
call_15_3_4_14, \
call_15_3_4_15, \
call_15_3_5_0, \
call_15_3_5_1, \
call_15_3_5_2, \
call_15_3_5_3, \
call_15_3_5_4, \
call_15_3_5_5, \
call_15_3_5_6, \
call_15_3_5_7, \
call_15_3_5_8, \
call_15_3_5_9, \
call_15_3_5_10, \
call_15_3_5_11, \
call_15_3_5_12, \
call_15_3_5_13, \
call_15_3_5_14, \
call_15_3_5_15, \
call_15_3_6_0, \
call_15_3_6_1, \
call_15_3_6_2, \
call_15_3_6_3, \
call_15_3_6_4, \
call_15_3_6_5, \
call_15_3_6_6, \
call_15_3_6_7, \
call_15_3_6_8, \
call_15_3_6_9, \
call_15_3_6_10, \
call_15_3_6_11, \
call_15_3_6_12, \
call_15_3_6_13, \
call_15_3_6_14, \
call_15_3_6_15, \
call_15_3_7_0, \
call_15_3_7_1, \
call_15_3_7_2, \
call_15_3_7_3, \
call_15_3_7_4, \
call_15_3_7_5, \
call_15_3_7_6, \
call_15_3_7_7, \
call_15_3_7_8, \
call_15_3_7_9, \
call_15_3_7_10, \
call_15_3_7_11, \
call_15_3_7_12, \
call_15_3_7_13, \
call_15_3_7_14, \
call_15_3_7_15, \
call_15_3_8_0, \
call_15_3_8_1, \
call_15_3_8_2, \
call_15_3_8_3, \
call_15_3_8_4, \
call_15_3_8_5, \
call_15_3_8_6, \
call_15_3_8_7, \
call_15_3_8_8, \
call_15_3_8_9, \
call_15_3_8_10, \
call_15_3_8_11, \
call_15_3_8_12, \
call_15_3_8_13, \
call_15_3_8_14, \
call_15_3_8_15, \
call_15_3_9_0, \
call_15_3_9_1, \
call_15_3_9_2, \
call_15_3_9_3, \
call_15_3_9_4, \
call_15_3_9_5, \
call_15_3_9_6, \
call_15_3_9_7, \
call_15_3_9_8, \
call_15_3_9_9, \
call_15_3_9_10, \
call_15_3_9_11, \
call_15_3_9_12, \
call_15_3_9_13, \
call_15_3_9_14, \
call_15_3_9_15, \
call_15_3_10_0, \
call_15_3_10_1, \
call_15_3_10_2, \
call_15_3_10_3, \
call_15_3_10_4, \
call_15_3_10_5, \
call_15_3_10_6, \
call_15_3_10_7, \
call_15_3_10_8, \
call_15_3_10_9, \
call_15_3_10_10, \
call_15_3_10_11, \
call_15_3_10_12, \
call_15_3_10_13, \
call_15_3_10_14, \
call_15_3_10_15, \
call_15_3_11_0, \
call_15_3_11_1, \
call_15_3_11_2, \
call_15_3_11_3, \
call_15_3_11_4, \
call_15_3_11_5, \
call_15_3_11_6, \
call_15_3_11_7, \
call_15_3_11_8, \
call_15_3_11_9, \
call_15_3_11_10, \
call_15_3_11_11, \
call_15_3_11_12, \
call_15_3_11_13, \
call_15_3_11_14, \
call_15_3_11_15, \
call_15_3_12_0, \
call_15_3_12_1, \
call_15_3_12_2, \
call_15_3_12_3, \
call_15_3_12_4, \
call_15_3_12_5, \
call_15_3_12_6, \
call_15_3_12_7, \
call_15_3_12_8, \
call_15_3_12_9, \
call_15_3_12_10, \
call_15_3_12_11, \
call_15_3_12_12, \
call_15_3_12_13, \
call_15_3_12_14, \
call_15_3_12_15, \
call_15_3_13_0, \
call_15_3_13_1, \
call_15_3_13_2, \
call_15_3_13_3, \
call_15_3_13_4, \
call_15_3_13_5, \
call_15_3_13_6, \
call_15_3_13_7, \
call_15_3_13_8, \
call_15_3_13_9, \
call_15_3_13_10, \
call_15_3_13_11, \
call_15_3_13_12, \
call_15_3_13_13, \
call_15_3_13_14, \
call_15_3_13_15, \
call_15_3_14_0, \
call_15_3_14_1, \
call_15_3_14_2, \
call_15_3_14_3, \
call_15_3_14_4, \
call_15_3_14_5, \
call_15_3_14_6, \
call_15_3_14_7, \
call_15_3_14_8, \
call_15_3_14_9, \
call_15_3_14_10, \
call_15_3_14_11, \
call_15_3_14_12, \
call_15_3_14_13, \
call_15_3_14_14, \
call_15_3_14_15, \
call_15_3_15_0, \
call_15_3_15_1, \
call_15_3_15_2, \
call_15_3_15_3, \
call_15_3_15_4, \
call_15_3_15_5, \
call_15_3_15_6, \
call_15_3_15_7, \
call_15_3_15_8, \
call_15_3_15_9, \
call_15_3_15_10, \
call_15_3_15_11, \
call_15_3_15_12, \
call_15_3_15_13, \
call_15_3_15_14, \
call_15_3_15_15, \
call_15_4_0_0, \
call_15_4_0_1, \
call_15_4_0_2, \
call_15_4_0_3, \
call_15_4_0_4, \
call_15_4_0_5, \
call_15_4_0_6, \
call_15_4_0_7, \
call_15_4_0_8, \
call_15_4_0_9, \
call_15_4_0_10, \
call_15_4_0_11, \
call_15_4_0_12, \
call_15_4_0_13, \
call_15_4_0_14, \
call_15_4_0_15, \
call_15_4_1_0, \
call_15_4_1_1, \
call_15_4_1_2, \
call_15_4_1_3, \
call_15_4_1_4, \
call_15_4_1_5, \
call_15_4_1_6, \
call_15_4_1_7, \
call_15_4_1_8, \
call_15_4_1_9, \
call_15_4_1_10, \
call_15_4_1_11, \
call_15_4_1_12, \
call_15_4_1_13, \
call_15_4_1_14, \
call_15_4_1_15, \
call_15_4_2_0, \
call_15_4_2_1, \
call_15_4_2_2, \
call_15_4_2_3, \
call_15_4_2_4, \
call_15_4_2_5, \
call_15_4_2_6, \
call_15_4_2_7, \
call_15_4_2_8, \
call_15_4_2_9, \
call_15_4_2_10, \
call_15_4_2_11, \
call_15_4_2_12, \
call_15_4_2_13, \
call_15_4_2_14, \
call_15_4_2_15, \
call_15_4_3_0, \
call_15_4_3_1, \
call_15_4_3_2, \
call_15_4_3_3, \
call_15_4_3_4, \
call_15_4_3_5, \
call_15_4_3_6, \
call_15_4_3_7, \
call_15_4_3_8, \
call_15_4_3_9, \
call_15_4_3_10, \
call_15_4_3_11, \
call_15_4_3_12, \
call_15_4_3_13, \
call_15_4_3_14, \
call_15_4_3_15, \
call_15_4_4_0, \
call_15_4_4_1, \
call_15_4_4_2, \
call_15_4_4_3, \
call_15_4_4_4, \
call_15_4_4_5, \
call_15_4_4_6, \
call_15_4_4_7, \
call_15_4_4_8, \
call_15_4_4_9, \
call_15_4_4_10, \
call_15_4_4_11, \
call_15_4_4_12, \
call_15_4_4_13, \
call_15_4_4_14, \
call_15_4_4_15, \
call_15_4_5_0, \
call_15_4_5_1, \
call_15_4_5_2, \
call_15_4_5_3, \
call_15_4_5_4, \
call_15_4_5_5, \
call_15_4_5_6, \
call_15_4_5_7, \
call_15_4_5_8, \
call_15_4_5_9, \
call_15_4_5_10, \
call_15_4_5_11, \
call_15_4_5_12, \
call_15_4_5_13, \
call_15_4_5_14, \
call_15_4_5_15, \
call_15_4_6_0, \
call_15_4_6_1, \
call_15_4_6_2, \
call_15_4_6_3, \
call_15_4_6_4, \
call_15_4_6_5, \
call_15_4_6_6, \
call_15_4_6_7, \
call_15_4_6_8, \
call_15_4_6_9, \
call_15_4_6_10, \
call_15_4_6_11, \
call_15_4_6_12, \
call_15_4_6_13, \
call_15_4_6_14, \
call_15_4_6_15, \
call_15_4_7_0, \
call_15_4_7_1, \
call_15_4_7_2, \
call_15_4_7_3, \
call_15_4_7_4, \
call_15_4_7_5, \
call_15_4_7_6, \
call_15_4_7_7, \
call_15_4_7_8, \
call_15_4_7_9, \
call_15_4_7_10, \
call_15_4_7_11, \
call_15_4_7_12, \
call_15_4_7_13, \
call_15_4_7_14, \
call_15_4_7_15, \
call_15_4_8_0, \
call_15_4_8_1, \
call_15_4_8_2, \
call_15_4_8_3, \
call_15_4_8_4, \
call_15_4_8_5, \
call_15_4_8_6, \
call_15_4_8_7, \
call_15_4_8_8, \
call_15_4_8_9, \
call_15_4_8_10, \
call_15_4_8_11, \
call_15_4_8_12, \
call_15_4_8_13, \
call_15_4_8_14, \
call_15_4_8_15, \
call_15_4_9_0, \
call_15_4_9_1, \
call_15_4_9_2, \
call_15_4_9_3, \
call_15_4_9_4, \
call_15_4_9_5, \
call_15_4_9_6, \
call_15_4_9_7, \
call_15_4_9_8, \
call_15_4_9_9, \
call_15_4_9_10, \
call_15_4_9_11, \
call_15_4_9_12, \
call_15_4_9_13, \
call_15_4_9_14, \
call_15_4_9_15, \
call_15_4_10_0, \
call_15_4_10_1, \
call_15_4_10_2, \
call_15_4_10_3, \
call_15_4_10_4, \
call_15_4_10_5, \
call_15_4_10_6, \
call_15_4_10_7, \
call_15_4_10_8, \
call_15_4_10_9, \
call_15_4_10_10, \
call_15_4_10_11, \
call_15_4_10_12, \
call_15_4_10_13, \
call_15_4_10_14, \
call_15_4_10_15, \
call_15_4_11_0, \
call_15_4_11_1, \
call_15_4_11_2, \
call_15_4_11_3, \
call_15_4_11_4, \
call_15_4_11_5, \
call_15_4_11_6, \
call_15_4_11_7, \
call_15_4_11_8, \
call_15_4_11_9, \
call_15_4_11_10, \
call_15_4_11_11, \
call_15_4_11_12, \
call_15_4_11_13, \
call_15_4_11_14, \
call_15_4_11_15, \
call_15_4_12_0, \
call_15_4_12_1, \
call_15_4_12_2, \
call_15_4_12_3, \
call_15_4_12_4, \
call_15_4_12_5, \
call_15_4_12_6, \
call_15_4_12_7, \
call_15_4_12_8, \
call_15_4_12_9, \
call_15_4_12_10, \
call_15_4_12_11, \
call_15_4_12_12, \
call_15_4_12_13, \
call_15_4_12_14, \
call_15_4_12_15, \
call_15_4_13_0, \
call_15_4_13_1, \
call_15_4_13_2, \
call_15_4_13_3, \
call_15_4_13_4, \
call_15_4_13_5, \
call_15_4_13_6, \
call_15_4_13_7, \
call_15_4_13_8, \
call_15_4_13_9, \
call_15_4_13_10, \
call_15_4_13_11, \
call_15_4_13_12, \
call_15_4_13_13, \
call_15_4_13_14, \
call_15_4_13_15, \
call_15_4_14_0, \
call_15_4_14_1, \
call_15_4_14_2, \
call_15_4_14_3, \
call_15_4_14_4, \
call_15_4_14_5, \
call_15_4_14_6, \
call_15_4_14_7, \
call_15_4_14_8, \
call_15_4_14_9, \
call_15_4_14_10, \
call_15_4_14_11, \
call_15_4_14_12, \
call_15_4_14_13, \
call_15_4_14_14, \
call_15_4_14_15, \
call_15_4_15_0, \
call_15_4_15_1, \
call_15_4_15_2, \
call_15_4_15_3, \
call_15_4_15_4, \
call_15_4_15_5, \
call_15_4_15_6, \
call_15_4_15_7, \
call_15_4_15_8, \
call_15_4_15_9, \
call_15_4_15_10, \
call_15_4_15_11, \
call_15_4_15_12, \
call_15_4_15_13, \
call_15_4_15_14, \
call_15_4_15_15, \
call_15_5_0_0, \
call_15_5_0_1, \
call_15_5_0_2, \
call_15_5_0_3, \
call_15_5_0_4, \
call_15_5_0_5, \
call_15_5_0_6, \
call_15_5_0_7, \
call_15_5_0_8, \
call_15_5_0_9, \
call_15_5_0_10, \
call_15_5_0_11, \
call_15_5_0_12, \
call_15_5_0_13, \
call_15_5_0_14, \
call_15_5_0_15, \
call_15_5_1_0, \
call_15_5_1_1, \
call_15_5_1_2, \
call_15_5_1_3, \
call_15_5_1_4, \
call_15_5_1_5, \
call_15_5_1_6, \
call_15_5_1_7, \
call_15_5_1_8, \
call_15_5_1_9, \
call_15_5_1_10, \
call_15_5_1_11, \
call_15_5_1_12, \
call_15_5_1_13, \
call_15_5_1_14, \
call_15_5_1_15, \
call_15_5_2_0, \
call_15_5_2_1, \
call_15_5_2_2, \
call_15_5_2_3, \
call_15_5_2_4, \
call_15_5_2_5, \
call_15_5_2_6, \
call_15_5_2_7, \
call_15_5_2_8, \
call_15_5_2_9, \
call_15_5_2_10, \
call_15_5_2_11, \
call_15_5_2_12, \
call_15_5_2_13, \
call_15_5_2_14, \
call_15_5_2_15, \
call_15_5_3_0, \
call_15_5_3_1, \
call_15_5_3_2, \
call_15_5_3_3, \
call_15_5_3_4, \
call_15_5_3_5, \
call_15_5_3_6, \
call_15_5_3_7, \
call_15_5_3_8, \
call_15_5_3_9, \
call_15_5_3_10, \
call_15_5_3_11, \
call_15_5_3_12, \
call_15_5_3_13, \
call_15_5_3_14, \
call_15_5_3_15, \
call_15_5_4_0, \
call_15_5_4_1, \
call_15_5_4_2, \
call_15_5_4_3, \
call_15_5_4_4, \
call_15_5_4_5, \
call_15_5_4_6, \
call_15_5_4_7, \
call_15_5_4_8, \
call_15_5_4_9, \
call_15_5_4_10, \
call_15_5_4_11, \
call_15_5_4_12, \
call_15_5_4_13, \
call_15_5_4_14, \
call_15_5_4_15, \
call_15_5_5_0, \
call_15_5_5_1, \
call_15_5_5_2, \
call_15_5_5_3, \
call_15_5_5_4, \
call_15_5_5_5, \
call_15_5_5_6, \
call_15_5_5_7, \
call_15_5_5_8, \
call_15_5_5_9, \
call_15_5_5_10, \
call_15_5_5_11, \
call_15_5_5_12, \
call_15_5_5_13, \
call_15_5_5_14, \
call_15_5_5_15, \
call_15_5_6_0, \
call_15_5_6_1, \
call_15_5_6_2, \
call_15_5_6_3, \
call_15_5_6_4, \
call_15_5_6_5, \
call_15_5_6_6, \
call_15_5_6_7, \
call_15_5_6_8, \
call_15_5_6_9, \
call_15_5_6_10, \
call_15_5_6_11, \
call_15_5_6_12, \
call_15_5_6_13, \
call_15_5_6_14, \
call_15_5_6_15, \
call_15_5_7_0, \
call_15_5_7_1, \
call_15_5_7_2, \
call_15_5_7_3, \
call_15_5_7_4, \
call_15_5_7_5, \
call_15_5_7_6, \
call_15_5_7_7, \
call_15_5_7_8, \
call_15_5_7_9, \
call_15_5_7_10, \
call_15_5_7_11, \
call_15_5_7_12, \
call_15_5_7_13, \
call_15_5_7_14, \
call_15_5_7_15, \
call_15_5_8_0, \
call_15_5_8_1, \
call_15_5_8_2, \
call_15_5_8_3, \
call_15_5_8_4, \
call_15_5_8_5, \
call_15_5_8_6, \
call_15_5_8_7, \
call_15_5_8_8, \
call_15_5_8_9, \
call_15_5_8_10, \
call_15_5_8_11, \
call_15_5_8_12, \
call_15_5_8_13, \
call_15_5_8_14, \
call_15_5_8_15, \
call_15_5_9_0, \
call_15_5_9_1, \
call_15_5_9_2, \
call_15_5_9_3, \
call_15_5_9_4, \
call_15_5_9_5, \
call_15_5_9_6, \
call_15_5_9_7, \
call_15_5_9_8, \
call_15_5_9_9, \
call_15_5_9_10, \
call_15_5_9_11, \
call_15_5_9_12, \
call_15_5_9_13, \
call_15_5_9_14, \
call_15_5_9_15, \
call_15_5_10_0, \
call_15_5_10_1, \
call_15_5_10_2, \
call_15_5_10_3, \
call_15_5_10_4, \
call_15_5_10_5, \
call_15_5_10_6, \
call_15_5_10_7, \
call_15_5_10_8, \
call_15_5_10_9, \
call_15_5_10_10, \
call_15_5_10_11, \
call_15_5_10_12, \
call_15_5_10_13, \
call_15_5_10_14, \
call_15_5_10_15, \
call_15_5_11_0, \
call_15_5_11_1, \
call_15_5_11_2, \
call_15_5_11_3, \
call_15_5_11_4, \
call_15_5_11_5, \
call_15_5_11_6, \
call_15_5_11_7, \
call_15_5_11_8, \
call_15_5_11_9, \
call_15_5_11_10, \
call_15_5_11_11, \
call_15_5_11_12, \
call_15_5_11_13, \
call_15_5_11_14, \
call_15_5_11_15, \
call_15_5_12_0, \
call_15_5_12_1, \
call_15_5_12_2, \
call_15_5_12_3, \
call_15_5_12_4, \
call_15_5_12_5, \
call_15_5_12_6, \
call_15_5_12_7, \
call_15_5_12_8, \
call_15_5_12_9, \
call_15_5_12_10, \
call_15_5_12_11, \
call_15_5_12_12, \
call_15_5_12_13, \
call_15_5_12_14, \
call_15_5_12_15, \
call_15_5_13_0, \
call_15_5_13_1, \
call_15_5_13_2, \
call_15_5_13_3, \
call_15_5_13_4, \
call_15_5_13_5, \
call_15_5_13_6, \
call_15_5_13_7, \
call_15_5_13_8, \
call_15_5_13_9, \
call_15_5_13_10, \
call_15_5_13_11, \
call_15_5_13_12, \
call_15_5_13_13, \
call_15_5_13_14, \
call_15_5_13_15, \
call_15_5_14_0, \
call_15_5_14_1, \
call_15_5_14_2, \
call_15_5_14_3, \
call_15_5_14_4, \
call_15_5_14_5, \
call_15_5_14_6, \
call_15_5_14_7, \
call_15_5_14_8, \
call_15_5_14_9, \
call_15_5_14_10, \
call_15_5_14_11, \
call_15_5_14_12, \
call_15_5_14_13, \
call_15_5_14_14, \
call_15_5_14_15, \
call_15_5_15_0, \
call_15_5_15_1, \
call_15_5_15_2, \
call_15_5_15_3, \
call_15_5_15_4, \
call_15_5_15_5, \
call_15_5_15_6, \
call_15_5_15_7, \
call_15_5_15_8, \
call_15_5_15_9, \
call_15_5_15_10, \
call_15_5_15_11, \
call_15_5_15_12, \
call_15_5_15_13, \
call_15_5_15_14, \
call_15_5_15_15, \
call_15_6_0_0, \
call_15_6_0_1, \
call_15_6_0_2, \
call_15_6_0_3, \
call_15_6_0_4, \
call_15_6_0_5, \
call_15_6_0_6, \
call_15_6_0_7, \
call_15_6_0_8, \
call_15_6_0_9, \
call_15_6_0_10, \
call_15_6_0_11, \
call_15_6_0_12, \
call_15_6_0_13, \
call_15_6_0_14, \
call_15_6_0_15, \
call_15_6_1_0, \
call_15_6_1_1, \
call_15_6_1_2, \
call_15_6_1_3, \
call_15_6_1_4, \
call_15_6_1_5, \
call_15_6_1_6, \
call_15_6_1_7, \
call_15_6_1_8, \
call_15_6_1_9, \
call_15_6_1_10, \
call_15_6_1_11, \
call_15_6_1_12, \
call_15_6_1_13, \
call_15_6_1_14, \
call_15_6_1_15, \
call_15_6_2_0, \
call_15_6_2_1, \
call_15_6_2_2, \
call_15_6_2_3, \
call_15_6_2_4, \
call_15_6_2_5, \
call_15_6_2_6, \
call_15_6_2_7, \
call_15_6_2_8, \
call_15_6_2_9, \
call_15_6_2_10, \
call_15_6_2_11, \
call_15_6_2_12, \
call_15_6_2_13, \
call_15_6_2_14, \
call_15_6_2_15, \
call_15_6_3_0, \
call_15_6_3_1, \
call_15_6_3_2, \
call_15_6_3_3, \
call_15_6_3_4, \
call_15_6_3_5, \
call_15_6_3_6, \
call_15_6_3_7, \
call_15_6_3_8, \
call_15_6_3_9, \
call_15_6_3_10, \
call_15_6_3_11, \
call_15_6_3_12, \
call_15_6_3_13, \
call_15_6_3_14, \
call_15_6_3_15, \
call_15_6_4_0, \
call_15_6_4_1, \
call_15_6_4_2, \
call_15_6_4_3, \
call_15_6_4_4, \
call_15_6_4_5, \
call_15_6_4_6, \
call_15_6_4_7, \
call_15_6_4_8, \
call_15_6_4_9, \
call_15_6_4_10, \
call_15_6_4_11, \
call_15_6_4_12, \
call_15_6_4_13, \
call_15_6_4_14, \
call_15_6_4_15, \
call_15_6_5_0, \
call_15_6_5_1, \
call_15_6_5_2, \
call_15_6_5_3, \
call_15_6_5_4, \
call_15_6_5_5, \
call_15_6_5_6, \
call_15_6_5_7, \
call_15_6_5_8, \
call_15_6_5_9, \
call_15_6_5_10, \
call_15_6_5_11, \
call_15_6_5_12, \
call_15_6_5_13, \
call_15_6_5_14, \
call_15_6_5_15, \
call_15_6_6_0, \
call_15_6_6_1, \
call_15_6_6_2, \
call_15_6_6_3, \
call_15_6_6_4, \
call_15_6_6_5, \
call_15_6_6_6, \
call_15_6_6_7, \
call_15_6_6_8, \
call_15_6_6_9, \
call_15_6_6_10, \
call_15_6_6_11, \
call_15_6_6_12, \
call_15_6_6_13, \
call_15_6_6_14, \
call_15_6_6_15, \
call_15_6_7_0, \
call_15_6_7_1, \
call_15_6_7_2, \
call_15_6_7_3, \
call_15_6_7_4, \
call_15_6_7_5, \
call_15_6_7_6, \
call_15_6_7_7, \
call_15_6_7_8, \
call_15_6_7_9, \
call_15_6_7_10, \
call_15_6_7_11, \
call_15_6_7_12, \
call_15_6_7_13, \
call_15_6_7_14, \
call_15_6_7_15, \
call_15_6_8_0, \
call_15_6_8_1, \
call_15_6_8_2, \
call_15_6_8_3, \
call_15_6_8_4, \
call_15_6_8_5, \
call_15_6_8_6, \
call_15_6_8_7, \
call_15_6_8_8, \
call_15_6_8_9, \
call_15_6_8_10, \
call_15_6_8_11, \
call_15_6_8_12, \
call_15_6_8_13, \
call_15_6_8_14, \
call_15_6_8_15, \
call_15_6_9_0, \
call_15_6_9_1, \
call_15_6_9_2, \
call_15_6_9_3, \
call_15_6_9_4, \
call_15_6_9_5, \
call_15_6_9_6, \
call_15_6_9_7, \
call_15_6_9_8, \
call_15_6_9_9, \
call_15_6_9_10, \
call_15_6_9_11, \
call_15_6_9_12, \
call_15_6_9_13, \
call_15_6_9_14, \
call_15_6_9_15, \
call_15_6_10_0, \
call_15_6_10_1, \
call_15_6_10_2, \
call_15_6_10_3, \
call_15_6_10_4, \
call_15_6_10_5, \
call_15_6_10_6, \
call_15_6_10_7, \
call_15_6_10_8, \
call_15_6_10_9, \
call_15_6_10_10, \
call_15_6_10_11, \
call_15_6_10_12, \
call_15_6_10_13, \
call_15_6_10_14, \
call_15_6_10_15, \
call_15_6_11_0, \
call_15_6_11_1, \
call_15_6_11_2, \
call_15_6_11_3, \
call_15_6_11_4, \
call_15_6_11_5, \
call_15_6_11_6, \
call_15_6_11_7, \
call_15_6_11_8, \
call_15_6_11_9, \
call_15_6_11_10, \
call_15_6_11_11, \
call_15_6_11_12, \
call_15_6_11_13, \
call_15_6_11_14, \
call_15_6_11_15, \
call_15_6_12_0, \
call_15_6_12_1, \
call_15_6_12_2, \
call_15_6_12_3, \
call_15_6_12_4, \
call_15_6_12_5, \
call_15_6_12_6, \
call_15_6_12_7, \
call_15_6_12_8, \
call_15_6_12_9, \
call_15_6_12_10, \
call_15_6_12_11, \
call_15_6_12_12, \
call_15_6_12_13, \
call_15_6_12_14, \
call_15_6_12_15, \
call_15_6_13_0, \
call_15_6_13_1, \
call_15_6_13_2, \
call_15_6_13_3, \
call_15_6_13_4, \
call_15_6_13_5, \
call_15_6_13_6, \
call_15_6_13_7, \
call_15_6_13_8, \
call_15_6_13_9, \
call_15_6_13_10, \
call_15_6_13_11, \
call_15_6_13_12, \
call_15_6_13_13, \
call_15_6_13_14, \
call_15_6_13_15, \
call_15_6_14_0, \
call_15_6_14_1, \
call_15_6_14_2, \
call_15_6_14_3, \
call_15_6_14_4, \
call_15_6_14_5, \
call_15_6_14_6, \
call_15_6_14_7, \
call_15_6_14_8, \
call_15_6_14_9, \
call_15_6_14_10, \
call_15_6_14_11, \
call_15_6_14_12, \
call_15_6_14_13, \
call_15_6_14_14, \
call_15_6_14_15, \
call_15_6_15_0, \
call_15_6_15_1, \
call_15_6_15_2, \
call_15_6_15_3, \
call_15_6_15_4, \
call_15_6_15_5, \
call_15_6_15_6, \
call_15_6_15_7, \
call_15_6_15_8, \
call_15_6_15_9, \
call_15_6_15_10, \
call_15_6_15_11, \
call_15_6_15_12, \
call_15_6_15_13, \
call_15_6_15_14, \
call_15_6_15_15, \
call_15_7_0_0, \
call_15_7_0_1, \
call_15_7_0_2, \
call_15_7_0_3, \
call_15_7_0_4, \
call_15_7_0_5, \
call_15_7_0_6, \
call_15_7_0_7, \
call_15_7_0_8, \
call_15_7_0_9, \
call_15_7_0_10, \
call_15_7_0_11, \
call_15_7_0_12, \
call_15_7_0_13, \
call_15_7_0_14, \
call_15_7_0_15, \
call_15_7_1_0, \
call_15_7_1_1, \
call_15_7_1_2, \
call_15_7_1_3, \
call_15_7_1_4, \
call_15_7_1_5, \
call_15_7_1_6, \
call_15_7_1_7, \
call_15_7_1_8, \
call_15_7_1_9, \
call_15_7_1_10, \
call_15_7_1_11, \
call_15_7_1_12, \
call_15_7_1_13, \
call_15_7_1_14, \
call_15_7_1_15, \
call_15_7_2_0, \
call_15_7_2_1, \
call_15_7_2_2, \
call_15_7_2_3, \
call_15_7_2_4, \
call_15_7_2_5, \
call_15_7_2_6, \
call_15_7_2_7, \
call_15_7_2_8, \
call_15_7_2_9, \
call_15_7_2_10, \
call_15_7_2_11, \
call_15_7_2_12, \
call_15_7_2_13, \
call_15_7_2_14, \
call_15_7_2_15, \
call_15_7_3_0, \
call_15_7_3_1, \
call_15_7_3_2, \
call_15_7_3_3, \
call_15_7_3_4, \
call_15_7_3_5, \
call_15_7_3_6, \
call_15_7_3_7, \
call_15_7_3_8, \
call_15_7_3_9, \
call_15_7_3_10, \
call_15_7_3_11, \
call_15_7_3_12, \
call_15_7_3_13, \
call_15_7_3_14, \
call_15_7_3_15, \
call_15_7_4_0, \
call_15_7_4_1, \
call_15_7_4_2, \
call_15_7_4_3, \
call_15_7_4_4, \
call_15_7_4_5, \
call_15_7_4_6, \
call_15_7_4_7, \
call_15_7_4_8, \
call_15_7_4_9, \
call_15_7_4_10, \
call_15_7_4_11, \
call_15_7_4_12, \
call_15_7_4_13, \
call_15_7_4_14, \
call_15_7_4_15, \
call_15_7_5_0, \
call_15_7_5_1, \
call_15_7_5_2, \
call_15_7_5_3, \
call_15_7_5_4, \
call_15_7_5_5, \
call_15_7_5_6, \
call_15_7_5_7, \
call_15_7_5_8, \
call_15_7_5_9, \
call_15_7_5_10, \
call_15_7_5_11, \
call_15_7_5_12, \
call_15_7_5_13, \
call_15_7_5_14, \
call_15_7_5_15, \
call_15_7_6_0, \
call_15_7_6_1, \
call_15_7_6_2, \
call_15_7_6_3, \
call_15_7_6_4, \
call_15_7_6_5, \
call_15_7_6_6, \
call_15_7_6_7, \
call_15_7_6_8, \
call_15_7_6_9, \
call_15_7_6_10, \
call_15_7_6_11, \
call_15_7_6_12, \
call_15_7_6_13, \
call_15_7_6_14, \
call_15_7_6_15, \
call_15_7_7_0, \
call_15_7_7_1, \
call_15_7_7_2, \
call_15_7_7_3, \
call_15_7_7_4, \
call_15_7_7_5, \
call_15_7_7_6, \
call_15_7_7_7, \
call_15_7_7_8, \
call_15_7_7_9, \
call_15_7_7_10, \
call_15_7_7_11, \
call_15_7_7_12, \
call_15_7_7_13, \
call_15_7_7_14, \
call_15_7_7_15, \
call_15_7_8_0, \
call_15_7_8_1, \
call_15_7_8_2, \
call_15_7_8_3, \
call_15_7_8_4, \
call_15_7_8_5, \
call_15_7_8_6, \
call_15_7_8_7, \
call_15_7_8_8, \
call_15_7_8_9, \
call_15_7_8_10, \
call_15_7_8_11, \
call_15_7_8_12, \
call_15_7_8_13, \
call_15_7_8_14, \
call_15_7_8_15, \
call_15_7_9_0, \
call_15_7_9_1, \
call_15_7_9_2, \
call_15_7_9_3, \
call_15_7_9_4, \
call_15_7_9_5, \
call_15_7_9_6, \
call_15_7_9_7, \
call_15_7_9_8, \
call_15_7_9_9, \
call_15_7_9_10, \
call_15_7_9_11, \
call_15_7_9_12, \
call_15_7_9_13, \
call_15_7_9_14, \
call_15_7_9_15, \
call_15_7_10_0, \
call_15_7_10_1, \
call_15_7_10_2, \
call_15_7_10_3, \
call_15_7_10_4, \
call_15_7_10_5, \
call_15_7_10_6, \
call_15_7_10_7, \
call_15_7_10_8, \
call_15_7_10_9, \
call_15_7_10_10, \
call_15_7_10_11, \
call_15_7_10_12, \
call_15_7_10_13, \
call_15_7_10_14, \
call_15_7_10_15, \
call_15_7_11_0, \
call_15_7_11_1, \
call_15_7_11_2, \
call_15_7_11_3, \
call_15_7_11_4, \
call_15_7_11_5, \
call_15_7_11_6, \
call_15_7_11_7, \
call_15_7_11_8, \
call_15_7_11_9, \
call_15_7_11_10, \
call_15_7_11_11, \
call_15_7_11_12, \
call_15_7_11_13, \
call_15_7_11_14, \
call_15_7_11_15, \
call_15_7_12_0, \
call_15_7_12_1, \
call_15_7_12_2, \
call_15_7_12_3, \
call_15_7_12_4, \
call_15_7_12_5, \
call_15_7_12_6, \
call_15_7_12_7, \
call_15_7_12_8, \
call_15_7_12_9, \
call_15_7_12_10, \
call_15_7_12_11, \
call_15_7_12_12, \
call_15_7_12_13, \
call_15_7_12_14, \
call_15_7_12_15, \
call_15_7_13_0, \
call_15_7_13_1, \
call_15_7_13_2, \
call_15_7_13_3, \
call_15_7_13_4, \
call_15_7_13_5, \
call_15_7_13_6, \
call_15_7_13_7, \
call_15_7_13_8, \
call_15_7_13_9, \
call_15_7_13_10, \
call_15_7_13_11, \
call_15_7_13_12, \
call_15_7_13_13, \
call_15_7_13_14, \
call_15_7_13_15, \
call_15_7_14_0, \
call_15_7_14_1, \
call_15_7_14_2, \
call_15_7_14_3, \
call_15_7_14_4, \
call_15_7_14_5, \
call_15_7_14_6, \
call_15_7_14_7, \
call_15_7_14_8, \
call_15_7_14_9, \
call_15_7_14_10, \
call_15_7_14_11, \
call_15_7_14_12, \
call_15_7_14_13, \
call_15_7_14_14, \
call_15_7_14_15, \
call_15_7_15_0, \
call_15_7_15_1, \
call_15_7_15_2, \
call_15_7_15_3, \
call_15_7_15_4, \
call_15_7_15_5, \
call_15_7_15_6, \
call_15_7_15_7, \
call_15_7_15_8, \
call_15_7_15_9, \
call_15_7_15_10, \
call_15_7_15_11, \
call_15_7_15_12, \
call_15_7_15_13, \
call_15_7_15_14, \
call_15_7_15_15, \
call_15_8_0_0, \
call_15_8_0_1, \
call_15_8_0_2, \
call_15_8_0_3, \
call_15_8_0_4, \
call_15_8_0_5, \
call_15_8_0_6, \
call_15_8_0_7, \
call_15_8_0_8, \
call_15_8_0_9, \
call_15_8_0_10, \
call_15_8_0_11, \
call_15_8_0_12, \
call_15_8_0_13, \
call_15_8_0_14, \
call_15_8_0_15, \
call_15_8_1_0, \
call_15_8_1_1, \
call_15_8_1_2, \
call_15_8_1_3, \
call_15_8_1_4, \
call_15_8_1_5, \
call_15_8_1_6, \
call_15_8_1_7, \
call_15_8_1_8, \
call_15_8_1_9, \
call_15_8_1_10, \
call_15_8_1_11, \
call_15_8_1_12, \
call_15_8_1_13, \
call_15_8_1_14, \
call_15_8_1_15, \
call_15_8_2_0, \
call_15_8_2_1, \
call_15_8_2_2, \
call_15_8_2_3, \
call_15_8_2_4, \
call_15_8_2_5, \
call_15_8_2_6, \
call_15_8_2_7, \
call_15_8_2_8, \
call_15_8_2_9, \
call_15_8_2_10, \
call_15_8_2_11, \
call_15_8_2_12, \
call_15_8_2_13, \
call_15_8_2_14, \
call_15_8_2_15, \
call_15_8_3_0, \
call_15_8_3_1, \
call_15_8_3_2, \
call_15_8_3_3, \
call_15_8_3_4, \
call_15_8_3_5, \
call_15_8_3_6, \
call_15_8_3_7, \
call_15_8_3_8, \
call_15_8_3_9, \
call_15_8_3_10, \
call_15_8_3_11, \
call_15_8_3_12, \
call_15_8_3_13, \
call_15_8_3_14, \
call_15_8_3_15, \
call_15_8_4_0, \
call_15_8_4_1, \
call_15_8_4_2, \
call_15_8_4_3, \
call_15_8_4_4, \
call_15_8_4_5, \
call_15_8_4_6, \
call_15_8_4_7, \
call_15_8_4_8, \
call_15_8_4_9, \
call_15_8_4_10, \
call_15_8_4_11, \
call_15_8_4_12, \
call_15_8_4_13, \
call_15_8_4_14, \
call_15_8_4_15, \
call_15_8_5_0, \
call_15_8_5_1, \
call_15_8_5_2, \
call_15_8_5_3, \
call_15_8_5_4, \
call_15_8_5_5, \
call_15_8_5_6, \
call_15_8_5_7, \
call_15_8_5_8, \
call_15_8_5_9, \
call_15_8_5_10, \
call_15_8_5_11, \
call_15_8_5_12, \
call_15_8_5_13, \
call_15_8_5_14, \
call_15_8_5_15, \
call_15_8_6_0, \
call_15_8_6_1, \
call_15_8_6_2, \
call_15_8_6_3, \
call_15_8_6_4, \
call_15_8_6_5, \
call_15_8_6_6, \
call_15_8_6_7, \
call_15_8_6_8, \
call_15_8_6_9, \
call_15_8_6_10, \
call_15_8_6_11, \
call_15_8_6_12, \
call_15_8_6_13, \
call_15_8_6_14, \
call_15_8_6_15, \
call_15_8_7_0, \
call_15_8_7_1, \
call_15_8_7_2, \
call_15_8_7_3, \
call_15_8_7_4, \
call_15_8_7_5, \
call_15_8_7_6, \
call_15_8_7_7, \
call_15_8_7_8, \
call_15_8_7_9, \
call_15_8_7_10, \
call_15_8_7_11, \
call_15_8_7_12, \
call_15_8_7_13, \
call_15_8_7_14, \
call_15_8_7_15, \
call_15_8_8_0, \
call_15_8_8_1, \
call_15_8_8_2, \
call_15_8_8_3, \
call_15_8_8_4, \
call_15_8_8_5, \
call_15_8_8_6, \
call_15_8_8_7, \
call_15_8_8_8, \
call_15_8_8_9, \
call_15_8_8_10, \
call_15_8_8_11, \
call_15_8_8_12, \
call_15_8_8_13, \
call_15_8_8_14, \
call_15_8_8_15, \
call_15_8_9_0, \
call_15_8_9_1, \
call_15_8_9_2, \
call_15_8_9_3, \
call_15_8_9_4, \
call_15_8_9_5, \
call_15_8_9_6, \
call_15_8_9_7, \
call_15_8_9_8, \
call_15_8_9_9, \
call_15_8_9_10, \
call_15_8_9_11, \
call_15_8_9_12, \
call_15_8_9_13, \
call_15_8_9_14, \
call_15_8_9_15, \
call_15_8_10_0, \
call_15_8_10_1, \
call_15_8_10_2, \
call_15_8_10_3, \
call_15_8_10_4, \
call_15_8_10_5, \
call_15_8_10_6, \
call_15_8_10_7, \
call_15_8_10_8, \
call_15_8_10_9, \
call_15_8_10_10, \
call_15_8_10_11, \
call_15_8_10_12, \
call_15_8_10_13, \
call_15_8_10_14, \
call_15_8_10_15, \
call_15_8_11_0, \
call_15_8_11_1, \
call_15_8_11_2, \
call_15_8_11_3, \
call_15_8_11_4, \
call_15_8_11_5, \
call_15_8_11_6, \
call_15_8_11_7, \
call_15_8_11_8, \
call_15_8_11_9, \
call_15_8_11_10, \
call_15_8_11_11, \
call_15_8_11_12, \
call_15_8_11_13, \
call_15_8_11_14, \
call_15_8_11_15, \
call_15_8_12_0, \
call_15_8_12_1, \
call_15_8_12_2, \
call_15_8_12_3, \
call_15_8_12_4, \
call_15_8_12_5, \
call_15_8_12_6, \
call_15_8_12_7, \
call_15_8_12_8, \
call_15_8_12_9, \
call_15_8_12_10, \
call_15_8_12_11, \
call_15_8_12_12, \
call_15_8_12_13, \
call_15_8_12_14, \
call_15_8_12_15, \
call_15_8_13_0, \
call_15_8_13_1, \
call_15_8_13_2, \
call_15_8_13_3, \
call_15_8_13_4, \
call_15_8_13_5, \
call_15_8_13_6, \
call_15_8_13_7, \
call_15_8_13_8, \
call_15_8_13_9, \
call_15_8_13_10, \
call_15_8_13_11, \
call_15_8_13_12, \
call_15_8_13_13, \
call_15_8_13_14, \
call_15_8_13_15, \
call_15_8_14_0, \
call_15_8_14_1, \
call_15_8_14_2, \
call_15_8_14_3, \
call_15_8_14_4, \
call_15_8_14_5, \
call_15_8_14_6, \
call_15_8_14_7, \
call_15_8_14_8, \
call_15_8_14_9, \
call_15_8_14_10, \
call_15_8_14_11, \
call_15_8_14_12, \
call_15_8_14_13, \
call_15_8_14_14, \
call_15_8_14_15, \
call_15_8_15_0, \
call_15_8_15_1, \
call_15_8_15_2, \
call_15_8_15_3, \
call_15_8_15_4, \
call_15_8_15_5, \
call_15_8_15_6, \
call_15_8_15_7, \
call_15_8_15_8, \
call_15_8_15_9, \
call_15_8_15_10, \
call_15_8_15_11, \
call_15_8_15_12, \
call_15_8_15_13, \
call_15_8_15_14, \
call_15_8_15_15, \
call_15_9_0_0, \
call_15_9_0_1, \
call_15_9_0_2, \
call_15_9_0_3, \
call_15_9_0_4, \
call_15_9_0_5, \
call_15_9_0_6, \
call_15_9_0_7, \
call_15_9_0_8, \
call_15_9_0_9, \
call_15_9_0_10, \
call_15_9_0_11, \
call_15_9_0_12, \
call_15_9_0_13, \
call_15_9_0_14, \
call_15_9_0_15, \
call_15_9_1_0, \
call_15_9_1_1, \
call_15_9_1_2, \
call_15_9_1_3, \
call_15_9_1_4, \
call_15_9_1_5, \
call_15_9_1_6, \
call_15_9_1_7, \
call_15_9_1_8, \
call_15_9_1_9, \
call_15_9_1_10, \
call_15_9_1_11, \
call_15_9_1_12, \
call_15_9_1_13, \
call_15_9_1_14, \
call_15_9_1_15, \
call_15_9_2_0, \
call_15_9_2_1, \
call_15_9_2_2, \
call_15_9_2_3, \
call_15_9_2_4, \
call_15_9_2_5, \
call_15_9_2_6, \
call_15_9_2_7, \
call_15_9_2_8, \
call_15_9_2_9, \
call_15_9_2_10, \
call_15_9_2_11, \
call_15_9_2_12, \
call_15_9_2_13, \
call_15_9_2_14, \
call_15_9_2_15, \
call_15_9_3_0, \
call_15_9_3_1, \
call_15_9_3_2, \
call_15_9_3_3, \
call_15_9_3_4, \
call_15_9_3_5, \
call_15_9_3_6, \
call_15_9_3_7, \
call_15_9_3_8, \
call_15_9_3_9, \
call_15_9_3_10, \
call_15_9_3_11, \
call_15_9_3_12, \
call_15_9_3_13, \
call_15_9_3_14, \
call_15_9_3_15, \
call_15_9_4_0, \
call_15_9_4_1, \
call_15_9_4_2, \
call_15_9_4_3, \
call_15_9_4_4, \
call_15_9_4_5, \
call_15_9_4_6, \
call_15_9_4_7, \
call_15_9_4_8, \
call_15_9_4_9, \
call_15_9_4_10, \
call_15_9_4_11, \
call_15_9_4_12, \
call_15_9_4_13, \
call_15_9_4_14, \
call_15_9_4_15, \
call_15_9_5_0, \
call_15_9_5_1, \
call_15_9_5_2, \
call_15_9_5_3, \
call_15_9_5_4, \
call_15_9_5_5, \
call_15_9_5_6, \
call_15_9_5_7, \
call_15_9_5_8, \
call_15_9_5_9, \
call_15_9_5_10, \
call_15_9_5_11, \
call_15_9_5_12, \
call_15_9_5_13, \
call_15_9_5_14, \
call_15_9_5_15, \
call_15_9_6_0, \
call_15_9_6_1, \
call_15_9_6_2, \
call_15_9_6_3, \
call_15_9_6_4, \
call_15_9_6_5, \
call_15_9_6_6, \
call_15_9_6_7, \
call_15_9_6_8, \
call_15_9_6_9, \
call_15_9_6_10, \
call_15_9_6_11, \
call_15_9_6_12, \
call_15_9_6_13, \
call_15_9_6_14, \
call_15_9_6_15, \
call_15_9_7_0, \
call_15_9_7_1, \
call_15_9_7_2, \
call_15_9_7_3, \
call_15_9_7_4, \
call_15_9_7_5, \
call_15_9_7_6, \
call_15_9_7_7, \
call_15_9_7_8, \
call_15_9_7_9, \
call_15_9_7_10, \
call_15_9_7_11, \
call_15_9_7_12, \
call_15_9_7_13, \
call_15_9_7_14, \
call_15_9_7_15, \
call_15_9_8_0, \
call_15_9_8_1, \
call_15_9_8_2, \
call_15_9_8_3, \
call_15_9_8_4, \
call_15_9_8_5, \
call_15_9_8_6, \
call_15_9_8_7, \
call_15_9_8_8, \
call_15_9_8_9, \
call_15_9_8_10, \
call_15_9_8_11, \
call_15_9_8_12, \
call_15_9_8_13, \
call_15_9_8_14, \
call_15_9_8_15, \
call_15_9_9_0, \
call_15_9_9_1, \
call_15_9_9_2, \
call_15_9_9_3, \
call_15_9_9_4, \
call_15_9_9_5, \
call_15_9_9_6, \
call_15_9_9_7, \
call_15_9_9_8, \
call_15_9_9_9, \
call_15_9_9_10, \
call_15_9_9_11, \
call_15_9_9_12, \
call_15_9_9_13, \
call_15_9_9_14, \
call_15_9_9_15, \
call_15_9_10_0, \
call_15_9_10_1, \
call_15_9_10_2, \
call_15_9_10_3, \
call_15_9_10_4, \
call_15_9_10_5, \
call_15_9_10_6, \
call_15_9_10_7, \
call_15_9_10_8, \
call_15_9_10_9, \
call_15_9_10_10, \
call_15_9_10_11, \
call_15_9_10_12, \
call_15_9_10_13, \
call_15_9_10_14, \
call_15_9_10_15, \
call_15_9_11_0, \
call_15_9_11_1, \
call_15_9_11_2, \
call_15_9_11_3, \
call_15_9_11_4, \
call_15_9_11_5, \
call_15_9_11_6, \
call_15_9_11_7, \
call_15_9_11_8, \
call_15_9_11_9, \
call_15_9_11_10, \
call_15_9_11_11, \
call_15_9_11_12, \
call_15_9_11_13, \
call_15_9_11_14, \
call_15_9_11_15, \
call_15_9_12_0, \
call_15_9_12_1, \
call_15_9_12_2, \
call_15_9_12_3, \
call_15_9_12_4, \
call_15_9_12_5, \
call_15_9_12_6, \
call_15_9_12_7, \
call_15_9_12_8, \
call_15_9_12_9, \
call_15_9_12_10, \
call_15_9_12_11, \
call_15_9_12_12, \
call_15_9_12_13, \
call_15_9_12_14, \
call_15_9_12_15, \
call_15_9_13_0, \
call_15_9_13_1, \
call_15_9_13_2, \
call_15_9_13_3, \
call_15_9_13_4, \
call_15_9_13_5, \
call_15_9_13_6, \
call_15_9_13_7, \
call_15_9_13_8, \
call_15_9_13_9, \
call_15_9_13_10, \
call_15_9_13_11, \
call_15_9_13_12, \
call_15_9_13_13, \
call_15_9_13_14, \
call_15_9_13_15, \
call_15_9_14_0, \
call_15_9_14_1, \
call_15_9_14_2, \
call_15_9_14_3, \
call_15_9_14_4, \
call_15_9_14_5, \
call_15_9_14_6, \
call_15_9_14_7, \
call_15_9_14_8, \
call_15_9_14_9, \
call_15_9_14_10, \
call_15_9_14_11, \
call_15_9_14_12, \
call_15_9_14_13, \
call_15_9_14_14, \
call_15_9_14_15, \
call_15_9_15_0, \
call_15_9_15_1, \
call_15_9_15_2, \
call_15_9_15_3, \
call_15_9_15_4, \
call_15_9_15_5, \
call_15_9_15_6, \
call_15_9_15_7, \
call_15_9_15_8, \
call_15_9_15_9, \
call_15_9_15_10, \
call_15_9_15_11, \
call_15_9_15_12, \
call_15_9_15_13, \
call_15_9_15_14, \
call_15_9_15_15, \
call_15_10_0_0, \
call_15_10_0_1, \
call_15_10_0_2, \
call_15_10_0_3, \
call_15_10_0_4, \
call_15_10_0_5, \
call_15_10_0_6, \
call_15_10_0_7, \
call_15_10_0_8, \
call_15_10_0_9, \
call_15_10_0_10, \
call_15_10_0_11, \
call_15_10_0_12, \
call_15_10_0_13, \
call_15_10_0_14, \
call_15_10_0_15, \
call_15_10_1_0, \
call_15_10_1_1, \
call_15_10_1_2, \
call_15_10_1_3, \
call_15_10_1_4, \
call_15_10_1_5, \
call_15_10_1_6, \
call_15_10_1_7, \
call_15_10_1_8, \
call_15_10_1_9, \
call_15_10_1_10, \
call_15_10_1_11, \
call_15_10_1_12, \
call_15_10_1_13, \
call_15_10_1_14, \
call_15_10_1_15, \
call_15_10_2_0, \
call_15_10_2_1, \
call_15_10_2_2, \
call_15_10_2_3, \
call_15_10_2_4, \
call_15_10_2_5, \
call_15_10_2_6, \
call_15_10_2_7, \
call_15_10_2_8, \
call_15_10_2_9, \
call_15_10_2_10, \
call_15_10_2_11, \
call_15_10_2_12, \
call_15_10_2_13, \
call_15_10_2_14, \
call_15_10_2_15, \
call_15_10_3_0, \
call_15_10_3_1, \
call_15_10_3_2, \
call_15_10_3_3, \
call_15_10_3_4, \
call_15_10_3_5, \
call_15_10_3_6, \
call_15_10_3_7, \
call_15_10_3_8, \
call_15_10_3_9, \
call_15_10_3_10, \
call_15_10_3_11, \
call_15_10_3_12, \
call_15_10_3_13, \
call_15_10_3_14, \
call_15_10_3_15, \
call_15_10_4_0, \
call_15_10_4_1, \
call_15_10_4_2, \
call_15_10_4_3, \
call_15_10_4_4, \
call_15_10_4_5, \
call_15_10_4_6, \
call_15_10_4_7, \
call_15_10_4_8, \
call_15_10_4_9, \
call_15_10_4_10, \
call_15_10_4_11, \
call_15_10_4_12, \
call_15_10_4_13, \
call_15_10_4_14, \
call_15_10_4_15, \
call_15_10_5_0, \
call_15_10_5_1, \
call_15_10_5_2, \
call_15_10_5_3, \
call_15_10_5_4, \
call_15_10_5_5, \
call_15_10_5_6, \
call_15_10_5_7, \
call_15_10_5_8, \
call_15_10_5_9, \
call_15_10_5_10, \
call_15_10_5_11, \
call_15_10_5_12, \
call_15_10_5_13, \
call_15_10_5_14, \
call_15_10_5_15, \
call_15_10_6_0, \
call_15_10_6_1, \
call_15_10_6_2, \
call_15_10_6_3, \
call_15_10_6_4, \
call_15_10_6_5, \
call_15_10_6_6, \
call_15_10_6_7, \
call_15_10_6_8, \
call_15_10_6_9, \
call_15_10_6_10, \
call_15_10_6_11, \
call_15_10_6_12, \
call_15_10_6_13, \
call_15_10_6_14, \
call_15_10_6_15, \
call_15_10_7_0, \
call_15_10_7_1, \
call_15_10_7_2, \
call_15_10_7_3, \
call_15_10_7_4, \
call_15_10_7_5, \
call_15_10_7_6, \
call_15_10_7_7, \
call_15_10_7_8, \
call_15_10_7_9, \
call_15_10_7_10, \
call_15_10_7_11, \
call_15_10_7_12, \
call_15_10_7_13, \
call_15_10_7_14, \
call_15_10_7_15, \
call_15_10_8_0, \
call_15_10_8_1, \
call_15_10_8_2, \
call_15_10_8_3, \
call_15_10_8_4, \
call_15_10_8_5, \
call_15_10_8_6, \
call_15_10_8_7, \
call_15_10_8_8, \
call_15_10_8_9, \
call_15_10_8_10, \
call_15_10_8_11, \
call_15_10_8_12, \
call_15_10_8_13, \
call_15_10_8_14, \
call_15_10_8_15, \
call_15_10_9_0, \
call_15_10_9_1, \
call_15_10_9_2, \
call_15_10_9_3, \
call_15_10_9_4, \
call_15_10_9_5, \
call_15_10_9_6, \
call_15_10_9_7, \
call_15_10_9_8, \
call_15_10_9_9, \
call_15_10_9_10, \
call_15_10_9_11, \
call_15_10_9_12, \
call_15_10_9_13, \
call_15_10_9_14, \
call_15_10_9_15, \
call_15_10_10_0, \
call_15_10_10_1, \
call_15_10_10_2, \
call_15_10_10_3, \
call_15_10_10_4, \
call_15_10_10_5, \
call_15_10_10_6, \
call_15_10_10_7, \
call_15_10_10_8, \
call_15_10_10_9, \
call_15_10_10_10, \
call_15_10_10_11, \
call_15_10_10_12, \
call_15_10_10_13, \
call_15_10_10_14, \
call_15_10_10_15, \
call_15_10_11_0, \
call_15_10_11_1, \
call_15_10_11_2, \
call_15_10_11_3, \
call_15_10_11_4, \
call_15_10_11_5, \
call_15_10_11_6, \
call_15_10_11_7, \
call_15_10_11_8, \
call_15_10_11_9, \
call_15_10_11_10, \
call_15_10_11_11, \
call_15_10_11_12, \
call_15_10_11_13, \
call_15_10_11_14, \
call_15_10_11_15, \
call_15_10_12_0, \
call_15_10_12_1, \
call_15_10_12_2, \
call_15_10_12_3, \
call_15_10_12_4, \
call_15_10_12_5, \
call_15_10_12_6, \
call_15_10_12_7, \
call_15_10_12_8, \
call_15_10_12_9, \
call_15_10_12_10, \
call_15_10_12_11, \
call_15_10_12_12, \
call_15_10_12_13, \
call_15_10_12_14, \
call_15_10_12_15, \
call_15_10_13_0, \
call_15_10_13_1, \
call_15_10_13_2, \
call_15_10_13_3, \
call_15_10_13_4, \
call_15_10_13_5, \
call_15_10_13_6, \
call_15_10_13_7, \
call_15_10_13_8, \
call_15_10_13_9, \
call_15_10_13_10, \
call_15_10_13_11, \
call_15_10_13_12, \
call_15_10_13_13, \
call_15_10_13_14, \
call_15_10_13_15, \
call_15_10_14_0, \
call_15_10_14_1, \
call_15_10_14_2, \
call_15_10_14_3, \
call_15_10_14_4, \
call_15_10_14_5, \
call_15_10_14_6, \
call_15_10_14_7, \
call_15_10_14_8, \
call_15_10_14_9, \
call_15_10_14_10, \
call_15_10_14_11, \
call_15_10_14_12, \
call_15_10_14_13, \
call_15_10_14_14, \
call_15_10_14_15, \
call_15_10_15_0, \
call_15_10_15_1, \
call_15_10_15_2, \
call_15_10_15_3, \
call_15_10_15_4, \
call_15_10_15_5, \
call_15_10_15_6, \
call_15_10_15_7, \
call_15_10_15_8, \
call_15_10_15_9, \
call_15_10_15_10, \
call_15_10_15_11, \
call_15_10_15_12, \
call_15_10_15_13, \
call_15_10_15_14, \
call_15_10_15_15, \
call_15_11_0_0, \
call_15_11_0_1, \
call_15_11_0_2, \
call_15_11_0_3, \
call_15_11_0_4, \
call_15_11_0_5, \
call_15_11_0_6, \
call_15_11_0_7, \
call_15_11_0_8, \
call_15_11_0_9, \
call_15_11_0_10, \
call_15_11_0_11, \
call_15_11_0_12, \
call_15_11_0_13, \
call_15_11_0_14, \
call_15_11_0_15, \
call_15_11_1_0, \
call_15_11_1_1, \
call_15_11_1_2, \
call_15_11_1_3, \
call_15_11_1_4, \
call_15_11_1_5, \
call_15_11_1_6, \
call_15_11_1_7, \
call_15_11_1_8, \
call_15_11_1_9, \
call_15_11_1_10, \
call_15_11_1_11, \
call_15_11_1_12, \
call_15_11_1_13, \
call_15_11_1_14, \
call_15_11_1_15, \
call_15_11_2_0, \
call_15_11_2_1, \
call_15_11_2_2, \
call_15_11_2_3, \
call_15_11_2_4, \
call_15_11_2_5, \
call_15_11_2_6, \
call_15_11_2_7, \
call_15_11_2_8, \
call_15_11_2_9, \
call_15_11_2_10, \
call_15_11_2_11, \
call_15_11_2_12, \
call_15_11_2_13, \
call_15_11_2_14, \
call_15_11_2_15, \
call_15_11_3_0, \
call_15_11_3_1, \
call_15_11_3_2, \
call_15_11_3_3, \
call_15_11_3_4, \
call_15_11_3_5, \
call_15_11_3_6, \
call_15_11_3_7, \
call_15_11_3_8, \
call_15_11_3_9, \
call_15_11_3_10, \
call_15_11_3_11, \
call_15_11_3_12, \
call_15_11_3_13, \
call_15_11_3_14, \
call_15_11_3_15, \
call_15_11_4_0, \
call_15_11_4_1, \
call_15_11_4_2, \
call_15_11_4_3, \
call_15_11_4_4, \
call_15_11_4_5, \
call_15_11_4_6, \
call_15_11_4_7, \
call_15_11_4_8, \
call_15_11_4_9, \
call_15_11_4_10, \
call_15_11_4_11, \
call_15_11_4_12, \
call_15_11_4_13, \
call_15_11_4_14, \
call_15_11_4_15, \
call_15_11_5_0, \
call_15_11_5_1, \
call_15_11_5_2, \
call_15_11_5_3, \
call_15_11_5_4, \
call_15_11_5_5, \
call_15_11_5_6, \
call_15_11_5_7, \
call_15_11_5_8, \
call_15_11_5_9, \
call_15_11_5_10, \
call_15_11_5_11, \
call_15_11_5_12, \
call_15_11_5_13, \
call_15_11_5_14, \
call_15_11_5_15, \
call_15_11_6_0, \
call_15_11_6_1, \
call_15_11_6_2, \
call_15_11_6_3, \
call_15_11_6_4, \
call_15_11_6_5, \
call_15_11_6_6, \
call_15_11_6_7, \
call_15_11_6_8, \
call_15_11_6_9, \
call_15_11_6_10, \
call_15_11_6_11, \
call_15_11_6_12, \
call_15_11_6_13, \
call_15_11_6_14, \
call_15_11_6_15, \
call_15_11_7_0, \
call_15_11_7_1, \
call_15_11_7_2, \
call_15_11_7_3, \
call_15_11_7_4, \
call_15_11_7_5, \
call_15_11_7_6, \
call_15_11_7_7, \
call_15_11_7_8, \
call_15_11_7_9, \
call_15_11_7_10, \
call_15_11_7_11, \
call_15_11_7_12, \
call_15_11_7_13, \
call_15_11_7_14, \
call_15_11_7_15, \
call_15_11_8_0, \
call_15_11_8_1, \
call_15_11_8_2, \
call_15_11_8_3, \
call_15_11_8_4, \
call_15_11_8_5, \
call_15_11_8_6, \
call_15_11_8_7, \
call_15_11_8_8, \
call_15_11_8_9, \
call_15_11_8_10, \
call_15_11_8_11, \
call_15_11_8_12, \
call_15_11_8_13, \
call_15_11_8_14, \
call_15_11_8_15, \
call_15_11_9_0, \
call_15_11_9_1, \
call_15_11_9_2, \
call_15_11_9_3, \
call_15_11_9_4, \
call_15_11_9_5, \
call_15_11_9_6, \
call_15_11_9_7, \
call_15_11_9_8, \
call_15_11_9_9, \
call_15_11_9_10, \
call_15_11_9_11, \
call_15_11_9_12, \
call_15_11_9_13, \
call_15_11_9_14, \
call_15_11_9_15, \
call_15_11_10_0, \
call_15_11_10_1, \
call_15_11_10_2, \
call_15_11_10_3, \
call_15_11_10_4, \
call_15_11_10_5, \
call_15_11_10_6, \
call_15_11_10_7, \
call_15_11_10_8, \
call_15_11_10_9, \
call_15_11_10_10, \
call_15_11_10_11, \
call_15_11_10_12, \
call_15_11_10_13, \
call_15_11_10_14, \
call_15_11_10_15, \
call_15_11_11_0, \
call_15_11_11_1, \
call_15_11_11_2, \
call_15_11_11_3, \
call_15_11_11_4, \
call_15_11_11_5, \
call_15_11_11_6, \
call_15_11_11_7, \
call_15_11_11_8, \
call_15_11_11_9, \
call_15_11_11_10, \
call_15_11_11_11, \
call_15_11_11_12, \
call_15_11_11_13, \
call_15_11_11_14, \
call_15_11_11_15, \
call_15_11_12_0, \
call_15_11_12_1, \
call_15_11_12_2, \
call_15_11_12_3, \
call_15_11_12_4, \
call_15_11_12_5, \
call_15_11_12_6, \
call_15_11_12_7, \
call_15_11_12_8, \
call_15_11_12_9, \
call_15_11_12_10, \
call_15_11_12_11, \
call_15_11_12_12, \
call_15_11_12_13, \
call_15_11_12_14, \
call_15_11_12_15, \
call_15_11_13_0, \
call_15_11_13_1, \
call_15_11_13_2, \
call_15_11_13_3, \
call_15_11_13_4, \
call_15_11_13_5, \
call_15_11_13_6, \
call_15_11_13_7, \
call_15_11_13_8, \
call_15_11_13_9, \
call_15_11_13_10, \
call_15_11_13_11, \
call_15_11_13_12, \
call_15_11_13_13, \
call_15_11_13_14, \
call_15_11_13_15, \
call_15_11_14_0, \
call_15_11_14_1, \
call_15_11_14_2, \
call_15_11_14_3, \
call_15_11_14_4, \
call_15_11_14_5, \
call_15_11_14_6, \
call_15_11_14_7, \
call_15_11_14_8, \
call_15_11_14_9, \
call_15_11_14_10, \
call_15_11_14_11, \
call_15_11_14_12, \
call_15_11_14_13, \
call_15_11_14_14, \
call_15_11_14_15, \
call_15_11_15_0, \
call_15_11_15_1, \
call_15_11_15_2, \
call_15_11_15_3, \
call_15_11_15_4, \
call_15_11_15_5, \
call_15_11_15_6, \
call_15_11_15_7, \
call_15_11_15_8, \
call_15_11_15_9, \
call_15_11_15_10, \
call_15_11_15_11, \
call_15_11_15_12, \
call_15_11_15_13, \
call_15_11_15_14, \
call_15_11_15_15, \
call_15_12_0_0, \
call_15_12_0_1, \
call_15_12_0_2, \
call_15_12_0_3, \
call_15_12_0_4, \
call_15_12_0_5, \
call_15_12_0_6, \
call_15_12_0_7, \
call_15_12_0_8, \
call_15_12_0_9, \
call_15_12_0_10, \
call_15_12_0_11, \
call_15_12_0_12, \
call_15_12_0_13, \
call_15_12_0_14, \
call_15_12_0_15, \
call_15_12_1_0, \
call_15_12_1_1, \
call_15_12_1_2, \
call_15_12_1_3, \
call_15_12_1_4, \
call_15_12_1_5, \
call_15_12_1_6, \
call_15_12_1_7, \
call_15_12_1_8, \
call_15_12_1_9, \
call_15_12_1_10, \
call_15_12_1_11, \
call_15_12_1_12, \
call_15_12_1_13, \
call_15_12_1_14, \
call_15_12_1_15, \
call_15_12_2_0, \
call_15_12_2_1, \
call_15_12_2_2, \
call_15_12_2_3, \
call_15_12_2_4, \
call_15_12_2_5, \
call_15_12_2_6, \
call_15_12_2_7, \
call_15_12_2_8, \
call_15_12_2_9, \
call_15_12_2_10, \
call_15_12_2_11, \
call_15_12_2_12, \
call_15_12_2_13, \
call_15_12_2_14, \
call_15_12_2_15, \
call_15_12_3_0, \
call_15_12_3_1, \
call_15_12_3_2, \
call_15_12_3_3, \
call_15_12_3_4, \
call_15_12_3_5, \
call_15_12_3_6, \
call_15_12_3_7, \
call_15_12_3_8, \
call_15_12_3_9, \
call_15_12_3_10, \
call_15_12_3_11, \
call_15_12_3_12, \
call_15_12_3_13, \
call_15_12_3_14, \
call_15_12_3_15, \
call_15_12_4_0, \
call_15_12_4_1, \
call_15_12_4_2, \
call_15_12_4_3, \
call_15_12_4_4, \
call_15_12_4_5, \
call_15_12_4_6, \
call_15_12_4_7, \
call_15_12_4_8, \
call_15_12_4_9, \
call_15_12_4_10, \
call_15_12_4_11, \
call_15_12_4_12, \
call_15_12_4_13, \
call_15_12_4_14, \
call_15_12_4_15, \
call_15_12_5_0, \
call_15_12_5_1, \
call_15_12_5_2, \
call_15_12_5_3, \
call_15_12_5_4, \
call_15_12_5_5, \
call_15_12_5_6, \
call_15_12_5_7, \
call_15_12_5_8, \
call_15_12_5_9, \
call_15_12_5_10, \
call_15_12_5_11, \
call_15_12_5_12, \
call_15_12_5_13, \
call_15_12_5_14, \
call_15_12_5_15, \
call_15_12_6_0, \
call_15_12_6_1, \
call_15_12_6_2, \
call_15_12_6_3, \
call_15_12_6_4, \
call_15_12_6_5, \
call_15_12_6_6, \
call_15_12_6_7, \
call_15_12_6_8, \
call_15_12_6_9, \
call_15_12_6_10, \
call_15_12_6_11, \
call_15_12_6_12, \
call_15_12_6_13, \
call_15_12_6_14, \
call_15_12_6_15, \
call_15_12_7_0, \
call_15_12_7_1, \
call_15_12_7_2, \
call_15_12_7_3, \
call_15_12_7_4, \
call_15_12_7_5, \
call_15_12_7_6, \
call_15_12_7_7, \
call_15_12_7_8, \
call_15_12_7_9, \
call_15_12_7_10, \
call_15_12_7_11, \
call_15_12_7_12, \
call_15_12_7_13, \
call_15_12_7_14, \
call_15_12_7_15, \
call_15_12_8_0, \
call_15_12_8_1, \
call_15_12_8_2, \
call_15_12_8_3, \
call_15_12_8_4, \
call_15_12_8_5, \
call_15_12_8_6, \
call_15_12_8_7, \
call_15_12_8_8, \
call_15_12_8_9, \
call_15_12_8_10, \
call_15_12_8_11, \
call_15_12_8_12, \
call_15_12_8_13, \
call_15_12_8_14, \
call_15_12_8_15, \
call_15_12_9_0, \
call_15_12_9_1, \
call_15_12_9_2, \
call_15_12_9_3, \
call_15_12_9_4, \
call_15_12_9_5, \
call_15_12_9_6, \
call_15_12_9_7, \
call_15_12_9_8, \
call_15_12_9_9, \
call_15_12_9_10, \
call_15_12_9_11, \
call_15_12_9_12, \
call_15_12_9_13, \
call_15_12_9_14, \
call_15_12_9_15, \
call_15_12_10_0, \
call_15_12_10_1, \
call_15_12_10_2, \
call_15_12_10_3, \
call_15_12_10_4, \
call_15_12_10_5, \
call_15_12_10_6, \
call_15_12_10_7, \
call_15_12_10_8, \
call_15_12_10_9, \
call_15_12_10_10, \
call_15_12_10_11, \
call_15_12_10_12, \
call_15_12_10_13, \
call_15_12_10_14, \
call_15_12_10_15, \
call_15_12_11_0, \
call_15_12_11_1, \
call_15_12_11_2, \
call_15_12_11_3, \
call_15_12_11_4, \
call_15_12_11_5, \
call_15_12_11_6, \
call_15_12_11_7, \
call_15_12_11_8, \
call_15_12_11_9, \
call_15_12_11_10, \
call_15_12_11_11, \
call_15_12_11_12, \
call_15_12_11_13, \
call_15_12_11_14, \
call_15_12_11_15, \
call_15_12_12_0, \
call_15_12_12_1, \
call_15_12_12_2, \
call_15_12_12_3, \
call_15_12_12_4, \
call_15_12_12_5, \
call_15_12_12_6, \
call_15_12_12_7, \
call_15_12_12_8, \
call_15_12_12_9, \
call_15_12_12_10, \
call_15_12_12_11, \
call_15_12_12_12, \
call_15_12_12_13, \
call_15_12_12_14, \
call_15_12_12_15, \
call_15_12_13_0, \
call_15_12_13_1, \
call_15_12_13_2, \
call_15_12_13_3, \
call_15_12_13_4, \
call_15_12_13_5, \
call_15_12_13_6, \
call_15_12_13_7, \
call_15_12_13_8, \
call_15_12_13_9, \
call_15_12_13_10, \
call_15_12_13_11, \
call_15_12_13_12, \
call_15_12_13_13, \
call_15_12_13_14, \
call_15_12_13_15, \
call_15_12_14_0, \
call_15_12_14_1, \
call_15_12_14_2, \
call_15_12_14_3, \
call_15_12_14_4, \
call_15_12_14_5, \
call_15_12_14_6, \
call_15_12_14_7, \
call_15_12_14_8, \
call_15_12_14_9, \
call_15_12_14_10, \
call_15_12_14_11, \
call_15_12_14_12, \
call_15_12_14_13, \
call_15_12_14_14, \
call_15_12_14_15, \
call_15_12_15_0, \
call_15_12_15_1, \
call_15_12_15_2, \
call_15_12_15_3, \
call_15_12_15_4, \
call_15_12_15_5, \
call_15_12_15_6, \
call_15_12_15_7, \
call_15_12_15_8, \
call_15_12_15_9, \
call_15_12_15_10, \
call_15_12_15_11, \
call_15_12_15_12, \
call_15_12_15_13, \
call_15_12_15_14, \
call_15_12_15_15, \
call_15_13_0_0, \
call_15_13_0_1, \
call_15_13_0_2, \
call_15_13_0_3, \
call_15_13_0_4, \
call_15_13_0_5, \
call_15_13_0_6, \
call_15_13_0_7, \
call_15_13_0_8, \
call_15_13_0_9, \
call_15_13_0_10, \
call_15_13_0_11, \
call_15_13_0_12, \
call_15_13_0_13, \
call_15_13_0_14, \
call_15_13_0_15, \
call_15_13_1_0, \
call_15_13_1_1, \
call_15_13_1_2, \
call_15_13_1_3, \
call_15_13_1_4, \
call_15_13_1_5, \
call_15_13_1_6, \
call_15_13_1_7, \
call_15_13_1_8, \
call_15_13_1_9, \
call_15_13_1_10, \
call_15_13_1_11, \
call_15_13_1_12, \
call_15_13_1_13, \
call_15_13_1_14, \
call_15_13_1_15, \
call_15_13_2_0, \
call_15_13_2_1, \
call_15_13_2_2, \
call_15_13_2_3, \
call_15_13_2_4, \
call_15_13_2_5, \
call_15_13_2_6, \
call_15_13_2_7, \
call_15_13_2_8, \
call_15_13_2_9, \
call_15_13_2_10, \
call_15_13_2_11, \
call_15_13_2_12, \
call_15_13_2_13, \
call_15_13_2_14, \
call_15_13_2_15, \
call_15_13_3_0, \
call_15_13_3_1, \
call_15_13_3_2, \
call_15_13_3_3, \
call_15_13_3_4, \
call_15_13_3_5, \
call_15_13_3_6, \
call_15_13_3_7, \
call_15_13_3_8, \
call_15_13_3_9, \
call_15_13_3_10, \
call_15_13_3_11, \
call_15_13_3_12, \
call_15_13_3_13, \
call_15_13_3_14, \
call_15_13_3_15, \
call_15_13_4_0, \
call_15_13_4_1, \
call_15_13_4_2, \
call_15_13_4_3, \
call_15_13_4_4, \
call_15_13_4_5, \
call_15_13_4_6, \
call_15_13_4_7, \
call_15_13_4_8, \
call_15_13_4_9, \
call_15_13_4_10, \
call_15_13_4_11, \
call_15_13_4_12, \
call_15_13_4_13, \
call_15_13_4_14, \
call_15_13_4_15, \
call_15_13_5_0, \
call_15_13_5_1, \
call_15_13_5_2, \
call_15_13_5_3, \
call_15_13_5_4, \
call_15_13_5_5, \
call_15_13_5_6, \
call_15_13_5_7, \
call_15_13_5_8, \
call_15_13_5_9, \
call_15_13_5_10, \
call_15_13_5_11, \
call_15_13_5_12, \
call_15_13_5_13, \
call_15_13_5_14, \
call_15_13_5_15, \
call_15_13_6_0, \
call_15_13_6_1, \
call_15_13_6_2, \
call_15_13_6_3, \
call_15_13_6_4, \
call_15_13_6_5, \
call_15_13_6_6, \
call_15_13_6_7, \
call_15_13_6_8, \
call_15_13_6_9, \
call_15_13_6_10, \
call_15_13_6_11, \
call_15_13_6_12, \
call_15_13_6_13, \
call_15_13_6_14, \
call_15_13_6_15, \
call_15_13_7_0, \
call_15_13_7_1, \
call_15_13_7_2, \
call_15_13_7_3, \
call_15_13_7_4, \
call_15_13_7_5, \
call_15_13_7_6, \
call_15_13_7_7, \
call_15_13_7_8, \
call_15_13_7_9, \
call_15_13_7_10, \
call_15_13_7_11, \
call_15_13_7_12, \
call_15_13_7_13, \
call_15_13_7_14, \
call_15_13_7_15, \
call_15_13_8_0, \
call_15_13_8_1, \
call_15_13_8_2, \
call_15_13_8_3, \
call_15_13_8_4, \
call_15_13_8_5, \
call_15_13_8_6, \
call_15_13_8_7, \
call_15_13_8_8, \
call_15_13_8_9, \
call_15_13_8_10, \
call_15_13_8_11, \
call_15_13_8_12, \
call_15_13_8_13, \
call_15_13_8_14, \
call_15_13_8_15, \
call_15_13_9_0, \
call_15_13_9_1, \
call_15_13_9_2, \
call_15_13_9_3, \
call_15_13_9_4, \
call_15_13_9_5, \
call_15_13_9_6, \
call_15_13_9_7, \
call_15_13_9_8, \
call_15_13_9_9, \
call_15_13_9_10, \
call_15_13_9_11, \
call_15_13_9_12, \
call_15_13_9_13, \
call_15_13_9_14, \
call_15_13_9_15, \
call_15_13_10_0, \
call_15_13_10_1, \
call_15_13_10_2, \
call_15_13_10_3, \
call_15_13_10_4, \
call_15_13_10_5, \
call_15_13_10_6, \
call_15_13_10_7, \
call_15_13_10_8, \
call_15_13_10_9, \
call_15_13_10_10, \
call_15_13_10_11, \
call_15_13_10_12, \
call_15_13_10_13, \
call_15_13_10_14, \
call_15_13_10_15, \
call_15_13_11_0, \
call_15_13_11_1, \
call_15_13_11_2, \
call_15_13_11_3, \
call_15_13_11_4, \
call_15_13_11_5, \
call_15_13_11_6, \
call_15_13_11_7, \
call_15_13_11_8, \
call_15_13_11_9, \
call_15_13_11_10, \
call_15_13_11_11, \
call_15_13_11_12, \
call_15_13_11_13, \
call_15_13_11_14, \
call_15_13_11_15, \
call_15_13_12_0, \
call_15_13_12_1, \
call_15_13_12_2, \
call_15_13_12_3, \
call_15_13_12_4, \
call_15_13_12_5, \
call_15_13_12_6, \
call_15_13_12_7, \
call_15_13_12_8, \
call_15_13_12_9, \
call_15_13_12_10, \
call_15_13_12_11, \
call_15_13_12_12, \
call_15_13_12_13, \
call_15_13_12_14, \
call_15_13_12_15, \
call_15_13_13_0, \
call_15_13_13_1, \
call_15_13_13_2, \
call_15_13_13_3, \
call_15_13_13_4, \
call_15_13_13_5, \
call_15_13_13_6, \
call_15_13_13_7, \
call_15_13_13_8, \
call_15_13_13_9, \
call_15_13_13_10, \
call_15_13_13_11, \
call_15_13_13_12, \
call_15_13_13_13, \
call_15_13_13_14, \
call_15_13_13_15, \
call_15_13_14_0, \
call_15_13_14_1, \
call_15_13_14_2, \
call_15_13_14_3, \
call_15_13_14_4, \
call_15_13_14_5, \
call_15_13_14_6, \
call_15_13_14_7, \
call_15_13_14_8, \
call_15_13_14_9, \
call_15_13_14_10, \
call_15_13_14_11, \
call_15_13_14_12, \
call_15_13_14_13, \
call_15_13_14_14, \
call_15_13_14_15, \
call_15_13_15_0, \
call_15_13_15_1, \
call_15_13_15_2, \
call_15_13_15_3, \
call_15_13_15_4, \
call_15_13_15_5, \
call_15_13_15_6, \
call_15_13_15_7, \
call_15_13_15_8, \
call_15_13_15_9, \
call_15_13_15_10, \
call_15_13_15_11, \
call_15_13_15_12, \
call_15_13_15_13, \
call_15_13_15_14, \
call_15_13_15_15, \
call_15_14_0_0, \
call_15_14_0_1, \
call_15_14_0_2, \
call_15_14_0_3, \
call_15_14_0_4, \
call_15_14_0_5, \
call_15_14_0_6, \
call_15_14_0_7, \
call_15_14_0_8, \
call_15_14_0_9, \
call_15_14_0_10, \
call_15_14_0_11, \
call_15_14_0_12, \
call_15_14_0_13, \
call_15_14_0_14, \
call_15_14_0_15, \
call_15_14_1_0, \
call_15_14_1_1, \
call_15_14_1_2, \
call_15_14_1_3, \
call_15_14_1_4, \
call_15_14_1_5, \
call_15_14_1_6, \
call_15_14_1_7, \
call_15_14_1_8, \
call_15_14_1_9, \
call_15_14_1_10, \
call_15_14_1_11, \
call_15_14_1_12, \
call_15_14_1_13, \
call_15_14_1_14, \
call_15_14_1_15, \
call_15_14_2_0, \
call_15_14_2_1, \
call_15_14_2_2, \
call_15_14_2_3, \
call_15_14_2_4, \
call_15_14_2_5, \
call_15_14_2_6, \
call_15_14_2_7, \
call_15_14_2_8, \
call_15_14_2_9, \
call_15_14_2_10, \
call_15_14_2_11, \
call_15_14_2_12, \
call_15_14_2_13, \
call_15_14_2_14, \
call_15_14_2_15, \
call_15_14_3_0, \
call_15_14_3_1, \
call_15_14_3_2, \
call_15_14_3_3, \
call_15_14_3_4, \
call_15_14_3_5, \
call_15_14_3_6, \
call_15_14_3_7, \
call_15_14_3_8, \
call_15_14_3_9, \
call_15_14_3_10, \
call_15_14_3_11, \
call_15_14_3_12, \
call_15_14_3_13, \
call_15_14_3_14, \
call_15_14_3_15, \
call_15_14_4_0, \
call_15_14_4_1, \
call_15_14_4_2, \
call_15_14_4_3, \
call_15_14_4_4, \
call_15_14_4_5, \
call_15_14_4_6, \
call_15_14_4_7, \
call_15_14_4_8, \
call_15_14_4_9, \
call_15_14_4_10, \
call_15_14_4_11, \
call_15_14_4_12, \
call_15_14_4_13, \
call_15_14_4_14, \
call_15_14_4_15, \
call_15_14_5_0, \
call_15_14_5_1, \
call_15_14_5_2, \
call_15_14_5_3, \
call_15_14_5_4, \
call_15_14_5_5, \
call_15_14_5_6, \
call_15_14_5_7, \
call_15_14_5_8, \
call_15_14_5_9, \
call_15_14_5_10, \
call_15_14_5_11, \
call_15_14_5_12, \
call_15_14_5_13, \
call_15_14_5_14, \
call_15_14_5_15, \
call_15_14_6_0, \
call_15_14_6_1, \
call_15_14_6_2, \
call_15_14_6_3, \
call_15_14_6_4, \
call_15_14_6_5, \
call_15_14_6_6, \
call_15_14_6_7, \
call_15_14_6_8, \
call_15_14_6_9, \
call_15_14_6_10, \
call_15_14_6_11, \
call_15_14_6_12, \
call_15_14_6_13, \
call_15_14_6_14, \
call_15_14_6_15, \
call_15_14_7_0, \
call_15_14_7_1, \
call_15_14_7_2, \
call_15_14_7_3, \
call_15_14_7_4, \
call_15_14_7_5, \
call_15_14_7_6, \
call_15_14_7_7, \
call_15_14_7_8, \
call_15_14_7_9, \
call_15_14_7_10, \
call_15_14_7_11, \
call_15_14_7_12, \
call_15_14_7_13, \
call_15_14_7_14, \
call_15_14_7_15, \
call_15_14_8_0, \
call_15_14_8_1, \
call_15_14_8_2, \
call_15_14_8_3, \
call_15_14_8_4, \
call_15_14_8_5, \
call_15_14_8_6, \
call_15_14_8_7, \
call_15_14_8_8, \
call_15_14_8_9, \
call_15_14_8_10, \
call_15_14_8_11, \
call_15_14_8_12, \
call_15_14_8_13, \
call_15_14_8_14, \
call_15_14_8_15, \
call_15_14_9_0, \
call_15_14_9_1, \
call_15_14_9_2, \
call_15_14_9_3, \
call_15_14_9_4, \
call_15_14_9_5, \
call_15_14_9_6, \
call_15_14_9_7, \
call_15_14_9_8, \
call_15_14_9_9, \
call_15_14_9_10, \
call_15_14_9_11, \
call_15_14_9_12, \
call_15_14_9_13, \
call_15_14_9_14, \
call_15_14_9_15, \
call_15_14_10_0, \
call_15_14_10_1, \
call_15_14_10_2, \
call_15_14_10_3, \
call_15_14_10_4, \
call_15_14_10_5, \
call_15_14_10_6, \
call_15_14_10_7, \
call_15_14_10_8, \
call_15_14_10_9, \
call_15_14_10_10, \
call_15_14_10_11, \
call_15_14_10_12, \
call_15_14_10_13, \
call_15_14_10_14, \
call_15_14_10_15, \
call_15_14_11_0, \
call_15_14_11_1, \
call_15_14_11_2, \
call_15_14_11_3, \
call_15_14_11_4, \
call_15_14_11_5, \
call_15_14_11_6, \
call_15_14_11_7, \
call_15_14_11_8, \
call_15_14_11_9, \
call_15_14_11_10, \
call_15_14_11_11, \
call_15_14_11_12, \
call_15_14_11_13, \
call_15_14_11_14, \
call_15_14_11_15, \
call_15_14_12_0, \
call_15_14_12_1, \
call_15_14_12_2, \
call_15_14_12_3, \
call_15_14_12_4, \
call_15_14_12_5, \
call_15_14_12_6, \
call_15_14_12_7, \
call_15_14_12_8, \
call_15_14_12_9, \
call_15_14_12_10, \
call_15_14_12_11, \
call_15_14_12_12, \
call_15_14_12_13, \
call_15_14_12_14, \
call_15_14_12_15, \
call_15_14_13_0, \
call_15_14_13_1, \
call_15_14_13_2, \
call_15_14_13_3, \
call_15_14_13_4, \
call_15_14_13_5, \
call_15_14_13_6, \
call_15_14_13_7, \
call_15_14_13_8, \
call_15_14_13_9, \
call_15_14_13_10, \
call_15_14_13_11, \
call_15_14_13_12, \
call_15_14_13_13, \
call_15_14_13_14, \
call_15_14_13_15, \
call_15_14_14_0, \
call_15_14_14_1, \
call_15_14_14_2, \
call_15_14_14_3, \
call_15_14_14_4, \
call_15_14_14_5, \
call_15_14_14_6, \
call_15_14_14_7, \
call_15_14_14_8, \
call_15_14_14_9, \
call_15_14_14_10, \
call_15_14_14_11, \
call_15_14_14_12, \
call_15_14_14_13, \
call_15_14_14_14, \
call_15_14_14_15, \
call_15_14_15_0, \
call_15_14_15_1, \
call_15_14_15_2, \
call_15_14_15_3, \
call_15_14_15_4, \
call_15_14_15_5, \
call_15_14_15_6, \
call_15_14_15_7, \
call_15_14_15_8, \
call_15_14_15_9, \
call_15_14_15_10, \
call_15_14_15_11, \
call_15_14_15_12, \
call_15_14_15_13, \
call_15_14_15_14, \
call_15_14_15_15, \
call_15_15_0_0, \
call_15_15_0_1, \
call_15_15_0_2, \
call_15_15_0_3, \
call_15_15_0_4, \
call_15_15_0_5, \
call_15_15_0_6, \
call_15_15_0_7, \
call_15_15_0_8, \
call_15_15_0_9, \
call_15_15_0_10, \
call_15_15_0_11, \
call_15_15_0_12, \
call_15_15_0_13, \
call_15_15_0_14, \
call_15_15_0_15, \
call_15_15_1_0, \
call_15_15_1_1, \
call_15_15_1_2, \
call_15_15_1_3, \
call_15_15_1_4, \
call_15_15_1_5, \
call_15_15_1_6, \
call_15_15_1_7, \
call_15_15_1_8, \
call_15_15_1_9, \
call_15_15_1_10, \
call_15_15_1_11, \
call_15_15_1_12, \
call_15_15_1_13, \
call_15_15_1_14, \
call_15_15_1_15, \
call_15_15_2_0, \
call_15_15_2_1, \
call_15_15_2_2, \
call_15_15_2_3, \
call_15_15_2_4, \
call_15_15_2_5, \
call_15_15_2_6, \
call_15_15_2_7, \
call_15_15_2_8, \
call_15_15_2_9, \
call_15_15_2_10, \
call_15_15_2_11, \
call_15_15_2_12, \
call_15_15_2_13, \
call_15_15_2_14, \
call_15_15_2_15, \
call_15_15_3_0, \
call_15_15_3_1, \
call_15_15_3_2, \
call_15_15_3_3, \
call_15_15_3_4, \
call_15_15_3_5, \
call_15_15_3_6, \
call_15_15_3_7, \
call_15_15_3_8, \
call_15_15_3_9, \
call_15_15_3_10, \
call_15_15_3_11, \
call_15_15_3_12, \
call_15_15_3_13, \
call_15_15_3_14, \
call_15_15_3_15, \
call_15_15_4_0, \
call_15_15_4_1, \
call_15_15_4_2, \
call_15_15_4_3, \
call_15_15_4_4, \
call_15_15_4_5, \
call_15_15_4_6, \
call_15_15_4_7, \
call_15_15_4_8, \
call_15_15_4_9, \
call_15_15_4_10, \
call_15_15_4_11, \
call_15_15_4_12, \
call_15_15_4_13, \
call_15_15_4_14, \
call_15_15_4_15, \
call_15_15_5_0, \
call_15_15_5_1, \
call_15_15_5_2, \
call_15_15_5_3, \
call_15_15_5_4, \
call_15_15_5_5, \
call_15_15_5_6, \
call_15_15_5_7, \
call_15_15_5_8, \
call_15_15_5_9, \
call_15_15_5_10, \
call_15_15_5_11, \
call_15_15_5_12, \
call_15_15_5_13, \
call_15_15_5_14, \
call_15_15_5_15, \
call_15_15_6_0, \
call_15_15_6_1, \
call_15_15_6_2, \
call_15_15_6_3, \
call_15_15_6_4, \
call_15_15_6_5, \
call_15_15_6_6, \
call_15_15_6_7, \
call_15_15_6_8, \
call_15_15_6_9, \
call_15_15_6_10, \
call_15_15_6_11, \
call_15_15_6_12, \
call_15_15_6_13, \
call_15_15_6_14, \
call_15_15_6_15, \
call_15_15_7_0, \
call_15_15_7_1, \
call_15_15_7_2, \
call_15_15_7_3, \
call_15_15_7_4, \
call_15_15_7_5, \
call_15_15_7_6, \
call_15_15_7_7, \
call_15_15_7_8, \
call_15_15_7_9, \
call_15_15_7_10, \
call_15_15_7_11, \
call_15_15_7_12, \
call_15_15_7_13, \
call_15_15_7_14, \
call_15_15_7_15, \
call_15_15_8_0, \
call_15_15_8_1, \
call_15_15_8_2, \
call_15_15_8_3, \
call_15_15_8_4, \
call_15_15_8_5, \
call_15_15_8_6, \
call_15_15_8_7, \
call_15_15_8_8, \
call_15_15_8_9, \
call_15_15_8_10, \
call_15_15_8_11, \
call_15_15_8_12, \
call_15_15_8_13, \
call_15_15_8_14, \
call_15_15_8_15, \
call_15_15_9_0, \
call_15_15_9_1, \
call_15_15_9_2, \
call_15_15_9_3, \
call_15_15_9_4, \
call_15_15_9_5, \
call_15_15_9_6, \
call_15_15_9_7, \
call_15_15_9_8, \
call_15_15_9_9, \
call_15_15_9_10, \
call_15_15_9_11, \
call_15_15_9_12, \
call_15_15_9_13, \
call_15_15_9_14, \
call_15_15_9_15, \
call_15_15_10_0, \
call_15_15_10_1, \
call_15_15_10_2, \
call_15_15_10_3, \
call_15_15_10_4, \
call_15_15_10_5, \
call_15_15_10_6, \
call_15_15_10_7, \
call_15_15_10_8, \
call_15_15_10_9, \
call_15_15_10_10, \
call_15_15_10_11, \
call_15_15_10_12, \
call_15_15_10_13, \
call_15_15_10_14, \
call_15_15_10_15, \
call_15_15_11_0, \
call_15_15_11_1, \
call_15_15_11_2, \
call_15_15_11_3, \
call_15_15_11_4, \
call_15_15_11_5, \
call_15_15_11_6, \
call_15_15_11_7, \
call_15_15_11_8, \
call_15_15_11_9, \
call_15_15_11_10, \
call_15_15_11_11, \
call_15_15_11_12, \
call_15_15_11_13, \
call_15_15_11_14, \
call_15_15_11_15, \
call_15_15_12_0, \
call_15_15_12_1, \
call_15_15_12_2, \
call_15_15_12_3, \
call_15_15_12_4, \
call_15_15_12_5, \
call_15_15_12_6, \
call_15_15_12_7, \
call_15_15_12_8, \
call_15_15_12_9, \
call_15_15_12_10, \
call_15_15_12_11, \
call_15_15_12_12, \
call_15_15_12_13, \
call_15_15_12_14, \
call_15_15_12_15, \
call_15_15_13_0, \
call_15_15_13_1, \
call_15_15_13_2, \
call_15_15_13_3, \
call_15_15_13_4, \
call_15_15_13_5, \
call_15_15_13_6, \
call_15_15_13_7, \
call_15_15_13_8, \
call_15_15_13_9, \
call_15_15_13_10, \
call_15_15_13_11, \
call_15_15_13_12, \
call_15_15_13_13, \
call_15_15_13_14, \
call_15_15_13_15, \
call_15_15_14_0, \
call_15_15_14_1, \
call_15_15_14_2, \
call_15_15_14_3, \
call_15_15_14_4, \
call_15_15_14_5, \
call_15_15_14_6, \
call_15_15_14_7, \
call_15_15_14_8, \
call_15_15_14_9, \
call_15_15_14_10, \
call_15_15_14_11, \
call_15_15_14_12, \
call_15_15_14_13, \
call_15_15_14_14, \
call_15_15_14_15, \
call_15_15_15_0, \
call_15_15_15_1, \
call_15_15_15_2, \
call_15_15_15_3, \
call_15_15_15_4, \
call_15_15_15_5, \
call_15_15_15_6, \
call_15_15_15_7, \
call_15_15_15_8, \
call_15_15_15_9, \
call_15_15_15_10, \
call_15_15_15_11, \
call_15_15_15_12, \
call_15_15_15_13, \
call_15_15_15_14, \
call_15_15_15_15, \
};
